"""Data downloaded from NCBI Gene data converted into Python data."""

# Copyright (C) 2014-2018, DV Klopfenstein. All rights reserved

downloaded = "2016_01_05" # 59542 items

import collections as cx

NtData = cx.namedtuple('NtData', 'tax_id Org_name GeneID CurrentID Status Symbol Aliases description other_designations map_location chromosome genomic_nucleotide_accession_version start_position_on_the_genomic_accession end_position_on_the_genomic_accession orientation exon_count OMIM no_hdr0')
GeneID2nt = {
  1 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1, CurrentID=0, Status='live', Symbol='A1BG', Aliases='A1B, ABG, GAB, HYST2477', description='alpha-1-B glycoprotein', other_designations='HEL-S-163pA|epididymis secretory sperm binding protein Li 163pA', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58346806, end_position_on_the_genomic_accession=58353499, orientation='minus', exon_count=8, OMIM=138670, no_hdr0=''),
  2 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2, CurrentID=0, Status='live', Symbol='A2M', Aliases='A2MD, CPAMD5, FWP007, S863-7', description='alpha-2-macroglobulin', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 5|alpha-2-M', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9067708, end_position_on_the_genomic_accession=9115962, orientation='minus', exon_count=36, OMIM=103950, no_hdr0=''),
  3 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3, CurrentID=0, Status='live', Symbol='A2MP1', Aliases='A2MP', description='alpha-2-macroglobulin pseudogene 1', other_designations='pregnancy-zone protein pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9228533, end_position_on_the_genomic_accession=9234207, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  131076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131076, CurrentID=0, Status='live', Symbol='CCDC58', Aliases='', description='coiled-coil domain containing 58', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122359589, end_position_on_the_genomic_accession=122383281, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  645949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645949, CurrentID=0, Status='live', Symbol='LOC645949', Aliases='', description='uncharacterized LOC645949', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21687434, end_position_on_the_genomic_accession=21710652, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  9 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9, CurrentID=0, Status='live', Symbol='NAT1', Aliases='AAC1, MNAT, NAT-1, NATI', description='N-acetyltransferase 1 (arylamine N-acetyltransferase)', other_designations='N-acetyltransferase type 1|arylamide acetylase 1|monomorphic arylamine N-acetyltransferase', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18170462, end_position_on_the_genomic_accession=18223689, orientation='plus', exon_count=11, OMIM=108345, no_hdr0=''),
  10 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10, CurrentID=0, Status='live', Symbol='NAT2', Aliases='AAC2, NAT-2, PNAT', description='N-acetyltransferase 2 (arylamine N-acetyltransferase)', other_designations='N-acetyltransferase type 2|arylamide acetylase 2', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18391245, end_position_on_the_genomic_accession=18401219, orientation='plus', exon_count=3, OMIM=612182, no_hdr0=''),
  11 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11, CurrentID=0, Status='live', Symbol='NATP', Aliases='AACP1, NATP', description='N-acetyltransferase pseudogene', other_designations='arylamide acetylase pseudogene', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18369910, end_position_on_the_genomic_accession=18371878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  12 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=12, CurrentID=0, Status='live', Symbol='SERPINA3', Aliases='AACT, ACT, GIG24, GIG25', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3', other_designations='cell growth-inhibiting gene 24/25 protein|growth-inhibiting protein 24|growth-inhibiting protein 25|serine (or cysteine) proteinase inhibitor, clade A, member 3|serpin A3', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94612377, end_position_on_the_genomic_accession=94624053, orientation='plus', exon_count=5, OMIM=107280, no_hdr0=''),
  13 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=13, CurrentID=0, Status='live', Symbol='AADAC', Aliases='CES5A1, DAC', description='arylacetamide deacetylase', other_designations='arylacetamide deacetylase (esterase)', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151814025, end_position_on_the_genomic_accession=151828488, orientation='plus', exon_count=5, OMIM=600338, no_hdr0=''),
  14 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=14, CurrentID=0, Status='live', Symbol='AAMP', Aliases='', description='angio associated migratory cell protein', other_designations='angio-associated, migratory cell protein', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218264129, end_position_on_the_genomic_accession=218270209, orientation='minus', exon_count=11, OMIM=603488, no_hdr0=''),
  15 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=15, CurrentID=0, Status='live', Symbol='AANAT', Aliases='DSPS, SNAT', description='aralkylamine N-acetyltransferase', other_designations='arylalkylamine N-acetyltransferase|serotonin acetylase', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76452853, end_position_on_the_genomic_accession=76470790, orientation='plus', exon_count=11, OMIM=600950, no_hdr0=''),
  16 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=16, CurrentID=0, Status='live', Symbol='AARS', Aliases='CMT2N, EIEE29', description='alanyl-tRNA synthetase', other_designations='alaRS|alanine tRNA ligase 1, cytoplasmic|alanyl-tRNA synthetase, cytoplasmic|renal carcinoma antigen NY-REN-42', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70252295, end_position_on_the_genomic_accession=70289550, orientation='minus', exon_count=21, OMIM=601065, no_hdr0=''),
  17 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=17, CurrentID=0, Status='live', Symbol='AAVS1', Aliases='AAV', description='adeno-associated virus integration site 1', other_designations='', map_location='19q13-qter', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=102699, no_hdr0=''),
  18 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=18, CurrentID=0, Status='live', Symbol='ABAT', Aliases='GABA-AT, GABAT, NPD009', description='4-aminobutyrate aminotransferase', other_designations='(S)-3-amino-2-methylpropionate transaminase|4-aminobutyrate transaminase|GABA aminotransferase|GABA transaminase|GABA transferase|gamma-amino-N-butyrate transaminase', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8674587, end_position_on_the_genomic_accession=8784575, orientation='plus', exon_count=20, OMIM=137150, no_hdr0=''),
  19 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=19, CurrentID=0, Status='live', Symbol='ABCA1', Aliases='ABC-1, ABC1, CERP, HDLDT1, TGD', description='ATP binding cassette subfamily A member 1', other_designations='ATP-binding cassette transporter A1|ATP-binding cassette, sub-family A (ABC1), member 1|cholesterol efflux regulatory protein|membrane-bound', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104781002, end_position_on_the_genomic_accession=104928246, orientation='minus', exon_count=53, OMIM=600046, no_hdr0=''),
  20 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=20, CurrentID=0, Status='live', Symbol='ABCA2', Aliases='ABC2', description='ATP binding cassette subfamily A member 2', other_designations='ATP-binding cassette 2|ATP-binding cassette transporter 2|ATP-binding cassette, sub-family A (ABC1), member 2|ATP-binding cassette, sub-family A, member 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137007234, end_position_on_the_genomic_accession=137028974, orientation='minus', exon_count=50, OMIM=600047, no_hdr0=''),
  21 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=21, CurrentID=0, Status='live', Symbol='ABCA3', Aliases='ABC-C, ABC3, EST111653, LBM180, SMDP3', description='ATP binding cassette subfamily A member 3', other_designations='ABC transporter 3|ABC-C transporter|ATP-binding cassette transporter 3|ATP-binding cassette, sub-family A (ABC1), member 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2275878, end_position_on_the_genomic_accession=2340746, orientation='minus', exon_count=33, OMIM=601615, no_hdr0=''),
  22 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22, CurrentID=0, Status='live', Symbol='ABCB7', Aliases='ABC7, ASAT, Atm1p, EST140535', description='ATP binding cassette subfamily B member 7', other_designations='ABC transporter 7 protein|ATP-binding cassette sub-family B member 7|ATP-binding cassette transporter 7|ATP-binding cassette, sub-family B (MDR/TAP), member 7', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75053172, end_position_on_the_genomic_accession=75156340, orientation='minus', exon_count=16, OMIM=300135, no_hdr0=''),
  23 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23, CurrentID=0, Status='live', Symbol='ABCF1', Aliases='ABC27, ABC50', description='ATP binding cassette subfamily F member 1', other_designations='ATP-binding cassette 50 (TNF-alpha stimulated)|ATP-binding cassette, sub-family F (GCN20), member 1|TNF-alpha-stimulated ABC protein|TNFalpha-inducible ATP-binding protein', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30571393, end_position_on_the_genomic_accession=30591532, orientation='plus', exon_count=25, OMIM=603429, no_hdr0=''),
  24 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24, CurrentID=0, Status='live', Symbol='ABCA4', Aliases='ABC10, ABCR, ARMD2, CORD3, FFM, RMP, RP19, STGD, STGD1', description='ATP binding cassette subfamily A member 4', other_designations='ATP binding cassette transporter|ATP-binding cassette sub-family A member 4|ATP-binding cassette transporter, retinal-specific|ATP-binding cassette, sub-family A (ABC1), member 4|ATP-binding transporter, retina-specific|RIM ABC transporter|RIM protein|photoreceptor rim protein|retina-specific ABC transporter|stargardt disease protein', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93992837, end_position_on_the_genomic_accession=94121149, orientation='minus', exon_count=50, OMIM=601691, no_hdr0=''),
  25 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25, CurrentID=0, Status='live', Symbol='ABL1', Aliases='ABL, JTK7, bcr/abl, c-ABL, c-ABL1, p150, v-abl', description='ABL proto-oncogene 1, non-receptor tyrosine kinase', other_designations='Abelson tyrosine-protein kinase 1|bcr/c-abl oncogene protein|c-abl oncogene 1, receptor tyrosine kinase|proto-oncogene c-Abl|proto-oncogene tyrosine-protein kinase ABL1|v-abl Abelson murine leukemia viral oncogene homolog 1', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130713881, end_position_on_the_genomic_accession=130887675, orientation='plus', exon_count=12, OMIM=189980, no_hdr0=''),
  26 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26, CurrentID=0, Status='live', Symbol='AOC1', Aliases='ABP, ABP1, DAO, DAO1, KAO', description='amine oxidase, copper containing 1', other_designations='amiloride binding protein 1 (amine oxidase (copper-containing))|amiloride-binding protein 1|amiloride-sensitive amine oxidase|amine oxidase copper domain-containing protein 1|diamine oxidase|histaminase|kidney amine oxidase', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150826393, end_position_on_the_genomic_accession=150861291, orientation='plus', exon_count=9, OMIM=104610, no_hdr0=''),
  27 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27, CurrentID=0, Status='live', Symbol='ABL2', Aliases='ABLL, ARG', description='ABL proto-oncogene 2, non-receptor tyrosine kinase', other_designations='abelson-related gene protein|c-abl oncogene 2, non-receptor tyrosine kinase|tyrosine-protein kinase ARG|v-abl Abelson murine leukemia viral oncogene homolog 2', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179099327, end_position_on_the_genomic_accession=179229693, orientation='minus', exon_count=15, OMIM=164690, no_hdr0=''),
  28 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28, CurrentID=0, Status='live', Symbol='ABO', Aliases='A3GALNT, A3GALT1, GTB, NAGAT', description='ABO blood group (transferase A, alpha 1-3-N-acetylgalactosaminyltransferase; transferase B, alpha 1-3-galactosyltransferase)', other_designations='A1-specific alpha 1-3-N-acetylgalactosaminyltransferase|ABO A3 transferase|ABO glycosyltransferase|ABO weak transfer|B(A) alpha-1,3-galactosyltransferase|fucosylglycoprotein|fucosylglycoprotein 3-alpha-galactosyltransferase|fucosylglycoprotein alpha-N-acetylgalactosaminyltransferase|glycoprotein-fucosylgalactoside alpha-N-acetylgalactosaminyltransferase|glycoprotein-fucosylgalactoside alpha-galactosyltransferase|glycosyltransferase A|glycosyltransferase B|histo-blood group A transferase|histo-blood group A2 transferase|histo-blood group B transferase', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133255176, end_position_on_the_genomic_accession=133275214, orientation='minus', exon_count=7, OMIM=110300, no_hdr0=''),
  29 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29, CurrentID=0, Status='live', Symbol='ABR', Aliases='MDB', description='active BCR-related', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1003518, end_position_on_the_genomic_accession=1187322, orientation='minus', exon_count=33, OMIM=600365, no_hdr0=''),
  30 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30, CurrentID=0, Status='live', Symbol='ACAA1', Aliases='ACAA, PTHIO, THIO', description='acetyl-CoA acyltransferase 1', other_designations='acetyl-Coenzyme A acyltransferase 1|beta-ketothiolase|peroxisomal 3-oxoacyl-CoA thiolase|peroxisomal 3-oxoacyl-Coenzyme A thiolase', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38122710, end_position_on_the_genomic_accession=38137242, orientation='minus', exon_count=12, OMIM=604054, no_hdr0=''),
  31 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=31, CurrentID=0, Status='live', Symbol='ACACA', Aliases='ACACD, ACC, ACC1, ACCA, ACACA', description='acetyl-CoA carboxylase alpha', other_designations='ACC-alpha|acetyl-Coenzyme A carboxylase alpha', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37084992, end_position_on_the_genomic_accession=37406822, orientation='minus', exon_count=65, OMIM=200350, no_hdr0=''),
  32 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=32, CurrentID=0, Status='live', Symbol='ACACB', Aliases='ACC2, ACCB, HACC275', description='acetyl-CoA carboxylase beta', other_designations='ACC-beta|acetyl-Coenzyme A carboxylase beta', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109111184, end_position_on_the_genomic_accession=109268226, orientation='plus', exon_count=58, OMIM=601557, no_hdr0=''),
  33 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=33, CurrentID=0, Status='live', Symbol='ACADL', Aliases='ACAD4, LCAD', description='acyl-CoA dehydrogenase, long chain', other_designations='acyl-Coenzyme A dehydrogenase, long chain', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210187990, end_position_on_the_genomic_accession=210225491, orientation='minus', exon_count=12, OMIM=609576, no_hdr0=''),
  34 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=34, CurrentID=0, Status='live', Symbol='ACADM', Aliases='ACAD1, MCAD, MCADH', description='acyl-CoA dehydrogenase, C-4 to C-12 straight chain', other_designations='acyl-CoA dehydrogenase C-4 to C-12 straight chain|acyl-Coenzyme A dehydrogenase, C-4 to C-12 straight chain', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75724347, end_position_on_the_genomic_accession=75763679, orientation='plus', exon_count=13, OMIM=607008, no_hdr0=''),
  35 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=35, CurrentID=0, Status='live', Symbol='ACADS', Aliases='ACAD3, SCAD', description='acyl-CoA dehydrogenase, C-2 to C-3 short chain', other_designations='acyl-Coenzyme A dehydrogenase, C-2 to C-3 short chain|butyryl-CoA dehydrogenase|mitochondrial short-chain specific acyl-CoA dehydrogenase|unsaturated acyl-CoA reductase', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120725738, end_position_on_the_genomic_accession=120740008, orientation='plus', exon_count=11, OMIM=606885, no_hdr0=''),
  36 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=36, CurrentID=0, Status='live', Symbol='ACADSB', Aliases='2-MEBCAD, ACAD7, SBCAD', description='acyl-CoA dehydrogenase, short/branched chain', other_designations='2-methyl branched chain acyl-CoA dehydrogenase|2-methylbutyryl-coenzyme A dehydrogenase|acyl-CoA dehydrogenase short/branched chain|acyl-Coenzyme A dehydrogenase, short/branched chain', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123008913, end_position_on_the_genomic_accession=123058290, orientation='plus', exon_count=11, OMIM=600301, no_hdr0=''),
  37 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=37, CurrentID=0, Status='live', Symbol='ACADVL', Aliases='ACAD6, LCACD, VLCAD', description='acyl-CoA dehydrogenase, very long chain', other_designations='acyl-Coenzyme A dehydrogenase, very long chain', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7217125, end_position_on_the_genomic_accession=7225267, orientation='plus', exon_count=22, OMIM=609575, no_hdr0=''),
  38 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=38, CurrentID=0, Status='live', Symbol='ACAT1', Aliases='ACAT, MAT, T2, THIL', description='acetyl-CoA acetyltransferase 1', other_designations='acetoacetyl Coenzyme A thiolase|acetoacetyl-CoA thiolase|acetyl-Coenzyme A acetyltransferase 1|mitochondrial acetoacetyl-CoA thiolase', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108121531, end_position_on_the_genomic_accession=108148168, orientation='plus', exon_count=14, OMIM=607809, no_hdr0=''),
  39 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=39, CurrentID=0, Status='live', Symbol='ACAT2', Aliases='', description='acetyl-CoA acetyltransferase 2', other_designations='acetoacetyl Coenzyme A thiolase|acetyl-CoA transferase-like protein|cytosolic acetoacetyl-CoA thiolase', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159761957, end_position_on_the_genomic_accession=159779112, orientation='plus', exon_count=10, OMIM=100678, no_hdr0=''),
  40 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=40, CurrentID=0, Status='live', Symbol='ASIC2', Aliases='ACCN, ACCN1a, BNC1, BNaC1, MDEG, hBNaC1, ASIC2', description='acid sensing ion channel subunit 2', other_designations='acid sensing (proton gated) ion channel 2|acid sensing ion channel 2|acid-sensing (proton-gated) ion channel 2|amiloride-sensitive cation channel 1, neuronal|brain sodium channel 1|mammalian degenerin homolog|neuronal amiloride-sensitive cation channel 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=33013087, end_position_on_the_genomic_accession=34156806, orientation='minus', exon_count=11, OMIM=601784, no_hdr0=''),
  41 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=41, CurrentID=0, Status='live', Symbol='ASIC1', Aliases='ACCN2, ASIC, BNaC2', description='acid sensing ion channel subunit 1', other_designations='Cation channel, amiloride-sensitive, neuronal, 2|acid sensing (proton gated) ion channel 1|acid sensing ion channel 1|acid-sensing (proton-gated) ion channel 1|acid-sensing ion channel 1a protein|amiloride-sensitive cation channel 2, neuronal|brain sodium channel 2', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50057637, end_position_on_the_genomic_accession=50083622, orientation='plus', exon_count=14, OMIM=602866, no_hdr0=''),
  43 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=43, CurrentID=0, Status='live', Symbol='ACHE', Aliases='ACEE, ARACHE, N-ACHE, YT', description='acetylcholinesterase (Yt blood group)', other_designations='Yt blood group|apoptosis-related acetylcholinesterase', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100889994, end_position_on_the_genomic_accession=100896686, orientation='minus', exon_count=7, OMIM=100740, no_hdr0=''),
  105373674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373674, CurrentID=0, Status='live', Symbol='LOC105373674', Aliases='', description='uncharacterized LOC105373674', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147983185, end_position_on_the_genomic_accession=147993777, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  131118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131118, CurrentID=0, Status='live', Symbol='DNAJC19', Aliases='PAM18, TIM14, TIMM14', description='DnaJ heat shock protein family (Hsp40) member C19', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19|DnaJ-like protein subfamily C member 19|homolog of yeast TIM14', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180983709, end_position_on_the_genomic_accession=180989774, orientation='minus', exon_count=7, OMIM=608977, no_hdr0=''),
  47 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=47, CurrentID=0, Status='live', Symbol='ACLY', Aliases='ACL, ATPCL, CLATP', description='ATP citrate lyase', other_designations='ATP citrate synthase|ATP-citrate (pro-S-)-lyase|citrate cleavage enzyme', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41866916, end_position_on_the_genomic_accession=41930542, orientation='minus', exon_count=30, OMIM=108728, no_hdr0=''),
  48 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=48, CurrentID=0, Status='live', Symbol='ACO1', Aliases='ACONS, HEL60, IREB1, IREBP, IREBP1, IRP1', description='aconitase 1', other_designations='IRE-BP 1|aconitase 1, soluble|aconitate hydratase, cytoplasmic|citrate hydro-lyase|cytosplasmic aconitase|epididymis luminal protein 60|ferritin repressor protein|iron regulatory protein 1|iron-responsive element binding protein 1|iron-responsive element-binding protein 1|soluble aconitase', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32384603, end_position_on_the_genomic_accession=32450836, orientation='plus', exon_count=23, OMIM=100880, no_hdr0=''),
  49 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49, CurrentID=0, Status='live', Symbol='ACR', Aliases='', description='acrosin', other_designations='acrosin light and heavy chain prepropeptide|preproacrosin|proacrosin', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50738224, end_position_on_the_genomic_accession=50745299, orientation='plus', exon_count=5, OMIM=102480, no_hdr0=''),
  50 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50, CurrentID=0, Status='live', Symbol='ACO2', Aliases='ACONM, HEL-S-284, ICRD, OCA8', description='aconitase 2', other_designations='aconitase 2, mitochondrial|citrate hydro-lyase|epididymis secretory sperm binding protein Li 284|mitochondrial aconitase', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41469125, end_position_on_the_genomic_accession=41528989, orientation='plus', exon_count=18, OMIM=100850, no_hdr0=''),
  51 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51, CurrentID=0, Status='live', Symbol='ACOX1', Aliases='ACOX, PALMCOX, SCOX', description='acyl-CoA oxidase 1, palmitoyl', other_designations='AOX|acyl-CoA oxidase, straight-chain|acyl-Coenzyme A oxidase 1, palmitoyl|palmitoyl-CoA oxidase|peroxisomal fatty acyl-CoA oxidase|straight-chain acyl-CoA oxidase', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75941507, end_position_on_the_genomic_accession=75979434, orientation='minus', exon_count=15, OMIM=609751, no_hdr0=''),
  52 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=52, CurrentID=0, Status='live', Symbol='ACP1', Aliases='HAAP, LMW-PTP', description='acid phosphatase 1, soluble', other_designations='LMW-PTPase|acid phosphatase of erythrocyte|adipocyte acid phosphatase|cytoplasmic phosphotyrosyl protein phosphatase|low molecular weight cytosolic acid phosphatase|protein tyrosine phosphatase|red cell acid phosphatase 1', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=264140, end_position_on_the_genomic_accession=278283, orientation='plus', exon_count=8, OMIM=171500, no_hdr0=''),
  53 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53, CurrentID=0, Status='live', Symbol='ACP2', Aliases='LAP', description='acid phosphatase 2, lysosomal', other_designations='', map_location='11p12-p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47239302, end_position_on_the_genomic_accession=47248847, orientation='minus', exon_count=12, OMIM=171650, no_hdr0=''),
  54 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54, CurrentID=0, Status='live', Symbol='ACP5', Aliases='HPAP, SPENCDI, TRAP', description='acid phosphatase 5, tartrate resistant', other_designations='TrATPase|human purple acid phosphatase|tartrate-resistant acid ATPase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11574660, end_position_on_the_genomic_accession=11578986, orientation='minus', exon_count=8, OMIM=171640, no_hdr0=''),
  55 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55, CurrentID=0, Status='live', Symbol='ACPP', Aliases="5'-NT, ACP-3, ACP3", description='acid phosphatase, prostate', other_designations="5'-nucleotidase|TMPase|ecto-5'-nucleotidase|prostatic acid phosphotase|thiamine monophosphatase", map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132317367, end_position_on_the_genomic_accession=132368302, orientation='plus', exon_count=12, OMIM=171790, no_hdr0=''),
  56 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56, CurrentID=0, Status='live', Symbol='ACRV1', Aliases='D11S4365, SP-10, SPACA2', description='acrosomal vesicle protein 1', other_designations='sperm protein 10', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125672334, end_position_on_the_genomic_accession=125680898, orientation='minus', exon_count=5, OMIM=102525, no_hdr0=''),
  58 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58, CurrentID=0, Status='live', Symbol='ACTA1', Aliases='ACTA, ASMA, CFTD, CFTD1, CFTDM, MPFD, NEM1, NEM2, NEM3', description='actin, alpha 1, skeletal muscle', other_designations='nemaline myopathy type 3', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229431245, end_position_on_the_genomic_accession=229434096, orientation='minus', exon_count=7, OMIM=102610, no_hdr0=''),
  59 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59, CurrentID=0, Status='live', Symbol='ACTA2', Aliases='AAT6, ACTSA, MYMY5', description='actin, alpha 2, smooth muscle, aorta', other_designations='alpha-cardiac actin|cell growth-inhibiting gene 46 protein', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88935074, end_position_on_the_genomic_accession=88991390, orientation='minus', exon_count=10, OMIM=102620, no_hdr0=''),
  60 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60, CurrentID=0, Status='live', Symbol='ACTB', Aliases='BRWS1, PS1TP5BP1', description='actin, beta', other_designations='PS1TP5-binding protein 1|beta cytoskeletal actin', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5527148, end_position_on_the_genomic_accession=5530601, orientation='minus', exon_count=6, OMIM=102630, no_hdr0=''),
  62 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=62, CurrentID=0, Status='live', Symbol='ACTBP2', Aliases='', description='actin, beta pseudogene 2', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77784803, end_position_on_the_genomic_accession=77786597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  63 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63, CurrentID=0, Status='live', Symbol='ACTBP3', Aliases='', description='actin, beta pseudogene 3', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59693046, end_position_on_the_genomic_accession=59693473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64, CurrentID=0, Status='live', Symbol='ACTBP4', Aliases='', description='actin, beta pseudogene 4', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131658440, end_position_on_the_genomic_accession=131660199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  66 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66, CurrentID=0, Status='live', Symbol='ACTBP6', Aliases='H8-PSI-BETA-AC3', description='actin, beta pseudogene 6', other_designations='actin, gamma 1 pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84948509, end_position_on_the_genomic_accession=84950268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  67 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=67, CurrentID=0, Status='live', Symbol='ACTBP7', Aliases='', description='actin, beta pseudogene 7', other_designations='actin, gamma 1 pseudogene', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43988478, end_position_on_the_genomic_accession=43990265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  68 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=68, CurrentID=0, Status='live', Symbol='ACTBP8', Aliases='ACTBP2', description='actin, beta pseudogene 8', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88275305, end_position_on_the_genomic_accession=88277073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  69 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=69, CurrentID=0, Status='live', Symbol='ACTBP9', Aliases='', description='actin, beta pseudogene 9', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62441924, end_position_on_the_genomic_accession=62443715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  70 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=70, CurrentID=0, Status='live', Symbol='ACTC1', Aliases='ACTC, ASD5, CMD1R, CMH11, LVNC4', description='actin, alpha, cardiac muscle 1', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34788096, end_position_on_the_genomic_accession=34795726, orientation='minus', exon_count=7, OMIM=102540, no_hdr0=''),
  71 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=71, CurrentID=0, Status='live', Symbol='ACTG1', Aliases='ACT, ACTG, BRWS2, DFNA20, DFNA26, HEL-176', description='actin gamma 1', other_designations='cytoskeletal gamma-actin|epididymis luminal protein 176', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81509971, end_position_on_the_genomic_accession=81512866, orientation='minus', exon_count=7, OMIM=102560, no_hdr0=''),
  72 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=72, CurrentID=0, Status='live', Symbol='ACTG2', Aliases='ACT, ACTA3, ACTE, ACTL3, ACTSG, VSCM', description='actin, gamma 2, smooth muscle, enteric', other_designations='actin-like protein|alpha-actin-3', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73892966, end_position_on_the_genomic_accession=73919653, orientation='plus', exon_count=9, OMIM=102545, no_hdr0=''),
  73 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=73, CurrentID=0, Status='live', Symbol='ACTG1P1', Aliases='ACT1GP1, ACTGP1, HY-psi-gamma-AC6', description='actin gamma 1 pseudogene 1', other_designations='actin, gamma 1 pseudogene 1|actin, gamma pseudogene 1', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139493740, end_position_on_the_genomic_accession=139495682, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  74 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=74, CurrentID=0, Status='live', Symbol='ACTG1P2', Aliases='ACTGP2, ACTL2', description='actin gamma 1 pseudogene 2', other_designations='actin, gamma 1 pseudogene 2', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17756197, end_position_on_the_genomic_accession=17758298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  75 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=75, CurrentID=0, Status='live', Symbol='ACTG1P3', Aliases='ACTGP3, dJ545L17.4', description='actin gamma 1 pseudogene 3', other_designations='actin, gamma 1 pseudogene 3|actin, gamma pseudogene 3', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1160509, end_position_on_the_genomic_accession=1162097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131149, CurrentID=0, Status='live', Symbol='OTOL1', Aliases='C1QTNF15', description='otolin 1', other_designations='C1q and tumor necrosis factor related protein 15|otolin 1 homolog', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161429601, end_position_on_the_genomic_accession=161503975, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  78 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78, CurrentID=0, Status='live', Symbol='ACTG1P6', Aliases='ACTGP6', description='actin gamma 1 pseudogene 6', other_designations='actin, gamma 1 pseudogene 6|actin, gamma pseudogene 6', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79, CurrentID=0, Status='live', Symbol='ACTG1P7', Aliases='ACTGP7', description='actin gamma 1 pseudogene 7', other_designations='actin, gamma 1 pseudogene 7|actin, gamma pseudogene 7', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  80 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80, CurrentID=0, Status='live', Symbol='ACTG1P8', Aliases='ACTGP8', description='actin gamma 1 pseudogene 8', other_designations='actin, gamma 1 pseudogene 8|actin, gamma pseudogene 8', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81, CurrentID=0, Status='live', Symbol='ACTN4', Aliases='ACTININ-4, FSGS, FSGS1', description='actinin, alpha 4', other_designations='F-actin cross-linking protein|actinin alpha 4|actinin alpha4 isoform|focal segmental glomerulosclerosis 1|non-muscle alpha-actinin 4', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38647616, end_position_on_the_genomic_accession=38730531, orientation='plus', exon_count=23, OMIM=604638, no_hdr0=''),
  82 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=82, CurrentID=0, Status='live', Symbol='ACTG1P9', Aliases='ACTGP9', description='actin gamma 1 pseudogene 9', other_designations='actin, gamma 1 pseudogene 9|actin, gamma pseudogene 9', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46204912, end_position_on_the_genomic_accession=46206561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83, CurrentID=0, Status='live', Symbol='ACTG1P10', Aliases='ACTGP10, ACTL1, ACTP1', description='actin gamma 1 pseudogene 10', other_designations='actin, gamma 1 pseudogene 10|actin, gamma pseudogene 10|actin-like 1|actin-like sequence-1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53141688, end_position_on_the_genomic_accession=53143984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  86 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=86, CurrentID=0, Status='live', Symbol='ACTL6A', Aliases='ACTL6, ARPN-BETA, Arp4, BAF53A, INO80K', description='actin like 6A', other_designations='53 kDa BRG1-associated factor A|BAF complex 53 kDa subunit|BAF53|BRG1-associated factor 53A|INO80 complex subunit K|actin-like 6A|actin-related protein 4|actin-related protein Baf53a|arpNbeta|hArpN beta', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179562880, end_position_on_the_genomic_accession=179588405, orientation='plus', exon_count=14, OMIM=604958, no_hdr0=''),
  87 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=87, CurrentID=0, Status='live', Symbol='ACTN1', Aliases='BDPLT15', description='actinin, alpha 1', other_designations='F-actin cross-linking protein|actinin 1 smooth muscle|alpha-actinin cytoskeletal isoform|non-muscle alpha-actinin-1', map_location='14q22-q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68874123, end_position_on_the_genomic_accession=68979366, orientation='minus', exon_count=24, OMIM=102575, no_hdr0=''),
  88 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=88, CurrentID=0, Status='live', Symbol='ACTN2', Aliases='CMD1AA, CMH23', description='actinin, alpha 2', other_designations='F-actin cross-linking protein|alpha-actinin skeletal muscle', map_location='1q42-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236686454, end_position_on_the_genomic_accession=236764631, orientation='plus', exon_count=23, OMIM=102573, no_hdr0=''),
  89 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89, CurrentID=0, Status='live', Symbol='ACTN3', Aliases='', description='actinin, alpha 3 (gene/pseudogene)', other_designations='F-actin cross-linking protein|alpha-actinin skeletal muscle', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66546395, end_position_on_the_genomic_accession=66563329, orientation='plus', exon_count=22, OMIM=102574, no_hdr0=''),
  90 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90, CurrentID=0, Status='live', Symbol='ACVR1', Aliases='ACTRIA, ACVRLK2, ALK2, FOP, SKR1, TSRI, ACVR1', description='activin A receptor type I', other_designations='TGF-B superfamily receptor type I|activin A receptor, type I|activin A receptor, type II-like kinase 2|activin receptor type I|activin receptor-like kinase 2|hydroxyalkyl-protein kinase|serine/threonine-protein kinase receptor R1', map_location='2q23-q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157736444, end_position_on_the_genomic_accession=157875862, orientation='minus', exon_count=18, OMIM=102576, no_hdr0=''),
  91 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91, CurrentID=0, Status='live', Symbol='ACVR1B', Aliases='ACTRIB, ACVRLK4, ALK4, SKR2', description='activin A receptor type IB', other_designations='activin A receptor, type IB|activin A receptor, type II-like kinase 4|activin receptor-like kinase 4|serine/threonine-protein kinase receptor R2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51951667, end_position_on_the_genomic_accession=51997079, orientation='plus', exon_count=12, OMIM=601300, no_hdr0=''),
  92 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92, CurrentID=0, Status='live', Symbol='ACVR2A', Aliases='ACTRII, ACVR2', description='activin A receptor type IIA', other_designations='activin A receptor, type IIA', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147844517, end_position_on_the_genomic_accession=147930827, orientation='plus', exon_count=12, OMIM=102581, no_hdr0=''),
  93 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93, CurrentID=0, Status='live', Symbol='ACVR2B', Aliases='ACTRIIB, ActR-IIB, HTX4', description='activin A receptor type IIB', other_designations='activin A receptor, type IIB', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38454299, end_position_on_the_genomic_accession=38493142, orientation='plus', exon_count=12, OMIM=602730, no_hdr0=''),
  94 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94, CurrentID=0, Status='live', Symbol='ACVRL1', Aliases='ACVRLK1, ALK-1, ALK1, HHT, HHT2, ORW2, SKR3, TSR-I', description='activin A receptor type IL', other_designations='TGF-B superfamily receptor type I|activin A receptor type II-like 1|activin A receptor, type II-like kinase 1', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51906913, end_position_on_the_genomic_accession=51923361, orientation='plus', exon_count=11, OMIM=601284, no_hdr0=''),
  95 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=95, CurrentID=0, Status='live', Symbol='ACY1', Aliases='ACY-1D, HEL-S-5, ACY1', description='aminoacylase 1', other_designations='N-acyl-L-amino-acid amidohydrolase|acylase|epididymis secretory protein Li 5', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51983284, end_position_on_the_genomic_accession=51989202, orientation='plus', exon_count=15, OMIM=104620, no_hdr0=''),
  97 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=97, CurrentID=0, Status='live', Symbol='ACYP1', Aliases='ACYPE', description='acylphosphatase 1, erythrocyte (common) type', other_designations='acylphosphatase, erythrocyte isozyme|acylphosphatase, organ-common type isozyme|acylphosphate phosphohydrolase 1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75053221, end_position_on_the_genomic_accession=75069483, orientation='minus', exon_count=5, OMIM=600875, no_hdr0=''),
  98 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=98, CurrentID=0, Status='live', Symbol='ACYP2', Aliases='ACYM, ACYP', description='acylphosphatase 2, muscle type', other_designations='acylphosphatase, muscle type isozyme|acylphosphate phosphohydrolase 2', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54115273, end_position_on_the_genomic_accession=54305300, orientation='plus', exon_count=4, OMIM=102595, no_hdr0=''),
  100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100, CurrentID=0, Status='live', Symbol='ADA', Aliases='', description='adenosine deaminase', other_designations='adenosine aminohydrolase|mutant adenosine deaminase', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44619522, end_position_on_the_genomic_accession=44652233, orientation='minus', exon_count=13, OMIM=608958, no_hdr0=''),
  101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101, CurrentID=0, Status='live', Symbol='ADAM8', Aliases='CD156, CD156a, MS2', description='ADAM metallopeptidase domain 8', other_designations='a disintegrin and metalloproteinase domain 8|cell surface antigen MS2|human leukocyte differentiation antigen', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133262416, end_position_on_the_genomic_accession=133276903, orientation='minus', exon_count=25, OMIM=602267, no_hdr0=''),
  102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102, CurrentID=0, Status='live', Symbol='ADAM10', Aliases='AD10, AD18, CD156c, HsT18717, MADM, RAK, kuz', description='ADAM metallopeptidase domain 10', other_designations='CDw156|a disintegrin and metalloprotease domain 10|kuzbanian protein homolog|mammalian disintegrin-metalloprotease', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58595204, end_position_on_the_genomic_accession=58749978, orientation='minus', exon_count=16, OMIM=602192, no_hdr0=''),
  103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103, CurrentID=0, Status='live', Symbol='ADAR', Aliases='ADAR1, AGS6, DRADA, DSH, DSRAD, G1P1, IFI-4, IFI4, K88DSRBP, P136', description='adenosine deaminase, RNA-specific', other_designations='136 kDa double-stranded RNA-binding protein|adenosine deaminase acting on RNA 1-A|dsRNA adenosine deaminase|interferon-induced protein 4|interferon-inducible protein 4', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154582057, end_position_on_the_genomic_accession=154628013, orientation='minus', exon_count=18, OMIM=146920, no_hdr0=''),
  104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104, CurrentID=0, Status='live', Symbol='ADARB1', Aliases='ADAR2, DRABA2, DRADA2, RED1', description='adenosine deaminase, RNA-specific, B1', other_designations='RED1 homolog|RNA editase|RNA editing deaminase 1|RNA-editing deaminase 1|RNA-editing enzyme 1|adenosine deaminase, RNA-specific, B1 (RED1 homolog rat)|adenosine deaminase, RNA-specific, B1 (homolog of rat RED1)|dsRNA adenosine deaminase DRADA2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45074578, end_position_on_the_genomic_accession=45226563, orientation='plus', exon_count=19, OMIM=601218, no_hdr0=''),
  105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105, CurrentID=0, Status='live', Symbol='ADARB2', Aliases='ADAR3, RED2', description='adenosine deaminase, RNA-specific, B2 (inactive)', other_designations='RED2 homolog|RNA-dependent adenosine deaminase 3|RNA-editing deaminase 2|RNA-editing enzyme 2|adenosine deaminase, RNA-specific, B2 (RED1 homolog rat)|adenosine deaminase, RNA-specific, B2 (RED2 homolog rat)|adenosine deaminase, RNA-specific, B2 (non-functional)|dsRNA adenosine deaminase B2|homolog of rat BLUE', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1177313, end_position_on_the_genomic_accession=1737498, orientation='minus', exon_count=11, OMIM=602065, no_hdr0=''),
  106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106, CurrentID=0, Status='live', Symbol='ADCP1', Aliases='', description='adenosine deaminase complexing protein 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=107, CurrentID=0, Status='live', Symbol='ADCY1', Aliases='AC1, DFNB44', description='adenylate cyclase 1 (brain)', other_designations="3',5'-cyclic AMP synthetase|ATP pyrophosphate-lyase 1|Ca(2+)/calmodulin-activated adenylyl cyclase|adenyl cyclase|adenylate cyclase type I|adenylyl cyclase 1", map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45574140, end_position_on_the_genomic_accession=45723116, orientation='plus', exon_count=22, OMIM=103072, no_hdr0=''),
  108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=108, CurrentID=0, Status='live', Symbol='ADCY2', Aliases='AC2, HBAC2', description='adenylate cyclase 2 (brain)', other_designations="3',5'-cyclic AMP synthetase|ATP pyrophosphate-lyase 2|adenylate cyclase II|adenylate cyclase type II|adenylyl cyclase 2|type II adenylate cyclase", map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7396230, end_position_on_the_genomic_accession=7830081, orientation='plus', exon_count=25, OMIM=103071, no_hdr0=''),
  109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=109, CurrentID=0, Status='live', Symbol='ADCY3', Aliases='AC3', description='adenylate cyclase 3', other_designations='AC-III|ATP pyrophosphate-lyase 3|adenylate cyclase type III|adenylate cyclase, olfactive type|adenylyl cyclase 3|adenylyl cyclase, type III', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24819169, end_position_on_the_genomic_accession=24919733, orientation='minus', exon_count=26, OMIM=600291, no_hdr0=''),
  111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=111, CurrentID=0, Status='live', Symbol='ADCY5', Aliases='AC5, FDFM', description='adenylate cyclase 5', other_designations='ATP pyrophosphate-lyase 5|adenylate cyclase type V|adenylyl cyclase 5', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123282296, end_position_on_the_genomic_accession=123448638, orientation='minus', exon_count=29, OMIM=600293, no_hdr0=''),
  112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112, CurrentID=0, Status='live', Symbol='ADCY6', Aliases='AC6, LCCS8', description='adenylate cyclase 6', other_designations='ATP pyrophosphate-lyase 6|adenylate cyclase type VI|ca(2+)-inhibitable adenylyl cyclase', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48766192, end_position_on_the_genomic_accession=48789037, orientation='minus', exon_count=23, OMIM=600294, no_hdr0=''),
  113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113, CurrentID=0, Status='live', Symbol='ADCY7', Aliases='AC7', description='adenylate cyclase 7', other_designations='ATP pyrophosphate-lyase 7|adenylate cyclase type VII|adenylyl cyclase 7', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50244749, end_position_on_the_genomic_accession=50318135, orientation='plus', exon_count=31, OMIM=600385, no_hdr0=''),
  114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114, CurrentID=0, Status='live', Symbol='ADCY8', Aliases='AC8, ADCY3, HBAC1', description='adenylate cyclase 8 (brain)', other_designations='ATP pyrophosphate-lyase 8|HEL-S-172mP|adenylate cyclase type VIII|adenylyl cyclase 8|adenylyl cyclase-8, brain|ca(2+)/calmodulin-activated adenylyl cyclase|epididymis secretory sperm binding protein Li 172mP', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130780300, end_position_on_the_genomic_accession=131041598, orientation='minus', exon_count=18, OMIM=103070, no_hdr0=''),
  115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115, CurrentID=0, Status='live', Symbol='ADCY9', Aliases='AC9, ACIX', description='adenylate cyclase 9', other_designations='ATP pyrophosphate-lyase 9|adenylate cyclase type IX|adenylyl cyclase 9|type IX ATP pyrophosphate-lyase', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3953387, end_position_on_the_genomic_accession=4116185, orientation='minus', exon_count=12, OMIM=603302, no_hdr0=''),
  116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116, CurrentID=0, Status='live', Symbol='ADCYAP1', Aliases='PACAP', description='adenylate cyclase activating polypeptide 1 (pituitary)', other_designations='prepro-PACAP', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=904345, end_position_on_the_genomic_accession=912172, orientation='plus', exon_count=6, OMIM=102980, no_hdr0=''),
  117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117, CurrentID=0, Status='live', Symbol='ADCYAP1R1', Aliases='PAC1, PAC1R, PACAPR, PACAPRI', description='adenylate cyclase activating polypeptide 1 (pituitary) receptor type I', other_designations='PACAP receptor 1|PACAP type I receptor|PACAP-R1|pituitary adenylate cyclase activating polypeptide 1 receptor type I Hiphop', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31052341, end_position_on_the_genomic_accession=31111479, orientation='plus', exon_count=18, OMIM=102981, no_hdr0=''),
  118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118, CurrentID=0, Status='live', Symbol='ADD1', Aliases='ADDA', description='adducin 1', other_designations='adducin 1 (alpha)|erythrocyte adducin alpha subunit', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2843727, end_position_on_the_genomic_accession=2930076, orientation='plus', exon_count=19, OMIM=102680, no_hdr0=''),
  119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119, CurrentID=0, Status='live', Symbol='ADD2', Aliases='ADDB', description='adducin 2', other_designations='adducin 2 (beta)|erythrocyte adducin subunit beta', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70661202, end_position_on_the_genomic_accession=70768243, orientation='minus', exon_count=20, OMIM=102681, no_hdr0=''),
  120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120, CurrentID=0, Status='live', Symbol='ADD3', Aliases='ADDL', description='adducin 3', other_designations='adducin 3 (gamma)|adducin-like protein 70', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109996365, end_position_on_the_genomic_accession=110135565, orientation='plus', exon_count=23, OMIM=601568, no_hdr0=''),
  122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122, CurrentID=0, Status='live', Symbol='ADFN', Aliases='ALDS', description='albinism-deafness syndrome', other_designations='', map_location='Xq26.3-q27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300700, no_hdr0=''),
  123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123, CurrentID=0, Status='live', Symbol='PLIN2', Aliases='ADFP, ADRP', description='perilipin 2', other_designations='adipophilin|adipose differentiation-related protein', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19108388, end_position_on_the_genomic_accession=19127606, orientation='minus', exon_count=9, OMIM=103195, no_hdr0=''),
  124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124, CurrentID=0, Status='live', Symbol='ADH1A', Aliases='ADH1', description='alcohol dehydrogenase 1A (class I), alpha polypeptide', other_designations='ADH, alpha subunit|alcohol dehydrogenase 1 (class I), alpha polypeptide|alcohol dehydrogenase subunit alpha|aldehyde reductase', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99276366, end_position_on_the_genomic_accession=99291028, orientation='minus', exon_count=9, OMIM=103700, no_hdr0=''),
  125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125, CurrentID=0, Status='live', Symbol='ADH1B', Aliases='ADH2, HEL-S-117', description='alcohol dehydrogenase 1B (class I), beta polypeptide', other_designations='ADH, beta subunit|alcohol dehydrogenase 2 (class I), beta polypeptide|alcohol dehydrogenase subunit beta|aldehyde reductase|epididymis secretory protein Li 117', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99306387, end_position_on_the_genomic_accession=99321442, orientation='minus', exon_count=10, OMIM=103720, no_hdr0=''),
  126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126, CurrentID=0, Status='live', Symbol='ADH1C', Aliases='ADH3', description='alcohol dehydrogenase 1C (class I), gamma polypeptide', other_designations='ADH, gamma subunit|alcohol dehydrogenase 3 (class I), gamma polypeptide|aldehyde reductase', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99336492, end_position_on_the_genomic_accession=99353045, orientation='minus', exon_count=10, OMIM=103730, no_hdr0=''),
  127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127, CurrentID=0, Status='live', Symbol='ADH4', Aliases='ADH-2, HEL-S-4', description='alcohol dehydrogenase 4 (class II), pi polypeptide', other_designations='alcohol dehydrogenase class II pi chain|aldehyde reductase|epididymis secretory protein Li 4', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99123681, end_position_on_the_genomic_accession=99144314, orientation='minus', exon_count=11, OMIM=103740, no_hdr0=''),
  128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128, CurrentID=0, Status='live', Symbol='ADH5', Aliases='ADH-3, ADHX, FALDH, FDH, GSH-FDH, GSNOR, HEL-S-60p', description='alcohol dehydrogenase 5 (class III), chi polypeptide', other_designations='S-(hydroxymethyl)glutathione dehydrogenase|alcohol dehydrogenase (class III), chi polypeptide|alcohol dehydrogenase class chi chain|alcohol dehydrogenase class-III|epididymis secretory sperm binding protein Li 60p|formaldehyde dehydrogenase|glutathione-dependent formaldehyde dehydrogenase', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99070978, end_position_on_the_genomic_accession=99088788, orientation='minus', exon_count=9, OMIM=103710, no_hdr0=''),
  130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130, CurrentID=0, Status='live', Symbol='ADH6', Aliases='ADH-5', description='alcohol dehydrogenase 6 (class V)', other_designations='aldehyde reductase', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99202638, end_position_on_the_genomic_accession=99219246, orientation='minus', exon_count=9, OMIM=103735, no_hdr0=''),
  131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131, CurrentID=0, Status='live', Symbol='ADH7', Aliases='ADH4', description='alcohol dehydrogenase 7 (class IV), mu or sigma polypeptide', other_designations='alcohol dehydrogenase VII|alcohol dehydrogenase class IV mu/sigma chain|alcohol dehydrogenase-7|class IV sigma-1 alcohol dehydrogenase|class IV sigmasigma alcohol dehydrogenase|gastric alcohol dehydrogenase|retinol dehydrogenase', map_location='4q23-q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99412261, end_position_on_the_genomic_accession=99435510, orientation='minus', exon_count=10, OMIM=600086, no_hdr0=''),
  132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132, CurrentID=0, Status='live', Symbol='ADK', Aliases='AK', description='adenosine kinase', other_designations="adenosine 5'-phosphotransferase", map_location='10q11-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74151185, end_position_on_the_genomic_accession=74709303, orientation='plus', exon_count=13, OMIM=102750, no_hdr0=''),
  133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133, CurrentID=0, Status='live', Symbol='ADM', Aliases='AM, PAMP', description='adrenomedullin', other_designations='preproadrenomedullin|proadrenomedullin N-20 terminal peptide', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10304980, end_position_on_the_genomic_accession=10307402, orientation='plus', exon_count=4, OMIM=103275, no_hdr0=''),
  134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134, CurrentID=0, Status='live', Symbol='ADORA1', Aliases='RDC7', description='adenosine A1 receptor', other_designations='adenosine A1 receptor variant 1|adenosine A1 receptor variant 2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203127705, end_position_on_the_genomic_accession=203167405, orientation='plus', exon_count=6, OMIM=102775, no_hdr0=''),
  135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135, CurrentID=0, Status='live', Symbol='ADORA2A', Aliases='A2aR, ADORA2, RDC8', description='adenosine A2a receptor', other_designations='adenosine receptor subtype A2a', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24423597, end_position_on_the_genomic_accession=24442360, orientation='plus', exon_count=6, OMIM=102776, no_hdr0=''),
  136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136, CurrentID=0, Status='live', Symbol='ADORA2B', Aliases='ADORA2', description='adenosine A2b receptor', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15927783, end_position_on_the_genomic_accession=15975896, orientation='plus', exon_count=6, OMIM=600446, no_hdr0=''),
  137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137, CurrentID=0, Status='live', Symbol='ADORA2BP1', Aliases='ADORA2BP', description='adenosine A2b receptor pseudogene 1', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209744350, end_position_on_the_genomic_accession=209745676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140, CurrentID=0, Status='live', Symbol='ADORA3', Aliases='A3AR', description='adenosine A3 receptor', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111499429, end_position_on_the_genomic_accession=111504121, orientation='minus', exon_count=2, OMIM=600445, no_hdr0=''),
  141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=141, CurrentID=0, Status='live', Symbol='ADPRH', Aliases='ARH1', description='ADP-ribosylarginine hydrolase', other_designations='ADP-ribose-L-arginine cleaving enzyme|[Protein ADP-ribosylarginine] hydrolase', map_location='3q13.31-q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119579433, end_position_on_the_genomic_accession=119589945, orientation='plus', exon_count=5, OMIM=603081, no_hdr0=''),
  142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142, CurrentID=0, Status='live', Symbol='PARP1', Aliases='ADPRT, ADPRT 1, ADPRT1, ARTD1, PARP, PARP-1, PPOL, pADPRT-1', description='poly(ADP-ribose) polymerase 1', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)|ADP-ribosyltransferase NAD(+)|ADP-ribosyltransferase diphtheria toxin-like 1|NAD(+) ADP-ribosyltransferase 1|poly (ADP-ribose) polymerase 1|poly (ADP-ribose) polymerase family, member 1|poly(ADP-ribose) polymerase|poly(ADP-ribose) synthetase|poly(ADP-ribosyl)transferase|poly[ADP-ribose] synthase 1', map_location='1q41-q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226360691, end_position_on_the_genomic_accession=226408100, orientation='minus', exon_count=23, OMIM=173870, no_hdr0=''),
  143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143, CurrentID=0, Status='live', Symbol='PARP4', Aliases='ADPRTL1, ARTD4, PARP-4, PARPL, PH5P, VAULT3, VPARP, VWA5C, p193', description='poly(ADP-ribose) polymerase family member 4', other_designations='193 kDa vault protein|ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 1|ADP-ribosyltransferase diphtheria toxin-like 4|H5 proline-rich|I-alpha-I-related|PARP-related|PARP-related/IalphaI-related H5/proline-rich|poly (ADP-ribose) polymerase family, member 4|poly(ADP-ribose) synthetase|poly(ADP-ribosyl)transferase-like 1|vault poly(ADP-ribose) polymerase|vault protein, 193-kDa|von Willebrand factor A domain containing 5C', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24420930, end_position_on_the_genomic_accession=24512993, orientation='minus', exon_count=36, OMIM=607519, no_hdr0=''),
  144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144, CurrentID=0, Status='live', Symbol='PARP1P1', Aliases='ADPRTP1, PPOLP1', description='poly(ADP-ribose) polymerase family member 1 pseudogene 1', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase) pseudogene 1|poly (ADP-ribose) polymerase 1 pseudogene|poly (ADP-ribose) polymerase family, member 1 pseudogene 1|poly(ADP-ribose) polymerase family, member 1 pseudogene 1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110936624, end_position_on_the_genomic_accession=110940232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131096, CurrentID=0, Status='live', Symbol='KCNH8', Aliases='ELK, ELK1, Kv12.1, elk3', description='potassium channel, voltage gated eag related subfamily H, member 8', other_designations='ELK channel 3|ether-a-go-go-like potassium channel 1|ether-a-go-go-like potassium channel 3|hElk1|potassium voltage-gated channel, subfamily H (eag-related), member 8|voltage-gated potassium channel subunit Kv12.1', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19147093, end_position_on_the_genomic_accession=19535643, orientation='plus', exon_count=19, OMIM=608260, no_hdr0=''),
  146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146, CurrentID=0, Status='live', Symbol='ADRA1D', Aliases='ADRA1, ADRA1A, ADRA1R, ALPHA1, DAR, dJ779E11.2', description='adrenoceptor alpha 1D', other_designations='adrenergic, alpha -1D-, receptor|adrenergic, alpha-1A-, receptor|adrenergic, alpha-1D-, receptor|alpha-1A adrenergic receptor|alpha-1D adrenoceptor|alpha-1D adrenoreceptor|alpha-1D-adrenergic receptor|alpha-adrenergic receptor 1a', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4220631, end_position_on_the_genomic_accession=4249012, orientation='minus', exon_count=2, OMIM=104219, no_hdr0=''),
  147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147, CurrentID=0, Status='live', Symbol='ADRA1B', Aliases='ADRA1, ALPHA1BAR', description='adrenoceptor alpha 1B', other_designations='adrenergic, alpha-1B-, receptor|alpha-1B adrenoceptor|alpha-1B adrenoreceptor|alpha-1B-adrenergic receptor', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159916733, end_position_on_the_genomic_accession=159986178, orientation='plus', exon_count=8, OMIM=104220, no_hdr0=''),
  148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148, CurrentID=0, Status='live', Symbol='ADRA1A', Aliases='ADRA1C, ADRA1L1, ALPHA1AAR', description='adrenoceptor alpha 1A', other_designations='G protein coupled receptor|adrenergic, alpha-1A-, receptor variant 1|adrenergic, alpha-1A-, receptor variant 11|adrenergic, alpha-1A-, receptor variant 2|adrenergic, alpha-1A-, receptor variant 3|adrenergic, alpha-1A-, receptor variant 4|adrenergic, alpha-1A-, receptor variant 5|adrenergic, alpha-1A-, receptor variant 6|adrenergic, alpha-1A-, receptor variant 8|adrenergic, alpha-1A-, receptor variant 9|alpha-1A adrenoceptor|alpha-1A adrenoreceptor|alpha-1C adrenergic receptor', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26742847, end_position_on_the_genomic_accession=26871008, orientation='minus', exon_count=8, OMIM=104221, no_hdr0=''),
  150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150, CurrentID=0, Status='live', Symbol='ADRA2A', Aliases='ADRA2, ADRA2R, ADRAR, ALPHA2AAR, ZNF32', description='adrenoceptor alpha 2A', other_designations='adrenergic, alpha-2A-, receptor|alpha-2 adrenergic receptor subtype C10|alpha-2-adrenergic receptor, platelet type|alpha-2A adrenoceptor|alpha-2A adrenoreceptor|alpha-2AAR subtype C10', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111077032, end_position_on_the_genomic_accession=111080907, orientation='plus', exon_count=1, OMIM=104210, no_hdr0=''),
  151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151, CurrentID=0, Status='live', Symbol='ADRA2B', Aliases='ADRA2L1, ADRA2RL1, ADRARL1, ALPHA2BAR, FAME2, alpha-2BAR', description='adrenoceptor alpha 2B', other_designations='ADRA2B adrenergic, alpha-2B-, receptor|G-protein coupled receptor|alpha-2 adrenergic receptor subtype C2|alpha-2-adrenergic receptor-like 1|alpha-2B adrenoreceptor', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96112875, end_position_on_the_genomic_accession=96116245, orientation='minus', exon_count=1, OMIM=104260, no_hdr0=''),
  152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152, CurrentID=0, Status='live', Symbol='ADRA2C', Aliases='ADRA2L2, ADRA2RL2, ADRARL2, ALPHA2CAR', description='adrenoceptor alpha 2C', other_designations='adrenergic, alpha-2C-, receptor|alpha-2 adrenergic receptor subtype C4|alpha-2C adrenoceptor|alpha-2C adrenoreceptor|alpha-2C-adrenergic receptor|alpha-2CAR|alpha2-AR-C4', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3766569, end_position_on_the_genomic_accession=3768526, orientation='plus', exon_count=1, OMIM=104250, no_hdr0=''),
  153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153, CurrentID=0, Status='live', Symbol='ADRB1', Aliases='ADRB1R, B1AR, BETA1AR, RHR', description='adrenoceptor beta 1', other_designations='adrenergic, beta-1-, receptor|beta-1 adrenoceptor|beta-1 adrenoreceptor', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114044047, end_position_on_the_genomic_accession=114046908, orientation='plus', exon_count=1, OMIM=109630, no_hdr0=''),
  154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154, CurrentID=0, Status='live', Symbol='ADRB2', Aliases='ADRB2R, ADRBR, B2AR, BAR, BETA2AR', description='adrenoceptor beta 2', other_designations='adrenergic, beta-2-, receptor, surface|adrenoceptor beta 2 surface|adrenoceptor beta 2, surface|beta-2 adrenoceptor|beta-2 adrenoreceptor|catecholamine receptor', map_location='5q31-q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148826593, end_position_on_the_genomic_accession=148828634, orientation='plus', exon_count=1, OMIM=109690, no_hdr0=''),
  155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155, CurrentID=0, Status='live', Symbol='ADRB3', Aliases='BETA3AR', description='adrenoceptor beta 3', other_designations='adrenergic, beta-3-, receptor|beta-3 adrenoceptor|beta-3 adrenoreceptor', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37962995, end_position_on_the_genomic_accession=37966666, orientation='minus', exon_count=2, OMIM=109691, no_hdr0=''),
  156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=156, CurrentID=0, Status='live', Symbol='ADRBK1', Aliases='BARK1, BETA-ARK1, GRK2', description='adrenergic, beta, receptor kinase 1', other_designations='G-protein coupled receptor kinase 2|adrenergic beta receptor kinase 1|beta-ARK-1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67266400, end_position_on_the_genomic_accession=67286558, orientation='plus', exon_count=22, OMIM=109635, no_hdr0=''),
  157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157, CurrentID=0, Status='live', Symbol='ADRBK2', Aliases='BARK2, GRK3', description='adrenergic, beta, receptor kinase 2', other_designations='G-protein-coupled receptor kinase 3|beta-ARK-2', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25517856, end_position_on_the_genomic_accession=25729294, orientation='plus', exon_count=25, OMIM=109636, no_hdr0=''),
  158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158, CurrentID=0, Status='live', Symbol='ADSL', Aliases='AMPS, ASASE, ASL', description='adenylosuccinate lyase', other_designations='adenylosuccinase', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40346500, end_position_on_the_genomic_accession=40387408, orientation='plus', exon_count=16, OMIM=608222, no_hdr0=''),
  159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159, CurrentID=0, Status='live', Symbol='ADSS', Aliases='ADEH 2, ADSS', description='adenylosuccinate synthase', other_designations='AMPSase 2|IMP--aspartate ligase 2|L-type adenylosuccinate synthetase|adSS 2|adenylosuccinate synthetase (Ade(-)H-complementing)|adenylosuccinate synthetase, acidic isozyme|adenylosuccinate synthetase, liver isozyme', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244408492, end_position_on_the_genomic_accession=244452134, orientation='minus', exon_count=14, OMIM=103060, no_hdr0=''),
  160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160, CurrentID=0, Status='live', Symbol='AP2A1', Aliases='ADTAA, AP2-ALPHA, CLAPA1', description='adaptor related protein complex 2, alpha 1 subunit', other_designations='100 kDa coated vesicle protein A|adapter-related protein complex 2 alpha-1 subunit|adapter-related protein complex 2 subunit alpha-1|adaptin, alpha A|adaptor protein complex AP-2 subunit alpha-1|adaptor-related protein complex 2 subunit alpha-1|adaptor-related protein complex 2, alpha 1 subunit|alpha-adaptin A|alpha1-adaptin|clathrin assembly protein complex 2 alpha-A large chain|clathrin-associated/assembly/adaptor protein, large, alpha 1|plasma membrane adaptor HA2/AP2 adaptin alpha A subunit', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49766923, end_position_on_the_genomic_accession=49807112, orientation='plus', exon_count=25, OMIM=601026, no_hdr0=''),
  161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161, CurrentID=0, Status='live', Symbol='AP2A2', Aliases='ADTAB, CLAPA2, HIP-9, HIP9, HYPJ', description='adaptor related protein complex 2, alpha 2 subunit', other_designations='100 kDa coated vesicle protein C|adapter-related protein complex 2 subunit alpha-2|adaptin, alpha B|adaptor-related protein complex 2 subunit alpha-2|adaptor-related protein complex 2, alpha 2 subunit|alpha-adaptin C; Huntingtin interacting protein J|alpha2-adaptin|clathrin assembly protein complex 2 alpha-C large chain|clathrin-associated/assembly/adaptor protein, large, alpha 2|huntingtin yeast partner J|huntingtin-interacting protein 9|plasma membrane adaptor HA2/AP2 adaptin alpha C subunit', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=925809, end_position_on_the_genomic_accession=1012245, orientation='plus', exon_count=25, OMIM=607242, no_hdr0=''),
  162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162, CurrentID=0, Status='live', Symbol='AP1B1', Aliases='ADTB1, AP105A, BAM22, CLAPB2', description='adaptor related protein complex 1 beta 1 subunit', other_designations='Golgi adaptor HA1/AP1 adaptin beta subunit|adapter-related protein complex 1 subunit beta-1|adaptor protein complex AP-1 subunit beta-1|adaptor-related protein complex 1 beta 1 subunit|adaptor-related protein complex 1 subunit beta-1|adaptor-related protein complex 1, beta 1 subunit|beta-1-adaptin|beta-adaptin 1|beta-prime-adaptin|beta1-adaptin|clathrin assembly protein complex 1 beta large chain|plasma membrane adaptor HA2/AP2 adaptor beta subunit', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29327680, end_position_on_the_genomic_accession=29388583, orientation='minus', exon_count=23, OMIM=600157, no_hdr0=''),
  163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163, CurrentID=0, Status='live', Symbol='AP2B1', Aliases='ADTB2, AP105B, AP2-BETA, CLAPB1', description='adaptor related protein complex 2, beta 1 subunit', other_designations='adapter-related protein complex 2 beta subunit|adapter-related protein complex 2 subunit beta|adaptin, beta 2 (beta)|adaptor protein complex AP-2 subunit beta|adaptor-related protein complex 2 subunit beta|adaptor-related protein complex 2, beta 1 subunit|beta-2-adaptin|beta-adaptin|clathrin assembly protein complex 2 beta large chain|clathrin-associated/assembly/adaptor protein, large, beta 1|plasma membrane adaptor HA2/AP2 adaptin beta subunit', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35586899, end_position_on_the_genomic_accession=35726417, orientation='plus', exon_count=26, OMIM=601025, no_hdr0=''),
  164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164, CurrentID=0, Status='live', Symbol='AP1G1', Aliases='ADTG, CLAPG1', description='adaptor related protein complex 1 gamma 1 subunit', other_designations='adapter-related protein complex 1 subunit gamma-1|adaptor protein complex AP-1 subunit gamma-1|adaptor-related protein complex 1 gamma 1 subunit|adaptor-related protein complex 1 subunit gamma-1|adaptor-related protein complex 1, gamma 1 subunit|clathrin assembly protein complex 1 gamma large chain|clathrin assembly protein complex 1 gamma-1 large chain|clathrin-associated/assembly/adaptor protein, large, gamma 1|gamma adaptin|gamma1-adaptin|golgi adaptor HA1/AP1 adaptin gamma subunit|golgi adaptor HA1/AP1 adaptin subunit gamma-1', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71729000, end_position_on_the_genomic_accession=71809073, orientation='minus', exon_count=24, OMIM=603533, no_hdr0=''),
  165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165, CurrentID=0, Status='live', Symbol='AEBP1', Aliases='ACLP', description='AE binding protein 1', other_designations='aortic carboxypeptidase-like protein', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44104320, end_position_on_the_genomic_accession=44114565, orientation='plus', exon_count=21, OMIM=602981, no_hdr0=''),
  166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166, CurrentID=0, Status='live', Symbol='AES', Aliases='AES-1-2, ESP1, GRG, GRG5, Grg-5, TLE5, AES', description='amino-terminal enhancer of split', other_designations='gp130-associated protein GAM|groucho-related protein 5', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3052910, end_position_on_the_genomic_accession=3063107, orientation='minus', exon_count=8, OMIM=600188, no_hdr0=''),
  167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167, CurrentID=0, Status='live', Symbol='CRISP1', Aliases='AEGL1, ARP, CRISP-1, HEL-S-57, HSCRISP1D, HSCRISP1G, HUMARP', description='cysteine-rich secretory protein 1', other_designations='AEG-like protein|AEG-related protein|acidic epididymal glycoprotein homolog|acidic epididymal glycoprotein-like 1|cysteine-rich secretory protein-1 delta|epididymis secretory protein Li 57', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49834257, end_position_on_the_genomic_accession=49877096, orientation='minus', exon_count=10, OMIM=601193, no_hdr0=''),
  169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169, CurrentID=0, Status='live', Symbol='AF8T', Aliases='', description='AF8 temperature sensitivity complementing', other_designations='cell cycle controller G1|tsAF8 complementing', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=116950, no_hdr0=''),
  170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170, CurrentID=0, Status='live', Symbol='AFA', Aliases='', description='ankyloblepharon filiforme adnatum', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=106250, no_hdr0=''),
  171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171, CurrentID=0, Status='live', Symbol='AFD1', Aliases='AFDN', description='acrofacial dysostosis 1, Nager type', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=154400, no_hdr0=''),
  172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=172, CurrentID=0, Status='live', Symbol='AFG3L1P', Aliases='AFG3, AFG3L1', description='AFG3-like AAA ATPase 1, pseudogene', other_designations='AFG3 (ATPase family gene 3, yeast)-like 1|AFG3 ATPase family gene 3-like 1, pseudogene|AFG3 ATPase family member 3-like 1, pseudogene', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89972580, end_position_on_the_genomic_accession=90000787, orientation='plus', exon_count=14, OMIM=603020, no_hdr0=''),
  173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=173, CurrentID=0, Status='live', Symbol='AFM', Aliases='ALB2, ALBA, ALF', description='afamin', other_designations='alpha-Alb|alpha-albumin', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73481745, end_position_on_the_genomic_accession=73504001, orientation='plus', exon_count=15, OMIM=104145, no_hdr0=''),
  174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=174, CurrentID=0, Status='live', Symbol='AFP', Aliases='AFPD, FETA, HPAFP', description='alpha-fetoprotein', other_designations='alpha-1-fetoprotein|alpha-fetoglobulin', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73423104, end_position_on_the_genomic_accession=73455785, orientation='plus', exon_count=16, OMIM=104150, no_hdr0=''),
  175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=175, CurrentID=0, Status='live', Symbol='AGA', Aliases='AGU, ASRG, GA', description='aspartylglucosaminidase', other_designations='N4-(N-acetyl-beta-glucosaminyl)-L-asparagine amidase|aspartylglucosylamine deaspartylase|glycosylasparaginase', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177430774, end_position_on_the_genomic_accession=177442503, orientation='minus', exon_count=9, OMIM=613228, no_hdr0=''),
  176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=176, CurrentID=0, Status='live', Symbol='ACAN', Aliases='AGC1, AGCAN, CSPG1, CSPGCP, MSK16, SEDK', description='aggrecan', other_designations='cartilage-specific proteoglycan core protein|chondroitin sulfate proteoglycan core protein 1|large aggregating proteoglycan', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88803442, end_position_on_the_genomic_accession=88875354, orientation='plus', exon_count=19, OMIM=155760, no_hdr0=''),
  177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=177, CurrentID=0, Status='live', Symbol='AGER', Aliases='RAGE', description='advanced glycosylation end product-specific receptor', other_designations='RAGE isoform NtRAGE-delta|RAGE isoform sRAGE-delta|receptor for advanced glycation end-products variant 20', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32180968, end_position_on_the_genomic_accession=32184322, orientation='minus', exon_count=11, OMIM=600214, no_hdr0=''),
  178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=178, CurrentID=0, Status='live', Symbol='AGL', Aliases='GDE', description='amylo-alpha-1, 6-glucosidase, 4-alpha-glucanotransferase', other_designations='amylo-1, 6-glucosidase, 4-alpha-glucanotransferase|amylo-alpha-1, 6-glucosidase, 4-alpha-glucanotransferaseprovided|glycogen debrancher|glycogen debranching protein', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99850077, end_position_on_the_genomic_accession=99924023, orientation='plus', exon_count=36, OMIM=610860, no_hdr0=''),
  179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=179, CurrentID=0, Status='live', Symbol='AGMX2', Aliases='IMD6, XLA2', description='agammaglobulinemia, X-linked 2 (with growth hormone deficiency)', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300310, no_hdr0=''),
  181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=181, CurrentID=0, Status='live', Symbol='AGRP', Aliases='AGRT, ART, ASIP2', description='agouti related neuropeptide', other_designations='agouti related protein homolog', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67482571, end_position_on_the_genomic_accession=67484083, orientation='minus', exon_count=5, OMIM=602311, no_hdr0=''),
  182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=182, CurrentID=0, Status='live', Symbol='JAG1', Aliases='AGS, AHD, AWS, CD339, HJ1, JAGL1', description='jagged 1', other_designations='', map_location='20p12.1-p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10637684, end_position_on_the_genomic_accession=10674046, orientation='minus', exon_count=26, OMIM=601920, no_hdr0=''),
  183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=183, CurrentID=0, Status='live', Symbol='AGT', Aliases='ANHU, SERPINA8', description='angiotensinogen (serpin peptidase inhibitor, clade A, member 8)', other_designations='alpha-1 antiproteinase, antitrypsin|angiotensin I|angiotensin II|pre-angiotensinogen|serine (or cysteine) proteinase inhibitor|serpin A8', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230702523, end_position_on_the_genomic_accession=230714590, orientation='minus', exon_count=5, OMIM=106150, no_hdr0=''),
  185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=185, CurrentID=0, Status='live', Symbol='AGTR1', Aliases='AG2SB, AT1, AT1AR, AT1B, AT1BR, AT1R, AT2R1, HAT1R, AGTR1', description='angiotensin II receptor, type 1', other_designations='type-1B angiotensin II receptor', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148697871, end_position_on_the_genomic_accession=148743003, orientation='plus', exon_count=5, OMIM=106165, no_hdr0=''),
  186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=186, CurrentID=0, Status='live', Symbol='AGTR2', Aliases='AT2, ATGR2, MRX88', description='angiotensin II receptor, type 2', other_designations='angiotensin II type-2 receptor', map_location='Xq22-q23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116170705, end_position_on_the_genomic_accession=116174972, orientation='plus', exon_count=3, OMIM=300034, no_hdr0=''),
  187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=187, CurrentID=0, Status='live', Symbol='APLNR', Aliases='AGTRL1, APJ, APJR, HG11', description='apelin receptor', other_designations='APJ (apelin) receptor|APJ receptor|G protein-coupled receptor APJ|G-protein coupled receptor APJ|G-protein coupled receptor HG11|HG11 orphan receptor|angiotensin II receptor-like 1|angiotensin receptor-like 1', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57233577, end_position_on_the_genomic_accession=57237453, orientation='minus', exon_count=1, OMIM=600052, no_hdr0=''),
  101056700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101056700, CurrentID=0, Status='live', Symbol='NADK2-AS1', Aliases='NADKD1-AS1', description='NADK2 antisense RNA 1', other_designations='NADKD1 antisense RNA 1', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=189, CurrentID=0, Status='live', Symbol='AGXT', Aliases='AGT, AGT11, PH1, SPAT, SPT, TLH6, AGXT', description='alanine-glyoxylate aminotransferase', other_designations='L-alanine: glyoxylate aminotransferase 1|alanine--glyoxylate aminotransferase|hepatic peroxisomal alanine:glyoxylate aminotransferase|serine-pyruvate aminotransferase|serine:pyruvate aminotransferase', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240868745, end_position_on_the_genomic_accession=240879119, orientation='plus', exon_count=11, OMIM=604285, no_hdr0=''),
  190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=190, CurrentID=0, Status='live', Symbol='NR0B1', Aliases='AHC, AHCH, AHX, DAX-1, DAX1, DSS, GTD, HHG, NROB1, SRXY2', description='nuclear receptor subfamily 0 group B member 1', other_designations='DSS-AHC critical region on the X chromosome protein 1|nuclear hormone receptor|nuclear receptor DAX-1|nuclear receptor subfamily 0, group B, member 1', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30304422, end_position_on_the_genomic_accession=30309378, orientation='minus', exon_count=2, OMIM=300473, no_hdr0=''),
  191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=191, CurrentID=0, Status='live', Symbol='AHCY', Aliases='SAHH, adoHcyase', description='adenosylhomocysteinase', other_designations='S-adenosyl-L-homocysteine hydrolase|S-adenosylhomocysteine hydrolase', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34234840, end_position_on_the_genomic_accession=34311976, orientation='minus', exon_count=17, OMIM=180960, no_hdr0=''),
  192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192, CurrentID=0, Status='live', Symbol='AIC', Aliases='', description='Aicardi syndrome', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=304050, no_hdr0=''),
  196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196, CurrentID=0, Status='live', Symbol='AHR', Aliases='bHLHe76', description='aryl hydrocarbon receptor', other_designations='AH-receptor|ah receptor|aromatic hydrocarbon receptor|class E basic helix-loop-helix protein 76', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17298652, end_position_on_the_genomic_accession=17346151, orientation='plus', exon_count=11, OMIM=600253, no_hdr0=''),
  197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197, CurrentID=0, Status='live', Symbol='AHSG', Aliases='A2HS, AHS, FETUA, HSGA', description='alpha-2-HS-glycoprotein', other_designations='alpha-2-Z-globulin|ba-alpha-2-glycoprotein|fetuin-A', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186613061, end_position_on_the_genomic_accession=186621318, orientation='plus', exon_count=7, OMIM=138680, no_hdr0=''),
  199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199, CurrentID=0, Status='live', Symbol='AIF1', Aliases='AIF-1, IBA1, IRT-1, IRT1', description='allograft inflammatory factor 1', other_designations='allograft inflammatory factor-1 splice variant Hara-1|interferon gamma responsive transcript|ionized calcium-binding adapter molecule 1|protein G1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31615192, end_position_on_the_genomic_accession=31617021, orientation='plus', exon_count=5, OMIM=601833, no_hdr0=''),
  201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201, CurrentID=0, Status='live', Symbol='AIH3', Aliases='', description='amelogenesis imperfecta 3, hypomaturation or hypoplastic type', other_designations='', map_location='Xq22-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301201, no_hdr0=''),
  202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202, CurrentID=0, Status='live', Symbol='AIM1', Aliases='CRYBG1, ST4', description='absent in melanoma 1', other_designations='beta-gamma crystallin domain containing 1|beta/gamma crystallin domain-containing protein 1|suppression of tumorigenicity 4 (malignant melanoma)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106360808, end_position_on_the_genomic_accession=106570451, orientation='plus', exon_count=22, OMIM=601797, no_hdr0=''),
  203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203, CurrentID=0, Status='live', Symbol='AK1', Aliases='HTL-S-58j', description='adenylate kinase 1', other_designations='ATP-AMP transphosphorylase 1|ATP:AMP phosphotransferase|adenylate monophosphate kinase|myokinase|testis secretory sperm binding protein Li 58j', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127866480, end_position_on_the_genomic_accession=127877743, orientation='minus', exon_count=9, OMIM=103000, no_hdr0=''),
  204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204, CurrentID=0, Status='live', Symbol='AK2', Aliases='ADK2, AK 2', description='adenylate kinase 2', other_designations='AK 2|ATP-AMP transphosphorylase 2|ATP:AMP phosphotransferase|adenylate kinase isoenzyme 2, mitochondrial|adenylate monophosphate kinase', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33007940, end_position_on_the_genomic_accession=33036911, orientation='minus', exon_count=8, OMIM=103020, no_hdr0=''),
  205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205, CurrentID=0, Status='live', Symbol='AK4', Aliases='AK 4, AK3, AK3L1, AK3L2', description='adenylate kinase 4', other_designations='AK 4|ATP-AMP transphosphorylase|GTP:AMP phosphotransferase AK4, mitochondrial|adenylate kinase 3-like 1|adenylate kinase isoenzyme 4, mitochondrial|mitochondrial adenylate kinase-3|nucleoside-triphosphate-adenylate kinase', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65147549, end_position_on_the_genomic_accession=65232145, orientation='plus', exon_count=7, OMIM=103030, no_hdr0=''),
  206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=206, CurrentID=0, Status='live', Symbol='AK4P1', Aliases='AK3P1', description='adenylate kinase 4 pseudogene 1', other_designations='adenylate kinase 3 pseudogene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31345314, end_position_on_the_genomic_accession=31347501, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207, CurrentID=0, Status='live', Symbol='AKT1', Aliases='AKT, CWS6, PKB, PKB-ALPHA, PRKBA, RAC, RAC-ALPHA', description='v-akt murine thymoma viral oncogene homolog 1', other_designations='AKT1m|PKB alpha|RAC-PK-alpha|protein kinase B alpha|proto-oncogene c-Akt|rac protein kinase alpha|v-akt murine thymoma viral oncogene-like protein 1', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104769349, end_position_on_the_genomic_accession=104795743, orientation='minus', exon_count=17, OMIM=164730, no_hdr0=''),
  208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=208, CurrentID=0, Status='live', Symbol='AKT2', Aliases='HIHGHH, PKBB, PKBBETA, PRKBB, RAC-BETA', description='v-akt murine thymoma viral oncogene homolog 2', other_designations='PKB beta|RAC-PK-beta|murine thymoma viral (v-akt) homolog-2|protein kinase Akt-2|protein kinase B beta|rac protein kinase beta', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40230317, end_position_on_the_genomic_accession=40285531, orientation='minus', exon_count=23, OMIM=164731, no_hdr0=''),
  210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=210, CurrentID=0, Status='live', Symbol='ALAD', Aliases='ALADH, PBGS', description='aminolevulinate dehydratase', other_designations='aminolevulinate, delta-, dehydratase|porphobilinogen synthase', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113386312, end_position_on_the_genomic_accession=113401338, orientation='minus', exon_count=16, OMIM=125270, no_hdr0=''),
  211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=211, CurrentID=0, Status='live', Symbol='ALAS1', Aliases='ALAS, ALAS-H, ALAS3, ALASH, MIG4', description="5'-aminolevulinate synthase 1", other_designations='5-aminolevulinic acid synthase 1|aminolevulinate, delta-, synthase 1|delta-ALA synthase 1|delta-aminolevulinate synthase 1|migration-inducing protein 4', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52198083, end_position_on_the_genomic_accession=52214327, orientation='plus', exon_count=12, OMIM=125290, no_hdr0=''),
  212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=212, CurrentID=0, Status='live', Symbol='ALAS2', Aliases='ALAS-E, ALASE, ANH1, ASB, XLDPP, XLEPP, XLSA', description="5'-aminolevulinate synthase 2", other_designations='5-aminolevulinic acid synthase 2|aminolevulinate, delta-, synthase 2|delta-ALA synthase 2|delta-ALA synthetase|delta-aminolevulinate synthase 2', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55009055, end_position_on_the_genomic_accession=55031064, orientation='minus', exon_count=12, OMIM=301300, no_hdr0=''),
  213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=213, CurrentID=0, Status='live', Symbol='ALB', Aliases='ANALBA, FDAH, PRO0883, PRO0903, PRO1341', description='albumin', other_designations='albumin (32 AA)|albumin (AA 34)|cell growth inhibiting protein 42|growth-inhibiting protein 20', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73404255, end_position_on_the_genomic_accession=73421412, orientation='plus', exon_count=15, OMIM=103600, no_hdr0=''),
  214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=214, CurrentID=0, Status='live', Symbol='ALCAM', Aliases='CD166, MEMD', description='activated leukocyte cell adhesion molecule', other_designations='', map_location='3q13.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=105366713, end_position_on_the_genomic_accession=105576913, orientation='plus', exon_count=16, OMIM=601662, no_hdr0=''),
  215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=215, CurrentID=0, Status='live', Symbol='ABCD1', Aliases='ABC42, ALD, ALDP, AMN', description='ATP binding cassette subfamily D member 1', other_designations='ATP-binding cassette, sub-family D (ALD), member 1|adrenoleukodystrophy protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153724851, end_position_on_the_genomic_accession=153744762, orientation='plus', exon_count=11, OMIM=300371, no_hdr0=''),
  216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=216, CurrentID=0, Status='live', Symbol='ALDH1A1', Aliases='ALDC, ALDH-E1, ALDH1, ALDH11, HEL-9, HEL-S-53e, HEL12, PUMB1, RALDH1', description='aldehyde dehydrogenase 1 family member A1', other_designations='ALDH class 1|ALHDII|RALDH 1|acetaldehyde dehydrogenase 1|aldehyde dehydrogenase 1 family, member A1|aldehyde dehydrogenase 1, soluble|aldehyde dehydrogenase, liver cytosolic|epididymis luminal protein 12|epididymis luminal protein 9|epididymis secretory sperm binding protein Li 53e|retinaldehyde dehydrogenase 1', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72900662, end_position_on_the_genomic_accession=72953317, orientation='minus', exon_count=13, OMIM=100640, no_hdr0=''),
  217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=217, CurrentID=0, Status='live', Symbol='ALDH2', Aliases='ALDH-E2, ALDHI, ALDM', description='aldehyde dehydrogenase 2 family (mitochondrial)', other_designations='ALDH class 2|acetaldehyde dehydrogenase 2|liver mitochondrial ALDH|nucleus-encoded mitochondrial aldehyde dehydrogenase 2', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111766887, end_position_on_the_genomic_accession=111809985, orientation='plus', exon_count=13, OMIM=100650, no_hdr0=''),
  218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=218, CurrentID=0, Status='live', Symbol='ALDH3A1', Aliases='ALDH3, ALDHIII', description='aldehyde dehydrogenase 3 family member A1', other_designations='aldehyde dehydrogenase 3 family, member A1|aldehyde dehydrogenase isozyme 3|aldehyde dehydrogenase type III|stomach aldehyde dehydrogenase', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19737984, end_position_on_the_genomic_accession=19748433, orientation='minus', exon_count=13, OMIM=100660, no_hdr0=''),
  219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219, CurrentID=0, Status='live', Symbol='ALDH1B1', Aliases='ALDH5, ALDHX', description='aldehyde dehydrogenase 1 family member B1', other_designations='ALDH class 2|acetaldehyde dehydrogenase 5|aldehyde dehydrogenase 1 family, member B1|aldehyde dehydrogenase 5', map_location='9p11.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38392664, end_position_on_the_genomic_accession=38398665, orientation='plus', exon_count=3, OMIM=100670, no_hdr0=''),
  220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220, CurrentID=0, Status='live', Symbol='ALDH1A3', Aliases='ALDH1A6, ALDH6, MCOP8, RALDH3', description='aldehyde dehydrogenase 1 family member A3', other_designations='acetaldehyde dehydrogenase 6|aldehyde dehydrogenase 1 family, member A3|aldehyde dehydrogenase 6|retinaldehyde dehydrogenase 3', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100879692, end_position_on_the_genomic_accession=100916626, orientation='plus', exon_count=13, OMIM=600463, no_hdr0=''),
  221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221, CurrentID=0, Status='live', Symbol='ALDH3B1', Aliases='ALDH4, ALDH7', description='aldehyde dehydrogenase 3 family member B1', other_designations='aldehyde dehydrogenase 3 family, member B1|aldehyde dehydrogenase 3B1|aldehyde dehydrogenase 7', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68008547, end_position_on_the_genomic_accession=68029282, orientation='plus', exon_count=11, OMIM=600466, no_hdr0=''),
  222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222, CurrentID=0, Status='live', Symbol='ALDH3B2', Aliases='ALDH8', description='aldehyde dehydrogenase 3 family member B2', other_designations='acetaldehyde dehydrogenase 8|aldehyde dehydrogenase 3 family, member B2|aldehyde dehydrogenase 8', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67662155, end_position_on_the_genomic_accession=67681214, orientation='minus', exon_count=11, OMIM=601917, no_hdr0=''),
  223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=223, CurrentID=0, Status='live', Symbol='ALDH9A1', Aliases='ALDH4, ALDH7, ALDH9, E3, TMABADH', description='aldehyde dehydrogenase 9 family member A1', other_designations='R-aminobutyraldehyde dehydrogenase|aldehyde dehydrogenase (NAD+)|aldehyde dehydrogenase 9 family, member A1|aldehyde dehydrogenase E3 isozyme|aldehyde dehydrogenase family 9 member A1|gamma-aminobutyraldehyde dehydrogenase', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165662212, end_position_on_the_genomic_accession=165698663, orientation='minus', exon_count=11, OMIM=602733, no_hdr0=''),
  224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=224, CurrentID=0, Status='live', Symbol='ALDH3A2', Aliases='ALDH10, FALDH, SLS', description='aldehyde dehydrogenase 3 family member A2', other_designations='aldehyde dehydrogenase 10|aldehyde dehydrogenase 3 family, member A2|aldehyde dehydrogenase family 3 member A2|microsomal aldehyde dehydrogenase', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19648136, end_position_on_the_genomic_accession=19677596, orientation='plus', exon_count=13, OMIM=609523, no_hdr0=''),
  225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=225, CurrentID=0, Status='live', Symbol='ABCD2', Aliases='ABC39, ALDL1, ALDR, ALDRP, hALDR', description='ATP binding cassette subfamily D member 2', other_designations='ATP-binding cassette, sub-family D (ALD), member 2|adrenoleukodystrophy-like 1|adrenoleukodystrophy-related protein', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39550353, end_position_on_the_genomic_accession=39651328, orientation='minus', exon_count=12, OMIM=601081, no_hdr0=''),
  226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=226, CurrentID=0, Status='live', Symbol='ALDOA', Aliases='ALDA, GSD12, HEL-S-87p', description='aldolase, fructose-bisphosphate A', other_designations='aldolase A, fructose-bisphosphate|epididymis secretory sperm binding protein Li 87p|fructose-1,6-bisphosphate triosephosphate-lyase|lung cancer antigen NY-LU-1|muscle-type aldolase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30053090, end_position_on_the_genomic_accession=30070420, orientation='plus', exon_count=17, OMIM=103850, no_hdr0=''),
  228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=228, CurrentID=0, Status='live', Symbol='ALDOAP2', Aliases='', description='aldolase, fructose-bisphosphate A pseudogene 2', other_designations='aldolase A, fructose-bisphosphate pseudogene 2', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125666680, end_position_on_the_genomic_accession=125668056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=229, CurrentID=0, Status='live', Symbol='ALDOB', Aliases='ALDB, ALDO2', description='aldolase, fructose-bisphosphate B', other_designations='aldolase 2|aldolase B, fructose-bisphosphatase|aldolase B, fructose-bisphosphate|liver-type aldolase', map_location='9q21.3-q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101420560, end_position_on_the_genomic_accession=101435780, orientation='minus', exon_count=9, OMIM=612724, no_hdr0=''),
  230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=230, CurrentID=0, Status='live', Symbol='ALDOC', Aliases='ALDC', description='aldolase, fructose-bisphosphate C', other_designations='aldolase 3|aldolase C, fructose-bisphosphate|brain-type aldolase|fructoaldolase C|fructose-1,6-biphosphate triosephosphate lyase', map_location='17cen-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28573115, end_position_on_the_genomic_accession=28576948, orientation='minus', exon_count=10, OMIM=103870, no_hdr0=''),
  231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=231, CurrentID=0, Status='live', Symbol='AKR1B1', Aliases='ADR, ALDR1, ALR2, AR', description='aldo-keto reductase family 1, member B1 (aldose reductase)', other_designations='Lii5-2 CTCL tumor antigen|aldehyde reductase 1|aldo-keto reductase family 1 member B1|low Km aldose reductase', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134442351, end_position_on_the_genomic_accession=134459136, orientation='minus', exon_count=10, OMIM=103880, no_hdr0=''),
  152956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152956, CurrentID=0, Status='live', Symbol='FLJ12425', Aliases='', description='uncharacterized protein FLJ12425', other_designations='', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=236, CurrentID=0, Status='live', Symbol='AKR1B1P2', Aliases='ALDRP', description='aldo-keto reductase family 1, member B1 pseudogene 2', other_designations='aldehyde reductase (aldose reductase) pseudogene|aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='3p12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74135927, end_position_on_the_genomic_accession=74137588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=238, CurrentID=0, Status='live', Symbol='ALK', Aliases='CD246, NBLST3', description='anaplastic lymphoma receptor tyrosine kinase', other_designations='CD246 antigen|mutant anaplastic lymphoma kinase', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29192774, end_position_on_the_genomic_accession=29921611, orientation='minus', exon_count=29, OMIM=105590, no_hdr0=''),
  239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=239, CurrentID=0, Status='live', Symbol='ALOX12', Aliases='12-LOX, 12S-LOX, LOG12', description='arachidonate 12-lipoxygenase', other_designations='12(S)-lipoxygenase|12S-lipoxygenase|lipoxin synthase 12-LO|platelet 12-LOX|platelet-type 12-lipoxygenase|platelet-type lipoxygenase 12', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6994196, end_position_on_the_genomic_accession=7010754, orientation='plus', exon_count=14, OMIM=152391, no_hdr0=''),
  240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=240, CurrentID=0, Status='live', Symbol='ALOX5', Aliases='5-LO, 5-LOX, 5LPG, LOG5', description='arachidonate 5-lipoxygenase', other_designations='LOX-5|arachidonic 5-lipoxygenase alpha-10 isoform|arachidonic 5-lipoxygenase delta-10-13 isoform|arachidonic 5-lipoxygenase delta-13 isoform|arachidonic 5-lipoxygenase delta-p10 isoform|arachidonic acid 5-lipoxygenase|leukotriene A4 synthase', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45374176, end_position_on_the_genomic_accession=45446121, orientation='plus', exon_count=14, OMIM=152390, no_hdr0=''),
  241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=241, CurrentID=0, Status='live', Symbol='ALOX5AP', Aliases='FLAP', description='arachidonate 5-lipoxygenase-activating protein', other_designations='MK-886-binding protein', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30713478, end_position_on_the_genomic_accession=30764428, orientation='plus', exon_count=8, OMIM=603700, no_hdr0=''),
  242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=242, CurrentID=0, Status='live', Symbol='ALOX12B', Aliases='12R-LOX, ARCI2', description='arachidonate 12-lipoxygenase, 12R type', other_designations='12R-lipoxygenase|epidermis-type lipoxygenase 12', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8072636, end_position_on_the_genomic_accession=8087703, orientation='minus', exon_count=15, OMIM=603741, no_hdr0=''),
  243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=243, CurrentID=0, Status='live', Symbol='ALOX12P1', Aliases='ALOX12P', description='arachidonate 12-lipoxygenase pseudogene 1', other_designations='', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30529673, end_position_on_the_genomic_accession=30542433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245, CurrentID=0, Status='live', Symbol='ALOX12P2', Aliases='ALOX12E', description='arachidonate 12-lipoxygenase pseudogene 2', other_designations='12-lipoxygenase-related protein|hair and skin epidermal-type 12-lipoxygenase-related pseudogene', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6853576, end_position_on_the_genomic_accession=6900349, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246, CurrentID=0, Status='live', Symbol='ALOX15', Aliases='12-LOX, 15-LOX-1, 15LOX-1', description='arachidonate 15-lipoxygenase', other_designations='12/15-lipoxygenase|15-LOX|15-lipooxygenase-1|arachidonate 12-lipoxygenase, leukocyte-type|arachidonate omega-6 lipoxygenase', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4630919, end_position_on_the_genomic_accession=4641665, orientation='minus', exon_count=14, OMIM=152392, no_hdr0=''),
  247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=247, CurrentID=0, Status='live', Symbol='ALOX15B', Aliases='15-LOX-2', description='arachidonate 15-lipoxygenase, type B', other_designations='15-LOX-B|15S-lipoxygenase|arachidonate 15-lipoxygenase 2|arachidonate 15-lipoxygenase type II|arachidonate 15-lipoxygenase, second type|arachidonate omega(6) lipoxygenase|linoleate 13-lipoxygenase 15-LOb', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8039040, end_position_on_the_genomic_accession=8049134, orientation='plus', exon_count=14, OMIM=603697, no_hdr0=''),
  248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=248, CurrentID=0, Status='live', Symbol='ALPI', Aliases='IAP', description='alkaline phosphatase, intestinal', other_designations='Kasahara isozyme|alkaline phosphomonoesterase|glycerophosphatase|intestinal alkaline phosphatase', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232456112, end_position_on_the_genomic_accession=232460745, orientation='plus', exon_count=11, OMIM=171740, no_hdr0=''),
  249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=249, CurrentID=0, Status='live', Symbol='ALPL', Aliases='AP-TNAP, APTNAP, HOPS, TNAP, TNSALP', description='alkaline phosphatase, liver/bone/kidney', other_designations='alkaline phosphatase liver/bone/kidney isozyme|liver/bone/kidney-type alkaline phosphatase|tissue-nonspecific ALP', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21508982, end_position_on_the_genomic_accession=21578412, orientation='plus', exon_count=15, OMIM=171760, no_hdr0=''),
  250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=250, CurrentID=0, Status='live', Symbol='ALPP', Aliases='ALP, PALP, PLAP, PLAP-1', description='alkaline phosphatase, placental', other_designations='alkaline phosphatase Regan isozyme|alkaline phosphomonoesterase|glycerophosphatase|placental alkaline phosphatase 1', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232378534, end_position_on_the_genomic_accession=232382889, orientation='plus', exon_count=11, OMIM=171800, no_hdr0=''),
  251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=251, CurrentID=0, Status='live', Symbol='ALPPL2', Aliases='ALPG, ALPPL, GCAP', description='alkaline phosphatase, placental like 2', other_designations='ALP-1|Nagao isozyme|alkaline phosphatase Nagao isozyme|alkaline phosphatase, placental-like 2|germ cell alkaline phosphatase|placental-like alkaline phosphatase|testicular and thymus alkaline phosphatase', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232406842, end_position_on_the_genomic_accession=232410714, orientation='plus', exon_count=11, OMIM=171810, no_hdr0=''),
  253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253, CurrentID=0, Status='live', Symbol='ALS3', Aliases='ALS6', description='amyotrophic lateral sclerosis 3 (autosomal dominant)', other_designations='', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606640, no_hdr0=''),
  255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255, CurrentID=0, Status='live', Symbol='ALS5', Aliases='', description='amyotrophic lateral sclerosis 5', other_designations='', map_location='15q15.1-q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602099, no_hdr0=''),
  257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257, CurrentID=0, Status='live', Symbol='ALX3', Aliases='FND, FND1', description='ALX homeobox 3', other_designations='aristaless-like homeobox 3|frontonasal dysplasia|proline-rich transcription factor ALX3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110060375, end_position_on_the_genomic_accession=110070700, orientation='minus', exon_count=4, OMIM=606014, no_hdr0=''),
  258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=258, CurrentID=0, Status='live', Symbol='AMBN', Aliases='AI1F', description='ameloblastin', other_designations='ameloblastin (enamel matrix protein)|enamel matrix protein', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70592258, end_position_on_the_genomic_accession=70607288, orientation='plus', exon_count=13, OMIM=601259, no_hdr0=''),
  259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259, CurrentID=0, Status='live', Symbol='AMBP', Aliases='A1M, EDC1, HCP, HI30, IATIL, ITI, ITIL, ITILC, UTI', description='alpha-1-microglobulin/bikunin precursor', other_designations='bikunin|complex-forming glycoprotein heterogeneous in charge|growth-inhibiting protein 19|inter-alpha-trypsin inhibitor light chain|protein HC|trypstatin|uristatin|uronic-acid-rich protein', map_location='9q32-q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114060127, end_position_on_the_genomic_accession=114078472, orientation='minus', exon_count=10, OMIM=176870, no_hdr0=''),
  261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261, CurrentID=0, Status='live', Symbol='AMCN', Aliases='AMC, AMC11, AMCN', description='arthrogryposis multiplex congenita, neurogenic', other_designations='Arthrogryposis multiplex congenital, neurogenic', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=208100, no_hdr0=''),
  262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=262, CurrentID=0, Status='live', Symbol='AMD1', Aliases='ADOMETDC, AMD, SAMDC', description='adenosylmethionine decarboxylase 1', other_designations='S-adenosylmethionine decarboxylase 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110814621, end_position_on_the_genomic_accession=110895713, orientation='plus', exon_count=10, OMIM=180980, no_hdr0=''),
  263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=263, CurrentID=0, Status='live', Symbol='AMD1P2', Aliases='AMD, AMD2, AMDP1, AMDPX, AMDPY', description='adenosylmethionine decarboxylase 1 pseudogene 2', other_designations='S-adenosylmethionine decarboxylase 2 (pseudogene)|S-adenosylmethionine decarboxylase pseudogene 1|adenosylmethionine decarboxylase pseudogene 1', map_location='Xq28 and Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57014613, end_position_on_the_genomic_accession=57016181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=265, CurrentID=0, Status='live', Symbol='AMELX', Aliases='AI1E, AIH1, ALGN, AMG, AMGL, AMGX', description='amelogenin, X-linked', other_designations='amelogenesis imperfecta 1|amelogenin (X chromosome, amelogenesis imperfecta 1)|amelogenin (amelogenesis imperfecta 1, X-linked)', map_location='Xp22.31-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11292578, end_position_on_the_genomic_accession=11300761, orientation='plus', exon_count=8, OMIM=300391, no_hdr0=''),
  266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266, CurrentID=0, Status='live', Symbol='AMELY', Aliases='AMGL, AMGY', description='amelogenin, Y-linked', other_designations='amelogenin (Y chromosome)', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6865918, end_position_on_the_genomic_accession=6911937, orientation='minus', exon_count=8, OMIM=410000, no_hdr0=''),
  267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267, CurrentID=0, Status='live', Symbol='AMFR', Aliases='GP78, RNF45', description='autocrine motility factor receptor, E3 ubiquitin protein ligase', other_designations='RING finger protein 45', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56361452, end_position_on_the_genomic_accession=56425538, orientation='minus', exon_count=16, OMIM=603243, no_hdr0=''),
  268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=268, CurrentID=0, Status='live', Symbol='AMH', Aliases='MIF, MIS', description='anti-Mullerian hormone', other_designations='Mullerian inhibiting factor|Mullerian inhibiting substance|anti-Muellerian hormone|muellerian-inhibiting substance', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2249114, end_position_on_the_genomic_accession=2252073, orientation='plus', exon_count=5, OMIM=600957, no_hdr0=''),
  269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=269, CurrentID=0, Status='live', Symbol='AMHR2', Aliases='AMHR, MISR2, MISRII, MRII', description='anti-Mullerian hormone receptor, type II', other_designations='AMH type II receptor|MIS type II receptor|Muellerian inhibiting substance type II receptor|Mullerian inhibiting substance type II receptor|anti-Muellerian hormone type II receptor', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53423855, end_position_on_the_genomic_accession=53431534, orientation='plus', exon_count=13, OMIM=600956, no_hdr0=''),
  270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=270, CurrentID=0, Status='live', Symbol='AMPD1', Aliases='MAD, MADA, MMDD', description='adenosine monophosphate deaminase 1', other_designations='AMPD|adenosine monophosphate deaminase 1 (isoform M)|adenosine monophosphate deaminase-1 (muscle)|myoadenylate deaminase|skeletal muscle AMPD', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114673098, end_position_on_the_genomic_accession=114695618, orientation='minus', exon_count=16, OMIM=102770, no_hdr0=''),
  271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=271, CurrentID=0, Status='live', Symbol='AMPD2', Aliases='PCH9, SPG63', description='adenosine monophosphate deaminase 2', other_designations='AMPD|adenosine monophosphate deaminase 2 (isoform L)', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109619659, end_position_on_the_genomic_accession=109632055, orientation='plus', exon_count=20, OMIM=102771, no_hdr0=''),
  272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=272, CurrentID=0, Status='live', Symbol='AMPD3', Aliases='', description='adenosine monophosphate deaminase 3', other_designations='AMP aminohydrolase|adenosine monophosphate deaminase (isoform E)|erythrocyte AMP deaminase|erythrocyte type AMP deaminase|erythrocyte-specific AMP deaminase|myoadenylate deaminase', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10450321, end_position_on_the_genomic_accession=10507579, orientation='plus', exon_count=19, OMIM=102772, no_hdr0=''),
  273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=273, CurrentID=0, Status='live', Symbol='AMPH', Aliases='AMPH1', description='amphiphysin', other_designations='amphiphysin (Stiff-Mann syndrome with breast cancer 128kD autoantigen)|amphiphysin I', map_location='7p14-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38383694, end_position_on_the_genomic_accession=38631567, orientation='minus', exon_count=23, OMIM=600418, no_hdr0=''),
  274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=274, CurrentID=0, Status='live', Symbol='BIN1', Aliases='AMPH2, AMPHL, SH3P9', description='bridging integrator 1', other_designations='amphiphysin II|amphiphysin-like protein|box dependant MYC interacting protein 1|box-dependent myc-interacting protein 1', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127048023, end_position_on_the_genomic_accession=127107400, orientation='minus', exon_count=22, OMIM=601248, no_hdr0=''),
  275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=275, CurrentID=0, Status='live', Symbol='AMT', Aliases='GCE, GCST, GCVT, NKH', description='aminomethyltransferase', other_designations='glycine cleavage system T protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49416778, end_position_on_the_genomic_accession=49422678, orientation='minus', exon_count=9, OMIM=238310, no_hdr0=''),
  276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=276, CurrentID=0, Status='live', Symbol='AMY1A', Aliases='AMY1', description='amylase, alpha 1A (salivary)', other_designations='1,4-alpha-D-glucan glucanohydrolase 1|amylase, salivary, alpha-1A|glycogenase|salivary alpha-amylase|salivary amylase alpha 1A', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103655519, end_position_on_the_genomic_accession=103664551, orientation='plus', exon_count=11, OMIM=104700, no_hdr0=''),
  277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=277, CurrentID=0, Status='live', Symbol='AMY1B', Aliases='AMY1', description='amylase, alpha 1B (salivary)', other_designations='1,4-alpha-D-glucan glucanohydrolase 1|amylase, salivary, alpha-1B|glycogenase|salivary alpha-amylase|salivary amylase alpha 1B', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103687415, end_position_on_the_genomic_accession=103696452, orientation='minus', exon_count=12, OMIM=104701, no_hdr0=''),
  278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=278, CurrentID=0, Status='live', Symbol='AMY1C', Aliases='AMY1', description='amylase, alpha 1C (salivary)', other_designations='1,4-alpha-D-glucan glucanohydrolase 1|amylase, salivary, alpha-1C|glycogenase|salivary alpha-amylase|salivary amylase alpha 1C', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103749656, end_position_on_the_genomic_accession=103758692, orientation='plus', exon_count=11, OMIM=104702, no_hdr0=''),
  279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=279, CurrentID=0, Status='live', Symbol='AMY2A', Aliases='AMY2, PA', description='amylase, alpha 2A (pancreatic)', other_designations='1,4-alpha-D-glucan glucanohydrolase|glycogenase|pancreatic amylase alpha 2A', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103617332, end_position_on_the_genomic_accession=103625780, orientation='plus', exon_count=10, OMIM=104650, no_hdr0=''),
  280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280, CurrentID=0, Status='live', Symbol='AMY2B', Aliases='AMY2, AMY3, HXA', description='amylase, alpha 2B (pancreatic)', other_designations='1,4-alpha-D-glucan glucanohydrolase 2B|carcinoid alpha-amylase|glycogenase', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103554644, end_position_on_the_genomic_accession=103579534, orientation='plus', exon_count=12, OMIM=104660, no_hdr0=''),
  281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=281, CurrentID=0, Status='live', Symbol='AMYP1', Aliases='AMY2P', description='amylase, alpha pseudogene 1', other_designations='', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103713720, end_position_on_the_genomic_accession=103719903, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282, CurrentID=0, Status='live', Symbol='ANCR', Aliases='', description='Angelman syndrome chromosome region', other_designations='', map_location='15q11-q12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283, CurrentID=0, Status='live', Symbol='ANG', Aliases='ALS9, HEL168, RAA1, RNASE4, RNASE5', description='angiogenin, ribonuclease, RNase A family, 5', other_designations='RNase 5|epididymis luminal protein 168|ribonuclease 5|ribonuclease A A1', map_location='14q11.1-q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20684177, end_position_on_the_genomic_accession=20694186, orientation='plus', exon_count=3, OMIM=105850, no_hdr0=''),
  284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284, CurrentID=0, Status='live', Symbol='ANGPT1', Aliases='AGP1, AGPT, ANG1', description='angiopoietin 1', other_designations='ANG-1', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107249482, end_position_on_the_genomic_accession=107498026, orientation='minus', exon_count=9, OMIM=601667, no_hdr0=''),
  285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285, CurrentID=0, Status='live', Symbol='ANGPT2', Aliases='AGPT2, ANG2', description='angiopoietin 2', other_designations='ANG-2|Tie2-ligand|angiopoietin-2B|angiopoietin-2a', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6499651, end_position_on_the_genomic_accession=6563263, orientation='minus', exon_count=9, OMIM=601922, no_hdr0=''),
  286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286, CurrentID=0, Status='live', Symbol='ANK1', Aliases='ANK, SPH1, SPH2', description='ankyrin 1', other_designations='ANK-1|ankyrin 1, erythrocytic|ankyrin-R|erythrocyte ankyrin', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41653225, end_position_on_the_genomic_accession=41896762, orientation='minus', exon_count=49, OMIM=612641, no_hdr0=''),
  287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=287, CurrentID=0, Status='live', Symbol='ANK2', Aliases='ANK-2, LQT4, brank-2', description='ankyrin 2, neuronal', other_designations='ankyrin B|ankyrin, brain|ankyrin-2, nonerythrocytic|non-erythroid ankyrin', map_location='4q25-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112818083, end_position_on_the_genomic_accession=113383740, orientation='plus', exon_count=55, OMIM=106410, no_hdr0=''),
  288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=288, CurrentID=0, Status='live', Symbol='ANK3', Aliases='ANKYRIN-G, MRT37', description='ankyrin 3, node of Ranvier (ankyrin G)', other_designations='ankyrin G119', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60026298, end_position_on_the_genomic_accession=60733526, orientation='minus', exon_count=60, OMIM=600465, no_hdr0=''),
  289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=289, CurrentID=0, Status='live', Symbol='ANOP1', Aliases='MCOPS4', description='anophthalmos 1 (with mental retardation, without limb anomalies or dental or urogenital abnormalities)', other_designations='Anophthalmos-1 (with mental retardation but without anomalies)', map_location='Xq27-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301590, no_hdr0=''),
  290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=290, CurrentID=0, Status='live', Symbol='ANPEP', Aliases='APN, CD13, GP150, LAP1, P150, PEPN', description='alanyl (membrane) aminopeptidase', other_designations='AP-M|AP-N|alanyl aminopeptidase|aminopeptidase M|hAPN|microsomal aminopeptidase|myeloid plasma membrane glycoprotein CD13', map_location='15q25-q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89784895, end_position_on_the_genomic_accession=89814854, orientation='minus', exon_count=21, OMIM=151530, no_hdr0=''),
  291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=291, CurrentID=0, Status='live', Symbol='SLC25A4', Aliases='AAC1, ANT, ANT 1, ANT1, MTDPS12, PEO2, PEO3, T1', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 4', other_designations='ADP,ATP carrier protein 1|ADP,ATP carrier protein, heart/skeletal muscle|adenine nucleotide translocator 1 (skeletal muscle)|heart/skeletal muscle ATP/ADP translocator|solute carrier family 25 member 4', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185143263, end_position_on_the_genomic_accession=185150384, orientation='plus', exon_count=4, OMIM=103220, no_hdr0=''),
  292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=292, CurrentID=0, Status='live', Symbol='SLC25A5', Aliases='2F1, AAC2, ANT2, T2, T3', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5', other_designations='ADP,ATP carrier protein, fibroblast isoform|adenine nucleotide translocator 2 (fibroblast)', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119468400, end_position_on_the_genomic_accession=119471396, orientation='plus', exon_count=4, OMIM=300150, no_hdr0=''),
  293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=293, CurrentID=0, Status='live', Symbol='SLC25A6', Aliases='AAC3, ANT, ANT 2, ANT 3, ANT3, ANT3Y', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6', other_designations='ADP,ATP carrier protein|ADP,ATP carrier protein 3|ADP,ATP carrier protein, liver|ADP/ATP translocator of liver|adenine nucleotide translocator 3|solute carrier family 25 member 6', map_location='Xp22.32 and Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1386152, end_position_on_the_genomic_accession=1392146, orientation='minus', exon_count=4, OMIM='300151,403000', no_hdr0=''),
  131368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131368, CurrentID=0, Status='live', Symbol='ZPLD1', Aliases='', description='zona pellucida-like domain containing 1', other_designations='ZP domain-containing protein 1', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102357481, end_position_on_the_genomic_accession=102479841, orientation='plus', exon_count=14, OMIM=615915, no_hdr0=''),
  105371057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371057, CurrentID=0, Status='live', Symbol='LOC105371057', Aliases='', description='uncharacterized LOC105371057', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3038006, end_position_on_the_genomic_accession=3045919, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=301, CurrentID=0, Status='live', Symbol='ANXA1', Aliases='ANX1, LPC1', description='annexin A1', other_designations='annexin I (lipocortin I)|annexin-1|calpactin II|calpactin-2|chromobindin-9|phospholipase A2 inhibitory protein', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73151731, end_position_on_the_genomic_accession=73170393, orientation='plus', exon_count=15, OMIM=151690, no_hdr0=''),
  302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=302, CurrentID=0, Status='live', Symbol='ANXA2', Aliases='ANX2, ANX2L4, CAL1H, HEL-S-270, LIP2, LPC2, LPC2D, P36, PAP-IV', description='annexin A2', other_designations='annexin II|annexin-2|calpactin I heavy chain|calpactin I heavy polypeptide|calpactin-1 heavy chain|chromobindin 8|chromobindin-8|epididymis secretory protein Li 270|lipocortin II|placental anticoagulant protein IV|protein I', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60347151, end_position_on_the_genomic_accession=60397986, orientation='minus', exon_count=17, OMIM=151740, no_hdr0=''),
  131375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131375, CurrentID=0, Status='live', Symbol='LYZL4', Aliases='LYC4, LYZA', description='lysozyme like 4', other_designations='lysozyme A|lysozyme-4|lysozyme-like 4', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42360768, end_position_on_the_genomic_accession=42410652, orientation='minus', exon_count=6, OMIM=612750, no_hdr0=''),
  304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=304, CurrentID=0, Status='live', Symbol='ANXA2P2', Aliases='ANX2L2, ANX2P2, LPC2B', description='annexin A2 pseudogene 2', other_designations='annexin II (lipocortin II) pseudogene 2', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33624225, end_position_on_the_genomic_accession=33625534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  131377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131377, CurrentID=0, Status='live', Symbol='KLHL40', Aliases='KBTBD5, NEM8, SRYP, SYRP', description='kelch like family member 40', other_designations='kelch repeat and BTB (POZ) domain containing 5|kelch repeat and BTB domain-containing protein 5|kelch-like family member 40|nemaline myopathy type 8|sarcosynapsin', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42685519, end_position_on_the_genomic_accession=42692446, orientation='plus', exon_count=6, OMIM=615340, no_hdr0=''),
  306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=306, CurrentID=0, Status='live', Symbol='ANXA3', Aliases='ANX3', description='annexin A3', other_designations='35-alpha calcimedin|Annexin III (lipocortin III)|PAP-III|annexin III (lipocortin III, 1,2-cyclic-inositol-phosphate phosphodiesterase, placental anticoagulant protein III, calcimedin 35-alpha)|annexin-3|calcimedin 35-alpha|inositol 1,2-cyclic phosphate 2-phosphohydrolase|lipocortin III|placental anticoagulant protein III', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78551588, end_position_on_the_genomic_accession=78610451, orientation='plus', exon_count=15, OMIM=106490, no_hdr0=''),
  307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=307, CurrentID=0, Status='live', Symbol='ANXA4', Aliases='ANX4, HEL-S-274, PIG28, ZAP36', description='annexin A4', other_designations='35-beta calcimedin|P32.5|PAP-II|PP4-X|annexin IV (placental anticoagulant protein II)|annexin-4|carbohydrate-binding protein p33/p41|chromobindin-4|endonexin I|epididymis secretory protein Li 274|lipocortin IV|placental anticoagulant protein II|proliferation-inducing gene 28|proliferation-inducing protein 28|protein II', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69644100, end_position_on_the_genomic_accession=69826464, orientation='plus', exon_count=16, OMIM=106491, no_hdr0=''),
  308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=308, CurrentID=0, Status='live', Symbol='ANXA5', Aliases='ANX5, ENX2, HEL-S-7, PP4, RPRGL3', description='annexin A5', other_designations='CBP-I|PAP-I|VAC-alpha|anchorin CII|annexin V|annexin-5|calphobindin I|endonexin II|epididymis secretory protein Li 7|lipocortin V|placental anticoagulant protein 4|placental anticoagulant protein I|thromboplastin inhibitor|vascular anticoagulant-alpha', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121667997, end_position_on_the_genomic_accession=121696992, orientation='minus', exon_count=13, OMIM=131230, no_hdr0=''),
  309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=309, CurrentID=0, Status='live', Symbol='ANXA6', Aliases='ANX6, CBP68', description='annexin A6', other_designations='67 kDa calelectrin|CPB-II|annexin VI (p68)|annexin-6|calcium-binding protein p68|calelectrin|calphobindin II|calphobindin-II|chromobindin-20|lipocortin VI|p68|p70', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151100706, end_position_on_the_genomic_accession=151157882, orientation='minus', exon_count=27, OMIM=114070, no_hdr0=''),
  310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=310, CurrentID=0, Status='live', Symbol='ANXA7', Aliases='ANX7, SNX, SYNEXIN', description='annexin A7', other_designations='annexin VII|annexin-7', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73375431, end_position_on_the_genomic_accession=73414083, orientation='minus', exon_count=14, OMIM=186360, no_hdr0=''),
  311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=311, CurrentID=0, Status='live', Symbol='ANXA11', Aliases='ANX11, CAP50', description='annexin A11', other_designations='56 kDa autoantigen|CAP-50|annexin XI|annexin-11|autoantigen, 56-kD|calcyclin-associated annexin 50', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80152610, end_position_on_the_genomic_accession=80205677, orientation='minus', exon_count=21, OMIM=602572, no_hdr0=''),
  312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=312, CurrentID=0, Status='live', Symbol='ANXA13', Aliases='ANX13, ISA', description='annexin A13', other_designations='annexin XIII|annexin, intestine-specific|annexin-13|intestine-specific annexin', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123680794, end_position_on_the_genomic_accession=123737407, orientation='minus', exon_count=12, OMIM=602573, no_hdr0=''),
  313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=313, CurrentID=0, Status='live', Symbol='AOAH', Aliases='', description='acyloxyacyl hydrolase', other_designations='acyloxyacyl hydrolase (neutrophil)', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36512941, end_position_on_the_genomic_accession=36724549, orientation='minus', exon_count=29, OMIM=102593, no_hdr0=''),
  314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=314, CurrentID=0, Status='live', Symbol='AOC2', Aliases='DAO2, RAO', description='amine oxidase, copper containing 2', other_designations='SSAO|amine oxidase copper containing 2|amine oxidase, copper containing 2 (retina-specific)|semicarbazide-sensitive amine oxidase', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42844592, end_position_on_the_genomic_accession=42850707, orientation='plus', exon_count=4, OMIM=602268, no_hdr0=''),
  316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=316, CurrentID=0, Status='live', Symbol='AOX1', Aliases='AO, AOH1', description='aldehyde oxidase 1', other_designations='azaheterocycle hydroxylase', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200585952, end_position_on_the_genomic_accession=200682241, orientation='plus', exon_count=38, OMIM=602841, no_hdr0=''),
  317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317, CurrentID=0, Status='live', Symbol='APAF1', Aliases='APAF-1, CED4', description='apoptotic peptidase activating factor 1', other_designations='', map_location='12q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98645228, end_position_on_the_genomic_accession=98735433, orientation='plus', exon_count=29, OMIM=602233, no_hdr0=''),
  318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=318, CurrentID=0, Status='live', Symbol='NUDT2', Aliases='APAH1', description='nudix hydrolase 2', other_designations="Ap4A hydrolase 1|Ap4Aase|bis(5'-nucleosyl)-tetraphosphatase (asymmetrical)|diadenosine 5',5'''-P1,P4-tetraphosphate asymmetrical hydrolase|diadenosine 5',5''-P1,P4-tetraphosphate pyrophosphohydrolase|diadenosine tetraphosphatase|nucleoside diphosphate-linked moiety X motif 2|nudix (nucleoside diphosphate linked moiety X)-type motif 2|nudix motif 2", map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34329506, end_position_on_the_genomic_accession=34343713, orientation='plus', exon_count=5, OMIM=602852, no_hdr0=''),
  319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319, CurrentID=0, Status='live', Symbol='APOF', Aliases='Apo-F, LTIP', description='apolipoprotein F', other_designations='lipid transfer inhibitor protein', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56360571, end_position_on_the_genomic_accession=56362799, orientation='minus', exon_count=2, OMIM=107760, no_hdr0=''),
  320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=320, CurrentID=0, Status='live', Symbol='APBA1', Aliases='D9S411E, LIN10, MINT1, X11, X11A, X11ALPHA', description='amyloid beta (A4) precursor protein-binding, family A, member 1', other_designations='adapter protein X11alpha|adaptor protein X11alpha|amyloid beta (A4) precursor protein-binding, family A, member 1 (X11)|mint-1|neuron-specific X11 protein|neuronal munc18-1-interacting protein 1|phosphotyrosine-binding/-interacting domain (PTB)-bearing protein', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69427532, end_position_on_the_genomic_accession=69673617, orientation='minus', exon_count=16, OMIM=602414, no_hdr0=''),
  321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=321, CurrentID=0, Status='live', Symbol='APBA2', Aliases='D15S1518E, HsT16821, LIN-10, MGC:14091, MINT2, X11-BETA, X11L', description='amyloid beta (A4) precursor protein-binding, family A, member 2', other_designations='X11-like protein|adapter protein X11beta|amyloid beta (A4) precursor protein-binding, family A, member 2 (X11-like)|mint-2|neuron-specific X11L protein|neuronal munc18-1-interacting protein 2|phosphotyrosine-binding/-interacting domain (PTB)-bearing protein', map_location='15q11-q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28885994, end_position_on_the_genomic_accession=29118315, orientation='plus', exon_count=21, OMIM=602712, no_hdr0=''),
  322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=322, CurrentID=0, Status='live', Symbol='APBB1', Aliases='FE65, MGC:9072, RIR', description='amyloid beta (A4) precursor protein-binding, family B, member 1 (Fe65)', other_designations='adaptor protein FE65a2|stat-like protein', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6395124, end_position_on_the_genomic_accession=6419472, orientation='minus', exon_count=19, OMIM=602709, no_hdr0=''),
  323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=323, CurrentID=0, Status='live', Symbol='APBB2', Aliases='FE65L, FE65L1', description='amyloid beta (A4) precursor protein-binding, family B, member 2', other_designations='Fe65-like 1|amyloid beta precursor protein-binding family B member 2|protein Fe65-like 1', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40810027, end_position_on_the_genomic_accession=41216714, orientation='minus', exon_count=25, OMIM=602710, no_hdr0=''),
  324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=324, CurrentID=0, Status='live', Symbol='APC', Aliases='BTPS2, DP2, DP2.5, DP3, GS, PPP1R46', description='adenomatous polyposis coli', other_designations='adenomatosis polyposis coli tumor suppressor|deleted in polyposis 2.5|protein phosphatase 1, regulatory subunit 46|truncated adenomatosis polyposis coli', map_location='5q21-q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112707505, end_position_on_the_genomic_accession=112846239, orientation='plus', exon_count=18, OMIM=611731, no_hdr0=''),
  325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=325, CurrentID=0, Status='live', Symbol='APCS', Aliases='HEL-S-92n, PTX2, SAP', description='amyloid P component, serum', other_designations='9.5S alpha-1-glycoprotein|epididymis secretory sperm binding protein Li 92n|pentaxin-related|pentraxin-related', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159587826, end_position_on_the_genomic_accession=159588871, orientation='plus', exon_count=2, OMIM=104770, no_hdr0=''),
  326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326, CurrentID=0, Status='live', Symbol='AIRE', Aliases='AIRE1, APECED, APS1, APSI, PGA1', description='autoimmune regulator', other_designations='autoimmune polyendocrinopathy candidiasis ectodermal dystrophy protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44285838, end_position_on_the_genomic_accession=44298219, orientation='plus', exon_count=14, OMIM=607358, no_hdr0=''),
  327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=327, CurrentID=0, Status='live', Symbol='APEH', Aliases='AARE, ACPH, APH, D3F15S2, D3S48E, DNF15S2, OPH', description='acylaminoacyl-peptide hydrolase', other_designations='N-acylaminoacyl-peptide hydrolase|acyl-peptide hydrolase|acylaminoacyl-peptidase|oxidized protein hydrolase', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49673130, end_position_on_the_genomic_accession=49683504, orientation='plus', exon_count=24, OMIM=102645, no_hdr0=''),
  328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=328, CurrentID=0, Status='live', Symbol='APEX1', Aliases='APE, APE1, APEN, APEX, APX, HAP1, REF1', description='APEX nuclease (multifunctional DNA repair enzyme) 1', other_designations='AP endonuclease class I|AP lyase|apurinic-apyrimidinic endonuclease 1|apurinic/apyrimidinic (abasic) endonuclease|deoxyribonuclease (apurinic or apyrimidinic)|protein REF-1|redox factor-1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20455131, end_position_on_the_genomic_accession=20457772, orientation='plus', exon_count=5, OMIM=107748, no_hdr0=''),
  329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=329, CurrentID=0, Status='live', Symbol='BIRC2', Aliases='API1, HIAP2, Hiap-2, MIHB, RNF48, c-IAP1, cIAP1', description='baculoviral IAP repeat containing 2', other_designations='IAP homolog B|IAP-2|NFR2-TRAF signalling complex protein|RING finger protein 48|TNFR2-TRAF-signaling complex protein 2|apoptosis inhibitor 1|baculoviral IAP repeat-containing 2|inhibitor of apoptosis protein 2', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102347182, end_position_on_the_genomic_accession=102378670, orientation='plus', exon_count=9, OMIM=601712, no_hdr0=''),
  330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=330, CurrentID=0, Status='live', Symbol='BIRC3', Aliases='AIP1, API2, CIAP2, HAIP1, HIAP1, MALT2, MIHC, RNF49, c-IAP2', description='baculoviral IAP repeat containing 3', other_designations='IAP homolog C|IAP-1|RING finger protein 49|TNFR2-TRAF signaling complex protein|TNFR2-TRAF-signaling complex protein 1|apoptosis inhibitor 2|baculoviral IAP repeat-containing 3|inhibitor of apoptosis protein 1|mammalian IAP homolog C', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102317373, end_position_on_the_genomic_accession=102339403, orientation='plus', exon_count=9, OMIM=601721, no_hdr0=''),
  331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=331, CurrentID=0, Status='live', Symbol='XIAP', Aliases='API3, BIRC4, IAP-3, ILP1, MIHA, XLP2, hIAP-3, hIAP3', description='X-linked inhibitor of apoptosis', other_designations='IAP-like protein|X-linked IAP|X-linked inhibitor of apoptosis, E3 ubiquitin protein ligase|baculoviral IAP repeat-containing protein 4|inhibitor of apoptosis protein 3', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123859812, end_position_on_the_genomic_accession=123913979, orientation='plus', exon_count=9, OMIM=300079, no_hdr0=''),
  332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=332, CurrentID=0, Status='live', Symbol='BIRC5', Aliases='API4, EPR-1', description='baculoviral IAP repeat containing 5', other_designations='apoptosis inhibitor 4|apoptosis inhibitor survivin|survivin variant 3 alpha', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78214196, end_position_on_the_genomic_accession=78225635, orientation='plus', exon_count=6, OMIM=603352, no_hdr0=''),
  333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=333, CurrentID=0, Status='live', Symbol='APLP1', Aliases='APLP', description='amyloid beta (A4) precursor-like protein 1', other_designations='APLP-1|amyloid precursor-like protein 1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35868499, end_position_on_the_genomic_accession=35879797, orientation='plus', exon_count=17, OMIM=104775, no_hdr0=''),
  334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=334, CurrentID=0, Status='live', Symbol='APLP2', Aliases='APLP-2, APPH, APPL2, CDEBP', description='amyloid beta (A4) precursor-like protein 2', other_designations='CDEI box-binding protein|amyloid precursor protein homolog HSD-2', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130069821, end_position_on_the_genomic_accession=130144811, orientation='plus', exon_count=19, OMIM=104776, no_hdr0=''),
  335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=335, CurrentID=0, Status='live', Symbol='APOA1', Aliases='apo(a)', description='apolipoprotein A-I', other_designations='apo-AI', map_location='11q23-q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116835751, end_position_on_the_genomic_accession=116837647, orientation='minus', exon_count=4, OMIM=107680, no_hdr0=''),
  336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=336, CurrentID=0, Status='live', Symbol='APOA2', Aliases='Apo-AII, ApoA-II, apoAII', description='apolipoprotein A-II', other_designations='apolipoprotein A2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161222293, end_position_on_the_genomic_accession=161223628, orientation='minus', exon_count=4, OMIM=107670, no_hdr0=''),
  337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337, CurrentID=0, Status='live', Symbol='APOA4', Aliases='', description='apolipoprotein A-IV', other_designations='apo-AIV|apoA-IV|apolipoprotein A4', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116820702, end_position_on_the_genomic_accession=116823295, orientation='minus', exon_count=3, OMIM=107690, no_hdr0=''),
  338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338, CurrentID=0, Status='live', Symbol='APOB', Aliases='FLDB, LDLCQ4, apoB-100, apoB-48', description='apolipoprotein B', other_designations='apolipoprotein B (including Ag(x) antigen)|apolipoprotein B48', map_location='2p24-p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20999218, end_position_on_the_genomic_accession=21044073, orientation='minus', exon_count=31, OMIM=107730, no_hdr0=''),
  339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339, CurrentID=0, Status='live', Symbol='APOBEC1', Aliases='APOBEC-1, BEDP, CDAR1, HEPR', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide 1', other_designations='apolipoprotein B mRNA editing enzyme complex-1|apolipoprotein B mRNA-editing enzyme 1|mRNA(cytosine(6666)) deaminase 1', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7649400, end_position_on_the_genomic_accession=7670599, orientation='minus', exon_count=6, OMIM=600130, no_hdr0=''),
  341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341, CurrentID=0, Status='live', Symbol='APOC1', Aliases='Apo-CI, ApoC-I', description='apolipoprotein C-I', other_designations='apo-CIB|apoC-IB|apolipoprotein C-I variant I|apolipoprotein C-I variant II|apolipoprotein C1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44914325, end_position_on_the_genomic_accession=44919349, orientation='plus', exon_count=5, OMIM=107710, no_hdr0=''),
  342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342, CurrentID=0, Status='live', Symbol='APOC1P1', Aliases='', description='apolipoprotein C-I pseudogene 1', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44926803, end_position_on_the_genomic_accession=44931386, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343, CurrentID=0, Status='live', Symbol='AQP8', Aliases='AQP-8', description='aquaporin 8', other_designations='', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25216916, end_position_on_the_genomic_accession=25228932, orientation='plus', exon_count=6, OMIM=603750, no_hdr0=''),
  344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344, CurrentID=0, Status='live', Symbol='APOC2', Aliases='APO-CII, APOC-II', description='apolipoprotein C-II', other_designations='apolipoprotein C2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44945982, end_position_on_the_genomic_accession=44949565, orientation='plus', exon_count=4, OMIM=608083, no_hdr0=''),
  345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345, CurrentID=0, Status='live', Symbol='APOC3', Aliases='APOCIII, HALP2', description='apolipoprotein C-III', other_designations='apo-CIII|apoC-III|apolipoprotein C3', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116829908, end_position_on_the_genomic_accession=116833071, orientation='plus', exon_count=4, OMIM=107720, no_hdr0=''),
  346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346, CurrentID=0, Status='live', Symbol='APOC4', Aliases='APO-CIV, APOC-IV', description='apolipoprotein C-IV', other_designations='apolipoprotein C4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44942238, end_position_on_the_genomic_accession=44945496, orientation='plus', exon_count=3, OMIM=600745, no_hdr0=''),
  347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347, CurrentID=0, Status='live', Symbol='APOD', Aliases='', description='apolipoprotein D', other_designations='apo-D', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195568702, end_position_on_the_genomic_accession=195584205, orientation='minus', exon_count=5, OMIM=107740, no_hdr0=''),
  348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348, CurrentID=0, Status='live', Symbol='APOE', Aliases='AD2, APO-E, LDLCQ5, LPG', description='apolipoprotein E', other_designations='apolipoprotein E3', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44905749, end_position_on_the_genomic_accession=44909395, orientation='plus', exon_count=6, OMIM=107741, no_hdr0=''),
  101929219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929219, CurrentID=0, Status='live', Symbol='LOC101929219', Aliases='', description='uncharacterized LOC101929219', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=350, CurrentID=0, Status='live', Symbol='APOH', Aliases='B2G1, B2GP1, BG', description='apolipoprotein H', other_designations='APC inhibitor|B2GPI|activated protein C-binding protein|anticardiolipin cofactor|apo-H|apolipoprotein H (beta-2-glycoprotein I)|beta(2)GPI', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66212029, end_position_on_the_genomic_accession=66229438, orientation='minus', exon_count=8, OMIM=138700, no_hdr0=''),
  351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=351, CurrentID=0, Status='live', Symbol='APP', Aliases='AAA, ABETA, ABPP, AD1I, CTFgamma, CVAP, PN-II, PN2, APP', description='amyloid beta (A4) precursor protein', other_designations='alzheimer disease amyloid protein|beta-amyloid peptide|beta-amyloid peptide(1-40)|beta-amyloid peptide(1-42)|cerebral vascular amyloid peptide|peptidase nexin-II|preA4|protease nexin-II', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25880550, end_position_on_the_genomic_accession=26171128, orientation='minus', exon_count=20, OMIM=104760, no_hdr0=''),
  353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353, CurrentID=0, Status='live', Symbol='APRT', Aliases='AMPD, APRT', description='adenine phosphoribosyltransferase', other_designations='AMP diphosphorylase|AMP pyrophosphorylase|transphosphoribosidase', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88809469, end_position_on_the_genomic_accession=88811934, orientation='minus', exon_count=5, OMIM=102600, no_hdr0=''),
  354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=354, CurrentID=0, Status='live', Symbol='KLK3', Aliases='APS, KLK2A1, PSA, hK3', description='kallikrein related peptidase 3', other_designations='P-30 antigen|gamma-seminoprotein|kallikrein-3|kallikrein-related peptidase 3|prostate specific antigen|semenogelase|seminin', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50854915, end_position_on_the_genomic_accession=50877702, orientation='plus', exon_count=6, OMIM=176820, no_hdr0=''),
  355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=355, CurrentID=0, Status='live', Symbol='FAS', Aliases='ALPS1A, APO-1, APT1, CD951, FASTM, TNFRSF6, FAS', description='Fas cell surface death receptor', other_designations='APO-1 cell surface antigen|CD95 antigen|Delta Fas/APO-1/CD95|FAS 827dupA|FAS receptor variant 9|FASLG receptor|Fas (TNF receptor superfamily, member 6)|Fas AMA|TNF receptor superfamily member 6|apoptosis antigen 1|apoptosis-mediating surface antigen FAS|tumor necrosis factor receptor superfamily, member 6', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88969801, end_position_on_the_genomic_accession=89017059, orientation='plus', exon_count=14, OMIM=134637, no_hdr0=''),
  356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=356, CurrentID=0, Status='live', Symbol='FASLG', Aliases='ALPS1B, APT1LG1, APTL, CD178, CD95-L, CD95L, FASL, TNFSF6', description='Fas ligand', other_designations='CD95 ligand|Fas ligand (TNF superfamily, member 6)|apoptosis (APO-1) antigen ligand 1|apoptosis antigen ligand|fas antigen ligand|mutant tumor necrosis factor family member 6', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172659008, end_position_on_the_genomic_accession=172666873, orientation='plus', exon_count=4, OMIM=134638, no_hdr0=''),
  357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=357, CurrentID=0, Status='live', Symbol='SHROOM2', Aliases='APXL, HSAPXL', description='shroom family member 2', other_designations='APX homolog of Xenopus|apical-like protein', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9786402, end_position_on_the_genomic_accession=9949443, orientation='plus', exon_count=11, OMIM=300103, no_hdr0=''),
  358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=358, CurrentID=0, Status='live', Symbol='AQP1', Aliases='AQP-CHIP, CHIP28, CO', description='aquaporin 1 (Colton blood group)', other_designations='aquaporin 1 (channel-forming integral protein, 28kDa, CO blood group)|aquaporin 1, Colton blood group antigen|aquaporin-CHIP|channel-like integral membrane protein, 28-kDa|urine water channel|water channel protein for red blood cells and kidney proximal tubule', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30911800, end_position_on_the_genomic_accession=30925516, orientation='plus', exon_count=7, OMIM=107776, no_hdr0=''),
  359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359, CurrentID=0, Status='live', Symbol='AQP2', Aliases='AQP-CD, WCH-CD', description='aquaporin 2 (collecting duct)', other_designations='ADH water channel|AQP-2|aquaporin-CD|collecting duct water channel protein|water channel protein for renal collecting duct|water-channel aquaporin 2', map_location='12q12-q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49950741, end_position_on_the_genomic_accession=49958881, orientation='plus', exon_count=4, OMIM=107777, no_hdr0=''),
  360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360, CurrentID=0, Status='live', Symbol='AQP3', Aliases='AQP-3, GIL', description='aquaporin 3 (Gill blood group)', other_designations='aquaglyceroporin-3|aquaporin 3 (GIL blood group)', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33441154, end_position_on_the_genomic_accession=33447633, orientation='minus', exon_count=6, OMIM=600170, no_hdr0=''),
  361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=361, CurrentID=0, Status='live', Symbol='AQP4', Aliases='MIWC, WCH4', description='aquaporin 4', other_designations='aquaporin type4|mercurial-insensitive water channel', map_location='18q11.2-q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26852043, end_position_on_the_genomic_accession=26865752, orientation='minus', exon_count=6, OMIM=600308, no_hdr0=''),
  362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=362, CurrentID=0, Status='live', Symbol='AQP5', Aliases='AQP-5, PPKB', description='aquaporin 5', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49961496, end_position_on_the_genomic_accession=49965682, orientation='plus', exon_count=5, OMIM=600442, no_hdr0=''),
  363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=363, CurrentID=0, Status='live', Symbol='AQP6', Aliases='AQP2L, KID', description='aquaporin 6', other_designations='AQP-6|aquaporin 2-like, kidney specific|aquaporin 6, kidney specific|aquaporin-6, kidney specific|hKID|kidney-specific aquaporin', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49972837, end_position_on_the_genomic_accession=49977139, orientation='plus', exon_count=4, OMIM=601383, no_hdr0=''),
  364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=364, CurrentID=0, Status='live', Symbol='AQP7', Aliases='AQP7L, AQPap, GLYCQTL', description='aquaporin 7', other_designations='aquaglyceroporin-7|aquaporin adipose', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33384947, end_position_on_the_genomic_accession=33402645, orientation='minus', exon_count=11, OMIM=602974, no_hdr0=''),
  366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=366, CurrentID=0, Status='live', Symbol='AQP9', Aliases='AQP-9, HsT17287, SSC1', description='aquaporin 9', other_designations='aquaglyceroporin-9|small solute channel 1', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58138209, end_position_on_the_genomic_accession=58185911, orientation='plus', exon_count=6, OMIM=602914, no_hdr0=''),
  367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=367, CurrentID=0, Status='live', Symbol='AR', Aliases='AIS8, DHTR, HUMARA, HYSP1, KD, NR3C4, SBMA, SMAX1, TFM, AR', description='androgen receptor', other_designations='androgen nuclear receptor variant 2|androgen receptor splice variant 4b|dihydrotestosterone receptor|nuclear receptor subfamily 3 group C member 4', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=67544032, end_position_on_the_genomic_accession=67730619, orientation='plus', exon_count=9, OMIM=313700, no_hdr0=''),
  368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=368, CurrentID=0, Status='live', Symbol='ABCC6', Aliases='ABC34, ARA, EST349056, GACI2, MLP1, MOAT-E, MOATE, MRP6, PXE, PXE1, URG7', description='ATP binding cassette subfamily C member 6', other_designations='ATP-binding cassette sub-family C member 6|ATP-binding cassette, sub-family C (CFTR/MRP), member 6|anthracycline resistance-associated protein|multi-specific organic anion transporter E', map_location='16p13.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16149565, end_position_on_the_genomic_accession=16223673, orientation='minus', exon_count=34, OMIM=603234, no_hdr0=''),
  369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=369, CurrentID=0, Status='live', Symbol='ARAF', Aliases='A-RAF1, PKS2, RAFA1, ARAF', description='A-Raf proto-oncogene, serine/threonine kinase', other_designations='A-Raf proto-oncogene serine/threonine-protein kinase|Oncogene ARAF1|Ras-binding protein DA-Raf|proto-oncogene A-Raf-1|proto-oncogene Pks|v-raf murine sarcoma 3611 viral oncogene homolog 1|v-raf murine sarcoma 3611 viral oncogene-like protein', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47561100, end_position_on_the_genomic_accession=47571921, orientation='plus', exon_count=18, OMIM=311010, no_hdr0=''),
  104355219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355219, CurrentID=0, Status='live', Symbol='LHX5-AS1', Aliases='locus 4010, locus4010', description='LHX5 antisense RNA 1 (head to head)', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113472003, end_position_on_the_genomic_accession=113480481, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=372, CurrentID=0, Status='live', Symbol='ARCN1', Aliases='COPD', description='archain 1', other_designations='archain vesicle transport protein 1|coatomer delta subunit|coatomer protein complex, subunit delta|coatomer protein delta-COP|delta-COP|delta-coat protein', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118572387, end_position_on_the_genomic_accession=118603033, orientation='plus', exon_count=10, OMIM=600820, no_hdr0=''),
  373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373, CurrentID=0, Status='live', Symbol='TRIM23', Aliases='ARD1, ARFD1, RNF46', description='tripartite motif containing 23', other_designations='ADP-ribosylation factor domain protein 1, 64kDa|ADP-ribosylation factor domain-containing protein 1|ARF domain protein 1|GTP-binding protein ARD-1|RING finger protein 46|tripartite motif protein TRIM23|tripartite motif-containing 23|tripartite motif-containing protein 23', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65589680, end_position_on_the_genomic_accession=65624360, orientation='minus', exon_count=14, OMIM=601747, no_hdr0=''),
  374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374, CurrentID=0, Status='live', Symbol='AREG', Aliases='ARB, CRDGF, SDGF, AREG', description='amphiregulin', other_designations='amphiregulin B|colorectum cell-derived growth factor|schwannoma-derived growth factor', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74445098, end_position_on_the_genomic_accession=74455009, orientation='plus', exon_count=6, OMIM=104640, no_hdr0=''),
  375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375, CurrentID=0, Status='live', Symbol='ARF1', Aliases='', description='ADP ribosylation factor 1', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228082660, end_position_on_the_genomic_accession=228099212, orientation='plus', exon_count=6, OMIM=103180, no_hdr0=''),
  377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377, CurrentID=0, Status='live', Symbol='ARF3', Aliases='', description='ADP ribosylation factor 3', other_designations='small GTP binding protein', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48931156, end_position_on_the_genomic_accession=48957526, orientation='minus', exon_count=8, OMIM=103190, no_hdr0=''),
  131450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131450, CurrentID=0, Status='live', Symbol='CD200R1', Aliases='CD200R, HCRTR2, MOX2R, OX2R', description='CD200 receptor 1', other_designations='CD200 cell surface glycoprotein receptor|MOX2 receptor|cell surface glycoprotein OX2 receptor 1|cell surface glycoprotein receptor CD200', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112921057, end_position_on_the_genomic_accession=112975122, orientation='minus', exon_count=8, OMIM=607546, no_hdr0=''),
  379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379, CurrentID=0, Status='live', Symbol='ARL4D', Aliases='ARF4L, ARL6', description='ADP ribosylation factor like GTPase 4D', other_designations='ADP-ribosylation factor like GTPase 4D|ADP-ribosylation factor-like 4D|ADP-ribosylation factor-like 6|ADP-ribosylation factor-like protein 4L', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43398985, end_position_on_the_genomic_accession=43401136, orientation='plus', exon_count=3, OMIM=600732, no_hdr0=''),
  380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=380, CurrentID=0, Status='live', Symbol='ARF4P1', Aliases='ARF4P', description='ADP ribosylation factor 4 pseudogene 1', other_designations='ADP-ribosylation factor 4 pseudogene 1', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134078631, end_position_on_the_genomic_accession=134079747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=381, CurrentID=0, Status='live', Symbol='ARF5', Aliases='', description='ADP ribosylation factor 5', other_designations='', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127588352, end_position_on_the_genomic_accession=127591705, orientation='plus', exon_count=6, OMIM=103188, no_hdr0=''),
  382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=382, CurrentID=0, Status='live', Symbol='ARF6', Aliases='', description='ADP ribosylation factor 6', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49893018, end_position_on_the_genomic_accession=49897054, orientation='plus', exon_count=2, OMIM=600464, no_hdr0=''),
  383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=383, CurrentID=0, Status='live', Symbol='ARG1', Aliases='', description='arginase 1', other_designations='arginase, liver|liver-type arginase|type I arginase', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131573203, end_position_on_the_genomic_accession=131584332, orientation='plus', exon_count=8, OMIM=608313, no_hdr0=''),
  384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=384, CurrentID=0, Status='live', Symbol='ARG2', Aliases='', description='arginase 2', other_designations='L-arginine amidinohydrolase|L-arginine ureahydrolase|arginase, type II|kidney arginase|kidney-type arginase|non-hepatic arginase|nonhepatic arginase|type II arginase', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67619862, end_position_on_the_genomic_accession=67651720, orientation='plus', exon_count=8, OMIM=107830, no_hdr0=''),
  387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387, CurrentID=0, Status='live', Symbol='RHOA', Aliases='ARH12, ARHA, RHO12, RHOH12', description='ras homolog family member A', other_designations='Aplysia ras-related homolog 12|oncogene RHO H12|small GTP binding protein RhoA', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49359136, end_position_on_the_genomic_accession=49412093, orientation='minus', exon_count=6, OMIM=165390, no_hdr0=''),
  388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388, CurrentID=0, Status='live', Symbol='RHOB', Aliases='ARH6, ARHB, MST081, MSTP081, RHOH6', description='ras homolog family member B', other_designations='Aplysia RAS-related homolog 6|h6|oncogene RHO H6|ras homolog gene family, member B|rho cDNA clone 6', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20447071, end_position_on_the_genomic_accession=20449445, orientation='plus', exon_count=1, OMIM=165370, no_hdr0=''),
  389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389, CurrentID=0, Status='live', Symbol='RHOC', Aliases='ARH9, ARHC, H9, RHOH9', description='ras homolog family member C', other_designations='RAS-related homolog 9|oncogene RHO H9|ras homolog gene family, member C|rho cDNA clone 9|rhoC GTPase|small GTP binding protein RhoC', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112701127, end_position_on_the_genomic_accession=112707403, orientation='minus', exon_count=7, OMIM=165380, no_hdr0=''),
  390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390, CurrentID=0, Status='live', Symbol='RND3', Aliases='ARHE, Rho8, RhoE, memB', description='Rho family GTPase 3', other_designations='protein MemB|ras homolog gene family, member E|rho-related GTP-binding protein Rho8|small GTP binding protein Rho8', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150468193, end_position_on_the_genomic_accession=150487695, orientation='minus', exon_count=6, OMIM=602924, no_hdr0=''),
  391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391, CurrentID=0, Status='live', Symbol='RHOG', Aliases='ARHG', description='ras homolog family member G', other_designations='ras homolog gene family, member G (rho G)', map_location='11p15.5-p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3826978, end_position_on_the_genomic_accession=3840983, orientation='minus', exon_count=3, OMIM=179505, no_hdr0=''),
  392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392, CurrentID=0, Status='live', Symbol='ARHGAP1', Aliases='CDC42GAP, RHOGAP, RHOGAP1, p50rhoGAP', description='Rho GTPase activating protein 1', other_designations='CDC42 GTPase-activating protein|GTPase-activating protein rhoGAP|GTPase-activating protein rhoOGAP|GTPase-activating protein, Rho, 1|rho-related small GTPase protein activator|rho-type GTPase-activating protein 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46677075, end_position_on_the_genomic_accession=46700665, orientation='minus', exon_count=14, OMIM=602732, no_hdr0=''),
  393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=393, CurrentID=0, Status='live', Symbol='ARHGAP4', Aliases='C1, RGC1, RhoGAP4, SrGAP4, p115', description='Rho GTPase activating protein 4', other_designations='Rho-GAP hematopoietic protein C1|rho-type GTPase-activating protein 4', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153907376, end_position_on_the_genomic_accession=153926260, orientation='minus', exon_count=23, OMIM=300023, no_hdr0=''),
  394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394, CurrentID=0, Status='live', Symbol='ARHGAP5', Aliases='GFI2, RhoGAP5, p190-B, p190BRhoGAP', description='Rho GTPase activating protein 5', other_designations='growth factor independent 2|p100 RasGAP-associated p105 protein|p105 RhoGAP|rho-type GTPase-activating protein 5', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32076458, end_position_on_the_genomic_accession=32159728, orientation='plus', exon_count=9, OMIM=602680, no_hdr0=''),
  395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=395, CurrentID=0, Status='live', Symbol='ARHGAP6', Aliases='RHOGAP6, RHOGAPX-1', description='Rho GTPase activating protein 6', other_designations='Rho-type GTPase-activating protein RhoGAPX-1|rho-type GTPase-activating protein 6', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11137543, end_position_on_the_genomic_accession=11665701, orientation='minus', exon_count=17, OMIM=300118, no_hdr0=''),
  396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=396, CurrentID=0, Status='live', Symbol='ARHGDIA', Aliases='GDIA1, HEL-S-47e, NPHS8, RHOGDI, RHOGDI-1', description='Rho GDP dissociation inhibitor (GDI) alpha', other_designations='GDP-dissociation inhibitor, aplysia RAS-related 1|Rho GDP dissociation inhibitor alpha|epididymis secretory sperm binding protein Li 47e', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81867719, end_position_on_the_genomic_accession=81871406, orientation='minus', exon_count=7, OMIM=601925, no_hdr0=''),
  397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=397, CurrentID=0, Status='live', Symbol='ARHGDIB', Aliases='D4, GDIA2, GDID4, LYGDI, Ly-GDI, RAP1GN1, RhoGDI2', description='Rho GDP dissociation inhibitor (GDI) beta', other_designations='Rho GDI 2|rho-GDI beta', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14942015, end_position_on_the_genomic_accession=14961628, orientation='minus', exon_count=8, OMIM=602843, no_hdr0=''),
  398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=398, CurrentID=0, Status='live', Symbol='ARHGDIG', Aliases='RHOGDI-3', description='Rho GDP dissociation inhibitor (GDI) gamma', other_designations='RhoGDI gamma|rho GDI 3|rho-GDI gamma', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=280606, end_position_on_the_genomic_accession=283003, orientation='plus', exon_count=6, OMIM=602844, no_hdr0=''),
  399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399, CurrentID=0, Status='live', Symbol='RHOH', Aliases='ARHH, TTF', description='ras homolog family member H', other_designations='GTP-binding protein TTF|TTF, translocation three four|ras homolog gene family, member H', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40191011, end_position_on_the_genomic_accession=40244764, orientation='plus', exon_count=10, OMIM=602037, no_hdr0=''),
  400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400, CurrentID=0, Status='live', Symbol='ARL1', Aliases='ARFL1', description='ADP ribosylation factor like GTPase 1', other_designations='ADP-ribosylation factor like GTPase 1|ADP-ribosylation factor-like 1', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101393120, end_position_on_the_genomic_accession=101407820, orientation='minus', exon_count=6, OMIM=603425, no_hdr0=''),
  401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401, CurrentID=0, Status='live', Symbol='PHOX2A', Aliases='ARIX, CFEOM2, FEOM2, NCAM2, PMX2A', description='paired-like homeobox 2a', other_designations='ARIX1 homeodomain protein|aristaless homeobox homolog|aristaless homeobox protein homolog|arix homeodomain protein', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72239077, end_position_on_the_genomic_accession=72244176, orientation='minus', exon_count=3, OMIM=602753, no_hdr0=''),
  402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402, CurrentID=0, Status='live', Symbol='ARL2', Aliases='ARFL2', description='ADP ribosylation factor like GTPase 2', other_designations='ADP-ribosylation factor like GTPase 2|ADP-ribosylation factor-like 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65014113, end_position_on_the_genomic_accession=65022185, orientation='plus', exon_count=5, OMIM=601175, no_hdr0=''),
  403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403, CurrentID=0, Status='live', Symbol='ARL3', Aliases='ARFL3', description='ADP ribosylation factor like GTPase 3', other_designations='ADP-ribosylation factor like GTPase 3|ADP-ribosylation factor-like 3|ARF-like 3', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102673727, end_position_on_the_genomic_accession=102714433, orientation='minus', exon_count=6, OMIM=604695, no_hdr0=''),
  405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=405, CurrentID=0, Status='live', Symbol='ARNT', Aliases='HIF-1-beta, HIF-1beta, HIF1-beta, HIF1B, HIF1BETA, TANGO, bHLHe2', description='aryl hydrocarbon receptor nuclear translocator', other_designations='class E basic helix-loop-helix protein 2|dioxin receptor, nuclear translocator|hypoxia-inducible factor 1, beta subunit', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150809705, end_position_on_the_genomic_accession=150876768, orientation='minus', exon_count=25, OMIM=126110, no_hdr0=''),
  406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406, CurrentID=0, Status='live', Symbol='ARNTL', Aliases='BMAL1, BMAL1c, JAP3, MOP3, PASD3, TIC, bHLHe5', description='aryl hydrocarbon receptor nuclear translocator like', other_designations='ARNT-like protein 1, brain and muscle|PAS domain-containing protein 3|bHLH-PAS protein JAP3|basic-helix-loop-helix-PAS orphan MOP3|basic-helix-loop-helix-PAS protein MOP3|brain and muscle ARNT-like 1|class E basic helix-loop-helix protein 5|member of PAS protein 3|member of PAS superfamily 3', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13277727, end_position_on_the_genomic_accession=13387811, orientation='plus', exon_count=21, OMIM=602550, no_hdr0=''),
  407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407, CurrentID=0, Status='live', Symbol='ARR3', Aliases='ARRX', description='arrestin 3, retinal (X-arrestin)', other_designations='C-arrestin|X-arrestin|arrestin 4|cArr|cone arrestin|retinal cone arrestin-3', map_location='Xcen-q21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70263503, end_position_on_the_genomic_accession=70281840, orientation='plus', exon_count=17, OMIM=301770, no_hdr0=''),
  408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408, CurrentID=0, Status='live', Symbol='ARRB1', Aliases='ARB1, ARR1', description='arrestin, beta 1', other_designations='arrestin 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75260122, end_position_on_the_genomic_accession=75351831, orientation='minus', exon_count=19, OMIM=107940, no_hdr0=''),
  409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=409, CurrentID=0, Status='live', Symbol='ARRB2', Aliases='ARB2, ARR2, BARR2', description='arrestin, beta 2', other_designations='arrestin 3|arrestin beta-2', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4710494, end_position_on_the_genomic_accession=4721500, orientation='plus', exon_count=17, OMIM=107941, no_hdr0=''),
  410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=410, CurrentID=0, Status='live', Symbol='ARSA', Aliases='MLD', description='arylsulfatase A', other_designations='ASA|cerebroside-sulfatase', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50622754, end_position_on_the_genomic_accession=50628173, orientation='minus', exon_count=9, OMIM=607574, no_hdr0=''),
  411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=411, CurrentID=0, Status='live', Symbol='ARSB', Aliases='ASB, G4S, MPS6', description='arylsulfatase B', other_designations='N-acetylgalactosamine-4-sulfatase', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78777209, end_position_on_the_genomic_accession=78986534, orientation='minus', exon_count=14, OMIM=611542, no_hdr0=''),
  412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=412, CurrentID=0, Status='live', Symbol='STS', Aliases='ARSC, ARSC1, ASC, ES, SSDD, XLI', description='steroid sulfatase (microsomal), isozyme S', other_designations='arylsulfatase C|estrone sulfatase|steryl-sulfate sulfohydrolase', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=7146517, end_position_on_the_genomic_accession=7354641, orientation='plus', exon_count=16, OMIM=300747, no_hdr0=''),
  413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=413, CurrentID=0, Status='live', Symbol='ARSC2', Aliases='ARSC', description='arylsulfatase C, isozyme F', other_designations='Arylsulfatase C, f form', map_location='Xpter-p22.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301780, no_hdr0=''),
  414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414, CurrentID=0, Status='live', Symbol='ARSD', Aliases='ASD', description='arylsulfatase D', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2903970, end_position_on_the_genomic_accession=2929375, orientation='minus', exon_count=11, OMIM=300002, no_hdr0=''),
  415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=415, CurrentID=0, Status='live', Symbol='ARSE', Aliases='ASE, CDPX, CDPX1, CDPXR', description='arylsulfatase E (chondrodysplasia punctata 1)', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2934632, end_position_on_the_genomic_accession=2968310, orientation='minus', exon_count=13, OMIM=300180, no_hdr0=''),
  416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=416, CurrentID=0, Status='live', Symbol='ARSF', Aliases='ASF', description='arylsulfatase F', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3040234, end_position_on_the_genomic_accession=3112729, orientation='plus', exon_count=14, OMIM=300003, no_hdr0=''),
  417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=417, CurrentID=0, Status='live', Symbol='ART1', Aliases='ART2, ARTC1, CD296, RT6', description='ADP-ribosyltransferase 1', other_designations='ADP-ribosyltransferase 2|ADP-ribosyltransferase C2 and C3 toxin-like 1|mono(ADP-ribosyl)transferase 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3642234, end_position_on_the_genomic_accession=3665075, orientation='plus', exon_count=7, OMIM=601625, no_hdr0=''),
  418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=418, CurrentID=0, Status='live', Symbol='ART2P', Aliases='ART1P, RT6', description='ADP-ribosyltransferase 2, pseudogene', other_designations='ADP-ribosyltransferase 2 (RT6 antigen homolog, rat), pseudogene|ADP-ribosyltransferase 2 pseudogene (RT6 antigen homolog)|ADP-ribosyltransferase 2 pseudogene (RT6 antigen homolog, rat)|RT6 antigen homolog', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72520951, end_position_on_the_genomic_accession=72521640, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=419, CurrentID=0, Status='live', Symbol='ART3', Aliases='ARTC3', description='ADP-ribosyltransferase 3', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 3|NAD(P)(+)--arginine ADP-ribosyltransferase 3|mono(ADP-ribosyl)transferase 3|mono-ADP-ribosyltransferase', map_location='4p15.1-p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76011180, end_position_on_the_genomic_accession=76112802, orientation='plus', exon_count=16, OMIM=603086, no_hdr0=''),
  420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=420, CurrentID=0, Status='live', Symbol='ART4', Aliases='ARTC4, CD297, DO, DOK1', description='ADP-ribosyltransferase 4 (Dombrock blood group)', other_designations='ADP-ribosyltransferase 4 (DO blood group)|ADP-ribosyltransferase C2 and C3 toxin-like 4|Dombrock blood group carrier molecule|NAD(P)(+)--arginine ADP-ribosyltransferase 4|mono-ADP-ribosyltransferase 4', map_location='12p13-p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14829311, end_position_on_the_genomic_accession=14843479, orientation='minus', exon_count=3, OMIM=110600, no_hdr0=''),
  421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=421, CurrentID=0, Status='live', Symbol='ARVCF', Aliases='', description='armadillo repeat gene deleted in velocardiofacial syndrome', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19966603, end_position_on_the_genomic_accession=20016802, orientation='minus', exon_count=27, OMIM=602269, no_hdr0=''),
  424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=424, CurrentID=0, Status='live', Symbol='ARVD3', Aliases='ARVD2', description='arrhythmogenic right ventricular dysplasia 3', other_designations='', map_location='14q12-q22', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602086, no_hdr0=''),
  425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=425, CurrentID=0, Status='live', Symbol='ARVD4', Aliases='', description='arrhythmogenic right ventricular dysplasia 4', other_designations='', map_location='2q32.1-q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602087, no_hdr0=''),
  100126474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126474, CurrentID=0, Status='live', Symbol='TRNAF-GAA', Aliases='', description='transfer RNA phenylalanine (anticodon GAA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=427, CurrentID=0, Status='live', Symbol='ASAH1', Aliases='AC, ACDase, ASAH, PHP, PHP32, SMAPME', description='N-acylsphingosine amidohydrolase (acid ceramidase) 1', other_designations='acid CDase|acylsphingosine deacylase|putative 32 kDa heart protein', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18056416, end_position_on_the_genomic_accession=18084998, orientation='minus', exon_count=16, OMIM=613468, no_hdr0=''),
  429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=429, CurrentID=0, Status='live', Symbol='ASCL1', Aliases='ASH1, HASH1, MASH1, bHLHa46', description='achaete-scute family bHLH transcription factor 1', other_designations='ASH-1|achaete scute protein|achaete-scute complex homolog 1|achaete-scute complex-like 1|class A basic helix-loop-helix protein 46', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102957674, end_position_on_the_genomic_accession=102960516, orientation='plus', exon_count=2, OMIM=100790, no_hdr0=''),
  430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=430, CurrentID=0, Status='live', Symbol='ASCL2', Aliases='ASH2, HASH2, MASH2, bHLHa45', description='achaete-scute family bHLH transcription factor 2', other_designations='ASH-2|achaete-scute complex homolog 2|achaete-scute complex-like 2|class A basic helix-loop-helix protein 45|mammalian achaete/scute homologue 2', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2268498, end_position_on_the_genomic_accession=2270952, orientation='minus', exon_count=2, OMIM=601886, no_hdr0=''),
  431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431, CurrentID=0, Status='live', Symbol='ASD1', Aliases='', description='atrial septal defect 1', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=108800, no_hdr0=''),
  432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=432, CurrentID=0, Status='live', Symbol='ASGR1', Aliases='ASGPR, ASGPR1, CLEC4H1, HL-1', description='asialoglycoprotein receptor 1', other_designations='C-type lectin domain family 4 member H1|hepatic lectin H1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7173431, end_position_on_the_genomic_accession=7179564, orientation='minus', exon_count=10, OMIM=108360, no_hdr0=''),
  433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=433, CurrentID=0, Status='live', Symbol='ASGR2', Aliases='ASGP-R2, ASGPR2, CLEC4H2, HBXBP, HL-2', description='asialoglycoprotein receptor 2', other_designations='C-type lectin domain family 4 member H2|HBxAg-binding protein|hepatic lectin H2', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7101322, end_position_on_the_genomic_accession=7115522, orientation='minus', exon_count=11, OMIM=108361, no_hdr0=''),
  434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=434, CurrentID=0, Status='live', Symbol='ASIP', Aliases='AGSW, AGTI, AGTIL, ASP, SHEP9', description='agouti signaling protein', other_designations='agouti signaling protein, nonagouti homolog|agouti switch protein|nonagouti homolog', map_location='20q11.2-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34186493, end_position_on_the_genomic_accession=34269344, orientation='plus', exon_count=8, OMIM=600201, no_hdr0=''),
  435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=435, CurrentID=0, Status='live', Symbol='ASL', Aliases='ASAL', description='argininosuccinate lyase', other_designations='argininosuccinase|arginosuccinase', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66075789, end_position_on_the_genomic_accession=66093343, orientation='plus', exon_count=17, OMIM=608310, no_hdr0=''),
  436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=436, CurrentID=0, Status='live', Symbol='ASLP1', Aliases='ASLL', description='argininosuccinate lyase pseudogene 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23648968, end_position_on_the_genomic_accession=23653359, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=438, CurrentID=0, Status='live', Symbol='ASMT', Aliases='ASMTY, HIOMT, HIOMTY', description='acetylserotonin O-methyltransferase', other_designations='acetylserotonin N-methyltransferase|acetylserotonin methyltransferase (Y chromosome)|hydroxyindole O-methyltransferase', map_location='Xp22.3 or Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1595455, end_position_on_the_genomic_accession=1643081, orientation='plus', exon_count=10, OMIM='300015,402500', no_hdr0=''),
  439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439, CurrentID=0, Status='live', Symbol='ASNA1', Aliases='ARSA-I, ARSA1, ASNA-I, GET3, TRC40, hASNA-I', description='arsA arsenite transporter, ATP-binding, homolog 1 (bacterial)', other_designations='golgi to ER traffic 3 homolog|transmembrane domain recognition complex 40 kDa ATPase subunit|transmembrane domain recognition complex, 40kDa', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12737492, end_position_on_the_genomic_accession=12748323, orientation='plus', exon_count=7, OMIM=601913, no_hdr0=''),
  440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440, CurrentID=0, Status='live', Symbol='ASNS', Aliases='ASNSD, TS11', description='asparagine synthetase (glutamine-hydrolyzing)', other_designations='TS11 cell cycle control protein|glutamine-dependent asparagine synthetase', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97852117, end_position_on_the_genomic_accession=97872542, orientation='minus', exon_count=15, OMIM=108370, no_hdr0=''),
  442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442, CurrentID=0, Status='live', Symbol='ASNSP2', Aliases='ASNSL2', description='asparagine synthetase pseudogene 2', other_designations='asparagine synthetase-like 2', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8853355, end_position_on_the_genomic_accession=8853510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=443, CurrentID=0, Status='live', Symbol='ASPA', Aliases='ACY2, ASP', description='aspartoacylase', other_designations='ACY-2|aminoacylase 2|aminoacylase-2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3474110, end_position_on_the_genomic_accession=3499406, orientation='plus', exon_count=9, OMIM=608034, no_hdr0=''),
  444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=444, CurrentID=0, Status='live', Symbol='ASPH', Aliases='AAH, BAH, CASQ2BP1, FDLAB, HAAH, JCTN, junctin', description='aspartate beta-hydroxylase', other_designations='A beta H-J-J|ASP beta-hydroxylase|aspartyl/asparaginyl-beta-hydroxylase|cardiac junctin|humbug|junctate|peptide-aspartate beta-dioxygenase', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61500556, end_position_on_the_genomic_accession=61714640, orientation='minus', exon_count=32, OMIM=600582, no_hdr0=''),
  445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445, CurrentID=0, Status='live', Symbol='ASS1', Aliases='ASS, CTLN1', description='argininosuccinate synthase 1', other_designations='argininosuccinate synthetase 1|citrulline--aspartate ligase|citrulline-aspartate ligase', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130444707, end_position_on_the_genomic_accession=130501274, orientation='plus', exon_count=18, OMIM=603470, no_hdr0=''),
  447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=447, CurrentID=0, Status='live', Symbol='ASS1P2', Aliases='ASSP2', description='argininosuccinate synthetase 1 pseudogene 2', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 2', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38810215, end_position_on_the_genomic_accession=38811756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449, CurrentID=0, Status='live', Symbol='ASS1P4', Aliases='ASSP4', description='argininosuccinate synthetase 1 pseudogene 4', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 4', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3367045, end_position_on_the_genomic_accession=3368609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450, CurrentID=0, Status='live', Symbol='ASS1P5', Aliases='ASSP5', description='argininosuccinate synthetase 1 pseudogene 5', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 5', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115664789, end_position_on_the_genomic_accession=115666339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=451, CurrentID=0, Status='live', Symbol='ASS1P6', Aliases='ASSP6', description='argininosuccinate synthetase 1 pseudogene 6', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 6', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11922285, end_position_on_the_genomic_accession=11923873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379787, CurrentID=0, Status='live', Symbol='LOC105379787', Aliases='', description='uncharacterized LOC105379787', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=453, CurrentID=0, Status='live', Symbol='ASS1P8', Aliases='ASSP8', description='argininosuccinate synthetase 1 pseudogene 8', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 8', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148526260, end_position_on_the_genomic_accession=148527154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=454, CurrentID=0, Status='live', Symbol='ASS1P9', Aliases='ASSP9', description='argininosuccinate synthetase 1 pseudogene 9', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 9', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53859064, end_position_on_the_genomic_accession=53860632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=455, CurrentID=0, Status='live', Symbol='ASS1P10', Aliases='ASSP10', description='argininosuccinate synthetase 1 pseudogene 10', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 10', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145228711, end_position_on_the_genomic_accession=145230285, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=459, CurrentID=0, Status='live', Symbol='ASS1P14', Aliases='ASSP14', description='argininosuccinate synthetase 1 pseudogene 14', other_designations='argininosuccinate synthetase pseudogene 14', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32985738, end_position_on_the_genomic_accession=32987301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=460, CurrentID=0, Status='live', Symbol='ASTN1', Aliases='ASTN', description='astrotactin 1', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176857305, end_position_on_the_genomic_accession=177164904, orientation='minus', exon_count=26, OMIM=600904, no_hdr0=''),
  462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=462, CurrentID=0, Status='live', Symbol='SERPINC1', Aliases='AT3, AT3D, ATIII, THPH7', description='serpin peptidase inhibitor, clade C (antithrombin), member 1', other_designations='serine (or cysteine) proteinase inhibitor, clade C (antithrombin), member 1|serpin peptidase inhibitor clade C member 1', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173903800, end_position_on_the_genomic_accession=173917378, orientation='minus', exon_count=8, OMIM=107300, no_hdr0=''),
  463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=463, CurrentID=0, Status='live', Symbol='ZFHX3', Aliases='ATBF1, ATBT, ZNF927', description='zinc finger homeobox 3', other_designations='AT motif-binding factor 1|AT-binding transcription factor 1|ZFH-3|alpha-fetoprotein enhancer binding protein|alpha-fetoprotein enhancer-binding protein|zinc finger homeodomain protein 3', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72782885, end_position_on_the_genomic_accession=73058635, orientation='minus', exon_count=11, OMIM=104155, no_hdr0=''),
  465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=465, CurrentID=0, Status='live', Symbol='ATD', Aliases='ATD1', description='asphixiating thoracic dystrophy (chondroectodermal dysplasia-like syndrome)', other_designations='', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=208500, no_hdr0=''),
  466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=466, CurrentID=0, Status='live', Symbol='ATF1', Aliases='EWS-ATF1, FUS/ATF-1, TREB36', description='activating transcription factor 1', other_designations='cAMP-dependent transcription factor ATF-1', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50763710, end_position_on_the_genomic_accession=50821167, orientation='plus', exon_count=9, OMIM=123803, no_hdr0=''),
  467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=467, CurrentID=0, Status='live', Symbol='ATF3', Aliases='', description='activating transcription factor 3', other_designations='cAMP-dependent transcription factor ATF-3', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212565334, end_position_on_the_genomic_accession=212620777, orientation='plus', exon_count=7, OMIM=603148, no_hdr0=''),
  468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=468, CurrentID=0, Status='live', Symbol='ATF4', Aliases='CREB-2, CREB2, TAXREB67, TXREB', description='activating transcription factor 4', other_designations='DNA-binding protein TAXREB67|cAMP response element-binding protein 2|cAMP-dependent transcription factor ATF-4|cAMP-responsive element-binding protein 2|cyclic AMP-responsive element-binding protein 2|tax-responsive enhancer element B67|tax-responsive enhancer element-binding protein 67', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39520559, end_position_on_the_genomic_accession=39522686, orientation='plus', exon_count=3, OMIM=604064, no_hdr0=''),
  469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=469, CurrentID=0, Status='live', Symbol='ATF4P1', Aliases='ATF4P, TAXREB67P, TXREBP', description='activating transcription factor 4 pseudogene 1', other_designations='activating transcription factor 4 (tax-responsive enhancer element B67) pseudogene|activating transcription factor 4 pseudogene (tax-responsive enhancer element B67 pseudogene)|cyclic-AMP-dependent transcription factor ATF-4 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154622340, end_position_on_the_genomic_accession=154623780, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=470, CurrentID=0, Status='live', Symbol='ATHS', Aliases='ALP', description='atherosclerosis susceptibility (lipoprotein associated)', other_designations='', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=108725, no_hdr0=''),
  471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=471, CurrentID=0, Status='live', Symbol='ATIC', Aliases='AICAR, AICARFT, HEL-S-70p, IMPCHASE, PURH', description='5-aminoimidazole-4-carboxamide ribonucleotide formyltransferase/IMP cyclohydrolase', other_designations='5-aminoimidazole-4-carboxamide-1-beta-D-ribonucleotide transformylase/inosinicase|AICAR formyltransferase/IMP cyclohydrolase bifunctional enzyme|AICARFT/IMPCHASE|epididymis secretory sperm binding protein Li 70p|phosphoribosylaminoimidazolecarboxamide formyltransferase/IMP cyclohydrolase', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215311956, end_position_on_the_genomic_accession=215349773, orientation='plus', exon_count=16, OMIM=601731, no_hdr0=''),
  472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=472, CurrentID=0, Status='live', Symbol='ATM', Aliases='AT1, ATA, ATC, ATD, ATDC, ATE, TEL1, TELO1', description='ATM serine/threonine kinase', other_designations='A-T mutated|AT mutated|TEL1, telomere maintenance 1, homolog|ataxia telangiectasia mutated', map_location='11q22-q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108222500, end_position_on_the_genomic_accession=108369102, orientation='plus', exon_count=69, OMIM=607585, no_hdr0=''),
  473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=473, CurrentID=0, Status='live', Symbol='RERE', Aliases='ARG, ARP, ATN1L, DNB1', description='arginine-glutamic acid dipeptide (RE) repeats', other_designations='atrophin 2|atrophin-1 like protein|atrophin-1 related protein|atrophin-1-like protein|atrophin-1-related protein', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8352404, end_position_on_the_genomic_accession=8817691, orientation='minus', exon_count=28, OMIM=605226, no_hdr0=''),
  474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474, CurrentID=0, Status='live', Symbol='ATOH1', Aliases='ATH1, HATH1, MATH-1, bHLHa14', description='atonal bHLH transcription factor 1', other_designations='atonal homolog 1|atonal homolog bHLH transcription factor 1|class A basic helix-loop-helix protein 14|helix-loop-helix protein hATH-1', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=93828927, end_position_on_the_genomic_accession=93829991, orientation='plus', exon_count=1, OMIM=601461, no_hdr0=''),
  475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=475, CurrentID=0, Status='live', Symbol='ATOX1', Aliases='ATX1, HAH1', description='antioxidant 1 copper chaperone', other_designations='ATX1 antioxidant protein 1 homolog|metal transport protein ATX1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151742822, end_position_on_the_genomic_accession=151758649, orientation='minus', exon_count=4, OMIM=602270, no_hdr0=''),
  476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=476, CurrentID=0, Status='live', Symbol='ATP1A1', Aliases='', description='ATPase, Na+/K+ transporting, alpha 1 polypeptide', other_designations='Na(+)/K(+) ATPase alpha-1 subunit|Na+, K+ ATPase alpha subunit|Na+/K+ ATPase 1|Na, K-ATPase, alpha-A catalytic polypeptide|Na,K-ATPase alpha-1 subunit|Na,K-ATPase catalytic subunit alpha-A protein|sodium pump 1|sodium pump subunit alpha-1|sodium-potassium ATPase catalytic subunit alpha-1|sodium-potassium-ATPase, alpha 1 polypeptide', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116372185, end_position_on_the_genomic_accession=116404774, orientation='plus', exon_count=26, OMIM=182310, no_hdr0=''),
  477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=477, CurrentID=0, Status='live', Symbol='ATP1A2', Aliases='FHM2, MHP2', description='ATPase, Na+/K+ transporting, alpha 2 polypeptide', other_designations='ATPase Na+/K+ transporting alpha 2 polypeptide|Na(+)/K(+) ATPase alpha-2 subunit|Na+/K+ ATPase, alpha-A(+) catalytic polypeptide|Na+/K+ ATPase, alpha-B polypeptide|sodium pump subunit alpha-2|sodium-potassium ATPase catalytic subunit alpha-2|sodium/potassium-transporting ATPase alpha-2 chain', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160115730, end_position_on_the_genomic_accession=160143591, orientation='plus', exon_count=23, OMIM=182340, no_hdr0=''),
  478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=478, CurrentID=0, Status='live', Symbol='ATP1A3', Aliases='AHC2, CAPOS, DYT12, RDP', description='ATPase, Na+/K+ transporting, alpha 3 polypeptide', other_designations='Na(+)/K(+) ATPase alpha(III) subunit|Na(+)/K(+) ATPase alpha-3 subunit|Na+/K+ ATPase 3|sodium pump subunit alpha-3|sodium-potassium ATPase catalytic subunit alpha-3|sodium-potassium-ATPase, alpha 3 polypeptide|sodium/potassium-transporting ATPase alpha-3 chain', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41966582, end_position_on_the_genomic_accession=41994276, orientation='minus', exon_count=24, OMIM=182350, no_hdr0=''),
  479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=479, CurrentID=0, Status='live', Symbol='ATP12A', Aliases='ATP1AL1', description='ATPase, H+/K+ transporting, nongastric, alpha polypeptide', other_designations='ATPase, Na+/K+ transporting, alpha polypeptide-like 1|non-gastric H(+)/K(+) ATPase alpha subunit|non-gastric H(+)/K(+) ATPase subunit alpha|proton pump', map_location='13q12.1-q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24680411, end_position_on_the_genomic_accession=24711785, orientation='plus', exon_count=23, OMIM=182360, no_hdr0=''),
  480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=480, CurrentID=0, Status='live', Symbol='ATP1A4', Aliases='ATP1A1, ATP1AL2', description='ATPase, Na+/K+ transporting, alpha 4 polypeptide', other_designations='ATPase, Na+/K+ transporting, alpha polypeptide-like 2|Na(+)/K(+) ATPase alpha-4 subunit|Na+/K+ ATPase 4|Na+/K+ ATPase, alpha-D polypeptide|Na,K-ATPase subunit alpha-C|sodium pump 4|sodium pump subunit alpha-4|sodium-potassium ATPase catalytic subunit alpha-4|sodium/potassium-transporting ATPase alpha-4 chain', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160151562, end_position_on_the_genomic_accession=160186977, orientation='plus', exon_count=23, OMIM=607321, no_hdr0=''),
  481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=481, CurrentID=0, Status='live', Symbol='ATP1B1', Aliases='ATP1B', description='ATPase, Na+/K+ transporting, beta 1 polypeptide', other_designations='Beta 1-subunit of Na(+),K(+)-ATPase|Na, K-ATPase beta-1 polypeptide|adenosinetriphosphatase|sodium pump subunit beta-1|sodium-potassium ATPase subunit beta 1 (non-catalytic)|sodium/potassium-dependent ATPase beta-1 subunit|sodium/potassium-transporting ATPase beta-1 chain', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169106709, end_position_on_the_genomic_accession=169132722, orientation='plus', exon_count=6, OMIM=182330, no_hdr0=''),
  482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=482, CurrentID=0, Status='live', Symbol='ATP1B2', Aliases='AMOG', description='ATPase, Na+/K+ transporting, beta 2 polypeptide', other_designations='Na, K-ATPase beta-2 polypeptide|adhesion molecule in glia|adhesion molecule on glia|sodium pump subunit beta-2|sodium-potassium ATPase subunit beta 2 (non-catalytic)|sodium/potassium-dependent ATPase beta-2 subunit|sodium/potassium-dependent ATPase subunit beta-2|sodium/potassium-transporting ATPase beta-2 chain', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7646627, end_position_on_the_genomic_accession=7657771, orientation='plus', exon_count=8, OMIM=182331, no_hdr0=''),
  483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=483, CurrentID=0, Status='live', Symbol='ATP1B3', Aliases='ATPB-3, CD298', description='ATPase, Na+/K+ transporting, beta 3 polypeptide', other_designations='Na, K-ATPase beta-3 polypeptide|sodium pump subunit beta-3|sodium-potassium ATPase subunit beta 3 (non-catalytic)|sodium/potassium-dependent ATPase beta-3 subunit|sodium/potassium-dependent ATPase subunit beta-3|sodium/potassium-transporting ATPase beta-3 chain', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141876628, end_position_on_the_genomic_accession=141926540, orientation='plus', exon_count=7, OMIM=601867, no_hdr0=''),
  484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=484, CurrentID=0, Status='live', Symbol='ATP1B3P1', Aliases='', description='ATPase, Na+/K+ transporting, beta 3 pseudogene', other_designations='ATPase, Na+/K+ transporting, beta 3 polypeptide pseudogene|ATPase, Na+/K+ transporting, beta 3 pseudogene 1', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60734390, end_position_on_the_genomic_accession=60735927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=485, CurrentID=0, Status='live', Symbol='ATP1B1P1', Aliases='ATP1BL1', description='ATPase, Na+/K+ transporting, beta 1 polypeptide pseudogene 1', other_designations='ATPase, Na+/K+ transporting, beta polypeptide-like 1', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42029085, end_position_on_the_genomic_accession=42031265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=486, CurrentID=0, Status='live', Symbol='FXYD2', Aliases='ATP1G1, HOMG2', description='FXYD domain containing ion transport regulator 2', other_designations='ATPase, Na+/K+ transporting, gamma 1 polypeptide|Na(+)/K(+) ATPase subunit gamma|sodium pump gamma chain', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117820075, end_position_on_the_genomic_accession=117828092, orientation='minus', exon_count=7, OMIM=601814, no_hdr0=''),
  487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=487, CurrentID=0, Status='live', Symbol='ATP2A1', Aliases='ATP2A, SERCA1', description='ATPase, Ca++ transporting, cardiac muscle, fast twitch 1', other_designations='SR Ca(2+)-ATPase 1|calcium pump 1|calcium-transporting ATPase sarcoplasmic reticulum type, fast twitch skeletal muscle isoform|endoplasmic reticulum class 1/2 Ca(2+) ATPase', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28878488, end_position_on_the_genomic_accession=28904509, orientation='plus', exon_count=24, OMIM=108730, no_hdr0=''),
  488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=488, CurrentID=0, Status='live', Symbol='ATP2A2', Aliases='ATP2B, DAR, DD, SERCA2', description='ATPase, Ca++ transporting, cardiac muscle, slow twitch 2', other_designations='ATPase Ca++ transporting cardiac muscle slow twitch 2|ATPase, Ca++ dependent, slow-twitch, cardiac muscle-2|SR Ca(2+)-ATPase 2|calcium pump 2|calcium-transporting ATPase sarcoplasmic reticulum type, slow twitch skeletal muscle isoform|cardiac Ca2+ ATPase|endoplasmic reticulum class 1/2 Ca(2+) ATPase', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110281227, end_position_on_the_genomic_accession=110351093, orientation='plus', exon_count=21, OMIM=108740, no_hdr0=''),
  489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=489, CurrentID=0, Status='live', Symbol='ATP2A3', Aliases='SERCA3', description='ATPase, Ca++ transporting, ubiquitous', other_designations='ATPase, Ca(2+)-transporting, ubiquitous|SR Ca(2+)-ATPase 3|adenosine triphosphatase, calcium|calcium pump 3|calcium-translocating P-type ATPase|sarco/endoplasmic reticulum Ca2+ -ATPase', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3923869, end_position_on_the_genomic_accession=3964464, orientation='minus', exon_count=26, OMIM=601929, no_hdr0=''),
  490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=490, CurrentID=0, Status='live', Symbol='ATP2B1', Aliases='PMCA1, PMCA1kb', description='ATPase, Ca++ transporting, plasma membrane 1', other_designations='plasma membrane calcium pump', map_location='12q21.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89588049, end_position_on_the_genomic_accession=89709353, orientation='minus', exon_count=24, OMIM=108731, no_hdr0=''),
  491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=491, CurrentID=0, Status='live', Symbol='ATP2B2', Aliases='PMCA2, PMCA2a, PMCA2i', description='ATPase, Ca++ transporting, plasma membrane 2', other_designations='plasma membrane Ca(2+)-ATPase|plasma membrane Ca2+ pump 2|plasma membrane calcium ATPase|plasma membrane calcium pump', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10324023, end_position_on_the_genomic_accession=10707993, orientation='minus', exon_count=31, OMIM=108733, no_hdr0=''),
  492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=492, CurrentID=0, Status='live', Symbol='ATP2B3', Aliases='CFAP39, CLA2, OPCA, PMCA3, PMCA3a, SCAX1', description='ATPase, Ca++ transporting, plasma membrane 3', other_designations='cilia and flagella associated protein 39|plasma membrane calcium ATPase|plasma membrane calcium pump', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153517665, end_position_on_the_genomic_accession=153582929, orientation='plus', exon_count=25, OMIM=300014, no_hdr0=''),
  493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493, CurrentID=0, Status='live', Symbol='ATP2B4', Aliases='ATP2B2, MXRA1, PMCA4, PMCA4b, PMCA4x', description='ATPase, Ca++ transporting, plasma membrane 4', other_designations='matrix-remodeling-associated protein 1|sarcolemmal calcium pump', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203626787, end_position_on_the_genomic_accession=203744081, orientation='plus', exon_count=22, OMIM=108732, no_hdr0=''),
  131566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131566, CurrentID=0, Status='live', Symbol='DCBLD2', Aliases='CLCP1, ESDN', description='discoidin, CUB and LCCL domain containing 2', other_designations='1700055P21Rik|CUB, LCCL and coagulation factor V/VIII-homology domains protein 1|coagulation factor V/VIII-homology domains protein 1|endothelial and smooth muscle cell-derived neuropilin-like protein', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98795970, end_position_on_the_genomic_accession=98901689, orientation='minus', exon_count=16, OMIM=608698, no_hdr0=''),
  495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=495, CurrentID=0, Status='live', Symbol='ATP4A', Aliases='ATP6A', description='ATPase, H+/K+ exchanging, alpha polypeptide', other_designations='ATPase, H+/K+ transporting, alpha polypeptide|H(+)-K(+)-ATPase alpha subunit|gastric H(+)/K(+) ATPase subunit alpha|gastric H+/K+ ATPase alpha subunit|gastric H,K-ATPase alpha subunit|gastric H,K-ATPase catalytic subunit|gastric hydrogen-potassium ATPase|proton pump', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35550193, end_position_on_the_genomic_accession=35563658, orientation='minus', exon_count=22, OMIM=137216, no_hdr0=''),
  496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=496, CurrentID=0, Status='live', Symbol='ATP4B', Aliases='ATP6B', description='ATPase, H+/K+ exchanging, beta polypeptide', other_designations='ATPase, H+/K+ transporting, beta polypeptide|gastric H(+)/K(+) ATPase subunit beta|gastric H+/K+ ATPase beta subunit|gastric hydrogen-potassium ATPase, beta|potassium-transporting ATPase beta chain|proton pump beta chain', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113648804, end_position_on_the_genomic_accession=113658198, orientation='minus', exon_count=7, OMIM=137217, no_hdr0=''),
  498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=498, CurrentID=0, Status='live', Symbol='ATP5A1', Aliases='ATP5A, ATP5AL2, ATPM, COXPD22, HEL-S-123m, MC5DN4, MOM2, OMR, ORM, hATP1', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1, cardiac muscle', other_designations='ATP synthase alpha chain, mitochondrial|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, isoform 1, cardiac muscle|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, isoform 2, non-cardiac muscle-like 2|ATP sythase (F1-ATPase) alpha subunit|epididymis secretory sperm binding protein Li 123m|mitochondrial ATP synthetase, oligomycin-resistant', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46084144, end_position_on_the_genomic_accession=46104233, orientation='minus', exon_count=13, OMIM=164360, no_hdr0=''),
  100532724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532724, CurrentID=0, Status='live', Symbol='NPHP3-ACAD11', Aliases='', description='NPHP3-ACAD11 readthrough (NMD candidate)', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132558138, end_position_on_the_genomic_accession=132722459, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=501, CurrentID=0, Status='live', Symbol='ALDH7A1', Aliases='ATQ1, EPD, PDE', description='aldehyde dehydrogenase 7 family member A1', other_designations='26g turgor protein homolog|P6c dehydrogenase|aldehyde dehydrogenase 7 family, member A1|alpha-AASA dehydrogenase|antiquitin-1|betaine aldehyde dehydrogenase|delta1-piperideine-6-carboxylate dehydrogenase', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126541841, end_position_on_the_genomic_accession=126595390, orientation='minus', exon_count=19, OMIM=107323, no_hdr0=''),
  100532726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532726, CurrentID=0, Status='live', Symbol='NDUFC2-KCTD14', Aliases='', description='NDUFC2-KCTD14 readthrough', other_designations='NDUFC2-KCTD14 readthrough transcript protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78015715, end_position_on_the_genomic_accession=78080219, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100532727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532727, CurrentID=0, Status='live', Symbol='DDX11L12', Aliases='', description='DEAD/H box polypeptide 11 like 12', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504, CurrentID=0, Status='live', Symbol='ATP5A1P2', Aliases='ATP5AP2', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, pseudogene 2', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155269631, end_position_on_the_genomic_accession=155270037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=506, CurrentID=0, Status='live', Symbol='ATP5B', Aliases='ATPMB, ATPSB, HEL-S-271', description='ATP synthase, H+ transporting, mitochondrial F1 complex, beta polypeptide', other_designations='epididymis secretory protein Li 271|mitochondrial ATP synthase beta subunit|mitochondrial ATP synthetase, beta subunit', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56638175, end_position_on_the_genomic_accession=56646068, orientation='minus', exon_count=10, OMIM=102910, no_hdr0=''),
  507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=507, CurrentID=0, Status='live', Symbol='ATP5BP1', Aliases='ATP5BL1, ATPMBL1, ATPSBL1', description='ATP synthase, H+ transporting, mitochondrial F1 complex, beta polypeptide pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, beta polypeptide-like 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98206121, end_position_on_the_genomic_accession=98207904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100532732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532732, CurrentID=0, Status='live', Symbol='MSH5-SAPCD1', Aliases='MSH5-C6orf26', description='MSH5-SAPCD1 readthrough (NMD candidate)', other_designations='MSH5-C6orf26 readthrough (non-protein coding)|MSH5-SAPCD1 readthrough (non-protein coding)', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31739948, end_position_on_the_genomic_accession=31764850, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=509, CurrentID=0, Status='live', Symbol='ATP5C1', Aliases='ATP5C, ATP5CL1', description='ATP synthase, H+ transporting, mitochondrial F1 complex, gamma polypeptide 1', other_designations='ATP synthase gamma chain, mitochondrial|F-ATPase gamma subunit|mitochondrial ATP synthase, gamma subunit 1', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7788130, end_position_on_the_genomic_accession=7807801, orientation='plus', exon_count=10, OMIM=108729, no_hdr0=''),
  131583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131583, CurrentID=0, Status='live', Symbol='FAM43A', Aliases='', description='family with sequence similarity 43 member A', other_designations='family with sequence similarity 43, member A', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194685893, end_position_on_the_genomic_accession=194689037, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100532736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532736, CurrentID=0, Status='live', Symbol='MINOS1-NBL1', Aliases='C1orf151-NBL1', description='MINOS1-NBL1 readthrough', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19596977, end_position_on_the_genomic_accession=19658456, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=513, CurrentID=0, Status='live', Symbol='ATP5D', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F1 complex, delta subunit', other_designations='F-ATPase delta subunit|mitochondrial ATP synthase complex delta-subunit precusor|mitochondrial ATP synthase, delta subunit', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1241750, end_position_on_the_genomic_accession=1244825, orientation='plus', exon_count=5, OMIM=603150, no_hdr0=''),
  514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=514, CurrentID=0, Status='live', Symbol='ATP5E', Aliases='ATPE, MC5DN3', description='ATP synthase, H+ transporting, mitochondrial F1 complex, epsilon subunit', other_designations='F(0)F(1)-ATPase|H(+)-transporting two-sector ATPase|mitochondrial ATP synthase epsilon chain|mitochondrial ATPase', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59028678, end_position_on_the_genomic_accession=59032367, orientation='minus', exon_count=3, OMIM=606153, no_hdr0=''),
  515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=515, CurrentID=0, Status='live', Symbol='ATP5F1', Aliases='PIG47', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1', other_designations='ATP synthase B chain, mitochondrial|ATP synthase proton-transporting mitochondrial F(0) complex subunit B1|ATP synthase subunit b, mitochondrial|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit b|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1|ATPase subunit b|H+-ATP synthase subunit b|cell proliferation-inducing protein 47', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111449121, end_position_on_the_genomic_accession=111461918, orientation='plus', exon_count=7, OMIM=603270, no_hdr0=''),
  516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=516, CurrentID=0, Status='live', Symbol='ATP5G1', Aliases='ATP5A, ATP5G', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9)', other_designations='ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P1|ATP synthase proton-transporting mitochondrial F(0) complex subunit C1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C1 (subunit 9)|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9)|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9)|ATPase protein 9|ATPase subunit 9|ATPase subunit C|mitochondrial ATP synthase, subunit 9|mitochondrial ATP synthase, subunit C', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48892786, end_position_on_the_genomic_accession=48895871, orientation='plus', exon_count=5, OMIM=603192, no_hdr0=''),
  517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=517, CurrentID=0, Status='live', Symbol='ATP5G2', Aliases='ATP5A', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9)', other_designations='ATP synthase c subunit|ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P2|ATP synthase proton-transporting mitochondrial F(0) complex subunit C2|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C2 (subunit 9)|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9)|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9)|ATPase protein 9|ATPase subunit C|mitochondrial ATP synthase, subunit C (subunit 9)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53665160, end_position_on_the_genomic_accession=53676728, orientation='minus', exon_count=6, OMIM=603193, no_hdr0=''),
  518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=518, CurrentID=0, Status='live', Symbol='ATP5G3', Aliases='P3', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C3 (subunit 9)', other_designations='ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P3|ATP synthase proton-transporting mitochondrial F(0) complex subunit C3|ATP synthase subunit 9|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C3 (subunit 9)|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C3 (subunit 9)|ATP synthase, mitochondrial, C subunit-3|ATPase protein 9|ATPase subunit C', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175176258, end_position_on_the_genomic_accession=175181762, orientation='minus', exon_count=4, OMIM=602736, no_hdr0=''),
  520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=520, CurrentID=0, Status='live', Symbol='ATP5G2P1', Aliases='ATP5GP1', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9) pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C2 (subunit 9) pseudogene 1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9) pseudogene 1|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9) pseudogene 1|pseudogene 1 for mitochondrial ATP synthase c subunit (P2 form)', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209267650, end_position_on_the_genomic_accession=209268283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=521, CurrentID=0, Status='live', Symbol='ATP5I', Aliases='ATP5K', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit E', other_designations='ATP synthase e chain, mitochondrial|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit E|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit E|ATPase subunit e|F1F0-ATP synthase, murine e subunit', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=672436, end_position_on_the_genomic_accession=674338, orientation='minus', exon_count=4, OMIM=601519, no_hdr0=''),
  522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=522, CurrentID=0, Status='live', Symbol='ATP5J', Aliases='ATP5, ATP5A, ATPM, CF6, F6', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F6', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F6|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F6|ATPase subunit F6|coupling factor 6|mitochondrial ATP synthase, coupling factor 6|mitochondrial ATP synthase, subunit F6|mitochondrial ATPase coupling factor 6|proliferation-inducing protein 36', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25724480, end_position_on_the_genomic_accession=25735654, orientation='minus', exon_count=6, OMIM=603152, no_hdr0=''),
  523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=523, CurrentID=0, Status='live', Symbol='ATP6V1A', Aliases='ATP6A11, HO68, VA68, VPP2, Vma1, ATP6V1A', description='ATPase, H+ transporting, lysosomal 70kDa, V1 subunit A', other_designations='ATPase, H+ transporting, lysosomal, subunit A1|H(+)-transporting two-sector ATPase, subunit A|H+-transporting ATPase chain A, vacuolar (VA68 type)|V-ATPase 69 kDa subunit 1|V-ATPase A subunit 1|V-ATPase subunit A|vacuolar ATP synthase catalytic subunit A, ubiquitous isoform|vacuolar ATPase isoform VA68|vacuolar proton pump alpha subunit 1|vacuolar proton pump subunit alpha', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113747019, end_position_on_the_genomic_accession=113812058, orientation='plus', exon_count=15, OMIM=607027, no_hdr0=''),
  525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=525, CurrentID=0, Status='live', Symbol='ATP6V1B1', Aliases='ATP6B1, RTA1B, VATB, VMA2, VPP3', description='ATPase, H+ transporting, lysosomal 56/58kDa, V1 subunit B1', other_designations='H(+)-transporting two-sector ATPase, 58kD subunit|H+-ATPase beta 1 subunit|V-ATPase B1 subunit|V-ATPase subunit B 1|endomembrane proton pump 58 kDa subunit|vacuolar proton pump 3|vacuolar proton pump subunit B 1|vacuolar proton pump, subunit 3', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70935868, end_position_on_the_genomic_accession=70965431, orientation='plus', exon_count=15, OMIM=192132, no_hdr0=''),
  526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=526, CurrentID=0, Status='live', Symbol='ATP6V1B2', Aliases='ATP6B1B2, ATP6B2, HO57, VATB, VPP3, Vma2', description='ATPase, H+ transporting, lysosomal 56/58kDa, V1 subunit B2', other_designations='H+ transporting two-sector ATPase|V-ATPase B2 subunit|V-ATPase subunit B 2|endomembrane proton pump 58 kDa subunit|vacuolar H+-ATPase 56,000 subunit|vacuolar proton pump subunit B 2', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20197193, end_position_on_the_genomic_accession=20221696, orientation='plus', exon_count=14, OMIM=606939, no_hdr0=''),
  527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=527, CurrentID=0, Status='live', Symbol='ATP6V0C', Aliases='ATP6C, ATP6L, ATPL, VATL, VPPC, Vma3', description='ATPase, H+ transporting, lysosomal 16kDa, V0 subunit c', other_designations='H(+)-transporting two-sector ATPase, 16 kDa subunit|V-ATPase 16 kDa proteolipid subunit|vacuolar ATP synthase 16 kDa proteolipid subunit|vacuolar H+ ATPase proton channel subunit|vacuolar proton pump 16 kDa proteolipid subunit', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2513726, end_position_on_the_genomic_accession=2520223, orientation='plus', exon_count=4, OMIM=108745, no_hdr0=''),
  528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=528, CurrentID=0, Status='live', Symbol='ATP6V1C1', Aliases='ATP6C, ATP6D, VATC, Vma5', description='ATPase, H+ transporting, lysosomal 42kDa, V1 subunit C1', other_designations='H(+)-transporting two-sector ATPase, subunit C|H+ -ATPase C subunit|H+-transporting ATPase chain C, vacuolar|V-ATPase C subunit|V-ATPase subunit C 1|subunit C of vacuolar proton-ATPase V1 domain|vacuolar ATP synthase subunit C|vacuolar proton pump C subunit|vacuolar proton pump subunit C 1|vacuolar proton pump, 42-kD subunit|vacuolar proton-ATPase, subunit C, VI domain', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103021020, end_position_on_the_genomic_accession=103073057, orientation='plus', exon_count=13, OMIM=603097, no_hdr0=''),
  529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=529, CurrentID=0, Status='live', Symbol='ATP6V1E1', Aliases='ATP6E, ATP6E2, ATP6V1E, P31, Vma4', description='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E1', other_designations='H(+)-transporting two-sector ATPase, 31kDa subunit|H+-transporting ATP synthase chain E, vacuolar|V-ATPase 31 kDa subunit|V-ATPase subunit E 1|V-ATPase, subunit E|vacuolar proton pump subunit E 1', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17592136, end_position_on_the_genomic_accession=17628822, orientation='minus', exon_count=9, OMIM=108746, no_hdr0=''),
  533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=533, CurrentID=0, Status='live', Symbol='ATP6V0B', Aliases='ATP6F, HATPL, VMA16', description='ATPase, H+ transporting, lysosomal 21kDa, V0 subunit b', other_designations="ATPase, H+ transporting, lysosomal, 21-KD, V0 subunit C-prime, prime|ATPase, H+ transporting, lysosomal, subunit F|H(+)-transporting two-sector ATPase, subunit F|V-ATPase 21 kDa proteolipid subunit|V-ATPase subunit c''|vacuolar ATP synthase 21 kDa proteolipid subunit|vacuolar proton pump 21 kDa proteolipid subunit", map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43974648, end_position_on_the_genomic_accession=43978300, orientation='plus', exon_count=8, OMIM=603717, no_hdr0=''),
  534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=534, CurrentID=0, Status='live', Symbol='ATP6V1G2', Aliases='ATP6G, ATP6G2, NG38, VMA10', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G2', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump)|H(+)-transporting two-sector ATPase, subunit G2|V-ATPase 13 kDa subunit 2|vacuolar ATP synthase subunit G 2|vacuolar proton pump G subunit 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31544451, end_position_on_the_genomic_accession=31546848, orientation='minus', exon_count=3, OMIM=606853, no_hdr0=''),
  535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=535, CurrentID=0, Status='live', Symbol='ATP6V0A1', Aliases='ATP6N1, ATP6N1A, Stv1, VPP1, Vph1, a1', description='ATPase, H+ transporting, lysosomal V0 subunit a1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) non-catalytic accessory protein 1A (110/116kD)|ATPase, H+ transporting, lysosomal non-catalytic accessory protein 1 (110/116kD)|H(+)-transporting two-sector ATPase, 116 kDa accessory protein A1|V-ATPase 116 kDa|V-type proton ATPase 116 kDa subunit a|clathrin-coated vesicle/synaptic vesicle proton pump 116 kDa subunit|vacuolar adenosine triphosphatase subunit Ac116|vacuolar proton pump subunit 1|vacuolar proton pump, subunit 1|vacuolar proton translocating ATPase 116 kDa subunit A|vacuolar-type H(+)-ATPase 115 kDa subunit', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42458844, end_position_on_the_genomic_accession=42522579, orientation='plus', exon_count=23, OMIM=192130, no_hdr0=''),
  537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=537, CurrentID=0, Status='live', Symbol='ATP6AP1', Aliases='16A, ATP6IP1, ATP6S1, Ac45, CF2, VATPS1, XAP-3, XAP3', description='ATPase, H+ transporting, lysosomal accessory protein 1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump), subunit 1|ATPase, H+ transporting, lysosomal interacting protein 1|H-ATPase subunit|V-ATPase Ac45 subunit|V-ATPase S1 accessory protein|V-ATPase subunit S1|protein XAP-3|vacuolar proton pump subunit S1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154428632, end_position_on_the_genomic_accession=154436517, orientation='plus', exon_count=11, OMIM=300197, no_hdr0=''),
  538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=538, CurrentID=0, Status='live', Symbol='ATP7A', Aliases='DSMAX, MK, MNK, SMAX3', description='ATPase, Cu++ transporting, alpha polypeptide', other_designations='Cu++-transporting P-type ATPase|Menkes disease-associated protein|copper pump 1', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77910656, end_position_on_the_genomic_accession=78050395, orientation='plus', exon_count=23, OMIM=300011, no_hdr0=''),
  539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=539, CurrentID=0, Status='live', Symbol='ATP5O', Aliases='ATPO, HMC08D05, OSCP', description='ATP synthase, H+ transporting, mitochondrial F1 complex, O subunit', other_designations='human ATP synthase OSCP subunit|oligomycin sensitivity conferral protein oscp-like protein|oligomycin sensitivity conferring protein', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33903453, end_position_on_the_genomic_accession=33915854, orientation='minus', exon_count=7, OMIM=600828, no_hdr0=''),
  540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=540, CurrentID=0, Status='live', Symbol='ATP7B', Aliases='PWD, WC1, WD, WND', description='ATPase, Cu++ transporting, beta polypeptide', other_designations='ATPase, Cu(2+)- transporting, beta polypeptide|Wilson disease-associated protein|copper pump 2', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51891086, end_position_on_the_genomic_accession=52012099, orientation='minus', exon_count=34, OMIM=606882, no_hdr0=''),
  541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541, CurrentID=0, Status='live', Symbol='ALDH7A1P1', Aliases='ATQL1', description='aldehyde dehydrogenase 7 family member A1 pseudogene 1', other_designations='aldehyde dehydrogenase 7 family, member A1 pseudogene 1|antiquitin-like 1', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77290184, end_position_on_the_genomic_accession=77291994, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=542, CurrentID=0, Status='live', Symbol='ALDH7A1P2', Aliases='ATQL2', description='aldehyde dehydrogenase 7 family member A1 pseudogene 2', other_designations='aldehyde dehydrogenase 7 family, member A1 pseudogene 2|antiquitin-like 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172893763, end_position_on_the_genomic_accession=172894383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=543, CurrentID=0, Status='live', Symbol='ALDH7A1P3', Aliases='ATQL3', description='aldehyde dehydrogenase 7 family member A1 pseudogene 3', other_designations='aldehyde dehydrogenase 7 family, member A1 pseudogene 3|antiquitin-like 3', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150541632, end_position_on_the_genomic_accession=150542155, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  131616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131616, CurrentID=0, Status='live', Symbol='TMEM42', Aliases='', description='transmembrane protein 42', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44861723, end_position_on_the_genomic_accession=44865667, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=545, CurrentID=0, Status='live', Symbol='ATR', Aliases='FCTCS, FRP1, MEC1, SCKL, SCKL1', description='ATR serine/threonine kinase', other_designations='FRAP-related protein-1|MEC1, mitosis entry checkpoint 1, homolog|ataxia telangiectasia and Rad3-related protein', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142449235, end_position_on_the_genomic_accession=142578826, orientation='minus', exon_count=49, OMIM=601215, no_hdr0=''),
  546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=546, CurrentID=0, Status='live', Symbol='ATRX', Aliases='ATR2, JMS, MRX52, MRXHF1, RAD54, RAD54L, SFM1, SHS, XH2, XNP, ZNF-HX', description='alpha thalassemia/mental retardation syndrome X-linked', other_designations='ATP-dependent helicase ATRX|DNA dependent ATPase and helicase|X-linked helicase II|X-linked nuclear protein|Zinc finger helicase|alpha thalassemia/mental retardation syndrome X-linked (RAD54 homolog, S. cerevisiae)|helicase 2, X-linked|mental retardation, X-linked 52', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77504878, end_position_on_the_genomic_accession=77786269, orientation='minus', exon_count=37, OMIM='300032,300504', no_hdr0=''),
  547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=547, CurrentID=0, Status='live', Symbol='KIF1A', Aliases='ATSV, C2orf20, HSN2C, MRD9, SPG30, UNC104', description='kinesin family member 1A', other_designations='axonal transporter of synaptic vesicles|kinesin, heavy chain, member 1A, homolog of mouse|microtubule-based motor KIF1A|tmp_locus_27|unc-104- and KIF1A-related protein', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240713764, end_position_on_the_genomic_accession=240821025, orientation='minus', exon_count=54, OMIM=601255, no_hdr0=''),
  549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=549, CurrentID=0, Status='live', Symbol='AUH', Aliases='', description='AU RNA binding protein/enoyl-CoA hydratase', other_designations='3-methylglutaconyl-CoA hydratase|AU RNA-binding protein/enoyl-Coenzyme A hydratase|AU-binding protein/Enoyl-CoA hydratase|AU-specific RNA-binding enoyl-CoA hydratase', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91213815, end_position_on_the_genomic_accession=91361939, orientation='minus', exon_count=18, OMIM=600529, no_hdr0=''),
  550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550, CurrentID=0, Status='live', Symbol='AUP1', Aliases='', description='ancient ubiquitous protein 1', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74526648, end_position_on_the_genomic_accession=74529897, orientation='minus', exon_count=12, OMIM=602434, no_hdr0=''),
  551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=551, CurrentID=0, Status='live', Symbol='AVP', Aliases='ADH, ARVP-NPII, AVRP, VP, AVP', description='arginine vasopressin', other_designations='antidiuretic hormone|copeptin|neurohypophyseal|prepro-AVP-NP II|prepro-arginine-vasopressin-neurophysin II|vasopressin-neurophysin II-copeptin', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3082555, end_position_on_the_genomic_accession=3093521, orientation='minus', exon_count=4, OMIM=192340, no_hdr0=''),
  552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552, CurrentID=0, Status='live', Symbol='AVPR1A', Aliases='AVPR V1a, AVPR1, V1aR', description='arginine vasopressin receptor 1A', other_designations='SCCL vasopressin subtype 1a receptor|V1-vascular vasopressin receptor AVPR1A|V1a vasopressin receptor|antidiuretic hormone receptor 1A|vascular/hepatic-type arginine vasopressin receptor', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63142287, end_position_on_the_genomic_accession=63153860, orientation='minus', exon_count=3, OMIM=600821, no_hdr0=''),
  553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553, CurrentID=0, Status='live', Symbol='AVPR1B', Aliases='AVPR3', description='arginine vasopressin receptor 1B', other_designations='AVPR V1b|AVPR V3|V1bR|antidiuretic hormone receptor 1B|arginine vasopressin receptor 3|pituitary vasopressin receptor 3|vasopressin V3 receptor', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206109849, end_position_on_the_genomic_accession=206117048, orientation='minus', exon_count=2, OMIM=600264, no_hdr0=''),
  554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554, CurrentID=0, Status='live', Symbol='AVPR2', Aliases='ADHR, DI1, DIR, DIR3, NDI, V2R', description='arginine vasopressin receptor 2', other_designations='AVPR V2|antidiuretic hormone receptor|renal-type arginine vasopressin receptor', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153902531, end_position_on_the_genomic_accession=153907166, orientation='plus', exon_count=5, OMIM=300538, no_hdr0=''),
  558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=558, CurrentID=0, Status='live', Symbol='AXL', Aliases='ARK, JTK11, Tyro7, UFO', description='AXL receptor tyrosine kinase', other_designations='AXL oncogene|AXL transforming sequence/gene', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41219199, end_position_on_the_genomic_accession=41261767, orientation='plus', exon_count=21, OMIM=109135, no_hdr0=''),
  100270639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270639, CurrentID=0, Status='live', Symbol='IHPS5', Aliases='', description='Pyloric stenosis, infantile hypertrophic, 5', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612525, no_hdr0=''),
  560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=560, CurrentID=0, Status='live', Symbol='AZF1', Aliases='AZF, AZFA, SP3', description='azoospermia factor 1', other_designations='', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=415000, no_hdr0=''),
  100270641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270641, CurrentID=0, Status='live', Symbol='MYP16', Aliases='', description='myopia 16', other_designations='', map_location='5p15.33-p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612554, no_hdr0=''),
  100270642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270642, CurrentID=0, Status='live', Symbol='CLLS3', Aliases='', description='Leukemia, chronic lymphocytic, susceptibility to, 3', other_designations='', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612557, no_hdr0=''),
  563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=563, CurrentID=0, Status='live', Symbol='AZGP1', Aliases='ZA2G, ZAG', description='alpha-2-glycoprotein 1, zinc-binding', other_designations='Alpha-2-glycoprotein, zinc|Zn-alpha2-glycoprotein|zn-alpha-2-GP|zn-alpha-2-glycoprotein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99966727, end_position_on_the_genomic_accession=99976112, orientation='minus', exon_count=4, OMIM=194460, no_hdr0=''),
  100270644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270644, CurrentID=0, Status='live', Symbol='CLLS5', Aliases='', description='Leukemia, chronic lymphocytic susceptiblity to, 5', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612559, no_hdr0=''),
  566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=566, CurrentID=0, Status='live', Symbol='AZU1', Aliases='AZAMP, AZU, CAP37, HBP, HUMAZUR, NAZC', description='azurocidin 1', other_designations='cationic antimicrobial protein 37|cationic antimicrobial protein CAP37|heparin-binding protein|neutrophil azurocidin', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=827831, end_position_on_the_genomic_accession=832017, orientation='plus', exon_count=5, OMIM=162815, no_hdr0=''),
  567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=567, CurrentID=0, Status='live', Symbol='B2M', Aliases='', description='beta-2-microglobulin', other_designations='beta chain of MHC class I molecules|beta-2-microglobin', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44711487, end_position_on_the_genomic_accession=44718159, orientation='plus', exon_count=4, OMIM=109700, no_hdr0=''),
  568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=568, CurrentID=0, Status='live', Symbol='B2MR', Aliases='', description='beta-2-microglobulin regulator', other_designations='', map_location='15q13-q15', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109710, no_hdr0=''),
  570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=570, CurrentID=0, Status='live', Symbol='BAAT', Aliases='BACAT, BAT', description='bile acid-CoA:amino acid N-acyltransferase', other_designations='bile acid CoA: amino acid N-acyltransferase (glycine N-choloyltransferase)|bile acid CoA:amino acid N-acyltransferase|bile acid Coenzyme A: amino acid N-acyltransferase (glycine N-choloyltransferase)|long-chain fatty-acyl-CoA hydrolase', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101360417, end_position_on_the_genomic_accession=101385005, orientation='minus', exon_count=5, OMIM=602938, no_hdr0=''),
  571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=571, CurrentID=0, Status='live', Symbol='BACH1', Aliases='BACH-1, BTBD24', description='BTB and CNC homology 1, basic leucine zipper transcription factor 1', other_designations='basic region leucine zipper transcriptional regulator BACH1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29298795, end_position_on_the_genomic_accession=29361896, orientation='plus', exon_count=11, OMIM=602751, no_hdr0=''),
  572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=572, CurrentID=0, Status='live', Symbol='BAD', Aliases='BBC2, BCL2L8', description='BCL2 associated agonist of cell death', other_designations='BCL-X/BCL-2 binding protein|BCL2-antagonist of cell death protein|BCL2-associated agonist of cell death|BCL2-binding component 6|BCL2-binding protein|bcl-2-binding component 6|bcl-2-like protein 8|bcl-XL/Bcl-2-associated death promoter|bcl2 antagonist of cell death|bcl2-L-8', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64269828, end_position_on_the_genomic_accession=64284704, orientation='minus', exon_count=3, OMIM=603167, no_hdr0=''),
  573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=573, CurrentID=0, Status='live', Symbol='BAG1', Aliases='BAG-1, HAP, RAP46', description='BCL2 associated athanogene 1', other_designations='BCL2-associated athanogene|Bcl-2 associating athanogene-1 protein|Bcl-2-binding protein|glucocortoid receptor-associated protein RAP46|receptor-associated protein, 46-KD', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33252471, end_position_on_the_genomic_accession=33264761, orientation='minus', exon_count=7, OMIM=601497, no_hdr0=''),
  574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574, CurrentID=0, Status='live', Symbol='BAGE', Aliases='BAGE1, CT2.1', description='B melanoma antigen', other_designations='antigen MZ2-BA|cancer/testis antigen 2.1|cancer/testis antigen family 2, member 1', map_location='21p11.1 not on reference assembly', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605167, no_hdr0=''),
  575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=575, CurrentID=0, Status='live', Symbol='ADGRB1', Aliases='BAI1, GDAIF', description='adhesion G protein-coupled receptor B1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142449888, end_position_on_the_genomic_accession=142545008, orientation='plus', exon_count=31, OMIM=602682, no_hdr0=''),
  576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=576, CurrentID=0, Status='live', Symbol='ADGRB2', Aliases='BAI2', description='adhesion G protein-coupled receptor B2', other_designations='brain-specific angiogenesis inhibitor 2', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31727105, end_position_on_the_genomic_accession=31764063, orientation='minus', exon_count=33, OMIM=602683, no_hdr0=''),
  577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=577, CurrentID=0, Status='live', Symbol='ADGRB3', Aliases='BAI3', description='adhesion G protein-coupled receptor B3', other_designations='brain-specific angiogenesis inhibitor 3|dJ91B17.1 (brain-specific angiogenesis inhibitor 3)', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68635282, end_position_on_the_genomic_accession=69389511, orientation='plus', exon_count=35, OMIM=602684, no_hdr0=''),
  578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=578, CurrentID=0, Status='live', Symbol='BAK1', Aliases='BAK, BAK-LIKE, BCL2L7, CDN1', description='BCL2-antagonist/killer 1', other_designations='BCL2-like 7 protein|apoptosis regulator BAK|bcl-2-like protein 7|bcl2-L-7|pro-apoptotic protein BAK', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33572546, end_position_on_the_genomic_accession=33580299, orientation='minus', exon_count=7, OMIM=600516, no_hdr0=''),
  579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=579, CurrentID=0, Status='live', Symbol='NKX3-2', Aliases='BAPX1, NKX3.2, NKX3B, SMMD', description='NK3 homeobox 2', other_designations='bagpipe homeobox protein homolog 1|homeobox protein NK-3 homolog B', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13540830, end_position_on_the_genomic_accession=13544490, orientation='minus', exon_count=2, OMIM=602183, no_hdr0=''),
  580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=580, CurrentID=0, Status='live', Symbol='BARD1', Aliases='', description='BRCA1 associated RING domain 1', other_designations='BRCA1-associated RING domain gene 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214725646, end_position_on_the_genomic_accession=214809711, orientation='minus', exon_count=13, OMIM=601593, no_hdr0=''),
  581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=581, CurrentID=0, Status='live', Symbol='BAX', Aliases='BCL2L4', description='BCL2-associated X protein', other_designations='BCL2-associated X protein omega|Baxdelta2G9|Baxdelta2G9omega|Baxdelta2omega|bcl-2-like protein 4|bcl2-L-4', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48954860, end_position_on_the_genomic_accession=48961798, orientation='plus', exon_count=7, OMIM=600040, no_hdr0=''),
  582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=582, CurrentID=0, Status='live', Symbol='BBS1', Aliases='BBS2L2', description='Bardet-Biedl syndrome 1', other_designations='BBS2-like protein 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66510648, end_position_on_the_genomic_accession=66533613, orientation='plus', exon_count=17, OMIM=209901, no_hdr0=''),
  583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=583, CurrentID=0, Status='live', Symbol='BBS2', Aliases='BBS', description='Bardet-Biedl syndrome 2', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56470389, end_position_on_the_genomic_accession=56520096, orientation='minus', exon_count=20, OMIM=606151, no_hdr0=''),
  585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=585, CurrentID=0, Status='live', Symbol='BBS4', Aliases='', description='Bardet-Biedl syndrome 4', other_designations='', map_location='15q22.3-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72686179, end_position_on_the_genomic_accession=72738476, orientation='plus', exon_count=17, OMIM=600374, no_hdr0=''),
  586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=586, CurrentID=0, Status='live', Symbol='BCAT1', Aliases='BCATC, BCT1, ECA39, MECA39, PNAS121, PP18', description='branched chain amino-acid transaminase 1, cytosolic', other_designations='branched chain aminotransferase 1, cytosolic|placental protein 18', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24810024, end_position_on_the_genomic_accession=24949459, orientation='minus', exon_count=13, OMIM=113520, no_hdr0=''),
  587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=587, CurrentID=0, Status='live', Symbol='BCAT2', Aliases='BCAM, BCATM, BCT2, PP18', description='branched chain amino-acid transaminase 2, mitochondrial', other_designations='branched chain aminotransferase 2, mitochondrial|placental protein 18', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48795062, end_position_on_the_genomic_accession=48811063, orientation='minus', exon_count=12, OMIM=113530, no_hdr0=''),
  590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=590, CurrentID=0, Status='live', Symbol='BCHE', Aliases='CHE1, CHE2, E1', description='butyrylcholinesterase', other_designations='acylcholine acylhydrolase|butyrylcholine esterase|choline esterase II|cholinesterase (serum) 2|cholinesterase 1|pseudocholinesterase', map_location='3q26.1-q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=165772904, end_position_on_the_genomic_accession=165837472, orientation='minus', exon_count=4, OMIM=177400, no_hdr0=''),
  593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=593, CurrentID=0, Status='live', Symbol='BCKDHA', Aliases='BCKDE1A, MSU, MSUD1, OVD1A', description='branched chain keto acid dehydrogenase E1, alpha polypeptide', other_designations='2-oxoisovalerate dehydrogenase (lipoamide)|BCKDH E1-alpha|branched chain keto acid dehydrogenase E1 alpha protein|branched-chain alpha-keto acid dehydrogenase E1 component alpha chain', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41397789, end_position_on_the_genomic_accession=41425005, orientation='plus', exon_count=9, OMIM=608348, no_hdr0=''),
  594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594, CurrentID=0, Status='live', Symbol='BCKDHB', Aliases='E1B, dJ279A18.1', description='branched chain keto acid dehydrogenase E1, beta polypeptide', other_designations='2-oxoisovalerate dehydrogenase beta subunit|BCKDE1B|BCKDH E1-beta|E1b-beta subunit of the branched-chain complex|branched chain alpha-ketoacid dehydrogenase E1-beta subunit|branched-chain alpha-keto acid dehydrogenase E1 component beta chain', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80106627, end_position_on_the_genomic_accession=80346270, orientation='plus', exon_count=16, OMIM=248611, no_hdr0=''),
  595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595, CurrentID=0, Status='live', Symbol='CCND1', Aliases='BCL1, D11S287E, PRAD1, U21B31', description='cyclin D1', other_designations='B-cell CLL/lymphoma 1|B-cell lymphoma 1 protein|BCL-1 oncogene|PRAD1 oncogene', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69641105, end_position_on_the_genomic_accession=69654474, orientation='plus', exon_count=6, OMIM=168461, no_hdr0=''),
  596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=596, CurrentID=0, Status='live', Symbol='BCL2', Aliases='Bcl-2, PPP1R50', description='B-cell CLL/lymphoma 2', other_designations='protein phosphatase 1, regulatory subunit 50', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63123346, end_position_on_the_genomic_accession=63319778, orientation='minus', exon_count=6, OMIM=151430, no_hdr0=''),
  597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=597, CurrentID=0, Status='live', Symbol='BCL2A1', Aliases='ACC-1, ACC-2, ACC1, ACC2, BCL2L5, BFL1, GRS, HBPA1', description='BCL2-related protein A1', other_designations='bcl-2-like protein 5|bcl2-L-5|hematopoietic BCL2-related protein A1|hemopoietic-specific early response protein|protein BFL-1', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79960890, end_position_on_the_genomic_accession=79971301, orientation='minus', exon_count=3, OMIM=601056, no_hdr0=''),
  598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=598, CurrentID=0, Status='live', Symbol='BCL2L1', Aliases='BCL-XL/S, BCL2L, BCLX, Bcl-X, PPP1R52', description='BCL2-like 1', other_designations='apoptosis regulator Bcl-X|protein phosphatase 1, regulatory subunit 52', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31664452, end_position_on_the_genomic_accession=31723964, orientation='minus', exon_count=8, OMIM=600039, no_hdr0=''),
  599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=599, CurrentID=0, Status='live', Symbol='BCL2L2', Aliases='BCL-W, BCL2-L-2, BCLW, PPP1R51', description='BCL2-like 2', other_designations='apoptosis regulator BCL-W|protein phosphatase 1, regulatory subunit 51', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23306762, end_position_on_the_genomic_accession=23311759, orientation='plus', exon_count=4, OMIM=601931, no_hdr0=''),
  100270680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270680, CurrentID=0, Status='live', Symbol='CASC11', Aliases='CARLo-7, LINC00990, TCONS_00014535', description='cancer susceptibility candidate 11 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 990', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127700608, end_position_on_the_genomic_accession=127733967, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=601, CurrentID=0, Status='live', Symbol='BAK1P2', Aliases='BAK3, BAK3P, BCL2L7P2', description='BCL2-antagonist/killer 1 pseudogene 2', other_designations='BCL2-like 7 pseudogene 2', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130545543, end_position_on_the_genomic_accession=130547654, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=602, CurrentID=0, Status='live', Symbol='BCL3', Aliases='BCL4, D19S37', description='B-cell CLL/lymphoma 3', other_designations='B-cell leukemia/lymphoma 3|B-cell lymphoma 3-encoded protein|BCL-3|chronic lymphatic leukemia protein|proto-oncogene BCL3', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44748045, end_position_on_the_genomic_accession=44760044, orientation='plus', exon_count=11, OMIM=109560, no_hdr0=''),
  603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=603, CurrentID=0, Status='live', Symbol='BCL5', Aliases='', description='B-cell CLL/lymphoma 5', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151441, no_hdr0=''),
  604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=604, CurrentID=0, Status='live', Symbol='BCL6', Aliases='BCL5A, LAZ3, ZBTB27, ZNF51, BCL6', description='B-cell CLL/lymphoma 6', other_designations='B-cell lymphoma 5 protein|B-cell lymphoma 6 protein transcript|BCL-5|BCL-6|cys-his2 zinc finger transcription factor|lymphoma-associated zinc finger gene on chromosome 3|protein LAZ-3|zinc finger and BTB domain-containing protein 27|zinc finger protein 51|zinc finger transcription factor BCL6S', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187721377, end_position_on_the_genomic_accession=187745725, orientation='minus', exon_count=12, OMIM=109565, no_hdr0=''),
  605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=605, CurrentID=0, Status='live', Symbol='BCL7A', Aliases='BCL7', description='B-cell CLL/lymphoma 7A', other_designations='B-cell CLL/lymphoma-7', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122021886, end_position_on_the_genomic_accession=122062044, orientation='plus', exon_count=6, OMIM=601406, no_hdr0=''),
  606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606, CurrentID=0, Status='live', Symbol='NBEAP1', Aliases='BCL8, BCL8A', description='neurobeachin pseudogene 1', other_designations='B-cell CLL/lymphoma 8', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20669468, end_position_on_the_genomic_accession=20756151, orientation='minus', exon_count=3, OMIM=601889, no_hdr0=''),
  607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=607, CurrentID=0, Status='live', Symbol='BCL9', Aliases='LGS', description='B-cell CLL/lymphoma 9', other_designations='B-cell lymphoma 9 protein|bcl-9|protein legless homolog', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147541401, end_position_on_the_genomic_accession=147626219, orientation='plus', exon_count=12, OMIM=602597, no_hdr0=''),
  608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=608, CurrentID=0, Status='live', Symbol='TNFRSF17', Aliases='BCM, BCMA, CD269, TNFRSF13A', description='tumor necrosis factor receptor superfamily member 17', other_designations='B cell maturation antigen|B-cell maturation factor|B-cell maturation protein|tumor necrosis factor receptor superfamily, member 17', map_location='16p13.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11965107, end_position_on_the_genomic_accession=11968068, orientation='plus', exon_count=3, OMIM=109545, no_hdr0=''),
  610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=610, CurrentID=0, Status='live', Symbol='HCN2', Aliases='BCNG-2, BCNG2, HAC-1', description='hyperpolarization activated cyclic nucleotide gated potassium channel 2', other_designations='brain cyclic nucleotide gated channel 2|brain cyclic nucleotide-gated channel 2|hyperpolarization activated cyclic nucleotide-gated potassium channel 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=589893, end_position_on_the_genomic_accession=617159, orientation='plus', exon_count=8, OMIM=602781, no_hdr0=''),
  611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=611, CurrentID=0, Status='live', Symbol='OPN1SW', Aliases='BCP, BOP, CBT', description='opsin 1 (cone pigments), short-wave-sensitive', other_designations='blue cone photoreceptor pigment|blue-sensitive opsin', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128772489, end_position_on_the_genomic_accession=128775790, orientation='minus', exon_count=5, OMIM=613522, no_hdr0=''),
  613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613, CurrentID=0, Status='live', Symbol='BCR', Aliases='ALL1, CML, D22S11, D22S662, PHL, BCR', description='breakpoint cluster region', other_designations='BCR/FGFR1 chimera protein|FGFR1/BCR chimera protein|renal carcinoma antigen NY-REN-26', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23180365, end_position_on_the_genomic_accession=23318037, orientation='plus', exon_count=23, OMIM=151410, no_hdr0=''),
  616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=616, CurrentID=0, Status='live', Symbol='BCRP4', Aliases='BCR-4, BCR4, BCRL4', description='breakpoint cluster region pseudogene 4', other_designations='breakpoint cluster region-like 4', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22630065, end_position_on_the_genomic_accession=22636584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=617, CurrentID=0, Status='live', Symbol='BCS1L', Aliases='BCS, BCS1, BJS, FLNMS, GRACILE, Hs.6719, MC3DN1, PTD, h-BCS', description='BCS1 homolog, ubiquinol-cytochrome c reductase complex chaperone', other_designations='BC1 (ubiquinol-cytochrome c reductase) synthesis-like|BCS1-like protein|h-BCS1|mitochondrial complex III assembly', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218659656, end_position_on_the_genomic_accession=218663443, orientation='plus', exon_count=10, OMIM=603647, no_hdr0=''),
  618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=618, CurrentID=0, Status='live', Symbol='BCYRN1', Aliases='BC200, BC200a, LINC00004, NCRNA00004', description='brain cytoplasmic RNA 1', other_designations='BC200-alpha|brain cytoplasmic RNA 1 (analog of mouse Bc1)|brain cytoplasmic RNA 1 (non-protein coding)|brain cytoplasmic RNA 1, Bc1 analog|brain cytoplasmic RNA 200-alpha|long intergenic non-protein coding RNA 4|non-protein coding RNA 4', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47335315, end_position_on_the_genomic_accession=47335514, orientation='plus', exon_count=1, OMIM=606089, no_hdr0=''),
  619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619, CurrentID=0, Status='live', Symbol='BCYRN1P1', Aliases='BC200b', description='brain cytoplasmic RNA 1, pseudogene 1', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139720579, end_position_on_the_genomic_accession=139720775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=620, CurrentID=0, Status='live', Symbol='BCYRN1P2', Aliases='BC200g', description='brain cytoplasmic RNA 1, pseudogene 2', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=622, CurrentID=0, Status='live', Symbol='BDH1', Aliases='BDH, SDR9C1', description='3-hydroxybutyrate dehydrogenase, type 1', other_designations='(R)-3-hydroxybutyrate dehydrogenase|3-hydroxybutyrate dehydrogenase (heart, mitochondrial)|short chain dehydrogenase/reductase family 9C member 1|short chain dehydrogenase/reductase family 9C, member 1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197509783, end_position_on_the_genomic_accession=197573323, orientation='minus', exon_count=13, OMIM=603063, no_hdr0=''),
  623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=623, CurrentID=0, Status='live', Symbol='BDKRB1', Aliases='B1BKR, B1R, BDKRB2, BKB1R, BKR1, BRADYB1', description='bradykinin receptor B1', other_designations='BK-1 receptor|bradykinin B1 receptor|bradykinin receptor 1|bradykinin receptor B2', map_location='14q32.1-q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96256210, end_position_on_the_genomic_accession=96264763, orientation='plus', exon_count=3, OMIM=600337, no_hdr0=''),
  624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=624, CurrentID=0, Status='live', Symbol='BDKRB2', Aliases='B2R, BK-2, BK2, BKR2, BRB2', description='bradykinin receptor B2', other_designations='BK-2 receptor', map_location='14q32.1-q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96204798, end_position_on_the_genomic_accession=96244329, orientation='plus', exon_count=3, OMIM=113503, no_hdr0=''),
  627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=627, CurrentID=0, Status='live', Symbol='BDNF', Aliases='ANON2, BULN2', description='brain-derived neurotrophic factor', other_designations='abrineurin|neurotrophin', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27654893, end_position_on_the_genomic_accession=27722058, orientation='minus', exon_count=12, OMIM=113505, no_hdr0=''),
  628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=628, CurrentID=0, Status='live', Symbol='BEVI', Aliases='', description='baboon M7 virus integration site', other_designations='Baboon M7 virus replication', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109180, no_hdr0=''),
  629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=629, CurrentID=0, Status='live', Symbol='CFB', Aliases='AHUS4, ARMD14, BF, BFD, CFABD, FB, FBI12, GBG, H2-Bf, PBF2, CFB', description='complement factor B', other_designations='B-factor, properdin|C3 proaccelerator|C3 proactivator|C3/C5 convertase|glycine-rich beta glycoprotein|glycine-rich beta-glycoprotein|properdin factor B', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31945944, end_position_on_the_genomic_accession=31952084, orientation='plus', exon_count=18, OMIM=138470, no_hdr0=''),
  631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=631, CurrentID=0, Status='live', Symbol='BFSP1', Aliases='CP115, CP94, CTRCT33, LIFL-H', description='beaded filament structural protein 1', other_designations='beaded filament structural protein 1, filensin|cytoskeletal protein, 115 KD|lens fiber cell beaded-filament structural protein CP 115|lens intermediate filament-like heavy', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17493905, end_position_on_the_genomic_accession=17569220, orientation='minus', exon_count=14, OMIM=603307, no_hdr0=''),
  632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=632, CurrentID=0, Status='live', Symbol='BGLAP', Aliases='BGP, OC, OCN', description='bone gamma-carboxyglutamate (gla) protein', other_designations='bone gamma-carboxyglutamate (gla) protein (osteocalcin)|gamma-carboxyglutamic acid-containing protein', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156242160, end_position_on_the_genomic_accession=156243332, orientation='plus', exon_count=4, OMIM=112260, no_hdr0=''),
  633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=633, CurrentID=0, Status='live', Symbol='BGN', Aliases='DSPG1, PG-S1, PGI, SLRR1A', description='biglycan', other_designations='bone/cartilage proteoglycan-I|dermatan sulphate proteoglycan I|small leucine-rich protein 1A', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153494889, end_position_on_the_genomic_accession=153509546, orientation='plus', exon_count=8, OMIM=301870, no_hdr0=''),
  634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=634, CurrentID=0, Status='live', Symbol='CEACAM1', Aliases='BGP, BGP1, BGPI', description='carcinoembryonic antigen-related cell adhesion molecule 1 (biliary glycoprotein)', other_designations='CD66a antigen|antigen CD66', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42507306, end_position_on_the_genomic_accession=42528509, orientation='minus', exon_count=10, OMIM=109770, no_hdr0=''),
  635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=635, CurrentID=0, Status='live', Symbol='BHMT', Aliases='BHMT1, HEL-S-61p', description='betaine--homocysteine S-methyltransferase', other_designations='epididymis secretory sperm binding protein Li 61p', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79111781, end_position_on_the_genomic_accession=79132290, orientation='plus', exon_count=8, OMIM=602888, no_hdr0=''),
  636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=636, CurrentID=0, Status='live', Symbol='BICD1', Aliases='BICD', description='bicaudal D homolog 1 (Drosophila)', other_designations='bic-D 1|cytoskeleton-like bicaudal D protein homolog 1', map_location='12p11.2-p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32106736, end_position_on_the_genomic_accession=32383650, orientation='plus', exon_count=12, OMIM=602204, no_hdr0=''),
  637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=637, CurrentID=0, Status='live', Symbol='BID', Aliases='FP497', description='BH3 interacting domain death agonist', other_designations='BID isoform ES(1b)|BID isoform L(2)|BID isoform Si6|Human BID coding sequence|apoptic death agonist|desmocollin type 4|p22 BID', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17734140, end_position_on_the_genomic_accession=17774665, orientation='minus', exon_count=8, OMIM=601997, no_hdr0=''),
  638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=638, CurrentID=0, Status='live', Symbol='BIK', Aliases='BIP1, BP4, NBK', description='BCL2-interacting killer', other_designations='BCL2-interacting killer (apoptosis-inducing)|apoptosis inducer NBK|apoptosis-inducing NBK|natural born killer', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43110748, end_position_on_the_genomic_accession=43129712, orientation='plus', exon_count=5, OMIM=603392, no_hdr0=''),
  639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=639, CurrentID=0, Status='live', Symbol='PRDM1', Aliases='BLIMP1, PRDI-BF1', description='PR domain containing 1, with ZNF domain', other_designations='B-lymphocyte-induced maturation protein 1|BLIMP-1|PRDI-binding factor-1|beta-interferon gene positive-regulatory domain I binding factor', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106086320, end_position_on_the_genomic_accession=106109939, orientation='plus', exon_count=10, OMIM=603423, no_hdr0=''),
  640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=640, CurrentID=0, Status='live', Symbol='BLK', Aliases='MODY11', description='BLK proto-oncogene, Src family tyrosine kinase', other_designations='B lymphoid tyrosine kinase|BLK nonreceptor tyrosine kinase|b lymphocyte kinase|p55-Blk', map_location='8p23-p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11493991, end_position_on_the_genomic_accession=11564599, orientation='plus', exon_count=16, OMIM=191305, no_hdr0=''),
  641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641, CurrentID=0, Status='live', Symbol='BLM', Aliases='BS, RECQ2, RECQL2, RECQL3', description='Bloom syndrome, RecQ helicase-like', other_designations='DNA helicase, RecQ-like type 2|recQ protein-like 3', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90717327, end_position_on_the_genomic_accession=90815462, orientation='plus', exon_count=25, OMIM=604610, no_hdr0=''),
  642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642, CurrentID=0, Status='live', Symbol='BLMH', Aliases='BH, BMH', description='bleomycin hydrolase', other_designations='BLM hydrolase', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30248195, end_position_on_the_genomic_accession=30292166, orientation='minus', exon_count=12, OMIM=602403, no_hdr0=''),
  643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643, CurrentID=0, Status='live', Symbol='CXCR5', Aliases='BLR1, CD185, MDR15', description='chemokine (C-X-C motif) receptor 5', other_designations='Burkitt lymphoma receptor 1, GTP binding protein (chemokine (C-X-C motif) receptor 5)|Burkitt lymphoma receptor 1, GTP-binding protein|CXC-R5|CXCR-5|MDR-15|monocyte-derived receptor 15', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118883766, end_position_on_the_genomic_accession=118896271, orientation='plus', exon_count=2, OMIM=601613, no_hdr0=''),
  644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644, CurrentID=0, Status='live', Symbol='BLVRA', Aliases='BLVR, BVR, BVRA', description='biliverdin reductase A', other_designations='BVR A|biliverdin-IX alpha-reductase', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43758192, end_position_on_the_genomic_accession=43807342, orientation='plus', exon_count=10, OMIM=109750, no_hdr0=''),
  645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645, CurrentID=0, Status='live', Symbol='BLVRB', Aliases='BVRB, FLR, HEL-S-10, SDR43U1', description='biliverdin reductase B', other_designations='BVR-B|FR|GHBP|NADPH-dependent diaphorase|NADPH-flavin reductase|biliverdin reductase B (flavin reductase (NADPH))|biliverdin-IX beta-reductase|epididymis secretory protein Li 10|green heme-binding protein|short chain dehydrogenase/reductase family 43U, member 1', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40447784, end_position_on_the_genomic_accession=40465818, orientation='minus', exon_count=5, OMIM=600941, no_hdr0=''),
  646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646, CurrentID=0, Status='live', Symbol='BNC1', Aliases='BNC, BSN1, HsT19447', description='basonuclin 1', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83255884, end_position_on_the_genomic_accession=83284716, orientation='minus', exon_count=7, OMIM=601930, no_hdr0=''),
  648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648, CurrentID=0, Status='live', Symbol='BMI1', Aliases='FLVI2/BMI1, PCGF4, RNF51, flvi-2/bmi-1', description='BMI1 proto-oncogene, polycomb ring finger', other_designations='B lymphoma Mo-MLV insertion region 1 homolog|BMI1 polycomb ring finger oncogene|BMI1 polycomb ring finger proto-oncogene|murine leukemia viral (bmi-1) oncogene homolog|polycomb group RING finger protein 4|polycomb group protein Bmi1|ring finger protein 51', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22321210, end_position_on_the_genomic_accession=22331485, orientation='plus', exon_count=10, OMIM=164831, no_hdr0=''),
  649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649, CurrentID=0, Status='live', Symbol='BMP1', Aliases='OI13, PCOLC, PCP, PCP2, TLD', description='bone morphogenetic protein 1', other_designations='mammalian tolloid protein|procollagen C-endopeptidase|procollagen C-proteinase 3', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22165140, end_position_on_the_genomic_accession=22212326, orientation='plus', exon_count=25, OMIM=112264, no_hdr0=''),
  650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650, CurrentID=0, Status='live', Symbol='BMP2', Aliases='BDA2A, BMP2', description='bone morphogenetic protein 2', other_designations='bone morphogenetic protein 2A', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6767694, end_position_on_the_genomic_accession=6780263, orientation='plus', exon_count=3, OMIM=112261, no_hdr0=''),
  651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651, CurrentID=0, Status='live', Symbol='BMP3', Aliases='BMP-3A', description='bone morphogenetic protein 3', other_designations='BMP-3|bone morphogenetic protein 3 (osteogenic)|bone morphogenetic protein 3A|bone morphogenetic protein-3|osteogenin', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81030622, end_position_on_the_genomic_accession=81057531, orientation='plus', exon_count=3, OMIM=112263, no_hdr0=''),
  652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652, CurrentID=0, Status='live', Symbol='BMP4', Aliases='BMP2B, BMP2B1, MCOPS6, OFC11, ZYME', description='bone morphogenetic protein 4', other_designations='bone morphogenetic protein 2B', map_location='14q22-q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53949736, end_position_on_the_genomic_accession=53956836, orientation='minus', exon_count=6, OMIM=112262, no_hdr0=''),
  653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653, CurrentID=0, Status='live', Symbol='BMP5', Aliases='', description='bone morphogenetic protein 5', other_designations='BMP-5|bone morphogenenic protein-5', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55753653, end_position_on_the_genomic_accession=55875593, orientation='minus', exon_count=8, OMIM=112265, no_hdr0=''),
  654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654, CurrentID=0, Status='live', Symbol='BMP6', Aliases='VGR, VGR1', description='bone morphogenetic protein 6', other_designations='BMP-6|VG-1-R|VG-1-related protein|VGR-1|Vg1-related sequence|vegetal related growth factor (TGFB-related)|vegetal-related (TGFB related) cytokine', map_location='6p24-p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7726778, end_position_on_the_genomic_accession=7881728, orientation='plus', exon_count=7, OMIM=112266, no_hdr0=''),
  655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=655, CurrentID=0, Status='live', Symbol='BMP7', Aliases='OP-1', description='bone morphogenetic protein 7', other_designations='osteogenic protein 1', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57168753, end_position_on_the_genomic_accession=57266651, orientation='minus', exon_count=7, OMIM=112267, no_hdr0=''),
  656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=656, CurrentID=0, Status='live', Symbol='BMP8B', Aliases='BMP8, OP2', description='bone morphogenetic protein 8b', other_designations='BMP-8|BMP-8B|bone morphogenetic protein 8 (osteogenic protein 2)|dJ118J21.1 (bone morphogenetic protein 8 (osteogenic protein 2))|osteogenic protein 2', map_location='1p35-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39757182, end_position_on_the_genomic_accession=39788861, orientation='minus', exon_count=12, OMIM=602284, no_hdr0=''),
  657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=657, CurrentID=0, Status='live', Symbol='BMPR1A', Aliases='10q23del, ACVRLK3, ALK3, CD292, SKR5', description='bone morphogenetic protein receptor type IA', other_designations='ALK-3|BMP type-1A receptor|BMPR-1A|activin A receptor, type II-like kinase 3|activin receptor-like kinase 3|bone morphogenetic protein receptor, type IA|serine/threonine-protein kinase receptor R5', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86755786, end_position_on_the_genomic_accession=86925188, orientation='plus', exon_count=15, OMIM=601299, no_hdr0=''),
  658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=658, CurrentID=0, Status='live', Symbol='BMPR1B', Aliases='ALK-6, ALK6, AMDD, CDw293', description='bone morphogenetic protein receptor type IB', other_designations='BMP type-1B receptor|BMPR-1B|activin receptor-like kinase 6|bone morphogenetic protein receptor, type IB|serine/threonine receptor kinase', map_location='4q22-q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94757977, end_position_on_the_genomic_accession=95158450, orientation='plus', exon_count=17, OMIM=603248, no_hdr0=''),
  659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=659, CurrentID=0, Status='live', Symbol='BMPR2', Aliases='BMPR-II, BMPR3, BMR2, BRK-3, POVD1, PPH1, T-ALK', description='bone morphogenetic protein receptor type II', other_designations='BMP type II receptor|BMP type-2 receptor|BMPR-2|bone morphogenetic protein receptor, type II (serine/threonine kinase)|type II activin receptor-like kinase|type II receptor for bone morphogenetic protein-4', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202376310, end_position_on_the_genomic_accession=202567751, orientation='plus', exon_count=13, OMIM=600799, no_hdr0=''),
  660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=660, CurrentID=0, Status='live', Symbol='BMX', Aliases='ETK, PSCTK2, PSCTK3', description='BMX non-receptor tyrosine kinase', other_designations='BTK-like on X chromosome|Etk/Bmx cytosolic tyrosine kinase|NTK38 tyrosine kinase|bone marrow tyrosine kinase gene in chromosome X protein|epithelial and endothelial tyrosine kinase', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15500777, end_position_on_the_genomic_accession=15556529, orientation='plus', exon_count=20, OMIM=300101, no_hdr0=''),
  661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=661, CurrentID=0, Status='live', Symbol='POLR3D', Aliases='BN51T, RPC4, RPC53, TSBN51', description='polymerase (RNA) III (DNA directed) polypeptide D, 44kDa', other_designations='BN51 (BHK21) temperature sensitivity complementing|DNA-directed RNA polymerase III 47 kDa polypeptide|DNA-directed RNA polymerase III subunit D|RNA polymerase III 47 kDa subunit|RNA polymerase III subunit C4|RPC53 homolog|temperature sensitive complementation, cell cycle specific, tsBN51', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22245104, end_position_on_the_genomic_accession=22254355, orientation='plus', exon_count=9, OMIM=187280, no_hdr0=''),
  662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=662, CurrentID=0, Status='live', Symbol='BNIP1', Aliases='NIP1, SEC20, TRG-8', description='BCL2/adenovirus E1B 19kDa interacting protein 1', other_designations='BCL2/adenovirus E1B 19 kDa protein-interacting protein 1|transformation-related gene 8 protein', map_location='5q33-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173144442, end_position_on_the_genomic_accession=173164387, orientation='plus', exon_count=9, OMIM=603291, no_hdr0=''),
  663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=663, CurrentID=0, Status='live', Symbol='BNIP2', Aliases='BNIP-2, NIP2', description='BCL2/adenovirus E1B 19kDa interacting protein 2', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59659146, end_position_on_the_genomic_accession=59689529, orientation='minus', exon_count=12, OMIM=603292, no_hdr0=''),
  664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664, CurrentID=0, Status='live', Symbol='BNIP3', Aliases='NIP3', description='BCL2/adenovirus E1B 19kDa interacting protein 3', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131967683, end_position_on_the_genomic_accession=131982013, orientation='minus', exon_count=6, OMIM=603293, no_hdr0=''),
  665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=665, CurrentID=0, Status='live', Symbol='BNIP3L', Aliases='BNIP3a, NIX', description='BCL2/adenovirus E1B 19kDa interacting protein 3-like', other_designations='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3A|BCL2/adenovirus E1B 19-kd protein-interacting protein 3a|NIP3-like protein X|NIP3L|adenovirus E1B19k-binding protein B5', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26383007, end_position_on_the_genomic_accession=26413129, orientation='plus', exon_count=8, OMIM=605368, no_hdr0=''),
  666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=666, CurrentID=0, Status='live', Symbol='BOK', Aliases='BCL2L9L, BOK', description='BCL2-related ovarian killer', other_designations='bcl-2-like protein 9|bcl2-L-9|hBOK', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241558064, end_position_on_the_genomic_accession=241574138, orientation='plus', exon_count=8, OMIM=605404, no_hdr0=''),
  667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=667, CurrentID=0, Status='live', Symbol='DST', Aliases='BP240, BPA, BPAG1, CATX-15, CATX15, D6S1101, DMH, DT, EBSB2, HSAN6, MACF2', description='dystonin', other_designations='bullous pemphigoid antigen 1|dystonia musculorum protein|hemidesmosomal plaque protein|trabeculin-beta', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56457987, end_position_on_the_genomic_accession=56954628, orientation='minus', exon_count=111, OMIM=113810, no_hdr0=''),
  668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=668, CurrentID=0, Status='live', Symbol='FOXL2', Aliases='BPES, BPES1, PFRK, PINTO, POF3', description='forkhead box L2', other_designations='forkhead transcription factor FOXL2', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138944224, end_position_on_the_genomic_accession=138947140, orientation='minus', exon_count=1, OMIM=605597, no_hdr0=''),
  669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=669, CurrentID=0, Status='live', Symbol='BPGM', Aliases='DPGM', description='bisphosphoglycerate mutase', other_designations='2,3-bisphosphoglycerate mutase|2,3-bisphosphoglycerate mutase, erythrocyte|2,3-bisphosphoglycerate synthase|2,3-diphosphoglycerate mutase|BPG-dependent PGAM|erythrocyte 2,3-bisphosphoglycerate mutase', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134646779, end_position_on_the_genomic_accession=134679816, orientation='plus', exon_count=6, OMIM=613896, no_hdr0=''),
  670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=670, CurrentID=0, Status='live', Symbol='BPHL', Aliases='BPH-RP, MCNAA, VACVASE', description='biphenyl hydrolase-like (serine hydrolase)', other_designations='biphenyl hydrolase-related protein|breast epithelial mucin-associated antigen|valacyclovirase', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3118376, end_position_on_the_genomic_accession=3153198, orientation='plus', exon_count=9, OMIM=603156, no_hdr0=''),
  671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=671, CurrentID=0, Status='live', Symbol='BPI', Aliases='BPIFD1, rBPI', description='bactericidal/permeability-increasing protein', other_designations='BPI fold containing family D, member 1|CAP 57|recombinant BPI holoprotein, rBPI', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38304150, end_position_on_the_genomic_accession=38337505, orientation='plus', exon_count=16, OMIM=109195, no_hdr0=''),
  672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=672, CurrentID=0, Status='live', Symbol='BRCA1', Aliases='BRCAI, BRCC1, BROVCA1, FANCS, IRIS, PNCA4, PPP1R53, PSCP, RNF53', description='breast cancer 1', other_designations='BRCA1/BRCA2-containing complex, subunit 1|Fanconi anemia, complementation group S|RING finger protein 53|breast and ovarian cancer susceptibility protein 1|breast and ovarian cancer sususceptibility protein 1|breast cancer 1, early onset|early onset breast cancer 1|protein phosphatase 1, regulatory subunit 53|truncated breast and ovarian cancer susceptibility protein 1|truncated breast and ovarian cancer sususceptibility protein 1|truncated breast cancer type 1 susceptibility protein', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43044295, end_position_on_the_genomic_accession=43125483, orientation='minus', exon_count=24, OMIM=113705, no_hdr0=''),
  673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=673, CurrentID=0, Status='live', Symbol='BRAF', Aliases='B-RAF11, NS7, RAFB1, BRAF', description='B-Raf proto-oncogene, serine/threonine kinase', other_designations='94 kDa B-raf protein|B-Raf proto-oncogene serine/threonine-protein kinase (p94)|murine sarcoma viral (v-raf) oncogene homolog B1|proto-oncogene B-Raf|v-raf murine sarcoma viral oncogene homolog B|v-raf murine sarcoma viral oncogene homolog B1', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140715951, end_position_on_the_genomic_accession=140924764, orientation='minus', exon_count=22, OMIM=164757, no_hdr0=''),
  100270754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270754, CurrentID=0, Status='live', Symbol='RPL7AP72', Aliases='RPL7A_34_1809', description='ribosomal protein L7a pseudogene 72', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128313281, end_position_on_the_genomic_accession=128314048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=675, CurrentID=0, Status='live', Symbol='BRCA2', Aliases='BRCC2, BROVCA2, FACD, FAD, FAD1, FANCD, FANCD1, GLM3, PNCA2, XRCC11', description='breast cancer 2', other_designations='BRCA1/BRCA2-containing complex, subunit 2|Fanconi anemia group D1 protein|breast and ovarian cancer susceptibility gene, early onset|breast and ovarian cancer susceptibility protein 2|breast cancer 2 tumor suppressor|breast cancer 2, early onset|mutant BRCA2', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32314862, end_position_on_the_genomic_accession=32399850, orientation='plus', exon_count=28, OMIM=600185, no_hdr0=''),
  676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=676, CurrentID=0, Status='live', Symbol='BRDT', Aliases='BRD6, CT9', description='bromodomain, testis-specific', other_designations='RING3-like protein|cancer/testis antigen 9', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91949371, end_position_on_the_genomic_accession=92014428, orientation='plus', exon_count=22, OMIM=602144, no_hdr0=''),
  677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677, CurrentID=0, Status='live', Symbol='ZFP36L1', Aliases='BRF1, Berg36, ERF-1, ERF1, RNF162B, TIS11B, cMG1', description='ZFP36 ring finger protein-like 1', other_designations='EGF-response factor 1|butyrate response factor 1|early response factor Berg36|zinc finger protein 36, C3H type-like 1|zinc finger protein, C3H type, 36-like 1', map_location='14q22-q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68787655, end_position_on_the_genomic_accession=68796243, orientation='minus', exon_count=4, OMIM=601064, no_hdr0=''),
  678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=678, CurrentID=0, Status='live', Symbol='ZFP36L2', Aliases='BRF2, ERF-2, ERF2, RNF162C, TIS11D', description='ZFP36 ring finger protein-like 2', other_designations='EGF-response factor 2|ZFP36-like 2|butyrate response factor 2|zinc finger protein 36, C3H type-like 1|zinc finger protein 36, C3H type-like 2|zinc finger protein, C3H type, 36-like 2', map_location='2p22.3-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43222402, end_position_on_the_genomic_accession=43226606, orientation='minus', exon_count=2, OMIM=612053, no_hdr0=''),
  131185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131185, CurrentID=0, Status='live', Symbol='RAD23BP1', Aliases='', description='RAD23 homolog B, nucleotide excision repair protein pseudogene 1', other_designations='RAD23 homolog B (S. cerevisiae) pseudogene 1', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18538673, end_position_on_the_genomic_accession=18539896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=680, CurrentID=0, Status='live', Symbol='BRS3', Aliases='BB3', description='bombesin like receptor 3', other_designations='BRS-3|G-protein coupled receptor|bombesin receptor subtype 3|bombesin-like receptor 3', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136463815, end_position_on_the_genomic_accession=136493774, orientation='plus', exon_count=5, OMIM=300107, no_hdr0=''),
  682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=682, CurrentID=0, Status='live', Symbol='BSG', Aliases='5F7, CD147, EMMPRIN, M6, OK, TCSF', description='basigin (Ok blood group)', other_designations='CD147 antigen|OK blood group antigen|collagenase stimulatory factor|extracellular matrix metalloproteinase inducer|leukocyte activation antigen M6|tumor cell-derived collagenase stimulatory factor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=571277, end_position_on_the_genomic_accession=583493, orientation='plus', exon_count=10, OMIM=109480, no_hdr0=''),
  683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=683, CurrentID=0, Status='live', Symbol='BST1', Aliases='CD157', description='bone marrow stromal cell antigen 1', other_designations='ADP-ribosyl cyclase 2|BST-1|NAD(+) nucleosidase|bone marrow stromal antigen 1|cADPr hydrolase 2|cyclic ADP-ribose hydrolase 2', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15702950, end_position_on_the_genomic_accession=15774190, orientation='plus', exon_count=15, OMIM=600387, no_hdr0=''),
  684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=684, CurrentID=0, Status='live', Symbol='BST2', Aliases='CD317, TETHERIN', description='bone marrow stromal cell antigen 2', other_designations='BST-2|HM1.24 antigen|NPC-A-7', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17402939, end_position_on_the_genomic_accession=17405648, orientation='minus', exon_count=5, OMIM=600534, no_hdr0=''),
  685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=685, CurrentID=0, Status='live', Symbol='BTC', Aliases='', description='betacellulin', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74744759, end_position_on_the_genomic_accession=74794672, orientation='minus', exon_count=7, OMIM=600345, no_hdr0=''),
  686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=686, CurrentID=0, Status='live', Symbol='BTD', Aliases='', description='biotinidase', other_designations='biotinase', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15601352, end_position_on_the_genomic_accession=15647642, orientation='plus', exon_count=9, OMIM=609019, no_hdr0=''),
  687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=687, CurrentID=0, Status='live', Symbol='KLF9', Aliases='BTEB, BTEB1', description='Kruppel-like factor 9', other_designations='BTE-binding protein 1|GC-box-binding protein 1|basic transcription element binding protein 1|basic transcription element-binding protein 1|transcription factor BTEB1', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70384597, end_position_on_the_genomic_accession=70414657, orientation='minus', exon_count=2, OMIM=602902, no_hdr0=''),
  688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=688, CurrentID=0, Status='live', Symbol='KLF5', Aliases='BTEB2, CKLF, IKLF', description='Kruppel-like factor 5 (intestinal)', other_designations='BTE-binding protein 2|GC box binding protein 2|Klf5C isoform|basic transcription element binding protein 2|colon krueppel-like factor|colon kruppel-like factor|intestinal-enriched krueppel-like factor|intestinal-enriched kruppel-like factor|transcription factor BTEB2', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73054976, end_position_on_the_genomic_accession=73077542, orientation='plus', exon_count=5, OMIM=602903, no_hdr0=''),
  689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=689, CurrentID=0, Status='live', Symbol='BTF3', Aliases='BETA-NACa, BTF3b, NACB, BTF3', description='basic transcription factor 3', other_designations='NAC-beta|RNA polymerase B transcription factor 3|nascent polypeptide-associated complex subunit beta|nascent-polypeptide-associated complex beta polypeptide', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73498425, end_position_on_the_genomic_accession=73505623, orientation='plus', exon_count=6, OMIM=602542, no_hdr0=''),
  690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=690, CurrentID=0, Status='live', Symbol='BTF3P11', Aliases='BTF3L1, HUMBTFB, OCIF, OPG, TNFRSF11B', description='basic transcription factor 3 pseudogene 11', other_designations='BTF3 homologue|basic transcription factor 3, like 1 pseudogene|basic transcription factor 3-like 1, pseudogene|lambda h27A', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76928451, end_position_on_the_genomic_accession=76929090, orientation='plus', exon_count=1, OMIM=602543, no_hdr0=''),
  693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693, CurrentID=0, Status='live', Symbol='BTF3P1', Aliases='HUMBTFA', description='basic transcription factor 3, pseudogene 1', other_designations='lambda h16', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51721256, end_position_on_the_genomic_accession=51723046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=694, CurrentID=0, Status='live', Symbol='BTG1', Aliases='', description='B-cell translocation gene 1, anti-proliferative', other_designations='B-cell translocation gene 1 protein', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92140278, end_position_on_the_genomic_accession=92145897, orientation='minus', exon_count=2, OMIM=109580, no_hdr0=''),
  695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=695, CurrentID=0, Status='live', Symbol='BTK', Aliases='AGMX1, AT, ATK, BPK, IMD1, PSCTK1, XLA', description='Bruton tyrosine kinase', other_designations="B-cell progenitor kinase|Bruton agammaglobulinemia tyrosine kinase|Bruton's tyrosine kinase|agammaglobulinaemia tyrosine kinase|dominant-negative kinase-deficient Brutons tyrosine kinase|truncated Bruton agammaglobulinemia tyrosine kinase|tyrosine-protein kinase BTK isoform (lacking exon 13 to 17)|tyrosine-protein kinase BTK isoform (lacking exon 14)", map_location='Xq21.33-q22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101349447, end_position_on_the_genomic_accession=101390796, orientation='minus', exon_count=21, OMIM=300300, no_hdr0=''),
  696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=696, CurrentID=0, Status='live', Symbol='BTN1A1', Aliases='BT, BTN, BTN1', description='butyrophilin subfamily 1 member A1', other_designations='bK14H9.2 (butyrophilin, subfamily 1, member A1)|butyrophilin, subfamily 1, member A1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26499306, end_position_on_the_genomic_accession=26510425, orientation='plus', exon_count=8, OMIM=601610, no_hdr0=''),
  698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=698, CurrentID=0, Status='live', Symbol='BTS1', Aliases='', description='bladder tumor susceptibility 1', other_designations='bladded tumor susceptibility 1', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=699, CurrentID=0, Status='live', Symbol='BUB1', Aliases='BUB1AL, hBUB1, BUB1', description='BUB1, mitotic checkpoint serine/threonine kinase', other_designations='BUB1 budding uninhibited by benzimidazoles 1 homolog|BUB1 mitotic checkpoint serine/threonine kinase|budding uninhibited by benzimidazoles 1 homolog|mitotic spindle checkpoint kinase|putative serine/threonine-protein kinase', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110637527, end_position_on_the_genomic_accession=110678107, orientation='minus', exon_count=25, OMIM=602452, no_hdr0=''),
  700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=700, CurrentID=0, Status='live', Symbol='BVR1', Aliases='', description='Burkitt lymphoma variant rearranging region 1', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=701, CurrentID=0, Status='live', Symbol='BUB1B', Aliases='BUB1beta, BUBR1, Bub1A, MAD3L, MVA1, SSK1, hBUBR1', description='BUB1B, mitotic checkpoint serine/threonine kinase', other_designations='BUB1 mitotic checkpoint serine/threonine kinase B|MAD3/BUB1-related protein kinase|budding uninhibited by benzimidazoles 1 homolog beta|mitotic checkpoint kinase MAD3L', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40161009, end_position_on_the_genomic_accession=40221136, orientation='plus', exon_count=23, OMIM=602860, no_hdr0=''),
  705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=705, CurrentID=0, Status='live', Symbol='BYSL', Aliases='BYSTIN', description='bystin like', other_designations='by the ribosomal protein s6 gene, drosophila, homolog-like|bystin-like', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41907847, end_position_on_the_genomic_accession=41933046, orientation='plus', exon_count=8, OMIM=603871, no_hdr0=''),
  706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=706, CurrentID=0, Status='live', Symbol='TSPO', Aliases='BPBS, BZRP, DBI, IBP, MBR, PBR, PBS, PKBS, PTBR, mDRC, pk18', description='translocator protein', other_designations='benzodiazepine peripheral binding site|mitochondrial benzodiazepine receptor|peripheral-type benzodiazepine receptor|translocator protein (18kDa)', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43151514, end_position_on_the_genomic_accession=43163242, orientation='plus', exon_count=5, OMIM=109610, no_hdr0=''),
  708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=708, CurrentID=0, Status='live', Symbol='C1QBP', Aliases='GC1QBP, HABP1, SF2p32, gC1Q-R, gC1qR, p32', description='complement component 1, q subcomponent binding protein', other_designations='ASF/SF2-associated protein p32|C1q globular domain-binding protein|glycoprotein gC1qBP|hyaluronan-binding protein 1|mitochondrial matrix protein p32|p33|splicing factor SF2-associated protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5432779, end_position_on_the_genomic_accession=5439151, orientation='minus', exon_count=6, OMIM=601269, no_hdr0=''),
  102723373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723373, CurrentID=0, Status='live', Symbol='LOC102723373', Aliases='', description='uncharacterized LOC102723373', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53373493, end_position_on_the_genomic_accession=53384259, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=710, CurrentID=0, Status='live', Symbol='SERPING1', Aliases='C1IN, C1INH, C1NH, HAE1, HAE2', description='serpin peptidase inhibitor, clade G (C1 inhibitor), member 1', other_designations='C1 esterase inhibitor|C1-inhibiting factor|complement component 1 inhibitor|serine/cysteine proteinase inhibitor clade G member 1|serpin G1|serpin peptidase inhibitor clade G member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57597554, end_position_on_the_genomic_accession=57614853, orientation='plus', exon_count=8, OMIM=606860, no_hdr0=''),
  711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=711, CurrentID=0, Status='live', Symbol='ERC2-IT1', Aliases='C1orf1, C3orf51, Po42', description='ERC2 intronic transcript 1', other_designations='ERC2 intronic transcript 1 (non-protein coding)', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55657206, end_position_on_the_genomic_accession=55659469, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=712, CurrentID=0, Status='live', Symbol='C1QA', Aliases='', description='complement component 1, q subcomponent, A chain', other_designations='complement component 1, q subcomponent, alpha polypeptide|complement component C1q, A chain', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22636625, end_position_on_the_genomic_accession=22639682, orientation='plus', exon_count=3, OMIM=120550, no_hdr0=''),
  713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=713, CurrentID=0, Status='live', Symbol='C1QB', Aliases='', description='complement component 1, q subcomponent, B chain', other_designations='complement component 1, q subcomponent, beta polypeptide|complement component C1q, B chain|complement subcomponent C1q chain B', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22653179, end_position_on_the_genomic_accession=22661552, orientation='plus', exon_count=4, OMIM=120570, no_hdr0=''),
  714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=714, CurrentID=0, Status='live', Symbol='C1QC', Aliases='C1Q-C, C1QG', description='complement component 1, q subcomponent, C chain', other_designations='complement component 1, q subcomponent, gamma polypeptide', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22643625, end_position_on_the_genomic_accession=22648110, orientation='plus', exon_count=3, OMIM=120575, no_hdr0=''),
  715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=715, CurrentID=0, Status='live', Symbol='C1R', Aliases='', description='complement component 1, r subcomponent', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7085860, end_position_on_the_genomic_accession=7092447, orientation='minus', exon_count=11, OMIM=613785, no_hdr0=''),
  716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=716, CurrentID=0, Status='live', Symbol='C1S', Aliases='', description='complement component 1, s subcomponent', other_designations='C1 esterase|basic proline-rich peptide IB-1|complement component 1 subcomponent s', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7060676, end_position_on_the_genomic_accession=7071032, orientation='plus', exon_count=13, OMIM=120580, no_hdr0=''),
  717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=717, CurrentID=0, Status='live', Symbol='C2', Aliases='ARMD14, CO2', description='complement component 2', other_designations='C3/C5 convertase|complement component C2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31897785, end_position_on_the_genomic_accession=31945674, orientation='plus', exon_count=20, OMIM=613927, no_hdr0=''),
  718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=718, CurrentID=0, Status='live', Symbol='C3', Aliases='AHUS5, ARMD9, ASPa, C3b, CPAMD1, HEL-S-62p, C3', description='complement component 3', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 1|C3a anaphylatoxin|acylation-stimulating protein cleavage product|complement component C3a|complement component C3b|epididymis secretory sperm binding protein Li 62p|prepro-C3', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6677835, end_position_on_the_genomic_accession=6720682, orientation='minus', exon_count=41, OMIM=120700, no_hdr0=''),
  719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=719, CurrentID=0, Status='live', Symbol='C3AR1', Aliases='AZ3B, C3AR, HNFAG09', description='complement component 3a receptor 1', other_designations='C3a-R|complement component 3 receptor 1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8058323, end_position_on_the_genomic_accession=8066359, orientation='minus', exon_count=2, OMIM=605246, no_hdr0=''),
  720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=720, CurrentID=0, Status='live', Symbol='C4A', Aliases='C42, C4A3, C4A4, C4A6, C4AD, C4S, CO4, CPAMD2, RG, C4A', description='complement component 4A (Rodgers blood group)', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 2|C4A anaphylatoxin|Rodgers form of C4|acidic C4|acidic complement C4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31982057, end_position_on_the_genomic_accession=32002680, orientation='plus', exon_count=41, OMIM=120810, no_hdr0=''),
  721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=721, CurrentID=0, Status='live', Symbol='C4B', Aliases='C4B112, C4B2, C4B3, C4B5, C4BD, C4B_2, C4F, CH, CO4, CPAMD3, C4B', description='complement component 4B (Chido blood group)', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 3|Chido form of C4|basic complement C4|complement C4B1a', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32014795, end_position_on_the_genomic_accession=32035418, orientation='plus', exon_count=41, OMIM=120820, no_hdr0=''),
  722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=722, CurrentID=0, Status='live', Symbol='C4BPA', Aliases='C4BP, PRP', description='complement component 4 binding protein, alpha', other_designations='proline-rich protein', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207104238, end_position_on_the_genomic_accession=207144972, orientation='plus', exon_count=12, OMIM=120830, no_hdr0=''),
  100270803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270803, CurrentID=0, Status='live', Symbol='DEL6PTER', Aliases='C6DELpter', description='Chromosome 6pter deletion syndrome', other_designations='', map_location='6pter-p24', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612582, no_hdr0=''),
  724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724, CurrentID=0, Status='live', Symbol='C4BPAP2', Aliases='C4BPAL2', description='complement component 4 binding protein, alpha pseudogene 2', other_designations='complement component 4 binding protein, alpha-like 2', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207225798, end_position_on_the_genomic_accession=207229202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=725, CurrentID=0, Status='live', Symbol='C4BPB', Aliases='C4BP', description='complement component 4 binding protein, beta', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207088850, end_position_on_the_genomic_accession=207099993, orientation='plus', exon_count=8, OMIM=120831, no_hdr0=''),
  726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=726, CurrentID=0, Status='live', Symbol='CAPN5', Aliases='ADNIV, HTRA3, VRNI, nCL-3', description='calpain 5', other_designations='calpain htra-3|new calpain 3', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77066945, end_position_on_the_genomic_accession=77126155, orientation='plus', exon_count=14, OMIM=602537, no_hdr0=''),
  727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727, CurrentID=0, Status='live', Symbol='C5', Aliases='C5Da, C5b, CPAMD4, ECLZB, C5', description='complement component 5', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 4|C5a anaphylatoxin|anaphylatoxin C5a analog|prepro-C5', map_location='9q33-q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120952335, end_position_on_the_genomic_accession=121074925, orientation='minus', exon_count=45, OMIM=120900, no_hdr0=''),
  728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728, CurrentID=0, Status='live', Symbol='C5AR1', Aliases='C5A, C5AR, C5R1, CD88', description='complement component 5a receptor 1', other_designations='C5a anaphylatoxin receptor|C5a ligand|C5a-R|complement component 5 receptor 1', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47307745, end_position_on_the_genomic_accession=47322070, orientation='plus', exon_count=4, OMIM=113995, no_hdr0=''),
  729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729, CurrentID=0, Status='live', Symbol='C6', Aliases='', description='complement component 6', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41142146, end_position_on_the_genomic_accession=41261486, orientation='minus', exon_count=23, OMIM=217050, no_hdr0=''),
  730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730, CurrentID=0, Status='live', Symbol='C7', Aliases='', description='complement component 7', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40909473, end_position_on_the_genomic_accession=41016004, orientation='plus', exon_count=19, OMIM=217070, no_hdr0=''),
  731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731, CurrentID=0, Status='live', Symbol='C8A', Aliases='', description='complement component 8, alpha polypeptide', other_designations='complement component 8 subunit alpha', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56854770, end_position_on_the_genomic_accession=56918221, orientation='plus', exon_count=12, OMIM=120950, no_hdr0=''),
  732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732, CurrentID=0, Status='live', Symbol='C8B', Aliases='C82', description='complement component 8, beta polypeptide', other_designations='complement component 8 subunit beta', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56929210, end_position_on_the_genomic_accession=56966140, orientation='minus', exon_count=14, OMIM=120960, no_hdr0=''),
  733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=733, CurrentID=0, Status='live', Symbol='C8G', Aliases='C8C', description='complement component 8, gamma polypeptide', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136944875, end_position_on_the_genomic_accession=136946974, orientation='plus', exon_count=7, OMIM=120930, no_hdr0=''),
  734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=734, CurrentID=0, Status='live', Symbol='OSGIN2', Aliases='C8orf1, hT41', description='oxidative stress induced growth inhibitor family member 2', other_designations='', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89901868, end_position_on_the_genomic_accession=89927868, orientation='plus', exon_count=9, OMIM=604598, no_hdr0=''),
  735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=735, CurrentID=0, Status='live', Symbol='C9', Aliases='ARMD15D, C9', description='complement component 9', other_designations='', map_location='5p14-p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39284275, end_position_on_the_genomic_accession=39364553, orientation='minus', exon_count=11, OMIM=120940, no_hdr0=''),
  738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=738, CurrentID=0, Status='live', Symbol='VPS51', Aliases='ANG2, ANG3, C11orf2, C11orf3, FFR', description='VPS51 GARP complex subunit', other_designations='another new gene 2 protein|protein fat-free homolog|vacuolar protein sorting 51 homolog', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65096115, end_position_on_the_genomic_accession=65111860, orientation='plus', exon_count=10, OMIM=615738, no_hdr0=''),
  740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=740, CurrentID=0, Status='live', Symbol='MRPL49', Aliases='C11orf4, L49mt, MRP-L49, NOF, NOF1', description='mitochondrial ribosomal protein L49', other_designations='neighbor of FAU|next to FAU', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65122183, end_position_on_the_genomic_accession=65127369, orientation='plus', exon_count=4, OMIM=606866, no_hdr0=''),
  741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=741, CurrentID=0, Status='live', Symbol='ZNHIT2', Aliases='C11orf5, FON', description='zinc finger, HIT-type containing 2', other_designations='zinc finger, HIT domain containing 2|zinc finger, HIT type 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65116403, end_position_on_the_genomic_accession=65117738, orientation='minus', exon_count=1, OMIM=604575, no_hdr0=''),
  100270823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270823, CurrentID=0, Status='live', Symbol='RPL7P11', Aliases='RPL7_1_6', description='ribosomal protein L7 pseudogene 11', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8750410, end_position_on_the_genomic_accession=8751091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=744, CurrentID=0, Status='live', Symbol='MPPED2', Aliases='239FB, C11orf8', description='metallophosphoesterase domain containing 2', other_designations='fetal brain protein 239|metallophosphoesterase domain-containing protein 2', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30384493, end_position_on_the_genomic_accession=30586714, orientation='minus', exon_count=14, OMIM=600911, no_hdr0=''),
  745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=745, CurrentID=0, Status='live', Symbol='MYRF', Aliases='C11orf9, MRF, Ndt80, pqn-47', description='myelin regulatory factor', other_designations='myelin gene regulatory factor', map_location='11q12-q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61752637, end_position_on_the_genomic_accession=61788518, orientation='plus', exon_count=28, OMIM=608329, no_hdr0=''),
  746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=746, CurrentID=0, Status='live', Symbol='TMEM258', Aliases='C11orf10', description='transmembrane protein 258', other_designations='UPF0197 transmembrane protein C11orf10', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61789130, end_position_on_the_genomic_accession=61792613, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=747, CurrentID=0, Status='live', Symbol='DAGLA', Aliases='C11orf11, DAGL(ALPHA)LPHA, NSDDR, DAGLA', description='diacylglycerol lipase, alpha', other_designations='DGL-alpha|neural stem cell-derived dendrite regulator', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61680433, end_position_on_the_genomic_accession=61747002, orientation='plus', exon_count=23, OMIM=614015, no_hdr0=''),
  100270828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270828, CurrentID=0, Status='live', Symbol='RPL37P9', Aliases='RPL37_1_4', description='ribosomal protein L37 pseudogene 9', other_designations='', map_location='1p36.31-p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6834328, end_position_on_the_genomic_accession=6834600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270829, CurrentID=0, Status='live', Symbol='RPL39P12', Aliases='RPL39_1_3', description='ribosomal protein L39 pseudogene 12', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1008074, end_position_on_the_genomic_accession=1008229, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=750, CurrentID=0, Status='live', Symbol='GAS8-AS1', Aliases='C16orf3', description='GAS8 antisense RNA 1', other_designations='uncharacterized protein C16orf3', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90028908, end_position_on_the_genomic_accession=90029901, orientation='minus', exon_count=1, OMIM=605179, no_hdr0=''),
  752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=752, CurrentID=0, Status='live', Symbol='FMNL1', Aliases='C17orf1, C17orf1B, FHOD4, FMNL, KW-13', description='formin like 1', other_designations='CLL-associated antigen KW-13|formin-like 1|leukocyte formin', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45221906, end_position_on_the_genomic_accession=45247320, orientation='plus', exon_count=29, OMIM=604656, no_hdr0=''),
  753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=753, CurrentID=0, Status='live', Symbol='LDLRAD4', Aliases='C18orf1', description='low density lipoprotein receptor class A domain containing 4', other_designations='clone 22', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13218730, end_position_on_the_genomic_accession=13652754, orientation='plus', exon_count=14, OMIM=606571, no_hdr0=''),
  754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=754, CurrentID=0, Status='live', Symbol='PTTG1IP', Aliases='C21orf1, C21orf3, PBF', description='pituitary tumor-transforming 1 interacting protein', other_designations='PTTG-binding factor|pituitary tumor-transforming gene protein-binding factor', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44849585, end_position_on_the_genomic_accession=44873903, orientation='minus', exon_count=6, OMIM=603784, no_hdr0=''),
  755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=755, CurrentID=0, Status='live', Symbol='C21orf2', Aliases='LRRC76, YF5/A2', description='chromosome 21 open reading frame 2', other_designations='leucine rich repeat containing 76', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44328944, end_position_on_the_genomic_accession=44339402, orientation='minus', exon_count=9, OMIM=603191, no_hdr0=''),
  100270836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270836, CurrentID=0, Status='live', Symbol='RPL5P15', Aliases='RPL5_6_566', description='ribosomal protein L5 pseudogene 15', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58822427, end_position_on_the_genomic_accession=58823248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=757, CurrentID=0, Status='live', Symbol='TMEM50B', Aliases='C21orf4, HCVP7TP3', description='transmembrane protein 50B', other_designations='HCV p7-trans-regulated protein 3|HCV p7-transregulated protein 3', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33432486, end_position_on_the_genomic_accession=33480009, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=758, CurrentID=0, Status='live', Symbol='MPPED1', Aliases='239AB, C22orf1, FAM1A', description='metallophosphoesterase domain containing 1', other_designations='adult brain protein 239', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43412014, end_position_on_the_genomic_accession=43507808, orientation='plus', exon_count=8, OMIM=602112, no_hdr0=''),
  759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=759, CurrentID=0, Status='live', Symbol='CA1', Aliases='CA-I, CAB, Car1, HEL-S-11', description='carbonic anhydrase I', other_designations='carbonate dehydratase I|carbonic anhydrase B|epididymis secretory protein Li 11', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85328229, end_position_on_the_genomic_accession=85378154, orientation='minus', exon_count=11, OMIM=114800, no_hdr0=''),
  760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=760, CurrentID=0, Status='live', Symbol='CA2', Aliases='CA-II, CAC, CAII, Car2, HEL-76, HEL-S-282', description='carbonic anhydrase II', other_designations='carbonate dehydratase II|carbonic anhydrase B|carbonic anhydrase C|carbonic dehydratase|epididymis luminal protein 76|epididymis secretory protein Li 282', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85463902, end_position_on_the_genomic_accession=85481492, orientation='plus', exon_count=7, OMIM=611492, no_hdr0=''),
  761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=761, CurrentID=0, Status='live', Symbol='CA3', Aliases='CAIII, Car3', description='carbonic anhydrase III', other_designations='CA-III|HEL-S-167mP|carbonate dehydratase III|carbonic anhydrase III, muscle specific|epididymis secretory sperm binding protein Li 167mP', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85438827, end_position_on_the_genomic_accession=85449040, orientation='plus', exon_count=7, OMIM=114750, no_hdr0=''),
  762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=762, CurrentID=0, Status='live', Symbol='CA4', Aliases='CAIV, Car4, RP17', description='carbonic anhydrase IV', other_designations='CA-IV|carbonate dehydratase IV|carbonic dehydratase IV', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60149941, end_position_on_the_genomic_accession=60159546, orientation='plus', exon_count=9, OMIM=114760, no_hdr0=''),
  763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=763, CurrentID=0, Status='live', Symbol='CA5A', Aliases='CA5D, CAV, CAVA, GS1-21A4.1, CA5A', description='carbonic anhydrase VA, mitochondrial', other_designations='CA-VA|carbonate dehydratase VA|carbonic anhydrase V, mitochondrial|carbonic dehydratase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87888019, end_position_on_the_genomic_accession=87936569, orientation='minus', exon_count=10, OMIM=114761, no_hdr0=''),
  764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=764, CurrentID=0, Status='live', Symbol='CA5AP1', Aliases='CA5P', description='carbonic anhydrase VA pseudogene 1', other_designations='carbonic anhydrase V pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29618675, end_position_on_the_genomic_accession=29636331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=765, CurrentID=0, Status='live', Symbol='CA6', Aliases='CA-VI, GUSTIN', description='carbonic anhydrase VI', other_designations='carbonate dehydratase VI|carbonic anhydrase VI nirs variant 2|salivary carbonic anhydrase|secreted carbonic anhydrase', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8945785, end_position_on_the_genomic_accession=8975092, orientation='plus', exon_count=10, OMIM=114780, no_hdr0=''),
  766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=766, CurrentID=0, Status='live', Symbol='CA7', Aliases='CAVII', description='carbonic anhydrase VII', other_designations='CA-VII|carbonate dehydratase VII|carbonic dehydratase VII', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66844379, end_position_on_the_genomic_accession=66854149, orientation='plus', exon_count=9, OMIM=114770, no_hdr0=''),
  767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767, CurrentID=0, Status='live', Symbol='CA8', Aliases='CA-VIII, CALS, CAMRQ3, CARP', description='carbonic anhydrase VIII', other_designations='CA-related protein|carbonate dehydratase|carbonic anhydrase-like sequence', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60187347, end_position_on_the_genomic_accession=60281395, orientation='minus', exon_count=13, OMIM=114815, no_hdr0=''),
  768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768, CurrentID=0, Status='live', Symbol='CA9', Aliases='CAIX, MN', description='carbonic anhydrase IX', other_designations='CA-IX|P54/58N|RCC-associated antigen G250|RCC-associated protein G250|carbonate dehydratase IX|carbonic dehydratase|membrane antigen MN|pMW1|renal cell carcinoma-associated antigen G250', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35673918, end_position_on_the_genomic_accession=35681159, orientation='plus', exon_count=12, OMIM=603179, no_hdr0=''),
  770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=770, CurrentID=0, Status='live', Symbol='CA11', Aliases='CARPX1', description='carbonic anhydrase XI', other_designations='CA-RP II|CA-XI|CARP XI|CARP-2|carbonic anhydrase-related protein 2|carbonic anhydrase-related protein XI', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48638015, end_position_on_the_genomic_accession=48646194, orientation='minus', exon_count=9, OMIM=604644, no_hdr0=''),
  771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=771, CurrentID=0, Status='live', Symbol='CA12', Aliases='CAXII, HsT18816', description='carbonic anhydrase XII', other_designations='CA-XII|carbonate dehydratase XII|carbonic dehydratase|tumor antigen HOM-RCC-3.1.3', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63323531, end_position_on_the_genomic_accession=63382110, orientation='minus', exon_count=11, OMIM=603263, no_hdr0=''),
  772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=772, CurrentID=0, Status='live', Symbol='CACD', Aliases='', description='central areolar choroidal dystrophy', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=215500, no_hdr0=''),
  773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=773, CurrentID=0, Status='live', Symbol='CACNA1A', Aliases='APCA, BI, CACNL1A4, CAV2.1, EA2, FHM, HPCA, MHP, MHP1, SCA6', description='calcium channel, voltage-dependent, P/Q type, alpha 1A subunit', other_designations='brain calcium channel 1|brain calcium channel I|calcium channel, L type, alpha-1 polypeptide|fetal brain Ca2+ voltage-gated channel alpha1A pore-forming subunit|voltage-gated calcium channel subunit alpha Cav2.1', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13206442, end_position_on_the_genomic_accession=13506460, orientation='minus', exon_count=49, OMIM=601011, no_hdr0=''),
  774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=774, CurrentID=0, Status='live', Symbol='CACNA1B', Aliases='BIII, CACNL1A5, CACNN, Cav2.2, DYT23', description='calcium channel, voltage-dependent, N type, alpha 1B subunit', other_designations='Cav2.2 voltage-gated Ca2+ channel|brain calcium channel III|calcium channel alpha12.2 subunit|calcium channel, L type, alpha-1 polypeptide|calcium channel, voltage-dependent, L type, alpha 1B subunit|voltage-gated calcium channel alpha subunit Cav2.2|voltage-gated calcium channel subunit alpha Cav2.2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137877789, end_position_on_the_genomic_accession=138124624, orientation='plus', exon_count=51, OMIM=601012, no_hdr0=''),
  775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=775, CurrentID=0, Status='live', Symbol='CACNA1C', Aliases='CACH2, CACN2, CACNL1A1, CCHL1A1, CaV1.2, LQT8, TS', description='calcium channel, voltage-dependent, L type, alpha 1C subunit', other_designations='DHPR, alpha-1 subunit|calcium channel, L type, alpha-1 polypeptide, isoform 1, cardiac muscle|calcium channel, cardic dihydropyridine-sensitive, alpha-1 subunit|voltage-dependent L-type Ca2+ channel alpha 1 subunit|voltage-gated L-type calcium channel Cav1.2 alpha 1 subunit, splice variant 10*', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1970786, end_position_on_the_genomic_accession=2697949, orientation='plus', exon_count=56, OMIM=114205, no_hdr0=''),
  776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=776, CurrentID=0, Status='live', Symbol='CACNA1D', Aliases='CACH3, CACN4, CACNL1A2, CCHL1A2, Cav1.3, PASNA, SANDD', description='calcium channel, voltage-dependent, L type, alpha 1D subunit', other_designations='calcium channel, L type, alpha-1 polypeptide|calcium channel, neuroendocrine/brain-type, alpha 1 subunit|voltage-gated calcium channel alpha 1 subunit|voltage-gated calcium channel alpha subunit Cav1.3', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53494976, end_position_on_the_genomic_accession=53813151, orientation='plus', exon_count=55, OMIM=114206, no_hdr0=''),
  777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777, CurrentID=0, Status='live', Symbol='CACNA1E', Aliases='BII, CACH6, CACNL1A6, Cav2.3', description='calcium channel, voltage-dependent, R type, alpha 1E subunit', other_designations='brain calcium channel II|calcium channel, L type, alpha-1 polypeptide|calcium channel, R type, alpha-1 polypeptide|calcium channel, voltage-dependent, alpha 1E subunit|voltage-dependent calcium channel alpha 1E subunit|voltage-gated calcium channel alpha 1E subunit|voltage-gated calcium channel alpha subunit Cav2.3|voltage-gated calcium channel subunit alpha Cav2.3', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181483311, end_position_on_the_genomic_accession=181806784, orientation='plus', exon_count=48, OMIM=601013, no_hdr0=''),
  778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=778, CurrentID=0, Status='live', Symbol='CACNA1F', Aliases='AIED, COD3, COD4, CORDX, CORDX3, CSNB2, CSNB2A, CSNBX2, Cav1.4, Cav1.4alpha1, JM8, JMC8, OA2', description='calcium channel, voltage-dependent, L type, alpha 1F subunit', other_designations='voltage-gated calcium channel subunit alpha Cav1.4', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49205063, end_position_on_the_genomic_accession=49233671, orientation='minus', exon_count=48, OMIM=300110, no_hdr0=''),
  779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=779, CurrentID=0, Status='live', Symbol='CACNA1S', Aliases='CACNL1A3, CCHL1A3, Cav1.1, HOKPP, HOKPP1, MHS5, TTPP1, hypoPP', description='calcium channel, voltage-dependent, L type, alpha 1S subunit', other_designations='calcium channel, L type, alpha 1 polypeptide, isoform 3 (skeletal muscle, hypokalemic periodic paralysis)|dihydropyridine receptor|dihydropyridine-sensitive L-type calcium channel alpha-1 subunit|voltage-gated calcium channel subunit alpha Cav1.1', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201039507, end_position_on_the_genomic_accession=201112566, orientation='minus', exon_count=44, OMIM=114208, no_hdr0=''),
  780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780, CurrentID=0, Status='live', Symbol='DDR1', Aliases='CAK, CD167, DDR, EDDR1, HGK2, MCK10, NEP, NTRK4, PTK3, PTK3A, RTK6, TRKE', description='discoidin domain receptor tyrosine kinase 1', other_designations='CD167 antigen-like family member A|PTK3A protein tyrosine kinase 3A|cell adhesion kinase|mammary carcinoma kinase 10|neuroepithelial tyrosine kinase|neurotrophic tyrosine kinase, receptor, type 4|protein-tyrosine kinase RTK-6|tyrosine kinase DDR|tyrosine-protein kinase CAK', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30880909, end_position_on_the_genomic_accession=30900156, orientation='plus', exon_count=23, OMIM=600408, no_hdr0=''),
  781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=781, CurrentID=0, Status='live', Symbol='CACNA2D1', Aliases='CACNA2, CACNL2A, CCHL2A, LINC01112, lncRNA-N3', description='calcium channel, voltage-dependent, alpha 2/delta subunit 1', other_designations='calcium channel, L type, alpha 2 polypeptide|dihydropyridine-sensitive L-type, calcium channel alpha-2/delta subunit|voltage-gated calcium channel subunit alpha-2/delta-1', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81946444, end_position_on_the_genomic_accession=82443806, orientation='minus', exon_count=47, OMIM=114204, no_hdr0=''),
  782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=782, CurrentID=0, Status='live', Symbol='CACNB1', Aliases='CAB1, CACNLB1, CCHLB1', description='calcium channel, voltage-dependent, beta 1 subunit', other_designations='calcium channel voltage-dependent subunit beta 1|calcium channel, L type, beta 1 polypeptide|dihydropyridine-sensitive L-type, calcium channel beta-1 subunit', map_location='17q21-q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39173456, end_position_on_the_genomic_accession=39197703, orientation='minus', exon_count=18, OMIM=114207, no_hdr0=''),
  783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=783, CurrentID=0, Status='live', Symbol='CACNB2', Aliases='CACNLB2, CAVB2, MYSB', description='calcium channel, voltage-dependent, beta 2 subunit', other_designations='CAB2|calcium channel voltage-dependent subunit beta 2|lambert-Eaton myasthenic syndrome antigen B|myasthenic (Lambert-Eaton) syndrome antigen B', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18140677, end_position_on_the_genomic_accession=18541759, orientation='plus', exon_count=20, OMIM=600003, no_hdr0=''),
  784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=784, CurrentID=0, Status='live', Symbol='CACNB3', Aliases='CAB3, CACNLB3', description='calcium channel, voltage-dependent, beta 3 subunit', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48814432, end_position_on_the_genomic_accession=48828943, orientation='plus', exon_count=19, OMIM=601958, no_hdr0=''),
  785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=785, CurrentID=0, Status='live', Symbol='CACNB4', Aliases='CAB4, CACNLB4, EA5, EIG9, EJM, EJM4, EJM6', description='calcium channel, voltage-dependent, beta 4 subunit', other_designations='calcium channel voltage-dependent subunit beta 4|dihydropyridine-sensitive L-type, calcium channel beta-4 subunit', map_location='2q22-q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151832771, end_position_on_the_genomic_accession=152099988, orientation='minus', exon_count=21, OMIM=601949, no_hdr0=''),
  786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=786, CurrentID=0, Status='live', Symbol='CACNG1', Aliases='CACNLG', description='calcium channel, voltage-dependent, gamma subunit 1', other_designations='L-type calcium channel gamma polypeptide|dihydropyridine-sensitive L-type, skeletal muscle calcium channel subunit gamma|neuronal dihydropyridine-sensitive calcium channel gamma subunit', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67044536, end_position_on_the_genomic_accession=67056797, orientation='plus', exon_count=4, OMIM=114209, no_hdr0=''),
  100270867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270867, CurrentID=0, Status='live', Symbol='RPS6P12', Aliases='RPS6_5_991', description='ribosomal protein S6 pseudogene 12', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82744055, end_position_on_the_genomic_accession=82744792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=788, CurrentID=0, Status='live', Symbol='SLC25A20', Aliases='CAC, CACT', description='solute carrier family 25 (carnitine/acylcarnitine translocase), member 20', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48856923, end_position_on_the_genomic_accession=48898993, orientation='minus', exon_count=9, OMIM=613698, no_hdr0=''),
  100270869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270869, CurrentID=0, Status='live', Symbol='RPS7P6', Aliases='RPS7_3_462', description='ribosomal protein S7 pseudogene 6', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17427665, end_position_on_the_genomic_accession=17428349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790, CurrentID=0, Status='live', Symbol='CAD', Aliases='', description='carbamoyl-phosphate synthetase 2, aspartate transcarbamylase, and dihydroorotase', other_designations='CAD trifunctional protein|multifunctional protein CAD', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27217390, end_position_on_the_genomic_accession=27243792, orientation='plus', exon_count=45, OMIM=114010, no_hdr0=''),
  100270871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270871, CurrentID=0, Status='live', Symbol='RPS8P6', Aliases='RPS8_1_339', description='ribosomal protein S8 pseudogene 6', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=308717, end_position_on_the_genomic_accession=309319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270872, CurrentID=0, Status='live', Symbol='RPSAP16', Aliases='RPSA_1_135', description='ribosomal protein SA pseudogene 16', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179968595, end_position_on_the_genomic_accession=179969581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=793, CurrentID=0, Status='live', Symbol='CALB1', Aliases='CALB, D-28K', description='calbindin 1', other_designations='RTVL-H protein|calbindin 1, (28kD)|calbindin 1, 28kDa|calbindin D28|vitamin D-dependent calcium-binding protein, avian-type', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90058608, end_position_on_the_genomic_accession=90082881, orientation='minus', exon_count=11, OMIM=114050, no_hdr0=''),
  794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=794, CurrentID=0, Status='live', Symbol='CALB2', Aliases='CAB29, CAL2, CR', description='calbindin 2', other_designations='29 kDa calbindin|calbindin 2, (29kD, calretinin)|calbindin D29K', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71358713, end_position_on_the_genomic_accession=71390438, orientation='plus', exon_count=11, OMIM=114051, no_hdr0=''),
  795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=795, CurrentID=0, Status='live', Symbol='S100G', Aliases='CABP, CABP1, CABP9K, CALB3', description='S100 calcium binding protein G', other_designations='S100 calcium-binding protein G|calbindin 3, (vitamin D-dependent calcium-binding protein)|calbindin D9K|calbindin-D9K|vitamin D-dependent calcium-binding protein, intestinal', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16650158, end_position_on_the_genomic_accession=16654668, orientation='plus', exon_count=3, OMIM=302020, no_hdr0=''),
  796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=796, CurrentID=0, Status='live', Symbol='CALCA', Aliases='CALC1, CGRP, CGRP-I, CGRP1, CT, KC', description='calcitonin-related polypeptide alpha', other_designations='alpha-type CGRP|calcitonin 1|calcitonin gene-related peptide I|calcitonin/calcitonin-related polypeptide, alpha|katacalcin', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14966669, end_position_on_the_genomic_accession=14972286, orientation='minus', exon_count=6, OMIM=114130, no_hdr0=''),
  797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=797, CurrentID=0, Status='live', Symbol='CALCB', Aliases='CALC2, CGRP-II, CGRP2', description='calcitonin-related polypeptide beta', other_designations='beta-CGRP|beta-type CGRP|calcitonin 2|calcitonin gene-related peptide II', map_location='11p15.2-p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15073600, end_position_on_the_genomic_accession=15078631, orientation='plus', exon_count=5, OMIM=114160, no_hdr0=''),
  131870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131870, CurrentID=0, Status='live', Symbol='NUDT16', Aliases='', description='nudix hydrolase 16', other_designations='IDP phosphatase|IDPase|U8 snoRNA-binding protein H29K|inosine diphosphate phosphatase|m7GpppN-mRNA hydrolase|nucleoside diphosphate-linked moiety X motif 16|nudix (nucleoside diphosphate linked moiety X)-type motif 16|nudix motif 16', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131381671, end_position_on_the_genomic_accession=131388830, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=799, CurrentID=0, Status='live', Symbol='CALCR', Aliases='CRT, CT-R, CTR, CTR1', description='calcitonin receptor', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93424487, end_position_on_the_genomic_accession=93574730, orientation='minus', exon_count=17, OMIM=114131, no_hdr0=''),
  800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=800, CurrentID=0, Status='live', Symbol='CALD1', Aliases='CDM, H-CAD, HCAD, L-CAD, LCAD, NAG22', description='caldesmon 1', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134779413, end_position_on_the_genomic_accession=134970729, orientation='plus', exon_count=18, OMIM=114213, no_hdr0=''),
  801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=801, CurrentID=0, Status='live', Symbol='CALM1', Aliases='CALML2, CAMI, CPVT4, DD132, LQT14, PHKD, caM', description='calmodulin 1 (phosphorylase kinase, delta)', other_designations='phosphorylase kinase, delta subunit|prepro-calmodulin 1', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90396983, end_position_on_the_genomic_accession=90408275, orientation='plus', exon_count=7, OMIM=114180, no_hdr0=''),
  802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=802, CurrentID=0, Status='live', Symbol='CALM1P1', Aliases='CALML1', description='calmodulin 1 (phosphorylase kinase, delta) pseudogene 1', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95499808, end_position_on_the_genomic_accession=95499915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270883, CurrentID=0, Status='live', Symbol='RPL26P7', Aliases='RPL26_1_10', description='ribosomal protein L26 pseudogene 7', other_designations='', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9662288, end_position_on_the_genomic_accession=9662628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=804, CurrentID=0, Status='live', Symbol='CALM1P2', Aliases='', description='calmodulin 1 (phosphorylase kinase, delta) pseudogene 2', other_designations='TCAG_1643085', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55259718, end_position_on_the_genomic_accession=55260101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=805, CurrentID=0, Status='live', Symbol='CALM2', Aliases='CAMII, LQT15, PHKD, PHKD2, caM', description='calmodulin 2 (phosphorylase kinase, delta)', other_designations='LP7057 protein|phosphorylase kinase delta|prepro-calmodulin 2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47160082, end_position_on_the_genomic_accession=47176601, orientation='minus', exon_count=6, OMIM=114182, no_hdr0=''),
  806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=806, CurrentID=0, Status='live', Symbol='CALM2P2', Aliases='', description='calmodulin 2 pseudogene 2', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70163609, end_position_on_the_genomic_accession=70164741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=807, CurrentID=0, Status='live', Symbol='CALM2P3', Aliases='', description='calmodulin 2 pseudogene 3', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41170588, end_position_on_the_genomic_accession=41171206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=808, CurrentID=0, Status='live', Symbol='CALM3', Aliases='HEL-S-72, PHKD, PHKD3', description='calmodulin 3 (phosphorylase kinase, delta)', other_designations='caM|epididymis secretory protein Li 72|prepro-calmodulin 3', map_location='19q13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46601255, end_position_on_the_genomic_accession=46610782, orientation='plus', exon_count=6, OMIM=114183, no_hdr0=''),
  104355292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355292, CurrentID=0, Status='live', Symbol='LINC01525', Aliases='lnc-MAN1A2-1', description='long intergenic non-protein coding RNA 1525', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117295466, end_position_on_the_genomic_accession=117321336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=810, CurrentID=0, Status='live', Symbol='CALML3', Aliases='CLP', description='calmodulin like 3', other_designations='caM-like protein|calmodulin-like 3|calmodulin-related protein NB-1', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5524889, end_position_on_the_genomic_accession=5526268, orientation='plus', exon_count=1, OMIM=114184, no_hdr0=''),
  811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=811, CurrentID=0, Status='live', Symbol='CALR', Aliases='CRT, HEL-S-99n, RO, SSA, cC1qR', description='calreticulin', other_designations='CRP55|ERp60|HACBP|Sicca syndrome antigen A (autoantigen Ro; calreticulin)|calregulin|endoplasmic reticulum resident protein 60|epididymis secretory sperm binding protein Li 99n|grp60', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12938600, end_position_on_the_genomic_accession=12944490, orientation='plus', exon_count=9, OMIM=109091, no_hdr0=''),
  100270892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270892, CurrentID=0, Status='live', Symbol='RPL39P7', Aliases='RPL39_2_23', description='ribosomal protein L39 pseudogene 7', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25888916, end_position_on_the_genomic_accession=25889301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=813, CurrentID=0, Status='live', Symbol='CALU', Aliases='', description='calumenin', other_designations='IEF SSP 9302|crocalbin-like protein|multiple EF-hand protein', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128739292, end_position_on_the_genomic_accession=128773423, orientation='plus', exon_count=9, OMIM=603420, no_hdr0=''),
  814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=814, CurrentID=0, Status='live', Symbol='CAMK4', Aliases='CaMK IV, CaMK-GR, IV, caMK', description='calcium/calmodulin-dependent protein kinase IV', other_designations='CAM kinase IV|CAM kinase- GR|brain Ca(2+)-calmodulin-dependent protein kinase type IV|brain Ca++-calmodulin-dependent protein kinase type IV|caM kinase-GR|calcium/calmodulin-dependent protein kinase type IV catalytic chain', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111224249, end_position_on_the_genomic_accession=111494886, orientation='plus', exon_count=12, OMIM=114080, no_hdr0=''),
  815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=815, CurrentID=0, Status='live', Symbol='CAMK2A', Aliases='CAMKA', description='calcium/calmodulin-dependent protein kinase II alpha', other_designations='CaM kinase II alpha subunit|CaM-kinase II alpha chain|CaMK-II alpha subunit|CaMKIINalpha|caM kinase II subunit alpha|caMK-II subunit alpha|calcium/calmodulin-dependent protein kinase (CaM kinase) II alpha|calcium/calmodulin-dependent protein kinase II alpha-B subunit|calcium/calmodulin-dependent protein kinase type II alpha chain', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150219491, end_position_on_the_genomic_accession=150289840, orientation='minus', exon_count=19, OMIM=114078, no_hdr0=''),
  816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=816, CurrentID=0, Status='live', Symbol='CAMK2B', Aliases='CAM2, CAMK2, CAMKB', description='calcium/calmodulin-dependent protein kinase II beta', other_designations='CaM kinase II beta subunit|CaM-kinase II beta chain|caMK-II subunit beta|proline rich calmodulin-dependent protein kinase', map_location='7p14.3-p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44217150, end_position_on_the_genomic_accession=44325994, orientation='minus', exon_count=29, OMIM=607707, no_hdr0=''),
  817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=817, CurrentID=0, Status='live', Symbol='CAMK2D', Aliases='CAMKD', description='calcium/calmodulin-dependent protein kinase II delta', other_designations='CaM kinase II delta subunit|CaM-kinase II delta chain|CaMK-II delta subunit|caM kinase II subunit delta|caMK-II subunit delta|calcium/calmodulin-dependent protein kinase (CaM kinase) II delta|calcium/calmodulin-dependent protein kinase type II delta chain', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113451032, end_position_on_the_genomic_accession=113762133, orientation='minus', exon_count=27, OMIM=607708, no_hdr0=''),
  818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=818, CurrentID=0, Status='live', Symbol='CAMK2G', Aliases='CAMK, CAMK-II, CAMKG', description='calcium/calmodulin-dependent protein kinase II gamma', other_designations='caMK-II subunit gamma|calcium/calmodulin-dependent protein kinase (CaM kinase) II gamma', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73812501, end_position_on_the_genomic_accession=73874591, orientation='minus', exon_count=27, OMIM=602123, no_hdr0=''),
  819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=819, CurrentID=0, Status='live', Symbol='CAMLG', Aliases='CAML', description='calcium modulating ligand', other_designations='calcium-modulating cyclophilin ligand|calcium-signal modulating cyclophilin ligand|cyclophilin B-binding protein', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134737927, end_position_on_the_genomic_accession=134752160, orientation='plus', exon_count=4, OMIM=601118, no_hdr0=''),
  820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=820, CurrentID=0, Status='live', Symbol='CAMP', Aliases='CAP-18, CAP18, CRAMP, FALL-39, FALL39, HSD26, LL37', description='cathelicidin antimicrobial peptide', other_designations='18 kDa cationic antimicrobial protein', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48223347, end_position_on_the_genomic_accession=48225491, orientation='plus', exon_count=4, OMIM=600474, no_hdr0=''),
  821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=821, CurrentID=0, Status='live', Symbol='CANX', Aliases='CNX, IP90, P90', description='calnexin', other_designations='major histocompatibility complex class I antigen-binding protein p88', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179678646, end_position_on_the_genomic_accession=179731641, orientation='plus', exon_count=17, OMIM=114217, no_hdr0=''),
  822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=822, CurrentID=0, Status='live', Symbol='CAPG', Aliases='AFCP, HEL-S-66, MCP', description='capping protein (actin filament), gelsolin-like', other_designations='actin regulatory protein CAP-G|actin-regulatory protein CAP-G|epididymis secretory protein Li 66|gelsolin-like capping protein|macrophage capping protein', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85394748, end_position_on_the_genomic_accession=85418467, orientation='minus', exon_count=14, OMIM=153615, no_hdr0=''),
  823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=823, CurrentID=0, Status='live', Symbol='CAPN1', Aliases='CANP, CANP1, CANPL1, muCANP, muCL', description='calpain 1, (mu/I) large subunit', other_designations='CANP 1|calcium-activated neutral proteinase 1|calpain mu-type|calpain, large polypeptide L1|calpain-1 large subunit|cell proliferation-inducing gene 30 protein|cell proliferation-inducing protein 30|micromolar-calpain', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65181215, end_position_on_the_genomic_accession=65212006, orientation='plus', exon_count=24, OMIM=114220, no_hdr0=''),
  824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=824, CurrentID=0, Status='live', Symbol='CAPN2', Aliases='CANP2, CANPL2, CANPml, mCANP', description='calpain 2, (m/II) large subunit', other_designations='CANP 2|M-calpain|calcium-activated neutral proteinase 2|calpain 2, large [catalytic] subunit|calpain 2, large subunit|calpain M-type|calpain large polypeptide L2|calpain, large polypeptide L2|calpain-2 large subunit|millimolar-calpain', map_location='1q41-q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223701593, end_position_on_the_genomic_accession=223776018, orientation='plus', exon_count=23, OMIM=114230, no_hdr0=''),
  825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=825, CurrentID=0, Status='live', Symbol='CAPN3', Aliases='CANP3, CANPL3, LGMD2, LGMD2A, nCL-1, p94', description='calpain 3', other_designations='calpain 3, (p94)|calpain p94, large [catalytic] subunit|calpain, large polypeptide L3|muscle-specific calcium-activated neutral protease 3 large subunit|new calpain 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42359500, end_position_on_the_genomic_accession=42412317, orientation='plus', exon_count=26, OMIM=114240, no_hdr0=''),
  826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=826, CurrentID=0, Status='live', Symbol='CAPNS1', Aliases='CALPAIN4, CANP, CANPS, CAPN4, CDPS, CSS1', description='calpain, small subunit 1', other_designations='CANP small subunit|calcium-activated neutral proteinase small subunit|calcium-dependent protease small subunit 1|calcium-dependent protease, small subunit|calpain 4, small subunit (30K)|calpain regulatory subunit|calpain, small polypeptide', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36139926, end_position_on_the_genomic_accession=36150353, orientation='plus', exon_count=13, OMIM=114170, no_hdr0=''),
  827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=827, CurrentID=0, Status='live', Symbol='CAPN6', Aliases='CANPX, CAPNX, CalpM, DJ914P14.1', description='calpain 6', other_designations='calpain-like protease X-linked|calpamodulin', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111245099, end_position_on_the_genomic_accession=111270546, orientation='minus', exon_count=13, OMIM=300146, no_hdr0=''),
  828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=828, CurrentID=0, Status='live', Symbol='CAPS', Aliases='CAPS1', description='calcyphosine', other_designations='calcyphosine 1|thyroid protein p24', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5913648, end_position_on_the_genomic_accession=5916211, orientation='plus', exon_count=5, OMIM=114212, no_hdr0=''),
  829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=829, CurrentID=0, Status='live', Symbol='CAPZA1', Aliases='CAPPA1, CAPZ, CAZ1', description='capping protein (actin filament) muscle Z-line, alpha 1', other_designations='Cap Z|capZ alpha-1', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112619453, end_position_on_the_genomic_accession=112671619, orientation='plus', exon_count=11, OMIM=601580, no_hdr0=''),
  830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=830, CurrentID=0, Status='live', Symbol='CAPZA2', Aliases='CAPPA2, CAPZ', description='capping protein (actin filament) muscle Z-line, alpha 2', other_designations='F-actin capping protein alpha-2 subunit|capZ alpha-2', map_location='7q31.2-q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116862509, end_position_on_the_genomic_accession=116919259, orientation='plus', exon_count=10, OMIM=601571, no_hdr0=''),
  831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=831, CurrentID=0, Status='live', Symbol='CAST', Aliases='BS-17, PLACK', description='calpastatin', other_designations='calpain inhibitor|sperm BS-17 component', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96662037, end_position_on_the_genomic_accession=96774683, orientation='plus', exon_count=38, OMIM=114090, no_hdr0=''),
  832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=832, CurrentID=0, Status='live', Symbol='CAPZB', Aliases='CAPB, CAPPB, CAPZ', description='capping protein (actin filament) muscle Z-line, beta', other_designations='capZ beta', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19338773, end_position_on_the_genomic_accession=19485641, orientation='minus', exon_count=14, OMIM=601572, no_hdr0=''),
  833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=833, CurrentID=0, Status='live', Symbol='CARS', Aliases='CARS1, CYSRS, MGC:11246', description='cysteinyl-tRNA synthetase', other_designations='cysteine tRNA ligase 1, cytoplasmic|cysteine translase', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3000922, end_position_on_the_genomic_accession=3057451, orientation='minus', exon_count=23, OMIM=123859, no_hdr0=''),
  834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=834, CurrentID=0, Status='live', Symbol='CASP1', Aliases='ICE, IL1BC, P45', description='caspase 1', other_designations='CASP1 nirs variant 1|IL-1 beta-converting enzyme|IL1B-convertase|caspase 1, apoptosis-related cysteine peptidase (interleukin 1, beta, convertase)|interleukin 1, beta, convertase|interleukin 1-B converting enzyme', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105025508, end_position_on_the_genomic_accession=105036662, orientation='minus', exon_count=11, OMIM=147678, no_hdr0=''),
  835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=835, CurrentID=0, Status='live', Symbol='CASP2', Aliases='CASP-2, ICH1, NEDD-2, NEDD2, PPP1R57', description='caspase 2', other_designations='caspase 2 apoptosis-related cysteine peptidase|neural precursor cell expressed developmentally down-regulated protein 2|protease ICH-1|protein phosphatase 1, regulatory subunit 57', map_location='7q34-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143288215, end_position_on_the_genomic_accession=143307696, orientation='plus', exon_count=13, OMIM=600639, no_hdr0=''),
  836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=836, CurrentID=0, Status='live', Symbol='CASP3', Aliases='CPP32, CPP32B, SCA-1', description='caspase 3', other_designations='CASP-3|CPP-32|PARP cleavage protease|SREBP cleavage activity 1|apopain|caspase 3, apoptosis-related cysteine peptidase|caspase 3, apoptosis-related cysteine protease|cysteine protease CPP32|procaspase3|protein Yama', map_location='4q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184627696, end_position_on_the_genomic_accession=184649475, orientation='minus', exon_count=8, OMIM=600636, no_hdr0=''),
  837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=837, CurrentID=0, Status='live', Symbol='CASP4', Aliases='ICE(rel)II, ICEREL-II, ICH-2, Mih1/TX, TX', description='caspase 4', other_designations='CASP-4|ICE(rel)-II|apoptotic cysteine protease Mih1/TX|caspase 4, apoptosis-related cysteine peptidase|caspase 4, apoptosis-related cysteine protease|protease ICH-2|protease TX', map_location='11q22.2-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104942866, end_position_on_the_genomic_accession=104968598, orientation='minus', exon_count=10, OMIM=602664, no_hdr0=''),
  838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=838, CurrentID=0, Status='live', Symbol='CASP5', Aliases='ICE(rel)III, ICEREL-III, ICH-3', description='caspase 5', other_designations='CASP-5|ICE(rel)-III|TY protease|caspase 5, apoptosis-related cysteine peptidase|caspase 5, apoptosis-related cysteine protease|protease ICH-3|protease TY', map_location='11q22.2-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104992633, end_position_on_the_genomic_accession=105023221, orientation='minus', exon_count=10, OMIM=602665, no_hdr0=''),
  839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=839, CurrentID=0, Status='live', Symbol='CASP6', Aliases='MCH2', description='caspase 6', other_designations='apoptotic protease MCH-2|caspase 6, apoptosis-related cysteine peptidase|caspase 6, apoptosis-related cysteine protease', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109688628, end_position_on_the_genomic_accession=109713904, orientation='minus', exon_count=8, OMIM=601532, no_hdr0=''),
  840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=840, CurrentID=0, Status='live', Symbol='CASP7', Aliases='CASP-7, CMH-1, ICE-LAP3, LICE2, MCH3', description='caspase 7', other_designations='ICE-like apoptotic protease 3|apoptotic protease MCH-3|caspase 7, apoptosis-related cysteine peptidase|caspase 7, apoptosis-related cysteine protease', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113679162, end_position_on_the_genomic_accession=113730909, orientation='plus', exon_count=12, OMIM=601761, no_hdr0=''),
  841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=841, CurrentID=0, Status='live', Symbol='CASP8', Aliases='ALPS2B, CAP4, Casp-8, FLICE, MACH, MCH5', description='caspase 8', other_designations='FADD-homologous ICE/CED-3-like protease|FADD-like ICE|ICE-like apoptotic protease 5|MACH-alpha-1/2/3 protein|MACH-beta-1/2/3/4 protein|MORT1-associated ced-3 homolog|apoptotic cysteine protease|apoptotic protease Mch-5|caspase 8, apoptosis-related cysteine peptidase|caspase 8, apoptosis-related cysteine protease', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201233443, end_position_on_the_genomic_accession=201287711, orientation='plus', exon_count=16, OMIM=601763, no_hdr0=''),
  842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=842, CurrentID=0, Status='live', Symbol='CASP9', Aliases='APAF-3, APAF3, ICE-LAP6, MCH6, PPP1R56', description='caspase 9', other_designations='ICE-like apoptotic protease 6|apoptotic protease MCH-6|apoptotic protease activating factor 3|caspase 9, apoptosis-related cysteine peptidase|protein phosphatase 1, regulatory subunit 56', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15491401, end_position_on_the_genomic_accession=15526529, orientation='minus', exon_count=14, OMIM=602234, no_hdr0=''),
  843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=843, CurrentID=0, Status='live', Symbol='CASP10', Aliases='ALPS2, FLICE2, MCH4', description='caspase 10', other_designations='CASP-10|FADD-like ICE2|FAS-associated death domain protein interleukin-1B-converting enzyme 2|ICE-like apoptotic protease 4|apoptotic protease MCH-4|caspase 10 apoptosis-related cysteine peptidase|caspase 10, apoptosis-related cysteine peptidase|caspase 10, apoptosis-related cysteine protease|interleukin-1B-converting enzyme 2', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201182898, end_position_on_the_genomic_accession=201229406, orientation='plus', exon_count=13, OMIM=601762, no_hdr0=''),
  844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=844, CurrentID=0, Status='live', Symbol='CASQ1', Aliases='CASQ, PDIB1, VMCQA', description='calsequestrin 1', other_designations='calmitin|calmitine|calsequestrin 1 (fast-twitch, skeletal muscle)|calsequestrin, skeletal muscle isoform', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160190495, end_position_on_the_genomic_accession=160201886, orientation='plus', exon_count=11, OMIM=114250, no_hdr0=''),
  845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=845, CurrentID=0, Status='live', Symbol='CASQ2', Aliases='PDIB2', description='calsequestrin 2', other_designations='calsequestrin 2 (cardiac muscle)|calsequestrin 2, fast-twitch, cardiac muscle|calsequestrin, cardiac muscle isoform', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115700003, end_position_on_the_genomic_accession=115768805, orientation='minus', exon_count=11, OMIM=114251, no_hdr0=''),
  846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=846, CurrentID=0, Status='live', Symbol='CASR', Aliases='CAR, EIG8, FHH, FIH, GPRC2A, HHC, HHC1, HYPOC1, NSHPT, PCAR1', description='calcium-sensing receptor', other_designations='parathyroid Ca(2+)-sensing receptor 1|parathyroid cell calcium-sensing receptor 1', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122183667, end_position_on_the_genomic_accession=122286503, orientation='plus', exon_count=12, OMIM=601199, no_hdr0=''),
  847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=847, CurrentID=0, Status='live', Symbol='CAT', Aliases='', description='catalase', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34438925, end_position_on_the_genomic_accession=34472060, orientation='plus', exon_count=13, OMIM=115500, no_hdr0=''),
  131920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131920, CurrentID=0, Status='live', Symbol='TMEM207', Aliases='UNQ846', description='transmembrane protein 207', other_designations='SRSR846', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190428655, end_position_on_the_genomic_accession=190449876, orientation='minus', exon_count=6, OMIM=614786, no_hdr0=''),
  100270930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270930, CurrentID=0, Status='live', Symbol='RPL14P2', Aliases='RPL14_1_377', description='ribosomal protein L14 pseudogene 2', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62779159, end_position_on_the_genomic_accession=62779813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270933, CurrentID=0, Status='live', Symbol='RPL15P6', Aliases='RPL15_1_425', description='ribosomal protein L15 pseudogene 6', other_designations='', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158275809, end_position_on_the_genomic_accession=158276421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270934, CurrentID=0, Status='live', Symbol='RPL15P8', Aliases='RPL15_2_518', description='ribosomal protein L15 pseudogene 8', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128069020, end_position_on_the_genomic_accession=128069612, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270935, CurrentID=0, Status='live', Symbol='RPL15P9', Aliases='RPL15P10, RPL15_3_755', description='ribosomal protein L15 pseudogene 9', other_designations='ribosomal protein L15 pseudogene 10', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131780670, end_position_on_the_genomic_accession=131781348, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=856, CurrentID=0, Status='live', Symbol='CATR1', Aliases='', description='CATR tumorigenicity conversion 1', other_designations='CATR tumorigenic conversion 1', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600676, no_hdr0=''),
  857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=857, CurrentID=0, Status='live', Symbol='CAV1', Aliases='BSCL3, CGL3, LCCNS, MSTP085, PPH3, VIP21', description='caveolin 1', other_designations='caveolin 1, caveolae protein, 22kDa|cell growth-inhibiting protein 32', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116524785, end_position_on_the_genomic_accession=116561185, orientation='plus', exon_count=4, OMIM=601047, no_hdr0=''),
  858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=858, CurrentID=0, Status='live', Symbol='CAV2', Aliases='CAV', description='caveolin 2', other_designations='caveolae protein, 20-kD|caveolin 2 isoform a and b', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116499601, end_position_on_the_genomic_accession=116508541, orientation='plus', exon_count=3, OMIM=601048, no_hdr0=''),
  859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=859, CurrentID=0, Status='live', Symbol='CAV3', Aliases='LGMD1C, LQT9, VIP-21, VIP21', description='caveolin 3', other_designations='M-caveolin', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8733800, end_position_on_the_genomic_accession=8746765, orientation='plus', exon_count=2, OMIM=601253, no_hdr0=''),
  860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=860, CurrentID=0, Status='live', Symbol='RUNX2', Aliases='AML3, CBF-alpha-1, CBFA1, CCD, CCD1, CLCD, OSF-2, OSF2, PEA2aA, PEBP2aA', description='runt-related transcription factor 2', other_designations='PEA2-alpha A|PEBP2-alpha A|SL3-3 enhancer factor 1 alpha A subunit|SL3/AKV core-binding factor alpha A subunit|acute myeloid leukemia 3 protein|core-binding factor, runt domain, alpha subunit 1|oncogene AML-3|osteoblast-specific transcription factor 2|polyomavirus enhancer-binding protein 2 alpha A subunit', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45327800, end_position_on_the_genomic_accession=45664032, orientation='plus', exon_count=16, OMIM=600211, no_hdr0=''),
  861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=861, CurrentID=0, Status='live', Symbol='RUNX1', Aliases='AML1, AML1-EVI-1, AMLCR1, CBF2alpha, CBFA2, EVI-1, PEBP2aB, PEBP2alpha', description='runt-related transcription factor 1', other_designations='AML1-EVI-1 fusion protein|CBF-alpha-2|PEA2-alpha B|PEBP2-alpha B|SL3-3 enhancer factor 1 alpha B subunit|SL3/AKV core-binding factor alpha B subunit|acute myeloid leukemia 1 protein|core-binding factor, runt domain, alpha subunit 2|oncogene AML-1|polyomavirus enhancer-binding protein 2 alpha B subunit|transcription factor', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34787801, end_position_on_the_genomic_accession=35049310, orientation='minus', exon_count=13, OMIM=151385, no_hdr0=''),
  862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=862, CurrentID=0, Status='live', Symbol='RUNX1T1', Aliases='AML1-MTG8, AML1T1, CBFA2T1, CDR, ETO, MTG8, ZMYND2', description='runt-related transcription factor 1; translocated to, 1 (cyclin D-related)', other_designations='CBFA2T1 isoform r1t1-11a61|CBFA2T1 isoform r1t1-11a62|CBFA2T1 isoform r1t1-11a63|CBFA2T1 isoform r1t1-11a64|CBFA2T1 isoform r1t1-11a65|CBFA2T1 isoform r1t1-7a47|CBFA2T1 isoform r1t1-7a48|CBFA2T1 isoform r1t1-7a49|CBFA2T1 isoform r1t1-7a50|CBFA2T1 isoform r1t1-7a51|CBFA2T1 isoform r1t1-7a52|CBFA2T1 isoform r1t1-7d53|CBFA2T1 isoform r1t1-7d54|CBFA2T1 isoform r1t1-7d55|CBFA2T1 isoform r1t1-7d56|CBFA2T1 isoform r1t1-8a57|CBFA2T1 isoform r1t1-8a58|CBFA2T1 isoform r1t1-8a59|CBFA2T1 isoform r1t1-8a60|acute myelogenous leukemia 1 translocation 1, cyclin-D related|core-binding factor, runt domain, alpha subunit 2; translocated to, 1; cyclin D-related|eight twenty one protein|myeloid translocation gene on 8q22|zinc finger MYND domain-containing protein 2', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91954967, end_position_on_the_genomic_accession=92103226, orientation='minus', exon_count=18, OMIM=133435, no_hdr0=''),
  863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=863, CurrentID=0, Status='live', Symbol='CBFA2T3', Aliases='ETO2, MTG16, MTGR2, ZMYND4', description='core-binding factor, runt domain, alpha subunit 2; translocated to, 3', other_designations='MTG8-related gene 2|MTG8-related protein 2|myeloid translocation gene 8 and 16b|myeloid translocation gene on chromosome 16 protein|zinc finger MYND domain-containing protein 4', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88874855, end_position_on_the_genomic_accession=88977199, orientation='minus', exon_count=16, OMIM=603870, no_hdr0=''),
  864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=864, CurrentID=0, Status='live', Symbol='RUNX3', Aliases='AML2, CBFA3, PEBP2aC', description='runt-related transcription factor 3', other_designations='CBF-alpha-3|PEA2 alpha C|PEA2-alpha C|PEBP2 alpha C|PEBP2-alpha C|SL3-3 enhancer factor 1 alpha C subunit|SL3/AKV core-binding factor alpha C subunit|acute myeloid leukemia 2 protein|acute myeloid leukemia gene 2|core-binding factor subunit alpha-3|core-binding factor, runt domain, alpha subunit 3|oncogene AML-2|polyomavirus enhancer-binding protein 2 alpha C subunit|transcription factor AML2', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24899511, end_position_on_the_genomic_accession=24965157, orientation='minus', exon_count=7, OMIM=600210, no_hdr0=''),
  865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=865, CurrentID=0, Status='live', Symbol='CBFB', Aliases='PEBP2B', description='core-binding factor, beta subunit', other_designations='CBF-beta|PEA2-beta|PEBP2-beta|SL3-3 enhancer factor 1 beta subunit|SL3-3 enhancer factor 1 subunit beta|SL3/AKV core-binding factor beta subunit|polyomavirus enhancer binding protein 2, beta subunit|polyomavirus enhancer-binding protein 2 beta subunit', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67029147, end_position_on_the_genomic_accession=67101058, orientation='plus', exon_count=6, OMIM=121360, no_hdr0=''),
  866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=866, CurrentID=0, Status='live', Symbol='SERPINA6', Aliases='CBG', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 6', other_designations='corticosteroid binding globulin|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 6|serpin A6|transcortin', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94304248, end_position_on_the_genomic_accession=94323351, orientation='minus', exon_count=5, OMIM=122500, no_hdr0=''),
  867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=867, CurrentID=0, Status='live', Symbol='CBL', Aliases='C-CBL2, FRA11B, NSLL, RNF55, CBL', description='Cbl proto-oncogene, E3 ubiquitin protein ligase', other_designations='Cas-Br-M (murine) ecotropic retroviral transforming sequence|RING finger protein 55|casitas B-lineage lymphoma proto-oncogene|fragile site, folic acid type, rare, fra(11)(q23.3)|oncogene CBL2|proto-oncogene c-Cbl|signal transduction protein CBL', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119206276, end_position_on_the_genomic_accession=119308149, orientation='plus', exon_count=17, OMIM=165360, no_hdr0=''),
  868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=868, CurrentID=0, Status='live', Symbol='CBLB', Aliases='Cbl-b, Nbla00127, RNF56', description='Cbl proto-oncogene B, E3 ubiquitin protein ligase', other_designations='Cas-Br-M (murine) ecotropic retroviral transforming sequence b|Cas-Br-M (murine) ectropic retroviral transforming sequence b|Cbl proto-oncogene, E3 ubiquitin protein ligase B|RING finger protein 56|SH3-binding protein CBL-B|casitas B-lineage lymphoma proto-oncogene b|signal transduction protein CBL-B', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=105655461, end_position_on_the_genomic_accession=105869552, orientation='minus', exon_count=22, OMIM=604491, no_hdr0=''),
  869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=869, CurrentID=0, Status='live', Symbol='CBLN1', Aliases='', description='cerebellin 1 precursor', other_designations='precerebellin', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49277917, end_position_on_the_genomic_accession=49281831, orientation='minus', exon_count=3, OMIM=600432, no_hdr0=''),
  100270950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270950, CurrentID=0, Status='live', Symbol='RPL19P11', Aliases='RPL19_6_547', description='ribosomal protein L19 pseudogene 11', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31053523, end_position_on_the_genomic_accession=31054248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=871, CurrentID=0, Status='live', Symbol='SERPINH1', Aliases='AsTP3, CBP1, CBP2, HSP47, OI10, PIG14, PPROM, RA-A47, SERPINH2, gp46', description='serpin peptidase inhibitor, clade H (heat shock protein 47), member 1, (collagen binding protein 1)', other_designations='47 kDa heat shock protein|arsenic-transactivated protein 3|cell proliferation-inducing gene 14 protein|colligin-1|colligin-2|rheumatoid arthritis antigen A-47|rheumatoid arthritis-related antigen RA-A47|serine (or cysteine) proteinase inhibitor, clade H (heat shock protein 47), member 1, (collagen binding protein 1)|serine (or cysteine) proteinase inhibitor, clade H (heat shock protein 47), member 2, (collagen-binding protein 2)', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75562056, end_position_on_the_genomic_accession=75572804, orientation='plus', exon_count=8, OMIM=600943, no_hdr0=''),
  100270952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270952, CurrentID=0, Status='live', Symbol='RPL21P27', Aliases='RPL21_3_112', description='ribosomal protein L21 pseudogene 27', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165676272, end_position_on_the_genomic_accession=165676816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=873, CurrentID=0, Status='live', Symbol='CBR1', Aliases='CBR, SDR21C1, hCBR1', description='carbonyl reductase 1', other_designations='15-hydroxyprostaglandin dehydrogenase|NADPH-dependent carbonyl reductase 1|carbonyl reductase (NADPH) 1|prostaglandin 9-ketoreductase|prostaglandin-E(2) 9-reductase|short chain dehydrogenase/reductase family 21C member 1|short chain dehydrogenase/reductase family 21C, member 1', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36069924, end_position_on_the_genomic_accession=36073177, orientation='plus', exon_count=3, OMIM=114830, no_hdr0=''),
  874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=874, CurrentID=0, Status='live', Symbol='CBR3', Aliases='HEL-S-25, SDR21C2, hCBR3', description='carbonyl reductase 3', other_designations='NADPH-dependent carbonyl reductase 3|carbonyl reductase (NADPH) 3|epididymis secretory protein Li 25|short chain dehydrogenase/reductase family 21C member 2|short chain dehydrogenase/reductase family 21C, member 2', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36134900, end_position_on_the_genomic_accession=36146562, orientation='plus', exon_count=4, OMIM=603608, no_hdr0=''),
  875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=875, CurrentID=0, Status='live', Symbol='CBS', Aliases='HIP4', description='cystathionine-beta-synthase', other_designations='beta-thionase|methylcysteine synthase|serine sulfhydrase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43053191, end_position_on_the_genomic_accession=43076868, orientation='minus', exon_count=24, OMIM=613381, no_hdr0=''),
  100270956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270956, CurrentID=0, Status='live', Symbol='RPL21P36', Aliases='RPL21_9_206', description='ribosomal protein L21 pseudogene 36', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36299359, end_position_on_the_genomic_accession=36299884, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270957, CurrentID=0, Status='live', Symbol='RPL22P8', Aliases='RPL22_4_256', description='ribosomal protein L22 pseudogene 8', other_designations='', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107914910, end_position_on_the_genomic_accession=107915295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=878, CurrentID=0, Status='live', Symbol='CCA1', Aliases='CTRCT7', description='cataract, congenital, cerulean type, 1', other_designations='', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=115660, no_hdr0=''),
  100270959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270959, CurrentID=0, Status='live', Symbol='RPL22P16', Aliases='RPL22_6_780', description='ribosomal protein L22 pseudogene 16', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4865430, end_position_on_the_genomic_accession=4865802, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270960, CurrentID=0, Status='live', Symbol='RPL22P14', Aliases='RPL22_8_723', description='ribosomal protein L22 pseudogene 14', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89876511, end_position_on_the_genomic_accession=89876878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533105, CurrentID=0, Status='live', Symbol='C8orf44-SGK3', Aliases='', description='C8orf44-SGK3 readthrough', other_designations='cytokine-independent survival kinase|serum/glucocorticoid-regulated kinase 3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66667552, end_position_on_the_genomic_accession=66862022, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=882, CurrentID=0, Status='live', Symbol='CCAL1', Aliases='CPDD', description='chondrocalcinosis 1 (calcium pyrophosphate-deposition disease, early onset osteoarthritis)', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600668, no_hdr0=''),
  883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=883, CurrentID=0, Status='live', Symbol='CCBL1', Aliases='GTK, KAT1, KATI', description='cysteine conjugate-beta lyase, cytoplasmic', other_designations='beta-lysase, kidney|cysteine conjugate-beta lyase; cytoplasmic (glutamine transaminase K, kyneurenine aminotransferase)|cysteine-S-conjugate beta-lyase|glutamine transaminase K|glutamine--phenylpyruvate transaminase|glutamine-phenylpyruvate aminotransferase|kyneurenine aminotransferase|kynurenine aminotransferase I|kynurenine--oxoglutarate transaminase I', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128832942, end_position_on_the_genomic_accession=128882513, orientation='minus', exon_count=19, OMIM=600547, no_hdr0=''),
  100270964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270964, CurrentID=0, Status='live', Symbol='RPL24P8', Aliases='RPL24_4_982', description='ribosomal protein L24 pseudogene 8', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70217157, end_position_on_the_genomic_accession=70217706, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=885, CurrentID=0, Status='live', Symbol='CCK', Aliases='', description='cholecystokinin', other_designations='cholecystokinin triacontatriapeptide|prepro-cholecystokinin', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42257826, end_position_on_the_genomic_accession=42266170, orientation='minus', exon_count=6, OMIM=118440, no_hdr0=''),
  886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=886, CurrentID=0, Status='live', Symbol='CCKAR', Aliases='CCK-A, CCK1-R, CCK1R, CCKRA', description='cholecystokinin A receptor', other_designations='CCK-A receptor|CCK-AR|cholecystokinin type-A receptor|cholecystokinin-1 receptor', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26481396, end_position_on_the_genomic_accession=26490420, orientation='minus', exon_count=5, OMIM=118444, no_hdr0=''),
  887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=887, CurrentID=0, Status='live', Symbol='CCKBR', Aliases='CCK-B, CCK2R, GASR', description='cholecystokinin B receptor', other_designations='CCK-B receptor|CCK-BR|CCK2 receptor|CCK2-R|cholecystokinin-2 receptor|gastrin receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6259611, end_position_on_the_genomic_accession=6272127, orientation='plus', exon_count=5, OMIM=118445, no_hdr0=''),
  100270968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270968, CurrentID=0, Status='live', Symbol='RPL26P18', Aliases='RPL26_9_632', description='ribosomal protein L26 pseudogene 18', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157422404, end_position_on_the_genomic_accession=157422924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=889, CurrentID=0, Status='live', Symbol='KRIT1', Aliases='CAM, CCM1', description='KRIT1, ankyrin repeat containing', other_designations='ankyrin repeat-containing protein Krit1|cerebral cavernous malformations 1 protein|krev interaction trapped 1', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92198969, end_position_on_the_genomic_accession=92246128, orientation='minus', exon_count=20, OMIM=604214, no_hdr0=''),
  890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=890, CurrentID=0, Status='live', Symbol='CCNA2', Aliases='CCN1, CCNA', description='cyclin A2', other_designations='cyclin-A', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121816444, end_position_on_the_genomic_accession=121823933, orientation='minus', exon_count=8, OMIM=123835, no_hdr0=''),
  891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=891, CurrentID=0, Status='live', Symbol='CCNB1', Aliases='CCNB', description='cyclin B1', other_designations='G2/mitotic-specific cyclin B1', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69167010, end_position_on_the_genomic_accession=69178245, orientation='plus', exon_count=9, OMIM=123836, no_hdr0=''),
  892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=892, CurrentID=0, Status='live', Symbol='CCNC', Aliases='CycC', description='cyclin C', other_designations='SRB11 homolog|hSRB11', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99542387, end_position_on_the_genomic_accession=99568973, orientation='minus', exon_count=14, OMIM=123838, no_hdr0=''),
  131965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131965, CurrentID=0, Status='live', Symbol='METTL6', Aliases='', description='methyltransferase like 6', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15381063, end_position_on_the_genomic_accession=15427547, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=894, CurrentID=0, Status='live', Symbol='CCND2', Aliases='KIAK0002, MPPH3', description='cyclin D2', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4273733, end_position_on_the_genomic_accession=4305356, orientation='plus', exon_count=6, OMIM=123833, no_hdr0=''),
  895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=895, CurrentID=0, Status='live', Symbol='CCND2P1', Aliases='CCND2P', description='cyclin D2 pseudogene 1', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63243085, end_position_on_the_genomic_accession=63244037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=896, CurrentID=0, Status='live', Symbol='CCND3', Aliases='', description='cyclin D3', other_designations='D3-type cyclin', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41934933, end_position_on_the_genomic_accession=42048894, orientation='minus', exon_count=7, OMIM=123834, no_hdr0=''),
  897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=897, CurrentID=0, Status='live', Symbol='CCND3P1', Aliases='CCND3P', description='cyclin D3 pseudogene 1', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30403194, end_position_on_the_genomic_accession=30403821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=898, CurrentID=0, Status='live', Symbol='CCNE1', Aliases='CCNE, pCCNE1', description='cyclin E1', other_designations='cyclin E variant ex5del|cyclin E variant ex7del|cyclin Es|cyclin Et|pluripotent specific cyclin E1', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29811898, end_position_on_the_genomic_accession=29824312, orientation='plus', exon_count=12, OMIM=123837, no_hdr0=''),
  899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=899, CurrentID=0, Status='live', Symbol='CCNF', Aliases='FBX1, FBXO1', description='cyclin F', other_designations='F-box only protein 1|G2/mitotic-specific cyclin-F', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2429394, end_position_on_the_genomic_accession=2458858, orientation='plus', exon_count=17, OMIM=600227, no_hdr0=''),
  900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=900, CurrentID=0, Status='live', Symbol='CCNG1', Aliases='CCNG', description='cyclin G1', other_designations='cyclin-G', map_location='5q32-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163437270, end_position_on_the_genomic_accession=163445016, orientation='plus', exon_count=8, OMIM=601578, no_hdr0=''),
  901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=901, CurrentID=0, Status='live', Symbol='CCNG2', Aliases='', description='cyclin G2', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77157204, end_position_on_the_genomic_accession=77170060, orientation='plus', exon_count=10, OMIM=603203, no_hdr0=''),
  902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=902, CurrentID=0, Status='live', Symbol='CCNH', Aliases='CAK, CycH, p34, p37', description='cyclin H', other_designations='CAK complex subunit|CDK-activating kinase complex subunit|MO15-associated protein|cyclin-dependent kinase-activating kinase complex subunit', map_location='5q13.3-q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87391493, end_position_on_the_genomic_accession=87413033, orientation='minus', exon_count=10, OMIM=601953, no_hdr0=''),
  100270983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270983, CurrentID=0, Status='live', Symbol='RPL31P13', Aliases='RPL31_2_164', description='ribosomal protein L31 pseudogene 13', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213428717, end_position_on_the_genomic_accession=213429064, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=904, CurrentID=0, Status='live', Symbol='CCNT1', Aliases='CCNT, CYCT1, HIVE1', description='cyclin T1', other_designations='CDK9-associated C-type protein|cyclin C-related protein|human immunodeficiency virus type 1 (HIV-1) expression (elevated) 1', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48688458, end_position_on_the_genomic_accession=48716998, orientation='minus', exon_count=9, OMIM=143055, no_hdr0=''),
  905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=905, CurrentID=0, Status='live', Symbol='CCNT2', Aliases='CYCT2', description='cyclin T2', other_designations='SDS-stable vimentin-bound DNA fragment HEF42VIM22|cyclin T2a|cyclin T2b|subunit of positive elongation transcription factor b', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134918822, end_position_on_the_genomic_accession=134959345, orientation='plus', exon_count=12, OMIM=603862, no_hdr0=''),
  100270986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270986, CurrentID=0, Status='live', Symbol='RPL31P20', Aliases='RPL31_6_349', description='ribosomal protein L31 pseudogene 20', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24430766, end_position_on_the_genomic_accession=24431108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=907, CurrentID=0, Status='live', Symbol='CCT', Aliases='', description='cataract, congenital, total', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302200, no_hdr0=''),
  908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=908, CurrentID=0, Status='live', Symbol='CCT6A', Aliases='CCT-zeta, CCT-zeta-1, CCT6, Cctz, HTR3, MoDP-2, TCP-1-zeta, TCP20, TCPZ, TTCP20', description='chaperonin containing TCP1, subunit 6A (zeta 1)', other_designations='T-complex protein 1, zeta subunit|acute morphine dependence related protein 2|acute morphine dependence-related protein 2|amino acid transport defect-complementing|chaperonin containing T-complex subunit 6|histidine transport regulator 3', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56051685, end_position_on_the_genomic_accession=56063989, orientation='plus', exon_count=14, OMIM=104613, no_hdr0=''),
  909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=909, CurrentID=0, Status='live', Symbol='CD1A', Aliases='CD1, FCB6, HTA1, R4, T6', description='CD1a molecule', other_designations='CD1A antigen, a polypeptide|T-cell surface antigen T6/Leu-6|cluster of differentiation 1 A|cortical thymocyte antigen CD1A|differentiation antigen CD1-alpha-3|epidermal dendritic cell marker CD1a|hTa1 thymocyte antigen', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158248329, end_position_on_the_genomic_accession=158258269, orientation='plus', exon_count=7, OMIM=188370, no_hdr0=''),
  910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=910, CurrentID=0, Status='live', Symbol='CD1B', Aliases='CD1, CD1A, R1', description='CD1b molecule', other_designations='CD1B antigen, b polypeptide|cortical thymocyte antigen CD1B|differentiation antigen CD1-alpha-3', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158327950, end_position_on_the_genomic_accession=158331531, orientation='minus', exon_count=6, OMIM=188360, no_hdr0=''),
  911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=911, CurrentID=0, Status='live', Symbol='CD1C', Aliases='BDCA1, CD1, CD1A, R7', description='CD1c molecule', other_designations='CD1C antigen, c polypeptide|cortical thymocyte antigen CD1C|differentiation antigen CD1-alpha-3', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158289773, end_position_on_the_genomic_accession=158294774, orientation='plus', exon_count=6, OMIM=188340, no_hdr0=''),
  912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=912, CurrentID=0, Status='live', Symbol='CD1D', Aliases='CD1A, R3', description='CD1d molecule', other_designations='CD1D antigen, d polypeptide|HMC class I antigen-like glycoprotein CD1D|R3G1|T-cell surface glycoprotein CD1d|differentiation antigen CD1-alpha-3|thymocyte antigen CD1D', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158178038, end_position_on_the_genomic_accession=158186427, orientation='plus', exon_count=8, OMIM=188410, no_hdr0=''),
  913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=913, CurrentID=0, Status='live', Symbol='CD1E', Aliases='CD1A, R2', description='CD1e molecule', other_designations='CD1E antigen, e polypeptide|R2G1|differentiation antigen CD1-alpha-3|hCD1e|leukocyte differentiation antigen|thymocyte antigen CD1E', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158353435, end_position_on_the_genomic_accession=158357553, orientation='plus', exon_count=6, OMIM=188411, no_hdr0=''),
  914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=914, CurrentID=0, Status='live', Symbol='CD2', Aliases='LFA-2, SRBC, T11', description='CD2 molecule', other_designations='CD2 antigen (p50), sheep red blood cell receptor|LFA-3 receptor|T-cell surface antigen T11/Leu-5|erythrocyte receptor|lymphocyte-function antigen-2|rosette receptor', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116754435, end_position_on_the_genomic_accession=116769229, orientation='plus', exon_count=5, OMIM=186990, no_hdr0=''),
  915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=915, CurrentID=0, Status='live', Symbol='CD3D', Aliases='CD3-DELTA, IMD19, T3D', description='CD3d molecule, delta (CD3-TCR complex)', other_designations='CD3 antigen, delta subunit|CD3 delta|CD3d antigen, delta polypeptide (TiT3 complex)|OKT3, delta chain|T-cell receptor T3 delta chain', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118339074, end_position_on_the_genomic_accession=118342744, orientation='minus', exon_count=5, OMIM=186790, no_hdr0=''),
  916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=916, CurrentID=0, Status='live', Symbol='CD3E', Aliases='IMD18, T3E, TCRE', description='CD3e molecule, epsilon (CD3-TCR complex)', other_designations='CD3-epsilon|CD3e antigen, epsilon polypeptide (TiT3 complex)|T-cell antigen receptor complex, epsilon subunit of T3|T-cell surface antigen T3/Leu-4 epsilon chain', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118304580, end_position_on_the_genomic_accession=118316175, orientation='plus', exon_count=9, OMIM=186830, no_hdr0=''),
  917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=917, CurrentID=0, Status='live', Symbol='CD3G', Aliases='CD3-GAMMA, IMD17, T3G', description='CD3g molecule, gamma (CD3-TCR complex)', other_designations='CD3g antigen, gamma polypeptide (TiT3 complex)|CD3g molecule, epsilon (CD3-TCR complex)|T-cell antigen receptor complex, gamma subunit of T3|T-cell receptor T3 gamma chain', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118344316, end_position_on_the_genomic_accession=118353782, orientation='plus', exon_count=7, OMIM=186740, no_hdr0=''),
  100270998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270998, CurrentID=0, Status='live', Symbol='RPL34P6', Aliases='RPL34_4_146', description='ribosomal protein L34 pseudogene 6', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200863808, end_position_on_the_genomic_accession=200864143, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=919, CurrentID=0, Status='live', Symbol='CD247', Aliases='CD3-ZETA, CD3H, CD3Q, CD3Z, IMD25, T3Z, TCRZ', description='CD247 molecule', other_designations='CD247 antigen, zeta subunit|CD3Z antigen, zeta polypeptide (TiT3 complex)|CD3zeta chain|T-cell antigen receptor complex, zeta subunit of CD3|T-cell receptor T3 zeta chain|TCR zeta chain', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167430640, end_position_on_the_genomic_accession=167518616, orientation='minus', exon_count=9, OMIM=186780, no_hdr0=''),
  920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=920, CurrentID=0, Status='live', Symbol='CD4', Aliases='CD4mut', description='CD4 molecule', other_designations='CD4 antigen (p55)|CD4 receptor|T-cell surface antigen T4/Leu-3', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6789472, end_position_on_the_genomic_accession=6820810, orientation='plus', exon_count=10, OMIM=186940, no_hdr0=''),
  921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=921, CurrentID=0, Status='live', Symbol='CD5', Aliases='LEU1, T1', description='CD5 molecule', other_designations='CD5 antigen (p56-62)|lymphocyte antigen T1/Leu-1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61102458, end_position_on_the_genomic_accession=61127852, orientation='plus', exon_count=11, OMIM=153340, no_hdr0=''),
  922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=922, CurrentID=0, Status='live', Symbol='CD5L', Aliases='AIM, API6, CT-2, PRO229, SP-ALPHA, Spalpha', description='CD5 molecule like', other_designations='CD5 antigen-like (scavenger receptor cysteine rich family)|CD5 molecule-like|apoptosis inhibitor 6|apoptosis inhibitor of macrophage|igM-associated peptide', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157827071, end_position_on_the_genomic_accession=157841844, orientation='minus', exon_count=7, OMIM=602592, no_hdr0=''),
  923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=923, CurrentID=0, Status='live', Symbol='CD6', Aliases='TP120', description='CD6 molecule', other_designations='CD6 antigen|T12', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60971641, end_position_on_the_genomic_accession=61020377, orientation='plus', exon_count=13, OMIM=186720, no_hdr0=''),
  924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=924, CurrentID=0, Status='live', Symbol='CD7', Aliases='GP40, LEU-9, TP41, Tp40', description='CD7 molecule', other_designations='CD7 antigen (p41)|T-cell leukemia antigen|T-cell surface antigen Leu-9|p41 protein', map_location='17q25.2-q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82314865, end_position_on_the_genomic_accession=82317604, orientation='minus', exon_count=4, OMIM=186820, no_hdr0=''),
  925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=925, CurrentID=0, Status='live', Symbol='CD8A', Aliases='CD8, Leu2, MAL, p32', description='CD8a molecule', other_designations='CD8 antigen, alpha polypeptide (p32)|Leu2 T-lymphocyte antigen|OKT8 T-cell antigen|T cell co-receptor|T-cell antigen Leu2|T-lymphocyte differentiation antigen T8/Leu-2|T8 T-cell antigen', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86784605, end_position_on_the_genomic_accession=86808396, orientation='minus', exon_count=10, OMIM=186910, no_hdr0=''),
  926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=926, CurrentID=0, Status='live', Symbol='CD8B', Aliases='CD8B1, LEU2, LY3, LYT3, P37', description='CD8b molecule', other_designations='CD8 antigen, beta polypeptide 1 (p37)|T lymphocyte surface glycoprotein beta chain', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86815337, end_position_on_the_genomic_accession=86861924, orientation='minus', exon_count=8, OMIM=186730, no_hdr0=''),
  927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=927, CurrentID=0, Status='live', Symbol='CD8BP', Aliases='CD8B2', description='CD8b molecule pseudogene', other_designations='CD8 antigen, beta polypeptide pseudogene', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106487368, end_position_on_the_genomic_accession=106507681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=928, CurrentID=0, Status='live', Symbol='CD9', Aliases='BTCC-1, DRAP-27, MIC3, MRP-1, TSPAN-29, TSPAN29', description='CD9 molecule', other_designations='5H9 antigen|BA-2/p24 antigen|CD9 antigen (p24)|cell growth-inhibiting gene 2 protein|leukocyte antigen MIC3|motility related protein-1|tetraspanin-29', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6199844, end_position_on_the_genomic_accession=6238271, orientation='plus', exon_count=9, OMIM=143030, no_hdr0=''),
  929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=929, CurrentID=0, Status='live', Symbol='CD14', Aliases='', description='CD14 molecule', other_designations='myeloid cell-specific leucine-rich glycoprotein', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140631728, end_position_on_the_genomic_accession=140633701, orientation='minus', exon_count=3, OMIM=158120, no_hdr0=''),
  930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=930, CurrentID=0, Status='live', Symbol='CD19', Aliases='B4, CVID3', description='CD19 molecule', other_designations='B-lymphocyte surface antigen B4|T-cell surface antigen Leu-12|differentiation antigen CD19', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28931916, end_position_on_the_genomic_accession=28939347, orientation='plus', exon_count=15, OMIM=107265, no_hdr0=''),
  931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=931, CurrentID=0, Status='live', Symbol='MS4A1', Aliases='B1, Bp35, CD20, CVID5, LEU-16, MS4A2, S7', description='membrane-spanning 4-domains, subfamily A, member 1', other_designations='B-lymphocyte cell-surface antigen B1|CD20 antigen|CD20 receptor|leukocyte surface antigen Leu-16', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60455809, end_position_on_the_genomic_accession=60470752, orientation='plus', exon_count=8, OMIM=112210, no_hdr0=''),
  932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=932, CurrentID=0, Status='live', Symbol='MS4A3', Aliases='CD20L, HTM4', description='membrane-spanning 4-domains, subfamily A, member 3 (hematopoietic cell-specific)', other_designations='CD20 antigen homolog|CD20 antigen-like protein|IgE receptor beta chain|IgE receptor beta subunit|hematopoietic cell 4 transmembrane protein|hematopoietic-specific transmembrane protein 4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60056628, end_position_on_the_genomic_accession=60071116, orientation='plus', exon_count=8, OMIM=606498, no_hdr0=''),
  933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=933, CurrentID=0, Status='live', Symbol='CD22', Aliases='SIGLEC-2, SIGLEC2', description='CD22 molecule', other_designations='B-lymphocyte cell adhesion molecule|BL-CAM|CD22 antigen|T-cell surface antigen Leu-14|sialic acid binding Ig-like lectin 2|sialic acid-binding Ig-like lectin 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35329166, end_position_on_the_genomic_accession=35347361, orientation='plus', exon_count=14, OMIM=107266, no_hdr0=''),
  100271014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271014, CurrentID=0, Status='live', Symbol='RPL37P8', Aliases='RPL37_2_190', description='ribosomal protein L37 pseudogene 8', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244694430, end_position_on_the_genomic_accession=244694720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=935, CurrentID=0, Status='live', Symbol='CD24P1', Aliases='CD24L1', description='CD24 molecule pseudogene 1', other_designations='CD24 antigen-like 1|CD24 molecule-like 1', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15609982, end_position_on_the_genomic_accession=15614946, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=936, CurrentID=0, Status='live', Symbol='CD24P2', Aliases='CD24L2', description='CD24 molecule pseudogene 2', other_designations='CD24 antigen-like 2|CD24 molecule-like 2', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56071339, end_position_on_the_genomic_accession=56074088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=937, CurrentID=0, Status='live', Symbol='CD24P3', Aliases='CD24L3', description='CD24 molecule pseudogene 3', other_designations='CD24 antigen-like 3|CD24 molecule-like 3', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31510275, end_position_on_the_genomic_accession=31511768, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=938, CurrentID=0, Status='live', Symbol='CD24P4', Aliases='CD24L4', description='CD24 molecule pseudogene 4', other_designations='CD24 antigen pseudogene|CD24 antigen-like 4|CD24 molecule-like 4 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18990640, end_position_on_the_genomic_accession=18992818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=939, CurrentID=0, Status='live', Symbol='CD27', Aliases='S152, S152. LPFS2, T14, TNFRSF7, Tp55', description='CD27 molecule', other_designations='CD27L receptor|T cell activation antigen S152|T-cell activation antigen CD27|tumor necrosis factor receptor superfamily, member 7', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6444885, end_position_on_the_genomic_accession=6451718, orientation='plus', exon_count=7, OMIM=186711, no_hdr0=''),
  940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=940, CurrentID=0, Status='live', Symbol='CD28', Aliases='Tp44', description='CD28 molecule', other_designations='CD28 antigen', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203706475, end_position_on_the_genomic_accession=203738912, orientation='plus', exon_count=4, OMIM=186760, no_hdr0=''),
  941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=941, CurrentID=0, Status='live', Symbol='CD80', Aliases='B7, B7-1, B7.1, BB1, CD28LG, CD28LG1, LAB7', description='CD80 molecule', other_designations='B-lymphocyte activation antigen B7|CD80 antigen (CD28 antigen ligand 1, B7-1 antigen)|CTLA-4 counter-receptor B7.1|activation B7-1 antigen|costimulatory factor CD80|costimulatory molecule variant IgV-CD80', map_location='3q13.3-q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119523948, end_position_on_the_genomic_accession=119559634, orientation='minus', exon_count=10, OMIM=112203, no_hdr0=''),
  942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=942, CurrentID=0, Status='live', Symbol='CD86', Aliases='B7-2, B7.2, B70, CD28LG2, LAB72', description='CD86 molecule', other_designations='B-lymphocyte activation antigen B7-2|BU63|CD86 antigen (CD28 antigen ligand 2, B7-2 antigen)|CTLA-4 counter-receptor B7.2|FUN-1', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122055362, end_position_on_the_genomic_accession=122121143, orientation='plus', exon_count=8, OMIM=601020, no_hdr0=''),
  943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=943, CurrentID=0, Status='live', Symbol='TNFRSF8', Aliases='CD30, D1S166E, Ki-1', description='tumor necrosis factor receptor superfamily member 8', other_designations='CD30L receptor|Ki-1 antigen|cytokine receptor CD30|lymphocyte activation antigen CD30|tumor necrosis factor receptor superfamily, member 8', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12063330, end_position_on_the_genomic_accession=12144207, orientation='plus', exon_count=18, OMIM=153243, no_hdr0=''),
  944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=944, CurrentID=0, Status='live', Symbol='TNFSF8', Aliases='CD153, CD30L, CD30LG', description='tumor necrosis factor superfamily member 8', other_designations='CD153 antigen|CD30 antigen ligand|CD30 ligand|CD30-L|tumor necrosis factor (ligand) superfamily member 8|tumor necrosis factor (ligand) superfamily, member 8', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114893343, end_position_on_the_genomic_accession=114930595, orientation='minus', exon_count=5, OMIM=603875, no_hdr0=''),
  945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=945, CurrentID=0, Status='live', Symbol='CD33', Aliases='SIGLEC-3, SIGLEC3, p67', description='CD33 molecule', other_designations='CD33 antigen (gp67)|gp67|sialic acid binding Ig-like lectin 3|sialic acid-binding Ig-like lectin 3', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51215537, end_position_on_the_genomic_accession=51240019, orientation='plus', exon_count=11, OMIM=159590, no_hdr0=''),
  946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=946, CurrentID=0, Status='live', Symbol='SIGLEC6', Aliases='CD327, CD33L, CD33L1, CD33L2, CDW327, OBBP1', description='sialic acid binding Ig-like lectin 6', other_designations='CD33 antigen-like 1|obesity-binding protein 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51519530, end_position_on_the_genomic_accession=51531863, orientation='minus', exon_count=11, OMIM=604405, no_hdr0=''),
  947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=947, CurrentID=0, Status='live', Symbol='CD34', Aliases='', description='CD34 molecule', other_designations='CD34 antigen', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207881028, end_position_on_the_genomic_accession=207911338, orientation='minus', exon_count=9, OMIM=142230, no_hdr0=''),
  948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=948, CurrentID=0, Status='live', Symbol='CD36', Aliases='BDPLT10, CHDS7, FAT, GP3B, GP4, GPIV, PASIV, SCARB3', description='CD36 molecule (thrombospondin receptor)', other_designations='CD36 antigen (collagen type I receptor, thrombospondin receptor)|GPIIIB|PAS IV|PAS-4 protein|cluster determinant 36|fatty acid translocase|glycoprotein IIIb|leukocyte differentiation antigen CD36|platelet glycoprotein IV|scavenger receptor class B, member 3', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80602188, end_position_on_the_genomic_accession=80679277, orientation='plus', exon_count=18, OMIM=173510, no_hdr0=''),
  949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=949, CurrentID=0, Status='live', Symbol='SCARB1', Aliases='CD36L1, CLA-1, CLA1, HDLQTL6, SR-BI, SRB1', description='scavenger receptor class B member 1', other_designations='CD36 and LIMPII analogous 1|CD36 antigen (collagen type I receptor, thrombospondin receptor)-like 1|scavenger receptor class B type III|scavenger receptor class B, member 1', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124777628, end_position_on_the_genomic_accession=124863973, orientation='minus', exon_count=13, OMIM=601040, no_hdr0=''),
  950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=950, CurrentID=0, Status='live', Symbol='SCARB2', Aliases='AMRF, CD36L2, EPM4, HLGP85, LGP85, LIMP-2, LIMPII, SR-BII', description='scavenger receptor class B member 2', other_designations='85 kDa lysosomal membrane sialoglycoprotein|85 kDa lysosomal sialoglycoprotein scavenger receptor class B, member 2|CD36 antigen (collagen type I receptor, thrombospondin receptor)-like 2 (lysosomal integral membrane protein II)|CD36 antigen-like 2|LIMP II|lysosome membrane protein II|scavenger receptor class B, member 2', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76158737, end_position_on_the_genomic_accession=76213899, orientation='minus', exon_count=12, OMIM=602257, no_hdr0=''),
  951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=951, CurrentID=0, Status='live', Symbol='CD37', Aliases='GP52-40, TSPAN26', description='CD37 molecule', other_designations='cell differentiation antigen 37|leukocyte surface antigen CD37|tetraspanin-26|tspan-26', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49335365, end_position_on_the_genomic_accession=49340606, orientation='plus', exon_count=8, OMIM=151523, no_hdr0=''),
  952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=952, CurrentID=0, Status='live', Symbol='CD38', Aliases='ADPRC 1, ADPRC1', description='CD38 molecule', other_designations="2'-phospho-ADP-ribosyl cyclase|2'-phospho-cyclic-ADP-ribose transferase|ADP-ribosyl cyclase 1|CD38 antigen (p45)|NAD(+) nucleosidase|cluster of differentiation 38|cyclic ADP-ribose hydrolase 1|ecto-nicotinamide adenine dinucleotide glycohydrolase", map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15778298, end_position_on_the_genomic_accession=15849083, orientation='plus', exon_count=8, OMIM=107270, no_hdr0=''),
  953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=953, CurrentID=0, Status='live', Symbol='ENTPD1', Aliases='ATPDase, CD39, NTPDase-1, SPG64', description='ectonucleoside triphosphate diphosphohydrolase 1', other_designations='CD39 antigen|NTPDase 1|ecto-ATP diphosphohydrolase 1|ecto-ATPDase 1|ecto-ATPase 1|ecto-apyrase|lymphoid cell activation antigen', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95710906, end_position_on_the_genomic_accession=95877266, orientation='plus', exon_count=16, OMIM=601752, no_hdr0=''),
  954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=954, CurrentID=0, Status='live', Symbol='ENTPD2', Aliases='CD39L1, NTPDase-2', description='ectonucleoside triphosphate diphosphohydrolase 2', other_designations='CD39 antigen-like 1|NTPDase 2|ecto-ATP diphosphohydrolase 2|ecto-ATPDase 2|ecto-ATPase 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137048098, end_position_on_the_genomic_accession=137054069, orientation='minus', exon_count=9, OMIM=602012, no_hdr0=''),
  955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=955, CurrentID=0, Status='live', Symbol='ENTPD6', Aliases='CD39L2, IL-6SAG, IL6ST2, NTPDase-6', description='ectonucleoside triphosphate diphosphohydrolase 6 (putative)', other_designations='CD39 antigen-like 2|CD39-like 2|HCV-E2 binding protein 1|NTPDase 6|ectonucleoside triphosphate diphosphohydrolase 6 (putative function)|interleukin 6 signal transducer-2', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25195670, end_position_on_the_genomic_accession=25226724, orientation='plus', exon_count=18, OMIM=603160, no_hdr0=''),
  956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=956, CurrentID=0, Status='live', Symbol='ENTPD3', Aliases='CD39L3, HB6, NTPDase-3', description='ectonucleoside triphosphate diphosphohydrolase 3', other_designations='CD39 antigen-like 3|NTPDase 3|ecto-ATP diphosphohydrolase 3|ecto-ATPDase 3|ecto-ATPase 3|ecto-apyrase 3', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40387156, end_position_on_the_genomic_accession=40428626, orientation='plus', exon_count=13, OMIM=603161, no_hdr0=''),
  957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=957, CurrentID=0, Status='live', Symbol='ENTPD5', Aliases='CD39L4, NTPDase-5, PCPH', description='ectonucleoside triphosphate diphosphohydrolase 5', other_designations='CD39 antigen-like 4|CD39-like 4|ER-UDPase|GDPase ENTPD5|NTPDase 5|Pcph proto-oncogene protein|UDPase ENTPD5|guanosine-diphosphatase ENTPD5|nucleoside diphosphatase|proto-oncogene CPH|uridine-diphosphatase ENTPD5', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73959467, end_position_on_the_genomic_accession=74019326, orientation='minus', exon_count=22, OMIM=603162, no_hdr0=''),
  958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=958, CurrentID=0, Status='live', Symbol='CD40', Aliases='Bp50, CDW40, TNFRSF5, p50', description='CD40 molecule, TNF receptor superfamily member 5', other_designations='B cell surface antigen CD40|B cell-associated molecule|CD40 antigen (TNF receptor superfamily member 5)|CD40 molecule TNF receptor superfamily member 5|CD40 type II isoform|CD40L receptor|nerve growth factor receptor-related B-lymphocyte activation molecule', map_location='20q12-q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46118250, end_position_on_the_genomic_accession=46129745, orientation='plus', exon_count=9, OMIM=109535, no_hdr0=''),
  959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=959, CurrentID=0, Status='live', Symbol='CD40LG', Aliases='CD154, CD40L, HIGM1, IGM, IMD3, T-BAM, TNFSF5, TRAP, gp39, hCD40L', description='CD40 ligand', other_designations='CD40 antigen ligand|CD40-L|T-B cell-activating molecule|T-cell antigen Gp39|TNF-related activation protein|tumor necrosis factor (ligand) superfamily member 5', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136648177, end_position_on_the_genomic_accession=136660390, orientation='plus', exon_count=5, OMIM=300386, no_hdr0=''),
  960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=960, CurrentID=0, Status='live', Symbol='CD44', Aliases='CDW44, CSPG8, ECMR-III, HCELL, HUTCH-I, IN, LHR, MC56, MDU2, MDU3, MIC4, Pgp1', description='CD44 molecule (Indian blood group)', other_designations='GP90 lymphocyte homing/adhesion receptor|Hermes antigen|cell surface glycoprotein CD44|chondroitin sulfate proteoglycan 8|epican|extracellular matrix receptor III|hematopoietic cell E- and L-selectin ligand|heparan sulfate proteoglycan|homing function and Indian blood group system|hyaluronate receptor|phagocytic glycoprotein 1|soluble CD44', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35138870, end_position_on_the_genomic_accession=35232402, orientation='plus', exon_count=21, OMIM=107269, no_hdr0=''),
  961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=961, CurrentID=0, Status='live', Symbol='CD47', Aliases='IAP, MER6, OA3', description='CD47 molecule', other_designations='CD47 antigen (Rh-related antigen, integrin-associated signal transducer)|CD47 glycoprotein|Rh-related antigen|antigen identified by monoclonal antibody 1D8|antigenic surface determinant protein OA3|integrin associated protein|integrin-associated protein|integrin-associated signal transducer', map_location='3q13.1-q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108043094, end_position_on_the_genomic_accession=108091088, orientation='minus', exon_count=13, OMIM=601028, no_hdr0=''),
  962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=962, CurrentID=0, Status='live', Symbol='CD48', Aliases='BCM1, BLAST, BLAST1, MEM-102, SLAMF2, hCD48, mCD48', description='CD48 molecule', other_designations='B-lymphocyte activation marker BLAST-1|BCM1 surface antigen|CD48 antigen (B-cell membrane protein)|SLAM family member 2|TCT.1|leukocyte antigen MEM-102|signaling lymphocytic activation molecule 2', map_location='1q21.3-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160678746, end_position_on_the_genomic_accession=160711851, orientation='minus', exon_count=5, OMIM=109530, no_hdr0=''),
  963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=963, CurrentID=0, Status='live', Symbol='CD53', Aliases='MOX44, TSPAN25', description='CD53 molecule', other_designations='CD53 antigen|CD53 glycoprotein|CD53 tetraspan antigen|antigen MOX44 identified by monoclonal antibody MRC-OX44|cell surface antigen|cell surface glycoprotein CD53|tetraspanin-25|transmembrane glycoprotein|tspan-25', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110871199, end_position_on_the_genomic_accession=110899936, orientation='plus', exon_count=9, OMIM=151525, no_hdr0=''),
  100271044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271044, CurrentID=0, Status='live', Symbol='RPL8P5', Aliases='RPL8_2_1740', description='ribosomal protein L8 pseudogene 5', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19887289, end_position_on_the_genomic_accession=19887970, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=965, CurrentID=0, Status='live', Symbol='CD58', Aliases='LFA-3, LFA3, ag3', description='CD58 molecule', other_designations='CD58 antigen, (lymphocyte function-associated antigen 3)|surface glycoprotein LFA-3', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116514534, end_position_on_the_genomic_accession=116571093, orientation='minus', exon_count=6, OMIM=153420, no_hdr0=''),
  966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=966, CurrentID=0, Status='live', Symbol='CD59', Aliases='16.3A5, 1F5, EJ16, EJ30, EL32, G344, HRF-20, HRF20, MAC-IP, MACIF, MEM43, MIC11, MIN1, MIN2, MIN3, MIRL, MSK21, p18-20', description='CD59 molecule, complement regulatory protein', other_designations='1F5 antigen|20 kDa homologous restriction factor|CD59 antigen p18-20 (antigen identified by monoclonal antibodies 16.3A5, EJ16, EJ30, EL32 and G344)|Ly-6-like protein|MEM43 antigen|T cell-activating protein|human leukocyte antigen MIC11|lymphocytic antigen CD59/MEM43|membrane attack complex (MAC) inhibition factor|membrane attack complex inhibition factor|membrane inhibitor of reactive lysis|protectin|surface anitgen recognized by monoclonal antibody 16.3A5', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33703010, end_position_on_the_genomic_accession=33736479, orientation='minus', exon_count=7, OMIM=107271, no_hdr0=''),
  967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=967, CurrentID=0, Status='live', Symbol='CD63', Aliases='LAMP-3, ME491, MLA1, OMA81H, TSPAN30', description='CD63 molecule', other_designations='CD63 antigen (melanoma 1 antigen)|granulophysin|lysosomal-associated membrane protein 3|lysosome-associated membrane glycoprotein 3|melanoma-associated antigen ME491|melanoma-associated antigen MLA1|ocular melanoma-associated antigen|tetraspanin-30|tspan-30', map_location='12q12-q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55725443, end_position_on_the_genomic_accession=55729673, orientation='minus', exon_count=12, OMIM=155740, no_hdr0=''),
  968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=968, CurrentID=0, Status='live', Symbol='CD68', Aliases='GP110, LAMP4, SCARD1', description='CD68 molecule', other_designations='CD68 antigen|macrophage antigen CD68|scavenger receptor class D, member 1', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7579487, end_position_on_the_genomic_accession=7582111, orientation='plus', exon_count=6, OMIM=153634, no_hdr0=''),
  969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=969, CurrentID=0, Status='live', Symbol='CD69', Aliases='AIM, BL-AC/P26, CLEC2C, EA1, GP32/28, MLR-3', description='CD69 molecule', other_designations='C-type lectin domain family 2, member C|CD69 antigen (p60, early T-cell activation antigen)|activation inducer molecule (AIM/CD69)|early T-cell activation antigen p60|early lymphocyte activation antigen|leukocyte surface antigen Leu-23', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9752486, end_position_on_the_genomic_accession=9760901, orientation='minus', exon_count=5, OMIM=107273, no_hdr0=''),
  970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=970, CurrentID=0, Status='live', Symbol='CD70', Aliases='CD27L, CD27LG, TNFSF7', description='CD70 molecule', other_designations='CD27 ligand|CD27-L|Ki-24 antigen|surface antigen CD70|tumor necrosis factor (ligand) superfamily, member 7|tumor necrosis factor ligand superfamily member 7', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6585839, end_position_on_the_genomic_accession=6591152, orientation='minus', exon_count=3, OMIM=602840, no_hdr0=''),
  971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=971, CurrentID=0, Status='live', Symbol='CD72', Aliases='CD72b, LYB2', description='CD72 molecule', other_designations='CD72 antigen|lyb-2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35609979, end_position_on_the_genomic_accession=35646857, orientation='minus', exon_count=11, OMIM=107272, no_hdr0=''),
  972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=972, CurrentID=0, Status='live', Symbol='CD74', Aliases='DHLAG, HLADG, II, Ia-GAMMA', description='CD74 molecule, major histocompatibility complex, class II invariant chain', other_designations='CD74 antigen (invariant polypeptide of major histocompatibility complex, class II antigen-associated)|HLA-DR antigens-associated invariant chain|HLA-DR-gamma|Ia-associated invariant chain|MHC HLA-DR gamma chain|gamma chain of class II antigens|p33', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150401637, end_position_on_the_genomic_accession=150412936, orientation='minus', exon_count=9, OMIM=142790, no_hdr0=''),
  973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=973, CurrentID=0, Status='live', Symbol='CD79A', Aliases='IGA, MB-1', description='CD79a molecule', other_designations='CD79a antigen (immunoglobulin-associated alpha)|CD79a molecule, immunoglobulin-associated alpha|MB-1 membrane glycoprotein|ig-alpha|membrane-bound immunoglobulin-associated protein|surface IgM-associated protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41877120, end_position_on_the_genomic_accession=41881372, orientation='plus', exon_count=5, OMIM=112205, no_hdr0=''),
  974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=974, CurrentID=0, Status='live', Symbol='CD79B', Aliases='AGM6, B29, IGB', description='CD79b molecule', other_designations='B-cell-specific glycoprotein B29|CD79b antigen (immunoglobulin-associated beta)|CD79b molecule, immunoglobulin-associated beta|Ig-beta|immunoglobulin-associated B29 protein', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63928738, end_position_on_the_genomic_accession=63932353, orientation='minus', exon_count=6, OMIM=147245, no_hdr0=''),
  975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=975, CurrentID=0, Status='live', Symbol='CD81', Aliases='CVID6, S5.7, TAPA1, TSPAN28', description='CD81 molecule', other_designations='26 kDa cell surface protein TAPA-1|CD81 antigen (target of antiproliferative antibody 1)|tetraspanin-28|tspan-28', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2376177, end_position_on_the_genomic_accession=2397419, orientation='plus', exon_count=10, OMIM=186845, no_hdr0=''),
  976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=976, CurrentID=0, Status='live', Symbol='ADGRE5', Aliases='CD97, TM7LN1', description='adhesion G protein-coupled receptor E5', other_designations='CD97 molecule|leukocyte antigen CD97|seven transmembrane helix receptor|seven-span transmembrane protein|seven-transmembrane, heterodimeric receptor associated with inflammation', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14381144, end_position_on_the_genomic_accession=14408725, orientation='plus', exon_count=20, OMIM=601211, no_hdr0=''),
  977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=977, CurrentID=0, Status='live', Symbol='CD151', Aliases='GP27, MER2, PETA-3, RAPH, SFA1, TSPAN24', description='CD151 molecule (Raph blood group)', other_designations='CD151 antigen (Raph blood group)|hemidesmosomal tetraspanin CD151|membrane glycoprotein SFA-1|platelet surface glycoprotein gp27|platelet-endothelial cell tetraspan antigen 3|platelet-endothelial tetraspan antigen 3|tetraspanin-24|tspan-24', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=832952, end_position_on_the_genomic_accession=838835, orientation='plus', exon_count=9, OMIM=602243, no_hdr0=''),
  978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=978, CurrentID=0, Status='live', Symbol='CDA', Aliases='CDD', description='cytidine deaminase', other_designations='cytidine aminohydrolase|cytosine nucleoside deaminase|small cytidine deaminase', map_location='1p36.2-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20588951, end_position_on_the_genomic_accession=20618908, orientation='plus', exon_count=4, OMIM=123920, no_hdr0=''),
  100271059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271059, CurrentID=0, Status='live', Symbol='RPS12P15', Aliases='RPS12_7_934', description='ribosomal protein S12 pseudogene 15', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102503755, end_position_on_the_genomic_accession=102504138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271060, CurrentID=0, Status='live', Symbol='RPS13P4', Aliases='RPS13_2_199', description='ribosomal protein S13 pseudogene 4', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23900693, end_position_on_the_genomic_accession=23901149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=981, CurrentID=0, Status='live', Symbol='CDAN3', Aliases='CDA3, CDAIII', description='congenital dyserythropoietic anemia, type III', other_designations='', map_location='15q21-q25', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=105600, no_hdr0=''),
  982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=982, CurrentID=0, Status='live', Symbol='CDB2', Aliases='CDBII, CDTB', description='corneal dystrophy of Bowman layer type II (Thiel-Behnke)', other_designations="corneal dystrophy of Bowman's layer type II (Thiel-Behnke)", map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602082, no_hdr0=''),
  983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=983, CurrentID=0, Status='live', Symbol='CDK1', Aliases='CDC2, CDC28A, P34CDC2', description='cyclin-dependent kinase 1', other_designations='cell cycle controller CDC2|cell division control protein 2 homolog|cell division cycle 2, G1 to S and G2 to M|cell division protein kinase 1|p34 protein kinase', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60778331, end_position_on_the_genomic_accession=60794852, orientation='plus', exon_count=9, OMIM=116940, no_hdr0=''),
  984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=984, CurrentID=0, Status='live', Symbol='CDK11B', Aliases='CDC2L1, CDK11, CDK11-p110, CDK11-p46, CDK11-p58, CLK-1, PITSLREA, PK58, p58, p58CDC2L1, p58CLK-1', description='cyclin-dependent kinase 11B', other_designations='CDC-related protein kinase p58|PITSLRE serine/threonine-protein kinase CDC2L1|cell division cycle 2-like 1 (PITSLRE proteins)|cell division protein kinase 11B|galactosyltransferase-associated protein kinase p58/GTA|p58 CLK-1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1635226, end_position_on_the_genomic_accession=1659839, orientation='minus', exon_count=22, OMIM=176873, no_hdr0=''),
  100271065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271065, CurrentID=0, Status='live', Symbol='RPS14P7', Aliases='RPS14_4_530', description='ribosomal protein S14 pseudogene 7', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=160818422, end_position_on_the_genomic_accession=160818854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271066, CurrentID=0, Status='live', Symbol='RPS15P2', Aliases='RPS15_1_177', description='ribosomal protein S15 pseudogene 2', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234356673, end_position_on_the_genomic_accession=234357170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=987, CurrentID=0, Status='live', Symbol='LRBA', Aliases='BGL, CDC4L, CVID8, LAB300, LBA', description='LPS-responsive vesicle trafficking, beach and anchor containing', other_designations='CDC4-like protein|vesicle trafficking, beach and anchor containing', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150264659, end_position_on_the_genomic_accession=151015727, orientation='minus', exon_count=60, OMIM=606453, no_hdr0=''),
  988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=988, CurrentID=0, Status='live', Symbol='CDC5L', Aliases='CDC5, CDC5-LIKE, CEF1, PCDC5RP, dJ319D22.1', description='cell division cycle 5 like', other_designations='CDC5 cell division cycle 5-like|Cdc5-related protein|dJ319D22.1 (CDC5-like protein)|pombe cdc5-related protein', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44387514, end_position_on_the_genomic_accession=44450424, orientation='plus', exon_count=18, OMIM=602868, no_hdr0=''),
  989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=989, CurrentID=0, Status='live', Symbol='SEPT7', Aliases='CDC10, CDC3, NBLA02942A, SEPT7', description='septin 7', other_designations='CDC10 (cell division cycle 10, S. cerevisiae, homolog)|CDC10 protein homolog|septin 7 variant 4', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35800986, end_position_on_the_genomic_accession=35915763, orientation='plus', exon_count=19, OMIM=603151, no_hdr0=''),
  990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=990, CurrentID=0, Status='live', Symbol='CDC6', Aliases='CDC18L, HsCDC18, HsCDC6', description='cell division cycle 6', other_designations='CDC6 cell division cycle 6 homolog|CDC6-related protein|cdc18-related protein|cell division cycle 6 homolog|p62(cdc6)', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40287689, end_position_on_the_genomic_accession=40303161, orientation='plus', exon_count=13, OMIM=602627, no_hdr0=''),
  991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=991, CurrentID=0, Status='live', Symbol='CDC20', Aliases='CDC20A, bA276H19.3, p55CDC', description='cell division cycle 20', other_designations='CDC20 cell division cycle 20 homolog|cell division cycle 20 homolog', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43358955, end_position_on_the_genomic_accession=43363203, orientation='plus', exon_count=11, OMIM=603618, no_hdr0=''),
  100271072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271072, CurrentID=0, Status='live', Symbol='RPS17P11', Aliases='RPS17_4_560', description='ribosomal protein S17 pseudogene 11', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52282511, end_position_on_the_genomic_accession=52282862, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=993, CurrentID=0, Status='live', Symbol='CDC25A', Aliases='CDC25A2', description='cell division cycle 25A', other_designations='CDC25 isoform A1-CAG|CDC25A2-CAG isoform|dual specificity phosphatase CDC25A', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48157178, end_position_on_the_genomic_accession=48188658, orientation='minus', exon_count=15, OMIM=116947, no_hdr0=''),
  994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=994, CurrentID=0, Status='live', Symbol='CDC25B', Aliases='', description='cell division cycle 25B', other_designations='dual specificity phosphatase Cdc25B', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3786772, end_position_on_the_genomic_accession=3806121, orientation='plus', exon_count=18, OMIM=116949, no_hdr0=''),
  995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=995, CurrentID=0, Status='live', Symbol='CDC25C', Aliases='CDC25, PPP1R60', description='cell division cycle 25C', other_designations='dual specificity phosphatase CDC25C|mitosis inducer CDC25|phosphotyrosine phosphatase|protein phosphatase 1, regulatory subunit 60', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138285265, end_position_on_the_genomic_accession=138338355, orientation='minus', exon_count=15, OMIM=157680, no_hdr0=''),
  996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=996, CurrentID=0, Status='live', Symbol='CDC27', Aliases='ANAPC3, APC3Hs, D0S1430E, D17S978E, H-NUC, HNUC, NUC2, CDC27', description='cell division cycle 27', other_designations='anaphase promoting complex subunit 3|anaphase-promoting complex, protein 3|cell division cycle 27 homolog|nuc2 homolog', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47117697, end_position_on_the_genomic_accession=47189312, orientation='minus', exon_count=20, OMIM=116946, no_hdr0=''),
  997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=997, CurrentID=0, Status='live', Symbol='CDC34', Aliases='E2-CDC34, UBC3, UBCH3, UBE2R1', description='cell division cycle 34', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme R1|E2 ubiquitin-conjugating enzyme R1|cell division cycle 34 homolog|ubiquitin carrier protein|ubiquitin-conjugating enzyme E2-32 KDA complementing|ubiquitin-conjugating enzyme E2-CDC34|ubiquitin-protein ligase R1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=531719, end_position_on_the_genomic_accession=542088, orientation='plus', exon_count=6, OMIM=116948, no_hdr0=''),
  998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=998, CurrentID=0, Status='live', Symbol='CDC42', Aliases='CDC42Hs, G25K', description='cell division cycle 42', other_designations='G25K GTP-binding protein|GTP binding protein, 25kDa|GTP-binding protein, 25kD|dJ224A6.1.1 (cell division cycle 42 (GTP-binding protein, 25kD))|dJ224A6.1.2 (cell division cycle 42 (GTP-binding protein, 25kD))|growth-regulating protein|small GTP binding protein CDC42', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22052627, end_position_on_the_genomic_accession=22092943, orientation='plus', exon_count=8, OMIM=116952, no_hdr0=''),
  999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=999, CurrentID=0, Status='live', Symbol='CDH1', Aliases='Arc-1, CD324, CDHE, ECAD, LCAM, UVO', description='cadherin 1, type 1', other_designations='CAM 120/80|E-Cadherin|E-cadherin 1|cadherin 1, E-cadherin (epithelial)|cadherin 1, type 1, E-cadherin (epithelial)|calcium-dependent adhesion protein, epithelial|cell-CAM 120/80|epithelial cadherin|uvomorulin', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68737292, end_position_on_the_genomic_accession=68835542, orientation='plus', exon_count=18, OMIM=192090, no_hdr0=''),
  1000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1000, CurrentID=0, Status='live', Symbol='CDH2', Aliases='CD325, CDHN, CDw325, NCAD', description='cadherin 2, type 1, N-cadherin (neuronal)', other_designations='N-cadherin 1|calcium-dependent adhesion protein, neuronal|neural cadherin', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27932879, end_position_on_the_genomic_accession=28177481, orientation='minus', exon_count=20, OMIM=114020, no_hdr0=''),
  1001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1001, CurrentID=0, Status='live', Symbol='CDH3', Aliases='CDHP, HJMD, PCAD', description='cadherin 3, type 1, P-cadherin (placental)', other_designations='calcium-dependent adhesion protein, placental', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68644248, end_position_on_the_genomic_accession=68727577, orientation='plus', exon_count=18, OMIM=114021, no_hdr0=''),
  1002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1002, CurrentID=0, Status='live', Symbol='CDH4', Aliases='CAD4, R-CAD, RCAD', description='cadherin 4, type 1, R-cadherin (retinal)', other_designations='R-cadherin|cadherin 4, type 1, preproprotein|retinal cadherin', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61252426, end_position_on_the_genomic_accession=61940617, orientation='plus', exon_count=18, OMIM=603006, no_hdr0=''),
  1003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1003, CurrentID=0, Status='live', Symbol='CDH5', Aliases='7B4, CD144', description='cadherin 5, type 2 (vascular endothelium)', other_designations='7B4 antigen|VE-cadherin|cadherin 5, type 2, VE-cadherin (vascular epithelium)|cd144 antigen|endothelial-specific cadherin|vascular endothelial cadherin', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66366622, end_position_on_the_genomic_accession=66404786, orientation='plus', exon_count=12, OMIM=601120, no_hdr0=''),
  1004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1004, CurrentID=0, Status='live', Symbol='CDH6', Aliases='CAD6, KCAD', description='cadherin 6, type 2, K-cadherin (fetal kidney)', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31193655, end_position_on_the_genomic_accession=31329146, orientation='plus', exon_count=13, OMIM=603007, no_hdr0=''),
  1005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1005, CurrentID=0, Status='live', Symbol='CDH7', Aliases='CDH7L1', description='cadherin 7, type 2', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=65750252, end_position_on_the_genomic_accession=65880939, orientation='plus', exon_count=14, OMIM=605806, no_hdr0=''),
  1006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1006, CurrentID=0, Status='live', Symbol='CDH8', Aliases='Nbla04261', description='cadherin 8, type 2', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61647244, end_position_on_the_genomic_accession=62037035, orientation='minus', exon_count=14, OMIM=603008, no_hdr0=''),
  1007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1007, CurrentID=0, Status='live', Symbol='CDH9', Aliases='', description='cadherin 9, type 2 (T1-cadherin)', other_designations='T1-cadherin', map_location='5p14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=26880600, end_position_on_the_genomic_accession=27121198, orientation='minus', exon_count=13, OMIM=609974, no_hdr0=''),
  1008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1008, CurrentID=0, Status='live', Symbol='CDH10', Aliases='', description='cadherin 10, type 2 (T2-cadherin)', other_designations='T2-cadherin|cadherin 10 type 2', map_location='5p14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24487100, end_position_on_the_genomic_accession=24645024, orientation='minus', exon_count=14, OMIM=604555, no_hdr0=''),
  1009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1009, CurrentID=0, Status='live', Symbol='CDH11', Aliases='CAD11, CDHOB, OB, OSF-4', description='cadherin 11, type 2, OB-cadherin (osteoblast)', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=64945226, end_position_on_the_genomic_accession=65126112, orientation='minus', exon_count=18, OMIM=600023, no_hdr0=''),
  1010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1010, CurrentID=0, Status='live', Symbol='CDH12', Aliases='CDHB', description='cadherin 12, type 2 (N-cadherin 2)', other_designations='Br-cadherin|N-cadherin 2|brain cadherin|brain-cadherin|cadherin, neural type, 2|neural type cadherin 2|neuronal cadherin 2', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21750864, end_position_on_the_genomic_accession=22853622, orientation='minus', exon_count=16, OMIM=600562, no_hdr0=''),
  1011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1011, CurrentID=0, Status='live', Symbol='CDH12P2', Aliases='C41-CAD, CDH12P', description='cadherin 12 (N-cadherin 2) pseudogene 2', other_designations='Br-cadherin pseudogene', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69985099, end_position_on_the_genomic_accession=69985600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1012, CurrentID=0, Status='live', Symbol='CDH13', Aliases='CDHH, P105', description='cadherin 13', other_designations='H-cadherin (heart)|T-cad|T-cadherin|cadherin 13, H-cadherin (heart)|heart cadherin', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82626794, end_position_on_the_genomic_accession=83796610, orientation='plus', exon_count=20, OMIM=601364, no_hdr0=''),
  1013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1013, CurrentID=0, Status='live', Symbol='CDH15', Aliases='CDH14, CDH3, CDHM, MCAD, MRD3', description='cadherin 15, type 1, M-cadherin (myotubule)', other_designations='cadherin-14|cadherin-3|muscle-cadherin', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89171755, end_position_on_the_genomic_accession=89195492, orientation='plus', exon_count=14, OMIM=114019, no_hdr0=''),
  1014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1014, CurrentID=0, Status='live', Symbol='CDH16', Aliases='', description='cadherin 16', other_designations='KSP-cadherin|cadherin 16, KSP-cadherin|kidney-specific cadherin', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66908122, end_position_on_the_genomic_accession=66918984, orientation='minus', exon_count=18, OMIM=603118, no_hdr0=''),
  1015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1015, CurrentID=0, Status='live', Symbol='CDH17', Aliases='CDH16, HPT-1, HPT1', description='cadherin 17, LI cadherin (liver-intestine)', other_designations='HPT-1 cadherin|LI cadherin|LI-cadherin|cadherin-16|human intestinal peptide-associated transporter HPT-1|human peptide transporter 1|intestinal peptide-associated transporter HPT-1|liver-intestine cadherin', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94127162, end_position_on_the_genomic_accession=94217303, orientation='minus', exon_count=19, OMIM=603017, no_hdr0=''),
  1016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1016, CurrentID=0, Status='live', Symbol='CDH18', Aliases='CDH14, CDH14L, CDH24', description='cadherin 18, type 2', other_designations='cadherin-14|cadherin-like 24|ey-cadherin', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=19472930, end_position_on_the_genomic_accession=20575873, orientation='minus', exon_count=21, OMIM=603019, no_hdr0=''),
  1017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1017, CurrentID=0, Status='live', Symbol='CDK2', Aliases='CDKN2, p33(CDK2)', description='cyclin-dependent kinase 2', other_designations='cdc2-related protein kinase|cell division protein kinase 2|p33 protein kinase', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55966769, end_position_on_the_genomic_accession=55972789, orientation='plus', exon_count=8, OMIM=116953, no_hdr0=''),
  1018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1018, CurrentID=0, Status='live', Symbol='CDK3', Aliases='', description='cyclin-dependent kinase 3', other_designations='cell division protein kinase 3', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76000906, end_position_on_the_genomic_accession=76005999, orientation='plus', exon_count=8, OMIM=123828, no_hdr0=''),
  1019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1019, CurrentID=0, Status='live', Symbol='CDK4', Aliases='CMM3, PSK-J3', description='cyclin-dependent kinase 4', other_designations='cell division protein kinase 4', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57747727, end_position_on_the_genomic_accession=57752447, orientation='minus', exon_count=8, OMIM=123829, no_hdr0=''),
  1020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1020, CurrentID=0, Status='live', Symbol='CDK5', Aliases='LIS7, PSSALRE', description='cyclin-dependent kinase 5', other_designations='TPKII catalytic subunit|cell division protein kinase 5|protein kinase CDK5 splicing|serine/threonine-protein kinase PSSALRE|tau protein kinase II catalytic subunit', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151053812, end_position_on_the_genomic_accession=151057965, orientation='minus', exon_count=12, OMIM=123831, no_hdr0=''),
  1021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1021, CurrentID=0, Status='live', Symbol='CDK6', Aliases='MCPH12, PLSTIRE', description='cyclin-dependent kinase 6', other_designations='cell division protein kinase 6|serine/threonine-protein kinase PLSTIRE', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92604921, end_position_on_the_genomic_accession=92836627, orientation='minus', exon_count=11, OMIM=603368, no_hdr0=''),
  1022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1022, CurrentID=0, Status='live', Symbol='CDK7', Aliases='CAK1, CDKN7, HCAK, MO15, STK1, p39MO15', description='cyclin-dependent kinase 7', other_designations='39 KDa protein kinase|CAK|CDK-activating kinase 1|TFIIH basal transcription factor complex kinase subunit|cell division protein kinase 7|cyclin-dependent kinase 7 (MO15 homolog, Xenopus laevis, cdk-activating kinase)|homolog of Xenopus MO15 Cdk-activating kinase|kinase subunit of CAK|p39 Mo15|protein kinase|serine/threonine kinase stk1|serine/threonine protein kinase 1|serine/threonine protein kinase MO15|serine/threonine-protein kinase 1', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69234795, end_position_on_the_genomic_accession=69277430, orientation='plus', exon_count=12, OMIM=601955, no_hdr0=''),
  100271103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271103, CurrentID=0, Status='live', Symbol='RPS25P7', Aliases='RPS25_5_649', description='ribosomal protein S25 pseudogene 7', other_designations='', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3913875, end_position_on_the_genomic_accession=3914321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1024, CurrentID=0, Status='live', Symbol='CDK8', Aliases='K35', description='cyclin-dependent kinase 8', other_designations='CDK8 protein kinase|cell division protein kinase 8|mediator complex subunit CDK8|mediator of RNA polymerase II transcription subunit CDK8|protein kinase K35', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26254125, end_position_on_the_genomic_accession=26405238, orientation='plus', exon_count=16, OMIM=603184, no_hdr0=''),
  1025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1025, CurrentID=0, Status='live', Symbol='CDK9', Aliases='C-2k, CDC2L4, CTK1, PITALRE, TAK', description='cyclin-dependent kinase 9', other_designations='CDC2-related kinase|cell division cycle 2-like protein kinase 4|cell division protein kinase 9|serine/threonine protein kinase PITALRE|tat-associated kinase complex catalytic subunit', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127786026, end_position_on_the_genomic_accession=127790773, orientation='plus', exon_count=7, OMIM=603251, no_hdr0=''),
  1026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1026, CurrentID=0, Status='live', Symbol='CDKN1A', Aliases='CAP20, CDKN1, CIP1, MDA-6, P21, SDI1, WAF1, p21CIP1', description='cyclin-dependent kinase inhibitor 1A (p21, Cip1)', other_designations='CDK-interacting protein 1|CDK-interaction protein 1|DNA synthesis inhibitor|melanoma differentiation associated protein 6|wild-type p53-activated fragment 1', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36676460, end_position_on_the_genomic_accession=36687339, orientation='plus', exon_count=6, OMIM=116899, no_hdr0=''),
  1027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1027, CurrentID=0, Status='live', Symbol='CDKN1B', Aliases='CDKN4, KIP1, MEN1B, MEN4, P27KIP1', description='cyclin-dependent kinase inhibitor 1B (p27, Kip1)', other_designations='', map_location='12p13.1-p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12717270, end_position_on_the_genomic_accession=12722383, orientation='plus', exon_count=3, OMIM=600778, no_hdr0=''),
  1028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1028, CurrentID=0, Status='live', Symbol='CDKN1C', Aliases='BWCR, BWS, KIP2, WBS, p57, p57Kip2', description='cyclin-dependent kinase inhibitor 1C (p57, Kip2)', other_designations='cyclin-dependent kinase inhibitor p57', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2883218, end_position_on_the_genomic_accession=2885816, orientation='minus', exon_count=4, OMIM=600856, no_hdr0=''),
  1029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1029, CurrentID=0, Status='live', Symbol='CDKN2A', Aliases='ARF, CDK4I, CDKN2, CMM2, INK4, INK4A, MLM, MTS-1, MTS1, P14, P14ARF, P16, P16-INK4A, P16INK4, P16INK4A, P19, P19ARF, TP16', description='cyclin-dependent kinase inhibitor 2A', other_designations='CDK4 inhibitor p16-INK4|alternative reading frame|cell cycle negative regulator beta|cyclin-dependent kinase 4 inhibitor A|cyclin-dependent kinase inhibitor 2A (melanoma, p16, inhibits CDK4)|multiple tumor suppressor 1', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21967752, end_position_on_the_genomic_accession=21995043, orientation='minus', exon_count=12, OMIM=600160, no_hdr0=''),
  1030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1030, CurrentID=0, Status='live', Symbol='CDKN2B', Aliases='CDK4I, INK4B, MTS2, P15, TP15, p15INK4b', description='cyclin-dependent kinase inhibitor 2B (p15, inhibits CDK4)', other_designations='CDK inhibitory protein|CDK4B inhibitor|MTS-2|cyclin-dependent kinases 4 and 6 binding protein|multiple tumor suppressor 2|p14-INK4b|p14_CDK inhibitor|p14_INK4B|p15 CDK inhibitor|p15-INK4b|p15_INK4B', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22002903, end_position_on_the_genomic_accession=22009313, orientation='minus', exon_count=2, OMIM=600431, no_hdr0=''),
  1031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1031, CurrentID=0, Status='live', Symbol='CDKN2C', Aliases='INK4C, p18, p18-INK4C', description='cyclin-dependent kinase inhibitor 2C (p18, inhibits CDK4)', other_designations='CDK6 inhibitor p18|cyclin-dependent inhibitor|cyclin-dependent kinase 6 inhibitor p18|p18-INK6', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50968695, end_position_on_the_genomic_accession=50974637, orientation='plus', exon_count=3, OMIM=603369, no_hdr0=''),
  1032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1032, CurrentID=0, Status='live', Symbol='CDKN2D', Aliases='INK4D, p19, p19-INK4D', description='cyclin-dependent kinase inhibitor 2D (p19, inhibits CDK4)', other_designations='CDK inhibitor p19INK4d|cell cycle inhibitor, Nur77 associating protein|cyclin-dependent kinase 4 inhibitor D p19|inhibitor of cyclin-dependent kinase 4d', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10566462, end_position_on_the_genomic_accession=10568979, orientation='minus', exon_count=2, OMIM=600927, no_hdr0=''),
  1033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1033, CurrentID=0, Status='live', Symbol='CDKN3', Aliases='CDI1, CIP2, KAP, KAP1', description='cyclin-dependent kinase inhibitor 3', other_designations='CDK2-associated dual specificity phosphatase|CDK2-associated dual-specificity phosphatase|Cdk-associated protein phosphatase|cyclin-dependent kinase interacting protein 2|cyclin-dependent kinase interactor 1|cyclin-dependent kinase-interacting protein 2|kinase associated phosphatase|kinase-associated phosphatase', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54396934, end_position_on_the_genomic_accession=54420218, orientation='plus', exon_count=9, OMIM=123832, no_hdr0=''),
  100271114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271114, CurrentID=0, Status='live', Symbol='RPS26P34', Aliases='RPS26_9_922', description='ribosomal protein S26 pseudogene 34', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81627247, end_position_on_the_genomic_accession=81627683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271115, CurrentID=0, Status='live', Symbol='RPS27P11', Aliases='RPS27_3_353', description='ribosomal protein S27 pseudogene 11', other_designations='', map_location='3p21.3-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27624145, end_position_on_the_genomic_accession=27624488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1036, CurrentID=0, Status='live', Symbol='CDO1', Aliases='', description='cysteine dioxygenase type 1', other_designations='CDO|CDO-I|cysteine dioxygenase, type I', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115804733, end_position_on_the_genomic_accession=115816954, orientation='minus', exon_count=6, OMIM=603943, no_hdr0=''),
  100271117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271117, CurrentID=0, Status='live', Symbol='RPS27P12', Aliases='RPS27_4_403', description='ribosomal protein S27 pseudogene 12', other_designations='ribosomal protein S27-like pseudogene', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129218036, end_position_on_the_genomic_accession=129218372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1038, CurrentID=0, Status='live', Symbol='CDR1', Aliases='CDR, CDR34, CDR62A', description='cerebellar degeneration related protein 1', other_designations='cerebellar degeneration-related protein 1, 34kDa', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140783260, end_position_on_the_genomic_accession=140784558, orientation='minus', exon_count=1, OMIM=302650, no_hdr0=''),
  1039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1039, CurrentID=0, Status='live', Symbol='CDR2', Aliases='CDR62, Yo', description='cerebellar degeneration related protein 2', other_designations='Yo paraneoplastic antigen|cerebellar degeneration-related protein 2, 62kDa|major Yo paraneoplastic antigen|paraneoplastic cerebellar degeneration-associated antigen', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22345936, end_position_on_the_genomic_accession=22374617, orientation='minus', exon_count=5, OMIM=117340, no_hdr0=''),
  1040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1040, CurrentID=0, Status='live', Symbol='CDS1', Aliases='CDS', description='CDP-diacylglycerol synthase 1', other_designations='CDP-DAG synthase 1|CDP-DG synthase 1|CDP-DG synthetase 1|CDP-diacylglycerol synthase (phosphatidate cytidylyltransferase) 1|CDP-diglyceride pyrophosphorylase 1|CDP-diglyceride synthase 1|CDP-diglyceride synthetase 1|CDS 1|CTP:phosphatidate cytidylyltransferase 1', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84582904, end_position_on_the_genomic_accession=84651340, orientation='plus', exon_count=13, OMIM=603548, no_hdr0=''),
  1041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1041, CurrentID=0, Status='live', Symbol='CDSN', Aliases='HTSS, HTSS1, HYPT2, PSS, PSS1, S', description='corneodesmosin', other_designations='differentiated keratinocyte S protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31115088, end_position_on_the_genomic_accession=31120475, orientation='minus', exon_count=2, OMIM=602593, no_hdr0=''),
  100271122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271122, CurrentID=0, Status='live', Symbol='RPS29P13', Aliases='RPS29_2_744', description='ribosomal protein S29 pseudogene 13', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117048670, end_position_on_the_genomic_accession=117048841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1043, CurrentID=0, Status='live', Symbol='CD52', Aliases='CDW52', description='CD52 molecule', other_designations='CD52 antigen (CAMPATH-1 antigen)|CDW52 antigen (CAMPATH-1 antigen)|HEL-S-171mP|cambridge pathology 1 antigen|epididymal secretory protein E5|epididymis secretory sperm binding protein Li 171mP|he5|human epididymis-specific protein 5', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26317920, end_position_on_the_genomic_accession=26320523, orientation='plus', exon_count=2, OMIM=114280, no_hdr0=''),
  1044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1044, CurrentID=0, Status='live', Symbol='CDX1', Aliases='', description='caudal type homeobox 1', other_designations='caudal type homeo box transcription factor 1|caudal type homeobox transcription factor 1|caudal-type homeobox protein 1|caudal-type homeobox protein CDX1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150166781, end_position_on_the_genomic_accession=150184558, orientation='plus', exon_count=3, OMIM=600746, no_hdr0=''),
  1045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1045, CurrentID=0, Status='live', Symbol='CDX2', Aliases='CDX-3/AS, CDX3, CDX2', description='caudal type homeobox 2', other_designations='caudal type homeo box transcription factor 2|caudal type homeobox transcription factor 2|caudal-type homeobox protein 2|homeobox protein miniCDX2', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27960057, end_position_on_the_genomic_accession=27969368, orientation='minus', exon_count=5, OMIM=600297, no_hdr0=''),
  1046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1046, CurrentID=0, Status='live', Symbol='CDX4', Aliases='', description='caudal type homeobox 4', other_designations='caudal type homeobox transcription factor 4|caudal-type homeobox protein 4', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73447254, end_position_on_the_genomic_accession=73454585, orientation='plus', exon_count=3, OMIM=300025, no_hdr0=''),
  1047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1047, CurrentID=0, Status='live', Symbol='CLGN', Aliases='', description='calmegin', other_designations='', map_location='4q28.3-q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140388453, end_position_on_the_genomic_accession=140427661, orientation='minus', exon_count=16, OMIM=601858, no_hdr0=''),
  1048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1048, CurrentID=0, Status='live', Symbol='CEACAM5', Aliases='CD66e, CEA', description='carcinoembryonic antigen-related cell adhesion molecule 5', other_designations='meconium antigen 100', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41708611, end_position_on_the_genomic_accession=41730433, orientation='plus', exon_count=10, OMIM=114890, no_hdr0=''),
  100271129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271129, CurrentID=0, Status='live', Symbol='RPS3AP9', Aliases='RPS3A_4_140', description='ribosomal protein S3a pseudogene 9', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=188694294, end_position_on_the_genomic_accession=188695155, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1050, CurrentID=0, Status='live', Symbol='CEBPA', Aliases='C/EBP-alpha, CEBP', description='CCAAT/enhancer binding protein (C/EBP), alpha', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33299934, end_position_on_the_genomic_accession=33302564, orientation='minus', exon_count=1, OMIM=116897, no_hdr0=''),
  1051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1051, CurrentID=0, Status='live', Symbol='CEBPB', Aliases='C/EBP-beta, IL6DBP, NF-IL6, TCF5', description='CCAAT/enhancer binding protein (C/EBP), beta', other_designations='interleukin 6-dependent DNA-binding protein|nuclear factor NF-IL6|nuclear factor of interleukin 6|transcription factor 5|transcription factor C/EBP beta', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50190583, end_position_on_the_genomic_accession=50192690, orientation='plus', exon_count=1, OMIM=189965, no_hdr0=''),
  1052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1052, CurrentID=0, Status='live', Symbol='CEBPD', Aliases='C/EBP-delta, CELF, CRP3, NF-IL6-beta', description='CCAAT/enhancer binding protein (C/EBP), delta', other_designations='c/EBP delta|nuclear factor NF-IL6-beta', map_location='8p11.2-p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47736914, end_position_on_the_genomic_accession=47738164, orientation='minus', exon_count=1, OMIM=116898, no_hdr0=''),
  1053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1053, CurrentID=0, Status='live', Symbol='CEBPE', Aliases='C/EBP-epsilon, CRP1', description='CCAAT/enhancer binding protein (C/EBP), epsilon', other_designations='c/EBP epsilon', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23117306, end_position_on_the_genomic_accession=23119985, orientation='minus', exon_count=3, OMIM=600749, no_hdr0=''),
  1054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1054, CurrentID=0, Status='live', Symbol='CEBPG', Aliases='GPE1BP, IG/EBP-1', description='CCAAT/enhancer binding protein (C/EBP), gamma', other_designations='c/EBP gamma', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33373669, end_position_on_the_genomic_accession=33382686, orientation='plus', exon_count=3, OMIM=138972, no_hdr0=''),
  1055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1055, CurrentID=0, Status='live', Symbol='CECR', Aliases='CES', description='cat eye syndrome chromosome region', other_designations='', map_location='22pter-q11', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=115470, no_hdr0=''),
  1056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1056, CurrentID=0, Status='live', Symbol='CEL', Aliases='BAL, BSDL, BSSLL, CEase, FAP, FAPP, LIPA, MODY8, CEL', description='carboxyl ester lipase', other_designations='bile salt-dependent lipase, oncofetal isoform|bucelipase|carboxyl ester hydrolase|carboxyl ester lipase (bile salt-stimulated lipase)|cholesterol esterase|fetoacinar pancreatic protein|lysophospholipase, pancreatic|sterol esterase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133061978, end_position_on_the_genomic_accession=133071863, orientation='plus', exon_count=11, OMIM=114840, no_hdr0=''),
  1057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1057, CurrentID=0, Status='live', Symbol='CELP', Aliases='CELL, cell1, cell2, cell3', description='carboxyl ester lipase pseudogene', other_designations='carboxyl ester lipase-like (bile salt-stimulated lipase-like)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133082539, end_position_on_the_genomic_accession=133087091, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  1058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1058, CurrentID=0, Status='live', Symbol='CENPA', Aliases='CENP-A, CenH3', description='centromere protein A', other_designations='centromere autoantigen A|centromere protein A, 17kDa|centromere-specific histone', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26786014, end_position_on_the_genomic_accession=26794589, orientation='plus', exon_count=5, OMIM=117139, no_hdr0=''),
  1059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1059, CurrentID=0, Status='live', Symbol='CENPB', Aliases='', description='centromere protein B', other_designations='CENP-B|centromere autoantigen B|centromere protein B, 80kDa', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3783851, end_position_on_the_genomic_accession=3786690, orientation='minus', exon_count=1, OMIM=117140, no_hdr0=''),
  1060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1060, CurrentID=0, Status='live', Symbol='CENPC', Aliases='CENP-C1, MIF2, hcp-4, CENPC', description='centromere protein C', other_designations='CENP-C 1|centromere autoantigen C|centromere autoantigen C1|centromere protein C 1|interphase centromere complex protein 7', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67472271, end_position_on_the_genomic_accession=67545538, orientation='minus', exon_count=21, OMIM=117141, no_hdr0=''),
  1061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1061, CurrentID=0, Status='live', Symbol='CENPCP1', Aliases='CENPC, CENPC1P1, CENPC2', description='centromere protein C pseudogene 1', other_designations='centromere protein C 1 pseudogene 1|centromere protein C2, 140kDa pseudogene', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89499908, end_position_on_the_genomic_accession=89503082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1062, CurrentID=0, Status='live', Symbol='CENPE', Aliases='CENP-E, KIF10, MCPH13, PPP1R61', description='centromere protein E', other_designations='Centromere autoantigen E (312kD)|centromere protein E, 312kDa|kinesin family member 10|kinesin-related protein CENPE|protein phosphatase 1, regulatory subunit 61', map_location='4q24-q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103105806, end_position_on_the_genomic_accession=103198409, orientation='minus', exon_count=49, OMIM=117143, no_hdr0=''),
  1063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1063, CurrentID=0, Status='live', Symbol='CENPF', Aliases='CENF, CILD31, PRO1779, hcp-1', description='centromere protein F', other_designations='AH antigen|CENP-F kinetochore protein|cell-cycle-dependent 350K nuclear protein|centromere protein F, 350/400kDa (mitosin)|centromere protein F, 350/400ka (mitosin)|kinetochore protein CENPF|mitosin', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214603179, end_position_on_the_genomic_accession=214664571, orientation='plus', exon_count=21, OMIM=600236, no_hdr0=''),
  100271144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271144, CurrentID=0, Status='live', Symbol='RPS27AP6', Aliases='RPS27A_1_97', description='ribosomal protein S27a pseudogene 6', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150881232, end_position_on_the_genomic_accession=150881634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271145, CurrentID=0, Status='live', Symbol='RPL7P53', Aliases='RPL7_30_1785', description='ribosomal protein L7 pseudogene 53', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73535503, end_position_on_the_genomic_accession=73536241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1066, CurrentID=0, Status='live', Symbol='CES1', Aliases='ACAT, CE-1, CEH, CES2, HMSE, HMSE1, PCE-1, REH, SES1, TGH, hCE-1', description='carboxylesterase 1', other_designations='acyl coenzyme A:cholesterol acyltransferase|acyl-coenzyme A:cholesterol acyltransferase|brain carboxylesterase hBr1|carboxylesterase 1 (monocyte/macrophage serine esterase 1)|carboxylesterase 2 (liver)|cholesteryl ester hydrolase|cocaine carboxylesterase|egasyn|human monocyte/macrophage serine esterase 1|methylumbelliferyl-acetate deacetylase 1|monocyte/macrophage serine esterase|retinyl ester hydrolase|serine esterase 1|triacylglycerol hydrolase', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55802851, end_position_on_the_genomic_accession=55833186, orientation='minus', exon_count=15, OMIM=114835, no_hdr0=''),
  100271147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271147, CurrentID=0, Status='live', Symbol='RPL7P55', Aliases='RPL7_32_1793', description='ribosomal protein L7 pseudogene 55', other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93573832, end_position_on_the_genomic_accession=93574645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1068, CurrentID=0, Status='live', Symbol='CETN1', Aliases='CEN1, CETN', description='centrin 1', other_designations='EF-hand protein|calcium binding protein|caltractin|centrin, EF-hand protein, 1', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=580343, end_position_on_the_genomic_accession=582020, orientation='plus', exon_count=1, OMIM=603187, no_hdr0=''),
  1069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1069, CurrentID=0, Status='live', Symbol='CETN2', Aliases='CALT, CEN2', description='centrin 2', other_designations='caltractin (20kD calcium-binding protein)|centrin, EF-hand protein, 2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152827327, end_position_on_the_genomic_accession=152830757, orientation='minus', exon_count=5, OMIM=300006, no_hdr0=''),
  1070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1070, CurrentID=0, Status='live', Symbol='CETN3', Aliases='CDC31, CEN3', description='centrin 3', other_designations='CDC31 yeast homolog|EF-hand superfamily member|centrin, EF-hand protein, 3 (CDC31 homolog, yeast)', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90393335, end_position_on_the_genomic_accession=90409786, orientation='minus', exon_count=6, OMIM=602907, no_hdr0=''),
  1071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1071, CurrentID=0, Status='live', Symbol='CETP', Aliases='BPIFF, HDLCQ10', description='cholesteryl ester transfer protein, plasma', other_designations='BPI fold containing family F|cholesteryl ester transfer protein plasma|lipid transfer protein I', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56961923, end_position_on_the_genomic_accession=56983844, orientation='plus', exon_count=17, OMIM=118470, no_hdr0=''),
  1072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1072, CurrentID=0, Status='live', Symbol='CFL1', Aliases='CFL, HEL-S-15, cofilin', description='cofilin 1 (non-muscle)', other_designations='18 kDa phosphoprotein|cofilin, non-muscle isoform|epididymis secretory protein Li 15|p18', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65854811, end_position_on_the_genomic_accession=65858333, orientation='minus', exon_count=4, OMIM=601442, no_hdr0=''),
  1073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1073, CurrentID=0, Status='live', Symbol='CFL2', Aliases='NEM7', description='cofilin 2 (muscle)', other_designations='cofilin, muscle isoform|nemaline myopathy type 7', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34710382, end_position_on_the_genomic_accession=34714823, orientation='minus', exon_count=6, OMIM=601443, no_hdr0=''),
  100271154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271154, CurrentID=0, Status='live', Symbol='RPL17P29', Aliases='RPL17_18_874', description='ribosomal protein L17 pseudogene 29', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11175010, end_position_on_the_genomic_accession=11175616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1075, CurrentID=0, Status='live', Symbol='CTSC', Aliases='CPPI, DPP-I, DPP1, DPPI, HMS, JP, JPD, PALS, PDON1, PLS', description='cathepsin C', other_designations='cathepsin J|dipeptidyl peptidase I|dipeptidyl transferase|dipeptidyl-peptidase I', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88293592, end_position_on_the_genomic_accession=88337773, orientation='minus', exon_count=9, OMIM=602365, no_hdr0=''),
  100271156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271156, CurrentID=0, Status='live', Symbol='RPL17P31', Aliases='RPL17_21_910', description='ribosomal protein L17 pseudogene 31', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67406469, end_position_on_the_genomic_accession=67407107, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271157, CurrentID=0, Status='live', Symbol='RPL21P30', Aliases='RPL21_10_219', description='ribosomal protein L21 pseudogene 30', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54029480, end_position_on_the_genomic_accession=54029867, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271158, CurrentID=0, Status='live', Symbol='RPL21P40', Aliases='RPL21_10_356', description='ribosomal protein L21 pseudogene 40', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32030154, end_position_on_the_genomic_accession=32030711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271159, CurrentID=0, Status='live', Symbol='RPL21P41', Aliases='RPL21_11_380', description='ribosomal protein L21 pseudogene 41', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66637149, end_position_on_the_genomic_accession=66637542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1080, CurrentID=0, Status='live', Symbol='CFTR', Aliases='ABC35, ABCC7, CF/MRP, MRP7, TNR-CFTR, dJ760C5.1, CFTR', description='cystic fibrosis transmembrane conductance regulator', other_designations='cAMP-dependent chloride channel|channel conductance-controlling ATPase|cystic fibrosis transmembrane conductance regulator (ATP-binding cassette sub-family C, member 7)', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117478340, end_position_on_the_genomic_accession=117668665, orientation='plus', exon_count=33, OMIM=602421, no_hdr0=''),
  1081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1081, CurrentID=0, Status='live', Symbol='CGA', Aliases='CG-ALPHA, FSHA, GPHA1, GPHa, HCG, LHA, TSHA', description='glycoprotein hormones, alpha polypeptide', other_designations='FSH-alpha|LSH-alpha|TSH-alpha|anterior pituitary glycoprotein hormones common subunit alpha|choriogonadotropin alpha chain|chorionic gonadotrophin subunit alpha|chorionic gonadotropin, alpha polypeptide|follicle-stimulating hormone alpha chain|follicle-stimulating hormone alpha subunit|follitropin alpha chain|luteinizing hormone alpha chain|lutropin alpha chain|thyroid-stimulating hormone alpha chain|thyrotropin alpha chain', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87085498, end_position_on_the_genomic_accession=87095147, orientation='minus', exon_count=5, OMIM=118850, no_hdr0=''),
  1082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1082, CurrentID=0, Status='live', Symbol='CGB', Aliases='CGB35, CGB7, CGB8, hCGB, CGB', description='chorionic gonadotropin, beta polypeptide', other_designations='CG-beta|chorionic gonadotrophin chain beta|chorionic gonadotropin beta 3 subunit|chorionic gonadotropin beta chain|chorionic gonadotropin beta subunit', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49022869, end_position_on_the_genomic_accession=49024375, orientation='minus', exon_count=3, OMIM=118860, no_hdr0=''),
  1083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1083, CurrentID=0, Status='live', Symbol='CGF1', Aliases='', description='cognitive function 1, social', other_designations='', map_location='Xp11.22-qter', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300082, no_hdr0=''),
  1084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1084, CurrentID=0, Status='live', Symbol='CEACAM3', Aliases='CD66D, CEA, CGM1, W264, W282', description='carcinoembryonic antigen-related cell adhesion molecule 3', other_designations='CD66d antigen|carcinoembryonic antigen CGM1|carcinoembryonic antigen gene family member 1|nonspecific cross-reacting antigen', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41796590, end_position_on_the_genomic_accession=41811553, orientation='plus', exon_count=8, OMIM=609142, no_hdr0=''),
  1085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1085, CurrentID=0, Status='live', Symbol='CEACAMP5', Aliases='CEACAM26P, CGM12', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 5', other_designations='carcinoembryonic antigen gene family member 12|carcinoembryonic antigen-related cell adhesion molecule 26, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42654152, end_position_on_the_genomic_accession=42663017, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132158, CurrentID=0, Status='live', Symbol='GLYCTK', Aliases='HBEBP2, HBEBP4, HBeAgBP4A', description='glycerate kinase', other_designations='HBeAg binding protein 4|HBeAg-binding protein 2', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52287820, end_position_on_the_genomic_accession=52295256, orientation='plus', exon_count=5, OMIM=610516, no_hdr0=''),
  1087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1087, CurrentID=0, Status='live', Symbol='CEACAM7', Aliases='CGM2', description='carcinoembryonic antigen-related cell adhesion molecule 7', other_designations='carcinoembryonic antigen CGM2|carcinoembryonic antigen gene family member 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41673307, end_position_on_the_genomic_accession=41688277, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  1088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1088, CurrentID=0, Status='live', Symbol='CEACAM8', Aliases='CD66b, CD67, CGM6, NCA-95', description='carcinoembryonic antigen-related cell adhesion molecule 8', other_designations='CD67 antigen|carcinoembryonic antigen CGM6|carcinoembryonic antigen gene family member 6|non-specific cross-reacting antigen NCA-95', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42580243, end_position_on_the_genomic_accession=42594939, orientation='minus', exon_count=8, OMIM=615747, no_hdr0=''),
  1089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1089, CurrentID=0, Status='live', Symbol='CEACAM4', Aliases='CGM7, CGM7_HUMAN, NCA', description='carcinoembryonic antigen-related cell adhesion molecule 4', other_designations='Nonspecific cross-reacting antigen (NCA)|carcinoembryonic antigen CGM7|carcinoembryonic antigen gene family member 7|carcinoembryonic antigen-related cell adhesion molecule 4-sv1|carcinoembryonic antigen-related cell adhesion molecule 4-sv2|non-specific cross-reacting antigen W236|nonspecific cross-reacting antigen W236', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41616986, end_position_on_the_genomic_accession=41627084, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  1090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1090, CurrentID=0, Status='live', Symbol='CEACAMP1', Aliases='CEACAM22P, CGM8', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 1', other_designations='carcinoembryonic antigen gene family member 8 (pseudogene)|carcinoembryonic antigen-related cell adhesion molecule 22, pseudogene|pregnancy specific beta-1-glycoprotein 9 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42620173, end_position_on_the_genomic_accession=42633136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1091, CurrentID=0, Status='live', Symbol='CEACAMP2', Aliases='CEACAM23P, CGM9', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 2', other_designations='carcinoembryonic antigen gene family member 9 (pseudogene)|carcinoembryonic antigen-related cell adhesion molecule 23, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42560467, end_position_on_the_genomic_accession=42560755, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1092, CurrentID=0, Status='live', Symbol='CEACAMP3', Aliases='CEACAM24P, CGM10', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 3', other_designations='carcinoembryonic antigen gene family member 10 (pseudogene)|carcinoembryonic antigen-related cell adhesion molecule 24, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41601256, end_position_on_the_genomic_accession=41601503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1093, CurrentID=0, Status='live', Symbol='CEACAMP4', Aliases='CEACAM25P, CGM11', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 4', other_designations='carcinoembryonic antigen gene family member 11 (pseudogene)|carcinoembryonic antigen-related cell adhesion molecule 25, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43303202, end_position_on_the_genomic_accession=43303498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1094, CurrentID=0, Status='live', Symbol='CEACAMP6', Aliases='CEACAM27P, CGM13', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 6', other_designations='carcinoembryonic antigen gene family member 13|carcinoembryonic antigen-related cell adhesion molecule 27, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42785711, end_position_on_the_genomic_accession=42786621, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1095, CurrentID=0, Status='live', Symbol='CEACAMP7', Aliases='CEACAM28P, CGM14', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 7', other_designations='carcinoembryonic antigen gene family member 14|carcinoembryonic antigen-related cell adhesion molecule 28, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42949503, end_position_on_the_genomic_accession=42950415, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1096, CurrentID=0, Status='live', Symbol='CEACAMP8', Aliases='CEACAM29P, CGM15', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 8', other_designations='carcinoembryonic antigen gene family member 15|carcinoembryonic antigen-related cell adhesion molecule 29, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43040374, end_position_on_the_genomic_accession=43041274, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1097, CurrentID=0, Status='live', Symbol='CEACAMP9', Aliases='CEACAM30P, CGM16', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 9', other_designations='carcinoembryonic antigen gene family member 16|carcinoembryonic antigen-related cell adhesion molecule 30, pseudogene|pregnancy specific beta-1-glycoprotein 9 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43096449, end_position_on_the_genomic_accession=43097360, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1098, CurrentID=0, Status='live', Symbol='CEACAMP10', Aliases='CEACAM31P, CGM17', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 10', other_designations='carcinoembryonic antigen gene family member 17|carcinoembryonic antigen-related cell adhesion molecule 31, pseudogene|pregnancy specific beta-1-glycoprotein pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43219988, end_position_on_the_genomic_accession=43220899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271179, CurrentID=0, Status='live', Symbol='RPL21P66', Aliases='RPL21_26_756', description='ribosomal protein L21 pseudogene 66', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132518810, end_position_on_the_genomic_accession=132519331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271180, CurrentID=0, Status='live', Symbol='RPL21P67', Aliases='RPL21_27_754', description='ribosomal protein L21 pseudogene 67', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131469019, end_position_on_the_genomic_accession=131469568, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1101, CurrentID=0, Status='live', Symbol='CHAD', Aliases='SLRR4A', description='chondroadherin', other_designations='cartilage leucine-rich protein|chondroadherin proteoglycan', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50464492, end_position_on_the_genomic_accession=50469881, orientation='minus', exon_count=4, OMIM=602178, no_hdr0=''),
  1102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1102, CurrentID=0, Status='live', Symbol='RCBTB2', Aliases='CHC1L, RLG', description='regulator of chromosome condensation (RCC1) and BTB (POZ) domain containing protein 2', other_designations='RCC1-like G exchanging factor RLG', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48488963, end_position_on_the_genomic_accession=48533256, orientation='minus', exon_count=16, OMIM=603524, no_hdr0=''),
  1103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1103, CurrentID=0, Status='live', Symbol='CHAT', Aliases='CHOACTASE, CMS1A, CMS1A2, CMS6', description='choline O-acetyltransferase', other_designations='acetyl CoA:choline O-acetyltransferase|choline acetylase', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49609095, end_position_on_the_genomic_accession=49665104, orientation='plus', exon_count=18, OMIM=118490, no_hdr0=''),
  1104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1104, CurrentID=0, Status='live', Symbol='RCC1', Aliases='CHC1-I, SNHG3-RCC1, RCC1', description='regulator of chromosome condensation 1', other_designations='SNHG3-RCC1 readthrough|cell cycle regulatory protein|guanine nucleotide-releasing protein', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28505943, end_position_on_the_genomic_accession=28539196, orientation='plus', exon_count=15, OMIM=179710, no_hdr0=''),
  1105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1105, CurrentID=0, Status='live', Symbol='CHD1', Aliases='', description='chromodomain helicase DNA binding protein 1', other_designations='ATP-dependent helicase CHD1|CHD-1', map_location='5q15-q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98855204, end_position_on_the_genomic_accession=98928623, orientation='minus', exon_count=37, OMIM=602118, no_hdr0=''),
  1106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1106, CurrentID=0, Status='live', Symbol='CHD2', Aliases='EEOC', description='chromodomain helicase DNA binding protein 2', other_designations='ATP-dependent helicase CHD2|CHD-2', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92900321, end_position_on_the_genomic_accession=93028007, orientation='plus', exon_count=39, OMIM=602119, no_hdr0=''),
  1107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1107, CurrentID=0, Status='live', Symbol='CHD3', Aliases='Mi-2a, Mi2-ALPHA, ZFH', description='chromodomain helicase DNA binding protein 3', other_designations='ATP-dependent helicase CHD3|CHD-3|hZFH|mi-2 autoantigen 240 kDa protein|zinc finger helicase|zinc-finger helicase (Snf2-like)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7884776, end_position_on_the_genomic_accession=7912757, orientation='plus', exon_count=45, OMIM=602120, no_hdr0=''),
  1108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1108, CurrentID=0, Status='live', Symbol='CHD4', Aliases='CHD-4, Mi-2b, Mi2-BETA', description='chromodomain helicase DNA binding protein 4', other_designations='ATP-dependent helicase CHD4|Mi-2 autoantigen 218 kDa protein', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6570082, end_position_on_the_genomic_accession=6607439, orientation='minus', exon_count=40, OMIM=603277, no_hdr0=''),
  1109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1109, CurrentID=0, Status='live', Symbol='AKR1C4', Aliases='3-alpha-HSD, C11, CDR, CHDR, DD-4, DD4, HAKRA', description='aldo-keto reductase family 1, member C4', other_designations='3-alpha-HSD1|chlordecone reductase; 3-alpha hydroxysteroid dehydrogenase, type I; dihydrodiol dehydrogenase 4|dihydrodiol dehydrogenase isozyme DD4|type I 3-alpha-hydroxysteroid dehydrogenase', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5196809, end_position_on_the_genomic_accession=5220633, orientation='plus', exon_count=9, OMIM=600451, no_hdr0=''),
  100271190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271190, CurrentID=0, Status='live', Symbol='RPL21P86', Aliases='RPL21_34_998', description='ribosomal protein L21 pseudogene 86', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92676897, end_position_on_the_genomic_accession=92677447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1111, CurrentID=0, Status='live', Symbol='CHEK1', Aliases='CHK1', description='checkpoint kinase 1', other_designations='CHK1 checkpoint homolog|Checkpoint, S. pombe, homolog of, 1|Chk1-S|cell cycle checkpoint kinase|checkpoint kinase-1', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125624920, end_position_on_the_genomic_accession=125676259, orientation='plus', exon_count=16, OMIM=603078, no_hdr0=''),
  1112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1112, CurrentID=0, Status='live', Symbol='FOXN3', Aliases='C14orf116, CHES1, PRO1635', description='forkhead box N3', other_designations='checkpoint suppressor 1', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89156172, end_position_on_the_genomic_accession=89619150, orientation='minus', exon_count=8, OMIM=602628, no_hdr0=''),
  1113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1113, CurrentID=0, Status='live', Symbol='CHGA', Aliases='CGA', description='chromogranin A', other_designations='SP-I|betagranin (N-terminal fragment of chromogranin A)|catestatin|chromofungin|parathyroid secretory protein 1|pituitary secretory protein I', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92922589, end_position_on_the_genomic_accession=92935293, orientation='plus', exon_count=9, OMIM=118910, no_hdr0=''),
  1114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1114, CurrentID=0, Status='live', Symbol='CHGB', Aliases='SCG1', description='chromogranin B', other_designations='cgB|chromogranin-B|secretogranin B|secretogranin I|sgI', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5911328, end_position_on_the_genomic_accession=5925361, orientation='plus', exon_count=5, OMIM=118920, no_hdr0=''),
  1116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1116, CurrentID=0, Status='live', Symbol='CHI3L1', Aliases='ASRT7, CGP-39, GP-39, GP39, HC-gp39, HCGP-3P, YKL-40, YKL40, YYL-40, hCGP-39', description='chitinase 3 like 1', other_designations='39 kDa synovial protein|cartilage glycoprotein 39|chitinase 3-like 1 (cartilage glycoprotein-39)', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203178931, end_position_on_the_genomic_accession=203186960, orientation='minus', exon_count=10, OMIM=601525, no_hdr0=''),
  1117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1117, CurrentID=0, Status='live', Symbol='CHI3L2', Aliases='CHIL2, YKL-39, YKL39', description='chitinase 3 like 2', other_designations='chitinase 3-like 2|chondrocyte protein 39', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111227659, end_position_on_the_genomic_accession=111243440, orientation='plus', exon_count=11, OMIM=601526, no_hdr0=''),
  1118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1118, CurrentID=0, Status='live', Symbol='CHIT1', Aliases='CHI3, CHIT, CHITD', description='chitinase 1', other_designations='chitinase 1 (chitotriosidase)|plasma methylumbelliferyl tetra-N-acetylchitotetraoside hydrolase', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203212823, end_position_on_the_genomic_accession=203230099, orientation='minus', exon_count=15, OMIM=600031, no_hdr0=''),
  1119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1119, CurrentID=0, Status='live', Symbol='CHKA', Aliases='CHK, CK, CKI, EK', description='choline kinase alpha', other_designations='CHETK-alpha|ethanolamine kinase', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68052859, end_position_on_the_genomic_accession=68121391, orientation='minus', exon_count=17, OMIM=118491, no_hdr0=''),
  1120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1120, CurrentID=0, Status='live', Symbol='CHKB', Aliases='CHETK, CHKL, CK, CKB, CKEKB, EK, EKB, MDCMC', description='choline kinase beta', other_designations='choline kinase-like protein|ethanolamine kinase beta', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50578958, end_position_on_the_genomic_accession=50582999, orientation='minus', exon_count=11, OMIM=612395, no_hdr0=''),
  1121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1121, CurrentID=0, Status='live', Symbol='CHM', Aliases='DXS540, GGTA, HSD-32, REP-1, TCD', description='choroideremia (Rab escort protein 1)', other_designations='', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85861180, end_position_on_the_genomic_accession=86047562, orientation='minus', exon_count=18, OMIM=300390, no_hdr0=''),
  1122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1122, CurrentID=0, Status='live', Symbol='CHML', Aliases='REP2', description='choroideremia-like (Rab escort protein 2)', other_designations='choroideraemia-like protein|choroideremia-like protein', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241628864, end_position_on_the_genomic_accession=241635930, orientation='minus', exon_count=1, OMIM=118825, no_hdr0=''),
  1123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1123, CurrentID=0, Status='live', Symbol='CHN1', Aliases='ARHGAP2, CHN, DURS2, NC, RHOGAP2', description='chimerin 1', other_designations='A-chimaerin|Rho GTPase-activating protein 2|a2-chimaerin|alpha-chimerin|chimaerin 1|chimerin (chimaerin) 1|n-chimerin', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174799314, end_position_on_the_genomic_accession=175005379, orientation='minus', exon_count=14, OMIM=118423, no_hdr0=''),
  1124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1124, CurrentID=0, Status='live', Symbol='CHN2', Aliases='ARHGAP3, BCH-3, RHOGAP3, CHN2', description='chimerin 2', other_designations='beta-chimerin|beta3-chimaerin|chimerin (chimaerin) 2|chimerin, beta-2|chimerin, testis-specific|rho-GTPase-activating protein 3', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29146547, end_position_on_the_genomic_accession=29514335, orientation='plus', exon_count=21, OMIM=602857, no_hdr0=''),
  1125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1125, CurrentID=0, Status='live', Symbol='CHR', Aliases='', description='chromate resistance; sulfate transport', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=118840, no_hdr0=''),
  100271206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271206, CurrentID=0, Status='live', Symbol='RPL5P26', Aliases='RPL5_13_1059', description='ribosomal protein L5 pseudogene 26', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69778890, end_position_on_the_genomic_accession=69779896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271207, CurrentID=0, Status='live', Symbol='RPL5P32', Aliases='RPL5_17_1422', description='ribosomal protein L5 pseudogene 32', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27895660, end_position_on_the_genomic_accession=27896640, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1128, CurrentID=0, Status='live', Symbol='CHRM1', Aliases='HM1, M1, M1R', description='cholinergic receptor, muscarinic 1', other_designations='acetylcholine receptor, muscarinic 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62908679, end_position_on_the_genomic_accession=62921861, orientation='minus', exon_count=3, OMIM=118510, no_hdr0=''),
  1129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1129, CurrentID=0, Status='live', Symbol='CHRM2', Aliases='HM2', description='cholinergic receptor, muscarinic 2', other_designations='7TM receptor|acetylcholine receptor, muscarinic 2|muscarinic M2 receptor', map_location='7q31-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136868652, end_position_on_the_genomic_accession=137020255, orientation='plus', exon_count=9, OMIM=118493, no_hdr0=''),
  1130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1130, CurrentID=0, Status='live', Symbol='LYST', Aliases='CHS, CHS1', description='lysosomal trafficking regulator', other_designations='Chediak-Higashi syndrome 1|beige homolog', map_location='1q42.1-q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235661031, end_position_on_the_genomic_accession=235883708, orientation='minus', exon_count=61, OMIM=606897, no_hdr0=''),
  1131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1131, CurrentID=0, Status='live', Symbol='CHRM3', Aliases='EGBRS, HM3, PBS', description='cholinergic receptor, muscarinic 3', other_designations='acetylcholine receptor, muscarinic 3|m3 muscarinic receptor', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239386565, end_position_on_the_genomic_accession=239911462, orientation='plus', exon_count=12, OMIM=118494, no_hdr0=''),
  1132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1132, CurrentID=0, Status='live', Symbol='CHRM4', Aliases='HM4, M4R', description='cholinergic receptor, muscarinic 4', other_designations='acetylcholine receptor, muscarinic 4', map_location='11p12-p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46384792, end_position_on_the_genomic_accession=46386608, orientation='minus', exon_count=1, OMIM=118495, no_hdr0=''),
  1133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1133, CurrentID=0, Status='live', Symbol='CHRM5', Aliases='HM5', description='cholinergic receptor, muscarinic 5', other_designations='acetylcholine receptor, muscarinic 5', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33968245, end_position_on_the_genomic_accession=34065094, orientation='plus', exon_count=4, OMIM=118496, no_hdr0=''),
  1134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1134, CurrentID=0, Status='live', Symbol='CHRNA1', Aliases='ACHRA, ACHRD, CHRNA, CMS1A, CMS1B, CMS2A, FCCMS, SCCMS', description='cholinergic receptor, nicotinic alpha 1', other_designations='acetylcholine receptor, nicotinic, alpha 1 (muscle)|cholinergic receptor, nicotinic, alpha 1 (muscle)|cholinergic receptor, nicotinic, alpha polypeptide 1 (muscle)|muscle nicotinic acetylcholine receptor|nicotinic acetylcholine receptor alpha subunit|nicotinic cholinergic receptor alpha 1', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174747592, end_position_on_the_genomic_accession=174764472, orientation='minus', exon_count=10, OMIM=100690, no_hdr0=''),
  1135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1135, CurrentID=0, Status='live', Symbol='CHRNA2', Aliases='', description='cholinergic receptor, nicotinic alpha 2', other_designations='acetylcholine receptor, nicotinic, alpha 2 (neuronal)|cholinergic receptor, nicotinic, alpha 2 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 2 (neuronal)', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27459761, end_position_on_the_genomic_accession=27479296, orientation='minus', exon_count=8, OMIM=118502, no_hdr0=''),
  1136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1136, CurrentID=0, Status='live', Symbol='CHRNA3', Aliases='LNCR2, NACHRA3, PAOD2', description='cholinergic receptor, nicotinic alpha 3', other_designations='cholinergic receptor, nicotinic, alpha 3 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 3|neuronal nicotinic acetylcholine receptor, alpha3 subunit', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78593052, end_position_on_the_genomic_accession=78621295, orientation='minus', exon_count=10, OMIM=118503, no_hdr0=''),
  1137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1137, CurrentID=0, Status='live', Symbol='CHRNA4', Aliases='BFNC, EBN, EBN1, NACHR, NACHRA4, NACRA4', description='cholinergic receptor, nicotinic alpha 4', other_designations='cholinergic receptor, nicotinic, alpha 4 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 4|neuronal nicotinic acetylcholine receptor alpha-4 subunit', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63343310, end_position_on_the_genomic_accession=63361396, orientation='minus', exon_count=7, OMIM=118504, no_hdr0=''),
  1138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1138, CurrentID=0, Status='live', Symbol='CHRNA5', Aliases='LNCR2', description='cholinergic receptor, nicotinic alpha 5', other_designations='Cholinergic receptor, neuronal nicotinic, alpha polypeptide-5|acetylcholine receptor, nicotinic, alpha 5 (neuronal)|cholinergic receptor, nicotinic, alpha 5 (neuronal)|neuronal nicotinic acetylcholine receptor, alpha5 subunit', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78565520, end_position_on_the_genomic_accession=78595269, orientation='plus', exon_count=6, OMIM=118505, no_hdr0=''),
  1139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1139, CurrentID=0, Status='live', Symbol='CHRNA7', Aliases='CHRNA7-2, NACHRA7', description='cholinergic receptor, nicotinic alpha 7', other_designations='a7 nicotinic acetylcholine receptor|alpha 7 neuronal nicotinic acetylcholine receptor|alpha-7 nicotinic cholinergic receptor subunit|cholinergic receptor, nicotinic, alpha 7 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 7|neuronal acetylcholine receptor protein, alpha-7 chain', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32030483, end_position_on_the_genomic_accession=32172521, orientation='plus', exon_count=13, OMIM=118511, no_hdr0=''),
  1140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1140, CurrentID=0, Status='live', Symbol='CHRNB1', Aliases='ACHRB, CHRNB, CMS1D, CMS2A, CMS2C, SCCMS', description='cholinergic receptor, nicotinic beta 1', other_designations='acetylcholine receptor, nicotinic, beta 1 (muscle)|cholinergic receptor, nicotinic, beta 1 (muscle)|cholinergic receptor, nicotinic, beta polypeptide 1 (muscle)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7445087, end_position_on_the_genomic_accession=7457613, orientation='plus', exon_count=11, OMIM=100710, no_hdr0=''),
  1141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1141, CurrentID=0, Status='live', Symbol='CHRNB2', Aliases='EFNL3, nAChRB2', description='cholinergic receptor, nicotinic beta 2', other_designations='acetylcholine receptor, nicotinic, beta 2 (neuronal)|beta2 human neuronal nicotinic acetylcholine receptor|cholinergic receptor, nicotinic, beta 2 (neuronal)|cholinergic receptor, nicotinic, beta polypeptide 2 (neuronal)|neuronal nicotinic acetylcholine receptor beta 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154567781, end_position_on_the_genomic_accession=154579878, orientation='plus', exon_count=6, OMIM=118507, no_hdr0=''),
  1142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1142, CurrentID=0, Status='live', Symbol='CHRNB3', Aliases='', description='cholinergic receptor, nicotinic beta 3', other_designations='acetylcholine receptor, neuronal nicotinic, beta-3 subunit|acetylcholine receptor, nicotinic, beta 3 (neuronal)|cholinergic receptor, nicotinic, beta 3 (neuronal)|cholinergic receptor, nicotinic, beta polypeptide 3', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42697419, end_position_on_the_genomic_accession=42737407, orientation='plus', exon_count=7, OMIM=118508, no_hdr0=''),
  1143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1143, CurrentID=0, Status='live', Symbol='CHRNB4', Aliases='', description='cholinergic receptor, nicotinic beta 4', other_designations='acetylcholine receptor, nicotinic, beta 4 (neuronal)|cholinergic receptor, nicotinic, beta 4 (neuronal)|cholinergic receptor, nicotinic, beta polypeptide 4|neuronal nicotinic receptor beta 4 subunit', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78623297, end_position_on_the_genomic_accession=78655586, orientation='minus', exon_count=18, OMIM=118509, no_hdr0=''),
  1144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1144, CurrentID=0, Status='live', Symbol='CHRND', Aliases='ACHRD, CMS2A, CMS3A, CMS3B, CMS3C, FCCMS, SCCMS', description='cholinergic receptor, nicotinic delta', other_designations='acetylcholine receptor, nicotinic, delta (muscle)|cholinergic receptor, nicotinic, delta (muscle)|cholinergic receptor, nicotinic, delta polypeptide', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232526160, end_position_on_the_genomic_accession=232536665, orientation='plus', exon_count=12, OMIM=100720, no_hdr0=''),
  1145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1145, CurrentID=0, Status='live', Symbol='CHRNE', Aliases='ACHRE, CMS1D, CMS1E, CMS2A, CMS4A, CMS4B, CMS4C, FCCMS, SCCMS', description='cholinergic receptor, nicotinic epsilon', other_designations='AchR epsilon subunit|acetylcholine receptor, nicotinic, epsilon (muscle)|cholinergic receptor, nicotinic, epsilon (muscle)|cholinergic receptor, nicotinic, epsilon polypeptide', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4897769, end_position_on_the_genomic_accession=4903128, orientation='minus', exon_count=12, OMIM=100725, no_hdr0=''),
  1146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1146, CurrentID=0, Status='live', Symbol='CHRNG', Aliases='ACHRG', description='cholinergic receptor, nicotinic gamma', other_designations='acetylcholine receptor, muscle, gamma subunit|acetylcholine receptor, nicotinic, gamma (muscle)|cholinergic receptor, nicotinic, gamma (muscle)|cholinergic receptor, nicotinic, gamma polypeptide', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232539727, end_position_on_the_genomic_accession=232546328, orientation='plus', exon_count=12, OMIM=100730, no_hdr0=''),
  1147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1147, CurrentID=0, Status='live', Symbol='CHUK', Aliases='IKBKA, IKK-alpha, IKK1, IKKA, NFKBIKA, TCF16', description='conserved helix-loop-helix ubiquitous kinase', other_designations='I-kappa-B kinase 1|I-kappa-B kinase-alpha|IKK-a kinase|IkB kinase alpha subunit|Nuclear factor NFkappaB inhibitor kinase alpha|TCF-16|transcription factor 16', map_location='10q24-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100186113, end_position_on_the_genomic_accession=100229610, orientation='minus', exon_count=23, OMIM=600664, no_hdr0=''),
  100271228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271228, CurrentID=0, Status='live', Symbol='RPL9P31', Aliases='RPL9_12_1606', description='ribosomal protein L9 pseudogene 31', other_designations='', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58631232, end_position_on_the_genomic_accession=58631815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1149, CurrentID=0, Status='live', Symbol='CIDEA', Aliases='CIDE-A', description='cell death-inducing DFFA-like effector a', other_designations='', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12254319, end_position_on_the_genomic_accession=12277595, orientation='plus', exon_count=6, OMIM=604440, no_hdr0=''),
  100271230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271230, CurrentID=0, Status='live', Symbol='RPS17P17', Aliases='RPS17_8_1818', description='ribosomal protein S17 pseudogene 17', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140470866, end_position_on_the_genomic_accession=140471274, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271231, CurrentID=0, Status='live', Symbol='RPS20P18', Aliases='RPS20_10_738', description='ribosomal protein S20 pseudogene 18', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110700440, end_position_on_the_genomic_accession=110700923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1152, CurrentID=0, Status='live', Symbol='CKB', Aliases='B-CK, BCKB, HEL-211, HEL-S-29, CKB', description='creatine kinase, brain', other_designations='creatine kinase B chain|creatine kinase brain|creatine kinase-B|epididymis luminal protein 211|epididymis secretory protein Li 29', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103519658, end_position_on_the_genomic_accession=103522859, orientation='minus', exon_count=8, OMIM=123280, no_hdr0=''),
  1153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1153, CurrentID=0, Status='live', Symbol='CIRBP', Aliases='CIRP', description='cold inducible RNA binding protein', other_designations='A18 hnRNP|cold inducible RNA-binding protein|glycine-rich RNA binding protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1269266, end_position_on_the_genomic_accession=1274810, orientation='plus', exon_count=9, OMIM=602649, no_hdr0=''),
  1154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1154, CurrentID=0, Status='live', Symbol='CISH', Aliases='BACTS2, CIS, CIS-1, G18, SOCS', description='cytokine inducible SH2-containing protein', other_designations='cytokine-inducible inhibitor of signaling type 1B|suppressor of cytokine signaling', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50606454, end_position_on_the_genomic_accession=50611831, orientation='minus', exon_count=4, OMIM=602441, no_hdr0=''),
  1155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1155, CurrentID=0, Status='live', Symbol='TBCB', Aliases='CG22, CKAP1, CKAPI', description='tubulin folding cofactor B', other_designations='cytoskeleton associated protein 1|cytoskeleton-associated protein CKAPI|tubulin-specific chaperone B', map_location='19q13.11-q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36114986, end_position_on_the_genomic_accession=36125948, orientation='plus', exon_count=7, OMIM=601303, no_hdr0=''),
  1156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1156, CurrentID=0, Status='live', Symbol='CKBE', Aliases='', description='creatine kinase, ectopic expression', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=123270, no_hdr0=''),
  1157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1157, CurrentID=0, Status='live', Symbol='CKBP1', Aliases='', description='creatine kinase B pseudogene 1', other_designations='brain creatine kinase pseudogene', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46842458, end_position_on_the_genomic_accession=46844619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1158, CurrentID=0, Status='live', Symbol='CKM', Aliases='CKMM, M-CK', description='creatine kinase, muscle', other_designations='creatine kinase M chain', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45306413, end_position_on_the_genomic_accession=45322977, orientation='minus', exon_count=8, OMIM=123310, no_hdr0=''),
  1159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1159, CurrentID=0, Status='live', Symbol='CKMT1B', Aliases='CKMT, CKMT1, UMTCK', description='creatine kinase, mitochondrial 1B', other_designations='U-MtCK|acidic-type mitochondrial creatine kinase|creatine kinase, mitochondrial 1 (ubiquitous)|mia-CK|ubiquitous mitochondrial creatine kinase', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43592862, end_position_on_the_genomic_accession=43599406, orientation='plus', exon_count=10, OMIM=123290, no_hdr0=''),
  1160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1160, CurrentID=0, Status='live', Symbol='CKMT2', Aliases='SMTCK', description='creatine kinase, mitochondrial 2', other_designations='S-MtCK|basic-type mitochondrial creatine kinase|creatine kinase, mitochondrial 2 (sarcomeric)|mib-CK|sarcomeric mitochondrial creatine kinase', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81233320, end_position_on_the_genomic_accession=81266398, orientation='plus', exon_count=11, OMIM=123295, no_hdr0=''),
  1161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1161, CurrentID=0, Status='live', Symbol='ERCC8', Aliases='CKN1, CSA, UVSS2', description='excision repair cross-complementation group 8', other_designations='Cockayne syndrome WD-repeat protein CSA|cockayne syndrome WD repeat protein CSA|excision repair cross-complementing rodent repair deficiency complementation group 8|excision repair cross-complementing rodent repair deficiency, complementation group 8', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60873832, end_position_on_the_genomic_accession=60945078, orientation='minus', exon_count=13, OMIM=609412, no_hdr0=''),
  1163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1163, CurrentID=0, Status='live', Symbol='CKS1B', Aliases='CKS1, PNAS-16, PNAS-18, ckshs1', description='CDC28 protein kinase regulatory subunit 1B', other_designations='CDC2-associated protein CKS1|CDC28 protein kinase 1|CDC28 protein kinase 1B|CKS-1|NB4 apoptosis/differentiation related protein|PNAS-143|cell division control protein CKS1', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154974642, end_position_on_the_genomic_accession=154979249, orientation='plus', exon_count=3, OMIM=116900, no_hdr0=''),
  1164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1164, CurrentID=0, Status='live', Symbol='CKS2', Aliases='CKSHS2', description='CDC28 protein kinase regulatory subunit 2', other_designations='CDC28 protein kinase 2|CKS-2|CKS1(S. cerevisiae Cdc28/Cdc2 kinase subunit) homolog-2', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89311195, end_position_on_the_genomic_accession=89316703, orientation='plus', exon_count=3, OMIM=116901, no_hdr0=''),
  1165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1165, CurrentID=0, Status='live', Symbol='SCAR2', Aliases='CLA1, CPD3', description='spinocerebellar ataxia, autosomal recessive 2', other_designations='Cerebellar ataxia-1|cerebellar ataxia 1 (autosomal recessive)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=213200, no_hdr0=''),
  100271246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271246, CurrentID=0, Status='live', Symbol='RPS2P43', Aliases='RPS2_18_1391', description='ribosomal protein S2 pseudogene 43', other_designations='', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75838528, end_position_on_the_genomic_accession=75839138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1167, CurrentID=0, Status='live', Symbol='CLA3', Aliases='CPD1, SCAR6', description='cerebellar ataxia 3 (cerebellar parenchyma disorder 1)', other_designations='cerebellar ataxia, infantile nonprogressive, autosomal recessive|cerebellar parenchyma disorder 1', map_location='20q11-q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608029, no_hdr0=''),
  100271248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271248, CurrentID=0, Status='live', Symbol='RPS2P44', Aliases='RPS2_21_1482', description='ribosomal protein S2 pseudogene 44', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48577487, end_position_on_the_genomic_accession=48578392, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132241, CurrentID=0, Status='live', Symbol='RPL32P3', Aliases='', description='ribosomal protein L32 pseudogene 3', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129382834, end_position_on_the_genomic_accession=129399439, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271250, CurrentID=0, Status='live', Symbol='RPS2P52', Aliases='RPS2_23_1615', description='ribosomal protein S2 pseudogene 52', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=690442, end_position_on_the_genomic_accession=691211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132243, CurrentID=0, Status='live', Symbol='H1FOO', Aliases='H1.8, H1oo, osH1', description='H1 histone family member O, oocyte specific', other_designations='H1 histone family, member O, oocyte-specific|oocyte-specific histone H1|oocyte-specific linker histone H1', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129543214, end_position_on_the_genomic_accession=129551467, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100271252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271252, CurrentID=0, Status='live', Symbol='RPS2P54', Aliases='RPS2_28_1709', description='ribosomal protein S2 pseudogene 54', other_designations='', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47160790, end_position_on_the_genomic_accession=47161700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1173, CurrentID=0, Status='live', Symbol='AP2M1', Aliases='AP50, CLAPM1, mu2', description='adaptor related protein complex 2, mu 1 subunit', other_designations='AP-2 mu 2 chain|HA2 50 kDA subunit|adaptin-mu2|adaptor protein complex AP-2 subunit mu|adaptor-related protein complex 2 subunit mu|adaptor-related protein complex 2, mu 1 subunit|clathrin adaptor complex AP2, mu subunit|clathrin assembly protein complex 2 medium chain|clathrin assembly protein complex 2 mu medium chain|clathrin coat adaptor protein AP50|clathrin coat assembly protein AP50|clathrin coat-associated protein AP50|clathrin-associated/assembly/adaptor protein, medium 1|plasma membrane adaptor AP-2 50 kDa protein|plasma membrane adaptor AP-2 50kDA protein', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184174846, end_position_on_the_genomic_accession=184184091, orientation='plus', exon_count=12, OMIM=601024, no_hdr0=''),
  1174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1174, CurrentID=0, Status='live', Symbol='AP1S1', Aliases='AP19, CLAPS1, EKV3, MEDNIK, SIGMA1A', description='adaptor related protein complex 1 sigma 1 subunit', other_designations='HA1 19 kDa subunit|adapter-related protein complex 1 sigma-1A subunit|adaptor protein complex AP-1 subunit sigma-1A|adaptor-related protein complex 1 sigma 1 subunit|adaptor-related protein complex 1 subunit sigma-1A|adaptor-related protein complex 1, sigma 1 subunit|clathrin assembly protein complex 1 sigma-1A small chain|clathrin coat assembly protein AP19|clathrin-associated/assembly/adaptor protein, small 1 (19kD)|golgi adaptor HA1/AP1 adaptin sigma-1A subunit|sigma1A subunit of AP-1 clathrin adaptor complex|sigma1A-adaptin', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101154405, end_position_on_the_genomic_accession=101161276, orientation='plus', exon_count=5, OMIM=603531, no_hdr0=''),
  1175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1175, CurrentID=0, Status='live', Symbol='AP2S1', Aliases='AP17, CLAPS2, FBH3, FBHOk, HHC3', description='adaptor related protein complex 2 sigma 1 subunit', other_designations='HA2 17 kDa subunit|adaptor protein complex AP-2 subunit sigma|adaptor-related protein complex 2 sigma 1 subunit|adaptor-related protein complex 2, sigma 1 subunit|clathrin assembly protein 2 sigma small chain|clathrin coat assembly protein AP17|clathrin coat-associated protein AP17|clathrin-associated/assembly/adaptor protein, small 2 (17kD)|plasma membrane adaptor AP-2 17 kDa protein|sigma2-adaptin', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46838158, end_position_on_the_genomic_accession=46850995, orientation='minus', exon_count=7, OMIM=602242, no_hdr0=''),
  1176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1176, CurrentID=0, Status='live', Symbol='AP3S1', Aliases='CLAPS3, Sigma3A', description='adaptor related protein complex 3, sigma 1 subunit', other_designations='adapter-related protein complex 3 subunit sigma-1|adaptor-related protein complex 3 subunit sigma-1|clathrin adaptor complex AP3, sigma-3A subunit|clathrin-associated/assembly/adapter protein, small 3|clathrin-associated/assembly/adaptor protein, small 3 (22kD)|sigma-adaptin 3a', map_location='5q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115841922, end_position_on_the_genomic_accession=115914081, orientation='plus', exon_count=8, OMIM=601507, no_hdr0=''),
  100271257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271257, CurrentID=0, Status='live', Symbol='RPS3AP28', Aliases='RPS3A_13_855', description='ribosomal protein S3a pseudogene 28', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138490543, end_position_on_the_genomic_accession=138491373, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1178, CurrentID=0, Status='live', Symbol='CLC', Aliases='GAL10, Gal-10, LGALS10, LGALS10A, LPPL_HUMAN', description='Charcot-Leyden crystal galectin', other_designations='Charcot-Leyden crystal protein|eosinophil lysophospholipase|galectin 10|lectin, galactoside-binding, soluble, 10|lysolecithin acylhydrolase', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39731253, end_position_on_the_genomic_accession=39738029, orientation='minus', exon_count=4, OMIM=153310, no_hdr0=''),
  1179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1179, CurrentID=0, Status='live', Symbol='CLCA1', Aliases='CACC, CACC1, CLCRG1, CaCC-1, GOB5, hCLCA1, hCaCC-1', description='chloride channel accessory 1', other_designations='CLCA family member 1, chloride channel regulator|calcium-activated chloride channel family member 1|calcium-activated chloride channel protein 1|calcium-dependent chloride channel-1|chloride channel regulator 1|chloride channel, calcium activated, family member 1', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86468843, end_position_on_the_genomic_accession=86500294, orientation='plus', exon_count=14, OMIM=603906, no_hdr0=''),
  1180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1180, CurrentID=0, Status='live', Symbol='CLCN1', Aliases='CLC1', description='chloride channel, voltage-sensitive 1', other_designations='chloride channel 1, skeletal muscle|chloride channel protein, skeletal muscle|clC-1', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143316126, end_position_on_the_genomic_accession=143353862, orientation='plus', exon_count=24, OMIM=118425, no_hdr0=''),
  1181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1181, CurrentID=0, Status='live', Symbol='CLCN2', Aliases='CIC-2, CLC2, ECA2, ECA3, EGI11, EGI3, EGMA, EJM6, EJM8, LKPAT, clC-2', description='chloride channel, voltage-sensitive 2', other_designations='chloride channel 2', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184346185, end_position_on_the_genomic_accession=184361651, orientation='minus', exon_count=25, OMIM=600570, no_hdr0=''),
  1182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1182, CurrentID=0, Status='live', Symbol='CLCN3', Aliases='CLC3, ClC-3', description='chloride channel, voltage-sensitive 3', other_designations='chloride channel 3|chloride channel protein 3|chloride transporter ClC-3', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169620521, end_position_on_the_genomic_accession=169723187, orientation='plus', exon_count=16, OMIM=600580, no_hdr0=''),
  1183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1183, CurrentID=0, Status='live', Symbol='CLCN4', Aliases='CLC4, ClC-4, ClC-4A', description='chloride channel, voltage-sensitive 4', other_designations='chloride channel 4|chloride channel protein 4|chloride transporter ClC-4', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=10156945, end_position_on_the_genomic_accession=10237660, orientation='plus', exon_count=13, OMIM=302910, no_hdr0=''),
  1184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1184, CurrentID=0, Status='live', Symbol='CLCN5', Aliases='CLC5, CLCK2, ClC-5, DENTS, NPHL1, NPHL2, XLRH, XRN, hCIC-K2', description='chloride channel, voltage-sensitive 5', other_designations='chloride transporter ClC-5|voltage-gated chloride ion channel CLCN5', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49922615, end_position_on_the_genomic_accession=50099235, orientation='plus', exon_count=18, OMIM=300008, no_hdr0=''),
  1185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1185, CurrentID=0, Status='live', Symbol='CLCN6', Aliases='CLC-6', description='chloride channel, voltage-sensitive 6', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11806096, end_position_on_the_genomic_accession=11843144, orientation='plus', exon_count=23, OMIM=602726, no_hdr0=''),
  1186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1186, CurrentID=0, Status='live', Symbol='CLCN7', Aliases='CLC-7, CLC7, OPTA2, OPTB4, PPP1R63', description='chloride channel, voltage-sensitive 7', other_designations='chloride channel 7 alpha subunit|chloride channel protein 7|protein phosphatase 1, regulatory subunit 63', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1444933, end_position_on_the_genomic_accession=1475084, orientation='minus', exon_count=26, OMIM=602727, no_hdr0=''),
  1187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1187, CurrentID=0, Status='live', Symbol='CLCNKA', Aliases='CLCK1, ClC-K1, hClC-Ka', description='chloride channel, voltage-sensitive Ka', other_designations='chloride channel Ka|chloride channel, kidney, A', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16021991, end_position_on_the_genomic_accession=16034050, orientation='plus', exon_count=20, OMIM=602024, no_hdr0=''),
  1188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1188, CurrentID=0, Status='live', Symbol='CLCNKB', Aliases='CLCKB, ClC-K2, ClC-Kb', description='chloride channel, voltage-sensitive Kb', other_designations='chloride channel, kidney, B', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16043736, end_position_on_the_genomic_accession=16057326, orientation='plus', exon_count=23, OMIM=602023, no_hdr0=''),
  1191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1191, CurrentID=0, Status='live', Symbol='CLU', Aliases='AAG4, APO-J, APOJ, CLI1, CLU2, KUB1, NA1/NA2, SGP-2, SGP2, SP-40, TRPM-2, TRPM2, CLU', description='clusterin', other_designations='aging-associated protein 4|apolipoprotein J|complement cytolysis inhibitor|complement lysis inhibitor|complement-associated protein SP-40,40|ku70-binding protein 1|sulfated glycoprotein 2|testosterone-repressed prostate message 2', map_location='8p21-p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27596917, end_position_on_the_genomic_accession=27615031, orientation='minus', exon_count=11, OMIM=185430, no_hdr0=''),
  1192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1192, CurrentID=0, Status='live', Symbol='CLIC1', Aliases='G6, NCC27', description='chloride intracellular channel 1', other_designations='RNCC protein|chloride channel ABP|hRNCC|nuclear chloride ion channel 27|nuclear chloride ion channel protein|p64CLCP|regulatory nuclear chloride ion channel protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31730581, end_position_on_the_genomic_accession=31737318, orientation='minus', exon_count=7, OMIM=602872, no_hdr0=''),
  1193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1193, CurrentID=0, Status='live', Symbol='CLIC2', Aliases='CLIC2b, MRXS32, XAP121', description='chloride intracellular channel 2', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155276207, end_position_on_the_genomic_accession=155334681, orientation='minus', exon_count=6, OMIM=300138, no_hdr0=''),
  100271274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271274, CurrentID=0, Status='live', Symbol='RPL12P29', Aliases='RPL12_8_1073', description='ribosomal protein L12 pseudogene 29', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84389120, end_position_on_the_genomic_accession=84389719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1195, CurrentID=0, Status='live', Symbol='CLK1', Aliases='CLK, CLK/STY, STY', description='CDC like kinase 1', other_designations='CDC-like kinase 1|CDC28/CDC2-like kinase|protein tyrosine kinase STY', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200853009, end_position_on_the_genomic_accession=200864744, orientation='minus', exon_count=14, OMIM=601951, no_hdr0=''),
  1196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1196, CurrentID=0, Status='live', Symbol='CLK2', Aliases='', description='CDC like kinase 2', other_designations='CDC-like kinase 2|CLK kinase', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155262868, end_position_on_the_genomic_accession=155273529, orientation='minus', exon_count=14, OMIM=602989, no_hdr0=''),
  1197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1197, CurrentID=0, Status='live', Symbol='CLK2P1', Aliases='CLK2P', description='CDC like kinase 2, pseudogene 1', other_designations='CDC-like kinase 2, pseudogene 1', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23584716, end_position_on_the_genomic_accession=23586527, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  1198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1198, CurrentID=0, Status='live', Symbol='CLK3', Aliases='PHCLK3, PHCLK3/152', description='CDC like kinase 3', other_designations='CDC-like kinase 3', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74598476, end_position_on_the_genomic_accession=74645418, orientation='plus', exon_count=21, OMIM=602990, no_hdr0=''),
  100271279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271279, CurrentID=0, Status='live', Symbol='RPL13AP15', Aliases='RPL13A_4_746', description='ribosomal protein L13a pseudogene 15', other_designations='', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119269133, end_position_on_the_genomic_accession=119269711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1200, CurrentID=0, Status='live', Symbol='TPP1', Aliases='CLN2, GIG1, LPIC, SCAR7, TPP-1', description='tripeptidyl peptidase I', other_designations='cell growth-inhibiting gene 1 protein|growth-inhibiting protein 1|lysosomal pepstatin insensitive protease|tripeptidyl aminopeptidase', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6612766, end_position_on_the_genomic_accession=6619461, orientation='minus', exon_count=13, OMIM=607998, no_hdr0=''),
  1201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1201, CurrentID=0, Status='live', Symbol='CLN3', Aliases='BTS, JNCL', description='ceroid-lipofuscinosis, neuronal 3', other_designations='batten disease protein', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28466653, end_position_on_the_genomic_accession=28492302, orientation='minus', exon_count=16, OMIM=607042, no_hdr0=''),
  100271282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271282, CurrentID=0, Status='live', Symbol='RPL14P4', Aliases='RPL14_2_1267', description='ribosomal protein L14 pseudogene 4', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95125735, end_position_on_the_genomic_accession=95126423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1203, CurrentID=0, Status='live', Symbol='CLN5', Aliases='NCL', description='ceroid-lipofuscinosis, neuronal 5', other_designations='', map_location='13q21.1-q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76991924, end_position_on_the_genomic_accession=77002517, orientation='plus', exon_count=5, OMIM=608102, no_hdr0=''),
  1204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1204, CurrentID=0, Status='live', Symbol='CLNS1AP1', Aliases='CLNS1B, ICln', description='chloride channel, nucleotide-sensitive, 1A pseudogene 1', other_designations='chloride channel, nucleotide-sensitive, 1B', map_location='6p12-q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54484499, end_position_on_the_genomic_accession=54485853, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271285, CurrentID=0, Status='live', Symbol='RPL18AP10', Aliases='RPL18A_4_806', description='ribosomal protein L18a pseudogene 10', other_designations='', map_location='7p14.1-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43274129, end_position_on_the_genomic_accession=43274634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271286, CurrentID=0, Status='live', Symbol='RPL18P10', Aliases='RPL18_6_1311', description='ribosomal protein L18 pseudogene 10', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50099334, end_position_on_the_genomic_accession=50099875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1207, CurrentID=0, Status='live', Symbol='CLNS1A', Aliases='CLCI, CLNS1B, ICln', description='chloride channel, nucleotide-sensitive, 1A', other_designations='chloride channel regulatory protein|chloride channel, nucleotide sensitive 1A|chloride conductance regulatory protein ICln|chloride ion current inducer protein|i(Cln)|reticulocyte pICln|reticulocyte protein ICln', map_location='11q13.5-q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77616151, end_position_on_the_genomic_accession=77637806, orientation='minus', exon_count=7, OMIM=602158, no_hdr0=''),
  1208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1208, CurrentID=0, Status='live', Symbol='CLPS', Aliases='', description='colipase', other_designations='colipase, pancreatic|pancreatic colipase preproprotein', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35794982, end_position_on_the_genomic_accession=35797344, orientation='minus', exon_count=4, OMIM=120105, no_hdr0=''),
  1209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1209, CurrentID=0, Status='live', Symbol='CLPTM1', Aliases='', description='cleft lip and palate associated transmembrane protein 1', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44954585, end_position_on_the_genomic_accession=44993346, orientation='plus', exon_count=16, OMIM=604783, no_hdr0=''),
  100271290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271290, CurrentID=0, Status='live', Symbol='RPL22P18', Aliases='RPL22_9_1070', description='ribosomal protein L22 pseudogene 18', other_designations='', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80125941, end_position_on_the_genomic_accession=80126314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1211, CurrentID=0, Status='live', Symbol='CLTA', Aliases='LCA', description='clathrin, light chain A', other_designations='clathrin, light polypeptide (Lca)', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36190856, end_position_on_the_genomic_accession=36212062, orientation='plus', exon_count=7, OMIM=118960, no_hdr0=''),
  1212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1212, CurrentID=0, Status='live', Symbol='CLTB', Aliases='LCB', description='clathrin, light chain B', other_designations='clathrin, light chain (Lcb)|clathrin, light polypeptide (Lcb)', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176392455, end_position_on_the_genomic_accession=176416569, orientation='minus', exon_count=8, OMIM=118970, no_hdr0=''),
  1213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1213, CurrentID=0, Status='live', Symbol='CLTC', Aliases='CHC, CHC17, CLH-17L2, Hc, CLTC', description='clathrin, heavy chain (Hc)', other_designations='clathrin heavy chain on chromosome 17|clathrin, heavy polypeptide (Hc)|clathrin, heavy polypeptide-like 2', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59619689, end_position_on_the_genomic_accession=59696956, orientation='plus', exon_count=36, OMIM=118955, no_hdr0=''),
  100271294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271294, CurrentID=0, Status='live', Symbol='RPL23AP34', Aliases='RPL23A_8_204', description='ribosomal protein L23a pseudogene 34', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28308161, end_position_on_the_genomic_accession=28308570, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1215, CurrentID=0, Status='live', Symbol='CMA1', Aliases='CYH, MCT1, chymase', description='chymase 1, mast cell', other_designations='alpha-chymase|chymase 1 preproprotein transcript E|chymase 1 preproprotein transcript I|chymase, heart|chymase, mast cell|mast cell protease I', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24505346, end_position_on_the_genomic_accession=24508690, orientation='minus', exon_count=5, OMIM=118938, no_hdr0=''),
  100271296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271296, CurrentID=0, Status='live', Symbol='RPL23AP36', Aliases='RPL23A_9_314', description='ribosomal protein L23a pseudogene 36', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203074685, end_position_on_the_genomic_accession=203075131, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271297, CurrentID=0, Status='live', Symbol='RPL23P11', Aliases='RPL23_6_1032', description='ribosomal protein L23 pseudogene 11', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33864313, end_position_on_the_genomic_accession=33864723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1218, CurrentID=0, Status='live', Symbol='CMD1B', Aliases='CMPD1, FDC', description='cardiomyopathy, dilated 1B (autosomal dominant)', other_designations='', map_location='9q13-q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600884, no_hdr0=''),
  100271299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271299, CurrentID=0, Status='live', Symbol='RPL27AP4', Aliases='RPL27A_2_255', description='ribosomal protein L27a pseudogene 4', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106304755, end_position_on_the_genomic_accession=106305204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271300, CurrentID=0, Status='live', Symbol='RPL27AP5', Aliases='RPL27A_2_577', description='ribosomal protein L27a pseudogene 5', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74990189, end_position_on_the_genomic_accession=74990637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271301, CurrentID=0, Status='live', Symbol='RPL27AP7', Aliases='RPL27A_3_699', description='ribosomal protein L27a pseudogene 7', other_designations='', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47752831, end_position_on_the_genomic_accession=47753234, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1222, CurrentID=0, Status='live', Symbol='CMD1F', Aliases='CDCD3, LGMD1D, LGMD1E', description='cardiomyopathy, dilated 1F (autosomal dominant)', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602067, no_hdr0=''),
  100271303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271303, CurrentID=0, Status='live', Symbol='RPL30P11', Aliases='RPL30_6_1213', description='ribosomal protein L30 pseudogene 11', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14356865, end_position_on_the_genomic_accession=14357137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271304, CurrentID=0, Status='live', Symbol='RPL30P12', Aliases='RPL30_6_1274', description='ribosomal protein L30 pseudogene 12', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106910757, end_position_on_the_genomic_accession=106911112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271305, CurrentID=0, Status='live', Symbol='RPL30P13', Aliases='RPL30_7_1228', description='ribosomal protein L30 pseudogene 13', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40068240, end_position_on_the_genomic_accession=40068663, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271306, CurrentID=0, Status='live', Symbol='RPL32P27', Aliases='RPL32_7_1234', description='ribosomal protein L32 pseudogene 27', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48903415, end_position_on_the_genomic_accession=48903813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132299, CurrentID=0, Status='live', Symbol='OCIAD2', Aliases='', description='OCIA domain containing 2', other_designations='ovarian carcinoma immunoreactive antigen-like protein', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48885380, end_position_on_the_genomic_accession=48906828, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100271308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271308, CurrentID=0, Status='live', Symbol='RPL34P20', Aliases='RPL34_8_1088', description='ribosomal protein L34 pseudogene 20', other_designations='', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97393198, end_position_on_the_genomic_accession=97393546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271309, CurrentID=0, Status='live', Symbol='RPL34P24', Aliases='RPL34_9_1123', description='ribosomal protein L34 pseudogene 24', other_designations='', map_location='11p15.5-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17137976, end_position_on_the_genomic_accession=17138265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1230, CurrentID=0, Status='live', Symbol='CCR1', Aliases='CD191, CKR-1, CKR1, CMKBR1, HM145, MIP1aR, SCYAR1', description='chemokine (C-C motif) receptor 1', other_designations='C-C CKR-1|CC-CKR-1|CCR-1|LD78 receptor|MIP-1alpha-R|RANTES receptor|RANTES-R|macrophage inflammatory protein 1-alpha receptor', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46201709, end_position_on_the_genomic_accession=46208341, orientation='minus', exon_count=2, OMIM=601159, no_hdr0=''),
  100271311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271311, CurrentID=0, Status='live', Symbol='RPL35AP9', Aliases='RPL35A_2_427', description='ribosomal protein L35a pseudogene 9', other_designations='', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158763366, end_position_on_the_genomic_accession=158763805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1232, CurrentID=0, Status='live', Symbol='CCR3', Aliases='CC-CKR-3, CD193, CKR3, CMKBR3', description='chemokine (C-C motif) receptor 3', other_designations='C-C CKR-3|CC chemokine receptor 3|CCR-3|b-chemokine receptor|eosinophil CC chemokine receptor 3|eosinophil eotaxin receptor', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46210695, end_position_on_the_genomic_accession=46266706, orientation='plus', exon_count=5, OMIM=601268, no_hdr0=''),
  1233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1233, CurrentID=0, Status='live', Symbol='CCR4', Aliases='CC-CKR-4, CD194, CKR4, CMKBR4, ChemR13, HGCN:14099, K5-5', description='chemokine (C-C motif) receptor 4', other_designations='C-C CKR-4|CCR-4|chemokine (C-C) receptor 4', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32951574, end_position_on_the_genomic_accession=32954911, orientation='plus', exon_count=2, OMIM=604836, no_hdr0=''),
  1234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1234, CurrentID=0, Status='live', Symbol='CCR5', Aliases='CC-CKR-5, CCCKR5, CCR-5, CD195, CKR-5, CKR5, CMKBR5, IDDM22', description='chemokine (C-C motif) receptor 5 (gene/pseudogene)', other_designations='C-C motif chemokine receptor 5 A159A|HIV-1 fusion coreceptor|chemokine receptor CCR5|chemr13', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46370142, end_position_on_the_genomic_accession=46376206, orientation='plus', exon_count=3, OMIM=601373, no_hdr0=''),
  1235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1235, CurrentID=0, Status='live', Symbol='CCR6', Aliases='BN-1, C-C CKR-6, CC-CKR-6, CCR-6, CD196, CKR-L3, CKRL3, CMKBR6, DCR2, DRY6, GPR29, GPRCY4, STRL22', description='chemokine (C-C motif) receptor 6', other_designations='G protein-coupled receptor 29|G-protein coupled receptor 29|LARC receptor|chemokine (C-C) receptor 6|chemokine receptor-like 3|seven-transmembrane receptor, lymphocyte, 22', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167111807, end_position_on_the_genomic_accession=167139141, orientation='plus', exon_count=4, OMIM=601835, no_hdr0=''),
  1236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1236, CurrentID=0, Status='live', Symbol='CCR7', Aliases='BLR2, CC-CKR-7, CCR-7, CD197, CDw197, CMKBR7, EBI1', description='chemokine (C-C motif) receptor 7', other_designations="Bukitt's lymphoma receptor 2|CC chemokine receptor 7|EBV-induced G protein-coupled receptor 1|Epstein-Barr virus induced gene 1|Epstein-Barr virus-induced G-protein coupled receptor 1|MIP-3 beta receptor|lymphocyte-specific G protein-coupled peptide receptor", map_location='17q12-q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40553769, end_position_on_the_genomic_accession=40565484, orientation='minus', exon_count=5, OMIM=600242, no_hdr0=''),
  1237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1237, CurrentID=0, Status='live', Symbol='CCR8', Aliases='CC-CKR-8, CCR-8, CDw198, CKRL1, CMKBR8, CMKBRL2, CY6, GPRCY6, TER1', description='chemokine (C-C motif) receptor 8', other_designations='CC chemokine receptor 8|CC chemokine receptor CHEMR1|CC-chemokine receptor chemr1|chemokine (C-C) receptor 8|chemokine (C-C) receptor-like 2|chemokine receptor-like 1', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39329706, end_position_on_the_genomic_accession=39333680, orientation='plus', exon_count=2, OMIM=601834, no_hdr0=''),
  1238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1238, CurrentID=0, Status='live', Symbol='ACKR2', Aliases='CCBP2, CCR10, CCR9, CMKBR9, D6, hD6', description='atypical chemokine receptor 2', other_designations='C-C chemokine receptor D6|CC-chemokine-binding receptor JAB61|chemokine (C-C motif) receptor 9|chemokine (C-C) receptor 9|chemokine binding protein 2|chemokine receptor CCR-10|chemokine receptor CCR-9|chemokine receptor D6|chemokine-binding protein 2|chemokine-binding protein D6', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42809472, end_position_on_the_genomic_accession=42867283, orientation='plus', exon_count=3, OMIM=602648, no_hdr0=''),
  100271319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271319, CurrentID=0, Status='live', Symbol='RPL35AP17', Aliases='RPL35A_7_625', description='ribosomal protein L35a pseudogene 17', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146138761, end_position_on_the_genomic_accession=146139008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1240, CurrentID=0, Status='live', Symbol='CMKLR1', Aliases='CHEMERINR, ChemR23, DEZ, RVER1', description='chemerin chemokine-like receptor 1', other_designations='G-protein coupled receptor ChemR23|G-protein coupled receptor DEZ|chemerin receptor|chemokine receptor-like 1|orphan G-protein coupled receptor, Dez|resolvin E1 receptor', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108288044, end_position_on_the_genomic_accession=108339317, orientation='minus', exon_count=5, OMIM=602351, no_hdr0=''),
  1241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1241, CurrentID=0, Status='live', Symbol='LTB4R', Aliases='BLT1, BLTR, CMKRL1, GPR161, LTBR1, P2RY7, P2Y7, LTB4R', description='leukotriene B4 receptor', other_designations='G protein-coupled receptor 16|G-protein coupled receptor 16|LTB4-R 1|LTB4-R1|P2Y purinoceptor 7|chemoattractant receptor-like 1|chemokine receptor-like 1|purinergic receptor P2Y, G-protein coupled, 7', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24311499, end_position_on_the_genomic_accession=24318036, orientation='plus', exon_count=3, OMIM=601531, no_hdr0=''),
  100271322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271322, CurrentID=0, Status='live', Symbol='RPL35AP19', Aliases='RPL35A_8_947', description='ribosomal protein L35a pseudogene 19', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121380093, end_position_on_the_genomic_accession=121380539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1243, CurrentID=0, Status='live', Symbol='CMM', Aliases='CMM1, DNS, FAMMM, MLM', description='cutaneous malignant melanoma/dysplastic nevus', other_designations='dysplastic nevus syndrome', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=155600, no_hdr0=''),
  1244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1244, CurrentID=0, Status='live', Symbol='ABCC2', Aliases='ABC30, CMOAT, DJS, MRP2, cMRP', description='ATP binding cassette subfamily C member 2', other_designations='ATP-binding cassette, sub-family C (CFTR/MRP), member 2|canalicular multidrug resistance protein|multidrug resistance-associated protein 2', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99782598, end_position_on_the_genomic_accession=99852576, orientation='plus', exon_count=34, OMIM=601107, no_hdr0=''),
  100271325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271325, CurrentID=0, Status='live', Symbol='RPL36AP15', Aliases='RPL36A_4_216', description='ribosomal protein L36a pseudogene 15', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47797754, end_position_on_the_genomic_accession=47798140, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271326, CurrentID=0, Status='live', Symbol='RPL36AP17', Aliases='RPL36A_4_361', description='ribosomal protein L36a pseudogene 17', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35871858, end_position_on_the_genomic_accession=35872216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271327, CurrentID=0, Status='live', Symbol='RPL36AP16', Aliases='RPL36A_5_238', description='ribosomal protein L36a pseudogene 16', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69594689, end_position_on_the_genomic_accession=69595086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132320, CurrentID=0, Status='live', Symbol='SCLT1', Aliases='CAP-1A, CAP1A', description='sodium channel and clathrin linker 1', other_designations='sodium channel-associated protein 1', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128874418, end_position_on_the_genomic_accession=129093607, orientation='minus', exon_count=27, OMIM=611399, no_hdr0=''),
  132321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132321, CurrentID=0, Status='live', Symbol='C4orf33', Aliases='', description='chromosome 4 open reading frame 33', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129093674, end_position_on_the_genomic_accession=129112688, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100271330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271330, CurrentID=0, Status='live', Symbol='RPL36AP19', Aliases='RPL36A_7_509', description='ribosomal protein L36a pseudogene 19', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112171180, end_position_on_the_genomic_accession=112171578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271331, CurrentID=0, Status='live', Symbol='RPL36AP25', Aliases='RPL36A_7_666', description='ribosomal protein L36a pseudogene 25', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20722328, end_position_on_the_genomic_accession=20722569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271332, CurrentID=0, Status='live', Symbol='RPL36AP21', Aliases='RPL36A_8_545', description='ribosomal protein L36a pseudogene 21', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=18049544, end_position_on_the_genomic_accession=18049915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1253, CurrentID=0, Status='live', Symbol='CMTX2', Aliases='', description='Charcot-Marie-Tooth neuropathy, X-linked 2 (recessive)', other_designations='Charcot-Marie-Tooth disease, X-linked-2, recessive', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302801, no_hdr0=''),
  1254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1254, CurrentID=0, Status='live', Symbol='CMTX3', Aliases='', description='Charcot-Marie-Tooth neuropathy, X-linked 3 (dominant)', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302802, no_hdr0=''),
  1255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1255, CurrentID=0, Status='live', Symbol='CNA1', Aliases='', description='cornea plana 1 (autosomal dominant)', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=121400, no_hdr0=''),
  100271336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271336, CurrentID=0, Status='live', Symbol='RPL36P17', Aliases='RPL36_8_1743', description='ribosomal protein L36 pseudogene 17', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30105100, end_position_on_the_genomic_accession=30105408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1257, CurrentID=0, Status='live', Symbol='CNC2', Aliases='CNC', description='Carney complex type 2, multiple neoplasia and lentiginosis', other_designations='Carney complex, multiple neoplasia and lentiginosis', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605244, no_hdr0=''),
  1258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1258, CurrentID=0, Status='live', Symbol='CNGB1', Aliases='CNCG2, CNCG3L, CNCG4, CNG4B, GAR1, GARP, GARP2, RCNC2, RCNCb, RCNCbeta, RP45, CNGB1', description='cyclic nucleotide gated channel beta 1', other_designations='CNG channel beta-1|CNG-4|cyclic nucleotide-gated cation channel 4|cyclic nucleotide-gated cation channel gamma|cyclic nucleotide-gated cation channel modulatory subunit|cyclic nucleotide-gated channel beta-1|glutamic acid-rich protein|glutamic-acid-rich protein', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57882340, end_position_on_the_genomic_accession=57979295, orientation='minus', exon_count=37, OMIM=600724, no_hdr0=''),
  1259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1259, CurrentID=0, Status='live', Symbol='CNGA1', Aliases='CNCG, CNCG1, CNG-1, CNG1, RCNC1, RCNCa, RCNCalpha, RP49', description='cyclic nucleotide gated channel alpha 1', other_designations='CNG channel alpha-1|cyclic nucleotide-gated cation channel 1|cyclic nucleotide-gated channel alpha-1|cyclic nucleotide-gated channel, photoreceptor|interleukin-1 homologue|rod photoreceptor cGMP-gated channel subunit alpha', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47935015, end_position_on_the_genomic_accession=48016679, orientation='minus', exon_count=13, OMIM=123825, no_hdr0=''),
  1260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1260, CurrentID=0, Status='live', Symbol='CNGA2', Aliases='CNCA, CNCA1, CNG2, OCNC1, OCNCALPHA, OCNCa', description='cyclic nucleotide gated channel alpha 2', other_designations='CNG channel alpha-2|cyclic-nucleotide-gated cation channel 2', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151734746, end_position_on_the_genomic_accession=151745564, orientation='plus', exon_count=7, OMIM=300338, no_hdr0=''),
  1261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1261, CurrentID=0, Status='live', Symbol='CNGA3', Aliases='ACHM2, CCNC1, CCNCa, CCNCalpha, CNCG3, CNG3', description='cyclic nucleotide gated channel alpha 3', other_designations='CNG channel alpha-3|CNG-3|cone photoreceptor cGMP-gated channel alpha subunit|cone photoreceptor cGMP-gated channel subunit alpha|cyclic nucleotide-gated channel alpha-3', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98346155, end_position_on_the_genomic_accession=98399154, orientation='plus', exon_count=9, OMIM=600053, no_hdr0=''),
  1262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1262, CurrentID=0, Status='live', Symbol='CNGA4', Aliases='CNCA2, CNG-4, CNG4, CNG5, CNGB2, OCNC2, OCNCBETA, OCNCb', description='cyclic nucleotide gated channel alpha 4', other_designations='CNG channel alpha-4|cyclic nucleotide gated channel beta 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6234774, end_position_on_the_genomic_accession=6245924, orientation='plus', exon_count=7, OMIM=609472, no_hdr0=''),
  1263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1263, CurrentID=0, Status='live', Symbol='PLK3', Aliases='CNK, FNK, PLK-3, PRK', description='polo-like kinase 3', other_designations='FGF-inducible kinase|cytokine-inducible serine/threonine-protein kinase|proliferation-related kinase', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44799952, end_position_on_the_genomic_accession=44805995, orientation='plus', exon_count=15, OMIM=602913, no_hdr0=''),
  1264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1264, CurrentID=0, Status='live', Symbol='CNN1', Aliases='HEL-S-14, SMCC, Sm-Calp', description='calponin 1, basic, smooth muscle', other_designations='basic calponin|calponin H1, smooth muscle|calponins, basic|epididymis secretory protein Li 14', map_location='19p13.2-p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11538764, end_position_on_the_genomic_accession=11550323, orientation='plus', exon_count=11, OMIM=600806, no_hdr0=''),
  1265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1265, CurrentID=0, Status='live', Symbol='CNN2', Aliases='', description='calponin 2', other_designations='calponin H2, smooth muscle|neutral calponin', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1026275, end_position_on_the_genomic_accession=1039068, orientation='plus', exon_count=7, OMIM=602373, no_hdr0=''),
  1266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1266, CurrentID=0, Status='live', Symbol='CNN3', Aliases='', description='calponin 3, acidic', other_designations='calponin, acidic isoform|dJ639P13.2.2 (acidic calponin 3)', map_location='1p22-p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94896949, end_position_on_the_genomic_accession=94927223, orientation='minus', exon_count=8, OMIM=602374, no_hdr0=''),
  1267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1267, CurrentID=0, Status='live', Symbol='CNP', Aliases='CNP1', description="2',3'-cyclic nucleotide 3' phosphodiesterase", other_designations="2', 3' cyclic nucleotide 3' phosphohydrolase|CNPase", map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41966741, end_position_on_the_genomic_accession=41977736, orientation='plus', exon_count=5, OMIM=123830, no_hdr0=''),
  1268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1268, CurrentID=0, Status='live', Symbol='CNR1', Aliases='CANN6, CB-R, CB1, CB1A, CB1K5, CB1R, CNR', description='cannabinoid receptor 1 (brain)', other_designations='central cannabinoid receptor', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88139864, end_position_on_the_genomic_accession=88166745, orientation='minus', exon_count=6, OMIM=114610, no_hdr0=''),
  1269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1269, CurrentID=0, Status='live', Symbol='CNR2', Aliases='CB-2, CB2, CX5', description='cannabinoid receptor 2', other_designations='cannabinoid receptor 2 (macrophage)|testis-dominant CNR2 isoform CB2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23873463, end_position_on_the_genomic_accession=23913327, orientation='minus', exon_count=5, OMIM=605051, no_hdr0=''),
  1270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1270, CurrentID=0, Status='live', Symbol='CNTF', Aliases='HCNTF', description='ciliary neurotrophic factor', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58622673, end_position_on_the_genomic_accession=58625733, orientation='plus', exon_count=2, OMIM=118945, no_hdr0=''),
  1271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1271, CurrentID=0, Status='live', Symbol='CNTFR', Aliases='', description='ciliary neurotrophic factor receptor', other_designations='CNTF receptor subunit alpha|CNTFR-alpha', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34551432, end_position_on_the_genomic_accession=34590140, orientation='minus', exon_count=12, OMIM=118946, no_hdr0=''),
  1272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1272, CurrentID=0, Status='live', Symbol='CNTN1', Aliases='F3, GP135, MYPCN', description='contactin 1', other_designations='glycoprotein gP135|neural cell surface protein F3', map_location='12q11-q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40692442, end_position_on_the_genomic_accession=41072412, orientation='plus', exon_count=28, OMIM=600016, no_hdr0=''),
  100271353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271353, CurrentID=0, Status='live', Symbol='RPS12P17', Aliases='RPS12_8_1057', description='ribosomal protein S12 pseudogene 17', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69161514, end_position_on_the_genomic_accession=69161981, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271354, CurrentID=0, Status='live', Symbol='RPS12P18', Aliases='RPS12P19, RPS12_8_1068, RPS12_9_1069', description='ribosomal protein S12 pseudogene 18', other_designations='ribosomal protein S12 pseudogene 19', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79544079, end_position_on_the_genomic_accession=79544356, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1275, CurrentID=0, Status='live', Symbol='COD2', Aliases='CORDX2', description='cone dystrophy 2 (X-linked)', other_designations='', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300085, no_hdr0=''),
  1277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1277, CurrentID=0, Status='live', Symbol='COL1A1', Aliases='EDSC, OI1, OI2, OI3, OI4', description='collagen, type I, alpha 1', other_designations='alpha-1 type I collagen|alpha1(I) procollagen|collagen alpha 1 chain type I|collagen alpha-1(I) chain preproprotein|collagen of skin, tendon and bone, alpha-1 chain|pro-alpha-1 collagen type 1|type I proalpha 1|type I procollagen alpha 1 chain', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50184096, end_position_on_the_genomic_accession=50201648, orientation='minus', exon_count=51, OMIM=120150, no_hdr0=''),
  1278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1278, CurrentID=0, Status='live', Symbol='COL1A2', Aliases='OI4', description='collagen, type I, alpha 2', other_designations='alpha 2 type I procollagen|alpha 2(I) procollagen|alpha 2(I)-collagen|alpha-2 type I collagen|collagen I, alpha-2 polypeptide|collagen of skin, tendon and bone, alpha-2 chain|collagen type I alpha 2|type I procollagen', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94394561, end_position_on_the_genomic_accession=94431232, orientation='plus', exon_count=52, OMIM=120160, no_hdr0=''),
  1279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1279, CurrentID=0, Status='live', Symbol='COL1AR', Aliases='COLR', description='collagen, type I, alpha, receptor', other_designations='Collagen I, alpha, receptor', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=120340, no_hdr0=''),
  1280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1280, CurrentID=0, Status='live', Symbol='COL2A1', Aliases='ANFH, AOM, COL11A3, SEDC, STL1', description='collagen, type II, alpha 1', other_designations='alpha-1 type II collagen|arthroophthalmopathy, progressive (Stickler syndrome)|cartilage collagen|chondrocalcin|collagen II, alpha-1 polypeptide', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47972965, end_position_on_the_genomic_accession=48025286, orientation='minus', exon_count=67, OMIM=120140, no_hdr0=''),
  1281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1281, CurrentID=0, Status='live', Symbol='COL3A1', Aliases='EDS4A', description='collagen, type III, alpha 1', other_designations='Ehlers-Danlos syndrome type IV, autosomal dominant|alpha1 (III) collagen|collagen, fetal', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188974373, end_position_on_the_genomic_accession=189012746, orientation='plus', exon_count=51, OMIM=120180, no_hdr0=''),
  1282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1282, CurrentID=0, Status='live', Symbol='COL4A1', Aliases='BSVD, RATOR', description='collagen, type IV, alpha 1', other_designations='COL4A1 NC1 domain|arresten|collagen IV, alpha-1 polypeptide|collagen of basement membrane, alpha-1 chain', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110148958, end_position_on_the_genomic_accession=110307157, orientation='minus', exon_count=54, OMIM=120130, no_hdr0=''),
  1283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1283, CurrentID=0, Status='live', Symbol='CTD', Aliases='', description='Coats disease', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300216, no_hdr0=''),
  1284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1284, CurrentID=0, Status='live', Symbol='COL4A2', Aliases='ICH, POREN2', description='collagen, type IV, alpha 2', other_designations='canstatin', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110307284, end_position_on_the_genomic_accession=110513027, orientation='plus', exon_count=48, OMIM=120090, no_hdr0=''),
  1285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1285, CurrentID=0, Status='live', Symbol='COL4A3', Aliases='', description='collagen, type IV, alpha 3 (Goodpasture antigen)', other_designations='collagen IV, alpha-3 polypeptide|tumstatin', map_location='2q36-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227164565, end_position_on_the_genomic_accession=227314792, orientation='plus', exon_count=53, OMIM=120070, no_hdr0=''),
  1286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1286, CurrentID=0, Status='live', Symbol='COL4A4', Aliases='CA44', description='collagen, type IV, alpha 4', other_designations='Collagen IV, alpha-4 polypeptide|collagen of basement membrane, alpha-4 chain', map_location='2q35-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226967532, end_position_on_the_genomic_accession=227164559, orientation='minus', exon_count=58, OMIM=120131, no_hdr0=''),
  1287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1287, CurrentID=0, Status='live', Symbol='COL4A5', Aliases='ASLN, ATS, CA54', description='collagen, type IV, alpha 5', other_designations='collagen IV, alpha-5 polypeptide|collagen of basement membrane, alpha-5 chain|collagen type IV alpha 5|dA149D17.3|dA24A23.1', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108439736, end_position_on_the_genomic_accession=108697545, orientation='plus', exon_count=57, OMIM=303630, no_hdr0=''),
  1288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1288, CurrentID=0, Status='live', Symbol='COL4A6', Aliases='CXDELq22.3, DELXq22.3, DFNX6', description='collagen, type IV, alpha 6', other_designations='collagen IV, alpha-6 polypeptide|collagen of basement membrane, alpha-6|dJ889N15.4 (Collagen Alpha 6(IV))', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108155607, end_position_on_the_genomic_accession=108439497, orientation='minus', exon_count=49, OMIM=303631, no_hdr0=''),
  1289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1289, CurrentID=0, Status='live', Symbol='COL5A1', Aliases='EDSC', description='collagen, type V, alpha 1', other_designations='', map_location='9q34.2-q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134641786, end_position_on_the_genomic_accession=134844843, orientation='plus', exon_count=67, OMIM=120215, no_hdr0=''),
  1290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1290, CurrentID=0, Status='live', Symbol='COL5A2', Aliases='EDSC', description='collagen, type V, alpha 2', other_designations='AB collagen|collagen, fetal membrane, A polypeptide|type V preprocollagen alpha 2 chain', map_location='2q14-q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189031915, end_position_on_the_genomic_accession=189225314, orientation='minus', exon_count=55, OMIM=120190, no_hdr0=''),
  1291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1291, CurrentID=0, Status='live', Symbol='COL6A1', Aliases='OPLL', description='collagen, type VI, alpha 1', other_designations='alpha 1 (VI) chain (61 AA)|collagen VI, alpha-1 polypeptide', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45981749, end_position_on_the_genomic_accession=46005049, orientation='plus', exon_count=35, OMIM=120220, no_hdr0=''),
  1292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1292, CurrentID=0, Status='live', Symbol='COL6A2', Aliases='PP3610', description='collagen, type VI, alpha 2', other_designations='collagen VI, alpha-2 polypeptide|human mRNA for collagen VI alpha-2 C-terminal globular domain', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46097606, end_position_on_the_genomic_accession=46132849, orientation='plus', exon_count=31, OMIM=120240, no_hdr0=''),
  1293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1293, CurrentID=0, Status='live', Symbol='COL6A3', Aliases='DYT27', description='collagen, type VI, alpha 3', other_designations='collagen VI, alpha-3 polypeptide', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237324012, end_position_on_the_genomic_accession=237414370, orientation='minus', exon_count=44, OMIM=120250, no_hdr0=''),
  1294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1294, CurrentID=0, Status='live', Symbol='COL7A1', Aliases='EBD1, EBDCT, EBR1, NDNC8', description='collagen, type VII, alpha 1', other_designations='LC collagen|collagen VII, alpha-1 polypeptide|long-chain collagen', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48564073, end_position_on_the_genomic_accession=48595302, orientation='minus', exon_count=120, OMIM=120120, no_hdr0=''),
  1295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1295, CurrentID=0, Status='live', Symbol='COL8A1', Aliases='C3orf7', description='collagen, type VIII, alpha 1', other_designations='cell proliferation-inducing protein 41|collagen VIII, alpha-1 polypeptide|endothelial collagen|smag-64|smooth muscle cell-expressed and macrophage conditioned medium-induced protein smag-64', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99638596, end_position_on_the_genomic_accession=99796733, orientation='plus', exon_count=5, OMIM=120251, no_hdr0=''),
  1296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1296, CurrentID=0, Status='live', Symbol='COL8A2', Aliases='FECD, FECD1, PPCD, PPCD2', description='collagen, type VIII, alpha 2', other_designations='collagen VIII, alpha-2 polypeptide|endothelial collagen', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36095236, end_position_on_the_genomic_accession=36125770, orientation='minus', exon_count=5, OMIM=120252, no_hdr0=''),
  1297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1297, CurrentID=0, Status='live', Symbol='COL9A1', Aliases='DJ149L1.1.2, EDM6, MED, STL4', description='collagen, type IX, alpha 1', other_designations='alpha-1(IX) collagen chain|cartilage-specific short collagen|collagen IX, alpha-1 polypeptide', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70215061, end_position_on_the_genomic_accession=70303083, orientation='minus', exon_count=41, OMIM=120210, no_hdr0=''),
  1298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1298, CurrentID=0, Status='live', Symbol='COL9A2', Aliases='DJ39G22.4, EDM2, MED, STL5', description='collagen, type IX, alpha 2', other_designations='alpha 2 type IX collagen|collagen IX, alpha-2 polypeptide', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40300487, end_position_on_the_genomic_accession=40317690, orientation='minus', exon_count=36, OMIM=120260, no_hdr0=''),
  1299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1299, CurrentID=0, Status='live', Symbol='COL9A3', Aliases='DJ885L7.4.1, EDM3, IDD, MED', description='collagen, type IX, alpha 3', other_designations='collagen IX, alpha-3 polypeptide|collagen type IX proteoglycan', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62817062, end_position_on_the_genomic_accession=62841159, orientation='plus', exon_count=35, OMIM=120270, no_hdr0=''),
  1300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1300, CurrentID=0, Status='live', Symbol='COL10A1', Aliases='', description='collagen, type X, alpha 1', other_designations='Schmid metaphyseal chondrodysplasia|collagen X, alpha-1 polypeptide', map_location='6q21-q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116118909, end_position_on_the_genomic_accession=116218169, orientation='minus', exon_count=6, OMIM=120110, no_hdr0=''),
  1301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1301, CurrentID=0, Status='live', Symbol='COL11A1', Aliases='CO11A1, COLL6, STL2', description='collagen, type XI, alpha 1', other_designations='collagen XI, alpha-1 polypeptide', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=102876467, end_position_on_the_genomic_accession=103108576, orientation='minus', exon_count=71, OMIM=120280, no_hdr0=''),
  1302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1302, CurrentID=0, Status='live', Symbol='COL11A2', Aliases='DFNA13, DFNB53, FBCG2, HKE5, PARP, STL3', description='collagen, type XI, alpha 2', other_designations='pro-a2 chain of collagen type XI', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33162692, end_position_on_the_genomic_accession=33193152, orientation='minus', exon_count=69, OMIM=120290, no_hdr0=''),
  1303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1303, CurrentID=0, Status='live', Symbol='COL12A1', Aliases='BA209D8.1L, DJ234P15.1, COL12A1', description='collagen, type XII, alpha 1', other_designations='collagen type XII proteoglycan', map_location='6q12-q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75084326, end_position_on_the_genomic_accession=75206202, orientation='minus', exon_count=66, OMIM=120320, no_hdr0=''),
  100271384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271384, CurrentID=0, Status='live', Symbol='RPS29P22', Aliases='RPS29_6_1530', description='ribosomal protein S29 pseudogene 22', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28116239, end_position_on_the_genomic_accession=28116406, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1305, CurrentID=0, Status='live', Symbol='COL13A1', Aliases='COLXIIIA1', description='collagen, type XIII, alpha 1', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69801863, end_position_on_the_genomic_accession=69959148, orientation='plus', exon_count=42, OMIM=120350, no_hdr0=''),
  1306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1306, CurrentID=0, Status='live', Symbol='COL15A1', Aliases='', description='collagen, type XV, alpha 1', other_designations='collagen XV, alpha-1 polypeptide|collagen type XV proteoglycan|endostatin-XV|restin', map_location='9q21-q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98943348, end_position_on_the_genomic_accession=99070792, orientation='plus', exon_count=43, OMIM=120325, no_hdr0=''),
  1307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1307, CurrentID=0, Status='live', Symbol='COL16A1', Aliases='447AA, FP1572', description='collagen, type XVI, alpha 1', other_designations='alpha 1 type XVI collagen|collagen XVI, alpha-1 polypeptide', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31652247, end_position_on_the_genomic_accession=31704167, orientation='minus', exon_count=74, OMIM=120326, no_hdr0=''),
  1308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1308, CurrentID=0, Status='live', Symbol='COL17A1', Aliases='BA16H23.2, BP180, BPA-2, BPAG2, ERED, LAD-1', description='collagen, type XVII, alpha 1', other_designations='180 kDa bullous pemphigoid antigen 2|alpha 1 type XVII collagen|bA16H23.2 (collagen, type XVII, alpha 1 (BP180))|bullous pemphigoid antigen 2 (180kD)|collagen XVII, alpha-1 polypeptide|type XVII collagen alpha-1', map_location='10q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104031288, end_position_on_the_genomic_accession=104085880, orientation='minus', exon_count=56, OMIM=113811, no_hdr0=''),
  100271389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271389, CurrentID=0, Status='live', Symbol='RPL21P81', Aliases='RPL21_35_1017', description='ribosomal protein L21 pseudogene 81', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133338950, end_position_on_the_genomic_accession=133339498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1310, CurrentID=0, Status='live', Symbol='COL19A1', Aliases='COL9A1L, D6S228E', description='collagen, type XIX, alpha 1', other_designations='a1 chain of type XIX collagen|collagen XIX, alpha-1 polypeptide|collagen alpha 1 (Y) chain|collagen alpha-1(Y) chain', map_location='6q12-q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=69866556, end_position_on_the_genomic_accession=70212454, orientation='plus', exon_count=53, OMIM=120165, no_hdr0=''),
  1311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1311, CurrentID=0, Status='live', Symbol='COMP', Aliases='EDM1, EPD1, MED, PSACH, THBS5', description='cartilage oligomeric matrix protein', other_designations='TSP5|cartilage oligomeric matrix protein (pseudoachondroplasia, epiphyseal dysplasia 1, multiple)|cartilage oligomeric matrix protein(pseudoachondroplasia, epiphyseal dysplasia 1, multiple)|pseudoachondroplasia (epiphyseal dysplasia 1, multiple)|thrombospondin-5', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18782773, end_position_on_the_genomic_accession=18791305, orientation='minus', exon_count=19, OMIM=600310, no_hdr0=''),
  1312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1312, CurrentID=0, Status='live', Symbol='COMT', Aliases='HEL-S-98n', description='catechol-O-methyltransferase', other_designations='catechol-O-methyltransferase isoform|epididymis secretory sperm binding protein Li 98n', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19941740, end_position_on_the_genomic_accession=19969975, orientation='plus', exon_count=11, OMIM=116790, no_hdr0=''),
  100271393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271393, CurrentID=0, Status='live', Symbol='RPL32P35', Aliases='RPL32_11_1810', description='ribosomal protein L32 pseudogene 35', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128838441, end_position_on_the_genomic_accession=128838841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1314, CurrentID=0, Status='live', Symbol='COPA', Aliases='AILJK, HEP-COP', description='coatomer protein complex subunit alpha', other_designations='HEPCOP|alpha coat protein|alpha-COP|alpha-coat protein|coatomer protein complex, subunit alpha|proxenin|xenin', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160288587, end_position_on_the_genomic_accession=160343564, orientation='minus', exon_count=33, OMIM=601924, no_hdr0=''),
  1315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1315, CurrentID=0, Status='live', Symbol='COPB1', Aliases='COPB', description='coatomer protein complex subunit beta 1', other_designations='beta coat protein|beta-coat protein|beta-cop|coatomer protein complex, subunit beta 1', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14457503, end_position_on_the_genomic_accession=14499895, orientation='minus', exon_count=22, OMIM=600959, no_hdr0=''),
  1316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1316, CurrentID=0, Status='live', Symbol='KLF6', Aliases='BCD1, CBA1, COPEB, CPBP, GBF, PAC1, ST12, ZF9', description='Kruppel-like factor 6', other_designations='B-cell-derived protein 1|GC-rich binding factor|GC-rich sites-binding factor GBF|Kruppel-like zinc finger protein Zf9|core promoter element binding protein|core promoter element-binding protein|proto-oncogene BCD1|protooncogene B-cell derived 1|suppression of tumorigenicity 12 (prostate)|suppressor of tumorigenicity 12 protein|transcription factor Zf9', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3775996, end_position_on_the_genomic_accession=3785281, orientation='minus', exon_count=4, OMIM=602053, no_hdr0=''),
  1317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1317, CurrentID=0, Status='live', Symbol='SLC31A1', Aliases='COPT1, CTR1', description='solute carrier family 31 (copper transporter), member 1', other_designations='copper transport 1 homolog|copper transporter 1|solute carrier family 31 (copper transporters), member 1|solute carrier family 31 member 1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113221528, end_position_on_the_genomic_accession=113264492, orientation='plus', exon_count=5, OMIM=603085, no_hdr0=''),
  1318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1318, CurrentID=0, Status='live', Symbol='SLC31A2', Aliases='COPT2, CTR2, hCTR2', description='solute carrier family 31 (copper transporter), member 2', other_designations='copper transporter 2|solute carrier family 31 (copper transporters), member 2|solute carrier family 31 member 2', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113150958, end_position_on_the_genomic_accession=113164142, orientation='plus', exon_count=6, OMIM=603088, no_hdr0=''),
  1319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1319, CurrentID=0, Status='live', Symbol='CORD1', Aliases='CORD, CRD1', description='cone rod dystrophy 1 (autosomal dominant)', other_designations='', map_location='18q21.1-q21.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600624, no_hdr0=''),
  100271400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271400, CurrentID=0, Status='live', Symbol='RPS26P56', Aliases='RPS26_23_1811', description='ribosomal protein S26 pseudogene 56', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129408323, end_position_on_the_genomic_accession=129408621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1321, CurrentID=0, Status='live', Symbol='CORD4', Aliases='', description='cone rod dystrophy 4', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271402, CurrentID=0, Status='live', Symbol='RPS29P28', Aliases='RPS29_22_1795', description='ribosomal protein S29 pseudogene 28', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96195477, end_position_on_the_genomic_accession=96195617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271404, CurrentID=0, Status='live', Symbol='RPL12P26', Aliases='RPL12_10_1101', description='ribosomal protein L12 pseudogene 26', other_designations='', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117196125, end_position_on_the_genomic_accession=117196732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1325, CurrentID=0, Status='live', Symbol='CORT', Aliases='CST-14, CST-17, CST-29', description='cortistatin', other_designations='cortistatin-14|cortistatin-17|cortistatin-29|prepro-cortistatin|preprocortistatin', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10449719, end_position_on_the_genomic_accession=10452003, orientation='plus', exon_count=2, OMIM=602784, no_hdr0=''),
  1326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1326, CurrentID=0, Status='live', Symbol='MAP3K8', Aliases='AURA2, COT, EST, ESTF, MEKK8, TPL2, Tpl-2, c-COT', description='mitogen-activated protein kinase kinase kinase 8', other_designations='Ewing sarcoma transformant|augmented in rheumatoid arthritis 2|cot (cancer Osaka thyroid) oncogene|proto-oncogene c-Cot|proto-oncogene serine/threoine protein kinase|tumor progression locus 2', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30434003, end_position_on_the_genomic_accession=30461833, orientation='plus', exon_count=13, OMIM=191195, no_hdr0=''),
  1327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1327, CurrentID=0, Status='live', Symbol='COX4I1', Aliases='COX4, COX4-1, COXIV', description='cytochrome c oxidase subunit IV isoform 1', other_designations='COX IV-1|cytochrome c oxidase polypeptide IV', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85799567, end_position_on_the_genomic_accession=85807002, orientation='plus', exon_count=6, OMIM=123864, no_hdr0=''),
  100271408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271408, CurrentID=0, Status='live', Symbol='RPL12P36', Aliases='RPL12_17_1486', description='ribosomal protein L12 pseudogene 36', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58825897, end_position_on_the_genomic_accession=58826346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1329, CurrentID=0, Status='live', Symbol='COX5B', Aliases='COXVB', description='cytochrome c oxidase subunit Vb', other_designations='cytochrome c oxidase polypeptide VB, mitochondrial', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97646058, end_position_on_the_genomic_accession=97648194, orientation='plus', exon_count=4, OMIM=123866, no_hdr0=''),
  1330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1330, CurrentID=0, Status='live', Symbol='COX5BP1', Aliases='COX5BL1', description='cytochrome c oxidase subunit Vb pseudogene 1', other_designations='cytochrome c oxidase subunit Vb-like 1', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81919922, end_position_on_the_genomic_accession=81920400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271411, CurrentID=0, Status='live', Symbol='RPL17P37', Aliases='RPL17_12_1285', description='ribosomal protein L17 pseudogene 37', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118737920, end_position_on_the_genomic_accession=118738536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1332, CurrentID=0, Status='live', Symbol='COX5BP3', Aliases='COX5BL3', description='cytochrome c oxidase subunit Vb pseudogene 3', other_designations='cytochrome c oxidase subunit Vb-like 3', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133727280, end_position_on_the_genomic_accession=133727768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271413, CurrentID=0, Status='live', Symbol='RPL17P40', Aliases='RPL17_15_1477', description='ribosomal protein L17 pseudogene 40', other_designations='', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16094033, end_position_on_the_genomic_accession=16094513, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271414, CurrentID=0, Status='live', Symbol='RPL17P45', Aliases='RPL17_16_1598', description='ribosomal protein L17 pseudogene 45', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=40143926, end_position_on_the_genomic_accession=40144504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271415, CurrentID=0, Status='live', Symbol='RPL17P47', Aliases='RPL17_17_1624', description='ribosomal protein L17 pseudogene 47', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12142381, end_position_on_the_genomic_accession=12142875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271416, CurrentID=0, Status='live', Symbol='RPL17P48', Aliases='RPL17_19_1717', description='ribosomal protein L17 pseudogene 48', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62037395, end_position_on_the_genomic_accession=62037987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1337, CurrentID=0, Status='live', Symbol='COX6A1', Aliases='CMTRID, COX6A, COX6AL', description='cytochrome c oxidase subunit VIa polypeptide 1', other_designations='COX VIa-L|cytochrome C oxidase subunit VIa homolog|cytochrome c oxidase polypeptide VIa-liver|cytochrome c oxidase subunit VIA-liver', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120438090, end_position_on_the_genomic_accession=120440742, orientation='plus', exon_count=3, OMIM=602072, no_hdr0=''),
  1338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1338, CurrentID=0, Status='live', Symbol='COX6A1P1', Aliases='COX6A1P', description='cytochrome c oxidase subunit VIa polypeptide 1 pseudogene 1', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=80495889, end_position_on_the_genomic_accession=80496422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1339, CurrentID=0, Status='live', Symbol='COX6A2', Aliases='COX6AH, COXVIAH', description='cytochrome c oxidase subunit VIa polypeptide 2', other_designations='COX VIa-M|cytochrome c oxidase polypeptide VIa-heart|cytochrome c oxidase subunit VIA-muscle', map_location='16p11.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31427731, end_position_on_the_genomic_accession=31428428, orientation='minus', exon_count=3, OMIM=602009, no_hdr0=''),
  1340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1340, CurrentID=0, Status='live', Symbol='COX6B1', Aliases='COX6B, COXG, COXVIb1', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous)', other_designations='COX VIb-1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35648223, end_position_on_the_genomic_accession=35658784, orientation='plus', exon_count=4, OMIM=124089, no_hdr0=''),
  100271421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271421, CurrentID=0, Status='live', Symbol='RPL21P88', Aliases='RPL21_36_1036', description='ribosomal protein L21 pseudogene 88', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43630947, end_position_on_the_genomic_accession=43631414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1342, CurrentID=0, Status='live', Symbol='COX6B1P2', Aliases='COX6BP-2, COX6BP2', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 2', other_designations='cytochrome c oxidase subunit VIb pseudogene 2', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48713647, end_position_on_the_genomic_accession=48713868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1343, CurrentID=0, Status='live', Symbol='COX6B1P3', Aliases='COX6BP-3, COX6BP3', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 3', other_designations='cytochrome c oxidase subunit VIb pseudogene 3', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40569189, end_position_on_the_genomic_accession=40569681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271424, CurrentID=0, Status='live', Symbol='RPL21P91', Aliases='RPL21_39_1094', description='ribosomal protein L21 pseudogene 91', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109807611, end_position_on_the_genomic_accession=109808167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1345, CurrentID=0, Status='live', Symbol='COX6C', Aliases='', description='cytochrome c oxidase subunit VIc', other_designations='cytochrome c oxidase polypeptide VIc|cytochrome c oxidase subunit VIc preprotein', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99877995, end_position_on_the_genomic_accession=99894014, orientation='minus', exon_count=4, OMIM=124090, no_hdr0=''),
  1346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1346, CurrentID=0, Status='live', Symbol='COX7A1', Aliases='COX7A, COX7AH, COX7AM', description='cytochrome c oxidase subunit VIIa polypeptide 1 (muscle)', other_designations='cytochrome c oxidase subunit VIIa heart/muscle isoform|cytochrome c oxidase subunit VIIa-H|cytochrome c oxidase subunit VIIa-M|cytochrome c oxidase subunit VIIa-heart|cytochrome c oxidase subunit VIIa-muscle', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36150922, end_position_on_the_genomic_accession=36152869, orientation='minus', exon_count=4, OMIM=123995, no_hdr0=''),
  1347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1347, CurrentID=0, Status='live', Symbol='COX7A2', Aliases='COX7AL, COX7AL1, COXVIIAL, COXVIIa-L, VIIAL', description='cytochrome c oxidase subunit VIIa polypeptide 2 (liver)', other_designations='cytochrome c oxidase polypeptide 7A2, mitochondrial|cytochrome c oxidase polypeptide VIIa-liver/heart|cytochrome c oxidase subunit VIIa-L|cytochrome c oxidase subunit VIIa-liver/heart|cytochrome c oxidase subunit VIIaL|hepatic cytochrome-c oxidase chain VIIa', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75237675, end_position_on_the_genomic_accession=75254885, orientation='minus', exon_count=5, OMIM=123996, no_hdr0=''),
  1348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1348, CurrentID=0, Status='live', Symbol='COX7A2P2', Aliases='COX7A3, COX7AL, COX7AL2, COX7AP2', description='cytochrome c oxidase subunit VIIa polypeptide 2 (liver) pseudogene 2', other_designations='H_RG162B04.1|WUGSC:H_RG162B04.1|cytochrome c oxidase subunit VIIa polypeptide 3 (liver)|cytochrome c oxidase subunit VIIa pseudogene 2', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=96902763, end_position_on_the_genomic_accession=96903189, orientation='plus', exon_count=0, OMIM=123997, no_hdr0=''),
  1349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1349, CurrentID=0, Status='live', Symbol='COX7B', Aliases='APLCC, LSDMCA2', description='cytochrome c oxidase subunit VIIb', other_designations='cytochrome c oxidase polypeptide VIIb|cytochrome-c oxidase chain VIIb', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77899464, end_position_on_the_genomic_accession=77905384, orientation='plus', exon_count=3, OMIM=300885, no_hdr0=''),
  1350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1350, CurrentID=0, Status='live', Symbol='COX7C', Aliases='', description='cytochrome c oxidase subunit VIIc', other_designations='cytochrome c oxidase polypeptide VIIc|cytochrome-c oxidase chain VIIc', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86617967, end_position_on_the_genomic_accession=86620766, orientation='plus', exon_count=3, OMIM=603774, no_hdr0=''),
  1351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1351, CurrentID=0, Status='live', Symbol='COX8A', Aliases='COX, COX8, COX8-2, COX8L, VIII, VIII-L', description='cytochrome c oxidase subunit VIIIA (ubiquitous)', other_designations='cytochrome c oxidase polypeptide VIII-liver/heart|cytochrome c oxidase subunit 8-2|cytochrome c oxidase subunit 8A (ubiquitous)|cytochrome c oxidase subunit VIII', map_location='11q12-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63974607, end_position_on_the_genomic_accession=63976543, orientation='plus', exon_count=2, OMIM=123870, no_hdr0=''),
  1352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1352, CurrentID=0, Status='live', Symbol='COX10', Aliases='', description='COX10 heme A:farnesyltransferase cytochrome c oxidase assembly factor', other_designations='COX10 homolog, cytochrome c oxidase assembly protein, heme A: farnesyltransferase|cytochrome c oxidase assembly homolog 10|cytochrome c oxidase assembly protein|cytochrome c oxidase subunit X|heme A: farnesyltransferase|heme O synthase', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14069402, end_position_on_the_genomic_accession=14217922, orientation='plus', exon_count=12, OMIM=602125, no_hdr0=''),
  1353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1353, CurrentID=0, Status='live', Symbol='COX11', Aliases='COX11P', description='COX11 cytochrome c oxidase copper chaperone', other_designations='COX11 cytochrome c oxidase assembly homolog|COX11 homolog, cytochrome c oxidase assembly protein|cytochrome c oxidase assembly homolog 11|cytochrome c oxidase subunit 11', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54951898, end_position_on_the_genomic_accession=54968780, orientation='minus', exon_count=7, OMIM=603648, no_hdr0=''),
  100271434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271434, CurrentID=0, Status='live', Symbol='RPL21P108', Aliases='RPL21_44_1324', description='ribosomal protein L21 pseudogene 108', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74377545, end_position_on_the_genomic_accession=74377991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1355, CurrentID=0, Status='live', Symbol='COX15', Aliases='CEMCOX2', description='cytochrome c oxidase assembly homolog 15 (yeast)', other_designations='COX15 homolog, cytochrome c oxidase assembly protein|cytochrome c oxidase subunit 15', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99696129, end_position_on_the_genomic_accession=99732666, orientation='minus', exon_count=10, OMIM=603646, no_hdr0=''),
  1356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1356, CurrentID=0, Status='live', Symbol='CP', Aliases='CP-2', description='ceruloplasmin (ferroxidase)', other_designations='', map_location='3q23-q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149162410, end_position_on_the_genomic_accession=149222050, orientation='minus', exon_count=23, OMIM=117700, no_hdr0=''),
  1357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1357, CurrentID=0, Status='live', Symbol='CPA1', Aliases='CPA', description='carboxypeptidase A1', other_designations='carboxypeptidase A1 (pancreatic)|pancreatic carboxypeptidase A', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130380371, end_position_on_the_genomic_accession=130388108, orientation='plus', exon_count=10, OMIM=114850, no_hdr0=''),
  1358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1358, CurrentID=0, Status='live', Symbol='CPA2', Aliases='', description='carboxypeptidase A2', other_designations='carboxypeptidase A2 (pancreatic)', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130266863, end_position_on_the_genomic_accession=130289798, orientation='plus', exon_count=11, OMIM=600688, no_hdr0=''),
  1359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1359, CurrentID=0, Status='live', Symbol='CPA3', Aliases='MC-CPA', description='carboxypeptidase A3', other_designations='carboxypeptidase A3 (mast cell)|tissue carboxypeptidase A', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148865256, end_position_on_the_genomic_accession=148897087, orientation='plus', exon_count=11, OMIM=114851, no_hdr0=''),
  1360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1360, CurrentID=0, Status='live', Symbol='CPB1', Aliases='CPB, PASP, PCPB', description='carboxypeptidase B1', other_designations='carboxypeptidase B1 (tissue)|pancreas-specific protein|pancreatic carboxypeptidase B|procarboxypeptidase B|protaminase|tissue carboxypeptidase B', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148791180, end_position_on_the_genomic_accession=148860192, orientation='plus', exon_count=13, OMIM=114852, no_hdr0=''),
  1361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1361, CurrentID=0, Status='live', Symbol='CPB2', Aliases='CPU, PCPB, TAFI', description='carboxypeptidase B2', other_designations='carboxypeptidase B-like protein|carboxypeptidase B2 (plasma)|carboxypeptidase B2 (plasma, carboxypeptidase U)|carboxypeptidase R|thrombin-activable fibrinolysis inhibitor|thrombin-activatable fibrinolysis inhibitor', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46053187, end_position_on_the_genomic_accession=46105076, orientation='minus', exon_count=11, OMIM=603101, no_hdr0=''),
  1362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1362, CurrentID=0, Status='live', Symbol='CPD', Aliases='GP180', description='carboxypeptidase D', other_designations='metallocarboxypeptidase D', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30378910, end_position_on_the_genomic_accession=30469657, orientation='plus', exon_count=22, OMIM=603102, no_hdr0=''),
  1363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1363, CurrentID=0, Status='live', Symbol='CPE', Aliases='CPH', description='carboxypeptidase E', other_designations='carboxypeptidase H|cobalt-stimulated chromaffin granule carboxypeptidase|enkephalin convertase|insulin granule-associated carboxypeptidase|prohormone-processing carboxypeptidase', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165378945, end_position_on_the_genomic_accession=165498330, orientation='plus', exon_count=9, OMIM=114855, no_hdr0=''),
  1364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1364, CurrentID=0, Status='live', Symbol='CLDN4', Aliases='CPE-R, CPER, CPETR, CPETR1, WBSCR8, hCPE-R', description='claudin 4', other_designations='CPE-receptor|Clostridium perfringens enterotoxin receptor 1|Williams-Beuren syndrome chromosomal region 8 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73830863, end_position_on_the_genomic_accession=73832693, orientation='plus', exon_count=1, OMIM=602909, no_hdr0=''),
  1365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1365, CurrentID=0, Status='live', Symbol='CLDN3', Aliases='C7orf1, CPE-R2, CPETR2, HRVP1, RVP1', description='claudin 3', other_designations='CPE-R 2|CPE-receptor 2|Clostridium perfringens enterotoxin receptor 2|ventral prostate.1 protein homolog|ventral prostate.1-like protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73768997, end_position_on_the_genomic_accession=73770270, orientation='minus', exon_count=1, OMIM=602910, no_hdr0=''),
  1366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1366, CurrentID=0, Status='live', Symbol='CLDN7', Aliases='CEPTRL2, CLDN-7, CPETRL2, Hs.84359, claudin-1', description='claudin 7', other_designations='clostridium perfringens enterotoxin receptor-like 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7259903, end_position_on_the_genomic_accession=7263193, orientation='minus', exon_count=5, OMIM=609131, no_hdr0=''),
  100271447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271447, CurrentID=0, Status='live', Symbol='RPL21P122', Aliases='RPL21_58_1515', description='ribosomal protein L21 pseudogene 122', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12010716, end_position_on_the_genomic_accession=12011254, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1368, CurrentID=0, Status='live', Symbol='CPM', Aliases='', description='carboxypeptidase M', other_designations='renal carboxypeptidase|urinary carboxypeptidase B', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68851175, end_position_on_the_genomic_accession=68963240, orientation='minus', exon_count=10, OMIM=114860, no_hdr0=''),
  1369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1369, CurrentID=0, Status='live', Symbol='CPN1', Aliases='CPN, SCPN', description='carboxypeptidase N subunit 1', other_designations='anaphylatoxin inactivator|arginine carboxypeptidase|carboxypeptidase K|carboxypeptidase N catalytic subunit|carboxypeptidase N polypeptide 1 50 kD|carboxypeptidase N small subunit|carboxypeptidase N, polypeptide 1|kininase I|kininase-1|lysine carboxypeptidase|plasma carboxypeptidase B|serum carboxypeptidase N', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100042308, end_position_on_the_genomic_accession=100081901, orientation='minus', exon_count=9, OMIM=603103, no_hdr0=''),
  1370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1370, CurrentID=0, Status='live', Symbol='CPN2', Aliases='ACBP', description='carboxypeptidase N subunit 2', other_designations='arginine carboxypeptidase|carboxypeptidase N 83 kDa chain|carboxypeptidase N large subunit|carboxypeptidase N regulatory subunit|carboxypeptidase N, polypeptide 2|carboxypeptidase N, polypeptide 2, 83kD', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194339765, end_position_on_the_genomic_accession=194351345, orientation='minus', exon_count=3, OMIM=603104, no_hdr0=''),
  1371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1371, CurrentID=0, Status='live', Symbol='CPOX', Aliases='CPO, CPX, HCP', description='coproporphyrinogen oxidase', other_designations='COX|coprogen oxidase|coproporphyrinogen-III oxidase, mitochondrial|coproporphyrinogenase', map_location='3q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98576414, end_position_on_the_genomic_accession=98593668, orientation='minus', exon_count=9, OMIM=612732, no_hdr0=''),
  100271452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271452, CurrentID=0, Status='live', Symbol='RPL21P127', Aliases='RPL21_61_1581', description='ribosomal protein L21 pseudogene 127', other_designations='', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3406310, end_position_on_the_genomic_accession=3406858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1373, CurrentID=0, Status='live', Symbol='CPS1', Aliases='CPSASE1, PHN', description='carbamoyl-phosphate synthase 1', other_designations='carbamoyl-phosphate synthase (ammonia)|carbamoyl-phosphate synthase 1, mitochondrial|carbamoylphosphate synthetase I', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210477682, end_position_on_the_genomic_accession=210679107, orientation='plus', exon_count=43, OMIM=608307, no_hdr0=''),
  1374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1374, CurrentID=0, Status='live', Symbol='CPT1A', Aliases='CPT1, CPT1-L, L-CPT1', description='carnitine palmitoyltransferase 1A (liver)', other_designations='CPT I|CPTI-L|carnitine O-palmitoyltransferase I, liver isoform|carnitine palmitoyltransferase I, liver', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68754620, end_position_on_the_genomic_accession=68841931, orientation='minus', exon_count=21, OMIM=600528, no_hdr0=''),
  1375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1375, CurrentID=0, Status='live', Symbol='CPT1B', Aliases='CPT1-M, CPT1M, CPTI, CPTI-M, M-CPT1, MCCPT1, MCPT1', description='carnitine palmitoyltransferase 1B', other_designations='carnitine O-palmitoyltransferase I, mitochondrial muscle isoform|carnitine palmitoyltransferase 1B (muscle)|carnitine palmitoyltransferase I-like protein', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50568861, end_position_on_the_genomic_accession=50578667, orientation='minus', exon_count=21, OMIM=601987, no_hdr0=''),
  1376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1376, CurrentID=0, Status='live', Symbol='CPT2', Aliases='CPT1, CPTASE, IIAE4', description='carnitine palmitoyltransferase 2', other_designations='CPT II|carnitine palmitoyltransferase II', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53196429, end_position_on_the_genomic_accession=53214197, orientation='plus', exon_count=5, OMIM=600650, no_hdr0=''),
  100271457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271457, CurrentID=0, Status='live', Symbol='RPL23AP28', Aliases='RPL23A_10_329', description='ribosomal protein L23a pseudogene 28', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222044525, end_position_on_the_genomic_accession=222044994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1378, CurrentID=0, Status='live', Symbol='CR1', Aliases='C3BR, C4BR, CD35, KN', description='complement component (3b/4b) receptor 1 (Knops blood group)', other_designations='C3-binding protein|C3b/C4b receptor|CD35 antigen|Knops blood group antigen', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207496128, end_position_on_the_genomic_accession=207641765, orientation='plus', exon_count=48, OMIM=120620, no_hdr0=''),
  1379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1379, CurrentID=0, Status='live', Symbol='CR1L', Aliases='', description='complement component (3b/4b) receptor 1-like', other_designations='complement C4b-binding protein CR-1-like protein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207645113, end_position_on_the_genomic_accession=207723691, orientation='plus', exon_count=12, OMIM=605886, no_hdr0=''),
  1380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1380, CurrentID=0, Status='live', Symbol='CR2', Aliases='C3DR, CD21, CR, CVID7, SLEB9', description='complement component (3d/Epstein Barr virus) receptor 2', other_designations='EBV receptor|complement C3d receptor', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207454300, end_position_on_the_genomic_accession=207489897, orientation='plus', exon_count=21, OMIM=120650, no_hdr0=''),
  1381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1381, CurrentID=0, Status='live', Symbol='CRABP1', Aliases='CRABP, CRABP-I, CRABPI, RBP5', description='cellular retinoic acid binding protein 1', other_designations='cellular retinoic acid-binding protein I', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78340324, end_position_on_the_genomic_accession=78348230, orientation='plus', exon_count=4, OMIM=180230, no_hdr0=''),
  1382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1382, CurrentID=0, Status='live', Symbol='CRABP2', Aliases='CRABP-II, RBP6', description='cellular retinoic acid binding protein 2', other_designations='cellular retinoic acid-binding protein II', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156699606, end_position_on_the_genomic_accession=156705816, orientation='minus', exon_count=5, OMIM=180231, no_hdr0=''),
  1384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1384, CurrentID=0, Status='live', Symbol='CRAT', Aliases='CAT1', description='carnitine O-acetyltransferase', other_designations='CAT|carnitine acetylase|carnitine acetyltransferase', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129094794, end_position_on_the_genomic_accession=129110791, orientation='minus', exon_count=17, OMIM=600184, no_hdr0=''),
  1385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1385, CurrentID=0, Status='live', Symbol='CREB1', Aliases='CREB', description='cAMP responsive element binding protein 1', other_designations='CREB-1|active transcription factor CREB|cAMP-response element-binding protein-1|cAMP-responsive element-binding protein 1|transactivator protein', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207529892, end_position_on_the_genomic_accession=207605560, orientation='plus', exon_count=15, OMIM=123810, no_hdr0=''),
  1386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1386, CurrentID=0, Status='live', Symbol='ATF2', Aliases='CRE-BP1, CREB-2, CREB2, HB16, TREB7', description='activating transcription factor 2', other_designations='activating transcription factor 2 splice variant ATF2-var2|cAMP response element-binding protein CRE-BP1|cAMP responsive element binding protein 2, formerly|cAMP-dependent transcription factor ATF-2|cAMP-responsive element-binding protein 2|cyclic AMP-responsive element-binding protein 2|histone acetyltransferase ATF2', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175072250, end_position_on_the_genomic_accession=175168206, orientation='minus', exon_count=17, OMIM=123811, no_hdr0=''),
  1387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1387, CurrentID=0, Status='live', Symbol='CREBBP', Aliases='CBP, KAT3A, RSTS', description='CREB binding protein', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3725054, end_position_on_the_genomic_accession=3880120, orientation='minus', exon_count=34, OMIM=600140, no_hdr0=''),
  1388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1388, CurrentID=0, Status='live', Symbol='ATF6B', Aliases='CREB-RP, CREBL1, G13', description='activating transcription factor 6 beta', other_designations='Creb-related protein|cAMP response element-binding protein-related protein|cAMP responsive element binding protein-like 1|cAMP-dependent transcription factor ATF-6 beta|cAMP-responsive element-binding protein-like 1|protein G13', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32115268, end_position_on_the_genomic_accession=32128240, orientation='minus', exon_count=18, OMIM=600984, no_hdr0=''),
  1389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1389, CurrentID=0, Status='live', Symbol='CREBL2', Aliases='', description='cAMP responsive element binding protein-like 2', other_designations='MHBs-binding protein 1', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12611833, end_position_on_the_genomic_accession=12645108, orientation='plus', exon_count=4, OMIM=603476, no_hdr0=''),
  1390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1390, CurrentID=0, Status='live', Symbol='CREM', Aliases='CREM-2, ICER, hCREM-2', description='cAMP responsive element modulator', other_designations='CREM 2alpha-b protein|CREM 2beta-a protein|cAMP response element modulator|inducible cAMP early repressor ICER', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35126841, end_position_on_the_genomic_accession=35212958, orientation='plus', exon_count=18, OMIM=123812, no_hdr0=''),
  100271471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271471, CurrentID=0, Status='live', Symbol='RPL26P27', Aliases='RPL26_15_1055', description='ribosomal protein L26 pseudogene 27', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68499488, end_position_on_the_genomic_accession=68500009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1392, CurrentID=0, Status='live', Symbol='CRH', Aliases='CRF1, CRH', description='corticotropin releasing hormone', other_designations='corticotropin-releasing factor', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66176377, end_position_on_the_genomic_accession=66178611, orientation='minus', exon_count=2, OMIM=122560, no_hdr0=''),
  1393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1393, CurrentID=0, Status='live', Symbol='CRHBP', Aliases='CRF-BP, CRFBP', description='corticotropin releasing hormone binding protein', other_designations='CRF-binding protein|CRH-BP|corticotropin releasing hormone-binding protein|corticotropin-releasing hormone-binding protein', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76952835, end_position_on_the_genomic_accession=76981147, orientation='plus', exon_count=9, OMIM=122559, no_hdr0=''),
  1394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1394, CurrentID=0, Status='live', Symbol='CRHR1', Aliases='CRF-R, CRF-R-1, CRF-R1, CRF1, CRFR-1, CRFR1, CRH-R-1, CRH-R1, CRHRL, CRHR1', description='corticotropin releasing hormone receptor 1', other_designations='CRH receptor 1|corticotropin-releasing factor type 1 receptor|seven transmembrane helix receptor', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45784280, end_position_on_the_genomic_accession=45835828, orientation='plus', exon_count=14, OMIM=122561, no_hdr0=''),
  1395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1395, CurrentID=0, Status='live', Symbol='CRHR2', Aliases='CRF-RB, CRF2, CRFR2, HM-CRF', description='corticotropin releasing hormone receptor 2', other_designations='CRF2 receptor, beta isoform|CRH receptor 2 variant B|CRH-R2', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30651843, end_position_on_the_genomic_accession=30700103, orientation='minus', exon_count=16, OMIM=602034, no_hdr0=''),
  1396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1396, CurrentID=0, Status='live', Symbol='CRIP1', Aliases='CRHP, CRIP, CRP-1, CRP1', description='cysteine-rich protein 1 (intestinal)', other_designations='cysteine-rich heart protein|cysteine-rich intestinal protein', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105486920, end_position_on_the_genomic_accession=105488789, orientation='plus', exon_count=6, OMIM=123875, no_hdr0=''),
  1397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1397, CurrentID=0, Status='live', Symbol='CRIP2', Aliases='CRIP, CRP2, ESP1', description='cysteine-rich protein 2', other_designations='Cysteine-rich intestinal protein|LIM domain protein ESP1/CRP2', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105472938, end_position_on_the_genomic_accession=105480170, orientation='plus', exon_count=11, OMIM=601183, no_hdr0=''),
  1398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1398, CurrentID=0, Status='live', Symbol='CRK', Aliases='CRKII, p38', description='v-crk avian sarcoma virus CT10 oncogene homolog', other_designations='proto-oncogene c-Crk|v-crk sarcoma virus CT10 oncogene-like protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1421353, end_position_on_the_genomic_accession=1456267, orientation='minus', exon_count=3, OMIM=164762, no_hdr0=''),
  1399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1399, CurrentID=0, Status='live', Symbol='CRKL', Aliases='', description='v-crk avian sarcoma virus CT10 oncogene homolog-like', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20917398, end_position_on_the_genomic_accession=20953749, orientation='plus', exon_count=4, OMIM=602007, no_hdr0=''),
  1400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1400, CurrentID=0, Status='live', Symbol='CRMP1', Aliases='CRMP-1, DPYSL1, DRP-1, DRP1, ULIP-3', description='collapsin response mediator protein 1', other_designations='dihydropyrimidinase-like 1|unc-33-like phosphoprotein 3', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5820764, end_position_on_the_genomic_accession=5893083, orientation='minus', exon_count=17, OMIM=602462, no_hdr0=''),
  1401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1401, CurrentID=0, Status='live', Symbol='CRP', Aliases='PTX1', description='C-reactive protein, pentraxin-related', other_designations='pentraxin 1', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159712289, end_position_on_the_genomic_accession=159714609, orientation='minus', exon_count=3, OMIM=123260, no_hdr0=''),
  100271482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271482, CurrentID=0, Status='live', Symbol='RPL31P62', Aliases='RPL31_26_1736', description='ribosomal protein L31 pseudogene 62', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17093615, end_position_on_the_genomic_accession=17093937, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271483, CurrentID=0, Status='live', Symbol='RPL31P54', Aliases='RPL31_27_1326', description='ribosomal protein L31 pseudogene 54', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78422823, end_position_on_the_genomic_accession=78423174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1404, CurrentID=0, Status='live', Symbol='HAPLN1', Aliases='CRT1, CRTL1', description='hyaluronan and proteoglycan link protein 1', other_designations='Cartilage link protein|cartilage linking protein 1|cartilage-link protein|cartilage-linking protein 1|proteoglycan link protein', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83638198, end_position_on_the_genomic_accession=83721077, orientation='minus', exon_count=7, OMIM=115435, no_hdr0=''),
  100271485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271485, CurrentID=0, Status='live', Symbol='RPL31P58', Aliases='RPL31_32_1533', description='ribosomal protein L31 pseudogene 58', other_designations='', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28802427, end_position_on_the_genomic_accession=28802856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1406, CurrentID=0, Status='live', Symbol='CRX', Aliases='CORD2, CRD, LCA7, OTX3', description='cone-rod homeobox', other_designations='orthodenticle homeobox 3', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47821842, end_position_on_the_genomic_accession=47843330, orientation='plus', exon_count=4, OMIM=602225, no_hdr0=''),
  1407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1407, CurrentID=0, Status='live', Symbol='CRY1', Aliases='PHLL1', description='cryptochrome circadian clock 1', other_designations='cryptochrome 1 (photolyase-like)', map_location='12q23-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106991364, end_position_on_the_genomic_accession=107093857, orientation='minus', exon_count=14, OMIM=601933, no_hdr0=''),
  1408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1408, CurrentID=0, Status='live', Symbol='CRY2', Aliases='HCRY2, PHLL2', description='cryptochrome circadian clock 2', other_designations='cryptochrome 2 (photolyase-like)|growth-inhibiting protein 37', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45847118, end_position_on_the_genomic_accession=45883248, orientation='plus', exon_count=13, OMIM=603732, no_hdr0=''),
  1409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1409, CurrentID=0, Status='live', Symbol='CRYAA', Aliases='CRYA1, CTRCT9, HSPB4', description='crystallin alpha A', other_designations='crystallin, alpha A|crystallin, alpha-1|heat shock protein beta-4|human alphaA-crystallin (CRYA1)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43169008, end_position_on_the_genomic_accession=43172810, orientation='plus', exon_count=4, OMIM=123580, no_hdr0=''),
  1410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1410, CurrentID=0, Status='live', Symbol='CRYAB', Aliases='CMD1II, CRYA2, CTPP2, CTRCT16, HEL-S-101, HSPB5, MFM2', description='crystallin alpha B', other_designations='crystallin, alpha B|epididymis secretory protein Li 101|heat shock protein beta-5|heat-shock 20 kD like-protein|renal carcinoma antigen NY-REN-27|rosenthal fiber component', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111908620, end_position_on_the_genomic_accession=111913213, orientation='minus', exon_count=6, OMIM=123590, no_hdr0=''),
  1411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1411, CurrentID=0, Status='live', Symbol='CRYBA1', Aliases='CRYB1, CTRCT10', description='crystallin beta A1', other_designations='beta crystallin A3 chain transcript CN|beta crystallin A3 chain transcript LAM|beta crystallin A3 chain transcript PS|beta crystallin A3 chain transcript TC|crystallin beta A3|crystallin beta A3/A1|crystallin, beta A1|crystallin, beta A3|eye lens structural protein|truncated beta-crystallin A3', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29246857, end_position_on_the_genomic_accession=29254494, orientation='plus', exon_count=6, OMIM=123610, no_hdr0=''),
  1412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1412, CurrentID=0, Status='live', Symbol='CRYBA2', Aliases='CTRCT42', description='crystallin beta A2', other_designations='crystallin, beta A2|eye lens structural protein', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218990190, end_position_on_the_genomic_accession=218993556, orientation='minus', exon_count=6, OMIM=600836, no_hdr0=''),
  1413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1413, CurrentID=0, Status='live', Symbol='CRYBA4', Aliases='CTRCT23, MCOPCT4', description='crystallin beta A4', other_designations='beta crystallin A4 chain transcript PS|beta-A4 crystallin|crystallin, beta A4|crystallin, beta polypeptide A4|eye lens structural protein', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26607987, end_position_on_the_genomic_accession=26630672, orientation='plus', exon_count=7, OMIM=123631, no_hdr0=''),
  1414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1414, CurrentID=0, Status='live', Symbol='CRYBB1', Aliases='CATCN3, CTRCT17', description='crystallin beta B1', other_designations='beta-B1 crystallin|crystallin, beta B1|eye lens structural protein', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26599278, end_position_on_the_genomic_accession=26618087, orientation='minus', exon_count=6, OMIM=600929, no_hdr0=''),
  1415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1415, CurrentID=0, Status='live', Symbol='CRYBB2', Aliases='CCA2, CRYB2, CRYB2A, CTRCT3, D22S665', description='crystallin beta B2', other_designations='CTA-221G9.7|beta-B2 crystallin|beta-crystallin Bp|crystallin, beta B2|eye lens structural protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25211670, end_position_on_the_genomic_accession=25231869, orientation='plus', exon_count=7, OMIM=123620, no_hdr0=''),
  1416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1416, CurrentID=0, Status='live', Symbol='CRYBB2P1', Aliases='CRYB2B', description='crystallin beta B2 pseudogene 1', other_designations='crystallin, beta B2 pseudogene 1', map_location='22q11.2-q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25448087, end_position_on_the_genomic_accession=25461678, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  1417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1417, CurrentID=0, Status='live', Symbol='CRYBB3', Aliases='CATCN2, CRYB3, CTRCT22', description='crystallin beta B3', other_designations='crystallin, beta B3|eye lens structural protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25198686, end_position_on_the_genomic_accession=25207363, orientation='plus', exon_count=6, OMIM=123630, no_hdr0=''),
  1418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1418, CurrentID=0, Status='live', Symbol='CRYGA', Aliases='CRY-g-A, CRYG1, CRYG5', description='crystallin gamma A', other_designations='crystallin, gamma 1|crystallin, gamma A|gamma-crystallin 5', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208160740, end_position_on_the_genomic_accession=208163573, orientation='minus', exon_count=3, OMIM=123660, no_hdr0=''),
  1419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1419, CurrentID=0, Status='live', Symbol='CRYGB', Aliases='CRYG2, CTRCT39', description='crystallin gamma B', other_designations='crystallin, gamma 1-2|crystallin, gamma B', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208142573, end_position_on_the_genomic_accession=208146153, orientation='minus', exon_count=3, OMIM=123670, no_hdr0=''),
  1420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1420, CurrentID=0, Status='live', Symbol='CRYGC', Aliases='CCL, CRYG3, CTRCT2', description='crystallin gamma C', other_designations='crystallin, gamma C|gamma-crystallin 2-1|gamma-crystallin 3', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208128137, end_position_on_the_genomic_accession=208139738, orientation='minus', exon_count=5, OMIM=123680, no_hdr0=''),
  1421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1421, CurrentID=0, Status='live', Symbol='CRYGD', Aliases='CACA, CCA3, CCP, CRYG4, CTRCT4, PCC, cry-g-D', description='crystallin gamma D', other_designations='crystallin, gamma D|gamma crystallin 4|gamma-D-crystallin|gamma-crystallin 4', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208121607, end_position_on_the_genomic_accession=208124589, orientation='minus', exon_count=3, OMIM=123690, no_hdr0=''),
  100533646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533646, CurrentID=0, Status='live', Symbol='ZNF840P', Aliases='C20orf157, ZNF840, dJ981L23.3', description='zinc finger protein 840, pseudogene', other_designations='putative zinc finger protein 840|zinc finger protein 135 pseudogene', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46490940, end_position_on_the_genomic_accession=46492292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1423, CurrentID=0, Status='live', Symbol='CRYGFP', Aliases='CRYG61, p1, CRYGFP', description='crystallin gamma F, pseudogene', other_designations='crystallin, gamma D pseudogene|crystallin, gamma F pseudogene 1', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209145128, end_position_on_the_genomic_accession=209147806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1424, CurrentID=0, Status='live', Symbol='CRYGGP', Aliases='CRYGGP1', description='crystallin gamma G, pseudogene', other_designations='crystallin, gamma G pseudogene 1|crystallin, gamma G, pseudogene', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51774998, end_position_on_the_genomic_accession=51775714, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271505, CurrentID=0, Status='live', Symbol='RPL34P30', Aliases='RPL34_17_1547', description='ribosomal protein L34 pseudogene 30', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42714299, end_position_on_the_genomic_accession=42714639, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271506, CurrentID=0, Status='live', Symbol='RPL35AP21', Aliases='RPL35A_11_983', description='ribosomal protein L35a pseudogene 21', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=71589540, end_position_on_the_genomic_accession=71589977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1427, CurrentID=0, Status='live', Symbol='CRYGS', Aliases='CRYG8, CTRCT20', description='crystallin gamma S', other_designations='crystallin, gamma 8|crystallin, gamma S|gamma-S-crystallin|gamma-crystallin S', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186538443, end_position_on_the_genomic_accession=186544378, orientation='minus', exon_count=3, OMIM=123730, no_hdr0=''),
  1428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1428, CurrentID=0, Status='live', Symbol='CRYM', Aliases='DFNA40, THBP', description='crystallin mu', other_designations='NADP-regulated thyroid-hormone binding protein|crystallin, mu|mu-crystallin homolog|thiomorpholine-carboxylate dehydrogenase', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21258518, end_position_on_the_genomic_accession=21303136, orientation='minus', exon_count=10, OMIM=123740, no_hdr0=''),
  1429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1429, CurrentID=0, Status='live', Symbol='CRYZ', Aliases='', description='crystallin zeta', other_designations='NADPH:quinone reductase|crystallin, zeta (quinone reductase)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74705482, end_position_on_the_genomic_accession=74733408, orientation='minus', exon_count=11, OMIM=123691, no_hdr0=''),
  1430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1430, CurrentID=0, Status='live', Symbol='CRYZP1', Aliases='', description='crystallin zeta pseudogene 1', other_designations='crystallin, zeta (quinone reductase) pseudogene 1', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103087701, end_position_on_the_genomic_accession=103089775, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1431, CurrentID=0, Status='live', Symbol='CS', Aliases='', description='citrate synthase', other_designations='citrate (Si)-synthase', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56271699, end_position_on_the_genomic_accession=56300391, orientation='minus', exon_count=11, OMIM=118950, no_hdr0=''),
  1432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1432, CurrentID=0, Status='live', Symbol='MAPK14', Aliases='CSBP, CSBP1, CSBP2, CSPB1, EXIP, Mxi2, PRKM14, PRKM15, RK, SAPK2A, p38, p38ALPHA', description='mitogen-activated protein kinase 14', other_designations='CSAID-binding protein|Csaids binding protein|MAP kinase 14|MAP kinase Mxi2|MAP kinase p38 alpha|MAX-interacting protein 2|cytokine suppressive anti-inflammatory drug binding protein|cytokine suppressive anti-inflammatory drug-binding protein|mitogen-activated protein kinase p38 alpha|p38 MAP kinase|p38 mitogen activated protein kinase|p38alpha Exip|stress-activated protein kinase 2A', map_location='6p21.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36027635, end_position_on_the_genomic_accession=36111236, orientation='plus', exon_count=19, OMIM=600289, no_hdr0=''),
  1433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1433, CurrentID=0, Status='live', Symbol='CSE', Aliases='DYT9', description='choreoathetosis/spasticity, episodic (paroxysmal choreoathetosis/spasticity)', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601042, no_hdr0=''),
  1434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1434, CurrentID=0, Status='live', Symbol='CSE1L', Aliases='CAS, CSE1, XPO2', description='CSE1 chromosome segregation 1-like (yeast)', other_designations='cellular apoptosis susceptibility protein|chromosome segregation 1-like protein|exp2|importin-alpha re-exporter', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49046246, end_position_on_the_genomic_accession=49096960, orientation='plus', exon_count=25, OMIM=601342, no_hdr0=''),
  1435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1435, CurrentID=0, Status='live', Symbol='CSF1', Aliases='CSF-1, MCSF', description='colony stimulating factor 1 (macrophage)', other_designations='lanimostim', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109910611, end_position_on_the_genomic_accession=109930994, orientation='plus', exon_count=10, OMIM=120420, no_hdr0=''),
  1436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1436, CurrentID=0, Status='live', Symbol='CSF1R', Aliases='C-FMS, CD115, CSF-1R, CSFR, FIM2, FMS, HDLS, M-CSF-R', description='colony stimulating factor 1 receptor', other_designations='CD115 antigen|CSF-1 receptor|FMS proto-oncogene|McDonough feline sarcoma viral (v-fms) oncogene homolog|macrophage colony stimulating factor I receptor|proto-oncogene c-Fms', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150053291, end_position_on_the_genomic_accession=150113372, orientation='minus', exon_count=22, OMIM=164770, no_hdr0=''),
  1437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1437, CurrentID=0, Status='live', Symbol='CSF2', Aliases='GMCSF', description='colony stimulating factor 2 (granulocyte-macrophage)', other_designations='CSF|granulocyte macrophage-colony stimulating factor|molgramostin|sargramostim', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132073792, end_position_on_the_genomic_accession=132076170, orientation='plus', exon_count=4, OMIM=138960, no_hdr0=''),
  1438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1438, CurrentID=0, Status='live', Symbol='CSF2RA', Aliases='CD116, CDw116, CSF2RX, CSF2RAY, CSF2RX, CSF2RY, GM-CSF-R-alpha, GMCSFR, GMR, SMDP4, CSF2RA', description='colony stimulating factor 2 receptor, alpha, low-affinity (granulocyte-macrophage)', other_designations='CD116 antigen|GM-CSF receptor alpha subunit|GMCSFR-alpha|GMR-alpha|colony stimulating factor 2 receptor alpha subunit|granulocyte-macrophage colony-stimulating factor receptor alpha chain', map_location='Xp22.32 and Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1268800, end_position_on_the_genomic_accession=1325078, orientation='plus', exon_count=19, OMIM='306250,425000', no_hdr0=''),
  1439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1439, CurrentID=0, Status='live', Symbol='CSF2RB', Aliases='CD131, CDw131, IL3RB, IL5RB, SMDP5', description='colony stimulating factor 2 receptor, beta, low-affinity (granulocyte-macrophage)', other_designations='GM-CSF/IL-3/IL-5 receptor common beta subunit|GM-CSF/IL-3/IL-5 receptor common beta-chain|colony-stimulating factor-2 receptor, beta, low-affinity|interleukin 3 receptor/granulocyte-macrophage colony stimulating factor 3 receptor, beta (high affinity)', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36913575, end_position_on_the_genomic_accession=36940449, orientation='plus', exon_count=14, OMIM=138981, no_hdr0=''),
  1440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1440, CurrentID=0, Status='live', Symbol='CSF3', Aliases='C17orf33OS, GCSF, CSF3', description='colony stimulating factor 3', other_designations='colony stimulating factor 3 (granulocyte)|filgrastim|granulocyte-colony stimulating factor|lenograstim|pluripoietin', map_location='17q11.2-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40015361, end_position_on_the_genomic_accession=40017813, orientation='plus', exon_count=5, OMIM=138970, no_hdr0=''),
  1441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1441, CurrentID=0, Status='live', Symbol='CSF3R', Aliases='CD114, GCSFR', description='colony stimulating factor 3 receptor', other_designations='CD114 antigen|G-CSF receptor|G-CSF-R|colony stimulating factor 3 receptor (granulocyte)', map_location='1p35-p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36466043, end_position_on_the_genomic_accession=36483314, orientation='minus', exon_count=18, OMIM=138971, no_hdr0=''),
  1442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1442, CurrentID=0, Status='live', Symbol='CSH1', Aliases='CS-1, CSA, CSMT, GHB3, PL, hCS-1, hCS-A', description='chorionic somatomammotropin hormone 1 (placental lactogen)', other_designations='choriomammotropin|chorionic somatomammotropin A|chorionic somatomammotropin-1|growth hormone B3|placental lactogen', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63894908, end_position_on_the_genomic_accession=63896661, orientation='minus', exon_count=5, OMIM=150200, no_hdr0=''),
  1443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1443, CurrentID=0, Status='live', Symbol='CSH2', Aliases='CS-2, CSB, GHB1, PL, hCS-B', description='chorionic somatomammotropin hormone 2', other_designations='choriomammotropin|chorionic somatomammotropin B|growth hormone B1|lactogen|placental lactogen', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63872012, end_position_on_the_genomic_accession=63873729, orientation='minus', exon_count=5, OMIM=118820, no_hdr0=''),
  1444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1444, CurrentID=0, Status='live', Symbol='CSHL1', Aliases='CS-5, CSHP1, CSL, GHB4, hCS-L', description='chorionic somatomammotropin hormone-like 1', other_designations='chorionic somatomammotropin CS-5|growth hormone B4|growth hormone cluster', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63909602, end_position_on_the_genomic_accession=63911323, orientation='minus', exon_count=4, OMIM=603515, no_hdr0=''),
  1445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1445, CurrentID=0, Status='live', Symbol='CSK', Aliases='', description='c-src tyrosine kinase', other_designations='C-Src kinase|protein-tyrosine kinase CYL', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74782084, end_position_on_the_genomic_accession=74803198, orientation='plus', exon_count=15, OMIM=124095, no_hdr0=''),
  1446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1446, CurrentID=0, Status='live', Symbol='CSN1S1', Aliases='CASA, CSN1', description='casein alpha s1', other_designations='casein, alpha', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69931062, end_position_on_the_genomic_accession=69946570, orientation='plus', exon_count=16, OMIM=115450, no_hdr0=''),
  1447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1447, CurrentID=0, Status='live', Symbol='CSN2', Aliases='CASB', description='casein beta', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69955256, end_position_on_the_genomic_accession=69993629, orientation='minus', exon_count=8, OMIM=115460, no_hdr0=''),
  1448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1448, CurrentID=0, Status='live', Symbol='CSN3', Aliases='CNS10, CSN10, CSNK, KCA', description='casein kappa', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70238306, end_position_on_the_genomic_accession=70251432, orientation='plus', exon_count=8, OMIM=601695, no_hdr0=''),
  100271529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271529, CurrentID=0, Status='live', Symbol='RPL39P38', Aliases='RPL39_19_1635', description='ribosomal protein L39 pseudogene 38', other_designations='', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18285166, end_position_on_the_genomic_accession=18285545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271530, CurrentID=0, Status='live', Symbol='RPL39P39', Aliases='RPL39_19_1716', description='ribosomal protein L39 pseudogene 39', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57328500, end_position_on_the_genomic_accession=57328886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271531, CurrentID=0, Status='live', Symbol='RPL39P40', Aliases='RPL39_21_1719', description='ribosomal protein L39 pseudogene 40', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17500480, end_position_on_the_genomic_accession=17500873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1452, CurrentID=0, Status='live', Symbol='CSNK1A1', Aliases='CK1, CK1a, CKIa, HEL-S-77p, HLCDGP1, PRO2975', description='casein kinase 1, alpha 1', other_designations='CKI-alpha|clock regulator kinase|down-regulated in lung cancer|epididymis secretory sperm binding protein Li 77p', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149495894, end_position_on_the_genomic_accession=149551552, orientation='minus', exon_count=11, OMIM=600505, no_hdr0=''),
  1453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1453, CurrentID=0, Status='live', Symbol='CSNK1D', Aliases='ASPS, CKIdelta, FASPS2, HCKID', description='casein kinase 1, delta', other_designations='CKI-delta|CKId|tau-protein kinase CSNK1D', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82239725, end_position_on_the_genomic_accession=82273742, orientation='minus', exon_count=13, OMIM=600864, no_hdr0=''),
  1454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1454, CurrentID=0, Status='live', Symbol='CSNK1E', Aliases='CKIepsilon, HCKIE', description='casein kinase 1, epsilon', other_designations='CKI-epsilon|CKIe', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38290691, end_position_on_the_genomic_accession=38318084, orientation='minus', exon_count=12, OMIM=600863, no_hdr0=''),
  1455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1455, CurrentID=0, Status='live', Symbol='CSNK1G2', Aliases='CK1g2', description='casein kinase 1, gamma 2', other_designations='CKI-gamma 2|casein kinase 1 isoform gamma-2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1941149, end_position_on_the_genomic_accession=1981338, orientation='plus', exon_count=13, OMIM=602214, no_hdr0=''),
  1456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1456, CurrentID=0, Status='live', Symbol='CSNK1G3', Aliases='CKI-gamma 3L, CSNK1G3', description='casein kinase 1, gamma 3', other_designations='', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123512099, end_position_on_the_genomic_accession=123617045, orientation='plus', exon_count=14, OMIM=604253, no_hdr0=''),
  1457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1457, CurrentID=0, Status='live', Symbol='CSNK2A1', Aliases='CK2A1, CKII, CSNK2A3', description='casein kinase 2, alpha 1 polypeptide', other_designations='CK II alpha 3|CK2 catalytic subunit alpha|casein kinase II alpha 1 polypeptide pseudogene|casein kinase II alpha 1 subunit|protein kinase CK2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=472969, end_position_on_the_genomic_accession=543838, orientation='minus', exon_count=16, OMIM=115440, no_hdr0=''),
  100271538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271538, CurrentID=0, Status='live', Symbol='RPL7AP60', Aliases='RPL7A_25_1279', description='ribosomal protein L7a pseudogene 60', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112301812, end_position_on_the_genomic_accession=112302685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1459, CurrentID=0, Status='live', Symbol='CSNK2A2', Aliases='CK2A2, CSNK2A1', description='casein kinase 2, alpha prime polypeptide', other_designations="CK II alpha'", map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58157907, end_position_on_the_genomic_accession=58197878, orientation='minus', exon_count=12, OMIM=115442, no_hdr0=''),
  1460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1460, CurrentID=0, Status='live', Symbol='CSNK2B', Aliases='CK2B, CK2N, CSK2B, G5A', description='casein kinase 2, beta polypeptide', other_designations='CK II beta|CSNK2B-LY6G5B-1072|CSNK2B-LY6G5B-1103|CSNK2B-LY6G5B-532|CSNK2B-LY6G5B-560|CSNK2B-LY6G5B-562|Casein kinase II beta subunit|alternative name: G5a, phosvitin|casein kinase II b subunit splicing isoform 132|casein kinase II b subunit splicing isoform 247|casein kinase II b subunit splicing isoform 318|casein kinase II b subunit splicing isoform 660|casein kinase II b subunit splicing isoform 806|chimera CSNK2B-LY6G5B splicing isoform 1072|chimera CSNK2B-LY6G5B splicing isoform 1103|chimera CSNK2B-LY6G5B splicing isoform 532|chimera CSNK2B-LY6G5B splicing isoform 560|chimera CSNK2B-LY6G5B splicing isoform 562|phosvitin|protein G5a', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31665880, end_position_on_the_genomic_accession=31670070, orientation='plus', exon_count=7, OMIM=115441, no_hdr0=''),
  100653049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653049, CurrentID=0, Status='live', Symbol='LOC100653049', Aliases='', description='keratin, type I cuticular Ha4', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1462, CurrentID=0, Status='live', Symbol='VCAN', Aliases='CSPG2, ERVR, GHAP, PG-M, WGN, WGN1', description='versican', other_designations='chondroitin sulfate proteoglycan 2|chondroitin sulfate proteoglycan core protein 2|glial hyaluronate-binding protein|large fibroblast proteoglycan|versican proteoglycan', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83471674, end_position_on_the_genomic_accession=83582303, orientation='plus', exon_count=15, OMIM=118661, no_hdr0=''),
  1463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1463, CurrentID=0, Status='live', Symbol='NCAN', Aliases='CSPG3', description='neurocan', other_designations='chondroitin sulfate proteoglycan 3 (neurocan)|neurocan proteoglycan', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19211964, end_position_on_the_genomic_accession=19252252, orientation='plus', exon_count=15, OMIM=600826, no_hdr0=''),
  1464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1464, CurrentID=0, Status='live', Symbol='CSPG4', Aliases='HMW-MAA, MCSP, MCSPG, MEL-CSPG, MSK16, NG2', description='chondroitin sulfate proteoglycan 4', other_designations='chondroitin sulfate proteoglycan 4 (melanoma-associated)|chondroitin sulfate proteoglycan NG2|melanoma chondroitin sulfate proteoglycan|melanoma-associated chondroitin sulfate proteoglycan', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75674322, end_position_on_the_genomic_accession=75712848, orientation='minus', exon_count=10, OMIM=601172, no_hdr0=''),
  1465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1465, CurrentID=0, Status='live', Symbol='CSRP1', Aliases='CRP, CRP1, CSRP, CYRP, D1S181E, HEL-141, HEL-S-286', description='cysteine and glycine rich protein 1', other_designations='LIM-domain protein|cysteine-rich protein 1|epididymis luminal protein 141|epididymis secretory protein Li 286', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201483530, end_position_on_the_genomic_accession=201507259, orientation='minus', exon_count=8, OMIM=123876, no_hdr0=''),
  1466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1466, CurrentID=0, Status='live', Symbol='CSRP2', Aliases='CRP2, LMO5, SmLIM', description='cysteine and glycine rich protein 2', other_designations='LIM domain only 5, smooth muscle|LIM domain only protein 5|LMO-5|cysteine-rich protein 2|smooth muscle cell LIM protein', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76858715, end_position_on_the_genomic_accession=76879046, orientation='minus', exon_count=7, OMIM=601871, no_hdr0=''),
  1467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1467, CurrentID=0, Status='live', Symbol='CSRP2P1', Aliases='CSRP2P', description='cysteine and glycine rich protein 2 pseudogene 1', other_designations='cysteine and glycine-rich protein 2 pseudogene', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119530830, end_position_on_the_genomic_accession=119531611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1468, CurrentID=0, Status='live', Symbol='SLC25A10', Aliases='DIC', description='solute carrier family 25 (mitochondrial carrier; dicarboxylate transporter), member 10', other_designations='dicarboxylate ion carrier', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81712236, end_position_on_the_genomic_accession=81721016, orientation='plus', exon_count=12, OMIM=606794, no_hdr0=''),
  1469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1469, CurrentID=0, Status='live', Symbol='CST1', Aliases='', description='cystatin SN', other_designations='cystain-SA-I|cystatin 1|cystatin SA-I|cystatin-1|cysteine proteinase inhibitor, type 2 family|salivary cystatin-SA-1', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23747553, end_position_on_the_genomic_accession=23750937, orientation='minus', exon_count=3, OMIM=123855, no_hdr0=''),
  1470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1470, CurrentID=0, Status='live', Symbol='CST2', Aliases='', description='cystatin SA', other_designations='cystatin 2|cystatin S5|cystatin-2|cystatin-S5|cysteine-proteinase inhibitor|salivary cysteine (thiol) protease inhibitor', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23823767, end_position_on_the_genomic_accession=23826675, orientation='minus', exon_count=3, OMIM=123856, no_hdr0=''),
  1471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1471, CurrentID=0, Status='live', Symbol='CST3', Aliases='ARMD11, HEL-S-2', description='cystatin C', other_designations='bA218C14.4 (cystatin C)|cystatin 3|cystatin-3|epididymis secretory protein Li 2|gamma-trace|neuroendocrine basic polypeptide|post-gamma-globulin', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23627897, end_position_on_the_genomic_accession=23638048, orientation='minus', exon_count=4, OMIM=604312, no_hdr0=''),
  1472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1472, CurrentID=0, Status='live', Symbol='CST4', Aliases='', description='cystatin S', other_designations='cystatin 4|cystatin-4|cystatin-SA-III|salivary acidic protein 1', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23685640, end_position_on_the_genomic_accession=23689025, orientation='minus', exon_count=3, OMIM=123857, no_hdr0=''),
  1473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1473, CurrentID=0, Status='live', Symbol='CST5', Aliases='', description='cystatin D', other_designations='cystatin 5|cystatin-5|cysteine-proteinase inhibitor', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23875935, end_position_on_the_genomic_accession=23879743, orientation='minus', exon_count=3, OMIM=123858, no_hdr0=''),
  1474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1474, CurrentID=0, Status='live', Symbol='CST6', Aliases='', description='cystatin E/M', other_designations='cystatin 6|cystatin M|cystatin M/E|cystatin-6|cystatin-E|cysteine proteinase inhibitor', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66011991, end_position_on_the_genomic_accession=66013505, orientation='plus', exon_count=3, OMIM=601891, no_hdr0=''),
  1475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1475, CurrentID=0, Status='live', Symbol='CSTA', Aliases='AREI, STF1, STFA', description='cystatin A', other_designations='cystatin A (stefin A)|cystatin AS', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122325164, end_position_on_the_genomic_accession=122341969, orientation='plus', exon_count=3, OMIM=184600, no_hdr0=''),
  1476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1476, CurrentID=0, Status='live', Symbol='CSTB', Aliases='CST6, EPM1, EPM1A, PME, STFB, ULD', description='cystatin B', other_designations='CPI-B|cystatin B (stefin B)|liver thiol proteinase inhibitor', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43773665, end_position_on_the_genomic_accession=43776375, orientation='minus', exon_count=3, OMIM=601145, no_hdr0=''),
  1477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1477, CurrentID=0, Status='live', Symbol='CSTF1', Aliases='CstF-50, CstFp50', description="cleavage stimulation factor, 3' pre-RNA, subunit 1", other_designations="CF-1 50 kDa subunit|CSTF 50 kDa subunit|cleavage stimulation factor 50 kDa subunit|cleavage stimulation factor, 3' pre-RNA, subunit 1, 50kD|cleavage stimulation factor, 3' pre-RNA, subunit 1, 50kDa", map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56392371, end_position_on_the_genomic_accession=56404526, orientation='plus', exon_count=6, OMIM=600369, no_hdr0=''),
  1478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1478, CurrentID=0, Status='live', Symbol='CSTF2', Aliases='CstF-64', description="cleavage stimulation factor, 3' pre-RNA, subunit 2", other_designations="CF-1 64 kDa subunit|CSTF 64 kDa subunit|betaCstF-64 variant 2|cleavage stimulation factor 64 kDa subunit|cleavage stimulation factor, 3' pre-RNA, subunit 2, 64kD|cleavage stimulation factor, 3' pre-RNA, subunit 2, 64kDa", map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100820359, end_position_on_the_genomic_accession=100841520, orientation='plus', exon_count=16, OMIM=300907, no_hdr0=''),
  1479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1479, CurrentID=0, Status='live', Symbol='CSTF3', Aliases='CSTF-77', description="cleavage stimulation factor, 3' pre-RNA, subunit 3", other_designations="CF-1 77 kDa subunit|CSTF 77 kDa subunit|cleavage stimulation factor 77 kDa subunit|cleavage stimulation factor subunit 3|cleavage stimulation factor, 3' pre-RNA, subunit 3, 77kD|cleavage stimulation factor, 3' pre-RNA, subunit 3, 77kDa", map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33084584, end_position_on_the_genomic_accession=33161491, orientation='minus', exon_count=22, OMIM=600367, no_hdr0=''),
  1480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1480, CurrentID=0, Status='live', Symbol='CSTP1', Aliases='dJ850N15.1', description='cystatin pseudogene 1', other_designations='cystatin S pseudogene', map_location='20p11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23918771, end_position_on_the_genomic_accession=23922169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1481, CurrentID=0, Status='live', Symbol='CSTP2', Aliases='', description='cystatin pseudogene 2', other_designations='type 2 cystatin pseudogene', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23772859, end_position_on_the_genomic_accession=23776169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1482, CurrentID=0, Status='live', Symbol='NKX2-5', Aliases='CHNG5, CSX, CSX1, HLHS2, NKX2.5, NKX2E, NKX4-1, VSD3', description='NK2 homeobox 5', other_designations='NK2 transcription factor related locus 5|NK2 transcription factor related, locus 5|cardiac-specific homeobox 1|homeobox protein CSX|homeobox protein NK-2 homolog E|tinman paralog', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173232104, end_position_on_the_genomic_accession=173235312, orientation='minus', exon_count=2, OMIM=600584, no_hdr0=''),
  1483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1483, CurrentID=0, Status='live', Symbol='CTAA1', Aliases='CAP, CTPP5', description='cataract, anterior polar 1', other_designations='', map_location='14q24-qter', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=115650, no_hdr0=''),
  1484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1484, CurrentID=0, Status='live', Symbol='CTAA2', Aliases='', description='cataract, anterior polar 2', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601202, no_hdr0=''),
  1485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1485, CurrentID=0, Status='live', Symbol='CTAG1B', Aliases='CT6.1, CTAG, CTAG1, ESO1, LAGE-2, LAGE2B, NY-ESO-1', description='cancer/testis antigen 1B', other_designations='New York esophageal squamous cell carcinoma 1|autoimmunogenic cancer/testis antigen NY-ESO-1|cancer antigen 3|cancer/testis antigen 6.1|l antigen family member 2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154617609, end_position_on_the_genomic_accession=154619271, orientation='minus', exon_count=3, OMIM=300156, no_hdr0=''),
  1486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1486, CurrentID=0, Status='live', Symbol='CTBS', Aliases='CTB', description='chitobiase, di-N-acetyl-', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84553121, end_position_on_the_genomic_accession=84574480, orientation='minus', exon_count=7, OMIM=600873, no_hdr0=''),
  1487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1487, CurrentID=0, Status='live', Symbol='CTBP1', Aliases='BARS', description='C-terminal binding protein 1', other_designations='brefeldin A-ribosylated substrate', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1211440, end_position_on_the_genomic_accession=1250248, orientation='minus', exon_count=14, OMIM=602618, no_hdr0=''),
  1488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1488, CurrentID=0, Status='live', Symbol='CTBP2', Aliases='', description='C-terminal binding protein 2', other_designations='ribeye', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124987849, end_position_on_the_genomic_accession=125161055, orientation='minus', exon_count=21, OMIM=602619, no_hdr0=''),
  1489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1489, CurrentID=0, Status='live', Symbol='CTF1', Aliases='CT-1, CT1', description='cardiotrophin 1', other_designations='cardiophin 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30895828, end_position_on_the_genomic_accession=30903560, orientation='plus', exon_count=4, OMIM=600435, no_hdr0=''),
  1490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1490, CurrentID=0, Status='live', Symbol='CTGF', Aliases='CCN2, HCS24, IGFBP8, NOV2', description='connective tissue growth factor', other_designations='CCN family member 2|IBP-8|IGF-binding protein 8|IGFBP-8|hypertrophic chondrocyte-specific protein 24|insulin-like growth factor-binding protein 8', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131948176, end_position_on_the_genomic_accession=131951378, orientation='minus', exon_count=5, OMIM=121009, no_hdr0=''),
  1491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1491, CurrentID=0, Status='live', Symbol='CTH', Aliases='', description='cystathionine gamma-lyase', other_designations='cystathionase (cystathionine gamma-lyase)|cysteine desulfhydrase|cysteine-protein sulfhydrase|gamma-cystathionase|homoserine deaminase|homoserine dehydratase', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70411218, end_position_on_the_genomic_accession=70440114, orientation='plus', exon_count=13, OMIM=607657, no_hdr0=''),
  100271572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271572, CurrentID=0, Status='live', Symbol='RPS26P46', Aliases='RPS26_23_1212', description='ribosomal protein S26 pseudogene 46', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13021265, end_position_on_the_genomic_accession=13021603, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1493, CurrentID=0, Status='live', Symbol='CTLA4', Aliases='ALPS5, CD, CD152, CELIAC3, CTLA-4, GRD4, GSE, IDDM12', description='cytotoxic T-lymphocyte-associated protein 4', other_designations='CD152 isoform|celiac disease 3|cytotoxic T lymphocyte associated antigen 4 short spliced form|cytotoxic T-lymphocyte-associated serine esterase-4|insulin-dependent diabetes mellitus 12|ligand and transmembrane spliced cytotoxic T lymphocyte associated antigen 4', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203867788, end_position_on_the_genomic_accession=203873960, orientation='plus', exon_count=4, OMIM=123890, no_hdr0=''),
  100271574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271574, CurrentID=0, Status='live', Symbol='RPS26P48', Aliases='RPS26_26_1396', description='ribosomal protein S26 pseudogene 48', other_designations='', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78231277, end_position_on_the_genomic_accession=78231619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1495, CurrentID=0, Status='live', Symbol='CTNNA1', Aliases='CAP102', description='catenin alpha 1', other_designations='alpha E-catenin|alpha-E-catenin|alpha-catenin|alphaE-catenin|cadherin-associated protein,102kDa|catenin (cadherin-associated protein), alpha 1, 102kDa|renal carcinoma antigen NY-REN-13', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138753396, end_position_on_the_genomic_accession=138935034, orientation='plus', exon_count=22, OMIM=116805, no_hdr0=''),
  1496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1496, CurrentID=0, Status='live', Symbol='CTNNA2', Aliases='CAP-R, CAPR, CT114, CTNR', description='catenin alpha 2', other_designations='alpha N-catenin|alpha-N-catenin|alpha-catenin-related protein|cadherin-associated protein, related|cancer/testis antigen 114|catenin (cadherin-associated protein), alpha 2', map_location='2p12-p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79512934, end_position_on_the_genomic_accession=80648868, orientation='plus', exon_count=25, OMIM=114025, no_hdr0=''),
  1497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1497, CurrentID=0, Status='live', Symbol='CTNS', Aliases='CTNS-LSB, PQLC4', description='cystinosin, lysosomal cystine transporter', other_designations='cystinosis nephropathic', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3636468, end_position_on_the_genomic_accession=3663103, orientation='plus', exon_count=12, OMIM=606272, no_hdr0=''),
  1498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1498, CurrentID=0, Status='live', Symbol='CTNNA1P1', Aliases='CTNNAP1', description='catenin alpha 1 pseudogene 1', other_designations='catenin (cadherin-associated protein), alpha 1 pseudogene 1|catenin (cadherin-associated protein), alpha pseudogene 1', map_location='5q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115389024, end_position_on_the_genomic_accession=115392441, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1499, CurrentID=0, Status='live', Symbol='CTNNB1', Aliases='CTNNB, MRD19, armadillo', description='catenin beta 1', other_designations='catenin (cadherin-associated protein), beta 1, 88kDa', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41199451, end_position_on_the_genomic_accession=41240448, orientation='plus', exon_count=17, OMIM=116806, no_hdr0=''),
  1500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1500, CurrentID=0, Status='live', Symbol='CTNND1', Aliases='CAS, CTNND, P120CAS, P120CTN, p120, p120(CAS), p120(CTN)', description='catenin delta 1', other_designations='cadherin-associated Src substrate|catenin (cadherin-associated protein), delta 1|p120 catenin', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57761762, end_position_on_the_genomic_accession=57819180, orientation='plus', exon_count=21, OMIM=601045, no_hdr0=''),
  1501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1501, CurrentID=0, Status='live', Symbol='CTNND2', Aliases='GT24, NPRAP', description='catenin delta 2', other_designations='T-cell delta-catenin|catenin (cadherin-associated protein), delta 2 (neural plakophilin-related arm-repeat protein)|neurojungin', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10971840, end_position_on_the_genomic_accession=11904043, orientation='minus', exon_count=26, OMIM=604275, no_hdr0=''),
  1503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1503, CurrentID=0, Status='live', Symbol='CTPS1', Aliases='CTPS, IMD24', description='CTP synthase 1', other_designations="CTP synthetase 1|UTP--ammonia ligase 1|cytidine 5'-triphosphate synthetase|cytidine 5-prime triphosphate synthetase", map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40979299, end_position_on_the_genomic_accession=41012565, orientation='plus', exon_count=23, OMIM=123860, no_hdr0=''),
  1504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1504, CurrentID=0, Status='live', Symbol='CTRB1', Aliases='CTRB', description='chymotrypsinogen B1', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75218986, end_position_on_the_genomic_accession=75224924, orientation='plus', exon_count=7, OMIM=118890, no_hdr0=''),
  100271585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271585, CurrentID=0, Status='live', Symbol='RPS27P27', Aliases='RPS27_15_1487', description='ribosomal protein S27 pseudogene 27', other_designations='ribosomal protein S27-like pseudogene', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59088837, end_position_on_the_genomic_accession=59089091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1506, CurrentID=0, Status='live', Symbol='CTRL', Aliases='CTRL1', description='chymotrypsin-like', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67929570, end_position_on_the_genomic_accession=67931875, orientation='minus', exon_count=7, OMIM=118888, no_hdr0=''),
  100271587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271587, CurrentID=0, Status='live', Symbol='RPS29P20', Aliases='RPS29_15_1116', description='ribosomal protein S29 pseudogene 20', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4018948, end_position_on_the_genomic_accession=4019118, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1508, CurrentID=0, Status='live', Symbol='CTSB', Aliases='APPS, CPSB', description='cathepsin B', other_designations='APP secretase|amyloid precursor protein secretase|cathepsin B1|cysteine protease', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11842524, end_position_on_the_genomic_accession=11868137, orientation='minus', exon_count=15, OMIM=116810, no_hdr0=''),
  1509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1509, CurrentID=0, Status='live', Symbol='CTSD', Aliases='CLN10, CPSD, HEL-S-130P', description='cathepsin D', other_designations='ceroid-lipofuscinosis, neuronal 10|epididymis secretory sperm binding protein Li 130P|lysosomal aspartyl peptidase|lysosomal aspartyl protease', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1752752, end_position_on_the_genomic_accession=1763992, orientation='minus', exon_count=9, OMIM=116840, no_hdr0=''),
  1510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1510, CurrentID=0, Status='live', Symbol='CTSE', Aliases='CATE', description='cathepsin E', other_designations='erythrocyte membrane aspartic proteinase|slow-moving proteinase', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206009264, end_position_on_the_genomic_accession=206023909, orientation='minus', exon_count=9, OMIM=116890, no_hdr0=''),
  1511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1511, CurrentID=0, Status='live', Symbol='CTSG', Aliases='CATG, CG', description='cathepsin G', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24573518, end_position_on_the_genomic_accession=24576260, orientation='minus', exon_count=5, OMIM=116830, no_hdr0=''),
  1512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1512, CurrentID=0, Status='live', Symbol='CTSH', Aliases='ACC-4, ACC-5, ACC4, ACC5, CPSB, minichain', description='cathepsin H', other_designations='N-benzoylarginine-beta-naphthylamide hydrolase|aleurain|cathepsin B3|cathepsin BA', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78921750, end_position_on_the_genomic_accession=78945089, orientation='minus', exon_count=13, OMIM=116820, no_hdr0=''),
  1513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1513, CurrentID=0, Status='live', Symbol='CTSK', Aliases='CTS02, CTSO, CTSO1, CTSO2, PKND, PYCD', description='cathepsin K', other_designations='cathepsin O|cathepsin O1|cathepsin O2|cathepsin X', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150796208, end_position_on_the_genomic_accession=150808441, orientation='minus', exon_count=8, OMIM=601105, no_hdr0=''),
  1514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1514, CurrentID=0, Status='live', Symbol='CTSL', Aliases='CATL1, MEP, CTSL', description='cathepsin L', other_designations='major excreted protein', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87725519, end_position_on_the_genomic_accession=87731469, orientation='plus', exon_count=9, OMIM=116880, no_hdr0=''),
  1515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1515, CurrentID=0, Status='live', Symbol='CTSV', Aliases='CATL2, CTSL2, CTSU', description='cathepsin V', other_designations='cathepsin L2, preproprotein|cathepsin U', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97029677, end_position_on_the_genomic_accession=97039643, orientation='minus', exon_count=9, OMIM=603308, no_hdr0=''),
  100271596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271596, CurrentID=0, Status='live', Symbol='RPS3AP46', Aliases='RPS3A_19_1365', description='ribosomal protein S3a pseudogene 46', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53612798, end_position_on_the_genomic_accession=53613381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1517, CurrentID=0, Status='live', Symbol='CTSLP2', Aliases='CTSL1P2, CTSL1P5, CTSL1P7, CTSLL2, CTSLL5, CTSLL7, CTSLP5, CTSLP7, bA144G6.6, bA342C24.4', description='cathepsin L pseudogene 2', other_designations='cathepsin L pseudogene 5|cathepsin L pseudogene 7|cathepsin L-like 2|cathepsin L1 pseudogene 2|cathepsin L1 pseudogene 5|cathepsin L1 pseudogene 7', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47580671, end_position_on_the_genomic_accession=47583419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  1518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1518, CurrentID=0, Status='live', Symbol='CTSLP8', Aliases='CTSL1P8', description='cathepsin L pseudogene 8', other_designations='cathepsin L1 pseudogene 8', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87844745, end_position_on_the_genomic_accession=87847424, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  1519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1519, CurrentID=0, Status='live', Symbol='CTSO', Aliases='CTSO1', description='cathepsin O', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155924118, end_position_on_the_genomic_accession=155953896, orientation='minus', exon_count=8, OMIM=600550, no_hdr0=''),
  1520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1520, CurrentID=0, Status='live', Symbol='CTSS', Aliases='', description='cathepsin S', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150730196, end_position_on_the_genomic_accession=150765957, orientation='minus', exon_count=8, OMIM=116845, no_hdr0=''),
  1521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1521, CurrentID=0, Status='live', Symbol='CTSW', Aliases='LYPN', description='cathepsin W', other_designations='lymphopain', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65879813, end_position_on_the_genomic_accession=65883741, orientation='plus', exon_count=10, OMIM=602364, no_hdr0=''),
  1522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1522, CurrentID=0, Status='live', Symbol='CTSZ', Aliases='CTSX', description='cathepsin Z', other_designations='carboxypeptidase LB|cathepsin B2|cathepsin IV|cathepsin P|cathepsin X|cathepsin Y|cathepsin Z1|cysteine-type carboxypeptidase|lysosomal carboxypeptidase B|preprocathepsin P', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58995187, end_position_on_the_genomic_accession=59007254, orientation='minus', exon_count=6, OMIM=603169, no_hdr0=''),
  1523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1523, CurrentID=0, Status='live', Symbol='CUX1', Aliases='CASP, CDP, CDP/Cut, CDP1, COY1, CUTL1, CUX, Clox, Cux/CDP, GOLIM6, Nbla10317, p100, p110, p200, p75', description='cut-like homeobox 1', other_designations='CCAAT displacement protein|cut homolog|golgi integral membrane protein 6|homeobox protein cux-1|putative protein product of Nbla10317', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101815904, end_position_on_the_genomic_accession=102283958, orientation='plus', exon_count=34, OMIM=116896, no_hdr0=''),
  1524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1524, CurrentID=0, Status='live', Symbol='CX3CR1', Aliases='CCRL1, CMKBRL1, CMKDR1, GPR13, GPRV28, V28', description='chemokine (C-X3-C motif) receptor 1', other_designations='C-X3-C CKR-1|CMK-BRL-1|CMK-BRL1|G protein-coupled receptor 13|G-protein coupled receptor 13|beta chemokine receptor-like 1|chemokine (C-C) receptor-like 1|chemokine (C-X3-C) receptor 1|fractalkine receptor', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39263494, end_position_on_the_genomic_accession=39281735, orientation='minus', exon_count=5, OMIM=601470, no_hdr0=''),
  1525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1525, CurrentID=0, Status='live', Symbol='CXADR', Aliases='CAR, CAR4/6, HCAR', description='coxsackie virus and adenovirus receptor', other_designations='46 kD coxsackievirus and adenovirus receptor (CAR) protein|CVB3-binding protein|HCVADR|coxsackievirus B-adenovirus receptor', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17512906, end_position_on_the_genomic_accession=17633323, orientation='plus', exon_count=10, OMIM=602621, no_hdr0=''),
  1526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1526, CurrentID=0, Status='live', Symbol='CXB3S', Aliases='CB3S', description='coxsackie virus B3 sensitivity', other_designations='', map_location='19pter-q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=120050, no_hdr0=''),
  1527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1527, CurrentID=0, Status='live', Symbol='TEX28', Aliases='CXorf2, MRX99P1, TEX28P2, fTEX, TEX28', description='testis expressed 28', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154271265, end_position_on_the_genomic_accession=154295351, orientation='minus', exon_count=9, OMIM=300092, no_hdr0=''),
  1528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1528, CurrentID=0, Status='live', Symbol='CYB5A', Aliases='CYB5, MCB5', description='cytochrome b5 type A (microsomal)', other_designations='type 1 cyt-b5', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74253292, end_position_on_the_genomic_accession=74292016, orientation='minus', exon_count=6, OMIM=613218, no_hdr0=''),
  100271610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271610, CurrentID=0, Status='live', Symbol='RPL36AP35', Aliases='RPL36A_16_1006', description='ribosomal protein L36a pseudogene 35', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108821697, end_position_on_the_genomic_accession=108822018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1531, CurrentID=0, Status='live', Symbol='CYB5AP3', Aliases='CYB5P3, Psgb-5-1', description='cytochrome b5 type A (microsomal) pseudogene 3', other_designations='cytochrome b-5 pseudogene 1|cytochrome b-5 pseudogene 3', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93137955, end_position_on_the_genomic_accession=93138777, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1532, CurrentID=0, Status='live', Symbol='CYB5AP4', Aliases='CYB5P4, Psgb-5-2, bA189G24.1', description='cytochrome b5 type A (microsomal) pseudogene 4', other_designations='cytochrome b-5 pseudogene 4', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22885490, end_position_on_the_genomic_accession=22886068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271613, CurrentID=0, Status='live', Symbol='RPS15AP27', Aliases='RPS15A_15_1003', description='ribosomal protein S15a pseudogene 27', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107963661, end_position_on_the_genomic_accession=107964075, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1534, CurrentID=0, Status='live', Symbol='CYB561', Aliases='CYB561A1, FRRS2', description='cytochrome b561', other_designations='cytochrome b-561|cytochrome b561 family, member A1|ferric-chelate reductase 2', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63432304, end_position_on_the_genomic_accession=63446361, orientation='minus', exon_count=8, OMIM=600019, no_hdr0=''),
  1535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1535, CurrentID=0, Status='live', Symbol='CYBA', Aliases='p22-PHOX', description='cytochrome b-245, alpha polypeptide', other_designations='cytochrome b light chain|cytochrome b(558) alpha chain|cytochrome b(558) alpha-subunit|cytochrome b, alpha polypeptide|cytochrome b558 subunit alpha|flavocytochrome b-558 alpha polypeptide|neutrophil cytochrome b 22 kDa polypeptide|p22 phagocyte B-cytochrome|p22phox|superoxide-generating NADPH oxidase light chain subunit', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88643289, end_position_on_the_genomic_accession=88651085, orientation='minus', exon_count=7, OMIM=608508, no_hdr0=''),
  1536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1536, CurrentID=0, Status='live', Symbol='CYBB', Aliases='AMCBX2, CGD, GP91-1, GP91-PHOX, GP91PHOX, IMD34, NOX2, p91-PHOX', description='cytochrome b-245, beta polypeptide', other_designations='CGD91-phox|NADPH oxidase 2|cytochrome b(558) subunit beta|cytochrome b-245 beta polypeptide|cytochrome b558 subunit beta|heme-binding membrane glycoprotein gp91phox|neutrophil cytochrome b 91 kDa polypeptide|p22 phagocyte B-cytochrome|superoxide-generating NADPH oxidase heavy chain subunit', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37780017, end_position_on_the_genomic_accession=37813465, orientation='plus', exon_count=14, OMIM=300481, no_hdr0=''),
  1537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1537, CurrentID=0, Status='live', Symbol='CYC1', Aliases='MC3DN6, UQCR4', description='cytochrome c-1', other_designations='complex III subunit 4|complex III subunit IV|cytochrome b-c1 complex subunit 4|ubiquinol-cytochrome-c reductase complex cytochrome c1 subunit', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144095035, end_position_on_the_genomic_accession=144097527, orientation='plus', exon_count=7, OMIM=123980, no_hdr0=''),
  1538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1538, CurrentID=0, Status='live', Symbol='CYLC1', Aliases='CYCL1', description='cylicin, basic protein of sperm head cytoskeleton 1', other_designations='cylicin I|multiple-band polypeptide I', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=83861126, end_position_on_the_genomic_accession=83886700, orientation='plus', exon_count=5, OMIM=300768, no_hdr0=''),
  1539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1539, CurrentID=0, Status='live', Symbol='CYLC2', Aliases='', description='cylicin, basic protein of sperm head cytoskeleton 2', other_designations='cylicin II|multiple-band polypeptide II', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102995311, end_position_on_the_genomic_accession=103018488, orientation='plus', exon_count=8, OMIM=604035, no_hdr0=''),
  1540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1540, CurrentID=0, Status='live', Symbol='CYLD', Aliases='BRSS, CDMT1, CYLDI, EAC, MFT, MFT1, SBS, TEM, USPL2, CYLD', description='cylindromatosis (turban tumor syndrome)', other_designations='deubiquitinating enzyme CYLD|probable ubiquitin carboxyl-terminal hydrolase CYLD|ubiquitin specific peptidase like 2|ubiquitin thioesterase CYLD|ubiquitin thiolesterase CYLD|ubiquitin-specific-processing protease CYLD', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50742050, end_position_on_the_genomic_accession=50801936, orientation='plus', exon_count=21, OMIM=605018, no_hdr0=''),
  1541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1541, CurrentID=0, Status='live', Symbol='CYMD', Aliases='DCMD, MDDC', description='cystoid macular dystrophy', other_designations='', map_location='7p21-p15', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=153880, no_hdr0=''),
  100271622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271622, CurrentID=0, Status='live', Symbol='RPL23AP67', Aliases='RPL23A_31_1299', description='ribosomal protein L23a pseudogene 67', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133125479, end_position_on_the_genomic_accession=133125952, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1543, CurrentID=0, Status='live', Symbol='CYP1A1', Aliases='AHH, AHRR, CP11, CYP1, P1-450, P450-C, P450DX', description='cytochrome P450 family 1 subfamily A member 1', other_designations='CYPIA1|aryl hydrocarbon hydroxylase|cytochrome P1-450, dioxin-inducible|cytochrome P450 form 6|cytochrome P450, family 1, subfamily A, polypeptide 1|cytochrome P450, subfamily I (aromatic compound-inducible), polypeptide 1|cytochrome P450-C|cytochrome P450-P1|flavoprotein-linked monooxygenase|xenobiotic monooxygenase', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74719542, end_position_on_the_genomic_accession=74725610, orientation='minus', exon_count=7, OMIM=108330, no_hdr0=''),
  1544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1544, CurrentID=0, Status='live', Symbol='CYP1A2', Aliases='CP12, P3-450, P450(PA)', description='cytochrome P450 family 1 subfamily A member 2', other_designations='CYPIA2|P450 form 4|aryl hydrocarbon hydroxylase|cholesterol 25-hydroxylase|cytochrome P(3)450|cytochrome P450 4|cytochrome P450, family 1, subfamily A, polypeptide 2|cytochrome P450, subfamily I (aromatic compound-inducible), polypeptide 2|cytochrome P450-P3|dioxin-inducible P3-450|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74748843, end_position_on_the_genomic_accession=74756600, orientation='plus', exon_count=7, OMIM=124060, no_hdr0=''),
  1545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1545, CurrentID=0, Status='live', Symbol='CYP1B1', Aliases='CP1B, CYPIB1, GLC3A, P4501B1', description='cytochrome P450 family 1 subfamily B member 1', other_designations='aryl hydrocarbon hydroxylase|cytochrome P450, family 1, subfamily B, polypeptide 1|cytochrome P450, subfamily I (dioxin-inducible), polypeptide 1 (glaucoma 3, primary infantile)|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38067603, end_position_on_the_genomic_accession=38076181, orientation='minus', exon_count=3, OMIM=601771, no_hdr0=''),
  1546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1546, CurrentID=0, Status='live', Symbol='CYP2A', Aliases='CYP1, CYP2, P450C2, P450C2A, P450PB', description='cytochrome P450, family 2, subfamily A', other_designations='cytochrome P450, subfamily IIA (phenobarbital-inducible)', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=41314400, end_position_on_the_genomic_accession=41712393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271627, CurrentID=0, Status='live', Symbol='RPL23AP81', Aliases='RPL23A_40_1702', description='ribosomal protein L23a pseudogene 81', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41196686, end_position_on_the_genomic_accession=41197157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1548, CurrentID=0, Status='live', Symbol='CYP2A6', Aliases='CPA6, CYP2A, CYP2A3, CYPIIA6, P450C2A, P450PB', description='cytochrome P450 family 2 subfamily A member 6', other_designations='1,4-cineole 2-exo-monooxygenase|coumarin 7-hydroxylase|cytochrome P450 IIA3|cytochrome P450(I)|cytochrome P450, family 2, subfamily A, polypeptide 6|cytochrome P450, subfamily IIA (phenobarbital-inducible), polypeptide 6|flavoprotein-linked monooxygenase|xenobiotic monooxygenase', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40843538, end_position_on_the_genomic_accession=40850447, orientation='minus', exon_count=9, OMIM=122720, no_hdr0=''),
  1549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1549, CurrentID=0, Status='live', Symbol='CYP2A7', Aliases='CPA7, CPAD, CYP2A, CYPIIA7, P450-IIA4', description='cytochrome P450 family 2 subfamily A member 7', other_designations='cytochrome P450 IIA4|cytochrome P450, family 2, subfamily A, polypeptide 7|cytochrome P450, subfamily IIA (phenobarbital-inducible), polypeptide 7|cytochrome P450IIA4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40875439, end_position_on_the_genomic_accession=40882752, orientation='minus', exon_count=9, OMIM=608054, no_hdr0=''),
  1550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1550, CurrentID=0, Status='live', Symbol='CYP2A7P1', Aliases='CYP2A18PC, CYP2A18PN, CYP2A7P2, CYP2A7PT', description='cytochrome P450 family 2 subfamily A member 7 pseudogene 1', other_designations='cytochrome P450, family 2, subfamily A, polypeptide 7 pseudogene 1|cytochrome P450, family 2, subfamily A, polypeptide 7 pseudogene 2|cytochrome P450, subfamily IIA (phenobarbital-inducible), polypeptide 7, pseudogene 1|cytochrome P450, subfamily IIA (phenobarbital-inducible), polypeptide 7, pseudogene 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41024264, end_position_on_the_genomic_accession=41028084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1551, CurrentID=0, Status='live', Symbol='CYP3A7', Aliases='CP37, CYPIIIA7, P-450(HFL33), P-450111A7, P450-HFLA', description='cytochrome P450 family 3 subfamily A member 7', other_designations='aryl hydrocarbon hydroxylase|cytochrome P450, family 3, subfamily A, polypeptide 7|cytochrome P450, subfamily IIIA, polypeptide 7|cytochrome P450-HFLA|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99705037, end_position_on_the_genomic_accession=99735200, orientation='minus', exon_count=13, OMIM=605340, no_hdr0=''),
  100271632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271632, CurrentID=0, Status='live', Symbol='RPL35AP32', Aliases='RPL35A_12_1446', description='ribosomal protein L35a pseudogene 32', other_designations='', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66446453, end_position_on_the_genomic_accession=66446840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1553, CurrentID=0, Status='live', Symbol='CYP2A13', Aliases='CPAD, CYP2A, CYPIIA13', description='cytochrome P450 family 2 subfamily A member 13', other_designations='cytochrome P450, family 2, subfamily A, polypeptide 13|cytochrome P450, subfamily IIA (phenobarbital-inducible), polypeptide 13', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41088451, end_position_on_the_genomic_accession=41096195, orientation='plus', exon_count=9, OMIM=608055, no_hdr0=''),
  100271634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271634, CurrentID=0, Status='live', Symbol='RPL35AP33', Aliases='RPL35A_13_1491', description='ribosomal protein L35a pseudogene 33', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68477192, end_position_on_the_genomic_accession=68477659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1555, CurrentID=0, Status='live', Symbol='CYP2B6', Aliases='CPB6, CYP2B, CYP2B7, CYP2B7P, CYPIIB6, EFVM, IIB1, P450', description='cytochrome P450 family 2 subfamily B member 6', other_designations='1,4-cineole 2-exo-monooxygenase|cytochrome P450 IIB1|cytochrome P450, family 2, subfamily B, polypeptide 6|cytochrome P450, subfamily IIB (phenobarbital-inducible), polypeptide 6', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40991229, end_position_on_the_genomic_accession=41018398, orientation='plus', exon_count=11, OMIM=123930, no_hdr0=''),
  1556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1556, CurrentID=0, Status='live', Symbol='CYP2B7P', Aliases='CYP2B, CYP2B71, CYP2B7P', description='cytochrome P450 family 2 subfamily B member 7, pseudogene', other_designations='cytochrome P450 2B7 isoform|cytochrome P450 2B7 short isoform|cytochrome P450, family 2, subfamily B, polypeptide 7 pseudogene 1|cytochrome P450, family 2, subfamily B, polypeptide 7, pseudogene|cytochrome P450, subfamily IIB (phenobarbital-inducible), polypeptide 7', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40924265, end_position_on_the_genomic_accession=40950660, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  1557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1557, CurrentID=0, Status='live', Symbol='CYP2C19', Aliases='CPCJ, CYP2C, CYPIIC17, CYPIIC19, P450C2C, P450IIC19', description='cytochrome P450 family 2 subfamily C member 19', other_designations="(R)-limonene 6-monooxygenase|(S)-limonene 6-monooxygenase|(S)-limonene 7-monooxygenase|S-mephenytoin 4-hydroxylase|cytochrome P-450 II C|cytochrome P450, family 2, subfamily C, polypeptide 19|cytochrome P450, subfamily IIC (mephenytoin 4-hydroxylase), polypeptide 19|cytochrome P450-11A|cytochrome P450-254C|flavoprotein-linked monooxygenase|mephenytoin 4'-hydroxylase|mephenytoin 4-hydroxylase|microsomal monooxygenase|xenobiotic monooxygenase", map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94762681, end_position_on_the_genomic_accession=94853205, orientation='plus', exon_count=9, OMIM=124020, no_hdr0=''),
  1558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1558, CurrentID=0, Status='live', Symbol='CYP2C8', Aliases='CPC8, CYPIIC8, MP-12/MP-20', description='cytochrome P450 family 2 subfamily C member 8', other_designations='P450 form 1|cytochrome P450 IIC2|cytochrome P450 MP-12|cytochrome P450 MP-20|cytochrome P450 form 1|cytochrome P450, family 2, subfamily C, polypeptide 8|cytochrome P450, subfamily IIC (mephenytoin 4-hydroxylase), polypeptide 8|flavoprotein-linked monooxygenase|microsomal monooxygenase|s-mephenytoin 4-hydroxylase|xenobiotic monooxygenase', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95036772, end_position_on_the_genomic_accession=95069498, orientation='minus', exon_count=11, OMIM=601129, no_hdr0=''),
  1559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1559, CurrentID=0, Status='live', Symbol='CYP2C9', Aliases='CPC9, CYP2C, CYP2C10, CYPIIC9, P450IIC9', description='cytochrome P450 family 2 subfamily C member 9', other_designations='cholesterol 25-hydroxylase|cytochrome P-450 S-mephenytoin 4-hydroxylase|cytochrome P-450MP|cytochrome P450 PB-1|cytochrome P450, family 2, subfamily C, polypeptide 9|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94938658, end_position_on_the_genomic_accession=94989391, orientation='plus', exon_count=9, OMIM=601130, no_hdr0=''),
  100271640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271640, CurrentID=0, Status='live', Symbol='RPL35AP31', Aliases='RPL35A_18_1321', description='ribosomal protein L35a pseudogene 31', other_designations='', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71845920, end_position_on_the_genomic_accession=71846352, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271641, CurrentID=0, Status='live', Symbol='RPL35AP34', Aliases='RPL35A_19_1474', description='ribosomal protein L35a pseudogene 34', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12881136, end_position_on_the_genomic_accession=12881540, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1562, CurrentID=0, Status='live', Symbol='CYP2C18', Aliases='CPCI, CYP2C, CYP2C17, P450-6B/29C, P450IIC17', description='cytochrome P450 family 2 subfamily C member 18', other_designations='(S)-mephenytoin hydroxylase associated cytochrome P450|CYPIIC18|cytochrome P450, family 2, subfamily C, polypeptide 18|cytochrome P450, subfamily IIC (mephenytoin 4-hydroxylase), polypeptide 17|cytochrome P450, subfamily IIC (mephenytoin 4-hydroxylase), polypeptide 18|cytochrome P450-6B/29C|flavoprotein-linked monooxygenase|microsomal monooxygenase|unspecific monooxygenase', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94683494, end_position_on_the_genomic_accession=94736190, orientation='plus', exon_count=9, OMIM=601131, no_hdr0=''),
  100271643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271643, CurrentID=0, Status='live', Symbol='RPL36AP39', Aliases='RPL36A_17_1112', description='ribosomal protein L36a pseudogene 39', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1792525, end_position_on_the_genomic_accession=1792920, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1564, CurrentID=0, Status='live', Symbol='CYP2D7', Aliases='CYP2DAP, CYP2D7P, CYP2D7P1, CYP2D@, P450C2D, P450DB1, RNA40057, CYP2D7', description='cytochrome P450 family 2 subfamily D member 7 (gene/pseudogene)', other_designations='Putative cytochrome P450 2D7|cytochrome P450, family 2, subfamily D, polypeptide 7 pseudogene 1|cytochrome P450, family 2, subfamily D, polypeptide 7, pseudogene|cytochrome P450, subfamily II (debrisoquine, sparteine, etc., -metabolising), polypeptide 7 pseudogene 1|cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolizing) cluster', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42140203, end_position_on_the_genomic_accession=42144575, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  1565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1565, CurrentID=0, Status='live', Symbol='CYP2D6', Aliases='CPD6, CYP2D, CYP2D7AP, CYP2D7BP, CYP2D7P2, CYP2D8P2, CYP2DL1, CYPIID6, P450-DB1, P450C2D, P450DB1', description='cytochrome P450 family 2 subfamily D member 6', other_designations='cholesterol 25-hydroxylase|cytochrome P450, family 2, subfamily D, polypeptide 6|cytochrome P450, family 2, subfamily D, polypeptide 7 pseudogene 2|cytochrome P450, family 2, subfamily D, polypeptide 8 pseudogene 2|cytochrome P450, subfamily II (debrisoquine, sparteine, etc., -metabolising), polypeptide 7 pseudogene 2|cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolising), polypeptide 8 pseudogene 2|cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolizing), polypeptide 6|cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolizing)-like 1|cytochrome P450-DB1|debrisoquine 4-hydroxylase|flavoprotein-linked monooxygenase|microsomal monooxygenase|nonfunctional cytochrome P450 2D6|xenobiotic monooxygenase', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42125656, end_position_on_the_genomic_accession=42135378, orientation='minus', exon_count=12, OMIM=124030, no_hdr0=''),
  1566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1566, CurrentID=0, Status='live', Symbol='CYP2D7BP', Aliases='', description='cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolising), polypeptide 7b (pseudogene)', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271647, CurrentID=0, Status='live', Symbol='RPL36AP47', Aliases='RPL36A_22_1567', description='ribosomal protein L36a pseudogene 47', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62319880, end_position_on_the_genomic_accession=62320254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1568, CurrentID=0, Status='live', Symbol='CYP2D8P', Aliases='CYP2D8P1, CYP2DP1', description='ccytochrome P450 family 2 subfamily D member 8, pseudogene', other_designations='cytochrome P450, family 2, subfamily D, polypeptide 8 pseudogene 1|cytochrome P450, family 2, subfamily D, polypeptide 8, pseudogene|cytochrome P450, subfamily IID (debrisoquine, sparteine, etc., -metabolizing), polypeptide 8 pseudogene 1', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42149883, end_position_on_the_genomic_accession=42155091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271649, CurrentID=0, Status='live', Symbol='RPS15AP32', Aliases='RPS15A_10_1280', description='ribosomal protein S15a pseudogene 32', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113054062, end_position_on_the_genomic_accession=113054438, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271650, CurrentID=0, Status='live', Symbol='RPS15AP36', Aliases='RPS15A_12_1675', description='ribosomal protein S15a pseudogene 36', other_designations='', map_location='19qter', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58216147, end_position_on_the_genomic_accession=58216615, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1571, CurrentID=0, Status='live', Symbol='CYP2E1', Aliases='CPE1, CYP2E, P450-J, P450C2E', description='cytochrome P450 family 2 subfamily E member 1', other_designations='4-nitrophenol 2-hydroxylase|CYPIIE1|cytochrome P450, family 2, subfamily E, polypeptide 1|cytochrome P450, subfamily IIE (ethanol-inducible), polypeptide 1|cytochrome P450-J|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133527363, end_position_on_the_genomic_accession=133539116, orientation='plus', exon_count=9, OMIM=124040, no_hdr0=''),
  1572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1572, CurrentID=0, Status='live', Symbol='CYP2F1', Aliases='C2F1, CYP2F', description='cytochrome P450 family 2 subfamily F member 1', other_designations='CYPIIF1|cytochrome P450, family 2, subfamily F, polypeptide 1|cytochrome P450, subfamily IIF, polypeptide 1|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41114292, end_position_on_the_genomic_accession=41128381, orientation='plus', exon_count=11, OMIM=124070, no_hdr0=''),
  1573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1573, CurrentID=0, Status='live', Symbol='CYP2J2', Aliases='CPJ2', description='cytochrome P450 family 2 subfamily J member 2', other_designations='CYPIIJ2|arachidonic acid epoxygenase|cytochrome P450, family 2, subfamily J, polypeptide 2|cytochrome P450, subfamily IIJ (arachidonic acid epoxygenase) polypeptide 2|flavoprotein-linked monooxygenase|microsomal monooxygenase', map_location='1p31.3-p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59893308, end_position_on_the_genomic_accession=59926773, orientation='minus', exon_count=10, OMIM=601258, no_hdr0=''),
  1574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1574, CurrentID=0, Status='live', Symbol='CYP3A', Aliases='CYP3', description='cytochrome P450, family 3, subfamily A', other_designations='cytochrome P450, subfamily IIIA (niphedipine oxidase)', map_location='7q21.3-q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271655, CurrentID=0, Status='live', Symbol='RPS15AP38', Aliases='RPS15A_25_1749', description='ribosomal protein S15a pseudogene 38', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36421230, end_position_on_the_genomic_accession=36421661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1576, CurrentID=0, Status='live', Symbol='CYP3A4', Aliases='CP33, CP34, CYP3A, CYP3A3, CYPIIIA3, CYPIIIA4, HLP, NF-25, P450C3, P450PCN1', description='cytochrome P450 family 3 subfamily A member 4', other_designations='1,8-cineole 2-exo-monooxygenase|P450-III, steroid inducible|albendazole monooxygenase|albendazole sulfoxidase|cholesterol 25-hydroxylase|cytochrome P450 3A3|cytochrome P450 HLp|cytochrome P450 NF-25|cytochrome P450, family 3, subfamily A, polypeptide 4|cytochrome P450, subfamily IIIA (niphedipine oxidase), polypeptide 3|cytochrome P450, subfamily IIIA (niphedipine oxidase), polypeptide 4|cytochrome P450-PCN1|glucocorticoid-inducible P450|nifedipine oxidase|quinine 3-monooxygenase|taurochenodeoxycholate 6-alpha-hydroxylase', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99756960, end_position_on_the_genomic_accession=99784188, orientation='minus', exon_count=14, OMIM=124010, no_hdr0=''),
  1577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1577, CurrentID=0, Status='live', Symbol='CYP3A5', Aliases='CP35, CYPIIIA5, P450PCN3, PCN3', description='cytochrome P450 family 3 subfamily A member 5', other_designations='aryl hydrocarbon hydroxylase|cytochrome P450 HLp2|cytochrome P450, family 3, subfamily A, polypeptide 5|cytochrome P450, subfamily IIIA (niphedipine oxidase), polypeptide 5|cytochrome P450-PCN3|flavoprotein-linked monooxygenase|microsomal monooxygenase|xenobiotic monooxygenase', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99648189, end_position_on_the_genomic_accession=99680026, orientation='minus', exon_count=18, OMIM=605325, no_hdr0=''),
  1578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1578, CurrentID=0, Status='live', Symbol='CYP3A51P', Aliases='CYP3A5-de13c, CYP3A5-de1b2b, CYP3A5P1, CYP3AP1', description='cytochrome P450 family 3 subfamily A member 51, pseudogene', other_designations='cytochrome P450, family 3, subfamily A, polypeptide 5 pseudogene 1|cytochrome P450, family 3, subfamily A, polypeptide 51, pseudogene|cytochrome P450, subfamily IIIA (niphedipine oxidase), pseudogene 1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99685024, end_position_on_the_genomic_accession=99700120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1579, CurrentID=0, Status='live', Symbol='CYP4A11', Aliases='CP4Y, CYP4A2, CYP4AII', description='cytochrome P450 family 4 subfamily A member 11', other_designations='20-HETE synthase|20-hydroxyeicosatetraenoic acid synthase|CYPIVA11|P450HL-omega|alkane-1 monooxygenase|cytochrome P-450HK-omega|cytochrome P450, family 4, subfamily A, polypeptide 11|cytochrome P450, subfamily IVA, polypeptide 11|cytochrome P450HL-omega|fatty acid omega-hydroxylase|lauric acid omega-hydroxylase|long-chain fatty acid omega-monooxygenase', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46929174, end_position_on_the_genomic_accession=46941499, orientation='minus', exon_count=13, OMIM=601310, no_hdr0=''),
  1580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1580, CurrentID=0, Status='live', Symbol='CYP4B1', Aliases='CYPIVB1, P-450HP', description='cytochrome P450 family 4 subfamily B member 1', other_designations='cytochrome P450, family 4, subfamily B, polypeptide 1|cytochrome P450, subfamily IVB, polypeptide 1|cytochrome P450-HP|microsomal monooxygenase', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46798995, end_position_on_the_genomic_accession=46819349, orientation='plus', exon_count=12, OMIM=124075, no_hdr0=''),
  1581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1581, CurrentID=0, Status='live', Symbol='CYP7A1', Aliases='CP7A, CYP7, CYPVII', description='cytochrome P450 family 7 subfamily A member 1', other_designations='cholesterol 7-alpha-hydroxylase|cytochrome P450 7A1|cytochrome P450, family 7, subfamily A, polypeptide 1|cytochrome P450, subfamily VIIA polypeptide 1', map_location='8q11-q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58490178, end_position_on_the_genomic_accession=58500959, orientation='minus', exon_count=7, OMIM=118455, no_hdr0=''),
  1582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1582, CurrentID=0, Status='live', Symbol='CYP8B1', Aliases='CP8B, CYP12', description='cytochrome P450 family 8 subfamily B member 1', other_designations='7 alpha-hydroxy-4-cholesten-3-one 12-alpha-hydroxylase|7-alpha-hydroxy-4-cholesten-3-one 12-alpha-hydroxylase|CYPVIIIB1|cytochrome P450 8B1|cytochrome P450, family 8, subfamily B, polypeptide 1|cytochrome P450, subfamily VIIIB (sterol 12-alpha-hydroxylase), polypeptide 1|sterol 12-alpha-hydroxylase', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42872192, end_position_on_the_genomic_accession=42876141, orientation='minus', exon_count=1, OMIM=602172, no_hdr0=''),
  1583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1583, CurrentID=0, Status='live', Symbol='CYP11A1', Aliases='CYP11A, CYPXIA1, P450SCC', description='cytochrome P450 family 11 subfamily A member 1', other_designations='cholesterol 20-22 desmolase|cholesterol monooxygenase (side-chain cleaving)|cytochrome P450 11A1|cytochrome P450 family 11 subfamily A polypeptide 1|cytochrome P450(scc)|cytochrome P450, family 11, subfamily A, polypeptide 1|cytochrome P450, subfamily XIA (cholesterol side chain cleavage)|cytochrome P450C11A1|steroid 20-22-lyase', map_location='15q23-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74337762, end_position_on_the_genomic_accession=74367740, orientation='minus', exon_count=10, OMIM=118485, no_hdr0=''),
  1584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1584, CurrentID=0, Status='live', Symbol='CYP11B1', Aliases='CPN1, CYP11B, FHI, P450C11', description='cytochrome P450 family 11 subfamily B member 1', other_designations='CYPXIB1|cytochrome P-450c11|cytochrome P450, family 11, subfamily B, polypeptide 1|cytochrome P450, subfamily XIB (steroid 11-beta-hydroxylase), polypeptide 1|cytochrome P450C11|cytochrome p450 XIB1|steroid 11-beta-hydroxylase|steroid 11-beta-monooxygenase', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142872354, end_position_on_the_genomic_accession=142879846, orientation='minus', exon_count=11, OMIM=610613, no_hdr0=''),
  1585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1585, CurrentID=0, Status='live', Symbol='CYP11B2', Aliases='ALDOS, CPN2, CYP11B, CYP11BL, CYPXIB2, P-450C18, P450C18, P450aldo', description='cytochrome P450 family 11 subfamily B member 2', other_designations='aldosterone synthase|aldosterone-synthesizing enzyme|cytochrome P-450Aldo|cytochrome P-450C18|cytochrome P450, family 11, subfamily B, polypeptide 2|cytochrome P450, subfamily XIB (steroid 11-beta-hydroxylase), polypeptide 2|mitochondrial cytochrome P450, family 11, subfamily B, polypeptide 2|steroid 11-beta-monooxygenase|steroid 11-beta/18-hydroxylase|steroid 18-hydroxylase, aldosterone synthase, P450C18, P450aldo', map_location='8q21-q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142910558, end_position_on_the_genomic_accession=142917862, orientation='minus', exon_count=10, OMIM=124080, no_hdr0=''),
  1586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1586, CurrentID=0, Status='live', Symbol='CYP17A1', Aliases='CPT7, CYP17, P450C17, S17AH', description='cytochrome P450 family 17 subfamily A member 1', other_designations='17-alpha-hydroxyprogesterone aldolase|CYPXVII|cytochrome P450 17A1|cytochrome P450, family 17, subfamily A, polypeptide 1|cytochrome P450, subfamily XVII (steroid 17-alpha-hydroxylase), adrenal hyperplasia|cytochrome P450-C17|cytochrome P450c17|cytochrome p450 XVIIA1|steroid 17-alpha-monooxygenase', map_location='10q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102830531, end_position_on_the_genomic_accession=102837533, orientation='minus', exon_count=8, OMIM=609300, no_hdr0=''),
  1587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1587, CurrentID=0, Status='live', Symbol='ADAM3A', Aliases='ADAM3, CYRN1, tMDCI', description='ADAM metallopeptidase domain 3A (pseudogene)', other_designations='ADAM metallopeptidase domain 3A, pseudogene|a disintegrin and metalloproteinase domain 3a (cyritestin 1)|cyritestin 1', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39451045, end_position_on_the_genomic_accession=39522989, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  1588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1588, CurrentID=0, Status='live', Symbol='CYP19A1', Aliases='ARO, ARO1, CPV1, CYAR, CYP19, CYPXIX, P-450AROM', description='cytochrome P450 family 19 subfamily A member 1', other_designations='cytochrome P-450AROM|cytochrome P450 19A1|cytochrome P450, family 19, subfamily A, polypeptide 1|cytochrome P450, subfamily XIX (aromatization of androgens)|estrogen synthase|estrogen synthetase|flavoprotein-linked monooxygenase|microsomal monooxygenase', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51208057, end_position_on_the_genomic_accession=51338598, orientation='minus', exon_count=13, OMIM=107910, no_hdr0=''),
  1589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1589, CurrentID=0, Status='live', Symbol='CYP21A2', Aliases='CA21H, CAH1, CPS1, CYP21, CYP21B, P450c21B', description='cytochrome P450 family 21 subfamily A member 2', other_designations='21-OHase|cytochrome P450 XXI|cytochrome P450, family 21, subfamily A, polypeptide 2|cytochrome P450, subfamily XXIA (steroid 21-hydroxylase, congenital adrenal hyperplasia), polypeptide 2|cytochrome P450-C21B|steroid 21 hydroxylase|steroid 21-monooxygenase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32038316, end_position_on_the_genomic_accession=32041670, orientation='plus', exon_count=10, OMIM=613815, no_hdr0=''),
  1590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1590, CurrentID=0, Status='live', Symbol='CYP21A1P', Aliases='CYP21A, CYP21P, P450c21A', description='cytochrome P450 family 21 subfamily A member 1, pseudogene', other_designations='cytochrome P450, family 21, subfamily A, polypeptide 1 pseudogene|cytochrome P450, subfamily XXI (steroid 21-hydroxylase) pseudogene|cytochrome P450, subfamily XXIA (steroid 21-hydroxylase), polypeptide 1 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32005636, end_position_on_the_genomic_accession=32008909, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  1591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1591, CurrentID=0, Status='live', Symbol='CYP24A1', Aliases='CP24, CYP24, HCAI, P450-CC24', description='cytochrome P450 family 24 subfamily A member 1', other_designations='1,25-@dihydroxyvitamin D3 24-hydroxylase|24-OHase|cytochrome P450 24A1|cytochrome P450, family 24, subfamily A, polypeptide 1|cytochrome P450, subfamily XXIV (vitamin D 24-hydroxylase)|cytochrome P450-CC24|exo-mitochondrial protein|vitamin D 24-hydroxylase|vitamin D(3) 24-hydroxylase', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=54153449, end_position_on_the_genomic_accession=54173985, orientation='minus', exon_count=12, OMIM=126065, no_hdr0=''),
  1592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1592, CurrentID=0, Status='live', Symbol='CYP26A1', Aliases='CP26, CYP26, P450RAI, P450RAI1', description='cytochrome P450 family 26 subfamily A member 1', other_designations='P450, retinoic acid-inactivating, 1|cytochrome P450 retinoic acid-inactivating 1|cytochrome P450, family 26, subfamily A, polypeptide 1|cytochrome P450, subfamily XXVIA, polypeptide 1|cytochrome P450RAI|hP450RAI|retinoic acid 4-hydroxylase|retinoic acid-metabolizing cytochrome', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93073475, end_position_on_the_genomic_accession=93077884, orientation='plus', exon_count=8, OMIM=602239, no_hdr0=''),
  1593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1593, CurrentID=0, Status='live', Symbol='CYP27A1', Aliases='CP27, CTX, CYP27', description='cytochrome P450 family 27 subfamily A member 1', other_designations='5-beta-cholestane-3-alpha, 7-alpha, 12-alpha-triol 26-hydroxylase|5-beta-cholestane-3-alpha, 7-alpha, 12-alpha-triol 27-hydroxylase|5-beta-cholestane-3-alpha,7-alpha,12-alpha-triol 27-hydroxylase|cholestanetriol 26-monooxygenase|cytochrome P-450C27/25|cytochrome P450 27|cytochrome P450, family 27, subfamily A, polypeptide 1|cytochrome P450, subfamily XXVIIA (steroid 27-hydroxylase, cerebrotendinous xanthomatosis), polypeptide 1|sterol 27-hydroxylase|vitamin D(3) 25-hydroxylase', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218781749, end_position_on_the_genomic_accession=218815293, orientation='plus', exon_count=9, OMIM=606530, no_hdr0=''),
  1594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1594, CurrentID=0, Status='live', Symbol='CYP27B1', Aliases='CP2B, CYP1, CYP1alpha, CYP27B, P450c1, PDDR, VDD1, VDDR, VDDRI, VDR', description='cytochrome P450 family 27 subfamily B member 1', other_designations='1alpha(OH)ase|25 hydroxyvitamin D3-1-alpha hydroxylase|25-OHD-1 alpha-hydroxylase|VD3 1A hydroxylase|calcidiol 1-monooxygenase|cytochrome P450 subfamily XXVIIB polypeptide 1|cytochrome P450, family 27, subfamily B, polypeptide 1|cytochrome P450C1 alpha|cytochrome P450VD1-alpha|cytochrome p450 27B1', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57762334, end_position_on_the_genomic_accession=57767193, orientation='minus', exon_count=9, OMIM=609506, no_hdr0=''),
  1595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1595, CurrentID=0, Status='live', Symbol='CYP51A1', Aliases='CP51, CYP51, CYPL1, LDM, P450-14DM, P450L1', description='cytochrome P450 family 51 subfamily A member 1', other_designations='CYPLI|cytochrome P450 51A1|cytochrome P450, 51 (lanosterol 14-alpha-demethylase)|cytochrome P450, family 51, subfamily A, polypeptide 1|cytochrome P450-14DM|cytochrome P45014DM|cytochrome P450LI|lanosterol 14-alpha-demethylase|sterol 14-alpha demethylase', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92112149, end_position_on_the_genomic_accession=92134745, orientation='minus', exon_count=11, OMIM=601637, no_hdr0=''),
  1596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1596, CurrentID=0, Status='live', Symbol='ADAM3B', Aliases='CYRN2', description='ADAM metallopeptidase domain 3B (pseudogene)', other_designations='ADAM metallopeptidase domain 3B (cyritestin 2)|ADAM metallopeptidase domain 3B (non-functional)|a disintegrin and metalloproteinase domain 3b (cyritestin 2)|cyritestin 2', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49517646, end_position_on_the_genomic_accession=49518352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132671, CurrentID=0, Status='live', Symbol='SPATA18', Aliases='Mieap, SPETEX1', description='spermatogenesis associated 18', other_designations='spermatogenesis associated 18 homolog|spermatogenesis-associated protein 18', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52051331, end_position_on_the_genomic_accession=52097305, orientation='plus', exon_count=15, OMIM=612814, no_hdr0=''),
  1600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1600, CurrentID=0, Status='live', Symbol='DAB1', Aliases='', description='Dab, reelin signal transducer, homolog 1 (Drosophila)', other_designations='Dab reelin signal transducer 1', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56997906, end_position_on_the_genomic_accession=58250539, orientation='minus', exon_count=17, OMIM=603448, no_hdr0=''),
  1601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1601, CurrentID=0, Status='live', Symbol='DAB2', Aliases='DOC-2, DOC2', description='Dab, mitogen-responsive phosphoprotein, homolog 2 (Drosophila)', other_designations='adaptor molecule disabled-2|differentially expressed in ovarian carcinoma 2|differentially-expressed protein 2|disabled homolog 2, mitogen-responsive phosphoprotein', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39371674, end_position_on_the_genomic_accession=39425233, orientation='minus', exon_count=15, OMIM=601236, no_hdr0=''),
  1602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1602, CurrentID=0, Status='live', Symbol='DACH1', Aliases='DACH', description='dachshund family transcription factor 1', other_designations='dac homolog', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71437966, end_position_on_the_genomic_accession=71869113, orientation='minus', exon_count=15, OMIM=603803, no_hdr0=''),
  1603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1603, CurrentID=0, Status='live', Symbol='DAD1', Aliases='OST2', description='defender against cell death 1', other_designations='DAD-1|oligosaccharyl transferase subunit DAD1|oligosaccharyltransferase 2 homolog|oligosaccharyltransferase subunit 2 (non-catalytic)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22564907, end_position_on_the_genomic_accession=22589237, orientation='minus', exon_count=3, OMIM=600243, no_hdr0=''),
  1604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1604, CurrentID=0, Status='live', Symbol='CD55', Aliases='CR, CROM, DAF, TC', description='CD55 molecule, decay accelerating factor for complement (Cromer blood group)', other_designations='CD55 antigen', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207321472, end_position_on_the_genomic_accession=207360966, orientation='plus', exon_count=14, OMIM=125240, no_hdr0=''),
  1605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1605, CurrentID=0, Status='live', Symbol='DAG1', Aliases='156DAG, A3a, AGRNR, DAG, MDDGC7, MDDGC9', description='dystroglycan 1', other_designations='dystroglycan 1 (dystrophin-associated glycoprotein 1)', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49468703, end_position_on_the_genomic_accession=49535618, orientation='plus', exon_count=9, OMIM=128239, no_hdr0=''),
  1606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1606, CurrentID=0, Status='live', Symbol='DGKA', Aliases='DAGK, DAGK1, DGK-alpha', description='diacylglycerol kinase alpha', other_designations='80 kDa diacylglycerol kinase|DAG kinase alpha|diacylglycerol kinase, alpha 80kDa|diglyceride kinase alpha', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55931162, end_position_on_the_genomic_accession=55954027, orientation='plus', exon_count=30, OMIM=125855, no_hdr0=''),
  1607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1607, CurrentID=0, Status='live', Symbol='DGKB', Aliases='DAGK2, DGK, DGK-BETA', description='diacylglycerol kinase beta', other_designations='90 kDa diacylglycerol kinase|DAG kinase beta|diacylglycerol kinase, beta 90kDa|diglyceride kinase beta', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14145049, end_position_on_the_genomic_accession=14974777, orientation='minus', exon_count=31, OMIM=604070, no_hdr0=''),
  1608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1608, CurrentID=0, Status='live', Symbol='DGKG', Aliases='DAGK3, DGK-GAMMA', description='diacylglycerol kinase gamma', other_designations='DAG kinase gamma|diacylglycerol kinase, gamma 90kDa|diacylglyerol kinase gamma|diglyceride kinase gamma', map_location='3q27.2-q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186147201, end_position_on_the_genomic_accession=186362234, orientation='minus', exon_count=25, OMIM=601854, no_hdr0=''),
  1609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1609, CurrentID=0, Status='live', Symbol='DGKQ', Aliases='DAGK, DAGK4, DAGK7', description='diacylglycerol kinase theta', other_designations='DAG kinase theta|DGK-theta|diacylglycerol kinase, theta 110kDa|diglyceride kinase theta', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=958884, end_position_on_the_genomic_accession=973560, orientation='minus', exon_count=25, OMIM=601207, no_hdr0=''),
  1610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1610, CurrentID=0, Status='live', Symbol='DAO', Aliases='DAAO, DAMOX, OXDA', description='D-amino-acid oxidase', other_designations='D-amino acid oxidase', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108880030, end_position_on_the_genomic_accession=108901043, orientation='plus', exon_count=12, OMIM=124050, no_hdr0=''),
  1611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1611, CurrentID=0, Status='live', Symbol='DAP', Aliases='', description='death-associated protein', other_designations='DAP-1', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10679230, end_position_on_the_genomic_accession=10761275, orientation='minus', exon_count=4, OMIM=600954, no_hdr0=''),
  1612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1612, CurrentID=0, Status='live', Symbol='DAPK1', Aliases='DAPK', description='death-associated protein kinase 1', other_designations='DAP kinase 1', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87497228, end_position_on_the_genomic_accession=87708634, orientation='plus', exon_count=29, OMIM=600831, no_hdr0=''),
  1613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1613, CurrentID=0, Status='live', Symbol='DAPK3', Aliases='DLK, ZIP, ZIPK', description='death-associated protein kinase 3', other_designations='DAP kinase 3|DAP-like kinase|MYPT1 kinase|ZIP-kinase|zipper-interacting protein kinase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3958453, end_position_on_the_genomic_accession=3971123, orientation='minus', exon_count=9, OMIM=603289, no_hdr0=''),
  100271694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271694, CurrentID=0, Status='live', Symbol='MS2', Aliases='', description='Multiple sclerosis, susceptibility to, 2', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612594, no_hdr0=''),
  1615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1615, CurrentID=0, Status='live', Symbol='DARS', Aliases='HBSL, aspRS', description='aspartyl-tRNA synthetase', other_designations='aspartate tRNA ligase 1, cytoplasmic|aspartyl-tRNA synthetase, cytoplasmic|cell proliferation-inducing gene 40 protein', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135905881, end_position_on_the_genomic_accession=135985684, orientation='minus', exon_count=16, OMIM=603084, no_hdr0=''),
  1616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1616, CurrentID=0, Status='live', Symbol='DAXX', Aliases='BING2, DAP6, EAP1', description='death-domain associated protein', other_designations='CENP-C binding protein|ETS1-associated protein 1|Fas-binding protein|death-associated protein 6|fas death domain-associated protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33318558, end_position_on_the_genomic_accession=33323016, orientation='minus', exon_count=8, OMIM=603186, no_hdr0=''),
  1617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1617, CurrentID=0, Status='live', Symbol='DAZ1', Aliases='DAZ, SPGY', description='deleted in azoospermia 1', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23129355, end_position_on_the_genomic_accession=23199123, orientation='minus', exon_count=28, OMIM=400003, no_hdr0=''),
  1618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1618, CurrentID=0, Status='live', Symbol='DAZL', Aliases='DAZH1, DAZLA, SPGYLA, DAZL', description='deleted in azoospermia-like', other_designations='DAZ homolog|DAZ-like autosomal|SPGY-like-autosomal|deleted in azoospermia-like 1|germline specific RNA binding protein|spermatogenesis gene on the Y-like autosomal', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16586792, end_position_on_the_genomic_accession=16605499, orientation='minus', exon_count=12, OMIM=601486, no_hdr0=''),
  100533843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533843, CurrentID=0, Status='live', Symbol='LOC100533843', Aliases='', description='karyopherin alpha 2 (RAG cohort 1, importin alpha 1) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65184951, end_position_on_the_genomic_accession=65186216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1620, CurrentID=0, Status='live', Symbol='BRINP1', Aliases='DBC1, DBCCR1, FAM5A', description='bone morphogenetic protein/retinoic acid inducible neural-specific 1', other_designations='bA574M5.1 (deleted in bladder cancer chromosome region candidate 1 (IB3089A))|bone morphogenic protein/retinoic acid inducible neural-specific 1|deleted in bladder cancer 1|deleted in bladder cancer chromosome region candidate 1|deleted in bladder cancer protein 1', map_location='9q32-q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=119166630, end_position_on_the_genomic_accession=119369461, orientation='minus', exon_count=8, OMIM=602865, no_hdr0=''),
  1621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1621, CurrentID=0, Status='live', Symbol='DBH', Aliases='DBM', description='dopamine beta-hydroxylase (dopamine beta-monooxygenase)', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133636363, end_position_on_the_genomic_accession=133659344, orientation='plus', exon_count=12, OMIM=609312, no_hdr0=''),
  1622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1622, CurrentID=0, Status='live', Symbol='DBI', Aliases='ACBD1, ACBP, CCK-RP, EP', description='diazepam binding inhibitor (GABA receptor modulator, acyl-CoA binding protein)', other_designations='GABA receptor modulator|acyl coenzyme A binding protein|acyl-Coenzyme A binding domain containing 1|cholecystokinin-releasing peptide, trypsin-sensitive|diazepam binding inhibitor (GABA receptor modulator, acyl-Coenzyme A binding protein)|diazepam-binding inhibitor|endozepine', map_location='2q12-q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119366924, end_position_on_the_genomic_accession=119372551, orientation='plus', exon_count=7, OMIM=125950, no_hdr0=''),
  100533847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533847, CurrentID=0, Status='live', Symbol='TYW1P1', Aliases='', description='tRNA-yW synthesizing protein 1 homolog pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33117264, end_position_on_the_genomic_accession=33119106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1624, CurrentID=0, Status='live', Symbol='DBIP1', Aliases='DBIL2', description='diazepam binding inhibitor (GABA receptor modulator, acyl-CoA binding protein) pseudogene 1', other_designations='diazepam binding inhibitor-like 2 (a pseudogene)|diazepam binding inhibitor-like 2 (pseudogene)', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79436685, end_position_on_the_genomic_accession=79437228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533849, CurrentID=0, Status='live', Symbol='MAPKAPK5P1', Aliases='', description='mitogen-activated protein kinase-activated protein kinase 5 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=108848203, end_position_on_the_genomic_accession=108849559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533850, CurrentID=0, Status='live', Symbol='LOC100533850', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28362787, end_position_on_the_genomic_accession=28363975, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1627, CurrentID=0, Status='live', Symbol='DBN1', Aliases='D0S117E', description='drebrin 1', other_designations='developmentally-regulated brain protein|drebrin E|drebrin E2|drebrin-1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177456613, end_position_on_the_genomic_accession=177480336, orientation='minus', exon_count=19, OMIM=126660, no_hdr0=''),
  1628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1628, CurrentID=0, Status='live', Symbol='DBP', Aliases='DABP', description='D site of albumin promoter (albumin D-box) binding protein', other_designations='albumin D box-binding protein|albumin D-element-binding protein|tax-responsive enhancer element-binding protein 302|taxREB302', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48630560, end_position_on_the_genomic_accession=48637550, orientation='minus', exon_count=4, OMIM=124097, no_hdr0=''),
  1629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1629, CurrentID=0, Status='live', Symbol='DBT', Aliases='BCATE2, BCKAD-E2, BCKADE2, BCOADC-E2, E2, E2B', description='dihydrolipoamide branched chain transacylase E2', other_designations='52 kDa mitochondrial autoantigen of primary biliary cirrhosis|BCKAD E2 subunit|E2 component of branched chain alpha-keto acid dehydrogenase complex|branched chain 2-oxo-acid dehydrogenase complex component E2|branched chain acyltransferase, E2 component|branched-chain alpha-keto acid dehydrogenase complex component E2|dihydrolipoamide acetyltransferase component of branched-chain alpha-keto acid dehydrogenase complex|dihydrolipoyl transacylase|dihydrolipoyllysine-residue (2-methylpropanoyl)transferase|lipoamide acyltransferase component of mitochondrial branched-chain alpha-keto acid dehydrogenase complex|mitochondrial branched chain alpha-keto acid dehydrogenase transacylase subunit (E2b)', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100186922, end_position_on_the_genomic_accession=100249863, orientation='minus', exon_count=13, OMIM=248610, no_hdr0=''),
  1630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1630, CurrentID=0, Status='live', Symbol='DCC', Aliases='CRC18, CRCR1, IGDCC1, MRMV1, NTN1R1', description='DCC netrin 1 receptor', other_designations='colorectal cancer suppressor|colorectal tumor suppressor|deleted in colorectal cancer protein|deleted in colorectal carcinoma|immunoglobulin superfamily DCC subclass member 1|immunoglobulin superfamily, DCC subclass, member 1|tumor suppressor protein DCC', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=52340172, end_position_on_the_genomic_accession=53535903, orientation='plus', exon_count=33, OMIM=120470, no_hdr0=''),
  1632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1632, CurrentID=0, Status='live', Symbol='ECI1', Aliases='DCI', description='enoyl-CoA delta isomerase 1', other_designations='3,2 trans-enoyl-CoA isomerase|3,2 trans-enoyl-Coenzyme A isomerase|3,2-trans-enoyl-CoA isomerase, mitochondrial|D3,D2-enoyl-CoA isomerase|acetylene-allene isomerase|delta(3),Delta(2)-enoyl-CoA isomerase|delta3, delta2-enoyl-CoA isomerase|dodecenoyl-CoA delta isomerase (3,2 trans-enoyl-CoA isomerase)|dodecenoyl-CoA isomerase|dodecenoyl-Coenzyme A delta isomerase (3,2 trans-enoyl-Coenzyme A isomerase)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2239872, end_position_on_the_genomic_accession=2251601, orientation='minus', exon_count=7, OMIM=600305, no_hdr0=''),
  1633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1633, CurrentID=0, Status='live', Symbol='DCK', Aliases='', description='deoxycytidine kinase', other_designations='deoxynucleoside kinase', map_location='4q13.3-q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70993548, end_position_on_the_genomic_accession=71030914, orientation='plus', exon_count=7, OMIM=125450, no_hdr0=''),
  1634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1634, CurrentID=0, Status='live', Symbol='DCN', Aliases='CSCD, DSPG2, PG40, PGII, PGS2, SLRR1B', description='decorin', other_designations='bone proteoglycan II|dermatan sulphate proteoglycans II|proteoglycan core protein|small leucine-rich protein 1B', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91145258, end_position_on_the_genomic_accession=91183029, orientation='minus', exon_count=11, OMIM=125255, no_hdr0=''),
  1635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1635, CurrentID=0, Status='live', Symbol='DCTD', Aliases='', description='dCMP deaminase', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182890087, end_position_on_the_genomic_accession=182917477, orientation='minus', exon_count=9, OMIM=607638, no_hdr0=''),
  1636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1636, CurrentID=0, Status='live', Symbol='ACE', Aliases='ACE1, CD143, DCP, DCP1, ICH, MVCD3', description='angiotensin I converting enzyme', other_designations='CD143 antigen|angiotensin I converting enzyme (peptidyl-dipeptidase A) 1|angiotensin I converting enzyme peptidyl-dipeptidase A 1 transcript|angiotensin converting enzyme, somatic isoform|carboxycathepsin|dipeptidyl carboxypeptidase 1|dipeptidyl carboxypeptidase I|kininase II|peptidase P|testicular ECA', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63477061, end_position_on_the_genomic_accession=63498380, orientation='plus', exon_count=26, OMIM=106180, no_hdr0=''),
  1637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1637, CurrentID=0, Status='live', Symbol='DCR', Aliases='DSCR', description='Down syndrome chromosome region', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=190685, no_hdr0=''),
  1638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1638, CurrentID=0, Status='live', Symbol='DCT', Aliases='TRP-2, TYRP2', description='dopachrome tautomerase', other_designations='DT|L-dopachrome Delta-isomerase|L-dopachrome isomerase|TRP2|dopachrome delta-isomerase|dopachrome tautomerase (dopachrome delta-isomerase, tyrosine-related protein 2)|tyrosinase related protein-2|tyrosinase-related protein 2|tyrosine-related protein 2', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94437304, end_position_on_the_genomic_accession=94550265, orientation='minus', exon_count=13, OMIM=191275, no_hdr0=''),
  1639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1639, CurrentID=0, Status='live', Symbol='DCTN1', Aliases='DAP-150, DP-150, P135', description='dynactin subunit 1', other_designations='150 kDa dynein-associated polypeptide|dynactin 1 (p150, glued homolog, Drosophila)', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74361154, end_position_on_the_genomic_accession=74392087, orientation='minus', exon_count=33, OMIM=601143, no_hdr0=''),
  1641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1641, CurrentID=0, Status='live', Symbol='DCX', Aliases='DBCN, DC, LISX, SCLH, XLIS', description='doublecortin', other_designations='doublecortex|doublin|lis-X|lissencephalin-X', map_location='Xq22.3-q23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111293779, end_position_on_the_genomic_accession=111412232, orientation='minus', exon_count=8, OMIM=300121, no_hdr0=''),
  1642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1642, CurrentID=0, Status='live', Symbol='DDB1', Aliases='DDBA, UV-DDB1, XAP1, XPCE, XPE, XPE-BF', description='damage-specific DNA binding protein 1', other_designations='DDB p127 subunit|DNA damage-binding protein a|HBV X-associated protein 1|UV-DDB 1|UV-damaged DNA-binding factor|UV-damaged DNA-binding protein 1|XAP-1|XPE-binding factor|damage-specific DNA binding protein 1, 127kDa|xeroderma pigmentosum group E-complementing protein', map_location='11q12-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61299447, end_position_on_the_genomic_accession=61333212, orientation='minus', exon_count=27, OMIM=600045, no_hdr0=''),
  1643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1643, CurrentID=0, Status='live', Symbol='DDB2', Aliases='DDBB, UV-DDB2, XPE', description='damage-specific DNA binding protein 2', other_designations='DDB p48 subunit|UV-DDB 2|UV-damaged DNA-binding protein 2|damage-specific DNA binding protein 2, 48kDa|damage-specific DNA-binding protein 2|xeroderma pigmentosum group E protein', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47214942, end_position_on_the_genomic_accession=47239218, orientation='plus', exon_count=10, OMIM=600811, no_hdr0=''),
  1644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1644, CurrentID=0, Status='live', Symbol='DDC', Aliases='AADC', description='dopa decarboxylase', other_designations='aromatic L-amino acid decarboxylase|dopa decarboxylase (aromatic L-amino acid decarboxylase)', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50458436, end_position_on_the_genomic_accession=50565457, orientation='minus', exon_count=18, OMIM=107930, no_hdr0=''),
  1645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1645, CurrentID=0, Status='live', Symbol='AKR1C1', Aliases='2-ALPHA-HSD, 20-ALPHA-HSD, C9, DD1, DD1/DD2, DDH, DDH1, H-37, HAKRC, HBAB, MBAB', description='aldo-keto reductase family 1, member C1', other_designations='20 alpha-hydroxysteroid dehydrogenase|20-alpha-hydroxysteroid dehydrogenase|aldo-keto reductase C|chlordecone reductase homolog HAKRC|dihydrodiol dehydrogenase 1|dihydrodiol dehydrogenase 1/2|dihydrodiol dehydrogenase 1; 20-alpha (3-alpha)-hydroxysteroid dehydrogenase|dihydrodiol dehydrogenase isoform DD1|hepatic dihydrodiol dehydrogenase|high-affinity hepatic bile acid-binding protein|indanol dehydrogenase|trans-1,2-dihydrobenzene-1,2-diol dehydrogenase|type II 3-alpha-hydroxysteroid dehydrogenase', map_location='10p15-p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4963262, end_position_on_the_genomic_accession=4977966, orientation='plus', exon_count=9, OMIM=600449, no_hdr0=''),
  1646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1646, CurrentID=0, Status='live', Symbol='AKR1C2', Aliases='AKR1C-pseudo, BABP, DD, DD-2, DD/BABP, DD2, DDH2, HAKRD, HBAB, MCDR2, SRXY8, TDD', description='aldo-keto reductase family 1, member C2', other_designations='3-alpha-HSD3|chlordecone reductase homolog HAKRD|dihydrodiol dehydrogenase 2; bile acid binding protein; 3-alpha hydroxysteroid dehydrogenase, type III|pseudo-chlordecone reductase|testicular 17,20-desmolase deficiency|trans-1,2-dihydrobenzene-1,2-diol dehydrogenase|type II dihydrodiol dehydrogenase', map_location='10p15-p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4987775, end_position_on_the_genomic_accession=5018033, orientation='minus', exon_count=12, OMIM=600450, no_hdr0=''),
  1647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1647, CurrentID=0, Status='live', Symbol='GADD45A', Aliases='DDIT1, GADD45', description='growth arrest and DNA damage inducible alpha', other_designations='DDIT-1|DNA damage-inducible transcript 1 protein|DNA damage-inducible transcript-1|DNA-damage-inducible transcript 1|growth arrest and DNA-damage-inducible 45 alpha|growth arrest and DNA-damage-inducible, alpha', map_location='1p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67685177, end_position_on_the_genomic_accession=67688338, orientation='plus', exon_count=4, OMIM=126335, no_hdr0=''),
  132720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132720, CurrentID=0, Status='live', Symbol='C4orf32', Aliases='', description='chromosome 4 open reading frame 32', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112145397, end_position_on_the_genomic_accession=112189081, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  1649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1649, CurrentID=0, Status='live', Symbol='DDIT3', Aliases='CEBPZ, CHOP, CHOP-10, CHOP10, GADD153', description='DNA damage inducible transcript 3', other_designations='C/EBP zeta|CCAAT/enhancer-binding protein homologous protein|DDIT-3|DNA-damage-inducible transcript 3|c/EBP-homologous protein 10|growth arrest and DNA damage-inducible protein GADD153', map_location='12q13.1-q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57516588, end_position_on_the_genomic_accession=57521698, orientation='minus', exon_count=5, OMIM=126337, no_hdr0=''),
  1650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1650, CurrentID=0, Status='live', Symbol='DDOST', Aliases='AGER1, CDG1R, OKSWcl45, OST, OST48, WBP1', description='dolichyl-diphosphooligosaccharide--protein glycosyltransferase subunit (non-catalytic)', other_designations='advanced glycation endproduct receptor 1|dolichyl-diphosphooligosaccharide-protein glycotransferase|oligosaccharyl transferase 48 kDa subunit|oligosaccharyltransferase 48 kDa subunit|oligosaccharyltransferase subunit 48', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20651767, end_position_on_the_genomic_accession=20661544, orientation='minus', exon_count=11, OMIM=602202, no_hdr0=''),
  1652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1652, CurrentID=0, Status='live', Symbol='DDT', Aliases='DDCT', description='D-dopachrome tautomerase', other_designations='phenylpyruvate tautomerase II', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23971365, end_position_on_the_genomic_accession=23979828, orientation='minus', exon_count=4, OMIM=602750, no_hdr0=''),
  1653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1653, CurrentID=0, Status='live', Symbol='DDX1', Aliases='DBP-RB, UKVH5d', description='DEAD (Asp-Glu-Ala-Asp) box helicase 1', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 1|DEAD box polypeptide 1|DEAD box protein 1|DEAD box protein retinoblastoma|DEAD box-1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 1', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15591621, end_position_on_the_genomic_accession=15631111, orientation='plus', exon_count=26, OMIM=601257, no_hdr0=''),
  1654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1654, CurrentID=0, Status='live', Symbol='DDX3X', Aliases='CAP-Rf, DBX, DDX14, DDX3, HLP2', description='DEAD (Asp-Glu-Ala-Asp) box helicase 3, X-linked', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, X-linked|DEAD box protein 3, X-chromosomal|DEAD box, X isoform|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 3|DEAD/H box-3|helicase-like protein 2', map_location='Xp11.3-p11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41333308, end_position_on_the_genomic_accession=41364472, orientation='plus', exon_count=20, OMIM=300160, no_hdr0=''),
  1655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1655, CurrentID=0, Status='live', Symbol='DDX5', Aliases='G17P1, HLR1, HUMP68, p68', description='DEAD (Asp-Glu-Ala-Asp) box helicase 5', other_designations='ATP-dependent RNA helicase DDX5|DEAD (Asp-Glu-Ala-Asp) box polypeptide 5|DEAD box protein 5|DEAD box-5|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 5 (RNA helicase, 68kD)|RNA helicase p68', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64498256, end_position_on_the_genomic_accession=64506924, orientation='minus', exon_count=16, OMIM=180630, no_hdr0=''),
  1656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1656, CurrentID=0, Status='live', Symbol='DDX6', Aliases='HLR2, P54, RCK', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6', other_designations='ATP-dependent RNA helicase p54|DEAD (Asp-Glu-Ala-Asp) box polypeptide 6|DEAD box protein 6|DEAD box-6|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 6 (RNA helicase, 54kD)|oncogene RCK', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118747763, end_position_on_the_genomic_accession=118791263, orientation='minus', exon_count=15, OMIM=600326, no_hdr0=''),
  1657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1657, CurrentID=0, Status='live', Symbol='DMXL1', Aliases='', description='Dmx-like 1', other_designations='', map_location='5q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119071014, end_position_on_the_genomic_accession=119249129, orientation='plus', exon_count=46, OMIM=605671, no_hdr0=''),
  1659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1659, CurrentID=0, Status='live', Symbol='DHX8', Aliases='DDX8, HRH1, PRP22, PRPF22', description='DEAH (Asp-Glu-Ala-His) box polypeptide 8', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 8 (RNA helicase)|DEAH box protein 8|RNA helicase HRH1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43483917, end_position_on_the_genomic_accession=43526796, orientation='plus', exon_count=24, OMIM=600396, no_hdr0=''),
  1660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1660, CurrentID=0, Status='live', Symbol='DHX9', Aliases='DDX9, LKP, NDH2, NDHII, RHA', description='DEAH (Asp-Glu-Ala-His) box helicase 9', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 9|DEAH (Asp-Glu-Ala-His) box polypeptide 9|DEAH box protein 9|RNA helicase A|leukophysin|nuclear DNA helicase II', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182839304, end_position_on_the_genomic_accession=182887982, orientation='plus', exon_count=29, OMIM=603115, no_hdr0=''),
  1661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1661, CurrentID=0, Status='live', Symbol='DHX9P1', Aliases='DDX9P, DHX9P', description='DEAH (Asp-Glu-Ala-His) box polypeptide 9 pseudogene 1', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 9 (RNA helicase A) pseudogene', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76984742, end_position_on_the_genomic_accession=76988969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1662, CurrentID=0, Status='live', Symbol='DDX10', Aliases='HRH-J8', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 10', other_designations='DDX10-NUP98 fusion protein type 2|DEAD box protein 10|DEAD box-10|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 10 (RNA helicase)', map_location='11q22-q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108665025, end_position_on_the_genomic_accession=108940930, orientation='plus', exon_count=19, OMIM=601235, no_hdr0=''),
  1663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1663, CurrentID=0, Status='live', Symbol='DDX11', Aliases='CHL1, CHLR1, KRG2, WABS', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11', other_designations='CHL1-like helicase homolog|CHL1-related helicase gene-1|CHL1-related protein 1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 (CHL1-like helicase homolog, S. cerevisiae)|DEAD/H box protein 11|KRG-2|hCHLR1|keratinocyte growth factor-regulated gene 2 protein', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31073581, end_position_on_the_genomic_accession=31104799, orientation='plus', exon_count=29, OMIM=601150, no_hdr0=''),
  100533888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533888, CurrentID=0, Status='live', Symbol='LOC100533888', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1665, CurrentID=0, Status='live', Symbol='DHX15', Aliases='DBP1, DDX15, HRH2, PRP43, PRPF43, PrPp43p', description='DEAH (Asp-Glu-Ala-His) box helicase 15', other_designations='ATP-dependent RNA helicase #46|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 15|DEAD/H box-15|DEAH (Asp-Glu-Ala-His) box polypeptide 15|DEAH box protein 15|RNA helicase 2|putative pre-mRNA-splicing factor ATP-dependent RNA helicase DHX15', map_location='4p15.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24527464, end_position_on_the_genomic_accession=24584603, orientation='minus', exon_count=14, OMIM=603403, no_hdr0=''),
  1666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1666, CurrentID=0, Status='live', Symbol='DECR1', Aliases='DECR, NADPH, SDR18C1', description='2,4-dienoyl-CoA reductase 1, mitochondrial', other_designations='2,4-dienoyl CoA reductase 1, mitochondrial|4-enoyl-CoA reductase|short chain dehydrogenase/reductase family 18C member 1|short chain dehydrogenase/reductase family 18C, member 1', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90001352, end_position_on_the_genomic_accession=90052004, orientation='plus', exon_count=11, OMIM=222745, no_hdr0=''),
  1667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1667, CurrentID=0, Status='live', Symbol='DEFA1', Aliases='DEF1, DEFA2, HNP-1, HP-1, HP1, MRS', description='defensin alpha 1', other_designations='defensin, alpha 1, myeloid-related sequence|defensin, alpha 2|myeloid-related sequence', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6977649, end_position_on_the_genomic_accession=6980122, orientation='minus', exon_count=4, OMIM=125220, no_hdr0=''),
  1668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1668, CurrentID=0, Status='live', Symbol='DEFA3', Aliases='DEF3, HNP-3, HNP3, HP-3, HP3', description='defensin alpha 3', other_designations='defensin 3, neutrophil-specific|defensin, alpha 3, neutrophil-specific|neutrophil peptide 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7015869, end_position_on_the_genomic_accession=7018339, orientation='minus', exon_count=4, OMIM=604522, no_hdr0=''),
  1669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1669, CurrentID=0, Status='live', Symbol='DEFA4', Aliases='DEF4, HNP-4, HP-4, HP4', description='defensin alpha 4', other_designations='corticostatin|defensin, alpha 4, corticostatin|defensin, alpha 4, preproprotein', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6935820, end_position_on_the_genomic_accession=6938338, orientation='minus', exon_count=3, OMIM=601157, no_hdr0=''),
  1670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1670, CurrentID=0, Status='live', Symbol='DEFA5', Aliases='DEF5, HD-5', description='defensin alpha 5', other_designations='HD5(20-94)|defensin 5|defensin, alpha 5, Paneth cell-specific|defensin, alpha 5, preproprotein', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7055300, end_position_on_the_genomic_accession=7056739, orientation='minus', exon_count=2, OMIM=600472, no_hdr0=''),
  1671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1671, CurrentID=0, Status='live', Symbol='DEFA6', Aliases='DEF6, HD-6', description='defensin alpha 6', other_designations='defensin 6|defensin, alpha 6, Paneth cell-specific', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6924694, end_position_on_the_genomic_accession=6926076, orientation='minus', exon_count=2, OMIM=600471, no_hdr0=''),
  1672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1672, CurrentID=0, Status='live', Symbol='DEFB1', Aliases='BD1, DEFB-101, HBD1, DEFB1', description='defensin beta 1', other_designations='BD-1|beta defensin 1|beta-defensin-1|defensin, beta 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6870575, end_position_on_the_genomic_accession=6878007, orientation='minus', exon_count=2, OMIM=602056, no_hdr0=''),
  1673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1673, CurrentID=0, Status='live', Symbol='DEFB4A', Aliases='BD-2, DEFB-2, DEFB102, DEFB2, DEFB4, HBD-2, SAP1', description='defensin beta 4A', other_designations='beta defensin 2|beta defensin-2|defensin, beta 2|defensin, beta 4|defensin, beta 4A|skin-antimicrobial peptide 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7894565, end_position_on_the_genomic_accession=7896716, orientation='plus', exon_count=2, OMIM=602215, no_hdr0=''),
  1674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1674, CurrentID=0, Status='live', Symbol='DES', Aliases='CSM1, CSM2, LGMD2R', description='desmin', other_designations='intermediate filament protein|mutant desmin p.K241E', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219418377, end_position_on_the_genomic_accession=219426739, orientation='plus', exon_count=9, OMIM=125660, no_hdr0=''),
  1675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1675, CurrentID=0, Status='live', Symbol='CFD', Aliases='ADIPSIN, ADN, DF, PFD', description='complement factor D (adipsin)', other_designations='C3 convertase activator|D component of complement (adipsin)|complement factor D preproprotein|properdin factor D', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=859665, end_position_on_the_genomic_accession=863624, orientation='plus', exon_count=5, OMIM=134350, no_hdr0=''),
  1676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1676, CurrentID=0, Status='live', Symbol='DFFA', Aliases='DFF-45, DFF1, ICAD', description='DNA fragmentation factor, 45kDa, alpha polypeptide', other_designations='DFF45|DNA fragmentation factor 45 kDa subunit|inhibitor of CAD', map_location='1p36.3-p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10460531, end_position_on_the_genomic_accession=10472556, orientation='minus', exon_count=6, OMIM=601882, no_hdr0=''),
  1677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1677, CurrentID=0, Status='live', Symbol='DFFB', Aliases='CAD, CPAN, DFF-40, DFF2, DFF40', description='DNA fragmentation factor, 40kDa, beta polypeptide (caspase-activated DNase)', other_designations='DNA fragmentation factor 40 kDa subunit|caspase-activated DNase|caspase-activated deoxyribonuclease|caspase-activated nuclease', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3857143, end_position_on_the_genomic_accession=3885429, orientation='plus', exon_count=10, OMIM=601883, no_hdr0=''),
  1678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1678, CurrentID=0, Status='live', Symbol='TIMM8A', Aliases='DDP, DDP1, DFN1, MTS, TIM8', description='translocase of inner mitochondrial membrane 8 homolog A (yeast)', other_designations='X-linked deafness dystonia protein|deafness dystonia protein 1|deafness/dystonia peptide', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101345656, end_position_on_the_genomic_accession=101348969, orientation='minus', exon_count=3, OMIM=300356, no_hdr0=''),
  100533903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533903, CurrentID=0, Status='live', Symbol='LOC100533903', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1680, CurrentID=0, Status='live', Symbol='DFNX3', Aliases='DFN4', description='deafness, X-linked 3', other_designations='deafness 4, congenital sensorineural', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300030, no_hdr0=''),
  1681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1681, CurrentID=0, Status='live', Symbol='DFN5', Aliases='', description='deafness, X-linked 5', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1684, CurrentID=0, Status='live', Symbol='DFN8', Aliases='', description='deafness, X-linked 8', other_designations='deafness, X-linked 8, progressive', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1687, CurrentID=0, Status='live', Symbol='DFNA5', Aliases='ICERE-1', description='deafness, autosomal dominant 5', other_designations='inversely correlated with estrogen receptor expression 1|nonsyndromic hearing impairment protein', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24698355, end_position_on_the_genomic_accession=24758020, orientation='minus', exon_count=11, OMIM=608798, no_hdr0=''),
  100533912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533912, CurrentID=0, Status='live', Symbol='LOC100533912', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1689, CurrentID=0, Status='live', Symbol='DFNA7', Aliases='', description='deafness, autosomal dominant 7', other_designations='', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601412, no_hdr0=''),
  1690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1690, CurrentID=0, Status='live', Symbol='COCH', Aliases='COCH-5B25B2, DFNA9, COCH', description='cochlin', other_designations='coagulation factor C homolog, cochlin (Limulus polyphemus)', map_location='14q11.2-q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30874535, end_position_on_the_genomic_accession=30895079, orientation='plus', exon_count=12, OMIM=603196, no_hdr0=''),
  100533915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533915, CurrentID=0, Status='live', Symbol='LOC100533915', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533916, CurrentID=0, Status='live', Symbol='LOC100533916', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533917, CurrentID=0, Status='live', Symbol='LOC100533917', Aliases='', description='N-ethylmaleimide-sensitive factor pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1694, CurrentID=0, Status='live', Symbol='DFNA16', Aliases='', description='deafness, autosomal dominant 16', other_designations='', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603964, no_hdr0=''),
  100533919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533919, CurrentID=0, Status='live', Symbol='LOC100533919', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1696, CurrentID=0, Status='live', Symbol='DFNA19', Aliases='', description='deafness, autosomal dominant 19', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1697, CurrentID=0, Status='live', Symbol='DFNB5', Aliases='', description='deafness, autosomal recessive 5', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600792, no_hdr0=''),
  100533922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533922, CurrentID=0, Status='live', Symbol='LOC100533922', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533923, CurrentID=0, Status='live', Symbol='LOC100533923', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533924, CurrentID=0, Status='live', Symbol='LOC100533924', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533925, CurrentID=0, Status='live', Symbol='LOC100533925', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533926, CurrentID=0, Status='live', Symbol='LOC100533926', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284273, CurrentID=0, Status='live', Symbol='ZADH2', Aliases='', description='zinc binding alcohol dehydrogenase domain containing 2', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75197323, end_position_on_the_genomic_accession=75209326, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100533928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533928, CurrentID=0, Status='live', Symbol='LOC100533928', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1705, CurrentID=0, Status='live', Symbol='DFNB13', Aliases='', description='deafness, autosomal recessive 13', other_designations='', map_location='7q34-q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603098, no_hdr0=''),
  1706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1706, CurrentID=0, Status='live', Symbol='DFNB14', Aliases='', description='deafness, autosomal recessive 14', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603678, no_hdr0=''),
  100533931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533931, CurrentID=0, Status='live', Symbol='LOC100533931', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533932, CurrentID=0, Status='live', Symbol='LOC100533932', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1709, CurrentID=0, Status='live', Symbol='DFNB17', Aliases='', description='deafness, autosomal recessive 17', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603010, no_hdr0=''),
  100533934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533934, CurrentID=0, Status='live', Symbol='LOC100533934', Aliases='', description='leucine rich repeat containing 37A pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533935, CurrentID=0, Status='live', Symbol='LOC100533935', Aliases='', description='leucine rich repeat containing 37, member A2 pseudogene', other_designations='leucine rich repeat containing 37A pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1712, CurrentID=0, Status='live', Symbol='DFNB19', Aliases='', description='deafness, autosomal recessive 19', other_designations='', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1713, CurrentID=0, Status='live', Symbol='DFNB20', Aliases='', description='deafness, autosomal recessive 20', other_designations='', map_location='11q25-qter', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604060, no_hdr0=''),
  1714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1714, CurrentID=0, Status='live', Symbol='DGCR', Aliases='CATCH22, DGS, VCF', description='DiGeorge syndrome chromosome region', other_designations='DiGeorge syndrome chromosome region (velocardiofacial syndrome)', map_location='22q11.21-q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=188400, no_hdr0=''),
  100533939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533939, CurrentID=0, Status='live', Symbol='LOC100533939', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1716, CurrentID=0, Status='live', Symbol='DGUOK', Aliases='MTDPS3, dGK', description='deoxyguanosine kinase', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73926826, end_position_on_the_genomic_accession=73958961, orientation='plus', exon_count=8, OMIM=601465, no_hdr0=''),
  1717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1717, CurrentID=0, Status='live', Symbol='DHCR7', Aliases='SLOS', description='7-dehydrocholesterol reductase', other_designations='7-DHC reductase|delta-7-dehydrocholesterol reductase|putative sterol reductase SR-2|sterol delta-7-reductase', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71434411, end_position_on_the_genomic_accession=71448431, orientation='minus', exon_count=9, OMIM=602858, no_hdr0=''),
  1718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1718, CurrentID=0, Status='live', Symbol='DHCR24', Aliases='DCE, Nbla03646, SELADIN1, seladin-1', description='24-dehydrocholesterol reductase', other_designations='3 beta-hydroxysterol delta 24-reductase|3-beta-hydroxysterol delta-24-reductase|desmosterol-to-cholesterol enzyme|diminuto/dwarf1 homolog|seladin 1|selective AD indicator 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54849627, end_position_on_the_genomic_accession=54887248, orientation='minus', exon_count=9, OMIM=606418, no_hdr0=''),
  1719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1719, CurrentID=0, Status='live', Symbol='DHFR', Aliases='DHFRP1, DYR', description='dihydrofolate reductase', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80626226, end_position_on_the_genomic_accession=80654981, orientation='minus', exon_count=6, OMIM=126060, no_hdr0=''),
  1720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1720, CurrentID=0, Status='live', Symbol='LOC1720', Aliases='', description='dihydrofolate reductase pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82856803, end_position_on_the_genomic_accession=82857769, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100533945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533945, CurrentID=0, Status='live', Symbol='LOC100533945', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533946, CurrentID=0, Status='live', Symbol='LOC100533946', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1723, CurrentID=0, Status='live', Symbol='DHODH', Aliases='DHOdehase, POADS, URA1', description='dihydroorotate dehydrogenase (quinone)', other_designations='dihydroorotate oxidase|human complement of yeast URA1', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72008744, end_position_on_the_genomic_accession=72025417, orientation='plus', exon_count=11, OMIM=126064, no_hdr0=''),
  100533948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533948, CurrentID=0, Status='live', Symbol='LOC100533948', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1725, CurrentID=0, Status='live', Symbol='DHPS', Aliases='DHS, DS, MIG13', description='deoxyhypusine synthase', other_designations='migration-inducing gene 13', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12672858, end_position_on_the_genomic_accession=12681887, orientation='minus', exon_count=13, OMIM=600944, no_hdr0=''),
  100533950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533950, CurrentID=0, Status='live', Symbol='LOC100533950', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1727, CurrentID=0, Status='live', Symbol='CYB5R3', Aliases='B5R, DIA1', description='cytochrome b5 reductase 3', other_designations='NADH-cytochrome b5 reductase 3 membrane-bound form|NADH-cytochrome b5 reductase 3 soluble form|diaphorase-1|mutant NADH-cytochrome b5 reductase', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42617840, end_position_on_the_genomic_accession=42649399, orientation='minus', exon_count=12, OMIM=613213, no_hdr0=''),
  1728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1728, CurrentID=0, Status='live', Symbol='NQO1', Aliases='DHQU, DIA4, DTD, NMOR1, NMORI, QR1', description='NAD(P)H dehydrogenase, quinone 1', other_designations='DT-diaphorase|NAD(P)H:Quinone acceptor oxidoreductase type 1|NAD(P)H:menadione oxidoreductase 1|NAD(P)H:quinone oxidoreductase 1|NAD(P)H:quinone oxireductase|azoreductase|diaphorase (NADH/NADPH) (cytochrome b-5 reductase)|diaphorase-4|dioxin-inducible 1|menadione reductase|phylloquinone reductase|quinone reductase 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69709401, end_position_on_the_genomic_accession=69726668, orientation='minus', exon_count=6, OMIM=125860, no_hdr0=''),
  1729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1729, CurrentID=0, Status='live', Symbol='DIAPH1', Aliases='DFNA1, DIA1, DRF1, LFHL1, SCBMS, hDIA1', description='diaphanous related formin 1', other_designations='diaphanous-related formin 1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141515016, end_position_on_the_genomic_accession=141619055, orientation='minus', exon_count=31, OMIM=602121, no_hdr0=''),
  1730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1730, CurrentID=0, Status='live', Symbol='DIAPH2', Aliases='DIA, DIA2, DRF2, POF, POF2', description='diaphanous related formin 2', other_designations='diaphanous homolog 2|diaphanous-related formin 2|diaphanous-related formin-2|diaphorase-2', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96684663, end_position_on_the_genomic_accession=97600598, orientation='plus', exon_count=28, OMIM=300108, no_hdr0=''),
  1731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1731, CurrentID=0, Status='live', Symbol='SEPT1', Aliases='DIFF6, LARP, PNUTL3, SEP1', description='septin 1', other_designations='differentiation 6 (deoxyguanosine triphosphate triphosphohydrolase)|peanut-like protein 3|serologically defined breast cancer antigen NY-BR-24', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30378133, end_position_on_the_genomic_accession=30382850, orientation='minus', exon_count=12, OMIM=612897, no_hdr0=''),
  1732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1732, CurrentID=0, Status='live', Symbol='DIH1', Aliases='HCD', description='diaphragmatic hernia 1', other_designations='', map_location='15q26.1-q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=142340, no_hdr0=''),
  1733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1733, CurrentID=0, Status='live', Symbol='DIO1', Aliases='5DI, TXDI1', description='deiodinase, iodothyronine, type I', other_designations="DIOI|iodothyronine deiodinase type 1|thyroxine deiodinase type I (selenoprotein)|type 1 DI|type-I 5'-deiodinase", map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53894188, end_position_on_the_genomic_accession=53911086, orientation='plus', exon_count=4, OMIM=147892, no_hdr0=''),
  1734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1734, CurrentID=0, Status='live', Symbol='DIO2', Aliases='5DII, D2, DIOII, SelY, TXDI2', description='deiodinase, iodothyronine, type II', other_designations="deiodinase-2|deiodonase-2|thyroxine deiodinase, type II|type 2 DI|type 2 iodothyronine deiodinase|type-II 5'-deiodinase|type-II 5'deiodinase", map_location='14q24.2-q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80197525, end_position_on_the_genomic_accession=80231054, orientation='minus', exon_count=6, OMIM=601413, no_hdr0=''),
  1735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1735, CurrentID=0, Status='live', Symbol='DIO3', Aliases='5DIII, D3, DIOIII, TXDI3', description='deiodinase, iodothyronine, type III', other_designations="iodothyronine deiodinase, placental type|thyroxine deiodinase type III (selenoprotein)|type 3 DI|type 3 iodothyronine selenodeiodinase|type-III 5' deiodinase|type-III 5'-deiodinase", map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101561351, end_position_on_the_genomic_accession=101563452, orientation='plus', exon_count=1, OMIM=601038, no_hdr0=''),
  1736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1736, CurrentID=0, Status='live', Symbol='DKC1', Aliases='CBF5, DKC, DKCX, NAP57, NOLA4, XAP101', description='dyskeratosis congenita 1, dyskerin', other_designations='CBF5 homolog|nopp140-associated protein of 57 kDa|nucleolar protein NAP57|nucleolar protein family A member 4|snoRNP protein DKC1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154762742, end_position_on_the_genomic_accession=154777689, orientation='plus', exon_count=14, OMIM=300126, no_hdr0=''),
  1737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1737, CurrentID=0, Status='live', Symbol='DLAT', Aliases='DLTA, PDC-E2, PDCE2', description='dihydrolipoamide S-acetyltransferase', other_designations='70 kDa mitochondrial autoantigen of primary biliary cirrhosis|E2 component of pyruvate dehydrogenase complex|M2 antigen complex 70 kDa subunit|PBC|dihydrolipoamide acetyltransferase component of pyruvate dehydrogenase complex|dihydrolipoyllysine-residue acetyltransferase component of pyruvate dehydrogenase complex mitochondrial|pyruvate dehydrogenase complex component E2', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112024814, end_position_on_the_genomic_accession=112064278, orientation='plus', exon_count=14, OMIM=608770, no_hdr0=''),
  1738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1738, CurrentID=0, Status='live', Symbol='DLD', Aliases='DLDDH, E3, GCSL, LAD, PHE3, DLD', description='dihydrolipoamide dehydrogenase', other_designations='E3 component of pyruvate dehydrogenase complex, 2-oxo-glutarate complex, branched chain keto acid dehydrogenase complex|diaphorase|glycine cleavage system L protein|glycine cleavage system protein L|lipoamide dehydrogenase|lipoamide reductase|lipoyl dehydrogenase', map_location='7q31-q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107891107, end_position_on_the_genomic_accession=107921198, orientation='plus', exon_count=14, OMIM=238331, no_hdr0=''),
  1739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1739, CurrentID=0, Status='live', Symbol='DLG1', Aliases='DLGH1, SAP-97, SAP97, dJ1061C18.1.1, hdlg', description='discs, large homolog 1 (Drosophila)', other_designations='presynaptic protein SAP97|synapse-associated protein 97', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197042560, end_position_on_the_genomic_accession=197299272, orientation='minus', exon_count=34, OMIM=601014, no_hdr0=''),
  1740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1740, CurrentID=0, Status='live', Symbol='DLG2', Aliases='PPP1R58, PSD-93, PSD93, chapsyn-110', description='discs, large homolog 2 (Drosophila)', other_designations='channel-associated protein of synapse-110|channel-associated protein of synapses, 110kDa|discs large 2|discs, large homolog 2, chapsyn-110|postsynaptic density protein PSD-93|protein phosphatase 1, regulatory subunit 58', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83455012, end_position_on_the_genomic_accession=85628534, orientation='minus', exon_count=39, OMIM=603583, no_hdr0=''),
  1741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1741, CurrentID=0, Status='live', Symbol='DLG3', Aliases='MRX, MRX90, NEDLG, PPP1R82, SAP102, XLMR', description='discs, large homolog 3 (Drosophila)', other_designations='neuroendocrine-DLG|protein phosphatase 1, regulatory subunit 82|synapse-associated protein 102', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70444850, end_position_on_the_genomic_accession=70505490, orientation='plus', exon_count=23, OMIM=300189, no_hdr0=''),
  1742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1742, CurrentID=0, Status='live', Symbol='DLG4', Aliases='PSD95, SAP-90, SAP90', description='discs, large homolog 4 (Drosophila)', other_designations='Tax interaction protein 15|discs large homolog 4|post-synaptic density protein 95|postsynaptic density protein 95|synapse-associated protein 90', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7189890, end_position_on_the_genomic_accession=7220050, orientation='minus', exon_count=26, OMIM=602887, no_hdr0=''),
  1743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1743, CurrentID=0, Status='live', Symbol='DLST', Aliases='DLTS', description='dihydrolipoamide S-succinyltransferase (E2 component of 2-oxo-glutarate complex)', other_designations='E2K|OGDC-E2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74881891, end_position_on_the_genomic_accession=74903747, orientation='plus', exon_count=15, OMIM=126063, no_hdr0=''),
  1744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1744, CurrentID=0, Status='live', Symbol='DLSTP1', Aliases='DLSTP', description='dihydrolipoamide S-succinyltransferase pseudogene 1', other_designations='dihydrolipoamide S-succinyltransferase (E2 component of 2-oxo-glutarate complex) pseudogene|dihydrolipoamide S-succinyltransferase pseudogene (E2 component of 2-oxo-glutarate complex)', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75742002, end_position_on_the_genomic_accession=75745013, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  1745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1745, CurrentID=0, Status='live', Symbol='DLX1', Aliases='', description='distal-less homeobox 1', other_designations='distal-less homeo box 1', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172085480, end_position_on_the_genomic_accession=172089674, orientation='plus', exon_count=3, OMIM=600029, no_hdr0=''),
  1746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1746, CurrentID=0, Status='live', Symbol='DLX2', Aliases='TES-1, TES1', description='distal-less homeobox 2', other_designations='distal-less homeo box 2', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172099438, end_position_on_the_genomic_accession=172102750, orientation='minus', exon_count=3, OMIM=126255, no_hdr0=''),
  1747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1747, CurrentID=0, Status='live', Symbol='DLX3', Aliases='AI4, TDO', description='distal-less homeobox 3', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49977079, end_position_on_the_genomic_accession=49995224, orientation='minus', exon_count=4, OMIM=600525, no_hdr0=''),
  1748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1748, CurrentID=0, Status='live', Symbol='DLX4', Aliases='BP1, DLX7, DLX8, DLX9', description='distal-less homeobox 4', other_designations='beta protein 1|distal-less homeo box 7|distal-less homeo box 9|homeobox protein DLX-7|homeobox protein DLX-8', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49969198, end_position_on_the_genomic_accession=49974959, orientation='plus', exon_count=3, OMIM=601911, no_hdr0=''),
  1749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1749, CurrentID=0, Status='live', Symbol='DLX5', Aliases='SHFM1D', description='distal-less homeobox 5', other_designations='distal-less homeo box 5|split hand/foot malformation type 1 with sensorineural hearing loss', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97020390, end_position_on_the_genomic_accession=97024831, orientation='minus', exon_count=4, OMIM=600028, no_hdr0=''),
  1750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1750, CurrentID=0, Status='live', Symbol='DLX6', Aliases='', description='distal-less homeobox 6', other_designations='distal-less homeo box 6', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97005978, end_position_on_the_genomic_accession=97011040, orientation='plus', exon_count=3, OMIM=600030, no_hdr0=''),
  100533975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533975, CurrentID=0, Status='live', Symbol='SLMO2-ATP5E', Aliases='', description='SLMO2-ATP5E readthrough', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59028678, end_position_on_the_genomic_accession=59042846, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271832, CurrentID=0, Status='live', Symbol='LOC100271832', Aliases='', description='uncharacterized LOC100271832', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32927127, end_position_on_the_genomic_accession=32946136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  1755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1755, CurrentID=0, Status='live', Symbol='DMBT1', Aliases='GP340, muclin', description='deleted in malignant brain tumors 1', other_designations='SAG|glycoprotein 340|gp-340|hensin|salivary agglutinin|surfactant pulmonary-associated D-binding protein', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122560665, end_position_on_the_genomic_accession=122643736, orientation='plus', exon_count=56, OMIM=601969, no_hdr0=''),
  1756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1756, CurrentID=0, Status='live', Symbol='DMD', Aliases='BMD, CMD3B, DXS142, DXS164, DXS206, DXS230, DXS239, DXS268, DXS269, DXS270, DXS272, MRX85', description='dystrophin', other_designations='', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=31119219, end_position_on_the_genomic_accession=33339609, orientation='minus', exon_count=87, OMIM=300377, no_hdr0=''),
  1757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1757, CurrentID=0, Status='live', Symbol='SARDH', Aliases='BPR-2, DMGDHL1, SAR, SARD, SDH', description='sarcosine dehydrogenase', other_designations='dimethylglycine dehydrogenase-like 1', map_location='9q33-q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133661740, end_position_on_the_genomic_accession=133739955, orientation='minus', exon_count=24, OMIM=604455, no_hdr0=''),
  1758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1758, CurrentID=0, Status='live', Symbol='DMP1', Aliases='ARHP, ARHR, DMP-1', description='dentin matrix acidic phosphoprotein 1', other_designations='dentin matrix protein 1', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87650302, end_position_on_the_genomic_accession=87664361, orientation='plus', exon_count=7, OMIM=600980, no_hdr0=''),
  1759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1759, CurrentID=0, Status='live', Symbol='DNM1', Aliases='DNM, EIEE31', description='dynamin 1', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128203355, end_position_on_the_genomic_accession=128255249, orientation='plus', exon_count=26, OMIM=602377, no_hdr0=''),
  1760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1760, CurrentID=0, Status='live', Symbol='DMPK', Aliases='DM, DM1, DM1PK, DMK, MDPK, MT-PK', description='dystrophia myotonica protein kinase', other_designations='DM protein kinase|DM1 protein kinase|dystrophia myotonica-protein kinase|myotonic dystrophy associated protein kinase|myotonin protein kinase A|thymopoietin homolog', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45769709, end_position_on_the_genomic_accession=45782557, orientation='minus', exon_count=16, OMIM=605377, no_hdr0=''),
  1761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1761, CurrentID=0, Status='live', Symbol='DMRT1', Aliases='CT154, DMT1', description='doublesex and mab-3 related transcription factor 1', other_designations='DM domain expressed in testis 1|DM domain expressed in testis protein 1', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=841647, end_position_on_the_genomic_accession=969090, orientation='plus', exon_count=7, OMIM=602424, no_hdr0=''),
  1762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1762, CurrentID=0, Status='live', Symbol='DMWD', Aliases='D19S593E, DMR-N9, DMRN9, gene59', description='dystrophia myotonica, WD repeat containing', other_designations='dystrophia myotonica-containing WD repeat motif protein|protein 59', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45783006, end_position_on_the_genomic_accession=45792802, orientation='minus', exon_count=5, OMIM=609857, no_hdr0=''),
  1763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1763, CurrentID=0, Status='live', Symbol='DNA2', Aliases='DNA2L, hDNA2', description='DNA replication helicase/nuclease 2', other_designations='DNA replication ATP-dependent helicase-like homolog|DNA replication helicase 2 homolog|DNA2-like helicase', map_location='10q21.3-q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68414064, end_position_on_the_genomic_accession=68472516, orientation='minus', exon_count=23, OMIM=601810, no_hdr0=''),
  100271846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271846, CurrentID=0, Status='live', Symbol='ERVV-2', Aliases='ENVV2, HERV-V2', description='endogenous retrovirus group V member 2', other_designations='HERV-V_19q13.41 provirus ancestral Env polyprotein 2|endogenous retrovirus HSV2|endogenous retrovirus group V, member 2|envelope glycoprotein ENVV2|envelope protein ENVV2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53044738, end_position_on_the_genomic_accession=53051127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  1767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1767, CurrentID=0, Status='live', Symbol='DNAH5', Aliases='CILD3, DNAHC5, HL1, KTGNR, PCD', description='dynein, axonemal, heavy chain 5', other_designations='axonemal beta dynein heavy chain 5|ciliary dynein heavy chain 5|dynein, axonemal, heavy polypeptide 5', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=13690328, end_position_on_the_genomic_accession=14011759, orientation='minus', exon_count=80, OMIM=603335, no_hdr0=''),
  1768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1768, CurrentID=0, Status='live', Symbol='DNAH6', Aliases='DNHL1, Dnahc6, HL-2, HL2', description='dynein, axonemal, heavy chain 6', other_designations='axonemal beta dynein heavy chain 6|ciliary dynein heavy chain 6|dynein heavy chain-like 1|dynein, axonemal, heavy polypeptide 6', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84459540, end_position_on_the_genomic_accession=84819589, orientation='plus', exon_count=83, OMIM=603336, no_hdr0=''),
  100271849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271849, CurrentID=0, Status='live', Symbol='MEF2B', Aliases='RSRFR2', description='myocyte enhancer factor 2B', other_designations='MADS box transcription enhancer factor 2, polypeptide B (myocyte enhancer factor 2B)|serum response factor-like protein 2', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19145567, end_position_on_the_genomic_accession=19170289, orientation='minus', exon_count=9, OMIM=600661, no_hdr0=''),
  1770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1770, CurrentID=0, Status='live', Symbol='DNAH9', Aliases='DNAH17L, DNEL1, DYH9, Dnahc9, HL-20, HL20', description='dynein, axonemal, heavy chain 9', other_designations='DNAH9 variant protein|axonemal beta dynein heavy chain 9|ciliary dynein heavy chain 9|dynein, axonemal, heavy polypeptide 9', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11598431, end_position_on_the_genomic_accession=11970168, orientation='plus', exon_count=71, OMIM=603330, no_hdr0=''),
  1773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1773, CurrentID=0, Status='live', Symbol='DNASE1', Aliases='DNL1, DRNI', description='deoxyribonuclease I', other_designations='DNase I, lysosomal|Dornase alfa|human urine deoxyribonuclease I', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3611745, end_position_on_the_genomic_accession=3662875, orientation='plus', exon_count=15, OMIM=125505, no_hdr0=''),
  1774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1774, CurrentID=0, Status='live', Symbol='DNASE1L1', Aliases='DNAS1L1, DNASEX, DNL1L, G4.8, XIB', description='deoxyribonuclease I-like 1', other_designations='DNase I, lysosomal-like|DNase I-like 1|DNase I-like, muscle-specific', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154401236, end_position_on_the_genomic_accession=154412091, orientation='minus', exon_count=10, OMIM=300081, no_hdr0=''),
  1775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1775, CurrentID=0, Status='live', Symbol='DNASE1L2', Aliases='DNAS1L2', description='deoxyribonuclease I-like 2', other_designations='DNase I homolog protein DHP1|DNase I-like 2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2236252, end_position_on_the_genomic_accession=2238711, orientation='plus', exon_count=7, OMIM=602622, no_hdr0=''),
  1776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1776, CurrentID=0, Status='live', Symbol='DNASE1L3', Aliases='DHP2, DNAS1L3, LSD, SLEB16', description='deoxyribonuclease I-like 3', other_designations='DNase I homolog protein 2|DNase I homolog protein DHP2|DNase I-like 3|DNase gamma|LS-DNase|Liver and spleen DNase|deoxyribonuclease I-like III', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58192626, end_position_on_the_genomic_accession=58211003, orientation='minus', exon_count=8, OMIM=602244, no_hdr0=''),
  1777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1777, CurrentID=0, Status='live', Symbol='DNASE2', Aliases='DNASE2A, DNL, DNL2', description='deoxyribonuclease II, lysosomal', other_designations='DNase II alpha|DNase II, lysosomal|R31240_2|acid DNase|deoxyribonuclease II alpha|lysosomal DNase II', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12875211, end_position_on_the_genomic_accession=12881521, orientation='minus', exon_count=6, OMIM=126350, no_hdr0=''),
  1778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1778, CurrentID=0, Status='live', Symbol='DYNC1H1', Aliases='CMT2O, DHC1, DHC1a, DNCH1, DNCL, DNECL, DYHC, Dnchc1, HL-3, SMALED1, p22', description='dynein, cytoplasmic 1, heavy chain 1', other_designations='dynein heavy chain, cytosolic|dynein, cytoplasmic, heavy polypeptide 1', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101964528, end_position_on_the_genomic_accession=102050798, orientation='plus', exon_count=78, OMIM=600112, no_hdr0=''),
  132851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132851, CurrentID=0, Status='live', Symbol='SPATA4', Aliases='SPEF1B, TSARG2', description='spermatogenesis associated 4', other_designations='testis and spermatogenesis cell related protein 2|testis and spermatogenesis cell-related protein 2|testis spermatocyte apoptosis-related gene 2 protein', map_location='4q34.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176184574, end_position_on_the_genomic_accession=176195671, orientation='minus', exon_count=6, OMIM=609879, no_hdr0=''),
  1780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1780, CurrentID=0, Status='live', Symbol='DYNC1I1', Aliases='DNCI1, DNCIC1', description='dynein, cytoplasmic 1, intermediate chain 1', other_designations='DH IC-1|cytoplasmic dynein intermediate chain 1|dynein intermediate chain 1, cytosolic|dynein, cytoplasmic, intermediate polypeptide 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95772506, end_position_on_the_genomic_accession=96110322, orientation='plus', exon_count=21, OMIM=603772, no_hdr0=''),
  1781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1781, CurrentID=0, Status='live', Symbol='DYNC1I2', Aliases='DNCI2, IC2', description='dynein, cytoplasmic 1, intermediate chain 2', other_designations='DH IC-2|dynein intermediate chain 2, cytosolic|dynein, cytoplasmic, intermediate polypeptide 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171687409, end_position_on_the_genomic_accession=171750158, orientation='plus', exon_count=19, OMIM=603331, no_hdr0=''),
  1783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1783, CurrentID=0, Status='live', Symbol='DYNC1LI2', Aliases='DNCLI2, LIC2', description='dynein, cytoplasmic 1, light intermediate chain 2', other_designations='LIC-2|LIC53/55|dynein light intermediate chain 2, cytosolic|dynein, cytoplasmic, light intermediate polypeptide 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66720893, end_position_on_the_genomic_accession=66751622, orientation='minus', exon_count=14, OMIM=611406, no_hdr0=''),
  1784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1784, CurrentID=0, Status='live', Symbol='DNCM', Aliases='cmDNA', description='DNA associated with cytoplasmic membrane', other_designations='Cytoplasmic membrane DNA', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=126330, no_hdr0=''),
  1785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1785, CurrentID=0, Status='live', Symbol='DNM2', Aliases='CMT2M, CMTDI1, CMTDIB, DI-CMTB, DYN2, DYNII, LCCS5', description='dynamin 2', other_designations='dynamin II', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10718053, end_position_on_the_genomic_accession=10831910, orientation='plus', exon_count=22, OMIM=602378, no_hdr0=''),
  1786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1786, CurrentID=0, Status='live', Symbol='DNMT1', Aliases='ADCADN, AIM, CXXC9, DNMT, HSN1E, MCMT', description='DNA (cytosine-5-)-methyltransferase 1', other_designations='CXXC-type zinc finger protein 9|DNA MTase HsaI|DNA methyltransferase HsaI|m.HsaI', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10133345, end_position_on_the_genomic_accession=10195079, orientation='minus', exon_count=42, OMIM=126375, no_hdr0=''),
  1787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1787, CurrentID=0, Status='live', Symbol='TRDMT1', Aliases='DMNT2, DNMT2, MHSAIIP, PUMET, RNMT1', description='tRNA aspartic acid methyltransferase 1', other_designations='DNA (cytosine-5)-methyltransferase-like protein 2|DNA MTase homolog HsaIIP|DNA cytosine-5 methyltransferase 2|DNA methyltransferase-2|tRNA (cytosine-5-)-methyltransferase|tRNA aspartic acid methyltransferase 1 variant 6|tRNA aspartic acid methyltransferase 1 variant 7', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17137802, end_position_on_the_genomic_accession=17202071, orientation='minus', exon_count=16, OMIM=602478, no_hdr0=''),
  1788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1788, CurrentID=0, Status='live', Symbol='DNMT3A', Aliases='DNMT3A2, M.HsaIIIA, TBRS', description='DNA (cytosine-5-)-methyltransferase 3 alpha', other_designations='DNA MTase HsaIIIA|DNA cytosine methyltransferase 3A2', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25232961, end_position_on_the_genomic_accession=25342590, orientation='minus', exon_count=33, OMIM=602769, no_hdr0=''),
  1789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1789, CurrentID=0, Status='live', Symbol='DNMT3B', Aliases='ICF, ICF1, M.HsaIIIB', description='DNA (cytosine-5-)-methyltransferase 3 beta', other_designations='DNA MTase HsaIIIB|DNA methyltransferase HsaIIIB', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32762385, end_position_on_the_genomic_accession=32809356, orientation='plus', exon_count=24, OMIM=602900, no_hdr0=''),
  100271870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271870, CurrentID=0, Status='live', Symbol='RBM22P4', Aliases='', description='RNA binding motif protein 22 pseudogene 4', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45097959, end_position_on_the_genomic_accession=45099349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1791, CurrentID=0, Status='live', Symbol='DNTT', Aliases='TDT', description='DNA nucleotidylexotransferase', other_designations='deoxynucleotidyltransferase, terminal|nucleosidetriphosphate:DNA deoxynucleotidylexotransferase|terminal addition enzyme|terminal deoxynucleotidyltransferase|terminal deoxyribonucleotidyltransferase|terminal transferase', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96304328, end_position_on_the_genomic_accession=96338564, orientation='plus', exon_count=11, OMIM=187410, no_hdr0=''),
  132864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132864, CurrentID=0, Status='live', Symbol='CPEB2', Aliases='CPE-BP2, CPEB-2, hCPEB-2', description='cytoplasmic polyadenylation element binding protein 2', other_designations='CPE-binding protein 2', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15002451, end_position_on_the_genomic_accession=15070153, orientation='plus', exon_count=14, OMIM=610605, no_hdr0=''),
  1793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1793, CurrentID=0, Status='live', Symbol='DOCK1', Aliases='DOCK180, ced5', description='dedicator of cytokinesis 1', other_designations='180 kDa protein downstream of CRK|DOwnstream of CrK', map_location='10q26.13-q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126905409, end_position_on_the_genomic_accession=127452517, orientation='plus', exon_count=57, OMIM=601403, no_hdr0=''),
  1794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1794, CurrentID=0, Status='live', Symbol='DOCK2', Aliases='IMD40', description='dedicator of cytokinesis 2', other_designations='dedicator of cyto-kinesis 2', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169637247, end_position_on_the_genomic_accession=170083382, orientation='plus', exon_count=58, OMIM=603122, no_hdr0=''),
  1795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1795, CurrentID=0, Status='live', Symbol='DOCK3', Aliases='MOCA, PBP', description='dedicator of cytokinesis 3', other_designations='modifier of cell adhesion|presenilin-binding protein', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50674969, end_position_on_the_genomic_accession=51384198, orientation='plus', exon_count=59, OMIM=603123, no_hdr0=''),
  1796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1796, CurrentID=0, Status='live', Symbol='DOK1', Aliases='P62DOK', description='docking protein 1', other_designations='Downstream of tyrosine kinase 1|docking protein 1 (downstream of tyrosine kinase 1)|docking protein 1, 62kDa (downstream of tyrosine kinase 1)|p62(dok)|pp62', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74548205, end_position_on_the_genomic_accession=74557551, orientation='plus', exon_count=6, OMIM=602919, no_hdr0=''),
  1797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1797, CurrentID=0, Status='live', Symbol='DXO', Aliases='DOM3L, DOM3Z, NG6, RAI1', description='decapping exoribonuclease', other_designations='protein Dom3Z', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31969811, end_position_on_the_genomic_accession=31972290, orientation='minus', exon_count=7, OMIM=605996, no_hdr0=''),
  1798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1798, CurrentID=0, Status='live', Symbol='DPAGT1', Aliases='ALG7, CDG-Ij, CDG1J, CMS13, CMSTA2, D11S366, DGPT, DPAGT, DPAGT2, G1PT, GPT, UAGT, UGAT', description='dolichyl-phosphate (UDP-N-acetylglucosamine) N-acetylglucosaminephosphotransferase 1 (GlcNAc-1-P transferase)', other_designations='GlcNAc-1-P transferase|N-acetylglucosamine-1-phosphate transferase|UDP-GlcNAc:dolichyl-phosphate N-acetylglucosaminephosphotransferase|dolichyl-phosphate (UDP-N-acetylglucosamine) N-acetylglucosaminephosphotransferase 1 (GlcNAc-1-P tra|dolichyl-phosphate alpha-N-acetylglucosaminyltransferase', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119096503, end_position_on_the_genomic_accession=119102414, orientation='minus', exon_count=9, OMIM=191350, no_hdr0=''),
  1800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1800, CurrentID=0, Status='live', Symbol='DPEP1', Aliases='MBD1, MDP, RDP', description='dipeptidase 1 (renal)', other_designations='dehydropeptidase-I|hRDP|microsomal dipeptidase|renal dipeptidase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89613308, end_position_on_the_genomic_accession=89640862, orientation='plus', exon_count=16, OMIM=179780, no_hdr0=''),
  1801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1801, CurrentID=0, Status='live', Symbol='DPH1', Aliases='DPH2L, DPH2L1, OVCA1', description='diphthamide biosynthesis 1', other_designations='DPH-like 1|DPH1 homolog|DPH2-like 1|candidate tumor suppressor in ovarian cancer 1|diphthamide biosynthesis protein 2 homolog-like 1|diptheria toxin resistance protein required for diphthamide biosynthesis (Saccharomyces)-like 1|diptheria toxin resistance protein required for diphthamide biosynthesis-like 1|ovarian cancer-associated gene 1 protein|ovarian tumor suppressor candidate 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2030137, end_position_on_the_genomic_accession=2043431, orientation='plus', exon_count=13, OMIM=603527, no_hdr0=''),
  1802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1802, CurrentID=0, Status='live', Symbol='DPH2', Aliases='DPH2L2', description='DPH2 homolog', other_designations='DPH2-like 2|diphthamide biosynthesis protein 2 homolog-like 2|diphthamide biosynthesis-like protein 2|diptheria toxin resistance protein required for diphthamide biosynthesis-like 2', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43969981, end_position_on_the_genomic_accession=43973371, orientation='plus', exon_count=6, OMIM=603456, no_hdr0=''),
  1803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1803, CurrentID=0, Status='live', Symbol='DPP4', Aliases='ADABP, ADCP2, CD26, DPPIV, TP103', description='dipeptidyl-peptidase 4', other_designations='ADCP-2|DPP IV|T-cell activation antigen CD26|adenosine deaminase complexing protein 2|dipeptidyl peptidase IV|dipeptidylpeptidase 4|dipeptidylpeptidase IV (CD26, adenosine deaminase complexing protein 2)', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161992241, end_position_on_the_genomic_accession=162074542, orientation='minus', exon_count=26, OMIM=102720, no_hdr0=''),
  1804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1804, CurrentID=0, Status='live', Symbol='DPP6', Aliases='DPPX, MRD33, VF2', description='dipeptidyl-peptidase 6', other_designations='DPP VI|dipeptidyl aminopeptidase IV-related protein|dipeptidyl peptidase IV-related protein|dipeptidyl peptidase VI', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153887097, end_position_on_the_genomic_accession=154894290, orientation='plus', exon_count=30, OMIM=126141, no_hdr0=''),
  1805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1805, CurrentID=0, Status='live', Symbol='DPT', Aliases='TRAMP', description='dermatopontin', other_designations='tyrosine-rich acidic matrix protein', map_location='1q12-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168695457, end_position_on_the_genomic_accession=168729204, orientation='minus', exon_count=4, OMIM=125597, no_hdr0=''),
  1806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1806, CurrentID=0, Status='live', Symbol='DPYD', Aliases='DHP, DHPDHASE, DPD', description='dihydropyrimidine dehydrogenase', other_designations='dihydrothymine dehydrogenase|dihydrouracil dehydrogenase', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97077743, end_position_on_the_genomic_accession=97921059, orientation='minus', exon_count=25, OMIM=612779, no_hdr0=''),
  1807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1807, CurrentID=0, Status='live', Symbol='DPYS', Aliases='DHP, DHPase', description='dihydropyrimidinase', other_designations='dihydropyrimidine amidohydrolase|hydantoinase', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104379424, end_position_on_the_genomic_accession=104467049, orientation='minus', exon_count=13, OMIM=613326, no_hdr0=''),
  1808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1808, CurrentID=0, Status='live', Symbol='DPYSL2', Aliases='CRMP-2, CRMP2, DHPRP2, DRP-2, DRP2, N2A3, ULIP-2, ULIP2', description='dihydropyrimidinase like 2', other_designations='collapsin response mediator protein hCRMP-2|unc-33-like phosphoprotein 2', map_location='8p22-p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26514193, end_position_on_the_genomic_accession=26658177, orientation='plus', exon_count=16, OMIM=602463, no_hdr0=''),
  1809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1809, CurrentID=0, Status='live', Symbol='DPYSL3', Aliases='CRMP-4, CRMP4, DRP-3, DRP3, LCRMP, ULIP, ULIP-1', description='dihydropyrimidinase like 3', other_designations='collapsin response mediator protein 4 long|dihydropyrimidinase-like 3|unc-33-like phosphoprotein 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147390808, end_position_on_the_genomic_accession=147510260, orientation='minus', exon_count=16, OMIM=601168, no_hdr0=''),
  1810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1810, CurrentID=0, Status='live', Symbol='DR1', Aliases='NC2, NC2-BETA', description='down-regulator of transcription 1', other_designations='TATA-binding protein-associated phosphoprotein|down-regulator of transcription 1, TBP-binding (negative cofactor 2)|negative cofactor 2-beta', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93345921, end_position_on_the_genomic_accession=93362592, orientation='plus', exon_count=3, OMIM=601482, no_hdr0=''),
  1811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1811, CurrentID=0, Status='live', Symbol='SLC26A3', Aliases='CLD, DRA', description='solute carrier family 26 (anion exchanger), member 3', other_designations='down-regulated in adenoma protein|solute carrier family 26, member 3', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107765467, end_position_on_the_genomic_accession=107803233, orientation='minus', exon_count=21, OMIM=126650, no_hdr0=''),
  1812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1812, CurrentID=0, Status='live', Symbol='DRD1', Aliases='DADRA, DRD1', description='dopamine receptor D1', other_designations='dopamine D1 receptor', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175440672, end_position_on_the_genomic_accession=175444160, orientation='minus', exon_count=2, OMIM=126449, no_hdr0=''),
  1813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1813, CurrentID=0, Status='live', Symbol='DRD2', Aliases='D2DR, D2R', description='dopamine receptor D2', other_designations='dopamine D2 receptor|dopamine receptor D2 isoform|seven transmembrane helix receptor', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113409595, end_position_on_the_genomic_accession=113475279, orientation='minus', exon_count=8, OMIM=126450, no_hdr0=''),
  1814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1814, CurrentID=0, Status='live', Symbol='DRD3', Aliases='D3DR, ETM1, FET1', description='dopamine receptor D3', other_designations='essential tremor 1', map_location='3q13.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114127797, end_position_on_the_genomic_accession=114199407, orientation='minus', exon_count=12, OMIM=126451, no_hdr0=''),
  1815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1815, CurrentID=0, Status='live', Symbol='DRD4', Aliases='D4DR', description='dopamine receptor D4', other_designations='D(2C) dopamine receptor|dopamine D4 receptor|seven transmembrane helix receptor', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=637305, end_position_on_the_genomic_accession=640706, orientation='plus', exon_count=4, OMIM=126452, no_hdr0=''),
  1816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1816, CurrentID=0, Status='live', Symbol='DRD5', Aliases='DBDR, DRD1B, DRD1L2', description='dopamine receptor D5', other_designations='D1beta dopamine receptor|d(5) dopamine receptor|dopamine D5 receptor|dopamine receptor D1B', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9781634, end_position_on_the_genomic_accession=9784009, orientation='plus', exon_count=1, OMIM=126453, no_hdr0=''),
  1817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1817, CurrentID=0, Status='live', Symbol='DRD5P1', Aliases='', description='dopamine receptor D5 pseudogene 1', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91683885, end_position_on_the_genomic_accession=91686278, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1818, CurrentID=0, Status='live', Symbol='DRD5P2', Aliases='', description='dopamine receptor D5 pseudogene 2', other_designations='psi DRD5-2|seven transmembrane helix receptor', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144689614, end_position_on_the_genomic_accession=144691842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  1819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1819, CurrentID=0, Status='live', Symbol='DRG2', Aliases='', description='developmentally regulated GTP binding protein 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18087867, end_position_on_the_genomic_accession=18107985, orientation='plus', exon_count=13, OMIM=602986, no_hdr0=''),
  1820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1820, CurrentID=0, Status='live', Symbol='ARID3A', Aliases='BRIGHT, DRIL1, DRIL3, E2FBP1', description='AT-rich interaction domain 3A', other_designations='ARID domain-containing 3A|ARID domain-containing protein 3A|AT rich interactive domain 3A (BRIGHT- like) protein|AT rich interactive domain 3A (BRIGHT-like)|B-cell regulator of IgH transcription|E2F-binding protein 1|dead ringer-like 1|dead ringer-like protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=926037, end_position_on_the_genomic_accession=972816, orientation='plus', exon_count=11, OMIM=603265, no_hdr0=''),
  1821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1821, CurrentID=0, Status='live', Symbol='DRP2', Aliases='DRP-2', description='dystrophin related protein 2', other_designations='', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101219944, end_position_on_the_genomic_accession=101264496, orientation='plus', exon_count=24, OMIM=300052, no_hdr0=''),
  1822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1822, CurrentID=0, Status='live', Symbol='ATN1', Aliases='B37, D12S755E, DRPLA, HRS, NOD', description='atrophin 1', other_designations='dentatorubral-pallidoluysian atrophy protein', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6924463, end_position_on_the_genomic_accession=6942321, orientation='plus', exon_count=11, OMIM=607462, no_hdr0=''),
  1823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1823, CurrentID=0, Status='live', Symbol='DSC1', Aliases='CDHF1, DG2/DG3', description='desmocollin 1', other_designations='cadherin family member 1|desmosomal glycoprotein 2/3', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31129236, end_position_on_the_genomic_accession=31162856, orientation='minus', exon_count=17, OMIM=125643, no_hdr0=''),
  1824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1824, CurrentID=0, Status='live', Symbol='DSC2', Aliases='ARVD11, CDHF2, DG2, DGII/III, DSC3', description='desmocollin 2', other_designations='cadherin family member 2|desmosomal glycoprotein II/III', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31065974, end_position_on_the_genomic_accession=31102432, orientation='minus', exon_count=18, OMIM=125645, no_hdr0=''),
  1825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1825, CurrentID=0, Status='live', Symbol='DSC3', Aliases='CDHF3, DSC, DSC1, DSC2, DSC4, HT-CP', description='desmocollin 3', other_designations='cadherin family member 3|desmocollin-4', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=30989365, end_position_on_the_genomic_accession=31042815, orientation='minus', exon_count=17, OMIM=600271, no_hdr0=''),
  1826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1826, CurrentID=0, Status='live', Symbol='DSCAM', Aliases='CHD2, CHD2-42, CHD2-52', description='Down syndrome cell adhesion molecule', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40010999, end_position_on_the_genomic_accession=40847113, orientation='minus', exon_count=33, OMIM=602523, no_hdr0=''),
  1827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1827, CurrentID=0, Status='live', Symbol='RCAN1', Aliases='ADAPT78, CSP1, DSC1, DSCR1, MCIP1, RCN1', description='regulator of calcineurin 1', other_designations='Down syndrome candidate region 1|Down syndrome critical region gene 1|calcium and oxidant-inducible mRNA|modulatory calcineurin-interacting protein 1|myocyte-enriched calcineurin-interacting protein 1|near DSCR proline-rich protein', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34516442, end_position_on_the_genomic_accession=34615142, orientation='minus', exon_count=9, OMIM=602917, no_hdr0=''),
  1828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1828, CurrentID=0, Status='live', Symbol='DSG1', Aliases='CDHF4, DG1, DSG, EPKHE, EPKHIA, PPKS1, SPPK1', description='desmoglein 1', other_designations='cadherin family member 4|desmosomal glycoprotein 1|pemphigus foliaceus antigen', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31318089, end_position_on_the_genomic_accession=31359246, orientation='plus', exon_count=15, OMIM=125670, no_hdr0=''),
  1829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1829, CurrentID=0, Status='live', Symbol='DSG2', Aliases='ARVC10, ARVD10, CDHF5, CMD1BB, HDGC', description='desmoglein 2', other_designations='HDGC|cadherin family member 5', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31498064, end_position_on_the_genomic_accession=31548851, orientation='plus', exon_count=15, OMIM=125671, no_hdr0=''),
  1830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1830, CurrentID=0, Status='live', Symbol='DSG3', Aliases='CDHF6, PVA', description='desmoglein 3', other_designations='130 kDa pemphigus vulgaris antigen|130-kD pemphigus vulgaris antigen|cadherin family member 6|pemphigus vulgaris antigen', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31447769, end_position_on_the_genomic_accession=31478702, orientation='plus', exon_count=16, OMIM=169615, no_hdr0=''),
  1831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1831, CurrentID=0, Status='live', Symbol='TSC22D3', Aliases='DIP, DSIPI, GILZ, TSC-22R, hDIP', description='TSC22 domain family member 3', other_designations='DSIP-immunoreactive leucine zipper protein|DSIP-immunoreactive peptide|TSC-22 related protein|TSC-22-like protein|TSC-22-related protein|TSC22 domain family, member 3|delta sleep inducing peptide, immunoreactor|delta sleep-inducing peptide immunoreactor|glucocorticoid-induced leucine zipper protein', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107713221, end_position_on_the_genomic_accession=107777329, orientation='minus', exon_count=8, OMIM=300506, no_hdr0=''),
  1832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1832, CurrentID=0, Status='live', Symbol='DSP', Aliases='DCWHKTA, DP, DPI, DPII', description='desmoplakin', other_designations='250/210 kDa paraneoplastic pemphigus antigen|desmoplakin I|desmoplakin II', map_location='6p24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7541575, end_position_on_the_genomic_accession=7586713, orientation='plus', exon_count=24, OMIM=125647, no_hdr0=''),
  1833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1833, CurrentID=0, Status='live', Symbol='EPYC', Aliases='DSPG3, PGLB, Pg-Lb, SLRR3B', description='epiphycan', other_designations='dermatan sulfate proteoglycan 3|epiphycan proteoglycan|proteoglycan-lb|small chondroitin/dermatan sulfate proteoglycan', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90963679, end_position_on_the_genomic_accession=91007906, orientation='minus', exon_count=9, OMIM=601657, no_hdr0=''),
  1834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1834, CurrentID=0, Status='live', Symbol='DSPP', Aliases='DFNA39, DGI1, DMP3, DPP, DSP, DTDP2', description='dentin sialophosphoprotein', other_designations='dentin phosphophoryn|dentin phosphoprotein|dentin phosphoryn|dentin sialoprotein', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87608529, end_position_on_the_genomic_accession=87616873, orientation='plus', exon_count=5, OMIM=125485, no_hdr0=''),
  1836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1836, CurrentID=0, Status='live', Symbol='SLC26A2', Aliases='D5S1708, DTD, DTDST, EDM4, MST153, MSTP157', description='solute carrier family 26 (anion exchanger), member 2', other_designations='diastrophic dysplasia protein|solute carrier family 26 (sulfate transporter), member 2|sulfate anion transporter 1', map_location='5q31-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149960737, end_position_on_the_genomic_accession=149987400, orientation='plus', exon_count=3, OMIM=606718, no_hdr0=''),
  1837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1837, CurrentID=0, Status='live', Symbol='DTNA', Aliases='D18S892E, DRP3, DTN, DTN-A, LVNC1', description='dystrobrevin alpha', other_designations='dystrobrevin, alpha|dystrophin-related protein 3', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34493290, end_position_on_the_genomic_accession=34891844, orientation='plus', exon_count=32, OMIM=601239, no_hdr0=''),
  1838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1838, CurrentID=0, Status='live', Symbol='DTNB', Aliases='', description='dystrobrevin beta', other_designations='DTN-B|beta-dystrobrevin|dystrobrevin, beta', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25377198, end_position_on_the_genomic_accession=25673647, orientation='minus', exon_count=27, OMIM=602415, no_hdr0=''),
  1839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1839, CurrentID=0, Status='live', Symbol='HBEGF', Aliases='DTR, DTS, DTSF, HEGFL', description='heparin-binding EGF-like growth factor', other_designations='diphtheria toxin receptor (heparin-binding EGF-like growth factor)|diphtheria toxin receptor (heparin-binding epidermal growth factor-like growth factor)|heparin-binding epidermal growth factor', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140332843, end_position_on_the_genomic_accession=140346603, orientation='minus', exon_count=6, OMIM=126150, no_hdr0=''),
  1840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1840, CurrentID=0, Status='live', Symbol='DTX1', Aliases='hDx-1', description='deltex 1, E3 ubiquitin ligase', other_designations='deltex homolog 1|deltex1|hDTX1|protein deltex-1', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113056714, end_position_on_the_genomic_accession=113098028, orientation='plus', exon_count=10, OMIM=602582, no_hdr0=''),
  1841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1841, CurrentID=0, Status='live', Symbol='DTYMK', Aliases='CDC8, PP3731, TMPK, TYMK', description='deoxythymidylate kinase', other_designations='dTMP kinase|deoxythymidylate kinase (thymidylate kinase)', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241675742, end_position_on_the_genomic_accession=241686968, orientation='minus', exon_count=5, OMIM=188345, no_hdr0=''),
  1842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1842, CurrentID=0, Status='live', Symbol='ECM2', Aliases='', description='extracellular matrix protein 2, female organ and adipocyte specific', other_designations='matrix glycoprotein SC1/ECM2', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92493547, end_position_on_the_genomic_accession=92536092, orientation='minus', exon_count=11, OMIM=603479, no_hdr0=''),
  1843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1843, CurrentID=0, Status='live', Symbol='DUSP1', Aliases='CL100, HVH1, MKP-1, MKP1, PTPN10', description='dual specificity phosphatase 1', other_designations='CL 100|MAP kinase phosphatase 1|dual specificity protein phosphatase hVH1|mitogen-activated protein kinase phosphatase 1|protein-tyrosine phosphatase CL100|serine/threonine specific protein phosphatase', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172768090, end_position_on_the_genomic_accession=172771200, orientation='minus', exon_count=4, OMIM=600714, no_hdr0=''),
  1844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1844, CurrentID=0, Status='live', Symbol='DUSP2', Aliases='PAC-1, PAC1', description='dual specificity phosphatase 2', other_designations='dual specificity protein phosphatase PAC-1|dual-specificity phosphatase 2|serine/threonine specific protein phosphatase', map_location='2q11', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96143169, end_position_on_the_genomic_accession=96145440, orientation='minus', exon_count=4, OMIM=603068, no_hdr0=''),
  1845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1845, CurrentID=0, Status='live', Symbol='DUSP3', Aliases='VHR', description='dual specificity phosphatase 3', other_designations='dual specificity protein phosphatase VHR|serine/threonine specific protein phosphatase|vaccinia H1-related phosphatase|vaccinia virus phosphatase VH1-related', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43766121, end_position_on_the_genomic_accession=43779000, orientation='minus', exon_count=3, OMIM=600183, no_hdr0=''),
  1846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1846, CurrentID=0, Status='live', Symbol='DUSP4', Aliases='HVH2, MKP-2, MKP2, TYP', description='dual specificity phosphatase 4', other_designations='MAP kinase phosphatase 2|VH1 homologous phosphatase 2|dual specificity protein phosphatase hVH2|mitogen-activated protein kinase phosphatase 2|serine/threonine specific protein phosphatase', map_location='8p12-p11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29333062, end_position_on_the_genomic_accession=29350750, orientation='minus', exon_count=6, OMIM=602747, no_hdr0=''),
  1847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1847, CurrentID=0, Status='live', Symbol='DUSP5', Aliases='DUSP, HVH3', description='dual specificity phosphatase 5', other_designations='VH1-like phosphatase 3|dual specificity protein phosphatase hVH3|serine/threonine specific protein phosphatase', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110497867, end_position_on_the_genomic_accession=110511544, orientation='plus', exon_count=4, OMIM=603069, no_hdr0=''),
  1848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1848, CurrentID=0, Status='live', Symbol='DUSP6', Aliases='HH19, MKP3, PYST1', description='dual specificity phosphatase 6', other_designations='MAP kinase phosphatase 3|dual specificity protein phosphatase PYST1|mitogen-activated protein kinase phosphatase 3|serine/threonine specific protein phosphatase', map_location='12q22-q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89347825, end_position_on_the_genomic_accession=89352859, orientation='minus', exon_count=3, OMIM=602748, no_hdr0=''),
  1849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1849, CurrentID=0, Status='live', Symbol='DUSP7', Aliases='MKPX, PYST2', description='dual specificity phosphatase 7', other_designations='dual specificity protein phosphatase PYST2|dual-specificity phosphatase-7', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52048919, end_position_on_the_genomic_accession=52056445, orientation='minus', exon_count=3, OMIM=602749, no_hdr0=''),
  1850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1850, CurrentID=0, Status='live', Symbol='DUSP8', Aliases='C11orf81, HB5, HVH-5, HVH8', description='dual specificity phosphatase 8', other_designations='H1 phosphatase, vaccinia virus homolog|dual specificity protein phosphatase hVH-5|serine/threonine specific protein phosphatase', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1554051, end_position_on_the_genomic_accession=1572831, orientation='minus', exon_count=9, OMIM=602038, no_hdr0=''),
  284577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284577, CurrentID=0, Status='live', Symbol='LOC284577', Aliases='', description='uncharacterized LOC284577', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205455922, end_position_on_the_genomic_accession=205469106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  1852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1852, CurrentID=0, Status='live', Symbol='DUSP9', Aliases='MKP-4, MKP4', description='dual specificity phosphatase 9', other_designations='map kinase phosphatase 4|mitogen-activated protein kinase phosphatase 4|serine/threonine specific protein phosphatase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153642443, end_position_on_the_genomic_accession=153651326, orientation='plus', exon_count=7, OMIM=300134, no_hdr0=''),
  1854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1854, CurrentID=0, Status='live', Symbol='DUT', Aliases='dUTPase', description='deoxyuridine triphosphatase', other_designations='dUTP diphosphatase|dUTP nucleotidohydrolase|dUTP pyrophosphatase', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48331100, end_position_on_the_genomic_accession=48343373, orientation='plus', exon_count=8, OMIM=601266, no_hdr0=''),
  1855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1855, CurrentID=0, Status='live', Symbol='DVL1', Aliases='DRS2, DVLL1, DVL1P1, DVL1', description='dishevelled segment polarity protein 1', other_designations='DSH homolog 1|dishevelled 1 (homologous to Drosophila dsh)|dishevelled, dsh homolog 1|dishevelled-1', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1335278, end_position_on_the_genomic_accession=1349158, orientation='minus', exon_count=16, OMIM=601365, no_hdr0=''),
  1856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1856, CurrentID=0, Status='live', Symbol='DVL2', Aliases='', description='dishevelled segment polarity protein 2', other_designations='dishevelled 2 (homologous to Drosophila dsh)|dishevelled, dsh homolog 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7225342, end_position_on_the_genomic_accession=7234548, orientation='minus', exon_count=16, OMIM=602151, no_hdr0=''),
  1857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1857, CurrentID=0, Status='live', Symbol='DVL3', Aliases='', description='dishevelled segment polarity protein 3', other_designations='DSH homolog 3|dishevelled 3 (homologous to Drosophila dsh)|dishevelled, dsh homolog 3|dishevelled-3', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184155311, end_position_on_the_genomic_accession=184173526, orientation='plus', exon_count=15, OMIM=601368, no_hdr0=''),
  1858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1858, CurrentID=0, Status='live', Symbol='DWS', Aliases='C3DELq22q24, DEL3q22q24', description='dandy-walker syndrome', other_designations='', map_location='3q22-q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=220200, no_hdr0=''),
  1859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1859, CurrentID=0, Status='live', Symbol='DYRK1A', Aliases='DYRK, DYRK1, HP86, MNB, MNBH, MRD7', description='dual specificity tyrosine-(Y)-phosphorylation regulated kinase 1A', other_designations='MNB/DYRK protein kinase|dual specificity YAK1-related kinase|dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 1A|hMNB|mnb protein kinase homolog hp86|protein kinase minibrain homolog|serine/threonine kinase MNB|serine/threonine-specific protein kinase', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37365790, end_position_on_the_genomic_accession=37515376, orientation='plus', exon_count=18, OMIM=600855, no_hdr0=''),
  1861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1861, CurrentID=0, Status='live', Symbol='TOR1A', Aliases='DQ2, DYT1', description='torsin family 1, member A (torsin A)', other_designations='dystonia 1 protein|dystonia 1, torsion (autosomal dominant; torsin A)|torsin ATPase 1|torsin ATPase-1A', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129812942, end_position_on_the_genomic_accession=129824245, orientation='minus', exon_count=5, OMIM=605204, no_hdr0=''),
  1862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1862, CurrentID=0, Status='live', Symbol='DYT2', Aliases='', description='dystonia 2, torsion (autosomal recessive)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=224500, no_hdr0=''),
  1866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1866, CurrentID=0, Status='live', Symbol='DYT7', Aliases='', description='dystonia 7, torsion (autosomal dominant)', other_designations='Dystonia-7 (torsion dystonia, adult-onset, focal)', map_location='18p', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602124, no_hdr0=''),
  1867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1867, CurrentID=0, Status='live', Symbol='DYX1', Aliases='RD', description='dyslexia susceptibility 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=127700, no_hdr0=''),
  1869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1869, CurrentID=0, Status='live', Symbol='E2F1', Aliases='E2F-1, RBAP1, RBBP3, RBP3', description='E2F transcription factor 1', other_designations='PBR3|PRB-binding protein E2F-1|RBAP-1|RBBP-3|retinoblastoma-associated protein 1|retinoblastoma-binding protein 3', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33675486, end_position_on_the_genomic_accession=33686404, orientation='minus', exon_count=7, OMIM=189971, no_hdr0=''),
  1870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1870, CurrentID=0, Status='live', Symbol='E2F2', Aliases='E2F-2', description='E2F transcription factor 2', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23505974, end_position_on_the_genomic_accession=23531250, orientation='minus', exon_count=10, OMIM=600426, no_hdr0=''),
  1871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1871, CurrentID=0, Status='live', Symbol='E2F3', Aliases='E2F-3', description='E2F transcription factor 3', other_designations='', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20401906, end_position_on_the_genomic_accession=20493714, orientation='plus', exon_count=11, OMIM=600427, no_hdr0=''),
  1872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1872, CurrentID=0, Status='live', Symbol='E2F3P1', Aliases='', description='E2F transcription factor 3 pseudogene 1', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35490032, end_position_on_the_genomic_accession=35494381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1874, CurrentID=0, Status='live', Symbol='E2F4', Aliases='E2F-4', description='E2F transcription factor 4, p107/p130-binding', other_designations='p107/p130-binding protein', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67192165, end_position_on_the_genomic_accession=67198918, orientation='plus', exon_count=10, OMIM=600659, no_hdr0=''),
  1875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1875, CurrentID=0, Status='live', Symbol='E2F5', Aliases='E2F-5', description='E2F transcription factor 5, p130-binding', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85177056, end_position_on_the_genomic_accession=85214518, orientation='plus', exon_count=10, OMIM=600967, no_hdr0=''),
  1876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1876, CurrentID=0, Status='live', Symbol='E2F6', Aliases='E2F-6', description='E2F transcription factor 6', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11444375, end_position_on_the_genomic_accession=11466177, orientation='minus', exon_count=9, OMIM=602944, no_hdr0=''),
  1877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1877, CurrentID=0, Status='live', Symbol='E4F1', Aliases='E4F', description='E4F transcription factor 1', other_designations='p120E4F|p50E4F|putative E3 ubiquitin-protein ligase E4F1|transcription factor E4F', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2223488, end_position_on_the_genomic_accession=2235742, orientation='plus', exon_count=14, OMIM=603022, no_hdr0=''),
  1878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1878, CurrentID=0, Status='live', Symbol='E11S', Aliases='', description='ECHO virus (serotypes 4, 6, 11, 19) sensitivity', other_designations='', map_location='19q13.1-q13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=129150, no_hdr0=''),
  1879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1879, CurrentID=0, Status='live', Symbol='EBF1', Aliases='COE1, EBF, O/E-1, OLF1', description='early B-cell factor 1', other_designations='Collier, Olf and EBF transcription factor 1|OE-1|olfactory neuronal transcription factor 1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158695908, end_position_on_the_genomic_accession=159099780, orientation='minus', exon_count=21, OMIM=164343, no_hdr0=''),
  1880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1880, CurrentID=0, Status='live', Symbol='GPR183', Aliases='EBI2', description='G protein-coupled receptor 183', other_designations='EBV-induced G-protein coupled receptor 2|Epstein-Barr virus induced gene 2 (lymphocyte-specific G protein-coupled receptor)|epstein-Barr virus-induced G-protein coupled receptor 2', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99294535, end_position_on_the_genomic_accession=99307495, orientation='minus', exon_count=2, OMIM=605741, no_hdr0=''),
  1881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1881, CurrentID=0, Status='live', Symbol='EBM', Aliases='', description='epidermolysis bullosa, macular type', other_designations='', map_location='Xq27.3-qter', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302000, no_hdr0=''),
  132954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132954, CurrentID=0, Status='live', Symbol='PDCL2', Aliases='GCPHLP', description='phosducin like 2', other_designations='phosducin-like 2', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55556523, end_position_on_the_genomic_accession=55592268, orientation='minus', exon_count=7, OMIM=611676, no_hdr0=''),
  1884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1884, CurrentID=0, Status='live', Symbol='EBR3', Aliases='ULG4', description='epidermolysis bullosa 3, progressiva', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=226500, no_hdr0=''),
  1885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1885, CurrentID=0, Status='live', Symbol='EBR4', Aliases='', description='epidermolysis bullosa 4, pseudojunctional (intraepidermal)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1888, CurrentID=0, Status='live', Symbol='EBVS1', Aliases='', description='Epstein Barr virus integration site 1', other_designations='Epstein Barr virus insertion site 1|Epstein-Barr virus integration site', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=132850, no_hdr0=''),
  1889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1889, CurrentID=0, Status='live', Symbol='ECE1', Aliases='ECE', description='endothelin converting enzyme 1', other_designations='ECE-1', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21217247, end_position_on_the_genomic_accession=21345541, orientation='minus', exon_count=23, OMIM=600423, no_hdr0=''),
  1890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1890, CurrentID=0, Status='live', Symbol='TYMP', Aliases='ECGF, ECGF1, MEDPS1, MNGIE, MTDPS1, PDECGF, TP, hPD-ECGF', description='thymidine phosphorylase', other_designations='gliostatin|tdRPase', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50525752, end_position_on_the_genomic_accession=50530085, orientation='minus', exon_count=10, OMIM=131222, no_hdr0=''),
  1891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1891, CurrentID=0, Status='live', Symbol='ECH1', Aliases='HPXEL', description='enoyl-CoA hydratase 1, peroxisomal', other_designations='delta3,5-delta2,4-dienoyl-CoA isomerase|dienoyl-CoA isomerase|enoyl CoA hydratase 1, peroxisomal|enoyl Coenzyme A hydratase 1, peroxisomal|peroxisomal enoyl-CoA hydratase 1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38815422, end_position_on_the_genomic_accession=38831857, orientation='minus', exon_count=10, OMIM=600696, no_hdr0=''),
  1892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1892, CurrentID=0, Status='live', Symbol='ECHS1', Aliases='ECHS1D, SCEH', description='enoyl-CoA hydratase, short chain, 1, mitochondrial', other_designations='enoyl CoA hydratase, short chain, 1, mitochondrial|enoyl Coenzyme A hydratase, short chain, 1, mitochondrial', map_location='10q26.2-q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133362483, end_position_on_the_genomic_accession=133373404, orientation='minus', exon_count=8, OMIM=602292, no_hdr0=''),
  1893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1893, CurrentID=0, Status='live', Symbol='ECM1', Aliases='URBWD', description='extracellular matrix protein 1', other_designations='secretory component p85', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150508011, end_position_on_the_genomic_accession=150513789, orientation='plus', exon_count=10, OMIM=602201, no_hdr0=''),
  1894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1894, CurrentID=0, Status='live', Symbol='ECT2', Aliases='ARHGEF31', description='epithelial cell transforming 2', other_designations='epithelial cell transforming sequence 2 oncogene|epithelial cell-transforming sequence 2 oncogene', map_location='3q26.1-q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172750682, end_position_on_the_genomic_accession=172829265, orientation='plus', exon_count=29, OMIM=600586, no_hdr0=''),
  1896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1896, CurrentID=0, Status='live', Symbol='EDA', Aliases='ECTD1, ED1, ED1-A1, ED1-A2-A1, EDA-A2, EDA1, EDA2, HED, HED1, ODT1, STHAGX1, XHED, XLHED, EDA', description='ectodysplasin A', other_designations='X-linked anhidroitic ectodermal dysplasia protein|oligodontia 1', map_location='Xq12-q13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69616067, end_position_on_the_genomic_accession=70039472, orientation='plus', exon_count=11, OMIM=300451, no_hdr0=''),
  1901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1901, CurrentID=0, Status='live', Symbol='S1PR1', Aliases='CD363, CHEDG1, D1S3362, ECGF1, EDG-1, EDG1, S1P1', description='sphingosine-1-phosphate receptor 1', other_designations='S1P receptor 1|S1P receptor Edg-1|endothelial differentiation G-protein coupled receptor 1|endothelial differentiation, sphingolipid G-protein-coupled receptor, 1|sphingosine 1-phosphate receptor EDG1|sphingosine 1-phosphate receptor Edg-1', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101236749, end_position_on_the_genomic_accession=101241520, orientation='plus', exon_count=2, OMIM=601974, no_hdr0=''),
  1902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1902, CurrentID=0, Status='live', Symbol='LPAR1', Aliases='EDG2, GPR26, Gpcr26, LPA1, Mrec1.3, VZG1, edg-2, rec.1.3, vzg-1', description='lysophosphatidic acid receptor 1', other_designations='LPA receptor 1|LPA-1|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor, 2|lysophosphatidic acid receptor Edg-2|ventricular zone gene 1', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110873261, end_position_on_the_genomic_accession=111039243, orientation='minus', exon_count=7, OMIM=602282, no_hdr0=''),
  1903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1903, CurrentID=0, Status='live', Symbol='S1PR3', Aliases='EDG-3, EDG3, LPB3, S1P3', description='sphingosine-1-phosphate receptor 3', other_designations='G protein-coupled receptor, endothelial differentiation gene-3|S1P receptor 3|S1P receptor EDG3|S1P receptor Edg-3|endothelial differentiation G-protein coupled receptor 3|endothelial differentiation, sphingolipid G-protein-coupled receptor, 3|sphingosine 1-phosphate receptor Edg-3', map_location='9q22.1-q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88991409, end_position_on_the_genomic_accession=89005155, orientation='plus', exon_count=2, OMIM=601965, no_hdr0=''),
  1906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1906, CurrentID=0, Status='live', Symbol='EDN1', Aliases='ARCND3, ET1, HDLCQ7, PPET1, QME', description='endothelin 1', other_designations='preproendothelin-1', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12261214, end_position_on_the_genomic_accession=12297194, orientation='plus', exon_count=7, OMIM=131240, no_hdr0=''),
  1907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1907, CurrentID=0, Status='live', Symbol='EDN2', Aliases='ET-2, ET2, PPET2', description='endothelin 2', other_designations='preproendothelin 2', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41478775, end_position_on_the_genomic_accession=41484683, orientation='minus', exon_count=5, OMIM=131241, no_hdr0=''),
  1908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1908, CurrentID=0, Status='live', Symbol='EDN3', Aliases='ET-3, ET3, HSCR4, PPET3, WS4B', description='endothelin 3', other_designations='preproendothelin-3', map_location='20q13.2-q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59300424, end_position_on_the_genomic_accession=59325992, orientation='plus', exon_count=6, OMIM=131242, no_hdr0=''),
  1909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1909, CurrentID=0, Status='live', Symbol='EDNRA', Aliases='ET-A, ETA, ETA-R, ETAR, ETRA, MFDA, hET-AR', description='endothelin receptor type A', other_designations='G protein-coupled receptor|endothelin receptor subtype A|endothelin-1-specific receptor', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147480917, end_position_on_the_genomic_accession=147544954, orientation='plus', exon_count=8, OMIM=131243, no_hdr0=''),
  1910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1910, CurrentID=0, Status='live', Symbol='EDNRB', Aliases='ABCDS, ET-B, ET-BR, ETB, ETB1, ETBR, ETRB, HSCR, HSCR2, WS4A', description='endothelin receptor type B', other_designations='endothelin receptor non-selective type|endothelin receptor subtype B1', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77895481, end_position_on_the_genomic_accession=77975529, orientation='minus', exon_count=10, OMIM=131244, no_hdr0=''),
  1911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1911, CurrentID=0, Status='live', Symbol='PHC1', Aliases='EDR1, HPH1, MCPH11, RAE28', description='polyhomeotic homolog 1 (Drosophila)', other_designations='early development regulator 1 (homolog of polyhomeotic 1)|early development regulatory protein 1|polyhomeotic-like 1', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8914684, end_position_on_the_genomic_accession=8941467, orientation='plus', exon_count=16, OMIM=602978, no_hdr0=''),
  1912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1912, CurrentID=0, Status='live', Symbol='PHC2', Aliases='EDR2, HPH2, PH2', description='polyhomeotic homolog 2 (Drosophila)', other_designations='early development regulator 2 (homolog of polyhomeotic 2)|early development regulatory protein 2|polyhomeotic 2|polyhomeotic-like 2', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33323623, end_position_on_the_genomic_accession=33430610, orientation='minus', exon_count=24, OMIM=602979, no_hdr0=''),
  1913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1913, CurrentID=0, Status='live', Symbol='EEC1', Aliases='EEC', description='ectrodactyly, ectodermal dysplasia and cleft lip/palate syndrome 1', other_designations='Ectrodactyly, ectodermal dysplasia, cleft lip/palate, 1', map_location='7q11.2-q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=129900, no_hdr0=''),
  1914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1914, CurrentID=0, Status='live', Symbol='EEC2', Aliases='', description='ectrodactyly, ectodermal dysplasia and cleft lip/palate syndrome 2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1915, CurrentID=0, Status='live', Symbol='EEF1A1', Aliases='CCS-3, CCS3, EE1A1, EEF-1, EEF1A, EF-Tu, EF1A, GRAF-1EF, HNGC:16303, LENG7, PTI1, eEF1A-1', description='eukaryotic translation elongation factor 1 alpha 1', other_designations='CTCL tumor antigen|EF-1-alpha-1|EF1a-like protein|cervical cancer suppressor 3|elongation factor 1 alpha subunit|elongation factor Tu|eukaryotic elongation factor 1 A-1|eukaryotic translation elongation factor 1 alpha 1-like 14|glucocorticoid receptor AF-1 specific elongation factor|leukocyte receptor cluster (LRC) member 7|leukocyte receptor cluster member 7|prostate tumor-inducing protein 1|translation elongation factor 1 alpha 1-like 14', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73515750, end_position_on_the_genomic_accession=73521032, orientation='minus', exon_count=9, OMIM=130590, no_hdr0=''),
  1917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1917, CurrentID=0, Status='live', Symbol='EEF1A2', Aliases='EEF1AL, EF-1-alpha-2, EF1A, EIEE33, HS1, MRD38, STN, STNL', description='eukaryotic translation elongation factor 1 alpha 2', other_designations='eukaryotic elongation factor 1 A-2|statin-S1', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63488012, end_position_on_the_genomic_accession=63499315, orientation='minus', exon_count=8, OMIM=602959, no_hdr0=''),
  1918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1918, CurrentID=0, Status='live', Symbol='EEF1A1P43', Aliases='EEF1A3', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 43', other_designations='eukaryotic translation elongation factor 1 alpha 3', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17554807, end_position_on_the_genomic_accession=17586147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1932, CurrentID=0, Status='live', Symbol='EEF1B2P1', Aliases='EEF1B1, EF1B', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 1', other_designations='eukaryotic translation elongation factor 1 beta 1', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52505029, end_position_on_the_genomic_accession=52505908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1933, CurrentID=0, Status='live', Symbol='EEF1B2', Aliases='EEF1B, EEF1B1, EF1B', description='eukaryotic translation elongation factor 1 beta 2', other_designations='EF-1-beta|eukaryotic translation elongation factor 1 beta 1', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206159594, end_position_on_the_genomic_accession=206162929, orientation='plus', exon_count=7, OMIM=600655, no_hdr0=''),
  1934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1934, CurrentID=0, Status='live', Symbol='EEF1B2P2', Aliases='BETA-3, EEF-1, EEF-1BETA5A, EEF1B, EEF1B3', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 2', other_designations='eukaryotic translation elongation factor 1 beta 3', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68159175, end_position_on_the_genomic_accession=68159962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1936, CurrentID=0, Status='live', Symbol='EEF1D', Aliases='EF-1D, EF1D, FP1047', description='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein)', other_designations='EF-1-delta|antigen NY-CO-4|guanine nucleotide exchange protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143579697, end_position_on_the_genomic_accession=143597675, orientation='minus', exon_count=15, OMIM=130592, no_hdr0=''),
  1937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1937, CurrentID=0, Status='live', Symbol='EEF1G', Aliases='EF1G, GIG35', description='eukaryotic translation elongation factor 1 gamma', other_designations='EF-1-gamma|PRO1608|eEF-1B gamma|pancreatic tumor-related protein|translation elongation factor eEF-1 gamma chain', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62559601, end_position_on_the_genomic_accession=62573988, orientation='minus', exon_count=10, OMIM=130593, no_hdr0=''),
  1938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1938, CurrentID=0, Status='live', Symbol='EEF2', Aliases='EEF-2, EF-2, EF2, SCA26', description='eukaryotic translation elongation factor 2', other_designations='polypeptidyl-tRNA translocase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3976056, end_position_on_the_genomic_accession=3985463, orientation='minus', exon_count=15, OMIM=130610, no_hdr0=''),
  1939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1939, CurrentID=0, Status='live', Symbol='EIF2D', Aliases='HCA56, LGTN', description='eukaryotic translation initiation factor 2D', other_designations='hepatocellular carcinoma-associated antigen 56|ligatin', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206571297, end_position_on_the_genomic_accession=206612551, orientation='minus', exon_count=19, OMIM=613709, no_hdr0=''),
  1940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1940, CurrentID=0, Status='live', Symbol='EEGV1', Aliases='EEGL', description='electro-encephalographic variant pattern 1', other_designations='', map_location='20q', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=130180, no_hdr0=''),
  1942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1942, CurrentID=0, Status='live', Symbol='EFNA1', Aliases='B61, ECKLG, EFL1, EPLG1, LERK-1, LERK1, TNFAIP4', description='ephrin-A1', other_designations='TNF alpha-induced protein 4|eph-related receptor tyrosine kinase ligand 1|immediate early response protein B61|ligand of eph-related kinase 1|tumor necrosis factor alpha-induced protein 4|tumor necrosis factor, alpha-induced protein 4', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155127873, end_position_on_the_genomic_accession=155134910, orientation='plus', exon_count=6, OMIM=191164, no_hdr0=''),
  1943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1943, CurrentID=0, Status='live', Symbol='EFNA2', Aliases='ELF-1, EPLG6, HEK7-L, LERK-6, LERK6', description='ephrin-A2', other_designations='HEK7 ligand|eph-related receptor tyrosine kinase ligand 6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1286154, end_position_on_the_genomic_accession=1301431, orientation='plus', exon_count=4, OMIM=602756, no_hdr0=''),
  1944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1944, CurrentID=0, Status='live', Symbol='EFNA3', Aliases='EFL2, EPLG3, Ehk1-L, LERK3', description='ephrin-A3', other_designations='EFL-2|EHK1 ligand|LERK-3|eph-related receptor tyrosine kinase ligand 3|ligand of eph-related kinase 3', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155078872, end_position_on_the_genomic_accession=155087538, orientation='plus', exon_count=5, OMIM=601381, no_hdr0=''),
  1945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1945, CurrentID=0, Status='live', Symbol='EFNA4', Aliases='EFL4, EPLG4, LERK4', description='ephrin-A4', other_designations='LERK-4|eph-related receptor tyrosine kinase ligand 4|ligand of eph-related kinase 4', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155063737, end_position_on_the_genomic_accession=155069553, orientation='plus', exon_count=4, OMIM=601380, no_hdr0=''),
  1946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1946, CurrentID=0, Status='live', Symbol='EFNA5', Aliases='AF1, EFL5, EPLG7, GLC1M, LERK7, RAGS', description='ephrin-A5', other_designations='AL-1|LERK-7|eph-related receptor tyrosine kinase ligand 7', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=107376889, end_position_on_the_genomic_accession=107670895, orientation='minus', exon_count=7, OMIM=601535, no_hdr0=''),
  1947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1947, CurrentID=0, Status='live', Symbol='EFNB1', Aliases='CFND, CFNS, EFB1, EFL3, EPLG2, Elk-L, LERK2', description='ephrin-B1', other_designations='ELK ligand|eph-related receptor tyrosine kinase ligand 2', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68828997, end_position_on_the_genomic_accession=68842164, orientation='plus', exon_count=5, OMIM=300035, no_hdr0=''),
  1948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1948, CurrentID=0, Status='live', Symbol='EFNB2', Aliases='EPLG5, HTKL, Htk-L, LERK5', description='ephrin-B2', other_designations='HTK ligand|LERK-5|eph-related receptor tyrosine kinase ligand 5|ligand of eph-related kinase 5', map_location='13q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106489731, end_position_on_the_genomic_accession=106535040, orientation='minus', exon_count=5, OMIM=600527, no_hdr0=''),
  1949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1949, CurrentID=0, Status='live', Symbol='EFNB3', Aliases='EFL6, EPLG8, LERK8', description='ephrin-B3', other_designations='EPH-related receptor transmembrane ligand ELK-L3|Ephrin B3|eph-related receptor tyrosine kinase ligand 8', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7705202, end_position_on_the_genomic_accession=7711375, orientation='plus', exon_count=5, OMIM=602297, no_hdr0=''),
  1950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1950, CurrentID=0, Status='live', Symbol='EGF', Aliases='HOMG4, URG', description='epidermal growth factor', other_designations='beta-urogastrone', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109912883, end_position_on_the_genomic_accession=110012962, orientation='plus', exon_count=26, OMIM=131530, no_hdr0=''),
  1951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1951, CurrentID=0, Status='live', Symbol='CELSR3', Aliases='ADGRC3, CDHF11, EGFL1, FMI1, HFMI1, MEGF2, RESDA1', description='cadherin, EGF LAG seven-pass G-type receptor 3', other_designations='EGF-like protein 1|EGF-like-domain, multiple 1|adhesion G protein-coupled receptor C3|anchor protein|cadherin family member 11|cadherin, EGF LAG seven-pass G-type receptor 3 (flamingo homolog, Drosophila)|epidermal growth factor-like protein 1|flamingo homolog 1|multiple EGF-like domains protein 2|multiple epidermal growth factor-like domains protein 2', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48636463, end_position_on_the_genomic_accession=48662915, orientation='minus', exon_count=35, OMIM=604264, no_hdr0=''),
  1952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1952, CurrentID=0, Status='live', Symbol='CELSR2', Aliases='ADGRC2, CDHF10, EGFL2, Flamingo1, MEGF3', description='cadherin, EGF LAG seven-pass G-type receptor 2', other_designations='EGF-like protein 2|EGF-like-domain, multiple 2|adhesion G protein-coupled receptor C2|cadherin family member 10|cadherin, EGF LAG seven-pass G-type receptor 2 (flamingo homolog, Drosophila)|epidermal growth factor-like 2|epidermal growth factor-like protein 2|flamingo homolog 3|multiple EGF-like domains protein 3|multiple epidermal growth factor-like domains 3|multiple epidermal growth factor-like domains protein 3', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109249715, end_position_on_the_genomic_accession=109275756, orientation='plus', exon_count=34, OMIM=604265, no_hdr0=''),
  1953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1953, CurrentID=0, Status='live', Symbol='MEGF6', Aliases='EGFL3', description='multiple EGF-like-domains 6', other_designations='EGF-like protein 3|EGF-like-domain, multiple 3|epidermal growth factor-like protein 3|multiple EGF-like domains protein 6', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3487942, end_position_on_the_genomic_accession=3624757, orientation='minus', exon_count=41, OMIM=604266, no_hdr0=''),
  1954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1954, CurrentID=0, Status='live', Symbol='MEGF8', Aliases='C19orf49, CRPT2, EGFL4, SBP1', description='multiple EGF-like-domains 8', other_designations='EGF-like domain-containing protein 4|EGF-like-domain, multiple 4|HBV pre-S2-binding protein 1|HBV pre-s2 binding protein 1|epidermal growth factor-like protein 4|hepatitis B virus pre-S2-binding protein 1', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42325609, end_position_on_the_genomic_accession=42378769, orientation='plus', exon_count=42, OMIM=604267, no_hdr0=''),
  1955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1955, CurrentID=0, Status='live', Symbol='MEGF9', Aliases='EGFL5', description='multiple EGF-like-domains 9', other_designations='EGF-like-domain, multiple 5|epidermal growth factor-like protein 5', map_location='9q32-q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120600918, end_position_on_the_genomic_accession=120714487, orientation='minus', exon_count=6, OMIM=604268, no_hdr0=''),
  1956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1956, CurrentID=0, Status='live', Symbol='EGFR', Aliases='ERBB, ERBB1, HER1, NISBD2, PIG61, mENA', description='epidermal growth factor receptor', other_designations='avian erythroblastic leukemia viral (v-erb-b) oncogene homolog|cell growth inhibiting protein 40|cell proliferation-inducing protein 61|erb-b2 receptor tyrosine kinase 1|proto-oncogene c-ErbB-1|receptor tyrosine-protein kinase erbB-1', map_location='7p12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55019032, end_position_on_the_genomic_accession=55207338, orientation='plus', exon_count=30, OMIM=131550, no_hdr0=''),
  1957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1957, CurrentID=0, Status='live', Symbol='EGI', Aliases='EIG, EIG1', description='epilepsy, generalized, idiopathic', other_designations='epilepsy, idiopathic generalized, susceptibility to 1', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600669, no_hdr0=''),
  1958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1958, CurrentID=0, Status='live', Symbol='EGR1', Aliases='AT225, G0S30, KROX-24, NGFI-A, TIS8, ZIF-268, ZNF225', description='early growth response 1', other_designations='EGR-1|nerve growth factor-induced protein A|transcription factor ETR103|transcription factor Zif268|zinc finger protein 225|zinc finger protein Krox-24', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138465492, end_position_on_the_genomic_accession=138469315, orientation='plus', exon_count=2, OMIM=128990, no_hdr0=''),
  1959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1959, CurrentID=0, Status='live', Symbol='EGR2', Aliases='AT591, CMT1D, CMT4E, KROX20', description='early growth response 2', other_designations='KROX-20, Drosophila, homolog (early growth response-2)|early growth response protein 2|zinc finger protein Krox-20', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62811996, end_position_on_the_genomic_accession=62919895, orientation='minus', exon_count=11, OMIM=129010, no_hdr0=''),
  1960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1960, CurrentID=0, Status='live', Symbol='EGR3', Aliases='EGR-3, PILOT', description='early growth response 3', other_designations='zinc finger protein pilot', map_location='8p23-p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22687659, end_position_on_the_genomic_accession=22693302, orientation='minus', exon_count=5, OMIM=602419, no_hdr0=''),
  1961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1961, CurrentID=0, Status='live', Symbol='EGR4', Aliases='NGFI-C, NGFIC, PAT133', description='early growth response 4', other_designations='AT133|EGR-4', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73290929, end_position_on_the_genomic_accession=73293701, orientation='minus', exon_count=2, OMIM=128992, no_hdr0=''),
  1962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1962, CurrentID=0, Status='live', Symbol='EHHADH', Aliases='ECHD, FRTS3, L-PBE, LBFP, LBP, PBFE', description='enoyl-CoA, hydratase/3-hydroxyacyl CoA dehydrogenase', other_designations='3,2-trans-enoyl-CoA isomerase|L-3-hydroxyacyl-CoA dehydrogenase|L-bifunctional protein, peroxisomal|PBE|enoyl-Coenzyme A, hydratase/3-hydroxyacyl Coenzyme A dehydrogenase|peroxisomal enoyl-CoA hydratase', map_location='3q26.3-q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185190624, end_position_on_the_genomic_accession=185282886, orientation='minus', exon_count=9, OMIM=607037, no_hdr0=''),
  1964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1964, CurrentID=0, Status='live', Symbol='EIF1AX', Aliases='EIF1A, EIF1AP1, EIF4C, eIF-1A, eIF-4C', description='eukaryotic translation initiation factor 1A, X-linked', other_designations='Putative eukaryotic translation initiation factor 1A|eIF-1A X isoform|eukaryotic translation initiation factor 1A, X chromosome|eukaryotic translation initiation factor 4C', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20124518, end_position_on_the_genomic_accession=20141848, orientation='minus', exon_count=7, OMIM=300186, no_hdr0=''),
  1965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1965, CurrentID=0, Status='live', Symbol='EIF2S1', Aliases='EIF-2, EIF-2A, EIF-2alpha, EIF2, EIF2A', description='eukaryotic translation initiation factor 2 subunit alpha', other_designations='eIF-2-alpha|eukaryotic translation initiation factor 2, subunit 1 alpha, 35kDa', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67360317, end_position_on_the_genomic_accession=67386516, orientation='plus', exon_count=8, OMIM=603907, no_hdr0=''),
  1967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1967, CurrentID=0, Status='live', Symbol='EIF2B1', Aliases='EIF2B, EIF2BA', description='eukaryotic translation initiation factor 2B subunit alpha', other_designations='eIF-2B GDP-GTP exchange factor subunit alpha|eukaryotic translation initiation factor 2B, subunit 1 alpha, 26kDa', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123621023, end_position_on_the_genomic_accession=123633776, orientation='minus', exon_count=9, OMIM=606686, no_hdr0=''),
  1968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1968, CurrentID=0, Status='live', Symbol='EIF2S3', Aliases='EIF2, EIF2G, EIF2gamma, eIF-2gA', description='eukaryotic translation initiation factor 2 subunit gamma', other_designations='eIF-2-gamma X|eIF-2gX|eukaryotic translation initiation factor 2 subunit gamma X|eukaryotic translation initiation factor 2, subunit 3 gamma, 52kDa|eukaryotic translation initiation factor 2G', map_location='Xp22.2-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24054948, end_position_on_the_genomic_accession=24078810, orientation='plus', exon_count=12, OMIM=300161, no_hdr0=''),
  1969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1969, CurrentID=0, Status='live', Symbol='EPHA2', Aliases='ARCC2, CTPA, CTPP1, CTRCT6, ECK', description='EPH receptor A2', other_designations='epithelial cell receptor protein tyrosine kinase|soluble EPHA2 variant 1|tyrosine-protein kinase receptor ECK', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16124337, end_position_on_the_genomic_accession=16156087, orientation='minus', exon_count=17, OMIM=176946, no_hdr0=''),
  102723415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723415, CurrentID=0, Status='live', Symbol='LOC102723415', Aliases='', description='uncharacterized LOC102723415', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38594516, end_position_on_the_genomic_accession=38610466, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  1973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1973, CurrentID=0, Status='live', Symbol='EIF4A1', Aliases='DDX2A, EIF-4A, EIF4A, eIF-4A-I, eIF4A-I', description='eukaryotic translation initiation factor 4A1', other_designations='ATP-dependent RNA helicase eIF4A-1|eukaryotic initiation factor 4AI|eukaryotic translation initiation factor 4A', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7572706, end_position_on_the_genomic_accession=7579006, orientation='plus', exon_count=11, OMIM=602641, no_hdr0=''),
  1974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1974, CurrentID=0, Status='live', Symbol='EIF4A2', Aliases='BM-010, DDX2B, EIF4A, EIF4F, eIF-4A-II, eIF4A-II', description='eukaryotic translation initiation factor 4A2', other_designations='ATP-dependent RNA helicase eIF4A-2', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186783572, end_position_on_the_genomic_accession=186789897, orientation='plus', exon_count=11, OMIM=601102, no_hdr0=''),
  1975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1975, CurrentID=0, Status='live', Symbol='EIF4B', Aliases='EIF-4B, PRO1843', description='eukaryotic translation initiation factor 4B', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53006258, end_position_on_the_genomic_accession=53042209, orientation='plus', exon_count=16, OMIM=603928, no_hdr0=''),
  105372679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372679, CurrentID=0, Status='live', Symbol='LOC105372679', Aliases='', description='uncharacterized LOC105372679', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56036767, end_position_on_the_genomic_accession=56040487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  1977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1977, CurrentID=0, Status='live', Symbol='EIF4E', Aliases='AUTS19, CBP1, EIF4EL1, EIF4F, eIF-4E, EIF4E', description='eukaryotic translation initiation factor 4E', other_designations='eIF-4F 25 kDa subunit|eukaryotic translation initiation factor 4E-like 1|mRNA cap-binding protein', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98878456, end_position_on_the_genomic_accession=98930635, orientation='minus', exon_count=10, OMIM=133440, no_hdr0=''),
  1978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1978, CurrentID=0, Status='live', Symbol='EIF4EBP1', Aliases='4E-BP1, 4EBP1, BP-1, PHAS-I', description='eukaryotic translation initiation factor 4E binding protein 1', other_designations='eIF4E-binding protein 1|phosphorylated heat- and acid-stable protein regulated by insulin 1', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38030502, end_position_on_the_genomic_accession=38060365, orientation='plus', exon_count=3, OMIM=602223, no_hdr0=''),
  1979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1979, CurrentID=0, Status='live', Symbol='EIF4EBP2', Aliases='4EBP2, PHASII', description='eukaryotic translation initiation factor 4E binding protein 2', other_designations='4E-BP2|eIF4E-binding protein 2|phosphorylated, heat and acid stable regulated by insulin protein II', map_location='10q21-q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70404105, end_position_on_the_genomic_accession=70428618, orientation='plus', exon_count=3, OMIM=602224, no_hdr0=''),
  1980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1980, CurrentID=0, Status='live', Symbol='EIF4EP1', Aliases='EIF4EL2, dJ1022P6.3', description='eukaryotic translation initiation factor 4E pseudogene 1', other_designations='eukaryotic translation initiation factor 4E-like 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5548852, end_position_on_the_genomic_accession=5549645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1981, CurrentID=0, Status='live', Symbol='EIF4G1', Aliases='EIF-4G1, EIF4F, EIF4G, EIF4GI, P220, PARK18', description='eukaryotic translation initiation factor 4 gamma, 1', other_designations='EIF4-gamma|eIF-4-gamma 1|eucaryotic translation initiation factor 4G', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184314495, end_position_on_the_genomic_accession=184335358, orientation='plus', exon_count=34, OMIM=600495, no_hdr0=''),
  1982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1982, CurrentID=0, Status='live', Symbol='EIF4G2', Aliases='AAG1, DAP5, NAT1, P97', description='eukaryotic translation initiation factor 4 gamma, 2', other_designations='DAP-5|aging-associated protein 1|death-associated protein 5|eIF-4-gamma 2|eIF-4G 2|eIF4G 2|eukaryotic translation initiation factor 4G-like 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10797046, end_position_on_the_genomic_accession=10809035, orientation='minus', exon_count=23, OMIM=602325, no_hdr0=''),
  1983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1983, CurrentID=0, Status='live', Symbol='EIF5', Aliases='EIF-5, EIF-5A', description='eukaryotic translation initiation factor 5', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103334002, end_position_on_the_genomic_accession=103345025, orientation='plus', exon_count=13, OMIM=601710, no_hdr0=''),
  1984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1984, CurrentID=0, Status='live', Symbol='EIF5A', Aliases='EIF-5A1, eIF5AI, EIF5A', description='eukaryotic translation initiation factor 5A', other_designations='eIF-4D|eIF-5A-1|eIF-5A1|eukaryotic initiation factor 5A|rev-binding factor', map_location='17p13-p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7306999, end_position_on_the_genomic_accession=7312463, orientation='plus', exon_count=11, OMIM=600187, no_hdr0=''),
  101928357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928357, CurrentID=0, Status='live', Symbol='LOC101928357', Aliases='', description='uncharacterized LOC101928357', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=126495194, end_position_on_the_genomic_accession=126537276, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  1986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1986, CurrentID=0, Status='live', Symbol='EIF5AP2', Aliases='', description='eukaryotic translation initiation factor 5A pseudogene 2', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78157359, end_position_on_the_genomic_accession=78159135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1987, CurrentID=0, Status='live', Symbol='EIF5AP3', Aliases='', description='eukaryotic translation initiation factor 5A pseudogene 3', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45128525, end_position_on_the_genomic_accession=45129082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  133060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133060, CurrentID=0, Status='live', Symbol='OTOP1', Aliases='', description='otopetrin 1', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4188803, end_position_on_the_genomic_accession=4226894, orientation='minus', exon_count=6, OMIM=607806, no_hdr0=''),
  100510710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100510710, CurrentID=0, Status='live', Symbol='LOC100510710', Aliases='', description='glucosylceramidase-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1990, CurrentID=0, Status='live', Symbol='CELA1', Aliases='ELA1', description='chymotrypsin like elastase family member 1', other_designations='chymotrypsin-like elastase family, member 1|elastase 1, pancreatic|elastase-1|pancreatic elastase 1', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51328443, end_position_on_the_genomic_accession=51346679, orientation='minus', exon_count=8, OMIM=130120, no_hdr0=''),
  1991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1991, CurrentID=0, Status='live', Symbol='ELANE', Aliases='ELA2, GE, HLE, HNE, NE, PMN-E, SCN1', description='elastase, neutrophil expressed', other_designations='PMN elastase|bone marrow serine protease|elastase 2, neutrophil|elastase-2|granulocyte-derived elastase|human leukocyte elastase|leukocyte elastase|medullasin|polymorphonuclear elastase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=850997, end_position_on_the_genomic_accession=856250, orientation='plus', exon_count=6, OMIM=130130, no_hdr0=''),
  1992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1992, CurrentID=0, Status='live', Symbol='SERPINB1', Aliases='EI, ELANH2, HEL-S-27, HEL57, LEI, M/NEI, MNEI, PI-2, PI2', description='serpin peptidase inhibitor, clade B (ovalbumin), member 1', other_designations='epididymis luminal protein 57|epididymis secretory protein Li 27|peptidase inhibitor 2|protease inhibitor 2 (anti-elastase), monocyte/neutrophil derived|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2832332, end_position_on_the_genomic_accession=2842049, orientation='minus', exon_count=9, OMIM=130135, no_hdr0=''),
  1993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1993, CurrentID=0, Status='live', Symbol='ELAVL2', Aliases='HEL-N1, HELN1, HUB', description='ELAV like neuron-specific RNA binding protein 2', other_designations='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 2 (Hu antigen B)|ELAV-like neuronal protein 1 isoform Hel-N2|hu-antigen B|nervous system-specific RNA-binding protein Hel-N1', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23690104, end_position_on_the_genomic_accession=23850706, orientation='minus', exon_count=17, OMIM=601673, no_hdr0=''),
  1994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1994, CurrentID=0, Status='live', Symbol='ELAVL1', Aliases='ELAV1, HUR, Hua, MelG', description='ELAV like RNA binding protein 1', other_designations='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 1 (Hu antigen R)|Hu antigen R|embryonic lethal, abnormal vision, drosophila, homolog-like 1|hu-antigen R', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7958573, end_position_on_the_genomic_accession=8005645, orientation='minus', exon_count=7, OMIM=603466, no_hdr0=''),
  1995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1995, CurrentID=0, Status='live', Symbol='ELAVL3', Aliases='HUC, HUCL, PLE21', description='ELAV like neuron-specific RNA binding protein 3', other_designations='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 3 (Hu antigen C)|Hu antigen C|hu-antigen C|paraneoplastic cerebellar degeneration-associated antigen|paraneoplastic limbic encephalitis antigen 21', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11451326, end_position_on_the_genomic_accession=11481046, orientation='minus', exon_count=7, OMIM=603458, no_hdr0=''),
  1996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1996, CurrentID=0, Status='live', Symbol='ELAVL4', Aliases='HUD, PNEM', description='ELAV like neuron-specific RNA binding protein 4', other_designations='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 4 (Hu antigen D)|Hu antigen D|paraneoplastic encephalomyelitis antigen HuD', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50048014, end_position_on_the_genomic_accession=50203786, orientation='plus', exon_count=15, OMIM=168360, no_hdr0=''),
  1997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1997, CurrentID=0, Status='live', Symbol='ELF1', Aliases='', description='E74-like factor 1 (ets domain transcription factor)', other_designations='', map_location='13q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40931919, end_position_on_the_genomic_accession=41061378, orientation='minus', exon_count=14, OMIM=189973, no_hdr0=''),
  1998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1998, CurrentID=0, Status='live', Symbol='ELF2', Aliases='EU32, NERF, NERF-1A, NERF-1B, NERF-1a,b, NERF-2', description='E74-like factor 2 (ets domain transcription factor)', other_designations='ets family transcription factor ELF2C|new Ets-related factor', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139057717, end_position_on_the_genomic_accession=139177218, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  1999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1999, CurrentID=0, Status='live', Symbol='ELF3', Aliases='EPR-1, ERT, ESE-1, ESX', description='E74-like factor 3 (ets domain transcription factor, epithelial-specific )', other_designations='E74-like factor 3 (ETS domain transcription factor, serine box, epithelial-specific)|E74-like factor 3 (ets domain transcription factor)|epithelial-restricted with serine box|epithelium-restricted Ets protein ESX|epithelium-specific Ets transcription factor 1|ets domain transcription factor, serine box (epithelial-specific)', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202010519, end_position_on_the_genomic_accession=202017188, orientation='plus', exon_count=9, OMIM=602191, no_hdr0=''),
  2000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2000, CurrentID=0, Status='live', Symbol='ELF4', Aliases='ELFR, MEF', description='E74-like factor 4 (ets domain transcription factor)', other_designations='myeloid Elf-1-like factor', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130064920, end_position_on_the_genomic_accession=130110713, orientation='minus', exon_count=11, OMIM=300775, no_hdr0=''),
  2001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2001, CurrentID=0, Status='live', Symbol='ELF5', Aliases='ESE2', description='E74-like factor 5 (ets domain transcription factor)', other_designations='epithelium-restricted ESE-1-related Ets factor|epithelium-specific Ets transcription factor 2', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34478795, end_position_on_the_genomic_accession=34513800, orientation='minus', exon_count=8, OMIM=605169, no_hdr0=''),
  2002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2002, CurrentID=0, Status='live', Symbol='ELK1', Aliases='', description='ELK1, member of ETS oncogene family', other_designations='ETS-like gene 1|tyrosine kinase (ELK1) oncogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47635520, end_position_on_the_genomic_accession=47650604, orientation='minus', exon_count=7, OMIM=311040, no_hdr0=''),
  2003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2003, CurrentID=0, Status='live', Symbol='ELK2AP', Aliases='ELK2, ELK2.1, ELK2P1', description='ELK2A, member of ETS oncogene family, pseudogene', other_designations='ELK1, member of ETS oncogene family pseudogene|ELK2, member of ETS oncogene family, pseudogene 1', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105669577, end_position_on_the_genomic_accession=105672807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2004, CurrentID=0, Status='live', Symbol='ELK3', Aliases='ERP, NET, SAP-2, SAP2', description='ELK3, ETS-domain protein (SRF accessory protein 2)', other_designations='ETS-related protein ERP|ETS-related protein NET|SRF accessory protein 2|serum response factor accessory protein 2', map_location='12q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96194382, end_position_on_the_genomic_accession=96269835, orientation='plus', exon_count=7, OMIM=600247, no_hdr0=''),
  2005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2005, CurrentID=0, Status='live', Symbol='ELK4', Aliases='SAP1', description='ELK4, ETS-domain protein (SRF accessory protein 1)', other_designations='SAP-1|serum response factor accessory protein 1', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205593466, end_position_on_the_genomic_accession=205634085, orientation='minus', exon_count=12, OMIM=600246, no_hdr0=''),
  2006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2006, CurrentID=0, Status='live', Symbol='ELN', Aliases='SVAS, WBS, WS', description='elastin', other_designations='tropoelastin', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74027772, end_position_on_the_genomic_accession=74069907, orientation='plus', exon_count=34, OMIM=130160, no_hdr0=''),
  2009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2009, CurrentID=0, Status='live', Symbol='EML1', Aliases='ELP79, EMAP, EMAPL, HuEMAP', description='echinoderm microtubule associated protein like 1', other_designations='EMAP-1|huEMAP-1', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99737722, end_position_on_the_genomic_accession=99942060, orientation='plus', exon_count=27, OMIM=602033, no_hdr0=''),
  2010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2010, CurrentID=0, Status='live', Symbol='EMD', Aliases='EDMD, LEMD5, STA', description='emerin', other_designations='LEM domain containing 5', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154379237, end_position_on_the_genomic_accession=154381523, orientation='plus', exon_count=6, OMIM=300384, no_hdr0=''),
  2011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2011, CurrentID=0, Status='live', Symbol='MARK2', Aliases='EMK-1, EMK1, PAR-1, Par-1b, Par1b', description='MAP/microtubule affinity-regulating kinase 2', other_designations='ELKL motif kinase 1|PAR1 homolog b|Ser/Thr protein kinase PAR-1B|serine/threonine protein kinase EMK', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63838928, end_position_on_the_genomic_accession=63911020, orientation='plus', exon_count=21, OMIM=600526, no_hdr0=''),
  2012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2012, CurrentID=0, Status='live', Symbol='EMP1', Aliases='CL-20, EMP-1, TMP', description='epithelial membrane protein 1', other_designations='tumor-associated membrane protein', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13196668, end_position_on_the_genomic_accession=13216774, orientation='plus', exon_count=5, OMIM=602333, no_hdr0=''),
  2013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2013, CurrentID=0, Status='live', Symbol='EMP2', Aliases='XMP', description='epithelial membrane protein 2', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10528422, end_position_on_the_genomic_accession=10580682, orientation='minus', exon_count=6, OMIM=602334, no_hdr0=''),
  2014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2014, CurrentID=0, Status='live', Symbol='EMP3', Aliases='YMP', description='epithelial membrane protein 3', other_designations='EMP-3|HNMP-1|hematopoietic neural membrane protein 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48325370, end_position_on_the_genomic_accession=48330555, orientation='plus', exon_count=6, OMIM=602335, no_hdr0=''),
  2015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2015, CurrentID=0, Status='live', Symbol='ADGRE1', Aliases='EMR1, TM7LN3', description='adhesion G protein-coupled receptor E1', other_designations='EGF-like module receptor 1|EGF-like module-containing mucin-like hormone receptor-like 1|EMR1 hormone receptor|egf-like module containing, mucin-like, hormone receptor-like 1|egf-like module containing, mucin-like, hormone receptor-like sequence 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6887549, end_position_on_the_genomic_accession=6940453, orientation='plus', exon_count=23, OMIM=600493, no_hdr0=''),
  2016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2016, CurrentID=0, Status='live', Symbol='EMX1', Aliases='', description='empty spiracles homeobox 1', other_designations='empty spiracles homolog 1|empty spiracles-like protein 1', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72911148, end_position_on_the_genomic_accession=72934891, orientation='plus', exon_count=7, OMIM=600034, no_hdr0=''),
  2017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2017, CurrentID=0, Status='live', Symbol='CTTN', Aliases='EMS1', description='cortactin', other_designations='1110020L01Rik|amplaxin|ems1 sequence (mammary tumor and squamous cell carcinoma-associated (p80/85 src substrate)|oncogene EMS1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70398506, end_position_on_the_genomic_accession=70436584, orientation='plus', exon_count=19, OMIM=164765, no_hdr0=''),
  2018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2018, CurrentID=0, Status='live', Symbol='EMX2', Aliases='', description='empty spiracles homeobox 2', other_designations='empty spiracles homolog 2|empty spiracles-like protein 2', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117542445, end_position_on_the_genomic_accession=117549546, orientation='plus', exon_count=3, OMIM=600035, no_hdr0=''),
  2019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2019, CurrentID=0, Status='live', Symbol='EN1', Aliases='', description='engrailed homeobox 1', other_designations='engrailed homolog 1|homeobox protein en-1|hu-En-1', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118842171, end_position_on_the_genomic_accession=118848183, orientation='minus', exon_count=2, OMIM=131290, no_hdr0=''),
  2020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2020, CurrentID=0, Status='live', Symbol='EN2', Aliases='', description='engrailed homeobox 2', other_designations='engrailed homolog 2|engrailed-2|homeobox protein en-2|hu-En-2', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155458129, end_position_on_the_genomic_accession=155464831, orientation='plus', exon_count=2, OMIM=131310, no_hdr0=''),
  2021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2021, CurrentID=0, Status='live', Symbol='ENDOG', Aliases='', description='endonuclease G', other_designations='endo G|mitochondrial endonuclease G', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128818411, end_position_on_the_genomic_accession=128822678, orientation='plus', exon_count=4, OMIM=600440, no_hdr0=''),
  2022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2022, CurrentID=0, Status='live', Symbol='ENG', Aliases='END, HHT1, ORW1', description='endoglin', other_designations='CD105 antigen', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127815012, end_position_on_the_genomic_accession=127854773, orientation='minus', exon_count=16, OMIM=131195, no_hdr0=''),
  2023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2023, CurrentID=0, Status='live', Symbol='ENO1', Aliases='ENO1L1, HEL-S-17, MPB1, NNE, PPH', description='enolase 1, (alpha)', other_designations='2-phospho-D-glycerate hydro-lyase|MYC promoter-binding protein 1|alpha enolase like 1|enolase-alpha|epididymis secretory protein Li 17|non-neural enolase|phosphopyruvate hydratase|plasminogen-binding protein|tau-crystallin', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8861000, end_position_on_the_genomic_accession=8879092, orientation='minus', exon_count=13, OMIM=172430, no_hdr0=''),
  2025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2025, CurrentID=0, Status='live', Symbol='ENO1P1', Aliases='ENO1P', description='enolase 1, (alpha) pseudogene 1', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236483098, end_position_on_the_genomic_accession=236484905, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2026, CurrentID=0, Status='live', Symbol='ENO2', Aliases='HEL-S-279, NSE', description='enolase 2 (gamma, neuronal)', other_designations='2-phospho-D-glycerate hydro-lyase|2-phospho-D-glycerate hydrolyase|epididymis secretory protein Li 279|neural enolase|neuron specific gamma enolase|neuron-specific enolase|neurone-specific enolase', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6914450, end_position_on_the_genomic_accession=6923696, orientation='plus', exon_count=12, OMIM=131360, no_hdr0=''),
  2027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2027, CurrentID=0, Status='live', Symbol='ENO3', Aliases='GSD13, MSE', description='enolase 3 (beta, muscle)', other_designations='2-phospho-D-glycerate hydrolyase|muscle-specific enolase|skeletal muscle enolase', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4949182, end_position_on_the_genomic_accession=4957131, orientation='plus', exon_count=14, OMIM=131370, no_hdr0=''),
  2028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2028, CurrentID=0, Status='live', Symbol='ENPEP', Aliases='APA, CD249, gp160', description='glutamyl aminopeptidase (aminopeptidase A)', other_designations='AP-A|EAP|aminopeptidase A|differentiation antigen gp160', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110476073, end_position_on_the_genomic_accession=110563337, orientation='plus', exon_count=20, OMIM=138297, no_hdr0=''),
  2029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2029, CurrentID=0, Status='live', Symbol='ENSA', Aliases='ARPP-19e', description='endosulfine alpha', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150622123, end_position_on_the_genomic_accession=150629622, orientation='minus', exon_count=6, OMIM=603061, no_hdr0=''),
  2030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2030, CurrentID=0, Status='live', Symbol='SLC29A1', Aliases='ENT1', description='solute carrier family 29 (equilibrative nucleoside transporter), member 1', other_designations='equilibrative nitrobenzylmercaptopurine riboside (NBMPR)-sensitive nucleoside transporter|equilibrative nucleoside transporter 1 variant ata|nucleoside transporter, es-type|solute carrier family 29 (nucleoside transporters), member 1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44219505, end_position_on_the_genomic_accession=44234151, orientation='plus', exon_count=18, OMIM=602193, no_hdr0=''),
  2031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2031, CurrentID=0, Status='live', Symbol='ENUR1', Aliases='', description='enuresis, nocturnal 1', other_designations='', map_location='13q13-q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600631, no_hdr0=''),
  2032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2032, CurrentID=0, Status='live', Symbol='ENUR2', Aliases='', description='enuresis, nocturnal 2', other_designations='', map_location='12q13-q21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600808, no_hdr0=''),
  2033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2033, CurrentID=0, Status='live', Symbol='EP300', Aliases='KAT3B, RSTS2, p300', description='E1A binding protein p300', other_designations='E1A-associated protein p300|E1A-binding protein, 300kD|p300 HAT', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41092610, end_position_on_the_genomic_accession=41180077, orientation='plus', exon_count=31, OMIM=602700, no_hdr0=''),
  2034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2034, CurrentID=0, Status='live', Symbol='EPAS1', Aliases='ECYT4, HIF2A, HLF, MOP2, PASD2, bHLHe73', description='endothelial PAS domain protein 1', other_designations='EPAS-1|HIF-1-alpha-like factor|HIF-1alpha-like factor|HIF-2-alpha|HIF2-alpha|PAS domain-containing protein 2|basic-helix-loop-helix-PAS protein MOP2|class E basic helix-loop-helix protein 73|hypoxia-inducible factor 2 alpha|hypoxia-inducible factor 2-alpha|member of PAS protein 2', map_location='2p21-p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46297402, end_position_on_the_genomic_accession=46386703, orientation='plus', exon_count=17, OMIM=603349, no_hdr0=''),
  2035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2035, CurrentID=0, Status='live', Symbol='EPB41', Aliases='4.1R, EL1, HE', description='erythrocyte membrane protein band 4.1', other_designations='EPB4.1|P4.1|band 4.1|elliptocytosis 1, RH-linked|erythrocyte surface protein band 4.1', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28887091, end_position_on_the_genomic_accession=29120046, orientation='plus', exon_count=29, OMIM=130500, no_hdr0=''),
  2036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2036, CurrentID=0, Status='live', Symbol='EPB41L1', Aliases='4.1N, MRD11', description='erythrocyte membrane protein band 4.1-like 1', other_designations='neuron-type nonerythroid protein 4.1|neuronal protein 4.1', map_location='20q11.2-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36064891, end_position_on_the_genomic_accession=36232799, orientation='plus', exon_count=37, OMIM=602879, no_hdr0=''),
  2037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2037, CurrentID=0, Status='live', Symbol='EPB41L2', Aliases='4.1-G, 4.1G', description='erythrocyte membrane protein band 4.1-like 2', other_designations='erythrocyte membrane protein band 4.1 like-protein 2|generally expressed protein 4.1', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130828173, end_position_on_the_genomic_accession=131063322, orientation='minus', exon_count=30, OMIM=603237, no_hdr0=''),
  2038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2038, CurrentID=0, Status='live', Symbol='EPB42', Aliases='PA, SPH5', description='erythrocyte membrane protein band 4.2', other_designations='P4.2|erythrocyte protein 4.2|erythrocyte surface protein band 4.2', map_location='15q15-q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43197227, end_position_on_the_genomic_accession=43225861, orientation='minus', exon_count=16, OMIM=177070, no_hdr0=''),
  2039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2039, CurrentID=0, Status='live', Symbol='DMTN', Aliases='DMT, EPB49', description='dematin actin binding protein', other_designations='erythrocyte membrane protein band 4.9 (dematin)', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22049149, end_position_on_the_genomic_accession=22082527, orientation='plus', exon_count=21, OMIM=125305, no_hdr0=''),
  2040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2040, CurrentID=0, Status='live', Symbol='STOM', Aliases='BND7, EPB7, EPB72', description='stomatin', other_designations='erythrocyte membrane protein band 7.2 (stomatin)|erythrocyte surface protein band 7.2|protein 7.2b', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121338988, end_position_on_the_genomic_accession=121370304, orientation='minus', exon_count=8, OMIM=133090, no_hdr0=''),
  2041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2041, CurrentID=0, Status='live', Symbol='EPHA1', Aliases='EPH, EPHT, EPHT1', description='EPH receptor A1', other_designations='eph tyrosine kinase 1|erythropoietin-producing hepatoma amplified sequence|erythropoietin-producing hepatoma receptor|hEpha1|oncogene EPH|soluble EPHA1 variant 1|soluble EPHA1 variant 2|tyrosine-protein kinase receptor EPH', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143390782, end_position_on_the_genomic_accession=143408892, orientation='minus', exon_count=20, OMIM=179610, no_hdr0=''),
  2042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2042, CurrentID=0, Status='live', Symbol='EPHA3', Aliases='EK4, ETK, ETK1, HEK, HEK4, TYRO4', description='EPH receptor A3', other_designations='EPH-like kinase 4|TYRO4 protein tyrosine kinase|eph-like tyrosine kinase 1|human embryo kinase 1|tyrosine-protein kinase receptor ETK1', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=89107524, end_position_on_the_genomic_accession=89482134, orientation='plus', exon_count=17, OMIM=179611, no_hdr0=''),
  2043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2043, CurrentID=0, Status='live', Symbol='EPHA4', Aliases='HEK8, SEK, TYRO1', description='EPH receptor A4', other_designations='EK8|EPH-like kinase 8|TYRO1 protein tyrosine kinase|receptor protein-tyrosine kinase HEK8|tyrosine-protein kinase TYRO1|tyrosine-protein kinase receptor SEK', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221418027, end_position_on_the_genomic_accession=221574202, orientation='minus', exon_count=19, OMIM=602188, no_hdr0=''),
  2044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2044, CurrentID=0, Status='live', Symbol='EPHA5', Aliases='CEK7, EHK-1, EHK1, EK7, HEK7, TYRO4', description='EPH receptor A5', other_designations='EPH homology kinase 1|EPH-like kinase 7|brain-specific kinase|receptor protein-tyrosine kinase HEK7|tyrosine-protein kinase receptor EHK-1', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65319563, end_position_on_the_genomic_accession=65670495, orientation='minus', exon_count=21, OMIM=600004, no_hdr0=''),
  2045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2045, CurrentID=0, Status='live', Symbol='EPHA7', Aliases='EHK-3, EHK3, EK11, HEK11', description='EPH receptor A7', other_designations='EPH homology kinase 3|EPH-like kinase 11|Eph homology kinase-3|receptor protein-tyrosine kinase HEK11|tyrosine-protein kinase receptor EHK-3', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93240020, end_position_on_the_genomic_accession=93419584, orientation='minus', exon_count=18, OMIM=602190, no_hdr0=''),
  2046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2046, CurrentID=0, Status='live', Symbol='EPHA8', Aliases='EEK, EK3, HEK3', description='EPH receptor A8', other_designations='EPH- and ELK-related kinase|EPH- and ELK-related tyrosine kinase|EPH-like kinase 3|hydroxyaryl-protein kinase|protein-tyrosine kinase|tyrosine-protein kinase receptor EEK|tyrosylprotein kinase', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22563508, end_position_on_the_genomic_accession=22603594, orientation='plus', exon_count=26, OMIM=176945, no_hdr0=''),
  2047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2047, CurrentID=0, Status='live', Symbol='EPHB1', Aliases='ELK, EPHT2, Hek6, NET', description='EPH receptor B1', other_designations='EK6|EPH-like kinase 6|eph tyrosine kinase 2|neuronally-expressed EPH-related tyrosine kinase|soluble EPHB1 variant 1|tyrosine-protein kinase receptor EPH-2', map_location='3q21-q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134795257, end_position_on_the_genomic_accession=135260467, orientation='plus', exon_count=21, OMIM=600600, no_hdr0=''),
  2048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2048, CurrentID=0, Status='live', Symbol='EPHB2', Aliases='CAPB, DRT, EK5, EPHT3, ERK, Hek5, PCBC, Tyro5', description='EPH receptor B2', other_designations='EPH-like kinase 5|developmentally-regulated Eph-related tyrosine kinase|elk-related tyrosine kinase|eph tyrosine kinase 3|protein-tyrosine kinase HEK5|renal carcinoma antigen NY-REN-47|tyrosine-protein kinase TYRO5|tyrosine-protein kinase receptor EPH-3', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22710770, end_position_on_the_genomic_accession=22915330, orientation='plus', exon_count=17, OMIM=600997, no_hdr0=''),
  2049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2049, CurrentID=0, Status='live', Symbol='EPHB3', Aliases='ETK2, HEK2, TYRO6', description='EPH receptor B3', other_designations='EK2|EPH-like kinase 2|EPH-like tyrosine kinase 2|EPH-like tyrosine kinase-2|embryonic kinase 2|human embryo kinase 2|tyrosine-protein kinase TYRO6', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184561799, end_position_on_the_genomic_accession=184582408, orientation='plus', exon_count=16, OMIM=601839, no_hdr0=''),
  2050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2050, CurrentID=0, Status='live', Symbol='EPHB4', Aliases='HTK, MYK1, TYRO11', description='EPH receptor B4', other_designations='ephrin receptor EphB4|hepatoma transmembrane kinase|soluble EPHB4 variant 1|soluble EPHB4 variant 2|soluble EPHB4 variant 3|tyrosine-protein kinase TYRO11|tyrosine-protein kinase receptor HTK', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100802565, end_position_on_the_genomic_accession=100827521, orientation='minus', exon_count=17, OMIM=600011, no_hdr0=''),
  2051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2051, CurrentID=0, Status='live', Symbol='EPHB6', Aliases='HEP', description='EPH receptor B6', other_designations='human kinase-defective Eph-family receptor protein|tyrosine-protein kinase-defective receptor EPH-6', map_location='7q33-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142854643, end_position_on_the_genomic_accession=142871093, orientation='plus', exon_count=21, OMIM=602757, no_hdr0=''),
  2052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2052, CurrentID=0, Status='live', Symbol='EPHX1', Aliases='EPHX, EPOX, HYL1, MEH', description='epoxide hydrolase 1, microsomal (xenobiotic)', other_designations='epoxide hydratase|epoxide hydrolase 1 microsomal', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225810074, end_position_on_the_genomic_accession=225845563, orientation='plus', exon_count=10, OMIM=132810, no_hdr0=''),
  2053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2053, CurrentID=0, Status='live', Symbol='EPHX2', Aliases='CEH, SEH', description='epoxide hydrolase 2, cytoplasmic', other_designations='epoxide hydratase|epoxide hydrolase 2, cytosolic|epoxide hydrolase, soluble', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27491002, end_position_on_the_genomic_accession=27544922, orientation='plus', exon_count=19, OMIM=132811, no_hdr0=''),
  2054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2054, CurrentID=0, Status='live', Symbol='STX2', Aliases='EPIM, EPMA, STX2B, STX2C, STX2', description='syntaxin 2', other_designations='epimorphin', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130774883, end_position_on_the_genomic_accession=130839274, orientation='minus', exon_count=14, OMIM=132350, no_hdr0=''),
  2055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2055, CurrentID=0, Status='live', Symbol='CLN8', Aliases='C8orf61, EPMR', description='ceroid-lipofuscinosis, neuronal 8', other_designations='ceroid-lipofuscinosis, neuronal 8 (epilepsy, progressive with mental retardation)', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1755778, end_position_on_the_genomic_accession=1786570, orientation='plus', exon_count=6, OMIM=607837, no_hdr0=''),
  2056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2056, CurrentID=0, Status='live', Symbol='EPO', Aliases='EP, MVCD2', description='erythropoietin', other_designations='epoetin', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100720800, end_position_on_the_genomic_accession=100723700, orientation='plus', exon_count=5, OMIM=133170, no_hdr0=''),
  2057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2057, CurrentID=0, Status='live', Symbol='EPOR', Aliases='EPO-R', description='erythropoietin receptor', other_designations='', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11377205, end_position_on_the_genomic_accession=11384342, orientation='minus', exon_count=8, OMIM=133171, no_hdr0=''),
  2058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2058, CurrentID=0, Status='live', Symbol='EPRS', Aliases='EARS, GLUPRORS, PARS, PIG32, QARS, QPRS', description='glutamyl-prolyl-tRNA synthetase', other_designations='bifunctional aminoacyl-tRNA synthetase|cell proliferation-inducing gene 32 protein|glutamate tRNA ligase|glutamatyl-prolyl-tRNA synthetase|glutaminyl-tRNA synthetase|proliferation-inducing gene 32 protein|proliferation-inducing protein 32|proline tRNA ligase|proline-tRNA ligase|prolyl-tRNA synthetase', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219968600, end_position_on_the_genomic_accession=220046658, orientation='minus', exon_count=32, OMIM=138295, no_hdr0=''),
  2059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2059, CurrentID=0, Status='live', Symbol='EPS8', Aliases='DFNB102', description='epidermal growth factor receptor pathway substrate 8', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15620141, end_position_on_the_genomic_accession=15789576, orientation='minus', exon_count=25, OMIM=600206, no_hdr0=''),
  2060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2060, CurrentID=0, Status='live', Symbol='EPS15', Aliases='AF-1P, AF1P, MLLT5', description='epidermal growth factor receptor pathway substrate 15', other_designations='ALL1 fused gene from chromosome 1|protein AF-1p', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51354263, end_position_on_the_genomic_accession=51519354, orientation='minus', exon_count=27, OMIM=600051, no_hdr0=''),
  2063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2063, CurrentID=0, Status='live', Symbol='NR2F6', Aliases='EAR-2, EAR2, ERBAL2', description='nuclear receptor subfamily 2 group F member 6', other_designations='ERBA-related gene-2|V-erbA-related protein 2|nuclear receptor V-erbA-related|nuclear receptor subfamily 2, group F, member 6|v-erb-a avian erythroblastic leukemia viral oncogene homolog-like 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17231885, end_position_on_the_genomic_accession=17245342, orientation='minus', exon_count=4, OMIM=132880, no_hdr0=''),
  2064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2064, CurrentID=0, Status='live', Symbol='ERBB2', Aliases='CD340, HER-2, HER-2/neu, HER2, MLN 19, NEU, NGL, TKR1', description='erb-b2 receptor tyrosine kinase 2', other_designations='c-erb B2/neu protein|herstatin|human epidermal growth factor receptor 2|metastatic lymph node gene 19 protein|neuro/glioblastoma derived oncogene homolog|neuroblastoma/glioblastoma derived oncogene homolog|p185erbB2|proto-oncogene Neu|proto-oncogene c-ErbB-2|tyrosine kinase-type cell surface receptor HER2|v-erb-b2 avian erythroblastic leukemia viral oncogene homolog 2|v-erb-b2 avian erythroblastic leukemia viral oncoprotein 2|v-erb-b2 erythroblastic leukemia viral oncogene homolog 2, neuro/glioblastoma derived oncogene homolog', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39688084, end_position_on_the_genomic_accession=39728662, orientation='plus', exon_count=32, OMIM=164870, no_hdr0=''),
  2065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2065, CurrentID=0, Status='live', Symbol='ERBB3', Aliases='ErbB-3, HER3, LCCS2, MDA-BF-1, c-erbB-3, c-erbB3, erbB3-S, p180-ErbB3, p45-sErbB3, p85-sErbB3', description='erb-b2 receptor tyrosine kinase 3', other_designations='human epidermal growth factor receptor 3|proto-oncogene-like protein c-ErbB-3|tyrosine kinase-type cell surface receptor HER3|v-erb-b2 avian erythroblastic leukemia viral oncogene homolog 3', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56080025, end_position_on_the_genomic_accession=56103507, orientation='plus', exon_count=28, OMIM=190151, no_hdr0=''),
  2066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2066, CurrentID=0, Status='live', Symbol='ERBB4', Aliases='ALS19, HER4, p180erbB4', description='erb-b2 receptor tyrosine kinase 4', other_designations='ERBB4 transcript variant I12DEL|ERBB4 transcript variant I20DEL|avian erythroblastic leukemia viral (v-erb-b2) oncogene homolog 4|human epidermal growth factor receptor 4|proto-oncogene-like protein c-ErbB-4|tyrosine kinase-type cell surface receptor HER4|v-erb-a erythroblastic leukemia viral oncogene homolog 4|v-erb-b2 avian erythroblastic leukemia viral oncogene homolog 4', map_location='2q33.3-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=211375717, end_position_on_the_genomic_accession=212539155, orientation='minus', exon_count=29, OMIM=600543, no_hdr0=''),
  2067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2067, CurrentID=0, Status='live', Symbol='ERCC1', Aliases='COFS4, RAD10, UV20', description='excision repair cross-complementation group 1', other_designations='excision repair cross-complementing rodent repair deficiency, complementation group 1 (includes overlapping antisense sequence)', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45407333, end_position_on_the_genomic_accession=45478820, orientation='minus', exon_count=12, OMIM=126380, no_hdr0=''),
  2068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2068, CurrentID=0, Status='live', Symbol='ERCC2', Aliases='COFS2, EM9, TFIIH, TTD, TTD1, XPD', description='excision repair cross-complementation group 2', other_designations='BTF2 p80|CXPD|DNA excision repair protein ERCC-2|DNA repair protein complementing XP-D cells|TFIIH 80 kDa subunit|TFIIH basal transcription factor complex 80 kDa subunit|TFIIH basal transcription factor complex helicase XPB subunit|TFIIH basal transcription factor complex helicase subunit|TFIIH p80|basic transcription factor 2 80 kDa subunit|excision repair cross-complementing rodent repair deficiency, complementation group 2|xeroderma pigmentosum complementary group D|xeroderma pigmentosum group D-complementing protein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45351391, end_position_on_the_genomic_accession=45370646, orientation='minus', exon_count=24, OMIM=126340, no_hdr0=''),
  2069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2069, CurrentID=0, Status='live', Symbol='EREG', Aliases='EPR, ER, Ep', description='epiregulin', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74365143, end_position_on_the_genomic_accession=74388760, orientation='plus', exon_count=5, OMIM=602061, no_hdr0=''),
  2070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2070, CurrentID=0, Status='live', Symbol='EYA4', Aliases='CMD1J, DFNA10', description='EYA transcriptional coactivator and phosphatase 4', other_designations='dJ78N10.1 (eyes absent)|eyes absent-like protein 4', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133241335, end_position_on_the_genomic_accession=133532120, orientation='plus', exon_count=22, OMIM=603550, no_hdr0=''),
  2071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2071, CurrentID=0, Status='live', Symbol='ERCC3', Aliases='BTF2, GTF2H, RAD25, TFIIH, TTD2, XPB', description='excision repair cross-complementation group 3', other_designations='BTF2 p89|DNA excision repair protein ERCC-3|DNA repair helicase|DNA repair protein complementing XP-B cells|TFIIH 89 kDa subunit|TFIIH basal transcription factor complex 89 kDa subunit|TFIIH p89|basic transcription factor 2 89 kDa subunit|excision repair cross-complementing rodent repair deficiency, complementation group 3|xeroderma pigmentosum group B-complementing protein|xeroderma pigmentosum, complementation group B', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127257290, end_position_on_the_genomic_accession=127294176, orientation='minus', exon_count=15, OMIM=133510, no_hdr0=''),
  2072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2072, CurrentID=0, Status='live', Symbol='ERCC4', Aliases='ERCC11, FANCQ, RAD1, XFEPS, XPF', description='excision repair cross-complementation group 4', other_designations='DNA excision repair protein ERCC-4|DNA repair protein complementing XP-F cells|excision repair cross-complementing rodent repair deficiency, complementation group 4|excision-repair, complementing defective, in Chinese hamster|xeroderma pigmentosum group F-complementing protein|xeroderma pigmentosum, complementation group F', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=13920145, end_position_on_the_genomic_accession=13952348, orientation='plus', exon_count=14, OMIM=133520, no_hdr0=''),
  2073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2073, CurrentID=0, Status='live', Symbol='ERCC5', Aliases='COFS3-201, ERCM2, UVDR, XPG, XPGC, ERCC5', description='excision repair cross-complementation group 5', other_designations='DNA excision repair protein ERCC-5|XPG-complementing protein|excision repair cross-complementing rodent repair deficiency, complementation group 5|xeroderma pigmentosum, complementation group G', map_location='13q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102845841, end_position_on_the_genomic_accession=102876001, orientation='plus', exon_count=15, OMIM=133530, no_hdr0=''),
  2074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2074, CurrentID=0, Status='live', Symbol='ERCC6', Aliases='ARMD5, CKN2, COFS, COFS1, CSB, RAD26, UVSS1', description='excision repair cross-complementation group 6', other_designations='ATP-dependent helicase ERCC6|Cockayne syndrome group B protein|cockayne syndrome protein CSB|excision repair cross-complementing rodent repair deficiency, complementation group 6', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49454480, end_position_on_the_genomic_accession=49539123, orientation='minus', exon_count=21, OMIM=609413, no_hdr0=''),
  2076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2076, CurrentID=0, Status='live', Symbol='ERCM1', Aliases='', description='excision repair complementing defective repair in mouse cells', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=133535, no_hdr0=''),
  2077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2077, CurrentID=0, Status='live', Symbol='ERF', Aliases='CRS4, PE-2, PE2', description='Ets2 repressor factor', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42247561, end_position_on_the_genomic_accession=42255164, orientation='minus', exon_count=8, OMIM=611888, no_hdr0=''),
  133150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133150, CurrentID=0, Status='live', Symbol='SULT1D1P', Aliases='SULT1D1', description='sulfotransferase family 1D member 1, pseudogene', other_designations='SULT1D pseudogene|amine N-sulfotransferase|sulfotransferase family, cytosolic, 1C, member 3 pseudogene|sulfotransferase family, cytosolic, 1D, member 1, pseudogene|tyrosine-ester sulfotransferase', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69791873, end_position_on_the_genomic_accession=69813872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2079, CurrentID=0, Status='live', Symbol='ERH', Aliases='DROER', description='enhancer of rudimentary homolog (Drosophila)', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69380123, end_position_on_the_genomic_accession=69398304, orientation='minus', exon_count=4, OMIM=601191, no_hdr0=''),
  2081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2081, CurrentID=0, Status='live', Symbol='ERN1', Aliases='IRE1, IRE1P, IRE1a, hIRE1p', description='endoplasmic reticulum to nucleus signaling 1', other_designations='ER to nucleus signalling 1|endoplasmic reticulum-to-nucleus signaling 1|inositol-requiring 1|inositol-requiring enzyme 1|inositol-requiring protein 1|ire1-alpha|protein kinase/endoribonuclease', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64043030, end_position_on_the_genomic_accession=64130142, orientation='minus', exon_count=22, OMIM=604033, no_hdr0=''),
  2086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2086, CurrentID=0, Status='live', Symbol='ERV3-1', Aliases='ERV-R, ERV3, ERVR, HERV-R, HERVR, envR', description='endogenous retrovirus group 3 member 1', other_designations='ERV-3 envelope protein|ERV-R envelope protein|ERV3-1 envelope protein|HERV-R envelope protein|HERV-R_7q21.2 provirus ancestral Env polyprotein|endogenous retroviral sequence 3|endogenous retrovirus group 3, member 1', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64990355, end_position_on_the_genomic_accession=65006746, orientation='minus', exon_count=2, OMIM=131170, no_hdr0=''),
  100272169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272169, CurrentID=0, Status='live', Symbol='L1RE4', Aliases='', description='LINE1 retrotransposable element 4', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2090, CurrentID=0, Status='live', Symbol='ESA4', Aliases='', description='esterase A4', other_designations='', map_location='11q13-q22', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=133220, no_hdr0=''),
  2091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2091, CurrentID=0, Status='live', Symbol='FBL', Aliases='FIB, FLRN, RNU3IP1', description='fibrillarin', other_designations='34 kDa nucleolar scleroderma antigen|34-kD nucleolar scleroderma antigen|RNA, U3 small nucleolar interacting protein 1|histone-glutamine methyltransferase', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39834453, end_position_on_the_genomic_accession=39846414, orientation='minus', exon_count=9, OMIM=134795, no_hdr0=''),
  2096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2096, CurrentID=0, Status='live', Symbol='ESAT', Aliases='', description='esterase activator', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=133250, no_hdr0=''),
  2097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2097, CurrentID=0, Status='live', Symbol='ESB3', Aliases='', description='esterase B3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=133290, no_hdr0=''),
  2098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2098, CurrentID=0, Status='live', Symbol='ESD', Aliases='FGH', description='esterase D', other_designations='esterase 10|esterase D/formylglutathione hydrolase|methylumbelliferyl-acetate deacetylase', map_location='13q14.1-q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46771256, end_position_on_the_genomic_accession=46797681, orientation='minus', exon_count=14, OMIM=133280, no_hdr0=''),
  2099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2099, CurrentID=0, Status='live', Symbol='ESR1', Aliases='ER, ESR, ESRA, ESTRR, Era, NR3A1', description='estrogen receptor 1', other_designations='ER-alpha|estradiol receptor|estrogen nuclear receptor alpha|estrogen receptor alpha E1-E2-1-2|estrogen receptor alpha E1-N2-E2-1-2|nuclear receptor subfamily 3 group A member 1', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151690496, end_position_on_the_genomic_accession=152103274, orientation='plus', exon_count=18, OMIM=133430, no_hdr0=''),
  2100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2100, CurrentID=0, Status='live', Symbol='ESR2', Aliases='ER-BETA, ESR-BETA, ESRB, ESTRB, Erb, NR3A2', description='estrogen receptor 2 (ER beta)', other_designations='estrogen receptor beta 4|nuclear receptor subfamily 3 group A member 2', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64172925, end_position_on_the_genomic_accession=64338550, orientation='minus', exon_count=20, OMIM=601663, no_hdr0=''),
  2101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2101, CurrentID=0, Status='live', Symbol='ESRRA', Aliases='ERR1, ERRa, ERRalpha, ESRL1, NR3B1', description='estrogen-related receptor alpha', other_designations='ERR-alpha|estrogen receptor-like 1|estrogen-related nuclear receptor alpha|nuclear receptor subfamily 3 group B member 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64305528, end_position_on_the_genomic_accession=64316743, orientation='plus', exon_count=8, OMIM=601998, no_hdr0=''),
  2103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2103, CurrentID=0, Status='live', Symbol='ESRRB', Aliases='DFNB35, ERR2, ERRb, ESRL2, NR3B2', description='estrogen-related receptor beta', other_designations='ERR beta-2|ERR-beta|ERRbeta-2|estrogen receptor-like 2|estrogen-related nuclear receptor beta|nuclear receptor ERRB2|nuclear receptor subfamily 3 group B member 2|orphan nuclear receptor', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76371299, end_position_on_the_genomic_accession=76501839, orientation='plus', exon_count=17, OMIM=602167, no_hdr0=''),
  2104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2104, CurrentID=0, Status='live', Symbol='ESRRG', Aliases='ERR3, ERRgamma, NR3B3', description='estrogen-related receptor gamma', other_designations='ERR gamma-2|estrogen receptor-related protein 3|nuclear receptor subfamily 3 group B member 3', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=216503246, end_position_on_the_genomic_accession=217137755, orientation='minus', exon_count=26, OMIM=602969, no_hdr0=''),
  2107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2107, CurrentID=0, Status='live', Symbol='ETF1', Aliases='D5S1995, ERF, ERF1, RF1, SUP45L1, TB3-1', description='eukaryotic translation termination factor 1', other_designations='polypeptide chain release factor 1|protein Cl1|sup45 (yeast omnipotent suppressor 45) homolog-like 1', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138506093, end_position_on_the_genomic_accession=138543300, orientation='minus', exon_count=13, OMIM=600285, no_hdr0=''),
  2108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2108, CurrentID=0, Status='live', Symbol='ETFA', Aliases='EMA, GA2, MADD', description='electron-transfer-flavoprotein, alpha polypeptide', other_designations='alpha-ETF|electron transfer flavoprotein alpha-subunit|electron transfer flavoprotein, alpha polypeptide|electron-transfer-flavoprotein alpha polypeptide|glutaric aciduria II', map_location='15q23-q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76216287, end_position_on_the_genomic_accession=76311469, orientation='minus', exon_count=13, OMIM=608053, no_hdr0=''),
  2109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2109, CurrentID=0, Status='live', Symbol='ETFB', Aliases='FP585, MADD', description='electron-transfer-flavoprotein, beta polypeptide', other_designations='beta-ETF|electron transfer flavoprotein beta subunit|electron transfer flavoprotein beta-subunit|electron transfer flavoprotein, beta polypeptide|electron-transferring-flavoprotein, beta polypeptide', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51345155, end_position_on_the_genomic_accession=51366418, orientation='minus', exon_count=6, OMIM=130410, no_hdr0=''),
  2110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2110, CurrentID=0, Status='live', Symbol='ETFDH', Aliases='ETFQO, MADD', description='electron-transferring-flavoprotein dehydrogenase', other_designations='ETF dehydrogenase|ETF-QO|ETF-ubiquinone oxidoreductase', map_location='4q32-q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158672101, end_position_on_the_genomic_accession=158708713, orientation='plus', exon_count=13, OMIM=231675, no_hdr0=''),
  2112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2112, CurrentID=0, Status='live', Symbol='ETM2', Aliases='ETM', description='essential tremor 2', other_designations='', map_location='2p25-p22', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602134, no_hdr0=''),
  2113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2113, CurrentID=0, Status='live', Symbol='ETS1', Aliases='ETS-1, EWSR2, c-ets-1, p54', description='v-ets avian erythroblastosis virus E26 oncogene homolog 1', other_designations='Avian erythroblastosis virus E26 (v-ets) oncogene homolog-1|v-ets avian erythroblastosis virus E2 oncogene homolog 1', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128458761, end_position_on_the_genomic_accession=128587584, orientation='minus', exon_count=12, OMIM=164720, no_hdr0=''),
  2114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2114, CurrentID=0, Status='live', Symbol='ETS2', Aliases='ETS2IT1', description='v-ets avian erythroblastosis virus E26 oncogene homolog 2', other_designations='oncogene ETS-2|v-ets avian erythroblastosis virus E2 oncogene homolog 2|v-ets erythroblastosis virus E26 oncogene homolog 2', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38805307, end_position_on_the_genomic_accession=38824955, orientation='plus', exon_count=12, OMIM=164740, no_hdr0=''),
  2115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2115, CurrentID=0, Status='live', Symbol='ETV1', Aliases='ER81', description='ets variant 1', other_designations='ets variant gene 1|ets-related protein 81', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=13891229, end_position_on_the_genomic_accession=13991425, orientation='minus', exon_count=21, OMIM=600541, no_hdr0=''),
  2116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2116, CurrentID=0, Status='live', Symbol='ETV2', Aliases='ER71, ETSRP71', description='ets variant 2', other_designations='ets variant gene 2|ets-related protein 71', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35641213, end_position_on_the_genomic_accession=35644871, orientation='plus', exon_count=7, OMIM=609358, no_hdr0=''),
  2117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2117, CurrentID=0, Status='live', Symbol='ETV3', Aliases='METS, PE-1, PE1, bA110J1.4', description='ets variant 3', other_designations='ETS domain transcriptional repressor PE1|ets variant gene 3, ETS family transcriptional repressor|mitogenic Ets transcriptional suppressor', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157121191, end_position_on_the_genomic_accession=157138591, orientation='minus', exon_count=5, OMIM=164873, no_hdr0=''),
  2118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2118, CurrentID=0, Status='live', Symbol='ETV4', Aliases='E1A-F, E1AF, PEA3, PEAS3', description='ets variant 4', other_designations='EWS protein/E1A enhancer binding protein chimera|adenovirus E1A enhancer-binding protein|ets variant gene 4 (E1A enhancer-binding protein, E1AF)|polyomavirus enhancer activator 3 homolog', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43527843, end_position_on_the_genomic_accession=43546432, orientation='minus', exon_count=14, OMIM=600711, no_hdr0=''),
  2119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2119, CurrentID=0, Status='live', Symbol='ETV5', Aliases='ERM', description='ets variant 5', other_designations='ets-related molecule|ets-related protein ERM', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186046317, end_position_on_the_genomic_accession=186109112, orientation='minus', exon_count=13, OMIM=601600, no_hdr0=''),
  2120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2120, CurrentID=0, Status='live', Symbol='ETV6', Aliases='TEL, TEL/ABL, THC5', description='ets variant 6', other_designations='ETS translocation variant 6|ETS-related protein Tel1|TEL1 oncogene|ets variant gene 6 (TEL oncogene)', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11649854, end_position_on_the_genomic_accession=11895402, orientation='plus', exon_count=13, OMIM=600618, no_hdr0=''),
  2121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2121, CurrentID=0, Status='live', Symbol='EVC', Aliases='DWF-11, EVCL, EVC', description='Ellis van Creveld protein', other_designations='Ellis van Creveld syndrome', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5711189, end_position_on_the_genomic_accession=5829045, orientation='plus', exon_count=29, OMIM=604831, no_hdr0=''),
  2122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2122, CurrentID=0, Status='live', Symbol='MECOM', Aliases='AML1-EVI-1, EVI1, MDS1, MDS1-EVI1, PRDM3', description='MDS1 and EVI1 complex locus', other_designations='AML1-EVI-1 fusion protein|MDS1 and EVI1 complex locus protein MDS1|ecotropic virus integration site 1 protein homolog|myelodysplasia syndrome-associated protein 1|oncogene EVI1|zinc finger protein Evi1', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169083499, end_position_on_the_genomic_accession=169663786, orientation='minus', exon_count=22, OMIM=165215, no_hdr0=''),
  2123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2123, CurrentID=0, Status='live', Symbol='EVI2A', Aliases='EVDA, EVI-2A, EVI2', description='ecotropic viral integration site 2A', other_designations='ecotropic viral integration site 2A protein homolog', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31316410, end_position_on_the_genomic_accession=31321749, orientation='minus', exon_count=3, OMIM=158380, no_hdr0=''),
  2124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2124, CurrentID=0, Status='live', Symbol='EVI2B', Aliases='CD361, D17S376, EVDB', description='ecotropic viral integration site 2B', other_designations='EVI-2B|ecotropic viral integration site 2B protein homolog', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31302870, end_position_on_the_genomic_accession=31314389, orientation='minus', exon_count=4, OMIM=158381, no_hdr0=''),
  2125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2125, CurrentID=0, Status='live', Symbol='EVPL', Aliases='EVPK', description='envoplakin', other_designations='210 kDa cornified envelope precursor protein|210 kDa paraneoplastic pemphigus antigen|p210', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76006845, end_position_on_the_genomic_accession=76027452, orientation='minus', exon_count=22, OMIM=601590, no_hdr0=''),
  2128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2128, CurrentID=0, Status='live', Symbol='EVX1', Aliases='EVX-1', description='even-skipped homeobox 1', other_designations='eve, even-skipped homeo box homolog 1|eve, even-skipped homeobox homolog 1|even-skipped homeo box 1 (homolog of Drosophila)', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27242545, end_position_on_the_genomic_accession=27247819, orientation='plus', exon_count=5, OMIM=142996, no_hdr0=''),
  2130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2130, CurrentID=0, Status='live', Symbol='EWSR1', Aliases='EWS, EWS-FLI1, bK984G1.4', description='EWS RNA binding protein 1', other_designations='EWS RNA-binding protein 1|EWS RNA-binding protein variant 6|Ewing sarcoma breakpoint region 1|Ewings sarcoma EWS-Fli1 (type 1) oncogene', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29268009, end_position_on_the_genomic_accession=29300525, orientation='plus', exon_count=19, OMIM=133450, no_hdr0=''),
  2131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2131, CurrentID=0, Status='live', Symbol='EXT1', Aliases='EXT, LGCR, LGS, TRPS2, TTV', description='exostosin glycosyltransferase 1', other_designations='Glucuronosyl-N-acetylglucosaminyl-proteoglycan 4-alpha-N- acetylglucosaminyltransferase|Langer-Giedion syndrome chromosome region|N-acetylglucosaminyl-proteoglycan 4-beta-glucuronosyltransferase|exostoses (multiple) 1|exostosin 1|glucuronosyl-N-acetylglucosaminyl-proteoglycan/N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostoses protein 1|putative tumor suppressor protein EXT1', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117799363, end_position_on_the_genomic_accession=118111819, orientation='minus', exon_count=11, OMIM=608177, no_hdr0=''),
  2132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2132, CurrentID=0, Status='live', Symbol='EXT2', Aliases='SOTV, SSMS', description='exostosin glycosyltransferase 2', other_designations='N-acetylglucosaminyl-proteoglycan 4-beta-glucuronosyltransferase|glucuronosyl-N-acetylglucosaminyl-proteoglycan/N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostoses protein 2|putative tumor suppressor protein EXT2', map_location='11p12-p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44095549, end_position_on_the_genomic_accession=44245430, orientation='plus', exon_count=17, OMIM=608210, no_hdr0=''),
  2133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2133, CurrentID=0, Status='live', Symbol='EXT3', Aliases='EXT2', description='exostoses (multiple) 3', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600209, no_hdr0=''),
  2134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2134, CurrentID=0, Status='live', Symbol='EXTL1', Aliases='EXTL', description='exostosin-like glycosyltransferase 1', other_designations='alpha-N-acetylglucosaminyltransferase II|exostoses (multiple)-like 1|exostosin-L|glucuronosyl-N-acetylglucosaminyl-proteoglycan 4-alpha-N- acetylglucosaminyltransferase|glucuronosyl-N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|glucuronyl-N-acetylglucosaminylproteoglycan alpha-1,4-N- acetylglucosaminyltransferase|multiple exostosis-like protein', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26021741, end_position_on_the_genomic_accession=26036464, orientation='plus', exon_count=12, OMIM=601738, no_hdr0=''),
  2135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2135, CurrentID=0, Status='live', Symbol='EXTL2', Aliases='EXTR2', description='exostosin-like glycosyltransferase 2', other_designations='EXT-related protein 2|alpha-1,4-N-acetylhexosaminyltransferase EXTL2|alpha-GalNAcT EXTL2|exostoses (multiple)-like 2|glucuronyl-galactosyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|processed exostosin-like 2', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100872372, end_position_on_the_genomic_accession=100895323, orientation='minus', exon_count=8, OMIM=602411, no_hdr0=''),
  2137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2137, CurrentID=0, Status='live', Symbol='EXTL3', Aliases='BOTV, EXTL1L, EXTR1, REGR, RPR', description='exostosin-like glycosyltransferase 3', other_designations='EXT-related 1|exostoses (multiple)-like 3|exostosin tumor-like 3|glucuronyl-galactosyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|hereditary multiple exostoses gene isolog|putative tumor suppressor protein EXTL3|reg receptor', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28701473, end_position_on_the_genomic_accession=28753690, orientation='plus', exon_count=7, OMIM=605744, no_hdr0=''),
  2138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2138, CurrentID=0, Status='live', Symbol='EYA1', Aliases='BOP, BOR, BOS1, OFC1', description='EYA transcriptional coactivator and phosphatase 1', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71197433, end_position_on_the_genomic_accession=71548109, orientation='minus', exon_count=24, OMIM=601653, no_hdr0=''),
  2139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2139, CurrentID=0, Status='live', Symbol='EYA2', Aliases='EAB1', description='EYA transcriptional coactivator and phosphatase 2', other_designations='', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46894624, end_position_on_the_genomic_accession=47188844, orientation='plus', exon_count=17, OMIM=601654, no_hdr0=''),
  2140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2140, CurrentID=0, Status='live', Symbol='EYA3', Aliases='', description='EYA transcriptional coactivator and phosphatase 3', other_designations='eyes absent 3', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27970344, end_position_on_the_genomic_accession=28088675, orientation='minus', exon_count=22, OMIM=601655, no_hdr0=''),
  2141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2141, CurrentID=0, Status='live', Symbol='EYCL1', Aliases='GEY', description='eye color 1 (green/blue)', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100272224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272224, CurrentID=0, Status='live', Symbol='UAQTL3', Aliases='GOUT3', description='Uric acid concentration, serum, QTL3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612670, no_hdr0=''),
  2145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2145, CurrentID=0, Status='live', Symbol='EZH1', Aliases='KMT6B', description='enhancer of zeste 1 polycomb repressive complex 2 subunit', other_designations='ENX-2|enhancer of zeste homolog 1', map_location='17q21.1-q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42700275, end_position_on_the_genomic_accession=42745078, orientation='minus', exon_count=21, OMIM=601674, no_hdr0=''),
  2146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2146, CurrentID=0, Status='live', Symbol='EZH2', Aliases='ENX-1, ENX1, EZH1b, KMT6, KMT6A, WVS, WVS2, EZH2', description='enhancer of zeste 2 polycomb repressive complex 2 subunit', other_designations='enhancer of zeste homolog 2|lysine N-methyltransferase 6', map_location='7q35-q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148807372, end_position_on_the_genomic_accession=148884349, orientation='minus', exon_count=25, OMIM=601573, no_hdr0=''),
  2147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2147, CurrentID=0, Status='live', Symbol='F2', Aliases='PT, RPRGL2, THPH1', description='coagulation factor II (thrombin)', other_designations='prepro-coagulation factor II|prothrombin B-chain|serine protease', map_location='11p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46719192, end_position_on_the_genomic_accession=46739506, orientation='plus', exon_count=14, OMIM=176930, no_hdr0=''),
  100272228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272228, CurrentID=0, Status='live', Symbol='LINC00894', Aliases='', description='long intergenic non-protein coding RNA 894', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149938548, end_position_on_the_genomic_accession=150016787, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  2149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2149, CurrentID=0, Status='live', Symbol='F2R', Aliases='CF2R, HTR, PAR-1, PAR1, TR', description='coagulation factor II (thrombin) receptor', other_designations='protease-activated receptor 1', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76716043, end_position_on_the_genomic_accession=76735770, orientation='plus', exon_count=2, OMIM=187930, no_hdr0=''),
  2150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2150, CurrentID=0, Status='live', Symbol='F2RL1', Aliases='GPR11, PAR2', description='coagulation factor II (thrombin) receptor-like 1', other_designations='G-protein coupled receptor 11|coagulation factor II receptor-like 1|protease-activated receptor 2|thrombin receptor-like 1', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76819008, end_position_on_the_genomic_accession=76835315, orientation='plus', exon_count=2, OMIM=600933, no_hdr0=''),
  2151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2151, CurrentID=0, Status='live', Symbol='F2RL2', Aliases='PAR-3, PAR3', description='coagulation factor II (thrombin) receptor-like 2', other_designations='Coagulation factor II receptor-like 2 (protease-actovated receptor 3)|protease-activated receptor 3|proteinase-activated receptor-3|thrombin receptor-like 2', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76615482, end_position_on_the_genomic_accession=76623434, orientation='minus', exon_count=2, OMIM=601919, no_hdr0=''),
  2152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2152, CurrentID=0, Status='live', Symbol='F3', Aliases='CD142, TF, TFA', description='coagulation factor III (thromboplastin, tissue factor)', other_designations='', map_location='1p22-p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94529176, end_position_on_the_genomic_accession=94541857, orientation='minus', exon_count=6, OMIM=134390, no_hdr0=''),
  2153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2153, CurrentID=0, Status='live', Symbol='F5', Aliases='FVL, PCCF, RPRGL1, THPH2', description='coagulation factor V (proaccelerin, labile factor)', other_designations='activated protein c cofactor|coagulation factor V jinjiang A2 domain|factor V Leiden', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169511954, end_position_on_the_genomic_accession=169586531, orientation='minus', exon_count=25, OMIM=612309, no_hdr0=''),
  2155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2155, CurrentID=0, Status='live', Symbol='F7', Aliases='SPCA', description='coagulation factor VII (serum prothrombin conversion accelerator)', other_designations='FVII coagulation protein|eptacog alfa|proconvertin', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113105788, end_position_on_the_genomic_accession=113120681, orientation='plus', exon_count=11, OMIM=613878, no_hdr0=''),
  2156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2156, CurrentID=0, Status='live', Symbol='F7R', Aliases='F7E', description='coagulation factor VII regulator', other_designations='', map_location='8p23.2-p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=134450, no_hdr0=''),
  2157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2157, CurrentID=0, Status='live', Symbol='F8', Aliases='AHF, DXS1253EB, F8C, FVIII, HEMA, F8', description='coagulation factor VIII, procoagulant component', other_designations='antihemophilic factor|coagulation factor VIIIc|factor VIII F8B', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154835788, end_position_on_the_genomic_accession=155026934, orientation='minus', exon_count=28, OMIM=300841, no_hdr0=''),
  2158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2158, CurrentID=0, Status='live', Symbol='F9', Aliases='F9 p22, FIX, HEMB, P19, PTC, THPH8', description='coagulation factor IX', other_designations='Christmas factor|factor 9|factor IX F9|plasma thromboplastic component|plasma thromboplastin component', map_location='Xq27.1-q27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139530734, end_position_on_the_genomic_accession=139563464, orientation='plus', exon_count=8, OMIM=300746, no_hdr0=''),
  2159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2159, CurrentID=0, Status='live', Symbol='F10', Aliases='FX, FXA', description='coagulation factor X', other_designations='Stuart-Prower factor|factor Xa|prothrombinase', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113122799, end_position_on_the_genomic_accession=113149529, orientation='plus', exon_count=8, OMIM=613872, no_hdr0=''),
  2160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2160, CurrentID=0, Status='live', Symbol='F11', Aliases='FXI', description='coagulation factor XI', other_designations='PTA|plasma thromboplastin antecedent', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186265945, end_position_on_the_genomic_accession=186289681, orientation='plus', exon_count=17, OMIM=264900, no_hdr0=''),
  2161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2161, CurrentID=0, Status='live', Symbol='F12', Aliases='HAE3, HAEX, HAF', description='coagulation factor XII (Hageman factor)', other_designations='Hageman factor|beta-factor XIIa part 1|beta-factor XIIa part 2|coagulation factor XIIa heavy chain|coagulation factor XIIa light chain', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177402138, end_position_on_the_genomic_accession=177416524, orientation='minus', exon_count=16, OMIM=610619, no_hdr0=''),
  2162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2162, CurrentID=0, Status='live', Symbol='F13A1', Aliases='F13A', description='coagulation factor XIII, A1 polypeptide', other_designations='FSF, A subunit|TGase|bA525O21.1 (coagulation factor XIII, A1 polypeptide)|coagulation factor XIII, A polypeptide|coagulation factor XIIIa|factor XIIIa|fibrin stabilizing factor, A subunit|fibrinoligase|protein-glutamine gamma-glutamyltransferase A chain|transglutaminase A chain|transglutaminase. plasma', map_location='6p25.3-p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6144078, end_position_on_the_genomic_accession=6321013, orientation='minus', exon_count=17, OMIM=134570, no_hdr0=''),
  2165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2165, CurrentID=0, Status='live', Symbol='F13B', Aliases='FXIIIB', description='coagulation factor XIII, B polypeptide', other_designations='TGase|fibrin-stabilizing factor B subunit|protein-glutamine gamma-glutamyltransferase B chain|transglutaminase B chain', map_location='1q31-q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197039191, end_position_on_the_genomic_accession=197067267, orientation='minus', exon_count=13, OMIM=134580, no_hdr0=''),
  2166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2166, CurrentID=0, Status='live', Symbol='FAAH', Aliases='FAAH-1, PSAB', description='fatty acid amide hydrolase', other_designations='anandamide amidohydrolase 1|oleamide hydrolase 1', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46394267, end_position_on_the_genomic_accession=46413848, orientation='plus', exon_count=15, OMIM=602935, no_hdr0=''),
  2167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2167, CurrentID=0, Status='live', Symbol='FABP4', Aliases='A-FABP, AFABP, ALBP, HEL-S-104, aP2', description='fatty acid binding protein 4, adipocyte', other_designations='adipocyte lipid-binding protein|adipocyte-type fatty acid-binding protein|epididymis secretory protein Li 104|fatty acid-binding protein 4', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81478497, end_position_on_the_genomic_accession=81483238, orientation='minus', exon_count=4, OMIM=600434, no_hdr0=''),
  2168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2168, CurrentID=0, Status='live', Symbol='FABP1', Aliases='FABPL, L-FABP', description='fatty acid binding protein 1, liver', other_designations='fatty acid-binding protein 1|liver-type fatty acid-binding protein', map_location='2p11', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88122982, end_position_on_the_genomic_accession=88128131, orientation='minus', exon_count=4, OMIM=134650, no_hdr0=''),
  2169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2169, CurrentID=0, Status='live', Symbol='FABP2', Aliases='FABPI, I-FABP', description='fatty acid binding protein 2, intestinal', other_designations='fatty acid-binding protein 2|intestinal-type fatty acid-binding protein', map_location='4q28-q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119317250, end_position_on_the_genomic_accession=119322161, orientation='minus', exon_count=4, OMIM=134640, no_hdr0=''),
  2170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2170, CurrentID=0, Status='live', Symbol='FABP3', Aliases='FABP11, H-FABP, M-FABP, MDGI, O-FABP', description='fatty acid binding protein 3, muscle and heart', other_designations='Fatty acid-binding protein 3, muscle|fatty acid binding protein 11|heart-type fatty acid-binding protein|mammary-derived growth inhibitor|muscle fatty acid-binding protein', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31360418, end_position_on_the_genomic_accession=31373283, orientation='minus', exon_count=5, OMIM=134651, no_hdr0=''),
  2171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2171, CurrentID=0, Status='live', Symbol='FABP5', Aliases='E-FABP, EFABP, KFABP, PA-FABP, PAFABP', description='fatty acid binding protein 5 (psoriasis-associated)', other_designations='epidermal-type fatty acid-binding protein|psoriasis-associated fatty acid-binding protein homolog', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81280483, end_position_on_the_genomic_accession=81284777, orientation='plus', exon_count=4, OMIM=605168, no_hdr0=''),
  2172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2172, CurrentID=0, Status='live', Symbol='FABP6', Aliases='I-15P, I-BABP, I-BALB, I-BAP, ILBP, ILBP3, ILLBP', description='fatty acid binding protein 6, ileal', other_designations='GT|ileal bile acid binding protein|ileal lipid-binding protein|illeal lipid-binding protein|intestinal 15 kDa protein', map_location='5q33.3-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160187367, end_position_on_the_genomic_accession=160238722, orientation='plus', exon_count=8, OMIM=600422, no_hdr0=''),
  2173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2173, CurrentID=0, Status='live', Symbol='FABP7', Aliases='B-FABP, BLBP, FABPB, LTR2-FABP7, MRG', description='fatty acid binding protein 7, brain', other_designations='Hypothetical protein DKFZp547J2313|brain lipid binding protein|brain lipid-binding protein|brain-type fatty acid-binding protein|fatty acid-binding protein 7|mammary-derived growth inhibitor related|mammary-derived growth inhibitor-related', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122779501, end_position_on_the_genomic_accession=122784074, orientation='plus', exon_count=4, OMIM=602965, no_hdr0=''),
  2175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2175, CurrentID=0, Status='live', Symbol='FANCA', Aliases='FA, FA-H, FA1, FAA, FACA, FAH, FANCH', description='Fanconi anemia complementation group A', other_designations='Fanconi anemia, complementation group A|Fanconi anemia, complementation group H|Fanconi anemia, type 1', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89737551, end_position_on_the_genomic_accession=89816658, orientation='minus', exon_count=44, OMIM=607139, no_hdr0=''),
  2176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2176, CurrentID=0, Status='live', Symbol='FANCC', Aliases='FA3, FAC, FACC', description='Fanconi anemia complementation group C', other_designations='Fanconi anemia, complementation group C', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95099054, end_position_on_the_genomic_accession=95317709, orientation='minus', exon_count=22, OMIM=613899, no_hdr0=''),
  2177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2177, CurrentID=0, Status='live', Symbol='FANCD2', Aliases='FA-D2, FA4, FACD, FAD, FAD2, FANCD', description='Fanconi anemia complementation group D2', other_designations='Fanconi anemia, complementation group D2', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10026387, end_position_on_the_genomic_accession=10101932, orientation='plus', exon_count=47, OMIM=613984, no_hdr0=''),
  2178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2178, CurrentID=0, Status='live', Symbol='FANCE', Aliases='FACE, FAE', description='Fanconi anemia complementation group E', other_designations='Fanconi anemia, complementation group E', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35452339, end_position_on_the_genomic_accession=35467106, orientation='plus', exon_count=11, OMIM=613976, no_hdr0=''),
  2180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2180, CurrentID=0, Status='live', Symbol='ACSL1', Aliases='ACS1, FACL1, FACL2, LACS, LACS1, LACS2', description='acyl-CoA synthetase long-chain family member 1', other_designations='LACS 1|LACS 2|acyl-CoA synthetase 1|fatty-acid-Coenzyme A ligase, long-chain 1|fatty-acid-Coenzyme A ligase, long-chain 2|lignoceroyl-CoA synthase|long-chain acyl-CoA synthetase 1|long-chain acyl-CoA synthetase 2|long-chain fatty acid-CoA ligase 2|long-chain fatty-acid-coenzyme A ligase 1|palmitoyl-CoA ligase 1|palmitoyl-CoA ligase 2|paltimoyl-CoA ligase 1', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184755595, end_position_on_the_genomic_accession=184826114, orientation='minus', exon_count=28, OMIM=152425, no_hdr0=''),
  2181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2181, CurrentID=0, Status='live', Symbol='ACSL3', Aliases='ACS3, FACL3, PRO2194', description='acyl-CoA synthetase long-chain family member 3', other_designations='LACS 3|fatty-acid-Coenzyme A ligase, long-chain 3|lignoceroyl-CoA synthase|long-chain acyl-CoA synthetase 3', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222860984, end_position_on_the_genomic_accession=222943401, orientation='plus', exon_count=17, OMIM=602371, no_hdr0=''),
  2182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2182, CurrentID=0, Status='live', Symbol='ACSL4', Aliases='ACS4, FACL4, LACS4, MRX63, MRX68', description='acyl-CoA synthetase long-chain family member 4', other_designations='LACS 4|acyl-CoA synthetase 4|fatty-acid-Coenzyme A ligase, long-chain 4|lignoceroyl-CoA synthase|long-chain acyl-CoA synthetase 4|long-chain fatty-acid-Coenzyme A ligase 4', map_location='Xq22.3-q23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109641335, end_position_on_the_genomic_accession=109733392, orientation='minus', exon_count=17, OMIM=300157, no_hdr0=''),
  2184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2184, CurrentID=0, Status='live', Symbol='FAH', Aliases='', description='fumarylacetoacetate hydrolase (fumarylacetoacetase)', other_designations='FAA|beta-diketonase', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80152891, end_position_on_the_genomic_accession=80186582, orientation='plus', exon_count=14, OMIM=613871, no_hdr0=''),
  2185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2185, CurrentID=0, Status='live', Symbol='PTK2B', Aliases='CADTK, CAKB, FADK2, FAK2, PKB, PTK, PYK2, RAFTK', description='protein tyrosine kinase 2 beta', other_designations='CAK-beta|FADK 2|PTK2B protein tyrosine kinase 2 beta|calcium-dependent tyrosine kinase|calcium-regulated non-receptor proline-rich tyrosine kinase|cell adhesion kinase beta|focal adhesion kinase 2|proline-rich tyrosine kinase 2|protein kinase B|related adhesion focal tyrosine kinase', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27311482, end_position_on_the_genomic_accession=27459391, orientation='plus', exon_count=38, OMIM=601212, no_hdr0=''),
  2186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2186, CurrentID=0, Status='live', Symbol='BPTF', Aliases='FAC1, FALZ, NURF301', description='bromodomain PHD finger transcription factor', other_designations='bromodomain and PHD domain transcription factor|bromodomain and PHD finger-containing transcription factor|fetal Alz-50 clone 1 protein|fetal Alz-50 reactive clone 1|fetal Alzheimer antigen|nucleosome remodeling factor, large subunit', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67825528, end_position_on_the_genomic_accession=67984378, orientation='plus', exon_count=34, OMIM=601819, no_hdr0=''),
  2187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2187, CurrentID=0, Status='live', Symbol='FANCB', Aliases='FA2, FAAP90, FAAP95, FAB, FACB', description='Fanconi anemia complementation group B', other_designations='Fanconi anemia, complementation group B|Fanconi anemia-associated polypeptide of 95 kDa', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14796153, end_position_on_the_genomic_accession=14873317, orientation='minus', exon_count=13, OMIM=300515, no_hdr0=''),
  2188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2188, CurrentID=0, Status='live', Symbol='FANCF', Aliases='FAF', description='Fanconi anemia complementation group F', other_designations='Fanconi anemia, complementation group F', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22622533, end_position_on_the_genomic_accession=22625841, orientation='minus', exon_count=1, OMIM=613897, no_hdr0=''),
  2189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2189, CurrentID=0, Status='live', Symbol='FANCG', Aliases='FAG, XRCC9', description='Fanconi anemia complementation group G', other_designations='DNA repair protein XRCC9|Fanconi anemia, complementation group G|X-ray repair complementing defective repair in Chinese hamster cells 9|X-ray repair, complementing defective, in Chinese hamster, 9|truncated Fanconi anemia group G protein', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35073838, end_position_on_the_genomic_accession=35080016, orientation='minus', exon_count=14, OMIM=602956, no_hdr0=''),
  2191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2191, CurrentID=0, Status='live', Symbol='FAP', Aliases='DPPIVA, SIMP, FAP', description='fibroblast activation protein alpha', other_designations='170 kDa melanoma membrane-bound gelatinase|FAPalpha|dipeptidyl peptidase FAP|fibroblast activation protein, alpha|gelatine degradation protease FAP|integral membrane serine protease|post-proline cleaving enzyme|seprase|serine integral membrane protease|surface-expressed protease', map_location='2q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162170684, end_position_on_the_genomic_accession=162243535, orientation='minus', exon_count=30, OMIM=600403, no_hdr0=''),
  2192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2192, CurrentID=0, Status='live', Symbol='FBLN1', Aliases='FBLN, FIBL1', description='fibulin 1', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45502839, end_position_on_the_genomic_accession=45601134, orientation='plus', exon_count=20, OMIM=135820, no_hdr0=''),
  2193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2193, CurrentID=0, Status='live', Symbol='FARSA', Aliases='CML33, FARSL, FARSLA, FRSA, PheHA', description='phenylalanyl-tRNA synthetase alpha subunit', other_designations='pheRS|phenylalanine tRNA ligase 1, alpha, cytoplasmic|phenylalanine--tRNA ligase alpha chain|phenylalanine-tRNA synthetase alpha-subunit|phenylalanine-tRNA synthetase-like, alpha subunit|phenylalanyl-tRNA synthetase alpha chain|phenylalanyl-tRNA synthetase, alpha subunit|phenylalanyl-tRNA synthetase-like, alpha subunit', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12922470, end_position_on_the_genomic_accession=12933744, orientation='minus', exon_count=13, OMIM=602918, no_hdr0=''),
  2194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2194, CurrentID=0, Status='live', Symbol='FASN', Aliases='FAS, OA-519, SDR27X1', description='fatty acid synthase', other_designations='short chain dehydrogenase/reductase family 27X, member 1', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82078338, end_position_on_the_genomic_accession=82098289, orientation='minus', exon_count=43, OMIM=600212, no_hdr0=''),
  2195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2195, CurrentID=0, Status='live', Symbol='FAT1', Aliases='CDHF7, CDHR8, FAT, ME5, hFat1', description='FAT atypical cadherin 1', other_designations='FAT tumor suppressor 1|FAT tumor suppressor homolog 1|cadherin ME5|cadherin family member 7|cadherin-related family member 8|cadherin-related tumor suppressor homolog|protein fat homolog', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186587783, end_position_on_the_genomic_accession=186726915, orientation='minus', exon_count=29, OMIM=600976, no_hdr0=''),
  2196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2196, CurrentID=0, Status='live', Symbol='FAT2', Aliases='CDHF8, CDHR9, HFAT2, MEGF1', description='FAT atypical cadherin 2', other_designations='FAT tumor suppressor homolog 2|cadherin family member 8|cadherin-related family member 9|multiple EGF-like domains protein 1|multiple epidermal growth factor-like domains 1|multiple epidermal growth factor-like domains protein 1|protocadherin FAT2', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151504092, end_position_on_the_genomic_accession=151652140, orientation='minus', exon_count=34, OMIM=604269, no_hdr0=''),
  2197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2197, CurrentID=0, Status='live', Symbol='FAU', Aliases='FAU1, Fub1, Fubi, MNSFbeta, RPS30, S30, asr1', description='Finkel-Biskis-Reilly murine sarcoma virus (FBR-MuSV) ubiquitously expressed', other_designations='40S ribosomal protein S30|FAU-encoded ubiquitin-like protein|FBR-MuSV-associated ubiquitously expressed|Finkel-Biskis-Reilly murine sarcoma virus (FBR-MuSV) ubiquitously expressed (fox derived)|monoclonal nonspecific suppressor factor beta|ribosomal protein S30|ubiquitin-like-S30 fusion protein', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65120627, end_position_on_the_genomic_accession=65122200, orientation='minus', exon_count=5, OMIM=134690, no_hdr0=''),
  2199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2199, CurrentID=0, Status='live', Symbol='FBLN2', Aliases='', description='fibulin 2', other_designations='FIBL-2', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13549125, end_position_on_the_genomic_accession=13638422, orientation='plus', exon_count=19, OMIM=135821, no_hdr0=''),
  2200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2200, CurrentID=0, Status='live', Symbol='FBN1', Aliases='ACMICD, ECTOL1, FBN, GPHYSD2, MASS, MFS1, OCTD, SGS, SSKS, WMS, WMS2', description='fibrillin 1', other_designations='fibrillin 15', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48408306, end_position_on_the_genomic_accession=48645788, orientation='minus', exon_count=66, OMIM=134797, no_hdr0=''),
  2201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2201, CurrentID=0, Status='live', Symbol='FBN2', Aliases='CCA, DA9, EOMD', description='fibrillin 2', other_designations='fibrillin 5', map_location='5q23-q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128257909, end_position_on_the_genomic_accession=128538042, orientation='minus', exon_count=65, OMIM=612570, no_hdr0=''),
  2202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2202, CurrentID=0, Status='live', Symbol='EFEMP1', Aliases='DHRD, DRAD, FBLN3, FBNL, FIBL-3, MLVT, MTLV, S1-5', description='EGF containing fibulin-like extracellular matrix protein 1', other_designations='extracellular protein S1-5|fibulin-3', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55865962, end_position_on_the_genomic_accession=55924163, orientation='minus', exon_count=12, OMIM=601548, no_hdr0=''),
  2203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2203, CurrentID=0, Status='live', Symbol='FBP1', Aliases='FBP', description='fructose-bisphosphatase 1', other_designations='D-fructose-1,6-bisphosphate 1-phosphohydrolase 1|FBPase 1|growth-inhibiting protein 17|liver FBPase', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94603133, end_position_on_the_genomic_accession=94640249, orientation='minus', exon_count=8, OMIM=611570, no_hdr0=''),
  2204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2204, CurrentID=0, Status='live', Symbol='FCAR', Aliases='CD89, CTB-61M7.2, FcalphaRI', description='Fc fragment of IgA receptor', other_designations='Fc alpha receptor|Fc fragment of IgA, receptor for', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54874093, end_position_on_the_genomic_accession=54891420, orientation='plus', exon_count=6, OMIM=147045, no_hdr0=''),
  2205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2205, CurrentID=0, Status='live', Symbol='FCER1A', Aliases='FCE1A, FcERI', description='Fc fragment of IgE, high affinity I, receptor for; alpha polypeptide', other_designations='Fc IgE receptor, alpha polypeptide|Fc epsilon RI alpha-chain|Fc-epsilon RI-alpha|high affinity immunoglobulin epsilon receptor alpha-subunit|igE Fc receptor subunit alpha|immunoglobulin E receptor, high-affinity, of mast cells, alpha polypeptide', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159283888, end_position_on_the_genomic_accession=159308224, orientation='plus', exon_count=7, OMIM=147140, no_hdr0=''),
  2206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2206, CurrentID=0, Status='live', Symbol='MS4A2', Aliases='APY, ATOPY, FCER1B, FCERI, IGEL, IGER, IGHER, MS4A1', description='membrane-spanning 4-domains, subfamily A, member 2', other_designations='Fc fragment of IgE, high affinity I, receptor for; beta polypeptide|High affinity immunoglobulin epsilon receptor beta-subunit (FcERI) (IgE Fc receptor, beta-subunit) (Fc epsilon receptor I beta-chain)|high affinity IgE receptor beta subunit|igE Fc receptor subunit beta|immunoglobulin E receptor, high affinity, beta polypeptide', map_location='11q12-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60088261, end_position_on_the_genomic_accession=60098467, orientation='plus', exon_count=8, OMIM=147138, no_hdr0=''),
  2207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2207, CurrentID=0, Status='live', Symbol='FCER1G', Aliases='FCRG', description='Fc fragment of IgE, high affinity I, receptor for; gamma polypeptide', other_designations='Fc receptor gamma-chain|fc-epsilon RI-gamma|fcRgamma|fceRI gamma|immunoglobulin E receptor, high affinity, gamma chain', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161215297, end_position_on_the_genomic_accession=161219248, orientation='plus', exon_count=5, OMIM=147139, no_hdr0=''),
  2208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2208, CurrentID=0, Status='live', Symbol='FCER2', Aliases='BLAST-2, CD23, CD23A, CLEC4J, FCE2, IGEBF', description='Fc fragment of IgE, low affinity II, receptor for (CD23)', other_designations='C-type lectin domain family 4, member J|CD23 antigen|fc-epsilon-RII|immunoglobulin E-binding factor|immunoglobulin epsilon-chain|lymphocyte IgE receptor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7688757, end_position_on_the_genomic_accession=7702310, orientation='minus', exon_count=12, OMIM=151445, no_hdr0=''),
  2209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2209, CurrentID=0, Status='live', Symbol='FCGR1A', Aliases='CD64, CD64A, FCRI, IGFR1', description='Fc fragment of IgG, high affinity Ia, receptor (CD64)', other_designations='Fc fragment of IgG, high affinity Ia, receptor for (CD64)|Fc gamma receptor|Fc-gamma RI|Fc-gamma receptor I A1|IgG Fc receptor I|fc-gamma RIA|fcgammaRIa', map_location='1q21.2-q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149781275, end_position_on_the_genomic_accession=149792518, orientation='plus', exon_count=6, OMIM=146760, no_hdr0=''),
  2210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2210, CurrentID=0, Status='live', Symbol='FCGR1B', Aliases='CD64b, IGFRB', description='Fc fragment of IgG, high affinity Ib, receptor (CD64)', other_designations='Fc-gamma receptor I B2|fc-gamma RIB|fcRIB|hFcgammaRIB|igG Fc receptor IB', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121087345, end_position_on_the_genomic_accession=121097161, orientation='plus', exon_count=6, OMIM=601502, no_hdr0=''),
  133283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133283, CurrentID=0, Status='live', Symbol='EEF1A1P13', Aliases='EEF1AL11', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 13', other_designations='eukaryotic translation elongation factor 1 alpha-like 11', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14651643, end_position_on_the_genomic_accession=14653227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2212, CurrentID=0, Status='live', Symbol='FCGR2A', Aliases='CD32, CD32A, CDw32, FCG2, FCGR21, FcGR, IGFR2, FCGR2A', description='Fc fragment of IgG, low affinity IIa, receptor (CD32)', other_designations='Immunoglobulin G Fc receptor II|fc-gamma RII-a|fc-gamma-RIIa|fcRII-a|igG Fc receptor II-a', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161505415, end_position_on_the_genomic_accession=161524048, orientation='plus', exon_count=10, OMIM=146790, no_hdr0=''),
  2213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2213, CurrentID=0, Status='live', Symbol='FCGR2B', Aliases='CD32, CD32B, FCG2, FCGR2, IGFR2', description='Fc fragment of IgG, low affinity IIb, receptor (CD32)', other_designations='CDw32|Fc fragment of IgG, low affinity II, receptor for (CD32)|Fc fragment of IgG, low affinity IIb, receptor for (CD32)|Fc gamma RIIb|fc-gamma RII-b|fc-gamma-RIIb|fcRII-b|igG Fc receptor II-b', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161663115, end_position_on_the_genomic_accession=161678654, orientation='plus', exon_count=9, OMIM=604590, no_hdr0=''),
  2214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2214, CurrentID=0, Status='live', Symbol='FCGR3A', Aliases='CD16, CD16A, FCG3, FCGR3, FCGRIII, FCR-10, FCRIII, FCRIIIA, IGFR3, IMD20', description='Fc fragment of IgG, low affinity IIIa, receptor (CD16a)', other_designations='CD16a antigen|Fc fragment of IgG low affinity IIIa receptor|Fc fragment of IgG, low affinity III, receptor for (CD16)|Fc gamma receptor III-A|Fc-gamma RIII-alpha|Fc-gamma RIIIa|Fc-gamma receptor III-2 (CD 16)|Fc-gamma receptor IIIb (CD16)|FcgammaRIIIA|fc-gamma RIII|igG Fc receptor III-2|immunoglobulin G Fc receptor III|neutrophil-specific antigen NA', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161541759, end_position_on_the_genomic_accession=161550623, orientation='minus', exon_count=6, OMIM=146740, no_hdr0=''),
  2215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2215, CurrentID=0, Status='live', Symbol='FCGR3B', Aliases='CD16, CD16b, FCG3, FCGR3, FCR-10, FCRIII, FCRIIIb', description='Fc fragment of IgG, low affinity IIIb, receptor (CD16b)', other_designations='Fc-gamma receptor IIIb (CD 16)|fc-gamma RIII-beta|fc-gamma RIIIb|igG Fc receptor III-1', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161623196, end_position_on_the_genomic_accession=161631963, orientation='minus', exon_count=7, OMIM=610665, no_hdr0=''),
  2216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2216, CurrentID=0, Status='live', Symbol='FDX1P2', Aliases='FDXP2', description='ferredoxin 1 pseudogene 2', other_designations='adrenodoxin pseudogene|ferredoxin pseudogene 2', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25692171, end_position_on_the_genomic_accession=25693145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2217, CurrentID=0, Status='live', Symbol='FCGRT', Aliases='FCRN, alpha-chain', description='Fc fragment of IgG, receptor, transporter, alpha', other_designations='FcRn alpha chain|IgG Fc fragment receptor transporter alpha chain|immunoglobulin receptor, intestinal, heavy chain|major histocompatibility complex class I-like Fc receptor|neonatal Fc receptor|neonatal Fc-receptor for Ig', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49512279, end_position_on_the_genomic_accession=49526428, orientation='plus', exon_count=9, OMIM=601437, no_hdr0=''),
  2218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2218, CurrentID=0, Status='live', Symbol='FKTN', Aliases='CMD1X, FCMD, LGMD2M, MDDGA4, MDDGB4, MDDGC4', description='fukutin', other_designations='Fukuyama type congenital muscular dystrophy protein|patient fukutin', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105558119, end_position_on_the_genomic_accession=105641118, orientation='plus', exon_count=24, OMIM=607440, no_hdr0=''),
  2219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2219, CurrentID=0, Status='live', Symbol='FCN1', Aliases='FCNM', description='ficolin 1', other_designations='M-ficolin|collagen/fibrinogen domain-containing protein 1|ficolin (collagen/fibrinogen domain containing) 1|ficolin (collagen/fibrinogen domain-containing) 1|ficolin-A|ficolin-alpha', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134909585, end_position_on_the_genomic_accession=134917960, orientation='minus', exon_count=9, OMIM=601252, no_hdr0=''),
  2220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2220, CurrentID=0, Status='live', Symbol='FCN2', Aliases='EBP-37, FCNL, P35, ficolin-2', description='ficolin 2', other_designations='37 kDa elastin-binding protein|L-ficolin|collagen/fibrinogen domain-containing protein 2|ficolin (collagen/fibrinogen domain containing lectin) 2 (hucolin)|ficolin B|ficolin-B|ficolin-beta|serum lectin p35', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134864156, end_position_on_the_genomic_accession=134889286, orientation='plus', exon_count=9, OMIM=601624, no_hdr0=''),
  2221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2221, CurrentID=0, Status='live', Symbol='FCP1', Aliases='FCP, FCPX, HBFQTL3', description='F-cell production 1', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=305435, no_hdr0=''),
  2222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2222, CurrentID=0, Status='live', Symbol='FDFT1', Aliases='DGPT, ERG9, SQS, SS', description='farnesyl-diphosphate farnesyltransferase 1', other_designations='FPP:FPP farnesyltransferase|presqualene-di-diphosphate synthase|squalene synthetase', map_location='8p23.1-p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11795573, end_position_on_the_genomic_accession=11839309, orientation='plus', exon_count=14, OMIM=184420, no_hdr0=''),
  2224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2224, CurrentID=0, Status='live', Symbol='FDPS', Aliases='FPPS, FPS', description='farnesyl diphosphate synthase', other_designations='(2E,6E)-farnesyl diphosphate synthase|FPP synthase|FPP synthetase|farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155308748, end_position_on_the_genomic_accession=155320666, orientation='plus', exon_count=11, OMIM=134629, no_hdr0=''),
  2227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2227, CurrentID=0, Status='live', Symbol='FDPSP3', Aliases='FDPSL3, FPSL3', description='farnesyl diphosphate synthase pseudogene 3', other_designations='farnesyl diphosphate synthase-like 3 (farnesyl pyrophosphate synthetase-like 3)', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54884807, end_position_on_the_genomic_accession=54886244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2228, CurrentID=0, Status='live', Symbol='FDPSP4', Aliases='CHR39B, FDPSL4, FPSL4', description='farnesyl diphosphate synthase pseudogene 4', other_designations='cholesterol-repressible protein 39B|farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene|farnesyl diphosphate synthase-like 4 (farnesyl pyrophosphate synthetase-like 4)', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42939253, end_position_on_the_genomic_accession=42940404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2229, CurrentID=0, Status='live', Symbol='FDPSP5', Aliases='CHR39C, FDPSL5, FPSL5', description='farnesyl diphosphate synthase pseudogene 5', other_designations='cholesterol-repressible protein 39C|farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) (FDPS) pseudogene|farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene|farnesyl diphosphate synthase-like 5 (farnesyl pyrophosphate synthetase-like 5) pseudogene', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44475971, end_position_on_the_genomic_accession=44477480, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2230, CurrentID=0, Status='live', Symbol='FDX1', Aliases='ADX, FDX, LOH11CR1D', description='ferredoxin 1', other_designations='adrenal ferredoxin|ferredoxin-1|hepatoredoxin|mitochondrial adrenodoxin', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110429937, end_position_on_the_genomic_accession=110464884, orientation='plus', exon_count=4, OMIM=103260, no_hdr0=''),
  2231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2231, CurrentID=0, Status='live', Symbol='FDX1P1', Aliases='ADXP1, FDXP1', description='ferredoxin 1 pseudogene 1', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34473311, end_position_on_the_genomic_accession=34476490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2232, CurrentID=0, Status='live', Symbol='FDXR', Aliases='ADXR', description='ferredoxin reductase', other_designations='AR|adrenodoxin reductase|adrenodoxin-NADP(+) reductase|ferredoxin--NADP(+) reductase', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74862497, end_position_on_the_genomic_accession=74873031, orientation='minus', exon_count=15, OMIM=103270, no_hdr0=''),
  2233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2233, CurrentID=0, Status='live', Symbol='FEB1', Aliases='', description='febrile convulsions 1', other_designations='', map_location='8q13-q21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=121210, no_hdr0=''),
  2234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2234, CurrentID=0, Status='live', Symbol='FEB2', Aliases='', description='febrile convulsions 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602477, no_hdr0=''),
  2235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2235, CurrentID=0, Status='live', Symbol='FECH', Aliases='EPP, FCE', description='ferrochelatase', other_designations='heme synthase|heme synthetase|protoheme ferro-lyase|protoporphyria', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57544841, end_position_on_the_genomic_accession=57586737, orientation='minus', exon_count=12, OMIM=612386, no_hdr0=''),
  133308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133308, CurrentID=0, Status='live', Symbol='SLC9B2', Aliases='NHA2, NHE10, NHEDC2', description='solute carrier family 9, subfamily B (NHA2, cation proton antiporter 2), member 2', other_designations='NHE domain-containing protein 2|Na(+)/H(+) exchanger-like domain-containing protein 2|Na+/H+ exchanger domain containing 2|mitochondrial Na(+)/H(+) exchanger NHA2|mitochondrial sodium/hydrogen exchanger NHA2|sodium/hydrogen exchanger-like domain-containing protein 2|solute carrier family 9 subfamily B member 2|solute carrier family 9, subfamily B (cation proton antiporter 2), member 2', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103020230, end_position_on_the_genomic_accession=103077371, orientation='minus', exon_count=14, OMIM=611789, no_hdr0=''),
  2237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2237, CurrentID=0, Status='live', Symbol='FEN1', Aliases='FEN-1, MF1, RAD2', description='flap structure-specific endonuclease 1', other_designations='DNase IV|FEN-1|maturation factor 1|maturation factor-1', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61792637, end_position_on_the_genomic_accession=61797244, orientation='plus', exon_count=2, OMIM=600393, no_hdr0=''),
  2239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2239, CurrentID=0, Status='live', Symbol='GPC4', Aliases='K-glypican', description='glypican 4', other_designations='dJ900E8.1 (glypican 4)|glypican proteoglycan 4', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133301036, end_position_on_the_genomic_accession=133415177, orientation='minus', exon_count=9, OMIM=300168, no_hdr0=''),
  2241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2241, CurrentID=0, Status='live', Symbol='FER', Aliases='PPP1R74, TYK3, p94-Fer', description='fer (fps/fes related) tyrosine kinase', other_designations='feline encephalitis virus-related kinase FER|fujinami poultry sarcoma/Feline sarcoma-related protein Fer|phosphoprotein NCP94|protein phosphatase 1, regulatory subunit 74|proto-oncogene c-Fer|tyrosine kinase 3', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108746085, end_position_on_the_genomic_accession=109187674, orientation='plus', exon_count=29, OMIM=176942, no_hdr0=''),
  2242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2242, CurrentID=0, Status='live', Symbol='FES', Aliases='FPS', description='FES proto-oncogene, tyrosine kinase', other_designations='Oncogene FES, feline sarcoma virus|feline sarcoma (Snyder-Theilen) viral (v-fes)/Fujinami avian sarcoma (PRCII) viral (v-fps) oncogene homolog|feline sarcoma oncogene|feline sarcoma/Fujinami avian sarcoma oncogene homolog|p93c-fes|proto-oncogene c-Fes|proto-oncogene c-Fps|proto-oncogene tyrosine-protein kinase Fes/Fps', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90884435, end_position_on_the_genomic_accession=90895776, orientation='plus', exon_count=19, OMIM=190030, no_hdr0=''),
  2243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2243, CurrentID=0, Status='live', Symbol='FGA', Aliases='Fib2', description='fibrinogen alpha chain', other_designations='fibrinogen, A alpha polypeptide', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154583128, end_position_on_the_genomic_accession=154590745, orientation='minus', exon_count=6, OMIM=134820, no_hdr0=''),
  2244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2244, CurrentID=0, Status='live', Symbol='FGB', Aliases='HEL-S-78p', description='fibrinogen beta chain', other_designations='epididymis secretory sperm binding protein Li 78p|fibrinogen, B beta polypeptide', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154562980, end_position_on_the_genomic_accession=154572763, orientation='plus', exon_count=8, OMIM=134830, no_hdr0=''),
  2245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2245, CurrentID=0, Status='live', Symbol='FGD1', Aliases='AAS, FGDY, MRXS16, ZFYVE3', description='FYVE, RhoGEF and PH domain containing 1', other_designations='faciogenital dysplasia 1 protein|rho/Rac GEF|rho/Rac guanine nucleotide exchange factor FGD1|zinc finger FYVE domain-containing protein 3', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54445454, end_position_on_the_genomic_accession=54496166, orientation='minus', exon_count=18, OMIM=300546, no_hdr0=''),
  2246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2246, CurrentID=0, Status='live', Symbol='FGF1', Aliases='AFGF, ECGF, ECGF-beta, ECGFA, ECGFB, FGF-1, FGF-alpha, FGFA, GLIO703, HBGF-1, HBGF1', description='fibroblast growth factor 1 (acidic)', other_designations='beta-endothelial cell growth factor|endothelial cell growth factor, alpha|endothelial cell growth factor, beta|heparin-binding growth factor 1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142592178, end_position_on_the_genomic_accession=142698070, orientation='minus', exon_count=10, OMIM=131220, no_hdr0=''),
  2247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2247, CurrentID=0, Status='live', Symbol='FGF2', Aliases='BFGF, FGF-2, FGFB, HBGF-2', description='fibroblast growth factor 2 (basic)', other_designations='basic fibroblast growth factor bFGF|heparin-binding growth factor 2|prostatropin', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122826708, end_position_on_the_genomic_accession=122898235, orientation='plus', exon_count=3, OMIM=134920, no_hdr0=''),
  2248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2248, CurrentID=0, Status='live', Symbol='FGF3', Aliases='HBGF-3, INT2', description='fibroblast growth factor 3', other_designations='FGF-3|INT-2 proto-oncogene protein|V-INT2 murine mammary tumor virus integration site oncogene homolog|fibroblast growth factor 3 (murine mammary tumor virus integration site (v-int-2) oncogene homolog)|heparin-binding growth factor 3|murine mammary tumor virus integration site 2, mouse|oncogene INT2|proto-oncogene Int-2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69809968, end_position_on_the_genomic_accession=69819424, orientation='minus', exon_count=3, OMIM=164950, no_hdr0=''),
  2249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2249, CurrentID=0, Status='live', Symbol='FGF4', Aliases='HBGF-4, HST, HST-1, HSTF1, K-FGF, KFGF', description='fibroblast growth factor 4', other_designations='FGF-4|HSTF-1|fibroblast growth factor 4 splice isoform|heparin secretory transforming protein 1|heparin secretory-transforming protein 1|heparin-binding growth factor 4|human stomach cancer, transforming factor from FGF-related oncogene|kaposi sarcoma oncogene|oncogene HST|transforming protein KS3', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69759043, end_position_on_the_genomic_accession=69775403, orientation='minus', exon_count=5, OMIM=164980, no_hdr0=''),
  2250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2250, CurrentID=0, Status='live', Symbol='FGF5', Aliases='HBGF-5, Smag-82, TCMGLY', description='fibroblast growth factor 5', other_designations='heparin-binding growth factor 5', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80266588, end_position_on_the_genomic_accession=80291017, orientation='plus', exon_count=4, OMIM=165190, no_hdr0=''),
  2251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2251, CurrentID=0, Status='live', Symbol='FGF6', Aliases='HBGF-6, HST2', description='fibroblast growth factor 6', other_designations='FGF-6|HST-2|HSTF-2|heparin secretory-transforming protein 2|heparin-binding growth factor 6', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4434142, end_position_on_the_genomic_accession=4445614, orientation='minus', exon_count=3, OMIM=134921, no_hdr0=''),
  2252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2252, CurrentID=0, Status='live', Symbol='FGF7', Aliases='HBGF-7, KGF', description='fibroblast growth factor 7', other_designations='FGF-7|heparin-binding growth factor 7|keratinocyte growth factor', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49423178, end_position_on_the_genomic_accession=49487326, orientation='plus', exon_count=4, OMIM=148180, no_hdr0=''),
  2253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2253, CurrentID=0, Status='live', Symbol='FGF8', Aliases='AIGF, FGF-8, HBGF-8, HH6, KAL6', description='fibroblast growth factor 8', other_designations='androgen-induced growth factor|fibroblast growth factor 8 (androgen-induced)|heparin-binding growth factor 8', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101770130, end_position_on_the_genomic_accession=101780369, orientation='minus', exon_count=8, OMIM=600483, no_hdr0=''),
  2254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2254, CurrentID=0, Status='live', Symbol='FGF9', Aliases='FGF-9, GAF, HBFG-9, HBGF-9, SYNS3', description='fibroblast growth factor 9', other_designations='fibroblast growth factor 9 (glia-activating factor)|heparin-binding growth factor 9', map_location='13q11-q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21671076, end_position_on_the_genomic_accession=21704501, orientation='plus', exon_count=4, OMIM=600921, no_hdr0=''),
  2255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2255, CurrentID=0, Status='live', Symbol='FGF10', Aliases='', description='fibroblast growth factor 10', other_designations='FGF-10|keratinocyte growth factor 2|produced by fibroblasts of urinary bladder lamina propria', map_location='5p13-p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=44303541, end_position_on_the_genomic_accession=44389780, orientation='minus', exon_count=4, OMIM=602115, no_hdr0=''),
  2256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2256, CurrentID=0, Status='live', Symbol='FGF11', Aliases='FGF-11, FHF-3, FHF3', description='fibroblast growth factor 11', other_designations='fibroblast growth factor homologous factor 3', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7438273, end_position_on_the_genomic_accession=7444932, orientation='plus', exon_count=7, OMIM=601514, no_hdr0=''),
  2257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2257, CurrentID=0, Status='live', Symbol='FGF12', Aliases='FGF12B, FHF1', description='fibroblast growth factor 12', other_designations='FGF-12|FHF-1|fibroblast growth factor 12B|fibroblast growth factor FGF-12b|fibroblast growth factor homologous factor 1|myocyte-activating factor', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192139244, end_position_on_the_genomic_accession=192727599, orientation='minus', exon_count=10, OMIM=601513, no_hdr0=''),
  2258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2258, CurrentID=0, Status='live', Symbol='FGF13', Aliases='FGF-13, FGF2, FHF-2, FHF2', description='fibroblast growth factor 13', other_designations='fibroblast growth factor homologous factor 2', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138631573, end_position_on_the_genomic_accession=139222889, orientation='minus', exon_count=13, OMIM=300070, no_hdr0=''),
  2259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2259, CurrentID=0, Status='live', Symbol='FGF14', Aliases='FGF-14, FHF-4, FHF4, SCA27', description='fibroblast growth factor 14', other_designations='bA397O8.2|fibroblast growth factor homologous factor 4', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101719784, end_position_on_the_genomic_accession=102401774, orientation='minus', exon_count=7, OMIM=601515, no_hdr0=''),
  2260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2260, CurrentID=0, Status='live', Symbol='FGFR1', Aliases='BFGFR, CD331, CEK, FGFBR, FGFR-1, FLG, FLT-2, FLT2, HBGFR, HH2, HRTFDS, KAL2, N-SAM, OGD, bFGF-R-1', description='fibroblast growth factor receptor 1', other_designations='FGFR1/PLAG1 fusion|FMS-like tyrosine kinase 2|basic fibroblast growth factor receptor 1|fms-related tyrosine kinase 2|heparin-binding growth factor receptor|hydroxyaryl-protein kinase|proto-oncogene c-Fgr', map_location='8p11.23-p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38411138, end_position_on_the_genomic_accession=38468834, orientation='minus', exon_count=24, OMIM=136350, no_hdr0=''),
  2261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2261, CurrentID=0, Status='live', Symbol='FGFR3', Aliases='ACH, CD333, CEK2, HSFGFR3EX, JTK4', description='fibroblast growth factor receptor 3', other_designations='FGFR-3|fibroblast growth factor receptor 3 variant 4|hydroxyaryl-protein kinase|tyrosine kinase JTK4', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1793299, end_position_on_the_genomic_accession=1808872, orientation='plus', exon_count=19, OMIM=134934, no_hdr0=''),
  2262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2262, CurrentID=0, Status='live', Symbol='GPC5', Aliases='', description='glypican 5', other_designations='bA93M14.1|glypican proteoglycan 5', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91398619, end_position_on_the_genomic_accession=92867237, orientation='plus', exon_count=13, OMIM=602446, no_hdr0=''),
  2263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2263, CurrentID=0, Status='live', Symbol='FGFR2', Aliases='BBDS, BEK, BFR-1, CD332, CEK3, CFD1, ECT1, JWS, K-SAM, KGFR, TK14, TK25', description='fibroblast growth factor receptor 2', other_designations='BEK fibroblast growth factor receptor|FGF receptor|FGFR2-AHCYL1 fusion kinase protein|bacteria-expressed kinase|fibroblast growth factor receptor 2 isoform IgIIIc isoform|hydroxyaryl-protein kinase|keratinocyte growth factor receptor|protein tyrosine kinase, receptor like 14|soluble FGFR4 variant 4', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121478330, end_position_on_the_genomic_accession=121598656, orientation='minus', exon_count=23, OMIM=176943, no_hdr0=''),
  2264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2264, CurrentID=0, Status='live', Symbol='FGFR4', Aliases='CD334, JTK2, TKF', description='fibroblast growth factor receptor 4', other_designations='hydroxyaryl-protein kinase|protein-tyrosine kinase|tyrosine kinase related to fibroblast growth factor receptor|tyrosylprotein kinase', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177086872, end_position_on_the_genomic_accession=177098142, orientation='plus', exon_count=19, OMIM=134935, no_hdr0=''),
  2266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2266, CurrentID=0, Status='live', Symbol='FGG', Aliases='', description='fibrinogen gamma chain', other_designations='fibrinogen, gamma polypeptide', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154604134, end_position_on_the_genomic_accession=154612750, orientation='minus', exon_count=10, OMIM=134850, no_hdr0=''),
  2267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2267, CurrentID=0, Status='live', Symbol='FGL1', Aliases='HFREP1, HP-041, LFIRE-1, LFIRE1', description='fibrinogen like 1', other_designations='fibrinogen-like 1|hepassocin|hepatocellular carcinoma-related sequence|hepatocyte-derived fibrinogen-related protein 1|liver fibrinogen-related protein 1', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17864391, end_position_on_the_genomic_accession=17895538, orientation='minus', exon_count=10, OMIM=605776, no_hdr0=''),
  2268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2268, CurrentID=0, Status='live', Symbol='FGR', Aliases='SRC2, c-fgr, c-src2, p55-Fgr, p55c-fgr, p58-Fgr, p58c-fgr', description='FGR proto-oncogene, Src family tyrosine kinase', other_designations='Gardner-Rasheed feline sarcoma viral (v-fgr) oncogene homolog|c-fgr protooncogene|c-src-2 proto-oncogene|feline Gardner-Rasheed sarcoma viral oncogene homolog|p55-c-fgr protein|proto-oncogene c-Fgr|proto-oncogene tyrosine-protein kinase FGR|v-fgr feline Gardner-Rasheed sarcoma viral oncogene homolog', map_location='1p36.2-p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27612289, end_position_on_the_genomic_accession=27635561, orientation='minus', exon_count=18, OMIM=164940, no_hdr0=''),
  378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378, CurrentID=0, Status='live', Symbol='ARF4', Aliases='ARF2', description='ADP ribosylation factor 4', other_designations='ADP-ribosylation factor 2', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57571363, end_position_on_the_genomic_accession=57597488, orientation='minus', exon_count=6, OMIM=601177, no_hdr0=''),
  2270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2270, CurrentID=0, Status='live', Symbol='FGS2', Aliases='', description='FG syndrome 2', other_designations='', map_location='Xq11-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300321, no_hdr0=''),
  2271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2271, CurrentID=0, Status='live', Symbol='FH', Aliases='FMRD, HLRCC, LRCC, MCL, MCUL1', description='fumarate hydratase', other_designations='fumarase', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241497557, end_position_on_the_genomic_accession=241519785, orientation='minus', exon_count=11, OMIM=136850, no_hdr0=''),
  2272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2272, CurrentID=0, Status='live', Symbol='FHIT', Aliases='AP3Aase, FRA3B', description='fragile histidine triad', other_designations="AP3A hydrolase|diadenosine 5',5'''-P1,P3-triphosphate hydrolase|dinucleosidetriphosphatase|tumor suppressor protein", map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59747587, end_position_on_the_genomic_accession=61251459, orientation='minus', exon_count=14, OMIM=601153, no_hdr0=''),
  2273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2273, CurrentID=0, Status='live', Symbol='FHL1', Aliases='FHL-1A, FHL1B, FLH1A, KYOT, SLIM, SLIM-1, SLIM1, SLIMMER, XMPMA, FHL1', description='four and a half LIM domains 1', other_designations='LIM protein SLIMMER|four-and-a-half Lin11, Isl-1 and Mec-3 domains 1|skeletal muscle LIM-protein 1', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136146702, end_position_on_the_genomic_accession=136211359, orientation='plus', exon_count=15, OMIM=300163, no_hdr0=''),
  2274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2274, CurrentID=0, Status='live', Symbol='FHL2', Aliases='AAG11, DRAL, FHL-2, SLIM-3, SLIM3', description='four and a half LIM domains 2', other_designations='LIM domain protein DRAL|aging-associated gene 11|down-regulated in rhabdomyosarcoma LIM protein|skeletal muscle LIM-protein 3', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105360815, end_position_on_the_genomic_accession=105438773, orientation='minus', exon_count=11, OMIM=602633, no_hdr0=''),
  2275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2275, CurrentID=0, Status='live', Symbol='FHL3', Aliases='SLIM2', description='four and a half LIM domains 3', other_designations='FHL-3|LIM-only protein FHL3|SLIM-2|skeletal muscle LIM-protein 2', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37996770, end_position_on_the_genomic_accession=38005515, orientation='minus', exon_count=6, OMIM=602790, no_hdr0=''),
  2277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2277, CurrentID=0, Status='live', Symbol='FIGF', Aliases='VEGF-D, VEGFD', description='c-fos induced growth factor (vascular endothelial growth factor D)', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15345591, end_position_on_the_genomic_accession=15384413, orientation='minus', exon_count=7, OMIM=300091, no_hdr0=''),
  2278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2278, CurrentID=0, Status='live', Symbol='FIM1', Aliases='', description='Friend-murine leukemia virus integration site 1 homolog', other_designations='', map_location='6p23-p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2279, CurrentID=0, Status='live', Symbol='FIM3', Aliases='', description='Friend-murine leukemia virus integration site 3 homolog', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2280, CurrentID=0, Status='live', Symbol='FKBP1A', Aliases='FKBP-12, FKBP-1A, FKBP1, FKBP12, PKC12, PKCI2, PPIASE', description='FK506 binding protein 1A', other_designations='12 kDa FK506-binding protein|12 kDa FKBP|FK506 binding protein 1A, 12kDa|FK506 binding protein12|FK506-binding protein 1|FK506-binding protein 12|FK506-binding protein 1A (12kD)|FK506-binding protein, T-cell, 12-kD|FKBP12-Exip3|PPIase FKBP1A|calstabin 1|calstabin-1|immunophilin FKBP12|protein kinase C inhibitor 2|rotamase', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1368977, end_position_on_the_genomic_accession=1393172, orientation='minus', exon_count=5, OMIM=186945, no_hdr0=''),
  2281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2281, CurrentID=0, Status='live', Symbol='FKBP1B', Aliases='FKBP12.6, FKBP1L, OTK4, PKBP1L, PPIase', description='FK506 binding protein 1B', other_designations='12.6 kDa FK506-binding protein|12.6 kDa FKBP|FK506 binding protein 1B, 12.6 kDa|FK506-binding protein 12.6|FK506-binding protein 1B|FKBP-12.6|FKBP-1B|PPIase FKBP1B|calstabin 2|h-FKBP-12|immunophilin FKBP12.6|rotamase', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24049706, end_position_on_the_genomic_accession=24067743, orientation='plus', exon_count=6, OMIM=600620, no_hdr0=''),
  2282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2282, CurrentID=0, Status='live', Symbol='FKBP1AP1', Aliases='FKBP12, FKBP1P1', description='FK506 binding protein 1A pseudogene 1', other_designations='FK506 binding protein 1 pseudogene 1|FK506 binding protein 1A, 12kDa pseudogene 1|FK506-binding protein 1 pseudogene 1', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57825280, end_position_on_the_genomic_accession=57827462, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2283, CurrentID=0, Status='live', Symbol='FKBP1AP2', Aliases='FKBP12, FKBP1P2', description='FK506 binding protein 1A pseudogene 2', other_designations='FK506 binding protein 1 pseudogene 2|FK506 binding protein 1A, 12kDa pseudogene 2|FK506-binding protein 1 pseudogene 2', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73142102, end_position_on_the_genomic_accession=73143465, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2284, CurrentID=0, Status='live', Symbol='FKBP1AP3', Aliases='FKBP12, FKBP1P3', description='FK506 binding protein 1A pseudogene 3', other_designations='FK506 binding protein 1 pseudogene 3|FK506 binding protein 1A, 12kDa pseudogene 3|FK506-binding protein 1 pseudogene 3', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46505079, end_position_on_the_genomic_accession=46508509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2285, CurrentID=0, Status='live', Symbol='FKBP1AP4', Aliases='FKBP12, FKBP1P4', description='FK506 binding protein 1A pseudogene 4', other_designations='FK506 binding protein 1 pseudogene 4|FK506 binding protein 1A, 12kDa pseudogene 4|FK506-binding protein 1 pseudogene 4|WD repeat domain 70 pseudogene', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149462599, end_position_on_the_genomic_accession=149464217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2286, CurrentID=0, Status='live', Symbol='FKBP2', Aliases='FKBP-13, PPIase', description='FK506 binding protein 2', other_designations='13 kDa FK506-binding protein|13 kDa FKBP|FK506 binding protein 2, 13kDa|FK506-binding protein 2 (13kD)|FKBP-2|PPIase FKBP2|immunophilin FKBP13|proline isomerase|rapamycin-binding protein|rotamase', map_location='11q13.1-q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64240941, end_position_on_the_genomic_accession=64244135, orientation='plus', exon_count=8, OMIM=186946, no_hdr0=''),
  2287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2287, CurrentID=0, Status='live', Symbol='FKBP3', Aliases='FKBP-25, FKBP-3, FKBP25, PPIase', description='FK506 binding protein 3', other_designations='25 kDa FK506-binding protein|25 kDa FKBP|FK506 binding protein 3, 25kDa|FK506-binding protein 25, T-cell|FK506-binding protein 3 (25kD)|PPIase FKBP3|immunophilin FKBP25|rapamycin binding protein|rapamycin-selective 25 kDa immunophilin|rotamase', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45115599, end_position_on_the_genomic_accession=45134529, orientation='minus', exon_count=8, OMIM=186947, no_hdr0=''),
  2288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2288, CurrentID=0, Status='live', Symbol='FKBP4', Aliases='FKBP51, FKBP52, FKBP59, HBI, Hsp56, PPIase, p52', description='FK506 binding protein 4', other_designations='51 kDa FK506-binding protein|FK506 binding protein 4, 59kDa|FK506-binding protein 4 (59kD)|HSP binding immunophilin|T-cell FK506-binding protein, 59kD|peptidylprolyl cis-trans isomerase|rotamase', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2794942, end_position_on_the_genomic_accession=2805423, orientation='plus', exon_count=11, OMIM=600611, no_hdr0=''),
  2289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2289, CurrentID=0, Status='live', Symbol='FKBP5', Aliases='AIG61, FKBP54, P54, PPIase, Ptg-10, FKBP5', description='FK506 binding protein 5', other_designations='51 kDa FK506-binding protein|51 kDa FKBP|54 kDa progesterone receptor-associated immunophilin|FF1 antigen|FKBP-51|HSP90-binding immunophilin|PPIase FKBP5|T-cell FK506-binding protein|androgen-regulated protein 6|peptidylprolyl cis-trans isomerase|rotamase', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35573585, end_position_on_the_genomic_accession=35728583, orientation='minus', exon_count=13, OMIM=602623, no_hdr0=''),
  2290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2290, CurrentID=0, Status='live', Symbol='FOXG1', Aliases='BF1, BF2, FHKL3, FKH2, FKHL1, FKHL2, FKHL3, FKHL4A, FOXG1B, FOXG1C, HBF-1, HBF-2, HBF-3, HBF-G2, HBF2, HFK1, HFK2, HFK3, KHL2, QIN, FOXG1', description='forkhead box G1', other_designations='brain factor 1|brain factor 2|forkhead-like 1|forkhead-like 2|forkhead-like 3|forkhead-like 4|oncogene QIN', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28767072, end_position_on_the_genomic_accession=28770277, orientation='plus', exon_count=1, OMIM=164874, no_hdr0=''),
  2294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2294, CurrentID=0, Status='live', Symbol='FOXF1', Aliases='ACDMPV, FKHL5, FREAC1', description='forkhead box F1', other_designations='FREAC-1|Forkhead, drosophila, homolog-like 5|forkhead-related activator 1|forkhead-related protein FKHL5|forkhead-related transcription factor 1', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86510527, end_position_on_the_genomic_accession=86514464, orientation='plus', exon_count=2, OMIM=601089, no_hdr0=''),
  2295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2295, CurrentID=0, Status='live', Symbol='FOXF2', Aliases='FKHL6, FREAC-2, FREAC2', description='forkhead box F2', other_designations='forkhead-like 6|forkhead-related activator 2|forkhead-related protein FKHL6|forkhead-related transcription factor 2', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1389834, end_position_on_the_genomic_accession=1395597, orientation='plus', exon_count=2, OMIM=603250, no_hdr0=''),
  2296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2296, CurrentID=0, Status='live', Symbol='FOXC1', Aliases='ARA, FKHL7, FREAC-3, FREAC3, IGDA, IHG1, IRID1, RIEG3', description='forkhead box C1', other_designations='forkhead box C1 protein|forkhead, drosophila, homolog-like 7|forkhead-related activator 3|forkhead-related protein FKHL7|forkhead-related transcription factor 3|forkhead/winged helix-like transcription factor 7|myeloid factor-delta', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1610446, end_position_on_the_genomic_accession=1613897, orientation='plus', exon_count=1, OMIM=601090, no_hdr0=''),
  2297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2297, CurrentID=0, Status='live', Symbol='FOXD1', Aliases='FKHL8, FREAC-4, FREAC4', description='forkhead box D1', other_designations='Forkhead, drosophila, homolog-like 8|forkhead-like 8|forkhead-related activator 4|forkhead-related protein FKHL8|forkhead-related transcription factor 4', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73446256, end_position_on_the_genomic_accession=73448527, orientation='minus', exon_count=1, OMIM=601091, no_hdr0=''),
  2298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2298, CurrentID=0, Status='live', Symbol='FOXD4', Aliases='FKHL9A, FREAC-5, FREAC5, FOXD4', description='forkhead box D4', other_designations='forkhead, Drosophila, homolog-like 9|forkhead-related activator 5|forkhead-related protein FKHL9|forkhead-related transcription factor 5|myeloid factor-alpha', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116231, end_position_on_the_genomic_accession=118417, orientation='minus', exon_count=1, OMIM=601092, no_hdr0=''),
  2299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2299, CurrentID=0, Status='live', Symbol='FOXI1', Aliases='FKH10, FKHL10, FREAC-6, FREAC6, HFH-3, HFH3', description='forkhead box I1', other_designations='HNF-3/fork-head homolog 3|HNF-3/fork-head homolog-3|forkhead-like 10|forkhead-related activator 6|forkhead-related protein FKHL10|forkhead-related transcription factor 6|hepatocyte nuclear factor 3 forkhead homolog 3', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170105897, end_position_on_the_genomic_accession=170109725, orientation='plus', exon_count=4, OMIM=601093, no_hdr0=''),
  2300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2300, CurrentID=0, Status='live', Symbol='FOXL1', Aliases='FKH6, FKHL11, FREAC7', description='forkhead box L1', other_designations='forkhead-like 11|forkhead-related protein FKHL11|forkhead-related transcription factor 7', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86578509, end_position_on_the_genomic_accession=86581698, orientation='plus', exon_count=1, OMIM=603252, no_hdr0=''),
  2301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2301, CurrentID=0, Status='live', Symbol='FOXE3', Aliases='FKHL12, FREAC8', description='forkhead box E3', other_designations='FREAC-8|forkhead, drosophila, homolog-like 12|forkhead-related protein FKHL12|forkhead-related transcription factor 8', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47416072, end_position_on_the_genomic_accession=47418052, orientation='plus', exon_count=1, OMIM=601094, no_hdr0=''),
  2302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2302, CurrentID=0, Status='live', Symbol='FOXJ1', Aliases='FKHL13, HFH-4, HFH4', description='forkhead box J1', other_designations='fork head homologue 4|forkhead transcription factor HFH-4|forkhead-like 13|forkhead-related protein FKHL13|hepatocyte nuclear factor 3 forkhead homolog 4', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76136333, end_position_on_the_genomic_accession=76141299, orientation='minus', exon_count=3, OMIM=602291, no_hdr0=''),
  2303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2303, CurrentID=0, Status='live', Symbol='FOXC2', Aliases='FKHL14, LD, MFH-1, MFH1', description='forkhead box C2', other_designations='MFH-1,mesenchyme forkhead 1|forkhead box C2 (MFH-1, mesenchyme forkhead 1)|forkhead, Drosophila, homolog-like 14|forkhead-related protein FKHL14|mesenchyme fork head protein 1|mesenchyme forkhead 1|transcription factor FKH-14', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86567251, end_position_on_the_genomic_accession=86568933, orientation='plus', exon_count=1, OMIM=602402, no_hdr0=''),
  2304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2304, CurrentID=0, Status='live', Symbol='FOXE1', Aliases='FKHL15, FOXE2, HFKH4, HFKL5, TITF2, TTF-2, TTF2', description='forkhead box E1', other_designations='HNF-3/fork head-like protein 5|forkhead box E2|forkhead box protein E2|forkhead, drosophila, homolog-like 15|forkhead-related protein FKHL15|thyroid transcription factor 2', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97853255, end_position_on_the_genomic_accession=97856715, orientation='plus', exon_count=1, OMIM=602617, no_hdr0=''),
  2305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2305, CurrentID=0, Status='live', Symbol='FOXM1', Aliases='FKHL16B, HFH-11, HFH11, HNF-3, INS-1, MPHOSPH2, MPP-2, MPP2, PIG29, TRIDENT, FOXM1', description='forkhead box M1', other_designations='Forkhead, drosophila, homolog-like 16|HNF-3/fork-head homolog 11|M-phase phosphoprotein 2|MPM-2 reactive phosphoprotein 2|forkhead-related protein FKHL16|hepatocyte nuclear factor 3 forkhead homolog 11|transcription factor Trident|winged-helix factor from INS-1 cells', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2857681, end_position_on_the_genomic_accession=2877155, orientation='minus', exon_count=11, OMIM=602341, no_hdr0=''),
  2306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2306, CurrentID=0, Status='live', Symbol='FOXD2', Aliases='FKHL17, FREAC-9, FREAC9', description='forkhead box D2', other_designations='forkhead, drosophila, homolog-like 17|forkhead-like 17|forkhead-related activator 9|forkhead-related protein FKHL17|forkhead-related transcription factor 9', map_location='1p34-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47436017, end_position_on_the_genomic_accession=47440691, orientation='plus', exon_count=1, OMIM=602211, no_hdr0=''),
  2307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2307, CurrentID=0, Status='live', Symbol='FOXS1', Aliases='FKHL18, FREAC10', description='forkhead box S1', other_designations='FREAC-10|forkhead-like 18 protein|forkhead-related activator 10|forkhead-related transcription factor 10|forkhead-related transcription factor FREAC-10', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31844300, end_position_on_the_genomic_accession=31845617, orientation='minus', exon_count=1, OMIM=602939, no_hdr0=''),
  2308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2308, CurrentID=0, Status='live', Symbol='FOXO1', Aliases='FKH1, FKHRA, FOXO1', description='forkhead box O1', other_designations='forkhead box protein O1A|forkhead, Drosophila, homolog of, in rhabdomyosarcoma', map_location='13q14.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40555664, end_position_on_the_genomic_accession=40666660, orientation='minus', exon_count=7, OMIM=136533, no_hdr0=''),
  2309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2309, CurrentID=0, Status='live', Symbol='FOXO3', Aliases='AF6q21, FKHRL1, FKHRL1P2, FOXO2A, FOXO3', description='forkhead box O3', other_designations='forkhead box O3A|forkhead homolog (rhabdomyosarcoma) like 1|forkhead in rhabdomyosarcoma-like 1|forkhead, Drosophila, homolog of, in rhabdomyosarcoma-like 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108559823, end_position_on_the_genomic_accession=108684769, orientation='plus', exon_count=9, OMIM=602681, no_hdr0=''),
  2310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2310, CurrentID=0, Status='live', Symbol='FOXO3B', Aliases='FKHRL1P1, ZNF286C', description='forkhead box O3B pseudogene', other_designations='forkhead homolog (rhabdomyosarcoma) like 1 pseudogene 1|forkhead homolog (rhabdomyosarcoma) pseudogene|forkhead(Drosophila) homolog (rhabdomyosarcoma) 1 like pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18665923, end_position_on_the_genomic_accession=18673181, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  133383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133383, CurrentID=0, Status='live', Symbol='SETD9', Aliases='C5orf35', description='SET domain containing 9', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56909259, end_position_on_the_genomic_accession=56925532, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  2312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2312, CurrentID=0, Status='live', Symbol='FLG', Aliases='ATOD2', description='filaggrin', other_designations='epidermal filaggrin', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152302163, end_position_on_the_genomic_accession=152325257, orientation='minus', exon_count=4, OMIM=135940, no_hdr0=''),
  2313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2313, CurrentID=0, Status='live', Symbol='FLI1', Aliases='EWSR2, SIC-1', description='Fli-1 proto-oncogene, ETS transcription factor', other_designations='Ewing sarcoma breakpoint region 2|Friend leukemia virus integration 1|transcription factor ERGB', map_location='11q24.1-q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128686535, end_position_on_the_genomic_accession=128813267, orientation='plus', exon_count=13, OMIM=193067, no_hdr0=''),
  2314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2314, CurrentID=0, Status='live', Symbol='FLII', Aliases='FLI, FLIL, Fli1', description='flightless I actin binding protein', other_designations='flightless I homolog', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18244815, end_position_on_the_genomic_accession=18258916, orientation='minus', exon_count=31, OMIM=600362, no_hdr0=''),
  2315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2315, CurrentID=0, Status='live', Symbol='MLANA', Aliases='MART-1, MART1', description='melan-A', other_designations='antigen LB39-AA|antigen SK29-AA|protein Melan-A', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5890909, end_position_on_the_genomic_accession=5909822, orientation='plus', exon_count=5, OMIM=605513, no_hdr0=''),
  2316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2316, CurrentID=0, Status='live', Symbol='FLNA', Aliases='ABP-280, ABPX, CSBS, CVD1, FLN, FLN-A, FLN1, FMD, MNS, NHBP, OPD, OPD1, OPD2, XLVD, XMVD', description='filamin A', other_designations='actin binding protein 280|alpha filamin|alpha-filamin|endothelial actin-binding protein|filamin-1|non-muscle filamin', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154348529, end_position_on_the_genomic_accession=154374638, orientation='minus', exon_count=48, OMIM=300017, no_hdr0=''),
  2317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2317, CurrentID=0, Status='live', Symbol='FLNB', Aliases='ABP-278, ABP-280, AOI, FH1, FLN-B, FLN1L, LRS1, SCT, TABP, TAP', description='filamin B', other_designations='ABP-280 homolog|Larsen syndrome 1 (autosomal dominant)|actin binding protein 278|actin-binding-like protein|beta-filamin|filamin homolog 1|filamin-3|thyroid autoantigen', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58008400, end_position_on_the_genomic_accession=58172255, orientation='plus', exon_count=48, OMIM=603381, no_hdr0=''),
  2318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2318, CurrentID=0, Status='live', Symbol='FLNC', Aliases='ABP-280, ABP280A, ABPA, ABPL, FLN2, MFM5, MPD4', description='filamin C', other_designations='ABP-280-like protein|ABP-L, gamma filamin|actin binding protein 280|filamin C, gamma|filamin-2|gamma filamin', map_location='7q32-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128830429, end_position_on_the_genomic_accession=128859274, orientation='plus', exon_count=48, OMIM=102565, no_hdr0=''),
  2319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2319, CurrentID=0, Status='live', Symbol='FLOT2', Aliases='ECS-1, ECS1, ESA, ESA1, M17S1', description='flotillin 2', other_designations='Flotillin 2 (epidermal surface antigen 1)|epidermal surface antigen|membrane component chromosome 17 surface marker 1|membrane component, chromosome 17, surface marker 1 (35kD protein identified by monoclonal antibody ECS-1)', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28879339, end_position_on_the_genomic_accession=28897697, orientation='minus', exon_count=12, OMIM=131560, no_hdr0=''),
  2321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2321, CurrentID=0, Status='live', Symbol='FLT1', Aliases='FLT, FLT-1, VEGFR-1, VEGFR1', description='fms-related tyrosine kinase 1', other_designations='fms-like tyrosine kinase 1|fms-related tyrosine kinase 1 (vascular endothelial growth factor/vascular permeability factor receptor)|tyrosine-protein kinase FRT|tyrosine-protein kinase receptor FLT|vascular permeability factor receptor', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28300346, end_position_on_the_genomic_accession=28495128, orientation='minus', exon_count=33, OMIM=165070, no_hdr0=''),
  2322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2322, CurrentID=0, Status='live', Symbol='FLT3', Aliases='CD135, FLK-2, FLK2, STK1', description='fms-related tyrosine kinase 3', other_designations='CD135 antigen|FL cytokine receptor|fetal liver kinase 2|fms-like tyrosine kinase 3|growth factor receptor tyrosine kinase type III|stem cell tyrosine kinase 1', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28003274, end_position_on_the_genomic_accession=28100592, orientation='minus', exon_count=27, OMIM=136351, no_hdr0=''),
  2323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2323, CurrentID=0, Status='live', Symbol='FLT3LG', Aliases='FL, FLT3L', description='fms-related tyrosine kinase 3 ligand', other_designations='flt3 ligand', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49474172, end_position_on_the_genomic_accession=49486234, orientation='plus', exon_count=12, OMIM=600007, no_hdr0=''),
  2324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2324, CurrentID=0, Status='live', Symbol='FLT4', Aliases='FLT-41, LMPH1A, PCL, VEGFR-3, VEGFR3, FLT4', description='fms-related tyrosine kinase 4', other_designations='fms-like tyrosine kinase 4|tyrosine-protein kinase receptor FLT4', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180601506, end_position_on_the_genomic_accession=180650271, orientation='minus', exon_count=34, OMIM=136352, no_hdr0=''),
  2326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2326, CurrentID=0, Status='live', Symbol='FMO1', Aliases='', description='flavin containing monooxygenase 1', other_designations='FMO 1|Flavin-containing monooxygenase 1 (fetal liver)|dimethylaniline oxidase 1|fetal hepatic flavin-containing monooxygenase 1', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171248471, end_position_on_the_genomic_accession=171285978, orientation='plus', exon_count=11, OMIM=136130, no_hdr0=''),
  2327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2327, CurrentID=0, Status='live', Symbol='FMO2', Aliases='FMO1B1', description='flavin containing monooxygenase 2', other_designations='FMO, pulmonary|dimethylaniline oxidase 2|flavin containing monooxygenase 2 (non-functional)|pulmonary flavin-containing monooxygenase 2', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171185208, end_position_on_the_genomic_accession=171212683, orientation='plus', exon_count=9, OMIM=603955, no_hdr0=''),
  2328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2328, CurrentID=0, Status='live', Symbol='FMO3', Aliases='FMOII, TMAU, dJ127D3.1', description='flavin containing monooxygenase 3', other_designations='FMO 3|FMO II|FMO form 2|dimethylaniline oxidase 3|hepatic flavin-containing monooxygenase 3|hepatic flavin-containing monooxygenase-3|trimethylamine monooxygenase', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171090877, end_position_on_the_genomic_accession=171117821, orientation='plus', exon_count=10, OMIM=136132, no_hdr0=''),
  2329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2329, CurrentID=0, Status='live', Symbol='FMO4', Aliases='FMO2', description='flavin containing monooxygenase 4', other_designations='FMO 4|dimethylaniline oxidase 4|hepatic flavin-containing monooxygenase 4', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171314137, end_position_on_the_genomic_accession=171342084, orientation='plus', exon_count=11, OMIM=136131, no_hdr0=''),
  2330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2330, CurrentID=0, Status='live', Symbol='FMO5', Aliases='', description='flavin containing monooxygenase 5', other_designations='FMO 5|dimethylaniline oxidase 5|hepatic flavin-containing monooxygenase 5', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147184300, end_position_on_the_genomic_accession=147225798, orientation='minus', exon_count=12, OMIM=603957, no_hdr0=''),
  2331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2331, CurrentID=0, Status='live', Symbol='FMOD', Aliases='FM, SLRR2E', description='fibromodulin', other_designations='KSPG fibromodulin|collagen-binding 59 kDa protein|keratan sulfate proteoglycan fibromodulin', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203340621, end_position_on_the_genomic_accession=203351429, orientation='minus', exon_count=3, OMIM=600245, no_hdr0=''),
  2332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2332, CurrentID=0, Status='live', Symbol='FMR1', Aliases='FMRP, FRAXA, POF, POF1', description='fragile X mental retardation 1', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147911951, end_position_on_the_genomic_accession=147951127, orientation='plus', exon_count=17, OMIM=309550, no_hdr0=''),
  2334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2334, CurrentID=0, Status='live', Symbol='AFF2', Aliases='FMR2, FMR2P, FRAXE, MRX2, OX19', description='AF4/FMR2 family member 2', other_designations='AF4/FMR2 family, member 2|fragile X E mental retardation syndrome protein|fragile X mental retardation 2 protein|protein FMR-2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=148500619, end_position_on_the_genomic_accession=149000663, orientation='plus', exon_count=22, OMIM=300806, no_hdr0=''),
  2335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2335, CurrentID=0, Status='live', Symbol='FN1', Aliases='CIG, ED-B, FINC, FN, FNZ, GFND, GFND2, LETS, MSF', description='fibronectin 1', other_designations='cold-insoluble globulin|migration-stimulating factor', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215360454, end_position_on_the_genomic_accession=215436167, orientation='minus', exon_count=47, OMIM=135600, no_hdr0=''),
  2339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2339, CurrentID=0, Status='live', Symbol='FNTA', Aliases='FPTA, PGGT1A, PTAR2', description='farnesyltransferase, CAAX box, alpha', other_designations='FTase-alpha|GGTase-I-alpha|farnesyl-protein transferase alpha-subunit|protein prenyltransferase alpha subunit repeat containing 2|ras proteins prenyltransferase subunit alpha|type I protein geranyl-geranyltransferase alpha subunit', map_location='8p11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43056299, end_position_on_the_genomic_accession=43085788, orientation='plus', exon_count=9, OMIM=134635, no_hdr0=''),
  2341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2341, CurrentID=0, Status='live', Symbol='FNTAP2', Aliases='FNTAL2', description='farnesyltransferase, CAAX box, alpha pseudogene 2', other_designations='farnesyltransferase, CAAX box, alpha-like 2', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21530686, end_position_on_the_genomic_accession=21531840, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2342, CurrentID=0, Status='live', Symbol='FNTB', Aliases='FPTB', description='farnesyltransferase, CAAX box, beta', other_designations='CAAX farnesyltransferase subunit beta|FTase-beta|ras proteins prenyltransferase subunit beta', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64986789, end_position_on_the_genomic_accession=65062652, orientation='plus', exon_count=12, OMIM=134636, no_hdr0=''),
  2346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2346, CurrentID=0, Status='live', Symbol='FOLH1', Aliases='FGCP, FOLH, GCP2, GCPII, NAALAD1, NAALAdase, PSM, PSMA, mGCP', description='folate hydrolase (prostate-specific membrane antigen) 1', other_designations='N-acetylated alpha-linked acidic dipeptidase 1|N-acetylated-alpha-linked acidic dipeptidase I|NAALADase I|cell growth-inhibiting gene 27 protein|folylpoly-gamma-glutamate carboxypeptidase|glutamate carboxylase II|glutamate carboxypeptidase II|membrane glutamate carboxypeptidase|prostate specific membrane antigen variant F|pteroylpoly-gamma-glutamate carboxypeptidase', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49146635, end_position_on_the_genomic_accession=49208670, orientation='minus', exon_count=20, OMIM=600934, no_hdr0=''),
  2348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2348, CurrentID=0, Status='live', Symbol='FOLR1', Aliases='FBP, FOLR', description='folate receptor 1 (adult)', other_designations='FR-alpha|KB cells FBP|adult folate-binding protein|folate binding protein|folate receptor, adult|ovarian tumor-associated antigen MOv18', map_location='11q13.3-q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72189558, end_position_on_the_genomic_accession=72196323, orientation='plus', exon_count=7, OMIM=136430, no_hdr0=''),
  2350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2350, CurrentID=0, Status='live', Symbol='FOLR2', Aliases='BETA-HFR, FBP, FBP/PL-1, FR-BETA, FR-P3', description='folate receptor 2 (fetal)', other_designations='folate receptor, beta|folate receptor, fetal/placental|folate-binding protein, fetal/placental|placental folate-binding protein', map_location='11q13.3-q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72216775, end_position_on_the_genomic_accession=72221950, orientation='plus', exon_count=6, OMIM=136425, no_hdr0=''),
  2352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2352, CurrentID=0, Status='live', Symbol='FOLR3', Aliases='FR-G, FR-gamma, gamma-hFR', description='folate receptor 3 (gamma)', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72135682, end_position_on_the_genomic_accession=72139892, orientation='plus', exon_count=6, OMIM=602469, no_hdr0=''),
  2353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2353, CurrentID=0, Status='live', Symbol='FOS', Aliases='AP-1, C-FOS, p55', description='FBJ murine osteosarcoma viral oncogene homolog', other_designations='FBJ murine osteosarcoma viral (v-fos) oncogene homolog (oncogene FOS)|G0/G1 switch regulatory protein 7|activator protein 1|cellular oncogene c-fos', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75278778, end_position_on_the_genomic_accession=75282234, orientation='plus', exon_count=4, OMIM=164810, no_hdr0=''),
  2354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2354, CurrentID=0, Status='live', Symbol='FOSB', Aliases='AP-1, G0S3, GOS3, GOSB', description='FBJ murine osteosarcoma viral oncogene homolog B', other_designations='G0/G1 switch regulatory protein 3|activator protein 1', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45467995, end_position_on_the_genomic_accession=45475179, orientation='plus', exon_count=4, OMIM=164772, no_hdr0=''),
  2355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2355, CurrentID=0, Status='live', Symbol='FOSL2', Aliases='FRA2', description='FOS-like antigen 2', other_designations='FRA-2', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28392759, end_position_on_the_genomic_accession=28417312, orientation='plus', exon_count=6, OMIM=601575, no_hdr0=''),
  2356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2356, CurrentID=0, Status='live', Symbol='FPGS', Aliases='', description='folylpolyglutamate synthase', other_designations='folylpoly-gamma-glutamate synthetase|tetrahydrofolate synthase|tetrahydrofolylpolyglutamate synthase', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127802858, end_position_on_the_genomic_accession=127814520, orientation='plus', exon_count=17, OMIM=136510, no_hdr0=''),
  2357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2357, CurrentID=0, Status='live', Symbol='FPR1', Aliases='FMLP, FPR', description='formyl peptide receptor 1', other_designations='N-formylpeptide chemoattractant receptor|fMLP receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51745770, end_position_on_the_genomic_accession=51751897, orientation='minus', exon_count=3, OMIM=136537, no_hdr0=''),
  2358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2358, CurrentID=0, Status='live', Symbol='FPR2', Aliases='ALXR, FMLP-R-II, FMLPXA, FPRH1, FPRH2, FPRL1, HM63, LXA4R, FPR2', description='formyl peptide receptor 2', other_designations='FMLP-R-I|FMLP-related receptor I|LXA4 receptor|RFP|formyl peptide receptor-like 1|lipoxin A4 receptor (formyl peptide receptor related)', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51760920, end_position_on_the_genomic_accession=51770670, orientation='plus', exon_count=3, OMIM=136538, no_hdr0=''),
  2359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2359, CurrentID=0, Status='live', Symbol='FPR3', Aliases='FML2_HUMAN, FMLP-R-II, FMLPY, FPRH1, FPRH2, FPRL2, RMLP-R-I', description='formyl peptide receptor 3', other_designations='FMLP-related receptor II|formyl peptide receptor-like 2', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51795158, end_position_on_the_genomic_accession=51826474, orientation='plus', exon_count=3, OMIM=136539, no_hdr0=''),
  2360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2360, CurrentID=0, Status='live', Symbol='FRA1A', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p36)', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2361, CurrentID=0, Status='live', Symbol='FRA1B', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p32)', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2362, CurrentID=0, Status='live', Symbol='FRA1C', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p31.2)', other_designations='', map_location='1p31.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2363, CurrentID=0, Status='live', Symbol='FRA1D', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p22)', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2364, CurrentID=0, Status='live', Symbol='FRA1E', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p21.2)', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534589, CurrentID=0, Status='live', Symbol='HOXA10-HOXA9', Aliases='', description='HOXA10-HOXA9 readthrough', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27162438, end_position_on_the_genomic_accession=27180261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2366, CurrentID=0, Status='live', Symbol='FRA1G', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(q25.1)', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2367, CurrentID=0, Status='live', Symbol='FRA1H', Aliases='', description='fragile site, 5-azacytidine type, common, fra(1)(q42)', other_designations='', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534592, CurrentID=0, Status='live', Symbol='URGCP-MRPS24', Aliases='', description='URGCP-MRPS24 readthrough', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43866558, end_position_on_the_genomic_accession=43906632, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100534593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534593, CurrentID=0, Status='live', Symbol='STX16-NPEPL1', Aliases='', description='STX16-NPEPL1 readthrough (NMD candidate)', other_designations='STX16-NPEPL1 readthrough (non-protein coding)', map_location='20q', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58651253, end_position_on_the_genomic_accession=58715844, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  2370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2370, CurrentID=0, Status='live', Symbol='FRA1K', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(q31)', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534595, CurrentID=0, Status='live', Symbol='HNRNPUL2-BSCL2', Aliases='', description='HNRNPUL2-BSCL2 readthrough (NMD candidate)', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62690262, end_position_on_the_genomic_accession=62727384, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  2372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2372, CurrentID=0, Status='live', Symbol='FRA1M', Aliases='', description='fragile site, folic acid type, rare, fra(1)(p21.3)', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2373, CurrentID=0, Status='live', Symbol='FRA2A', Aliases='', description='fragile site, folic acid type, rare, fra(2)(q11.2)', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136610, no_hdr0=''),
  2374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2374, CurrentID=0, Status='live', Symbol='FRA2B', Aliases='', description='fragile site, folic acid type, rare, fra(2)(q13)', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534599, CurrentID=0, Status='live', Symbol='ISY1-RAB43', Aliases='', description='ISY1-RAB43 readthrough', other_designations='', map_location='3q', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129087569, end_position_on_the_genomic_accession=129161230, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  2376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2376, CurrentID=0, Status='live', Symbol='FRA2D', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(p16.2)', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2377, CurrentID=0, Status='live', Symbol='FRA2E', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(p13)', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2378, CurrentID=0, Status='live', Symbol='FRA2F', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(q21.3)', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2379, CurrentID=0, Status='live', Symbol='FRA2G', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(q31)', other_designations='', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2380, CurrentID=0, Status='live', Symbol='FRA2H', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(q32.1)', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2381, CurrentID=0, Status='live', Symbol='FRA2I', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(q33)', other_designations='', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2382, CurrentID=0, Status='live', Symbol='FRA2J', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(q37.3)', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2383, CurrentID=0, Status='live', Symbol='FRA2K', Aliases='', description='fragile site, folic acid type, rare, fra(2)(q22.3)', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2384, CurrentID=0, Status='live', Symbol='FRA3A', Aliases='', description='fragile site, aphidicolin type, common, fra(3)(p24.2)', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2386, CurrentID=0, Status='live', Symbol='FRA3C', Aliases='', description='fragile site, aphidicolin type, common, fra(3)(q27)', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534611, CurrentID=0, Status='live', Symbol='TM4SF19-TCTEX1D2', Aliases='', description='TM4SF19-TCTEX1D2 readthrough (NMD candidate)', other_designations='', map_location='3q', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196316085, end_position_on_the_genomic_accession=196338420, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100534612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534612, CurrentID=0, Status='live', Symbol='C1QTNF3-AMACR', Aliases='', description='C1QTNF3-AMACR readthrough (NMD candidate)', other_designations='', map_location='5p', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33986986, end_position_on_the_genomic_accession=34124528, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2389, CurrentID=0, Status='live', Symbol='FRA4B', Aliases='', description='fragile site, BrdU type, common, fra(4)(q12)', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2390, CurrentID=0, Status='live', Symbol='FRA4C', Aliases='', description='fragile site, aphidicolin type, common, fra(4)(q31.1)', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2391, CurrentID=0, Status='live', Symbol='FRA4D', Aliases='', description='fragile site, aphidicolin type, common, fra(4)(p15)', other_designations='', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2393, CurrentID=0, Status='live', Symbol='FRA5A', Aliases='', description='fragile site, BrdU type, common, fra(5)(p13)', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2394, CurrentID=0, Status='live', Symbol='FRA5B', Aliases='', description='fragile site, BrdU type, common, fra(5)(q15)', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2395, CurrentID=0, Status='live', Symbol='FXN', Aliases='CyaY, FA, FARR, FRDA, X25', description='frataxin', other_designations='Friedreich ataxia protein', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69035563, end_position_on_the_genomic_accession=69100178, orientation='plus', exon_count=6, OMIM=606829, no_hdr0=''),
  2396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2396, CurrentID=0, Status='live', Symbol='FRA5C', Aliases='', description='fragile site, aphidicolin type, common, fra(5)(q31.1)', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2397, CurrentID=0, Status='live', Symbol='FRA5D', Aliases='', description='fragile site, aphidicolin type, common, fra(5)(q15)', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2398, CurrentID=0, Status='live', Symbol='FRA5E', Aliases='', description='fragile site, aphidicolin type, common, fra(5)(p14)', other_designations='', map_location='5p14', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2399, CurrentID=0, Status='live', Symbol='FRA5F', Aliases='', description='fragile site, aphidicolin type, common, fra(5)(q21)', other_designations='', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534624, CurrentID=0, Status='live', Symbol='LCR-OPSIN', Aliases='OPN1C', description='opsin locus control region', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300824, no_hdr0=''),
  2401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2401, CurrentID=0, Status='live', Symbol='FRA6A', Aliases='', description='fragile site, folic acid type, rare, fra(6)(p23)', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2402, CurrentID=0, Status='live', Symbol='FRA6B', Aliases='', description='fragile site, aphidicolin type, common, fra(6)(p25.1)', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2403, CurrentID=0, Status='live', Symbol='FRA6C', Aliases='', description='fragile site, aphidicolin type, common, fra(6)(p22.2)', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2404, CurrentID=0, Status='live', Symbol='FRA6D', Aliases='', description='fragile site, BrdU type, common, fra(6)(q13)', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2405, CurrentID=0, Status='live', Symbol='FRA6E', Aliases='', description='fragile site, aphidicolin type, common, fra(6)(q26)', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2406, CurrentID=0, Status='live', Symbol='FRA6F', Aliases='', description='fragile site, aphidicolin type, common, fra(6)(q21)', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2407, CurrentID=0, Status='live', Symbol='FRA6G', Aliases='', description='fragile site, aphidicolin type, common, fra(6)(q15)', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2408, CurrentID=0, Status='live', Symbol='FRA7A', Aliases='', description='fragile site, folic acid type, rare, fra(7)(p11.2)', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2409, CurrentID=0, Status='live', Symbol='FRA7B', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(p22)', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  133482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133482, CurrentID=0, Status='live', Symbol='SLCO6A1', Aliases='CT48, GST, OATP-I, OATP6A1, OATPY', description='solute carrier organic anion transporter family member 6A1', other_designations='cancer/testis antigen 48|gonad-specific transporter|organic anion-transporting polypeptide I|solute carrier family 21 member 19|solute carrier organic anion transporter family, member 6A1|testis-specific organic anion transporter', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102371782, end_position_on_the_genomic_accession=102499656, orientation='minus', exon_count=14, OMIM=613365, no_hdr0=''),
  2411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2411, CurrentID=0, Status='live', Symbol='FRA7D', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(p13)', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2412, CurrentID=0, Status='live', Symbol='FRA7E', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(q21.2)', other_designations='', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  131474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131474, CurrentID=0, Status='live', Symbol='CHCHD4', Aliases='MIA40, TIMM40', description='coiled-coil-helix-coiled-coil-helix domain containing 4', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 4|mitochondrial intermembrane space import and assembly 40 homolog|translocase of inner mitochondrial membrane 40 homolog', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14112077, end_position_on_the_genomic_accession=14124871, orientation='minus', exon_count=4, OMIM=611077, no_hdr0=''),
  2414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2414, CurrentID=0, Status='live', Symbol='FRA7G', Aliases='D7S486', description='fragile site, aphidicolin type, common, fra(7)(q31.2)', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2415, CurrentID=0, Status='live', Symbol='FRA7H', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(q32.3)', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2416, CurrentID=0, Status='live', Symbol='FRA7I', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(q36)', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2417, CurrentID=0, Status='live', Symbol='FRA7J', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(q11)', other_designations='', map_location='7q11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2418, CurrentID=0, Status='live', Symbol='FRA8A', Aliases='', description='fragile site, folic acid type, rare, fra(8)(q22.3)', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  133491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133491, CurrentID=0, Status='live', Symbol='C5orf47', Aliases='', description='chromosome 5 open reading frame 47', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173989159, end_position_on_the_genomic_accession=174006140, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2420, CurrentID=0, Status='live', Symbol='FRDA2', Aliases='', description='Friedreich ataxia 2', other_designations='', map_location='9p23-p11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601992, no_hdr0=''),
  2421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2421, CurrentID=0, Status='live', Symbol='FRA8C', Aliases='', description='fragile site, aphidicolin type, common, fra(8)(q24.1)', other_designations='', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2422, CurrentID=0, Status='live', Symbol='FRA8D', Aliases='', description='fragile site, aphidicolin type, common, fra(8)(q24.3)', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2423, CurrentID=0, Status='live', Symbol='FRA8E', Aliases='', description='fragile site, distamycin A type, rare, fra(8)(q24.1)', other_designations='', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2425, CurrentID=0, Status='live', Symbol='FRA9A', Aliases='', description='fragile site, folic acid type, rare, fra(9)(p21)', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2426, CurrentID=0, Status='live', Symbol='FRA9B', Aliases='', description='fragile site, folic acid type, rare, fra(9)(q32)', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136640, no_hdr0=''),
  2427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2427, CurrentID=0, Status='live', Symbol='FRA9C', Aliases='', description='fragile site, BrdU type, common, fra(9)(p21)', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2428, CurrentID=0, Status='live', Symbol='FRA9D', Aliases='', description='fragile site, aphidicolin type, common, fra(9)(q22.1)', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2429, CurrentID=0, Status='live', Symbol='FRA9E', Aliases='', description='fragile site, aphidicolin type, common, fra(9)(q32)', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2430, CurrentID=0, Status='live', Symbol='FRA9F', Aliases='', description='fragile site, 5-azacytidine type, common, fra(9)(q12)', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2431, CurrentID=0, Status='live', Symbol='FRA10A', Aliases='', description='fragile site, folic acid type, rare, fra(10)(q23.3) or fra(10)(q24.2)', other_designations='', map_location='10q23.3 or 10q24.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136620, no_hdr0=''),
  2432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2432, CurrentID=0, Status='live', Symbol='FRA10B', Aliases='', description='fragile site, BrdU type, rare, fra(10)(q25.2)', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2433, CurrentID=0, Status='live', Symbol='FRA10C', Aliases='', description='fragile site, BrdU type, common, fra(10)(q21)', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102238594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102238594, CurrentID=0, Status='live', Symbol='SPACA6P-AS', Aliases='LINC01129', description='SPACA6P antisense RNA', other_designations='long intergenic non-protein coding RNA 1129', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51685363, end_position_on_the_genomic_accession=51693456, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2435, CurrentID=0, Status='live', Symbol='FRA10E', Aliases='', description='fragile site, aphidicolin type, common, fra(10)(q25.2)', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2436, CurrentID=0, Status='live', Symbol='FRA10F', Aliases='', description='fragile site, aphidicolin type, common, fra(10)(q26.1)', other_designations='', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2437, CurrentID=0, Status='live', Symbol='FRA10G', Aliases='', description='fragile site, aphidicolin type, common, fra(10)(q11.2)', other_designations='', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2438, CurrentID=0, Status='live', Symbol='FRA11A', Aliases='', description='fragile site, folic acid type, rare, fra(11)(q13.3)', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2440, CurrentID=0, Status='live', Symbol='FRA11C', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(p15.1)', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2441, CurrentID=0, Status='live', Symbol='FRA11D', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(p14.2)', other_designations='', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2442, CurrentID=0, Status='live', Symbol='FRA11E', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(p13)', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2443, CurrentID=0, Status='live', Symbol='FRA11F', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(q14.2)', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2444, CurrentID=0, Status='live', Symbol='FRK', Aliases='GTK, PTK5, RAK', description='fyn-related Src family tyrosine kinase', other_designations='PTK5 protein tyrosine kinase 5|fyn-related kinase|nuclear tyrosine protein kinase RAK|protein-tyrosine kinase 5', map_location='6q21-q22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=115941528, end_position_on_the_genomic_accession=116100728, orientation='minus', exon_count=11, OMIM=606573, no_hdr0=''),
  2445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2445, CurrentID=0, Status='live', Symbol='FRA11G', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(q23.3)', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2446, CurrentID=0, Status='live', Symbol='FRA11H', Aliases='', description='fragile site, aphidicolin type, common, fra(11)(q13)', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2447, CurrentID=0, Status='live', Symbol='FRA11I', Aliases='', description='fragile site, distamycin A type, rare, fra(11)(p15.1)', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2448, CurrentID=0, Status='live', Symbol='FRA12A', Aliases='', description='fragile site, folic acid type, rare, fra(12)(q13.1)', other_designations='', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2449, CurrentID=0, Status='live', Symbol='FRA12B', Aliases='', description='fragile site, aphidicolin type, common, fra(12)(q21.3)', other_designations='', map_location='12q21.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  133522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133522, CurrentID=0, Status='live', Symbol='PPARGC1B', Aliases='ERRL1, PERC, PGC-1(beta), PGC1B', description='peroxisome proliferator-activated receptor gamma, coactivator 1 beta', other_designations='PGC-1-related estrogen receptor alpha coactivator|PPAR-gamma coactivator 1-beta|PPARGC-1-beta', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149730252, end_position_on_the_genomic_accession=149857874, orientation='plus', exon_count=16, OMIM=608886, no_hdr0=''),
  2451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2451, CurrentID=0, Status='live', Symbol='FRA12D', Aliases='', description='fragile site, folic acid type, rare, fra(12)(q24.13)', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2452, CurrentID=0, Status='live', Symbol='FRA12E', Aliases='', description='fragile site, aphidicolin type, common, fra(12)(q24)', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2453, CurrentID=0, Status='live', Symbol='FRA13A', Aliases='', description='fragile site, aphidicolin type, common, fra(13)(q13.2)', other_designations='', map_location='13q13.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2454, CurrentID=0, Status='live', Symbol='FRA13B', Aliases='', description='fragile site, BrdU type, common, fra(13)(q21)', other_designations='', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2455, CurrentID=0, Status='live', Symbol='FRA13C', Aliases='', description='fragile site, aphidicolin type, common, fra(13)(q21.2)', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2456, CurrentID=0, Status='live', Symbol='FRA13D', Aliases='', description='fragile site, aphidicolin type, common, fra(13)(q32)', other_designations='', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2457, CurrentID=0, Status='live', Symbol='FRA14B', Aliases='', description='fragile site, aphidicolin type, common, fra(14)(q23)', other_designations='', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2458, CurrentID=0, Status='live', Symbol='FRA14C', Aliases='', description='fragile site, aphidicolin type, common, fra(14)(q24.1)', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2459, CurrentID=0, Status='live', Symbol='FRA15A', Aliases='', description='fragile site, aphidicolin type, common, fra(15)(q22)', other_designations='', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2460, CurrentID=0, Status='live', Symbol='FRA16A', Aliases='', description='fragile site, folic acid type, rare, fra(16)(p13.11)', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2461, CurrentID=0, Status='live', Symbol='FRA16B', Aliases='', description='fragile site, distamycin A type, rare, fra(16)(q22.1)', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136580, no_hdr0=''),
  2462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2462, CurrentID=0, Status='live', Symbol='FRA16C', Aliases='', description='fragile site, aphidicolin type, common, fra(16)(q22.1)', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2463, CurrentID=0, Status='live', Symbol='FRA16D', Aliases='', description='fragile site, aphidicolin type, common, fra(16)(q23.2)', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2464, CurrentID=0, Status='live', Symbol='FRA16E', Aliases='C16DELp12.1', description='fragile site, distamycin A type, rare, fra(16)(p12.1)', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136570, no_hdr0=''),
  2465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2465, CurrentID=0, Status='live', Symbol='FRA17A', Aliases='', description='fragile site, distamycin A type, rare, fra(17)(p12)', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136660, no_hdr0=''),
  2466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2466, CurrentID=0, Status='live', Symbol='FRA17B', Aliases='', description='fragile site, aphidicolin type, common, fra(17)(q23.1)', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2467, CurrentID=0, Status='live', Symbol='FRA18A', Aliases='', description='fragile site, aphidicolin type, common, fra(18)(q12.2)', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2468, CurrentID=0, Status='live', Symbol='FRA18B', Aliases='', description='fragile site, aphidicolin type, common, fra(18)(q21.3)', other_designations='', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2469, CurrentID=0, Status='live', Symbol='FRA19A', Aliases='', description='fragile site, 5-azacytidine type, common, fra(19)(q13)', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2470, CurrentID=0, Status='live', Symbol='FRA19B', Aliases='', description='fragile site, folic acid type, rare, fra(19)(p13)', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2471, CurrentID=0, Status='live', Symbol='FRA20A', Aliases='', description='fragile site, folic acid type, rare, fra(20)(p11.23)', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136590, no_hdr0=''),
  2472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2472, CurrentID=0, Status='live', Symbol='FRA20B', Aliases='', description='fragile site, aphidicolin type, common, fra(20)(p12.2)', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2473, CurrentID=0, Status='live', Symbol='FRA22A', Aliases='', description='fragile site, folic acid type, rare, fra(22)(q13)', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2474, CurrentID=0, Status='live', Symbol='FRA22B', Aliases='', description='fragile site, aphidicolin type, common, fra(22)(q12.2)', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2475, CurrentID=0, Status='live', Symbol='MTOR', Aliases='FRAP, FRAP1, FRAP2, RAFT1, RAPT1, SKS', description='mechanistic target of rapamycin (serine/threonine kinase)', other_designations='FK506 binding protein 12-rapamycin associated protein 2|FK506-binding protein 12-rapamycin complex-associated protein 1|FKBP-rapamycin associated protein|FKBP12-rapamycin complex-associated protein 1|mammalian target of rapamycin|rapamycin and FKBP12 target 1|rapamycin associated protein FRAP2|rapamycin target protein 1', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11106531, end_position_on_the_genomic_accession=11262557, orientation='minus', exon_count=59, OMIM=601231, no_hdr0=''),
  2477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2477, CurrentID=0, Status='live', Symbol='FRAXA', Aliases='FMR1', description='fragile site, folic acid type, rare, fra(X)(q27.3) A (macroorchidism, mental retardation)', other_designations='fragile site, folic acid type, rare, fra(X)(q27.3), (macroorchidism, mental retardation)', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2478, CurrentID=0, Status='live', Symbol='FRAXB', Aliases='', description='fragile site, aphidicolin type, common, fra(X)(p22.31) B', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2479, CurrentID=0, Status='live', Symbol='FRAXC', Aliases='', description='fragile site, aphidicolin type, common, fra(X)(q22.1) C', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2480, CurrentID=0, Status='live', Symbol='FRAXD', Aliases='', description='fragile site, aphidicolin type, common, fra(X)(q27.2) D', other_designations='', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2481, CurrentID=0, Status='live', Symbol='FRAXE', Aliases='FMR2', description='fragile site, folic acid type, rare, fra(X)(q28) E', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2483, CurrentID=0, Status='live', Symbol='FRG1', Aliases='FRG1A, FSG1', description='FSHD region gene 1', other_designations='FSHD region gene 1 protein|facioscapulohumeral muscular dystrophy region gene-1', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189940819, end_position_on_the_genomic_accession=189963204, orientation='plus', exon_count=10, OMIM=601278, no_hdr0=''),
  133558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133558, CurrentID=0, Status='live', Symbol='MROH2B', Aliases='HEATR7B2', description='maestro heat-like repeat family member 2B', other_designations='HEAT repeat family member 7B2|HEAT repeat family member B2|HEAT repeat-containing protein 7B2', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40998020, end_position_on_the_genomic_accession=41071342, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  2487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2487, CurrentID=0, Status='live', Symbol='FRZB', Aliases='FRE, FRITZ, FRP-3-1, FRZB-PEN, FRZB1, FZRB, OS1, SFRP3, SRFP3, hFIZ, FRZB', description='frizzled-related protein', other_designations='frezzled|frizzled homolog-related|frizzled-related protein 1|sFRP-3', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182833275, end_position_on_the_genomic_accession=182866770, orientation='minus', exon_count=6, OMIM=605083, no_hdr0=''),
  2488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2488, CurrentID=0, Status='live', Symbol='FSHB', Aliases='HH24', description='follicle stimulating hormone, beta polypeptide', other_designations='FSH-B|FSH-beta|follicle-stimulating hormone beta subunit|follitropin beta chain|follitropin, beta chain', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30230688, end_position_on_the_genomic_accession=30235277, orientation='plus', exon_count=4, OMIM=136530, no_hdr0=''),
  2489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2489, CurrentID=0, Status='live', Symbol='FSHMD1A', Aliases='FMD, FSHD, FSHD1A, FSHMD', description='facioscapulohumeral muscular dystrophy 1A', other_designations='', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=158900, no_hdr0=''),
  2490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2490, CurrentID=0, Status='live', Symbol='FSHMD1B', Aliases='FSHD1B', description='facioscapulohumeral muscular dystrophy 1B', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=158901, no_hdr0=''),
  2491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2491, CurrentID=0, Status='live', Symbol='CENPI', Aliases='CENP-I, FSHPRH1, LRPR1, Mis6', description='centromere protein I', other_designations='FSH primary response 1|FSH primary response protein 1|follicle-stimulating hormone primary response protein|interphase centromere complex protein 19|leucine-rich primary response protein 1', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101098159, end_position_on_the_genomic_accession=101163678, orientation='plus', exon_count=25, OMIM=300065, no_hdr0=''),
  2492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2492, CurrentID=0, Status='live', Symbol='FSHR', Aliases='FSHR1O, LGR1, ODG1, FSHR', description='follicle stimulating hormone receptor', other_designations='FSH receptor|follitropin receptor', map_location='2p21-p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48866435, end_position_on_the_genomic_accession=49154527, orientation='minus', exon_count=17, OMIM=136435, no_hdr0=''),
  2494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2494, CurrentID=0, Status='live', Symbol='NR5A2', Aliases='B1F, B1F2, CPF, FTF, FTZ-F1, FTZ-F1beta, LRH-1, LRH1, hB1F-2', description='nuclear receptor subfamily 5 group A member 2', other_designations='CYP7A promoter-binding factor|alpha-1-fetoprotein transcription factor|b1-binding factor, hepatocyte transcription factor which activates enhancer II of hepatitis B virus|fetoprotein-alpha 1 (AFP) transcription factor|hepatocytic transcription factor hB1F-3|liver nuclear receptor homolog-1 variant 2|liver receptor homolog 1|liver receptor homolog-1|nuclear receptor NR5A2|nuclear receptor subfamily 5, group A, member 2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200027602, end_position_on_the_genomic_accession=200177424, orientation='plus', exon_count=13, OMIM=604453, no_hdr0=''),
  2495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2495, CurrentID=0, Status='live', Symbol='FTH1', Aliases='FHC, FTH, FTHL6, HFE5, PIG15, PLIF', description='ferritin, heavy polypeptide 1', other_designations='apoferritin|cell proliferation-inducing gene 15 protein|ferritin H subunit|placenta immunoregulatory factor|proliferation-inducing protein 15', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61964285, end_position_on_the_genomic_accession=61967660, orientation='minus', exon_count=4, OMIM=134770, no_hdr0=''),
  2496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2496, CurrentID=0, Status='live', Symbol='FTH1P1', Aliases='FTHL1', description='ferritin, heavy polypeptide 1 pseudogene', other_designations='ferritin, heavy polypeptide-like 1 pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37544563, end_position_on_the_genomic_accession=37545458, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2497, CurrentID=0, Status='live', Symbol='FTH1P2', Aliases='FTHL2', description='ferritin, heavy polypeptide 1 pseudogene 2', other_designations='ferritin, heavy polypeptide-like 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228687417, end_position_on_the_genomic_accession=228687879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2498, CurrentID=0, Status='live', Symbol='FTH1P3', Aliases='FTHL3, FTHL3P', description='ferritin, heavy polypeptide 1 pseudogene 3', other_designations='ferritin, heavy polypeptide-like 3 pseudogene', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27392623, end_position_on_the_genomic_accession=27393576, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2499, CurrentID=0, Status='live', Symbol='FTH1P4', Aliases='FTHL4', description='ferritin, heavy polypeptide 1 pseudogene 4', other_designations='ferritin, heavy polypeptide-like 4 pseudogene', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128764306, end_position_on_the_genomic_accession=128765221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2500, CurrentID=0, Status='live', Symbol='FTH1P7', Aliases='FTHL7', description='ferritin, heavy polypeptide 1 pseudogene 7', other_designations='ferritin, heavy polypeptide-like 7', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22695862, end_position_on_the_genomic_accession=22696791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2501, CurrentID=0, Status='live', Symbol='FTH1P8', Aliases='FTHL8', description='ferritin, heavy polypeptide 1 pseudogene 8', other_designations='ferritin, heavy polypeptide-like 8', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=148052012, end_position_on_the_genomic_accession=148052923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2502, CurrentID=0, Status='live', Symbol='FTH1P10', Aliases='FTHL10', description='ferritin, heavy polypeptide 1 pseudogene 10', other_designations='ferritin, heavy polypeptide-like 10', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17353713, end_position_on_the_genomic_accession=17354631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2503, CurrentID=0, Status='live', Symbol='FTH1P11', Aliases='FTHL11', description='ferritin, heavy polypeptide 1 pseudogene 11', other_designations='ferritin, heavy polypeptide-like 11', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81521520, end_position_on_the_genomic_accession=81522433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2504, CurrentID=0, Status='live', Symbol='FTH1P12', Aliases='FTHL12', description='ferritin, heavy polypeptide 1 pseudogene 12', other_designations='ferritin, heavy polypeptide-like 12', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15526972, end_position_on_the_genomic_accession=15527886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2507, CurrentID=0, Status='live', Symbol='FTH1P15', Aliases='FTHL15', description='ferritin, heavy polypeptide 1 pseudogene 15', other_designations='ferritin, heavy polypeptide-like 15', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57004347, end_position_on_the_genomic_accession=57004802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2508, CurrentID=0, Status='live', Symbol='FTH1P16', Aliases='FTH2, FTHL16', description='ferritin, heavy polypeptide 1 pseudogene 16', other_designations='ferritin, heavy polypeptide-like 16', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77734315, end_position_on_the_genomic_accession=77735232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2509, CurrentID=0, Status='live', Symbol='FTH1P5', Aliases='FTHL5, FTHP1', description='ferritin, heavy polypeptide 1 pseudogene 5', other_designations='ferritin, heavy polypeptide pseudogene 1|ferritin, heavy polypeptide-like 5', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50912550, end_position_on_the_genomic_accession=50913462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2510, CurrentID=0, Status='live', Symbol='FTH1P21', Aliases='FTHL21, FTHP2', description='ferritin, heavy polypeptide 1 pseudogene 21', other_designations='ferritin, heavy polypeptide pseudogene 2|ferritin, heavy polypeptide-like 21', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=156006256, end_position_on_the_genomic_accession=156007180, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2512, CurrentID=0, Status='live', Symbol='FTL', Aliases='LFTD, NBIA3', description='ferritin, light polypeptide', other_designations='ferritin L subunit|ferritin L-chain|ferritin light polypeptide-like 3', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48965309, end_position_on_the_genomic_accession=48966879, orientation='plus', exon_count=4, OMIM=134790, no_hdr0=''),
  2515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2515, CurrentID=0, Status='live', Symbol='ADAM2', Aliases='CRYN1, CRYN2, CT15, FTNB, PH-30b, PH30, PH30-beta', description='ADAM metallopeptidase domain 2', other_designations='cancer/testis antigen 15|fertilin subunit beta', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39743735, end_position_on_the_genomic_accession=39838289, orientation='minus', exon_count=21, OMIM=601533, no_hdr0=''),
  2516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2516, CurrentID=0, Status='live', Symbol='NR5A1', Aliases='AD4BP, ELP, FTZ1, FTZF1, POF7, SF-1, SF1, SPGF8, SRXY3', description='nuclear receptor subfamily 5 group A member 1', other_designations='STF-1|adrenal 4 binding protein|adrenal 4-binding protein|fushi tarazu factor homolog 1|nuclear receptor AdBP4|nuclear receptor subfamily 5, group A, member 1|steroid hormone receptor Ad4BP|steroidogenic factor 1 nuclear receptor|steroidogenic factor-1', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124481232, end_position_on_the_genomic_accession=124507448, orientation='minus', exon_count=9, OMIM=184757, no_hdr0=''),
  2517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2517, CurrentID=0, Status='live', Symbol='FUCA1', Aliases='FUCA', description='fucosidase, alpha-L- 1, tissue', other_designations='alpha-L-fucosidase 1|alpha-L-fucosidase I|alpha-L-fucoside fucohydrolase 1', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23845077, end_position_on_the_genomic_accession=23868369, orientation='minus', exon_count=9, OMIM=612280, no_hdr0=''),
  2518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2518, CurrentID=0, Status='live', Symbol='FUCA1P1', Aliases='FUCA1P', description='fucosidase, alpha-L- 1, tissue pseudogene 1', other_designations='', map_location='2q31-q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176804455, end_position_on_the_genomic_accession=176805939, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2519, CurrentID=0, Status='live', Symbol='FUCA2', Aliases='dJ20N2.5', description='fucosidase, alpha-L- 2, plasma', other_designations='alpha-L-fucosidase 2|alpha-L-fucoside fucohydrolase 2', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143494812, end_position_on_the_genomic_accession=143511883, orientation='minus', exon_count=7, OMIM=136820, no_hdr0=''),
  2520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2520, CurrentID=0, Status='live', Symbol='GAST', Aliases='GAS', description='gastrin', other_designations='preprogastrin', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41712326, end_position_on_the_genomic_accession=41715969, orientation='plus', exon_count=3, OMIM=137250, no_hdr0=''),
  2521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2521, CurrentID=0, Status='live', Symbol='FUS', Aliases='ALS6, ETM41, HNRNPP2, POMP75, TLS, FUS', description='FUS RNA binding protein', other_designations='75 kDa DNA-pairing protein|fus-like protein|fused in sarcoma|fusion gene in myxoid liposarcoma|heterogeneous nuclear ribonucleoprotein P2|oncogene FUS|oncogene TLS|translocated in liposarcoma protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31179704, end_position_on_the_genomic_accession=31194871, orientation='plus', exon_count=15, OMIM=137070, no_hdr0=''),
  2522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2522, CurrentID=0, Status='live', Symbol='FUSE', Aliases='', description='polykaryocytosis promoter', other_designations='Polykaryocytosis inducer', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=174750, no_hdr0=''),
  2523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2523, CurrentID=0, Status='live', Symbol='FUT1', Aliases='H, HH, HSC', description='fucosyltransferase 1 (H blood group)', other_designations='2-alpha-L-fucosyltransferase|GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase 1|alpha (1,2) fucosyltransferase|alpha(1,2) fucosyltransferase 1|alpha(1,2)FT 1|blood group H alpha 2-fucosyltransferase|fucosyltransferase 1 (galactoside 2-alpha-L-fucosyltransferase)|fucosyltransferase 1 (galactoside 2-alpha-L-fucosyltransferase, H blood group)', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48748011, end_position_on_the_genomic_accession=48755390, orientation='minus', exon_count=4, OMIM=211100, no_hdr0=''),
  2524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2524, CurrentID=0, Status='live', Symbol='FUT2', Aliases='B12QTL1, SE, SEC2, Se2, sej', description='fucosyltransferase 2', other_designations='GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase 2|alpha (1,2) fucosyltransferase|alpha(1,2)FT 2|alpha(1,2)FT2|fucosyltransferase 2 (secretor status included)|secretor blood group alpha-2-fucosyltransferase|secretor factor', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48695971, end_position_on_the_genomic_accession=48705934, orientation='plus', exon_count=2, OMIM=182100, no_hdr0=''),
  2525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2525, CurrentID=0, Status='live', Symbol='FUT3', Aliases='CD174, FT3B, FucT-III, LE, Les', description='fucosyltransferase 3 (Lewis blood group)', other_designations='Lewis FT|alpha-(1,3/1,4)-fucosyltransferase|blood group Lewis alpha-4-fucosyltransferase|fucosyltransferase 3 (galactoside 3(4)-L-fucosyltransferase, Lewis blood group)|fucosyltransferase III', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5842888, end_position_on_the_genomic_accession=5851474, orientation='minus', exon_count=5, OMIM=111100, no_hdr0=''),
  2526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2526, CurrentID=0, Status='live', Symbol='FUT4', Aliases='CD15, ELFT, FCT3A, FUC-TIV, FUTIV, LeX, SSEA-1', description='fucosyltransferase 4', other_designations='ELAM ligand fucosyltransferase|ELAM-1 ligand fucosyltransferase|Lewis X|fucT-IV|fucosyltransferase 4 (alpha (1,3) fucosyltransferase, myeloid-specific)|fucosyltransferase IV|galactoside 3-L-fucosyltransferase|stage-specific embryonic antigen 1', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94543851, end_position_on_the_genomic_accession=94549898, orientation='plus', exon_count=1, OMIM=104230, no_hdr0=''),
  2527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2527, CurrentID=0, Status='live', Symbol='FUT5', Aliases='FUC-TV', description='fucosyltransferase 5', other_designations='alpha (1,3) fucosyltransferase|fucT-V|fucosyltransferase 5 (alpha (1,3) fucosyltransferase)|fucosyltransferase V|galactoside 3-L-fucosyltransferase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5865826, end_position_on_the_genomic_accession=5870540, orientation='minus', exon_count=2, OMIM=136835, no_hdr0=''),
  2528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2528, CurrentID=0, Status='live', Symbol='FUT6', Aliases='FCT3A, FT1A, Fuc-TVI, FucT-VI', description='fucosyltransferase 6', other_designations='fucosyltransferase 6 (alpha (1,3) fucosyltransferase)|fucosyltransferase VI|galactoside 3-L-fucosyltransferase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5830626, end_position_on_the_genomic_accession=5839753, orientation='minus', exon_count=6, OMIM=136836, no_hdr0=''),
  2529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2529, CurrentID=0, Status='live', Symbol='FUT7', Aliases='FucT-VII', description='fucosyltransferase 7 (alpha (1,3) fucosyltransferase)', other_designations='fuc-TVII|fucosyltransferase VII|galactoside 3-L-fucosyltransferase|selectin ligand synthase|selectin-ligand synthase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137030174, end_position_on_the_genomic_accession=137032840, orientation='minus', exon_count=2, OMIM=602030, no_hdr0=''),
  2530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2530, CurrentID=0, Status='live', Symbol='FUT8', Aliases='', description='fucosyltransferase 8 (alpha (1,6) fucosyltransferase)', other_designations='GDP-L-Fuc:N-acetyl-beta-D-glucosaminide alpha1,6-fucosyltransferase|GDP-fucose--glycoprotein fucosyltransferase|alpha1-6FucT|glycoprotein 6-alpha-L-fucosyltransferase', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65410592, end_position_on_the_genomic_accession=65744121, orientation='plus', exon_count=13, OMIM=602589, no_hdr0=''),
  2531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2531, CurrentID=0, Status='live', Symbol='KDSR', Aliases='DHSR, FVT1, SDR35C1', description='3-ketodihydrosphingosine reductase', other_designations='3-dehydrosphinganine reductase|FVT-1|KDS reductase|follicular lymphoma variant translocation 1|follicular variant translocation protein 1|short chain dehydrogenase/reductase family 35C member 1|short chain dehydrogenase/reductase family 35C, member 1', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63327738, end_position_on_the_genomic_accession=63367273, orientation='minus', exon_count=11, OMIM=136440, no_hdr0=''),
  2532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2532, CurrentID=0, Status='live', Symbol='ACKR1', Aliases='CCBP1, CD234, DARC, Dfy, FY, GPD, GpFy, WBCQ1', description='atypical chemokine receptor 1 (Duffy blood group)', other_designations='Duffy antigen chemokine receptor|Duffy blood group antigen|Duffy blood group, atypical chemokine receptor|Duffy blood group, chemokine receptor|Fy glycoprotein|duffy blood group chemokine receptor|glycoprotein D|plasmodium vivax receptor', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159204013, end_position_on_the_genomic_accession=159206500, orientation='plus', exon_count=2, OMIM=613665, no_hdr0=''),
  2533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2533, CurrentID=0, Status='live', Symbol='FYB', Aliases='ADAP, PRO0823, SLAP-130, SLAP130', description='FYN binding protein', other_designations='FYB-120/130|FYN-T-binding protein|SLP-76-associated phosphoprotein|adhesion and degranulation-promoting adaptor protein|p120/p130', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39105252, end_position_on_the_genomic_accession=39274523, orientation='minus', exon_count=27, OMIM=602731, no_hdr0=''),
  2534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2534, CurrentID=0, Status='live', Symbol='FYN', Aliases='SLK, SYN, p59-FYN', description='FYN proto-oncogene, Src family tyrosine kinase', other_designations='FYN oncogene related to SRC, FGR, YES|OKT3-induced calcium influx regulator|c-syn protooncogene|proto-oncogene Syn|proto-oncogene c-Fyn|src-like kinase|src/yes-related novel|tyrosine kinase p59fyn(T)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111660332, end_position_on_the_genomic_accession=111873452, orientation='minus', exon_count=19, OMIM=137025, no_hdr0=''),
  2535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2535, CurrentID=0, Status='live', Symbol='FZD2', Aliases='Fz2, fz-2, fzE2, hFz2', description='frizzled class receptor 2', other_designations='frizzled 2, seven transmembrane spanning receptor|frizzled family receptor 2|frizzled homolog 2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44557444, end_position_on_the_genomic_accession=44561262, orientation='plus', exon_count=1, OMIM=600667, no_hdr0=''),
  2537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2537, CurrentID=0, Status='live', Symbol='IFI6', Aliases='6-16, FAM14C, G1P3, IFI-6-1616, IFI6', description='interferon, alpha-inducible protein 6', other_designations='interferon, alpha-inducible protein clone IFI-6-16|interferon-induced protein 6-16', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27666061, end_position_on_the_genomic_accession=27672213, orientation='minus', exon_count=5, OMIM=147572, no_hdr0=''),
  2538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2538, CurrentID=0, Status='live', Symbol='G6PC', Aliases='G6PC1, G6PT, GSD1, GSD1a', description='glucose-6-phosphatase, catalytic subunit', other_designations='G-6-Pase|G6Pase|G6Pase-alpha|glucose-6-phosphatase alpha', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42900797, end_position_on_the_genomic_accession=42914433, orientation='plus', exon_count=5, OMIM=613742, no_hdr0=''),
  2539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2539, CurrentID=0, Status='live', Symbol='G6PD', Aliases='G6PD1', description='glucose-6-phosphate dehydrogenase', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154531390, end_position_on_the_genomic_accession=154547586, orientation='minus', exon_count=14, OMIM=305900, no_hdr0=''),
  2541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2541, CurrentID=0, Status='live', Symbol='G6PR', Aliases='GSD1aSP', description='glucose-6-phosphatase regulator', other_designations='glucose-6-phosphatase, regulatory', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2542, CurrentID=0, Status='live', Symbol='SLC37A4', Aliases='G6PT1, G6PT2, G6PT3, GSD1b, GSD1c, GSD1d, PRO0685, TRG-19, TRG19', description='solute carrier family 37 (glucose-6-phosphate transporter), member 4', other_designations='glucose-5-phosphate transporter|glucose-6-phosphatase, transport (glucose) protein 3|glucose-6-phosphatase, transport (glucose-6-phosphate) protein 1|glucose-6-phosphatase, transport (phosphate/pyrophosphate) protein 2|microsomal glucose-6-phosphate transporter|transformation-related gene 19 protein', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119024351, end_position_on_the_genomic_accession=119030906, orientation='minus', exon_count=12, OMIM=602671, no_hdr0=''),
  2543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2543, CurrentID=0, Status='live', Symbol='GAGE1', Aliases='CT4.1, GAGE-1', description='G antigen 1', other_designations='MZ2-F antigen|cancer/testis antigen 4.1|cancer/testis antigen family 4, member 1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49598995, end_position_on_the_genomic_accession=49608536, orientation='plus', exon_count=7, OMIM=300594, no_hdr0=''),
  2547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2547, CurrentID=0, Status='live', Symbol='XRCC6', Aliases='CTC75, CTCBF, G22P1, KU70, ML8, TLAA', description='X-ray repair complementing defective repair in Chinese hamster cells 6', other_designations="5'-dRP lyase Ku70|5'-deoxyribose-5-phosphate lyase Ku70|70 kDa subunit of Ku antigen|ATP-dependent DNA helicase 2 subunit 1|ATP-dependent DNA helicase II 70 kDa subunit|ATP-dependent DNA helicase II, 70 kDa subunit|CTC box binding factor 75 kDa subunit|CTC box-binding factor 75 kDa subunit|DNA repair protein XRCC6|Ku autoantigen p70 subunit|Ku autoantigen, 70kDa|lupus Ku autoantigen protein p70|thyroid autoantigen 70kD (Ku antigen)|thyroid autoantigen 70kDa (Ku antigen)|thyroid-lupus autoantigen p70", map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41621163, end_position_on_the_genomic_accession=41664048, orientation='plus', exon_count=13, OMIM=152690, no_hdr0=''),
  2548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2548, CurrentID=0, Status='live', Symbol='GAA', Aliases='LYAG', description='glucosidase, alpha; acid', other_designations='acid maltase|aglucosidase alfa', map_location='17q25.2-q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80101531, end_position_on_the_genomic_accession=80119881, orientation='plus', exon_count=21, OMIM=606800, no_hdr0=''),
  2549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2549, CurrentID=0, Status='live', Symbol='GAB1', Aliases='', description='GRB2 associated binding protein 1', other_designations='GRB2-associated binder 1|GRB2-associated binding protein 1|growth factor receptor bound protein 2-associated protein 1', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143336762, end_position_on_the_genomic_accession=143474565, orientation='plus', exon_count=14, OMIM=604439, no_hdr0=''),
  2550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2550, CurrentID=0, Status='live', Symbol='GABBR1', Aliases='GABABR1-3, GB1, GPRC3A, dJ271M21.1.1, dJ271M21.1.2, GABBR1', description='gamma-aminobutyric acid (GABA) B receptor, 1', other_designations='GABA-BR1|GABAB, subunit 1c|seven transmembrane helix receptor', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29602228, end_position_on_the_genomic_accession=29633185, orientation='minus', exon_count=25, OMIM=603540, no_hdr0=''),
  2551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2551, CurrentID=0, Status='live', Symbol='GABPA', Aliases='E4TF1-60, E4TF1A, NFT2, NRF2, NRF2A, RCH04A07', description='GA binding protein transcription factor alpha subunit', other_designations='GA binding protein transcription factor alpha subunit 60kDa|GA binding protein transcription factor, alpha subunit 60kDa|GABP subunit alpha|human nuclear respiratory factor-2 subunit alpha|nuclear respiratory factor 2 alpha subunit|nuclear respiratory factor 2 subunit alpha|transcription factor E4TF1-60', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25734947, end_position_on_the_genomic_accession=25772460, orientation='plus', exon_count=12, OMIM=600609, no_hdr0=''),
  2553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2553, CurrentID=0, Status='live', Symbol='GABPB1', Aliases='BABPB2, E4TF1, E4TF1-47, E4TF1-53, E4TF1B, GABPB, GABPB2, NRF2B1, NRF2B2', description='GA binding protein transcription factor, beta subunit 1', other_designations='GA binding protein transcription factor beta subunit 1 transcript variant gamma-2|GABP subunit beta-2|GABPB-1|nuclear respiratory factor 2|transcription factor E4TF1-47|transcription factor E4TF1-53', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50277192, end_position_on_the_genomic_accession=50355408, orientation='minus', exon_count=12, OMIM=600610, no_hdr0=''),
  2554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2554, CurrentID=0, Status='live', Symbol='GABRA1', Aliases='ECA4, EIEE19, EJM, EJM5', description='gamma-aminobutyric acid (GABA) A receptor, alpha 1', other_designations='GABA(A) receptor subunit alpha-1|GABA(A) receptor, alpha 1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161847191, end_position_on_the_genomic_accession=161899959, orientation='plus', exon_count=13, OMIM=137160, no_hdr0=''),
  2555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2555, CurrentID=0, Status='live', Symbol='GABRA2', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, alpha 2', other_designations='GABA(A) receptor subunit alpha-2|GABA(A) receptor, alpha 2|gamma-aminobutyric acid A receptor alpha 2', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46243548, end_position_on_the_genomic_accession=46390039, orientation='minus', exon_count=12, OMIM=137140, no_hdr0=''),
  2556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2556, CurrentID=0, Status='live', Symbol='GABRA3', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, alpha 3', other_designations='GABA(A) receptor subunit alpha-3|GABA(A) receptor, alpha 3|gamma-animobutyric acid (GABA) A receptor, alpha 3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152166234, end_position_on_the_genomic_accession=152451359, orientation='minus', exon_count=11, OMIM=305660, no_hdr0=''),
  2557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2557, CurrentID=0, Status='live', Symbol='GABRA4', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, alpha 4', other_designations='GABA(A) receptor, alpha 4|gamma-aminobutyric acid A receptor alpha 4', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46918900, end_position_on_the_genomic_accession=46994407, orientation='minus', exon_count=9, OMIM=137141, no_hdr0=''),
  2558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2558, CurrentID=0, Status='live', Symbol='GABRA5', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, alpha 5', other_designations='GABA(A) receptor subunit alpha-5', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26866719, end_position_on_the_genomic_accession=26949210, orientation='plus', exon_count=13, OMIM=137142, no_hdr0=''),
  2559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2559, CurrentID=0, Status='live', Symbol='GABRA6', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, alpha 6', other_designations='GABA subunit A receptor alpha 6|GABA(A) receptor subunit alpha-6|GABA(A) receptor, alpha 6', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161685652, end_position_on_the_genomic_accession=161702592, orientation='plus', exon_count=9, OMIM=137143, no_hdr0=''),
  2560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2560, CurrentID=0, Status='live', Symbol='GABRB1', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, beta 1', other_designations='GABA(A) receptor subunit beta-1|GABA(A) receptor, beta 1|gamma-aminobutyric acid A receptor beta 1', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46993858, end_position_on_the_genomic_accession=47427343, orientation='plus', exon_count=11, OMIM=137190, no_hdr0=''),
  2561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2561, CurrentID=0, Status='live', Symbol='GABRB2', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, beta 2', other_designations='GABA(A) receptor subunit beta-2|GABA(A) receptor, beta 2|gamma-aminobutyric acid A receptor beta 2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161288419, end_position_on_the_genomic_accession=161548124, orientation='minus', exon_count=12, OMIM=600232, no_hdr0=''),
  2562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2562, CurrentID=0, Status='live', Symbol='GABRB3', Aliases='ECA5', description='gamma-aminobutyric acid (GABA) A receptor, beta 3', other_designations='GABA-alpha receptor beta-2 subunit|GABAA receptor beta-3 subunit|gamma-aminobutyric acid A receptor beta 3', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26543546, end_position_on_the_genomic_accession=26773788, orientation='minus', exon_count=18, OMIM=137192, no_hdr0=''),
  2563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2563, CurrentID=0, Status='live', Symbol='GABRD', Aliases='EIG10, EJM7, GEFSP5', description='gamma-aminobutyric acid (GABA) A receptor, delta', other_designations='GABA(A) receptor subunit delta|GABA(A) receptor, delta|GABA-A receptor, delta polypeptide', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2019329, end_position_on_the_genomic_accession=2030753, orientation='plus', exon_count=10, OMIM=137163, no_hdr0=''),
  2564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2564, CurrentID=0, Status='live', Symbol='GABRE', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, epsilon', other_designations='GABA(A) receptor subunit epsilon|GABA(A) receptor, epsilon', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151953124, end_position_on_the_genomic_accession=151974684, orientation='minus', exon_count=12, OMIM=300093, no_hdr0=''),
  2565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2565, CurrentID=0, Status='live', Symbol='GABRG1', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, gamma 1', other_designations='GABA(A) receptor subunit gamma-1|GABA(A) receptor, gamma|gamma-1 polypeptide', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46035769, end_position_on_the_genomic_accession=46124065, orientation='minus', exon_count=9, OMIM=137166, no_hdr0=''),
  2566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2566, CurrentID=0, Status='live', Symbol='GABRG2', Aliases='CAE2, ECA2, GEFSP3', description='gamma-aminobutyric acid (GABA) A receptor, gamma 2', other_designations='GABA(A) receptor subunit gamma-2|GABA(A) receptor, gamma 2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=162067642, end_position_on_the_genomic_accession=162155539, orientation='plus', exon_count=11, OMIM=137164, no_hdr0=''),
  2567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2567, CurrentID=0, Status='live', Symbol='GABRG3', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, gamma 3', other_designations='GABA(A) receptor subunit gamma-3|GABA(G) receptor, gamma 3', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26971282, end_position_on_the_genomic_accession=27533895, orientation='plus', exon_count=12, OMIM=600233, no_hdr0=''),
  2568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2568, CurrentID=0, Status='live', Symbol='GABRP', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, pi', other_designations='GABA(A) receptor subunit pi|GABA(A) receptor, pi', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170772566, end_position_on_the_genomic_accession=170814047, orientation='plus', exon_count=12, OMIM=602729, no_hdr0=''),
  2569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2569, CurrentID=0, Status='live', Symbol='GABRR1', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, rho 1', other_designations='GABA(A) receptor subunit rho-1|GABA(A) receptor, rho 1|GABA(C) receptor|bA135P14.1 (gamma-aminobutyric acid (GABA) receptor, rho 1)|gamma-aminobutyric acid (GABA) A receptor, rho-1|gamma-aminobutyric acid (GABA) receptor, rho 1', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89177504, end_position_on_the_genomic_accession=89231288, orientation='minus', exon_count=13, OMIM=137161, no_hdr0=''),
  2570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2570, CurrentID=0, Status='live', Symbol='GABRR2', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, rho 2', other_designations='GABA-C receptor, rho-2 subunit', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89256061, end_position_on_the_genomic_accession=89315299, orientation='minus', exon_count=10, OMIM=137162, no_hdr0=''),
  2571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2571, CurrentID=0, Status='live', Symbol='GAD1', Aliases='CPSQ1, GAD, SCP', description='glutamate decarboxylase 1', other_designations='67 kDa glutamic acid decarboxylase|GAD-67|glutamate decarboxylase 1 (brain, 67kDa)|glutamate decarboxylase 67 kDa isoform', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170813210, end_position_on_the_genomic_accession=170861151, orientation='plus', exon_count=19, OMIM=605363, no_hdr0=''),
  2572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2572, CurrentID=0, Status='live', Symbol='GAD2', Aliases='GAD65', description='glutamate decarboxylase 2', other_designations='65 kDa glutamic acid decarboxylase|GAD-65|Glutamate decarboxylase-2 (pancreas)|glutamate decarboxylase 2 (pancreatic islets and brain, 65kDa)|glutamate decarboxylase 65 kDa isoform', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26216307, end_position_on_the_genomic_accession=26304562, orientation='plus', exon_count=17, OMIM=138275, no_hdr0=''),
  2574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2574, CurrentID=0, Status='live', Symbol='GAGE2C', Aliases='CT4.2, GAGE-2, GAGE-2C, GAGE2', description='G antigen 2C', other_designations='G antigen 2|cancer/testis antigen 4.2', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300595, no_hdr0=''),
  2575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2575, CurrentID=0, Status='live', Symbol='GAGE3', Aliases='CT4.3', description='G antigen 3', other_designations='GAGE-3|cancer/testis antigen 4.3|cancer/testis antigen family 4, member 3', map_location='Xp11.23 not on reference assembly', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300596, no_hdr0=''),
  2576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2576, CurrentID=0, Status='live', Symbol='GAGE4', Aliases='CT4.4', description='G antigen 4', other_designations='GAGE-4|cancer/testis antigen 4.4|cancer/testis antigen family 4, member 4', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=49103593, end_position_on_the_genomic_accession=49110884, orientation='plus', exon_count=5, OMIM=300597, no_hdr0=''),
  2577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2577, CurrentID=0, Status='live', Symbol='GAGE5', Aliases='CT4.5', description='G antigen 5', other_designations='GAGE-5|cancer/testis antigen 4.5|cancer/testis antigen family 4, member 5', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=49103601, end_position_on_the_genomic_accession=49110887, orientation='plus', exon_count=5, OMIM=300598, no_hdr0=''),
  2578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2578, CurrentID=0, Status='live', Symbol='GAGE6', Aliases='CT4.6', description='G antigen 6', other_designations='GAGE-6|cancer/testis antigen 4.6|cancer/testis antigen family 4, member 6', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=49212424, end_position_on_the_genomic_accession=49219752, orientation='plus', exon_count=5, OMIM=300599, no_hdr0=''),
  2579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2579, CurrentID=0, Status='live', Symbol='GAGE7', Aliases='AL4, CT4.7, GAGE-7', description='G antigen 7', other_designations='cancer/testis antigen 4.7|cancer/testis antigen family 4, member 7', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=49103622, end_position_on_the_genomic_accession=49110884, orientation='plus', exon_count=5, OMIM=300601, no_hdr0=''),
  2580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2580, CurrentID=0, Status='live', Symbol='GAK', Aliases='DNAJ26, DNAJC26', description='cyclin G associated kinase', other_designations='auxilin-2', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=849277, end_position_on_the_genomic_accession=932390, orientation='minus', exon_count=31, OMIM=602052, no_hdr0=''),
  2581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2581, CurrentID=0, Status='live', Symbol='GALC', Aliases='', description='galactosylceramidase', other_designations='GALCERase|galactocerebroside beta-galactosidase|galactosylceramide beta-galactosidase|galactosylceraminidase', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=87933014, end_position_on_the_genomic_accession=87993665, orientation='minus', exon_count=19, OMIM=606890, no_hdr0=''),
  2582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2582, CurrentID=0, Status='live', Symbol='GALE', Aliases='SDR1E1', description='UDP-galactose-4-epimerase', other_designations="UDP galactose-4'-epimerase|UDP-GalNAc 4-epimerase|UDP-GlcNAc 4-epimerase|UDP-N-acetylgalactosamine 4-epimerase|UDP-N-acetylglucosamine 4-epimerase|UDP-galactose 4-epimerase|galactose-4-epimerase, UDP-|galactowaldenase|short chain dehydrogenase/reductase family 1E, member 1", map_location='1p36-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23795599, end_position_on_the_genomic_accession=23800804, orientation='minus', exon_count=13, OMIM=606953, no_hdr0=''),
  2583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2583, CurrentID=0, Status='live', Symbol='B4GALNT1', Aliases='GALGT, GALNACT, GalNAc-T, SPG26', description='beta-1,4-N-acetyl-galactosaminyl transferase 1', other_designations='GD2 synthase, GM2 synthase|GM2/GD2 synthase|UDP-Gal:betaGlcNAc beta-1,4-N-acetylgalactosaminyltransferase transferase 1|UDP-N-acetyl-alpha-D-galactosamine:(N-acetylneuraminyl)-galactosylglucosylceramide N-acetylgalactosaminyltransferase (GalNAc-T)|beta1,4GalNAc-T', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57625895, end_position_on_the_genomic_accession=57633239, orientation='minus', exon_count=13, OMIM=601873, no_hdr0=''),
  2584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2584, CurrentID=0, Status='live', Symbol='GALK1', Aliases='GALK, GK1, HEL-S-19', description='galactokinase 1', other_designations='epididymis secretory protein Li 19|galactose kinase', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75757937, end_position_on_the_genomic_accession=75765199, orientation='minus', exon_count=8, OMIM=604313, no_hdr0=''),
  2585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2585, CurrentID=0, Status='live', Symbol='GALK2', Aliases='GK2', description='galactokinase 2', other_designations='galNAc kinase', map_location='15q21.1-q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49155759, end_position_on_the_genomic_accession=49367894, orientation='plus', exon_count=18, OMIM=137028, no_hdr0=''),
  2587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2587, CurrentID=0, Status='live', Symbol='GALR1', Aliases='GALNR, GALNR1', description='galanin receptor 1', other_designations='GAL1-R|GALR-1', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77250052, end_position_on_the_genomic_accession=77270142, orientation='plus', exon_count=3, OMIM=600377, no_hdr0=''),
  2588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2588, CurrentID=0, Status='live', Symbol='GALNS', Aliases='GALNAC6S, GAS, GalN6S, MPS4A', description='galactosamine (N-acetyl)-6-sulfatase', other_designations='N-acetylgalactosamine-6-sulfate sulfatase|chondroitinase|chondroitinsulfatase|galNAc6S sulfatase|galactosamine (N-acetyl)-6-sulfate sulfatase|galactose-6-sulfate sulfatase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88855128, end_position_on_the_genomic_accession=88856579, orientation='minus', exon_count=18, OMIM=612222, no_hdr0=''),
  2589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2589, CurrentID=0, Status='live', Symbol='GALNT1', Aliases='GALNAC-T1', description='polypeptide N-acetylgalactosaminyltransferase 1', other_designations='GalNAc transferase 1|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 1|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 1 (GalNAc-T1)|polypeptide GalNAc transferase 1|pp-GaNTase 1|protein-UDP acetylgalactosaminyltransferase 1', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35627436, end_position_on_the_genomic_accession=35711834, orientation='plus', exon_count=12, OMIM=602273, no_hdr0=''),
  2590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2590, CurrentID=0, Status='live', Symbol='GALNT2', Aliases='GalNAc-T2', description='polypeptide N-acetylgalactosaminyltransferase 2', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 2|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 2 (GalNAc-T2)|polypeptide GalNAc transferase 2|pp-GaNTase 2', map_location='1q41-q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230057789, end_position_on_the_genomic_accession=230282130, orientation='plus', exon_count=20, OMIM=602274, no_hdr0=''),
  2591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2591, CurrentID=0, Status='live', Symbol='GALNT3', Aliases='GalNAc-T3, HFTC, HHS', description='polypeptide N-acetylgalactosaminyltransferase 3', other_designations='GalNAc transferase 3|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 3|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 3 (GalNAc-T3)|polypeptide GalNAc transferase 3|polypeptide GalNAc-transferase T3|pp-GaNTase 3|protein-UDP acetylgalactosaminyltransferase 3', map_location='2q24-q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165747803, end_position_on_the_genomic_accession=165794650, orientation='minus', exon_count=13, OMIM=601756, no_hdr0=''),
  2592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2592, CurrentID=0, Status='live', Symbol='GALT', Aliases='', description='galactose-1-phosphate uridylyltransferase', other_designations='UDP-glucose--hexose-1-phosphate uridylyltransferase|gal-1-P uridylyltransferase|galactose-1-phosphate uridyl transferase', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34646589, end_position_on_the_genomic_accession=34650598, orientation='plus', exon_count=11, OMIM=606999, no_hdr0=''),
  2593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2593, CurrentID=0, Status='live', Symbol='GAMT', Aliases='CCDS2, HEL-S-20, PIG2, TP53I2', description='guanidinoacetate N-methyltransferase', other_designations='epididymis secretory protein Li 20', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1397026, end_position_on_the_genomic_accession=1401570, orientation='minus', exon_count=6, OMIM=601240, no_hdr0=''),
  2595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2595, CurrentID=0, Status='live', Symbol='GANC', Aliases='', description='glucosidase, alpha; neutral C', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42273658, end_position_on_the_genomic_accession=42353666, orientation='plus', exon_count=28, OMIM=104180, no_hdr0=''),
  2596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2596, CurrentID=0, Status='live', Symbol='GAP43', Aliases='B-50, PP46', description='growth associated protein 43', other_designations='axonal membrane protein GAP-43|calmodulin-binding protein P-57|growth-associated protein 43|nerve growth-related peptide GAP43|neural phosphoprotein B-50|neuron growth-associated protein 43|protein F1', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115623304, end_position_on_the_genomic_accession=115721487, orientation='plus', exon_count=4, OMIM=162060, no_hdr0=''),
  2597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2597, CurrentID=0, Status='live', Symbol='GAPDH', Aliases='G3PD, GAPD, HEL-S-162eP', description='glyceraldehyde-3-phosphate dehydrogenase', other_designations='aging-associated gene 9 protein|epididymis secretory sperm binding protein Li 162eP|peptidyl-cysteine S-nitrosylase GAPDH', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6534405, end_position_on_the_genomic_accession=6538375, orientation='plus', exon_count=10, OMIM=138400, no_hdr0=''),
  2601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2601, CurrentID=0, Status='live', Symbol='GAPDHP60', Aliases='GAPDHL4, GAPDL4', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 60', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 4, pseudogene', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87207016, end_position_on_the_genomic_accession=87208290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2609, CurrentID=0, Status='live', Symbol='GAPDHP67', Aliases='GAPDHL12, GAPDL12', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 67', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 12, pseudogene', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136077877, end_position_on_the_genomic_accession=136078632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2611, CurrentID=0, Status='live', Symbol='GAPDHP69', Aliases='GAPDHL14, GAPDL14', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 69', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 14, pseudogene', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29306707, end_position_on_the_genomic_accession=29307974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2613, CurrentID=0, Status='live', Symbol='GAPDHP71', Aliases='GAPDHL16, GAPDL16', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 71', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 16, pseudogene', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174513233, end_position_on_the_genomic_accession=174514513, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  133686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133686, CurrentID=0, Status='live', Symbol='NADK2', Aliases='C5orf33, DECRD, MNADK, NADKD1', description='NAD kinase 2, mitochondrial', other_designations='NAD kinase domain-containing protein 1, mitochondrial', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36192589, end_position_on_the_genomic_accession=36242279, orientation='minus', exon_count=15, OMIM=615787, no_hdr0=''),
  2615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2615, CurrentID=0, Status='live', Symbol='LRRC32', Aliases='D11S833E, GARP', description='leucine rich repeat containing 32', other_designations='garpin|glycoprotein A repetitions predominant', map_location='11q13.5-q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76657524, end_position_on_the_genomic_accession=76670946, orientation='minus', exon_count=4, OMIM=137207, no_hdr0=''),
  133688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133688, CurrentID=0, Status='live', Symbol='UGT3A1', Aliases='', description='UDP glycosyltransferase 3 family, polypeptide A1', other_designations='UDPGT 3A1', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35953089, end_position_on_the_genomic_accession=36001045, orientation='minus', exon_count=11, OMIM=616383, no_hdr0=''),
  2617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2617, CurrentID=0, Status='live', Symbol='GARS', Aliases='CMT2D, DSMAV, GlyRS, HMN5, SMAD1', description='glycyl-tRNA synthetase', other_designations='AP-4-A synthetase|Charcot-Marie-Tooth neuropathy 2D|Charcot-Marie-Tooth neuropathy, neuronal type, D|diadenosine tetraphosphate synthetase|glycine tRNA ligase', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30594565, end_position_on_the_genomic_accession=30634033, orientation='plus', exon_count=18, OMIM=600287, no_hdr0=''),
  2618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2618, CurrentID=0, Status='live', Symbol='GART', Aliases='AIRS, GARSF, PAIS, PGFT, PRGS, GART', description='phosphoribosylglycinamide formyltransferase, phosphoribosylglycinamide synthetase, phosphoribosylaminoimidazole synthetase', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33503931, end_position_on_the_genomic_accession=33542917, orientation='minus', exon_count=23, OMIM=138440, no_hdr0=''),
  2619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2619, CurrentID=0, Status='live', Symbol='GAS1', Aliases='', description='growth arrest specific 1', other_designations='GAS-1|Growth arrest-specific gene-1|growth arrest-specific 1', map_location='9q21.3-q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86944362, end_position_on_the_genomic_accession=86947189, orientation='minus', exon_count=1, OMIM=139185, no_hdr0=''),
  2620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2620, CurrentID=0, Status='live', Symbol='GAS2', Aliases='', description='growth arrest specific 2', other_designations='GAS-2|growth arrest-specific 2', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22627053, end_position_on_the_genomic_accession=22813001, orientation='plus', exon_count=16, OMIM=602835, no_hdr0=''),
  2621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2621, CurrentID=0, Status='live', Symbol='GAS6', Aliases='AXLLG, AXSF', description='growth arrest specific 6', other_designations='AXL receptor tyrosine kinase ligand|AXL stimulatory factor|growth arrest-specific 6', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113820549, end_position_on_the_genomic_accession=113864103, orientation='minus', exon_count=15, OMIM=600441, no_hdr0=''),
  2622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2622, CurrentID=0, Status='live', Symbol='GAS8', Aliases='GAS11', description='growth arrest specific 8', other_designations='GAS-11|GAS-8|growth arrest specific 11|growth arrest-specific 11|growth arrest-specific 8|growth arrest-specific protein 11', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90019629, end_position_on_the_genomic_accession=90044971, orientation='plus', exon_count=15, OMIM=605178, no_hdr0=''),
  2623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2623, CurrentID=0, Status='live', Symbol='GATA1', Aliases='ERYF1, GATA-1, GF-1, GF1, NF-E1, NFE1, XLANP, XLTDA, XLTT', description='GATA binding protein 1 (globin transcription factor 1)', other_designations='GATA-binding factor 1|NF-E1 DNA-binding protein|erythroid transcription factor 1|globin transcription factor 1|nuclear factor, erythroid 1|transcription factor GATA1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48786560, end_position_on_the_genomic_accession=48794311, orientation='plus', exon_count=6, OMIM=305371, no_hdr0=''),
  2624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2624, CurrentID=0, Status='live', Symbol='GATA2', Aliases='DCML, IMD21, MONOMAC, NFE1B', description='GATA binding protein 2', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128479422, end_position_on_the_genomic_accession=128493187, orientation='minus', exon_count=8, OMIM=137295, no_hdr0=''),
  2625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2625, CurrentID=0, Status='live', Symbol='GATA3', Aliases='HDR, HDRS', description='GATA binding protein 3', other_designations='GATA-binding factor 3', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8045428, end_position_on_the_genomic_accession=8075201, orientation='plus', exon_count=8, OMIM=131320, no_hdr0=''),
  2626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2626, CurrentID=0, Status='live', Symbol='GATA4', Aliases='ASD2, TACHD, TOF, VSD1', description='GATA binding protein 4', other_designations='GATA-binding factor 4', map_location='8p23.1-p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11676924, end_position_on_the_genomic_accession=11760001, orientation='plus', exon_count=12, OMIM=600576, no_hdr0=''),
  2627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2627, CurrentID=0, Status='live', Symbol='GATA6', Aliases='', description='GATA binding protein 6', other_designations='GATA-binding factor 6', map_location='18q11.1-q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22169437, end_position_on_the_genomic_accession=22202528, orientation='plus', exon_count=7, OMIM=601656, no_hdr0=''),
  2628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2628, CurrentID=0, Status='live', Symbol='GATM', Aliases='AGAT, AT, CCDS3', description='glycine amidinotransferase (L-arginine:glycine amidinotransferase)', other_designations='transamidinase', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45361124, end_position_on_the_genomic_accession=45397043, orientation='minus', exon_count=12, OMIM=602360, no_hdr0=''),
  2629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2629, CurrentID=0, Status='live', Symbol='GBA', Aliases='GBA1, GCB, GLUC', description='glucosidase, beta, acid', other_designations='D-glucosyl-N-acylsphingosine glucohydrolase|acid beta-glucosidase|alglucerase|beta-GC|beta-glucocerebrosidase|glucosylceramidase-like protein|imiglucerase|lysosomal glucocerebrosidase', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155234448, end_position_on_the_genomic_accession=155244862, orientation='minus', exon_count=13, OMIM=606463, no_hdr0=''),
  2630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2630, CurrentID=0, Status='live', Symbol='GBAP1', Aliases='GBAP', description='glucosidase, beta, acid pseudogene 1', other_designations='glucosidase, beta; acid, pseudogene|glucosylceramidase-like protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155213825, end_position_on_the_genomic_accession=155227534, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  2631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2631, CurrentID=0, Status='live', Symbol='GBAS', Aliases='NIPSNAP2', description='glioblastoma amplified sequence', other_designations='4-nitrophenylphosphatase domain and non-neuronal SNAP25-like 2', map_location='7p12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55964577, end_position_on_the_genomic_accession=56000182, orientation='plus', exon_count=10, OMIM=603004, no_hdr0=''),
  2632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2632, CurrentID=0, Status='live', Symbol='GBE1', Aliases='APBD, GBE, GSD4', description='glucan (1,4-alpha-), branching enzyme 1', other_designations='amylo-(1,4 to 1,6) transglucosidase|amylo-(1,4 to 1,6) transglycosylase|brancher enzyme|glycogen branching enzyme|glycogen-branching enzyme', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=81489699, end_position_on_the_genomic_accession=81761799, orientation='minus', exon_count=16, OMIM=607839, no_hdr0=''),
  2633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2633, CurrentID=0, Status='live', Symbol='GBP1', Aliases='', description='guanylate binding protein 1, interferon-inducible', other_designations='GBP-1|GTP-binding protein 1|guanine nucleotide-binding protein 1|guanylate binding protein 1, interferon-inducible, 67kDa|huGBP-1|interferon-induced guanylate-binding protein 1', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89052304, end_position_on_the_genomic_accession=89065360, orientation='minus', exon_count=12, OMIM=600411, no_hdr0=''),
  2634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2634, CurrentID=0, Status='live', Symbol='GBP2', Aliases='', description='guanylate binding protein 2, interferon-inducible', other_designations='GTP-binding protein 2|guanine nucleotide-binding protein 2|interferon-induced guanylate-binding protein 2', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89106132, end_position_on_the_genomic_accession=89126159, orientation='minus', exon_count=11, OMIM=600412, no_hdr0=''),
  2635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2635, CurrentID=0, Status='live', Symbol='GBP3', Aliases='', description='guanylate binding protein 3', other_designations='GBP-3|GTP-binding protein 3|guanine nucleotide-binding protein 3|guanylate-binding protein 3 delta C', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89006677, end_position_on_the_genomic_accession=89022880, orientation='minus', exon_count=13, OMIM=600413, no_hdr0=''),
  2636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2636, CurrentID=0, Status='live', Symbol='GBX1', Aliases='Huh-17', description='gastrulation brain homeobox 1', other_designations='gastrulation and brain-specific homeobox protein 1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151146738, end_position_on_the_genomic_accession=151167548, orientation='minus', exon_count=4, OMIM=603354, no_hdr0=''),
  2637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2637, CurrentID=0, Status='live', Symbol='GBX2', Aliases='', description='gastrulation brain homeobox 2', other_designations='gastrulation and brain-specific homeobox protein 2', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236165236, end_position_on_the_genomic_accession=236168009, orientation='minus', exon_count=3, OMIM=601135, no_hdr0=''),
  2638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2638, CurrentID=0, Status='live', Symbol='GC', Aliases='DBP, DBP/GC, GRD3, Gc-MAF, GcMAF, HEL-S-51, VDBG, VDBP', description='group-specific component (vitamin D binding protein)', other_designations='DBP-maf|VDB|epididymis secretory protein Li 51|gc protein-derived macrophage activating factor|gc-globulin|vitamin D-binding alpha-globulin|vitamin D-binding protein-macrophage activating factor', map_location='4q12-q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71741693, end_position_on_the_genomic_accession=71805520, orientation='minus', exon_count=15, OMIM=139200, no_hdr0=''),
  2639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2639, CurrentID=0, Status='live', Symbol='GCDH', Aliases='ACAD5, GCD', description='glutaryl-CoA dehydrogenase', other_designations='glutaryl-Coenzyme A dehydrogenase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12891129, end_position_on_the_genomic_accession=12914630, orientation='plus', exon_count=15, OMIM=608801, no_hdr0=''),
  2641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2641, CurrentID=0, Status='live', Symbol='GCG', Aliases='GLP1, GLP2, GRPP', description='glucagon', other_designations='glicentin-related polypeptide|glucagon-like peptide 1|glucagon-like peptide 2|preproglucagon', map_location='2q36-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162142869, end_position_on_the_genomic_accession=162152404, orientation='minus', exon_count=6, OMIM=138030, no_hdr0=''),
  2642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2642, CurrentID=0, Status='live', Symbol='GCGR', Aliases='GGR, GL-R', description='glucagon receptor', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81804132, end_position_on_the_genomic_accession=81814013, orientation='plus', exon_count=15, OMIM=138033, no_hdr0=''),
  2643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2643, CurrentID=0, Status='live', Symbol='GCH1', Aliases='DYT14, DYT5, DYT5a, GCH, GTP-CH-1, GTPCH1, HPABH4B', description='GTP cyclohydrolase 1', other_designations="GTP cyclohydrolase I|GTP-CH-I|dystonia 14|guanosine 5'-triphosphate cyclohydrolase I", map_location='14q22.1-q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54842005, end_position_on_the_genomic_accession=54902885, orientation='minus', exon_count=9, OMIM=600225, no_hdr0=''),
  2644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2644, CurrentID=0, Status='live', Symbol='GCHFR', Aliases='GFRP, HsT16933, P35', description='GTP cyclohydrolase I feedback regulator', other_designations='GTP cyclohydrolase I feedback regulatory protein', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40764087, end_position_on_the_genomic_accession=40767713, orientation='plus', exon_count=3, OMIM=602437, no_hdr0=''),
  2645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2645, CurrentID=0, Status='live', Symbol='GCK', Aliases='FGQTL3, GK, GLK, HHF3, HK4, HKIV, HXKP, LGLK, MODY2', description='glucokinase (hexokinase 4)', other_designations='ATP:D-hexose 6-phosphotransferase|HK IV|hexokinase D, pancreatic isozyme|hexokinase type IV|hexokinase-4|hexokinase-D', map_location='7p15.3-p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44144271, end_position_on_the_genomic_accession=44189423, orientation='minus', exon_count=12, OMIM=138079, no_hdr0=''),
  2646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2646, CurrentID=0, Status='live', Symbol='GCKR', Aliases='FGQTL5, GKRP', description='glucokinase (hexokinase 4) regulator', other_designations='', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27496839, end_position_on_the_genomic_accession=27523689, orientation='plus', exon_count=19, OMIM=600842, no_hdr0=''),
  2647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2647, CurrentID=0, Status='live', Symbol='BLOC1S1', Aliases='BLOS1, GCN5L1, MICoA, RT14', description='biogenesis of lysosomal organelles complex-1, subunit 1', other_designations='BLOC subunit 1|BLOC-1 subunit 1|GCN5 (general control of amino-acid synthesis, yeast, homolog)-like 1|GCN5 general control of amino-acid synthesis 5-like 1|GCN5-like protein 1|MTA1-interacting coactivator', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55716034, end_position_on_the_genomic_accession=55719707, orientation='plus', exon_count=5, OMIM=601444, no_hdr0=''),
  2648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2648, CurrentID=0, Status='live', Symbol='KAT2A', Aliases='GCN5, GCN5L2, PCAF-b, hGCN5', description='K(lysine) acetyltransferase 2A', other_designations='GCN5 (general control of amino-acid synthesis, yeast, homolog)-like 2|General control of amino acid synthesis, yeast, homolog-like 2|STAF97|general control of amino acid synthesis protein 5-like 2|histone acetyltransferase GCN5|hsGCN5|lysine acetyltransferase 2A', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42113111, end_position_on_the_genomic_accession=42121393, orientation='minus', exon_count=18, OMIM=602301, no_hdr0=''),
  2649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2649, CurrentID=0, Status='live', Symbol='NR6A1', Aliases='CT150, GCNF, GCNF1, NR61, RTR, hGCNF, hRTR', description='nuclear receptor subfamily 6 group A member 1', other_designations='germ cell nuclear factor variant 1|nuclear receptor subfamily 6, group A, member 1|retinoic acid receptor-related testis-associated receptor|retinoid receptor-related testis-specific receptor', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124517275, end_position_on_the_genomic_accession=124771310, orientation='minus', exon_count=11, OMIM=602778, no_hdr0=''),
  2650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2650, CurrentID=0, Status='live', Symbol='GCNT1', Aliases='C2GNT, C2GNT-L, C2GNT1, G6NT, NACGT2, NAGCT2', description='glucosaminyl (N-acetyl) transferase 1, core 2', other_designations='beta-1,3-galactosyl-O-glycosyl-glycoprotein beta-1,6-N--acetylglucosaminyltransferase|beta-1,6-N-acetylglucosaminyltransferase|core 2 GnT|core 2 beta-1,6-N-acetylglucosaminyltransferase I|core 2 beta1,6 N-acetylglucosaminyltransferase-I|core 2 branching enzyme|core 2-branching enzyme|core2-GlcNAc-transferase|glucosaminyl (N-acetyl) transferase 1, core 2 (beta-1,6-N-acetylglucosaminyltransferase)', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76441666, end_position_on_the_genomic_accession=76507416, orientation='plus', exon_count=7, OMIM=600391, no_hdr0=''),
  2651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2651, CurrentID=0, Status='live', Symbol='GCNT2', Aliases='CCAT, CTRCT13C, GCNT5, IGNT, II, NACGT1, NAGCT1, ULG3, bA360O19.2, bA421M1.1, GCNT2', description='glucosaminyl (N-acetyl) transferase 2, I-branching enzyme (I blood group)', other_designations='I beta-1,6-N-acetylglucosaminyltransferase|Ii blood group|beta-1,6-N-acetylglucosaminyltransferase 2', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10521283, end_position_on_the_genomic_accession=10629368, orientation='plus', exon_count=14, OMIM=600429, no_hdr0=''),
  2652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2652, CurrentID=0, Status='live', Symbol='OPN1MW', Aliases='CBBM, CBD, COD5, GCP, GOP1, OPN1MW', description='opsin 1 (cone pigments), medium-wave-sensitive', other_designations='cone dystrophy 5 (X-linked)|green cone photoreceptor pigment|green cone pigment|green-sensitive opsin|photopigment apoprotein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154182596, end_position_on_the_genomic_accession=154196861, orientation='plus', exon_count=6, OMIM=300821, no_hdr0=''),
  2653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2653, CurrentID=0, Status='live', Symbol='GCSH', Aliases='GCE, NKH', description='glycine cleavage system protein H (aminomethyl carrier)', other_designations='lipoic acid-containing protein|mitochondrial glycine cleavage system H-protein', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81081947, end_position_on_the_genomic_accession=81096375, orientation='minus', exon_count=5, OMIM=238330, no_hdr0=''),
  2656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2656, CurrentID=0, Status='live', Symbol='GCY', Aliases='STA, TS, TSY', description='growth control, Y chromosome influenced', other_designations='', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=475000, no_hdr0=''),
  2657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2657, CurrentID=0, Status='live', Symbol='GDF1', Aliases='DORV, DTGA3, RAI', description='growth differentiation factor 1', other_designations='GDF-1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18868546, end_position_on_the_genomic_accession=18896144, orientation='minus', exon_count=8, OMIM=602880, no_hdr0=''),
  2658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2658, CurrentID=0, Status='live', Symbol='GDF2', Aliases='BMP-9, BMP9, HHT5', description='growth differentiation factor 2', other_designations='GDF-2|bone morphogenetic protein 9', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47322490, end_position_on_the_genomic_accession=47327594, orientation='plus', exon_count=3, OMIM=605120, no_hdr0=''),
  2660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2660, CurrentID=0, Status='live', Symbol='MSTN', Aliases='GDF8, MSLHP', description='myostatin', other_designations='', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190055700, end_position_on_the_genomic_accession=190062729, orientation='minus', exon_count=4, OMIM=601788, no_hdr0=''),
  2661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2661, CurrentID=0, Status='live', Symbol='GDF9', Aliases='', description='growth differentiation factor 9', other_designations='GDF-9', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132861181, end_position_on_the_genomic_accession=132866884, orientation='minus', exon_count=6, OMIM=601918, no_hdr0=''),
  2662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2662, CurrentID=0, Status='live', Symbol='GDF10', Aliases='BMP-3b, BMP3B', description='growth differentiation factor 10', other_designations='BIP|GDF-10|bone morphogenetic protein 3B|bone-inducing protein', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47300224, end_position_on_the_genomic_accession=47313583, orientation='plus', exon_count=3, OMIM=601361, no_hdr0=''),
  100130913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130913, CurrentID=0, Status='live', Symbol='LOC100130913', Aliases='', description='uncharacterized LOC100130913', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=55475032, end_position_on_the_genomic_accession=55475586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2664, CurrentID=0, Status='live', Symbol='GDI1', Aliases='1A, GDIL, MRX41, MRX48, OPHN2, RABGD1A, RABGDIA, XAP-4', description='GDP dissociation inhibitor 1', other_designations='GDI-1|guanosine diphosphate dissociation inhibitor 1|mental retardation, X-linked 48|oligophrenin-2|protein XAP-4|rab GDI alpha|rab GDP-dissociation inhibitor, alpha', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154436913, end_position_on_the_genomic_accession=154443467, orientation='plus', exon_count=11, OMIM=300104, no_hdr0=''),
  2665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2665, CurrentID=0, Status='live', Symbol='GDI2', Aliases='HEL-S-46e, RABGDIB', description='GDP dissociation inhibitor 2', other_designations='GDI-2|epididymis secretory sperm binding protein Li 46e|guanosine diphosphate dissociation inhibitor 2|rab GDI beta|rab GDP-dissociation inhibitor, beta', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5765223, end_position_on_the_genomic_accession=5813549, orientation='minus', exon_count=11, OMIM=600767, no_hdr0=''),
  2667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2667, CurrentID=0, Status='live', Symbol='GDI2P1', Aliases='GDI2P', description='GDP dissociation inhibitor 2 pseudogene 1', other_designations='rab GDP-dissociation inhibitor, beta pseudogene', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48901756, end_position_on_the_genomic_accession=48904003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2668, CurrentID=0, Status='live', Symbol='GDNF', Aliases='ATF1, ATF2, HFB1-GDNF, HSCR3', description='glial cell derived neurotrophic factor', other_designations='ATF|astrocyte-derived trophic factor', map_location='5p13.1-p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37812677, end_position_on_the_genomic_accession=37839680, orientation='minus', exon_count=7, OMIM=600837, no_hdr0=''),
  2669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2669, CurrentID=0, Status='live', Symbol='GEM', Aliases='KIR', description='GTP binding protein overexpressed in skeletal muscle', other_designations='GTP-binding mitogen-induced T-cell protein|GTP-binding protein overexpressed in skeletal muscle|RAS-like protein KIR|kinase-inducible Ras-like protein', map_location='8q13-q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94249253, end_position_on_the_genomic_accession=94262319, orientation='minus', exon_count=5, OMIM=600164, no_hdr0=''),
  2670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2670, CurrentID=0, Status='live', Symbol='GFAP', Aliases='ALXDRD', description='glial fibrillary acidic protein', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44905626, end_position_on_the_genomic_accession=44915552, orientation='minus', exon_count=10, OMIM=137780, no_hdr0=''),
  2671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2671, CurrentID=0, Status='live', Symbol='GFER', Aliases='ALR, ERV1, HERV1, HPO, HPO1, HPO2, HSS', description='growth factor, augmenter of liver regeneration', other_designations='ERV1 homolog|erv1-like growth factor|hepatic regenerative stimulation substance|hepatopoietin protein', map_location='16p13.3-p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1984149, end_position_on_the_genomic_accession=1987749, orientation='plus', exon_count=3, OMIM=600924, no_hdr0=''),
  2672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2672, CurrentID=0, Status='live', Symbol='GFI1', Aliases='GFI-1A, SCN2, ZNF163, GFI1', description='growth factor independent 1 transcription repressor', other_designations='growth factor independence-1|growth factor independent protein 1|zinc finger protein 163', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92473043, end_position_on_the_genomic_accession=92486876, orientation='minus', exon_count=11, OMIM=600871, no_hdr0=''),
  2673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2673, CurrentID=0, Status='live', Symbol='GFPT1', Aliases='CMS12, CMSTA1, GFA, GFAT, GFAT 1, GFAT1, GFAT1m, GFPTL, MSLG, GFPT1', description='glutamine--fructose-6-phosphate transaminase 1', other_designations='D-fructose-6-phosphate amidotransferase 1|glucosamine--fructose-6-phosphate aminotransferase [isomerizing] 1|glutamine:fructose-6-phosphate amidotransferase 1|hexosephosphate aminotransferase 1', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69319769, end_position_on_the_genomic_accession=69387254, orientation='minus', exon_count=20, OMIM=138292, no_hdr0=''),
  2674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2674, CurrentID=0, Status='live', Symbol='GFRA1', Aliases='GDNFR, GDNFRA, GFR-ALPHA-1, RET1L, RETL1, TRNR1', description='GDNF family receptor alpha 1', other_designations='GDNF receptor alpha-1|GDNFR-alpha-1|GPI-linked anchor protein|Glial cell line-derived neurotrophic factor receptor alpha|PI-linked cell-surface accessory protein|RET ligand 1|TGF-beta related neurotrophic factor receptor 1|TGF-beta-related neurotrophic factor receptor 1', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116056925, end_position_on_the_genomic_accession=116273614, orientation='minus', exon_count=13, OMIM=601496, no_hdr0=''),
  2675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2675, CurrentID=0, Status='live', Symbol='GFRA2', Aliases='GDNFRB, NRTNR-ALPHA, NTNRA, RETL2, TRNR2', description='GDNF family receptor alpha 2', other_designations='GDNF receptor beta|GDNFR-alpha-2|GDNFR-beta|GFR-alpha 2|NTNR-alpha|PI-linked cell-surface accessory protein|RET ligand 2|TGF-beta-related neurotrophic factor receptor 2|TRN receptor, GPI-anchored|glial cell line derived neurotrophic factor receptor, beta|neurturin receptor alpha', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21692014, end_position_on_the_genomic_accession=21789294, orientation='minus', exon_count=12, OMIM=601956, no_hdr0=''),
  2676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2676, CurrentID=0, Status='live', Symbol='GFRA3', Aliases='GDNFR3', description='GDNF family receptor alpha 3', other_designations='GDNF receptor alpha-3|GDNFR-alpha-3|GFR-alpha-3|GPI-linked receptor|glial cell line-derived neurotrophic factor receptor alpha-3', map_location='5q31.1-q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138252380, end_position_on_the_genomic_accession=138274564, orientation='minus', exon_count=8, OMIM=605710, no_hdr0=''),
  2677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2677, CurrentID=0, Status='live', Symbol='GGCX', Aliases='VKCFD1', description='gamma-glutamyl carboxylase', other_designations='peptidyl-glutamate 4-carboxylase', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85544855, end_position_on_the_genomic_accession=85561534, orientation='minus', exon_count=15, OMIM=137167, no_hdr0=''),
  2678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2678, CurrentID=0, Status='live', Symbol='GGT1', Aliases='CD224, D22S672, D22S732, GGT, GGT 1, GTG', description='gamma-glutamyltransferase 1', other_designations='glutathione hydrolase 1|leukotriene-C4 hydrolase', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24583750, end_position_on_the_genomic_accession=24629005, orientation='plus', exon_count=18, OMIM=612346, no_hdr0=''),
  2679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2679, CurrentID=0, Status='live', Symbol='GGT3P', Aliases='GGT3', description='gamma-glutamyltransferase 3 pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18773689, end_position_on_the_genomic_accession=18791961, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  2681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2681, CurrentID=0, Status='live', Symbol='GGTA1P', Aliases='GGTA, GGTA1, GLYT2, a1/3GTP', description='glycoprotein, alpha-galactosyltransferase 1 pseudogene', other_designations='alpha-1,3-galactosyltransferase pseudogene', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121455041, end_position_on_the_genomic_accession=121500027, orientation='minus', exon_count=7, OMIM=104175, no_hdr0=''),
  2683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2683, CurrentID=0, Status='live', Symbol='B4GALT1', Aliases='B4GAL-T1, CDG2D, GGTB2, GT1, GTB, beta4Gal-T1', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 1', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 1|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 1|beta-1,4-GalTase 1|glycoprotein-4-beta-galactosyltransferase 2|lactose synthase', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33110637, end_position_on_the_genomic_accession=33167358, orientation='minus', exon_count=6, OMIM=137060, no_hdr0=''),
  2686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2686, CurrentID=0, Status='live', Symbol='GGT7', Aliases='D20S101, GGT4, GGTL3, GGTL5, dJ18C9.2', description='gamma-glutamyltransferase 7', other_designations='GGT 7|gamma-glutamyltransferase 4|gamma-glutamyltransferase-like 3|gamma-glutamyltransferase-like 5|gamma-glutamyltranspeptidase 7|gamma-glutamyltranspeptidase-like 3|glutathione hydrolase 7', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34844720, end_position_on_the_genomic_accession=34872858, orientation='minus', exon_count=18, OMIM=612342, no_hdr0=''),
  2687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2687, CurrentID=0, Status='live', Symbol='GGT5', Aliases='GGL, GGT 5, GGT-REL, GGTLA1', description='gamma-glutamyltransferase 5', other_designations='gamma-glutamyl cleaving enzyme|gamma-glutamyl transpeptidase-related enzyme|gamma-glutamyl transpeptidase-related protein|gamma-glutamyltransferase-like activity 1|gamma-glutamyltranspeptidase 5|glutathione hydrolase 5|leukotriene-C4 hydrolase', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24219654, end_position_on_the_genomic_accession=24245142, orientation='minus', exon_count=14, OMIM=137168, no_hdr0=''),
  2688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2688, CurrentID=0, Status='live', Symbol='GH1', Aliases='GH, GH-N, GHB5, GHN, IGHD1B, hGH-N', description='growth hormone 1', other_designations='growth hormone B5|pituitary growth hormone', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63917193, end_position_on_the_genomic_accession=63918852, orientation='minus', exon_count=5, OMIM=139250, no_hdr0=''),
  2689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2689, CurrentID=0, Status='live', Symbol='GH2', Aliases='GH-V, GHB2, GHL, GHV, hGH-V', description='growth hormone 2', other_designations='growth hormone B2|placenta-specific growth hormone|placental-specific growth hormone', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63880212, end_position_on_the_genomic_accession=63881942, orientation='minus', exon_count=5, OMIM=139240, no_hdr0=''),
  2690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2690, CurrentID=0, Status='live', Symbol='GHR', Aliases='GHBP, GHIP', description='growth hormone receptor', other_designations='GH receptor|growth hormone binding protein|serum binding protein|somatotropin receptor', map_location='5p13-p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42423775, end_position_on_the_genomic_accession=42721878, orientation='plus', exon_count=18, OMIM=600946, no_hdr0=''),
  2691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2691, CurrentID=0, Status='live', Symbol='GHRH', Aliases='GHRF, GRF, INN', description='growth hormone releasing hormone', other_designations='growth hormone-releasing factor|growth hormone-releasing hormone|sermorelin|somatocrinin|somatorelin', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37251065, end_position_on_the_genomic_accession=37261878, orientation='minus', exon_count=8, OMIM=139190, no_hdr0=''),
  2692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2692, CurrentID=0, Status='live', Symbol='GHRHR', Aliases='GHRFR, GRFR, IGHD1B', description='growth hormone releasing hormone receptor', other_designations='GHRH receptor|GRF receptor', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30964021, end_position_on_the_genomic_accession=30979531, orientation='plus', exon_count=13, OMIM=139191, no_hdr0=''),
  2693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2693, CurrentID=0, Status='live', Symbol='GHSR', Aliases='GHDP', description='growth hormone secretagogue receptor', other_designations='GH-releasing peptide receptor|GHRP|GHS-R|ghrelin receptor', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172443291, end_position_on_the_genomic_accession=172448456, orientation='minus', exon_count=2, OMIM=601898, no_hdr0=''),
  2694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2694, CurrentID=0, Status='live', Symbol='GIF', Aliases='IF, IFMH, INF, TCN3', description='gastric intrinsic factor', other_designations='gastric intrinsic factor (vitamin B synthesis)|intrinsic factor', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59829268, end_position_on_the_genomic_accession=59845537, orientation='minus', exon_count=9, OMIM=609342, no_hdr0=''),
  2695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2695, CurrentID=0, Status='live', Symbol='GIP', Aliases='', description='gastric inhibitory polypeptide', other_designations='glucose-dependent insulinotropic polypeptide|incretin hormone', map_location='17q21.3-q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48958556, end_position_on_the_genomic_accession=48968593, orientation='minus', exon_count=6, OMIM=137240, no_hdr0=''),
  2696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2696, CurrentID=0, Status='live', Symbol='GIPR', Aliases='PGQTL2', description='gastric inhibitory polypeptide receptor', other_designations='GIP-R|glucose-dependent insulinotropic polypeptide receptor', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45668193, end_position_on_the_genomic_accession=45682504, orientation='plus', exon_count=14, OMIM=137241, no_hdr0=''),
  2697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2697, CurrentID=0, Status='live', Symbol='GJA1', Aliases='AVSD3, CMDR, CX43, EKVP, GJAL, HLHS1, HSS, ODDD, PPKCA', description='gap junction protein alpha 1', other_designations='connexin 43|connexin-43|gap junction 43 kDa heart protein', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121435577, end_position_on_the_genomic_accession=121449744, orientation='plus', exon_count=2, OMIM=121014, no_hdr0=''),
  2698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2698, CurrentID=0, Status='live', Symbol='GJA1P1', Aliases='', description='gap junction protein alpha 1 pseudogene 1', other_designations='gap junction protein, alpha 1, 43kDa pseudogene 1', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109049582, end_position_on_the_genomic_accession=109052667, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2700, CurrentID=0, Status='live', Symbol='GJA3', Aliases='CTRCT14, CX46, CZP3', description='gap junction protein alpha 3', other_designations='connexin 46|connexin-46', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20138255, end_position_on_the_genomic_accession=20161385, orientation='minus', exon_count=3, OMIM=121015, no_hdr0=''),
  2701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2701, CurrentID=0, Status='live', Symbol='GJA4', Aliases='CX37', description='gap junction protein alpha 4', other_designations='connexin 37|connexin-37', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34792958, end_position_on_the_genomic_accession=34795750, orientation='plus', exon_count=2, OMIM=121012, no_hdr0=''),
  2702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2702, CurrentID=0, Status='live', Symbol='GJA5', Aliases='ATFB11, CX40', description='gap junction protein alpha 5', other_designations='connexin 40|gap junction protein alpha 5 40kDa', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147756199, end_position_on_the_genomic_accession=147781067, orientation='minus', exon_count=5, OMIM=121013, no_hdr0=''),
  2703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2703, CurrentID=0, Status='live', Symbol='GJA8', Aliases='CAE, CAE1, CTRCT1, CX50, CZP1, MP70', description='gap junction protein alpha 8', other_designations='cell surface glycoprotein|connexin 50|connexin-50|gap junction alpha 8|gap junction membrane channel protein alpha-8|gap junction protein alpha 8 50kDa|lens fiber protein MP70|lens intrinsic membrane protein MP70', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147902820, end_position_on_the_genomic_accession=147915287, orientation='plus', exon_count=3, OMIM=600897, no_hdr0=''),
  2705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2705, CurrentID=0, Status='live', Symbol='GJB1', Aliases='CMTX, CMTX1, CX32', description='gap junction protein beta 1', other_designations='GAP junction 28 kDa liver protein|connexin 32|connexin-32', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71215212, end_position_on_the_genomic_accession=71225215, orientation='plus', exon_count=4, OMIM=304040, no_hdr0=''),
  2706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2706, CurrentID=0, Status='live', Symbol='GJB2', Aliases='CX26, DFNA3, DFNA3A, DFNB1, DFNB1A, HID, KID, NSRD1, PPK', description='gap junction protein beta 2', other_designations='connexin 26|mutant gap junction protein beta 2', map_location='13q11-q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20187463, end_position_on_the_genomic_accession=20192975, orientation='minus', exon_count=3, OMIM=121011, no_hdr0=''),
  2707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2707, CurrentID=0, Status='live', Symbol='GJB3', Aliases='CX31, DFNA2, DFNA2B, EKV', description='gap junction protein beta 3', other_designations='connexin 31|connexin-31', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34781189, end_position_on_the_genomic_accession=34786366, orientation='plus', exon_count=3, OMIM=603324, no_hdr0=''),
  2709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2709, CurrentID=0, Status='live', Symbol='GJB5', Aliases='CX31.1', description='gap junction protein beta 5', other_designations='gap junction protein, beta 5 (connexin 31.1)|gap junction protein, beta 5, 31.1kDa|gap junctional protein cx31.1', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34755046, end_position_on_the_genomic_accession=34758512, orientation='plus', exon_count=2, OMIM=604493, no_hdr0=''),
  2710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2710, CurrentID=0, Status='live', Symbol='GK', Aliases='GK1D, GK', description='glycerol kinase', other_designations='ATP:glycerol 3-phosphotransferase|glycerokinase', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30653359, end_position_on_the_genomic_accession=30731462, orientation='plus', exon_count=23, OMIM=300474, no_hdr0=''),
  2712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2712, CurrentID=0, Status='live', Symbol='GK2', Aliases='GKP2, GKTA', description='glycerol kinase 2', other_designations='ATP:glycerol 3-phosphotransferase 2|glycerokinase 2|glycerol kinase pseudogene 2|glycerol kinase testis specific 2|glycerol kinase, testis specific 2', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79406353, end_position_on_the_genomic_accession=79408218, orientation='minus', exon_count=1, OMIM=600148, no_hdr0=''),
  2713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2713, CurrentID=0, Status='live', Symbol='GK3P', Aliases='GKP3, GKTB', description='glycerol kinase 3 pseudogene', other_designations='glycerol kinase pseudogene 3', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165277792, end_position_on_the_genomic_accession=165280023, orientation='minus', exon_count=1, OMIM=600149, no_hdr0=''),
  2716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2716, CurrentID=0, Status='live', Symbol='GK4P', Aliases='GKP6', description='glycerol kinase 4 pseudogene', other_designations='glycerol kinase pseudogene 6', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101778466, end_position_on_the_genomic_accession=101780653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2717, CurrentID=0, Status='live', Symbol='GLA', Aliases='GALA', description='galactosidase alpha', other_designations='agalsidase alfa|alpha-D-galactosidase A|alpha-D-galactoside galactohydrolase 1|alpha-gal A|galactosidase, alpha|melibiase', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101397791, end_position_on_the_genomic_accession=101408013, orientation='minus', exon_count=8, OMIM=300644, no_hdr0=''),
  133790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133790, CurrentID=0, Status='live', Symbol='RPS18P7', Aliases='RPS18_4_595', description='ribosomal protein S18 pseudogene 7', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91354440, end_position_on_the_genomic_accession=91354971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2719, CurrentID=0, Status='live', Symbol='GPC3', Aliases='DGSX, GTR2-2, MXR7, OCI-5, SDYS, SGB, SGBS, SGBS1', description='glypican 3', other_designations='glypican proteoglycan 3|heparan sulphate proteoglycan|intestinal protein OCI-5|secreted glypican-3', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133535745, end_position_on_the_genomic_accession=133985646, orientation='minus', exon_count=9, OMIM=300037, no_hdr0=''),
  2720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2720, CurrentID=0, Status='live', Symbol='GLB1', Aliases='EBP, ELNR1, MPS4B', description='galactosidase beta 1', other_designations='acid beta-galactosidase|elastin receptor 1, 67kDa|lactase', map_location='3p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32996608, end_position_on_the_genomic_accession=33097202, orientation='minus', exon_count=17, OMIM=611458, no_hdr0=''),
  2722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2722, CurrentID=0, Status='live', Symbol='GLC1B', Aliases='', description='glaucoma 1, open angle, B (adult-onset)', other_designations='', map_location='2cen-q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606689, no_hdr0=''),
  2723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2723, CurrentID=0, Status='live', Symbol='GLC1C', Aliases='', description='glaucoma 1, open angle, C', other_designations='', map_location='3q21-q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601682, no_hdr0=''),
  2724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2724, CurrentID=0, Status='live', Symbol='GLC1D', Aliases='', description='glaucoma 1, open angle, D (adult-onset)', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602429, no_hdr0=''),
  2727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2727, CurrentID=0, Status='live', Symbol='GLC2A', Aliases='', description='glaucoma 2, angle closure', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2728, CurrentID=0, Status='live', Symbol='GLC3B', Aliases='', description='glaucoma 3, primary infantile, B', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600975, no_hdr0=''),
  2729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2729, CurrentID=0, Status='live', Symbol='GCLC', Aliases='GCL, GCS, GLCL, GLCLC', description='glutamate-cysteine ligase, catalytic subunit', other_designations='GCS heavy chain|gamma-ECS|gamma-glutamylcysteine synthetase', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53497341, end_position_on_the_genomic_accession=53545129, orientation='minus', exon_count=16, OMIM=606857, no_hdr0=''),
  2730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2730, CurrentID=0, Status='live', Symbol='GCLM', Aliases='GLCLR', description='glutamate-cysteine ligase, modifier subunit', other_designations='GCS light chain|GSC light chain|gamma-ECS regulatory subunit|gamma-glutamylcysteine synthetase regulatory subunit|glutamate-cysteine ligase (gamma-glutamylcysteine synthetase), regulatory (30.8kD)|glutamate-cysteine ligase modifier subunit delta2 alternative splicing|glutamate-cysteine ligase regulatory protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93885205, end_position_on_the_genomic_accession=93909598, orientation='minus', exon_count=8, OMIM=601176, no_hdr0=''),
  2731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2731, CurrentID=0, Status='live', Symbol='GLDC', Aliases='GCE, GCSP, HYGN1', description='glycine dehydrogenase (decarboxylating)', other_designations='glycine cleavage system protein P|glycine decarboxylase P-protein|glycine dehydrogenase (aminomethyl-transferring)|glycine dehydrogenase [decarboxylating], mitochondrial', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6532464, end_position_on_the_genomic_accession=6645692, orientation='minus', exon_count=25, OMIM=238300, no_hdr0=''),
  2732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2732, CurrentID=0, Status='live', Symbol='GLDCP1', Aliases='GLDCP', description='glycine dehydrogenase (decarboxylase) pseudogene 1', other_designations='glycine dehydrogenase (decarboxylating) pseudogene', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56592465, end_position_on_the_genomic_accession=56596265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2733, CurrentID=0, Status='live', Symbol='GLE1', Aliases='GLE1L, LCCS, LCCS1, hGLE1', description='GLE1, RNA export mediator', other_designations='GLE1 RNA export mediator homolog|GLE1-like protein|GLE1-like, RNA export mediator', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128504692, end_position_on_the_genomic_accession=128542301, orientation='plus', exon_count=19, OMIM=603371, no_hdr0=''),
  2734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2734, CurrentID=0, Status='live', Symbol='GLG1', Aliases='CFR-1, ESL-1, MG-160, MG160', description='golgi glycoprotein 1', other_designations='E-selectin ligand 1|cysteine-rich fibroblast growth factor receptor|golgi sialoglycoprotein MG-160', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74447427, end_position_on_the_genomic_accession=74607144, orientation='minus', exon_count=27, OMIM=600753, no_hdr0=''),
  2735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2735, CurrentID=0, Status='live', Symbol='GLI1', Aliases='GLI', description='GLI family zinc finger 1', other_designations='GLI-Kruppel family member GLI1|glioma-associated oncogene 1|glioma-associated oncogene homolog 1 (zinc finger protein)|oncogene GLI', map_location='12q13.2-q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57460135, end_position_on_the_genomic_accession=57472268, orientation='plus', exon_count=13, OMIM=165220, no_hdr0=''),
  2736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2736, CurrentID=0, Status='live', Symbol='GLI2', Aliases='CJS, HPE9, PHS2, THP1, THP2', description='GLI family zinc finger 2', other_designations='GLI family zinc finger protein 2|GLI-Kruppel family member GLI2|glioma-associated oncogene family zinc finger 2|oncogene GLI2|tax helper protein 1|tax helper protein 2|tax-responsive element-2 holding protein|tax-responsive element-25-bp sequence binding protein', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120797291, end_position_on_the_genomic_accession=120992653, orientation='plus', exon_count=17, OMIM=165230, no_hdr0=''),
  2737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2737, CurrentID=0, Status='live', Symbol='GLI3', Aliases='ACLS, GCPS-190, GLI3FL, PAP-A, PAPA, PAPA1, PAPB, PHS, PPDIV, GLI3', description='GLI family zinc finger 3', other_designations='GLI-Kruppel family member GLI3|glioma-associated oncogene family zinc finger 3|oncogene GLI3|zinc finger protein GLI3', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41960949, end_position_on_the_genomic_accession=42264112, orientation='minus', exon_count=20, OMIM=165240, no_hdr0=''),
  2738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2738, CurrentID=0, Status='live', Symbol='GLI4', Aliases='HKR4, ZNF928', description='GLI family zinc finger 4', other_designations='GLI-Kruppel family member GLI4 (oncogene HKR4)|glioma-associated oncogene family zinc finger 4|krueppel-related zinc finger protein 4', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143267437, end_position_on_the_genomic_accession=143276931, orientation='plus', exon_count=4, OMIM=165280, no_hdr0=''),
  2739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2739, CurrentID=0, Status='live', Symbol='GLO1', Aliases='GLOD1, GLYI, HEL-S-74', description='glyoxalase I', other_designations='S-D-lactoylglutathione methylglyoxal lyase|aldoketomutase|epididymis secretory protein Li 74|glx I|glyoxalase domain containing 1|ketone-aldehyde mutase|lactoyl glutathione lyase|methylglyoxalase', map_location='6p21.3-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38675925, end_position_on_the_genomic_accession=38703176, orientation='minus', exon_count=6, OMIM=138750, no_hdr0=''),
  2740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2740, CurrentID=0, Status='live', Symbol='GLP1R', Aliases='', description='glucagon like peptide 1 receptor', other_designations='GLP-1 receptor|GLP-1-R|GLP-1R', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39048781, end_position_on_the_genomic_accession=39091303, orientation='plus', exon_count=14, OMIM=138032, no_hdr0=''),
  2741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2741, CurrentID=0, Status='live', Symbol='GLRA1', Aliases='HKPX1, STHE', description='glycine receptor alpha 1', other_designations='glycine receptor 48 kDa subunit|glycine receptor strychnine-binding subunit', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151822513, end_position_on_the_genomic_accession=151924866, orientation='minus', exon_count=10, OMIM=138491, no_hdr0=''),
  2742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2742, CurrentID=0, Status='live', Symbol='GLRA2', Aliases='GLR', description='glycine receptor alpha 2', other_designations='glycine receptor alpha 2 subunit|glycine receptor, alpha 2|glycine receptor, alpha-2 polypeptide', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14448779, end_position_on_the_genomic_accession=14731812, orientation='plus', exon_count=13, OMIM=305990, no_hdr0=''),
  2743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2743, CurrentID=0, Status='live', Symbol='GLRB', Aliases='HKPX2', description='glycine receptor beta', other_designations='glycine receptor 58 kDa subunit|glycine receptor, beta subunit', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157076125, end_position_on_the_genomic_accession=157172099, orientation='plus', exon_count=12, OMIM=138492, no_hdr0=''),
  2744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2744, CurrentID=0, Status='live', Symbol='GLS', Aliases='AAD20, GAC, GAM1, KGA, GLS', description='glutaminase', other_designations='K-glutaminase|L-glutamine amidohydrolase|glutaminase C|glutaminase, phosphate-activated', map_location='2q32-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190880821, end_position_on_the_genomic_accession=190965552, orientation='plus', exon_count=21, OMIM=138280, no_hdr0=''),
  2745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2745, CurrentID=0, Status='live', Symbol='GLRX', Aliases='GRX, GRX1', description='glutaredoxin', other_designations='TTase-1|glutaredoxin (thioltransferase)|thioltransferase|thioltransferase-1', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95813849, end_position_on_the_genomic_accession=95822873, orientation='minus', exon_count=3, OMIM=600443, no_hdr0=''),
  2746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2746, CurrentID=0, Status='live', Symbol='GLUD1', Aliases='GDH, GDH1, GLUD', description='glutamate dehydrogenase 1', other_designations='GDH 1|glutamate dehydrogenase (NAD(P)+)', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87050202, end_position_on_the_genomic_accession=87095019, orientation='minus', exon_count=17, OMIM=138130, no_hdr0=''),
  2747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2747, CurrentID=0, Status='live', Symbol='GLUD2', Aliases='GDH2, GLUDP1', description='glutamate dehydrogenase 2', other_designations='GDH 2|glutamate dehydrogenase pseudogene 1', map_location='Xq24-q25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121047608, end_position_on_the_genomic_accession=121049942, orientation='plus', exon_count=1, OMIM=300144, no_hdr0=''),
  2748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2748, CurrentID=0, Status='live', Symbol='LOC2748', Aliases='', description='glutamate dehydrogenase 1 pseudogene', other_designations='', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47558349, end_position_on_the_genomic_accession=47558750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2749, CurrentID=0, Status='live', Symbol='GLUD1P3', Aliases='C10orf102, GLUDP3', description='glutamate dehydrogenase 1 pseudogene 3', other_designations='glutamate dehydrogenase pseudogene 3', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73731541, end_position_on_the_genomic_accession=73735609, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2750, CurrentID=0, Status='live', Symbol='GLUD1P4', Aliases='GLUDP4', description='glutamate dehydrogenase 1 pseudogene 4', other_designations='glutamate dehydrogenase pseudogene 4', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59778519, end_position_on_the_genomic_accession=59780607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2751, CurrentID=0, Status='live', Symbol='GLUD1P5', Aliases='GLUDP5', description='glutamate dehydrogenase 1 pseudogene 5', other_designations='glutamate dehydrogenase pseudogene 5', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31617611, end_position_on_the_genomic_accession=31618201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2752, CurrentID=0, Status='live', Symbol='GLUL', Aliases='GLNS, GS, PIG43, PIG59', description='glutamate-ammonia ligase', other_designations='cell proliferation-inducing protein 59|glutamate decarboxylase|glutamine synthase|proliferation-inducing protein 43', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182378093, end_position_on_the_genomic_accession=182392206, orientation='minus', exon_count=9, OMIM=138290, no_hdr0=''),
  2753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2753, CurrentID=0, Status='live', Symbol='GLULP1', Aliases='GLULL1', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 1', other_designations='glutamate-ammonia ligase (glutamine synthase)-like 1|glutamate-ammonia ligase (glutamine synthetase)-like 1', map_location='5q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151848743, end_position_on_the_genomic_accession=151850918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2754, CurrentID=0, Status='live', Symbol='GLULP2', Aliases='GLULL2', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 2', other_designations='glutamate-ammonia ligase (glutamine synthase)-like 2|glutamate-ammonia ligase (glutamine synthetase)-like 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9956585, end_position_on_the_genomic_accession=9959914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2757, CurrentID=0, Status='live', Symbol='GLYB', Aliases='', description='glycine B complementing', other_designations='Glycine auxotroph B, complementation of hamster', map_location='8q21.1-qter', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=138480, no_hdr0=''),
  2759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2759, CurrentID=0, Status='live', Symbol='GLYS1', Aliases='', description='glycosuria 1, renal', other_designations='glucosuria, renal', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=233100, no_hdr0=''),
  2760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2760, CurrentID=0, Status='live', Symbol='GM2A', Aliases='GM2-AP, SAP-3', description='GM2 ganglioside activator', other_designations='cerebroside sulfate activator protein|shingolipid activator protein 3|sphingolipid activator protein 3', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151253052, end_position_on_the_genomic_accession=151270394, orientation='plus', exon_count=4, OMIM=613109, no_hdr0=''),
  2761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2761, CurrentID=0, Status='live', Symbol='GM2AP1', Aliases='GM2AP', description='GM2 ganglioside activator pseudogene 1', other_designations='GM2 ganglioside activator protein pseudogene', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145684489, end_position_on_the_genomic_accession=145688498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2762, CurrentID=0, Status='live', Symbol='GMDS', Aliases='GMD, SDR3E1', description='GDP-mannose 4,6-dehydratase', other_designations='GDP-D-mannose dehydratase|short chain dehydrogenase/reductase family 3E, member 1', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1623800, end_position_on_the_genomic_accession=2245634, orientation='minus', exon_count=20, OMIM=602884, no_hdr0=''),
  2763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2763, CurrentID=0, Status='live', Symbol='GMFA', Aliases='', description='glia maturation factor, alpha', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2764, CurrentID=0, Status='live', Symbol='GMFB', Aliases='GMF', description='glia maturation factor, beta', other_designations='GMF-beta', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54474484, end_position_on_the_genomic_accession=54489067, orientation='minus', exon_count=8, OMIM=601713, no_hdr0=''),
  2765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2765, CurrentID=0, Status='live', Symbol='GML', Aliases='LY6DL', description='glycosylphosphatidylinositol anchored molecule like', other_designations='GPI anchored molecule like protein|Glycosylphosphatidylinositol-anchored molecule-like protein|glycosylphosphatidylinositol anchored molecule like protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142833998, end_position_on_the_genomic_accession=142917044, orientation='plus', exon_count=8, OMIM=602370, no_hdr0=''),
  2766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2766, CurrentID=0, Status='live', Symbol='GMPR', Aliases='GMPR 11, GMPR', description='guanosine monophosphate reductase', other_designations="guanine monophosphate reductase|guanosine 5'-monophosphate oxidoreductase 1|guanosine monophosphate reductase 1", map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16238580, end_position_on_the_genomic_accession=16295549, orientation='plus', exon_count=10, OMIM=139265, no_hdr0=''),
  2767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2767, CurrentID=0, Status='live', Symbol='GNA11', Aliases='FBH, FBH2, FHH2, GNA-11, HHC2, HYPOC2', description='guanine nucleotide binding protein (G protein), alpha 11 (Gq class)', other_designations='G-protein subunit alpha-11|g alpha-11|guanine nucleotide-binding protein G(y) subunit alpha', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3094410, end_position_on_the_genomic_accession=3124002, orientation='plus', exon_count=7, OMIM=139313, no_hdr0=''),
  2768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2768, CurrentID=0, Status='live', Symbol='GNA12', Aliases='NNX3, RMP, gep', description='guanine nucleotide binding protein (G protein) alpha 12', other_designations='G-protein subunit alpha-12|WUGSC:H_GS165O14.2|g alpha-12', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2728105, end_position_on_the_genomic_accession=2870656, orientation='minus', exon_count=8, OMIM=604394, no_hdr0=''),
  2769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2769, CurrentID=0, Status='live', Symbol='GNA15', Aliases='GNA16', description='guanine nucleotide binding protein (G protein), alpha 15 (Gq class)', other_designations='G-protein subunit alpha-15|G-protein subunit alpha-16|epididymis tissue protein Li 17E|g alpha-15|g alpha-16|guanine nucleotide-binding protein subunit alpha-16', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3136032, end_position_on_the_genomic_accession=3163769, orientation='plus', exon_count=7, OMIM=139314, no_hdr0=''),
  2770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2770, CurrentID=0, Status='live', Symbol='GNAI1', Aliases='Gi', description='guanine nucleotide binding protein (G protein), alpha inhibiting activity polypeptide 1', other_designations='Gi1 protein alpha subunit|adenylate cyclase-inhibiting G alpha protein', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80134824, end_position_on_the_genomic_accession=80219409, orientation='plus', exon_count=9, OMIM=139310, no_hdr0=''),
  2771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2771, CurrentID=0, Status='live', Symbol='GNAI2', Aliases='GIPB, H_LUCA15.1, H_LUCA16.1, GNAI2', description='guanine nucleotide binding protein (G protein), alpha inhibiting activity polypeptide 2', other_designations='GTP-binding regulatory protein Gi alpha-2 chain|adenylate cyclase-inhibiting G alpha protein|guanine nucleotide-binding protein G(i), alpha-2 subunit', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50226688, end_position_on_the_genomic_accession=50259354, orientation='plus', exon_count=13, OMIM=139360, no_hdr0=''),
  2772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2772, CurrentID=0, Status='live', Symbol='GNAI2P1', Aliases='GNAI2A, GNAI2L', description='guanine nucleotide binding protein (G protein), alpha inhibiting activity polypeptide 2 pseudogene 1', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14254826, end_position_on_the_genomic_accession=14255586, orientation='plus', exon_count=0, OMIM=139180, no_hdr0=''),
  2773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2773, CurrentID=0, Status='live', Symbol='GNAI3', Aliases='87U6, ARCND1', description='guanine nucleotide binding protein (G protein), alpha inhibiting activity polypeptide 3', other_designations='g(i) alpha-3', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109548564, end_position_on_the_genomic_accession=109595843, orientation='plus', exon_count=9, OMIM=139370, no_hdr0=''),
  2774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2774, CurrentID=0, Status='live', Symbol='GNAL', Aliases='DYT25', description='guanine nucleotide binding protein (G protein), alpha activating activity polypeptide, olfactory type', other_designations='adenylate cyclase-stimulating G alpha protein, olfactory type|guanine nucleotide binding protein (G protein), alpha stimulating activity polypeptide, olfactory type', map_location='18p11.22-p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11689015, end_position_on_the_genomic_accession=11885685, orientation='plus', exon_count=16, OMIM=139312, no_hdr0=''),
  2775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2775, CurrentID=0, Status='live', Symbol='GNAO1', Aliases='EIEE17, G-ALPHA-o, GNAO', description='guanine nucleotide binding protein (G protein), alpha activating activity polypeptide O', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56191339, end_position_on_the_genomic_accession=56357444, orientation='plus', exon_count=12, OMIM=139311, no_hdr0=''),
  2776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2776, CurrentID=0, Status='live', Symbol='GNAQ', Aliases='CMC1, G-ALPHA-q, GAQ, SWS', description='guanine nucleotide binding protein (G protein), q polypeptide', other_designations='guanine nucleotide-binding protein alpha-q', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77716274, end_position_on_the_genomic_accession=78031449, orientation='minus', exon_count=7, OMIM=600998, no_hdr0=''),
  2777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2777, CurrentID=0, Status='live', Symbol='GNAQP1', Aliases='GNAQP', description='guanine nucleotide binding protein (G protein), q polypeptide pseudogene 1', other_designations='quanine nucleotide binding protein (G protein), q polypeptide pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131422742, end_position_on_the_genomic_accession=131424935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2778, CurrentID=0, Status='live', Symbol='GNAS', Aliases='AHO, C20orf451, GPSA, GSA, GSP, NESP, POH, SCG6, SgVI, GNAS', description='GNAS complex locus', other_designations='adenylate cyclase-stimulating G alpha protein|alternative gene product encoded by XL-exon|extra large alphas protein|guanine nucleotide binding protein (G protein), alpha stimulating activity polypeptide 1|guanine nucleotide regulatory protein|guanine nucleotide-binding protein G(s) subunit alpha isoforms XLas|neuroendocrine secretory protein|secretogranin VI', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58839740, end_position_on_the_genomic_accession=58911195, orientation='plus', exon_count=16, OMIM=139320, no_hdr0=''),
  2779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2779, CurrentID=0, Status='live', Symbol='GNAT1', Aliases='CSNB1G, CSNBAD3, GBT1, GNATR', description='guanine nucleotide binding protein (G protein), alpha transducing activity polypeptide 1', other_designations='guanine nucleotide-binding protein G(T), alpha-1 subunit|rod-type transducin alpha subunit|transducin alpha-1 chain|transducin, rod-specific', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50191610, end_position_on_the_genomic_accession=50197696, orientation='plus', exon_count=10, OMIM=139330, no_hdr0=''),
  2780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2780, CurrentID=0, Status='live', Symbol='GNAT2', Aliases='ACHM4, GNATC', description='guanine nucleotide binding protein (G protein), alpha transducing activity polypeptide 2', other_designations='cone-type transducin alpha subunit|transducin alpha-2 chain|transducin, cone-specific, alpha polypeptide', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109602906, end_position_on_the_genomic_accession=109619662, orientation='minus', exon_count=9, OMIM=139340, no_hdr0=''),
  2781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2781, CurrentID=0, Status='live', Symbol='GNAZ', Aliases='', description='guanine nucleotide binding protein (G protein), alpha z polypeptide', other_designations='g(x) alpha chain|gz-alpha|transducin alpha', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23070483, end_position_on_the_genomic_accession=23125034, orientation='plus', exon_count=7, OMIM=139160, no_hdr0=''),
  2782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2782, CurrentID=0, Status='live', Symbol='GNB1', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 1', other_designations='G protein, beta-1 subunit|beta subunit, signal-transducing proteins GS/GI|transducin beta chain 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1785285, end_position_on_the_genomic_accession=1891117, orientation='minus', exon_count=12, OMIM=139380, no_hdr0=''),
  2783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2783, CurrentID=0, Status='live', Symbol='GNB2', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 2', other_designations='G protein, beta-2 subunit|g protein subunit beta-2|guanine nucleotide-binding protein G(I)/G(S)/G(T) beta subunit 2|signal-transducing guanine nucleotide-binding regulatory protein beta subunit|transducin beta chain 2', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100673740, end_position_on_the_genomic_accession=100679169, orientation='plus', exon_count=10, OMIM=139390, no_hdr0=''),
  2784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2784, CurrentID=0, Status='live', Symbol='GNB3', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 3', other_designations='G protein, beta-3 subunit|GTP-binding regulatory protein beta-3 chain|guanine nucleotide-binding protein G(I)/G(S)/G(T) beta subunit 3|hypertension associated protein|transducin beta chain 3', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6840854, end_position_on_the_genomic_accession=6847395, orientation='plus', exon_count=10, OMIM=139130, no_hdr0=''),
  2785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2785, CurrentID=0, Status='live', Symbol='GNG3', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 3', other_designations='NBP gamma-3|guanine nucleotide-binding protein gamma-3 subunit', map_location='11p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62706422, end_position_on_the_genomic_accession=62709206, orientation='plus', exon_count=5, OMIM=608941, no_hdr0=''),
  2786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2786, CurrentID=0, Status='live', Symbol='GNG4', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 4', other_designations='guanine nucleotide binding protein 4', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235547685, end_position_on_the_genomic_accession=235650754, orientation='minus', exon_count=10, OMIM=604388, no_hdr0=''),
  2787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2787, CurrentID=0, Status='live', Symbol='GNG5', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 5', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84498323, end_position_on_the_genomic_accession=84506579, orientation='minus', exon_count=4, OMIM=600874, no_hdr0=''),
  2788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2788, CurrentID=0, Status='live', Symbol='GNG7', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 7', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2511220, end_position_on_the_genomic_accession=2702748, orientation='minus', exon_count=5, OMIM=604430, no_hdr0=''),
  2789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2789, CurrentID=0, Status='live', Symbol='GNG5P1', Aliases='GNG5P', description='guanine nucleotide binding protein (G protein), gamma 5 pseudogene 1', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166239928, end_position_on_the_genomic_accession=166240539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2790, CurrentID=0, Status='live', Symbol='GNG10', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 10', other_designations='guanine nucleotide binding protein 10', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111661571, end_position_on_the_genomic_accession=111670246, orientation='plus', exon_count=3, OMIM=604389, no_hdr0=''),
  2791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2791, CurrentID=0, Status='live', Symbol='GNG11', Aliases='GNGT11', description='guanine nucleotide binding protein (G protein), gamma 11', other_designations='G protein gamma-11 subunit|guanine nucleotide-binding protein G(I)/G(S)/G(O) gamma-11 subunit', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93921704, end_position_on_the_genomic_accession=93926514, orientation='plus', exon_count=2, OMIM=604390, no_hdr0=''),
  2792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2792, CurrentID=0, Status='live', Symbol='GNGT1', Aliases='GNG1', description='guanine nucleotide binding protein (G protein), gamma transducing activity polypeptide 1', other_designations='transducin gamma chain', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93906508, end_position_on_the_genomic_accession=93911173, orientation='plus', exon_count=3, OMIM=189970, no_hdr0=''),
  2793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2793, CurrentID=0, Status='live', Symbol='GNGT2', Aliases='G-GAMMA-8, G-GAMMA-C, GNG8, GNG9, GNGT8', description='guanine nucleotide binding protein (G protein), gamma transducing activity polypeptide 2', other_designations='G protein cone gamma 8 subunit|G-gamma-9|g gamma-C|gamma-T2 subunit|guanine nucleotide binding protein gamma 9|guanine nucleotide binding protein gamma transducing activity polypeptide 2', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49204442, end_position_on_the_genomic_accession=49210574, orientation='minus', exon_count=5, OMIM=139391, no_hdr0=''),
  2794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2794, CurrentID=0, Status='live', Symbol='GNL1', Aliases='HSR1', description='guanine nucleotide binding protein-like 1', other_designations='GTP-binding protein HSR1|HSR1 GTP-binding protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30541377, end_position_on_the_genomic_accession=30557594, orientation='minus', exon_count=13, OMIM=143024, no_hdr0=''),
  2796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2796, CurrentID=0, Status='live', Symbol='GNRH1', Aliases='GNRH, GRH, LHRH, LNRH', description='gonadotropin releasing hormone 1', other_designations='GnRH-associated peptide 1|gonadotropin-releasing hormone 1 (luteinizing-releasing hormone)|leuteinizing-releasing hormone|luliberin I|prolactin release-inhibiting factor', map_location='8p21-p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25419258, end_position_on_the_genomic_accession=25425040, orientation='minus', exon_count=4, OMIM=152760, no_hdr0=''),
  2797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2797, CurrentID=0, Status='live', Symbol='GNRH2', Aliases='GnRH-II, LH-RHII', description='gonadotropin releasing hormone 2', other_designations='GnRH-associated peptide 2|GnRH-associated peptide II|Gonadoliberin II|Gonadoliberin-2|gonadotropin-releasing hormone 2|luliberin II|luteinizing hormone-releasing hormone II|progonadoliberin II', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3039061, end_position_on_the_genomic_accession=3045860, orientation='plus', exon_count=5, OMIM=602352, no_hdr0=''),
  2798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2798, CurrentID=0, Status='live', Symbol='GNRHR', Aliases='GNRHR1, GRHR, HH7, LHRHR, LRHR', description='gonadotropin releasing hormone receptor', other_designations='gnRH receptor|gnRH-R|gonadotropin-releasing hormone (type 1) receptor 1|leutinizing hormone releasing horomone receptor|leutinizing-releasing hormone receptor|luliberin receptor|type I GnRH receptor', map_location='4q21.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67737375, end_position_on_the_genomic_accession=67756086, orientation='minus', exon_count=3, OMIM=138850, no_hdr0=''),
  2799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2799, CurrentID=0, Status='live', Symbol='GNS', Aliases='G6S', description='glucosamine (N-acetyl)-6-sulfatase', other_designations='glucosamine -6-sulfatase|glucosamine-6-sulfatase', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64713442, end_position_on_the_genomic_accession=64759446, orientation='minus', exon_count=14, OMIM=607664, no_hdr0=''),
  2800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2800, CurrentID=0, Status='live', Symbol='GOLGA1', Aliases='golgin-97', description='golgin A1', other_designations='gap junction protein, alpha 4, 37kD|golgi autoantigen, golgin subfamily a, 1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124878294, end_position_on_the_genomic_accession=124941114, orientation='minus', exon_count=24, OMIM=602502, no_hdr0=''),
  2801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2801, CurrentID=0, Status='live', Symbol='GOLGA2', Aliases='GM130', description='golgin A2', other_designations='130 kDa cis-Golgi matrix protein|GM130 autoantigen|Golgi matrix protein GM130|SY11 protein|golgi autoantigen, golgin subfamily a, 2|golgin-95', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128255829, end_position_on_the_genomic_accession=128276007, orientation='minus', exon_count=27, OMIM=602580, no_hdr0=''),
  2802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2802, CurrentID=0, Status='live', Symbol='GOLGA3', Aliases='GCP170, MEA-2', description='golgin A3', other_designations='Golgi membrane associated protein|Golgi peripheral membrane protein|SY2/SY10 protein|golgi autoantigen, golgin subfamily a, 3|golgi complex-associated protein of 170 kDa|golgin-160|golgin-165|male enhanced antigen-2', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132768909, end_position_on_the_genomic_accession=132829118, orientation='minus', exon_count=31, OMIM=602581, no_hdr0=''),
  2803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2803, CurrentID=0, Status='live', Symbol='GOLGA4', Aliases='CRPF46, GCP2, GOLG, MU-RMS-40.18, p230', description='golgin A4', other_designations='256 kDa golgin|72.1 protein|centrosome-related protein F46|golgi autoantigen, golgin subfamily a, 4|golgin-240|golgin-245|protein 72.1|trans-Golgi p230', map_location='3p22-p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37243191, end_position_on_the_genomic_accession=37366879, orientation='plus', exon_count=28, OMIM=602509, no_hdr0=''),
  2804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2804, CurrentID=0, Status='live', Symbol='GOLGB1', Aliases='GCP, GCP372, GOLIM1', description='golgin B1', other_designations='372 kDa Golgi complex-associated protein|giantin|golgi autoantigen, golgin subfamily b, macrogolgin (with transmembrane signal), 1|golgi integral membrane protein 1|golgin B1, golgi integral membrane protein|macrogolgin', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121663199, end_position_on_the_genomic_accession=121750811, orientation='minus', exon_count=26, OMIM=602500, no_hdr0=''),
  2805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2805, CurrentID=0, Status='live', Symbol='GOT1', Aliases='AST1, ASTQTL1, GIG18, cAspAT, cCAT', description='glutamic-oxaloacetic transaminase 1, soluble', other_designations='aspartate aminotransferase 1|aspartate transaminase 1|cysteine aminotransferase, cytoplasmic|cysteine transaminase, cytoplasmic|glutamate oxaloacetate transaminase 1|growth-inhibiting protein 18|transaminase A', map_location='10q24.1-q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99396870, end_position_on_the_genomic_accession=99430773, orientation='minus', exon_count=9, OMIM=138180, no_hdr0=''),
  2806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2806, CurrentID=0, Status='live', Symbol='GOT2', Aliases='KAT4, KATIV, mitAAT', description='glutamic-oxaloacetic transaminase 2', other_designations='FABP-1|FABPpm|aspartate aminotransferase 2|aspartate transaminase 2|fatty acid-binding protein|glutamate oxaloacetate transaminase 2|glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2)|kynurenine aminotransferase 4|kynurenine aminotransferase IV|kynurenine--oxoglutarate transaminase 4|kynurenine--oxoglutarate transaminase IV|mAspAT|plasma membrane-associated fatty acid-binding protein|transaminase A', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58707131, end_position_on_the_genomic_accession=58734357, orientation='minus', exon_count=10, OMIM=138150, no_hdr0=''),
  131540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131540, CurrentID=0, Status='live', Symbol='ZDHHC19', Aliases='DHHC19', description='zinc finger, DHHC-type containing 19', other_designations='DHHC-19|zinc finger DHHC domain-containing protein 19|zinc finger, DHHC domain containing 19', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196197443, end_position_on_the_genomic_accession=196211437, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  2810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2810, CurrentID=0, Status='live', Symbol='SFN', Aliases='YWHAS', description='stratifin', other_designations='14-3-3 sigma|epithelial cell marker protein 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26863142, end_position_on_the_genomic_accession=26864456, orientation='plus', exon_count=1, OMIM=601290, no_hdr0=''),
  2811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2811, CurrentID=0, Status='live', Symbol='GP1BA', Aliases='BDPLT1, BDPLT3, BSS, CD42B, CD42b-alpha, DBPLT3, GP1B, GPIbA, VWDP', description='glycoprotein Ib (platelet), alpha polypeptide', other_designations='GP-Ib alpha|antigen CD42b-alpha|platelet membrane glycoprotein 1b-alpha subunit', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4932275, end_position_on_the_genomic_accession=4935030, orientation='plus', exon_count=2, OMIM=606672, no_hdr0=''),
  2812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2812, CurrentID=0, Status='live', Symbol='GP1BB', Aliases='BDPLT1, BS, CD42C, GPIBB', description='glycoprotein Ib (platelet), beta polypeptide', other_designations='GP-Ib beta|antigen CD42b-beta|nuclear localization signal deleted in velocardiofacial syndrome', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19723543, end_position_on_the_genomic_accession=19724774, orientation='plus', exon_count=2, OMIM=138720, no_hdr0=''),
  2813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2813, CurrentID=0, Status='live', Symbol='GP2', Aliases='ZAP75', description='glycoprotein 2 (zymogen granule membrane)', other_designations='pancreatic zymogen granule membrane associated protein GP2', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20309574, end_position_on_the_genomic_accession=20327820, orientation='minus', exon_count=12, OMIM=602977, no_hdr0=''),
  2814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2814, CurrentID=0, Status='live', Symbol='GP5', Aliases='CD42d, GPV', description='glycoprotein V (platelet)', other_designations='glycoprotein 5', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194394821, end_position_on_the_genomic_accession=194399266, orientation='minus', exon_count=2, OMIM=173511, no_hdr0=''),
  2815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2815, CurrentID=0, Status='live', Symbol='GP9', Aliases='CD42a, GPIX', description='glycoprotein IX (platelet)', other_designations='glycoprotein 9', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129055449, end_position_on_the_genomic_accession=129062411, orientation='plus', exon_count=6, OMIM=173515, no_hdr0=''),
  2817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2817, CurrentID=0, Status='live', Symbol='GPC1', Aliases='glypican', description='glypican 1', other_designations='glypican proteoglycan 1', map_location='2q35-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240435698, end_position_on_the_genomic_accession=240468078, orientation='plus', exon_count=10, OMIM=600395, no_hdr0=''),
  2819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2819, CurrentID=0, Status='live', Symbol='GPD1', Aliases='GPD-C, GPDH-C, HTGTI', description='glycerol-3-phosphate dehydrogenase 1', other_designations='glycerol-3-phosphate dehydrogenase 1 (soluble)|glycerol-3-phosphate dehydrogenase [NAD+], cytoplasmic|glycerophosphate dehydrogenase', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50103819, end_position_on_the_genomic_accession=50111320, orientation='plus', exon_count=8, OMIM=138420, no_hdr0=''),
  2820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2820, CurrentID=0, Status='live', Symbol='GPD2', Aliases='GDH2, GPDM, mGPDH', description='glycerol-3-phosphate dehydrogenase 2', other_designations='GPD-M|GPDH-M|glycerol-3-phosphate dehydrogenase 2 (mitochondrial)|mitochondrial glycerophosphate dehydrogenase|mtGPD', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156435290, end_position_on_the_genomic_accession=156586403, orientation='plus', exon_count=23, OMIM=138430, no_hdr0=''),
  2821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2821, CurrentID=0, Status='live', Symbol='GPI', Aliases='AMF, GNPI, NLK, PGI, PHI, SA-36, SA36', description='glucose-6-phosphate isomerase', other_designations='autocrine motility factor|glucose phosphate isomerase|hexose monophosphate isomerase|hexosephosphate isomerase|neuroleukin|oxoisomerase|phosphoglucose isomerase|phosphohexomutase|phosphohexose isomerase|phosphosaccharomutase|sperm antigen 36|sperm antigen-36', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34359592, end_position_on_the_genomic_accession=34402413, orientation='plus', exon_count=22, OMIM=172400, no_hdr0=''),
  2822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2822, CurrentID=0, Status='live', Symbol='GPLD1', Aliases='GPIPLD, GPIPLDM, PIGPLD, PIGPLD1, PLD', description='glycosylphosphatidylinositol specific phospholipase D1', other_designations='GPI-PLD|GPI-specific phospholipase D|PI-G PLD|glycoprotein phospholipase D', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24425832, end_position_on_the_genomic_accession=24495043, orientation='minus', exon_count=26, OMIM=602515, no_hdr0=''),
  2823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2823, CurrentID=0, Status='live', Symbol='GPM6A', Aliases='GPM6, M6A', description='glycoprotein M6A', other_designations='', map_location='4q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175632937, end_position_on_the_genomic_accession=176002691, orientation='minus', exon_count=11, OMIM=601275, no_hdr0=''),
  2824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2824, CurrentID=0, Status='live', Symbol='GPM6B', Aliases='M6B', description='glycoprotein M6B', other_designations='protolipid M6B', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13770939, end_position_on_the_genomic_accession=13938732, orientation='minus', exon_count=10, OMIM=300051, no_hdr0=''),
  2825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2825, CurrentID=0, Status='live', Symbol='GPR1', Aliases='', description='G protein-coupled receptor 1', other_designations='probable G-protein coupled receptor 1', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206175316, end_position_on_the_genomic_accession=206218047, orientation='minus', exon_count=6, OMIM=600239, no_hdr0=''),
  2826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2826, CurrentID=0, Status='live', Symbol='CCR10', Aliases='GPR2', description='chemokine (C-C motif) receptor 10', other_designations='C-C CKR-10|CC chemokine receptor 10|CC-CKR-10|CCR-10|G protein-coupled receptor 2|G-protein coupled receptor 2', map_location='17q21.1-q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42679402, end_position_on_the_genomic_accession=42681827, orientation='minus', exon_count=2, OMIM=600240, no_hdr0=''),
  2827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2827, CurrentID=0, Status='live', Symbol='GPR3', Aliases='ACCA', description='G protein-coupled receptor 3', other_designations='ACCA orphan receptor|adenylate cyclase constitutive activator', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27392644, end_position_on_the_genomic_accession=27395814, orientation='plus', exon_count=2, OMIM=600241, no_hdr0=''),
  2828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2828, CurrentID=0, Status='live', Symbol='GPR4', Aliases='', description='G protein-coupled receptor 4', other_designations='G-protein coupled receptor 19', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45589764, end_position_on_the_genomic_accession=45602208, orientation='minus', exon_count=2, OMIM=600551, no_hdr0=''),
  2829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2829, CurrentID=0, Status='live', Symbol='XCR1', Aliases='CCXCR1, GPR5', description='chemokine (C motif) receptor 1', other_designations='G protein-coupled receptor 5|G-protein coupled receptor 5|XC chemokine receptor 1|chemokine (C motif) XC receptor 1|lymphotactin receptor', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46020799, end_position_on_the_genomic_accession=46027487, orientation='minus', exon_count=3, OMIM=600552, no_hdr0=''),
  2830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2830, CurrentID=0, Status='live', Symbol='GPR6', Aliases='', description='G protein-coupled receptor 6', other_designations='sphingosine 1-phosphate receptor GPR6', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109978256, end_position_on_the_genomic_accession=109980720, orientation='plus', exon_count=3, OMIM=600553, no_hdr0=''),
  2831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2831, CurrentID=0, Status='live', Symbol='NPBWR1', Aliases='GPR7', description='neuropeptides B/W receptor 1', other_designations='G protein-coupled receptor 7|G-protein coupled receptor 7|neuropeptide B/W receptor 1|opioid-somatostatin-like receptor 7', map_location='8p22-q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52939908, end_position_on_the_genomic_accession=52940894, orientation='plus', exon_count=1, OMIM=600730, no_hdr0=''),
  2832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2832, CurrentID=0, Status='live', Symbol='NPBWR2', Aliases='GPR8', description='neuropeptides B/W receptor 2', other_designations='G protein-coupled receptor 8|G-protein coupled receptor 8|opioid-somatostatin-like receptor 8', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64105830, end_position_on_the_genomic_accession=64106831, orientation='minus', exon_count=1, OMIM=600731, no_hdr0=''),
  2833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2833, CurrentID=0, Status='live', Symbol='CXCR3', Aliases='CD182, CD183, CKR-L2, CMKAR3, GPR9, IP10-R, Mig-R, MigR', description='chemokine (C-X-C motif) receptor 3', other_designations='G protein-coupled receptor 9|IP-10 receptor|Mig receptor|chemokine receptor 3|interferon-inducible protein 10 receptor', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71615914, end_position_on_the_genomic_accession=71618517, orientation='minus', exon_count=3, OMIM=300574, no_hdr0=''),
  2834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2834, CurrentID=0, Status='live', Symbol='PRLHR', Aliases='GPR10, GR3, PrRPR', description='prolactin releasing hormone receptor', other_designations='G protein-coupled receptor 10|G-protein coupled receptor 10|hGR3|prRP receptor|prolactin releasing peptide receptor|prolactin-releasing hormone receptor', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118591925, end_position_on_the_genomic_accession=118599895, orientation='minus', exon_count=3, OMIM=600895, no_hdr0=''),
  2835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2835, CurrentID=0, Status='live', Symbol='GPR12', Aliases='GPCR12, GPCR21, PPP1R84', description='G protein-coupled receptor 12', other_designations='protein phosphatase 1, regulatory subunit 84', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26755200, end_position_on_the_genomic_accession=26760785, orientation='minus', exon_count=2, OMIM=600752, no_hdr0=''),
  2837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2837, CurrentID=0, Status='live', Symbol='UTS2R', Aliases='GPR14, UR-2-R, UTR, UTR2', description='urotensin 2 receptor', other_designations='G protein-coupled receptor 14|G-protein coupled receptor 14|UR-II-R|urotensin II receptor', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82374277, end_position_on_the_genomic_accession=82375586, orientation='plus', exon_count=1, OMIM=600896, no_hdr0=''),
  2838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2838, CurrentID=0, Status='live', Symbol='GPR15', Aliases='BOB', description='G protein-coupled receptor 15', other_designations='brother of Bonzo', map_location='3q11.2-q13.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98531899, end_position_on_the_genomic_accession=98533150, orientation='plus', exon_count=1, OMIM=601166, no_hdr0=''),
  2840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2840, CurrentID=0, Status='live', Symbol='GPR17', Aliases='', description='G protein-coupled receptor 17', other_designations='G-protein coupled receptor 17|P2Y-like receptor|R12|UDP/CysLT receptor', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127645864, end_position_on_the_genomic_accession=127652639, orientation='plus', exon_count=4, OMIM=603071, no_hdr0=''),
  2841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2841, CurrentID=0, Status='live', Symbol='GPR18', Aliases='', description='G protein-coupled receptor 18', other_designations='NAGly receptor', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99254713, end_position_on_the_genomic_accession=99262448, orientation='minus', exon_count=5, OMIM=602042, no_hdr0=''),
  2842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2842, CurrentID=0, Status='live', Symbol='GPR19', Aliases='', description='G protein-coupled receptor 19', other_designations='GPR-NGA', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12660884, end_position_on_the_genomic_accession=12717272, orientation='minus', exon_count=10, OMIM=602927, no_hdr0=''),
  2843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2843, CurrentID=0, Status='live', Symbol='GPR20', Aliases='', description='G protein-coupled receptor 20', other_designations='CTD-3064M3.3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141353407, end_position_on_the_genomic_accession=141367271, orientation='minus', exon_count=2, OMIM=601908, no_hdr0=''),
  2844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2844, CurrentID=0, Status='live', Symbol='GPR21', Aliases='', description='G protein-coupled receptor 21', other_designations='', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123030013, end_position_on_the_genomic_accession=123039101, orientation='plus', exon_count=2, OMIM=601909, no_hdr0=''),
  2845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2845, CurrentID=0, Status='live', Symbol='GPR22', Aliases='', description='G protein-coupled receptor 22', other_designations='', map_location='7q22-q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107470057, end_position_on_the_genomic_accession=107476212, orientation='plus', exon_count=4, OMIM=601910, no_hdr0=''),
  2846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2846, CurrentID=0, Status='live', Symbol='LPAR4', Aliases='GPR23, LPA4, P2RY9, P2Y5-LIKE, P2Y9', description='lysophosphatidic acid receptor 4', other_designations='G protein-coupled receptor 23|G-protein coupled receptor 23|LPA receptor 4|LPA-4|P2Y purinoceptor 9|P2Y5-like receptor|purinergic receptor 9', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78747709, end_position_on_the_genomic_accession=78757754, orientation='plus', exon_count=6, OMIM=300086, no_hdr0=''),
  2847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2847, CurrentID=0, Status='live', Symbol='MCHR1', Aliases='GPR24, MCH-1R, MCH1R, SLC-1, SLC1', description='melanin concentrating hormone receptor 1', other_designations='G protein-coupled receptor 24|MCH receptor 1|somatostatin receptor-like protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40679178, end_position_on_the_genomic_accession=40682814, orientation='plus', exon_count=2, OMIM=601751, no_hdr0=''),
  2848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2848, CurrentID=0, Status='live', Symbol='GPR25', Aliases='', description='G protein-coupled receptor 25', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200872955, end_position_on_the_genomic_accession=200874178, orientation='plus', exon_count=1, OMIM=602174, no_hdr0=''),
  2849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2849, CurrentID=0, Status='live', Symbol='GPR26', Aliases='', description='G protein-coupled receptor 26', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123666355, end_position_on_the_genomic_accession=123697399, orientation='plus', exon_count=3, OMIM=604847, no_hdr0=''),
  2850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2850, CurrentID=0, Status='live', Symbol='GPR27', Aliases='SREB1', description='G protein-coupled receptor 27', other_designations='super conserved receptor expressed in brain 1', map_location='3p21-p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71754050, end_position_on_the_genomic_accession=71755177, orientation='plus', exon_count=1, OMIM=605187, no_hdr0=''),
  133923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133923, CurrentID=0, Status='live', Symbol='ZNF474', Aliases='', description='zinc finger protein 474', other_designations='4933409D10Rik|TSZFP|testis-specific zinc finger protein', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122129513, end_position_on_the_genomic_accession=122153571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2852, CurrentID=0, Status='live', Symbol='GPER1', Aliases='CEPR, CMKRL2, DRY12, FEG-1, GPCR-Br, GPER, GPR30, LERGU, LERGU2, LyGPR, mER', description='G protein-coupled estrogen receptor 1', other_designations="G protein-coupled receptor 30|IL8-related receptor DRY12|chemoattractant receptor-like 2|chemokine receptor-like 2|constitutively expressed peptide-like receptor|flow-induced endothelial G-protein coupled receptor 1|heptahelix receptor|leucine rich protein in GPR30 3'UTR|lymphocyte-derived G-protein coupled receptor|membrane estrogen receptor", map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1086807, end_position_on_the_genomic_accession=1093815, orientation='plus', exon_count=3, OMIM=601805, no_hdr0=''),
  2853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2853, CurrentID=0, Status='live', Symbol='GPR31', Aliases='12-HETER, HETER, HETER1', description='G protein-coupled receptor 31', other_designations='12-(S)-HETE acid receptor|12-(S)-HETE receptor|12-(S)-hydroxy-5,8,10,14-eicosatetraenoic acid (HETE) receptor|G-protein coupled receptor 31|bA517H2.2 (G protein-coupled receptor 31)|hydroxyeicosatetraenoic (HETE) acid receptor 1|probable G-protein coupled receptor 31', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167156872, end_position_on_the_genomic_accession=167157831, orientation='minus', exon_count=1, OMIM=602043, no_hdr0=''),
  2854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2854, CurrentID=0, Status='live', Symbol='GPR32', Aliases='RVDR1', description='G protein-coupled receptor 32', other_designations='resolvin D1 receptor', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50770464, end_position_on_the_genomic_accession=50771732, orientation='plus', exon_count=1, OMIM=603195, no_hdr0=''),
  2855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2855, CurrentID=0, Status='live', Symbol='GPR32P1', Aliases='GPR32P, hGPCR39', description='G protein-coupled receptor 32, pseudogene 1', other_designations='Putative G-protein coupled receptor GPCR39', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50758233, end_position_on_the_genomic_accession=50759501, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2856, CurrentID=0, Status='live', Symbol='GPR33', Aliases='', description='G protein-coupled receptor 33 (gene/pseudogene)', other_designations='G protein-coupled receptor 33 pseudogene', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31482944, end_position_on_the_genomic_accession=31487945, orientation='minus', exon_count=2, OMIM=610118, no_hdr0=''),
  2857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2857, CurrentID=0, Status='live', Symbol='GPR34', Aliases='LYPSR1', description='G protein-coupled receptor 34', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41688959, end_position_on_the_genomic_accession=41697277, orientation='plus', exon_count=3, OMIM=300241, no_hdr0=''),
  646044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646044, CurrentID=0, Status='live', Symbol='LOC646044', Aliases='', description='single stranded DNA binding protein 4 pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12066734, end_position_on_the_genomic_accession=12068620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2859, CurrentID=0, Status='live', Symbol='GPR35', Aliases='', description='G protein-coupled receptor 35', other_designations='KYNA receptor|kynurenic acid receptor', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240605408, end_position_on_the_genomic_accession=240631259, orientation='plus', exon_count=7, OMIM=602646, no_hdr0=''),
  2860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2860, CurrentID=0, Status='live', Symbol='GPR36', Aliases='', description='G protein-coupled receptor 36', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  2861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2861, CurrentID=0, Status='live', Symbol='GPR37', Aliases='EDNRBL, PAELR, hET(B)R-LP', description='G protein-coupled receptor 37 (endothelin receptor type B-like)', other_designations='ETBR-LP-1|G-protein coupled receptor 37|Parkin-associated endothelin receptor-like receptor|endothelin B receptor-like protein 1|probable G-protein coupled receptor 37', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124745601, end_position_on_the_genomic_accession=124766025, orientation='minus', exon_count=2, OMIM=602583, no_hdr0=''),
  2862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2862, CurrentID=0, Status='live', Symbol='MLNR', Aliases='GPR38, MTLR1', description='motilin receptor', other_designations='G protein-coupled receptor 38', map_location='13q14-q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49220338, end_position_on_the_genomic_accession=49222377, orientation='plus', exon_count=2, OMIM=602885, no_hdr0=''),
  2863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2863, CurrentID=0, Status='live', Symbol='GPR39', Aliases='', description='G protein-coupled receptor 39', other_designations='', map_location='2q21-q22', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132416161, end_position_on_the_genomic_accession=132646596, orientation='plus', exon_count=3, OMIM=602886, no_hdr0=''),
  2864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2864, CurrentID=0, Status='live', Symbol='FFAR1', Aliases='FFA1R, GPCR40, GPR40', description='free fatty acid receptor 1', other_designations='G protein-coupled receptor 40|G-protein coupled receptor 40', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35351542, end_position_on_the_genomic_accession=35352464, orientation='plus', exon_count=1, OMIM=603820, no_hdr0=''),
  2865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2865, CurrentID=0, Status='live', Symbol='FFAR3', Aliases='FFA3R, GPR41, GPR42', description='free fatty acid receptor 3', other_designations='G protein-coupled receptor 41|G-protein coupled receptor 41', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35358209, end_position_on_the_genomic_accession=35360489, orientation='plus', exon_count=3, OMIM=603821, no_hdr0=''),
  2866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2866, CurrentID=0, Status='live', Symbol='GPR42', Aliases='FFAR1L, FFAR3L, GPR41LP, GPR42', description='G protein-coupled receptor 42 (gene/pseudogene)', other_designations='G protein-coupled receptor 42 pseudogene', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35371346, end_position_on_the_genomic_accession=35372600, orientation='plus', exon_count=0, OMIM=603822, no_hdr0=''),
  2867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2867, CurrentID=0, Status='live', Symbol='FFAR2', Aliases='FFA2R, GPR43', description='free fatty acid receptor 2', other_designations='G protein-coupled receptor 43|G-protein coupled receptor 43|fatty acid receptor 2|free fatty acid activated receptor 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35447902, end_position_on_the_genomic_accession=35448862, orientation='plus', exon_count=3, OMIM=603823, no_hdr0=''),
  2868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2868, CurrentID=0, Status='live', Symbol='GRK4', Aliases='GPRK2L, GPRK4a, IT11, GRK4', description='G protein-coupled receptor kinase 4', other_designations='ITI1', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2963505, end_position_on_the_genomic_accession=3040752, orientation='plus', exon_count=21, OMIM=137026, no_hdr0=''),
  2869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2869, CurrentID=0, Status='live', Symbol='GRK5', Aliases='GPRK5', description='G protein-coupled receptor kinase 5', other_designations='FP2025|g protein-coupled receptor kinase GRK5', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119207685, end_position_on_the_genomic_accession=119455619, orientation='plus', exon_count=17, OMIM=600870, no_hdr0=''),
  2870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2870, CurrentID=0, Status='live', Symbol='GRK6', Aliases='GPRK6', description='G protein-coupled receptor kinase 6', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177426496, end_position_on_the_genomic_accession=177442849, orientation='plus', exon_count=18, OMIM=600869, no_hdr0=''),
  2871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2871, CurrentID=0, Status='live', Symbol='GRK6P1', Aliases='GPRK6L, GPRK6P, GRK6PS', description='G protein-coupled receptor kinase 6 pseudogene 1', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21319027, end_position_on_the_genomic_accession=21320826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2872, CurrentID=0, Status='live', Symbol='MKNK2', Aliases='GPRK7, MNK2', description='MAP kinase interacting serine/threonine kinase 2', other_designations='G protein-coupled receptor kinase 7|MAP kinase signal-integrating kinase 2|MAPK signal-integrating kinase 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2037471, end_position_on_the_genomic_accession=2051244, orientation='minus', exon_count=14, OMIM=605069, no_hdr0=''),
  2873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2873, CurrentID=0, Status='live', Symbol='GPS1', Aliases='COPS1, CSN1', description='G protein pathway suppressor 1', other_designations='COP9 signalosome subunit 1|GPS-1|JAB1-containing signalosome subunit 1|SGN1|signalosome subunit 1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82051351, end_position_on_the_genomic_accession=82057470, orientation='plus', exon_count=15, OMIM=601934, no_hdr0=''),
  2874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2874, CurrentID=0, Status='live', Symbol='GPS2', Aliases='AMF-1', description='G protein pathway suppressor 2', other_designations='GPS-2', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7312659, end_position_on_the_genomic_accession=7315339, orientation='minus', exon_count=11, OMIM=601935, no_hdr0=''),
  2875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2875, CurrentID=0, Status='live', Symbol='GPT', Aliases='AAT1, ALT11, GPT', description='glutamic-pyruvate transaminase (alanine aminotransferase)', other_designations='GPT 1|glutamate pyruvate transaminase 1|glutamic--alanine transaminase 1|glutamic--pyruvic transaminase 1|glutamic-alanine transaminase 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144502221, end_position_on_the_genomic_accession=144507174, orientation='plus', exon_count=12, OMIM=138200, no_hdr0=''),
  2876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2876, CurrentID=0, Status='live', Symbol='GPX1', Aliases='GPXD, GSHPX1', description='glutathione peroxidase 1', other_designations='GPx-1|GSHPx-1|cellular glutathione peroxidase', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49357176, end_position_on_the_genomic_accession=49358358, orientation='minus', exon_count=2, OMIM=138320, no_hdr0=''),
  2877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2877, CurrentID=0, Status='live', Symbol='GPX2', Aliases='GI-GPx, GPRP, GPRP-2, GPx-2, GPx-GI, GSHPX-GI, GSHPx-2', description='glutathione peroxidase 2', other_designations='gastrointestinal glutathione peroxidase|glutathione peroxidase 2 (gastrointestinal)|glutathione peroxidase-related protein 2', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64939152, end_position_on_the_genomic_accession=64942905, orientation='minus', exon_count=4, OMIM=138319, no_hdr0=''),
  2878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2878, CurrentID=0, Status='live', Symbol='GPX3', Aliases='GPx-P, GSHPx-3, GSHPx-P', description='glutathione peroxidase 3', other_designations='GPx-3|extracellular glutathione peroxidase|glutathione peroxidase 3 (plasma)|plasma glutathione peroxidase', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151020438, end_position_on_the_genomic_accession=151028993, orientation='plus', exon_count=5, OMIM=138321, no_hdr0=''),
  2879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2879, CurrentID=0, Status='live', Symbol='GPX4', Aliases='GPx-4, GSHPx-4, MCSP, PHGPx, snGPx, snPHGPx', description='glutathione peroxidase 4', other_designations='phospholipid hydroperoxidase|sperm nucleus glutathione peroxidase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1103926, end_position_on_the_genomic_accession=1106789, orientation='plus', exon_count=8, OMIM=138322, no_hdr0=''),
  2880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2880, CurrentID=0, Status='live', Symbol='GPX5', Aliases='HEL-S-75p', description='glutathione peroxidase 5', other_designations='EGLP|GPx-5|GSHPx-5|epididymal androgen-related protein|epididymis secretory sperm binding protein Li 75p|epididymis-specific glutathione peroxidase-like protein', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28526012, end_position_on_the_genomic_accession=28534951, orientation='plus', exon_count=7, OMIM=603435, no_hdr0=''),
  2882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2882, CurrentID=0, Status='live', Symbol='GPX7', Aliases='CL683, GPX6, GPx-7, GSHPx-7, NPGPx', description='glutathione peroxidase 7', other_designations='glutathione peroxidase 6|non-selenocysteine containing phospholipid hydroperoxide glutathione peroxidase', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52602371, end_position_on_the_genomic_accession=52609051, orientation='plus', exon_count=3, OMIM=615784, no_hdr0=''),
  2884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2884, CurrentID=0, Status='live', Symbol='GPX1P2', Aliases='GPXL1, GPXL2, GPXP2, GPXP2P', description='glutathione peroxidase pseudogene 2', other_designations='human glutathione peroxidase (GPXP2) pseudogene', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27143123, end_position_on_the_genomic_accession=27143999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2885, CurrentID=0, Status='live', Symbol='GRB2', Aliases='ASH, EGFRBP-GRB2, Grb3-3, MST084, MSTP084, NCKAP2', description='growth factor receptor bound protein 2', other_designations='HT027|SH2/SH3 adapter GRB2|abundant SRC homology|epidermal growth factor receptor-binding protein GRB2|growth factor receptor-bound protein 3|protein Ash', map_location='17q24-q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75318076, end_position_on_the_genomic_accession=75405709, orientation='minus', exon_count=6, OMIM=108355, no_hdr0=''),
  2886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2886, CurrentID=0, Status='live', Symbol='GRB7', Aliases='', description='growth factor receptor bound protein 7', other_designations='B47|GRB7 adapter protein|epidermal growth factor receptor GRB-7', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39737909, end_position_on_the_genomic_accession=39747285, orientation='plus', exon_count=18, OMIM=601522, no_hdr0=''),
  2887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2887, CurrentID=0, Status='live', Symbol='GRB10', Aliases='GRB-IR, Grb-10, IRBP, MEG1, RSS', description='growth factor receptor bound protein 10', other_designations='GRB10 adapter protein|GRB10 adaptor protein|insulin receptor-binding protein Grb-IR|maternally expressed gene 1', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50590063, end_position_on_the_genomic_accession=50793462, orientation='minus', exon_count=35, OMIM=601523, no_hdr0=''),
  2888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2888, CurrentID=0, Status='live', Symbol='GRB14', Aliases='', description='growth factor receptor bound protein 14', other_designations='GRB14 adapter protein', map_location='2q22-q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164492406, end_position_on_the_genomic_accession=164621850, orientation='minus', exon_count=16, OMIM=601524, no_hdr0=''),
  2889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2889, CurrentID=0, Status='live', Symbol='RAPGEF1', Aliases='C3G, GRF2', description='Rap guanine nucleotide exchange factor 1', other_designations='CRK SH3-binding GNRP|Rap guanine nucleotide exchange factor (GEF) 1|guanine nucleotide-releasing factor 2 (specific for crk proto-oncogene)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131576770, end_position_on_the_genomic_accession=131739944, orientation='minus', exon_count=33, OMIM=600303, no_hdr0=''),
  2890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2890, CurrentID=0, Status='live', Symbol='GRIA1', Aliases='GLUH1, GLUR1, GLURA, GluA1, HBGR1', description='glutamate receptor, ionotropic, AMPA 1', other_designations='AMPA 1|AMPA-selective glutamate receptor 1|gluR-1|gluR-A|gluR-K1', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=153489696, end_position_on_the_genomic_accession=153813873, orientation='plus', exon_count=21, OMIM=138248, no_hdr0=''),
  2891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2891, CurrentID=0, Status='live', Symbol='GRIA2', Aliases='GLUR2, GLURB, GluA2, GluR-K2, HBGR2', description='glutamate receptor, ionotropic, AMPA 2', other_designations='AMPA-selective glutamate receptor 2|gluR-2|gluR-B|glutamate receptor B flip isoform', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157220584, end_position_on_the_genomic_accession=157366075, orientation='plus', exon_count=17, OMIM=138247, no_hdr0=''),
  2892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2892, CurrentID=0, Status='live', Symbol='GRIA3', Aliases='GLUR-C, GLUR-K3, GLUR3, GLURC, GluA3, MRX94', description='glutamate receptor, ionotropic, AMPA 3', other_designations='AMPA-selective glutamate receptor 3|dJ1171F9.1|gluR-3|glutamate receptor C|glutamate receptor subunit 3|glutamate receptor, ionotrophic, AMPA 3', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123184243, end_position_on_the_genomic_accession=123490915, orientation='plus', exon_count=19, OMIM=305915, no_hdr0=''),
  2893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2893, CurrentID=0, Status='live', Symbol='GRIA4', Aliases='GLUR4, GLUR4C, GLURD, GluA4', description='glutamate receptor, ionotropic, AMPA 4', other_designations='AMPA-selective glutamate receptor 4|gluR-4|gluR-D|glutamate receptor, ionotrophic, AMPA 4', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105609994, end_position_on_the_genomic_accession=105982092, orientation='plus', exon_count=23, OMIM=138246, no_hdr0=''),
  2894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2894, CurrentID=0, Status='live', Symbol='GRID1', Aliases='GluD1', description='glutamate receptor, ionotropic, delta 1', other_designations='gluR delta-1 subunit|glutamate receptor delta-1 subunit', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85599555, end_position_on_the_genomic_accession=86366493, orientation='minus', exon_count=17, OMIM=610659, no_hdr0=''),
  2895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2895, CurrentID=0, Status='live', Symbol='GRID2', Aliases='GluD2, SCAR18', description='glutamate receptor, ionotropic, delta 2', other_designations='gluR delta-2 subunit|glutamate receptor delta-2 subunit', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92304399, end_position_on_the_genomic_accession=93774556, orientation='plus', exon_count=20, OMIM=602368, no_hdr0=''),
  2896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2896, CurrentID=0, Status='live', Symbol='GRN', Aliases='CLN11, GEP, GP88, PCDGF, PEPI, PGRN', description='granulin', other_designations='PC cell-derived growth factor|acrogranin|granulin-epithelin|proepithelin|progranulin', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44345086, end_position_on_the_genomic_accession=44353106, orientation='plus', exon_count=13, OMIM=138945, no_hdr0=''),
  2897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2897, CurrentID=0, Status='live', Symbol='GRIK1', Aliases='EAA3, EEA3, GLR5, GLUR5, GluK1', description='glutamate receptor, ionotropic, kainate 1', other_designations='excitatory amino acid receptor 3|gluR-5|glutamate receptor 5', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29536933, end_position_on_the_genomic_accession=29939964, orientation='minus', exon_count=19, OMIM=138245, no_hdr0=''),
  2898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2898, CurrentID=0, Status='live', Symbol='GRIK2', Aliases='EAA4, GLR6, GLUK6, GLUR6, GluK2, MRT6', description='glutamate receptor, ionotropic, kainate 2', other_designations='bA487F5.1|excitatory amino acid receptor 4|gluR-6|glutamate receptor 6|glutamate receptor form A|glutamate receptor form B|glutamate receptor form C|glutamate receptor form D|glutamate receptor form E', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=101393708, end_position_on_the_genomic_accession=102070083, orientation='plus', exon_count=21, OMIM=138244, no_hdr0=''),
  2899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2899, CurrentID=0, Status='live', Symbol='GRIK3', Aliases='EAA5, GLR7, GLUR7, GluK3, GluR7a', description='glutamate receptor, ionotropic, kainate 3', other_designations='dJ1090M5.1 (glutamate receptor, ionotropic, kainate 3 (GLUR7))|excitatory amino acid receptor 5|gluR-7|glutamate receptor 7', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36795527, end_position_on_the_genomic_accession=37034548, orientation='minus', exon_count=18, OMIM=138243, no_hdr0=''),
  2900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2900, CurrentID=0, Status='live', Symbol='GRIK4', Aliases='EAA1, GRIK, GluK4, KA1', description='glutamate receptor, ionotropic, kainate 4', other_designations='excitatory amino acid receptor 1|glutamate receptor KA1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120511746, end_position_on_the_genomic_accession=120988805, orientation='plus', exon_count=27, OMIM=600282, no_hdr0=''),
  2901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2901, CurrentID=0, Status='live', Symbol='GRIK5', Aliases='EAA2, GRIK2, GluK5, KA2', description='glutamate receptor, ionotropic, kainate 5', other_designations='excitatory amino acid receptor 2|glutamate receptor KA2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41998316, end_position_on_the_genomic_accession=42071804, orientation='minus', exon_count=26, OMIM=600283, no_hdr0=''),
  2902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2902, CurrentID=0, Status='live', Symbol='GRIN1', Aliases='GluN1, MRD8, NMD-R1, NMDA1, NMDAR1, NR1', description='glutamate receptor, ionotropic, N-methyl D-aspartate 1', other_designations='N-methyl-D-aspartate receptor channel, subunit zeta-1|N-methyl-D-aspartate receptor subunit NR1|glutamate [NMDA] receptor subunit zeta 1|glutamate [NMDA] receptor subunit zeta-1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137139112, end_position_on_the_genomic_accession=137168762, orientation='plus', exon_count=22, OMIM=138249, no_hdr0=''),
  2903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2903, CurrentID=0, Status='live', Symbol='GRIN2A', Aliases='EPND, FESD, GluN2A, LKS, NMDAR2A, NR2A', description='glutamate receptor, ionotropic, N-methyl D-aspartate 2A', other_designations='N-methyl D-aspartate receptor subtype 2A|N-methyl-D-aspartate receptor channel, subunit epsilon-1|N-methyl-D-aspartate receptor subunit 2A', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9753404, end_position_on_the_genomic_accession=10183364, orientation='minus', exon_count=20, OMIM=138253, no_hdr0=''),
  2904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2904, CurrentID=0, Status='live', Symbol='GRIN2B', Aliases='EIEE27, GluN2B, MRD6, NMDAR2B, NR2B, hNR3', description='glutamate receptor, ionotropic, N-methyl D-aspartate 2B', other_designations='N-methyl D-aspartate receptor subtype 2B|N-methyl-D-aspartate receptor subunit 3|NR3|glutamate [NMDA] receptor subunit epsilon-2|glutamate receptor subunit epsilon-2', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13537337, end_position_on_the_genomic_accession=13982002, orientation='minus', exon_count=16, OMIM=138252, no_hdr0=''),
  2905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2905, CurrentID=0, Status='live', Symbol='GRIN2C', Aliases='GluN2C, NMDAR2C, NR2C', description='glutamate receptor, ionotropic, N-methyl D-aspartate 2C', other_designations='N-methyl D-aspartate receptor subtype 2C|N-methyl-D-aspartate receptor subunit 2C|glutamate [NMDA] receptor subunit epsilon-3', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74842021, end_position_on_the_genomic_accession=74861514, orientation='minus', exon_count=16, OMIM=138254, no_hdr0=''),
  2906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2906, CurrentID=0, Status='live', Symbol='GRIN2D', Aliases='EB11, GluN2D, NMDAR2D, NR2D', description='glutamate receptor, ionotropic, N-methyl D-aspartate 2D', other_designations='N-methyl D-aspartate receptor subtype 2D|N-methyl-d-aspartate receptor subunit 2D|estrogen receptor binding CpG island|glutamate [NMDA] receptor subunit epsilon-4', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48394875, end_position_on_the_genomic_accession=48444937, orientation='plus', exon_count=13, OMIM=602717, no_hdr0=''),
  2907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2907, CurrentID=0, Status='live', Symbol='GRINA', Aliases='HNRGW, LFG1, NMDARA1, TMBIM3', description='glutamate receptor, ionotropic, N-methyl D-aspartate-associated protein 1 (glutamate binding)', other_designations='NMDA receptor glutamate-binding subunit|glutamate [NMDA] receptor-associated protein 1|glutamate receptor, NMDA subtype, glutamate-binding subunit|putative MAPK-activating protein PM02|transmembrane BAX inhibitor motif containing 3|transmembrane BAX inhibitor motif-containing protein 3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143990058, end_position_on_the_genomic_accession=143993419, orientation='plus', exon_count=8, OMIM=138251, no_hdr0=''),
  2908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2908, CurrentID=0, Status='live', Symbol='NR3C1', Aliases='GCCR, GCR, GCRST, GR, GRL', description='nuclear receptor subfamily 3 group C member 1', other_designations='glucocorticoid nuclear receptor variant 1|nuclear receptor subfamily 3, group C, member 1 (glucocorticoid receptor)', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143277931, end_position_on_the_genomic_accession=143435512, orientation='minus', exon_count=16, OMIM=138040, no_hdr0=''),
  2909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2909, CurrentID=0, Status='live', Symbol='ARHGAP35', Aliases='GRF-1, GRLF1, P190-A, P190A, p190ARhoGAP, p190RhoGAP', description='Rho GTPase activating protein 35', other_designations='glucocorticoid receptor DNA binding factor 1|glucocorticoid receptor DNA-binding factor 1|glucocorticoid receptor repression factor 1|rho GAP p190A', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46918567, end_position_on_the_genomic_accession=47005077, orientation='plus', exon_count=8, OMIM=605277, no_hdr0=''),
  2910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2910, CurrentID=0, Status='live', Symbol='NR3C1P1', Aliases='GCR2, GCRL, GRLL1, NR3C1P', description='nuclear receptor subfamily 3 group C member 1 pseudogene 1', other_designations='glucocorticoid receptor-like 1|nuclear receptor subfamily 3, group C, member 1 pseudogene 1', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87467136, end_position_on_the_genomic_accession=87467685, orientation='plus', exon_count=0, OMIM=138060, no_hdr0=''),
  2911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2911, CurrentID=0, Status='live', Symbol='GRM1', Aliases='GPRC1A, MGLU1, MGLUR1, PPP1R85, SCAR13', description='glutamate receptor, metabotropic 1', other_designations='protein phosphatase 1, regulatory subunit 85', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145964769, end_position_on_the_genomic_accession=146437598, orientation='plus', exon_count=11, OMIM=604473, no_hdr0=''),
  2912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2912, CurrentID=0, Status='live', Symbol='GRM2', Aliases='GLUR2, GPRC1B, MGLUR2, mGlu2', description='glutamate receptor, metabotropic 2', other_designations='glutamate metabotropic receptor 2|glutamate receptor homolog', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51707033, end_position_on_the_genomic_accession=51718616, orientation='plus', exon_count=8, OMIM=604099, no_hdr0=''),
  2913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2913, CurrentID=0, Status='live', Symbol='GRM3', Aliases='GLUR3, GPRC1C, MGLUR3, mGlu3', description='glutamate receptor, metabotropic 3', other_designations='glutamate metabotropic receptor 3', map_location='7q21.1-q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=86643914, end_position_on_the_genomic_accession=86864879, orientation='plus', exon_count=8, OMIM=601115, no_hdr0=''),
  2914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2914, CurrentID=0, Status='live', Symbol='GRM4', Aliases='GPRC1D, MGLUR4, mGlu4', description='glutamate receptor, metabotropic 4', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34018643, end_position_on_the_genomic_accession=34155622, orientation='minus', exon_count=16, OMIM=604100, no_hdr0=''),
  2915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2915, CurrentID=0, Status='live', Symbol='GRM5', Aliases='GPRC1E, MGLUR5, PPP1R86, mGlu5', description='glutamate receptor, metabotropic 5', other_designations='protein phosphatase 1, regulatory subunit 86', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88504576, end_position_on_the_genomic_accession=89065961, orientation='minus', exon_count=13, OMIM=604102, no_hdr0=''),
  2916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2916, CurrentID=0, Status='live', Symbol='GRM6', Aliases='CSNB1B, GPRC1F, MGLUR6, mGlu6', description='glutamate receptor, metabotropic 6', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178978327, end_position_on_the_genomic_accession=178995123, orientation='minus', exon_count=10, OMIM=604096, no_hdr0=''),
  2917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2917, CurrentID=0, Status='live', Symbol='GRM7', Aliases='GLUR7, GPRC1G, MGLU7, MGLUR7, PPP1R87', description='glutamate receptor, metabotropic 7', other_designations='protein phosphatase 1, regulatory subunit 87', map_location='3p26.1-p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6861097, end_position_on_the_genomic_accession=7741531, orientation='plus', exon_count=17, OMIM=604101, no_hdr0=''),
  2918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2918, CurrentID=0, Status='live', Symbol='GRM8', Aliases='GLUR8, GPRC1H, MGLUR8, mGlu8', description='glutamate receptor, metabotropic 8', other_designations='', map_location='7q31.3-q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=126438598, end_position_on_the_genomic_accession=127253294, orientation='minus', exon_count=27, OMIM=601116, no_hdr0=''),
  2919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2919, CurrentID=0, Status='live', Symbol='CXCL1', Aliases='FSP, GRO1, GROa, MGSA, MGSA-a, NAP-3, SCYB1', description='chemokine (C-X-C motif) ligand 1 (melanoma growth stimulating activity, alpha)', other_designations='C-X-C motif chemokine 1|GRO-alpha(1-73)|GRO1 oncogene (melanoma growth stimulating activity, alpha)|GRO1 oncogene (melanoma growth-stimulating activity)|MGSA alpha|fibroblast secretory protein|melanoma growth stimulatory activity alpha|neutrophil-activating protein 3', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73869392, end_position_on_the_genomic_accession=73871302, orientation='plus', exon_count=4, OMIM=155730, no_hdr0=''),
  2920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2920, CurrentID=0, Status='live', Symbol='CXCL2', Aliases='CINC-2a, GRO2, GROb, MGSA-b, MIP-2a, MIP2, MIP2A, SCYB2', description='chemokine (C-X-C motif) ligand 2', other_designations='GRO2 oncogene|MGSA beta|MIP2-alpha|gro-beta|growth-regulated protein beta|macrophage inflammatory protein 2-alpha|melanoma growth stimulatory activity beta', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74097035, end_position_on_the_genomic_accession=74099280, orientation='minus', exon_count=4, OMIM=139110, no_hdr0=''),
  2921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2921, CurrentID=0, Status='live', Symbol='CXCL3', Aliases='CINC-2b, GRO3, GROg, MIP-2b, MIP2B, SCYB3', description='chemokine (C-X-C motif) ligand 3', other_designations='GRO-gamma|GRO-gamma(1-73)|GRO3 oncogene|MGSA gamma|MIP2-beta|growth-regulated protein gamma|macrophage inflammatory protein 2-beta|melanoma growth stimulatory activity gamma', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74036589, end_position_on_the_genomic_accession=74038773, orientation='minus', exon_count=4, OMIM=139111, no_hdr0=''),
  2922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2922, CurrentID=0, Status='live', Symbol='GRP', Aliases='BN-10, preproGRP, proGRP, GRP', description='gastrin-releasing peptide', other_designations='bombesin|neuromedin C|pre-progastrin releasing peptide|prepro-GRP', map_location='18q21.1-q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59219118, end_position_on_the_genomic_accession=59230771, orientation='plus', exon_count=7, OMIM=137260, no_hdr0=''),
  2923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2923, CurrentID=0, Status='live', Symbol='PDIA3', Aliases='ER60, ERp57, ERp60, ERp61, GRP57, GRP58, HEL-S-269, HEL-S-93n, HsT17083, P58, PI-PLC', description='protein disulfide isomerase family A member 3', other_designations='58 kDa glucose-regulated protein|58 kDa microsomal protein|ER protein 57|ER protein 60|disulfide isomerase ER-60|endoplasmic reticulum P58|endoplasmic reticulum resident protein 57|endoplasmic reticulum resident protein 60|epididymis secretory protein Li 269|epididymis secretory sperm binding protein Li 93n|glucose regulated protein, 58kDa|phospholipase C-alpha|protein disulfide isomerase family A, member 3|protein disulfide isomerase-associated 3', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43746392, end_position_on_the_genomic_accession=43772606, orientation='plus', exon_count=13, OMIM=602046, no_hdr0=''),
  2925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2925, CurrentID=0, Status='live', Symbol='GRPR', Aliases='BB2', description='gastrin releasing peptide receptor', other_designations='GRP-R|GRP-preferring bombesin receptor|bombesin receptor 2', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16123301, end_position_on_the_genomic_accession=16153518, orientation='plus', exon_count=3, OMIM=305670, no_hdr0=''),
  2926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2926, CurrentID=0, Status='live', Symbol='GRSF1', Aliases='', description='G-rich RNA sequence binding factor 1', other_designations='GRSF-1', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70815782, end_position_on_the_genomic_accession=70843274, orientation='minus', exon_count=12, OMIM=604851, no_hdr0=''),
  2928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2928, CurrentID=0, Status='live', Symbol='GSC2', Aliases='GSCL', description='goosecoid homeobox 2', other_designations='GSC-2|GSC-L', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19148991, end_position_on_the_genomic_accession=19150283, orientation='minus', exon_count=3, OMIM=601845, no_hdr0=''),
  2931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2931, CurrentID=0, Status='live', Symbol='GSK3A', Aliases='', description='glycogen synthase kinase 3 alpha', other_designations='GSK-3 alpha|serine/threonine-protein kinase GSK3A', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42230186, end_position_on_the_genomic_accession=42242584, orientation='minus', exon_count=11, OMIM=606784, no_hdr0=''),
  2932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2932, CurrentID=0, Status='live', Symbol='GSK3B', Aliases='', description='glycogen synthase kinase 3 beta', other_designations='GSK-3 beta|GSK3beta isoform|serine/threonine-protein kinase GSK3B', map_location='3q13.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119821321, end_position_on_the_genomic_accession=120094417, orientation='minus', exon_count=12, OMIM=605004, no_hdr0=''),
  2933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2933, CurrentID=0, Status='live', Symbol='GSM1', Aliases='GSP', description='geniospasm 1', other_designations='', map_location='9q13-q21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=190100, no_hdr0=''),
  2934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2934, CurrentID=0, Status='live', Symbol='GSN', Aliases='ADF, AGEL', description='gelsolin', other_designations='actin-depolymerizing factor|brevin', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121210803, end_position_on_the_genomic_accession=121332844, orientation='plus', exon_count=32, OMIM=137350, no_hdr0=''),
  2935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2935, CurrentID=0, Status='live', Symbol='GSPT1', Aliases='551G9.2, ETF3A, GST1, eRF3a', description='G1 to S phase transition 1', other_designations='G1 to S phase transition protein 1 homolog|eukaryotic peptide chain release factor subunit 3a', map_location='16p13.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11868128, end_position_on_the_genomic_accession=11916662, orientation='minus', exon_count=16, OMIM=139259, no_hdr0=''),
  2936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2936, CurrentID=0, Status='live', Symbol='GSR', Aliases='HEL-75, HEL-S-122m', description='glutathione reductase', other_designations='GR|GRase|epididymis luminal protein 75|epididymis secretory sperm binding protein Li 122m', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30678061, end_position_on_the_genomic_accession=30727969, orientation='minus', exon_count=13, OMIM=138300, no_hdr0=''),
  2937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2937, CurrentID=0, Status='live', Symbol='GSS', Aliases='GSHS, HEL-S-64p, HEL-S-88n', description='glutathione synthetase', other_designations='GSH synthetase|GSH-S|epididymis secretory sperm binding protein Li 64p|epididymis secretory sperm binding protein Li 88n|glutathione synthase', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34928433, end_position_on_the_genomic_accession=34955992, orientation='minus', exon_count=15, OMIM=601002, no_hdr0=''),
  2938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2938, CurrentID=0, Status='live', Symbol='GSTA1', Aliases='GST2-1, GTH1, GSTA1', description='glutathione S-transferase alpha 1', other_designations='GST HA subunit 1|GST class-alpha member 1|GST, class alpha, 1|GST-epsilon|S-(hydroxyalkyl)glutathione lyase A1|glutathione S-alkyltransferase A1|glutathione S-aryltransferase A1|glutathione S-transferase 2|glutathione S-transferase Ha subunit 1', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52791380, end_position_on_the_genomic_accession=52803866, orientation='minus', exon_count=7, OMIM=138359, no_hdr0=''),
  2939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2939, CurrentID=0, Status='live', Symbol='GSTA2', Aliases='GST2-2, GTA2, GTH2, GSTA2', description='glutathione S-transferase alpha 2', other_designations='GST HA subunit 2|GST class-alpha member 2|GST-gamma|S-(hydroxyalkyl)glutathione lyase A2|glutathione S-alkyltransferase A2|glutathione S-aralkyltransferase A2|glutathione S-aryltransferase A2|liver GST2', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52750087, end_position_on_the_genomic_accession=52763563, orientation='minus', exon_count=8, OMIM=138360, no_hdr0=''),
  2940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2940, CurrentID=0, Status='live', Symbol='GSTA3', Aliases='GSTA3-3, GTA3', description='glutathione S-transferase alpha 3', other_designations='GST class-alpha member 3|S-(hydroxyalkyl)glutathione lyase A3|glutathione S-alkyltransferase A3|glutathione S-aralkyltransferase A3|glutathione S-aryltransferase A3|glutathione S-transferase A3-3', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52896639, end_position_on_the_genomic_accession=52909798, orientation='minus', exon_count=9, OMIM=605449, no_hdr0=''),
  2941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2941, CurrentID=0, Status='live', Symbol='GSTA4', Aliases='GSTA4-4, GTA4', description='glutathione S-transferase alpha 4', other_designations='GST class-alpha member 4|S-(hydroxyalkyl)glutathione lyase A4|glutathione S-alkyltransferase A4|glutathione S-aralkyltransferase A4|glutathione S-aryltransferase A4|glutathione S-transferase A4-4|glutathione transferase A4-4', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52977948, end_position_on_the_genomic_accession=52995380, orientation='minus', exon_count=7, OMIM=605450, no_hdr0=''),
  2942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2942, CurrentID=0, Status='live', Symbol='GSTA6P', Aliases='GSTAP1', description='glutathione S-transferase alpha 6, pseudogene', other_designations='glutathione S-transferase A pseudogene 1|glutathione S-transferase alpha pseudogene 1', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52805446, end_position_on_the_genomic_accession=52813462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2944, CurrentID=0, Status='live', Symbol='GSTM1', Aliases='GST1-1, GSTM1a-1a, GSTM1b-1b, GTH4, GTM1, H-B, MU, MU-1, GSTM1', description='glutathione S-transferase mu 1', other_designations='GST HB subunit 4|GST class-mu 1|HB subunit 4|S-(hydroxyalkyl)glutathione lyase|glutathione S-alkyltransferase|glutathione S-aralkyltransferase|glutathione S-aryltransferase|glutathione S-transferase M1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109687796, end_position_on_the_genomic_accession=109693745, orientation='plus', exon_count=8, OMIM=138350, no_hdr0=''),
  2946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2946, CurrentID=0, Status='live', Symbol='GSTM2', Aliases='GST4, GSTM-2, GTHMUS, GSTM2', description='glutathione S-transferase mu 2 (muscle)', other_designations='GST class-mu 2|GST, muscle|S-(hydroxyalkyl)glutathione lyase M2|glutathione S-alkyltransferase M2|glutathione S-aralkyltransferase M2|glutathione S-aryltransferase M2|glutathione S-transferase 4|glutathione S-transferase M1|glutathione S-transferase M2 (muscle)', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109668022, end_position_on_the_genomic_accession=109683997, orientation='plus', exon_count=10, OMIM=138380, no_hdr0=''),
  2947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2947, CurrentID=0, Status='live', Symbol='GSTM3', Aliases='GST5, GSTB-3, GTM3, GSTM3', description='glutathione S-transferase mu 3 (brain)', other_designations='GST class-mu 3|S-(hydroxyalkyl)glutathione lyase M3|brain GST|brain type mu-glutathione S-transferase|glutathione S-alkyltransferase M3|glutathione S-aralkyltransferase M3|glutathione S-aryltransferase M3|glutathione S-transferase M3 (brain)|glutathione S-transferase, Mu-3|hGSTM3-3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109733932, end_position_on_the_genomic_accession=109741038, orientation='minus', exon_count=9, OMIM=138390, no_hdr0=''),
  2948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2948, CurrentID=0, Status='live', Symbol='GSTM4', Aliases='GSTM4-4, GTM4', description='glutathione S-transferase mu 4', other_designations='GST class-mu 4|GST-Mu2|GTS-Mu2|S-(hydroxyalkyl)glutathione lyase M4|glutathione S-alkyltransferase M4|glutathione S-aralkyltransferase M4|glutathione S-aryltransferase M4|glutathione S-transferase M4', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109656076, end_position_on_the_genomic_accession=109665501, orientation='plus', exon_count=10, OMIM=138333, no_hdr0=''),
  2949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2949, CurrentID=0, Status='live', Symbol='GSTM5', Aliases='GSTM5-5, GTM5', description='glutathione S-transferase mu 5', other_designations='GST class-mu 5|S-(hydroxyalkyl)glutathione lyase M5|glutathione S-alkyltransferase M5|glutathione S-aralkyltransferase M5|glutathione S-aryltransferase M5|glutathione S-transferase M5', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109711761, end_position_on_the_genomic_accession=109718268, orientation='plus', exon_count=9, OMIM=138385, no_hdr0=''),
  2950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2950, CurrentID=0, Status='live', Symbol='GSTP1', Aliases='DFN7, FAEES3, GST3, GSTP, HEL-S-22, PI', description='glutathione S-transferase pi 1', other_designations='GST class-pi|GSTP1-1|deafness, X-linked 7|epididymis secretory protein Li 22|fatty acid ethyl ester synthase III', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67583595, end_position_on_the_genomic_accession=67586653, orientation='plus', exon_count=7, OMIM=134660, no_hdr0=''),
  2951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2951, CurrentID=0, Status='live', Symbol='GSTP1P1', Aliases='GST3L, GSTP1P, GSTPL, GSTPP', description='glutathione S-transferase pi 1 pseudogene 1', other_designations='glutathione S-transferase pi pseudogene', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55900300, end_position_on_the_genomic_accession=55900711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2952, CurrentID=0, Status='live', Symbol='GSTT1', Aliases='', description='glutathione S-transferase theta 1', other_designations='GST class-theta-1|glutathione transferase T1-1', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession=270308, end_position_on_the_genomic_accession=278486, orientation='minus', exon_count=6, OMIM=600436, no_hdr0=''),
  2953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2953, CurrentID=0, Status='live', Symbol='GSTT2', Aliases='', description='glutathione S-transferase theta 2 (gene/pseudogene)', other_designations='GST class-theta-2', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23980028, end_position_on_the_genomic_accession=23983915, orientation='plus', exon_count=5, OMIM=600437, no_hdr0=''),
  2954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2954, CurrentID=0, Status='live', Symbol='GSTZ1', Aliases='GSTZ1-1, MAAI, MAI', description='glutathione S-transferase zeta 1', other_designations='S-(hydroxyalkyl)glutathione lyase|glutathione S-alkyltransferase|glutathione S-aralkyltransferase|glutathione S-aryltransferase|glutathione transferase zeta 1|maleylacetone isomerase', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77320887, end_position_on_the_genomic_accession=77331597, orientation='plus', exon_count=12, OMIM=603758, no_hdr0=''),
  2956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2956, CurrentID=0, Status='live', Symbol='MSH6', Aliases='GTBP, GTMBP, HNPCC5, HSAP, p160', description='mutS homolog 6', other_designations='G/T mismatch-binding protein|mutS-alpha 160 kDa subunit|sperm-associated protein', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47783082, end_position_on_the_genomic_accession=47806953, orientation='plus', exon_count=13, OMIM=600678, no_hdr0=''),
  2957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2957, CurrentID=0, Status='live', Symbol='GTF2A1', Aliases='TF2A1, TFIIA, TFIIA-42, TFIIAL', description='general transcription factor IIA 1', other_designations='TFIIA alpha, p55|TFIIA alpha/beta subunits|general transcription factor IIA subunit 1|general transcription factor IIA, 1, 19/37kDa|glucose regulated protein, 58kD pseudogene|transcription initiation factor TFIIA 42 kDa subunit', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81175452, end_position_on_the_genomic_accession=81221231, orientation='minus', exon_count=11, OMIM=600520, no_hdr0=''),
  2958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2958, CurrentID=0, Status='live', Symbol='GTF2A2', Aliases='HsT18745, TF2A2, TFIIA', description='general transcription factor IIA 2', other_designations='TFIIA gamma subunit|TFIIA p12 subunit|TFIIA-12|TFIIA-gamma|TFIIAS|general transcription factor IIA subunit 2|general transcription factor IIA, 2, 12kDa|transcription initiation factor IIA gamma chain', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59638062, end_position_on_the_genomic_accession=59657539, orientation='minus', exon_count=7, OMIM=600519, no_hdr0=''),
  2959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2959, CurrentID=0, Status='live', Symbol='GTF2B', Aliases='TF2B, TFIIB', description='general transcription factor IIB', other_designations='RNA polymerase II transcription factor IIB|S300-II|general transcription factor TFIIB', map_location='1p22-p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88852638, end_position_on_the_genomic_accession=88891944, orientation='minus', exon_count=9, OMIM=189963, no_hdr0=''),
  2960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2960, CurrentID=0, Status='live', Symbol='GTF2E1', Aliases='FE, TF2E1, TFIIE-A', description='general transcription factor IIE subunit 1', other_designations='TFIIE alpha subunit|TFIIE-alpha|general transcription factor IIE 56 kDa subunit|general transcription factor IIE, polypeptide 1, alpha 56kDa|transcription initiation factor IIE subunit alpha', map_location='3q21-q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120742711, end_position_on_the_genomic_accession=120783070, orientation='plus', exon_count=7, OMIM=189962, no_hdr0=''),
  2961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2961, CurrentID=0, Status='live', Symbol='GTF2E2', Aliases='FE, TF2E2, TFIIE-B', description='general transcription factor IIE subunit 2', other_designations='TFIIE beta subunit|TFIIE-beta|general transcription factor IIE, polypeptide 2, beta 34kDa', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30578514, end_position_on_the_genomic_accession=30658241, orientation='minus', exon_count=10, OMIM=189964, no_hdr0=''),
  2962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2962, CurrentID=0, Status='live', Symbol='GTF2F1', Aliases='BTF4, RAP74, TF2F1, TFIIF', description='general transcription factor IIF subunit 1', other_designations='TFIIF-alpha|general transcription factor IIF 74 kDa subunit|general transcription factor IIF, polypeptide 1, 74kDa|transcription initiation factor IIF subunit alpha|transcription initiation factor RAP74', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6379569, end_position_on_the_genomic_accession=6393280, orientation='minus', exon_count=13, OMIM=189968, no_hdr0=''),
  2963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2963, CurrentID=0, Status='live', Symbol='GTF2F2', Aliases='BTF4, RAP30, TF2F2, TFIIF', description='general transcription factor IIF subunit 2', other_designations='ATP-dependent helicase GTF2F2|TFIIF-beta|general transcription factor IIF 30 kDa subunit|general transcription factor IIF, polypeptide 2, 30kDa|transcription initiation factor IIF subunit beta|transcription initiation factor RAP30', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45120464, end_position_on_the_genomic_accession=45284889, orientation='plus', exon_count=10, OMIM=189969, no_hdr0=''),
  2964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2964, CurrentID=0, Status='live', Symbol='GTF2F2P1', Aliases='GTF2F2L, TFIIF', description='general transcription factor IIF subunit 2 pseudogene 1', other_designations='general transcription factor IIF, polypeptide 2 pseudogene 1|general transcription factor IIF, polypeptide 2, 30kDa pseudogene', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147506089, end_position_on_the_genomic_accession=147507227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2965, CurrentID=0, Status='live', Symbol='GTF2H1', Aliases='BTF2, P62, TFB1, TFIIH', description='general transcription factor IIH subunit 1', other_designations='BTF2 p62|TFIIH basal transcription factor complex p62 subunit|basic transcription factor 2 62 kDa subunit|general transcription factor IIH polypeptide 1|general transcription factor IIH, polypeptide 1, 62kDa', map_location='11p15.1-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18322269, end_position_on_the_genomic_accession=18367043, orientation='plus', exon_count=17, OMIM=189972, no_hdr0=''),
  2966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2966, CurrentID=0, Status='live', Symbol='GTF2H2', Aliases='BTF2, BTF2P44, T-BTF2P44, TFIIH, p44', description='general transcription factor IIH subunit 2', other_designations='BTF2 p44|TFIIH basal transcription factor complex p44 subunit|basic transcription factor 2 44 kDa subunit|general transcription factor IIH polypeptide 2|general transcription factor IIH, polypeptide 2, 44kD subunit|general transcription factor IIH, polypeptide 2, 44kDa', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71035124, end_position_on_the_genomic_accession=71067670, orientation='minus', exon_count=16, OMIM=601748, no_hdr0=''),
  2967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2967, CurrentID=0, Status='live', Symbol='GTF2H3', Aliases='BTF2, P34, TFB4, TFIIH', description='general transcription factor IIH subunit 3', other_designations='TFIIH basal transcription factor complex p34 subunit|basic transcription factor 2 34 kDa subunit|general transcription factor IIH, polypeptide 3, 34kDa', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123633739, end_position_on_the_genomic_accession=123662604, orientation='plus', exon_count=13, OMIM=601750, no_hdr0=''),
  2968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2968, CurrentID=0, Status='live', Symbol='GTF2H4', Aliases='P52, TFB2, TFIIH', description='general transcription factor IIH subunit 4', other_designations='BTF2 p52|TFIIH basal transcription factor complex p52 subunit|basic transcription factor 2 52 kDa subunit|general transcription factor IIH polypeptide 4|general transcription factor IIH, polypeptide 4, 52kDa', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30908200, end_position_on_the_genomic_accession=30914103, orientation='plus', exon_count=14, OMIM=601760, no_hdr0=''),
  2969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2969, CurrentID=0, Status='live', Symbol='GTF2I', Aliases='BAP135, BTKAP1, DIWS, GTFII-I, IB291, SPIN, TFII-I, WBS, WBSCR6', description='general transcription factor IIi', other_designations='BTK-associated protein, 135kD|Bruton tyrosine kinase-associated protein 135|SRF-Phox1-interacting protein|Williams-Beuren syndrome chromosome region 6', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74657665, end_position_on_the_genomic_accession=74760692, orientation='plus', exon_count=36, OMIM=601679, no_hdr0=''),
  2970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2970, CurrentID=0, Status='live', Symbol='GTF2IP1', Aliases='WBSCR7', description='general transcription factor IIi pseudogene 1', other_designations='Williams-Beuren syndrome chromosome region 7|general transcription factor IIi, pseudogene 1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75185383, end_position_on_the_genomic_accession=75237705, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  2971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2971, CurrentID=0, Status='live', Symbol='GTF3A', Aliases='AP2, TFIIIA', description='general transcription factor IIIA', other_designations='', map_location='13q12.3-q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27424544, end_position_on_the_genomic_accession=27435709, orientation='plus', exon_count=9, OMIM=600860, no_hdr0=''),
  2972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2972, CurrentID=0, Status='live', Symbol='BRF1', Aliases='BRF, BRF-1, CFDS, GTF3B, HEL-S-76p, TAF3B2, TAF3C, TAFIII90, TF3B90, TFIIIB90, hBRF', description='BRF1, RNA polymerase III transcription initiation factor 90 kDa subunit', other_designations='B - related factor 1|BRF1 homolog, subunit of RNA polymerase III transcription initiation factor IIIB|TATA box binding protein (TBP)-associated factor 3C|TATA box binding protein (TBP)-associated factor, RNA polymerase III, GTF3B subunit 2|TBP - associated factor, RNA polymerase III, 90kD|epididymis secretory sperm binding protein Li 76p|general transcription factor IIIB, 90kD subunit', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105209286, end_position_on_the_genomic_accession=105315577, orientation='minus', exon_count=23, OMIM=604902, no_hdr0=''),
  2973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2973, CurrentID=0, Status='live', Symbol='GTS', Aliases='', description='Gilles de la Tourette syndrome', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=137580, no_hdr0=''),
  2974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2974, CurrentID=0, Status='live', Symbol='GUCY1B2', Aliases='GC-SB2', description='guanylate cyclase 1, soluble, beta 2 (pseudogene)', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50994511, end_position_on_the_genomic_accession=51066157, orientation='minus', exon_count=17, OMIM=603695, no_hdr0=''),
  2975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2975, CurrentID=0, Status='live', Symbol='GTF3C1', Aliases='TFIIIC, TFIIIC220, TFIIICalpha', description='general transcription factor IIIC subunit 1', other_designations='TF3C-alpha|TFIIIC 220 kDa subunit|TFIIIC box B-binding subunit|general transcription factor IIIC, polypeptide 1 (alpha subunit)|general transcription factor IIIC, polypeptide 1, alpha 220kDa|transcription factor IIIC 220 kDa subunit|transcription factor IIIC subunit alpha', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27460613, end_position_on_the_genomic_accession=27549930, orientation='minus', exon_count=37, OMIM=603246, no_hdr0=''),
  2976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2976, CurrentID=0, Status='live', Symbol='GTF3C2', Aliases='TFIIIC-BETA, TFIIIC110', description='general transcription factor IIIC subunit 2', other_designations='TF3C-beta|TFIIIC 110 kDa subunit|general transcription factor IIIC, polypeptide 2, beta 110kDa|transcription factor IIIC 110 kDa subunit|transcription factor IIIC subunit beta', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27325849, end_position_on_the_genomic_accession=27357034, orientation='minus', exon_count=21, OMIM=604883, no_hdr0=''),
  2977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2977, CurrentID=0, Status='live', Symbol='GUCY1A2', Aliases='GC-SA2, GUC1A2', description='guanylate cyclase 1, soluble, alpha 2', other_designations='GCS-alpha-2', map_location='11q21-q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106674012, end_position_on_the_genomic_accession=107018445, orientation='minus', exon_count=9, OMIM=601244, no_hdr0=''),
  2978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2978, CurrentID=0, Status='live', Symbol='GUCA1A', Aliases='C6orf131, COD3, CORD14, GCAP, GCAP1, GUCA, GUCA1, dJ139D8.6', description='guanylate cyclase activator 1A', other_designations='GCAP 1|cone dystrophy 3|guanylate cyclase activator 1A (retina)|guanylate cyclase-activating protein, photoreceptor 1|guanylin 1, retina', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42155386, end_position_on_the_genomic_accession=42180056, orientation='plus', exon_count=7, OMIM=600364, no_hdr0=''),
  2979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2979, CurrentID=0, Status='live', Symbol='GUCA1B', Aliases='GCAP2, GUCA2, RP48', description='guanylate cyclase activator 1B', other_designations='GCAP 2|guanylate cyclase activator 1B (retina)|guanylate cyclase-activating protein, photoreceptor 2', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42183284, end_position_on_the_genomic_accession=42206551, orientation='minus', exon_count=5, OMIM=602275, no_hdr0=''),
  2980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2980, CurrentID=0, Status='live', Symbol='GUCA2A', Aliases='GCAP-I, GUCA2, STARA', description='guanylate cyclase activator 2A', other_designations='guanylate cyclase activator 2A (guanylin 2, intestinal, heat-stable)|guanylate cyclase activator 2A (guanylin)|guanylate cyclase-activating protein 1|guanylate cyclase-activating protein I|prepro-guanylin', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42162691, end_position_on_the_genomic_accession=42164724, orientation='minus', exon_count=3, OMIM=139392, no_hdr0=''),
  2981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2981, CurrentID=0, Status='live', Symbol='GUCA2B', Aliases='GCAP-II, UGN', description='guanylate cyclase activator 2B', other_designations='guanylate cyclase activator 2B (uroguanylin)|prepro-uroguanylin|uroguanylin', map_location='1p34-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42153421, end_position_on_the_genomic_accession=42155824, orientation='plus', exon_count=3, OMIM=601271, no_hdr0=''),
  2982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2982, CurrentID=0, Status='live', Symbol='GUCY1A3', Aliases='GC-SA3, GUC1A3, GUCA3, GUCSA3, GUCY1A1, MYMY6', description='guanylate cyclase 1, soluble, alpha 3', other_designations='GC-S-alpha-1|GCS-alpha-3|soluble guanylate cyclase large subunit', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155666568, end_position_on_the_genomic_accession=155737062, orientation='plus', exon_count=15, OMIM=139396, no_hdr0=''),
  2983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2983, CurrentID=0, Status='live', Symbol='GUCY1B3', Aliases='GC-S-beta-1, GC-SB3, GUC1B3, GUCB3, GUCSB3, GUCY1B1', description='guanylate cyclase 1, soluble, beta 3', other_designations='GCS-beta-1|GCS-beta-3|guanylate cyclase soluble subunit beta-3|soluble guanylate cyclase small subunit', map_location='4q31.3-q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155758973, end_position_on_the_genomic_accession=155807642, orientation='plus', exon_count=17, OMIM=139397, no_hdr0=''),
  2984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2984, CurrentID=0, Status='live', Symbol='GUCY2C', Aliases='DIAR6, GUC2C, MECIL, MUCIL, STAR', description='guanylate cyclase 2C', other_designations='GC-C|STA receptor|guanylyl cyclase C|hSTAR|heat stable enterotoxin receptor|intestinal guanylate cyclase', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14612632, end_position_on_the_genomic_accession=14696618, orientation='minus', exon_count=27, OMIM=601330, no_hdr0=''),
  2986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2986, CurrentID=0, Status='live', Symbol='GUCY2F', Aliases='CYGF, GC-F, GUC2DL, GUC2F, RETGC-2, ROS-GC2', description='guanylate cyclase 2F, retinal', other_designations='guanylate cyclase 2D-like, membrane (retina-specific)|guanylate cyclase F|rod outer segment membrane guanylate cyclase 2', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109372017, end_position_on_the_genomic_accession=109482056, orientation='minus', exon_count=22, OMIM=300041, no_hdr0=''),
  2987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2987, CurrentID=0, Status='live', Symbol='GUK1', Aliases='GMK', description='guanylate kinase 1', other_designations='ATP:GMP phosphotransferase|GMP kinase', map_location='1q32-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228139643, end_position_on_the_genomic_accession=228148955, orientation='plus', exon_count=12, OMIM=139270, no_hdr0=''),
  2988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2988, CurrentID=0, Status='live', Symbol='GUK2', Aliases='', description='guanylate kinase 2', other_designations='', map_location='1q32.1-q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=139280, no_hdr0=''),
  2989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2989, CurrentID=0, Status='live', Symbol='GULOP', Aliases='GULO, SCURVY', description='gulonolactone (L-) oxidase, pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27577611, end_position_on_the_genomic_accession=27589073, orientation='plus', exon_count=0, OMIM=240400, no_hdr0=''),
  2990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2990, CurrentID=0, Status='live', Symbol='GUSB', Aliases='BG, MPS7', description='glucuronidase, beta', other_designations='beta-D-glucuronidase|beta-G1', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65960684, end_position_on_the_genomic_accession=65982314, orientation='minus', exon_count=12, OMIM=611499, no_hdr0=''),
  2992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2992, CurrentID=0, Status='live', Symbol='GYG1', Aliases='GSD15, GYG', description='glycogenin 1', other_designations='GN-1|glycogenin glucosyltransferase', map_location='3q24-q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148991408, end_position_on_the_genomic_accession=149027669, orientation='plus', exon_count=8, OMIM=603942, no_hdr0=''),
  2993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2993, CurrentID=0, Status='live', Symbol='GYPA', Aliases='CD235a, GPA, GPErik, GPSAT, HGpMiV, HGpMiXI, HGpSta(C), MN, MNS, PAS-2', description='glycophorin A (MNS blood group)', other_designations='MN sialoglycoprotein|Mi.V glycoprotein (24 AA)|erythroid-lineage-specific membrane sialoglycoprotein|glycophorin A (MN blood group)|glycophorin A, GPA|glycophorin Erik|glycophorin MiI|glycophorin MiV|glycophorin SAT|glycophorin Sta type C|recombinant glycophorin A-B Miltenberger-DR|sialoglycoprotein alpha', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144109303, end_position_on_the_genomic_accession=144140752, orientation='minus', exon_count=7, OMIM=111300, no_hdr0=''),
  2994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2994, CurrentID=0, Status='live', Symbol='GYPB', Aliases='CD235b, GPB, GYP, MNS, PAS-3, SS', description='glycophorin B (MNS blood group)', other_designations='GYPB/GYPA fusion|SS-active sialoglycoprotein|Ss blood group|glycophorin B/glycophorin A fusion protein|glycophorin Hop|hybrid glycophorin Kip|sialoglycoprotein delta', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143995188, end_position_on_the_genomic_accession=144019427, orientation='minus', exon_count=8, OMIM=111740, no_hdr0=''),
  2995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2995, CurrentID=0, Status='live', Symbol='GYPC', Aliases="CD236, CD236R, GE, GPC, GPD, GYPD, PAS-2, PAS-2'", description='glycophorin C (Gerbich blood group)', other_designations='glycoconnectin|glycophorin-D|glycoprotein beta|sialoglycoprotein D', map_location='2q14-q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126655935, end_position_on_the_genomic_accession=126696675, orientation='plus', exon_count=6, OMIM=110750, no_hdr0=''),
  2996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2996, CurrentID=0, Status='live', Symbol='GYPE', Aliases='GPE, MNS, MiIX', description='glycophorin E (MNS blood group)', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143870866, end_position_on_the_genomic_accession=143905563, orientation='minus', exon_count=5, OMIM=138590, no_hdr0=''),
  2997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2997, CurrentID=0, Status='live', Symbol='GYS1', Aliases='GSY, GYS', description='glycogen synthase 1', other_designations='glycogen synthase 1 (muscle)', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48968125, end_position_on_the_genomic_accession=48993353, orientation='minus', exon_count=16, OMIM=138570, no_hdr0=''),
  2998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2998, CurrentID=0, Status='live', Symbol='GYS2', Aliases='', description='glycogen synthase 2', other_designations='glycogen synthase 2 (liver)', map_location='12p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21531527, end_position_on_the_genomic_accession=21630773, orientation='minus', exon_count=20, OMIM=138571, no_hdr0=''),
  2999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2999, CurrentID=0, Status='live', Symbol='GZMH', Aliases='CCP-X, CGL-2, CSP-C, CTLA1, CTSGL2', description='granzyme H', other_designations='cytotoxic T-lymphocyte proteinase|cytotoxic T-lymphocyte-associated serine esterase 1|cytotoxin serine protease-C|granzyme H (cathepsin G-like 2, protein h-CCPX)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24606480, end_position_on_the_genomic_accession=24609774, orientation='minus', exon_count=5, OMIM=116831, no_hdr0=''),
  3000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3000, CurrentID=0, Status='live', Symbol='GUCY2D', Aliases='CORD5, CORD6, CYGD, GUC1A4, GUC2D, LCA, LCA1, RCD2, RETGC-1, ROS-GC1, ROSGC, retGC', description='guanylate cyclase 2D, membrane (retina-specific)', other_designations='ROS-GC|cone rod dystrophy 6|guanylate cyclase 2D, retinal|retinal guanylate cyclase 1|rod outer segment membrane guanylate cyclase', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8002670, end_position_on_the_genomic_accession=8020340, orientation='plus', exon_count=20, OMIM=600179, no_hdr0=''),
  3001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3001, CurrentID=0, Status='live', Symbol='GZMA', Aliases='CTLA3, HFSP', description='granzyme A', other_designations='CTL tryptase|Granzyme A (Cytotoxic T-lymphocyte-associated serine esterase-3; Hanukah factor serine protease)|HF|Hanukah factor serine protease)|cytotoxic T-lymphocyte proteinase 1|fragmentin-1|granzyme A (granzyme 1, cytotoxic T-lymphocyte-associated serine esterase 3)|h factor', map_location='5q11-q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55102646, end_position_on_the_genomic_accession=55110252, orientation='plus', exon_count=5, OMIM=140050, no_hdr0=''),
  3002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3002, CurrentID=0, Status='live', Symbol='GZMB', Aliases='CCPI, CGL-1, CGL1, CSP-B, CSPB, CTLA1, CTSGL1, HLP, SECT', description='granzyme B', other_designations='C11|CTLA-1|T-cell serine protease 1-3E|cathepsin G-like 1|cytotoxic T-lymphocyte proteinase 2|cytotoxic serine protease B|fragmentin 2|fragmentin-2|granzyme B (granzyme 2, cytotoxic T-lymphocyte-associated serine esterase 1)|human lymphocyte protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24630948, end_position_on_the_genomic_accession=24634226, orientation='minus', exon_count=5, OMIM=123910, no_hdr0=''),
  3003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3003, CurrentID=0, Status='live', Symbol='GZMK', Aliases='TRYP2', description='granzyme K', other_designations='NK-Tryp-2|NK-tryptase-2|fragmentin-3|granzyme 3|granzyme K (granzyme 3; tryptase II)|granzyme K (serine protease, granzyme 3; tryptase II)|granzyme-3|tryptase II', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55024279, end_position_on_the_genomic_accession=55034132, orientation='plus', exon_count=5, OMIM=600784, no_hdr0=''),
  3004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3004, CurrentID=0, Status='live', Symbol='GZMM', Aliases='LMET1, MET1', description='granzyme M', other_designations='HU-Met-1|Met-1 serine protease|granzyme M (lymphocyte met-ase 1)|lymphocyte met-ase 1|met-ase|natural killer cell granular protease', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=544034, end_position_on_the_genomic_accession=549920, orientation='plus', exon_count=5, OMIM=600311, no_hdr0=''),
  3005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3005, CurrentID=0, Status='live', Symbol='H1F0', Aliases='H10, H1FV', description='H1 histone family member 0', other_designations="H1 histone family, member 0|H1.0, H1(0), H1-0|histone H1'|histone H1(0)", map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37805107, end_position_on_the_genomic_accession=37807436, orientation='plus', exon_count=1, OMIM=142708, no_hdr0=''),
  3006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3006, CurrentID=0, Status='live', Symbol='HIST1H1C', Aliases='H1.2, H1C, H1F2, H1s-1', description='histone cluster 1, H1c', other_designations='H1 histone family, member 2|histone 1, H1c|histone H1c|histone H1d|histone H1s-1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26055740, end_position_on_the_genomic_accession=26056471, orientation='minus', exon_count=1, OMIM=142710, no_hdr0=''),
  3007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3007, CurrentID=0, Status='live', Symbol='HIST1H1D', Aliases='H1.3, H1D, H1F3, H1s-2', description='histone cluster 1, H1d', other_designations='H1 histone family, member 3|histone 1, H1d|histone H1c|histone H1s-2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26234212, end_position_on_the_genomic_accession=26234988, orientation='minus', exon_count=1, OMIM=142210, no_hdr0=''),
  3008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3008, CurrentID=0, Status='live', Symbol='HIST1H1E', Aliases='H1.4, H1E, H1F4, H1s-4, dJ221C16.5', description='histone cluster 1, H1e', other_designations='H1 histone family, member 4|histone 1, H1e|histone H1b|histone H1s-4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26156331, end_position_on_the_genomic_accession=26157115, orientation='plus', exon_count=1, OMIM=142220, no_hdr0=''),
  3009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3009, CurrentID=0, Status='live', Symbol='HIST1H1B', Aliases='H1, H1.5, H1B, H1F5, H1s-3', description='histone cluster 1, H1b', other_designations='H1 histone family, member 5|histone 1, H1b|histone H1a|histone H1b|histone H1s-3', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27866792, end_position_on_the_genomic_accession=27867581, orientation='minus', exon_count=1, OMIM=142711, no_hdr0=''),
  3010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3010, CurrentID=0, Status='live', Symbol='HIST1H1T', Aliases='H1.6, H1FT, H1t, dJ221C16.2', description='histone cluster 1, H1t', other_designations='H1 histone family, member T (testis-specific)|histone 1, H1t|testicular H1 histone', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26107412, end_position_on_the_genomic_accession=26108136, orientation='minus', exon_count=1, OMIM=142712, no_hdr0=''),
  134083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134083, CurrentID=0, Status='live', Symbol='OR2Y1', Aliases='OR5-2', description='olfactory receptor family 2 subfamily Y member 1', other_designations='olfactory receptor OR5-2|olfactory receptor, family 2, subfamily Y, member 1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180739123, end_position_on_the_genomic_accession=180740058, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3012, CurrentID=0, Status='live', Symbol='HIST1H2AE', Aliases='H2A.1, H2A.2, H2A/a, H2AFA', description='histone cluster 1, H2ae', other_designations='H2A histone family, member A|histone 1, H2ae|histone H2A/a|histone H2AE', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26216920, end_position_on_the_genomic_accession=26217483, orientation='plus', exon_count=1, OMIM=602786, no_hdr0=''),
  3013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3013, CurrentID=0, Status='live', Symbol='HIST1H2AD', Aliases='H2A.3, H2A/g, H2AFG', description='histone cluster 1, H2ad', other_designations='H2A histone family, member G|histone 1, H2ad|histone H2A.3|histone H2A/g|histone H2AD', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26198784, end_position_on_the_genomic_accession=26199293, orientation='minus', exon_count=1, OMIM=602792, no_hdr0=''),
  3014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3014, CurrentID=0, Status='live', Symbol='H2AFX', Aliases='H2A.X, H2A/X, H2AX', description='H2A histone family member X', other_designations='H2A histone family, member X|H2AX histone|histone H2A.x', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119093874, end_position_on_the_genomic_accession=119095467, orientation='minus', exon_count=1, OMIM=601772, no_hdr0=''),
  3015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3015, CurrentID=0, Status='live', Symbol='H2AFZ', Aliases='H2A.Z-1, H2A.z, H2A/z, H2AZ', description='H2A histone family member Z', other_designations='H2A histone family, member Z|H2AZ histone', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99948087, end_position_on_the_genomic_accession=99950355, orientation='minus', exon_count=5, OMIM=142763, no_hdr0=''),
  3017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3017, CurrentID=0, Status='live', Symbol='HIST1H2BD', Aliases='H2B.1B, H2B/b, H2BFB, HIRIP2, dJ221C16.6', description='histone cluster 1, H2bd', other_designations='H2B histone family, member B|HIRA-interacting protein 2|histone 1, H2bd|histone H2B.1 B|histone H2B.b', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26157843, end_position_on_the_genomic_accession=26171349, orientation='plus', exon_count=3, OMIM=602799, no_hdr0=''),
  3018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3018, CurrentID=0, Status='live', Symbol='HIST1H2BB', Aliases='H2B.1, H2B/f, H2BFF', description='histone cluster 1, H2bb', other_designations='H2B histone family, member F|histone 1, H2bb|histone H2B.1|histone H2B.f', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26043227, end_position_on_the_genomic_accession=26043657, orientation='minus', exon_count=1, OMIM=602803, no_hdr0=''),
  3020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3020, CurrentID=0, Status='live', Symbol='H3F3A', Aliases='H3.3A, H3F3', description='H3 histone, family 3A', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226062707, end_position_on_the_genomic_accession=226072002, orientation='plus', exon_count=4, OMIM=601128, no_hdr0=''),
  3021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3021, CurrentID=0, Status='live', Symbol='H3F3B', Aliases='H3.3B', description='H3 histone, family 3B (H3.3B)', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75776434, end_position_on_the_genomic_accession=75779935, orientation='minus', exon_count=4, OMIM=601058, no_hdr0=''),
  3024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3024, CurrentID=0, Status='live', Symbol='HIST1H1A', Aliases='H1.1, H1A, H1F1, HIST1', description='histone cluster 1, H1a', other_designations='H1 histone family, member 1|histone 1, H1a|histone H1a', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26017032, end_position_on_the_genomic_accession=26017812, orientation='minus', exon_count=1, OMIM=142709, no_hdr0=''),
  3026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3026, CurrentID=0, Status='live', Symbol='HABP2', Aliases='FSAP, HABP, HGFAL, PHBP', description='hyaluronan binding protein 2', other_designations='factor VII activating protein|factor VII-activating protease|factor seven-activating protease|hepatocyte growth factor activator-like protein|hyaluronic acid binding protein 2|plasma hyaluronan binding protein|plasma hyaluronan-binding protein', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113550831, end_position_on_the_genomic_accession=113589602, orientation='plus', exon_count=14, OMIM=603924, no_hdr0=''),
  3028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3028, CurrentID=0, Status='live', Symbol='HSD17B10', Aliases='17b-HSD10, ABAD, CAMR, DUPXp11.22, ERAB, HADH2, HCD2, MHBD, MRPP2, MRX17, MRX31, MRXS10, SCHAD, SDR5C1', description='hydroxysteroid (17-beta) dehydrogenase 10', other_designations='3-hydroxy-2-methylbutyryl-CoA dehydrogenase|AB-binding alcohol dehydrogenase|amyloid-beta peptide binding alcohol dehydrogenase|endoplasmic reticulum-associated amyloid beta-peptide-binding protein|hydroxysteroid dehydrogenase 10|mitochondrial RNase P subunit 2|mitochondrial ribonuclease P protein 2|short chain L-3-hydroxyacyl-CoA dehydrogenase type 2|short chain dehydrogenase/reductase family 5C member 1|short chain dehydrogenase/reductase family 5C, member 1|short chain type dehydrogenase/reductase XH98G2', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53431258, end_position_on_the_genomic_accession=53434376, orientation='minus', exon_count=6, OMIM=300256, no_hdr0=''),
  3029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3029, CurrentID=0, Status='live', Symbol='HAGH', Aliases='GLO2, GLX2, GLXII1, HAGH', description='hydroxyacylglutathione hydrolase', other_designations='glx II|glyoxalase II|hydroxyacylglutathione hydroxylase', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1809103, end_position_on_the_genomic_accession=1827194, orientation='minus', exon_count=10, OMIM=138760, no_hdr0=''),
  3030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3030, CurrentID=0, Status='live', Symbol='HADHA', Aliases='ECHA, GBP, HADH, LCEH, LCHAD, MTPA, TP-ALPHA', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), alpha subunit', other_designations='3-ketoacyl-Coenzyme A (CoA) thiolase, alpha subunit|3-oxoacyl-CoA thiolase|78 kDa gastrin-binding protein|gastrin-binding protein|hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), alpha subunit|long-chain 2-enoyl-CoA hydratase|long-chain-3-hydroxyacyl-CoA dehydrogenase|mitochondrial long-chain 2-enoyl-Coenzyme A (CoA) hydratase, alpha subunit|mitochondrial long-chain L-3-hydroxyacyl-Coenzyme A (CoA) dehydrogenase, alpha subunit|mitochondrial trifunctional enzyme, alpha subunit|mitochondrial trifunctional protein, alpha subunit', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26190635, end_position_on_the_genomic_accession=26244726, orientation='minus', exon_count=20, OMIM=600890, no_hdr0=''),
  3031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3031, CurrentID=0, Status='live', Symbol='HADHAP1', Aliases='HADHAP', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), alpha subunit pseudogene 1', other_designations='gastrin binding protein pseudogene|hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase pseudogene (gastrin binding protein pseudogene)|hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), alpha subunit pseudogene|hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase pseudogene (gastrin binding protein pseudogene)', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165403687, end_position_on_the_genomic_accession=165406384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3032, CurrentID=0, Status='live', Symbol='HADHB', Aliases='ECHB, MSTP029, MTPB, TP-BETA', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), beta subunit', other_designations='2-enoyl-Coenzyme A (CoA) hydratase, beta subunit|3-ketoacyl-Coenzyme A (CoA) thiolase of mitochondrial trifunctional protein, beta subunit|acetyl-CoA acyltransferase|beta-ketothiolase|hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), beta subunit', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26244748, end_position_on_the_genomic_accession=26290465, orientation='plus', exon_count=17, OMIM=143450, no_hdr0=''),
  3033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3033, CurrentID=0, Status='live', Symbol='HADH', Aliases='HAD1, HADHSC, HCDH, HHF4, MSCHAD, SCHAD, HADH', description='hydroxyacyl-CoA dehydrogenase', other_designations='L-3-hydroxyacyl-Coenzyme A dehydrogenase, short chain|medium and short-chain L-3-hydroxyacyl-coenzyme A dehydrogenase|short-chain 3-hydroxyacyl-CoA dehydrogenase', map_location='4q22-q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107989714, end_position_on_the_genomic_accession=108035175, orientation='plus', exon_count=10, OMIM=601609, no_hdr0=''),
  3034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3034, CurrentID=0, Status='live', Symbol='HAL', Aliases='HIS, HSTD', description='histidine ammonia-lyase', other_designations='histidase', map_location='12q22-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95972662, end_position_on_the_genomic_accession=95996365, orientation='minus', exon_count=21, OMIM=609457, no_hdr0=''),
  3035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3035, CurrentID=0, Status='live', Symbol='HARS', Aliases='HRS, USH3B', description='histidyl-tRNA synthetase', other_designations='HisRS|histidine translase', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140673904, end_position_on_the_genomic_accession=140691727, orientation='minus', exon_count=13, OMIM=142810, no_hdr0=''),
  3036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3036, CurrentID=0, Status='live', Symbol='HAS1', Aliases='HAS', description='hyaluronan synthase 1', other_designations='HA synthase 1|hyaluronate synthase 1|hyaluronic acid synthase 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51713112, end_position_on_the_genomic_accession=51723992, orientation='minus', exon_count=5, OMIM=601463, no_hdr0=''),
  3037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3037, CurrentID=0, Status='live', Symbol='HAS2', Aliases='', description='hyaluronan synthase 2', other_designations='HA synthase 2|hyaluronate synthase 2|hyaluronic acid synthase 2', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121612116, end_position_on_the_genomic_accession=121641390, orientation='minus', exon_count=7, OMIM=601636, no_hdr0=''),
  3038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3038, CurrentID=0, Status='live', Symbol='HAS3', Aliases='', description='hyaluronan synthase 3', other_designations='HA synthase 3|hyaluronate synthase 3|hyaluronic acid synthase 3', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69105564, end_position_on_the_genomic_accession=69118719, orientation='plus', exon_count=8, OMIM=602428, no_hdr0=''),
  3039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3039, CurrentID=0, Status='live', Symbol='HBA1', Aliases='HBA-T3, HBH', description='hemoglobin subunit alpha 1', other_designations='alpha one globin|alpha-2 globin chain|delta globin|hemoglobin alpha 1 globin chain|hemoglobin, alpha 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=176651, end_position_on_the_genomic_accession=177522, orientation='plus', exon_count=3, OMIM=141800, no_hdr0=''),
  3040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3040, CurrentID=0, Status='live', Symbol='HBA2', Aliases='HBA-T2, HBH', description='hemoglobin subunit alpha 2', other_designations='alpha globin|alpha-2 globin|alpha-globin|hemoglobin alpha chain|hemoglobin, alpha 2|mutant hemoglobin alpha 2 globin chain', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=172847, end_position_on_the_genomic_accession=173710, orientation='plus', exon_count=3, OMIM=141850, no_hdr0=''),
  3041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3041, CurrentID=0, Status='live', Symbol='HBAP1', Aliases='HBA-T1, HBA3, HBA3P', description='hemoglobin subunit alpha pseudogene 1', other_designations='alpha globin pseudogene 1|hemoglobin, alpha pseudogene 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=168641, end_position_on_the_genomic_accession=169452, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3042, CurrentID=0, Status='live', Symbol='HBM', Aliases='HBAP2, HBK', description='hemoglobin subunit mu', other_designations='alpha globin pseudogene 2|hemoglobin mu chain|hemoglobin, alpha pseudogene 2|hemoglobin, mu|mu-globin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=165974, end_position_on_the_genomic_accession=166768, orientation='plus', exon_count=3, OMIM=609639, no_hdr0=''),
  3043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3043, CurrentID=0, Status='live', Symbol='HBB', Aliases='CD113t-C, beta-globin', description='hemoglobin subunit beta', other_designations='beta globin chain|hemoglobin, beta', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5225466, end_position_on_the_genomic_accession=5227071, orientation='minus', exon_count=3, OMIM=141900, no_hdr0=''),
  3044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3044, CurrentID=0, Status='live', Symbol='HBBP1', Aliases='HBH1, HBHP', description='hemoglobin subunit beta pseudogene 1', other_designations='beta globin pseudogene|hemoglobin, beta pseudogene 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5241954, end_position_on_the_genomic_accession=5243592, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3045, CurrentID=0, Status='live', Symbol='HBD', Aliases='', description='hemoglobin subunit delta', other_designations='delta globin|delta-globin chain|hemoglobin delta chain|hemoglobin, delta', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5232829, end_position_on_the_genomic_accession=5234628, orientation='minus', exon_count=3, OMIM=142000, no_hdr0=''),
  3046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3046, CurrentID=0, Status='live', Symbol='HBE1', Aliases='HBE', description='hemoglobin subunit epsilon 1', other_designations='epsilon globin|epsilon-globin|hemoglobin epsilon chain|hemoglobin, epsilon 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5268345, end_position_on_the_genomic_accession=5281945, orientation='minus', exon_count=5, OMIM=142100, no_hdr0=''),
  3047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3047, CurrentID=0, Status='live', Symbol='HBG1', Aliases='HBG-T2, HBGA, HBGR, HSGGL1, PRO2979', description='hemoglobin subunit gamma 1', other_designations='A-gamma globin|gamma A hemoglobin|gamma globin|gamma-1-globin|hb F Agamma|hemoglobin gamma-1 chain|hemoglobin gamma-a chain|hemoglobin, gamma A|hemoglobin, gamma, regulator of', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5248272, end_position_on_the_genomic_accession=5249857, orientation='minus', exon_count=3, OMIM=142200, no_hdr0=''),
  3048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3048, CurrentID=0, Status='live', Symbol='HBG2', Aliases='HBG-T1, TNCY', description='hemoglobin subunit gamma 2', other_designations='G-gamma globin Paulinia|abnormal hemoglobin|gamma-2-globin|hb F Ggamma|hemoglobin gamma-2 chain|hemoglobin gamma-G chain|hemoglobin, gamma G|methemoglobin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5253191, end_position_on_the_genomic_accession=5254781, orientation='minus', exon_count=3, OMIM=142250, no_hdr0=''),
  3049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3049, CurrentID=0, Status='live', Symbol='HBQ1', Aliases='HBQ', description='hemoglobin subunit theta 1', other_designations='hemoglobin theta-1 chain|hemoglobin, theta 1|theta-1-globin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=180334, end_position_on_the_genomic_accession=181179, orientation='plus', exon_count=3, OMIM=142240, no_hdr0=''),
  3050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3050, CurrentID=0, Status='live', Symbol='HBZ', Aliases='HBZ-T11, HBZ', description='hemoglobin subunit zeta', other_designations='HBAZ|hemoglobin zeta chain|hemoglobin, zeta|zeta-globin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=142053, end_position_on_the_genomic_accession=154505, orientation='plus', exon_count=4, OMIM=142310, no_hdr0=''),
  3051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3051, CurrentID=0, Status='live', Symbol='HBZP1', Aliases='HBZ-T2, HBZ2, HBZ2P, HBZP', description='hemoglobin subunit zeta pseudogene 1', other_designations='hemoglobin, zeta pseudogene 1|zeta globin pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=161199, end_position_on_the_genomic_accession=165526, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3052, CurrentID=0, Status='live', Symbol='HCCS', Aliases='CCHL, LSDMCA1, MCOPS7, MLS', description='holocytochrome c synthase', other_designations='cytochrome c heme-lyase|holocytochrome c-type synthase|microphthalamia with linear skin defects', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11111286, end_position_on_the_genomic_accession=11123086, orientation='plus', exon_count=7, OMIM=300056, no_hdr0=''),
  3053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3053, CurrentID=0, Status='live', Symbol='SERPIND1', Aliases='D22S673, HC2, HCF2, HCII, HLS2, LS2, THPH10', description='serpin peptidase inhibitor, clade D (heparin cofactor), member 1', other_designations='leuserpin 2|protease inhibitor leuserpin-2|serine (or cysteine) proteinase inhibitor, clade D (heparin cofactor), member 1|serpin D1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20774095, end_position_on_the_genomic_accession=20787720, orientation='plus', exon_count=5, OMIM=142360, no_hdr0=''),
  3054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3054, CurrentID=0, Status='live', Symbol='HCFC1', Aliases='CFF, HCF, HCF-1, HCF1, HFC1, MRX3, PPP1R89, VCAF', description='host cell factor C1', other_designations='VP16-accessory protein|protein phosphatase 1, regulatory subunit 89', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153947557, end_position_on_the_genomic_accession=153971833, orientation='minus', exon_count=26, OMIM=300019, no_hdr0=''),
  3055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3055, CurrentID=0, Status='live', Symbol='HCK', Aliases='JTK9, p59Hck, p61Hck', description='HCK proto-oncogene, Src family tyrosine kinase', other_designations='hematopoietic cell kinase|hemopoietic cell kinase|p59-HCK/p60-HCK', map_location='20q11-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32052188, end_position_on_the_genomic_accession=32101856, orientation='plus', exon_count=15, OMIM=142370, no_hdr0=''),
  3056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3056, CurrentID=0, Status='live', Symbol='HCL1', Aliases='BRHC', description='hair color 1 (brown)', other_designations='Brown hair color', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3057, CurrentID=0, Status='live', Symbol='HCL2', Aliases='RHA, RHC', description='hair color 2 (red)', other_designations='', map_location='4q28-q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=266300, no_hdr0=''),
  3059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3059, CurrentID=0, Status='live', Symbol='HCLS1', Aliases='CTTNL, HS1, lckBP1, p75', description='hematopoietic cell-specific Lyn substrate 1', other_designations='', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121631398, end_position_on_the_genomic_accession=121660948, orientation='minus', exon_count=14, OMIM=601306, no_hdr0=''),
  3060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3060, CurrentID=0, Status='live', Symbol='HCRT', Aliases='NRCLP1, OX, PPOX', description='hypocretin (orexin) neuropeptide precursor', other_designations='prepro-orexin', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42184060, end_position_on_the_genomic_accession=42185452, orientation='minus', exon_count=2, OMIM=602358, no_hdr0=''),
  3061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3061, CurrentID=0, Status='live', Symbol='HCRTR1', Aliases='OX1R', description='hypocretin (orexin) receptor 1', other_designations='hypocretin receptor 1|hypocretin receptor type 1|hypocretin receptor-1|orexin receptor 1|orexin receptor-1|ox-1-R|ox1-R', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31617689, end_position_on_the_genomic_accession=31632519, orientation='plus', exon_count=11, OMIM=602392, no_hdr0=''),
  3062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3062, CurrentID=0, Status='live', Symbol='HCRTR2', Aliases='OX2R', description='hypocretin (orexin) receptor 2', other_designations='hypocretin receptor type 2|orexin type-2 receptor|ox-2-R|ox2-R', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55174273, end_position_on_the_genomic_accession=55282620, orientation='plus', exon_count=8, OMIM=602393, no_hdr0=''),
  3063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3063, CurrentID=0, Status='live', Symbol='HCVS', Aliases='', description='human coronavirus sensitivity', other_designations='', map_location='15q11-qter', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=122460, no_hdr0=''),
  3064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3064, CurrentID=0, Status='live', Symbol='HTT', Aliases='HD, IT15', description='huntingtin', other_designations='huntington disease protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3074510, end_position_on_the_genomic_accession=3243960, orientation='plus', exon_count=67, OMIM=613004, no_hdr0=''),
  3065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3065, CurrentID=0, Status='live', Symbol='HDAC1', Aliases='GON-10, HD1, RPD3, RPD3L1', description='histone deacetylase 1', other_designations='reduced potassium dependency, yeast homolog-like 1', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32292107, end_position_on_the_genomic_accession=32333628, orientation='plus', exon_count=15, OMIM=601241, no_hdr0=''),
  3066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3066, CurrentID=0, Status='live', Symbol='HDAC2', Aliases='HD2, RPD3, YAF1', description='histone deacetylase 2', other_designations='YY1-associated factor 1|transcriptional regulator homolog RPD3', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113936156, end_position_on_the_genomic_accession=113971195, orientation='minus', exon_count=17, OMIM=605164, no_hdr0=''),
  3067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3067, CurrentID=0, Status='live', Symbol='HDC', Aliases='', description='histidine decarboxylase', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50241945, end_position_on_the_genomic_accession=50266091, orientation='minus', exon_count=14, OMIM=142704, no_hdr0=''),
  3068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3068, CurrentID=0, Status='live', Symbol='HDGF', Aliases='HMG1L2', description='hepatoma-derived growth factor', other_designations='HMG-1L2|high mobility group protein 1-like 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156742107, end_position_on_the_genomic_accession=156752448, orientation='minus', exon_count=11, OMIM=600339, no_hdr0=''),
  3069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3069, CurrentID=0, Status='live', Symbol='HDLBP', Aliases='HBP, PRO2900, VGL', description='high density lipoprotein binding protein', other_designations='HDL-binding protein|high density lipoprotein-binding protein', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241227267, end_position_on_the_genomic_accession=241315700, orientation='minus', exon_count=36, OMIM=142695, no_hdr0=''),
  3070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3070, CurrentID=0, Status='live', Symbol='HELLS', Aliases='LSH, Nbla10143, PASG, SMARCA6', description='helicase, lymphoid-specific', other_designations='SWI/SNF2-related matrix-associated actin-dependent regulator of chromatin subfamily A member 6|SWI/SNF2-related, matrix-associated, actin-dependent regulator of chromatin, subfamily A, member 6|proliferation-associated SNF2-like protein', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94545767, end_position_on_the_genomic_accession=94602099, orientation='plus', exon_count=25, OMIM=603946, no_hdr0=''),
  3071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3071, CurrentID=0, Status='live', Symbol='NCKAP1L', Aliases='HEM1', description='NCK-associated protein 1-like', other_designations='hematopoietic protein 1|membrane-associated protein hem-1', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54497711, end_position_on_the_genomic_accession=54543115, orientation='plus', exon_count=32, OMIM=141180, no_hdr0=''),
  3073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3073, CurrentID=0, Status='live', Symbol='HEXA', Aliases='TSD', description='hexosaminidase A (alpha polypeptide)', other_designations='N-acetyl-beta-glucosaminidase subunit alpha|beta-N-acetylhexosaminidase subunit alpha|hexosaminidase subunit A', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72343437, end_position_on_the_genomic_accession=72376179, orientation='minus', exon_count=14, OMIM=606869, no_hdr0=''),
  3074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3074, CurrentID=0, Status='live', Symbol='HEXB', Aliases='ENC-1AS, HEL-248, HEL-S-111', description='hexosaminidase B (beta polypeptide)', other_designations='HCC-7|N-acetyl-beta-glucosaminidase subunit beta|beta-N-acetylhexosaminidase subunit beta|cervical cancer proto-oncogene 7 protein|epididymis luminal protein 248|epididymis secretory protein Li 111|hexosaminidase subunit B', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74640023, end_position_on_the_genomic_accession=74721288, orientation='plus', exon_count=15, OMIM=606873, no_hdr0=''),
  3075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3075, CurrentID=0, Status='live', Symbol='CFH', Aliases='AHUS1, AMBP1, ARMD4, ARMS1L3, FH, FHL1, HF, HF1, HF2, HUS, CFH', description='complement factor H', other_designations='H factor 1 (complement)|H factor 2 (complement)|adrenomedullin binding protein|age-related maculopathy susceptibility 1|beta-1-H-globulin|beta-1H|factor H|factor H-like 1', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196651878, end_position_on_the_genomic_accession=196747504, orientation='plus', exon_count=23, OMIM=134370, no_hdr0=''),
  3077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3077, CurrentID=0, Status='live', Symbol='HFE', Aliases='HFE1, HH, HLA-H, MVCD7, TFQTL2', description='hemochromatosis', other_designations='MHC class I-like protein HFE|hereditary hemochromatosis protein HLA-H|high Fe', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26087281, end_position_on_the_genomic_accession=26096117, orientation='plus', exon_count=7, OMIM=613609, no_hdr0=''),
  3078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3078, CurrentID=0, Status='live', Symbol='CFHR1', Aliases='CFHL, CFHL1, CFHL1PP, FHR1, H36-1, H36-2, HFL1, HFL2, CFHR1', description='complement factor H-related 1', other_designations='FHR-1|H factor (complement)-like 1|H factor (complement)-like 2|H-factor-like 1|H36|complement factor H-related 1 pseudogene|h factor-like protein 1', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196819731, end_position_on_the_genomic_accession=196832189, orientation='plus', exon_count=6, OMIM=134371, no_hdr0=''),
  100532737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532737, CurrentID=0, Status='live', Symbol='ATP6V1G2-DDX39B', Aliases='', description='ATP6V1G2-DDX39B readthrough (NMD candidate)', other_designations='ATP6V1G2-DDX39B readthrough (non-protein coding)', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31530219, end_position_on_the_genomic_accession=31546848, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  3080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3080, CurrentID=0, Status='live', Symbol='CFHR2', Aliases='CFHL2, FHR2, HFL3', description='complement factor H-related 2', other_designations='DDESK59|FHR-2|H factor (complement)-like 3|factor H-related gene 2|h factor-like 3|h factor-like protein 2', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196943758, end_position_on_the_genomic_accession=196959226, orientation='plus', exon_count=5, OMIM=600889, no_hdr0=''),
  3081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3081, CurrentID=0, Status='live', Symbol='HGD', Aliases='AKU, HGO', description='homogentisate 1,2-dioxygenase', other_designations='homogentisate oxidase|homogentisate oxygenase|homogentisic acid oxidase|homogentisicase', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120628168, end_position_on_the_genomic_accession=120682571, orientation='minus', exon_count=16, OMIM=607474, no_hdr0=''),
  3082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3082, CurrentID=0, Status='live', Symbol='HGF', Aliases='DFNB39, F-TCFB, HPTA, SF, HGF', description='hepatocyte growth factor (hepapoietin A; scatter factor)', other_designations='fibroblast-derived tumor cytotoxic factor|hepatopoietin-A|lung fibroblast-derived mitogen', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81699006, end_position_on_the_genomic_accession=81770438, orientation='minus', exon_count=20, OMIM=142409, no_hdr0=''),
  3083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3083, CurrentID=0, Status='live', Symbol='HGFAC', Aliases='HGFA', description='HGF activator', other_designations='', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3441933, end_position_on_the_genomic_accession=3449495, orientation='plus', exon_count=15, OMIM=604552, no_hdr0=''),
  3084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3084, CurrentID=0, Status='live', Symbol='NRG1', Aliases='ARIA, GGF, GGF2, HGL, HRG, HRG1, HRGA, MST131, MSTP131, NDF-IT2, SMDF, NRG1', description='neuregulin 1', other_designations='glial growth factor|heregulin, alpha (45kD, ERBB2 p185-activator)|neu differentiation factor|pro-NRG1|sensory and motor neuron derived factor', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31639304, end_position_on_the_genomic_accession=32767957, orientation='plus', exon_count=19, OMIM=142445, no_hdr0=''),
  100532738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532738, CurrentID=0, Status='live', Symbol='DEL1P32P31', Aliases='C1DELp32p31', description='Chromosome 1p32-p31 deletion syndrome', other_designations='', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613735, no_hdr0=''),
  3087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3087, CurrentID=0, Status='live', Symbol='HHEX', Aliases='HEX, HMPH, HOX11L-PEN, PRH, PRHX', description='hematopoietically expressed homeobox', other_designations='homeobox protein HEX|homeobox protein PRH|homeobox, hematopoietically expressed|proline-rich homeodomain-containing transcription factor', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92689924, end_position_on_the_genomic_accession=92695651, orientation='plus', exon_count=4, OMIM=604420, no_hdr0=''),
  3090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3090, CurrentID=0, Status='live', Symbol='HIC1', Aliases='ZBTB29, ZNF901, hic-1', description='hypermethylated in cancer 1', other_designations='zinc finger and BTB domain-containing protein 29', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2055099, end_position_on_the_genomic_accession=2059687, orientation='plus', exon_count=3, OMIM=603825, no_hdr0=''),
  3091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3091, CurrentID=0, Status='live', Symbol='HIF1A', Aliases='HIF-1-alpha, HIF-1A, HIF-1alpha, HIF1, HIF1-ALPHA, MOP1, PASD8, bHLHe78', description='hypoxia inducible factor 1, alpha subunit (basic helix-loop-helix transcription factor)', other_designations='ARNT interacting protein|ARNT-interacting protein|PAS domain-containing protein 8|basic-helix-loop-helix-PAS protein MOP1|class E basic helix-loop-helix protein 78|hypoxia-inducible factor 1 alpha isoform I.3|hypoxia-inducible factor 1, alpha subunit (basic helix-loop-helix transcription factor)|hypoxia-inducible factor1alpha|member of PAS protein 1|member of PAS superfamily 1', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61695401, end_position_on_the_genomic_accession=61748259, orientation='plus', exon_count=16, OMIM=603348, no_hdr0=''),
  3092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3092, CurrentID=0, Status='live', Symbol='HIP1', Aliases='HIP-I, ILWEQ, SHON, SHONbeta, SHONgamma', description='huntingtin interacting protein 1', other_designations='huntingtin-interacting protein I', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75533298, end_position_on_the_genomic_accession=75738972, orientation='minus', exon_count=33, OMIM=601767, no_hdr0=''),
  3093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3093, CurrentID=0, Status='live', Symbol='UBE2K', Aliases='E2-25K, HIP2, HYPG, LIG, UBC1', description='ubiquitin conjugating enzyme E2K', other_designations='E2 ubiquitin-conjugating enzyme K|E2(25K)|HIP-2|huntingtin interacting protein 2|huntingtin-interacting protein 2|ubiquitin carrier protein|ubiquitin-conjugating enzyme E2(25K)|ubiquitin-conjugating enzyme E2-25 KDA|ubiquitin-conjugating enzyme E2-25K|ubiquitin-conjugating enzyme E2K (UBC1 homolog, yeast)|ubiquitin-protein ligase', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39698044, end_position_on_the_genomic_accession=39782792, orientation='plus', exon_count=7, OMIM=602846, no_hdr0=''),
  3094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3094, CurrentID=0, Status='live', Symbol='HINT1', Aliases='HINT, NMAN, PKCI-1, PRKCNH1', description='histidine triad nucleotide binding protein 1', other_designations="adenosine 5'-monophosphoramidase|protein kinase C inhibitor 1|protein kinase C-interacting protein 1", map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131159283, end_position_on_the_genomic_accession=131165348, orientation='minus', exon_count=5, OMIM=601314, no_hdr0=''),
  3096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3096, CurrentID=0, Status='live', Symbol='HIVEP1', Aliases='CIRIP, CRYBP1, GAAP, MBP-1, PRDII-BF1, Schnurri-1, ZAS1, ZNF40, ZNF40A', description='human immunodeficiency virus type I enhancer binding protein 1', other_designations='cirhin interaction protein|gate keeper of apoptosis-activating protein|major histocompatibility complex binding protein 1|positive regulatory domain II binding factor 1', map_location='6p24-p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12007653, end_position_on_the_genomic_accession=12164999, orientation='plus', exon_count=17, OMIM=194540, no_hdr0=''),
  3097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3097, CurrentID=0, Status='live', Symbol='HIVEP2', Aliases='HIV-EP2, MBP-2, MIBP1, SHN2, ZAS2, ZNF40B', description='human immunodeficiency virus type I enhancer binding protein 2', other_designations='MHC binding protein-2|MHC-binding protein 2|Schnurri-2|c-myc intron binding protein 1|human immunodeficiency virus type I enhancer-binding protein 2', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142751467, end_position_on_the_genomic_accession=142945201, orientation='minus', exon_count=10, OMIM=143054, no_hdr0=''),
  3098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3098, CurrentID=0, Status='live', Symbol='HK1', Aliases='HK1-ta-tb, HK1-tc, HKD, HKI, HMSNR, HXK1, HK1', description='hexokinase 1', other_designations='HK I|brain form hexokinase|glycolytic enzyme|hexokinase type I', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69269984, end_position_on_the_genomic_accession=69401882, orientation='plus', exon_count=28, OMIM=142600, no_hdr0=''),
  3099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3099, CurrentID=0, Status='live', Symbol='HK2', Aliases='HKII, HXK2', description='hexokinase 2', other_designations='HK II|hexokinase type II|hexokinase-2, muscle|muscle form hexokinase', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74832655, end_position_on_the_genomic_accession=74893354, orientation='plus', exon_count=18, OMIM=601125, no_hdr0=''),
  3101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3101, CurrentID=0, Status='live', Symbol='HK3', Aliases='HKIII, HXK3', description='hexokinase 3 (white cell)', other_designations='ATP:D-hexose 6-phosphotransferase|HK III|hexokinase type III', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176880869, end_position_on_the_genomic_accession=176899373, orientation='minus', exon_count=21, OMIM=142570, no_hdr0=''),
  3104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3104, CurrentID=0, Status='live', Symbol='ZBTB48', Aliases='HKR3, ZNF855, pp9964', description='zinc finger and BTB domain containing 48', other_designations='GLI-Kruppel family member HKR3|krueppel-related zinc finger protein 3|zinc finger protein 855', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6579991, end_position_on_the_genomic_accession=6589280, orientation='plus', exon_count=11, OMIM=165270, no_hdr0=''),
  3105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3105, CurrentID=0, Status='live', Symbol='HLA-A', Aliases='HLAA', description='major histocompatibility complex, class I, A', other_designations='MHC class I antigen HLA-A heavy chain|antigen presenting molecule|leukocyte antigen class I-A', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29942470, end_position_on_the_genomic_accession=29945884, orientation='plus', exon_count=8, OMIM=142800, no_hdr0=''),
  3106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3106, CurrentID=0, Status='live', Symbol='HLA-B', Aliases='AS, HLAB, SPDA1', description='major histocompatibility complex, class I, B', other_designations='HLA class I histocompatibility antigen, B alpha chain|MHC Class I HLA heavy chain|MHC HLA-B cell surface glycoprotein|MHC HLA-B transmembrane glycoprotein|MHC class I antigen GN00104|MHC class I antigen HLA-B heavy chain|MHC class I antigen SHCHA|leukocyte antigen class I-B|lymphocyte antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31353868, end_position_on_the_genomic_accession=31357212, orientation='minus', exon_count=8, OMIM=142830, no_hdr0=''),
  3107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3107, CurrentID=0, Status='live', Symbol='HLA-C', Aliases='D6S204, HLA-JY3, HLC-C, PSORS1', description='major histocompatibility complex, class I, C', other_designations='HLA class I histocompatibility antigen, C alpha chain|MHC class I antigen heavy chain HLA-C|human leukocyte antigen-C alpha chain|major histocompatibility antigen HLA-C', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31268749, end_position_on_the_genomic_accession=31272136, orientation='minus', exon_count=8, OMIM=142840, no_hdr0=''),
  3108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3108, CurrentID=0, Status='live', Symbol='HLA-DMA', Aliases='D6S222E, DMA, HLADM, RING6', description='major histocompatibility complex, class II, DM alpha', other_designations='MHC class II antigen DMA|class II histocompatibility antigen, M alpha chain|really interesting new gene 6 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32948614, end_position_on_the_genomic_accession=32953122, orientation='minus', exon_count=5, OMIM=142855, no_hdr0=''),
  3109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3109, CurrentID=0, Status='live', Symbol='HLA-DMB', Aliases='D6S221E, RING7', description='major histocompatibility complex, class II, DM beta', other_designations='MHC class II HLA-DMB|MHC class II antigen DMB|MHC class II antigen HLA-DM beta chain|class II histocompatibility antigen, M beta chain|really interesting new gene 7 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32934629, end_position_on_the_genomic_accession=32941070, orientation='minus', exon_count=6, OMIM=142856, no_hdr0=''),
  3110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3110, CurrentID=0, Status='live', Symbol='MNX1', Aliases='HB9, HLXB9, HOXHB9, SCRA1', description='motor neuron and pancreas homeobox 1', other_designations='homeobox HB9|homeobox protein HB9', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157004853, end_position_on_the_genomic_accession=157010653, orientation='minus', exon_count=4, OMIM=142994, no_hdr0=''),
  3111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3111, CurrentID=0, Status='live', Symbol='HLA-DOA', Aliases='HLA-DNA, HLA-DZA, HLADZ', description='major histocompatibility complex, class II, DO alpha', other_designations='HLA-D0-alpha|MHC DN-alpha|MHC DZ alpha|MHC class II antigen DOA|lymphocyte antigen|major histocompatibility complex, class II, DN alpha', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33004182, end_position_on_the_genomic_accession=33009612, orientation='minus', exon_count=5, OMIM=142930, no_hdr0=''),
  3112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3112, CurrentID=0, Status='live', Symbol='HLA-DOB', Aliases='DOB', description='major histocompatibility complex, class II, DO beta', other_designations='MHC class II antigen DOB', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32812763, end_position_on_the_genomic_accession=32817048, orientation='minus', exon_count=6, OMIM=600629, no_hdr0=''),
  3113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3113, CurrentID=0, Status='live', Symbol='HLA-DPA1', Aliases='DP(W3), DP(W4), HLA-DP1A, HLADP, HLASB, PLT1', description='major histocompatibility complex, class II, DP alpha 1', other_designations='HLA-SB alpha chain|MHC class II DP3-alpha|MHC class II HLA-DPA1 antigen|MHC class II antigen|Primed lymphocyte test-1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33064569, end_position_on_the_genomic_accession=33080778, orientation='minus', exon_count=6, OMIM=142880, no_hdr0=''),
  3115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3115, CurrentID=0, Status='live', Symbol='HLA-DPB1', Aliases='DPB1, HLA-DP, HLA-DP1B, HLA-DPB', description='major histocompatibility complex, class II, DP beta 1', other_designations='HLA DP14-beta chain|HLA class II histocompatibility antigen, DP(W4) beta chain|HLA-DP histocompatibility type, beta-1 subunit|MHC HLA DPB1|MHC class II HLA-DP-beta-1|MHC class II HLA-DRB1|MHC class II antigen DP beta 1 chain|MHC class II antigen DPB1|MHC class II antigen DPbeta1|MHC class II antigen beta chain|beta1 domain MHC class II HLA DPB|class II HLA beta chain|major histocompatibility complex class II antigen beta chain', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33075926, end_position_on_the_genomic_accession=33089696, orientation='plus', exon_count=7, OMIM=142858, no_hdr0=''),
  3116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3116, CurrentID=0, Status='live', Symbol='HLA-DPB2', Aliases='DP2B, DPB2, DPbeta2, HLA-DP2B', description='major histocompatibility complex, class II, DP beta 2 (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33112516, end_position_on_the_genomic_accession=33129113, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3117, CurrentID=0, Status='live', Symbol='HLA-DQA1', Aliases='CD, CELIAC1, DQ-A1, GSE, HLA-DQA', description='major histocompatibility complex, class II, DQ alpha 1', other_designations='DC-1 alpha chain|DC-alpha|HLA class II histocompatibility antigen, DQ(W3) alpha chain|HLA-DCA|MHC HLA-DQ alpha|MHC class II DQA1|MHC class II HLA-D alpha glycoprotein|MHC class II HLA-DQ-alpha-1|MHC class II antigen|MHC class II surface glycoprotein|leucocyte antigen DQA1|leukocyte antigen alpha chain', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32637396, end_position_on_the_genomic_accession=32654774, orientation='plus', exon_count=6, OMIM=146880, no_hdr0=''),
  3118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3118, CurrentID=0, Status='live', Symbol='HLA-DQA2', Aliases='DX-ALPHA, HLA-DXA', description='major histocompatibility complex, class II, DQ alpha 2', other_designations='DX alpha chain|HLA class II histocompatibility antigen, DQ(6) alpha chain|MHC class II DQA2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32741386, end_position_on_the_genomic_accession=32746887, orientation='plus', exon_count=5, OMIM=613503, no_hdr0=''),
  3119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3119, CurrentID=0, Status='live', Symbol='HLA-DQB1', Aliases='CELIAC1, HLA-DQB, IDDM1', description='major histocompatibility complex, class II, DQ beta 1', other_designations='MHC DQ beta|MHC class II DQ beta chain|MHC class II HLA-DQ beta glycoprotein|MHC class II antigen DQB1|MHC class II antigen HLA-DQ-beta-1|MHC class2 antigen|lymphocyte antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32659464, end_position_on_the_genomic_accession=32666689, orientation='minus', exon_count=6, OMIM=604305, no_hdr0=''),
  3120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3120, CurrentID=0, Status='live', Symbol='HLA-DQB2', Aliases='HLA-DQB1, HLA-DXB', description='major histocompatibility complex, class II, DQ beta 2', other_designations='DV19.1 (major histocompatibility complex, class II, DQ beta 2 (HLA-DXB))|HLA class II histocompatibility antigen, DX beta chain|MHC class II antigen DQB2|major histocompatibility complex, class II, DQ beta 1', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32756094, end_position_on_the_genomic_accession=32763553, orientation='minus', exon_count=6, OMIM=615161, no_hdr0=''),
  3121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3121, CurrentID=0, Status='live', Symbol='HLA-DQB3', Aliases='D6S205, D6S205E, DQB3, DVB, DVbeta, HLA-DVB', description='major histocompatibility complex, class II, DQ beta 3', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3122, CurrentID=0, Status='live', Symbol='HLA-DRA', Aliases='HLA-DRA1, MLRW', description='major histocompatibility complex, class II, DR alpha', other_designations='MHC cell surface glycoprotein|MHC class II antigen DRA|histocompatibility antigen HLA-DR alpha', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32439842, end_position_on_the_genomic_accession=32445046, orientation='plus', exon_count=5, OMIM=142860, no_hdr0=''),
  3123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3123, CurrentID=0, Status='live', Symbol='HLA-DRB1', Aliases='DRB1, DRw10, HLA-DR1B, HLA-DRB, SS1', description='major histocompatibility complex, class II, DR beta 1', other_designations='DW2.2/DR2.2|HLA class II histocompatibility antigen, DR-1 beta chain|MHC class II HLA-DR beta 1 chain|MHC class II HLA-DR-beta cell surface glycoprotein|MHC class II HLA-DRw10-beta|MHC class II antigen|human leucocyte antigen DRB1|lymphocyte antigen DRB1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32578769, end_position_on_the_genomic_accession=32589836, orientation='minus', exon_count=6, OMIM=142857, no_hdr0=''),
  3124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3124, CurrentID=0, Status='live', Symbol='HLA-DRB2', Aliases='HLA-DR2B', description='major histocompatibility complex, class II, DR beta 2 (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3125, CurrentID=0, Status='live', Symbol='HLA-DRB3', Aliases='DRB1, DRw10, HLA-DR1B, HLA-DR3B, HLA-DRB, SS1', description='major histocompatibility complex, class II, DR beta 3', other_designations='DR7|HLA class II histocompatibility antigen, DR beta 3 chain|HLA class II histocompatibility antigen, DRB1-7 beta chain|MHC Class II antigen|MHC class II HLA-DR beta 3 chain|MHC class II antigen DR beta 3 chain|MHC class II antigen DRB3|human leucocyte antigen DRB3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612735, no_hdr0=''),
  3126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3126, CurrentID=0, Status='live', Symbol='HLA-DRB4', Aliases='DR-4, DR4, DRB4, HLA-DR4B', description='major histocompatibility complex, class II, DR beta 4', other_designations='DRB1 transplantation antigen|HLA class II histocompatibility antigen, DR beta 4 chain|MHC HLA DR-beta chain|MHC class II antigen DRB4|MHC class II antigen HLA-DR-beta|MHC class2 antigen|human leucocyte antigen DRB4|leukocyte antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3127, CurrentID=0, Status='live', Symbol='HLA-DRB5', Aliases='HLA-DRB', description='major histocompatibility complex, class II, DR beta 5', other_designations='DR beta-5|DR-9|DR2-beta-2|DR9|HLA class II histocompatibility antigen, DR beta 5 chain|HLA class II histocompatibility antigen, DR-5 beta chain|HLA class II histocompatibility antigen, DRB1-9 beta chain|MHC HLA-DR-beta cell surface glycoprotein|MHC HLA-DR-beta chain|MHC class II HLA beta chain|MHC class II antigen DRB1*9|MHC class II antigen DRB5|dw2|human leucocyte antigen DRB5|leukocyte antigen class II', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32517374, end_position_on_the_genomic_accession=32530316, orientation='minus', exon_count=7, OMIM=604776, no_hdr0=''),
  3128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3128, CurrentID=0, Status='live', Symbol='HLA-DRB6', Aliases='', description='major histocompatibility complex, class II, DR beta 6 (pseudogene)', other_designations='MHC class II antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32552713, end_position_on_the_genomic_accession=32560002, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3129, CurrentID=0, Status='live', Symbol='HLA-DRB7', Aliases='', description='major histocompatibility complex, class II, DR beta 7 (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3130, CurrentID=0, Status='live', Symbol='HLA-DRB8', Aliases='', description='major histocompatibility complex, class II, DR beta 8 (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3131, CurrentID=0, Status='live', Symbol='HLF', Aliases='', description='hepatic leukemia factor', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55264960, end_position_on_the_genomic_accession=55325065, orientation='plus', exon_count=5, OMIM=142385, no_hdr0=''),
  3132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3132, CurrentID=0, Status='live', Symbol='HLA-DRB9', Aliases='D6S206, D6S206E, HLA-DR1BL, HLA-DRB1L', description='major histocompatibility complex, class II, DR beta 9 (pseudogene)', other_designations='MHC class II antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32459820, end_position_on_the_genomic_accession=32460089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3133, CurrentID=0, Status='live', Symbol='HLA-E', Aliases='EA1.2, EA2.1, HLA-6.2, MHC, QA1', description='major histocompatibility complex, class I, E', other_designations='HLA class I histocompatibility antigen, E alpha chain|MHC HLA-E alpha-1|MHC HLA-E alpha-2.1|MHC class I antigen E|lymphocyte antigen', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30489406, end_position_on_the_genomic_accession=30494205, orientation='plus', exon_count=8, OMIM=143010, no_hdr0=''),
  3134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3134, CurrentID=0, Status='live', Symbol='HLA-F', Aliases='CDA12, HLA-5.4, HLA-CDA12, HLAF', description='major histocompatibility complex, class I, F', other_designations='HLA F antigen|HLA class I molecule|MHC class I antigen F|MHC class Ib antigen|leukocyte antigen F', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29723340, end_position_on_the_genomic_accession=29740355, orientation='plus', exon_count=9, OMIM=143110, no_hdr0=''),
  3135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3135, CurrentID=0, Status='live', Symbol='HLA-G', Aliases='MHC-G', description='major histocompatibility complex, class I, G', other_designations='HLA G antigen|HLA class I molecule|HLA-G histocompatibility antigen, class I, G|MHC class I antigen G|b2 microglobulin', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29826979, end_position_on_the_genomic_accession=29831122, orientation='plus', exon_count=8, OMIM=142871, no_hdr0=''),
  3136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3136, CurrentID=0, Status='live', Symbol='HLA-H', Aliases='HLAHP', description='major histocompatibility complex, class I, H (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29887760, end_position_on_the_genomic_accession=29891080, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3137, CurrentID=0, Status='live', Symbol='HLA-J', Aliases='CDA12, D6S203, HLA-59, HLA-CDA12', description='major histocompatibility complex, class I, J (pseudogene)', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30005971, end_position_on_the_genomic_accession=30009956, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3138, CurrentID=0, Status='live', Symbol='HLA-K', Aliases='HLA-70, HLA70, HLAK', description='major histocompatibility complex, class I, K (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29926659, end_position_on_the_genomic_accession=29929839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3139, CurrentID=0, Status='live', Symbol='HLA-L', Aliases='HLA-92, HLA92, HLAL', description='major histocompatibility complex, class I, L (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30259562, end_position_on_the_genomic_accession=30266951, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3140, CurrentID=0, Status='live', Symbol='MR1', Aliases='HLALS', description='major histocompatibility complex, class I-related', other_designations='MHC class I-like antigen MR-1|MHC class-I related-gene protein|class I histocompatibility antigen-like protein|major histocompatibility complex, class I-like sequence', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181033379, end_position_on_the_genomic_accession=181061938, orientation='plus', exon_count=8, OMIM=600764, no_hdr0=''),
  3141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3141, CurrentID=0, Status='live', Symbol='HLCS', Aliases='HCS', description='holocarboxylase synthetase (biotin-(proprionyl-CoA-carboxylase (ATP-hydrolysing)) ligase)', other_designations='biotin apo-protein ligase|biotin--[acetyl-CoA-carboxylase] ligase|biotin--[methylcrotonoyl-CoA-carboxylase] ligase|biotin--[methylmalonyl-CoA-carboxytransferase] ligase|holocarboxylase synthetase (biotin-(proprionyl-Coenzyme A-carboxylase (ATP-hydrolysing)) ligase)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36750888, end_position_on_the_genomic_accession=36990246, orientation='minus', exon_count=20, OMIM=609018, no_hdr0=''),
  3142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3142, CurrentID=0, Status='live', Symbol='HLX', Aliases='HB241, HLX', description='H2.0-like homeobox', other_designations='H2.0-like homeo box 1|H2.0-like homeo box-1|H2.0-like homeobox 1|homeobox protein HB24|homeobox protein HLX1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220879401, end_position_on_the_genomic_accession=220885059, orientation='plus', exon_count=4, OMIM=142995, no_hdr0=''),
  401554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401554, CurrentID=0, Status='live', Symbol='LOC401554', Aliases='', description='uncharacterized LOC401554', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130140315, end_position_on_the_genomic_accession=130144225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3145, CurrentID=0, Status='live', Symbol='HMBS', Aliases='PBG-D, PBGD, PORC, UPS', description='hydroxymethylbilane synthase', other_designations='porphyria, acute; Chester type|pre-uroporphyrinogen synthase|uroporphyrinogen I synthase|uroporphyrinogen I synthetase', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119084876, end_position_on_the_genomic_accession=119093549, orientation='plus', exon_count=15, OMIM=609806, no_hdr0=''),
  3146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3146, CurrentID=0, Status='live', Symbol='HMGB1', Aliases='HMG-1, HMG1, HMG3, SBP-1', description='high mobility group box 1', other_designations='Amphoterin|Sulfoglucuronyl carbohydrate binding protein|high-mobility group (nonhistone chromosomal) protein 1|high-mobility group box 1', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30458740, end_position_on_the_genomic_accession=30465944, orientation='minus', exon_count=8, OMIM=163905, no_hdr0=''),
  3148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3148, CurrentID=0, Status='live', Symbol='HMGB2', Aliases='HMG2', description='high mobility group box 2', other_designations='HMG-2|high mobility group protein 2|high-mobility group (nonhistone chromosomal) protein 2|high-mobility group box 2', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173331376, end_position_on_the_genomic_accession=173334444, orientation='minus', exon_count=5, OMIM=163906, no_hdr0=''),
  3149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3149, CurrentID=0, Status='live', Symbol='HMGB3', Aliases='HMG-2a, HMG-4, HMG2A, HMG4', description='high mobility group box 3', other_designations='high mobility group protein 2a|high-mobility group (nonhistone chromosomal) protein 4', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150980508, end_position_on_the_genomic_accession=150990775, orientation='plus', exon_count=7, OMIM=300193, no_hdr0=''),
  3150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3150, CurrentID=0, Status='live', Symbol='HMGN1', Aliases='HMG14', description='high mobility group nucleosome binding domain 1', other_designations='high mobility group nucleosome-binding domain-containing protein 1|high-mobility group (nonhistone chromosomal) protein 14|high-mobility group nucleosome binding 1|high-mobility group nucleosome binding domain 1|nonhistone chromosomal protein HMG-14', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39342315, end_position_on_the_genomic_accession=39349121, orientation='minus', exon_count=10, OMIM=163920, no_hdr0=''),
  3151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3151, CurrentID=0, Status='live', Symbol='HMGN2', Aliases='HMG17', description='high mobility group nucleosomal binding domain 2', other_designations='high mobility group nucleosome-binding domain-containing protein 2|high mobility group protein N2|high-mobility group (nonhistone chromosomal) protein 17|high-mobility group nucleosomal binding domain 2|nonhistone chromosomal protein HMG-17', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26472411, end_position_on_the_genomic_accession=26476642, orientation='plus', exon_count=6, OMIM=163910, no_hdr0=''),
  3152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3152, CurrentID=0, Status='live', Symbol='HMGN2P11', Aliases='28H, HMG17P1, HMGN2L11', description='high mobility group nucleosomal binding domain 2 pseudogene 11', other_designations='high-mobility group (nonhistone chromosomal) protein 17 pseudogene 1|high-mobility group nucleosomal binding domain 2 pseudogene 11|high-mobility group nucleosomal binding domain 2-like 11', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84876484, end_position_on_the_genomic_accession=84877824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3153, CurrentID=0, Status='live', Symbol='HMGN2P12', Aliases='60H, HMG17P2, HMGN2L12', description='high mobility group nucleosomal binding domain 2 pseudogene 12', other_designations='high-mobility group (nonhistone chromosomal) protein 17 pseudogene 2', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106930962, end_position_on_the_genomic_accession=106932128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3155, CurrentID=0, Status='live', Symbol='HMGCL', Aliases='HL', description='3-hydroxymethyl-3-methylglutaryl-CoA lyase', other_designations='3-hydroxy-3-methylglutarate-CoA lyase|3-hydroxy-3-methylglutaryl-CoA lyase|3-hydroxymethyl-3-methylglutaryl-Coenzyme A lyase|HMG-CoA lyase|hydroxymethylglutaricaciduria|mitochondrial 3-hydroxy-3-methylglutaryl-CoA lyase', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23801877, end_position_on_the_genomic_accession=23825459, orientation='minus', exon_count=9, OMIM=613898, no_hdr0=''),
  3156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3156, CurrentID=0, Status='live', Symbol='HMGCR', Aliases='LDLCQ3', description='3-hydroxy-3-methylglutaryl-CoA reductase', other_designations='3-hydroxy-3-methylglutaryl CoA reductase (NADPH)|HMG-CoA reductase|hydroxymethylglutaryl-CoA reductase', map_location='5q13.3-q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75336334, end_position_on_the_genomic_accession=75362101, orientation='plus', exon_count=22, OMIM=142910, no_hdr0=''),
  3157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3157, CurrentID=0, Status='live', Symbol='HMGCS1', Aliases='HMGCS', description='3-hydroxy-3-methylglutaryl-CoA synthase 1', other_designations='3-hydroxy-3-methylglutaryl coenzyme A (HMG-CoA) synthase|3-hydroxy-3-methylglutaryl-CoA synthase 1 (soluble)|3-hydroxy-3-methylglutaryl-Coenzyme A synthase 1 (soluble)', map_location='5p14-p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43287470, end_position_on_the_genomic_accession=43313512, orientation='minus', exon_count=12, OMIM=142940, no_hdr0=''),
  3158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3158, CurrentID=0, Status='live', Symbol='HMGCS2', Aliases='', description='3-hydroxy-3-methylglutaryl-CoA synthase 2 (mitochondrial)', other_designations='3-hydroxy-3-methylglutaryl-Coenzyme A synthase 2 (mitochondrial)|HMG-CoA synthase', map_location='1p13-p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119747996, end_position_on_the_genomic_accession=119769125, orientation='minus', exon_count=10, OMIM=600234, no_hdr0=''),
  3159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3159, CurrentID=0, Status='live', Symbol='HMGA1', Aliases='HMG-RA, HMGIY, HMGA1', description='high mobility group AT-hook 1', other_designations='HMG-I(Y)|high mobility group protein A1|high mobility group protein R|high-mobility group (nonhistone chromosomal) protein isoforms I and Y|nonhistone chromosomal high-mobility group protein HMG-I/HMG-Y', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34236800, end_position_on_the_genomic_accession=34246231, orientation='plus', exon_count=8, OMIM=600701, no_hdr0=''),
  3161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3161, CurrentID=0, Status='live', Symbol='HMMR', Aliases='CD168, IHABP, RHAMM', description='hyaluronan-mediated motility receptor (RHAMM)', other_designations='intracellular hyaluronic acid-binding protein|receptor for hyaluronan-mediated motility', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163460511, end_position_on_the_genomic_accession=163491945, orientation='plus', exon_count=18, OMIM=600936, no_hdr0=''),
  3162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3162, CurrentID=0, Status='live', Symbol='HMOX1', Aliases='HMOX1D, HO-1, HSP32, bK286B10', description='heme oxygenase 1', other_designations='heat shock protein, 32-kD|heme oxygenase (decycling) 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35381067, end_position_on_the_genomic_accession=35394214, orientation='plus', exon_count=5, OMIM=141250, no_hdr0=''),
  3163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3163, CurrentID=0, Status='live', Symbol='HMOX2', Aliases='HO-2', description='heme oxygenase 2', other_designations='heme oxygenase (decycling) 2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4474703, end_position_on_the_genomic_accession=4510347, orientation='plus', exon_count=12, OMIM=141251, no_hdr0=''),
  3164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3164, CurrentID=0, Status='live', Symbol='NR4A1', Aliases='GFRP1, HMR, N10, NAK-1, NGFIB, NP10, NUR77, TR3', description='nuclear receptor subfamily 4 group A member 1', other_designations='ST-59|TR3 orphan receptor|early response protein NAK1|growth factor-inducible nuclear protein N10|hormone receptor|nerve growth factor IB nuclear receptor variant 1|nuclear hormone receptor NUR/77|nuclear receptor subfamily 4, group A, member 1|orphan nuclear receptor HMR|orphan nuclear receptor TR3|steroid receptor TR3|testicular receptor 3', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52022832, end_position_on_the_genomic_accession=52059507, orientation='plus', exon_count=13, OMIM=139139, no_hdr0=''),
  3166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3166, CurrentID=0, Status='live', Symbol='HMX1', Aliases='H6, NKX5-3', description='H6 family homeobox 1', other_designations='H6 homeodomain protein|homeobox protein H6', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8846077, end_position_on_the_genomic_accession=8871817, orientation='minus', exon_count=6, OMIM=142992, no_hdr0=''),
  3167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3167, CurrentID=0, Status='live', Symbol='HMX2', Aliases='H6L, Nkx5-2', description='H6 family homeobox 2', other_designations='homeo box (H6 family) 2|homeobox protein H6 family member 2', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123142421, end_position_on_the_genomic_accession=123150672, orientation='plus', exon_count=3, OMIM=600647, no_hdr0=''),
  3169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3169, CurrentID=0, Status='live', Symbol='FOXA1', Aliases='HNF3A, TCF3A', description='forkhead box A1', other_designations='HNF-3-alpha|HNF-3A|TCF-3A|forkhead box protein A1|transcription factor 3A', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=37589552, end_position_on_the_genomic_accession=37595120, orientation='minus', exon_count=2, OMIM=602294, no_hdr0=''),
  3170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3170, CurrentID=0, Status='live', Symbol='FOXA2', Aliases='HNF3B, TCF3B', description='forkhead box A2', other_designations='HNF-3-beta|HNF-3B|TCF-3B|forkhead box protein A2|hepatic nuclear factor-3-beta|hepatocyte nuclear factor 3, beta|transcription factor 3B', map_location='20p11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22581004, end_position_on_the_genomic_accession=22585463, orientation='minus', exon_count=5, OMIM=600288, no_hdr0=''),
  3171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3171, CurrentID=0, Status='live', Symbol='FOXA3', Aliases='FKHH3, HNF3G, TCF3G', description='forkhead box A3', other_designations='HNF-3-gamma|HNF-3G|TCF-3G|fork head-related protein FKH H3|forkhead box protein A3|hepatocyte nuclear factor 3, gamma|transcription factor 3G', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45864260, end_position_on_the_genomic_accession=45873797, orientation='plus', exon_count=2, OMIM=602295, no_hdr0=''),
  3172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3172, CurrentID=0, Status='live', Symbol='HNF4A', Aliases='FRTS4, HNF4, HNF4a7, HNF4a8, HNF4a9, HNF4alpha, MODY, MODY1, NR2A1, NR2A21, TCF, TCF14', description='hepatocyte nuclear factor 4, alpha', other_designations='HNF4alpha10/11/12|TCF-14|hepatic nuclear factor 4 alpha|nuclear receptor subfamily 2 group A member 1|transcription factor 14|transcription factor HNF-4', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44355801, end_position_on_the_genomic_accession=44432845, orientation='plus', exon_count=13, OMIM=600281, no_hdr0=''),
  3174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3174, CurrentID=0, Status='live', Symbol='HNF4G', Aliases='NR2A2, NR2A3', description='hepatocyte nuclear factor 4, gamma', other_designations='HNF-4-gamma|nuclear receptor subfamily 2 group A member 2', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=75407681, end_position_on_the_genomic_accession=75566843, orientation='plus', exon_count=17, OMIM=605966, no_hdr0=''),
  3175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3175, CurrentID=0, Status='live', Symbol='ONECUT1', Aliases='HNF-6, HNF6, HNF6A', description='one cut homeobox 1', other_designations='hepatocyte nuclear factor 6, alpha|one cut domain family member 1', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52756963, end_position_on_the_genomic_accession=52790277, orientation='minus', exon_count=7, OMIM=604164, no_hdr0=''),
  3176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3176, CurrentID=0, Status='live', Symbol='HNMT', Aliases='HMT-S1, HNMT-S2, HNMT', description='histamine N-methyltransferase', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=137964238, end_position_on_the_genomic_accession=138016364, orientation='plus', exon_count=8, OMIM=605238, no_hdr0=''),
  3177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3177, CurrentID=0, Status='live', Symbol='SLC29A2', Aliases='DER12, ENT2, HNP36', description='solute carrier family 29 (equilibrative nucleoside transporter), member 2', other_designations='36 kDa nucleolar protein HNP36|delayed-early response protein 12|equilibrative NBMPR-insensitive nucleoside transporter|equilibrative nitrobenzylmercaptopurine riboside-insensitive nucleoside transporter|hydrophobic nucleolar protein, 36 kDa|hydrophobic nucleolar protein, 36kD|nucleoside transporter, ei-type|solute carrier family 29 (nucleoside transporters), member 2|solute carrier family 29 member 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66362521, end_position_on_the_genomic_accession=66372476, orientation='minus', exon_count=14, OMIM=602110, no_hdr0=''),
  3178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3178, CurrentID=0, Status='live', Symbol='HNRNPA1', Aliases='ALS19, ALS20, HNRPA1, HNRPA1L3, IBMPFD3, hnRNP A1, hnRNP-A1', description='heterogeneous nuclear ribonucleoprotein A1', other_designations='Putative heterogeneous nuclear ribonucleoprotein A1-like 3|helix-destabilizing protein|heterogeneous nuclear ribonucleoprotein A1B protein|heterogeneous nuclear ribonucleoprotein B2 protein|heterogeneous nuclear ribonucleoprotein core protein A1|hnRNP A1-like 3|hnRNP core protein A1|hnRNP core protein A1-like 3|nuclear ribonucleoprotein particle A1 protein|single-strand DNA-binding protein UP1|single-strand RNA-binding protein', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54280696, end_position_on_the_genomic_accession=54287087, orientation='plus', exon_count=12, OMIM=164017, no_hdr0=''),
  3181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3181, CurrentID=0, Status='live', Symbol='HNRNPA2B1', Aliases='HNRNPA2, HNRNPB1, HNRPA2, HNRPA2B1, HNRPB1, IBMPFD2, RNPA2, SNRPB1', description='heterogeneous nuclear ribonucleoprotein A2/B1', other_designations='hnRNP A2 / hnRNP B1|nuclear ribonucleoprotein particle A2 protein', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26189927, end_position_on_the_genomic_accession=26200793, orientation='minus', exon_count=15, OMIM=600124, no_hdr0=''),
  3182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3182, CurrentID=0, Status='live', Symbol='HNRNPAB', Aliases='ABBP1, HNRPAB', description='heterogeneous nuclear ribonucleoprotein A/B', other_designations='ABBP-1|APOBEC1-binding protein 1|apobec-1 binding protein 1|apolipoprotein B mRNA editing enzyme, catalytic polypeptide 1-binding protein 1|hnRNP A/B|hnRNP type A/B protein', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178204507, end_position_on_the_genomic_accession=178211183, orientation='plus', exon_count=8, OMIM=602688, no_hdr0=''),
  3183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3183, CurrentID=0, Status='live', Symbol='HNRNPC', Aliases='C1, C2, HNRNP, HNRPC, SNRPC', description='heterogeneous nuclear ribonucleoprotein C (C1/C2)', other_designations='hnRNP C1/C2|nuclear ribonucleoprotein particle C1 protein|nuclear ribonucleoprotein particle C2 protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21209136, end_position_on_the_genomic_accession=21269479, orientation='minus', exon_count=10, OMIM=164020, no_hdr0=''),
  3184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3184, CurrentID=0, Status='live', Symbol='HNRNPD', Aliases='AUF1, AUF1A, HNRPD, P37, hnRNPD0', description='heterogeneous nuclear ribonucleoprotein D', other_designations='ARE-binding protein AUFI, type A|heterogeneous nuclear ribonucleoprotein D (AU-rich element RNA binding protein 1, 37kDa)|hnRNP D0', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82353314, end_position_on_the_genomic_accession=82373996, orientation='minus', exon_count=9, OMIM=601324, no_hdr0=''),
  3185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3185, CurrentID=0, Status='live', Symbol='HNRNPF', Aliases='HNRPF, OK/SW-cl.23, mcs94-1', description='heterogeneous nuclear ribonucleoprotein F', other_designations='HnRNP F protein|nucleolin-like protein mcs94-1', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43385617, end_position_on_the_genomic_accession=43409248, orientation='minus', exon_count=7, OMIM=601037, no_hdr0=''),
  3186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3186, CurrentID=0, Status='live', Symbol='RBMXP1', Aliases='HNRNP-G, HNRPG', description='RNA binding motif protein, X-linked pseudogene 1', other_designations='Heterogeneous nuclear ribonucleoprotein polypeptide G|RNA binding motif protein, X chromosome, pseudogene 1|heterogeneous nuclear ribonucleoprotein G', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=48212903, end_position_on_the_genomic_accession=48214969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3187, CurrentID=0, Status='live', Symbol='HNRNPH1', Aliases='HNRPH, HNRPH1, hnRNPH', description='heterogeneous nuclear ribonucleoprotein H1 (H)', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179614178, end_position_on_the_genomic_accession=179634784, orientation='minus', exon_count=17, OMIM=601035, no_hdr0=''),
  3188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3188, CurrentID=0, Status='live', Symbol='HNRNPH2', Aliases="FTP3, HNRPH', HNRPH2, hnRNPH'", description="heterogeneous nuclear ribonucleoprotein H2 (H')", other_designations="FTP-3|heterogeneous nuclear ribonucleoprotein H-prime|hnRNP H'|hnRNP H2", map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101408133, end_position_on_the_genomic_accession=101414140, orientation='plus', exon_count=2, OMIM=300610, no_hdr0=''),
  3189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3189, CurrentID=0, Status='live', Symbol='HNRNPH3', Aliases='2H9, HNRPH3', description='heterogeneous nuclear ribonucleoprotein H3 (2H9)', other_designations='heterogeneous nuclear ribonucleoprotein 2H9|hnRNP 2H9|hnRNP H3', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68332011, end_position_on_the_genomic_accession=68343196, orientation='plus', exon_count=11, OMIM=602324, no_hdr0=''),
  3190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3190, CurrentID=0, Status='live', Symbol='HNRNPK', Aliases='AUKS, CSBP, HNRPK, TUNP', description='heterogeneous nuclear ribonucleoprotein K', other_designations='dC-stretch binding protein|transformation upregulated nuclear protein', map_location='9q21.32-q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83968083, end_position_on_the_genomic_accession=83980654, orientation='minus', exon_count=18, OMIM=600712, no_hdr0=''),
  3191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3191, CurrentID=0, Status='live', Symbol='HNRNPL', Aliases='HNRPL, P/OKcl.14, hnRNP-L', description='heterogeneous nuclear ribonucleoprotein L', other_designations='hnRNP L', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38836378, end_position_on_the_genomic_accession=38852339, orientation='minus', exon_count=17, OMIM=603083, no_hdr0=''),
  3192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3192, CurrentID=0, Status='live', Symbol='HNRNPU', Aliases='HNRNPU-AS1, HNRPU, SAF-A, SAFA, U21.1, hnRNP U', description='heterogeneous nuclear ribonucleoprotein U (scaffold attachment factor A)', other_designations='p120 nuclear protein|pp120', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244840638, end_position_on_the_genomic_accession=244846941, orientation='minus', exon_count=16, OMIM=602869, no_hdr0=''),
  134265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134265, CurrentID=0, Status='live', Symbol='AFAP1L1', Aliases='', description='actin filament associated protein 1 like 1', other_designations='AFAP1-like protein 1|actin filament associated protein 1-like 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149271838, end_position_on_the_genomic_accession=149343637, orientation='plus', exon_count=20, OMIM=614410, no_hdr0=''),
  134266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134266, CurrentID=0, Status='live', Symbol='GRPEL2', Aliases='Mt-GrpE#2', description='GrpE-like 2, mitochondrial (E. coli)', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149345414, end_position_on_the_genomic_accession=149354583, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3195, CurrentID=0, Status='live', Symbol='TLX1', Aliases='HOX11, TCL3', description='T-cell leukemia homeobox 1', other_designations='T-cell leukemia/lymphoma protein 3|homeo box-11 (T-cell leukemia-3 associated breakpoint, homologous to Drosophila Notch)|homeobox protein Hox-11|proto-oncogene TCL-3', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101130780, end_position_on_the_genomic_accession=101137789, orientation='plus', exon_count=4, OMIM=186770, no_hdr0=''),
  3196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3196, CurrentID=0, Status='live', Symbol='TLX2', Aliases='HOX11L1, NCX', description='T-cell leukemia homeobox 2', other_designations='homeo box 11-like 1|homeobox protein Hox-11L1|neural crest homeobox protein', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74514469, end_position_on_the_genomic_accession=74517148, orientation='plus', exon_count=3, OMIM=604240, no_hdr0=''),
  3197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3197, CurrentID=0, Status='live', Symbol='HOXA@', Aliases='HOX1@', description='homeobox A cluster', other_designations='homeo box A cluster', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3198, CurrentID=0, Status='live', Symbol='HOXA1', Aliases='BSAS, HOX1, HOX1F', description='homeobox A1', other_designations='HOX A1 homeodomain protein|Hox 1.6-like protein|homeo box A1|homeobox 1F|homeobox protein Hox-1F|lab-like protein', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27092993, end_position_on_the_genomic_accession=27096006, orientation='minus', exon_count=3, OMIM=142955, no_hdr0=''),
  3199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3199, CurrentID=0, Status='live', Symbol='HOXA2', Aliases='HOX1K, MCOHI', description='homeobox A2', other_designations='homeobox protein Hox-1K', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27100354, end_position_on_the_genomic_accession=27102775, orientation='minus', exon_count=2, OMIM=604685, no_hdr0=''),
  3200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3200, CurrentID=0, Status='live', Symbol='HOXA3', Aliases='HOX1, HOX1E', description='homeobox A3', other_designations='Hox-1.5-like protein|homeo box 1E|homeo box A3|homeobox protein Hox-1E', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27106190, end_position_on_the_genomic_accession=27152581, orientation='minus', exon_count=10, OMIM=142954, no_hdr0=''),
  3201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3201, CurrentID=0, Status='live', Symbol='HOXA4', Aliases='HOX1, HOX1D', description='homeobox A4', other_designations='Dfd-like protein|Hox-1.4-like protein|homeo box A4|homeobox protein Hox-1.4|homeobox protein Hox-1D', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27128507, end_position_on_the_genomic_accession=27130780, orientation='minus', exon_count=2, OMIM=142953, no_hdr0=''),
  3202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3202, CurrentID=0, Status='live', Symbol='HOXA5', Aliases='HOX1, HOX1.3, HOX1C', description='homeobox A5', other_designations='homeo box 1C|homeo box A5|homeobox protein HOXA5|homeobox protein Hox-1C', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27141052, end_position_on_the_genomic_accession=27143668, orientation='minus', exon_count=2, OMIM=142952, no_hdr0=''),
  3203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3203, CurrentID=0, Status='live', Symbol='HOXA6', Aliases='HOX1, HOX1.2, HOX1B', description='homeobox A6', other_designations='homeo box 1B|homeo box A6|homeobox protein HOXA6|homeobox protein Hox-1B', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27145583, end_position_on_the_genomic_accession=27147774, orientation='minus', exon_count=2, OMIM=142951, no_hdr0=''),
  3204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3204, CurrentID=0, Status='live', Symbol='HOXA7', Aliases='ANTP, HOX1, HOX1.1, HOX1A', description='homeobox A7', other_designations='homeo box A7|homeobox protein HOX-1A|homeobox protein Hox 1.1', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27153716, end_position_on_the_genomic_accession=27156677, orientation='minus', exon_count=2, OMIM=142950, no_hdr0=''),
  3205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3205, CurrentID=0, Status='live', Symbol='HOXA9', Aliases='ABD-B, HOX1, HOX1.7, HOX1G', description='homeobox A9', other_designations='homeobox protein Hox-1G|homeodomain protein HOXA9', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27162438, end_position_on_the_genomic_accession=27165530, orientation='minus', exon_count=2, OMIM=142956, no_hdr0=''),
  3206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3206, CurrentID=0, Status='live', Symbol='HOXA10', Aliases='HOX1, HOX1.8, HOX1H, PL', description='homeobox A10', other_designations='homeo box A10|homeobox protein 1H|homeobox protein HOXA10|homeobox protein Hox-1.8|homeobox protein Hox-1H', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27170591, end_position_on_the_genomic_accession=27180261, orientation='minus', exon_count=3, OMIM=142957, no_hdr0=''),
  3207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3207, CurrentID=0, Status='live', Symbol='HOXA11', Aliases='HOX1, HOX1I', description='homeobox A11', other_designations='homeo box 1I|homeobox protein HOXA11|homeobox protein Hox-1I', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27181157, end_position_on_the_genomic_accession=27185216, orientation='minus', exon_count=2, OMIM=142958, no_hdr0=''),
  3208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3208, CurrentID=0, Status='live', Symbol='HPCA', Aliases='BDR2, DYT2', description='hippocalcin', other_designations='calcium-binding protein BDR-2', map_location='1p35-p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32886087, end_position_on_the_genomic_accession=32894646, orientation='plus', exon_count=5, OMIM=142622, no_hdr0=''),
  3209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3209, CurrentID=0, Status='live', Symbol='HOXA13', Aliases='HOX1, HOX1J', description='homeobox A13', other_designations='homeo box 1J|homeo box A13|homeobox protein HOXA13|homeobox protein Hox-1J|transcription factor HOXA13', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27193419, end_position_on_the_genomic_accession=27200266, orientation='minus', exon_count=3, OMIM=142959, no_hdr0=''),
  3210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3210, CurrentID=0, Status='live', Symbol='HOXB@', Aliases='HOX2@', description='homeobox B cluster', other_designations='homeo box B cluster', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3211, CurrentID=0, Status='live', Symbol='HOXB1', Aliases='HCFP3, HOX2, HOX2I, Hox-2.9', description='homeobox B1', other_designations='homeobox protein Hox-2I', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48529445, end_position_on_the_genomic_accession=48530910, orientation='minus', exon_count=2, OMIM=142968, no_hdr0=''),
  3212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3212, CurrentID=0, Status='live', Symbol='HOXB2', Aliases='HOX2, HOX2H, Hox-2.8, K8', description='homeobox B2', other_designations='K8 home protein|homeo box 2H|homeo box B2|homeobox protein Hox-2.8|homeobox protein Hox-2H', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48542655, end_position_on_the_genomic_accession=48545031, orientation='minus', exon_count=4, OMIM=142967, no_hdr0=''),
  3213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3213, CurrentID=0, Status='live', Symbol='HOXB3', Aliases='HOX2, HOX2G, Hox-2.7', description='homeobox B3', other_designations='homeo box 2G|homeo box B3|homeobox protein Hox-2.7|homeobox protein Hox-2G', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48548870, end_position_on_the_genomic_accession=48590282, orientation='minus', exon_count=15, OMIM=142966, no_hdr0=''),
  3214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3214, CurrentID=0, Status='live', Symbol='HOXB4', Aliases='HOX-2.6, HOX2, HOX2F', description='homeobox B4', other_designations='homeo box 2F|homeo box B4|homeobox protein Hox-2.6|homeobox protein Hox-2F', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48575507, end_position_on_the_genomic_accession=48578381, orientation='minus', exon_count=2, OMIM=142965, no_hdr0=''),
  3215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3215, CurrentID=0, Status='live', Symbol='HOXB5', Aliases='HHO.C10, HOX2, HOX2A, HU-1, Hox2.1', description='homeobox B5', other_designations='homeo box 2A|homeo box B5|homeobox protein HHO.C10|homeobox protein Hox-2A|homeobox protein Hu-1', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48591257, end_position_on_the_genomic_accession=48593741, orientation='minus', exon_count=2, OMIM=142960, no_hdr0=''),
  134288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134288, CurrentID=0, Status='live', Symbol='TMEM174', Aliases='', description='transmembrane protein 174', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73173196, end_position_on_the_genomic_accession=73175143, orientation='plus', exon_count=2, OMIM=614909, no_hdr0=''),
  3217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3217, CurrentID=0, Status='live', Symbol='HOXB7', Aliases='HHO.C1, HOX2, HOX2C, Hox-2.3', description='homeobox B7', other_designations='homeo box 2C|homeo box B7|homeo box c1 protein|homeobox protein HHO.C1|homeobox protein Hox-2C', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48607232, end_position_on_the_genomic_accession=48611021, orientation='minus', exon_count=2, OMIM=142962, no_hdr0=''),
  3218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3218, CurrentID=0, Status='live', Symbol='HOXB8', Aliases='HOX2, HOX2D, Hox-2.4', description='homeobox B8', other_designations='homeo box 2D|homeo box B8|homeobox protein Hox-2.4|homeobox protein Hox-2D', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48612346, end_position_on_the_genomic_accession=48615112, orientation='minus', exon_count=2, OMIM=142963, no_hdr0=''),
  3219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3219, CurrentID=0, Status='live', Symbol='HOXB9', Aliases='HOX-2.5, HOX2, HOX2E', description='homeobox B9', other_designations='homeo box 2E|homeo box B9|homeobox protein Hox-2.5|homeobox protein Hox-2E', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48615128, end_position_on_the_genomic_accession=48626473, orientation='minus', exon_count=3, OMIM=142964, no_hdr0=''),
  3220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3220, CurrentID=0, Status='live', Symbol='HOXC@', Aliases='HOX3@', description='homeobox C cluster', other_designations='homeo box C cluster', map_location='12q12-q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3221, CurrentID=0, Status='live', Symbol='HOXC4', Aliases='HOX3, HOX3E, cp19', description='homeobox C4', other_designations='homeo box 3E|homeo box C4|homeobox protein CP19|homeobox protein Hox-3E', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54016852, end_position_on_the_genomic_accession=54056030, orientation='plus', exon_count=4, OMIM=142974, no_hdr0=''),
  3222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3222, CurrentID=0, Status='live', Symbol='HOXC5', Aliases='CP11, HOX3, HOX3D', description='homeobox C5', other_designations='homeo box C5|homeobox protein CP11|homeobox protein Hox-3D', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54016852, end_position_on_the_genomic_accession=54035361, orientation='plus', exon_count=3, OMIM=142973, no_hdr0=''),
  3223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3223, CurrentID=0, Status='live', Symbol='HOXC6', Aliases='CP25, HHO.C8, HOX3, HOX3C', description='homeobox C6', other_designations='homeo box 3C|homeo box C6|homeo box C8 protein|homeobox protein CP25|homeobox protein HHO.C8|homeobox protein Hox-3C', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54016852, end_position_on_the_genomic_accession=54030823, orientation='plus', exon_count=3, OMIM=142972, no_hdr0=''),
  3224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3224, CurrentID=0, Status='live', Symbol='HOXC8', Aliases='HOX3, HOX3A', description='homeobox C8', other_designations='Hox-3.1, mouse, homolog of|homeo box 3A|homeo box C8|homeobox protein Hox-3A', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54009106, end_position_on_the_genomic_accession=54012763, orientation='plus', exon_count=2, OMIM=142970, no_hdr0=''),
  3225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3225, CurrentID=0, Status='live', Symbol='HOXC9', Aliases='HOX3, HOX3B', description='homeobox C9', other_designations='homeo box C9|homeobox protein Hox-3B', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54000093, end_position_on_the_genomic_accession=54003337, orientation='plus', exon_count=2, OMIM=142971, no_hdr0=''),
  3226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3226, CurrentID=0, Status='live', Symbol='HOXC10', Aliases='HOX3I', description='homeobox C10', other_designations='homeo box C10|homeobox protein Hox-3I|homeoprotein C10', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53985162, end_position_on_the_genomic_accession=53990279, orientation='plus', exon_count=2, OMIM=605560, no_hdr0=''),
  3227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3227, CurrentID=0, Status='live', Symbol='HOXC11', Aliases='HOX3H', description='homeobox C11', other_designations='homeo box C11|homeobox protein Hox-3H', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53973126, end_position_on_the_genomic_accession=53976419, orientation='plus', exon_count=2, OMIM=605559, no_hdr0=''),
  3228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3228, CurrentID=0, Status='live', Symbol='HOXC12', Aliases='HOC3F, HOX3, HOX3F', description='homeobox C12', other_designations='homeo box 3F|homeo box C12|homeobox protein Hox-3F', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53954930, end_position_on_the_genomic_accession=53956566, orientation='plus', exon_count=2, OMIM=142975, no_hdr0=''),
  3229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3229, CurrentID=0, Status='live', Symbol='HOXC13', Aliases='ECTD9, HOX3, HOX3G', description='homeobox C13', other_designations='NUP98/HOXC13|homeo box 3G|homeobox protein Hox-3G', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53938792, end_position_on_the_genomic_accession=53946544, orientation='plus', exon_count=2, OMIM=142976, no_hdr0=''),
  3230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3230, CurrentID=0, Status='live', Symbol='HOXD@', Aliases='HOX4@', description='homeobox D cluster', other_designations='homeo box D cluster', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3231, CurrentID=0, Status='live', Symbol='HOXD1', Aliases='HOX4, HOX4G, Hox-4.7', description='homeobox D1', other_designations='homeo box 4G|homeo box D1|homeobox protein Hox-GG', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176188579, end_position_on_the_genomic_accession=176190907, orientation='plus', exon_count=3, OMIM=142987, no_hdr0=''),
  3232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3232, CurrentID=0, Status='live', Symbol='HOXD3', Aliases='HOX1D, HOX4, HOX4A, Hox-4.1', description='homeobox D3', other_designations='Hox-4.1, mouse, homolog of|homeo box D3|homeobox protein Hox-4A', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176136921, end_position_on_the_genomic_accession=176173102, orientation='plus', exon_count=9, OMIM=142980, no_hdr0=''),
  3233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3233, CurrentID=0, Status='live', Symbol='HOXD4', Aliases='HHO.C13, HOX-5.1, HOX4, HOX4B, Hox-4.2', description='homeobox D4', other_designations='Hox-4.2, mouse, homolog of homeo box X|homeobox protein HHO.C13|homeobox protein Hox-4B|homeobox protein Hox-5.1', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176150394, end_position_on_the_genomic_accession=176153223, orientation='plus', exon_count=2, OMIM=142981, no_hdr0=''),
  3234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3234, CurrentID=0, Status='live', Symbol='HOXD8', Aliases='HOX4, HOX4E, HOX5.4', description='homeobox D8', other_designations='Hox-4.5|homeo box 4E|homeo box D8|homeobox protein 5.4|homeobox protein Hox-4E|homeobox protein Hox-5.4', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176129694, end_position_on_the_genomic_accession=176132695, orientation='plus', exon_count=2, OMIM=142985, no_hdr0=''),
  3235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3235, CurrentID=0, Status='live', Symbol='HOXD9', Aliases='HOX4, HOX4C, Hox-4.3, Hox-5.2', description='homeobox D9', other_designations='Hox-4.3, mouse, homolog of|homeo box D9|homeobox protein Hox-4C|homeobox protein Hox-5.2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176122685, end_position_on_the_genomic_accession=176124917, orientation='plus', exon_count=2, OMIM=142982, no_hdr0=''),
  3236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3236, CurrentID=0, Status='live', Symbol='HOXD10', Aliases='HOX4, HOX4D, HOX4E, Hox-4.4', description='homeobox D10', other_designations='homeo box 4D|homeo box D10|homeobox protein Hox-4D|homeobox protein Hox-4E', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176116764, end_position_on_the_genomic_accession=176119942, orientation='plus', exon_count=2, OMIM=142984, no_hdr0=''),
  3237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3237, CurrentID=0, Status='live', Symbol='HOXD11', Aliases='HOX4, HOX4F', description='homeobox D11', other_designations='Hox-4.6, mouse, homolog of|homeo box 4F|homeo box D11|homeobox protein Hox-4F', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176107356, end_position_on_the_genomic_accession=176109588, orientation='plus', exon_count=2, OMIM=142986, no_hdr0=''),
  3238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3238, CurrentID=0, Status='live', Symbol='HOXD12', Aliases='HOX4H', description='homeobox D12', other_designations='Hox-4.7, mouse, homolog of|homeo box D12|homeobox protein Hox-4H', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176098513, end_position_on_the_genomic_accession=176103770, orientation='plus', exon_count=3, OMIM=142988, no_hdr0=''),
  3239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3239, CurrentID=0, Status='live', Symbol='HOXD13', Aliases='BDE, BDSD, HOX4I, SPD', description='homeobox D13', other_designations='homeo box 4I|homeo box D13|homeobox protein Hox-4I', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176087505, end_position_on_the_genomic_accession=176095938, orientation='plus', exon_count=4, OMIM=142989, no_hdr0=''),
  3240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3240, CurrentID=0, Status='live', Symbol='HP', Aliases='BP2ALPHA2, HPA1S, HP', description='haptoglobin', other_designations='binding peptide|haptoglobin alpha(1S)-beta|haptoglobin alpha(2FS)-beta|haptoglobin, alpha polypeptide|haptoglobin, beta polypeptide|zonulin', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72054609, end_position_on_the_genomic_accession=72061056, orientation='plus', exon_count=7, OMIM=140100, no_hdr0=''),
  3241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3241, CurrentID=0, Status='live', Symbol='HPCAL1', Aliases='BDR1, HLP2, VILIP-3', description='hippocalcin-like 1', other_designations='calcium-binding protein BDR-1|visinin-like protein 3', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10302889, end_position_on_the_genomic_accession=10427617, orientation='plus', exon_count=15, OMIM=600207, no_hdr0=''),
  3242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3242, CurrentID=0, Status='live', Symbol='HPD', Aliases='4-HPPD, 4HPPD, GLOD3, HPPDASE, PPD', description='4-hydroxyphenylpyruvate dioxygenase', other_designations='4-hydroxyphenylpyruvic acid oxidase|glyoxalase domain containing 3', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121839527, end_position_on_the_genomic_accession=121888611, orientation='minus', exon_count=17, OMIM=609695, no_hdr0=''),
  3244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3244, CurrentID=0, Status='live', Symbol='HPE1', Aliases='HPEC', description='holoprosencephaly 1, alobar', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=236100, no_hdr0=''),
  3247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3247, CurrentID=0, Status='live', Symbol='HPFH2', Aliases='', description='hereditary persistence of fetal hemoglobin, heterocellular, Indian type', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=142335, no_hdr0=''),
  3248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3248, CurrentID=0, Status='live', Symbol='HPGD', Aliases='15-PGDH, PGDH, PGDH1, PHOAR1, SDR36C1', description='hydroxyprostaglandin dehydrogenase 15-(NAD)', other_designations='NAD+-dependent 15-hydroxyprostaglandin dehydrogenase|prostaglandin dehydrogenase 1|short chain dehydrogenase/reductase family 36C member 1|short chain dehydrogenase/reductase family 36C, member 1', map_location='4q34-q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174490177, end_position_on_the_genomic_accession=174522898, orientation='minus', exon_count=9, OMIM=601688, no_hdr0=''),
  3249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3249, CurrentID=0, Status='live', Symbol='HPN', Aliases='TMPRSS1', description='hepsin', other_designations='transmembrane protease serine 1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35040506, end_position_on_the_genomic_accession=35066577, orientation='plus', exon_count=15, OMIM=142440, no_hdr0=''),
  3250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3250, CurrentID=0, Status='live', Symbol='HPR', Aliases='A-259H10.2, HP', description='haptoglobin-related protein', other_designations='Haptoglobin-related locus', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72063226, end_position_on_the_genomic_accession=72077246, orientation='plus', exon_count=7, OMIM=140210, no_hdr0=''),
  3251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3251, CurrentID=0, Status='live', Symbol='HPRT1', Aliases='HGPRT, HPRT', description='hypoxanthine phosphoribosyltransferase 1', other_designations='HGPRTase|hypoxanthine-guanine phosphoribosyltransferase 1', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134452842, end_position_on_the_genomic_accession=134500668, orientation='plus', exon_count=11, OMIM=308000, no_hdr0=''),
  100418991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418991, CurrentID=0, Status='live', Symbol='LOC100418991', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20904883, end_position_on_the_genomic_accession=20905643, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3254, CurrentID=0, Status='live', Symbol='HPRT1P2', Aliases='HPRTP2', description='hypoxanthine phosphoribosyltransferase 1 pseudogene 2', other_designations='hypoxanthine phosphoribosyltransferase pseudogene 2', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=30248374, end_position_on_the_genomic_accession=30249481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3255, CurrentID=0, Status='live', Symbol='HPRT1P3', Aliases='HPRTP3, HPRTP4', description='hypoxanthine phosphoribosyltransferase 1 pseudogene 3', other_designations='hypoxanthine phosphoribosyltransferase pseudogene 3|hypoxanthine phosphoribosyltransferase pseudogene 4', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93990676, end_position_on_the_genomic_accession=93991522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3257, CurrentID=0, Status='live', Symbol='HPS1', Aliases='HPS', description='Hermansky-Pudlak syndrome 1', other_designations='', map_location='10q23.1-q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98416198, end_position_on_the_genomic_accession=98446963, orientation='minus', exon_count=21, OMIM=604982, no_hdr0=''),
  3258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3258, CurrentID=0, Status='live', Symbol='HPT', Aliases='HPTX, HYPX', description='hypoparathyroidism', other_designations='', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=307700, no_hdr0=''),
  3259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3259, CurrentID=0, Status='live', Symbol='HPV6AI1', Aliases='', description='human papillomavirus (type 6a) integration site 1', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604461, no_hdr0=''),
  3260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3260, CurrentID=0, Status='live', Symbol='HPV18I1', Aliases='D8S37, HPV18I', description='human papilloma virus (type 18) integration site 1', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=167959, no_hdr0=''),
  3261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3261, CurrentID=0, Status='live', Symbol='HPV18I2', Aliases='PAL2, pal2A, pal2B', description='human papillomavirus (type 18) integration site 2', other_designations='papillomavirus-associated locus 2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=167960, no_hdr0=''),
  3262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3262, CurrentID=0, Status='live', Symbol='HPVC1', Aliases='HPV18E5L, PE5L', description='human papillomavirus (type 18) E5 central sequence-like 1', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54201224, end_position_on_the_genomic_accession=54202421, orientation='minus', exon_count=1, OMIM=600762, no_hdr0=''),
  3263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3263, CurrentID=0, Status='live', Symbol='HPX', Aliases='HX', description='hemopexin', other_designations='beta-1B-glycoprotein', map_location='11p15.5-p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6431038, end_position_on_the_genomic_accession=6441024, orientation='minus', exon_count=11, OMIM=142290, no_hdr0=''),
  3265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3265, CurrentID=0, Status='live', Symbol='HRAS', Aliases='C-BAS/HAS, C-H-RAS, C-HA-RAS1, CTLO, H-RASIDX, HAMSV1, RASH1, p21ras, HRAS', description='Harvey rat sarcoma viral oncogene homolog', other_designations='GTP- and GDP-binding peptide B|Ha-Ras1 proto-oncoprotein|Harvey rat sarcoma viral oncoprotein|Ras family small GTP binding protein H-Ras|c-has/bas p21 protein|c-ras-Ki-2 activated oncogene|p19 H-RasIDX protein|transformation gene: oncogene HAMSV|transforming protein p21|v-Ha-ras Harvey rat sarcoma viral oncogene homolog', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=532242, end_position_on_the_genomic_accession=535561, orientation='minus', exon_count=7, OMIM=190020, no_hdr0=''),
  3266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3266, CurrentID=0, Status='live', Symbol='ERAS', Aliases='HRAS2, HRASP', description='ES cell expressed Ras', other_designations='E-Ras|embryonic stem cell-expressed Ras|small GTPase protein E-Ras|v-Ha-ras Harvey rat sarcoma viral oncogene homolog 2|v-Ha-ras Harvey rat sarcoma viral oncogene homolog pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48826513, end_position_on_the_genomic_accession=48829869, orientation='plus', exon_count=2, OMIM=300437, no_hdr0=''),
  3267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3267, CurrentID=0, Status='live', Symbol='AGFG1', Aliases='HRB, RAB, RIP', description='ArfGAP with FG repeats 1', other_designations='HIV-1 Rev binding protein|HIV-1 Rev-binding protein|Rab, Rev/Rex activation domain-binding protein|arf-GAP domain and FG repeats-containing protein 1|hRIP, Rev interacting protein|nucleoporin-like protein RIP|rev-interacting protein|rev/Rex activation domain-binding protein', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227472172, end_position_on_the_genomic_accession=227561222, orientation='plus', exon_count=17, OMIM=600862, no_hdr0=''),
  3268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3268, CurrentID=0, Status='live', Symbol='AGFG2', Aliases='HRBL, RABR', description='ArfGAP with FG repeats 2', other_designations='HIV-1 Rev-binding protein-like protein|Rev/Rex activation domain binding protein-related|arf-GAP domain and FG repeats-containing protein 2|nucleoporin|rev/Rex activation domain-binding protein related', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100539211, end_position_on_the_genomic_accession=100568220, orientation='plus', exon_count=13, OMIM=604019, no_hdr0=''),
  3269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3269, CurrentID=0, Status='live', Symbol='HRH1', Aliases='H1-R, H1R, HH1R, hisH1', description='histamine receptor H1', other_designations='histamine receptor, subclass H1', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11137093, end_position_on_the_genomic_accession=11263253, orientation='plus', exon_count=8, OMIM=600167, no_hdr0=''),
  3270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3270, CurrentID=0, Status='live', Symbol='HRC', Aliases='', description='histidine rich calcium binding protein', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49151198, end_position_on_the_genomic_accession=49155429, orientation='minus', exon_count=6, OMIM=142705, no_hdr0=''),
  3272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3272, CurrentID=0, Status='live', Symbol='HRES1', Aliases='HRES-1, HRES-1/p28', description='HTLV-1 related endogenous sequence', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=227473294, end_position_on_the_genomic_accession=227473969, orientation='minus', exon_count=1, OMIM=143025, no_hdr0=''),
  3273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3273, CurrentID=0, Status='live', Symbol='HRG', Aliases='HPRGP, THPH11, HRG', description='histidine-rich glycoprotein', other_designations='histidine-proline-rich glycoprotein', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186665958, end_position_on_the_genomic_accession=186678235, orientation='plus', exon_count=7, OMIM=142640, no_hdr0=''),
  3274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3274, CurrentID=0, Status='live', Symbol='HRH2', Aliases='H2R', description='histamine receptor H2', other_designations='HH2R|gastric receptor 1|gastric receptor I', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175657796, end_position_on_the_genomic_accession=175710756, orientation='plus', exon_count=8, OMIM=142703, no_hdr0=''),
  3275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3275, CurrentID=0, Status='live', Symbol='PRMT2', Aliases='HRMT1L1', description='protein arginine methyltransferase 2', other_designations='HMT1 (hnRNP methyltransferase, S. cerevisiae)-like 1|HMT1 hnRNP methyltransferase-like 1|PRMT2 alpha|PRMT2 beta|PRMT2 gamma|histone-arginine N-methyltransferase PRMT2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46635153, end_position_on_the_genomic_accession=46665685, orientation='plus', exon_count=13, OMIM=601961, no_hdr0=''),
  3276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3276, CurrentID=0, Status='live', Symbol='PRMT1', Aliases='ANM1, HCP1, HRMT1L2, IR1B4', description='protein arginine methyltransferase 1', other_designations='HMT1 (hnRNP methyltransferase, S. cerevisiae)-like 2|heterogeneous nuclear ribonucleoprotein methyltransferase 1-like 2|histone-arginine N-methyltransferase PRMT1|interferon receptor 1-bound protein 4', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49677152, end_position_on_the_genomic_accession=49688450, orientation='plus', exon_count=12, OMIM=602950, no_hdr0=''),
  646058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646058, CurrentID=0, Status='live', Symbol='LOC646058', Aliases='', description='chromosome 4 open reading frame 27 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180085607, end_position_on_the_genomic_accession=180087055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3280, CurrentID=0, Status='live', Symbol='HES1', Aliases='HES-1, HHL, HRY, bHLHb39', description='hes family bHLH transcription factor 1', other_designations='class B basic helix-loop-helix protein 39|hairy homolog|hairy-like protein', map_location='3q28-q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194136142, end_position_on_the_genomic_accession=194138612, orientation='plus', exon_count=4, OMIM=139605, no_hdr0=''),
  3281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3281, CurrentID=0, Status='live', Symbol='HSBP1', Aliases='NPC-A-13', description='heat shock factor binding protein 1', other_designations='nasopharyngeal carcinoma-associated antigen 13', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83807903, end_position_on_the_genomic_accession=83813002, orientation='plus', exon_count=4, OMIM=604553, no_hdr0=''),
  3283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3283, CurrentID=0, Status='live', Symbol='HSD3B1', Aliases='3BETAHSD, HSD3B, HSDB3, HSDB3A, I, SDR11E1', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 1', other_designations='3 beta-hydroxysteroid dehydrogenase/Delta 5-->4-isomerase type I|3-beta-HSD I|3-beta-hydroxy-5-ene steroid dehydrogenase|3-beta-hydroxy-Delta(5)-steroid dehydrogenase|delta-5-3-ketosteroid isomerase|hydroxy-delta-5-steroid dehydrogenase, 3 beta-and steroid delta-isomerase 1|progesterone reductase|short chain dehydrogenase/reductase family 11E, member 1|steroid Delta-isomerase|trophoblast antigen FDO161G', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119507124, end_position_on_the_genomic_accession=119515060, orientation='plus', exon_count=4, OMIM=109715, no_hdr0=''),
  3284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3284, CurrentID=0, Status='live', Symbol='HSD3B2', Aliases='HSD3B, HSDB, SDR11E2', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 2', other_designations='3 beta-HSD type II|3 beta-hydroxysteroid dehydrogenase type II, delta 5-delta 4-isomerase type II, 3 beta-HSD type II|3 beta-hydroxysteroid dehydrogenase/Delta 5-->4-isomerase type II|3-beta-HSD II|3-beta-HSD adrenal and gonadal type|3-beta-hydroxy-5-ene steroid dehydrogenase|3-beta-hydroxy-Delta(5)-steroid dehydrogenase|delta 5-delta 4-isomerase type II|progesterone reductase|short chain dehydrogenase/reductase family 11E, member 2', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119414931, end_position_on_the_genomic_accession=119423039, orientation='plus', exon_count=5, OMIM=613890, no_hdr0=''),
  134359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134359, CurrentID=0, Status='live', Symbol='POC5', Aliases='C5orf37', description='POC5 centriolar protein', other_designations='POC5 centriolar protein homolog|hPOC5|protein of centriole 5|proteome of centriole 5', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75674198, end_position_on_the_genomic_accession=75717488, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3290, CurrentID=0, Status='live', Symbol='HSD11B1', Aliases='11-DH, 11-beta-HSD1, CORTRD2, HDL, HSD11, HSD11B, HSD11L, SDR26C1', description='hydroxysteroid (11-beta) dehydrogenase 1', other_designations='short chain dehydrogenase/reductase family 26C member 1|short chain dehydrogenase/reductase family 26C, member 1', map_location='1q32-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209686180, end_position_on_the_genomic_accession=209734950, orientation='plus', exon_count=7, OMIM=600713, no_hdr0=''),
  3291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3291, CurrentID=0, Status='live', Symbol='HSD11B2', Aliases='AME, AME1, HSD11K, HSD2, SDR9C3', description='hydroxysteroid (11-beta) dehydrogenase 2', other_designations='-HSD11 type II|11-DH2|11-HSD type II|11-beta-HSD type II|11-beta-HSD2|11-beta-hydroxysteroid dehydrogenase type 2|11-beta-hydroxysteroid dehydrogenase type II|NAD-dependent 11-beta-hydroxysteroid dehydrogenase|short chain dehydrogenase/reductase family 9C member 3', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67431133, end_position_on_the_genomic_accession=67437553, orientation='plus', exon_count=5, OMIM=614232, no_hdr0=''),
  3292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3292, CurrentID=0, Status='live', Symbol='HSD17B1', Aliases='EDH17B2, EDHB17, HSD17, SDR28C1', description='hydroxysteroid (17-beta) dehydrogenase 1', other_designations='17-beta-HSD 1|17-beta-hydroxysteroid dehydrogenase type 1|20 alpha-hydroxysteroid dehydrogenase|20-alpha-HSD|E2DH|estradiol 17-beta-dehydrogenase-1|hydroxysteroid (17-beta) dehydrogenase 1 isoform|placental 17-beta-hydroxysteroid dehydrogenase|short chain dehydrogenase/reductase family 28C member 1|short chain dehydrogenase/reductase family 28CE, member 1', map_location='17q11-q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42551943, end_position_on_the_genomic_accession=42555214, orientation='plus', exon_count=5, OMIM=109684, no_hdr0=''),
  3293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3293, CurrentID=0, Status='live', Symbol='HSD17B3', Aliases='EDH17B3, SDR12C2', description='hydroxysteroid (17-beta) dehydrogenase 3', other_designations='17-beta-HSD 3|17-beta-HSD3|17-beta-hydroxysteroid dehydrogenase type 3|hydroxysteroid dehydrogenase 3|short chain dehydrogenase/reductase family 12C member 2|short chain dehydrogenase/reductase family 12C, member 2|testicular 17-beta-hydroxysteroid dehydrogenase', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96235306, end_position_on_the_genomic_accession=96313983, orientation='minus', exon_count=16, OMIM=605573, no_hdr0=''),
  3294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3294, CurrentID=0, Status='live', Symbol='HSD17B2', Aliases='EDH17B2, HSD17, SDR9C2', description='hydroxysteroid (17-beta) dehydrogenase 2', other_designations='17-beta-HSD 2|17-beta-hydroxysteroid dehydrogenase type 2|20 alpha-hydroxysteroid dehydrogenase|20-alpha-HSD|E2DH|microsomal 17-beta-hydroxysteroid dehydrogenase|short chain dehydrogenase/reductase family 9C member 2|short chain dehydrogenase/reductase family 9C, member 2|testosterone 17-beta-dehydrogenase', map_location='16q24.1-q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82035237, end_position_on_the_genomic_accession=82098534, orientation='plus', exon_count=6, OMIM=109685, no_hdr0=''),
  3295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3295, CurrentID=0, Status='live', Symbol='HSD17B4', Aliases='DBP, MFE-2, MPF-2, PRLTS1, SDR8C1', description='hydroxysteroid (17-beta) dehydrogenase 4', other_designations='17-beta-HSD 4|17-beta-HSD IV|17-beta-hydroxysteroid dehydrogenase 4|17beta-estradiol dehydrogenase type IV|3-alpha,7-alpha,12-alpha-trihydroxy-5-beta-cholest-24-enoyl-CoA hydratase|D-3-hydroxyacyl-CoA dehydratase|D-bifunctional protein, peroxisomal|beta-hydroxyacyl dehydrogenase|beta-keto-reductase|hydroxysteroid dehydrogenase 4|multifunctional protein 2|peroxisomal multifunctional protein 2|short chain dehydrogenase/reductase family 8C member 1|short chain dehydrogenase/reductase family 8C, member 1', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119452443, end_position_on_the_genomic_accession=119542335, orientation='plus', exon_count=25, OMIM=601860, no_hdr0=''),
  3297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3297, CurrentID=0, Status='live', Symbol='HSF1', Aliases='HSTF1', description='heat shock transcription factor 1', other_designations='HSF 1|HSTF 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144291588, end_position_on_the_genomic_accession=144314722, orientation='plus', exon_count=14, OMIM=140580, no_hdr0=''),
  3298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3298, CurrentID=0, Status='live', Symbol='HSF2', Aliases='HSF 2, HSTF 2', description='heat shock transcription factor 2', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122399551, end_position_on_the_genomic_accession=122433119, orientation='plus', exon_count=13, OMIM=140581, no_hdr0=''),
  3299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3299, CurrentID=0, Status='live', Symbol='HSF4', Aliases='CTM, CTRCT5', description='heat shock transcription factor 4', other_designations='HSF 4|HSTF 4|hHSF4', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67163385, end_position_on_the_genomic_accession=67169945, orientation='plus', exon_count=15, OMIM=602438, no_hdr0=''),
  3300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3300, CurrentID=0, Status='live', Symbol='DNAJB2', Aliases='CMT2T, DSMA5, HSJ-1, HSJ1, HSPF3', description='DnaJ heat shock protein family (Hsp40) member B2', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 2|dnaJ protein homolog 1|heat shock 40 kDa protein 3|heat shock protein J1|heat shock protein, neuronal DNAJ-like 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219279318, end_position_on_the_genomic_accession=219286900, orientation='plus', exon_count=9, OMIM=604139, no_hdr0=''),
  3301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3301, CurrentID=0, Status='live', Symbol='DNAJA1', Aliases='DJ-2, DjA1, HDJ2, HSDJ, HSJ-2, HSJ2, HSPF4, NEDD7, hDJ-2', description='DnaJ heat shock protein family (Hsp40) member A1', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1|dnaJ protein homolog 2|heat shock 40 kDa protein 4|heat shock protein J2|heat shock protein, DNAJ-like 2|human DnaJ protein 2|neural precursor cell expressed, developmentally down-regulated 7', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33025211, end_position_on_the_genomic_accession=33039064, orientation='plus', exon_count=9, OMIM=602837, no_hdr0=''),
  3303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3303, CurrentID=0, Status='live', Symbol='HSPA1A', Aliases='HEL-S-103, HSP70-1, HSP70-1A, HSP70.1, HSP70I, HSP72, HSPA1', description='heat shock protein family A (Hsp70) member 1A', other_designations='HSP70-1/HSP70-2|HSP70.1/HSP70.2|dnaK-type molecular chaperone HSP70-1|epididymis secretory protein Li 103|heat shock 70 kDa protein 1|heat shock 70 kDa protein 1/2|heat shock 70 kDa protein 1A/1B|heat shock 70kD protein 1A|heat shock 70kDa protein 1A|heat shock-induced protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31815514, end_position_on_the_genomic_accession=31817942, orientation='plus', exon_count=1, OMIM=140550, no_hdr0=''),
  3304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3304, CurrentID=0, Status='live', Symbol='HSPA1B', Aliases='HSP70-1B, HSP70-2, HSP70.2', description='heat shock protein family A (Hsp70) member 1B', other_designations='HSP70-1/HSP70-2|HSP70.1/HSP70.2|heat shock 70 kDa protein 1/2|heat shock 70 kDa protein 1A/1B|heat shock 70 kDa protein 2|heat shock 70kD protein 1B|heat shock 70kDa protein 1B', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31827735, end_position_on_the_genomic_accession=31830254, orientation='plus', exon_count=1, OMIM=603012, no_hdr0=''),
  3305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3305, CurrentID=0, Status='live', Symbol='HSPA1L', Aliases='HSP70-1L, HSP70-HOM, HSP70T, hum70t', description='heat shock protein family A (Hsp70) member 1 like', other_designations='heat shock 10kDa protein 1-like|heat shock 70 kDa protein 1-Hom|heat shock 70 kDa protein 1L|heat shock 70kD protein-like 1|heat shock 70kDa protein 1-like', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31809619, end_position_on_the_genomic_accession=31821999, orientation='minus', exon_count=5, OMIM=140559, no_hdr0=''),
  3306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3306, CurrentID=0, Status='live', Symbol='HSPA2', Aliases='HSP70-2, HSP70-3', description='heat shock protein family A (Hsp70) member 2', other_designations='heat shock 70kD protein 2|heat shock 70kDa protein 2', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64540468, end_position_on_the_genomic_accession=64543237, orientation='plus', exon_count=1, OMIM=140560, no_hdr0=''),
  3308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3308, CurrentID=0, Status='live', Symbol='HSPA4', Aliases='APG-2, HEL-S-5a, HS24/P52, HSPH2, RY, hsp70, hsp70RY', description='heat shock protein family A (Hsp70) member 4', other_designations='epididymis secretory sperm binding protein Li 5a|heat shock 70-related protein APG-2|heat shock 70kD protein 4|heat shock 70kDa protein 4|heat shock protein, 110 kDa|hsp70 RY', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133051970, end_position_on_the_genomic_accession=133105017, orientation='plus', exon_count=19, OMIM=601113, no_hdr0=''),
  3309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3309, CurrentID=0, Status='live', Symbol='HSPA5', Aliases='BIP, GRP78, HEL-S-89n, MIF2', description='heat shock protein family A (Hsp70) member 5', other_designations='endoplasmic reticulum lumenal Ca(2+)-binding protein grp78|epididymis secretory sperm binding protein Li 89n|heat shock 70kDa protein 5 (glucose-regulated protein, 78kDa)|immunoglobulin heavy chain-binding protein', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125234848, end_position_on_the_genomic_accession=125241387, orientation='minus', exon_count=8, OMIM=138120, no_hdr0=''),
  3310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3310, CurrentID=0, Status='live', Symbol='HSPA6', Aliases='', description='heat shock protein family A (Hsp70) member 6', other_designations="heat shock 70 kDa protein B'|heat shock 70kD protein 6 (HSP70B')|heat shock 70kDa protein 6 (HSP70B')", map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161524540, end_position_on_the_genomic_accession=161526897, orientation='plus', exon_count=1, OMIM=140555, no_hdr0=''),
  3311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3311, CurrentID=0, Status='live', Symbol='HSPA7', Aliases='HSP70B', description='heat shock protein family A (Hsp70) member 7', other_designations='heat shock 70kD protein 7 (HSP70B)|heat shock 70kDa protein 7 (HSP70B)', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161606059, end_position_on_the_genomic_accession=161608551, orientation='plus', exon_count=1, OMIM=140556, no_hdr0=''),
  3312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3312, CurrentID=0, Status='live', Symbol='HSPA8', Aliases='HEL-33, HEL-S-72p, HSC54, HSC70, HSC71, HSP71, HSP73, HSPA10, LAP-1, LAP1, NIP71', description='heat shock protein family A (Hsp70) member 8', other_designations='LPS-associated protein 1|N-myristoyltransferase inhibitor protein 71|constitutive heat shock protein 70|epididymis luminal protein 33|epididymis secretory sperm binding protein Li 72p|heat shock 70kDa protein 8|heat shock 70kd protein 10|heat shock cognate protein 54|lipopolysaccharide-associated protein 1', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123057492, end_position_on_the_genomic_accession=123062366, orientation='minus', exon_count=10, OMIM=600816, no_hdr0=''),
  3313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3313, CurrentID=0, Status='live', Symbol='HSPA9', Aliases='CRP40, CSA, GRP-75, GRP75, HEL-S-124mB, MOT, MOT2, MTHSP75, PBP74, HSPA9', description='heat shock protein family A (Hsp70) member 9', other_designations='75 kDa glucose-regulated protein|catecholamine-regulated protein 40|epididymis secretory sperm binding protein Li 124m|heat shock 70kD protein 9B|heat shock 70kDa protein 9 (mortalin)|mortalin, perinuclear|mortalin-2|p66-mortalin|peptide-binding protein 74', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138554882, end_position_on_the_genomic_accession=138575629, orientation='minus', exon_count=17, OMIM=600548, no_hdr0=''),
  3314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3314, CurrentID=0, Status='live', Symbol='HSPA8P1', Aliases='HSPAP1', description='heat shock protein family A (Hsp70) member 8 pseudogene 1', other_designations='heat shock 70kD protein pseudogene 1|heat shock 70kDa protein 8 pseudogene 1|heat shock 70kDa protein pseudogene 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121202843, end_position_on_the_genomic_accession=121205092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3315, CurrentID=0, Status='live', Symbol='HSPB1', Aliases='CMT2F, HEL-S-102, HMN2B, HS.76067, HSP27, HSP28, Hsp25, SRP27', description='heat shock protein family B (small) member 1', other_designations='28 kDa heat shock protein|HSP 27|epididymis secretory protein Li 102|estrogen-regulated 24 kDa protein|heat shock 27 kDa protein|heat shock 27kD protein 1|heat shock 27kDa protein 1|stress-responsive protein 27', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76302558, end_position_on_the_genomic_accession=76304297, orientation='plus', exon_count=3, OMIM=602195, no_hdr0=''),
  3316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3316, CurrentID=0, Status='live', Symbol='HSPB2', Aliases='HSP27, Hs.78846, LOH11CR1K, MKBP', description='heat shock protein family B (small) member 2', other_designations='DMPK-binding protein|heat shock 27kDa protein 2|heat-shock protein beta-2', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111912736, end_position_on_the_genomic_accession=111914093, orientation='plus', exon_count=2, OMIM=602179, no_hdr0=''),
  134391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134391, CurrentID=0, Status='live', Symbol='GPR151', Aliases='GALR4, GALRL, GPCR, PGR7', description='G protein-coupled receptor 151', other_designations='G-protein coupled receptor PGR7|GPCR-2037|galanin receptor 4|putative G-protein coupled receptor', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146514854, end_position_on_the_genomic_accession=146516113, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3320, CurrentID=0, Status='live', Symbol='HSP90AA1', Aliases='EL52, HEL-S-65p, HSP86, HSP89A, HSP90A, HSP90N, HSPC1, HSPCA, HSPCAL1, HSPCAL4, HSPN, Hsp89, Hsp90, LAP-2, LAP2', description='heat shock protein 90kDa alpha family class A member 1', other_designations='HSP 86|LPS-associated protein 2|epididymis luminal secretory protein 52|epididymis secretory sperm binding protein Li 65p|heat shock 86 kDa|heat shock 90kD protein 1, alpha|heat shock 90kD protein 1, alpha-like 4|heat shock 90kD protein, alpha-like 4|heat shock 90kDa protein 1, alpha|heat shock protein 90kDa alpha (cytosolic), class A member 1|lipopolysaccharide-associated protein 2|renal carcinoma antigen NY-REN-38', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102080738, end_position_on_the_genomic_accession=102139759, orientation='minus', exon_count=13, OMIM=140571, no_hdr0=''),
  3321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3321, CurrentID=0, Status='live', Symbol='IGSF3', Aliases='EWI-3, LCDD, V8', description='immunoglobulin superfamily member 3', other_designations='glu-Trp-Ile EWI motif-containing protein 3|immunoglobin superfamily, member 3|immunoglobulin superfamily, member 3', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116574398, end_position_on_the_genomic_accession=116667755, orientation='minus', exon_count=15, OMIM=603491, no_hdr0=''),
  3322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3322, CurrentID=0, Status='live', Symbol='HSP90AA3P', Aliases='HSPCAL1', description='heat shock protein 90kDa alpha family class A member 3, pseudogene', other_designations='Heat-shock 90kD protein, alpha-like 1|heat shock 90kD protein 1, alpha-like 1|heat shock 90kDa protein 1, alpha-like 1|heat shock protein 90kDa alpha (cytosolic), class A member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class A member 3 (pseudogene)|heat shock protein 90kDa alpha (cytosolic), class A member 3, pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158523613, end_position_on_the_genomic_accession=158526480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3323, CurrentID=0, Status='live', Symbol='HSP90AA4P', Aliases='HSP90Ad, HSPCAL2', description='heat shock protein 90kDa alpha family class A member 4, pseudogene', other_designations='heat shock 90kD protein 1, alpha-like 2|heat shock 90kDa protein 1, alpha-like 2|heat shock protein 90Ad|heat shock protein 90kDa alpha (cytosolic), class A member 4, pseudogene', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189472906, end_position_on_the_genomic_accession=189475853, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3324, CurrentID=0, Status='live', Symbol='HSP90AA2P', Aliases='HSP90AA2, HSP90ALPHA, HSPCA, HSPCAL3', description='heat shock protein 90kDa alpha family class A member 2, pseudogene', other_designations='heat shock 90kD protein, alpha-like 3|heat shock 90kDa protein 1, alpha-like 3|heat shock protein 86|heat shock protein 90kDa alpha (cytosolic), class A member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class A member 2, pseudogene', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27888171, end_position_on_the_genomic_accession=27891092, orientation='minus', exon_count=0, OMIM=140575, no_hdr0=''),
  3326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3326, CurrentID=0, Status='live', Symbol='HSP90AB1', Aliases='D6S182, HSP84, HSP90B, HSPC2, HSPCB', description='heat shock protein 90kDa alpha family class B member 1', other_designations='HSP90-beta|heat shock 84 kDa|heat shock 90kD protein 1, beta|heat shock protein 90 kDa|heat shock protein 90kDa alpha (cytosolic), class B member 1', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44246166, end_position_on_the_genomic_accession=44253888, orientation='plus', exon_count=13, OMIM=140572, no_hdr0=''),
  3327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3327, CurrentID=0, Status='live', Symbol='HSP90AB3P', Aliases='HSP90BC, HSPCP1', description='heat shock protein 90kDa alpha family class B member 3, pseudogene', other_designations='heat shock 90kD protein 1, beta pseudogene 1|heat shock 90kDa protein 1, beta pseudogene 1|heat shock protein 90Bc|heat shock protein 90kDa alpha (cytosolic), class B member 3, pseudogene', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87891748, end_position_on_the_genomic_accession=87894288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3329, CurrentID=0, Status='live', Symbol='HSPD1', Aliases='CPN60, GROEL, HLD4, HSP-60, HSP60, HSP65, HuCHA60, SPG13', description='heat shock protein family D (Hsp60) member 1', other_designations='60 kDa chaperonin|P60 lymphocyte protein|chaperonin 60|heat shock 60kDa protein 1 (chaperonin)|heat shock protein 65|mitochondrial matrix protein P1|short heat shock protein 60 Hsp60s1', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197486584, end_position_on_the_genomic_accession=197500274, orientation='minus', exon_count=13, OMIM=118190, no_hdr0=''),
  3332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3332, CurrentID=0, Status='live', Symbol='HSPD1P3', Aliases='HSP60P1, HSP60P3, HSP60P4, HSPD1-7P, HSPDP1, HSPDP3, HSPDP4', description='heat shock protein family D (Hsp60) member 1 pseudogene 3', other_designations='heat shock 60kD protein 1 (chaperonin) pseudogene 3|heat shock 60kDa protein 1 (chaperonin) pseudogene 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7418786, end_position_on_the_genomic_accession=7421032, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103156999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103156999, CurrentID=0, Status='live', Symbol='LOC103156999', Aliases='', description='dynein, light chain, Tctex-type 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12115519, end_position_on_the_genomic_accession=12116243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3336, CurrentID=0, Status='live', Symbol='HSPE1', Aliases='CPN10, EPF, GROES, HSP10', description='heat shock protein family E (Hsp10) member 1', other_designations='10 kDa chaperonin|chaperonin 10|early pregnancy factor|early-pregnancy factor|heat shock 10kD protein 1 (chaperonin 10)', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197499997, end_position_on_the_genomic_accession=197503463, orientation='plus', exon_count=4, OMIM=600141, no_hdr0=''),
  3337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3337, CurrentID=0, Status='live', Symbol='DNAJB1', Aliases='HSPF1, Hdj1, Hsp40, RSPH16B, Sis1', description='DnaJ heat shock protein family (Hsp40) member B1', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 1|dnaJ protein homolog 1|heat shock 40 kDa protein 1|human DnaJ protein 1|radial spoke 16 homolog B', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14514764, end_position_on_the_genomic_accession=14529711, orientation='minus', exon_count=7, OMIM=604572, no_hdr0=''),
  3338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3338, CurrentID=0, Status='live', Symbol='DNAJC4', Aliases='DANJC4, HSPF2, MCG18', description='DnaJ heat shock protein family (Hsp40) member C4', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 4|dnaJ-like protein HSPF2|heat shock 40kD protein 2|multiple endocrine neoplasia type 1 candidate protein number 18', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64230234, end_position_on_the_genomic_accession=64234281, orientation='plus', exon_count=7, OMIM=604189, no_hdr0=''),
  3339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3339, CurrentID=0, Status='live', Symbol='HSPG2', Aliases='HSPG, PLC, PRCAN, SJA, SJS, SJS1', description='heparan sulfate proteoglycan 2', other_designations='endorepellin (domain V region)|perlecan proteoglycan', map_location='1p36.1-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21822232, end_position_on_the_genomic_accession=21937297, orientation='minus', exon_count=104, OMIM=142461, no_hdr0=''),
  3340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3340, CurrentID=0, Status='live', Symbol='NDST1', Aliases='HSST, MRT46, NST1', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 1', other_designations='HSNST 1|N-HSST 1|N-heparan sulfate sulfotransferase 1|NDST-1|[Heparan sulfate]-glucosamine N-sulfotransferase 1|glucosaminyl N-deacetylase/N-sulfotransferase 1|heparan sulfate-N-deacetylase/N-sulfotransferase', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150485823, end_position_on_the_genomic_accession=150558211, orientation='plus', exon_count=19, OMIM=600853, no_hdr0=''),
  3341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3341, CurrentID=0, Status='live', Symbol='HTC1', Aliases='', description='hypertrichosis 1 (universalis, congenital)', other_designations='Ambras syndrome|hypertrichosis universalis (universalis, congenital)', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145701, no_hdr0=''),
  3342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3342, CurrentID=0, Status='live', Symbol='HTC2', Aliases='CGH, CXINSq27.1, HCG', description='hypertrichosis 2 (generalized, congenital)', other_designations='hypertrichosis 2 (generalised, congenital)', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=307150, no_hdr0=''),
  3343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3343, CurrentID=0, Status='live', Symbol='HTL', Aliases='HLT, LEUT', description='high L-leucine transport', other_designations='Leucine transport, high', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151310, no_hdr0=''),
  3344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3344, CurrentID=0, Status='live', Symbol='FOXN2', Aliases='HTLF', description='forkhead box N2', other_designations='human T-cell leukemia virus enhancer factor', map_location='2p22-p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48314644, end_position_on_the_genomic_accession=48379295, orientation='plus', exon_count=10, OMIM=143089, no_hdr0=''),
  3346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3346, CurrentID=0, Status='live', Symbol='HTN1', Aliases='HIS1', description='histatin 1', other_designations='PPB|histidine-rich protein 1|post-PB protein', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70050415, end_position_on_the_genomic_accession=70058848, orientation='plus', exon_count=6, OMIM=142701, no_hdr0=''),
  3347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3347, CurrentID=0, Status='live', Symbol='HTN3', Aliases='HIS2, HTN2, HTN5', description='histatin 3', other_designations='PB|basic histidine-rich protein|histatin 5|histatin-4|histatin-5|histatin-6|histatin-7|histatin-8|histatin-9|histidine-rich protein 3|hst', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70028413, end_position_on_the_genomic_accession=70036538, orientation='plus', exon_count=6, OMIM=142702, no_hdr0=''),
  3349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3349, CurrentID=0, Status='live', Symbol='HTOR', Aliases='', description='5-hydroxytryptamine (serotonin) oxygenase regulator', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=143460, no_hdr0=''),
  3350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3350, CurrentID=0, Status='live', Symbol='HTR1A', Aliases='5-HT-1A, 5-HT1A, 5HT1a, ADRB2RL1, ADRBRL1, G-21, PFMCD', description='5-hydroxytryptamine (serotonin) receptor 1A, G protein-coupled', other_designations='5-HT1a receptor|guanine nucleotide-binding regulatory protein-coupled receptor', map_location='5q11.2-q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=63960048, end_position_on_the_genomic_accession=63962292, orientation='minus', exon_count=1, OMIM=109760, no_hdr0=''),
  3351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3351, CurrentID=0, Status='live', Symbol='HTR1B', Aliases='5-HT-1B, 5-HT-1D-beta, 5-HT1B, 5-HT1DB, HTR1D2, HTR1DB, S12', description='5-hydroxytryptamine (serotonin) receptor 1B, G protein-coupled', other_designations='serotonin 1D beta receptor|serotonin receptor 1B', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77462231, end_position_on_the_genomic_accession=77463403, orientation='minus', exon_count=1, OMIM=182131, no_hdr0=''),
  3352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3352, CurrentID=0, Status='live', Symbol='HTR1D', Aliases='5-HT1D, HT1DAA, HTRL, RDC4, HTR1D', description='5-hydroxytryptamine (serotonin) receptor 1D, G protein-coupled', other_designations='5-HT-1D|5-HT-1D-alpha|serotonin 1D alpha receptor|serotonin receptor 1D', map_location='1p36.3-p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23191895, end_position_on_the_genomic_accession=23194729, orientation='minus', exon_count=1, OMIM=182133, no_hdr0=''),
  100130936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130936, CurrentID=0, Status='live', Symbol='LOC100130936', Aliases='', description='acireductone dioxygenase 1 pseudogene', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41331122, end_position_on_the_genomic_accession=41333241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3354, CurrentID=0, Status='live', Symbol='HTR1E', Aliases='5-HT1E', description='5-hydroxytryptamine (serotonin) receptor 1E, G protein-coupled', other_designations='5-HT-1E|S31|serotonin receptor 1E', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86937306, end_position_on_the_genomic_accession=87016679, orientation='plus', exon_count=3, OMIM=182132, no_hdr0=''),
  3355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3355, CurrentID=0, Status='live', Symbol='HTR1F', Aliases='5-HT-1F, 5-HT1F, 5HT6, HTR1EL, MR77', description='5-hydroxytryptamine (serotonin) receptor 1F, G protein-coupled', other_designations='serotonin receptor 1F', map_location='3p12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87792767, end_position_on_the_genomic_accession=87994856, orientation='plus', exon_count=7, OMIM=182134, no_hdr0=''),
  3356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3356, CurrentID=0, Status='live', Symbol='HTR2A', Aliases='5-HT2A, HTR2', description='5-hydroxytryptamine (serotonin) receptor 2A, G protein-coupled', other_designations='5-HT2 receptor|serotonin 5-HT-2A receptor', map_location='13q14-q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46831542, end_position_on_the_genomic_accession=46897076, orientation='minus', exon_count=4, OMIM=182135, no_hdr0=''),
  3357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3357, CurrentID=0, Status='live', Symbol='HTR2B', Aliases='5-HT(2B), 5-HT2B', description='5-hydroxytryptamine (serotonin) receptor 2B, G protein-coupled', other_designations='5-HT 2B receptor|5-HT-2B|5-hydroxytryptamine 2B receptor|5-hydroxytryptamine receptor 2B variant b|serotonin receptor 2B', map_location='2q36.3-q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231108230, end_position_on_the_genomic_accession=231125778, orientation='minus', exon_count=4, OMIM=601122, no_hdr0=''),
  3358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3358, CurrentID=0, Status='live', Symbol='HTR2C', Aliases='5-HT1C, 5-HT2C, 5-HTR2C, 5HTR2C, HTR1C', description='5-hydroxytryptamine (serotonin) receptor 2C, G protein-coupled', other_designations='5-hydroxytryptamine receptor 1C|serotonin 5-HT-1C receptor|serotonin 5-HT-2C receptor', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114584078, end_position_on_the_genomic_accession=114910061, orientation='plus', exon_count=7, OMIM=312861, no_hdr0=''),
  3359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3359, CurrentID=0, Status='live', Symbol='HTR3A', Aliases='5-HT-3, 5-HT3A, 5-HT3R, 5HT3R, HTR3', description='5-hydroxytryptamine (serotonin) receptor 3A, ionotropic', other_designations='5-HT3-A|5-hydroxytryptamine receptor 3|5HT3 serotonin receptor|serotonin receptor 3A|serotonin-gated ion channel receptor', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113975075, end_position_on_the_genomic_accession=113990313, orientation='plus', exon_count=9, OMIM=182139, no_hdr0=''),
  3360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3360, CurrentID=0, Status='live', Symbol='HTR4', Aliases='5-HT4, 5-HT4R', description='5-hydroxytryptamine (serotonin) receptor 4, G protein-coupled', other_designations='cardiac 5-HT4 receptor', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148451032, end_position_on_the_genomic_accession=148654527, orientation='minus', exon_count=12, OMIM=602164, no_hdr0=''),
  3361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3361, CurrentID=0, Status='live', Symbol='HTR5A', Aliases='5-HT5A', description='5-hydroxytryptamine (serotonin) receptor 5A, G protein-coupled', other_designations='5-HT-5|5-HT-5A', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155070324, end_position_on_the_genomic_accession=155087392, orientation='plus', exon_count=2, OMIM=601305, no_hdr0=''),
  3362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3362, CurrentID=0, Status='live', Symbol='HTR6', Aliases='5-HT6, 5-HT6R', description='5-hydroxytryptamine (serotonin) receptor 6, G protein-coupled', other_designations='serotonin receptor 6', map_location='1p36-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19665287, end_position_on_the_genomic_accession=19680966, orientation='plus', exon_count=3, OMIM=601109, no_hdr0=''),
  3363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3363, CurrentID=0, Status='live', Symbol='HTR7', Aliases='5-HT7', description='5-hydroxytryptamine (serotonin) receptor 7, adenylate cyclase-coupled', other_designations='5-HT-7|5-HT-X|5-hydroxytryptamine (serotonin) receptor 7 (adenylate cyclase-coupled)', map_location='10q21-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90740818, end_position_on_the_genomic_accession=90857914, orientation='minus', exon_count=4, OMIM=182137, no_hdr0=''),
  3364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3364, CurrentID=0, Status='live', Symbol='HUS1', Aliases='hHUS1', description='HUS1 checkpoint clamp component', other_designations='HUS1 checkpoint homolog|hus1+-like protein', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47963288, end_position_on_the_genomic_accession=47979625, orientation='minus', exon_count=9, OMIM=603760, no_hdr0=''),
  153478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153478, CurrentID=0, Status='live', Symbol='PLEKHG4B', Aliases='', description='pleckstrin homology and RhoGEF domain containing G4B', other_designations='CTD-2231H16.1|PH domain-containing family G member 4B|pleckstrin homology domain containing, family G (with RhoGef domain) member 4B', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54015, end_position_on_the_genomic_accession=189972, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  3369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3369, CurrentID=0, Status='live', Symbol='HVBS7', Aliases='', description='hepatitis B virus integration site 7', other_designations='', map_location='18q11.1-q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3370, CurrentID=0, Status='live', Symbol='HVBS8', Aliases='', description='hepatitis B virus integration site 8', other_designations='', map_location='17p12-p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3371, CurrentID=0, Status='live', Symbol='TNC', Aliases='150-225, DFNA56, GMEM, GP, HXB, JI, TN, TN-C', description='tenascin C', other_designations='GP 150-225|cytotactin|deafness, autosomal dominant 56|glioma-associated-extracellular matrix antigen|hexabrachion (tenascin)|myotendinous antigen|neuronectin|tenascin-C additional domain 1|tenascin-C isoform 14/AD1/16', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115019575, end_position_on_the_genomic_accession=115118257, orientation='minus', exon_count=30, OMIM=187380, no_hdr0=''),
  3373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3373, CurrentID=0, Status='live', Symbol='HYAL1', Aliases='HYAL-1, LUCA1, MPS9, NAT6', description='hyaluronoglucosaminidase 1', other_designations='luCa-1|lung carcinoma protein 1|plasma hyaluronidase|tumor suppressor LUCA-1', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50299889, end_position_on_the_genomic_accession=50312419, orientation='minus', exon_count=7, OMIM=607071, no_hdr0=''),
  3375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3375, CurrentID=0, Status='live', Symbol='IAPP', Aliases='DAP, IAP', description='islet amyloid polypeptide', other_designations='Islet amyloid polypeptide (diabetes-associated peptide; amylin)|amylin|diabetes-associated peptide|insulinoma amyloid peptide', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21372868, end_position_on_the_genomic_accession=21379980, orientation='plus', exon_count=3, OMIM=147940, no_hdr0=''),
  3376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3376, CurrentID=0, Status='live', Symbol='IARS', Aliases='IARS1, ILERS, ILRS, IRS, PRO0785', description='isoleucyl-tRNA synthetase', other_designations='isoleucine tRNA ligase 1, cytoplasmic|isoleucyl-tRNA synthetase, cytoplasmic', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92210207, end_position_on_the_genomic_accession=92293756, orientation='minus', exon_count=34, OMIM=600709, no_hdr0=''),
  3378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3378, CurrentID=0, Status='live', Symbol='IBD2', Aliases='', description='inflammatory bowel disease 2', other_designations='', map_location='12p13.2-q24.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601458, no_hdr0=''),
  100270643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270643, CurrentID=0, Status='live', Symbol='CLLS4', Aliases='', description='Leukemia, chronic lymphocytic susceptibility to, 4', other_designations='', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612558, no_hdr0=''),
  3381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3381, CurrentID=0, Status='live', Symbol='IBSP', Aliases='BNSP, BSP, BSP-II, SP-II', description='integrin-binding sialoprotein', other_designations='BSP II|bone sialoprotein II|cell-binding sialoprotein', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87799550, end_position_on_the_genomic_accession=87812449, orientation='plus', exon_count=7, OMIM=147563, no_hdr0=''),
  3382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3382, CurrentID=0, Status='live', Symbol='ICA1', Aliases='ICA69, ICAp69', description='islet cell autoantigen 1', other_designations='69 kDa islet cell autoantigen|diabetes mellitus type I autoantigen|islet cell autoantigen 1 isoform|islet cell autoantigen 1, 69kDa|islet cell autoantigen p69|p69', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8113184, end_position_on_the_genomic_accession=8262687, orientation='minus', exon_count=22, OMIM=147625, no_hdr0=''),
  3383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3383, CurrentID=0, Status='live', Symbol='ICAM1', Aliases='BB2, CD54, P3.58', description='intercellular adhesion molecule 1', other_designations='ICAM-1|cell surface glycoprotein P3.58|intercellular adhesion molecule 1 (CD54), human rhinovirus receptor|major group rhinovirus receptor', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10270841, end_position_on_the_genomic_accession=10286615, orientation='plus', exon_count=7, OMIM=147840, no_hdr0=''),
  3384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3384, CurrentID=0, Status='live', Symbol='ICAM2', Aliases='CD102', description='intercellular adhesion molecule 2', other_designations='ICAM-2', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64002595, end_position_on_the_genomic_accession=64020634, orientation='minus', exon_count=6, OMIM=146630, no_hdr0=''),
  3385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3385, CurrentID=0, Status='live', Symbol='ICAM3', Aliases='CD50, CDW50, ICAM-R', description='intercellular adhesion molecule 3', other_designations='ICAM-3|intercellular adhesion molecule-3', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10333776, end_position_on_the_genomic_accession=10339829, orientation='minus', exon_count=7, OMIM=146631, no_hdr0=''),
  3386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3386, CurrentID=0, Status='live', Symbol='ICAM4', Aliases='CD242, LW', description='intercellular adhesion molecule 4 (Landsteiner-Wiener blood group)', other_designations='CD242 antigen|LW blood group protein|Landsteiner-Wiener blood group antigen a|Landsteiner-Wiener blood group glycoprotein|intercellular adhesion molecule 4 (LW blood group)|intercellular adhesion molecule 4 isoform 3 anitgen LWa', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10286967, end_position_on_the_genomic_accession=10288584, orientation='plus', exon_count=2, OMIM=614088, no_hdr0=''),
  3388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3388, CurrentID=0, Status='live', Symbol='ICR1', Aliases='', description='ichthyosis congenita I, erythromatous lamellar ichthyosis, ichthyosiform erythroderma', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060580, CurrentID=0, Status='live', Symbol='LOC101060580', Aliases='', description='uncharacterized LOC101060580', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3390, CurrentID=0, Status='live', Symbol='ICR3', Aliases='', description='ichthyosis congenita III, vulgaris-like recessive ichthyosis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101059903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059903, CurrentID=0, Status='live', Symbol='HMSN5', Aliases='', description='Hereditary motor and sensory neuropathy V', other_designations='', map_location='4q34.3-q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600361, no_hdr0=''),
  101059905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059905, CurrentID=0, Status='live', Symbol='TCEB1P26', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 26', other_designations='info transcription elongation factor B (SIII), polypeptide 1 pseudogene 26', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18178932, end_position_on_the_genomic_accession=18179260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3394, CurrentID=0, Status='live', Symbol='IRF8', Aliases='H-ICSBP, ICSBP, ICSBP1, IMD32A, IMD32B, IRF-8', description='interferon regulatory factor 8', other_designations='interferon consensus sequence binding protein 1', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85899168, end_position_on_the_genomic_accession=85922606, orientation='plus', exon_count=9, OMIM=601565, no_hdr0=''),
  101059907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059907, CurrentID=0, Status='live', Symbol='LOC101059907', Aliases='', description='protein FAM207A pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6918341, end_position_on_the_genomic_accession=6928057, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3396, CurrentID=0, Status='live', Symbol='ICT1', Aliases='DS-1, DS1, MRP-L58', description='immature colon carcinoma transcript 1', other_designations='39S ribosomal protein L58, mitochondrial|digestion substraction 1|immature colon carcinoma transcript 1 protein', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75012664, end_position_on_the_genomic_accession=75021261, orientation='plus', exon_count=6, OMIM=603000, no_hdr0=''),
  3397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3397, CurrentID=0, Status='live', Symbol='ID1', Aliases='ID, bHLHb24', description='inhibitor of DNA binding 1, dominant negative helix-loop-helix protein', other_designations='class B basic helix-loop-helix protein 24|dJ857M17.1.2 (inhibitor of DNA binding 1, dominant negative helix-loop-helix protein)|inhibitor of differentiation 1', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31605283, end_position_on_the_genomic_accession=31606515, orientation='plus', exon_count=1, OMIM=600349, no_hdr0=''),
  3398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3398, CurrentID=0, Status='live', Symbol='ID2', Aliases='GIG8A, ID2H, bHLHb26, ID2', description='inhibitor of DNA binding 2, dominant negative helix-loop-helix protein', other_designations='DNA-binding protein inhibitor ID2|cell growth-inhibiting gene 8|class B basic helix-loop-helix protein 26|helix-loop-helix protein ID2|inhibitor of differentiation 2', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8681983, end_position_on_the_genomic_accession=8684453, orientation='plus', exon_count=3, OMIM=600386, no_hdr0=''),
  3399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3399, CurrentID=0, Status='live', Symbol='ID3', Aliases='HEIR-1, bHLHb25', description='inhibitor of DNA binding 3, dominant negative helix-loop-helix protein', other_designations='ID-like protein inhibitor HLH 1R21|class B basic helix-loop-helix protein 25|helix-loop-helix protein HEIR-1|inhibitor of differentiation 3', map_location='1p36.13-p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23557930, end_position_on_the_genomic_accession=23559794, orientation='minus', exon_count=3, OMIM=600277, no_hdr0=''),
  3400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3400, CurrentID=0, Status='live', Symbol='ID4', Aliases='IDB4, bHLHb27', description='inhibitor of DNA binding 4, dominant negative helix-loop-helix protein', other_designations='class B basic helix-loop-helix protein 27|inhibitor of differentiation 4', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19837370, end_position_on_the_genomic_accession=19842200, orientation='plus', exon_count=3, OMIM=600581, no_hdr0=''),
  3402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3402, CurrentID=0, Status='live', Symbol='IDDM3', Aliases='', description='insulin-dependent diabetes mellitus 3', other_designations='', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600318, no_hdr0=''),
  3403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3403, CurrentID=0, Status='live', Symbol='IDDM4', Aliases='', description='insulin-dependent diabetes mellitus 4', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600319, no_hdr0=''),
  101059916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059916, CurrentID=0, Status='live', Symbol='LOC101059916', Aliases='', description='nucleoporin 35kDa pseudogene', other_designations='nucleoporin NUP53-like', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79543308, end_position_on_the_genomic_accession=79544878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3405, CurrentID=0, Status='live', Symbol='IDDM6', Aliases='', description='insulin-dependent diabetes mellitus 6', other_designations='', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601941, no_hdr0=''),
  3406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3406, CurrentID=0, Status='live', Symbol='IDDM7', Aliases='', description='insulin-dependent diabetes mellitus 7', other_designations='', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600321, no_hdr0=''),
  3407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3407, CurrentID=0, Status='live', Symbol='IDDM8', Aliases='', description='insulin-dependent diabetes mellitus 8', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600883, no_hdr0=''),
  3408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3408, CurrentID=0, Status='live', Symbol='IDDM9', Aliases='', description='insulin-dependent diabetes mellitus 9', other_designations='', map_location='3q22.3-q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270648, CurrentID=0, Status='live', Symbol='NIFKP3', Aliases='MKI67IPP3', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 3', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 3', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7604254, end_position_on_the_genomic_accession=7605273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3410, CurrentID=0, Status='live', Symbol='IDDM11', Aliases='', description='insulin-dependent diabetes mellitus 11', other_designations='', map_location='14q24.3-q31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601208, no_hdr0=''),
  3412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3412, CurrentID=0, Status='live', Symbol='IDDM13', Aliases='', description='insulin-dependent diabetes mellitus 13', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601318, no_hdr0=''),
  3413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3413, CurrentID=0, Status='live', Symbol='IDDM14', Aliases='', description='insulin-dependent diabetes mellitus 14', other_designations='', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3414, CurrentID=0, Status='live', Symbol='IDDM15', Aliases='', description='insulin-dependent diabetes mellitus 15', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601666, no_hdr0=''),
  3415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3415, CurrentID=0, Status='live', Symbol='IDDM16', Aliases='', description='insulin-dependent diabetes mellitus 16', other_designations='', map_location='14q32.32-q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3416, CurrentID=0, Status='live', Symbol='IDE', Aliases='INSULYSIN', description='insulin-degrading enzyme', other_designations='Abeta-degrading protease|insulin protease|insulinase', map_location='10q23-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92451684, end_position_on_the_genomic_accession=92574095, orientation='minus', exon_count=28, OMIM=146680, no_hdr0=''),
  3417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3417, CurrentID=0, Status='live', Symbol='IDH1', Aliases='HEL-216, HEL-S-26, IDCD, IDH, IDP, IDPC, PICD', description='isocitrate dehydrogenase 1 (NADP+)', other_designations='NADP(+)-specific ICDH|NADP-dependent isocitrate dehydrogenase, cytosolic|NADP-dependent isocitrate dehydrogenase, peroxisomal|epididymis luminal protein 216|epididymis secretory protein Li 26|isocitrate dehydrogenase 1 (NADP+), soluble|oxalosuccinate decarboxylase', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208236227, end_position_on_the_genomic_accession=208255143, orientation='minus', exon_count=12, OMIM=147700, no_hdr0=''),
  3418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3418, CurrentID=0, Status='live', Symbol='IDH2', Aliases='D2HGA2, ICD-M, IDH, IDHM, IDP, IDPM, mNADP-IDH', description='isocitrate dehydrogenase 2 (NADP+), mitochondrial', other_designations='NADP(+)-specific ICDH|oxalosuccinate decarboxylase', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90083978, end_position_on_the_genomic_accession=90102554, orientation='minus', exon_count=12, OMIM=147650, no_hdr0=''),
  3419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3419, CurrentID=0, Status='live', Symbol='IDH3A', Aliases='', description='isocitrate dehydrogenase 3 (NAD+) alpha', other_designations='H-IDH alpha|NAD(+)-specific ICDH subunit alpha|NAD(H)-specific isocitrate dehydrogenase alpha subunit|NAD+-specific ICDH|isocitrate dehydrogenase (NAD+) alpha chain|isocitric dehydrogenase subunit alpha', map_location='15q25.1-q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78149362, end_position_on_the_genomic_accession=78170542, orientation='plus', exon_count=12, OMIM=601149, no_hdr0=''),
  3420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3420, CurrentID=0, Status='live', Symbol='IDH3B', Aliases='H-IDHB, RP46', description='isocitrate dehydrogenase 3 (NAD+) beta', other_designations='NAD(+)-specific ICDH subunit beta|NAD+-specific ICDH|NAD+-specific isocitrate dehydrogenase b subunit|NAD+-specific isocitrate dehydrogenase beta|isocitrate dehydrogenase, NAD(+)-specific, mitochondrial, beta subunit|isocitric dehydrogenase subunit beta', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2658394, end_position_on_the_genomic_accession=2664219, orientation='minus', exon_count=14, OMIM=604526, no_hdr0=''),
  3421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3421, CurrentID=0, Status='live', Symbol='IDH3G', Aliases='H-IDHG', description='isocitrate dehydrogenase 3 (NAD+) gamma', other_designations='IDH-gamma|NAD (H)-specific isocitrate dehydrogenase gamma subunit|NAD(+)-specific ICDH subunit gamma|NAD+-specific ICDH|isocitrate dehydrogenase 3 gamma|isocitrate dehydrogenase, NAD(+)-specific, mitochondrial, gamma subunit|isocitric dehydrogenase subunit gamma', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153785766, end_position_on_the_genomic_accession=153794523, orientation='minus', exon_count=12, OMIM=300089, no_hdr0=''),
  3422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3422, CurrentID=0, Status='live', Symbol='IDI1', Aliases='IPP1, IPPI1', description='isopentenyl-diphosphate delta isomerase 1', other_designations='IPP isomerase 1|isopentenyl diphosphate dimethylallyl diphosphate isomerase 1|isopentenyl pyrophosphate isomerase 1', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1040023, end_position_on_the_genomic_accession=1056699, orientation='minus', exon_count=7, OMIM=604055, no_hdr0=''),
  3423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3423, CurrentID=0, Status='live', Symbol='IDS', Aliases='MPS2, SIDS', description='iduronate 2-sulfatase', other_designations='alpha-L-iduronate sulfate sulfatase|iduronate 2-sulfatase 14 kDa chain|iduronate 2-sulfatase 42 kDa chain|idursulfase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149478764, end_position_on_the_genomic_accession=149505354, orientation='minus', exon_count=11, OMIM=300823, no_hdr0=''),
  3424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3424, CurrentID=0, Status='live', Symbol='IDSP1', Aliases='IDS2', description='iduronate 2-sulfatase pseudogene 1', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149525000, end_position_on_the_genomic_accession=149526417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3425, CurrentID=0, Status='live', Symbol='IDUA', Aliases='IDA, MPS1', description='iduronidase, alpha-L-', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=986997, end_position_on_the_genomic_accession=1004557, orientation='plus', exon_count=17, OMIM=252800, no_hdr0=''),
  3426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3426, CurrentID=0, Status='live', Symbol='CFI', Aliases='AHUS3, ARMD13, C3BINA, C3b-INA, FI, IF, KAF', description='complement factor I', other_designations='C3B/C4B inactivator|C3b-inactivator|Konglutinogen-activating factor|complement component I|complement control protein factor I|complement factor I heavy chain|light chain of factor I', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109731221, end_position_on_the_genomic_accession=109802225, orientation='minus', exon_count=16, OMIM=217030, no_hdr0=''),
  3428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3428, CurrentID=0, Status='live', Symbol='IFI16', Aliases='IFNGIP1, PYHIN2', description='interferon, gamma-inducible protein 16', other_designations='interferon-gamma induced protein IFI 16|interferon-inducible myeloid differentiation transcriptional activator', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158999973, end_position_on_the_genomic_accession=159055155, orientation='plus', exon_count=13, OMIM=147586, no_hdr0=''),
  3429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3429, CurrentID=0, Status='live', Symbol='IFI27', Aliases='FAM14D, ISG12, ISG12A, P27', description='interferon, alpha-inducible protein 27', other_designations='2310061N23Rik|ISG12(a)|interferon alpha-induced 11.5 kDa protein|interferon-stimulated gene 12a protein', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94110733, end_position_on_the_genomic_accession=94116699, orientation='plus', exon_count=6, OMIM=600009, no_hdr0=''),
  3430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3430, CurrentID=0, Status='live', Symbol='IFI35', Aliases='IFP35', description='interferon induced protein 35', other_designations='IFP 35|ifi-35|interferon-induced protein 35', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43006725, end_position_on_the_genomic_accession=43014459, orientation='plus', exon_count=7, OMIM=600735, no_hdr0=''),
  3431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3431, CurrentID=0, Status='live', Symbol='SP110', Aliases='IFI41, IFI75, IPR1, VODI', description='SP110 nuclear body protein', other_designations='interferon-induced protein 41, 30kD|interferon-induced protein 41/75|interferon-induced protein 75, 52kD|phosphoprotein 41|phosphoprotein 75|speckled 110 kDa|transcriptional coactivator Sp110', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230167842, end_position_on_the_genomic_accession=230225729, orientation='minus', exon_count=23, OMIM=604457, no_hdr0=''),
  728989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728989, CurrentID=0, Status='live', Symbol='LOC728989', Aliases='', description='phosphodiesterase 4D interacting protein pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147019336, end_position_on_the_genomic_accession=147043039, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3433, CurrentID=0, Status='live', Symbol='IFIT2', Aliases='G10P2, GARG-39, IFI-54, IFI-54K, IFI54, IFIT-2, ISG-54 K, ISG-54K, ISG54, P54, cig42', description='interferon induced protein with tetratricopeptide repeats 2', other_designations='Interferon, alpha-inducible protein (MW 54kD)|interferon-induced 54 kDa protein|interferon-induced protein 54', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89301949, end_position_on_the_genomic_accession=89309276, orientation='plus', exon_count=2, OMIM=147040, no_hdr0=''),
  3434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3434, CurrentID=0, Status='live', Symbol='IFIT1', Aliases='C56, G10P1, IFI-56, IFI-56K, IFI56, IFIT-1, IFNAI1, ISG56, P56, RNM561', description='interferon induced protein with tetratricopeptide repeats 1', other_designations='interferon, alpha-inducible protein (MW 56kD)|interferon-induced 56 kDa protein|interferon-inducible mRNA 561', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89392546, end_position_on_the_genomic_accession=89406487, orientation='plus', exon_count=4, OMIM=147690, no_hdr0=''),
  101059948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059948, CurrentID=0, Status='live', Symbol='LOC101059948', Aliases='', description='uncharacterized LOC101059948', other_designations='CTB-60B18.10', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49018248, end_position_on_the_genomic_accession=49019484, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3437, CurrentID=0, Status='live', Symbol='IFIT3', Aliases='CIG-49, GARG-49, IFI60, IFIT4, IRG2, ISG60, P60, RIG-G, cig41', description='interferon induced protein with tetratricopeptide repeats 3', other_designations='CIG49|IFI-60K|IFIT-3|IFIT-4|ISG-60|interferon-induced 60 kDa protein|interferon-induced protein with tetratricopeptide repeats 4|retinoic acid-induced gene G protein', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89327819, end_position_on_the_genomic_accession=89340968, orientation='plus', exon_count=4, OMIM=604650, no_hdr0=''),
  3438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3438, CurrentID=0, Status='live', Symbol='IFN1@', Aliases='IFNA', description='interferon, type 1, cluster', other_designations='', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3439, CurrentID=0, Status='live', Symbol='IFNA1', Aliases='IFL, IFN, IFN-ALPHA, IFN-alphaD3, IFNA@, IFNA1', description='interferon, alpha 1', other_designations='IFN-alpha 1b|IFN-alpha-1/13|interferon alpha 1b|interferon alpha-D|interferon-alpha1|leIF D', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21440454, end_position_on_the_genomic_accession=21441316, orientation='plus', exon_count=1, OMIM=147660, no_hdr0=''),
  3440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3440, CurrentID=0, Status='live', Symbol='IFNA2', Aliases='IFN-alphaA, IFNAB, INFA2, IFNA2', description='interferon, alpha 2', other_designations='IFN-alpha-2|alpha-2a interferon|interferon alpha 2a|interferon alpha 2b|interferon alpha A|leIF A', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21384255, end_position_on_the_genomic_accession=21385397, orientation='minus', exon_count=1, OMIM=147562, no_hdr0=''),
  3441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3441, CurrentID=0, Status='live', Symbol='IFNA4', Aliases='IFN-alpha4a, INFA4', description='interferon, alpha 4', other_designations='IFN-alpha-4|interferon alpha-4B|interferon alpha-76|interferon alpha-M1', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21186618, end_position_on_the_genomic_accession=21187599, orientation='minus', exon_count=1, OMIM=147564, no_hdr0=''),
  3442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3442, CurrentID=0, Status='live', Symbol='IFNA5', Aliases='IFN-alpha-5, IFN-alphaG, INA5, INFA5, leIF G', description='interferon, alpha 5', other_designations='interferon alpha-61|interferon alpha-G', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21304614, end_position_on_the_genomic_accession=21305313, orientation='minus', exon_count=1, OMIM=147565, no_hdr0=''),
  3443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3443, CurrentID=0, Status='live', Symbol='IFNA6', Aliases='IFN-alphaK', description='interferon, alpha 6', other_designations='IFN-alpha-6|interferon alpha-54|interferon alpha-K|leIF K', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21350318, end_position_on_the_genomic_accession=21350887, orientation='minus', exon_count=1, OMIM=147566, no_hdr0=''),
  3444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3444, CurrentID=0, Status='live', Symbol='IFNA7', Aliases='IFN-alphaJ, IFNA-J', description='interferon, alpha 7', other_designations='IFN-alpha 7|IFN-alpha-7|IFN-alpha-J1|interferon alpha-J|interferon alpha-J1|leIF J', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21201469, end_position_on_the_genomic_accession=21202205, orientation='minus', exon_count=1, OMIM=147567, no_hdr0=''),
  3445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3445, CurrentID=0, Status='live', Symbol='IFNA8', Aliases='IFN-alphaB', description='interferon, alpha 8', other_designations="IFN-alpha-8|interferon alpha type 201|interferon alpha-B|interferon alpha-B'|interferon alpha-B2|leIF B", map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21409147, end_position_on_the_genomic_accession=21410185, orientation='plus', exon_count=1, OMIM=147568, no_hdr0=''),
  3446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3446, CurrentID=0, Status='live', Symbol='IFNA10', Aliases='IFN-alphaC', description='interferon, alpha 10', other_designations='IFN-alpha-10|interferon alpha-6L|interferon alpha-C|leIF C', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21206181, end_position_on_the_genomic_accession=21207143, orientation='minus', exon_count=1, OMIM=147577, no_hdr0=''),
  3447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3447, CurrentID=0, Status='live', Symbol='IFNA13', Aliases='', description='interferon, alpha 13', other_designations='IFN-alpha-1/13|interferon alpha-D|leIF D', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21367372, end_position_on_the_genomic_accession=21368076, orientation='minus', exon_count=1, OMIM=147578, no_hdr0=''),
  3448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3448, CurrentID=0, Status='live', Symbol='IFNA14', Aliases='IFN-alphaH, LEIF2H', description='interferon, alpha 14', other_designations='IFN-alpha-14|interferon alpha-H|interferon lambda-2-H|leIF H', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21239202, end_position_on_the_genomic_accession=21239979, orientation='minus', exon_count=1, OMIM=147579, no_hdr0=''),
  3449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3449, CurrentID=0, Status='live', Symbol='IFNA16', Aliases='IFN-alpha-16, IFN-alphaO', description='interferon, alpha 16', other_designations='IFN-alpha-N-protein|interferon alpha-WA', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21216373, end_position_on_the_genomic_accession=21217311, orientation='minus', exon_count=1, OMIM=147580, no_hdr0=''),
  101059962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059962, CurrentID=0, Status='live', Symbol='LOC101059962', Aliases='', description='serine/threonine-protein phosphatase 1 regulatory subunit 10-like', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10193930, end_position_on_the_genomic_accession=10197199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3451, CurrentID=0, Status='live', Symbol='IFNA17', Aliases='IFN-alphaI, IFNA, INFA, LEIF2C1', description='interferon, alpha 17', other_designations="IFN-alpha-17|interferon alpha-88|interferon alpha-I'|interferon alpha-T|interferon alpha-WA|leIF I", map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21227243, end_position_on_the_genomic_accession=21228222, orientation='minus', exon_count=1, OMIM=147583, no_hdr0=''),
  3452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3452, CurrentID=0, Status='live', Symbol='IFNA21', Aliases='IFN-alphaI, LeIF F, leIF-F', description='interferon, alpha 21', other_designations='IFN-alpha-21|interferon alpha-F|leukocyte interferon protein', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21165637, end_position_on_the_genomic_accession=21166660, orientation='minus', exon_count=1, OMIM=147584, no_hdr0=''),
  3453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3453, CurrentID=0, Status='live', Symbol='IFNA22P', Aliases='IFNAP22', description='interferon, alpha 22, pseudogene', other_designations='interferon, alpha 7 pseudogene|interferon, alpha pseudogene 22', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21277688, end_position_on_the_genomic_accession=21278563, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3454, CurrentID=0, Status='live', Symbol='IFNAR1', Aliases='AVP, IFN-alpha-REC, IFNAR, IFNBR, IFRC', description='interferon (alpha, beta and omega) receptor 1', other_designations='CRF2-1|IFN-R-1|IFN-alpha/beta receptor 1|alpha-type antiviral protein|beta-type antiviral protein|cytokine receptor class-II member 1|cytokine receptor family 2 member 1|interferon receptor 1|interferon-alpha/beta receptor alpha chain|interferon-beta receptor 1|type I interferon receptor 1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33324443, end_position_on_the_genomic_accession=33359823, orientation='plus', exon_count=13, OMIM=107450, no_hdr0=''),
  3455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3455, CurrentID=0, Status='live', Symbol='IFNAR2', Aliases='IFN-R, IFN-alpha-REC, IFNABR, IFNARB', description='interferon (alpha, beta and omega) receptor 2', other_designations='IFN-R-2|IFN-alpha/beta receptor 2|human interferon alpha/beta receptor|interferon alpha binding protein|interferon-alpha/beta receptor beta chain|type I interferon receptor 2', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33229895, end_position_on_the_genomic_accession=33264513, orientation='plus', exon_count=10, OMIM=602376, no_hdr0=''),
  3456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3456, CurrentID=0, Status='live', Symbol='IFNB1', Aliases='IFB, IFF, IFN-beta, IFNB', description='interferon, beta 1, fibroblast', other_designations='fibroblast interferon', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21077105, end_position_on_the_genomic_accession=21077944, orientation='minus', exon_count=1, OMIM=147640, no_hdr0=''),
  3458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3458, CurrentID=0, Status='live', Symbol='IFNG', Aliases='IFG, IFI', description='interferon, gamma', other_designations='IFN-gamma|immune interferon', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68154770, end_position_on_the_genomic_accession=68159741, orientation='minus', exon_count=4, OMIM=147570, no_hdr0=''),
  3459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3459, CurrentID=0, Status='live', Symbol='IFNGR1', Aliases='CD119, IFNGR, IMD27A, IMD27B', description='interferon gamma receptor 1', other_designations='AVP, type 2|CD119 antigen|CDw119|IFN-gamma receptor 1|IFN-gamma-R1|antiviral protein, type 2|immune interferon receptor 1|interferon-gamma receptor alpha chain', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137197484, end_position_on_the_genomic_accession=137219844, orientation='minus', exon_count=9, OMIM=107470, no_hdr0=''),
  3460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3460, CurrentID=0, Status='live', Symbol='IFNGR2', Aliases='AF-1, IFGR2, IFNGT1, IMD28', description='interferon gamma receptor 2 (interferon gamma transducer 1)', other_designations='IFN-gamma receptor 2|IFN-gamma-R2|interferon gamma receptor accessory factor 1|interferon gamma receptor accessory factor-1|interferon gamma receptor beta chain|interferon gamma transducer 1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33402896, end_position_on_the_genomic_accession=33437521, orientation='plus', exon_count=9, OMIM=147569, no_hdr0=''),
  3461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3461, CurrentID=0, Status='live', Symbol='IFNA11P', Aliases='IFNP11', description='interferon, alpha 11, pseudogene', other_designations='interferon pseudogene 11', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21398614, end_position_on_the_genomic_accession=21399139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059974, CurrentID=0, Status='live', Symbol='LOC101059974', Aliases='', description='p53 and DNA damage-regulated protein 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85567878, end_position_on_the_genomic_accession=85568242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3463, CurrentID=0, Status='live', Symbol='IFNA20P', Aliases='IFNP20', description='interferon, alpha 20, pseudogene', other_designations='interferon pseudogene 20', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21321300, end_position_on_the_genomic_accession=21321827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059977, CurrentID=0, Status='live', Symbol='LOC101059977', Aliases='', description='predicted GPI-anchored protein 58', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43246257, end_position_on_the_genomic_accession=43247108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3466, CurrentID=0, Status='live', Symbol='IFNR', Aliases='IFNGM, IFNGM2', description='interferon production regulator', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147573, no_hdr0=''),
  3467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3467, CurrentID=0, Status='live', Symbol='IFNW1', Aliases='', description='interferon, omega 1', other_designations='IFN-omega 1, interferon omega-1|interferon alpha-II-1', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21140632, end_position_on_the_genomic_accession=21142145, orientation='minus', exon_count=1, OMIM=147553, no_hdr0=''),
  101059982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059982, CurrentID=0, Status='live', Symbol='TCEAL8P1', Aliases='', description='transcription elongation factor A (SII)-like 8 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=66673116, end_position_on_the_genomic_accession=66673527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3474, CurrentID=0, Status='live', Symbol='IFNWP19', Aliases='', description='interferon, omega 1 pseudogene 19', other_designations='interferon, omega 19 (pseudogene)', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21455439, end_position_on_the_genomic_accession=21456068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3475, CurrentID=0, Status='live', Symbol='IFRD1', Aliases='PC4, TIS7', description='interferon-related developmental regulator 1', other_designations='12-O-tetradecanoylphorbol-13-acetate-induced sequence 7|TPA induced sequence 7|nerve growth factor-inducible protein PC4|pheochromocytoma cell-4', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112423144, end_position_on_the_genomic_accession=112477203, orientation='plus', exon_count=17, OMIM=603502, no_hdr0=''),
  3476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3476, CurrentID=0, Status='live', Symbol='IGBP1', Aliases='ALPHA-4, IBP1', description='immunoglobulin (CD79A) binding protein 1', other_designations='B cell signal transduction molecule alpha 4|B-cell signal transduction molecule alpha 4|CD79a-binding protein 1|bA351K23.1 (immunoglobulin binding protein 1 (CD79A) )|protein alpha-4|protein phosphatase 2/4/6 regulatory subunit|protein phosphatase 2A, regulatory subunit alpha-4|renal carcinoma antigen NY-REN-16', map_location='Xq13.1-q13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70133449, end_position_on_the_genomic_accession=70166324, orientation='plus', exon_count=7, OMIM=300139, no_hdr0=''),
  134549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134549, CurrentID=0, Status='live', Symbol='SHROOM1', Aliases='APXL2', description='shroom family member 1', other_designations='apical protein 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132822141, end_position_on_the_genomic_accession=132831621, orientation='minus', exon_count=10, OMIM=611179, no_hdr0=''),
  3478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3478, CurrentID=0, Status='live', Symbol='IGES', Aliases='', description='immunoglobulin E concentration, serum', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147061, no_hdr0=''),
  3479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3479, CurrentID=0, Status='live', Symbol='IGF1', Aliases='IGF-I, IGFI, MGF', description='insulin like growth factor 1', other_designations='insulin-like growth factor 1 (somatomedin C)|insulin-like growth factor IB|mechano growth factor|somatomedin-C', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102395867, end_position_on_the_genomic_accession=102481786, orientation='minus', exon_count=7, OMIM=147440, no_hdr0=''),
  3480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3480, CurrentID=0, Status='live', Symbol='IGF1R', Aliases='CD221, IGFIR, IGFR, JTK13', description='insulin like growth factor 1 receptor', other_designations='IGF-I receptor|soluble IGF1R variant 1|soluble IGF1R variant 2', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98648539, end_position_on_the_genomic_accession=98964530, orientation='plus', exon_count=24, OMIM=147370, no_hdr0=''),
  3481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3481, CurrentID=0, Status='live', Symbol='IGF2', Aliases='C11orf43, GRDF, IGF-II, PP9974', description='insulin like growth factor 2', other_designations='T3M-11-derived growth factor|insulin-like growth factor 2 (somatomedin A)|insulin-like growth factor type 2|preptin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2129112, end_position_on_the_genomic_accession=2149603, orientation='minus', exon_count=9, OMIM=147470, no_hdr0=''),
  3482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3482, CurrentID=0, Status='live', Symbol='IGF2R', Aliases='CD222, CIMPR, M6P-R, M6P/IGF2R, MPR 300, MPR1, MPRI', description='insulin like growth factor 2 receptor', other_designations='300 kDa mannose 6-phosphate receptor|CI Man-6-P receptor|IGF-II receptor|M6P/IGF2 receptor|insulin-like growth factor 2 receptor|insulin-like growth factor II receptor', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159969099, end_position_on_the_genomic_accession=160106551, orientation='plus', exon_count=48, OMIM=147280, no_hdr0=''),
  3483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3483, CurrentID=0, Status='live', Symbol='IGFALS', Aliases='ACLSD, ALS', description='insulin like growth factor binding protein acid labile subunit', other_designations='insulin-like growth factor binding protein complex acid labile chain|insulin-like growth factor binding protein, acid labile subunit', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1790413, end_position_on_the_genomic_accession=1794908, orientation='minus', exon_count=4, OMIM=601489, no_hdr0=''),
  3484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3484, CurrentID=0, Status='live', Symbol='IGFBP1', Aliases='AFBP, IBP1, IGF-BP25, PP12, hIGFBP-1', description='insulin like growth factor binding protein 1', other_designations='IBP-1|IGF-binding protein 1|IGFBP-1|alpha-pregnancy-associated endometrial globulin|amniotic fluid binding protein|binding protein-25|binding protein-26|binding protein-28|growth hormone independent-binding protein|insulin-like growth factor binding protein 1|placental protein 12', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45888360, end_position_on_the_genomic_accession=45893668, orientation='plus', exon_count=4, OMIM=146730, no_hdr0=''),
  3485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3485, CurrentID=0, Status='live', Symbol='IGFBP2', Aliases='IBP2, IGF-BP53', description='insulin like growth factor binding protein 2', other_designations='IGF-binding protein 2|insulin-like growth factor binding protein 2, 36kDa', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216633404, end_position_on_the_genomic_accession=216664436, orientation='plus', exon_count=4, OMIM=146731, no_hdr0=''),
  3486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3486, CurrentID=0, Status='live', Symbol='IGFBP3', Aliases='BP-53, IBP3', description='insulin like growth factor binding protein 3', other_designations='IBP-3|IGF-binding protein 3|IGFBP-3|acid stable subunit of the 140 K IGF complex|binding protein 29|binding protein 53|growth hormone-dependent binding protein|insulin-like growth factor binding protein 3', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45912245, end_position_on_the_genomic_accession=45921272, orientation='minus', exon_count=5, OMIM=146732, no_hdr0=''),
  3487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3487, CurrentID=0, Status='live', Symbol='IGFBP4', Aliases='BP-4, HT29-IGFBP, IBP4, IGFBP-4', description='insulin like growth factor binding protein 4', other_designations='IBP-4|IGF-binding protein 4|insulin-like growth factor binding protein 4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40443424, end_position_on_the_genomic_accession=40457730, orientation='plus', exon_count=4, OMIM=146733, no_hdr0=''),
  3488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3488, CurrentID=0, Status='live', Symbol='IGFBP5', Aliases='IBP5', description='insulin like growth factor binding protein 5', other_designations='IBP-5|IGF-binding protein 5|IGFBP-5|insulin-like growth factor binding protein 5', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216672105, end_position_on_the_genomic_accession=216695549, orientation='minus', exon_count=4, OMIM=146734, no_hdr0=''),
  3489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3489, CurrentID=0, Status='live', Symbol='IGFBP6', Aliases='IBP6', description='insulin like growth factor binding protein 6', other_designations='IBP-6|IGF binding protein 6|IGF-binding protein 6|IGFBP-6|insulin-like growth factor binding protein 6', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53097652, end_position_on_the_genomic_accession=53102344, orientation='plus', exon_count=4, OMIM=146735, no_hdr0=''),
  3490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3490, CurrentID=0, Status='live', Symbol='IGFBP7', Aliases='AGM, FSTL2, IBP-7, IGFBP-7, IGFBP-7v, IGFBPRP1, MAC25, PSF, RAMSVPS, TAF', description='insulin like growth factor binding protein 7', other_designations='IGF-binding protein 7|IGFBP-rP1|PGI2-stimulating factor|angiomodulin|insulin-like growth factor binding protein 7|prostacyclin-stimulating factor|tumor-derived adhesion factor', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57031071, end_position_on_the_genomic_accession=57110385, orientation='minus', exon_count=5, OMIM=602867, no_hdr0=''),
  3491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3491, CurrentID=0, Status='live', Symbol='CYR61', Aliases='CCN1, GIG1, IGFBP10', description='cysteine-rich, angiogenic inducer, 61', other_designations='CCN family member 1|IBP-10|IGF-binding protein 10|IGFBP-10|cysteine-rich heparin-binding protein 61|cysteine-rich, anigogenic inducer, 61|insulin-like growth factor-binding protein 10', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85580761, end_position_on_the_genomic_accession=85583967, orientation='plus', exon_count=5, OMIM=602369, no_hdr0=''),
  3492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3492, CurrentID=0, Status='live', Symbol='IGH', Aliases='IGD1.1@, IGH@, IGHD@, IGHDY1, IGHJ, IGHJ@, IGHV, IGHV@, IGH', description='immunoglobulin heavy locus', other_designations='D (diversity) region of heavy chains|J (joining) region of heavy chains|immunglobulin heavy chain variable region|immunoglobulin heavy diversity cluster|immunoglobulin heavy diversity group|immunoglobulin heavy diversity locus|immunoglobulin heavy joining cluster|immunoglobulin heavy joining group|immunoglobulin heavy polypeptide, joining region|immunoglobulin heavy variable cluster|immunoglobulin heavy variable group', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105566277, end_position_on_the_genomic_accession=106879844, orientation='minus', exon_count=0, OMIM='146910,147010,147070', no_hdr0=''),
  3493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3493, CurrentID=0, Status='live', Symbol='IGHA1', Aliases='IgA1', description='immunoglobulin heavy constant alpha 1', other_designations='immunoglobulin alpha 1|immunoglobulin heavy chain constant region alpha 1', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105707168, end_position_on_the_genomic_accession=105708664, orientation='minus', exon_count=0, OMIM=146900, no_hdr0=''),
  3494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3494, CurrentID=0, Status='live', Symbol='IGHA2', Aliases='', description='immunoglobulin heavy constant alpha 2 (A2m marker)', other_designations='constant region of heavy chain of IgA2|immunoglobulin alpha 2 (A2M marker)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105586937, end_position_on_the_genomic_accession=105588394, orientation='minus', exon_count=0, OMIM=147000, no_hdr0=''),
  3495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3495, CurrentID=0, Status='live', Symbol='IGHD', Aliases='', description='immunoglobulin heavy constant delta', other_designations='constant region of heavy chain of IgD', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105838401, end_position_on_the_genomic_accession=105845677, orientation='minus', exon_count=0, OMIM=147170, no_hdr0=''),
  3496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3496, CurrentID=0, Status='live', Symbol='IGHDOR15@', Aliases='IGD2, IGHD/OR15, IGHDY2', description='immunoglobulin heavy diversity orphans on chromosome 15', other_designations='immunoglobulin heavy chain diversity region-2|immunoglobulin heavy diversity orphons on chromosome 15|immunoglobulin heavy polypeptide, diversity region 2', map_location='15q11-q12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146990, no_hdr0=''),
  3497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3497, CurrentID=0, Status='live', Symbol='IGHE', Aliases='IgE', description='immunoglobulin heavy constant epsilon', other_designations='constant region of heavy chain of IgE|immunoglobulin epsilon', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105600066, end_position_on_the_genomic_accession=105601727, orientation='minus', exon_count=0, OMIM=147180, no_hdr0=''),
  3498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3498, CurrentID=0, Status='live', Symbol='IGHEP1', Aliases='', description='immunoglobulin heavy constant epsilon P1 (pseudogene)', other_designations='immunoglobulin epsilon pseudogene 1|immunoglobulin heavy constant epsilon P1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105721794, end_position_on_the_genomic_accession=105722527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3499, CurrentID=0, Status='live', Symbol='IGHEP2', Aliases='', description='immunoglobulin heavy constant epsilon P2 (pseudogene)', other_designations='immunoglobulin epsilon pseudogene 2|immunoglobulin heavy constant epsilon P2 pseudogene', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5113452, end_position_on_the_genomic_accession=5114909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3500, CurrentID=0, Status='live', Symbol='IGHG1', Aliases='', description='immunoglobulin heavy constant gamma 1 (G1m marker)', other_designations='constant region of heavy chain of IgG1|immunoglobulin gamma 1 (Gm marker)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105741473, end_position_on_the_genomic_accession=105743070, orientation='minus', exon_count=0, OMIM=147100, no_hdr0=''),
  3501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3501, CurrentID=0, Status='live', Symbol='IGHG2', Aliases='', description='immunoglobulin heavy constant gamma 2 (G2m marker)', other_designations='Constant region of heavy chain of IgG2|immunoglobulin gamma 2 (Gm marker)|immunoglobulin heavy constant gamma 2 (Gm marker)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105643203, end_position_on_the_genomic_accession=105644789, orientation='minus', exon_count=0, OMIM=147110, no_hdr0=''),
  3502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3502, CurrentID=0, Status='live', Symbol='IGHG3', Aliases='IgG3', description='immunoglobulin heavy constant gamma 3 (G3m marker)', other_designations='C gamma 3|constant region of heavy chain of IgG3|immunoglobulin gamma 3 (Gm marker)|immunoglobulin heavy constant gamma 3 (Gm marker)|secrete-type Ig gamma heavy-chain', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105765914, end_position_on_the_genomic_accession=105771405, orientation='minus', exon_count=0, OMIM=147120, no_hdr0=''),
  3503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3503, CurrentID=0, Status='live', Symbol='IGHG4', Aliases='', description='immunoglobulin heavy constant gamma 4 (G4m marker)', other_designations='immunoglobulin heavy chain constant region gamma 4', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105624476, end_position_on_the_genomic_accession=105626065, orientation='minus', exon_count=0, OMIM=147130, no_hdr0=''),
  101060016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060016, CurrentID=0, Status='live', Symbol='LOC101060016', Aliases='', description='signal peptidase complex subunit 3-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226331985, end_position_on_the_genomic_accession=226333655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3505, CurrentID=0, Status='live', Symbol='IGHGP', Aliases='IGHGP1', description='immunoglobulin heavy constant gamma P (non-functional)', other_designations='immunoglobulin gamma pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105668249, end_position_on_the_genomic_accession=105670040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3507, CurrentID=0, Status='live', Symbol='IGHM', Aliases='AGM1, MU, VH', description='immunoglobulin heavy constant mu', other_designations='constant region of heavy chain of IgM|immunoglobulin mu chain', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105851966, end_position_on_the_genomic_accession=105856217, orientation='minus', exon_count=0, OMIM=147020, no_hdr0=''),
  3508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3508, CurrentID=0, Status='live', Symbol='IGHMBP2', Aliases='CATF1, CMT2S, HCSA, HMN6, SMARD1, SMUBP2, ZFAND7', description='immunoglobulin mu binding protein 2', other_designations='ATP-dependent helicase IGHMBP2|GF-1|cardiac transcription factor 1|glial factor 1|immunoglobulin mu-binding protein 2|zinc finger, AN1-type domain 7', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68903851, end_position_on_the_genomic_accession=68940601, orientation='plus', exon_count=18, OMIM=600502, no_hdr0=''),
  101060021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060021, CurrentID=0, Status='live', Symbol='LOC101060021', Aliases='', description='ADP-ribosylation factor 1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58364394, end_position_on_the_genomic_accession=58395140, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060022, CurrentID=0, Status='live', Symbol='LOC101060022', Aliases='', description='nascent polypeptide-associated complex subunit alpha, muscle-specific form-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227969793, end_position_on_the_genomic_accession=227971378, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3512, CurrentID=0, Status='live', Symbol='JCHAIN', Aliases='IGCJ, IGJ, JCH', description='joining chain of multimeric IgA and IgM', other_designations='IgJ chain|J chain|immunoglobulin J polypeptide, linker protein for immunoglobulin alpha and mu polypeptides', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70655541, end_position_on_the_genomic_accession=70676901, orientation='minus', exon_count=8, OMIM=147790, no_hdr0=''),
  3513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3513, CurrentID=0, Status='live', Symbol='IGJP1', Aliases='', description='immunoglobulin J polypeptide pseudogene 1', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84979577, end_position_on_the_genomic_accession=84980592, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3514, CurrentID=0, Status='live', Symbol='IGKC', Aliases='HCAK1D, Km, IGKC', description='immunoglobulin kappa constant', other_designations='immunoglobulin kappa (invariant region)|immunoglobulin kappa constant region|immunoglobulin kappa light chain (VJ)', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88857361, end_position_on_the_genomic_accession=88857683, orientation='minus', exon_count=0, OMIM=147200, no_hdr0=''),
  3515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3515, CurrentID=0, Status='live', Symbol='IGKDEL', Aliases='IGKDE', description='immunoglobulin kappa deleting element or like', other_designations='immunoglobulin kappa polypeptide deleting element', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88832222, end_position_on_the_genomic_accession=88833331, orientation='minus', exon_count=0, OMIM=146780, no_hdr0=''),
  3516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3516, CurrentID=0, Status='live', Symbol='RBPJ', Aliases='AOS3, CBF1, IGKJRB, IGKJRB1, KBF2, RBP-JK, RBPSUH, SUH, csl, RBPJ', description='recombination signal binding protein for immunoglobulin kappa J region', other_designations='CBF-1|H-2K binding factor-2|RBP-J kappa|RBP-JK|immunoglobulin kappa J region recombination signal binding protein 1|renal carcinoma antigen NY-REN-30|suppressor of hairless homolog', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26163494, end_position_on_the_genomic_accession=26435131, orientation='plus', exon_count=16, OMIM=147183, no_hdr0=''),
  3517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3517, CurrentID=0, Status='live', Symbol='RBPJP2', Aliases='IGKJRBP2, RBPSUHP2', description='RBPJ pseudogene 2', other_designations='immunoglobulin kappa J region recombination signal binding protein pseudogene 2|recombination signal binding protein for immunoglobulin kappa J region pseudogene|recombining binding protein suppressor of hairless pseudogene 2', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66635579, end_position_on_the_genomic_accession=66637852, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3519, CurrentID=0, Status='live', Symbol='IGKV@', Aliases='IGKV, IGKV1, IGKV1@, IGKV2, IGKV2@, IGKV3, IGKV3@', description='immunoglobulin kappa variable cluster', other_designations='immunoglobulin kappa variable group|immunoglobulin kappa variable region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146980, no_hdr0=''),
  101060032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060032, CurrentID=0, Status='live', Symbol='RHEBP2', Aliases='', description='Ras-homolog enriched in brain pseudogene 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47705812, end_position_on_the_genomic_accession=47707174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284576, CurrentID=0, Status='live', Symbol='LEMD1-AS1', Aliases='', description='LEMD1 antisense RNA 1', other_designations='LEMD1 antisense RNA 1 (non-protein coding)', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205373252, end_position_on_the_genomic_accession=205387440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3523, CurrentID=0, Status='live', Symbol='IGKV3OR2-268', Aliases='IGKV268A, IGKV3OR2268, IGKV3OR2268A, IGKV3OR2-268', description='immunoglobulin kappa variable 3/OR2-268 (non-functional)', other_designations='IGKV3/OR2-268|IGKV3/OR2-268A|immunoglobulin kappa variable 3/OR2-268a|immunoglobulin kappa variable region 268', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3525, CurrentID=0, Status='live', Symbol='IGKV1OR1-1', Aliases='IGKV1/OR1-1, IGKV1OR11, IGKVP1', description='immunoglobulin kappa variable 1/OR1-1 (pseudogene)', other_designations='immunoglobulin kappa variable region pseudogene 1', map_location='1p13-q12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060038, CurrentID=0, Status='live', Symbol='LOC101060038', Aliases='', description='uncharacterized LOC101060038', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10552661, end_position_on_the_genomic_accession=10575720, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060039, CurrentID=0, Status='live', Symbol='HNRNPA1P57', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 57', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41156565, end_position_on_the_genomic_accession=41157615, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060040, CurrentID=0, Status='live', Symbol='CHCHD2P11', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 11', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42181283, end_position_on_the_genomic_accession=42182318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3529, CurrentID=0, Status='live', Symbol='IGKV2OR22-3', Aliases='IGKV2OR223, IGKVP4', description='immunoglobulin kappa variable 2/OR22-3 (pseudogene)', other_designations='IGKV2/OR22-3|immunoglobulin kappa variable 2/OR22-3 pseudogene|immunoglobulin kappa variable region pseudogene 4', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16921324, end_position_on_the_genomic_accession=16921937, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3530, CurrentID=0, Status='live', Symbol='IGKV1OR22-1', Aliases='IGKV1OR221, IGKVP5', description='immunoglobulin kappa variable 1/OR22-1 (pseudogene)', other_designations='IGKV1/OR22-1|immunoglobulin kappa variable 1/OR22-1 pseudogene|immunoglobulin kappa variable region pseudogene 5', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16932727, end_position_on_the_genomic_accession=16934653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3531, CurrentID=0, Status='live', Symbol='IGKV1OR2-1', Aliases='IGKV1/OR-1, IGKV1/OR2-1, IGKV1OR-1, IGKV1OR1, IGKVPZ1', description='immunoglobulin kappa variable 1/OR2-1 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR-1 (pseudogene)|immunoglobulin kappa variable 1/OR-1 pseudogene|immunoglobulin kappa variable region pseudogene Z1', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3532, CurrentID=0, Status='live', Symbol='IGKV1OR-2', Aliases='IGKV1/OR-2, IGKV1OR2, IGKVPZ2', description='immunoglobulin kappa variable 1/OR-2 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR-2 pseudogene|immunoglobulin kappa variable region pseudogene Z2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64765056, end_position_on_the_genomic_accession=64765539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060045, CurrentID=0, Status='live', Symbol='PCBP2P1', Aliases='', description='poly(rC) binding protein 2 pseudogene 1', other_designations='poly(rC)-binding protein 2-like', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39170282, end_position_on_the_genomic_accession=39172368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3534, CurrentID=0, Status='live', Symbol='IGKV1OR-4', Aliases='IGKV1/OR-4, IGKV1OR4, IGKVPZ4', description='immunoglobulin kappa variable 1/OR-4 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR-4 pseudogene|immunoglobulin kappa variable region pseudogene Z4', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3535, CurrentID=0, Status='live', Symbol='IGL', Aliases='IGL@C6, IGL', description='immunoglobulin lambda locus', other_designations='immunoglobulin lambda gene cluster', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22026076, end_position_on_the_genomic_accession=22922913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060048, CurrentID=0, Status='live', Symbol='LOC101060048', Aliases='', description='uncharacterized LOC101060048', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32810505, end_position_on_the_genomic_accession=32812001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3537, CurrentID=0, Status='live', Symbol='IGLC1', Aliases='IGLC', description='immunoglobulin lambda constant 1 (Mcg marker)', other_designations='C1 segment|immunoglobulin lambda constant region 1 (Mcg marker)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22895375, end_position_on_the_genomic_accession=22895694, orientation='plus', exon_count=0, OMIM=147220, no_hdr0=''),
  3538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3538, CurrentID=0, Status='live', Symbol='IGLC2', Aliases='IGLC', description='immunoglobulin lambda constant 2 (Kern-Oz- marker)', other_designations='C2 segment|Ig light-chain, partial Ke-Oz- polypeptide, C-term|immunoglobulin lambda constant region 2 (Kern- Oz- marker)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22900976, end_position_on_the_genomic_accession=22901295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3539, CurrentID=0, Status='live', Symbol='IGLC3', Aliases='IGLC', description='immunoglobulin lambda constant 3 (Kern-Oz+ marker)', other_designations='C3 segment|Ig lambda-3 chain C regions|immunoglobulin lambda constant region 3 (Kern- Oz+ marker)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22906342, end_position_on_the_genomic_accession=22906661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3540, CurrentID=0, Status='live', Symbol='IGLC4', Aliases='IGLC', description='immunoglobulin lambda constant 4 (pseudogene)', other_designations='immunoglobulin lambda constant 4 pseudogene|immunoglobulin lambda constant region 4 (pseudogene)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22910828, end_position_on_the_genomic_accession=22911080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3541, CurrentID=0, Status='live', Symbol='IGLC5', Aliases='IGLC', description='immunoglobulin lambda constant 5 (pseudogene)', other_designations='immunoglobulin lambda constant 5 pseudogene|immunoglobulin lambda constant region 5 (pseudogene)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22915624, end_position_on_the_genomic_accession=22915932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3542, CurrentID=0, Status='live', Symbol='IGLC6', Aliases='IGLC', description='immunoglobulin lambda constant 6 (Kern+Oz- marker, gene/pseudogene)', other_designations='Ig lambda-6 chain C region|immunoglobulin lambda constant 6 (Kern+Oz- marker)|immunoglobulin lambda constant region 6 (Kern+ Oz- marker)', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22919535, end_position_on_the_genomic_accession=22919858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3543, CurrentID=0, Status='live', Symbol='IGLL1', Aliases='14.1, AGM2, CD179b, IGL1, IGL5, IGLJ14.1, IGLL, IGO, IGVPB, VPREB2', description='immunoglobulin lambda-like polypeptide 1', other_designations='CD179 antigen-like family member B|CD179b antigen|Pre-B lymphocyte-specific protein-2|ig lambda-5|immunoglobulin omega polypeptide chain|immunoglobulin-related 14.1 protein|immunoglobulin-related protein 14.1|lambda5', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23573125, end_position_on_the_genomic_accession=23580472, orientation='minus', exon_count=3, OMIM=146770, no_hdr0=''),
  3544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3544, CurrentID=0, Status='live', Symbol='IGLL2P', Aliases='16.2, FLAMBDA1, IGLL1P, IGLL2', description='immunoglobulin lambda-like polypeptide 2, pseudogene', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3546, CurrentID=0, Status='live', Symbol='IGLV@', Aliases='IGLV', description='immunoglobulin lambda variable cluster', other_designations='Variable region of lambda light chains|immunoglobulin lambda variable group|immunoglobulin lambda variable region gene cluster', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147240, no_hdr0=''),
  3547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3547, CurrentID=0, Status='live', Symbol='IGSF1', Aliases='CHTE, IGCD1, IGDC1, INHBP, PGSF2, p120', description='immunoglobulin superfamily member 1', other_designations='immunoglobulin superfamily, member 1|immunoglobulin-like domain-containing protein 1|inhibin-binding protein|pituitary gland-specific factor 2', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131273506, end_position_on_the_genomic_accession=131289457, orientation='minus', exon_count=23, OMIM=300137, no_hdr0=''),
  3548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3548, CurrentID=0, Status='live', Symbol='IHG1', Aliases='IHG', description='iris hypoplasia with glaucoma 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=308500, no_hdr0=''),
  3549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3549, CurrentID=0, Status='live', Symbol='IHH', Aliases='BDA1, HHG2', description='indian hedgehog', other_designations='Indian hedgehog homolog', map_location='2q33-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219054420, end_position_on_the_genomic_accession=219060516, orientation='minus', exon_count=3, OMIM=600726, no_hdr0=''),
  3550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3550, CurrentID=0, Status='live', Symbol='IK', Aliases='CSA2, RED', description='IK cytokine, down-regulator of HLA II', other_designations='IK factor|RD element|chondrosarcoma-associated protein 2|cytokine IK|prer protein', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140647799, end_position_on_the_genomic_accession=140662480, orientation='plus', exon_count=20, OMIM=600549, no_hdr0=''),
  3551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3551, CurrentID=0, Status='live', Symbol='IKBKB', Aliases='IKK-beta, IKK2, IKKB, IMD15, NFKBIKB', description='inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase beta', other_designations='I-kappa-B kinase 2|I-kappa-B-kinase beta|IKK-B|nuclear factor NF-kappa-B inhibitor kinase beta', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42271302, end_position_on_the_genomic_accession=42332653, orientation='plus', exon_count=24, OMIM=603258, no_hdr0=''),
  3552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3552, CurrentID=0, Status='live', Symbol='IL1A', Aliases='IL-1A, IL1, IL1-ALPHA, IL1F1', description='interleukin 1 alpha', other_designations='IL-1 alpha|hematopoietin-1|interleukin 1, alpha|preinterleukin 1 alpha|pro-interleukin-1-alpha', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112773915, end_position_on_the_genomic_accession=112785394, orientation='minus', exon_count=7, OMIM=147760, no_hdr0=''),
  3553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3553, CurrentID=0, Status='live', Symbol='IL1B', Aliases='IL-1, IL1-BETA, IL1F2', description='interleukin 1 beta', other_designations='IL-1 beta|catabolin|interleukin 1, beta|preinterleukin 1 beta|pro-interleukin-1-beta', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112829751, end_position_on_the_genomic_accession=112836779, orientation='minus', exon_count=7, OMIM=147720, no_hdr0=''),
  3554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3554, CurrentID=0, Status='live', Symbol='IL1R1', Aliases='CD121A, D2S1473, IL-1R-alpha, IL1R, IL1RA, P80', description='interleukin 1 receptor, type I', other_designations='CD121 antigen-like family member A|IL-1R-1|IL-1RT-1|IL-1RT1|interleukin 1 receptor alpha, type I|interleukin-1 receptor alpha|interleukin-1 receptor type I', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102069638, end_position_on_the_genomic_accession=102179874, orientation='plus', exon_count=23, OMIM=147810, no_hdr0=''),
  101060067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060067, CurrentID=0, Status='live', Symbol='LOC101060067', Aliases='', description='40S ribosomal protein S3 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73152410, end_position_on_the_genomic_accession=73153736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3556, CurrentID=0, Status='live', Symbol='IL1RAP', Aliases='C3orf13, IL-1RAcP, IL1R3', description='interleukin 1 receptor accessory protein', other_designations='IL-1 receptor accessory protein|IL-1R3|interleukin-1 receptor 3|interleukin-1 receptor accessory protein beta', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190514051, end_position_on_the_genomic_accession=190657197, orientation='plus', exon_count=15, OMIM=602626, no_hdr0=''),
  3557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3557, CurrentID=0, Status='live', Symbol='IL1RN', Aliases='DIRA, ICIL-1RA, IL-1RN, IL-1ra, IL-1ra3, IL1F3, IL1RA, IRAP, MVCD4', description='interleukin 1 receptor antagonist', other_designations='IL1 inhibitor|IL1RN (IL1F3)|intracellular IL-1 receptor antagonist type II|intracellular interleukin-1 receptor antagonist (icIL-1ra)|type II interleukin-1 receptor antagonist', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113099365, end_position_on_the_genomic_accession=113134016, orientation='plus', exon_count=11, OMIM=147679, no_hdr0=''),
  3558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3558, CurrentID=0, Status='live', Symbol='IL2', Aliases='IL-2, TCGF, lymphokine', description='interleukin 2', other_designations='T cell growth factor|aldesleukin|involved in regulation of T-cell clonal expansion', map_location='4q26-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122451470, end_position_on_the_genomic_accession=122456495, orientation='minus', exon_count=4, OMIM=147680, no_hdr0=''),
  3559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3559, CurrentID=0, Status='live', Symbol='IL2RA', Aliases='CD25, IDDM10, IL2R, TCGFR, p55', description='interleukin 2 receptor subunit alpha', other_designations='IL-2 receptor subunit alpha|IL-2R subunit alpha|TAC antigen|interleukin 2 receptor, alpha', map_location='10p15-p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6010694, end_position_on_the_genomic_accession=6062370, orientation='minus', exon_count=8, OMIM=147730, no_hdr0=''),
  3560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3560, CurrentID=0, Status='live', Symbol='IL2RB', Aliases='CD122, IL15RB, P70-75', description='interleukin 2 receptor subunit beta', other_designations='CD122 antigen|IL-2 receptor subunit beta|IL-2R subunit beta|IL-2RB|high affinity IL-2 receptor beta subunit|high affinity IL-2 receptor subunit beta|interleukin 15 receptor, beta|interleukin 2 receptor, beta|p75', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37125838, end_position_on_the_genomic_accession=37149922, orientation='minus', exon_count=10, OMIM=146710, no_hdr0=''),
  3561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3561, CurrentID=0, Status='live', Symbol='IL2RG', Aliases='CD132, CIDX, IL-2RG, IMD4, P64, SCIDX, SCIDX1', description='interleukin 2 receptor subunit gamma', other_designations='CD132 antigen|IL-2 receptor subunit gamma|IL-2R subunit gamma|common cytokine receptor gamma chain|gammaC', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71107404, end_position_on_the_genomic_accession=71111631, orientation='minus', exon_count=8, OMIM=308380, no_hdr0=''),
  3562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3562, CurrentID=0, Status='live', Symbol='IL3', Aliases='IL-3, MCGF, MULTI-CSF', description='interleukin 3', other_designations='P-cell stimulating factor|P-cell-stimulating factor|colony-stimulating factor, multiple|hematopoietic growth factor|mast cell growth factor|mast-cell growth factor|multilineage-colony-stimulating factor|multipotential colony-stimulating factor', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132060654, end_position_on_the_genomic_accession=132063203, orientation='plus', exon_count=5, OMIM=147740, no_hdr0=''),
  3563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3563, CurrentID=0, Status='live', Symbol='IL3RA', Aliases='CD123, IL3RY, IL3RX, IL3RY, hIL-3Ra, IL3RA', description='interleukin 3 receptor subunit alpha', other_designations='CD123 antigen|IL-3 receptor alpha SP2 isoform|IL-3 receptor subunit alpha|IL-3R subunit alpha|IL-3R-alpha|IL-3RA|interleukin 3 receptor, alpha (low affinity)', map_location='Xp22.3 or Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1336587, end_position_on_the_genomic_accession=1382689, orientation='plus', exon_count=12, OMIM='308385,430000', no_hdr0=''),
  101060076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060076, CurrentID=0, Status='live', Symbol='LOC101060076', Aliases='', description='60S ribosomal protein L14 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85210038, end_position_on_the_genomic_accession=85216823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3565, CurrentID=0, Status='live', Symbol='IL4', Aliases='BCGF-1, BCGF1, BSF-1, BSF1, IL-4', description='interleukin 4', other_designations='B cell growth factor 1|B_cell stimulatory factor 1|binetrakin|interleukin 4 variant 2|lymphocyte stimulatory factor 1|pitrakinra', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132673986, end_position_on_the_genomic_accession=132682678, orientation='plus', exon_count=4, OMIM=147780, no_hdr0=''),
  3566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3566, CurrentID=0, Status='live', Symbol='IL4R', Aliases='CD124, IL-4RAA, IL4R', description='interleukin 4 receptor', other_designations='IL-4 receptor subunit alpha|IL4R nirs variant 1|interleukin-4 receptor alpha chain', map_location='16p12.1-p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27313668, end_position_on_the_genomic_accession=27364778, orientation='plus', exon_count=16, OMIM=147781, no_hdr0=''),
  3567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3567, CurrentID=0, Status='live', Symbol='IL5', Aliases='EDF, IL-5, TRF', description='interleukin 5', other_designations='B-cell differentiation factor I|T-cell replacing factor|colony-stimulating factor, eosinophil|eosinophil differentiation factor', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132541444, end_position_on_the_genomic_accession=132556890, orientation='minus', exon_count=6, OMIM=147850, no_hdr0=''),
  3568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3568, CurrentID=0, Status='live', Symbol='IL5RA', Aliases='CD125, CDw125, HSIL5R3, IL5R', description='interleukin 5 receptor subunit alpha', other_designations='CD125 antigen|IL-5 receptor subunit alpha|IL-5R subunit alpha|interleukin 5 receptor type 3|interleukin 5 receptor, alpha|interleukin-5 receptor alpha chain', map_location='3p26-p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3066324, end_position_on_the_genomic_accession=3110374, orientation='minus', exon_count=15, OMIM=147851, no_hdr0=''),
  3569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3569, CurrentID=0, Status='live', Symbol='IL6', Aliases='BSF-2, BSF2, CDF, HGF, HSF, IFN-beta-2, IFNB2, IL-6', description='interleukin 6', other_designations='B-cell differentiation factor|B-cell stimulatory factor 2|CTL differentiation factor|hybridoma growth factor|interferon beta-2|interleukin BSF-2', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22725442, end_position_on_the_genomic_accession=22732002, orientation='plus', exon_count=6, OMIM=147620, no_hdr0=''),
  3570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3570, CurrentID=0, Status='live', Symbol='IL6R', Aliases='CD126, IL-6R-1, IL-6RA, IL6QA, IL6RQ, gp80, IL6R', description='interleukin 6 receptor', other_designations='CD126 antigen|IL-6 receptor subunit alpha|IL-6R 1|membrane glycoprotein 80', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154405193, end_position_on_the_genomic_accession=154469450, orientation='plus', exon_count=12, OMIM=147880, no_hdr0=''),
  3572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3572, CurrentID=0, Status='live', Symbol='IL6ST', Aliases='CD130, CDW130, GP130, IL-6RB', description='interleukin 6 signal transducer', other_designations='CD130 antigen|IL-6 receptor subunit beta|IL-6R subunit beta|gp130 of the rheumatoid arthritis antigenic peptide-bearing soluble form|gp130, oncostatin M receptor|interleukin receptor beta chain|membrane glycoprotein 130|membrane glycoprotein gp130|oncostatin-M receptor subunit alpha', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55935095, end_position_on_the_genomic_accession=55994993, orientation='minus', exon_count=19, OMIM=600694, no_hdr0=''),
  3573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3573, CurrentID=0, Status='live', Symbol='IL6STP1', Aliases='IL6STP', description='interleukin 6 signal transducer (gp130, oncostatin M receptor) pseudogene 1', other_designations='', map_location='17p11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15782922, end_position_on_the_genomic_accession=15791633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3574, CurrentID=0, Status='live', Symbol='IL7', Aliases='IL-7', description='interleukin 7', other_designations='', map_location='8q12-q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78675872, end_position_on_the_genomic_accession=78805523, orientation='minus', exon_count=9, OMIM=146660, no_hdr0=''),
  3575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3575, CurrentID=0, Status='live', Symbol='IL7R', Aliases='CD127, CDW127, IL-7R-alphaA, ILRA, IL7R', description='interleukin 7 receptor', other_designations='CD127 antigen|IL-7 receptor subunit alpha|IL-7R subunit alpha|interleukin 7 receptor alpha chain|interleukin 7 receptor isoform H5-6', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35856875, end_position_on_the_genomic_accession=35879603, orientation='plus', exon_count=8, OMIM=146661, no_hdr0=''),
  3576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3576, CurrentID=0, Status='live', Symbol='CXCL8', Aliases='GCP-1, GCP1, IL8, LECT, LUCT, LYNAP, MDNCF, MONAP, NAF, NAP-1, NAP1', description='chemokine (C-X-C motif) ligand 8', other_designations='T-cell chemotactic factor|alveolar macrophage chemotactic factor I|beta endothelial cell-derived neutrophil activating peptide|beta-thromboglobulin-like protein|emoctakin|granulocyte chemotactic protein 1|interleukin 8|lung giant cell carcinoma-derived chemotactic protein|lymphocyte derived neutrophil activating peptide|monocyte-derived neutrophil chemotactic factor|monocyte-derived neutrophil-activating peptide|neutrophil-activating peptide 1|small inducible cytokine subfamily B, member 8|tumor necrosis factor-induced gene 1', map_location='4q13-q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73740506, end_position_on_the_genomic_accession=73743716, orientation='plus', exon_count=4, OMIM=146930, no_hdr0=''),
  3577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3577, CurrentID=0, Status='live', Symbol='CXCR1', Aliases='C-C, C-C-CKR-1, CD128, CD181, CDw128a, CKR-1, CMKAR1, IL8R1, IL8RA, IL8RBA', description='chemokine (C-X-C motif) receptor 1', other_designations='CXC-R1|CXCR-1|IL-8 receptor type 1|IL-8R A|high affinity interleukin-8 receptor A|interleukin 8 receptor, alpha|interleukin-8 receptor type 1|interleukin-8 receptor type A', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218162845, end_position_on_the_genomic_accession=218166993, orientation='minus', exon_count=2, OMIM=146929, no_hdr0=''),
  3578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3578, CurrentID=0, Status='live', Symbol='IL9', Aliases='HP40, IL-9, P40', description='interleukin 9', other_designations='T-cell growth factor p40|cytokine P40|homolog of mouse T cell and mast cell growth factor 40|p40 T-cell and mast cell growth factor|p40 cytokine', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135892246, end_position_on_the_genomic_accession=135895827, orientation='minus', exon_count=5, OMIM=146931, no_hdr0=''),
  3579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3579, CurrentID=0, Status='live', Symbol='CXCR2', Aliases='CD182, CDw128b, CMKAR2, IL8R2, IL8RA, IL8RB', description='chemokine (C-X-C motif) receptor 2', other_designations='CXC-R2|CXCR-2|CXCR2 gene for IL8 receptor type B|GRO/MGSA receptor|IL-8 receptor type 2|IL-8R B|chemokine (CXC) receptor 2|high affinity interleukin-8 receptor B|interleukin 8 receptor B|interleukin 8 receptor type 2|interleukin 8 receptor, beta|interleukin-8 receptor type B', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218125290, end_position_on_the_genomic_accession=218137253, orientation='plus', exon_count=4, OMIM=146928, no_hdr0=''),
  3580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3580, CurrentID=0, Status='live', Symbol='CXCR2P1', Aliases='CXCR2P, IL8RBP', description='chemokine (C-X-C motif) receptor 2 pseudogene 1', other_designations='interleukin 8 receptor (IL8RP) pseudogene|interleukin 8 receptor, beta pseudogene|pseudogene for the low affinity IL-8 receptor', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218059155, end_position_on_the_genomic_accession=218061290, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3581, CurrentID=0, Status='live', Symbol='IL9R', Aliases='CD129, IL-9R', description='interleukin 9 receptor', other_designations='IL-9 receptor', map_location='Xq28 and Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57183856, end_position_on_the_genomic_accession=57197869, orientation='plus', exon_count=11, OMIM=300007, no_hdr0=''),
  131669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131669, CurrentID=0, Status='live', Symbol='UROC1', Aliases='HMFN0320, UROCD', description='urocanate hydratase 1', other_designations='imidazolonepropionate hydrolase|urocanase 1|urocanase domain containing 1', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126481165, end_position_on_the_genomic_accession=126517773, orientation='minus', exon_count=21, OMIM=613012, no_hdr0=''),
  101060097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060097, CurrentID=0, Status='live', Symbol='LOC101060097', Aliases='', description='uncharacterized LOC101060097', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32082480, end_position_on_the_genomic_accession=32085096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3586, CurrentID=0, Status='live', Symbol='IL10', Aliases='CSIF, GVHDS, IL-10A, TGIF, IL10', description='interleukin 10', other_designations='T-cell growth inhibitory factor|cytokine synthesis inhibitory factor', map_location='1q31-q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206767602, end_position_on_the_genomic_accession=206774607, orientation='minus', exon_count=6, OMIM=124092, no_hdr0=''),
  3587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3587, CurrentID=0, Status='live', Symbol='IL10RA', Aliases='CD210, CD210a, CDW210A, HIL-10R, IL-10R1, IL10R', description='interleukin 10 receptor subunit alpha', other_designations='IL-10 receptor subunit alpha|IL-10R subunit 1|IL-10R subunit alpha|IL-10RA|interleukin 10 receptor, alpha|interleukin-10 receptor alpha chain|interleukin-10 receptor subunit 1', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117986391, end_position_on_the_genomic_accession=118001483, orientation='plus', exon_count=8, OMIM=146933, no_hdr0=''),
  3588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3588, CurrentID=0, Status='live', Symbol='IL10RB', Aliases='CDW210B, CRF2-4, CRFB4, D21S58, D21S66, IL-10R2', description='interleukin 10 receptor subunit beta', other_designations='IL-10 receptor subunit beta|IL-10R subunit 2|IL-10R subunit beta|IL-10RB|cytokine receptor class-II CRF2-4|cytokine receptor class-II member 4|cytokine receptor family 2 member 4|cytokine receptor family II, member 4|interleukin 10 receptor, beta|interleukin-10 receptor subunit 2', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33266360, end_position_on_the_genomic_accession=33297234, orientation='plus', exon_count=7, OMIM=123889, no_hdr0=''),
  3589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3589, CurrentID=0, Status='live', Symbol='IL11', Aliases='AGIF, IL-11', description='interleukin 11', other_designations='adipogenesis inhibitory factor|oprelvekin', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55364382, end_position_on_the_genomic_accession=55370463, orientation='minus', exon_count=5, OMIM=147681, no_hdr0=''),
  3590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3590, CurrentID=0, Status='live', Symbol='IL11RA', Aliases='CRSDA', description='interleukin 11 receptor subunit alpha', other_designations='IL-11 receptor subunit alpha|IL-11R subunit alpha|interleukin 11 receptor, alpha|interleukin-11 receptor alpha chain', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34652185, end_position_on_the_genomic_accession=34661902, orientation='plus', exon_count=14, OMIM=600939, no_hdr0=''),
  3592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3592, CurrentID=0, Status='live', Symbol='IL12A', Aliases='CLMF, IL-12A, NFSK, NKSF1, P35', description='interleukin 12A', other_designations='CLMF p35|IL-12, subunit p35|IL35 subunit|NF cell stimulatory factor chain 1|NK cell stimulatory factor chain 1|cytotoxic lymphocyte maturation factor 1, p35|cytotoxic lymphocyte maturation factor 35 kDa subunit|interleukin 12, p35|interleukin 12A (natural killer cell stimulatory factor 1, cytotoxic lymphocyte maturation factor 1, p35)|interleukin-12 alpha chain', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159988836, end_position_on_the_genomic_accession=159996019, orientation='plus', exon_count=7, OMIM=161560, no_hdr0=''),
  3593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3593, CurrentID=0, Status='live', Symbol='IL12B', Aliases='CLMF, CLMF2, IL-12B, IMD28, IMD29, NKSF, NKSF2', description='interleukin 12B', other_designations='CLMF p40|IL-12 subunit p40|IL12, subunit p40|NK cell stimulatory factor chain 2|cytotoxic lymphocyte maturation factor 40 kDa subunit|interleukin 12, p40|interleukin 12B (natural killer cell stimulatory factor 2, cytotoxic lymphocyte maturation factor 2, p40)|interleukin-12 beta chain|natural killer cell stimulatory factor, 40 kD subunit', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159314783, end_position_on_the_genomic_accession=159330473, orientation='minus', exon_count=8, OMIM=161561, no_hdr0=''),
  3594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3594, CurrentID=0, Status='live', Symbol='IL12RB1', Aliases='CD212, IL-12R-BETA1, IL12RB, IMD30', description='interleukin 12 receptor subunit beta 1', other_designations='IL-12 receptor beta component|IL-12 receptor subunit beta-1|IL-12R subunit beta-1|cluster of differentiation 212|interleukin 12 receptor, beta 1|interleukin-12 receptor beta-1 chain', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18058995, end_position_on_the_genomic_accession=18098986, orientation='minus', exon_count=21, OMIM=601604, no_hdr0=''),
  3595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3595, CurrentID=0, Status='live', Symbol='IL12RB2', Aliases='', description='interleukin 12 receptor subunit beta 2', other_designations='IL-12 receptor subunit beta-2|IL-12R subunit beta-2|interleukin-12 receptor beta-2 chain', map_location='1p31.3-p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67307351, end_position_on_the_genomic_accession=67396900, orientation='plus', exon_count=19, OMIM=601642, no_hdr0=''),
  3596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3596, CurrentID=0, Status='live', Symbol='IL13', Aliases='IL-13, P600', description='interleukin 13', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132658173, end_position_on_the_genomic_accession=132661109, orientation='plus', exon_count=4, OMIM=147683, no_hdr0=''),
  3597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3597, CurrentID=0, Status='live', Symbol='IL13RA1', Aliases='CD213A1, CT19, IL-13Ra, NR4', description='interleukin 13 receptor subunit alpha 1', other_designations='CD213a1 antigen|IL-13 receptor subunit alpha-1|IL-13R subunit alpha-1|IL13 receptor alpha-1 chain|cancer/testis antigen 19|interleukin 13 receptor, alpha 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118726954, end_position_on_the_genomic_accession=118794533, orientation='plus', exon_count=13, OMIM=300119, no_hdr0=''),
  3598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3598, CurrentID=0, Status='live', Symbol='IL13RA2', Aliases='CD213A2, CT19, IL-13R, IL13BP', description='interleukin 13 receptor subunit alpha 2', other_designations='IL-13 receptor subunit alpha-2|IL-13R subunit alpha-2|IL-13R-alpha-2|IL-13RA2|cancer/testis antigen 19|interleukin 13 binding protein|interleukin 13 receptor alpha 2 chain|interleukin 13 receptor, alpha 2|interleukin-13-binding protein', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115003975, end_position_on_the_genomic_accession=115017644, orientation='minus', exon_count=10, OMIM=300130, no_hdr0=''),
  101060111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060111, CurrentID=0, Status='live', Symbol='LUZP4P1', Aliases='', description='leucine zipper protein 4 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=21536681, end_position_on_the_genomic_accession=21537001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3600, CurrentID=0, Status='live', Symbol='IL15', Aliases='IL-15', description='interleukin 15', other_designations='', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=141636596, end_position_on_the_genomic_accession=141733987, orientation='plus', exon_count=10, OMIM=600554, no_hdr0=''),
  3601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3601, CurrentID=0, Status='live', Symbol='IL15RA', Aliases='CD215', description='interleukin 15 receptor subunit alpha', other_designations='interleukin 15 receptor alpha isoform EM2|interleukin 15 receptor alpha isoform IC2|interleukin 15 receptor alpha isoform IC3|interleukin 15 receptor alpha isoform IC4|interleukin 15 receptor alpha isoform IC5|interleukin 15 receptor alpha isoform IC6|interleukin 15 receptor alpha isoform IC7|interleukin 15 receptor alpha isoform IC8|interleukin 15 receptor, alpha', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5948900, end_position_on_the_genomic_accession=5978634, orientation='minus', exon_count=11, OMIM=601070, no_hdr0=''),
  3603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3603, CurrentID=0, Status='live', Symbol='IL16', Aliases='LCF, NIL16, PRIL16, prIL-16', description='interleukin 16', other_designations='lymphocyte chemoattractant factor|neuronal interleukin 16|prointerleukin 16', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81182600, end_position_on_the_genomic_accession=81312763, orientation='plus', exon_count=21, OMIM=603035, no_hdr0=''),
  3604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3604, CurrentID=0, Status='live', Symbol='TNFRSF9', Aliases='4-1BB, CD137, CDw137, ILA', description='tumor necrosis factor receptor superfamily member 9', other_designations='4-1BB ligand receptor|CD137 antigen|T cell antigen ILA|T-cell antigen 4-1BB homolog|T-cell antigen ILA|homolog of mouse 4-1BB|induced by lymphocyte activation (ILA)|interleukin-activated receptor, homolog of mouse Ly63|receptor protein 4-1BB|tumor necrosis factor receptor superfamily, member 9', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7915871, end_position_on_the_genomic_accession=7943165, orientation='minus', exon_count=10, OMIM=602250, no_hdr0=''),
  3605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3605, CurrentID=0, Status='live', Symbol='IL17A', Aliases='CTLA-8, CTLA8, IL-17, IL-17A, IL17', description='interleukin 17A', other_designations='cytotoxic T-lymphocyte-associated antigen 8|cytotoxic T-lymphocyte-associated protein 8|interleukin 17 (cytotoxic T-lymphocyte-associated serine esterase 8)', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52186387, end_position_on_the_genomic_accession=52190638, orientation='plus', exon_count=3, OMIM=603149, no_hdr0=''),
  3606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3606, CurrentID=0, Status='live', Symbol='IL18', Aliases='IGIF, IL-18, IL-1g, IL1F4', description='interleukin 18', other_designations='IFN-gamma-inducing factor|IL-1 gamma|iboctadekin|interleukin 18 (interferon-gamma-inducing factor)|interleukin-1 gamma', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112143251, end_position_on_the_genomic_accession=112164117, orientation='minus', exon_count=6, OMIM=600953, no_hdr0=''),
  3607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3607, CurrentID=0, Status='live', Symbol='FOXK2', Aliases='ILF, ILF-1, ILF1', description='forkhead box K2', other_designations='FOXK1|cellular transcription factor ILF-1|interleukin enhancer binding factor 1|interleukin enhancer-binding factor 1', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82519718, end_position_on_the_genomic_accession=82604607, orientation='plus', exon_count=11, OMIM=147685, no_hdr0=''),
  3608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3608, CurrentID=0, Status='live', Symbol='ILF2', Aliases='NF45, PRO3063', description='interleukin enhancer binding factor 2', other_designations='interleukin enhancer binding factor 2, 45kD|interleukin enhancer binding factor 2, 45kDa|nuclear factor of activated T-cells, 45-kDa', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153661788, end_position_on_the_genomic_accession=153671028, orientation='minus', exon_count=14, OMIM=603181, no_hdr0=''),
  3609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3609, CurrentID=0, Status='live', Symbol='ILF3', Aliases='CBTF, DRBF, DRBP76, MMP4, MPHOSPH4, MPP4, NF-AT-90, NF110, NF110b, NF90, NF90a, NF90b, NFAR, NFAR-1, NFAR2, TCP110, TCP80', description='interleukin enhancer binding factor 3', other_designations='M-phase phosphoprotein 4|double-stranded RNA-binding protein, 76 kD|dsRNA binding protein NFAR-2/MPP4|interleukin enhancer binding factor 3, 90kD|interleukin enhancer binding factor 3, 90kDa|nuclear factor associated with dsRNA|nuclear factor of activated T-cells 90 kDa|nuclear factor of activated T-cells, 90 kD|translational control protein 80', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10654261, end_position_on_the_genomic_accession=10692419, orientation='plus', exon_count=21, OMIM=603182, no_hdr0=''),
  3611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3611, CurrentID=0, Status='live', Symbol='ILK', Aliases='HEL-S-28-1, ILK-2, P59, p59ILK, ILK', description='integrin linked kinase', other_designations='59 kDa serine/threonine-protein kinase|epididymis secretory protein Li 28|integrin-linked kinase|integrin-linked kinase-2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6603708, end_position_on_the_genomic_accession=6610874, orientation='plus', exon_count=13, OMIM=602366, no_hdr0=''),
  3612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3612, CurrentID=0, Status='live', Symbol='IMPA1', Aliases='IMP, IMPA', description='inositol(myo)-1(or 4)-monophosphatase 1', other_designations='D-galactose 1-phosphate phosphatase|IMP 1|IMPase 1|inositol-1(or 4)-monophosphatase 1|lithium-sensitive myo-inositol monophosphatase A1', map_location='8q21.13-q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81656916, end_position_on_the_genomic_accession=81686354, orientation='minus', exon_count=10, OMIM=602064, no_hdr0=''),
  3613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3613, CurrentID=0, Status='live', Symbol='IMPA2', Aliases='', description='inositol(myo)-1(or 4)-monophosphatase 2', other_designations='IMP 2|IMPase 2|inosine monophosphatase 2|inositol monophosphatase 2 variant 1|inositol monophosphatase 2 variant 2|myo-inositol monophosphatase A2', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11981428, end_position_on_the_genomic_accession=12030886, orientation='plus', exon_count=10, OMIM=605922, no_hdr0=''),
  3614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3614, CurrentID=0, Status='live', Symbol='IMPDH1', Aliases='IMPD, IMPD1, IMPDH-I, LCA11, RP10, sWSS2608', description="IMP (inosine 5'-monophosphate) dehydrogenase 1", other_designations='IMP (inosine monophosphate) dehydrogenase 1|IMPD 1|IMPDH 1', map_location='7q31.3-q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128392277, end_position_on_the_genomic_accession=128409987, orientation='minus', exon_count=18, OMIM=146690, no_hdr0=''),
  3615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3615, CurrentID=0, Status='live', Symbol='IMPDH2', Aliases='IMPD2, IMPDH-II', description="IMP (inosine 5'-monophosphate) dehydrogenase 2", other_designations="IMP (inosine monophosphate) dehydrogenase 2|IMP oxireductase 2|IMPD 2|IMPDH 2|inosine 5' phosphate dehydrogenase 2|inosine monophosphate dehydrogenase type II", map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49024325, end_position_on_the_genomic_accession=49029701, orientation='minus', exon_count=15, OMIM=146691, no_hdr0=''),
  101060128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060128, CurrentID=0, Status='live', Symbol='LOC101060128', Aliases='', description='putative uncharacterized protein encoded by LINC00174', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63744446, end_position_on_the_genomic_accession=63747240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3617, CurrentID=0, Status='live', Symbol='IMPG1', Aliases='GP147, IPM150, SPACR, VMD4', description='interphotoreceptor matrix proteoglycan 1', other_designations='interphotoreceptor matrix proteoglycan of 150 kDa|sialoprotein associated with cones and rods', map_location='6q14.2-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75921114, end_position_on_the_genomic_accession=76072678, orientation='minus', exon_count=17, OMIM=602870, no_hdr0=''),
  3619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3619, CurrentID=0, Status='live', Symbol='INCENP', Aliases='', description='inner centromere protein', other_designations='binds and activates aurora-B and -C in vivo and in vitro|chromosomal passenger protein|inner centromere protein INCENP|inner centromere protein antigens 135/155kDa', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62123973, end_position_on_the_genomic_accession=62153163, orientation='plus', exon_count=19, OMIM=604411, no_hdr0=''),
  3620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3620, CurrentID=0, Status='live', Symbol='IDO1', Aliases='IDO, IDO-1, INDO', description='indoleamine 2,3-dioxygenase 1', other_designations='indolamine 2,3 dioxygenase|indole 2,3-dioxygenase|indoleamine-pyrrole 2,3-dioxygenase', map_location='8p12-p11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39913809, end_position_on_the_genomic_accession=39928790, orientation='plus', exon_count=10, OMIM=147435, no_hdr0=''),
  3621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3621, CurrentID=0, Status='live', Symbol='ING1', Aliases='p24ING1c, p33, p33ING1, p33ING1b, p47, p47ING1a', description='inhibitor of growth family member 1', other_designations='growth inhibitor ING1|growth inhibitory protein ING1|inhibitor of growth family, member 1|tumor suppressor ING1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110712623, end_position_on_the_genomic_accession=110721074, orientation='plus', exon_count=4, OMIM=601566, no_hdr0=''),
  3622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3622, CurrentID=0, Status='live', Symbol='ING2', Aliases='ING1L, p33ING2', description='inhibitor of growth family member 2', other_designations='ING1Lp|inhibitor of growth 1-like protein|inhibitor of growth family, member 2|p32', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183505050, end_position_on_the_genomic_accession=183512428, orientation='plus', exon_count=4, OMIM=604215, no_hdr0=''),
  3623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3623, CurrentID=0, Status='live', Symbol='INHA', Aliases='', description='inhibin alpha', other_designations='A-inhibin subunit|inhibin, alpha', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219572232, end_position_on_the_genomic_accession=219575713, orientation='plus', exon_count=2, OMIM=147380, no_hdr0=''),
  3624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3624, CurrentID=0, Status='live', Symbol='INHBA', Aliases='EDF, FRP', description='inhibin beta A', other_designations='FSH-releasing protein|Inhibin, beta-1|activin beta-A chain|erythroid differentiation factor|erythroid differentiation protein|follicle-stimulating hormone-releasing protein|inhibin, beta A (activin A, activin AB alpha polypeptide)', map_location='7p15-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41689003, end_position_on_the_genomic_accession=41703108, orientation='minus', exon_count=3, OMIM=147290, no_hdr0=''),
  3625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3625, CurrentID=0, Status='live', Symbol='INHBB', Aliases='', description='inhibin beta B', other_designations='Inhibin, beta-2|activin AB beta polypeptide|activin beta-B chain', map_location='2cen-q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120346143, end_position_on_the_genomic_accession=120351807, orientation='plus', exon_count=2, OMIM=147390, no_hdr0=''),
  3626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3626, CurrentID=0, Status='live', Symbol='INHBC', Aliases='IHBC', description='inhibin beta C', other_designations='activin beta-C chain|inhibin, beta C', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57434685, end_position_on_the_genomic_accession=57452062, orientation='plus', exon_count=2, OMIM=601233, no_hdr0=''),
  3627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3627, CurrentID=0, Status='live', Symbol='CXCL10', Aliases='C7, IFI10, INP10, IP-10, SCYB10, crg-2, gIP-10, mob-1', description='chemokine (C-X-C motif) ligand 10', other_designations='10 kDa interferon gamma-induced protein|gamma IP10|gamma-IP10|interferon-inducible cytokine IP-10|protein 10 from interferon (gamma)-induced cell line|small inducible cytokine subfamily B (Cys-X-Cys), member 10|small-inducible cytokine B10', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76021116, end_position_on_the_genomic_accession=76023536, orientation='minus', exon_count=4, OMIM=147310, no_hdr0=''),
  3628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3628, CurrentID=0, Status='live', Symbol='INPP1', Aliases='', description='inositol polyphosphate-1-phosphatase', other_designations='IPP|IPPase', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190343470, end_position_on_the_genomic_accession=190371665, orientation='plus', exon_count=8, OMIM=147263, no_hdr0=''),
  134701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134701, CurrentID=0, Status='live', Symbol='RIPPLY2', Aliases='C6orf159, dJ237I15.1', description='ripply transcriptional repressor 2', other_designations='ripply2 homolog', map_location='6q14.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83853266, end_position_on_the_genomic_accession=83857515, orientation='plus', exon_count=4, OMIM=609891, no_hdr0=''),
  3630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3630, CurrentID=0, Status='live', Symbol='INS', Aliases='IDDM, IDDM1, IDDM2, ILPR, IRDN, MODY10', description='insulin', other_designations='preproinsulin|proinsulin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2159779, end_position_on_the_genomic_accession=2161209, orientation='minus', exon_count=3, OMIM=176730, no_hdr0=''),
  3631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3631, CurrentID=0, Status='live', Symbol='INPP4A', Aliases='INPP4, TVAS1', description='inositol polyphosphate-4-phosphatase type I A', other_designations='inositol polyphosphate-4-phosphatase, type I, 107kD|inositol polyphosphate-4-phosphatase, type I, 107kDa|type I inositol-3,4-bisphosphate 4-phosphatase', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98444858, end_position_on_the_genomic_accession=98594067, orientation='plus', exon_count=34, OMIM=600916, no_hdr0=''),
  3632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3632, CurrentID=0, Status='live', Symbol='INPP5A', Aliases='5PTASE', description='inositol polyphosphate-5-phosphatase A', other_designations='43 kDa inositol polyphosphate 5-phophatase|CTCL tumor antigen HD-CL-02|InsP3 5-phosphatase|inositol polyphosphate-5-phosphatase, 40kD|inositol polyphosphate-5-phosphatase, 40kDa|inositol trisphosphate-5-phosphatase, 40kD|type I inositol-1,4,5-trisphosphate 5-phophatase|type I inositol-1,4,5-trisphosphate 5-phosphatase', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132537784, end_position_on_the_genomic_accession=132783480, orientation='plus', exon_count=20, OMIM=600106, no_hdr0=''),
  3633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3633, CurrentID=0, Status='live', Symbol='INPP5B', Aliases='5PTase', description='inositol polyphosphate-5-phosphatase B', other_designations='inositol polyphosphate-5-phosphatase, 75kDa|phosphoinositide 5-phosphatase', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37860697, end_position_on_the_genomic_accession=37947084, orientation='minus', exon_count=32, OMIM=147264, no_hdr0=''),
  101060146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060146, CurrentID=0, Status='live', Symbol='CHMP1B2P', Aliases='', description='charged multivesicular body protein 1B2, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80228489, end_position_on_the_genomic_accession=80335318, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3635, CurrentID=0, Status='live', Symbol='INPP5D', Aliases='SHIP, SHIP-1, SHIP1, SIP-145, hp51CN, p150Ship', description='inositol polyphosphate-5-phosphatase D', other_designations="SH2 domain-containing inositol 5'-phosphatase 1|inositol polyphosphate-5-phosphatase, 145kD|inositol polyphosphate-5-phosphatase, 145kDa|signaling inositol polyphosphate 5 phosphatase SIP-145|signaling inositol polyphosphate phosphatase SHIP II", map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233059967, end_position_on_the_genomic_accession=233207903, orientation='plus', exon_count=28, OMIM=601582, no_hdr0=''),
  3636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3636, CurrentID=0, Status='live', Symbol='INPPL1', Aliases='OPSMD, SHIP2', description='inositol polyphosphate phosphatase like 1', other_designations="51C protein|INPPL-1|SH2 domain-containing inositol 5'-phosphatase 2|SH2 domain-containing inositol-5'-phosphatase 2|SHIP-2|inositol polyphosphate phosphatase-like 1|phosphatidylinositol-3,4,5-trisphosphate 5-phosphatase 2|protein 51C", map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72223866, end_position_on_the_genomic_accession=72239147, orientation='plus', exon_count=29, OMIM=600829, no_hdr0=''),
  101060149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060149, CurrentID=0, Status='live', Symbol='LOC101060149', Aliases='', description='40S ribosomal protein S24 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5260644, end_position_on_the_genomic_accession=5261044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3638, CurrentID=0, Status='live', Symbol='INSIG1', Aliases='CL-6, CL6', description='insulin induced gene 1', other_designations='INSIG-1 membrane protein', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155297776, end_position_on_the_genomic_accession=155310235, orientation='plus', exon_count=7, OMIM=602055, no_hdr0=''),
  3640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3640, CurrentID=0, Status='live', Symbol='INSL3', Aliases='RLF, RLNL, ley-I-L', description='insulin like 3', other_designations='insulin-like 3 (Leydig cell)|leydig insulin -like hormone|leydig insulin-like peptide|prepro-INSL3|relaxin-like factor b', map_location='19p13.2-p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17816513, end_position_on_the_genomic_accession=17821574, orientation='minus', exon_count=3, OMIM=146738, no_hdr0=''),
  3641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3641, CurrentID=0, Status='live', Symbol='INSL4', Aliases='EPIL, PLACENTIN', description='insulin like 4', other_designations='early placenta insulin-like peptide (EPIL)|insulin-like 4 (placenta)|insulin-like peptide 4', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5231419, end_position_on_the_genomic_accession=5233967, orientation='plus', exon_count=3, OMIM=600910, no_hdr0=''),
  3642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3642, CurrentID=0, Status='live', Symbol='INSM1', Aliases='IA-1, IA1', description='insulinoma associated 1', other_designations='insulinoma-associated 1|zinc finger protein IA-1', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20368121, end_position_on_the_genomic_accession=20370949, orientation='plus', exon_count=1, OMIM=600010, no_hdr0=''),
  3643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3643, CurrentID=0, Status='live', Symbol='INSR', Aliases='CD220, HHF5', description='insulin receptor', other_designations='IR', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7112255, end_position_on_the_genomic_accession=7294000, orientation='minus', exon_count=22, OMIM=147670, no_hdr0=''),
  3645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3645, CurrentID=0, Status='live', Symbol='INSRR', Aliases='IRR', description='insulin receptor-related receptor', other_designations='IR-related receptor', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156840873, end_position_on_the_genomic_accession=156858920, orientation='minus', exon_count=22, OMIM=147671, no_hdr0=''),
  3646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3646, CurrentID=0, Status='live', Symbol='EIF3E', Aliases='EIF3-P48, EIF3S6, INT6, eIF3-p46', description='eukaryotic translation initiation factor 3 subunit E', other_designations='eIF-3 p48|eukaryotic translation initiation factor 3 subunit 6|eukaryotic translation initiation factor 3, subunit 6 (48kD)|eukaryotic translation initiation factor 3, subunit 6 48kDa|eukaryotic translation initiation factor 3, subunit E|mammary tumor-associated protein INT6|murine mammary tumor integration site 6 (oncogene homolog)|viral integration site protein INT-6 homolog', map_location='8q22-q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108201743, end_position_on_the_genomic_accession=108248730, orientation='minus', exon_count=13, OMIM=602210, no_hdr0=''),
  3647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3647, CurrentID=0, Status='live', Symbol='EIF3EP1', Aliases='EIF3EP, EIF3S6P1, INT6P1', description='eukaryotic translation initiation factor 3 subunit E pseudogene 1', other_designations='eukaryotic translation initiation factor 3, subunit 6 48kDa pseudogene 1|eukaryotic translation initiation factor 3, subunit E pseudogene 1|murine mammary tumor integration site 6 (oncogene homolog) pseudogene 1', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73291936, end_position_on_the_genomic_accession=73293415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060160, CurrentID=0, Status='live', Symbol='DDTP1', Aliases='', description='D-dopachrome tautomerase pseudogene 1', other_designations='D-dopachrome decarboxylase-like', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=38556936, end_position_on_the_genomic_accession=38557321, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101060162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060162, CurrentID=0, Status='live', Symbol='LOC101060162', Aliases='', description='UDP-glucuronosyltransferase 2A1-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69144739, end_position_on_the_genomic_accession=69147642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3651, CurrentID=0, Status='live', Symbol='PDX1', Aliases='GSF, IDX-1, IPF1, IUF1, MODY4, PAGEN1, PDX-1, STF-1', description='pancreatic and duodenal homeobox 1', other_designations='IPF-1|IUF-1|glucose-sensitive factor|insulin promoter factor 1, homeodomain transcription factor|insulin upstream factor 1|islet/duodenum homeobox-1|pancreatic-duodenal homeobox factor 1|somatostatin transcription factor 1|somatostatin-transactivating factor 1', map_location='13q12.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27919994, end_position_on_the_genomic_accession=27926314, orientation='plus', exon_count=2, OMIM=600733, no_hdr0=''),
  3652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3652, CurrentID=0, Status='live', Symbol='IPP', Aliases='KLHL27', description='intracisternal A particle-promoted polypeptide', other_designations='kelch-like family member 27|kelch-like protein 27', map_location='1p34-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45694324, end_position_on_the_genomic_accession=45750813, orientation='minus', exon_count=11, OMIM=147485, no_hdr0=''),
  3653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3653, CurrentID=0, Status='live', Symbol='IPW', Aliases='NCRNA00002', description='imprinted in Prader-Willi syndrome (non-protein coding)', other_designations='non-protein coding RNA 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25116545, end_position_on_the_genomic_accession=25122476, orientation='plus', exon_count=3, OMIM=601491, no_hdr0=''),
  3654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3654, CurrentID=0, Status='live', Symbol='IRAK1', Aliases='IRAK, pelle', description='interleukin 1 receptor associated kinase 1', other_designations='IRAK-1|Pelle homolog', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154010506, end_position_on_the_genomic_accession=154024584, orientation='minus', exon_count=14, OMIM=300283, no_hdr0=''),
  3655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3655, CurrentID=0, Status='live', Symbol='ITGA6', Aliases='CD49fB, VLA-6, ITGA6', description='integrin subunit alpha 6', other_designations='CD49 antigen-like family member F|integrin alpha 6|integrin alpha6B|integrin, alpha 6', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172427586, end_position_on_the_genomic_accession=172506453, orientation='plus', exon_count=26, OMIM=147556, no_hdr0=''),
  3656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3656, CurrentID=0, Status='live', Symbol='IRAK2', Aliases='IRAK-2', description='interleukin 1 receptor associated kinase 2', other_designations='interleukin-1 receptor associated kinase-2|interleukin-1 receptor-associated kinase 2', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10164879, end_position_on_the_genomic_accession=10243745, orientation='plus', exon_count=13, OMIM=603304, no_hdr0=''),
  3658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3658, CurrentID=0, Status='live', Symbol='IREB2', Aliases='ACO3, IRP2, IRP2AD', description='iron responsive element binding protein 2', other_designations='IRE-BP 2|iron regulatory protein 2|iron-responsive element binding protein 2', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78438176, end_position_on_the_genomic_accession=78501456, orientation='plus', exon_count=22, OMIM=147582, no_hdr0=''),
  3659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3659, CurrentID=0, Status='live', Symbol='IRF1', Aliases='IRF-1, MAR', description='interferon regulatory factor 1', other_designations='interferon regulatory factor 1 isoform +I9|interferon regulatory factor 1 isoform d78|interferon regulatory factor 1 isoform delta4|interferon regulatory factor 1 isoform delta7', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132481609, end_position_on_the_genomic_accession=132490777, orientation='minus', exon_count=10, OMIM=147575, no_hdr0=''),
  3660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3660, CurrentID=0, Status='live', Symbol='IRF2', Aliases='IRF-2', description='interferon regulatory factor 2', other_designations='', map_location='4q34.1-q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184387713, end_position_on_the_genomic_accession=184474580, orientation='minus', exon_count=11, OMIM=147576, no_hdr0=''),
  3661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3661, CurrentID=0, Status='live', Symbol='IRF3', Aliases='', description='interferon regulatory factor 3', other_designations='', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49659569, end_position_on_the_genomic_accession=49665875, orientation='minus', exon_count=11, OMIM=603734, no_hdr0=''),
  3662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3662, CurrentID=0, Status='live', Symbol='IRF4', Aliases='LSIRF, MUM1, NF-EM5, SHEP8', description='interferon regulatory factor 4', other_designations='lymphocyte-specific interferon regulatory factor|multiple myeloma oncogene 1', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=391739, end_position_on_the_genomic_accession=411443, orientation='plus', exon_count=10, OMIM=601900, no_hdr0=''),
  3663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3663, CurrentID=0, Status='live', Symbol='IRF5', Aliases='SLEB10', description='interferon regulatory factor 5', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128937737, end_position_on_the_genomic_accession=128950042, orientation='plus', exon_count=14, OMIM=607218, no_hdr0=''),
  3664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3664, CurrentID=0, Status='live', Symbol='IRF6', Aliases='LPS, OFC6, PIT, PPS, PPS1, VWS, VWS1', description='interferon regulatory factor 6', other_designations='', map_location='1q32.3-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209785623, end_position_on_the_genomic_accession=209806175, orientation='minus', exon_count=9, OMIM=607199, no_hdr0=''),
  3665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3665, CurrentID=0, Status='live', Symbol='IRF7', Aliases='IMD39, IRF-7HA, IRF7B, IRF7C, IRF7H, IRF7', description='interferon regulatory factor 7', other_designations='IRF-7|interferon regulatory factor-7H', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=612553, end_position_on_the_genomic_accession=615999, orientation='minus', exon_count=10, OMIM=605047, no_hdr0=''),
  3667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3667, CurrentID=0, Status='live', Symbol='IRS1', Aliases='HIRS-1', description='insulin receptor substrate 1', other_designations='IRS-1', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226731317, end_position_on_the_genomic_accession=226798790, orientation='minus', exon_count=2, OMIM=147545, no_hdr0=''),
  3669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3669, CurrentID=0, Status='live', Symbol='ISG20', Aliases='CD25, HEM45', description='interferon stimulated exonuclease gene 20kDa', other_designations='estrogen-regulated transcript 45 protein|promyelocytic leukemia nuclear body-associated protein ISG20', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88635622, end_position_on_the_genomic_accession=88656344, orientation='plus', exon_count=8, OMIM=604533, no_hdr0=''),
  3670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3670, CurrentID=0, Status='live', Symbol='ISL1', Aliases='ISLET1, Isl-1', description='ISL LIM homeobox 1', other_designations='ISL1 transcription factor, LIM/homeodomain|islet-1', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51383124, end_position_on_the_genomic_accession=51394738, orientation='plus', exon_count=7, OMIM=600366, no_hdr0=''),
  3671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3671, CurrentID=0, Status='live', Symbol='ISLR', Aliases='HsT17563', description='immunoglobulin superfamily containing leucine-rich repeat', other_designations='', map_location='15q23-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74173746, end_position_on_the_genomic_accession=74176871, orientation='plus', exon_count=3, OMIM=602059, no_hdr0=''),
  3672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3672, CurrentID=0, Status='live', Symbol='ITGA1', Aliases='CD49a, VLA1', description='integrin subunit alpha 1', other_designations='CD49 antigen-like family member A|VLA-1|integrin alpha 1|integrin, alpha 1|laminin and collagen receptor|very late activation protein 1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52788302, end_position_on_the_genomic_accession=52953655, orientation='plus', exon_count=29, OMIM=192968, no_hdr0=''),
  3673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3673, CurrentID=0, Status='live', Symbol='ITGA2', Aliases='BR, CD49B, GPIa, HPA-5, VLA-2, VLAA2', description='integrin subunit alpha 2', other_designations='CD49 antigen-like family member B|collagen receptor|human platelet alloantigen system 5|integrin, alpha 2 (CD49B, alpha 2 subunit of VLA-2 receptor)|platelet antigen Br|platelet glycoprotein GPIa|platelet membrane glycoprotein Ia|very late activation protein 2 receptor, alpha-2 subunit', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52989326, end_position_on_the_genomic_accession=53094779, orientation='plus', exon_count=30, OMIM=192974, no_hdr0=''),
  3674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3674, CurrentID=0, Status='live', Symbol='ITGA2B', Aliases='BDPLT16, BDPLT2, CD41, CD41B, GP2B, GPIIb, GT, GTA, HPA3, PPP1R93', description='integrin subunit alpha 2b', other_designations='GPalpha IIb|alphaIIb protein|integrin alpha 2b|integrin, alpha 2b (platelet glycoprotein IIb of IIb/IIIa complex, antigen CD41)|platelet fibrinogen receptor, alpha subunit|platelet glycoprotein IIb of IIb/IIIa complex|platelet membrane glycoprotein IIb|platelet-specific antigen BAK|protein phosphatase 1, regulatory subunit 93', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44372181, end_position_on_the_genomic_accession=44389505, orientation='minus', exon_count=30, OMIM=607759, no_hdr0=''),
  3675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3675, CurrentID=0, Status='live', Symbol='ITGA3', Aliases='CD49C, FRP-2, GAP-B3, GAPB3, ILNEB, MSK18, VCA-2, VL3A, VLA3a', description='integrin subunit alpha 3', other_designations='CD49 antigen-like family member C|antigen identified by monoclonal antibody J143|galactoprotein B3|integrin alpha 3|integrin, alpha 3 (antigen CD49C, alpha 3 subunit of VLA-3 receptor)|very late activation protein 3 receptor, alpha-3 subunit', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50055976, end_position_on_the_genomic_accession=50090485, orientation='plus', exon_count=26, OMIM=605025, no_hdr0=''),
  3676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3676, CurrentID=0, Status='live', Symbol='ITGA4', Aliases='CD49D, IA4', description='integrin subunit alpha 4', other_designations='269C wild type|CD49 antigen-like family member D|VLA-4 subunit alpha|antigen CD49D, alpha-4 subunit of VLA-4 receptor|integrin alpha 4|integrin alpha-4 subunit|integrin alpha-IV|integrin, alpha 4 (antigen CD49D, alpha 4 subunit of VLA-4 receptor)|very late activation protein 4 receptor, alpha 4 subunit', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181456892, end_position_on_the_genomic_accession=181537747, orientation='plus', exon_count=28, OMIM=192975, no_hdr0=''),
  101060189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060189, CurrentID=0, Status='live', Symbol='UBE2FP2', Aliases='', description='ubiquitin conjugating enzyme E2F (putative) pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2F (putative) pseudogene|ubiquitin-conjugating enzyme E2F (putative) pseudogene 2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63180375, end_position_on_the_genomic_accession=63181084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3678, CurrentID=0, Status='live', Symbol='ITGA5', Aliases='CD49e, FNRA, VLA-5, VLA5A', description='integrin subunit alpha 5', other_designations='CD49 antigen-like family member E|fibronectin receptor subunit alpha|fibronectin receptor, alpha polypeptide|fibronectin receptor, alpha subunit|integrin alpha 5|integrin alpha-F|integrin, alpha 5 (fibronectin receptor, alpha polypeptide)|very late activation protein 5, alpha subunit', map_location='12q11-q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54395261, end_position_on_the_genomic_accession=54419266, orientation='minus', exon_count=30, OMIM=135620, no_hdr0=''),
  3679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3679, CurrentID=0, Status='live', Symbol='ITGA7', Aliases='', description='integrin subunit alpha 7', other_designations='integrin alpha 7 chain|integrin, alpha 7', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55684568, end_position_on_the_genomic_accession=55716037, orientation='minus', exon_count=32, OMIM=600536, no_hdr0=''),
  3680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3680, CurrentID=0, Status='live', Symbol='ITGA9', Aliases='ALPHA-RLC, ITGA4L, RLC', description='integrin subunit alpha 9', other_designations='integrin alpha 9|integrin alpha-RLC|integrin, alpha 9', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37452322, end_position_on_the_genomic_accession=37819790, orientation='plus', exon_count=28, OMIM=603963, no_hdr0=''),
  3681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3681, CurrentID=0, Status='live', Symbol='ITGAD', Aliases='ADB2, CD11D', description='integrin subunit alpha D', other_designations='CD11 antigen-like family member D|leukointegrin alpha D', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31393241, end_position_on_the_genomic_accession=31426513, orientation='plus', exon_count=32, OMIM=602453, no_hdr0=''),
  3682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3682, CurrentID=0, Status='live', Symbol='ITGAE', Aliases='CD103, HUMINAE', description='integrin subunit alpha E', other_designations='HML-1 antigen|antigen CD103, human mucosal lymphocyte antigen 1; alpha polypeptide|integrin alpha E|integrin alpha-IEL|integrin, alpha E (antigen CD103, human mucosal lymphocyte antigen 1; alpha polypeptide)|mucosal lymphocyte 1 antigen', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3714625, end_position_on_the_genomic_accession=3801243, orientation='minus', exon_count=31, OMIM=604682, no_hdr0=''),
  3683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3683, CurrentID=0, Status='live', Symbol='ITGAL', Aliases='CD11A, LFA-1, LFA1A', description='integrin subunit alpha L', other_designations='CD11 antigen-like family member A|LFA-1 alpha|LFA-1A|antigen CD11A (p180), lymphocyte function-associated antigen 1, alpha polypeptide|integrin alpha L|integrin gene promoter|integrin, alpha L (antigen CD11A (p180), lymphocyte function-associated antigen 1; alpha polypeptide)|leukocyte adhesion glycoprotein LFA-1 alpha chain|leukocyte function-associated molecule 1 alpha chain|lymphocyte function-associated antigen 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30472662, end_position_on_the_genomic_accession=30523185, orientation='plus', exon_count=31, OMIM=153370, no_hdr0=''),
  3684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3684, CurrentID=0, Status='live', Symbol='ITGAM', Aliases='CD11B, CR3A, MAC-1, MAC1A, MO1A, SLEB6', description='integrin subunit alpha M', other_designations='CD11 antigen-like family member B|CR-3 alpha chain|antigen CD11b (p170)|cell surface glycoprotein MAC-1 subunit alpha|integrin, alpha M (complement component 3 receptor 3 subunit)|leukocyte adhesion receptor MO1|macrophage antigen alpha polypeptide|neutrophil adherence receptor alpha-M subunit', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31259967, end_position_on_the_genomic_accession=31332892, orientation='plus', exon_count=30, OMIM=120980, no_hdr0=''),
  3685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3685, CurrentID=0, Status='live', Symbol='ITGAV', Aliases='CD51, MSK8, VNRA, VTNR', description='integrin subunit alpha V', other_designations='antigen identified by monoclonal antibody L230|integrin alpha V|integrin alphaVbeta3|integrin, alpha V (vitronectin receptor, alpha polypeptide, antigen CD51)|vitronectin receptor subunit alpha', map_location='2q31-q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186590063, end_position_on_the_genomic_accession=186680902, orientation='plus', exon_count=32, OMIM=193210, no_hdr0=''),
  3687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3687, CurrentID=0, Status='live', Symbol='ITGAX', Aliases='CD11C, SLEB6', description='integrin subunit alpha X', other_designations='CD11 antigen-like family member C|integrin alpha X|integrin, alpha X (antigen CD11C (p150), alpha polypeptide)|integrin, alpha X (complement component 3 receptor 4 subunit)|leu M5, alpha subunit|leukocyte adhesion glycoprotein p150,95 alpha chain|leukocyte adhesion receptor p150,95|leukocyte surface antigen p150,95, alpha subunit|myeloid membrane antigen, alpha subunit|p150 95 integrin alpha chain', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31355134, end_position_on_the_genomic_accession=31382997, orientation='plus', exon_count=30, OMIM=151510, no_hdr0=''),
  3688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3688, CurrentID=0, Status='live', Symbol='ITGB1', Aliases='CD29, FNRB, GPIIA, MDF2, MSK12, VLA-BETA, VLAB', description='integrin subunit beta 1', other_designations='glycoprotein IIa|integrin VLA-4 beta subunit|integrin beta 1|integrin, beta 1 (fibronectin receptor, beta polypeptide, antigen CD29 includes MDF2, MSK12)|very late activation protein, beta polypeptide', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32900318, end_position_on_the_genomic_accession=32958365, orientation='minus', exon_count=18, OMIM=135630, no_hdr0=''),
  3689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3689, CurrentID=0, Status='live', Symbol='ITGB2', Aliases='CD18, LAD, LCAMB, LFA-1, MAC-1, MF17, MFI7', description='integrin subunit beta 2', other_designations='cell surface adhesion glycoprotein (LFA-1/CR3/P150,959 beta subunit precursor)|complement receptor C3 beta-subunit|integrin beta chain, beta 2|integrin, beta 2 (complement component 3 receptor 3 and 4 subunit)|leukocyte cell adhesion molecule CD18|leukocyte-associated antigens CD18/11A, CD18/11B, CD18/11C', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44885949, end_position_on_the_genomic_accession=44928873, orientation='minus', exon_count=17, OMIM=600065, no_hdr0=''),
  3690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3690, CurrentID=0, Status='live', Symbol='ITGB3', Aliases='BDPLT16, BDPLT2, CD61, GP3A, GPIIIa, GT', description='integrin subunit beta 3', other_designations='integrin beta 3|integrin, beta 3 (platelet glycoprotein IIIa, antigen CD61)|platelet membrane glycoprotein IIIa', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47253842, end_position_on_the_genomic_accession=47312711, orientation='plus', exon_count=15, OMIM=173470, no_hdr0=''),
  3691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3691, CurrentID=0, Status='live', Symbol='ITGB4', Aliases='CD104', description='integrin subunit beta 4', other_designations='CD104 antigen|GP150|integrin beta 4|integrin beta-4 subunit|integrin, beta 4', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75721342, end_position_on_the_genomic_accession=75757819, orientation='plus', exon_count=43, OMIM=147557, no_hdr0=''),
  3692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3692, CurrentID=0, Status='live', Symbol='EIF6', Aliases='CAB, EIF3A, ITGB4BP, b(2)gcn, eIF-6, p27(BBP), p27BBP', description='eukaryotic translation initiation factor 6', other_designations='B4 integrin interactor|eIF-6|eukaryotic translation initiation factor 3A|p27 beta-4 integrin-binding protein', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35278906, end_position_on_the_genomic_accession=35284816, orientation='minus', exon_count=8, OMIM=602912, no_hdr0=''),
  3693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3693, CurrentID=0, Status='live', Symbol='ITGB5', Aliases='', description='integrin subunit beta 5', other_designations='integrin beta 5|integrin, beta 5', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124761948, end_position_on_the_genomic_accession=124887538, orientation='minus', exon_count=15, OMIM=147561, no_hdr0=''),
  3694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3694, CurrentID=0, Status='live', Symbol='ITGB6', Aliases='AI1H', description='integrin subunit beta 6', other_designations='integrin beta 6|integrin, beta 6', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160099666, end_position_on_the_genomic_accession=160200313, orientation='minus', exon_count=16, OMIM=147558, no_hdr0=''),
  3695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3695, CurrentID=0, Status='live', Symbol='ITGB7', Aliases='', description='integrin subunit beta 7', other_designations='gut homing receptor beta subunit|integrin beta 7 subunit|integrin, beta 7', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53191318, end_position_on_the_genomic_accession=53207310, orientation='minus', exon_count=16, OMIM=147559, no_hdr0=''),
  3696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3696, CurrentID=0, Status='live', Symbol='ITGB8', Aliases='', description='integrin subunit beta 8', other_designations='integrin beta 8|integrin, beta 8', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20329803, end_position_on_the_genomic_accession=20415759, orientation='plus', exon_count=18, OMIM=604160, no_hdr0=''),
  3697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3697, CurrentID=0, Status='live', Symbol='ITIH1', Aliases='H1P, IATIH, IGHEP1, ITI-HC1, ITIH, SHAP', description='inter-alpha-trypsin inhibitor heavy chain 1', other_designations='ITI heavy chain H1|inter-alpha (globulin) inhibitor, H1 polypeptide|inter-alpha-inhibitor heavy chain 1|inter-alpha-trypsin inhibitor complex component III|serum-derived hyaluronan-associated protein', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52777586, end_position_on_the_genomic_accession=52792068, orientation='plus', exon_count=23, OMIM=147270, no_hdr0=''),
  3698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3698, CurrentID=0, Status='live', Symbol='ITIH2', Aliases='H2P, SHAP', description='inter-alpha-trypsin inhibitor heavy chain 2', other_designations='ITI heavy chain H2|ITI-HC2|inter-alpha (globulin) inhibitor H2|inter-alpha (globulin) inhibitor, H2 polypeptide|inter-alpha-inhibitor heavy chain 2|inter-alpha-trypsin inhibitor complex component II|serum-derived hyaluronan-associated protein', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7703273, end_position_on_the_genomic_accession=7749520, orientation='plus', exon_count=21, OMIM=146640, no_hdr0=''),
  3699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3699, CurrentID=0, Status='live', Symbol='ITIH3', Aliases='H3P, ITI-HC3, SHAP', description='inter-alpha-trypsin inhibitor heavy chain 3', other_designations='ITI heavy chain H3|inter-alpha (globulin) inhibitor H3|inter-alpha (globulin) inhibitor, H3 polypeptide|inter-alpha-inhibitor heavy chain 3|pre-alpha (globulin) inhibitor, H3 polypeptide|serum-derived hyaluronan-associated protein', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52794768, end_position_on_the_genomic_accession=52809009, orientation='plus', exon_count=23, OMIM=146650, no_hdr0=''),
  3700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3700, CurrentID=0, Status='live', Symbol='ITIH4', Aliases='GP120, H4P, IHRP, ITI-HC4, ITIHL1, PK-120, PK120', description='inter-alpha-trypsin inhibitor heavy chain family member 4', other_designations='ITI heavy chain H4|inter-alpha (globulin) inhibitor H4 (plasma Kallikrein-sensitive glycoprotein)|inter-alpha-inhibitor heavy chain 4|inter-alpha-trypsin inhibitor family heavy chain-related protein|inter-alpha-trypsin inhibitor heavy chain family, member 4|inter-alpha-trypsin inhibitor, heavy chain-like, 1|plasma kallikrein-sensitive glycoprotein 120', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52812990, end_position_on_the_genomic_accession=52830701, orientation='minus', exon_count=24, OMIM=600564, no_hdr0=''),
  3702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3702, CurrentID=0, Status='live', Symbol='ITK', Aliases='EMT, LPFS1, LYK, PSCTK2', description='IL2-inducible T-cell kinase', other_designations='IL-2-inducible T cell kinase|IL-2-inducible T-cell kinase|T-cell-specific kinase|homolog of mouse T-cell itk/tsk|interleukin-2-inducible T cell kinase|interleukin-2-inducible T-cell kinase|kinase EMT|tyrosine-protein kinase LYK', map_location='5q31-q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157180896, end_position_on_the_genomic_accession=157255099, orientation='plus', exon_count=17, OMIM=186973, no_hdr0=''),
  3703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3703, CurrentID=0, Status='live', Symbol='STT3A', Aliases='ITM1, STT3-A, TMC', description='STT3A, subunit of the oligosaccharyltransferase complex (catalytic)', other_designations='B5|STT3, subunit of the oligosaccharyltransferase complex, homolog A|STT3A, cataylic subunit of the oligosaccharyltransferase complex|dolichyl-diphosphooligosaccharide protein glycotransferase|integral membrane protein 1|integral transmembrane protein 1|oligosaccharyl transferase subunit STT3A|transmembrane conserved|transmembrane protein TMC', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125592795, end_position_on_the_genomic_accession=125622759, orientation='plus', exon_count=21, OMIM=601134, no_hdr0=''),
  3704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3704, CurrentID=0, Status='live', Symbol='ITPA', Aliases='C20orf37, HLC14-06-P, My049, dJ794I6.3', description='inosine triphosphatase', other_designations='ITPase|NTPase|inosine triphosphatase (nucleoside triphosphate pyrophosphatase)|inosine triphosphatase-A|inosine triphosphate pyrophosphohydrolase|non-canonical purine NTP pyrophosphatase|non-standard purine NTP pyrophosphatase|nucleoside-triphosphate diphosphatase|putative oncogene protein HLC14-06-P', map_location='20p', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3208868, end_position_on_the_genomic_accession=3227449, orientation='plus', exon_count=12, OMIM=147520, no_hdr0=''),
  3705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3705, CurrentID=0, Status='live', Symbol='ITPK1', Aliases='ITRPK1', description='inositol-tetrakisphosphate 1-kinase', other_designations='inositol 1,3,4-triphosphate 5/6 kinase|inositol 1,3,4-trisphosphate 5/6 kinase|inositol 1,3,4-trisphosphate 5/6-kinase|inositol-triphosphate 5/6-kinase|ins(1,3,4)P(3) 5/6-kinase', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92936914, end_position_on_the_genomic_accession=93115918, orientation='minus', exon_count=12, OMIM=601838, no_hdr0=''),
  3706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3706, CurrentID=0, Status='live', Symbol='ITPKA', Aliases='IP3-3KA, IP3KA', description='inositol-trisphosphate 3-kinase A', other_designations='IP3 3-kinase A|IP3K A|inositol 1,4,5-trisphosphate 3-kinase A|insP 3-kinase A', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41493858, end_position_on_the_genomic_accession=41503559, orientation='plus', exon_count=8, OMIM=147521, no_hdr0=''),
  3707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3707, CurrentID=0, Status='live', Symbol='ITPKB', Aliases='IP3-3KB, IP3K, IP3K-B, IP3KB, PIG37', description='inositol-trisphosphate 3-kinase B', other_designations='IP3 3-kinase B|IP3K B|inositol 1,4,5-trisphosphate 3-kinase B|insP 3-kinase B|proliferation-inducing protein 37', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226631690, end_position_on_the_genomic_accession=226739327, orientation='minus', exon_count=9, OMIM=147522, no_hdr0=''),
  3708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3708, CurrentID=0, Status='live', Symbol='ITPR1', Aliases='ACV, CLA4, INSP3R1, IP3R, IP3R1, PPP1R94, SCA15, SCA16, SCA29', description='inositol 1,4,5-trisphosphate receptor, type 1', other_designations='IP3 receptor|IP3R 1|inositol 1,4,5-triphosphate receptor, type 1|protein phosphatase 1, regulatory subunit 94|type 1 InsP3 receptor|type 1 inositol 1,4,5-trisphosphate receptor', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4493348, end_position_on_the_genomic_accession=4847840, orientation='plus', exon_count=63, OMIM=147265, no_hdr0=''),
  3709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3709, CurrentID=0, Status='live', Symbol='ITPR2', Aliases='ANHD, CFAP48, INSP3R2, IP3R2', description='inositol 1,4,5-trisphosphate receptor, type 2', other_designations='IP3 receptor|IP3R 2|cilia and flagella associated protein 48|type 2 InsP3 receptor', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26335352, end_position_on_the_genomic_accession=26833198, orientation='minus', exon_count=60, OMIM=600144, no_hdr0=''),
  3710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3710, CurrentID=0, Status='live', Symbol='ITPR3', Aliases='IP3R, IP3R3', description='inositol 1,4,5-trisphosphate receptor, type 3', other_designations='IP3 receptor|inositol 1,4,5-triphosphate receptor, type 3|insP3R3|type 3 InsP3 receptor', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33621379, end_position_on_the_genomic_accession=33696574, orientation='plus', exon_count=61, OMIM=147267, no_hdr0=''),
  3712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3712, CurrentID=0, Status='live', Symbol='IVD', Aliases='ACAD2', description='isovaleryl-CoA dehydrogenase', other_designations='isovaleryl CoA dehydrogenase|isovaleryl Coenzyme A dehydrogenase', map_location='15q14-q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40405485, end_position_on_the_genomic_accession=40435948, orientation='plus', exon_count=16, OMIM=607036, no_hdr0=''),
  3713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3713, CurrentID=0, Status='live', Symbol='IVL', Aliases='', description='involucrin', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152908563, end_position_on_the_genomic_accession=152911886, orientation='plus', exon_count=3, OMIM=147360, no_hdr0=''),
  3714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3714, CurrentID=0, Status='live', Symbol='JAG2', Aliases='HJ2, SER2', description='jagged 2', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105140981, end_position_on_the_genomic_accession=105168824, orientation='minus', exon_count=26, OMIM=602570, no_hdr0=''),
  101060227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060227, CurrentID=0, Status='live', Symbol='LOC101060227', Aliases='', description='uncharacterized LOC101060227', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149558141, end_position_on_the_genomic_accession=149561039, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3716, CurrentID=0, Status='live', Symbol='JAK1', Aliases='JAK1AB, JTK3, JAK1', description='Janus kinase 1', other_designations='', map_location='1p32.3-p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64833223, end_position_on_the_genomic_accession=64966910, orientation='minus', exon_count=27, OMIM=147795, no_hdr0=''),
  3717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3717, CurrentID=0, Status='live', Symbol='JAK2', Aliases='JTK10, THCYT3', description='Janus kinase 2', other_designations='JAK-2|Janus kinase 2 (a protein tyrosine kinase)', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4985245, end_position_on_the_genomic_accession=5128183, orientation='plus', exon_count=25, OMIM=147796, no_hdr0=''),
  3718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3718, CurrentID=0, Status='live', Symbol='JAK3', Aliases='JAK-3_HUMAN, JAKL, L-JAK, LJAK, JAK3', description='Janus kinase 3', other_designations='Janus kinase 3 (a protein tyrosine kinase, leukocyte)|leukocyte Janus kinase', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17824780, end_position_on_the_genomic_accession=17848071, orientation='minus', exon_count=26, OMIM=600173, no_hdr0=''),
  3719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3719, CurrentID=0, Status='live', Symbol='JBS', Aliases='', description='Jacobsen syndrome', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147791, no_hdr0=''),
  3720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3720, CurrentID=0, Status='live', Symbol='JARID2', Aliases='JMJ', description='jumonji and AT-rich interaction domain containing 2', other_designations='jumonji homolog|jumonji, AT rich interactive domain 2|jumonji-like protein|jumonji/ARID domain-containing protein 2', map_location='6p24-p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15245975, end_position_on_the_genomic_accession=15522042, orientation='plus', exon_count=23, OMIM=601594, no_hdr0=''),
  101060233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060233, CurrentID=0, Status='live', Symbol='OPN1MW3', Aliases='', description='opsin 1 (cone pigments), medium-wave-sensitive 3', other_designations='medium-wave-sensitive opsin 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154257620, end_position_on_the_genomic_accession=154271805, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101060236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060236, CurrentID=0, Status='live', Symbol='LOC101060236', Aliases='', description='developmental pluripotency associated 3 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104076474, end_position_on_the_genomic_accession=104077535, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3725, CurrentID=0, Status='live', Symbol='JUN', Aliases='AP-1, AP1, c-Jun', description='jun proto-oncogene', other_designations='Jun activation domain binding protein|activator protein 1|enhancer-binding protein AP1|jun oncogene|p39|proto-oncogene c-Jun|v-jun avian sarcoma virus 17 oncogene homolog|v-jun sarcoma virus 17 oncogene homolog', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58780791, end_position_on_the_genomic_accession=58784113, orientation='minus', exon_count=1, OMIM=165160, no_hdr0=''),
  3726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3726, CurrentID=0, Status='live', Symbol='JUNB', Aliases='AP-1', description='jun B proto-oncogene', other_designations='activator protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12791496, end_position_on_the_genomic_accession=12793311, orientation='plus', exon_count=1, OMIM=165161, no_hdr0=''),
  3727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3727, CurrentID=0, Status='live', Symbol='JUND', Aliases='AP-1', description='jun D proto-oncogene', other_designations='JunD-FL isoform|activator protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18279694, end_position_on_the_genomic_accession=18281656, orientation='minus', exon_count=1, OMIM=165162, no_hdr0=''),
  3728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3728, CurrentID=0, Status='live', Symbol='JUP', Aliases='ARVD12, CTNNG, DP3, DPIII, PDGB, PKGB', description='junction plakoglobin', other_designations='catenin (cadherin-associated protein), gamma 80kDa|desmoplakin III|desmoplakin-3', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41754607, end_position_on_the_genomic_accession=41786712, orientation='minus', exon_count=19, OMIM=173325, no_hdr0=''),
  3730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3730, CurrentID=0, Status='live', Symbol='ANOS1', Aliases='ADMLX, HH1, HHA, KAL, KAL1, KALIG-1, KMS, WFDC19', description='anosmin 1', other_designations='Kallmann syndrome 1 sequence|Kallmann syndrome interval gene 1|Kallmann syndrome-1 sequence (anosmin-1)|WAP four-disulfide core domain 19|adhesion molecule-like X-linked|kallmann syndrome protein', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8528874, end_position_on_the_genomic_accession=8732186, orientation='minus', exon_count=15, OMIM=300836, no_hdr0=''),
  677827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677827, CurrentID=0, Status='live', Symbol='SNORA46', Aliases='ACA46', description='small nucleolar RNA, H/ACA box 46', other_designations='ACA46 snoRNA', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58548499, end_position_on_the_genomic_accession=58548633, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3732, CurrentID=0, Status='live', Symbol='CD82', Aliases='4F9, C33, GR15, IA4, KAI1, R2, SAR2, ST6, TSPAN27', description='CD82 molecule', other_designations='C33 antigen|inducible membrane protein R2|kangai 1 (suppression of tumorigenicity 6, prostate; CD82 antigen (R2 leukocyte antigen, antigen detected by monoclonal and antibody IA4))|metastasis suppressor Kangai-1|tetraspanin-27|tspan-27', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44564427, end_position_on_the_genomic_accession=44620363, orientation='plus', exon_count=13, OMIM=600623, no_hdr0=''),
  3734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3734, CurrentID=0, Status='live', Symbol='ANOS2P', Aliases='ADMLY, KAL-Y, KALP', description='anosmin 2, pseudogene', other_designations='Kallmann syndrome sequence pseudogene|WUGSC:H_NH0304C24.2', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13751656, end_position_on_the_genomic_accession=13920022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3735, CurrentID=0, Status='live', Symbol='KARS', Aliases='CMTRIB, DFNB891, KARS2, KRS, KARS', description='lysyl-tRNA synthetase', other_designations='lysRS|lysine tRNA ligase', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75627724, end_position_on_the_genomic_accession=75647687, orientation='minus', exon_count=15, OMIM=601421, no_hdr0=''),
  3736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3736, CurrentID=0, Status='live', Symbol='KCNA1', Aliases='AEMK, EA1, HBK1, HUK1, KV1.1, MBK1, MK1, RBK1', description='potassium channel, voltage gated shaker related subfamily A, member 1', other_designations='potassium voltage-gated channel, shaker-related subfamily, member 1 (episodic ataxia with myokymia)|voltage-gated K(+) channel HuKI|voltage-gated potassium channel HBK1|voltage-gated potassium channel subunit Kv1.1', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4909907, end_position_on_the_genomic_accession=4918256, orientation='plus', exon_count=2, OMIM=176260, no_hdr0=''),
  3737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3737, CurrentID=0, Status='live', Symbol='KCNA2', Aliases='EIEE32, HBK5, HK4, HUKIV, KV1.2, MK2, NGK1, RBK2', description='potassium channel, voltage gated shaker related subfamily A, member 2', other_designations='potassium voltage-gated channel, shaker-related subfamily, member 2|voltage-gated K(+) channel HuKIV|voltage-gated potassium channel HBK5|voltage-gated potassium channel protein Kv1.2|voltage-gated potassium channel subunit Kv1.2', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110593580, end_position_on_the_genomic_accession=110631514, orientation='minus', exon_count=9, OMIM=176262, no_hdr0=''),
  3738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3738, CurrentID=0, Status='live', Symbol='KCNA3', Aliases='HGK5, HLK3, HPCN3, HUKIII, KV1.3, MK3, PCN3', description='potassium channel, voltage gated shaker related subfamily A, member 3', other_designations='potassium channel 3|potassium voltage-gated channel, shaker-related subfamily, member 3|type n potassium channel|voltage-gated K(+) channel HuKIII|voltage-gated potassium channel protein Kv1.3|voltage-gated potassium channel subunit Kv1.3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110653560, end_position_on_the_genomic_accession=110675033, orientation='minus', exon_count=4, OMIM=176263, no_hdr0=''),
  3739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3739, CurrentID=0, Status='live', Symbol='KCNA4', Aliases='HBK4, HK1, HPCN2, HUKIIL, KCNA8, KV1.4, PCN2, KCNA4', description='potassium channel, voltage gated shaker related subfamily A, member 4', other_designations='cardiac potassium channel|fetal skeletal muscle potassium channel|potassium channel 2|rapidly inactivating potassium channel|shaker-related potassium channel Kv1.4|type A potassium channel|voltage-gated K(+) channel HuKII|voltage-gated potassium channel HBK4|voltage-gated potassium channel HK1|voltage-gated potassium channel subunit Kv1.4', map_location='11p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30009741, end_position_on_the_genomic_accession=30017030, orientation='minus', exon_count=2, OMIM=176266, no_hdr0=''),
  3741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3741, CurrentID=0, Status='live', Symbol='KCNA5', Aliases='ATFB7, HCK1, HK2, HPCN1, KV1.5, PCN1', description='potassium channel, voltage gated shaker related subfamily A, member 5', other_designations='cardiac potassium channel|insulinoma and islet potassium channel|potassium channel 1|potassium voltage-gated channel, shaker-related subfamily, member 5|voltage-gated potassium channel HK2|voltage-gated potassium channel protein Kv1.5|voltage-gated potassium channel subunit Kv1.5', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5043919, end_position_on_the_genomic_accession=5046788, orientation='plus', exon_count=1, OMIM=176267, no_hdr0=''),
  3742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3742, CurrentID=0, Status='live', Symbol='KCNA6', Aliases='HBK2, KV1.6, PPP1R96', description='potassium channel, voltage gated shaker related subfamily A, member 6', other_designations='human brain potassium channel-2|potassium voltage-gated channel, shaker-related subfamily, member 6|protein phosphatase 1, regulatory subunit 96|voltage-gated potassium channel HBK2|voltage-gated potassium channel protein Kv1.6|voltage-gated potassium channel subunit Kv1.6', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4806885, end_position_on_the_genomic_accession=4851114, orientation='plus', exon_count=3, OMIM=176257, no_hdr0=''),
  3743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3743, CurrentID=0, Status='live', Symbol='KCNA7', Aliases='HAK6, KV1.7', description='potassium channel, voltage gated shaker related subfamily A, member 7', other_designations='potassium voltage-gated channel, shaker-related subfamily, member 7|voltage-dependent potassium channel Kv1.7|voltage-gated potassium channel KCNA7|voltage-gated potassium channel subunit Kv1.7', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49067418, end_position_on_the_genomic_accession=49072941, orientation='minus', exon_count=2, OMIM=176268, no_hdr0=''),
  3744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3744, CurrentID=0, Status='live', Symbol='KCNA10', Aliases='Kcn1, Kv1.8', description='potassium channel, voltage gated shaker related subfamily A, member 10', other_designations='cyclic GMP gated potassium channel|potassium voltage-gated channel, shaker-related subfamily, member 10|voltage-gated potassium channel subunit Kv1.8', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110517217, end_position_on_the_genomic_accession=110519175, orientation='minus', exon_count=1, OMIM=602420, no_hdr0=''),
  3745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3745, CurrentID=0, Status='live', Symbol='KCNB1', Aliases='DRK1, EIEE26, KV2.1, h-DRK1', description='potassium channel, voltage gated Shab related subfamily B, member 1', other_designations='delayed rectifier potassium channel 1|delayed rectifier potassium channel Kv2.1|h-DRK1 K(+) channel|potassium channel protein DRK1|potassium voltage-gated channel, Shab-related subfamily, member 1|voltage-gated potassium channel subunit Kv2.1', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49363877, end_position_on_the_genomic_accession=49484042, orientation='minus', exon_count=4, OMIM=600397, no_hdr0=''),
  3746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3746, CurrentID=0, Status='live', Symbol='KCNC1', Aliases='EPM7, KV3.1, KV4, NGK2', description='potassium channel, voltage gated Shaw related subfamily C, member 1', other_designations='voltage-gated potassium channel protein KV3.1|voltage-gated potassium channel subunit Kv4', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17735910, end_position_on_the_genomic_accession=17864288, orientation='plus', exon_count=8, OMIM=176258, no_hdr0=''),
  3747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3747, CurrentID=0, Status='live', Symbol='KCNC2', Aliases='KV3.2', description='potassium channel, voltage gated Shaw related subfamily C, member 2', other_designations='potassium voltage-gated channel, Shaw-related subfamily, member 2|shaw-like potassium channel|voltage-gated potassium channel Kv3.2', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75040078, end_position_on_the_genomic_accession=75209748, orientation='minus', exon_count=10, OMIM=176256, no_hdr0=''),
  3748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3748, CurrentID=0, Status='live', Symbol='KCNC3', Aliases='KSHIIID, KV3.3, SCA13', description='potassium channel, voltage gated Shaw related subfamily C, member 3', other_designations='Shaw-related voltage-gated potassium channel protein 3|potassium voltage-gated channel, Shaw-related subfamily, member 3|voltage-gated potassium channel protein KV3.3|voltage-gated potassium channel subunit Kv3.3', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50311942, end_position_on_the_genomic_accession=50333515, orientation='minus', exon_count=8, OMIM=176264, no_hdr0=''),
  3749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3749, CurrentID=0, Status='live', Symbol='KCNC4', Aliases='C1orf30, HKSHIIIC, KSHIIIC, KV3.4', description='potassium channel, voltage gated Shaw related subfamily C, member 4', other_designations='K+ channel subunit|potassium voltage-gated channel, Shaw-related subfamily, member 4|voltage-gated potassium channel subunit KV3.4', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110210714, end_position_on_the_genomic_accession=110282617, orientation='plus', exon_count=7, OMIM=176265, no_hdr0=''),
  3750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3750, CurrentID=0, Status='live', Symbol='KCND1', Aliases='KV4.1', description='potassium channel, voltage gated Shal related subfamily D, member 1', other_designations='Shal-type potassium channel|potassium voltage-gated channel, Shal-related subfamily, member 1|voltage-gated potassium channel subunit Kv4.1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48960983, end_position_on_the_genomic_accession=48971844, orientation='minus', exon_count=7, OMIM=300281, no_hdr0=''),
  3751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3751, CurrentID=0, Status='live', Symbol='KCND2', Aliases='KV4.2, RK5', description='potassium channel, voltage gated Shal related subfamily D, member 2', other_designations='voltage-gated potassium channel subunit Kv4.2|voltage-sensitive potassium channel', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120273641, end_position_on_the_genomic_accession=120750333, orientation='plus', exon_count=10, OMIM=605410, no_hdr0=''),
  3752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3752, CurrentID=0, Status='live', Symbol='KCND3', Aliases='BRGDA9L, KCND3S, KSHIVB, KV4.3, SCA19, SCA22, KCND3', description='potassium channel, voltage gated Shal related subfamily D, member 3', other_designations='potassium ionic channel Kv4.3|potassium voltage-gated channel long|potassium voltage-gated channel, Shal-related subfamily, member 3|sha1-related potassium channel Kv4.3|voltage-gated K+ channel|voltage-gated potassium channel subunit Kv4.3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111775822, end_position_on_the_genomic_accession=111989576, orientation='minus', exon_count=14, OMIM=605411, no_hdr0=''),
  3753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3753, CurrentID=0, Status='live', Symbol='KCNE1', Aliases='ISK, JLNS, JLNS2, LQT2/5, LQT5, MinK', description='potassium channel, voltage gated subfamily E regulatory beta subunit 1', other_designations='IKs producing slow voltage-gated potassium channel subunit beta Mink|cardiac delayed rectifier potassium channel protein|delayed rectifier potassium channel subunit IsK|minimal potassium channel|potassium voltage-gated channel, Isk-related family, member 1|potassium voltage-gated channel, Isk-related subfamily, member 1|voltage gated potassiun channel accessory subunit', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34418715, end_position_on_the_genomic_accession=34512275, orientation='minus', exon_count=9, OMIM=176261, no_hdr0=''),
  3754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3754, CurrentID=0, Status='live', Symbol='KCNF1', Aliases='IK8, KCNF, KV5.1, kH1', description='potassium channel, voltage gated modifier subfamily F, member 1', other_designations='potassium channel KH1|potassium channel Kv5.1|potassium channel, voltage gated subfamily F, member 1|potassium voltage-gated channel, subfamily F, member 1|voltage-gated potassium channel subunit Kv5.1', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10911937, end_position_on_the_genomic_accession=10914225, orientation='plus', exon_count=1, OMIM=603787, no_hdr0=''),
  3755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3755, CurrentID=0, Status='live', Symbol='KCNG1', Aliases='K13, KCNG, KV6.1, kH2', description='potassium channel, voltage gated modifier subfamily G, member 1', other_designations='potassium channel KH2|potassium channel Kv6.1|potassium voltage-gated channel, subfamily G, member 1|voltage-gated potassium channel subunit Kv6.1', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51003656, end_position_on_the_genomic_accession=51023176, orientation='minus', exon_count=5, OMIM=603788, no_hdr0=''),
  3756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3756, CurrentID=0, Status='live', Symbol='KCNH1', Aliases='EAG, EAG1, Kv10.1, TMBTS, h-eag', description='potassium channel, voltage gated eag related subfamily H, member 1', other_designations='EAG channel 1|ether-a-go-go potassium channel 1|ether-a-go-go, Drosophila, homolog of|hEAG1|potassium voltage-gated channel subfamily H member 1 isoform E65|potassium voltage-gated channel subfamily H member 1 isoform E70|potassium voltage-gated channel, subfamily H (eag-related), member 1|voltage-gated potassium channel subunit Kv10.1', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210678314, end_position_on_the_genomic_accession=211134115, orientation='minus', exon_count=12, OMIM=603305, no_hdr0=''),
  3757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3757, CurrentID=0, Status='live', Symbol='KCNH2', Aliases='ERG-1, ERG1, H-ERG, HERG, HERG1, Kv11.1, LQT2, SQT1', description='potassium channel, voltage gated eag related subfamily H, member 2', other_designations='eag homolog|eag-related protein 1|ether-a-go-go-related gene potassium channel 1|ether-a-go-go-related potassium channel protein|ether-a-go-go-related protein 1|potassium voltage-gated channel, subfamily H (eag-related), member 2|voltage-gated potassium channel subunit Kv11.1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150944956, end_position_on_the_genomic_accession=150978314, orientation='minus', exon_count=17, OMIM=152427, no_hdr0=''),
  3758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3758, CurrentID=0, Status='live', Symbol='KCNJ1', Aliases='KIR1.1, ROMK, ROMK1', description='potassium channel, inwardly rectifying subfamily J, member 1', other_designations='ATP-regulated potassium channel ROM-K|inward rectifier K(+) channel Kir1.1|inwardly rectifying K+ channel|potassium channel, inwardly rectifying subfamily J member 1|potassium inwardly-rectifying channel, subfamily J, member 1', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128838014, end_position_on_the_genomic_accession=128867373, orientation='minus', exon_count=5, OMIM=600359, no_hdr0=''),
  3759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3759, CurrentID=0, Status='live', Symbol='KCNJ2', Aliases='ATFB9, HHBIRK1, HHIRK1, IRK1, KIR2.1, LQT7, SQT3', description='potassium channel, inwardly rectifying subfamily J, member 2', other_designations='IRK-1|cardiac inward rectifier potassium channel|hIRK1|inward rectifier K+ channel KIR2.1|potassium inwardly-rectifying channel, subfamily J, member 2', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70168616, end_position_on_the_genomic_accession=70180044, orientation='plus', exon_count=3, OMIM=600681, no_hdr0=''),
  3760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3760, CurrentID=0, Status='live', Symbol='KCNJ3', Aliases='GIRK1, KGA, KIR3.1', description='potassium channel, inwardly rectifying subfamily J, member 3', other_designations='GIRK-1|inward rectifier K(+) channel Kir3.1|inward rectifier K+ channel KIR3.1|potassium channel, inwardly rectifying subfamily J member 3|potassium inwardly-rectifying channel subfamily J member 3 splice variant 1e|potassium inwardly-rectifying channel, subfamily J, member 3', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154698581, end_position_on_the_genomic_accession=154858352, orientation='plus', exon_count=3, OMIM=601534, no_hdr0=''),
  3761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3761, CurrentID=0, Status='live', Symbol='KCNJ4', Aliases='HIR, HIRK2, HRK1, IRK-3, IRK3, Kir2.3', description='potassium channel, inwardly rectifying subfamily J, member 4', other_designations='hippocampal inward rectifier potassium channel|inward rectifier K(+) channel Kir2.3|inward rectifier K+ channel Kir2.3|potassium channel, inwardly rectifying subfamily J member 4|potassium inwardly-rectifying channel, subfamily J, member 4', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38426327, end_position_on_the_genomic_accession=38455199, orientation='minus', exon_count=3, OMIM=600504, no_hdr0=''),
  3762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3762, CurrentID=0, Status='live', Symbol='KCNJ5', Aliases='CIR, GIRK4, KATP1, KIR3.4, LQT13', description='potassium channel, inwardly rectifying subfamily J, member 5', other_designations='IRK-4|cardiac ATP-sensitive potassium channel|heart KATP channel|inward rectifier K+ channel KIR3.4|potassium inwardly-rectifying channel, subfamily J, member 5', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128891418, end_position_on_the_genomic_accession=128921163, orientation='plus', exon_count=5, OMIM=600734, no_hdr0=''),
  3763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3763, CurrentID=0, Status='live', Symbol='KCNJ6', Aliases='BIR1, GIRK-2, GIRK2, KATP-2, KATP2, KCNJ7, KIR3.2, KPLBS, hiGIRK2', description='potassium channel, inwardly rectifying subfamily J, member 6', other_designations='inward rectifier K(+) channel Kir3.2|inward rectifier potassium channel KIR3.2', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37617482, end_position_on_the_genomic_accession=37919731, orientation='minus', exon_count=6, OMIM=600877, no_hdr0=''),
  3764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3764, CurrentID=0, Status='live', Symbol='KCNJ8', Aliases='KIR6.1, uKATP-1', description='potassium channel, inwardly rectifying subfamily J, member 8', other_designations='inward rectifier K(+) channel Kir6.1|inwardly rectifying potassium channel KIR6.1|potassium channel, inwardly rectifying subfamily J member 8|potassium inwardly-rectifying channel, subfamily J, member 8', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21764955, end_position_on_the_genomic_accession=21775502, orientation='minus', exon_count=4, OMIM=600935, no_hdr0=''),
  3765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3765, CurrentID=0, Status='live', Symbol='KCNJ9', Aliases='GIRK3, KIR3.3', description='potassium channel, inwardly rectifying subfamily J, member 9', other_designations='G protein-coupled inward rectifier potassium channel|GIRK-3|inward rectifier K(+) channel Kir3.3|inwardly rectifier K+ channel KIR3.3|potassium channel, inwardly rectifying subfamily J member 9|potassium inwardly-rectifying channel, subfamily J, member 9', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160079953, end_position_on_the_genomic_accession=160090563, orientation='plus', exon_count=5, OMIM=600932, no_hdr0=''),
  3766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3766, CurrentID=0, Status='live', Symbol='KCNJ10', Aliases='BIRK-10, KCNJ13-PEN, KIR1.2, KIR4.1, SESAME', description='potassium channel, inwardly rectifying subfamily J, member 10', other_designations='ATP-dependent inwardly rectifying potassium channel Kir4.1|glial ATP-dependent inwardly rectifying potassium channel KIR4.1|inward rectifier K(+) channel Kir1.2|inward rectifier K+ channel KIR1.2|potassium channel, inwardly rectifying subfamily J member 10|potassium inwardly-rectifying channel, subfamily J, member 10', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160037467, end_position_on_the_genomic_accession=160070261, orientation='minus', exon_count=2, OMIM=602208, no_hdr0=''),
  3767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3767, CurrentID=0, Status='live', Symbol='KCNJ11', Aliases='BIR, HHF2, IKATP, KIR6.2, MODY13, PHHI, TNDM3', description='potassium channel, inwardly rectifying subfamily J, member 11', other_designations='beta-cell inward rectifier subunit|inward rectifier K(+) channel Kir6.2|inwardly rectifying potassium channel KIR6.2|potassium channel inwardly rectifing subfamily J member 11|potassium channel, inwardly rectifying subfamily J member 11|potassium inwardly-rectifying channel, subfamily J, member 11', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17364824, end_position_on_the_genomic_accession=17389331, orientation='minus', exon_count=4, OMIM=600937, no_hdr0=''),
  3768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3768, CurrentID=0, Status='live', Symbol='KCNJ12', Aliases='IRK-2, IRK2, KCNJN1, Kir2.2, Kir2.2v, hIRK, hIRK1, hkir2.2x, kcnj12x', description='potassium channel, inwardly rectifying subfamily J, member 12', other_designations='inward rectifier K(+) channel Kir2.2v|inward rectifier K(+) channel Kir2.6|potassium inwardly-rectifying channel, subfamily J, inhibitor 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21376387, end_position_on_the_genomic_accession=21419872, orientation='plus', exon_count=5, OMIM=602323, no_hdr0=''),
  3769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3769, CurrentID=0, Status='live', Symbol='KCNJ13', Aliases='KIR1.4, KIR7.1, LCA16, SVD', description='potassium channel, inwardly rectifying subfamily J, member 13', other_designations='inward rectifier K(+) channel Kir7.1|potassium inwardly-rectifying channel, subfamily J, member 13', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232765802, end_position_on_the_genomic_accession=232776565, orientation='minus', exon_count=3, OMIM=603208, no_hdr0=''),
  3770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3770, CurrentID=0, Status='live', Symbol='KCNJ14', Aliases='IRK4, KIR2.4', description='potassium channel, inwardly rectifying subfamily J, member 14', other_designations='IRK-4|inward rectifier K(+) channel Kir2.4|inwardly rectifying potassium channel KIR2.4|potassium channel, inwardly rectifying subfamily J member 14|potassium inwardly-rectifying channel, subfamily J, member 14', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48455707, end_position_on_the_genomic_accession=48466110, orientation='plus', exon_count=3, OMIM=603953, no_hdr0=''),
  3772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3772, CurrentID=0, Status='live', Symbol='KCNJ15', Aliases='IRKK, KIR1.3, KIR4.2', description='potassium channel, inwardly rectifying subfamily J, member 15', other_designations='inward rectifier K(+) channel Kir1.3|inward rectifier K(+) channel Kir4.2|inward rectifier K+ channel KIR4.2|potassium channel, inwardly rectifying subfamily J member 15|potassium inwardly-rectifying channel, subfamily J, member 15', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38229915, end_position_on_the_genomic_accession=38303126, orientation='plus', exon_count=10, OMIM=602106, no_hdr0=''),
  3773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3773, CurrentID=0, Status='live', Symbol='KCNJ16', Aliases='BIR9, KIR5.1', description='potassium channel, inwardly rectifying subfamily J, member 16', other_designations='inward rectifier K(+) channel Kir5.1|inward rectifier K+ channel KIR5.1|potassium channel, inwardly rectifying subfamily J member 16|potassium inwardly-rectifying channel, subfamily J, member 16', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70075225, end_position_on_the_genomic_accession=70135608, orientation='plus', exon_count=10, OMIM=605722, no_hdr0=''),
  3775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3775, CurrentID=0, Status='live', Symbol='KCNK1', Aliases='DPK, HOHO, K2P1, K2p1.1, KCNO1, TWIK-1, TWIK1', description='potassium channel, two pore domain subfamily K, member 1', other_designations='inward rectifying potassium channel protein TWIK-1|potassium channel K2P1|potassium channel KCNO1|potassium inwardly-rectifying channel, subfamily K, member 1|tandem of P domains in a weak inward rectifying K+ channel 1', map_location='1q42-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233614004, end_position_on_the_genomic_accession=233672512, orientation='plus', exon_count=6, OMIM=601745, no_hdr0=''),
  3776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3776, CurrentID=0, Status='live', Symbol='KCNK2', Aliases='K2p2.1, TPKC1, TREK, TREK-1, TREK1, hTREK-1c, hTREK-1e', description='potassium channel, two pore domain subfamily K, member 2', other_designations='K2P2.1 potassium channel|TREK-1 K(+) channel subunit|TWIK-related potassium channel 1|outward rectifying potassium channel protein TREK-1|potassium channel subfamily k member 2 variant 1|potassium channel subfamily k member 2 variant 2|potassium inwardly-rectifying channel, subfamily K, member 2|tandem-pore-domain potassium channel TREK-1|two pore domain potassium channel TREK-1|two pore potassium channel TPKC1|two-pore potassium channel 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215002639, end_position_on_the_genomic_accession=215237093, orientation='plus', exon_count=13, OMIM=603219, no_hdr0=''),
  3777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3777, CurrentID=0, Status='live', Symbol='KCNK3', Aliases='K2p3.1, OAT1, PPH4, TASK, TASK-1, TBAK1', description='potassium channel, two pore domain subfamily K, member 3', other_designations='TWIK-related acid-sensitive K(+) channel 1|TWIK-related acid-sensitive K+ channel|acid-sensitive potassium channel protein TASK|acid-sensitive potassium channel protein TASK-1|cardiac potassium channel|potassium inwardly-rectifying channel, subfamily K, member 3|two P domain potassium channel|two pore K(+) channel KT3.1|two pore potassium channel KT3.1', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26692713, end_position_on_the_genomic_accession=26731198, orientation='plus', exon_count=3, OMIM=603220, no_hdr0=''),
  3778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3778, CurrentID=0, Status='live', Symbol='KCNMA1', Aliases='BKTM, KCa1.1, MaxiK, SAKCA, SLO, SLO-ALPHA, SLO1, bA205K10.1, mSLO1', description='potassium channel, calcium activated large conductance subfamily M alpha, member 1', other_designations='BK channel alpha subunit|BKCA alpha subunit|big potassium channel alpha subunit|calcium-activated potassium channel, subfamily M subunit alpha-1|hSlo|k(VCA)alpha|maxi-K channel HSLO|potassium large conductance calcium-activated channel, subfamily M, alpha member 1|slo homolog|slowpoke homolog|stretch-activated Kca channel', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=76869601, end_position_on_the_genomic_accession=77637819, orientation='minus', exon_count=38, OMIM=600150, no_hdr0=''),
  3779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3779, CurrentID=0, Status='live', Symbol='KCNMB1', Aliases='BKbeta1, K(VCA)beta, SLO-BETA, hbeta1, hslo-beta, k(VCA)beta-1, slo-beta-1', description='potassium channel subfamily M regulatory beta subunit 1', other_designations='BK channel beta subunit 1|BK channel subunit beta-1|MaxiK channel beta-subunit 1|big potassium channel beta subunit 1|calcium-activated potassium channel, subfamily M subunit beta-1|charybdotoxin receptor subunit beta-1|large conductance Ca2+-activated K+ channel beta 1 subunit|maxi K channel subunit beta-1|potassium large conductance calcium-activated channel, subfamily M, beta member 1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170378161, end_position_on_the_genomic_accession=170389677, orientation='minus', exon_count=4, OMIM=603951, no_hdr0=''),
  3780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3780, CurrentID=0, Status='live', Symbol='KCNN1', Aliases='KCa2.1, SK1, SKCA1, hSK1', description='potassium channel, calcium activated intermediate/small conductance subfamily N alpha, member 1', other_designations='potassium intermediate/small conductance calcium-activated channel, subfamily N, member 1|small conductance calcium-activated potassium channel 1', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17951302, end_position_on_the_genomic_accession=17999121, orientation='plus', exon_count=14, OMIM=602982, no_hdr0=''),
  3781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3781, CurrentID=0, Status='live', Symbol='KCNN2', Aliases='KCa2.2, SK2, SKCA2, SKCa 2, hSK2', description='potassium channel, calcium activated intermediate/small conductance subfamily N alpha, member 2', other_designations='apamin-sensitive small-conductance Ca2+-activated potassium channel|potassium intermediate/small conductance calcium-activated channel, subfamily N, member 2|small conductance calcium-activated potassium channel 2', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=114056006, end_position_on_the_genomic_accession=114496500, orientation='plus', exon_count=15, OMIM=605879, no_hdr0=''),
  3782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3782, CurrentID=0, Status='live', Symbol='KCNN3', Aliases='KCa2.3, SK3, SKCA3, hSK3', description='potassium channel, calcium activated intermediate/small conductance subfamily N alpha, member 3', other_designations='SKCa 3|potassium intermediate/small conductance calcium-activated channel, subfamily N, member 3|small conductance calcium-activated potassium channel 3', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154697462, end_position_on_the_genomic_accession=154870278, orientation='minus', exon_count=10, OMIM=602983, no_hdr0=''),
  3783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3783, CurrentID=0, Status='live', Symbol='KCNN4', Aliases='IK, IK1, IKCA1, KCA4, KCa3.1, SK4, hIKCa1, hKCa4, hSK4', description='potassium channel, calcium activated intermediate/small conductance subfamily N alpha, member 4', other_designations='SKCa 4|SKCa4|potassium intermediate/small conductance calcium-activated channel, subfamily N, member 4|putative Gardos channel|putative erythrocyte intermediate conductance calcium-activated potassium Gardos channel|small conductance calcium-activated potassium channel 4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43766533, end_position_on_the_genomic_accession=43782139, orientation='minus', exon_count=10, OMIM=602754, no_hdr0=''),
  3784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3784, CurrentID=0, Status='live', Symbol='KCNQ1', Aliases='ATFB1, ATFB3, JLNS1, KCNA8, KCNA9, KVLQT1, Kv1.9, Kv7.1, LQT, LQT1, RWS, SQT2, WRS', description='potassium channel, voltage gated KQT-like subfamily Q, member 1', other_designations='IKs producing slow voltage-gated potassium channel subunit alpha KvLQT1|kidney and cardiac voltage dependend K+ channel|potassium voltage-gated channel, KQT-like subfamily, member 1|slow delayed rectifier channel subunit|voltage-gated potassium channel subunit Kv7.1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2444991, end_position_on_the_genomic_accession=2849110, orientation='plus', exon_count=17, OMIM=607542, no_hdr0=''),
  3785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3785, CurrentID=0, Status='live', Symbol='KCNQ2', Aliases='BFNC, BFNS1, EBN, EBN1, EIEE7, ENB1, HNSPC, KCNA11, KV7.2, KVEBN1', description='potassium channel, voltage gated KQT-like subfamily Q, member 2', other_designations='KQT-like 2|neuroblastoma-specific potassium channel protein|neuroblastoma-specific potassium channel subunit alpha KvLQT2|potassium voltage-gated channel, KQT-like subfamily, member 2|voltage-gated potassium channel subunit Kv7.2', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63400208, end_position_on_the_genomic_accession=63472640, orientation='minus', exon_count=21, OMIM=602235, no_hdr0=''),
  3786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3786, CurrentID=0, Status='live', Symbol='KCNQ3', Aliases='BFNC2, EBN2, KV7.3', description='potassium channel, voltage gated KQT-like subfamily Q, member 3', other_designations='potassium channel subunit alpha KvLQT3|potassium channel, voltage-gated, subfamily Q, member 3|potassium voltage-gated channel, KQT-like subfamily, member 3|voltage-gated potassium channel subunit Kv7.3', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132120858, end_position_on_the_genomic_accession=132480757, orientation='minus', exon_count=17, OMIM=602232, no_hdr0=''),
  3787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3787, CurrentID=0, Status='live', Symbol='KCNS1', Aliases='KV9.1', description='potassium voltage-gated channel, modifier subfamily S, member 1', other_designations='delayed-rectifier K(+) channel alpha subunit 1|delayed-rectifier K+ channel alpha subunit 1|potassium voltage-gated channel, delayed-rectifier, subfamily S, member 1|voltage-gated potassium channel protein Kv9.1|voltage-gated potassium channel subunit Kv9.1', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45092072, end_position_on_the_genomic_accession=45101112, orientation='minus', exon_count=5, OMIM=602905, no_hdr0=''),
  134860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134860, CurrentID=0, Status='live', Symbol='TAAR9', Aliases='TA3, TAR3, TAR9, TRAR3', description='trace amine associated receptor 9 (gene/pseudogene)', other_designations='trace amine receptor 3|trace amine receptor 9', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132538288, end_position_on_the_genomic_accession=132539336, orientation='plus', exon_count=1, OMIM=608282, no_hdr0=''),
  101060301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060301, CurrentID=0, Status='live', Symbol='HNRNPCL4', Aliases='', description='heterogeneous nuclear ribonucleoprotein C-like 4', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13164555, end_position_on_the_genomic_accession=13165482, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3790, CurrentID=0, Status='live', Symbol='KCNS3', Aliases='KV9.3', description='potassium voltage-gated channel, modifier subfamily S, member 3', other_designations='Shab-related delayed-rectifier K+ channel alpha subunit 3|delayed-rectifier K(+) channel alpha subunit 3|potassium voltage-gated channel delayed-rectifier protein S3|potassium voltage-gated channel, delayed-rectifier, subfamily S, member 3|voltage-gated potassium channel protein Kv9.3|voltage-gated potassium channel subunit Kv9.3', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17877847, end_position_on_the_genomic_accession=17932961, orientation='plus', exon_count=5, OMIM=603888, no_hdr0=''),
  3791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3791, CurrentID=0, Status='live', Symbol='KDR', Aliases='CD309, FLK1, VEGFR, VEGFR2', description='kinase insert domain receptor', other_designations='fetal liver kinase 1|fetal liver kinase-1|kinase insert domain receptor (a type III receptor tyrosine kinase)|protein-tyrosine kinase receptor Flk-1|soluble VEGFR2|tyrosine kinase growth factor receptor', map_location='4q11-q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55078259, end_position_on_the_genomic_accession=55125595, orientation='minus', exon_count=30, OMIM=191306, no_hdr0=''),
  3792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3792, CurrentID=0, Status='live', Symbol='KEL', Aliases='CD238, ECE3', description='Kell blood group, metallo-endopeptidase', other_designations='kell blood group antigen', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142941114, end_position_on_the_genomic_accession=142962416, orientation='minus', exon_count=20, OMIM=613883, no_hdr0=''),
  401662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401662, CurrentID=0, Status='live', Symbol='OR51A8P', Aliases='', description='olfactory receptor family 51 subfamily A member 8 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 8 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4832133, end_position_on_the_genomic_accession=4833069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3795, CurrentID=0, Status='live', Symbol='KHK', Aliases='', description='ketohexokinase', other_designations='fructokinase|hepatic fructokinase|ketohexokinase (fructokinase)', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27086743, end_position_on_the_genomic_accession=27100751, orientation='plus', exon_count=9, OMIM=614058, no_hdr0=''),
  3796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3796, CurrentID=0, Status='live', Symbol='KIF2A', Aliases='CDCBM3, HK2, KIF2', description='kinesin heavy chain member 2A', other_designations='Kinesin, heavy chain, 2|kinesin-2', map_location='5q12-q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62306162, end_position_on_the_genomic_accession=62387184, orientation='plus', exon_count=22, OMIM=602591, no_hdr0=''),
  3797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3797, CurrentID=0, Status='live', Symbol='KIF3C', Aliases='', description='kinesin family member 3C', other_designations='KIF3C variant protein', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25926586, end_position_on_the_genomic_accession=25982574, orientation='minus', exon_count=8, OMIM=602845, no_hdr0=''),
  3798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3798, CurrentID=0, Status='live', Symbol='KIF5A', Aliases='D12S1889, MY050, NKHC, SPG10', description='kinesin family member 5A', other_designations='KIF5A variant protein|kinesin heavy chain neuron-specific 1|kinesin, heavy chain, neuron-specific|neuronal kinesin heavy chain', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57550064, end_position_on_the_genomic_accession=57584771, orientation='plus', exon_count=29, OMIM=602821, no_hdr0=''),
  3799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3799, CurrentID=0, Status='live', Symbol='KIF5B', Aliases='HEL-S-61, KINH, KNS, KNS1, UKHC', description='kinesin family member 5B', other_designations='conventional kinesin heavy chain|epididymis secretory protein Li 61|kinesin 1 (110-120kD)|kinesin heavy chain|ubiquitous kinesin heavy chain', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32009010, end_position_on_the_genomic_accession=32056443, orientation='minus', exon_count=26, OMIM=602809, no_hdr0=''),
  3800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3800, CurrentID=0, Status='live', Symbol='KIF5C', Aliases='CDCBM2, KINN, NKHC, NKHC-2, NKHC2', description='kinesin family member 5C', other_designations='kinesin heavy chain neuron-specific 2', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148875223, end_position_on_the_genomic_accession=149026759, orientation='plus', exon_count=28, OMIM=604593, no_hdr0=''),
  3801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3801, CurrentID=0, Status='live', Symbol='KIFC3', Aliases='', description='kinesin family member C3', other_designations='', map_location='16q13-q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57758217, end_position_on_the_genomic_accession=57863053, orientation='minus', exon_count=31, OMIM=604535, no_hdr0=''),
  3802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3802, CurrentID=0, Status='live', Symbol='KIR2DL1', Aliases='CD158A, KIR-K64, KIR221, NKAT, NKAT-1, NKAT1, p58.1', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 1', other_designations='CD158 antigen-like family member A|MHC class I NK cell receptor|killer Ig receptor|killer inhibitory receptor 2-2-1|natural killer-associated transcript 1|p58 NK cell inhibitory receptor NKR-K6|p58 NK receptor CL-42/47.11|p58 killer cell inhibitory receptor KIR-K64|p58 natural killer cell receptor clones CL-42/47.11|p58.1 MHC class-I-specific NK receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54769805, end_position_on_the_genomic_accession=54784326, orientation='plus', exon_count=9, OMIM=604936, no_hdr0=''),
  3803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3803, CurrentID=0, Status='live', Symbol='KIR2DL2', Aliases='CD158B1, CD158b, NKAT-6, NKAT6, p58.2', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 2', other_designations='CD158 antigen-like family member B1|MHC class I NK cell receptor|natural killer-associated transcript 6|p58 NK receptor CL-43|p58 natural killer cell receptor clone CL-43', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604937, no_hdr0=''),
  3804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3804, CurrentID=0, Status='live', Symbol='KIR2DL3', Aliases='CD158B2, CD158b, GL183, KIR-023GB, KIR-K7b, KIR-K7c, KIR2DS5, KIRCL23, NKAT, NKAT2, NKAT2A, NKAT2B, p58', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 3', other_designations='CD158 antigen-like family member B2|MHC class I NK cell receptor|NK-receptor|NKAT-2|killer Ig-like inhibitory receptor|killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 5|killer inhibitory receptor cl 2-3|natural killer associated transcript 2|natural killer cell inhibitory receptor KIR2DL3|natural killer-associated transcript 2|p58 NK receptor CL-6|p58 natural killer cell receptor clone CL-6|p58.2 MHC class-I specific NK receptor|p58.2 MHC class-I-specific NK receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54738509, end_position_on_the_genomic_accession=54753052, orientation='plus', exon_count=8, OMIM=604938, no_hdr0=''),
  3805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3805, CurrentID=0, Status='live', Symbol='KIR2DL4', Aliases='CD158D, G9P, KIR-103AS, KIR103, KIR103AS', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 4', other_designations='CD158 antigen-like family member D|MHC class I NK cell receptor KIR103AS|NK cell receptor|killer Ig receptor|killer cell inhibitory receptor 103AS|natural killer cell inhibitory receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54803612, end_position_on_the_genomic_accession=54814517, orientation='plus', exon_count=8, OMIM=604945, no_hdr0=''),
  3806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3806, CurrentID=0, Status='live', Symbol='KIR2DS1', Aliases='CD158H, CD158a, KIR2DP1DL1, KIR2DS4, p50.1', description='killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 1', other_designations='CD158 antigen-like family member H|MHC class I NK cell receptor Eb6 ActI|killer cell immunoglobulin-like receptor KIRDS1|killer-cell immunoglobulin-like receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604952, no_hdr0=''),
  3808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3808, CurrentID=0, Status='live', Symbol='KIR2DS3', Aliases='NKAT7', description='killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 3', other_designations='MHC class I NK cell receptor|NKAT-7|natural killer cell inhibitory receptor|natural killer-associated transcript 7', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604954, no_hdr0=''),
  3809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3809, CurrentID=0, Status='live', Symbol='KIR2DS4', Aliases='CD158I, KIR1D, KIR2DS1, KIR412, KKA3, NKAT-8, NKAT8', description='killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 4', other_designations='CD158 antigen-like family member I|KIR antigen 2DS4|MHC class I NK cell receptor|P58 natural killer cell receptor clones CL-39/CL-17|killer Ig receptor|killer cell immunoglobulin-like receptor two domains short cytoplasmic tail 1|killer inhibitory receptor 4-1-2|natural killer cell inhibitory receptor|natural killer-associated transcript 8|p50 killer cell activating receptor KAR-K1d|p58 NK receptor CL-39/CL-17', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54832676, end_position_on_the_genomic_accession=54848569, orientation='plus', exon_count=8, OMIM=604955, no_hdr0=''),
  3810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3810, CurrentID=0, Status='live', Symbol='KIR2DS5', Aliases='CD158G, NKAT9', description='killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 5', other_designations='CD158 antigen-like family member G|MHC class I NK cell receptor|NKAT-9|natural killer cell inhibitory receptor|natural killer-associated transcript 9', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604956, no_hdr0=''),
  3811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3811, CurrentID=0, Status='live', Symbol='KIR3DL1', Aliases='CD158E1, KIR/S1, KIR3DL2, NKAT-3, NKAT3, NKB1, NKB1B, KIR3DL1', description='killer cell immunoglobulin-like receptor, three domains, long cytoplasmic tail, 1', other_designations='CD158 antigen-like family member E|HLA-BW4-specific inhibitory NK cell receptor|KIR antigen 3DL1|MHC class I NK cell receptor|NK-receptor|killer Ig receptor|killer cell immunoglobulin-like receptor three domains long cytoplasmic tail 2|killer cell immunoglobulin-like receptor three domains short cytoplasmic tail 1|natural killer-associated transcript 3|p70 NK receptor CL-2/CL-11|p70 killer cell inhibitory receptor|p70 natural killer cell receptor clones CL-2/CL-11', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54816438, end_position_on_the_genomic_accession=54830778, orientation='plus', exon_count=9, OMIM=604946, no_hdr0=''),
  3812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3812, CurrentID=0, Status='live', Symbol='KIR3DL2', Aliases='CD158K, NKAT-4, NKAT4, NKAT4B, p140', description='killer cell immunoglobulin-like receptor, three domains, long cytoplasmic tail, 2', other_designations='CD158 antigen-like family member K|KIR antigen 3DL2|MHC class I NK cell receptor|killer Ig receptor|natural killer-associated transcript 4|p70 NK receptor CL-5|p70 killer cell inhibitory receptor|p70 natural killer cell receptor clone CL-5', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54850443, end_position_on_the_genomic_accession=54867215, orientation='plus', exon_count=9, OMIM=604947, no_hdr0=''),
  3813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3813, CurrentID=0, Status='live', Symbol='KIR3DS1', Aliases='CD158E2, KIR-123FM, KIR-G1, NKAT-10, NKAT10', description='killer cell immunoglobulin-like receptor, three domains, short cytoplasmic tail, 1', other_designations='MHC class I NK cell receptor|killer Ig receptor|killer cell immunoglobulin-like receptor KIR3DS1|natural killer-associated transcript 10', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3814, CurrentID=0, Status='live', Symbol='KISS1', Aliases='HH13, KiSS-1', description='KiSS-1 metastasis-suppressor', other_designations='kisspeptin-1|malignant melanoma metastasis-suppressor|metastin|prepro-kisspeptin', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204190341, end_position_on_the_genomic_accession=204196491, orientation='minus', exon_count=4, OMIM=603286, no_hdr0=''),
  3815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3815, CurrentID=0, Status='live', Symbol='KIT', Aliases='C-Kit, CD117, PBT, SCFR', description='v-kit Hardy-Zuckerman 4 feline sarcoma viral oncogene homolog', other_designations='p145 c-kit|piebald trait protein|proto-oncogene c-Kit|proto-oncogene tyrosine-protein kinase Kit|soluble KIT variant 1|tyrosine-protein kinase Kit|v-kit Hardy-Zuckerman 4 feline sarcoma viral oncogene-like protein', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54657928, end_position_on_the_genomic_accession=54740715, orientation='plus', exon_count=21, OMIM=164920, no_hdr0=''),
  3816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3816, CurrentID=0, Status='live', Symbol='KLK1', Aliases='KLKR, Klk6, hK1', description='kallikrein 1', other_designations='glandular kallikrein 1|kallikrein 1, renal/pancreas/salivary|kallikrein serine protease 1|kidney/pancreas/salivary gland kallikrein|tissue kallikrein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50819146, end_position_on_the_genomic_accession=50823787, orientation='minus', exon_count=6, OMIM=147910, no_hdr0=''),
  3817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3817, CurrentID=0, Status='live', Symbol='KLK2', Aliases='KLK2A2, hGK-1, hK2', description='kallikrein related peptidase 2', other_designations='glandular kallikrein 2|glandular kallikrein-1|kallikrein 2, prostatic|kallikrein-related peptidase 2|tissue kallikrein-2', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50873433, end_position_on_the_genomic_accession=50880567, orientation='plus', exon_count=7, OMIM=147960, no_hdr0=''),
  3818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3818, CurrentID=0, Status='live', Symbol='KLKB1', Aliases='KLK3, PKK, PKKD, PPK', description='kallikrein B1', other_designations='Fletcher factor|kallikrein B, plasma (Fletcher factor) 1|kininogenin|plasma kallikrein heavy chain|plasma kallikrein light chain|plasma prekallikrein', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186210753, end_position_on_the_genomic_accession=186258474, orientation='plus', exon_count=18, OMIM=229000, no_hdr0=''),
  3820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3820, CurrentID=0, Status='live', Symbol='KLRB1', Aliases='CD161, CLEC5B, NKR, NKR-P1, NKR-P1A, NKRP1A, hNKR-P1A', description='killer cell lectin-like receptor subfamily B, member 1', other_designations='C-type lectin domain family 5 member B|natural killer cell surface protein P1A', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9595274, end_position_on_the_genomic_accession=9607901, orientation='minus', exon_count=6, OMIM=602890, no_hdr0=''),
  3821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3821, CurrentID=0, Status='live', Symbol='KLRC1', Aliases='CD159A, NKG2, NKG2A', description='killer cell lectin-like receptor subfamily C, member 1', other_designations='C-lectin type II protein|CD159 antigen-like family member A|NK cell receptor A|NKG2-1/B activating NK receptor|NKG2-A/B type II integral membrane protein|NKG2-A/B-activating NK receptor|natural killer cell lectin|natural killer group protein 2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10442264, end_position_on_the_genomic_accession=10454685, orientation='minus', exon_count=9, OMIM=161555, no_hdr0=''),
  3822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3822, CurrentID=0, Status='live', Symbol='KLRC2', Aliases='CD159c, NKG2-C, NKG2C', description='killer cell lectin-like receptor subfamily C, member 2', other_designations='CD159 antigen-like family member C|NK cell receptor C|NKG2-C-activating NK receptor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10430599, end_position_on_the_genomic_accession=10435993, orientation='minus', exon_count=6, OMIM=602891, no_hdr0=''),
  3823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3823, CurrentID=0, Status='live', Symbol='KLRC3', Aliases='NKG2-E, NKG2E', description='killer cell lectin-like receptor subfamily C, member 3', other_designations='NK cell receptor E|NKG2-E-activating NK receptor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10412315, end_position_on_the_genomic_accession=10420595, orientation='minus', exon_count=7, OMIM=602892, no_hdr0=''),
  3824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3824, CurrentID=0, Status='live', Symbol='KLRD1', Aliases='CD94', description='killer cell lectin-like receptor subfamily D, member 1', other_designations='CD94 antigen|KP43|NK cell receptor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10304446, end_position_on_the_genomic_accession=10360876, orientation='plus', exon_count=13, OMIM=602894, no_hdr0=''),
  3827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3827, CurrentID=0, Status='live', Symbol='KNG1', Aliases='BDK, BK, KNG', description='kininogen 1', other_designations='HMWK|alpha-2-thiol proteinase inhibitor|bradykinin|fitzgerald factor|high molecular weight kininogen|williams-Fitzgerald-Flaujeac factor', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186717309, end_position_on_the_genomic_accession=186744410, orientation='plus', exon_count=11, OMIM=612358, no_hdr0=''),
  103504726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504726, CurrentID=0, Status='live', Symbol='MIR3670-3', Aliases='hsa-mir-3670-3', description='microRNA 3670-3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18405698, end_position_on_the_genomic_accession=18405762, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101060341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060341, CurrentID=0, Status='live', Symbol='LOC101060341', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55746453, end_position_on_the_genomic_accession=55750700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3831, CurrentID=0, Status='live', Symbol='KLC1', Aliases='KLC, KNS2, KNS2A', description='kinesin light chain 1', other_designations='kinesin 2 60/70kDa|medulloblastoma antigen MU-MB-2.50', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103629188, end_position_on_the_genomic_accession=103701551, orientation='plus', exon_count=18, OMIM=600025, no_hdr0=''),
  3832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3832, CurrentID=0, Status='live', Symbol='KIF11', Aliases='EG5, HKSP, KNSL1, MCLMR, TRIP5', description='kinesin family member 11', other_designations='TR-interacting protein 5|TRIP-5|kinesin-like protein 1|kinesin-like spindle protein HKSP|kinesin-related motor protein Eg5|thyroid receptor-interacting protein 5', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92593068, end_position_on_the_genomic_accession=92655395, orientation='plus', exon_count=22, OMIM=148760, no_hdr0=''),
  3833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3833, CurrentID=0, Status='live', Symbol='KIFC1', Aliases='HSET, KNSL2', description='kinesin family member C1', other_designations='kinesin-like 2|kinesin-like protein 2|kinesin-related protein HSET', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33391524, end_position_on_the_genomic_accession=33410226, orientation='plus', exon_count=12, OMIM=603763, no_hdr0=''),
  3834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3834, CurrentID=0, Status='live', Symbol='KIF25', Aliases='KNSL3', description='kinesin family member 25', other_designations='kinesin-like 3|kinesin-like protein 3', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168002543, end_position_on_the_genomic_accession=168045089, orientation='plus', exon_count=11, OMIM=603815, no_hdr0=''),
  3835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3835, CurrentID=0, Status='live', Symbol='KIF22', Aliases='A-328A3.2, KID, KNSL4, OBP, OBP-1, OBP-2, SEMDJL2', description='kinesin family member 22', other_designations='kinesin-like DNA-binding protein pseudogene|kinesin-like protein 4|oriP binding protein|origin of plasmid DNA replication-binding protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29790062, end_position_on_the_genomic_accession=29805385, orientation='plus', exon_count=16, OMIM=603213, no_hdr0=''),
  3836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3836, CurrentID=0, Status='live', Symbol='KPNA1', Aliases='IPOA5, NPI-1, RCH2, SRP1', description='karyopherin alpha 1', other_designations='RAG cohort protein 2|SRP1-beta|importin alpha 5|importin subunit alpha-1|importin-alpha-S1|karyopherin alpha 1 (importin alpha 5)|karyopherin subunit alpha-1|nucleoprotein interactor 1|recombination activating gene cohort 2', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122421901, end_position_on_the_genomic_accession=122514939, orientation='minus', exon_count=18, OMIM=600686, no_hdr0=''),
  3837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3837, CurrentID=0, Status='live', Symbol='KPNB1', Aliases='IMB1, IPO1, IPOB, Impnb, NTF97', description='karyopherin (importin) beta 1', other_designations='PTAC97|importin 1|importin 90|importin beta-1 subunit|importin-90|karyopherin subunit beta-1|nuclear factor p97|pore targeting complex 97 kDa subunit', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47649838, end_position_on_the_genomic_accession=47683638, orientation='plus', exon_count=23, OMIM=602738, no_hdr0=''),
  3838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3838, CurrentID=0, Status='live', Symbol='KPNA2', Aliases='IPOA1, QIP2, RCH1, SRP1alpha', description='karyopherin alpha 2 (RAG cohort 1, importin alpha 1)', other_designations='RAG cohort 1|RAG cohort protein 1|SRP1-alpha|importin alpha 1|importin alpha 2|importin subunit alpha-2|importin-alpha-P1|karyopherin subunit alpha-2|pendulin', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68035732, end_position_on_the_genomic_accession=68046854, orientation='plus', exon_count=12, OMIM=600685, no_hdr0=''),
  3839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3839, CurrentID=0, Status='live', Symbol='KPNA3', Aliases='IPOA4, SRP1, SRP1gamma, SRP4, hSRP1', description='karyopherin alpha 3 (importin alpha 4)', other_designations='SRP1-gamma|importin alpha 4|importin alpha Q2|importin alpha-3|importin subunit alpha-3|importin-alpha-Q2|karyopherin subunit alpha-3|qip2', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49699307, end_position_on_the_genomic_accession=49792921, orientation='minus', exon_count=17, OMIM=601892, no_hdr0=''),
  3840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3840, CurrentID=0, Status='live', Symbol='KPNA4', Aliases='IPOA3, QIP1, SRP3', description='karyopherin alpha 4 (importin alpha 3)', other_designations='importin alpha Q1|importin subunit alpha-4|karyopherin subunit alpha-4', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160494995, end_position_on_the_genomic_accession=160565588, orientation='minus', exon_count=17, OMIM=602970, no_hdr0=''),
  3841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3841, CurrentID=0, Status='live', Symbol='KPNA5', Aliases='IPOA6, SRP6', description='karyopherin alpha 5 (importin alpha 6)', other_designations='importin alpha 6|karyopherin subunit alpha-5', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116681086, end_position_on_the_genomic_accession=116741867, orientation='plus', exon_count=17, OMIM=604545, no_hdr0=''),
  3842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3842, CurrentID=0, Status='live', Symbol='TNPO1', Aliases='IPO2, KPNB2, MIP, MIP1, TRN', description='transportin 1', other_designations='M9 region interaction protein|importin 2|importin beta 2|importin beta-2|karyopherin (importin) beta 2|karyopherin beta-2', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72816591, end_position_on_the_genomic_accession=72914388, orientation='plus', exon_count=26, OMIM=602901, no_hdr0=''),
  3843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3843, CurrentID=0, Status='live', Symbol='IPO5', Aliases='IMB3, KPNB3, Pse1, RANBP5, imp5', description='importin 5', other_designations='RAN binding protein 5|Ran_GTP binding protein 5|importin beta-3 subunit|importin subunit beta-3|karyopherin (importin) beta 3|karyopherin beta-3|ran-binding protein 5', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97953638, end_position_on_the_genomic_accession=98024296, orientation='plus', exon_count=31, OMIM=602008, no_hdr0=''),
  3844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3844, CurrentID=0, Status='live', Symbol='KRASP1', Aliases='KRAS1P, c-Kras1', description='Kirsten rat sarcoma viral oncogene homolog pseudogene 1 (functional)', other_designations='Kirsten rat sarcoma-1 viral (v-Ki-ras1) oncogene homolog, processed|v-Ki-ras1 Kirsten rat sarcoma 1 viral oncogene homolog, processed pseudogene|v-Ki-ras2 Kirsten rat sarcoma viral oncogene homolog pseudogene 1', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54770474, end_position_on_the_genomic_accession=54775731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3845, CurrentID=0, Status='live', Symbol='KRAS', Aliases='C-K-RAS, CFC2, K-RAS2A, K-RAS2B, K-RAS4A, K-RAS4B, KI-RAS1, KRAS2, NS, NS3, RALD, RASK2, KRAS', description='Kirsten rat sarcoma viral oncogene homolog', other_designations='K-Ras 2|K-ras p21 protein|PR310 c-K-ras oncogene|c-Ki-ras|c-Kirsten-ras protein|cellular c-Ki-ras2 proto-oncogene|oncogene KRAS2|transforming protein p21|v-Ki-ras2 Kirsten rat sarcoma 2 viral oncogene homolog', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25204789, end_position_on_the_genomic_accession=25250931, orientation='minus', exon_count=6, OMIM=190070, no_hdr0=''),
  3846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3846, CurrentID=0, Status='live', Symbol='KRTAP5-9', Aliases='KRN1, KRTAP5-1, KRTAP5.9', description='keratin associated protein 5-9', other_designations='UHS KerA|UHS keratin A|keratin, cuticle, ultrahigh sulfur 1|keratin, cuticle, ultrahigh sulphur 1|keratin, ultra high-sulfur matrix protein A|keratin-associated protein 5.9|ultrahigh sulfur keratin-associated protein 5.9', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71548420, end_position_on_the_genomic_accession=71549607, orientation='plus', exon_count=1, OMIM=148021, no_hdr0=''),
  3848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3848, CurrentID=0, Status='live', Symbol='KRT1', Aliases='CK1, EHK, EHK1, EPPK, K1A, NEPPK, KRT1', description='keratin 1, type II', other_designations='67 kDa cytokeratin|CK-1|cytokeratin 1|cytokeratin-1|epidermolytic hyperkeratosis 1|hair alpha protein|type-II keratin Kb1', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52674736, end_position_on_the_genomic_accession=52680407, orientation='minus', exon_count=9, OMIM=139350, no_hdr0=''),
  3849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3849, CurrentID=0, Status='live', Symbol='KRT2', Aliases='CK-2e, K2eA, KRT2E, KRTE, KRT2', description='keratin 2, type II', other_designations='cytokeratin-2e|epithelial keratin-2e|keratin-2 epidermis|keratin-2e|type-II keratin Kb2', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52644558, end_position_on_the_genomic_accession=52652175, orientation='minus', exon_count=9, OMIM=600194, no_hdr0=''),
  3850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3850, CurrentID=0, Status='live', Symbol='KRT3', Aliases='CK3, K3', description='keratin 3, type II', other_designations='65 kDa cytokeratin|CK-3|cytokeratin 3|cytokeratin-3|type-II keratin Kb3', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52789685, end_position_on_the_genomic_accession=52805717, orientation='minus', exon_count=10, OMIM=148043, no_hdr0=''),
  3851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3851, CurrentID=0, Status='live', Symbol='KRT4', Aliases='CK-4, CK4, CYK4, K4, WSN1', description='keratin 4, type II', other_designations='cytokeratin 4|type-II keratin Kb4', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52806543, end_position_on_the_genomic_accession=52814116, orientation='minus', exon_count=9, OMIM=123940, no_hdr0=''),
  3852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3852, CurrentID=0, Status='live', Symbol='KRT5', Aliases='CK5, DDD, DDD1, EBS2, K5A, KRT5', description='keratin 5, type II', other_designations='58 kda cytokeratin|CK-5|cytokeratin-5|epidermolysis bullosa simplex 2 Dowling-Meara/Kobner/Weber-Cockayne types|keratin 5 (epidermolysis bullosa simplex, Dowling-Meara/Kobner/Weber-Cockayne types)|type-II keratin Kb5', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52514575, end_position_on_the_genomic_accession=52520459, orientation='minus', exon_count=9, OMIM=148040, no_hdr0=''),
  3853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3853, CurrentID=0, Status='live', Symbol='KRT6A', Aliases='CK-6C, CK-6E, CK6A, CK6C, CK6D, K6A, K6C, K6D, KRT6C, KRT6D, PC3', description='keratin 6A, type II', other_designations='KDAMP|cytokeratin 6A|cytokeratin 6C|cytokeratin 6D|cytokeratin-6C|cytokeratin-6E|kamp-keratin derived antimicrobial peptide|keratin 6A, , type II|keratin K6h|keratin, epidermal type II, K6A|keratin, type II cytoskeletal 6A|keratin-6C|type-II keratin Kb12|type-II keratin Kb6', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52487174, end_position_on_the_genomic_accession=52493397, orientation='minus', exon_count=9, OMIM=148041, no_hdr0=''),
  3854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3854, CurrentID=0, Status='live', Symbol='KRT6B', Aliases='CK-6B, CK6B, K6B, KRTL1, PC2, PC4', description='keratin 6B, type II', other_designations='cytokeratin 6B|cytokeratin-6B|keratin, epidermal, type II, K6B|keratin-6B|keratin-like 1 (a type II keratin sequence)|type-II keratin Kb10', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52446651, end_position_on_the_genomic_accession=52452126, orientation='minus', exon_count=9, OMIM=148042, no_hdr0=''),
  3855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3855, CurrentID=0, Status='live', Symbol='KRT7', Aliases='CK7, K2C7, K7, SCL', description='keratin 7, type II', other_designations='CK-7|cytokeratin 7|cytokeratin-7|keratin, 55K type II cytoskeletal|keratin, simple epithelial type I, K7|keratin-7|sarcolectin|type II mesothelial keratin K7|type-II keratin Kb7', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52233170, end_position_on_the_genomic_accession=52252667, orientation='plus', exon_count=10, OMIM=148059, no_hdr0=''),
  3856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3856, CurrentID=0, Status='live', Symbol='KRT8', Aliases='CARD2, CK-8, CK8, CYK8, K2C8, K8, KO', description='keratin 8, type II', other_designations='cytokeratin 8|cytokeratin-8|type-II keratin Kb8', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52897187, end_position_on_the_genomic_accession=52949866, orientation='minus', exon_count=10, OMIM=148060, no_hdr0=''),
  3857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3857, CurrentID=0, Status='live', Symbol='KRT9', Aliases='CK-9, EPPK, K9', description='keratin 9, type I', other_designations='cytokeratin 9|cytokeratin-9|keratin-9|type I cytoskeletal 9', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41565841, end_position_on_the_genomic_accession=41572058, orientation='minus', exon_count=8, OMIM=607606, no_hdr0=''),
  3858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3858, CurrentID=0, Status='live', Symbol='KRT10', Aliases='BCIE, BIE, CK10, EHK, K10, KPP', description='keratin 10, type I', other_designations='CK-10|cytokeratin 10|keratin-10', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40818117, end_position_on_the_genomic_accession=40822618, orientation='minus', exon_count=8, OMIM=148080, no_hdr0=''),
  3859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3859, CurrentID=0, Status='live', Symbol='KRT12', Aliases='K12', description='keratin 12, type I', other_designations='CK-12|cytokeratin-12|keratin-12', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40861178, end_position_on_the_genomic_accession=40867210, orientation='minus', exon_count=8, OMIM=601687, no_hdr0=''),
  3860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3860, CurrentID=0, Status='live', Symbol='KRT13', Aliases='CK13, K13, WSN2', description='keratin 13, type I', other_designations='CK-13|cytokeratin 13|cytokeratin-13|keratin-13', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41500981, end_position_on_the_genomic_accession=41505613, orientation='minus', exon_count=8, OMIM=148065, no_hdr0=''),
  3861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3861, CurrentID=0, Status='live', Symbol='KRT14', Aliases='CK14, EBS3, EBS4, K14, NFJ', description='keratin 14, type I', other_designations='cytokeratin 14|cytokeratin-14|keratin-14', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41582279, end_position_on_the_genomic_accession=41586895, orientation='minus', exon_count=8, OMIM=148066, no_hdr0=''),
  101060376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060376, CurrentID=0, Status='live', Symbol='TBC1D3L', Aliases='', description='TBC1 domain family member 3L', other_designations='TBC1 domain family, member 3L', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37978155, end_position_on_the_genomic_accession=37989062, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3866, CurrentID=0, Status='live', Symbol='KRT15', Aliases='CK15, K15, K1CO', description='keratin 15, type I', other_designations='CK-15|cytokeratin 15|cytokeratin-15|keratin-15, basic|keratin-15, beta|type I cytoskeletal 15', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41513745, end_position_on_the_genomic_accession=41522413, orientation='minus', exon_count=9, OMIM=148030, no_hdr0=''),
  3868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3868, CurrentID=0, Status='live', Symbol='KRT16', Aliases='CK16, FNEPPK, K16, K1CPA, NEPPK, PC1, KRT16', description='keratin 16, type I', other_designations='cytokeratin 16|cytokeratin-16|focal non-epidermolytic palmoplantar keratoderma|keratin-16', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41609778, end_position_on_the_genomic_accession=41612827, orientation='minus', exon_count=8, OMIM=148067, no_hdr0=''),
  677850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677850, CurrentID=0, Status='live', Symbol='SNORD1C', Aliases='R38C, snR38C', description='small nucleolar RNA, C/D box 1C', other_designations='R38c snoRNA', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76558792, end_position_on_the_genomic_accession=76558869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3872, CurrentID=0, Status='live', Symbol='KRT17', Aliases='39.1, CK-17, K17, PC, PC2, PCHC1', description='keratin 17, type I', other_designations='cytokeratin-17|keratin-17', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41619440, end_position_on_the_genomic_accession=41624630, orientation='minus', exon_count=8, OMIM=148069, no_hdr0=''),
  101060385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060385, CurrentID=0, Status='live', Symbol='LOC101060385', Aliases='', description='uncharacterized LOC101060385', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=900216, end_position_on_the_genomic_accession=905451, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101060386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060386, CurrentID=0, Status='live', Symbol='LOC101060386', Aliases='', description='putative uncharacterized SMG1-like protein', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3875, CurrentID=0, Status='live', Symbol='KRT18', Aliases='CK-18, CYK18, K18', description='keratin 18, type I', other_designations='cell proliferation-inducing gene 46 protein|cytokeratin 18', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52948871, end_position_on_the_genomic_accession=52952901, orientation='plus', exon_count=8, OMIM=148070, no_hdr0=''),
  3876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3876, CurrentID=0, Status='live', Symbol='KRT18P11', Aliases='KRT18L1', description='keratin 18 pseudogene 11', other_designations='keratin 18-like 1', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=92459635, end_position_on_the_genomic_accession=92461009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060389, CurrentID=0, Status='live', Symbol='LOC101060389', Aliases='TBC1D3D', description='TBC1 domain family member-like', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38004033, end_position_on_the_genomic_accession=38014908, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101060390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060390, CurrentID=0, Status='live', Symbol='LOC101060390', Aliases='', description='uncharacterized LOC101060390', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3879, CurrentID=0, Status='live', Symbol='KRT18P1', Aliases='dJ25J6.3', description='keratin 18 pseudogene 1', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28969072, end_position_on_the_genomic_accession=28970467, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3880, CurrentID=0, Status='live', Symbol='KRT19', Aliases='CK19, K19, K1CS', description='keratin 19, type I', other_designations='40-kDa keratin intermediate filament|CK-19|cytokeratin 19|cytokeratin-19|keratin, type I, 40-kd|keratin-19', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41523617, end_position_on_the_genomic_accession=41528389, orientation='minus', exon_count=6, OMIM=148020, no_hdr0=''),
  3881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3881, CurrentID=0, Status='live', Symbol='KRT31', Aliases='HA1, Ha-1, KRTHA1, hHa1', description='keratin 31, type I', other_designations='K31|hair keratin, type I Ha1|hard keratin, type I, 1|keratin, hair, acidic,1|keratin-31', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41393724, end_position_on_the_genomic_accession=41397592, orientation='minus', exon_count=7, OMIM=601077, no_hdr0=''),
  3882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3882, CurrentID=0, Status='live', Symbol='KRT32', Aliases='HA2, HKA2, KRTHA2, hHa2', description='keratin 32, type I', other_designations='K32|hair keratin, type I Ha2|hard keratin, type I, 2|keratin, hair, acidic, 2|keratin-32|type I cuticular hair keratin', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41459513, end_position_on_the_genomic_accession=41469252, orientation='minus', exon_count=7, OMIM=602760, no_hdr0=''),
  3883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3883, CurrentID=0, Status='live', Symbol='KRT33A', Aliases='HA3I, Ha-3I, K33A, KRTHA3A, Krt1-3, hHa3-I', description='keratin 33A, type I', other_designations='hair keratin, type I Ha3-I|hard keratin, type I,3I|keratin, hair, acidic, 3A', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41345705, end_position_on_the_genomic_accession=41350812, orientation='minus', exon_count=8, OMIM=602761, no_hdr0=''),
  3884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3884, CurrentID=0, Status='live', Symbol='KRT33B', Aliases='HA3II, Ha-3II, K33B, KRTHA3A, KRTHA3B, hHa3-II', description='keratin 33B, type I', other_designations='hair keratin, type I Ha3-II|hard keratin, type I, 3II|keratin, hair, acidic, 3B|type I hair keratin 3B', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41363494, end_position_on_the_genomic_accession=41369800, orientation='minus', exon_count=7, OMIM=602762, no_hdr0=''),
  134957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134957, CurrentID=0, Status='live', Symbol='STXBP5', Aliases='LGL3, LLGL3, Nbla04300', description='syntaxin binding protein 5 (tomosyn)', other_designations='lethal(2) giant larvae protein homolog 3|putative protein product of Nbla04300|tomosyn|tomosyn-1', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=147204358, end_position_on_the_genomic_accession=147390476, orientation='plus', exon_count=34, OMIM=604586, no_hdr0=''),
  3886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3886, CurrentID=0, Status='live', Symbol='KRT35', Aliases='HA5, Ha-5, KRTHA5, hHa5', description='keratin 35, type I', other_designations='HHa5 hair keratin type I intermediate filament|K35|hair keratin, type I Ha5|hard keratin, type I, 5|keratin, hair, acidic, 5|keratin-35', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41476689, end_position_on_the_genomic_accession=41482743, orientation='minus', exon_count=8, OMIM=602764, no_hdr0=''),
  3887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3887, CurrentID=0, Status='live', Symbol='KRT81', Aliases='HB1, Hb-1, KRTHB1, MLN137, ghHkb1, hHAKB2-1', description='keratin 81, type II', other_designations='K81|MLN 137|ghHb1|hair keratin K2.9|hard keratin, type II, 1|keratin, hair, basic, 1|keratin-81|metastatic lymph node 137 gene protein|type II hair keratin Hb1|type-II keratin Kb21', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52285913, end_position_on_the_genomic_accession=52308410, orientation='minus', exon_count=11, OMIM=602153, no_hdr0=''),
  3888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3888, CurrentID=0, Status='live', Symbol='KRT82', Aliases='HB2, Hb-2, KRTHB2', description='keratin 82, type II', other_designations='K82|hard keratin, type II, 2|keratin, hair, basic, 2|keratin-82|type II hair keratin Hb2|type-II keratin Kb22', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52393951, end_position_on_the_genomic_accession=52406392, orientation='minus', exon_count=9, OMIM=601078, no_hdr0=''),
  3889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3889, CurrentID=0, Status='live', Symbol='KRT83', Aliases='HB3, Hb-3, KRTHB3', description='keratin 83, type II', other_designations='K83|hHb3|hair keratin K2.10|hard keratin, type II, 3|keratin, hair, basic, 3|keratin-83|type II hair keratin Hb3|type-II keratin Kb23', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52314301, end_position_on_the_genomic_accession=52321398, orientation='minus', exon_count=9, OMIM=602765, no_hdr0=''),
  3890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3890, CurrentID=0, Status='live', Symbol='KRT84', Aliases='HB4, KRTHB4', description='keratin 84, type II', other_designations='K84|hard keratin, type II, 4|keratin, hair, basic, 4|keratin-84|type II hair keratin 4|type II hair keratin Hb4|type-II keratin Kb24', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52377812, end_position_on_the_genomic_accession=52386906, orientation='minus', exon_count=10, OMIM=602766, no_hdr0=''),
  3891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3891, CurrentID=0, Status='live', Symbol='KRT85', Aliases='ECTD4, HB5, Hb-5, K85, KRTHB5, hHb5', description='keratin 85, type II', other_designations='hair keratin K2.12|hard keratin, type II, 5|keratin, hair, basic, 5|keratin-85|type II hair keratin Hb5|type-II keratin Kb25', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52360006, end_position_on_the_genomic_accession=52367525, orientation='minus', exon_count=10, OMIM=602767, no_hdr0=''),
  3892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3892, CurrentID=0, Status='live', Symbol='KRT86', Aliases='HB6, Hb1, KRTHB1, KRTHB6, MNX, hHb6', description='keratin 86, type II', other_designations='K86|hair keratin K2.11|hard keratin, type II, 6|keratin protein HB6|keratin, hair, basic, 6 (monilethrix)|keratin-86|type II hair keratin Hb6|type-II keratin Kb26', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52274631, end_position_on_the_genomic_accession=52309163, orientation='plus', exon_count=12, OMIM=601928, no_hdr0=''),
  101060405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060405, CurrentID=0, Status='live', Symbol='LOC101060405', Aliases='', description='putative RRN3-like protein RRN3P1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3895, CurrentID=0, Status='live', Symbol='KTN1', Aliases='CG1, KNT, MU-RMS-40.19', description='kinectin 1', other_designations='CG-1 antigen|kinectin 1 (kinesin receptor)|kinesin receptor', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55580207, end_position_on_the_genomic_accession=55684584, orientation='plus', exon_count=48, OMIM=600381, no_hdr0=''),
  3896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3896, CurrentID=0, Status='live', Symbol='KWE', Aliases='', description='keratolytic winter erythema (Oudtshoorn skin disease)', other_designations='', map_location='8p23-p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=148370, no_hdr0=''),
  3897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3897, CurrentID=0, Status='live', Symbol='L1CAM', Aliases='CAML1, CD171, HSAS, HSAS1, MASA, MIC5, N-CAM-L1, N-CAML1, NCAM-L1, S10, SPG1', description='L1 cell adhesion molecule', other_designations='antigen identified by monoclonal antibody R1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153861514, end_position_on_the_genomic_accession=153886174, orientation='minus', exon_count=29, OMIM=308840, no_hdr0=''),
  3898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3898, CurrentID=0, Status='live', Symbol='LAD1', Aliases='LadA', description='ladinin 1', other_designations='lad-1|linear IgA bullous dermatosis antigen|linear IgA disease antigen homolog', map_location='1q25.1-q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201380838, end_position_on_the_genomic_accession=201399541, orientation='minus', exon_count=11, OMIM=602314, no_hdr0=''),
  3899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3899, CurrentID=0, Status='live', Symbol='AFF3', Aliases='LAF4, MLLT2-like', description='AF4/FMR2 family member 3', other_designations='AF4/FMR2 family, member 3|MLLT2-related protein|lymphoid nuclear protein 4|lymphoid nuclear protein related to AF4|protein LAF-4', map_location='2q11.2-q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99545864, end_position_on_the_genomic_accession=100142583, orientation='minus', exon_count=32, OMIM=601464, no_hdr0=''),
  100008586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100008586, CurrentID=0, Status='live', Symbol='GAGE12F', Aliases='', description='G antigen 12F', other_designations='GAGE-12F|g antigen 12F/G/I', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=49325416, end_position_on_the_genomic_accession=49332811, orientation='plus', exon_count=5, OMIM=300730, no_hdr0=''),
  3902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3902, CurrentID=0, Status='live', Symbol='LAG3', Aliases='CD223', description='lymphocyte-activation gene 3', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6772483, end_position_on_the_genomic_accession=6778455, orientation='plus', exon_count=8, OMIM=153337, no_hdr0=''),
  3903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3903, CurrentID=0, Status='live', Symbol='LAIR1', Aliases='CD305, LAIR-1', description='leukocyte-associated immunoglobulin-like receptor 1', other_designations='leukocyte-associated Ig-like receptor 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54353624, end_position_on_the_genomic_accession=54370556, orientation='minus', exon_count=12, OMIM=602992, no_hdr0=''),
  3904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3904, CurrentID=0, Status='live', Symbol='LAIR2', Aliases='CD306', description='leukocyte-associated immunoglobulin-like receptor 2', other_designations='LAIR-2|leukocyte-associated Ig-like receptor 2|leukocyte-associated Ig-like receptor-2', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54502797, end_position_on_the_genomic_accession=54510690, orientation='plus', exon_count=5, OMIM=602993, no_hdr0=''),
  3905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3905, CurrentID=0, Status='live', Symbol='LAKLG', Aliases='LAKL', description='lymphokine-activated killer cell ligand', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=153435, no_hdr0=''),
  3906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3906, CurrentID=0, Status='live', Symbol='LALBA', Aliases='', description='lactalbumin, alpha-', other_designations='lactose synthase B protein|lysozyme-like protein 7', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48567684, end_position_on_the_genomic_accession=48600111, orientation='minus', exon_count=6, OMIM=149750, no_hdr0=''),
  100008587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100008587, CurrentID=0, Status='live', Symbol='RNA5-8S5', Aliases='RN5-8S1', description='RNA, 5.8S ribosomal 5', other_designations='5.8S ribosomal RNA|RNA, 5.8S ribosomal 1', map_location='22p12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3908, CurrentID=0, Status='live', Symbol='LAMA2', Aliases='LAMM', description='laminin subunit alpha 2', other_designations='laminin M chain|laminin, alpha 2|laminin-12 subunit alpha|laminin-2 subunit alpha|laminin-4 subunit alpha|merosin heavy chain', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128883141, end_position_on_the_genomic_accession=129516569, orientation='plus', exon_count=66, OMIM=156225, no_hdr0=''),
  3909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3909, CurrentID=0, Status='live', Symbol='LAMA3', Aliases='BM600, E170, LAMNA, LOCS', description='laminin subunit alpha 3', other_designations='BM600 150kD subunit|LAM3, alpha-3 subunit|epiligrin 170 kda subunit|epiligrin alpha 3 subunit|kalinin 165kD subunit|kalinin subunit alpha|laminin 5, alpha-3 subunit|laminin A3|laminin, alpha 3 (nicein (150kD), kalinin (165kD), BM600 (150kD), epilegrin)|laminin-5 alpha 3 chain|laminin-5 subunit alpha|laminin-6 subunit alpha|laminin-7 subunit alpha|nicein 150kD subunit|nicein subunit alpha', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23689443, end_position_on_the_genomic_accession=23955066, orientation='plus', exon_count=77, OMIM=600805, no_hdr0=''),
  3910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3910, CurrentID=0, Status='live', Symbol='LAMA4', Aliases='CMD1JJ, LAMA3*-1, LAMA4', description='laminin subunit alpha 4', other_designations='laminin alpha 4 chain|laminin, alpha 4', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112107931, end_position_on_the_genomic_accession=112254722, orientation='minus', exon_count=39, OMIM=600133, no_hdr0=''),
  3911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3911, CurrentID=0, Status='live', Symbol='LAMA5', Aliases='', description='laminin subunit alpha 5', other_designations='laminin alpha-5 chain|laminin, alpha 5|laminin-10 subunit alpha|laminin-11 subunit alpha|laminin-15 subunit alpha', map_location='20q13.2-q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62309060, end_position_on_the_genomic_accession=62367320, orientation='minus', exon_count=81, OMIM=601033, no_hdr0=''),
  3912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3912, CurrentID=0, Status='live', Symbol='LAMB1', Aliases='CLM, LIS5', description='laminin subunit beta 1', other_designations='laminin B1 chain|laminin, beta 1', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107923801, end_position_on_the_genomic_accession=108003359, orientation='minus', exon_count=35, OMIM=150240, no_hdr0=''),
  3913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3913, CurrentID=0, Status='live', Symbol='LAMB2', Aliases='LAMS, NPHS5', description='laminin subunit beta 2', other_designations='S-LAM beta|S-laminin subunit beta|laminin B1s chain|laminin S|laminin, beta 2 (laminin S)', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49121114, end_position_on_the_genomic_accession=49133166, orientation='minus', exon_count=32, OMIM=150325, no_hdr0=''),
  3914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3914, CurrentID=0, Status='live', Symbol='LAMB3', Aliases='AI1A, BM600-125KDA, LAM5, LAMNB1', description='laminin subunit beta 3', other_designations='epiligrin subunit bata|kalinin B1 chain|kalinin subunit beta|kalinin-140kDa|laminin B1k chain|laminin S B3 chain|laminin beta 3|laminin, beta 3 (nicein (125kD), kalinin (140kD), BM600 (125kD))|laminin-5 subunit beta|nicein subunit beta|nicein-125kDa', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209614870, end_position_on_the_genomic_accession=209652475, orientation='minus', exon_count=25, OMIM=150310, no_hdr0=''),
  3915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3915, CurrentID=0, Status='live', Symbol='LAMC1', Aliases='LAMB2', description='laminin subunit gamma 1', other_designations='S-LAM gamma|S-laminin subunit gamma|laminin B2 chain|laminin, gamma 1 (formerly LAMB2)|laminin-10 subunit gamma|laminin-11 subunit gamma|laminin-2 subunit gamma|laminin-3 subunit gamma|laminin-4 subunit gamma|laminin-6 subunit gamma|laminin-7 subunit gamma|laminin-8 subunit gamma|laminin-9 subunit gamma', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183023460, end_position_on_the_genomic_accession=183145592, orientation='plus', exon_count=28, OMIM=150290, no_hdr0=''),
  3916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3916, CurrentID=0, Status='live', Symbol='LAMP1', Aliases='CD107a, LAMPA, LGP120', description='lysosomal-associated membrane protein 1', other_designations='CD107 antigen-like family member A|LAMP-1|lysosome-associated membrane protein 1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113297121, end_position_on_the_genomic_accession=113323431, orientation='plus', exon_count=10, OMIM=153330, no_hdr0=''),
  101060429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060429, CurrentID=0, Status='live', Symbol='LOC101060429', Aliases='', description='polycystin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3918, CurrentID=0, Status='live', Symbol='LAMC2', Aliases='B2T, BM600, CSF, EBR2, EBR2A, LAMB2T, LAMNB2', description='laminin subunit gamma 2', other_designations='BM600-100kDa|CSF 140 kDa subunit|cell-scattering factor 140 kDa subunit|epiligrin subunit gamma|kalinin subunit gamma|ladsin 140 kDa subunit|laminin B2t chain|large adhesive scatter factor 140 kDa subunit|nicein subunit gamma', map_location='1q25-q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183186039, end_position_on_the_genomic_accession=183245127, orientation='plus', exon_count=23, OMIM=150292, no_hdr0=''),
  100008589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100008589, CurrentID=0, Status='live', Symbol='RNA28S5', Aliases='RN28S1', description='RNA, 28S ribosomal 5', other_designations='28S ribosomal RNA|RNA, 28S ribosomal 1', map_location='22p12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3920, CurrentID=0, Status='live', Symbol='LAMP2', Aliases='CD107b, LAMP-2, LAMPB, LGP110', description='lysosomal-associated membrane protein 2', other_designations='CD107 antigen-like family member B', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120426148, end_position_on_the_genomic_accession=120469349, orientation='minus', exon_count=10, OMIM=309060, no_hdr0=''),
  3921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3921, CurrentID=0, Status='live', Symbol='RPSA', Aliases='37LRP, 67LR, ICAS, LAMBR, LAMR1, LBP, LBP/p40, LRP, LRP/LR, NEM/1CHD4, SA, lamR, p40', description='ribosomal protein SA', other_designations='37 kDa laminin receptor|37/67 kDa laminin receptor|67 kDa laminin receptor|colon carcinoma laminin-binding protein|laminin receptor 1 (67kD, ribosomal protein SA)|laminin-binding protein precursor p40|multidrug resistance-associated protein MGr1-Ag', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39406689, end_position_on_the_genomic_accession=39412542, orientation='plus', exon_count=7, OMIM=150370, no_hdr0=''),
  3923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3923, CurrentID=0, Status='live', Symbol='RPSAP13', Aliases='LAMR1P13, LAMRL3, LAMRP3, RPSA_20_1366', description='ribosomal protein SA pseudogene 13', other_designations='laminin receptor 1 pseudogene 13|laminin receptor-like 3', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55004755, end_position_on_the_genomic_accession=55005763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3924, CurrentID=0, Status='live', Symbol='RPSAP14', Aliases='LAMR1P14, LAMRL4, LAMRP4', description='ribosomal protein SA pseudogene 14', other_designations='laminin receptor 1 pseudogene 14|laminin receptor-like 4', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74031461, end_position_on_the_genomic_accession=74031839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3925, CurrentID=0, Status='live', Symbol='STMN1', Aliases='C1orf215, LAP18, Lag, OP18, PP17, PP19, PR22, SMN', description='stathmin 1', other_designations='leukemia-associated phosphoprotein p18|metablastin|oncoprotein 18|phosphoprotein 19|phosphoprotein p19|prosolin|stathmin 1/oncoprotein 18|transmembrane protein C1orf215', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25884186, end_position_on_the_genomic_accession=25906877, orientation='minus', exon_count=8, OMIM=151442, no_hdr0=''),
  3927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3927, CurrentID=0, Status='live', Symbol='LASP1', Aliases='Lasp-1, MLN50', description='LIM and SH3 protein 1', other_designations='metastatic lymph node gene 50 protein', map_location='17q11-q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38869859, end_position_on_the_genomic_accession=38921770, orientation='plus', exon_count=8, OMIM=602920, no_hdr0=''),
  3929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3929, CurrentID=0, Status='live', Symbol='LBP', Aliases='BPIFD2', description='lipopolysaccharide binding protein', other_designations='BPI fold containing family D, member 2|LPS-binding protein', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38346411, end_position_on_the_genomic_accession=38377011, orientation='plus', exon_count=15, OMIM=151990, no_hdr0=''),
  3930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3930, CurrentID=0, Status='live', Symbol='LBR', Aliases='DHCR14B, LMN2R, PHA, TDRD18', description='lamin B receptor', other_designations='integral nuclear envelope inner membrane protein|tudor domain containing 18', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225401502, end_position_on_the_genomic_accession=225428855, orientation='minus', exon_count=16, OMIM=600024, no_hdr0=''),
  3931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3931, CurrentID=0, Status='live', Symbol='LCAT', Aliases='', description='lecithin-cholesterol acyltransferase', other_designations='phosphatidylcholine--sterol O-acyltransferase|phospholipid-cholesterol acyltransferase', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67939884, end_position_on_the_genomic_accession=67944112, orientation='minus', exon_count=6, OMIM=606967, no_hdr0=''),
  3932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3932, CurrentID=0, Status='live', Symbol='LCK', Aliases='IMD22, LSK, YT16, p56lck, pp58lck', description='LCK proto-oncogene, Src family tyrosine kinase', other_designations='T-lymphocyte specific protein tyrosine kinase p56lck|leukocyte C-terminal Src kinase|lymphocyte cell-specific protein-tyrosine kinase|lymphocyte-specific protein tyrosine kinase|p56(LSTRA) protein-tyrosine kinase|proto-oncogene tyrosine-protein kinase LCK|t cell-specific protein-tyrosine kinase', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32251239, end_position_on_the_genomic_accession=32286170, orientation='plus', exon_count=13, OMIM=153390, no_hdr0=''),
  3933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3933, CurrentID=0, Status='live', Symbol='LCN1', Aliases='PMFA, TLC, TP, VEGP', description='lipocalin 1', other_designations='VEG protein|Von Ebner gland protein|lipocalin 1 (tear prealbumin)|protein migrating faster than albumin|tear prealbumin', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135521297, end_position_on_the_genomic_accession=135526540, orientation='plus', exon_count=7, OMIM=151675, no_hdr0=''),
  3934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3934, CurrentID=0, Status='live', Symbol='LCN2', Aliases='24p3, MSFI, NGAL, p25', description='lipocalin 2', other_designations='25 kDa alpha-2-microglobulin-related subunit of MMP-9|lipocalin-2|migration-stimulating factor inhibitor|oncogene 24p3|siderocalin LCN2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128149453, end_position_on_the_genomic_accession=128153455, orientation='plus', exon_count=7, OMIM=600181, no_hdr0=''),
  3936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3936, CurrentID=0, Status='live', Symbol='LCP1', Aliases='CP64, HEL-S-37, L-PLASTIN, LC64P, LPL, PLS2', description='lymphocyte cytosolic protein 1 (L-plastin)', other_designations='L-plastin (Lymphocyte cytosolic protein 1) (LCP-1) (LC64P)|LCP-1|Lymphocyte cytosolic protein-1 (plasmin)|bA139H14.1 (lymphocyte cytosolic protein 1 (L-plastin))|epididymis secretory protein Li 37|plastin 2', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46125923, end_position_on_the_genomic_accession=46182324, orientation='minus', exon_count=17, OMIM=153430, no_hdr0=''),
  3937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3937, CurrentID=0, Status='live', Symbol='LCP2', Aliases='SLP-76, SLP76', description='lymphocyte cytosolic protein 2', other_designations='76 kDa tyrosine phosphoprotein|SH2 domain-containing leukocyte protein of 76 kDa|SH2 domain-containing leukocyte protein of 76kD|SLP-76 tyrosine phosphoprotein|lymphocyte cytosolic protein 2 (SH2 domain containing leukocyte protein of 76kDa)', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170247754, end_position_on_the_genomic_accession=170297818, orientation='minus', exon_count=21, OMIM=601603, no_hdr0=''),
  3938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3938, CurrentID=0, Status='live', Symbol='LCT', Aliases='LAC, LPH, LPH1', description='lactase', other_designations='lactase-glycosylceramidase|lactase-phlorizin hydrolase-1', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135787845, end_position_on_the_genomic_accession=135837180, orientation='minus', exon_count=17, OMIM=603202, no_hdr0=''),
  3939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3939, CurrentID=0, Status='live', Symbol='LDHA', Aliases='GSD11, HEL-S-133P, LDHM, PIG19', description='lactate dehydrogenase A', other_designations='LDH muscle subunit|LDH-A|LDH-M|cell proliferation-inducing gene 19 protein|epididymis secretory sperm binding protein Li 133P|lactate dehydrogenase M|proliferation-inducing gene 19|renal carcinoma antigen NY-REN-59', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18394389, end_position_on_the_genomic_accession=18408218, orientation='plus', exon_count=9, OMIM=150000, no_hdr0=''),
  3940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3940, CurrentID=0, Status='live', Symbol='LDHAP1', Aliases='LDHAL1', description='lactate dehydrogenase A pseudogene 1', other_designations='lactate dehydrogenase A-like 1', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4893624, end_position_on_the_genomic_accession=4895268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401667, CurrentID=0, Status='live', Symbol='OR51A2', Aliases='', description='olfactory receptor family 51 subfamily A member 2', other_designations='olfactory receptor, family 51, subfamily A, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4954772, end_position_on_the_genomic_accession=4955713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3945, CurrentID=0, Status='live', Symbol='LDHB', Aliases='HEL-S-281, LDH-B, LDH-HD, TRG-5, LDHB', description='lactate dehydrogenase B', other_designations='LDH heart subunit|epididymis secretory protein Li 281|lactate dehydrogenase H chain|renal carcinoma antigen NY-REN-46', map_location='12p12.2-p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21635341, end_position_on_the_genomic_accession=21657971, orientation='minus', exon_count=8, OMIM=150100, no_hdr0=''),
  101060458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060458, CurrentID=0, Status='live', Symbol='LOC101060458', Aliases='', description='peptidyl-prolyl cis-trans isomerase A-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  3947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3947, CurrentID=0, Status='live', Symbol='LDHBP2', Aliases='LDHBL2, LDHBP', description='lactate dehydrogenase B pseudogene 2', other_designations='lactate dehydrogenase B-like 2', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76334643, end_position_on_the_genomic_accession=76335910, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3948, CurrentID=0, Status='live', Symbol='LDHC', Aliases='CT32, LDH3, LDHX', description='lactate dehydrogenase C', other_designations='LDH testis subunit|LDH-C|LDH-X|cancer/testis antigen 32|lactate dehydrogenase C4|lactate dehydrogenase c variant 1|lactate dehydrogenase c variant 3|lactate dehydrogenase c variant 4|lactate dehydrogenase c variant 5', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18412306, end_position_on_the_genomic_accession=18451246, orientation='plus', exon_count=8, OMIM=150150, no_hdr0=''),
  3949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3949, CurrentID=0, Status='live', Symbol='LDLR', Aliases='FH, FHC, LDLCQ2', description='low density lipoprotein receptor', other_designations='LDL receptor|low-density lipoprotein receptor class A domain-containing protein 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11089362, end_position_on_the_genomic_accession=11133830, orientation='plus', exon_count=18, OMIM=606945, no_hdr0=''),
  3950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3950, CurrentID=0, Status='live', Symbol='LECT2', Aliases='chm-II, chm2', description='leukocyte cell-derived chemotaxin 2', other_designations='chondromodulin-II', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135946911, end_position_on_the_genomic_accession=135955034, orientation='minus', exon_count=4, OMIM=602882, no_hdr0=''),
  3952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3952, CurrentID=0, Status='live', Symbol='LEP', Aliases='LEPD, OB, OBS', description='leptin', other_designations='leptin (murine obesity homolog)|leptin (obesity homolog, mouse)|obese protein|obese, mouse, homolog of|obesity factor', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128241202, end_position_on_the_genomic_accession=128257629, orientation='plus', exon_count=3, OMIM=164160, no_hdr0=''),
  3953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3953, CurrentID=0, Status='live', Symbol='LEPR', Aliases='CD295, LEP-RD, OB-R, OBR, LEPR', description='leptin receptor', other_designations='OB receptor|huB219', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65420652, end_position_on_the_genomic_accession=65637493, orientation='plus', exon_count=24, OMIM=601007, no_hdr0=''),
  3954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3954, CurrentID=0, Status='live', Symbol='LETM1', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 1', other_designations='Mdm38 homolog', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1811479, end_position_on_the_genomic_accession=1856247, orientation='minus', exon_count=15, OMIM=604407, no_hdr0=''),
  3955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3955, CurrentID=0, Status='live', Symbol='LFNG', Aliases='SCDO3', description='LFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2512529, end_position_on_the_genomic_accession=2529177, orientation='plus', exon_count=13, OMIM=602576, no_hdr0=''),
  3956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3956, CurrentID=0, Status='live', Symbol='LGALS1', Aliases='GAL1, GBP', description='lectin, galactoside-binding, soluble, 1', other_designations='14 kDa laminin-binding protein|14 kDa lectin|HBL|HLBP14|HPL|S-Lac lectin 1|beta-galactoside-binding lectin L-14-I|beta-galactoside-binding protein 14kDa|gal-1|galaptin|galectin 1|lactose-binding lectin 1|putative MAPK-activating protein PM12', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37675606, end_position_on_the_genomic_accession=37679802, orientation='plus', exon_count=4, OMIM=150570, no_hdr0=''),
  3957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3957, CurrentID=0, Status='live', Symbol='LGALS2', Aliases='HL14', description='lectin, galactoside-binding, soluble, 2', other_designations='S-Lac lectin 2|beta-galactoside-binding lectin L-14-II|gal-2|galectin 2|lactose-binding lectin 2', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37570246, end_position_on_the_genomic_accession=37580017, orientation='minus', exon_count=4, OMIM=150571, no_hdr0=''),
  3958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3958, CurrentID=0, Status='live', Symbol='LGALS3', Aliases='CBP35, GAL3, GALBP, GALIG, L31, LGALS2, MAC2', description='lectin, galactoside-binding, soluble, 3', other_designations='35 kDa lectin|IgE-binding protein|MAC-2 antigen|carbohydrate-binding protein 35|galactose-specific lectin 3|laminin-binding protein|lectin L-29', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55129217, end_position_on_the_genomic_accession=55145430, orientation='plus', exon_count=7, OMIM=153619, no_hdr0=''),
  3959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3959, CurrentID=0, Status='live', Symbol='LGALS3BP', Aliases='90K, BTBD17B, CyCAP, M2BP, MAC-2-BP, TANGO10B, gp90', description='lectin, galactoside-binding, soluble, 3 binding protein', other_designations='L3 antigen|MAC2BP|Mac-2-binding protein|basement membrane autoantigen p105|lectin galactoside-binding soluble 3 binding protein|lectin galactoside-binding soluble 3-binding protein|mac-2 BP|serum protein 90K|transport and golgi organization 10 homolog B|tumor-associated antigen 90K', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78971253, end_position_on_the_genomic_accession=78979979, orientation='minus', exon_count=6, OMIM=600626, no_hdr0=''),
  3960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3960, CurrentID=0, Status='live', Symbol='LGALS4', Aliases='GAL4, L36LBP', description='lectin, galactoside-binding, soluble, 4', other_designations='L-36 lactose-binding protein|antigen NY-CO-27|gal-4|galectin 4|lactose-binding lectin 4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38801671, end_position_on_the_genomic_accession=38813544, orientation='minus', exon_count=10, OMIM=602518, no_hdr0=''),
  3963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3963, CurrentID=0, Status='live', Symbol='LGALS7', Aliases='GAL7A, LGALS7', description='lectin, galactoside-binding, soluble, 7', other_designations='galectin 7', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38770968, end_position_on_the_genomic_accession=38773517, orientation='minus', exon_count=4, OMIM=600615, no_hdr0=''),
  3964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3964, CurrentID=0, Status='live', Symbol='LGALS8', Aliases='Gal-8, PCTA-1, PCTA1, Po66-CBP', description='lectin, galactoside-binding, soluble, 8', other_designations='Po66 carbohydrate binding protein|galectin 8|galectin-8g|po66 carbohydrate-binding protein|prostate carcinoma tumor antigen 1', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236518214, end_position_on_the_genomic_accession=236552981, orientation='plus', exon_count=15, OMIM=606099, no_hdr0=''),
  3965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3965, CurrentID=0, Status='live', Symbol='LGALS9', Aliases='HUATA, LGALS9', description='lectin, galactoside-binding, soluble, 9', other_designations='ecalectin|gal-9|galectin 9|tumor antigen HOM-HD-21|urate transporter/channel protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27631148, end_position_on_the_genomic_accession=27649560, orientation='plus', exon_count=12, OMIM=601879, no_hdr0=''),
  3972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3972, CurrentID=0, Status='live', Symbol='LHB', Aliases='CGB4, HH23, LSH-B, hLHB', description='luteinizing hormone beta polypeptide', other_designations='LSH-beta|interstitial cell stimulating hormone, beta chain|luteinizing hormone beta subunit|lutropin beta chain', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49015980, end_position_on_the_genomic_accession=49017090, orientation='minus', exon_count=3, OMIM=152780, no_hdr0=''),
  3973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3973, CurrentID=0, Status='live', Symbol='LHCGR', Aliases='HHG, LCGR, LGR2, LH/CG-R, LH/CGR, LHR, LHRHR, LSH-R, ULG5', description='luteinizing hormone/choriogonadotropin receptor', other_designations='hypergonadotropic hypogonadism|lutropin/choriogonadotropin receptor', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48686774, end_position_on_the_genomic_accession=48755741, orientation='minus', exon_count=14, OMIM=152790, no_hdr0=''),
  3975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3975, CurrentID=0, Status='live', Symbol='LHX1', Aliases='LIM-1, LIM1', description='LIM homeobox 1', other_designations='LIM homeobox protein 1|homeobox protein Lim-1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36937475, end_position_on_the_genomic_accession=36944615, orientation='plus', exon_count=5, OMIM=601999, no_hdr0=''),
  3976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3976, CurrentID=0, Status='live', Symbol='LIF', Aliases='CDF, DIA, HILDA, MLPLI', description='leukemia inhibitory factor', other_designations='D factor|cholinergic differentiation factor|differentiation inhibitory activity|differentiation-inducing factor|differentiation-stimulating factor|hepatocyte-stimulating factor III|human interleukin in DA cells|melanoma-derived LPL inhibitor', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30240447, end_position_on_the_genomic_accession=30257923, orientation='minus', exon_count=5, OMIM=159540, no_hdr0=''),
  3977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3977, CurrentID=0, Status='live', Symbol='LIFR', Aliases='CD118, LIF-R, SJS2, STWS, SWS', description='leukemia inhibitory factor receptor alpha', other_designations='CD118 antigen|LIF receptor', map_location='5p13-p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38474963, end_position_on_the_genomic_accession=38595405, orientation='minus', exon_count=24, OMIM=151443, no_hdr0=''),
  3978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3978, CurrentID=0, Status='live', Symbol='LIG1', Aliases='', description='ligase I, DNA, ATP-dependent', other_designations='polydeoxyribonucleotide synthase [ATP] 1', map_location='19q13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48115444, end_position_on_the_genomic_accession=48170603, orientation='minus', exon_count=28, OMIM=126391, no_hdr0=''),
  3980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3980, CurrentID=0, Status='live', Symbol='LIG3', Aliases='LIG2', description='ligase III, DNA, ATP-dependent', other_designations='ligase II, DNA, ATP-dependent|polydeoxyribonucleotide synthase [ATP] 3', map_location='17q11.2-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34980498, end_position_on_the_genomic_accession=35008448, orientation='plus', exon_count=23, OMIM=600940, no_hdr0=''),
  3981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3981, CurrentID=0, Status='live', Symbol='LIG4', Aliases='LIG4S', description='ligase IV, DNA, ATP-dependent', other_designations='DNA joinase|DNA ligase IV|DNA repair enzyme|polydeoxyribonucleotide synthase [ATP] 4|polynucleotide ligase|sealase', map_location='13q33-q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108207442, end_position_on_the_genomic_accession=108218368, orientation='minus', exon_count=5, OMIM=601837, no_hdr0=''),
  3982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3982, CurrentID=0, Status='live', Symbol='LIM2', Aliases='CTRCT19, MP17, MP19', description='lens intrinsic membrane protein 2', other_designations='MP18|MP20|lens intrinsic membrane protein 2, 19kDa', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51379909, end_position_on_the_genomic_accession=51387956, orientation='minus', exon_count=5, OMIM=154045, no_hdr0=''),
  3983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3983, CurrentID=0, Status='live', Symbol='ABLIM1', Aliases='ABLIM, LIMAB1, LIMATIN, abLIM-1', description='actin binding LIM protein 1', other_designations='actin-binding LIM protein family member 1|actin-binding double-zinc-finger protein', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114431110, end_position_on_the_genomic_accession=114821643, orientation='minus', exon_count=34, OMIM=602330, no_hdr0=''),
  3984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3984, CurrentID=0, Status='live', Symbol='LIMK1', Aliases='LIMK, LIMK-1', description='LIM domain kinase 1', other_designations='LIM motif-containing protein kinase', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74083777, end_position_on_the_genomic_accession=74122525, orientation='plus', exon_count=17, OMIM=601329, no_hdr0=''),
  3985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3985, CurrentID=0, Status='live', Symbol='LIMK2', Aliases='', description='LIM domain kinase 2', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31212264, end_position_on_the_genomic_accession=31280080, orientation='plus', exon_count=18, OMIM=601988, no_hdr0=''),
  101060498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060498, CurrentID=0, Status='live', Symbol='LOC101060498', Aliases='', description='uncharacterized LOC101060498', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40307185, end_position_on_the_genomic_accession=40330419, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3987, CurrentID=0, Status='live', Symbol='LIMS1', Aliases='PINCH, PINCH-1, PINCH1', description='LIM and senescent cell antigen-like domains 1', other_designations='particularly interesting new Cys-His protein 1|renal carcinoma antigen NY-REN-48|senescent cell antigen', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108534355, end_position_on_the_genomic_accession=108687246, orientation='plus', exon_count=16, OMIM=602567, no_hdr0=''),
  3988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3988, CurrentID=0, Status='live', Symbol='LIPA', Aliases='CESD, LAL', description='lipase A, lysosomal acid, cholesterol esterase', other_designations='acid cholesteryl ester hydrolase|cholesterol ester hydrolase|cholesteryl esterase|lysosomal acid lipase|sterol esterase', map_location='10q23.2-q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89213569, end_position_on_the_genomic_accession=89252039, orientation='minus', exon_count=10, OMIM=613497, no_hdr0=''),
  3989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3989, CurrentID=0, Status='live', Symbol='LIPB', Aliases='', description='lipase B, lysosomal acid', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=247980, no_hdr0=''),
  3990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3990, CurrentID=0, Status='live', Symbol='LIPC', Aliases='HDLCQ12, HL, HTGL, LIPH', description='lipase, hepatic', other_designations='hepatic lipase|lipase member C', map_location='15q21-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58410754, end_position_on_the_genomic_accession=58568874, orientation='plus', exon_count=13, OMIM=151670, no_hdr0=''),
  3991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3991, CurrentID=0, Status='live', Symbol='LIPE', Aliases='AOMS4, FPLD6, HSL, LHS', description='lipase, hormone-sensitive', other_designations='hormone-sensitive lipase testicular isoform', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42401507, end_position_on_the_genomic_accession=42427426, orientation='minus', exon_count=16, OMIM=151750, no_hdr0=''),
  3992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3992, CurrentID=0, Status='live', Symbol='FADS1', Aliases='D5D, FADS6, FADSD5, LLCDL1, TU12', description='fatty acid desaturase 1', other_designations='delta(5) desaturase|delta(5) fatty acid desaturase|delta-5 desaturase|delta-5 fatty acid desaturase|linoleoyl-CoA desaturase (delta-6-desaturase)-like 1', map_location='11q12.2-q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61799625, end_position_on_the_genomic_accession=61817057, orientation='minus', exon_count=13, OMIM=606148, no_hdr0=''),
  3993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3993, CurrentID=0, Status='live', Symbol='LLGL2', Aliases='HGL, Hugl-2, LGL2', description='lethal giant larvae homolog 2 (Drosophila)', other_designations='human giant larvae homolog', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75525080, end_position_on_the_genomic_accession=75575209, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  3995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3995, CurrentID=0, Status='live', Symbol='FADS3', Aliases='CYB5RP, LLCDL3', description='fatty acid desaturase 3', other_designations='cytochrome b5-related protein|delta-9 fatty acid desaturase|delta-9-desaturase|linoleoyl-CoA desaturase (delta-9-desaturase)-like 3', map_location='11q12-q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61873523, end_position_on_the_genomic_accession=61891545, orientation='minus', exon_count=12, OMIM=606150, no_hdr0=''),
  3996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3996, CurrentID=0, Status='live', Symbol='LLGL1', Aliases='DLG4, HUGL, HUGL-1, HUGL1, LLGL', description='lethal giant larvae homolog 1 (Drosophila)', other_designations='human homolog to the D-lgl gene protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18225593, end_position_on_the_genomic_accession=18244875, orientation='plus', exon_count=23, OMIM=600966, no_hdr0=''),
  100270746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270746, CurrentID=0, Status='live', Symbol='LOC100270746', Aliases='', description='uncharacterized LOC100270746', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27019366, end_position_on_the_genomic_accession=27020306, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3998, CurrentID=0, Status='live', Symbol='LMAN1', Aliases='ERGIC-53, ERGIC53, F5F8D, FMFD1, MCFD1, MR60, gp58', description='lectin, mannose-binding, 1', other_designations='ER-Golgi intermediate compartment 53 kDa protein|endoplasmic reticulum-golgi intermediate compartment protein 53|intracellular mannose-specific lectin MR60', map_location='18q21.3-q22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59327823, end_position_on_the_genomic_accession=59359276, orientation='minus', exon_count=13, OMIM=601567, no_hdr0=''),
  4000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4000, CurrentID=0, Status='live', Symbol='LMNA', Aliases='CDCD1, CDDC, CMD1A, CMT2B1, EMD2, FPL, FPLD, FPLD2, HGPS, IDC, LDP1, LFP, LGMD1B, LMN1, LMNC, LMNL1, PRO1', description='lamin A/C', other_designations='70 kDa lamin|lamin A/C-like 1|prelamin-A/C|renal carcinoma antigen NY-REN-32', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156082546, end_position_on_the_genomic_accession=156140089, orientation='plus', exon_count=17, OMIM=150330, no_hdr0=''),
  4001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4001, CurrentID=0, Status='live', Symbol='LMNB1', Aliases='ADLD, LMN, LMN2, LMNB', description='lamin B1', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126776623, end_position_on_the_genomic_accession=126837020, orientation='plus', exon_count=12, OMIM=150340, no_hdr0=''),
  105377111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377111, CurrentID=0, Status='live', Symbol='LOC105377111', Aliases='', description='uncharacterized LOC105377111', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60989254, end_position_on_the_genomic_accession=60990633, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4004, CurrentID=0, Status='live', Symbol='LMO1', Aliases='RBTN1, RHOM1, TTG1', description='LIM domain only 1', other_designations='LIM domain only 1 (rhombotin 1)|LIM domain only protein 1|LMO-1|T-cell translocation gene 1|T-cell translocation protein 1|cysteine-rich protein TTG-1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8224304, end_position_on_the_genomic_accession=8268635, orientation='minus', exon_count=10, OMIM=186921, no_hdr0=''),
  4005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4005, CurrentID=0, Status='live', Symbol='LMO2', Aliases='RBTN2, RBTNL1, RHOM2, TTG2', description='LIM domain only 2', other_designations='LIM domain only 2 (rhombotin-like 1)|LIM domain only protein 2|LMO-2|T-cell translocation gene 2|T-cell translocation protein 2|cysteine-rich protein TTG-2|rhombotin 2|rhombotin-like 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33858576, end_position_on_the_genomic_accession=33892289, orientation='minus', exon_count=8, OMIM=180385, no_hdr0=''),
  4007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4007, CurrentID=0, Status='live', Symbol='PRICKLE3', Aliases='LMO6', description='prickle homolog 3', other_designations='LIM domain only protein 6|LMO-6|triple LIM domain protein 6', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49174841, end_position_on_the_genomic_accession=49186528, orientation='minus', exon_count=10, OMIM=300111, no_hdr0=''),
  4008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4008, CurrentID=0, Status='live', Symbol='LMO7', Aliases='FBX20, FBXO20, LOMP', description='LIM domain 7', other_designations='F-box only protein 20|F-box protein Fbx20|LIM domain only 7 protein|LMO-7|zinc-finger domain-containing protein', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75620434, end_position_on_the_genomic_accession=75859870, orientation='plus', exon_count=37, OMIM=604362, no_hdr0=''),
  4009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4009, CurrentID=0, Status='live', Symbol='LMX1A', Aliases='LMX1, LMX1.1', description='LIM homeobox transcription factor 1, alpha', other_designations='LIM/homeobox protein 1.1', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165201867, end_position_on_the_genomic_accession=165356715, orientation='minus', exon_count=13, OMIM=600298, no_hdr0=''),
  4010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4010, CurrentID=0, Status='live', Symbol='LMX1B', Aliases='LMX1.2, NPS1', description='LIM homeobox transcription factor 1, beta', other_designations='LIM homeobox transcription factor 1 beta|LIM/homeobox protein 1.2|LMX-1.2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126614443, end_position_on_the_genomic_accession=126701032, orientation='plus', exon_count=8, OMIM=602575, no_hdr0=''),
  4012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4012, CurrentID=0, Status='live', Symbol='LNPEP', Aliases='CAP, IRAP, P-LAP, PLAP', description='leucyl/cystinyl aminopeptidase', other_designations='AT (4) receptor|angiotensin IV receptor|cystinyl aminopeptidase|insulin-regulated aminopeptidase|insulin-regulated membrane aminopeptidase|insulin-responsive aminopeptidase|otase|oxytocinase|placental leucine aminopeptidase|vasopressinase', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96935642, end_position_on_the_genomic_accession=97029411, orientation='plus', exon_count=19, OMIM=151300, no_hdr0=''),
  4013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4013, CurrentID=0, Status='live', Symbol='VWA5A', Aliases='BCSC-1, BCSC1, LOH11CR2A', description='von Willebrand factor A domain containing 5A', other_designations='breast cancer suppressor candidate 1|loss of heterozygosity 11 chromosomal region 2 gene A protein|loss of heterozygosity, 11, chromosomal region 2, gene A|ortholog of mouse AW551984', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124115404, end_position_on_the_genomic_accession=124146912, orientation='plus', exon_count=19, OMIM=602929, no_hdr0=''),
  4014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4014, CurrentID=0, Status='live', Symbol='LOR', Aliases='', description='loricrin', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153259635, end_position_on_the_genomic_accession=153262125, orientation='plus', exon_count=3, OMIM=152445, no_hdr0=''),
  4015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4015, CurrentID=0, Status='live', Symbol='LOX', Aliases='', description='lysyl oxidase', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122063195, end_position_on_the_genomic_accession=122078360, orientation='minus', exon_count=7, OMIM=153455, no_hdr0=''),
  4016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4016, CurrentID=0, Status='live', Symbol='LOXL1', Aliases='LOL, LOXL', description='lysyl oxidase like 1', other_designations='lysyl oxidase-like 1|lysyl oxidase-like protein 1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73926448, end_position_on_the_genomic_accession=73952137, orientation='plus', exon_count=9, OMIM=153456, no_hdr0=''),
  4017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4017, CurrentID=0, Status='live', Symbol='LOXL2', Aliases='LOR2, WS9-14', description='lysyl oxidase like 2', other_designations='lysyl oxidase related 2|lysyl oxidase-like 2 delta e13|lysyl oxidase-like 2 protein|lysyl oxidase-like protein 2|lysyl oxidase-related protein 2|lysyl oxidase-related protein WS9-14', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23296897, end_position_on_the_genomic_accession=23404209, orientation='minus', exon_count=14, OMIM=606663, no_hdr0=''),
  4018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4018, CurrentID=0, Status='live', Symbol='LPA', Aliases='AK38, APOA, LP', description='lipoprotein, Lp(a)', other_designations='antiangiogenic AK38 protein|apo(a)|lp(a)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160531483, end_position_on_the_genomic_accession=160666375, orientation='minus', exon_count=40, OMIM=152200, no_hdr0=''),
  4019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4019, CurrentID=0, Status='live', Symbol='LPAL1', Aliases='', description='lipoprotein, Lp(a)-like 1', other_designations='', map_location='6q26-q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4023, CurrentID=0, Status='live', Symbol='LPL', Aliases='HDLCQ11, LIPD', description='lipoprotein lipase', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19939071, end_position_on_the_genomic_accession=19967259, orientation='plus', exon_count=10, OMIM=609708, no_hdr0=''),
  4025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4025, CurrentID=0, Status='live', Symbol='LPO', Aliases='SPO', description='lactoperoxidase', other_designations='salivary peroxidase', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58237737, end_position_on_the_genomic_accession=58268536, orientation='plus', exon_count=15, OMIM=150205, no_hdr0=''),
  4026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4026, CurrentID=0, Status='live', Symbol='LPP', Aliases='', description='LIM domain containing preferred translocation partner in lipoma', other_designations='LIM protein|lipoma preferred partner', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188152143, end_position_on_the_genomic_accession=188890671, orientation='plus', exon_count=22, OMIM=600700, no_hdr0=''),
  4029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4029, CurrentID=0, Status='live', Symbol='L1RE1', Aliases='L1.2, LRE1', description='LINE1 retrotransposable element 1', other_designations='LINE retrotransposable element 1|Line-1.2 retrotransposable element', map_location='22q11.1-q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151626, no_hdr0=''),
  4030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4030, CurrentID=0, Status='live', Symbol='L1RE2', Aliases='LRE2', description='LINE1 retrotransposable element 2', other_designations='LINE retrotransposable element 2', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151628, no_hdr0=''),
  101060544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060544, CurrentID=0, Status='live', Symbol='LINC01563', Aliases='HP08942', description='long intergenic non-protein coding RNA 1563', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21075556, end_position_on_the_genomic_accession=21090613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  4033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4033, CurrentID=0, Status='live', Symbol='LRMP', Aliases='JAW1', description='lymphoid-restricted membrane protein', other_designations='protein Jaw1', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25052101, end_position_on_the_genomic_accession=25108335, orientation='plus', exon_count=25, OMIM=602003, no_hdr0=''),
  4034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4034, CurrentID=0, Status='live', Symbol='LRCH4', Aliases='LRN, LRRN1, LRRN4, PP14183', description='leucine-rich repeats and calponin homology (CH) domain containing 4', other_designations='leucine rich repeat neuronal 4|leucine-rich neuronal protein', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100574011, end_position_on_the_genomic_accession=100586188, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  4035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4035, CurrentID=0, Status='live', Symbol='LRP1', Aliases='A2MR, APOER, APR, CD91, IGFBP3R, LRPA, TGFBR5, LRP1', description='LDL receptor related protein 1', other_designations='TbetaR-V/LRP-1/IGFBP-3 receptor|alpha-2-macroglobulin receptor|apolipoprotein E receptor|low density lipoprotein receptor-related protein 1|type V tgf-beta receptor', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57128499, end_position_on_the_genomic_accession=57213359, orientation='plus', exon_count=89, OMIM=107770, no_hdr0=''),
  4036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4036, CurrentID=0, Status='live', Symbol='LRP2', Aliases='DBS, GP330', description='LDL receptor related protein 2', other_designations='Heymann nephritis antigen homolog|LRP-2|calcium sensor protein|glycoprotein 330|low density lipoprotein receptor-related protein 2|megalin', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169127109, end_position_on_the_genomic_accession=169362613, orientation='minus', exon_count=81, OMIM=600073, no_hdr0=''),
  4037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4037, CurrentID=0, Status='live', Symbol='LRP3', Aliases='', description='LDL receptor related protein 3', other_designations='105 kDa low-density lipoprotein receptor-related protein|CTD-2540B15.13|LRP-3|hLRp105|low density lipoprotein receptor-related protein 3', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33194693, end_position_on_the_genomic_accession=33208867, orientation='plus', exon_count=7, OMIM=603159, no_hdr0=''),
  4038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4038, CurrentID=0, Status='live', Symbol='LRP4', Aliases='CLSS, CMS17, LRP-4, LRP10, MEGF7, SOST2', description='LDL receptor related protein 4', other_designations='low density lipoprotein receptor-related protein 4|multiple epidermal growth factor-like domains 7', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46856717, end_position_on_the_genomic_accession=46918622, orientation='minus', exon_count=41, OMIM=604270, no_hdr0=''),
  100270753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270753, CurrentID=0, Status='live', Symbol='RPL21P112', Aliases='RPL21_45_1329', description='ribosomal protein L21 pseudogene 112', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94760514, end_position_on_the_genomic_accession=94761053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4040, CurrentID=0, Status='live', Symbol='LRP6', Aliases='ADCAD2', description='LDL receptor related protein 6', other_designations='LRP-6|low density lipoprotein receptor-related protein 6', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12116025, end_position_on_the_genomic_accession=12267007, orientation='minus', exon_count=23, OMIM=603507, no_hdr0=''),
  4041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4041, CurrentID=0, Status='live', Symbol='LRP5', Aliases='BMND1, EVR1, EVR4, HBM, LR3, LRP-5, LRP7, OPPG, OPS, OPTA1, VBCH2', description='LDL receptor related protein 5', other_designations='low density lipoprotein receptor-related protein 5|low density lipoprotein receptor-related protein 7', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68298866, end_position_on_the_genomic_accession=68449275, orientation='plus', exon_count=27, OMIM=603506, no_hdr0=''),
  135114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135114, CurrentID=0, Status='live', Symbol='HINT3', Aliases='HINT4', description='histidine triad nucleotide binding protein 3', other_designations='HINT-3|HIT-like protein', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125956715, end_position_on_the_genomic_accession=125980244, orientation='plus', exon_count=5, OMIM=609998, no_hdr0=''),
  4043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4043, CurrentID=0, Status='live', Symbol='LRPAP1', Aliases='A2MRAP, A2RAP, HBP44, MRAP, MYP23, RAP, alpha-2-MRAP', description='LDL receptor related protein associated protein 1', other_designations='39 kDa receptor-associated protein|low density lipoprotein receptor-related protein associated protein 1|low density lipoprotein-related protein-associated protein 1 (alpha-2-macroglobulin receptor-associated protein 1)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3503597, end_position_on_the_genomic_accession=3532497, orientation='minus', exon_count=9, OMIM=104225, no_hdr0=''),
  4045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4045, CurrentID=0, Status='live', Symbol='LSAMP', Aliases='IGLON3, LAMP', description='limbic system-associated membrane protein', other_designations='IgLON family member 3', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115802363, end_position_on_the_genomic_accession=116445542, orientation='minus', exon_count=14, OMIM=603241, no_hdr0=''),
  4046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4046, CurrentID=0, Status='live', Symbol='LSP1', Aliases='WP34, pp52', description='lymphocyte-specific protein 1', other_designations='47 kDa actin binding protein|47 kDa actin-binding protein|52 kDa phosphoprotein|F-actin binding and cytoskeleton associated protein|leufactin (leukocyte F-actin binding protein)|leukocyte-specific protein 1|lymphocyte-specific antigen WP34', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1852970, end_position_on_the_genomic_accession=1892263, orientation='plus', exon_count=16, OMIM=153432, no_hdr0=''),
  4047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4047, CurrentID=0, Status='live', Symbol='LSS', Aliases='OSC', description='lanosterol synthase (2,3-oxidosqualene-lanosterol cyclase)', other_designations='2,3-epoxysqualene--lanosterol cyclase|2,3-epoxysqualene-lanosterol cyclase|hOSC|oxidosqualene--lanosterol cyclase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46188446, end_position_on_the_genomic_accession=46228824, orientation='minus', exon_count=25, OMIM=600909, no_hdr0=''),
  4048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4048, CurrentID=0, Status='live', Symbol='LTA4H', Aliases='', description='leukotriene A4 hydrolase', other_designations='LTA-4 hydrolase', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96000753, end_position_on_the_genomic_accession=96043520, orientation='minus', exon_count=21, OMIM=151570, no_hdr0=''),
  4049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4049, CurrentID=0, Status='live', Symbol='LTA', Aliases='LT, TNFB, TNFSF1', description='lymphotoxin alpha', other_designations='LT-alpha|TNF superfamily, member 1|TNF-beta|tumor necrosis factor beta|tumor necrosis factor ligand superfamily member 1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31560550, end_position_on_the_genomic_accession=31574324, orientation='plus', exon_count=8, OMIM=153440, no_hdr0=''),
  4050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4050, CurrentID=0, Status='live', Symbol='LTB', Aliases='TNFC, TNFSF3, p33', description='lymphotoxin beta', other_designations='LT-beta|TNF-C|lymphotoxin beta (TNF superfamily, member 3)|tumor necrosis factor C|tumor necrosis factor ligand superfamily member 3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31580558, end_position_on_the_genomic_accession=31582425, orientation='minus', exon_count=4, OMIM=600978, no_hdr0=''),
  4051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4051, CurrentID=0, Status='live', Symbol='CYP4F3', Aliases='CPF3, CYP4F, LTB4H', description='cytochrome P450 family 4 subfamily F member 3', other_designations='20-HETE synthase|20-hydroxyeicosatetraenoic acid synthase|CYPIVF3|cytochrome P-450|cytochrome P450 4F3|cytochrome P450, family 4, subfamily F, polypeptide 3|cytochrome P450, subfamily IVF, polypeptide 3 (leukotriene B4 omega hydroxylase)|cytochrome P450-LTB-omega|leukotriene B4 omega hydroxylase|leukotriene-B(4) 20-monooxygenase 2|leukotriene-B(4) omega-hydroxylase 2|leukotriene-B4 20-monooxygenase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15640897, end_position_on_the_genomic_accession=15660760, orientation='plus', exon_count=14, OMIM=601270, no_hdr0=''),
  4052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4052, CurrentID=0, Status='live', Symbol='LTBP1', Aliases='', description='latent transforming growth factor beta binding protein 1', other_designations='LTBP-1|TGF-beta1-BP-1|transforming growth factor beta-1-binding protein 1', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32947300, end_position_on_the_genomic_accession=33399509, orientation='plus', exon_count=36, OMIM=150390, no_hdr0=''),
  4053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4053, CurrentID=0, Status='live', Symbol='LTBP2', Aliases='C14orf141, GLC3D, LTBP3, MSPKA, MSTP031, WMS3', description='latent transforming growth factor beta binding protein 2', other_designations='LTBP-2', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74498183, end_position_on_the_genomic_accession=74618609, orientation='minus', exon_count=38, OMIM=602091, no_hdr0=''),
  4054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4054, CurrentID=0, Status='live', Symbol='LTBP3', Aliases='DASS, LTBP-3, LTBP2, STHAG6, pp6425', description='latent transforming growth factor beta binding protein 3', other_designations='latent TGF beta binding protein 3', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65538559, end_position_on_the_genomic_accession=65558361, orientation='minus', exon_count=28, OMIM=602090, no_hdr0=''),
  4055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4055, CurrentID=0, Status='live', Symbol='LTBR', Aliases='CD18, D12S370, LT-BETA-R, TNF-R-III, TNFCR, TNFR-RP, TNFR2-RP, TNFR3, TNFRSF3', description='lymphotoxin beta receptor', other_designations='lymphotoxin B receptor|lymphotoxin beta receptor (TNFR superfamily, member 3)|tumor necrosis factor C receptor|tumor necrosis factor receptor 2-related protein|tumor necrosis factor receptor type III', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6375368, end_position_on_the_genomic_accession=6391571, orientation='plus', exon_count=12, OMIM=600979, no_hdr0=''),
  4056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4056, CurrentID=0, Status='live', Symbol='LTC4S', Aliases='', description='leukotriene C4 synthase', other_designations='LTC4 synthase', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179793985, end_position_on_the_genomic_accession=179796512, orientation='plus', exon_count=5, OMIM=246530, no_hdr0=''),
  4057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4057, CurrentID=0, Status='live', Symbol='LTF', Aliases='GIG12, HEL110, HLF2, LF', description='lactotransferrin', other_designations='epididymis luminal protein 110|growth-inhibiting protein 12|kaliocin-1|lactoferricin|lactoferroxin|neutrophil lactoferrin|talalactoferrin', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46436005, end_position_on_the_genomic_accession=46465142, orientation='minus', exon_count=18, OMIM=150210, no_hdr0=''),
  4058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4058, CurrentID=0, Status='live', Symbol='LTK', Aliases='TYK1', description='leukocyte receptor tyrosine kinase', other_designations='protein tyrosine kinase 1', map_location='15q15.1-q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41503637, end_position_on_the_genomic_accession=41513887, orientation='minus', exon_count=21, OMIM=151520, no_hdr0=''),
  4059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4059, CurrentID=0, Status='live', Symbol='BCAM', Aliases='AU, CD239, LU, MSK19', description='basal cell adhesion molecule (Lutheran blood group)', other_designations='Auberger b antigen|B-CAM cell surface glycoprotein|B-cell adhesion molecule|F8/G253 antigen|antigen identified by monoclonal antibody F8|basal cell adhesion molecule (Lu and Au blood groups)|glycoprotein 95kDa', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44809059, end_position_on_the_genomic_accession=44821421, orientation='plus', exon_count=14, OMIM=612773, no_hdr0=''),
  4060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4060, CurrentID=0, Status='live', Symbol='LUM', Aliases='LDC, SLRR2D', description='lumican', other_designations='KSPG lumican|keratan sulfate proteoglycan lumican|lumican proteoglycan', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91103455, end_position_on_the_genomic_accession=91111765, orientation='minus', exon_count=3, OMIM=600616, no_hdr0=''),
  4061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4061, CurrentID=0, Status='live', Symbol='LY6E', Aliases='RIG-E, RIGE, SCA-2, SCA2, TSA-1', description='lymphocyte antigen 6 complex, locus E', other_designations='ly-6E|retinoic acid induced gene E|retinoic acid-induced gene E protein|stem cell antigen 2|thymic shared antigen 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143018485, end_position_on_the_genomic_accession=143022410, orientation='plus', exon_count=4, OMIM=601384, no_hdr0=''),
  4062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4062, CurrentID=0, Status='live', Symbol='LY6H', Aliases='NMLY6', description='lymphocyte antigen 6 complex, locus H', other_designations='ly-6H', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143157914, end_position_on_the_genomic_accession=143160636, orientation='minus', exon_count=6, OMIM=603625, no_hdr0=''),
  4063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4063, CurrentID=0, Status='live', Symbol='LY9', Aliases='CD229, SLAMF3, hly9, mLY9', description='lymphocyte antigen 9', other_designations='SLAM family member 3|cell surface molecule Ly-9|signaling lymphocytic activation molecule 3', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160796074, end_position_on_the_genomic_accession=160828256, orientation='plus', exon_count=12, OMIM=600684, no_hdr0=''),
  4064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4064, CurrentID=0, Status='live', Symbol='CD180', Aliases='LY64, Ly78, RP105', description='CD180 molecule', other_designations='lymphocyte antigen 64|lymphocyte antigen-64, radioprotective, 105kDa|radioprotective 105 kDa protein', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67181009, end_position_on_the_genomic_accession=67196789, orientation='minus', exon_count=4, OMIM=602226, no_hdr0=''),
  4065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4065, CurrentID=0, Status='live', Symbol='LY75', Aliases='CD205, CLEC13B, DEC-205, GP200-MR6, LY-75', description='lymphocyte antigen 75', other_designations='C-type lectin domain family 13 member B', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159803355, end_position_on_the_genomic_accession=159904756, orientation='minus', exon_count=35, OMIM=604524, no_hdr0=''),
  135138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135138, CurrentID=0, Status='live', Symbol='PACRG', Aliases='GLUP, HAK005771, PARK2CRG', description='PARK2 co-regulated', other_designations='molecular chaperone/chaperonin-binding protein|parkin co-regulated gene protein', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162727084, end_position_on_the_genomic_accession=163315492, orientation='plus', exon_count=27, OMIM=608427, no_hdr0=''),
  4067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4067, CurrentID=0, Status='live', Symbol='LYN', Aliases='JTK8, p53Lyn, p56Lyn', description='LYN proto-oncogene, Src family tyrosine kinase', other_designations='lck/Yes-related novel protein tyrosine kinase|v-yes-1 Yamaguchi sarcoma viral related oncogene homolog', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55878463, end_position_on_the_genomic_accession=56012447, orientation='plus', exon_count=16, OMIM=165120, no_hdr0=''),
  4068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4068, CurrentID=0, Status='live', Symbol='SH2D1A', Aliases='DSHP, EBVS, IMD5, LYP, MTCP1, SAP, SAP/SH2D1A, XLP, XLPD, XLPD1', description='SH2 domain containing 1A', other_designations='Duncan disease SH2-protein|SLAM associated protein/SH2 domain protein 1A|SLAM-associated protein|T cell signal transduction molecule SAP|T-cell signal transduction molecule SAP|signaling lymphocyte activation molecule-associated protein|signaling lymphocytic activation molecule-associated protein', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124346282, end_position_on_the_genomic_accession=124373160, orientation='plus', exon_count=4, OMIM=300490, no_hdr0=''),
  4069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4069, CurrentID=0, Status='live', Symbol='LYZ', Aliases='LZM', description='lysozyme', other_designations='1,4-beta-N-acetylmuramidase C|c-type lysozyme', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69348354, end_position_on_the_genomic_accession=69354233, orientation='plus', exon_count=4, OMIM=153450, no_hdr0=''),
  4070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4070, CurrentID=0, Status='live', Symbol='TACSTD2', Aliases='EGP-1, EGP1, GA733-1, GA7331, GP50, M1S1, TROP2', description='tumor-associated calcium signal transducer 2', other_designations='40kD glycoprotein, identified by monoclonal antibody GA733|cell surface glycoprotein TROP2|cell surface glycoprotein Trop-2|epithelial glycoprotein-1|gastrointestinal tumor-associated antigen GA7331|membrane component chromosome 1 surface marker 1|membrane component, chromosome 1, surface marker 1|pancreatic carcinoma marker protein GA733-1|pancreatic carcinoma marker protein GA7331', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58575423, end_position_on_the_genomic_accession=58577494, orientation='minus', exon_count=1, OMIM=137290, no_hdr0=''),
  4071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4071, CurrentID=0, Status='live', Symbol='TM4SF1', Aliases='H-L6, L6, M3S1, TAAL6', description='transmembrane 4 L six family member 1', other_designations='membrane component chromosome 3 surface marker 1|membrane component, chromosome 3, surface marker 1|transmembrane 4 superfamily member 1|tumor-associated antigen L6', map_location='3q21-q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149369018, end_position_on_the_genomic_accession=149377860, orientation='minus', exon_count=5, OMIM=191155, no_hdr0=''),
  4072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4072, CurrentID=0, Status='live', Symbol='EPCAM', Aliases='DIAR5, EGP-2, EGP314, EGP40, ESA, HNPCC8, KS1/4, KSA, M4S1, MIC18, MK-1, TACSTD1, TROP1', description='epithelial cell adhesion molecule', other_designations='adenocarcinoma-associated antigen|cell surface glycoprotein Trop-1|epithelial glycoprotein 314|human epithelial glycoprotein-2|major gastrointestinal tumor-associated protein GA733-2|membrane component, chromosome 4, surface marker (35kD glycoprotein)|tumor-associated calcium signal transducer 1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47369148, end_position_on_the_genomic_accession=47387028, orientation='plus', exon_count=9, OMIM=185535, no_hdr0=''),
  4074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4074, CurrentID=0, Status='live', Symbol='M6PR', Aliases='CD-MPR, MPR 46, MPR-46, MPR46, SMPR', description='mannose-6-phosphate receptor (cation dependent)', other_designations='46-kDa mannose 6-phosphate receptor|CD Man-6-P receptor|Mr 46,000 Man6PR|small mannose 6-phosphate receptor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8940361, end_position_on_the_genomic_accession=8949761, orientation='minus', exon_count=9, OMIM=154540, no_hdr0=''),
  101060587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060587, CurrentID=0, Status='live', Symbol='LOC101060587', Aliases='', description='uncharacterized LOC101060587', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28341768, end_position_on_the_genomic_accession=28345732, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4076, CurrentID=0, Status='live', Symbol='CAPRIN1', Aliases='GPIAP1, GPIP137, GRIP137, M11S1, RNG105, p137GPI', description='cell cycle associated protein 1', other_designations='GPI-anchored membrane protein 1|GPI-anchored protein p137|GPI-p137|RNA granule protein 105|activation/proliferation-associated protein 1|caprin 1|cell cycle-associated protein 1|cytoplasmic activation- and proliferation-associated protein 1|cytoplasmic activation/proliferation-associated protein-1|membrane component chromosome 11 surface marker 1|membrane component, chromosome 11, surface marker 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34051617, end_position_on_the_genomic_accession=34102610, orientation='plus', exon_count=19, OMIM=601178, no_hdr0=''),
  4077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4077, CurrentID=0, Status='live', Symbol='NBR1', Aliases='1A1-3B, IAI3B, M17S2, MIG19', description='neighbor of BRCA1 gene 1', other_designations='B-box protein|cell migration-inducing gene 19 protein|membrane component, chromosome 17, surface marker 2 (ovarian carcinoma antigen CA125)|migration-inducing protein 19', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43170471, end_position_on_the_genomic_accession=43211689, orientation='plus', exon_count=23, OMIM=166945, no_hdr0=''),
  284669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284669, CurrentID=0, Status='live', Symbol='LOC284669', Aliases='', description='uncharacterized LOC284669', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  135152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135152, CurrentID=0, Status='live', Symbol='B3GAT2', Aliases='GLCATS', description='beta-1,3-glucuronyltransferase 2', other_designations='UDP-glucuronosyltransferase S|UDP-glucuronyltransferase S|glcAT-D|glucuronosyltransferase S|uridine diphosphate glucuronic acid:acceptor glucuronosyltransferase', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70861366, end_position_on_the_genomic_accession=70957189, orientation='minus', exon_count=3, OMIM=607497, no_hdr0=''),
  4081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4081, CurrentID=0, Status='live', Symbol='MAB21L1', Aliases='CAGR1, Nbla00126', description='mab-21-like 1 (C. elegans)', other_designations='mab-21-like protein 1', map_location='13q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35473789, end_position_on_the_genomic_accession=35476695, orientation='minus', exon_count=1, OMIM=601280, no_hdr0=''),
  4082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4082, CurrentID=0, Status='live', Symbol='MARCKS', Aliases='80K-L, MACS, PKCSL, PRKCSL', description='myristoylated alanine-rich protein kinase C substrate', other_designations='myristoylated alanine-rich protein kinase C substrate (MARCKS, 80K-L)|phosphomyristin|protein kinase C substrate, 80 kDa protein, light chain', map_location='6q22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113857335, end_position_on_the_genomic_accession=113863475, orientation='plus', exon_count=2, OMIM=177061, no_hdr0=''),
  4083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4083, CurrentID=0, Status='live', Symbol='MARCKSP1', Aliases='MACSL1', description='myristoylated alanine-rich protein kinase C substrate pseudogene 1', other_designations='myristoylated alanine-rich protein kinase C substrate (MARCKS, 80K-L)-like 1', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26227560, end_position_on_the_genomic_accession=26235914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4084, CurrentID=0, Status='live', Symbol='MXD1', Aliases='BHLHC58, MAD, MAD1', description='MAX dimerization protein 1', other_designations='MAX-binding protein|antagonizer of myc transcriptional activity|max dimerizer 1', map_location='2p13-p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69915041, end_position_on_the_genomic_accession=69942945, orientation='plus', exon_count=6, OMIM=600021, no_hdr0=''),
  4085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4085, CurrentID=0, Status='live', Symbol='MAD2L1', Aliases='HSMAD2, MAD2', description='MAD2 mitotic arrest deficient-like 1 (yeast)', other_designations='MAD2 (mitotic arrest deficient, yeast, homolog)-like 1|MAD2-like protein 1|mitotic arrest deficient 2-like protein 1|mitotic arrest deficient, yeast, homolog-like 1', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120059422, end_position_on_the_genomic_accession=120066858, orientation='minus', exon_count=5, OMIM=601467, no_hdr0=''),
  4086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4086, CurrentID=0, Status='live', Symbol='SMAD1', Aliases='BSP-1, BSP1, JV4-1, JV41, MADH1, MADR1', description='SMAD family member 1', other_designations='MAD, mothers against decapentaplegic homolog 1|Mad-related protein 1|SMAD, mothers against DPP homolog 1|TGF-beta signaling protein 1|mothers against DPP homolog 1|transforming growth factor-beta signaling protein 1', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145481306, end_position_on_the_genomic_accession=145559176, orientation='plus', exon_count=14, OMIM=601595, no_hdr0=''),
  4087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4087, CurrentID=0, Status='live', Symbol='SMAD2', Aliases='JV18, JV18-1, MADH2, MADR2, hMAD-2, hSMAD2', description='SMAD family member 2', other_designations='MAD homolog 2|SMAD, mothers against DPP homolog 2|Sma- and Mad-related protein 2|mother against DPP homolog 2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47833095, end_position_on_the_genomic_accession=47931193, orientation='minus', exon_count=16, OMIM=601366, no_hdr0=''),
  4088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4088, CurrentID=0, Status='live', Symbol='SMAD3', Aliases='HSPC193, HsT17436, JV15-2, LDS1C, LDS3, MADH3', description='SMAD family member 3', other_designations='MAD homolog 3|MAD, mothers against decapentaplegic homolog 3|SMA- and MAD-related protein 3|SMAD, mothers against DPP homolog 3|hMAD-3|hSMAD3|mad homolog JV15-2|mad protein homolog|mad3|mothers against DPP homolog 3', map_location='15q22.33', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67065627, end_position_on_the_genomic_accession=67195195, orientation='plus', exon_count=13, OMIM=603109, no_hdr0=''),
  4089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4089, CurrentID=0, Status='live', Symbol='SMAD4', Aliases='DPC4, JIP, MADH4, MYHRS', description='SMAD family member 4', other_designations='MAD homolog 4|SMAD, mothers against DPP homolog 4|deleted in pancreatic carcinoma locus 4|deletion target in pancreatic carcinoma 4|mothers against decapentaplegic, Drosophila, homolog of, 4', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51030213, end_position_on_the_genomic_accession=51085042, orientation='plus', exon_count=12, OMIM=600993, no_hdr0=''),
  4090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4090, CurrentID=0, Status='live', Symbol='SMAD5', Aliases='DWFC, JV5-1, MADH5', description='SMAD family member 5', other_designations='MAD, mothers against decapentaplegic homolog 5|SMA- and MAD-related protein 5|SMAD, mothers against DPP homolog 5|mothers against decapentaplegic, drosophila, homolog of, 5', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136132845, end_position_on_the_genomic_accession=136182734, orientation='plus', exon_count=9, OMIM=603110, no_hdr0=''),
  4091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4091, CurrentID=0, Status='live', Symbol='SMAD6', Aliases='AOVD2, HsT17432, MADH6, MADH7', description='SMAD family member 6', other_designations='MAD homolog 6|SMAD, mothers against DPP homolog 6|mothers against decapentaplegic, drosophila, homolog of, 6', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66702100, end_position_on_the_genomic_accession=66782000, orientation='plus', exon_count=7, OMIM=602931, no_hdr0=''),
  4092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4092, CurrentID=0, Status='live', Symbol='SMAD7', Aliases='CRCS3, MADH7, MADH8', description='SMAD family member 7', other_designations='MAD (mothers against decapentaplegic, Drosophila) homolog 7|MAD homolog 8|Mothers against decapentaplegic, drosophila, homolog of, 7|SMAD, mothers against DPP homolog 7|hSMAD7|mothers against DPP homolog 8', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48919853, end_position_on_the_genomic_accession=48950711, orientation='minus', exon_count=5, OMIM=602932, no_hdr0=''),
  4093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4093, CurrentID=0, Status='live', Symbol='SMAD9', Aliases='MADH6, MADH9, PPH2, SMAD8, SMAD8A, SMAD8B', description='SMAD family member 9', other_designations='MAD homolog 9|Mothers against decapentaplegic, drosophila, homolog of, 9|SMAD, mothers against DPP homolog 9', map_location='13q12-q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36815914, end_position_on_the_genomic_accession=36920721, orientation='minus', exon_count=10, OMIM=603295, no_hdr0=''),
  4094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4094, CurrentID=0, Status='live', Symbol='MAF', Aliases='AYGRP, CCA4, CTRCT21, c-MAF', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog', other_designations='Avian musculoaponeurotic fibrosarcoma (MAF) protooncogene|T lymphocyte c-maf long form|c-maf proto-oncogene|proto-oncogene c-Maf', map_location='16q22-q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79212772, end_position_on_the_genomic_accession=79600786, orientation='minus', exon_count=4, OMIM=177075, no_hdr0=''),
  4095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4095, CurrentID=0, Status='live', Symbol='MAFD1', Aliases='BPAD, MD1', description='major affective disorder 1', other_designations='', map_location='18p', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=125480, no_hdr0=''),
  4096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4096, CurrentID=0, Status='live', Symbol='MAFD2', Aliases='BPAD, MDI, MDX', description='major affective disorder 2', other_designations='', map_location='Xq27-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309200, no_hdr0=''),
  4097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4097, CurrentID=0, Status='live', Symbol='MAFG', Aliases='hMAF', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog G', other_designations='basic leucine zipper transcription factor MafG', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81918269, end_position_on_the_genomic_accession=81931560, orientation='minus', exon_count=6, OMIM=602020, no_hdr0=''),
  4099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4099, CurrentID=0, Status='live', Symbol='MAG', Aliases='GMA, S-MAG, SIGLEC-4A, SIGLEC4A', description='myelin associated glycoprotein', other_designations='sialic acid binding Ig-like lectin 4A|sialic acid-binding immunoglobulin-like lectin 4A', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35292086, end_position_on_the_genomic_accession=35313807, orientation='plus', exon_count=13, OMIM=159460, no_hdr0=''),
  4100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4100, CurrentID=0, Status='live', Symbol='MAGEA1', Aliases='CT1.1, MAGE1', description='MAGE family member A1', other_designations='MAGE-1 antigen|antigen MZ2-E|cancer/testis antigen 1.1|cancer/testis antigen family 1, member 1|melanoma antigen MAGE-1|melanoma antigen family A 1|melanoma antigen family A, 1 (directs expression of antigen MZ2-E)|melanoma antigen family A1|melanoma-associated antigen MZ2-E', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153179284, end_position_on_the_genomic_accession=153183878, orientation='plus', exon_count=3, OMIM=300016, no_hdr0=''),
  4101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4101, CurrentID=0, Status='live', Symbol='MAGEA2', Aliases='CT1.2, MAGE2A, MAGEA2', description='MAGE family member A2', other_designations='MAGE-2 antigen|cancer/testis antigen 1.2|cancer/testis antigen family 1, member 2|melanoma antigen 2|melanoma antigen family A, 2|melanoma antigen family A2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152749863, end_position_on_the_genomic_accession=152753884, orientation='minus', exon_count=7, OMIM=300173, no_hdr0=''),
  4102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4102, CurrentID=0, Status='live', Symbol='MAGEA3', Aliases='CT1.3, HIP8, HYPD, MAGE3, MAGEA6', description='MAGE family member A3', other_designations='MAGE-3 antigen|antigen MZ2-D|cancer/testis antigen 1.3|cancer/testis antigen family 1, member 3|melanoma antigen family A, 3|melanoma antigen family A3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152698732, end_position_on_the_genomic_accession=152702347, orientation='plus', exon_count=6, OMIM=300174, no_hdr0=''),
  4103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4103, CurrentID=0, Status='live', Symbol='MAGEA4', Aliases='CT1.4, MAGE-41, MAGE-X2, MAGE4, MAGE4A, MAGE4B', description='MAGE family member A4', other_designations='MAGE-4 antigen|MAGE-41 antigen|MAGE-X2 antigen|cancer/testis antigen 1.4|cancer/testis antigen family 1, member 4|melanoma antigen family A, 4|melanoma antigen family A4', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151912889, end_position_on_the_genomic_accession=151925170, orientation='plus', exon_count=8, OMIM=300175, no_hdr0=''),
  4104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4104, CurrentID=0, Status='live', Symbol='MAGEA5', Aliases='CT1.5, MAGE5, MAGEA4', description='MAGE family member A5', other_designations='MAGE-5 antigen|MAGE-5a antigen|MAGE-5b antigen|cancer/testis antigen 1.5|cancer/testis antigen family 1, member 5|melanoma antigen family A, 5|melanoma antigen family A5', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152114049, end_position_on_the_genomic_accession=152117939, orientation='minus', exon_count=3, OMIM=300340, no_hdr0=''),
  4105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4105, CurrentID=0, Status='live', Symbol='MAGEA6', Aliases='CT1.6, MAGE-3b, MAGE3B, MAGE6', description='MAGE family member A6', other_designations='MAGE-6 antigen|MAGE3B antigen|cancer/testis antigen 1.6|cancer/testis antigen family 1, member 6|melanoma antigen family A, 6|melanoma antigen family A6', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152766136, end_position_on_the_genomic_accession=152769716, orientation='minus', exon_count=4, OMIM=300176, no_hdr0=''),
  4106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4106, CurrentID=0, Status='live', Symbol='MAGEA7P', Aliases='MAGE7, MAGEA7', description='MAGE family member A7, pseudogene', other_designations='melanoma antigen family A, 7, pseudogene|melanoma antigen family A7, pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149808344, end_position_on_the_genomic_accession=149809010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4107, CurrentID=0, Status='live', Symbol='MAGEA8', Aliases='CT1.8, MAGE8', description='MAGE family member A8', other_designations='MAGE-8 antigen|cancer/testis antigen 1.8|cancer/testis antigen family 1, member 8|melanoma antigen family A, 8|melanoma antigen family A8', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149881141, end_position_on_the_genomic_accession=149885835, orientation='plus', exon_count=4, OMIM=300341, no_hdr0=''),
  4108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4108, CurrentID=0, Status='live', Symbol='MAGEA9', Aliases='CT1.9, MAGE9', description='MAGE family member A9', other_designations='MAGE-9 antigen|cancer/testis antigen 1.9|cancer/testis antigen family 1, member 9|melanoma antigen family A, 9|melanoma antigen family A9', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149781921, end_position_on_the_genomic_accession=149787739, orientation='plus', exon_count=5, OMIM=300342, no_hdr0=''),
  4109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4109, CurrentID=0, Status='live', Symbol='MAGEA10', Aliases='CT1.10, MAGE10', description='MAGE family member A10', other_designations='MAGE-10 antigen|cancer/testis antigen 1.10|cancer/testis antigen family 1, member 10|melanoma antigen family A, 10|melanoma antigen family A10', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152133310, end_position_on_the_genomic_accession=152138578, orientation='minus', exon_count=5, OMIM=300343, no_hdr0=''),
  4110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4110, CurrentID=0, Status='live', Symbol='MAGEA11', Aliases='CT1.11, MAGE-11, MAGE11, MAGEA-11', description='MAGE family member A11', other_designations='MAGE-11 antigen|cancer/testis antigen 1.11|cancer/testis antigen family 1, member 11|melanoma antigen family A, 11|melanoma antigen family A11', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149688198, end_position_on_the_genomic_accession=149717268, orientation='plus', exon_count=10, OMIM=300344, no_hdr0=''),
  4111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4111, CurrentID=0, Status='live', Symbol='MAGEA12', Aliases='CT1.12, MAGE12', description='MAGE family member A12', other_designations='MAGE12F antigen|cancer/testis antigen 1.12|cancer/testis antigen family 1, member 12|melanoma antigen family A, 12|melanoma antigen family A12', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152733779, end_position_on_the_genomic_accession=152737669, orientation='plus', exon_count=4, OMIM=300177, no_hdr0=''),
  4112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4112, CurrentID=0, Status='live', Symbol='MAGEB1', Aliases='CT3.1, DAM10, MAGE-Xp, MAGEL1', description='MAGE family member B1', other_designations='DSS-AHC critical interval MAGE superfamily 10|DSS/AHC critical interval MAGE superfamily 10|MAGE-B1 antigen|MAGE-XP antigen|cancer/testis antigen 3.1|cancer/testis antigen family 3, member 1|melanoma antigen family B, 1|melanoma antigen family B1', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30243731, end_position_on_the_genomic_accession=30252038, orientation='plus', exon_count=5, OMIM=300097, no_hdr0=''),
  4113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4113, CurrentID=0, Status='live', Symbol='MAGEB2', Aliases='CT3.2, DAM6, MAGE-XP-2', description='MAGE family member B2', other_designations='DSS-AHC critical interval MAGE superfamily 6|DSS/AHC critical interval MAGE superfamily 6|MAGE XP-2 antigen|MAGE-B2 antigen|cancer/testis antigen 3.2|cancer/testis antigen family 3, member 2|melanoma antigen family B, 2|melanoma antigen family B2', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30215558, end_position_on_the_genomic_accession=30220089, orientation='plus', exon_count=2, OMIM=300098, no_hdr0=''),
  4114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4114, CurrentID=0, Status='live', Symbol='MAGEB3', Aliases='CT3.5', description='MAGE family member B3', other_designations='MAGE-B3 antigen|cancer/testis antigen family 3, member 5|melanoma antigen family B, 3|melanoma antigen family B3', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30230436, end_position_on_the_genomic_accession=30237493, orientation='plus', exon_count=6, OMIM=300152, no_hdr0=''),
  4115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4115, CurrentID=0, Status='live', Symbol='MAGEB4', Aliases='CT3.6', description='MAGE family member B4', other_designations='MAGE-B4 antigen|cancer/testis antigen family 3, member 6|melanoma antigen family B, 4|melanoma antigen family B4', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30241940, end_position_on_the_genomic_accession=30244193, orientation='plus', exon_count=1, OMIM=300153, no_hdr0=''),
  4116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4116, CurrentID=0, Status='live', Symbol='MAGOH', Aliases='MAGOH1A, MAGOH', description='mago homolog, exon junction complex core component', other_designations='mago-nashi homolog, proliferation-associated', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53226892, end_position_on_the_genomic_accession=53238610, orientation='minus', exon_count=5, OMIM=602603, no_hdr0=''),
  4117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4117, CurrentID=0, Status='live', Symbol='MAK', Aliases='RP62, dJ417M14.2', description='male germ cell-associated kinase', other_designations='male germ cell-associated kinase retinal-enriched isoform|serine/threonine protein kinase MAK', map_location='6p24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10762723, end_position_on_the_genomic_accession=10838605, orientation='minus', exon_count=20, OMIM=154235, no_hdr0=''),
  4118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4118, CurrentID=0, Status='live', Symbol='MAL', Aliases='', description='mal, T-cell differentiation protein', other_designations='T-lymphocyte maturation-associated protein', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95025655, end_position_on_the_genomic_accession=95053992, orientation='plus', exon_count=4, OMIM=188860, no_hdr0=''),
  101060632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060632, CurrentID=0, Status='live', Symbol='LOC101060632', Aliases='', description='putative ankyrin repeat domain-containing protein 20A12 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4121, CurrentID=0, Status='live', Symbol='MAN1A1', Aliases='HUMM3, HUMM9, MAN9', description='mannosidase, alpha, class 1A, member 1', other_designations='Man9-mannosidase|alpha-1,2-mannosidase IA|man(9)-alpha-mannosidase|processing alpha-1,2-mannosidase IA', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119177201, end_position_on_the_genomic_accession=119351633, orientation='minus', exon_count=16, OMIM=604344, no_hdr0=''),
  4122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4122, CurrentID=0, Status='live', Symbol='MAN2A2', Aliases='MANA2X', description='mannosidase, alpha, class 2A, member 2', other_designations='MAN IIx|alpha-mannosidase IIx|mannosidase, alpha type II-X|mannosyl-oligosaccharide 1,3-1,6-alpha-mannosidase|manosidase, alpha-, type II, isozyme X', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90902218, end_position_on_the_genomic_accession=90922585, orientation='plus', exon_count=26, OMIM=600988, no_hdr0=''),
  4123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4123, CurrentID=0, Status='live', Symbol='MAN2C1', Aliases='MAN6A8, MANA, MANA1', description='mannosidase, alpha, class 2C, member 1', other_designations='alpha mannosidase 6A8B|alpha-D-mannoside mannohydrolase|mannosidase, alpha 6A8|mannosidase, alpha A, cytoplasmic', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75355792, end_position_on_the_genomic_accession=75369114, orientation='minus', exon_count=27, OMIM=154580, no_hdr0=''),
  4124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4124, CurrentID=0, Status='live', Symbol='MAN2A1', Aliases='AMan II, GOLIM7, MANA2, MANII', description='mannosidase, alpha, class 2A, member 1', other_designations='Golgi alpha-mannosidase II|golgi integral membrane protein 7|mannosidase, alpha type II|mannosyl-oligosaccharide 1,3-1,6-alpha-mannosidase', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109689366, end_position_on_the_genomic_accession=109869625, orientation='plus', exon_count=25, OMIM=154582, no_hdr0=''),
  4125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4125, CurrentID=0, Status='live', Symbol='MAN2B1', Aliases='LAMAN, MANB', description='mannosidase, alpha, class 2B, member 1', other_designations='lysosomal acid alpha-mannosidase|mannosidase alpha-B|mannosidase, alpha B, lysosomal', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12646508, end_position_on_the_genomic_accession=12666777, orientation='minus', exon_count=24, OMIM=609458, no_hdr0=''),
  4126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4126, CurrentID=0, Status='live', Symbol='MANBA', Aliases='MANB1', description='mannosidase, beta A, lysosomal', other_designations='lysosomal beta A mannosidase|mannanase|mannase', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102631486, end_position_on_the_genomic_accession=102760994, orientation='minus', exon_count=19, OMIM=609489, no_hdr0=''),
  4127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4127, CurrentID=0, Status='live', Symbol='MANBB', Aliases='', description='mannosidase, beta B, soluble', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4128, CurrentID=0, Status='live', Symbol='MAOA', Aliases='MAO-A', description='monoamine oxidase A', other_designations='monoamine oxidase type A', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43654907, end_position_on_the_genomic_accession=43746824, orientation='plus', exon_count=16, OMIM=309850, no_hdr0=''),
  4129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4129, CurrentID=0, Status='live', Symbol='MAOB', Aliases='', description='monoamine oxidase B', other_designations='MAO, brain|MAO, platelet|MAO-B|adrenalin oxidase|monoamine oxidase type B|tyramine oxidase', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43766610, end_position_on_the_genomic_accession=43882475, orientation='minus', exon_count=17, OMIM=309860, no_hdr0=''),
  4130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4130, CurrentID=0, Status='live', Symbol='MAP1A', Aliases='MAP1L, MTAP1A', description='microtubule associated protein 1A', other_designations='MAP-1A|proliferation-related protein p80', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43517608, end_position_on_the_genomic_accession=43531620, orientation='plus', exon_count=6, OMIM=600178, no_hdr0=''),
  4131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4131, CurrentID=0, Status='live', Symbol='MAP1B', Aliases='FUTSCH, MAP5, PPP1R102', description='microtubule associated protein 1B', other_designations='MAP-1B|protein phosphatase 1, regulatory subunit 102', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72107291, end_position_on_the_genomic_accession=72209570, orientation='plus', exon_count=8, OMIM=157129, no_hdr0=''),
  101060644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060644, CurrentID=0, Status='live', Symbol='LOC101060644', Aliases='', description='TAR DNA-binding protein 43 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4133, CurrentID=0, Status='live', Symbol='MAP2', Aliases='MAP2AB, MAP2C, MAP2', description='microtubule associated protein 2', other_designations='MAP-2', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209424047, end_position_on_the_genomic_accession=209734118, orientation='plus', exon_count=22, OMIM=157130, no_hdr0=''),
  4134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4134, CurrentID=0, Status='live', Symbol='MAP4', Aliases='', description='microtubule associated protein 4', other_designations='MAP-4', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47850690, end_position_on_the_genomic_accession=48089279, orientation='minus', exon_count=27, OMIM=157132, no_hdr0=''),
  4135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4135, CurrentID=0, Status='live', Symbol='MAP6', Aliases='MAP6-N, MTAP6, N-STOP, STOP', description='microtubule associated protein 6', other_designations='CTD-2530H12.7|MAP-6|stable tubule-only polypeptide', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75586914, end_position_on_the_genomic_accession=75668730, orientation='minus', exon_count=10, OMIM=601783, no_hdr0=''),
  4137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4137, CurrentID=0, Status='live', Symbol='MAPT', Aliases='DDPAC, FTDP-17L, MSTD, MTBT1, MTBT2, PPND, PPP1R103, TAU, MAPT', description='microtubule associated protein tau', other_designations='G protein beta1/gamma2 subunit-interacting factor 1|PHF-tau|microtubule-associated protein tau|neurofibrillary tangle protein|paired helical filament-tau|protein phosphatase 1, regulatory subunit 103', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45894382, end_position_on_the_genomic_accession=46028334, orientation='plus', exon_count=15, OMIM=157140, no_hdr0=''),
  4139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4139, CurrentID=0, Status='live', Symbol='MARK1', Aliases='MARK, Par-1c, Par1c', description='MAP/microtubule affinity-regulating kinase 1', other_designations='PAR1 homolog c', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220528183, end_position_on_the_genomic_accession=220664457, orientation='plus', exon_count=21, OMIM=606511, no_hdr0=''),
  4140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4140, CurrentID=0, Status='live', Symbol='MARK3', Aliases='CTAK1, KP78, PAR1A, Par-1a', description='MAP/microtubule affinity-regulating kinase 3', other_designations='C-TAK1|ELKL motif kinase 2|EMK-2|cdc25C-associated protein kinase 1|protein kinase STK10|ser/Thr protein kinase PAR-1|serine/threonine-protein kinase p78', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103385364, end_position_on_the_genomic_accession=103503831, orientation='plus', exon_count=20, OMIM=602678, no_hdr0=''),
  4141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4141, CurrentID=0, Status='live', Symbol='MARS', Aliases='CMT2U, ILFS2, ILLD, METRS, MRS, MTRNS, SPG70', description='methionyl-tRNA synthetase', other_designations='cytosolic methionyl-tRNA synthetase', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57487953, end_position_on_the_genomic_accession=57516655, orientation='plus', exon_count=21, OMIM=156560, no_hdr0=''),
  4142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4142, CurrentID=0, Status='live', Symbol='MAS1', Aliases='MAS, MGRA', description='MAS1 proto-oncogene, G protein-coupled receptor', other_designations='MAS1 oncogene|Mas-related G protein-coupled receptor A', map_location='6q25.3-q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159906942, end_position_on_the_genomic_accession=159908076, orientation='plus', exon_count=1, OMIM=165180, no_hdr0=''),
  4143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4143, CurrentID=0, Status='live', Symbol='MAT1A', Aliases='MAT, MATA1, SAMS, SAMS1', description='methionine adenosyltransferase I, alpha', other_designations='MAT 1|MAT-I/III|S-adenosylmethionine synthetase isoform type-1|adoMet synthase 1|adoMet synthetase 1|methionine adenosyltransferase 1|methionine adenosyltransferase I/III', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80271820, end_position_on_the_genomic_accession=80289979, orientation='minus', exon_count=10, OMIM=610550, no_hdr0=''),
  4144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4144, CurrentID=0, Status='live', Symbol='MAT2A', Aliases='MATA2, MATII, SAMS2', description='methionine adenosyltransferase II, alpha', other_designations='MAT 2|MAT-II|adoMet synthase 2|adoMet synthetase 2|methionine adenosyltransferase 2', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85538978, end_position_on_the_genomic_accession=85545280, orientation='plus', exon_count=9, OMIM=601468, no_hdr0=''),
  4145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4145, CurrentID=0, Status='live', Symbol='MATK', Aliases='CHK, CTK, HHYLTK, HYL, HYLTK, Lsk', description='megakaryocyte-associated tyrosine kinase', other_designations='CSK homologous kinase|Csk-homologous kinase|Csk-type protein tyrosine kinase|HYL tyrosine kinase|hematopoietic consensus tyrosine-lacking kinase|hydroxyaryl-protein kinase|leukocyte carboxyl-terminal src kinase related|protein kinase HYL|tyrosine kinase MATK|tyrosine-protein kinase CTK|tyrosylprotein kinase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3777969, end_position_on_the_genomic_accession=3801812, orientation='minus', exon_count=17, OMIM=600038, no_hdr0=''),
  4146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4146, CurrentID=0, Status='live', Symbol='MATN1', Aliases='CMP, CRTM', description='matrilin 1, cartilage matrix protein', other_designations='', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30711277, end_position_on_the_genomic_accession=30723585, orientation='minus', exon_count=8, OMIM=115437, no_hdr0=''),
  4147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4147, CurrentID=0, Status='live', Symbol='MATN2', Aliases='', description='matrilin 2', other_designations='', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97869056, end_position_on_the_genomic_accession=98036720, orientation='plus', exon_count=19, OMIM=602108, no_hdr0=''),
  4148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4148, CurrentID=0, Status='live', Symbol='MATN3', Aliases='DIPOA, EDM5, HOA, OADIP, OS2', description='matrilin 3', other_designations='', map_location='2p24-p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19992052, end_position_on_the_genomic_accession=20012694, orientation='minus', exon_count=8, OMIM=602109, no_hdr0=''),
  4149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4149, CurrentID=0, Status='live', Symbol='MAX', Aliases='bHLHd4', description='MYC associated factor X', other_designations='class D basic helix-loop-helix protein 4', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65006101, end_position_on_the_genomic_accession=65102695, orientation='minus', exon_count=9, OMIM=154950, no_hdr0=''),
  4150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4150, CurrentID=0, Status='live', Symbol='MAZ', Aliases='PUR1, Pur-1, SAF-1, SAF-2, SAF-3, ZF87, ZNF801, Zif87', description='MYC-associated zinc finger protein (purine-binding transcription factor)', other_designations='MAZI|purine-binding transcription factor|serum amyloid A activating factor 1|serum amyloid A activating factor 2|serum amyloid A-activating factor-1|transcription factor Zif87|zinc finger protein 801|zinc-finger protein, 87 kilodaltons', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29806083, end_position_on_the_genomic_accession=29811183, orientation='plus', exon_count=8, OMIM=600999, no_hdr0=''),
  4151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4151, CurrentID=0, Status='live', Symbol='MB', Aliases='PVALB, myoglobgin', description='myoglobin', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35606764, end_position_on_the_genomic_accession=35623354, orientation='minus', exon_count=6, OMIM=160000, no_hdr0=''),
  4152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4152, CurrentID=0, Status='live', Symbol='MBD1', Aliases='CXXC3, PCM1, RFT', description='methyl-CpG binding domain protein 1', other_designations='CXXC-type zinc finger protein 3|methyl-CpG binding domain protein 1 isoform PCM1|protein containing methyl-CpG-binding domain 1|the regulator of fibroblast growth factor 2 (FGF-2) transcription', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50266623, end_position_on_the_genomic_accession=50281774, orientation='minus', exon_count=21, OMIM=156535, no_hdr0=''),
  4153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4153, CurrentID=0, Status='live', Symbol='MBL2', Aliases='COLEC1, HSMBPC, MBLD, MBP, MBP-C, MBP1, MBPD, MBL2', description='mannose-binding lectin (protein C) 2, soluble', other_designations='collectin-1|mannan-binding lectin|mannose-binding lectin (protein C) 2, soluble (opsonic defect)|mannose-binding lectin 2, soluble (opsonic defect)', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52765380, end_position_on_the_genomic_accession=52772841, orientation='minus', exon_count=5, OMIM=154545, no_hdr0=''),
  4154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4154, CurrentID=0, Status='live', Symbol='MBNL1', Aliases='EXP, MBNL', description='muscleblind like splicing regulator 1', other_designations='muscleblind-like splicing regulator 1|triplet-expansion RNA-binding protein', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152244321, end_position_on_the_genomic_accession=152465780, orientation='plus', exon_count=20, OMIM=606516, no_hdr0=''),
  4155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4155, CurrentID=0, Status='live', Symbol='MBP', Aliases='', description='myelin basic protein', other_designations='myelin A1 protein|myelin membrane encephalitogenic protein', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76978833, end_position_on_the_genomic_accession=77133708, orientation='minus', exon_count=13, OMIM=159430, no_hdr0=''),
  135228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135228, CurrentID=0, Status='live', Symbol='CD109', Aliases='CPAMD7, p180, r150', description='CD109 molecule', other_designations='150 kDa TGF-beta-1-binding protein|C3 and PZP-like alpha-2-macroglobulin domain-containing protein 7|Gov platelet alloantigens|activated T-cell marker CD109|platelet-specific Gov antigen', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73672182, end_position_on_the_genomic_accession=73828317, orientation='plus', exon_count=36, OMIM=608859, no_hdr0=''),
  4157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4157, CurrentID=0, Status='live', Symbol='MC1R', Aliases='CMM5, MSH-R, SHEP2', description='melanocortin 1 receptor (alpha melanocyte stimulating hormone receptor)', other_designations='MC1-R|melanotropin receptor', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89917879, end_position_on_the_genomic_accession=89920977, orientation='plus', exon_count=1, OMIM=155555, no_hdr0=''),
  4158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4158, CurrentID=0, Status='live', Symbol='MC2R', Aliases='ACTHR', description='melanocortin 2 receptor (adrenocorticotropic hormone)', other_designations='ACTH receptor|MC2 receptor|adrenocorticotropin receptor|corticotropin receptor', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13882044, end_position_on_the_genomic_accession=13915707, orientation='minus', exon_count=3, OMIM=607397, no_hdr0=''),
  4159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4159, CurrentID=0, Status='live', Symbol='MC3R', Aliases='BMIQ9, MC3, MC3-R, OB20, OQTL', description='melanocortin 3 receptor', other_designations='obesity quantitative trait locus', map_location='20q13.2-q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56248732, end_position_on_the_genomic_accession=56249815, orientation='plus', exon_count=1, OMIM=155540, no_hdr0=''),
  4160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4160, CurrentID=0, Status='live', Symbol='MC4R', Aliases='', description='melanocortin 4 receptor', other_designations='MC4-R', map_location='18q22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60371331, end_position_on_the_genomic_accession=60372768, orientation='minus', exon_count=1, OMIM=155541, no_hdr0=''),
  4161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4161, CurrentID=0, Status='live', Symbol='MC5R', Aliases='MC2', description='melanocortin 5 receptor', other_designations='melanocortin-5 receptor', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13825544, end_position_on_the_genomic_accession=13826862, orientation='plus', exon_count=1, OMIM=600042, no_hdr0=''),
  4162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4162, CurrentID=0, Status='live', Symbol='MCAM', Aliases='CD146, MUC18', description='melanoma cell adhesion molecule', other_designations='Gicerin|S-endo 1 endothelial-associated antigen|cell surface glycoprotein P1H12|melanoma adhesion molecule|melanoma-associated antigen A32|melanoma-associated antigen MUC18', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119308524, end_position_on_the_genomic_accession=119317130, orientation='minus', exon_count=16, OMIM=155735, no_hdr0=''),
  4163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4163, CurrentID=0, Status='live', Symbol='MCC', Aliases='MCC1', description='mutated in colorectal cancers', other_designations='', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=113022099, end_position_on_the_genomic_accession=113488830, orientation='minus', exon_count=21, OMIM=159350, no_hdr0=''),
  4166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4166, CurrentID=0, Status='live', Symbol='CHST6', Aliases='MCDC1', description='carbohydrate (N-acetylglucosamine 6-O) sulfotransferase 6', other_designations='C-GlcNAc6ST|GST4-beta|N-acetylglucosamine 6-O-sulfotransferase 5|corneal N-acetylglucosamine 6-sulfotransferase|corneal N-acetylglucosamine-6-O-sulfotransferase|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 4-beta|glcNAc6ST-5|gn6st-5|hCGn6ST', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75473124, end_position_on_the_genomic_accession=75495434, orientation='minus', exon_count=5, OMIM=605294, no_hdr0=''),
  4167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4167, CurrentID=0, Status='live', Symbol='MCDR1', Aliases='CAPED, NCMD', description='macular dystrophy, retinal, 1 (North Carolina type)', other_designations='', map_location='6q14-q16.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=136550, no_hdr0=''),
  4168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4168, CurrentID=0, Status='live', Symbol='MCF2', Aliases='ARHGEF21, DBL', description='MCF.2 cell line derived transforming sequence', other_designations='Oncogene MCF2 (oncogene DBL)|proto-oncogene MCF-2', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139581771, end_position_on_the_genomic_accession=139708243, orientation='minus', exon_count=33, OMIM=311030, no_hdr0=''),
  4170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4170, CurrentID=0, Status='live', Symbol='MCL1', Aliases='BCL2L3, EAT-ES, MCL1L, MCL1S, Mcl-1, TM, bcl2-L-3, mcl1/EAT, MCL1', description='myeloid cell leukemia 1', other_designations='bcl-2-like protein 3|bcl-2-related protein EAT/mcl1|myeloid cell leukemia ES|myeloid cell leukemia sequence 1 (BCL2-related)', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150574551, end_position_on_the_genomic_accession=150579738, orientation='minus', exon_count=4, OMIM=159552, no_hdr0=''),
  4171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4171, CurrentID=0, Status='live', Symbol='MCM2', Aliases='BM28, CCNL1, CDCL1, D3S3194, MITOTIN, cdc19', description='minichromosome maintenance complex component 2', other_designations='MCM2 minichromosome maintenance deficient 2, mitotin|cell devision cycle-like 1|cyclin-like 1|minichromosome maintenance deficient 2 (mitotin)|minichromosome maintenance protein 2 homolog|nuclear protein BM28', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127598357, end_position_on_the_genomic_accession=127622436, orientation='plus', exon_count=16, OMIM=116945, no_hdr0=''),
  4172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4172, CurrentID=0, Status='live', Symbol='MCM3', Aliases='HCC5, P1-MCM3, P1.h, RLFB', description='minichromosome maintenance complex component 3', other_designations='DNA polymerase alpha holoenzyme-associated protein P1|DNA replication factor MCM3|MCM3 minichromosome maintenance deficient 3|RLF subunit beta|cervical cancer proto-oncogene 5|hRlf beta subunit|minichromosome maintenance deficient 3|p102|replication licensing factor, beta subunit', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52264014, end_position_on_the_genomic_accession=52284881, orientation='minus', exon_count=17, OMIM=602693, no_hdr0=''),
  4173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4173, CurrentID=0, Status='live', Symbol='MCM4', Aliases='CDC21, CDC54, NKCD, NKGCD, P1-CDC21, hCdc21', description='minichromosome maintenance complex component 4', other_designations='CDC21 homolog|homolog of S. pombe cell devision cycle 21|minichromosome maintenance deficient 4', map_location='8q11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47960203, end_position_on_the_genomic_accession=47978160, orientation='plus', exon_count=16, OMIM=602638, no_hdr0=''),
  4174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4174, CurrentID=0, Status='live', Symbol='MCM5', Aliases='CDC46, P1-CDC46', description='minichromosome maintenance complex component 5', other_designations='CDC46 homolog|MCM5 minichromosome maintenance deficient 5, cell division cycle 46|minichromosome maintenance deficient 5 (cell division cycle 46)', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35400063, end_position_on_the_genomic_accession=35454889, orientation='plus', exon_count=19, OMIM=602696, no_hdr0=''),
  4175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4175, CurrentID=0, Status='live', Symbol='MCM6', Aliases='MCG40308, Mis5, P105MCM', description='minichromosome maintenance complex component 6', other_designations='MCM6 minichromosome maintenance deficient 6 (MIS5 homolog, S. pombe)|minichromosome maintenance deficient (mis5, S. pombe) 6', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135839626, end_position_on_the_genomic_accession=135876477, orientation='minus', exon_count=17, OMIM=601806, no_hdr0=''),
  4176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4176, CurrentID=0, Status='live', Symbol='MCM7', Aliases='CDC47, MCM2, P1.1-MCM3, P1CDC47, P85MCM, PNAS146, PPP1R104', description='minichromosome maintenance complex component 7', other_designations='CDC47 homolog|homolog of S. cerevisiae Cdc47|minichromosome maintenance deficient 7|protein phosphatase 1, regulatory subunit 104', map_location='7q21.3-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100092728, end_position_on_the_genomic_accession=100101940, orientation='minus', exon_count=15, OMIM=600592, no_hdr0=''),
  101060689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060689, CurrentID=0, Status='live', Symbol='LOC101060689', Aliases='', description='protein FAM115C-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  135250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135250, CurrentID=0, Status='live', Symbol='RAET1E', Aliases='LETAL, N2DL-4, NKG2DL42, RL-4, ULBP4, bA350J20.7, RAET1E', description='retinoic acid early transcript 1E', other_designations='RAE-1-like transcript 4|lymphocyte effector toxicity activation ligand', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149864397, end_position_on_the_genomic_accession=149898321, orientation='minus', exon_count=9, OMIM=609243, no_hdr0=''),
  4179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4179, CurrentID=0, Status='live', Symbol='CD46', Aliases='AHUS2, MCP, MIC10, TLX, TRA2.10', description='CD46 molecule, complement regulatory protein', other_designations='CD46 antigen, complement regulatory protein|antigen identified by monoclonal antibody TRA-2-10|complement membrane cofactor protein|measles virus receptor|membrane cofactor protein (CD46, trophoblast-lymphocyte cross-reactive antigen)|trophoblast leucocyte common antigen|trophoblast leukocyte common antigen|trophoblast-lymphocyte cross-reactive antigen', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207752038, end_position_on_the_genomic_accession=207795516, orientation='plus', exon_count=14, OMIM=120920, no_hdr0=''),
  4182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4182, CurrentID=0, Status='live', Symbol='CD46P1', Aliases='CD46P, MCPL, MCPL1', description='CD46 molecule, complement regulatory protein pseudogene 1', other_designations='membrane cofactor protein-like (CD46-like, trophoblast-lymphocyte cross-reactive antigen-like)', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207645084, end_position_on_the_genomic_accession=207657319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4183, CurrentID=0, Status='live', Symbol='MCS', Aliases='MRXS4', description='Miles-Carpenter X-linked mental retardation syndrome', other_designations='Miles-Carpenter syndrome (mental retardation, X-linked, syndromic-4, with congenital contractures and low fingertip arches)', map_location='Xq13-q22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309605, no_hdr0=''),
  4184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4184, CurrentID=0, Status='live', Symbol='SMCP', Aliases='HSMCSGEN1, MCS, MCSP', description='sperm mitochondria-associated cysteine-rich protein', other_designations='mitochondrial capsule selenoprotein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152878322, end_position_on_the_genomic_accession=152885047, orientation='plus', exon_count=2, OMIM=601148, no_hdr0=''),
  4185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4185, CurrentID=0, Status='live', Symbol='ADAM11', Aliases='MDC', description='ADAM metallopeptidase domain 11', other_designations='ADAM 11', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44759037, end_position_on_the_genomic_accession=44781846, orientation='plus', exon_count=27, OMIM=155120, no_hdr0=''),
  4188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4188, CurrentID=0, Status='live', Symbol='MDFI', Aliases='I-MF, I-mfa', description='MyoD family inhibitor', other_designations='inhibitor of MyoD family a|myogenic repressor I-mf', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41637037, end_position_on_the_genomic_accession=41654246, orientation='plus', exon_count=8, OMIM=604971, no_hdr0=''),
  4189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4189, CurrentID=0, Status='live', Symbol='DNAJB9', Aliases='ERdj4, MDG-1, MDG1, MST049, MSTP049', description='DnaJ heat shock protein family (Hsp40) member B9', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 9|ER-resident protein ERdj4|endoplasmic reticulum DNA J domain-containing protein 4|endoplasmic reticulum DnaJ homolog 4|microvascular endothelial differentiation gene 1 protein', map_location='14q24.2-q24.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108569745, end_position_on_the_genomic_accession=108574850, orientation='plus', exon_count=3, OMIM=602634, no_hdr0=''),
  4190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4190, CurrentID=0, Status='live', Symbol='MDH1', Aliases='HEL-S-32, MDH-s, MDHA, MGC:1375, MOR2', description='malate dehydrogenase 1', other_designations='cytosolic malate dehydrogenase|diiodophenylpyruvate reductase|epididymis secretory protein Li 32|malate dehydrogenase 1, NAD (soluble)|soluble malate dehydrogenase', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63588609, end_position_on_the_genomic_accession=63607197, orientation='plus', exon_count=10, OMIM=154200, no_hdr0=''),
  4191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4191, CurrentID=0, Status='live', Symbol='MDH2', Aliases='M-MDH, MDH, MGC:3559, MOR1', description='malate dehydrogenase 2', other_designations='malate dehydrogenase 2, NAD (mitochondrial)', map_location='7cen-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76048019, end_position_on_the_genomic_accession=76067508, orientation='plus', exon_count=10, OMIM=154100, no_hdr0=''),
  4192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4192, CurrentID=0, Status='live', Symbol='MDK', Aliases='ARAP, MK, NEGF2', description='midkine (neurite growth-promoting factor 2)', other_designations='amphiregulin-associated protein|midgestation and kidney protein|neurite outgrowth-promoting factor 2|retinoic acid inducible factor', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46380784, end_position_on_the_genomic_accession=46383837, orientation='plus', exon_count=6, OMIM=162096, no_hdr0=''),
  4193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4193, CurrentID=0, Status='live', Symbol='MDM2', Aliases='ACTFS, HDMX, hdm2', description='MDM2 proto-oncogene, E3 ubiquitin protein ligase', other_designations='MDM2 oncogene, E3 ubiquitin protein ligase|Mdm2, p53 E3 ubiquitin protein ligase homolog|Mdm2, transformed 3T3 cell double minute 2, p53 binding protein|double minute 2, human homolog of; p53-binding protein|oncoprotein Mdm2', map_location='12q14.3-q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68808168, end_position_on_the_genomic_accession=68845544, orientation='plus', exon_count=13, OMIM=164785, no_hdr0=''),
  4194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4194, CurrentID=0, Status='live', Symbol='MDM4', Aliases='HDMX, MDMX, MRP1', description='MDM4, p53 regulator', other_designations='MDM4 protein variant G|MDM4 protein variant Y|MDM4-related protein 1|Mdm4 p53 binding protein homolog|double minute 4, human homolog of; p53-binding protein|mdm2-like p53-binding protein|protein Mdmx', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204516379, end_position_on_the_genomic_accession=204581681, orientation='plus', exon_count=17, OMIM=602704, no_hdr0=''),
  4195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4195, CurrentID=0, Status='live', Symbol='MDRV', Aliases='NIPA2', description='muscular dystrophy, with rimmed vacuoles', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601846, no_hdr0=''),
  135270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135270, CurrentID=0, Status='live', Symbol='COPS5P1', Aliases='COPS5P', description='COP9 signalosome subunit 5 pseudogene 1', other_designations='COP9 constitutive photomorphogenic homolog subunit 5 pseudogene', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93091827, end_position_on_the_genomic_accession=93092986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4199, CurrentID=0, Status='live', Symbol='ME1', Aliases='HUMNDME, MES', description='malic enzyme 1, NADP(+)-dependent, cytosolic', other_designations='Malic enzyme, cytoplasmic|NADP-ME|malate dehydrogenase|malic enzyme 1, soluble|pyruvic-malic carboxylase', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83210389, end_position_on_the_genomic_accession=83431741, orientation='minus', exon_count=14, OMIM=154250, no_hdr0=''),
  4200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4200, CurrentID=0, Status='live', Symbol='ME2', Aliases='ODS1', description='malic enzyme 2, NAD(+)-dependent, mitochondrial', other_designations='NAD-ME|malate dehydrogenase|pyruvic-malic carboxylase', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50878604, end_position_on_the_genomic_accession=50949795, orientation='plus', exon_count=16, OMIM=154270, no_hdr0=''),
  4201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4201, CurrentID=0, Status='live', Symbol='MEA1', Aliases='HYS, MEA', description='male-enhanced antigen 1', other_designations='MEA-1|Male-enhanced antigen (H-Y structural gene)', map_location='6p21.3-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42980344, end_position_on_the_genomic_accession=43017087, orientation='minus', exon_count=8, OMIM=143170, no_hdr0=''),
  4204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4204, CurrentID=0, Status='live', Symbol='MECP2', Aliases='AUTSX3, MRX16, MRX79, MRXS13, MRXSL, PPMX, RS, RTS, RTT', description='methyl-CpG binding protein 2', other_designations='meCp-2 protein|methyl CpG binding protein 2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154021813, end_position_on_the_genomic_accession=154097731, orientation='minus', exon_count=10, OMIM=300005, no_hdr0=''),
  4205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4205, CurrentID=0, Status='live', Symbol='MEF2A', Aliases='ADCAD1, RSRFC4, RSRFC9, mef2', description='myocyte enhancer factor 2A', other_designations='MADS box transcription enhancer factor 2, polypeptide A (myocyte enhancer factor 2A)|serum response factor-like protein 1', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99565533, end_position_on_the_genomic_accession=99716483, orientation='plus', exon_count=24, OMIM=600660, no_hdr0=''),
  4206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4206, CurrentID=0, Status='live', Symbol='MEF2AP1', Aliases='MEF2AP', description='myocyte enhancer factor 2A pseudogene 1', other_designations='MADS box transcription enhancer factor 2, polypeptide A pseudogene', map_location='1q24-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179447588, end_position_on_the_genomic_accession=179448440, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4207, CurrentID=0, Status='live', Symbol='MEF2BNB-MEF2B', Aliases='LOC729991-MEF2B, MEF2B', description='MEF2BNB-MEF2B readthrough', other_designations='LOC729991-MEF2B readthrough', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19145567, end_position_on_the_genomic_accession=19192591, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  4208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4208, CurrentID=0, Status='live', Symbol='MEF2C', Aliases='C5DELq14.3, DEL5q14.3', description='myocyte enhancer factor 2C', other_designations='MADS box transcription enhancer factor 2, polypeptide C', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88718241, end_position_on_the_genomic_accession=88904105, orientation='minus', exon_count=16, OMIM=600662, no_hdr0=''),
  4209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4209, CurrentID=0, Status='live', Symbol='MEF2D', Aliases='', description='myocyte enhancer factor 2D', other_designations='MADS box transcription enhancer factor 2, polypeptide D', map_location='1q12-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156463721, end_position_on_the_genomic_accession=156500842, orientation='minus', exon_count=17, OMIM=600663, no_hdr0=''),
  4210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4210, CurrentID=0, Status='live', Symbol='MEFV', Aliases='FMF, MEF, TRIM20', description='Mediterranean fever', other_designations='marenostrin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3242028, end_position_on_the_genomic_accession=3256627, orientation='minus', exon_count=10, OMIM=608107, no_hdr0=''),
  4211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4211, CurrentID=0, Status='live', Symbol='MEIS1', Aliases='', description='Meis homeobox 1', other_designations='Meis1, myeloid ecotropic viral integration site 1 homolog|WUGSC:H_NH0444B04.1|leukemogenic homolog protein', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66435125, end_position_on_the_genomic_accession=66572759, orientation='plus', exon_count=15, OMIM=601739, no_hdr0=''),
  4212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4212, CurrentID=0, Status='live', Symbol='MEIS2', Aliases='HsT18361, MRG1', description='Meis homeobox 2', other_designations='Meis homolog 2|Meis1, myeloid ecotropic viral integration site 1 homolog 2|Meis1-related gene 1|TALE homeobox protein Meis2|meis1-related protein 1', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36891021, end_position_on_the_genomic_accession=37101334, orientation='minus', exon_count=16, OMIM=601740, no_hdr0=''),
  4213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4213, CurrentID=0, Status='live', Symbol='MEIS3P1', Aliases='MEIS3, MEIS4, MRG2', description='Meis homeobox 3 pseudogene 1', other_designations='CTD-3157E16.2|Meis homolog 3|Meis1 homolog 3 pseudogene 1|Meis1, myeloid ecotropic viral integration site 1 homolog 3|Meis1, myeloid ecotropic viral integration site 1 homolog 4|Putative homeobox protein Meis3-like 1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15786850, end_position_on_the_genomic_accession=15789705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4214, CurrentID=0, Status='live', Symbol='MAP3K1', Aliases='MAPKKK1, MEKK, MEKK 1, MEKK1, SRXY6', description='mitogen-activated protein kinase kinase kinase 1, E3 ubiquitin protein ligase', other_designations='MAP/ERK kinase kinase 1|MAPK/ERK kinase kinase 1|MEK kinase 1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56815073, end_position_on_the_genomic_accession=56896152, orientation='plus', exon_count=23, OMIM=600982, no_hdr0=''),
  4215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4215, CurrentID=0, Status='live', Symbol='MAP3K3', Aliases='MAPKKK3, MEKK3', description='mitogen-activated protein kinase kinase kinase 3', other_designations='MAP/ERK kinase kinase 3|MAPK/ERK kinase kinase 3|MEK kinase 3|MEKK 3', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63622209, end_position_on_the_genomic_accession=63696310, orientation='plus', exon_count=17, OMIM=602539, no_hdr0=''),
  4216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4216, CurrentID=0, Status='live', Symbol='MAP3K4', Aliases='MAPKKK4, MEKK 4, MEKK4, MTK1, PRO0412', description='mitogen-activated protein kinase kinase kinase 4', other_designations='MAP three kinase 1|MAP/ERK kinase kinase 4|MAPK/ERK kinase kinase 4|MEK kinase 4|SSK2/SSK22 MAP kinase kinase kinase, yeast, homolog of|dJ473J16.1 (mitogen-activated protein kinase kinase kinase 4)|predicted protein of HQ0412', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160991727, end_position_on_the_genomic_accession=161117385, orientation='plus', exon_count=29, OMIM=602425, no_hdr0=''),
  4217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4217, CurrentID=0, Status='live', Symbol='MAP3K5', Aliases='ASK1, MAPKKK5, MEKK5', description='mitogen-activated protein kinase kinase kinase 5', other_designations='ASK-1|MAP/ERK kinase kinase 5|MAPK/ERK kinase kinase 5|MEK kinase 5|MEKK 5|apoptosis signal regulating kinase 1|apoptosis signal-regulating kinase 1', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136557046, end_position_on_the_genomic_accession=136793051, orientation='minus', exon_count=31, OMIM=602448, no_hdr0=''),
  4218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4218, CurrentID=0, Status='live', Symbol='RAB8A', Aliases='MEL, RAB8', description='RAB8A, member RAS oncogene family', other_designations='mel transforming oncogene (RAB8 homolog)|mel transforming oncogene (derived from cell line NK14)|mel transforming oncogene (derived from cell line NK14)- RAB8 homolog|oncogene c-mel|ras-associated protein RAB8', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16111680, end_position_on_the_genomic_accession=16133635, orientation='plus', exon_count=9, OMIM=165040, no_hdr0=''),
  4221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4221, CurrentID=0, Status='live', Symbol='MEN1', Aliases='MEAI, SCG2', description='menin 1', other_designations='multiple endocrine neoplasia I', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64803514, end_position_on_the_genomic_accession=64811294, orientation='minus', exon_count=13, OMIM=613733, no_hdr0=''),
  4222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4222, CurrentID=0, Status='live', Symbol='MEOX1', Aliases='KFS2, MOX1', description='mesenchyme homeobox 1', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43640389, end_position_on_the_genomic_accession=43661954, orientation='minus', exon_count=4, OMIM=600147, no_hdr0=''),
  4223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4223, CurrentID=0, Status='live', Symbol='MEOX2', Aliases='GAX, MOX2', description='mesenchyme homeobox 2', other_designations='growth arrest-specific homeobox', map_location='7p22.1-p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15611212, end_position_on_the_genomic_accession=15686683, orientation='minus', exon_count=3, OMIM=600535, no_hdr0=''),
  4224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4224, CurrentID=0, Status='live', Symbol='MEP1A', Aliases='PPHA', description='meprin A subunit alpha', other_designations='N-benzoyl-L-tyrosyl-P-amino-benzoic acid hydrolase subunit alpha|PABA peptide hydrolase|PPH alpha|bA268F1.1 (meprin A alpha (PABA peptide hydrolase))|endopeptidase-2|meprin A, alpha (PABA peptide hydrolase)', map_location='6p12-p11', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46793357, end_position_on_the_genomic_accession=46845987, orientation='plus', exon_count=15, OMIM=600388, no_hdr0=''),
  4225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4225, CurrentID=0, Status='live', Symbol='MEP1B', Aliases='', description='meprin A subunit beta', other_designations='N-benzoyl-L-tyrosyl-P-amino-benzoic acid hydrolase subunit beta|N-benzoyl-L-tyrosyl-p-amino-benzoic acid hydrolase beta subunit|PABA peptide hydrolase|PPH beta|endopeptidase-2|meprin A, beta|meprin B', map_location='18q12.2-q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32190024, end_position_on_the_genomic_accession=32220404, orientation='plus', exon_count=15, OMIM=600389, no_hdr0=''),
  4228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4228, CurrentID=0, Status='live', Symbol='MER5', Aliases='', description='antigen identified by monoclonal antibody 2D8', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4232, CurrentID=0, Status='live', Symbol='MEST', Aliases='PEG1', description='mesoderm specific transcript', other_designations='paternally-expressed gene 1 protein', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130486175, end_position_on_the_genomic_accession=130506297, orientation='plus', exon_count=14, OMIM=601029, no_hdr0=''),
  4233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4233, CurrentID=0, Status='live', Symbol='MET', Aliases='AUTS9, DFNB97, HGFR, RCCP2, c-Met', description='MET proto-oncogene, receptor tyrosine kinase', other_designations='HGF receptor|HGF/SF receptor|SF receptor|met proto-oncogene tyrosine kinase|proto-oncogene c-Met|scatter factor receptor|soluble MET variant 12|tyrosine-protein kinase Met', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116672405, end_position_on_the_genomic_accession=116798386, orientation='plus', exon_count=24, OMIM=164860, no_hdr0=''),
  4234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4234, CurrentID=0, Status='live', Symbol='METTL1', Aliases='C12orf1, TRM8, TRMT8, YDL201w', description='methyltransferase like 1', other_designations='D1075-like gene product|methyltransferase-like 1|methyltransferase-like protein 1|tRNA (guanine(46)-N(7))-methyltransferase|tRNA(m7G46)-methyltransferase', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57768471, end_position_on_the_genomic_accession=57772131, orientation='minus', exon_count=7, OMIM=604466, no_hdr0=''),
  4236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4236, CurrentID=0, Status='live', Symbol='MFAP1', Aliases='', description='microfibrillar associated protein 1', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43804535, end_position_on_the_genomic_accession=43824753, orientation='minus', exon_count=9, OMIM=600215, no_hdr0=''),
  4237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4237, CurrentID=0, Status='live', Symbol='MFAP2', Aliases='MAGP, MAGP-1, MAGP1', description='microfibrillar associated protein 2', other_designations='microfibril-associated glycoprotein 1', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16974500, end_position_on_the_genomic_accession=16981643, orientation='minus', exon_count=10, OMIM=156790, no_hdr0=''),
  4238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4238, CurrentID=0, Status='live', Symbol='MFAP3', Aliases='', description='microfibrillar associated protein 3', other_designations='microfibrillar-associated protein 3', map_location='5q32-q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154038959, end_position_on_the_genomic_accession=154057454, orientation='plus', exon_count=3, OMIM=600491, no_hdr0=''),
  4239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4239, CurrentID=0, Status='live', Symbol='MFAP4', Aliases='', description='microfibrillar associated protein 4', other_designations='microfibrillar-associated protein 4', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19383442, end_position_on_the_genomic_accession=19387219, orientation='minus', exon_count=6, OMIM=600596, no_hdr0=''),
  4240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4240, CurrentID=0, Status='live', Symbol='MFGE8', Aliases='BA46, EDIL1, HMFG, HsT19888, MFG-E8, MFGM, OAcGD3S, SED1, SPAG10, hP47', description='milk fat globule-EGF factor 8 protein', other_designations='O-acetyl disialoganglioside synthase|breast epithelial antigen BA46|medin|sperm associated antigen 10|sperm surface protein hP47', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88898683, end_position_on_the_genomic_accession=88913454, orientation='minus', exon_count=10, OMIM=602281, no_hdr0=''),
  4241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4241, CurrentID=0, Status='live', Symbol='MFI2', Aliases='CD228, MAP97, MTF1, MTf', description='antigen p97 (melanoma associated) identified by monoclonal antibodies 133.2 and 96.5', other_designations='melanoma-associated antigen p97|membrane-bound transferrin-like protein', map_location='3q28-q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196987559, end_position_on_the_genomic_accession=197029816, orientation='minus', exon_count=19, OMIM=155750, no_hdr0=''),
  4242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4242, CurrentID=0, Status='live', Symbol='MFNG', Aliases='', description='MFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37469063, end_position_on_the_genomic_accession=37486440, orientation='minus', exon_count=8, OMIM=602577, no_hdr0=''),
  401677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401677, CurrentID=0, Status='live', Symbol='LOC401677', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29275655, end_position_on_the_genomic_accession=29276585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4245, CurrentID=0, Status='live', Symbol='MGAT1', Aliases='GLCNAC-TI, GLCT1, GLYT1, GNT-1, GNT-I, MGAT', description='mannosyl (alpha-1,3-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase', other_designations='N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase I|glcNAc-T I', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180790541, end_position_on_the_genomic_accession=180815652, orientation='minus', exon_count=9, OMIM=160995, no_hdr0=''),
  4246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4246, CurrentID=0, Status='live', Symbol='SCGB2A1', Aliases='LPHC, LPNC, MGB2, UGB3', description='secretoglobin, family 2A, member 1', other_designations='lacryglobin|lipophilin C|lipophilin-C|mammaglobin 2|mammaglobin B|mammaglobin-2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62208668, end_position_on_the_genomic_accession=62213939, orientation='plus', exon_count=3, OMIM=604398, no_hdr0=''),
  4247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4247, CurrentID=0, Status='live', Symbol='MGAT2', Aliases='CDG2A, CDGS2, GLCNACTII, GNT-II, GNT2', description='mannosyl (alpha-1,6-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase', other_designations='Beta-1,2-N-acetylglucosaminyltransferase II|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase II|UDP-N-acetylglucosamine:alpha-6-D-mannoside beta-1,2-N-acetylglucosaminyltransferase II|glcNAc-T II|mannoside acetylglucosaminyltransferase 2', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49620771, end_position_on_the_genomic_accession=49623481, orientation='plus', exon_count=1, OMIM=602616, no_hdr0=''),
  4248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4248, CurrentID=0, Status='live', Symbol='MGAT3', Aliases='GNT-III, GNT3', description='mannosyl (beta-1,4-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase', other_designations='GlcNAc-T III|N-acetylglucosaminyltransferase III|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase III', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39447182, end_position_on_the_genomic_accession=39492194, orientation='plus', exon_count=13, OMIM=604621, no_hdr0=''),
  4249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4249, CurrentID=0, Status='live', Symbol='MGAT5', Aliases='GNT-V, GNT-VA', description='mannosyl (alpha-1,6-)-glycoprotein beta-1,6-N-acetyl-glucosaminyltransferase', other_designations='N-acetylglucosaminyl-transferase V|alpha-mannoside beta-1,6-N-acetylglucosaminyltransferase|glcNAc-T V|mannoside acetylglucosaminyltransferase 5', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134119930, end_position_on_the_genomic_accession=134454621, orientation='plus', exon_count=24, OMIM=601774, no_hdr0=''),
  4250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4250, CurrentID=0, Status='live', Symbol='SCGB2A2', Aliases='MGB1, UGB2', description='secretoglobin, family 2A, member 2', other_designations='mammaglobin 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62269933, end_position_on_the_genomic_accession=62273160, orientation='plus', exon_count=3, OMIM=605562, no_hdr0=''),
  4253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4253, CurrentID=0, Status='live', Symbol='CTAGE5', Aliases='MEA6, MGEA, MGEA11, MGEA6', description='CTAGE family member 5', other_designations='CTAGE family, member 5|cutaneous T-cell lymphoma-associated antigen 5|meningioma expressed antigen 6 (coiled-coil proline-rich)|meningioma-expressed antigen 6/11|protein cTAGE-5', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39265272, end_position_on_the_genomic_accession=39351193, orientation='plus', exon_count=31, OMIM=602132, no_hdr0=''),
  4254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4254, CurrentID=0, Status='live', Symbol='KITLG', Aliases='DCUA, FPH2, FPHH, KL-1, Kitl, MGF, SCF, SF, SHEP7', description='KIT ligand', other_designations='c-Kit ligand|familial progressive hyperpigmentation 2|mast cell growth factor|steel factor|stem cell factor', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88492793, end_position_on_the_genomic_accession=88580473, orientation='minus', exon_count=10, OMIM=184745, no_hdr0=''),
  4255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4255, CurrentID=0, Status='live', Symbol='MGMT', Aliases='', description='O-6-methylguanine-DNA methyltransferase', other_designations='6-O-methylguanine-DNA methyltransferase|O-6-methylguanine-DNA-alkyltransferase|O6-methylguanine-DNA methyltransferase|methylated DNA protein cysteine methyltransferase|methylguanine-DNA methyltransferase', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129467190, end_position_on_the_genomic_accession=129768042, orientation='plus', exon_count=6, OMIM=156569, no_hdr0=''),
  4256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4256, CurrentID=0, Status='live', Symbol='MGP', Aliases='GIG36, MGLAP, NTI', description='matrix Gla protein', other_designations='cell growth-inhibiting gene 36 protein', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14881181, end_position_on_the_genomic_accession=14885919, orientation='minus', exon_count=5, OMIM=154870, no_hdr0=''),
  4257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4257, CurrentID=0, Status='live', Symbol='MGST1', Aliases='GST12, MGST, MGST-I', description='microsomal glutathione S-transferase 1', other_designations='glutathione S-transferase 12|microsomal GST-1|microsomal GST-I', map_location='12p12.3-p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16347142, end_position_on_the_genomic_accession=16377189, orientation='plus', exon_count=11, OMIM=138330, no_hdr0=''),
  4258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4258, CurrentID=0, Status='live', Symbol='MGST2', Aliases='GST2, MGST-II', description='microsomal glutathione S-transferase 2', other_designations='microsomal GST-2|microsomal GST-II', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139665768, end_position_on_the_genomic_accession=139740745, orientation='plus', exon_count=7, OMIM=601733, no_hdr0=''),
  4259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4259, CurrentID=0, Status='live', Symbol='MGST3', Aliases='GST-III', description='microsomal glutathione S-transferase 3', other_designations='microsomal GST-3|microsomal GST-III|microsomal glutathione S-transferase III', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165630873, end_position_on_the_genomic_accession=165656136, orientation='plus', exon_count=7, OMIM=604564, no_hdr0=''),
  4261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4261, CurrentID=0, Status='live', Symbol='CIITA', Aliases='C2TAIV, MHC2TA, NLRA, CIITA', description='class II, major histocompatibility complex, transactivator', other_designations='MHC class II transactivator type I|MHC class II transactivator type III|NLR family, acid domain containing|nucleotide-binding oligomerization domain, leucine rich repeat and acid domain containing', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10866212, end_position_on_the_genomic_accession=10941449, orientation='plus', exon_count=29, OMIM=600005, no_hdr0=''),
  4264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4264, CurrentID=0, Status='live', Symbol='MHS2', Aliases='', description='malignant hyperthermia susceptibility 2', other_designations='', map_location='17q11.2-q24', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=154275, no_hdr0=''),
  4265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4265, CurrentID=0, Status='live', Symbol='MHS4', Aliases='', description='malignant hyperthermia susceptibility 4', other_designations='', map_location='3q13.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600467, no_hdr0=''),
  4266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4266, CurrentID=0, Status='live', Symbol='MHS6', Aliases='', description='malignant hyperthermia susceptibility 6', other_designations='', map_location='5p', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601888, no_hdr0=''),
  4267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4267, CurrentID=0, Status='live', Symbol='CD99', Aliases='HBA71, MIC2, MIC2X, MIC2Y, MSK5X', description='CD99 molecule', other_designations='E2 antigen|MIC2 (monoclonal antibody 12E7)|T-cell surface glycoprotein E2|antigen identified by monoclonal 12E7, Y homolog|antigen identified by monoclonal antibodies 12E7, F21 and O13|surface antigen MIC2', map_location='Xp22.32 and Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2691187, end_position_on_the_genomic_accession=2741309, orientation='plus', exon_count=12, OMIM='313470,450000', no_hdr0=''),
  4270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4270, CurrentID=0, Status='live', Symbol='MIC7', Aliases='', description='antigen identified by monoclonal antibody 28.3.7', other_designations='Attached cell antigen 28.3.7', map_location='15q11-q22', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=108990, no_hdr0=''),
  4273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4273, CurrentID=0, Status='live', Symbol='MIC12', Aliases='', description='antigen identified by monoclonal antibody 30.2A8', other_designations='leukocyte antigen MIC12', map_location='15q11-q22', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=107254, no_hdr0=''),
  285591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285591, CurrentID=0, Status='live', Symbol='LOC285591', Aliases='', description='pyrophosphatase (inorganic) 1 pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172975446, end_position_on_the_genomic_accession=172976693, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4277, CurrentID=0, Status='live', Symbol='MICB', Aliases='PERB11.2', description='MHC class I polypeptide-related sequence B', other_designations='MHC class I chain-related protein B|MHC class I mic-B antigen|MHC class I-like molecule PERB11.2-IMX|stress inducible class I homolog', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31494277, end_position_on_the_genomic_accession=31511124, orientation='plus', exon_count=8, OMIM=602436, no_hdr0=''),
  4279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4279, CurrentID=0, Status='live', Symbol='MICD', Aliases='PERB11.4', description='MHC class I polypeptide-related sequence D (pseudogene)', other_designations='MHC class I polypeptide-related sequence B pseudogene|MHC class I polypeptide-related sequence D, pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29970372, end_position_on_the_genomic_accession=29975745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4280, CurrentID=0, Status='live', Symbol='MICE', Aliases='PERB11.5, dJ377H14.7', description='MHC class I polypeptide-related sequence E (pseudogene)', other_designations='MHC class I polypeptide-related sequence E, pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29741457, end_position_on_the_genomic_accession=29749103, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4281, CurrentID=0, Status='live', Symbol='MID1', Aliases='BBBG1, FXY, GBBB1, MIDIN, OGS1, OS, OSX, RNF59, TRIM18, XPRF, ZNFXY', description='midline 1', other_designations='Opitz/BBB syndrome|RING finger protein 59|RING finger protein Midline-1|midline 1 RING finger protein|putative transcription factor XPRF|tripartite motif protein TRIM18|tripartite motif-containing protein 18|zinc finger on X and Y, mouse, homolog of', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=10445310, end_position_on_the_genomic_accession=11111177, orientation='minus', exon_count=19, OMIM=300552, no_hdr0=''),
  4282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4282, CurrentID=0, Status='live', Symbol='MIF', Aliases='GIF, GLIF, MMIF', description='macrophage migration inhibitory factor (glycosylation-inhibiting factor)', other_designations='L-dopachrome isomerase|L-dopachrome tautomerase|phenylpyruvate tautomerase', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23894378, end_position_on_the_genomic_accession=23895222, orientation='plus', exon_count=3, OMIM=153620, no_hdr0=''),
  4283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4283, CurrentID=0, Status='live', Symbol='CXCL9', Aliases='CMK, Humig, MIG, SCYB9, crg-10', description='chemokine (C-X-C motif) ligand 9', other_designations='gamma-interferon-induced monokine|monokine induced by gamma interferon|monokine induced by interferon-gamma|small-inducible cytokine B9', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76001342, end_position_on_the_genomic_accession=76007523, orientation='minus', exon_count=4, OMIM=601704, no_hdr0=''),
  4284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4284, CurrentID=0, Status='live', Symbol='MIP', Aliases='AQP0, CTRCT15, LIM126, MP26, MIP', description='major intrinsic protein of lens fiber', other_designations='aquaporin 0', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56449502, end_position_on_the_genomic_accession=56456553, orientation='minus', exon_count=7, OMIM=154050, no_hdr0=''),
  4285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4285, CurrentID=0, Status='live', Symbol='MIPEP', Aliases='HMIP, MIP', description='mitochondrial intermediate peptidase', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23730188, end_position_on_the_genomic_accession=23889448, orientation='minus', exon_count=21, OMIM=602241, no_hdr0=''),
  4286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4286, CurrentID=0, Status='live', Symbol='MITF', Aliases='CMM8, MI, WS2, WS2A, bHLHe32', description='microphthalmia-associated transcription factor', other_designations='class E basic helix-loop-helix protein 32|microphtalmia-associated transcription factor', map_location='3p14.2-p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69739435, end_position_on_the_genomic_accession=69968337, orientation='plus', exon_count=16, OMIM=156845, no_hdr0=''),
  4287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4287, CurrentID=0, Status='live', Symbol='ATXN3', Aliases='AT3, ATX3, JOS, MJD, MJD1, SCA3', description='ataxin 3', other_designations='Machado-Joseph disease (spinocerebellar ataxia 3, olivopontocerebellar ataxia 3, autosomal dominant, ataxin 3)|Machado-Joseph disease protein 1|ataxin 3 variant an|ataxin 3 variant ao|ataxin 3 variant at|ataxin 3 variant e|ataxin 3 variant h|ataxin 3 variant m|ataxin 3 variant r|ataxin 3 variant ref|ataxin 3 variant y|josephin|olivopontocerebellar ataxia 3|spinocerebellar ataxia type 3 protein', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92058552, end_position_on_the_genomic_accession=92106621, orientation='minus', exon_count=13, OMIM=607047, no_hdr0=''),
  4288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4288, CurrentID=0, Status='live', Symbol='MKI67', Aliases='KIA, MIB-, MIB-1, PPP1R105', description='marker of proliferation Ki-67', other_designations='antigen identified by monoclonal antibody Ki-67|proliferation-related Ki-67 antigen|protein phosphatase 1, regulatory subunit 105', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128096659, end_position_on_the_genomic_accession=128126363, orientation='minus', exon_count=16, OMIM=176741, no_hdr0=''),
  4289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4289, CurrentID=0, Status='live', Symbol='MKLN1', Aliases='TWA2', description='muskelin 1', other_designations='muskelin 1, intracellular mediator containing kelch motifs', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131110096, end_position_on_the_genomic_accession=131496639, orientation='plus', exon_count=20, OMIM=605623, no_hdr0=''),
  4291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4291, CurrentID=0, Status='live', Symbol='MLF1', Aliases='', description='myeloid leukemia factor 1', other_designations='myelodysplasia-myeloid leukemia factor 1', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158571163, end_position_on_the_genomic_accession=158606460, orientation='plus', exon_count=10, OMIM=601402, no_hdr0=''),
  4292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4292, CurrentID=0, Status='live', Symbol='MLH1', Aliases='COCA2, FCC2, HNPCC, HNPCC2, hMLH1', description='mutL homolog 1', other_designations='mutL homolog 1, colon cancer, nonpolyposis type 2', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36993350, end_position_on_the_genomic_accession=37050846, orientation='plus', exon_count=21, OMIM=120436, no_hdr0=''),
  4293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4293, CurrentID=0, Status='live', Symbol='MAP3K9', Aliases='MEKK9, MLK1, PRKE1', description='mitogen-activated protein kinase kinase kinase 9', other_designations='mixed lineage kinase 1 (tyr and ser/thr specificity)', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70722526, end_position_on_the_genomic_accession=70809269, orientation='minus', exon_count=17, OMIM=600136, no_hdr0=''),
  4294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4294, CurrentID=0, Status='live', Symbol='MAP3K10', Aliases='MEKK10, MLK2, MST', description='mitogen-activated protein kinase kinase kinase 10', other_designations='MKN28 derived nonreceptor_type serine/threonine kinase|MKN28 kinase|mixed lineage kinase 2|protein kinase MST', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40191650, end_position_on_the_genomic_accession=40215580, orientation='plus', exon_count=12, OMIM=600137, no_hdr0=''),
  4295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4295, CurrentID=0, Status='live', Symbol='MLN', Aliases='', description='motilin', other_designations='prepromotilin', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33794672, end_position_on_the_genomic_accession=33804016, orientation='minus', exon_count=5, OMIM=158270, no_hdr0=''),
  4296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4296, CurrentID=0, Status='live', Symbol='MAP3K11', Aliases='MEKK11, MLK-3, MLK3, PTK1, SPRK', description='mitogen-activated protein kinase kinase kinase 11', other_designations='SH3 domain-containing proline-rich kinase|mixed lineage kinase 3|protein-tyrosine kinase PTK1|src-homology 3 domain-containing proline-rich kinase', map_location='11q13.1-q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65597755, end_position_on_the_genomic_accession=65614338, orientation='minus', exon_count=11, OMIM=600050, no_hdr0=''),
  4297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4297, CurrentID=0, Status='live', Symbol='KMT2A', Aliases='ALL-1, CXXC7, HRX, HTRX1, MLL, MLL-AF9, MLL/GAS7, MLL1, MLL1A, TET1-MLL, TRX1, WDSTS', description='lysine (K)-specific methyltransferase 2A', other_designations='CDK6/MLL fusion protein|CXXC-type zinc finger protein 7|MLL-AF4 der(11) fusion protein|MLL/ENL fusion protein|MLL/GAS7 fusion protein|MLL/GMPS fusion protein|MLL/hCDCrel fusion protein|lysine N-methyltransferase 2A|mixed lineage leukemia 1|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila)|rearranged MLL protein|trithorax-like protein|zinc finger protein HRX', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118436490, end_position_on_the_genomic_accession=118526824, orientation='plus', exon_count=37, OMIM=159555, no_hdr0=''),
  4298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4298, CurrentID=0, Status='live', Symbol='MLLT1', Aliases='ENL, LTG19, YEATS1', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 1', other_designations='CTC-503J8.6|ENL/MLL fusion|MLL/ENL fusion protein|MLLT1/MLL fusion|YEATS domain-containing protein 1|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog); translocated to, 1|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6210381, end_position_on_the_genomic_accession=6279948, orientation='minus', exon_count=14, OMIM=159556, no_hdr0=''),
  4299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4299, CurrentID=0, Status='live', Symbol='AFF1', Aliases='AF4, MLLT2, PBM1', description='AF4/FMR2 family member 1', other_designations='AF4/FMR2 family, member 1|ALL1-fused gene from chromosome 4 protein|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 2|pre-B-cell monocytic leukemia partner 1|proto-oncogene AF4', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86934511, end_position_on_the_genomic_accession=87141054, orientation='plus', exon_count=24, OMIM=159557, no_hdr0=''),
  4300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4300, CurrentID=0, Status='live', Symbol='MLLT3', Aliases='AF9, YEATS3', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 3', other_designations='ALL1-fused gene from chromosome 9 protein|YEATS domain-containing protein 3|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog); translocated to, 3|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 3|myeloid/lymphoid or mixed-lineage leukemia translocated to chromosome 3 protein|myeloid/lymphoid or mixed-lineage leukemia; translocated to 3', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20341665, end_position_on_the_genomic_accession=20622543, orientation='minus', exon_count=12, OMIM=159558, no_hdr0=''),
  4301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4301, CurrentID=0, Status='live', Symbol='MLLT4', Aliases='AF6, MLL-AF6', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 4', other_designations='ALL1-fused gene from chromosome 6 protein|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 4|protein AF-6', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167826922, end_position_on_the_genomic_accession=167972023, orientation='plus', exon_count=37, OMIM=159559, no_hdr0=''),
  4302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4302, CurrentID=0, Status='live', Symbol='MLLT6', Aliases='AF17', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 6', other_designations='ALL1-fused gene from chromosome 17 protein|Myeloid/lymphoid or mixed-lineage leukemia, translocated to, 6|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog); translocated to, 6|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 6|trithorax homolog', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38705620, end_position_on_the_genomic_accession=38729803, orientation='plus', exon_count=22, OMIM=600328, no_hdr0=''),
  4303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4303, CurrentID=0, Status='live', Symbol='FOXO4', Aliases='AFX, AFX1, MLLT7', description='forkhead box O4', other_designations='fork head domain transcription factor AFX1|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 7', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71096149, end_position_on_the_genomic_accession=71103534, orientation='plus', exon_count=4, OMIM=300033, no_hdr0=''),
  401679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401679, CurrentID=0, Status='live', Symbol='LOC401679', Aliases='', description='filamin binding LIM protein 1 pseudogene', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46142716, end_position_on_the_genomic_accession=46143591, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4306, CurrentID=0, Status='live', Symbol='NR3C2', Aliases='MCR, MLR, MRVIT, NR3C2', description='nuclear receptor subfamily 3 group C member 2', other_designations='aldosterone receptor|mineralocorticoid receptor 1|mineralocorticoid receptor 2|mineralocorticoid receptor delta|nuclear receptor subfamily 3, group C, member 2 variant 3', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148078764, end_position_on_the_genomic_accession=148445278, orientation='minus', exon_count=12, OMIM=600983, no_hdr0=''),
  4308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4308, CurrentID=0, Status='live', Symbol='TRPM1', Aliases='CSNB1C, LTRPC1, MLSN1', description='transient receptor potential cation channel, subfamily M, member 1', other_designations='long transient receptor potential channel 1|melastatin-1|transient receptor potential melastatin family', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31001061, end_position_on_the_genomic_accession=31161273, orientation='minus', exon_count=29, OMIM=603576, no_hdr0=''),
  4309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4309, CurrentID=0, Status='live', Symbol='MLVI2', Aliases='', description='Moloney murine leukemia virus (MoMuLV) integration site 2 homolog', other_designations='Moloney leukemia virus integration site-2', map_location='5p14', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=157960, no_hdr0=''),
  4311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4311, CurrentID=0, Status='live', Symbol='MME', Aliases='CALLA, CD10, NEP, SFE', description='membrane metallo-endopeptidase', other_designations='atriopeptidase|common acute lymphocytic leukemia antigen|enkephalinase|membrane metallo-endopeptidase (neutral endopeptidase, enkephalinase)|membrane metallo-endopeptidase (neutral endopeptidase, enkephalinase, CALLA, CD10)|membrane metallo-endopeptidase variant 1|membrane metallo-endopeptidase variant 2|neprilysin-390|neprilysin-411|neutral endopeptidase 24.11|skin fibroblast elastase', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155024124, end_position_on_the_genomic_accession=155183729, orientation='plus', exon_count=30, OMIM=120520, no_hdr0=''),
  4312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4312, CurrentID=0, Status='live', Symbol='MMP1', Aliases='CLG, CLGN', description='matrix metallopeptidase 1', other_designations='fibroblast collagenase|matrix metallopeptidase 1 (interstitial collagenase)|matrix metalloprotease 1|matrix metalloproteinase 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102789910, end_position_on_the_genomic_accession=102798235, orientation='minus', exon_count=10, OMIM=120353, no_hdr0=''),
  4313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4313, CurrentID=0, Status='live', Symbol='MMP2', Aliases='CLG4, CLG4A, MMP-2, MMP-II, MONA, TBE-1', description='matrix metallopeptidase 2', other_designations='collagenase type IV-A|matrix metallopeptidase 2 (gelatinase A, 72kDa gelatinase, 72kDa type IV collagenase)|matrix metalloproteinase-2|matrix metalloproteinase-II|neutrophil gelatinase', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55478830, end_position_on_the_genomic_accession=55506691, orientation='plus', exon_count=17, OMIM=120360, no_hdr0=''),
  4314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4314, CurrentID=0, Status='live', Symbol='MMP3', Aliases='CHDS6, MMP-3, SL-1, STMY, STMY1, STR1', description='matrix metallopeptidase 3', other_designations='matrix metalloproteinase 3 (stromelysin 1, progelatinase)|proteoglycanase|transin-1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102835797, end_position_on_the_genomic_accession=102843611, orientation='minus', exon_count=10, OMIM=185250, no_hdr0=''),
  100270799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270799, CurrentID=0, Status='live', Symbol='IBD25', Aliases='', description='Inflammatory bowel disease-25', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612567, no_hdr0=''),
  4316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4316, CurrentID=0, Status='live', Symbol='MMP7', Aliases='MMP-7, MPSL1, PUMP-1', description='matrix metallopeptidase 7', other_designations='matrin|matrix metallopeptidase 7 (matrilysin, uterine)|matrix metalloproteinase 7 (matrilysin, uterine)|matrix metalloproteinase-7|pump-1 protease|uterine matrilysin|uterine metalloproteinase', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102520508, end_position_on_the_genomic_accession=102530747, orientation='minus', exon_count=6, OMIM=178990, no_hdr0=''),
  4317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4317, CurrentID=0, Status='live', Symbol='MMP8', Aliases='CLG1, HNC, MMP-8, PMNL-CL', description='matrix metallopeptidase 8', other_designations='PMN leukocyte collagenase|PMNL collagenase|collagenase 2|matrix metalloproteinase 8 (neutrophil collagenase)|matrix metalloproteinase-8', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102711795, end_position_on_the_genomic_accession=102724964, orientation='minus', exon_count=12, OMIM=120355, no_hdr0=''),
  4318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4318, CurrentID=0, Status='live', Symbol='MMP9', Aliases='CLG4B, GELB, MANDP2, MMP-9', description='matrix metallopeptidase 9', other_designations='macrophage gelatinase|matrix metallopeptidase 9 (gelatinase B, 92kDa gelatinase, 92kDa type IV collagenase)|matrix metalloproteinase 9 (gelatinase B, 92kDa gelatinase, 92kDa type IV collagenase)|type V collagenase', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46008908, end_position_on_the_genomic_accession=46016561, orientation='plus', exon_count=13, OMIM=120361, no_hdr0=''),
  4319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4319, CurrentID=0, Status='live', Symbol='MMP10', Aliases='SL-2, STMY2', description='matrix metallopeptidase 10', other_designations='MMP-10|matrix metalloprotease 10|matrix metalloproteinase 10 (stromelysin 2)|matrix metalloproteinase-10|stromelysin 2|transin 2|transin-2', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102770502, end_position_on_the_genomic_accession=102780628, orientation='minus', exon_count=10, OMIM=185260, no_hdr0=''),
  4320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4320, CurrentID=0, Status='live', Symbol='MMP11', Aliases='SL-3, ST3, STMY3', description='matrix metallopeptidase 11', other_designations='matrix metallopeptidase 11 (stromelysin 3)|stromelysin III', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23772849, end_position_on_the_genomic_accession=23784316, orientation='plus', exon_count=8, OMIM=185261, no_hdr0=''),
  4321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4321, CurrentID=0, Status='live', Symbol='MMP12', Aliases='HME, ME, MME, MMP-12', description='matrix metallopeptidase 12', other_designations='matrix metallopeptidase 12 (macrophage elastase)|matrix metalloproteinase 12 (macrophage elastase)', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102862733, end_position_on_the_genomic_accession=102875034, orientation='minus', exon_count=10, OMIM=601046, no_hdr0=''),
  4322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4322, CurrentID=0, Status='live', Symbol='MMP13', Aliases='CLG3, MANDP1, MMP-13', description='matrix metallopeptidase 13', other_designations='matrix metallopeptidase 13 (collagenase 3)|matrix metalloproteinase 13 (collagenase 3)', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102942992, end_position_on_the_genomic_accession=102955734, orientation='minus', exon_count=10, OMIM=600108, no_hdr0=''),
  4323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4323, CurrentID=0, Status='live', Symbol='MMP14', Aliases='MMP-14, MMP-X1, MT-MMP, MT-MMP 1, MT1-MMP, MT1MMP, MTMMP1, WNCHRS', description='matrix metallopeptidase 14', other_designations='matrix metallopeptidase 14 (membrane-inserted)|membrane type 1 metalloprotease|membrane-type-1 matrix metalloproteinase', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22836533, end_position_on_the_genomic_accession=22847600, orientation='plus', exon_count=10, OMIM=600754, no_hdr0=''),
  4324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4324, CurrentID=0, Status='live', Symbol='MMP15', Aliases='MT2-MMP, MTMMP2, SMCP-2', description='matrix metallopeptidase 15', other_designations='MMP-15|MT-MMP 2|MT2MMP|matrix metallopeptidase 15 (membrane-inserted)|matrix metalloproteinase 15 (membrane-inserted)|membrane-type matrix metalloproteinase 2|membrane-type-2 matrix metalloproteinase', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58025378, end_position_on_the_genomic_accession=58046901, orientation='plus', exon_count=10, OMIM=602261, no_hdr0=''),
  4325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4325, CurrentID=0, Status='live', Symbol='MMP16', Aliases='C8orf57, MMP-X2, MT-MMP2, MT-MMP3, MT3-MMP', description='matrix metallopeptidase 16', other_designations='MMP-16|MT-MMP 3|MT3MMP|MTMMP3|Putative transmembrane protein C8orf57|matrix metallopeptidase 16 (membrane-inserted)|membrane-type matrix metalloproteinase 3|membrane-type-3 matrix metalloproteinase', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88003095, end_position_on_the_genomic_accession=88327488, orientation='minus', exon_count=15, OMIM=602262, no_hdr0=''),
  4326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4326, CurrentID=0, Status='live', Symbol='MMP17', Aliases='MT4-MMP', description='matrix metallopeptidase 17', other_designations='MMP-17|MT-MMP 4|MT4MMP|MTMMP4|matrix metallopeptidase 17 (membrane-inserted)|matrix metalloproteinase 17 (membrane-inserted)|membrane-type matrix metalloproteinase 4|membrane-type-4 matrix metalloproteinase', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131828396, end_position_on_the_genomic_accession=131851771, orientation='plus', exon_count=13, OMIM=602285, no_hdr0=''),
  4327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4327, CurrentID=0, Status='live', Symbol='MMP19', Aliases='CODA, MMP18, RASI-1', description='matrix metallopeptidase 19', other_designations='matrix metalloproteinase RASI|matrix metalloproteinase-18|matrix metalloproteinase-beta19', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55835430, end_position_on_the_genomic_accession=55842983, orientation='minus', exon_count=9, OMIM=601807, no_hdr0=''),
  4329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4329, CurrentID=0, Status='live', Symbol='ALDH6A1', Aliases='MMSADHA, MMSDH', description='aldehyde dehydrogenase 6 family member A1', other_designations='aldehyde dehydrogenase 6 family, member A1|malonate-semialdehyde dehydrogenase|mitochondrial acylating methylmalonate-semialdehyde dehydrogenase', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74057662, end_position_on_the_genomic_accession=74084493, orientation='minus', exon_count=12, OMIM=603178, no_hdr0=''),
  4330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4330, CurrentID=0, Status='live', Symbol='MN1', Aliases='MGCR, MGCR1, MGCR1-PEN, dJ353E16.2', description='meningioma (disrupted in balanced translocation) 1', other_designations='meningioma (translocation balanced)|meningioma chromosome region 1', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27748277, end_position_on_the_genomic_accession=27801498, orientation='minus', exon_count=2, OMIM=156100, no_hdr0=''),
  4331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4331, CurrentID=0, Status='live', Symbol='MNAT1', Aliases='CAP35, MAT1, RNF66, TFB3', description='MNAT CDK-activating kinase assembly factor 1', other_designations='CDK7/cyclin-H assembly factor|RING finger protein 66|RING finger protein MAT1|cyclin G1 interacting protein|cyclin-G1-interacting protein|menage a trois 1 (CAK assembly factor)|menage a trois homolog 1, cyclin H assembly factor|menage a trois-like protein 1 cyclin H assembly factor', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60734741, end_position_on_the_genomic_accession=60968680, orientation='plus', exon_count=9, OMIM=602659, no_hdr0=''),
  4332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4332, CurrentID=0, Status='live', Symbol='MNDA', Aliases='PYHIN3', description='myeloid cell nuclear differentiation antigen', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158831378, end_position_on_the_genomic_accession=158849480, orientation='plus', exon_count=7, OMIM=159553, no_hdr0=''),
  100270802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270802, CurrentID=0, Status='live', Symbol='STQTL16', Aliases='', description='Stature QTL 16', other_designations='', map_location='15q22.32', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612579, no_hdr0=''),
  4335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4335, CurrentID=0, Status='live', Symbol='MNT', Aliases='MAD6, MXD6, ROX, bHLHd3', description='MAX network transcriptional repressor', other_designations='MAX binding protein|MNT, MAX dimerization protein|Max-interacting protein|class D basic helix-loop-helix protein 3|myc antagonist MNT', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2384060, end_position_on_the_genomic_accession=2400964, orientation='minus', exon_count=9, OMIM=603039, no_hdr0=''),
  4336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4336, CurrentID=0, Status='live', Symbol='MOBP', Aliases='', description='myelin-associated oligodendrocyte basic protein', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39467573, end_position_on_the_genomic_accession=39529497, orientation='plus', exon_count=7, OMIM=600948, no_hdr0=''),
  4337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4337, CurrentID=0, Status='live', Symbol='MOCS1', Aliases='MIG11, MOCOD, MOCODA', description='molybdenum cofactor synthesis 1', other_designations='MOCS1A enzyme|cell migration-inducing gene 11 protein|migration-inducing gene 11 protein|molybdenum cofactor biosynthesis protein A|molybdenum cofactor synthesis-step 1 protein A-B', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39903705, end_position_on_the_genomic_accession=39934551, orientation='minus', exon_count=12, OMIM=603707, no_hdr0=''),
  4338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4338, CurrentID=0, Status='live', Symbol='MOCS2', Aliases='MCBPE, MOCO1, MOCODB, MPTS', description='molybdenum cofactor synthesis 2', other_designations='MOCS2A|MOCS2B|molybdenum cofactor biosynthesis protein E|molybdopterin synthase small and large subunit', map_location='5q11', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53095679, end_position_on_the_genomic_accession=53109772, orientation='minus', exon_count=7, OMIM=603708, no_hdr0=''),
  4340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4340, CurrentID=0, Status='live', Symbol='MOG', Aliases='BTN6, BTNL11IG2, NRCLP7, MOG', description='myelin oligodendrocyte glycoprotein', other_designations='MOG AluA|MOG AluB|MOG Ig-AluB|MOG alpha-5', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29656981, end_position_on_the_genomic_accession=29672372, orientation='plus', exon_count=10, OMIM=159465, no_hdr0=''),
  4342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4342, CurrentID=0, Status='live', Symbol='MOS', Aliases='MSV', description='v-mos Moloney murine sarcoma viral oncogene homolog', other_designations='c-mos|oncogene MOS, Moloney murine sarcoma virus|oocyte maturation factor mos|proto-oncogene c-Mos', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56112942, end_position_on_the_genomic_accession=56113982, orientation='minus', exon_count=1, OMIM=190060, no_hdr0=''),
  4343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4343, CurrentID=0, Status='live', Symbol='MOV10', Aliases='fSAP113, gb110', description='Mov10 RISC complex RNA helicase', other_designations='Mov10, Moloney leukemia virus 10, homolog|functional spliceosome-associated protein 113|moloney leukemia virus 10 protein', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112674312, end_position_on_the_genomic_accession=112700746, orientation='plus', exon_count=23, OMIM=610742, no_hdr0=''),
  4345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4345, CurrentID=0, Status='live', Symbol='CD200', Aliases='MOX1, MOX2, MRC, OX-2', description='CD200 molecule', other_designations='CD200 antigen|MRC OX-2 antigen|antigen identified by monoclonal antibody MRC OX-2', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112332347, end_position_on_the_genomic_accession=112362812, orientation='plus', exon_count=10, OMIM=155970, no_hdr0=''),
  4350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4350, CurrentID=0, Status='live', Symbol='MPG', Aliases='AAG, ADPG, APNG, CRA36.1, MDG, Mid1, PIG11, PIG16, anpg', description='N-methylpurine DNA glycosylase', other_designations="3' end of the Mid1 gene, localized 68 kb upstream the humanzeta globin gene on 16p|3-alkyladenine DNA glycosylase|3-methyladenine DNA glycosidase|CRA36.1 (3-methyl-adenine DNA glycosylase)|N-methylpurine-DNA glycosylase, MPG|proliferation-inducing protein 11|proliferation-inducing protein 16", map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77019, end_position_on_the_genomic_accession=85851, orientation='plus', exon_count=6, OMIM=156565, no_hdr0=''),
  4351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4351, CurrentID=0, Status='live', Symbol='MPI', Aliases='CDG1B, PMI, PMI1', description='mannose phosphate isomerase', other_designations='phosphohexomutase|phosphomannose isomerase 1', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74890011, end_position_on_the_genomic_accession=74899457, orientation='plus', exon_count=8, OMIM=154550, no_hdr0=''),
  4352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4352, CurrentID=0, Status='live', Symbol='MPL', Aliases='C-MPL, CD110V, THCYT2, TPOR, MPL', description='MPL proto-oncogene, thrombopoietin receptor', other_designations='TPO-R|myeloproliferative leukemia protein|myeloproliferative leukemia virus oncogene|proto-oncogene c-Mpl', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43337418, end_position_on_the_genomic_accession=43354464, orientation='plus', exon_count=12, OMIM=159530, no_hdr0=''),
  4353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4353, CurrentID=0, Status='live', Symbol='MPO', Aliases='', description='myeloperoxidase', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58269855, end_position_on_the_genomic_accession=58284313, orientation='minus', exon_count=14, OMIM=606989, no_hdr0=''),
  4354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4354, CurrentID=0, Status='live', Symbol='MPP1', Aliases='AAG12, DXS552E, EMP55, MRG1, PEMP', description='membrane protein, palmitoylated 1', other_designations='aging-associated gene 12|erythrocyte membrane protein p55|membrane protein, palmitoylated 1, 55kDa|migration-related gene 1|p55|palmitoylated erythrocyte membrane protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154778684, end_position_on_the_genomic_accession=154805527, orientation='minus', exon_count=14, OMIM=305360, no_hdr0=''),
  4355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4355, CurrentID=0, Status='live', Symbol='MPP2', Aliases='DLG2', description='membrane protein, palmitoylated 2', other_designations='discs large, homolog 2|membrane protein, palmitoylated 2 (MAGUK p55 subfamily member 2)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43875357, end_position_on_the_genomic_accession=43909711, orientation='minus', exon_count=19, OMIM=600723, no_hdr0=''),
  4356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4356, CurrentID=0, Status='live', Symbol='MPP3', Aliases='DLG3', description='membrane protein, palmitoylated 3', other_designations='discs large homolog 3|discs, large homolog 3|membrane protein palmitoylated 3|membrane protein, palmitoylated 3 (MAGUK p55 subfamily member 3)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43800799, end_position_on_the_genomic_accession=43833179, orientation='minus', exon_count=23, OMIM=601114, no_hdr0=''),
  4357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4357, CurrentID=0, Status='live', Symbol='MPST', Aliases='MST, TST2, TUM1', description='mercaptopyruvate sulfurtransferase', other_designations='human liver rhodanese|tRNA thiouridin modification protein 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37019642, end_position_on_the_genomic_accession=37029822, orientation='plus', exon_count=6, OMIM=602496, no_hdr0=''),
  4358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4358, CurrentID=0, Status='live', Symbol='MPV17', Aliases='MTDPS6, SYM1', description='MpV17 mitochondrial inner membrane protein', other_designations='Mpv17, human homolog of glomerulosclerosis and nephrotic syndrome', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27309490, end_position_on_the_genomic_accession=27325680, orientation='minus', exon_count=10, OMIM=137960, no_hdr0=''),
  4359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4359, CurrentID=0, Status='live', Symbol='MPZ', Aliases='CHM, CMT1, CMT1B, CMT2I, CMT2J, CMT4E, CMTDI3, CMTDID, DSS, HMSNIB, MPP, P0', description='myelin protein zero', other_designations='Charcot-Marie-Tooth neuropathy 1B|myelin peripheral protein', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161304735, end_position_on_the_genomic_accession=161309972, orientation='minus', exon_count=6, OMIM=159440, no_hdr0=''),
  4360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4360, CurrentID=0, Status='live', Symbol='MRC1', Aliases='CD206, CLEC13D, CLEC13DL, MMRL1, bA541I19.1, hMR, MRC1', description='mannose receptor, C type 1', other_designations='C-type lectin domain family 13 member D|human mannose receptor|macrophage mannose receptor 1-like protein 1|mannose receptor, C type 1-like 1', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17809343, end_position_on_the_genomic_accession=17911162, orientation='plus', exon_count=30, OMIM=153618, no_hdr0=''),
  4361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4361, CurrentID=0, Status='live', Symbol='MRE11A', Aliases='ATLD, HNGS1, MRE11, MRE11B', description='MRE11 homolog A, double strand break repair nuclease', other_designations='AT-like disease|DNA recombination and repair protein|MRE11 double strand break repair nuclease A|MRE11 homolog 1|MRE11 homolog A|MRE11 homolog, double strand break repair nuclease A|MRE11 meiotic recombination 11 homolog A|MRE11 meiotic recombination 11-like protein A|endo/exonuclease Mre11|meiotic recombination 11 homolog 1|meiotic recombination 11 homolog A', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94415578, end_position_on_the_genomic_accession=94512185, orientation='minus', exon_count=22, OMIM=600814, no_hdr0=''),
  4362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4362, CurrentID=0, Status='live', Symbol='MRE11B', Aliases='hMRE11B', description='MRE11 homolog B, double strand break repair nuclease (pseudogene)', other_designations='MRE11 double strand break repair nuclease B (pseudogene)|MRE11 homolog, double strand break repair nuclease B (pseudogene)|MRE11 meiotic recombination 11 homolog B (pseudogene)|meiotic recombination 11 homolog B', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156110591, end_position_on_the_genomic_accession=156114449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4363, CurrentID=0, Status='live', Symbol='ABCC1', Aliases='ABC29, ABCC, GS-X, MRP, MRP1', description='ATP binding cassette subfamily C member 1', other_designations='ATP-binding cassette transporter variant ABCC1delta-ex13|ATP-binding cassette transporter variant ABCC1delta-ex13&14|ATP-binding cassette transporter variant ABCC1delta-ex25|ATP-binding cassette transporter variant ABCC1delta-ex25&26|ATP-binding cassette, sub-family C (CFTR/MRP), member 1|LTC4 transporter|leukotriene C(4) transporter', map_location='16p13.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15949577, end_position_on_the_genomic_accession=16143074, orientation='plus', exon_count=33, OMIM=158343, no_hdr0=''),
  4364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4364, CurrentID=0, Status='live', Symbol='MRSD', Aliases='CHRS', description='mental retardation-skeletal dysplasia', other_designations='', map_location='Xq27-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309620, no_hdr0=''),
  4366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4366, CurrentID=0, Status='live', Symbol='MRX4', Aliases='', description='mental retardation, X-linked 4', other_designations='', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371957, CurrentID=0, Status='live', Symbol='LOC105371957', Aliases='', description='uncharacterized LOC105371957', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2020959, end_position_on_the_genomic_accession=2038349, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  4369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4369, CurrentID=0, Status='live', Symbol='MRX5', Aliases='', description='mental retardation, X-linked 5', other_designations='', map_location='Xp', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4370, CurrentID=0, Status='live', Symbol='MRX6', Aliases='', description='mental retardation, X-linked 6 (Okinawa type)', other_designations='', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4371, CurrentID=0, Status='live', Symbol='MRX7', Aliases='', description='mental retardation, X-linked 7', other_designations='', map_location='Xp11.23-q12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4372, CurrentID=0, Status='live', Symbol='MRX8', Aliases='', description='mental retardation, X-linked 8', other_designations='', map_location='Xq21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4375, CurrentID=0, Status='live', Symbol='MRX11', Aliases='', description='mental retardation, X-linked 11', other_designations='', map_location='Xp21.3-p11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414775, CurrentID=0, Status='live', Symbol='HCG21', Aliases='NCRNA00150', description='HLA complex group 21 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4378, CurrentID=0, Status='live', Symbol='MRX14', Aliases='', description='mental retardation, X-linked 14', other_designations='', map_location='Xp11.3-q13.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300062, no_hdr0=''),
  4379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4379, CurrentID=0, Status='live', Symbol='MRX15', Aliases='', description='mental retardation, X-linked 15', other_designations='', map_location='Xp21.1-p11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4381, CurrentID=0, Status='live', Symbol='MRX17', Aliases='', description='mental retardation, X-linked 17', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4382, CurrentID=0, Status='live', Symbol='MRX18', Aliases='', description='mental retardation, X-linked 18', other_designations='', map_location='Xp21.2-p11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4384, CurrentID=0, Status='live', Symbol='MRX20', Aliases='', description='mental retardation, X-linked 20', other_designations='', map_location='Xp11-q21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300047, no_hdr0=''),
  135458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135458, CurrentID=0, Status='live', Symbol='HUS1B', Aliases='', description='HUS1 checkpoint clamp component B', other_designations='HUS1 checkpoint homolog b|hHUS1B', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=655939, end_position_on_the_genomic_accession=656964, orientation='minus', exon_count=1, OMIM=609713, no_hdr0=''),
  4387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4387, CurrentID=0, Status='live', Symbol='MRX23', Aliases='', description='mental retardation, X-linked 23', other_designations='', map_location='Xq23-q24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300046, no_hdr0=''),
  4388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4388, CurrentID=0, Status='live', Symbol='MRX24', Aliases='', description='mental retardation, X-linked 24', other_designations='', map_location='Xp22.3-p22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4389, CurrentID=0, Status='live', Symbol='MRX25', Aliases='', description='mental retardation, X-linked 25', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4390, CurrentID=0, Status='live', Symbol='MRX26', Aliases='', description='mental retardation, X-linked 26', other_designations='', map_location='Xp11.3-q21.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4391, CurrentID=0, Status='live', Symbol='MRX27', Aliases='', description='mental retardation, X-linked 27', other_designations='', map_location='Xq24-q26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4392, CurrentID=0, Status='live', Symbol='MRX28', Aliases='', description='mental retardation, X-linked 28', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4395, CurrentID=0, Status='live', Symbol='MRX31', Aliases='', description='mental retardation, X-linked 31', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4397, CurrentID=0, Status='live', Symbol='MS', Aliases='MS1', description='multiple sclerosis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=126200, no_hdr0=''),
  4402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4402, CurrentID=0, Status='live', Symbol='MRX37', Aliases='', description='mental retardation, X-linked 37', other_designations='', map_location='Xp22.32-p22.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4404, CurrentID=0, Status='live', Symbol='MRX39', Aliases='', description='mental retardation, X-linked 39', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4405, CurrentID=0, Status='live', Symbol='MRX40', Aliases='', description='mental retardation, X-linked 40', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4406, CurrentID=0, Status='live', Symbol='MRX42', Aliases='', description='mental retardation, X-linked 42', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300372, no_hdr0=''),
  266553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266553, CurrentID=0, Status='live', Symbol='OFCC1', Aliases='MRDS1', description='orofacial cleft 1 candidate 1', other_designations='orofacial clefting chromosomal breakpoint region candidate 1 protein', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=9845518, end_position_on_the_genomic_accession=10159438, orientation='minus', exon_count=21, OMIM=614287, no_hdr0=''),
  4412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4412, CurrentID=0, Status='live', Symbol='MRX49', Aliases='', description='mental retardation, X-linked 49', other_designations='', map_location='Xp22.3-p22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300114, no_hdr0=''),
  4413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4413, CurrentID=0, Status='live', Symbol='MRX50', Aliases='', description='mental retardation, X-linked 50', other_designations='', map_location='Xp11.3-p11.21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300115, no_hdr0=''),
  4414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4414, CurrentID=0, Status='live', Symbol='MRX51', Aliases='', description='mental retardation, X-linked 51', other_designations='', map_location='Xp11.3-p11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4416, CurrentID=0, Status='live', Symbol='MRX53', Aliases='', description='mental retardation, X-linked 53', other_designations='', map_location='Xq22.2-q24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300324, no_hdr0=''),
  4419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4419, CurrentID=0, Status='live', Symbol='MRX56', Aliases='', description='mental retardation, X-linked 56', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4420, CurrentID=0, Status='live', Symbol='MRX57', Aliases='', description='mental retardation, X-linked 57', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4424, CurrentID=0, Status='live', Symbol='MRX61', Aliases='', description='mental retardation, X-linked 61', other_designations='', map_location='Xq13.1-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4427, CurrentID=0, Status='live', Symbol='MRX64', Aliases='', description='mental retardation, X-linked 64', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4428, CurrentID=0, Status='live', Symbol='MRX65', Aliases='', description='mental retardation, X-linked 65', other_designations='', map_location='Xp11.23-q21.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4429, CurrentID=0, Status='live', Symbol='MRXS7', Aliases='', description='mental retardation, X-linked, syndromic 7', other_designations='', map_location='Xp11.3-q22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300218, no_hdr0=''),
  4430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4430, CurrentID=0, Status='live', Symbol='MYO1B', Aliases='myr1', description='myosin IB', other_designations='MMI-alpha|MMIa|MYH-1c|MYO1B variant protein|myosin I alpha|myosin-I alpha|myosin-Ib', map_location='2q12-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191245185, end_position_on_the_genomic_accession=191425389, orientation='plus', exon_count=34, OMIM=606537, no_hdr0=''),
  4431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4431, CurrentID=0, Status='live', Symbol='MSBP1', Aliases='', description='minisatellite binding protein 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4432, CurrentID=0, Status='live', Symbol='MSBP2', Aliases='', description='minisatellite binding protein 2', other_designations='minisatellite binding protein 2, 77kDa', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723497, CurrentID=0, Status='live', Symbol='LOC102723497', Aliases='', description='uncharacterized LOC102723497', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41108501, end_position_on_the_genomic_accession=41131238, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  4434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4434, CurrentID=0, Status='live', Symbol='MSD', Aliases='', description='microcephaly with spastic diplegia (Paine syndrome)', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=311400, no_hdr0=''),
  4435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4435, CurrentID=0, Status='live', Symbol='CITED1', Aliases='MSG1', description='Cbp/p300-interacting transactivator, with Glu/Asp rich carboxy-terminal domain, 1', other_designations='Cbp/p300-interacting transactivator, with Glu/Asp-rich carboxy-terminal domain, 1|melanocyte-specific gene 1|melanocyte-specific protein 1', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72301638, end_position_on_the_genomic_accession=72307187, orientation='minus', exon_count=6, OMIM=300149, no_hdr0=''),
  4436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4436, CurrentID=0, Status='live', Symbol='MSH2', Aliases='COCA1, FCC1, HNPCC, HNPCC1, LCFS2', description='mutS homolog 2', other_designations='hMSH2|mutS homolog 2, colon cancer, nonpolyposis type 1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47403067, end_position_on_the_genomic_accession=47564579, orientation='plus', exon_count=19, OMIM=609309, no_hdr0=''),
  4437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4437, CurrentID=0, Status='live', Symbol='MSH3', Aliases='DUP, MRP1', description='mutS homolog 3', other_designations='divergent upstream protein|hMSH3|mismatch repair protein 1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80654648, end_position_on_the_genomic_accession=80876815, orientation='plus', exon_count=24, OMIM=600887, no_hdr0=''),
  4438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4438, CurrentID=0, Status='live', Symbol='MSH4', Aliases='', description='mutS homolog 4', other_designations='hMSH4', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75796871, end_position_on_the_genomic_accession=75913238, orientation='plus', exon_count=20, OMIM=602105, no_hdr0=''),
  4439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4439, CurrentID=0, Status='live', Symbol='MSH5', Aliases='G7, MUTSH5, NG23', description='mutS homolog 5', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31739948, end_position_on_the_genomic_accession=31762678, orientation='plus', exon_count=25, OMIM=603382, no_hdr0=''),
  4440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4440, CurrentID=0, Status='live', Symbol='MSI1', Aliases='', description='musashi RNA binding protein 1', other_designations='musashi RNA-binding protein 1|musashi-1|musashi1', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120339650, end_position_on_the_genomic_accession=120369180, orientation='minus', exon_count=22, OMIM=603328, no_hdr0=''),
  4447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4447, CurrentID=0, Status='live', Symbol='MSK9', Aliases='', description='antigen identified by monoclonal antibody K15', other_designations='', map_location='3pter-p14', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4448, CurrentID=0, Status='live', Symbol='MSK10', Aliases='', description='antigen identified by monoclonal antibody AJ425', other_designations='', map_location='3pter-p14', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130175, CurrentID=0, Status='live', Symbol='LOC100130175', Aliases='', description='uncharacterized LOC100130175', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=112653427, end_position_on_the_genomic_accession=112655020, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377195, CurrentID=0, Status='live', Symbol='LOC105377195', Aliases='', description='uncharacterized LOC105377195', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29359966, end_position_on_the_genomic_accession=29434513, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  4467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4467, CurrentID=0, Status='live', Symbol='MSK32', Aliases='', description='antigen identified by monoclonal antibody K66', other_designations='', map_location='3pter-p14', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372762, CurrentID=0, Status='live', Symbol='LOC105372762', Aliases='', description='uncharacterized LOC105372762', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27419015, end_position_on_the_genomic_accession=27448644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270825, CurrentID=0, Status='live', Symbol='RPL9P12', Aliases='RPL9_1_50', description='ribosomal protein L9 pseudogene 12', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52189905, end_position_on_the_genomic_accession=52190309, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4473, CurrentID=0, Status='live', Symbol='MSK38', Aliases='', description='antigen identified by monoclonal antibody O5', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266619, CurrentID=0, Status='live', Symbol='NOS2P4', Aliases='', description='nitric oxide synthase 2 pseudogene 4', other_designations='nitric oxide synthase 2, inducible pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16812441, end_position_on_the_genomic_accession=16812651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421751, CurrentID=0, Status='live', Symbol='LOC100421751', Aliases='', description='ATPase, Na+/K+ transporting, beta 3 polypeptide pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19907999, end_position_on_the_genomic_accession=19908645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4477, CurrentID=0, Status='live', Symbol='MSMB', Aliases='HPC13, IGBF, MSP, MSPB, PN44, PRPS, PSP, PSP-94, PSP57, PSP94', description='microseminoprotein, beta-', other_designations='immunoglobulin binding factor|prostate secreted seminal plasma protein|prostate secretory protein of 94 amino acids|seminal plasma beta-inhibin', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46033305, end_position_on_the_genomic_accession=46046269, orientation='minus', exon_count=4, OMIM=157145, no_hdr0=''),
  4478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4478, CurrentID=0, Status='live', Symbol='MSN', Aliases='HEL70', description='moesin', other_designations='epididymis luminal protein 70|membrane-organizing extension spike protein', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65588382, end_position_on_the_genomic_accession=65741931, orientation='plus', exon_count=16, OMIM=309845, no_hdr0=''),
  4479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4479, CurrentID=0, Status='live', Symbol='MSNP1', Aliases='MSNL1', description='moesin pseudogene 1', other_designations='moesin-like 1', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=25909307, end_position_on_the_genomic_accession=25913278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4481, CurrentID=0, Status='live', Symbol='MSR1', Aliases='CD204, SCARA1, SR-A, SRA, phSR1, phSR2', description='macrophage scavenger receptor 1', other_designations='macrophage acetylated LDL receptor I and II|macrophage scavenger receptor type III|scavenger receptor class A member 1|scavenger receptor class A, member 1', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16107878, end_position_on_the_genomic_accession=16192791, orientation='minus', exon_count=11, OMIM=153622, no_hdr0=''),
  4482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4482, CurrentID=0, Status='live', Symbol='MSRA', Aliases='PMSR', description='methionine sulfoxide reductase A', other_designations='cytosolic methionine-S-sulfoxide reductase|peptide Met(O) reductase|peptide met (O) reductase|peptide-methionine (S)-S-oxide reductase', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10054224, end_position_on_the_genomic_accession=10428891, orientation='plus', exon_count=10, OMIM=601250, no_hdr0=''),
  100270827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270827, CurrentID=0, Status='live', Symbol='RPS2P14', Aliases='RPS2_2_88', description='ribosomal protein S2 pseudogene 14', other_designations='', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113698884, end_position_on_the_genomic_accession=113699606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4485, CurrentID=0, Status='live', Symbol='MST1', Aliases='D3F15S2, DNF15S2, HGFL, MSP, NF15S2', description='macrophage stimulating 1', other_designations='hepatocyte growth factor-like protein homolog|macrophage stimulating 1 (hepatocyte growth factor-like)|macrophage-stimulating protein', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49683947, end_position_on_the_genomic_accession=49689528, orientation='minus', exon_count=17, OMIM=142408, no_hdr0=''),
  4486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4486, CurrentID=0, Status='live', Symbol='MST1R', Aliases='CD136, CDw136, PTK8, RON', description='macrophage stimulating 1 receptor', other_designations='MSP receptor|MST1R variant RON30|MST1R variant RON62|PTK8 protein tyrosine kinase 8|RON variant 21|RON variant E2E3|c-met-related tyrosine kinase|macrophage stimulating 1 receptor (c-met-related tyrosine kinase)|p185-Ron|soluble RON variant 1|soluble RON variant 2|soluble RON variant 3|soluble RON variant 4', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49886981, end_position_on_the_genomic_accession=49903873, orientation='minus', exon_count=20, OMIM=600168, no_hdr0=''),
  4487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4487, CurrentID=0, Status='live', Symbol='MSX1', Aliases='ECTD3, HOX7, HYD1, STHAG1', description='msh homeobox 1', other_designations='homeobox 7|homeobox protein Hox-7|msh homeo box 1|msh homeobox 1-like protein|msh homeobox homolog 1', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4859665, end_position_on_the_genomic_accession=4863936, orientation='plus', exon_count=2, OMIM=142983, no_hdr0=''),
  4488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4488, CurrentID=0, Status='live', Symbol='MSX2', Aliases='CRS2, FPP, HOX8, MSH, PFM, PFM1', description='msh homeobox 2', other_designations='homeobox protein Hox-8|msh homeo box 2|msh homeobox homolog 2', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174724572, end_position_on_the_genomic_accession=174730899, orientation='plus', exon_count=2, OMIM=123101, no_hdr0=''),
  4489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4489, CurrentID=0, Status='live', Symbol='MT1A', Aliases='MT1, MT1S, MTC', description='metallothionein 1A', other_designations='MT-1A|MT-IA|metallothionein 1A (functional)|metallothionein 1S|metallothionein-IA', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56638666, end_position_on_the_genomic_accession=56640087, orientation='plus', exon_count=3, OMIM=156350, no_hdr0=''),
  4490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4490, CurrentID=0, Status='live', Symbol='MT1B', Aliases='MT1, MT1Q, MTP', description='metallothionein 1B', other_designations='MT-1B|MT-IB|metallothionein 1B (functional)|metallothionein 1Q|metallothionein-IB', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56651899, end_position_on_the_genomic_accession=56653204, orientation='plus', exon_count=3, OMIM=156349, no_hdr0=''),
  4493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4493, CurrentID=0, Status='live', Symbol='MT1E', Aliases='MT-1E, MT-IE, MT1, MTD', description='metallothionein 1E', other_designations='metallothionein 1E (functional)|metallothionein D|metallothionein-IE', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56625673, end_position_on_the_genomic_accession=56627112, orientation='plus', exon_count=3, OMIM=156351, no_hdr0=''),
  4494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4494, CurrentID=0, Status='live', Symbol='MT1F', Aliases='MT1', description='metallothionein 1F', other_designations='MT-1F|MT-IF|metallothionein 1F (functional)|metallothionein-IF', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56657943, end_position_on_the_genomic_accession=56659303, orientation='plus', exon_count=3, OMIM=156352, no_hdr0=''),
  4495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4495, CurrentID=0, Status='live', Symbol='MT1G', Aliases='MT1, MT1K', description='metallothionein 1G', other_designations='MT-1G|MT-1K|MT-IG|metallothionein 1K|metallothionein-1K|metallothionein-IG', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56666735, end_position_on_the_genomic_accession=56668065, orientation='minus', exon_count=3, OMIM=156353, no_hdr0=''),
  4496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4496, CurrentID=0, Status='live', Symbol='MT1H', Aliases='MT-0, MT-1H, MT-IH, MT1', description='metallothionein 1H', other_designations='metallothionein-0|metallothionein-IH', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56669814, end_position_on_the_genomic_accession=56671129, orientation='plus', exon_count=3, OMIM=156354, no_hdr0=''),
  4498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4498, CurrentID=0, Status='live', Symbol='MT1JP', Aliases='MT1, MT1J, MT1NP, MTB', description='metallothionein 1J, pseudogene', other_designations='metallothionein 1J (pseudogene)|metallothionein 1M pseudogene|metallothionein 1N, pseudogene|metallothionein B', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56635739, end_position_on_the_genomic_accession=56637086, orientation='plus', exon_count=3, OMIM=156356, no_hdr0=''),
  4499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4499, CurrentID=0, Status='live', Symbol='MT1M', Aliases='MT-1M, MT-IM, MT1, MT1K', description='metallothionein 1M', other_designations='metallothionein 1K|metallothionein 1Y', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56632622, end_position_on_the_genomic_accession=56633986, orientation='plus', exon_count=3, OMIM=156357, no_hdr0=''),
  4500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4500, CurrentID=0, Status='live', Symbol='MT1L', Aliases='MT1, MT1R, MTF', description='metallothionein 1L (gene/pseudogene)', other_designations='metallothionein 1L (pseudogene)|metallothionein 1R', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56617461, end_position_on_the_genomic_accession=56618818, orientation='plus', exon_count=3, OMIM=156358, no_hdr0=''),
  4501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4501, CurrentID=0, Status='live', Symbol='MT1X', Aliases='MT-1l, MT1', description='metallothionein 1X', other_designations='MT-1X|MT-IX|metallothionein-IX', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56682470, end_position_on_the_genomic_accession=56684196, orientation='plus', exon_count=3, OMIM=156359, no_hdr0=''),
  4502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4502, CurrentID=0, Status='live', Symbol='MT2A', Aliases='MT2', description='metallothionein 2A', other_designations='MT-2|MT-II|metallothionein-2A|metallothionein-II', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56608566, end_position_on_the_genomic_accession=56609497, orientation='plus', exon_count=3, OMIM=156360, no_hdr0=''),
  4503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4503, CurrentID=0, Status='live', Symbol='MT2P1', Aliases='', description='metallothionein 2 pseudogene 1', other_designations='metallothionein 2 pseudogene 1 (processed)', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68376251, end_position_on_the_genomic_accession=68376665, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4504, CurrentID=0, Status='live', Symbol='MT3', Aliases='GIF, GIFB, GRIF, ZnMT3', description='metallothionein 3', other_designations='MT-3|MT-III|growth inhibitory factor|metallothionein 3 (growth inhibitory factor (neurotrophic))|metallothionein-III', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56589355, end_position_on_the_genomic_accession=56591088, orientation='plus', exon_count=3, OMIM=139255, no_hdr0=''),
  4507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4507, CurrentID=0, Status='live', Symbol='MTAP', Aliases='BDMF, DMSFH, DMSMFH, HEL-249, LGMBF, MSAP, c86fus', description='methylthioadenosine phosphorylase', other_designations="5'-methylthioadenosine phosphorylase|MTA phosphorylase|MTAPase|MeSAdo phosphorylase|epididymis luminal protein 249", map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21802636, end_position_on_the_genomic_accession=21865971, orientation='plus', exon_count=8, OMIM=156540, no_hdr0=''),
  4508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4508, CurrentID=0, Status='live', Symbol='MT-ATP6', Aliases='ATPase6, MTATP6, ATP6', description='mitochondrially encoded ATP synthase 6', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=8527, end_position_on_the_genomic_accession=9207, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4509, CurrentID=0, Status='live', Symbol='MT-ATP8', Aliases='ATPase8, MTATP8, ATP8', description='mitochondrially encoded ATP synthase 8', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=8366, end_position_on_the_genomic_accession=8572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4511, CurrentID=0, Status='live', Symbol='MT-TC', Aliases='MTTC, TRNC', description='mitochondrially encoded tRNA cysteine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5761, end_position_on_the_genomic_accession=5826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4512, CurrentID=0, Status='live', Symbol='MT-CO1', Aliases='COI, MTCO1, COX1', description='mitochondrially encoded cytochrome c oxidase I', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5904, end_position_on_the_genomic_accession=7445, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4513, CurrentID=0, Status='live', Symbol='MT-CO2', Aliases='COII, MTCO2, COX2', description='mitochondrially encoded cytochrome c oxidase II', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=7586, end_position_on_the_genomic_accession=8269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4514, CurrentID=0, Status='live', Symbol='MT-CO3', Aliases='COIII, MTCO3, COX3', description='mitochondrially encoded cytochrome c oxidase III', other_designations='cytochrome c oxidase subunit III', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=9207, end_position_on_the_genomic_accession=9990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4515, CurrentID=0, Status='live', Symbol='MTCP1', Aliases='P13MTCP1, p8MTCP1', description='mature T-cell proliferation 1', other_designations='MTCP-1 type B1|mature T-cell proliferation 1 isoform p13|mature T-cell proliferation-1 type B1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155064034, end_position_on_the_genomic_accession=155071272, orientation='minus', exon_count=5, OMIM=300116, no_hdr0=''),
  266661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266661, CurrentID=0, Status='live', Symbol='TAS2R22', Aliases='T2R22', description='taste 2 receptor member 22', other_designations='taste receptor, type 2, member 22', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4519, CurrentID=0, Status='live', Symbol='MT-CYB', Aliases='MTCYB, CYTB', description='mitochondrially encoded cytochrome b', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=14747, end_position_on_the_genomic_accession=15887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4520, CurrentID=0, Status='live', Symbol='MTF1', Aliases='MTF-1, ZRF', description='metal-regulatory transcription factor 1', other_designations='MRE-binding transcription factor|MRE-binding transcription factor-1|metal-responsive transcription factor 1|transcription factor MTF-1|zinc regulatory factor', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37809567, end_position_on_the_genomic_accession=37859624, orientation='minus', exon_count=15, OMIM=600172, no_hdr0=''),
  4521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4521, CurrentID=0, Status='live', Symbol='NUDT1', Aliases='MTH1', description='nudix hydrolase 1', other_designations='2-hydroxy-dATP diphosphatase|8-oxo-7,8-dihydrodeoxyguanosine triphosphatase|8-oxo-7,8-dihydroguanosine triphosphatase|8-oxo-dGTPase|mutT human homolog 1|nucleoside diphosphate-linked moiety X motif 1|nucleoside diphosphate-linked moiety X-type motif 1|nudix (nucleoside diphosphate linked moiety X)-type motif 1|nudix motif 1', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2242222, end_position_on_the_genomic_accession=2251145, orientation='plus', exon_count=6, OMIM=600312, no_hdr0=''),
  4522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4522, CurrentID=0, Status='live', Symbol='MTHFD1', Aliases='MTHFC, MTHFD', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1, methenyltetrahydrofolate cyclohydrolase, formyltetrahydrofolate synthetase', other_designations='5,10-methylenetetrahydrofolate dehydrogenase, 5,10-methylenetetrahydrofolate cyclohydrolase, 10-formyltetrahydrofolate synthetase|C1-THF synthase|cytoplasmic C-1-tetrahydrofolate synthase', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64388041, end_position_on_the_genomic_accession=64460007, orientation='plus', exon_count=28, OMIM=172460, no_hdr0=''),
  790955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790955, CurrentID=0, Status='live', Symbol='UQCC3', Aliases='C11orf83, CCDS41658.1, MC3DN9, UNQ655', description='ubiquinol-cytochrome c reductase complex assembly factor 3', other_designations='UPF0723 protein C11orf83|assembly factor CBP4 homolog', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62671654, end_position_on_the_genomic_accession=62673690, orientation='plus', exon_count=2, OMIM=616097, no_hdr0=''),
  4524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4524, CurrentID=0, Status='live', Symbol='MTHFR', Aliases='', description='methylenetetrahydrofolate reductase (NAD(P)H)', other_designations='5,10-methylenetetrahydrofolate reductase (NADPH)', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11785730, end_position_on_the_genomic_accession=11806103, orientation='minus', exon_count=14, OMIM=607093, no_hdr0=''),
  100270834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270834, CurrentID=0, Status='live', Symbol='RPL5P13', Aliases='RPL5_5_523', description='ribosomal protein L5 pseudogene 13', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=142286917, end_position_on_the_genomic_accession=142287625, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266671, CurrentID=0, Status='live', Symbol='MAGI2-IT1', Aliases='MAGI2-IT, MAGI2IT, MAGI2IT1, PR47', description='MAGI2 intronic transcript 1', other_designations='MAGI2 intronic transcript (non-protein coding)|MAGI2 intronic transcript 1 (non-protein coding)|platelet receptor for type III collagen, 47 kDa', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608950, no_hdr0=''),
  4528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4528, CurrentID=0, Status='live', Symbol='MTIF2', Aliases='', description='mitochondrial translational initiation factor 2', other_designations='IF-2(Mt)|IF-2mt|IF2(mt)', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55236620, end_position_on_the_genomic_accession=55269303, orientation='minus', exon_count=19, OMIM=603766, no_hdr0=''),
  266673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266673, CurrentID=0, Status='live', Symbol='ZNF402P', Aliases='', description='zinc finger protein 402, pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16856938, end_position_on_the_genomic_accession=16860731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266675, CurrentID=0, Status='live', Symbol='BEST4', Aliases='VMD2L2', description='bestrophin 4', other_designations='vitelliform macular dystrophy 2-like 2|vitelliform macular dystrophy 2-like protein 2', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44782045, end_position_on_the_genomic_accession=44792828, orientation='minus', exon_count=14, OMIM=607336, no_hdr0=''),
  4534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4534, CurrentID=0, Status='live', Symbol='MTM1', Aliases='CNM, MTMX, XLMTM', description='myotubularin 1', other_designations='phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150562658, end_position_on_the_genomic_accession=150673143, orientation='plus', exon_count=19, OMIM=300415, no_hdr0=''),
  4535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4535, CurrentID=0, Status='live', Symbol='MT-ND1', Aliases='MTND1, ND1', description='mitochondrially encoded NADH dehydrogenase 1', other_designations='NADH dehydrogenase subunit 1', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=3307, end_position_on_the_genomic_accession=4262, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4536, CurrentID=0, Status='live', Symbol='MT-ND2', Aliases='MTND2, ND2', description='mitochondrially encoded NADH dehydrogenase 2', other_designations='NADH dehydrogenase subunit 2', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=4470, end_position_on_the_genomic_accession=5511, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4537, CurrentID=0, Status='live', Symbol='MT-ND3', Aliases='MTND3, ND3', description='mitochondrially encoded NADH dehydrogenase 3', other_designations='NADH dehydrogenase subunit 3', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=10059, end_position_on_the_genomic_accession=10404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4538, CurrentID=0, Status='live', Symbol='MT-ND4', Aliases='MTND4, ND4', description='mitochondrially encoded NADH dehydrogenase 4', other_designations='NADH dehydrogenase subunit 4', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=10760, end_position_on_the_genomic_accession=12137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4539, CurrentID=0, Status='live', Symbol='MT-ND4L', Aliases='MTND4L, ND4L', description='mitochondrially encoded NADH 4L dehydrogenase', other_designations='NADH dehydrogenase subunit 4L', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=10470, end_position_on_the_genomic_accession=10766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4540, CurrentID=0, Status='live', Symbol='MT-ND5', Aliases='MTND5, ND5', description='mitochondrially encoded NADH dehydrogenase 5', other_designations='NADH dehydrogenase subunit 5', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=12337, end_position_on_the_genomic_accession=14148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4541, CurrentID=0, Status='live', Symbol='MT-ND6', Aliases='MTND6, ND6', description='mitochondrially encoded NADH dehydrogenase 6', other_designations='NADH dehydrogenase subunit 6', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=14149, end_position_on_the_genomic_accession=14673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4542, CurrentID=0, Status='live', Symbol='MYO1F', Aliases='', description='myosin IF', other_designations='myosin-ID|myosin-Ie', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8520790, end_position_on_the_genomic_accession=8577447, orientation='minus', exon_count=28, OMIM=601480, no_hdr0=''),
  4543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4543, CurrentID=0, Status='live', Symbol='MTNR1A', Aliases='MEL-1A-R, MT1', description='melatonin receptor 1A', other_designations='mel1a receptor', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186533000, end_position_on_the_genomic_accession=186555383, orientation='minus', exon_count=3, OMIM=600665, no_hdr0=''),
  4544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4544, CurrentID=0, Status='live', Symbol='MTNR1B', Aliases='FGQTL2, MEL-1B-R, MT2', description='melatonin receptor 1B', other_designations='mel1b receptor|melatonin receptor 1B variant b|melatonin receptor MEL1B', map_location='11q21-q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92969623, end_position_on_the_genomic_accession=92986240, orientation='plus', exon_count=3, OMIM=600804, no_hdr0=''),
  4547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4547, CurrentID=0, Status='live', Symbol='MTTP', Aliases='ABL, MTP', description='microsomal triglyceride transfer protein', other_designations='microsomal triglyceride transfer protein (large polypeptide, 88kDa)', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99564078, end_position_on_the_genomic_accession=99623997, orientation='plus', exon_count=19, OMIM=157147, no_hdr0=''),
  4548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4548, CurrentID=0, Status='live', Symbol='MTR', Aliases='HMAG, MS, cblG', description='5-methyltetrahydrofolate-homocysteine methyltransferase', other_designations='5-methyltetrahydrofolate-homocysteine methyltransferase 1|cobalamin-dependent methionine synthase|vitamin-B12 dependent methionine synthase', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236794304, end_position_on_the_genomic_accession=236903981, orientation='plus', exon_count=34, OMIM=156570, no_hdr0=''),
  4549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4549, CurrentID=0, Status='live', Symbol='MT-RNR1', Aliases='MTRNR1, RNR1', description='mitochondrially encoded 12S RNA', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=648, end_position_on_the_genomic_accession=1601, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4550, CurrentID=0, Status='live', Symbol='MT-RNR2', Aliases='MTRNR2, RNR2', description='mitochondrially encoded 16S RNA', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=1671, end_position_on_the_genomic_accession=3229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266695, CurrentID=0, Status='live', Symbol='PHF2P1', Aliases='DKFZp686A1627', description='PHD finger protein 2 pseudogene 1', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19045349, end_position_on_the_genomic_accession=19119662, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  4552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4552, CurrentID=0, Status='live', Symbol='MTRR', Aliases='MSR, cblE', description='5-methyltetrahydrofolate-homocysteine methyltransferase reductase', other_designations='[methionine synthase]-cobalamin methyltransferase (cob(II)alamin reducing)|methionine synthase reductase, mitochondrial', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7851186, end_position_on_the_genomic_accession=7901124, orientation='plus', exon_count=21, OMIM=602568, no_hdr0=''),
  4553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4553, CurrentID=0, Status='live', Symbol='MT-TA', Aliases='MTTA, TRNA', description='mitochondrially encoded tRNA alanine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5587, end_position_on_the_genomic_accession=5655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266698, CurrentID=0, Status='live', Symbol='ARHGAP42P3', Aliases='OPHN1P1', description='Rho GTPase activating protein 42 pseudogene 3', other_designations='oligophrenin 1 pseudogene 1', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15533785, end_position_on_the_genomic_accession=15553838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  131831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131831, CurrentID=0, Status='live', Symbol='ERICH6', Aliases='C3orf44A, FAM194A, ERICH6', description='glutamate rich 6', other_designations='family with sequence similarity 194, member A|glutamate-rich 6|glutamate-rich 6A|protein FAM194A', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150659882, end_position_on_the_genomic_accession=150704563, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  4556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4556, CurrentID=0, Status='live', Symbol='MT-TE', Aliases='MTTE, TRNE', description='mitochondrially encoded tRNA glutamic acid', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=14674, end_position_on_the_genomic_accession=14742, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266701, CurrentID=0, Status='live', Symbol='SOCS2P2', Aliases='', description='suppressor of cytokine signaling 2 pseudogene 2', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22169956, end_position_on_the_genomic_accession=22171602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4558, CurrentID=0, Status='live', Symbol='MT-TF', Aliases='TRNF', description='mitochondrially encoded tRNA phenylalanine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=577, end_position_on_the_genomic_accession=647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266703, CurrentID=0, Status='live', Symbol='ASH2LP1', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila) pseudogene 1', other_designations='', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22420377, end_position_on_the_genomic_accession=22421787, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270840, CurrentID=0, Status='live', Symbol='RPL5P23', Aliases='RPL5_8_899', description='ribosomal protein L5 pseudogene 23', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42341709, end_position_on_the_genomic_accession=42342683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4563, CurrentID=0, Status='live', Symbol='MT-TG', Aliases='MTTG, TRNG', description='mitochondrially encoded tRNA glycine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=9991, end_position_on_the_genomic_accession=10058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4564, CurrentID=0, Status='live', Symbol='MT-TH', Aliases='MTTH, TRNH', description='mitochondrially encoded tRNA histidine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=12138, end_position_on_the_genomic_accession=12206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4565, CurrentID=0, Status='live', Symbol='MT-TI', Aliases='MTTI, TRNI', description='mitochondrially encoded tRNA isoleucine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=4263, end_position_on_the_genomic_accession=4331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4566, CurrentID=0, Status='live', Symbol='MT-TK', Aliases='MTTK, TRNK', description='mitochondrially encoded tRNA lysine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=8295, end_position_on_the_genomic_accession=8364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4567, CurrentID=0, Status='live', Symbol='MT-TL1', Aliases='MTTL1, TRNL1', description='mitochondrially encoded tRNA leucine 1 (UUA/G)', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=3230, end_position_on_the_genomic_accession=3304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4568, CurrentID=0, Status='live', Symbol='MT-TL2', Aliases='MTTL2, TRNL2', description='mitochondrially encoded tRNA leucine 2 (CUN)', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=12266, end_position_on_the_genomic_accession=12336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4569, CurrentID=0, Status='live', Symbol='MT-TM', Aliases='MTTM, TRNM', description='mitochondrially encoded tRNA methionine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=4402, end_position_on_the_genomic_accession=4469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4570, CurrentID=0, Status='live', Symbol='MT-TN', Aliases='MTTN, TRNN', description='mitochondrially encoded tRNA asparagine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5657, end_position_on_the_genomic_accession=5729, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4571, CurrentID=0, Status='live', Symbol='MT-TP', Aliases='MTTP, TRNP', description='mitochondrially encoded tRNA proline', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=15956, end_position_on_the_genomic_accession=16023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4572, CurrentID=0, Status='live', Symbol='MT-TQ', Aliases='MTTQ, TRNQ', description='mitochondrially encoded tRNA glutamine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=4329, end_position_on_the_genomic_accession=4400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4573, CurrentID=0, Status='live', Symbol='MT-TR', Aliases='MTTR, TRNR', description='mitochondrially encoded tRNA arginine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=10405, end_position_on_the_genomic_accession=10469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4574, CurrentID=0, Status='live', Symbol='MT-TS1', Aliases='MTTS1, TRNS1', description='mitochondrially encoded tRNA serine 1 (UCN)', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=7446, end_position_on_the_genomic_accession=7514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4575, CurrentID=0, Status='live', Symbol='MT-TS2', Aliases='MTTS2, TRNS2', description='mitochondrially encoded tRNA serine 2 (AGU/C)', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=12207, end_position_on_the_genomic_accession=12265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4576, CurrentID=0, Status='live', Symbol='MT-TT', Aliases='MTTT, TRNT', description='mitochondrially encoded tRNA threonine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=15888, end_position_on_the_genomic_accession=15953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4577, CurrentID=0, Status='live', Symbol='MT-TV', Aliases='MTTV, TRNV', description='mitochondrially encoded tRNA valine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=1602, end_position_on_the_genomic_accession=1670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266722, CurrentID=0, Status='live', Symbol='HS6ST3', Aliases='HS6ST-3', description='heparan sulfate 6-O-sulfotransferase 3', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96090536, end_position_on_the_genomic_accession=96839562, orientation='plus', exon_count=8, OMIM=609401, no_hdr0=''),
  4579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4579, CurrentID=0, Status='live', Symbol='MT-TY', Aliases='MTTY, TRNY', description='mitochondrially encoded tRNA tyrosine', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5826, end_position_on_the_genomic_accession=5891, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4580, CurrentID=0, Status='live', Symbol='MTX1', Aliases='MTX, MTXN', description='metaxin 1', other_designations='mitochondrial outer membrane import complex protein 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155208699, end_position_on_the_genomic_accession=155213839, orientation='plus', exon_count=9, OMIM=600605, no_hdr0=''),
  4581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4581, CurrentID=0, Status='live', Symbol='MTX1P1', Aliases='MTX1P, MTXP, psMTX', description='metaxin 1 pseudogene 1', other_designations='metaxin pseudogene', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155230976, end_position_on_the_genomic_accession=155234451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4582, CurrentID=0, Status='live', Symbol='MUC1', Aliases='ADMCKD, ADMCKD1, CA 15-3, CD227, EMA, H23AG, KL-6, MAM6, MCD, MCKD, MCKD1, MUC-1, MUC-1/SEC, MUC-1/X/ZD, PEM, PEMT, PUM, MUC1', description='mucin 1, cell surface associated', other_designations='DF3 antigen|H23 antigen|Medullary cystic kidney disease, autosomal dominant|breast carcinoma-associated antigen DF3|cancer antigen 15-3|carcinoma-associated mucin|episialin|krebs von den Lungen-6|mucin 1, transmembrane|peanut-reactive urinary mucin|polymorphic epithelial mucin|tumor associated epithelial mucin|tumor-associated epithelial membrane antigen', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155185824, end_position_on_the_genomic_accession=155192915, orientation='minus', exon_count=8, OMIM=158340, no_hdr0=''),
  4583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4583, CurrentID=0, Status='live', Symbol='MUC2', Aliases='MLP, MUC-2, SMUC', description='mucin 2, oligomeric mucus/gel-forming', other_designations='intestinal mucin-2|mucin 2, intestinal/tracheal|mucin-like protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1074875, end_position_on_the_genomic_accession=1110508, orientation='plus', exon_count=51, OMIM=158370, no_hdr0=''),
  4584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4584, CurrentID=0, Status='live', Symbol='MUC3A', Aliases='MUC-3A, MUC3', description='mucin 3A, cell surface associated', other_designations='intestinal mucin|mucin 3, intestinal', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100949117, end_position_on_the_genomic_accession=100968347, orientation='plus', exon_count=13, OMIM=158371, no_hdr0=''),
  4585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4585, CurrentID=0, Status='live', Symbol='MUC4', Aliases='ASGP, HSA276359, MUC-4', description='mucin 4, cell surface associated', other_designations='ascites sialoglycoprotein|mucin 4, tracheobronchial|pancreatic adenocarcinoma mucin|testis mucin|tracheobronchial mucin', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195746765, end_position_on_the_genomic_accession=195811973, orientation='minus', exon_count=25, OMIM=158372, no_hdr0=''),
  4586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4586, CurrentID=0, Status='live', Symbol='MUC5AC', Aliases='MUC5, TBM, leB, mucin', description='mucin 5AC, oligomeric mucus/gel-forming', other_designations='gastric mucin|lewis B blood group antigen|major airway glycoprotein|mucin 5, subtypes A and C, tracheobronchial/gastric|mucin 5AC, oligomeric mucus/gel-forming pseudogene|mucin-5 subtype AC, tracheobronchial|tracheobronchial mucin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1157953, end_position_on_the_genomic_accession=1201138, orientation='plus', exon_count=49, OMIM=158373, no_hdr0=''),
  4588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4588, CurrentID=0, Status='live', Symbol='MUC6', Aliases='MUC-6', description='mucin 6, oligomeric mucus/gel-forming', other_designations='gastric mucin-6', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1012824, end_position_on_the_genomic_accession=1036706, orientation='minus', exon_count=33, OMIM=158374, no_hdr0=''),
  4589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4589, CurrentID=0, Status='live', Symbol='MUC7', Aliases='MG2', description='mucin 7, secreted', other_designations='MUC-7|apo-MG2|mucin 7, salivary|salivary mucin-7', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70430492, end_position_on_the_genomic_accession=70482997, orientation='plus', exon_count=5, OMIM=158375, no_hdr0=''),
  4591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4591, CurrentID=0, Status='live', Symbol='TRIM37', Aliases='MUL, POB1, TEF3', description='tripartite motif containing 37', other_designations='RING-B-box-coiled-coil protein|mulibrey nanism protein|tripartite motif-containing protein 37', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58982638, end_position_on_the_genomic_accession=59106905, orientation='minus', exon_count=26, OMIM=605073, no_hdr0=''),
  4593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4593, CurrentID=0, Status='live', Symbol='MUSK', Aliases='CMS9, FADS', description='muscle, skeletal, receptor tyrosine kinase', other_designations='muscle-specific kinase receptor|muscle-specific tyrosine-protein kinase receptor', map_location='9q31.3-q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110668545, end_position_on_the_genomic_accession=110806542, orientation='plus', exon_count=17, OMIM=601296, no_hdr0=''),
  4594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4594, CurrentID=0, Status='live', Symbol='MUT', Aliases='MCM', description='methylmalonyl-CoA mutase', other_designations='methylmalonyl CoA mutase|methylmalonyl Coenzyme A mutase|methylmalonyl-CoA isomerase|mutant methylmalonyl CoA mutase|truncated methylmalonyl CoA mutase', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49430360, end_position_on_the_genomic_accession=49463328, orientation='minus', exon_count=13, OMIM=609058, no_hdr0=''),
  4595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4595, CurrentID=0, Status='live', Symbol='MUTYH', Aliases='MYH', description='mutY DNA glycosylase', other_designations='mutY homolog|mutY-like protein', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45329242, end_position_on_the_genomic_accession=45340470, orientation='minus', exon_count=22, OMIM=604933, no_hdr0=''),
  266740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266740, CurrentID=0, Status='live', Symbol='MAGEA2B', Aliases='MAGE2, MAGEA2', description='MAGE family member A2B', other_designations='CT1.2|MAGE-2 antigen|cancer/testis antigen 1.2|melanoma antigen 2|melanoma antigen family A, 2B|melanoma antigen family A2B|melanoma antigen, family A, 2, copy b', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152714502, end_position_on_the_genomic_accession=152718607, orientation='plus', exon_count=7, OMIM=300549, no_hdr0=''),
  4597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4597, CurrentID=0, Status='live', Symbol='MVD', Aliases='FP17780, MDDase, MPD', description='mevalonate diphosphate decarboxylase', other_designations='mevalonate (diphospho) decarboxylase|mevalonate pyrophosphate decarboxylase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88651935, end_position_on_the_genomic_accession=88663157, orientation='minus', exon_count=14, OMIM=603236, no_hdr0=''),
  4598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4598, CurrentID=0, Status='live', Symbol='MVK', Aliases='LRBP, MK, MVLK, POROK3', description='mevalonate kinase', other_designations='LH receptor mRNA-binding protein|mevalonate kinase 1', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109573353, end_position_on_the_genomic_accession=109597270, orientation='plus', exon_count=12, OMIM=251170, no_hdr0=''),
  4599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4599, CurrentID=0, Status='live', Symbol='MX1', Aliases='IFI-78K, IFI78, MX, MxA', description='MX dynamin-like GTPase 1', other_designations='interferon-induced protein p78|interferon-inducible protein p78|interferon-regulated resistance GTP-binding protein MxA|myxoma resistance protein 1|myxovirus (influenza virus) resistance 1, interferon-inducible protein p78', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41420329, end_position_on_the_genomic_accession=41459214, orientation='plus', exon_count=24, OMIM=147150, no_hdr0=''),
  4600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4600, CurrentID=0, Status='live', Symbol='MX2', Aliases='MXB', description='MX dynamin-like GTPase 2', other_designations='interferon-regulated resistance GTP-binding protein MXB|myxovirus (influenza virus) resistance 2|myxovirus resistance protein 2|p78-related protein|second interferon-induced protein p78', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41361986, end_position_on_the_genomic_accession=41408943, orientation='plus', exon_count=17, OMIM=147890, no_hdr0=''),
  4601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4601, CurrentID=0, Status='live', Symbol='MXI1', Aliases='MAD2, MXD2, MXI, bHLHc11', description='MAX interactor 1, dimerization protein', other_designations='MAX dimerization protein 2|Max-related transcription factor|class C basic helix-loop-helix protein 11', map_location='10q24-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110207605, end_position_on_the_genomic_accession=110287365, orientation='plus', exon_count=8, OMIM=600020, no_hdr0=''),
  4602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4602, CurrentID=0, Status='live', Symbol='MYB', Aliases='Cmyb, c-myb, c-myb_CDS, efg', description='v-myb avian myeloblastosis viral oncogene homolog', other_designations='c-myb protein (140 AA)|c-myb10A_CDS|c-myb13A_CDS|c-myb14A_CDS|c-myb8B_CDS|proto-oncogene c-Myb', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135180981, end_position_on_the_genomic_accession=135219173, orientation='plus', exon_count=18, OMIM=189990, no_hdr0=''),
  4603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4603, CurrentID=0, Status='live', Symbol='MYBL1', Aliases='A-MYB, AMYB', description='v-myb avian myeloblastosis viral oncogene homolog-like 1', other_designations='myb-like protein 1', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66562175, end_position_on_the_genomic_accession=66613249, orientation='minus', exon_count=17, OMIM=159405, no_hdr0=''),
  4604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4604, CurrentID=0, Status='live', Symbol='MYBPC1', Aliases='LCCS4, MYBPCC, MYBPCS', description='myosin binding protein C, slow type', other_designations='C-protein, skeletal muscle slow isoform|skeletal muscle C-protein|slow MyBP-C', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101594931, end_position_on_the_genomic_accession=101685883, orientation='plus', exon_count=33, OMIM=160794, no_hdr0=''),
  4605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4605, CurrentID=0, Status='live', Symbol='MYBL2', Aliases='B-MYB, BMYB', description='v-myb avian myeloblastosis viral oncogene homolog-like 2', other_designations='myb-like protein 2|v-myb myeloblastosis viral oncogene homolog-like 2', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43667019, end_position_on_the_genomic_accession=43716496, orientation='plus', exon_count=14, OMIM=601415, no_hdr0=''),
  4606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4606, CurrentID=0, Status='live', Symbol='MYBPC2', Aliases='MYBPC, MYBPCF', description='myosin binding protein C, fast type', other_designations='C-protein, skeletal muscle fast isoform|fast MyBP-C|fast-type muscle myosin-binding-protein C', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50432903, end_position_on_the_genomic_accession=50466326, orientation='plus', exon_count=28, OMIM=160793, no_hdr0=''),
  4607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4607, CurrentID=0, Status='live', Symbol='MYBPC3', Aliases='CMD1MM, CMH4, FHC, LVNC10, MYBP-C', description='myosin binding protein C, cardiac', other_designations='C-protein, cardiac muscle isoform', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47331406, end_position_on_the_genomic_accession=47352702, orientation='minus', exon_count=35, OMIM=600958, no_hdr0=''),
  4608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4608, CurrentID=0, Status='live', Symbol='MYBPH', Aliases='', description='myosin binding protein H', other_designations='H-protein|myBP-H', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203167811, end_position_on_the_genomic_accession=203175966, orientation='minus', exon_count=12, OMIM=160795, no_hdr0=''),
  4609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4609, CurrentID=0, Status='live', Symbol='MYC', Aliases='MRTLC, bHLHe39, c-Myc, MYC', description='v-myc avian myelocytomatosis viral oncogene homolog', other_designations='avian myelocytomatosis viral oncogene homolog|class E basic helix-loop-helix protein 39|myc-related translation/localization regulatory factor|proto-oncogene c-Myc|transcription factor p64|v-myc myelocytomatosis viral oncogene homolog', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127736069, end_position_on_the_genomic_accession=127741434, orientation='plus', exon_count=3, OMIM=190080, no_hdr0=''),
  4610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4610, CurrentID=0, Status='live', Symbol='MYCL', Aliases='L-Myc, LMYC1, bHLHe38, MYCL', description='v-myc avian myelocytomatosis viral oncogene lung carcinoma derived homolog', other_designations='class E basic helix-loop-helix protein 38|l-myc-1 proto-oncogene|myc-related gene from lung cancer|protein L-Myc-1|v-myc myelocytomatosis viral oncogene homolog 1, lung carcinoma derived', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39895424, end_position_on_the_genomic_accession=39902015, orientation='minus', exon_count=2, OMIM=164850, no_hdr0=''),
  4611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4611, CurrentID=0, Status='live', Symbol='MYCLP1', Aliases='MYCL1P1, MYCL2, bHLHe38', description='MYCL pseudogene 1', other_designations='MYCL-related processed|v-myc avian myelocytomatosis viral oncogene homolog 1 pseudogene 1|v-myc avian myelocytomatosis viral oncogene homolog 2|v-myc myelocytomatosis viral oncogene homolog 1 pseudogene 1|v-myc myelocytomatosis viral oncogene homolog 2', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107272464, end_position_on_the_genomic_accession=107276559, orientation='plus', exon_count=0, OMIM=310310, no_hdr0=''),
  4612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4612, CurrentID=0, Status='live', Symbol='MYCLK1', Aliases='', description='v-myc avian myelocytomatosis viral oncogene homolog-like 1', other_designations='Avian myelocytomatosis viral (v-myc) oncogene homolog like 1|v-myc myelocytomatosis viral oncogene homolog-like 1', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164865, no_hdr0=''),
  4613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4613, CurrentID=0, Status='live', Symbol='MYCN', Aliases='MODED, N-myc, NMYC, ODED, bHLHe37', description='v-myc avian myelocytomatosis viral oncogene neuroblastoma derived homolog', other_designations='class E basic helix-loop-helix protein 37|neuroblastoma MYC oncogene|neuroblastoma-derived v-myc avian myelocytomatosis viral related oncogene|oncogene NMYC|pp65/67', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15940438, end_position_on_the_genomic_accession=15947007, orientation='plus', exon_count=3, OMIM=164840, no_hdr0=''),
  4615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4615, CurrentID=0, Status='live', Symbol='MYD88', Aliases='MYD88D', description='myeloid differentiation primary response 88', other_designations='myeloid differentiation primary response gene (88)', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38138478, end_position_on_the_genomic_accession=38143022, orientation='plus', exon_count=5, OMIM=602170, no_hdr0=''),
  4616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4616, CurrentID=0, Status='live', Symbol='GADD45B', Aliases='GADD45BETA, MYD118', description='growth arrest and DNA damage inducible beta', other_designations='growth arrest and DNA-damage-inducible, beta|myeloid differentiation primary response protein MyD118|negative growth regulatory protein MyD118', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2476125, end_position_on_the_genomic_accession=2478259, orientation='plus', exon_count=4, OMIM=604948, no_hdr0=''),
  4617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4617, CurrentID=0, Status='live', Symbol='MYF5', Aliases='bHLHc2', description='myogenic factor 5', other_designations='class C basic helix-loop-helix protein 2|myf-5', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80716929, end_position_on_the_genomic_accession=80719668, orientation='plus', exon_count=3, OMIM=159990, no_hdr0=''),
  4618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4618, CurrentID=0, Status='live', Symbol='MYF6', Aliases='CNM3, MRF4, bHLHc4, myf-6', description='myogenic factor 6', other_designations='class C basic helix-loop-helix protein 4|muscle-specific regulatory factor 4|myogenic factor 6 (herculin)', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80707629, end_position_on_the_genomic_accession=80709478, orientation='plus', exon_count=3, OMIM=159991, no_hdr0=''),
  4619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4619, CurrentID=0, Status='live', Symbol='MYH1', Aliases='HEL71, MYHSA1, MYHa, MyHC-2X/D, MyHC-2x', description='myosin, heavy chain 1, skeletal muscle, adult', other_designations='epididymis luminal protein 71|myHC-IIx/d|myosin heavy chain 1|myosin heavy chain 2x|myosin heavy chain IIx/d|myosin heavy chain, skeletal muscle, adult 1|myosin, heavy polypeptide 1, skeletal muscle, adult', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10492307, end_position_on_the_genomic_accession=10518542, orientation='minus', exon_count=40, OMIM=160730, no_hdr0=''),
  4620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4620, CurrentID=0, Status='live', Symbol='MYH2', Aliases='IBM3A, MYHSA2, MYHas8, MYPOP, MyHC-2A, MyHC-IIa, MYH2', description='myosin, heavy chain 2, skeletal muscle, adult', other_designations='fast 2a myosin heavy chain|inclusion body myopathy 3, autosomal dominant|myosin heavy chain 2|myosin heavy chain 2a|myosin heavy chain IIa|myosin heavy chain, skeletal muscle, adult 2|myosin, heavy polypeptide 2, skeletal muscle, adult|type IIA myosin heavy chain', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10521148, end_position_on_the_genomic_accession=10549700, orientation='minus', exon_count=41, OMIM=160740, no_hdr0=''),
  4621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4621, CurrentID=0, Status='live', Symbol='MYH3', Aliases='HEMHC, MYHC-EMB, MYHSE1, SMHCE', description='myosin, heavy chain 3, skeletal muscle, embryonic', other_designations='myosin heavy chain, fast skeletal muscle, embryonic|myosin, heavy polypeptide 3, skeletal muscle, embryonic|myosin, skeletal, heavy chain, embryonic 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10628526, end_position_on_the_genomic_accession=10679030, orientation='minus', exon_count=44, OMIM=160720, no_hdr0=''),
  4622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4622, CurrentID=0, Status='live', Symbol='MYH4', Aliases='MYH2B, MyHC-2B, MyHC-IIb', description='myosin, heavy chain 4, skeletal muscle', other_designations='myosin heavy chain 2b|myosin heavy chain 4|myosin heavy chain IIb|myosin heavy chain, skeletal muscle, fetal|myosin, heavy polypeptide 4, skeletal muscle', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10443290, end_position_on_the_genomic_accession=10469559, orientation='minus', exon_count=40, OMIM=160742, no_hdr0=''),
  4624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4624, CurrentID=0, Status='live', Symbol='MYH6', Aliases='ASD3, CMD1EE, CMH14, MYHC, MYHCA, SSS3, alpha-MHC', description='myosin, heavy chain 6, cardiac muscle, alpha', other_designations='myHC-alpha|myosin heavy chain 6|myosin heavy chain, cardiac muscle alpha isoform|myosin, heavy polypeptide 6, cardiac muscle, alpha (cardiomyopathy, hypertrophic 1)', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23381990, end_position_on_the_genomic_accession=23408277, orientation='minus', exon_count=39, OMIM=160710, no_hdr0=''),
  4625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4625, CurrentID=0, Status='live', Symbol='MYH7', Aliases='CMD1S, CMH1, MPD1, MYHCB, SPMD, SPMM', description='myosin, heavy chain 7, cardiac muscle, beta', other_designations='myHC-beta|myhc-slow|myopathy, distal 1|myosin heavy chain slow isoform|myosin heavy chain, cardiac muscle beta isoform|myosin, heavy polypeptide 7, cardiac muscle, beta|rhabdomyosarcoma antigen MU-RMS-40.7A', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23412738, end_position_on_the_genomic_accession=23435686, orientation='minus', exon_count=41, OMIM=160760, no_hdr0=''),
  4626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4626, CurrentID=0, Status='live', Symbol='MYH8', Aliases='DA7, MyHC-peri, MyHC-pn, gtMHC-F', description='myosin, heavy chain 8, skeletal muscle, perinatal', other_designations='fetal-myosin heavy chain|myHC-perinatal|myosin heavy chain 8|myosin heavy chain, skeletal muscle, perinatal|myosin, heavy polypeptide 8, skeletal muscle, perinatal', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10390325, end_position_on_the_genomic_accession=10421950, orientation='minus', exon_count=40, OMIM=160741, no_hdr0=''),
  4627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4627, CurrentID=0, Status='live', Symbol='MYH9', Aliases='BDPLT6, DFNA17, EPSTS, FTNS, MHA, NMHC-II-A, NMMHC-IIA, NMMHCA', description='myosin, heavy chain 9, non-muscle', other_designations='cellular myosin heavy chain, type A|non-muscle myosin heavy chain A|non-muscle myosin heavy chain IIa|non-muscle myosin heavy polypeptide 9|nonmuscle myosin heavy chain II-A', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36281277, end_position_on_the_genomic_accession=36388067, orientation='minus', exon_count=43, OMIM=160775, no_hdr0=''),
  4628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4628, CurrentID=0, Status='live', Symbol='MYH10', Aliases='NMMHC-IIB, NMMHCB', description='myosin, heavy chain 10, non-muscle', other_designations='cellular myosin heavy chain, type B|myosin heavy chain, nonmuscle type B|myosin, heavy polypeptide 10, non-muscle|nonmuscle myosin II heavy chain-B|nonmuscle myosin heavy chain IIB', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8474205, end_position_on_the_genomic_accession=8631012, orientation='minus', exon_count=45, OMIM=160776, no_hdr0=''),
  4629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4629, CurrentID=0, Status='live', Symbol='MYH11', Aliases='AAT4, FAA4, SMHC, SMMHC', description='myosin, heavy chain 11, smooth muscle', other_designations='myosin heavy chain 11|myosin heavy chain, smooth muscle isoform|myosin, heavy polypeptide 11, smooth muscle', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15703135, end_position_on_the_genomic_accession=15857033, orientation='minus', exon_count=43, OMIM=160745, no_hdr0=''),
  350297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=350297, CurrentID=0, Status='live', Symbol='MRPS21P8', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 8', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49764453, end_position_on_the_genomic_accession=49764712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4632, CurrentID=0, Status='live', Symbol='MYL1', Aliases='MLC1F, MLC3F', description='myosin light chain 1', other_designations='A1 catalytic|A2 catalytic|MLC1/MLC3|MLC1F/MLC3F|myosin light chain A1/A2|myosin light chain alkali 1/2|myosin, light chain 1, alkali; skeletal, fast|myosin, light polypeptide 1, alkali; skeletal, fast', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210290144, end_position_on_the_genomic_accession=210315171, orientation='minus', exon_count=9, OMIM=160780, no_hdr0=''),
  4633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4633, CurrentID=0, Status='live', Symbol='MYL2', Aliases='CMH10, MLC-2s/v, MLC2', description='myosin light chain 2', other_designations='MLC-2|MLC-2v|RLC of myosin|cardiac myosin light chain 2|cardiac ventricular myosin light chain 2|myosin light chain 2, slow skeletal/ventricular muscle isoform|myosin, light chain 2, regulatory, cardiac, slow|myosin, light polypeptide 2, regulatory, cardiac, slow|regulatory light chain of myosin|slow cardiac myosin regulatory light chain 2|ventricular myosin light chain 2', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110910819, end_position_on_the_genomic_accession=110920600, orientation='minus', exon_count=7, OMIM=160781, no_hdr0=''),
  4634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4634, CurrentID=0, Status='live', Symbol='MYL3', Aliases='CMH8, MLC-lV/sb, MLC1SB, MLC1V, VLC1, VLCl', description='myosin light chain 3', other_designations='CMLC1|cardiac myosin light chain 1|myosin light chain 1, slow-twitch muscle B/ventricular isoform|myosin, light chain 3, alkali; ventricular, skeletal, slow|myosin, light polypeptide 3, alkali; ventricular, skeletal, slow|ventricular myosin alkali light chain|ventricular myosin light chain 1|ventricular/slow twitch myosin alkali light chain', map_location='3p21.3-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46857867, end_position_on_the_genomic_accession=46863483, orientation='minus', exon_count=7, OMIM=160790, no_hdr0=''),
  4635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4635, CurrentID=0, Status='live', Symbol='MYL4', Aliases='ALC1, AMLC, GT1, PRO1957', description='myosin light chain 4', other_designations='myosin light chain 1, embryonic muscle/atrial isoform|myosin light chain alkali GT-1 isoform|myosin, atrial/fetal muscle, light chain|myosin, light chain 4, alkali; atrial, embryonic|myosin, light polypeptide 4, alkali; atrial, embryonic', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47209013, end_position_on_the_genomic_accession=47227697, orientation='plus', exon_count=10, OMIM=160770, no_hdr0=''),
  4636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4636, CurrentID=0, Status='live', Symbol='MYL5', Aliases='', description='myosin light chain 5', other_designations='MYLC2|myLC-2|myosin regulatory light chain 5|myosin, light chain 5, regulatory|myosin, light polypeptide 5, regulatory|superfast myosin regulatory light chain 2', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=674212, end_position_on_the_genomic_accession=682028, orientation='plus', exon_count=8, OMIM=160782, no_hdr0=''),
  4637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4637, CurrentID=0, Status='live', Symbol='MYL6', Aliases='ESMLC, LC17, LC17-GI, LC17-NM, LC17A, LC17B, MLC-3, MLC1SM, MLC3NM, MLC3SM', description='myosin light chain 6', other_designations='17 kDa myosin light chain|myosin light chain A3|myosin light chain alkali 3|myosin, light chain 6, alkali, smooth muscle and non-muscle|myosin, light polypeptide 6, alkali, smooth muscle and non-muscle', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56158261, end_position_on_the_genomic_accession=56161582, orientation='plus', exon_count=7, OMIM=609931, no_hdr0=''),
  4638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4638, CurrentID=0, Status='live', Symbol='MYLK', Aliases='AAT7, KRP, MLCK, MLCK1, MLCK108, MLCK210, MSTP0831, smMLCK, MYLK', description='myosin light chain kinase', other_designations='kinase-related protein|myosin, light polypeptide kinase|smooth muscle myosin light chain kinase|telokin', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123612296, end_position_on_the_genomic_accession=123884302, orientation='minus', exon_count=34, OMIM=600922, no_hdr0=''),
  100270853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270853, CurrentID=0, Status='live', Symbol='RPL7P19', Aliases='RPL7_9_617', description='ribosomal protein L7 pseudogene 19', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138473744, end_position_on_the_genomic_accession=138474417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4640, CurrentID=0, Status='live', Symbol='MYO1A', Aliases='BBMI, DFNA48, MIHC, MYHL', description='myosin IA', other_designations='brush border myosin I|myosin I heavy chain|myosin, heavy polypeptide-like (100kD)', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57028517, end_position_on_the_genomic_accession=57050765, orientation='minus', exon_count=30, OMIM=601478, no_hdr0=''),
  4641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4641, CurrentID=0, Status='live', Symbol='MYO1C', Aliases='MMI-beta, MMIb, NMI, myr2', description='myosin IC', other_designations='myosin I beta|myosin-I beta|myosin-Ic|nuclear myosin I', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1464186, end_position_on_the_genomic_accession=1492707, orientation='minus', exon_count=34, OMIM=606538, no_hdr0=''),
  4642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4642, CurrentID=0, Status='live', Symbol='MYO1D', Aliases='PPP1R108, myr4', description='myosin ID', other_designations='myosin-I gamma|protein phosphatase 1, regulatory subunit 108', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32492522, end_position_on_the_genomic_accession=32877173, orientation='minus', exon_count=25, OMIM=606539, no_hdr0=''),
  4643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4643, CurrentID=0, Status='live', Symbol='MYO1E', Aliases='FSGS6, HuncM-IC, MYO1C', description='myosin IE', other_designations='MYO1E variant protein|myosin-IC|unconventional myosin 1E', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59135969, end_position_on_the_genomic_accession=59372872, orientation='minus', exon_count=28, OMIM=601479, no_hdr0=''),
  4644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4644, CurrentID=0, Status='live', Symbol='MYO5A', Aliases='GS1, MYH12, MYO5, MYR12', description='myosin VA', other_designations='dilute myosin heavy chain, non-muscle|myosin V|myosin VA (heavy chain 12, myoxin)|myosin, heavy polypeptide kinase|myosin-12|myosin-Va|myoxin', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52307283, end_position_on_the_genomic_accession=52529050, orientation='minus', exon_count=44, OMIM=160777, no_hdr0=''),
  4645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4645, CurrentID=0, Status='live', Symbol='MYO5B', Aliases='', description='myosin VB', other_designations='MYO5B variant protein|myosin-Vb', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49822786, end_position_on_the_genomic_accession=50195081, orientation='minus', exon_count=40, OMIM=606540, no_hdr0=''),
  4646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4646, CurrentID=0, Status='live', Symbol='MYO6', Aliases='DFNA22, DFNB37', description='myosin VI', other_designations='myosin-VI|unconventional myosin-6', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75749176, end_position_on_the_genomic_accession=75919537, orientation='plus', exon_count=36, OMIM=600970, no_hdr0=''),
  4647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4647, CurrentID=0, Status='live', Symbol='MYO7A', Aliases='DFNA11, DFNB2, MYOVIIA, MYU7A, NSRD2, USH1B', description='myosin VIIA', other_designations='myosin VIIA (Usher syndrome 1B (autosomal recessive, severe))', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77128214, end_position_on_the_genomic_accession=77215241, orientation='plus', exon_count=55, OMIM=276903, no_hdr0=''),
  4648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4648, CurrentID=0, Status='live', Symbol='MYO7B', Aliases='', description='myosin VIIB', other_designations='myosin-VIIb', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127535693, end_position_on_the_genomic_accession=127637728, orientation='plus', exon_count=48, OMIM=606541, no_hdr0=''),
  4649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4649, CurrentID=0, Status='live', Symbol='MYO9A', Aliases='', description='myosin IXA', other_designations='myosin-IXa|unconventional myosin-9a', map_location='15q22-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71822291, end_position_on_the_genomic_accession=72118464, orientation='minus', exon_count=48, OMIM=604875, no_hdr0=''),
  4650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4650, CurrentID=0, Status='live', Symbol='MYO9B', Aliases='CELIAC4, MYR5', description='myosin IXB', other_designations='myosin-IXb|unconventional myosin IXb|unconventional myosin-9b', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17075779, end_position_on_the_genomic_accession=17215277, orientation='plus', exon_count=43, OMIM=602129, no_hdr0=''),
  4651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4651, CurrentID=0, Status='live', Symbol='MYO10', Aliases='', description='myosin X', other_designations='myosin-X|unconventional myosin-10|unconventionnal myosin-X', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16661907, end_position_on_the_genomic_accession=16936276, orientation='minus', exon_count=44, OMIM=601481, no_hdr0=''),
  4653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4653, CurrentID=0, Status='live', Symbol='MYOC', Aliases='GLC1A, GPOA, JOAG, JOAG1, TIGR, myocilin', description='myocilin, trabecular meshwork inducible glucocorticoid response', other_designations='mutated trabecular meshwork-induced glucocorticoid response protein|myocilin 55 kDa subunit|myocilin trabecular meshwork inducible glucocorticoid response', map_location='1q23-q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171635417, end_position_on_the_genomic_accession=171652633, orientation='minus', exon_count=3, OMIM=601652, no_hdr0=''),
  4654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4654, CurrentID=0, Status='live', Symbol='MYOD1', Aliases='MYF3, MYOD, PUM, bHLHc1', description='myogenic differentiation 1', other_designations='class C basic helix-loop-helix protein 1|myf-3|myogenic factor 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17719563, end_position_on_the_genomic_accession=17722131, orientation='plus', exon_count=3, OMIM=159970, no_hdr0=''),
  791087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791087, CurrentID=0, Status='live', Symbol='HHT4', Aliases='', description='Telangiectasia, hereditary hemorrhagic, type 4', other_designations='', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610655, no_hdr0=''),
  4656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4656, CurrentID=0, Status='live', Symbol='MYOG', Aliases='MYF4, bHLHc3, myf-4', description='myogenin (myogenic factor 4)', other_designations='class C basic helix-loop-helix protein 3|myogenic factor 4', map_location='1q31-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203083129, end_position_on_the_genomic_accession=203086038, orientation='minus', exon_count=3, OMIM=159980, no_hdr0=''),
  4657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4657, CurrentID=0, Status='live', Symbol='MYP1', Aliases='BED', description='myopia 1 (X-linked)', other_designations='Myopia, X-linked (Bornholm eye disease)|Myopia-1 (Bornholm eye disease)|myopia 1 (X-linked, Bornholm eye disease included)', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=310460, no_hdr0=''),
  4658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4658, CurrentID=0, Status='live', Symbol='MYP2', Aliases='', description='myopia 2 (high grade, autosomal dominant)', other_designations='Myopia, high grade, autosomal dominant 1|pathologic myopia', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=160700, no_hdr0=''),
  4659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4659, CurrentID=0, Status='live', Symbol='PPP1R12A', Aliases='M130, MBS, MYPT1', description='protein phosphatase 1 regulatory subunit 12A', other_designations='myosin binding subunit|myosin phosphatase, target subunit 1|myosin phosphatase-targeting subunit 1|protein phosphatase 1, regulatory (inhibitor) subunit 12A|protein phosphatase 1, regulatory subunit 12A|protein phosphatase myosin-binding subunit', map_location='12q15-q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79773563, end_position_on_the_genomic_accession=79935455, orientation='minus', exon_count=31, OMIM=602021, no_hdr0=''),
  4660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4660, CurrentID=0, Status='live', Symbol='PPP1R12B', Aliases='MYPT2, PP1bp55', description='protein phosphatase 1 regulatory subunit 12B', other_designations='myosin phosphatase target subunit 2|myosin phosphatase, target subunit 2|myosin phosphatase-targeting subunit 2|protein phosphatase 1, regulatory (inhibitor) subunit 12B|protein phosphatase 1, regulatory subunit 12B', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202348702, end_position_on_the_genomic_accession=202588569, orientation='plus', exon_count=35, OMIM=603768, no_hdr0=''),
  4661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4661, CurrentID=0, Status='live', Symbol='MYT1', Aliases='C20orf36, MTF1, MYTI, NZF2, PLPB1, ZC2HC4A', description='myelin transcription factor 1', other_designations='myelin transcription factor I|neural zinc finger transcription factor 2|proteolipid protein binding protein|proteolipid protein-binding protein', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64164474, end_position_on_the_genomic_accession=64242253, orientation='plus', exon_count=23, OMIM=600379, no_hdr0=''),
  100270857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270857, CurrentID=0, Status='live', Symbol='RPL9P15', Aliases='RPL9_4_422', description='ribosomal protein L9 pseudogene 15', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154676588, end_position_on_the_genomic_accession=154677132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4664, CurrentID=0, Status='live', Symbol='NAB1', Aliases='', description='NGFI-A binding protein 1', other_designations='EGR-1-binding protein 1|EGR1 binding protein 1|NGFI-A binding protein 1 (EGR1 binding protein 1)|transcriptional regulatory protein p54', map_location='2q32.3-q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190648895, end_position_on_the_genomic_accession=190692766, orientation='plus', exon_count=12, OMIM=600800, no_hdr0=''),
  4665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4665, CurrentID=0, Status='live', Symbol='NAB2', Aliases='MADER', description='NGFI-A binding protein 2 (EGR1 binding protein 2)', other_designations='EGR-1-binding protein 2|EGR1 binding protein 2|melanoma-associated delayed early response protein', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57088894, end_position_on_the_genomic_accession=57095476, orientation='plus', exon_count=7, OMIM=602381, no_hdr0=''),
  4666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4666, CurrentID=0, Status='live', Symbol='NACA', Aliases='HSD481, skNAC, NACA', description='nascent polypeptide-associated complex alpha subunit', other_designations='NAC-alpha|alpha-NAC, muscle-specific form|nascent-polypeptide-associated complex alpha polypeptide', map_location='12q23-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56712427, end_position_on_the_genomic_accession=56726275, orientation='minus', exon_count=9, OMIM=601234, no_hdr0=''),
  266811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266811, CurrentID=0, Status='live', Symbol='NPSA', Aliases='', description='novel prostate-specific antigen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4668, CurrentID=0, Status='live', Symbol='NAGA', Aliases='D22S674, GALB', description='N-acetylgalactosaminidase, alpha-', other_designations='Acetylgalactosaminidase, alpha-N- (alpha-galactosidase B)|alpha-galactosidase B', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42058334, end_position_on_the_genomic_accession=42070871, orientation='minus', exon_count=9, OMIM=104170, no_hdr0=''),
  4669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4669, CurrentID=0, Status='live', Symbol='NAGLU', Aliases='MPS-IIIB, MPS3B, NAG, UFHSD', description='N-acetylglucosaminidase, alpha', other_designations='N-acetyl-alpha-glucosaminidase', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42535933, end_position_on_the_genomic_accession=42544449, orientation='plus', exon_count=6, OMIM=609701, no_hdr0=''),
  4670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4670, CurrentID=0, Status='live', Symbol='HNRNPM', Aliases='CEAR4, HNRPM, HNRPM4, HTGR1, NAGR1, hnRNP M, HNRNPM', description='heterogeneous nuclear ribonucleoprotein M', other_designations='CEA receptor|N-acetylglucosamine receptor 1|heterogenous nuclear ribonucleoprotein M4|hnRNA-binding protein M4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8444575, end_position_on_the_genomic_accession=8489118, orientation='plus', exon_count=16, OMIM=160994, no_hdr0=''),
  4671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4671, CurrentID=0, Status='live', Symbol='NAIP', Aliases='BIRC1, NLRB1, psiNAIP', description='NLR family, apoptosis inhibitory protein', other_designations='neuronal apoptosis inhibitory protein|nucleotide-binding oligomerization domain, leucine rich repeat and BIR domain containing 1|psi neuronal apoptosis inhibitory protein', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70968483, end_position_on_the_genomic_accession=71025114, orientation='minus', exon_count=20, OMIM=600355, no_hdr0=''),
  4673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4673, CurrentID=0, Status='live', Symbol='NAP1L1', Aliases='NAP1, NAP1L, NRP', description='nucleosome assembly protein 1-like 1', other_designations='HSP22-like protein interacting protein|NAP-1-related protein', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76044892, end_position_on_the_genomic_accession=76084958, orientation='minus', exon_count=19, OMIM=164060, no_hdr0=''),
  4674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4674, CurrentID=0, Status='live', Symbol='NAP1L2', Aliases='BPX', description='nucleosome assembly protein 1-like 2', other_designations='brain specific gene BPX|brain-specific protein, X-linked', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73212299, end_position_on_the_genomic_accession=73214874, orientation='minus', exon_count=1, OMIM=300026, no_hdr0=''),
  4675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4675, CurrentID=0, Status='live', Symbol='NAP1L3', Aliases='MB20, NPL3', description='nucleosome assembly protein 1-like 3', other_designations='', map_location='Xq21.3-q22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93670926, end_position_on_the_genomic_accession=93673683, orientation='minus', exon_count=1, OMIM=300117, no_hdr0=''),
  4676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4676, CurrentID=0, Status='live', Symbol='NAP1L4', Aliases='NAP1L4b, NAP2, NAP2L, hNAP2', description='nucleosome assembly protein 1-like 4', other_designations='NAP-2|nucleosome assembly protein 1-like 4b|nucleosome assembly protein 2', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2944430, end_position_on_the_genomic_accession=2992377, orientation='minus', exon_count=16, OMIM=601651, no_hdr0=''),
  4677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4677, CurrentID=0, Status='live', Symbol='NARS', Aliases='ASNRS1, NARS', description='asparaginyl-tRNA synthetase', other_designations='asparagine tRNA ligase 1, cytoplasmic|asparaginyl-tRNA synthetase, cytoplasmic', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57600656, end_position_on_the_genomic_accession=57621945, orientation='minus', exon_count=14, OMIM=108410, no_hdr0=''),
  4678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4678, CurrentID=0, Status='live', Symbol='NASP', Aliases='FLB7527, HMDRA1, PRO1999', description='nuclear autoantigenic sperm protein (histone-binding)', other_designations='histone H1-binding protein', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45583988, end_position_on_the_genomic_accession=45618906, orientation='plus', exon_count=16, OMIM=603185, no_hdr0=''),
  4680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4680, CurrentID=0, Status='live', Symbol='CEACAM6', Aliases='CD66c, CEAL, NCA', description='carcinoembryonic antigen-related cell adhesion molecule 6 (non-specific cross reacting antigen)', other_designations='Cluster of Differentiation 66c|normal cross-reacting antigen', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41755421, end_position_on_the_genomic_accession=41772211, orientation='plus', exon_count=6, OMIM=163980, no_hdr0=''),
  4681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4681, CurrentID=0, Status='live', Symbol='NBL1', Aliases='D1S1733E, DAN, DAND1, NB, NO3', description='neuroblastoma 1, DAN family BMP antagonist', other_designations='DAN domain family member 1|differential screening-selected gene aberrant in neuroblastoma|neuroblastoma candidate region, suppression of tumorigenicity 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19643229, end_position_on_the_genomic_accession=19658456, orientation='plus', exon_count=10, OMIM=600613, no_hdr0=''),
  791114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791114, CurrentID=0, Status='live', Symbol='PWRN1', Aliases='NCRNA00198', description='Prader-Willi region non-protein coding RNA 1', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24558157, end_position_on_the_genomic_accession=24587780, orientation='plus', exon_count=9, OMIM=611215, no_hdr0=''),
  4683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4683, CurrentID=0, Status='live', Symbol='NBN', Aliases='AT-V1, AT-V2, ATV, NBS, NBS1, P95', description='nibrin', other_designations='Nijmegen breakage syndrome 1 (nibrin)|cell cycle regulatory protein p95|p95 protein of the MRE11/RAD50 complex', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89933336, end_position_on_the_genomic_accession=89984733, orientation='minus', exon_count=19, OMIM=602667, no_hdr0=''),
  4684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4684, CurrentID=0, Status='live', Symbol='NCAM1', Aliases='CD56, MSK39, NCAM', description='neural cell adhesion molecule 1', other_designations='antigen recognized by monoclonal antibody 5.1H11|neural cell adhesion molecule, NCAM', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112961247, end_position_on_the_genomic_accession=113278436, orientation='plus', exon_count=25, OMIM=116930, no_hdr0=''),
  4685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4685, CurrentID=0, Status='live', Symbol='NCAM2', Aliases='NCAM21', description='neural cell adhesion molecule 2', other_designations='N-CAM-2|NCAM-2', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20998315, end_position_on_the_genomic_accession=21540630, orientation='plus', exon_count=26, OMIM=602040, no_hdr0=''),
  4686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4686, CurrentID=0, Status='live', Symbol='NCBP1', Aliases='CBP80, NCBP, Sto1', description='nuclear cap binding protein subunit 1', other_designations='80 kDa nuclear cap-binding protein|NCBP 80 kDa subunit|nuclear cap binding protein subunit 1, 80kD|nuclear cap binding protein subunit 1, 80kDa', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97633423, end_position_on_the_genomic_accession=97673748, orientation='plus', exon_count=26, OMIM=600469, no_hdr0=''),
  100270861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270861, CurrentID=0, Status='live', Symbol='RPS2P22', Aliases='RPS2_9_556', description='ribosomal protein S2 pseudogene 22', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41894939, end_position_on_the_genomic_accession=41895668, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4688, CurrentID=0, Status='live', Symbol='NCF2', Aliases='NCF-2, NOXA2, P67-PHOX, P67PHOX', description='neutrophil cytosolic factor 2', other_designations='67 kDa neutrophil oxidase factor|NADPH oxidase activator 2|chronic granulomatous disease, autosomal 2|neutrophil NADPH oxidase factor 2|neutrophil cytosolic factor 2 (65kD, chronic granulomatous disease, autosomal 2)', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183555562, end_position_on_the_genomic_accession=183590921, orientation='minus', exon_count=17, OMIM=608515, no_hdr0=''),
  4689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4689, CurrentID=0, Status='live', Symbol='NCF4', Aliases='NCF, P40PHOX, SH3PXD4', description='neutrophil cytosolic factor 4', other_designations='NCF-4|SH3 and PX domain-containing protein 4|neutrophil NADPH oxidase factor 4|neutrophil cytosolic factor 4, 40kDa|p40-phox', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36860988, end_position_on_the_genomic_accession=36878017, orientation='plus', exon_count=12, OMIM=601488, no_hdr0=''),
  4690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4690, CurrentID=0, Status='live', Symbol='NCK1', Aliases='NCK, NCKalpha, nck-1', description='NCK adaptor protein 1', other_designations='NCK tyrosine kinase|SH2/SH3 adaptor protein NCK-alpha|melanoma NCK protein|non-catalytic region of tyrosine kinase', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136862208, end_position_on_the_genomic_accession=136951610, orientation='plus', exon_count=5, OMIM=600508, no_hdr0=''),
  4691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4691, CurrentID=0, Status='live', Symbol='NCL', Aliases='C23', description='nucleolin', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231454748, end_position_on_the_genomic_accession=231464494, orientation='minus', exon_count=14, OMIM=164035, no_hdr0=''),
  4692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4692, CurrentID=0, Status='live', Symbol='NDN', Aliases='HsT16328, PWCR', description='necdin, MAGE family member', other_designations='Prader-Willi syndrome chromosome region|necdin homolog|necdin, melanoma antigen (MAGE) family member|necdin-like protein', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23685407, end_position_on_the_genomic_accession=23687303, orientation='minus', exon_count=1, OMIM=602117, no_hdr0=''),
  4693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4693, CurrentID=0, Status='live', Symbol='NDP', Aliases='EVR2, FEVR, ND', description='Norrie disease (pseudoglioma)', other_designations='X-linked exudative vitreoretinopathy 2 protein|norrie disease protein', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43948776, end_position_on_the_genomic_accession=43973675, orientation='minus', exon_count=3, OMIM=300658, no_hdr0=''),
  4694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4694, CurrentID=0, Status='live', Symbol='NDUFA1', Aliases='CI-MWFE, MWFE, ZNF183', description='NADH:ubiquinone oxidoreductase subunit A1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 1, 7.5kDa|NADH oxidoreductase subunit MWFE|NADH-ubiquinone oxidoreductase MWFE subunit|NADH:ubiquinone oxidoreductase (complex 1)|complex I MWFE subunit|complex I-MWFE|type I dehydrogenase', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119871771, end_position_on_the_genomic_accession=119876666, orientation='plus', exon_count=3, OMIM=300078, no_hdr0=''),
  4695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4695, CurrentID=0, Status='live', Symbol='NDUFA2', Aliases='B8, CD14, CIB8', description='NADH:ubiquinone oxidoreductase subunit A2', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 2, 8kDa|NADH-ubiquinone oxidoreductase subunit CI-B8|complex I B8 subunit', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140645363, end_position_on_the_genomic_accession=140647785, orientation='minus', exon_count=3, OMIM=602137, no_hdr0=''),
  4696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4696, CurrentID=0, Status='live', Symbol='NDUFA3', Aliases='B9, CI-B9', description='NADH:ubiquinone oxidoreductase subunit A3', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3, 9kDa|NADH-ubiquinone oxidoreductase B9 subunit|complex I B9 subunit|complex I-B9', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54102852, end_position_on_the_genomic_accession=54106974, orientation='plus', exon_count=4, OMIM=603832, no_hdr0=''),
  4697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4697, CurrentID=0, Status='live', Symbol='NDUFA4', Aliases='CI-9k, CI-MLRQ, MLRQ', description='NDUFA4, mitochondrial complex associated', other_designations='Complex I 9kDa subunit|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 4|NADH-ubiquinone oxidoreductase MLRQ subunit|complex I-MLRQ', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10931953, end_position_on_the_genomic_accession=10940186, orientation='minus', exon_count=4, OMIM=603833, no_hdr0=''),
  4698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4698, CurrentID=0, Status='live', Symbol='NDUFA5', Aliases='B13, CI-13KD-B, CI-13kB, NUFM, UQOR13', description='NADH:ubiquinone oxidoreductase subunit A5', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa|NADH-ubiquinone oxidoreductase 13 kDa-B subunit|complex I 13kDa subunit B|complex I subunit B13|type I dehydrogenase|ubiquinone reductase', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123536997, end_position_on_the_genomic_accession=123601651, orientation='minus', exon_count=13, OMIM=601677, no_hdr0=''),
  4699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4699, CurrentID=0, Status='live', Symbol='NDUFA5P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 1|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, pseudogene 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3321869, end_position_on_the_genomic_accession=3322790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4700, CurrentID=0, Status='live', Symbol='NDUFA6', Aliases='B14, CI-B14, LYRM6, NADHB14', description='NADH:ubiquinone oxidoreductase subunit A6', other_designations='Complex I-B14|LYR motif-containing protein 6|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 6, 14kDa|NADH-ubiquinone oxidoreductase 1 alpha subcomplex, 6|NADH-ubiquinone oxidoreductase B14 subunit|complex I B14 subunit', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42085526, end_position_on_the_genomic_accession=42090884, orientation='minus', exon_count=3, OMIM=602138, no_hdr0=''),
  4701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4701, CurrentID=0, Status='live', Symbol='NDUFA7', Aliases='B14.5a', description='NADH:ubiquinone oxidoreductase subunit A7', other_designations='CI-B14.5a|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 7, 14.5kDa|NADH-ubiquinone oxidoreductase subunit B14.5a|complex I B14.5a subunit', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8308762, end_position_on_the_genomic_accession=8321396, orientation='minus', exon_count=5, OMIM=602139, no_hdr0=''),
  4702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4702, CurrentID=0, Status='live', Symbol='NDUFA8', Aliases='CI-19KD, CI-PGIV, PGIV', description='NADH:ubiquinone oxidoreductase subunit A8', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 8, 19kDa|NADH-ubiquinone oxidoreductase 19 kDa subunit|NADH:ubiquinone oxidoreductase PGIV subunit|complex I-19kD|complex I-PGIV', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122132466, end_position_on_the_genomic_accession=122159819, orientation='minus', exon_count=5, OMIM=603359, no_hdr0=''),
  4703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4703, CurrentID=0, Status='live', Symbol='NEB', Aliases='NEB177D, NEM2', description='nebulin', other_designations='nemaline myopathy type 2', map_location='2q22', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151485339, end_position_on_the_genomic_accession=151734487, orientation='minus', exon_count=183, OMIM=161650, no_hdr0=''),
  4704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4704, CurrentID=0, Status='live', Symbol='NDUFA9', Aliases='CC6, CI-39k, CI39k, NDUFS2L, SDR22E1', description='NADH:ubiquinone oxidoreductase subunit A9', other_designations='CI-39kD|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 9, 39kDa|NADH dehydrogenase (ubiquinone) Fe-S protein 2-like (NADH-coenzyme Q reductase)|NADH-ubiquinone oxidoreductase 39 kDa subunit|complex I 39kDa subunit|short chain dehydrogenase/reductase family 22E, member 1', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4649098, end_position_on_the_genomic_accession=4687554, orientation='plus', exon_count=11, OMIM=603834, no_hdr0=''),
  4705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4705, CurrentID=0, Status='live', Symbol='NDUFA10', Aliases='CI-42KD, CI-42k', description='NADH:ubiquinone oxidoreductase subunit A10', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 10, 42kDa|NADH-ubiquinone oxidoreductase 42 kDa subunit|complex I 42kDa subunit', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239957372, end_position_on_the_genomic_accession=240025402, orientation='minus', exon_count=12, OMIM=603835, no_hdr0=''),
  4706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4706, CurrentID=0, Status='live', Symbol='NDUFAB1', Aliases='ACP, FASN2A, SDAP', description='NADH:ubiquinone oxidoreductase subunit AB1', other_designations='CI-SDAP|NADH dehydrogenase (ubiquinone) 1, alpha/beta subcomplex, 1, 8kDa|NADH-ubiquinone oxidoreductase 9.6 kDa subunit|NADH:ubiquinone oxidoreductase SDAP subunit|complex I SDAP subunit|mitochondrial acyl carrier protein', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23581014, end_position_on_the_genomic_accession=23596341, orientation='minus', exon_count=6, OMIM=603836, no_hdr0=''),
  4707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4707, CurrentID=0, Status='live', Symbol='NDUFB1', Aliases='CI-MNLL, CI-SGDH, MNLL', description='NADH:ubiquinone oxidoreductase subunit B1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 1, 7kDa|NADH-ubiquinone oxidoreductase MNLL subunit|complex I MNLL subunit|complex I-MNLL', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92116123, end_position_on_the_genomic_accession=92121809, orientation='minus', exon_count=3, OMIM=603837, no_hdr0=''),
  4708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4708, CurrentID=0, Status='live', Symbol='NDUFB2', Aliases='AGGG, CI-AGGG', description='NADH:ubiquinone oxidoreductase subunit B2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 2, 8kDa|NADH-ubiquinone oxidoreductase AGGG subunit|complex I AGGG subunit|complex I-AGGG', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140696681, end_position_on_the_genomic_accession=140706646, orientation='plus', exon_count=4, OMIM=603838, no_hdr0=''),
  4709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4709, CurrentID=0, Status='live', Symbol='NDUFB3', Aliases='B12, CI-B12', description='NADH:ubiquinone oxidoreductase subunit B3', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa|NADH-ubiquinone oxidoreductase B12 subunit|complex I-B12', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201071739, end_position_on_the_genomic_accession=201085750, orientation='plus', exon_count=4, OMIM=603839, no_hdr0=''),
  4710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4710, CurrentID=0, Status='live', Symbol='NDUFB4', Aliases='B15, CI-B15', description='NADH:ubiquinone oxidoreductase subunit B4', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 4, 15kDa|NADH-ubiquinone oxidoreductase B15 subunit|complex I B15 subunit|complex I-B15', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120596281, end_position_on_the_genomic_accession=120602411, orientation='plus', exon_count=3, OMIM=603840, no_hdr0=''),
  4711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4711, CurrentID=0, Status='live', Symbol='NDUFB5', Aliases='CISGDH, SGDH', description='NADH:ubiquinone oxidoreductase subunit B5', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 5, 16kDa|NADH-ubiquinone oxidoreductase SGDH subunit|complex I SGDH subunit|complex I-SGDH', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179604787, end_position_on_the_genomic_accession=179624500, orientation='plus', exon_count=6, OMIM=603841, no_hdr0=''),
  4712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4712, CurrentID=0, Status='live', Symbol='NDUFB6', Aliases='B17, CI', description='NADH:ubiquinone oxidoreductase subunit B6', other_designations='CI-B17|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 6, 17kDa|NADH-ubiquinone oxidoreductase B17 subunit|NADH-ubiquinone oxidoreductase beta subunit, 6|complex I, mitochondrial respiratory chain, B17 subunit|complex I-B17', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32553526, end_position_on_the_genomic_accession=32573184, orientation='minus', exon_count=4, OMIM=603322, no_hdr0=''),
  4713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4713, CurrentID=0, Status='live', Symbol='NDUFB7', Aliases='B18, CI-B18', description='NADH:ubiquinone oxidoreductase subunit B7', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 7, 18kDa|NADH-ubiquinone oxidoreductase B18 subunit|cell adhesion protein SQM1|complex I B18 subunit|complex I-B18', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14566078, end_position_on_the_genomic_accession=14572077, orientation='minus', exon_count=4, OMIM=603842, no_hdr0=''),
  4714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4714, CurrentID=0, Status='live', Symbol='NDUFB8', Aliases='ASHI, CI-ASHI', description='NADH:ubiquinone oxidoreductase subunit B8', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, 19kDa|NADH-ubiquinone oxidoreductase ASHI subunit|NADH:ubiquinone oxidoreductase ASHI subunit|complex I ASHI subunit|complex I-ASHI', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100523729, end_position_on_the_genomic_accession=100529923, orientation='minus', exon_count=5, OMIM=602140, no_hdr0=''),
  4715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4715, CurrentID=0, Status='live', Symbol='NDUFB9', Aliases='B22, CI-B22, LYRM3, UQOR22', description='NADH:ubiquinone oxidoreductase subunit B9', other_designations='LYR motif-containing protein 3|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 9, 22kDa|NADH-ubiquinone oxidoreductase B22 subunit|complex I B22 subunit', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124539102, end_position_on_the_genomic_accession=124549986, orientation='plus', exon_count=4, OMIM=601445, no_hdr0=''),
  4716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4716, CurrentID=0, Status='live', Symbol='NDUFB10', Aliases='PDSW', description='NADH:ubiquinone oxidoreductase subunit B10', other_designations='CI-PDSW|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 10, 22kDa|NADH ubiquinone oxidoreductase PDSW subunit (RH 16p13.3)|NADH-ubiquinone oxidoreductase PDSW subunit|complex I PDSW subunit|complex I-PDSW', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1959516, end_position_on_the_genomic_accession=1961975, orientation='plus', exon_count=4, OMIM=603843, no_hdr0=''),
  4717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4717, CurrentID=0, Status='live', Symbol='NDUFC1', Aliases='KFYI', description='NADH:ubiquinone oxidoreductase subunit C1', other_designations='CI-KFYI|NADH dehydrogenase (ubiquinone) 1, subcomplex unknown, 1, 6kDa|NADH-ubiquinone oxidoreductase KFYI subunit|complex I KFYI subunit|complex I-KFYI', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139289917, end_position_on_the_genomic_accession=139302551, orientation='minus', exon_count=8, OMIM=603844, no_hdr0=''),
  4718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4718, CurrentID=0, Status='live', Symbol='NDUFC2', Aliases='B14.5b, CI-B14.5b, HLC-1, NADHDH2', description='NADH:ubiquinone oxidoreductase subunit C2', other_designations='NADH dehydrogenase (ubiquinone) 1, subcomplex unknown, 2, 14.5kDa|NADH-ubiquinone oxidoreductase subunit B14.5b|complex I subunit B14.5b|complex I-B14.5b|human lung cancer oncogene 1 protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78068347, end_position_on_the_genomic_accession=78080219, orientation='minus', exon_count=4, OMIM=603845, no_hdr0=''),
  4719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4719, CurrentID=0, Status='live', Symbol='NDUFS1', Aliases='CI-75Kd, CI-75k, PRO1304', description='NADH:ubiquinone oxidoreductase core subunit S1', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 1, 75kDa (NADH-coenzyme Q reductase)|complex I 75kDa subunit|complex I, mitochondrial respiratory chain, 75-kD subunit|mitochondrial NADH-ubiquinone oxidoreductase 75 kDa subunit', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206123079, end_position_on_the_genomic_accession=206159519, orientation='minus', exon_count=20, OMIM=157655, no_hdr0=''),
  4720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4720, CurrentID=0, Status='live', Symbol='NDUFS2', Aliases='CI-49', description='NADH:ubiquinone oxidoreductase core subunit S2', other_designations='CI-49kD|NADH dehydrogenase (ubiquinone) Fe-S protein 2, 49kDa (NADH-coenzyme Q reductase)|NADH-ubiquinone oxidoreductase 49 kDa subunit|NADH-ubiquinone oxidoreductase NDUFS2 subunit|complex 1, mitochondrial respiratory chain, 49-KD subunit|complex I 49kDa subunit|complex I-49kD', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161197419, end_position_on_the_genomic_accession=161214395, orientation='plus', exon_count=16, OMIM=602985, no_hdr0=''),
  4722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4722, CurrentID=0, Status='live', Symbol='NDUFS3', Aliases='CI-30', description='NADH:ubiquinone oxidoreductase core subunit S3', other_designations='CI-30kD|NADH dehydrogenase (ubiquinone) Fe-S protein 3, 30kDa (NADH-coenzyme Q reductase)|NADH dehydrogenase-ubiquinone 30 kDa subunit|NADH-ubiquinone oxidoreductase 30 kDa subunit|complex I 30kDa subunit|complex I-30kD', map_location='11p11.11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47579010, end_position_on_the_genomic_accession=47584563, orientation='plus', exon_count=7, OMIM=603846, no_hdr0=''),
  4723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4723, CurrentID=0, Status='live', Symbol='NDUFV1', Aliases='CI-51K, CI51KD, UQOR1', description='NADH:ubiquinone oxidoreductase core subunit V1', other_designations='NADH dehydrogenase (ubiquinone) flavoprotein 1, 51kDa|NADH-ubiquinone oxidoreductase 51 kDa subunit|complex I 51 kda subunit|complex I 51kDa subunit|complex I, mitochondrial respiratory chain|mitochondrial NADH dehydrogenase ubiquinone flavoprotein 1|mitochondrial NADH:ubiquinone oxidoreductase 51 kda subunit', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67606852, end_position_on_the_genomic_accession=67612541, orientation='plus', exon_count=10, OMIM=161015, no_hdr0=''),
  4724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4724, CurrentID=0, Status='live', Symbol='NDUFS4', Aliases='AQDQ, CI-18, CI-18 kDa, CI-AQDQ', description='NADH:ubiquinone oxidoreductase subunit S4', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 4, 18kDa (NADH-coenzyme Q reductase)|NADH-ubiquinone oxidoreductase 18 kDa subunit|complex I 18kDa subunit|complex I-AQDQ|mitochondrial respiratory chain complex I (18-KD subunit)', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53560623, end_position_on_the_genomic_accession=53683341, orientation='plus', exon_count=7, OMIM=602694, no_hdr0=''),
  4725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4725, CurrentID=0, Status='live', Symbol='NDUFS5', Aliases='CI-15k, CI15K', description='NADH:ubiquinone oxidoreductase subunit S5', other_designations='CI-15 kDa|NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase)|NADH:ubiquinone oxidoreductase 15 kDa IP subunit|complex I-15 kDa', map_location='1p34.2-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39026295, end_position_on_the_genomic_accession=39034636, orientation='plus', exon_count=3, OMIM=603847, no_hdr0=''),
  4726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4726, CurrentID=0, Status='live', Symbol='NDUFS6', Aliases='CI-13kA, CI-13kD-A, CI13KDA', description='NADH:ubiquinone oxidoreductase subunit S6', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 6, 13kDa (NADH-coenzyme Q reductase)|NADH-ubiquinone oxidoreductase 13 kDa-A subunit|NADH:ubiquinone oxidoreductase NDUFS6 subunit|complex I 13kDa subunit A|complex I, mitochondrial respiratory chain, 13-kD subunit', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1801382, end_position_on_the_genomic_accession=1816051, orientation='plus', exon_count=4, OMIM=603848, no_hdr0=''),
  4728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4728, CurrentID=0, Status='live', Symbol='NDUFS8', Aliases='CI-23k, CI23KD, TYKY', description='NADH:ubiquinone oxidoreductase core subunit S8', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 8, 23kDa (NADH-coenzyme Q reductase)|NADH-ubiquinone oxidoreductase 23 kDa subunit|complex I 23kDa subunit|complex I-23kD', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68030617, end_position_on_the_genomic_accession=68036647, orientation='plus', exon_count=9, OMIM=602141, no_hdr0=''),
  4729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4729, CurrentID=0, Status='live', Symbol='NDUFV2', Aliases='CI-24k', description='NADH:ubiquinone oxidoreductase core subunit V2', other_designations='NADH dehydrogenase (ubiquinone) flavoprotein 2, 24kDa|NADH dehydrogenase ubiquinone flavoprotein 2, mitochondrial|NADH-ubiquinone oxidoreductase 24 kDa subunit|NADH-ubiquinone oxidoreductase flavoprotein 2|complex I 24kDa subunit|complex I, mitochondrial respitory chain, 24 kD subunit|nuclear-encoded mitochondrial NADH-ubiquinone reductase 24Kd subunit', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9102630, end_position_on_the_genomic_accession=9134345, orientation='plus', exon_count=8, OMIM=600532, no_hdr0=''),
  4730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4730, CurrentID=0, Status='live', Symbol='NDUFV2P1', Aliases='', description='NADH:ubiquinone oxidoreductase core subunit V2 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) flavoprotein 2 pseudogene 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53223908, end_position_on_the_genomic_accession=53224879, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4731, CurrentID=0, Status='live', Symbol='NDUFV3', Aliases='CI-10k, CI-9KD', description='NADH:ubiquinone oxidoreductase subunit V3', other_designations='NADH dehydrogenase (ubiquinone) flavoprotein 3, 10kDa|NADH-ubiquinone oxidoreductase 9 kD subunit|NADH-ubiquinone oxidoreductase 9 kDa subunit|NADH-ubiquinone oxidoreductase flavoprotein 3, 10kD|complex I 10kDa subunit|complex I, mitochondrial respiratory chain, 10-kD subunit|complex I-9kD|mitochondrial NADH oxidoreductase-like protein|renal carcinoma antigen NY-REN-4', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42893268, end_position_on_the_genomic_accession=42909670, orientation='plus', exon_count=5, OMIM=602184, no_hdr0=''),
  4733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4733, CurrentID=0, Status='live', Symbol='DRG1', Aliases='NEDD3', description='developmentally regulated GTP binding protein 1', other_designations='DRG-1|NEDD-3|developmentally regulated GTP-binding protein 1|neural precursor cell expressed developmentally down-regulated protein 3|neural precursor cell expressed, developmentally down-regulated 3', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31399553, end_position_on_the_genomic_accession=31434186, orientation='plus', exon_count=9, OMIM=603952, no_hdr0=''),
  4734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4734, CurrentID=0, Status='live', Symbol='NEDD4', Aliases='NEDD4-1, RPF1', description='neural precursor cell expressed, developmentally down-regulated 4, E3 ubiquitin protein ligase', other_designations='cell proliferation-inducing gene 53 protein|neural precursor cell expressed developmentally down-regulated protein 4|receptor-potentiating factor 1', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55826917, end_position_on_the_genomic_accession=55993746, orientation='minus', exon_count=33, OMIM=602278, no_hdr0=''),
  4735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4735, CurrentID=0, Status='live', Symbol='SEPT2', Aliases='DIFF6, NEDD-5, NEDD5, Pnutl3, hNedd5', description='septin 2', other_designations='neural precursor cell expressed developmentally down-regulated protein 5|neural precursor cell expressed, developmentally down-regulated 5', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241315187, end_position_on_the_genomic_accession=241354027, orientation='plus', exon_count=20, OMIM=601506, no_hdr0=''),
  4736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4736, CurrentID=0, Status='live', Symbol='RPL10A', Aliases='CSA19, Csa-19, L10A, NEDD6', description='ribosomal protein L10a', other_designations='NEDD-6|neural precursor cell expressed developmentally down-regulated protein 6', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35468401, end_position_on_the_genomic_accession=35470781, orientation='plus', exon_count=6, OMIM=615660, no_hdr0=''),
  4738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4738, CurrentID=0, Status='live', Symbol='NEDD8', Aliases='NEDD-8', description='neural precursor cell expressed, developmentally down-regulated 8', other_designations='neddylin|ubiquitin-like protein Nedd8', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24216851, end_position_on_the_genomic_accession=24232370, orientation='minus', exon_count=4, OMIM=603171, no_hdr0=''),
  4739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4739, CurrentID=0, Status='live', Symbol='NEDD9', Aliases='CAS-L, CAS2, CASL, CASS2, HEF1', description='neural precursor cell expressed, developmentally down-regulated 9', other_designations='Cas scaffolding protein family member 2|Crk-associated substrate related protein Cas-L|Enhancer of filamentation 1 p55|cas-like docking|neural precursor cell expressed developmentally down-regulated protein 9|p130Cas-related protein|renal carcinoma antigen NY-REN-12', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11183298, end_position_on_the_genomic_accession=11382348, orientation='minus', exon_count=11, OMIM=602265, no_hdr0=''),
  4741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4741, CurrentID=0, Status='live', Symbol='NEFM', Aliases='NEF3, NF-M, NFM', description='neurofilament, medium polypeptide', other_designations='160 kDa neurofilament protein|neurofilament 3|neurofilament triplet M protein|neurofilament, medium polypeptide 150kDa|neurofilament-3 (150 kD medium)', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24913761, end_position_on_the_genomic_accession=24919093, orientation='plus', exon_count=4, OMIM=162250, no_hdr0=''),
  4744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4744, CurrentID=0, Status='live', Symbol='NEFH', Aliases='NFH', description='neurofilament, heavy polypeptide', other_designations='200 kDa neurofilament protein|NF-H|neurofilament triplet H protein|neurofilament, heavy polypeptide 200kDa', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29480192, end_position_on_the_genomic_accession=29491290, orientation='plus', exon_count=4, OMIM=162230, no_hdr0=''),
  4745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4745, CurrentID=0, Status='live', Symbol='NELL1', Aliases='IDH3GL, NRP1', description='neural EGFL like 1', other_designations='nel-related protein 1|neural epidermal growth factor-like 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20669551, end_position_on_the_genomic_accession=21575686, orientation='plus', exon_count=22, OMIM=602319, no_hdr0=''),
  4747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4747, CurrentID=0, Status='live', Symbol='NEFL', Aliases='CMT1F, CMT2E, NF-L, NF68, NFL, PPP1R110', description='neurofilament, light polypeptide', other_designations='light molecular weight neurofilament protein|neurofilament protein, light chain|neurofilament subunit NF-L|neurofilament triplet L protein|neurofilament, light polypeptide 68kDa|protein phosphatase 1, regulatory subunit 110', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24950955, end_position_on_the_genomic_accession=24956869, orientation='minus', exon_count=4, OMIM=162280, no_hdr0=''),
  4750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4750, CurrentID=0, Status='live', Symbol='NEK1', Aliases='NY-REN-55, SRPS2, SRPS2A, SRTD6', description='NIMA-related kinase 1', other_designations='NIMA (never in mitosis gene a)-related kinase 1|never in mitosis A-related kinase 1|nimA-related protein kinase 1|protein-serine/threonine kinase|renal carcinoma antigen NY-REN-55', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169393270, end_position_on_the_genomic_accession=169612629, orientation='minus', exon_count=37, OMIM=604588, no_hdr0=''),
  4751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4751, CurrentID=0, Status='live', Symbol='NEK2', Aliases='HsPK21A, NLK1, PPP1R111, RP67, NEK2', description='NIMA-related kinase 2', other_designations='NIMA (never in mitosis gene a)-related kinase 2|nimA-like protein kinase 1|nimA-related protein kinase 2|protein phosphatase 1, regulatory subunit 111', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211658256, end_position_on_the_genomic_accession=211675630, orientation='minus', exon_count=9, OMIM=604043, no_hdr0=''),
  4752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4752, CurrentID=0, Status='live', Symbol='NEK3', Aliases='HSPK36', description='NIMA-related kinase 3', other_designations='HSPK 36|NIMA (never in mitosis gene a)-related kinase 3|glycogen synthase A kinase|hydroxyalkyl-protein kinase|never in mitosis A-related kinase 3|nimA-related protein kinase 3|phosphorylase B kinase kinase', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52132643, end_position_on_the_genomic_accession=52159861, orientation='minus', exon_count=17, OMIM=604044, no_hdr0=''),
  4753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4753, CurrentID=0, Status='live', Symbol='NELL2', Aliases='NRP2', description='neural EGFL like 2', other_designations='NEL-like protein 2|NEL-related protein 2|neural epidermal growth factor-like 2', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=44508275, end_position_on_the_genomic_accession=44921848, orientation='minus', exon_count=26, OMIM=602320, no_hdr0=''),
  101929165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929165, CurrentID=0, Status='live', Symbol='LOC101929165', Aliases='', description='uncharacterized LOC101929165', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74912672, end_position_on_the_genomic_accession=74944808, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  4756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4756, CurrentID=0, Status='live', Symbol='NEO1', Aliases='IGDCC2, NGN, NTN1R2', description='neogenin 1', other_designations='immunoglobulin superfamily DCC subclass member 2|neogenin homolog 1', map_location='15q22.3-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73051715, end_position_on_the_genomic_accession=73305206, orientation='plus', exon_count=30, OMIM=601907, no_hdr0=''),
  4758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4758, CurrentID=0, Status='live', Symbol='NEU1', Aliases='NANH, NEU, SIAL1', description='neuraminidase 1 (lysosomal sialidase)', other_designations='G9 sialidase|N-acetyl-alpha-neuraminidase 1|acetylneuraminyl hydrolase|exo-alpha-sialidase|lysosomal sialidase|sialidase 1 (lysosomal sialidase)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31859052, end_position_on_the_genomic_accession=31862932, orientation='minus', exon_count=6, OMIM=608272, no_hdr0=''),
  4759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4759, CurrentID=0, Status='live', Symbol='NEU2', Aliases='SIAL2', description='neuraminidase 2 (cytosolic sialidase)', other_designations='N-acetyl-alpha-neuraminidase 2|cytosolic sialidase|neuraminidase 2|sialidase 2 (cytosolic sialidase)', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233023056, end_position_on_the_genomic_accession=233035532, orientation='plus', exon_count=3, OMIM=605528, no_hdr0=''),
  4760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4760, CurrentID=0, Status='live', Symbol='NEUROD1', Aliases='BETA2, BHF-1, MODY6, NEUROD, bHLHa3', description='neuronal differentiation 1', other_designations='basic helix-loop-helix transcription factor|beta-cell E-box transactivator 2|class A basic helix-loop-helix protein 3|neurogenic helix-loop-helix protein NEUROD', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181676106, end_position_on_the_genomic_accession=181680665, orientation='minus', exon_count=2, OMIM=601724, no_hdr0=''),
  4761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4761, CurrentID=0, Status='live', Symbol='NEUROD2', Aliases='NDRF, bHLHa1', description='neuronal differentiation 2', other_designations='class A basic helix-loop-helix protein 1|neuroD-related factor|neurogenic basic-helix-loop-helix protein|neurogenic differentiation 2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39603768, end_position_on_the_genomic_accession=39607922, orientation='minus', exon_count=4, OMIM=601725, no_hdr0=''),
  4762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4762, CurrentID=0, Status='live', Symbol='NEUROG1', Aliases='AKA, Math4C, NEUROD3, bHLHa6, ngn1', description='neurogenin 1', other_designations='NGN-1|class A basic helix-loop-helix protein 6|neurogenic basic-helix-loop-helix protein|neurogenic differentiation 3|neurogenic differentiation factor 3', map_location='5q23-q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135534282, end_position_on_the_genomic_accession=135535949, orientation='minus', exon_count=1, OMIM=601726, no_hdr0=''),
  4763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4763, CurrentID=0, Status='live', Symbol='NF1', Aliases='NFNS, VRNF, WSS', description='neurofibromin 1', other_designations='neurofibromatosis-related protein NF-1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31007873, end_position_on_the_genomic_accession=31377677, orientation='plus', exon_count=73, OMIM=613113, no_hdr0=''),
  4764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4764, CurrentID=0, Status='live', Symbol='NF1P3', Aliases='NF1L1', description='neurofibromin 1 pseudogene 3', other_designations='neurofibromin 1-like 1 (pseudogene)', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14000925, end_position_on_the_genomic_accession=14005279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270874, CurrentID=0, Status='live', Symbol='RPSAP22', Aliases='RPSA_5_248', description='ribosomal protein SA pseudogene 22', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85490925, end_position_on_the_genomic_accession=85491780, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4768, CurrentID=0, Status='live', Symbol='NF1P5', Aliases='NF1L5', description='neurofibromin 1 pseudogene 5', other_designations='Putative neurofibromin 1-like protein 5|neurofibromin 1-like 5', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14152997, end_position_on_the_genomic_accession=14157344, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101978785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101978785, CurrentID=0, Status='live', Symbol='PCAT14', Aliases='', description='prostate cancer associated transcript 14 (non-protein coding)', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23538102, end_position_on_the_genomic_accession=23547306, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  4771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4771, CurrentID=0, Status='live', Symbol='NF2', Aliases='ACN, BANF, SCH', description='neurofibromin 2 (merlin)', other_designations='moesin-ezrin-radixin like|moesin-ezrin-radixin-like protein|moesin-ezrin-radizin-like protein|neurofibromin 2 (bilateral acoustic neuroma)|neurofibromin-2|schwannomerlin|schwannomin', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29603556, end_position_on_the_genomic_accession=29698600, orientation='plus', exon_count=18, OMIM=607379, no_hdr0=''),
  4772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4772, CurrentID=0, Status='live', Symbol='NFATC1', Aliases='NF-ATC, NF-ATc1.2, NFAT2, NFATc', description='nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 1', other_designations="NFAT transcription complex cytosolic component|nuclear factor of activated T-cells 'c'", map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79395772, end_position_on_the_genomic_accession=79529323, orientation='plus', exon_count=11, OMIM=600489, no_hdr0=''),
  4773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4773, CurrentID=0, Status='live', Symbol='NFATC2', Aliases='NFAT1, NFATP', description='nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 2', other_designations='NF-ATc2|NFAT pre-existing subunit|NFAT transcription complex, preexisting component|T cell transcription factor NFAT1|T-cell transcription factor NFAT1|nuclear factor of activated T-cells, preexisting component|preexisting nuclear factor of activated T-cells 2', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51386957, end_position_on_the_genomic_accession=51562858, orientation='minus', exon_count=13, OMIM=600490, no_hdr0=''),
  4774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4774, CurrentID=0, Status='live', Symbol='NFIA', Aliases='CTF, NF-I/A, NF1-A, NFI-A, NFI-L', description='nuclear factor I/A', other_designations='CCAAT-box-binding transcription factor|TGGCA-binding protein', map_location='1p31.3-p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61077257, end_position_on_the_genomic_accession=61462788, orientation='plus', exon_count=15, OMIM=600727, no_hdr0=''),
  4775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4775, CurrentID=0, Status='live', Symbol='NFATC3', Aliases='NFAT4, NFATX', description='nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 3', other_designations='NF-ATc3|T cell transcription factor NFAT4|nuclear factor of activated T-cells c3 isoform IE-Xa', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68085366, end_position_on_the_genomic_accession=68229259, orientation='plus', exon_count=12, OMIM=602698, no_hdr0=''),
  4776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4776, CurrentID=0, Status='live', Symbol='NFATC4', Aliases='NF-AT3, NF-ATC4, NFAT3', description='nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 4', other_designations='T-cell transcription factor NFAT3', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24366911, end_position_on_the_genomic_accession=24379604, orientation='plus', exon_count=12, OMIM=602699, no_hdr0=''),
  100270876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270876, CurrentID=0, Status='live', Symbol='RPSAP34', Aliases='RPSA_8_506', description='ribosomal protein SA pseudogene 34', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108407839, end_position_on_the_genomic_accession=108408561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4778, CurrentID=0, Status='live', Symbol='NFE2', Aliases='NF-E2, p45', description='nuclear factor, erythroid 2', other_designations='leucine zipper protein NF-E2|nuclear factor (erythroid-derived 2), 45kD|nuclear factor (erythroid-derived 2), 45kDa|nuclear factor, erythroid-derived 2 45 kDa subunit|p45 NF-E2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54292107, end_position_on_the_genomic_accession=54301037, orientation='minus', exon_count=7, OMIM=601490, no_hdr0=''),
  4779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4779, CurrentID=0, Status='live', Symbol='NFE2L1', Aliases='LCR-F1, NRF1, TCF11', description='nuclear factor, erythroid 2 like 1', other_designations='NF-E2-related factor 1|NFE2-related factor 1|TCF-11|locus control region-factor 1|nuclear factor (erythroid-derived 2)-like 1|nuclear factor, erythroid 2-like 1|nuclear factor, erythroid derived 2, like 1|transcription factor 11 (basic leucine zipper type)|transcription factor HBZ17|transcription factor LCR-F1', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48048324, end_position_on_the_genomic_accession=48061545, orientation='plus', exon_count=7, OMIM=163260, no_hdr0=''),
  4780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4780, CurrentID=0, Status='live', Symbol='NFE2L2', Aliases='HEBP1, NRF2', description='nuclear factor, erythroid 2 like 2', other_designations='nuclear factor erythroid-derived 2-like 2|nuclear factor, erythroid 2-like 2', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177230303, end_position_on_the_genomic_accession=177265131, orientation='minus', exon_count=6, OMIM=600492, no_hdr0=''),
  4781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4781, CurrentID=0, Status='live', Symbol='NFIB', Aliases='CTF, HMGIC/NFIB, NF-I/B, NF1-B, NFI-B, NFI-RED2, NFIB3, NFIB', description='nuclear factor I/B', other_designations='CCAAT-box-binding transcription factor|TGGCA-binding protein|nuclear factor 1/B', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14081843, end_position_on_the_genomic_accession=14398983, orientation='minus', exon_count=16, OMIM=600728, no_hdr0=''),
  4782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4782, CurrentID=0, Status='live', Symbol='NFIC', Aliases='CTF, CTF5, NF-I, NFI', description='nuclear factor I/C (CCAAT-binding transcription factor)', other_designations='CCAAT-box-binding transcription factor|NF-I/C|NF1-C|TGGCA-binding protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3359563, end_position_on_the_genomic_accession=3469217, orientation='plus', exon_count=12, OMIM=600729, no_hdr0=''),
  4783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4783, CurrentID=0, Status='live', Symbol='NFIL3', Aliases='E4BP4, IL3BP1, NF-IL3AA, NFIL3', description='nuclear factor, interleukin 3 regulated', other_designations='E4 promoter-binding protein 4|interleukin-3 promoter transcriptional activator|interleukin-3-binding protein 1|nuclear factor interleukin 3 regulated protein|transcriptional activator NF-IL3A', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91409045, end_position_on_the_genomic_accession=91424626, orientation='minus', exon_count=3, OMIM=605327, no_hdr0=''),
  4784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4784, CurrentID=0, Status='live', Symbol='NFIX', Aliases='MRSHSS, NF1A, SOTOS2', description='nuclear factor I/X (CCAAT-binding transcription factor)', other_designations='CCAAT-box-binding transcription factor|CTF|NF-I/X|NF1-X|TGGCA-binding protein|nuclear factor 1/X', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12995756, end_position_on_the_genomic_accession=13098796, orientation='plus', exon_count=13, OMIM=164005, no_hdr0=''),
  798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=798, CurrentID=0, Status='live', Symbol='CALCP', Aliases='CALC3', description='calcitonin pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14907499, end_position_on_the_genomic_accession=14907999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4790, CurrentID=0, Status='live', Symbol='NFKB1', Aliases='EBP-1, KBF1, NF-kB1, NF-kappa-B, NF-kappaB, NFKB-p105, NFKB-p50, NFkappaB, p105, p50', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells 1', other_designations='DNA binding factor KBF1|DNA-binding factor KBF1|NF-kappabeta|nuclear factor NF-kappa-B p50 subunit|nuclear factor kappa-B DNA binding subunit', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102501329, end_position_on_the_genomic_accession=102617302, orientation='plus', exon_count=26, OMIM=164011, no_hdr0=''),
  4791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4791, CurrentID=0, Status='live', Symbol='NFKB2', Aliases='CVID10, H2TF1, LYT-10, LYT10, NF-kB2, p100, p52', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells 2 (p49/p100)', other_designations='DNA-binding factor KBF2|NFKB, p52/p100 subunit|lymphocyte translocation chromosome 10 protein|nuclear factor Kappa-B, subunit 2|nuclear factor NF-kappa-B p52 subunit|nuclear factor of Kappa light chain gene enhancer in B cells 2|oncogene Lyt-10|transcription factor NFKB2', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102394110, end_position_on_the_genomic_accession=102402529, orientation='plus', exon_count=23, OMIM=164012, no_hdr0=''),
  4792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4792, CurrentID=0, Status='live', Symbol='NFKBIA', Aliases='IKBA, MAD-3, NFKBI', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, alpha', other_designations='I-kappa-B-alpha|IkappaBalpha|ikB-alpha|major histocompatibility complex enhancer-binding protein MAD3|nuclear factor of kappa light chain gene enhancer in B-cells', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35401510, end_position_on_the_genomic_accession=35404754, orientation='minus', exon_count=6, OMIM=164008, no_hdr0=''),
  4793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4793, CurrentID=0, Status='live', Symbol='NFKBIB', Aliases='IKBB, TRIP9', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, beta', other_designations='I-kappa-B-beta|NF-kappa-BIB|TR-interacting protein 9|TRIP-9|ikB-B|ikB-beta|ikappaBbeta|thyroid receptor-interacting protein 9', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38899657, end_position_on_the_genomic_accession=38908894, orientation='plus', exon_count=7, OMIM=604495, no_hdr0=''),
  4794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4794, CurrentID=0, Status='live', Symbol='NFKBIE', Aliases='IKBE', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, epsilon', other_designations='I-kappa-B-epsilon|NF-kappa-BIE|ikB-E|ikB-epsilon|ikappaBepsilon', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44258166, end_position_on_the_genomic_accession=44265788, orientation='minus', exon_count=6, OMIM=604548, no_hdr0=''),
  4795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4795, CurrentID=0, Status='live', Symbol='NFKBIL1', Aliases='IKBL, LST1, NFKBIL', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor-like 1', other_designations='I-kappa-B-like protein|ikappaBL|inhibitor of kappa B-like protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31546851, end_position_on_the_genomic_accession=31558829, orientation='plus', exon_count=5, OMIM=601022, no_hdr0=''),
  4796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4796, CurrentID=0, Status='live', Symbol='TONSL', Aliases='IKBR, NFKBIL2', description='tonsoku-like, DNA repair protein', other_designations='I-kappa-B-related protein|NF-kappa-B inhibitor-like protein 2|ikappaBR|inhibitor of kappa B-related protein|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor-like 2', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144428780, end_position_on_the_genomic_accession=144444488, orientation='minus', exon_count=27, OMIM=604546, no_hdr0=''),
  791229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791229, CurrentID=0, Status='live', Symbol='MACOM', Aliases='', description='macrophthalmia, colobomatous, with microcornea', other_designations='', map_location='2p23-p16', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602499, no_hdr0=''),
  4798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4798, CurrentID=0, Status='live', Symbol='NFRKB', Aliases='INO80G', description='nuclear factor related to kappaB binding protein', other_designations='DNA-binding protein R kappa B|DNA-binding protein R kappa-B|INO80 complex subunit G', map_location='11q24-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129863773, end_position_on_the_genomic_accession=129895595, orientation='minus', exon_count=29, OMIM=164013, no_hdr0=''),
  4799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4799, CurrentID=0, Status='live', Symbol='NFX1', Aliases='NFX2, TEG-42, Tex42', description='nuclear transcription factor, X-box binding 1', other_designations='nuclear transcription factor, X box-binding protein 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33290420, end_position_on_the_genomic_accession=33371157, orientation='plus', exon_count=25, OMIM=603255, no_hdr0=''),
  4800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4800, CurrentID=0, Status='live', Symbol='NFYA', Aliases='CBF-A, CBF-B, HAP2, NF-YA', description='nuclear transcription factor Y subunit alpha', other_designations='CAAT box DNA-binding protein subunit A|CAAT-box DNA binding protein subunit A|CCAAT-binding transcription factor subunit B|HAP2 CCAAT-binding protein|Transcription factor NF-Y, A subunit|nuclear transcription factor Y subunit A|nuclear transcription factor Y, alpha', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41072968, end_position_on_the_genomic_accession=41102407, orientation='plus', exon_count=10, OMIM=189903, no_hdr0=''),
  4801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4801, CurrentID=0, Status='live', Symbol='NFYB', Aliases='CBF-A, CBF-B, HAP3, NF-YB', description='nuclear transcription factor Y subunit beta', other_designations='CAAT box DNA-binding protein subunit B|CCAAT-binding transcription factor subunit A|Transcription factor NF-Y, B subunit|nuclear transcription factor Y subunit B|nuclear transcription factor Y, beta', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104117080, end_position_on_the_genomic_accession=104138262, orientation='minus', exon_count=10, OMIM=189904, no_hdr0=''),
  4802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4802, CurrentID=0, Status='live', Symbol='NFYC', Aliases='CBF-C, CBFC, H1TF2A, HAP5, HSM, NF-YC', description='nuclear transcription factor Y subunit gamma', other_designations='CAAT box DNA-binding protein subunit C|CCAAT binding factor subunit C|CCAAT transcription binding factor subunit gamma|histone H1 transcription factor large subunit 2A|nuclear transcription factor Y subunit C|nuclear transcription factor Y, gamma|transactivator HSM-1|transactivator HSM-1/2|transcription factor NF-Y, C subunit', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40691570, end_position_on_the_genomic_accession=40771603, orientation='plus', exon_count=20, OMIM=605344, no_hdr0=''),
  4803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4803, CurrentID=0, Status='live', Symbol='NGF', Aliases='Beta-NGF, HSAN5B, NGF', description='nerve growth factor (beta polypeptide)', other_designations='nerve growth factor, beta subunit', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115285916, end_position_on_the_genomic_accession=115338384, orientation='minus', exon_count=5, OMIM=162030, no_hdr0=''),
  4804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4804, CurrentID=0, Status='live', Symbol='NGFR', Aliases='CD271, Gp80-LNGFR, TNFRSF16, p75(NTR), p75NTR', description='nerve growth factor receptor', other_designations='NGF receptor|TNFR superfamily, member 16|low affinity nerve growth factor receptor|low affinity neurotrophin receptor p75NTR|low-affinity nerve growth factor receptor|p75 ICD', map_location='17q21-q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49495293, end_position_on_the_genomic_accession=49515020, orientation='plus', exon_count=6, OMIM=162010, no_hdr0=''),
  4805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4805, CurrentID=0, Status='live', Symbol='NHCP1', Aliases='', description='non-histone chromosome protein 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=118870, no_hdr0=''),
  131873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131873, CurrentID=0, Status='live', Symbol='COL6A6', Aliases='', description='collagen, type VI, alpha 6', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130517177, end_position_on_the_genomic_accession=130678135, orientation='plus', exon_count=44, OMIM=616613, no_hdr0=''),
  4808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4808, CurrentID=0, Status='live', Symbol='NHLH2', Aliases='HEN2, NSCL2, bHLHa34', description='nescient helix-loop-helix 2', other_designations='HEN-2|NSCL-2|class A basic helix-loop-helix protein 34|nescient helix loop helix 2', map_location='1p12-p11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115831911, end_position_on_the_genomic_accession=115841127, orientation='minus', exon_count=5, OMIM=162361, no_hdr0=''),
  4809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4809, CurrentID=0, Status='live', Symbol='SNU13', Aliases='15.5K, FA-1, FA1, NHP2L1, NHPX, OTK27, SNRNP15-5, SPAG12, SSFA1', description='SNU13 homolog, small nuclear ribonucleoprotein (U4/U6.U5)', other_designations='NHP2 non-histone chromosome protein 2-like 1|U4/U6.U5 small nuclear ribonucleoprotein SNU13|U4/U6.U5 tri-snRNP 15.5 kDa protein|[U4/U6.U5] tri-snRNP 15.5 kD RNA binding protein|high mobility group-like nuclear protein 2 homolog 1|non-histone chromosome protein 2-like 1|small nuclear ribonucleoprotein 15.5kDa (U4/U6.U5)|sperm specific antigen 1', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41673933, end_position_on_the_genomic_accession=41689292, orientation='minus', exon_count=5, OMIM=601304, no_hdr0=''),
  4810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4810, CurrentID=0, Status='live', Symbol='NHS', Aliases='CTRCT40, CXN, SCML1', description='NHS actin remodeling regulator', other_designations='Nance-Horan syndrome (congenital cataracts and dental anomalies)|congenital cataracts and dental anomalies protein', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17375420, end_position_on_the_genomic_accession=17735994, orientation='plus', exon_count=11, OMIM=300457, no_hdr0=''),
  4811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4811, CurrentID=0, Status='live', Symbol='NID1', Aliases='NID', description='nidogen 1', other_designations='NID-1|enactin|entactin', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235975830, end_position_on_the_genomic_accession=236065181, orientation='minus', exon_count=20, OMIM=131390, no_hdr0=''),
  4812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4812, CurrentID=0, Status='live', Symbol='NIDDM1', Aliases='', description='non-insulin-dependent diabetes mellitus (common, type 2) 1', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601283, no_hdr0=''),
  4813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4813, CurrentID=0, Status='live', Symbol='NIDDM2', Aliases='', description='non-insulin-dependent diabetes mellitus (common, type 2) 2', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601407, no_hdr0=''),
  4814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4814, CurrentID=0, Status='live', Symbol='NINJ1', Aliases='NIN1, NINJURIN', description='ninjurin 1', other_designations='nerve injury-induced protein 1|nerve injury-induced protein-1', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93121489, end_position_on_the_genomic_accession=93134304, orientation='minus', exon_count=5, OMIM=602062, no_hdr0=''),
  4815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4815, CurrentID=0, Status='live', Symbol='NINJ2', Aliases='', description='ninjurin 2', other_designations='nerve injury-induced protein 2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=564296, end_position_on_the_genomic_accession=663741, orientation='minus', exon_count=6, OMIM=607297, no_hdr0=''),
  4817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4817, CurrentID=0, Status='live', Symbol='NIT1', Aliases='', description='nitrilase 1', other_designations='', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161118072, end_position_on_the_genomic_accession=161125445, orientation='plus', exon_count=8, OMIM=604618, no_hdr0=''),
  4818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4818, CurrentID=0, Status='live', Symbol='NKG7', Aliases='GIG1, GMP-17, p15-TIA-1', description='natural killer cell granule protein 7', other_designations='G-CSF-induced gene 1 protein|GIG-1 protein|granule membrane protein 17|granule membrane protein of 17 kDa|natural killer cell group 7 sequence|natural killer cell protein 7', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51371615, end_position_on_the_genomic_accession=51372728, orientation='minus', exon_count=4, OMIM=606008, no_hdr0=''),
  4819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4819, CurrentID=0, Status='live', Symbol='NKS1', Aliases='EC-1, EC1', description='natural killer cell susceptibility 1', other_designations='Susceptibility to lysis by alloreactive natural killer cells', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=272370, no_hdr0=''),
  4820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4820, CurrentID=0, Status='live', Symbol='NKTR', Aliases='p104', description='natural killer cell triggering receptor', other_designations='NK-TR protein|natural killer triggering receptor|natural killer-tumor recognition sequence|natural-killer cells cyclophilin-related protein', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42600655, end_position_on_the_genomic_accession=42648741, orientation='plus', exon_count=19, OMIM=161565, no_hdr0=''),
  4821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4821, CurrentID=0, Status='live', Symbol='NKX2-2', Aliases='NKX2.2, NKX2B', description='NK2 homeobox 2', other_designations='NK2 transcription factor related, locus 2|NK2 transcription factor-like protein B|homeobox protein NK-2 homolog B', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21499260, end_position_on_the_genomic_accession=21514026, orientation='minus', exon_count=6, OMIM=604612, no_hdr0=''),
  791254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791254, CurrentID=0, Status='live', Symbol='EDS8', Aliases='', description='Ehlers-Danlos syndrome, type VIII', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=130080, no_hdr0=''),
  791255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791255, CurrentID=0, Status='live', Symbol='SPDA2', Aliases='', description='Spondyloarthropathy, susceptibility to, 2', other_designations='', map_location='9q31-q34', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=183840, no_hdr0=''),
  4824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4824, CurrentID=0, Status='live', Symbol='NKX3-1', Aliases='BAPX2, NKX3, NKX3.1, NKX3A', description='NK3 homeobox 1', other_designations='NK homeobox, family 3, A|NK3 transcription factor homolog A|NK3 transcription factor related, locus 1|homeobox protein NK-3 homolog A', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23678693, end_position_on_the_genomic_accession=23682937, orientation='minus', exon_count=4, OMIM=602041, no_hdr0=''),
  4825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4825, CurrentID=0, Status='live', Symbol='NKX6-1', Aliases='NKX6.1, NKX6A', description='NK6 homeobox 1', other_designations='NK homeo box, family 6, member A|NK homeobox, family 6, A|NK6 transcription factor homolog A|NK6 transcription factor related, locus 1|homeobox protein NK-6 homolog A', map_location='4q21.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84491987, end_position_on_the_genomic_accession=84499141, orientation='minus', exon_count=4, OMIM=602563, no_hdr0=''),
  4826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4826, CurrentID=0, Status='live', Symbol='NNAT', Aliases='Peg5', description='neuronatin', other_designations='', map_location='20q11.2-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37521205, end_position_on_the_genomic_accession=37523690, orientation='plus', exon_count=3, OMIM=603106, no_hdr0=''),
  4827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4827, CurrentID=0, Status='live', Symbol='NM', Aliases='GP130, NCR', description='neutrophil migration', other_designations='Neutrophil migration (granulocyte glycoprotein)', map_location='7q22-qter', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=162820, no_hdr0=''),
  4828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4828, CurrentID=0, Status='live', Symbol='NMB', Aliases='', description='neuromedin B', other_designations='Neuromedin-B-32|neuromedin-beta', map_location='15q22-qter', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84655129, end_position_on_the_genomic_accession=84658571, orientation='minus', exon_count=3, OMIM=162340, no_hdr0=''),
  4829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4829, CurrentID=0, Status='live', Symbol='NMBR', Aliases='BB1', description='neuromedin B receptor', other_designations='NMB-R|bombesin receptor 1|epididymis tissue protein Li 185a|neuromedin-B-preferring bombesin receptor', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142074484, end_position_on_the_genomic_accession=142147118, orientation='minus', exon_count=6, OMIM=162341, no_hdr0=''),
  4830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4830, CurrentID=0, Status='live', Symbol='NME1', Aliases='AWD, GAAD, NB, NBS, NDKA, NDPK-A, NDPKA, NM23, NM23-H1', description='NME/NM23 nucleoside diphosphate kinase 1', other_designations='NDP kinase A|granzyme A-activated DNase|metastasis inhibition factor nm23|non-metastatic cells 1, protein (NM23A) expressed in|tumor metastatic process-associated protein', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51153559, end_position_on_the_genomic_accession=51162089, orientation='plus', exon_count=6, OMIM=156490, no_hdr0=''),
  4831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4831, CurrentID=0, Status='live', Symbol='NME2', Aliases='NDKB, NDPK-B, NDPKB, NM23-H2, NM23B, PUF', description='NME/NM23 nucleoside diphosphate kinase 2', other_designations='HEL-S-155an|NDP kinase B|c-myc purine-binding transcription factor PUF|c-myc transcription factor|epididymis secretory sperm binding protein Li 155an|histidine protein kinase NDKB|non-metastatic cells 2, protein (NM23) expressed in|non-metastatic cells 2, protein (NM23B) expressed in|nucleotide diphosphate kinase B', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51165435, end_position_on_the_genomic_accession=51171744, orientation='plus', exon_count=7, OMIM=156491, no_hdr0=''),
  4832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4832, CurrentID=0, Status='live', Symbol='NME3', Aliases='DR-nm23, NDPK-C, NDPKC, NM23-H3, NM23H3, c371H6.2', description='NME/NM23 nucleoside diphosphate kinase 3', other_designations='NDK 3|NDP kinase 3|NDP kinase C|non-metastatic cells 3, protein expressed in|nucleoside diphosphate kinase C', map_location='16q13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1770320, end_position_on_the_genomic_accession=1771758, orientation='minus', exon_count=5, OMIM=601817, no_hdr0=''),
  4833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4833, CurrentID=0, Status='live', Symbol='NME4', Aliases='NDPK-D, NM23H4, nm23-H4', description='NME/NM23 nucleoside diphosphate kinase 4', other_designations='NDK|NDP kinase D|NDP kinase, mitochondrial|NDPKD|non-metastatic cells 4, protein expressed in|nucleoside diphosphate kinase D', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=396755, end_position_on_the_genomic_accession=400754, orientation='plus', exon_count=8, OMIM=601818, no_hdr0=''),
  4835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4835, CurrentID=0, Status='live', Symbol='NQO2', Aliases='DHQV, DIA6, NMOR2, QR2', description='NAD(P)H dehydrogenase, quinone 2', other_designations='NAD(P)H menadione oxidoreductase-1, dioxin-inducible-2|NRH:quinone oxidoreductase 2|quinone reductase 2|ribosyldihydronicotinamide dehydrogenase', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2988640, end_position_on_the_genomic_accession=3019876, orientation='plus', exon_count=15, OMIM=160998, no_hdr0=''),
  4836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4836, CurrentID=0, Status='live', Symbol='NMT1', Aliases='NMT', description='N-myristoyltransferase 1', other_designations='alternative, short form NMT-S|long form, NMT-L|myristoyl-CoA:protein N-myristoyltransferase 1|type I N-myristoyltransferase', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45061283, end_position_on_the_genomic_accession=45109016, orientation='plus', exon_count=12, OMIM=160993, no_hdr0=''),
  4837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4837, CurrentID=0, Status='live', Symbol='NNMT', Aliases='', description='nicotinamide N-methyltransferase', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114295813, end_position_on_the_genomic_accession=114312516, orientation='plus', exon_count=3, OMIM=600008, no_hdr0=''),
  4838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4838, CurrentID=0, Status='live', Symbol='NODAL', Aliases='HTX5', description='nodal growth differentiation factor', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70431936, end_position_on_the_genomic_accession=70441709, orientation='minus', exon_count=3, OMIM=601265, no_hdr0=''),
  4839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4839, CurrentID=0, Status='live', Symbol='NOP2', Aliases='NOL1, NOP120, NSUN1, p120', description='NOP2 nucleolar protein', other_designations='NOL1/NOP2/Sun domain family, member 1|NOP2 nucleolar protein homolog|nucleolar protein 1, 120kDa|nucleolar protein 2 homolog|proliferating-cell nucleolar antigen p120|proliferation-associated nucleolar protein p120|putative ribosomal RNA methyltransferase NOP2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6556870, end_position_on_the_genomic_accession=6568332, orientation='minus', exon_count=18, OMIM=164031, no_hdr0=''),
  4841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4841, CurrentID=0, Status='live', Symbol='NONO', Aliases='NMT55, NRB54, P54, P54NRB, PPP1R114', description='non-POU domain containing, octamer-binding', other_designations='54 kDa nuclear RNA- and DNA-binding protein|55 kDa nuclear protein|DNA-binding p52/p100 complex, 52 kDa subunit|non-POU domain containing octamer-binding|non-POU domain-containing octamer (ATGCAAAT) binding protein|p54(nrb)|protein phosphatase 1, regulatory subunit 114', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71283192, end_position_on_the_genomic_accession=71301168, orientation='plus', exon_count=13, OMIM=300084, no_hdr0=''),
  4842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4842, CurrentID=0, Status='live', Symbol='NOS1', Aliases='IHPS1, N-NOS, NC-NOS, NOS, bNOS, nNOS', description='nitric oxide synthase 1', other_designations='NOS type I|constitutive NOS|neuronal NOS|nitric oxide synthase 1 (neuronal)|peptidyl-cysteine S-nitrosylase NOS1', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117208142, end_position_on_the_genomic_accession=117361802, orientation='minus', exon_count=35, OMIM=163731, no_hdr0=''),
  4843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4843, CurrentID=0, Status='live', Symbol='NOS2', Aliases='HEP-NOS, INOS, NOSA, NOS2', description='nitric oxide synthase 2', other_designations='NOS type II|NOS, type II|hepatocyte NOS|inducible NO synthase|inducible NOS|nitric oxide synthase 2, inducible|nitric oxide synthase 2A (inducible, hepatocytes)|nitric oxide synthase, macrophage|peptidyl-cysteine S-nitrosylase NOS2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27756766, end_position_on_the_genomic_accession=27804341, orientation='minus', exon_count=29, OMIM=163730, no_hdr0=''),
  4846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4846, CurrentID=0, Status='live', Symbol='NOS3', Aliases='ECNOS, eNOS', description='nitric oxide synthase 3', other_designations='EC-NOS|NOS type III|NOSIII|cNOS|constitutive NOS|endothelial NOS|nitric oxide synthase 3 (endothelial cell)|nitric oxide synthase 3 transcript variant eNOS-delta20|nitric oxide synthase 3 transcript variant eNOS-delta20-21|nitric oxide synthase 3 transcript variant eNOS-delta21', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150991056, end_position_on_the_genomic_accession=151014599, orientation='plus', exon_count=29, OMIM=163729, no_hdr0=''),
  4848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4848, CurrentID=0, Status='live', Symbol='CNOT2', Aliases='CDC36, HSPC131, NOT2, NOT2H', description='CCR4-NOT transcription complex subunit 2', other_designations='CCR4-NOT transcription complex, subunit 2|CCR4-associated factor 2|negative regulator of transcription 2', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70242994, end_position_on_the_genomic_accession=70354993, orientation='plus', exon_count=21, OMIM=604909, no_hdr0=''),
  4849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4849, CurrentID=0, Status='live', Symbol='CNOT3', Aliases='LENG2, NOT3, NOT3H', description='CCR4-NOT transcription complex subunit 3', other_designations='CCR4-NOT transcription complex, subunit 3|CCR4-associated factor 3|NOT3 (negative regulator of transcription 3, yeast) homolog|leukocyte receptor cluster member 2', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54137689, end_position_on_the_genomic_accession=54155708, orientation='plus', exon_count=20, OMIM=604910, no_hdr0=''),
  4850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4850, CurrentID=0, Status='live', Symbol='CNOT4', Aliases='CLONE243, NOT4, NOT4H', description='CCR4-NOT transcription complex subunit 4', other_designations='CCR4-NOT transcription complex, subunit 4|CCR4-associated factor 4|E3 ubiquitin-protein ligase CNOT4|NOT4 (negative regulator of transcription 4, yeast) homolog|potential transcriptional repressor NOT4Hp', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135361795, end_position_on_the_genomic_accession=135510127, orientation='minus', exon_count=15, OMIM=604911, no_hdr0=''),
  4851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4851, CurrentID=0, Status='live', Symbol='NOTCH1', Aliases='AOS5, AOVD1, TAN1, hN1', description='notch 1', other_designations='Notch homolog 1, translocation-associated|translocation-associated notch protein TAN-1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136494444, end_position_on_the_genomic_accession=136545786, orientation='minus', exon_count=34, OMIM=190198, no_hdr0=''),
  4852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4852, CurrentID=0, Status='live', Symbol='NPY', Aliases='PYY4', description='neuropeptide Y', other_designations='prepro-neuropeptide Y', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24284188, end_position_on_the_genomic_accession=24291865, orientation='plus', exon_count=4, OMIM=162640, no_hdr0=''),
  4853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4853, CurrentID=0, Status='live', Symbol='NOTCH2', Aliases='AGS2, HJCYS, hN2', description='notch 2', other_designations='Notch homolog 2', map_location='1p13-p11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119911553, end_position_on_the_genomic_accession=120069703, orientation='minus', exon_count=37, OMIM=600275, no_hdr0=''),
  4854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4854, CurrentID=0, Status='live', Symbol='NOTCH3', Aliases='CADASIL, CASIL, IMF2, LMNS', description='notch 3', other_designations='Notch homolog 3', map_location='19p13.2-p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15159633, end_position_on_the_genomic_accession=15200981, orientation='minus', exon_count=33, OMIM=600276, no_hdr0=''),
  4855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4855, CurrentID=0, Status='live', Symbol='NOTCH4', Aliases='INT3', description='notch 4', other_designations='Notch homolog 4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32194843, end_position_on_the_genomic_accession=32224067, orientation='minus', exon_count=30, OMIM=164951, no_hdr0=''),
  4856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4856, CurrentID=0, Status='live', Symbol='NOV', Aliases='CCN3, IBP-9, IGFBP-9, IGFBP9h, NOV', description='nephroblastoma overexpressed', other_designations='CCN family member 3|IGF-binding protein 9|insulin-like growth factor-binding protein 9|nephroblastoma-overexpressed gene protein homolog', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119416312, end_position_on_the_genomic_accession=119424438, orientation='plus', exon_count=5, OMIM=164958, no_hdr0=''),
  4857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4857, CurrentID=0, Status='live', Symbol='NOVA1', Aliases='Nova-1', description='neuro-oncological ventral antigen 1', other_designations='onconeural ventral antigen 1|paraneoplastic Ri antigen|ventral neuron-specific protein 1', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26445883, end_position_on_the_genomic_accession=26597754, orientation='minus', exon_count=7, OMIM=602157, no_hdr0=''),
  4858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4858, CurrentID=0, Status='live', Symbol='NOVA2', Aliases='ANOVA, NOVA3', description='neuro-oncological ventral antigen 2', other_designations='astrocytic NOVA1-like RNA-binding protein|neuro-oncological ventral antigen 3', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45933734, end_position_on_the_genomic_accession=45973399, orientation='minus', exon_count=8, OMIM=601991, no_hdr0=''),
  4860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4860, CurrentID=0, Status='live', Symbol='PNP', Aliases='NP, PRO1837, PUNP', description='purine nucleoside phosphorylase', other_designations='HEL-S-156an|epididymis secretory sperm binding protein Li 156an|inosine phosphorylase|inosine-guanosine phosphorylase|purine-nucleoside:orthophosphate ribosyltransferase', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20469379, end_position_on_the_genomic_accession=20478006, orientation='plus', exon_count=6, OMIM=164050, no_hdr0=''),
  4861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4861, CurrentID=0, Status='live', Symbol='NPAS1', Aliases='MOP5, PASD5, bHLHe11', description='neuronal PAS domain protein 1', other_designations='PAS domain-containing protein 5|basic-helix-loop-helix-PAS protein MOP5|class E basic helix-loop-helix protein 11|member of PAS protein 5|member of PAS superfamily 5|neuronal PAS1', map_location='19q13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47019820, end_position_on_the_genomic_accession=47045775, orientation='plus', exon_count=14, OMIM=603346, no_hdr0=''),
  4862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4862, CurrentID=0, Status='live', Symbol='NPAS2', Aliases='MOP4, PASD4, bHLHe9', description='neuronal PAS domain protein 2', other_designations='PAS domain-containing protein 4|basic-helix-loop-helix-PAS protein MOP4|class E basic helix-loop-helix protein 9|member of PAS protein 4|member of PAS superfamily 4|neuronal PAS2', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100820151, end_position_on_the_genomic_accession=100996827, orientation='plus', exon_count=25, OMIM=603347, no_hdr0=''),
  4863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4863, CurrentID=0, Status='live', Symbol='NPAT', Aliases='E14, E14/NPAT, p220', description='nuclear protein, ataxia-telangiectasia locus', other_designations='E14|nuclear protein of the ATM locus|nuclear protein of the ataxia telangiectasia mutated locus', map_location='11q22-q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108157215, end_position_on_the_genomic_accession=108222638, orientation='minus', exon_count=18, OMIM=601448, no_hdr0=''),
  4864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4864, CurrentID=0, Status='live', Symbol='NPC1', Aliases='NPC', description='Niemann-Pick disease, type C1', other_designations='truncated Niemann-Pick C1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23506184, end_position_on_the_genomic_accession=23586617, orientation='minus', exon_count=28, OMIM=607623, no_hdr0=''),
  267009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267009, CurrentID=0, Status='live', Symbol='RPS9P2', Aliases='RPS9_3_1753, bA12M9.1', description='ribosomal protein S9 pseudogene 2', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41063996, end_position_on_the_genomic_accession=41064705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  267010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267010, CurrentID=0, Status='live', Symbol='RNU12', Aliases='RNU12-1L, RNU12P, dJ222E13.7, RNU12', description='RNA, U12 small nuclear', other_designations='RNA, U12 small nuclear 1|RNA, U12 small nuclear pseudogene', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42615245, end_position_on_the_genomic_accession=42615393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4867, CurrentID=0, Status='live', Symbol='NPHP1', Aliases='JBTS4, NPH1, SLSN1', description='nephronophthisis 1 (juvenile)', other_designations='juvenile nephronophthisis 1 protein|nephrocystin 1', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110123336, end_position_on_the_genomic_accession=110205062, orientation='minus', exon_count=23, OMIM=607100, no_hdr0=''),
  4868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4868, CurrentID=0, Status='live', Symbol='NPHS1', Aliases='CNF, NPHN, nephrin', description='nephrosis 1, congenital, Finnish type (nephrin)', other_designations='renal glomerulus-specific cell adhesion receptor', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35825372, end_position_on_the_genomic_accession=35851993, orientation='minus', exon_count=29, OMIM=602716, no_hdr0=''),
  4869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4869, CurrentID=0, Status='live', Symbol='NPM1', Aliases='B23, NPM', description='nucleophosmin (nucleolar phosphoprotein B23, numatrin)', other_designations='nucleolar protein NO38|nucleophosmin/nucleoplasmin family, member 1', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171387648, end_position_on_the_genomic_accession=171410884, orientation='plus', exon_count=12, OMIM=164040, no_hdr0=''),
  267014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267014, CurrentID=0, Status='live', Symbol='HLA-N', Aliases='HLA-30', description='major histocompatibility complex, class I, N (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30351074, end_position_on_the_genomic_accession=30352038, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4871, CurrentID=0, Status='live', Symbol='NPM1P2', Aliases='NG2-6, NPMP2', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 2', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 2', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26204780, end_position_on_the_genomic_accession=26206543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  267016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267016, CurrentID=0, Status='live', Symbol='HLA-X', Aliases='', description='major histocompatibility complex, class I, X (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31461846, end_position_on_the_genomic_accession=31462490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  267017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267017, CurrentID=0, Status='live', Symbol='HLA-Z', Aliases='HLA-Z1', description='major histocompatibility complex, class I, Z (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32896402, end_position_on_the_genomic_accession=32896489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  135946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135946, CurrentID=0, Status='live', Symbol='OR6B1', Aliases='OR7-3, OR7-9', description='olfactory receptor family 6 subfamily B member 1', other_designations='olfactory receptor 7-3|olfactory receptor OR7-9|olfactory receptor, family 6, subfamily B, member 1', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144003997, end_position_on_the_genomic_accession=144004932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4875, CurrentID=0, Status='live', Symbol='NPM1P6', Aliases='NG6-4, NPMP6', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 6', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 6', map_location='8q12.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61202255, end_position_on_the_genomic_accession=61203545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  267020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267020, CurrentID=0, Status='live', Symbol='ATP5L2', Aliases='ATP5K2', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit G2 pseudogene|ATP synthase, H+ transporting, mitochondrial F1F0, subunit g|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit G2|ATPase subunit g 2', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42639803, end_position_on_the_genomic_accession=42640601, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4877, CurrentID=0, Status='live', Symbol='NT3', Aliases='', description="3'-nucleotidase", other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4878, CurrentID=0, Status='live', Symbol='NPPA', Aliases='ANF, ANP, ATFB6, ATRST2, CDD, CDD-ANF, CDP, PND', description='natriuretic peptide A', other_designations='atriopeptin|cardiodilatin-related peptide|cardionatrin|natriuretic peptide precursor A variant 1|prepronatriodilatin', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11845709, end_position_on_the_genomic_accession=11847783, orientation='minus', exon_count=3, OMIM=108780, no_hdr0=''),
  4879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4879, CurrentID=0, Status='live', Symbol='NPPB', Aliases='BNP', description='natriuretic peptide B', other_designations='brain type natriuretic peptide|gamma-brain natriuretic peptide|natriuretic peptide precursor B|natriuretic protein', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11857464, end_position_on_the_genomic_accession=11858935, orientation='minus', exon_count=3, OMIM=600295, no_hdr0=''),
  4880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4880, CurrentID=0, Status='live', Symbol='NPPC', Aliases='CNP, CNP2', description='natriuretic peptide C', other_designations='natriuretic peptide precursor type C', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231922095, end_position_on_the_genomic_accession=231926389, orientation='minus', exon_count=4, OMIM=600296, no_hdr0=''),
  4881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4881, CurrentID=0, Status='live', Symbol='NPR1', Aliases='ANPRA, ANPa, GUC2A, GUCY2A, NPRA', description='natriuretic peptide receptor 1', other_designations='ANP-A|ANPR-A|GC-A|NPR-A|atrial natriuretic peptide receptor type A|atrionatriuretic peptide receptor A|guanylate cyclase A|natriuretic peptide A type receptor|natriuretic peptide receptor A|natriuretic peptide receptor A/guanylate cyclase A (atrionatriuretic peptide receptor A)', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153678617, end_position_on_the_genomic_accession=153693992, orientation='plus', exon_count=22, OMIM=108960, no_hdr0=''),
  4882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4882, CurrentID=0, Status='live', Symbol='NPR2', Aliases='AMDM, ANPRB, ANPb, ECDM, GUC2B, GUCY2B, NPRB, NPRBi, SNSK', description='natriuretic peptide receptor 2', other_designations='ANP-B|ANPR-B|GC-B|NPR-B|atrial natriuretic peptide B-type receptor|atrial natriuretic peptide receptor type B|guanylate cyclase B|natriuretic peptide receptor B/guanylate cyclase B (atrionatriuretic peptide receptor B)', map_location='9p21-p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35752945, end_position_on_the_genomic_accession=35809731, orientation='plus', exon_count=27, OMIM=108961, no_hdr0=''),
  4883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4883, CurrentID=0, Status='live', Symbol='NPR3', Aliases='ANP-C, ANPR-C, ANPRC, C5orf23, GUCY2B, NPR-C, NPRC', description='natriuretic peptide receptor 3', other_designations='atrial natriuretic peptide clearance receptor|atrial natriuretic peptide receptor type C|atrionatriuretic peptide receptor C|guanylate cyclase C|natriuretic peptide receptor C/guanylate cyclase C (atrionatriuretic peptide receptor C)', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32689076, end_position_on_the_genomic_accession=32791724, orientation='plus', exon_count=14, OMIM=108962, no_hdr0=''),
  4884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4884, CurrentID=0, Status='live', Symbol='NPTX1', Aliases='NP1', description='neuronal pentraxin I', other_designations='NP-I', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80466833, end_position_on_the_genomic_accession=80476604, orientation='minus', exon_count=5, OMIM=602367, no_hdr0=''),
  4885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4885, CurrentID=0, Status='live', Symbol='NPTX2', Aliases='NARP, NP-II, NP2', description='neuronal pentraxin II', other_designations='apexin|neuronal activity-regulated pentaxin', map_location='7q21.3-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98617285, end_position_on_the_genomic_accession=98629869, orientation='plus', exon_count=5, OMIM=600750, no_hdr0=''),
  4886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4886, CurrentID=0, Status='live', Symbol='NPY1R', Aliases='NPY1-R, NPYR', description='neuropeptide Y receptor Y1', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163323962, end_position_on_the_genomic_accession=163344832, orientation='minus', exon_count=5, OMIM=162641, no_hdr0=''),
  4887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4887, CurrentID=0, Status='live', Symbol='NPY2R', Aliases='NPY2-R', description='neuropeptide Y receptor Y2', other_designations='NPY-Y2 receptor', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155173754, end_position_on_the_genomic_accession=155217096, orientation='plus', exon_count=3, OMIM=162642, no_hdr0=''),
  4888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4888, CurrentID=0, Status='live', Symbol='NPY6R', Aliases='NPY1RLP, PP2, Y2B, NPY6R', description='neuropeptide Y receptor Y6 (pseudogene)', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137801193, end_position_on_the_genomic_accession=137810751, orientation='plus', exon_count=2, OMIM=601770, no_hdr0=''),
  4889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4889, CurrentID=0, Status='live', Symbol='NPY5R', Aliases='NPY5-R, NPYR5, NPYY5-R', description='neuropeptide Y receptor Y5', other_designations='NPY-Y5 receptor|Y5 receptor', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163343868, end_position_on_the_genomic_accession=163352403, orientation='plus', exon_count=8, OMIM=602001, no_hdr0=''),
  4891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4891, CurrentID=0, Status='live', Symbol='SLC11A2', Aliases='DCT1, DMT1, NRAMP2', description='solute carrier family 11 (proton-coupled divalent metal ion transporter), member 2', other_designations='DMT-1|NRAMP 2|divalent cation transporter 1|solute carrier family 11 (proton-coupled divalent metal ion transporters), member 2|solute carrier family 11 member 2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50979410, end_position_on_the_genomic_accession=51028335, orientation='minus', exon_count=24, OMIM=600523, no_hdr0=''),
  4892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4892, CurrentID=0, Status='live', Symbol='NRAP', Aliases='N-RAP', description='nebulin related anchoring protein', other_designations='nebulin-related anchoring protein', map_location='10q24-q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113588824, end_position_on_the_genomic_accession=113664070, orientation='minus', exon_count=42, OMIM=602873, no_hdr0=''),
  4893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4893, CurrentID=0, Status='live', Symbol='NRAS', Aliases='ALPS4, CMNS, N-ras, NCMS1, NS6, NRAS', description='neuroblastoma RAS viral (v-ras) oncogene homolog', other_designations='N-ras protein part 4|transforming protein N-Ras|v-ras neuroblastoma RAS viral oncogene homolog', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114704464, end_position_on_the_genomic_accession=114716894, orientation='minus', exon_count=7, OMIM=164790, no_hdr0=''),
  284805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284805, CurrentID=0, Status='live', Symbol='C20orf203', Aliases='', description='chromosome 20 open reading frame 203', other_designations='alugen', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32631625, end_position_on_the_genomic_accession=32651981, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  4897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4897, CurrentID=0, Status='live', Symbol='NRCAM', Aliases='', description='neuronal cell adhesion molecule', other_designations='NgCAM-related cell adhesion molecule|neuronal surface protein Bravo', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108147626, end_position_on_the_genomic_accession=108456397, orientation='minus', exon_count=37, OMIM=601581, no_hdr0=''),
  4898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4898, CurrentID=0, Status='live', Symbol='NRDC', Aliases='NRD1, hNRD1, hNRD2', description='nardilysin convertase', other_designations='NRD convertase|NRD-C|nardilysin (N-arginine dibasic convertase)', map_location='1p32.2-p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51789192, end_position_on_the_genomic_accession=51878937, orientation='minus', exon_count=36, OMIM=602651, no_hdr0=''),
  4899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4899, CurrentID=0, Status='live', Symbol='NRF1', Aliases='ALPHA-PAL', description='nuclear respiratory factor 1', other_designations='alpha palindromic-binding protein', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129611702, end_position_on_the_genomic_accession=129757082, orientation='plus', exon_count=13, OMIM=600879, no_hdr0=''),
  4900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4900, CurrentID=0, Status='live', Symbol='NRGN', Aliases='RC3, hng', description='neurogranin', other_designations='calmodulin-binding protein|neurogranin (protein kinase C substrate, RC3)|ng|protein kinase C substrate', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124739933, end_position_on_the_genomic_accession=124747206, orientation='plus', exon_count=4, OMIM=602350, no_hdr0=''),
  4901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4901, CurrentID=0, Status='live', Symbol='NRL', Aliases='D14S46E-MAF, RP27, NRL', description='neural retina leucine zipper', other_designations='neural retinal-specific leucine zipper', map_location='14q11.1-q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24078693, end_position_on_the_genomic_accession=24114802, orientation='minus', exon_count=8, OMIM=162080, no_hdr0=''),
  4902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4902, CurrentID=0, Status='live', Symbol='NRTN', Aliases='NTN', description='neurturin', other_designations='prepro-neurturin', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5805359, end_position_on_the_genomic_accession=5828324, orientation='plus', exon_count=3, OMIM=602018, no_hdr0=''),
  100270897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270897, CurrentID=0, Status='live', Symbol='RPL7AP17', Aliases='RPL7A_3_78', description='ribosomal protein L7a pseudogene 17', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100586604, end_position_on_the_genomic_accession=100587460, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4904, CurrentID=0, Status='live', Symbol='YBX1', Aliases='BP-8, CBF-A, CSDA2, CSDB, DBPB, EFI-A, MDR-NF1, NSEP-1, NSEP1, YB-1, YB1', description='Y-box binding protein 1', other_designations='CCAAT-binding transcription factor I subunit A|DNA-binding protein B|Y box binding protein 1|Y-box transcription factor|Y-box-binding protein 1|enhancer factor I subunit A|major histocompatibility complex, class II, Y box-binding protein I|nuclease sensitive element binding protein 1', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42682389, end_position_on_the_genomic_accession=42702363, orientation='plus', exon_count=8, OMIM=154030, no_hdr0=''),
  4905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4905, CurrentID=0, Status='live', Symbol='NSF', Aliases='SKD2', description='N-ethylmaleimide sensitive factor', other_designations='N-ethylmaleimide-sensitive factor|N-ethylmaleimide-sensitive factor-like protein|N-ethylmaleimide-sensitive fusion protein|NEM-sensitive fusion protein|vesicular-fusion protein NSF', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46590669, end_position_on_the_genomic_accession=46757464, orientation='plus', exon_count=22, OMIM=601633, no_hdr0=''),
  4907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4907, CurrentID=0, Status='live', Symbol='NT5E', Aliases='CALJA, CD73, E5NT, NT, NT5, NTE, eN, eNT', description="5'-nucleotidase, ecto (CD73)", other_designations="5'-NT|Purine 5-Prime-Nucleotidase|ecto-5'-nucleotidase", map_location='6q14-q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85449584, end_position_on_the_genomic_accession=85495791, orientation='plus', exon_count=9, OMIM=129190, no_hdr0=''),
  4908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4908, CurrentID=0, Status='live', Symbol='NTF3', Aliases='HDNF, NGF-2, NGF2, NT-3, NT3', description='neurotrophin 3', other_designations='nerve growth factor 2|neurotrophic factor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5432114, end_position_on_the_genomic_accession=5495299, orientation='plus', exon_count=3, OMIM=162660, no_hdr0=''),
  4909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4909, CurrentID=0, Status='live', Symbol='NTF4', Aliases='GLC10, GLC1O, NT-4, NT-4/5, NT-5, NT4, NT5, NTF5', description='neurotrophin 4', other_designations='neurotrophic factor 4|neurotrophic factor 5|neurotrophin 5 (neurotrophin 4/5)|neurotrophin-5|neutrophic factor 4', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49061080, end_position_on_the_genomic_accession=49065076, orientation='minus', exon_count=6, OMIM=162662, no_hdr0=''),
  4910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4910, CurrentID=0, Status='live', Symbol='NTF6A', Aliases='NTF5P, NTF6, PSI-NT4', description='neurotrophin 6 alpha (pseudogene)', other_designations='NT4 pseudogene|neurotrophin 5 pseudogene', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49025225, end_position_on_the_genomic_accession=49026122, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4911, CurrentID=0, Status='live', Symbol='NTF6B', Aliases='', description='neurotrophin 6 beta (pseudogene)', other_designations='NT6-beta|acidic neurotrophin 6 beta', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49038234, end_position_on_the_genomic_accession=49039138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4912, CurrentID=0, Status='live', Symbol='NTF6G', Aliases='', description='neurotrophin 6 gamma (pseudogene)', other_designations='NT6-gamma', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49029691, end_position_on_the_genomic_accession=49030585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4913, CurrentID=0, Status='live', Symbol='NTHL1', Aliases='FAP3, NTH1, OCTS3, hNTH1', description='nth-like DNA glycosylase 1', other_designations='DNA glycoslyase/AP lyase|bifunctional DNA N-glycoslyase/DNA-(apurinic or apyrimidinic site) lyase|nth endonuclease III-like 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2039815, end_position_on_the_genomic_accession=2047866, orientation='minus', exon_count=6, OMIM=602656, no_hdr0=''),
  4914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4914, CurrentID=0, Status='live', Symbol='NTRK1', Aliases='MTC, TRK, TRK1, TRKA, Trk-A, p140-TrkA', description='neurotrophic tyrosine kinase, receptor, type 1', other_designations='Oncogene TRK|TRK1-transforming tyrosine kinase protein|gp140trk|tropomyosin-related kinase A|tyrosine kinase receptor A', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156815750, end_position_on_the_genomic_accession=156881850, orientation='plus', exon_count=19, OMIM=191315, no_hdr0=''),
  4915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4915, CurrentID=0, Status='live', Symbol='NTRK2', Aliases='GP145-TrkB, TRKB, trk-B', description='neurotrophic tyrosine kinase, receptor, type 2', other_designations='BDNF-tropomyosine receptor kinase B|tropomyosin-related kinase B|tyrosine kinase receptor B', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84668458, end_position_on_the_genomic_accession=85027070, orientation='plus', exon_count=23, OMIM=600456, no_hdr0=''),
  4916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4916, CurrentID=0, Status='live', Symbol='NTRK3', Aliases='GP145-TrkC, TRKC, gp145(trkC)', description='neurotrophic tyrosine kinase, receptor, type 3', other_designations='ETS related protein-neurotrophic receptor tyrosine kinase fusion protein|ETV6-NTRK3 fusion|tyrosine kinase receptor C', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=87874999, end_position_on_the_genomic_accession=88256796, orientation='minus', exon_count=31, OMIM=191316, no_hdr0=''),
  4917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4917, CurrentID=0, Status='live', Symbol='NTN3', Aliases='NTN2L', description='netrin 3', other_designations='Netrin-2, chicken, homolog of, like|netrin-2-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2471499, end_position_on_the_genomic_accession=2474145, orientation='plus', exon_count=6, OMIM=602349, no_hdr0=''),
  4919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4919, CurrentID=0, Status='live', Symbol='ROR1', Aliases='NTRKR1, dJ537F10.1', description='receptor tyrosine kinase-like orphan receptor 1', other_designations='neurotrophic tyrosine kinase, receptor-related 1|tyrosine-protein kinase transmembrane receptor ROR1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63774019, end_position_on_the_genomic_accession=64181498, orientation='plus', exon_count=10, OMIM=602336, no_hdr0=''),
  4920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4920, CurrentID=0, Status='live', Symbol='ROR2', Aliases='BDB, BDB1, NTRKR2', description='receptor tyrosine kinase-like orphan receptor 2', other_designations='neurotrophic tyrosine kinase receptor-related 2', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91722596, end_position_on_the_genomic_accession=91950162, orientation='minus', exon_count=14, OMIM=602337, no_hdr0=''),
  4921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4921, CurrentID=0, Status='live', Symbol='DDR2', Aliases='MIG20a, NTRKR3, TKT, TYRO10', description='discoidin domain receptor tyrosine kinase 2', other_designations='CD167 antigen-like family member B|cell migration-inducing protein 20|discoidin domain receptor 2|discoidin domain receptor family, member 2|discoidin domain-containing receptor tyrosine kinase 2|hydroxyaryl-protein kinase|migration-inducing gene 16 protein|neurotrophic tyrosine kinase receptor related 3|neurotrophic tyrosine kinase, receptor-related 3|receptor protein-tyrosine kinase TKT|tyrosine-protein kinase TYRO10|tyrosylprotein kinase', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162631265, end_position_on_the_genomic_accession=162786591, orientation='plus', exon_count=22, OMIM=191311, no_hdr0=''),
  4922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4922, CurrentID=0, Status='live', Symbol='NTS', Aliases='NMN-125, NN, NT, NT/N1, NTS', description='neurotensin', other_designations='pro-neurotensin/neuromedin', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85874295, end_position_on_the_genomic_accession=85882992, orientation='plus', exon_count=4, OMIM=162650, no_hdr0=''),
  4923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4923, CurrentID=0, Status='live', Symbol='NTSR1', Aliases='NTR', description='neurotensin receptor 1 (high affinity)', other_designations='NT-R-1|NTR1|NTRH|high-affinity levocabastine-insensitive neurotensin receptor', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62708783, end_position_on_the_genomic_accession=62762771, orientation='plus', exon_count=4, OMIM=162651, no_hdr0=''),
  4924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4924, CurrentID=0, Status='live', Symbol='NUCB1', Aliases='CALNUC, NUC', description='nucleobindin 1', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48900050, end_position_on_the_genomic_accession=48923283, orientation='plus', exon_count=13, OMIM=601323, no_hdr0=''),
  4925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4925, CurrentID=0, Status='live', Symbol='NUCB2', Aliases='HEL-S-109, NEFA', description='nucleobindin 2', other_designations='DNA-binding protein NEFA|epididymis secretory protein Li 109|gastric cancer antigen Zg4|nesfatin 1|nesfatin-1|nucleobinding 2|prepronefastin|prepronesfatin', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17260337, end_position_on_the_genomic_accession=17349980, orientation='plus', exon_count=23, OMIM=608020, no_hdr0=''),
  4926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4926, CurrentID=0, Status='live', Symbol='NUMA1', Aliases='NMP-22, NUMA', description='nuclear mitotic apparatus protein 1', other_designations='SP-H antigen|centrophilin stabilizes mitotic spindle in mitotic cells|nuclear matrix protein-22|structural nuclear protein', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72002864, end_position_on_the_genomic_accession=72080693, orientation='minus', exon_count=38, OMIM=164009, no_hdr0=''),
  4927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4927, CurrentID=0, Status='live', Symbol='NUP88', Aliases='', description='nucleoporin 88kDa', other_designations='88 kDa nucleoporin|karyoporin|nuclear pore complex protein 88|nucleoporin 88kD|nucleoporin Nup88', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5386026, end_position_on_the_genomic_accession=5419739, orientation='minus', exon_count=18, OMIM=602552, no_hdr0=''),
  4928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4928, CurrentID=0, Status='live', Symbol='NUP98', Aliases='ADIR2, NUP196, NUP96', description='nucleoporin 98kDa', other_designations='GLFG-repeat containing nucleoporin|Nup98-Nup96|nucleoporin 98kD', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3675010, end_position_on_the_genomic_accession=3797792, orientation='minus', exon_count=35, OMIM=601021, no_hdr0=''),
  4929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4929, CurrentID=0, Status='live', Symbol='NR4A2', Aliases='HZF-3, NOT, NURR1, RNR1, TINUR', description='nuclear receptor subfamily 4 group A member 2', other_designations='NGFI-B/nur77 beta-type transcription factor homolog|T-cell nuclear receptor NOT|immediate-early response protein NOT|intermediate-early receptor protein|nuclear receptor related 1|nuclear receptor subfamily 4, group A, member 2|nur related protein-1, human homolog of|orphan nuclear receptor NR4A2|orphan nuclear receptor NURR1|transcriptionally inducible nuclear receptor related 1|transcriptionally-inducible nuclear receptor', map_location='2q22-q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156324432, end_position_on_the_genomic_accession=156334683, orientation='minus', exon_count=8, OMIM=601828, no_hdr0=''),
  4931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4931, CurrentID=0, Status='live', Symbol='NVL', Aliases='', description='nuclear VCP-like', other_designations='NVLp', map_location='1q41-q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224227334, end_position_on_the_genomic_accession=224330201, orientation='minus', exon_count=27, OMIM=602426, no_hdr0=''),
  4933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4933, CurrentID=0, Status='live', Symbol='NYS2', Aliases='NYSA', description='nystagmus 2, congenital autosomal dominant', other_designations='', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164100, no_hdr0=''),
  4934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4934, CurrentID=0, Status='live', Symbol='NYS3', Aliases='CN', description='nystagmus 3, congenital autosomal dominant', other_designations='', map_location='7p11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608345, no_hdr0=''),
  4935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4935, CurrentID=0, Status='live', Symbol='GPR143', Aliases='NYS6, OA1', description='G protein-coupled receptor 143', other_designations='ocular albinism 1|ocular albinism type 1 protein', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9725413, end_position_on_the_genomic_accession=9765965, orientation='minus', exon_count=10, OMIM=300808, no_hdr0=''),
  4936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4936, CurrentID=0, Status='live', Symbol='OA3', Aliases='OAR', description='ocular albinism 3 (autosomal recessive)', other_designations='Ocular albinism, autosomal recessive', map_location='6q13-q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4937, CurrentID=0, Status='live', Symbol='OAP', Aliases='OA, OASF', description='osteoarthrosis, precocious', other_designations='', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4938, CurrentID=0, Status='live', Symbol='OAS1', Aliases='IFI-4, OIAS, OIASI', description="2'-5'-oligoadenylate synthetase 1", other_designations="(2-5')oligo(A) synthase 1|(2-5')oligo(A) synthetase 1|2',5'-oligo A synthetase 1|2',5'-oligoadenylate synthetase 1, 40/46kDa|2'-5' oligoadenylate synthetase 1 p41 isoform|2'-5' oligoadenylate synthetase 1 p48 isoform|2'-5' oligoadenylate synthetase 1 p49 isoform|2'-5' oligoadenylate synthetase 1 p52 isoform|2'-5'-oligoadenylate synthetase 1, 40/46kDa|2'-5'-oligoisoadenylate synthetase 1|2-5A synthase 1|2-5A synthetase 1|E18/E16|p46/p42 OAS", map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112906934, end_position_on_the_genomic_accession=112932049, orientation='plus', exon_count=8, OMIM=164350, no_hdr0=''),
  4939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4939, CurrentID=0, Status='live', Symbol='OAS2', Aliases='', description="2'-5'-oligoadenylate synthetase 2", other_designations="(2'-5')oligo(A) synthetase 2|2'-5'-oligoadenylate synthetase 2, 69/71kDa|2-5A synthase 2|p69 OAS / p71 OAS", map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112970819, end_position_on_the_genomic_accession=113011723, orientation='plus', exon_count=13, OMIM=603350, no_hdr0=''),
  4940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4940, CurrentID=0, Status='live', Symbol='OAS3', Aliases='p100, p100OAS', description="2'-5'-oligoadenylate synthetase 3", other_designations="(2-5')oligo(A) synthase 3|(2-5')oligo(A) synthetase 3|2'-5'-oligoadenylate synthetase 3, 100kDa|2'-5'oligoadenylate synthetase p100|2-5A synthase 3|2-5A synthetase 3|p100 OAS", map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112938433, end_position_on_the_genomic_accession=112973249, orientation='plus', exon_count=16, OMIM=603351, no_hdr0=''),
  4941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4941, CurrentID=0, Status='live', Symbol='OASD', Aliases='', description='ocular albinism and sensorineural deafness', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300650, no_hdr0=''),
  4942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4942, CurrentID=0, Status='live', Symbol='OAT', Aliases='GACR, HOGAASE, OKT, OAT', description='ornithine aminotransferase', other_designations='gyrate atrophy|ornithine delta-aminotransferase|ornithine-oxo-acid aminotransferase', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124397303, end_position_on_the_genomic_accession=124418976, orientation='minus', exon_count=12, OMIM=613349, no_hdr0=''),
  4943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4943, CurrentID=0, Status='live', Symbol='TBC1D25', Aliases='MG81, OATL1', description='TBC1 domain family member 25', other_designations='5SN3 snoRNA|TBC1 domain family, member 25|ornithine aminotransferase-like 1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48539687, end_position_on_the_genomic_accession=48562609, orientation='plus', exon_count=6, OMIM=311240, no_hdr0=''),
  100422627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422627, CurrentID=0, Status='live', Symbol='LOC100422627', Aliases='', description='TATA-box binding protein associated factor 4b pseudogene', other_designations='TAF4b RNA polymerase II, TATA box binding protein (TBP)-associated factor, 105kDa pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112826302, end_position_on_the_genomic_accession=112827723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4945, CurrentID=0, Status='live', Symbol='OATP1', Aliases='OATL3', description='ornithine aminotransferase pseudogene 1', other_designations='ornithine aminotransferase-like 3 (pseudogene)', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124443784, end_position_on_the_genomic_accession=124444243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4946, CurrentID=0, Status='live', Symbol='OAZ1', Aliases='AZI, OAZ', description='ornithine decarboxylase antizyme 1', other_designations='ODC-Az|antizyme 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2269486, end_position_on_the_genomic_accession=2273488, orientation='plus', exon_count=5, OMIM=601579, no_hdr0=''),
  4947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4947, CurrentID=0, Status='live', Symbol='OAZ2', Aliases='AZ2', description='ornithine decarboxylase antizyme 2', other_designations='ODC-Az 2', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64687574, end_position_on_the_genomic_accession=64703281, orientation='minus', exon_count=5, OMIM=604152, no_hdr0=''),
  4948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4948, CurrentID=0, Status='live', Symbol='OCA2', Aliases='BEY, BEY1, BEY2, BOCA, D15S12, EYCL, EYCL2, EYCL3, HCL3, P, PED, SHEP1', description='oculocutaneous albinism II', other_designations='P-protein|eye color 2 (central brown)|eye color 3 (brown)|hair color 3 (brown)|melanocyte-specific transporter protein|oculocutaneous albinism II (pink-eye dilution homolog, mouse)|pink-eyed dilution protein homolog|total brown iris pigmentation', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27754873, end_position_on_the_genomic_accession=28099340, orientation='minus', exon_count=29, OMIM=611409, no_hdr0=''),
  4951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4951, CurrentID=0, Status='live', Symbol='OCM2', Aliases='OCM', description='oncomodulin 2', other_designations='OM|parvalbumin beta', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97984700, end_position_on_the_genomic_accession=97990104, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4952, CurrentID=0, Status='live', Symbol='OCRL', Aliases='INPP5F, LOCR, NPHL2-1, OCRL1, OCRL', description='oculocerebrorenal syndrome of Lowe', other_designations='Lowe oculocerebrorenal syndrome protein|phosphatidylinositol polyphosphate 5-phosphatase', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129532740, end_position_on_the_genomic_accession=129592556, orientation='plus', exon_count=29, OMIM=300535, no_hdr0=''),
  4953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4953, CurrentID=0, Status='live', Symbol='ODC1', Aliases='ODC', description='ornithine decarboxylase 1', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10440371, end_position_on_the_genomic_accession=10448554, orientation='minus', exon_count=13, OMIM=165640, no_hdr0=''),
  4954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4954, CurrentID=0, Status='live', Symbol='ODCP', Aliases='ODC2', description='ornithine decarboxylase pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129028498, end_position_on_the_genomic_accession=129031037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4956, CurrentID=0, Status='live', Symbol='ODF1', Aliases='CT133, HSPB10, ODF, ODF2, ODF27, ODFP, ODFPG, ODFPGA, ODFPGB, RT7, SODF', description='outer dense fiber of sperm tails 1', other_designations='cancer/testis antigen 133|heat shock protein beta-10|outer dense fiber of sperm tails, 27-kD', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102551620, end_position_on_the_genomic_accession=102561017, orientation='plus', exon_count=2, OMIM=182878, no_hdr0=''),
  4957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4957, CurrentID=0, Status='live', Symbol='ODF2', Aliases='CT134/1, ODF2/2, ODF84, ODF2', description='outer dense fiber of sperm tails 2', other_designations='cancer/testis antigen 134|cenexin 1|outer dense fiber of sperm tails, 84-kD|sperm tail structural protein', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128455155, end_position_on_the_genomic_accession=128501292, orientation='plus', exon_count=26, OMIM=602015, no_hdr0=''),
  4958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4958, CurrentID=0, Status='live', Symbol='OMD', Aliases='OSAD, SLRR2C', description='osteomodulin', other_designations='KSPG osteomodulin|keratan sulfate proteoglycan osteomodulin|osteoadherin proteoglycan', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92414245, end_position_on_the_genomic_accession=92424554, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4961, CurrentID=0, Status='live', Symbol='OED', Aliases='', description='Oregon eye disease', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4963, CurrentID=0, Status='live', Symbol='OFC1', Aliases='CL, OFC', description='orofacial cleft 1', other_designations='Orofacial cleft-1 (cleft lip with or without cleft palate; isolated cleft palate)', map_location='6p24', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=119530, no_hdr0=''),
  4964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4964, CurrentID=0, Status='live', Symbol='OFC2', Aliases='', description='orofacial cleft 2', other_designations='', map_location='2p14-p13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602966, no_hdr0=''),
  4965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4965, CurrentID=0, Status='live', Symbol='OFC3', Aliases='', description='orofacial cleft 3', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600757, no_hdr0=''),
  4967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4967, CurrentID=0, Status='live', Symbol='OGDH', Aliases='AKGDH, E1k, OGDC', description='oxoglutarate (alpha-ketoglutarate) dehydrogenase (lipoamide)', other_designations='2-oxoglutarate dehydrogenase complex component E1|OGDC-E1|oxoglutarate decarboxylase|oxoglutarate dehydrogenase (succinyl-transferring)', map_location='7p14-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44606522, end_position_on_the_genomic_accession=44709070, orientation='plus', exon_count=26, OMIM=613022, no_hdr0=''),
  4968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4968, CurrentID=0, Status='live', Symbol='OGG1', Aliases='HMMH, HOGG1, MUTM, OGH1', description='8-oxoguanine DNA glycosylase', other_designations='8-hydroxyguanine DNA glycosylase|AP lyase|DNA-apurinic or apyrimidinic site lyase|OGG1 type 1f', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9749944, end_position_on_the_genomic_accession=9788246, orientation='plus', exon_count=11, OMIM=601982, no_hdr0=''),
  4969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4969, CurrentID=0, Status='live', Symbol='OGN', Aliases='OG, OIF, SLRR3A', description='osteoglycin', other_designations='corneal keratan sulfate proteoglycan|mimecan proteoglycan|osteoinductive factor', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92383967, end_position_on_the_genomic_accession=92404655, orientation='minus', exon_count=7, OMIM=602383, no_hdr0=''),
  4973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4973, CurrentID=0, Status='live', Symbol='OLR1', Aliases='CLEC8A, LOX1, LOXIN, SCARE1, SLOX1', description='oxidized low density lipoprotein (lectin-like) receptor 1', other_designations='C-type lectin domain family 8 member A|hLOX-1|lectin-type oxidized LDL receptor 1|ox LDL receptor 1|oxidized low-density lipoprotein receptor 1, soluble form|scavenger receptor class E, member 1', map_location='12p13.2-p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10158300, end_position_on_the_genomic_accession=10176261, orientation='minus', exon_count=7, OMIM=602601, no_hdr0=''),
  4974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4974, CurrentID=0, Status='live', Symbol='OMG', Aliases='OMGP', description='oligodendrocyte myelin glycoprotein', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31294650, end_position_on_the_genomic_accession=31297362, orientation='minus', exon_count=2, OMIM=164345, no_hdr0=''),
  4975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4975, CurrentID=0, Status='live', Symbol='OMP', Aliases='', description='olfactory marker protein', other_designations='olfactory neuronal-specific protein', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77102840, end_position_on_the_genomic_accession=77103331, orientation='plus', exon_count=1, OMIM=164340, no_hdr0=''),
  4976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4976, CurrentID=0, Status='live', Symbol='OPA1', Aliases='MGM1, NPG, NTG, largeG', description='optic atrophy 1 (autosomal dominant)', other_designations='dynamin-like guanosine triphosphatase|mitochondrial dynamin-like GTPase|optic atrophy protein 1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193593144, end_position_on_the_genomic_accession=193697811, orientation='plus', exon_count=32, OMIM=605290, no_hdr0=''),
  4977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4977, CurrentID=0, Status='live', Symbol='OPA2', Aliases='', description='optic atrophy 2 (obscure)', other_designations='', map_location='Xp11.4-p11.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=311050, no_hdr0=''),
  4978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4978, CurrentID=0, Status='live', Symbol='OPCML', Aliases='IGLON1, OBCAM, OPCM', description='opioid binding protein/cell adhesion molecule-like', other_designations='IgLON family member 1|opioid binding protein/cell adhesion molecule-like preprotein', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=132414981, end_position_on_the_genomic_accession=133532508, orientation='minus', exon_count=9, OMIM=600632, no_hdr0=''),
  136051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136051, CurrentID=0, Status='live', Symbol='ZNF786', Aliases='', description='zinc finger protein 786', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149069641, end_position_on_the_genomic_accession=149090777, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4981, CurrentID=0, Status='live', Symbol='OPEM', Aliases='', description='ophthalmoplegia, external, with myopia', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=311000, no_hdr0=''),
  4982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4982, CurrentID=0, Status='live', Symbol='TNFRSF11B', Aliases='OCIF, OPG, PDB5, TR1', description='tumor necrosis factor receptor superfamily member 11b', other_designations='osteoclastogenesis inhibitory factor|osteoprotegerin|tumor necrosis factor receptor superfamily, member 11b', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118923557, end_position_on_the_genomic_accession=118952144, orientation='minus', exon_count=5, OMIM=602643, no_hdr0=''),
  4983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4983, CurrentID=0, Status='live', Symbol='OPHN1', Aliases='ARHGAP41, MRX60, OPN1', description='oligophrenin 1', other_designations='mental retardation, X-linked 60|oligophrenin-1, Rho-GTPase activating protein', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68042344, end_position_on_the_genomic_accession=68433841, orientation='minus', exon_count=26, OMIM=300127, no_hdr0=''),
  4985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4985, CurrentID=0, Status='live', Symbol='OPRD1', Aliases='OPRD', description='opioid receptor, delta 1', other_designations='D-OR-1|DOR-1|delta opioid receptor 1', map_location='1p36.1-p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28812142, end_position_on_the_genomic_accession=28863696, orientation='plus', exon_count=3, OMIM=165195, no_hdr0=''),
  4986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4986, CurrentID=0, Status='live', Symbol='OPRK1', Aliases='K-OR-1, KOR, KOR-1, OPRK', description='opioid receptor, kappa 1', other_designations='Opiate receptor, kappa-1|kappa opioid receptor', map_location='8q11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53225716, end_position_on_the_genomic_accession=53251697, orientation='minus', exon_count=5, OMIM=165196, no_hdr0=''),
  4987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4987, CurrentID=0, Status='live', Symbol='OPRL1', Aliases='KOR-3, NOCIR, OOR, ORL1', description='opiate receptor-like 1', other_designations='kappa-type 3 opioid receptor|kappa3-related opioid receptor|orphanin FQ receptor', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64080118, end_position_on_the_genomic_accession=64100643, orientation='plus', exon_count=7, OMIM=602548, no_hdr0=''),
  4988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4988, CurrentID=0, Status='live', Symbol='OPRM1', Aliases='LMOR, M-OR-1, MOP, MOR, MOR1, OPRM', description='opioid receptor, mu 1', other_designations='mu opiate receptor|mu opioid receptor hMOR-1a', map_location='6q24-q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154010496, end_position_on_the_genomic_accession=154246867, orientation='plus', exon_count=20, OMIM=600018, no_hdr0=''),
  4990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4990, CurrentID=0, Status='live', Symbol='SIX6', Aliases='MCOPCT2, ODRMD, OPTX2, Six9', description='SIX homeobox 6', other_designations='homeodomain protein OPTX2|optic homeobox 2|sine oculis homeobox homolog 6|sine oculis homeobox protein 6', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60509220, end_position_on_the_genomic_accession=60511807, orientation='plus', exon_count=2, OMIM=606326, no_hdr0=''),
  4991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4991, CurrentID=0, Status='live', Symbol='OR1D2', Aliases='OLFR1, OR17-4', description='olfactory receptor family 1 subfamily D member 2', other_designations='olfactory receptor 17-4|olfactory receptor OR17-6|olfactory receptor, family 1, subfamily D, member 2|olfactory receptor-like protein HGMP07E', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3092058, end_position_on_the_genomic_accession=3092996, orientation='minus', exon_count=1, OMIM=164342, no_hdr0=''),
  4992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4992, CurrentID=0, Status='live', Symbol='OR1F1', Aliases='OLFMF, OR16-36, OR16-37, OR16-88, OR16-89, OR16-900, OR1F13P, OR1F4, OR1F5, OR1F6, OR1F7, OR1F8, OR1F9, OR3-145, ORL1023, OR1F1', description='olfactory receptor family 1 subfamily F member 1', other_designations='olfactory receptor, family 1, subfamily F, member 1|olfactory receptor, family 1, subfamily F, member 10|olfactory receptor, family 1, subfamily F, member 13 pseudogene|olfactory receptor, family 1, subfamily F, member 4|olfactory receptor, family 1, subfamily F, member 5|olfactory receptor, family 1, subfamily F, member 6|olfactory receptor, family 1, subfamily F, member 7|olfactory receptor, family 1, subfamily F, member 9', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3191095, end_position_on_the_genomic_accession=3205344, orientation='plus', exon_count=4, OMIM=603232, no_hdr0=''),
  4993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4993, CurrentID=0, Status='live', Symbol='OR2C1', Aliases='OLFmf3, OR2C2P', description='olfactory receptor family 2 subfamily C member 1', other_designations='olfactory receptor 2C2|olfactory receptor OR16-1|olfactory receptor OR16-2|olfactory receptor, family 2, subfamily C, member 1|olfactory receptor, family 2, subfamily C, member 2 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3355889, end_position_on_the_genomic_accession=3356924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4994, CurrentID=0, Status='live', Symbol='OR3A1', Aliases='OLFRA03, OR17-40, OR17-82, OR40', description='olfactory receptor family 3 subfamily A member 1', other_designations='olfactory receptor 17-40|olfactory receptor OR17-15|olfactory receptor, family 3, subfamily A, member 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3288318, end_position_on_the_genomic_accession=3386209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4995, CurrentID=0, Status='live', Symbol='OR3A2', Aliases='OLFRA04, OR17-14, OR17-228, OR228', description='olfactory receptor family 3 subfamily A member 2', other_designations='olfactory receptor 17-228|olfactory receptor OR17-14|olfactory receptor, family 3, subfamily A, member 2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3277890, end_position_on_the_genomic_accession=3278974, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4998, CurrentID=0, Status='live', Symbol='ORC1', Aliases='HSORC1L, PARC1, ORC1', description='origin recognition complex subunit 1', other_designations='origin recognition complex, subunit 1 homolog|replication control protein 1', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52372829, end_position_on_the_genomic_accession=52404471, orientation='minus', exon_count=17, OMIM=601902, no_hdr0=''),
  4999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4999, CurrentID=0, Status='live', Symbol='ORC2', Aliases='ORC2L', description='origin recognition complex subunit 2', other_designations='origin recognition complex protein 2 homolog|origin recognition complex, subunit 2 homolog', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200910171, end_position_on_the_genomic_accession=200963703, orientation='minus', exon_count=20, OMIM=601182, no_hdr0=''),
  5000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5000, CurrentID=0, Status='live', Symbol='ORC4', Aliases='ORC4LP, ORC4', description='origin recognition complex subunit 4', other_designations='origin recognition complex, subunit 4 homolog', map_location='2q22-q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147930397, end_position_on_the_genomic_accession=148021604, orientation='minus', exon_count=17, OMIM=603056, no_hdr0=''),
  5001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5001, CurrentID=0, Status='live', Symbol='ORC5', Aliases='ORC5LP, ORC5T, PPP1R117, ORC5', description='origin recognition complex subunit 5', other_designations='protein phosphatase 1, regulatory subunit 117', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104126341, end_position_on_the_genomic_accession=104208047, orientation='minus', exon_count=15, OMIM=602331, no_hdr0=''),
  5002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5002, CurrentID=0, Status='live', Symbol='SLC22A18', Aliases='BWR1A, BWSCR1A, HET, IMPT1, ITM, ORCTL2, SLC22A1L, TSSC5, p45-BWR1A', description='solute carrier family 22 member 18', other_designations='beckwith-Wiedemann syndrome chromosomal region 1 candidate gene A protein|efflux transporter-like protein|imprinted multi-membrane-spanning polyspecific transporter-related protein 1|organic cation transporter-like protein 2|p45 Beckwith-Wiedemann region 1A|solute carrier family 22, member 18|tumor-suppressing STF cDNA 5 protein|tumor-suppressing subchromosomal transferable fragment candidate gene 5 protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2899721, end_position_on_the_genomic_accession=2925246, orientation='plus', exon_count=18, OMIM=602631, no_hdr0=''),
  5003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5003, CurrentID=0, Status='live', Symbol='SLC22A18AS', Aliases='BWR1B, BWSCR1B, ORCTL2S, SLC22A1LS, p27-BWR1B', description='solute carrier family 22 (organic cation transporter), member 18 antisense', other_designations='Beckwith-Wiedemann region 1B|Beckwith-Wiedemann syndrome chromosome region 1, candidate b|organic cation transporter-like 2 antisense|organic cation transporter-like protein 2 antisense protein|p27-Beckwith-Wiedemann region 1 B|solute carrier family 22 (organic cation transporter), member 1-like antisense|solute carrier family 22 member 1-like antisense protein|solute carrier family 22 member 18 antisense protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2887344, end_position_on_the_genomic_accession=2903945, orientation='minus', exon_count=4, OMIM=603240, no_hdr0=''),
  5004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5004, CurrentID=0, Status='live', Symbol='ORM1', Aliases='AGP-A, AGP1, HEL-S-153w, ORM', description='orosomucoid 1', other_designations='AGP 1|OMD 1|epididymis secretory sperm binding protein Li 153w|orosomucoid-1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114323023, end_position_on_the_genomic_accession=114326479, orientation='plus', exon_count=6, OMIM=138600, no_hdr0=''),
  5005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5005, CurrentID=0, Status='live', Symbol='ORM2', Aliases="AGP-B, AGP-B', AGP2", description='orosomucoid 2', other_designations='AGP 2|OMD 2|alpha-1-acid glycoprotein, type 2|orosomucoid-2', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114329789, end_position_on_the_genomic_accession=114333256, orientation='plus', exon_count=6, OMIM=138610, no_hdr0=''),
  5007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5007, CurrentID=0, Status='live', Symbol='OSBP', Aliases='OSBP1', description='oxysterol binding protein', other_designations='', map_location='11q12-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59574398, end_position_on_the_genomic_accession=59616144, orientation='minus', exon_count=14, OMIM=167040, no_hdr0=''),
  5008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5008, CurrentID=0, Status='live', Symbol='OSM', Aliases='', description='oncostatin M', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30262826, end_position_on_the_genomic_accession=30266840, orientation='minus', exon_count=4, OMIM=165095, no_hdr0=''),
  5009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5009, CurrentID=0, Status='live', Symbol='OTC', Aliases='OCTD', description='ornithine carbamoyltransferase', other_designations='OTCase|ornithine transcarbamylase', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38352483, end_position_on_the_genomic_accession=38421450, orientation='plus', exon_count=10, OMIM=300461, no_hdr0=''),
  5010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5010, CurrentID=0, Status='live', Symbol='CLDN11', Aliases='OSP, OTM', description='claudin 11', other_designations='oligodendrocyte transmembrane protein|oligodendrocyte-specific protein', map_location='3q26.2-q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170418865, end_position_on_the_genomic_accession=170434691, orientation='plus', exon_count=4, OMIM=601326, no_hdr0=''),
  100270915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270915, CurrentID=0, Status='live', Symbol='RPSAP37', Aliases='RPSA_10_610', description='ribosomal protein SA pseudogene 37', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125966738, end_position_on_the_genomic_accession=125967502, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5012, CurrentID=0, Status='live', Symbol='OTSC1', Aliases='OTS', description='otosclerosis 1', other_designations='', map_location='15q25-q26', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=166800, no_hdr0=''),
  5013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5013, CurrentID=0, Status='live', Symbol='OTX1', Aliases='', description='orthodenticle homeobox 1', other_designations='orthodenticle homolog 1', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63050057, end_position_on_the_genomic_accession=63057831, orientation='plus', exon_count=6, OMIM=600036, no_hdr0=''),
  5015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5015, CurrentID=0, Status='live', Symbol='OTX2', Aliases='CPHD6, MCOPS5', description='orthodenticle homeobox 2', other_designations='orthodenticle homolog 2', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56800707, end_position_on_the_genomic_accession=56810476, orientation='minus', exon_count=5, OMIM=600037, no_hdr0=''),
  5016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5016, CurrentID=0, Status='live', Symbol='OVGP1', Aliases='CHIT5, EGP, MUC9, OGP', description='oviductal glycoprotein 1', other_designations='estrogen-dependent oviduct protein|mucin 9|mucin-9|oviduct glycoprotein|oviductal glycoprotein 1, 120kDa|oviductin', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111414315, end_position_on_the_genomic_accession=111427777, orientation='minus', exon_count=11, OMIM=603578, no_hdr0=''),
  5017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5017, CurrentID=0, Status='live', Symbol='OVOL1', Aliases='HOVO1', description='ovo-like zinc finger 1', other_designations='ovo homolog-like 1|ovo-like 1(Drosophila)', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65787034, end_position_on_the_genomic_accession=65797219, orientation='plus', exon_count=7, OMIM=602313, no_hdr0=''),
  5018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5018, CurrentID=0, Status='live', Symbol='OXA1L', Aliases='OXA1', description='oxidase (cytochrome c) assembly 1-like', other_designations='OXA1-like protein|OXA1Hs|hsa|oxidase assembly 1-like protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22766522, end_position_on_the_genomic_accession=22771789, orientation='plus', exon_count=10, OMIM=601066, no_hdr0=''),
  5019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5019, CurrentID=0, Status='live', Symbol='OXCT1', Aliases='OXCT, SCOT', description='3-oxoacid CoA-transferase 1', other_designations='3-oxoacid CoA transferase 1|3-oxoacid-CoA transferase 1|SCOT-s|somatic-type succinyl CoA:3-oxoacid CoA-transferase|somatic-type succinyl-CoA:3-oxoacid CoA-transferase|somatic-type succinyl-CoA:3-oxoacid-CoA-transferase|succinyl CoA:3-oxoacid CoA transferase|succinyl-CoA:3-ketoacid-CoA transferase|succinyl-CoA:3-ketoacid-coenzyme A transferase 1, mitochondrial', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41730065, end_position_on_the_genomic_accession=41870689, orientation='minus', exon_count=17, OMIM=601424, no_hdr0=''),
  5020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5020, CurrentID=0, Status='live', Symbol='OXT', Aliases='OT, OT-NPI-NPI, OXT', description='oxytocin/neurophysin I prepropeptide', other_designations='neurophysin I|oxytocin, prepro- (neurophysin I)|oxytocin, prepropeptide|oxytocin-neurophysin I, preproprotein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3068871, end_position_on_the_genomic_accession=3072517, orientation='plus', exon_count=4, OMIM=167050, no_hdr0=''),
  5021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5021, CurrentID=0, Status='live', Symbol='OXTR', Aliases='OT-R', description='oxytocin receptor', other_designations='', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8750408, end_position_on_the_genomic_accession=8854549, orientation='minus', exon_count=8, OMIM=167055, no_hdr0=''),
  5023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5023, CurrentID=0, Status='live', Symbol='P2RX1', Aliases='P2X1', description='purinergic receptor P2X, ligand gated ion channel, 1', other_designations='ATP receptor|P2X receptor, subunit 1|P2X1 receptor|purinergic receptor P2X, ligand-gated ion channel, 1|purinergic receptor P2X1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3896591, end_position_on_the_genomic_accession=3916666, orientation='minus', exon_count=13, OMIM=600845, no_hdr0=''),
  5024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5024, CurrentID=0, Status='live', Symbol='P2RX3', Aliases='P2X3', description='purinergic receptor P2X, ligand gated ion channel, 3', other_designations='ATP receptor|P2X receptor, subunit 3|purinergic receptor P2X, ligand-gated ion channel, 3|purinergic receptor P2X3|purinoceptor P2X3', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57335946, end_position_on_the_genomic_accession=57372399, orientation='plus', exon_count=16, OMIM=600843, no_hdr0=''),
  5025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5025, CurrentID=0, Status='live', Symbol='P2RX4', Aliases='P2X4, P2X4R', description='purinergic receptor P2X, ligand gated ion channel, 4', other_designations='ATP receptor|ATP-gated cation channel protein|P2X receptor, subunit 4|purinergic receptor P2X, ligand-gated ion channel, 4|purinergic receptor P2X4|purinoceptor P2X4', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121209861, end_position_on_the_genomic_accession=121234106, orientation='plus', exon_count=14, OMIM=600846, no_hdr0=''),
  5026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5026, CurrentID=0, Status='live', Symbol='P2RX5', Aliases='LRH-1, P2X5, P2X5R', description='purinergic receptor P2X, ligand gated ion channel, 5', other_designations='ATP receptor subunit|ionotropic ATP receptor P2X5|lymphoid-restricted histocompatibility antigen-1|purinergic receptor P2X, ligand-gated ion channel, 5', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3673227, end_position_on_the_genomic_accession=3696404, orientation='minus', exon_count=12, OMIM=602836, no_hdr0=''),
  5027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5027, CurrentID=0, Status='live', Symbol='P2RX7', Aliases='P2X7', description='purinergic receptor P2X, ligand gated ion channel, 7', other_designations='ATP receptor|P2X7 receptor|P2Z receptor|purinergic receptor P2X, ligand-gated ion channel, 7|purinergic receptor P2X7 variant A', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121132819, end_position_on_the_genomic_accession=121186551, orientation='plus', exon_count=13, OMIM=602566, no_hdr0=''),
  5028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5028, CurrentID=0, Status='live', Symbol='P2RY1', Aliases='P2Y1', description='purinergic receptor P2Y, G-protein coupled, 1', other_designations='ATP receptor|P2 purinoceptor subtype Y1|platelet ADP receptor', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152834947, end_position_on_the_genomic_accession=152838056, orientation='plus', exon_count=1, OMIM=601167, no_hdr0=''),
  5029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5029, CurrentID=0, Status='live', Symbol='P2RY2', Aliases='HP2U, P2RU1, P2U, P2U1, P2UR, P2Y2, P2Y2R', description='purinergic receptor P2Y, G-protein coupled, 2', other_designations='ATP receptor|P2U nucleotide receptor|P2U purinoceptor 1|P2U receptor 1|purinergic receptor P2Y2|purinoceptor P2Y2', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73216448, end_position_on_the_genomic_accession=73242427, orientation='plus', exon_count=7, OMIM=600041, no_hdr0=''),
  5030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5030, CurrentID=0, Status='live', Symbol='P2RY4', Aliases='NRU, P2P, P2Y4, UNR', description='pyrimidinergic receptor P2Y, G-protein coupled, 4', other_designations='pyrimidinergic receptor P2Y4|uridine nucleotide receptor', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70258166, end_position_on_the_genomic_accession=70259804, orientation='minus', exon_count=1, OMIM=300038, no_hdr0=''),
  5031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5031, CurrentID=0, Status='live', Symbol='P2RY6', Aliases='P2Y6', description='pyrimidinergic receptor P2Y, G-protein coupled, 6', other_designations='G-coupled nucleotide receptor|P2 purinoceptor|P2Y6 receptor', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73264496, end_position_on_the_genomic_accession=73298625, orientation='plus', exon_count=8, OMIM=602451, no_hdr0=''),
  5032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5032, CurrentID=0, Status='live', Symbol='P2RY11', Aliases='P2Y11', description='purinergic receptor P2Y, G-protein coupled, 11', other_designations='purinergic receptor P2Y11', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10111521, end_position_on_the_genomic_accession=10115389, orientation='plus', exon_count=2, OMIM=602697, no_hdr0=''),
  5033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5033, CurrentID=0, Status='live', Symbol='P4HA1', Aliases='P4HA', description='prolyl 4-hydroxylase, alpha polypeptide I', other_designations='C-P4Halpha(I)|collagen prolyl 4-hydroxylase alpha(I)|procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha polypeptide I|procollagen-proline,2-oxoglutarate-4-dioxygenase subunit alpha-1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73007217, end_position_on_the_genomic_accession=73096974, orientation='minus', exon_count=17, OMIM=176710, no_hdr0=''),
  5034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5034, CurrentID=0, Status='live', Symbol='P4HB', Aliases='CLCRP1, DSI, ERBA2L, GIT, P4Hbeta, PDI, PDIA1, PHDB, PO4DB, PO4HB, PROHB', description='prolyl 4-hydroxylase, beta polypeptide', other_designations='cellular thyroid hormone-binding protein|collagen prolyl 4-hydroxylase beta|glutathione-insulin transhydrogenase|p55|procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), beta polypeptide|prolyl 4-hydroxylase subunit beta|protein disulfide isomerase family A, member 1|protein disulfide isomerase-associated 1|protein disulfide isomerase/oxidoreductase|protocollagen hydroxylase|thyroid hormone-binding protein p55', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81843158, end_position_on_the_genomic_accession=81860668, orientation='minus', exon_count=11, OMIM=176790, no_hdr0=''),
  100270919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270919, CurrentID=0, Status='live', Symbol='RPSAP42', Aliases='RPSA_18_761', description='ribosomal protein SA pseudogene 42', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137995270, end_position_on_the_genomic_accession=137996147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5036, CurrentID=0, Status='live', Symbol='PA2G4', Aliases='EBP1, HG4-1, p38-2G4', description='proliferation-associated 2G4', other_designations='ErbB-3 binding protein 1|ErbB3-binding protein Ebp1|cell cycle protein p38-2G4 homolog|erbB3-binding protein 1|proliferation-associated 2G4, 38kD|proliferation-associated 2G4, 38kDa', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56104319, end_position_on_the_genomic_accession=56113910, orientation='plus', exon_count=13, OMIM=602145, no_hdr0=''),
  5037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5037, CurrentID=0, Status='live', Symbol='PEBP1', Aliases='HCNP, HCNPpp, HEL-210, HEL-S-34, HEL-S-96, PBP, PEBP, PEBP-1, RKIP', description='phosphatidylethanolamine binding protein 1', other_designations='Raf kinase inhibitory protein|epididymis luminal protein 210|epididymis secretory protein Li 34|epididymis secretory protein Li 96|hippocampal cholinergic neurostimulating peptide|neuropolypeptide h3|prostatic binding protein|raf kinase inhibitor protein', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118136065, end_position_on_the_genomic_accession=118145585, orientation='plus', exon_count=4, OMIM=604591, no_hdr0=''),
  100270920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270920, CurrentID=0, Status='live', Symbol='RPL12P15', Aliases='RPL12_1_272', description='ribosomal protein L12 pseudogene 15', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121658416, end_position_on_the_genomic_accession=121659060, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5042, CurrentID=0, Status='live', Symbol='PABPC3', Aliases='PABP3, PABPL3, tPABP', description='poly(A) binding protein, cytoplasmic 3', other_designations='PABP-3|poly(A)-binding protein 3|poly(A)-binding protein-like 3|testis PABP|testis-specific poly(A)-binding protein (PABP)|testis-specific poly(A)-binding protein 3', map_location='13q12-q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25096138, end_position_on_the_genomic_accession=25098567, orientation='plus', exon_count=1, OMIM=604680, no_hdr0=''),
  5043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5043, CurrentID=0, Status='live', Symbol='PABX', Aliases='', description='pseudoautosomal boundary region, X-linked', other_designations='', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5044, CurrentID=0, Status='live', Symbol='PABY', Aliases='', description='pseudoautosomal boundary region, Y-linked', other_designations='', map_location='Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5045, CurrentID=0, Status='live', Symbol='FURIN', Aliases='FUR, PACE, PCSK3, SPC1', description='furin (paired basic amino acid cleaving enzyme)', other_designations='FES upstream region|dibasic processing enzyme|furin, membrane associated receptor protein|paired basic amino acid residue-cleaving enzyme|proprotein convertase subtilisin/kexin type 3', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90868592, end_position_on_the_genomic_accession=90883458, orientation='plus', exon_count=18, OMIM=136950, no_hdr0=''),
  5046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5046, CurrentID=0, Status='live', Symbol='PCSK6', Aliases='PACE4, SPC4', description='proprotein convertase subtilisin/kexin type 6', other_designations='paired basic amino acid cleaving enzyme 4|paired basic amino acid cleaving system 4|subtilisin-like proprotein convertase 4|subtilisin/kexin-like protease PACE4', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101303928, end_position_on_the_genomic_accession=101489984, orientation='minus', exon_count=24, OMIM=167405, no_hdr0=''),
  5047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5047, CurrentID=0, Status='live', Symbol='PAEP', Aliases='GD, GdA, GdF, GdS, PAEG, PEP, PP14', description='progestagen-associated endometrial protein', other_designations='PEG|PP14 protein (placental protein 14)|alpha uterine protein|glycodelin-A|glycodelin-F|glycodelin-S|placental protein 14|pregnancy-associated endometrial alpha-2 globulin|pregnancy-associated endometrial alpha-2-globulin|progestagen-associated endometrial protein (placental protein 14, pregnancy-associated endometrial a|progesterone-associated endometrial protein', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135561754, end_position_on_the_genomic_accession=135566776, orientation='plus', exon_count=7, OMIM=173310, no_hdr0=''),
  5048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5048, CurrentID=0, Status='live', Symbol='PAFAH1B1', Aliases='LIS1, LIS2, MDCR, MDS, PAFAH', description='platelet activating factor acetylhydrolase 1b regulatory subunit 1', other_designations='lissencephaly 1 protein|platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 (45kDa)|platelet-activating factor acetylhydrolase, isoform Ib, alpha subunit (45kD)|platelet-activating factor acetylhydrolase, isoform Ib, subunit 1 (45kDa)', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2593629, end_position_on_the_genomic_accession=2685617, orientation='plus', exon_count=14, OMIM=601545, no_hdr0=''),
  5049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5049, CurrentID=0, Status='live', Symbol='PAFAH1B2', Aliases='HEL-S-303', description='platelet activating factor acetylhydrolase 1b catalytic subunit 2', other_designations='PAF acetylhydrolase 30 kDa subunit|PAF-AH 30 kDa subunit|PAF-AH subunit beta|PAF-AH1b alpha 2 subunit|PAFAH subunit beta|epididymis secretory protein Li 303|intracellular platelet-activating factor acetylhydrolase alpha 2 subunit|platelet-activating factor acetylhydrolase 1b, catalytic subunit 2 (30kDa)|platelet-activating factor acetylhydrolase, isoform Ib, subunit 2 (30kDa)', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117144284, end_position_on_the_genomic_accession=117178173, orientation='plus', exon_count=9, OMIM=602508, no_hdr0=''),
  5050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5050, CurrentID=0, Status='live', Symbol='PAFAH1B3', Aliases='PAFAHG', description='platelet activating factor acetylhydrolase 1b catalytic subunit 3', other_designations='PAF acetylhydrolase 29 kDa subunit|PAF-AH 29 kDa subunit|PAF-AH subunit gamma|PAF-AH1b alpha 1 subunit|PAFAH subunit gamma|platelet-activating factor acetylhydrolase 1b, catalytic subunit 3 (29kDa)|platelet-activating factor acetylhydrolase, isoform Ib, subunit 3 (29kDa)', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42297033, end_position_on_the_genomic_accession=42302800, orientation='minus', exon_count=6, OMIM=603074, no_hdr0=''),
  5051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5051, CurrentID=0, Status='live', Symbol='PAFAH2', Aliases='HSD-PLA2', description='platelet-activating factor acetylhydrolase 2', other_designations='SD-PLA2|platelet-activating factor acetylhydrolase 2, 40kDa|serine-dependent phospholipase A2', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25959764, end_position_on_the_genomic_accession=25998581, orientation='minus', exon_count=13, OMIM=602344, no_hdr0=''),
  5052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5052, CurrentID=0, Status='live', Symbol='PRDX1', Aliases='MSP23, NKEF-A, NKEFA, PAG, PAGA, PAGB, PRX1, PRXI, TDPX2', description='peroxiredoxin 1', other_designations='natural killer cell-enhancing factor A|natural killer-enhancing factor A|proliferation-associated gene A|proliferation-associated gene protein|thioredoxin peroxidase 2|thioredoxin-dependent peroxide reductase 2', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45511035, end_position_on_the_genomic_accession=45522890, orientation='minus', exon_count=7, OMIM=176763, no_hdr0=''),
  5053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5053, CurrentID=0, Status='live', Symbol='PAH', Aliases='PH, PKU, PKU1', description='phenylalanine hydroxylase', other_designations='phe-4-monooxygenase|phenylalanine 4-monooxygenase', map_location='12q22-q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102838321, end_position_on_the_genomic_accession=102917603, orientation='minus', exon_count=13, OMIM=612349, no_hdr0=''),
  5054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5054, CurrentID=0, Status='live', Symbol='SERPINE1', Aliases='PAI, PAI-1, PAI1, PLANH1', description='serpin peptidase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 1', other_designations='endothelial plasminogen activator inhibitor|serine (or cysteine) proteinase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 1|serpin E1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101127089, end_position_on_the_genomic_accession=101139266, orientation='plus', exon_count=9, OMIM=173360, no_hdr0=''),
  5055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5055, CurrentID=0, Status='live', Symbol='SERPINB2', Aliases='HsT1201, PAI, PAI-2, PAI2, PLANH2', description='serpin peptidase inhibitor, clade B (ovalbumin), member 2', other_designations='monocyte Arg-serpin|placental plasminogen activator inhibitor|plasminogen activator inhibitor, type II (arginine-serpin)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 2|serpin B2|urokinase inhibitor', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63887705, end_position_on_the_genomic_accession=63903890, orientation='plus', exon_count=9, OMIM=173390, no_hdr0=''),
  100533624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533624, CurrentID=0, Status='live', Symbol='SNX33P1', Aliases='', description='sorting nexin 33 pseudogene 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16034829, end_position_on_the_genomic_accession=16036095, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5058, CurrentID=0, Status='live', Symbol='PAK1', Aliases='PAKalpha', description='p21 protein (Cdc42/Rac)-activated kinase 1', other_designations='STE20 homolog, yeast|alpha-PAK|p21/Cdc42/Rac1-activated kinase 1 (STE20 homolog, yeast)|p21/Cdc42/Rac1-activated kinase 1 (yeast Ste20-related)|p65-PAK', map_location='11q13-q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77321551, end_position_on_the_genomic_accession=77474063, orientation='minus', exon_count=19, OMIM=602590, no_hdr0=''),
  100270923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270923, CurrentID=0, Status='live', Symbol='RPL12P20', Aliases='RPL12_5_471', description='ribosomal protein L12 pseudogene 20', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41389115, end_position_on_the_genomic_accession=41389593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373327, CurrentID=0, Status='live', Symbol='LOC105373327', Aliases='', description='uncharacterized LOC105373327', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120575903, end_position_on_the_genomic_accession=120595990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  391092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391092, CurrentID=0, Status='live', Symbol='LOC391092', Aliases='', description='PRAME family member 17 pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147840952, end_position_on_the_genomic_accession=147842497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5062, CurrentID=0, Status='live', Symbol='PAK2', Aliases='PAK65, PAKgamma', description='p21 protein (Cdc42/Rac)-activated kinase 2', other_designations='PAK-2|S6/H4 kinase|gamma-PAK|p21 (CDKN1A)-activated kinase 2|p21-activated kinase 2|p58', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196739857, end_position_on_the_genomic_accession=196832647, orientation='plus', exon_count=16, OMIM=605022, no_hdr0=''),
  5063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5063, CurrentID=0, Status='live', Symbol='PAK3', Aliases='MRX30, MRX47, OPHN3, PAK-3beta, bPAK, beta-PAK, PAK3', description='p21 protein (Cdc42/Rac)-activated kinase 3', other_designations='oligophrenin-3|p21 (CDKN1A)-activated kinase 3', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110944285, end_position_on_the_genomic_accession=111227361, orientation='plus', exon_count=29, OMIM=300142, no_hdr0=''),
  5064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5064, CurrentID=0, Status='live', Symbol='PALM', Aliases='', description='paralemmin', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=708953, end_position_on_the_genomic_accession=748330, orientation='plus', exon_count=13, OMIM=608134, no_hdr0=''),
  100270924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270924, CurrentID=0, Status='live', Symbol='RPL12P24', Aliases='RPL12_6_940', description='ribosomal protein L12 pseudogene 24', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105737280, end_position_on_the_genomic_accession=105737756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5066, CurrentID=0, Status='live', Symbol='PAM', Aliases='PAL, PHM', description='peptidylglycine alpha-amidating monooxygenase', other_designations='pancreatic peptidylglycine alpha-amidating monooxygenase|peptidyl alpha-amidating enzyme|peptidyl-alpha-hydroxyglycine alpha-amidating lyase|peptidylamidoglycolate lyase|peptidylglycine 2-hydroxylase|peptidylglycine alpha-hydroxylating monooxygenase', map_location='5q14-q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102755276, end_position_on_the_genomic_accession=103031105, orientation='plus', exon_count=28, OMIM=170270, no_hdr0=''),
  5067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5067, CurrentID=0, Status='live', Symbol='CNTN3', Aliases='BIG-1, PANG, PCS', description='contactin 3', other_designations='brain-derived immunoglobulin superfamily protein 1|contactin 3 (plasmacytoma associated)|plasmacytoma-associated neuronal glycoprotein', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74262568, end_position_on_the_genomic_accession=74614445, orientation='minus', exon_count=25, OMIM=601325, no_hdr0=''),
  5068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5068, CurrentID=0, Status='live', Symbol='REG3A', Aliases='HIP, HIP/PAP, INGAP, PAP, PAP-H, PAP1, PBCGF, REG-III, REG3', description='regenerating family member 3 alpha', other_designations='PAP homologous protein|REG-3-alpha|hepatocarcinoma-intestine-pancreas|hepatointestinal pancreatic protein|human proislet peptide|pancreatic beta cell growth factor|pancreatitis-associated protein 1|proliferation-inducing protein 34|proliferation-inducing protein 42|reg III-alpha|regenerating islet-derived 3 alpha|regenerating islet-derived protein III-alpha', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79157006, end_position_on_the_genomic_accession=79159754, orientation='minus', exon_count=6, OMIM=167805, no_hdr0=''),
  5069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5069, CurrentID=0, Status='live', Symbol='PAPPA', Aliases='ASBABP2, DIPLA1, IGFBP-4ase, PAPA, PAPP-A1, PAPPA', description='pregnancy-associated plasma protein A, pappalysin 1', other_designations='IGF-dependent IGFBP-4 protease|aspecific BCL2 ARE-binding protein 2|differentially placenta 1 expressed protein|insulin-like growth factor-dependent IGF binding protein-4 protease|insulin-like growth factor-dependent IGF-binding protein 4 protease|pregnacy-associated plasma protein A', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116153792, end_position_on_the_genomic_accession=116402321, orientation='plus', exon_count=22, OMIM=176385, no_hdr0=''),
  5071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5071, CurrentID=0, Status='live', Symbol='PARK2', Aliases='AR-JP, LPRS2, PDJ, PRKN', description='parkin RBR E3 ubiquitin protein ligase', other_designations='Parkinson disease (autosomal recessive, juvenile) 2, parkin|parkinson juvenile disease protein 2|parkinson protein 2 E3 ubiquitin protein ligase|parkinson protein 2, E3 ubiquitin protein ligase (parkin)', map_location='6q25.2-q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=161347417, end_position_on_the_genomic_accession=162727802, orientation='minus', exon_count=16, OMIM=602544, no_hdr0=''),
  5072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5072, CurrentID=0, Status='live', Symbol='PARK3', Aliases='', description='Parkinson disease 3 (autosomal dominant, Lewy body)', other_designations='Parkinson disease (autosomal dominant, Lewy body) 3|Parkinson disease (autosomal dominant, Lewy-body) 3|Parkinson disease, dominant Lewy-body, 3', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602404, no_hdr0=''),
  5073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5073, CurrentID=0, Status='live', Symbol='PARN', Aliases='DAN, DKCB6, PFBMFT4', description='poly(A)-specific ribonuclease', other_designations='deadenylating nuclease|deadenylation nuclease|polyadenylate-specific ribonuclease', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14435700, end_position_on_the_genomic_accession=14630275, orientation='minus', exon_count=27, OMIM=604212, no_hdr0=''),
  5074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5074, CurrentID=0, Status='live', Symbol='PAWR', Aliases='PAR4, Par-4', description='PRKC, apoptosis, WT1, regulator', other_designations='WT1-interacting protein|prostate apoptosis response 4 protein|prostate apoptosis response protein 4|prostate apoptosis response protein PAR-4|prostate apoptosis response-4|transcriptional repressor PAR4', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79587791, end_position_on_the_genomic_accession=79691010, orientation='minus', exon_count=11, OMIM=601936, no_hdr0=''),
  5075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5075, CurrentID=0, Status='live', Symbol='PAX1', Aliases='HUP48, OFC2', description='paired box 1', other_designations='paired box gene 1|paired domain gene HuP48', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21705659, end_position_on_the_genomic_accession=21718486, orientation='plus', exon_count=5, OMIM=167411, no_hdr0=''),
  5076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5076, CurrentID=0, Status='live', Symbol='PAX2', Aliases='FSGS7, PAPRS', description='paired box 2', other_designations='paired box homeotic gene 2', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100735709, end_position_on_the_genomic_accession=100829941, orientation='plus', exon_count=13, OMIM=167409, no_hdr0=''),
  5077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5077, CurrentID=0, Status='live', Symbol='PAX3', Aliases='CDHS, HUP2, WS1, WS3', description='paired box 3', other_designations='paired box homeotic gene 3|paired domain gene 3|paired domain gene HuP2', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222199887, end_position_on_the_genomic_accession=222321263, orientation='minus', exon_count=14, OMIM=606597, no_hdr0=''),
  5078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5078, CurrentID=0, Status='live', Symbol='PAX4', Aliases='KPD, MODY9', description='paired box 4', other_designations='paired box gene 4|paired domain gene 4', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127610292, end_position_on_the_genomic_accession=127618192, orientation='minus', exon_count=12, OMIM=167413, no_hdr0=''),
  5079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5079, CurrentID=0, Status='live', Symbol='PAX5', Aliases='ALL3, BSAP', description='paired box 5', other_designations='B-cell lineage specific activator|paired box homeotic gene 5|paired domain gene 5|transcription factor PAX 5', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36833275, end_position_on_the_genomic_accession=37035319, orientation='minus', exon_count=12, OMIM=167414, no_hdr0=''),
  5080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5080, CurrentID=0, Status='live', Symbol='PAX6', Aliases='AN, AN2, D11S812E, FVH1, MGDA, WAGR', description='paired box 6', other_designations='aniridia type II protein|oculorhombin|paired box homeotic gene-6', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31784792, end_position_on_the_genomic_accession=31817961, orientation='minus', exon_count=20, OMIM=607108, no_hdr0=''),
  5081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5081, CurrentID=0, Status='live', Symbol='PAX7', Aliases='HUP1B, RMS2, PAX7', description='paired box 7', other_designations='PAX7 transcriptional factor|paired box homeotic gene 7|paired domain gene 7', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18631006, end_position_on_the_genomic_accession=18748866, orientation='plus', exon_count=9, OMIM=167410, no_hdr0=''),
  5082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5082, CurrentID=0, Status='live', Symbol='PDCL', Aliases='PhLP', description='phosducin like', other_designations='phosducin-like', map_location='9q12-q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122818097, end_position_on_the_genomic_accession=122828656, orientation='minus', exon_count=4, OMIM=604421, no_hdr0=''),
  5083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5083, CurrentID=0, Status='live', Symbol='PAX9', Aliases='STHAG3', description='paired box 9', other_designations='paired domain gene 9', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36657568, end_position_on_the_genomic_accession=36677807, orientation='plus', exon_count=5, OMIM=167416, no_hdr0=''),
  5084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5084, CurrentID=0, Status='live', Symbol='PBG1', Aliases='', description='pancreatic beta cell glycoprotein 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  136157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136157, CurrentID=0, Status='live', Symbol='SSU72P8', Aliases='', description='SSU72 pseudogene 8', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 8', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124476371, end_position_on_the_genomic_accession=124476955, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5087, CurrentID=0, Status='live', Symbol='PBX1', Aliases='', description='pre-B-cell leukemia homeobox 1', other_designations='homeobox protein PBX1|homeobox protein PRL', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164559360, end_position_on_the_genomic_accession=164886047, orientation='plus', exon_count=17, OMIM=176310, no_hdr0=''),
  5088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5088, CurrentID=0, Status='live', Symbol='PBX2P1', Aliases='PBX2, PBXP1', description='pre-B-cell leukemia homeobox 2 pseudogene 1', other_designations='pre-B-cell leukemia transcription factor pseudogene 1', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143176062, end_position_on_the_genomic_accession=143179253, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5089, CurrentID=0, Status='live', Symbol='PBX2', Aliases='G17, HOX12MHC, PBX2', description='pre-B-cell leukemia homeobox 2', other_designations='XXbac-BPG300A18.13|homeobox 12|homeobox protein PBX2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32184733, end_position_on_the_genomic_accession=32190186, orientation='minus', exon_count=9, OMIM=176311, no_hdr0=''),
  5090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5090, CurrentID=0, Status='live', Symbol='PBX3', Aliases='', description='pre-B-cell leukemia homeobox 3', other_designations='homeobox protein PBX3', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125747338, end_position_on_the_genomic_accession=125967377, orientation='plus', exon_count=13, OMIM=176312, no_hdr0=''),
  5091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5091, CurrentID=0, Status='live', Symbol='PC', Aliases='PCB', description='pyruvate carboxylase', other_designations='pyruvic carboxylase', map_location='11q13.4-q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66848522, end_position_on_the_genomic_accession=66958376, orientation='minus', exon_count=32, OMIM=608786, no_hdr0=''),
  5092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5092, CurrentID=0, Status='live', Symbol='PCBD1', Aliases='DCOH, PCBD, PCD, PHS', description='pterin-4 alpha-carbinolamine dehydratase/dimerization cofactor of hepatocyte nuclear factor 1 alpha', other_designations='4-alpha-hydroxy-tetrahydropterin dehydratase|6-pyruvoyl-tetrahydropterin synthase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1)|dimerizing cofactor for HNF1|phenylalanine hydroxylase-stimulating protein', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70882280, end_position_on_the_genomic_accession=70888786, orientation='minus', exon_count=6, OMIM=126090, no_hdr0=''),
  5093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5093, CurrentID=0, Status='live', Symbol='PCBP1', Aliases='HEL-S-85, HNRPE1, HNRPX, hnRNP-E1, hnRNP-X', description='poly(rC) binding protein 1', other_designations='alpha-CP1|epididymis secretory protein Li 85|heterogeneous nuclear ribonucleoprotein E1|heterogenous nuclear ribonucleoprotein E1|heterogenous nuclear ribonucleoprotein X|nucleic acid-binding protein SUB2.3', map_location='2p13-p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70087453, end_position_on_the_genomic_accession=70089203, orientation='plus', exon_count=1, OMIM=601209, no_hdr0=''),
  5094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5094, CurrentID=0, Status='live', Symbol='PCBP2', Aliases='HNRNPE2, HNRPE2, hnRNP-E2', description='poly(rC) binding protein 2', other_designations='alpha-CP2|heterogeneous nuclear ribonucleoprotein E2|heterogenous nuclear ribonucleoprotein E2|hnRNP E2', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53452102, end_position_on_the_genomic_accession=53481162, orientation='plus', exon_count=15, OMIM=601210, no_hdr0=''),
  5095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5095, CurrentID=0, Status='live', Symbol='PCCA', Aliases='', description='propionyl-CoA carboxylase alpha subunit', other_designations='PCCase alpha subunit|pccA complementation group|propanoyl-CoA:carbon dioxide ligase alpha subunit|propionyl CoA carboxylase, alpha polypeptide|propionyl Coenzyme A carboxylase, alpha polypeptide', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100089015, end_position_on_the_genomic_accession=100530437, orientation='plus', exon_count=27, OMIM=232000, no_hdr0=''),
  5096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5096, CurrentID=0, Status='live', Symbol='PCCB', Aliases='', description='propionyl-CoA carboxylase beta subunit', other_designations='PCCase subunit beta|propanoyl-CoA:carbon dioxide ligase subunit beta|propionyl CoA carboxylase, beta polypeptide|propionyl Coenzyme A carboxylase, beta polypeptide', map_location='3q21-q22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136250325, end_position_on_the_genomic_accession=136330171, orientation='plus', exon_count=17, OMIM=232050, no_hdr0=''),
  5097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5097, CurrentID=0, Status='live', Symbol='PCDH1', Aliases='PC42, PCDH42', description='protocadherin 1', other_designations='cadherin-like 1|cadherin-like protein 1|protocadherin 42|protocadherin-42', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141852319, end_position_on_the_genomic_accession=141878813, orientation='minus', exon_count=9, OMIM=603626, no_hdr0=''),
  5098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5098, CurrentID=0, Status='live', Symbol='PCDHGC3', Aliases='PC43, PCDH-GAMMA-C3, PCDH2', description='protocadherin gamma subfamily C, 3', other_designations='cadherin-like 2|protocadherin 2|protocadherin 43|protocadherin-2|protocadherin-43', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141476002, end_position_on_the_genomic_accession=141512977, orientation='plus', exon_count=4, OMIM=603627, no_hdr0=''),
  5099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5099, CurrentID=0, Status='live', Symbol='PCDH7', Aliases='BH-Pcdh, BHPCDH, PPP1R120', description='protocadherin 7', other_designations='BH-protocadherin (brain-heart)|brain-heart protocadherin|protein phosphatase 1, regulatory subunit 120', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=30720319, end_position_on_the_genomic_accession=31146801, orientation='plus', exon_count=8, OMIM=602988, no_hdr0=''),
  5100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5100, CurrentID=0, Status='live', Symbol='PCDH8', Aliases='ARCADLIN, PAPC', description='protocadherin 8', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52843974, end_position_on_the_genomic_accession=52848640, orientation='minus', exon_count=3, OMIM=603580, no_hdr0=''),
  5101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5101, CurrentID=0, Status='live', Symbol='PCDH9', Aliases='', description='protocadherin 9', other_designations='cadherin superfamily protein VR4-11', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66302834, end_position_on_the_genomic_accession=67230336, orientation='minus', exon_count=7, OMIM=603581, no_hdr0=''),
  5104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5104, CurrentID=0, Status='live', Symbol='SERPINA5', Aliases='PAI-3, PAI3, PCI, PCI-B, PLANH3, PROCI', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 5', other_designations='acrosomal serine protease inhibitor|plasminogen activator inhibitor III|plasminogen activator inhibitor-3|protein C inhibitor|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 5', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94581369, end_position_on_the_genomic_accession=94593120, orientation='plus', exon_count=6, OMIM=601841, no_hdr0=''),
  5105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5105, CurrentID=0, Status='live', Symbol='PCK1', Aliases='PEPCK-C, PEPCK1, PEPCKC', description='phosphoenolpyruvate carboxykinase 1', other_designations='PEP carboxykinase|phosphoenolpyruvate carboxykinase 1 (soluble)|phosphoenolpyruvate carboxykinase, cytosolic|phosphoenolpyruvate carboxylase|phosphopyruvate carboxylase', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57561081, end_position_on_the_genomic_accession=57566461, orientation='plus', exon_count=10, OMIM=614168, no_hdr0=''),
  5106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5106, CurrentID=0, Status='live', Symbol='PCK2', Aliases='PEPCK, PEPCK-M, PEPCK2', description='phosphoenolpyruvate carboxykinase 2 (mitochondrial)', other_designations='PEP carboxykinase|phosphopyruvate carboxylase', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24094131, end_position_on_the_genomic_accession=24104132, orientation='plus', exon_count=11, OMIM=614095, no_hdr0=''),
  5108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5108, CurrentID=0, Status='live', Symbol='PCM1', Aliases='PTC4, RET/PCM-1', description='pericentriolar material 1', other_designations='PCM-1|hPCM-1|pericentriolar material 1, PCM1', map_location='8p22-p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17922857, end_position_on_the_genomic_accession=18029948, orientation='plus', exon_count=46, OMIM=600299, no_hdr0=''),
  5110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5110, CurrentID=0, Status='live', Symbol='PCMT1', Aliases='PIMT', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase', other_designations='L-isoaspartyl protein carboxyl methyltransferase|protein L-isoaspartyl/D-aspartyl methyltransferase|protein-beta-aspartate methyltransferase', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149749695, end_position_on_the_genomic_accession=149811421, orientation='plus', exon_count=9, OMIM=176851, no_hdr0=''),
  5111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5111, CurrentID=0, Status='live', Symbol='PCNA', Aliases='ATLD2', description='proliferating cell nuclear antigen', other_designations='DNA polymerase delta auxiliary protein|cyclin', map_location='20pter-p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5114953, end_position_on_the_genomic_accession=5126622, orientation='minus', exon_count=7, OMIM=176740, no_hdr0=''),
  5116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5116, CurrentID=0, Status='live', Symbol='PCNT', Aliases='KEN, MOPD2, PCN2, PCNTB, PCTN2, SCKL4, PCNT', description='pericentrin', other_designations='kendrin|pericentrin-2|pericentrin-380|pericentrin-B', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46324100, end_position_on_the_genomic_accession=46445769, orientation='plus', exon_count=50, OMIM=605925, no_hdr0=''),
  5118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5118, CurrentID=0, Status='live', Symbol='PCOLCE', Aliases='PCPE, PCPE-1, PCPE1', description='procollagen C-endopeptidase enhancer', other_designations='procollagen C-proteinase enhancer 1|procollagen COOH-terminal proteinase enhancer 1|procollagen, type 1, COOH-terminal proteinase enhancer|type 1 procollagen C-proteinase enhancer protein|type I procollagen COOH-terminal proteinase enhancer', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100602259, end_position_on_the_genomic_accession=100608175, orientation='plus', exon_count=9, OMIM=600270, no_hdr0=''),
  5119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5119, CurrentID=0, Status='live', Symbol='CHMP1A', Aliases='CHMP1, PCH8, PCOLN3, PRSM1, VPS46-1, VPS46A', description='charged multivesicular body protein 1A', other_designations='charged multivesicular body protein 1/chromatin modifying protein 1|chromatin modifying protein 1A|procollagen (type III) N-endopeptidase|protease, metallo, 1, 33kD|vacuolar protein sorting-associated protein 46-1', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89644431, end_position_on_the_genomic_accession=89657785, orientation='minus', exon_count=9, OMIM=164010, no_hdr0=''),
  5120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5120, CurrentID=0, Status='live', Symbol='PCOS1', Aliases='PCO, PCO1', description='polycystic ovary syndrome 1', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=184700, no_hdr0=''),
  5121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5121, CurrentID=0, Status='live', Symbol='PCP4', Aliases='PEP-19', description='Purkinje cell protein 4', other_designations='brain specific polypeptide PEP19|brain-specific antigen PCP-4|brain-specific polypeptide PEP-19', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39867421, end_position_on_the_genomic_accession=39929397, orientation='plus', exon_count=3, OMIM=601629, no_hdr0=''),
  5122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5122, CurrentID=0, Status='live', Symbol='PCSK1', Aliases='BMIQ12, NEC1, PC1, PC3, SPC3', description='proprotein convertase subtilisin/kexin type 1', other_designations='prohormone convertase 1|prohormone convertase 3', map_location='5q15-q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96390336, end_position_on_the_genomic_accession=96433281, orientation='minus', exon_count=15, OMIM=162150, no_hdr0=''),
  102503427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102503427, CurrentID=0, Status='live', Symbol='LOC102503427', Aliases='', description='uncharacterized LOC102503427', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72557077, end_position_on_the_genomic_accession=72571743, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102503428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102503428, CurrentID=0, Status='live', Symbol='DUP22Q13', Aliases='C22DUPq13', description='Chromosome 22q13 duplication syndrome', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615538, no_hdr0=''),
  5125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5125, CurrentID=0, Status='live', Symbol='PCSK5', Aliases='PC5, PC6, PC6A, SPC6', description='proprotein convertase subtilisin/kexin type 5', other_designations='prohormone convertase 5|proprotein convertase 6|protease PC6|subtilase|subtilisin/kexin-like protease PC5', map_location='9q21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75890601, end_position_on_the_genomic_accession=76362339, orientation='plus', exon_count=40, OMIM=600488, no_hdr0=''),
  5126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5126, CurrentID=0, Status='live', Symbol='PCSK2', Aliases='NEC 2, NEC-2, NEC2, PC2, SPC2', description='proprotein convertase subtilisin/kexin type 2', other_designations='KEX2-like endoprotease 2|prohormone convertase 2', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17226107, end_position_on_the_genomic_accession=17484578, orientation='plus', exon_count=13, OMIM=162151, no_hdr0=''),
  5127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5127, CurrentID=0, Status='live', Symbol='CDK16', Aliases='PCTAIRE, PCTAIRE1, PCTGAIRE, PCTK1', description='cyclin-dependent kinase 16', other_designations='PCTAIRE-motif protein kinase 1|cell division protein kinase 16|serine/threonine-protein kinase PCTAIRE-1', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47217881, end_position_on_the_genomic_accession=47229997, orientation='plus', exon_count=19, OMIM=311550, no_hdr0=''),
  5128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5128, CurrentID=0, Status='live', Symbol='CDK17', Aliases='PCTAIRE2, PCTK2', description='cyclin-dependent kinase 17', other_designations='PCTAIRE-motif protein kinase 2|cell division protein kinase 17|protein kinase cdc2-related PCTAIRE-2|serine/threonine-protein kinase PCTAIRE-2', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96278261, end_position_on_the_genomic_accession=96400588, orientation='minus', exon_count=17, OMIM=603440, no_hdr0=''),
  5129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5129, CurrentID=0, Status='live', Symbol='CDK18', Aliases='PCTAIRE, PCTAIRE3, PCTK3', description='cyclin-dependent kinase 18', other_designations='PCTAIRE protein kinase 3|PCTAIRE-motif protein kinase 3|cell division protein kinase 18|serine/threonine-protein kinase PCTAIRE-3', map_location='1q31-q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205504556, end_position_on_the_genomic_accession=205532793, orientation='plus', exon_count=18, OMIM=169190, no_hdr0=''),
  5130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5130, CurrentID=0, Status='live', Symbol='PCYT1A', Aliases='CCTA, CT, CTA, CTPCT, PCYT1, SMDCRD', description='phosphate cytidylyltransferase 1, choline, alpha', other_designations='CCT A|CCT-alpha|CT A|CTP:phosphocholine cytidylyltransferase A|phosphorylcholine transferase A', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196238382, end_position_on_the_genomic_accession=196287713, orientation='minus', exon_count=10, OMIM=123695, no_hdr0=''),
  5131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5131, CurrentID=0, Status='live', Symbol='PDB1', Aliases='PDB', description='Paget disease of bone 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=167250, no_hdr0=''),
  5132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5132, CurrentID=0, Status='live', Symbol='PDC', Aliases='MEKA, PHD, PhLOP, PhLP', description='phosducin', other_designations='33 kDa phototransducing protein|G beta gamma binding protein|phosducin-like orphan protein', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186443566, end_position_on_the_genomic_accession=186461108, orientation='minus', exon_count=5, OMIM=171490, no_hdr0=''),
  5133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5133, CurrentID=0, Status='live', Symbol='PDCD1', Aliases='CD279, PD-1, PD1, SLEB2, hPD-1, hPD-l, hSLE1', description='programmed cell death 1', other_designations='programmed cell death 1 protein|protein PD-1|systemic lupus erythematosus susceptibility 2', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241849881, end_position_on_the_genomic_accession=241858908, orientation='minus', exon_count=6, OMIM=600244, no_hdr0=''),
  5134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5134, CurrentID=0, Status='live', Symbol='PDCD2', Aliases='RP8, ZMYND7', description='programmed cell death 2', other_designations='zinc finger MYND domain-containing protein 7|zinc finger protein Rp-8', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170575572, end_position_on_the_genomic_accession=170584774, orientation='minus', exon_count=9, OMIM=600866, no_hdr0=''),
  5136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5136, CurrentID=0, Status='live', Symbol='PDE1A', Aliases='CAM-PDE-1A, HCAM-1, HCAM1, HSPDE1A', description='phosphodiesterase 1A', other_designations='61 kDa Cam-PDE|calcium/calmodulin-stimulated cyclic nucleotide phosphodiesterase|cam-PDE 1A|phosphodiesterase 1A, calmodulin-dependent', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182140035, end_position_on_the_genomic_accession=182522845, orientation='minus', exon_count=23, OMIM=171890, no_hdr0=''),
  5137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5137, CurrentID=0, Status='live', Symbol='PDE1C', Aliases='Hcam3, cam-PDE 1C, hCam-3', description='phosphodiesterase 1C', other_designations="Human 3',5' cyclic nucleotide phosphodiesterase (HSPDE1C1A)|phosphodiesterase 1C, calmodulin-dependent 70kDa", map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31616777, end_position_on_the_genomic_accession=32299404, orientation='minus', exon_count=28, OMIM=602987, no_hdr0=''),
  5138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5138, CurrentID=0, Status='live', Symbol='PDE2A', Aliases='CGS-PDE1, PED2A4, cGSPDE, PDE2A', description='phosphodiesterase 2A', other_designations='cGMP-stimulated phosphodiesterase 1|cGMP-stimulated phosphodiesterase 2|cGMP-stimulated phosphodiesterase 4|cyclic GMP-stimulated phosphodiesterase|phosphodiesterase 2A, cGMP-stimulated', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72576140, end_position_on_the_genomic_accession=72674453, orientation='minus', exon_count=35, OMIM=602658, no_hdr0=''),
  5139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5139, CurrentID=0, Status='live', Symbol='PDE3A', Aliases='CGI-PDE, CGI-PDE A, CGI-PDE-A, HTNB', description='phosphodiesterase 3A', other_designations='cAMP phosphodiesterase, myocardial cGMP-inhibited|cyclic GMP-inhibited phosphodiesterase A|phosphodiesterase 3A, cGMP-inhibited', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20368510, end_position_on_the_genomic_accession=20687641, orientation='plus', exon_count=18, OMIM=123805, no_hdr0=''),
  5140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5140, CurrentID=0, Status='live', Symbol='PDE3B', Aliases='HcGIP1, cGIPDE1', description='phosphodiesterase 3B', other_designations='CGI-PDE B|CGIP1|cyclic GMP-inhibited phosphodiesterase B|cyclic nucleotide phosphodiesterase|phosphodiesterase 3B, cGMP-inhibited', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14643656, end_position_on_the_genomic_accession=14872059, orientation='plus', exon_count=16, OMIM=602047, no_hdr0=''),
  5141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5141, CurrentID=0, Status='live', Symbol='PDE4A', Aliases='DPDE2, PDE4, PDE46', description='phosphodiesterase 4A', other_designations='phosphodiesterase 4A, cAMP-specific (dunce|phosphodiesterase E2 dunce homolog, Drosophila|phosphodiesterase isozyme 4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10416773, end_position_on_the_genomic_accession=10469631, orientation='plus', exon_count=22, OMIM=600126, no_hdr0=''),
  5142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5142, CurrentID=0, Status='live', Symbol='PDE4B', Aliases='DPDE4, PDEIVB', description='phosphodiesterase 4B', other_designations='dunce-like phosphodiesterase E4|phosphodiesterase 4B, cAMP-specific (phosphodiesterase E4 dunce homolog, Drosophila)', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65792510, end_position_on_the_genomic_accession=66374579, orientation='plus', exon_count=26, OMIM=600127, no_hdr0=''),
  5143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5143, CurrentID=0, Status='live', Symbol='PDE4C', Aliases='DPDE1', description='phosphodiesterase 4C', other_designations='PDE21|phosphodiesterase 4C, cAMP-specific (phosphodiesterase E1 dunce homolog, Drosophila)|phosphodiesterase E1 dunce homolog', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18207961, end_position_on_the_genomic_accession=18248200, orientation='minus', exon_count=20, OMIM=600128, no_hdr0=''),
  5144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5144, CurrentID=0, Status='live', Symbol='PDE4D', Aliases='ACRDYS2, DPDE3, HSPDE4D, PDE43N2, STRK1, PDE4D', description='phosphodiesterase 4D', other_designations='cAMP-specific phosphodiesterase PDE4D6|phosphodiesterase 4D, cAMP-specific (phosphodiesterase E3 dunce homolog, Drosophila)', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58969038, end_position_on_the_genomic_accession=60522120, orientation='minus', exon_count=37, OMIM=600129, no_hdr0=''),
  5145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5145, CurrentID=0, Status='live', Symbol='PDE6A', Aliases='CGPR-A, PDEA, RP43', description='phosphodiesterase 6A', other_designations='GMP-PDE alpha|PDE V-B1|cGMP phosphodiesterase alpha subunit|phosphodiesterase 6A, cGMP-specific, rod, alpha|rod photoreceptor cGMP phosphodiesterase alpha subunit', map_location='5q31.2-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149857956, end_position_on_the_genomic_accession=149959518, orientation='minus', exon_count=25, OMIM=180071, no_hdr0=''),
  5146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5146, CurrentID=0, Status='live', Symbol='PDE6C', Aliases='ACHM5, COD4, PDEA2', description='phosphodiesterase 6C', other_designations='cGMP phosphodiesterase 6C|phosphodiesterase 6C, cGMP-specific, cone, alpha prime', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93612588, end_position_on_the_genomic_accession=93665673, orientation='plus', exon_count=22, OMIM=600827, no_hdr0=''),
  5147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5147, CurrentID=0, Status='live', Symbol='PDE6D', Aliases='JBTS22, PDED', description='phosphodiesterase 6D', other_designations='GMP-PDE delta|phosphodiesterase 6D, cGMP-specific, rod, delta|protein p17', map_location='2q35-q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231732425, end_position_on_the_genomic_accession=231781327, orientation='minus', exon_count=6, OMIM=602676, no_hdr0=''),
  5148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5148, CurrentID=0, Status='live', Symbol='PDE6G', Aliases='PDEG, RP57', description='phosphodiesterase 6G', other_designations='GMP-PDE gamma|phosphodiesterase 6G, cGMP-specific, rod, gamma|rod cG-PDE G', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81650459, end_position_on_the_genomic_accession=81657054, orientation='minus', exon_count=4, OMIM=180073, no_hdr0=''),
  5149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5149, CurrentID=0, Status='live', Symbol='PDE6H', Aliases='ACHM6, RCD3', description='phosphodiesterase 6H', other_designations='GMP-PDE gamma|phosphodiesterase 6H, cGMP-specific, cone, gamma', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14973022, end_position_on_the_genomic_accession=14981865, orientation='plus', exon_count=4, OMIM=601190, no_hdr0=''),
  5150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5150, CurrentID=0, Status='live', Symbol='PDE7A', Aliases='HCP1, PDE7', description='phosphodiesterase 7A', other_designations='TM22|phosphodiesterase isozyme 7', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65714334, end_position_on_the_genomic_accession=65841734, orientation='minus', exon_count=14, OMIM=171885, no_hdr0=''),
  5151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5151, CurrentID=0, Status='live', Symbol='PDE8A', Aliases='HsT19550', description='phosphodiesterase 8A', other_designations='cAMP-specific cyclic nucleotide phosphodiesterase 8A', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84980513, end_position_on_the_genomic_accession=85139145, orientation='plus', exon_count=26, OMIM=602972, no_hdr0=''),
  5152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5152, CurrentID=0, Status='live', Symbol='PDE9A', Aliases='HSPDE9A2', description='phosphodiesterase 9A', other_designations="CGMP-specific 3',5'-cyclic phosphodiesterase type 9|phosphodiesterase PDE9A21", map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42653752, end_position_on_the_genomic_accession=42775520, orientation='plus', exon_count=25, OMIM=602973, no_hdr0=''),
  5153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5153, CurrentID=0, Status='live', Symbol='PDE1B', Aliases='HEL-S-79p1, PDES1B, PDE1B', description='phosphodiesterase 1B', other_designations='63 kDa Cam-PDE|calcium/calmodulin-stimulated cyclic nucleotide phosphodiesterase|calmodulin-stimulated phosphodiesterase PDE1B1|cam-PDE 1B|epididymis secretory sperm binding protein Li 79p|phosphodiesterase 1B, calmodulin-dependent|presumed 63kDa form of the type 1 cyclic nucleotide phosphodiesterase family known as PDE1B', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54549393, end_position_on_the_genomic_accession=54579239, orientation='plus', exon_count=19, OMIM=171891, no_hdr0=''),
  5154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5154, CurrentID=0, Status='live', Symbol='PDGFA', Aliases='PDGF-A, PDGF1', description='platelet-derived growth factor alpha polypeptide', other_designations='PDGF A-chain|PDGF subunit A|platelet-derived growth factor A chain|platelet-derived growth factor alpha chain', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=497258, end_position_on_the_genomic_accession=520668, orientation='minus', exon_count=9, OMIM=173430, no_hdr0=''),
  5155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5155, CurrentID=0, Status='live', Symbol='PDGFB', Aliases='IBGC5, PDGF-2, PDGF2, SIS, SSV, c-sis', description='platelet-derived growth factor beta polypeptide', other_designations='PDGF subunit B|PDGF, B chain|becaplermin|platelet-derived growth factor 2|platelet-derived growth factor B chain|platelet-derived growth factor beta polypeptide (simian sarcoma viral (v-sis) oncogene homolog)|platelet-derived growth factor, beta polypeptide (oncogene SIS)|proto-oncogene c-Sis', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39223680, end_position_on_the_genomic_accession=39244952, orientation='minus', exon_count=8, OMIM=190040, no_hdr0=''),
  5156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5156, CurrentID=0, Status='live', Symbol='PDGFRA', Aliases='CD140A, PDGFR-2, PDGFR2, RHEPDGFRA', description='platelet-derived growth factor receptor, alpha polypeptide', other_designations='CD140 antigen-like family member A|CD140a antigen|PDGF-R-alpha|PDGFRA/BCR fusion|alpha-type platelet-derived growth factor receptor|platelet-derived growth factor receptor 2|rearranged-in-hypereosinophilia-platelet derived growth factor receptor alpha fusion protein', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54229097, end_position_on_the_genomic_accession=54298245, orientation='plus', exon_count=28, OMIM=173490, no_hdr0=''),
  5157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5157, CurrentID=0, Status='live', Symbol='PDGFRL', Aliases='PDGRL, PRLTS', description='platelet-derived growth factor receptor-like', other_designations='PDGF receptor beta-like tumor suppressor|PDGFR-like protein|platelet-derived growth factor-beta-like tumor suppressor', map_location='8p22-p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17576433, end_position_on_the_genomic_accession=17643133, orientation='plus', exon_count=7, OMIM=604584, no_hdr0=''),
  5158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5158, CurrentID=0, Status='live', Symbol='PDE6B', Aliases='CSNB3, CSNBAD2, PDEB, RP40, rd1', description='phosphodiesterase 6B', other_designations='GMP-PDE beta|rod cGMP-phosphodiesterase beta-subunit', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=587325, end_position_on_the_genomic_accession=670892, orientation='plus', exon_count=27, OMIM=180072, no_hdr0=''),
  5159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5159, CurrentID=0, Status='live', Symbol='PDGFRB', Aliases='CD140B, IBGC4, IMF1, JTK12, KOGS, PDGFR, PDGFR-1, PDGFR1, PENTT', description='platelet-derived growth factor receptor, beta polypeptide', other_designations='CD140 antigen-like family member B|PDGF-R-beta|PDGFR-beta|beta-type platelet-derived growth factor receptor|platelet-derived growth factor receptor 1', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150113839, end_position_on_the_genomic_accession=150155884, orientation='minus', exon_count=26, OMIM=173410, no_hdr0=''),
  5160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5160, CurrentID=0, Status='live', Symbol='PDHA1', Aliases='PDHA, PDHAD, PDHCE1A, PHE1A', description='pyruvate dehydrogenase (lipoamide) alpha 1', other_designations='PDHE1-A type I|pyruvate dehydrogenase E1 subunit|pyruvate dehydrogenase complex, E1-alpha polypeptide 1', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19343893, end_position_on_the_genomic_accession=19361707, orientation='plus', exon_count=12, OMIM=300502, no_hdr0=''),
  5161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5161, CurrentID=0, Status='live', Symbol='PDHA2', Aliases='PDHAL', description='pyruvate dehydrogenase (lipoamide) alpha 2', other_designations='PDHE1-A type II|pyruvate dehydrogenase, E1-alpha polypeptide, testis specific', map_location='4q22-q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95840088, end_position_on_the_genomic_accession=95841474, orientation='plus', exon_count=1, OMIM=179061, no_hdr0=''),
  5162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5162, CurrentID=0, Status='live', Symbol='PDHB', Aliases='PDHBD, PDHE1-B, PHE1B', description='pyruvate dehydrogenase (lipoamide) beta', other_designations='pyruvate dehydrogenase, E1 beta polypeptide', map_location='3p21.1-p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58427630, end_position_on_the_genomic_accession=58433852, orientation='minus', exon_count=11, OMIM=179060, no_hdr0=''),
  5163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5163, CurrentID=0, Status='live', Symbol='PDK1', Aliases='', description='pyruvate dehydrogenase kinase, isozyme 1', other_designations='PDH kinase 1|mitochondrial pyruvate dehydrogenase, lipoamide, kinase isoenzyme 1|pyruvate dehydrogenase kinase, isoenzyme 1', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172555373, end_position_on_the_genomic_accession=172670881, orientation='plus', exon_count=24, OMIM=602524, no_hdr0=''),
  5164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5164, CurrentID=0, Status='live', Symbol='PDK2', Aliases='PDHK2, PDKII', description='pyruvate dehydrogenase kinase, isozyme 2', other_designations='PDH kinase 2|[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 2, mitochondrial|pyruvate dehydrogenase kinase, isoenzyme 2|pyruvate dehydrogenase, lipoamide, kinase isozyme 2, mitochondrial', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50094737, end_position_on_the_genomic_accession=50111369, orientation='plus', exon_count=14, OMIM=602525, no_hdr0=''),
  5165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5165, CurrentID=0, Status='live', Symbol='PDK3', Aliases='CMTX6, GS1-358P8.4', description='pyruvate dehydrogenase kinase, isozyme 3', other_designations='[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 3, mitochondrial|pyruvate dehydrogenase kinase, isoenzyme 3|pyruvate dehydrogenase, lipoamide, kinase isozyme 3, mitochondrial', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24465227, end_position_on_the_genomic_accession=24550466, orientation='plus', exon_count=12, OMIM=300906, no_hdr0=''),
  5166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5166, CurrentID=0, Status='live', Symbol='PDK4', Aliases='', description='pyruvate dehydrogenase kinase, isozyme 4', other_designations='[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 4, mitochondrial|[Pyruvate dehydrogenase [lipoamide]] kinase isozyme 4, mitochondrial|pyruvate dehydrogenase kinase, isoenzyme 4|pyruvate dehydrogenase, lipoamide, kinase isozyme 4, mitochondrial', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95583497, end_position_on_the_genomic_accession=95596613, orientation='minus', exon_count=11, OMIM=602527, no_hdr0=''),
  5167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5167, CurrentID=0, Status='live', Symbol='ENPP1', Aliases='ARHR2, COLED, M6S1, NPP1, NPPS, PC-1, PCA1, PDNP1', description='ectonucleotide pyrophosphatase/phosphodiesterase 1', other_designations='E-NPP 1|Ly-41 antigen|alkaline phosphodiesterase 1|membrane component chromosome 6 surface marker 1|membrane component, chromosome 6, surface marker 1|phosphodiesterase I/nucleotide pyrophosphatase 1|plasma-cell membrane glycoprotein 1|plasma-cell membrane glycoprotein PC-1', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131808016, end_position_on_the_genomic_accession=131895155, orientation='plus', exon_count=25, OMIM=173335, no_hdr0=''),
  5168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5168, CurrentID=0, Status='live', Symbol='ENPP2', Aliases='ATX, ATX-X, AUTOTAXIN, LysoPLD, NPP2, PD-IALPHA, PDNP2', description='ectonucleotide pyrophosphatase/phosphodiesterase 2', other_designations='E-NPP 2|autotaxin-t|extracellular lysophospholipase D|phosphodiesterase I/nucleotide pyrophosphatase 2|plasma lysophospholipase D', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119557077, end_position_on_the_genomic_accession=119673404, orientation='minus', exon_count=29, OMIM=601060, no_hdr0=''),
  5169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5169, CurrentID=0, Status='live', Symbol='ENPP3', Aliases='B10, CD203c, NPP3, PD-IBETA, PDNP3', description='ectonucleotide pyrophosphatase/phosphodiesterase 3', other_designations='dJ1005H11.3 (phosphodiesterase I/nucleotide pyrophosphatase 3)|dJ914N13.3 (phosphodiesterase I/nucleotide pyrophosphatase 3)|gp130RB13-6|phosphodiesterase I/nucleotide pyrophosphatase 3|phosphodiesterase-I beta', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131637235, end_position_on_the_genomic_accession=131747413, orientation='plus', exon_count=25, OMIM=602182, no_hdr0=''),
  5170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5170, CurrentID=0, Status='live', Symbol='PDPK1', Aliases='PDK1, PDPK2, PDPK2P, PRO0461', description='3-phosphoinositide dependent protein kinase 1', other_designations='3-phosphoinositide dependent protein kinase-1|3-phosphoinositide-dependent protein kinase 2 pseudogene|PkB kinase like gene 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2537964, end_position_on_the_genomic_accession=2603191, orientation='plus', exon_count=18, OMIM=605213, no_hdr0=''),
  5171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5171, CurrentID=0, Status='live', Symbol='PDR', Aliases='', description='pigment disorder, reticulate', other_designations='', map_location='Xp22-p21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301220, no_hdr0=''),
  5172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5172, CurrentID=0, Status='live', Symbol='SLC26A4', Aliases='DFNB4, EVA, PDS, TDH2B', description='solute carrier family 26 (anion exchanger), member 4', other_designations='sodium-independent chloride/iodide transporter|truncated solute carrier family 26', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107660635, end_position_on_the_genomic_accession=107717809, orientation='plus', exon_count=23, OMIM=605646, no_hdr0=''),
  5173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5173, CurrentID=0, Status='live', Symbol='PDYN', Aliases='ADCA, PENKB, SCA23', description='prodynorphin', other_designations='beta-neoendorphin-dynorphin|leu-enkephalin|leumorphin|neoendorphin-dynorphin-enkephalin prepropeptide|preprodynorphin|preproenkephalin B|rimorphin', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1978756, end_position_on_the_genomic_accession=1994285, orientation='minus', exon_count=10, OMIM=131340, no_hdr0=''),
  5174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5174, CurrentID=0, Status='live', Symbol='PDZK1', Aliases='CAP70, CLAMP, NHERF-3, NHERF3, PDZD1', description='PDZ domain containing 1', other_designations='CFTR-associated protein of 70 kDa|PDZ-containing kidney protein 1|na/Pi cotransporter C-terminal-associated protein 1|naPi-Cap1|sodium-hydrogen exchanger regulatory factor 3', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145670852, end_position_on_the_genomic_accession=145707400, orientation='minus', exon_count=13, OMIM=603831, no_hdr0=''),
  5175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5175, CurrentID=0, Status='live', Symbol='PECAM1', Aliases="CD31, CD31/EndoCAM, GPIIA', PECA1, PECAM-1, endoCAM", description='platelet/endothelial cell adhesion molecule 1', other_designations='CD31 antigen', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64319415, end_position_on_the_genomic_accession=64413801, orientation='minus', exon_count=18, OMIM=173445, no_hdr0=''),
  5176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5176, CurrentID=0, Status='live', Symbol='SERPINF1', Aliases='EPC-1, OI12, OI6, PEDF, PIG35', description='serpin peptidase inhibitor, clade F (alpha-2 antiplasmin, pigment epithelium derived factor), member 1', other_designations='cell proliferation-inducing gene 35 protein|serine (or cysteine) proteinase inhibitor, clade F (alpha-2 antiplasmin, pigment epithelium derived factor), member 1|serpin F1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1761965, end_position_on_the_genomic_accession=1777565, orientation='plus', exon_count=8, OMIM=172860, no_hdr0=''),
  5177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5177, CurrentID=0, Status='live', Symbol='PEE1', Aliases='PEE, PIH, PREG1', description='preeclampsia/eclampsia 1', other_designations='pregnancy-induced hypertension (pre-eclampsia, eclampsia, toxemia of pregnancy included)', map_location='4q25-qter', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=189800, no_hdr0=''),
  5178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5178, CurrentID=0, Status='live', Symbol='PEG3', Aliases='PW1, ZKSCAN22, ZNF904, ZSCAN24', description='paternally expressed 3', other_designations='Kruppel-type zinc finger protein|zinc finger and SCAN domain-containing protein 24', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56810077, end_position_on_the_genomic_accession=56840726, orientation='minus', exon_count=10, OMIM=601483, no_hdr0=''),
  5179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5179, CurrentID=0, Status='live', Symbol='PENK', Aliases='PE-A, PENK', description='proenkephalin', other_designations='enkephalin A|preproenkephalin', map_location='8q23-q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56440954, end_position_on_the_genomic_accession=56446723, orientation='minus', exon_count=4, OMIM=131330, no_hdr0=''),
  5182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5182, CurrentID=0, Status='live', Symbol='PEPB', Aliases='', description='peptidase B', other_designations='', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=169900, no_hdr0=''),
  5183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5183, CurrentID=0, Status='live', Symbol='PEPC', Aliases='', description='peptidase C', other_designations='', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=170000, no_hdr0=''),
  5184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5184, CurrentID=0, Status='live', Symbol='PEPD', Aliases='PROLIDASE', description='peptidase D', other_designations='X-Pro dipeptidase|aminoacyl-L-proline hydrolase|imidodipeptidase|proline dipeptidase', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33386949, end_position_on_the_genomic_accession=33521893, orientation='minus', exon_count=15, OMIM=613230, no_hdr0=''),
  5185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5185, CurrentID=0, Status='live', Symbol='PEPE', Aliases='', description='peptidase E', other_designations='', map_location='17q23-qter', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=170200, no_hdr0=''),
  5187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5187, CurrentID=0, Status='live', Symbol='PER1', Aliases='PER, RIGUI, hPER', description='period circadian clock 1', other_designations='Period, drosophila, homolog of|circadian clock protein PERIOD 1|circadian pacemaker protein RIGUI|hPER1|period homolog 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8140470, end_position_on_the_genomic_accession=8156405, orientation='minus', exon_count=24, OMIM=602260, no_hdr0=''),
  5188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5188, CurrentID=0, Status='live', Symbol='GATB', Aliases='HSPC199, PET112, PET112L', description='glutamyl-tRNA(Gln) amidotransferase, subunit B', other_designations='PET112 homolog|cytochrome c oxidase assembly factor PET112 homolog|cytochrome oxidase assembly factor PET112 homolog|glu-AdT subunit B', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151670052, end_position_on_the_genomic_accession=151761023, orientation='minus', exon_count=14, OMIM=603645, no_hdr0=''),
  5189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5189, CurrentID=0, Status='live', Symbol='PEX1', Aliases='PBD1A, PBD1B, ZWS, ZWS1', description='peroxisomal biogenesis factor 1', other_designations='Zellweger syndrome|peroxin-1|peroxisome biogenesis disorder protein 1', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92487023, end_position_on_the_genomic_accession=92528531, orientation='minus', exon_count=24, OMIM=602136, no_hdr0=''),
  5190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5190, CurrentID=0, Status='live', Symbol='PEX6', Aliases='PAF-2, PAF2, PBD4A, PDB4B, PXAAA1', description='peroxisomal biogenesis factor 6', other_designations='peroxin-6|peroxisomal AAA-type ATPase 1|peroxisomal-type ATPase 1|peroxisome assembly factor 2', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42963873, end_position_on_the_genomic_accession=42979243, orientation='minus', exon_count=17, OMIM=601498, no_hdr0=''),
  5191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5191, CurrentID=0, Status='live', Symbol='PEX7', Aliases='PBD9B, PTS2R, RCDP1, RD', description='peroxisomal biogenesis factor 7', other_designations='PTS2 receptor|peroxin-7|peroxisomal PTS2 receptor|peroxisomal targeting signal 2 receptor|peroxisome targeting signal 2 receptor', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136822220, end_position_on_the_genomic_accession=136913934, orientation='plus', exon_count=11, OMIM=601757, no_hdr0=''),
  5192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5192, CurrentID=0, Status='live', Symbol='PEX10', Aliases='NALD, PBD6A, PBD6B, RNF69', description='peroxisomal biogenesis factor 10', other_designations='RING finger protein 69|peroxin 10|peroxin-10|peroxisome assembly protein 10', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2404802, end_position_on_the_genomic_accession=2413780, orientation='minus', exon_count=9, OMIM=602859, no_hdr0=''),
  5193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5193, CurrentID=0, Status='live', Symbol='PEX12', Aliases='PAF-3, PBD3A', description='peroxisomal biogenesis factor 12', other_designations='peroxin 12|peroxin-12|peroxisome assembly factor 3', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35574795, end_position_on_the_genomic_accession=35578637, orientation='minus', exon_count=3, OMIM=601758, no_hdr0=''),
  5194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5194, CurrentID=0, Status='live', Symbol='PEX13', Aliases='NALD, PBD11A, PBD11B, ZWS', description='peroxisomal biogenesis factor 13', other_designations='peroxin-13|peroxisomal membrane protein PEX13', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61017677, end_position_on_the_genomic_accession=61051990, orientation='plus', exon_count=4, OMIM=601789, no_hdr0=''),
  5195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5195, CurrentID=0, Status='live', Symbol='PEX14', Aliases='NAPP2, PBD13A, Pex14p, dJ734G22.2', description='peroxisomal biogenesis factor 14', other_designations='NF-E2 associated polypeptide 2|PTS1 receptor docking protein|PTS1 receptor-docking protein|peroxin-14|peroxisomal membrane anchor protein PEX14|peroxisomal membrane anchor protein Pex14p', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10474946, end_position_on_the_genomic_accession=10630758, orientation='plus', exon_count=13, OMIM=601791, no_hdr0=''),
  5196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5196, CurrentID=0, Status='live', Symbol='PF4', Aliases='CXCL4, PF-4, SCYB4', description='platelet factor 4', other_designations='C-X-C motif chemokine 4|chemokine (C-X-C motif) ligand 4|iroplact|oncostatin-A', map_location='4q12-q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73980825, end_position_on_the_genomic_accession=73982124, orientation='minus', exon_count=4, OMIM=173460, no_hdr0=''),
  5197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5197, CurrentID=0, Status='live', Symbol='PF4V1', Aliases='CXCL4L1, CXCL4V1, PF4-ALT, PF4A, SCYB4V1', description='platelet factor 4 variant 1', other_designations='C-X-C motif chemokine 4|PF4alt|PF4var1|platelet factor 4, variant 1 (PF4-like)', map_location='4q12-q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73853296, end_position_on_the_genomic_accession=73854481, orientation='plus', exon_count=3, OMIM=173461, no_hdr0=''),
  5198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5198, CurrentID=0, Status='live', Symbol='PFAS', Aliases='FGAMS, FGAR-AT, FGARAT, PURL', description='phosphoribosylformylglycinamidine synthase', other_designations='FGAM synthase|FGAR amidotransferase|formylglycinamide ribonucleotide amidotransferase|formylglycinamide ribotide amidotransferase|formylglycinamide ribotide synthetase', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8249255, end_position_on_the_genomic_accession=8270491, orientation='plus', exon_count=29, OMIM=602133, no_hdr0=''),
  5199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5199, CurrentID=0, Status='live', Symbol='CFP', Aliases='BFD, PFC, PFD, PROPERDIN', description='complement factor properdin', other_designations='complement factor P|properdin P factor, complement', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47624213, end_position_on_the_genomic_accession=47630305, orientation='minus', exon_count=10, OMIM=300383, no_hdr0=''),
  5201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5201, CurrentID=0, Status='live', Symbol='PFDN1', Aliases='PDF, PFD1', description='prefoldin subunit 1', other_designations='prefoldin 1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140245050, end_position_on_the_genomic_accession=140303104, orientation='minus', exon_count=5, OMIM=604897, no_hdr0=''),
  5202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5202, CurrentID=0, Status='live', Symbol='PFDN2', Aliases='PFD2', description='prefoldin subunit 2', other_designations='prefoldin 2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161100556, end_position_on_the_genomic_accession=161118076, orientation='minus', exon_count=5, OMIM=613466, no_hdr0=''),
  5203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5203, CurrentID=0, Status='live', Symbol='PFDN4', Aliases='C1, PFD4', description='prefoldin subunit 4', other_designations='prefoldin 4|protein C-1', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=54207963, end_position_on_the_genomic_accession=54219953, orientation='plus', exon_count=4, OMIM=604898, no_hdr0=''),
  5204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5204, CurrentID=0, Status='live', Symbol='PFDN5', Aliases='MM-1, MM1, PFD5', description='prefoldin subunit 5', other_designations='c-myc binding protein|myc modulator-1', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53295451, end_position_on_the_genomic_accession=53299450, orientation='plus', exon_count=6, OMIM=604899, no_hdr0=''),
  5205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5205, CurrentID=0, Status='live', Symbol='ATP8B1', Aliases='ATPIC, BRIC, FIC1, ICP1, PFIC, PFIC1', description='ATPase, aminophospholipid transporter, class I, type 8B, member 1', other_designations='ATPase, class I, type 8B, member 1|E1-E2 ATPase|P4-ATPase flippase complex alpha subunit ATP8B1|familial intrahepatic cholestasis type 1|probable phospholipid-transporting ATPase IC', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57646426, end_position_on_the_genomic_accession=57803822, orientation='minus', exon_count=31, OMIM=602397, no_hdr0=''),
  5207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5207, CurrentID=0, Status='live', Symbol='PFKFB1', Aliases='F6PK, HL2K, PFRX', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 1', other_designations='6PF-2-K/Fru-2,6-P2ASE liver isozyme|6PF-2-K/Fru-2,6-P2ase 1|PFK/FBPase 1|fructose-6-phosphate,2-kinase:fructose-2,6-bisphosphatase', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54933134, end_position_on_the_genomic_accession=54999597, orientation='minus', exon_count=20, OMIM=311790, no_hdr0=''),
  5208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5208, CurrentID=0, Status='live', Symbol='PFKFB2', Aliases='PFK-2/FBPase-2', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 2', other_designations='6PF-2-K/Fru-2,6-P2ASE heart-type isozyme|6PF-2-K/Fru-2,6-P2ase 2|PFK/FBPase 2|PFKFB, cardiac|fructose-2,6-bisphosphatase, cardiac isozyme', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207034416, end_position_on_the_genomic_accession=207081023, orientation='plus', exon_count=21, OMIM=171835, no_hdr0=''),
  5209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5209, CurrentID=0, Status='live', Symbol='PFKFB3', Aliases='IPFK2, PFK2, iPFK-2', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 3', other_designations='6-phosphofructo-2-kinase/ fructose-2,6-bisphosphatase|6PF-2-K/Fru-2,6-P2ase 3|6PF-2-K/Fru-2,6-P2ase brain/placenta-type isozyme|PFK/FBPase 3|fructose-6-phosphate,2-kinase/fructose-2, 6-bisphosphatase|inducible 6-phosphofructo-2-kinase/fructose-2,6-bisphosphatase|renal carcinoma antigen NY-REN-56', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6144880, end_position_on_the_genomic_accession=6235545, orientation='plus', exon_count=18, OMIM=605319, no_hdr0=''),
  5210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5210, CurrentID=0, Status='live', Symbol='PFKFB4', Aliases='', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 4', other_designations='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase-4|6PF-2-K/Fru-2,6-P2ase 4|6PF-2-K/Fru-2,6-P2ase testis-type isozyme|PFK/FBPase 4|bifunctional enzyme with kinase and biphosphatase activities', map_location='3p22-p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48517684, end_position_on_the_genomic_accession=48562774, orientation='minus', exon_count=18, OMIM=605320, no_hdr0=''),
  5211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5211, CurrentID=0, Status='live', Symbol='PFKL', Aliases='ATP-PFK, PFK-B, PFK-L', description='phosphofructokinase, liver', other_designations='6-phosphofructokinase type B|6-phosphofructokinase, liver type|ATP-PFK|liver-type 1-phosphofructokinase|phosphofructo-1-kinase isozyme B|phosphohexokinase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44300034, end_position_on_the_genomic_accession=44327381, orientation='plus', exon_count=27, OMIM=171860, no_hdr0=''),
  5212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5212, CurrentID=0, Status='live', Symbol='VIT', Aliases='', description='vitrin', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36696690, end_position_on_the_genomic_accession=36814794, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  5213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5213, CurrentID=0, Status='live', Symbol='PFKM', Aliases='ATP-PFK, GSD7, PFK-1, PFK1, PFKA, PFKX, PPP1R122', description='phosphofructokinase, muscle', other_designations='6-phosphofructo-1-kinase|6-phosphofructokinase type A|6-phosphofructokinase, muscle type|ATP-PFK|PFK-A|phosphofructo-1-kinase isozyme A|phosphofructokinase 1|phosphofructokinase muscle|phosphofructokinase, polypeptide X|phosphofructokinase-M|phosphohexokinase|protein phosphatase 1, regulatory subunit 122', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48105402, end_position_on_the_genomic_accession=48146404, orientation='plus', exon_count=31, OMIM=610681, no_hdr0=''),
  5214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5214, CurrentID=0, Status='live', Symbol='PFKP', Aliases='ATP-PFK, PFK-C, PFK-P, PFKF', description='phosphofructokinase, platelet', other_designations='6-phosphofructokinase type C|6-phosphofructokinase, platelet type|Phosphofructokinase, platelet type|phosphofructo-1-kinase isozyme C|phosphofructokinase 1|phosphohexokinase', map_location='10p15.3-p15.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3067474, end_position_on_the_genomic_accession=3136805, orientation='plus', exon_count=26, OMIM=171840, no_hdr0=''),
  5216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5216, CurrentID=0, Status='live', Symbol='PFN1', Aliases='ALS18', description='profilin 1', other_designations='epididymis tissue protein Li 184a|profilin I', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4945650, end_position_on_the_genomic_accession=4949086, orientation='minus', exon_count=3, OMIM=176610, no_hdr0=''),
  5217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5217, CurrentID=0, Status='live', Symbol='PFN2', Aliases='D3S1319E, PFL', description='profilin 2', other_designations='profilin II', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149964904, end_position_on_the_genomic_accession=149970954, orientation='minus', exon_count=3, OMIM=176590, no_hdr0=''),
  5218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5218, CurrentID=0, Status='live', Symbol='CDK14', Aliases='PFTAIRE1, PFTK1', description='cyclin-dependent kinase 14', other_designations='PFTAIRE protein kinase 1|cell division protein kinase 14|serine/threonine-protein kinase PFTAIRE-1', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90596362, end_position_on_the_genomic_accession=91210590, orientation='plus', exon_count=23, OMIM=610679, no_hdr0=''),
  5222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5222, CurrentID=0, Status='live', Symbol='PGA5', Aliases='Pg5', description='pepsinogen 5, group I (pepsinogen A)', other_designations='pepsin A|pepsinogen A5|pepsinogen-5', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61241157, end_position_on_the_genomic_accession=61251457, orientation='plus', exon_count=9, OMIM=169730, no_hdr0=''),
  5223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5223, CurrentID=0, Status='live', Symbol='PGAM1', Aliases='HEL-S-35, PGAM-B, PGAMA', description='phosphoglycerate mutase 1', other_designations='BPG-dependent PGAM 1|epididymis secretory protein Li 35|phosphoglycerate mutase 1 (brain)|phosphoglycerate mutase A, nonmuscle form|phosphoglycerate mutase isozyme B', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97426270, end_position_on_the_genomic_accession=97433441, orientation='plus', exon_count=4, OMIM=172250, no_hdr0=''),
  5224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5224, CurrentID=0, Status='live', Symbol='PGAM2', Aliases='GSD10, PGAM-M, PGAMM', description='phosphoglycerate mutase 2', other_designations='BPG-dependent PGAM 2|muscle-specific phosphoglycerate mutase|phosphoglycerate mutase 2 (muscle)|phosphoglycerate mutase isozyme M', map_location='7p13-p12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44062727, end_position_on_the_genomic_accession=44065972, orientation='minus', exon_count=3, OMIM=612931, no_hdr0=''),
  5225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5225, CurrentID=0, Status='live', Symbol='PGC', Aliases='PEPC, PGII', description='progastricsin (pepsinogen C)', other_designations='pepsin C|pepsinogen C|pepsinogen group II|preprogastricsin', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41736711, end_position_on_the_genomic_accession=41747401, orientation='minus', exon_count=11, OMIM=169740, no_hdr0=''),
  5226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5226, CurrentID=0, Status='live', Symbol='PGD', Aliases='6PGD', description='phosphogluconate dehydrogenase', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10398992, end_position_on_the_genomic_accession=10420511, orientation='plus', exon_count=13, OMIM=172200, no_hdr0=''),
  100270951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270951, CurrentID=0, Status='live', Symbol='RPL19P14', Aliases='RPL19_9_931', description='ribosomal protein L19 pseudogene 14', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98782012, end_position_on_the_genomic_accession=98782710, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5228, CurrentID=0, Status='live', Symbol='PGF', Aliases='D12S1900L, PLGF, PlGF-2, SHGC-10760, PGF', description='placental growth factor', other_designations='placental growth factor, vascular endothelial growth factor-related protein', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74941830, end_position_on_the_genomic_accession=74955764, orientation='minus', exon_count=7, OMIM=601121, no_hdr0=''),
  5229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5229, CurrentID=0, Status='live', Symbol='PGGT1B', Aliases='BGGI, GGTI', description='protein geranylgeranyltransferase type I, beta subunit', other_designations='CTC-428G20.3|GGTase-I-beta|geranylgeranyl transferase type I subunit beta|geranylgeranyltransferase type I beta subunit|type I protein geranyl-geranyltransferase subunit beta', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115200241, end_position_on_the_genomic_accession=115262887, orientation='minus', exon_count=11, OMIM=602031, no_hdr0=''),
  5230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5230, CurrentID=0, Status='live', Symbol='PGK1', Aliases='HEL-S-68p, MIG10, PGKA', description='phosphoglycerate kinase 1', other_designations='PRP 2|cell migration-inducing gene 10 protein|epididymis secretory sperm binding protein Li 68p|primer recognition protein 2', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78104169, end_position_on_the_genomic_accession=78126827, orientation='plus', exon_count=11, OMIM=311800, no_hdr0=''),
  5231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5231, CurrentID=0, Status='live', Symbol='PGK1P1', Aliases='', description='phosphoglycerate kinase 1, pseudogene 1', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68070077, end_position_on_the_genomic_accession=68071850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5232, CurrentID=0, Status='live', Symbol='PGK2', Aliases='HEL-S-272, PGKB, PGKPS, dJ417L20.2', description='phosphoglycerate kinase 2', other_designations='epididymis secretory protein Li 272|phosphoglycerate kinase autosomal pseudogene|phosphoglycerate kinase, testis specific', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49785651, end_position_on_the_genomic_accession=49787340, orientation='minus', exon_count=1, OMIM=172270, no_hdr0=''),
  5233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5233, CurrentID=0, Status='live', Symbol='PGK1P2', Aliases='PGK2', description='phosphoglycerate kinase 1, pseudogene 2', other_designations='phosphoglycerate kinase-2 (testicular PGK)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12559485, end_position_on_the_genomic_accession=12561278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  136306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136306, CurrentID=0, Status='live', Symbol='SVOPL', Aliases='', description='SVOP-like', other_designations='SV2 related protein homolog-like|SVOP-like protein', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138594285, end_position_on_the_genomic_accession=138701354, orientation='minus', exon_count=17, OMIM=611700, no_hdr0=''),
  136307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136307, CurrentID=0, Status='live', Symbol='IMPDH1P3', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 3', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138440593, end_position_on_the_genomic_accession=138442940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5236, CurrentID=0, Status='live', Symbol='PGM1', Aliases='CDG1T, GSD14', description='phosphoglucomutase 1', other_designations='PGM 1|glucose phosphomutase 1', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63593276, end_position_on_the_genomic_accession=63660245, orientation='plus', exon_count=13, OMIM=171900, no_hdr0=''),
  5238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5238, CurrentID=0, Status='live', Symbol='PGM3', Aliases='AGM1, IMD23, PAGM, PGM 3', description='phosphoglucomutase 3', other_designations='N-acetylglucosamine-phosphate mutase 1|acetylglucosamine phosphomutase', map_location='6q14.1-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83151972, end_position_on_the_genomic_accession=83193936, orientation='minus', exon_count=19, OMIM=172100, no_hdr0=''),
  5239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5239, CurrentID=0, Status='live', Symbol='PGM5', Aliases='PGMRP', description='phosphoglucomutase 5', other_designations='PGM-RP|aciculin|phosphoglucomutase-related protein', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68356899, end_position_on_the_genomic_accession=68531061, orientation='plus', exon_count=14, OMIM=600981, no_hdr0=''),
  5241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5241, CurrentID=0, Status='live', Symbol='PGR', Aliases='NR3C3, PR', description='progesterone receptor', other_designations='nuclear receptor subfamily 3 group C member 3', map_location='11q22-q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101029624, end_position_on_the_genomic_accession=101130668, orientation='minus', exon_count=10, OMIM=607311, no_hdr0=''),
  5243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5243, CurrentID=0, Status='live', Symbol='ABCB1', Aliases='ABC20, CD243, CLCS, GP170, MDR1, P-GP, PGY1', description='ATP binding cassette subfamily B member 1', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 1|P-glycoprotein 1|colchicin sensitivity|doxorubicin resistance', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87503863, end_position_on_the_genomic_accession=87713323, orientation='minus', exon_count=29, OMIM=171050, no_hdr0=''),
  5244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5244, CurrentID=0, Status='live', Symbol='ABCB4', Aliases='ABC21, GBD1, ICP3, MDR2, MDR2/3, MDR3, PFIC-3, PGY3', description='ATP binding cassette subfamily B member 4', other_designations='ATP-binding cassette sub-family B member 4|ATP-binding cassette, sub-family B (MDR/TAP), member 4|P glycoprotein 3/multiple drug resistance 3|P-glycoprotein 3|P-glycoprotein-3/multiple drug resistance-3|multidrug resistance protein 3|multiple drug resistance 3', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87402044, end_position_on_the_genomic_accession=87477523, orientation='minus', exon_count=30, OMIM=171060, no_hdr0=''),
  5245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5245, CurrentID=0, Status='live', Symbol='PHB', Aliases='HEL-215, HEL-S-54e1, PHB', description='prohibitin', other_designations='epididymis luminal protein 215|epididymis secretory sperm binding protein Li 54e', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49404048, end_position_on_the_genomic_accession=49414905, orientation='minus', exon_count=8, OMIM=176705, no_hdr0=''),
  5246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5246, CurrentID=0, Status='live', Symbol='PHBP1', Aliases='', description='prohibitin pseudogene 1', other_designations='', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150042486, end_position_on_the_genomic_accession=150044311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  136319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136319, CurrentID=0, Status='live', Symbol='MTPN', Aliases='GCDP, V-1', description='myotrophin', other_designations='granule cell differentiation protein|protein V-1', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135926755, end_position_on_the_genomic_accession=135977456, orientation='minus', exon_count=4, OMIM=606484, no_hdr0=''),
  105372788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372788, CurrentID=0, Status='live', Symbol='LOC105372788', Aliases='', description='uncharacterized LOC105372788', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33967079, end_position_on_the_genomic_accession=33969220, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5250, CurrentID=0, Status='live', Symbol='SLC25A3', Aliases='OK/SW-cl.48, PHC, PTP', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 3', other_designations='phosphate transport protein', map_location='12q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98593625, end_position_on_the_genomic_accession=98602000, orientation='plus', exon_count=8, OMIM=600370, no_hdr0=''),
  5251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5251, CurrentID=0, Status='live', Symbol='PHEX', Aliases='HPDR, HPDR1, HYP, HYP1, LXHR, PEX, XLH', description='phosphate regulating endopeptidase homolog, X-linked', other_designations='X-linked hypophosphatemia protein|metalloendopeptidase homolog PEX|phosphate regulating gene with homologies to endopeptidases on the X chromosome (hypophosphatemia, vitamin D resistant rickets)|vitamin D-resistant hypophosphatemic rickets protein', map_location='Xp22.2-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22032327, end_position_on_the_genomic_accession=22251310, orientation='plus', exon_count=26, OMIM=300550, no_hdr0=''),
  5252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5252, CurrentID=0, Status='live', Symbol='PHF1', Aliases='MTF2L2, PCL1, PHF2, TDRD19C, hPHF1', description='PHD finger protein 1', other_designations='hPCl1|polycomb-like protein 1|tudor domain containing 19C', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33410412, end_position_on_the_genomic_accession=33416453, orientation='plus', exon_count=18, OMIM=602881, no_hdr0=''),
  5253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5253, CurrentID=0, Status='live', Symbol='PHF2', Aliases='CENP-35, GRC5, JHDM1E', description='PHD finger protein 2', other_designations='centromere protein 35|jumonji C domain-containing histone demethylase 1E', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93576627, end_position_on_the_genomic_accession=93679587, orientation='plus', exon_count=23, OMIM=604351, no_hdr0=''),
  5255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5255, CurrentID=0, Status='live', Symbol='PHKA1', Aliases='PHKA', description='phosphorylase kinase, alpha 1 (muscle)', other_designations='phosphorylase b kinase regulatory subunit alpha skeletal muscle isoform|phosphorylase kinase alpha M subunit|phosphorylase kinase, alpha 1 (muscle), muscle glycogenosis', map_location='Xq12-q13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72578814, end_position_on_the_genomic_accession=72714181, orientation='minus', exon_count=32, OMIM=311870, no_hdr0=''),
  5256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5256, CurrentID=0, Status='live', Symbol='PHKA2', Aliases='GSD9A, PHK, PYK, PYKL, XLG, XLG2', description='phosphorylase kinase, alpha 2 (liver)', other_designations='phosphorylase b kinase regulatory subunit alpha liver isoform|phosphorylase kinase alpha L subunit|phosphorylase kinase alpha-subunit', map_location='Xp22.2-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18892298, end_position_on_the_genomic_accession=18984362, orientation='minus', exon_count=33, OMIM=300798, no_hdr0=''),
  5257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5257, CurrentID=0, Status='live', Symbol='PHKB', Aliases='', description='phosphorylase kinase, beta', other_designations='phosphorylase kinase beta-subunit|phosphorylase kinase subunit beta', map_location='16q12-q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47461268, end_position_on_the_genomic_accession=47701523, orientation='plus', exon_count=34, OMIM=172490, no_hdr0=''),
  5258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5258, CurrentID=0, Status='live', Symbol='PHKBP1', Aliases='', description='phosphorylase kinase, beta pseudogene 1', other_designations='', map_location='20p12.3-p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8097724, end_position_on_the_genomic_accession=8099886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5259, CurrentID=0, Status='live', Symbol='PHKBP2', Aliases='', description='phosphorylase kinase, beta pseudogene 2', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36533827, end_position_on_the_genomic_accession=36534353, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5260, CurrentID=0, Status='live', Symbol='PHKG1', Aliases='PHKG', description='phosphorylase kinase, gamma 1 (muscle)', other_designations='PHK-gamma-M|phosphorylase b kinase gamma catalytic chain, skeletal muscle isoform|phosphorylase kinase gamma|phosphorylase kinase subunit gamma-1|serine/threonine-protein kinase PHKG1', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56080283, end_position_on_the_genomic_accession=56093396, orientation='minus', exon_count=11, OMIM=172470, no_hdr0=''),
  5261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5261, CurrentID=0, Status='live', Symbol='PHKG2', Aliases='GSD9C', description='phosphorylase kinase, gamma 2 (testis)', other_designations='PHK-gamma-LT|PHK-gamma-T|PSK-C3|Phosphorylase kinase, gamma 2 (testis/liver)|phosphorylase b kinase gamma catalytic chain, testis/liver isoform|phosphorylase kinase subunit gamma-2|serine/threonine-protein kinase PHKG2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30748299, end_position_on_the_genomic_accession=30761176, orientation='plus', exon_count=10, OMIM=172471, no_hdr0=''),
  5262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5262, CurrentID=0, Status='live', Symbol='PHKG1P3', Aliases='PHKGL, PHKg1ps2', description='phosphorylase kinase, gamma 1 pseudogene 3', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50246477, end_position_on_the_genomic_accession=50247762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5264, CurrentID=0, Status='live', Symbol='PHYH', Aliases='LN1, LNAP1, PAHX1, RD, PHYH', description='phytanoyl-CoA 2-hydroxylase', other_designations='phytanic acid oxidase|phytanoil-CoA alpha hydroxylase|phytanoyl-CoA 2 oxoglutarate dioxygenase|phytanoyl-CoA alpha-hydroxylase', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13277796, end_position_on_the_genomic_accession=13300138, orientation='minus', exon_count=10, OMIM=602026, no_hdr0=''),
  5265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5265, CurrentID=0, Status='live', Symbol='SERPINA1', Aliases='A1A, A1AT, AAT, PI, PI1, PRO2275, alpha1AT', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 1', other_designations='alpha-1 protease inhibitor|alpha-1-antiproteinase|alpha-1-antitrypsin null|protease inhibitor 1 (anti-elastase), alpha-1-antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 1|serpin A1|serpin peptidase inhibitor clade A member 1', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94376747, end_position_on_the_genomic_accession=94390692, orientation='minus', exon_count=7, OMIM=107400, no_hdr0=''),
  5266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5266, CurrentID=0, Status='live', Symbol='PI3', Aliases='ESI, SKALP, WAP3, WFDC14, cementoin', description='peptidase inhibitor 3', other_designations='PI-3|WAP four-disulfide core domain 14|WAP four-disulfide core domain protein 14|elastase-specific inhibitor|pre-elafin|protease inhibitor 3, skin-derived (SKALP)|protease inhibitor WAP3|skin-derived antileukoproteinase|trappin-2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45174899, end_position_on_the_genomic_accession=45176544, orientation='plus', exon_count=3, OMIM=182257, no_hdr0=''),
  5267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5267, CurrentID=0, Status='live', Symbol='SERPINA4', Aliases='KAL, KLST, KST, PI-4, PI4, kallistatin', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 4', other_designations='kallikrein inhibitor|peptidase inhibitor 4|protease inhibitor 4 (kallistatin)|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 4|serpin A4', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94561420, end_position_on_the_genomic_accession=94569913, orientation='plus', exon_count=5, OMIM=147935, no_hdr0=''),
  5268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5268, CurrentID=0, Status='live', Symbol='SERPINB5', Aliases='PI5, maspin', description='serpin peptidase inhibitor, clade B (ovalbumin), member 5', other_designations='PI-5|peptidase inhibitor 5|protease inhibitor 5 (maspin)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 5', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63476911, end_position_on_the_genomic_accession=63505085, orientation='plus', exon_count=8, OMIM=154790, no_hdr0=''),
  5269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5269, CurrentID=0, Status='live', Symbol='SERPINB6', Aliases='CAP, DFNB91, MSTP057, PI-6, PI6, PTI, SPI3', description='serpin peptidase inhibitor, clade B (ovalbumin), member 6', other_designations='cytoplasmic antiproteinase|protease inhibitor 6 (placental thrombin inhibitor)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 6', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2948159, end_position_on_the_genomic_accession=2972165, orientation='minus', exon_count=13, OMIM=173321, no_hdr0=''),
  5270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5270, CurrentID=0, Status='live', Symbol='SERPINE2', Aliases='GDN, GDNPF, PI-7, PI7, PN-1, PN1, PNI', description='serpin peptidase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 2', other_designations='glial-derived neurite promoting factor|peptidase inhibitor 7|protease nexin I|serpin E2', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223975048, end_position_on_the_genomic_accession=224039319, orientation='minus', exon_count=14, OMIM=177010, no_hdr0=''),
  5271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5271, CurrentID=0, Status='live', Symbol='SERPINB8', Aliases='CAP2, PI8', description='serpin peptidase inhibitor, clade B (ovalbumin), member 8', other_designations='PI-8|cytoplasmic antiproteinase 2|peptidase inhibitor 8|protease inhibitor 8 (ovalbumin type)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 8', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63970028, end_position_on_the_genomic_accession=64006821, orientation='plus', exon_count=10, OMIM=601697, no_hdr0=''),
  5272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5272, CurrentID=0, Status='live', Symbol='SERPINB9', Aliases='CAP-3, CAP3, PI-9, PI9', description='serpin peptidase inhibitor, clade B (ovalbumin), member 9', other_designations='cytoplasmic antiproteinase 3|peptidase inhibitor 9|protease inhibitor 9 (ovalbumin type)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 9|serpin peptidase inhibitor, clade B, member 9', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2887265, end_position_on_the_genomic_accession=2903312, orientation='minus', exon_count=9, OMIM=601799, no_hdr0=''),
  5273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5273, CurrentID=0, Status='live', Symbol='SERPINB10', Aliases='PI-10, PI10', description='serpin peptidase inhibitor, clade B (ovalbumin), member 10', other_designations='bomapin|peptidase inhibitor 10|protease inhibitor 10 (ovalbumin type, bomapin)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 10', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63907969, end_position_on_the_genomic_accession=63936111, orientation='plus', exon_count=10, OMIM=602058, no_hdr0=''),
  5274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5274, CurrentID=0, Status='live', Symbol='SERPINI1', Aliases='PI12, neuroserpin', description='serpin peptidase inhibitor, clade I (neuroserpin), member 1', other_designations='PI-12|peptidase inhibitor 12|serine (or cysteine) proteinase inhibitor, clade I (neuroserpin), member 1|serpin I1|serpin peptidase inhibitor clade I member 1', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167735644, end_position_on_the_genomic_accession=167825569, orientation='plus', exon_count=9, OMIM=602445, no_hdr0=''),
  5275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5275, CurrentID=0, Status='live', Symbol='SERPINB13', Aliases='HSHUR7SEQ, HUR7, PI13, headpin', description='serpin peptidase inhibitor, clade B (ovalbumin), member 13', other_designations='HURPIN|PI-13|UV-B repressed sequence, HUR 7|haCaT UV-repressible serpin|peptidase inhibitor 13|protease inhibitor 13 (hurpin, headpin)|proteinase inhibitor 13|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 13', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63587202, end_position_on_the_genomic_accession=63599199, orientation='plus', exon_count=9, OMIM=604445, no_hdr0=''),
  5276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5276, CurrentID=0, Status='live', Symbol='SERPINI2', Aliases='MEPI, PANCPIN, PI14, TSA2004', description='serpin peptidase inhibitor, clade I (pancpin), member 2', other_designations='myoepithelium-derived serine protease inhibitor|pancreas-specific protein TSA2004|peptidase inhibitor 14|protease inhibitor 14|serine (or cysteine) proteinase inhibitor, clade I (neuroserpin), member 2|serine (or cysteine) proteinase inhibitor, clade I (pancpin), member 2', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167441789, end_position_on_the_genomic_accession=167474132, orientation='minus', exon_count=10, OMIM=605587, no_hdr0=''),
  5277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5277, CurrentID=0, Status='live', Symbol='PIGA', Aliases='GPI3, MCAHS2, PIG-A, PNH1', description='phosphatidylinositol glycan anchor biosynthesis class A', other_designations='GLCNAC-PI synthesis protein|GPI anchor biosynthesis|class A GlcNAc-inositol phospholipid assembly protein|phosphatidylinositol glycan anchor biosynthesis, class A|phosphatidylinositol-glycan biosynthesis, class A protein', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15319451, end_position_on_the_genomic_accession=15335554, orientation='minus', exon_count=7, OMIM=311770, no_hdr0=''),
  5278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5278, CurrentID=0, Status='live', Symbol='PIGAP1', Aliases='PIG-A-II', description='phosphatidylinositol glycan anchor biosynthesis class A pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class A pseudogene|phosphatidylinositol glycan anchor biosynthesis, class A, pseudogene 1|phosphatidylinositol glycan, class A, pseudogene 1', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100578634, end_position_on_the_genomic_accession=100582672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5279, CurrentID=0, Status='live', Symbol='PIGC', Aliases='GPI2', description='phosphatidylinositol glycan anchor biosynthesis class C', other_designations='PIG-C|phosphatidylinositol glycan anchor biosynthesis, class C|phosphatidylinositol-glycan biosynthesis class C protein|phosphatidylinositol-glycan biosynthesis, class C protein', map_location='1q23-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172441457, end_position_on_the_genomic_accession=172444090, orientation='minus', exon_count=2, OMIM=601730, no_hdr0=''),
  5280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5280, CurrentID=0, Status='live', Symbol='PIGCP1', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class C pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class C pseudogene|phosphatidylinositol glycan anchor biosynthesis, class C, pseudogene 1|phosphatidylinositol glycan, class C, pseudogene 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33075086, end_position_on_the_genomic_accession=33076718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5281, CurrentID=0, Status='live', Symbol='PIGF', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class F', other_designations='GPI11 homolog|PIG-F|phosphatidylinositol glycan anchor biosynthesis, class F', map_location='2p21-p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46581274, end_position_on_the_genomic_accession=46617112, orientation='minus', exon_count=8, OMIM=600153, no_hdr0=''),
  5282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5282, CurrentID=0, Status='live', Symbol='PIGFP1', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class F pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class F pseudogene|phosphatidylinositol glycan anchor biosynthesis, class F, pseudogene 1|phosphatidylinositol glycan, class F, pseudogene 1', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178922020, end_position_on_the_genomic_accession=178922914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5283, CurrentID=0, Status='live', Symbol='PIGH', Aliases='GPI-H', description='phosphatidylinositol glycan anchor biosynthesis class H', other_designations='PIG-H|phosphatidylinositol glycan anchor biosynthesis, class H|phosphatidylinositol glycan, class H|phosphatidylinositol-glycan biosynthesis class H protein|phosphatidylinositol-glycan biosynthesis, class H protein', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67589306, end_position_on_the_genomic_accession=67600300, orientation='minus', exon_count=5, OMIM=600154, no_hdr0=''),
  5284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5284, CurrentID=0, Status='live', Symbol='PIGR', Aliases='', description='polymeric immunoglobulin receptor', other_designations='hepatocellular carcinoma associated protein TB6|hepatocellular carcinoma-associated protein TB6|poly-Ig receptor', map_location='1q31-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206928522, end_position_on_the_genomic_accession=206946466, orientation='minus', exon_count=12, OMIM=173880, no_hdr0=''),
  5286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5286, CurrentID=0, Status='live', Symbol='PIK3C2A', Aliases='CPK, PI3-K-C2(ALPHA), PI3-K-C2A', description='phosphatidylinositol-4-phosphate 3-kinase catalytic subunit type 2 alpha', other_designations='C2-containing phosphatidylinositol kinase|PI3K-C2-alpha|PI3K-C2alpha|phosphatidylinositol-4-phosphate 3-kinase C2 domain-containing subunit alpha|phosphatidylinositol-4-phosphate 3-kinase, catalytic subunit type 2 alpha|phosphoinositide 3-kinase-C2-alpha|phosphoinositide-3-kinase, class 2, alpha polypeptide|ptdIns-3-kinase C2 subunit alpha', map_location='11p15.5-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17086561, end_position_on_the_genomic_accession=17207996, orientation='minus', exon_count=34, OMIM=603601, no_hdr0=''),
  5287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5287, CurrentID=0, Status='live', Symbol='PIK3C2B', Aliases='C2-PI3K', description='phosphatidylinositol-4-phosphate 3-kinase catalytic subunit type 2 beta', other_designations='PI3K-C2-beta|PI3K-C2beta|PTDINS-3-kinase C2 beta|phosphatidylinositol 3-kinase C2 domain-containing beta polypeptide|phosphatidylinositol-4-phosphate 3-kinase C2 domain-containing subunit beta|phosphatidylinositol-4-phosphate 3-kinase, catalytic subunit type 2 beta|phosphoinositide 3-kinase-C2-beta|phosphoinositide-3-kinase, class 2, beta polypeptide|ptdIns-3-kinase C2 subunit beta', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204422628, end_position_on_the_genomic_accession=204494815, orientation='minus', exon_count=37, OMIM=602838, no_hdr0=''),
  5288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5288, CurrentID=0, Status='live', Symbol='PIK3C2G', Aliases='PI3K-C2-gamma, PI3K-C2GAMMA', description='phosphatidylinositol-4-phosphate 3-kinase catalytic subunit type 2 gamma', other_designations='PTDINS-3-kinase C2 gamma|phosphatidylinositol-4-phosphate 3-kinase C2 domain-containing gamma polypeptide|phosphatidylinositol-4-phosphate 3-kinase C2 domain-containing subunit gamma|phosphatidylinositol-4-phosphate 3-kinase, catalytic subunit type 2 gamma|phosphoinositide-3-kinase, class 2, gamma polypeptide|ptdIns-3-kinase C2 subunit gamma', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18242965, end_position_on_the_genomic_accession=18699660, orientation='plus', exon_count=43, OMIM=609001, no_hdr0=''),
  5289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5289, CurrentID=0, Status='live', Symbol='PIK3C3', Aliases='VPS34, Vps34, hVps34', description='phosphatidylinositol 3-kinase catalytic subunit type 3', other_designations='PI3-kinase type 3|PI3K type 3|PtdIns-3-kinase type 3|phosphatidylinositol 3-kinase p100 subunit|phosphatidylinositol 3-kinase, catalytic subunit type 3|phosphoinositide-3-kinase, class 3', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41955199, end_position_on_the_genomic_accession=42084606, orientation='plus', exon_count=26, OMIM=602609, no_hdr0=''),
  5290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5290, CurrentID=0, Status='live', Symbol='PIK3CA', Aliases='CLOVE, CWS5, MCAP, MCM, MCMTC, PI3K, p110-alpha', description='phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit alpha', other_designations='PI3-kinase p110 subunit alpha|PI3K-alpha|phosphatidylinositol 3-kinase, catalytic, 110-KD, alpha|phosphatidylinositol 3-kinase, catalytic, alpha polypeptide|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit alpha|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit, alpha isoform|phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit alpha|phosphoinositide-3-kinase, catalytic, alpha polypeptide|ptdIns-3-kinase subunit p110-alpha|serine/threonine protein kinase PIK3CA', map_location='3q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179148114, end_position_on_the_genomic_accession=179240084, orientation='plus', exon_count=23, OMIM=171834, no_hdr0=''),
  5291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5291, CurrentID=0, Status='live', Symbol='PIK3CB', Aliases='P110BETA, PI3K, PI3KBETA, PIK3C1', description='phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit beta', other_designations='PI3-kinase p110 subunit beta|PI3-kinase subunit beta|PI3K-beta|PtdIns-3-kinase p110|phosphatidylinositol 4,5-bisphosphate 3-kinase 110 kDa catalytic subunit beta|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit beta|phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit beta|phosphoinositide-3-kinase, catalytic, beta polypeptide|ptdIns-3-kinase subunit beta|ptdIns-3-kinase subunit p110-beta', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138652698, end_position_on_the_genomic_accession=138834938, orientation='minus', exon_count=25, OMIM=602925, no_hdr0=''),
  5292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5292, CurrentID=0, Status='live', Symbol='PIM1', Aliases='PIM', description='Pim-1 proto-oncogene, serine/threonine kinase', other_designations='Oncogene PIM1|pim-1 kinase 44 kDa isoform|pim-1 oncogene (proviral integration site 1)|proto-oncogene serine/threonine-protein kinase pim-1', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37170146, end_position_on_the_genomic_accession=37175428, orientation='plus', exon_count=6, OMIM=164960, no_hdr0=''),
  5293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5293, CurrentID=0, Status='live', Symbol='PIK3CD', Aliases='APDS, IMD14, P110DELTA, PI3K, p110D', description='phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit delta', other_designations='PI3-kinase p110 subunit delta|PI3Kdelta|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit delta|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit delta isoform|phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit delta|phosphoinositide-3-kinase C|phosphoinositide-3-kinase, catalytic, delta polypeptide variant p37delta|ptdIns-3-kinase subunit p110-delta', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9629889, end_position_on_the_genomic_accession=9729114, orientation='plus', exon_count=30, OMIM=602839, no_hdr0=''),
  5294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5294, CurrentID=0, Status='live', Symbol='PIK3CG', Aliases='PI3CG, PI3K, PI3Kgamma, PIK3, p110gamma, p120-PI3K', description='phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit gamma', other_designations='1-phosphatidylinositol 3-kinase|PI3-kinase subunit gamma|phosphatidylinositol 3 kinase gamma, p110 gamma|phosphatidylinositol 3-kinase catalytic 110-kD gamma|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit gamma|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit gamma isoform|phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit gamma|phosphoinositide-3-kinase gamma catalytic subunit|ptdIns-3-kinase subunit gamma|ptdIns-3-kinase subunit p110-gamma|serine/threonine protein kinase PIK3CG', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106865278, end_position_on_the_genomic_accession=106908978, orientation='plus', exon_count=14, OMIM=601232, no_hdr0=''),
  5295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5295, CurrentID=0, Status='live', Symbol='PIK3R1', Aliases='AGM7, GRB1, IMD36, p85, p85-ALPHA', description='phosphoinositide-3-kinase regulatory subunit 1', other_designations='PI3-kinase subunit p85-alpha|PI3K regulatory subunit alpha|phosphatidylinositol 3-kinase 85 kDa regulatory subunit alpha|phosphatidylinositol 3-kinase, regulatory subunit, polypeptide 1 (p85 alpha)|phosphatidylinositol 3-kinase-associated p-85 alpha|phosphoinositide-3-kinase regulatory subunit|phosphoinositide-3-kinase, regulatory subunit 1 (alpha)|ptdIns-3-kinase regulatory subunit alpha', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68215756, end_position_on_the_genomic_accession=68301821, orientation='plus', exon_count=20, OMIM=171833, no_hdr0=''),
  5296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5296, CurrentID=0, Status='live', Symbol='PIK3R2', Aliases='MPPH, MPPH1, P85B, p85, p85-BETA', description='phosphoinositide-3-kinase regulatory subunit 2', other_designations='PI3-kinase subunit p85-beta|PI3K regulatory subunit beta|phosphatidylinositol 3-kinase 85 kDa regulatory subunit beta|phosphatidylinositol 3-kinase, regulatory subunit, polypeptide 2 (p85 beta)|phosphoinositide-3-kinase, regulatory subunit 2 (beta)|ptdIns-3-kinase regulatory subunit p85-beta', map_location='19q13.2-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18153178, end_position_on_the_genomic_accession=18170533, orientation='plus', exon_count=16, OMIM=603157, no_hdr0=''),
  5297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5297, CurrentID=0, Status='live', Symbol='PI4KA', Aliases='PI4K-ALPHA, PIK4CA, PMGYCHA, pi4K230', description='phosphatidylinositol 4-kinase, catalytic, alpha', other_designations='PI4-kinase alpha|phosphatidylinositol 4-kinase 230|phosphatidylinositol 4-kinase, type III, alpha|ptdIns-4-kinase alpha', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20707691, end_position_on_the_genomic_accession=20858812, orientation='minus', exon_count=57, OMIM=600286, no_hdr0=''),
  5298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5298, CurrentID=0, Status='live', Symbol='PI4KB', Aliases='NPIK, PI4K-BETA, PI4K92ETA, PI4KIIIBETA, PIK4CB, PI4KB', description='phosphatidylinositol 4-kinase, catalytic, beta', other_designations='PtdIns 4-kinase beta|phosphatidylinositol 4-kinase, wortmannin-sensitive|type III phosphatidylinositol 4-kinase beta', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151291797, end_position_on_the_genomic_accession=151327715, orientation='minus', exon_count=15, OMIM=602758, no_hdr0=''),
  136371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136371, CurrentID=0, Status='live', Symbol='ASB10', Aliases='GLC1F', description='ankyrin repeat and SOCS box containing 10', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151175698, end_position_on_the_genomic_accession=151187832, orientation='minus', exon_count=8, OMIM=615054, no_hdr0=''),
  5300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5300, CurrentID=0, Status='live', Symbol='PIN1', Aliases='DOD, UBL5', description='peptidylprolyl cis/trans isomerase, NIMA-interacting 1', other_designations='PPIase Pin1|protein (peptidyl-prolyl cis/trans isomerase) NIMA-interacting 1|rotamase Pin1', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9835207, end_position_on_the_genomic_accession=9849689, orientation='plus', exon_count=6, OMIM=601052, no_hdr0=''),
  5301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5301, CurrentID=0, Status='live', Symbol='PIN1P1', Aliases='PIN1L', description='peptidylprolyl cis/trans isomerase, NIMA-interacting 1 pseudogene 1', other_designations='peptidylprolyl cis/trans isomerase, NIMA-interacting 1-like (pseudogene)|protein (peptidylprolyl cis/trans isomerase) NIMA-interacting 1-like (pseudogene)', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69919322, end_position_on_the_genomic_accession=69920317, orientation='plus', exon_count=1, OMIM=602051, no_hdr0=''),
  5303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5303, CurrentID=0, Status='live', Symbol='PIN4', Aliases='EPVH, PAR14, PAR17', description='peptidylprolyl cis/trans isomerase, NIMA-interacting 4', other_designations='PPIase PIN4|eukaryotic parvulin homolog|hEPVH|hPar14|hPar17|parvulin|parvulin-14|parvulin-17|peptidyl-prolyl cis-trans isomerase Pin4|peptidyl-prolyl cis/trans isomerase EPVH|peptidylprolyl cis/trans isomerase NIMA-interacting, 4|protein (peptidylprolyl cis/trans isomerase) NIMA-interacting, 4 (parvulin)|rotamase PIN4', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72181676, end_position_on_the_genomic_accession=72263964, orientation='plus', exon_count=5, OMIM=300252, no_hdr0=''),
  5304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5304, CurrentID=0, Status='live', Symbol='PIP', Aliases='GCDFP-15, GCDFP15, GPIP4', description='prolactin-induced protein', other_designations='SABP|gross cystic disease fluid protein 15|secretory actin-binding protein', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143132081, end_position_on_the_genomic_accession=143139741, orientation='plus', exon_count=4, OMIM=176720, no_hdr0=''),
  5305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5305, CurrentID=0, Status='live', Symbol='PIP4K2A', Aliases='PI5P4KA, PIP5K2A, PIP5KII-alpha, PIP5KIIA, PIPK', description='phosphatidylinositol-5-phosphate 4-kinase, type II, alpha', other_designations='1-phosphatidylinositol 5-phosphate 4-kinase 2-alpha|1-phosphatidylinositol-4-phosphate kinase|1-phosphatidylinositol-4-phosphate-5-kinase|1-phosphatidylinositol-5-phosphate 4-kinase 2-alpha|PI(5)P 4-kinase type II alpha|PIP4KII-alpha|PIP5KIII|PIP5KIIalpha|PtdIns(4)P-5-kinase B isoform|diphosphoinositide kinase 2-alpha|phosphatidylinositol 5-phosphate 4-kinase type II alpha|phosphatidylinositol-4-phosphate 5-kinase, type II, alpha|phosphatidylinositol-5-phosphate 4-kinase type-2 alpha|ptdIns(4)P-5-kinase C isoform|ptdIns(5)P-4-kinase isoform 2-alpha|type II phosphatidylinositol-4-phosphate 5-kinase 53 K isoform', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22534837, end_position_on_the_genomic_accession=22714574, orientation='minus', exon_count=14, OMIM=603140, no_hdr0=''),
  5306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5306, CurrentID=0, Status='live', Symbol='PITPNA', Aliases='HEL-S-36, PI-TPalpha, PITPN, VIB1A', description='phosphatidylinositol transfer protein alpha', other_designations='PI-TP-alpha|epididymis secretory protein Li 36|phosphatidylinositol transfer protein, alpha|ptdIns transfer protein alpha|ptdInsTP alpha', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1517989, end_position_on_the_genomic_accession=1562816, orientation='minus', exon_count=13, OMIM=600174, no_hdr0=''),
  5307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5307, CurrentID=0, Status='live', Symbol='PITX1', Aliases='BFT, CCF, LBNBG, POTX, PTX1', description='paired-like homeodomain 1', other_designations='hindlimb expressed homeobox protein backfoot|hindlimb-expressed homeobox protein backfoot|homeobox protein PITX1|paired-like homeodomain transcription factor 1|pituitary homeo box 1|pituitary otx-related factor', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135027734, end_position_on_the_genomic_accession=135034274, orientation='minus', exon_count=3, OMIM=602149, no_hdr0=''),
  5308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5308, CurrentID=0, Status='live', Symbol='PITX2', Aliases='ARP1, Brx1, IDG2, IGDS, IGDS2, IHG2, IRID2, Otlx2, PTX2, RGS, RIEG, RIEG1, RS', description='paired-like homeodomain 2', other_designations='ALL1-responsive protein ARP1|all1-responsive gene 1|homeobox protein PITX2|paired-like homeodomain transcription factor 2|rieg bicoid-related homeobox transcription factor 1|solurshin', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110617423, end_position_on_the_genomic_accession=110642123, orientation='minus', exon_count=9, OMIM=601542, no_hdr0=''),
  5309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5309, CurrentID=0, Status='live', Symbol='PITX3', Aliases='ASMD, ASOD, CTPP4, CTRCT11, PTX3', description='paired-like homeodomain 3', other_designations='homeobox protein PITX3', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102230189, end_position_on_the_genomic_accession=102241474, orientation='minus', exon_count=5, OMIM=602669, no_hdr0=''),
  5310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5310, CurrentID=0, Status='live', Symbol='PKD1', Aliases='PBP, Pc-1, TRPP1', description='polycystic kidney disease 1 (autosomal dominant)', other_designations='autosomal dominant polycystic kidney disease 1 protein|polycystic kidney disease-associated protein|polycystin 1|transient receptor potential cation channel, subfamily P, member 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2088708, end_position_on_the_genomic_accession=2135898, orientation='minus', exon_count=53, OMIM=601313, no_hdr0=''),
  5311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5311, CurrentID=0, Status='live', Symbol='PKD2', Aliases='APKD2, PC2, PKD4, Pc-2, TRPP2', description='polycystic kidney disease 2 (autosomal dominant)', other_designations='R48321|autosomal dominant polycystic kidney disease type II protein|polycystwin|transient receptor potential cation channel subfamily P member 2|transient receptor potential cation channel, subfamily P, member 2', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88007639, end_position_on_the_genomic_accession=88077779, orientation='plus', exon_count=17, OMIM=173910, no_hdr0=''),
  5312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5312, CurrentID=0, Status='live', Symbol='PKD3', Aliases='', description='polycystic kidney disease 3 (autosomal dominant)', other_designations='', map_location='2p', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600666, no_hdr0=''),
  5313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5313, CurrentID=0, Status='live', Symbol='PKLR', Aliases='PK1, PKL, PKR, PKRL, RPK', description='pyruvate kinase, liver and RBC', other_designations='R-type/L-type pyruvate kinase|pyruvate kinase 1|pyruvate kinase isozyme R/L|pyruvate kinase isozymes L/R|pyruvate kinase isozymes R/L|pyruvate kinase type L|pyruvate kinase, liver and blood cell|red cell/liver pyruvate kinase', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155289293, end_position_on_the_genomic_accession=155308709, orientation='minus', exon_count=13, OMIM=609712, no_hdr0=''),
  5314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5314, CurrentID=0, Status='live', Symbol='PKHD1', Aliases='ARPKD, FCYT, TIGM1', description='polycystic kidney and hepatic disease 1 (autosomal recessive)', other_designations='TIG multiple domains 1|polycystic kidney and hepatic disease 1 protein|polyductin|tigmin', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51614685, end_position_on_the_genomic_accession=52087625, orientation='minus', exon_count=75, OMIM=606702, no_hdr0=''),
  5315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5315, CurrentID=0, Status='live', Symbol='PKM', Aliases='CTHBP, HEL-S-30, OIP3, PK32, TCB, THBP1, PKM', description='pyruvate kinase, muscle', other_designations='OIP-3|OPA-interacting protein 3|PK, muscle type|cytosolic thyroid hormone-binding protein|epididymis secretory protein Li 30|p58|pyruvate kinase 2/3|pyruvate kinase isozymes M1/M2|pyruvate kinase muscle isozyme|thyroid hormone-binding protein 1|thyroid hormone-binding protein, cytosolic|tumor M2-PK', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72199029, end_position_on_the_genomic_accession=72231812, orientation='minus', exon_count=16, OMIM=179050, no_hdr0=''),
  5316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5316, CurrentID=0, Status='live', Symbol='PKNOX1', Aliases='PREP1, pkonx1c', description='PBX/knotted 1 homeobox 1', other_designations='PBX/knotted homeobox 1|Pbx regulating protein-1|homeobox protein PREP-1|human homeobox-containing protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42974510, end_position_on_the_genomic_accession=43033931, orientation='plus', exon_count=11, OMIM=602100, no_hdr0=''),
  5317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5317, CurrentID=0, Status='live', Symbol='PKP1', Aliases='B6P', description='plakophilin 1', other_designations='band 6 protein', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201283452, end_position_on_the_genomic_accession=201332993, orientation='plus', exon_count=15, OMIM=601975, no_hdr0=''),
  5318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5318, CurrentID=0, Status='live', Symbol='PKP2', Aliases='ARVD9', description='plakophilin 2', other_designations='', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32790746, end_position_on_the_genomic_accession=32896846, orientation='minus', exon_count=14, OMIM=602861, no_hdr0=''),
  5319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5319, CurrentID=0, Status='live', Symbol='PLA2G1B', Aliases='PLA2, PLA2A, PPLA2', description='phospholipase A2 group IB', other_designations='phosphatidylcholine 2-acylhydrolase 1B|phospholipase A2, group IB (pancreas)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120322111, end_position_on_the_genomic_accession=120327789, orientation='minus', exon_count=4, OMIM=172410, no_hdr0=''),
  5320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5320, CurrentID=0, Status='live', Symbol='PLA2G2A', Aliases='MOM1, PLA2, PLA2B, PLA2L, PLA2S, PLAS1, sPLA2', description='phospholipase A2 group IIA', other_designations='GIIC sPLA2|NPS-PLA2|group IIA phospholipase A2|non-pancreatic secretory phospholipase A2|phosphatidylcholine 2-acylhydrolase 2A|phospholipase A2, group IIA (platelets, synovial fluid)', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19975431, end_position_on_the_genomic_accession=19980439, orientation='minus', exon_count=6, OMIM=172411, no_hdr0=''),
  5321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5321, CurrentID=0, Status='live', Symbol='PLA2G4A', Aliases='PLA2G4, cPLA2, cPLA2-alpha', description='phospholipase A2 group IVA', other_designations='calcium-dependent phospholipid-binding protein|lysophospholipase|phosphatidylcholine 2-acylhydrolase|phospholipase A2, group IVA (cytosolic, calcium-dependent)', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186828900, end_position_on_the_genomic_accession=186988981, orientation='plus', exon_count=21, OMIM=600522, no_hdr0=''),
  5322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5322, CurrentID=0, Status='live', Symbol='PLA2G5', Aliases='FRFB, GV-PLA2, PLA2-10, hVPLA(2)', description='phospholipase A2 group V', other_designations='Ca2+-dependent phospholipase A2|phosphatidylcholine 2-acylhydrolase 5|phospholipase A2, group V', map_location='1p36-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20028339, end_position_on_the_genomic_accession=20091901, orientation='plus', exon_count=14, OMIM=601192, no_hdr0=''),
  100533111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533111, CurrentID=0, Status='live', Symbol='ZNF670-ZNF695', Aliases='', description='ZNF670-ZNF695 readthrough (NMD candidate)', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246945547, end_position_on_the_genomic_accession=247078813, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5324, CurrentID=0, Status='live', Symbol='PLAG1', Aliases='PSA, SGPA, ZNF912', description='PLAG1 zinc finger', other_designations='COL1A2/PLAG1 fusion|HAS2/PLAG1 fusion', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56160909, end_position_on_the_genomic_accession=56211314, orientation='minus', exon_count=6, OMIM=603026, no_hdr0=''),
  5325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5325, CurrentID=0, Status='live', Symbol='PLAGL1', Aliases='LOT1, ZAC, ZAC1', description='PLAG1 like zinc finger 1', other_designations='PLAG-like 1|lost on transformation 1|pleiomorphic adenoma gene-like 1|tumor suppressor ZAC', map_location='6q24-q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143940300, end_position_on_the_genomic_accession=144064599, orientation='minus', exon_count=10, OMIM=603044, no_hdr0=''),
  5326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5326, CurrentID=0, Status='live', Symbol='PLAGL2', Aliases='ZNF900', description='PLAG1 like zinc finger 2', other_designations='C2H2-type zinc finger protein|pleiomorphic adenoma gene-like 2|pleiomorphic adenoma-like protein 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32192504, end_position_on_the_genomic_accession=32207743, orientation='minus', exon_count=6, OMIM=604866, no_hdr0=''),
  5327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5327, CurrentID=0, Status='live', Symbol='PLAT', Aliases='T-PA, TPA', description='plasminogen activator, tissue', other_designations='alteplase|plasminogen activator, tissue type|plasminogen/activator kringle|reteplase|t-plasminogen activator|tissue plasminogen activator (t-PA)', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42174718, end_position_on_the_genomic_accession=42207676, orientation='minus', exon_count=14, OMIM=173370, no_hdr0=''),
  5328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5328, CurrentID=0, Status='live', Symbol='PLAU', Aliases='ATF, BDPLT5, QPD, UPA, URK, u-PA', description='plasminogen activator, urokinase', other_designations='U-plasminogen activator|plasminogen activator, urinary', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73909182, end_position_on_the_genomic_accession=73917501, orientation='plus', exon_count=12, OMIM=191840, no_hdr0=''),
  5329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5329, CurrentID=0, Status='live', Symbol='PLAUR', Aliases='CD87, U-PAR, UPAR, URKR', description='plasminogen activator, urokinase receptor', other_designations='monocyte activation antigen Mo3|u-plasminogen activator receptor form 2|urokinase-type plasminogen activator (uPA) receptor', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43646095, end_position_on_the_genomic_accession=43670346, orientation='minus', exon_count=9, OMIM=173391, no_hdr0=''),
  5330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5330, CurrentID=0, Status='live', Symbol='PLCB2', Aliases='PLC-beta-2', description='phospholipase C beta 2', other_designations='phosphoinositide phospholipase C-beta-2|phospholipase C, beta 2', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40284260, end_position_on_the_genomic_accession=40307973, orientation='minus', exon_count=36, OMIM=604114, no_hdr0=''),
  5331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5331, CurrentID=0, Status='live', Symbol='PLCB3', Aliases='', description='phospholipase C beta 3', other_designations='PLC beta 3|PLC-beta-3|phosphoinositide phospholipase C-beta-3|phospholipase C, beta 3 (phosphatidylinositol-specific)', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64251523, end_position_on_the_genomic_accession=64269452, orientation='plus', exon_count=31, OMIM=600230, no_hdr0=''),
  5332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5332, CurrentID=0, Status='live', Symbol='PLCB4', Aliases='ARCND2, PI-PLC', description='phospholipase C beta 4', other_designations='1-phosphatidyl-D-myo-inositol-4,5-bisphosphate|PLC-beta-4|dJ1119D9.2 (Phopholipase C, beta 4 (1-Phosphatidylinositol-4,5-Bisphosphate Phosphodiesterase Beta 4))|inositoltrisphosphohydrolase|monophosphatidylinositol phosphodiesterase|phosphoinositidase C|phosphoinositide phospholipase C-beta-4|phospholipase C, beta 4|triphosphoinositide phosphodiesterase', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9068710, end_position_on_the_genomic_accession=9480816, orientation='plus', exon_count=43, OMIM=600810, no_hdr0=''),
  5333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5333, CurrentID=0, Status='live', Symbol='PLCD1', Aliases='NDNC3, PLC-III', description='phospholipase C delta 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase delta-1|PLC-delta-1|phosphoinositide phospholipase C-delta-1|phospholipase C, delta 1|phospholipase C-III', map_location='3p22-p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38007496, end_position_on_the_genomic_accession=38029663, orientation='minus', exon_count=16, OMIM=602142, no_hdr0=''),
  5334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5334, CurrentID=0, Status='live', Symbol='PLCL1', Aliases='PLCE, PLCL, PLDL1, PPP1R127, PRIP', description='phospholipase C like 1', other_designations='PLC-L1|phospholipase C related, but catalytically inactive protein|phospholipase C, epsilon|phospholipase C-deleted in lung carcinoma|phospholipase C-like 1|phospholipase C-related but catalytically inactive protein|protein phosphatase 1, regulatory subunit 127', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197804602, end_position_on_the_genomic_accession=198149884, orientation='plus', exon_count=9, OMIM=600597, no_hdr0=''),
  5335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5335, CurrentID=0, Status='live', Symbol='PLCG1', Aliases='NCKAP3, PLC-II, PLC1, PLC148, PLCgamma1', description='phospholipase C gamma 1', other_designations='1-phosphatidyl-D-myo-inositol-4,5-bisphosphate|1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase gamma 1|1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase gamma-1|PLC-148|PLC-gamma-1|inositoltrisphosphohydrolase|monophosphatidylinositol phosphodiesterase|phosphatidylinositol phospholipase C|phosphoinositidase C|phosphoinositide phospholipase C|phosphoinositide phospholipase C-gamma-1|phospholipase C, gamma 1 (formerly subtype 148)|phospholipase C-148|phospholipase C-II|phospholipase C-gamma-1|triphosphoinositide phosphodiesterase', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41137519, end_position_on_the_genomic_accession=41175721, orientation='plus', exon_count=32, OMIM=172420, no_hdr0=''),
  5336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5336, CurrentID=0, Status='live', Symbol='PLCG2', Aliases='APLAID, FCAS3, PLC-IV, PLC-gamma-2', description='phospholipase C gamma 2', other_designations='phosphoinositide phospholipase C-gamma-2|phospholipase C, gamma 2 (phosphatidylinositol-specific)|phospholipase C-IV', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81779258, end_position_on_the_genomic_accession=81962693, orientation='plus', exon_count=35, OMIM=600220, no_hdr0=''),
  5337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5337, CurrentID=0, Status='live', Symbol='PLD1', Aliases='', description='phospholipase D1', other_designations='choline phosphatase 1|phosphatidylcholine-hydrolyzing phospholipase D1|phospholipase D1, phosphatidylcholine-specific', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171600405, end_position_on_the_genomic_accession=171810494, orientation='minus', exon_count=33, OMIM=602382, no_hdr0=''),
  5338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5338, CurrentID=0, Status='live', Symbol='PLD2', Aliases='', description='phospholipase D2', other_designations='PLD1C|choline phosphatase 2|hPLD2|phosphatidylcholine-hydrolyzing phospholipase D2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4807101, end_position_on_the_genomic_accession=4823432, orientation='plus', exon_count=25, OMIM=602384, no_hdr0=''),
  5339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5339, CurrentID=0, Status='live', Symbol='PLEC', Aliases='EBS1, EBSMD, EBSND, EBSO, EBSOG, EBSPA, HD1, LGMD2Q, PCN1, PLEC1b, PLTN, PLEC', description='plectin', other_designations='hemidesmosomal protein 1|plectin 1, intermediate filament binding protein 500kDa', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143915147, end_position_on_the_genomic_accession=143976792, orientation='minus', exon_count=43, OMIM=601282, no_hdr0=''),
  5340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5340, CurrentID=0, Status='live', Symbol='PLG', Aliases='', description='plasminogen', other_designations='plasmin', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160702193, end_position_on_the_genomic_accession=160754054, orientation='plus', exon_count=19, OMIM=173350, no_hdr0=''),
  5341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5341, CurrentID=0, Status='live', Symbol='PLEK', Aliases='P47', description='pleckstrin', other_designations='platelet 47 kDa protein', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68365190, end_position_on_the_genomic_accession=68397453, orientation='plus', exon_count=9, OMIM=173570, no_hdr0=''),
  5342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5342, CurrentID=0, Status='live', Symbol='PLGLB2', Aliases='PLGP1', description='plasminogen-like B2', other_designations='plasminogen pseudogene 1|plasminogen-related protein B|type B plasminogen related', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87748087, end_position_on_the_genomic_accession=87759476, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5343, CurrentID=0, Status='live', Symbol='PLGLB1', Aliases='PLGL, PLGP1, PRGB, PRP-B', description='plasminogen-like B1', other_designations='plasminogen-related protein B|type B plasminogen related', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87010464, end_position_on_the_genomic_accession=87021846, orientation='minus', exon_count=4, OMIM=173340, no_hdr0=''),
  5345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5345, CurrentID=0, Status='live', Symbol='SERPINF2', Aliases='A2AP, AAP, ALPHA-2-PI, API, PLI', description='serpin peptidase inhibitor, clade F (alpha-2 antiplasmin, pigment epithelium derived factor), member 2', other_designations='alpha-2-AP|alpha-2-plasmin inhibitor|serine (or cysteine) proteinase inhibitor, clade F (alpha-2 antiplasmin, pigment epithelium derived factor), member 2|serpin F2', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1742807, end_position_on_the_genomic_accession=1755265, orientation='plus', exon_count=13, OMIM=613168, no_hdr0=''),
  5346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5346, CurrentID=0, Status='live', Symbol='PLIN1', Aliases='FPLD4, PERI, PLIN', description='perilipin 1', other_designations='lipid droplet-associated protein', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89664365, end_position_on_the_genomic_accession=89679417, orientation='minus', exon_count=10, OMIM=170290, no_hdr0=''),
  5347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5347, CurrentID=0, Status='live', Symbol='PLK1', Aliases='PLK, STPK13', description='polo-like kinase 1', other_designations='PLK-1|cell cycle regulated protein kinase|polo (Drosophia)-like kinase|polo like kinase|serine/threonine-protein kinase 13', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23678822, end_position_on_the_genomic_accession=23690367, orientation='plus', exon_count=10, OMIM=602098, no_hdr0=''),
  5348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5348, CurrentID=0, Status='live', Symbol='FXYD1', Aliases='PLM', description='FXYD domain containing ion transport regulator 1', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35138789, end_position_on_the_genomic_accession=35143055, orientation='plus', exon_count=11, OMIM=602359, no_hdr0=''),
  5349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5349, CurrentID=0, Status='live', Symbol='FXYD3', Aliases='MAT8, PLML', description='FXYD domain containing ion transport regulator 3', other_designations='chloride conductance inducer protein Mat-8|mammary tumor 8 kDa protein|phospholemman-like protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35115816, end_position_on_the_genomic_accession=35124324, orientation='plus', exon_count=12, OMIM=604996, no_hdr0=''),
  5350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5350, CurrentID=0, Status='live', Symbol='PLN', Aliases='CMD1P, CMH18, PLB', description='phospholamban', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118548279, end_position_on_the_genomic_accession=118560424, orientation='plus', exon_count=2, OMIM=172405, no_hdr0=''),
  5351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5351, CurrentID=0, Status='live', Symbol='PLOD1', Aliases='EDS6, LH, LH1, LLH, PLOD', description='procollagen-lysine, 2-oxoglutarate 5-dioxygenase 1', other_designations='lysine hydroxylase|lysyl hydroxlase 1|procollagen-lysine 1, 2-oxoglutarate 5-dioxygenase 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11934667, end_position_on_the_genomic_accession=11975542, orientation='plus', exon_count=21, OMIM=153454, no_hdr0=''),
  5352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5352, CurrentID=0, Status='live', Symbol='PLOD2', Aliases='BRKS2, LH2, TLH', description='procollagen-lysine, 2-oxoglutarate 5-dioxygenase 2', other_designations='lysine hydroxylase 2|lysyl hydroxlase 2|lysyl hydroxylase 2|telopeptide lysyl hydroxylase', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146069436, end_position_on_the_genomic_accession=146161495, orientation='minus', exon_count=22, OMIM=601865, no_hdr0=''),
  100270972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270972, CurrentID=0, Status='live', Symbol='RPL29P7', Aliases='RPL29_3_121', description='ribosomal protein L29 pseudogene 7', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168938416, end_position_on_the_genomic_accession=168939047, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5354, CurrentID=0, Status='live', Symbol='PLP1', Aliases='GPM6C, HLD1, MMPL, PLP, PLP/DM20, PMD, SPG2', description='proteolipid protein 1', other_designations='lipophilin|major myelin proteolipid protein', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103776506, end_position_on_the_genomic_accession=103792619, orientation='plus', exon_count=8, OMIM=300401, no_hdr0=''),
  5355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5355, CurrentID=0, Status='live', Symbol='PLP2', Aliases='A4, A4LSB', description='proteolipid protein 2 (colonic epithelium-enriched)', other_designations='A4 differentiation-dependent protein|differentiation-dependent protein A4|intestinal membrane A4 protein', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49171837, end_position_on_the_genomic_accession=49175120, orientation='plus', exon_count=5, OMIM=300112, no_hdr0=''),
  5356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5356, CurrentID=0, Status='live', Symbol='PLRG1', Aliases='Cwc1, PRL1, PRP46, PRPF46, TANGO4', description='pleiotropic regulator 1', other_designations='pleiotropic regulator 1 (PRL1 homolog, Arabidopsis)|transport and golgi organization 4 homolog', map_location='4q31.2-q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154534997, end_position_on_the_genomic_accession=154550433, orientation='minus', exon_count=15, OMIM=605961, no_hdr0=''),
  5357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5357, CurrentID=0, Status='live', Symbol='PLS1', Aliases='', description='plastin 1', other_designations='I plastin|fimbrin|intestine specific plastin|intestine-specific plastin', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142596387, end_position_on_the_genomic_accession=142713664, orientation='plus', exon_count=22, OMIM=602734, no_hdr0=''),
  5358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5358, CurrentID=0, Status='live', Symbol='PLS3', Aliases='BMND18, T-plastin', description='plastin 3', other_designations='T fimbrin|T plastin', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115560850, end_position_on_the_genomic_accession=115650861, orientation='plus', exon_count=21, OMIM=300131, no_hdr0=''),
  5359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5359, CurrentID=0, Status='live', Symbol='PLSCR1', Aliases='MMTRA1B', description='phospholipid scramblase 1', other_designations='PL scramblase 1|ca(2+)-dependent phospholipid scramblase 1|erythrocyte phospholipid scramblase', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146515178, end_position_on_the_genomic_accession=146544841, orientation='minus', exon_count=10, OMIM=604170, no_hdr0=''),
  5360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5360, CurrentID=0, Status='live', Symbol='PLTP', Aliases='BPIFE, HDLCQ9', description='phospholipid transfer protein', other_designations='BPI fold containing family E|lipid transfer protein II', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45898620, end_position_on_the_genomic_accession=45912364, orientation='minus', exon_count=17, OMIM=172425, no_hdr0=''),
  5361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5361, CurrentID=0, Status='live', Symbol='PLXNA1', Aliases='NOV, NOVP, PLEXIN-A1, PLXN1', description='plexin A1', other_designations='plexin 1|semaphorin receptor NOV', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126982119, end_position_on_the_genomic_accession=127037392, orientation='plus', exon_count=33, OMIM=601055, no_hdr0=''),
  5362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5362, CurrentID=0, Status='live', Symbol='PLXNA2', Aliases='OCT, PLXN2', description='plexin A2', other_designations='plexin 2|semaphorin receptor OCT|transmembrane protein OCT', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208022242, end_position_on_the_genomic_accession=208244384, orientation='minus', exon_count=35, OMIM=601054, no_hdr0=''),
  5364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5364, CurrentID=0, Status='live', Symbol='PLXNB1', Aliases='PLEXIN-B1, PLXN5, SEP', description='plexin B1', other_designations='plexin 5|semaphorin receptor SEP', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48403854, end_position_on_the_genomic_accession=48430324, orientation='minus', exon_count=43, OMIM=601053, no_hdr0=''),
  5365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5365, CurrentID=0, Status='live', Symbol='PLXNB3', Aliases='PLEXB3, PLEXR, PLXN6', description='plexin B3', other_designations='plexin 6', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153764196, end_position_on_the_genomic_accession=153779346, orientation='plus', exon_count=37, OMIM=300214, no_hdr0=''),
  5366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5366, CurrentID=0, Status='live', Symbol='PMAIP1', Aliases='APR, NOXA', description='phorbol-12-myristate-13-acetate-induced protein 1', other_designations='PMA-induced protein 1|adult T cell leukemia-derived PMA-responsive|immediate-early-response protein APR|protein Noxa', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59899960, end_position_on_the_genomic_accession=59904306, orientation='plus', exon_count=2, OMIM=604959, no_hdr0=''),
  5367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5367, CurrentID=0, Status='live', Symbol='PMCH', Aliases='MCH, ppMCH', description='pro-melanin concentrating hormone', other_designations='prepro-MCH|prepro-melanin-concentrating hormone|pro-melanin-concentrating hormone', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102196459, end_position_on_the_genomic_accession=102197836, orientation='minus', exon_count=3, OMIM=176795, no_hdr0=''),
  5368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5368, CurrentID=0, Status='live', Symbol='PNOC', Aliases='N/OFQ, NOP, OFQ, PPNOC, ppN/OFQ', description='prepronociceptin', other_designations='nociceptin|nocistatin|orphanin FQ|pre-pro-N/OFQ|pronociceptin', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28316991, end_position_on_the_genomic_accession=28343355, orientation='plus', exon_count=5, OMIM=601459, no_hdr0=''),
  5369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5369, CurrentID=0, Status='live', Symbol='PMCHL1', Aliases='', description='pro-melanin concentrating hormone like 1 (pseudogene)', other_designations='pro-melanin-concentrating hormone-like 1, pseudogene', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=22142352, end_position_on_the_genomic_accession=22152272, orientation='plus', exon_count=4, OMIM=176793, no_hdr0=''),
  5370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5370, CurrentID=0, Status='live', Symbol='PMCHL2', Aliases='', description='pro-melanin concentrating hormone like 2 (pseudogene)', other_designations='pro-melanin-concentrating hormone-like 2, pseudogene', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71375785, end_position_on_the_genomic_accession=71385993, orientation='plus', exon_count=4, OMIM=176794, no_hdr0=''),
  5371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5371, CurrentID=0, Status='live', Symbol='PML', Aliases='MYL, PP8675, RNF71, TRIM19', description='promyelocytic leukemia', other_designations='PML/RARA fusion|RING finger protein 71|probable transcription factor PML|promyelocytic leukemia protein|promyelocytic leukemia, inducer of|tripartite motif protein TRIM19|tripartite motif-containing protein 19', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73994673, end_position_on_the_genomic_accession=74047819, orientation='plus', exon_count=10, OMIM=102578, no_hdr0=''),
  5372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5372, CurrentID=0, Status='live', Symbol='PMM1', Aliases='Sec53', description='phosphomannomutase 1', other_designations='PMM 1|PMMH-22|brain glucose-1,6-bisphosphatase', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41576886, end_position_on_the_genomic_accession=41589867, orientation='minus', exon_count=11, OMIM=601786, no_hdr0=''),
  5373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5373, CurrentID=0, Status='live', Symbol='PMM2', Aliases='CDG1, CDG1a, CDGS, PMI, PMI1, PMM 2', description='phosphomannomutase 2', other_designations='phosphomannose isomerase 1', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8797813, end_position_on_the_genomic_accession=8855197, orientation='plus', exon_count=10, OMIM=601785, no_hdr0=''),
  5375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5375, CurrentID=0, Status='live', Symbol='PMP2', Aliases='FABP8, M-FABP, MP2, P2', description='peripheral myelin protein 2', other_designations='', map_location='8q21.3-q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81440326, end_position_on_the_genomic_accession=81447484, orientation='minus', exon_count=4, OMIM=170715, no_hdr0=''),
  5376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5376, CurrentID=0, Status='live', Symbol='PMP22', Aliases='CMT1A, CMT1E, DSS, GAS-3, HMSNIA, HNPP, Sp110', description='peripheral myelin protein 22', other_designations='growth arrest-specific protein 3|peripheral myelin protein 22 kDa', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15229777, end_position_on_the_genomic_accession=15265357, orientation='minus', exon_count=7, OMIM=601097, no_hdr0=''),
  100270976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270976, CurrentID=0, Status='live', Symbol='RPL30P2', Aliases='RPL30_1_286', description='ribosomal protein L30 pseudogene 2', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152131498, end_position_on_the_genomic_accession=152131889, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5378, CurrentID=0, Status='live', Symbol='PMS1', Aliases='HNPCC3, MLH2, PMSL1, hPMS1', description='PMS1 homolog 1, mismatch repair system component', other_designations='DNA mismatch repair protein PMS1|PMS1 postmeiotic segregation increased 1|human homolog of yeast mutL|mismatch repair gene PMSL1|rhabdomyosarcoma antigen MU-RMS-40.10B|rhabdomyosarcoma antigen MU-RMS-40.10E', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189784085, end_position_on_the_genomic_accession=189877629, orientation='plus', exon_count=16, OMIM=600258, no_hdr0=''),
  5379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5379, CurrentID=0, Status='live', Symbol='PMS2P1', Aliases='PMS2L1, PMS2L13, PMS2L6, PMS2L7, PMS2L8, PMS3, PMS8, PMSR1, PMSR2', description='PMS1 homolog 2, mismatch repair system component pseudogene 1', other_designations='postmeiotic segregation increased 2 pseudogene 1|postmeiotic segregation increased 2-like 1 pseudogene|postmeiotic segregation increased 2-like 5|postmeiotic segregation increased 2-like 6|postmeiotic segregation increased 2-like 7|postmeiotic segregation increased 2-like 8', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100320640, end_position_on_the_genomic_accession=100336307, orientation='minus', exon_count=7, OMIM=605038, no_hdr0=''),
  5380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5380, CurrentID=0, Status='live', Symbol='PMS2P2', Aliases='PMS2L2, PMS4', description='PMS1 homolog 2, mismatch repair system component pseudogene 2', other_designations='postmeiotic segregation increased 2 pseudogene 2|postmeiotic segregation increased 2-like 2 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75343937, end_position_on_the_genomic_accession=75358997, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5382, CurrentID=0, Status='live', Symbol='PMS2P4', Aliases='PMS2L4, PMS6', description='PMS1 homolog 2, mismatch repair system component pseudogene 4', other_designations='PMS2 pseudogene|postmeiotic segregation increased 2 pseudogene 4|postmeiotic segregation increased 2-like 4 pseudogene', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67276131, end_position_on_the_genomic_accession=67302442, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5383, CurrentID=0, Status='live', Symbol='PMS2P5', Aliases='PMS2L5, PMS7', description='PMS1 homolog 2, mismatch repair system component pseudogene 5', other_designations='postmeiotic segregation increased 2 pseudogene 5|postmeiotic segregation increased 2-like 5', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74890768, end_position_on_the_genomic_accession=74921138, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  5387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5387, CurrentID=0, Status='live', Symbol='PMS2P3', Aliases='PMS2L3, PMS2L9, PMS5, PMSR3', description='PMS1 homolog 2, mismatch repair system component pseudogene 3', other_designations='postmeiotic segregation increased 2 pseudogene 3|postmeiotic segregation increased 2-like 3, pseudogene|postmeiotic segregation increased 2-like 9', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75507747, end_position_on_the_genomic_accession=75528123, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100270978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270978, CurrentID=0, Status='live', Symbol='RPL30P5', Aliases='RPL30_2_497', description='ribosomal protein L30 pseudogene 5', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83502699, end_position_on_the_genomic_accession=83503034, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5393, CurrentID=0, Status='live', Symbol='EXOSC9', Aliases='PM/Scl-75, PMSCL1, RRP45, Rrp45p, p5, p6', description='exosome component 9', other_designations='P75 polymyositis-scleroderma overlap syndrome associated autoantigen|P75 polymyositis-scleroderma overlap syndrome-associated autoantigen|PMSCL autoantigen, 75kD|autoantigen PM/Scl 1|exosome complex exonuclease RRP45|polymyositis/scleroderma autoantigen 1, 75kDa', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121801295, end_position_on_the_genomic_accession=121817021, orientation='plus', exon_count=14, OMIM=606180, no_hdr0=''),
  5394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5394, CurrentID=0, Status='live', Symbol='EXOSC10', Aliases='PM-Scl, PM/Scl-100, PMSCL, PMSCL2, RRP6, Rrp6p, p2, p3, p4', description='exosome component 10', other_designations='P100 polymyositis-scleroderma overlap syndrome-associated autoantigen|autoantigen PM-SCL|polymyositis/scleroderma autoantigen 100 kDa|polymyositis/scleroderma autoantigen 2', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11066613, end_position_on_the_genomic_accession=11099910, orientation='minus', exon_count=26, OMIM=605960, no_hdr0=''),
  5395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5395, CurrentID=0, Status='live', Symbol='PMS2', Aliases='HNPCC4, MLH4CL, PMSL2, PMS2', description='PMS1 homolog 2, mismatch repair system component', other_designations='DNA mismatch repair protein PMS2|H_DJ0042M02.9|PMS1 homolog 2, mismatch repair protein|PMS1 protein homolog 2|PMS2 postmeiotic segregation increased 2|postmeiotic segregation increased 2 nirs variant 6', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5970925, end_position_on_the_genomic_accession=6009106, orientation='minus', exon_count=18, OMIM=600259, no_hdr0=''),
  5396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5396, CurrentID=0, Status='live', Symbol='PRRX1', Aliases='AGOTC, PHOX1, PMX1, PRX-1, PRX1', description='paired related homeobox 1', other_designations='homeobox protein PHOX1|paired mesoderm homeobox 1 isoform pmx-1b', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170662728, end_position_on_the_genomic_accession=170739419, orientation='plus', exon_count=6, OMIM=167420, no_hdr0=''),
  100270980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270980, CurrentID=0, Status='live', Symbol='RPL30P6', Aliases='RPL30_4_501', description='ribosomal protein L30 pseudogene 6', other_designations='', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95644952, end_position_on_the_genomic_accession=95645290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422695, CurrentID=0, Status='live', Symbol='LOC100422695', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 5 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77071728, end_position_on_the_genomic_accession=77072491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5406, CurrentID=0, Status='live', Symbol='PNLIP', Aliases='PLD, PTL, PNLIP', description='pancreatic lipase', other_designations='triacylglycerol acylhydrolase', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116545916, end_position_on_the_genomic_accession=116567855, orientation='plus', exon_count=13, OMIM=246600, no_hdr0=''),
  5407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5407, CurrentID=0, Status='live', Symbol='PNLIPRP1', Aliases='PLRP1', description='pancreatic lipase-related protein 1', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116590415, end_position_on_the_genomic_accession=116609175, orientation='plus', exon_count=15, OMIM=604422, no_hdr0=''),
  5408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5408, CurrentID=0, Status='live', Symbol='PNLIPRP2', Aliases='PLRP2', description='pancreatic lipase-related protein 2', other_designations='PL-RP2|galactolipase', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116620953, end_position_on_the_genomic_accession=116645143, orientation='plus', exon_count=13, OMIM=604423, no_hdr0=''),
  5409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5409, CurrentID=0, Status='live', Symbol='PNMT', Aliases='PENTase, PNMT', description='phenylethanolamine N-methyltransferase', other_designations='noradrenaline N-methyltransferase|phenylethanolamine N-methylase', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39667981, end_position_on_the_genomic_accession=39670475, orientation='plus', exon_count=5, OMIM=171190, no_hdr0=''),
  5411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5411, CurrentID=0, Status='live', Symbol='PNN', Aliases='DRS, DRSP, SDK3, memA', description='pinin, desmosome associated protein', other_designations='140 kDa nuclear and cell adhesion-related phosphoprotein|SR-like protein|desmosome-associated protein|domain-rich serine protein|melanoma metastasis clone A protein|neutrophil protein|nuclear protein SDK3', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39175183, end_position_on_the_genomic_accession=39183218, orientation='plus', exon_count=9, OMIM=603154, no_hdr0=''),
  5412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5412, CurrentID=0, Status='live', Symbol='UBL3', Aliases='HCG-1, PNSC1', description='ubiquitin like 3', other_designations='MUB|hsMUB|membrane-anchored ubiquitin-fold protein|protein HCG-1|ubiquitin-like 3', map_location='13q12-q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29764405, end_position_on_the_genomic_accession=29850683, orientation='minus', exon_count=6, OMIM=604711, no_hdr0=''),
  5413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5413, CurrentID=0, Status='live', Symbol='SEPT5', Aliases='CDCREL, CDCREL-1, CDCREL1, H5, HCDCREL-1, PNUTL1', description='septin 5', other_designations='cell division control related protein 1|peanut-like 1|platelet glycoprotein Ib beta chain', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19714464, end_position_on_the_genomic_accession=19723322, orientation='plus', exon_count=13, OMIM=602724, no_hdr0=''),
  5414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5414, CurrentID=0, Status='live', Symbol='SEPT4', Aliases='ARTS, BRADEION, CE5B3, H5, MART, PNUTL2, SEP4, hCDCREL-2, hucep-7', description='septin 4', other_designations='CE5B3 beta|apoptosis-related protein in the TGF-beta signaling pathway|bradeion beta|brain protein H5|cell division control-related protein 2|cerebral protein 7|peanut-like protein 2|septin-M', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58520250, end_position_on_the_genomic_accession=58540818, orientation='minus', exon_count=18, OMIM=603696, no_hdr0=''),
  5420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5420, CurrentID=0, Status='live', Symbol='PODXL', Aliases='Gp200, PC, PCLP, PCLP-1', description='podocalyxin-like', other_designations='GCTM-2 antigen|podocalyxin-like protein 1', map_location='7q32-q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131500262, end_position_on_the_genomic_accession=131556617, orientation='minus', exon_count=9, OMIM=602632, no_hdr0=''),
  5422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5422, CurrentID=0, Status='live', Symbol='POLA1', Aliases='NSX, POLA, p180', description='polymerase (DNA directed), alpha 1, catalytic subunit', other_designations='DNA polymerase alpha 1 catalytic subunit|DNA polymerase alpha catalytic subunit p180|DNA polymerase alpha p180 subunit|polymerase (DNA-directed), alpha (70kD)', map_location='Xp22.1-p21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24693874, end_position_on_the_genomic_accession=24996986, orientation='plus', exon_count=38, OMIM=312040, no_hdr0=''),
  5423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5423, CurrentID=0, Status='live', Symbol='POLB', Aliases='', description='polymerase (DNA directed), beta', other_designations='DNA pol beta|DNA polymerase beta subunit', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42338455, end_position_on_the_genomic_accession=42371813, orientation='plus', exon_count=16, OMIM=174760, no_hdr0=''),
  5424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5424, CurrentID=0, Status='live', Symbol='POLD1', Aliases='CDC2, CRCS10, MDPL, POLD', description='polymerase (DNA directed), delta 1, catalytic subunit', other_designations='CDC2 homolog|DNA polymerase subunit delta p125|polymerase (DNA directed), delta 1, catalytic subunit 125kDa', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50384290, end_position_on_the_genomic_accession=50418018, orientation='plus', exon_count=29, OMIM=174761, no_hdr0=''),
  5425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5425, CurrentID=0, Status='live', Symbol='POLD2', Aliases='', description='polymerase (DNA directed), delta 2, accessory subunit', other_designations='DNA polymerase delta subunit p50|Pol delta B subunit (p50)|polymerase (DNA directed), delta 2, regulatory subunit 50kDa', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44114680, end_position_on_the_genomic_accession=44123668, orientation='minus', exon_count=13, OMIM=600815, no_hdr0=''),
  5426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5426, CurrentID=0, Status='live', Symbol='POLE', Aliases='CRCS12, FILS1, POLE', description='polymerase (DNA directed), epsilon, catalytic subunit', other_designations='DNA polymerase II subunit A', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132623762, end_position_on_the_genomic_accession=132687524, orientation='minus', exon_count=54, OMIM=174762, no_hdr0=''),
  5427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5427, CurrentID=0, Status='live', Symbol='POLE2', Aliases='DPE2', description='polymerase (DNA directed), epsilon 2, accessory subunit', other_designations='DNA polymerase II subunit 2|DNA polymerase epsilon subunit B|polymerase (DNA directed), epsilon 2 (p59 subunit)', map_location='14q21-q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49643552, end_position_on_the_genomic_accession=49688417, orientation='minus', exon_count=20, OMIM=602670, no_hdr0=''),
  5428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5428, CurrentID=0, Status='live', Symbol='POLG', Aliases='MDP1, MIRAS, MTDPS4A, MTDPS4B, PEO1, POLGA, SANDO, SCAE, POLG', description='polymerase (DNA directed), gamma', other_designations='PolG-alpha|mitochondrial DNA polymerase catalytic subunit|mitochondrial DNA polymerase gamma catalytic subunit|truncated mitochondrial DNA polymerase gamma catalytic subunit', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89316305, end_position_on_the_genomic_accession=89334795, orientation='minus', exon_count=23, OMIM=174763, no_hdr0=''),
  5429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5429, CurrentID=0, Status='live', Symbol='POLH', Aliases='RAD30, RAD30A, XP-V, XPV', description='polymerase (DNA directed), eta', other_designations='RAD30 homolog A|xeroderma pigmentosum variant type protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43576141, end_position_on_the_genomic_accession=43620523, orientation='plus', exon_count=11, OMIM=603968, no_hdr0=''),
  5430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5430, CurrentID=0, Status='live', Symbol='POLR2A', Aliases='POLR2, POLRA, RPB1, RPBh1, RPO2, RPOL2, RpIILS, hRPB220, hsRPB1', description='polymerase (RNA) II (DNA directed) polypeptide A, 220kDa', other_designations='DNA-directed RNA polymerase II largest subunit, RNA polymerase II 220 kd subunit|DNA-directed RNA polymerase II subunit A|DNA-directed RNA polymerase III largest subunit|RNA polymerase II subunit B1|RNA-directed RNA polymerase II subunit RPB1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7484379, end_position_on_the_genomic_accession=7514618, orientation='plus', exon_count=29, OMIM=180660, no_hdr0=''),
  5431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5431, CurrentID=0, Status='live', Symbol='POLR2B', Aliases='POL2RB, RPB2, hRPB140', description='polymerase (RNA) II (DNA directed) polypeptide B, 140kDa', other_designations='DNA-directed RNA polymerase II 140 kDa polypeptide|DNA-directed RNA polymerase II subunit B|RNA polymerase II second largest subunit|RNA polymerase II subunit 2|RNA polymerase II subunit B2', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56978640, end_position_on_the_genomic_accession=57031162, orientation='plus', exon_count=26, OMIM=180661, no_hdr0=''),
  5432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5432, CurrentID=0, Status='live', Symbol='POLR2C', Aliases='RPB3, RPB31, hRPB33, hsRPB3', description='polymerase (RNA) II (DNA directed) polypeptide C, 33kDa', other_designations='DNA-directed RNA polymerase II 33 kDa polypeptide|DNA-directed RNA polymerase II subunit C|RNA polymerase II subunit 3|RNA polymerase II subunit B3|RPB33', map_location='16q13-q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57462287, end_position_on_the_genomic_accession=57472009, orientation='plus', exon_count=9, OMIM=180663, no_hdr0=''),
  5433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5433, CurrentID=0, Status='live', Symbol='POLR2D', Aliases='HSRBP4, HSRPB4, RBP4, RPB16', description='polymerase (RNA) II (DNA directed) polypeptide D', other_designations='DNA-directed RNA polymerase II 16 kDa polypeptide|DNA-directed RNA polymerase II subunit D|RNA polymerase II 16 kDa subunit|RNA polymerase II subunit B4|RNA polymerase II subunit hsRBP4', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127846266, end_position_on_the_genomic_accession=127858155, orientation='minus', exon_count=4, OMIM=606017, no_hdr0=''),
  5434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5434, CurrentID=0, Status='live', Symbol='POLR2E', Aliases='RPABC1, RPB5, XAP4, hRPB25, hsRPB5', description='polymerase (RNA) II (DNA directed) polypeptide E, 25kDa', other_designations='DNA directed RNA polymerase II 23 kda polypeptide|DNA-directed RNA polymerase II 23 kDa polypeptide|DNA-directed RNA polymerase II subunit E|DNA-directed RNA polymerase subunit RPABC1|RNA polymerases I, II, and III subunit ABC1|RPB5 homolog', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1086579, end_position_on_the_genomic_accession=1095392, orientation='minus', exon_count=8, OMIM=180664, no_hdr0=''),
  5435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5435, CurrentID=0, Status='live', Symbol='POLR2F', Aliases='HRBP14.4, POLRF, RPABC14.4, RPABC2, RPB14.4, RPB6, RPC15', description='polymerase (RNA) II (DNA directed) polypeptide F', other_designations='DNA-directed RNA polymerase II subunit F|DNA-directed RNA polymerases I, II, and III 14.4 kDa polypeptide|RNA Polymerase II subunit 14.4 kD|RNA polymerases I, II, and III subunit ABC2', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37953663, end_position_on_the_genomic_accession=38041915, orientation='plus', exon_count=9, OMIM=604414, no_hdr0=''),
  5436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5436, CurrentID=0, Status='live', Symbol='POLR2G', Aliases='RPB19, RPB7, hRPB19, hsRPB7', description='polymerase (RNA) II (DNA directed) polypeptide G', other_designations='DNA directed RNA polymerase II 19 kda polypeptide|DNA-directed RNA polymerase II 19 kDa polypeptide|DNA-directed RNA polymerase II subunit G|RNA polymerase II 19 kDa subunit|RNA polymerase II seventh subunit|RNA polymerase II subunit B7', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62761539, end_position_on_the_genomic_accession=62766715, orientation='plus', exon_count=8, OMIM=602013, no_hdr0=''),
  5437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5437, CurrentID=0, Status='live', Symbol='POLR2H', Aliases='RPABC3, RPB17, RPB8', description='polymerase (RNA) II (DNA directed) polypeptide H', other_designations='DNA-directed RNA polymerase II subunit H|DNA-directed RNA polymerases I, II, and III 17.1 kDa polypeptide|RPB8 homolog', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184361714, end_position_on_the_genomic_accession=184368595, orientation='plus', exon_count=6, OMIM=606023, no_hdr0=''),
  5438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5438, CurrentID=0, Status='live', Symbol='POLR2I', Aliases='RPB9, hRPB14.5', description='polymerase (RNA) II (DNA directed) polypeptide I, 14.5kDa', other_designations='DNA-directed RNA polymerase II 14.5 kDa polypeptide|DNA-directed RNA polymerase II subunit I|RNA polymerase II 14.5 kDa subunit|RNA polymerase II subunit B9|RPB14.5', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36113709, end_position_on_the_genomic_accession=36115304, orientation='minus', exon_count=6, OMIM=180662, no_hdr0=''),
  5439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5439, CurrentID=0, Status='live', Symbol='POLR2J', Aliases='POLR2J1, RPB11, RPB11A, RPB11m, hRPB14', description='polymerase (RNA) II (DNA directed) polypeptide J, 13.3kDa', other_designations='DNA-directed RNA polymerase II subunit J-1|RNA polymerase II 13.3 kDa subunit|RNA polymerase II subunit B11-a', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102473100, end_position_on_the_genomic_accession=102478999, orientation='minus', exon_count=4, OMIM=604150, no_hdr0=''),
  5440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5440, CurrentID=0, Status='live', Symbol='POLR2K', Aliases='ABC10-alpha, RPABC4, RPB10alpha, RPB12, RPB7.0, hRPB7.0, hsRPB10a', description='polymerase (RNA) II (DNA directed) polypeptide K, 7.0kDa', other_designations='DNA directed RNA polymerases I, II, and III 7.0 kda polypeptide|DNA-directed RNA polymerase II subunit K|RNA polymerase II 7.0 kDa subunit|RNA polymerases I, II, and III subunit ABC4', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100150611, end_position_on_the_genomic_accession=100154002, orientation='plus', exon_count=4, OMIM=606033, no_hdr0=''),
  5441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5441, CurrentID=0, Status='live', Symbol='POLR2L', Aliases='RBP10, RPABC5, RPB10, RPB10beta, RPB7.6, hRPB7.6', description='polymerase (RNA) II (DNA directed) polypeptide L, 7.6kDa', other_designations='DNA-directed RNA polymerase III subunit L|RNA polymerase II 7.6 kDa subunit|RNA polymerases I, II, and III subunit ABC5|RPB10 homolog', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=839721, end_position_on_the_genomic_accession=842529, orientation='minus', exon_count=2, OMIM=601189, no_hdr0=''),
  5442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5442, CurrentID=0, Status='live', Symbol='POLRMT', Aliases='APOLMT, MTRNAP, MTRPOL, h-mtRPOL', description='polymerase (RNA) mitochondrial (DNA directed)', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=617223, end_position_on_the_genomic_accession=633592, orientation='minus', exon_count=21, OMIM=601778, no_hdr0=''),
  5443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5443, CurrentID=0, Status='live', Symbol='POMC', Aliases='ACTH, CLIP, LPH, MSH, NPP, POC', description='proopiomelanocortin', other_designations='adrenocorticotropic hormone|adrenocorticotropin|alpha-MSH|alpha-melanocyte-stimulating hormone|beta-LPH|beta-MSH|beta-endorphin|beta-melanocyte-stimulating hormone|corticotropin-like intermediary peptide|corticotropin-lipotropin|gamma-LPH|gamma-MSH|lipotropin beta|lipotropin gamma|melanotropin alpha|melanotropin beta|melanotropin gamma|met-enkephalin|opiomelanocortin prepropeptide|pro-ACTH-endorphin|proopiomelanocortin preproprotein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25160853, end_position_on_the_genomic_accession=25168871, orientation='minus', exon_count=4, OMIM=176830, no_hdr0=''),
  5444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5444, CurrentID=0, Status='live', Symbol='PON1', Aliases='ESA, MVCD5, PON', description='paraoxonase 1', other_designations='A-esterase 1|K-45|PON 1|aromatic esterase 1|arylesterase 1|arylesterase B-type|esterase A|paraoxonase B-type|serum aryldiakylphosphatase|serum aryldialkylphosphatase 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95298357, end_position_on_the_genomic_accession=95324572, orientation='minus', exon_count=9, OMIM=168820, no_hdr0=''),
  5445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5445, CurrentID=0, Status='live', Symbol='PON2', Aliases='', description='paraoxonase 2', other_designations='A-esterase 2|PON 2|aromatic esterase 2|arylesterase 2|paraoxonase nirs|serum aryldialkylphosphatase 2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95404862, end_position_on_the_genomic_accession=95435072, orientation='minus', exon_count=9, OMIM=602447, no_hdr0=''),
  5446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5446, CurrentID=0, Status='live', Symbol='PON3', Aliases='', description='paraoxonase 3', other_designations='arylesterase 3|paraoxanase-3', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95359872, end_position_on_the_genomic_accession=95396375, orientation='minus', exon_count=9, OMIM=602720, no_hdr0=''),
  5447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5447, CurrentID=0, Status='live', Symbol='POR', Aliases='CPR, CYPOR, P450R', description='P450 (cytochrome) oxidoreductase', other_designations='NADPH-dependent cytochrome P450 reductase', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75915102, end_position_on_the_genomic_accession=75986855, orientation='plus', exon_count=16, OMIM=124015, no_hdr0=''),
  5449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5449, CurrentID=0, Status='live', Symbol='POU1F1', Aliases='CPHD1, GHF-1, PIT1a, Pit-1, POU1F1', description='POU class 1 homeobox 1', other_designations='POU domain, class 1, transcription factor 1|growth hormone factor 1', map_location='3p11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87259633, end_position_on_the_genomic_accession=87276587, orientation='minus', exon_count=6, OMIM=173110, no_hdr0=''),
  5450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5450, CurrentID=0, Status='live', Symbol='POU2AF1', Aliases='BOB1, OBF-1, OBF1, OCAB', description='POU class 2 associating factor 1', other_designations='B-cell-specific coactivator OBF-1|BOB-1|OCA-B|OCT-binding factor 1|POU domain, class 2, associating factor 1', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111352251, end_position_on_the_genomic_accession=111379432, orientation='minus', exon_count=9, OMIM=601206, no_hdr0=''),
  5451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5451, CurrentID=0, Status='live', Symbol='POU2F1', Aliases='OCT1, OTF1, oct-1B', description='POU class 2 homeobox 1', other_designations='NF-A1|OTF-1|Octamer-binding transcription factor-1|oct-1|octamer-binding protein 1|octamer-binding transcription factor 1', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167220829, end_position_on_the_genomic_accession=167427345, orientation='plus', exon_count=20, OMIM=164175, no_hdr0=''),
  5452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5452, CurrentID=0, Status='live', Symbol='POU2F2', Aliases='OCT2, OTF2, Oct-2', description='POU class 2 homeobox 2', other_designations='OTF-2|homeobox protein|lymphoid-restricted immunoglobulin octamer-binding protein NF-A2|octamer-binding protein 2|octamer-binding transcription factor 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42086110, end_position_on_the_genomic_accession=42132473, orientation='minus', exon_count=15, OMIM=164176, no_hdr0=''),
  5453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5453, CurrentID=0, Status='live', Symbol='POU3F1', Aliases='OCT6, OTF6, SCIP', description='POU class 3 homeobox 1', other_designations='OTF-6|POU domain transcription factor SCIP|octamer-binding protein 6|octamer-binding transcription factor 6', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38043851, end_position_on_the_genomic_accession=38046778, orientation='minus', exon_count=1, OMIM=602479, no_hdr0=''),
  5454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5454, CurrentID=0, Status='live', Symbol='POU3F2', Aliases='BRN2, N-Oct3, OCT7, OTF-7, OTF7, POUF3, brn-2, oct-7', description='POU class 3 homeobox 2', other_designations='brain-2|brain-specific homeobox/POU domain protein 2|nervous system-specific octamer-binding transcription factor N-Oct-3|octamer-binding protein 7|octamer-binding transcription factor 7', map_location='6q16', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=98834704, end_position_on_the_genomic_accession=98838790, orientation='plus', exon_count=1, OMIM=600494, no_hdr0=''),
  5455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5455, CurrentID=0, Status='live', Symbol='POU3F3', Aliases='BRN1, OTF8, brain-1, oct-8', description='POU class 3 homeobox 3', other_designations='brain-specific homeobox/POU domain protein 1|octamer-binding protein 8|octamer-binding transcription factor 8', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104855511, end_position_on_the_genomic_accession=104857013, orientation='plus', exon_count=1, OMIM=602480, no_hdr0=''),
  5456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5456, CurrentID=0, Status='live', Symbol='POU3F4', Aliases='BRAIN-4, BRN-4, BRN4, DFN3, DFNX2, OCT-9, OTF-9, OTF9', description='POU class 3 homeobox 4', other_designations='brain-specific homeobox/POU domain protein 4|octamer-binding transcription factor 9', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=83508261, end_position_on_the_genomic_accession=83509767, orientation='plus', exon_count=1, OMIM=300039, no_hdr0=''),
  5457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5457, CurrentID=0, Status='live', Symbol='POU4F1', Aliases='BRN3A, Oct-T1, RDC-1, brn-3A', description='POU class 4 homeobox 1', other_designations='brain-3A|brain-specific homeobox/POU domain protein 3A|homeobox/POU domain protein RDC-1', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78599092, end_position_on_the_genomic_accession=78603560, orientation='minus', exon_count=3, OMIM=601632, no_hdr0=''),
  5458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5458, CurrentID=0, Status='live', Symbol='POU4F2', Aliases='BRN3.2, BRN3B, Brn-3b', description='POU class 4 homeobox 2', other_designations='Brn3b POU domain transcription factor|POU domain class 4 transcription factor 2|POU domain protein|brain-3B|brain-specific homeobox/POU domain protein 3B', map_location='4q31.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146638893, end_position_on_the_genomic_accession=146642471, orientation='plus', exon_count=2, OMIM=113725, no_hdr0=''),
  5459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5459, CurrentID=0, Status='live', Symbol='POU4F3', Aliases='BRN3C, DFNA15', description='POU class 4 homeobox 3', other_designations='brain-3C|brain-specific homeobox/POU domain protein 3C|brn-3C', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146339024, end_position_on_the_genomic_accession=146340520, orientation='plus', exon_count=2, OMIM=602460, no_hdr0=''),
  5460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5460, CurrentID=0, Status='live', Symbol='POU5F1', Aliases='OCT3, OCT4, OTF-3, OTF3, OTF4, Oct-3, Oct-4', description='POU class 5 homeobox 1', other_designations='POU class 5 homeobox 1 transcript variant OCT4B1|POU class 5 homeobox 1 transcript variant OCT4B2|POU class 5 homeobox 1 transcript variant OCT4B3|POU class 5 homeobox 1 transcript variant OCT4B4|POU class 5 homeobox 1 transcript variant OCT4B5|POU class 5 homeobox 1 transcript variant OCT4B6|POU domain class 5 transcription factor 1|POU domain transcription factor OCT4|POU-type homeodomain-containing DNA-binding protein|octamer-binding protein 3|octamer-binding protein 4|octamer-binding transcription factor 3', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31164337, end_position_on_the_genomic_accession=31170693, orientation='minus', exon_count=6, OMIM=164177, no_hdr0=''),
  100270990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270990, CurrentID=0, Status='live', Symbol='RPL31P29', Aliases='RPL31_9_773', description='ribosomal protein L31 pseudogene 29', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154605625, end_position_on_the_genomic_accession=154605928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5462, CurrentID=0, Status='live', Symbol='POU5F1B', Aliases='OCT4-PG1, OCT4PG1, OTF3C, OTF3P1, POU5F1P1, POU5F1P4, POU5FLC20, POU5FLC8', description='POU class 5 homeobox 1B', other_designations='POU 5 domain protein|POU class 5 homeobox 1 pseudogene 1|POU domain transcription factor OCT4-pg1|POU domain transcription factor Oct-4|POU domain, class 5, transcription factor 1 pseudogene 1|octamer binding protein 3_like sequence', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127415612, end_position_on_the_genomic_accession=127417210, orientation='plus', exon_count=1, OMIM=615739, no_hdr0=''),
  5463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5463, CurrentID=0, Status='live', Symbol='POU6F1', Aliases='BRN5, MPOU, TCFB1', description='POU class 6 homeobox 1', other_designations='brain-5|brain-specific homeobox/POU domain protein 5|brn-5|mPOU homeobox protein', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51186936, end_position_on_the_genomic_accession=51217693, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  5464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5464, CurrentID=0, Status='live', Symbol='PPA1', Aliases='HEL-S-66p, IOPPP, PP, PP1, SID6-8061', description='pyrophosphatase (inorganic) 1', other_designations='PPase|cytosolic inorganic pyrophosphatase|diphosphate phosphohydrolase|epididymis secretory sperm binding protein Li 66p|inorganic diphosphatase|inorganic pyrophosphatase 1|pyrophosphatase 1|pyrophosphate phospho-hydrolase', map_location='10q11.1-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70202830, end_position_on_the_genomic_accession=70233434, orientation='minus', exon_count=11, OMIM=179030, no_hdr0=''),
  5465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5465, CurrentID=0, Status='live', Symbol='PPARA', Aliases='NR1C1, PPAR, PPARalpha, hPPAR', description='peroxisome proliferator-activated receptor alpha', other_designations='PPAR-alpha|nuclear receptor subfamily 1 group C member 1|peroxisome proliferative activated receptor, alpha|peroxisome proliferator-activated nuclear receptor alpha variant 3', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46150547, end_position_on_the_genomic_accession=46243756, orientation='plus', exon_count=13, OMIM=170998, no_hdr0=''),
  5467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5467, CurrentID=0, Status='live', Symbol='PPARD', Aliases='FAAR, NR1C2, NUC1, NUCI, NUCII, PPARB', description='peroxisome proliferator-activated receptor delta', other_designations='PPAR-beta|PPAR-delta|nuclear hormone receptor 1|nuclear receptor subfamily 1 group C member 2|peroxisome proliferator-activated nuclear receptor beta/delta variant 2|peroxisome proliferator-activated receptor beta', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35342558, end_position_on_the_genomic_accession=35428191, orientation='plus', exon_count=11, OMIM=600409, no_hdr0=''),
  5468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5468, CurrentID=0, Status='live', Symbol='PPARG', Aliases='CIMT1, GLM1, NR1C31, PPARG2, PPARgamma, PPARG', description='peroxisome proliferator-activated receptor gamma', other_designations='PPAR-gamma|nuclear receptor subfamily 1 group C member 3|peroxisome proliferator-activated nuclear receptor gamma variant 1', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12287850, end_position_on_the_genomic_accession=12471013, orientation='plus', exon_count=11, OMIM=601487, no_hdr0=''),
  5469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5469, CurrentID=0, Status='live', Symbol='MED1', Aliases='CRSP1, CRSP200, DRIP205, DRIP230, PBP, PPARBP, PPARGBP, RB18A, TRAP220, TRIP2', description='mediator complex subunit 1', other_designations='ARC205|PPAR binding protein|PPAR-binding protein|PPARG binding protein|TR-interacting protein 2|TRIP-2|activator-recruited cofactor 205 kDa component|p53 regulatory protein RB18A|peroxisome proliferator-activated receptor-binding protein|thyroid hormone receptor-associated protein complex 220 kDa component|thyroid hormone receptor-associated protein complex component TRAP220|thyroid receptor interacting protein 2|thyroid receptor-interacting protein 2|vitamin D receptor-interacting protein 230 kD|vitamin D receptor-interacting protein complex component DRIP205', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39404285, end_position_on_the_genomic_accession=39451280, orientation='minus', exon_count=17, OMIM=604311, no_hdr0=''),
  5470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5470, CurrentID=0, Status='live', Symbol='PPEF2', Aliases='PPP7CB', description='protein phosphatase, EF-hand calcium binding domain 2', other_designations='protein phosphatase 7, catalytic subunit, beta isozyme|protein phosphatase with EF hands 2|protein phosphatase, EF hand calcium-binding domain 2', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75859872, end_position_on_the_genomic_accession=75902528, orientation='minus', exon_count=17, OMIM=602256, no_hdr0=''),
  5471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5471, CurrentID=0, Status='live', Symbol='PPAT', Aliases='ATASE, GPAT, PRAT', description='phosphoribosyl pyrophosphate amidotransferase', other_designations='glutamine PRPP amidotransferase|glutamine phosphoribosylpyrophosphatate amidotransferase|glutamine phosphoribosylpyrophosphate amidotransferase', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56393362, end_position_on_the_genomic_accession=56435636, orientation='minus', exon_count=11, OMIM=172450, no_hdr0=''),
  5473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5473, CurrentID=0, Status='live', Symbol='PPBP', Aliases='B-TG1, Beta-TG, CTAP-III, CTAP3, CTAPIII, CXCL7, LA-PF4, LDGF, MDGF, NAP-2, PBP, SCYB7, TC1, TC2, TGB, TGB1, THBGB, THBGB1', description='pro-platelet basic protein', other_designations='C-X-C motif chemokine 7|CXC chemokine ligand 7|beta-thromboglobulin|connective tissue-activating peptide III|leukocyte-derived growth factor|low-affinity platelet factor IV|macrophage-derived growth factor|neutrophil-activating peptide 2|neutrophil-activating peptide-2|pro-platelet basic protein (chemokine (C-X-C motif) ligand 7)|small inducible cytokine B7|small inducible cytokine subfamily B, member 7|small-inducible cytokine B7|thrombocidin 1|thrombocidin 2|thromboglobulin, beta-1', map_location='4q12-q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73986439, end_position_on_the_genomic_accession=73988190, orientation='minus', exon_count=3, OMIM=121010, no_hdr0=''),
  5475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5475, CurrentID=0, Status='live', Symbol='PPEF1', Aliases='PP7, PPEF, PPP7C, PPP7CA', description='protein phosphatase, EF-hand calcium binding domain 1', other_designations='protein phosphatase 7, catalytic subunit, alpha isozyme|protein phosphatase with EF calcium-binding domain|protein phosphatase, serine/threonine type, with EF-hands|serine/threonine protein phosphatase 7|serine/threonine-protein phosphatase 7', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18675909, end_position_on_the_genomic_accession=18827921, orientation='plus', exon_count=22, OMIM=300109, no_hdr0=''),
  5476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5476, CurrentID=0, Status='live', Symbol='CTSA', Aliases='GLB2, GSL, NGBE, PPCA, PPGB', description='cathepsin A', other_designations='beta-galactosidase 2|beta-galactosidase protective protein|carboxypeptidase C|carboxypeptidase Y-like kininase|carboxypeptidase-L|deamidase|lysosomal carboxypeptidase A|protective protein cathepsin A|urinary kininase', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45890952, end_position_on_the_genomic_accession=45898820, orientation='plus', exon_count=15, OMIM=613111, no_hdr0=''),
  5478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5478, CurrentID=0, Status='live', Symbol='PPIA', Aliases='CYPA, CYPH, HEL-S-69p', description='peptidylprolyl isomerase A (cyclophilin A)', other_designations='PPIase A|T cell cyclophilin|cyclosporin A-binding protein|epididymis secretory sperm binding protein Li 69p|rotamase A', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44796636, end_position_on_the_genomic_accession=44803123, orientation='plus', exon_count=6, OMIM=123840, no_hdr0=''),
  5479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5479, CurrentID=0, Status='live', Symbol='PPIB', Aliases='CYP-S1, CYPB, HEL-S-39, OI9, SCYLP', description='peptidylprolyl isomerase B (cyclophilin B)', other_designations='PPIase B|S-cyclophilin|cyclophilin-like protein|epididymis secretory protein Li 39|rotamase B', map_location='15q21-q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64155815, end_position_on_the_genomic_accession=64163155, orientation='minus', exon_count=5, OMIM=123841, no_hdr0=''),
  5480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5480, CurrentID=0, Status='live', Symbol='PPIC', Aliases='CYPC', description='peptidylprolyl isomerase C (cyclophilin C)', other_designations='PPIase C|cyclophilin C|parvulin|rotamase C', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123023383, end_position_on_the_genomic_accession=123036730, orientation='minus', exon_count=5, OMIM=123842, no_hdr0=''),
  5481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5481, CurrentID=0, Status='live', Symbol='PPID', Aliases='CYP-40, CYPD', description='peptidylprolyl isomerase D', other_designations='40 kDa peptidyl-prolyl cis-trans isomerase D|PPIase D|cyclophilin 40|cyclophilin D|cyclophilin-40|cyclophilin-related protein|rotamase D', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158709127, end_position_on_the_genomic_accession=158723400, orientation='minus', exon_count=10, OMIM=601753, no_hdr0=''),
  5483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5483, CurrentID=0, Status='live', Symbol='PPIAP11', Aliases='PPIP1', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 11', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82009390, end_position_on_the_genomic_accession=82010140, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5485, CurrentID=0, Status='live', Symbol='PPIAP13', Aliases='PPIP3', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 13', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 3', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75089032, end_position_on_the_genomic_accession=75089772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5486, CurrentID=0, Status='live', Symbol='PPIAP14', Aliases='PPIP4', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 14', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 4', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5002552, end_position_on_the_genomic_accession=5003296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5487, CurrentID=0, Status='live', Symbol='PPIAP15', Aliases='PPIP5', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 15', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 5', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109471120, end_position_on_the_genomic_accession=109471857, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5488, CurrentID=0, Status='live', Symbol='PPIAP16', Aliases='PPIP6', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 16', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 6', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57941776, end_position_on_the_genomic_accession=57942514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5489, CurrentID=0, Status='live', Symbol='PPIAP7', Aliases='PPIP7', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 7', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 7', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101270665, end_position_on_the_genomic_accession=101271410, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5491, CurrentID=0, Status='live', Symbol='PPIAP9', Aliases='PPIP9', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 9', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 9', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31518977, end_position_on_the_genomic_accession=31520330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5493, CurrentID=0, Status='live', Symbol='PPL', Aliases='', description='periplakin', other_designations='190 kDa paraneoplastic pemphigus antigen|195 kDa cornified envelope precursor protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4882507, end_position_on_the_genomic_accession=4937135, orientation='minus', exon_count=23, OMIM=602871, no_hdr0=''),
  5494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5494, CurrentID=0, Status='live', Symbol='PPM1A', Aliases='PP2C-ALPHA, PP2CA, PP2Calpha', description='protein phosphatase, Mg2+/Mn2+ dependent 1A', other_designations='protein phosphatase 1A (formerly 2C), magnesium-dependent, alpha isoform|protein phosphatase, Mg2+/Mn2+ dependent, 1A', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60245752, end_position_on_the_genomic_accession=60299087, orientation='plus', exon_count=17, OMIM=606108, no_hdr0=''),
  5495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5495, CurrentID=0, Status='live', Symbol='PPM1B', Aliases='PP2C-beta, PP2C-beta-X, PP2CB, PP2CBETA, PPC2BETAX', description='protein phosphatase, Mg2+/Mn2+ dependent 1B', other_designations='Ser/Thr protein phosphatase type 2C beta 2 isoform|protein phosphatase 1B (formerly 2C), magnesium-dependent, beta isoform|protein phosphatase 2C-like protein|protein phosphatase, Mg2+/Mn2+ dependent, 1B', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44168803, end_position_on_the_genomic_accession=44244425, orientation='plus', exon_count=8, OMIM=603770, no_hdr0=''),
  5496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5496, CurrentID=0, Status='live', Symbol='PPM1G', Aliases='PP2CG, PP2CGAMMA, PPP2CG', description='protein phosphatase, Mg2+/Mn2+ dependent 1G', other_designations='PP2C, gamma|PP2C-gamma|protein phosphatase 1C|protein phosphatase 1G (formerly 2C), magnesium-dependent, gamma isoform|protein phosphatase 2, catalytic subunit, gamma isoform|protein phosphatase 2C gamma isoform|protein phosphatase 2C isoform gamma|protein phosphatase magnesium-dependent 1 gamma|protein phosphatase, Mg2+/Mn2+ dependent, 1G', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27381199, end_position_on_the_genomic_accession=27409683, orientation='minus', exon_count=10, OMIM=605119, no_hdr0=''),
  100270996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270996, CurrentID=0, Status='live', Symbol='RPL34P12', Aliases='RPL34_2_514', description='ribosomal protein L34 pseudogene 12', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122815782, end_position_on_the_genomic_accession=122816117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5498, CurrentID=0, Status='live', Symbol='PPOX', Aliases='PPO, V290M, VP', description='protoporphyrinogen oxidase', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161165728, end_position_on_the_genomic_accession=161178277, orientation='plus', exon_count=19, OMIM=600923, no_hdr0=''),
  5499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5499, CurrentID=0, Status='live', Symbol='PPP1CA', Aliases='PP-1A, PP1A, PP1alpha, PPP1A', description='protein phosphatase 1, catalytic subunit, alpha isozyme', other_designations='protein phosphatase 1, catalytic subunit, alpha isoform|serine/threonine protein phosphatase PP1-alpha 1 catalytic subunit', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67398181, end_position_on_the_genomic_accession=67401905, orientation='minus', exon_count=7, OMIM=176875, no_hdr0=''),
  5500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5500, CurrentID=0, Status='live', Symbol='PPP1CB', Aliases='HEL-S-80p, PP-1B, PP1B, PP1beta, PPP1CD', description='protein phosphatase 1, catalytic subunit, beta isozyme', other_designations='epididymis secretory sperm binding protein Li 80p|protein phosphatase 1, catalytic subunit, beta isoform|protein phosphatase 1, catalytic subunit, delta isoform|protein phosphatase 1-beta|protein phosphatase 1-delta|serine/threonine protein phosphatase PP1-beta catalytic subunit', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28751748, end_position_on_the_genomic_accession=28802940, orientation='plus', exon_count=9, OMIM=600590, no_hdr0=''),
  5501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5501, CurrentID=0, Status='live', Symbol='PPP1CC', Aliases='PP-1G, PP1C, PPP1G', description='protein phosphatase 1, catalytic subunit, gamma isozyme', other_designations='protein phosphatase 1, catalytic subunit, gamma isoform|protein phosphatase 1C catalytic subunit|serine/threonine phosphatase 1 gamma', map_location='12q24.1-q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110708970, end_position_on_the_genomic_accession=110742978, orientation='minus', exon_count=10, OMIM=176914, no_hdr0=''),
  5502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5502, CurrentID=0, Status='live', Symbol='PPP1R1A', Aliases='I1, IPP1', description='protein phosphatase 1 regulatory inhibitor subunit 1A', other_designations='I-1|IPP-1|inhibitor-1|protein phosphatase 1, regulatory (inhibitor) subunit 1A|protein phosphatase inhibitor 1|protein phosphatase inhibitor-1', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54579240, end_position_on_the_genomic_accession=54588677, orientation='minus', exon_count=8, OMIM=613246, no_hdr0=''),
  100270997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270997, CurrentID=0, Status='live', Symbol='RPL34P13', Aliases='RPL34_3_616', description='ribosomal protein L34 pseudogene 13', other_designations='', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134804777, end_position_on_the_genomic_accession=134805132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5504, CurrentID=0, Status='live', Symbol='PPP1R2', Aliases='IPP-2, IPP2', description='protein phosphatase 1 regulatory inhibitor subunit 2', other_designations='phosphoprotein phosphatase|protein phosphatase 1, regulatory (inhibitor) subunit 2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195514422, end_position_on_the_genomic_accession=195543401, orientation='minus', exon_count=7, OMIM=601792, no_hdr0=''),
  5506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5506, CurrentID=0, Status='live', Symbol='PPP1R3A', Aliases='GM, PP1G, PPP1R3', description='protein phosphatase 1 regulatory subunit 3A', other_designations='RG1|glycogen and sarcoplasmic reticulum binding subunit, skeletal muscle|glycogen-associated regulatory subunit of protein phosphatase-1|protein phosphatase 1 glycogen- associated regulatory subunit|protein phosphatase 1 glycogen-associated regulatory subunit|protein phosphatase 1 regulatory subunit GM|protein phosphatase 1, regulatory (inhibitor) subunit 3A|protein phosphatase 1, regulatory subunit 3A|protein phosphatase type-1 glycogen targeting subunit|serine /threonine specific protein phosphatase|type-1 protein phosphatase skeletal muscle glycogen targeting subunit', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=113876827, end_position_on_the_genomic_accession=113919027, orientation='minus', exon_count=4, OMIM=600917, no_hdr0=''),
  5507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5507, CurrentID=0, Status='live', Symbol='PPP1R3C', Aliases='PPP1R5', description='protein phosphatase 1 regulatory subunit 3C', other_designations='PP1 subunit R5|PTG|Phosphatase 1, regulatory inhibitor subunit 5|protein phosphatase 1 regulatory subunit 5|protein phosphatase 1, regulatory (inhibitor) subunit 3C|protein phosphatase 1, regulatory subunit 3C|protein targeting to glycogen', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91628440, end_position_on_the_genomic_accession=91633101, orientation='minus', exon_count=2, OMIM=602999, no_hdr0=''),
  5509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5509, CurrentID=0, Status='live', Symbol='PPP1R3D', Aliases='PPP1R6', description='protein phosphatase 1 regulatory subunit 3D', other_designations='PP1 subunit R6|protein phosphatase 1 regulatory subunit 6|protein phosphatase 1, regulatory (inhibitor) subunit 3D|protein phosphatase 1, regulatory subunit 3D|protein phosphatase 1, regulatory subunit 6, spinophilin|protein phosphatase 1-binding subunit R6', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59936832, end_position_on_the_genomic_accession=59940297, orientation='minus', exon_count=1, OMIM=603326, no_hdr0=''),
  5510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5510, CurrentID=0, Status='live', Symbol='PPP1R7', Aliases='SDS22', description='protein phosphatase 1 regulatory subunit 7', other_designations='protein phosphatase 1 regulatory subunit 22|protein phosphatase 1, regulatory (inhibitor) subunit 7|protein phosphatase 1, regulatory subunit 7', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241149573, end_position_on_the_genomic_accession=241183650, orientation='plus', exon_count=12, OMIM=602877, no_hdr0=''),
  5511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5511, CurrentID=0, Status='live', Symbol='PPP1R8', Aliases='ARD-1, ARD1, NIPP-1, NIPP1, PRO2047', description='protein phosphatase 1 regulatory subunit 8', other_designations='RNase E|activator of RNA decay|nuclear inhibitor of protein phosphatase-1 alpha|nuclear subunit of PP-1|protein phosphatase 1, regulatory (inhibitor) subunit 8|protein phosphatase 1, regulatory subunit 8', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27830741, end_position_on_the_genomic_accession=27851672, orientation='plus', exon_count=7, OMIM=602636, no_hdr0=''),
  5512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5512, CurrentID=0, Status='live', Symbol='PPP1R8P1', Aliases='PPP1R8P', description='protein phosphatase 1 regulatory subunit 8 pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 8 pseudogene 1|protein phosphatase 1, regulatory subunit 8 pseudogene 1', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48324974, end_position_on_the_genomic_accession=48326146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5514, CurrentID=0, Status='live', Symbol='PPP1R10', Aliases='CAT53, FB19, PNUTS, PP1R10, R111, p99', description='protein phosphatase 1 regulatory subunit 10', other_designations='HLA-C associated transcript 53|MHC class I region proline-rich protein CAT53|PP1-binding protein of 114 kDa|phosphatase 1 nuclear targeting subunit|protein phosphatase 1, regulatory (inhibitor) subunit 10|protein phosphatase 1, regulatory subunit 10', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30600400, end_position_on_the_genomic_accession=30618607, orientation='minus', exon_count=22, OMIM=603771, no_hdr0=''),
  5515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5515, CurrentID=0, Status='live', Symbol='PPP2CA', Aliases='PP2Ac, PP2CA, PP2Calpha, RP-C', description='protein phosphatase 2, catalytic subunit, alpha isozyme', other_designations='PP2A-alpha|protein phosphatase 2 (formerly 2A), catalytic subunit, alpha isoform|protein phosphatase 2A catalytic subunit, alpha isoform|replication protein C|serine/threonine protein phosphatase 2A, catalytic subunit, alpha isoform', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134196457, end_position_on_the_genomic_accession=134226259, orientation='minus', exon_count=7, OMIM=176915, no_hdr0=''),
  5516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5516, CurrentID=0, Status='live', Symbol='PPP2CB', Aliases='PP2Abeta, PP2CB', description='protein phosphatase 2, catalytic subunit, beta isozyme', other_designations='PP2A-beta|protein phosphatase 2 (formerly 2A), catalytic subunit, beta isoform|protein phosphatase 2, catalytic subunit, beta isoform|protein phosphatase 2A catalytic subunit, beta isoform|protein phosphatase type 2A catalytic subunit|serine/threonine protein phosphatase 2A, catalytic subunit, beta isoform', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30785610, end_position_on_the_genomic_accession=30812836, orientation='minus', exon_count=7, OMIM=176916, no_hdr0=''),
  5517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5517, CurrentID=0, Status='live', Symbol='PPP2CBP1', Aliases='PPP2CBP', description='protein phosphatase 2, catalytic subunit, beta isozyme pseudogene 1', other_designations='protein phosphatase 2 (formerly 2A), catalytic subunit, beta isoform pseudogene|protein phosphatase 2, catalytic subunit, beta isoform pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31805872, end_position_on_the_genomic_accession=31807321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5518, CurrentID=0, Status='live', Symbol='PPP2R1A', Aliases='MRD36, PP2A-Aalpha, PP2AAALPHA, PR65A', description='protein phosphatase 2 regulatory subunit A, alpha', other_designations='PP2A subunit A isoform PR65-alpha|PP2A subunit A isoform R1-alpha|medium tumor antigen-associated 61 KDA protein|protein phosphatase 2 (formerly 2A), regulatory subunit A (PR 65), alpha isoform|protein phosphatase 2, regulatory subunit A, alpha|serine/threonine protein phosphatase 2A, 65 kDa regulatory subunit A, alpha isoform', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52189802, end_position_on_the_genomic_accession=52226425, orientation='plus', exon_count=15, OMIM=605983, no_hdr0=''),
  5519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5519, CurrentID=0, Status='live', Symbol='PPP2R1B', Aliases='PP2A-Abeta, PR65B', description='protein phosphatase 2 regulatory subunit A, beta', other_designations='PP2A, subunit A, PR65-beta isoform|PP2A, subunit A, R1-beta isoform|protein phosphatase 2 (formerly 2A), regulatory subunit A, beta isoform|protein phosphatase 2, regulatory subunit A, beta|protein phosphatase 2, structural/regulatory subunit A, beta', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111695074, end_position_on_the_genomic_accession=111766445, orientation='minus', exon_count=22, OMIM=603113, no_hdr0=''),
  5520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5520, CurrentID=0, Status='live', Symbol='PPP2R2A', Aliases='B55A, B55ALPHA, PR52A, PR55A', description='protein phosphatase 2 regulatory subunit B, alpha', other_designations='protein phosphatase 2 (formerly 2A), regulatory subunit B (PR52), alpha isoform|protein phosphatase 2, regulatory subunit B, alpha|serine/threonine protein phosphatase 2A, 55 KDA regulatory subunit B, alpha isoform', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26291491, end_position_on_the_genomic_accession=26372680, orientation='plus', exon_count=15, OMIM=604941, no_hdr0=''),
  5521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5521, CurrentID=0, Status='live', Symbol='PPP2R2B', Aliases='B55BETA, PP2AB55BETA, PP2ABBETA, PP2APR55B, PP2APR55BETA, PR2AB55BETA, PR2ABBETA, PR2APR55BETA, PR52B, PR55-BETA, PR55BETA, SCA12', description='protein phosphatase 2 regulatory subunit B, beta', other_designations='PP2A subunit B isoform B55-beta|PP2A subunit B isoform PR55-beta|PP2A subunit B isoform R2-beta|PP2A, subunit B, B-beta isoform|protein phosphatase 2 (formerly 2A), regulatory subunit B (PR 52), beta isoform|protein phosphatase 2 (formerly 2A), regulatory subunit B, beta isoform|protein phosphatase 2, regulatory subunit B, beta|serine/threonine protein phosphatase 2A, 55 kDa regulatory subunit B, beta isoform|serine/threonine protein phosphatase 2A, neuronal isoform', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146589504, end_position_on_the_genomic_accession=147081520, orientation='minus', exon_count=14, OMIM=604325, no_hdr0=''),
  5522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5522, CurrentID=0, Status='live', Symbol='PPP2R2C', Aliases='B55-GAMMA, IMYPNO, IMYPNO1, PR52, PR55G', description='protein phosphatase 2 regulatory subunit B, gamma', other_designations='PP2A, subunit B, B-gamma isoform|PP2A, subunit B, B55-gamma isoform|PP2A, subunit B, PR55-gamma isoform|PP2A, subunit B, R2-gamma isoform|gamma isoform of regulatory subunit B55, protein phosphatase 2|phosphoprotein phosphatase 2A BR gamma regulatory chain|protein phosphatase 2 (formerly 2A), regulatory subunit B (PR 52), gamma isoform|protein phosphatase 2A1 B gamma subunit|serine/threonine-protein phosphatase 2A 55 kDa regulatory subunit B gamma isoform', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6320578, end_position_on_the_genomic_accession=6563600, orientation='minus', exon_count=16, OMIM=605997, no_hdr0=''),
  5523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5523, CurrentID=0, Status='live', Symbol='PPP2R3A', Aliases='PPP2R3, PR130, PR72', description="protein phosphatase 2 regulatory subunit B'', alpha", other_designations="PP2A subunit B isoforms B72/B130|PP2A, subunit B, R3 isoform|protein phosphatase 2 (formerly 2A), regulatory subunit B'' (PR 72), alpha isoform and (PR 130), beta isoform|protein phosphatase 2, regulatory subunit B'', alpha|serine/threonine-protein phosphatase 2A 72/130 kDa regulatory subunit B", map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135965673, end_position_on_the_genomic_accession=136147910, orientation='plus', exon_count=18, OMIM=604944, no_hdr0=''),
  5524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5524, CurrentID=0, Status='live', Symbol='PPP2R4', Aliases='PP2A, PR53, PTPA', description='protein phosphatase 2A regulatory subunit 4', other_designations="PP2A phosphatase activator|PP2A subunit B' isoform PR53|phosphotyrosyl phosphatase activator|protein phosphatase 2A activator, regulatory subunit 4|protein phosphatase 2A, regulatory subunit B' (PR 53)|serine/threonine-protein phosphatase 2A regulatory subunit B'", map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129110945, end_position_on_the_genomic_accession=129148946, orientation='plus', exon_count=13, OMIM=600756, no_hdr0=''),
  5525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5525, CurrentID=0, Status='live', Symbol='PPP2R5A', Aliases='B56A, PR61A', description="protein phosphatase 2 regulatory subunit B', alpha", other_designations="PP2A B subunit isoform B'-alpha|PP2A B subunit isoform B56-alpha|PP2A B subunit isoform PR61-alpha|PP2A B subunit isoform R5-alpha|PP2A, B subunit, B' alpha isoform|PP2A, B subunit, B56 alpha isoform|PP2A, B subunit, PR61 alpha isoform|PP2A, B subunit, R5 alpha isoform|PR61alpha|protein phosphatase 2, regulatory subunit B (B56), alpha isoform|protein phosphatase 2, regulatory subunit B', alpha isoform|serine/threonine protein phosphatase 2A, 56 kDa regulatory subunit, alpha isoform", map_location='1q32.2-q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212285537, end_position_on_the_genomic_accession=212361863, orientation='plus', exon_count=14, OMIM=601643, no_hdr0=''),
  5526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5526, CurrentID=0, Status='live', Symbol='PPP2R5B', Aliases='B56B, PR61B', description="protein phosphatase 2 regulatory subunit B', beta", other_designations="PP2A B subunit isoform B'-beta|PP2A B subunit isoform B56-beta|PP2A B subunit isoform PR61-beta|PP2A B subunit isoform R5-beta|protein phosphatase 2, regulatory subunit B', beta isoform|serine/threonine protein phosphatase 2A, 56 kDa regulatory subunit, beta isoform", map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64917553, end_position_on_the_genomic_accession=64934478, orientation='plus', exon_count=17, OMIM=601644, no_hdr0=''),
  5527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5527, CurrentID=0, Status='live', Symbol='PPP2R5C', Aliases='B56G, PR61G', description="protein phosphatase 2 regulatory subunit B', gamma", other_designations="B' alpha regulatory subunit|PP2A B subunit isoform B'-gamma|PP2A B subunit isoform B56-gamma|PP2A B subunit isoform PR61-gamma|PP2A B subunit isoform R5-gamma|protein phosphatase 2, regulatory subunit B (B56), gamma isoform|protein phosphatase 2, regulatory subunit B', gamma isoform|renal carcinoma antigen NY-REN-29", map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101760573, end_position_on_the_genomic_accession=101927992, orientation='plus', exon_count=25, OMIM=601645, no_hdr0=''),
  5528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5528, CurrentID=0, Status='live', Symbol='PPP2R5D', Aliases='B56D, MRD35', description="protein phosphatase 2 regulatory subunit B', delta", other_designations="PP2A B subunit isoform B'-delta|PP2A B subunit isoform B56-delta|PP2A B subunit isoform PR61-delta|PP2A B subunit isoform R5-delta|PP2A, B subunit, B' delta isoform|PP2A, B subunit, B56 delta isoform|PP2A, B subunit, PR61 delta isoform|PP2A, B subunit, R5 delta isoform|Serine/threonine protein phosphatase 2A, 56 kDa regulatory subunit, delta isoform|protein phosphatase 2, regulatory subunit B (B56), delta isoform|protein phosphatase 2, regulatory subunit B', delta isoform", map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42984499, end_position_on_the_genomic_accession=43012345, orientation='plus', exon_count=16, OMIM=601646, no_hdr0=''),
  5529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5529, CurrentID=0, Status='live', Symbol='PPP2R5E', Aliases='', description="protein phosphatase 2 regulatory subunit B', epsilon", other_designations="PP2A B subunit isoform B'-epsilon|PP2A B subunit isoform B56-epsilon|PP2A B subunit isoform PR61-epsilon|PP2A B subunit isoform R5-epsilon|PP2A, B subunit, B' epsilon|PP2A, B subunit, B56 epsilon|PP2A, B subunit, PR61 epsilon|PP2A, B subunit, R5 epsilon|epsilon isoform of regulatory subunit B56, protein phosphatase 2A|protein phosphatase 2, regulatory subunit B (B56), epsilon isoform|protein phosphatase 2, regulatory subunit B', epsilon isoform|serine/threonine protein phosphatase 2A, 56 kDa regulatory subunit, epsilon", map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63372479, end_position_on_the_genomic_accession=63543361, orientation='minus', exon_count=15, OMIM=601647, no_hdr0=''),
  5530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5530, CurrentID=0, Status='live', Symbol='PPP3CA', Aliases='CALN, CALNA, CALNA1, CCN1, CNA1, PPP2B', description='protein phosphatase 3, catalytic subunit, alpha isozyme', other_designations='CAM-PRP catalytic subunit|calcineurin A alpha|calmodulin-dependent calcineurin A subunit alpha isoform|protein phosphatase 2B, catalytic subunit, alpha isoform|protein phosphatase 3 (formerly 2B), catalytic subunit, alpha isoform|protein phosphatase 3 catalytic subunit alpha isozyme', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101023430, end_position_on_the_genomic_accession=101347471, orientation='minus', exon_count=14, OMIM=114105, no_hdr0=''),
  5531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5531, CurrentID=0, Status='live', Symbol='PPP4C', Aliases='PP-X, PP4, PP4C, PPH3, PPP4, PPX', description='protein phosphatase 4 catalytic subunit', other_designations='protein phosphatase 4 (formerly X), catalytic subunit|protein phosphatase 4, catalytic subunit|protein phosphatase X, catalytic subunit', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30075976, end_position_on_the_genomic_accession=30085377, orientation='plus', exon_count=9, OMIM=602035, no_hdr0=''),
  5532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5532, CurrentID=0, Status='live', Symbol='PPP3CB', Aliases='CALNA2, CALNB, CNA2, PP2Bbeta', description='protein phosphatase 3, catalytic subunit, beta isozyme', other_designations='CAM-PRP catalytic subunit|calcineurin A beta|calcineurin A2|calmodulin-dependent calcineurin A subunit beta isoform|protein phosphatase 2B, catalytic subunit, beta isoform|protein phosphatase 3 (formerly 2B), catalytic subunit, beta isoform|protein phosphatase from PCR fragment H32', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73436428, end_position_on_the_genomic_accession=73496024, orientation='minus', exon_count=17, OMIM=114106, no_hdr0=''),
  5533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5533, CurrentID=0, Status='live', Symbol='PPP3CC', Aliases='CALNA3, CNA3, PP2Bgamma', description='protein phosphatase 3, catalytic subunit, gamma isozyme', other_designations='CAM-PRP catalytic subunit|calcineurin, testis-specific catalytic subunit|calmodulin-dependent calcineurin A subunit gamma isoform|protein phosphatase 2B, catalytic subunit, gamma isoform|protein phosphatase 3 (formerly 2B), catalytic subunit, gamma isoform', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22440970, end_position_on_the_genomic_accession=22541144, orientation='plus', exon_count=17, OMIM=114107, no_hdr0=''),
  5534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5534, CurrentID=0, Status='live', Symbol='PPP3R1', Aliases='CALNB1, CNB, CNB1', description='protein phosphatase 3 regulatory subunit B, alpha', other_designations='calcineurin B, type I (19kDa)|protein phosphatase 2B regulatory subunit 1|protein phosphatase 2B regulatory subunit B alpha|protein phosphatase 3 (formerly 2B), regulatory subunit B (19kD), alpha isoform (calcineurin B, type I)|protein phosphatase 3 (formerly 2B), regulatory subunit B, 19kDa, alpha isoform (calcineurin B, type I)|protein phosphatase 3 (formerly 2B), regulatory subunit B, alpha isoform|protein phosphatase 3, regulatory subunit B, alpha', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68178857, end_position_on_the_genomic_accession=68252519, orientation='minus', exon_count=6, OMIM=601302, no_hdr0=''),
  5535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5535, CurrentID=0, Status='live', Symbol='PPP3R2', Aliases='PPP3RL', description='protein phosphatase 3 regulatory subunit B, beta', other_designations='CBLP|CNBII|calcineurin B, type II (19kDa)|calcineurin B-like protein|calcineurin BII|protein phosphatase 2B regulatory subunit 2|protein phosphatase 3 (formerly 2B), regulatory subunit B (19kD), beta isoform (calcineurin B, type II)|protein phosphatase 3 (formerly 2B), regulatory subunit B, 19kDa, beta isoform (calcineurin B, type II)|protein phosphatase 3 (formerly 2B), regulatory subunit B, beta isoform|protein phosphatase 3 regulatory subunit B beta isoform|protein phosphatase 3, regulatory subunit B, beta', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101591615, end_position_on_the_genomic_accession=101595001, orientation='minus', exon_count=1, OMIM=613821, no_hdr0=''),
  5536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5536, CurrentID=0, Status='live', Symbol='PPP5C', Aliases='PP5, PPP5, PPT', description='protein phosphatase 5, catalytic subunit', other_designations='PP-T|protein phosphatase T', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46346994, end_position_on_the_genomic_accession=46390975, orientation='plus', exon_count=13, OMIM=600658, no_hdr0=''),
  5537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5537, CurrentID=0, Status='live', Symbol='PPP6C', Aliases='PP6, PP6C', description='protein phosphatase 6, catalytic subunit', other_designations='protein phosphatase 6 catalytic subunit|serine/threonine protein phosphatase catalytic subunit', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125146573, end_position_on_the_genomic_accession=125189939, orientation='minus', exon_count=9, OMIM=612725, no_hdr0=''),
  5538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5538, CurrentID=0, Status='live', Symbol='PPT1', Aliases='CLN1, INCL, PPT', description='palmitoyl-protein thioesterase 1', other_designations='ceroid-palmitoyl-palmitoyl-protein thioesterase 1|palmitoyl-protein hydrolase 1', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40072710, end_position_on_the_genomic_accession=40097470, orientation='minus', exon_count=9, OMIM=600722, no_hdr0=''),
  5539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5539, CurrentID=0, Status='live', Symbol='PPY', Aliases='PNP', description='pancreatic polypeptide', other_designations='PP|pancreatic polypeptide Y|prepro-PP (prepropancreatic polypeptide)', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43940804, end_position_on_the_genomic_accession=43944248, orientation='minus', exon_count=5, OMIM=167780, no_hdr0=''),
  5540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5540, CurrentID=0, Status='live', Symbol='NPY4R', Aliases='NPY4-R, PP1, PPYR1, Y4', description='neuropeptide Y receptor Y4', other_designations='pancreatic polypeptide receptor 1', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46458551, end_position_on_the_genomic_accession=46470647, orientation='minus', exon_count=8, OMIM=601790, no_hdr0=''),
  5541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5541, CurrentID=0, Status='live', Symbol='PR@', Aliases='SPC', description='proline-rich protein gene cluster', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5542, CurrentID=0, Status='live', Symbol='PRB1', Aliases='PM, PMF, PMSL, PRB1M, PRB1', description='proline-rich protein BstNI subfamily 1', other_designations='BstNI type basic salivary proline-rich protein 1|parotid middle band protein|salivary protein Pe', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11351823, end_position_on_the_genomic_accession=11355590, orientation='minus', exon_count=5, OMIM=180989, no_hdr0=''),
  5544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5544, CurrentID=0, Status='live', Symbol='PRB3', Aliases='G1, PRG', description='proline-rich protein BstNI subfamily 3', other_designations='BstNI type basic salivary proline-rich protein 3|G1 parotid salivary glycoprotein|parotid salivary glycoprotein G1|proline-rich protein G1', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11265914, end_position_on_the_genomic_accession=11269707, orientation='minus', exon_count=5, OMIM=168840, no_hdr0=''),
  5545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5545, CurrentID=0, Status='live', Symbol='PRB4', Aliases='Po', description='proline-rich protein BstNI subfamily 4', other_designations='BstNI type basic salivary proline-rich protein 4|parotid o protein|salivary proline-rich protein II-1|salivary proline-rich protein Po', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11307081, end_position_on_the_genomic_accession=11310435, orientation='minus', exon_count=4, OMIM=180990, no_hdr0=''),
  5546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5546, CurrentID=0, Status='live', Symbol='PRCC', Aliases='RCCP1, TPRC', description='papillary renal cell carcinoma (translocation-associated)', other_designations='papillary renal cell carcinoma translocation-associated gene protein', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156767482, end_position_on_the_genomic_accession=156800817, orientation='plus', exon_count=9, OMIM=179755, no_hdr0=''),
  5547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5547, CurrentID=0, Status='live', Symbol='PRCP', Aliases='HUMPCP, PCP', description='prolylcarboxypeptidase', other_designations='angiotensinase C|lysosomal carboxypeptidase C|proline carboxypeptidase|prolylcarboxypeptidase (angiotensinase C)|prolylcarboxypeptidase isoform 1 preproprotein', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82824367, end_position_on_the_genomic_accession=82901691, orientation='minus', exon_count=11, OMIM=176785, no_hdr0=''),
  5548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5548, CurrentID=0, Status='live', Symbol='PRD', Aliases='', description='primary retinal dysplasia', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=312550, no_hdr0=''),
  5549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5549, CurrentID=0, Status='live', Symbol='PRELP', Aliases='MST161, MSTP161, SLRR2A', description='proline/arginine-rich end leucine-rich repeat protein', other_designations='55 kDa leucine-rich repeat protein of articular cartilage|prolargin proteoglycan|proline-arginine-rich end leucine-rich repeat protein', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203475755, end_position_on_the_genomic_accession=203491352, orientation='plus', exon_count=3, OMIM=601914, no_hdr0=''),
  5550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5550, CurrentID=0, Status='live', Symbol='PREP', Aliases='PE, PEP', description='prolyl endopeptidase', other_designations='dJ355L5.1 (prolyl endopeptidase)|post-proline cleaving enzyme|prolyl oligopeptidase', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105277567, end_position_on_the_genomic_accession=105403124, orientation='minus', exon_count=17, OMIM=600400, no_hdr0=''),
  5551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5551, CurrentID=0, Status='live', Symbol='PRF1', Aliases='FLH2, HPLH2, P1, PFN1, PFP', description='perforin 1 (pore forming protein)', other_designations='cytolysin|lymphocyte pore forming protein|lymphocyte pore-forming protein', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70597348, end_position_on_the_genomic_accession=70602775, orientation='minus', exon_count=3, OMIM=170280, no_hdr0=''),
  5552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5552, CurrentID=0, Status='live', Symbol='SRGN', Aliases='PPG, PRG, PRG1', description='serglycin', other_designations='hematopoetic proteoglycan core peptide|hematopoetic proteoglycan core protein|hematopoietic proteoglycan core protein|p.PG|platelet proteoglycan core protein|proteoglycan 1, secretory granule|proteoglycan protein core for mast cell secretory granule|secretory granule proteoglycan core peptide|secretory granule proteoglycan core protein|serglycin proteoglycan', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69088072, end_position_on_the_genomic_accession=69104811, orientation='plus', exon_count=3, OMIM=177040, no_hdr0=''),
  5553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5553, CurrentID=0, Status='live', Symbol='PRG2', Aliases='BMPG, MBP, MBP1', description='proteoglycan 2, bone marrow (natural killer cell activator, eosinophil granule major basic protein)', other_designations='bone-marrow proteoglycan|eosinophil granule major basic protein|eosinophil major basic protein|natural killer cell activator|proteoglycan 2 preproprotein', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57386787, end_position_on_the_genomic_accession=57390657, orientation='minus', exon_count=6, OMIM=605601, no_hdr0=''),
  5554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5554, CurrentID=0, Status='live', Symbol='PRH1', Aliases='Db-s, PA, PIF-S, PRH2, Pr1/Pr2', description='proline-rich protein HaeIII subfamily 1', other_designations='PRP-1/PRP-2|parotid acidic protein|parotid double-band protein|parotid isoelectric focusing variant protein|parotid proline-rich protein 1/2|protein C', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10880961, end_position_on_the_genomic_accession=11171623, orientation='minus', exon_count=7, OMIM=168730, no_hdr0=''),
  5555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5555, CurrentID=0, Status='live', Symbol='PRH2', Aliases='PRP-1/PRP-2, Pr, pr1/Pr2', description='proline-rich protein HaeIII subfamily 2', other_designations='acidic salivary proline-rich protein, HaeIII type, 2|parotid acidic protein|parotid double-band protein|parotid proline-rich protein 1/2|protein C', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10929235, end_position_on_the_genomic_accession=10934845, orientation='plus', exon_count=4, OMIM=168790, no_hdr0=''),
  5557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5557, CurrentID=0, Status='live', Symbol='PRIM1', Aliases='p49', description='primase, DNA, polypeptide 1 (49kDa)', other_designations='DNA primase 1|DNA primase 49 kDa subunit|DNA primase subunit 48|primase p49 subunit|primase polypeptide 1, 49kDa', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56731580, end_position_on_the_genomic_accession=56752362, orientation='minus', exon_count=13, OMIM=176635, no_hdr0=''),
  5558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5558, CurrentID=0, Status='live', Symbol='PRIM2', Aliases='PRIM2A, p58', description='primase, DNA, polypeptide 2 (58kDa)', other_designations='DNA primase 58 kDa subunit', map_location='6p12-p11.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57314805, end_position_on_the_genomic_accession=57646850, orientation='plus', exon_count=19, OMIM=176636, no_hdr0=''),
  5562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5562, CurrentID=0, Status='live', Symbol='PRKAA1', Aliases='AMPK, AMPKa1', description='protein kinase, AMP-activated, alpha 1 catalytic subunit', other_designations="5'-AMP-activated protein kinase, catalytic alpha-1 chain|ACACA kinase|AMP -activate kinase alpha 1 subunit|AMP-activated protein kinase, catalytic, alpha-1|AMPK alpha 1|AMPK subunit alpha-1|HMGCR kinase|acetyl-CoA carboxylase kinase|hydroxymethylglutaryl-CoA reductase kinase|tau-protein kinase PRKAA1", map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40759379, end_position_on_the_genomic_accession=40798195, orientation='minus', exon_count=10, OMIM=602739, no_hdr0=''),
  5563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5563, CurrentID=0, Status='live', Symbol='PRKAA2', Aliases='AMPK, AMPK2, AMPKa2, PRKAA', description='protein kinase, AMP-activated, alpha 2 catalytic subunit', other_designations="5'-AMP-activated protein kinase, catalytic alpha-2 chain|ACACA kinase|AMP-activated protein kinase alpha-2 subunit variant 2|AMP-activated protein kinase alpha-2 subunit variant 3|AMPK alpha 2|AMPK subunit alpha-2|AMPK-alpha-2 chain|HMGCR kinase|acetyl-CoA carboxylase kinase|hydroxymethylglutaryl-CoA reductase kinase", map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56645317, end_position_on_the_genomic_accession=56715335, orientation='plus', exon_count=9, OMIM=600497, no_hdr0=''),
  5564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5564, CurrentID=0, Status='live', Symbol='PRKAB1', Aliases='AMPK, HAMPKb', description='protein kinase, AMP-activated, beta 1 non-catalytic subunit', other_designations="5'-AMP-activated protein kinase beta-1 subunit|AMP-activated protein kinase beta subunit|AMPK beta -1 chain|AMPK beta 1|AMPK subunit beta-1|AMPKb|protein kinase, AMP-activated, noncatalytic, beta-1", map_location='12q24.1-q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119667864, end_position_on_the_genomic_accession=119681624, orientation='plus', exon_count=7, OMIM=602740, no_hdr0=''),
  5565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5565, CurrentID=0, Status='live', Symbol='PRKAB2', Aliases='', description='protein kinase, AMP-activated, beta 2 non-catalytic subunit', other_designations="5'-AMP-activated protein kinase, beta-2 subunit|AMP-activated protein kinase beta 2 non-catalytic subunit|AMPK beta 2|AMPK beta-2 chain|AMPK subunit beta-2", map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147155106, end_position_on_the_genomic_accession=147172589, orientation='minus', exon_count=9, OMIM=602741, no_hdr0=''),
  5566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5566, CurrentID=0, Status='live', Symbol='PRKACA', Aliases='PKACA, PPNAD4', description='protein kinase, cAMP-dependent, alpha catalytic subunit', other_designations='PKA C-alpha|cAMP-dependent protein kinase catalytic subunit alpha|protein kinase A catalytic subunit|protein kinase, cAMP-dependent, catalytic, alpha', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14091688, end_position_on_the_genomic_accession=14117747, orientation='minus', exon_count=12, OMIM=601639, no_hdr0=''),
  5567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5567, CurrentID=0, Status='live', Symbol='PRKACB', Aliases='PKA C-beta, PKACB', description='protein kinase, cAMP-dependent, beta catalytic subunit', other_designations='cAMP-dependent protein kinase catalytic beta subunit isoform 4ab|protein kinase A catalytic subunit beta|protein kinase, cAMP-dependent, catalytic, beta', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84077975, end_position_on_the_genomic_accession=84238498, orientation='plus', exon_count=18, OMIM=176892, no_hdr0=''),
  5568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5568, CurrentID=0, Status='live', Symbol='PRKACG', Aliases='BDPLT19, KAPG, PKACg', description='protein kinase, cAMP-dependent, gamma catalytic subunit', other_designations='PKA C-gamma|protein kinase, cAMP-dependent, catalytic, gamma|serine(threonine) protein kinase', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69012529, end_position_on_the_genomic_accession=69014123, orientation='minus', exon_count=1, OMIM=176893, no_hdr0=''),
  5569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5569, CurrentID=0, Status='live', Symbol='PKIA', Aliases='PRKACN1', description='protein kinase (cAMP-dependent, catalytic) inhibitor alpha', other_designations='PKI-alpha|cAMP-dependent protein kinase inhibitor, muscle/brain isoform', map_location='8q21.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78516101, end_position_on_the_genomic_accession=78605267, orientation='plus', exon_count=4, OMIM=606059, no_hdr0=''),
  5570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5570, CurrentID=0, Status='live', Symbol='PKIB', Aliases='PRKACN2', description='protein kinase (cAMP-dependent, catalytic) inhibitor beta', other_designations='PKI-beta|cAMP-dependent protein kinase inhibitor 2', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122471917, end_position_on_the_genomic_accession=122726373, orientation='plus', exon_count=18, OMIM=606914, no_hdr0=''),
  5571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5571, CurrentID=0, Status='live', Symbol='PRKAG1', Aliases='AMPKG', description='protein kinase, AMP-activated, gamma 1 non-catalytic subunit', other_designations="5'-AMP-activated protein kinase, gamma-1 subunit|AMPK gamma-1 chain|AMPK gamma1", map_location='12q12-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49002272, end_position_on_the_genomic_accession=49018980, orientation='minus', exon_count=14, OMIM=602742, no_hdr0=''),
  5573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5573, CurrentID=0, Status='live', Symbol='PRKAR1A', Aliases='ACRDYS1, ADOHR, CAR, CNC, CNC1, PKR1, PPNAD1, PRKAR1, TSE1', description='protein kinase, cAMP-dependent, regulatory subunit type I alpha', other_designations='Carney complex type 1|cAMP-dependent protein kinase regulatory subunit RIalpha|cAMP-dependent protein kinase type I-alpha regulatory chain|protein kinase A type 1a regulatory subunit|protein kinase, cAMP-dependent, regulatory, type I, alpha|tissue-specific extinguisher 1', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68413623, end_position_on_the_genomic_accession=68551316, orientation='plus', exon_count=16, OMIM=188830, no_hdr0=''),
  5574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5574, CurrentID=0, Status='live', Symbol='PRKAR1AP', Aliases='', description='protein kinase, cAMP-dependent, regulatory subunit type I alpha pseudogene', other_designations='protein kinase, cAMP-dependent, regulatory, type I, alpha (tissue specific extinguisher 1) pseudogene|protein kinase, cAMP-dependent, regulatory, type I, alpha pseudogene', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92125297, end_position_on_the_genomic_accession=92126301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5575, CurrentID=0, Status='live', Symbol='PRKAR1B', Aliases='PRKAR1', description='protein kinase, cAMP-dependent, regulatory subunit type I beta', other_designations='protein kinase, cAMP-dependent, regulatory, type I, beta', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=549196, end_position_on_the_genomic_accession=727676, orientation='minus', exon_count=18, OMIM=176911, no_hdr0=''),
  5576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5576, CurrentID=0, Status='live', Symbol='PRKAR2A', Aliases='PKR2, PRKAR2', description='protein kinase, cAMP-dependent, regulatory subunit type II alpha', other_designations='cAMP-dependent protein kinase regulatory subunit RII alpha|protein kinase A, RII-alpha subunit|protein kinase cAMP-dependent regulatory type II alpha|protein kinase, cAMP-dependent, regulatory, type II, alpha', map_location='3p21.3-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48744601, end_position_on_the_genomic_accession=48847837, orientation='minus', exon_count=15, OMIM=176910, no_hdr0=''),
  5577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5577, CurrentID=0, Status='live', Symbol='PRKAR2B', Aliases='PRKAR2, RII-BETA', description='protein kinase, cAMP-dependent, regulatory subunit type II beta', other_designations='H_RG363E19.2|WUGSC:H_RG363E19.2|cAMP-dependent protein kinase type II-beta regulatory chain|protein kinase, cAMP-dependent, regulatory, type II, beta', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107044728, end_position_on_the_genomic_accession=107161811, orientation='plus', exon_count=12, OMIM=176912, no_hdr0=''),
  5578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5578, CurrentID=0, Status='live', Symbol='PRKCA', Aliases='AAG6, PKC-alpha, PKCA, PRKACA', description='protein kinase C, alpha', other_designations='PKC-A|aging-associated gene 6', map_location='17q22-q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66302782, end_position_on_the_genomic_accession=66810744, orientation='plus', exon_count=21, OMIM=176960, no_hdr0=''),
  5579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5579, CurrentID=0, Status='live', Symbol='PRKCB', Aliases='PKC-beta, PKCB1, PRKCB2, PRKCB', description='protein kinase C, beta', other_designations='PKC-B|protein kinase C, beta 1 polypeptide', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23835979, end_position_on_the_genomic_accession=24220611, orientation='plus', exon_count=17, OMIM=176970, no_hdr0=''),
  5580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5580, CurrentID=0, Status='live', Symbol='PRKCD', Aliases='ALPS3, CVID9, MAY1, PKCD, nPKC-delta', description='protein kinase C, delta', other_designations='protein kinase C delta VIII|tyrosine-protein kinase PRKCD', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53161207, end_position_on_the_genomic_accession=53192717, orientation='plus', exon_count=22, OMIM=176977, no_hdr0=''),
  5581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5581, CurrentID=0, Status='live', Symbol='PRKCE', Aliases='PKCE, nPKC-epsilon', description='protein kinase C, epsilon', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45651315, end_position_on_the_genomic_accession=46187990, orientation='plus', exon_count=34, OMIM=176975, no_hdr0=''),
  5582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5582, CurrentID=0, Status='live', Symbol='PRKCG', Aliases='PKC-gamma, PKCC, PKCG, SCA14', description='protein kinase C, gamma', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53882213, end_position_on_the_genomic_accession=53907652, orientation='plus', exon_count=19, OMIM=176980, no_hdr0=''),
  5583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5583, CurrentID=0, Status='live', Symbol='PRKCH', Aliases='PKC-L, PKCL, PRKCL, nPKC-eta', description='protein kinase C, eta', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61321443, end_position_on_the_genomic_accession=61550980, orientation='plus', exon_count=18, OMIM=605437, no_hdr0=''),
  5584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5584, CurrentID=0, Status='live', Symbol='PRKCI', Aliases='DXS1179E, PKCI, nPKC-iota', description='protein kinase C, iota', other_designations='PRKC-lambda/iota|aPKC-lambda/iota|atypical protein kinase C-lambda/iota', map_location='3q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170222432, end_position_on_the_genomic_accession=170305982, orientation='plus', exon_count=18, OMIM=600539, no_hdr0=''),
  5585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5585, CurrentID=0, Status='live', Symbol='PKN1', Aliases='DBK, PAK-1, PAK1, PKN, PKN-ALPHA, PRK1, PRKCL1', description='protein kinase N1', other_designations='protease-activated kinase 1|protein kinase C-like 1|protein kinase C-like PKN|protein kinase C-related kinase 1|protein kinase PKN-alpha|protein-kinase C-related kinase 1|serine-threonine kinase N|serine-threonine protein kinase N|serine/threonine protein kinase N', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14433053, end_position_on_the_genomic_accession=14471867, orientation='plus', exon_count=25, OMIM=601032, no_hdr0=''),
  5586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5586, CurrentID=0, Status='live', Symbol='PKN2', Aliases='PAK2, PRK2, PRKCL2, PRO2042, Pak-2, STK7', description='protein kinase N2', other_designations='PKN gamma|cardiolipin-activated protein kinase Pak2|protein kinase C-like 2|protein-kinase C-related kinase 2', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88684193, end_position_on_the_genomic_accession=88836255, orientation='plus', exon_count=24, OMIM=602549, no_hdr0=''),
  5587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5587, CurrentID=0, Status='live', Symbol='PRKD1', Aliases='PKC-MU, PKCM, PKD, PRKCM', description='protein kinase D1', other_designations='nPKC-D1|nPKC-mu|protein kinase C mu type|protein kinase C, mu|protein kinase D', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=29576479, end_position_on_the_genomic_accession=29927693, orientation='minus', exon_count=22, OMIM=605435, no_hdr0=''),
  5588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5588, CurrentID=0, Status='live', Symbol='PRKCQ', Aliases='PRKCT, nPKC-theta', description='protein kinase C, theta', other_designations='', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6393595, end_position_on_the_genomic_accession=6580646, orientation='minus', exon_count=22, OMIM=600448, no_hdr0=''),
  5589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5589, CurrentID=0, Status='live', Symbol='PRKCSH', Aliases='AGE-R2, G19P1, PCLD, PKCSH, PLD1', description='protein kinase C substrate 80K-H', other_designations='AGE-binding receptor 2|glucosidase II subunit beta|hepatocystin|protein kinase C substrate 60.1 kDa protein heavy chain|protein kinase C substrate, 80 Kda protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11435257, end_position_on_the_genomic_accession=11450968, orientation='plus', exon_count=19, OMIM=177060, no_hdr0=''),
  5590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5590, CurrentID=0, Status='live', Symbol='PRKCZ', Aliases='PKC-ZETA, PKC2', description='protein kinase C, zeta', other_designations='nPKC-zeta', map_location='1p36.33-p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2050470, end_position_on_the_genomic_accession=2185395, orientation='plus', exon_count=27, OMIM=176982, no_hdr0=''),
  5591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5591, CurrentID=0, Status='live', Symbol='PRKDC', Aliases='DNA-PKcs, DNAPK, DNPK1, HYRC, HYRC1, IMD26, XRCC7, p350', description='protein kinase, DNA-activated, catalytic polypeptide', other_designations='DNA-PK catalytic subunit|hyper-radiosensitivity of murine scid mutation, complementing 1|p460', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47773108, end_position_on_the_genomic_accession=47960183, orientation='minus', exon_count=86, OMIM=600899, no_hdr0=''),
  5592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5592, CurrentID=0, Status='live', Symbol='PRKG1', Aliases='AAT8, PKGB, PRKGR1B, cGK, cGK 1, cGK1, cGKI, cGKI-BETA, cGKI-alpha, PRKG1', description='protein kinase, cGMP-dependent, type I', other_designations='protein kinase, cGMP-dependent, regulatory, type I, beta', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50991151, end_position_on_the_genomic_accession=52298350, orientation='plus', exon_count=21, OMIM=176894, no_hdr0=''),
  5593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5593, CurrentID=0, Status='live', Symbol='PRKG2', Aliases='PRKGR2, cGK2, cGKII', description='protein kinase, cGMP-dependent, type II', other_designations='cGMP-dependent protein kinase II', map_location='4q13.1-q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81087370, end_position_on_the_genomic_accession=81216743, orientation='minus', exon_count=24, OMIM=601591, no_hdr0=''),
  5594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5594, CurrentID=0, Status='live', Symbol='MAPK1', Aliases='ERK, ERK-2, ERK2, ERT1, MAPK2, P42MAPK, PRKM1, PRKM2, p38, p40, p41, p41mapk, p42-MAPK', description='mitogen-activated protein kinase 1', other_designations='MAP kinase 1|MAP kinase 2|MAP kinase isoform p42|MAPK 2|extracellular signal-regulated kinase 2|mitogen-activated protein kinase 2|protein tyrosine kinase ERK2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21759657, end_position_on_the_genomic_accession=21867680, orientation='minus', exon_count=9, OMIM=176948, no_hdr0=''),
  5595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5595, CurrentID=0, Status='live', Symbol='MAPK3', Aliases='ERK-1, ERK1, ERT2, HS44KDAP, HUMKER1A, P44ERK1, P44MAPK, PRKM3, p44-ERK1, p44-MAPK', description='mitogen-activated protein kinase 3', other_designations='MAP kinase isoform p44|MAPK 1|extracellular signal-regulated kinase 1|extracellular signal-related kinase 1|insulin-stimulated MAP2 kinase|microtubule-associated protein 2 kinase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30114105, end_position_on_the_genomic_accession=30123309, orientation='minus', exon_count=10, OMIM=601795, no_hdr0=''),
  5596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5596, CurrentID=0, Status='live', Symbol='MAPK4', Aliases='ERK-4, ERK4, PRKM4, p63-MAPK, p63MAPK', description='mitogen-activated protein kinase 4', other_designations='Erk3-related|MAP kinase isoform p63|MAPK 4|extracellular signal-regulated kinase 4', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50560114, end_position_on_the_genomic_accession=50731826, orientation='plus', exon_count=12, OMIM=176949, no_hdr0=''),
  5597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5597, CurrentID=0, Status='live', Symbol='MAPK6', Aliases='ERK3, HsT17250, PRKM6, p97MAPK', description='mitogen-activated protein kinase 6', other_designations='ERK-3|MAP kinase 6|MAP kinase isoform p97|MAPK 6|extracellular signal-regulated kinase 3|extracellular signal-regulated kinase, p97|p97-MAPK|protein kinase, mitogen-activated 5|protein kinase, mitogen-activated 6', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52004265, end_position_on_the_genomic_accession=52066265, orientation='plus', exon_count=9, OMIM=602904, no_hdr0=''),
  5598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5598, CurrentID=0, Status='live', Symbol='MAPK7', Aliases='BMK1, ERK4, ERK5, PRKM7', description='mitogen-activated protein kinase 7', other_designations='BMK-1|BMK1 kinase|ERK-5|MAP kinase 7|MAPK 7|big MAP kinase 1|extracellular signal-regulated kinase 5|extracellular-signal-regulated kinase 5', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19377721, end_position_on_the_genomic_accession=19383544, orientation='plus', exon_count=9, OMIM=602521, no_hdr0=''),
  5599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5599, CurrentID=0, Status='live', Symbol='MAPK8', Aliases='JNK, JNK-46, JNK1, JNK1A2, JNK21B1/2, PRKM8, SAPK1, SAPK1c', description='mitogen-activated protein kinase 8', other_designations='JUN N-terminal kinase|MAP kinase 8|c-Jun N-terminal kinase 1|mitogen-activated protein kinase 8 isoform JNK1 alpha1|mitogen-activated protein kinase 8 isoform JNK1 beta2|stress-activated protein kinase 1|stress-activated protein kinase 1c', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48306639, end_position_on_the_genomic_accession=48439360, orientation='plus', exon_count=15, OMIM=601158, no_hdr0=''),
  5600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5600, CurrentID=0, Status='live', Symbol='MAPK11', Aliases='P38B, P38BETA2, PRKM11, SAPK2, SAPK2B, p38-2, p38Beta', description='mitogen-activated protein kinase 11', other_designations='MAP kinase 11|MAP kinase p38 beta|mitogen-activated protein kinase p38 beta|mitogen-activated protein kinase p38-2|stress-activated protein kinase-2|stress-activated protein kinase-2b', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50263713, end_position_on_the_genomic_accession=50270393, orientation='minus', exon_count=13, OMIM=602898, no_hdr0=''),
  5601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5601, CurrentID=0, Status='live', Symbol='MAPK9', Aliases='JNK-55, JNK2, JNK2A, JNK2ALPHA, JNK2B, JNK2BETA, PRKM9, SAPK, SAPK1a, p54a, p54aSAPK', description='mitogen-activated protein kinase 9', other_designations='Jun kinase|MAP kinase 9|MAPK 9|c-Jun N-terminal kinase 2|c-Jun kinase 2|stress-activated protein kinase 1a|stress-activated protein kinase JNK2', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180233594, end_position_on_the_genomic_accession=180292071, orientation='minus', exon_count=16, OMIM=602896, no_hdr0=''),
  5602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5602, CurrentID=0, Status='live', Symbol='MAPK10', Aliases='JNK3, JNK3A, PRKM10, SAPK1b, p493F12, p54bSAPK', description='mitogen-activated protein kinase 10', other_designations='JNK3 alpha protein kinase|MAP kinase 10|MAP kinase p49 3F12|MAPK 10|c-Jun N-terminal kinase 3|stress activated protein kinase beta|stress-activated protein kinase 1b|stress-activated protein kinase JNK3', map_location='4q22.1-q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86012299, end_position_on_the_genomic_accession=86594110, orientation='minus', exon_count=22, OMIM=602897, no_hdr0=''),
  5603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5603, CurrentID=0, Status='live', Symbol='MAPK13', Aliases='MAPK 13, MAPK-13, PRKM13, SAPK4, p38delta', description='mitogen-activated protein kinase 13', other_designations='MAP kinase 13|MAP kinase p38 delta|mitogen-activated protein kinase p38 delta|stress-activated protein kinase 4', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36130484, end_position_on_the_genomic_accession=36144524, orientation='plus', exon_count=12, OMIM=602899, no_hdr0=''),
  5604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5604, CurrentID=0, Status='live', Symbol='MAP2K1', Aliases='CFC3, MAPKK1, MEK1, MKK1, PRKMK1', description='mitogen-activated protein kinase kinase 1', other_designations='ERK activator kinase 1|MAPK/ERK kinase 1|MAPKK 1|MEK 1|protein kinase, mitogen-activated, kinase 1 (MAP kinase kinase 1)', map_location='15q22.1-q22.33', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66386873, end_position_on_the_genomic_accession=66491544, orientation='plus', exon_count=12, OMIM=176872, no_hdr0=''),
  5605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5605, CurrentID=0, Status='live', Symbol='MAP2K2', Aliases='CFC4, MAPKK2, MEK2, MKK2, PRKMK2', description='mitogen-activated protein kinase kinase 2', other_designations='ERK activator kinase 2|MAP kinase kinase 2|MAPK/ERK kinase 2|mitogen-activated protein kinase kinase 2, p45', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4090321, end_position_on_the_genomic_accession=4124184, orientation='minus', exon_count=12, OMIM=601263, no_hdr0=''),
  5606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5606, CurrentID=0, Status='live', Symbol='MAP2K3', Aliases='MAPKK3, MEK3, MKK3, PRKMK3, SAPKK-2, SAPKK2', description='mitogen-activated protein kinase kinase 3', other_designations='MAP kinase kinase 3|MAPK/ERK kinase 3|MAPKK 3|MEK 3|SAPK kinase 2|stress-activated protein kinase kinase 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21284656, end_position_on_the_genomic_accession=21315240, orientation='plus', exon_count=15, OMIM=602315, no_hdr0=''),
  5607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5607, CurrentID=0, Status='live', Symbol='MAP2K5', Aliases='HsT17454, MAPKK5, MEK5, PRKMK5', description='mitogen-activated protein kinase kinase 5', other_designations='MAP kinase kinase 5|MAP kinase kinase MEK5b|MAPK/ERK kinase 5|MAPKK 5|MEK 5', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67542606, end_position_on_the_genomic_accession=67807117, orientation='plus', exon_count=27, OMIM=602520, no_hdr0=''),
  5608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5608, CurrentID=0, Status='live', Symbol='MAP2K6', Aliases='MAPKK6, MEK6, MKK6, PRKMK6, SAPKK-3, SAPKK3', description='mitogen-activated protein kinase kinase 6', other_designations='MAPK/ERK kinase 6|MAPKK 6|MEK 6|SAPK kinase 3|protein kinase, mitogen-activated, kinase 6 (MAP kinase kinase 6)|stress-activated protein kinase kinase 3', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69414697, end_position_on_the_genomic_accession=69542329, orientation='plus', exon_count=16, OMIM=601254, no_hdr0=''),
  5609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5609, CurrentID=0, Status='live', Symbol='MAP2K7', Aliases='JNKK2, MAPKK7, MEK, MEK 7, MKK7, PRKMK7, SAPKK-4, SAPKK4', description='mitogen-activated protein kinase kinase 7', other_designations='JNK-activating kinase 2|MAP kinase kinase 7|MAPK/ERK kinase 7|SAPK kinase 4|c-Jun N-terminal kinase kinase 2|stress-activated protein kinase kinase 4', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7903780, end_position_on_the_genomic_accession=7914483, orientation='plus', exon_count=12, OMIM=603014, no_hdr0=''),
  5610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5610, CurrentID=0, Status='live', Symbol='EIF2AK2', Aliases='EIF2AK1, PKR, PPP1R83, PRKR', description='eukaryotic translation initiation factor 2 alpha kinase 2', other_designations='P1/eIF-2A protein kinase|double stranded RNA activated protein kinase|eIF-2A protein kinase 2|eukaryotic translation initiation factor 2-alpha kinase 2|interferon-inducible elF2alpha kinase|p68 kinase|protein kinase R|protein kinase, interferon-inducible double stranded RNA dependent|protein phosphatase 1, regulatory subunit 83|tyrosine-protein kinase EIF2AK2', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37104683, end_position_on_the_genomic_accession=37157065, orientation='minus', exon_count=17, OMIM=176871, no_hdr0=''),
  5611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5611, CurrentID=0, Status='live', Symbol='DNAJC3', Aliases='ACPHD, ERdj6, HP58, P58, P58IPK, PRKRI', description='DnaJ heat shock protein family (Hsp40) member C3', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 3|ER-resident protein ERdj6|endoplasmic reticulum DNA J domain-containing protein 6|interferon-induced, double-stranded RNA-activated protein kinase inhibitor|protein kinase inhibitor of 58 kDa|protein kinase inhibitor p58|protein-kinase, interferon-inducible double stranded RNA dependent inhibitor', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95677139, end_position_on_the_genomic_accession=95794989, orientation='plus', exon_count=14, OMIM=601184, no_hdr0=''),
  5612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5612, CurrentID=0, Status='live', Symbol='PRKRIR', Aliases='DAP4, P52rIPK, THAP0, THAP12', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor)', other_designations='58 kDa interferon-induced protein kinase-interacting protein|THAP domain containing 12|THAP domain-containing protein 0|death-associated protein 4|inhibitor of protein kinase PKR|p58IPK-interacting protein', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76349956, end_position_on_the_genomic_accession=76380965, orientation='minus', exon_count=6, OMIM=607374, no_hdr0=''),
  5613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5613, CurrentID=0, Status='live', Symbol='PRKX', Aliases='PKX1', description='protein kinase, X-linked', other_designations='protein kinase PKX1|protein kinase X|serine/threonine-protein kinase PRKX', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3604343, end_position_on_the_genomic_accession=3713644, orientation='minus', exon_count=11, OMIM=300083, no_hdr0=''),
  5615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5615, CurrentID=0, Status='live', Symbol='PRKXP2', Aliases='', description='protein kinase, X-linked, pseudogene 2', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69878345, end_position_on_the_genomic_accession=69878744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5616, CurrentID=0, Status='live', Symbol='PRKY', Aliases='PRKXP3P, PRKY', description='protein kinase, Y-linked, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7273972, end_position_on_the_genomic_accession=7381547, orientation='plus', exon_count=8, OMIM=400008, no_hdr0=''),
  5617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5617, CurrentID=0, Status='live', Symbol='PRL', Aliases='GHA1', description='prolactin', other_designations='decidual prolactin|growth hormone A1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22287244, end_position_on_the_genomic_accession=22302883, orientation='minus', exon_count=7, OMIM=176760, no_hdr0=''),
  5618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5618, CurrentID=0, Status='live', Symbol='PRLR', Aliases='HPRL, MFAB, hPRLrI', description='prolactin receptor', other_designations='hPRL receptor|secreted prolactin binding protein', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35048756, end_position_on_the_genomic_accession=35230589, orientation='minus', exon_count=15, OMIM=176761, no_hdr0=''),
  5619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5619, CurrentID=0, Status='live', Symbol='PRM1', Aliases='CT94.1, P1', description='protamine 1', other_designations='cancer/testis antigen family 94, member 1|cysteine-rich protamine|testis specific protamine 1', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11280836, end_position_on_the_genomic_accession=11281335, orientation='minus', exon_count=2, OMIM=182880, no_hdr0=''),
  5620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5620, CurrentID=0, Status='live', Symbol='PRM2', Aliases='CT94.2', description='protamine 2', other_designations='cancer/testis antigen family 94, member 2|sperm histone P2|sperm protamine P2', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11275636, end_position_on_the_genomic_accession=11276480, orientation='minus', exon_count=2, OMIM=182890, no_hdr0=''),
  5621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5621, CurrentID=0, Status='live', Symbol='PRNP', Aliases='ASCR, AltPrP, CD230, CJD, GSS, KURU, PRIP, PrP, PrP27-30, PrP33-35C, PrPc, p27-30', description='prion protein', other_designations='CD230 antigen|prion-related protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4686151, end_position_on_the_genomic_accession=4701588, orientation='plus', exon_count=2, OMIM=176640, no_hdr0=''),
  5623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5623, CurrentID=0, Status='live', Symbol='PSPN', Aliases='PSP', description='persephin', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6375294, end_position_on_the_genomic_accession=6375849, orientation='minus', exon_count=2, OMIM=602921, no_hdr0=''),
  5624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5624, CurrentID=0, Status='live', Symbol='PROC', Aliases='APC, PC1, THPH3, THPH4, PROC', description='protein C, inactivator of coagulation factors Va and VIIIa', other_designations='anticoagulant protein C|autoprothrombin IIA|blood coagulation factor XIV|prepro-protein C|protein C (inactivator of coagulation factors Va and VIIIa)|type I protein C', map_location='2q13-q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127418405, end_position_on_the_genomic_accession=127429246, orientation='plus', exon_count=8, OMIM=612283, no_hdr0=''),
  5625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5625, CurrentID=0, Status='live', Symbol='PRODH', Aliases='HSPOX2, PIG6, POX1, PRODH2, TP53I6, PRODH', description='proline dehydrogenase (oxidase) 1', other_designations='p53-induced gene 6 protein|proline oxidase 2|proline oxidase, mitochondrial|tumor protein p53 inducible protein 6', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18912695, end_position_on_the_genomic_accession=18936553, orientation='minus', exon_count=16, OMIM=606810, no_hdr0=''),
  5626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5626, CurrentID=0, Status='live', Symbol='PROP1', Aliases='CPHD2, PROP-1', description='PROP paired-like homeobox 1', other_designations='pituitary-specific homeodomain factor|prophet of Pit1, paired-like homeodomain transcription factor', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177992235, end_position_on_the_genomic_accession=177996242, orientation='minus', exon_count=3, OMIM=601538, no_hdr0=''),
  5627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5627, CurrentID=0, Status='live', Symbol='PROS1', Aliases='PROS, PS21, PS22, PS23, PS24, PS25, PSA, THPH5, THPH6', description='protein S (alpha)', other_designations='protein Sa|vitamin K-dependent plasma protein S', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=93873037, end_position_on_the_genomic_accession=93974090, orientation='minus', exon_count=16, OMIM=176880, no_hdr0=''),
  5628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5628, CurrentID=0, Status='live', Symbol='PROS2P', Aliases='PROS2, PROSP', description='protein S (beta) pseudogene', other_designations='PS beta|protein S 2|protein S pseudogene (beta)', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=90201106, end_position_on_the_genomic_accession=90242064, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5629, CurrentID=0, Status='live', Symbol='PROX1', Aliases='', description='prospero homeobox 1', other_designations='homeobox prospero-like protein PROX1|prospero-related homeobox 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213983226, end_position_on_the_genomic_accession=214041510, orientation='plus', exon_count=9, OMIM=601546, no_hdr0=''),
  5630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5630, CurrentID=0, Status='live', Symbol='PRPH', Aliases='NEF41, PRPH', description='peripherin', other_designations='neurofilament 4 (57kD)', map_location='12q12-q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49295013, end_position_on_the_genomic_accession=49298698, orientation='plus', exon_count=9, OMIM=170710, no_hdr0=''),
  5631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5631, CurrentID=0, Status='live', Symbol='PRPS1', Aliases='ARTS, CMTX5, DFN2, DFNX1, PPRibP, PRS-I, PRSI', description='phosphoribosyl pyrophosphate synthetase 1', other_designations='dJ1070B1.2 (phosphoribosyl pyrophosphate synthetase 1)|deafness 2, perceptive, congenital|deafness, X-linked 2, perceptive, congenital|phosphoribosyl pyrophosphate synthase I|ribose-phosphate diphosphokinase 1', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107628424, end_position_on_the_genomic_accession=107651026, orientation='plus', exon_count=7, OMIM=311850, no_hdr0=''),
  5634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5634, CurrentID=0, Status='live', Symbol='PRPS2', Aliases='PRSII', description='phosphoribosyl pyrophosphate synthetase 2', other_designations='PPRibP synthetase|PRS-II|phosphoribosyl pyrophosphate synthase II|ribose-phosphate diphosphokinase 2', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12791355, end_position_on_the_genomic_accession=12824227, orientation='plus', exon_count=7, OMIM=311860, no_hdr0=''),
  5635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5635, CurrentID=0, Status='live', Symbol='PRPSAP1', Aliases='PAP39', description='phosphoribosyl pyrophosphate synthetase-associated protein 1', other_designations='39 kDa phosphoribosypyrophosphate synthase-associated protein|PRPP synthase-associated protein 1', map_location='17q24-q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76310787, end_position_on_the_genomic_accession=76354149, orientation='minus', exon_count=11, OMIM=601249, no_hdr0=''),
  5636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5636, CurrentID=0, Status='live', Symbol='PRPSAP2', Aliases='PAP41', description='phosphoribosyl pyrophosphate synthetase-associated protein 2', other_designations='41 kDa phosphoribosypyrophosphate synthetase-associated protein|PRPP synthase-associated protein 2', map_location='17p11.2-p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18856299, end_position_on_the_genomic_accession=18931287, orientation='plus', exon_count=14, OMIM=603762, no_hdr0=''),
  5638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5638, CurrentID=0, Status='live', Symbol='PRRG1', Aliases='PRGP1', description='proline rich Gla (G-carboxyglutamic acid) 1', other_designations='proline-rich Gla (G-carboxyglutamic acid) polypeptide 1', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37349275, end_position_on_the_genomic_accession=37457295, orientation='plus', exon_count=6, OMIM=300935, no_hdr0=''),
  5639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5639, CurrentID=0, Status='live', Symbol='PRRG2', Aliases='PRGP2', description='proline rich Gla (G-carboxyglutamic acid) 2', other_designations='proline-rich Gla (G-carboxglutamic acid) polypeptide 2|proline-rich Gla (G-carboxyglutamic acid) polypeptide 2|proline-rich Gla protein 2|proline-rich gamma-carboxyglutamic acid protein 2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49580617, end_position_on_the_genomic_accession=49591008, orientation='plus', exon_count=8, OMIM=604429, no_hdr0=''),
  5640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5640, CurrentID=0, Status='live', Symbol='PRS', Aliases='MRXS2', description='Prieto X-linked mental retardation syndrome', other_designations='Prieto syndrome (mental retardation, X-linked, syndromic-2, with dysmorphism and cerebral atrophy)', map_location='Xp21-p11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309610, no_hdr0=''),
  5641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5641, CurrentID=0, Status='live', Symbol='LGMN', Aliases='AEP1, PRSC1, LGMN', description='legumain', other_designations='asparaginyl endopeptidase|cysteine protease 1|protease, cysteine 1|protease, cysteine, 1 (legumain)', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92703807, end_position_on_the_genomic_accession=92748702, orientation='minus', exon_count=15, OMIM=602620, no_hdr0=''),
  5643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5643, CurrentID=0, Status='live', Symbol='PRSM2', Aliases='', description='protease, metallo, 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5644, CurrentID=0, Status='live', Symbol='PRSS1', Aliases='TRP1, TRY1, TRY4, TRYP1', description='protease, serine 1', other_designations='TCR V beta 4.1|beta-trypsin|cationic trypsinogen|digestive zymogen|nonfunctional trypsin 1|trypsinogen 1|trypsinogen A', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142740235, end_position_on_the_genomic_accession=142753076, orientation='plus', exon_count=7, OMIM=276000, no_hdr0=''),
  5645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5645, CurrentID=0, Status='live', Symbol='PRSS2', Aliases='TRY2, TRY8, TRYP2', description='protease, serine 2', other_designations='anionic trypsinogen|protease serine 2 preproprotein|protease, serine, 2 (trypsin 2)|trypsin II|trypsinogen 2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142770943, end_position_on_the_genomic_accession=142774564, orientation='plus', exon_count=6, OMIM=601564, no_hdr0=''),
  5646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5646, CurrentID=0, Status='live', Symbol='PRSS3', Aliases='MTG, PRSS4, T9, TRY3, TRY4', description='protease, serine 3', other_designations='brain trypsinogen|mesotrypsin|mesotrypsinogen|pancreatic trypsinogen III|protease, serine, 4 (trypsin 4, brain)|trypsin III|trypsin IV|trypsinogen 4|trypsinogen 5|trypsinogen IV', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33750466, end_position_on_the_genomic_accession=33799231, orientation='plus', exon_count=9, OMIM=613578, no_hdr0=''),
  100271021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271021, CurrentID=0, Status='live', Symbol='RPL37P16', Aliases='RPL37_6_843', description='ribosomal protein L37 pseudogene 16', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128433422, end_position_on_the_genomic_accession=128433713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5648, CurrentID=0, Status='live', Symbol='MASP1', Aliases='3MC1, CRARF, CRARF1, MAP1, MASP, MASP3, MAp44, PRSS5, RaRF', description='mannan-binding lectin serine peptidase 1 (C4/C2 activating component of Ra-reactive factor)', other_designations='Ra-reactive factor serine protease p100|complement factor MASP-3|complement-activating component of Ra-reactive factor|mannose-binding lectin-associated serine protease 1|mannose-binding protein-associated serine protease|serine protease 5', map_location='3q27-q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187216085, end_position_on_the_genomic_accession=187292022, orientation='minus', exon_count=19, OMIM=600521, no_hdr0=''),
  5649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5649, CurrentID=0, Status='live', Symbol='RELN', Aliases='ETL7, LIS2, PRO1598, RL', description='reelin', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103471784, end_position_on_the_genomic_accession=103989516, orientation='minus', exon_count=65, OMIM=600514, no_hdr0=''),
  5650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5650, CurrentID=0, Status='live', Symbol='KLK7', Aliases='PRSS6, SCCE, hK7', description='kallikrein related peptidase 7', other_designations='kallikrein 7 (chymotryptic, stratum corneum)|kallikrein-related peptidase 7|serine protease 6|signal protein|stratum corneum chymotryptic enzyme', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50976479, end_position_on_the_genomic_accession=50984064, orientation='minus', exon_count=7, OMIM=604438, no_hdr0=''),
  5651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5651, CurrentID=0, Status='live', Symbol='TMPRSS15', Aliases='ENTK, PRSS7', description='transmembrane protease, serine 15', other_designations='enterokinase catalytic subunit|protease, serine, 7 (enterokinase)|serine protease 7', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=18269107, end_position_on_the_genomic_accession=18403806, orientation='minus', exon_count=28, OMIM=606635, no_hdr0=''),
  5652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5652, CurrentID=0, Status='live', Symbol='PRSS8', Aliases='CAP1, PROSTASIN', description='protease, serine 8', other_designations='channel-activating protease 1|serine protease 8', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31131433, end_position_on_the_genomic_accession=31135830, orientation='minus', exon_count=6, OMIM=600823, no_hdr0=''),
  5653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5653, CurrentID=0, Status='live', Symbol='KLK6', Aliases='Bssp, Klk7, PRSS18, PRSS9, SP59, hK6', description='kallikrein related peptidase 6', other_designations='kallikrein-related peptidase 6|neurosin|protease M|serine protease 18|serine protease 9|zyme', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50958631, end_position_on_the_genomic_accession=50969673, orientation='minus', exon_count=7, OMIM=602652, no_hdr0=''),
  5654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5654, CurrentID=0, Status='live', Symbol='HTRA1', Aliases='ARMD7, CARASIL, HtrA, L56, ORF480, PRSS11', description='HtrA serine peptidase 1', other_designations='IGFBP5-protease|high-temperature requirement A serine peptidase 1|protease, serine, 11 (IGF binding)', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122461525, end_position_on_the_genomic_accession=122514908, orientation='plus', exon_count=9, OMIM=602194, no_hdr0=''),
  5655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5655, CurrentID=0, Status='live', Symbol='KLK10', Aliases='NES1, PRSSL1', description='kallikrein related peptidase 10', other_designations='breast normal epithelial cell associated serine protease|kallikrein 10 protein 1|kallikrein 10 protein 12|kallikrein 10 protein 13|kallikrein 10 protein 2|kallikrein 10 protein 3|kallikrein 10 protein 4|kallikrein 10 protein 5|kallikrein 10 protein 7|kallikrein 10 protein 8|kallikrein 10 protein 9|kallikrein-related peptidase 10|normal epithelial cell-specific 1|protease serine-like 1|protease, serine-like, 1', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51012744, end_position_on_the_genomic_accession=51020175, orientation='minus', exon_count=9, OMIM=602673, no_hdr0=''),
  5657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5657, CurrentID=0, Status='live', Symbol='PRTN3', Aliases='ACPA, AGP7, C-ANCA, CANCA, MBN, MBT, NP-4, NP4, P29, PR-3, PR3', description='proteinase 3', other_designations='C-ANCA antigen|Wegener granulomatosis autoantigen|azurophil granule protein 7|leukocyte proteinase 3|neutrophil proteinase 4|serine proteinase, neutrophil|wegener autoantigen', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=840965, end_position_on_the_genomic_accession=848175, orientation='plus', exon_count=5, OMIM=177020, no_hdr0=''),
  100271023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271023, CurrentID=0, Status='live', Symbol='RPL39P16', Aliases='RPL39_5_258', description='ribosomal protein L39 pseudogene 16', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108878308, end_position_on_the_genomic_accession=108878463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5660, CurrentID=0, Status='live', Symbol='PSAP', Aliases='GLBA, SAP1', description='prosaposin', other_designations='proactivator polypeptide|sphingolipid activator protein-1', map_location='10q21-q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71816298, end_position_on_the_genomic_accession=71851325, orientation='minus', exon_count=15, OMIM=176801, no_hdr0=''),
  5661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5661, CurrentID=0, Status='live', Symbol='PSBP1', Aliases='', description='prostatic steroid binding protein 1', other_designations='prostatic steroid-binding protein 1', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5662, CurrentID=0, Status='live', Symbol='PSD', Aliases='EFA6, EFA6A1, TYL, PSD', description='pleckstrin and Sec7 domain containing', other_designations='exchange factor for ARF6|pleckstrin homology and SEC7 domain-containing protein 1|pleckstrin-Sec7 domains-containing', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102402617, end_position_on_the_genomic_accession=102420161, orientation='minus', exon_count=21, OMIM=602327, no_hdr0=''),
  5663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5663, CurrentID=0, Status='live', Symbol='PSEN1', Aliases='AD3, FAD, PS-1, PS1, S182', description='presenilin 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73136435, end_position_on_the_genomic_accession=73223691, orientation='plus', exon_count=14, OMIM=104311, no_hdr0=''),
  5664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5664, CurrentID=0, Status='live', Symbol='PSEN2', Aliases='AD3L, AD4, CMD1V, PS2, STM2', description='presenilin 2', other_designations='AD3LP|AD5|Alzheimer disease 4|E5-1|PS-2|STM-2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226870572, end_position_on_the_genomic_accession=226896103, orientation='plus', exon_count=15, OMIM=600759, no_hdr0=''),
  100271024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271024, CurrentID=0, Status='live', Symbol='RPL39P17', Aliases='RPL39_6_343', description='ribosomal protein L39 pseudogene 17', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13052774, end_position_on_the_genomic_accession=13053182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5669, CurrentID=0, Status='live', Symbol='PSG1', Aliases='B1G1, CD66f, DHFRP2, FL-NCA-1/2, PBG1, PS-beta-C/D, PS-beta-G-1, PSBG-1, PSBG1, PSG95, PSGGA, PSGIIA, SP1', description='pregnancy specific beta-1-glycoprotein 1', other_designations='CD66 antigen-like family member F|fetal liver non-specific cross-reactive antigen 1/2|pregnancy-specific B-1 glycoprotein|pregnancy-specific beta-1 glycoprotein C/D', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42866461, end_position_on_the_genomic_accession=42879719, orientation='minus', exon_count=7, OMIM=176390, no_hdr0=''),
  5670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5670, CurrentID=0, Status='live', Symbol='PSG2', Aliases='CEA, PSBG2, PSG1', description='pregnancy specific beta-1-glycoprotein 2', other_designations='PS-beta-E|PS-beta-G-2|PSBG-2|carcinoembryonic antigen SG8|pregnancy-specific beta-1 glycoprotein E|pregnancy-specific beta-1-glycoprotein 7|pregnancy-specific beta-1-glycoprotein-2|pregnancy-specific glycoprotein 2', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43064210, end_position_on_the_genomic_accession=43082741, orientation='minus', exon_count=6, OMIM=176391, no_hdr0=''),
  5671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5671, CurrentID=0, Status='live', Symbol='PSG3', Aliases='', description='pregnancy specific beta-1-glycoprotein 3', other_designations='PS-beta-G-3|PSBG-3|carcinoembryonic antigen SG5|pregnancy-specific glycoprotein 3', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42721642, end_position_on_the_genomic_accession=42740516, orientation='minus', exon_count=7, OMIM=176392, no_hdr0=''),
  5672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5672, CurrentID=0, Status='live', Symbol='PSG4', Aliases='PSBG-4, PSBG-9, PSG9', description='pregnancy specific beta-1-glycoprotein 4', other_designations='PS-beta-G-4|PS-beta-G-9|Pregnancy-specific beta-1-glycoprotein-4|pregnancy-specific beta-1-glycoprotein 9|pregnancy-specific glycoprotein 4|pregnancy-specific glycoprotein 9', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43192702, end_position_on_the_genomic_accession=43205774, orientation='minus', exon_count=7, OMIM=176393, no_hdr0=''),
  5673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5673, CurrentID=0, Status='live', Symbol='PSG5', Aliases='FL-NCA-3, PSG', description='pregnancy specific beta-1-glycoprotein 5', other_designations='PS-beta-G-5|PSBG-5|Pregnancy-specific beta-1-glycoprotein-5|fetal liver non-specific cross-reactive antigen 3|pregnancy-specific beta 1 glycoprotein|pregnancy-specific beta-1 glycoprotein|pregnancy-specific glycoprotein 5', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43032810, end_position_on_the_genomic_accession=43186536, orientation='minus', exon_count=8, OMIM=176394, no_hdr0=''),
  5675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5675, CurrentID=0, Status='live', Symbol='PSG6', Aliases='PSBG-10, PSBG-12, PSBG-6, PSG10, PSGGB', description='pregnancy specific beta-1-glycoprotein 6', other_designations='PS-beta-G-10|PS-beta-G-12|PS-beta-G-6|pregnancy-specific beta-1-glycoprotein 10|pregnancy-specific beta-1-glycoprotein 12', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42902082, end_position_on_the_genomic_accession=42917924, orientation='minus', exon_count=7, OMIM=176395, no_hdr0=''),
  5676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5676, CurrentID=0, Status='live', Symbol='PSG7', Aliases='PSBG-7, PSG1, PSGGA', description='pregnancy specific beta-1-glycoprotein 7 (gene/pseudogene)', other_designations='PS-beta-G-7|pregnancy-specific glycoprotein 7', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42924132, end_position_on_the_genomic_accession=42937178, orientation='minus', exon_count=6, OMIM=176396, no_hdr0=''),
  5678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5678, CurrentID=0, Status='live', Symbol='PSG9', Aliases='PS34, PSBG-11, PSBG-9, PSG11, PSGII', description='pregnancy specific beta-1-glycoprotein 9', other_designations='PS-beta-B|PS-beta-G-11|PS-beta-G-9|pregnancy specific beta-1-glycoprotein 11|pregnancy-specific beta-1 glycoprotein B', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43253282, end_position_on_the_genomic_accession=43269563, orientation='minus', exon_count=6, OMIM=176398, no_hdr0=''),
  5680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5680, CurrentID=0, Status='live', Symbol='PSG11', Aliases='PSBG-11, PSBG-13, PSG13, PSG14', description='pregnancy specific beta-1-glycoprotein 11', other_designations='PS-beta-G-11|PS-beta-G-13|pregnancy-specific beta-1-glycoprotein 13', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43007657, end_position_on_the_genomic_accession=43026479, orientation='minus', exon_count=6, OMIM=176401, no_hdr0=''),
  5681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5681, CurrentID=0, Status='live', Symbol='PSKH1', Aliases='', description='protein serine kinase H1', other_designations='PSK-H1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67893272, end_position_on_the_genomic_accession=67929678, orientation='plus', exon_count=3, OMIM=177015, no_hdr0=''),
  5682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5682, CurrentID=0, Status='live', Symbol='PSMA1', Aliases='HC2, HEL-S-275, NU, PROS30', description='proteasome subunit alpha 1', other_designations='30 kDa prosomal protein|PROS-30|epididymis secretory protein Li 275|macropain subunit C2|macropain subunit nu|multicatalytic endopeptidase complex subunit C2|proteasome (prosome, macropain) subunit, alpha type, 1|proteasome component C2|proteasome nu chain|proteasome subunit nu|proteasome subunit, alpha-type, 1|protein P30-33K', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14504876, end_position_on_the_genomic_accession=14643634, orientation='minus', exon_count=12, OMIM=602854, no_hdr0=''),
  5683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5683, CurrentID=0, Status='live', Symbol='PSMA2', Aliases='HC3, MU, PMSA2, PSC2', description='proteasome subunit alpha 2', other_designations='macropain subunit C3|multicatalytic endopeptidase complex subunit C3|proteasome (prosome, macropain) subunit, alpha type, 2|proteasome component C3|proteasome subunit HC3', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42916861, end_position_on_the_genomic_accession=42932206, orientation='minus', exon_count=8, OMIM=176842, no_hdr0=''),
  5684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5684, CurrentID=0, Status='live', Symbol='PSMA3', Aliases='HC8, PSC3', description='proteasome subunit alpha 3', other_designations='macropain subunit C8|multicatalytic endopeptidase complex subunit C8|proteasome (prosome, macropain) subunit, alpha type, 3|proteasome component C8|proteasome subunit C8', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58244805, end_position_on_the_genomic_accession=58272009, orientation='plus', exon_count=11, OMIM=176843, no_hdr0=''),
  5685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5685, CurrentID=0, Status='live', Symbol='PSMA4', Aliases='HC9, HsT17706, PSC9', description='proteasome subunit alpha 4', other_designations='macropain subunit C9|multicatalytic endopeptidase complex subunit C9|proteasome (prosome, macropain) subunit, alpha type, 4|proteasome component C9|proteasome subunit HC9|proteasome subunit L', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78540405, end_position_on_the_genomic_accession=78549220, orientation='plus', exon_count=9, OMIM=176846, no_hdr0=''),
  5686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5686, CurrentID=0, Status='live', Symbol='PSMA5', Aliases='PSC5, ZETA', description='proteasome subunit alpha 5', other_designations='macropain subunit zeta|macropain zeta chain|multicatalytic endopeptidase complex zeta chain|proteasome (prosome, macropain) subunit, alpha type, 5|proteasome alpha 5 subunit|proteasome component 5|proteasome subunit zeta|proteasome zeta chain', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109399031, end_position_on_the_genomic_accession=109426486, orientation='minus', exon_count=9, OMIM=176844, no_hdr0=''),
  5687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5687, CurrentID=0, Status='live', Symbol='PSMA6', Aliases='IOTA, PROS27, p27K', description='proteasome subunit alpha 6', other_designations='27 kDa prosomal protein|PROS-27|macropain iota chain|macropain subunit iota|multicatalytic endopeptidase complex iota chain|prosomal P27K protein|proteasome (prosome, macropain) subunit, alpha type, 6|proteasome iota chain|proteasome subunit iota', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35278558, end_position_on_the_genomic_accession=35317479, orientation='plus', exon_count=8, OMIM=602855, no_hdr0=''),
  5688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5688, CurrentID=0, Status='live', Symbol='PSMA7', Aliases='C6, HEL-S-276, HSPC, RC6-1, XAPC7', description='proteasome subunit alpha 7', other_designations='epididymis secretory protein Li 276|proteasome (prosome, macropain) subunit, alpha type, 7|proteasome subunit RC6-1|proteasome subunit XAPC7|proteasome subunit alpha 4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62136727, end_position_on_the_genomic_accession=62143458, orientation='minus', exon_count=7, OMIM=606607, no_hdr0=''),
  5689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5689, CurrentID=0, Status='live', Symbol='PSMB1', Aliases='HC5, PMSB1, PSC5', description='proteasome subunit beta 1', other_designations='macropain subunit C5|multicatalytic endopeptidase complex subunit C5|proteasome (prosome, macropain) subunit, beta type, 1|proteasome beta 1 subunit|proteasome component C5|proteasome gamma chain|proteasome subunit HC5', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170535116, end_position_on_the_genomic_accession=170553329, orientation='minus', exon_count=6, OMIM=602017, no_hdr0=''),
  5690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5690, CurrentID=0, Status='live', Symbol='PSMB2', Aliases='HC7-I', description='proteasome subunit beta 2', other_designations='macropain subunit C7-I|multicatalytic endopeptidase complex subunit C7-1|multicatalytic endopeptidase complex subunit C7-I|proteasome (prosome, macropain) subunit, beta type, 2|proteasome beta 2 subunit|proteasome component C7-I|proteasome subunit, beta type, 2', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35599541, end_position_on_the_genomic_accession=35641844, orientation='minus', exon_count=7, OMIM=602175, no_hdr0=''),
  5691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5691, CurrentID=0, Status='live', Symbol='PSMB3', Aliases='HC10-II', description='proteasome subunit beta 3', other_designations='proteasome (prosome, macropain) subunit, beta type, 3|proteasome chain 13|proteasome component C10-II|proteasome theta chain', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38752713, end_position_on_the_genomic_accession=38764231, orientation='plus', exon_count=6, OMIM=602176, no_hdr0=''),
  5692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5692, CurrentID=0, Status='live', Symbol='PSMB4', Aliases='HN3, HsN3, PROS-26, PROS26', description='proteasome subunit beta 4', other_designations='26 kDa prosomal protein|hsBPROS26|macropain beta chain|multicatalytic endopeptidase complex beta chain|proteasome (prosome, macropain) subunit, beta type, 4|proteasome beta chain|proteasome chain 3|proteasome subunit HsN3|proteasome subunit, beta type, 4', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151399565, end_position_on_the_genomic_accession=151401936, orientation='plus', exon_count=7, OMIM=602177, no_hdr0=''),
  5693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5693, CurrentID=0, Status='live', Symbol='PSMB5', Aliases='LMPX, MB1, X', description='proteasome subunit beta 5', other_designations='PSX large multifunctional protease X|macropain epsilon chain|multicatalytic endopeptidase complex epsilon chain|proteasome (prosome, macropain) subunit, beta type, 5|proteasome beta 5 subunit|proteasome catalytic subunit 3|proteasome chain 6|proteasome epsilon chain|proteasome subunit MB1|proteasome subunit X|proteasome subunit, beta type, 5', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23016543, end_position_on_the_genomic_accession=23035220, orientation='minus', exon_count=5, OMIM=600306, no_hdr0=''),
  5694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5694, CurrentID=0, Status='live', Symbol='PSMB6', Aliases='DELTA, LMPY, Y', description='proteasome subunit beta 6', other_designations='PSY large multifunctional protease Y|macropain delta chain|multicatalytic endopeptidase complex delta chain|proteasome (prosome, macropain) subunit, beta type, 6|proteasome catalytic subunit 1|proteasome delta chain|proteasome subunit Y|proteasome subunit delta', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4796144, end_position_on_the_genomic_accession=4798503, orientation='plus', exon_count=6, OMIM=600307, no_hdr0=''),
  5695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5695, CurrentID=0, Status='live', Symbol='PSMB7', Aliases='Z', description='proteasome subunit beta 7', other_designations='macropain chain Z|multicatalytic endopeptidase complex chain Z|proteasome (prosome, macropain) subunit, beta type, 7|proteasome catalytic subunit 2|proteasome subunit Z|proteasome subunit alpha|protein serine kinase c17', map_location='9q34.11-q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124353465, end_position_on_the_genomic_accession=124415473, orientation='minus', exon_count=8, OMIM=604030, no_hdr0=''),
  5696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5696, CurrentID=0, Status='live', Symbol='PSMB8', Aliases='ALDD, D6S216, D6S216E, JMP, LMP7, NKJO, PSMB5i, RING10', description='proteasome subunit beta 8', other_designations='low molecular mass protein 7|low molecular weight protein 7|macropain subunit C13|multicatalytic endopeptidase complex subunit C13|protease component C13|proteasome (prosome, macropain) subunit, beta type, 8 (large multifunctional peptidase 7)|proteasome catalytic subunit 3i|proteasome component C13|proteasome subunit Y2|proteasome subunit beta 5i|proteasome-related gene 7|really interesting new gene 10 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32840717, end_position_on_the_genomic_accession=32844935, orientation='minus', exon_count=7, OMIM=177046, no_hdr0=''),
  5697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5697, CurrentID=0, Status='live', Symbol='PYY', Aliases='PYY-I1, PYY', description='peptide YY', other_designations='peptide tyrosine tyrosine|prepro-PYY', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43952733, end_position_on_the_genomic_accession=44024260, orientation='minus', exon_count=8, OMIM=600781, no_hdr0=''),
  5698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5698, CurrentID=0, Status='live', Symbol='PSMB9', Aliases='LMP2, PSMB6i, RING12, beta1i', description='proteasome subunit beta 9', other_designations='large multifunctional peptidase 2|low molecular mass protein 2|macropain chain 7|multicatalytic endopeptidase complex chain 7|proteasome (prosome, macropain) subunit, beta type, 9 (large multifunctional peptidase 2)|proteasome catalytic subunit 1i|proteasome chain 7|proteasome subunit beta 6i|proteasome subunit beta-1i|proteasome-related gene 2|really interesting new gene 12 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32854161, end_position_on_the_genomic_accession=32859851, orientation='plus', exon_count=6, OMIM=177045, no_hdr0=''),
  5699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5699, CurrentID=0, Status='live', Symbol='PSMB10', Aliases='LMP10, MECL1, beta2i', description='proteasome subunit beta 10', other_designations='low molecular mass protein 10|macropain subunit MECl-1|multicatalytic endopeptidase complex subunit MECl-1|proteasome (prosome, macropain) subunit, beta type, 10|proteasome MECl-1|proteasome catalytic subunit 2i|proteasome subunit MECL1|proteasome subunit beta 7i|proteasome subunit beta-2i', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67934504, end_position_on_the_genomic_accession=67936877, orientation='minus', exon_count=8, OMIM=176847, no_hdr0=''),
  5700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5700, CurrentID=0, Status='live', Symbol='PSMC1', Aliases='P26S4, S4, p56', description='proteasome 26S subunit, ATPase 1', other_designations='26S proteasome AAA-ATPase subunit RPT2|proteasome (prosome, macropain) 26S subunit, ATPase, 1|proteasome 26S ATPase subunit 1|proteasome 26S subunit ATPase 1|proteasome 26S subunit, ATPase, 1', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90256550, end_position_on_the_genomic_accession=90272622, orientation='plus', exon_count=11, OMIM=602706, no_hdr0=''),
  5701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5701, CurrentID=0, Status='live', Symbol='PSMC2', Aliases='MSS1, Nbla10058, S7', description='proteasome 26S subunit, ATPase 2', other_designations='26S proteasome AAA-ATPase subunit RPT1|mammalian suppressor of sgv-1 of yeast|protease 26S subunit 7|proteasome (prosome, macropain) 26S subunit, ATPase, 2|proteasome 26S subunit ATPase 2|proteasome 26S subunit, ATPase, 2|putative protein product of Nbla10058', map_location='7q22.1-q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103344405, end_position_on_the_genomic_accession=103369395, orientation='plus', exon_count=13, OMIM=154365, no_hdr0=''),
  5702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5702, CurrentID=0, Status='live', Symbol='PSMC3', Aliases='TBP1', description='proteasome 26S subunit, ATPase 3', other_designations='26S proteasome AAA-ATPase subunit RPT5|TBP-1|Tat-binding protein 1|human immunodeficiency virus tat transactivator binding protein-1|proteasome (prosome, macropain) 26S subunit, ATPase, 3|proteasome 26S subunit ATPase 3|proteasome subunit P50', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47418769, end_position_on_the_genomic_accession=47426473, orientation='minus', exon_count=12, OMIM=186852, no_hdr0=''),
  5703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5703, CurrentID=0, Status='live', Symbol='PSMC3P', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 3 pseudogene', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19026887, end_position_on_the_genomic_accession=19028191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5704, CurrentID=0, Status='live', Symbol='PSMC4', Aliases='MIP224, RPT3, S6, TBP-7, TBP7', description='proteasome 26S subunit, ATPase 4', other_designations='26S proteasome AAA-ATPase subunit RPT3|MB67-interacting protein|Tat-binding protein 7|protease 26S subunit 6|proteasome (prosome, macropain) 26S subunit, ATPase, 4', map_location='19q13.11-q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39971005, end_position_on_the_genomic_accession=39981764, orientation='plus', exon_count=11, OMIM=602707, no_hdr0=''),
  5705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5705, CurrentID=0, Status='live', Symbol='PSMC5', Aliases='S8, SUG-1, SUG1, TBP10, TRIP1, p45, p45/SUG', description='proteasome 26S subunit, ATPase 5', other_designations='26S proteasome AAA-ATPase subunit RPT6|MSUG1 protein|Tat-binding protein homolog 10|proteasome (prosome, macropain) 26S subunit, ATPase, 5|proteasome 26S ATPase subunit 5|proteasome 26S subunit ATPase 5|proteasome subunit p45|thyroid hormone receptor-interacting protein 1|thyroid receptor interactor 1', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63827410, end_position_on_the_genomic_accession=63832027, orientation='plus', exon_count=13, OMIM=601681, no_hdr0=''),
  5706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5706, CurrentID=0, Status='live', Symbol='PSMC6', Aliases='CADP44, HEL-S-73, P44, SUG2, p42', description='proteasome 26S subunit, ATPase 6', other_designations='26S protease regulatory subunit S10B|26S proteasome AAA-ATPase subunit RPT4|conserved ATPase domain protein 44|epididymis secretory protein Li 73|proteasome (prosome, macropain) 26S subunit, ATPase, 6|proteasome 26S subunit ATPase 6|proteasome subunit p42', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52707163, end_position_on_the_genomic_accession=52727998, orientation='plus', exon_count=15, OMIM=602708, no_hdr0=''),
  5707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5707, CurrentID=0, Status='live', Symbol='PSMD1', Aliases='P112, Rpn2, S1', description='proteasome 26S subunit, non-ATPase 1', other_designations='26S proteasome regulatory subunit RPN2|26S proteasome regulatory subunit S1|26S proteasome subunit p112|proteasome (prosome, macropain) 26S subunit, non-ATPase, 1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231056864, end_position_on_the_genomic_accession=231172827, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  5708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5708, CurrentID=0, Status='live', Symbol='PSMD2', Aliases='P97, RPN1, S2, TRAP2', description='proteasome 26S subunit, non-ATPase 2', other_designations='55.11 protein|TNFR-associated protein 2|proteasome (prosome, macropain) 26S subunit, non-ATPase, 2|protein 55.11|tumor necrosis factor type 1 receptor-associated protein 2', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184299098, end_position_on_the_genomic_accession=184309054, orientation='plus', exon_count=22, OMIM=606223, no_hdr0=''),
  5709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5709, CurrentID=0, Status='live', Symbol='PSMD3', Aliases='P58, RPN3, S3, TSTA2', description='proteasome 26S subunit, non-ATPase 3', other_designations='26S proteasome regulatory subunit RPN3|26S proteasome regulatory subunit S3|proteasome (prosome, macropain) 26S subunit, non-ATPase, 3|proteasome subunit p58|tissue specific transplantation antigen 2', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39980768, end_position_on_the_genomic_accession=39997960, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  5710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5710, CurrentID=0, Status='live', Symbol='PSMD4', Aliases='AF, AF-1, ASF, MCB1, Rpn10, S5A, pUB-R5', description='proteasome 26S subunit, non-ATPase 4', other_designations='26S proteasome regulatory subunit S5A|S5a/antisecretory factor protein|angiocidin|antisecretory factor 1|multiubiquitin chain-binding protein|proteasome (prosome, macropain) 26S subunit, non-ATPase, 4', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151254700, end_position_on_the_genomic_accession=151267479, orientation='plus', exon_count=10, OMIM=601648, no_hdr0=''),
  5711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5711, CurrentID=0, Status='live', Symbol='PSMD5', Aliases='S5B', description='proteasome 26S subunit, non-ATPase 5', other_designations='26S protease subunit S5 basic|26S proteasome subunit S5B|proteasome (prosome, macropain) 26S subunit, non-ATPase, 5', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120814779, end_position_on_the_genomic_accession=120843021, orientation='minus', exon_count=11, OMIM=604452, no_hdr0=''),
  5713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5713, CurrentID=0, Status='live', Symbol='PSMD7', Aliases='MOV34, P40, Rpn8, S12', description='proteasome 26S subunit, non-ATPase 7', other_designations='26S proteasome regulatory subunit S12|26S proteasome regulatory subunit rpn8|Moloney leukemia virus-34 proviral integration|Mov34 homolog|proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 (Mov34 homolog)|proteasome subunit p40', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74296775, end_position_on_the_genomic_accession=74306288, orientation='plus', exon_count=7, OMIM=157970, no_hdr0=''),
  5714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5714, CurrentID=0, Status='live', Symbol='PSMD8', Aliases='HEL-S-91n, HIP6, HYPF, Nin1p, Rpn12, S14, p31', description='proteasome 26S subunit, non-ATPase 8', other_designations='26S proteasome regulatory subunit RPN12|26S proteasome regulatory subunit S14|26S proteasome regulatory subunit p31|epididymis secretory sperm binding protein Li 91n|proteasome (prosome, macropain) 26S subunit, non-ATPase, 8', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38374550, end_position_on_the_genomic_accession=38383824, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5715, CurrentID=0, Status='live', Symbol='PSMD9', Aliases='Rpn4, p27', description='proteasome 26S subunit, non-ATPase 9', other_designations='26S proteasome regulatory subunit p27|homolog of rat Bridge 1|proteasome (prosome, macropain) 26S subunit, non-ATPase, 9|proteasome 26S non-ATPase regulatory subunit 9', map_location='12q24.31-q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121888731, end_position_on_the_genomic_accession=121917865, orientation='plus', exon_count=6, OMIM=603146, no_hdr0=''),
  5716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5716, CurrentID=0, Status='live', Symbol='PSMD10', Aliases='dJ889N15.2, p28, p28(GANK)', description='proteasome 26S subunit, non-ATPase 10', other_designations='26S proteasome regulatory subunit p28|ankyrin repeat protein|gankyrin|hepatocellular carcinoma-associated protein p28-II|proteasome (prosome, macropain) 26S subunit, non-ATPase, 10', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108084205, end_position_on_the_genomic_accession=108091644, orientation='minus', exon_count=5, OMIM=300880, no_hdr0=''),
  5717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5717, CurrentID=0, Status='live', Symbol='PSMD11', Aliases='Rpn6, S9, p44.5', description='proteasome 26S subunit, non-ATPase 11', other_designations='26S proteasome regulatory subunit 9|26S proteasome regulatory subunit RPN6|26S proteasome regulatory subunit S9|26S proteasome regulatory subunit p44.5|proteasome (prosome, macropain) 26S subunit, non-ATPase, 11', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32444463, end_position_on_the_genomic_accession=32483319, orientation='plus', exon_count=14, OMIM=604449, no_hdr0=''),
  5718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5718, CurrentID=0, Status='live', Symbol='PSMD12', Aliases='Rpn5, p55', description='proteasome 26S subunit, non-ATPase 12', other_designations='26S proteasome regulatory subunit RPN5|26S proteasome regulatory subunit p55|proteasome (prosome, macropain) 26S subunit, non-ATPase, 12|proteasome 26S subunit non-ATPase 12', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67340486, end_position_on_the_genomic_accession=67366605, orientation='minus', exon_count=13, OMIM=604450, no_hdr0=''),
  5719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5719, CurrentID=0, Status='live', Symbol='PSMD13', Aliases='HSPC027, Rpn9, S11, p40.5', description='proteasome 26S subunit, non-ATPase 13', other_designations='26S proteasome regulatory subunit RPN9|26S proteasome regulatory subunit S11|26S proteasome regulatory subunit p40.5|26S proteasome subunit p40.5|proteasome (prosome, macropain) 26S subunit, non-ATPase, 13', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=236808, end_position_on_the_genomic_accession=252984, orientation='plus', exon_count=12, OMIM=603481, no_hdr0=''),
  5720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5720, CurrentID=0, Status='live', Symbol='PSME1', Aliases='HEL-S-129m, IFI5111, PA28A, PA28alpha, REGalpha', description='proteasome activator subunit 1', other_designations='11S regulator complex subunit alpha|29-kD MCP activator subunit|IGUP I-5111|activator of multicatalytic protease subunit 1|epididymis secretory sperm binding protein Li 129m|interferon gamma up-regulated I-5111 protein|interferon-gamma IEF SSP 5111|interferon-gamma-inducible protein 5111|proteasome (prosome, macropain) activator subunit 1 (PA28 alpha)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24136158, end_position_on_the_genomic_accession=24138967, orientation='plus', exon_count=11, OMIM=600654, no_hdr0=''),
  5721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5721, CurrentID=0, Status='live', Symbol='PSME2', Aliases='PA28B, PA28beta, REGbeta', description='proteasome activator subunit 2', other_designations='11S regulator complex beta subunit|11S regulator complex subunit beta|MCP activator, 31-kD subunit|REG-beta|activator of multicatalytic protease subunit 2|cell migration-inducing protein 22|proteasome (prosome, macropain) activator subunit 2 (PA28 beta)|proteasome activator 28 subunit beta|proteasome activator 28-beta|proteasome activator hPA28 subunit beta', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24143365, end_position_on_the_genomic_accession=24147316, orientation='minus', exon_count=11, OMIM=602161, no_hdr0=''),
  5723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5723, CurrentID=0, Status='live', Symbol='PSPH', Aliases='PSPD, PSPH', description='phosphoserine phosphatase', other_designations='L-3-phosphoserine phosphatase|O-phosphoserine phosphohydrolase|PSPase', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56010363, end_position_on_the_genomic_accession=56051604, orientation='minus', exon_count=12, OMIM=172480, no_hdr0=''),
  5724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5724, CurrentID=0, Status='live', Symbol='PTAFR', Aliases='PAFR', description='platelet-activating factor receptor', other_designations='PAF-R', map_location='1p35-p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28147166, end_position_on_the_genomic_accession=28193936, orientation='minus', exon_count=4, OMIM=173393, no_hdr0=''),
  5725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5725, CurrentID=0, Status='live', Symbol='PTBP1', Aliases='HNRNP-I, HNRNPI, HNRPI, PTB, PTB-1, PTB-T, PTB2, PTB3, PTB4, pPTB', description='polypyrimidine tract binding protein 1', other_designations='57 kDa RNA-binding protein PPTB-1|RNA-binding protein|heterogeneous nuclear ribonucleoprotein I|heterogeneous nuclear ribonucleoprotein polypeptide I|hnRNP I|polypyrimidine tract binding protein (heterogeneous nuclear ribonucleoprotein I)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=797392, end_position_on_the_genomic_accession=812327, orientation='plus', exon_count=16, OMIM=600693, no_hdr0=''),
  5726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5726, CurrentID=0, Status='live', Symbol='TAS2R38', Aliases='PTC, T2R38, T2R61', description='taste 2 receptor member 38', other_designations='PTC bitter taste receptor|taste receptor type 2 member 61|taste receptor, type 2, member 38', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141972631, end_position_on_the_genomic_accession=141973773, orientation='minus', exon_count=1, OMIM=607751, no_hdr0=''),
  5727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5727, CurrentID=0, Status='live', Symbol='PTCH1', Aliases='BCNS, HPE7, NBCCS, PTC, PTC1, PTCH1, PTCH1', description='patched 1', other_designations="PTCH protein +12b|PTCH protein +4'|PTCH protein -10|PTCH protein -3,4,5", map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95442982, end_position_on_the_genomic_accession=95516965, orientation='minus', exon_count=33, OMIM=601309, no_hdr0=''),
  5728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5728, CurrentID=0, Status='live', Symbol='PTEN', Aliases='10q23del, BZS, CWS1, DEC, GLM2, MHAM, MMAC11, TEP1, PTEN', description='phosphatase and tensin homolog', other_designations='MMAC1 phosphatase and tensin homolog deleted on chromosome 10|mitochondrial PTENalpha|mitochondrial phosphatase and tensin protein alpha|mutated in multiple advanced cancers 1|phosphatase and tensin-like protein|phosphatidylinositol-3,4,5-trisphosphate 3-phosphatase and dual-specificity protein phosphatase PTEN', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87863438, end_position_on_the_genomic_accession=87971930, orientation='plus', exon_count=16, OMIM=601728, no_hdr0=''),
  5729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5729, CurrentID=0, Status='live', Symbol='PTGDR', Aliases='AS1, ASRT1, DP, DP11, PTGDR', description='prostaglandin D2 receptor (DP)', other_designations='PGD2 receptor', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52267256, end_position_on_the_genomic_accession=52276724, orientation='plus', exon_count=4, OMIM=604687, no_hdr0=''),
  5730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5730, CurrentID=0, Status='live', Symbol='PTGDS', Aliases='L-PGDS, LPGDS, PDS, PGD2, PGDS, PGDS2', description='prostaglandin D2 synthase', other_designations='PGD2 synthase|beta-trace protein|cerebrin-28|glutathione-independent PGD synthase|glutathione-independent PGD synthetase|lipocalin-type prostaglandin D synthase|lipocalin-type prostaglandin-D synthase|prostaglandin D synthase|prostaglandin D2 synthase 21kDa (brain)|prostaglandin-D2 synthase', map_location='9q34.2-q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136977504, end_position_on_the_genomic_accession=136981742, orientation='plus', exon_count=7, OMIM=176803, no_hdr0=''),
  5731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5731, CurrentID=0, Status='live', Symbol='PTGER1', Aliases='EP1', description='prostaglandin E receptor 1', other_designations='PGE receptor EP1 subtype|PGE receptor, EP1 subtype|PGE2 receptor EP1 subtype|prostaglandin E receptor 1 (subtype EP1), 42kD|prostaglandin E receptor 1 (subtype EP1), 42kDa|prostaglandin E receptor 1, subtype EP1|prostanoid EP1 receptor', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14472466, end_position_on_the_genomic_accession=14475362, orientation='minus', exon_count=3, OMIM=176802, no_hdr0=''),
  5732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5732, CurrentID=0, Status='live', Symbol='PTGER2', Aliases='EP2', description='prostaglandin E receptor 2', other_designations='PGE receptor EP2 subtype|PGE2 receptor EP2 subtype|prostaglandin E receptor 2 (subtype EP2), 53kD|prostaglandin E receptor 2 (subtype EP2), 53kDa|prostanoid EP2 receptor', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52314298, end_position_on_the_genomic_accession=52328606, orientation='plus', exon_count=2, OMIM=176804, no_hdr0=''),
  5733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5733, CurrentID=0, Status='live', Symbol='PTGER3', Aliases='EP3, EP3-I, EP3-II, EP3-III, EP3-IV, EP3-VI, EP3e, PGE2-R', description='prostaglandin E receptor 3', other_designations='PGE receptor, EP3 subtype|PGE2 receptor EP3 subtype|prostaglandin E receotor EP3 subtype 3 isoform|prostaglandin E receptor 3 (subtype EP3)|prostaglandin receptor (PGE-2)|prostanoid EP3 receptor', map_location='1p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70852353, end_position_on_the_genomic_accession=71047814, orientation='minus', exon_count=11, OMIM=176806, no_hdr0=''),
  5734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5734, CurrentID=0, Status='live', Symbol='PTGER4', Aliases='EP4, EP4R', description='prostaglandin E receptor 4', other_designations='PGE receptor EP4 subtype|PGE receptor, EP4 subtype|PGE2 receptor EP4 subtype|prostaglandin E receptor 4 (subtype EP4)|prostanoid EP4 receptor', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40679498, end_position_on_the_genomic_accession=40741764, orientation='plus', exon_count=7, OMIM=601586, no_hdr0=''),
  5735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5735, CurrentID=0, Status='live', Symbol='PTGER4P1', Aliases='', description='prostaglandin E receptor 4 pseudogene 1', other_designations='prostaglandin E receptor 4 (subtype EP4) pseudogene 1', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41279459, end_position_on_the_genomic_accession=41279908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5737, CurrentID=0, Status='live', Symbol='PTGFR', Aliases='FP', description='prostaglandin F receptor (FP)', other_designations='PGF receptor|PGF2 alpha receptor|PGF2-alpha receptor|prostaglandin F2 alpha receptor|prostaglandin receptor (2-alpha)|prostanoid FP receptor', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78476240, end_position_on_the_genomic_accession=78540701, orientation='plus', exon_count=7, OMIM=600563, no_hdr0=''),
  5738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5738, CurrentID=0, Status='live', Symbol='PTGFRN', Aliases='CD315, CD9P-1, EWI-F, FPRP, SMAP-6', description='prostaglandin F2 receptor inhibitor', other_designations='CD9 partner 1|glu-Trp-Ile EWI motif-containing protein F|prostaglandin F2-alpha receptor regulatory protein|prostaglandin F2-alpha receptor-associated protein', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116909923, end_position_on_the_genomic_accession=116990358, orientation='plus', exon_count=9, OMIM=601204, no_hdr0=''),
  5739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5739, CurrentID=0, Status='live', Symbol='PTGIR', Aliases='IP, PRIPR', description='prostaglandin I2 (prostacyclin) receptor (IP)', other_designations='PGI receptor|PGI2 receptor|prostaglandin I2 receptor|prostanoid IP receptor', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46610040, end_position_on_the_genomic_accession=46625118, orientation='minus', exon_count=6, OMIM=600022, no_hdr0=''),
  5740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5740, CurrentID=0, Status='live', Symbol='PTGIS', Aliases='CYP8, CYP8A1, PGIS, PTGI', description='prostaglandin I2 (prostacyclin) synthase', other_designations='cytochrome P450, family 8, subfamily A, polypeptide 1|prostaglandin I2 synthase', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49503874, end_position_on_the_genomic_accession=49568170, orientation='minus', exon_count=10, OMIM=601699, no_hdr0=''),
  5741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5741, CurrentID=0, Status='live', Symbol='PTH', Aliases='PTH1', description='parathyroid hormone', other_designations='parathormone|parathyrin|parathyroid hormone 1|prepro-PTH|preproparathyroid hormone', map_location='11p15.3-p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13492054, end_position_on_the_genomic_accession=13496020, orientation='minus', exon_count=3, OMIM=168450, no_hdr0=''),
  5742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5742, CurrentID=0, Status='live', Symbol='PTGS1', Aliases='COX1, COX3, PCOX1, PES-1, PGG/HS, PGHS-1, PGHS1, PHS1, PTGHS', description='prostaglandin-endoperoxide synthase 1 (prostaglandin G/H synthase and cyclooxygenase)', other_designations='PGH synthase 1|cyclooxygenase-1|prostaglandin H2 synthase 1', map_location='9q32-q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122370530, end_position_on_the_genomic_accession=122395705, orientation='plus', exon_count=13, OMIM=176805, no_hdr0=''),
  5743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5743, CurrentID=0, Status='live', Symbol='PTGS2', Aliases='COX-2, COX2, GRIPGHS, PGG/HS, PGHS-2, PHS-2, hCox-2', description='prostaglandin-endoperoxide synthase 2 (prostaglandin G/H synthase and cyclooxygenase)', other_designations='PGH synthase 2|PHS II|cyclooxygenase 2b|prostaglandin H2 synthase 2', map_location='1q25.2-q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186671812, end_position_on_the_genomic_accession=186680427, orientation='minus', exon_count=10, OMIM=600262, no_hdr0=''),
  5744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5744, CurrentID=0, Status='live', Symbol='PTHLH', Aliases='BDE2, HHM, PLP, PTHR, PTHRP', description='parathyroid hormone-like hormone', other_designations='PTH-rP|PTH-related protein|osteostatin|parathyroid hormone-like hormone preproprotein|parathyroid hormone-like related protein', map_location='12p12.1-p11.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27958084, end_position_on_the_genomic_accession=27972731, orientation='minus', exon_count=8, OMIM=168470, no_hdr0=''),
  5745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5745, CurrentID=0, Status='live', Symbol='PTH1R', Aliases='PFE, PTHR, PTHR1', description='parathyroid hormone 1 receptor', other_designations='PTH/PTHr receptor|PTH/PTHrP type I receptor|PTH1 receptor|parathyroid hormone receptor 1|parathyroid hormone/parathyroid hormone-related protein receptor|seven transmembrane helix receptor', map_location='3p22-p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46877717, end_position_on_the_genomic_accession=46903799, orientation='plus', exon_count=18, OMIM=168468, no_hdr0=''),
  5746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5746, CurrentID=0, Status='live', Symbol='PTH2R', Aliases='PTHR2', description='parathyroid hormone 2 receptor', other_designations='PTH2 receptor|parathyroid hormone receptor 2', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208359845, end_position_on_the_genomic_accession=208494506, orientation='plus', exon_count=15, OMIM=601469, no_hdr0=''),
  5747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5747, CurrentID=0, Status='live', Symbol='PTK2', Aliases='FADK, FAK, FAK1, FRNK, PPP1R71, p125FAK, pp125FAK', description='protein tyrosine kinase 2', other_designations='FADK 1|FAK-related non-kinase polypeptide|PTK2 protein tyrosine kinase 2|focal adhesion kinase isoform FAK-Del33|focal adhesion kinase-related nonkinase|protein phosphatase 1 regulatory subunit 71|protein phosphatase 1, regulatory subunit 71', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140658382, end_position_on_the_genomic_accession=141001313, orientation='minus', exon_count=41, OMIM=600758, no_hdr0=''),
  100533182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533182, CurrentID=0, Status='live', Symbol='SEC24B-AS1', Aliases='1/2-SBSRNA4', description='SEC24B antisense RNA 1', other_designations='SEC24B antisense RNA 1 (non-protein coding)|lncRNA_BC009800', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109429963, end_position_on_the_genomic_accession=109433817, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5753, CurrentID=0, Status='live', Symbol='PTK6', Aliases='BRK', description='protein tyrosine kinase 6', other_designations='PTK6 protein tyrosine kinase 6|breast tumor kinase|protein-tyrosine kinase BRK|tyrosine-protein kinase BRK', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63528413, end_position_on_the_genomic_accession=63537370, orientation='minus', exon_count=8, OMIM=602004, no_hdr0=''),
  5754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5754, CurrentID=0, Status='live', Symbol='PTK7', Aliases='CCK-4, CCK4', description='protein tyrosine kinase 7 (inactive)', other_designations='PTK7 protein tyrosine kinase 7|colon carcinoma kinase 4|pseudo tyrosine kinase receptor 7|tyrosine-protein kinase-like 7', map_location='6p21.1-p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43076268, end_position_on_the_genomic_accession=43161720, orientation='plus', exon_count=21, OMIM=601890, no_hdr0=''),
  100533183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533183, CurrentID=0, Status='live', Symbol='ZNF664-FAM101A', Aliases='', description='filamin-interacting protein FAM101A', other_designations='ZNF664-FAM101A readthrough|protein FAM101A|refilinA|regulator of filamin protein A', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123973215, end_position_on_the_genomic_accession=124316024, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5756, CurrentID=0, Status='live', Symbol='TWF1', Aliases='A6, PTK9', description='twinfilin actin binding protein 1', other_designations='A6 protein tyrosine kinase|PTK9 protein tyrosine kinase 9|protein A6|protein tyrosine kinase 9|twinfilin actin-binding protein 1|twinfilin, actin-binding protein, homolog 1', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43793723, end_position_on_the_genomic_accession=43806375, orientation='minus', exon_count=11, OMIM=610932, no_hdr0=''),
  5757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5757, CurrentID=0, Status='live', Symbol='PTMA', Aliases='TMSA', description='prothymosin, alpha', other_designations='gene sequence 28|prothymosin alpha protein', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231708525, end_position_on_the_genomic_accession=231713541, orientation='plus', exon_count=6, OMIM=188390, no_hdr0=''),
  5758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5758, CurrentID=0, Status='live', Symbol='PTMAP1', Aliases='', description='prothymosin, alpha pseudogene 1', other_designations='prothymosin, alpha pseudogene 1 (gene sequence 26)', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30633450, end_position_on_the_genomic_accession=30635237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5759, CurrentID=0, Status='live', Symbol='PTMAP2', Aliases='', description='prothymosin, alpha pseudogene 2', other_designations='prothymosin, alpha pseudogene 2 (gene sequence 32)', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118973562, end_position_on_the_genomic_accession=118974820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5760, CurrentID=0, Status='live', Symbol='PTMAP3', Aliases='', description='prothymosin, alpha pseudogene 3', other_designations='prothymosin, alpha pseudogene 3 (gene sequence 34)', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18011792, end_position_on_the_genomic_accession=18013057, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5761, CurrentID=0, Status='live', Symbol='PTMAP4', Aliases='PTMA', description='prothymosin, alpha pseudogene 4', other_designations='prothymosin alpha|prothymosin, alpha pseudogene 4 (gene sequence 112)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9239326, end_position_on_the_genomic_accession=9240496, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5763, CurrentID=0, Status='live', Symbol='PTMS', Aliases='ParaT', description='parathymosin', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6766139, end_position_on_the_genomic_accession=6770952, orientation='plus', exon_count=5, OMIM=168440, no_hdr0=''),
  5764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5764, CurrentID=0, Status='live', Symbol='PTN', Aliases='HARP, HBGF8, HBNF, NEGF1', description='pleiotrophin', other_designations='HB-GAM|HBBM|HBGF-8|HBNF-1|OSF-1|heparin affin regulatory protein|heparin binding growth factor 8|heparin-binding brain mitogen|heparin-binding growth factor 8|heparin-binding growth-associated molecule|heparin-binding neurite outgrowth-promoting factor 1|neurite growth-promoting factor 1|osteoblast-specific factor 1|pleiotrophin (heparin binding growth factor 8, neurite growth-promoting factor 1)', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137227341, end_position_on_the_genomic_accession=137343860, orientation='minus', exon_count=6, OMIM=162095, no_hdr0=''),
  5765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5765, CurrentID=0, Status='live', Symbol='PTOS1', Aliases='', description='ptosis, congenital 1 (autosomal dominant)', other_designations='', map_location='1p34.1-p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=178300, no_hdr0=''),
  5767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5767, CurrentID=0, Status='live', Symbol='PTP4A2P1', Aliases='PTP4AP1, h-PRL-1', description='protein tyrosine phosphatase type IVA, member 2 pseudogene 1', other_designations='protein tyrosine phosphatase IVA pseudogene 1|protein tyrosine phosphatase type IVA pseudogene 1', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42533133, end_position_on_the_genomic_accession=42534214, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5768, CurrentID=0, Status='live', Symbol='QSOX1', Aliases='Q6, QSCN6', description='quiescin sulfhydryl oxidase 1', other_designations='hQSOX|quiescin Q6 sulfhydryl oxidase 1', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180154833, end_position_on_the_genomic_accession=180198034, orientation='plus', exon_count=13, OMIM=603120, no_hdr0=''),
  5769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5769, CurrentID=0, Status='live', Symbol='PTP4A2P2', Aliases='PTP4AP2', description='protein tyrosine phosphatase type IVA, member 2 pseudogene 2', other_designations='protein tyrosine phosphatase IVA pseudogene 2|protein tyrosine phosphatase type IVA pseudogene 2', map_location='11q24-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134123427, end_position_on_the_genomic_accession=134125218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5770, CurrentID=0, Status='live', Symbol='PTPN1', Aliases='PTP1B', description='protein tyrosine phosphatase, non-receptor type 1', other_designations='protein tyrosine phosphatase, placental|protein-tyrosine phosphatase 1B', map_location='20q13.1-q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50510321, end_position_on_the_genomic_accession=50584762, orientation='plus', exon_count=10, OMIM=176885, no_hdr0=''),
  5771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5771, CurrentID=0, Status='live', Symbol='PTPN2', Aliases='PTN2, PTPT, TC-PTP, TCELLPTP, TCPTP', description='protein tyrosine phosphatase, non-receptor type 2', other_designations='T-cell protein tyrosine phosphatase', map_location='18p11.3-p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12785478, end_position_on_the_genomic_accession=12884335, orientation='minus', exon_count=15, OMIM=176887, no_hdr0=''),
  100271042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271042, CurrentID=0, Status='live', Symbol='RPL7AP25', Aliases='RPL7A_6_441', description='ribosomal protein L7a pseudogene 25', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181505009, end_position_on_the_genomic_accession=181505903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5774, CurrentID=0, Status='live', Symbol='PTPN3', Aliases='PTP-H1, PTPH1', description='protein tyrosine phosphatase, non-receptor type 3', other_designations='cytoskeletal-associated protein tyrosine phosphatase|protein-tyrosine phosphatase H1', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109375694, end_position_on_the_genomic_accession=109538353, orientation='minus', exon_count=32, OMIM=176877, no_hdr0=''),
  5775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5775, CurrentID=0, Status='live', Symbol='PTPN4', Aliases='MEG, PTPMEG, PTPMEG1', description='protein tyrosine phosphatase, non-receptor type 4 (megakaryocyte)', other_designations='PTPase-MEG1|megakaryocyte phosphatase|megakaryocyte protein-tyrosine phosphatase|protein tyrosine phosphatase MEG1|protein-tyrosine phosphatase MEG1', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119759631, end_position_on_the_genomic_accession=119984899, orientation='plus', exon_count=28, OMIM=176878, no_hdr0=''),
  5777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5777, CurrentID=0, Status='live', Symbol='PTPN6', Aliases='HCP, HCPH, HPTP1C, PTP-1C, SH-PTP1, SHP-1, SHP-1L, SHP1', description='protein tyrosine phosphatase, non-receptor type 6', other_designations='hematopoietic cell phosphatase|hematopoietic cell protein-tyrosine phosphatase|protein-tyrosine phosphatase 1C|protein-tyrosine phosphatase SHP-1', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6946576, end_position_on_the_genomic_accession=6961316, orientation='plus', exon_count=17, OMIM=176883, no_hdr0=''),
  5778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5778, CurrentID=0, Status='live', Symbol='PTPN7', Aliases='BPTP-4, HEPTP, LC-PTP, LPTP, PTPNI', description='protein tyrosine phosphatase, non-receptor type 7', other_designations='dual specificity phosphatase 1|hematopoietic protein-tyrosine phosphatase|protein-tyrosine phosphatase LC-PTP|protein-tyrosine phosphatase, nonreceptor-type, stress induced', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202147012, end_position_on_the_genomic_accession=202161592, orientation='minus', exon_count=12, OMIM=176889, no_hdr0=''),
  100271043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271043, CurrentID=0, Status='live', Symbol='RPL7AP29', Aliases='RPL7A_7_455', description='ribosomal protein L7a pseudogene 29', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3323926, end_position_on_the_genomic_accession=3324782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5780, CurrentID=0, Status='live', Symbol='PTPN9', Aliases='MEG2, PTPMEG2', description='protein tyrosine phosphatase, non-receptor type 9', other_designations='PTPase MEG2|PTPase-MEG2|protein-tyrosine phosphatase MEG2', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75467121, end_position_on_the_genomic_accession=75579291, orientation='minus', exon_count=13, OMIM=600768, no_hdr0=''),
  136853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136853, CurrentID=0, Status='live', Symbol='SSC4D', Aliases='S4D-SRCRB, SRCRB-S4D, SRCRB4D', description='scavenger receptor cysteine rich family, 4 domains', other_designations='S4-SRCRB|SRCRB4D|four scavenger receptor cysteine-rich domains-containing protein|scavenger receptor cysteine rich domain containing, group B (4 domains)|scavenger receptor cysteine-rich protein SRCRB-S4D', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76389329, end_position_on_the_genomic_accession=76409695, orientation='minus', exon_count=11, OMIM=607639, no_hdr0=''),
  5782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5782, CurrentID=0, Status='live', Symbol='PTPN12', Aliases='PTP-PEST, PTPG1', description='protein tyrosine phosphatase, non-receptor type 12', other_designations='protein-tyrosine phosphatase G1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77537456, end_position_on_the_genomic_accession=77640071, orientation='plus', exon_count=19, OMIM=600079, no_hdr0=''),
  5783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5783, CurrentID=0, Status='live', Symbol='PTPN13', Aliases='FAP-1, PNP1, PTP-BAS, PTP-BL, PTP1E, PTPL1, PTPLE, hPTP1E', description='protein tyrosine phosphatase, non-receptor type 13 (APO-1/CD95 (Fas)-associated phosphatase)', other_designations='fas-associated protein-tyrosine phosphatase 1|hPTPE1|protein-tyrosine phosphatase 1E|protein-tyrosine phosphatase PTPL1', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86594307, end_position_on_the_genomic_accession=86815179, orientation='plus', exon_count=49, OMIM=600267, no_hdr0=''),
  5784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5784, CurrentID=0, Status='live', Symbol='PTPN14', Aliases='PEZ, PTP36', description='protein tyrosine phosphatase, non-receptor type 14', other_designations='cytoskeletal-associated protein tyrosine phosphatase|protein-tyrosine phosphatase pez', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214336942, end_position_on_the_genomic_accession=214551681, orientation='minus', exon_count=22, OMIM=603155, no_hdr0=''),
  5786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5786, CurrentID=0, Status='live', Symbol='PTPRA', Aliases='HEPTP, HLPR, HPTPA, HPTPalpha, LRP, PTPA, PTPRL2, R-PTP-alpha, RPTPA', description='protein tyrosine phosphatase, receptor type, A', other_designations='Leukocyte common antigen-related peptide (protein tyrosine phosphate)|PTPLCA-related phosphatase|PTPase-alpha|protein tyrosine phosphatase, receptor type, alpha polypeptide|protein-tyrosine phosphatase alpha|tyrosine phosphatase alpha', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2864195, end_position_on_the_genomic_accession=3038669, orientation='plus', exon_count=30, OMIM=176884, no_hdr0=''),
  5787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5787, CurrentID=0, Status='live', Symbol='PTPRB', Aliases='HPTP-BETA, HPTPB, PTPB, R-PTP-BETA, VEPTP', description='protein tyrosine phosphatase, receptor type, B', other_designations='VE-PTP|protein tyrosine phosphatase, receptor type, beta polypeptide|vascular endothelial protein tyrosine phosphatase', map_location='12q15-q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70516850, end_position_on_the_genomic_accession=70637440, orientation='minus', exon_count=36, OMIM=176882, no_hdr0=''),
  5788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5788, CurrentID=0, Status='live', Symbol='PTPRC', Aliases='B220, CD45, CD45R, GP180, L-CA, LCA, LY5, T200', description='protein tyrosine phosphatase, receptor type, C', other_designations='CD45 antigen|T200 glycoprotein|T200 leukocyte common antigen|protein tyrosine phosphatase, receptor type, c polypeptide', map_location='1q31-q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198638968, end_position_on_the_genomic_accession=198757476, orientation='plus', exon_count=34, OMIM=151460, no_hdr0=''),
  5789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5789, CurrentID=0, Status='live', Symbol='PTPRD', Aliases='HPTP, HPTPD, HPTPDELTA, PTPD, RPTPDELTA', description='protein tyrosine phosphatase, receptor type, D', other_designations='R-PTP-delta|protein tyrosine phosphatase, receptor type, delta polypeptide', map_location='9p23-p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=8314246, end_position_on_the_genomic_accession=10612723, orientation='minus', exon_count=49, OMIM=601598, no_hdr0=''),
  5790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5790, CurrentID=0, Status='live', Symbol='PTPRCAP', Aliases='CD45-AP, LPAP', description='protein tyrosine phosphatase, receptor type, C-associated protein', other_designations='CD45 associated protein|CD45-associated protein|PTPRC-associated protein|lymphocyte phosphatase-associated phosphoprotein|protein tyrosine phosphatase, receptor type, c polypeptide-associated protein', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67435510, end_position_on_the_genomic_accession=67437682, orientation='minus', exon_count=2, OMIM=601577, no_hdr0=''),
  5791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5791, CurrentID=0, Status='live', Symbol='PTPRE', Aliases='HPTPE, PTPE, R-PTP-EPSILON', description='protein tyrosine phosphatase, receptor type, E', other_designations='protein tyrosine phosphatase, receptor type, epsilon polypeptide', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127907061, end_position_on_the_genomic_accession=128085916, orientation='plus', exon_count=27, OMIM=600926, no_hdr0=''),
  5792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5792, CurrentID=0, Status='live', Symbol='PTPRF', Aliases='BNAH2, LAR', description='protein tyrosine phosphatase, receptor type, F', other_designations='LCA-homolog|leukocyte antigen-related (LAR) PTP receptor|leukocyte antigen-related tyrosine phosphatase|leukocyte common antigen related|protein tyrosine phosphatase, receptor type, F polypeptide|receptor-linked protein-tyrosine phosphatase LAR', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43522238, end_position_on_the_genomic_accession=43623672, orientation='plus', exon_count=40, OMIM=179590, no_hdr0=''),
  5793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5793, CurrentID=0, Status='live', Symbol='PTPRG', Aliases='HPTPG, PTPG, R-PTP-GAMMA, RPTPG', description='protein tyrosine phosphatase, receptor type, G', other_designations='H_RG317H01.1|protein tyrosine phosphatase gamma|protein tyrosine phosphatase, receptor type, gamma polypeptide|protein-tyrosine phosphatase gamma|receptor type protein tyrosine phosphatase gamma|receptor tyrosine phosphatase gamma|receptor-type protein phosphatase gamma', map_location='3p21-p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=61561569, end_position_on_the_genomic_accession=62297609, orientation='plus', exon_count=31, OMIM=176886, no_hdr0=''),
  5794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5794, CurrentID=0, Status='live', Symbol='PTPRH', Aliases='R-PTP-H, SAP1', description='protein tyrosine phosphatase, receptor type, H', other_designations='stomach cancer-associated protein tyrosine phosphatase 1|transmembrane-type protein-tyrosine phosphatase type H', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55181247, end_position_on_the_genomic_accession=55209506, orientation='minus', exon_count=23, OMIM=602510, no_hdr0=''),
  5795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5795, CurrentID=0, Status='live', Symbol='PTPRJ', Aliases='CD148, DEP1, HPTPeta, R-PTP-ETA, SCC1', description='protein tyrosine phosphatase, receptor type, J', other_designations='CD148 antigen|DEP-1|HPTP eta|R-PTP-J|density-enhanced phosphatase 1|human density enhanced phosphatase-1|protein tyrosine phosphatase, receptor type, J polypeptide|protein-tyrosine phosphatase eta|protein-tyrosine phosphatase receptor type J|susceptibility to colon cancer 1, mouse, homolog of', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47980558, end_position_on_the_genomic_accession=48170842, orientation='plus', exon_count=27, OMIM=600925, no_hdr0=''),
  5796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5796, CurrentID=0, Status='live', Symbol='PTPRK', Aliases='R-PTP-kappa', description='protein tyrosine phosphatase, receptor type, K', other_designations='dJ480J14.2.1 (protein tyrosine phosphatase, receptor type, K (R-PTP-KAPPA, protein tyrosine phosphatase kappa , protein tyrosine phosphatase kappa|protein-tyrosine phosphatase kappa|protein-tyrosine phosphatase, receptor type, kappa', map_location='6q22.2-q22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127968779, end_position_on_the_genomic_accession=128520847, orientation='minus', exon_count=41, OMIM=602545, no_hdr0=''),
  5797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5797, CurrentID=0, Status='live', Symbol='PTPRM', Aliases='PTPRL1, R-PTP-MU, RPTPM, RPTPU, hR-PTPu', description='protein tyrosine phosphatase, receptor type, M', other_designations='protein tyrosine phosphatase mu|protein tyrosine phosphatase, receptor type, mu polypeptide|protein-tyrosine phosphatase mu', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7567316, end_position_on_the_genomic_accession=8406861, orientation='plus', exon_count=43, OMIM=176888, no_hdr0=''),
  5798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5798, CurrentID=0, Status='live', Symbol='PTPRN', Aliases='IA-2, IA-2/PTP, IA2, ICA512, R-PTP-N', description='protein tyrosine phosphatase, receptor type, N', other_designations='ICA 512|PTP IA-2|islet cell antigen 2|islet cell antigen 512|islet cell autoantigen 3|protein tyrosine phosphatase-like N', map_location='2q35-q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219289623, end_position_on_the_genomic_accession=219309573, orientation='minus', exon_count=24, OMIM=601773, no_hdr0=''),
  5799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5799, CurrentID=0, Status='live', Symbol='PTPRN2', Aliases='IA-2beta, IAR, ICAAR, PTPRP, R-PTP-N2', description='protein tyrosine phosphatase, receptor type, N polypeptide 2', other_designations='IAR/receptor-like protein-tyrosine phosphatase|islet cell autoantigen-related protein|phogrin|protein tyrosine phosphatase receptor pi|tyrosine phosphatase IA-2 beta', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157539056, end_position_on_the_genomic_accession=158587796, orientation='minus', exon_count=27, OMIM=601698, no_hdr0=''),
  5800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5800, CurrentID=0, Status='live', Symbol='PTPRO', Aliases='GLEPP1, NPHS6, PTP-OC, PTP-U2T, PTPU2, R-PTP-O, PTPRO', description='protein tyrosine phosphatase, receptor type, O', other_designations='PTP phi|PTPase U2|glomerular epithelial protein 1|osteoclastic transmembrane protein-tyrosine phosphatase|phosphotyrosine phosphatase U2|protein tyrosine phosphatase PTP-U2', map_location='12p13-p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15322257, end_position_on_the_genomic_accession=15598331, orientation='plus', exon_count=27, OMIM=600579, no_hdr0=''),
  5801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5801, CurrentID=0, Status='live', Symbol='PTPRR', Aliases='EC-PTP, PCPTP1, PTP-SL, PTPBR7, PTPRQ', description='protein tyrosine phosphatase, receptor type, R', other_designations='Ch-1 PTPase|NC-PTPCOM1|R-PTP-R|ch-1PTPase|protein tyrosine phosphatase Cr1PTPase|protein-tyrosine phosphatase NC-PTPCOM1|protein-tyrosine phosphatase PCPTP1', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70638073, end_position_on_the_genomic_accession=70920804, orientation='minus', exon_count=19, OMIM=602853, no_hdr0=''),
  5802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5802, CurrentID=0, Status='live', Symbol='PTPRS', Aliases='PTPSIGMA', description='protein tyrosine phosphatase, receptor type, S', other_designations='R-PTP-S|R-PTP-sigma|protein tyrosine phosphatase PTPsigma|receptor-type tyrosine-protein phosphatase sigma', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5205503, end_position_on_the_genomic_accession=5340803, orientation='minus', exon_count=40, OMIM=601576, no_hdr0=''),
  5803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5803, CurrentID=0, Status='live', Symbol='PTPRZ1', Aliases='HPTPZ, HPTPzeta, PTP-ZETA, PTP18, PTPRZ, PTPZ, R-PTP-zeta-2, RPTPB, RPTPbeta, phosphacan', description='protein tyrosine phosphatase, receptor-type, Z polypeptide 1', other_designations='protein tyrosine phosphatase, receptor-type, zeta polypeptide 1|protein-tyrosine phosphatase receptor type Z polypeptide 2|receptor-type tyrosine phosphatase beta/zeta', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121873105, end_position_on_the_genomic_accession=122062036, orientation='plus', exon_count=30, OMIM=176891, no_hdr0=''),
  5804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5804, CurrentID=0, Status='live', Symbol='PTPRZ2', Aliases='HTPZP2, PTPRZ', description='protein tyrosine phosphatase, receptor-type, Z polypeptide 2', other_designations='protein tyrosine phosphatase, receptor-type, zeta polypeptide 2', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604008, no_hdr0=''),
  5805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5805, CurrentID=0, Status='live', Symbol='PTS', Aliases='PTPS', description='6-pyruvoyltetrahydropterin synthase', other_designations='PTP synthase', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112226365, end_position_on_the_genomic_accession=112233973, orientation='plus', exon_count=7, OMIM=612719, no_hdr0=''),
  5806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5806, CurrentID=0, Status='live', Symbol='PTX3', Aliases='TNFAIP5, TSG-14', description='pentraxin 3', other_designations='TNF alpha-induced protein 5|pentaxin-related gene, rapidly induced by IL-1 beta, tumor necrosis factor, alpha-induced protein 5|pentaxin-related protein PTX3|pentraxin 3, long|pentraxin-3|pentraxin-related gene, rapidly induced by IL-1 beta|tumor necrosis factor alpha-induced protein 5|tumor necrosis factor, alpha-induced protein 5|tumor necrosis factor-inducible gene 14 protein|tumor necrosis factor-inducible protein TSG-14', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157436791, end_position_on_the_genomic_accession=157443628, orientation='plus', exon_count=3, OMIM=602492, no_hdr0=''),
  283887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283887, CurrentID=0, Status='live', Symbol='LOC283887', Aliases='', description='uncharacterized LOC283887', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25067038, end_position_on_the_genomic_accession=25068955, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5810, CurrentID=0, Status='live', Symbol='RAD1', Aliases='HRAD1, REC1', description='RAD1 checkpoint DNA exonuclease', other_designations='DNA repair exonuclease REC1|DNA repair exonuclease rad1 homolog|RAD1 checkpoint clamp component|RAD1 homolog|cell cycle checkpoint protein Hrad1|checkpoint control protein HRAD1|exonuclease homolog RAD1|rad1-like DNA damage checkpoint protein', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34905260, end_position_on_the_genomic_accession=34915675, orientation='minus', exon_count=6, OMIM=603153, no_hdr0=''),
  5813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5813, CurrentID=0, Status='live', Symbol='PURA', Aliases='MRD31, PUR-ALPHA, PUR1LPHA, PURA', description='purine-rich element binding protein A', other_designations='purine-rich single-stranded DNA-binding protein alpha', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140114123, end_position_on_the_genomic_accession=140119416, orientation='plus', exon_count=1, OMIM=600473, no_hdr0=''),
  5814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5814, CurrentID=0, Status='live', Symbol='PURB', Aliases='PURBETA', description='purine-rich element binding protein B', other_designations='purine-rich element-binding protein B', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44876293, end_position_on_the_genomic_accession=44885385, orientation='minus', exon_count=1, OMIM=608887, no_hdr0=''),
  100271049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271049, CurrentID=0, Status='live', Symbol='RPLP1P9', Aliases='RPLP1_3_885', description='ribosomal protein, large, P1 pseudogene 9', other_designations='', map_location='8p21-p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26771796, end_position_on_the_genomic_accession=26772129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5816, CurrentID=0, Status='live', Symbol='PVALB', Aliases='D22S749', description='parvalbumin', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36800701, end_position_on_the_genomic_accession=36819473, orientation='minus', exon_count=6, OMIM=168890, no_hdr0=''),
  5817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5817, CurrentID=0, Status='live', Symbol='PVR', Aliases='CD155, HVED, NECL5, Necl-5, PVS, TAGE4', description='poliovirus receptor', other_designations='nectin-like protein 5', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44643798, end_position_on_the_genomic_accession=44666161, orientation='plus', exon_count=8, OMIM=173850, no_hdr0=''),
  5818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5818, CurrentID=0, Status='live', Symbol='PVRL1', Aliases='CD111, CLPED1, ED4, HIgR, HV1S, HVEC, OFC7, PRR, PRR1, PVRR, PVRR1, SK-12, nectin-1', description='poliovirus receptor-related 1 (herpesvirus entry mediator C)', other_designations='ectodermal dysplasia 4 (Margarita Island type)|herpes simplex virus type 1 sensitivity|herpes virus entry mediator C|herpesvirus Ig-like receptor|nectin 1|poliovirus receptor-like 1|poliovirus receptor-related protein 1', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119638098, end_position_on_the_genomic_accession=119728725, orientation='minus', exon_count=10, OMIM=600644, no_hdr0=''),
  5819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5819, CurrentID=0, Status='live', Symbol='PVRL2', Aliases='CD112, HVEB, PRR2, PVRR2', description='poliovirus receptor-related 2 (herpesvirus entry mediator B)', other_designations='herpesvirus entry protein B|poliovirus receptor-like 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44846136, end_position_on_the_genomic_accession=44889228, orientation='plus', exon_count=10, OMIM=600798, no_hdr0=''),
  5820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5820, CurrentID=0, Status='live', Symbol='PVT1', Aliases='LINC00079, NCRNA00079, onco-lncRNA-100', description='Pvt1 oncogene (non-protein coding)', other_designations='Oncogene PVT-1 (MYC activator)|long intergenic non-protein coding RNA 79|plasmacytoma variant translocation 1|pvt-1 (murine) oncogene homolog, MYC activator', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127890628, end_position_on_the_genomic_accession=128101253, orientation='plus', exon_count=8, OMIM=165140, no_hdr0=''),
  100271050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271050, CurrentID=0, Status='live', Symbol='RPS10P16', Aliases='RPS10_4_945', description='ribosomal protein S10 pseudogene 16', other_designations='', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117523217, end_position_on_the_genomic_accession=117523726, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5822, CurrentID=0, Status='live', Symbol='PWP2', Aliases='EHOC-17H, UTP1, PWP2', description='PWP2 periodic tryptophan protein homolog (yeast)', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44107290, end_position_on_the_genomic_accession=44131181, orientation='plus', exon_count=22, OMIM=601475, no_hdr0=''),
  136895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136895, CurrentID=0, Status='live', Symbol='C7orf31', Aliases='', description='chromosome 7 open reading frame 31', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25125200, end_position_on_the_genomic_accession=25180748, orientation='minus', exon_count=11, OMIM=616071, no_hdr0=''),
  5824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5824, CurrentID=0, Status='live', Symbol='PEX19', Aliases='D1S2223E, HK33, PBD12A, PMP1, PMPI, PXF, PXMP1', description='peroxisomal biogenesis factor 19', other_designations='33 kDa housekeeping protein|housekeeping gene, 33kD|peroxin-19|peroxisomal farnesylated protein', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160276809, end_position_on_the_genomic_accession=160285151, orientation='minus', exon_count=8, OMIM=600279, no_hdr0=''),
  5825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5825, CurrentID=0, Status='live', Symbol='ABCD3', Aliases='ABC43, CBAS5, PMP70, PXMP1, ZWS2', description='ATP binding cassette subfamily D member 3', other_designations='70 kDa peroxisomal membrane protein|ATP-binding cassette, sub-family D (ALD), member 3|Peroxisomal membrane protein-1 (70kD)|dJ824O18.1 (ATP-binding cassette, sub-family D (ALD), member 3 (PMP70, PXMP1))|peroxisomal membrane protein 1 (70kD, Zellweger syndrome)', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94418086, end_position_on_the_genomic_accession=94518663, orientation='plus', exon_count=27, OMIM=170995, no_hdr0=''),
  5826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5826, CurrentID=0, Status='live', Symbol='ABCD4', Aliases='ABC41, EST352188, MAHCJ, P70R, P79R, PMP69, PXMP1L', description='ATP binding cassette subfamily D member 4', other_designations='69 kDa peroxisomal ABC-transporter|ATP-binding cassette, sub-family D (ALD), member 4|PMP70-related protein|PXMP1-L|peroxisomal membrane protein 69', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74285277, end_position_on_the_genomic_accession=74303064, orientation='minus', exon_count=20, OMIM=603214, no_hdr0=''),
  5827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5827, CurrentID=0, Status='live', Symbol='PXMP2', Aliases='PMP22', description='peroxisomal membrane protein 2', other_designations='22 kDa peroxisomal membrane protein|peroxisomal membrane protein 2, 22kDa', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132687606, end_position_on_the_genomic_accession=132704991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5828, CurrentID=0, Status='live', Symbol='PEX2', Aliases='PAF1, PBD5A, PBD5B, PMP3, PMP35, PXMP3, RNF72, ZWS3', description='peroxisomal biogenesis factor 2', other_designations='35 kDa peroxisomal membrane protein|RING finger protein 72|peroxisomal membrane protein 3, 35kDa|peroxisome assembly factor 1|peroxisome assembly factor-1', map_location='8q21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76980258, end_position_on_the_genomic_accession=77001044, orientation='minus', exon_count=5, OMIM=170993, no_hdr0=''),
  5829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5829, CurrentID=0, Status='live', Symbol='PXN', Aliases='', description='paxillin', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120210439, end_position_on_the_genomic_accession=120265771, orientation='minus', exon_count=20, OMIM=602505, no_hdr0=''),
  5830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5830, CurrentID=0, Status='live', Symbol='PEX5', Aliases='PBD2A, PBD2B, PTS1-BP, PTS1R, PXR1', description='peroxisomal biogenesis factor 5', other_designations='PTS1 receptor|peroxin-5|peroxisomal C-terminal targeting signal import receptor|peroxisomal targeting signal 1 (SKL type) receptor|peroxisomal targeting signal 1 receptor|peroxisomal targeting signal import receptor|peroxisomal targeting signal receptor 1|peroxisome receptor 1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7189163, end_position_on_the_genomic_accession=7218574, orientation='plus', exon_count=21, OMIM=600414, no_hdr0=''),
  5831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5831, CurrentID=0, Status='live', Symbol='PYCR1', Aliases='ARCL2B, ARCL3B, P5C, P5CR, PIG45, PP222, PRO3, PYCR', description='pyrroline-5-carboxylate reductase 1', other_designations='P5C reductase 1|proliferation-inducing protein 45', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81932384, end_position_on_the_genomic_accession=81937328, orientation='minus', exon_count=9, OMIM=179035, no_hdr0=''),
  5832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5832, CurrentID=0, Status='live', Symbol='ALDH18A1', Aliases='ARCL3A, GSAS, P5CS, PYCS', description='aldehyde dehydrogenase 18 family member A1', other_designations='aldehyde dehydrogenase 18 family, member A1|aldehyde dehydrogenase family 18 member A1|delta-1-pyrroline-5-carboxylate synthetase|delta1-pyrroline-5-carboxlate synthetase|pyrroline-5-carboxylate synthetase (glutamate gamma-semialdehyde synthetase)', map_location='10q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95605929, end_position_on_the_genomic_accession=95656810, orientation='minus', exon_count=18, OMIM=138250, no_hdr0=''),
  5833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5833, CurrentID=0, Status='live', Symbol='PCYT2', Aliases='ET', description='phosphate cytidylyltransferase 2, ethanolamine', other_designations='CTP:phosphoethanolamine cytidylyltransferase|phosphorylethanolamine transferase', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81902901, end_position_on_the_genomic_accession=81911477, orientation='minus', exon_count=16, OMIM=602679, no_hdr0=''),
  5834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5834, CurrentID=0, Status='live', Symbol='PYGB', Aliases='GPBB', description='phosphorylase, glycogen; brain', other_designations='glycogen phosphorylase B', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25248070, end_position_on_the_genomic_accession=25298012, orientation='plus', exon_count=20, OMIM=138550, no_hdr0=''),
  5836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5836, CurrentID=0, Status='live', Symbol='PYGL', Aliases='GSD6', description='phosphorylase, glycogen, liver', other_designations='', map_location='14q21-q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50905217, end_position_on_the_genomic_accession=50944530, orientation='minus', exon_count=20, OMIM=613741, no_hdr0=''),
  5837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5837, CurrentID=0, Status='live', Symbol='PYGM', Aliases='', description='phosphorylase, glycogen, muscle', other_designations='myophosphorylase', map_location='11q12-q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64746389, end_position_on_the_genomic_accession=64760715, orientation='minus', exon_count=20, OMIM=608455, no_hdr0=''),
  100271053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271053, CurrentID=0, Status='live', Symbol='RPS12P5', Aliases='RPS12_1_355', description='ribosomal protein S12 pseudogene 5', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29390815, end_position_on_the_genomic_accession=29391241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271054, CurrentID=0, Status='live', Symbol='RPS12P6', Aliases='RPS12_2_384', description='ribosomal protein S12 pseudogene 6', other_designations='', map_location='3p12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=78758978, end_position_on_the_genomic_accession=78759369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271055, CurrentID=0, Status='live', Symbol='RPS12P7', Aliases='RPS12_3_426', description='ribosomal protein S12 pseudogene 7', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158496566, end_position_on_the_genomic_accession=158496929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  153893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153893, CurrentID=0, Status='live', Symbol='LOC153893', Aliases='', description='DCN1, defective in cullin neddylation 1, domain containing 1 (S. cerevisiae) pseudogene', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100530262, end_position_on_the_genomic_accession=100531130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271056, CurrentID=0, Status='live', Symbol='RPS12P10', Aliases='RPS12_4_622', description='ribosomal protein S12 pseudogene 10', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142566700, end_position_on_the_genomic_accession=142567082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5858, CurrentID=0, Status='live', Symbol='PZP', Aliases='CPAMD6', description='pregnancy-zone protein', other_designations='C3 and PZP-like alpha-2-macroglobulin domain-containing protein 6|Pregnancy zone protein', map_location='12p13-p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9142819, end_position_on_the_genomic_accession=9208370, orientation='minus', exon_count=38, OMIM=176420, no_hdr0=''),
  5859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5859, CurrentID=0, Status='live', Symbol='QARS', Aliases='GLNRS, MSCCA, PRO2195', description='glutaminyl-tRNA synthetase', other_designations='glutamine-tRNA synthetase', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49095932, end_position_on_the_genomic_accession=49105129, orientation='minus', exon_count=24, OMIM=603727, no_hdr0=''),
  5860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5860, CurrentID=0, Status='live', Symbol='QDPR', Aliases='DHPR, PKU2, SDR33C1', description='quinoid dihydropteridine reductase', other_designations='6,7-dihydropteridine reductase|HDHPR|short chain dehydrogenase/reductase family 33C member 1|short chain dehydrogenase/reductase family 33C, member 1', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17486393, end_position_on_the_genomic_accession=17512234, orientation='minus', exon_count=7, OMIM=612676, no_hdr0=''),
  5861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5861, CurrentID=0, Status='live', Symbol='RAB1A', Aliases='RAB1, YPT1', description='RAB1A, member RAS oncogene family', other_designations='GTP binding protein Rab1a|RAB1, member RAS oncogene family|Rab GTPase YPT1 homolog|YPT1-related protein', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65086854, end_position_on_the_genomic_accession=65130301, orientation='minus', exon_count=7, OMIM=179508, no_hdr0=''),
  5862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5862, CurrentID=0, Status='live', Symbol='RAB2A', Aliases='LHX, RAB2', description='RAB2A, member RAS oncogene family', other_designations='RAB2, member RAS oncogene family|small GTP binding protein RAB2A', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60516910, end_position_on_the_genomic_accession=60623644, orientation='plus', exon_count=8, OMIM=179509, no_hdr0=''),
  5863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5863, CurrentID=0, Status='live', Symbol='RGL2', Aliases='HKE1.5, KE1.5, RAB2L', description='ral guanine nucleotide dissociation stimulator-like 2', other_designations='GDS-related protein|ralGDS-like 2|ralGDS-like factor|ras-associated protein RAB2L', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33291654, end_position_on_the_genomic_accession=33299452, orientation='minus', exon_count=20, OMIM=602306, no_hdr0=''),
  5864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5864, CurrentID=0, Status='live', Symbol='RAB3A', Aliases='', description='RAB3A, member RAS oncogene family', other_designations='RAS-associated protein RAB3A', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18196799, end_position_on_the_genomic_accession=18204064, orientation='minus', exon_count=6, OMIM=179490, no_hdr0=''),
  5865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5865, CurrentID=0, Status='live', Symbol='RAB3B', Aliases='', description='RAB3B, member RAS oncogene family', other_designations='brain antigen RAB3B', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51907956, end_position_on_the_genomic_accession=51990764, orientation='minus', exon_count=6, OMIM=179510, no_hdr0=''),
  5866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5866, CurrentID=0, Status='live', Symbol='RAB3IL1', Aliases='GRAB', description='RAB3A interacting protein (rabin3)-like 1', other_designations='rab3A-interacting-like protein 1|rabin3-like 1', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61897240, end_position_on_the_genomic_accession=61946311, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  5867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5867, CurrentID=0, Status='live', Symbol='RAB4A', Aliases='HRES-1, HRES-1/RAB4, HRES1, RAB4', description='RAB4A, member RAS oncogene family', other_designations='HTLV-1 related endogenous sequence', map_location='1q42-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229271062, end_position_on_the_genomic_accession=229305894, orientation='plus', exon_count=8, OMIM=179511, no_hdr0=''),
  5868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5868, CurrentID=0, Status='live', Symbol='RAB5A', Aliases='RAB5', description='RAB5A, member RAS oncogene family', other_designations='RAS-associated protein RAB5A', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19947080, end_position_on_the_genomic_accession=19985175, orientation='plus', exon_count=6, OMIM=179512, no_hdr0=''),
  5869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5869, CurrentID=0, Status='live', Symbol='RAB5B', Aliases='', description='RAB5B, member RAS oncogene family', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55973913, end_position_on_the_genomic_accession=55996683, orientation='plus', exon_count=8, OMIM=179514, no_hdr0=''),
  5870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5870, CurrentID=0, Status='live', Symbol='RAB6A', Aliases='RAB6', description='RAB6A, member RAS oncogene family', other_designations='RAB6, member RAS oncogene family|Rab GTPase', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73675638, end_position_on_the_genomic_accession=73761156, orientation='minus', exon_count=10, OMIM=179513, no_hdr0=''),
  5871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5871, CurrentID=0, Status='live', Symbol='MAP4K2', Aliases='BL44, GCK, RAB8IP', description='mitogen-activated protein kinase kinase kinase kinase 2', other_designations='B lymphocyte serine/threonine protein kinase|MAPK/ERK kinase kinase kinase 2|MEK kinase kinase 2|MEKKK 2|Rab8 interacting protein|germinal centre kinase (GC kinase)', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64789137, end_position_on_the_genomic_accession=64803248, orientation='minus', exon_count=32, OMIM=603166, no_hdr0=''),
  5872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5872, CurrentID=0, Status='live', Symbol='RAB13', Aliases='GIG4', description='RAB13, member RAS oncogene family', other_designations='RAS-associated protein RAB13|cell growth-inhibiting gene 4 protein|growth-inhibiting gene 4 protein', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153981617, end_position_on_the_genomic_accession=153986377, orientation='minus', exon_count=9, OMIM=602672, no_hdr0=''),
  5873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5873, CurrentID=0, Status='live', Symbol='RAB27A', Aliases='GS2, HsT18676, RAB27, RAM', description='RAB27A, member RAS oncogene family', other_designations='GTP-binding protein Ram|mutant Ras-related protein Rab-27A|rab-27', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55202966, end_position_on_the_genomic_accession=55289828, orientation='minus', exon_count=12, OMIM=603868, no_hdr0=''),
  5874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5874, CurrentID=0, Status='live', Symbol='RAB27B', Aliases='C25KG', description='RAB27B, member RAS oncogene family', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54717866, end_position_on_the_genomic_accession=54895516, orientation='plus', exon_count=11, OMIM=603869, no_hdr0=''),
  5875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5875, CurrentID=0, Status='live', Symbol='RABGGTA', Aliases='PTAR3', description='Rab geranylgeranyltransferase, alpha subunit', other_designations='Rab GG transferase alpha|Rab GGTase alpha|geranylgeranyl transferase type II subunit alpha|protein prenyltransferase alpha subunit repeat containing 3|rab geranyl-geranyltransferase subunit alpha', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24265538, end_position_on_the_genomic_accession=24271627, orientation='minus', exon_count=16, OMIM=601905, no_hdr0=''),
  5876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5876, CurrentID=0, Status='live', Symbol='RABGGTB', Aliases='GGTB', description='Rab geranylgeranyltransferase, beta subunit', other_designations='GGTase-II-beta|rab GG transferase beta|rab GGTase beta|type II protein geranyl-geranyltransferase subunit beta', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75786194, end_position_on_the_genomic_accession=75795090, orientation='plus', exon_count=11, OMIM=179080, no_hdr0=''),
  5877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5877, CurrentID=0, Status='live', Symbol='RABIF', Aliases='MSS4, RASGFR3, RASGRF3', description='RAB interacting factor', other_designations='Ras-specific guanine-releasing factor 3|mammalian suppressor of SEC4|rab-interacting factor', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202878282, end_position_on_the_genomic_accession=202889257, orientation='minus', exon_count=2, OMIM=603417, no_hdr0=''),
  5878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5878, CurrentID=0, Status='live', Symbol='RAB5C', Aliases='L1880L, RAB5L, RABL, RAB5C', description='RAB5C, member RAS oncogene family', other_designations='RAB5C, member of RAS oncogene family', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42124976, end_position_on_the_genomic_accession=42155044, orientation='minus', exon_count=9, OMIM=604037, no_hdr0=''),
  5879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5879, CurrentID=0, Status='live', Symbol='RAC1', Aliases='MIG5, Rac-1, TC-25, p21-Rac1', description='ras-related C3 botulinum toxin substrate 1 (rho family, small GTP binding protein Rac1)', other_designations='cell migration-inducing gene 5 protein|ras-like protein TC25', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6374495, end_position_on_the_genomic_accession=6403967, orientation='plus', exon_count=7, OMIM=602048, no_hdr0=''),
  5880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5880, CurrentID=0, Status='live', Symbol='RAC2', Aliases='EN-7, Gx, HSPC022, p21-Rac2', description='ras-related C3 botulinum toxin substrate 2 (rho family, small GTP binding protein Rac2)', other_designations='Ras-related C3 botulinum toxin substrate 3 (rho family, small GTP-binding protein Rac2)|small G protein', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37225261, end_position_on_the_genomic_accession=37244299, orientation='minus', exon_count=8, OMIM=602049, no_hdr0=''),
  5881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5881, CurrentID=0, Status='live', Symbol='RAC3', Aliases='', description='ras-related C3 botulinum toxin substrate 3 (rho family, small GTP binding protein Rac3)', other_designations='p21-Rac3|rho family, small GTP binding protein Rac3', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82031656, end_position_on_the_genomic_accession=82034204, orientation='plus', exon_count=6, OMIM=602050, no_hdr0=''),
  5883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5883, CurrentID=0, Status='live', Symbol='RAD9A', Aliases='RAD9', description='RAD9 checkpoint clamp component A', other_designations='DNA repair exonuclease rad9 homolog A|RAD9 homolog A|hRAD9', map_location='11q13.1-q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67391952, end_position_on_the_genomic_accession=67398412, orientation='plus', exon_count=11, OMIM=603761, no_hdr0=''),
  5884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5884, CurrentID=0, Status='live', Symbol='RAD17', Aliases='CCYC, HRAD17, R24LSP, RAD24, RAD17', description='RAD17 checkpoint clamp loader component', other_designations='RAD1 homolog|RAD17 homolog|RF-C activator 1 homolog|RF-C/activator 1 homolog|Rad17-like protein|cell cycle checkpoint protein (RAD17)', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69369297, end_position_on_the_genomic_accession=69414801, orientation='plus', exon_count=21, OMIM=603139, no_hdr0=''),
  5885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5885, CurrentID=0, Status='live', Symbol='RAD21', Aliases='CDLS4, HR21, HRAD21, MCD1, NXP1, SCC1, hHR21', description='RAD21 cohesin complex component', other_designations='NXP-1|RAD21 homolog|SCC1 homolog|kleisin|nuclear matrix protein 1|protein involved in DNA double-strand break repair|sister chromatid cohesion 1', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116845934, end_position_on_the_genomic_accession=116874866, orientation='minus', exon_count=14, OMIM=606462, no_hdr0=''),
  5886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5886, CurrentID=0, Status='live', Symbol='RAD23A', Aliases='HHR23A, HR23A', description='RAD23 homolog A, nucleotide excision repair protein', other_designations='RAD23, yeast homolog, A', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12945814, end_position_on_the_genomic_accession=12953643, orientation='plus', exon_count=9, OMIM=600061, no_hdr0=''),
  5887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5887, CurrentID=0, Status='live', Symbol='RAD23B', Aliases='HHR23B, HR23B, P58', description='RAD23 homolog B, nucleotide excision repair protein', other_designations='RAD23, yeast homolog of, B|XP-C repair complementing complex 58 kDa|XP-C repair complementing protein|XP-C repair-complementing complex 58 kDa protein', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107283236, end_position_on_the_genomic_accession=107332194, orientation='plus', exon_count=12, OMIM=600062, no_hdr0=''),
  5888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5888, CurrentID=0, Status='live', Symbol='RAD51', Aliases='BRCC5, FANCR, HRAD51, HsRad51, HsT16930, MRMV2A, RECA, RAD51', description='RAD51 recombinase', other_designations='BRCA1/BRCA2-containing complex, subunit 5|RAD51 homolog A|RecA, E. coli, homolog of|RecA-like protein|recombination protein A', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40694774, end_position_on_the_genomic_accession=40732158, orientation='plus', exon_count=13, OMIM=179617, no_hdr0=''),
  5889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5889, CurrentID=0, Status='live', Symbol='RAD51C', Aliases='BROVCA3, FANCO, R51H3, RAD51L2', description='RAD51 paralog C', other_designations='RAD51-like protein 2|yeast RAD51 homolog 3', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58692536, end_position_on_the_genomic_accession=58734342, orientation='plus', exon_count=12, OMIM=602774, no_hdr0=''),
  5890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5890, CurrentID=0, Status='live', Symbol='RAD51B', Aliases='R51H2, RAD51L1, REC2', description='RAD51 paralog B', other_designations='RAD51 homolog B|RecA-like protein|recombination repair protein', map_location='14q23-q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67819779, end_position_on_the_genomic_accession=68683210, orientation='plus', exon_count=21, OMIM=602948, no_hdr0=''),
  5891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5891, CurrentID=0, Status='live', Symbol='MOK', Aliases='RAGE, RAGE-1, RAGE1, STK30', description='MOK protein kinase', other_designations='renal cell carcinoma antigen|renal tumor antigen 1', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102224450, end_position_on_the_genomic_accession=102305200, orientation='minus', exon_count=18, OMIM=605762, no_hdr0=''),
  5892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5892, CurrentID=0, Status='live', Symbol='RAD51D', Aliases='BROVCA4, R51H3, RAD51L3, TRAD', description='RAD51 paralog D', other_designations='RAD51 homolog D|RAD51-like protein 3|recombination repair protein', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35099792, end_position_on_the_genomic_accession=35119869, orientation='minus', exon_count=11, OMIM=602954, no_hdr0=''),
  5893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5893, CurrentID=0, Status='live', Symbol='RAD52', Aliases='', description='RAD52 homolog, DNA repair protein', other_designations='recombination protein RAD52|rhabdomyosarcoma antigen MU-RMS-40.23', map_location='12p13-p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=911259, end_position_on_the_genomic_accession=990608, orientation='minus', exon_count=16, OMIM=600392, no_hdr0=''),
  5894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5894, CurrentID=0, Status='live', Symbol='RAF1', Aliases='CMD1NN, CRAF, NS5, Raf-1, c-Raf', description='Raf-1 proto-oncogene, serine/threonine kinase', other_designations='C-Raf proto-oncogene, serine/threonine kinase|Oncogene RAF1|proto-oncogene c-RAF|raf proto-oncogene serine/threonine protein kinase|v-raf-1 murine leukemia viral oncogene homolog 1|v-raf-1 murine leukemia viral oncogene-like protein 1', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12583601, end_position_on_the_genomic_accession=12664201, orientation='minus', exon_count=20, OMIM=164760, no_hdr0=''),
  101929203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929203, CurrentID=0, Status='live', Symbol='LOC101929203', Aliases='', description='RING finger and transmembrane domain-containing protein 1-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20751329, end_position_on_the_genomic_accession=20754588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5896, CurrentID=0, Status='live', Symbol='RAG1', Aliases='RAG-1, RNF74', description='recombination activating gene 1', other_designations='RING finger protein 74|recombination activating protein 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36510400, end_position_on_the_genomic_accession=36579762, orientation='plus', exon_count=5, OMIM=179615, no_hdr0=''),
  5897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5897, CurrentID=0, Status='live', Symbol='RAG2', Aliases='RAG-2', description='recombination activating gene 2', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36591943, end_position_on_the_genomic_accession=36598279, orientation='minus', exon_count=3, OMIM=179616, no_hdr0=''),
  5898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5898, CurrentID=0, Status='live', Symbol='RALA', Aliases='RAL', description='v-ral simian leukemia viral oncogene homolog A (ras related)', other_designations='RAS-like protein A|Ras family small GTP binding protein RALA|ras related GTP binding protein A', map_location='7p15-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39622955, end_position_on_the_genomic_accession=39708124, orientation='plus', exon_count=8, OMIM=179550, no_hdr0=''),
  5899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5899, CurrentID=0, Status='live', Symbol='RALB', Aliases='', description='v-ral simian leukemia viral oncogene homolog B', other_designations='RAS-like protein B|ras related GTP binding protein B|v-ral simian leukemia viral oncogene homolog B (ras related; GTP binding protein)', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120240070, end_position_on_the_genomic_accession=120294711, orientation='plus', exon_count=8, OMIM=179551, no_hdr0=''),
  5900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5900, CurrentID=0, Status='live', Symbol='RALGDS', Aliases='RGDS, RGF, RalGEF', description='ral guanine nucleotide dissociation stimulator', other_designations='ral guanine nucleotide exchange factor', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133097720, end_position_on_the_genomic_accession=133149220, orientation='minus', exon_count=21, OMIM=601619, no_hdr0=''),
  5901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5901, CurrentID=0, Status='live', Symbol='RAN', Aliases='ARA24, Gsp1, TC4', description='RAN, member RAS oncogene family', other_designations='GTPase Ran|OK/SW-cl.81|RanGTPase|androgen receptor-associated protein 24|guanosine triphosphatase Ran|member RAS oncogene family|ras-like protein TC4|ras-related nuclear protein', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130871994, end_position_on_the_genomic_accession=130877678, orientation='plus', exon_count=7, OMIM=601179, no_hdr0=''),
  5902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5902, CurrentID=0, Status='live', Symbol='RANBP1', Aliases='HTF9A', description='RAN binding protein 1', other_designations='HpaII tiny fragments locus 9A', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20115938, end_position_on_the_genomic_accession=20127357, orientation='plus', exon_count=9, OMIM=601180, no_hdr0=''),
  5903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5903, CurrentID=0, Status='live', Symbol='RANBP2', Aliases='ADANE, ANE1, IIAE3, NUP358, TRP1, TRP2', description='RAN binding protein 2', other_designations='358 kDa nucleoporin|P270|acute necrotizing encephalopathy 1 (autosomal dominant)|nuclear pore complex protein Nup358|nucleoporin 358|nucleoporin Nup358|ran-binding protein 2|transformation-related protein 2', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108719446, end_position_on_the_genomic_accession=108785811, orientation='plus', exon_count=31, OMIM=601181, no_hdr0=''),
  5905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5905, CurrentID=0, Status='live', Symbol='RANGAP1', Aliases='Fug1, RANGAP, SD', description='Ran GTPase activating protein 1', other_designations='segregation distorter homolog|segregation distortion', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41244777, end_position_on_the_genomic_accession=41302210, orientation='minus', exon_count=22, OMIM=602362, no_hdr0=''),
  5906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5906, CurrentID=0, Status='live', Symbol='RAP1A', Aliases='C21KG, G-22K, KREV-1, KREV1, RAP1, SMGP21', description='RAP1A, member of RAS oncogene family', other_designations='GTP-binding protein smg p21A|Ras-related protein Krev-1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111619783, end_position_on_the_genomic_accession=111716695, orientation='plus', exon_count=11, OMIM=179520, no_hdr0=''),
  5907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5907, CurrentID=0, Status='live', Symbol='RAP1AP', Aliases='KREV1P, SMGP21A', description='RAP1A, member of RAS oncogene family pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74482847, end_position_on_the_genomic_accession=74484166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5908, CurrentID=0, Status='live', Symbol='RAP1B', Aliases='K-REV, RAL1B', description='RAP1B, member of RAS oncogene family', other_designations='GTP-binding protein smg p21B|RAS-related protein RAP1B|Ras family small GTP binding protein RAP1B|small GTP binding protein', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68610839, end_position_on_the_genomic_accession=68660605, orientation='plus', exon_count=8, OMIM=179530, no_hdr0=''),
  5909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5909, CurrentID=0, Status='live', Symbol='RAP1GAP', Aliases='RAP1GA11, RAP1GAPII, RAPGAP, RAP1GAP', description='RAP1 GTPase activating protein', other_designations='', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21596215, end_position_on_the_genomic_accession=21669375, orientation='minus', exon_count=30, OMIM=600278, no_hdr0=''),
  5910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5910, CurrentID=0, Status='live', Symbol='RAP1GDS1', Aliases='GDS1, SmgGDS', description='RAP1, GTP-GDP dissociation stimulator 1', other_designations='SMG GDS protein|SMG P21 stimulatory GDP/GTP exchange protein|exchange factor smgGDS', map_location='4q23-q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98261376, end_position_on_the_genomic_accession=98443861, orientation='plus', exon_count=16, OMIM=179502, no_hdr0=''),
  5911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5911, CurrentID=0, Status='live', Symbol='RAP2A', Aliases='K-REV, KREV, RAP2, RbBP-30', description='RAP2A, member of RAS oncogene family', other_designations='RAP2, member of RAS oncogene family (K-rev)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97434221, end_position_on_the_genomic_accession=97467998, orientation='plus', exon_count=2, OMIM=179540, no_hdr0=''),
  5912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5912, CurrentID=0, Status='live', Symbol='RAP2B', Aliases='', description='RAP2B, member of RAS oncogene family', other_designations='small GTP binding protein', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153162212, end_position_on_the_genomic_accession=153170624, orientation='plus', exon_count=1, OMIM=179541, no_hdr0=''),
  5913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5913, CurrentID=0, Status='live', Symbol='RAPSN', Aliases='CMS11, CMS4C, FADS, RAPSYN, RNF205', description='receptor associated protein of the synapse', other_designations='43 kda postsynaptic protein|RING finger protein 205|acetylcholine receptor-associated 43 kda protein|receptor-associated protein of the synapse', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47437757, end_position_on_the_genomic_accession=47449178, orientation='minus', exon_count=8, OMIM=601592, no_hdr0=''),
  5914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5914, CurrentID=0, Status='live', Symbol='RARA', Aliases='NR1B1, RAR', description='retinoic acid receptor, alpha', other_designations='RAR-alpha|nuclear receptor subfamily 1 group B member 1|nucleophosmin-retinoic acid receptor alpha fusion protein NPM-RAR long form|retinoic acid nuclear receptor alpha variant 1|retinoic acid nuclear receptor alpha variant 2|retinoic acid receptor, alpha polypeptide', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40309171, end_position_on_the_genomic_accession=40357643, orientation='plus', exon_count=16, OMIM=180240, no_hdr0=''),
  5915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5915, CurrentID=0, Status='live', Symbol='RARB', Aliases='HAP, MCOPS12, NR1B2, RRB2', description='retinoic acid receptor, beta', other_designations='HBV-activated protein|RAR-beta|RAR-epsilon|hepatitis B virus activated protein|nuclear receptor subfamily 1 group B member 2|retinoic acid receptor, beta polypeptide', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25174332, end_position_on_the_genomic_accession=25597932, orientation='plus', exon_count=11, OMIM=180220, no_hdr0=''),
  5916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5916, CurrentID=0, Status='live', Symbol='RARG', Aliases='NR1B3, RARC', description='retinoic acid receptor, gamma', other_designations='RAR-gamma|nuclear receptor subfamily 1 group B member 3|retinoic acid nuclear receptor gamma variant 1|retinoic acid nuclear receptor gamma variant 2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53210566, end_position_on_the_genomic_accession=53232256, orientation='minus', exon_count=13, OMIM=180190, no_hdr0=''),
  5917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5917, CurrentID=0, Status='live', Symbol='RARS', Aliases='ArgRS, DALRD1, HLD9', description='arginyl-tRNA synthetase', other_designations='arginine tRNA ligase 1, cytoplasmic|arginyl-tRNA synthetase, cytoplasmic', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168486458, end_position_on_the_genomic_accession=168519306, orientation='plus', exon_count=15, OMIM=107820, no_hdr0=''),
  5918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5918, CurrentID=0, Status='live', Symbol='RARRES1', Aliases='LXNL, PERG-1, TIG1', description='retinoic acid receptor responder (tazarotene induced) 1', other_designations='RAR-responsive protein TIG1|latexin-like|phorbol ester-induced gene 1 protein|tazarotene-induced gene 1 protein', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158696892, end_position_on_the_genomic_accession=158732691, orientation='minus', exon_count=6, OMIM=605090, no_hdr0=''),
  5919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5919, CurrentID=0, Status='live', Symbol='RARRES2', Aliases='HP10433, TIG2', description='retinoic acid receptor responder (tazarotene induced) 2', other_designations='RAR-responsive protein TIG2|chemerin|tazarotene-induced gene 2 protein', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150338318, end_position_on_the_genomic_accession=150341674, orientation='minus', exon_count=6, OMIM=601973, no_hdr0=''),
  5920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5920, CurrentID=0, Status='live', Symbol='RARRES3', Aliases='HRASLS4, HRSL4, PLA1/2-3, RIG1, TIG3', description='retinoic acid receptor responder (tazarotene induced) 3', other_designations='HRAS-like suppressor 4|RAR-responsive protein TIG3|retinoic acid-inducible gene 1|retinoid-inducible gene 1 protein|tazarotene-induced gene 3 protein', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63536801, end_position_on_the_genomic_accession=63546458, orientation='plus', exon_count=4, OMIM=605092, no_hdr0=''),
  5921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5921, CurrentID=0, Status='live', Symbol='RASA1', Aliases='CM-AVM, CMAVM, GAP, PKWS, RASA, RASGAP, p120GAP, p120RASGAP', description='RAS p21 protein activator (GTPase activating protein) 1', other_designations='triphosphatase-activating protein', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87267801, end_position_on_the_genomic_accession=87391926, orientation='plus', exon_count=27, OMIM=139150, no_hdr0=''),
  5922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5922, CurrentID=0, Status='live', Symbol='RASA2', Aliases='GAP1M', description='RAS p21 protein activator 2', other_designations='GTPase-activating protein 1m|GTPase-activating protein of RAS', map_location='3q22-q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141487047, end_position_on_the_genomic_accession=141615363, orientation='plus', exon_count=27, OMIM=601589, no_hdr0=''),
  5923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5923, CurrentID=0, Status='live', Symbol='RASGRF1', Aliases='CDC25, CDC25L, GNRP, GRF1, GRF55, H-GRF55, PP13187, ras-GRF1', description='Ras protein specific guanine nucleotide releasing factor 1', other_designations='Ras protein-specific guanine nucleotide-releasing factor 1|Ras-specific guanine nucleotide-releasing factor, CDC25 homolog|Ras-specific nucleotide exchange factor CDC25|guanine nucleotide exchange factor|guanine nucleotide-releasing factor 1|guanine nucleotide-releasing factor, 55 kD|guanine nucleotide-releasing protein', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78959947, end_position_on_the_genomic_accession=79090873, orientation='minus', exon_count=31, OMIM=606600, no_hdr0=''),
  5924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5924, CurrentID=0, Status='live', Symbol='RASGRF2', Aliases='GRF2, RAS-GRF2', description='Ras protein specific guanine nucleotide releasing factor 2', other_designations='Ras protein-specific guanine nucleotide-releasing factor 2|ras guanine nucleotide exchange factor 2', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80960672, end_position_on_the_genomic_accession=81230162, orientation='plus', exon_count=28, OMIM=606614, no_hdr0=''),
  5925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5925, CurrentID=0, Status='live', Symbol='RB1', Aliases='OSRC, PPP1R130, RB, p105-Rb, pRb, pp110', description='retinoblastoma 1', other_designations='prepro-retinoblastoma-associated protein|protein phosphatase 1, regulatory subunit 130|retinoblastoma suspectibility protein', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48303747, end_position_on_the_genomic_accession=48481890, orientation='plus', exon_count=28, OMIM=614041, no_hdr0=''),
  5926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5926, CurrentID=0, Status='live', Symbol='ARID4A', Aliases='RBBP-1, RBBP1, RBP-1, RBP1', description='AT-rich interaction domain 4A', other_designations='ARID domain-containing protein 4A|AT rich interactive domain 4A (RBP1-like)|retinoblastoma binding protein 1|retinoblastoma-binding protein 1', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58298385, end_position_on_the_genomic_accession=58373995, orientation='plus', exon_count=26, OMIM=180201, no_hdr0=''),
  5927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5927, CurrentID=0, Status='live', Symbol='KDM5A', Aliases='RBBP-2, RBBP2, RBP2', description='lysine (K)-specific demethylase 5A', other_designations='Jumonji, AT rich interactive domain 1A (RBP2-like)|histone demethylase JARID1A|jumonji/ARID domain-containing protein 1A|retinoblastoma binding protein 2', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=280057, end_position_on_the_genomic_accession=389455, orientation='minus', exon_count=28, OMIM=180202, no_hdr0=''),
  5928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5928, CurrentID=0, Status='live', Symbol='RBBP4', Aliases='NURF55, RBAP48, lin-53', description='retinoblastoma binding protein 4', other_designations='CAF-1 subunit C|CAF-I 48 kDa subunit|CAF-I p48|MSI1 protein homolog|RBBP-4|chromatin assembly factor 1 subunit C|chromatin assembly factor I p48 subunit|chromatin assembly factor/CAF-1 p48 subunit|nucleosome-remodeling factor subunit RBAP48|retinoblastoma-binding protein 4|retinoblastoma-binding protein p48', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32651148, end_position_on_the_genomic_accession=32686211, orientation='plus', exon_count=13, OMIM=602923, no_hdr0=''),
  5929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5929, CurrentID=0, Status='live', Symbol='RBBP5', Aliases='RBQ3, SWD1', description='retinoblastoma binding protein 5', other_designations='RBBP-5|SWD1, Set1c WD40 repeat protein, homolog|retinoblastoma-binding protein RBQ-3', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205086142, end_position_on_the_genomic_accession=205122022, orientation='minus', exon_count=14, OMIM=600697, no_hdr0=''),
  5930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5930, CurrentID=0, Status='live', Symbol='RBBP6', Aliases='MY038, P2P-R, PACT, RBQ-1, SNAMA', description='retinoblastoma binding protein 6', other_designations='RB-binding Q-protein 1|p53-associated cellular protein of testis|proliferation potential-related protein|protein P2P-R|retinoblastoma-binding Q protein 1|retinoblastoma-binding protein 6', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24539587, end_position_on_the_genomic_accession=24572863, orientation='plus', exon_count=19, OMIM=600938, no_hdr0=''),
  5931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5931, CurrentID=0, Status='live', Symbol='RBBP7', Aliases='RbAp46', description='retinoblastoma binding protein 7', other_designations='G1/S transition control protein-binding protein RbAp46|RBBP-7|histone acetyltransferase type B subunit 2|nucleosome-remodeling factor subunit RBAP46|retinoblastoma-binding protein 7|retinoblastoma-binding protein RbAp46|retinoblastoma-binding protein p46', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16844650, end_position_on_the_genomic_accession=16870442, orientation='minus', exon_count=14, OMIM=300825, no_hdr0=''),
  5932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5932, CurrentID=0, Status='live', Symbol='RBBP8', Aliases='COM1, CTIP, JWDS, RIM, SAE2, SCKL2', description='retinoblastoma binding protein 8', other_designations='CTBP-interacting protein|RBBP-8|sporulation in the absence of SPO11 protein 2 homolog', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22914121, end_position_on_the_genomic_accession=23026488, orientation='plus', exon_count=24, OMIM=604124, no_hdr0=''),
  5933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5933, CurrentID=0, Status='live', Symbol='RBL1', Aliases='CP107, PRB1, p107', description='retinoblastoma-like 1', other_designations='107 kDa retinoblastoma-associated protein|cellular protein 107|retinoblastoma-like 1 (p107)', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36996349, end_position_on_the_genomic_accession=37096000, orientation='minus', exon_count=24, OMIM=116957, no_hdr0=''),
  5934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5934, CurrentID=0, Status='live', Symbol='RBL2', Aliases='P130, Rb2', description='retinoblastoma-like 2', other_designations='130 kDa retinoblastoma-associated protein|PRB2|RBR-2|retinoblastoma-related protein 2', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53434439, end_position_on_the_genomic_accession=53491648, orientation='plus', exon_count=25, OMIM=180203, no_hdr0=''),
  5935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5935, CurrentID=0, Status='live', Symbol='RBM3', Aliases='IS1-RNPL, RNPL', description='RNA binding motif (RNP1, RRM) protein 3', other_designations='RNA-binding motif protein 3|putative RNA-binding protein 3', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48574353, end_position_on_the_genomic_accession=48581165, orientation='plus', exon_count=7, OMIM=300027, no_hdr0=''),
  5936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5936, CurrentID=0, Status='live', Symbol='RBM4', Aliases='LARKA, ZCCHC21, ZCRB3A, RBM4', description='RNA binding motif protein 4', other_designations='RNA-binding motif protein 4a|lark homolog|transcriptional coactivator CoAZ|zinc finger CCHC-type and RNA binding motif 3A', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66638617, end_position_on_the_genomic_accession=66668385, orientation='plus', exon_count=5, OMIM=602571, no_hdr0=''),
  5937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5937, CurrentID=0, Status='live', Symbol='RBMS1', Aliases='C2orf12, HCC-4, MSSP, MSSP-1, MSSP-2, MSSP-3, SCR2, YC1', description='RNA binding motif, single stranded interacting protein 1', other_designations='RNA binding motif single stranded interacting protein 1|c-myc gene single strand binding protein 2|cervical cancer oncogene 4|single-stranded DNA-binding protein MSSP-1|suppressor of CDC2 with RNA-binding motif 2|suppressor of cdc 2 (cdc13) with RNA binding motif 2', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160272151, end_position_on_the_genomic_accession=160493856, orientation='minus', exon_count=17, OMIM=602310, no_hdr0=''),
  5938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5938, CurrentID=0, Status='live', Symbol='RBMS1P1', Aliases='MSSP1, RBMS1P', description='RNA binding motif, single stranded interacting protein 1 pseudogene 1', other_designations='RNA binding motif, single stranded interacting protein 1, pseudogene', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66233016, end_position_on_the_genomic_accession=66235226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5939, CurrentID=0, Status='live', Symbol='RBMS2', Aliases='SCR3', description='RNA binding motif, single stranded interacting protein 2', other_designations='suppressor of CDC2 with RNA-binding motif 3', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56521824, end_position_on_the_genomic_accession=56596196, orientation='plus', exon_count=18, OMIM=602387, no_hdr0=''),
  5940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5940, CurrentID=0, Status='live', Symbol='RBMY1A1', Aliases='RBM, RBM1, RBM2, RBMY, RBMY1C, YRRM1, YRRM2', description='RNA binding motif protein, Y-linked, family 1, member A1', other_designations='RNA binding motif protein, Y chromosome, family 1, member A1|RNA binding protein|RNA-binding motif protein 1|RNA-binding motif protein 2|RNA-binding motif protein, Y chromosome, family 1 member A1/C|y chromosome RNA recognition motif 1', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21534902, end_position_on_the_genomic_accession=21561014, orientation='plus', exon_count=15, OMIM=400006, no_hdr0=''),
  100271070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271070, CurrentID=0, Status='live', Symbol='RPS17P8', Aliases='RPS17_2_307', description='ribosomal protein S17 pseudogene 8', other_designations='', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=192774082, end_position_on_the_genomic_accession=192774424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5942, CurrentID=0, Status='live', Symbol='RBMY1C', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member C', other_designations='RNA binding motif protein, Y chromosome, family 1, member C', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5944, CurrentID=0, Status='live', Symbol='RBMY1HP', Aliases='RBMY1H, YRRM', description='RNA binding motif protein, Y-linked, family 1, member H, pseudogene', other_designations='Putative RNA-binding motif protein, Y chromosome, family 1 member H|RNA binding motif protein, Y chromosome, family 1, member H', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21493419, end_position_on_the_genomic_accession=21502337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  5945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5945, CurrentID=0, Status='live', Symbol='RBMY2AP', Aliases='RBMY2A', description='RNA binding motif protein, Y-linked, family 2, member A pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 2, member A', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21927115, end_position_on_the_genomic_accession=21939239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5947, CurrentID=0, Status='live', Symbol='RBP1', Aliases='CRABP-I, CRBP, CRBP1, CRBPI, RBPC', description='retinol binding protein 1', other_designations='CRBP-I|cellular retinol-binding protein I|retinol binding protein 1, cellular|retinol-binding protein 1, cellular', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139517434, end_position_on_the_genomic_accession=139539829, orientation='minus', exon_count=6, OMIM=180260, no_hdr0=''),
  5948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5948, CurrentID=0, Status='live', Symbol='RBP2', Aliases='CRABP-II, CRBP2, CRBPII, RBPC2', description='retinol binding protein 2', other_designations='CRBP-II|cellular retinol-binding protein II|retinol binding protein 2, cellular|retinol-binding protein 2, cellular', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139452884, end_position_on_the_genomic_accession=139476510, orientation='minus', exon_count=4, OMIM=180280, no_hdr0=''),
  5949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5949, CurrentID=0, Status='live', Symbol='RBP3', Aliases='D10S64, D10S65, D10S66, IRBP, RBPI, RP66', description='retinol binding protein 3', other_designations='interphotoreceptor retinoid-binding protein|interstitial retinol binding protein 3|retinol binding protein 3, interstitial', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47348371, end_position_on_the_genomic_accession=47357875, orientation='plus', exon_count=4, OMIM=180290, no_hdr0=''),
  5950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5950, CurrentID=0, Status='live', Symbol='RBP4', Aliases='MCOPCB10, RDCCAS', description='retinol binding protein 4', other_designations='PRBP|RBP|plasma retinol-binding protein|retinol binding protein 4, plasma|retinol-binding protein 4, interstitial', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93591836, end_position_on_the_genomic_accession=93601236, orientation='minus', exon_count=6, OMIM=180250, no_hdr0=''),
  5953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5953, CurrentID=0, Status='live', Symbol='RCD1', Aliases='', description='retinal cone dystrophy 1', other_designations='', map_location='6q25-q26', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180020, no_hdr0=''),
  5954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5954, CurrentID=0, Status='live', Symbol='RCN1', Aliases='HEL-S-84, PIG20, RCAL, RCN', description='reticulocalbin 1', other_designations='epididymis secretory protein Li 84|proliferation-inducing gene 20|reticulocalbin 1, EF-hand calcium binding domain', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32090931, end_position_on_the_genomic_accession=32105726, orientation='plus', exon_count=6, OMIM=602735, no_hdr0=''),
  5955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5955, CurrentID=0, Status='live', Symbol='RCN2', Aliases='E6BP, ERC-55, ERC55, TCBP49', description='reticulocalbin 2', other_designations='E6-binding protein|calcium-binding protein ERC-55|reticulocalbin 2, EF-hand calcium binding domain (endoplasmic reticulum calcium-binding protein, 55kD)', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76931621, end_position_on_the_genomic_accession=76950260, orientation='plus', exon_count=8, OMIM=602584, no_hdr0=''),
  5956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5956, CurrentID=0, Status='live', Symbol='OPN1LW', Aliases='CBBM, CBP, COD5, RCP, ROP', description='opsin 1 (cone pigments), long-wave-sensitive', other_designations='cone dystrophy 5 (X-linked)|red cone opsin|red cone photoreceptor pigment|red-sensitive opsin', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154144224, end_position_on_the_genomic_accession=154159032, orientation='plus', exon_count=6, OMIM=300822, no_hdr0=''),
  5957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5957, CurrentID=0, Status='live', Symbol='RCVRN', Aliases='RCV1', description='recoverin', other_designations='cancer associated retinopathy antigen|cancer-associated retinopathy protein', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9897708, end_position_on_the_genomic_accession=9905367, orientation='minus', exon_count=4, OMIM=179618, no_hdr0=''),
  5959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5959, CurrentID=0, Status='live', Symbol='RDH5', Aliases='9cRDH, HSD17B9, RDH1, SDR9C5', description='retinol dehydrogenase 5 (11-cis/9-cis)', other_designations='11-cis RDH|11-cis RoDH|9-cis retinol dehydrogenase|9-cis-retinol specific dehydrogenase|retinol dehydrogenase 1|retinol dehydrogenase 5 (11-cis and 9-cis)|short chain dehydrogenase/reductase family 9C member 5|short chain dehydrogenase/reductase family 9C, member 5', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55720367, end_position_on_the_genomic_accession=55724742, orientation='plus', exon_count=5, OMIM=601617, no_hdr0=''),
  5961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5961, CurrentID=0, Status='live', Symbol='PRPH2', Aliases='AOFMD, AVMD, CACD2, DS, MDBS1, PRPH, RDS, RP7, TSPAN22, rd2', description='peripherin 2 (retinal degeneration, slow)', other_designations='peripherin 2, homolog of mouse|peripherin, photoreceptor type|retinal degeneration slow protein|retinal peripherin|tetraspanin-22|tspan-22', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42696595, end_position_on_the_genomic_accession=42722989, orientation='minus', exon_count=5, OMIM=179605, no_hdr0=''),
  5962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5962, CurrentID=0, Status='live', Symbol='RDX', Aliases='DFNB24', description='radixin', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110174880, end_position_on_the_genomic_accession=110296712, orientation='minus', exon_count=17, OMIM=179410, no_hdr0=''),
  5964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5964, CurrentID=0, Status='live', Symbol='RDXP2', Aliases='', description='radixin pseudogene 2', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27515171, end_position_on_the_genomic_accession=27517883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5965, CurrentID=0, Status='live', Symbol='RECQL', Aliases='RECQL1, RecQ1', description='RecQ helicase-like', other_designations='DNA helicase, RecQ-like type 1|DNA-dependent ATPase Q1|RecQ protein-like (DNA helicase Q1-like)|recQ protein-like 1', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21468910, end_position_on_the_genomic_accession=21501669, orientation='minus', exon_count=15, OMIM=600537, no_hdr0=''),
  5966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5966, CurrentID=0, Status='live', Symbol='REL', Aliases='C-Rel', description='v-rel avian reticuloendotheliosis viral oncogene homolog', other_designations='oncogene REL, avian reticuloendotheliosis', map_location='2p13-p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60881495, end_position_on_the_genomic_accession=60928171, orientation='plus', exon_count=12, OMIM=164910, no_hdr0=''),
  5967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5967, CurrentID=0, Status='live', Symbol='REG1A', Aliases='ICRF, P19, PSP, PSPS, PSPS1, PTP, REG', description='regenerating family member 1 alpha', other_designations='REG-1-alpha|islet cells regeneration factor|islet of langerhans regenerating protein|lithostathine 1 alpha|pancreatic stone protein, secretory|pancreatic thread protein|protein-X|regenerating islet-derived protein 1-alpha|regenerating protein I alpha', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79120458, end_position_on_the_genomic_accession=79123419, orientation='plus', exon_count=6, OMIM=167770, no_hdr0=''),
  5968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5968, CurrentID=0, Status='live', Symbol='REG1B', Aliases='PSPS2, REGH, REGI-BETA, REGL', description='regenerating family member 1 beta', other_designations='PSP-2|REG-1-beta|lithostathine 1 beta|pancreatic stone protein 2|regenerating islet-derived 1 beta (pancreatic stone protein, pancreatic thread protein)|regenerating islet-derived protein 1-beta|regenerating protein I beta|secretory pancreatic stone protein 2', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79085023, end_position_on_the_genomic_accession=79088024, orientation='minus', exon_count=6, OMIM=167771, no_hdr0=''),
  5969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5969, CurrentID=0, Status='live', Symbol='REG1CP', Aliases='REG1P, REGL, RS', description='regenerating family member 1 gamma, pseudogene', other_designations='regenerating islet-derived 1 pseudogene|regenerating islet-derived-like, human homolog (pancreatic stone protein-like, pancreatic thread protein-like)', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79135503, end_position_on_the_genomic_accession=79138427, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5970, CurrentID=0, Status='live', Symbol='RELA', Aliases='NFKB3, p65', description='v-rel avian reticuloendotheliosis viral oncogene homolog A', other_designations='NF-kappa-B p65delta3|NF-kappa-B transcription factor p65|nuclear factor NF-kappa-B p65 subunit|nuclear factor of kappa light polypeptide gene enhancer in B-cells 3|v-rel reticuloendotheliosis viral oncogene homolog A', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65653596, end_position_on_the_genomic_accession=65662972, orientation='minus', exon_count=11, OMIM=164014, no_hdr0=''),
  5971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5971, CurrentID=0, Status='live', Symbol='RELB', Aliases='I-REL, IREL, REL-B', description='v-rel avian reticuloendotheliosis viral oncogene homolog B', other_designations='v-rel avian reticuloendotheliosis viral oncogene homolog B (nuclear factor of kappa light polypeptide gene enhancer in B-cells 3)|v-rel reticuloendotheliosis viral oncogene homolog B, nuclear factor of kappa light polypeptide gene enhancer in B-cells 3', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45001449, end_position_on_the_genomic_accession=45038198, orientation='plus', exon_count=12, OMIM=604758, no_hdr0=''),
  5972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5972, CurrentID=0, Status='live', Symbol='REN', Aliases='HNFJ2', description='renin', other_designations='angiotensin-forming enzyme|angiotensinogenase|renin precursor, renal', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204154816, end_position_on_the_genomic_accession=204166337, orientation='minus', exon_count=10, OMIM=179820, no_hdr0=''),
  5973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5973, CurrentID=0, Status='live', Symbol='RENBP', Aliases='RBP, RNBP', description='renin binding protein', other_designations='AGE|GlcNAc 2-epimerase|N-acetyl-D-glucosamine 2-epimerase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153935269, end_position_on_the_genomic_accession=153944780, orientation='minus', exon_count=11, OMIM=312420, no_hdr0=''),
  5975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5975, CurrentID=0, Status='live', Symbol='RENS2', Aliases='MRXS9', description='Renpenning syndrome 2', other_designations='Renpenning syndrome 2 (Renpenning-like syndrome)|Renpenning-like syndrome', map_location='Xq12-q21.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5976, CurrentID=0, Status='live', Symbol='UPF1', Aliases='HUPF1, NORF1, RENT1, pNORF1, smg-2', description='UPF1 regulator of nonsense transcripts homolog (yeast)', other_designations='ATP-dependent helicase RENT1|delta helicase|nonsense mRNA reducing factor 1|smg-2 homolog, nonsense mediated mRNA decay factor|up-frameshift mutation 1 homolog|up-frameshift suppressor 1 homolog|yeast Upf1p homolog', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18831935, end_position_on_the_genomic_accession=18868232, orientation='plus', exon_count=24, OMIM=601430, no_hdr0=''),
  5977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5977, CurrentID=0, Status='live', Symbol='DPF2', Aliases='REQ, UBID4, ubi-d4', description='D4, zinc and double PHD fingers family 2', other_designations='BAF45D|BRG1-associated factor 45D|apoptosis response zinc finger protein|protein requiem|requiem, apoptosis response zinc finger', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65333754, end_position_on_the_genomic_accession=65352980, orientation='plus', exon_count=12, OMIM=601671, no_hdr0=''),
  5978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5978, CurrentID=0, Status='live', Symbol='REST', Aliases='NRSF, XBR', description='RE1-silencing transcription factor', other_designations='RE1-silencing transcription factor variant E1a/E2/E3/E4c|RE1-silencing transcription factor variant E1a/E2/E3/E5|RE1-silencing transcription factor variant E1a/E2/E3/N3a/E4i|RE1-silencing transcription factor variant E1a/E2/E3/N3c/E4|RE1-silencing transcription factor variant E1a/E2/E4|RE1-silencing transcription factor variant E1a/E2/E5|RE1-silencing transcription factor variant E1a/E2a/E2k|RE1-silencing transcription factor variant E1a/E2d/E4g|RE1-silencing transcription factor variant E1a/E2e/E4h|RE1-silencing transcription factor variant E1a/E2f/E4e|RE1-silencing transcription factor variant E1a/E2k/E2i/E3/E4j|RE1-silencing transcription factor variant E1b/E2/E3/E5|RE1-silencing transcription factor variant E1b/E2/E3/N3b/E4i|RE1-silencing transcription factor variant E1b/E2/E3/N3c/E4|RE1-silencing transcription factor variant E1b/E2a/E2k|RE1-silencing transcription factor variant E1b/E2c/E2j/E3/E4|RE1-silencing transcription factor variant E1b/E2e/E4h|RE1-silencing transcription factor variant E1c/E2/E3/E5|RE1-silencing transcription factor variant E1c/E2a/E2k|RE1-silencing transcription factor variant E1c/E2g/E3/E4|neural-restrictive silencer factor|neuron restrictive silencer factor|repressor binding to the X2 box', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56907876, end_position_on_the_genomic_accession=56935845, orientation='plus', exon_count=8, OMIM=600571, no_hdr0=''),
  5979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5979, CurrentID=0, Status='live', Symbol='RET', Aliases='CDHF12, CDHR16, HSCR1, MEN2A, MEN2B, MTC1, PTC-ELE1, RET51, RET', description='ret proto-oncogene', other_designations='CUX1/RET fusion|RET receptor tyrosine kinase|RET transforming sequence|cadherin family member 12|cadherin-related family member 16|hydroxyaryl-protein kinase|proto-oncogene c-Ret|rearranged during transfection|receptor tyrosine kinase|ret proto-oncogene (multiple endocrine neoplasia and medullary thyroid carcinoma 1, Hirschsprung disease)', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43077027, end_position_on_the_genomic_accession=43130351, orientation='plus', exon_count=21, OMIM=164761, no_hdr0=''),
  5980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5980, CurrentID=0, Status='live', Symbol='REV3L', Aliases='POLZ, REV3', description='REV3 like, DNA directed polymerase zeta catalytic subunit', other_designations='REV3-like, polymerase (DNA directed), zeta, catalytic subunit|Rev-3, yeast, homolog-like (polymerase, DNA, zeta)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111299031, end_position_on_the_genomic_accession=111483715, orientation='minus', exon_count=40, OMIM=602776, no_hdr0=''),
  5981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5981, CurrentID=0, Status='live', Symbol='RFC1', Aliases='A1, MHCBFB, PO-GA, RECC1, RFC40, RFC1', description='replication factor C subunit 1', other_designations='A1 140 kDa subunit|DNA-binding protein PO-GA|MHC binding factor, beta|RF-C 140 kDa subunit|activator 1 140 kDa subunit|activator 1 large subunit|activator 1 subunit 1|replication factor C (activator 1) 1, 145kDa|replication factor C 140 kDa subunit|replication factor C large subunit|replication factor C1', map_location='4p14-p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39287449, end_position_on_the_genomic_accession=39366381, orientation='minus', exon_count=25, OMIM=102579, no_hdr0=''),
  5982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5982, CurrentID=0, Status='live', Symbol='RFC2', Aliases='RFC40', description='replication factor C subunit 2', other_designations='A1 40 kDa subunit|RF-C 40 kDa subunit|activator 1 subunit 2|replication factor C (activator 1) 2, 40kDa', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74231485, end_position_on_the_genomic_accession=74254458, orientation='minus', exon_count=11, OMIM=600404, no_hdr0=''),
  5983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5983, CurrentID=0, Status='live', Symbol='RFC3', Aliases='RFC38', description='replication factor C subunit 3', other_designations='A1 38 kDa subunit|RF-C 38 kDa subunit|RFC, 38 kD subunit|activator 1 38 kDa subunit|activator 1 subunit 3|replication factor C (activator 1) 3, 38kDa|replication factor C 38 kDa subunit', map_location='13q13.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33818049, end_position_on_the_genomic_accession=33973946, orientation='plus', exon_count=15, OMIM=600405, no_hdr0=''),
  5984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5984, CurrentID=0, Status='live', Symbol='RFC4', Aliases='A1, RFC37', description='replication factor C subunit 4', other_designations='A1 37 kDa subunit|RF-C 37 kDa subunit|RFC 37 kDa subunit|activator 1 37 kDa subunit|activator 1 subunit 4|replication factor C (activator 1) 4, 37kDa|replication factor C 37 kDa subunit', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186789892, end_position_on_the_genomic_accession=186806695, orientation='minus', exon_count=11, OMIM=102577, no_hdr0=''),
  5985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5985, CurrentID=0, Status='live', Symbol='RFC5', Aliases='RFC36', description='replication factor C subunit 5', other_designations='A1 36 kDa subunit|RF-C 36 kDa subunit|RFC, 36.5 kD subunit|replication factor C (activator 1) 5, 36.5kDa|replication factor C, 36-kDa subunit', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118016701, end_position_on_the_genomic_accession=118033564, orientation='plus', exon_count=14, OMIM=600407, no_hdr0=''),
  5986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5986, CurrentID=0, Status='live', Symbol='RFNG', Aliases='', description='RFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase|radical fringe homolog', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82047902, end_position_on_the_genomic_accession=82051774, orientation='minus', exon_count=9, OMIM=602578, no_hdr0=''),
  5987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5987, CurrentID=0, Status='live', Symbol='TRIM27', Aliases='RFP, RNF76', description='tripartite motif containing 27', other_designations='RFP transforming protein|RING finger protein 76|ret finger protein|tripartite motif protein TRIM27|tripartite motif-containing 27|tripartite motif-containing protein 27', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28903002, end_position_on_the_genomic_accession=28923991, orientation='minus', exon_count=8, OMIM=602165, no_hdr0=''),
  5988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5988, CurrentID=0, Status='live', Symbol='RFPL1', Aliases='RNF78', description='ret finger protein-like 1', other_designations='RING finger protein 78', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29434530, end_position_on_the_genomic_accession=29442455, orientation='plus', exon_count=5, OMIM=605968, no_hdr0=''),
  5989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5989, CurrentID=0, Status='live', Symbol='RFX1', Aliases='EFC, RFX', description='regulatory factor X1', other_designations='MHC class II regulatory factor RFX|enhancer factor C|regulatory factor X 1|regulatory factor X, 1 (influences HLA class II expression)|trans-acting regulatory factor 1|transcription factor RFX1', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13961530, end_position_on_the_genomic_accession=14007514, orientation='minus', exon_count=24, OMIM=600006, no_hdr0=''),
  5990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5990, CurrentID=0, Status='live', Symbol='RFX2', Aliases='', description='regulatory factor X2', other_designations='DNA binding protein RFX2|HLA class II regulatory factor RFX2|regulatory factor X 2|regulatory factor X, 2 (influences HLA class II expression)|trans-acting regulatory factor 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5993164, end_position_on_the_genomic_accession=6110653, orientation='minus', exon_count=19, OMIM=142765, no_hdr0=''),
  5991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5991, CurrentID=0, Status='live', Symbol='RFX3', Aliases='', description='regulatory factor X3', other_designations='DNA binding protein RFX3|regulatory factor X, 3 (influences HLA class II expression)', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3218297, end_position_on_the_genomic_accession=3526001, orientation='minus', exon_count=26, OMIM=601337, no_hdr0=''),
  5992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5992, CurrentID=0, Status='live', Symbol='RFX4', Aliases='NYD-SP10', description='regulatory factor X4', other_designations='regulatory factor X, 4 (influences HLA class II expression)|testis development protein NYD-SP10|winged-helix transcription factor RFX4', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106582907, end_position_on_the_genomic_accession=106762804, orientation='plus', exon_count=20, OMIM=603958, no_hdr0=''),
  5993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5993, CurrentID=0, Status='live', Symbol='RFX5', Aliases='', description='regulatory factor X5', other_designations='regulatory factor X 5|regulatory factor X, 5 (influences HLA class II expression)', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151340640, end_position_on_the_genomic_accession=151347293, orientation='minus', exon_count=12, OMIM=601863, no_hdr0=''),
  5994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5994, CurrentID=0, Status='live', Symbol='RFXAP', Aliases='', description='regulatory factor X associated protein', other_designations='RFX DNA-binding complex 36 kDa subunit|RFX-associated protein', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36819202, end_position_on_the_genomic_accession=36829603, orientation='plus', exon_count=3, OMIM=601861, no_hdr0=''),
  5995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5995, CurrentID=0, Status='live', Symbol='RGR', Aliases='RP44', description='retinal G protein coupled receptor', other_designations='RGR-opsin|RPE retinal G-protein coupled receptor', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84237083, end_position_on_the_genomic_accession=84259347, orientation='plus', exon_count=9, OMIM=600342, no_hdr0=''),
  5996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5996, CurrentID=0, Status='live', Symbol='RGS1', Aliases='1R20, BL34, HEL-S-87, IER1, IR20', description='regulator of G-protein signaling 1', other_designations='B-cell activation protein BL34|early response protein 1R20|epididymis secretory protein Li 87|immediate-early response 1, B-cell specific|regulator of G-protein signalling 1', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192575727, end_position_on_the_genomic_accession=192580029, orientation='plus', exon_count=5, OMIM=600323, no_hdr0=''),
  5997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5997, CurrentID=0, Status='live', Symbol='RGS2', Aliases='G0S8', description='regulator of G-protein signaling 2', other_designations='G0 to G1 switch regulatory 8, 24kD|G0/G1 switch regulatory protein 8|cell growth-inhibiting gene 31 protein|cell growth-inhibiting protein 31|regulator of G-protein signaling 2, 24kDa', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192809039, end_position_on_the_genomic_accession=192812277, orientation='plus', exon_count=5, OMIM=600861, no_hdr0=''),
  5998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5998, CurrentID=0, Status='live', Symbol='RGS3', Aliases='C2PA, RGP3', description='regulator of G-protein signaling 3', other_designations='regulator of G-protein signalling 3', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113444729, end_position_on_the_genomic_accession=113597743, orientation='plus', exon_count=36, OMIM=602189, no_hdr0=''),
  5999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5999, CurrentID=0, Status='live', Symbol='RGS4', Aliases='RGP4, SCZD9', description='regulator of G-protein signaling 4', other_designations='schizophrenia disorder 9', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163068606, end_position_on_the_genomic_accession=163076802, orientation='plus', exon_count=7, OMIM=602516, no_hdr0=''),
  6000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6000, CurrentID=0, Status='live', Symbol='RGS7', Aliases='', description='regulator of G-protein signaling 7', other_designations='regulator of G-protein signaling RGS7|regulator of G-protein signalling 7', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240775514, end_position_on_the_genomic_accession=241357230, orientation='minus', exon_count=23, OMIM=602517, no_hdr0=''),
  6001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6001, CurrentID=0, Status='live', Symbol='RGS10', Aliases='', description='regulator of G-protein signaling 10', other_designations='regulator of G-protein signalling 10', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119499827, end_position_on_the_genomic_accession=119542710, orientation='minus', exon_count=6, OMIM=602856, no_hdr0=''),
  6002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6002, CurrentID=0, Status='live', Symbol='RGS12', Aliases='', description='regulator of G-protein signaling 12', other_designations='regulator of G-protein signalling 12', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3292978, end_position_on_the_genomic_accession=3439913, orientation='plus', exon_count=21, OMIM=602512, no_hdr0=''),
  6003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6003, CurrentID=0, Status='live', Symbol='RGS13', Aliases='', description='regulator of G-protein signaling 13', other_designations='regulator of G-protein signalling 13', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192636138, end_position_on_the_genomic_accession=192660311, orientation='plus', exon_count=7, OMIM=607190, no_hdr0=''),
  6004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6004, CurrentID=0, Status='live', Symbol='RGS16', Aliases='A28-RGS14, A28-RGS14P, RGS-R', description='regulator of G-protein signaling 16', other_designations='hRGS-r|regulator of G-protein signalling 16|retinal-specific RGS|retinally abundant regulator of G-protein signaling', map_location='1q25-q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182598623, end_position_on_the_genomic_accession=182604413, orientation='minus', exon_count=5, OMIM=602514, no_hdr0=''),
  6005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6005, CurrentID=0, Status='live', Symbol='RHAG', Aliases='CD241, OHS, RH2, RH50A, Rh50, Rh50GP, SLC42A1', description='Rh-associated glycoprotein', other_designations='Rhesus associated polypeptide, 50-KD|Rhesus blood group-associated glycoprotein|erythrocyte membrane glycoprotein Rh50|erythrocyte plasma membrane 50 kDa glycoprotein|rh family type A glycoprotein|rh type A glycoprotein|rhesus blood group family type A glycoprotein|rhesus blood group-associated ammonia channel', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49605175, end_position_on_the_genomic_accession=49636874, orientation='minus', exon_count=11, OMIM=180297, no_hdr0=''),
  6006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6006, CurrentID=0, Status='live', Symbol='RHCE', Aliases='CD240CE, RH, RH30A, RHC, RHE, RHIXB, RHPI, Rh4, RhIVb(J), RhVI, RhVIII', description='Rh blood group, CcEe antigens', other_designations='(C)ces type 1 Rhesus blood group D antigen|RHCE blood group variant Crawford antigen Rh43|Rh blood group C antigen|Rh blood group antigen Evans|Rh blood group protein|Rh polypeptide I|Rhesus blood group CE protein|Rhesus blood group E antigen|Rhesus blood group Rhce antigen|Rhesus system C and E polypeptides|blood group RhCcEe antigen|rh polypeptide 1|rhesus C/E antigens|rhesus blood group antigen, RhC antigen|rhesus blood group little e antigen|silenced Rh blood group CcEe antigen|truncated RHCE', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25360659, end_position_on_the_genomic_accession=25430192, orientation='minus', exon_count=14, OMIM=111700, no_hdr0=''),
  6007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6007, CurrentID=0, Status='live', Symbol='RHD', Aliases='CD240D, DIIIc, RH, RH30, RHCEDVA(TT), RHDel, RHPII, RHXIII, Rh4, RhDCw, RhII, RhK562-II, RhPI, RHD', description='Rh blood group, D antigen', other_designations='D antigen (DCS)|RH polypeptide 2|Rh blood group antigen Evans|Rh blood group, D anitgen|RhD blood group antigen|RhD polypeptide|Rhesus blood group D antigen allele DIII type 7|Rhesus system D polypeptide|blood group antigen D|rhesus D antigen|truncated rhesus blood group D antigen', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25272393, end_position_on_the_genomic_accession=25330445, orientation='plus', exon_count=13, OMIM=111680, no_hdr0=''),
  6008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6008, CurrentID=0, Status='live', Symbol='RHEBP1', Aliases='RHEB, RHEB1', description='Ras-homolog enriched in brain pseudogene 1', other_designations='Ras homolog enriched in brain 1|Ras homolog enriched in brain pseudogene 1', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46634502, end_position_on_the_genomic_accession=46635846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6009, CurrentID=0, Status='live', Symbol='RHEB', Aliases='RHEB2', description='Ras homolog enriched in brain', other_designations='Ras homolog enriched in brain 2', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151466012, end_position_on_the_genomic_accession=151519924, orientation='minus', exon_count=9, OMIM=601293, no_hdr0=''),
  6010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6010, CurrentID=0, Status='live', Symbol='RHO', Aliases='CSNBAD1, OPN2, RP4', description='rhodopsin', other_designations='opsin 2, rod pigment|opsin-2', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129528639, end_position_on_the_genomic_accession=129535344, orientation='plus', exon_count=5, OMIM=180380, no_hdr0=''),
  6011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6011, CurrentID=0, Status='live', Symbol='GRK1', Aliases='GPRK1, RHOK, RK', description='G protein-coupled receptor kinase 1', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113723073, end_position_on_the_genomic_accession=113735664, orientation='plus', exon_count=7, OMIM=180381, no_hdr0=''),
  6012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6012, CurrentID=0, Status='live', Symbol='RIEG2', Aliases='ARS, RGS2', description='Rieger syndrome 2', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601499, no_hdr0=''),
  6013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6013, CurrentID=0, Status='live', Symbol='RLN1', Aliases='H1, H1RLX, RLXH1, bA12D24.3.1, bA12D24.3.2', description='relaxin 1', other_designations='preprorelaxin H1', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5313627, end_position_on_the_genomic_accession=5339873, orientation='minus', exon_count=4, OMIM=179730, no_hdr0=''),
  6014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6014, CurrentID=0, Status='live', Symbol='RIT2', Aliases='RIBA, RIN, ROC2', description='Ras-like without CAAX 2', other_designations='GTP-binding protein Roc2|Ric-like, expressed in neurons|ras-like protein expressed in neurons|ras-like without CAAX protein 2', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=42743218, end_position_on_the_genomic_accession=43115692, orientation='minus', exon_count=6, OMIM=609592, no_hdr0=''),
  6015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6015, CurrentID=0, Status='live', Symbol='RING1', Aliases='RING1A, RNF1', description='ring finger protein 1', other_designations='polycomb complex protein RING1|really interesting new gene 1 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33208509, end_position_on_the_genomic_accession=33212722, orientation='plus', exon_count=7, OMIM=602045, no_hdr0=''),
  6016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6016, CurrentID=0, Status='live', Symbol='RIT1', Aliases='NS8, RIBB, RIT, ROC1', description='Ras-like without CAAX 1', other_designations='GTP-binding protein Roc1|Ric-like, expressed in many tissues|ras-like protein expressed in many tissues|ras-like without CAAX protein 1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155897808, end_position_on_the_genomic_accession=155911402, orientation='minus', exon_count=7, OMIM=609591, no_hdr0=''),
  6017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6017, CurrentID=0, Status='live', Symbol='RLBP1', Aliases='CRALBP', description='retinaldehyde binding protein 1', other_designations='cellular retinaldehyde-binding protein|cellular retinaldehyde-binding protein-1', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89209867, end_position_on_the_genomic_accession=89221909, orientation='minus', exon_count=10, OMIM=180090, no_hdr0=''),
  6018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6018, CurrentID=0, Status='live', Symbol='RLF', Aliases='ZN-15L, ZNF292L', description='rearranged L-myc fusion', other_designations='Zn-15 related|rearranged L-myc fusion gene protein|rearranged L-myc fusion sequence|zn-15-related protein', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40161369, end_position_on_the_genomic_accession=40240921, orientation='plus', exon_count=9, OMIM=180610, no_hdr0=''),
  6019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6019, CurrentID=0, Status='live', Symbol='RLN2', Aliases='H2, H2-RLX, RLXH2, bA12D24.1.1, bA12D24.1.2', description='relaxin 2', other_designations='H2-preprorelaxin|relaxin, ovarian, of pregnancy', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5299864, end_position_on_the_genomic_accession=5339412, orientation='minus', exon_count=6, OMIM=179740, no_hdr0=''),
  101927511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927511, CurrentID=0, Status='live', Symbol='LOC101927511', Aliases='', description='uncharacterized LOC101927511', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11910412, end_position_on_the_genomic_accession=11948090, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  6022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6022, CurrentID=0, Status='live', Symbol='RMD1', Aliases='', description='rippling muscle disease 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600332, no_hdr0=''),
  6023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6023, CurrentID=0, Status='live', Symbol='RMRP', Aliases='CHH, NME1R, RRP2, RMRP', description='RNA component of mitochondrial RNA processing endoribonuclease', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35657751, end_position_on_the_genomic_accession=35658018, orientation='minus', exon_count=1, OMIM=157660, no_hdr0=''),
  6025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6025, CurrentID=0, Status='live', Symbol='RN5S1@', Aliases='RN5S', description='RNA, 5S ribosomal 1q42 cluster', other_designations='RNA, 5S cluster 1', map_location='1q42.11-q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180420, no_hdr0=''),
  6029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6029, CurrentID=0, Status='live', Symbol='RN7SL1', Aliases='7L1a, 7SL, RN7SL, RNSRP1', description='RNA, 7SL, cytoplasmic 1', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49586580, end_position_on_the_genomic_accession=49586878, orientation='plus', exon_count=1, OMIM=612177, no_hdr0=''),
  6030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6030, CurrentID=0, Status='live', Symbol='RN7SL4P', Aliases='7L28, RN7SL427P, RN7SLP1', description='RNA, 7SL, cytoplasmic 4, pseudogene', other_designations='RNA, 7SL, cytoplasmic 427, pseudogene|RNA, 7SL, cytoplasmic, pseudogene 1', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15738515, end_position_on_the_genomic_accession=15738786, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6031, CurrentID=0, Status='live', Symbol='RN7SL5P', Aliases='7LEM1, RN7SL249P, RN7SLP2', description='RNA, 7SL, cytoplasmic 5, pseudogene', other_designations='RNA, 7SL, cytoplasmic 249, pseudogene|RNA, 7SL, cytoplasmic, pseudogene 2', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=9441998, end_position_on_the_genomic_accession=9442310, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6032, CurrentID=0, Status='live', Symbol='RN7SL6P', Aliases='7L7, RN7SL848P, RN7SLP3', description='RNA, 7SL, cytoplasmic 6, pseudogene', other_designations='RNA, 7SL, cytoplasmic 848, pseudogene|RNA, 7SL, cytoplasmic, pseudogene 3', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23132704, end_position_on_the_genomic_accession=23133048, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6033, CurrentID=0, Status='live', Symbol='RN7SL7P', Aliases='7L23, RN7SL27P, RN7SLP4', description='RNA, 7SL, cytoplasmic 7, pseudogene', other_designations='RNA, 7SL, cytoplasmic 27, pseudogene|RNA, 7SL, cytoplasmic, pseudogene 4', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92970942, end_position_on_the_genomic_accession=92971326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6034, CurrentID=0, Status='live', Symbol='RN7SL8P', Aliases='7L63, RN7SL433P, RN7SLP5', description='RNA, 7SL, cytoplasmic 8, pseudogene', other_designations='RNA, 7SL, cytoplasmic 433, pseudogene|RNA, 7SL, cytoplasmic, pseudogene 5', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104911873, end_position_on_the_genomic_accession=104912145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6035, CurrentID=0, Status='live', Symbol='RNASE1', Aliases='RAC1, RIB1, RNS1', description='ribonuclease, RNase A family, 1 (pancreatic)', other_designations='HP-RNase|RIB-1|RNase 1|RNase A|RNase upI-1|ribonuclease 1|ribonuclease A C1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20801356, end_position_on_the_genomic_accession=20802877, orientation='minus', exon_count=3, OMIM=180440, no_hdr0=''),
  6036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6036, CurrentID=0, Status='live', Symbol='RNASE2', Aliases='EDN, RAF3, RNS2', description='ribonuclease, RNase A family, 2 (liver, eosinophil-derived neurotoxin)', other_designations='RNase 2|RNase UpI-2|eosinophil-derived neurotoxin|ribonuclease 2|ribonuclease A F3|ribonuclease US', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20955471, end_position_on_the_genomic_accession=20956435, orientation='plus', exon_count=2, OMIM=131410, no_hdr0=''),
  6037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6037, CurrentID=0, Status='live', Symbol='RNASE3', Aliases='ECP, RNS3', description='ribonuclease, RNase A family, 3', other_designations='RNase 3|cytotoxic ribonuclease|ribonuclease 3', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20891403, end_position_on_the_genomic_accession=20892348, orientation='plus', exon_count=2, OMIM=131398, no_hdr0=''),
  6038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6038, CurrentID=0, Status='live', Symbol='RNASE4', Aliases='RAB1, RNS4', description='ribonuclease, RNase A family, 4', other_designations='RNase 4|ribonuclease A B1', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20684213, end_position_on_the_genomic_accession=20700602, orientation='plus', exon_count=4, OMIM=601030, no_hdr0=''),
  6039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6039, CurrentID=0, Status='live', Symbol='RNASE6', Aliases='RAD1, RNS6, RNasek6', description='ribonuclease, RNase A family, k6', other_designations='ribonuclease A D1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20781051, end_position_on_the_genomic_accession=20782467, orientation='plus', exon_count=2, OMIM=601981, no_hdr0=''),
  6041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6041, CurrentID=0, Status='live', Symbol='RNASEL', Aliases='PRCA1, RNS4', description="ribonuclease L (2',5'-oligoisoadenylate synthetase-dependent)", other_designations="2',5'-oligoisoadenylate synthetase-dependent|2-5A-dependent RNase|RNase L|interferon-induced 2-5A-dependent RNase|ribonuclease 4", map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182573634, end_position_on_the_genomic_accession=182589285, orientation='minus', exon_count=8, OMIM=180435, no_hdr0=''),
  6043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6043, CurrentID=0, Status='live', Symbol='SNORA63', Aliases='E3, E3-2, RNE3, RNU107, elF-4AII', description='small nucleolar RNA, H/ACA box 63', other_designations='RNA, U107 small nucleolar|RNA, small nucleolar E3', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186787299, end_position_on_the_genomic_accession=186787433, orientation='plus', exon_count=1, OMIM=180647, no_hdr0=''),
  6044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6044, CurrentID=0, Status='live', Symbol='SNORA62', Aliases='E2, E2-1, RNE2, RNU108', description='small nucleolar RNA, H/ACA box 62', other_designations='RNA, U108 small nucleolar|RNA, small nucleolar E2', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39411054, end_position_on_the_genomic_accession=39411207, orientation='plus', exon_count=1, OMIM=180646, no_hdr0=''),
  6045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6045, CurrentID=0, Status='live', Symbol='RNF2', Aliases='BAP-1, BAP1, DING, HIPI3, RING1B, RING2', description='ring finger protein 2', other_designations='HIP2-interacting protein 3|RING finger protein 1B|RING finger protein BAP-1|huntingtin-interacting protein 2-interacting protein 3|protein DinG', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185045419, end_position_on_the_genomic_accession=185102608, orientation='plus', exon_count=9, OMIM=608985, no_hdr0=''),
  6046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6046, CurrentID=0, Status='live', Symbol='BRD2', Aliases='D6S113E, FSH, FSRG1, NAT, O27.1.1, RING3, RNF3', description='bromodomain containing 2', other_designations='female sterile homeotic-related gene 1|really interesting new gene 3 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32966705, end_position_on_the_genomic_accession=32972249, orientation='plus', exon_count=19, OMIM=601540, no_hdr0=''),
  6047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6047, CurrentID=0, Status='live', Symbol='RNF4', Aliases='RES4-26, SLX5, SNURF', description='ring finger protein 4', other_designations='E3 ubiquitin ligase RNF4|small nuclear RING finger protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2469068, end_position_on_the_genomic_accession=2515859, orientation='plus', exon_count=10, OMIM=602850, no_hdr0=''),
  6048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6048, CurrentID=0, Status='live', Symbol='RNF5', Aliases='RING5, RMA1', description='ring finger protein 5, E3 ubiquitin protein ligase', other_designations='protein G16|ram1 homolog', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32178385, end_position_on_the_genomic_accession=32180793, orientation='plus', exon_count=6, OMIM=602677, no_hdr0=''),
  6049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6049, CurrentID=0, Status='live', Symbol='RNF6', Aliases='', description='ring finger protein (C3H2C3 type) 6', other_designations='RING-H2 protein RNF-6', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26132115, end_position_on_the_genomic_accession=26222559, orientation='minus', exon_count=7, OMIM=604242, no_hdr0=''),
  6050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6050, CurrentID=0, Status='live', Symbol='RNH1', Aliases='RAI, RNH', description='ribonuclease/angiogenin inhibitor 1', other_designations='placental RNase inhibitor|placental ribonuclease inhibitor', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=494512, end_position_on_the_genomic_accession=507283, orientation='minus', exon_count=14, OMIM=173320, no_hdr0=''),
  6051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6051, CurrentID=0, Status='live', Symbol='RNPEP', Aliases='', description='arginyl aminopeptidase (aminopeptidase B)', other_designations='AP-B|arginine aminopeptidase', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201982638, end_position_on_the_genomic_accession=202006147, orientation='plus', exon_count=12, OMIM=602675, no_hdr0=''),
  6052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6052, CurrentID=0, Status='live', Symbol='RNR1', Aliases='', description='RNA, ribosomal cluster 1', other_designations='RNA, ribosomal 1|Ribosomal RNA-1', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180450, no_hdr0=''),
  6053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6053, CurrentID=0, Status='live', Symbol='RNR2', Aliases='', description='RNA, ribosomal cluster 2', other_designations='RNA, ribosomal 2|Ribosomal RNA-2', map_location='14p12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180451, no_hdr0=''),
  6054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6054, CurrentID=0, Status='live', Symbol='RNR3', Aliases='', description='RNA, ribosomal cluster 3', other_designations='RNA, ribosomal 3|Ribosomal RNA-3', map_location='15p12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180452, no_hdr0=''),
  6055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6055, CurrentID=0, Status='live', Symbol='RNR4', Aliases='', description='RNA, ribosomal cluster 4', other_designations='RNA, ribosomal 4|Ribosomal RNA-4', map_location='21p12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180453, no_hdr0=''),
  6056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6056, CurrentID=0, Status='live', Symbol='RNR5', Aliases='', description='RNA, ribosomal cluster 5', other_designations='RNA, ribosomal 5|Ribosomal RNA-5', map_location='22p12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180454, no_hdr0=''),
  6059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6059, CurrentID=0, Status='live', Symbol='ABCE1', Aliases='ABC38, OABP, RLI, RNASEL1, RNASELI, RNS4I', description='ATP binding cassette subfamily E member 1', other_designations="2'-5'-oligoadenylate-binding protein|ATP-binding cassette, sub-family E (OABP), member 1|RNase L inhibitor|huHP68|ribonuclease 4 inhibitor|ribonuclease L (2',5'-oligoisoadenylate synthetase-dependent) inhibitor", map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145098004, end_position_on_the_genomic_accession=145129524, orientation='plus', exon_count=18, OMIM=601213, no_hdr0=''),
  6060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6060, CurrentID=0, Status='live', Symbol='RNU1-4', Aliases='HSD2, HSD5, HSD6, HSD7, RNU1, RNU1D2, RNU1E2, RNU1F1, RNU1G1, RNU1G2, U1B2, U1D2, U1E2, U1F', description='RNA, U1 small nuclear 4', other_designations='RNA, U1D2 small nuclear|RNA, U1E2 small nuclear|RNA, U1F1 small nuclear|RNA, U1G1 small nuclear|RNA, U1G2 small nuclear', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16740516, end_position_on_the_genomic_accession=16740679, orientation='plus', exon_count=1, OMIM=180680, no_hdr0=''),
  137133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137133, CurrentID=0, Status='live', Symbol='PSAT1P1', Aliases='C8orf62', description='phosphoserine aminotransferase 1 pseudogene 1', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49739196, end_position_on_the_genomic_accession=49741406, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6063, CurrentID=0, Status='live', Symbol='RNVU1-11', Aliases='RNU1-12P, RNU1P2, U1P2A, vU1.11', description='RNA, variant U1 small nuclear 11', other_designations='RNA, U1 small nuclear 12 (pseudogene)|RNA, U1 small nuclear pseudogene 2', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144551671, end_position_on_the_genomic_accession=144552339, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285000, CurrentID=0, Status='live', Symbol='LOC285000', Aliases='', description='uncharacterized LOC285000', other_designations='', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105593097, end_position_on_the_genomic_accession=105610559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  6066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6066, CurrentID=0, Status='live', Symbol='RNU2-1', Aliases='RNU2, U2', description='RNA, U2 small nuclear 1', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43233790, end_position_on_the_genomic_accession=43233977, orientation='minus', exon_count=1, OMIM=180690, no_hdr0=''),
  100271091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271091, CurrentID=0, Status='live', Symbol='RPS23P3', Aliases='RPS23_1_489', description='ribosomal protein S23 pseudogene 3', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=66431092, end_position_on_the_genomic_accession=66431521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6068, CurrentID=0, Status='live', Symbol='RNU3P1', Aliases='U3', description='RNA, U3 small nucleolar pseudogene 1', other_designations='RNA, U3 small nuclear pseudogene 1', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58760466, end_position_on_the_genomic_accession=58760681, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6069, CurrentID=0, Status='live', Symbol='RNU4-7P', Aliases='RNU4P1, U4, U4/7', description='RNA, U4 small nuclear 7, pseudogene', other_designations='RNA, U4 small nuclear 7 (pseudogene)|RNA, U4 small nuclear pseudogene 1 (U4/7)', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150326564, end_position_on_the_genomic_accession=150327063, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6070, CurrentID=0, Status='live', Symbol='RNU4-8P', Aliases='RNU4P2, U4, U4/14', description='RNA, U4 small nuclear 8, pseudogene', other_designations='RNA, U4 small nuclear 8 (pseudogene)|RNA, U4 small nuclear pseudogene 2 (U4/14)', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97664292, end_position_on_the_genomic_accession=97664970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6071, CurrentID=0, Status='live', Symbol='RNU6V', Aliases='87U6, LH87', description='RNA, U6 small nuclear variant sequence with SNRPE pseudogene sequence', other_designations='small nuclear ribonucleoprotein E pseudogene', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109591305, end_position_on_the_genomic_accession=109591640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6072, CurrentID=0, Status='live', Symbol='RNU7-68P', Aliases='HSU7.732A, RNU7P1, U7, U7.68', description='RNA, U7 small nuclear 68 pseudogene', other_designations='RNA, U7 small nuclear pseudogene 1|RNA, small nuclear U7 pseudogene 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120503041, end_position_on_the_genomic_accession=120503097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6073, CurrentID=0, Status='live', Symbol='RNU7-86P', Aliases='HSU7.7/32B, RNU7P2, U7', description='RNA, U7 small nuclear 86 pseudogene', other_designations='RNA, U7 small nuclear pseudogene 2|RNA, small nuclear U7 pseudogene 2', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6074, CurrentID=0, Status='live', Symbol='RNU7-24P', Aliases='HSU7.24, HSU7.25, RNU7P3, U7, U7.24', description='RNA, U7 small nuclear 24 pseudogene', other_designations='RNA, U7 small nuclear pseudogene 3|RNA, small nuclear U7 pseudogene 3|RNA, small nuclear U7.24 pseudogene', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24400062, end_position_on_the_genomic_accession=24400116, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6075, CurrentID=0, Status='live', Symbol='RNU7-55P', Aliases='HSU7.36, HSU7.55, RNU7P4, U7, U7.55', description='RNA, U7 small nuclear 55 pseudogene', other_designations='RNA, U7 small nuclear pseudogene 4|RNA, small nuclear U7 pseudogene 4', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9072456, end_position_on_the_genomic_accession=9072514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6076, CurrentID=0, Status='live', Symbol='SNORD13P1', Aliases='RNU13P1, U13, U13.32A', description='small nucleolar RNA, C/D box 13 pseudogene 1', other_designations='RNA, U13 small nuclear pseudogene 1', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42076059, end_position_on_the_genomic_accession=42076161, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6077, CurrentID=0, Status='live', Symbol='SNORD13P2', Aliases='RNU13P2, U13, U13.12A', description='small nucleolar RNA, C/D box 13 pseudogene 2', other_designations='RNA, U13 small nuclear pseudogene 2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4689429, end_position_on_the_genomic_accession=4689633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6078, CurrentID=0, Status='live', Symbol='SNORD13P3', Aliases='RNU13P3, U13, U13.4B', description='small nucleolar RNA, C/D box 13 pseudogene 3', other_designations='RNA, U13 small nuclear pseudogene 3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47250525, end_position_on_the_genomic_accession=47250626, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6079, CurrentID=0, Status='live', Symbol='SNORD15A', Aliases='RNU15A, SNORNA, U15A', description='small nucleolar RNA, C/D box 15A', other_designations='U15A snRNA|U15A snoRNA', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75400391, end_position_on_the_genomic_accession=75400538, orientation='plus', exon_count=1, OMIM=600455, no_hdr0=''),
  6080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6080, CurrentID=0, Status='live', Symbol='SNORA73A', Aliases='E1, E1-7, E1b, RNE1, RNU17A, U17A', description='small nucleolar RNA, H/ACA box 73A', other_designations='RNA, U17a small nucleolar|RNA, small nucleolar E1', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28507365, end_position_on_the_genomic_accession=28507571, orientation='plus', exon_count=1, OMIM=180645, no_hdr0=''),
  6082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6082, CurrentID=0, Status='live', Symbol='SNORD20', Aliases='RNU20, U20', description='small nucleolar RNA, C/D box 20', other_designations='RNA, U20 small nuclear|RNA, U20 small nucleolar', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231456444, end_position_on_the_genomic_accession=231456523, orientation='minus', exon_count=1, OMIM=604012, no_hdr0=''),
  6083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6083, CurrentID=0, Status='live', Symbol='SNORD21', Aliases='RNU21, U21', description='small nucleolar RNA, C/D box 21', other_designations='RNA, U21 small nuclear|U21 small nucleolar RNA|U21 snoRNA', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92837289, end_position_on_the_genomic_accession=92837383, orientation='plus', exon_count=1, OMIM=603635, no_hdr0=''),
  6084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6084, CurrentID=0, Status='live', Symbol='RNY1', Aliases='HY1, Y1', description='RNA, Ro-associated Y1', other_designations='RNA, Y1 small cytoplasmic (associated with Ro protein)', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148987136, end_position_on_the_genomic_accession=148987248, orientation='minus', exon_count=1, OMIM=601821, no_hdr0=''),
  6085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6085, CurrentID=0, Status='live', Symbol='RNY3', Aliases='HY3, Y3', description='RNA, Ro-associated Y3', other_designations='RNA, Y3 small cytoplasmic (associated with Ro protein)', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148983755, end_position_on_the_genomic_accession=148983856, orientation='plus', exon_count=1, OMIM=601822, no_hdr0=''),
  6086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6086, CurrentID=0, Status='live', Symbol='RNY4', Aliases='HY4, Y4', description='RNA, Ro-associated Y4', other_designations='RNA, Y4 small cytoplasmic (associated with Ro protein)', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148963315, end_position_on_the_genomic_accession=148963410, orientation='plus', exon_count=1, OMIM=601823, no_hdr0=''),
  6087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6087, CurrentID=0, Status='live', Symbol='RNY4P1', Aliases='hY4.B8', description='RNA, Ro-associated Y4 pseudogene 1', other_designations='RNA, Y4 small cytoplasmic (associated with Ro protein) pseudogene 1', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74847506, end_position_on_the_genomic_accession=74847599, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6088, CurrentID=0, Status='live', Symbol='RNY4P2', Aliases='hY4.B7', description='RNA, Ro-associated Y4 pseudogene 2', other_designations='RNA, Y4 small cytoplasmic (associated with Ro protein) pseudogene 2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42932850, end_position_on_the_genomic_accession=42932942, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6089, CurrentID=0, Status='live', Symbol='RNY4P3', Aliases='hY4.F2', description='RNA, Ro-associated Y4 pseudogene 3', other_designations='RNA, Y4 small cytoplasmic (associated with Ro protein) pseudogene 3', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50098264, end_position_on_the_genomic_accession=50098356, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6090, CurrentID=0, Status='live', Symbol='RNY5', Aliases='Y5', description='RNA, Ro-associated Y5', other_designations='RNA, Y5 small cytoplasmic (associated with Ro protein)', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148941488, end_position_on_the_genomic_accession=148941571, orientation='plus', exon_count=1, OMIM=601824, no_hdr0=''),
  6091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6091, CurrentID=0, Status='live', Symbol='ROBO1', Aliases='DUTT1, SAX3', description='roundabout guidance receptor 1', other_designations='deleted in U twenty twenty|roundabout, axon guidance receptor, homolog 1', map_location='3p12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=78597238, end_position_on_the_genomic_accession=79767998, orientation='minus', exon_count=35, OMIM=602430, no_hdr0=''),
  6092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6092, CurrentID=0, Status='live', Symbol='ROBO2', Aliases='SAX3', description='roundabout guidance receptor 2', other_designations='roundabout, axon guidance receptor, homolog 2', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75906675, end_position_on_the_genomic_accession=77649964, orientation='plus', exon_count=34, OMIM=602431, no_hdr0=''),
  6093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6093, CurrentID=0, Status='live', Symbol='ROCK1', Aliases='P160ROCK, ROCK-I', description='Rho-associated, coiled-coil containing protein kinase 1', other_designations='p160 ROCK-1|renal carcinoma antigen NY-REN-35', map_location='18q11.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=20949740, end_position_on_the_genomic_accession=21111851, orientation='minus', exon_count=35, OMIM=601702, no_hdr0=''),
  6094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6094, CurrentID=0, Status='live', Symbol='ROM1', Aliases='ROM, ROSP1, RP7, TSPAN23', description='retinal outer segment membrane protein 1', other_designations='tetraspanin-23|tspan-23', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62612741, end_position_on_the_genomic_accession=62615120, orientation='plus', exon_count=3, OMIM=180721, no_hdr0=''),
  6095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6095, CurrentID=0, Status='live', Symbol='RORA', Aliases='NR1F1, ROR1, ROR2, ROR3, RZR-ALPHA, RZRA', description='RAR-related orphan receptor A', other_designations='ROR-alpha|nuclear receptor RZR-alpha|nuclear receptor subfamily 1 group F member 1|retinoic acid receptor-related orphan receptor alpha|retinoid-related orphan receptor alpha|thyroid hormone nuclear receptor alpha variant 4|transcription factor RZR-alpha', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60488284, end_position_on_the_genomic_accession=61229303, orientation='minus', exon_count=23, OMIM=600825, no_hdr0=''),
  6096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6096, CurrentID=0, Status='live', Symbol='RORB', Aliases='NR1F2, ROR-BETA, RZR-BETA, RZRB, bA133M9.1', description='RAR-related orphan receptor B', other_designations='RAR-related orphan receptor beta|nuclear receptor RZR-beta|nuclear receptor subfamily 1 group F member 2|retinoic acid-binding receptor beta|retinoid-related orphan receptor beta|retinoid-related orphan receptor-beta', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74497336, end_position_on_the_genomic_accession=74687201, orientation='plus', exon_count=10, OMIM=601972, no_hdr0=''),
  6097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6097, CurrentID=0, Status='live', Symbol='RORC', Aliases='NR1F3, RORG, RZR-GAMMA, RZRG, TOR', description='RAR-related orphan receptor C', other_designations='RAR-related orphan nuclear receptor variant 2|nuclear receptor RZR-gamma|nuclear receptor subfamily 1 group F member 3|retinoic acid-binding receptor gamma|retinoid-related orphan receptor gamma', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151806071, end_position_on_the_genomic_accession=151832238, orientation='minus', exon_count=13, OMIM=602943, no_hdr0=''),
  6098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6098, CurrentID=0, Status='live', Symbol='ROS1', Aliases='MCF3, ROS, c-ros-1', description='ROS proto-oncogene 1 , receptor tyrosine kinase', other_designations='c-ros oncogene 1 , receptor tyrosine kinase|proto-oncogene c-Ros-1|transmembrane tyrosine-specific protein kinase|v-ros avian UR2 sarcoma virus oncogene homolog 1', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117287918, end_position_on_the_genomic_accession=117425922, orientation='minus', exon_count=47, OMIM=165020, no_hdr0=''),
  6100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6100, CurrentID=0, Status='live', Symbol='RP9', Aliases='PAP-1, PAP1', description='retinitis pigmentosa 9 (autosomal dominant)', other_designations='Pim-1 kinase associated protein|pim-1-associated protein', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33094797, end_position_on_the_genomic_accession=33109390, orientation='minus', exon_count=7, OMIM=607331, no_hdr0=''),
  6101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6101, CurrentID=0, Status='live', Symbol='RP1', Aliases='DCDC4A, ORP1', description='retinitis pigmentosa 1 (autosomal dominant)', other_designations='retinitis pigmentosa 1 protein|retinitis pigmentosa RP1 protein', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54616067, end_position_on_the_genomic_accession=54630834, orientation='plus', exon_count=4, OMIM=603937, no_hdr0=''),
  6102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6102, CurrentID=0, Status='live', Symbol='RP2', Aliases='DELXp11.3, NM23-H10, NME10, TBCCD2, XRP2', description='retinitis pigmentosa 2 (X-linked recessive)', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46836912, end_position_on_the_genomic_accession=46882358, orientation='plus', exon_count=5, OMIM=300757, no_hdr0=''),
  6103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6103, CurrentID=0, Status='live', Symbol='RPGR', Aliases='COD1, CORDX1, CRD, PCDX, RP15, RP3, XLRP3, orf15', description='retinitis pigmentosa GTPase regulator', other_designations='retinitis pigmentosa 15|retinitis pigmentosa 3 GTPase regulator', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38269170, end_position_on_the_genomic_accession=38327564, orientation='minus', exon_count=19, OMIM=312610, no_hdr0=''),
  6104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6104, CurrentID=0, Status='live', Symbol='RP6', Aliases='', description='retinitis pigmentosa 6 (X-linked recessive)', other_designations='', map_location='Xp21.3-p21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=312612, no_hdr0=''),
  101929210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929210, CurrentID=0, Status='live', Symbol='LOC101929210', Aliases='', description='uncharacterized LOC101929210', other_designations='', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94117820, end_position_on_the_genomic_accession=94207556, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100271098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271098, CurrentID=0, Status='live', Symbol='RPS24P12', Aliases='RPS24_3_730', description='ribosomal protein S24 pseudogene 12', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107229759, end_position_on_the_genomic_accession=107230152, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6114, CurrentID=0, Status='live', Symbol='RP22', Aliases='', description='retinitis pigmentosa 22 (autosomal recessive)', other_designations='', map_location='16p12.3-p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602594, no_hdr0=''),
  6116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6116, CurrentID=0, Status='live', Symbol='RP24', Aliases='', description='retinitis pigmentosa 24 (X-linked recessive)', other_designations='', map_location='Xq26-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300155, no_hdr0=''),
  6117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6117, CurrentID=0, Status='live', Symbol='RPA1', Aliases='HSSB, MST075, REPA1, RF-A, RP-A, RPA70', description='replication protein A1', other_designations='MSTP075|RF-A protein 1|RP-A p70|replication factor A protein 1|replication protein A1, 70kDa|single-stranded DNA-binding protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1829979, end_position_on_the_genomic_accession=1899554, orientation='plus', exon_count=17, OMIM=179835, no_hdr0=''),
  6118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6118, CurrentID=0, Status='live', Symbol='RPA2', Aliases='REPA2, RP-A p32, RP-A p34, RPA32', description='replication protein A2', other_designations='RF-A protein 2|replication factor A protein 2|replication protein A 34 kDa subunit|replication protein A2, 32kDa', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27891524, end_position_on_the_genomic_accession=27914797, orientation='minus', exon_count=9, OMIM=179836, no_hdr0=''),
  6119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6119, CurrentID=0, Status='live', Symbol='RPA3', Aliases='REPA3, RP-A p14', description='replication protein A3', other_designations='RF-A protein 3|replication factor A protein 3|replication protein A3, 14kDa', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7636944, end_position_on_the_genomic_accession=7718607, orientation='minus', exon_count=8, OMIM=179837, no_hdr0=''),
  6120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6120, CurrentID=0, Status='live', Symbol='RPE', Aliases='RPE2-1', description='ribulose-5-phosphate-3-epimerase', other_designations='', map_location='2q32-q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210002565, end_position_on_the_genomic_accession=210024599, orientation='plus', exon_count=9, OMIM=180480, no_hdr0=''),
  6121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6121, CurrentID=0, Status='live', Symbol='RPE65', Aliases='BCO3, LCA2, RP20, mRPE65, rd12, sRPE65', description='retinal pigment epithelium-specific protein 65kDa', other_designations='BCO family, member 3|RBP-binding membrane protein|all-trans-retinyl-palmitate hydrolase|p63|retinal pigment epithelium-specific 65 kDa protein|retinitis pigmentosa 20 (autosomal recessive)|retinol isomerase', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68428824, end_position_on_the_genomic_accession=68449959, orientation='minus', exon_count=14, OMIM=180069, no_hdr0=''),
  6122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6122, CurrentID=0, Status='live', Symbol='RPL3', Aliases='ASC-1, L3, TARBP-B', description='ribosomal protein L3', other_designations='HIV-1 TAR RNA-binding protein B', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39312882, end_position_on_the_genomic_accession=39319665, orientation='minus', exon_count=10, OMIM=604163, no_hdr0=''),
  6123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6123, CurrentID=0, Status='live', Symbol='RPL3L', Aliases='', description='ribosomal protein L3 like', other_designations='ribosomal protein L3-like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1944579, end_position_on_the_genomic_accession=1954678, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  6124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6124, CurrentID=0, Status='live', Symbol='RPL4', Aliases='L4', description='ribosomal protein L4', other_designations='60S ribosomal protein L1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66499315, end_position_on_the_genomic_accession=66504855, orientation='minus', exon_count=10, OMIM=180479, no_hdr0=''),
  6125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6125, CurrentID=0, Status='live', Symbol='RPL5', Aliases='DBA6, L5, MSTP030, PPP1R135', description='ribosomal protein L5', other_designations='protein phosphatase 1, regulatory subunit 135', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92832037, end_position_on_the_genomic_accession=92841924, orientation='plus', exon_count=8, OMIM=603634, no_hdr0=''),
  6126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6126, CurrentID=0, Status='live', Symbol='RPL9P7', Aliases='RPL9L1, RPL9P1, RPL9_19_1768', description='ribosomal protein L9 pseudogene 7', other_designations='60S ribosomal protein L9 pseudogene 1|ribosomal protein L9 pseudogene 1|ribosomal protein L9-like 1', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23836644, end_position_on_the_genomic_accession=23837346, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271101, CurrentID=0, Status='live', Symbol='RPS25P3', Aliases='RPS25_1_198', description='ribosomal protein S25 pseudogene 3', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20606280, end_position_on_the_genomic_accession=20606654, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6128, CurrentID=0, Status='live', Symbol='RPL6', Aliases='L6, SHUJUN-2, TAXREB107, TXREB1', description='ribosomal protein L6', other_designations='DNA-binding protein TAXREB107|neoplasm-related protein C140|tax-responsive enhancer element-binding protein 107', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112405190, end_position_on_the_genomic_accession=112418838, orientation='minus', exon_count=13, OMIM=603703, no_hdr0=''),
  6129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6129, CurrentID=0, Status='live', Symbol='RPL7', Aliases='L7, humL7-1', description='ribosomal protein L7', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73290639, end_position_on_the_genomic_accession=73294494, orientation='minus', exon_count=8, OMIM=604166, no_hdr0=''),
  6130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6130, CurrentID=0, Status='live', Symbol='RPL7A', Aliases='L7A, SURF3, TRUP', description='ribosomal protein L7a', other_designations='PLA-X polypeptide|surfeit 3|surfeit locus protein 3|thyroid hormone receptor uncoupling protein', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133348214, end_position_on_the_genomic_accession=133351425, orientation='plus', exon_count=8, OMIM=185640, no_hdr0=''),
  6132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6132, CurrentID=0, Status='live', Symbol='RPL8', Aliases='L8', description='ribosomal protein L8', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144789767, end_position_on_the_genomic_accession=144792420, orientation='minus', exon_count=6, OMIM=604177, no_hdr0=''),
  6133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6133, CurrentID=0, Status='live', Symbol='RPL9', Aliases='L9, NPC-A-16', description='ribosomal protein L9', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39454124, end_position_on_the_genomic_accession=39458948, orientation='minus', exon_count=8, OMIM=603686, no_hdr0=''),
  6134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6134, CurrentID=0, Status='live', Symbol='RPL10', Aliases='AUTSX5, DXS648, DXS648E, L10, NOV, QM', description='ribosomal protein L10', other_designations='Wilms tumor-related protein|laminin receptor homolog|tumor suppressor QM', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154398065, end_position_on_the_genomic_accession=154402339, orientation='plus', exon_count=7, OMIM=312173, no_hdr0=''),
  6135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6135, CurrentID=0, Status='live', Symbol='RPL11', Aliases='DBA7, GIG34, L11', description='ribosomal protein L11', other_designations='CLL-associated antigen KW-12|cell growth-inhibiting protein 34', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23691779, end_position_on_the_genomic_accession=23696425, orientation='plus', exon_count=6, OMIM=604175, no_hdr0=''),
  6136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6136, CurrentID=0, Status='live', Symbol='RPL12', Aliases='L12', description='ribosomal protein L12', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127447674, end_position_on_the_genomic_accession=127451432, orientation='minus', exon_count=7, OMIM=180475, no_hdr0=''),
  6137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6137, CurrentID=0, Status='live', Symbol='RPL13', Aliases='BBC1, D16S444E, D16S44E, L13', description='ribosomal protein L13', other_designations='OK/SW-cl.46|breast basic conserved protein 1', map_location='17p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89560657, end_position_on_the_genomic_accession=89566829, orientation='plus', exon_count=7, OMIM=113703, no_hdr0=''),
  6138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6138, CurrentID=0, Status='live', Symbol='RPL15', Aliases='DBA12, EC45, L15, RPL10, RPLY10, RPYL10', description='ribosomal protein L15', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23916804, end_position_on_the_genomic_accession=23923696, orientation='plus', exon_count=6, OMIM=604174, no_hdr0=''),
  6139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6139, CurrentID=0, Status='live', Symbol='RPL17', Aliases='L17, PD-1, RPL23', description='ribosomal protein L17', other_designations='60S ribosomal protein L23|gene encoding putative NFkB activating protein', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49488481, end_position_on_the_genomic_accession=49492565, orientation='minus', exon_count=8, OMIM=603661, no_hdr0=''),
  6141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6141, CurrentID=0, Status='live', Symbol='RPL18', Aliases='L18', description='ribosomal protein L18', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48615327, end_position_on_the_genomic_accession=48619418, orientation='minus', exon_count=7, OMIM=604179, no_hdr0=''),
  6142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6142, CurrentID=0, Status='live', Symbol='RPL18A', Aliases='L18A', description='ribosomal protein L18a', other_designations='ribosomal protein L18a-like protein', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17859878, end_position_on_the_genomic_accession=17863324, orientation='plus', exon_count=5, OMIM=604178, no_hdr0=''),
  6143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6143, CurrentID=0, Status='live', Symbol='RPL19', Aliases='L19', description='ribosomal protein L19', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39200283, end_position_on_the_genomic_accession=39204727, orientation='plus', exon_count=6, OMIM=180466, no_hdr0=''),
  6144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6144, CurrentID=0, Status='live', Symbol='RPL21', Aliases='HYPT12, L21', description='ribosomal protein L21', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27251555, end_position_on_the_genomic_accession=27256568, orientation='plus', exon_count=6, OMIM=603636, no_hdr0=''),
  6145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6145, CurrentID=0, Status='live', Symbol='RPL21P1', Aliases='RPL21P', description='ribosomal protein L21 pseudogene 1', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122364742, end_position_on_the_genomic_accession=122365674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6146, CurrentID=0, Status='live', Symbol='RPL22', Aliases='EAP, HBP15, HBP15/L22, L22', description='ribosomal protein L22', other_designations='EBER-associated protein|Epstein-Barr virus small RNA-associated protein|Epstein-Barr-encoded RNA-associated protein|heparin-binding protein 15|heparin-binding protein HBp15', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6185020, end_position_on_the_genomic_accession=6199619, orientation='minus', exon_count=4, OMIM=180474, no_hdr0=''),
  6147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6147, CurrentID=0, Status='live', Symbol='RPL23A', Aliases='L23A, MDA20', description='ribosomal protein L23a', other_designations='melanoma differentiation-associated gene 20', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28719982, end_position_on_the_genomic_accession=28724356, orientation='plus', exon_count=5, OMIM=602326, no_hdr0=''),
  6148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6148, CurrentID=0, Status='live', Symbol='RPL23AP1', Aliases='RPL23A_15_676, dJ377H14.8', description='ribosomal protein L23a pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29726632, end_position_on_the_genomic_accession=29727154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6150, CurrentID=0, Status='live', Symbol='MRPL23', Aliases='L23MRP, RPL23, RPL23L', description='mitochondrial ribosomal protein L23', other_designations='L23 mitochondrial-related protein|L23mt|MRP-L23|ribosomal protein related to L23 (mitochondrial)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1947272, end_position_on_the_genomic_accession=2014017, orientation='plus', exon_count=12, OMIM=600789, no_hdr0=''),
  100271105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271105, CurrentID=0, Status='live', Symbol='RPS26P21', Aliases='RPS26_3_394', description='ribosomal protein S26 pseudogene 21', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119298665, end_position_on_the_genomic_accession=119299012, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6152, CurrentID=0, Status='live', Symbol='RPL24', Aliases='HEL-S-310, L24', description='ribosomal protein L24', other_designations='60S ribosomal protein L30|epididymis secretory protein Li 310|ribosomal protein L30', map_location='3q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101681090, end_position_on_the_genomic_accession=101686719, orientation='minus', exon_count=6, OMIM=604180, no_hdr0=''),
  6154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6154, CurrentID=0, Status='live', Symbol='RPL26', Aliases='DBA11, L26', description='ribosomal protein L26', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8377516, end_position_on_the_genomic_accession=8383247, orientation='minus', exon_count=4, OMIM=603704, no_hdr0=''),
  6155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6155, CurrentID=0, Status='live', Symbol='RPL27', Aliases='L27', description='ribosomal protein L27', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42998429, end_position_on_the_genomic_accession=43002959, orientation='plus', exon_count=5, OMIM=607526, no_hdr0=''),
  6156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6156, CurrentID=0, Status='live', Symbol='RPL30', Aliases='L30', description='ribosomal protein L30', other_designations='', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98041710, end_position_on_the_genomic_accession=98045590, orientation='minus', exon_count=5, OMIM=180467, no_hdr0=''),
  6157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6157, CurrentID=0, Status='live', Symbol='RPL27A', Aliases='L27A', description='ribosomal protein L27a', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8682448, end_position_on_the_genomic_accession=8689872, orientation='plus', exon_count=5, OMIM=603637, no_hdr0=''),
  6158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6158, CurrentID=0, Status='live', Symbol='RPL28', Aliases='L28', description='ribosomal protein L28', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55385101, end_position_on_the_genomic_accession=55403245, orientation='plus', exon_count=6, OMIM=603638, no_hdr0=''),
  6159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6159, CurrentID=0, Status='live', Symbol='RPL29', Aliases='HIP, HUMRPL29, L29P10, RPL29_3_370, RPL29', description='ribosomal protein L29', other_designations='HP/HS-interacting protein|cell surface heparin-binding protein HIP|heparin/heparan sulfate-binding protein|heparin/heparan sulfate-interacting protein|ribosomal protein YL43 homologue', map_location='3p21.3-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51993628, end_position_on_the_genomic_accession=51995942, orientation='minus', exon_count=4, OMIM=601832, no_hdr0=''),
  6160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6160, CurrentID=0, Status='live', Symbol='RPL31', Aliases='L31', description='ribosomal protein L31', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101002229, end_position_on_the_genomic_accession=101019693, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  6161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6161, CurrentID=0, Status='live', Symbol='RPL32', Aliases='L32, PP9932', description='ribosomal protein L32', other_designations='', map_location='3p25-p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12834945, end_position_on_the_genomic_accession=12841582, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  6163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6163, CurrentID=0, Status='live', Symbol='RPL32P1', Aliases='RPL32-L, RPL32P, RPL32_6_681', description='ribosomal protein L32 pseudogene 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33079299, end_position_on_the_genomic_accession=33080011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6164, CurrentID=0, Status='live', Symbol='RPL34', Aliases='L34', description='ribosomal protein L34', other_designations='leukemia-associated protein', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108620566, end_position_on_the_genomic_accession=108630484, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  6165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6165, CurrentID=0, Status='live', Symbol='RPL35A', Aliases='DBA5, L35A', description='ribosomal protein L35a', other_designations='cell growth-inhibiting gene 33 protein', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197950181, end_position_on_the_genomic_accession=197955851, orientation='plus', exon_count=5, OMIM=180468, no_hdr0=''),
  6166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6166, CurrentID=0, Status='live', Symbol='RPL36AL', Aliases='RPL36A', description='ribosomal protein L36a like', other_designations='ribosomal protein HL44|ribosomal protein L36a', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49618688, end_position_on_the_genomic_accession=49620685, orientation='minus', exon_count=2, OMIM=180469, no_hdr0=''),
  6167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6167, CurrentID=0, Status='live', Symbol='RPL37', Aliases='L37', description='ribosomal protein L37', other_designations='60S ribosomal protein L37a|G1.16', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40831328, end_position_on_the_genomic_accession=40835285, orientation='minus', exon_count=4, OMIM=604181, no_hdr0=''),
  6168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6168, CurrentID=0, Status='live', Symbol='RPL37A', Aliases='L37A', description='ribosomal protein L37a', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216498797, end_position_on_the_genomic_accession=216501467, orientation='plus', exon_count=4, OMIM=613314, no_hdr0=''),
  6169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6169, CurrentID=0, Status='live', Symbol='RPL38', Aliases='L38', description='ribosomal protein L38', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74203656, end_position_on_the_genomic_accession=74209880, orientation='plus', exon_count=5, OMIM=604182, no_hdr0=''),
  6170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6170, CurrentID=0, Status='live', Symbol='RPL39', Aliases='L39P42, RPL39_23_1806, RPL39', description='ribosomal protein L39', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119786504, end_position_on_the_genomic_accession=119791659, orientation='minus', exon_count=3, OMIM=300899, no_hdr0=''),
  6171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6171, CurrentID=0, Status='live', Symbol='RPL41', Aliases='L41', description='ribosomal protein L41', other_designations='HG12 protein|homologue of yeast ribosomal protein YL41', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56116590, end_position_on_the_genomic_accession=56117832, orientation='plus', exon_count=4, OMIM=613315, no_hdr0=''),
  6172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6172, CurrentID=0, Status='live', Symbol='RPL41P5', Aliases='RPL41L', description='ribosomal protein L41 pseudogene 5', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93083293, end_position_on_the_genomic_accession=93083716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6173, CurrentID=0, Status='live', Symbol='RPL36A', Aliases='L36A, L44L, MIG6, RPL44', description='ribosomal protein L36a', other_designations='60S ribosomal protein L44|L44-like ribosomal protein|cell growth-inhibiting gene 15 protein|cell migration-inducing gene 6 protein|dJ164F3.3 (ribosomal protein L44)', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101390890, end_position_on_the_genomic_accession=101396154, orientation='plus', exon_count=5, OMIM=300902, no_hdr0=''),
  6174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6174, CurrentID=0, Status='live', Symbol='RPL7P1', Aliases='RPL17P, RPL7P, RPL7_11_628', description='ribosomal protein L7 pseudogene 1', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150094285, end_position_on_the_genomic_accession=150095114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6175, CurrentID=0, Status='live', Symbol='RPLP0', Aliases='L10E, LP0, P0, PRLP0, RPP0', description='ribosomal protein, large, P0', other_designations='60S ribosomal protein L10E|acidic ribosomal phosphoprotein P0', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120196700, end_position_on_the_genomic_accession=120201211, orientation='minus', exon_count=8, OMIM=180510, no_hdr0=''),
  6176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6176, CurrentID=0, Status='live', Symbol='RPLP1', Aliases='LP1, P1, RPP1', description='ribosomal protein, large, P1', other_designations='acidic ribosomal phosphoprotein P1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69452820, end_position_on_the_genomic_accession=69455545, orientation='plus', exon_count=4, OMIM=180520, no_hdr0=''),
  6181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6181, CurrentID=0, Status='live', Symbol='RPLP2', Aliases='D11S2243E, LP2, P2, RPP2', description='ribosomal protein, large, P2', other_designations='acidic ribosomal phosphoprotein P2|renal carcinoma antigen NY-REN-44', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=809936, end_position_on_the_genomic_accession=812876, orientation='plus', exon_count=5, OMIM=180530, no_hdr0=''),
  6182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6182, CurrentID=0, Status='live', Symbol='MRPL12', Aliases='5c5-2, L12mt, MRP-L31/34, MRPL7, MRPL7/L12, RPML12', description='mitochondrial ribosomal protein L12', other_designations='MRP-L12', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81703370, end_position_on_the_genomic_accession=81707526, orientation='plus', exon_count=5, OMIM=602375, no_hdr0=''),
  6183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6183, CurrentID=0, Status='live', Symbol='MRPS12', Aliases='MPR-S12, MT-RPS12, RPMS12, RPS12, RPSM12', description='mitochondrial ribosomal protein S12', other_designations='MRP-S12|S12mt', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38930708, end_position_on_the_genomic_accession=38933020, orientation='plus', exon_count=3, OMIM=603021, no_hdr0=''),
  6184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6184, CurrentID=0, Status='live', Symbol='RPN1', Aliases='OST1, RBPH1', description='ribophorin I', other_designations='RPN-I|dolichyl-diphosphooligosaccharide--protein glycosyltransferase 67 kDa subunit|dolichyl-diphosphooligosaccharide-protein glycosyltransferase 67 kDa subunit|oligosaccharyltransferase 1 homolog|oligosaccharyltransferase complex subunit (non-catalytic)|ribophorin-1', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128619970, end_position_on_the_genomic_accession=128650876, orientation='minus', exon_count=10, OMIM=180470, no_hdr0=''),
  6185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6185, CurrentID=0, Status='live', Symbol='RPN2', Aliases='RIBIIR, RPN-II, RPNII, SWP1', description='ribophorin II', other_designations='dolichyl-diphosphooligosaccharide--protein glycosyltransferase 63 kDa subunit|oligosaccharyltransferase complex subunit (non-catalytic)|ribophorin-2', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37179051, end_position_on_the_genomic_accession=37241623, orientation='plus', exon_count=19, OMIM=180490, no_hdr0=''),
  6187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6187, CurrentID=0, Status='live', Symbol='RPS2', Aliases='LLREP3, S2', description='ribosomal protein S2', other_designations='40S ribosomal protein S4|OK/KNS-cl.6|protein LLRep3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1962061, end_position_on_the_genomic_accession=1964826, orientation='minus', exon_count=7, OMIM=603624, no_hdr0=''),
  6188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6188, CurrentID=0, Status='live', Symbol='RPS3', Aliases='S3', description='ribosomal protein S3', other_designations='IMR-90 ribosomal protein S3', map_location='11q13.3-q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75399491, end_position_on_the_genomic_accession=75422302, orientation='plus', exon_count=8, OMIM=600454, no_hdr0=''),
  6189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6189, CurrentID=0, Status='live', Symbol='RPS3A', Aliases='FTE1, MFTL, S3A', description='ribosomal protein S3A', other_designations='fte-1|v-fos transformation effector protein 1', map_location='4q31.2-q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151099573, end_position_on_the_genomic_accession=151104652, orientation='plus', exon_count=5, OMIM=180478, no_hdr0=''),
  6191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6191, CurrentID=0, Status='live', Symbol='RPS4X', Aliases='CCG2, DXS306, RPS4, S4, SCAR, SCR10', description='ribosomal protein S4, X-linked', other_designations='cell cycle gene 2|ribosomal protein S4X isoform|single copy abundant mRNA protein|single-copy abundant mRNA', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72272603, end_position_on_the_genomic_accession=72277291, orientation='minus', exon_count=7, OMIM=312760, no_hdr0=''),
  6192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6192, CurrentID=0, Status='live', Symbol='RPS4Y1', Aliases='RPS4Y, S4', description='ribosomal protein S4, Y-linked 1', other_designations='40S ribosomal protein S4, Y|ribosomal protein S4Y', map_location='Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2841582, end_position_on_the_genomic_accession=2866956, orientation='plus', exon_count=7, OMIM=470000, no_hdr0=''),
  6193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6193, CurrentID=0, Status='live', Symbol='RPS5', Aliases='S5', description='ribosomal protein S5', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58387269, end_position_on_the_genomic_accession=58394804, orientation='plus', exon_count=6, OMIM=603630, no_hdr0=''),
  6194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6194, CurrentID=0, Status='live', Symbol='RPS6', Aliases='S6', description='ribosomal protein S6', other_designations='phosphoprotein NP33', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19376256, end_position_on_the_genomic_accession=19380237, orientation='minus', exon_count=6, OMIM=180460, no_hdr0=''),
  6195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6195, CurrentID=0, Status='live', Symbol='RPS6KA1', Aliases='HU-1, MAPKAPK1A, RSK, RSK1, p90Rsk', description='ribosomal protein S6 kinase, 90kDa, polypeptide 1', other_designations='90 kDa ribosomal protein S6 kinase 1|MAP kinase-activated protein kinase 1a|MAPK-activated protein kinase 1a|MAPKAP kinase 1a|MAPKAPK-1a|RSK-1|S6K-alpha 1|S6K-alpha-1|dJ590P13.1 (ribosomal protein S6 kinase, 90kD, polypeptide 1)|p90-RSK 1|p90RSK1|p90S6K|ribosomal S6 kinase 1|ribosomal protein S6 kinase alpha 1', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26529758, end_position_on_the_genomic_accession=26575030, orientation='plus', exon_count=26, OMIM=601684, no_hdr0=''),
  6196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6196, CurrentID=0, Status='live', Symbol='RPS6KA2', Aliases='HU-2, MAPKAPK1C, RSK, RSK3, S6K-alpha, S6K-alpha2, p90-RSK3, pp90RSK3', description='ribosomal protein S6 kinase, 90kDa, polypeptide 2', other_designations='90 kDa ribosomal protein S6 kinase 2|MAP kinase-activated protein kinase 1c|MAPK-activated protein kinase 1c|MAPKAP kinase 1c|MAPKAPK-1c|RSK-3|S6K-alpha-2|p90-RSK 2|p90RSK2|ribosomal S6 kinase 3|ribosomal protein S6 kinase 90kDa polypeptide 2|ribosomal protein S6 kinase alpha 2', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166409366, end_position_on_the_genomic_accession=166862550, orientation='minus', exon_count=26, OMIM=601685, no_hdr0=''),
  6197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6197, CurrentID=0, Status='live', Symbol='RPS6KA3', Aliases='CLS, HU-3, ISPK-1, MAPKAPK1B, MRX19, RSK, RSK2, S6K-alpha3, p90-RSK2, pp90RSK2', description='ribosomal protein S6 kinase, 90kDa, polypeptide 3', other_designations='MAP kinase-activated protein kinase 1b|MAPK-activated protein kinase 1b|MAPKAP kinase 1b|MAPKAPK-1b|RSK-2|S6K-alpha-3|insulin-stimulated protein kinase 1|p90-RSK 3|ribosomal S6 kinase 2', map_location='Xp22.2-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20149911, end_position_on_the_genomic_accession=20267095, orientation='minus', exon_count=26, OMIM=300075, no_hdr0=''),
  6198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6198, CurrentID=0, Status='live', Symbol='RPS6KB1', Aliases='PS6K, S6K, S6K-beta-1, S6K1, STK14A, p70 S6KA, p70(S6K)-alpha, p70-S6K, p70-alpha', description='ribosomal protein S6 kinase, 70kDa, polypeptide 1', other_designations='p70 S6 kinase, alpha|ribosomal protein S6 kinase I|serine/threonine kinase 14 alpha|serine/threonine-protein kinase 14A', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59893046, end_position_on_the_genomic_accession=59950426, orientation='plus', exon_count=19, OMIM=608938, no_hdr0=''),
  6199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6199, CurrentID=0, Status='live', Symbol='RPS6KB2', Aliases='KLS, P70-beta, P70-beta-1, P70-beta-2, S6K-beta2, S6K2, SRK, STK14B, p70(S6K)-beta, p70S6Kb', description='ribosomal protein S6 kinase, 70kDa, polypeptide 2', other_designations='70 kDa ribosomal protein S6 kinase 2|P70S6K2|S6 kinase-related kinase|S6K-beta|S6K-beta-2|p70 S6 kinase beta|p70 S6K-beta|p70 S6KB|p70 ribosomal S6 kinase beta|p70-S6K 2|serine/threonine-protein kinase 14 beta|serine/threonine-protein kinase 14B', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67428460, end_position_on_the_genomic_accession=67435408, orientation='plus', exon_count=16, OMIM=608939, no_hdr0=''),
  6200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6200, CurrentID=0, Status='live', Symbol='RPS6KB3', Aliases='70kD', description='ribosomal protein S6 kinase, 70kDa, polypeptide 3', other_designations='polypeptide 3', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6201, CurrentID=0, Status='live', Symbol='RPS7', Aliases='DBA8, S7', description='ribosomal protein S7', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3575263, end_position_on_the_genomic_accession=3580919, orientation='plus', exon_count=7, OMIM=603658, no_hdr0=''),
  6202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6202, CurrentID=0, Status='live', Symbol='RPS8', Aliases='S8', description='ribosomal protein S8', other_designations='OK/SW-cl.83', map_location='1p34.1-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44775574, end_position_on_the_genomic_accession=44778740, orientation='plus', exon_count=6, OMIM=600357, no_hdr0=''),
  6203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6203, CurrentID=0, Status='live', Symbol='RPS9', Aliases='S9', description='ribosomal protein S9', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54200557, end_position_on_the_genomic_accession=54207647, orientation='plus', exon_count=6, OMIM=603631, no_hdr0=''),
  6204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6204, CurrentID=0, Status='live', Symbol='RPS10', Aliases='DBA9, S10', description='ribosomal protein S10', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34417454, end_position_on_the_genomic_accession=34426125, orientation='minus', exon_count=6, OMIM=603632, no_hdr0=''),
  6205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6205, CurrentID=0, Status='live', Symbol='RPS11', Aliases='S11', description='ribosomal protein S11', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49496365, end_position_on_the_genomic_accession=49499712, orientation='plus', exon_count=5, OMIM=180471, no_hdr0=''),
  6206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6206, CurrentID=0, Status='live', Symbol='RPS12', Aliases='S12', description='ribosomal protein S12', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132814569, end_position_on_the_genomic_accession=132817564, orientation='plus', exon_count=6, OMIM=603660, no_hdr0=''),
  6207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6207, CurrentID=0, Status='live', Symbol='RPS13', Aliases='S13', description='ribosomal protein S13', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17074392, end_position_on_the_genomic_accession=17077673, orientation='minus', exon_count=6, OMIM=180476, no_hdr0=''),
  6208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6208, CurrentID=0, Status='live', Symbol='RPS14', Aliases='EMTB, S14', description='ribosomal protein S14', other_designations='emetine resistance', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150444229, end_position_on_the_genomic_accession=150449756, orientation='minus', exon_count=5, OMIM=130620, no_hdr0=''),
  6209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6209, CurrentID=0, Status='live', Symbol='RPS15', Aliases='RIG, S15', description='ribosomal protein S15', other_designations='homolog of rat insulinoma|insulinoma protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1438364, end_position_on_the_genomic_accession=1440494, orientation='plus', exon_count=4, OMIM=180535, no_hdr0=''),
  6210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6210, CurrentID=0, Status='live', Symbol='RPS15A', Aliases='S15a', description='ribosomal protein S15a', other_designations='up-regulated by HBV X protein', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18782955, end_position_on_the_genomic_accession=18790334, orientation='minus', exon_count=5, OMIM=603674, no_hdr0=''),
  22881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22881, CurrentID=0, Status='live', Symbol='ANKRD6', Aliases='', description='ankyrin repeat domain 6', other_designations='diversin', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89433178, end_position_on_the_genomic_accession=89633834, orientation='plus', exon_count=21, OMIM=610583, no_hdr0=''),
  6217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6217, CurrentID=0, Status='live', Symbol='RPS16', Aliases='S16', description='ribosomal protein S16', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39433207, end_position_on_the_genomic_accession=39436020, orientation='minus', exon_count=5, OMIM=603675, no_hdr0=''),
  6218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6218, CurrentID=0, Status='live', Symbol='RPS17', Aliases='DBA4L, RPS17L1, RPS17L2, S17, RPS17', description='ribosomal protein S17', other_designations='', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82536750, end_position_on_the_genomic_accession=82540544, orientation='minus', exon_count=6, OMIM=180472, no_hdr0=''),
  6219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6219, CurrentID=0, Status='live', Symbol='RPS17P2', Aliases='RPS17A, RPS17L1, RPS17_5_604', description='ribosomal protein S17 pseudogene 2', other_designations='ribosomal protein S17-like 1|ribosomal protein S17a-like 1 (pseudogene)|ribosomal protein pseudogene S17a', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116716218, end_position_on_the_genomic_accession=116716696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6222, CurrentID=0, Status='live', Symbol='RPS18', Aliases='D6S218E, HKE3, KE-3, KE3, S18', description='ribosomal protein S18', other_designations='rhabdomyosarcoma antigen MU-RMS-40.21', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33272075, end_position_on_the_genomic_accession=33276504, orientation='plus', exon_count=6, OMIM=180473, no_hdr0=''),
  6223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6223, CurrentID=0, Status='live', Symbol='RPS19', Aliases='DBA, DBA1, S19', description='ribosomal protein S19', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41859918, end_position_on_the_genomic_accession=41871416, orientation='plus', exon_count=6, OMIM=603474, no_hdr0=''),
  6224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6224, CurrentID=0, Status='live', Symbol='RPS20', Aliases='S20', description='ribosomal protein S20', other_designations='', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56068180, end_position_on_the_genomic_accession=56074581, orientation='minus', exon_count=6, OMIM=603682, no_hdr0=''),
  6225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6225, CurrentID=0, Status='live', Symbol='RPS20P3', Aliases='RPS20A, RPS20_8_599', description='ribosomal protein S20 pseudogene 3', other_designations='ribosomal protein S20A (gene or pseudogene)', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108593484, end_position_on_the_genomic_accession=108594003, orientation='plus', exon_count=0, OMIM=180463, no_hdr0=''),
  6226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6226, CurrentID=0, Status='live', Symbol='RPS20P4', Aliases='RPS20B, RPS20_6_627', description='ribosomal protein S20 pseudogene 4', other_designations='ribosomal protein S20B (gene or pseudogene)', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150021527, end_position_on_the_genomic_accession=150022053, orientation='minus', exon_count=0, OMIM=180464, no_hdr0=''),
  6227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6227, CurrentID=0, Status='live', Symbol='RPS21', Aliases='HLDF, S21', description='ribosomal protein S21', other_designations='8.2 kDa differentiation factor|human leukemia differentiation factor', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62387065, end_position_on_the_genomic_accession=62388520, orientation='plus', exon_count=6, OMIM=180477, no_hdr0=''),
  6228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6228, CurrentID=0, Status='live', Symbol='RPS23', Aliases='S23', description='ribosomal protein S23', other_designations='homolog of yeast ribosomal protein S28', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82273320, end_position_on_the_genomic_accession=82278416, orientation='minus', exon_count=4, OMIM=603683, no_hdr0=''),
  6229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6229, CurrentID=0, Status='live', Symbol='RPS24', Aliases='DBA3, S24', description='ribosomal protein S24', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78033760, end_position_on_the_genomic_accession=78056813, orientation='plus', exon_count=10, OMIM=602412, no_hdr0=''),
  6230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6230, CurrentID=0, Status='live', Symbol='RPS25', Aliases='S25', description='ribosomal protein S25', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119015712, end_position_on_the_genomic_accession=119018347, orientation='minus', exon_count=5, OMIM=180465, no_hdr0=''),
  6231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6231, CurrentID=0, Status='live', Symbol='RPS26', Aliases='DBA10, S26', description='ribosomal protein S26', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56041902, end_position_on_the_genomic_accession=56044223, orientation='plus', exon_count=4, OMIM=603701, no_hdr0=''),
  6232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6232, CurrentID=0, Status='live', Symbol='RPS27', Aliases='MPS-1, MPS1, S27', description='ribosomal protein S27', other_designations='metallopan-stimulin 1|metallopanstimulin 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153990763, end_position_on_the_genomic_accession=153992155, orientation='plus', exon_count=4, OMIM=603702, no_hdr0=''),
  6233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6233, CurrentID=0, Status='live', Symbol='RPS27A', Aliases='CEP80, HEL112, S27A, UBA80, UBC, UBCEP1, UBCEP80', description='ribosomal protein S27a', other_designations='40S ribosomal protein S27a|epididymis luminal protein 112|ubiquitin C|ubiquitin and ribosomal protein S27a|ubiquitin carboxyl extension protein 80|ubiquitin-CEP80', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55231903, end_position_on_the_genomic_accession=55235853, orientation='plus', exon_count=7, OMIM=191343, no_hdr0=''),
  6234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6234, CurrentID=0, Status='live', Symbol='RPS28', Aliases='DBA15, S28', description='ribosomal protein S28', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8321500, end_position_on_the_genomic_accession=8322396, orientation='plus', exon_count=4, OMIM=603685, no_hdr0=''),
  6235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6235, CurrentID=0, Status='live', Symbol='RPS29', Aliases='DBA13, S29', description='ribosomal protein S29', other_designations='', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49576672, end_position_on_the_genomic_accession=49586416, orientation='minus', exon_count=4, OMIM=603633, no_hdr0=''),
  6236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6236, CurrentID=0, Status='live', Symbol='RRAD', Aliases='RAD, RAD1, REM3', description='Ras-related associated with diabetes', other_designations='RAS (RAD and GEM) like GTP binding 3|ras associated with diabetes', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66921679, end_position_on_the_genomic_accession=66925536, orientation='minus', exon_count=5, OMIM=179503, no_hdr0=''),
  6237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6237, CurrentID=0, Status='live', Symbol='RRAS', Aliases='', description='related RAS viral (r-ras) oncogene homolog', other_designations='oncogene RRAS|p23|ras family small GTP binding protein R-Ras', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49635295, end_position_on_the_genomic_accession=49640143, orientation='minus', exon_count=6, OMIM=165090, no_hdr0=''),
  6238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6238, CurrentID=0, Status='live', Symbol='RRBP1', Aliases='ES/130, ES130, RRp, hES', description='ribosome binding protein 1', other_designations='180 kDa ribosome receptor homolog|ES/130-related protein|ribosome binding protein 1 homolog 180kDa (dog)|ribosome receptor protein', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17613678, end_position_on_the_genomic_accession=17682283, orientation='minus', exon_count=26, OMIM=601418, no_hdr0=''),
  6239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6239, CurrentID=0, Status='live', Symbol='RREB1', Aliases='FINB, HNT, LZ321, RREB-1, Zep-1', description='ras responsive element binding protein 1', other_designations='DNA-binding protein|finger protein in nuclear bodies|hindsight homolog|raf-responsive zinc finger protein LZ321|zinc finger motif enhancer binding protein 1|zinc finger motif enhancer-binding protein 1|zinc finger motif-enhancer binding-protein 1', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7107597, end_position_on_the_genomic_accession=7251980, orientation='plus', exon_count=19, OMIM=602209, no_hdr0=''),
  6240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6240, CurrentID=0, Status='live', Symbol='RRM1', Aliases='R1, RIR1, RR1', description='ribonucleotide reductase M1', other_designations='ribonucleoside-diphosphate reductase subunit M1|ribonucleotide reductase M1 polypeptide|ribonucleotide reductase, R1 subunit', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4094561, end_position_on_the_genomic_accession=4138881, orientation='plus', exon_count=20, OMIM=180410, no_hdr0=''),
  6241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6241, CurrentID=0, Status='live', Symbol='RRM2', Aliases='R2, RR2, RR2M', description='ribonucleotide reductase M2', other_designations='ribonucleotide reductase M2 polypeptide|ribonucleotide reductase small chain|ribonucleotide reductase small subunit', map_location='2p25-p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10122568, end_position_on_the_genomic_accession=10131419, orientation='plus', exon_count=10, OMIM=180390, no_hdr0=''),
  6242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6242, CurrentID=0, Status='live', Symbol='RTKN', Aliases='', description='rhotekin', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74425861, end_position_on_the_genomic_accession=74441933, orientation='minus', exon_count=16, OMIM=602288, no_hdr0=''),
  6244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6244, CurrentID=0, Status='live', Symbol='RRM2P2', Aliases='', description='ribonucleotide reductase M2 polypeptide pseudogene 2', other_designations='ribonucleotide reductase M2 pseudogene', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161376801, end_position_on_the_genomic_accession=161379370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6246, CurrentID=0, Status='live', Symbol='RRM2P4', Aliases='', description='ribonucleotide reductase M2 polypeptide pseudogene 4', other_designations='ribonucleotide reductase M2 pseudogene', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=144309133, end_position_on_the_genomic_accession=144309698, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6247, CurrentID=0, Status='live', Symbol='RS1', Aliases='RS, XLRS1', description='retinoschisin 1', other_designations='X-linked juvenile retinoschisis protein', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18639688, end_position_on_the_genomic_accession=18672103, orientation='minus', exon_count=6, OMIM=300839, no_hdr0=''),
  6248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6248, CurrentID=0, Status='live', Symbol='RSC1A1', Aliases='RS1', description='regulatory solute carrier protein, family 1, member 1', other_designations='transporter regulator RS1', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15659869, end_position_on_the_genomic_accession=15661722, orientation='plus', exon_count=1, OMIM=601966, no_hdr0=''),
  6249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6249, CurrentID=0, Status='live', Symbol='CLIP1', Aliases='CLIP, CLIP-17070, CYLN1, RSN, CLIP1', description='CAP-Gly domain containing linker protein 1', other_designations='cytoplasmic linker protein 1|cytoplasmic linker protein 170 alpha-2|cytoplasmic linker protein CLIP-170|restin (Reed-Steinberg cell-expressed intermediate filament-associated protein)', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122271432, end_position_on_the_genomic_accession=122422632, orientation='minus', exon_count=27, OMIM=179838, no_hdr0=''),
  6251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6251, CurrentID=0, Status='live', Symbol='RSU1', Aliases='RSP-1', description='Ras suppressor protein 1', other_designations='ras suppressor protein 1 variant 1|ras suppressor protein 1 variant 2|ras suppressor protein 1 variant 3|ras suppressor protein 1 variant 5|rsu-1', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16590616, end_position_on_the_genomic_accession=16817454, orientation='minus', exon_count=10, OMIM=179555, no_hdr0=''),
  6252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6252, CurrentID=0, Status='live', Symbol='RTN1', Aliases='NSP', description='reticulon 1', other_designations='neuroendocrine-specific protein', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59595976, end_position_on_the_genomic_accession=59870839, orientation='minus', exon_count=11, OMIM=600865, no_hdr0=''),
  6253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6253, CurrentID=0, Status='live', Symbol='RTN2', Aliases='NSP2, NSPL1, NSPLI, SPG12', description='reticulon 2', other_designations='NSP-like protein 1|NSP-like protein I|neuroendocrine-specific protein-like I', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45485288, end_position_on_the_genomic_accession=45497055, orientation='minus', exon_count=11, OMIM=603183, no_hdr0=''),
  105377186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377186, CurrentID=0, Status='live', Symbol='LOC105377186', Aliases='', description='uncharacterized LOC105377186', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83818381, end_position_on_the_genomic_accession=83819967, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  6256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6256, CurrentID=0, Status='live', Symbol='RXRA', Aliases='NR2B1', description='retinoid X receptor alpha', other_designations='nuclear receptor subfamily 2 group B member 1|retinoid X nuclear receptor alpha|retinoid X receptor, alpha', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134326463, end_position_on_the_genomic_accession=134440586, orientation='plus', exon_count=12, OMIM=180245, no_hdr0=''),
  399473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399473, CurrentID=0, Status='live', Symbol='SPRED3', Aliases='Eve-3, spred-3', description='sprouty-related, EVH1 domain containing 3', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38388741, end_position_on_the_genomic_accession=38399884, orientation='plus', exon_count=8, OMIM=609293, no_hdr0=''),
  6258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6258, CurrentID=0, Status='live', Symbol='RXRG', Aliases='NR2B3, RXRC', description='retinoid X receptor gamma', other_designations='nuclear receptor subfamily 2 group B member 3|retinoid X receptor, gamma', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165400922, end_position_on_the_genomic_accession=165445355, orientation='minus', exon_count=12, OMIM=180247, no_hdr0=''),
  6259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6259, CurrentID=0, Status='live', Symbol='RYK', Aliases='D3S3195, JTK5, JTK5A1, RYK', description='receptor-like tyrosine kinase', other_designations='JTK5A protein tyrosine kinase|hydroxyaryl-protein kinase', map_location='3q22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134157133, end_position_on_the_genomic_accession=134250843, orientation='minus', exon_count=15, OMIM=600524, no_hdr0=''),
  6260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6260, CurrentID=0, Status='live', Symbol='RYKP1', Aliases='JTK5, JTK5B, RYKL1, RYKP', description='receptor-like tyrosine kinase pseudogene 1', other_designations='JTK5B protein tyrosine kinase|RYK receptor-like tyrosine kinase pseudogene 1|RYK receptor-like tyrosine kinase-like 1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4222066, end_position_on_the_genomic_accession=4225054, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6261, CurrentID=0, Status='live', Symbol='RYR1', Aliases='CCO, MHS, MHS1, PPP1R137, RYDR, RYR, RYR-1, SKRR', description='ryanodine receptor 1 (skeletal)', other_designations='central core disease of muscle|protein phosphatase 1, regulatory subunit 137|sarcoplasmic reticulum calcium release channel|skeletal muscle calcium release channel|skeletal muscle ryanodine receptor|type 1-like ryanodine receptor', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38433700, end_position_on_the_genomic_accession=38587564, orientation='plus', exon_count=106, OMIM=180901, no_hdr0=''),
  6262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6262, CurrentID=0, Status='live', Symbol='RYR2', Aliases='ARVC2, ARVD2, RYR-2, RyR, VTSIP', description='ryanodine receptor 2 (cardiac)', other_designations='cardiac muscle ryanodine receptor-calcium release channel|cardiac-type ryanodine receptor|islet-type ryanodine receptor|kidney-type ryanodine receptor|type 2 ryanodine receptor', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237042238, end_position_on_the_genomic_accession=237833988, orientation='plus', exon_count=107, OMIM=180902, no_hdr0=''),
  6263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6263, CurrentID=0, Status='live', Symbol='RYR3', Aliases='RYR-3', description='ryanodine receptor 3', other_designations='brain ryanodine receptor-calcium release channel|brain-type ryanodine receptor|type 3 ryanodine receptor', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33310783, end_position_on_the_genomic_accession=33866103, orientation='plus', exon_count=105, OMIM=180903, no_hdr0=''),
  6264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6264, CurrentID=0, Status='live', Symbol='S7', Aliases='SA7-2', description='surface antigen (chromosome 7) 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271124, CurrentID=0, Status='live', Symbol='RPS29P5', Aliases='RPS29_4_130', description='ribosomal protein S29 pseudogene 5', other_designations='', map_location='1q25.1-q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175921944, end_position_on_the_genomic_accession=175922231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6267, CurrentID=0, Status='live', Symbol='S11', Aliases='', description='surface antigen (X-linked) 2', other_designations='', map_location='Xq26-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6268, CurrentID=0, Status='live', Symbol='S12', Aliases='', description='surface antigen (X-linked) 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372822, CurrentID=0, Status='live', Symbol='LOC105372822', Aliases='', description='uncharacterized LOC105372822', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43284980, end_position_on_the_genomic_accession=43285295, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  6271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6271, CurrentID=0, Status='live', Symbol='S100A1', Aliases='S100, S100-alpha, S100A', description='S100 calcium binding protein A1', other_designations='S-100 protein alpha chain|S-100 protein subunit alpha|S100 alpha|S100 calcium-binding protein A1|S100 protein, alpha polypeptide', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153628397, end_position_on_the_genomic_accession=153632037, orientation='plus', exon_count=3, OMIM=176940, no_hdr0=''),
  6272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6272, CurrentID=0, Status='live', Symbol='SORT1', Aliases='Gp95, LDLCQ6, NT3, NTR3', description='sortilin 1', other_designations='100 kDa NT receptor|glycoprotein 95|neurotensin receptor 3', map_location='1p21.3-p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109309565, end_position_on_the_genomic_accession=109397941, orientation='minus', exon_count=23, OMIM=602458, no_hdr0=''),
  6273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6273, CurrentID=0, Status='live', Symbol='S100A2', Aliases='CAN19, S100L', description='S100 calcium binding protein A2', other_designations='S100 calcium-binding protein A2', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153561108, end_position_on_the_genomic_accession=153565830, orientation='minus', exon_count=3, OMIM=176993, no_hdr0=''),
  6274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6274, CurrentID=0, Status='live', Symbol='S100A3', Aliases='S100E', description='S100 calcium binding protein A3', other_designations='S100 calcium-binding protein A3', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153547333, end_position_on_the_genomic_accession=153549258, orientation='minus', exon_count=3, OMIM=176992, no_hdr0=''),
  6275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6275, CurrentID=0, Status='live', Symbol='S100A4', Aliases='18A2, 42A, CAPL, FSP1, MTS1, P9KA, PEL98', description='S100 calcium binding protein A4', other_designations='S100 calcium-binding protein A4 (calcium protein, calvasculin, metastasin, murine placental homolog)|fibroblast-specific protein-1|leukemia multidrug resistance associated protein|malignant transformation suppression 1|placental calcium-binding protein|protein Mts1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153543619, end_position_on_the_genomic_accession=153545806, orientation='minus', exon_count=4, OMIM=114210, no_hdr0=''),
  6276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6276, CurrentID=0, Status='live', Symbol='S100A5', Aliases='S100D', description='S100 calcium binding protein A5', other_designations='S100 calcium-binding protein A5', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153537147, end_position_on_the_genomic_accession=153542175, orientation='minus', exon_count=3, OMIM=176991, no_hdr0=''),
  6277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6277, CurrentID=0, Status='live', Symbol='S100A6', Aliases='2A9, 5B10, CABP, CACY, PRA', description='S100 calcium binding protein A6', other_designations='MLN 4|S100 calcium-binding protein A6 (calcyclin)|calcyclin|growth factor-inducible protein 2A9|prolactin receptor-associated protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153534600, end_position_on_the_genomic_accession=153536241, orientation='minus', exon_count=3, OMIM=114110, no_hdr0=''),
  6278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6278, CurrentID=0, Status='live', Symbol='S100A7', Aliases='PSOR1c, S100A7', description='S100 calcium binding protein A7', other_designations='S100 calcium-binding protein A7 (psoriasin 1)|psoriasin 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153457744, end_position_on_the_genomic_accession=153460661, orientation='minus', exon_count=3, OMIM=600353, no_hdr0=''),
  6279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6279, CurrentID=0, Status='live', Symbol='S100A8', Aliases='60B8AG, CAGA, CFAG, CGLA, CP-10, L1Ag, MA387, MIF, MRP8, NIF, P8', description='S100 calcium binding protein A8', other_designations='MRP-8|S100 calcium-binding protein A8 (calgranulin A)|calgranulin A|calgranulin-A|calprotectin L1L subunit|cystic fibrosis antigen|leukocyte L1 complex light chain|migration inhibitory factor-related protein 8|urinary stone protein band A', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153390032, end_position_on_the_genomic_accession=153422578, orientation='minus', exon_count=4, OMIM=123885, no_hdr0=''),
  6280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6280, CurrentID=0, Status='live', Symbol='S100A9', Aliases='60B8AG, CAGB, CFAG, CGLB, L1AG, LIAG, MAC387, MIF, MRP14, NIF, P14', description='S100 calcium binding protein A9', other_designations='MRP-14|S100 calcium-binding protein A9 (calgranulin B)|calgranulin B|calgranulin-B|calprotectin L1H subunit|leukocyte L1 complex heavy chain|migration inhibitory factor-related protein 14', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153357854, end_position_on_the_genomic_accession=153361027, orientation='plus', exon_count=3, OMIM=123886, no_hdr0=''),
  6281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6281, CurrentID=0, Status='live', Symbol='S100A10', Aliases='42C, ANX2L, ANX2LG, CAL1L, CLP11, Ca[1], GP11, P11, p10', description='S100 calcium binding protein A10', other_designations='S100 calcium binding protein A10 (annexin II ligand, calpactin I, light polypeptide (p11))|S100 calcium-binding protein A10 (annexin II ligand, calpactin I, light polypeptide (p11))|annexin II ligand, calpactin I, light polypeptide|annexin II tetramer (AIIt) p11 subunit|calpactin I light chain|calpactin-1 light chain|cellular ligand of annexin II', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151982910, end_position_on_the_genomic_accession=151994238, orientation='minus', exon_count=3, OMIM=114085, no_hdr0=''),
  6282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6282, CurrentID=0, Status='live', Symbol='S100A11', Aliases='HEL-S-43, MLN70, S100C', description='S100 calcium binding protein A11', other_designations='MLN 70|S100 calcium-binding protein A11 (calgizzarin)|calgizzarin|epididymis secretory protein Li 43|metastatic lymph node gene 70 protein|protein S100-C', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152032506, end_position_on_the_genomic_accession=152037035, orientation='minus', exon_count=3, OMIM=603114, no_hdr0=''),
  6283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6283, CurrentID=0, Status='live', Symbol='S100A12', Aliases='CAAF1, CAGC, CGRP, ENRAGE, MRP-6, MRP6, p6', description='S100 calcium binding protein A12', other_designations='EN-RAGE|S100 calcium-binding protein A12 (calgranulin C)|calcitermin|calcium-binding protein in amniotic fluid 1|calgranulin C|calgranulin-C|extracellular newly identified RAGE-binding protein|migration inhibitory factor-related protein 6|neutrophil S100 protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153373708, end_position_on_the_genomic_accession=153375599, orientation='minus', exon_count=3, OMIM=603112, no_hdr0=''),
  6284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6284, CurrentID=0, Status='live', Symbol='S100A13', Aliases='', description='S100 calcium binding protein A13', other_designations='S100 calcium-binding protein A13', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153618799, end_position_on_the_genomic_accession=153634092, orientation='minus', exon_count=10, OMIM=601989, no_hdr0=''),
  6285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6285, CurrentID=0, Status='live', Symbol='S100B', Aliases='NEF, S100, S100-B, S100beta', description='S100 calcium binding protein B', other_designations='S-100 calcium-binding protein, beta chain|S-100 protein subunit beta|S100 calcium-binding protein, beta (neural)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46598618, end_position_on_the_genomic_accession=46605122, orientation='minus', exon_count=3, OMIM=176990, no_hdr0=''),
  6286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6286, CurrentID=0, Status='live', Symbol='S100P', Aliases='MIG9', description='S100 calcium binding protein P', other_designations='S100 calcium-binding protein P|migration-inducing gene 9 protein|protein S100-E', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6693839, end_position_on_the_genomic_accession=6697170, orientation='plus', exon_count=2, OMIM=600614, no_hdr0=''),
  6287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6287, CurrentID=0, Status='live', Symbol='SAA@', Aliases='SAA, SAA4', description='serum amyloid A1 cluster', other_designations='serum amyloid A cluster', map_location='11p', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6288, CurrentID=0, Status='live', Symbol='SAA1', Aliases='PIG4, SAA, SAA2, TP53I4', description='serum amyloid A1', other_designations='serum amyloid A protein|tumor protein p53 inducible protein 4', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18266261, end_position_on_the_genomic_accession=18269977, orientation='plus', exon_count=4, OMIM=104750, no_hdr0=''),
  6289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6289, CurrentID=0, Status='live', Symbol='SAA2', Aliases='', description='serum amyloid A2', other_designations='', map_location='11p15.1-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18238233, end_position_on_the_genomic_accession=18248674, orientation='minus', exon_count=5, OMIM=104751, no_hdr0=''),
  6290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6290, CurrentID=0, Status='live', Symbol='SAA3P', Aliases='SAA3', description='serum amyloid A3 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18112472, end_position_on_the_genomic_accession=18116132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  6291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6291, CurrentID=0, Status='live', Symbol='SAA4', Aliases='C-SAA, CSAA', description='serum amyloid A4, constitutive', other_designations='constitutively expressed serum amyloid A protein', map_location='11p15.1-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18231355, end_position_on_the_genomic_accession=18236837, orientation='minus', exon_count=4, OMIM=104752, no_hdr0=''),
  6293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6293, CurrentID=0, Status='live', Symbol='VPS52', Aliases='ARE1, SAC2, SACM2L, dJ1033B10.5', description='VPS52 GARP complex subunit', other_designations='SAC2 suppressor of actin mutations 2-like protein|vacuolar protein sorting 52 homolog', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33250272, end_position_on_the_genomic_accession=33271965, orientation='minus', exon_count=21, OMIM=603443, no_hdr0=''),
  6294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6294, CurrentID=0, Status='live', Symbol='SAFB', Aliases='HAP, HET, SAF-B11, SAFB', description='scaffold attachment factor B', other_designations='HSP27 ERE-TATA-binding protein|HSP27 estrogen response element-TATA box-binding protein|Hsp27 ERE-TATA binding protein|SAB-B1|SAF-B|glutathione S-transferase fusion protein|heat-shock protein (HSP27) estrogen response element and TATA box-binding protein', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5623035, end_position_on_the_genomic_accession=5668478, orientation='plus', exon_count=21, OMIM=602895, no_hdr0=''),
  399511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399511, CurrentID=0, Status='live', Symbol='TCEA1P2', Aliases='', description='transcription elongation factor A (SII), 1 pseudogene 2', other_designations='', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37275537, end_position_on_the_genomic_accession=37278152, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6296, CurrentID=0, Status='live', Symbol='ACSM3', Aliases='SA, SAH', description='acyl-CoA synthetase medium-chain family member 3', other_designations='SA (rat hypertension-associated) homolog|SA hypertension-associated homolog|butyrate--CoA ligase 3|butyryl-coenzyme A synthetase 3|middle-chain acyl-CoA synthetase 3|protein SA homolog', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20763990, end_position_on_the_genomic_accession=20797157, orientation='plus', exon_count=15, OMIM=145505, no_hdr0=''),
  6297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6297, CurrentID=0, Status='live', Symbol='SALL2', Aliases='COLB, HSAL2, Sal-2, ZNF795, p150(Sal2)', description='spalt-like transcription factor 2', other_designations='zinc finger protein 795|zinc finger protein SALL2|zinc finger protein Spalt-2', map_location='14q11.1-q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21521080, end_position_on_the_genomic_accession=21537216, orientation='minus', exon_count=5, OMIM=602219, no_hdr0=''),
  6298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6298, CurrentID=0, Status='live', Symbol='SAI1', Aliases='MTS1, TFS1', description='suppression of anchorage independence 1', other_designations='malignant transformation suppression-1|transformation suppressor 1', map_location='1pter-p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=154280, no_hdr0=''),
  6299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6299, CurrentID=0, Status='live', Symbol='SALL1', Aliases='HEL-S-89, HSAL1, Sal-1, TBS, ZNF794', description='spalt-like transcription factor 1', other_designations='epididymis secretory protein Li 89|zinc finger protein 794|zinc finger protein SALL1|zinc finger protein Spalt-1', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51135975, end_position_on_the_genomic_accession=51152316, orientation='minus', exon_count=7, OMIM=602218, no_hdr0=''),
  6300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6300, CurrentID=0, Status='live', Symbol='MAPK12', Aliases='ERK-6, ERK3, ERK6, MAPK 12, P38GAMMA, PRKM12, SAPK-3, SAPK3', description='mitogen-activated protein kinase 12', other_designations='MAP kinase 12|MAP kinase p38 gamma|extracellular signal-regulated kinase 6|mitogen-activated protein kinase 3|mitogen-activated protein kinase p38 gamma|stress-activated protein kinase 3', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50252902, end_position_on_the_genomic_accession=50261759, orientation='minus', exon_count=12, OMIM=602399, no_hdr0=''),
  6301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6301, CurrentID=0, Status='live', Symbol='SARS', Aliases='SERRS, SERS', description='seryl-tRNA synthetase', other_designations='serine tRNA ligase 1, cytoplasmic|seryl-tRNA synthetase, cytoplasmic|seryl-tRNA(Ser/Sec) synthetase', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109213893, end_position_on_the_genomic_accession=109238182, orientation='plus', exon_count=13, OMIM=607529, no_hdr0=''),
  6302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6302, CurrentID=0, Status='live', Symbol='TSPAN31', Aliases='SAS', description='tetraspanin 31', other_designations='sarcoma amplified sequence|sarcoma-amplified sequence|transmembrane 4 protein|tspan-31', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57744090, end_position_on_the_genomic_accession=57748243, orientation='plus', exon_count=6, OMIM=181035, no_hdr0=''),
  6303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6303, CurrentID=0, Status='live', Symbol='SAT1', Aliases='DC21, KFSD, KFSDX, SAT, SSAT, SSAT-1', description='spermidine/spermine N1-acetyltransferase 1', other_designations='diamine N-acetyltransferase 1|polyamine N-acetyltransferase 1|putrescine acetyltransferase|spermidine/spermine N1-acetyltransferase alpha', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23783158, end_position_on_the_genomic_accession=23786223, orientation='plus', exon_count=7, OMIM=313020, no_hdr0=''),
  6304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6304, CurrentID=0, Status='live', Symbol='SATB1', Aliases='', description='SATB homeobox 1', other_designations='special AT-rich sequence binding protein 1 (binds to nuclear matrix/scaffold-associating DNA)|special AT-rich sequence-binding protein 1', map_location='3p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18345806, end_position_on_the_genomic_accession=18445588, orientation='minus', exon_count=16, OMIM=602075, no_hdr0=''),
  6305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6305, CurrentID=0, Status='live', Symbol='SBF1', Aliases='CMT4B3, DENND7A, MTMR5', description='SET binding factor 1', other_designations='DENN/MADD domain containing 7A|SET-binding factor 1', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50444630, end_position_on_the_genomic_accession=50480296, orientation='minus', exon_count=47, OMIM=603560, no_hdr0=''),
  6307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6307, CurrentID=0, Status='live', Symbol='MSMO1', Aliases='DESP4, ERG25, SC4MOL', description='methylsterol monooxygenase 1', other_designations='C-4 methylsterol oxidase', map_location='4q32-q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165327666, end_position_on_the_genomic_accession=165343164, orientation='plus', exon_count=6, OMIM=607545, no_hdr0=''),
  6308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6308, CurrentID=0, Status='live', Symbol='SC4MOP', Aliases='DESP4P1', description='sterol-C4-methyl oxidase pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37945884, end_position_on_the_genomic_accession=37946765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6309, CurrentID=0, Status='live', Symbol='SC5D', Aliases='ERG3, S5DESL, SC5D', description='sterol-C5-desaturase', other_designations='3beta-hydroxysteroid-delta5-desaturase|C-5 sterol desaturase|delta(7)-sterol 5-desaturase|delta(7)-sterol C5(6)-desaturase|fungal ERG3, delta-5-desaturase-like|lathosterol 5-desaturase|lathosterol dehydrogenase|sterol-C5-desaturase (ERG3 delta-5-desaturase homolog, S. cerevisiae)-like', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121292679, end_position_on_the_genomic_accession=121313410, orientation='plus', exon_count=6, OMIM=602286, no_hdr0=''),
  6310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6310, CurrentID=0, Status='live', Symbol='ATXN1', Aliases='ATX1, D6S504E, SCA1', description='ataxin 1', other_designations='alternative ataxin1|spinocerebellar ataxia type 1 protein', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16299112, end_position_on_the_genomic_accession=16761490, orientation='minus', exon_count=9, OMIM=601556, no_hdr0=''),
  6311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6311, CurrentID=0, Status='live', Symbol='ATXN2', Aliases='ASL13, ATX2, SCA2, TNRC13', description='ataxin 2', other_designations='spinocerebellar ataxia type 2 protein|trinucleotide repeat-containing gene 13 protein', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111452214, end_position_on_the_genomic_accession=111599676, orientation='minus', exon_count=25, OMIM=601517, no_hdr0=''),
  100271132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271132, CurrentID=0, Status='live', Symbol='RPS4XP5', Aliases='RPS4P5, RPS4X_1_232', description='ribosomal protein S4X pseudogene 5', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63643198, end_position_on_the_genomic_accession=63644068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6314, CurrentID=0, Status='live', Symbol='ATXN7', Aliases='ADCAII, OPCA3, SCA7', description='ataxin 7', other_designations='spinocerebellar ataxia type 7 protein', map_location='3p21.1-p12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63864557, end_position_on_the_genomic_accession=64003462, orientation='plus', exon_count=15, OMIM=607640, no_hdr0=''),
  6315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6315, CurrentID=0, Status='live', Symbol='ATXN8OS', Aliases='KLHL1AS, NCRNA00003, SCA8', description='ATXN8 opposite strand (non-protein coding)', other_designations='ataxin 8 opposite strand|kelch-like 1 antisense|non-protein coding RNA 3', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=70107213, end_position_on_the_genomic_accession=70139753, orientation='plus', exon_count=5, OMIM=603680, no_hdr0=''),
  6317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6317, CurrentID=0, Status='live', Symbol='SERPINB3', Aliases='HsT1196, SCC, SCCA-1, SCCA-PD, SCCA1, T4-A', description='serpin peptidase inhibitor, clade B (ovalbumin), member 3', other_designations='protein T4-A|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 3|squamous cell carcinoma antigen 1', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63655197, end_position_on_the_genomic_accession=63661963, orientation='minus', exon_count=8, OMIM=600517, no_hdr0=''),
  6318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6318, CurrentID=0, Status='live', Symbol='SERPINB4', Aliases='LEUPIN, PI11, SCCA-2, SCCA1, SCCA2', description='serpin peptidase inhibitor, clade B (ovalbumin), member 4', other_designations='SCCA2/SCCA1 fusion protein|peptidase inhibitor 11|protease inhibitor (leucine-serpin)|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 4|squamous cell carcinoma antigen 1|squamous cell carcinoma antigen 2', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63637259, end_position_on_the_genomic_accession=63644319, orientation='minus', exon_count=8, OMIM=600518, no_hdr0=''),
  394269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394269, CurrentID=0, Status='live', Symbol='BRCA1P1', Aliases='LBRCA1, PsiBRCA1, pseudo-BRCA1', description='BRCA1 pseudogene 1', other_designations='like BRCA1|like-BRCA1', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43168070, end_position_on_the_genomic_accession=43169954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6320, CurrentID=0, Status='live', Symbol='CLEC11A', Aliases='CLECSF3, LSLCL, P47, SCGF', description='C-type lectin domain family 11 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 3|C-type lectin domain family 11, member A|C-type lectin superfamily member 3|lymphocyte secreted C-type lectin|lymphocyte secreted long form of C-type lectin|stem cell growth factor|stem cell growth factor; lymphocyte secreted C-type lectin', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50723348, end_position_on_the_genomic_accession=50725724, orientation='plus', exon_count=4, OMIM=604713, no_hdr0=''),
  6322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6322, CurrentID=0, Status='live', Symbol='SCML1', Aliases='', description='sex comb on midleg-like 1 (Drosophila)', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17737449, end_position_on_the_genomic_accession=17754988, orientation='plus', exon_count=8, OMIM=300227, no_hdr0=''),
  6323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6323, CurrentID=0, Status='live', Symbol='SCN1A', Aliases='EIEE6, FEB3, FEB3A, FHM3, GEFSP2, HBSCI, NAC1, Nav1.1, SCN1, SMEI', description='sodium channel, voltage gated, type I alpha subunit', other_designations='sodium channel protein type I subunit alpha|sodium channel protein, brain I alpha subunit|sodium channel voltage gated type 1 alpha subunit|sodium channel, voltage-gated, type I, alpha polypeptide|sodium channel, voltage-gated, type I, alpha subunit|voltage-gated sodium channel subunit alpha Nav1.1', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165989160, end_position_on_the_genomic_accession=166149272, orientation='minus', exon_count=31, OMIM=182389, no_hdr0=''),
  6324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6324, CurrentID=0, Status='live', Symbol='SCN1B', Aliases='ATFB13, BRGDA5, GEFSP1', description='sodium channel, voltage gated, type I beta subunit', other_designations='sodium channel beta-1 subunit|sodium channel, voltage-gated, type I, beta subunit', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35030688, end_position_on_the_genomic_accession=35040449, orientation='plus', exon_count=7, OMIM=600235, no_hdr0=''),
  6326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6326, CurrentID=0, Status='live', Symbol='SCN2A', Aliases='BFIC3, BFIS3, BFNIS, EIEE11, HBA, HBSCI, HBSCII, NAC2, Na(v)1.2, Nav1.21, SCN2A2, SCN2A', description='sodium channel, voltage gated, type II alpha subunit', other_designations='HBSC II|sodium channel protein brain II subunit alpha|sodium channel protein type II subunit alpha|sodium channel protein, brain type 2 alpha subunit|sodium channel, voltage-gated, type II, alpha 1 polypeptide|sodium channel, voltage-gated, type II, alpha 2 polypeptide|sodium channel, voltage-gated, type II, alpha subunit|voltage-gated sodium channel subtype II|voltage-gated sodium channel subunit alpha Nav1.2', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165130149, end_position_on_the_genomic_accession=165392310, orientation='plus', exon_count=33, OMIM=182390, no_hdr0=''),
  6327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6327, CurrentID=0, Status='live', Symbol='SCN2B', Aliases='ATFB14', description='sodium channel, voltage gated, type II beta subunit', other_designations='neuronal voltage-gated sodium channel beta 2 subunit|sodium channel, voltage-gated, type II, beta polypeptide', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118162804, end_position_on_the_genomic_accession=118176622, orientation='minus', exon_count=4, OMIM=601327, no_hdr0=''),
  6328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6328, CurrentID=0, Status='live', Symbol='SCN3A', Aliases='NAC3, Nav1.3', description='sodium channel, voltage gated, type III alpha subunit', other_designations='brain III voltage-gated sodium channel|sodium channel protein brain III subunit alpha|sodium channel protein type III subunit alpha|sodium channel, voltage-gated, type III, alpha polypeptide|voltage-gated sodium channel subtype III|voltage-gated sodium channel subunit alpha Nav1.3', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165087520, end_position_on_the_genomic_accession=165204088, orientation='minus', exon_count=31, OMIM=182391, no_hdr0=''),
  6329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6329, CurrentID=0, Status='live', Symbol='SCN4A', Aliases='CMS16, HOKPP2, HYKPP, HYPP, NAC1A, Na(V)1.4, Nav1.4, SkM1', description='sodium channel, voltage gated, type IV alpha subunit', other_designations='CTC-264K15.6|skeletal muscle voltage-dependent sodium channel type IV alpha subunit|sodium channel protein skeletal muscle subunit alpha|sodium channel protein type IV subunit alpha|sodium channel, voltage-gated, type IV, alpha subunit|voltage-gated sodium channel subunit alpha Nav1.4', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63938554, end_position_on_the_genomic_accession=63990150, orientation='minus', exon_count=25, OMIM=603967, no_hdr0=''),
  6330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6330, CurrentID=0, Status='live', Symbol='SCN4B', Aliases='ATFB17, LQT10, Navbeta4', description='sodium channel, voltage gated, type IV beta subunit', other_designations='sodium channel, voltage-gated, type IV, beta subunit', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118133377, end_position_on_the_genomic_accession=118152915, orientation='minus', exon_count=5, OMIM=608256, no_hdr0=''),
  6331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6331, CurrentID=0, Status='live', Symbol='SCN5A', Aliases='CDCD2, CMD1E, CMPD2, HB1, HB2, HBBD, HH1, ICCD, IVF, LQT3, Nav1.5, PFHB1, SSS1, VF1', description='sodium channel, voltage gated, type V alpha subunit', other_designations='cardiac tetrodotoxin-insensitive voltage-dependent sodium channel alpha subunit|sodium channel protein cardiac muscle subunit alpha|sodium channel, voltage-gated, type V, alpha subunit|voltage-gated sodium channel subunit alpha Nav1.5', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38548058, end_position_on_the_genomic_accession=38649673, orientation='minus', exon_count=30, OMIM=600163, no_hdr0=''),
  6332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6332, CurrentID=0, Status='live', Symbol='SCN7A', Aliases='NaG, Nav2.1, Nav2.2, SCN6A', description='sodium channel, voltage gated, type VII alpha subunit', other_designations='putative voltage-gated sodium channel subunit alpha Nax|sodium channel protein cardiac and skeletal muscle subunit alpha|sodium channel protein type VII subunit alpha|sodium channel, voltage-gated, type VI, alpha polypeptide|sodium channel, voltage-gated, type VII, alpha subunit|voltage-dependent sodium channel alpha subunit', map_location='2q21-q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=166403468, end_position_on_the_genomic_accession=166494258, orientation='minus', exon_count=29, OMIM=182392, no_hdr0=''),
  6334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6334, CurrentID=0, Status='live', Symbol='SCN8A', Aliases='CERIII, CIAT, EIEE13, MED, NaCh6, Nav1.6, PN4', description='sodium channel, voltage gated, type VIII alpha subunit', other_designations='hNa6/Scn8a voltage-gated sodium channel|sodium channel, voltage gated, type VIII, alpha subunit|voltage-gated sodium channel subunit alpha Nav1.6|voltage-gated sodium channel type VIII alpha protein', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51590193, end_position_on_the_genomic_accession=51812864, orientation='plus', exon_count=30, OMIM=600702, no_hdr0=''),
  6335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6335, CurrentID=0, Status='live', Symbol='SCN9A', Aliases='ETHA, FEB3B, GEFSP7, HSAN2D, NE-NA, NENA, Nav1.7, PN1, SFNP', description='sodium channel, voltage gated, type IX alpha subunit', other_designations='hNE-Na|neuroendocrine sodium channel|peripheral sodium channel 1|sodium channel protein type IX subunit alpha|sodium channel, voltage-gated, type IX, alpha polypeptide|sodium channel, voltage-gated, type IX, alpha subunit|voltage-gated sodium channel alpha subunit Nav1.7|voltage-gated sodium channel subunit alpha Nav1.7', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=166195185, end_position_on_the_genomic_accession=166375987, orientation='minus', exon_count=28, OMIM=603415, no_hdr0=''),
  6336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6336, CurrentID=0, Status='live', Symbol='SCN10A', Aliases='FEPS2, Nav1.8, PN3, SNS, hPN3', description='sodium channel, voltage gated, type X alpha subunit', other_designations='peripheral nerve sodium channel 3|sodium channel protein type X subunit alpha|sodium channel, voltage-gated, type X, alpha polypeptide|sodium channel, voltage-gated, type X, alpha subunit|voltage-gated sodium channel subunit alpha Nav1.8', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38696891, end_position_on_the_genomic_accession=38794010, orientation='minus', exon_count=27, OMIM=604427, no_hdr0=''),
  6337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6337, CurrentID=0, Status='live', Symbol='SCNN1A', Aliases='BESC2, ENaCa, ENaCalpha, SCNEA, SCNN1', description='sodium channel, non voltage gated 1 alpha subunit', other_designations='alpha ENaC-2|alpha-ENaC|alpha-NaCH|amiloride-sensitive epithelial sodium channel alpha subunit|amiloride-sensitive sodium channel subunit alpha 2|epithelial Na(+) channel subunit alpha|nasal epithelial sodium channel alpha subunit|nonvoltage-gated sodium channel 1 subunit alpha|sodium channel, non-voltage-gated 1 alpha subunit|sodium channel, nonvoltage-gated 1 alpha', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6346843, end_position_on_the_genomic_accession=6377357, orientation='minus', exon_count=14, OMIM=600228, no_hdr0=''),
  6338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6338, CurrentID=0, Status='live', Symbol='SCNN1B', Aliases='BESC1, ENaCb, ENaCbeta, SCNEB', description='sodium channel, non voltage gated 1 beta subunit', other_designations='amiloride-sensitive sodium channel subunit beta 1|beta-ENaC|beta-NaCH|epithelial Na(+) channel subunit beta|epithelial sodium channel beta-2 subunit|epithelial sodium channel beta-3 subunit|nasal epithelial sodium channel beta subunit|nonvoltage-gated sodium channel 1 subunit beta|sodium channel, non-voltage-gated 1, beta subunit|sodium channel, nonvoltage-gated 1, beta', map_location='16p12.2-p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23302270, end_position_on_the_genomic_accession=23381299, orientation='plus', exon_count=15, OMIM=600760, no_hdr0=''),
  6339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6339, CurrentID=0, Status='live', Symbol='SCNN1D', Aliases='ENaCd, ENaCdelta, SCNED, dNaCh', description='sodium channel, non voltage gated 1 delta subunit', other_designations='delta-ENaC|delta-NaCH|epithelial Na(+) channel subunit delta|nonvoltage-gated sodium channel 1 subunit delta|sodium channel, non-voltage-gated 1, delta subunit|sodium channel, nonvoltage-gated 1, delta|sodium channel, voltage-gated, type I, delta polypeptide', map_location='1p36.3-p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1280415, end_position_on_the_genomic_accession=1292029, orientation='plus', exon_count=20, OMIM=601328, no_hdr0=''),
  6340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6340, CurrentID=0, Status='live', Symbol='SCNN1G', Aliases='BESC3, ENaCg, ENaCgamma, PHA1, SCNEG', description='sodium channel, non voltage gated 1 gamma subunit', other_designations='ENaC gamma subunit|amiloride-sensitive epithelial sodium channel gamma subunit|amiloride-sensitive sodium channel gamma-subunit|epithelial Na(+) channel subunit gamma|gamma-ENaC|gamma-NaCH|nonvoltage-gated sodium channel 1 subunit gamma|sodium channel, non-voltage-gated 1, gamma subunit|sodium channel, nonvoltage-gated 1, gamma', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23182719, end_position_on_the_genomic_accession=23216879, orientation='plus', exon_count=13, OMIM=600761, no_hdr0=''),
  6341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6341, CurrentID=0, Status='live', Symbol='SCO1', Aliases='SCOD1', description='SCO1 cytochrome c oxidase assembly protein', other_designations='SCO cytochrome oxidase deficient homolog 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10680332, end_position_on_the_genomic_accession=10697568, orientation='minus', exon_count=6, OMIM=603644, no_hdr0=''),
  6342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6342, CurrentID=0, Status='live', Symbol='SCP2', Aliases='NLTP, NSL-TP, SCP-2, SCP-CHI, SCP-X, SCPX', description='sterol carrier protein 2', other_designations='propanoyl-CoA C-acyltransferase|sterol carrier protein X', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52927229, end_position_on_the_genomic_accession=53051617, orientation='plus', exon_count=18, OMIM=184755, no_hdr0=''),
  6343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6343, CurrentID=0, Status='live', Symbol='SCT', Aliases='', description='secretin', other_designations='prepro-secretin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=624989, end_position_on_the_genomic_accession=627240, orientation='minus', exon_count=5, OMIM=182099, no_hdr0=''),
  6344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6344, CurrentID=0, Status='live', Symbol='SCTR', Aliases='SR', description='secretin receptor', other_designations='pancreatic secretin receptor', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119439843, end_position_on_the_genomic_accession=119524452, orientation='minus', exon_count=14, OMIM=182098, no_hdr0=''),
  6345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6345, CurrentID=0, Status='live', Symbol='SRL', Aliases='', description='sarcalumenin', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4189374, end_position_on_the_genomic_accession=4242080, orientation='minus', exon_count=6, OMIM=604992, no_hdr0=''),
  6346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6346, CurrentID=0, Status='live', Symbol='CCL1', Aliases='I-309, P500, SCYA1, SISe, TCA3', description='chemokine (C-C motif) ligand 1', other_designations='T lymphocyte-secreted protein I-309|inflammatory cytokine I-309|small inducible cytokine A1 (I-309, homologous to mouse Tca-3)', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34360328, end_position_on_the_genomic_accession=34363233, orientation='minus', exon_count=3, OMIM=182281, no_hdr0=''),
  6347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6347, CurrentID=0, Status='live', Symbol='CCL2', Aliases='GDCF-2, HC11, HSMCR30, MCAF, MCP-1, MCP1, SCYA2, SMC-CF', description='chemokine (C-C motif) ligand 2', other_designations='monocyte chemoattractant protein 1|monocyte chemoattractant protein-1|monocyte chemotactic and activating factor|monocyte chemotactic protein 1|monocyte secretory protein JE|small inducible cytokine A2 (monocyte chemotactic protein 1, homologous to mouse Sig-je)|small inducible cytokine subfamily A (Cys-Cys), member 2|small-inducible cytokine A2', map_location='17q11.2-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34255277, end_position_on_the_genomic_accession=34257203, orientation='plus', exon_count=3, OMIM=158105, no_hdr0=''),
  6348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6348, CurrentID=0, Status='live', Symbol='CCL3', Aliases='G0S19-1, LD78ALPHA, MIP-1-alpha, MIP1A, SCYA3', description='chemokine (C-C motif) ligand 3', other_designations='G0/G1 switch regulatory protein 19-1|PAT 464.1|SIS-beta|macrophage inflammatory protein 1-alpha|small inducible cytokine A3 (homologous to mouse Mip-1a)|tonsillar lymphocyte LD78 alpha protein', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36088256, end_position_on_the_genomic_accession=36090160, orientation='minus', exon_count=3, OMIM=182283, no_hdr0=''),
  6349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6349, CurrentID=0, Status='live', Symbol='CCL3L1', Aliases='464.2, D17S1718, G0S19-2, LD78, LD78-beta(1-70), LD78BETA, MIP1AP, SCYA3L, SCYA3L1', description='chemokine (C-C motif) ligand 3-like 1', other_designations='G0/G1 switch regulatory protein 19-2|small inducible cytokine A3-like 1|tonsillar lymphocyte LD78 beta protein', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601395, no_hdr0=''),
  6351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6351, CurrentID=0, Status='live', Symbol='CCL4', Aliases='ACT2, AT744.1, G-26, HC21, LAG-1, LAG1, MIP-1-beta, MIP1B, MIP1B1, SCYA2, SCYA4', description='chemokine (C-C motif) ligand 4', other_designations='G-26 T-lymphocyte-secreted protein|MIP-1-beta(1-69)|PAT 744|SIS-gamma|T-cell activation protein 2|lymphocyte activation gene 1 protein|macrophage inflammatory protein 1-beta|secreted protein G-26|small inducible cytokine A4 (homologous to mouse Mip-1b)', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36103827, end_position_on_the_genomic_accession=36105621, orientation='plus', exon_count=3, OMIM=182284, no_hdr0=''),
  6352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6352, CurrentID=0, Status='live', Symbol='CCL5', Aliases='D17S136E, RANTES, SCYA5, SIS-delta, SISd, TCP228, eoCP', description='chemokine (C-C motif) ligand 5', other_designations='T-cell specific protein p288|beta-chemokine RANTES|eosinophil chemotactic cytokine|regulated upon activation, normally T-expressed, and presumably secreted|small inducible cytokine subfamily A (Cys-Cys), member 5', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35871491, end_position_on_the_genomic_accession=35880373, orientation='minus', exon_count=4, OMIM=187011, no_hdr0=''),
  6354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6354, CurrentID=0, Status='live', Symbol='CCL7', Aliases='FIC, MARC, MCP-3, MCP3, NC28, SCYA6, SCYA7', description='chemokine (C-C motif) ligand 7', other_designations='monocyte chemoattractant protein 3|monocyte chemotactic protein 3|small inducible cytokine A7 (monocyte chemotactic protein 3)|small-inducible cytokine A7', map_location='17q11.2-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34270216, end_position_on_the_genomic_accession=34272242, orientation='plus', exon_count=3, OMIM=158106, no_hdr0=''),
  6355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6355, CurrentID=0, Status='live', Symbol='CCL8', Aliases='HC14, MCP-2, MCP2, SCYA10, SCYA8', description='chemokine (C-C motif) ligand 8', other_designations='monocyte chemoattractant protein 2|monocyte chemotactic protein 2|small inducible cytokine subfamily A (Cys-Cys), member 8 (monocyte chemotactic protein 2)|small-inducible cytokine A8', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34319047, end_position_on_the_genomic_accession=34321402, orientation='plus', exon_count=3, OMIM=602283, no_hdr0=''),
  6356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6356, CurrentID=0, Status='live', Symbol='CCL11', Aliases='SCYA11', description='chemokine (C-C motif) ligand 11', other_designations='eosinophil chemotactic protein|eotaxin-1|small inducible cytokine subfamily A (Cys-Cys), member 11 (eotaxin)', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34285668, end_position_on_the_genomic_accession=34288180, orientation='plus', exon_count=3, OMIM=601156, no_hdr0=''),
  6357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6357, CurrentID=0, Status='live', Symbol='CCL13', Aliases='CKb10, MCP-4, NCC-1, NCC1, SCYA13, SCYL1', description='chemokine (C-C motif) ligand 13', other_designations='CK-beta-10|monocyte chemoattractant protein 4|monocyte chemotactic protein 4|new CC chemokine 1|small inducible cytokine subfamily A (Cys-Cys), member 13|small-inducible cytokine A13', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34356452, end_position_on_the_genomic_accession=34358610, orientation='plus', exon_count=3, OMIM=601391, no_hdr0=''),
  6358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6358, CurrentID=0, Status='live', Symbol='CCL14', Aliases='CC-1, CC-3, CKB1, HCC-1, HCC-1(1-74), HCC-1/HCC-3, HCC-3, MCIF, NCC-2, NCC2, SCYA14, SCYL2, SY14', description='chemokine (C-C motif) ligand 14', other_designations='chemokine CC-1/CC-3|chemokine CC-3|hemofiltrate CC chemokine 1|new CC chemokine 2|small inducible cytokine subfamily A (Cys-Cys), member 14|small-inducible cytokine A14', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35983656, end_position_on_the_genomic_accession=35986728, orientation='minus', exon_count=4, OMIM=601392, no_hdr0=''),
  6359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6359, CurrentID=0, Status='live', Symbol='CCL15', Aliases='HCC-2, HMRP-2B, LKN-1, LKN1, MIP-1 delta, MIP-1D, MIP-5, MRP-2B, NCC-3, NCC3, SCYA15, SCYL3, SY15', description='chemokine (C-C motif) ligand 15', other_designations='chemokine CC-2|leukotactin 1|macrophage inflammatory protein 5|new CC chemokine 3|small inducible cytokine subfamily A (Cys-Cys), member 15|small-inducible cytokine A15', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35997582, end_position_on_the_genomic_accession=36002038, orientation='minus', exon_count=4, OMIM=601393, no_hdr0=''),
  6360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6360, CurrentID=0, Status='live', Symbol='CCL16', Aliases='CKb12, HCC-4, ILINCK, LCC-1, LEC, LMC, Mtn-1, NCC-4, NCC4, SCYA16, SCYL4', description='chemokine (C-C motif) ligand 16', other_designations='IL-10-inducible chemokine|chemokine LEC|liver CC chemokine-1|liver-expressed chemokine|lymphocyte and monocyte chemoattractant|monotactin-1|new CC chemokine 4|small inducible cytokine subfamily A (Cys-Cys), member 16|small-inducible cytokine A16', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35976493, end_position_on_the_genomic_accession=35982889, orientation='minus', exon_count=4, OMIM=601394, no_hdr0=''),
  6361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6361, CurrentID=0, Status='live', Symbol='CCL17', Aliases='A-152E5.3, ABCD-2, SCYA17, TARC', description='chemokine (C-C motif) ligand 17', other_designations='CC chemokine TARC|T cell-directed CC chemokine|small inducible cytokine subfamily A (Cys-Cys), member 17|small-inducible cytokine A17|thymus and activation-regulated chemokine', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57396190, end_position_on_the_genomic_accession=57416063, orientation='plus', exon_count=6, OMIM=601520, no_hdr0=''),
  6362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6362, CurrentID=0, Status='live', Symbol='CCL18', Aliases='AMAC-1, AMAC1, CKb7, DC-CK1, DCCK1, MIP-4, PARC, SCYA18', description='chemokine (C-C motif) ligand 18', other_designations='CC chemokine PARC|CC chemokine ligand 18|alternative macrophage activation-associated CC chemokine 1|chemokine (C-C motif) ligand 18 (pulmonary and activation-regulated)|chemokine (C-C), dendritic|dendritic cell chemokine 1|macrophage inflammatory protein 4|pulmonary and activation-regulated chemokine|small inducible cytokine A18|small inducible cytokine subfamily A (Cys-Cys), member 18, pulmonary and activation-regulated|small-inducible cytokine A18', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36064272, end_position_on_the_genomic_accession=36071481, orientation='plus', exon_count=3, OMIM=603757, no_hdr0=''),
  6363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6363, CurrentID=0, Status='live', Symbol='CCL19', Aliases='CKb11, ELC, MIP-3b, MIP3B, SCYA19', description='chemokine (C-C motif) ligand 19', other_designations='CC chemokine ligand 19|CK beta-11|EBI1 ligand chemokine|EBI1-ligand chemokine|MIP-3-beta|beta chemokine exodus-3|beta-chemokine exodus-3|epstein-Barr virus-induced molecule 1 ligand chemokine|exodus-3|macrophage inflammatory protein 3 beta|macrophage inflammatory protein 3-beta|small inducible cytokine subfamily A (Cys-Cys), member 19|small-inducible cytokine A19', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34689570, end_position_on_the_genomic_accession=34691277, orientation='minus', exon_count=4, OMIM=602227, no_hdr0=''),
  6364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6364, CurrentID=0, Status='live', Symbol='CCL20', Aliases='CKb4, Exodus, LARC, MIP-3-alpha, MIP-3a, MIP3A, SCYA20, ST38', description='chemokine (C-C motif) ligand 20', other_designations='CC chemokine LARC|beta chemokine exodus-1|liver and activation-regulated chemokine|macrophage inflammatory protein 3 alpha|small inducible cytokine subfamily A (Cys-Cys), member 20|small-inducible cytokine A20', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227813842, end_position_on_the_genomic_accession=227817564, orientation='plus', exon_count=4, OMIM=601960, no_hdr0=''),
  6365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6365, CurrentID=0, Status='live', Symbol='SCZD3', Aliases='', description='schizophrenia disorder 3', other_designations='Schizophrenia susceptibility locus, chromosome 6-related', map_location='6p24-p22', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600511, no_hdr0=''),
  6366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6366, CurrentID=0, Status='live', Symbol='CCL21', Aliases='6Ckine, CKb9, ECL, SCYA21, SLC, TCA4', description='chemokine (C-C motif) ligand 21', other_designations='Efficient Chemoattractant for Lymphocytes|beta chemokine exodus-2|secondary lymphoid tissue chemokine|small inducible cytokine subfamily A (Cys-Cys), member 21', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34709005, end_position_on_the_genomic_accession=34710167, orientation='minus', exon_count=4, OMIM=602737, no_hdr0=''),
  6367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6367, CurrentID=0, Status='live', Symbol='CCL22', Aliases='A-152E5.1, ABCD-1, DC/B-CK, MDC, SCYA22, STCP-1', description='chemokine (C-C motif) ligand 22', other_designations='CC chemokine STCP-1|MDC(1-69)|macrophage-derived chemokine|small inducible cytokine A22|small inducible cytokine subfamily A (Cys-Cys), member 22|small-inducible cytokine A22|stimulated T cell chemotactic protein 1|stimulated T-cell chemotactic protein 1', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57358783, end_position_on_the_genomic_accession=57366190, orientation='plus', exon_count=3, OMIM=602957, no_hdr0=''),
  6368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6368, CurrentID=0, Status='live', Symbol='CCL23', Aliases='CK-BETA-8, CKb8, Ckb-8, Ckb-8-1, MIP-3, MIP3, MPIF-1, SCYA23, hmrp-2a', description='chemokine (C-C motif) ligand 23', other_designations='C6 beta-chemokine|macrophage inflammatory protein 3|myeloid progenitor inhibitory factor 1|small inducible cytokine subfamily A (Cys-Cys), member 23', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36013056, end_position_on_the_genomic_accession=36017975, orientation='minus', exon_count=4, OMIM=602494, no_hdr0=''),
  6369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6369, CurrentID=0, Status='live', Symbol='CCL24', Aliases='Ckb-6, MPIF-2, MPIF2, SCYA24', description='chemokine (C-C motif) ligand 24', other_designations='CK-beta-6|eosinophil chemotactic protein 2|eotaxin-2|myeloid progenitor inhibitory factor 2|small inducible cytokine subfamily A (Cys-Cys), member 24|small-inducible cytokine A24', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75811429, end_position_on_the_genomic_accession=75823356, orientation='minus', exon_count=6, OMIM=602495, no_hdr0=''),
  6370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6370, CurrentID=0, Status='live', Symbol='CCL25', Aliases='Ckb15, SCYA25, TECK', description='chemokine (C-C motif) ligand 25', other_designations='Ck beta-15|TECKvar|chemokine TECK|small inducible cytokine subfamily A (Cys-Cys), member 25|small-inducible cytokine A25|thymus expressed chemokine|thymus-expressed chemokine', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8052340, end_position_on_the_genomic_accession=8062663, orientation='plus', exon_count=7, OMIM=602565, no_hdr0=''),
  22907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22907, CurrentID=0, Status='live', Symbol='DHX30', Aliases='DDX30, RETCOR', description='DEAH (Asp-Glu-Ala-His) box helicase 30', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 30|DEAH (Asp-Glu-Ala-His) box polypeptide 30|DEAH box protein 30|retina co-repressor', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47802909, end_position_on_the_genomic_accession=47850196, orientation='plus', exon_count=27, OMIM=616423, no_hdr0=''),
  6372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6372, CurrentID=0, Status='live', Symbol='CXCL6', Aliases='CKA-3, GCP-2, GCP2, SCYB6', description='chemokine (C-X-C motif) ligand 6', other_designations='Small inducible cytokine subfamily B (Cys-X-Cys), member b|chemokine (C-X-C motif) ligand 6 (granulocyte chemotactic protein 2)|chemokine alpha 3|granulocyte chemotactic protein 2|small inducible cytokine subfamily B (Cys-X-Cys), member 6 (granulocyte chemotactic protein 2)|small-inducible cytokine B6', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73836556, end_position_on_the_genomic_accession=73838760, orientation='plus', exon_count=4, OMIM=138965, no_hdr0=''),
  6373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6373, CurrentID=0, Status='live', Symbol='CXCL11', Aliases='H174, I-TAC, IP-9, IP9, SCYB11, SCYB9B, b-R1', description='chemokine (C-X-C motif) ligand 11', other_designations='beta-R1|interferon gamma-inducible protein 9|interferon-inducible T-cell alpha chemoattractant|small inducible cytokine B11|small inducible cytokine subfamily B (Cys-X-Cys), member 11|small inducible cytokine subfamily B (Cys-X-Cys), member 9B|small-inducible cytokine B11', map_location='4q21.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76033682, end_position_on_the_genomic_accession=76036197, orientation='minus', exon_count=4, OMIM=604852, no_hdr0=''),
  6374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6374, CurrentID=0, Status='live', Symbol='CXCL5', Aliases='ENA-78, SCYB5', description='chemokine (C-X-C motif) ligand 5', other_designations='epithelial-derived neutrophil-activating protein 78|neutrophil-activating peptide ENA-78|neutrophil-activating protein 78|small inducible cytokine subfamily B (Cys-X-Cys), member 5 (epithelial-derived neutrophil-activating peptide 78)', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73995642, end_position_on_the_genomic_accession=73998729, orientation='minus', exon_count=4, OMIM=600324, no_hdr0=''),
  6375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6375, CurrentID=0, Status='live', Symbol='XCL1', Aliases='ATAC, LPTN, LTN, SCM-1, SCM-1a, SCM1, SCM1A, SCYC1', description='chemokine (C motif) ligand 1', other_designations='SCM-1-alpha|XC chemokine ligand 1|c motif chemokine 1|cytokine SCM-1|lymphotaxin|single cysteine motif 1a|small inducible cytokine subfamily C, member 1 (lymphotactin)|small-inducible cytokine C1', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168576061, end_position_on_the_genomic_accession=168582077, orientation='plus', exon_count=4, OMIM=600250, no_hdr0=''),
  6376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6376, CurrentID=0, Status='live', Symbol='CX3CL1', Aliases='ABCD-3, C3Xkine, CXC3, CXC3C, NTN, NTT, SCYD1, fractalkine, neurotactin', description='chemokine (C-X3-C motif) ligand 1', other_designations='C-X3-C motif chemokine 1|CX3C membrane-anchored chemokine|small inducible cytokine subfamily D (Cys-X3-Cys), member 1 (fractalkine, neurotactin)|small inducible cytokine subfamily D (Cys-X3-Cys), member-1|small-inducible cytokine D1', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57372461, end_position_on_the_genomic_accession=57385048, orientation='plus', exon_count=3, OMIM=601880, no_hdr0=''),
  6377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6377, CurrentID=0, Status='live', Symbol='SCZD1', Aliases='', description='schizophrenia disorder 1', other_designations='Schizophrenia susceptibility locus, chromosome 5-related', map_location='5q11.2-q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=181510, no_hdr0=''),
  6378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6378, CurrentID=0, Status='live', Symbol='SCZD2', Aliases='', description='schizophrenia disorder 2', other_designations='Schizophrenia susceptibility locus, chromosome 11-related', map_location='11q14-q21', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603342, no_hdr0=''),
  100271143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271143, CurrentID=0, Status='live', Symbol='RPS15AP8', Aliases='RPS15A_2_39', description='ribosomal protein S15a pseudogene 8', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42412352, end_position_on_the_genomic_accession=42412824, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100820631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820631, CurrentID=0, Status='live', Symbol='MLSM7', Aliases='C7DELq, DEL7q', description='Myelodysplasia and leukemia syndrome with monosomy 7', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=252270, no_hdr0=''),
  6382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6382, CurrentID=0, Status='live', Symbol='SDC1', Aliases='CD138, SDC, SYND1, syndecan', description='syndecan 1', other_designations='CD138 antigen|heparan sulfate proteoglycan fibroblast growth factor receptor|syndecan proteoglycan 1', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20200797, end_position_on_the_genomic_accession=20225433, orientation='minus', exon_count=9, OMIM=186355, no_hdr0=''),
  6383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6383, CurrentID=0, Status='live', Symbol='SDC2', Aliases='CD362, HSPG, HSPG1, SYND2', description='syndecan 2', other_designations='cell surface-associated heparan sulfate proteoglycan 1|fibroglycan|heparan sulfate proteoglycan 1, cell surface-associated|heparan sulfate proteoglycan core protein|syndecan proteoglycan 2', map_location='8q22-q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96493351, end_position_on_the_genomic_accession=96611809, orientation='plus', exon_count=9, OMIM=142460, no_hdr0=''),
  6385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6385, CurrentID=0, Status='live', Symbol='SDC4', Aliases='SYND4', description='syndecan 4', other_designations='amphiglycan|ryudocan amphiglycan|ryudocan core protein|syndecan 4 (amphiglycan, ryudocan)|syndecan proteoglycan 4', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45325285, end_position_on_the_genomic_accession=45348424, orientation='minus', exon_count=5, OMIM=600017, no_hdr0=''),
  6386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6386, CurrentID=0, Status='live', Symbol='SDCBP', Aliases='MDA-9, MDA9, ST1, SYCL, TACIP18', description='syndecan binding protein', other_designations='melanoma differentiation associated protein-9|melanoma differentiation-associated protein 9|pro-TGF-alpha cytoplasmic domain-interacting protein 18|scaffold protein Pbp1|syndecan binding protein (syntenin)|syndecan-binding protein 1', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58553169, end_position_on_the_genomic_accession=58582860, orientation='plus', exon_count=10, OMIM=602217, no_hdr0=''),
  6387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6387, CurrentID=0, Status='live', Symbol='CXCL12', Aliases='IRH, PBSF, SCYB12, SDF1, TLSF, TPAR1', description='chemokine (C-X-C motif) ligand 12', other_designations='intercrine reduced in hepatomas|pre-B cell growth-stimulating factor', map_location='10q11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44370153, end_position_on_the_genomic_accession=44385097, orientation='minus', exon_count=6, OMIM=600835, no_hdr0=''),
  6388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6388, CurrentID=0, Status='live', Symbol='SDF2', Aliases='', description='stromal cell-derived factor 2', other_designations='SDF-2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28648356, end_position_on_the_genomic_accession=28662189, orientation='minus', exon_count=4, OMIM=602934, no_hdr0=''),
  6389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6389, CurrentID=0, Status='live', Symbol='SDHA', Aliases='CMD1GG, FP, PGL5, SDH1, SDH2, SDHF', description='succinate dehydrogenase complex subunit A, flavoprotein (Fp)', other_designations='flavoprotein subunit of complex II|succinate dehydrogenase complex, subunit A, flavoprotein (Fp)', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=218223, end_position_on_the_genomic_accession=264816, orientation='plus', exon_count=16, OMIM=600857, no_hdr0=''),
  6390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6390, CurrentID=0, Status='live', Symbol='SDHB', Aliases='CWS2, IP, PGL4, SDH, SDH1, SDH2, SDHIP', description='succinate dehydrogenase complex subunit B, iron sulfur (Ip)', other_designations='iron-sulfur subunit of complex II|succinate dehydrogenase complex, subunit B, iron sulfur (Ip)', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17018722, end_position_on_the_genomic_accession=17054170, orientation='minus', exon_count=8, OMIM=185470, no_hdr0=''),
  6391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6391, CurrentID=0, Status='live', Symbol='SDHC', Aliases='CYB560, CYBL, PGL3, QPS1, SDH3', description='succinate dehydrogenase complex subunit C', other_designations='cytochrome B large subunit of complex II|integral membrane protein CII-3b|succinate dehydrgenase cytochrome b|succinate dehydrogenase 3, integral membrane subunit|succinate dehydrogenase complex subunit C integral membrane protein 15kDa|succinate dehydrogenase complex, subunit C, integral membrane protein, 15kD|succinate dehydrogenase complex, subunit C, integral membrane protein, 15kDa|succinate-ubiquinone oxidoreductase cytochrome B large subunit', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161314376, end_position_on_the_genomic_accession=161364751, orientation='plus', exon_count=7, OMIM=602413, no_hdr0=''),
  6392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6392, CurrentID=0, Status='live', Symbol='SDHD', Aliases='CBT1, CII-4, CWS3, PGL, PGL1, QPs3, SDH4, cybS', description='succinate dehydrogenase complex subunit D, integral membrane protein', other_designations='succinate dehydrogenase complex subunit D integral membrane protein|succinate dehydrogenase complex, subunit D, integral membrane protein|succinate-ubiquinone oxidoreductase cytochrome b small subunit|succinate-ubiquinone reductase membrane anchor subunit', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112086824, end_position_on_the_genomic_accession=112095801, orientation='plus', exon_count=6, OMIM=602690, no_hdr0=''),
  6395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6395, CurrentID=0, Status='live', Symbol='SEA', Aliases='', description='S13 erythroblastosis (avian) oncogene homolog', other_designations='Oncogene SEA (S13 avian erythroblastosis)|S13 avian erythroblastosis oncogene homolog|S13 erythroblastosis oncogene homolog', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=165110, no_hdr0=''),
  6396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6396, CurrentID=0, Status='live', Symbol='SEC13', Aliases='D3S1231EL1, SEC13R, npp-20, SEC13', description='SEC13 homolog, nuclear pore and COPII coat complex component', other_designations='SEC13 homolog|SEC13 homolog, nuclear pore and COPII coating complex component|SEC13-like 1 isoform|SEC13-like protein 1|SEC13-related protein', map_location='3p25-p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10300929, end_position_on_the_genomic_accession=10321188, orientation='minus', exon_count=13, OMIM=600152, no_hdr0=''),
  6397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6397, CurrentID=0, Status='live', Symbol='SEC14L1', Aliases='PRELID4A, SEC14L', description='SEC14-like lipid binding 1', other_designations='', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77088643, end_position_on_the_genomic_accession=77217101, orientation='plus', exon_count=21, OMIM=601504, no_hdr0=''),
  6398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6398, CurrentID=0, Status='live', Symbol='SECTM1', Aliases='K12', description='secreted and transmembrane 1', other_designations='type 1a transmembrane protein', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82321024, end_position_on_the_genomic_accession=82334045, orientation='minus', exon_count=6, OMIM=602602, no_hdr0=''),
  6399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6399, CurrentID=0, Status='live', Symbol='TRAPPC2', Aliases='MIP2A, SEDL, SEDTP1, TRS20, ZNF547L, hYP38334, TRAPPC2', description='trafficking protein particle complex 2', other_designations='sedlin', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13712242, end_position_on_the_genomic_accession=13734635, orientation='minus', exon_count=7, OMIM=300202, no_hdr0=''),
  6400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6400, CurrentID=0, Status='live', Symbol='SEL1L', Aliases='PRO1063, SEL1-LIKE1, SEL1L', description='sel-1 suppressor of lin-12-like (C. elegans)', other_designations='Suppressor of lin 12 (sel-1), C. elegans, homolog of|sel-1 suppressor of lin-12-like 1|suppressor of lin-12-like protein 1', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81471547, end_position_on_the_genomic_accession=81533861, orientation='minus', exon_count=21, OMIM=602329, no_hdr0=''),
  6401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6401, CurrentID=0, Status='live', Symbol='SELE', Aliases='CD62E, ELAM, ELAM1, ESEL, LECAM2', description='selectin E', other_designations='CD62 antigen-like family member E|ELAM-1|endothelial adhesion molecule 1|endothelial leukocyte adhesion molecule 1|leukocyte endothelial cell adhesion molecule 2|leukocyte-endothelial cell adhesion molecule 2', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169722640, end_position_on_the_genomic_accession=169734079, orientation='minus', exon_count=14, OMIM=131210, no_hdr0=''),
  6402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6402, CurrentID=0, Status='live', Symbol='SELL', Aliases='CD62L, LAM1, LECAM1, LEU8, LNHR, LSEL, LYAM1, PLNHR, TQ1', description='selectin L', other_designations='CD62 antigen-like family member L|gp90-MEL|leukocyte surface antigen Leu-8|leukocyte-endothelial cell adhesion molecule 1|lymph node homing receptor|lymphocyte adhesion molecule 1|pln homing receptor', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169690665, end_position_on_the_genomic_accession=169711702, orientation='minus', exon_count=9, OMIM=153240, no_hdr0=''),
  6403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6403, CurrentID=0, Status='live', Symbol='SELP', Aliases='CD62, CD62P, GMP140, GRMP, LECAM3, PADGEM, PSEL', description='selectin P', other_designations='CD62 antigen-like family member P|GMP-140|granule membrane protein 140|granulocyte membrane protein|leukocyte-endothelial cell adhesion molecule 3|platelet activation dependent granule-external membrane protein|platelet alpha-granule membrane protein|selectin P (granule membrane protein 140kDa, antigen CD62)', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169588849, end_position_on_the_genomic_accession=169630139, orientation='minus', exon_count=18, OMIM=173610, no_hdr0=''),
  6404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6404, CurrentID=0, Status='live', Symbol='SELPLG', Aliases='CD162, CLA, PSGL-1, PSGL1', description='selectin P ligand', other_designations='cutaneous lymphocyte-associated associated antigen', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108621895, end_position_on_the_genomic_accession=108633954, orientation='minus', exon_count=3, OMIM=600738, no_hdr0=''),
  6405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6405, CurrentID=0, Status='live', Symbol='SEMA3F', Aliases='SEMA-IV, SEMA4, SEMAK', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3F', other_designations='sema III/F|sema IV|sema domain, immunoglobulin domain (Ig), short basic domain, secreted, 3F|semaphorin III/F|semaphorin IV', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50155063, end_position_on_the_genomic_accession=50189075, orientation='plus', exon_count=23, OMIM=601124, no_hdr0=''),
  6406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6406, CurrentID=0, Status='live', Symbol='SEMG1', Aliases='CT103, SEMG, SGI, dJ172H20.2', description='semenogelin I', other_designations='SgI-29|cancer/testis antigen 103|semen coagulating protein', map_location='20q12-q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45206997, end_position_on_the_genomic_accession=45209773, orientation='plus', exon_count=3, OMIM=182140, no_hdr0=''),
  6407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6407, CurrentID=0, Status='live', Symbol='SEMG2', Aliases='SGII', description='semenogelin II', other_designations='Semenogelin 2', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45221369, end_position_on_the_genomic_accession=45224458, orientation='plus', exon_count=3, OMIM=182141, no_hdr0=''),
  6409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6409, CurrentID=0, Status='live', Symbol='SEN2', Aliases='', description='senescence (cellular)-related 2', other_designations='', map_location='1q42-q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6410, CurrentID=0, Status='live', Symbol='SEN3', Aliases='', description='senescence (cellular)-related 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6411, CurrentID=0, Status='live', Symbol='SEN6', Aliases='', description='senescence (cellular)-related 6', other_designations='', map_location='6q26-q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6412, CurrentID=0, Status='live', Symbol='SEN6A', Aliases='', description='senescence (cellular)-related 6A', other_designations='', map_location='6q14-q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6413, CurrentID=0, Status='live', Symbol='SEN6B', Aliases='', description='senescence (cellular)-related 6B', other_designations='', map_location='6q21-q22', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6414, CurrentID=0, Status='live', Symbol='SEPP1', Aliases='SELP, SEPP, SeP', description='selenoprotein P, plasma, 1', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42799880, end_position_on_the_genomic_accession=42811922, orientation='minus', exon_count=7, OMIM=601484, no_hdr0=''),
  6415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6415, CurrentID=0, Status='live', Symbol='SEPW1', Aliases='selW', description='selenoprotein W, 1', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47778585, end_position_on_the_genomic_accession=47784682, orientation='plus', exon_count=6, OMIM=603235, no_hdr0=''),
  6416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6416, CurrentID=0, Status='live', Symbol='MAP2K4', Aliases='JNKK, JNKK1, MAPKK4, MEK4, MKK4, PRKMK4, SAPKK-1, SAPKK1, SEK1, SERK1, SKK1', description='mitogen-activated protein kinase kinase 4', other_designations='JNK-activated kinase 1|JNK-activating kinase 1|MAP kinase kinase 4|MAPK/ERK kinase 4|MAPKK 4|MEK 4|SAPK/ERK kinase 1|c-Jun N-terminal kinase kinase 1|stress-activated protein kinase kinase 1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12020818, end_position_on_the_genomic_accession=12143831, orientation='plus', exon_count=15, OMIM=601335, no_hdr0=''),
  6418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6418, CurrentID=0, Status='live', Symbol='SET', Aliases='2PP2A, I2PP2A, IGAAD, IPP2A2, PHAPII, TAF-I, TAF-IBETA', description='SET nuclear proto-oncogene', other_designations='HLA-DR-associated protein II|SET nuclear oncogene|SET translocation (myeloid leukemia-associated)|Template-Activating Factor-I, chromatin remodelling factor|inhibitor of granzyme A-activated DNase|inhibitor-2 of protein phosphatase-2A|phosphatase 2A inhibitor I2PP2A|protein phosphatase type 2A inhibitor', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128683432, end_position_on_the_genomic_accession=128696396, orientation='plus', exon_count=12, OMIM=600960, no_hdr0=''),
  6419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6419, CurrentID=0, Status='live', Symbol='SETMAR', Aliases='HsMar1, METNASE, Mar1', description='SET domain and mariner transposase fusion gene', other_designations='SET domain and mariner transposase fusion gene-containing protein|SET domain and mariner transposase fusion protein', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4303304, end_position_on_the_genomic_accession=4317265, orientation='plus', exon_count=6, OMIM=609834, no_hdr0=''),
  137492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137492, CurrentID=0, Status='live', Symbol='VPS37A', Aliases='HCRP1, PQBP2, SPG53', description='vacuolar protein sorting 37 homolog A (S. cerevisiae)', other_designations='ESCRT-I complex subunit VPS37A|hepatocellular carcinoma-related protein 1|polyglutamine binding protein 2', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17246892, end_position_on_the_genomic_accession=17298024, orientation='plus', exon_count=13, OMIM=609927, no_hdr0=''),
  6421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6421, CurrentID=0, Status='live', Symbol='SFPQ', Aliases='POMP100, PPP1R140, PSF', description='splicing factor proline/glutamine-rich', other_designations='100 kDa DNA-pairing protein|DNA-binding p52/p100 complex, 100 kDa subunit|PTB-associated splicing factor|PTB-associated-splicing factor|hPOMp100|polypyrimidine tract binding protein associated|polypyrimidine tract-binding protein-associated splicing factor|polypyrimidine tract-binding protein-associated-splicing factor|protein phosphatase 1, regulatory subunit 140|splicing factor proline/glutamine rich (polypyrimidine tract binding protein associated)|splicing factor proline/glutamine rich (polypyrimidine tract-binding protein-associated)', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35176380, end_position_on_the_genomic_accession=35193158, orientation='minus', exon_count=13, OMIM=605199, no_hdr0=''),
  6422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6422, CurrentID=0, Status='live', Symbol='SFRP1', Aliases='FRP, FRP-1, FRP1, FrzA, SARP2', description='secreted frizzled-related protein 1', other_designations='SARP-2|sFRP-1|secreted apoptosis-related protein 2', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41261957, end_position_on_the_genomic_accession=41309471, orientation='minus', exon_count=3, OMIM=604156, no_hdr0=''),
  6423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6423, CurrentID=0, Status='live', Symbol='SFRP2', Aliases='FRP-2, SARP1, SDF-5', description='secreted frizzled-related protein 2', other_designations='SARP-1|sFRP-2|secreted apoptosis related protein 1|secreted apoptosis-related protein 1', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153780590, end_position_on_the_genomic_accession=153789076, orientation='minus', exon_count=3, OMIM=604157, no_hdr0=''),
  6424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6424, CurrentID=0, Status='live', Symbol='SFRP4', Aliases='FRP-4, FRPHE, sFRP-4', description='secreted frizzled-related protein 4', other_designations='frizzled protein, human endometrium|secreted frizzled-related protein 4; secreted frizzled-related protein 4', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37905932, end_position_on_the_genomic_accession=37916923, orientation='minus', exon_count=6, OMIM=606570, no_hdr0=''),
  6425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6425, CurrentID=0, Status='live', Symbol='SFRP5', Aliases='SARP3', description='secreted frizzled-related protein 5', other_designations='FRP-1b|SARP-3|frizzled-related protein 1b|sFRP-5|secreted apoptosis related protein 3|secreted apoptosis-related protein 3', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97766751, end_position_on_the_genomic_accession=97771999, orientation='minus', exon_count=3, OMIM=604158, no_hdr0=''),
  6426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6426, CurrentID=0, Status='live', Symbol='SRSF1', Aliases='ASF, SF2, SF2p33, SFRS1, SRp30a', description='serine/arginine-rich splicing factor 1', other_designations='SR splicing factor 1|alternative-splicing factor 1|pre-mRNA-splicing factor SF2, P33 subunit|splicing factor 2|splicing factor, arginine/serine-rich, 30-KD, A', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58000919, end_position_on_the_genomic_accession=58007346, orientation='minus', exon_count=4, OMIM=600812, no_hdr0=''),
  6427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6427, CurrentID=0, Status='live', Symbol='SRSF2', Aliases='PR264, SC-35, SC35, SFRS2, SFRS2A, SRp30b', description='serine/arginine-rich splicing factor 2', other_designations='SR splicing factor 2|splicing component, 35 kDa|splicing factor SC35|splicing factor, arginine/serine-rich 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76734115, end_position_on_the_genomic_accession=76737629, orientation='minus', exon_count=5, OMIM=600813, no_hdr0=''),
  6428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6428, CurrentID=0, Status='live', Symbol='SRSF3', Aliases='SFRS3, SRp20', description='serine/arginine-rich splicing factor 3', other_designations='pre-mRNA splicing factor SRp20|pre-mRNA-splicing factor SRP20|splicing factor, arginine/serine-rich 3|splicing factor, arginine/serine-rich, 20-kD', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36594313, end_position_on_the_genomic_accession=36604467, orientation='plus', exon_count=7, OMIM=603364, no_hdr0=''),
  6429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6429, CurrentID=0, Status='live', Symbol='SRSF4', Aliases='SFRS4, SRP75', description='serine/arginine-rich splicing factor 4', other_designations='SR splicing factor 4|SRP001LB|pre-mRNA-splicing factor SRP75|splicing factor, arginine/serine-rich 4', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=29147738, end_position_on_the_genomic_accession=29182125, orientation='minus', exon_count=6, OMIM=601940, no_hdr0=''),
  6430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6430, CurrentID=0, Status='live', Symbol='SRSF5', Aliases='HRS, SFRS5, SRP40', description='serine/arginine-rich splicing factor 5', other_designations='SR splicing factor 5|delayed-early protein HRS|pre-mRNA-splicing factor SRP40|splicing factor, arginine/serine-rich 5', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69767095, end_position_on_the_genomic_accession=69772005, orientation='plus', exon_count=8, OMIM=600914, no_hdr0=''),
  6431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6431, CurrentID=0, Status='live', Symbol='SRSF6', Aliases='B52, HEL-S-91, SFRS6, SRP55', description='serine/arginine-rich splicing factor 6', other_designations='SR splicing factor 6|epididymis secretory protein Li 91|pre-mRNA-splicing factor SRP55|splicing factor, arginine/serine-rich, 55 kDa', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43457864, end_position_on_the_genomic_accession=43463605, orientation='plus', exon_count=7, OMIM=601944, no_hdr0=''),
  6432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6432, CurrentID=0, Status='live', Symbol='SRSF7', Aliases='9G8, AAG3, SFRS7', description='serine/arginine-rich splicing factor 7', other_designations='SR splicing factor 7|aging-associated protein 3|splicing factor 9G8|splicing factor, arginine/serine-rich 7, 35kDa', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38743599, end_position_on_the_genomic_accession=38751494, orientation='minus', exon_count=9, OMIM=600572, no_hdr0=''),
  6433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6433, CurrentID=0, Status='live', Symbol='SFSWAP', Aliases='SFRS8, SWAP', description='splicing factor, suppressor of white-apricot family', other_designations='splicing factor, arginine/serine-rich 8 (suppressor-of-white-apricot homolog, Drosophila)|splicing factor, arginine/serine-rich 8 (suppressor-of-white-apricot, Drosophila homolog)|suppressor of white apricot protein homolog', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131711087, end_position_on_the_genomic_accession=131799738, orientation='plus', exon_count=20, OMIM=601945, no_hdr0=''),
  6434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6434, CurrentID=0, Status='live', Symbol='TRA2B', Aliases='Htra2-beta, PPP1R156, SFRS10, SRFS10, TRA2-BETA, TRAN2B', description='transformer 2 beta homolog (Drosophila)', other_designations='TRA-2 beta|protein phosphatase 1, regulatory subunit 156|splicing factor, arginine/serine-rich 10 (transformer 2 homolog, Drosophila)|transformer-2 protein homolog B', map_location='3q26.2-q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185914568, end_position_on_the_genomic_accession=185938136, orientation='minus', exon_count=12, OMIM=602719, no_hdr0=''),
  6439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6439, CurrentID=0, Status='live', Symbol='SFTPB', Aliases='PSP-B, SFTB3, SFTP3, SMDP1, SP-B', description='surfactant protein B', other_designations='18 kDa pulmonary-surfactant protein|6 kDa protein|pulmonary surfactant-associated proteolipid SPL(Phe)', map_location='2p12-p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85657317, end_position_on_the_genomic_accession=85668741, orientation='minus', exon_count=14, OMIM=178640, no_hdr0=''),
  6440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6440, CurrentID=0, Status='live', Symbol='SFTPC', Aliases='BRICD6, PSP-C, SFTP2, SMDP2, SP-C', description='surfactant protein C', other_designations='BRICHOS domain containing 6|SP5|pulmonary surfactant apoprotein-2 SP-C|pulmonary surfactant-associated proteolipid SPL(Val)', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22161671, end_position_on_the_genomic_accession=22164479, orientation='plus', exon_count=6, OMIM=178620, no_hdr0=''),
  6441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6441, CurrentID=0, Status='live', Symbol='SFTPD', Aliases='COLEC7, PSP-D, SFTP4, SP-D', description='surfactant protein D', other_designations='collectin-7|lung surfactant protein D|pulmonary surfactant apoprotein|surfactant-associated protein, pulmonary 4', map_location='10q22.2-q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79937740, end_position_on_the_genomic_accession=79982236, orientation='minus', exon_count=9, OMIM=178635, no_hdr0=''),
  6442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6442, CurrentID=0, Status='live', Symbol='SGCA', Aliases='50-DAG, A2, ADL, DAG2, DMDA2, LGMD2D, SCARMD1, adhalin', description='sarcoglycan alpha', other_designations='50 kDa dystrophin-associated glycoprotein|50DAG|50kD DAG|alpha-SG|dystroglycan-2|sarcoglycan, alpha (50kDa dystrophin-associated glycoprotein)', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50165975, end_position_on_the_genomic_accession=50175932, orientation='plus', exon_count=10, OMIM=600119, no_hdr0=''),
  6443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6443, CurrentID=0, Status='live', Symbol='SGCB', Aliases='A3b, LGMD2E, SGC', description='sarcoglycan beta', other_designations='43 kDa dystrophin-associated glycoprotein|43DAG|beta-SG|beta-sarcoglycan(43kD dystrophin-associated glycoprotein)|limb girdle muscular dystrophy 2E (non-linked families)|sarcoglycan, beta (43kDa dystrophin-associated glycoprotein)', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52020695, end_position_on_the_genomic_accession=52038319, orientation='minus', exon_count=7, OMIM=600900, no_hdr0=''),
  6444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6444, CurrentID=0, Status='live', Symbol='SGCD', Aliases='35DAG, CMD1L, DAGD, SG-deltaP, SGD, SGCD', description='sarcoglycan delta', other_designations='35 kDa dystrophin-associated glycoprotein|delta-SG|dystrophin associated glycoprotein, delta sarcoglycan|placental delta sarcoglycan|sarcoglycan, delta (35kDa dystrophin-associated glycoprotein)', map_location='5q33-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156117934, end_position_on_the_genomic_accession=156767788, orientation='plus', exon_count=12, OMIM=601411, no_hdr0=''),
  6445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6445, CurrentID=0, Status='live', Symbol='SGCG', Aliases='35DAG, A4, DAGA4, DMDA, DMDA1, LGMD2C, MAM, SCARMD2, SCG3, gamma-SG', description='sarcoglycan gamma', other_designations='35 kDa dystrophin-associated glycoprotein|35kD dystrophin-associated glycoprotein|sarcoglycan, gamma (35kDa dystrophin-associated glycoprotein)', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23160506, end_position_on_the_genomic_accession=23325165, orientation='plus', exon_count=10, OMIM=608896, no_hdr0=''),
  6446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6446, CurrentID=0, Status='live', Symbol='SGK1', Aliases='SGK', description='serum/glucocorticoid regulated kinase 1', other_designations='Sgk1 variant i3|serine/threonine protein kinase SGK|serum/glucocorticoid-regulated kinase 1', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134169246, end_position_on_the_genomic_accession=134318112, orientation='minus', exon_count=18, OMIM=602958, no_hdr0=''),
  6447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6447, CurrentID=0, Status='live', Symbol='SCG5', Aliases='7B2, P7B2, SGNE1, SgV', description='secretogranin V', other_designations='pituitary polypeptide|prohormone convertase chaperone|secretogranin V (7B2 protein)|secretogranin-5|secretory granule endocrine protein I|secretory granule, neuroendocrine protein 1 (7B2 protein)', map_location='15q13-q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32641669, end_position_on_the_genomic_accession=32697097, orientation='plus', exon_count=6, OMIM=173120, no_hdr0=''),
  6448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6448, CurrentID=0, Status='live', Symbol='SGSH', Aliases='HSS, MPS3A, SFMD', description='N-sulfoglucosamine sulfohydrolase', other_designations='heparan sulfate sulfatase|mucopolysaccharidosis type IIIA|sulfoglucosamine sulfamidase|sulphamidase', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80206966, end_position_on_the_genomic_accession=80220400, orientation='minus', exon_count=12, OMIM=605270, no_hdr0=''),
  6449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6449, CurrentID=0, Status='live', Symbol='SGTA', Aliases='SGT, alphaSGT, hSGT', description='small glutamine-rich tetratricopeptide repeat (TPR)-containing, alpha', other_designations='UBP|alpha-SGT|protein containing three tetratricopeptide repeats|vpu-binding protein', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2754713, end_position_on_the_genomic_accession=2783356, orientation='minus', exon_count=13, OMIM=603419, no_hdr0=''),
  6450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6450, CurrentID=0, Status='live', Symbol='SH3BGR', Aliases='21-GARP', description='SH3 domain binding glutamate-rich protein', other_designations='21-glutamic acid-rich protein|SH3 domain binding glutamic acid-rich protein|SH3-binding domain and glutamic acid-rich protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39445856, end_position_on_the_genomic_accession=39515506, orientation='plus', exon_count=10, OMIM=602230, no_hdr0=''),
  6451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6451, CurrentID=0, Status='live', Symbol='SH3BGRL', Aliases='HEL-S-115, SH3BGR', description='SH3 domain binding glutamate-rich protein like', other_designations='SH3 domain binding glutamic acid-rich protein like|SH3-binding domain glutamic acid-rich protein like|epididymis secretory protein Li 115', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81201804, end_position_on_the_genomic_accession=81298547, orientation='plus', exon_count=7, OMIM=300190, no_hdr0=''),
  6452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6452, CurrentID=0, Status='live', Symbol='SH3BP2', Aliases='3BP-2, 3BP2, CRBM, CRPM, RES4-23', description='SH3-domain binding protein 2', other_designations='Abl-SH3 binding protein 2|TNFAIP3 interacting protein 2', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2793023, end_position_on_the_genomic_accession=2841096, orientation='plus', exon_count=22, OMIM=602104, no_hdr0=''),
  6453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6453, CurrentID=0, Status='live', Symbol='ITSN1', Aliases='ITSN, SH3D1A, SH3P17', description='intersectin 1', other_designations='SH3 domain-containing protein 1A|Src homology 3 domain-containing protein|human intersectin-SH3 domain-containing protein SH3P17|intersectin 1 (SH3 domain protein)|intersectin 1 short form variant 3|intersectin 1 short form variant, 11|intersectin short variant 12', map_location='21q22.1-q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33642400, end_position_on_the_genomic_accession=33889305, orientation='plus', exon_count=44, OMIM=602442, no_hdr0=''),
  399670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399670, CurrentID=0, Status='live', Symbol='RPL13AP17', Aliases='RPL13A_6_816', description='ribosomal protein L13a pseudogene 17', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=78347242, end_position_on_the_genomic_accession=78359458, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  6455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6455, CurrentID=0, Status='live', Symbol='SH3GL1', Aliases='CNSA1, EEN, SH3D2B, SH3P8', description='SH3-domain GRB2-like 1', other_designations='EEN fusion partner of MLL|SH3 domain protein 2B|SH3 domain-containing GRB2-like protein 1|SH3-containing Grb-2-like 1 protein|endophilin-2|extra 11-19 leukemia fusion|extra eleven-nineteen leukemia fusion gene protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4360367, end_position_on_the_genomic_accession=4400568, orientation='minus', exon_count=11, OMIM=601768, no_hdr0=''),
  6456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6456, CurrentID=0, Status='live', Symbol='SH3GL2', Aliases='CNSA2, EEN-B1, SH3D2A, SH3P4', description='SH3-domain GRB2-like 2', other_designations='Endophilin A1 BAR domain|SH3 domain protein 2A|SH3 domain-containing GRB2-like protein 2|bA335L15.1 (SH3-domain GRB2-like 2)|endophilin-1', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17578954, end_position_on_the_genomic_accession=17797129, orientation='plus', exon_count=11, OMIM=604465, no_hdr0=''),
  137529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137529, CurrentID=0, Status='live', Symbol='CKS1BP7', Aliases='CKS1, CKS1A', description='CDC28 protein kinase regulatory subunit 1B pseudogene 7', other_designations='CDC28 protein kinase 1|CDC28 protein kinase 1A|CDC28 protein kinase regulatory subunit 1A pseudogene|cell division control protein CKS1 pseudogene', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80644452, end_position_on_the_genomic_accession=80645216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6458, CurrentID=0, Status='live', Symbol='SH3GL1P1', Aliases='CNSA-P1, SH3GLP1', description='SH3-domain GRB2-like 1 pseudogene 1', other_designations='SH3-domain GRB2-like pseudogene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32040336, end_position_on_the_genomic_accession=32042832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  6459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6459, CurrentID=0, Status='live', Symbol='SH3GL1P2', Aliases='CNSA-P2, SH3GLP2', description='SH3-domain GRB2-like 1 pseudogene 2', other_designations='SH3-domain GRB2-like pseudogene 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30624318, end_position_on_the_genomic_accession=30626823, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  6461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6461, CurrentID=0, Status='live', Symbol='SHB', Aliases='bA3J10.2', description='Src homology 2 domain containing adaptor protein B', other_designations='SHB (Src homology 2 domain containing) adaptor protein B|SHB adaptor protein (a Src homology 2 protein)', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37915898, end_position_on_the_genomic_accession=38069213, orientation='minus', exon_count=6, OMIM=600314, no_hdr0=''),
  6462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6462, CurrentID=0, Status='live', Symbol='SHBG', Aliases='ABP, SBP, TEBG', description='sex hormone-binding globulin', other_designations='sex steroid-binding protein|testis-specific androgen-binding protein|testosterone-binding beta-globulin|testosterone-estradiol-binding globulin|testosterone-estrogen-binding globulin', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7614064, end_position_on_the_genomic_accession=7633383, orientation='plus', exon_count=10, OMIM=182205, no_hdr0=''),
  6463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6463, CurrentID=0, Status='live', Symbol='SHFM2', Aliases='SHFD2, SHSF2', description='split hand/foot malformation (ectrodactyly) type 2', other_designations='Split hand/foot malformation, type (ectrodactyly) 2', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=313350, no_hdr0=''),
  6464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6464, CurrentID=0, Status='live', Symbol='SHC1', Aliases='SHC, SHCA', description='SHC (Src homology 2 domain containing) transforming protein 1', other_designations='SH2 domain protein C1|SHC (Src homology 2 domain-containing) transforming protein 1|SHC-transforming protein 3|SHC-transforming protein A', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154962298, end_position_on_the_genomic_accession=154974492, orientation='minus', exon_count=13, OMIM=600560, no_hdr0=''),
  6465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6465, CurrentID=0, Status='live', Symbol='SHC1P1', Aliases='', description='SHC (Src homology 2 domain containing) transforming protein 1 pseudogene 1', other_designations='SHC (Src homology 2 domain-containing) transforming protein 1 pseudogene 1', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64432183, end_position_on_the_genomic_accession=64435674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6466, CurrentID=0, Status='live', Symbol='SHC1P2', Aliases='SHCL1', description='SHC (Src homology 2 domain containing) transforming protein 1 pseudogene 2', other_designations='SHC (Src homology 2 domain containing) transforming protein-like 1 pseudogene|SHC (Src homology 2 domain-containing) transforming protein-like 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44226961, end_position_on_the_genomic_accession=44230550, orientation='minus', exon_count=0, OMIM=600739, no_hdr0=''),
  6468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6468, CurrentID=0, Status='live', Symbol='FBXW4', Aliases='DAC, FBW4, FBWD4, SHFM3, SHSF3', description='F-box and WD repeat domain containing 4', other_designations='F-box and WD-40 domain protein 4|F-box and WD-40 domain-containing protein 4|F-box/WD repeat protein 4|dactylin', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101610664, end_position_on_the_genomic_accession=101695037, orientation='minus', exon_count=10, OMIM=608071, no_hdr0=''),
  6469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6469, CurrentID=0, Status='live', Symbol='SHH', Aliases='HHG1, HLP3, HPE3, MCOPCB5, SMMCI, TPT, TPTPS', description='sonic hedgehog', other_designations='sonic hedgehog homolog', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155799980, end_position_on_the_genomic_accession=155801870, orientation='minus', exon_count=8, OMIM=600725, no_hdr0=''),
  6470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6470, CurrentID=0, Status='live', Symbol='SHMT1', Aliases='CSHMT, SHMT', description='serine hydroxymethyltransferase 1 (soluble)', other_designations='cytoplasmic serine hydroxymethyltransferase|glycine hydroxymethyltransferase|serine methylase', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18327860, end_position_on_the_genomic_accession=18363563, orientation='minus', exon_count=13, OMIM=182144, no_hdr0=''),
  399687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399687, CurrentID=0, Status='live', Symbol='MYO18A', Aliases='MYSPDZ, SPR210', description='myosin XVIIIA', other_designations='MAJN|SP-A receptor subunit SP-R210 alphaS|molecule associated with JAK3 N-terminus|myosin 18A|myosin containing PDZ domain|myosin containing a PDZ domain|myosin-XVIIIa', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29073510, end_position_on_the_genomic_accession=29180389, orientation='minus', exon_count=42, OMIM=610067, no_hdr0=''),
  6472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6472, CurrentID=0, Status='live', Symbol='SHMT2', Aliases='GLYA, HEL-S-51e, SHMT', description='serine hydroxymethyltransferase 2 (mitochondrial)', other_designations='GLY A+|epididymis secretory sperm binding protein Li 51e|glycine auxotroph A, human complement for hamster|glycine hydroxymethyltransferase|serine aldolase|serine hydroxymethylase|serine methylase|threonine aldolase', map_location='12q12-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57229573, end_position_on_the_genomic_accession=57234935, orientation='plus', exon_count=14, OMIM=138450, no_hdr0=''),
  6473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6473, CurrentID=0, Status='live', Symbol='SHOX', Aliases='GCFX, PHOGY, SS, SHOX', description='short stature homeobox', other_designations='growth control factor, X-linked|pseudoautosomal homeobox-containing osteogenic protein', map_location='Xp22.33;Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=624344, end_position_on_the_genomic_accession=659411, orientation='plus', exon_count=7, OMIM='312865,400020', no_hdr0=''),
  6474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6474, CurrentID=0, Status='live', Symbol='SHOX2', Aliases='OG12, OG12X, SHOT', description='short stature homeobox 2', other_designations='SHOX homologous gene on chromosome 3|homeobox protein Og12X|paired-related homeobox protein SHOT', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158096011, end_position_on_the_genomic_accession=158106202, orientation='minus', exon_count=7, OMIM=602504, no_hdr0=''),
  6476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6476, CurrentID=0, Status='live', Symbol='SI', Aliases='', description='sucrase-isomaltase (alpha-glucosidase)', other_designations='oligosaccharide alpha-1,6-glucosidase', map_location='3q25.2-q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164978893, end_position_on_the_genomic_accession=165078495, orientation='minus', exon_count=48, OMIM=609845, no_hdr0=''),
  6477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6477, CurrentID=0, Status='live', Symbol='SIAH1', Aliases='SIAH1A', description='siah E3 ubiquitin protein ligase 1', other_designations='seven in absentia homolog 1|siah-1a', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48360531, end_position_on_the_genomic_accession=48448440, orientation='minus', exon_count=8, OMIM=602212, no_hdr0=''),
  6478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6478, CurrentID=0, Status='live', Symbol='SIAH2', Aliases='hSiah2', description='siah E3 ubiquitin protein ligase 2', other_designations='seven in absentia homolog 2|siah-2', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150741123, end_position_on_the_genomic_accession=150763476, orientation='minus', exon_count=2, OMIM=602213, no_hdr0=''),
  6480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6480, CurrentID=0, Status='live', Symbol='ST6GAL1', Aliases='SIAT1, ST6GalI, ST6N', description='ST6 beta-galactosamide alpha-2,6-sialyltranferase 1', other_designations='B-cell antigen CD75|CMP-N-acetylneuraminate beta-galactosamide alpha-2,6-sialyltransferase|CMP-N-acetylneuraminate-beta-galactosamide-alpha-2,6-sialyltransferase 1|ST6Gal I|alpha 2,6-ST 1|sialyltransferase 1 (beta-galactoside alpha-2,6-sialyltransferase)', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186930343, end_position_on_the_genomic_accession=187078553, orientation='plus', exon_count=16, OMIM=109675, no_hdr0=''),
  399697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399697, CurrentID=0, Status='live', Symbol='CTXN2', Aliases='', description='cortexin 2', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48178355, end_position_on_the_genomic_accession=48203754, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  6482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6482, CurrentID=0, Status='live', Symbol='ST3GAL1', Aliases='Gal-NAc6S, SIAT4A, SIATFL, ST3GalA, ST3GalA.1, ST3GalIA, ST3GalIA,1, ST3O', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 1', other_designations='Gal-beta-1,3-GalNAc-alpha-2,3-sialyltransferase|SIAT4-A|ST3GalI|alpha 2,3-ST 1|beta-galactoside alpha-2,3-sialyltransferase 1|sialyltransferase 4A (beta-galactosidase alpha-2,3-sialytransferase)|sialyltransferase 4A (beta-galactoside alpha-2,3-sialyltransferase)|sialyltransferase 4A (beta-galactoside alpha-2,3-sialytransferase)', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133454848, end_position_on_the_genomic_accession=133572454, orientation='minus', exon_count=13, OMIM=607187, no_hdr0=''),
  6483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6483, CurrentID=0, Status='live', Symbol='ST3GAL2', Aliases='Gal-NAc6S, SIAT4B, ST3GALII, ST3GalA.2', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 2', other_designations='Gal-beta-1,3-GalNAc-alpha-2,3-sialyltransferase|SIAT4-B|ST3Gal II|alpha 2,3-ST 2|beta-galactoside alpha-2,3-sialyltransferase 2|beta-galactoside alpha-2,3-sialytransferase|sialyltransferase 4B (beta-galactosidase alpha-2,3-sialytransferase)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70379435, end_position_on_the_genomic_accession=70439088, orientation='minus', exon_count=7, OMIM=607188, no_hdr0=''),
  6484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6484, CurrentID=0, Status='live', Symbol='ST3GAL4', Aliases='CGS23, NANTA3, SAT3, SIAT4, SIAT4C, ST3GalIV, STZ', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 4', other_designations='SAT-3|SIAT4-C|ST-4|ST3GalA.2|alpha 2,3-ST 4|alpha 2,3-sialyltransferase IV|alpha-3-N-acetylneuraminyltransferase|beta-galactoside alpha-2,3-sialyltransferase 4|gal-NAc6S|gal-beta-1,4-GalNAc-alpha-2,3-sialyltransferase|sialyltransferase 4C (beta-galactosidase alpha-2,3-sialytransferase)|sialyltransferase 4C (beta-galactoside alpha-2,3-sialytransferase)', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126355645, end_position_on_the_genomic_accession=126652917, orientation='plus', exon_count=19, OMIM=104240, no_hdr0=''),
  6487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6487, CurrentID=0, Status='live', Symbol='ST3GAL3', Aliases='EIEE15, MRT12, SIAT6, ST3GALII, ST3GalIII, ST3N', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 3', other_designations='Gal beta-1,3(4)GlcNAc alpha-2,3 sialyltransferase|ST3Gal III|alpha 2,3-ST 3|alpha 2,3-sialyltransferase III|alpha-2,3-sialyltransferase II|mental retardation, non-syndromic, autosomal recessive, 12|sialyltransferase 6 (N-acetyllacosaminide alpha 2,3-sialyltransferase)', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43705824, end_position_on_the_genomic_accession=43931165, orientation='plus', exon_count=27, OMIM=606494, no_hdr0=''),
  6489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6489, CurrentID=0, Status='live', Symbol='ST8SIA1', Aliases='GD3S, SIAT8, SIAT8-A, SIAT8A, ST8SiaI', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 1', other_designations='alpha-2,8-sialyltransferase 8A|disialoganglioside (GD3) synthase|ganglioside GD3 synthase|ganglioside GT3 synthase|ganglioside-specific alpha-2,8-polysialyltransferase|sialyltransferase 8 (alpha-N-acetylneuraminate: alpha-2,8-sialytransferase, GD3 synthase) A|sialyltransferase 8A (alpha-N-acetylneuraminate: alpha-2,8-sialyltransferase, GD3 synthase)|sialyltransferase St8Sia I|sialytransferase St8Sia I', map_location='12p12.1-p11.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22129135, end_position_on_the_genomic_accession=22334714, orientation='minus', exon_count=9, OMIM=601123, no_hdr0=''),
  6490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6490, CurrentID=0, Status='live', Symbol='PMEL', Aliases='D12S53E, ME20, ME20-M, ME20M, P1, P10017, SI, SIL, SILV, gp100, PMEL', description='premelanosome protein', other_designations='melanocyte protein Pmel 17|melanocyte protein mel 17|melanocytes lineage-specific antigen GP100|melanoma-associated ME20 antigen|melanosomal matrix protein17|silver locus protein homolog|silver, mouse, homolog of', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55954105, end_position_on_the_genomic_accession=55966712, orientation='minus', exon_count=13, OMIM=155550, no_hdr0=''),
  6491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6491, CurrentID=0, Status='live', Symbol='STIL', Aliases='MCPH7, SIL', description='SCL/TAL1 interrupting locus', other_designations='TAL-1-interrupting locus protein', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47250139, end_position_on_the_genomic_accession=47314781, orientation='minus', exon_count=20, OMIM=181590, no_hdr0=''),
  6492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6492, CurrentID=0, Status='live', Symbol='SIM1', Aliases='bHLHe14', description='single-minded family bHLH transcription factor 1', other_designations='class E basic helix-loop-helix protein 14', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100385009, end_position_on_the_genomic_accession=100464929, orientation='minus', exon_count=12, OMIM=603128, no_hdr0=''),
  6493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6493, CurrentID=0, Status='live', Symbol='SIM2', Aliases='HMC13F06, HMC29C01, SIM, bHLHe15', description='single-minded family bHLH transcription factor 2', other_designations='class E basic helix-loop-helix protein 15|transcription factor SIM2', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36699133, end_position_on_the_genomic_accession=36750219, orientation='plus', exon_count=13, OMIM=600892, no_hdr0=''),
  6494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6494, CurrentID=0, Status='live', Symbol='SIPA1', Aliases='SPA1', description='signal-induced proliferation-associated 1', other_designations='GTPase-activating protein Spa-1|p130 SPA-1|signal-induced proliferation-associated gene 1|sipa-1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65638107, end_position_on_the_genomic_accession=65650930, orientation='plus', exon_count=20, OMIM=602180, no_hdr0=''),
  6495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6495, CurrentID=0, Status='live', Symbol='SIX1', Aliases='BOS3, DFNA23, TIP39', description='SIX homeobox 1', other_designations='sine oculis homeobox homolog 1', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60644699, end_position_on_the_genomic_accession=60649437, orientation='minus', exon_count=2, OMIM=601205, no_hdr0=''),
  6496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6496, CurrentID=0, Status='live', Symbol='SIX3', Aliases='HPE2', description='SIX homeobox 3', other_designations='sine oculis homeobox homolog 3', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44941729, end_position_on_the_genomic_accession=44966751, orientation='plus', exon_count=3, OMIM=603714, no_hdr0=''),
  6497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6497, CurrentID=0, Status='live', Symbol='SKI', Aliases='SGS, SKV', description='SKI proto-oncogene', other_designations='proto-oncogene c-Ski|ski oncoprotein|v-ski avian sarcoma viral oncogene homolog', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2228695, end_position_on_the_genomic_accession=2310213, orientation='plus', exon_count=8, OMIM=164780, no_hdr0=''),
  6498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6498, CurrentID=0, Status='live', Symbol='SKIL', Aliases='SNO, SnoA, SnoI, SnoN', description='SKI-like proto-oncogene', other_designations='SKI-like oncogene|ski-related oncogene snoN', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170357678, end_position_on_the_genomic_accession=170396849, orientation='plus', exon_count=8, OMIM=165340, no_hdr0=''),
  6499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6499, CurrentID=0, Status='live', Symbol='SKIV2L', Aliases='170A, DDX13, HLP, SKI2, SKI2W, SKIV2, THES2', description='SKI2 homolog, superkiller viralicidic activity 2-like', other_designations='helicase-like protein|superkiller viralicidic activity 2-like', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31958804, end_position_on_the_genomic_accession=31969755, orientation='plus', exon_count=28, OMIM=600478, no_hdr0=''),
  6500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6500, CurrentID=0, Status='live', Symbol='SKP1', Aliases='EMC19, OCP-II, OCP2A, TCEB1L, p19A, SKP1', description='S-phase kinase-associated protein 1', other_designations='OCP-2|RNA polymerase II elongation factor-like protein OCP2|SIII|cyclin A/CDK2-associated p19|cyclin A/CDK2-associated protein p19|cyclin-A/CDK2-associated protein p19|organ of Corti protein 2|organ of Corti protein II|p19skp1|transcription elongation factor B polypeptide 1-like', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134156391, end_position_on_the_genomic_accession=134177033, orientation='minus', exon_count=5, OMIM=601434, no_hdr0=''),
  399717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399717, CurrentID=0, Status='live', Symbol='GATA3-AS1', Aliases='', description='GATA3 antisense RNA 1', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8050450, end_position_on_the_genomic_accession=8053484, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6502, CurrentID=0, Status='live', Symbol='SKP2', Aliases='FBL1, FBXL1, FLB1, p45', description='S-phase kinase-associated protein 2, E3 ubiquitin protein ligase', other_designations='CDK2/cyclin A-associated protein p45|F-box/LRR-repeat protein 1|S-phase kinase-associated protein 2 (p45)|p45skp2', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36151989, end_position_on_the_genomic_accession=36192788, orientation='plus', exon_count=12, OMIM=601436, no_hdr0=''),
  6503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6503, CurrentID=0, Status='live', Symbol='SLA', Aliases='SLA1P, SLA', description='Src-like-adaptor', other_designations='SLAP-1|hSLAP|src-like-adapter protein 1', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133036728, end_position_on_the_genomic_accession=133103066, orientation='minus', exon_count=10, OMIM=601099, no_hdr0=''),
  6504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6504, CurrentID=0, Status='live', Symbol='SLAMF1', Aliases='CD150, CDw150, SLAM', description='signaling lymphocytic activation molecule family member 1', other_designations='IPO-3', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160608111, end_position_on_the_genomic_accession=160647537, orientation='minus', exon_count=8, OMIM=603492, no_hdr0=''),
  6505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6505, CurrentID=0, Status='live', Symbol='SLC1A1', Aliases='DCBXA, EAAC1, EAAT3, SCZD18', description='solute carrier family 1 (neuronal/epithelial high affinity glutamate transporter, system Xag), member 1', other_designations='excitatory amino acid carrier 1|neuronal and epithelial glutamate transporter|sodium-dependent glutamate/aspartate transporter 3|solute carrier family 1 member 1', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4490427, end_position_on_the_genomic_accession=4587469, orientation='plus', exon_count=14, OMIM=133550, no_hdr0=''),
  6506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6506, CurrentID=0, Status='live', Symbol='SLC1A2', Aliases='EAAT2, GLT-1, HBGT', description='solute carrier family 1 (glial high affinity glutamate transporter), member 2', other_designations='excitotoxic amino acid transporter 2|glutamate/aspartate transporter II|sodium-dependent glutamate/aspartate transporter 2', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35251205, end_position_on_the_genomic_accession=35420063, orientation='minus', exon_count=16, OMIM=600300, no_hdr0=''),
  6507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6507, CurrentID=0, Status='live', Symbol='SLC1A3', Aliases='EA6, EAAT1, GLAST, GLAST1', description='solute carrier family 1 (glial high affinity glutamate transporter), member 3', other_designations='GLAST-1|sodium-dependent glutamate/aspartate transporter 1|solute carrier family 1 member 3', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36606355, end_position_on_the_genomic_accession=36688334, orientation='plus', exon_count=12, OMIM=600111, no_hdr0=''),
  6508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6508, CurrentID=0, Status='live', Symbol='SLC4A3', Aliases='AE3, SLC2C', description='solute carrier family 4 (anion exchanger), member 3', other_designations='AE 3|Anion exchanger 3, neuronal|CAE3/BAE3|anion exchanger 3 cardiac isoform|cardiac/brain band 3-like protein|neuronal band 3-like protein|solute carrier family 4 member 3', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219627570, end_position_on_the_genomic_accession=219641980, orientation='plus', exon_count=24, OMIM=106195, no_hdr0=''),
  6509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6509, CurrentID=0, Status='live', Symbol='SLC1A4', Aliases='ASCT1, SATT, SPATCCM', description='solute carrier family 1 (glutamate/neutral amino acid transporter), member 4', other_designations='ASCT-1|alanine/serine/cysteine/threonine transporter 1|glutamate/neutral amino acid transporter|solute carrier family 1 member 4', map_location='2p15-p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64988445, end_position_on_the_genomic_accession=65023865, orientation='plus', exon_count=9, OMIM=600229, no_hdr0=''),
  6510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6510, CurrentID=0, Status='live', Symbol='SLC1A5', Aliases='AAAT, ASCT2, ATBO, M7V1, M7VS1, R16, RDRC', description='solute carrier family 1 (neutral amino acid transporter), member 5', other_designations='ATB(0)|RD114 virus receptor|RD114/simian type D retrovirus receptor|baboon M7 virus receptor|neutral amino acid transporter B|sodium-dependent neutral amino acid transporter type 2|solute carrier family 1 member 5', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46759698, end_position_on_the_genomic_accession=46788649, orientation='minus', exon_count=11, OMIM=109190, no_hdr0=''),
  6511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6511, CurrentID=0, Status='live', Symbol='SLC1A6', Aliases='EAAT4', description='solute carrier family 1 (high affinity aspartate/glutamate transporter), member 6', other_designations='sodium-dependent glutamate/aspartate transporter|solute carrier family 1 member 6', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14950033, end_position_on_the_genomic_accession=15010643, orientation='minus', exon_count=15, OMIM=600637, no_hdr0=''),
  6512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6512, CurrentID=0, Status='live', Symbol='SLC1A7', Aliases='AAAT, EAAT5', description='solute carrier family 1 (glutamate transporter), member 7', other_designations='excitatory amino acid transporter 5 (retinal glutamate transporter)|retinal glutamate transporter', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53087179, end_position_on_the_genomic_accession=53142632, orientation='minus', exon_count=13, OMIM=604471, no_hdr0=''),
  6513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6513, CurrentID=0, Status='live', Symbol='SLC2A1', Aliases='CSE, DYT17, DYT18, DYT9, EIG12, GLUT, GLUT-1, GLUT1, GLUT1DS, HTLVR, PED, SDCHCN', description='solute carrier family 2 (facilitated glucose transporter), member 1', other_designations='choreoathetosis/spasticity, episodic (paroxysmal choreoathetosis/spasticity)|glucose transporter type 1, erythrocyte/brain|hepG2 glucose transporter|human T-cell leukemia virus (I and II) receptor|receptor for HTLV-1 and HTLV-2', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42925375, end_position_on_the_genomic_accession=42959176, orientation='minus', exon_count=10, OMIM='138140,143090', no_hdr0=''),
  6514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6514, CurrentID=0, Status='live', Symbol='SLC2A2', Aliases='GLUT2', description='solute carrier family 2 (facilitated glucose transporter), member 2', other_designations='glucose transporter type 2, liver', map_location='3q26.1-q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170995365, end_position_on_the_genomic_accession=171026979, orientation='minus', exon_count=13, OMIM=138160, no_hdr0=''),
  6515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6515, CurrentID=0, Status='live', Symbol='SLC2A3', Aliases='GLUT3', description='solute carrier family 2 (facilitated glucose transporter), member 3', other_designations='GLUT-3|glucose transporter type 3, brain', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7919228, end_position_on_the_genomic_accession=7936296, orientation='minus', exon_count=10, OMIM=138170, no_hdr0=''),
  6517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6517, CurrentID=0, Status='live', Symbol='SLC2A4', Aliases='GLUT4', description='solute carrier family 2 (facilitated glucose transporter), member 4', other_designations='GLUT-4|glucose transporter type 4, insulin-responsive|insulin-responsive glucose transporter type 4', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7281735, end_position_on_the_genomic_accession=7288048, orientation='plus', exon_count=11, OMIM=138190, no_hdr0=''),
  6518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6518, CurrentID=0, Status='live', Symbol='SLC2A5', Aliases='GLUT-5, GLUT5', description='solute carrier family 2 (facilitated glucose/fructose transporter), member 5', other_designations='glucose transporter type 5, small intestine|glucose transporter-like protein 5', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9036943, end_position_on_the_genomic_accession=9071704, orientation='minus', exon_count=14, OMIM=138230, no_hdr0=''),
  6519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6519, CurrentID=0, Status='live', Symbol='SLC3A1', Aliases='ATR1, CSNU1, D2H, NBAT, RBAT', description='solute carrier family 3 (amino acid transporter heavy chain), member 1', other_designations='B(0,+)-type amino acid transport protein|SLC3A1 variant B|SLC3A1 variant C|SLC3A1 variant D|SLC3A1 variant E|SLC3A1 variant F|SLC3A1 variant G|amino acid transporter 1|solute carrier family 3 (cystine, dibasic and neutral amino acid transporters), member 1|solute carrier family 3 (cystine, dibasic and neutral amino acid transporters, activator of cystine, dibasic and neutral amino acid transport), member 1|solute carrier family 3 member 1', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44275458, end_position_on_the_genomic_accession=44320824, orientation='plus', exon_count=11, OMIM=104614, no_hdr0=''),
  6520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6520, CurrentID=0, Status='live', Symbol='SLC3A2', Aliases='4F2, 4F2HC, 4T2HC, CD98, CD98HC, MDU1, NACAE', description='solute carrier family 3 (amino acid transporter heavy chain), member 2', other_designations='CD98 heavy chain|antigen defined by monoclonal antibody 4F2, heavy chain|antigen identified by monoclonal antibodies 4F2, TRA1.10, TROP4, and T43|lymphocyte activation antigen 4F2 large subunit|monoclonal antibody 44D7|solute carrier family 3 (activators of dibasic and neutral amino acid transport), member 2|solute carrier family 3 member 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62856012, end_position_on_the_genomic_accession=62888883, orientation='plus', exon_count=14, OMIM=158070, no_hdr0=''),
  6521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6521, CurrentID=0, Status='live', Symbol='SLC4A1', Aliases='AE1, BND3, CD233, CHC, DI, EMPB3, EPB3, FR, RTA1A, SAO, SW, WD, WD1, WR', description='solute carrier family 4 (anion exchanger), member 1 (Diego blood group)', other_designations='Diego blood group|Froese blood group|Swann blood group|Waldner blood group|Wright blood group|anion exchange protein 1|anion exchanger 1|anion exchanger-1|erythrocyte membrane protein band 3|erythroid anion exchange protein|solute carrier family 4, anion exchanger, member 1 (erythrocyte membrane protein band 3, Diego blood group)|solute carrier family 4, anion exchanger, number 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44248390, end_position_on_the_genomic_accession=44268161, orientation='minus', exon_count=22, OMIM=109270, no_hdr0=''),
  6522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6522, CurrentID=0, Status='live', Symbol='SLC4A2', Aliases='AE2, BND3L, EPB3L1, HKB3, NBND3', description='solute carrier family 4 (anion exchanger), member 2', other_designations='AE 2|anion exchanger 2 type a|anion exchanger 2 type b1|anion exchanger 2 type b2|erythrocyte membrane protein band 3-like 1|non-erythroid band 3-like protein|solute carrier family 4 member 2', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151058212, end_position_on_the_genomic_accession=151076527, orientation='plus', exon_count=26, OMIM=109280, no_hdr0=''),
  6523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6523, CurrentID=0, Status='live', Symbol='SLC5A1', Aliases='D22S675, NAGT, SGLT1', description='solute carrier family 5 (sodium/glucose cotransporter), member 1', other_designations='Na+/glucose cotransporter 1|high affinity sodium-glucose cotransporter', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32043032, end_position_on_the_genomic_accession=32113029, orientation='plus', exon_count=16, OMIM=182380, no_hdr0=''),
  6524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6524, CurrentID=0, Status='live', Symbol='SLC5A2', Aliases='SGLT2', description='solute carrier family 5 (sodium/glucose cotransporter), member 2', other_designations='Na(+)/glucose cotransporter 2|low affinity sodium-glucose cotransporter|solute carrier family 5 (sodium/glucose transporter), member 2|solute carrier family 5 member 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31482535, end_position_on_the_genomic_accession=31490770, orientation='plus', exon_count=13, OMIM=182381, no_hdr0=''),
  6525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6525, CurrentID=0, Status='live', Symbol='SMTN', Aliases='', description='smoothelin', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31064180, end_position_on_the_genomic_accession=31104624, orientation='plus', exon_count=27, OMIM=602127, no_hdr0=''),
  6526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6526, CurrentID=0, Status='live', Symbol='SLC5A3', Aliases='BCW2, SMIT, SMIT1, SMIT2', description='solute carrier family 5 (sodium/myo-inositol cotransporter), member 3', other_designations='Na(+)/myo-inositol cotransporter|sodium/myo-inositol cotransporter 1|sodium/myo-inositol transporter 1|solute carrier family 5 (inositol transporters), member 3', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34073523, end_position_on_the_genomic_accession=34106262, orientation='plus', exon_count=2, OMIM=600444, no_hdr0=''),
  6527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6527, CurrentID=0, Status='live', Symbol='SLC5A4', Aliases='DJ90G24.4, SAAT1, SGLT3', description='solute carrier family 5 (glucose activated ion channel), member 4', other_designations='Na(+)/glucose cotransporter 3|low affinity sodium glucose cotransporter|sodium transporter|sodium/glucose cotransporter 3|solute carrier family 5 (low affinity glucose cotransporter), member 4|solute carrier family 5 (neutral amino acid transporters, system A), member 4|solute carrier family 5 member 4', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32218421, end_position_on_the_genomic_accession=32353008, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  6528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6528, CurrentID=0, Status='live', Symbol='SLC5A5', Aliases='NIS, TDH1', description='solute carrier family 5 (sodium/iodide cotransporter), member 5', other_designations='Na(+)/I(-) cotransporter|Na(+)/I(-) symporter|solute carrier family 5 (sodium iodide symporter), member 5', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17871411, end_position_on_the_genomic_accession=17895174, orientation='plus', exon_count=16, OMIM=601843, no_hdr0=''),
  6529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6529, CurrentID=0, Status='live', Symbol='SLC6A1', Aliases='GABATHG, GABATR, GAT1, MAE', description='solute carrier family 6 (neurotransmitter transporter), member 1', other_designations='GABA transporter 1|GAT-1|solute carrier family 6 (neurotransmitter transporter, GABA), member 1|solute carrier family 6 member 1', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10992729, end_position_on_the_genomic_accession=11039249, orientation='plus', exon_count=17, OMIM=137165, no_hdr0=''),
  6530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6530, CurrentID=0, Status='live', Symbol='SLC6A2', Aliases='NAT1, NET, NET1, SLC6A5', description='solute carrier family 6 (neurotransmitter transporter), member 2', other_designations='neurotransmitter transporter|norepinephrine transporter|solute carrier family 6 (neurotransmitter transporter, noradrenalin), member 2|solute carrier family 6 member 5', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55655604, end_position_on_the_genomic_accession=55706192, orientation='plus', exon_count=19, OMIM=163970, no_hdr0=''),
  6531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6531, CurrentID=0, Status='live', Symbol='SLC6A3', Aliases='DAT, DAT1, PKDYS', description='solute carrier family 6 (neurotransmitter transporter), member 3', other_designations='DA transporter|dopamine transporter 1|solute carrier family 6 (neurotransmitter transporter, dopamine), member 3', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1392790, end_position_on_the_genomic_accession=1445428, orientation='minus', exon_count=15, OMIM=126455, no_hdr0=''),
  6532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6532, CurrentID=0, Status='live', Symbol='SLC6A4', Aliases='5-HTT, 5-HTTLPR, 5HTT, HTT, OCD1, SERT, SERT1, hSERT', description='solute carrier family 6 (neurotransmitter transporter), member 4', other_designations='5-hydroxytryptamine (serotonin) transporter|5HT transporter|Na+/Cl- dependent serotonin transporter|serotonin transporter 1|solute carrier family 6 (neurotransmitter transporter, serotonin), member 4', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30194319, end_position_on_the_genomic_accession=30235968, orientation='minus', exon_count=15, OMIM=182138, no_hdr0=''),
  6533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6533, CurrentID=0, Status='live', Symbol='SLC6A6', Aliases='TAUT', description='solute carrier family 6 (neurotransmitter transporter), member 6', other_designations='solute carrier family 6 (neurotransmitter transporter, taurine), member 6', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14402576, end_position_on_the_genomic_accession=14489349, orientation='plus', exon_count=20, OMIM=186854, no_hdr0=''),
  6534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6534, CurrentID=0, Status='live', Symbol='SLC6A7', Aliases='PROT', description='solute carrier family 6 (neurotransmitter transporter), member 7', other_designations='brain-specific L-proline transporter|solute carrier family 6 (neurotransmitter transporter, L-proline), member 7|solute carrier family 6 member 7', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150189957, end_position_on_the_genomic_accession=150211072, orientation='plus', exon_count=14, OMIM=606205, no_hdr0=''),
  6535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6535, CurrentID=0, Status='live', Symbol='SLC6A8', Aliases='CCDS1, CRT, CRTR, CT1, CTR5', description='solute carrier family 6 (neurotransmitter transporter), member 8', other_designations='creatine transporter 1|creatine transporter SLC6A8 variant D|solute carrier family 6 (neurotransmitter transporter, creatine), member 8|solute carrier family 6 member 8', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153688297, end_position_on_the_genomic_accession=153696593, orientation='plus', exon_count=14, OMIM=300036, no_hdr0=''),
  6536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6536, CurrentID=0, Status='live', Symbol='SLC6A9', Aliases='GLYT1', description='solute carrier family 6 (neurotransmitter transporter, glycine), member 9', other_designations='glyT-1|solute carrier family 6 member 9', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43996483, end_position_on_the_genomic_accession=44031492, orientation='minus', exon_count=17, OMIM=601019, no_hdr0=''),
  6538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6538, CurrentID=0, Status='live', Symbol='SLC6A11', Aliases='GAT-3, GAT3, GAT4', description='solute carrier family 6 (neurotransmitter transporter), member 11', other_designations='GABA transporter 3|solute carrier family 6 (neurotransmitter transporter, GABA), member 11|solute carrier family 6 member 11', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10816200, end_position_on_the_genomic_accession=10941471, orientation='plus', exon_count=16, OMIM=607952, no_hdr0=''),
  6539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6539, CurrentID=0, Status='live', Symbol='SLC6A12', Aliases='BGT-1, BGT1, GAT2', description='solute carrier family 6 (neurotransmitter transporter), member 12', other_designations='Na(+)/Cl(-) betaine/GABA transporter|betaine/GABA transporter-1|gamma-aminobutyric acid transporter|solute carrier family 6 (neurotransmitter transporter, betaine/GABA), member 12', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=187458, end_position_on_the_genomic_accession=214205, orientation='minus', exon_count=23, OMIM=603080, no_hdr0=''),
  6540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6540, CurrentID=0, Status='live', Symbol='SLC6A13', Aliases='GAT-2, GAT2, GAT3', description='solute carrier family 6 (neurotransmitter transporter), member 13', other_designations='GABA transport protein|GABA transporter 2|solute carrier family 6 (neurotransmitter transporter, GABA), member 13', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=220621, end_position_on_the_genomic_accession=264308, orientation='minus', exon_count=19, OMIM=615097, no_hdr0=''),
  6541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6541, CurrentID=0, Status='live', Symbol='SLC7A1', Aliases='ATRC1, CAT-1, ERR, HCAT1, REC1L', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 1', other_designations='CAT1|amino acid transporter, cationic 1|ecotropic retroviral leukemia receptor homolog|ecotropic retroviral receptor|ecotropic retrovirus receptor homolog|solute carrier family 7 member 1|system Y+ basic amino acid transporter', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29509410, end_position_on_the_genomic_accession=29595688, orientation='minus', exon_count=14, OMIM=104615, no_hdr0=''),
  6542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6542, CurrentID=0, Status='live', Symbol='SLC7A2', Aliases='ATRC2, CAT2, HCAT2', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 2', other_designations='low affinity cationic amino acid transporter 2', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17497088, end_position_on_the_genomic_accession=17570568, orientation='plus', exon_count=16, OMIM=601872, no_hdr0=''),
  6543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6543, CurrentID=0, Status='live', Symbol='SLC8A2', Aliases='NCX2', description='solute carrier family 8 (sodium/calcium exchanger), member 2', other_designations='Na(+)/Ca(2+)-exchange protein 2|Na+/Ca2+-exchanging protein Nac2|sodium-calcium exchanger 2|solute carrier family 8 (sodium-calcium exchanger), member 2|solute carrier family 8 member 2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47428017, end_position_on_the_genomic_accession=47472050, orientation='minus', exon_count=10, OMIM=601901, no_hdr0=''),
  6545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6545, CurrentID=0, Status='live', Symbol='SLC7A4', Aliases='CAT-4, CAT4, HCAT3, VH', description='solute carrier family 7 member 4', other_designations='Ig heavy chain variable region|VH 3 family|solute carrier family 7 (cationic amino acid transporter, y+ system), member 4|solute carrier family 7 (orphan transporter), member 4|solute carrier family 7, member 4', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21028718, end_position_on_the_genomic_accession=21032558, orientation='minus', exon_count=5, OMIM=603752, no_hdr0=''),
  6546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6546, CurrentID=0, Status='live', Symbol='SLC8A1', Aliases='NCX1', description='solute carrier family 8 (sodium/calcium exchanger), member 1', other_designations='Na(+)/Ca(2+)-exchange protein 1|Na+/Ca++ exchanger|Na+/Ca2+ exchanger|solute carrier family 8 member 1', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40112146, end_position_on_the_genomic_accession=40512486, orientation='minus', exon_count=20, OMIM=182305, no_hdr0=''),
  6547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6547, CurrentID=0, Status='live', Symbol='SLC8A3', Aliases='NCX3', description='solute carrier family 8 (sodium/calcium exchanger), member 3', other_designations='Na(+)/Ca(2+)-exchange protein 3|Na+/Ca2+ exchanger 3|sodium/calcium exchanger SLC8A3|solute carrier family 8 (sodium-calcium exchanger), member 3|solute carrier family 8 member 3', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70044217, end_position_on_the_genomic_accession=70223612, orientation='minus', exon_count=14, OMIM=607991, no_hdr0=''),
  6548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6548, CurrentID=0, Status='live', Symbol='SLC9A1', Aliases='APNH, LIKNS, NHE-1, NHE1, PPP1R143', description='solute carrier family 9, subfamily A (NHE1, cation proton antiporter 1), member 1', other_designations='Na(+)/H(+) exchanger 1|Na-Li countertransporter|protein phosphatase 1, regulatory subunit 143|solute carrier family 9 (sodium/hydrogen exchanger), isoform 1 (antiporter, Na+/H+, amiloride sensitive)|solute carrier family 9 (sodium/hydrogen exchanger), member 1 (antiporter, Na+/H+, amiloride sensitive)', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27098809, end_position_on_the_genomic_accession=27155130, orientation='minus', exon_count=13, OMIM=107310, no_hdr0=''),
  6549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6549, CurrentID=0, Status='live', Symbol='SLC9A2', Aliases='NHE2', description='solute carrier family 9, subfamily A (NHE2, cation proton antiporter 2), member 2', other_designations='Na(+)/H(+) exchanger 2|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 (sodium/hydrogen exchanger), member 2|solute carrier family 9 member 2', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102619689, end_position_on_the_genomic_accession=102711350, orientation='plus', exon_count=12, OMIM=600530, no_hdr0=''),
  6550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6550, CurrentID=0, Status='live', Symbol='SLC9A3', Aliases='NHE-3, NHE3', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3', other_designations='Na(+)/H(+) exchanger 3|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 (sodium/hydrogen exchanger), member 3|solute carrier family 9 member 3', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=470510, end_position_on_the_genomic_accession=570428, orientation='minus', exon_count=18, OMIM=182307, no_hdr0=''),
  6551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6551, CurrentID=0, Status='live', Symbol='SLC9A3P1', Aliases='NHE3P, SLC9A3P, SLC9A3P4', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 pseudogene 1', other_designations='Na+H+ exchanger 3 pseudogene|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 pseudogene|solute carrier family 9 (sodium/hydrogen exchanger), member 3 pseudogene 1|solute carrier family 9 (sodium/hydrogen exchanger), member 3 pseudogene 4|solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 pseudogene 4', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50150391, end_position_on_the_genomic_accession=50152586, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6553, CurrentID=0, Status='live', Symbol='SLC9A5', Aliases='NHE5', description='solute carrier family 9, subfamily A (NHE5, cation proton antiporter 5), member 5', other_designations='NHE-5|Na(+)/H(+) exchanger 5|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 (sodium/hydrogen exchanger), member 5|solute carrier family 9 member 5', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67247551, end_position_on_the_genomic_accession=67272191, orientation='plus', exon_count=17, OMIM=600477, no_hdr0=''),
  6554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6554, CurrentID=0, Status='live', Symbol='SLC10A1', Aliases='NTCP', description='solute carrier family 10 (sodium/bile acid cotransporter), member 1', other_designations='Na(+)/bile acid cotransporter|Na(+)/taurocholate transport protein|Na/taurocholate cotransporting polypeptide|cell growth-inhibiting gene 29 protein|growth-inhibiting protein 29|sodium/taurocholate cotransporter|sodium/taurocholate cotransporting polypeptide|solute carrier family 10 (sodium/bile acid cotransporter family), member 1|solute carrier family 10 member 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69775835, end_position_on_the_genomic_accession=69797289, orientation='minus', exon_count=5, OMIM=182396, no_hdr0=''),
  6555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6555, CurrentID=0, Status='live', Symbol='SLC10A2', Aliases='ASBT, IBAT, ISBT, NTCP2, PBAM', description='solute carrier family 10 (sodium/bile acid cotransporter), member 2', other_designations='Na(+)-dependent ileal bile acid transporter|ileal apical sodium-dependent bile acid transporter|ileal sodium-dependent bile acid transporter|sodium/taurocholate cotransporting polypeptide, ileal|solute carrier family 10 (sodium/bile acid cotransporter family), member 2', map_location='13q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=103043998, end_position_on_the_genomic_accession=103066846, orientation='minus', exon_count=6, OMIM=601295, no_hdr0=''),
  6556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6556, CurrentID=0, Status='live', Symbol='SLC11A1', Aliases='LSH, NRAMP, NRAMP1', description='solute carrier family 11 (proton-coupled divalent metal ion transporter), member 1', other_designations='NRAMP 1|solute carrier family 11 (proton-coupled divalent metal ion transporters), member 1|solute carrier family 11 (sodium/phosphate symporters), member 1|solute carrier family 11 member 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218381766, end_position_on_the_genomic_accession=218396894, orientation='plus', exon_count=16, OMIM=600266, no_hdr0=''),
  6557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6557, CurrentID=0, Status='live', Symbol='SLC12A1', Aliases='BSC1, NKCC2', description='solute carrier family 12 (sodium/potassium/chloride transporter), member 1', other_designations='NKCC2A variant A|Na-K-2Cl cotransporter|bumetanide-sensitive sodium-(potassium)-chloride cotransporter 2|kidney-specific Na-K-Cl symporter|solute carrier family 12 (sodium/potassium/chloride transporters), member 1', map_location='15q15-q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48206301, end_position_on_the_genomic_accession=48304078, orientation='plus', exon_count=29, OMIM=600839, no_hdr0=''),
  6558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6558, CurrentID=0, Status='live', Symbol='SLC12A2', Aliases='BSC, BSC2, NKCC1, PPP1R141', description='solute carrier family 12 (sodium/potassium/chloride transporter), member 2', other_designations='basolateral Na-K-Cl symporter|bumetanide-sensitive sodium-(potassium)-chloride cotransporter 1|protein phosphatase 1, regulatory subunit 141|solute carrier family 12 (sodium/potassium/chloride transporters), member 2', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128083791, end_position_on_the_genomic_accession=128189688, orientation='plus', exon_count=28, OMIM=600840, no_hdr0=''),
  6559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6559, CurrentID=0, Status='live', Symbol='SLC12A3', Aliases='NCC, NCCT, TSC', description='solute carrier family 12 (sodium/chloride transporter), member 3', other_designations='Na-Cl cotransporter|Na-Cl symporter|NaCl electroneutral thiazide-sensitive cotransporter|thiazide-sensitive Na-Cl cotransporter|thiazide-sensitive sodium-chloride cotransporter', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56865207, end_position_on_the_genomic_accession=56915850, orientation='plus', exon_count=26, OMIM=600968, no_hdr0=''),
  6560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6560, CurrentID=0, Status='live', Symbol='SLC12A4', Aliases='KCC1', description='solute carrier family 12 (potassium/chloride transporter), member 4', other_designations='CTC-479C5.17|electroneutral potassium-chloride cotransporter 1|erythroid K-Cl cotransporter 1|hKCC1|potassium/chloride cotransporter 1|solute carrier family 12 (potassium/chloride transporters), member 4', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67943474, end_position_on_the_genomic_accession=67968694, orientation='minus', exon_count=26, OMIM=604119, no_hdr0=''),
  6561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6561, CurrentID=0, Status='live', Symbol='SLC13A1', Aliases='NAS1, NaSi-1', description='solute carrier family 13 (sodium/sulfate symporter), member 1', other_designations='Na(+)/sulfate cotransporter|hNaSi-1|renal sodium/sulfate cotransporter|solute carrier family 13 (sodium/sulfate symporters), member 1|solute carrier family 13 (sodium/sulphate symporters), member 1', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123113534, end_position_on_the_genomic_accession=123199975, orientation='minus', exon_count=19, OMIM=606193, no_hdr0=''),
  6563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6563, CurrentID=0, Status='live', Symbol='SLC14A1', Aliases='HUT11, HsT1341, JK, RACH1, RACH2, UT-B1, UT1, UTE', description='solute carrier family 14 (urea transporter), member 1 (Kidd blood group)', other_designations='Kidd blood group|SLC14A1 JK|urea transporter JK glycoprotein|urea transporter, erythrocyte|urea transporter-B1', map_location='18q11-q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45723756, end_position_on_the_genomic_accession=45752520, orientation='plus', exon_count=14, OMIM=613868, no_hdr0=''),
  6564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6564, CurrentID=0, Status='live', Symbol='SLC15A1', Aliases='HPECT1, HPEPT1, PEPT1', description='solute carrier family 15 (oligopeptide transporter), member 1', other_designations='Caco-2 oligopeptide transporter|intestinal H+/peptide cotransporter|macrophage oligopeptide transporter PEPT1|oligopeptide transporter, small intestine isoform|peptide transporter 1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98683801, end_position_on_the_genomic_accession=98752675, orientation='minus', exon_count=23, OMIM=600544, no_hdr0=''),
  6565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6565, CurrentID=0, Status='live', Symbol='SLC15A2', Aliases='PEPT2', description='solute carrier family 15 (oligopeptide transporter), member 2', other_designations='kidney H(+)/peptide cotransporter|oligopeptide transporter, kidney isoform|peptide transporter 2|solute carrier family 15 (H+/peptide transporter), member 2', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121894324, end_position_on_the_genomic_accession=121944187, orientation='plus', exon_count=23, OMIM=602339, no_hdr0=''),
  6566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6566, CurrentID=0, Status='live', Symbol='SLC16A1', Aliases='HHF7, MCT, MCT1, MCT1D', description='solute carrier family 16 (monocarboxylate transporter), member 1', other_designations='MCT 1|solute carrier family 16 (monocarboxylic acid transporters), member 1|solute carrier family 16, member 1 (monocarboxylic acid transporter 1)', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112911847, end_position_on_the_genomic_accession=112956353, orientation='minus', exon_count=7, OMIM=600682, no_hdr0=''),
  6567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6567, CurrentID=0, Status='live', Symbol='SLC16A2', Aliases='AHDS, DXS128, DXS128E, MCT 7, MCT 8, MCT7, MCT8, MRX22, XPCT', description='solute carrier family 16, member 2 (thyroid hormone transporter)', other_designations='X-linked PEST-containing transporter|monocarboxylate transporter 7', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74421493, end_position_on_the_genomic_accession=74533929, orientation='plus', exon_count=7, OMIM=300095, no_hdr0=''),
  6568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6568, CurrentID=0, Status='live', Symbol='SLC17A1', Aliases='NAPI-1, NPT-1, NPT1', description='solute carrier family 17 (organic anion transporter), member 1', other_designations='Na(+)/PI cotransporter 1|na/Pi-4|renal Na(+)-dependent phosphate cotransporter 1|renal sodium-dependent phosphate transport protein 1|renal sodium-phosphate transport protein 1|sodium phosphate transporter|sodium/phosphate cotransporter 1|sodium/phosphate type I cotransporter|solute carrier family 17 (sodium phosphate), member 1|solute carrier family 17 (vesicular glutamate transporter), member 1|solute carrier family 17 member 1', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25781454, end_position_on_the_genomic_accession=25832070, orientation='minus', exon_count=13, OMIM=182308, no_hdr0=''),
  6569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6569, CurrentID=0, Status='live', Symbol='SLC34A1', Aliases='FRTS2, NAPI-3, NPHLOP1, NPT2, NPTIIa, SLC11, SLC17A2', description='solute carrier family 34 (type II sodium/phosphate cotransporter), member 1', other_designations='Na(+)-dependent phosphate cotransporter 2A|Na(+)/Pi cotransporter 2A|Na+-phosphate cotransporter type II|naPi-2a|renal sodium-dependent phosphate transporter|sodium-phosphate transport protein 2A|sodium/phosphate co-transporter|sodium/phosphate cotransporter 2A|solute carrier family 17 (sodium phosphate), member 2|solute carrier family 34 (sodium phosphate), member 1|solute carrier family 34 member 1', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177384431, end_position_on_the_genomic_accession=177398848, orientation='plus', exon_count=15, OMIM=182309, no_hdr0=''),
  6570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6570, CurrentID=0, Status='live', Symbol='SLC18A1', Aliases='CGAT, VAT1, VMAT1', description='solute carrier family 18 (vesicular monoamine transporter), member 1', other_designations='solute carrier family 18 (vesicular monoamine), member 1|solute carrier family 18 member 1|vesicular amine transporter 1', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20144855, end_position_on_the_genomic_accession=20183338, orientation='minus', exon_count=17, OMIM=193002, no_hdr0=''),
  6571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6571, CurrentID=0, Status='live', Symbol='SLC18A2', Aliases='SVAT, SVMT, VAT2, VMAT2', description='solute carrier family 18 (vesicular monoamine transporter), member 2', other_designations='monoamine neurotransmitter transporter|monoamine transporter|solute carrier family 18 (vesicular monoamine), member 2|solute carrier family 18 member 2|vesicle monoamine transporter type 2|vesicle monoamine/H+ antiporter|vesicular amine transporter 2', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117241073, end_position_on_the_genomic_accession=117279430, orientation='plus', exon_count=16, OMIM=193001, no_hdr0=''),
  6572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6572, CurrentID=0, Status='live', Symbol='SLC18A3', Aliases='VACHT', description='solute carrier family 18 (vesicular acetylcholine transporter), member 3', other_designations='solute carrier family 18 (vesicular acetylcholine), member 3|solute carrier family 18 member 3|solute carrier family 18, member 3', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49610301, end_position_on_the_genomic_accession=49612720, orientation='plus', exon_count=1, OMIM=600336, no_hdr0=''),
  6573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6573, CurrentID=0, Status='live', Symbol='SLC19A1', Aliases='CHMD, FOLT, IFC1, REFC, RFC1', description='solute carrier family 19 (folate transporter), member 1', other_designations='IFC-1|RFC|intestinal folate carrier 1|placental folate transporter|reduced folate carrier protein|solute carrier family 19 member 1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45502517, end_position_on_the_genomic_accession=45563126, orientation='minus', exon_count=18, OMIM=600424, no_hdr0=''),
  6574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6574, CurrentID=0, Status='live', Symbol='SLC20A1', Aliases='GLVR1, Glvr-1, PIT1, PiT-1', description='solute carrier family 20 (phosphate transporter), member 1', other_designations='gibbon ape leukemia virus receptor 1|leukemia virus receptor 1 homolog|solute carrier family 20 member 1', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112645857, end_position_on_the_genomic_accession=112663825, orientation='plus', exon_count=12, OMIM=137570, no_hdr0=''),
  6575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6575, CurrentID=0, Status='live', Symbol='SLC20A2', Aliases='GLVR-2, GLVR2, IBGC1, IBGC3, MLVAR, PIT-2, PIT2, RAM1', description='solute carrier family 20 (phosphate transporter), member 2', other_designations='gibbon ape leukemia virus receptor 2|murine leukemia virus, amphotropic, receptor for', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42416462, end_position_on_the_genomic_accession=42542213, orientation='minus', exon_count=16, OMIM=158378, no_hdr0=''),
  6576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6576, CurrentID=0, Status='live', Symbol='SLC25A1', Aliases='CTP, D2L2AD, SEA, SLC20A3', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1', other_designations='citrate transport protein|solute carrier family 20 (mitochondrial citrate transporter), member 3|tricarboxylate carrier protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19175575, end_position_on_the_genomic_accession=19178863, orientation='minus', exon_count=10, OMIM=190315, no_hdr0=''),
  100271176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271176, CurrentID=0, Status='live', Symbol='RPL21P63', Aliases='RPL21_24_657', description='ribosomal protein L21 pseudogene 63', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10574712, end_position_on_the_genomic_accession=10575239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6578, CurrentID=0, Status='live', Symbol='SLCO2A1', Aliases='MATR1, OATP2A1, PGT, PHOAR2, SLC21A2', description='solute carrier organic anion transporter family member 2A1', other_designations='matrin F/G 1|solute carrier family 21 (prostaglandin transporter), member 2|solute carrier organic anion transporter family, member 2A1', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133932696, end_position_on_the_genomic_accession=134030076, orientation='minus', exon_count=15, OMIM=601460, no_hdr0=''),
  6579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6579, CurrentID=0, Status='live', Symbol='SLCO1A2', Aliases='OATP, OATP-A, OATP1A2, SLC21A3', description='solute carrier organic anion transporter family member 1A2', other_designations='OATP-1|organic anion transporting polypeptide A|organic anion-transporting polypeptide 1|sodium-independent organic anion transporter|solute carrier family 21 (organic anion transporter), member 3|solute carrier family 21 member 3|solute carrier organic anion transporter family, member 1A2', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21264600, end_position_on_the_genomic_accession=21403657, orientation='minus', exon_count=21, OMIM=602883, no_hdr0=''),
  6580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6580, CurrentID=0, Status='live', Symbol='SLC22A1', Aliases='HOCT1, OCT1, oct1_cds', description='solute carrier family 22 (organic cation transporter), member 1', other_designations='organic cation transporter 1', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160121789, end_position_on_the_genomic_accession=160159198, orientation='plus', exon_count=15, OMIM=602607, no_hdr0=''),
  6581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6581, CurrentID=0, Status='live', Symbol='SLC22A3', Aliases='EMT, EMTH, OCT3', description='solute carrier family 22 (organic cation transporter), member 3', other_designations='EMT organic cation transporter 3|extraneuronal monoamine transporter|organic cation transporter 3|solute carrier family 22 (extraneuronal monoamine transporter), member 3', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160348373, end_position_on_the_genomic_accession=160452581, orientation='plus', exon_count=18, OMIM=604842, no_hdr0=''),
  6582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6582, CurrentID=0, Status='live', Symbol='SLC22A2', Aliases='OCT2', description='solute carrier family 22 (organic cation transporter), member 2', other_designations='organic cation transporter 2', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160216762, end_position_on_the_genomic_accession=160258931, orientation='minus', exon_count=11, OMIM=602608, no_hdr0=''),
  6583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6583, CurrentID=0, Status='live', Symbol='SLC22A4', Aliases='OCTN1', description='solute carrier family 22 (organic cation/zwitterion transporter), member 4', other_designations='ET transporter|ergothioneine transporter|integral membrane transport protein|organic cation/carnitine transporter 1|solute carrier family 22 (organic cation/ergothioneine transporter), member 4', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132294384, end_position_on_the_genomic_accession=132344206, orientation='plus', exon_count=10, OMIM=604190, no_hdr0=''),
  6584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6584, CurrentID=0, Status='live', Symbol='SLC22A5', Aliases='CDSP, OCTN2', description='solute carrier family 22 (organic cation/carnitine transporter), member 5', other_designations='high-affinity sodium dependent carnitine cotransporter|organic cation/carnitine transporter 2', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132369704, end_position_on_the_genomic_accession=132395614, orientation='plus', exon_count=11, OMIM=603377, no_hdr0=''),
  6585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6585, CurrentID=0, Status='live', Symbol='SLIT1', Aliases='MEGF4, SLIL1, SLIT-1, SLIT3', description='slit guidance ligand 1', other_designations='multiple EGF-like domains protein 4|multiple epidermal growth factor-like domains protein 4|slit homolog 1', map_location='10q23.3-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96998038, end_position_on_the_genomic_accession=97185926, orientation='minus', exon_count=37, OMIM=603742, no_hdr0=''),
  6586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6586, CurrentID=0, Status='live', Symbol='SLIT3', Aliases='MEGF5, SLIL2, SLIT1, Slit-3, slit2', description='slit guidance ligand 3', other_designations='multiple EGF-like domains protein 5|multiple epidermal growth factor-like domains protein 5|slit homolog 3', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168661733, end_position_on_the_genomic_accession=169301129, orientation='minus', exon_count=36, OMIM=603745, no_hdr0=''),
  6588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6588, CurrentID=0, Status='live', Symbol='SLN', Aliases='', description='sarcolipin', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107707375, end_position_on_the_genomic_accession=107712061, orientation='minus', exon_count=2, OMIM=602203, no_hdr0=''),
  100271178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271178, CurrentID=0, Status='live', Symbol='RPL21P65', Aliases='RPL21_26_729', description='ribosomal protein L21 pseudogene 65', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106642420, end_position_on_the_genomic_accession=106642959, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6590, CurrentID=0, Status='live', Symbol='SLPI', Aliases='ALK1, ALP, BLPI, HUSI, HUSI-I, MPI, WAP4, WFDC4', description='secretory leukocyte peptidase inhibitor', other_designations='HUSI-1|WAP four-disulfide core domain protein 4|mucus proteinase inhibitor|protease inhibitor WAP4|secretory leukocyte protease inhibitor (antileukoproteinase)|seminal proteinase inhibitor', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45252239, end_position_on_the_genomic_accession=45254564, orientation='minus', exon_count=4, OMIM=107285, no_hdr0=''),
  6591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6591, CurrentID=0, Status='live', Symbol='SNAI2', Aliases='SLUG, SLUGH1, SNAIL2, WS2D', description='snail family zinc finger 2', other_designations='neural crest transcription factor SLUG|protein snail homolog 2|slug (chicken homolog), zinc finger protein|snail homolog 2', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48917677, end_position_on_the_genomic_accession=48921440, orientation='minus', exon_count=3, OMIM=602150, no_hdr0=''),
  6594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6594, CurrentID=0, Status='live', Symbol='SMARCA1', Aliases='ISWI, NURF140, SNF2L, SNF2L1, SNF2LB, SNF2LT, SWI, SWI2', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 1', other_designations='ATP-dependent helicase SMARCA1|SNF2-like 1|global transcription activator homologous sequence|nucleosome-remodeling factor subunit SNF2L|sucrose nonfermenting 2-like protein 1', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129446501, end_position_on_the_genomic_accession=129523513, orientation='minus', exon_count=25, OMIM=300012, no_hdr0=''),
  6595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6595, CurrentID=0, Status='live', Symbol='SMARCA2', Aliases='BAF190, BRM, NCBRS, SNF2, SNF2L2, SNF2LA, SWI2, Sth1p, hBRM, hSNF2a', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 2', other_designations='ATP-dependent helicase SMARCA2|BAF190B|BRG1-associated factor 190B|SNF2-alpha|SNF2/SWI2-like protein 2|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin a2|global transcription activator homologous sequence|protein brahma homolog|sucrose nonfermenting 2-like protein 2', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2015219, end_position_on_the_genomic_accession=2193624, orientation='plus', exon_count=38, OMIM=600014, no_hdr0=''),
  6596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6596, CurrentID=0, Status='live', Symbol='HLTF', Aliases='HIP116, HIP116A1, RNF80, SMARCA3, SNF2L3, ZBU1, HLTF', description='helicase-like transcription factor', other_designations='DNA-binding protein/plasminogen activator inhibitor 1 regulator|DNA-binding protein/plasminogen activator inhibitor-1 regulator|RING finger protein 80|SNF2-like 3|SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 3|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily A member 3|sucrose nonfermenting protein 2-like 3|sucrose nonfermenting-like 3', map_location='3q25.1-q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149030117, end_position_on_the_genomic_accession=149086554, orientation='minus', exon_count=25, OMIM=603257, no_hdr0=''),
  6597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6597, CurrentID=0, Status='live', Symbol='SMARCA4', Aliases='BAF190, BAF190A, BRG1, MRD16, RTPS2, SNF2, SNF2L4, SNF2LB, SWI2, hSNF2b', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 4', other_designations='ATP-dependent helicase SMARCA4|BRG1-associated factor 190A|BRM/SWI2-related gene 1|SNF2-beta|SNF2-like 4|brahma protein-like 1|global transcription activator homologous sequence|homeotic gene regulator|mitotic growth and transcription activator|nuclear protein GRB1|protein BRG-1|protein brahma homolog 1|sucrose nonfermenting-like 4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10960922, end_position_on_the_genomic_accession=11062282, orientation='plus', exon_count=38, OMIM=603254, no_hdr0=''),
  6598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6598, CurrentID=0, Status='live', Symbol='SMARCB1', Aliases='BAF47, INI1, MRD15, PPP1R144, RDT, RTPS1, SNF5, SNF5L1, SWNTS1, Sfh1p, Snr1, hSNFS', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily b, member 1', other_designations='BRG1-associated factor 47|SNF5 homolog|SWI/SNF-related matrix-associated protein|hSNF5|integrase interactor 1 protein|malignant rhabdoid tumor suppressor|protein phosphatase 1, regulatory subunit 144|sucrose nonfermenting, yeast, homolog-like 1', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23786963, end_position_on_the_genomic_accession=23834518, orientation='plus', exon_count=9, OMIM=601607, no_hdr0=''),
  6599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6599, CurrentID=0, Status='live', Symbol='SMARCC1', Aliases='BAF155, CRACC1, Rsc8, SRG3, SWI3', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily c, member 1', other_designations='BRG1-associated factor 155|SWI/SNF complex 155 kDa subunit|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily C member 1|chromatin remodeling complex BAF155 subunit|mammalian chromatin remodeling complex BRG1-associated factor 155', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47585888, end_position_on_the_genomic_accession=47781915, orientation='minus', exon_count=30, OMIM=601732, no_hdr0=''),
  6601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6601, CurrentID=0, Status='live', Symbol='SMARCC2', Aliases='BAF170, CRACC2, Rsc8', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily c, member 2', other_designations='BRG1-associated factor 170|SWI/SNF complex 170 kDa subunit|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily C member 2|SWI3-like protein|chromatin remodeling complex BAF170 subunit|mammalian chromatin remodeling complex BRG1-associated factor 170', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56161852, end_position_on_the_genomic_accession=56189567, orientation='minus', exon_count=30, OMIM=601734, no_hdr0=''),
  6602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6602, CurrentID=0, Status='live', Symbol='SMARCD1', Aliases='BAF60A, CRACD1, Rsc6p', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 1', other_designations='60 kDa BRG-1/Brm-associated factor subunit A|BRG1-associated factor 60A|SWI/SNF complex 60 kDa subunit A|Swp73-like protein|chromatin remodeling complex BAF60A subunit|mammalian chromatin remodeling complex BRG1-associated factor 60A', map_location='12q13-q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50085200, end_position_on_the_genomic_accession=50100711, orientation='plus', exon_count=13, OMIM=601735, no_hdr0=''),
  6603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6603, CurrentID=0, Status='live', Symbol='SMARCD2', Aliases='BAF60B, CRACD2, PRO2451, Rsc6p', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 2', other_designations='60 kDa BRG-1/Brm-associated factor subunit B|BRG1-associated factor 60B|SWI/SNF complex 60 kDa subunit B|Swp73-like protein|chromatin remodeling complex BAF60B subunit|mammalian chromatin remodeling complex BRG1-associated factor 60B', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63832081, end_position_on_the_genomic_accession=63842991, orientation='minus', exon_count=13, OMIM=601736, no_hdr0=''),
  6604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6604, CurrentID=0, Status='live', Symbol='SMARCD3', Aliases='BAF60C, CRACD3, Rsc6p', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 3', other_designations='60 kDa BRG-1/Brm-associated factor subunit C|BRG1-associated factor 60C|SWI/SNF complex 60 kDa subunit C|Swp73-like protein|chromatin remodeling complex BAF60C subunit|mammalian chromatin remodeling complex BRG1-associated factor 60C', map_location='7q35-q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151238973, end_position_on_the_genomic_accession=151277149, orientation='minus', exon_count=17, OMIM=601737, no_hdr0=''),
  6605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6605, CurrentID=0, Status='live', Symbol='SMARCE1', Aliases='BAF57', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1', other_designations='BRG1-associated factor 57|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin e1|chromatin remodeling complex BRG1-associated factor 57', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40627720, end_position_on_the_genomic_accession=40647851, orientation='minus', exon_count=11, OMIM=603111, no_hdr0=''),
  6606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6606, CurrentID=0, Status='live', Symbol='SMN1', Aliases='BCD541, GEMIN1, SMA, SMA1, SMA2, SMA3, SMA4, SMA@, SMN, SMNT, T-BCD541, TDRD16A', description='survival of motor neuron 1, telomeric', other_designations='component of gems 1|gemin-1|survival motor neuron 1 protein|tudor domain containing 16A', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70924941, end_position_on_the_genomic_accession=70953015, orientation='plus', exon_count=10, OMIM=600354, no_hdr0=''),
  6607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6607, CurrentID=0, Status='live', Symbol='SMN2', Aliases='BCD541, C-BCD541, GEMIN1, SMNC, TDRD16B', description='survival of motor neuron 2, centromeric', other_designations='component of gems 1|gemin-1|tudor domain containing 16B', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70049523, end_position_on_the_genomic_accession=70077595, orientation='plus', exon_count=10, OMIM=601627, no_hdr0=''),
  6608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6608, CurrentID=0, Status='live', Symbol='SMO', Aliases='FZD11, GxH, SMO', description='smoothened, frizzled class receptor', other_designations='frizzled family member 11|protein Gx|seven transmembrane helix receptor|smoothened, frizzled family receptor|smoothened, seven transmembrane spanning receptor', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129188872, end_position_on_the_genomic_accession=129213546, orientation='plus', exon_count=13, OMIM=601500, no_hdr0=''),
  6609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6609, CurrentID=0, Status='live', Symbol='SMPD1', Aliases='ASM, ASMASE, NPD', description='sphingomyelin phosphodiesterase 1', other_designations='acid sphingomyelinase', map_location='11p15.4-p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6390301, end_position_on_the_genomic_accession=6394998, orientation='plus', exon_count=6, OMIM=607608, no_hdr0=''),
  6610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6610, CurrentID=0, Status='live', Symbol='SMPD2', Aliases='ISC1, NSMASE, NSMASE1', description='sphingomyelin phosphodiesterase 2', other_designations='N-SMase|lyso-PAF-PLC|lyso-platelet-activating factor-phospholipase C|neutral sphingomyelinase|sphingomyelin phosphodiesterase 2, neutral membrane (neutral sphingomyelinase)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109440646, end_position_on_the_genomic_accession=109443922, orientation='plus', exon_count=10, OMIM=603498, no_hdr0=''),
  6611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6611, CurrentID=0, Status='live', Symbol='SMS', Aliases='MRSR, SPMSY, SRS, SpS', description='spermine synthase', other_designations='spermidine aminopropyltransferase', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21940573, end_position_on_the_genomic_accession=21994837, orientation='plus', exon_count=13, OMIM=300105, no_hdr0=''),
  6612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6612, CurrentID=0, Status='live', Symbol='SUMO3', Aliases='SMT3A, SMT3H1, SUMO-3, Smt3B', description='small ubiquitin-like modifier 3', other_designations='SMT3 suppressor of mif two 3 homolog 1|SMT3 suppressor of mif two 3 homolog 3|ubiquitin-like protein SMT3B', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44805617, end_position_on_the_genomic_accession=44818129, orientation='minus', exon_count=4, OMIM=602231, no_hdr0=''),
  6613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6613, CurrentID=0, Status='live', Symbol='SUMO2', Aliases='HSMT3, SMT3B, SMT3H2, SUMO3, Smt3A', description='small ubiquitin-like modifier 2', other_designations='SMT3 homolog 2|SMT3 suppressor of mif two 3 homolog 2|sentrin 2|ubiquitin-like protein SMT3A|ubiquitin-like protein SMT3B', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75167730, end_position_on_the_genomic_accession=75183003, orientation='minus', exon_count=4, OMIM=603042, no_hdr0=''),
  6614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6614, CurrentID=0, Status='live', Symbol='SIGLEC1', Aliases='CD169, SIGLEC-1, SN, dJ1009E24.1', description='sialic acid binding Ig-like lectin 1, sialoadhesin', other_designations='sialic acid-binding Ig-like lectin 1|sialic acid-binding immunoglobulin-like lectin 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3686970, end_position_on_the_genomic_accession=3712617, orientation='minus', exon_count=23, OMIM=600751, no_hdr0=''),
  6615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6615, CurrentID=0, Status='live', Symbol='SNAI1', Aliases='SLUGH2, SNA, SNAH, SNAIL, SNAIL1, dJ710H13.1', description='snail family zinc finger 1', other_designations='protein sna|protein snail homolog 1|snail 1 homolog|snail 1 zinc finger protein|snail 1, zinc finger protein|snail homolog 1', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49982976, end_position_on_the_genomic_accession=49988886, orientation='plus', exon_count=3, OMIM=604238, no_hdr0=''),
  6616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6616, CurrentID=0, Status='live', Symbol='SNAP25', Aliases='CMS18, RIC-4, RIC4, SEC9, SNAP, SNAP-25, bA416N4.2, dJ1068F16.2', description='synaptosome associated protein 25kDa', other_designations='SUP|resistance to inhibitors of cholinesterase 4 homolog|super protein|synaptosomal-associated protein, 25kD|synaptosomal-associated protein, 25kDa', map_location='20p12-p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10218829, end_position_on_the_genomic_accession=10307420, orientation='plus', exon_count=10, OMIM=600322, no_hdr0=''),
  6617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6617, CurrentID=0, Status='live', Symbol='SNAPC1', Aliases='PTFgamma, SNAP43', description='small nuclear RNA activating complex polypeptide 1', other_designations='PSE-binding factor subunit gamma|PTF subunit gamma|SNAPc 43 kDa subunit|SNAPc subunit 1|proximal sequence element-binding transcription factor subunit gamma|small nuclear RNA activating complex, polypeptide 1, 43kD|small nuclear RNA activating complex, polypeptide 1, 43kDa|small nuclear RNA-activating complex polypeptide 1|snRNA-activating protein complex 43 kDa subunit', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61762357, end_position_on_the_genomic_accession=61796428, orientation='plus', exon_count=10, OMIM=600591, no_hdr0=''),
  6618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6618, CurrentID=0, Status='live', Symbol='SNAPC2', Aliases='PTFDELTA, SNAP45', description='small nuclear RNA activating complex polypeptide 2', other_designations='PSE-binding factor subunit delta|PTF subunit delta|SNAPC 45 kDa subunit|SNAPc subunit 2|proximal sequence element-binding transcription factor subunit delta|small nuclear RNA activating complex, polypeptide 2, 45kD|small nuclear RNA activating complex, polypeptide 2, 45kDa|small nuclear RNA-activating complex polypeptide 2|snRNA-activating protein complex 45 kDa subunit', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7920309, end_position_on_the_genomic_accession=7923251, orientation='plus', exon_count=6, OMIM=605076, no_hdr0=''),
  6619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6619, CurrentID=0, Status='live', Symbol='SNAPC3', Aliases='PTFbeta, SNAP50', description='small nuclear RNA activating complex polypeptide 3', other_designations='PSE-binding factor subunit beta|PTF subunit beta|SNAPc 50 kDa subunit|SNAPc subunit 3|proximal sequence element-binding transcription factor subunit beta|small nuclear RNA activating complex, polypeptide 3, 50kD|small nuclear RNA activating complex, polypeptide 3, 50kDa|small nuclear RNA-activating complex polypeptide 3|snRNA-activating protein complex 50 kDa subunit', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15422696, end_position_on_the_genomic_accession=15466751, orientation='plus', exon_count=12, OMIM=602348, no_hdr0=''),
  6620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6620, CurrentID=0, Status='live', Symbol='SNCB', Aliases='', description='synuclein beta', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176620209, end_position_on_the_genomic_accession=176630556, orientation='minus', exon_count=7, OMIM=602569, no_hdr0=''),
  6621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6621, CurrentID=0, Status='live', Symbol='SNAPC4', Aliases='PTFalpha, SNAP190', description='small nuclear RNA activating complex polypeptide 4', other_designations='PSE-binding factor subunit alpha|PTF subunit alpha|SNAPc 190 kDa subunit|SNAPc subunit 4|proximal sequence element-binding transcription factor subunit alpha|small nuclear RNA activating complex, polypeptide 4, 190kD|small nuclear RNA activating complex, polypeptide 4, 190kDa|snRNA-activating protein complex 190 kDa subunit', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136375571, end_position_on_the_genomic_accession=136400208, orientation='minus', exon_count=25, OMIM=602777, no_hdr0=''),
  6622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6622, CurrentID=0, Status='live', Symbol='SNCA', Aliases='NACP, PARK1, PARK4, PD1', description='synuclein alpha', other_designations='non A-beta component of AD amyloid|synuclein alpha-140|synuclein, alpha (non A4 component of amyloid precursor)', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89724099, end_position_on_the_genomic_accession=89838314, orientation='minus', exon_count=9, OMIM=163890, no_hdr0=''),
  6623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6623, CurrentID=0, Status='live', Symbol='SNCG', Aliases='BCSG1, SR', description='synuclein gamma', other_designations='breast cancer-specific gene 1 protein|persyn|synoretin|synuclein, gamma (breast cancer-specific protein 1)', map_location='10q23.2-q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86958531, end_position_on_the_genomic_accession=86963260, orientation='plus', exon_count=5, OMIM=602998, no_hdr0=''),
  6624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6624, CurrentID=0, Status='live', Symbol='FSCN1', Aliases='FAN1, HSN, SNL, p55', description='fascin actin-bundling protein 1', other_designations='55 kDa actin-bundling protein|fascin homolog 1, actin-bundling protein|singed-like (fascin homolog, sea urchin)', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5592805, end_position_on_the_genomic_accession=5606656, orientation='plus', exon_count=5, OMIM=602689, no_hdr0=''),
  6625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6625, CurrentID=0, Status='live', Symbol='SNRNP70', Aliases='RNPU1Z, RPU1, SNRP70, Snp1, U1-70K, U170K, U1AP, U1RNP', description='small nuclear ribonucleoprotein, U1 70kDa subunit', other_designations='U1 snRNP 70 kDa|small nuclear ribonucleoprotein 70kDa (U1)', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49085140, end_position_on_the_genomic_accession=49108613, orientation='plus', exon_count=13, OMIM=180740, no_hdr0=''),
  6626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6626, CurrentID=0, Status='live', Symbol='SNRPA', Aliases='Mud1, U1-A, U1A', description='small nuclear ribonucleoprotein polypeptide A', other_designations='U1 small nuclear RNP-specific A|U1 snRNP A|U1 snRNP-specific protein A', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40750854, end_position_on_the_genomic_accession=40765392, orientation='plus', exon_count=6, OMIM=182285, no_hdr0=''),
  6627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6627, CurrentID=0, Status='live', Symbol='SNRPA1', Aliases='Lea1', description="small nuclear ribonucleoprotein polypeptide A'", other_designations="U2 small nuclear ribonucleoprotein polypeptide A'|U2 snRNP A'", map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101281510, end_position_on_the_genomic_accession=101295277, orientation='minus', exon_count=9, OMIM=603521, no_hdr0=''),
  6628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6628, CurrentID=0, Status='live', Symbol='SNRPB', Aliases="CCMS, COD1, Sm-B/B', SmB/B', SmB/SmB', snRNP-B, SNRPB", description='small nuclear ribonucleoprotein polypeptides B and B1', other_designations="B polypeptide of Sm protein|Sm protein B/B'|sm-B/Sm-B'|small nuclear ribonucleoprotein polypeptide B|small nuclear ribonucleoprotein polypeptides B and B'", map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2461635, end_position_on_the_genomic_accession=2470853, orientation='minus', exon_count=7, OMIM=182282, no_hdr0=''),
  6629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6629, CurrentID=0, Status='live', Symbol='SNRPB2', Aliases='Msl1', description='small nuclear ribonucleoprotein polypeptide B', other_designations="U2 snRNP B''|small nuclear ribonucleoprotein polypeptide B2", map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16729964, end_position_on_the_genomic_accession=16741772, orientation='plus', exon_count=7, OMIM=603520, no_hdr0=''),
  6631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6631, CurrentID=0, Status='live', Symbol='SNRPC', Aliases='U1C, Yhc1', description='small nuclear ribonucleoprotein polypeptide C', other_designations='U1 small nuclear RNP specific C|U1 snRNP C|U1 snRNP protein C|U1-C|U1C', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34757094, end_position_on_the_genomic_accession=34773857, orientation='plus', exon_count=6, OMIM=603522, no_hdr0=''),
  6632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6632, CurrentID=0, Status='live', Symbol='SNRPD1', Aliases='HsT2456, SMD1, SNRPD, Sm-D1', description='small nuclear ribonucleoprotein D1 polypeptide', other_designations='Sm-D autoantigen|small nuclear ribonucleoprotein D1 polypeptide 16kDa pseudogene|snRNP core protein D1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21612269, end_position_on_the_genomic_accession=21630247, orientation='plus', exon_count=4, OMIM=601063, no_hdr0=''),
  6633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6633, CurrentID=0, Status='live', Symbol='SNRPD2', Aliases='SMD2, SNRPD1, Sm-D2', description='small nuclear ribonucleoprotein D2 polypeptide', other_designations='small nuclear ribonucleoprotein D2 polypeptide 16.5kDa|snRNP core protein D2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45687454, end_position_on_the_genomic_accession=45692311, orientation='minus', exon_count=4, OMIM=601061, no_hdr0=''),
  6634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6634, CurrentID=0, Status='live', Symbol='SNRPD3', Aliases='SMD3, Sm-D3', description='small nuclear ribonucleoprotein D3 polypeptide', other_designations='small nuclear ribonucleoprotein D3 polypeptide 18kDa|snRNP core protein D3', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24555650, end_position_on_the_genomic_accession=24574964, orientation='plus', exon_count=4, OMIM=601062, no_hdr0=''),
  6635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6635, CurrentID=0, Status='live', Symbol='SNRPE', Aliases='B-raf, HYPT11, SME, Sm-E', description='small nuclear ribonucleoprotein polypeptide E', other_designations='sm protein E|snRNP-E', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203861585, end_position_on_the_genomic_accession=203871152, orientation='plus', exon_count=5, OMIM=128260, no_hdr0=''),
  6636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6636, CurrentID=0, Status='live', Symbol='SNRPF', Aliases='SMF, Sm-F, snRNP-F', description='small nuclear ribonucleoprotein polypeptide F', other_designations='sm protein F', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95858931, end_position_on_the_genomic_accession=95866460, orientation='plus', exon_count=4, OMIM=603541, no_hdr0=''),
  6637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6637, CurrentID=0, Status='live', Symbol='SNRPG', Aliases='SMG, Sm-G', description='small nuclear ribonucleoprotein polypeptide G', other_designations='sm protein G|snRNP-G', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70281378, end_position_on_the_genomic_accession=70293737, orientation='minus', exon_count=4, OMIM=603542, no_hdr0=''),
  6638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6638, CurrentID=0, Status='live', Symbol='SNRPN', Aliases='HCERN3, PWCR, RT-LI, SM-D, SMN, SNRNP-N, SNURF-SNRPN, sm-N', description='small nuclear ribonucleoprotein polypeptide N', other_designations='SM protein N|sm protein D|tissue-specific splicing protein', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24823647, end_position_on_the_genomic_accession=24978582, orientation='plus', exon_count=15, OMIM=182279, no_hdr0=''),
  6640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6640, CurrentID=0, Status='live', Symbol='SNTA1', Aliases='LQT12, SNT1, TACIP1, dJ1187J4.5', description='syntrophin, alpha 1', other_designations='59 kDa dystrophin-associated protein A1 acidic component 1|acidic alpha 1 syntrophin|dystrophin-associated protein A1, 59kDa, acidic component|pro-TGF-alpha cytoplasmic domain-interacting protein 1|syntrophin, alpha 1 (dystrophin-associated protein A1, 59kDa, acidic component)|syntrophin-1', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33407957, end_position_on_the_genomic_accession=33443892, orientation='minus', exon_count=8, OMIM=601017, no_hdr0=''),
  6641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6641, CurrentID=0, Status='live', Symbol='SNTB1', Aliases='59-DAP, A1B, BSYN2, DAPA1B, SNT2, SNT2B1, TIP-43', description='syntrophin, beta 1 (dystrophin-associated protein A1, 59kDa, basic component 1)', other_designations='59 kDa dystrophin-associated protein A1 basic component 1|dystrophin-associated protein A1, 59kD, basic component 1|syntrophin-2|tax interaction protein 43', map_location='8q23-q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120535745, end_position_on_the_genomic_accession=120812069, orientation='minus', exon_count=8, OMIM=600026, no_hdr0=''),
  6642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6642, CurrentID=0, Status='live', Symbol='SNX1', Aliases='HsT17379, VPS5', description='sorting nexin 1', other_designations='sorting nexin 1A', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64094076, end_position_on_the_genomic_accession=64144234, orientation='plus', exon_count=16, OMIM=601272, no_hdr0=''),
  6643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6643, CurrentID=0, Status='live', Symbol='SNX2', Aliases='TRG-9', description='sorting nexin 2', other_designations='CTB-36H16.2|transformation-related gene 9 protein', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122774996, end_position_on_the_genomic_accession=122834539, orientation='plus', exon_count=15, OMIM=605929, no_hdr0=''),
  6644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6644, CurrentID=0, Status='live', Symbol='SRIP1', Aliases='SRIL', description='sorcin pseudogene 1', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=58103147, end_position_on_the_genomic_accession=58103629, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6645, CurrentID=0, Status='live', Symbol='SNTB2', Aliases='D16S2531E, EST25263, SNT2B2, SNT3, SNTL', description='syntrophin, beta 2 (dystrophin-associated protein A1, 59kDa, basic component 2)', other_designations='59 kDa dystrophin-associated protein A1 basic component 2|dystrophin-associated protein A1, 59kD, basic component 2|syntrophin-3', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69187147, end_position_on_the_genomic_accession=69309052, orientation='plus', exon_count=7, OMIM=600027, no_hdr0=''),
  6646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6646, CurrentID=0, Status='live', Symbol='SOAT1', Aliases='ACACT, ACAT, ACAT-1, ACAT1, SOAT, STAT', description='sterol O-acyltransferase 1', other_designations='acyl-coenzyme A:cholesterol acyltransferase 1|sterol O-acyltransferase (acyl-Coenzyme A: cholesterol acyltransferase) 1', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179293714, end_position_on_the_genomic_accession=179358680, orientation='plus', exon_count=17, OMIM=102642, no_hdr0=''),
  6647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6647, CurrentID=0, Status='live', Symbol='SOD1', Aliases='ALS, ALS1, HEL-S-44, IPOA, SOD, hSod1, homodimer', description='superoxide dismutase 1, soluble', other_designations='Cu/Zn superoxide dismutase|SOD, soluble|epididymis secretory protein Li 44|indophenoloxidase A|superoxide dismutase, cystolic', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31659622, end_position_on_the_genomic_accession=31668931, orientation='plus', exon_count=5, OMIM=147450, no_hdr0=''),
  6648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6648, CurrentID=0, Status='live', Symbol='SOD2', Aliases='IPOB, MNSOD, MVCD6', description='superoxide dismutase 2, mitochondrial', other_designations='Mn superoxide dismutase|indophenoloxidase B|manganese-containing superoxide dismutase|mangano-superoxide dismutase', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159679116, end_position_on_the_genomic_accession=159693321, orientation='minus', exon_count=6, OMIM=147460, no_hdr0=''),
  6649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6649, CurrentID=0, Status='live', Symbol='SOD3', Aliases='EC-SOD', description='superoxide dismutase 3, extracellular', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24795463, end_position_on_the_genomic_accession=24800845, orientation='plus', exon_count=2, OMIM=185490, no_hdr0=''),
  6650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6650, CurrentID=0, Status='live', Symbol='CAPN15', Aliases='SOLH', description='calpain 15', other_designations='small optic lobes homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=527756, end_position_on_the_genomic_accession=554636, orientation='plus', exon_count=17, OMIM=603267, no_hdr0=''),
  6651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6651, CurrentID=0, Status='live', Symbol='SON', Aliases='BASS1, C21orf50, DBP-5, NREBP3, SON', description='SON DNA binding protein', other_designations='Bax antagonist selected in Saccharomyces 1|NRE-binding protein|negative regulatory element-binding protein', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33543038, end_position_on_the_genomic_accession=33577514, orientation='plus', exon_count=15, OMIM=182465, no_hdr0=''),
  6652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6652, CurrentID=0, Status='live', Symbol='SORD', Aliases='HEL-S-95n1, SORD', description='sorbitol dehydrogenase', other_designations='L-iditol 2-dehydrogenase|epididymis secretory sperm binding protein Li 95n', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45023104, end_position_on_the_genomic_accession=45075089, orientation='plus', exon_count=9, OMIM=182500, no_hdr0=''),
  6653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6653, CurrentID=0, Status='live', Symbol='SORL1', Aliases='C11orf32, LR11, LRP9, SORLA, SorLA-1, gp250', description='sortilin-related receptor, L(DLR class) A repeats containing', other_designations='LDLR relative with 11 ligand-binding repeats|low-density lipoprotein receptor relative with 11 ligand-binding repeats|mosaic protein LR11|sortilin-related receptor, L(DLR class) A repeats-containing|sorting protein-related receptor containing LDLR class A repeats', map_location='11q23.2-q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121452203, end_position_on_the_genomic_accession=121633762, orientation='plus', exon_count=51, OMIM=602005, no_hdr0=''),
  6654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6654, CurrentID=0, Status='live', Symbol='SOS1', Aliases='GF1, GGF1, GINGF, HGF, NS4', description='SOS Ras/Rac guanine nucleotide exchange factor 1', other_designations='SOS-1|gingival fibromatosis, hereditary, 1|guanine nucleotide exchange factor', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38981549, end_position_on_the_genomic_accession=39124744, orientation='minus', exon_count=28, OMIM=182530, no_hdr0=''),
  6655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6655, CurrentID=0, Status='live', Symbol='SOS2', Aliases='', description='SOS Ras/Rho guanine nucleotide exchange factor 2', other_designations='SOS-2|guanine nucleotide exchange factor|guanine nucleotide releasing factor', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50117128, end_position_on_the_genomic_accession=50231878, orientation='minus', exon_count=25, OMIM=601247, no_hdr0=''),
  6656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6656, CurrentID=0, Status='live', Symbol='SOX1', Aliases='', description='SRY-box 1', other_designations='SRY (sex determining region Y)-box 1|SRY box 1|SRY-related HMG-box gene 1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112067599, end_position_on_the_genomic_accession=112071706, orientation='plus', exon_count=1, OMIM=602148, no_hdr0=''),
  6657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6657, CurrentID=0, Status='live', Symbol='SOX2', Aliases='ANOP3, MCOPS3', description='SRY-box 2', other_designations='SRY (sex determining region Y)-box 2|SRY box 2|SRY-related HMG-box gene 2|sex determining region Y-box 2|transcription factor SOX2', map_location='3q26.3-q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181711924, end_position_on_the_genomic_accession=181714436, orientation='plus', exon_count=1, OMIM=184429, no_hdr0=''),
  6658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6658, CurrentID=0, Status='live', Symbol='SOX3', Aliases='GHDX, MRGH, PHP, PHPX, SOXB', description='SRY-box 3', other_designations='SRY (sex determining region Y)-box 3|SRY box 3', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140502987, end_position_on_the_genomic_accession=140505060, orientation='minus', exon_count=1, OMIM=313430, no_hdr0=''),
  6659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6659, CurrentID=0, Status='live', Symbol='SOX4', Aliases='EVI16', description='SRY-box 4', other_designations='SRY (sex determining region Y)-box 4|SRY box 4|SRY-related HMG-box gene 4|ecotropic viral integration site 16', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21593741, end_position_on_the_genomic_accession=21598619, orientation='plus', exon_count=1, OMIM=184430, no_hdr0=''),
  6660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6660, CurrentID=0, Status='live', Symbol='SOX5', Aliases='L-SOX5, L-SOX5B, L-SOX5F', description='SRY-box 5', other_designations='SRY (sex determining region Y)-box 5|SRY box 5', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=23529495, end_position_on_the_genomic_accession=24562669, orientation='minus', exon_count=29, OMIM=604975, no_hdr0=''),
  6661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6661, CurrentID=0, Status='live', Symbol='SOX5P1', Aliases='SOX29, SOX5P', description='SRY-box 5 pseudogene 1', other_designations='SOX5 pseudogene|SRY (sex determining region Y)-box 5 pseudogene|SRY box 5 pseudogene 1', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87788235, end_position_on_the_genomic_accession=87789684, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6662, CurrentID=0, Status='live', Symbol='SOX9', Aliases='CMD1, CMPD1, SRA1, SRXX2, SRXY10', description='SRY-box 9', other_designations='SRY (sex determining region Y)-box9|SRY (sex-determining region Y)-box 9 protein|SRY box 9|SRY-related HMG-box, gene 9', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72121020, end_position_on_the_genomic_accession=72126420, orientation='plus', exon_count=3, OMIM=608160, no_hdr0=''),
  6663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6663, CurrentID=0, Status='live', Symbol='SOX10', Aliases='DOM, PCWH, WS2E, WS4, WS4C', description='SRY-box 10', other_designations='SRY (sex determining region Y)-box 10|SRY box 10|SRY-related HMG-box gene 10|dominant megacolon, mouse, human homolog of', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37972312, end_position_on_the_genomic_accession=37984532, orientation='minus', exon_count=4, OMIM=602229, no_hdr0=''),
  6664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6664, CurrentID=0, Status='live', Symbol='SOX11', Aliases='MRD27', description='SRY-box 11', other_designations='SRY (sex determining region Y)-box 11|SRY (sex-determining region Y)-box 11|SRY box 11|SRY-related HMG-box gene 11', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5692667, end_position_on_the_genomic_accession=5701385, orientation='plus', exon_count=1, OMIM=600898, no_hdr0=''),
  6665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6665, CurrentID=0, Status='live', Symbol='SOX15', Aliases='SOX20, SOX26, SOX27', description='SRY-box 15', other_designations='SRY (sex determining region Y)-box 15|SRY (sex determining region Y)-box 20|SRY box 15', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7588180, end_position_on_the_genomic_accession=7590170, orientation='minus', exon_count=2, OMIM=601297, no_hdr0=''),
  6666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6666, CurrentID=0, Status='live', Symbol='SOX12', Aliases='SOX22', description='SRY-box 12', other_designations='SOX-22 protein|SRY (sex determining region Y)-box 12|SRY box 12|SRY-related HMG-box gene 22', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=325571, end_position_on_the_genomic_accession=330228, orientation='plus', exon_count=1, OMIM=601947, no_hdr0=''),
  6667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6667, CurrentID=0, Status='live', Symbol='SP1', Aliases='', description='Sp1 transcription factor', other_designations='specificity protein 1', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53380195, end_position_on_the_genomic_accession=53416446, orientation='plus', exon_count=8, OMIM=189906, no_hdr0=''),
  6668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6668, CurrentID=0, Status='live', Symbol='SP2', Aliases='', description='Sp2 transcription factor', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47896054, end_position_on_the_genomic_accession=47932271, orientation='plus', exon_count=14, OMIM=601801, no_hdr0=''),
  6670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6670, CurrentID=0, Status='live', Symbol='SP3', Aliases='SPR2', description='Sp3 transcription factor', other_designations='GC-binding transcription factor Sp3|specificity protein 3', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173906459, end_position_on_the_genomic_accession=173965702, orientation='minus', exon_count=7, OMIM=601804, no_hdr0=''),
  6671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6671, CurrentID=0, Status='live', Symbol='SP4', Aliases='HF1B, SPR-1', description='Sp4 transcription factor', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21428071, end_position_on_the_genomic_accession=21514822, orientation='plus', exon_count=12, OMIM=600540, no_hdr0=''),
  6672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6672, CurrentID=0, Status='live', Symbol='SP100', Aliases='lysp100b', description='SP100 nuclear antigen', other_designations='SP100-HMG nuclear autoantigen|nuclear dot-associated Sp100 protein|speckled 100 kDa', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230416156, end_position_on_the_genomic_accession=230545602, orientation='plus', exon_count=32, OMIM=604585, no_hdr0=''),
  100271192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271192, CurrentID=0, Status='live', Symbol='RPL30P15', Aliases='RPL30_7_1765', description='ribosomal protein L30 pseudogene 15', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13574641, end_position_on_the_genomic_accession=13575071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6674, CurrentID=0, Status='live', Symbol='SPAG1', Aliases='CILD28, CT140, HEL-S-268, HSD-3.8, SP75, TPIS', description='sperm associated antigen 1', other_designations='TPR-containing protein involved in spermatogenesis|epididymis secretory protein Li 268|infertility-related sperm protein Spag-1|tetratricopeptide repeat-containing protein', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100158003, end_position_on_the_genomic_accession=100241904, orientation='plus', exon_count=21, OMIM=603395, no_hdr0=''),
  6675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6675, CurrentID=0, Status='live', Symbol='UAP1', Aliases='AGX, AGX1, AGX2, SPAG2', description='UDP-N-acetylglucosamine pyrophosphorylase 1', other_designations='antigen X|sperm associated antigen 2|sperm-associated antigen 2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162561506, end_position_on_the_genomic_accession=162599843, orientation='plus', exon_count=13, OMIM=602862, no_hdr0=''),
  6676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6676, CurrentID=0, Status='live', Symbol='SPAG4', Aliases='CT127, SUN4', description='sperm associated antigen 4', other_designations='SUN domain-containing protein 4|Sad1 and UNC84 domain containing 4|acrosomal protein ACR55|cancer/testis antigen 127|outer dense fiber-associated protein SPAG4|sperm tail protein', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35615853, end_position_on_the_genomic_accession=35621045, orientation='plus', exon_count=15, OMIM=603038, no_hdr0=''),
  6677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6677, CurrentID=0, Status='live', Symbol='SPAM1', Aliases='HEL-S-96n, HYA1, HYAL1, HYAL3, HYAL5, PH-20, PH20, SPAG15', description='sperm adhesion molecule 1 (PH-20 hyaluronidase, zona pellucida binding)', other_designations='epididymis secretory sperm binding protein Li 96n|hyal-PH20|hyaluronoglucosaminidase PH-20|sperm surface protein PH-20', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123925232, end_position_on_the_genomic_accession=123971410, orientation='plus', exon_count=10, OMIM=600930, no_hdr0=''),
  6678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6678, CurrentID=0, Status='live', Symbol='SPARC', Aliases='BM-40, ON', description='secreted protein, acidic, cysteine-rich (osteonectin)', other_designations='basement-membrane protein 40', map_location='5q31.3-q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151661096, end_position_on_the_genomic_accession=151687054, orientation='minus', exon_count=10, OMIM=182120, no_hdr0=''),
  100271193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271193, CurrentID=0, Status='live', Symbol='RPL31P34', Aliases='RPL31_10_781', description='ribosomal protein L31 pseudogene 34', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6490741, end_position_on_the_genomic_accession=6491101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6682, CurrentID=0, Status='live', Symbol='SPG3B', Aliases='', description='spastic paraplegia 3B', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6683, CurrentID=0, Status='live', Symbol='SPAST', Aliases='ADPSP, FSP2, SPG4', description='spastin', other_designations='spastic paraplegia 4 (autosomal dominant; spastin)|spastic paraplegia 4 protein', map_location='2p24-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32063552, end_position_on_the_genomic_accession=32157637, orientation='plus', exon_count=17, OMIM=604277, no_hdr0=''),
  399900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399900, CurrentID=0, Status='live', Symbol='LOC399900', Aliases='', description='uncharacterized LOC399900', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61967738, end_position_on_the_genomic_accession=61969275, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  6685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6685, CurrentID=0, Status='live', Symbol='SPG5B', Aliases='', description='spastic paraplegia 5B (autosomal recessive)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600146, no_hdr0=''),
  6687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6687, CurrentID=0, Status='live', Symbol='SPG7', Aliases='CAR, CMAR, PGN, SPG5C', description='spastic paraplegia 7 (pure and complicated autosomal recessive)', other_designations='cell matrix adhesion regulator|paraplegin|spastic paraplegia 7 protein', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89508381, end_position_on_the_genomic_accession=89557768, orientation='plus', exon_count=21, OMIM=602783, no_hdr0=''),
  6688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6688, CurrentID=0, Status='live', Symbol='SPI1', Aliases='OF, PU.1, SFPI1, SPI-1, SPI-A', description='Spi-1 proto-oncogene', other_designations='31 kDa transforming protein|31 kDa-transforming protein|hematopoietic transcription factor PU.1|spleen focus forming virus (SFFV) proviral integration oncogene spi1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47354858, end_position_on_the_genomic_accession=47395640, orientation='minus', exon_count=10, OMIM=165170, no_hdr0=''),
  6689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6689, CurrentID=0, Status='live', Symbol='SPIB', Aliases='SPI-B', description='Spi-B transcription factor (Spi-1/PU.1 related)', other_designations='', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50418938, end_position_on_the_genomic_accession=50431053, orientation='plus', exon_count=6, OMIM=606802, no_hdr0=''),
  6690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6690, CurrentID=0, Status='live', Symbol='SPINK1', Aliases='PCTT, PSTI, Spink3, TATI, TCP', description='serine peptidase inhibitor, Kazal type 1', other_designations='pancreatic secretory trypsin inhibitor|serine protease inhibitor, Kazal type 1|tumor-associated trypsin inhibitor', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147824580, end_position_on_the_genomic_accession=147839231, orientation='minus', exon_count=5, OMIM=167790, no_hdr0=''),
  6691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6691, CurrentID=0, Status='live', Symbol='SPINK2', Aliases='HUSI-II', description='serine peptidase inhibitor, Kazal type 2 (acrosin-trypsin inhibitor)', other_designations='epididymis tissue protein Li 172', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56809860, end_position_on_the_genomic_accession=56822176, orientation='minus', exon_count=5, OMIM=605753, no_hdr0=''),
  6692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6692, CurrentID=0, Status='live', Symbol='SPINT1', Aliases='HAI, HAI1, MANSC2', description='serine peptidase inhibitor, Kunitz type 1', other_designations='HAI-1|hepatocyte growth factor activator inhibitor type 1|serine protease inhibitor, Kunitz type 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40844018, end_position_on_the_genomic_accession=40857655, orientation='plus', exon_count=11, OMIM=605123, no_hdr0=''),
  6693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6693, CurrentID=0, Status='live', Symbol='SPN', Aliases='CD43, GALGP, GPL115, LSN', description='sialophorin', other_designations='galactoglycoprotein|leukocyte sialoglycoprotein|sialophorin (gpL115, leukosialin, CD43)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29662950, end_position_on_the_genomic_accession=29670507, orientation='plus', exon_count=3, OMIM=182160, no_hdr0=''),
  6694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6694, CurrentID=0, Status='live', Symbol='SPP2', Aliases='SPP-244, SPP2', description='secreted phosphoprotein 2', other_designations='secreted phosphoprotein 2, 24kD|secreted phosphoprotein 2, 24kDa', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234050690, end_position_on_the_genomic_accession=234077134, orientation='plus', exon_count=11, OMIM=602637, no_hdr0=''),
  6695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6695, CurrentID=0, Status='live', Symbol='SPOCK1', Aliases='SPOCK, TESTICAN, TIC1', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan (testican) 1', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136975298, end_position_on_the_genomic_accession=137499329, orientation='minus', exon_count=11, OMIM=602264, no_hdr0=''),
  6696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6696, CurrentID=0, Status='live', Symbol='SPP1', Aliases='BNSP, BSPI, ETA-1, OPN', description='secreted phosphoprotein 1', other_designations='SPP1/CALPHA1 fusion|early T-lymphocyte activation 1|nephropontin|osteopontin/immunoglobulin alpha 1 heavy chain constant region fusion protein|secreted phosphoprotein 1 (osteopontin, bone sialoprotein I, early T-lymphocyte activation 1)|secreted phosphoprotein 1 variant 6|urinary stone protein|uropontin', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87975650, end_position_on_the_genomic_accession=87983411, orientation='plus', exon_count=8, OMIM=166490, no_hdr0=''),
  6697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6697, CurrentID=0, Status='live', Symbol='SPR', Aliases='SDR38C1', description='sepiapterin reductase (7,8-dihydrobiopterin:NADP+ oxidoreductase)', other_designations='short chain dehydrogenase/reductase family 38C, member 1', map_location='2p14-p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72887383, end_position_on_the_genomic_accession=72892160, orientation='plus', exon_count=3, OMIM=182125, no_hdr0=''),
  6698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6698, CurrentID=0, Status='live', Symbol='SPRR1A', Aliases='SPRK', description='small proline-rich protein 1A', other_designations='19 kDa pancornulin|SPR-IA|cornifin A|small proline-rich protein IA', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152984088, end_position_on_the_genomic_accession=152985814, orientation='plus', exon_count=2, OMIM=182265, no_hdr0=''),
  6699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6699, CurrentID=0, Status='live', Symbol='SPRR1B', Aliases='CORNIFIN, GADD33, SPR-IB, SPRR1', description='small proline-rich protein 1B', other_designations='14.9 kDa pancornulin|cornifin B|small proline-rich protein IB', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153031203, end_position_on_the_genomic_accession=153032900, orientation='plus', exon_count=2, OMIM=182266, no_hdr0=''),
  6700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6700, CurrentID=0, Status='live', Symbol='SPRR2A', Aliases='', description='small proline-rich protein 2A', other_designations='2-1|SPR-2A', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153056120, end_position_on_the_genomic_accession=153057512, orientation='minus', exon_count=2, OMIM=182267, no_hdr0=''),
  6701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6701, CurrentID=0, Status='live', Symbol='SPRR2B', Aliases='', description='small proline-rich protein 2B', other_designations='SPR-2B', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153070228, end_position_on_the_genomic_accession=153071608, orientation='minus', exon_count=2, OMIM=182268, no_hdr0=''),
  6702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6702, CurrentID=0, Status='live', Symbol='SPRR2C', Aliases='', description='small proline-rich protein 2C (pseudogene)', other_designations='small proline-rich protein 2C', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153140118, end_position_on_the_genomic_accession=153141493, orientation='minus', exon_count=2, OMIM=182269, no_hdr0=''),
  6703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6703, CurrentID=0, Status='live', Symbol='SPRR2D', Aliases='', description='small proline-rich protein 2D', other_designations='SPR-2D|SPR-II|small proline-rich protein II', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153039725, end_position_on_the_genomic_accession=153041118, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6704, CurrentID=0, Status='live', Symbol='SPRR2E', Aliases='', description='small proline-rich protein 2E', other_designations='SPR-2E|SPR-II|small proline-rich protein II', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153093135, end_position_on_the_genomic_accession=153094528, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6705, CurrentID=0, Status='live', Symbol='SPRR2F', Aliases='', description='small proline-rich protein 2F', other_designations='SPR-2F', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153112124, end_position_on_the_genomic_accession=153113515, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6706, CurrentID=0, Status='live', Symbol='SPRR2G', Aliases='', description='small proline-rich protein 2G', other_designations='SPR-2G', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153149582, end_position_on_the_genomic_accession=153150951, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6707, CurrentID=0, Status='live', Symbol='SPRR3', Aliases='', description='small proline-rich protein 3', other_designations='22 kDa pancornulin|cornifin beta|esophagin', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153001747, end_position_on_the_genomic_accession=153003856, orientation='plus', exon_count=3, OMIM=182271, no_hdr0=''),
  6708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6708, CurrentID=0, Status='live', Symbol='SPTA1', Aliases='EL2, HPP, HS3, SPH3, SPTA', description='spectrin, alpha, erythrocytic 1', other_designations='alpha-I spectrin|elliptocytosis 2|erythroid alpha-spectrin|spectrin alpha chain, erythrocyte|spectrin, alpha, erythrocytic 1 (elliptocytosis 2)', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158610498, end_position_on_the_genomic_accession=158686716, orientation='minus', exon_count=53, OMIM=182860, no_hdr0=''),
  6709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6709, CurrentID=0, Status='live', Symbol='SPTAN1', Aliases='EIEE5, NEAS, SPTA2', description='spectrin, alpha, non-erythrocytic 1', other_designations='alpha-II spectrin|alpha-fodrin|fodrin alpha chain|spectrin, non-erythroid alpha chain|spectrin, non-erythroid alpha subunit', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128552558, end_position_on_the_genomic_accession=128633665, orientation='plus', exon_count=59, OMIM=182810, no_hdr0=''),
  6710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6710, CurrentID=0, Status='live', Symbol='SPTB', Aliases='EL3, HS2, HSPTB1, SPH2', description='spectrin, beta, erythrocytic', other_designations='Sp beta|beta-I spectrin|membrane cytoskeletal protein|spectrin beta Tandil|spectrin beta chain, erythrocyte', map_location='14q23-q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64746283, end_position_on_the_genomic_accession=64879893, orientation='minus', exon_count=37, OMIM=182870, no_hdr0=''),
  6711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6711, CurrentID=0, Status='live', Symbol='SPTBN1', Aliases='ELF, HEL102, SPTB2, betaSpII', description='spectrin, beta, non-erythrocytic 1', other_designations='beta-G spectrin|beta-II spectrin|beta-fodrin|beta-spectrin 2|beta-spectrin II|beta-spectrin non-erythrocytic 1|embryonic liver beta-fodrin|epididymis luminal protein 102|fodrin beta chain|spectrin beta chain, brain 1|spectrin, non-erythroid beta chain 1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54456317, end_position_on_the_genomic_accession=54671446, orientation='plus', exon_count=38, OMIM=182790, no_hdr0=''),
  6712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6712, CurrentID=0, Status='live', Symbol='SPTBN2', Aliases='GTRAP41, SCA5, SCAR14', description='spectrin, beta, non-erythrocytic 2', other_designations='beta-III spectrin|glutamate transporter EAAT4-associated protein 41|spectrin beta III sigma 2|spectrin beta chain, brain 2|spectrin, non-erythroid beta chain 2|spinocerebellar ataxia 5 protein', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66682496, end_position_on_the_genomic_accession=66729237, orientation='minus', exon_count=42, OMIM=604985, no_hdr0=''),
  6713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6713, CurrentID=0, Status='live', Symbol='SQLE', Aliases='', description='squalene epoxidase', other_designations='SE', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124998478, end_position_on_the_genomic_accession=125022284, orientation='plus', exon_count=12, OMIM=602019, no_hdr0=''),
  6714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6714, CurrentID=0, Status='live', Symbol='SRC', Aliases='ASV1, c-SRC, p60-Src, SRC', description='SRC proto-oncogene, non-receptor tyrosine kinase', other_designations='proto-oncogene c-Src|protooncogene SRC, Rous sarcoma|tyrosine kinase pp60c-src|tyrosine-protein kinase SRC-1|v-src avian sarcoma (Schmidt-Ruppin A-2) viral oncogene homolog', map_location='20q12-q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37344685, end_position_on_the_genomic_accession=37405432, orientation='plus', exon_count=16, OMIM=190090, no_hdr0=''),
  6715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6715, CurrentID=0, Status='live', Symbol='SRD5A1', Aliases='S5AR 1', description='steroid-5-alpha-reductase, alpha polypeptide 1 (3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha 1)', other_designations='3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha 1|5-alpha reductase|SR type 1|steroid 5-alpha-reductase 1|steroid 5-alpha-reductase type I', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6633360, end_position_on_the_genomic_accession=6669562, orientation='plus', exon_count=5, OMIM=184753, no_hdr0=''),
  6716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6716, CurrentID=0, Status='live', Symbol='SRD5A2', Aliases='', description='steroid-5-alpha-reductase, alpha polypeptide 2 (3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha 2)', other_designations='5 alpha-SR2|S5AR 2|SR type 2|steroid 5-alpha-reductase 2|type II 5-alpha reductase', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31522488, end_position_on_the_genomic_accession=31700784, orientation='minus', exon_count=13, OMIM=607306, no_hdr0=''),
  6717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6717, CurrentID=0, Status='live', Symbol='SRI', Aliases='CP-22, CP22, SCN, V19', description='sorcin', other_designations='22 kDa protein|H_RG167B05.1|calcium binding protein amplified in mutlidrug-resistant cells', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88205115, end_position_on_the_genomic_accession=88226993, orientation='minus', exon_count=9, OMIM=182520, no_hdr0=''),
  6718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6718, CurrentID=0, Status='live', Symbol='AKR1D1', Aliases='3o5bred, CBAS2, SRD5B1', description='aldo-keto reductase family 1, member D1', other_designations='delta 4-3-ketosteroid-5-beta-reductase|delta(4)-3-ketosteroid 5-beta-reductase|delta(4)-3-oxosteroid 5-beta-reductase|steroid-5-beta-reductase, beta polypeptide 1 (3-oxo-5 beta-steroid delta 4-dehydrogenase beta 1)', map_location='7q32-q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138076432, end_position_on_the_genomic_accession=138118304, orientation='plus', exon_count=9, OMIM=604741, no_hdr0=''),
  6719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6719, CurrentID=0, Status='live', Symbol='SRD5A1P1', Aliases='SRD5AP1', description='steroid-5-alpha-reductase, alpha polypeptide 1 pseudogene 1 (3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha pseudogene)', other_designations='steroid-5-alpha-reductase 1 pseudogene|steroid-5-alpha-reductase, alpha polypeptide 1 (3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha 1) pseudogene|steroid-5-alpha-reductase, alpha polypeptide pseudogene 1 (3-oxo-5 alpha-steroid delta 4-dehydrogenase alpha pseudogene)', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139446819, end_position_on_the_genomic_accession=139448973, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  6720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6720, CurrentID=0, Status='live', Symbol='SREBF1', Aliases='SREBP-1c, SREBP1, bHLHd1', description='sterol regulatory element binding transcription factor 1', other_designations='SREBP-1|class D basic helix-loop-helix protein 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17811349, end_position_on_the_genomic_accession=17837017, orientation='minus', exon_count=21, OMIM=184756, no_hdr0=''),
  6721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6721, CurrentID=0, Status='live', Symbol='SREBF2', Aliases='SREBP-2, SREBP2, bHLHd2', description='sterol regulatory element binding transcription factor 2', other_designations='class D basic helix-loop-helix protein 2|sterol regulatory element-binding transcription factor 2', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41833079, end_position_on_the_genomic_accession=41907308, orientation='plus', exon_count=23, OMIM=600481, no_hdr0=''),
  6722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6722, CurrentID=0, Status='live', Symbol='SRF', Aliases='MCM1', description='serum response factor', other_designations='c-fos serum response element-binding transcription factor|serum response factor (c-fos serum response element-binding transcription factor)', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43171295, end_position_on_the_genomic_accession=43181506, orientation='plus', exon_count=8, OMIM=600589, no_hdr0=''),
  6723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6723, CurrentID=0, Status='live', Symbol='SRM', Aliases='PAPT, SPDSY, SPS1L1, SRM', description='spermidine synthase', other_designations='putrescine aminopropyltransferase|spermidine synthase-1', map_location='1p36-p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11054592, end_position_on_the_genomic_accession=11060034, orientation='minus', exon_count=8, OMIM=182891, no_hdr0=''),
  399940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399940, CurrentID=0, Status='live', Symbol='TRIM51EP', Aliases='', description='tripartite motif-containing 51E, pseudogene', other_designations='SPRY domain containing 5 pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89981446, end_position_on_the_genomic_accession=89991691, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  6725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6725, CurrentID=0, Status='live', Symbol='SRMS', Aliases='C20orf148, SRM, dJ697K14.1', description='src-related kinase lacking C-terminal regulatory tyrosine and N-terminal myristylation sites', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63539924, end_position_on_the_genomic_accession=63547504, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  6726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6726, CurrentID=0, Status='live', Symbol='SRP9', Aliases='ALURBP', description='signal recognition particle 9kDa', other_designations='signal recognition particle 9kD', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225777813, end_position_on_the_genomic_accession=225790466, orientation='plus', exon_count=4, OMIM=600707, no_hdr0=''),
  6727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6727, CurrentID=0, Status='live', Symbol='SRP14', Aliases='ALURBP', description='signal recognition particle 14kDa', other_designations='18 kDa Alu RNA-binding protein|homologous Alu RNA binding protein|signal recognition particle 14kDa (homologous Alu RNA binding protein)', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40035690, end_position_on_the_genomic_accession=40039188, orientation='minus', exon_count=6, OMIM=600708, no_hdr0=''),
  6728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6728, CurrentID=0, Status='live', Symbol='SRP19', Aliases='', description='signal recognition particle 19kDa', other_designations='signal recognition particle 19kD', map_location='5q21-q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112861188, end_position_on_the_genomic_accession=112893079, orientation='plus', exon_count=6, OMIM=182175, no_hdr0=''),
  6729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6729, CurrentID=0, Status='live', Symbol='SRP54', Aliases='', description='signal recognition particle 54kDa', other_designations='signal recognition particle 54kD', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34982695, end_position_on_the_genomic_accession=35029567, orientation='plus', exon_count=17, OMIM=604857, no_hdr0=''),
  6730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6730, CurrentID=0, Status='live', Symbol='SRP68', Aliases='', description='signal recognition particle 68kDa', other_designations='signal recognition particle 68 kDa protein|signal recognition particle 68kD', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76038775, end_position_on_the_genomic_accession=76072526, orientation='minus', exon_count=17, OMIM=604858, no_hdr0=''),
  6731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6731, CurrentID=0, Status='live', Symbol='SRP72', Aliases='BMFF, BMFS1, HEL103', description='signal recognition particle 72kDa', other_designations='epididymis luminal protein 103|signal recognition particle 72 kDa protein|signal recognition particle 72kD', map_location='4q11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56467596, end_position_on_the_genomic_accession=56503681, orientation='plus', exon_count=20, OMIM=602122, no_hdr0=''),
  6732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6732, CurrentID=0, Status='live', Symbol='SRPK1', Aliases='SFRSK1', description='SRSF protein kinase 1', other_designations='SFRS protein kinase 1|SR-protein-specific kinase 1|serine/arginine-rich splicing factor kinase 1|serine/threonine-protein kinase SRPK1', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35833034, end_position_on_the_genomic_accession=35921180, orientation='minus', exon_count=16, OMIM=601939, no_hdr0=''),
  6733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6733, CurrentID=0, Status='live', Symbol='SRPK2', Aliases='SFRSK2', description='SRSF protein kinase 2', other_designations='SFRS protein kinase 2|SR protein kinase 2|SR-protein-specific kinase 2|serine kinase SRPK2|serine/arginine-rich protein-specific kinase 2|serine/arginine-rich splicing factor kinase 2|serine/threonine-protein kinase SRPK2', map_location='7q22-q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105113601, end_position_on_the_genomic_accession=105399366, orientation='minus', exon_count=25, OMIM=602980, no_hdr0=''),
  6734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6734, CurrentID=0, Status='live', Symbol='SRPRA', Aliases='DP, SRPR, Sralpha', description='SRP receptor alpha subunit', other_designations='DP-alpha|SR-alpha|SRP-alpha|docking protein alpha|signal recognition particle receptor (docking protein)', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126262919, end_position_on_the_genomic_accession=126268982, orientation='minus', exon_count=14, OMIM=182180, no_hdr0=''),
  6736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6736, CurrentID=0, Status='live', Symbol='SRY', Aliases='SRXX1, SRXY1, TDF, TDY', description='sex determining region Y', other_designations='essential protein for sex determination in human males|sex determining region Y protein|sex-determining region on Y|testis-determining factor on Y', map_location='Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2786855, end_position_on_the_genomic_accession=2787741, orientation='minus', exon_count=1, OMIM=480000, no_hdr0=''),
  6737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6737, CurrentID=0, Status='live', Symbol='TRIM21', Aliases='RNF81, RO52, Ro/SSA, SSA, SSA1', description='tripartite motif containing 21', other_designations='52 kDa Ro protein|52 kDa ribonucleoprotein autoantigen Ro/SS-A|RING finger protein 81|SS-A|Sicca syndrome antigen A|Sjogren syndrome antigen A1 (52kDa, ribonucleoprotein autoantigen SS-A/Ro)|ro(SS-A)|sjoegren syndrome type A antigen|tripartite motif-containing 21|tripartite motif-containing protein 21', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4384897, end_position_on_the_genomic_accession=4393712, orientation='minus', exon_count=7, OMIM=109092, no_hdr0=''),
  6738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6738, CurrentID=0, Status='live', Symbol='TROVE2', Aliases='RO60, RORNP, SSA2', description='TROVE domain family member 2', other_designations='60 kDa ribonucleoprotein Ro|Sjogren syndrome antigen A2 (60kD, ribonucleoprotein autoantigen SS-A/Ro)|TROVE domain family, member 2|gastric cancer multi-drug resistance protein|sjoegren syndrome type A antigen', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193059422, end_position_on_the_genomic_accession=193091777, orientation='plus', exon_count=14, OMIM=600063, no_hdr0=''),
  6739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6739, CurrentID=0, Status='live', Symbol='SSAV1', Aliases='GLV', description='simian sarcoma-associated virus 1/gibbon ape leukemia virus-related endogenous retroviral element 1', other_designations='Simian sarcoma-associated virus-1/Gibbon ape leukemia virus', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=182090, no_hdr0=''),
  6741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6741, CurrentID=0, Status='live', Symbol='SSB', Aliases='LARP3, La, La/SSB', description='Sjogren syndrome antigen B', other_designations='La ribonucleoprotein domain family, member 3|SS-B|SS-B/La protein|Sjogren syndrome antigen B (autoantigen La)|autoantigen La|la autoantigen|lupus La antigen|sjoegren syndrome type B antigen', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169798812, end_position_on_the_genomic_accession=169812064, orientation='plus', exon_count=13, OMIM=109090, no_hdr0=''),
  6742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6742, CurrentID=0, Status='live', Symbol='SSBP1', Aliases='Mt-SSB, SOSS-B1, SSBP, mtSSB', description='single-stranded DNA binding protein 1, mitochondrial', other_designations='PWP1-interacting protein 17', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141738321, end_position_on_the_genomic_accession=141750488, orientation='plus', exon_count=9, OMIM=600439, no_hdr0=''),
  399959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399959, CurrentID=0, Status='live', Symbol='MIR100HG', Aliases='AGD1, lncRNA-N2', description='mir-100-let-7a-2 cluster host gene', other_designations='adipogenesis down-regulated transcript 1|lncRNA neuronal 2|mir-100-let-7a-2 cluster host gene (non-protein coding)', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122089103, end_position_on_the_genomic_accession=122203062, orientation='minus', exon_count=4, OMIM=615965, no_hdr0=''),
  6744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6744, CurrentID=0, Status='live', Symbol='SSFA2', Aliases='CS-1, CS1, KRAP, SPAG13', description='sperm specific antigen 2', other_designations='KRAS-induced actin-interacting protein|cleavage signal-1 protein|ki-ras-induced actin-interacting protein|sperm associated antigen 13', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181891716, end_position_on_the_genomic_accession=181930738, orientation='plus', exon_count=23, OMIM=118990, no_hdr0=''),
  6745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6745, CurrentID=0, Status='live', Symbol='SSR1', Aliases='TRAPA', description='signal sequence receptor, alpha', other_designations='SSR alpha subunit|SSR-alpha|TRAP alpha|TRAP-alpha|signal sequence receptor subunit alpha|translocon-associated protein alpha subunit', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7281143, end_position_on_the_genomic_accession=7313314, orientation='minus', exon_count=8, OMIM=600868, no_hdr0=''),
  6746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6746, CurrentID=0, Status='live', Symbol='SSR2', Aliases='HSD25, TLAP, TRAP-BETA, TRAPB', description='signal sequence receptor, beta (translocon-associated protein beta)', other_designations='SSR-beta|signal sequence receptor subunit beta|translocon-associated protein beta', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156009048, end_position_on_the_genomic_accession=156020991, orientation='minus', exon_count=7, OMIM=600867, no_hdr0=''),
  6747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6747, CurrentID=0, Status='live', Symbol='SSR3', Aliases='TRAPG', description='signal sequence receptor, gamma (translocon-associated protein gamma)', other_designations='SSR gamma|SSR-gamma|TRAP-complex gamma subunit|TRAP-gamma|signal sequence receptor subunit gamma|translocon-associated protein gamma subunit', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156540140, end_position_on_the_genomic_accession=156555184, orientation='minus', exon_count=5, OMIM=606213, no_hdr0=''),
  6748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6748, CurrentID=0, Status='live', Symbol='SSR4', Aliases='CDG1Y, TRAPD', description='signal sequence receptor, delta', other_designations='SSR-delta|TRAP-delta', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153794175, end_position_on_the_genomic_accession=153798512, orientation='plus', exon_count=11, OMIM=300090, no_hdr0=''),
  6749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6749, CurrentID=0, Status='live', Symbol='SSRP1', Aliases='FACT, FACT80, T160', description='structure specific recognition protein 1', other_designations='FACT 80 kDa subunit|FACTp80|chromatin-specific transcription elongation factor 80 kDa subunit|cisplatin-DNA SSRP|facilitates chromatin remodeling 80 kDa subunit|facilitates chromatin transcription complex 80 kDa subunit|facilitates chromatin transcription complex subunit SSRP1|hSSRP1|high mobility group box|recombination signal sequence recognition protein 1|structure-specific recognition protein 1', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57325985, end_position_on_the_genomic_accession=57335877, orientation='minus', exon_count=18, OMIM=604328, no_hdr0=''),
  6750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6750, CurrentID=0, Status='live', Symbol='SST', Aliases='SMST', description='somatostatin', other_designations='growth hormone release-inhibiting factor|prepro-somatostatin|somatostatin-14|somatostatin-28', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187668906, end_position_on_the_genomic_accession=187670413, orientation='minus', exon_count=2, OMIM=182450, no_hdr0=''),
  6751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6751, CurrentID=0, Status='live', Symbol='SSTR1', Aliases='SRIF-2, SS-1-R, SS1-R, SS1R', description='somatostatin receptor 1', other_designations='', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38207999, end_position_on_the_genomic_accession=38213063, orientation='plus', exon_count=3, OMIM=182451, no_hdr0=''),
  6752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6752, CurrentID=0, Status='live', Symbol='SSTR2', Aliases='', description='somatostatin receptor 2', other_designations='SRIF-1|SS2R', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73165021, end_position_on_the_genomic_accession=73171955, orientation='plus', exon_count=2, OMIM=182452, no_hdr0=''),
  6753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6753, CurrentID=0, Status='live', Symbol='SSTR3', Aliases='SS-3-R, SS3-R, SS3R, SSR-28', description='somatostatin receptor 3', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37204237, end_position_on_the_genomic_accession=37220447, orientation='minus', exon_count=8, OMIM=182453, no_hdr0=''),
  6754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6754, CurrentID=0, Status='live', Symbol='SSTR4', Aliases='SS-4-R, SS4-R, SS4R', description='somatostatin receptor 4', other_designations='G-protein coupled receptor', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23035420, end_position_on_the_genomic_accession=23036677, orientation='plus', exon_count=1, OMIM=182454, no_hdr0=''),
  6755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6755, CurrentID=0, Status='live', Symbol='SSTR5', Aliases='SS-5-R', description='somatostatin receptor 5', other_designations='somatostatin receptor subtype 5', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1072756, end_position_on_the_genomic_accession=1081454, orientation='plus', exon_count=2, OMIM=182455, no_hdr0=''),
  6756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6756, CurrentID=0, Status='live', Symbol='SSX1', Aliases='CT5.1, SSRC', description='synovial sarcoma, X breakpoint 1', other_designations='cancer/testis antigen 5.1|cancer/testis antigen family 5, member 1|sarcoma, synovial, X-chromosome-related 1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48255317, end_position_on_the_genomic_accession=48267444, orientation='plus', exon_count=9, OMIM=312820, no_hdr0=''),
  6757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6757, CurrentID=0, Status='live', Symbol='SSX2', Aliases='CT5.2, CT5.2A, HD21, HOM-MEL-40, SSX', description='synovial sarcoma, X breakpoint 2', other_designations='cancer/testis antigen 5.2|cancer/testis antigen family 5, member 2a|sarcoma, synovial, X-chromosome-related 2|synovial sarcoma, X breakpoint 2B|tumor antigen HOM-MEL-40', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52696896, end_position_on_the_genomic_accession=52707227, orientation='minus', exon_count=9, OMIM=300192, no_hdr0=''),
  6758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6758, CurrentID=0, Status='live', Symbol='SSX5', Aliases='', description='synovial sarcoma, X breakpoint 5', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48186220, end_position_on_the_genomic_accession=48196798, orientation='minus', exon_count=10, OMIM=300327, no_hdr0=''),
  6759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6759, CurrentID=0, Status='live', Symbol='SSX4', Aliases='CT5.4', description='synovial sarcoma, X breakpoint 4', other_designations='cancer/testis antigen 5.4', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48383527, end_position_on_the_genomic_accession=48393343, orientation='plus', exon_count=8, OMIM=300326, no_hdr0=''),
  6760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6760, CurrentID=0, Status='live', Symbol='SS18', Aliases='SSXT, SYT', description='synovial sarcoma translocation, chromosome 18', other_designations='synovial sarcoma translocated to X chromosome protein|synovial sarcoma, translocated to X chromosome', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26016253, end_position_on_the_genomic_accession=26091212, orientation='minus', exon_count=18, OMIM=600192, no_hdr0=''),
  6761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6761, CurrentID=0, Status='live', Symbol='ST2', Aliases='', description='suppression of tumorigenicity 2', other_designations='', map_location='11p14.3-p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=185440, no_hdr0=''),
  6762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6762, CurrentID=0, Status='live', Symbol='ST3', Aliases='CCTS, TSHL', description='suppression of tumorigenicity 3', other_designations='tumor-suppressor, HELA cell type', map_location='11q13-q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=191181, no_hdr0=''),
  137835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137835, CurrentID=0, Status='live', Symbol='TMEM71', Aliases='', description='transmembrane protein 71', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132709945, end_position_on_the_genomic_accession=132760712, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  6764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6764, CurrentID=0, Status='live', Symbol='ST5', Aliases='DENND2B, HTS1, p126', description='suppression of tumorigenicity 5', other_designations='DENN/MADD domain containing 2B|heLa tumor suppression 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8693352, end_position_on_the_genomic_accession=8910951, orientation='minus', exon_count=38, OMIM=140750, no_hdr0=''),
  6765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6765, CurrentID=0, Status='live', Symbol='ST8', Aliases='OVC, OVCS', description='suppression of tumorigenicity 8 (ovarian)', other_designations='Oncogene OVC (ovarian adenocarcinoma oncogene)|ovarian cancer, familial', map_location='6q25-q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=167000, no_hdr0=''),
  6767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6767, CurrentID=0, Status='live', Symbol='ST13', Aliases='AAG2, FAM10A1, FAM10A4, HIP, HOP, HSPABP, HSPABP1, P48, PRO0786, SNC6', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein)', other_designations='Hsp70-interacting protein|aging-associated protein 2|heat shock 70kD protein binding protein|progesterone receptor-associated p48 protein|putative tumor suppressor ST13|renal carcinoma antigen NY-REN-33|suppression of tumorigenicity 13 protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40824535, end_position_on_the_genomic_accession=40857008, orientation='minus', exon_count=12, OMIM=606796, no_hdr0=''),
  6768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6768, CurrentID=0, Status='live', Symbol='ST14', Aliases='ARCI11, HAI, MT-SP1, MTSP1, PRSS14, SNC19, TADG15, TMPRSS14', description='suppression of tumorigenicity 14', other_designations='membrane-type serine protease 1|prostamin|serine protease 14|serine protease TADG-15|suppression of tumorigenicity 14 (colon carcinoma)|suppression of tumorigenicity 14 (colon carcinoma, matriptase, epithin)|tumor associated differentially expressed gene 15 protein|tumor-associated differentially-expressed gene 15 protein', map_location='11q24-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130159787, end_position_on_the_genomic_accession=130210362, orientation='plus', exon_count=19, OMIM=606797, no_hdr0=''),
  6769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6769, CurrentID=0, Status='live', Symbol='STAC', Aliases='STAC1', description='SH3 and cysteine rich domain', other_designations='SRC homology 3 and cysteine-rich domain-containing protein', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36380424, end_position_on_the_genomic_accession=36548011, orientation='plus', exon_count=13, OMIM=602317, no_hdr0=''),
  6770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6770, CurrentID=0, Status='live', Symbol='STAR', Aliases='STARD1', description='steroidogenic acute regulatory protein', other_designations='START domain containing 1|START domain-containing protein 1|StAR-related lipid transfer (START) domain containing 1|cholesterol trafficker|mitochondrial steroid acute regulatory protein|steroid acute regulatory protein|steroidogenic acute regulator', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38142700, end_position_on_the_genomic_accession=38151265, orientation='minus', exon_count=7, OMIM=600617, no_hdr0=''),
  6771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6771, CurrentID=0, Status='live', Symbol='STARP1', Aliases='STARD1P1', description='steroidogenic acute regulatory protein pseudogene 1', other_designations='START domain containing 1 pseudogene 1|steroidogenic acute regulator pseudogene 1', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65280055, end_position_on_the_genomic_accession=65311163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6772, CurrentID=0, Status='live', Symbol='STAT1', Aliases='CANDF7, IMD31A, IMD31B, IMD31C, ISGF-3, STAT91', description='signal transducer and activator of transcription 1', other_designations='signal transducer and activator of transcription 1, 91kD|signal transducer and activator of transcription 1, 91kDa|signal transducer and activator of transcription-1|transcription factor ISGF-3 components p91/p84', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190969036, end_position_on_the_genomic_accession=191014250, orientation='minus', exon_count=25, OMIM=600555, no_hdr0=''),
  6773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6773, CurrentID=0, Status='live', Symbol='STAT2', Aliases='ISGF-3, P113, STAT113', description='signal transducer and activator of transcription 2', other_designations='interferon alpha induced transcriptional activator|signal transducer and activator of transcription 2, 113kDa', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56341597, end_position_on_the_genomic_accession=56360253, orientation='minus', exon_count=25, OMIM=600556, no_hdr0=''),
  6774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6774, CurrentID=0, Status='live', Symbol='STAT3', Aliases='ADMIO, APRF, HIES', description='signal transducer and activator of transcription 3 (acute-phase response factor)', other_designations='DNA-binding protein APRF|acute-phase response factor', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42313324, end_position_on_the_genomic_accession=42388503, orientation='minus', exon_count=24, OMIM=102582, no_hdr0=''),
  6775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6775, CurrentID=0, Status='live', Symbol='STAT4', Aliases='SLEB11', description='signal transducer and activator of transcription 4', other_designations='', map_location='2q32.2-q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191029576, end_position_on_the_genomic_accession=191246172, orientation='minus', exon_count=36, OMIM=600558, no_hdr0=''),
  6776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6776, CurrentID=0, Status='live', Symbol='STAT5A', Aliases='MGF, STAT5', description='signal transducer and activator of transcription 5A', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42287547, end_position_on_the_genomic_accession=42311943, orientation='plus', exon_count=20, OMIM=601511, no_hdr0=''),
  6777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6777, CurrentID=0, Status='live', Symbol='STAT5B', Aliases='STAT5', description='signal transducer and activator of transcription 5B', other_designations='transcription factor STAT5B', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42199177, end_position_on_the_genomic_accession=42276460, orientation='minus', exon_count=20, OMIM=604260, no_hdr0=''),
  6778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6778, CurrentID=0, Status='live', Symbol='STAT6', Aliases='D12S1644, IL-4-STATB, STAT6C, STAT6', description='signal transducer and activator of transcription 6, interleukin-4 induced', other_designations='STAT, interleukin4-induced|transcription factor IL-4 STAT', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57095404, end_position_on_the_genomic_accession=57111433, orientation='minus', exon_count=24, OMIM=601512, no_hdr0=''),
  6779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6779, CurrentID=0, Status='live', Symbol='STATH', Aliases='STR', description='statherin', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69995931, end_position_on_the_genomic_accession=70002456, orientation='plus', exon_count=6, OMIM=184470, no_hdr0=''),
  6780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6780, CurrentID=0, Status='live', Symbol='STAU1', Aliases='PPP1R150, STAU', description='staufen double-stranded RNA binding protein 1', other_designations='protein phosphatase 1, regulatory subunit 150|staufen, RNA binding protein, homolog 1', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49113339, end_position_on_the_genomic_accession=49188370, orientation='minus', exon_count=18, OMIM=601716, no_hdr0=''),
  6781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6781, CurrentID=0, Status='live', Symbol='STC1', Aliases='STC', description='stanniocalcin 1', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23841921, end_position_on_the_genomic_accession=23854807, orientation='minus', exon_count=4, OMIM=601185, no_hdr0=''),
  6782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6782, CurrentID=0, Status='live', Symbol='HSPA13', Aliases='STCH', description='heat shock protein family A (Hsp70) member 13', other_designations='heat shock protein 70kDa family member 13|heat shock protein 70kDa family, member 13|microsomal stress 70 protein ATPase core|microsomal stress-70 protein ATPase core|stress 70 protein chaperone, microsome-associated, 60kD|stress 70 protein chaperone, microsome-associated, 60kDa|stress-70 protein chaperone microsome-associated 60 kDa protein', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14371115, end_position_on_the_genomic_accession=14383188, orientation='minus', exon_count=5, OMIM=601100, no_hdr0=''),
  6783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6783, CurrentID=0, Status='live', Symbol='SULT1E1', Aliases='EST, EST-1, ST1E1, STE', description='sulfotransferase family 1E member 1', other_designations='estrone sulfotransferase|sulfotransferase 1E1|sulfotransferase family 1E, estrogen-preferring, member 1|sulfotransferase, estrogen-preferring', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69841212, end_position_on_the_genomic_accession=69860765, orientation='minus', exon_count=10, OMIM=600043, no_hdr0=''),
  6785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6785, CurrentID=0, Status='live', Symbol='ELOVL4', Aliases='ADMD, CT118, ISQMR, SCA34, STGD2, STGD3', description='ELOVL fatty acid elongase 4', other_designations='3-keto acyl-CoA synthase ELOVL4|ELOVL FA elongase 4|cancer/testis antigen 118|elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 4|very-long-chain 3-oxoacyl-CoA synthase 4', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79914812, end_position_on_the_genomic_accession=79947598, orientation='minus', exon_count=6, OMIM=605512, no_hdr0=''),
  6786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6786, CurrentID=0, Status='live', Symbol='STIM1', Aliases='D11S4896E, GOK, IMD10, STRMK, TAM, TAM1', description='stromal interaction molecule 1', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3855703, end_position_on_the_genomic_accession=4093210, orientation='plus', exon_count=13, OMIM=605921, no_hdr0=''),
  6787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6787, CurrentID=0, Status='live', Symbol='NEK4', Aliases='NRK2, STK2, pp12301', description='NIMA-related kinase 4', other_designations='NIMA (never in mitosis gene a)-related kinase 4|never in mitosis A-related kinase 4|nimA-related protein kinase 4|serine/threonine kinase 2|serine/threonine protein kinase-2|serine/threonine-protein kinase 2|serine/threonine-protein kinase NRK2|serine/threonine-protein kinase Nek4 variant 2', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52710780, end_position_on_the_genomic_accession=52770949, orientation='minus', exon_count=18, OMIM=601959, no_hdr0=''),
  6788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6788, CurrentID=0, Status='live', Symbol='STK3', Aliases='KRS1, MST2', description='serine/threonine kinase 3', other_designations='KB-1458E12.1|MST-2|STE20-like kinase MST2|mammalian STE20-like protein kinase 2|serine/threonine kinase 3 (STE20 homolog, yeast)|serine/threonine kinase 3 (Ste20, yeast homolog)|serine/threonine-protein kinase Krs-1', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98371355, end_position_on_the_genomic_accession=98942571, orientation='minus', exon_count=28, OMIM=605030, no_hdr0=''),
  6789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6789, CurrentID=0, Status='live', Symbol='STK4', Aliases='KRS2, MST1, TIIAC, YSK3', description='serine/threonine kinase 4', other_designations='MST-1|STE20-like kinase MST1|dJ211D12.2 (serine/threonine kinase 4 (MST1, KRS2))|kinase responsive to stress 2|mammalian STE20-like protein kinase 1|mammalian sterile 20-like 1|serine/threonine-protein kinase Krs-2', map_location='20q11.2-q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44966479, end_position_on_the_genomic_accession=45079975, orientation='plus', exon_count=17, OMIM=604965, no_hdr0=''),
  6790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6790, CurrentID=0, Status='live', Symbol='AURKA', Aliases='AIK, ARK1, AURA, AURORA2, BTAK, PPP1R47, STK15, STK6, STK7', description='aurora kinase A', other_designations='ARK-1|Aurora-A kinase|IPL1-related kinase|aurora 2|aurora-related kinase 1|aurora/IPL1-like kinase|aurora/IPL1-related kinase 1|breast tumor-amplified kinase|breast-tumor-amplified kinase|hARK1|protein phosphatase 1, regulatory subunit 47|serine/threonine kinase 6|serine/threonine protein kinase 15|serine/threonine-protein kinase 15|serine/threonine-protein kinase 6|serine/threonine-protein kinase aurora-A', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56369389, end_position_on_the_genomic_accession=56392295, orientation='minus', exon_count=11, OMIM=603072, no_hdr0=''),
  6791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6791, CurrentID=0, Status='live', Symbol='AURKAPS1', Aliases='AurAps1, STK6P', description='aurora kinase A pseudogene 1', other_designations='serine/threonine kinase 6 pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220266179, end_position_on_the_genomic_accession=220267715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  6792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6792, CurrentID=0, Status='live', Symbol='CDKL5', Aliases='EIEE2, ISSX, STK9', description='cyclin-dependent kinase-like 5', other_designations='cyclin dependent kinase 5 transcript|serine/threonine kinase 9|serine/threonine-protein kinase 9', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18425605, end_position_on_the_genomic_accession=18660858, orientation='plus', exon_count=25, OMIM=300203, no_hdr0=''),
  6793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6793, CurrentID=0, Status='live', Symbol='STK10', Aliases='LOK, PRO2729', description='serine/threonine kinase 10', other_designations='lymphocyte-oriented kinase', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172042070, end_position_on_the_genomic_accession=172188342, orientation='minus', exon_count=20, OMIM=603919, no_hdr0=''),
  6794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6794, CurrentID=0, Status='live', Symbol='STK11', Aliases='LKB1, PJS, hLKB1', description='serine/threonine kinase 11', other_designations='liver kinase B1|polarization-related protein LKB1|renal carcinoma antigen NY-REN-19|serine/threonine-protein kinase 11|serine/threonine-protein kinase LKB1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1205799, end_position_on_the_genomic_accession=1228435, orientation='plus', exon_count=13, OMIM=602216, no_hdr0=''),
  6795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6795, CurrentID=0, Status='live', Symbol='AURKC', Aliases='AIE2, AIK3, ARK3, AurC, HEL-S-90, SPGF5, STK13, aurora-C', description='aurora kinase C', other_designations='ARK-3|aurora 3|aurora-related kinase 3|aurora/IPL1-related kinase 3|aurora/IPL1/EG2 protein 2|epididymis secretory protein Li 90|serine/threonine kinase 13 (aurora/IPL1-like)|serine/threonine-protein kinase 13|serine/threonine-protein kinase aurora-C', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57230360, end_position_on_the_genomic_accession=57235550, orientation='plus', exon_count=7, OMIM=603495, no_hdr0=''),
  137868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137868, CurrentID=0, Status='live', Symbol='SGCZ', Aliases='ZSG1', description='sarcoglycan zeta', other_designations='sarcoglycan, zeta|zeta-SG', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14089864, end_position_on_the_genomic_accession=15238283, orientation='minus', exon_count=8, OMIM=608113, no_hdr0=''),
  400013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400013, CurrentID=0, Status='live', Symbol='TCP1P3', Aliases='CCT1-1P, TCP1L3', description='t-complex 1 pseudogene 3', other_designations='t-complex 1-like 3', map_location='12p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19941199, end_position_on_the_genomic_accession=19943109, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6799, CurrentID=0, Status='live', Symbol='SULT1A2', Aliases='HAST4, P-PST, ST1A2, STP2, TSPST2', description='sulfotransferase family 1A member 2', other_designations='P-PST 2|aryl sulfotransferase 2|arylamine sulfotransferase|phenol sulfotransferase 2|phenol-preferring phenol sulfotransferase2|phenol-sulfating phenol sulfotransferase 2|phenolic-metabolizing (P) form of PST|sulfotransferase family, cytosolic, 1A, phenol-preferring, member 2|thermostable phenol sulfotransferase', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28591943, end_position_on_the_genomic_accession=28597083, orientation='minus', exon_count=8, OMIM=601292, no_hdr0=''),
  137872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137872, CurrentID=0, Status='live', Symbol='ADHFE1', Aliases='ADH8, HMFT2263, HOT', description='alcohol dehydrogenase, iron containing 1', other_designations='Fe-containing alcohol dehydrogenase 1|alcohol dehydrogenase 8|alcohol dehydrogenase iron-containing protein 1|alcohol dehydrogenase, iron containing, 1', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66432483, end_position_on_the_genomic_accession=66468809, orientation='plus', exon_count=14, OMIM=611083, no_hdr0=''),
  6801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6801, CurrentID=0, Status='live', Symbol='STRN', Aliases='SG2NA', description='striatin', other_designations='striatin, calmodulin binding protein|striatin, calmodulin-binding protein', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36837698, end_position_on_the_genomic_accession=36966557, orientation='minus', exon_count=19, OMIM=614765, no_hdr0=''),
  6803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6803, CurrentID=0, Status='live', Symbol='STSP1', Aliases='STS-Y, STSP', description='steroid sulfatase (microsomal) pseudogene 1', other_designations='steroid sulfatase (microsomal), isozyme S pseudogene', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=15547209, end_position_on_the_genomic_accession=15566909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6804, CurrentID=0, Status='live', Symbol='STX1A', Aliases='HPC-1, P35-1, STX1, SYN1A', description='syntaxin 1A', other_designations='neuron-specific antigen HPC-1|syntaxin 1A (brain)', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73699205, end_position_on_the_genomic_accession=73719702, orientation='minus', exon_count=11, OMIM=186590, no_hdr0=''),
  100271214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271214, CurrentID=0, Status='live', Symbol='RPL7AP49', Aliases='RPL7A_16_980', description='ribosomal protein L7a pseudogene 49', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39983821, end_position_on_the_genomic_accession=39984394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6809, CurrentID=0, Status='live', Symbol='STX3', Aliases='STX3A', description='syntaxin 3', other_designations='syntaxin 3A', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59755059, end_position_on_the_genomic_accession=59805882, orientation='plus', exon_count=12, OMIM=600876, no_hdr0=''),
  6810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6810, CurrentID=0, Status='live', Symbol='STX4', Aliases='STX4A, p35-2', description='syntaxin 4', other_designations='renal carcinoma antigen NY-REN-31|syntaxin 4A (placental)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31033095, end_position_on_the_genomic_accession=31040168, orientation='plus', exon_count=12, OMIM=186591, no_hdr0=''),
  6811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6811, CurrentID=0, Status='live', Symbol='STX5', Aliases='SED5A, STX5', description='syntaxin 5', other_designations='syntaxin 5A', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62806860, end_position_on_the_genomic_accession=62832091, orientation='minus', exon_count=12, OMIM=603189, no_hdr0=''),
  6812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6812, CurrentID=0, Status='live', Symbol='STXBP1', Aliases='MUNC18-1, NSEC1, P67, RBSEC1, UNC18', description='syntaxin binding protein 1', other_designations='N-Sec1|neuronal SEC1|protein unc-18 homolog 1|protein unc-18 homolog A|unc-18A|unc18-1', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127612207, end_position_on_the_genomic_accession=127692716, orientation='plus', exon_count=20, OMIM=602926, no_hdr0=''),
  6813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6813, CurrentID=0, Status='live', Symbol='STXBP2', Aliases='FHL5, Hunc18b, MUNC18-2, UNC18-2, UNC18B, pp10122', description='syntaxin binding protein 2', other_designations='protein unc-18 homolog B|unc-18B', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7637101, end_position_on_the_genomic_accession=7647874, orientation='plus', exon_count=21, OMIM=601717, no_hdr0=''),
  6814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6814, CurrentID=0, Status='live', Symbol='STXBP3', Aliases='MUNC18-3, MUNC18C, PSP, UNC-18C', description='syntaxin binding protein 3', other_designations='platelet Sec1 protein|protein unc-18 homolog 3|protein unc-18 homolog C|syntaxin 4 binding protein|unc18-3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108746663, end_position_on_the_genomic_accession=108809526, orientation='plus', exon_count=19, OMIM=608339, no_hdr0=''),
  6815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6815, CurrentID=0, Status='live', Symbol='STYX', Aliases='', description='serine/threonine/tyrosine interacting protein', other_designations='protein tyrosine phosphatase-like protein', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52730165, end_position_on_the_genomic_accession=52774989, orientation='plus', exon_count=12, OMIM=615814, no_hdr0=''),
  6817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6817, CurrentID=0, Status='live', Symbol='SULT1A1', Aliases='HAST1/HAST2, P-PST, PST, ST1A1, ST1A3, STP, STP1, TSPST1', description='sulfotransferase family 1A member 1', other_designations='P-PST 1|aryl sulfotransferase 1|phenol-sulfating phenol sulfotransferase 1|sulfotransferase family, cytosolic, 1A, phenol-preferring, member 1|thermostable phenol sulfotransferase1|ts-PST', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28605587, end_position_on_the_genomic_accession=28623586, orientation='minus', exon_count=11, OMIM=171150, no_hdr0=''),
  6818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6818, CurrentID=0, Status='live', Symbol='SULT1A3', Aliases='HAST, HAST3, M-PST, ST1A3, ST1A3/ST1A4, ST1A5, STM, TL-PST', description='sulfotransferase family 1A member 3', other_designations='aryl sulfotransferase 1A3/1A4|catecholamine-sulfating phenol sulfotransferase|dopamine-specific sulfotransferase|monoamine-sulfating phenosulfotransferase|phenol sulfotransferase 1A5|placental estrogen sulfotransferase|sulfokinase|sulfotransferase family, cytosolic, 1A, phenol-preferring, member 3|thermolabile (monoamine, M form) phenol sulfotransferase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30199228, end_position_on_the_genomic_accession=30204329, orientation='plus', exon_count=8, OMIM=600641, no_hdr0=''),
  6819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6819, CurrentID=0, Status='live', Symbol='SULT1C2', Aliases='ST1C1, ST1C2, SULT1C1, humSULTC2', description='sulfotransferase family 1C member 2', other_designations='SULT1C#1|sulfotransferase 1C1|sulfotransferase family, cytosolic, 1C, member 1|sulfotransferase family, cytosolic, 1C, member 2', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108288639, end_position_on_the_genomic_accession=108310569, orientation='plus', exon_count=9, OMIM=602385, no_hdr0=''),
  6820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6820, CurrentID=0, Status='live', Symbol='SULT2B1', Aliases='HSST2', description='sulfotransferase family 2B member 1', other_designations='ST2B1|alcohol sulfotransferase|hydroxysteroid sulfotransferase 2|sulfotransferase 2B1|sulfotransferase family, cytosolic, 2B, member 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48552172, end_position_on_the_genomic_accession=48599427, orientation='plus', exon_count=8, OMIM=604125, no_hdr0=''),
  6821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6821, CurrentID=0, Status='live', Symbol='SUOX', Aliases='', description='sulfite oxidase', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55997259, end_position_on_the_genomic_accession=56005525, orientation='plus', exon_count=7, OMIM=606887, no_hdr0=''),
  6822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6822, CurrentID=0, Status='live', Symbol='SULT2A1', Aliases='DHEA-ST, DHEAS, HST, ST2, ST2A1, ST2A3, STD, hSTa', description='sulfotransferase family 2A member 1', other_designations='alcohol/hydroxysteroid sulfotransferase|bile-salt sulfotranasferase 2A1|sulfotransferase family, cytosolic, 2A, dehydroepiandrosterone (DHEA)-preferring, member 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47870466, end_position_on_the_genomic_accession=47886397, orientation='minus', exon_count=6, OMIM=125263, no_hdr0=''),
  100271217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271217, CurrentID=0, Status='live', Symbol='RPL7P39', Aliases='RPL7_16_1250', description='ribosomal protein L7 pseudogene 39', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64867888, end_position_on_the_genomic_accession=64868720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6824, CurrentID=0, Status='live', Symbol='ETF1P1', Aliases='HCGVII, SUP45L2', description='eukaryotic translation termination factor 1 pseudogene 1', other_designations='sup45 (yeast omnipotent suppressor 45) homolog-like 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30031713, end_position_on_the_genomic_accession=30033877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6825, CurrentID=0, Status='live', Symbol='ETF1P2', Aliases='SUP45L3', description='eukaryotic translation termination factor 1 pseudogene 2', other_designations='sup45 (yeast omnipotent suppressor 45) homolog-like 3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151501725, end_position_on_the_genomic_accession=151504327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6827, CurrentID=0, Status='live', Symbol='SUPT4H1', Aliases='SPT4, SPT4H, SUPT4H', description='SPT4 homolog, DSIF elongation factor subunit', other_designations='DRB sensitivity-inducing factor 14 kDa subunit|DSIF p14|small hSpt4 subunit|suppressor of Ty 4 homolog 1', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58345175, end_position_on_the_genomic_accession=58352238, orientation='minus', exon_count=5, OMIM=603555, no_hdr0=''),
  6829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6829, CurrentID=0, Status='live', Symbol='SUPT5H', Aliases='SPT5, SPT5H, Tat-CT1', description='SPT5 homolog, DSIF elongation factor subunit', other_designations='DRB sensitivity-inducing factor 160 kDa subunit|DRB sensitivity-inducing factor large subunit|DSIF large subunit|DSIF p160|Tat-cotransactivator 1 protein|hSPT5|suppressor of Ty 5 homolog', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39445546, end_position_on_the_genomic_accession=39476670, orientation='plus', exon_count=30, OMIM=602102, no_hdr0=''),
  6830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6830, CurrentID=0, Status='live', Symbol='SUPT6H', Aliases='SPT6, SPT6H, emb-5', description='SPT6 homolog, histone chaperone', other_designations='hSPT6|histone chaperone suppressor of Ty6|suppressor of Ty 6 homolog|tat-CT2 protein|tat-cotransactivator 2 protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28662203, end_position_on_the_genomic_accession=28702237, orientation='plus', exon_count=41, OMIM=601333, no_hdr0=''),
  6832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6832, CurrentID=0, Status='live', Symbol='SUPV3L1', Aliases='SUV3', description='SUV3-like helicase', other_designations='SUV3-like protein 1|suppressor of var1, 3-like 1(SUV3)', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69180204, end_position_on_the_genomic_accession=69209098, orientation='plus', exon_count=16, OMIM=605122, no_hdr0=''),
  6833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6833, CurrentID=0, Status='live', Symbol='ABCC8', Aliases='ABC36, HHF1, HI, HRINS, MRP8, PHHI, SUR, SUR1, SUR1delta2, TNDM2', description='ATP binding cassette subfamily C member 8', other_designations='ATP-binding cassette transporter sub-family C member 8|ATP-binding cassette, sub-family C (CFTR/MRP), member 8|sulfonylurea receptor (hyperinsulinemia)|sulfonylurea receptor 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17392885, end_position_on_the_genomic_accession=17476845, orientation='minus', exon_count=40, OMIM=600509, no_hdr0=''),
  6834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6834, CurrentID=0, Status='live', Symbol='SURF1', Aliases='', description='surfeit 1', other_designations='', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133351805, end_position_on_the_genomic_accession=133356485, orientation='minus', exon_count=9, OMIM=185620, no_hdr0=''),
  6835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6835, CurrentID=0, Status='live', Symbol='SURF2', Aliases='SURF-2', description='surfeit 2', other_designations='', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133356545, end_position_on_the_genomic_accession=133361165, orientation='plus', exon_count=6, OMIM=185630, no_hdr0=''),
  6836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6836, CurrentID=0, Status='live', Symbol='SURF4', Aliases='ERV29', description='surfeit 4', other_designations='surface 4 integral membrane protein', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133361449, end_position_on_the_genomic_accession=133377949, orientation='minus', exon_count=10, OMIM=185660, no_hdr0=''),
  6837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6837, CurrentID=0, Status='live', Symbol='MED22', Aliases='MED24, SURF5, surf-5', description='mediator complex subunit 22', other_designations='surfeit locus protein 5', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133340901, end_position_on_the_genomic_accession=133348156, orientation='minus', exon_count=5, OMIM=185641, no_hdr0=''),
  6838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6838, CurrentID=0, Status='live', Symbol='SURF6', Aliases='RRP14', description='surfeit 6', other_designations='', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133330707, end_position_on_the_genomic_accession=133336210, orientation='minus', exon_count=5, OMIM=185642, no_hdr0=''),
  6839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6839, CurrentID=0, Status='live', Symbol='SUV39H1', Aliases='H3-K9-HMTase 1, KMT1A, MG44, SUV39H', description='suppressor of variegation 3-9 homolog 1', other_designations='Su(var)3-9 homolog 1|histone H3-K9 methyltransferase 1|histone-lysine N-methyltransferase, H3 lysine-9 specific 1|lysine N-methyltransferase 1A|position-effect variegation 3-9 homolog', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48695554, end_position_on_the_genomic_accession=48709016, orientation='plus', exon_count=7, OMIM=300254, no_hdr0=''),
  6840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6840, CurrentID=0, Status='live', Symbol='SVIL', Aliases='', description='supervillin', other_designations='archvillin|membrane-associated F-actin binding protein p205|p205/p250', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29457338, end_position_on_the_genomic_accession=29736935, orientation='minus', exon_count=47, OMIM=604126, no_hdr0=''),
  100271220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271220, CurrentID=0, Status='live', Symbol='RPL7P40', Aliases='RPL7_19_1215', description='ribosomal protein L7 pseudogene 40', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=17121716, end_position_on_the_genomic_accession=17122460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400058, CurrentID=0, Status='live', Symbol='MKRN9P', Aliases='MKRN5, MKRN9, MKRNP6, RNF65, ZNF127L3', description='makorin ring finger protein 9, pseudogene', other_designations='makorin ring finger protein 1 pseudogene|makorin ring finger protein pseudogene 6|makorin, ring finger protein, 1 pseudogene|makorin, ring finger protein, 5|makorin, ring finger protein, 9|zinc finger protein 127-like 3', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87782885, end_position_on_the_genomic_accession=87784711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  6843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6843, CurrentID=0, Status='live', Symbol='VAMP1', Aliases='SPAX1, SYB1, VAMP-1', description='vesicle associated membrane protein 1', other_designations='vesicle-associated membrane protein 1 (synaptobrevin 1)', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6462237, end_position_on_the_genomic_accession=6470899, orientation='minus', exon_count=5, OMIM=185880, no_hdr0=''),
  6844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6844, CurrentID=0, Status='live', Symbol='VAMP2', Aliases='SYB2, VAMP-2', description='vesicle associated membrane protein 2', other_designations='synaptobrevin 2|synaptobrevin-2|vesicle-associated membrane protein 2 (synaptobrevin 2)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8159147, end_position_on_the_genomic_accession=8162975, orientation='minus', exon_count=6, OMIM=185881, no_hdr0=''),
  6845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6845, CurrentID=0, Status='live', Symbol='VAMP7', Aliases='SYBL1, TI-VAMP, TIVAMP, VAMP-7', description='vesicle associated membrane protein 7', other_designations='synaptobrevin-like 1|synaptobrevin-like protein 1|tetanus neurotoxin-insensitive VAMP|tetanus-insensitive VAMP', map_location='Xq28 and Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57067800, end_position_on_the_genomic_accession=57130289, orientation='plus', exon_count=8, OMIM=300053, no_hdr0=''),
  6846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6846, CurrentID=0, Status='live', Symbol='XCL2', Aliases='SCM-1b, SCM1B, SCYC2', description='chemokine (C motif) ligand 2', other_designations='XC chemokine ligand 2|c motif chemokine 2|small inducible cytokine subfamily C, member 2', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168540765, end_position_on_the_genomic_accession=168543997, orientation='minus', exon_count=3, OMIM=604828, no_hdr0=''),
  6847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6847, CurrentID=0, Status='live', Symbol='SYCP1', Aliases='CT8, HOM-TES-14, SCP-1, SCP1', description='synaptonemal complex protein 1', other_designations='cancer/testis antigen 8', map_location='1p13-p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114854416, end_position_on_the_genomic_accession=114995370, orientation='plus', exon_count=36, OMIM=602162, no_hdr0=''),
  400064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400064, CurrentID=0, Status='live', Symbol='RPS4XP1', Aliases='RPS4P1, RPS4X_6_1269', description='ribosomal protein S4X pseudogene 1', other_designations='ribosomal protein S4 pseudogene 1|ribosomal protein S4, X-linked pseudogene', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100008987, end_position_on_the_genomic_accession=100009894, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6850, CurrentID=0, Status='live', Symbol='SYK', Aliases='p72-Syk', description='spleen tyrosine kinase', other_designations='', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90801680, end_position_on_the_genomic_accession=90898560, orientation='plus', exon_count=16, OMIM=600085, no_hdr0=''),
  6852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6852, CurrentID=0, Status='live', Symbol='SYM2', Aliases='', description='symphalangism 2 (distal)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6853, CurrentID=0, Status='live', Symbol='SYN1', Aliases='SYN1ab, SYNI, SYN1', description='synapsin I', other_designations='brain protein 4.1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47571901, end_position_on_the_genomic_accession=47619857, orientation='minus', exon_count=13, OMIM=313440, no_hdr0=''),
  6854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6854, CurrentID=0, Status='live', Symbol='SYN2', Aliases='SYNII', description='synapsin II', other_designations='', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12004360, end_position_on_the_genomic_accession=12192032, orientation='plus', exon_count=16, OMIM=600755, no_hdr0=''),
  6855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6855, CurrentID=0, Status='live', Symbol='SYP', Aliases='MRX96, MRXSYP', description='synaptophysin', other_designations='major synaptic vesicle protein P38', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49187812, end_position_on_the_genomic_accession=49200202, orientation='minus', exon_count=7, OMIM=313475, no_hdr0=''),
  6856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6856, CurrentID=0, Status='live', Symbol='SYPL1', Aliases='H-SP1, SYPL', description='synaptophysin-like 1', other_designations='pantophysin', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106090368, end_position_on_the_genomic_accession=106112647, orientation='minus', exon_count=7, OMIM=616665, no_hdr0=''),
  6857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6857, CurrentID=0, Status='live', Symbol='SYT1', Aliases='P65, SVP65, SYT', description='synaptotagmin 1', other_designations='synaptotagmin I|sytI', map_location='12cen-q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78863993, end_position_on_the_genomic_accession=79452008, orientation='plus', exon_count=13, OMIM=185605, no_hdr0=''),
  100271223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271223, CurrentID=0, Status='live', Symbol='RPL7P42', Aliases='RPL7_22_1255', description='ribosomal protein L7 pseudogene 42', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68685780, end_position_on_the_genomic_accession=68686308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6860, CurrentID=0, Status='live', Symbol='SYT4', Aliases='HsT1192', description='synaptotagmin 4', other_designations='synaptotagmin IV|sytIV', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=43267892, end_position_on_the_genomic_accession=43277650, orientation='minus', exon_count=4, OMIM=600103, no_hdr0=''),
  6861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6861, CurrentID=0, Status='live', Symbol='SYT5', Aliases='', description='synaptotagmin 5', other_designations='synaptotagmin V|sytV', map_location='11p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55173096, end_position_on_the_genomic_accession=55180441, orientation='minus', exon_count=10, OMIM=600782, no_hdr0=''),
  6862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6862, CurrentID=0, Status='live', Symbol='T', Aliases='SAVAFT, T', description='T brachyury transcription factor', other_designations='T brachyury homolog|T, brachyury homolog|protein T', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166157656, end_position_on_the_genomic_accession=166168669, orientation='minus', exon_count=9, OMIM=601397, no_hdr0=''),
  6863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6863, CurrentID=0, Status='live', Symbol='TAC1', Aliases='Hs.2563, NK2, NKNA, NPK, TAC2', description='tachykinin precursor 1', other_designations='PPT|neurokinin 1|neurokinin 2|neurokinin A|neurokinin alpha|neuromedin L|neuropeptide K|neuropeptide gamma|preprotachykinin|protachykinin|substance K|substance P|tachykinin 2|tachykinin, precursor 1 (substance K, substance P, neurokinin 1, neurokinin 2, neuromedin L, neurokinin alpha, neuropeptide K, neuropeptide gamma)', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97731959, end_position_on_the_genomic_accession=97740472, orientation='plus', exon_count=7, OMIM=162320, no_hdr0=''),
  6865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6865, CurrentID=0, Status='live', Symbol='TACR2', Aliases='NK2R, NKNAR, SKR, TAC2R', description='tachykinin receptor 2', other_designations='NK-2 receptor|NK-2R|neurokinin 2 receptor|neurokinin A receptor|seven transmembrane helix receptor|substance K receptor', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69404202, end_position_on_the_genomic_accession=69416918, orientation='minus', exon_count=5, OMIM=162321, no_hdr0=''),
  6866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6866, CurrentID=0, Status='live', Symbol='TAC3', Aliases='HH10, NKB, NKNB, PRO1155, ZNEUROK1', description='tachykinin 3', other_designations='gamma tachykinin 3|neurokinin b|neuromedin K|preprotachykinin B|preprotachykinin-B', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57009997, end_position_on_the_genomic_accession=57016560, orientation='minus', exon_count=9, OMIM=162330, no_hdr0=''),
  6867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6867, CurrentID=0, Status='live', Symbol='TACC1', Aliases='Ga55', description='transforming, acidic coiled-coil containing protein 1', other_designations='gastric cancer antigen Ga55|taxin-1', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38728186, end_position_on_the_genomic_accession=38853028, orientation='plus', exon_count=19, OMIM=605301, no_hdr0=''),
  6868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6868, CurrentID=0, Status='live', Symbol='ADAM17', Aliases='ADAM18, CD156B, CSVP, NISBD, NISBD1, TACE', description='ADAM metallopeptidase domain 17', other_designations='ADAM metallopeptidase domain 18|TNF-alpha convertase|TNF-alpha converting enzyme|snake venom-like protease|tumor necrosis factor, alpha, converting enzyme', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9489263, end_position_on_the_genomic_accession=9555788, orientation='minus', exon_count=21, OMIM=603639, no_hdr0=''),
  6869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6869, CurrentID=0, Status='live', Symbol='TACR1', Aliases='NK1R, NKIR, SPR, TAC1R', description='tachykinin receptor 1', other_designations='NK-1 receptor|NK-1R|tachykinin receptor 1 (substance P receptor; neurokinin-1 receptor)', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75046463, end_position_on_the_genomic_accession=75199519, orientation='minus', exon_count=5, OMIM=162323, no_hdr0=''),
  6870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6870, CurrentID=0, Status='live', Symbol='TACR3', Aliases='HH11, NK-3R, NK3R, NKR, TAC3RL', description='tachykinin receptor 3', other_designations='NK-3 receptor|neurokinin B receptor|neurokinin beta receptor', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103589468, end_position_on_the_genomic_accession=103719816, orientation='minus', exon_count=5, OMIM=162332, no_hdr0=''),
  6871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6871, CurrentID=0, Status='live', Symbol='TADA2A', Aliases='ADA2, ADA2A, KL04P, TADA2L, hADA2', description='transcriptional adaptor 2A', other_designations='ADA2-like protein|transcriptional adaptor 2 alpha', map_location='17q12-q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37406877, end_position_on_the_genomic_accession=37477481, orientation='plus', exon_count=18, OMIM=602276, no_hdr0=''),
  6872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6872, CurrentID=0, Status='live', Symbol='TAF1', Aliases='BA2R, CCG1, CCGS, DYT3, DYT3/TAF1, KAT4, N-TAF1, NSCL2, OF, P250, TAF(II)250, TAF2A, TAFII-250, TAFII250, XDP', description='TATA-box binding protein associated factor 1', other_designations='TAF1 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 250kDa|TATA box binding protein associated factor 1|TBP-associated factor 250 kDa|cell cycle gene 1 protein|cell cycle, G1 phase defect|complementation of cell cycle block, G1-to-S|transcription factor TFIID p250 polypeptide', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71366218, end_position_on_the_genomic_accession=71530525, orientation='plus', exon_count=45, OMIM=313650, no_hdr0=''),
  6873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6873, CurrentID=0, Status='live', Symbol='TAF2', Aliases='CIF150, MRT40B, TAFII150, TAF2', description='TATA-box binding protein associated factor 2', other_designations='150 kDa cofactor of initiator function|RNA polymerase II TBP-associated factor subunit B|TAF(II)150|TAF2 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 150kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, B, 150kD|TATA box binding protein associated factor 2|TBP-associated factor 150 kDa|cofactor of initiator function, 150kD subunit|transcription initiation factor TFIID 150 kDa subunit', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119730773, end_position_on_the_genomic_accession=119832858, orientation='minus', exon_count=27, OMIM=604912, no_hdr0=''),
  6874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6874, CurrentID=0, Status='live', Symbol='TAF4', Aliases='TAF2C, TAF2C1A, TAFII130, TAFII135, TAF4', description='TATA-box binding protein associated factor 4', other_designations='RNA polymerase II TBP-associated factor subunit C|TAF(II)130|TAF(II)135|TAF4 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 135kDa|TAF4A RNA polymerase II, TATA box binding protein (TBP)-associated factor, 135 kD|TAFII-130|TAFII-135|TATA box binding protein (TBP)-associated factor, RNA polymerase II, C1, 130kD|TATA box binding protein associated factor 4|TBP-associated factor 4|transcription initiation factor TFIID 130 kDa subunit|transcription initiation factor TFIID 135 kD subunit|transcription initiation factor TFIID 135 kDa subunit', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61974798, end_position_on_the_genomic_accession=62065810, orientation='minus', exon_count=15, OMIM=601796, no_hdr0=''),
  6875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6875, CurrentID=0, Status='live', Symbol='TAF4B', Aliases='SPGF13, TAF2C2, TAFII105', description='TATA-box binding protein associated factor 4b', other_designations='TAF(II)105|TAF4b RNA polymerase II, TATA box binding protein (TBP)-associated factor, 105kDa|TATA box binding protein (TBP)-associated factor 4B|TATA box binding protein (TBP)-associated factor, RNA polymerase II, C2, 105kD|TATA box binding protein associated factor 4b|transcription initiation factor TFIID 105 kDa subunit', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26226445, end_position_on_the_genomic_accession=26391685, orientation='plus', exon_count=18, OMIM=601689, no_hdr0=''),
  6876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6876, CurrentID=0, Status='live', Symbol='TAGLN', Aliases='SM22, SMCC1, WS3-10, TAGLN', description='transgelin', other_designations='22 kDa actin-binding protein|SM22-alpha|smooth muscle protein 22-alpha|transgelin variant 2', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117199324, end_position_on_the_genomic_accession=117204792, orientation='plus', exon_count=5, OMIM=600818, no_hdr0=''),
  6877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6877, CurrentID=0, Status='live', Symbol='TAF5', Aliases='TAF(II)100, TAF2D, TAFII-100, TAFII100', description='TATA-box binding protein associated factor 5', other_designations='TAF5 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 100kDa|TATA box binding protein (TBP)-associated factor 2D|TATA box binding protein (TBP)-associated factor, RNA polymerase II, D, 100kD|TATA box binding protein associated factor 5|transcription initiation factor TFIID 100 kD subunit|transcription initiation factor TFIID 100 kDa subunit', map_location='10q24-q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103367953, end_position_on_the_genomic_accession=103389065, orientation='plus', exon_count=11, OMIM=601787, no_hdr0=''),
  6878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6878, CurrentID=0, Status='live', Symbol='TAF6', Aliases='MGC:8964, TAF(II)70, TAF(II)80, TAF2E, TAFII-70, TAFII-80, TAFII70, TAFII80, TAFII85', description='TATA-box binding protein associated factor 6', other_designations='RNA polymerase II TBP-associated factor subunit E|TAF6 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 80kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, E, 70/85kD|TATA box binding protein associated factor 6|transcription initiation factor TFIID 70 kDa subunit', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100107070, end_position_on_the_genomic_accession=100119858, orientation='minus', exon_count=18, OMIM=602955, no_hdr0=''),
  6879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6879, CurrentID=0, Status='live', Symbol='TAF7', Aliases='TAF2F, TAFII55', description='TATA-box binding protein associated factor 7', other_designations='RNA polymerase II TBP-associated factor subunit F|TAF(II)55|TAF7 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 55kDa|TAFII-55|TATA box binding protein (TBP)-associated factor, RNA polymerase II, F, 55kD|TATA box binding protein associated factor 7|TBP-associated factor F|transcription factor IID subunit TAFII55|transcription initiation factor TFIID 55 kDa subunit|transcription initiation factor TFIID, 55 kDa subunit', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141318490, end_position_on_the_genomic_accession=141320784, orientation='minus', exon_count=1, OMIM=600573, no_hdr0=''),
  6880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6880, CurrentID=0, Status='live', Symbol='TAF9', Aliases='MGC:5067, STAF31/32, TAF2G, TAFII-31, TAFII-32, TAFII31, TAFII32, TAFIID32', description='TATA-box binding protein associated factor 9', other_designations='RNA polymerase II TBP-associated factor subunit G|TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa|TATA box binding protein associated factor 9|transcription initiation factor TFIID 31 kD subunit|transcription initiation factor TFIID 31 kDa subunit|transcription initiation factor TFIID 32 kDa subunit', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69364743, end_position_on_the_genomic_accession=69370013, orientation='minus', exon_count=4, OMIM=600822, no_hdr0=''),
  6881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6881, CurrentID=0, Status='live', Symbol='TAF10', Aliases='TAF2A, TAF2H, TAFII30', description='TATA-box binding protein associated factor 10', other_designations='STAF28|TAF(II)30|TAF10 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 30kDa|TAFII-30|TATA box binding protein (TBP)-associated factor, RNA polymerase II, H, 30kD|TATA box binding protein associated factor 10|transcription initiation factor TFIID 30 kD subunit|transcription initiation factor TFIID 30 kDa subunit', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6610817, end_position_on_the_genomic_accession=6612244, orientation='minus', exon_count=5, OMIM=600475, no_hdr0=''),
  6882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6882, CurrentID=0, Status='live', Symbol='TAF11', Aliases='MGC:15243, PRO2134, TAF2I, TAFII28', description='TATA-box binding protein associated factor 11', other_designations='TAF(II)28|TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa|TAFII-28|TATA box binding protein (TBP)-associated factor, RNA polymerase II, I, 28kD|TATA box binding protein associated factor 11|TFIID subunit p30-beta|transcription initiation factor TFIID 28 kD subunit|transcription initiation factor TFIID 28 kDa subunit', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34877778, end_position_on_the_genomic_accession=34888071, orientation='minus', exon_count=6, OMIM=600772, no_hdr0=''),
  6883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6883, CurrentID=0, Status='live', Symbol='TAF12', Aliases='TAF2J, TAFII20', description='TATA-box binding protein associated factor 12', other_designations='TAF12 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 20kDa|TAFII-20/TAFII-15|TAFII20/TAFII15|TATA box binding protein (TBP)-associated factor, RNA polymerase II, J, 20kD|TATA box binding protein associated factor 12|transcription initiation factor TFIID 20/15 kDa subunits', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28603096, end_position_on_the_genomic_accession=28643109, orientation='minus', exon_count=7, OMIM=600773, no_hdr0=''),
  6884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6884, CurrentID=0, Status='live', Symbol='TAF13', Aliases='TAF(II)18, TAF2K, TAFII-18, TAFII18', description='TATA-box binding protein associated factor 13', other_designations='TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 18kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, K, 18kD|TATA box binding protein associated factor 13|transcription initiation factor TFIID 18 kD subunit|transcription initiation factor TFIID 18 kDa subunit', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109064376, end_position_on_the_genomic_accession=109076002, orientation='minus', exon_count=4, OMIM=600774, no_hdr0=''),
  6885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6885, CurrentID=0, Status='live', Symbol='MAP3K7', Aliases='MEKK7, TAK1, TGF1a', description='mitogen-activated protein kinase kinase kinase 7', other_designations='TGF-beta activated kinase 1|TGF-beta-activated kinase 1|transforming growth factor-beta-activated kinase 1', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90513573, end_position_on_the_genomic_accession=90587301, orientation='minus', exon_count=17, OMIM=602614, no_hdr0=''),
  6886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6886, CurrentID=0, Status='live', Symbol='TAL1', Aliases='SCL, TCL5, bHLHa17, tal-1', description='T-cell acute lymphocytic leukemia 1', other_designations='T-cell leukemia/lymphoma protein 5|class A basic helix-loop-helix protein 17|stem cell protein|tal-1 product', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47216290, end_position_on_the_genomic_accession=47232373, orientation='minus', exon_count=10, OMIM=187040, no_hdr0=''),
  6887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6887, CurrentID=0, Status='live', Symbol='TAL2', Aliases='', description='T-cell acute lymphocytic leukemia 2', other_designations='TAL-2|class A basic helix-loop-helix protein 19', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105662457, end_position_on_the_genomic_accession=105663112, orientation='plus', exon_count=1, OMIM=186855, no_hdr0=''),
  6888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6888, CurrentID=0, Status='live', Symbol='TALDO1', Aliases='TAL, TAL-H, TALDOR, TALH', description='transaldolase 1', other_designations='dihydroxyacetone transferase|glycerone transferase', map_location='11p15.5-p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=747432, end_position_on_the_genomic_accession=765024, orientation='plus', exon_count=8, OMIM=602063, no_hdr0=''),
  6889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6889, CurrentID=0, Status='live', Symbol='TALDO1P1', Aliases='TAL-H, TALDO, TALDOP1', description='transaldolase 1 pseudogene 1', other_designations='transaldolase pseudogene 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54524386, end_position_on_the_genomic_accession=54525214, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6890, CurrentID=0, Status='live', Symbol='TAP1', Aliases='ABC17, ABCB2, APT1, D6S114E, PSF-1, PSF1, RING4*0102N, TAP1N, TAP1', description='transporter 1, ATP-binding cassette, sub-family B (MDR/TAP)', other_designations='ABC transporter, MHC 1|ATP-binding cassette sub-family B member 2|ATP-binding cassette, sub-family B (MDR/TAP), member 2|peptide supply factor 1|peptide transporter PSF1|peptide transporter TAP1|peptide transporter involved in antigen processing 1|really interesting new gene 4 protein|transporter 1 ATP-binding cassette sub-family B|transporter associated with antigen processing|transporter, ATP-binding cassette, major histocompatibility complex, 1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32845209, end_position_on_the_genomic_accession=32853971, orientation='minus', exon_count=12, OMIM=170260, no_hdr0=''),
  6891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6891, CurrentID=0, Status='live', Symbol='TAP2', Aliases='ABC18, ABCB3, APT2, D6S217E, PSF-2, PSF2, RING11', description='transporter 2, ATP-binding cassette, sub-family B (MDR/TAP)', other_designations='ABC transporter, MHC 2|ATP-binding cassette, sub-family B (MDR/TAP), member 3|peptide supply factor 2|peptide transporter PSF2|peptide transporter involved in antigen processing 2|really interesting new gene 11 protein|transporter 2, ABC (ATP binding cassette)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32821833, end_position_on_the_genomic_accession=32838823, orientation='minus', exon_count=13, OMIM=170261, no_hdr0=''),
  6892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6892, CurrentID=0, Status='live', Symbol='TAPBP', Aliases='NGS17, TAPA, TPN, TPSN', description='TAP binding protein (tapasin)', other_designations='NGS-17|TAP-associated protein|TAP-binding protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33299694, end_position_on_the_genomic_accession=33314387, orientation='minus', exon_count=8, OMIM=601962, no_hdr0=''),
  6893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6893, CurrentID=0, Status='live', Symbol='TAPVR1', Aliases='APVR, TAPVR', description='total anomalous pulmonary venous return 1', other_designations='', map_location='4p13-q11', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=106700, no_hdr0=''),
  6894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6894, CurrentID=0, Status='live', Symbol='TARBP1', Aliases='TRM3, TRP-185, TRP185', description='TAR (HIV-1) RNA binding protein 1', other_designations='TAR (HIV) RNA binding protein 1|TAR (HIV) RNA-binding protein 1|TAR RNA loop binding protein|TAR RNA-binding protein 1|TAR RNA-binding protein of 185 kDa|tRNA methyltransferase 3 homolog', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234391313, end_position_on_the_genomic_accession=234479158, orientation='minus', exon_count=34, OMIM=605052, no_hdr0=''),
  6895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6895, CurrentID=0, Status='live', Symbol='TARBP2', Aliases='LOQS, TRBP, TRBP1, TRBP2', description='TAR (HIV-1) RNA binding protein 2', other_designations='TAR (HIV) RNA binding protein 2|TAR (HIV) RNA-binding protein 2|TAR (HIV) RNA-binding protein TRBP1|TAR RNA binding protein 2|TAR RNA-binding protein 2|trans-activation responsive RNA-binding protein|trans-activation-responsive RNA-binding protein', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53500203, end_position_on_the_genomic_accession=53506431, orientation='plus', exon_count=10, OMIM=605053, no_hdr0=''),
  6896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6896, CurrentID=0, Status='live', Symbol='TARBP2P', Aliases='TRBP', description='TAR (HIV-1) RNA binding protein 2 pseudogene', other_designations='TAR (HIV) RNA binding protein 2 pseudogene|TAR (HIV) RNA-binding protein 2 pseudogene', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103659244, end_position_on_the_genomic_accession=103660588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6897, CurrentID=0, Status='live', Symbol='TARS', Aliases='ThrRS', description='threonyl-tRNA synthetase', other_designations='threonine tRNA ligase 1, cytoplasmic|threonyl-tRNA synthetase, cytoplasmic', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33440696, end_position_on_the_genomic_accession=33468091, orientation='plus', exon_count=21, OMIM=187790, no_hdr0=''),
  6898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6898, CurrentID=0, Status='live', Symbol='TAT', Aliases='', description='tyrosine aminotransferase', other_designations='L-tyrosine:2-oxoglutarate aminotransferase|tyrosine aminotransferase, cytosolic', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71566851, end_position_on_the_genomic_accession=71577095, orientation='minus', exon_count=12, OMIM=613018, no_hdr0=''),
  6899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6899, CurrentID=0, Status='live', Symbol='TBX1', Aliases='CAFS, CATCH22, CTHM, DGCR, DGS, DORVC, TGA, VCF, VCFS, TBX1', description='T-box 1', other_designations='T-box 1 transcription factor C|Testis-specific T-box protein|brachyury', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19756703, end_position_on_the_genomic_accession=19783593, orientation='plus', exon_count=13, OMIM=602054, no_hdr0=''),
  6900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6900, CurrentID=0, Status='live', Symbol='CNTN2', Aliases='AXT, FAME5, TAG-1, TAX, TAX1', description='contactin 2', other_designations='axonal glycoprotein TAG-1|axonin-1 cell adhesion molecule|contactin 2 (axonal)|contactin 2 (transiently expressed)|transient axonal glycoprotein 1|transiently-expressed axonal glycoprotein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205043202, end_position_on_the_genomic_accession=205078043, orientation='plus', exon_count=24, OMIM=190197, no_hdr0=''),
  6901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6901, CurrentID=0, Status='live', Symbol='TAZ', Aliases='BTHS, CMD3A, EFE, EFE2, G4.5, LVNCX, Taz1', description='tafazzin', other_designations='protein G4.5', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154411518, end_position_on_the_genomic_accession=154421726, orientation='plus', exon_count=12, OMIM=300394, no_hdr0=''),
  6902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6902, CurrentID=0, Status='live', Symbol='TBCA', Aliases='', description='tubulin folding cofactor A', other_designations='CFA|TCP1-chaperonin cofactor A|chaperonin cofactor A|tubulin-folding cofactor A', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77691169, end_position_on_the_genomic_accession=77776361, orientation='minus', exon_count=3, OMIM=610058, no_hdr0=''),
  6903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6903, CurrentID=0, Status='live', Symbol='TBCC', Aliases='CFC', description='tubulin folding cofactor C', other_designations='tubulin-folding cofactor C', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42744496, end_position_on_the_genomic_accession=42746146, orientation='minus', exon_count=1, OMIM=602971, no_hdr0=''),
  6904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6904, CurrentID=0, Status='live', Symbol='TBCD', Aliases='SSD-1, tfcD', description='tubulin folding cofactor D', other_designations='beta-tubulin cofactor D|tubulin-folding cofactor D', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82752060, end_position_on_the_genomic_accession=82943186, orientation='plus', exon_count=57, OMIM=604649, no_hdr0=''),
  6905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6905, CurrentID=0, Status='live', Symbol='TBCE', Aliases='HRD, KCS, KCS1, pac2', description='tubulin folding cofactor E', other_designations='hypoparathyroidism, growth and mental retardation, and dysmorphism', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235367360, end_position_on_the_genomic_accession=235448968, orientation='plus', exon_count=18, OMIM=604934, no_hdr0=''),
  6906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6906, CurrentID=0, Status='live', Symbol='SERPINA7', Aliases='TBG', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 7', other_designations='T4-binding globulin|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 7|serpin A7|thyroxin-binding globulin', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106033199, end_position_on_the_genomic_accession=106038727, orientation='minus', exon_count=5, OMIM=314200, no_hdr0=''),
  6907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6907, CurrentID=0, Status='live', Symbol='TBL1X', Aliases='EBI, SMAP55, TBL1', description='transducin (beta)-like 1X-linked', other_designations='transducin beta-like protein 1X|transducin-beta-like protein 1, X-linked', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9463295, end_position_on_the_genomic_accession=9719740, orientation='plus', exon_count=20, OMIM=300196, no_hdr0=''),
  6908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6908, CurrentID=0, Status='live', Symbol='TBP', Aliases='GTF2D, GTF2D1, HDL4, SCA17, TFIID', description='TATA-box binding protein', other_designations='TATA box binding protein|TATA sequence-binding protein|TATA-box binding protein N-terminal domain|TATA-box factor|transcription initiation factor TFIID TBP subunit', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170554333, end_position_on_the_genomic_accession=170572870, orientation='plus', exon_count=8, OMIM=600075, no_hdr0=''),
  6909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6909, CurrentID=0, Status='live', Symbol='TBX2', Aliases='', description='T-box 2', other_designations='T-box protein 2', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61394698, end_position_on_the_genomic_accession=61409466, orientation='plus', exon_count=9, OMIM=600747, no_hdr0=''),
  6910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6910, CurrentID=0, Status='live', Symbol='TBX5', Aliases='HOS', description='T-box 5', other_designations='T-box protein 5', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114353930, end_position_on_the_genomic_accession=114408442, orientation='minus', exon_count=10, OMIM=601620, no_hdr0=''),
  6911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6911, CurrentID=0, Status='live', Symbol='TBX6', Aliases='SCDO5', description='T-box 6', other_designations='T-box protein 6', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30085793, end_position_on_the_genomic_accession=30091937, orientation='minus', exon_count=9, OMIM=602427, no_hdr0=''),
  400128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400128, CurrentID=0, Status='live', Symbol='TUSC8', Aliases='LINC01071, XLOC_010588', description='tumor suppressor candidate 8 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1071', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44400250, end_position_on_the_genomic_accession=44405984, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  6913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6913, CurrentID=0, Status='live', Symbol='TBX15', Aliases='TBX14', description='T-box 15', other_designations='T-box 14|T-box protein 14|T-box protein 15|T-box transcription factor TBX14', map_location='1p11.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118883043, end_position_on_the_genomic_accession=118989556, orientation='minus', exon_count=11, OMIM=604127, no_hdr0=''),
  6914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6914, CurrentID=0, Status='live', Symbol='TBXA1R', Aliases='', description='thromboxane A1 receptor', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6915, CurrentID=0, Status='live', Symbol='TBXA2R', Aliases='BDPLT13, TXA2-R', description='thromboxane A2 receptor', other_designations='prostanoid TP receptor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3594506, end_position_on_the_genomic_accession=3606851, orientation='minus', exon_count=5, OMIM=188070, no_hdr0=''),
  6916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6916, CurrentID=0, Status='live', Symbol='TBXAS1', Aliases='BDPLT14, CYP5, CYP5A1, GHOSAL, THAS, TS, TXAS, TXS', description='thromboxane A synthase 1', other_designations='TXA synthase|cytochrome P450 5A1|cytochrome P450, family 5, subfamily A, polypeptide 1|platelet, cytochrome P450, subfamily V|thromboxane A synthase 1 (platelet)', map_location='7q34-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139778248, end_position_on_the_genomic_accession=140020325, orientation='plus', exon_count=19, OMIM=274180, no_hdr0=''),
  6917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6917, CurrentID=0, Status='live', Symbol='TCEA1', Aliases='GTF2S, SII, TCEA, TF2S, TFIIS', description='transcription elongation factor A (SII), 1', other_designations='transcription elongation factor S-II protein 1|transcription elongation factor TFIIS.o', map_location='8q11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53966554, end_position_on_the_genomic_accession=54022456, orientation='minus', exon_count=13, OMIM=601425, no_hdr0=''),
  6918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6918, CurrentID=0, Status='live', Symbol='TCEA1P1', Aliases='GTF2SP, TCEA1P', description='transcription elongation factor A (SII), 1 pseudogene 1', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97067594, end_position_on_the_genomic_accession=97070769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6919, CurrentID=0, Status='live', Symbol='TCEA2', Aliases='TFIIS', description='transcription elongation factor A (SII), 2', other_designations='testis-specific S-II|transcription elongation factor S-II protein 2|transcription elongation factor TFIIS.1|transcription elongation factor TFIIS.l', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64042350, end_position_on_the_genomic_accession=64072347, orientation='plus', exon_count=15, OMIM=604784, no_hdr0=''),
  6920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6920, CurrentID=0, Status='live', Symbol='TCEA3', Aliases='TFIIS, TFIIS.H', description='transcription elongation factor A (SII), 3', other_designations='rhabdomyosarcoma antigen MU-RMS-40.22|transcription elongation factor S-II protein 3|transcription elongation factor TFIIS.h', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23380909, end_position_on_the_genomic_accession=23424779, orientation='minus', exon_count=14, OMIM=604128, no_hdr0=''),
  6921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6921, CurrentID=0, Status='live', Symbol='TCEB1', Aliases='SIII, eloC', description='transcription elongation factor B (SIII), polypeptide 1 (15kDa, elongin C)', other_designations='RNA polymerase II transcription factor SIII subunit C|SIII p15|elongin 15 kDa subunit', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73945138, end_position_on_the_genomic_accession=73972287, orientation='minus', exon_count=9, OMIM=600788, no_hdr0=''),
  137994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137994, CurrentID=0, Status='live', Symbol='LETM2', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 2', other_designations='LETM1 and EF-hand domain-containing protein 2|leucine zipper-EF-hand containing transmembrane protein 1-like protein', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38386429, end_position_on_the_genomic_accession=38409527, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  6923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6923, CurrentID=0, Status='live', Symbol='TCEB2', Aliases='ELOB, SIII', description='transcription elongation factor B (SIII), polypeptide 2 (18kDa, elongin B)', other_designations='RNA polymerase II transcription factor SIII p18 subunit|RNA polymerase II transcription factor SIII subunit B|SIII p18|elongin 18 kDa subunit|elongin, 18-kD subunit|elongin-B', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2771414, end_position_on_the_genomic_accession=2777296, orientation='minus', exon_count=5, OMIM=600787, no_hdr0=''),
  6924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6924, CurrentID=0, Status='live', Symbol='TCEB3', Aliases='EloA, SIII, SIII p110A, TCEB3', description='transcription elongation factor B (SIII), polypeptide 3 (110kDa, elongin A)', other_designations='RNA polymerase II transcription factor SIII subunit A1|elongin 110 kDa subunit|elongin-A|transcription elongation factor B alpha subunit', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23743366, end_position_on_the_genomic_accession=23762059, orientation='plus', exon_count=11, OMIM=600786, no_hdr0=''),
  6925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6925, CurrentID=0, Status='live', Symbol='TCF4', Aliases='E2-2, ITF-2, ITF2, PTHS, SEF-2, SEF2, SEF2-1, SEF2-1A, SEF2-1B, SEF2-1D, TCF-4, bHLHb19', description='transcription factor 4', other_designations='SL3-3 enhancer factor 2|class B basic helix-loop-helix protein 19|immunoglobulin transcription factor 2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55222331, end_position_on_the_genomic_accession=55635993, orientation='minus', exon_count=34, OMIM=602272, no_hdr0=''),
  6926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6926, CurrentID=0, Status='live', Symbol='TBX3', Aliases='TBX3-ISO, UMS, XHL', description='T-box 3', other_designations='T-box protein 3|bladder cancer related protein XHL', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114670254, end_position_on_the_genomic_accession=114684164, orientation='minus', exon_count=8, OMIM=601621, no_hdr0=''),
  6927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6927, CurrentID=0, Status='live', Symbol='HNF1A', Aliases='HNF-1A, HNF1, IDDM20, LFB1, MODY3, TCF-1, TCF1', description='HNF1 homeobox A', other_designations='HNF-1-alpha|albumin proximal factor|hepatic nuclear factor 1|interferon production regulator factor|liver-specific transcription factor LF-B1|transcription factor 1, hepatic', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120977787, end_position_on_the_genomic_accession=121002512, orientation='plus', exon_count=9, OMIM=142410, no_hdr0=''),
  6928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6928, CurrentID=0, Status='live', Symbol='HNF1B', Aliases='FJHN, HNF-1-beta, HNF-1B, HNF1beta, HNF2, HPC11, LF-B3, LFB3, MODY5, TCF-2, TCF2, VHNF1', description='HNF1 homeobox B', other_designations='HNF1 beta A|homeoprotein LFB3|transcription factor 2, hepatic', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37686431, end_position_on_the_genomic_accession=37745078, orientation='minus', exon_count=11, OMIM=189907, no_hdr0=''),
  6929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6929, CurrentID=0, Status='live', Symbol='TCF3', Aliases='E2A, E47, ITF1, TCF-3, VDIR, bHLHb21', description='transcription factor 3', other_designations='NOL1-TCF3 fusion|VDR interacting repressor|class B basic helix-loop-helix protein 21|helix-loop-helix protein HE47|immunoglobulin transcription factor 1|kappa-E2-binding factor|negative vitamin D response element-binding protein|transcription factor 3 (E2A immunoglobulin enhancer binding factors E12/E47)|transcription factor 3 variant 3|transcription factor ITF-1|vitamin D receptor-interacting repressor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1609290, end_position_on_the_genomic_accession=1652426, orientation='minus', exon_count=22, OMIM=147141, no_hdr0=''),
  100271235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271235, CurrentID=0, Status='live', Symbol='RPS20P25', Aliases='RPS20_13_989', description='ribosomal protein S20 pseudogene 25', other_designations='', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81410279, end_position_on_the_genomic_accession=81410633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6932, CurrentID=0, Status='live', Symbol='TCF7', Aliases='TCF-1', description='transcription factor 7 (T-cell specific, HMG-box)', other_designations='T-cell-specific transcription factor 1', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134114703, end_position_on_the_genomic_accession=134148229, orientation='plus', exon_count=17, OMIM=189908, no_hdr0=''),
  6934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6934, CurrentID=0, Status='live', Symbol='TCF7L2', Aliases='TCF-4, TCF4', description='transcription factor 7-like 2 (T-cell specific, HMG-box)', other_designations='HMG box transcription factor 4|T-cell factor 4|T-cell factor-4 variant A|T-cell factor-4 variant B|T-cell factor-4 variant C|T-cell factor-4 variant D|T-cell factor-4 variant E|T-cell factor-4 variant F|T-cell factor-4 variant G|T-cell factor-4 variant H|T-cell factor-4 variant I|T-cell factor-4 variant J|T-cell factor-4 variant K|T-cell factor-4 variant L|T-cell factor-4 variant M|T-cell factor-4 variant X2|T-cell-specific transcription factor 4|hTCF-4', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112950219, end_position_on_the_genomic_accession=113167678, orientation='plus', exon_count=21, OMIM=602228, no_hdr0=''),
  6935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6935, CurrentID=0, Status='live', Symbol='ZEB1', Aliases='AREB6, BZP, DELTAEF1, FECD6, NIL2A, PPCD3, TCF8, ZFHEP, ZFHX1A', description='zinc finger E-box binding homeobox 1', other_designations='delta-crystallin enhancer binding factor 1|negative regulator of IL2|posterior polymorphous corneal dystrophy 3|transcription factor 8 (represses interleukin 2 expression)|zinc finger homeodomain enhancer-binding protein', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31318783, end_position_on_the_genomic_accession=31529814, orientation='plus', exon_count=17, OMIM=189909, no_hdr0=''),
  6936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6936, CurrentID=0, Status='live', Symbol='GCFC2', Aliases='C2orf3, DNABF, GCF, TCF9', description='GC-rich sequence DNA-binding factor 2', other_designations='GC binding factor|GC bindng factor|TCF-9|transcription factor 9 (binds GC-rich sequences)', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75662705, end_position_on_the_genomic_accession=75710986, orientation='minus', exon_count=17, OMIM=189901, no_hdr0=''),
  138009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138009, CurrentID=0, Status='live', Symbol='DCAF4L2', Aliases='WDR21C', description='DDB1 and CUL4 associated factor 4-like 2', other_designations='WD repeat domain 21C|WD repeat-containing protein 21C', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87870743, end_position_on_the_genomic_accession=87874068, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  6938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6938, CurrentID=0, Status='live', Symbol='TCF12', Aliases='CRS3, HEB, HTF4, HsT17266, TCF-12, bHLHb20', description='transcription factor 12', other_designations='DNA-binding protein HTF4|E-box-binding protein|class B basic helix-loop-helix protein 20|helix-loop-helix transcription factor 4|transcription factor HTF-4', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56918289, end_position_on_the_genomic_accession=57291129, orientation='plus', exon_count=25, OMIM=600480, no_hdr0=''),
  6939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6939, CurrentID=0, Status='live', Symbol='TCF15', Aliases='EC2, PARAXIS, bHLHa40', description='transcription factor 15 (basic helix-loop-helix)', other_designations='TCF-15|basic helix-loop-helix transcription factor 15|class A basic helix-loop-helix protein 40|protein bHLH-EC2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=603993, end_position_on_the_genomic_accession=610266, orientation='minus', exon_count=2, OMIM=601010, no_hdr0=''),
  6940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6940, CurrentID=0, Status='live', Symbol='ZNF354A', Aliases='EZNF, HEL104, HKL1, KID-1, KID1, TCF17', description='zinc finger protein 354A', other_designations='TCF-17|epididymis luminal protein 104|transcription factor 17|zinc finger protein eZNF', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178711512, end_position_on_the_genomic_accession=178730702, orientation='minus', exon_count=6, OMIM=602444, no_hdr0=''),
  6941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6941, CurrentID=0, Status='live', Symbol='TCF19', Aliases='SC1, TCF-19', description='transcription factor 19', other_designations='transcription factor SC1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31158526, end_position_on_the_genomic_accession=31164215, orientation='plus', exon_count=5, OMIM=600912, no_hdr0=''),
  6942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6942, CurrentID=0, Status='live', Symbol='TCF20', Aliases='AR1, SPBP, TCF-20', description='transcription factor 20 (AR1)', other_designations='SPRE-binding protein|nuclear factor SPBP|stromelysin-1 PDGF-responsive element-binding protein|stromelysin-1 platelet-derived growth factor-responsive element binding protein', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42160013, end_position_on_the_genomic_accession=42283922, orientation='minus', exon_count=9, OMIM=603107, no_hdr0=''),
  6943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6943, CurrentID=0, Status='live', Symbol='TCF21', Aliases='POD1, bHLHa23', description='transcription factor 21', other_designations='capsulin|class A basic helix-loop-helix protein 23|epicardin|pod-1|podocyte-expressed 1', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133889121, end_position_on_the_genomic_accession=133895537, orientation='plus', exon_count=3, OMIM=603306, no_hdr0=''),
  6944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6944, CurrentID=0, Status='live', Symbol='VPS72', Aliases='CFL1, Swc2, TCFL1, YL-1, YL1', description='vacuolar protein sorting 72 homolog (S. cerevisiae)', other_designations='transcription factor-like 1|transformation suppressor gene YL-1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151176300, end_position_on_the_genomic_accession=151190213, orientation='minus', exon_count=6, OMIM=600607, no_hdr0=''),
  6945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6945, CurrentID=0, Status='live', Symbol='MLX', Aliases='MAD7, MXD7, TCFL4, bHLHd13', description='MLX, MAX dimerization protein', other_designations='BigMax protein|MAX-like bHLHZIP protein|class D basic helix-loop-helix protein 13|transcription factor-like protein 4', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42567060, end_position_on_the_genomic_accession=42573203, orientation='plus', exon_count=8, OMIM=602976, no_hdr0=''),
  6946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6946, CurrentID=0, Status='live', Symbol='TCL4', Aliases='', description='T-cell leukemia/lymphoma 4', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=186860, no_hdr0=''),
  6947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6947, CurrentID=0, Status='live', Symbol='TCN1', Aliases='HC, TC-1, TC1, TCI', description='transcobalamin 1', other_designations='haptocorin|haptocorrin|protein R|transcobalamin I (vitamin B12 binding protein, R binder family)', map_location='11q11-q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59852808, end_position_on_the_genomic_accession=59866568, orientation='minus', exon_count=9, OMIM=189905, no_hdr0=''),
  6948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6948, CurrentID=0, Status='live', Symbol='TCN2', Aliases='D22S676, D22S750, II, TC, TC II, TC-2, TC2, TCII', description='transcobalamin 2', other_designations='macrocytic anemia|transcobalamin II|transcobalamin II; macrocytic anemia|vitamin B12-binding protein 2', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30607083, end_position_on_the_genomic_accession=30627060, orientation='plus', exon_count=9, OMIM=613441, no_hdr0=''),
  6949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6949, CurrentID=0, Status='live', Symbol='TCOF1', Aliases='MFD1, TCS, TCS1, treacle', description='Treacher Collins-Franceschetti syndrome 1', other_designations='Treacher Collins syndrome protein|nucleolar trafficking phosphoprotein', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150357639, end_position_on_the_genomic_accession=150400308, orientation='plus', exon_count=29, OMIM=606847, no_hdr0=''),
  6950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6950, CurrentID=0, Status='live', Symbol='TCP1', Aliases='CCT-alpha, CCT1, CCTa, D6S230E, TCP-1-alpha', description='t-complex 1', other_designations='T-complex protein 1, alpha subunit|tailless complex polypeptide 1', map_location='6q25.3-q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159778498, end_position_on_the_genomic_accession=159789703, orientation='minus', exon_count=12, OMIM=186980, no_hdr0=''),
  6952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6952, CurrentID=0, Status='live', Symbol='TCP1P2', Aliases='CCT1-2P, TCP1L2', description='t-complex 1 pseudogene 2', other_designations='t-complex 1-like 2', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41585767, end_position_on_the_genomic_accession=41587981, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6953, CurrentID=0, Status='live', Symbol='TCP10', Aliases='TCP10A', description='t-complex 10', other_designations='t-complex 10 (a murine tcp homolog)|t-complex 10 homolog', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167372379, end_position_on_the_genomic_accession=167384552, orientation='minus', exon_count=12, OMIM=187020, no_hdr0=''),
  6954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6954, CurrentID=0, Status='live', Symbol='TCP11', Aliases='D6S230E, FPPR', description='t-complex 11, testis-specific', other_designations='fertilization-promoting peptide receptor|t-complex 11 (a murine tcp homolog)|t-complex 11 homolog', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35118071, end_position_on_the_genomic_accession=35141642, orientation='minus', exon_count=11, OMIM=186982, no_hdr0=''),
  6955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6955, CurrentID=0, Status='live', Symbol='TRA', Aliases='IMD7, TCRA, TCRD@, TRAC, TRA', description='T-cell receptor alpha locus', other_designations='T cell receptor alpha locus|T-cell antigen receptor, alpha polypeptide|T-cell receptor, alpha (V,D,J,C)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21621904, end_position_on_the_genomic_accession=22552132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6956, CurrentID=0, Status='live', Symbol='TRAV6', Aliases='TCRAV5S1, TCRAV6S1', description='T cell receptor alpha variable 6', other_designations='T-cell receptor, alpha variable region 5, segment 1', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21768553, end_position_on_the_genomic_accession=21769080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6957, CurrentID=0, Status='live', Symbol='TRB', Aliases='TCRB@, TRB', description='T cell receptor beta locus', other_designations='T-cell antigen receptor, beta polypeptide, T-cell receptor, beta cluster|T-cell receptor, beta cluster', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142299011, end_position_on_the_genomic_accession=142813287, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6958, CurrentID=0, Status='live', Symbol='TRBV29OR9-2', Aliases='TCRBV29S2, TCRBV4O, TCRBV4S2O, TRBV29/OR9-2', description='T cell receptor beta variable 29/OR9-2 (non-functional)', other_designations='T cell receptor beta variable 29/OR9-2 pseudogene|T-cell receptor, beta cluster, variable region 4, orphon', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33786221, end_position_on_the_genomic_accession=33786828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6959, CurrentID=0, Status='live', Symbol='TRBV21OR9-2', Aliases='TCRBV10O, TCRBV10S2O, TCRBV21S2', description='T cell receptor beta variable 21/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable 21/OR9-2 (non-functional)|T cell receptor beta variable 21/OR9-2 pseudogene|T-cell receptor, beta variable region 10, orphon|TRBV21/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33629143, end_position_on_the_genomic_accession=33629588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6960, CurrentID=0, Status='live', Symbol='TRBV25OR9-2', Aliases='TCRBV11O, TCRBV11S2O, TCRBV25S2', description='T cell receptor beta variable 25/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable 25/OR9-2 pseudogene|T-cell receptor, beta variable region 11, orphon|TRBV25/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33662203, end_position_on_the_genomic_accession=33662664, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6961, CurrentID=0, Status='live', Symbol='TRBV24OR9-2', Aliases='TCRBV15O, TCRBV15S2O, TCRBV24S2', description='T cell receptor beta variable 24/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable 24/OR9-2 pseudogene|T-cell receptor, beta variable region 15, orphan, T-cell receptor, beta variable region 15, orphon|T-cell receptor, beta variable region 15, orphon|TRBV24/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33649131, end_position_on_the_genomic_accession=33649613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6962, CurrentID=0, Status='live', Symbol='TRBV20OR9-2', Aliases='CDR3, TCR, TCRB, TCRBV20S2, TCRBV2O, TCRBV2S2O', description='T cell receptor beta variable 20/OR9-2 (non-functional)', other_designations='T cell receptor beta variable 20/OR9-2 pseudogene|T-cell receptor, beta variable region 2, orphon|TRBV20/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33617845, end_position_on_the_genomic_accession=33618508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6964, CurrentID=0, Status='live', Symbol='TRD', Aliases='TCRD, TCRDV1@, TRD', description='T cell receptor delta locus', other_designations='T-cell antigen receptor, delta polypeptide|T-cell receptor, delta (V,D,J,C)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22422546, end_position_on_the_genomic_accession=22466577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6965, CurrentID=0, Status='live', Symbol='TRG', Aliases='TCRG@, TRG', description='T cell receptor gamma locus', other_designations='T-cell antigen receptor, gamma polypeptide|T-cell rearranging gene, gamma|T-cell receptor, gamma cluster', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38240024, end_position_on_the_genomic_accession=38368055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6966, CurrentID=0, Status='live', Symbol='TRGC1', Aliases='C1, TCRG, TCRGC1', description='T cell receptor gamma constant 1', other_designations='T-cell receptor, gamma, constant region C1', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38260088, end_position_on_the_genomic_accession=38265678, orientation='minus', exon_count=0, OMIM=186970, no_hdr0=''),
  106699575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106699575, CurrentID=0, Status='live', Symbol='DUP14Q32', Aliases='', description='Myeloproliferative neoplasms, familial, susceptibility to', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616604, no_hdr0=''),
  106699576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106699576, CurrentID=0, Status='live', Symbol='RC14S', Aliases='', description='Ring chromosome 14 syndrome', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616606, no_hdr0=''),
  6969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6969, CurrentID=0, Status='live', Symbol='TRGJ2', Aliases='J2, TCRGJ2', description='T cell receptor gamma joining 2', other_designations='T-cell receptor gamma joining segment J2', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38253380, end_position_on_the_genomic_accession=38253429, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6970, CurrentID=0, Status='live', Symbol='TRGJP', Aliases='JP, TCRGJP', description='T cell receptor gamma joining P', other_designations='T-cell receptor, gamma, joining segment JP', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38273587, end_position_on_the_genomic_accession=38273648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6971, CurrentID=0, Status='live', Symbol='TRGJP1', Aliases='JP1, TCRGJP1', description='T cell receptor gamma joining P1', other_designations='T-cell receptor gamma joining P1|T-cell receptor gamma joining segment JP1', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38276259, end_position_on_the_genomic_accession=38276318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6972, CurrentID=0, Status='live', Symbol='TRGJP2', Aliases='JP2, TCRGJP2', description='T cell receptor gamma joining P2', other_designations='T-cell receptor, gamma, joining segment JP2', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38256337, end_position_on_the_genomic_accession=38256396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6973, CurrentID=0, Status='live', Symbol='TRGV1', Aliases='TCRGV1, V1S1P', description='T cell receptor gamma variable 1 (non-functional)', other_designations='T cell receptor gamma variable 1 pseudogene|T-cell receptor, gamma, variable region V1 (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38367586, end_position_on_the_genomic_accession=38368055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  138046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138046, CurrentID=0, Status='live', Symbol='RALYL', Aliases='HNRPCL3', description='RALY RNA binding protein-like', other_designations='heterogeneous nuclear ribonucleoprotein C-like 3|hnRNP core protein C-like 3', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84183218, end_position_on_the_genomic_accession=84921844, orientation='plus', exon_count=15, OMIM=614648, no_hdr0=''),
  6975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6975, CurrentID=0, Status='live', Symbol='TECTB', Aliases='', description='tectorin beta', other_designations='', map_location='10q25-q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112283345, end_position_on_the_genomic_accession=112305035, orientation='plus', exon_count=11, OMIM=602653, no_hdr0=''),
  6976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6976, CurrentID=0, Status='live', Symbol='TRGV3', Aliases='TCRGV3, V1S3', description='T cell receptor gamma variable 3', other_designations='T-cell receptor, gamma, variable region V3', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38358512, end_position_on_the_genomic_accession=38358979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6977, CurrentID=0, Status='live', Symbol='TRGV4', Aliases='TCRGV4, V1S4', description='T cell receptor gamma variable 4', other_designations='T-cell receptor, gamma, variable region V4', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38353715, end_position_on_the_genomic_accession=38354183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  138050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138050, CurrentID=0, Status='live', Symbol='HGSNAT', Aliases='HGNAT, MPS3C, TMEM76', description='heparan-alpha-glucosaminide N-acetyltransferase', other_designations='transmembrane protein 76', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43140449, end_position_on_the_genomic_accession=43202827, orientation='plus', exon_count=20, OMIM=610453, no_hdr0=''),
  6979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6979, CurrentID=0, Status='live', Symbol='TRGV5P', Aliases='TCRGV5P, V1S5P', description='T cell receptor gamma variable 5P (pseudogene)', other_designations='T cell receptor gamma variable 5 pseudogene|T cell receptor gamma variable 5P pseudogene|T-cell receptor, gamma, variable region V5P (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38345030, end_position_on_the_genomic_accession=38345499, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6980, CurrentID=0, Status='live', Symbol='TRGV6', Aliases='TCRGV5P, TCRGV6, V1S5P, V1S6P', description='T cell receptor gamma variable 6 (pseudogene)', other_designations='T cell receptor gamma variable 6 pseudogene|T-cell receptor, gamma, variable region V5P (pseudogene)|T-cell receptor, gamma, variable region V6 (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38340700, end_position_on_the_genomic_accession=38341171, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6981, CurrentID=0, Status='live', Symbol='TRGV7', Aliases='TCRGV7, V1S7P', description='T cell receptor gamma variable 7 (pseudogene)', other_designations='T cell receptor gamma variable 7 pseudogene|T-cell receptor, gamma, variable region V7 (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38335041, end_position_on_the_genomic_accession=38335514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6982, CurrentID=0, Status='live', Symbol='TRGV8', Aliases='TCRGV8, V1S8', description='T cell receptor gamma variable 8', other_designations='T-cell receptor, gamma, variable region V8', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38330343, end_position_on_the_genomic_accession=38330821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6983, CurrentID=0, Status='live', Symbol='TRGV9', Aliases='TCRGV9, TRGC1, V2', description='T cell receptor gamma variable 9', other_designations='T-cell receptor, gamma, variable region V9', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38317017, end_position_on_the_genomic_accession=38317492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6984, CurrentID=0, Status='live', Symbol='TRGV10', Aliases='TCRGV10, V3P', description='T cell receptor gamma variable 10 (non-functional)', other_designations='T cell receptor gamma variable 10 pseudogene|T-cell receptor, gamma, variable region V10', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38299810, end_position_on_the_genomic_accession=38300280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6985, CurrentID=0, Status='live', Symbol='TRGV11', Aliases='TCRGV11, V4P', description='T cell receptor gamma variable 11 (non-functional)', other_designations='T cell receptor gamma variable 11 pseudogene|T-cell receptor, gamma, variable region V11', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38291616, end_position_on_the_genomic_accession=38292078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6986, CurrentID=0, Status='live', Symbol='TRGVA', Aliases='TCRGVA, V5P', description='T cell receptor gamma variable A (pseudogene)', other_designations='T cell receptor gamma variable A pseudogene|T-cell receptor, gamma, variable region VA (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38322446, end_position_on_the_genomic_accession=38322863, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6987, CurrentID=0, Status='live', Symbol='TRGVB', Aliases='TCRGVB, V6P', description='T cell receptor gamma variable B (pseudogene)', other_designations='T cell receptor gamma variable B pseudogene|T-cell receptor, gamma, variable region VB (pseudogene)', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38295763, end_position_on_the_genomic_accession=38296233, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6988, CurrentID=0, Status='live', Symbol='TCTA', Aliases='', description='T-cell leukemia translocation altered', other_designations='T-cell leukemia translocation-associated gene protein', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49412206, end_position_on_the_genomic_accession=49416476, orientation='plus', exon_count=3, OMIM=600690, no_hdr0=''),
  100293090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293090, CurrentID=0, Status='live', Symbol='MTND4P12', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 12', other_designations='MT-ND4 pseudogene 12', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134926661, end_position_on_the_genomic_accession=134928036, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6990, CurrentID=0, Status='live', Symbol='DYNLT3', Aliases='RP3, TCTE1L, TCTEX1L', description='dynein, light chain, Tctex-type 3', other_designations='protein 91/23', map_location='Xp21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37838836, end_position_on_the_genomic_accession=37847636, orientation='minus', exon_count=5, OMIM=300302, no_hdr0=''),
  6991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6991, CurrentID=0, Status='live', Symbol='TCTE3', Aliases='TCTEX1D3', description='t-complex-associated-testis-expressed 3', other_designations='T-complex testis-specific protein 3|T-complex-associated testis-expressed protein 3|Tctex1 domain containing 3|tcte-3', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169725090, end_position_on_the_genomic_accession=169751549, orientation='minus', exon_count=6, OMIM=186977, no_hdr0=''),
  6992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6992, CurrentID=0, Status='live', Symbol='PPP1R11', Aliases='HCG-V, HCGV, IPP3, TCTE5, TCTEX5', description='protein phosphatase 1 regulatory inhibitor subunit 11', other_designations='HCG V|hemochromatosis candidate gene V protein|inhibitor-3|protein phosphatase 1, regulatory (inhibitor) subunit 11|protein phosphatase inhibitor 3|t-complex-associated-testis-expressed 5', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30066880, end_position_on_the_genomic_accession=30070333, orientation='plus', exon_count=4, OMIM=606670, no_hdr0=''),
  6993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6993, CurrentID=0, Status='live', Symbol='DYNLT1', Aliases='CW-1, TCTEL1, tctex-1', description='dynein, light chain, Tctex-type 1', other_designations='T-complex testis-specific protein 1 homolog|t-complex-associated-testis-expressed 1-like 1', map_location='6q25.2-q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158636474, end_position_on_the_genomic_accession=158644786, orientation='minus', exon_count=5, OMIM=601554, no_hdr0=''),
  6996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6996, CurrentID=0, Status='live', Symbol='TDG', Aliases='hTDG', description='thymine DNA glycosylase', other_designations='thymine-DNA glycosylase', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103965815, end_position_on_the_genomic_accession=103988878, orientation='plus', exon_count=10, OMIM=601423, no_hdr0=''),
  6997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6997, CurrentID=0, Status='live', Symbol='TDGF1', Aliases='CR, CRGF, CRIPTO', description='teratocarcinoma-derived growth factor 1', other_designations='cripto-1 growth factor|epidermal growth factor-like cripto protein CR1', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46574555, end_position_on_the_genomic_accession=46582463, orientation='plus', exon_count=7, OMIM=187395, no_hdr0=''),
  6998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6998, CurrentID=0, Status='live', Symbol='TDGF1P3', Aliases='CR-3, CRIPTO, CRIPTO-3, CRIPTO3, TDGF1, TDGF2, TDGF3', description='teratocarcinoma-derived growth factor 1 pseudogene 3', other_designations='Teratocarcinoma-derived growth factor-2|teratocarcinoma-derived growth factor 2 (retroposin, possibly expressed)|teratocarcinoma-derived growth factor 3, pseudogene', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110520312, end_position_on_the_genomic_accession=110523021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  6999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6999, CurrentID=0, Status='live', Symbol='TDO2', Aliases='TDO, TO, TPH2, TRPO', description='tryptophan 2,3-dioxygenase', other_designations='TDO|TO|TRPO|tryptamin 2,3-dioxygenase|tryptophan oxygenase|tryptophan pyrrolase|tryptophanase', map_location='4q31-q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155903693, end_position_on_the_genomic_accession=155920406, orientation='plus', exon_count=12, OMIM=191070, no_hdr0=''),
  7001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7001, CurrentID=0, Status='live', Symbol='PRDX2', Aliases='HEL-S-2a, NKEF-B, NKEFB, PRP, PRX2, PRXII, PTX1, TDPX1, TPX1, TSA', description='peroxiredoxin 2', other_designations='epididymis secretory sperm binding protein Li 2a|natural killer cell-enhancing factor B|thiol-specific antioxidant 1|thioredoxin peroxidase 1|thioredoxin-dependent peroxide reductase 1|torin', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12796820, end_position_on_the_genomic_accession=12801910, orientation='minus', exon_count=6, OMIM=600538, no_hdr0=''),
  7002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7002, CurrentID=0, Status='live', Symbol='PRDX1P1', Aliases='PAGB, TDPX2', description='peroxiredoxin 1 pseudogene 1', other_designations='proliferation-associated gene B|thiol-specific antioxidant 2|thioredoxin-dependent peroxide reductase 2 (thiol-specific antioxidant 2, proliferation-associated gene B)', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12972560, end_position_on_the_genomic_accession=12973506, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7003, CurrentID=0, Status='live', Symbol='TEAD1', Aliases='AA, NTEF-1, REF1, TCF-13, TCF13, TEAD-1, TEF-1', description='TEA domain family member 1 (SV40 transcriptional enhancer factor)', other_designations='protein GT-IIC|transcription factor 13|transcriptional enhancer factor 1', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12674422, end_position_on_the_genomic_accession=12944737, orientation='plus', exon_count=13, OMIM=189967, no_hdr0=''),
  7004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7004, CurrentID=0, Status='live', Symbol='TEAD4', Aliases='EFTR-2, RTEF1, TCF13L1, TEF-3, TEF3, TEFR-1, hRTEF-1B', description='TEA domain family member 4', other_designations='related transcription enhancer factor 1B|transcription factor 13-like 1|transcription factor RTEF-1|transcriptional enhancer factor 1-related|transcriptional enhancer factor 3', map_location='12p13.3-p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2959312, end_position_on_the_genomic_accession=3040676, orientation='plus', exon_count=13, OMIM=601714, no_hdr0=''),
  7005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7005, CurrentID=0, Status='live', Symbol='TEAD3', Aliases='DTEF-1, ETFR-1, TEAD-3, TEAD5, TEF-5, TEF5', description='TEA domain family member 3', other_designations='TEA domain family member 5|transcriptional enhancer factor 5|transcriptional enhancer factor TEF-5 (DTEF-1)', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35473597, end_position_on_the_genomic_accession=35497084, orientation='minus', exon_count=13, OMIM=603170, no_hdr0=''),
  7006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7006, CurrentID=0, Status='live', Symbol='TEC', Aliases='PSCTK4', description='tec protein tyrosine kinase', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48135783, end_position_on_the_genomic_accession=48269797, orientation='minus', exon_count=20, OMIM=600583, no_hdr0=''),
  7007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7007, CurrentID=0, Status='live', Symbol='TECTA', Aliases='DFNA12, DFNA8, DFNB21', description='tectorin alpha', other_designations='', map_location='11q22-q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121102666, end_position_on_the_genomic_accession=121190806, orientation='plus', exon_count=23, OMIM=602574, no_hdr0=''),
  400224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400224, CurrentID=0, Status='live', Symbol='PLEKHD1', Aliases='UPF0639', description='pleckstrin homology and coiled-coil domain containing D1', other_designations='PH domain-containing family D member 1|UPF0639 protein|pleckstrin homology domain containing, family D (with M protein repeats) member 1|pleckstrin homology domain containing, family D (with coiled-coil domains) member 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69484754, end_position_on_the_genomic_accession=69529327, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  7009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7009, CurrentID=0, Status='live', Symbol='TMBIM6', Aliases='BAXI1, BI-1, TEGT', description='transmembrane BAX inhibitor motif containing 6', other_designations='testis enhanced gene transcript|testis-enhanced gene transcript protein|transmembrane BAX inhibitor motif-containing protein 6', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49741510, end_position_on_the_genomic_accession=49764934, orientation='plus', exon_count=11, OMIM=600748, no_hdr0=''),
  7010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7010, CurrentID=0, Status='live', Symbol='TEK', Aliases='CD202B, TIE-2, TIE2, VMCM, VMCM1', description='TEK tyrosine kinase, endothelial', other_designations='endothelial tyrosine kinase|tunica interna endothelial cell kinase|tyrosine kinase with Ig and EGF homology domains-2|tyrosine-protein kinase receptor TEK|tyrosine-protein kinase receptor TIE-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27109141, end_position_on_the_genomic_accession=27230178, orientation='plus', exon_count=23, OMIM=600221, no_hdr0=''),
  7011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7011, CurrentID=0, Status='live', Symbol='TEP1', Aliases='TLP1, TP1, TROVE1, VAULT2, p240', description='telomerase-associated protein 1', other_designations='TROVE domain family, member 1|p80 telomerase homolog|telomerase protein 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20365667, end_position_on_the_genomic_accession=20414645, orientation='minus', exon_count=55, OMIM=601686, no_hdr0=''),
  7012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7012, CurrentID=0, Status='live', Symbol='TERC', Aliases='DKCA1, PFBMFT2, SCARNA19, TR, TRC3, hTR', description='telomerase RNA component', other_designations='small Cajal body-specific RNA 19', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169764610, end_position_on_the_genomic_accession=169765060, orientation='minus', exon_count=1, OMIM=602322, no_hdr0=''),
  7013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7013, CurrentID=0, Status='live', Symbol='TERF1', Aliases='PIN2, TRBF1, TRF, TRF1, hTRF1-AS, t-TRF1', description='telomeric repeat binding factor (NIMA-interacting) 1', other_designations='NIMA-interacting protein 2|TTAGGG repeat-binding factor 1|telomeric protein Pin2/TRF1', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73008862, end_position_on_the_genomic_accession=73048123, orientation='plus', exon_count=12, OMIM=600951, no_hdr0=''),
  7014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7014, CurrentID=0, Status='live', Symbol='TERF2', Aliases='TRBF2, TRF2', description='telomeric repeat binding factor 2', other_designations='TTAGGG repeat-binding factor 2|telomeric DNA-binding protein|telomeric repeat binding protein 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69355561, end_position_on_the_genomic_accession=69386000, orientation='minus', exon_count=11, OMIM=602027, no_hdr0=''),
  7015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7015, CurrentID=0, Status='live', Symbol='TERT', Aliases='CMM9, DKCA2, DKCB4, EST2, PFBMFT1, TCS1, TP2, TRT, hEST2, hTRT', description='telomerase reverse transcriptase', other_designations='telomerase catalytic subunit|telomerase-associated protein 2', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1253167, end_position_on_the_genomic_accession=1295047, orientation='minus', exon_count=18, OMIM=187270, no_hdr0=''),
  7016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7016, CurrentID=0, Status='live', Symbol='TESK1', Aliases='', description='testis-specific kinase 1', other_designations='testicular protein kinase 1', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35605284, end_position_on_the_genomic_accession=35610041, orientation='plus', exon_count=10, OMIM=601782, no_hdr0=''),
  7018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7018, CurrentID=0, Status='live', Symbol='TF', Aliases='HEL-S-71p, PRO1557, PRO2086QTL1, TF', description='transferrin', other_designations='beta-1 metal-binding globulin|epididymis secretory sperm binding protein Li 71p|siderophilin', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133661998, end_position_on_the_genomic_accession=133779006, orientation='plus', exon_count=24, OMIM=190000, no_hdr0=''),
  7019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7019, CurrentID=0, Status='live', Symbol='TFAM', Aliases='MTTF1, MTTFA, TCF6, TCF6L1, TCF6L2, TCF6L3', description='transcription factor A, mitochondrial', other_designations='mitochondrial transcription factor 1|mitochondrial transcription factor A|transcription factor 6|transcription factor 6-like 1|transcription factor 6-like 2 (mitochondrial transcription factor)|transcription factor 6-like 3', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58385143, end_position_on_the_genomic_accession=58399230, orientation='plus', exon_count=9, OMIM=600438, no_hdr0=''),
  7020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7020, CurrentID=0, Status='live', Symbol='TFAP2A', Aliases='AP-2, AP-2alpha, AP2TF, BOFS, TFAP2', description='transcription factor AP-2 alpha (activating enhancer binding protein 2 alpha)', other_designations='AP-2 transcription factor|AP2-alpha|activating enhancer-binding protein 2-alpha|activator protein 2', map_location='6p24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10396677, end_position_on_the_genomic_accession=10419624, orientation='minus', exon_count=12, OMIM=107580, no_hdr0=''),
  7021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7021, CurrentID=0, Status='live', Symbol='TFAP2B', Aliases='AP-2B, AP2-B', description='transcription factor AP-2 beta (activating enhancer binding protein 2 beta)', other_designations='AP2-beta|activating enhancer binding protein 2 beta|activating enhancer-binding protein 2-beta', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50817710, end_position_on_the_genomic_accession=50847619, orientation='plus', exon_count=12, OMIM=601601, no_hdr0=''),
  7022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7022, CurrentID=0, Status='live', Symbol='TFAP2C', Aliases='AP2-GAMMA, ERF1, TFAP2G, hAP-2g', description='transcription factor AP-2 gamma (activating enhancer binding protein 2 gamma)', other_designations='activating enhancer-binding protein 2 gamma|estrogen receptor factor 1|transcription factor AP-2 gamma (activating enhancer-binding protein 2 gamma)|transcription factor ERF-1', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56629302, end_position_on_the_genomic_accession=56639283, orientation='plus', exon_count=7, OMIM=601602, no_hdr0=''),
  7023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7023, CurrentID=0, Status='live', Symbol='TFAP4', Aliases='AP-4, bHLHc41', description='transcription factor AP-4 (activating enhancer binding protein 4)', other_designations='activating enhancer-binding protein 4|class C basic helix-loop-helix protein 41|transcription factor AP-4 (activating enhancer-binding protein 4)', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4257186, end_position_on_the_genomic_accession=4273000, orientation='minus', exon_count=10, OMIM=600743, no_hdr0=''),
  7024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7024, CurrentID=0, Status='live', Symbol='TFCP2', Aliases='LBP1C, LSF, LSF1D, SEFC, TFCP2', description='transcription factor CP2', other_designations='SAA3 enhancer factor', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51093756, end_position_on_the_genomic_accession=51173143, orientation='minus', exon_count=15, OMIM=189889, no_hdr0=''),
  7025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7025, CurrentID=0, Status='live', Symbol='NR2F1', Aliases='BBOAS, BBSOAS, COUP-TFI, EAR-3, EAR3, ERBAL3, NR2F2, SVP44, TCFCOUP1, TFCOUP1', description='nuclear receptor subfamily 2 group F member 1', other_designations='COUP transcription factor I|COUP-TF I|COUP-TF1|V-erbA-related protein 3|chicken ovalbumin upstream promoter-transcription factor I|nuclear receptor subfamily 2, group F, member 1|transcription factor COUP 1 (chicken ovalbumin upstream promoter 1, v-erb-a homolog-like 3)', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93583337, end_position_on_the_genomic_accession=93594611, orientation='plus', exon_count=3, OMIM=132890, no_hdr0=''),
  7026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7026, CurrentID=0, Status='live', Symbol='NR2F2', Aliases='ARP1, CHTD4, COUPTFB, COUPTFII, NF-E3, NR2F1, SVP40, TFCOUP2', description='nuclear receptor subfamily 2 group F member 2', other_designations='ADP-ribosylation factor related protein 1|COUP transcription factor II|apolipoprotein A-I regulatory protein 1|apolipoprotein AI regulatory protein 1|chicken ovalbumin upstream promoter transcription factor 2|chicken ovalbumin upstream promoter-transcription factor I|nuclear receptor subfamily 2, group F, member 2', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96325928, end_position_on_the_genomic_accession=96340263, orientation='plus', exon_count=6, OMIM=107773, no_hdr0=''),
  7027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7027, CurrentID=0, Status='live', Symbol='TFDP1', Aliases='DP1, DRTF1, Dp-1', description='transcription factor Dp-1', other_designations='DRTF1-polypeptide 1|E2F dimerization partner 1|E2F-related transcription factor', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113584688, end_position_on_the_genomic_accession=113641473, orientation='plus', exon_count=15, OMIM=189902, no_hdr0=''),
  7029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7029, CurrentID=0, Status='live', Symbol='TFDP2', Aliases='DP2', description='transcription factor Dp-2 (E2F dimerization partner 2)', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141944428, end_position_on_the_genomic_accession=142149544, orientation='minus', exon_count=19, OMIM=602160, no_hdr0=''),
  7030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7030, CurrentID=0, Status='live', Symbol='TFE3', Aliases='RCCP2, RCCX1, TFEA, bHLHe33', description='transcription factor binding to IGHM enhancer 3', other_designations='class E basic helix-loop-helix protein 33|transcription factor E family, member A|transcription factor for IgH enhancer|transcription factor for immunoglobulin heavy-chain enhancer 3', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49028726, end_position_on_the_genomic_accession=49043517, orientation='minus', exon_count=11, OMIM=314310, no_hdr0=''),
  7031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7031, CurrentID=0, Status='live', Symbol='TFF1', Aliases='BCEI, D21S21, HP1.A, HPS2, pNR-2, pS2', description='trefoil factor 1', other_designations='breast cancer estrogen-inducible protein|breast cancer estrogen-inducible sequence|gastrointestinal trefoil protein pS2|polypeptide P1.A|protein pS2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42362282, end_position_on_the_genomic_accession=42366535, orientation='minus', exon_count=3, OMIM=113710, no_hdr0=''),
  7032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7032, CurrentID=0, Status='live', Symbol='TFF2', Aliases='SML1, SP', description='trefoil factor 2', other_designations='spasmolysin|spasmolytic polypeptide|spasmolytic protein 1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42346357, end_position_on_the_genomic_accession=42351099, orientation='minus', exon_count=4, OMIM=182590, no_hdr0=''),
  7033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7033, CurrentID=0, Status='live', Symbol='TFF3', Aliases='ITF, P1B, TFI', description='trefoil factor 3', other_designations='polypeptide P1.B|trefoil factor 3 (intestinal)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42311667, end_position_on_the_genomic_accession=42315596, orientation='minus', exon_count=3, OMIM=600633, no_hdr0=''),
  7035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7035, CurrentID=0, Status='live', Symbol='TFPI', Aliases='EPI, LACI, TFI1, TFPI', description='tissue factor pathway inhibitor', other_designations='anti-convertin|extrinsic pathway inhibitor|tissue factor pathway inhibitor (lipoprotein-associated coagulation inhibitor)', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187464230, end_position_on_the_genomic_accession=187554494, orientation='minus', exon_count=13, OMIM=152310, no_hdr0=''),
  7036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7036, CurrentID=0, Status='live', Symbol='TFR2', Aliases='HFE3, TFRC2', description='transferrin receptor 2', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100620416, end_position_on_the_genomic_accession=100642765, orientation='minus', exon_count=20, OMIM=604720, no_hdr0=''),
  7037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7037, CurrentID=0, Status='live', Symbol='TFRC', Aliases='CD71, T9, TFR, TFR1, TR, TRFR, p90', description='transferrin receptor', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196049284, end_position_on_the_genomic_accession=196082161, orientation='minus', exon_count=20, OMIM=190010, no_hdr0=''),
  7038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7038, CurrentID=0, Status='live', Symbol='TG', Aliases='AITD3N, TG', description='thyroglobulin', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132866935, end_position_on_the_genomic_accession=133134902, orientation='plus', exon_count=52, OMIM=188450, no_hdr0=''),
  7039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7039, CurrentID=0, Status='live', Symbol='TGFA', Aliases='TFGA', description='transforming growth factor alpha', other_designations='TGF-alpha|transforming growth factor, alpha', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70447280, end_position_on_the_genomic_accession=70554015, orientation='minus', exon_count=7, OMIM=190170, no_hdr0=''),
  7040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7040, CurrentID=0, Status='live', Symbol='TGFB1', Aliases='CED, DPD1, LAP, TGFB, TGFbeta', description='transforming growth factor beta 1', other_designations='TGF-beta-1|latency-associated peptide|prepro-transforming growth factor beta-1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41330531, end_position_on_the_genomic_accession=41353933, orientation='minus', exon_count=7, OMIM=190180, no_hdr0=''),
  7041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7041, CurrentID=0, Status='live', Symbol='TGFB1I1', Aliases='ARA55, HIC-5, HIC5, TSC-5', description='transforming growth factor beta 1 induced transcript 1', other_designations='androgen receptor coactivator 55 kDa protein|androgen receptor coactivator ARA55|androgen receptor-associated protein of 55 kDa|hydrogen peroxide-inducible clone 5 protein|hydrogen peroxide-inducible clone-5', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31472155, end_position_on_the_genomic_accession=31477960, orientation='plus', exon_count=13, OMIM=602353, no_hdr0=''),
  7042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7042, CurrentID=0, Status='live', Symbol='TGFB2', Aliases='LDS4, TGF-beta2', description='transforming growth factor beta 2', other_designations='BSC-1 cell growth inhibitor|G-TSF|cetermin|glioblastoma-derived T-cell suppressor factor|polyergin|prepro-transforming growth factor beta-2|transforming growth factor, beta 2', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218345334, end_position_on_the_genomic_accession=218444619, orientation='plus', exon_count=8, OMIM=190220, no_hdr0=''),
  7043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7043, CurrentID=0, Status='live', Symbol='TGFB3', Aliases='ARVD, ARVD1, RNHF, TGF-beta3', description='transforming growth factor beta 3', other_designations='TGF-beta-3|prepro-transforming growth factor beta-3|transforming growth factor, beta 3', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75958097, end_position_on_the_genomic_accession=75982991, orientation='minus', exon_count=8, OMIM=190230, no_hdr0=''),
  7044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7044, CurrentID=0, Status='live', Symbol='LEFTY2', Aliases='EBAF, LEFTA, LEFTYA, TGFB4', description='left-right determination factor 2', other_designations='TGF-beta-4|endometrial bleeding-associated factor|left-right determination factor A|protein lefty-2|protein lefty-A|transforming growth factor beta-4', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225936598, end_position_on_the_genomic_accession=225941383, orientation='minus', exon_count=6, OMIM=601877, no_hdr0=''),
  7045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7045, CurrentID=0, Status='live', Symbol='TGFBI', Aliases='BIGH3, CDB1, CDG2, CDGG1, CSD, CSD1, CSD2, CSD3, EBMD, LCD1', description='transforming growth factor beta induced', other_designations='RGD-CAP|RGD-containing collagen-associated protein|beta ig-h3|betaig-h3|kerato-epithelin|transforming growth factor beta-induced 68kDa|transforming growth factor, beta-induced, 68kD|transforming growth factor, beta-induced, 68kDa', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136028895, end_position_on_the_genomic_accession=136063818, orientation='plus', exon_count=17, OMIM=601692, no_hdr0=''),
  7046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7046, CurrentID=0, Status='live', Symbol='TGFBR1', Aliases='AAT5, ACVRLK4, ALK-5, ALK5, ESS1, LDS1, LDS1A, LDS2A, MSSE, SKR4, TGFR-1, tbetaR-I', description='transforming growth factor beta receptor I', other_designations='activin A receptor type II-like kinase, 53kDa|activin A receptor type II-like protein kinase of 53kD|activin receptor-like kinase 5|mutant transforming growth factor beta receptor I|serine/threonine-protein kinase receptor R4|transforming growth factor, beta receptor 1|transforming growth factor-beta receptor type I', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99104038, end_position_on_the_genomic_accession=99154192, orientation='plus', exon_count=11, OMIM=190181, no_hdr0=''),
  7047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7047, CurrentID=0, Status='live', Symbol='TGM4', Aliases='TGP, hTGP', description='transglutaminase 4', other_designations='TG(P)|TGase P|TGase-4|fibrinoligase|prostate transglutaminase|prostate-specific transglutaminase|transglutaminase 4 (prostate)|transglutaminase P|transglutaminase-4', map_location='3p22-p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44874592, end_position_on_the_genomic_accession=44914601, orientation='plus', exon_count=15, OMIM=600585, no_hdr0=''),
  7048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7048, CurrentID=0, Status='live', Symbol='TGFBR2', Aliases='AAT3, FAA3, LDS1B, LDS2, LDS2B, MFS2, RIIC, TAAD2, TGFR-2, TGFbeta-RII', description='transforming growth factor beta receptor II', other_designations='TGF-beta receptor type IIB|TGF-beta type II receptor|tbetaR-II|transforming growth factor beta receptor type IIC|transforming growth factor, beta receptor II alpha|transforming growth factor, beta receptor II beta|transforming growth factor, beta receptor II delta|transforming growth factor, beta receptor II epsilon|transforming growth factor, beta receptor II gamma', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30606493, end_position_on_the_genomic_accession=30694142, orientation='plus', exon_count=9, OMIM=190182, no_hdr0=''),
  7049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7049, CurrentID=0, Status='live', Symbol='TGFBR3', Aliases='BGCAN, betaglycan', description='transforming growth factor beta receptor III', other_designations='TGF-beta receptor type 3|TGF-beta receptor type III|TGFR-3|betaglycan proteoglycan', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91680343, end_position_on_the_genomic_accession=91906002, orientation='minus', exon_count=22, OMIM=600742, no_hdr0=''),
  7050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7050, CurrentID=0, Status='live', Symbol='TGIF1', Aliases='HPE4, TGIF', description='TGFB-induced factor homeobox 1', other_designations="5'-TG-3'-interacting factor 1|TALE homeobox TG-interacting factor|transforming growth factor-beta-induced factor", map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3411927, end_position_on_the_genomic_accession=3458411, orientation='plus', exon_count=10, OMIM=602630, no_hdr0=''),
  7051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7051, CurrentID=0, Status='live', Symbol='TGM1', Aliases='ARCI1, ICR2, KTG, LI, LI1, TGASE, TGK', description='transglutaminase 1', other_designations='K polypeptide epidermal type I, protein-glutamine-gamma-glutamyltransferase|TG(K)|TGase K|TGase-1|epidermal TGase|transglutaminase 1 isoform|transglutaminase K|transglutaminase, keratinocyte|transglutaminase-1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24249114, end_position_on_the_genomic_accession=24263210, orientation='minus', exon_count=15, OMIM=190195, no_hdr0=''),
  7052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7052, CurrentID=0, Status='live', Symbol='TGM2', Aliases='G-ALPHA-h, GNAH, HEL-S-45, TG2, TGC', description='transglutaminase 2', other_designations='C polypeptide, protein-glutamine-gamma-glutamyltransferase|TG(C)|TGase C|TGase H|TGase-2|TGase-H|epididymis secretory protein Li 45|protein-glutamine-gamma-glutamyltransferase|tissue transglutaminase|transglutaminase C|transglutaminase H|transglutaminase-2', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38128461, end_position_on_the_genomic_accession=38166509, orientation='minus', exon_count=14, OMIM=190196, no_hdr0=''),
  7053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7053, CurrentID=0, Status='live', Symbol='TGM3', Aliases='TGE', description='transglutaminase 3', other_designations='E polypeptide, protein-glutamine-gamma-glutamyltransferase|TG(E)|TGase E|TGase-3|transglutaminase E|transglutaminase-3', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2295967, end_position_on_the_genomic_accession=2341079, orientation='plus', exon_count=13, OMIM=600238, no_hdr0=''),
  7054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7054, CurrentID=0, Status='live', Symbol='TH', Aliases='DYT14, DYT5b, TYH', description='tyrosine hydroxylase', other_designations='dystonia 14|tyrosine 3-hydroxylase', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2163929, end_position_on_the_genomic_accession=2172137, orientation='minus', exon_count=14, OMIM=191290, no_hdr0=''),
  7055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7055, CurrentID=0, Status='live', Symbol='THAS', Aliases='TAS', description='thoracoabdominal syndrome', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=313850, no_hdr0=''),
  7056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7056, CurrentID=0, Status='live', Symbol='THBD', Aliases='AHUS6, BDCA3, CD141, THPH12, THRM, TM', description='thrombomodulin', other_designations='CD141 antigen|fetomodulin', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23045633, end_position_on_the_genomic_accession=23049664, orientation='minus', exon_count=1, OMIM=188040, no_hdr0=''),
  7057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7057, CurrentID=0, Status='live', Symbol='THBS1', Aliases='THBS, THBS-1, TSP, TSP-1, TSP1', description='thrombospondin 1', other_designations='thrombospondin-1p180|thrombospondin-p50', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39581079, end_position_on_the_genomic_accession=39597467, orientation='plus', exon_count=22, OMIM=188060, no_hdr0=''),
  7058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7058, CurrentID=0, Status='live', Symbol='THBS2', Aliases='TSP2', description='thrombospondin 2', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169215780, end_position_on_the_genomic_accession=169254114, orientation='minus', exon_count=23, OMIM=188061, no_hdr0=''),
  7059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7059, CurrentID=0, Status='live', Symbol='THBS3', Aliases='TSP3', description='thrombospondin 3', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155195584, end_position_on_the_genomic_accession=155209180, orientation='minus', exon_count=26, OMIM=188062, no_hdr0=''),
  7060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7060, CurrentID=0, Status='live', Symbol='THBS4', Aliases='TSP-4, TSP4', description='thrombospondin 4', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79991311, end_position_on_the_genomic_accession=80083637, orientation='plus', exon_count=24, OMIM=600715, no_hdr0=''),
  7062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7062, CurrentID=0, Status='live', Symbol='TCHH', Aliases='THH, THL, TRHY', description='trichohyalin', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152106317, end_position_on_the_genomic_accession=152115454, orientation='minus', exon_count=3, OMIM=190370, no_hdr0=''),
  7063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7063, CurrentID=0, Status='live', Symbol='THM', Aliases='', description='thymoma', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=274230, no_hdr0=''),
  7064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7064, CurrentID=0, Status='live', Symbol='THOP1', Aliases='EP24.15, MEPD_HUMAN, MP78, TOP', description='thimet oligopeptidase 1', other_designations='endopeptidase 24.15', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2785508, end_position_on_the_genomic_accession=2813601, orientation='plus', exon_count=15, OMIM=601117, no_hdr0=''),
  7066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7066, CurrentID=0, Status='live', Symbol='THPO', Aliases='MGDF, MKCSF, ML, MPLLG, THCYT1, TPO', description='thrombopoietin', other_designations='MPL ligand|c-mpl ligand|megakaryocyte colony-stimulating factor|megakaryocyte growth and development factor|megakaryocyte stimulating factor|myeloproliferative leukemia virus oncogene ligand|prepro-thrombopoietin', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184371935, end_position_on_the_genomic_accession=184379688, orientation='minus', exon_count=7, OMIM=600044, no_hdr0=''),
  7067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7067, CurrentID=0, Status='live', Symbol='THRA', Aliases='AR7, CHNG6, EAR7, ERB-T-1, ERBA, ERBA1, NR1A11, THRA2, c-ERBA-1, THRA', description='thyroid hormone receptor, alpha', other_designations='EAR-7|ERBA-related 7|V-erbA-related protein 7|c-erbA-alpha|nuclear receptor subfamily 1 group A member 1|thyroid hormone receptor alpha 1|thyroid hormone receptor, alpha (erythroblastic leukemia viral (v-erb-a) oncogene homolog, avian)|thyroid normone nuclear receptor alpha variant 1|triiodothyronine receptor', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40062193, end_position_on_the_genomic_accession=40093867, orientation='plus', exon_count=11, OMIM=190120, no_hdr0=''),
  7068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7068, CurrentID=0, Status='live', Symbol='THRB', Aliases='C-ERBA-2, C-ERBA-BETA, ERBA2, GRTH, NR1A2, PRTH, THR11, THRB2, THRB', description='thyroid hormone receptor, beta', other_designations='nuclear receptor subfamily 1 group A member 2|oncogene ERBA2|thyroid hormone nuclear receptor beta variant 1|thyroid hormone receptor, beta (erythroblastic leukemia viral (v-erb-a) oncogene homolog 2, avian)', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24117153, end_position_on_the_genomic_accession=24495281, orientation='minus', exon_count=17, OMIM=190160, no_hdr0=''),
  7069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7069, CurrentID=0, Status='live', Symbol='THRSP', Aliases='LPGP1, Lpgp, S14, SPOT14, THRP', description='thyroid hormone responsive', other_designations='SPOT14 homolog|lipogenic protein 1|spot 14 protein|thyroid hormone responsive (SPOT14 homolog, rat)|thyroid hormone responsive SPOT14', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78063861, end_position_on_the_genomic_accession=78068357, orientation='plus', exon_count=2, OMIM=601926, no_hdr0=''),
  7070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7070, CurrentID=0, Status='live', Symbol='THY1', Aliases='CD90, CDw90', description='Thy-1 cell surface antigen', other_designations='Thy-1 T-cell antigen|thy-1 antigen', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119417945, end_position_on_the_genomic_accession=119423536, orientation='minus', exon_count=4, OMIM=188230, no_hdr0=''),
  7071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7071, CurrentID=0, Status='live', Symbol='KLF10', Aliases='EGR-alpha, EGRA, TIEG, TIEG1', description='Kruppel-like factor 10', other_designations='TGFB-inducible early growth response protein 1|early growth response-alpha|transforming growth factor-beta-inducible early growth response protein 1|zinc finger transcription factor TIEG', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102648777, end_position_on_the_genomic_accession=102655902, orientation='minus', exon_count=5, OMIM=601878, no_hdr0=''),
  7072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7072, CurrentID=0, Status='live', Symbol='TIA1', Aliases='TIA-1, WDM', description='TIA1 cytotoxic granule-associated RNA binding protein', other_designations='T-cell-restricted intracellular antigen-1|p40-TIA-1 (containing p15-TIA-1)', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70209444, end_position_on_the_genomic_accession=70248707, orientation='minus', exon_count=16, OMIM=603518, no_hdr0=''),
  7073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7073, CurrentID=0, Status='live', Symbol='TIAL1', Aliases='TCBP, TIAR', description='TIA1 cytotoxic granule-associated RNA binding protein-like 1', other_designations='T-cluster binding protein|TIA-1-related nucleolysin|aging-associated gene 7 protein', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119573465, end_position_on_the_genomic_accession=119597029, orientation='minus', exon_count=15, OMIM=603413, no_hdr0=''),
  7074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7074, CurrentID=0, Status='live', Symbol='TIAM1', Aliases='', description='T-cell lymphoma invasion and metastasis 1', other_designations='TIAM-1|human T-lymphoma invasion and metastasis inducing TIAM1 protein', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31118418, end_position_on_the_genomic_accession=31559977, orientation='minus', exon_count=33, OMIM=600687, no_hdr0=''),
  7075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7075, CurrentID=0, Status='live', Symbol='TIE1', Aliases='JTK14, TIE', description='tyrosine kinase with immunoglobulin-like and EGF-like domains 1', other_designations='tyrosine kinase with immunoglobulin and epidermal growth factor homology domains 1', map_location='1p34-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43300895, end_position_on_the_genomic_accession=43323110, orientation='plus', exon_count=25, OMIM=600222, no_hdr0=''),
  7076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7076, CurrentID=0, Status='live', Symbol='TIMP1', Aliases='CLGI, EPA, EPO, HCI, TIMP', description='TIMP metallopeptidase inhibitor 1', other_designations='TIMP-1|collagenase inhibitor|erythroid potentiating activity|erythroid-potentiating activity|fibroblast collagenase inhibitor|tissue inhibitor of metalloproteinases 1', map_location='Xp11.3-p11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47582291, end_position_on_the_genomic_accession=47586791, orientation='plus', exon_count=6, OMIM=305370, no_hdr0=''),
  7077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7077, CurrentID=0, Status='live', Symbol='TIMP2', Aliases='CSC-21K, DDC8', description='TIMP metallopeptidase inhibitor 2', other_designations='tissue inhibitor of metalloproteinases 2', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78852977, end_position_on_the_genomic_accession=78925390, orientation='minus', exon_count=5, OMIM=188825, no_hdr0=''),
  7078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7078, CurrentID=0, Status='live', Symbol='TIMP3', Aliases='HSMRK222, K222, K222TA2, SFD', description='TIMP metallopeptidase inhibitor 3', other_designations='MIG-5 protein|TIMP-3|protein MIG-5|tissue inhibitor of metalloproteinases 3', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32800816, end_position_on_the_genomic_accession=32863041, orientation='plus', exon_count=5, OMIM=188826, no_hdr0=''),
  7079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7079, CurrentID=0, Status='live', Symbol='TIMP4', Aliases='', description='TIMP metallopeptidase inhibitor 4', other_designations='TIMP-4|tissue inhibitor of metalloproteinase 4|tissue inhibitor of metalloproteinases 4', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12153068, end_position_on_the_genomic_accession=12159351, orientation='minus', exon_count=5, OMIM=601915, no_hdr0=''),
  7080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7080, CurrentID=0, Status='live', Symbol='NKX2-1', Aliases='BCH, BHC, NK-2, NKX2.1, NKX2A, T/EBP, TEBP, TITF1, TTF-1, TTF1', description='NK2 homeobox 1', other_designations='NK-2 homolog A|homeobox protein NK-2 homolog A|thyroid nuclear factor 1|thyroid transcription factor 1|thyroid-specific enhancer-binding protein', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36516397, end_position_on_the_genomic_accession=36520225, orientation='minus', exon_count=3, OMIM=600635, no_hdr0=''),
  100271260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271260, CurrentID=0, Status='live', Symbol='RPS3AP34', Aliases='RPS3A_15_876', description='ribosomal protein S3a pseudogene 34', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12570306, end_position_on_the_genomic_accession=12571154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7082, CurrentID=0, Status='live', Symbol='TJP1', Aliases='ZO-1', description='tight junction protein 1', other_designations='zona occludens 1|zonula occludens 1 protein', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29699367, end_position_on_the_genomic_accession=29968799, orientation='minus', exon_count=35, OMIM=601009, no_hdr0=''),
  7083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7083, CurrentID=0, Status='live', Symbol='TK1', Aliases='TK2', description='thymidine kinase 1, soluble', other_designations='thymidine kinase 1 soluble isoform|thymidine kinase-1', map_location='17q23.2-q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78174079, end_position_on_the_genomic_accession=78187204, orientation='minus', exon_count=7, OMIM=188300, no_hdr0=''),
  7084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7084, CurrentID=0, Status='live', Symbol='TK2', Aliases='MTDPS2, MTTK, SCA31', description='thymidine kinase 2, mitochondrial', other_designations='mt-TK', map_location='16q22-q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66508003, end_position_on_the_genomic_accession=66550412, orientation='minus', exon_count=12, OMIM=188250, no_hdr0=''),
  7085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7085, CurrentID=0, Status='live', Symbol='TKCR', Aliases='TKC', description='torticollis, keloids, cryptorchidism and renal dysplasia', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=314300, no_hdr0=''),
  7086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7086, CurrentID=0, Status='live', Symbol='TKT', Aliases='HEL-S-48, HEL107, TK1, TKT', description='transketolase', other_designations='epididymis luminal protein 107|epididymis secretory protein Li 48', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53224707, end_position_on_the_genomic_accession=53256114, orientation='minus', exon_count=16, OMIM=606781, no_hdr0=''),
  7087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7087, CurrentID=0, Status='live', Symbol='ICAM5', Aliases='TLCN, TLN', description='intercellular adhesion molecule 5', other_designations='ICAM-5|intercellular adhesion molecule 5, telencephalin|telencephalin', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10289435, end_position_on_the_genomic_accession=10296778, orientation='plus', exon_count=12, OMIM=601852, no_hdr0=''),
  7088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7088, CurrentID=0, Status='live', Symbol='TLE1', Aliases='ESG, ESG1, GRG1', description='transducin like enhancer of split 1', other_designations='enhancer of split groucho-like protein 1|transducin-like enhancer of split 1 (E(sp1) homolog, Drosophila)', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81583683, end_position_on_the_genomic_accession=81689535, orientation='minus', exon_count=22, OMIM=600189, no_hdr0=''),
  7089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7089, CurrentID=0, Status='live', Symbol='TLE2', Aliases='ESG, ESG2, GRG2', description='transducin like enhancer of split 2', other_designations='enhancer of split groucho-like protein 2|transducin-like enhancer of split 2 (E(sp1) homolog, Drosophila)|transducin-like enhancer of split 2, homolog of Drosophila E(sp1)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2997638, end_position_on_the_genomic_accession=3047635, orientation='minus', exon_count=23, OMIM=601041, no_hdr0=''),
  7090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7090, CurrentID=0, Status='live', Symbol='TLE3', Aliases='ESG, ESG3, GRG3, HsT18976', description='transducin like enhancer of split 3', other_designations='enhancer of split groucho 3|enhancer of split groucho-like protein 3|transducin-like enhancer of split 3 (E(sp1) homolog, Drosophila)|transducin-like enhancer of split 3 short isoform|transducin-like enhancer of split 3, homolog of Drosophila E(sp1)', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70047790, end_position_on_the_genomic_accession=70097935, orientation='minus', exon_count=23, OMIM=600190, no_hdr0=''),
  7091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7091, CurrentID=0, Status='live', Symbol='TLE4', Aliases='BCE-1, BCE1, E(spI), ESG, ESG4, GRG4, Grg-4', description='transducin like enhancer of split 4', other_designations='B lymphocyte gene 1|enhancer of split groucho 4|groucho-related protein 4|transducin-like enhancer of split 4 (E(sp1) homolog, Drosophila)|transducin-like enhancer of split 4, homolog of Drosophila E(sp1)', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79571773, end_position_on_the_genomic_accession=79726882, orientation='plus', exon_count=29, OMIM=605132, no_hdr0=''),
  7092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7092, CurrentID=0, Status='live', Symbol='TLL1', Aliases='ASD6, TLL', description='tolloid like 1', other_designations='tolloid-like 1', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165873238, end_position_on_the_genomic_accession=166105178, orientation='plus', exon_count=24, OMIM=606742, no_hdr0=''),
  7093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7093, CurrentID=0, Status='live', Symbol='TLL2', Aliases='', description='tolloid like 2', other_designations='tolloid-like 2', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96364606, end_position_on_the_genomic_accession=96513926, orientation='minus', exon_count=21, OMIM=606743, no_hdr0=''),
  7094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7094, CurrentID=0, Status='live', Symbol='TLN1', Aliases='ILWEQ, TLN', description='talin 1', other_designations='', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35697337, end_position_on_the_genomic_accession=35732395, orientation='minus', exon_count=57, OMIM=186745, no_hdr0=''),
  7095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7095, CurrentID=0, Status='live', Symbol='SEC62', Aliases='Dtrp1, HTP1, TLOC1, TP-1', description='SEC62 homolog, preprotein translocation factor', other_designations='SEC62 homolog|SEC62 preprotein translocation factor|hTP-1|membrane protein SEC62, S.cerevisiae, homolog of|translocation protein 1', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169966792, end_position_on_the_genomic_accession=169998373, orientation='plus', exon_count=9, OMIM=602173, no_hdr0=''),
  7096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7096, CurrentID=0, Status='live', Symbol='TLR1', Aliases='CD281, TIL, TIL. LPRS5, rsc786', description='toll-like receptor 1', other_designations='toll/interleukin-1 receptor-like protein', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38787555, end_position_on_the_genomic_accession=38805800, orientation='minus', exon_count=8, OMIM=601194, no_hdr0=''),
  7097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7097, CurrentID=0, Status='live', Symbol='TLR2', Aliases='CD282, TIL4', description='toll-like receptor 2', other_designations='toll/interleukin 1 receptor-like 4|toll/interleukin-1 receptor-like protein 4', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153684256, end_position_on_the_genomic_accession=153706091, orientation='plus', exon_count=4, OMIM=603028, no_hdr0=''),
  7098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7098, CurrentID=0, Status='live', Symbol='TLR3', Aliases='CD283, IIAE2', description='toll-like receptor 3', other_designations='', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186069155, end_position_on_the_genomic_accession=186085098, orientation='plus', exon_count=5, OMIM=603029, no_hdr0=''),
  7099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7099, CurrentID=0, Status='live', Symbol='TLR4', Aliases='ARMD10, CD284, TLR-4, TOLL', description='toll-like receptor 4', other_designations='hToll|homolog of Drosophila toll', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117704175, end_position_on_the_genomic_accession=117717491, orientation='plus', exon_count=4, OMIM=603030, no_hdr0=''),
  7100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7100, CurrentID=0, Status='live', Symbol='TLR5', Aliases='MELIOS, SLE1, SLEB1, TIL3', description='toll-like receptor 5', other_designations='systemic lupus erythematosus susceptibility 1|toll/interleukin-1 receptor-like protein 3', map_location='1q41-q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223109404, end_position_on_the_genomic_accession=223143282, orientation='minus', exon_count=7, OMIM=603031, no_hdr0=''),
  7101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7101, CurrentID=0, Status='live', Symbol='NR2E1', Aliases='TLL, TLX, XTLL', description='nuclear receptor subfamily 2 group E member 1', other_designations='nuclear receptor TLX|nuclear receptor subfamily 2, group E, member 1|protein tailless homolog|tailes-related receptor', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108158115, end_position_on_the_genomic_accession=108188809, orientation='plus', exon_count=11, OMIM=603849, no_hdr0=''),
  7102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7102, CurrentID=0, Status='live', Symbol='TSPAN7', Aliases='A15, CCG-B7, CD231, DXS1692E, MRX58, MXS1, TALLA-1, TM4SF2, TM4SF2b', description='tetraspanin 7', other_designations='CD231 antigen|T-cell acute lymphoblastic leukemia associated antigen 1|T-cell acute lymphoblastic leukemia-associated antigen 1|cell surface glycoprotein A15|membrane component chromosome X surface marker 1|membrane component, X chromosome, surface marker 1|tetraspanin protein|transmembrane 4 superfamily 2b|transmembrane 4 superfamily member 2|transmembrane protein A15|tspan-7', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38561478, end_position_on_the_genomic_accession=38688918, orientation='plus', exon_count=8, OMIM=300096, no_hdr0=''),
  7103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7103, CurrentID=0, Status='live', Symbol='TSPAN8', Aliases='CO-029, TM4SF3', description='tetraspanin 8', other_designations='transmembrane 4 superfamily member 3|tspan-8|tumor-associated antigen CO-029', map_location='12q14.1-q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71125093, end_position_on_the_genomic_accession=71157999, orientation='minus', exon_count=10, OMIM=600769, no_hdr0=''),
  7104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7104, CurrentID=0, Status='live', Symbol='TM4SF4', Aliases='ILTMP, il-TMP', description='transmembrane 4 L six family member 4', other_designations='intestinal and liver (il) tetraspan membrane protein|intestine and liver tetraspan membrane protein|transmembrane 4 superfamily member 4', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149474581, end_position_on_the_genomic_accession=149503394, orientation='plus', exon_count=5, OMIM=606567, no_hdr0=''),
  7105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7105, CurrentID=0, Status='live', Symbol='TSPAN6', Aliases='T245, TM4SF6, TSPAN-6', description='tetraspanin 6', other_designations='A15 homolog|putative NF-kappa-B-activating protein 321|tetraspan TM4SF|tetraspanin TM4-D|transmembrane 4 superfamily member 6', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100627108, end_position_on_the_genomic_accession=100637104, orientation='minus', exon_count=9, OMIM=300191, no_hdr0=''),
  7106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7106, CurrentID=0, Status='live', Symbol='TSPAN4', Aliases='NAG-2, NAG2, TETRASPAN, TM4SF7, TSPAN-4', description='tetraspanin 4', other_designations='novel antigen 2|tetraspan TM4SF|transmembrane 4 superfamily member 7', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=842824, end_position_on_the_genomic_accession=867116, orientation='plus', exon_count=13, OMIM=602644, no_hdr0=''),
  7107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7107, CurrentID=0, Status='live', Symbol='GPR137B', Aliases='TM7SF1', description='G protein-coupled receptor 137B', other_designations='transmembrane 7 superfamily member 1 (upregulated in kidney)|transmembrane 7 superfamily member 1 protein', map_location='1q42-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236142502, end_position_on_the_genomic_accession=236208909, orientation='plus', exon_count=8, OMIM=604658, no_hdr0=''),
  7108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7108, CurrentID=0, Status='live', Symbol='TM7SF2', Aliases='ANG1, DHCR14A, NET47', description='transmembrane 7 superfamily member 2', other_designations='C-14 sterol reductase|another new gene 1 protein|delta-14-SR|putative sterol reductase SR-1|sterol C14-reductase', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65111854, end_position_on_the_genomic_accession=65116235, orientation='plus', exon_count=10, OMIM=603414, no_hdr0=''),
  7109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7109, CurrentID=0, Status='live', Symbol='TRAPPC10', Aliases='EHOC-1, EHOC1, GT334, TMEM1, TRS130, TRS30', description='trafficking protein particle complex 10', other_designations='TRAPP 130 kDa subunit|TRAPP subunit TMEM1|epilepsy holoprosencephaly candidate 1 protein|epilepsy holoprosencephaly candidate-1 protein|trafficking protein particle complex subunit 130|trafficking protein particle complex subunit TMEM1|transmembrane protein 1|transport protein particle subunit TMEM1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44012325, end_position_on_the_genomic_accession=44106552, orientation='plus', exon_count=28, OMIM=602103, no_hdr0=''),
  7110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7110, CurrentID=0, Status='live', Symbol='TMF1', Aliases='ARA160, TMF', description='TATA element modulatory factor 1', other_designations='androgen receptor coactivator 160 kDa protein|androgen receptor-associated protein of 160 kDa', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69019827, end_position_on_the_genomic_accession=69052333, orientation='minus', exon_count=17, OMIM=601126, no_hdr0=''),
  7111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7111, CurrentID=0, Status='live', Symbol='TMOD1', Aliases='D9S57E, ETMOD, TMOD', description='tropomodulin 1', other_designations='E-Tmod|e-tropomodulin|erythrocyte tropomodulin', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97501180, end_position_on_the_genomic_accession=97601743, orientation='plus', exon_count=11, OMIM=190930, no_hdr0=''),
  7112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7112, CurrentID=0, Status='live', Symbol='TMPO', Aliases='CMD1T, LAP2, LEMD4, PRO0868, TP', description='thymopoietin', other_designations='LEM domain containing 4|TP alpha|TP beta/gamma|lamina-associated polypeptide 2|thymopoietin-related peptide isoform alpha|thymopoietin-related peptide isoforms beta/gamma', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98515573, end_position_on_the_genomic_accession=98550379, orientation='plus', exon_count=10, OMIM=188380, no_hdr0=''),
  7113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7113, CurrentID=0, Status='live', Symbol='TMPRSS2', Aliases='PP9284, PRSS10', description='transmembrane protease, serine 2', other_designations='epitheliasin|serine protease 10', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41464551, end_position_on_the_genomic_accession=41508158, orientation='minus', exon_count=17, OMIM=602060, no_hdr0=''),
  7114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7114, CurrentID=0, Status='live', Symbol='TMSB4X', Aliases='FX, PTMB4, TB4X, TMSB4', description='thymosin beta 4, X-linked', other_designations='prothymosin beta-4|t beta-4|thymosin, beta 4, X chromosome', map_location='Xq21.3-q22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12975107, end_position_on_the_genomic_accession=12977227, orientation='plus', exon_count=3, OMIM=300159, no_hdr0=''),
  285175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285175, CurrentID=0, Status='live', Symbol='UNC80', Aliases='C2orf21, UNC-80', description='unc-80 homolog, NALCN activator', other_designations='WUGSC:H_NH0436C12.1|unc-80 homolog', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209771993, end_position_on_the_genomic_accession=209999300, orientation='plus', exon_count=66, OMIM=612636, no_hdr0=''),
  7116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7116, CurrentID=0, Status='live', Symbol='TMSB4XP2', Aliases='TMSL2', description='thymosin beta 4, X-linked pseudogene 2', other_designations='thymosin-like 2 (pseudogene)', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3617138, end_position_on_the_genomic_accession=3617760, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7117, CurrentID=0, Status='live', Symbol='TMSB4XP8', Aliases='TMSL3', description='thymosin beta 4, X-linked pseudogene 8', other_designations='thymosin beta-4-like protein 3|thymosin-like 3', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=90838485, end_position_on_the_genomic_accession=90839115, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7118, CurrentID=0, Status='live', Symbol='TMSB4XP4', Aliases='TMSL4', description='thymosin beta 4, X-linked pseudogene 4', other_designations='thymosin-like 4 (pseudogene)', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128342154, end_position_on_the_genomic_accession=128342774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7119, CurrentID=0, Status='live', Symbol='TMSB4XP5', Aliases='TMSL5', description='thymosin beta 4, X-linked pseudogene 5', other_designations='thymosin-like 5 (pseudogene)', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63520431, end_position_on_the_genomic_accession=63520952, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7120, CurrentID=0, Status='live', Symbol='TMSB4XP6', Aliases='TMSB4L, TMSL6', description='thymosin beta 4, X-linked pseudogene 6', other_designations='thymosin-like 6 (pseudogene)', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50840211, end_position_on_the_genomic_accession=50840827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7121, CurrentID=0, Status='live', Symbol='TMSB4XP7', Aliases='TMSL7', description='thymosin beta 4, X-linked pseudogene 7', other_designations='thymosin-like 7 (pseudogene)', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115132913, end_position_on_the_genomic_accession=115133499, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7122, CurrentID=0, Status='live', Symbol='CLDN5', Aliases='AWAL, BEC1, CPETRL1, TMVCF', description='claudin 5', other_designations='TMDVCF|transmembrane protein deleted in VCFS|transmembrane protein deleted in velocardiofacial syndrome', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19523024, end_position_on_the_genomic_accession=19525337, orientation='minus', exon_count=1, OMIM=602101, no_hdr0=''),
  7123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7123, CurrentID=0, Status='live', Symbol='CLEC3B', Aliases='TN, TNA', description='C-type lectin domain family 3 member B', other_designations='C-type lectin domain family 3, member B|plasminogen kringle 4-binding protein|tetranectin (plasminogen-binding protein)', map_location='3p22-p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45026267, end_position_on_the_genomic_accession=45036071, orientation='plus', exon_count=4, OMIM=187520, no_hdr0=''),
  7124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7124, CurrentID=0, Status='live', Symbol='TNF', Aliases='DIF-alpha, TNFA, TNFSF2, TNF', description='tumor necrosis factor', other_designations='APC1 protein|TNF, macrophage-derived|TNF, monocyte-derived|TNF-a|cachectin|tumor necrosis factor ligand superfamily member 2|tumor necrosis factor-alpha', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31575567, end_position_on_the_genomic_accession=31578336, orientation='plus', exon_count=4, OMIM=191160, no_hdr0=''),
  7125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7125, CurrentID=0, Status='live', Symbol='TNNC2', Aliases='', description='troponin C type 2 (fast)', other_designations='troponin C2, fast', map_location='20q12-q13.11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45823214, end_position_on_the_genomic_accession=45833323, orientation='minus', exon_count=7, OMIM=191039, no_hdr0=''),
  7126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7126, CurrentID=0, Status='live', Symbol='TNFAIP1', Aliases='B12, B61, BTBD34, EDP1, hBACURD2', description='TNF alpha induced protein 1', other_designations='BTB/POZ domain-containing protein TNFAIP1|tumor necrosis factor, alpha induced protein 1|tumor necrosis factor, alpha-induced protein 1 (endothelial)', map_location='17q22-q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28335522, end_position_on_the_genomic_accession=28347009, orientation='plus', exon_count=7, OMIM=191161, no_hdr0=''),
  7127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7127, CurrentID=0, Status='live', Symbol='TNFAIP2', Aliases='B94, EXOC3L3', description='TNF alpha induced protein 2', other_designations='TNF alpha-induced protein 2|exocyst complex component 3-like 3|primary response gene B94 protein|tumor necrosis factor, alpha induced protein 2|tumor necrosis factor, alpha-induced protein 2', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103121147, end_position_on_the_genomic_accession=103137439, orientation='plus', exon_count=15, OMIM=603300, no_hdr0=''),
  7128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7128, CurrentID=0, Status='live', Symbol='TNFAIP3', Aliases='A20, OTUD7C, TNFA1P2', description='TNF alpha induced protein 3', other_designations='OTU domain-containing protein 7C|TNF alpha-induced protein 3|putative DNA-binding protein A20|tumor necrosis factor inducible protein A20|tumor necrosis factor, alpha induced protein 3|tumor necrosis factor, alpha-induced protein 3|zinc finger protein A20', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137866317, end_position_on_the_genomic_accession=137883314, orientation='plus', exon_count=12, OMIM=191163, no_hdr0=''),
  7130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7130, CurrentID=0, Status='live', Symbol='TNFAIP6', Aliases='TSG-6, TSG6', description='TNF alpha induced protein 6', other_designations='TNF alpha-induced protein 6|TNF-stimulated gene 6 protein|hyaluronate-binding protein|tumor necrosis factor alpha-inducible protein 6|tumor necrosis factor, alpha induced protein 6|tumor necrosis factor, alpha-induced protein 6|tumor necrosis factor-stimulated gene-6 protein', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151357583, end_position_on_the_genomic_accession=151381245, orientation='plus', exon_count=7, OMIM=600410, no_hdr0=''),
  400347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400347, CurrentID=0, Status='live', Symbol='LOC400347', Aliases='', description="REX4 homolog, 3'-5' exonuclease pseudogene", other_designations='REX4, RNA exonuclease 4 homolog (S. cerevisiae) pseudogene', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31453198, end_position_on_the_genomic_accession=31454138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7132, CurrentID=0, Status='live', Symbol='TNFRSF1A', Aliases='CD120a, FPF, MS5, TBP1, TNF-R, TNF-R-I, TNF-R55, TNFAR, TNFR1, TNFR1-d2, TNFR55, TNFR60, p55, p55-R, p60', description='tumor necrosis factor receptor superfamily member 1A', other_designations='TNF-R1|TNF-RI|TNFR-I|tumor necrosis factor binding protein 1|tumor necrosis factor receptor 1A isoform beta|tumor necrosis factor receptor superfamily, member 1A|tumor necrosis factor receptor type 1|tumor necrosis factor-alpha receptor', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6328757, end_position_on_the_genomic_accession=6342117, orientation='minus', exon_count=10, OMIM=191190, no_hdr0=''),
  7133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7133, CurrentID=0, Status='live', Symbol='TNFRSF1B', Aliases='CD120b, TBPII, TNF-R-II, TNF-R75, TNFBR, TNFR1B, TNFR2, TNFR80, p75, p75TNFR', description='tumor necrosis factor receptor superfamily member 1B', other_designations='TNF-R2|TNF-RII|p75 TNF receptor|p80 TNF-alpha receptor|soluble TNFR1B variant 1|tumor necrosis factor beta receptor|tumor necrosis factor binding protein 2|tumor necrosis factor receptor 2|tumor necrosis factor receptor superfamily, member 1B|tumor necrosis factor receptor type II', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12166949, end_position_on_the_genomic_accession=12209222, orientation='plus', exon_count=12, OMIM=191191, no_hdr0=''),
  7134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7134, CurrentID=0, Status='live', Symbol='TNNC1', Aliases='CMD1Z, CMH13, TN-C, TNC, TNNC', description='troponin C type 1 (slow)', other_designations='cardiac troponin C|slow twitch skeletal/cardiac muscle troponin C|troponin C1, slow', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52451091, end_position_on_the_genomic_accession=52454041, orientation='minus', exon_count=6, OMIM=191040, no_hdr0=''),
  7135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7135, CurrentID=0, Status='live', Symbol='TNNI1', Aliases='SSTNI, TNN1', description='troponin I type 1 (skeletal, slow)', other_designations='troponin I, slow-twitch isoform', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201403767, end_position_on_the_genomic_accession=201421746, orientation='minus', exon_count=9, OMIM=191042, no_hdr0=''),
  7136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7136, CurrentID=0, Status='live', Symbol='TNNI2', Aliases='AMCD2B, DA2B, FSSV, fsTnI', description='troponin I type 2 (skeletal, fast)', other_designations='troponin I fast twitch 2|troponin I, fast-twitch isoform|troponin I, fast-twitch skeletal muscle isoform|troponin I, skeletal, fast', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1839003, end_position_on_the_genomic_accession=1841680, orientation='plus', exon_count=10, OMIM=191043, no_hdr0=''),
  7137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7137, CurrentID=0, Status='live', Symbol='TNNI3', Aliases='CMD1FF, CMD2A, CMH7, RCM1, TNNC1, cTnI', description='troponin I type 3 (cardiac)', other_designations='cardiomyopathy, dilated 2A (autosomal recessive)', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55151767, end_position_on_the_genomic_accession=55157732, orientation='minus', exon_count=8, OMIM=191044, no_hdr0=''),
  7138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7138, CurrentID=0, Status='live', Symbol='TNNT1', Aliases='ANM, NEM5, STNT, TNT, TNTS', description='troponin T type 1 (skeletal, slow)', other_designations='slow skeletal muscle troponin T|troponin-T1, skeletal, slow', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55132698, end_position_on_the_genomic_accession=55149354, orientation='minus', exon_count=15, OMIM=191041, no_hdr0=''),
  7139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7139, CurrentID=0, Status='live', Symbol='TNNT2', Aliases='CMD1D, CMH2, CMPD2, LVNC6, RCM3, TnTC, cTnT', description='troponin T type 2 (cardiac)', other_designations='cardiomyopathy, dilated 1D (autosomal dominant)|cardiomyopathy, hypertrophic 2|troponin T2, cardiac|truncated cardiac troponin T', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201359008, end_position_on_the_genomic_accession=201377765, orientation='minus', exon_count=18, OMIM=191045, no_hdr0=''),
  7140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7140, CurrentID=0, Status='live', Symbol='TNNT3', Aliases='TNTF', description='troponin T type 3 (skeletal, fast)', other_designations='beta TnTF|beta-TnTF|fTnT|fast skeletal muscle troponin T|troponin-T3, skeletal, fast', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1919562, end_position_on_the_genomic_accession=1938706, orientation='plus', exon_count=21, OMIM=600692, no_hdr0=''),
  7141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7141, CurrentID=0, Status='live', Symbol='TNP1', Aliases='TP1', description='transition protein 1', other_designations='STP-1|TP-1|transition protein 1 (during histone to protamine replacement)', map_location='2q35-q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216859458, end_position_on_the_genomic_accession=216860059, orientation='minus', exon_count=2, OMIM=190231, no_hdr0=''),
  7142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7142, CurrentID=0, Status='live', Symbol='TNP2', Aliases='TP2', description='transition protein 2 (during histone to protamine replacement)', other_designations='TP-2', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11267857, end_position_on_the_genomic_accession=11269303, orientation='minus', exon_count=2, OMIM=190232, no_hdr0=''),
  7143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7143, CurrentID=0, Status='live', Symbol='TNR', Aliases='TN-R', description='tenascin R', other_designations='janusin|restrictin', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175322799, end_position_on_the_genomic_accession=175743702, orientation='minus', exon_count=25, OMIM=601995, no_hdr0=''),
  400360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400360, CurrentID=0, Status='live', Symbol='C15orf54', Aliases='', description='chromosome 15 open reading frame 54', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39250669, end_position_on_the_genomic_accession=39254845, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7145, CurrentID=0, Status='live', Symbol='TNS1', Aliases='MST091, MST122, MST127, MSTP091, MSTP122, MSTP127, MXRA6, PPP1R155, TNS', description='tensin 1', other_designations='Matrix-remodelling-associated protein 6|matrix-remodelling associated 6|protein phosphatase 1, regulatory subunit 155', map_location='2q35-q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217799789, end_position_on_the_genomic_accession=217978827, orientation='minus', exon_count=41, OMIM=600076, no_hdr0=''),
  7146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7146, CurrentID=0, Status='live', Symbol='TNXA', Aliases='D6S103E, HXBL, TNX, XA', description='tenascin XA (pseudogene)', other_designations='tenascin XA pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32008420, end_position_on_the_genomic_accession=32013023, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100271271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271271, CurrentID=0, Status='live', Symbol='RPL11P3', Aliases='RPL11_2_1077', description='ribosomal protein L11 pseudogene 3', other_designations='', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87945480, end_position_on_the_genomic_accession=87946080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7148, CurrentID=0, Status='live', Symbol='TNXB', Aliases='EDS3, HXBL, TENX, TN-X, TNX1, TNXB2, TNXBS, VUR8, XB, XBS, TNXB', description='tenascin XB', other_designations='growth-inhibiting protein 45|hexabrachion-like protein|tenascin XB1|tenascin XB2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32041155, end_position_on_the_genomic_accession=32109374, orientation='minus', exon_count=44, OMIM=600985, no_hdr0=''),
  7150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7150, CurrentID=0, Status='live', Symbol='TOP1', Aliases='TOPI', description='topoisomerase (DNA) I', other_designations='type I DNA topoisomerase', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41028822, end_position_on_the_genomic_accession=41124487, orientation='plus', exon_count=23, OMIM=126420, no_hdr0=''),
  7151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7151, CurrentID=0, Status='live', Symbol='TOP1P1', Aliases='', description='topoisomerase (DNA) I pseudogene 1', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171338896, end_position_on_the_genomic_accession=171341326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7152, CurrentID=0, Status='live', Symbol='TOP1P2', Aliases='', description='topoisomerase (DNA) I pseudogene 2', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24764501, end_position_on_the_genomic_accession=24766021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7153, CurrentID=0, Status='live', Symbol='TOP2A', Aliases='TOP2, TP2A', description='topoisomerase (DNA) II alpha', other_designations='DNA gyrase|DNA topoisomerase (ATP-hydrolyzing)|DNA topoisomerase II, 170 kD|DNA topoisomerase II, alpha isozyme|topoisomerase (DNA) II alpha 170kDa', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40388521, end_position_on_the_genomic_accession=40417950, orientation='minus', exon_count=36, OMIM=126430, no_hdr0=''),
  7154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7154, CurrentID=0, Status='live', Symbol='TRA-TGC7-1', Aliases='TRAN, TRNAA1', description='transfer RNA-Ala (TGC) 7-1', other_designations='tRNA alanine (anticodon UGC)|tRNA-Ala-TGC-7-1|transfer RNA alanine (anticodon UGC)|transfer RNA alanine 1 (anticodon UGC)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28802800, end_position_on_the_genomic_accession=28802870, orientation='minus', exon_count=0, OMIM=601431, no_hdr0=''),
  7155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7155, CurrentID=0, Status='live', Symbol='TOP2B', Aliases='TOPIIB, top2beta', description='topoisomerase (DNA) II beta', other_designations='DNA topoisomerase II beta|DNA topoisomerase II, 180 kD|DNA topoisomerase II, beta isozyme|U937 associated antigen|antigen MLAA-44|topo II beta|topoisomerase (DNA) II beta 180kDa|topoisomerase II beta|topoisomerase IIb', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25597905, end_position_on_the_genomic_accession=25664905, orientation='minus', exon_count=36, OMIM=126431, no_hdr0=''),
  7156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7156, CurrentID=0, Status='live', Symbol='TOP3A', Aliases='TOP3, ZGRF7', description='topoisomerase (DNA) III alpha', other_designations='DNA topoisomerase III alpha|topo III-alpha|zinc finger, GRF-type containing 7', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18273921, end_position_on_the_genomic_accession=18315007, orientation='minus', exon_count=19, OMIM=601243, no_hdr0=''),
  7157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7157, CurrentID=0, Status='live', Symbol='TP53', Aliases='BCC7, LFS1, P53, TRP53', description='tumor protein p53', other_designations='antigen NY-CO-13|mutant tumor protein 53|p53 tumor suppressor|phosphoprotein p53|transformation-related protein 53|tumor protein 53|tumor supressor p53', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7668402, end_position_on_the_genomic_accession=7687550, orientation='minus', exon_count=12, OMIM=191170, no_hdr0=''),
  7158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7158, CurrentID=0, Status='live', Symbol='TP53BP1', Aliases='53BP1, TP53, p202, p53BP1', description='tumor protein p53 binding protein 1', other_designations='p53-binding protein 1|tumor protein 53-binding protein, 1|tumor protein p53-binding protein, 1', map_location='15q15-q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43403064, end_position_on_the_genomic_accession=43510728, orientation='minus', exon_count=33, OMIM=605230, no_hdr0=''),
  7159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7159, CurrentID=0, Status='live', Symbol='TP53BP2', Aliases='53BP2, ASPP2, BBP, P53BP2, PPP1R13A', description='tumor protein p53 binding protein 2', other_designations='BCL2-binding protein|apoptosis-stimulating protein of p53, 2|renal carcinoma antigen NY-REN-51|tumor protein p53 binding protein, 2|tumor suppressor p53-binding protein 2', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223779893, end_position_on_the_genomic_accession=223845981, orientation='minus', exon_count=20, OMIM=602143, no_hdr0=''),
  7161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7161, CurrentID=0, Status='live', Symbol='TP73', Aliases='P73', description='tumor protein p73', other_designations='p53-like transcription factor|p53-related protein', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3652565, end_position_on_the_genomic_accession=3736201, orientation='plus', exon_count=17, OMIM=601990, no_hdr0=''),
  7162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7162, CurrentID=0, Status='live', Symbol='TPBG', Aliases='5T4, 5T4AG, M6P1, WAIF1', description='trophoblast glycoprotein', other_designations='5T4 oncofetal antigen|5T4 oncofetal trophoblast glycoprotein|5T4 oncotrophoblast glycoprotein|wnt-activated inhibitory factor 1', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82363206, end_position_on_the_genomic_accession=82367417, orientation='plus', exon_count=4, OMIM=190920, no_hdr0=''),
  7163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7163, CurrentID=0, Status='live', Symbol='TPD52', Aliases='D52, N8L, PC-1, PrLZ, hD52', description='tumor protein D52', other_designations='prostate and colon associated protein|prostate leucine zipper|protein N8', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80034868, end_position_on_the_genomic_accession=80171659, orientation='minus', exon_count=12, OMIM=604068, no_hdr0=''),
  7164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7164, CurrentID=0, Status='live', Symbol='TPD52L1', Aliases='D53, hD53', description='tumor protein D52-like 1', other_designations='', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125153729, end_position_on_the_genomic_accession=125263498, orientation='plus', exon_count=11, OMIM=604069, no_hdr0=''),
  7165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7165, CurrentID=0, Status='live', Symbol='TPD52L2', Aliases='D54, TPD54', description='tumor protein D52-like 2', other_designations='HCCR-binding protein 2', map_location='20q13.2-q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63865228, end_position_on_the_genomic_accession=63891545, orientation='plus', exon_count=10, OMIM=603747, no_hdr0=''),
  7166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7166, CurrentID=0, Status='live', Symbol='TPH1', Aliases='TPRH, TRPH', description='tryptophan hydroxylase 1', other_designations='L-tryptophan hydroxylase|indoleacetic acid-5-hydroxylase|tryptophan 5-monooxygenase 1|tryptophan hydroxylase (tryptophan 5-monooxygenase)', map_location='11p15.3-p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18020537, end_position_on_the_genomic_accession=18040788, orientation='minus', exon_count=10, OMIM=191060, no_hdr0=''),
  7167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7167, CurrentID=0, Status='live', Symbol='TPI1', Aliases='HEL-S-49, TIM, TPI, TPID', description='triosephosphate isomerase 1', other_designations='epididymis secretory protein Li 49|triose-phosphate isomerase', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6867420, end_position_on_the_genomic_accession=6870946, orientation='plus', exon_count=8, OMIM=190450, no_hdr0=''),
  7168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7168, CurrentID=0, Status='live', Symbol='TPM1', Aliases='C15orf13, CMD1Y, CMH3, HEL-S-265, HTM-alpha, LVNC9, TMSA', description='tropomyosin 1 (alpha)', other_designations='alpha-tropomyosin|cardiomyopathy, hypertrophic 3|epididymis secretory protein Li 265|sarcomeric tropomyosin kappa', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63042639, end_position_on_the_genomic_accession=63071915, orientation='plus', exon_count=15, OMIM=191010, no_hdr0=''),
  7169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7169, CurrentID=0, Status='live', Symbol='TPM2', Aliases='AMCD1, DA1, DA2B, HEL-S-273, NEM4, TMSB', description='tropomyosin 2 (beta)', other_designations='epididymis secretory protein Li 273|nemaline myopathy type 4', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35681993, end_position_on_the_genomic_accession=35690056, orientation='minus', exon_count=11, OMIM=190990, no_hdr0=''),
  7170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7170, CurrentID=0, Status='live', Symbol='TPM3', Aliases='CAPM1, CFTD, HEL-189, HEL-S-82p, NEM1, OK/SW-cl.5, TM-5, TM3, TM30, TM30nm, TM5, TPMsk3, TRK, hscp30', description='tropomyosin 3', other_designations='alpha-tropomyosin, slow skeletal|cytoskeletal tropomyosin TM30|epididymis luminal protein 189|epididymis secretory sperm binding protein Li 82p|heat-stable cytoskeletal protein 30 kDa|tropomyosin gamma|tropomyosin-5', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154155304, end_position_on_the_genomic_accession=154192135, orientation='minus', exon_count=17, OMIM=191030, no_hdr0=''),
  7171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7171, CurrentID=0, Status='live', Symbol='TPM4', Aliases='HEL-S-108', description='tropomyosin 4', other_designations='TM30p1|epididymis secretory protein Li 108|tropomyosin-4', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16067507, end_position_on_the_genomic_accession=16103005, orientation='plus', exon_count=14, OMIM=600317, no_hdr0=''),
  7172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7172, CurrentID=0, Status='live', Symbol='TPMT', Aliases='', description='thiopurine S-methyltransferase', other_designations='S-adenosyl-L-methionine:thiopurine S-methyltransferase', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18128311, end_position_on_the_genomic_accession=18155165, orientation='minus', exon_count=9, OMIM=187680, no_hdr0=''),
  7173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7173, CurrentID=0, Status='live', Symbol='TPO', Aliases='MSA, TDH2A, TPX', description='thyroid peroxidase', other_designations='thyroid microsomal antigen|thyroperoxidase', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1413461, end_position_on_the_genomic_accession=1542727, orientation='plus', exon_count=19, OMIM=606765, no_hdr0=''),
  7174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7174, CurrentID=0, Status='live', Symbol='TPP2', Aliases='TPP-2, TPPII', description='tripeptidyl peptidase II', other_designations='TPP-II|tripeptidyl aminopeptidase|tripeptidyl-peptidase II', map_location='13q32-q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102596936, end_position_on_the_genomic_accession=102679174, orientation='plus', exon_count=32, OMIM=190470, no_hdr0=''),
  7175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7175, CurrentID=0, Status='live', Symbol='TPR', Aliases='', description='translocated promoter region, nuclear basket protein', other_designations='NPC-associated intranuclear protein|megator|nuclear pore complex-associated protein TPR|translocated promoter region (to activated MET oncogene)|translocated promoter region protein|tumor potentiating region', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186311652, end_position_on_the_genomic_accession=186375848, orientation='minus', exon_count=51, OMIM=189940, no_hdr0=''),
  7177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7177, CurrentID=0, Status='live', Symbol='TPSAB1', Aliases='TPS1, TPS2, TPSB1', description='tryptase alpha/beta 1', other_designations='mast cell alpha II tryptase|mast cell beta I tryptase|tryptase I|tryptase alpha II|tryptase alpha-1|tryptase beta 1|tryptase beta I|tryptase beta-1|tryptase-1|tryptase-I|tryptase-III', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1240677, end_position_on_the_genomic_accession=1242554, orientation='plus', exon_count=6, OMIM=191080, no_hdr0=''),
  7178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7178, CurrentID=0, Status='live', Symbol='TPT1', Aliases='HRF, TCTP, p02, p23', description='tumor protein, translationally-controlled 1', other_designations='fortilin|histamine-releasing factor', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45333471, end_position_on_the_genomic_accession=45341284, orientation='minus', exon_count=6, OMIM=600763, no_hdr0=''),
  7179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7179, CurrentID=0, Status='live', Symbol='TPTE', Aliases='CT44, PTEN2', description='transmembrane phosphatase with tensin homology', other_designations='PTEN-related tyrosine phosphatase|cancer/testis antigen 44|tensin, putative protein-tyrosine phosphatase|tumor antigen BJ-HCC-5', map_location='21p11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10521514, end_position_on_the_genomic_accession=10606270, orientation='plus', exon_count=24, OMIM=604336, no_hdr0=''),
  7180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7180, CurrentID=0, Status='live', Symbol='CRISP2', Aliases='CRISP-2, CT36, GAPDL5, TPX1, TSP1', description='cysteine-rich secretory protein 2', other_designations='cancer/testis antigen 36|glyceraldehyde-3-phosphate dehydrogenase-like 5|testis specific protein 1 (probe H4-1 p3-1)|testis-specific protein TPX-1', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49676659, end_position_on_the_genomic_accession=49714198, orientation='minus', exon_count=15, OMIM=187430, no_hdr0=''),
  7181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7181, CurrentID=0, Status='live', Symbol='NR2C1', Aliases='TR2', description='nuclear receptor subfamily 2 group C member 1', other_designations='TR2 nuclear hormone receptor|nuclear receptor subfamily 2, group C isoform|nuclear receptor subfamily 2, group C, member 1|orphan nuclear receptor TR2', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95020229, end_position_on_the_genomic_accession=95073628, orientation='minus', exon_count=16, OMIM=601529, no_hdr0=''),
  7182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7182, CurrentID=0, Status='live', Symbol='NR2C2', Aliases='TAK1, TR4', description='nuclear receptor subfamily 2 group C member 2', other_designations='Nuclear hormone receptor TR4|nuclear receptor subfamily 2, group C, member 2|orphan nuclear receptor TAK1|orphan nuclear receptor TR4|testicular nuclear receptor 4', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14947584, end_position_on_the_genomic_accession=15049279, orientation='plus', exon_count=20, OMIM=601426, no_hdr0=''),
  138255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138255, CurrentID=0, Status='live', Symbol='C9orf135', Aliases='', description='chromosome 9 open reading frame 135', other_designations='', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69820800, end_position_on_the_genomic_accession=69906237, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  7184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7184, CurrentID=0, Status='live', Symbol='HSP90B1', Aliases='ECGP, GP96, GRP94, HEL-S-125m, HEL35, TRA1', description='heat shock protein 90kDa beta family member 1', other_designations='94 kDa glucose-regulated protein|endothelial cell (HBMEC) glycoprotein|epididymis luminal protein 35|epididymis secretory sperm binding protein Li 125m|heat shock protein 90 kDa beta member 1|heat shock protein 90kDa beta (Grp94), member 1|stress-inducible tumor rejection antigen gp96|tumor rejection antigen (gp96) 1|tumor rejection antigen 1', map_location='12q24.2-q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103930334, end_position_on_the_genomic_accession=103947930, orientation='plus', exon_count=18, OMIM=191175, no_hdr0=''),
  7185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7185, CurrentID=0, Status='live', Symbol='TRAF1', Aliases='EBI6, MGC:10353', description='TNF receptor associated factor 1', other_designations='Epstein-Bar virus-induced protein 6', map_location='9q33-q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120902393, end_position_on_the_genomic_accession=120929173, orientation='minus', exon_count=10, OMIM=601711, no_hdr0=''),
  7186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7186, CurrentID=0, Status='live', Symbol='TRAF2', Aliases='MGC:45012, TRAP, TRAP3', description='TNF receptor associated factor 2', other_designations='E3 ubiquitin-protein ligase TRAF2|tumor necrosis factor type 2 receptor associated protein 3|tumor necrosis factor type 2 receptor-associated protein 3', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136881933, end_position_on_the_genomic_accession=136926621, orientation='plus', exon_count=17, OMIM=601895, no_hdr0=''),
  7187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7187, CurrentID=0, Status='live', Symbol='TRAF3', Aliases='CAP-1, CAP1, CD40bp, CRAF1, IIAE5, LAP1', description='TNF receptor associated factor 3', other_designations='CD40 associated protein 1|CD40 binding protein|CD40 receptor associated factor 1|LMP1-associated protein 1', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102777475, end_position_on_the_genomic_accession=102911500, orientation='plus', exon_count=13, OMIM=601896, no_hdr0=''),
  7188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7188, CurrentID=0, Status='live', Symbol='TRAF5', Aliases='MGC:39780, RNF84', description='TNF receptor associated factor 5', other_designations='RING finger protein 84', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211326579, end_position_on_the_genomic_accession=211374946, orientation='plus', exon_count=13, OMIM=602356, no_hdr0=''),
  7189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7189, CurrentID=0, Status='live', Symbol='TRAF6', Aliases='MGC:3310, RNF85', description='TNF receptor associated factor 6', other_designations='E3 ubiquitin-protein ligase TRAF6|RING finger protein 85|TNF receptor-associated factor 6, E3 ubiquitin protein ligase|interleukin-1 signal transducer', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36483767, end_position_on_the_genomic_accession=36510313, orientation='minus', exon_count=8, OMIM=602355, no_hdr0=''),
  400406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400406, CurrentID=0, Status='live', Symbol='ADAMTS7P3', Aliases='', description='ADAMTS7 pseudogene 3', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 7 pseudogene', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77976009, end_position_on_the_genomic_accession=77993241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503641, CurrentID=0, Status='live', Symbol='DPRXP1', Aliases='', description='divergent-paired related homeobox pseudogene 1', other_designations='divergent paired-related homeobox pseudogene 1', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186488556, end_position_on_the_genomic_accession=186489169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7192, CurrentID=0, Status='live', Symbol='TRD-GTC9-1', Aliases='TRD, TRNAD1', description='transfer RNA-Asp (GTC) 9-1', other_designations='tRNA aspartic acid|tRNA-Asp-GTC-9-1|transfer RNA aspartic acid 1 (anticodon GUC)', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7193, CurrentID=0, Status='live', Symbol='TRE-TTC3-1', Aliases='TRE, TRNAE1, TRNE', description='transfer RNA-Glu (TTC) 3-1', other_designations='tRNA glutamic acid (anticodon UUC)|tRNA-Glu-TTC-3-1|transfer RNA glutamic acid (anticodon UUC)|transfer RNA glutamic acid 1 (anticodon UUC)', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16872583, end_position_on_the_genomic_accession=16872654, orientation='plus', exon_count=0, OMIM=180640, no_hdr0=''),
  400410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400410, CurrentID=0, Status='live', Symbol='ST20', Aliases='HCCS-1', description='suppressor of tumorigenicity 20', other_designations='cervical cancer suppressor-1|human cervical cancer suppressor gene 1 protein', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79898840, end_position_on_the_genomic_accession=79923754, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7195, CurrentID=0, Status='live', Symbol='TRNAG1', Aliases='TRG1', description='tRNA glycine1 (anticodon CCC)', other_designations='tRNA glycine (CCC) 1|tRNA glycine-GCC-1|transfer RNA glycine 1 (anticodon CCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16545939, end_position_on_the_genomic_accession=16546009, orientation='minus', exon_count=0, OMIM=189911, no_hdr0=''),
  7196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7196, CurrentID=0, Status='live', Symbol='TRG-GCC2-4', Aliases='TRG2, TRNAG2', description='transfer RNA-Gly (GCC) 2-4', other_designations='tRNA glycine 2|tRNA-Gly-GCC-2-4|transfer RNA glycine 2 (anticodon GCC)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8125746, end_position_on_the_genomic_accession=8125816, orientation='plus', exon_count=0, OMIM=610407, no_hdr0=''),
  7197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7197, CurrentID=0, Status='live', Symbol='TRG-TCC1-1', Aliases='TRG3, TRNAG3', description='transfer RNA-Gly (TCC) 1-1', other_designations='tRNA glycine (UCC) 3|tRNA glycine 3 (anticodon UCC)|tRNA-Gly-TCC-1-1|transfer RNA glycine 3 (anticodon UCC)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4724070, end_position_on_the_genomic_accession=4724141, orientation='plus', exon_count=0, OMIM=615303, no_hdr0=''),
  7198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7198, CurrentID=0, Status='live', Symbol='TRG-GCC3-1', Aliases='TRG5, TRGL1, TRNAG5', description='transfer RNA-Gly (GCC) 3-1', other_designations='tRNA glycine (GCC)-like 1|tRNA glycine 5 (anticodon GCC)|tRNA-Gly-GCC-3-1|transfer RNA glycine 5 (anticodon GCC)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70778211, end_position_on_the_genomic_accession=70778281, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7200, CurrentID=0, Status='live', Symbol='TRH', Aliases='Pro-TRH, TRF', description='thyrotropin-releasing hormone', other_designations='TSH-releasing factor|pro-thyrotropin-releasing hormone|prothyroliberin|protirelin|thyrotropin-releasing factor', map_location='3q13.3-q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129974393, end_position_on_the_genomic_accession=129977938, orientation='plus', exon_count=3, OMIM=613879, no_hdr0=''),
  7201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7201, CurrentID=0, Status='live', Symbol='TRHR', Aliases='TRH-R', description='thyrotropin-releasing hormone receptor', other_designations='thyroliberin receptor', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109086342, end_position_on_the_genomic_accession=109119584, orientation='plus', exon_count=3, OMIM=188545, no_hdr0=''),
  7202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7202, CurrentID=0, Status='live', Symbol='TRI-AAT9-1', Aliases='TRI, TRNAI1', description='transfer RNA-Ile (AAT) 9-1', other_designations='tRNA isoleucine|tRNA-Ile-AAT-9-1|transfer RNA isoleucine 1 (anticodon AAU)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27273960, end_position_on_the_genomic_accession=27274033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7203, CurrentID=0, Status='live', Symbol='CCT3', Aliases='CCT-gamma, CCTG, PIG48, TCP-1-gamma, TRIC5', description='chaperonin containing TCP1, subunit 3 (gamma)', other_designations='T-complex protein 1, gamma subunit|TCP1 (t-complex-1) ring complex, polypeptide 5|hTRiC5', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156308961, end_position_on_the_genomic_accession=156338415, orientation='minus', exon_count=16, OMIM=600114, no_hdr0=''),
  7204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7204, CurrentID=0, Status='live', Symbol='TRIO', Aliases='ARHGEF23, tgat', description='trio Rho guanine nucleotide exchange factor', other_designations='PTPRF-interacting protein|triple functional domain (PTPRF interacting)', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14143708, end_position_on_the_genomic_accession=14510204, orientation='plus', exon_count=64, OMIM=601893, no_hdr0=''),
  7205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7205, CurrentID=0, Status='live', Symbol='TRIP6', Aliases='OIP-1, OIP1, TRIP-6i2, ZRP-1, TRIP6', description='thyroid hormone receptor interactor 6', other_designations='OPA-interacting protein 1|TR-interacting protein 6|thyroid hormone receptor interacting protein 6|zyxin related protein 1', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100866221, end_position_on_the_genomic_accession=100873455, orientation='plus', exon_count=10, OMIM=602933, no_hdr0=''),
  7206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7206, CurrentID=0, Status='live', Symbol='TRK-TTT3-4', Aliases='TRK1, TRNAK1', description='transfer RNA-Lys (TTT) 3-4', other_designations='tRNA lysine 1 (anticodon UUU)|tRNA-Lys-TTT-3-4|transfer RNA lysine 1 (anticodon UUU)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8119155, end_position_on_the_genomic_accession=8119227, orientation='plus', exon_count=0, OMIM=189918, no_hdr0=''),
  7207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7207, CurrentID=0, Status='live', Symbol='TRL-AAG2-1', Aliases='TRL1, TRNAL1, TRNP1', description='transfer RNA-Leu (AAG) 2-1', other_designations='tRNA leucine 1 (anticodon CUU)|tRNA-Leu-AAG-2-1|transfer RNA leucine 1 (anticodon AAG)|transfer RNA leucine 1 (anticodon CUU)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20610132, end_position_on_the_genomic_accession=20610213, orientation='plus', exon_count=0, OMIM=189932, no_hdr0=''),
  7208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7208, CurrentID=0, Status='live', Symbol='TRL-TAG1-1', Aliases='TRL2, TRNAL2', description='transfer RNA-Leu (TAG) 1-1', other_designations='tRNA leucine 2 (anticodon UAG)|tRNA-Leu-TAG-1-1|transfer RNA leucine 2 (anticodon UAG)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8120314, end_position_on_the_genomic_accession=8120395, orientation='minus', exon_count=0, OMIM=189920, no_hdr0=''),
  7209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7209, CurrentID=0, Status='live', Symbol='TRNAL47P', Aliases='TRLP1', description='transfer RNA leucine 47 (anticodon AAG) pseudogene', other_designations='tRNA leucine (AAG) pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28478573, end_position_on_the_genomic_accession=28478654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7210, CurrentID=0, Status='live', Symbol='TRX-CAT1-3', Aliases='RNTMI1, RNTMI2, RNTMT1, TRM1, TRM2, TRMI2, TRNAM1, TRNAMI2', description='transfer RNA-iMet (CAT) 1-3', other_designations='tRNA methionine 1 (anticodon CAU)|tRNA methionine initiator 2|tRNA-iMet-CAT-1-3|transfer RNA initiator methionine (CAT) 1-3|transfer RNA methionine 1 (anticodon CAU)|transfer RNA methionine initiator 2 (anticodon CAU)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26286526, end_position_on_the_genomic_accession=26286597, orientation='plus', exon_count=0, OMIM=180621, no_hdr0=''),
  7211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7211, CurrentID=0, Status='live', Symbol='TRMEP1', Aliases='', description='tRNA methionine elongator pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28480726, end_position_on_the_genomic_accession=28480798, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7212, CurrentID=0, Status='live', Symbol='TRX-CAT2-1', Aliases='RNTMI, RNTMI2, TRM1, TRMI1, TRNAM2, TRNAMI1, TRNAMI2', description='transfer RNA-iMet (CAT) 2-1', other_designations='tRNA methionine initiator 1|tRNA-iMet-CAT-2-1|transfer RNA initiator methionine (CAT) 2-1|transfer RNA methionine 2 (anticodon CAU)|transfer RNA methionine initiator 1 (anticodon CAU)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27777885, end_position_on_the_genomic_accession=27777956, orientation='plus', exon_count=0, OMIM=180620, no_hdr0=''),
  7214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7214, CurrentID=0, Status='live', Symbol='TRN-GTT2-1', Aliases='RNN, TRN, TRN1, TRNAN1', description='transfer RNA-Asn (GTT) 2-1', other_designations='tRNA asparagine 1 (anticodon GUU)|tRNA-Asn-GTT-2-1|transfer RNA asparagine 1 (anticodon GUU)', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145287766, end_position_on_the_genomic_accession=145287839, orientation='plus', exon_count=0, OMIM=189880, no_hdr0=''),
  7216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7216, CurrentID=0, Status='live', Symbol='TRO', Aliases='MAGE-d3, MAGED3', description='trophinin', other_designations='MAGE superfamily protein|MAGE-D3 antigen|magphinin|melanoma antigen, family D, 3', map_location='Xp11.22-p11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54920563, end_position_on_the_genomic_accession=54931436, orientation='plus', exon_count=14, OMIM=300132, no_hdr0=''),
  7217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7217, CurrentID=0, Status='live', Symbol='TRP-AGG2-3', Aliases='TRNAP1, TRNP1, TRP1', description='transfer RNA-Pro (AGG) 2-3', other_designations='tRNA proline 1 (anticodon AGG)|tRNA-Pro-AGG-2-3|transfer RNA proline 1 (anticodon AGG)', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20609336, end_position_on_the_genomic_accession=20609407, orientation='minus', exon_count=0, OMIM=189930, no_hdr0=''),
  7218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7218, CurrentID=0, Status='live', Symbol='TRP-AGG2-4', Aliases='TRNAP2, TRNP1, TRP2', description='transfer RNA-Pro (AGG) 2-4', other_designations='tRNA proline 2 (anticodon UGG)|tRNA-Pro-AGG-2-4|transfer RNA proline 2 (anticodon AGG)|transfer RNA proline 2 (anticodon UGG)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20613401, end_position_on_the_genomic_accession=20613472, orientation='minus', exon_count=0, OMIM=189931, no_hdr0=''),
  7219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7219, CurrentID=0, Status='live', Symbol='TRP-TGG3-5', Aliases='TRNAP3, TRP3', description='transfer RNA-Pro (TGG) 3-5', other_designations='tRNA proline (UGG) 3|tRNA-Pro-TGG-3-5|transfer RNA proline (UGG) 3|transfer RNA proline 3 (anticodon UGG)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181188854, end_position_on_the_genomic_accession=181188925, orientation='minus', exon_count=0, OMIM=189912, no_hdr0=''),
  7220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7220, CurrentID=0, Status='live', Symbol='TRPC1', Aliases='HTRP-1, TRP1', description='transient receptor potential cation channel, subfamily C, member 1', other_designations='TRP-1|capacitative calcium channel protein Trp1|transient receptor potential canonical 1|transient receptor protein 1', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142724032, end_position_on_the_genomic_accession=142807888, orientation='plus', exon_count=15, OMIM=602343, no_hdr0=''),
  7221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7221, CurrentID=0, Status='live', Symbol='TRPC2', Aliases='', description='transient receptor potential cation channel, subfamily C, member 2, pseudogene', other_designations='transient receptor potential channel 2 (pseudogene)', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3626460, end_position_on_the_genomic_accession=3637559, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  7222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7222, CurrentID=0, Status='live', Symbol='TRPC3', Aliases='SCA41, TRP3', description='transient receptor potential cation channel, subfamily C, member 3', other_designations='hTrp-3|transient receptor potential cation channel subfamily C member 3 variant c|transient receptor protein 3', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121879028, end_position_on_the_genomic_accession=121951891, orientation='minus', exon_count=14, OMIM=602345, no_hdr0=''),
  7223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7223, CurrentID=0, Status='live', Symbol='TRPC4', Aliases='HTRP-4, HTRP4, TRP4', description='transient receptor potential cation channel, subfamily C, member 4', other_designations='trp-related protein 4', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37633872, end_position_on_the_genomic_accession=37869802, orientation='minus', exon_count=11, OMIM=603651, no_hdr0=''),
  7224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7224, CurrentID=0, Status='live', Symbol='TRPC5', Aliases='PPP1R159, TRP5', description='transient receptor potential cation channel, subfamily C, member 5', other_designations='TRP-5|hTRP5|protein phosphatase 1, regulatory subunit 159|transient receptor potential channel 5|transient receptor protein 5', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111774314, end_position_on_the_genomic_accession=112082776, orientation='minus', exon_count=11, OMIM=300334, no_hdr0=''),
  7225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7225, CurrentID=0, Status='live', Symbol='TRPC6', Aliases='FSGS2, TRP6', description='transient receptor potential cation channel, subfamily C, member 6', other_designations='TRP-6|focal segmental glomerulosclerosis 2|transient receptor protein 6', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101451564, end_position_on_the_genomic_accession=101584018, orientation='minus', exon_count=14, OMIM=603652, no_hdr0=''),
  7226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7226, CurrentID=0, Status='live', Symbol='TRPM2', Aliases='EREG1, KNP3, LTRPC2, NUDT9H, NUDT9L1, TRPC7', description='transient receptor potential cation channel, subfamily M, member 2', other_designations='LTrpC-2|estrogen-responsive element-associated gene 1 protein|long transient receptor potential channel 2|transient receptor potential channel 7', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44350126, end_position_on_the_genomic_accession=44443081, orientation='plus', exon_count=36, OMIM=603749, no_hdr0=''),
  7227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7227, CurrentID=0, Status='live', Symbol='TRPS1', Aliases='GC79, LGCR', description='trichorhinophalangeal syndrome I', other_designations='tricho-rhino-phalangeal syndrome type I protein|zinc finger protein GC79', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=115408496, end_position_on_the_genomic_accession=115669580, orientation='minus', exon_count=11, OMIM=604386, no_hdr0=''),
  7228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7228, CurrentID=0, Status='live', Symbol='TRQ-CTG1-2', Aliases='TRNAQ1, TRQ1', description='transfer RNA-Gln (CTG) 1-2', other_designations='tRNA glutamine 1 (anticodon CUG)|tRNA-Gln-CTG-1-2|transfer RNA glutamine 1 (anticodon CUG)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8119752, end_position_on_the_genomic_accession=8119823, orientation='plus', exon_count=0, OMIM=189919, no_hdr0=''),
  7229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7229, CurrentID=0, Status='live', Symbol='TRQ-TTG1-1', Aliases='TRNAQ2, TRQ2', description='transfer RNA-Gln (TTG) 1-1', other_designations='tRNA glutamine 2|tRNA-Gln-TTG-1-1|transfer RNA glutamine 2 (anticodon UUG)', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49192528, end_position_on_the_genomic_accession=49192599, orientation='plus', exon_count=0, OMIM=189923, no_hdr0=''),
  7230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7230, CurrentID=0, Status='live', Symbol='TRR-TCT2-1', Aliases='TRNAR1, TRR1', description='transfer RNA-Arg (TCT) 2-1', other_designations='tRNA arginine (UCU) 1|tRNA arginine 1 (anticodon UCU)|transfer RNA arginine 1 (anticodon UCU)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8120925, end_position_on_the_genomic_accession=8121012, orientation='plus', exon_count=0, OMIM=610406, no_hdr0=''),
  7231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7231, CurrentID=0, Status='live', Symbol='TRR-ACG1-3', Aliases='TRNAR2, TRR2', description='transfer RNA-Arg (ACG) 1-3', other_designations='tRNA arginine (ACG) 2|tRNA arginine 2 (anticodon ACG)|tRNA-Arg-ACG-1-3|transfer RNA arginine 2 (anticodon ACG)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26537498, end_position_on_the_genomic_accession=26537570, orientation='plus', exon_count=0, OMIM=615305, no_hdr0=''),
  7232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7232, CurrentID=0, Status='live', Symbol='TRR-TCG4-1', Aliases='TRNAR3, TRR3', description='transfer RNA-Arg (TCG) 4-1', other_designations='tRNA arginine (UCG) 3|tRNA arginine 3 (anticodon UCG)|tRNA arginine-3|tRNA-Arg-TCG-4-1|transfer RNA arginine 3 (anticodon UCG)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26299677, end_position_on_the_genomic_accession=26299749, orientation='plus', exon_count=0, OMIM=601432, no_hdr0=''),
  7233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7233, CurrentID=0, Status='live', Symbol='TRR-CCG2-1', Aliases='TRNAR4, TRR4', description='transfer RNA-Arg (CCG) 2-1', other_designations='tRNA arginine (CCG) 4|tRNA-Arg-CCG-2-1|transfer RNA arginine (CCG) 4|transfer RNA arginine 4 (anticodon CCG)', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68019897, end_position_on_the_genomic_accession=68019969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7234, CurrentID=0, Status='live', Symbol='TRU-TCA1-1', Aliases='TRNAU1, TRSP, tRNA(Sec)', description='transfer RNA-SeC (TCA) 1-1', other_designations='tRNA phosphoserine (opal suppressor)|tRNA selenocysteine/phosphoserine (opal suppressor)|tRNA-SeC-TCA-1-1|transfer RNA selenocysteine 1 (anticodon UCA)|transfer RNA selenocysteine/phosphoserine (opal suppressor)', map_location='19q13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45478602, end_position_on_the_genomic_accession=45478687, orientation='minus', exon_count=0, OMIM=165060, no_hdr0=''),
  400451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400451, CurrentID=0, Status='live', Symbol='FAM174B', Aliases='', description='family with sequence similarity 174 member B', other_designations='family with sequence similarity 174, member B', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92617448, end_position_on_the_genomic_accession=92812338, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  7236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7236, CurrentID=0, Status='live', Symbol='TRT-TGT6-1', Aliases='TRNAT1, TRT1', description='transfer RNA-Thr (TGT) 6-1', other_designations='tRNA threonine (UGU) 1|tRNA threonine 1 (anticodon UGU)|tRNA-Thr-TGT-6-1|transfer RNA threonine 1 (anticodon UGU)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181191687, end_position_on_the_genomic_accession=181191758, orientation='minus', exon_count=0, OMIM=189913, no_hdr0=''),
  7237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7237, CurrentID=0, Status='live', Symbol='TRT-TGT3-1', Aliases='TRNAT2, TRT2', description='transfer RNA-Thr (TGT) 3-1', other_designations='tRNA threonine (UGU) 2|tRNA threonine-2|tRNA-Thr-TGT-3-1|transfer RNA threonine (UGU) 2|transfer RNA threonine 2 (anticodon UGU)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20613790, end_position_on_the_genomic_accession=20613862, orientation='minus', exon_count=0, OMIM=189933, no_hdr0=''),
  7238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7238, CurrentID=0, Status='live', Symbol='TRT-AGT7-1', Aliases='TRNAT3, TRT3', description='transfer RNA-Thr (AGT) 7-1', other_designations='tRNA threonine (AGU) 3|tRNA-Thr-AGT-7-1|transfer RNA threonine (AGU) 3|transfer RNA threonine 3 (anticodon AGU)', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138311, CurrentID=0, Status='live', Symbol='FAM69B', Aliases='C9orf136, pp6977', description='family with sequence similarity 69 member B', other_designations='family with sequence similarity 69, member B', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136712572, end_position_on_the_genomic_accession=136724718, orientation='plus', exon_count=7, OMIM=614543, no_hdr0=''),
  7240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7240, CurrentID=0, Status='live', Symbol='TRV-CAC1-2', Aliases='TRNAV2, TRV2', description='transfer RNA-Val (CAC) 1-2', other_designations='tRNA valine (CAC) 2|tRNA-Val-CAC-1-2|transfer RNA valine (CAC) 2|transfer RNA valine 2 (anticodon CAC)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181097070, end_position_on_the_genomic_accession=181097142, orientation='plus', exon_count=0, OMIM=612995, no_hdr0=''),
  7241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7241, CurrentID=0, Status='live', Symbol='TRV-AAC1-3', Aliases='TRNAV3, TRV3', description='transfer RNA-Val (AAC) 1-3', other_designations='tRNA valine (IAC) 3 (isoacceptor)|tRNA valine (IAC), isoacceptor 3|tRNA-Val-AAC-1-3|transfer RNA valine (IAC), isoacceptor 3|transfer RNA valine 3 (anticodon AAC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181169610, end_position_on_the_genomic_accession=181169682, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271287, CurrentID=0, Status='live', Symbol='RPL19P17', Aliases='RPL19_6_1208', description='ribosomal protein L19 pseudogene 17', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12568124, end_position_on_the_genomic_accession=12568798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7246, CurrentID=0, Status='live', Symbol='TRW-CCA7-1', Aliases='TRNAW1, TRW', description='transfer RNA-Trp (CCA) 7-1', other_designations='tRNA tryptophan|tRNA-Trp-CCA-7-1|transfer RNA tryptophan 1 (anticodon CCA)', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7247, CurrentID=0, Status='live', Symbol='TSN', Aliases='BCLF-1, C3PO, RCHF1, REHF-1, TBRBP, TRSLN', description='translin', other_designations='component 3 of promoter of RISC|recombination hotspot associated factor|recombination hotspot-binding protein|testis brain-RNA binding protein', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121755545, end_position_on_the_genomic_accession=121767853, orientation='plus', exon_count=6, OMIM=600575, no_hdr0=''),
  7248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7248, CurrentID=0, Status='live', Symbol='TSC1', Aliases='LAM, TSC', description='tuberous sclerosis 1', other_designations='tuberous sclerosis 1 protein|tumor suppressor', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132891348, end_position_on_the_genomic_accession=132945370, orientation='minus', exon_count=25, OMIM=605284, no_hdr0=''),
  7249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7249, CurrentID=0, Status='live', Symbol='TSC2', Aliases='LAM, PPP1R160, TSC4', description='tuberous sclerosis 2', other_designations='protein phosphatase 1, regulatory subunit 160|tuberous sclerosis 2 protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2047936, end_position_on_the_genomic_accession=2088712, orientation='plus', exon_count=42, OMIM=191092, no_hdr0=''),
  7251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7251, CurrentID=0, Status='live', Symbol='TSG101', Aliases='TSG10, VPS23', description='tumor susceptibility 101', other_designations='ESCRT-I complex subunit TSG101|tumor susceptibility gene 10|tumor susceptibility gene 101|tumor susceptibility protein', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18480311, end_position_on_the_genomic_accession=18526956, orientation='minus', exon_count=11, OMIM=601387, no_hdr0=''),
  7252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7252, CurrentID=0, Status='live', Symbol='TSHB', Aliases='TSH-B, TSH-BETA', description='thyroid stimulating hormone, beta', other_designations='thyrotropin beta chain', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115029824, end_position_on_the_genomic_accession=115034309, orientation='plus', exon_count=3, OMIM=188540, no_hdr0=''),
  7253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7253, CurrentID=0, Status='live', Symbol='TSHR', Aliases='CHNG1, LGR3, hTSHR-I', description='thyroid stimulating hormone receptor', other_designations='seven transmembrane helix receptor|thyrotropin receptor-I, hTSHR-I', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80954989, end_position_on_the_genomic_accession=81146302, orientation='plus', exon_count=12, OMIM=603372, no_hdr0=''),
  7254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7254, CurrentID=0, Status='live', Symbol='TSHRL1', Aliases='', description='thyroid stimulating hormone receptor-like 1', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  131177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131177, CurrentID=0, Status='live', Symbol='FAM3D', Aliases='EF7, OIT1', description='family with sequence similarity 3 member D', other_designations='cytokine-like protein EF-7|family with sequence similarity 3, member D', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58633943, end_position_on_the_genomic_accession=58666844, orientation='minus', exon_count=11, OMIM=608619, no_hdr0=''),
  7256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7256, CurrentID=0, Status='live', Symbol='TSHRL3', Aliases='', description='thyroid stimulating hormone receptor-like 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7257, CurrentID=0, Status='live', Symbol='TSNAX', Aliases='TRAX', description='translin-associated factor X', other_designations='translin-like protein', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231528653, end_position_on_the_genomic_accession=231566524, orientation='plus', exon_count=6, OMIM=602964, no_hdr0=''),
  7258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7258, CurrentID=0, Status='live', Symbol='TSPY1', Aliases='CT78, DYS14, TSPY, pJA923', description='testis specific protein, Y-linked 1', other_designations='cancer/testis antigen 78', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9466955, end_position_on_the_genomic_accession=9469749, orientation='plus', exon_count=6, OMIM=480100, no_hdr0=''),
  7259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7259, CurrentID=0, Status='live', Symbol='TSPYL1', Aliases='TSPYL', description='TSPY-like 1', other_designations='TSPY-like protein 1', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116274859, end_position_on_the_genomic_accession=116280117, orientation='minus', exon_count=1, OMIM=604714, no_hdr0=''),
  7260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7260, CurrentID=0, Status='live', Symbol='TSSC1', Aliases='', description='tumor suppressing subtransferable candidate 1', other_designations='tumor-suppressing STF cDNA 1 protein|tumor-suppressing subchromosomal transferable fragment candidate gene 1 protein', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3188970, end_position_on_the_genomic_accession=3377882, orientation='minus', exon_count=13, OMIM=608998, no_hdr0=''),
  7262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7262, CurrentID=0, Status='live', Symbol='PHLDA2', Aliases='BRW1C, BWR1C, HLDA2, IPL, TSSC3', description='pleckstrin homology-like domain, family A, member 2', other_designations='beckwith-Wiedemann syndrome chromosomal region 1 candidate gene C protein|imprinted in placenta and liver protein|p17-BWR1C|p17-Beckwith-Wiedemann region 1 C|p17-Beckwith-Wiedemann region 1C|tumor suppressing subchromosomal transferable fragment cDNA 3|tumor suppressing subtransferable candidate 3|tumor-suppressing subchromosomal transferable fragment candidate gene 3 protein|tumor-supressing STF cDNA 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2928273, end_position_on_the_genomic_accession=2929420, orientation='minus', exon_count=2, OMIM=602131, no_hdr0=''),
  7263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7263, CurrentID=0, Status='live', Symbol='TST', Aliases='RDS', description='thiosulfate sulfurtransferase (rhodanese)', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37010859, end_position_on_the_genomic_accession=37020183, orientation='minus', exon_count=4, OMIM=180370, no_hdr0=''),
  7264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7264, CurrentID=0, Status='live', Symbol='TSTA3', Aliases='FX, P35B, SDR4E1', description='tissue specific transplantation antigen P35B', other_designations='3-5 epimerase/4-reductase|GDP-4-keto-6-deoxy-D-mannose epimerase-reductase|GDP-4-keto-6-deoxy-D-mannose-3,5-epimerase-4-reductase|red cell NADP(H)-binding protein|short chain dehydrogenase/reductase family 4E, member 1|tissue specific transplantation antigen 3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143612618, end_position_on_the_genomic_accession=143618043, orientation='minus', exon_count=13, OMIM=137020, no_hdr0=''),
  7265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7265, CurrentID=0, Status='live', Symbol='TTC1', Aliases='TPR1', description='tetratricopeptide repeat domain 1', other_designations='TPR repeat protein 1', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160009100, end_position_on_the_genomic_accession=160065545, orientation='plus', exon_count=8, OMIM=601963, no_hdr0=''),
  7266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7266, CurrentID=0, Status='live', Symbol='DNAJC7', Aliases='DJ11, DJC7, TPR2, TTC2', description='DnaJ heat shock protein family (Hsp40) member C7', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 7|TPR repeat protein 2|tetratricopeptide repeat domain 2|tetratricopeptide repeat protein 2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41974005, end_position_on_the_genomic_accession=42021371, orientation='minus', exon_count=20, OMIM=601964, no_hdr0=''),
  7267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7267, CurrentID=0, Status='live', Symbol='TTC3', Aliases='DCRR1, RNF105, TPRDIII', description='tetratricopeptide repeat domain 3', other_designations='RING finger protein 105|TPR repeat protein 3|TPR repeat protein D|tetratricopeptide repeat protein 3', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37073254, end_position_on_the_genomic_accession=37203107, orientation='plus', exon_count=50, OMIM=602259, no_hdr0=''),
  7268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7268, CurrentID=0, Status='live', Symbol='TTC4', Aliases='', description='tetratricopeptide repeat domain 4', other_designations='TPR repeat protein 4', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54715822, end_position_on_the_genomic_accession=54742657, orientation='plus', exon_count=10, OMIM=606753, no_hdr0=''),
  7270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7270, CurrentID=0, Status='live', Symbol='TTF1', Aliases='TTF-1, TTF-I', description='transcription termination factor, RNA polymerase I', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132375548, end_position_on_the_genomic_accession=132406871, orientation='minus', exon_count=12, OMIM=600777, no_hdr0=''),
  7271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7271, CurrentID=0, Status='live', Symbol='TTIM1', Aliases='', description='T-cell tumor invasion and metastasis 1', other_designations='T-cell tumor invasion and metastasis-1 (invasion-metastasis of neoplasms, chromosome 7 determined)', map_location='7p12-cen', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147830, no_hdr0=''),
  7272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7272, CurrentID=0, Status='live', Symbol='TTK', Aliases='CT96, ESK, MPH1, MPS1, MPS1L1, PYT', description='TTK protein kinase', other_designations='cancer/testis antigen 96|monopolar spindle 1 kinase|monopolar spindle 1-like 1|phosphotyrosine picked threonine kinase|phosphotyrosine picked threonine-protein kinase', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80004605, end_position_on_the_genomic_accession=80042527, orientation='plus', exon_count=22, OMIM=604092, no_hdr0=''),
  7273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7273, CurrentID=0, Status='live', Symbol='TTN', Aliases='CMD1G, CMH9, CMPD4, EOMFC, HMERF, LGMD2J, MYLK5, TMD', description='titin', other_designations='connectin|rhabdomyosarcoma antigen MU-RMS-40.14', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178525989, end_position_on_the_genomic_accession=178830754, orientation='minus', exon_count=367, OMIM=188840, no_hdr0=''),
  7274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7274, CurrentID=0, Status='live', Symbol='TTPA', Aliases='ATTP, AVED, TTP1, alphaTTP', description='tocopherol (alpha) transfer protein', other_designations='alpha-TTP|tocopherol (alpha) transfer protein (ataxia (Friedreich-like) with vitamin E deficiency)', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63059488, end_position_on_the_genomic_accession=63086100, orientation='minus', exon_count=5, OMIM=600415, no_hdr0=''),
  7275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7275, CurrentID=0, Status='live', Symbol='TUB', Aliases='RDOB, rd5', description='tubby bipartite transcription factor', other_designations='tubby homolog|tubby homologue', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8019180, end_position_on_the_genomic_accession=8106241, orientation='plus', exon_count=16, OMIM=601197, no_hdr0=''),
  7276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7276, CurrentID=0, Status='live', Symbol='TTR', Aliases='CTS, CTS1, HEL111, HsT2651, PALB, TBPA', description='transthyretin', other_designations='ATTR|carpal tunnel syndrome 1|epididymis luminal protein 111|prealbumin, amyloidosis type I|thyroxine-binding prealbumin', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31591767, end_position_on_the_genomic_accession=31599024, orientation='plus', exon_count=4, OMIM=176300, no_hdr0=''),
  7277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7277, CurrentID=0, Status='live', Symbol='TUBA4A', Aliases='ALS22, H2-ALPHA, TUBA1', description='tubulin alpha 4a', other_designations='tubulin H2-alpha|tubulin alpha-1 chain|tubulin, alpha 1 (testis specific)|tubulin, alpha 4a', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219249711, end_position_on_the_genomic_accession=219254608, orientation='minus', exon_count=6, OMIM=191110, no_hdr0=''),
  7278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7278, CurrentID=0, Status='live', Symbol='TUBA3C', Aliases='TUBA2, bA408E5.3', description='tubulin alpha 3c', other_designations='alpha-tubulin 2|alpha-tubulin 3C/D|tubulin alpha-2 chain|tubulin, alpha 2|tubulin, alpha 3c', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19173770, end_position_on_the_genomic_accession=19181852, orientation='minus', exon_count=5, OMIM=602528, no_hdr0=''),
  100271293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271293, CurrentID=0, Status='live', Symbol='RPL23AP33', Aliases='RPL23A_7_304', description='ribosomal protein L23a pseudogene 33', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184902108, end_position_on_the_genomic_accession=184902562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7280, CurrentID=0, Status='live', Symbol='TUBB2A', Aliases='CDCBM5, TUBB, TUBB2', description='tubulin beta 2A class IIa', other_designations='class IIa beta-tubulin|tubulin, beta 2A class IIa|tubulin, beta polypeptide 2', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3153666, end_position_on_the_genomic_accession=3157549, orientation='minus', exon_count=4, OMIM=615101, no_hdr0=''),
  7283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7283, CurrentID=0, Status='live', Symbol='TUBG1', Aliases='CDCBM4, GCP-1, TUBG, TUBGCP1', description='tubulin gamma 1', other_designations='gamma-tubulin complex component 1|tubulin, gamma 1|tubulin, gamma polypeptide', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42609340, end_position_on_the_genomic_accession=42615238, orientation='plus', exon_count=11, OMIM=191135, no_hdr0=''),
  7284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7284, CurrentID=0, Status='live', Symbol='TUFM', Aliases='COXPD4, EF-TuMT, EFTU, P43', description='Tu translation elongation factor, mitochondrial', other_designations='EF-Tu', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28842411, end_position_on_the_genomic_accession=28846511, orientation='minus', exon_count=10, OMIM=602389, no_hdr0=''),
  7286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7286, CurrentID=0, Status='live', Symbol='TUFT1', Aliases='', description='tuftelin 1', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151540300, end_position_on_the_genomic_accession=151583583, orientation='plus', exon_count=14, OMIM=600087, no_hdr0=''),
  7287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7287, CurrentID=0, Status='live', Symbol='TULP1', Aliases='LCA15, RP14, TUBL1', description='tubby like protein 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35497874, end_position_on_the_genomic_accession=35512902, orientation='minus', exon_count=15, OMIM=602280, no_hdr0=''),
  7288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7288, CurrentID=0, Status='live', Symbol='TULP2', Aliases='CT65, TUBL2', description='tubby like protein 2', other_designations='cancer testis antigen 65|cancer/testis antigen 65|tubby-like protein 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48880822, end_position_on_the_genomic_accession=48900184, orientation='minus', exon_count=15, OMIM=602309, no_hdr0=''),
  7289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7289, CurrentID=0, Status='live', Symbol='TULP3', Aliases='TUBL3', description='tubby like protein 3', other_designations='tubby-like protein 3', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2890867, end_position_on_the_genomic_accession=2941140, orientation='plus', exon_count=12, OMIM=604730, no_hdr0=''),
  7290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7290, CurrentID=0, Status='live', Symbol='HIRA', Aliases='DGCR1, TUP1, TUPLE1', description='histone cell cycle regulator', other_designations='DiGeorge critical region gene 1|HIR histone cell cycle regulation defective homolog A|TUP1-like enhancer of split protein 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19330701, end_position_on_the_genomic_accession=19431696, orientation='minus', exon_count=25, OMIM=600237, no_hdr0=''),
  7291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7291, CurrentID=0, Status='live', Symbol='TWIST1', Aliases='ACS3, BPES2, BPES3, CRS, CRS1, CSO, SCS, TWIST, bHLHa38', description='twist family bHLH transcription factor 1', other_designations='B-HLH DNA binding protein|H-twist|TWIST homolog of drosophila|class A basic helix-loop-helix protein 38|twist basic helix-loop-helix transcription factor 1|twist homolog 1', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19113047, end_position_on_the_genomic_accession=19117672, orientation='minus', exon_count=4, OMIM=601622, no_hdr0=''),
  7292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7292, CurrentID=0, Status='live', Symbol='TNFSF4', Aliases='CD134L, CD252, GP34, OX-40L, OX4OL, TXGP1', description='tumor necrosis factor superfamily member 4', other_designations='CD134 ligand|OX40 antigen ligand|glycoprotein Gp34|tax-transcriptionally activated glycoprotein 1 (34kD)|tumor necrosis factor (ligand) superfamily member 4', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173183731, end_position_on_the_genomic_accession=173208072, orientation='minus', exon_count=5, OMIM=603594, no_hdr0=''),
  7293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7293, CurrentID=0, Status='live', Symbol='TNFRSF4', Aliases='ACT35, CD134, IMD16, OX40, TXGP1L', description='tumor necrosis factor receptor superfamily member 4', other_designations='ACT35 antigen|ATC35 antigen|CD134 antigen|OX40 antigen|OX40 cell surface antigen|OX40 homologue|OX40L receptor|TAX transcriptionally-activated glycoprotein 1 receptor|lymphoid activation antigene ACT35|tax-transcriptionally activated glycoprotein 1 receptor|tumor necrosis factor receptor superfamily, member 4', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1211326, end_position_on_the_genomic_accession=1214638, orientation='minus', exon_count=8, OMIM=600315, no_hdr0=''),
  7294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7294, CurrentID=0, Status='live', Symbol='TXK', Aliases='BTKL, PSCTK5, PTK4, RLK, TKL', description='TXK tyrosine kinase', other_designations='PTK4 protein tyrosine kinase 4|protein-tyrosine kinase 4|resting lymphocyte kinase|tyrosine kinase', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48066393, end_position_on_the_genomic_accession=48134256, orientation='minus', exon_count=16, OMIM=600058, no_hdr0=''),
  7295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7295, CurrentID=0, Status='live', Symbol='TXN', Aliases='TRDX, TRX, TRX1', description='thioredoxin', other_designations='ADF|ATL-derived factor|SASP|TXN delta 3|surface-associated sulphydryl protein|thioredoxin delta 3', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110243812, end_position_on_the_genomic_accession=110256640, orientation='minus', exon_count=5, OMIM=187700, no_hdr0=''),
  7296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7296, CurrentID=0, Status='live', Symbol='TXNRD1', Aliases='GRIM-12, TR, TR1, TRXR1, TXNR', description='thioredoxin reductase 1', other_designations='KM-102-derived reductase-like factor|gene associated with retinoic and IFN-induced mortality 12 protein|gene associated with retinoic and interferon-induced mortality 12 protein|oxidoreductase|thioredoxin reductase GRIM-12|thioredoxin reductase TR1', map_location='12q23-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104215779, end_position_on_the_genomic_accession=104350307, orientation='plus', exon_count=18, OMIM=601112, no_hdr0=''),
  7297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7297, CurrentID=0, Status='live', Symbol='TYK2', Aliases='IMD35, JTK1', description='tyrosine kinase 2', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10350528, end_position_on_the_genomic_accession=10380572, orientation='minus', exon_count=29, OMIM=176941, no_hdr0=''),
  7298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7298, CurrentID=0, Status='live', Symbol='TYMS', Aliases='HST422, TMS, TS', description='thymidylate synthetase', other_designations='TSase', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=657604, end_position_on_the_genomic_accession=673499, orientation='plus', exon_count=7, OMIM=188350, no_hdr0=''),
  7299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7299, CurrentID=0, Status='live', Symbol='TYR', Aliases='ATN, CMM8, OCA1, OCA1A, OCAIA, SHEP3', description='tyrosinase', other_designations='LB24-AB|SK29-AB|monophenol monooxygenase|oculocutaneous albinism IA|tumor rejection antigen AB', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89177858, end_position_on_the_genomic_accession=89295759, orientation='plus', exon_count=6, OMIM=606933, no_hdr0=''),
  7300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7300, CurrentID=0, Status='live', Symbol='TYRL', Aliases='', description='tyrosinase-like (pseudogene)', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49404692, end_position_on_the_genomic_accession=49415754, orientation='minus', exon_count=0, OMIM=191270, no_hdr0=''),
  7301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7301, CurrentID=0, Status='live', Symbol='TYRO3', Aliases='BYK, Dtk, Etk-2, RSE, Rek, Sky, Tif', description='TYRO3 protein tyrosine kinase', other_designations='tyrosine-protein kinase DTK|tyrosine-protein kinase RSE|tyrosine-protein kinase SKY|tyrosine-protein kinase TIF|tyrosine-protein kinase byk', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41557871, end_position_on_the_genomic_accession=41579338, orientation='plus', exon_count=21, OMIM=600341, no_hdr0=''),
  7302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7302, CurrentID=0, Status='live', Symbol='TYRO3P', Aliases='', description='TYRO3P protein tyrosine kinase pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76259289, end_position_on_the_genomic_accession=76260152, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7303, CurrentID=0, Status='live', Symbol='TYS', Aliases='HRZ', description='sclerotylosis', other_designations='', map_location='4q28-q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=181600, no_hdr0=''),
  7305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7305, CurrentID=0, Status='live', Symbol='TYROBP', Aliases='DAP12, KARAP, PLOSL', description='TYRO protein tyrosine kinase binding protein', other_designations='DNAX-activation protein 12|KAR-associated protein|killer-activating receptor-associated protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35904401, end_position_on_the_genomic_accession=35908309, orientation='minus', exon_count=5, OMIM=604142, no_hdr0=''),
  7306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7306, CurrentID=0, Status='live', Symbol='TYRP1', Aliases='CAS2, CATB, GP75, OCA3, TRP, TRP1, TYRP, b-PROTEIN', description='tyrosinase-related protein 1', other_designations='DHICA oxidase|catalase B|glycoprotein 75|melanoma antigen gp75', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12693386, end_position_on_the_genomic_accession=12710266, orientation='plus', exon_count=8, OMIM=115501, no_hdr0=''),
  7307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7307, CurrentID=0, Status='live', Symbol='U2AF1', Aliases='FP793, RN, RNU2AF1, U2AF35, U2AFBP', description='U2 small nuclear RNA auxiliary factor 1', other_designations='U2 small nuclear RNA auxillary factor 1|U2 small nuclear ribonucleoprotein auxillary factor, 35-KD subunit|U2 snRNP auxiliary factor small subunit|U2(RNU2) small nuclear RNA auxiliary factor binding protein|splicing factor U2AF 35kDa subunit', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43092956, end_position_on_the_genomic_accession=43107578, orientation='minus', exon_count=9, OMIM=191317, no_hdr0=''),
  100271298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271298, CurrentID=0, Status='live', Symbol='RPL27AP3', Aliases='RPL27A_1_302', description='ribosomal protein L27a pseudogene 3', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180938867, end_position_on_the_genomic_accession=180939298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7310, CurrentID=0, Status='live', Symbol='ZRSR1', Aliases='U2AF1-RS1, U2AF1L1, U2AF1P, U2AF1RS1, U2AFBPL, ZC3H21', description='zinc finger (CCCH type), RNA binding motif and serine/arginine rich 1', other_designations='U2 small nuclear RNA auxillary factor 1-like 1|U2 small nuclear ribonucleoprotein auxiliary factor, small subunit 1|U2(RNU2) small nuclear RNA auxiliary factor pseudogene 1|zinc finger (CCCH type), RNA-binding motif and serine/arginine rich 1', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112891616, end_position_on_the_genomic_accession=112893097, orientation='plus', exon_count=0, OMIM=601079, no_hdr0=''),
  7311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7311, CurrentID=0, Status='live', Symbol='UBA52', Aliases='CEP52, HUBCEP52, L40, RPL40', description='ubiquitin A-52 residue ribosomal protein fusion product 1', other_designations='60S ribosomal protein L40|ubiquitin carboxyl extension protein 52|ubiquitin-52 amino acid fusion protein|ubiquitin-CEP52', map_location='19p13.1-p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18563766, end_position_on_the_genomic_accession=18577460, orientation='plus', exon_count=6, OMIM=191321, no_hdr0=''),
  7312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7312, CurrentID=0, Status='live', Symbol='UBA52P1', Aliases='RPL40_1_796', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 1', other_designations='ubiquitin and ribosomal protein L40 precursor pseudogene 1|ubiquitin carboxyl extension protein 52 pseudogene 1|ubiquitin-52 amino acid fusion protein pseudogene 1', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25689396, end_position_on_the_genomic_accession=25690070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7313, CurrentID=0, Status='live', Symbol='UBA52P2', Aliases='RPL40_4_1339', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 2', other_designations='ubiquitin and ribosomal protein L40 precursor pseudogene 2|ubiquitin carboxyl extension protein 52 pseudogene 2|ubiquitin-52 amino acid fusion protein pseudogene 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21398788, end_position_on_the_genomic_accession=21399327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7314, CurrentID=0, Status='live', Symbol='UBB', Aliases='HEL-S-50', description='ubiquitin B', other_designations='epididymis secretory protein Li 50|polyubiquitin B', map_location='17p12-p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16380793, end_position_on_the_genomic_accession=16382745, orientation='plus', exon_count=5, OMIM=191339, no_hdr0=''),
  7315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7315, CurrentID=0, Status='live', Symbol='UBBP1', Aliases='', description='ubiquitin B pseudogene 1', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136329300, end_position_on_the_genomic_accession=136330002, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7316, CurrentID=0, Status='live', Symbol='UBC', Aliases='HMG20', description='ubiquitin C', other_designations='', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124911645, end_position_on_the_genomic_accession=124915041, orientation='minus', exon_count=2, OMIM=191340, no_hdr0=''),
  7317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7317, CurrentID=0, Status='live', Symbol='UBA1', Aliases='A1S9, A1S9T, A1ST, AMCX1, CFAP124, GXP1, POC20, SMAX2A, UBE1, UBE1X, UBA1', description='ubiquitin-like modifier activating enzyme 1', other_designations='A1S9T and BN75 temperature sensitivity complementing|POC20 centriolar protein homolog|UBA1, ubiquitin-activating enzyme E1 homolog A', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47190800, end_position_on_the_genomic_accession=47215128, orientation='plus', exon_count=31, OMIM=314370, no_hdr0=''),
  7318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7318, CurrentID=0, Status='live', Symbol='UBA7', Aliases='D8, UBA1B, UBE1L, UBE2', description='ubiquitin-like modifier activating enzyme 7', other_designations='UBA1, ubiquitin-activating enzyme E1 homolog B|UBA7, ubiquitin-activating enzyme E1|ubiquitin-activating enzyme 7|ubiquitin-activating enzyme E1 homolog|ubiquitin-activating enzyme E1-related protein|ubiquitin-activating enzyme-2', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49805205, end_position_on_the_genomic_accession=49813958, orientation='minus', exon_count=24, OMIM=191325, no_hdr0=''),
  7319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7319, CurrentID=0, Status='live', Symbol='UBE2A', Aliases='HHR6A, MRXS30, MRXSN, RAD6A, UBC2', description='ubiquitin conjugating enzyme E2A', other_designations='E2 ubiquitin-conjugating enzyme A|RAD6 homolog A|ubiquitin carrier protein A|ubiquitin-protein ligase A', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119574467, end_position_on_the_genomic_accession=119584429, orientation='plus', exon_count=6, OMIM=312180, no_hdr0=''),
  7320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7320, CurrentID=0, Status='live', Symbol='UBE2B', Aliases='E2-17kDa, HHR6B, HR6B, RAD6B, UBC2', description='ubiquitin conjugating enzyme E2B', other_designations='E2 protein|E2 ubiquitin-conjugating enzyme B|RAD6 homolog B|ubiquitin carrier protein B|ubiquitin-conjugating enzyme E2-17 kDa|ubiquitin-conjugating enzyme E2B (RAD6 homolog)|ubiquitin-protein ligase B', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134371176, end_position_on_the_genomic_accession=134392108, orientation='plus', exon_count=6, OMIM=179095, no_hdr0=''),
  7321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7321, CurrentID=0, Status='live', Symbol='UBE2D1', Aliases='E2(17)KB1, SFT, UBC4/5, UBCH5, UBCH5A', description='ubiquitin conjugating enzyme E2D 1', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D1|E2 ubiquitin-conjugating enzyme D1|UBC4/5 homolog|stimulator of Fe transport|ubiquitin carrier protein D1|ubiquitin-conjugating enzyme E2(17)KB 1|ubiquitin-conjugating enzyme E2-17 kDa 1|ubiquitin-conjugating enzyme E2D 1 (UBC4/5 homolog, yeast)|ubiquitin-protein ligase D1', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58334979, end_position_on_the_genomic_accession=58370753, orientation='plus', exon_count=7, OMIM=602961, no_hdr0=''),
  7322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7322, CurrentID=0, Status='live', Symbol='UBE2D2', Aliases='E2(17)KB2, PUBC1, UBC4, UBC4/5, UBCH4, UBCH5B', description='ubiquitin conjugating enzyme E2D 2', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D2|E2 ubiquitin-conjugating enzyme D2|p53-regulated ubiquitin-conjugating enzyme 1|ubiquitin carrier protein D2|ubiquitin-conjugating enzyme E2-17 kDa 2|ubiquitin-conjugating enzyme E2D 2 (homologous to yeast UBC4/5)|ubiquitin-protein ligase D2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139561166, end_position_on_the_genomic_accession=139628434, orientation='plus', exon_count=9, OMIM=602962, no_hdr0=''),
  7323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7323, CurrentID=0, Status='live', Symbol='UBE2D3', Aliases='E2(17)KB3, UBC4/5, UBCH5C', description='ubiquitin conjugating enzyme E2D 3', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D3|E2 ubiquitin-conjugating enzyme D3|E2(17)KB 3|PRO2116|ubiquitin carrier protein D3|ubiquitin-conjugating enzyme E2(17)KB 3|ubiquitin-conjugating enzyme E2-17 kDa 3|ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast)|ubiquitin-conjugating enzyme E2D 3 (homologous to yeast UBC4/5)|ubiquitin-protein ligase D3', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102794383, end_position_on_the_genomic_accession=102868893, orientation='minus', exon_count=14, OMIM=602963, no_hdr0=''),
  7324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7324, CurrentID=0, Status='live', Symbol='UBE2E1', Aliases='UBCH6', description='ubiquitin conjugating enzyme E2E 1', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme E1|E2 ubiquitin-conjugating enzyme E1|ubiquitin carrier protein E1|ubiquitin-conjugating enzyme E2E 1 (UBC4/5 homolog, yeast)|ubiquitin-conjugating enzyme E2E 1 (homologous to yeast UBC4/5)|ubiquitin-protein ligase E1', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23805893, end_position_on_the_genomic_accession=23891640, orientation='plus', exon_count=7, OMIM=602916, no_hdr0=''),
  7325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7325, CurrentID=0, Status='live', Symbol='UBE2E2', Aliases='UBCH8', description='ubiquitin conjugating enzyme E2E 2', other_designations='E2 ubiquitin-conjugating enzyme E2|ubiquitin carrier protein E2|ubiquitin-conjugating enzyme E2E 2 (UBC4/5 homolog, yeast)|ubiquitin-conjugating enzyme E2E 2 (homologous to yeast UBC4/5)|ubiquitin-protein ligase E2', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23203007, end_position_on_the_genomic_accession=23590805, orientation='plus', exon_count=12, OMIM=602163, no_hdr0=''),
  7326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7326, CurrentID=0, Status='live', Symbol='UBE2G1', Aliases='E217K, UBC7, UBE2G', description='ubiquitin conjugating enzyme E2G 1', other_designations='E2 ubiquitin-conjugating enzyme G1|ubiquitin carrier protein G1|ubiquitin-conjugating enzyme E2G 1 (UBC7 homolog, C. elegans)|ubiquitin-conjugating enzyme E2G 1 (UBC7 homolog, yeast)|ubiquitin-conjugating enzyme E2G 1 (homologous to C. elegans UBC7)|ubiquitin-protein ligase G1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4269217, end_position_on_the_genomic_accession=4366674, orientation='minus', exon_count=6, OMIM=601569, no_hdr0=''),
  7327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7327, CurrentID=0, Status='live', Symbol='UBE2G2', Aliases='UBC7', description='ubiquitin conjugating enzyme E2G 2', other_designations='E2 ubiquitin-conjugating enzyme G2|ubiquitin carrier protein G2|ubiquitin conjugating enzyme 7|ubiquitin conjugating enzyme G2|ubiquitin-conjugating enzyme E2G 2 (UBC7 homolog, yeast)|ubiquitin-conjugating enzyme E2G 2 (homologous to yeast UBC7)|ubiquitin-protein ligase G2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44768580, end_position_on_the_genomic_accession=44801836, orientation='minus', exon_count=7, OMIM=603124, no_hdr0=''),
  7328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7328, CurrentID=0, Status='live', Symbol='UBE2H', Aliases='E2-20K, GID3, UBC8, UBCH, UBCH2', description='ubiquitin conjugating enzyme E2H', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme H|E2 ubiquitin-conjugating enzyme H|GID complex subunit 3, UBC8 homolog|ubiquitin carrier protein H|ubiquitin-conjugating enzyme E2-20K|ubiquitin-conjugating enzyme E2H (UBC8 homolog, yeast)|ubiquitin-conjugating enzyme E2H (homologous to yeast UBC8)|ubiquitin-protein ligase H', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129830732, end_position_on_the_genomic_accession=129952960, orientation='minus', exon_count=9, OMIM=601082, no_hdr0=''),
  7329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7329, CurrentID=0, Status='live', Symbol='UBE2I', Aliases='C358B7.1, P18, UBC9', description='ubiquitin conjugating enzyme E2I', other_designations='SUMO-1-protein ligase|SUMO-protein ligase|ubiquitin carrier protein 9|ubiquitin carrier protein I|ubiquitin conjugating enzyme 9|ubiquitin-conjugating enzyme E2I (UBC9 homolog, yeast)|ubiquitin-conjugating enzyme E2I (homologous to yeast UBC9)|ubiquitin-conjugating enzyme UbcE2A|ubiquitin-like protein SUMO-1 conjugating enzyme|ubiquitin-protein ligase E2I|ubiquitin-protein ligase I', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1309153, end_position_on_the_genomic_accession=1327018, orientation='plus', exon_count=10, OMIM=601661, no_hdr0=''),
  7332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7332, CurrentID=0, Status='live', Symbol='UBE2L3', Aliases='E2-F1, L-UBC, UBCH7, UbcM4', description='ubiquitin conjugating enzyme E2L 3', other_designations='E2 ubiquitin-conjugating enzyme L3|ubiquitin carrier protein L3|ubiquitin-conjugating enzyme E2-F1|ubiquitin-conjugating enzyme E2L 3|ubiquitin-conjugating enzyme UBCH7|ubiquitin-protein ligase L3', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21549447, end_position_on_the_genomic_accession=21624034, orientation='plus', exon_count=6, OMIM=603721, no_hdr0=''),
  7333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7333, CurrentID=0, Status='live', Symbol='UBE2L4', Aliases='', description='ubiquitin conjugating enzyme E2L 4 (pseudogene)', other_designations='ubiquitin-conjugating enzyme E2L 4 (pseudogene)', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9774497, end_position_on_the_genomic_accession=9774960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400550, CurrentID=0, Status='live', Symbol='FENDRR', Aliases='FOXF1-AS1, TCONS_00024240, lincFOXF1, onco-lncRNA-21', description='FOXF1 adjacent non-coding developmental regulatory RNA', other_designations='FOXF1 antisense RNA 1 (head to head)|FOXF1 antisense RNA 1 (non-protein coding)', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86474525, end_position_on_the_genomic_accession=86508860, orientation='minus', exon_count=7, OMIM=614975, no_hdr0=''),
  7335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7335, CurrentID=0, Status='live', Symbol='UBE2V1', Aliases='CIR1, CROC-1, CROC1, UBE2V, UEV-1, UEV1, UEV1A', description='ubiquitin conjugating enzyme E2 variant 1', other_designations='DNA-binding protein|TRAF6-regulated IKK activator 1 beta Uev1A', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50081124, end_position_on_the_genomic_accession=50115959, orientation='minus', exon_count=11, OMIM=602995, no_hdr0=''),
  7336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7336, CurrentID=0, Status='live', Symbol='UBE2V2', Aliases='DDVIT1, DDVit-1, EDAF-1, EDPF-1, EDPF1, MMS2, UEV-2, UEV2', description='ubiquitin conjugating enzyme E2 variant 2', other_designations='1 alpha,25-dihydroxyvitamin D3-inducible|DDVit 1|MMS2 homolog|enterocyte differentiation promoting factor|enterocyte differentiation-associated factor 1|enterocyte differentiation-associated factor EDAF-1|enterocyte differentiation-promoting factor 1|methyl methanesulfonate sensitive 2, S. cerevisiae, homolog of|vitamin D3-inducible protein', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47998288, end_position_on_the_genomic_accession=48064708, orientation='plus', exon_count=6, OMIM=603001, no_hdr0=''),
  7337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7337, CurrentID=0, Status='live', Symbol='UBE3A', Aliases='ANCR, AS, E6-AP, EPVE6AP, HPVE6A', description='ubiquitin protein ligase E3A', other_designations='CTCL tumor antigen se37-2|E6AP ubiquitin-protein ligase|human papilloma virus E6-associated protein|human papillomavirus E6-associated protein|oncogenic protein-associated protein E6-AP|renal carcinoma antigen NY-REN-54', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25337234, end_position_on_the_genomic_accession=25439043, orientation='minus', exon_count=17, OMIM=601623, no_hdr0=''),
  7338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7338, CurrentID=0, Status='live', Symbol='UBE3AP1', Aliases='', description='ubiquitin protein ligase E3A pseudogene 1', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97351783, end_position_on_the_genomic_accession=97352319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7339, CurrentID=0, Status='live', Symbol='UBE3AP2', Aliases='', description='ubiquitin protein ligase E3A pseudogene 2', other_designations='ubiquitin protein ligase, processed pseudogene', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31059996, end_position_on_the_genomic_accession=31063168, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  138412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138412, CurrentID=0, Status='live', Symbol='SLC25A6P2', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene 2', other_designations='ADP/ATP translocase', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31253781, end_position_on_the_genomic_accession=31254943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7341, CurrentID=0, Status='live', Symbol='SUMO1', Aliases='DAP1, GMP1, OFC10, PIC1, SENP2, SMT3, SMT3C, SMT3H3, UBL1', description='small ubiquitin-like modifier 1', other_designations='GAP modifying protein 1|SMT3 homolog 3|SMT3 suppressor of mif two 3 homolog 1|sentrin|ubiquitin-homology domain protein PIC1|ubiquitin-like protein SMT3C|ubiquitin-like protein UBL1', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202206180, end_position_on_the_genomic_accession=202238599, orientation='minus', exon_count=5, OMIM=601912, no_hdr0=''),
  7342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7342, CurrentID=0, Status='live', Symbol='UBP1', Aliases='LBP-1B, LBP-1a, LBP1A, LBP1B', description='upstream binding protein 1 (LBP-1a)', other_designations='transcription factor LBP-1', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33388336, end_position_on_the_genomic_accession=33440813, orientation='minus', exon_count=19, OMIM=609784, no_hdr0=''),
  7343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7343, CurrentID=0, Status='live', Symbol='UBTF', Aliases='NOR-90, UBF, UBF-1, UBF1, UBF2', description='upstream binding transcription factor, RNA polymerase I', other_designations='90-kDa nucleolus organizer region autoantigen|autoantigen NOR-90', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44205033, end_position_on_the_genomic_accession=44221626, orientation='minus', exon_count=25, OMIM=600673, no_hdr0=''),
  7345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7345, CurrentID=0, Status='live', Symbol='UCHL1', Aliases='HEL-117, HEL-S-53, NDGOA, PARK5, PGP 9.5, PGP9.5, PGP95, Uch-L1', description='ubiquitin C-terminal hydrolase L1', other_designations='epididymis luminal protein 117|epididymis secretory protein Li 53|neuron cytoplasmic protein 9.5|ubiquitin C-terminal hydrolase|ubiquitin carboxyl-terminal esterase L1 (ubiquitin thiolesterase)|ubiquitin thioesterase L1|ubiquitin thiolesterase', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41256881, end_position_on_the_genomic_accession=41268429, orientation='plus', exon_count=9, OMIM=191342, no_hdr0=''),
  7347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7347, CurrentID=0, Status='live', Symbol='UCHL3', Aliases='UCH-L3', description='ubiquitin C-terminal hydrolase L3', other_designations='ubiquitin carboxyl-terminal esterase L3 (ubiquitin thiolesterase)|ubiquitin thioesterase L3|ubiquitin thiolesterase', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75548867, end_position_on_the_genomic_accession=75606020, orientation='plus', exon_count=11, OMIM=603090, no_hdr0=''),
  7348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7348, CurrentID=0, Status='live', Symbol='UPK1B', Aliases='TSPAN20, UPIB, UPK1', description='uroplakin 1B', other_designations='UP1b|tetraspan|tetraspanin-20|tspan-20|uroplakin Ib', map_location='3q13.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119173578, end_position_on_the_genomic_accession=119205153, orientation='plus', exon_count=8, OMIM=602380, no_hdr0=''),
  7349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7349, CurrentID=0, Status='live', Symbol='UCN', Aliases='UI, UROC', description='urocortin', other_designations='prepro-urocortin|urocortin, preproprotein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27307397, end_position_on_the_genomic_accession=27308262, orientation='minus', exon_count=2, OMIM=600945, no_hdr0=''),
  400566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400566, CurrentID=0, Status='live', Symbol='C17orf97', Aliases='CK20, LIAT1', description='chromosome 17 open reading frame 97', other_designations='ligand of ATE1 protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=410327, end_position_on_the_genomic_accession=414666, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7351, CurrentID=0, Status='live', Symbol='UCP2', Aliases='BMIQ4, SLC25A8, UCPH', description='uncoupling protein 2 (mitochondrial, proton carrier)', other_designations='solute carrier family 25 member 8', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73974671, end_position_on_the_genomic_accession=73982844, orientation='minus', exon_count=8, OMIM=601693, no_hdr0=''),
  7352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7352, CurrentID=0, Status='live', Symbol='UCP3', Aliases='SLC25A9', description='uncoupling protein 3 (mitochondrial, proton carrier)', other_designations='solute carrier family 25 member 9', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74000281, end_position_on_the_genomic_accession=74009237, orientation='minus', exon_count=7, OMIM=602044, no_hdr0=''),
  7353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7353, CurrentID=0, Status='live', Symbol='UFD1L', Aliases='UFD1', description='ubiquitin fusion degradation 1 like (yeast)', other_designations='UB fusion protein 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19449941, end_position_on_the_genomic_accession=19479296, orientation='minus', exon_count=12, OMIM=601754, no_hdr0=''),
  7355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7355, CurrentID=0, Status='live', Symbol='SLC35A2', Aliases='CDG2M, CDGX, UDP-Gal-Tr, UGALT, UGAT, UGT, UGT1, UGT2, UGTL', description='solute carrier family 35 (UDP-galactose transporter), member A2', other_designations='solute carrier family 35 (UDP-galactose transporter), member 2', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48903180, end_position_on_the_genomic_accession=48911958, orientation='minus', exon_count=6, OMIM=314375, no_hdr0=''),
  7356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7356, CurrentID=0, Status='live', Symbol='SCGB1A1', Aliases='CC10, CC16, CCPBP, CCSP, UGB, UP-1, UP1', description='secretoglobin, family 1A, member 1 (uteroglobin)', other_designations='blastokinin|clara cell phospholipid-binding protein|clara cells 10 kDa secretory protein|urinary protein 1|urine protein 1', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62419035, end_position_on_the_genomic_accession=62423206, orientation='plus', exon_count=3, OMIM=192020, no_hdr0=''),
  7357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7357, CurrentID=0, Status='live', Symbol='UGCG', Aliases='GCS, GLCT1', description='UDP-glucose ceramide glucosyltransferase', other_designations='UDP-glucose:N-acylsphingosine D-glucosyltransferase|glucosylceramide synthase', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111896766, end_position_on_the_genomic_accession=111935374, orientation='plus', exon_count=10, OMIM=602874, no_hdr0=''),
  7358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7358, CurrentID=0, Status='live', Symbol='UGDH', Aliases='GDH, UDP-GlcDH, UDPGDH, UGD', description='UDP-glucose 6-dehydrogenase', other_designations='UDP-Glc dehydrogenase|UDP-glucose dehydrogenase|uridine diphospho-glucose dehydrogenase', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39498755, end_position_on_the_genomic_accession=39527598, orientation='minus', exon_count=13, OMIM=603370, no_hdr0=''),
  7360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7360, CurrentID=0, Status='live', Symbol='UGP2', Aliases='UDPG, UDPGP, UDPGP2, UGP1, UGPP1, UGPP2, pHC379', description='UDP-glucose pyrophosphorylase 2', other_designations='UDP-glucose diphosphorylase|UDP-glucose pyrophosphorylase 1|UGPase 2|UTP--glucose-1-phosphate uridylyltransferase 2|UTP-glucose-1-phosphate uridyltransferase|Uridyl diphosphate glucose pyrophosphorylase-1|uridyl diphosphate glucose pyrophosphorylase 2', map_location='2p14-p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63840950, end_position_on_the_genomic_accession=63891562, orientation='plus', exon_count=13, OMIM=191760, no_hdr0=''),
  7361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7361, CurrentID=0, Status='live', Symbol='UGT1A', Aliases='GNT1, UGT, UGT1@, UGT1A', description='UDP glucuronosyltransferase 1 family, polypeptide A complex locus', other_designations='UDP glucuronosyltransferase 1 family, polypeptide A cluster|UDP glycosyltransferase 1 family, polypeptide A cluster', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233585439, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400578, CurrentID=0, Status='live', Symbol='KRT16P2', Aliases='', description='keratin 16 pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16830483, end_position_on_the_genomic_accession=16832833, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7363, CurrentID=0, Status='live', Symbol='UGT2B4', Aliases='HLUG25, UDPGT2B4, UDPGTH1, UDPGTh-1, UGT2B11', description='UDP glucuronosyltransferase 2 family, polypeptide B4', other_designations='UDP-glucuronyltransferase, family 2, beta-4|UDPGT 2B4|hyodeoxycholic acid|hyodeoxycholic acid-specific UDPGT', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69480165, end_position_on_the_genomic_accession=69526014, orientation='minus', exon_count=7, OMIM=600067, no_hdr0=''),
  7364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7364, CurrentID=0, Status='live', Symbol='UGT2B7', Aliases='UDPGT 2B9, UDPGT2B7, UDPGTH2, UGT2B9', description='UDP glucuronosyltransferase 2 family, polypeptide B7', other_designations='3,4-catechol estrogen specific|3,4-catechol estrogen-specific UDPGT|UDP glucuronosyltransferase 2B7|UDP-glucuronosyltransferase 2B9|UDP-glucuronyltransferase, family 2, beta-7|UDPGT 2B7|UDPGTh-2', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69051249, end_position_on_the_genomic_accession=69112987, orientation='plus', exon_count=10, OMIM=600068, no_hdr0=''),
  7365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7365, CurrentID=0, Status='live', Symbol='UGT2B10', Aliases='UDPGT2B10', description='UDP glucuronosyltransferase 2 family, polypeptide B10', other_designations='UDPGT 2B10', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68815993, end_position_on_the_genomic_accession=68832017, orientation='plus', exon_count=6, OMIM=600070, no_hdr0=''),
  7366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7366, CurrentID=0, Status='live', Symbol='UGT2B15', Aliases='HLUG4, UDPGT 2B8, UDPGT2B15, UDPGTH3, UGT2B8', description='UDP glucuronosyltransferase 2 family, polypeptide B15', other_designations='UDP glycosyltransferase 2B15|UDP-glucuronosyltransferase 2B8|UDP-glucuronosyltransferase UGT2B15|UDP-glucuronyltransferase, family 2, beta-15|UDPGT 2B15|UDPGTh-3', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68646597, end_position_on_the_genomic_accession=68670776, orientation='minus', exon_count=6, OMIM=600069, no_hdr0=''),
  7367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7367, CurrentID=0, Status='live', Symbol='UGT2B17', Aliases='BMND12, UDPGT2B17', description='UDP glucuronosyltransferase 2 family, polypeptide B17', other_designations='C19-steroid-specific UDP-glucuronosyltransferase|C19-steroid-specific UDPGT|UDP glycosyltransferase 2 family, member B17|UDP-glucuronyltransferase, family 2, beta-17', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68537184, end_position_on_the_genomic_accession=68568527, orientation='minus', exon_count=6, OMIM=601903, no_hdr0=''),
  7368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7368, CurrentID=0, Status='live', Symbol='UGT8', Aliases='CGT, UGT4', description='UDP glycosyltransferase 8', other_designations='UDP-galactose ceramide galactosyltransferase|UDP-galactose-ceramide galactosyltransferase|ceramide UDP-galactosyltransferase|cerebroside synthase|uridine diphosphate glycosyltransferase 8', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114598403, end_position_on_the_genomic_accession=114678225, orientation='plus', exon_count=8, OMIM=601291, no_hdr0=''),
  7369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7369, CurrentID=0, Status='live', Symbol='UMOD', Aliases='ADMCKD2, FJHN, HNFJ, HNFJ1, MCKD2, THGP, THP', description='uromodulin', other_designations='Tamm-Horsfall urinary glycoprotein|uromucoid', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20333051, end_position_on_the_genomic_accession=20356301, orientation='minus', exon_count=15, OMIM=191845, no_hdr0=''),
  7371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7371, CurrentID=0, Status='live', Symbol='UCK2', Aliases='TSA903, UK, UMPK', description='uridine-cytidine kinase 2', other_designations='cytidine monophosphokinase 2|testis-specific protein TSA903|uridine kinase|uridine monophosphate kinase|uridine monophosphokinase 2', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165827495, end_position_on_the_genomic_accession=165911618, orientation='plus', exon_count=9, OMIM=609329, no_hdr0=''),
  7372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7372, CurrentID=0, Status='live', Symbol='UMPS', Aliases='OPRT', description='uridine monophosphate synthetase', other_designations="OMPdecase|OPRTase|UMP synthase|orotate phosphoribosyltransferase|orotidine 5'-phosphate decarboxylase", map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124730366, end_position_on_the_genomic_accession=124749273, orientation='plus', exon_count=7, OMIM=613891, no_hdr0=''),
  7373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7373, CurrentID=0, Status='live', Symbol='COL14A1', Aliases='UND', description='collagen, type XIV, alpha 1', other_designations='undulin (fibronectin-tenascin-related)', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120125108, end_position_on_the_genomic_accession=120373573, orientation='plus', exon_count=50, OMIM=120324, no_hdr0=''),
  7374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7374, CurrentID=0, Status='live', Symbol='UNG', Aliases='DGU, HIGM4, HIGM5, UDG1, UNG15, UNG2, UNG', description='uracil DNA glycosylase', other_designations='UDG|uracil-DNA glycosylase 1, uracil-DNA glycosylase 2', map_location='12q23-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109097594, end_position_on_the_genomic_accession=109110993, orientation='plus', exon_count=7, OMIM=191525, no_hdr0=''),
  400591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400591, CurrentID=0, Status='live', Symbol='C17orf102', Aliases='', description='chromosome 17 open reading frame 102', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34574123, end_position_on_the_genomic_accession=34579369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7376, CurrentID=0, Status='live', Symbol='NR1H2', Aliases='LXR-b, LXRB, NER, NER-I, RIP15, UNR', description='nuclear receptor subfamily 1 group H member 2', other_designations='LX receptor beta|liver X nuclear receptor beta|nuclear orphan receptor LXR-beta|nuclear receptor NER|nuclear receptor subfamily 1, group H, member 2|steroid hormone-nuclear receptor NER|ubiquitously-expressed nuclear receptor', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50376423, end_position_on_the_genomic_accession=50383028, orientation='plus', exon_count=10, OMIM=600380, no_hdr0=''),
  7378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7378, CurrentID=0, Status='live', Symbol='UPP1', Aliases='UDRPASE, UP, UPASE, UPP', description='uridine phosphorylase 1', other_designations='UPase 1|urdPase 1', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48088603, end_position_on_the_genomic_accession=48108733, orientation='plus', exon_count=11, OMIM=191730, no_hdr0=''),
  7379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7379, CurrentID=0, Status='live', Symbol='UPK2', Aliases='UP2, UPII', description='uroplakin 2', other_designations='uroplakin II', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118956298, end_position_on_the_genomic_accession=118958559, orientation='plus', exon_count=5, OMIM=611558, no_hdr0=''),
  7380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7380, CurrentID=0, Status='live', Symbol='UPK3A', Aliases='UP3A, UPIII, UPIIIA, UPK3', description='uroplakin 3A', other_designations='uroplakin 3|uroplakin III', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45284987, end_position_on_the_genomic_accession=45295879, orientation='plus', exon_count=7, OMIM=611559, no_hdr0=''),
  7381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7381, CurrentID=0, Status='live', Symbol='UQCRB', Aliases='MC3DN3, QCR7, QP-C, QPC, UQBC, UQBP, UQCR6, UQPC', description='ubiquinol-cytochrome c reductase binding protein', other_designations='complex III subunit 7|complex III subunit VII|mitochondrial ubiquinone-binding protein|ubiquinol-cytochrome c reductase complex 14 kDa protein|ubiquinol-cytochrome c reductase, complex III subunit VI', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96226676, end_position_on_the_genomic_accession=96235634, orientation='minus', exon_count=6, OMIM=191330, no_hdr0=''),
  7384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7384, CurrentID=0, Status='live', Symbol='UQCRC1', Aliases='D3S3191, QCR1, UQCR1', description='ubiquinol-cytochrome c reductase core protein I', other_designations='complex III subunit 1|core protein I|ubiquinol-cytochrome-c reductase complex core protein 1', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48598999, end_position_on_the_genomic_accession=48609665, orientation='minus', exon_count=13, OMIM=191328, no_hdr0=''),
  7385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7385, CurrentID=0, Status='live', Symbol='UQCRC2', Aliases='MC3DN5, QCR2, UQCR2', description='ubiquinol-cytochrome c reductase core protein II', other_designations='complex III subunit 2|cytochrome bc-1 complex core protein II|ubiquinol-cytochrome-c reductase complex core protein 2', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21953288, end_position_on_the_genomic_accession=21983347, orientation='plus', exon_count=14, OMIM=191329, no_hdr0=''),
  7386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7386, CurrentID=0, Status='live', Symbol='UQCRFS1', Aliases='RIP1, RIS1, RISP, UQCR5', description='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide 1', other_designations='Rieske iron-sulfur protein|complex III subunit 5|cytochrome b-c1 complex subunit 5|ubiquinol-cytochrome c reductase iron-sulfur subunit', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29207260, end_position_on_the_genomic_accession=29213229, orientation='minus', exon_count=2, OMIM=191327, no_hdr0=''),
  386743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386743, CurrentID=0, Status='live', Symbol='CDY21P', Aliases='', description='chromodomain protein, Y-linked 21 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25807304, end_position_on_the_genomic_accession=25808634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7388, CurrentID=0, Status='live', Symbol='UQCRH', Aliases='QCR6, UQCR8', description='ubiquinol-cytochrome c reductase hinge protein', other_designations='complex III subunit 6|cytochrome c1 non-heme 11 kDa protein|mitochondrial hinge protein|ubiquinol-cytochrome c reductase complex 11 kDa protein|ubiquinol-cytochrome c reductase, complex III subunit VIII', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46303613, end_position_on_the_genomic_accession=46316777, orientation='plus', exon_count=5, OMIM=613844, no_hdr0=''),
  7389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7389, CurrentID=0, Status='live', Symbol='UROD', Aliases='PCT, UPD', description='uroporphyrinogen decarboxylase', other_designations='uroporphyrinogen III decarboxylase', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45012133, end_position_on_the_genomic_accession=45015669, orientation='plus', exon_count=10, OMIM=613521, no_hdr0=''),
  7390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7390, CurrentID=0, Status='live', Symbol='UROS', Aliases='UROIIIS', description='uroporphyrinogen III synthase', other_designations='hydroxymethylbilane hydrolyase|uroporphyrinogen-III cosynthase', map_location='10q25.2-q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125784980, end_position_on_the_genomic_accession=125823268, orientation='minus', exon_count=20, OMIM=606938, no_hdr0=''),
  7391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7391, CurrentID=0, Status='live', Symbol='USF1', Aliases='FCHL, FCHL1, HYPLIP1, MLTF, MLTFI, UEF, bHLHb11', description='upstream transcription factor 1', other_designations='class B basic helix-loop-helix protein 11|major late transcription factor 1', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161039251, end_position_on_the_genomic_accession=161045979, orientation='minus', exon_count=12, OMIM=191523, no_hdr0=''),
  7392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7392, CurrentID=0, Status='live', Symbol='USF2', Aliases='FIP, bHLHb12', description='upstream transcription factor 2, c-fos interacting', other_designations='FOS-interacting protein|c-fos interacting protein|class B basic helix-loop-helix protein 12|major late transcription factor 2', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35268993, end_position_on_the_genomic_accession=35279815, orientation='plus', exon_count=12, OMIM=600390, no_hdr0=''),
  100271312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271312, CurrentID=0, Status='live', Symbol='RPL35AP6', Aliases='RPL35A_3_191', description='ribosomal protein L35a pseudogene 6', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246755511, end_position_on_the_genomic_accession=246755832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7396, CurrentID=0, Status='live', Symbol='USH1E', Aliases='', description='Usher syndrome 1E (autosomal recessive, severe)', other_designations='Usher syndrome-2A, autosomal recessive, severe', map_location='21q21', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602097, no_hdr0=''),
  23078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23078, CurrentID=0, Status='live', Symbol='VWA8', Aliases='KIAA0564', description='von Willebrand factor A domain containing 8', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41566825, end_position_on_the_genomic_accession=41961103, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  7398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7398, CurrentID=0, Status='live', Symbol='USP1', Aliases='UBP', description='ubiquitin specific peptidase 1', other_designations='deubiquitinating enzyme 1|hUBP|ubiquitin carboxyl terminal hydrolase 1|ubiquitin specific processing protease 1|ubiquitin specific protease 1|ubiquitin thioesterase 1|ubiquitin thiolesterase 1|ubiquitin-specific-processing protease 1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62436304, end_position_on_the_genomic_accession=62451804, orientation='plus', exon_count=10, OMIM=603478, no_hdr0=''),
  7399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7399, CurrentID=0, Status='live', Symbol='USH2A', Aliases='RP39, US2, USH2, dJ1111A8.1', description='Usher syndrome 2A (autosomal recessive, mild)', other_designations='usher syndrome 2A|usher syndrome type IIa protein|usher syndrome type-2A protein', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215622894, end_position_on_the_genomic_accession=216423396, orientation='minus', exon_count=72, OMIM=608400, no_hdr0=''),
  7401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7401, CurrentID=0, Status='live', Symbol='CLRN1', Aliases='RP61, USH3, USH3A', description='clarin 1', other_designations='Usher syndrome type-3 protein', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150918911, end_position_on_the_genomic_accession=150973020, orientation='minus', exon_count=7, OMIM=606397, no_hdr0=''),
  7402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7402, CurrentID=0, Status='live', Symbol='UTRN', Aliases='DMDL, DRP, DRP1', description='utrophin', other_designations='DRP-1|dystrophin-related protein 1', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144285298, end_position_on_the_genomic_accession=144853034, orientation='plus', exon_count=84, OMIM=128240, no_hdr0=''),
  7403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7403, CurrentID=0, Status='live', Symbol='KDM6A', Aliases='KABUK2, UTX, bA386N14.2', description='lysine (K)-specific demethylase 6A', other_designations='bA386N14.2 (ubiquitously transcribed X chromosome tetratricopeptide repeat protein (UTX))|histone demethylase UTX|ubiquitously transcribed tetratricopeptide repeat protein X-linked|ubiquitously-transcribed TPR gene on the X chromosome', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44873175, end_position_on_the_genomic_accession=45112779, orientation='plus', exon_count=32, OMIM=300128, no_hdr0=''),
  7404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7404, CurrentID=0, Status='live', Symbol='UTY', Aliases='KDM6AL1, UTY', description='ubiquitously transcribed tetratricopeptide repeat containing, Y-linked', other_designations='ubiquitous TPR motif protein UTY|ubiquitously transcribed TPR gene on Y chromosome|ubiquitously transcribed tetratricopeptide repeat gene, Y chromosome|ubiquitously transcribed tetratricopeptide repeat gene, Y-linked|ubiquitously-transcribed TPR protein on the Y chromosome|ubiquitously-transcribed Y chromosome tetratricopeptide repeat protein', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13233920, end_position_on_the_genomic_accession=13480670, orientation='minus', exon_count=45, OMIM=400009, no_hdr0=''),
  7405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7405, CurrentID=0, Status='live', Symbol='UVRAG', Aliases='DHTX, VPS38, p63', description='UV radiation resistance associated', other_designations='beclin 1 binding protein|disrupted in heterotaxy', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75815167, end_position_on_the_genomic_accession=76144238, orientation='plus', exon_count=21, OMIM=602493, no_hdr0=''),
  400622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400622, CurrentID=0, Status='live', Symbol='LOC400622', Aliases='', description='uncharacterized LOC400622', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77526795, end_position_on_the_genomic_accession=77536598, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7407, CurrentID=0, Status='live', Symbol='VARS', Aliases='G7A1, VARS2, VARS', description='valyl-tRNA synthetase', other_designations='protein G7a|valRS|valine tRNA ligase 1, cytoplasmic|valyl-tRNA synthetase 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31777518, end_position_on_the_genomic_accession=31795935, orientation='minus', exon_count=30, OMIM=192150, no_hdr0=''),
  7408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7408, CurrentID=0, Status='live', Symbol='VASP', Aliases='', description='vasodilator-stimulated phosphoprotein', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45507430, end_position_on_the_genomic_accession=45526989, orientation='plus', exon_count=13, OMIM=601703, no_hdr0=''),
  7409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7409, CurrentID=0, Status='live', Symbol='VAV1', Aliases='VAV', description='vav guanine nucleotide exchange factor 1', other_designations='vav 1 guanine nucleotide exchange factor|vav 1 oncogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6772668, end_position_on_the_genomic_accession=6857366, orientation='plus', exon_count=28, OMIM=164875, no_hdr0=''),
  7410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7410, CurrentID=0, Status='live', Symbol='VAV2', Aliases='VAV-2', description='vav guanine nucleotide exchange factor 2', other_designations='vav 2 guanine nucleotide exchange factor|vav 2 oncogene', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133761894, end_position_on_the_genomic_accession=133992374, orientation='minus', exon_count=32, OMIM=600428, no_hdr0=''),
  7411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7411, CurrentID=0, Status='live', Symbol='VBP1', Aliases='HIBBJ46, PFD3, PFDN3, VBP-1', description='von Hippel-Lindau binding protein 1', other_designations='VHL binding protein-1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155197007, end_position_on_the_genomic_accession=155239841, orientation='plus', exon_count=8, OMIM=300133, no_hdr0=''),
  7412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7412, CurrentID=0, Status='live', Symbol='VCAM1', Aliases='CD106, INCAM-100', description='vascular cell adhesion molecule 1', other_designations='CD106 antigen', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100719640, end_position_on_the_genomic_accession=100739045, orientation='plus', exon_count=9, OMIM=192225, no_hdr0=''),
  400629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400629, CurrentID=0, Status='live', Symbol='TEX19', Aliases='', description='testis expressed 19', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82359247, end_position_on_the_genomic_accession=82363776, orientation='plus', exon_count=2, OMIM=615647, no_hdr0=''),
  7414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7414, CurrentID=0, Status='live', Symbol='VCL', Aliases='CMD1W, CMH15, HEL114, MV, MVCL', description='vinculin', other_designations='epididymis luminal protein 114|metavinculin', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73998078, end_position_on_the_genomic_accession=74120160, orientation='plus', exon_count=22, OMIM=193065, no_hdr0=''),
  7415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7415, CurrentID=0, Status='live', Symbol='VCP', Aliases='ALS14, HEL-220, HEL-S-70, IBMPFD, IBMPFD1, TERA, p97', description='valosin containing protein', other_designations='15S Mg(2+)-ATPase p97 subunit|TER ATPase|epididymis luminal protein 220|epididymis secretory protein Li 70|valosin-containing protein|yeast Cdc48p homolog', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35056068, end_position_on_the_genomic_accession=35072742, orientation='minus', exon_count=17, OMIM=601023, no_hdr0=''),
  7416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7416, CurrentID=0, Status='live', Symbol='VDAC1', Aliases='PORIN, VDAC-1', description='voltage-dependent anion channel 1', other_designations='outer mitochondrial membrane protein porin 1|plasmalemmal porin|porin 31HL|porin 31HM', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133971875, end_position_on_the_genomic_accession=134005389, orientation='minus', exon_count=9, OMIM=604492, no_hdr0=''),
  7417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7417, CurrentID=0, Status='live', Symbol='VDAC2', Aliases='POR', description='voltage-dependent anion channel 2', other_designations='outer mitochondrial membrane protein porin 2', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75210154, end_position_on_the_genomic_accession=75231448, orientation='plus', exon_count=13, OMIM=193245, no_hdr0=''),
  7418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7418, CurrentID=0, Status='live', Symbol='VDAC1P4', Aliases='VDAC4, VDAC4P', description='voltage-dependent anion channel 1 pseudogene 4', other_designations='voltage-dependent anion channel 4 pseudogene', map_location='1q24-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180434796, end_position_on_the_genomic_accession=180436802, orientation='plus', exon_count=0, OMIM=610030, no_hdr0=''),
  7419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7419, CurrentID=0, Status='live', Symbol='VDAC3', Aliases='HD-VDAC3, VDAC-3', description='voltage-dependent anion channel 3', other_designations='outer mitochondrial membrane protein porin 3', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42391624, end_position_on_the_genomic_accession=42405937, orientation='plus', exon_count=11, OMIM=610029, no_hdr0=''),
  7420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7420, CurrentID=0, Status='live', Symbol='VDI', Aliases='DIPI', description='vesicular stomatitis virus defective interfering particle suppression', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=125260, no_hdr0=''),
  7421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7421, CurrentID=0, Status='live', Symbol='VDR', Aliases='NR1I1, PPP1R163', description='vitamin D (1,25- dihydroxyvitamin D3) receptor', other_designations='1,25-dihydroxyvitamin D3 receptor|nuclear receptor subfamily 1 group I member 1|protein phosphatase 1, regulatory subunit 163|vitamin D nuclear receptor variant 1|vitamin D receptor', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47841537, end_position_on_the_genomic_accession=47905031, orientation='minus', exon_count=11, OMIM=601769, no_hdr0=''),
  7422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7422, CurrentID=0, Status='live', Symbol='VEGFA', Aliases='MVCD1, VEGF, VPF', description='vascular endothelial growth factor A', other_designations='vascular permeability factor', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43770209, end_position_on_the_genomic_accession=43786487, orientation='plus', exon_count=9, OMIM=192240, no_hdr0=''),
  7423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7423, CurrentID=0, Status='live', Symbol='VEGFB', Aliases='VEGFL, VRF', description='vascular endothelial growth factor B', other_designations='VEGF-related factor', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64234584, end_position_on_the_genomic_accession=64239264, orientation='plus', exon_count=7, OMIM=601398, no_hdr0=''),
  7424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7424, CurrentID=0, Status='live', Symbol='VEGFC', Aliases='Flt4-L, LMPH1D, VRP', description='vascular endothelial growth factor C', other_designations='FLT4 ligand DHM|vascular endothelial growth factor-related protein', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176683534, end_position_on_the_genomic_accession=176792745, orientation='minus', exon_count=7, OMIM=601528, no_hdr0=''),
  7425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7425, CurrentID=0, Status='live', Symbol='VGF', Aliases='SCG7, SgVII', description='VGF nerve growth factor inducible', other_designations='neuro-endocrine specific protein VGF', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101162509, end_position_on_the_genomic_accession=101169935, orientation='minus', exon_count=5, OMIM=602186, no_hdr0=''),
  400643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400643, CurrentID=0, Status='live', Symbol='LINC00668', Aliases='', description='long intergenic non-protein coding RNA 668', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6925474, end_position_on_the_genomic_accession=6929869, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7428, CurrentID=0, Status='live', Symbol='VHL', Aliases='HRCA1, RCA11, pVHL, VHL', description='von Hippel-Lindau tumor suppressor, E3 ubiquitin protein ligase', other_designations='elongin binding protein|protein G7', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10141635, end_position_on_the_genomic_accession=10153670, orientation='plus', exon_count=4, OMIM=608537, no_hdr0=''),
  7429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7429, CurrentID=0, Status='live', Symbol='VIL1', Aliases='D2S1471, VIL', description='villin 1', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218419115, end_position_on_the_genomic_accession=218449525, orientation='plus', exon_count=20, OMIM=193040, no_hdr0=''),
  7430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7430, CurrentID=0, Status='live', Symbol='EZR', Aliases='CVIL, CVL, HEL-S-105, VIL2', description='ezrin', other_designations='cytovillin 2|epididymis secretory protein Li 105|p81|villin 2 (ezrin)|villin-2', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158765741, end_position_on_the_genomic_accession=158819424, orientation='minus', exon_count=14, OMIM=123900, no_hdr0=''),
  7431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7431, CurrentID=0, Status='live', Symbol='VIM', Aliases='CTRCT30, HEL113', description='vimentin', other_designations='epididymis luminal protein 113', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17227935, end_position_on_the_genomic_accession=17237593, orientation='plus', exon_count=9, OMIM=193060, no_hdr0=''),
  7432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7432, CurrentID=0, Status='live', Symbol='VIP', Aliases='PHM27', description='vasoactive intestinal peptide', other_designations='prepro-VIP', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152747923, end_position_on_the_genomic_accession=152759765, orientation='plus', exon_count=7, OMIM=192320, no_hdr0=''),
  7433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7433, CurrentID=0, Status='live', Symbol='VIPR1', Aliases='HVR1, II, PACAP-R-2, PACAP-R2, RDC1, V1RG, VAPC1, VIP-R-1, VIPR, VIRG, VPAC1, VPAC1R, VPCAP1R', description='vasoactive intestinal peptide receptor 1', other_designations='PACAP type II receptor|VIP and PACAP receptor 1|VIP receptor, type I|VPAC1 receptor|pituitary adenylate cyclase activating polypeptide receptor, type II', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42489299, end_position_on_the_genomic_accession=42537573, orientation='plus', exon_count=16, OMIM=192321, no_hdr0=''),
  7434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7434, CurrentID=0, Status='live', Symbol='VIPR2', Aliases='C16DUPq36.3, DUP7q36.3, PACAP-R-3, PACAP-R3, VIP-R-2, VPAC2, VPAC2R, VPCAP2R', description='vasoactive intestinal peptide receptor 2', other_designations='PACAP type III receptor|VIP and PACAP receptor 2|helodermin-preferring VIP receptor|pituitary adenylate cyclase-activating polypeptide type III receptor', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=159028175, end_position_on_the_genomic_accession=159144957, orientation='minus', exon_count=17, OMIM=601970, no_hdr0=''),
  7435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7435, CurrentID=0, Status='live', Symbol='VIS1', Aliases='HIS-1, HIS1', description='viral integration site 1', other_designations='', map_location='2q14-q21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164755, no_hdr0=''),
  7436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7436, CurrentID=0, Status='live', Symbol='VLDLR', Aliases='CAMRQ1, CARMQ1, CHRMQ1CH, VLDLR', description='very low density lipoprotein receptor', other_designations='VLDL receptor|VLDL-R', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2621679, end_position_on_the_genomic_accession=2656100, orientation='plus', exon_count=19, OMIM=192977, no_hdr0=''),
  7438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7438, CurrentID=0, Status='live', Symbol='VMD1', Aliases='', description='vitelliform macular dystrophy, atypical', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=153840, no_hdr0=''),
  7439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7439, CurrentID=0, Status='live', Symbol='BEST1', Aliases='ARB, BEST, BMD, RP50, TU15B, VMD2', description='bestrophin 1', other_designations='Best disease|Best1V1Delta2|vitelliform macular dystrophy protein 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61946715, end_position_on_the_genomic_accession=61964463, orientation='plus', exon_count=14, OMIM=607854, no_hdr0=''),
  7441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7441, CurrentID=0, Status='live', Symbol='VPREB1', Aliases='CD179a, IGI, IGVPB, VPREB', description='pre-B lymphocyte 1', other_designations='CD179 antigen-like family member A|v(pre)B protein', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22244780, end_position_on_the_genomic_accession=22245515, orientation='plus', exon_count=2, OMIM=605141, no_hdr0=''),
  7442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7442, CurrentID=0, Status='live', Symbol='TRPV1', Aliases='VR1', description='transient receptor potential cation channel, subfamily V, member 1', other_designations='OTRPC1|capsaicin receptor|osm-9-like TRP channel 1|transient receptor potential vanilloid 1a|transient receptor potential vanilloid 1b|vanilloid receptor subtype 1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3565446, end_position_on_the_genomic_accession=3609411, orientation='minus', exon_count=19, OMIM=602076, no_hdr0=''),
  7443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7443, CurrentID=0, Status='live', Symbol='VRK1', Aliases='PCH1, PCH1A', description='vaccinia related kinase 1', other_designations='vaccinia virus B1R-related kinase 1|vaccinia-related kinase 1', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96797347, end_position_on_the_genomic_accession=96881614, orientation='plus', exon_count=14, OMIM=602168, no_hdr0=''),
  7444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7444, CurrentID=0, Status='live', Symbol='VRK2', Aliases='', description='vaccinia related kinase 2', other_designations='vaccinia virus B1R-related kinase 2', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57907651, end_position_on_the_genomic_accession=58159920, orientation='plus', exon_count=21, OMIM=602169, no_hdr0=''),
  400661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400661, CurrentID=0, Status='live', Symbol='LOC400661', Aliases='', description='uncharacterized LOC400661', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76690029, end_position_on_the_genomic_accession=76693636, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7446, CurrentID=0, Status='live', Symbol='AVSD1', Aliases='AVCD, AVCD1, AVSD, VSD1', description='atrioventricular septal defect 1', other_designations='ventricular septal defect 1', map_location='1p31-p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606215, no_hdr0=''),
  7447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7447, CurrentID=0, Status='live', Symbol='VSNL1', Aliases='HLP3, HPCAL3, HUVISL1, VILIP, VILIP-1', description='visinin like 1', other_designations='VLP-1|hippocalcin-like protein 3|visinin-like 1', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17540540, end_position_on_the_genomic_accession=17656439, orientation='plus', exon_count=4, OMIM=600817, no_hdr0=''),
  7448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7448, CurrentID=0, Status='live', Symbol='VTN', Aliases='V75, VN, VNT', description='vitronectin', other_designations='S-protein|complement S-protein|epibolin|serum spreading factor|serum-spreading factor|somatomedin B', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28367277, end_position_on_the_genomic_accession=28370352, orientation='minus', exon_count=8, OMIM=193190, no_hdr0=''),
  7450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7450, CurrentID=0, Status='live', Symbol='VWF', Aliases='F8VWF, VWD', description='von Willebrand factor', other_designations='coagulation factor VIII VWF', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5948874, end_position_on_the_genomic_accession=6124670, orientation='minus', exon_count=52, OMIM=613160, no_hdr0=''),
  285231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285231, CurrentID=0, Status='live', Symbol='FBXW12', Aliases='FBW12, FBXO12, FBXO35', description='F-box and WD repeat domain containing 12', other_designations='F-box and WD-40 domain protein 12|F-box and WD-40 domain-containing protein 12|F-box only protein 35|F-box- and WD40-repeat-containing protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48372175, end_position_on_the_genomic_accession=48394791, orientation='plus', exon_count=11, OMIM=609075, no_hdr0=''),
  400668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400668, CurrentID=0, Status='live', Symbol='PRSS57', Aliases='NSP4, PRSSL1, UNQ782', description='protease, serine 57', other_designations='GLGL782|protease, serine, 57|protease, serine-like 1|serine protease 1-like protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=685546, end_position_on_the_genomic_accession=695531, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7453, CurrentID=0, Status='live', Symbol='WARS', Aliases='GAMMA-2, IFI53, IFP53', description='tryptophanyl-tRNA synthetase', other_designations='hWRS|interferon-induced protein 53|trpRS|tryptophan tRNA ligase 1, cytoplasmic', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100333788, end_position_on_the_genomic_accession=100376343, orientation='minus', exon_count=14, OMIM=191050, no_hdr0=''),
  7454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7454, CurrentID=0, Status='live', Symbol='WAS', Aliases='IMD2, SCNX, THC, THC1P, WASPA, WAS', description='Wiskott-Aldrich syndrome', other_designations='eczema-thrombocytopenia|thrombocytopenia 1 (X-linked)', map_location='Xp11.4-p11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48683790, end_position_on_the_genomic_accession=48691427, orientation='plus', exon_count=12, OMIM=300392, no_hdr0=''),
  7455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7455, CurrentID=0, Status='live', Symbol='ZAN', Aliases='', description='zonadhesin (gene/pseudogene)', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100733583, end_position_on_the_genomic_accession=100797797, orientation='plus', exon_count=48, OMIM=602372, no_hdr0=''),
  7456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7456, CurrentID=0, Status='live', Symbol='WIPF1', Aliases='PRPL-2, WAS2, WASPIP, WIP', description='WAS/WASL interacting protein family member 1', other_designations='WAS/WASL interacting protein family, member 1|WASP-interacting protein|Wiskott-Aldrich syndrome protein interacting protein|protein PRPL-2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174559574, end_position_on_the_genomic_accession=174682899, orientation='minus', exon_count=11, OMIM=602357, no_hdr0=''),
  400673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400673, CurrentID=0, Status='live', Symbol='VMAC', Aliases='', description='vimentin-type intermediate filament associated coiled-coil protein', other_designations='vimentin-type IF-associated coiled-coil protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5904841, end_position_on_the_genomic_accession=5910252, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7458, CurrentID=0, Status='live', Symbol='EIF4H', Aliases='WBSCR1, WSCR1, eIF-4H', description='eukaryotic translation initiation factor 4H', other_designations='Williams-Beuren syndrome chromosome region 1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74174376, end_position_on_the_genomic_accession=74197101, orientation='plus', exon_count=8, OMIM=603431, no_hdr0=''),
  7459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7459, CurrentID=0, Status='live', Symbol='WBSCR2', Aliases='WSCR2', description='Williams-Beuren syndrome chromosome region 2', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7461, CurrentID=0, Status='live', Symbol='CLIP2', Aliases='CLIP, CLIP-115, CYLN2, WBSCR3, WBSCR4, WSCR3, WSCR4', description='CAP-Gly domain containing linker protein 2', other_designations='Williams-Beuren syndrome chromosome region 3|Williams-Beuren syndrome chromosome region 4|cytoplasmic linker 2|cytoplasmic linker protein 115|cytoplasmic linker protein 2|williams-Beuren syndrome chromosomal region 3 protein|williams-Beuren syndrome chromosomal region 4 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74289475, end_position_on_the_genomic_accession=74405943, orientation='plus', exon_count=17, OMIM=603432, no_hdr0=''),
  7462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7462, CurrentID=0, Status='live', Symbol='LAT2', Aliases='HSPC046, LAB, NTAL, WBSCR15, WBSCR5, WSCR5', description='linker for activation of T-cells family member 2', other_designations='Williams-Beuren syndrome chromosomal region 15 protein|Williams-Beuren syndrome chromosomal region 5 protein|linker for activation of B-cells|linker for activation of T cells family, member 2|linker for activation of T cells, transmembrane adaptor 2|membrane-associated adapter molecule|non-T-cell activation linker', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74209392, end_position_on_the_genomic_accession=74229834, orientation='plus', exon_count=14, OMIM=605719, no_hdr0=''),
  7464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7464, CurrentID=0, Status='live', Symbol='CORO2A', Aliases='CLIPINB, IR10, WDR2', description='coronin, actin binding protein, 2A', other_designations='WD protein IR10|WD repeat-containing protein 2|WD-repeat protein 2|coronin 2A|coronin, actin-binding protein, 2A|coronin-like protein B', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98120975, end_position_on_the_genomic_accession=98192674, orientation='minus', exon_count=15, OMIM=602159, no_hdr0=''),
  7465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7465, CurrentID=0, Status='live', Symbol='WEE1', Aliases='WEE1Ahu, WEE1', description='WEE1 G2 checkpoint kinase', other_designations='WEE1 homolog|WEE1+ homolog|wee1A kinase', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9573650, end_position_on_the_genomic_accession=9593458, orientation='plus', exon_count=13, OMIM=193525, no_hdr0=''),
  7466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7466, CurrentID=0, Status='live', Symbol='WFS1', Aliases='CTRCT41, WFRS, WFS, WFSL', description='Wolfram syndrome 1 (wolframin)', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6269850, end_position_on_the_genomic_accession=6303265, orientation='plus', exon_count=8, OMIM=606201, no_hdr0=''),
  7467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7467, CurrentID=0, Status='live', Symbol='WHCR', Aliases='PRDS', description='Wolf-Hirschhorn syndrome chromosome region', other_designations='', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194190, no_hdr0=''),
  7468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7468, CurrentID=0, Status='live', Symbol='WHSC1', Aliases='MMSET, NSD2, REIIBP, TRX5, WHS', description='Wolf-Hirschhorn syndrome candidate 1', other_designations='IL5 promoter REII region-binding protein|multiple myeloma SET domain containing protein type III|nuclear SET domain-containing protein 2|probable histone-lysine N-methyltransferase NSD2|trithorax/ash1-related protein 5', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1871396, end_position_on_the_genomic_accession=1982207, orientation='plus', exon_count=29, OMIM=602952, no_hdr0=''),
  7469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7469, CurrentID=0, Status='live', Symbol='NELFA', Aliases='NELF-A, P/OKcl.15, WHSC2', description='negative elongation factor complex member A', other_designations='wolf-Hirschhorn syndrome candidate 2 protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1982714, end_position_on_the_genomic_accession=2041914, orientation='minus', exon_count=15, OMIM=606026, no_hdr0=''),
  7471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7471, CurrentID=0, Status='live', Symbol='WNT1', Aliases='BMND16, INT1, OI15', description='wingless-type MMTV integration site family member 1', other_designations='proto-oncogene Int-1 homolog|wingless-type MMTV integration site family, member 1 (oncogene INT1)', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48978453, end_position_on_the_genomic_accession=48982613, orientation='plus', exon_count=4, OMIM=164820, no_hdr0=''),
  7472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7472, CurrentID=0, Status='live', Symbol='WNT2', Aliases='INT1L1, IRP', description='wingless-type MMTV integration site family member 2', other_designations='Int-1-related protein|int-1-like protein 1|secreted growth factor', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117276631, end_position_on_the_genomic_accession=117323289, orientation='minus', exon_count=5, OMIM=147870, no_hdr0=''),
  7473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7473, CurrentID=0, Status='live', Symbol='WNT3', Aliases='INT4, TETAMS', description='wingless-type MMTV integration site family member 3', other_designations='WNT-3 proto-oncogene protein|proto-oncogene Int-4 homolog|wingless-type MMTV integration site family, member 3', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46762506, end_position_on_the_genomic_accession=46818760, orientation='minus', exon_count=5, OMIM=165330, no_hdr0=''),
  7474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7474, CurrentID=0, Status='live', Symbol='WNT5A', Aliases='hWNT5A', description='wingless-type MMTV integration site family member 5A', other_designations='WNT-5A protein|wingless-type MMTV integration site family, member 5A', map_location='3p21-p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55465715, end_position_on_the_genomic_accession=55505261, orientation='minus', exon_count=16, OMIM=164975, no_hdr0=''),
  7475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7475, CurrentID=0, Status='live', Symbol='WNT6', Aliases='', description='wingless-type MMTV integration site family member 6', other_designations='wingless-type MMTV integration site family, member 6', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218859823, end_position_on_the_genomic_accession=218874233, orientation='plus', exon_count=5, OMIM=604663, no_hdr0=''),
  7476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7476, CurrentID=0, Status='live', Symbol='WNT7A', Aliases='', description='wingless-type MMTV integration site family member 7A', other_designations='proto-oncogene Wnt7a protein|wingless-type MMTV integration site family, member 7A', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13818217, end_position_on_the_genomic_accession=13880121, orientation='minus', exon_count=6, OMIM=601570, no_hdr0=''),
  7477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7477, CurrentID=0, Status='live', Symbol='WNT7B', Aliases='', description='wingless-type MMTV integration site family member 7B', other_designations='wingless-type MMTV integration site family, member 7B', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45920362, end_position_on_the_genomic_accession=45977128, orientation='minus', exon_count=5, OMIM=601967, no_hdr0=''),
  7478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7478, CurrentID=0, Status='live', Symbol='WNT8A', Aliases='WNT8D', description='wingless-type MMTV integration site family member 8A', other_designations='WNT8d|protein Wnt-8d|wingless-type MMTV integration site family, member 8A', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138083892, end_position_on_the_genomic_accession=138096730, orientation='plus', exon_count=10, OMIM=606360, no_hdr0=''),
  7479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7479, CurrentID=0, Status='live', Symbol='WNT8B', Aliases='', description='wingless-type MMTV integration site family member 8B', other_designations='wingless-type MMTV integration site family, member 8B', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100463055, end_position_on_the_genomic_accession=100483642, orientation='plus', exon_count=6, OMIM=601396, no_hdr0=''),
  7480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7480, CurrentID=0, Status='live', Symbol='WNT10B', Aliases='SHFM6, WNT-12', description='wingless-type MMTV integration site family member 10B', other_designations='WNT-10B protein|wingless-type MMTV integration site family, member 10B', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48965340, end_position_on_the_genomic_accession=48971858, orientation='minus', exon_count=6, OMIM=601906, no_hdr0=''),
  7481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7481, CurrentID=0, Status='live', Symbol='WNT11', Aliases='HWNT11', description='wingless-type MMTV integration site family member 11', other_designations='wingless-type MMTV integration site family, member 11', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76186326, end_position_on_the_genomic_accession=76214008, orientation='minus', exon_count=9, OMIM=603699, no_hdr0=''),
  7482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7482, CurrentID=0, Status='live', Symbol='WNT2B', Aliases='WNT13', description='wingless-type MMTV integration site family member 2B', other_designations='XWNT2, Xenopus, homolog of|wingless-type MMTV integration site family, member 13|wingless-type MMTV integration site family, member 2B', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112466541, end_position_on_the_genomic_accession=112521288, orientation='plus', exon_count=7, OMIM=601968, no_hdr0=''),
  7483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7483, CurrentID=0, Status='live', Symbol='WNT9A', Aliases='WNT14', description='wingless-type MMTV integration site family member 9A', other_designations='wingless-type MMTV integration site family, member 14|wingless-type MMTV integration site family, member 9A', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227918658, end_position_on_the_genomic_accession=227947975, orientation='minus', exon_count=5, OMIM=602863, no_hdr0=''),
  7484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7484, CurrentID=0, Status='live', Symbol='WNT9B', Aliases='WNT14B, WNT15', description='wingless-type MMTV integration site family member 9B', other_designations='protein Wnt-14b|wingless-type MMTV integration site family, member 15|wingless-type MMTV integration site family, member 9B', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46833201, end_position_on_the_genomic_accession=46886775, orientation='plus', exon_count=6, OMIM=602864, no_hdr0=''),
  7485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7485, CurrentID=0, Status='live', Symbol='WRB', Aliases='CHD5', description='tryptophan rich basic protein', other_designations='congenital heart disease 5 protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39379676, end_position_on_the_genomic_accession=39397889, orientation='plus', exon_count=8, OMIM=602915, no_hdr0=''),
  7486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7486, CurrentID=0, Status='live', Symbol='WRN', Aliases='RECQ3, RECQL2, RECQL3', description='Werner syndrome, RecQ helicase-like', other_designations='DNA helicase, RecQ-like type 3|exonuclease WRN|recQ protein-like 2', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31033262, end_position_on_the_genomic_accession=31173761, orientation='plus', exon_count=37, OMIM=604611, no_hdr0=''),
  23093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23093, CurrentID=0, Status='live', Symbol='TTLL5', Aliases='CORD19, KIAA0998, STAMP', description='tubulin tyrosine ligase like 5', other_designations='SRC1 and TIF2 associated binding protein|SRC1 and TIF2-associated modulatory protein|tubulin tyrosine ligase-like family member 5|tubulin tyrosine ligase-like family, member 5', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75661208, end_position_on_the_genomic_accession=75955082, orientation='plus', exon_count=32, OMIM=612268, no_hdr0=''),
  7488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7488, CurrentID=0, Status='live', Symbol='WS2B', Aliases='', description='Waardenburg syndrome, type 2B', other_designations='', map_location='1p21-p13.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600193, no_hdr0=''),
  7489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7489, CurrentID=0, Status='live', Symbol='WSN', Aliases='BGMR', description='Waisman syndrome', other_designations='Waisman syndrome (basal ganglion disorder with mental retardation)', map_location='Xq27.3-qter', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=311510, no_hdr0=''),
  7490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7490, CurrentID=0, Status='live', Symbol='WT1', Aliases='AWT1, EWS-WT1, GUD, NPHS4, WAGR, WIT-2, WT33', description='Wilms tumor 1', other_designations='Wilms tumor protein isoform Ex4a(+)', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32387775, end_position_on_the_genomic_accession=32435535, orientation='minus', exon_count=11, OMIM=607102, no_hdr0=''),
  7491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7491, CurrentID=0, Status='live', Symbol='WT2', Aliases='ADCR, MTACR1', description='Wilms tumor 2', other_designations='multiple tumor-associated chromosome region 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194071, no_hdr0=''),
  400709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400709, CurrentID=0, Status='live', Symbol='SIGLEC16', Aliases='SIGLECP16, Siglec-P16', description='sialic acid binding Ig-like lectin 16 (gene/pseudogene)', other_designations='sialic acid binding Ig-like lectin, pseudogene 16|sialic acid-binding immunoglobulin-like lectin 16', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49969655, end_position_on_the_genomic_accession=49975819, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7494, CurrentID=0, Status='live', Symbol='XBP1', Aliases='TREB-5, TREB5, XBP-1, XBP2', description='X-box binding protein 1', other_designations='tax-responsive element-binding protein 5', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28794560, end_position_on_the_genomic_accession=28800572, orientation='minus', exon_count=6, OMIM=194355, no_hdr0=''),
  7495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7495, CurrentID=0, Status='live', Symbol='XBP1P1', Aliases='XBP, XBP1, XBPP1', description='X-box binding protein 1 pseudogene 1', other_designations='X-box binding protein pseudogene 1', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112885122, end_position_on_the_genomic_accession=112886687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400713, CurrentID=0, Status='live', Symbol='ZNF880', Aliases='', description='zinc finger protein 880', other_designations='zinc finger protein LOC400713', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52369917, end_position_on_the_genomic_accession=52385795, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7498, CurrentID=0, Status='live', Symbol='XDH', Aliases='XO, XOR', description='xanthine dehydrogenase', other_designations='xanthine oxidoreductase', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31334309, end_position_on_the_genomic_accession=31414749, orientation='minus', exon_count=37, OMIM=607633, no_hdr0=''),
  7499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7499, CurrentID=0, Status='live', Symbol='XG', Aliases='PBDX', description='Xg blood group', other_designations='XG glycoprotein|Xg blood group (pseudoautosomal boundary-divided on the X chromosome)', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2751820, end_position_on_the_genomic_accession=2816500, orientation='plus', exon_count=12, OMIM=300879, no_hdr0=''),
  7500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7500, CurrentID=0, Status='live', Symbol='XGY1', Aliases='XGPY, XGPY1', description='Xg pseudogene, Y-linked 1', other_designations='XG pseudogene on the Y chromosome|Xg blood group pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12484479, end_position_on_the_genomic_accession=12507348, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7501, CurrentID=0, Status='live', Symbol='XGR', Aliases='YG', description='XG and CD99 regulator', other_designations='XG/MIC2 regulator|expression of XG and MIC2 on erythrocytes', map_location='Xp22.32 and Yp11', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='314705,489500', no_hdr0=''),
  7502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7502, CurrentID=0, Status='live', Symbol='XIC', Aliases='SXI1, XCE, XIST', description='X chromosome inactivation center', other_designations='X chromosome controlling element', map_location='Xq12-q13', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7503, CurrentID=0, Status='live', Symbol='XIST', Aliases='DXS1089, DXS399E, LINC00001, NCRNA00001, SXI1, swd66', description='X inactive specific transcript (non-protein coding)', other_designations='X (inactive)-specific transcript (non-protein coding)|long intergenic non-protein coding RNA 1', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73820651, end_position_on_the_genomic_accession=73852753, orientation='minus', exon_count=6, OMIM=314670, no_hdr0=''),
  7504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7504, CurrentID=0, Status='live', Symbol='XK', Aliases='KX, MCLDS, NA, NAC, X1kR1, XK', description='X-linked Kx blood group', other_designations='Kell blood group precursor (McLeod phenotype)|Kx antigen|XK, Kell blood group complex subunit (McLeod syndrome)|XK-related protein 1|kell complex 37 kDa component|neuroacanthocytosis|neurocanthocytosis', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37685781, end_position_on_the_genomic_accession=37732130, orientation='plus', exon_count=4, OMIM=314850, no_hdr0=''),
  7506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7506, CurrentID=0, Status='live', Symbol='XM', Aliases='', description='Xm(a) antigen', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=314900, no_hdr0=''),
  7507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7507, CurrentID=0, Status='live', Symbol='XPA', Aliases='XP1C, XPA', description='xeroderma pigmentosum, complementation group A', other_designations='excision repair-controlling|mutant xeroderma pigmentosum complementation group A|xeroderma pigmentosum group A-complementing protein', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97654398, end_position_on_the_genomic_accession=97697409, orientation='minus', exon_count=10, OMIM=611153, no_hdr0=''),
  7508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7508, CurrentID=0, Status='live', Symbol='XPC', Aliases='RAD4, XP3C, p125, XPC', description='xeroderma pigmentosum, complementation group C', other_designations='mutant xeroderma pigmentosum group C', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14145147, end_position_on_the_genomic_accession=14178672, orientation='minus', exon_count=18, OMIM=613208, no_hdr0=''),
  7511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7511, CurrentID=0, Status='live', Symbol='XPNPEP1', Aliases='APP1, SAMP, XPNPEP, XPNPEPL, XPNPEPL1', description='X-prolyl aminopeptidase (aminopeptidase P) 1, soluble', other_designations='X-Pro aminopeptidase 1|X-prolyl aminopeptidase 1, soluble|aminoacylproline aminopeptidase|aminopeptidase P, cytosolic|cytosolic aminopeptidase P|soluble aminopeptidase P', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109864766, end_position_on_the_genomic_accession=109923553, orientation='minus', exon_count=23, OMIM=602443, no_hdr0=''),
  7512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7512, CurrentID=0, Status='live', Symbol='XPNPEP2', Aliases='AEACEI, APP2', description='X-prolyl aminopeptidase (aminopeptidase P) 2, membrane-bound', other_designations='X-Pro aminopeptidase 2|aminoacylproline aminopeptidase|mAmP|membrane-bound APP|membrane-bound AmP|membrane-bound aminopeptidase P', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129738970, end_position_on_the_genomic_accession=129769549, orientation='plus', exon_count=21, OMIM=300145, no_hdr0=''),
  7514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7514, CurrentID=0, Status='live', Symbol='XPO1', Aliases='CRM1, emb, exp1', description='exportin 1', other_designations='chromosome region maintenance 1 homolog|chromosome region maintenance 1 protein homolog|exportin 1 (CRM1 homolog, yeast)|exportin-1 (required for chromosome region maintenance)', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61477934, end_position_on_the_genomic_accession=61538522, orientation='minus', exon_count=29, OMIM=602559, no_hdr0=''),
  7515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7515, CurrentID=0, Status='live', Symbol='XRCC1', Aliases='RCC', description='X-ray repair complementing defective repair in Chinese hamster cells 1', other_designations='X-ray repair cross-complementing protein 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43543312, end_position_on_the_genomic_accession=43575578, orientation='minus', exon_count=17, OMIM=194360, no_hdr0=''),
  7516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7516, CurrentID=0, Status='live', Symbol='XRCC2', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 2', other_designations='X-ray repair cross-complementing protein 2', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152646498, end_position_on_the_genomic_accession=152676165, orientation='minus', exon_count=3, OMIM=600375, no_hdr0=''),
  7517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7517, CurrentID=0, Status='live', Symbol='XRCC3', Aliases='CMM6', description='X-ray repair complementing defective repair in Chinese hamster cells 3', other_designations='X-ray repair cross-complementing protein 3', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103697608, end_position_on_the_genomic_accession=103715486, orientation='minus', exon_count=10, OMIM=600675, no_hdr0=''),
  7518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7518, CurrentID=0, Status='live', Symbol='XRCC4', Aliases='SSMED', description='X-ray repair complementing defective repair in Chinese hamster cells 4', other_designations='X-ray repair cross-complementing protein 4', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83077409, end_position_on_the_genomic_accession=83353760, orientation='plus', exon_count=12, OMIM=194363, no_hdr0=''),
  400735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400735, CurrentID=0, Status='live', Symbol='PRAMEF4', Aliases='', description='PRAME family member 4', other_designations='RP5-845O24.6', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12879212, end_position_on_the_genomic_accession=12886201, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  7520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7520, CurrentID=0, Status='live', Symbol='XRCC5', Aliases='KARP-1, KARP1, KU80, KUB2, Ku86, NFIV', description='X-ray repair complementing defective repair in Chinese hamster cells 5', other_designations='86 kDa subunit of Ku antigen|ATP-dependent DNA helicase 2 subunit 2|ATP-dependent DNA helicase II 80 kDa subunit|CTC box-binding factor 85 kDa subunit|CTC85|CTCBF|DNA repair protein XRCC5|Ku autoantigen, 80kDa|Ku86 autoantigen related protein 1|TLAA|X-ray repair complementing defective repair in Chinese hamster cells 5 (double-strand-break rejoining)|lupus Ku autoantigen protein p86|nuclear factor IV|thyroid-lupus autoantigen', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216109297, end_position_on_the_genomic_accession=216206293, orientation='plus', exon_count=21, OMIM=194364, no_hdr0=''),
  7521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7521, CurrentID=0, Status='live', Symbol='XRCC8', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 8', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7523, CurrentID=0, Status='live', Symbol='XS', Aliases='LUXS', description='X-linked suppressor of LU antigens', other_designations='Lutheran suppressor, X-linked', map_location='Xp21.2-q21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309050, no_hdr0=''),
  7524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7524, CurrentID=0, Status='live', Symbol='XWNPEP', Aliases='', description='X-tryptophanyl aminopeptidase', other_designations='X-tryptophanyl aminopeptidase (aminopeptidase W)|aminopeptidase W', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7525, CurrentID=0, Status='live', Symbol='YES1', Aliases='HsT441, P61-YES, Yes, c-yes', description='YES proto-oncogene 1, Src family tyrosine kinase', other_designations='YES1 proto-oncogene, Src family tyrosine kinase|Yamaguchi sarcoma oncogene|cellular yes-1 protein|proto-oncogene c-Yes|proto-oncogene tyrosine-protein kinase YES|v-yes-1 Yamaguchi sarcoma viral oncogene homolog 1', map_location='18p11.31-p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=721588, end_position_on_the_genomic_accession=812326, orientation='minus', exon_count=13, OMIM=164880, no_hdr0=''),
  7526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7526, CurrentID=0, Status='live', Symbol='YES1P1', Aliases='D22S670, SYR, YES2, YESP', description='YES1 pseudogene 1', other_designations='v-yes-1 Yamaguchi sarcoma viral oncogene homolog 1 pseudogene|v-yes-1 Yamaguchi sarcoma viral oncogene homolog pseudogene', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25647141, end_position_on_the_genomic_accession=25652035, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481452, CurrentID=0, Status='live', Symbol='RNU6-859P', Aliases='', description='RNA, U6 small nuclear 859, pseudogene', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7528, CurrentID=0, Status='live', Symbol='YY1', Aliases='DELTA, INO80S, NF-E1, UCRBP, YIN-YANG-1', description='YY1 transcription factor', other_designations='INO80 complex subunit S|YY-1|Yin and Yang 1 protein|delta transcription factor', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100238765, end_position_on_the_genomic_accession=100279034, orientation='plus', exon_count=5, OMIM=600013, no_hdr0=''),
  7529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7529, CurrentID=0, Status='live', Symbol='YWHAB', Aliases='GW128, HEL-S-1, HS1, KCIP-1, YWHAA', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta', other_designations='14-3-3 alpha|brain protein 14-3-3, beta isoform|epididymis secretory protein Li 1|protein 1054|protein kinase C inhibitor protein 1|protein kinase C inhibitor protein-1|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, alpha polypeptide|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta polypeptide', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44885599, end_position_on_the_genomic_accession=44908532, orientation='plus', exon_count=7, OMIM=601289, no_hdr0=''),
  400746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400746, CurrentID=0, Status='live', Symbol='NCMAP', Aliases='C1orf130, MP11', description='non-compact myelin associated protein', other_designations='myelin protein of 11 kDa|noncompact myelin associated protein', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24556076, end_position_on_the_genomic_accession=24609328, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7531, CurrentID=0, Status='live', Symbol='YWHAE', Aliases='14-3-3E, HEL2, KCIP-1, MDCR, MDS', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon', other_designations='14-3-3 epsilon|epididymis luminal protein 2|mitochondrial import stimulation factor L subunit|protein kinase C inhibitor protein-1|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide|tyrosine 3/tryptophan 5 -monooxygenase activation protein, epsilon polypeptide', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1344540, end_position_on_the_genomic_accession=1400375, orientation='minus', exon_count=8, OMIM=605066, no_hdr0=''),
  7532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7532, CurrentID=0, Status='live', Symbol='YWHAG', Aliases='14-3-3GAMMA, PPP1R170', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, gamma', other_designations='14-3-3 gamma|KCIP-1|protein kinase C inhibitor protein 1|protein phosphatase 1, regulatory subunit 170|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, gamma polypeptide', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76326791, end_position_on_the_genomic_accession=76359025, orientation='minus', exon_count=2, OMIM=605356, no_hdr0=''),
  7533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7533, CurrentID=0, Status='live', Symbol='YWHAH', Aliases='YWHA1', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, eta', other_designations='14-3-3 eta|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, eta polypeptide', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31944493, end_position_on_the_genomic_accession=31957603, orientation='plus', exon_count=2, OMIM=113508, no_hdr0=''),
  7534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7534, CurrentID=0, Status='live', Symbol='YWHAZ', Aliases='14-3-3-zeta, HEL-S-3, HEL-S-93, HEL4, KCIP-1, YWHAD', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta', other_designations='14-3-3 delta|14-3-3 protein/cytosolic phospholipase A2|14-3-3 zeta|epididymis luminal protein 4|epididymis secretory protein Li 3|epididymis secretory protein Li 93|phospholipase A2|protein kinase C inhibitor protein 1|protein kinase C inhibitor protein-1|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, delta polypeptide|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide|tyrosine 3/tryptophan 5 -monooxygenase activation protein, zeta polypeptide', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100918576, end_position_on_the_genomic_accession=100954596, orientation='minus', exon_count=12, OMIM=601288, no_hdr0=''),
  7535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7535, CurrentID=0, Status='live', Symbol='ZAP70', Aliases='SRK, STCD, STD, TZK, ZAP-70', description='zeta chain of T cell receptor associated protein kinase 70kDa', other_designations='70 kDa zeta-associated protein|70 kDa zeta-chain associated protein|syk-related tyrosine kinase|zeta-chain (TCR) associated protein kinase 70kDa|zeta-chain associated protein kinase, 70kD', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97713568, end_position_on_the_genomic_accession=97744327, orientation='plus', exon_count=18, OMIM=176947, no_hdr0=''),
  7536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7536, CurrentID=0, Status='live', Symbol='SF1', Aliases='BBP, D11S636, MBBP, ZCCHC25, ZFM1, ZNF162', description='splicing factor 1', other_designations='mammalian branch point-binding protein|transcription factor ZFM1|zinc finger gene in MEN1 locus|zinc finger protein 162', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64764604, end_position_on_the_genomic_accession=64778844, orientation='minus', exon_count=16, OMIM=601516, no_hdr0=''),
  7538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7538, CurrentID=0, Status='live', Symbol='ZFP36', Aliases='G0S24, GOS24, NUP475, RNF162A, TIS11, TTP, zfp-36', description='ZFP36 ring finger protein', other_designations='G0/G1 switch regulatory protein 24|growth factor-inducible nuclear protein NUP475|tristetraproline|zinc finger protein 36 homolog|zinc finger protein 36, C3H type, homolog|zinc finger protein, C3H type, 36 homolog', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39406847, end_position_on_the_genomic_accession=39409412, orientation='plus', exon_count=2, OMIM=190700, no_hdr0=''),
  7539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7539, CurrentID=0, Status='live', Symbol='ZFP37', Aliases='ZNF906, zfp-37', description='ZFP37 zinc finger protein', other_designations='zinc finger protein homologous to Zfp37 in mouse', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113041815, end_position_on_the_genomic_accession=113056791, orientation='minus', exon_count=4, OMIM=602951, no_hdr0=''),
  400756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400756, CurrentID=0, Status='live', Symbol='LOC400756', Aliases='', description='uncharacterized LOC400756', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62688263, end_position_on_the_genomic_accession=62710694, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7541, CurrentID=0, Status='live', Symbol='ZBTB14', Aliases='ZF5, ZFP-161, ZFP-5, ZFP161, ZNF478', description='zinc finger and BTB domain containing 14', other_designations='ZFP161 zinc finger protein|zinc finger protein 161 homolog|zinc finger protein 478|zinc finger protein 5 homolog|zinc finger protein homologous to Zfp161 in mouse', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5289019, end_position_on_the_genomic_accession=5297053, orientation='minus', exon_count=6, OMIM=602126, no_hdr0=''),
  7542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7542, CurrentID=0, Status='live', Symbol='ZFPL1', Aliases='D11S750, MCG4', description='zinc finger protein-like 1', other_designations='zinc finger protein MCG4|zinc-finger protein in MEN1 region', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65084222, end_position_on_the_genomic_accession=65088402, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  7543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7543, CurrentID=0, Status='live', Symbol='ZFX', Aliases='ZNF926', description='zinc finger protein, X-linked', other_designations='X-linked zinc finger protein', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24148934, end_position_on_the_genomic_accession=24216255, orientation='plus', exon_count=17, OMIM=314980, no_hdr0=''),
  7544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7544, CurrentID=0, Status='live', Symbol='ZFY', Aliases='ZNF911', description='zinc finger protein, Y-linked', other_designations='', map_location='Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2934416, end_position_on_the_genomic_accession=2982508, orientation='plus', exon_count=10, OMIM=490000, no_hdr0=''),
  7545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7545, CurrentID=0, Status='live', Symbol='ZIC1', Aliases='ZIC, ZNF201', description='Zic family member 1', other_designations='Zic family member 1 (odd-paired homolog, Drosophila)|Zinc finger protein of the cerebellum 1|zinc finger protein 201', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147409394, end_position_on_the_genomic_accession=147416719, orientation='plus', exon_count=3, OMIM=600470, no_hdr0=''),
  7546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7546, CurrentID=0, Status='live', Symbol='ZIC2', Aliases='HPE5', description='Zic family member 2', other_designations='Zic family member 2 (odd-paired homolog, Drosophila)|Zinc finger protein of the cerebellum 2', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99981772, end_position_on_the_genomic_accession=99986765, orientation='plus', exon_count=3, OMIM=603073, no_hdr0=''),
  7547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7547, CurrentID=0, Status='live', Symbol='ZIC3', Aliases='HTX, HTX1, VACTERLX, ZNF203', description='Zic family member 3', other_designations='Zic family member 3 (odd-paired homolog, Drosophila)|heterotaxy 1|zinc finger protein 203|zinc finger protein of the cerebellum 3', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=137566187, end_position_on_the_genomic_accession=137572100, orientation='plus', exon_count=3, OMIM=300265, no_hdr0=''),
  7549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7549, CurrentID=0, Status='live', Symbol='ZNF2', Aliases='A1-5, ZNF661, Zfp661', description='zinc finger protein 2', other_designations='zinc finger protein 2.2|zinc finger protein 661', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95165417, end_position_on_the_genomic_accession=95184317, orientation='plus', exon_count=5, OMIM=194500, no_hdr0=''),
  7551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7551, CurrentID=0, Status='live', Symbol='ZNF3', Aliases='A8-51, HF.12, KOX25, PP838, Zfp113', description='zinc finger protein 3', other_designations='C2-H2 type zinc finger protein|zinc finger protein HF.12|zinc finger protein HZF3.1|zinc finger protein KOX25', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100063847, end_position_on_the_genomic_accession=100082548, orientation='minus', exon_count=8, OMIM=194510, no_hdr0=''),
  7552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7552, CurrentID=0, Status='live', Symbol='ZNF711', Aliases='CMPX1, MRX97, ZNF4, ZNF5, ZNF6, Zfp711, dJ75N13.1', description='zinc finger protein 711', other_designations='dJ75N13.1 (znf6-like)|zinc finger protein 6 (CMPX1)', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85243820, end_position_on_the_genomic_accession=85273362, orientation='plus', exon_count=10, OMIM=314990, no_hdr0=''),
  7553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7553, CurrentID=0, Status='live', Symbol='ZNF7', Aliases='HF.16, KOX4, zf30', description='zinc finger protein 7', other_designations='C2-H2 type zinc finger protein|Zinc finger protein-7 (KOX4)|zinc finger protein 7 (KOX 4, clone HF.16)|zinc finger protein HF.16|zinc finger protein KOX4|zinc finger protein zfp30', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144827518, end_position_on_the_genomic_accession=144847508, orientation='plus', exon_count=9, OMIM=194531, no_hdr0=''),
  7554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7554, CurrentID=0, Status='live', Symbol='ZNF8', Aliases='HF.18, Zfp128', description='zinc finger protein 8', other_designations='zinc finger protein 272|zinc finger protein 8 (clone HF.18)|zinc finger protein HF.18', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58278952, end_position_on_the_genomic_accession=58295888, orientation='plus', exon_count=4, OMIM=194532, no_hdr0=''),
  7555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7555, CurrentID=0, Status='live', Symbol='CNBP', Aliases='CNBP1, DM2, PROMM, RNF163, ZCCHC22, ZNF9', description='CCHC-type zinc finger, nucleic acid binding protein', other_designations='CCHC-type zinc finger nucleic acid binding protein|cellular nucleic acid binding protein|erythroid differentiation-related|sterol regulatory element-binding protein|zinc finger protein 273|zinc finger protein 9 (a cellular retroviral nucleic acid binding protein)', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129167815, end_position_on_the_genomic_accession=129183967, orientation='minus', exon_count=5, OMIM=116955, no_hdr0=''),
  7556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7556, CurrentID=0, Status='live', Symbol='ZNF10', Aliases='KOX1', description='zinc finger protein 10', other_designations='zinc finger protein 10 (KOX 1)|zinc finger protein KOX1', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133130628, end_position_on_the_genomic_accession=133159465, orientation='plus', exon_count=5, OMIM=194538, no_hdr0=''),
  106481453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481453, CurrentID=0, Status='live', Symbol='RNU6-863P', Aliases='', description='RNA, U6 small nuclear 863, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7559, CurrentID=0, Status='live', Symbol='ZNF12', Aliases='GIOT-3, HZF11, KOX3, ZNF325', description='zinc finger protein 12', other_designations='gonadotropin inducible transcription repressor 3|gonadotropin-inducible ovary transcription repressor 3|zinc finger protein 11|zinc finger protein 325|zinc finger protein KOX3', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6688433, end_position_on_the_genomic_accession=6706969, orientation='minus', exon_count=6, OMIM=194536, no_hdr0=''),
  7561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7561, CurrentID=0, Status='live', Symbol='ZNF14', Aliases='GIOT-4, KOX6', description='zinc finger protein 14', other_designations='gonadotropin inducible transcription repressor-4|gonadotropin-inducible ovary transcription repressor 4|zinc finger protein 14 (KOX 6)|zinc finger protein KOX6', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19710472, end_position_on_the_genomic_accession=19733112, orientation='minus', exon_count=4, OMIM=194556, no_hdr0=''),
  7562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7562, CurrentID=0, Status='live', Symbol='ZNF708', Aliases='KOX8, ZNF15, ZNF15L1', description='zinc finger protein 708', other_designations='zinc finger protein 15-like 1 (KOX 8)', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21291160, end_position_on_the_genomic_accession=21329439, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7564, CurrentID=0, Status='live', Symbol='ZNF16', Aliases='HZF1, KOX9', description='zinc finger protein 16', other_designations='zinc finger protein KOX9|zinc finger protein Kruppel type 9', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144930358, end_position_on_the_genomic_accession=144950888, orientation='minus', exon_count=5, OMIM=601262, no_hdr0=''),
  7565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7565, CurrentID=0, Status='live', Symbol='ZNF17', Aliases='HPF3, KOX10', description='zinc finger protein 17', other_designations='zinc finger protein 17 (HPF3, KOX 10)|zinc finger protein HPF3|zinc finger protein KOX10', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57411161, end_position_on_the_genomic_accession=57421953, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7566, CurrentID=0, Status='live', Symbol='ZNF18', Aliases='HDSG1, KOX11, ZKSCAN6, ZNF535, ZSCAN38, Zfp535', description='zinc finger protein 18', other_designations='heart development-specific gene 1 protein|zinc finger protein 18 (KOX 11)|zinc finger protein 535|zinc finger protein with KRAB and SCAN domains 6', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11977439, end_position_on_the_genomic_accession=11997510, orientation='minus', exon_count=11, OMIM=194524, no_hdr0=''),
  7567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7567, CurrentID=0, Status='live', Symbol='ZNF19', Aliases='KOX12', description='zinc finger protein 19', other_designations='zinc finger protein 19 (KOX 12)|zinc finger protein KOX12', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71474073, end_position_on_the_genomic_accession=71489351, orientation='minus', exon_count=6, OMIM=194525, no_hdr0=''),
  7568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7568, CurrentID=0, Status='live', Symbol='ZNF20', Aliases='KOX13', description='zinc finger protein 20', other_designations='zinc finger protein KOX13', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12131350, end_position_on_the_genomic_accession=12140407, orientation='minus', exon_count=4, OMIM=194557, no_hdr0=''),
  7569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7569, CurrentID=0, Status='live', Symbol='ZNF182', Aliases='HHZ150, KOX14, ZNF21, Zfp182', description='zinc finger protein 182', other_designations='zinc finger protein 21 (KOX 14)|zinc finger protein KOX14', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47974851, end_position_on_the_genomic_accession=48003995, orientation='minus', exon_count=7, OMIM=314993, no_hdr0=''),
  7570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7570, CurrentID=0, Status='live', Symbol='ZNF22', Aliases='HKR-T1, KOX15, ZNF422, Zfp422', description='zinc finger protein 22', other_designations='krox-26 protein|zinc finger protein 22 (KOX 15)|zinc finger protein KOX15|zinc finger protein Krox-26', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45000825, end_position_on_the_genomic_accession=45005329, orientation='plus', exon_count=2, OMIM=194529, no_hdr0=''),
  7571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7571, CurrentID=0, Status='live', Symbol='ZNF23', Aliases='KOX16, ZNF359, ZNF612, Zfp612', description='zinc finger protein 23', other_designations='kruppel-like zinc finger factor X31|zinc finger protein 23 (KOX 16)|zinc finger protein 32|zinc finger protein 359|zinc finger protein 612', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71447597, end_position_on_the_genomic_accession=71462252, orientation='minus', exon_count=7, OMIM=194527, no_hdr0=''),
  7572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7572, CurrentID=0, Status='live', Symbol='ZNF24', Aliases='KOX17, RSG-A, ZNF191, ZSCAN3, Zfp191', description='zinc finger protein 24', other_designations='retinoic acid suppression protein A|zinc finger and SCAN domain-containing protein 3|zinc finger protein 191|zinc finger protein 24 (KOX 17)|zinc finger protein KOX17', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35332214, end_position_on_the_genomic_accession=35344462, orientation='minus', exon_count=5, OMIM=194534, no_hdr0=''),
  100271342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271342, CurrentID=0, Status='live', Symbol='RPL37P18', Aliases='RPL37_8_1033', description='ribosomal protein L37 pseudogene 18', other_designations='', map_location='10p11.22-p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34663877, end_position_on_the_genomic_accession=34664101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7574, CurrentID=0, Status='live', Symbol='ZNF26', Aliases='HEL-179, KOX20', description='zinc finger protein 26', other_designations='epididymis luminal protein 179|zinc finger protein KOX20', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132986348, end_position_on_the_genomic_accession=133012568, orientation='plus', exon_count=9, OMIM=194537, no_hdr0=''),
  105370110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370110, CurrentID=0, Status='live', Symbol='LOC105370110', Aliases='', description='uncharacterized LOC105370110', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22883885, end_position_on_the_genomic_accession=22894490, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7576, CurrentID=0, Status='live', Symbol='ZNF28', Aliases='KOX24', description='zinc finger protein 28', other_designations='zinc finger protein KOX24', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52797408, end_position_on_the_genomic_accession=52821678, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  7577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7577, CurrentID=0, Status='live', Symbol='ZNF29P', Aliases='KOX26, ZNF29', description='zinc finger protein 29, pseudogene', other_designations='zinc finger protein 29 (KOX 26)', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15675476, end_position_on_the_genomic_accession=15675643, orientation='plus', exon_count=0, OMIM=194535, no_hdr0=''),
  400794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400794, CurrentID=0, Status='live', Symbol='LOC400794', Aliases='', description='uncharacterized LOC400794', other_designations='', map_location='1q23.3-q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165476842, end_position_on_the_genomic_accession=165582104, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7579, CurrentID=0, Status='live', Symbol='ZSCAN20', Aliases='KOX29, ZFP-31, ZNF31, ZNF360', description='zinc finger and SCAN domain containing 20', other_designations='zinc finger protein 31|zinc finger protein 360|zinc finger protein KOX29', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33472631, end_position_on_the_genomic_accession=33500693, orientation='plus', exon_count=9, OMIM=611315, no_hdr0=''),
  7580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7580, CurrentID=0, Status='live', Symbol='ZNF32', Aliases='KOX30', description='zinc finger protein 32', other_designations='C2H2-546|zinc finger protein 32 (KOX 30)|zinc finger protein KOX30', map_location='10q22-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43643859, end_position_on_the_genomic_accession=43648878, orientation='minus', exon_count=4, OMIM=194539, no_hdr0=''),
  7581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7581, CurrentID=0, Status='live', Symbol='ZNF33A', Aliases='KOX2, KOX31, KOX5, NF11A, ZNF11, ZNF11A, ZNF33, ZZAPK', description='zinc finger protein 33A', other_designations='brain my041 protein|zinc finger and ZAK-associated protein with KRAB domain|zinc finger protein 11A|zinc finger protein 33a (KOX 31)|zinc finger protein KOX31', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38010483, end_position_on_the_genomic_accession=38060067, orientation='plus', exon_count=10, OMIM=194521, no_hdr0=''),
  7582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7582, CurrentID=0, Status='live', Symbol='ZNF33B', Aliases='KOX2, KOX31, ZNF11B', description='zinc finger protein 33B', other_designations='zinc finger protein 11b (KOX 2)|zinc finger protein 33b (KOX 31)', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42589107, end_position_on_the_genomic_accession=42638836, orientation='minus', exon_count=7, OMIM=194522, no_hdr0=''),
  400799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400799, CurrentID=0, Status='live', Symbol='LINC01344', Aliases='GS1-122H1.2', description='long intergenic non-protein coding RNA 1344', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182203945, end_position_on_the_genomic_accession=182314061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7584, CurrentID=0, Status='live', Symbol='ZNF35', Aliases='HF.10, HF10, Zfp105', description='zinc finger protein 35', other_designations='zinc finger protein 35 (clone HF.10)|zinc finger protein HF.10', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44648741, end_position_on_the_genomic_accession=44660791, orientation='plus', exon_count=4, OMIM=194533, no_hdr0=''),
  100271344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271344, CurrentID=0, Status='live', Symbol='UBA52P9', Aliases='RPL40_4_1591', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 9', other_designations='ribosomal protein L40 pseudogene', map_location='18q11.2-q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27355422, end_position_on_the_genomic_accession=27355764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7586, CurrentID=0, Status='live', Symbol='ZKSCAN1', Aliases='9130423L19Rik, KOX18, PHZ-37, ZNF139, ZNF36, ZSCAN33', description='zinc finger with KRAB and SCAN domains 1', other_designations='zinc finger protein 139|zinc finger protein 36 (KOX 18)|zinc finger protein KOX18', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100015572, end_position_on_the_genomic_accession=100045143, orientation='plus', exon_count=8, OMIM=601260, no_hdr0=''),
  7587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7587, CurrentID=0, Status='live', Symbol='ZNF37A', Aliases='KOX21, ZNF37', description='zinc finger protein 37A', other_designations='zinc finger protein 37a (KOX 21)|zinc finger protein KOX21', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38094336, end_position_on_the_genomic_accession=38123352, orientation='plus', exon_count=9, OMIM=616085, no_hdr0=''),
  400804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400804, CurrentID=0, Status='live', Symbol='C1orf140', Aliases='', description='uncharacterized LOC400804', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221329928, end_position_on_the_genomic_accession=221336296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7589, CurrentID=0, Status='live', Symbol='ZSCAN21', Aliases='NY-REN-21, ZNF38, Zipro1', description='zinc finger and SCAN domain containing 21', other_designations='renal carcinoma antigen NY-REN-21|zfp-38|zinc finger protein 38 (KOX 25)|zinc finger protein 38 homolog|zinc finger protein NY-REN-21 antigen', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100049767, end_position_on_the_genomic_accession=100065042, orientation='plus', exon_count=6, OMIM=601261, no_hdr0=''),
  7592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7592, CurrentID=0, Status='live', Symbol='ZNF41', Aliases='MRX89', description='zinc finger protein 41', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47444779, end_position_on_the_genomic_accession=47483234, orientation='minus', exon_count=6, OMIM=314995, no_hdr0=''),
  7593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7593, CurrentID=0, Status='live', Symbol='MZF1', Aliases='MZF-1B, ZFP98, ZNF42, ZSCAN6, MZF1', description='myeloid zinc finger 1', other_designations='zinc finger and SCAN domain-containing protein 6|zinc finger protein 42 (myeloid-specific retinoic acid-responsive)', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58561917, end_position_on_the_genomic_accession=58574439, orientation='minus', exon_count=8, OMIM=194550, no_hdr0=''),
  7594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7594, CurrentID=0, Status='live', Symbol='ZNF43', Aliases='HTF6, KOX27, ZNF39L1', description='zinc finger protein 43', other_designations='zinc finger protein 39-like 1 (KOX 27)|zinc finger protein HTF6|zinc finger protein KOX27', map_location='19p13.1-p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21804949, end_position_on_the_genomic_accession=21852068, orientation='minus', exon_count=11, OMIM=603972, no_hdr0=''),
  7596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7596, CurrentID=0, Status='live', Symbol='ZNF45', Aliases='KOX5, ZNF13', description='zinc finger protein 45', other_designations='BRC1744|zinc finger protein 13|zinc finger protein 45 (a Kruppel-associated box (KRAB) domain polypeptide)|zinc finger protein KOX5', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43912624, end_position_on_the_genomic_accession=43935280, orientation='minus', exon_count=12, OMIM=194554, no_hdr0=''),
  7597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7597, CurrentID=0, Status='live', Symbol='ZBTB25', Aliases='C14orf51, KUP, ZNF46', description='zinc finger and BTB domain containing 25', other_designations='zinc finger protein 46|zinc finger protein KUP', map_location='14q23-q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64449106, end_position_on_the_genomic_accession=64505213, orientation='minus', exon_count=12, OMIM=194541, no_hdr0=''),
  400818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400818, CurrentID=0, Status='live', Symbol='NBPF9', Aliases='AE01', description='neuroblastoma breakpoint family member 9', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149054033, end_position_on_the_genomic_accession=149090861, orientation='minus', exon_count=26, OMIM=613999, no_hdr0=''),
  100271347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271347, CurrentID=0, Status='live', Symbol='RPLP2P4', Aliases='RPLP2_3_1225', description='ribosomal protein, large, P2 pseudogene 4', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31897820, end_position_on_the_genomic_accession=31898129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400823, CurrentID=0, Status='live', Symbol='FAM177B', Aliases='', description='family with sequence similarity 177 member B', other_designations='RP11-452F19.2|family with sequence similarity 177, member B', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222737216, end_position_on_the_genomic_accession=222753727, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  7608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7608, CurrentID=0, Status='live', Symbol='ZNF56', Aliases='ZNF742', description='zinc finger protein 56', other_designations='putatibe zinc finger protein 742|putative zinc finger protein 56|putative zinc finger protein 742|zinc finger protein 742', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19833472, end_position_on_the_genomic_accession=19837620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271348, CurrentID=0, Status='live', Symbol='RPLP2P5', Aliases='RPLP2_4_1568', description='ribosomal protein, large, P2 pseudogene 5', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62381596, end_position_on_the_genomic_accession=62382052, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403289, CurrentID=0, Status='live', Symbol='OR6C73P', Aliases='', description='olfactory receptor family 6 subfamily C member 73 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 73 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55396276, end_position_on_the_genomic_accession=55397273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400830, CurrentID=0, Status='live', Symbol='DEFB132', Aliases='BD-32, DEFB-32, DEFB32, HEL-75, KFLL827, UNQ827', description='defensin beta 132', other_designations='RP5-1103G7.6|beta-defensin 32|defensin HEL-75|defensin, beta 132|defensin, beta 32', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=257736, end_position_on_the_genomic_accession=261096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400831, CurrentID=0, Status='live', Symbol='C20orf202', Aliases='', description='chromosome 20 open reading frame 202', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1203454, end_position_on_the_genomic_accession=1208274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7617, CurrentID=0, Status='live', Symbol='ZNF66', Aliases='ZNF66P', description='zinc finger protein 66', other_designations='zinc finger protein 486 pseudogene|zinc finger protein 66, pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20776270, end_position_on_the_genomic_accession=20837476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7618, CurrentID=0, Status='live', Symbol='ZNF726P1', Aliases='ZNF67, ZNF67P', description='zinc finger protein 726 pseudogene 1', other_designations='zinc finger protein 67, pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23832282, end_position_on_the_genomic_accession=23832470, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7620, CurrentID=0, Status='live', Symbol='ZNF69', Aliases='Cos5', description='zinc finger protein 69', other_designations='hZNF3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11887786, end_position_on_the_genomic_accession=11914551, orientation='plus', exon_count=7, OMIM=194543, no_hdr0=''),
  7621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7621, CurrentID=0, Status='live', Symbol='ZNF70', Aliases='Cos17', description='zinc finger protein 70', other_designations='zinc finger protein N27C7-1', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23741584, end_position_on_the_genomic_accession=23751092, orientation='minus', exon_count=2, OMIM=194544, no_hdr0=''),
  400839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400839, CurrentID=0, Status='live', Symbol='FLJ33581', Aliases='', description='uncharacterized LOC400839', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24199767, end_position_on_the_genomic_accession=24224588, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7624, CurrentID=0, Status='live', Symbol='ZNF73', Aliases='Cos12, ZNF186, hZNF2', description='zinc finger protein 73', other_designations='zinc finger protein 186', map_location='22p', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194547, no_hdr0=''),
  7625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7625, CurrentID=0, Status='live', Symbol='ZNF74', Aliases='COS52, ZFP520, ZNF520, hZNF7', description='zinc finger protein 74', other_designations='zinc finger protein 520', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20394115, end_position_on_the_genomic_accession=20408463, orientation='plus', exon_count=6, OMIM=194548, no_hdr0=''),
  7626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7626, CurrentID=0, Status='live', Symbol='ZNF75D', Aliases='D8C6, ZKSCAN24, ZNF75, ZNF82, ZSCAN28', description='zinc finger protein 75D', other_designations='zinc finger protein 75|zinc finger protein 82', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135248589, end_position_on_the_genomic_accession=135344618, orientation='minus', exon_count=11, OMIM=314997, no_hdr0=''),
  7627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7627, CurrentID=0, Status='live', Symbol='ZNF75A', Aliases='', description='zinc finger protein 75a', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3305406, end_position_on_the_genomic_accession=3318851, orientation='plus', exon_count=8, OMIM=601473, no_hdr0=''),
  7628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7628, CurrentID=0, Status='live', Symbol='ZNF75BP', Aliases='ZNF75B', description='zinc finger protein 75B, pseudogene', other_designations='zinc finger protein 75D pseudogene', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=44007985, end_position_on_the_genomic_accession=44010533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7629, CurrentID=0, Status='live', Symbol='ZNF76', Aliases='D6S229E, ZNF523, Zfp523', description='zinc finger protein 76', other_designations='zinc finger protein 523|zinc finger protein 76 (expressed in testis)', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35259488, end_position_on_the_genomic_accession=35295987, orientation='plus', exon_count=17, OMIM=194549, no_hdr0=''),
  100533496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533496, CurrentID=0, Status='live', Symbol='TVP23C-CDRT4', Aliases='FAM18B2-CDRT4', description='TVP23C-CDRT4 readthrough', other_designations='FAM18B2-CDRT4 readthrough', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15436015, end_position_on_the_genomic_accession=15563631, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7634, CurrentID=0, Status='live', Symbol='ZNF80', Aliases='pT17', description='zinc finger protein 80', other_designations='ZNFPT17', map_location='3q13.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114234631, end_position_on_the_genomic_accession=114237578, orientation='minus', exon_count=1, OMIM=194553, no_hdr0=''),
  7637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7637, CurrentID=0, Status='live', Symbol='ZNF84', Aliases='HPF2', description='zinc finger protein 84', other_designations='zinc finger protein HPF2', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133037286, end_position_on_the_genomic_accession=133063304, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  7638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7638, CurrentID=0, Status='live', Symbol='ZNF221', Aliases='', description='zinc finger protein 221', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43951223, end_position_on_the_genomic_accession=43967600, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7639, CurrentID=0, Status='live', Symbol='ZNF85', Aliases='HPF4, HTF1', description='zinc finger protein 85', other_designations='zinc finger protein 85 (HPF4, HTF1)', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20923221, end_position_on_the_genomic_accession=20950697, orientation='plus', exon_count=8, OMIM=603899, no_hdr0=''),
  7643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7643, CurrentID=0, Status='live', Symbol='ZNF90', Aliases='HTF9', description='zinc finger protein 90', other_designations='zinc finger protein HTF9', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20077994, end_position_on_the_genomic_accession=20121168, orientation='plus', exon_count=4, OMIM=603973, no_hdr0=''),
  7644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7644, CurrentID=0, Status='live', Symbol='ZNF91', Aliases='HPF7, HTF10', description='zinc finger protein 91', other_designations='zinc finger protein 91 (HPF7, HTF10)|zinc finger protein HPF7|zinc finger protein HTF10', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23357696, end_position_on_the_genomic_accession=23395560, orientation='minus', exon_count=4, OMIM=603971, no_hdr0=''),
  400866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400866, CurrentID=0, Status='live', Symbol='LINC00114', Aliases='C21orf24, NCRNA00114', description='long intergenic non-protein coding RNA 114', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38738955, end_position_on_the_genomic_accession=38773477, orientation='minus', exon_count=9, OMIM=611723, no_hdr0=''),
  400867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400867, CurrentID=0, Status='live', Symbol='LOC400867', Aliases='', description='uncharacterized LOC400867', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38877291, end_position_on_the_genomic_accession=38938429, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  7652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7652, CurrentID=0, Status='live', Symbol='ZNF99', Aliases='C19orf9, F8281', description='zinc finger protein 99', other_designations='CTC-451A6.4|zinc finger protein ENSP00000375192', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22752183, end_position_on_the_genomic_accession=22784171, orientation='minus', exon_count=4, OMIM=603981, no_hdr0=''),
  400870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400870, CurrentID=0, Status='live', Symbol='LOC400870', Aliases='', description='uncharacterized LOC400870', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43363324, end_position_on_the_genomic_accession=43363963, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7661, CurrentID=0, Status='live', Symbol='ZNF90P1', Aliases='ZNF108, ZNF108P', description='zinc finger protein 90 pseudogene 1', other_designations='zinc finger protein 108 (Y9)|zinc finger protein 108, pseudogene', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101357293, end_position_on_the_genomic_accession=101358633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271357, CurrentID=0, Status='live', Symbol='RPS15AP21', Aliases='RPS15A_6_747', description='ribosomal protein S15a pseudogene 21', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121381439, end_position_on_the_genomic_accession=121381896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271358, CurrentID=0, Status='live', Symbol='RPS15AP16', Aliases='RPS15A_7_402', description='ribosomal protein S15a pseudogene 16', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128798841, end_position_on_the_genomic_accession=128799235, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7673, CurrentID=0, Status='live', Symbol='ZNF222', Aliases='', description='zinc finger protein 222', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44025332, end_position_on_the_genomic_accession=44033110, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7675, CurrentID=0, Status='live', Symbol='ZNF121', Aliases='D19S204, ZHC32, ZNF20', description='zinc finger protein 121', other_designations='zinc finger protein 121 (clone ZHC32)|zinc finger protein 20', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9565616, end_position_on_the_genomic_accession=9584545, orientation='minus', exon_count=7, OMIM=194628, no_hdr0=''),
  400892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400892, CurrentID=0, Status='live', Symbol='BCRP2', Aliases='BCR-2, BCR2, BCRL2', description='breakpoint cluster region pseudogene 2', other_designations='breakpoint cluster region-like 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21103016, end_position_on_the_genomic_accession=21122286, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7678, CurrentID=0, Status='live', Symbol='ZNF124', Aliases='HZF-16, HZF16, ZK7', description='zinc finger protein 124', other_designations='zinc finger protein HZF-16', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247121975, end_position_on_the_genomic_accession=247172438, orientation='minus', exon_count=8, OMIM=194631, no_hdr0=''),
  154197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154197, CurrentID=0, Status='live', Symbol='PNLDC1', Aliases='', description='PARN like, ribonuclease domain containing 1', other_designations='poly(A)-specific ribonuclease (PARN)-like domain containing 1', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159800249, end_position_on_the_genomic_accession=159820704, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  7681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7681, CurrentID=0, Status='live', Symbol='MKRN3', Aliases='CPPB2, D15S9, RNF63, ZFP127, ZNF127', description='makorin ring finger protein 3', other_designations='RING finger protein 63|zinc finger protein 127', map_location='15q11-q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23565307, end_position_on_the_genomic_accession=23568020, orientation='plus', exon_count=1, OMIM=603856, no_hdr0=''),
  7682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7682, CurrentID=0, Status='live', Symbol='MKRN4P', Aliases='MKRN4, MKRNP5, RNF64, ZNF-Xp, ZNF127-Xp, ZNF127L1', description='makorin ring finger protein 4, pseudogene', other_designations='makorin ring finger protein 1 pseudogene|makorin ring finger protein pseudogene 5|makorin, ring finger protein, 4|zinc finger protein 127-like 1', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40834254, end_position_on_the_genomic_accession=40837807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7683, CurrentID=0, Status='live', Symbol='MKRN5P', Aliases='MKRNP1, ZNF-Xq, ZNF127L2', description='makorin ring finger protein 5, pseudogene', other_designations='makorin ring finger protein 1 pseudogene|makorin ring finger protein pseudogene 1|zinc finger protein 127-like 2', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74159607, end_position_on_the_genomic_accession=74162671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7686, CurrentID=0, Status='live', Symbol='MKRN7P', Aliases='MKRNP3, ZNF127L5, dJ981L23.2', description='makorin ring finger protein 7, pseudogene', other_designations='makorin ring finger protein 1 pseudogene|makorin ring finger protein pseudogene 3|zinc finger protein 127-like 5', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46463660, end_position_on_the_genomic_accession=46465292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271361, CurrentID=0, Status='live', Symbol='RPS15P7', Aliases='RPS15_4_1197', description='ribosomal protein S15 pseudogene 7', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4324347, end_position_on_the_genomic_accession=4324839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7690, CurrentID=0, Status='live', Symbol='ZNF131', Aliases='ZBTB35, pHZ-10', description='zinc finger protein 131', other_designations='zinc finger and BTB domain containing 35|zinc finger protein 131 (clone pHZ-10)', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43120888, end_position_on_the_genomic_accession=43176324, orientation='plus', exon_count=12, OMIM=604073, no_hdr0=''),
  7691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7691, CurrentID=0, Status='live', Symbol='ZNF132', Aliases='pHZ-12', description='zinc finger protein 132', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58432807, end_position_on_the_genomic_accession=58440222, orientation='minus', exon_count=4, OMIM=604074, no_hdr0=''),
  7692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7692, CurrentID=0, Status='live', Symbol='ZNF133', Aliases='ZNF150, pHZ-13, pHZ-66', description='zinc finger protein 133', other_designations='zinc finger protein 133 (clone pHZ-13)|zinc finger protein 150 (pHZ-66)', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18288283, end_position_on_the_genomic_accession=18316996, orientation='plus', exon_count=10, OMIM=604075, no_hdr0=''),
  7693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7693, CurrentID=0, Status='live', Symbol='ZNF134', Aliases='pHZ-15', description='zinc finger protein 134', other_designations='zinc finger protein 134 (clone pHZ-15)', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57614462, end_position_on_the_genomic_accession=57622268, orientation='plus', exon_count=3, OMIM=604076, no_hdr0=''),
  7694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7694, CurrentID=0, Status='live', Symbol='ZNF135', Aliases='ZNF61, ZNF78L1, pHZ-17, pT3', description='zinc finger protein 135', other_designations='zinc finger protein 135 (clone pHZ-17)|zinc finger protein 61|zinc finger protein 78-like 1 (pT3)', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58059239, end_position_on_the_genomic_accession=58069755, orientation='plus', exon_count=7, OMIM=604077, no_hdr0=''),
  7695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7695, CurrentID=0, Status='live', Symbol='ZNF136', Aliases='pHZ-20', description='zinc finger protein 136', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12163057, end_position_on_the_genomic_accession=12189249, orientation='plus', exon_count=5, OMIM=604078, no_hdr0=''),
  7696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7696, CurrentID=0, Status='live', Symbol='ZNF137P', Aliases='ZNF137, pHZ-30', description='zinc finger protein 137, pseudogene', other_designations='zinc finger protein 137 (clone pHZ-30)|zinc finger protein pseudogene', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52596684, end_position_on_the_genomic_accession=52600152, orientation='plus', exon_count=3, OMIM=604079, no_hdr0=''),
  7697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7697, CurrentID=0, Status='live', Symbol='ZNF138', Aliases='pHZ-32', description='zinc finger protein 138', other_designations='zinc finger protein 138 (clone pHZ-32)', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64794388, end_position_on_the_genomic_accession=64853800, orientation='plus', exon_count=10, OMIM=604080, no_hdr0=''),
  7699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7699, CurrentID=0, Status='live', Symbol='ZNF140', Aliases='pHZ-39', description='zinc finger protein 140', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133079687, end_position_on_the_genomic_accession=133108803, orientation='plus', exon_count=10, OMIM=604082, no_hdr0=''),
  400916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400916, CurrentID=0, Status='live', Symbol='CHCHD10', Aliases='C22orf16, FTDALS2, IMMD, N27C7-4, SMAJ', description='coiled-coil-helix-coiled-coil-helix domain containing 10', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23765834, end_position_on_the_genomic_accession=23767972, orientation='minus', exon_count=4, OMIM=615903, no_hdr0=''),
  7701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7701, CurrentID=0, Status='live', Symbol='ZNF142', Aliases='pHZ-49', description='zinc finger protein 142', other_designations='HA4654|zinc finger protein 142 (clone pHZ-49)', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218637916, end_position_on_the_genomic_accession=218659632, orientation='minus', exon_count=11, OMIM=604083, no_hdr0=''),
  7702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7702, CurrentID=0, Status='live', Symbol='ZNF143', Aliases='SBF, STAF, pHZ-1', description='zinc finger protein 143', other_designations='SPH-binding factor|hStaf|selenocysteine tRNA gene transcription-activating factor|transcriptional activator Staf', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9460965, end_position_on_the_genomic_accession=9528524, orientation='plus', exon_count=18, OMIM=603433, no_hdr0=''),
  7703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7703, CurrentID=0, Status='live', Symbol='PCGF2', Aliases='MEL-18, RNF110, ZNF144', description='polycomb group ring finger 2', other_designations='DNA-binding protein Mel-18|ring finger protein 110|zinc finger protein 144', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38733897, end_position_on_the_genomic_accession=38749817, orientation='minus', exon_count=13, OMIM=600346, no_hdr0=''),
  7704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7704, CurrentID=0, Status='live', Symbol='ZBTB16', Aliases='PLZF, ZNF145', description='zinc finger and BTB domain containing 16', other_designations='promyelocytic leukaemia zinc finger|zinc finger protein 145 (Kruppel-like, expressed in promyelocytic leukemia)|zinc finger protein PLZF', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114059579, end_position_on_the_genomic_accession=114250676, orientation='plus', exon_count=10, OMIM=176797, no_hdr0=''),
  7705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7705, CurrentID=0, Status='live', Symbol='ZNF146', Aliases='OZF', description='zinc finger protein 146', other_designations='only zinc finger protein', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36214602, end_position_on_the_genomic_accession=36238774, orientation='plus', exon_count=4, OMIM=601505, no_hdr0=''),
  7706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7706, CurrentID=0, Status='live', Symbol='TRIM25', Aliases='EFP, RNF147, Z147, ZNF147', description='tripartite motif containing 25', other_designations='RING finger protein 147|RING-type E3 ubiquitin transferase|estrogen-responsive finger protein|tripartite motif protein TRIM25|tripartite motif-containing 25|tripartite motif-containing protein 25|ubiquitin/ISG15-conjugating enzyme TRIM25|zinc finger protein 147 (estrogen-responsive finger protein)|zinc finger protein-147', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56887909, end_position_on_the_genomic_accession=56914048, orientation='minus', exon_count=9, OMIM=600453, no_hdr0=''),
  7707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7707, CurrentID=0, Status='live', Symbol='ZNF148', Aliases='BERF-1, BFCOL1, HT-BETA, ZBP-89, ZFP148, pHZ-52', description='zinc finger protein 148', other_designations='CACCC box-binding protein|CLL-associated antigen KW-10|transcription factor ZBP-89|zinc finger DNA-binding protein 89', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125225669, end_position_on_the_genomic_accession=125375354, orientation='minus', exon_count=9, OMIM=601897, no_hdr0=''),
  7709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7709, CurrentID=0, Status='live', Symbol='ZBTB17', Aliases='MIZ-1, ZNF151, ZNF60, pHZ-67', description='zinc finger and BTB domain containing 17', other_designations='Myc-interacting Zn finger protein-1|zinc finger protein 151 (pHZ-67)|zinc finger protein 60', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15941869, end_position_on_the_genomic_accession=15976132, orientation='minus', exon_count=16, OMIM=604084, no_hdr0=''),
  7710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7710, CurrentID=0, Status='live', Symbol='ZNF154', Aliases='pHZ-92', description='zinc finger protein 154', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57696275, end_position_on_the_genomic_accession=57709211, orientation='minus', exon_count=5, OMIM=604085, no_hdr0=''),
  7711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7711, CurrentID=0, Status='live', Symbol='ZNF155', Aliases='pHZ-96', description='zinc finger protein 155', other_designations='KRAB A domain', map_location='19q13.2-q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43967703, end_position_on_the_genomic_accession=43998326, orientation='plus', exon_count=7, OMIM=604086, no_hdr0=''),
  7712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7712, CurrentID=0, Status='live', Symbol='ZNF157', Aliases='HZF22', description='zinc finger protein 157', other_designations='zinc finger protein 22|zinc finger protein HZF22', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47370600, end_position_on_the_genomic_accession=47413699, orientation='plus', exon_count=4, OMIM=300024, no_hdr0=''),
  400930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400930, CurrentID=0, Status='live', Symbol='DKFZp547B139', Aliases='', description='uncharacterized protein DKFZp547B139', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400931, CurrentID=0, Status='live', Symbol='MIRLET7BHG', Aliases='linc-Ppara', description='MIRLET7B host gene', other_designations='MIRLET7B host gene (non-protein coding)', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46085997, end_position_on_the_genomic_accession=46113928, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7716, CurrentID=0, Status='live', Symbol='VEZF1', Aliases='DB1, ZNF161', description='vascular endothelial zinc finger 1', other_designations='putative transcription factor DB1|zinc finger protein 161', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57971547, end_position_on_the_genomic_accession=57988254, orientation='minus', exon_count=7, OMIM=606747, no_hdr0=''),
  100271366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271366, CurrentID=0, Status='live', Symbol='RPS21P7', Aliases='RPS21_4_1681', description='ribosomal protein S21 pseudogene 7', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5013473, end_position_on_the_genomic_accession=5013724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7718, CurrentID=0, Status='live', Symbol='ZNF165', Aliases='CT53, LD65, ZSCAN7', description='zinc finger protein 165', other_designations='cancer/testis antigen 53|zinc finger and SCAN domain-containing protein 7', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28080704, end_position_on_the_genomic_accession=28089563, orientation='plus', exon_count=4, OMIM=600834, no_hdr0=''),
  400935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400935, CurrentID=0, Status='live', Symbol='IL17REL', Aliases='', description='interleukin 17 receptor E-like', other_designations='IL-17 receptor E-like|IL-17RE-like', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49994513, end_position_on_the_genomic_accession=50013198, orientation='minus', exon_count=16, OMIM=613414, no_hdr0=''),
  285489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285489, CurrentID=0, Status='live', Symbol='DOK7', Aliases='C4orf25, CMS10, CMS1B', description='docking protein 7', other_designations='downstream of tyrosine kinase 7', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3463306, end_position_on_the_genomic_accession=3501476, orientation='plus', exon_count=12, OMIM=610285, no_hdr0=''),
  100271367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271367, CurrentID=0, Status='live', Symbol='RPS23P6', Aliases='RPS23_4_1473', description='ribosomal protein S23 pseudogene 6', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12081237, end_position_on_the_genomic_accession=12081532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400940, CurrentID=0, Status='live', Symbol='LOC400940', Aliases='', description='uncharacterized LOC400940', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5981978, end_position_on_the_genomic_accession=5988232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400941, CurrentID=0, Status='live', Symbol='LINC00487', Aliases='', description='long intergenic non-protein coding RNA 487', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6729168, end_position_on_the_genomic_accession=6770311, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  7726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7726, CurrentID=0, Status='live', Symbol='TRIM26', Aliases='AFP, RNF95, ZNF173', description='tripartite motif containing 26', other_designations='RING finger protein 95|acid finger protein|tripartite motif-containing 26|widely expressed acid zinc finger protein|zinc finger protein 173', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30184453, end_position_on_the_genomic_accession=30213494, orientation='minus', exon_count=10, OMIM=600830, no_hdr0=''),
  7727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7727, CurrentID=0, Status='live', Symbol='ZNF174', Aliases='ZSCAN8', description='zinc finger protein 174', other_designations='AW-1|zinc finger and SCAN domain-containing protein 8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3401190, end_position_on_the_genomic_accession=3430851, orientation='plus', exon_count=6, OMIM=603900, no_hdr0=''),
  7728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7728, CurrentID=0, Status='live', Symbol='ZNF175', Aliases='OTK18', description='zinc finger protein 175', other_designations='zinc finger protein OTK18', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51571278, end_position_on_the_genomic_accession=51592510, orientation='plus', exon_count=6, OMIM=601139, no_hdr0=''),
  400945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400945, CurrentID=0, Status='live', Symbol='LINC01376', Aliases='', description='long intergenic non-protein coding RNA 1376', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18984658, end_position_on_the_genomic_accession=19036011, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  7730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7730, CurrentID=0, Status='live', Symbol='ZNF177', Aliases='PIGX', description='zinc finger protein 177', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9363020, end_position_on_the_genomic_accession=9382617, orientation='plus', exon_count=9, OMIM=601276, no_hdr0=''),
  138803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138803, CurrentID=0, Status='live', Symbol='OR13C3', Aliases='OR37G, OR9-8', description='olfactory receptor family 13 subfamily C member 3', other_designations='olfactory receptor OR9-8|olfactory receptor, family 13, subfamily C, member 3', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104535770, end_position_on_the_genomic_accession=104536813, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  138804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138804, CurrentID=0, Status='live', Symbol='OR13C4', Aliases='HSHTPCRX17, HTPCRX17, OR2K1, OR37F, OR9-7', description='olfactory receptor family 13 subfamily C member 4', other_designations='olfactory receptor OR9-7|olfactory receptor, family 13, subfamily C, member 4|olfactory receptor, family 2, subfamily K, member 1', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104526253, end_position_on_the_genomic_accession=104527209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7733, CurrentID=0, Status='live', Symbol='ZNF180', Aliases='HHZ168', description='zinc finger protein 180', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44474428, end_position_on_the_genomic_accession=44501750, orientation='minus', exon_count=10, OMIM=606740, no_hdr0=''),
  400950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400950, CurrentID=0, Status='live', Symbol='C2orf91', Aliases='', description='chromosome 2 open reading frame 91', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41935368, end_position_on_the_genomic_accession=41953803, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400952, CurrentID=0, Status='live', Symbol='LINC01121', Aliases='UNQ6975', description='long intergenic non-protein coding RNA 1121', other_designations='NGNL6975|putative uncharacterized protein UNQ6975/PRO21958', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45174341, end_position_on_the_genomic_accession=45254941, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  7737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7737, CurrentID=0, Status='live', Symbol='RNF113A', Aliases='Cwc24, RNF113, TTD5, ZNF183', description='ring finger protein 113A', other_designations='zinc finger protein 183 (RING finger, C3HC4 type)', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119870532, end_position_on_the_genomic_accession=119871828, orientation='minus', exon_count=1, OMIM=300951, no_hdr0=''),
  7738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7738, CurrentID=0, Status='live', Symbol='ZNF184', Aliases='', description='zinc finger protein 184', other_designations='zinc finger protein 184 (Kruppel-like)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27450742, end_position_on_the_genomic_accession=27473118, orientation='minus', exon_count=7, OMIM=602277, no_hdr0=''),
  7739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7739, CurrentID=0, Status='live', Symbol='ZNF185', Aliases='', description='zinc finger protein 185 (LIM domain)', other_designations='P1-A', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152898132, end_position_on_the_genomic_accession=152973481, orientation='plus', exon_count=27, OMIM=300381, no_hdr0=''),
  7741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7741, CurrentID=0, Status='live', Symbol='ZSCAN26', Aliases='SRE-ZBP, SREZBP, ZNF187', description='zinc finger and SCAN domain containing 26', other_designations='zinc finger protein 187', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28267010, end_position_on_the_genomic_accession=28278224, orientation='plus', exon_count=6, OMIM=616474, no_hdr0=''),
  400958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400958, CurrentID=0, Status='live', Symbol='LOC400958', Aliases='', description='uncharacterized LOC400958', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64901840, end_position_on_the_genomic_accession=64932447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7743, CurrentID=0, Status='live', Symbol='ZNF189', Aliases='', description='zinc finger protein 189', other_designations='', map_location='9q22-q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101398808, end_position_on_the_genomic_accession=101410660, orientation='plus', exon_count=6, OMIM=603132, no_hdr0=''),
  400960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400960, CurrentID=0, Status='live', Symbol='PCBP1-AS1', Aliases='', description='PCBP1 antisense RNA 1', other_designations='PCBP1 antisense RNA 1 (non-protein coding)', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69962263, end_position_on_the_genomic_accession=70087015, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  400961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400961, CurrentID=0, Status='live', Symbol='PAIP2B', Aliases='', description='poly(A) binding protein interacting protein 2B', other_designations='PABP-interacting protein 2B|PAIP-2B|poly(A)-binding protein-interacting protein 2B', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71182738, end_position_on_the_genomic_accession=71227114, orientation='minus', exon_count=6, OMIM=611018, no_hdr0=''),
  7746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7746, CurrentID=0, Status='live', Symbol='ZSCAN9', Aliases='PRD51, ZNF193', description='zinc finger and SCAN domain containing 9', other_designations='cell proliferation-inducing gene 12 protein|cell proliferation-inducing protein 12|zinc finger protein 193', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28212665, end_position_on_the_genomic_accession=28233487, orientation='plus', exon_count=8, OMIM=602246, no_hdr0=''),
  100271371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271371, CurrentID=0, Status='live', Symbol='RPS24P18', Aliases='RPS24_8_1590', description='ribosomal protein S24 pseudogene 18', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26098794, end_position_on_the_genomic_accession=26099335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7748, CurrentID=0, Status='live', Symbol='ZNF195', Aliases='HRF1, ZNFP104', description='zinc finger protein 195', other_designations='hypoxia-regulated factor-1|krueppel family zinc finger protein (znfp104)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3357927, end_position_on_the_genomic_accession=3379222, orientation='minus', exon_count=9, OMIM=602187, no_hdr0=''),
  400965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400965, CurrentID=0, Status='live', Symbol='LOC400965', Aliases='', description='uncharacterized LOC400965', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=87052749, end_position_on_the_genomic_accession=87055359, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7750, CurrentID=0, Status='live', Symbol='ZMYM2', Aliases='FIM, MYM, RAMP, SCLL, ZNF198', description='zinc finger, MYM-type 2', other_designations='fused in myeloproliferative disorders protein|rearranged in an atypical myeloproliferative disorder|zinc finger protein 198', map_location='13q11-q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19957432, end_position_on_the_genomic_accession=20091845, orientation='plus', exon_count=36, OMIM=602221, no_hdr0=''),
  7752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7752, CurrentID=0, Status='live', Symbol='ZNF200', Aliases='', description='zinc finger protein 200', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3222325, end_position_on_the_genomic_accession=3236221, orientation='minus', exon_count=7, OMIM=603231, no_hdr0=''),
  7753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7753, CurrentID=0, Status='live', Symbol='ZNF202', Aliases='ZKSCAN10, ZSCAN42', description='zinc finger protein 202', other_designations='zinc finger protein with KRAB and SCAN domains 10', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123723927, end_position_on_the_genomic_accession=123741683, orientation='minus', exon_count=11, OMIM=603430, no_hdr0=''),
  7754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7754, CurrentID=0, Status='live', Symbol='ZNF204P', Aliases='ZNF184-Lp, ZNF204, ZNF315P, b24o18.1', description='zinc finger protein 204, pseudogene', other_designations='zinc finger protein 315 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27357823, end_position_on_the_genomic_accession=27375374, orientation='minus', exon_count=5, OMIM=603282, no_hdr0=''),
  7755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7755, CurrentID=0, Status='live', Symbol='ZNF205', Aliases='RhitH, ZNF210, Zfp13', description='zinc finger protein 205', other_designations='transcriptional repressor RhitH|zinc finger protein 210', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3112562, end_position_on_the_genomic_accession=3120527, orientation='plus', exon_count=8, OMIM=603436, no_hdr0=''),
  7756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7756, CurrentID=0, Status='live', Symbol='ZNF207', Aliases='BuGZ, hBuGZ', description='zinc finger protein 207', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32350138, end_position_on_the_genomic_accession=32408085, orientation='plus', exon_count=17, OMIM=603428, no_hdr0=''),
  7757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7757, CurrentID=0, Status='live', Symbol='ZNF208', Aliases='PMIDP, ZNF95', description='zinc finger protein 208', other_designations='zinc finger protein 95', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21966095, end_position_on_the_genomic_accession=22010943, orientation='minus', exon_count=4, OMIM=603977, no_hdr0=''),
  100271373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271373, CurrentID=0, Status='live', Symbol='RPS25P9', Aliases='RPS25_7_1060', description='ribosomal protein S25 pseudogene 9', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70198973, end_position_on_the_genomic_accession=70199239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7760, CurrentID=0, Status='live', Symbol='ZNF213', Aliases='CR53, ZKSCAN21, ZSCAN53', description='zinc finger protein 213', other_designations='putative transcription factor CR53|zinc finger protein with KRAB and SCAN domains 21', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3135056, end_position_on_the_genomic_accession=3142804, orientation='plus', exon_count=6, OMIM=608387, no_hdr0=''),
  7761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7761, CurrentID=0, Status='live', Symbol='ZNF214', Aliases='BAZ-1, BAZ1', description='zinc finger protein 214', other_designations='BAZ 1|BWSCR2-associated zinc finger protein 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6996613, end_position_on_the_genomic_accession=7020376, orientation='minus', exon_count=4, OMIM=605015, no_hdr0=''),
  7762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7762, CurrentID=0, Status='live', Symbol='ZNF215', Aliases='BAZ-2, BAZ2, ZKSCAN11, ZSCAN43', description='zinc finger protein 215', other_designations='BAZ 2|BWSCR2-associated zinc finger protein 2|zinc finger protein with KRAB and SCAN domains 11', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6926418, end_position_on_the_genomic_accession=7005354, orientation='plus', exon_count=13, OMIM=605016, no_hdr0=''),
  7763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7763, CurrentID=0, Status='live', Symbol='ZFAND5', Aliases='ZA20D2A, ZNF216, ZFAND5', description='zinc finger, AN1-type domain 5', other_designations='zinc finger A20 domain-containing protein 2|zinc finger protein 216|zinc finger, A20 domain containing 2', map_location='9q13-q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72351425, end_position_on_the_genomic_accession=72365247, orientation='minus', exon_count=8, OMIM=604761, no_hdr0=''),
  7764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7764, CurrentID=0, Status='live', Symbol='ZNF217', Aliases='ZABC1', description='zinc finger protein 217', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53567065, end_position_on_the_genomic_accession=53608969, orientation='minus', exon_count=10, OMIM=602967, no_hdr0=''),
  100271374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271374, CurrentID=0, Status='live', Symbol='RPS27AP5', Aliases='RPS27A_1_141', description='ribosomal protein S27a pseudogene 5', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192716135, end_position_on_the_genomic_accession=192716683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7766, CurrentID=0, Status='live', Symbol='ZNF223', Aliases='', description='zinc finger protein 223', other_designations='Homo sapiens zinc finger protein 223|KRAB A domain', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44052011, end_position_on_the_genomic_accession=44067994, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7767, CurrentID=0, Status='live', Symbol='ZNF224', Aliases='BMZF-2, BMZF2, KOX22, ZNF255, ZNF27', description='zinc finger protein 224', other_designations='bone marrow zinc finger 2|zinc finger 2, bone marrow|zinc finger protein 233|zinc finger protein 255|zinc finger protein 27 (KOX 22)|zinc finger protein KOX22|zinc finger protein ZNF255', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44094329, end_position_on_the_genomic_accession=44109831, orientation='plus', exon_count=6, OMIM=194555, no_hdr0=''),
  7768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7768, CurrentID=0, Status='live', Symbol='ZNF225', Aliases='', description='zinc finger protein 225', other_designations='C2-H2 type zinc finger protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44112076, end_position_on_the_genomic_accession=44134816, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  7769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7769, CurrentID=0, Status='live', Symbol='ZNF226', Aliases='', description='zinc finger protein 226', other_designations='Kruppel-associated box protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44165062, end_position_on_the_genomic_accession=44177685, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  7770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7770, CurrentID=0, Status='live', Symbol='ZNF227', Aliases='', description='zinc finger protein 227', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44207547, end_position_on_the_genomic_accession=44237268, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  7771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7771, CurrentID=0, Status='live', Symbol='ZNF112', Aliases='ZFP112, ZNF228', description='zinc finger protein 112', other_designations='zfp-112|zinc finger protein 112 (Y14)|zinc finger protein 112 homolog|zinc finger protein 228', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44326553, end_position_on_the_genomic_accession=44356704, orientation='minus', exon_count=5, OMIM=603994, no_hdr0=''),
  7772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7772, CurrentID=0, Status='live', Symbol='ZNF229', Aliases='', description='zinc finger protein 229', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44426251, end_position_on_the_genomic_accession=44448725, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7773, CurrentID=0, Status='live', Symbol='ZNF230', Aliases='FDZF2', description='zinc finger protein 230', other_designations='zinc finger protein FDZF2', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44002925, end_position_on_the_genomic_accession=44013926, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7775, CurrentID=0, Status='live', Symbol='ZNF232', Aliases='ZSCAN11', description='zinc finger protein 232', other_designations='zinc finger and SCAN domain-containing protein 11', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5105734, end_position_on_the_genomic_accession=5123102, orientation='minus', exon_count=6, OMIM=616463, no_hdr0=''),
  7776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7776, CurrentID=0, Status='live', Symbol='ZNF236', Aliases='ZNF236AB, ZNF236', description='zinc finger protein 236', other_designations='regulated by glucose', map_location='18q22-q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76822561, end_position_on_the_genomic_accession=76970732, orientation='plus', exon_count=38, OMIM=604760, no_hdr0=''),
  100271376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271376, CurrentID=0, Status='live', Symbol='RPS27AP9', Aliases='RPS27A_3_585', description='ribosomal protein S27a pseudogene 9', other_designations='', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80498530, end_position_on_the_genomic_accession=80498972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7779, CurrentID=0, Status='live', Symbol='SLC30A1', Aliases='ZNT1, ZRC1', description='solute carrier family 30 (zinc transporter), member 1', other_designations='solute carrier family 30 member 1|znT-1', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211575039, end_position_on_the_genomic_accession=211578757, orientation='minus', exon_count=2, OMIM=609521, no_hdr0=''),
  7780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7780, CurrentID=0, Status='live', Symbol='SLC30A2', Aliases='PP12488, TNZD, ZNT2, ZnT-2', description='solute carrier family 30 (zinc transporter), member 2', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26038022, end_position_on_the_genomic_accession=26046138, orientation='minus', exon_count=8, OMIM=609617, no_hdr0=''),
  7781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7781, CurrentID=0, Status='live', Symbol='SLC30A3', Aliases='ZNT3', description='solute carrier family 30 (zinc transporter), member 3', other_designations='solute carrier family 30 member 3|zinc transporter ZnT-3|znT-3', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27254572, end_position_on_the_genomic_accession=27275366, orientation='minus', exon_count=13, OMIM=602878, no_hdr0=''),
  7782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7782, CurrentID=0, Status='live', Symbol='SLC30A4', Aliases='ZNT4', description='solute carrier family 30 (zinc transporter), member 4', other_designations='solute carrier family 30 member 4|znT-4', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45482478, end_position_on_the_genomic_accession=45523669, orientation='minus', exon_count=9, OMIM=602095, no_hdr0=''),
  7783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7783, CurrentID=0, Status='live', Symbol='ZP2', Aliases='ZPA, Zp-2', description='zona pellucida glycoprotein 2 (sperm receptor)', other_designations='zona pellucida glycoprotein ZP2|zona pellucida protein A', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21197452, end_position_on_the_genomic_accession=21214510, orientation='minus', exon_count=20, OMIM=182888, no_hdr0=''),
  7784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7784, CurrentID=0, Status='live', Symbol='ZP3', Aliases='ZP3AB, ZPC, Zp-3, ZP3', description='zona pellucida glycoprotein 3 (sperm receptor)', other_designations='ZP3A/ZP3B|sperm receptor|zona pellucida glycoprotein 3A (sperm receptor)|zona pellucida glycoprotein 3B|zona pellucida glycoprotein ZP3|zona pellucida protein C', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76397524, end_position_on_the_genomic_accession=76442071, orientation='plus', exon_count=9, OMIM=182889, no_hdr0=''),
  7786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7786, CurrentID=0, Status='live', Symbol='MAP3K12', Aliases='DLK, MEKK12, MUK, ZPK, ZPKP1', description='mitogen-activated protein kinase kinase kinase 12', other_designations='MAPK-upstream kinase|dual leucine zipper bearing kinase|dual leucine zipper kinase DLK|leucine zipper protein kinase|mixed lineage kinase|protein kinase MUK', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53479669, end_position_on_the_genomic_accession=53500136, orientation='minus', exon_count=19, OMIM=600447, no_hdr0=''),
  23143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23143, CurrentID=0, Status='live', Symbol='LRCH1', Aliases='CHDC1, NP81', description='leucine-rich repeats and calponin homology (CH) domain containing 1', other_designations='neuronal protein 81', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46553138, end_position_on_the_genomic_accession=46753041, orientation='plus', exon_count=21, OMIM=610368, no_hdr0=''),
  7789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7789, CurrentID=0, Status='live', Symbol='ZXDA', Aliases='ZNF896', description='zinc finger, X-linked, duplicated A', other_designations='zinc finger protein 896', map_location='Xp11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57905430, end_position_on_the_genomic_accession=57910633, orientation='minus', exon_count=1, OMIM=300235, no_hdr0=''),
  7791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7791, CurrentID=0, Status='live', Symbol='ZYX', Aliases='ESP-2, HED-2', description='zyxin', other_designations='zyxin-2', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143378329, end_position_on_the_genomic_accession=143391111, orientation='plus', exon_count=10, OMIM=602002, no_hdr0=''),
  7792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7792, CurrentID=0, Status='live', Symbol='CCV', Aliases='CTRCT8', description='cataract, congenital, Volkmann type', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=115665, no_hdr0=''),
  401010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401010, CurrentID=0, Status='live', Symbol='LOC401010', Aliases='', description='NOC2-like nucleolar associated transcriptional repressor pseudogene', other_designations='nucleolar complex associated 2 homolog (S. cerevisiae) pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131442161, end_position_on_the_genomic_accession=131444894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7795, CurrentID=0, Status='live', Symbol='MEMO1', Aliases='', description='Methylation modifier for class I HLA', other_designations='', map_location='1p36.1-p35', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601201, no_hdr0=''),
  7798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7798, CurrentID=0, Status='live', Symbol='LUZP1', Aliases='LUZP', description='leucine zipper protein 1', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23084023, end_position_on_the_genomic_accession=23177808, orientation='minus', exon_count=6, OMIM=601422, no_hdr0=''),
  7799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7799, CurrentID=0, Status='live', Symbol='PRDM2', Aliases='HUMHOXY1, KMT8, MTB-ZF, RIZ, RIZ1, RIZ2', description='PR domain containing 2, with ZNF domain', other_designations='GATA-3 binding protein G3B|GATA-3-binding protein G3B|MTE-binding protein|PR domain-containing protein 2|lysine N-methyltransferase 8|retinoblastoma protein-binding zinc finger protein|retinoblastoma protein-interacting zinc finger protein|zinc finger protein RIZ|zinc-finger DNA-binding protein', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13700240, end_position_on_the_genomic_accession=13825079, orientation='plus', exon_count=15, OMIM=601196, no_hdr0=''),
  100271380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271380, CurrentID=0, Status='live', Symbol='RPS27P18', Aliases='RPS27_6_1109', description='ribosomal protein S27 pseudogene 18', other_designations='ribosomal protein S27-like pseudogene', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125473679, end_position_on_the_genomic_accession=125473935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7802, CurrentID=0, Status='live', Symbol='DNALI1', Aliases='P28, dJ423B22.5, hp28', description='dynein, axonemal, light intermediate chain 1', other_designations='dJ423B22.5 (axonemal dynein light chain (hp28))|dynein, axonemal, light intermediate polypeptide 1|inner dynein arm light chain, axonemal|inner dynein arm, homolog of clamydomonas', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37556919, end_position_on_the_genomic_accession=37566857, orientation='plus', exon_count=6, OMIM=602135, no_hdr0=''),
  7803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7803, CurrentID=0, Status='live', Symbol='PTP4A1', Aliases='HH72, PRL-1, PRL1, PTP(CAAX1), PTPCAAX1', description='protein tyrosine phosphatase type IVA, member 1', other_designations='protein tyrosine phosphatase type IVA protein 1|protein-tyrosine phosphatase 4a1|protein-tyrosine phosphatase of regenerating liver 1', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63572012, end_position_on_the_genomic_accession=63583588, orientation='plus', exon_count=7, OMIM=601585, no_hdr0=''),
  7804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7804, CurrentID=0, Status='live', Symbol='LRP8', Aliases='APOER2, HSZ75190, LRP-8, MCI1', description='LDL receptor related protein 8', other_designations='ApoE receptor 2|low density lipoprotein receptor-related protein 8, apolipoprotein e receptor', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53242364, end_position_on_the_genomic_accession=53328149, orientation='minus', exon_count=22, OMIM=602600, no_hdr0=''),
  7805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7805, CurrentID=0, Status='live', Symbol='LAPTM5', Aliases='CLAST6', description='lysosomal protein transmembrane 5', other_designations='CD40-ligand-activated specific transcripts|human retinoic acid-inducible E3 protein|lysosomal associated multispanning membrane protein 5|lysosomal multispanning membrane protein 5|lysosomal-associated multitransmembrane protein 5|retinoic acid-inducible E3 protein', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30732468, end_position_on_the_genomic_accession=30757836, orientation='minus', exon_count=8, OMIM=601476, no_hdr0=''),
  401022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401022, CurrentID=0, Status='live', Symbol='HAGLR', Aliases='HOXD-AS1, Mdgt', description='HOXD antisense growth-associated long non-coding RNA', other_designations='HOXD cluster antisense RNA 1 (non-protein coding)', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176173189, end_position_on_the_genomic_accession=176188958, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271381, CurrentID=0, Status='live', Symbol='RPS28P8', Aliases='RPS28_4_1309', description='ribosomal protein S28 pseudogene 8', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42003432, end_position_on_the_genomic_accession=42003636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401024, CurrentID=0, Status='live', Symbol='FSIP2', Aliases='', description='fibrous sheath interacting protein 2', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185738628, end_position_on_the_genomic_accession=185833290, orientation='plus', exon_count=23, OMIM=615796, no_hdr0=''),
  7809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7809, CurrentID=0, Status='live', Symbol='BSND', Aliases='BART, DFNB73', description='barttin CLCNK-type chloride channel accessory beta subunit', other_designations='Bartter syndrome, infantile, with sensorineural deafness (Barttin)|deafness, autosomal recessive 73', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54998944, end_position_on_the_genomic_accession=55008792, orientation='plus', exon_count=5, OMIM=606412, no_hdr0=''),
  138882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138882, CurrentID=0, Status='live', Symbol='OR1N2', Aliases='OR9-23', description='olfactory receptor family 1 subfamily N member 2', other_designations='olfactory receptor OR9-23|olfactory receptor, family 1, subfamily N, member 2', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122553170, end_position_on_the_genomic_accession=122554162, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  138883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138883, CurrentID=0, Status='live', Symbol='OR1N1', Aliases='OR1-26, OR1N3, OR9-22', description='olfactory receptor family 1 subfamily N member 1', other_designations='olfactory receptor 1-26|olfactory receptor 1N3|olfactory receptor OR9-22|olfactory receptor, family 1, subfamily N, member 1|olfactory receptor, family 1, subfamily N, member 3', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122526358, end_position_on_the_genomic_accession=122527293, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7812, CurrentID=0, Status='live', Symbol='CSDE1', Aliases='D1S155E, UNR', description='cold shock domain containing E1, RNA binding', other_designations='N-ras upstream gene protein|NRAS-related|cold shock domain containing E1, RNA-binding|upstream of NRAS', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114716913, end_position_on_the_genomic_accession=114758050, orientation='minus', exon_count=21, OMIM=191510, no_hdr0=''),
  7813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7813, CurrentID=0, Status='live', Symbol='EVI5', Aliases='EVI-5, NB4S', description='ecotropic viral integration site 5', other_designations='dJ846D11.1 (ecotropic viral integration site 5)|neuroblastoma stage 4S gene protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92508696, end_position_on_the_genomic_accession=92792404, orientation='minus', exon_count=28, OMIM=602942, no_hdr0=''),
  7818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7818, CurrentID=0, Status='live', Symbol='DAP3', Aliases='DAP-3, MRP-S29, MRPS29, bMRP-10', description='death associated protein 3', other_designations='S29mt|death-associated protein 3|ionizing radiation resistance conferring protein|mitochondrial 28S ribosomal protein S29', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155689089, end_position_on_the_genomic_accession=155739010, orientation='plus', exon_count=14, OMIM=602074, no_hdr0=''),
  100271383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271383, CurrentID=0, Status='live', Symbol='RPS29P19', Aliases='RPS29_4_1178', description='ribosomal protein S29 pseudogene 19', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113751116, end_position_on_the_genomic_accession=113751286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401036, CurrentID=0, Status='live', Symbol='ASB18', Aliases='ASB-18', description='ankyrin repeat and SOCS box containing 18', other_designations='ankyrin repeat and SOCS box-containing 18', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236194872, end_position_on_the_genomic_accession=236264345, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  401037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401037, CurrentID=0, Status='live', Symbol='LOC401037', Aliases='', description='uncharacterized LOC401037', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401039, CurrentID=0, Status='live', Symbol='FLJ43879', Aliases='', description='FLJ43879 protein', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238919302, end_position_on_the_genomic_accession=238926269, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  7824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7824, CurrentID=0, Status='live', Symbol='D1S111', Aliases='D1S61, MS336', description='Minisatellite 33.6', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=157560, no_hdr0=''),
  7827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7827, CurrentID=0, Status='live', Symbol='NPHS2', Aliases='PDCN, SRN1', description='nephrosis 2, idiopathic, steroid-resistant (podocin)', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179550539, end_position_on_the_genomic_accession=179575976, orientation='minus', exon_count=9, OMIM=604766, no_hdr0=''),
  401044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401044, CurrentID=0, Status='live', Symbol='FLJ40712', Aliases='', description='FLJ40712 protein', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=242577387, end_position_on_the_genomic_accession=242577946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  401045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401045, CurrentID=0, Status='live', Symbol='FLJ41327', Aliases='', description='FLJ41327 protein', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=242619101, end_position_on_the_genomic_accession=242619989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7830, CurrentID=0, Status='live', Symbol='PHA2A', Aliases='PHA2', description='Pseudohypoaldosteronism type II (gene A)', other_designations='', map_location='1q31-q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145260, no_hdr0=''),
  100271385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271385, CurrentID=0, Status='live', Symbol='RPS4XP12', Aliases='RPS4P12, RPS4X_4_1183', description='ribosomal protein S4X pseudogene 12', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121250264, end_position_on_the_genomic_accession=121250973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7832, CurrentID=0, Status='live', Symbol='BTG2', Aliases='PC3, TIS21', description='BTG family member 2', other_designations='B-cell translocation gene 2|BTG family, member 2|NGF-inducible anti-proliferative protein PC3|nerve growth factor-inducible anti-proliferative|pheochromacytoma cell-3', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203305536, end_position_on_the_genomic_accession=203309602, orientation='plus', exon_count=2, OMIM=601597, no_hdr0=''),
  7834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7834, CurrentID=0, Status='live', Symbol='PCAP', Aliases='HPC2, HPC8', description='predisposing for prostate cancer', other_designations='', map_location='1q42.2-q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602759, no_hdr0=''),
  401052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401052, CurrentID=0, Status='live', Symbol='LOC401052', Aliases='', description='uncharacterized LOC401052', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10006418, end_position_on_the_genomic_accession=10011095, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  7837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7837, CurrentID=0, Status='live', Symbol='PXDN', Aliases='COPOA, D2S448, D2S448E, MG50, PRG2, PXN, VPO', description='peroxidasin', other_designations='melanoma-associated antigen MG50|p53-responsive gene 2 protein|vascular peroxidase 1', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1631887, end_position_on_the_genomic_accession=1744852, orientation='minus', exon_count=26, OMIM=605158, no_hdr0=''),
  7839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7839, CurrentID=0, Status='live', Symbol='LEPQTL1', Aliases='LSL', description='Leptin, serum levels of', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601694, no_hdr0=''),
  7840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7840, CurrentID=0, Status='live', Symbol='ALMS1', Aliases='ALSS', description='ALMS1, centrosome and basal body associated protein', other_designations='Alstrom syndrome 1', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73385758, end_position_on_the_genomic_accession=73609919, orientation='plus', exon_count=23, OMIM=606844, no_hdr0=''),
  7841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7841, CurrentID=0, Status='live', Symbol='MOGS', Aliases='CDG2B, CWH41, DER7, GCS1', description='mannosyl-oligosaccharide glucosidase', other_designations='glucosidase I|processing A-glucosidase I', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74461057, end_position_on_the_genomic_accession=74465410, orientation='minus', exon_count=5, OMIM=601336, no_hdr0=''),
  7842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7842, CurrentID=0, Status='live', Symbol='IGKJ', Aliases='IGKJ@', description='immunoglobulin kappa joining cluster', other_designations='immunoglobulin kappa joining group|immunoglobulin kappa joining region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146970, no_hdr0=''),
  100271387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271387, CurrentID=0, Status='live', Symbol='RPL12P43', Aliases='RPL12_24_1805', description='ribosomal protein L12 pseudogene 43', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118146063, end_position_on_the_genomic_accession=118146561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7844, CurrentID=0, Status='live', Symbol='RNF103', Aliases='HKF-1, KF-1, KF1, ZFP-103, ZFP103', description='ring finger protein 103', other_designations='zinc finger protein 103 homolog|zinc finger protein expressed in cerebellum', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86603393, end_position_on_the_genomic_accession=86623877, orientation='minus', exon_count=5, OMIM=602507, no_hdr0=''),
  7845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7845, CurrentID=0, Status='live', Symbol='PAFAH1B1P1', Aliases='LIS2, PAFAH1P1', description='platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 pseudogene 1', other_designations='Lissencephaly gene 2|platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 (45kDa) pseudogene|platelet-activating factor acetylhydrolase, isoform Ib, pseudogene 1|platelet-activating factor acetylhydrolase, isoform Ib, subunit 1 pseudogene 1', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87565673, end_position_on_the_genomic_accession=87567250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7846, CurrentID=0, Status='live', Symbol='TUBA1A', Aliases='B-ALPHA-1, LIS3, TUBA3', description='tubulin alpha 1a', other_designations='hum-a-tub1|hum-a-tub2|tubulin B-alpha-1|tubulin alpha-3 chain|tubulin, alpha 1a|tubulin, alpha, brain-specific', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49184795, end_position_on_the_genomic_accession=49189324, orientation='minus', exon_count=4, OMIM=602529, no_hdr0=''),
  7849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7849, CurrentID=0, Status='live', Symbol='PAX8', Aliases='', description='paired box 8', other_designations='paired domain gene 8', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113215997, end_position_on_the_genomic_accession=113278950, orientation='minus', exon_count=12, OMIM=167415, no_hdr0=''),
  7850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7850, CurrentID=0, Status='live', Symbol='IL1R2', Aliases='CD121b, CDw121b, IL-1R-2, IL-1RT-2, IL-1RT2c, IL1RB, IL1R2', description='interleukin 1 receptor, type II', other_designations='CD121 antigen-like family member B|IL-1 type II receptor|IL-1R-beta|antigen CDw121b|interleukin 1 receptor type II variant 3|interleukin-1 receptor beta|interleukin-1 receptor type II|type II interleukin-1 receptor, beta', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101991816, end_position_on_the_genomic_accession=102028544, orientation='plus', exon_count=15, OMIM=147811, no_hdr0=''),
  7851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7851, CurrentID=0, Status='live', Symbol='MALL', Aliases='BENE', description='mal, T-cell differentiation protein-like', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110083870, end_position_on_the_genomic_accession=110118139, orientation='minus', exon_count=5, OMIM=602022, no_hdr0=''),
  7852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7852, CurrentID=0, Status='live', Symbol='CXCR4', Aliases='CD184, D2S201E, FB22, HM89, HSY3RR, LAP-3, LAP3, LCR1, LESTR, NPY3R, NPYR, NPYRL, NPYY3R, WHIM, WHIMS', description='chemokine (C-X-C motif) receptor 4', other_designations='CD184 antigen|LPS-associated protein 3|SDF-1 receptor|chemokine receptor|fusin|leukocyte-derived seven transmembrane domain receptor|lipopolysaccharide-associated protein 3|neuropeptide Y receptor Y3|neuropeptide Y3 receptor|seven transmembrane helix receptor|seven-transmembrane-segment receptor, spleen|stromal cell-derived factor 1 receptor', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136114349, end_position_on_the_genomic_accession=136118155, orientation='minus', exon_count=2, OMIM=162643, no_hdr0=''),
  101929269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929269, CurrentID=0, Status='live', Symbol='LOC101929269', Aliases='', description='uncharacterized LOC101929269', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11062652, end_position_on_the_genomic_accession=11067119, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7855, CurrentID=0, Status='live', Symbol='FZD5', Aliases='C2orf31, HFZ5', description='frizzled class receptor 5', other_designations='Wnt receptor|frizzled 5, seven transmembrane spanning receptor|frizzled family receptor 5|fz-5|fzE5|seven-transmembrane receptor frizzled-5', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207762586, end_position_on_the_genomic_accession=207769419, orientation='minus', exon_count=2, OMIM=601723, no_hdr0=''),
  7857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7857, CurrentID=0, Status='live', Symbol='SCG2', Aliases='CHGC, EM66, SN, SgII', description='secretogranin II', other_designations='chromogranin-C|secretoneurin', map_location='2q35-q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223596940, end_position_on_the_genomic_accession=223602499, orientation='minus', exon_count=2, OMIM=118930, no_hdr0=''),
  401074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401074, CurrentID=0, Status='live', Symbol='LINC00960', Aliases='', description='long intergenic non-protein coding RNA 960', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75672281, end_position_on_the_genomic_accession=75679303, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  285299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285299, CurrentID=0, Status='live', Symbol='LOC285299', Aliases='', description='FSHD region gene 2 family, member C-like', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NT_167216.1', start_position_on_the_genomic_accession=86423, end_position_on_the_genomic_accession=106274, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  401076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401076, CurrentID=0, Status='live', Symbol='CAP1P1', Aliases='', description='CAP1 pseudogene 1', other_designations='CAP, adenylate cyclase-associated protein 1 pseudogene', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=76434018, end_position_on_the_genomic_accession=76435425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271390, CurrentID=0, Status='live', Symbol='RPL21P132', Aliases='RPL21_62_1798', description='ribosomal protein L21 pseudogene 132', other_designations='', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101340589, end_position_on_the_genomic_accession=101340975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7862, CurrentID=0, Status='live', Symbol='BRPF1', Aliases='BR140', description='bromodomain and PHD finger containing 1', other_designations='bromodomain and PHD finger containing, 1|bromodomain and PHD finger-containing protein 1|bromodomain-containing protein, 140kD', map_location='3p26-p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9731729, end_position_on_the_genomic_accession=9748015, orientation='plus', exon_count=16, OMIM=602410, no_hdr0=''),
  7864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7864, CurrentID=0, Status='live', Symbol='SCLC1', Aliases='SCCL, SCLC', description='small cell cancer of the lung', other_designations='', map_location='3p23-p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=182280, no_hdr0=''),
  401081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401081, CurrentID=0, Status='live', Symbol='FLJ22763', Aliases='', description='uncharacterized LOC401081', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109136714, end_position_on_the_genomic_accession=109150109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7866, CurrentID=0, Status='live', Symbol='IFRD2', Aliases='IFNRP, SKMc15, SM15', description='interferon-related developmental regulator 2', other_designations='interferon-related protein', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50287732, end_position_on_the_genomic_accession=50292595, orientation='minus', exon_count=12, OMIM=602725, no_hdr0=''),
  7867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7867, CurrentID=0, Status='live', Symbol='MAPKAPK3', Aliases='3PK, MAPKAP-K3, MAPKAP3, MAPKAPK-3, MK-3', description='mitogen-activated protein kinase-activated protein kinase 3', other_designations='MAPK-activated protein kinase 3|MAPKAP kinase 3|chromosome 3p kinase', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50611862, end_position_on_the_genomic_accession=50649297, orientation='plus', exon_count=14, OMIM=602130, no_hdr0=''),
  7869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7869, CurrentID=0, Status='live', Symbol='SEMA3B', Aliases='LUCA-1, SEMA5, SEMAA, SemA, semaV', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3B', other_designations='sema A(V)|semaphorin A|semaphorin-V', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50267558, end_position_on_the_genomic_accession=50277171, orientation='plus', exon_count=18, OMIM=601281, no_hdr0=''),
  7870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7870, CurrentID=0, Status='live', Symbol='TRR', Aliases='', description='transfer RNA arginine', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601994, no_hdr0=''),
  7871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7871, CurrentID=0, Status='live', Symbol='SLMAP', Aliases='SLAP', description='sarcolemma associated protein', other_designations='', map_location='3p21.2-p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57756221, end_position_on_the_genomic_accession=57930003, orientation='plus', exon_count=26, OMIM=602701, no_hdr0=''),
  7873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7873, CurrentID=0, Status='live', Symbol='MANF', Aliases='ARMET, ARP', description='mesencephalic astrocyte-derived neurotrophic factor', other_designations='arginine-rich, mutated in early stage tumors', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51385237, end_position_on_the_genomic_accession=51389397, orientation='plus', exon_count=4, OMIM=601916, no_hdr0=''),
  7874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7874, CurrentID=0, Status='live', Symbol='USP7', Aliases='HAUSP, TEF1', description='ubiquitin specific peptidase 7 (herpes virus-associated)', other_designations='Herpes virus-associated ubiquitin-specific protease|deubiquitinating enzyme 7|herpesvirus-associated ubiquitin-specific protease|ubiquitin specific protease 7 (herpes virus-associated)|ubiquitin thioesterase 7|ubiquitin-specific-processing protease 7', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8892094, end_position_on_the_genomic_accession=8963484, orientation='minus', exon_count=33, OMIM=602519, no_hdr0=''),
  138948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138948, CurrentID=0, Status='live', Symbol='DBH-AS1', Aliases='NCRNA00118', description='DBH antisense RNA 1', other_designations='BPR|DBH antisense RNA 1 (non-protein coding)', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133654587, end_position_on_the_genomic_accession=133657408, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  401093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401093, CurrentID=0, Status='live', Symbol='MBNL1-AS1', Aliases='', description='MBNL1 antisense RNA 1', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152262616, end_position_on_the_genomic_accession=152269626, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7879, CurrentID=0, Status='live', Symbol='RAB7A', Aliases='PRO2706, RAB7', description='RAB7A, member RAS oncogene family', other_designations='RAB7, member RAS oncogene family|Ras-associated protein RAB7', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128726136, end_position_on_the_genomic_accession=128814798, orientation='plus', exon_count=6, OMIM=602298, no_hdr0=''),
  7880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7880, CurrentID=0, Status='live', Symbol='MBS2', Aliases='HCFP1', description='Moebius syndrome 2', other_designations='hereditary congenital facial paresis 1', map_location='3q21-q22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601471, no_hdr0=''),
  7881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7881, CurrentID=0, Status='live', Symbol='KCNAB1', Aliases='AKR6A3, KCNA1B, KV-BETA-1, Kvb1.3, hKvBeta3, hKvb3', description='potassium channel, voltage gated subfamily A regulatory beta subunit 1', other_designations='K(+) channel subunit beta-1|K+ channel Beta1a chain|potassium channel beta 3 chain|potassium channel beta3 subunit|potassium channel shaker chain beta 1a|potassium voltage-gated channel, shaker-related subfamily, beta member 1', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156120548, end_position_on_the_genomic_accession=156539138, orientation='plus', exon_count=18, OMIM=601141, no_hdr0=''),
  401098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401098, CurrentID=0, Status='live', Symbol='LOC401098', Aliases='', description='uncharacterized LOC401098', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160449782, end_position_on_the_genomic_accession=160452169, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  285303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285303, CurrentID=0, Status='live', Symbol='LOC285303', Aliases='', description='family with sequence similarity 151, member A pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194755474, end_position_on_the_genomic_accession=194757516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7884, CurrentID=0, Status='live', Symbol='SLBP', Aliases='HBP', description='stem-loop binding protein', other_designations='hairpin binding protein, histone|histone binding protein|histone stem-loop binding protein|histone stem-loop-binding protein|stem-loop (histone) binding protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1692800, end_position_on_the_genomic_accession=1712694, orientation='minus', exon_count=8, OMIM=602422, no_hdr0=''),
  7885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7885, CurrentID=0, Status='live', Symbol='CRSA', Aliases='CRS3', description='Craniosynostosis, Adelaide type', other_designations='', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600593, no_hdr0=''),
  106481464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481464, CurrentID=0, Status='live', Symbol='RNU6-903P', Aliases='', description='RNA, U6 small nuclear 903, pseudogene', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7889, CurrentID=0, Status='live', Symbol='PSORS3', Aliases='', description='psoriasis susceptibility 3', other_designations='', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601454, no_hdr0=''),
  401106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401106, CurrentID=0, Status='live', Symbol='LINC00884', Aliases='', description='long intergenic non-protein coding RNA 884', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194487140, end_position_on_the_genomic_accession=194488545, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271395, CurrentID=0, Status='live', Symbol='RPL37P24', Aliases='RPL37_11_1779', description='ribosomal protein L37 pseudogene 24', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54147366, end_position_on_the_genomic_accession=54147659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7892, CurrentID=0, Status='live', Symbol='FECB', Aliases='', description='Fecundity gene, Boorla, of sheep, homolog of', other_designations='', map_location='4q21-q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=134720, no_hdr0=''),
  401109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401109, CurrentID=0, Status='live', Symbol='LINC00885', Aliases='', description='long intergenic non-protein coding RNA 885', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196142636, end_position_on_the_genomic_accession=196160890, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7895, CurrentID=0, Status='live', Symbol='LAG5', Aliases='', description='Leukocyte antigen group 5', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151450, no_hdr0=''),
  7896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7896, CurrentID=0, Status='live', Symbol='TS13', Aliases='', description='Temperature sensitivity complementation, ts13', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=187320, no_hdr0=''),
  100271396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271396, CurrentID=0, Status='live', Symbol='RPL39P24', Aliases='RPL39_11_1016', description='ribosomal protein L39 pseudogene 24', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132962453, end_position_on_the_genomic_accession=132962771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401114, CurrentID=0, Status='live', Symbol='FLJ35816', Aliases='', description='FLJ35816 protein', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=1145160, end_position_on_the_genomic_accession=1147489, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  401115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401115, CurrentID=0, Status='live', Symbol='C4orf48', Aliases='CHR4_55', description='chromosome 4 open reading frame 48', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2035608, end_position_on_the_genomic_accession=2043970, orientation='plus', exon_count=9, OMIM=614690, no_hdr0=''),
  7900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7900, CurrentID=0, Status='live', Symbol='ZNF5', Aliases='', description='zinc finger protein-5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7903, CurrentID=0, Status='live', Symbol='ST8SIA4', Aliases='PST, PST1, SIAT8D, ST8SIA-IV', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 4', other_designations='CMP-N-acetylneuraminate-poly-alpha-2,8-sialyl transferase|SIAT8-D|ST8SiaIV|alpha-2,8-sialyltransferase 8D|polysialyltransferase-1|sialyltransferase 8 (alpha-2, 8-polysialytransferase) D|sialyltransferase 8D|sialyltransferase St8Sia IV|sialytransferase St8Sia IV', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100806935, end_position_on_the_genomic_accession=100903285, orientation='minus', exon_count=6, OMIM=602547, no_hdr0=''),
  7905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7905, CurrentID=0, Status='live', Symbol='REEP5', Aliases='C5orf18, D5S346, DP1, TB2, YOP1', description='receptor accessory protein 5', other_designations='deleted in polyposis 1|polyposis coli region hypothetical protein DP1|polyposis locus protein 1|receptor expression enhancing protein 5', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112876384, end_position_on_the_genomic_accession=112922334, orientation='minus', exon_count=5, OMIM=125265, no_hdr0=''),
  401123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401123, CurrentID=0, Status='live', Symbol='LOC401123', Aliases='', description='uncharacterized LOC401123', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=26818225, end_position_on_the_genomic_accession=26818804, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7908, CurrentID=0, Status='live', Symbol='EOS', Aliases='', description='eosinophilia, familial', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=131400, no_hdr0=''),
  7909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7909, CurrentID=0, Status='live', Symbol='HEMC', Aliases='HCI', description='hemangioma, capillary, hereditary', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602089, no_hdr0=''),
  7910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7910, CurrentID=0, Status='live', Symbol='PFBI', Aliases='', description='Plasmodium falciparum blood infection levels', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=248310, no_hdr0=''),
  7911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7911, CurrentID=0, Status='live', Symbol='SM1', Aliases='', description='Schistosoma mansoni, susceptibility/resistance to', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=181460, no_hdr0=''),
  7913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7913, CurrentID=0, Status='live', Symbol='DEK', Aliases='D6S231E', description='DEK proto-oncogene', other_designations='DEK oncogene (DNA binding)', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18224169, end_position_on_the_genomic_accession=18264568, orientation='minus', exon_count=12, OMIM=125264, no_hdr0=''),
  7915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7915, CurrentID=0, Status='live', Symbol='ALDH5A1', Aliases='SSADH, SSDH', description='aldehyde dehydrogenase 5 family member A1', other_designations='NAD(+)-dependent succinic semialdehyde dehydrogenase|mitochondrial succinate semialdehyde dehydrogenase', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24494969, end_position_on_the_genomic_accession=24537207, orientation='plus', exon_count=11, OMIM=610045, no_hdr0=''),
  7916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7916, CurrentID=0, Status='live', Symbol='PRRC2A', Aliases='BAT2, D6S51, D6S51E, G2', description='proline-rich coiled-coil 2A', other_designations='HLA-B associated transcript 2|HLA-B-associated transcript 2|large proline-rich protein BAT2|proline-rich and coiled-coil-containing protein 2A|protein G2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31620673, end_position_on_the_genomic_accession=31637777, orientation='plus', exon_count=31, OMIM=142580, no_hdr0=''),
  7917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7917, CurrentID=0, Status='live', Symbol='BAG6', Aliases='BAG-6, BAT3, D6S52E, G3', description='BCL2 associated athanogene 6', other_designations='BAG family molecular chaperone regulator 6|BCL2-associated athanogene 6|HLA-B associated transcript 3|HLA-B associated transcript-3|HLA-B-associated transcript 3|large proline-rich protein BAT3|protein G3|protein Scythe|scythe', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31639028, end_position_on_the_genomic_accession=31660767, orientation='minus', exon_count=29, OMIM=142590, no_hdr0=''),
  7918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7918, CurrentID=0, Status='live', Symbol='GPANK1', Aliases='ANKRD59, BAT4, D6S54E, G5, GPATCH10', description='G-patch domain and ankyrin repeats 1', other_designations='G patch domain and ankyrin repeats 1|G patch domain and ankyrin repeats-containing protein 1|G patch domain containing 10|HLA-B-associated transcript 4|ankyrin repeat domain-containing protein 59|g patch domain-containing protein 10|protein BAT4|protein G5', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31661229, end_position_on_the_genomic_accession=31666283, orientation='minus', exon_count=5, OMIM=142610, no_hdr0=''),
  7919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7919, CurrentID=0, Status='live', Symbol='DDX39B', Aliases='BAT1, D6S81E, UAP56', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39B', other_designations='56 kDa U2AF65-associated protein|ATP-dependent RNA helicase p47|HLA-B-associated transcript 1 protein|nuclear RNA helicase (DEAD family)|spliceosome RNA helicase BAT1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31530219, end_position_on_the_genomic_accession=31542475, orientation='minus', exon_count=11, OMIM=142560, no_hdr0=''),
  7920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7920, CurrentID=0, Status='live', Symbol='ABHD16A', Aliases='BAT5, D6S82E, NG26, PP199', description='abhydrolase domain containing 16A', other_designations='HLA-B associated transcript 5|protein G5', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31686949, end_position_on_the_genomic_accession=31703360, orientation='minus', exon_count=21, OMIM=142620, no_hdr0=''),
  401137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401137, CurrentID=0, Status='live', Symbol='PRR27', Aliases='C4orf40', description='proline rich 27', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70154187, end_position_on_the_genomic_accession=70166609, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7922, CurrentID=0, Status='live', Symbol='SLC39A7', Aliases='D6S115E, D6S2244E, H2-KE4, HKE4, KE4, RING5, ZIP7', description='solute carrier family 39 (zinc transporter), member 7', other_designations='HLA class II region expressed gene KE4|Ke4 gene, mouse, human homolog of|histidine-rich membrane protein Ke4|really interesting new gene 5 protein|solute carrier family 39 member 7|zrt-, Irt-like protein 7', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33200826, end_position_on_the_genomic_accession=33204437, orientation='plus', exon_count=8, OMIM=601416, no_hdr0=''),
  7923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7923, CurrentID=0, Status='live', Symbol='HSD17B8', Aliases='D6S2245E, FABG, FABGL, H2-KE6, HKE6, KE6, RING2, SDR30C1, dJ1033B10.9', description='hydroxysteroid (17-beta) dehydrogenase 8', other_designations='17-beta-HSD 8|17-beta-hydroxysteroid dehydrogenase 8|3-oxoacyl-[acyl-carrier-protein] reductase|estrogen 17-oxidoreductase|ke-6|protein Ke6|really interesting new gene 2 protein|short chain dehydrogenase/reductase family 30C member 1|short chain dehydrogenase/reductase family 30C, member 1|testosterone 17-beta-dehydrogenase 8', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33204637, end_position_on_the_genomic_accession=33206831, orientation='plus', exon_count=9, OMIM=601417, no_hdr0=''),
  7925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7925, CurrentID=0, Status='live', Symbol='IGLP1', Aliases='', description='immune response to synthetic polypeptides 1', other_designations='Immune response to synthetic polypeptides-1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147080, no_hdr0=''),
  7926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7926, CurrentID=0, Status='live', Symbol='IGLP2', Aliases='', description='immune response to synthetic polypeptides 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147090, no_hdr0=''),
  7927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7927, CurrentID=0, Status='live', Symbol='IGAT', Aliases='', description='Immune response to synthetic polypeptide--IRGAT', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146820, no_hdr0=''),
  7928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7928, CurrentID=0, Status='live', Symbol='ISCW', Aliases='ISSCW', description='Immune suppression to streptococcal antigen', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146850, no_hdr0=''),
  401145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401145, CurrentID=0, Status='live', Symbol='CCSER1', Aliases='FAM190A', description='coiled-coil serine rich protein 1', other_designations='coiled-coil serine-rich protein 1|family with sequence similarity 190, member A|protein FAM190A', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=90127394, end_position_on_the_genomic_accession=91602219, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  23167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23167, CurrentID=0, Status='live', Symbol='EFR3A', Aliases='', description='EFR3 homolog A', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=131904109, end_position_on_the_genomic_accession=132013642, orientation='plus', exon_count=25, OMIM=611798, no_hdr0=''),
  7932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7932, CurrentID=0, Status='live', Symbol='OR2H2', Aliases='FAT11, OLFR2, OLFR42B, OR2H3, dJ271M21.2, hs6M1-12', description='olfactory receptor family 2 subfamily H member 2', other_designations='Olfactory receptor 2|olfactory receptor 2H3|olfactory receptor OR6-36|olfactory receptor, family 2, subfamily H, member 2|olfactory receptor, family 2, subfamily H, member 3|olfactory receptor-like protein FAT11', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29587378, end_position_on_the_genomic_accession=29591012, orientation='plus', exon_count=2, OMIM=600578, no_hdr0=''),
  7933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7933, CurrentID=0, Status='live', Symbol='OPLL', Aliases='', description='ossification of posterior longitudinal ligament of spine', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602475, no_hdr0=''),
  7936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7936, CurrentID=0, Status='live', Symbol='NELFE', Aliases='D6S45, NELF-E, RD, RDBP, RDP', description='negative elongation factor complex member E', other_designations='RD RNA binding protein|RD RNA-binding protein|RNA-binding protein RD|major histocompatibility complex gene RD|negative elongation factor polypeptide E|nuclear protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31952087, end_position_on_the_genomic_accession=31959087, orientation='minus', exon_count=11, OMIM=154040, no_hdr0=''),
  7937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7937, CurrentID=0, Status='live', Symbol='RWS', Aliases='', description='Ragweed sensitivity', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=179450, no_hdr0=''),
  7938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7938, CurrentID=0, Status='live', Symbol='CP20', Aliases='LCP20', description='Lymphocyte cytosolic protein, molecular weight 20kD', other_designations='', map_location='6p21.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=153380, no_hdr0=''),
  7939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7939, CurrentID=0, Status='live', Symbol='LAP', Aliases='', description='Laryngeal adductor paralysis', other_designations='', map_location='6p21.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=150270, no_hdr0=''),
  7940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7940, CurrentID=0, Status='live', Symbol='LST1', Aliases='B144, D6S49E, LST-1', description='leukocyte specific transcript 1', other_designations='lymphocyte antigen 117', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31586179, end_position_on_the_genomic_accession=31588909, orientation='plus', exon_count=6, OMIM=109170, no_hdr0=''),
  7941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7941, CurrentID=0, Status='live', Symbol='PLA2G7', Aliases='LDL-PLA2, LP-PLA2, PAFAD, PAFAH', description='phospholipase A2 group VII', other_designations='1-alkyl-2-acetylglycerophosphocholine esterase|2-acetyl-1-alkylglycerophosphocholine esterase|LDL-PLA(2)|LDL-associated phospholipase A2|PAF 2-acylhydrolase|PAF acetylhydrolase|gVIIA-PLA2|group-VIIA phospholipase A2|lipoprotein-associated phospholipase A2|phospholipase A2, group VII (platelet-activating factor acetylhydrolase, plasma)', map_location='6p21.2-p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46704316, end_position_on_the_genomic_accession=46735701, orientation='minus', exon_count=13, OMIM=601690, no_hdr0=''),
  7942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7942, CurrentID=0, Status='live', Symbol='TFEB', Aliases='ALPHATFEB, BHLHE35, TCFEB', description='transcription factor EB', other_designations='T-cell transcription factor EB|class E basic helix-loop-helix protein 35', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41683978, end_position_on_the_genomic_accession=41736259, orientation='minus', exon_count=17, OMIM=600744, no_hdr0=''),
  285313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285313, CurrentID=0, Status='live', Symbol='IGSF10', Aliases='CMF608', description='immunoglobulin superfamily member 10', other_designations='calvaria mechanical force protein 608|immunoglobulin superfamily, member 10', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151435990, end_position_on_the_genomic_accession=151619926, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  7944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7944, CurrentID=0, Status='live', Symbol='CSCI', Aliases='', description='Corticosterone side-chain isomerase', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=122550, no_hdr0=''),
  7945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7945, CurrentID=0, Status='live', Symbol='PUJO', Aliases='', description='pelviureteric junction obstruction', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=143400, no_hdr0=''),
  7948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7948, CurrentID=0, Status='live', Symbol='PBCRA1', Aliases='CRAPB, PBCRA', description='progressive bifocal chorioretinal atrophy 1', other_designations='', map_location='6q14-q16.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600790, no_hdr0=''),
  23170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23170, CurrentID=0, Status='live', Symbol='TTLL12', Aliases='dJ526I14.2', description='tubulin tyrosine ligase like 12', other_designations='tubulin tyrosine ligase-like family member 12|tubulin tyrosine ligase-like family, member 12', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43166622, end_position_on_the_genomic_accession=43187131, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  401166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401166, CurrentID=0, Status='live', Symbol='FLJ44896', Aliases='', description='FLJ44896 protein', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=252099, end_position_on_the_genomic_accession=252575, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7951, CurrentID=0, Status='live', Symbol='BKMA1', Aliases='', description='Banded krait minor satellite DNA-1', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109780, no_hdr0=''),
  7952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7952, CurrentID=0, Status='live', Symbol='TNDM', Aliases='DMTN, TNDM1', description='diabetes mellitus, transient neonatal', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131099, CurrentID=0, Status='live', Symbol='LOC100131099', Aliases='', description='uncharacterized LOC100131099', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=60411356, end_position_on_the_genomic_accession=60411787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7954, CurrentID=0, Status='live', Symbol='HBFQTL2', Aliases='FCP, HPFH', description='hereditary persistence of fetal hemoglobin, heterocellular', other_designations='', map_location='6q22.3-q23.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=142470, no_hdr0=''),
  7955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7955, CurrentID=0, Status='live', Symbol='RNF217-AS1', Aliases='STL', description='RNF217 antisense RNA 1 (head to head)', other_designations='six-twelve leukemia', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=124908243, end_position_on_the_genomic_accession=124963027, orientation='minus', exon_count=7, OMIM=602532, no_hdr0=''),
  7956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7956, CurrentID=0, Status='live', Symbol='NTT', Aliases='', description='Noncoding transcript in T cells', other_designations='', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602154, no_hdr0=''),
  7957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7957, CurrentID=0, Status='live', Symbol='EPM2A', Aliases='EPM2, MELF', description='epilepsy, progressive myoclonus type 2A, Lafora disease (laforin)', other_designations='LAFPTPase|epilepsy, progressive myoclonus type 2, Lafora disease (laforin)|glucan phosphatase|lafora PTPase', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145500777, end_position_on_the_genomic_accession=145736303, orientation='minus', exon_count=10, OMIM=607566, no_hdr0=''),
  7959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7959, CurrentID=0, Status='live', Symbol='FEA', Aliases='', description='F9 embryonic antigen', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=137010, no_hdr0=''),
  7960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7960, CurrentID=0, Status='live', Symbol='MRBC', Aliases='', description='Monkey RBC receptor', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=158050, no_hdr0=''),
  401177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401177, CurrentID=0, Status='live', Symbol='LOC401177', Aliases='', description='uncharacterized LOC401177', other_designations='CTD-2139B15.4', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17378906, end_position_on_the_genomic_accession=17387310, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7962, CurrentID=0, Status='live', Symbol='PBCA', Aliases='', description='Pancreatic beta cell, agenesis of', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600089, no_hdr0=''),
  7963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7963, CurrentID=0, Status='live', Symbol='TS546', Aliases='', description='Temperature sensitivity complementation, cell cycle specific', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=187330, no_hdr0=''),
  646214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646214, CurrentID=0, Status='live', Symbol='LOC646214', Aliases='', description='p21 protein (Cdc42/Rac)-activated kinase 2 pseudogene', other_designations='p21-activated kinase 2 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21289604, end_position_on_the_genomic_accession=21297829, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7965, CurrentID=0, Status='live', Symbol='AIMP2', Aliases='JTV-1, JTV1, P38, PRO0992', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 2', other_designations='ARS-interacting multi-functional protein 2|multisynthase complex auxiliary component p38|multisynthetase complex auxiliary component p38|protein JTV-1', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6009251, end_position_on_the_genomic_accession=6023834, orientation='plus', exon_count=8, OMIM=600859, no_hdr0=''),
  139041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139041, CurrentID=0, Status='live', Symbol='PSMAGEA', Aliases='', description='MAGE family member A4 pseudogene', other_designations='melanoma antigen family A, 4 pseudogene|melanoma antigen family A4 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152784304, end_position_on_the_genomic_accession=152789417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7971, CurrentID=0, Status='live', Symbol='GHS', Aliases='', description='Goldenhar syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=141400, no_hdr0=''),
  401188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401188, CurrentID=0, Status='live', Symbol='LOC401188', Aliases='', description='uncharacterized LOC401188', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401190, CurrentID=0, Status='live', Symbol='RGS7BP', Aliases='R7BP', description='regulator of G-protein signaling 7 binding protein', other_designations='R7 family-binding protein', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64505539, end_position_on_the_genomic_accession=64614657, orientation='plus', exon_count=7, OMIM=610890, no_hdr0=''),
  7975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7975, CurrentID=0, Status='live', Symbol='MAFK', Aliases='NFE2U, P18', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog K', other_designations='basic-leucine zipper transcription factor MafK|erythroid transcription factor NF-E2 p18 subunit|nuclear factor erythroid-2, ubiquitous (p18)|v-maf avian musculoaponeurotic fibrosarcoma oncogene family, protein K|v-maf musculoaponeurotic fibrosarcoma oncogene homolog K', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1530732, end_position_on_the_genomic_accession=1543043, orientation='plus', exon_count=5, OMIM=600197, no_hdr0=''),
  7976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7976, CurrentID=0, Status='live', Symbol='FZD3', Aliases='Fz-3', description='frizzled class receptor 3', other_designations='frizzled 3, seven transmembrane spanning receptor|frizzled family receptor 3|frizzled homolog 3', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28494205, end_position_on_the_genomic_accession=28574268, orientation='plus', exon_count=15, OMIM=606143, no_hdr0=''),
  7977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7977, CurrentID=0, Status='live', Symbol='MHS3', Aliases='', description='Malignant hyperthermia susceptibility 3', other_designations='', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=154276, no_hdr0=''),
  7978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7978, CurrentID=0, Status='live', Symbol='MTERF1', Aliases='MTERF', description='mitochondrial transcription termination factor 1', other_designations='transcription termination factor, mitochondrial', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91872876, end_position_on_the_genomic_accession=91880720, orientation='minus', exon_count=4, OMIM=602318, no_hdr0=''),
  7979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7979, CurrentID=0, Status='live', Symbol='SHFM1', Aliases='DSS1, ECD, SEM1, SHFD1, SHSF1, Shfdg1', description='split hand/foot malformation (ectrodactyly) type 1', other_designations='deleted in split hand/split foot protein 1|deleted in split-hand/split-foot 1|split hand/foot deleted protein 1|split hand/foot malformation type 1 protein', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96688767, end_position_on_the_genomic_accession=96709891, orientation='minus', exon_count=3, OMIM=601285, no_hdr0=''),
  7980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7980, CurrentID=0, Status='live', Symbol='TFPI2', Aliases='PP5, REF1, TFPI-2', description='tissue factor pathway inhibitor 2', other_designations='placental protein 5|retinal pigment epithelium cell factor 1', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93885397, end_position_on_the_genomic_accession=93890991, orientation='minus', exon_count=5, OMIM=600033, no_hdr0=''),
  100271410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271410, CurrentID=0, Status='live', Symbol='RPL12P39', Aliases='RPL12_20_1611', description='ribosomal protein L12 pseudogene 39', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64072170, end_position_on_the_genomic_accession=64072651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7982, CurrentID=0, Status='live', Symbol='ST7', Aliases='ETS7q, FAM4A, FAM4A1, HELG, RAY1, SEN4, TSG7', description='suppression of tumorigenicity 7', other_designations='family with sequence similarity 4, subfamily A, member 1|suppression of tumorigenicity 7 (breast)', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116953327, end_position_on_the_genomic_accession=117230023, orientation='plus', exon_count=17, OMIM=600833, no_hdr0=''),
  7984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7984, CurrentID=0, Status='live', Symbol='ARHGEF5', Aliases='GEF5, P60, TIM, TIM1', description='Rho guanine nucleotide exchange factor 5', other_designations='Rho guanine nucleotide exchange factor (GEF) 5|ephexin-3|guanine nucleotide regulatory protein TIM|oncogene TIM|p60 TIM|transforming immortalized mammary oncogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144355396, end_position_on_the_genomic_accession=144380632, orientation='plus', exon_count=15, OMIM=600888, no_hdr0=''),
  7986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7986, CurrentID=0, Status='live', Symbol='GPDS1', Aliases='PDS1', description='glaucoma-related pigment dispersion syndrome 1', other_designations='Pigment dispersion syndrome-1', map_location='7q35-q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600510, no_hdr0=''),
  1331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1331, CurrentID=0, Status='live', Symbol='COX5BP2', Aliases='COX5BL2', description='cytochrome c oxidase subunit Vb pseudogene 2', other_designations='cytochrome c oxidase subunit Vb-like 2', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136034515, end_position_on_the_genomic_accession=136034886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7988, CurrentID=0, Status='live', Symbol='ZNF212', Aliases='C2H2-150, ZNF182, ZNFC150', description='zinc finger protein 212', other_designations='Zinc finger protein C2H2-150', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149239648, end_position_on_the_genomic_accession=149255609, orientation='plus', exon_count=6, OMIM=602386, no_hdr0=''),
  100270755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270755, CurrentID=0, Status='live', Symbol='RPSAP63', Aliases='RPSA_33_1813', description='ribosomal protein SA pseudogene 63', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133762720, end_position_on_the_genomic_accession=133763753, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401206, CurrentID=0, Status='live', Symbol='RPS25P6', Aliases='RPS25_4_602', description='ribosomal protein S25 pseudogene 6', other_designations='hCG1640785', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116051856, end_position_on_the_genomic_accession=116052342, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7991, CurrentID=0, Status='live', Symbol='TUSC3', Aliases='D8S1992, M33, MRT22, MRT7, N33, OST3A', description='tumor suppressor candidate 3', other_designations='Putative prostate cancer tumor suppressor|magnesium uptake/transporter TUSC3|oligosaccharyltransferase 3 homolog A', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=15540087, end_position_on_the_genomic_accession=15766649, orientation='plus', exon_count=12, OMIM=601385, no_hdr0=''),
  7993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7993, CurrentID=0, Status='live', Symbol='UBXN8', Aliases='D8S2298E, REP8, UBXD6', description='UBX domain protein 8', other_designations='Reproduction/chromosome 8|UBX domain-containing protein 6|rep-8 protein|reproduction 8 protein', map_location='8p12-p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30729107, end_position_on_the_genomic_accession=30767006, orientation='plus', exon_count=11, OMIM=602155, no_hdr0=''),
  7994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7994, CurrentID=0, Status='live', Symbol='KAT6A', Aliases='MOZ, MRD32, MYST-3, MYST3, RUNXBP2, ZC2HC6A, ZNF220', description='K(lysine) acetyltransferase 6A', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 3|MYST histone acetyltransferase (monocytic leukemia) 3|histone acetyltransferase MYST3|monocytic leukemia zinc finger protein|runt-related transcription factor binding protein 2|zinc finger protein 220', map_location='8p11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41929479, end_position_on_the_genomic_accession=42051989, orientation='minus', exon_count=19, OMIM=601408, no_hdr0=''),
  139067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139067, CurrentID=0, Status='live', Symbol='SPANXN3', Aliases='CT11.8, SPANX-N3', description='SPANX family member N3', other_designations='SPANX family, member N3|cancer/testis antigen family 11, member 8|nuclear-associated protein SPAN-Xn3', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143508735, end_position_on_the_genomic_accession=143517475, orientation='minus', exon_count=2, OMIM=300666, no_hdr0=''),
  101929273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929273, CurrentID=0, Status='live', Symbol='LOC101929273', Aliases='', description='otogelin-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419149, CurrentID=0, Status='live', Symbol='HCFC2P1', Aliases='', description='host cell factor C2 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108480262, end_position_on_the_genomic_accession=108481776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1333, CurrentID=0, Status='live', Symbol='COX5BP4', Aliases='COX5BL4', description='cytochrome c oxidase subunit Vb pseudogene 4', other_designations='cytochrome c oxidase subunit Vb-like 4', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83106488, end_position_on_the_genomic_accession=83106983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8000, CurrentID=0, Status='live', Symbol='PSCA', Aliases='PRO232', description='prostate stem cell antigen', other_designations='', map_location='8q24.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142670308, end_position_on_the_genomic_accession=142682727, orientation='plus', exon_count=4, OMIM=602470, no_hdr0=''),
  8001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8001, CurrentID=0, Status='live', Symbol='GLRA3', Aliases='', description='glycine receptor alpha 3', other_designations='glycine receptor, alpha 3|glycine receptor, alpha-3 polypeptide|ligand gated ion channel', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174624141, end_position_on_the_genomic_accession=174829315, orientation='minus', exon_count=13, OMIM=600421, no_hdr0=''),
  401218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401218, CurrentID=0, Status='live', Symbol='LOC401218', Aliases='', description='family with sequence similarity 58, member A pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172762121, end_position_on_the_genomic_accession=172763351, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401220, CurrentID=0, Status='live', Symbol='LOC401220', Aliases='', description='uncharacterized LOC401220', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=177363850, end_position_on_the_genomic_accession=177367314, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  139081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139081, CurrentID=0, Status='live', Symbol='MAGEC3', Aliases='CT7.2, HCA2, MAGE-C3, MAGEC4', description='MAGE family member C3', other_designations='MAGE family testis and tumor-specific protein|MAGE-C3 antigen|cancer/testis antigen 7.2|cancer/testis antigen family 7, member 2|hepatocellular carcinoma-associated antigen 2|hepatocellular carcinoma-associated protein HCA2|melanoma antigen family C, 3|melanoma antigen family C3', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141838316, end_position_on_the_genomic_accession=141897832, orientation='plus', exon_count=12, OMIM=300469, no_hdr0=''),
  8011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8011, CurrentID=0, Status='live', Symbol='MROS', Aliases='MRS', description='Melkersson-Rosenthal syndrome', other_designations='', map_location='9p11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=155900, no_hdr0=''),
  8013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8013, CurrentID=0, Status='live', Symbol='NR4A3', Aliases='CHN, CSMF, MINOR, NOR1, TEC', description='nuclear receptor subfamily 4 group A member 3', other_designations='chondrosarcoma, extraskeletal myxoid, fused to EWS|mitogen-induced nuclear orphan receptor|neuron-derived orphan receptor 1|nuclear hormone receptor NOR-1|nuclear receptor subfamily 4, group A, member 3|translocated in extraskeletal chondrosarcoma', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99821855, end_position_on_the_genomic_accession=99866893, orientation='plus', exon_count=11, OMIM=600542, no_hdr0=''),
  401230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401230, CurrentID=0, Status='live', Symbol='FLJ46552', Aliases='', description='FLJ46552 protein', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1617163, end_position_on_the_genomic_accession=1620757, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105667213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105667213, CurrentID=0, Status='live', Symbol='LINC01601', Aliases='', description='long intergenic non-protein coding RNA 1601', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401232, CurrentID=0, Status='live', Symbol='LINC01011', Aliases='DKFZP686I15217', description='long intergenic non-protein coding RNA 1011', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2987967, end_position_on_the_genomic_accession=2991171, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  401233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401233, CurrentID=0, Status='live', Symbol='HTATSF1P2', Aliases='', description='HIV-1 Tat specific factor 1 pseudogene 2', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3020156, end_position_on_the_genomic_accession=3024771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8019, CurrentID=0, Status='live', Symbol='BRD3', Aliases='ORFX, RING3L', description='bromodomain containing 3', other_designations='RING3-like protein|bromodomain-containing 3', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134030305, end_position_on_the_genomic_accession=134068533, orientation='minus', exon_count=14, OMIM=601541, no_hdr0=''),
  401236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401236, CurrentID=0, Status='live', Symbol='STMND1', Aliases='', description='stathmin domain containing 1', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17102258, end_position_on_the_genomic_accession=17131372, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8021, CurrentID=0, Status='live', Symbol='NUP214', Aliases='CAIN, CAN', description='nucleoporin 214kDa', other_designations='CAN protein, putative oncogene', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131124959, end_position_on_the_genomic_accession=131234670, orientation='plus', exon_count=38, OMIM=114350, no_hdr0=''),
  8022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8022, CurrentID=0, Status='live', Symbol='LHX3', Aliases='CPHD3, LIM3, M2-LHX3', description='LIM homeobox 3', other_designations='LIM homeobox protein 3|LIM/homeodomain protein LHX3', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136196250, end_position_on_the_genomic_accession=136205109, orientation='minus', exon_count=8, OMIM=600577, no_hdr0=''),
  8023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8023, CurrentID=0, Status='live', Symbol='H142T', Aliases='', description='Temperature sensitivity complementation, H142', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=187290, no_hdr0=''),
  8024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8024, CurrentID=0, Status='live', Symbol='RDPA', Aliases='', description='refsum disease, adult, with increased pipecolicacidemia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600964, no_hdr0=''),
  8026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8026, CurrentID=0, Status='live', Symbol='DGS2', Aliases='DGCR2', description='DiGeorge syndrome/velocardiofacial syndrome complex 2', other_designations='', map_location='10p14-p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601362, no_hdr0=''),
  8027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8027, CurrentID=0, Status='live', Symbol='STAM', Aliases='STAM-11, STAM', description='signal transducing adaptor molecule', other_designations='HSE1 homolog|signal transducing adaptor molecule (SH3 domain and ITAM motif) 1', map_location='10p14-p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17644125, end_position_on_the_genomic_accession=17716824, orientation='plus', exon_count=16, OMIM=601899, no_hdr0=''),
  8028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8028, CurrentID=0, Status='live', Symbol='MLLT10', Aliases='AF10', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 10', other_designations='ALL1-fused gene from chromosome 10 protein|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 10|type I AF10 protein|type III AF10 protein|type IV AF10 protein', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21533823, end_position_on_the_genomic_accession=21743633, orientation='plus', exon_count=34, OMIM=602409, no_hdr0=''),
  8029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8029, CurrentID=0, Status='live', Symbol='CUBN', Aliases='IFCR, MGA1, gp280', description='cubilin', other_designations='460 kDa receptor|cubilin (intrinsic factor-cobalamin receptor)|cubilin precursor variant 1|cubilin precursor variant 2|cubilin precursor variant 3|intestinal intrinsic factor receptor|intrinsic factor-vitamin B12 receptor', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16823966, end_position_on_the_genomic_accession=17130249, orientation='minus', exon_count=71, OMIM=602997, no_hdr0=''),
  8030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8030, CurrentID=0, Status='live', Symbol='CCDC6', Aliases='D10S170, H4, PTC, TPC, TST1', description='coiled-coil domain containing 6', other_designations='papillary thyroid carcinoma-encoded protein', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59788747, end_position_on_the_genomic_accession=59906656, orientation='minus', exon_count=9, OMIM=601985, no_hdr0=''),
  401247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401247, CurrentID=0, Status='live', Symbol='LINC00243', Aliases='C6orf214, NCRNA00243', description='long intergenic non-protein coding RNA 243', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30812866, end_position_on_the_genomic_accession=30830659, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  139105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139105, CurrentID=0, Status='live', Symbol='BEND2', Aliases='CXorf20', description='BEN domain containing 2', other_designations='', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18162931, end_position_on_the_genomic_accession=18220904, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  8034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8034, CurrentID=0, Status='live', Symbol='SLC25A16', Aliases='D10S105E, GDA, GDC, HGT.1, ML7, hML7', description='solute carrier family 25 (mitochondrial carrier), member 16', other_designations='graves disease autoantigen|mitochondrial solute carrier protein homolog|solute carrier family 25 (mitochondrial carrier; Graves disease autoantigen), member 16|solute carrier family 25 member 16', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68473387, end_position_on_the_genomic_accession=68527523, orientation='minus', exon_count=13, OMIM=139080, no_hdr0=''),
  401251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401251, CurrentID=0, Status='live', Symbol='SAPCD1', Aliases='C6orf26, NG23', description='suppressor APC domain containing 1', other_designations='protein G7d', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31762996, end_position_on_the_genomic_accession=31764850, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8036, CurrentID=0, Status='live', Symbol='SHOC2', Aliases='SIAA0862, SOC2, SUR8', description='SHOC2 leucine-rich repeat scaffold protein', other_designations='soc-2 suppressor of clear homolog', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110919543, end_position_on_the_genomic_accession=111013667, orientation='plus', exon_count=9, OMIM=602775, no_hdr0=''),
  401253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401253, CurrentID=0, Status='live', Symbol='LINC00336', Aliases='C6orf227, NCRNA00336', description='long intergenic non-protein coding RNA 336', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33586106, end_position_on_the_genomic_accession=33593338, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8038, CurrentID=0, Status='live', Symbol='ADAM12', Aliases='ADAM12-OT1, CAR10, MCMP, MCMPMltna, MLTN, MLTNA', description='ADAM metallopeptidase domain 12', other_designations='meltrin-alpha|metalloprotease-disintegrin 12 transmembrane', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126012385, end_position_on_the_genomic_accession=126388558, orientation='minus', exon_count=24, OMIM=602714, no_hdr0=''),
  8039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8039, CurrentID=0, Status='live', Symbol='PROA', Aliases='', description='Proline(-) auxotroph, complementation of', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=176770, no_hdr0=''),
  8041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8041, CurrentID=0, Status='live', Symbol='TP250', Aliases='', description='T-cell activation antigen p250', other_designations='', map_location='11pter-p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=186710, no_hdr0=''),
  401258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401258, CurrentID=0, Status='live', Symbol='RAB44', Aliases='RASD3, RASL13, dJ431A14.3', description='RAB44, member RAS oncogene family', other_designations='RAS-like, family 13|RASD family, member 3', map_location='6p21.31-p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36697827, end_position_on_the_genomic_accession=36733183, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  401260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401260, CurrentID=0, Status='live', Symbol='LINC00951', Aliases='', description='long intergenic non-protein coding RNA 951', other_designations='lincRNA-uc003opf.1', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40344345, end_position_on_the_genomic_accession=40356006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8045, CurrentID=0, Status='live', Symbol='RASSF7', Aliases='C11orf13, HRAS1, HRC1', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 7', other_designations='HRAS1-related cluster protein 1|HRAS1-related cluster-1|Ras association (RalGDS/AF-6) domain family 7', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=559776, end_position_on_the_genomic_accession=564025, orientation='plus', exon_count=8, OMIM=143023, no_hdr0=''),
  401262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401262, CurrentID=0, Status='live', Symbol='CRIP3', Aliases='TLP, TLP-A, bA480N24.2', description='cysteine-rich protein 3', other_designations='CRP-3|chromosome 6 LIM domain only protein|h6LIMo|thymus LIM protein TLP-A', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43299741, end_position_on_the_genomic_accession=43309614, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  8048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8048, CurrentID=0, Status='live', Symbol='CSRP3', Aliases='CLP, CMD1M, CMH12, CRP3, LMO4, MLP', description='cysteine and glycine rich protein 3', other_designations='LIM domain only 4|cardiac LIM domain protein|cysteine and glycine-rich protein 3 (cardiac LIM protein)|muscle lim protein isoform', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19182030, end_position_on_the_genomic_accession=19210571, orientation='minus', exon_count=7, OMIM=600824, no_hdr0=''),
  401265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401265, CurrentID=0, Status='live', Symbol='KLHL31', Aliases='BKLHD6, KBTBD1, KLHL, bA345L23.2', description='kelch like family member 31', other_designations='BTB and kelch domain-containing protein 6|kelch repeat and BTB (POZ) domain containing 1|kelch repeat and BTB domain-containing protein 1|kelch-like family member 31|kelch-like protein KLHL', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53647901, end_position_on_the_genomic_accession=53665846, orientation='minus', exon_count=3, OMIM=610749, no_hdr0=''),
  8050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8050, CurrentID=0, Status='live', Symbol='PDHX', Aliases='DLDBP, E3BP, OPDX, PDX1, proX', description='pyruvate dehydrogenase complex component X', other_designations='dihydrolipoamide dehydrogenase-binding protein of pyruvate dehydrogenase complex|lipoyl-containing pyruvate dehydrogenase complex component X|pyruvate dehydrogenase complex, E3-binding protein subunit|pyruvate dehydrogenase complex, component X|pyruvate dehydrogenase complex, lipoyl-containing component X', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34915829, end_position_on_the_genomic_accession=34996128, orientation='plus', exon_count=12, OMIM=608769, no_hdr0=''),
  23187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23187, CurrentID=0, Status='live', Symbol='PHLDB1', Aliases='LL5A', description='pleckstrin homology-like domain, family B, member 1', other_designations='LL5alpha|pleckstrin homology-like domain family B member 1 variant 3|pleckstrin homology-like domain family B member 1 variant 4|protein LL5-alpha', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118606439, end_position_on_the_genomic_accession=118658038, orientation='plus', exon_count=35, OMIM=612834, no_hdr0=''),
  100271422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271422, CurrentID=0, Status='live', Symbol='RPL21P89', Aliases='RPL21_37_1041', description='ribosomal protein L21 pseudogene 89', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49815058, end_position_on_the_genomic_accession=49815598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8056, CurrentID=0, Status='live', Symbol='ECB2', Aliases='FBP', description='erythrocytosis, autosomal recessive benign 2', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=263400, no_hdr0=''),
  100271423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271423, CurrentID=0, Status='live', Symbol='RPL21P90', Aliases='RPL21_37_1084', description='ribosomal protein L21 pseudogene 90', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95999734, end_position_on_the_genomic_accession=96000286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8060, CurrentID=0, Status='live', Symbol='FNL2', Aliases='', description='Fibronectin-like-2', other_designations='', map_location='11q12.1-q13.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=135610, no_hdr0=''),
  8061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8061, CurrentID=0, Status='live', Symbol='FOSL1', Aliases='FRA, FRA1, fra-1', description='FOS-like antigen 1', other_designations='FOS-like antigen-1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65892136, end_position_on_the_genomic_accession=65900526, orientation='minus', exon_count=4, OMIM=136515, no_hdr0=''),
  139135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139135, CurrentID=0, Status='live', Symbol='PASD1', Aliases='CT63, OXTES1', description='PAS domain containing 1', other_designations='cancer/testis antigen 63', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151563535, end_position_on_the_genomic_accession=151676739, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  8065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8065, CurrentID=0, Status='live', Symbol='CUL5', Aliases='VACM-1, VACM1', description='cullin 5', other_designations='CUL-5|Cullin-5 (vasopressin-activated calcium-mobilizing receptor-1)|Vasopressin-activated calcium-mobilizing receptor-1|vasopressin-activated calcium-mobilizing receptor 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108008682, end_position_on_the_genomic_accession=108107766, orientation='plus', exon_count=21, OMIM=601741, no_hdr0=''),
  8066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8066, CurrentID=0, Status='live', Symbol='ANC', Aliases='', description='Anal canal carcinoma', other_designations='', map_location='11q22-qter', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=105580, no_hdr0=''),
  8068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8068, CurrentID=0, Status='live', Symbol='BRCATA', Aliases='', description='Breast cancer, 11;22 translocation associated', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600048, no_hdr0=''),
  401285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401285, CurrentID=0, Status='live', Symbol='TCP10L2', Aliases='bA517H2.3', description='t-complex 10-like 2', other_designations='T-complex 10-like protein 2|putative t-complex protein 10A homolog 2', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167170593, end_position_on_the_genomic_accession=167193779, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  401286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401286, CurrentID=0, Status='live', Symbol='LOC401286', Aliases='', description='uncharacterized LOC401286', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167666840, end_position_on_the_genomic_accession=167679223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271425, CurrentID=0, Status='live', Symbol='RPL21P94', Aliases='RPL21_39_1117', description='ribosomal protein L21 pseudogene 94', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6685846, end_position_on_the_genomic_accession=6686411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401288, CurrentID=0, Status='live', Symbol='LINC00242', Aliases='C6orf122, NCRNA00242, dJ266L20.5', description='long intergenic non-protein coding RNA 242', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169788790, end_position_on_the_genomic_accession=169798825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8073, CurrentID=0, Status='live', Symbol='PTP4A2', Aliases='HH13, HH7-2, HU-PP-1, OV-1, PRL-2, PRL2, PTP4A, PTPCAAX2, ptp-IV1a, ptp-IV1b', description='protein tyrosine phosphatase type IVA, member 2', other_designations='PTP(CAAXII)|phosphatase of regenerating liver 2|protein tyrosine phosphatase IVA|protein tyrosine phosphatase IVA2|protein-tyrosine phosphatase 4a2|protein-tyrosine phosphatase of regenerating liver 2', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31906421, end_position_on_the_genomic_accession=31938387, orientation='minus', exon_count=9, OMIM=601584, no_hdr0=''),
  8074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8074, CurrentID=0, Status='live', Symbol='FGF23', Aliases='ADHR, FGFN, HPDR2, HYPF, PHPTC', description='fibroblast growth factor 23', other_designations='phosphatonin|tumor-derived hypophosphatemia inducing factor', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4368227, end_position_on_the_genomic_accession=4379728, orientation='minus', exon_count=3, OMIM=605380, no_hdr0=''),
  8075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8075, CurrentID=0, Status='live', Symbol='PCS', Aliases='', description='Parotid proline-rich salivary protein Pc', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=168710, no_hdr0=''),
  8076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8076, CurrentID=0, Status='live', Symbol='MFAP5', Aliases='AAT9, MAGP-2, MAGP2, MFAP-5, MP25', description='microfibrillar associated protein 5', other_designations='THE1A-MFAP5|microfibril-associated glycoprotein-2', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8645943, end_position_on_the_genomic_accession=8662888, orientation='minus', exon_count=11, OMIM=601103, no_hdr0=''),
  8078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8078, CurrentID=0, Status='live', Symbol='USP5', Aliases='ISOT', description='ubiquitin specific peptidase 5 (isopeptidase T)', other_designations='deubiquitinating enzyme 5|isopeptidase T|ubiquitin isopeptidase T|ubiquitin specific protease 5 (isopeptidase T)|ubiquitin thioesterase 5|ubiquitin thiolesterase 5|ubiquitin-specific protease-5 (ubiquitin isopeptidase T)|ubiquitin-specific-processing protease 5', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6852118, end_position_on_the_genomic_accession=6866632, orientation='plus', exon_count=20, OMIM=601447, no_hdr0=''),
  8079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8079, CurrentID=0, Status='live', Symbol='MLF2', Aliases='NTN4', description='myeloid leukemia factor 2', other_designations='myelodysplasia-myeloid leukemia factor 2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6747992, end_position_on_the_genomic_accession=6753470, orientation='minus', exon_count=10, OMIM=601401, no_hdr0=''),
  8080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8080, CurrentID=0, Status='live', Symbol='HTNB', Aliases='', description='hypertension with brachydactyly', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=112410, no_hdr0=''),
  8081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8081, CurrentID=0, Status='live', Symbol='AD5', Aliases='', description='Alzheimer disease 5', other_designations='Alzheimer disease, familial, type 5', map_location='12p11.23-q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602096, no_hdr0=''),
  8082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8082, CurrentID=0, Status='live', Symbol='SSPN', Aliases='DAGA5, KRAG, NSPN, SPN1, SPN2', description='sarcospan', other_designations='K-ras oncogene-associated protein|Kras oncogene-associated|kirsten-ras-associated protein|microspan|nanospan|sarcospan (Kras oncogene-associated gene)', map_location='12p11.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26121951, end_position_on_the_genomic_accession=26234777, orientation='plus', exon_count=7, OMIM=601599, no_hdr0=''),
  8083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8083, CurrentID=0, Status='live', Symbol='KAR', Aliases='', description='Aromatic alpha-keto acid reductase', other_designations='', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=107920, no_hdr0=''),
  8085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8085, CurrentID=0, Status='live', Symbol='KMT2D', Aliases='AAD10, ALR, CAGL114, KABUK1, KMS, MLL2, MLL4, TNRC21', description='lysine (K)-specific methyltransferase 2D', other_designations='ALL1-related protein|Kabuki make-up syndrome|Kabuki mental retardation syndrome|histone-lysine N-methyltransferase MLL2|lysine N-methyltransferase 2D|myeloid/lymphoid or mixed-lineage leukemia 2|trinucleotide repeat containing 21', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49018975, end_position_on_the_genomic_accession=49061895, orientation='minus', exon_count=56, OMIM=602113, no_hdr0=''),
  8086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8086, CurrentID=0, Status='live', Symbol='AAAS', Aliases='AAAb, ADRACALA, ADRACALIN, ALADIN, GL003, AAAS', description='aladin WD repeat nucleoporin', other_designations='Allgrove, triple-A|achalasia, adrenocortical insufficiency, alacrimia', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53307456, end_position_on_the_genomic_accession=53321628, orientation='minus', exon_count=18, OMIM=605378, no_hdr0=''),
  8087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8087, CurrentID=0, Status='live', Symbol='FXR1', Aliases='FXR1P', description='fragile X mental retardation, autosomal homolog 1', other_designations='fragile X mental retardation autosomal homolog variant p1K|fragile X mental retardation autosomal homolog variant p2K|fragile X mental retardation autosomal homolog variant p4K|fragile X mental retardation autosomal homolog variant p5FK|hFXR1p', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180912446, end_position_on_the_genomic_accession=180982753, orientation='plus', exon_count=20, OMIM=600819, no_hdr0=''),
  8089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8089, CurrentID=0, Status='live', Symbol='YEATS4', Aliases='4930573H17Rik, B230215M10Rik, GAS41, NUBI-1, YAF9', description='YEATS domain containing 4', other_designations='NuMA binding protein 1|glioma-amplified sequence 41|glioma-amplified sequence-41|nuBI1|nuMA-binding protein 1', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69359700, end_position_on_the_genomic_accession=69424838, orientation='plus', exon_count=9, OMIM=602116, no_hdr0=''),
  8090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8090, CurrentID=0, Status='live', Symbol='SPPM', Aliases='SPMD', description='scapuloperoneal syndrome, myopathic type', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=181430, no_hdr0=''),
  8091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8091, CurrentID=0, Status='live', Symbol='HMGA2', Aliases='BABL, HMGI-C, HMGIC, LIPO, STQTL9', description='high mobility group AT-hook 2', other_designations='High-mobility group protein HMGI-C|high-mobility group (nonhistone chromosomal) protein isoform I-C', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65824460, end_position_on_the_genomic_accession=65966291, orientation='plus', exon_count=7, OMIM=600698, no_hdr0=''),
  8092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8092, CurrentID=0, Status='live', Symbol='ALX1', Aliases='CART1, FND3, HEL23', description='ALX homeobox 1', other_designations='CART-1|cartilage paired-class homeoprotein 1|epididymis luminal protein 23', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85279795, end_position_on_the_genomic_accession=85301784, orientation='plus', exon_count=6, OMIM=601527, no_hdr0=''),
  8093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8093, CurrentID=0, Status='live', Symbol='MGCT', Aliases='', description='male germ cell tumor', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=273300, no_hdr0=''),
  8094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8094, CurrentID=0, Status='live', Symbol='SMAL', Aliases='', description='spinal muscular atrophy, congenital nonprogressive, of lower limbs', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600175, no_hdr0=''),
  100271429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271429, CurrentID=0, Status='live', Symbol='RPL21P99', Aliases='RPL21_41_1222', description='ribosomal protein L21 pseudogene 99', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29658341, end_position_on_the_genomic_accession=29658840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401312, CurrentID=0, Status='live', Symbol='LOC401312', Aliases='', description='uncharacterized LOC401312', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22649960, end_position_on_the_genomic_accession=22665533, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  139170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139170, CurrentID=0, Status='live', Symbol='DCAF12L1', Aliases='KIAA1892L, WDR40B', description='DDB1 and CUL4 associated factor 12-like 1', other_designations='WD repeat domain 40B|WD repeat-containing protein 40B', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126549383, end_position_on_the_genomic_accession=126552859, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8099, CurrentID=0, Status='live', Symbol='CDK2AP1', Aliases='DOC1, DORC1, ST19, doc-1, p12DOC-1', description='cyclin-dependent kinase 2 associated protein 1', other_designations='CDK2-associated protein 1|Deleted in oral cancer-1|deleted in oral cancer 1|putative oral cancer suppressor', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123260970, end_position_on_the_genomic_accession=123272316, orientation='minus', exon_count=8, OMIM=602198, no_hdr0=''),
  8100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8100, CurrentID=0, Status='live', Symbol='IFT88', Aliases='D13S1056E, DAF19, TG737, TTC10, hTg737', description='intraflagellar transport 88', other_designations='TPR repeat protein 10|intraflagellar transport 88 homolog|polaris homolog|probe hTg737 (polycystic kidney disease, autosomal recessive)|recessive polycystic kidney disease protein Tg737 homolog|tetratricopeptide repeat domain 10|tetratricopeptide repeat protein 10', map_location='13q12.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20566446, end_position_on_the_genomic_accession=20691437, orientation='plus', exon_count=31, OMIM=600595, no_hdr0=''),
  8101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8101, CurrentID=0, Status='live', Symbol='CLLS2', Aliases='D13S25, DBM', description='Disrupted in B-cell neoplasia', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109543, no_hdr0=''),
  8102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8102, CurrentID=0, Status='live', Symbol='XRS', Aliases='', description='X-ray sensitivity', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194370, no_hdr0=''),
  8103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8103, CurrentID=0, Status='live', Symbol='PAPA2', Aliases='', description='postaxial polydactyly, type A2', other_designations='', map_location='13q21-q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602085, no_hdr0=''),
  8104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8104, CurrentID=0, Status='live', Symbol='MCOR', Aliases='C13DELq32, DEL13q32', description='microcoria, congenital', other_designations='', map_location='13q31-q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=156600, no_hdr0=''),
  8105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8105, CurrentID=0, Status='live', Symbol='BRCD1', Aliases='BCDS1', description='cancer, familial, with in vitro radioresistance', other_designations='Breast cancer, ductal, suppressor-1', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=114450, no_hdr0=''),
  8106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8106, CurrentID=0, Status='live', Symbol='PABPN1', Aliases='OPMD, PAB2, PABII, PABP-2, PABP2', description='poly(A) binding protein, nuclear 1', other_designations='poly(A) binding protein 2|poly(A) binding protein II', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23320188, end_position_on_the_genomic_accession=23326185, orientation='plus', exon_count=7, OMIM=602279, no_hdr0=''),
  100271431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271431, CurrentID=0, Status='live', Symbol='RPL21P101', Aliases='RPL21_42_1230', description='ribosomal protein L21 pseudogene 101', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43560766, end_position_on_the_genomic_accession=43561336, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401324, CurrentID=0, Status='live', Symbol='LOC401324', Aliases='', description='uncharacterized LOC401324', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35313855, end_position_on_the_genomic_accession=35376476, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8110, CurrentID=0, Status='live', Symbol='DPF3', Aliases='BAF45C, CERD4', description='D4, zinc and double PHD fingers, family 3', other_designations='BRG1-associated factor 45C|zinc finger protein cer-d4', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72609032, end_position_on_the_genomic_accession=72894116, orientation='minus', exon_count=20, OMIM=601672, no_hdr0=''),
  8111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8111, CurrentID=0, Status='live', Symbol='GPR68', Aliases='GPR12A, OGR1', description='G protein-coupled receptor 68', other_designations='G-protein coupled receptor 68|ovarian cancer G protein-coupled receptor, 1|sphingosylphosphorylcholine receptor', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91232532, end_position_on_the_genomic_accession=91270790, orientation='minus', exon_count=12, OMIM=601404, no_hdr0=''),
  8113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8113, CurrentID=0, Status='live', Symbol='MCOP1', Aliases='MCOP', description='microphthalmia, autosomal recessive', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=251600, no_hdr0=''),
  8114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8114, CurrentID=0, Status='live', Symbol='IV', Aliases='SIV', description='inversus situs, viscerum', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=270100, no_hdr0=''),
  8115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8115, CurrentID=0, Status='live', Symbol='TCL1A', Aliases='TCL1', description='T-cell leukemia/lymphoma 1A', other_designations='T-cell lymphoma-1|oncogene TCL-1|oncogene TCL1|protein p14 TCL1', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95709967, end_position_on_the_genomic_accession=95714196, orientation='minus', exon_count=5, OMIM=186960, no_hdr0=''),
  139189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139189, CurrentID=0, Status='live', Symbol='DGKK', Aliases='', description='diacylglycerol kinase, kappa', other_designations='142 kDa diacylglycerol kinase|DAG kinase kappa|DGK-kappa|diglyceride kinase kappa', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50365407, end_position_on_the_genomic_accession=50470738, orientation='minus', exon_count=28, OMIM=300837, no_hdr0=''),
  8118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8118, CurrentID=0, Status='live', Symbol='LCH', Aliases='', description='Lentil agglutinin-binding', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151020, no_hdr0=''),
  401335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401335, CurrentID=0, Status='live', Symbol='C7orf65', Aliases='', description='chromosome 7 open reading frame 65', other_designations='FLJ44108 protein', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47655244, end_position_on_the_genomic_accession=47661648, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8120, CurrentID=0, Status='live', Symbol='AP3B2', Aliases='NAPTB', description='adaptor related protein complex 3, beta 2 subunit', other_designations='Neuronal adaptin-like protein, beta-subunit|adapter-related protein complex 3 subunit beta-2|adaptor protein complex AP-3 subunit beta-2|adaptor-related protein complex 3 subunit beta-2|adaptor-related protein complex 3, beta 2 subunit|beta-3B-adaptin|clathrin assembly protein complex 3 beta-2 large chain|neuron-specific vesicle coat protein beta-NAP', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82659281, end_position_on_the_genomic_accession=82709908, orientation='minus', exon_count=27, OMIM=602166, no_hdr0=''),
  401337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401337, CurrentID=0, Status='live', Symbol='LINC01446', Aliases='GS1-179L18.1', description='long intergenic non-protein coding RNA 1446', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53655509, end_position_on_the_genomic_accession=53811931, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23199, CurrentID=0, Status='live', Symbol='GSE1', Aliases='KIAA0182', description='Gse1 coiled-coil protein', other_designations='CTC-786C10.1|Gse1 coiled-coil protein homolog', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85170003, end_position_on_the_genomic_accession=85676206, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  8125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8125, CurrentID=0, Status='live', Symbol='ANP32A', Aliases='C15orf1, HPPCn, I1PP2A, LANP, MAPM, PHAP1, PHAPI, PP32', description='acidic nuclear phosphoprotein 32 family member A', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member A|acidic nuclear phosphoprotein pp32|cerebellar leucine rich acidic nuclear protein|hepatopoietin Cn|inhibitor-1 of protein phosphatase-2A|leucine-rich acidic nuclear protein|mapmodulin|potent heat-stable protein phosphatase 2A inhibitor I1PP2A|putative HLA-DR-associated protein I|putative human HLA class II associated protein I', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68778535, end_position_on_the_genomic_accession=68820922, orientation='minus', exon_count=7, OMIM=600832, no_hdr0=''),
  8126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8126, CurrentID=0, Status='live', Symbol='MRST', Aliases='', description='Mental retardation, severe, with spasticity and tapetoretinal', other_designations='', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602685, no_hdr0=''),
  8128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8128, CurrentID=0, Status='live', Symbol='ST8SIA2', Aliases='HsT19690, SIAT8B, ST8SIA-II, STX', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 2', other_designations='SIAT8-B|ST8SiaII|alpha-2,8-sialyltransferase 8B 1|sialyltransferase 8 (alpha-2, 8-sialytransferase) B|sialyltransferase 8B|sialyltransferase St8Sia II|sialyltransferase X|sialytransferase St8Sia II', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92393910, end_position_on_the_genomic_accession=92468728, orientation='plus', exon_count=6, OMIM=602546, no_hdr0=''),
  139201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139201, CurrentID=0, Status='live', Symbol='MAP2K4P1', Aliases='', description='mitogen-activated protein kinase kinase 4 pseudogene 1', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73524275, end_position_on_the_genomic_accession=73563085, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8130, CurrentID=0, Status='live', Symbol='MCOPCT1', Aliases='CATM', description='cataract, congenital, with microphthalmia', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=156850, no_hdr0=''),
  8131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8131, CurrentID=0, Status='live', Symbol='NPRL3', Aliases='C16orf35, CGTHBA, HS-40, MARE, NPR3, RMD11', description='NPR3-like, GATOR1 complex subunit', other_designations='-14 gene protein|alpha-globin regulatory element-containing gene protein|conserved gene telomeric to alpha globin cluster', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84271, end_position_on_the_genomic_accession=138698, orientation='minus', exon_count=17, OMIM=600928, no_hdr0=''),
  8132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8132, CurrentID=0, Status='live', Symbol='PKDTS', Aliases='', description='polycystic kidney disease, infantile severe, with tuberous sclerosis', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600273, no_hdr0=''),
  104472519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472519, CurrentID=0, Status='live', Symbol='STX18-IT1', Aliases='', description='STX18 intronic transcript 1', other_designations='STX18 intronic transcript 1 (non-protein coding)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4476121, end_position_on_the_genomic_accession=4481700, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8136, CurrentID=0, Status='live', Symbol='WT3', Aliases='', description='Wilms tumor-3', other_designations='', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194090, no_hdr0=''),
  8137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8137, CurrentID=0, Status='live', Symbol='HYD2', Aliases='', description='hypodontia 2 (autosomal recessive)', other_designations='hypodontia, autosomal recessive (2)', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602639, no_hdr0=''),
  8139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8139, CurrentID=0, Status='live', Symbol='GAN', Aliases='GAN1, KLHL16', description='gigaxonin', other_designations='kelch-like family member 16|kelch-like protein 16', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81314966, end_position_on_the_genomic_accession=81380198, orientation='plus', exon_count=11, OMIM=605379, no_hdr0=''),
  8140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8140, CurrentID=0, Status='live', Symbol='SLC7A5', Aliases='4F2LC, CD98, D16S469E, E16, LAT1, MPE16, hLAT1', description='solute carrier family 7 (amino acid transporter light chain, L system), member 5', other_designations='4F2 LC|4F2 light chain|CD98 light chain|L-type amino acid transporter 1|integral membrane protein E16|large neutral amino acids transporter 1|sodium-independent neutral amino acid transporter LAT1|solute carrier family 7 (cationic amino acid transporter, y+ system), member 5|solute carrier family 7 member 5|y+ system cationic amino acid transporter', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87830023, end_position_on_the_genomic_accession=87869497, orientation='minus', exon_count=12, OMIM=600182, no_hdr0=''),
  401357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401357, CurrentID=0, Status='live', Symbol='LOC401357', Aliases='', description='uncharacterized LOC401357', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56809124, end_position_on_the_genomic_accession=56812736, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8142, CurrentID=0, Status='live', Symbol='BCPR', Aliases='', description='Breast cancer-related regulator of TP53', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=113721, no_hdr0=''),
  100271437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271437, CurrentID=0, Status='live', Symbol='RPL21P109', Aliases='RPL21_49_1320', description='ribosomal protein L21 pseudogene 109', other_designations='', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71704140, end_position_on_the_genomic_accession=71704693, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8144, CurrentID=0, Status='live', Symbol='CMS1A1', Aliases='FIM1, FIMG, FIMG1, MGI', description='myasthenia gravis, familial infantile, 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605809, no_hdr0=''),
  8148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8148, CurrentID=0, Status='live', Symbol='TAF15', Aliases='Npl3, RBP56, TAF2N, TAFII68', description='TATA-box binding protein associated factor 15', other_designations='RBP56/CSMF fusion|TAF15 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 68kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, N, 68kD (RNA-binding protein 56)|TATA box binding protein associated factor 15|TATA box-binding protein-associated factor 2N (RNA-binding protein 56)|TBP-associated factor 15', map_location='17q11.1-q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35809455, end_position_on_the_genomic_accession=35847242, orientation='plus', exon_count=17, OMIM=601574, no_hdr0=''),
  139221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139221, CurrentID=0, Status='live', Symbol='MUM1L1', Aliases='', description='melanoma associated antigen (mutated) 1-like 1', other_designations='MUM1-like protein 1|mutated melanoma-associated antigen 1-like protein 1', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106168278, end_position_on_the_genomic_accession=106208961, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8151, CurrentID=0, Status='live', Symbol='WT4', Aliases='FWT1', description='Wilms tumor-4', other_designations='', map_location='17q12-q21', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601363, no_hdr0=''),
  8153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8153, CurrentID=0, Status='live', Symbol='RND2', Aliases='ARHN, RHO7, RhoN', description='Rho family GTPase 2', other_designations='CTD-3199J23.4|GTP-binding protein Rho7|ras homolog gene family, member N|rho-related GTP-binding protein Rho7', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43025241, end_position_on_the_genomic_accession=43032041, orientation='plus', exon_count=7, OMIM=601555, no_hdr0=''),
  100271439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271439, CurrentID=0, Status='live', Symbol='RPL21P111', Aliases='RPL21_51_1327', description='ribosomal protein L21 pseudogene 111', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78761134, end_position_on_the_genomic_accession=78761686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8156, CurrentID=0, Status='live', Symbol='GPSC', Aliases='', description='gliosis, familial progressive subcortical', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=221820, no_hdr0=''),
  139231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139231, CurrentID=0, Status='live', Symbol='FAM199X', Aliases='CXorf39', description='family with sequence similarity 199, X-linked', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104166475, end_position_on_the_genomic_accession=104195902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8160, CurrentID=0, Status='live', Symbol='MVWF', Aliases='', description='Modifier of von Willebrand factor', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8161, CurrentID=0, Status='live', Symbol='COIL', Aliases='CLN80, p80-coilin', description='coilin', other_designations='coilin p80|p80', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56938199, end_position_on_the_genomic_accession=56961050, orientation='minus', exon_count=7, OMIM=600272, no_hdr0=''),
  8163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8163, CurrentID=0, Status='live', Symbol='CDR3', Aliases='', description='Cerebellar degeneration-related autoantigen-3', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602197, no_hdr0=''),
  8165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8165, CurrentID=0, Status='live', Symbol='AKAP1', Aliases='AKAP21, AKAP149, AKAP84, D-AKAP1, PPP1R43, PRKA1, SAKAP84, TDRD17, AKAP1', description='A-kinase anchoring protein 1', other_designations='A kinase (PRKA) anchor protein 1|A-kinase anchor protein 149 kDa|AKAP 149|D-AKAP-1|dual-specificity A-kinase anchoring protein 1|protein kinase A anchoring protein 1|protein kinase A1|protein phosphatase 1, regulatory subunit 43|spermatid A-kinase anchor protein 84|tudor domain containing 17', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57085192, end_position_on_the_genomic_accession=57121349, orientation='plus', exon_count=15, OMIM=602449, no_hdr0=''),
  100271441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271441, CurrentID=0, Status='live', Symbol='RPL21P114', Aliases='RPL21_52_1438', description='ribosomal protein L21 pseudogene 114', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59525385, end_position_on_the_genomic_accession=59525924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8168, CurrentID=0, Status='live', Symbol='TW', Aliases='', description='Twirler mouse mutation, human homolog of', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8170, CurrentID=0, Status='live', Symbol='SLC14A2', Aliases='HUT2, UT-A2, UT2, UTA, UTR, hUT-A6', description='solute carrier family 14 (urea transporter), member 2', other_designations='solute carrier family 14 member 2|urea transporter, kidney|urea transporter-2', map_location='18q12.1-q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45212957, end_position_on_the_genomic_accession=45683107, orientation='plus', exon_count=23, OMIM=601611, no_hdr0=''),
  401387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401387, CurrentID=0, Status='live', Symbol='LRRD1', Aliases='', description='leucine-rich repeats and death domain containing 1', other_designations='leucine-rich repeat and death domain-containing protein LOC401387', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92144884, end_position_on_the_genomic_accession=92165276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  401388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401388, CurrentID=0, Status='live', Symbol='C7orf76', Aliases='', description='chromosome 7 open reading frame 76', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96481626, end_position_on_the_genomic_accession=96503523, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8173, CurrentID=0, Status='live', Symbol='CNSN', Aliases='', description='Carnosinemia (carnosinase)', other_designations='', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=212200, no_hdr0=''),
  8174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8174, CurrentID=0, Status='live', Symbol='MADCAM1', Aliases='MACAM1', description='mucosal vascular addressin cell adhesion molecule 1', other_designations='MAdCAM-1|hMAdCAM-1|mucosal addressin cell adhesion molecule-1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=496490, end_position_on_the_genomic_accession=505343, orientation='plus', exon_count=5, OMIM=102670, no_hdr0=''),
  8175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8175, CurrentID=0, Status='live', Symbol='SF3A2', Aliases='PRP11, PRPF11, SAP62, SF3a66', description='splicing factor 3a subunit 2', other_designations='SAP 62|pre-mRNA splicing factor SF3A, subunit 2|spliceosome associated protein 62|spliceosome-associated protein 62|splicing factor 3a, subunit 2, 66kD|splicing factor 3a, subunit 2, 66kDa', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2236817, end_position_on_the_genomic_accession=2248679, orientation='plus', exon_count=9, OMIM=600796, no_hdr0=''),
  401393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401393, CurrentID=0, Status='live', Symbol='AZGP1P2', Aliases='', description='alpha-2-glycoprotein 1, zinc-binding pseudogene 2', other_designations='Putative zinc-alpha-2-glycoprotein-like 2|alpha-2-glycoprotein 1, zinc pseudogene 2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101287479, end_position_on_the_genomic_accession=101289808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8178, CurrentID=0, Status='live', Symbol='ELL', Aliases='C19orf171, MEN, PPP1R68, ELL', description='elongation factor RNA polymerase II', other_designations='ELL gene (11-19 lysine-rich leukemia gene)|eleven-nineteen lysine-rich leukemia protein|protein phosphatase 1, regulatory subunit 68', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18442663, end_position_on_the_genomic_accession=18522129, orientation='minus', exon_count=14, OMIM=600284, no_hdr0=''),
  100271443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271443, CurrentID=0, Status='live', Symbol='RPL21P115', Aliases='RPL21_54_1451', description='ribosomal protein L21 pseudogene 115', other_designations='', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72134602, end_position_on_the_genomic_accession=72135055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401396, CurrentID=0, Status='live', Symbol='LOC401396', Aliases='', description='uncharacterized LOC401396', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8181, CurrentID=0, Status='live', Symbol='BFIS1', Aliases='BFIC, BFIC1', description='benign familial infantile convulsions', other_designations='', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601764, no_hdr0=''),
  401398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401398, CurrentID=0, Status='live', Symbol='POT1-AS1', Aliases='', description='POT1 antisense RNA 1', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124929873, end_position_on_the_genomic_accession=125145234, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  401399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401399, CurrentID=0, Status='live', Symbol='PRRT4', Aliases='', description='proline-rich transmembrane protein 4', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128350325, end_position_on_the_genomic_accession=128362069, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  8184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8184, CurrentID=0, Status='live', Symbol='HBN1', Aliases='HB1, PFHB1, PFHB1B', description='Progressive familial heart block, type I, locus 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604559, no_hdr0=''),
  100271444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271444, CurrentID=0, Status='live', Symbol='RPL21P116', Aliases='RPL21_55_1456', description='ribosomal protein L21 pseudogene 116', other_designations='', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78863212, end_position_on_the_genomic_accession=78863771, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8187, CurrentID=0, Status='live', Symbol='ZNF239', Aliases='HOK-2, MOK2', description='zinc finger protein 239', other_designations='zinc finger protein (C2H2) homologous to mouse MOK-2|zinc finger protein HOK-2|zinc finger protein MOK-2', map_location='10q11.22-q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43556344, end_position_on_the_genomic_accession=43574618, orientation='minus', exon_count=9, OMIM=601069, no_hdr0=''),
  8188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8188, CurrentID=0, Status='live', Symbol='NGFG', Aliases='', description='Nerve growth factor, gamma subunit', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=162040, no_hdr0=''),
  8189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8189, CurrentID=0, Status='live', Symbol='SYMPK', Aliases='SPK, SYM', description='symplekin', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45815442, end_position_on_the_genomic_accession=45863290, orientation='minus', exon_count=29, OMIM=602388, no_hdr0=''),
  8190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8190, CurrentID=0, Status='live', Symbol='MIA', Aliases='CD-RAP', description='melanoma inhibitory activity', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40775177, end_position_on_the_genomic_accession=40777493, orientation='plus', exon_count=5, OMIM=601340, no_hdr0=''),
  100271445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271445, CurrentID=0, Status='live', Symbol='RPL21P121', Aliases='RPL21_56_1523', description='ribosomal protein L21 pseudogene 121', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18313145, end_position_on_the_genomic_accession=18313531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8192, CurrentID=0, Status='live', Symbol='CLPP', Aliases='DFNB81, PRLTS3', description='caseinolytic mitochondrial matrix peptidase proteolytic subunit', other_designations='ATP-dependent protease ClpAP, proteolytic subunit, human|ClpP caseinolytic peptidase ATP-dependent, proteolytic subunit|ClpP caseinolytic peptidase, ATP-dependent, proteolytic subunit homolog|ClpP caseinolytic protease, ATP-dependent, proteolytic subunit homolog|endopeptidase Clp|putative ATP-dependent Clp protease proteolytic subunit, mitochondrial', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6361452, end_position_on_the_genomic_accession=6368904, orientation='plus', exon_count=6, OMIM=601119, no_hdr0=''),
  401409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401409, CurrentID=0, Status='live', Symbol='RAB19', Aliases='RAB19B', description='RAB19, member RAS oncogene family', other_designations='GTP-binding protein RAB19B', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140403822, end_position_on_the_genomic_accession=140426878, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  401410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401410, CurrentID=0, Status='live', Symbol='LOC401410', Aliases='', description='uncharacterized LOC401410', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8195, CurrentID=0, Status='live', Symbol='MKKS', Aliases='BBS6, HMCS, KMS, MKS', description='McKusick-Kaufman syndrome', other_designations='Bardet-Biedl syndrome 6 protein', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10404780, end_position_on_the_genomic_accession=10434239, orientation='minus', exon_count=7, OMIM=604896, no_hdr0=''),
  8197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8197, CurrentID=0, Status='live', Symbol='CHED1', Aliases='CHED, CHED2', description='corneal endothelial dystrophy 1 (autosomal dominant)', other_designations='', map_location='20p11.2-q11.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=121700, no_hdr0=''),
  8200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8200, CurrentID=0, Status='live', Symbol='GDF5', Aliases='BDA1C, BMP-14, BMP14, CDMP1, LAP-4, LAP4, OS5, SYM1B, SYNS2', description='growth differentiation factor 5', other_designations='CDMP-1|LPS-associated protein 4|bone morphogenetic protein 14|cartilage-derived morphogenetic protein-1|lipopolysaccharide-associated protein 4|radotermin', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35433347, end_position_on_the_genomic_accession=35441387, orientation='minus', exon_count=4, OMIM=601146, no_hdr0=''),
  8201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8201, CurrentID=0, Status='live', Symbol='MLRL', Aliases='MLRG, MTS', description='Myeloid leukemia-related gene (myeloid tumor suppressor)', other_designations='', map_location='20q11.2-q12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601308, no_hdr0=''),
  8202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8202, CurrentID=0, Status='live', Symbol='NCOA3', Aliases='ACTR, AIB-1, AIB1, CAGH16, CTG26, KAT13B, RAC3, SRC-3, SRC3, TNRC14, TNRC16, TRAM-1, bHLHe42, pCIP', description='nuclear receptor coactivator 3', other_designations='CBP-interacting protein|amplified in breast cancer 1 protein|class E basic helix-loop-helix protein 42|receptor-associated coactivator 3|steroid receptor coactivator protein 3|thyroid hormone receptor activator molecule 1', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47501857, end_position_on_the_genomic_accession=47656877, orientation='plus', exon_count=23, OMIM=601937, no_hdr0=''),
  8204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8204, CurrentID=0, Status='live', Symbol='NRIP1', Aliases='RIP140', description='nuclear receptor interacting protein 1', other_designations='nuclear factor RIP140|receptor interacting protein 140|receptor-interacting protein 140', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14961235, end_position_on_the_genomic_accession=15065903, orientation='minus', exon_count=9, OMIM=602490, no_hdr0=''),
  8205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8205, CurrentID=0, Status='live', Symbol='TAM', Aliases='MST', description='Myeloproliferative syndrome, transient (transient abnormal', other_designations='Myeloproliferative syndrome, transient (transient abnormal myelopoiesis)', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=159595, no_hdr0=''),
  8208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8208, CurrentID=0, Status='live', Symbol='CHAF1B', Aliases='CAF-1, CAF-IP60, CAF1, CAF1A, CAF1P60, MPHOSPH7, MPP7', description='chromatin assembly factor 1 subunit B', other_designations='CAF-1 subunit B|CAF-I 60 kDa subunit|CAF-I p60|M-phase phosphoprotein 7|chromatin assembly factor 1, subunit B (p60)|chromatin assembly factor I p60 subunit|human chromatin assembly factor-I p60 subunit', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36377057, end_position_on_the_genomic_accession=36416827, orientation='plus', exon_count=19, OMIM=601245, no_hdr0=''),
  8209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8209, CurrentID=0, Status='live', Symbol='C21orf33', Aliases='ES1, GT335, HES1, KNPH, KNPI', description='chromosome 21 open reading frame 33', other_designations='Keio novel protein I|human HES1 protein, homolog to E.coli and zebrafish ES1 protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44133612, end_position_on_the_genomic_accession=44145723, orientation='plus', exon_count=9, OMIM=601659, no_hdr0=''),
  401427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401427, CurrentID=0, Status='live', Symbol='OR2A7', Aliases='HSDJ0798C17, OR2A21', description='olfactory receptor family 2 subfamily A member 7', other_designations='olfactory receptor OR7-18|olfactory receptor OR7-20|olfactory receptor, family 2, subfamily A, member 21|olfactory receptor, family 2, subfamily A, member 7', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144258696, end_position_on_the_genomic_accession=144259628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8212, CurrentID=0, Status='live', Symbol='AABT', Aliases='', description='Beta-amino acids, renal transport of', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109660, no_hdr0=''),
  139285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139285, CurrentID=0, Status='live', Symbol='AMER1', Aliases='FAM123B, OSCS, WTX', description='APC membrane recruitment protein 1', other_designations='RP11-403E24.2|Wilms tumor gene on the X chromosome protein|Wilms tumor on the X|adenomatous polyposis coli membrane recruitment 1|family with sequence similarity 123B|protein FAM123B', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64185117, end_position_on_the_genomic_accession=64206398, orientation='minus', exon_count=3, OMIM=300647, no_hdr0=''),
  8214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8214, CurrentID=0, Status='live', Symbol='DGCR6', Aliases='', description='DiGeorge syndrome critical region gene 6', other_designations='DiGeorge syndrome critical region protein 6', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18906223, end_position_on_the_genomic_accession=18912088, orientation='plus', exon_count=6, OMIM=601279, no_hdr0=''),
  8215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8215, CurrentID=0, Status='live', Symbol='DVL1P1', Aliases='DSH, DVL, DVL-22, DVL1, DVL1L1, DVL4', description='dishevelled segment polarity protein 1 pseudogene 1', other_designations='dishevelled 1-like 1 (homologous to Drosophila dsh)|dishevelled 4, Drosophila dsh homolog|dishevelled, dsh homolog 1 pseudogene 1|dishevelled, dsh homolog 1-like 1', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19253121, end_position_on_the_genomic_accession=19253469, orientation='minus', exon_count=0, OMIM=601225, no_hdr0=''),
  8216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8216, CurrentID=0, Status='live', Symbol='LZTR1', Aliases='BTBD29, LZTR-1, SWNTS2', description='leucine-zipper-like transcription regulator 1', other_designations='', map_location='22q11.1-q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20982269, end_position_on_the_genomic_accession=20999037, orientation='plus', exon_count=21, OMIM=600574, no_hdr0=''),
  8217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8217, CurrentID=0, Status='live', Symbol='IGLJ', Aliases='IGLJ@', description='immunoglobulin lambda joining cluster', other_designations='J region of lambda light chains|immunoglobulin lambda joining group', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147230, no_hdr0=''),
  8218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8218, CurrentID=0, Status='live', Symbol='CLTCL1', Aliases='CHC22, CLH22, CLTCL, CLTD', description='clathrin, heavy chain-like 1', other_designations='CLH-22|Clathrin, heavy polypeptide D|clathrin heavy chain on chromosome 22|clathrin, heavy polypeptide-like 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19179473, end_position_on_the_genomic_accession=19291725, orientation='minus', exon_count=36, OMIM=601273, no_hdr0=''),
  285359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285359, CurrentID=0, Status='live', Symbol='PDCL3P4', Aliases='', description='phosducin-like 3 pseudogene 4', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101712434, end_position_on_the_genomic_accession=101713416, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8220, CurrentID=0, Status='live', Symbol='DGCR14', Aliases='DGCR13, DGS-H, DGS-I, DGSH, DGSI, ES2, Es2el', description='DiGeorge syndrome critical region gene 14', other_designations='DiGeorge syndrome critical region gene 13|DiGeorge syndrome critical region gene DGSI|DiGeorge syndrome gene H|DiGeorge syndrome gene I|Protein DGCR13|diGeorge syndrome protein H', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19130279, end_position_on_the_genomic_accession=19144726, orientation='minus', exon_count=11, OMIM=601755, no_hdr0=''),
  8221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8221, CurrentID=0, Status='live', Symbol='OGS2', Aliases='BBBG2, GBBB2', description='Opitz G syndrome, type II', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145410, no_hdr0=''),
  105372887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372887, CurrentID=0, Status='live', Symbol='LOC105372887', Aliases='', description='uncharacterized LOC105372887', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208094566, end_position_on_the_genomic_accession=208110988, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  8224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8224, CurrentID=0, Status='live', Symbol='SYN3', Aliases='', description='synapsin III', other_designations='cN28H9.2 (synapsin III)', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32508022, end_position_on_the_genomic_accession=33058391, orientation='minus', exon_count=21, OMIM=602705, no_hdr0=''),
  8225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8225, CurrentID=0, Status='live', Symbol='GTPBP6', Aliases='PGPL', description='GTP binding protein 6 (putative)', other_designations='', map_location='Xp22.33; Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=303357, end_position_on_the_genomic_accession=318819, orientation='minus', exon_count=13, OMIM=300124, no_hdr0=''),
  8226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8226, CurrentID=0, Status='live', Symbol='PUDP', Aliases='DXF68S1E, FAM16AX, GS1, HDHD1, HDHD1A', description="pseudouridine 5'-phosphatase", other_designations="5'-PsiMPase|family with sequence similarity 16, member A, X-linked|haloacid dehalogenase-like hydrolase domain containing 1|haloacid dehalogenase-like hydrolase domain containing 1A|haloacid dehalogenase-like hydrolase domain-containing protein 1|haloacid dehalogenase-like hydrolase domain-containing protein 1A|pseudouridine-5'-monophosphatase", map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=7048920, end_position_on_the_genomic_accession=7148190, orientation='minus', exon_count=6, OMIM=306480, no_hdr0=''),
  8227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8227, CurrentID=0, Status='live', Symbol='AKAP17A', Aliases='721P, AKAP-17A, CCDC133, CXYorf3, DXYS155E, PRKA17A, SFRS17A, XE7, XE7Y', description='A-kinase anchoring protein 17A', other_designations='A kinase (PRKA) anchor protein 17A|B-lymphocyte surface antigen|protein kinase A-anchoring protein 17A|pseudoautosomal gene XE7|splicing factor, arginine/serine-rich 17A', map_location='Xp22.33 and Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1591593, end_position_on_the_genomic_accession=1602520, orientation='plus', exon_count=6, OMIM='312095,465000', no_hdr0=''),
  8228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8228, CurrentID=0, Status='live', Symbol='PNPLA4', Aliases='DXS1283E, GS2, iPLA2eta', description='patatin like phospholipase domain containing 4', other_designations='IPLA2 eta|calcium independent phospholipases A2 eta|patatin-like phospholipase domain containing 4', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=7898763, end_position_on_the_genomic_accession=7928587, orientation='minus', exon_count=8, OMIM=300102, no_hdr0=''),
  401445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401445, CurrentID=0, Status='live', Symbol='LOC401445', Aliases='', description='uncharacterized LOC401445', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401447, CurrentID=0, Status='live', Symbol='USP17L1', Aliases='USP17L1P', description='ubiquitin specific peptidase 17-like family member 1', other_designations='deubiquitinating enzyme 17-like protein 1|putative ubiquitin carboxyl-terminal hydrolase 17-like protein 1|ubiquitin thioesterase 17-like protein 1|ubiquitin-specific-processing protease 17-like protein 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7332387, end_position_on_the_genomic_accession=7333979, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8233, CurrentID=0, Status='live', Symbol='ZRSR2', Aliases='U2AF1-RS2, U2AF1L2, U2AF1RS2, URP, ZC3H22', description='zinc finger (CCCH type), RNA binding motif and serine/arginine rich 2', other_designations='U2 small nuclear ribonucleoprotein auxiliary factor, small subunit 2|U2(RNU2) small nuclear RNA auxiliary factor 1-like 2|U2AF35-related protein|renal carcinoma antigen NY-REN-20|zinc finger (CCCH type), RNA-binding motif and serine/arginine rich 2', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15790269, end_position_on_the_genomic_accession=15823260, orientation='plus', exon_count=12, OMIM=300028, no_hdr0=''),
  401450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401450, CurrentID=0, Status='live', Symbol='OR7E96P', Aliases='', description='olfactory receptor family 7 subfamily E member 96 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 96 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8039979, end_position_on_the_genomic_accession=8040947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8237, CurrentID=0, Status='live', Symbol='USP11', Aliases='UHX1', description='ubiquitin specific peptidase 11', other_designations='deubiquitinating enzyme 11|ubiquitin carboxyl-terminal hydrolase, X-linked|ubiquitin thioesterase 11|ubiquitin thiolesterase 11|ubiquitin-specific processing protease 11|ubiquitin-specific-processing protease 11', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47232915, end_position_on_the_genomic_accession=47248328, orientation='plus', exon_count=23, OMIM=300050, no_hdr0=''),
  8239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8239, CurrentID=0, Status='live', Symbol='USP9X', Aliases='DFFRX, FAF, FAM, MRX99', description='ubiquitin specific peptidase 9, X-linked', other_designations='Drosophila fat facets related, X-linked|deubiquitinating enzyme FAF-X|fat facets in mammals|fat facets protein related, X-linked|fat facets protein-related, X-linked|hFAM|ubiquitin specific protease 9, X chromosome (fat facets-like Drosophila)|ubiquitin thioesterase FAF-X|ubiquitin thiolesterase FAF-X|ubiquitin-specific processing protease FAF-X|ubiquitin-specific protease 9, X chromosome|ubiquitin-specific-processing protease FAF-X', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41085433, end_position_on_the_genomic_accession=41236579, orientation='plus', exon_count=45, OMIM=300072, no_hdr0=''),
  8241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8241, CurrentID=0, Status='live', Symbol='RBM10', Aliases='DXS8237E, GPATC9, GPATCH9, S1-1, TARPS, ZRANB5', description='RNA binding motif protein 10', other_designations='RNA-binding protein S1-1|g patch domain-containing protein 9', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47145218, end_position_on_the_genomic_accession=47186815, orientation='plus', exon_count=25, OMIM=300080, no_hdr0=''),
  8242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8242, CurrentID=0, Status='live', Symbol='KDM5C', Aliases='DXS1272E, JARID1C, MRX13, MRXJ, MRXSCJ, MRXSJ, SMCX, XE169', description='lysine (K)-specific demethylase 5C', other_designations='JmjC domain-containing protein SMCX|Jumonji, AT rich interactive domain 1C (RBP2-like)|Jumonji/ARID domain-containing protein 1C|Smcx homolog, X chromosome|Smcy homolog, X-linked|histone demethylase JARID1C', map_location='Xp11.22-p11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53158726, end_position_on_the_genomic_accession=53225422, orientation='minus', exon_count=32, OMIM=314690, no_hdr0=''),
  8243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8243, CurrentID=0, Status='live', Symbol='SMC1A', Aliases='CDLS2, DXS423E, SB1.8, SMC1, SMC1L1, SMC1alpha, SMCB', description='structural maintenance of chromosomes 1A', other_designations='SMC protein 1A|SMC-1-alpha|SMC1 (structural maintenance of chromosomes 1, yeast)-like 1|segregation of mitotic chromosomes 1', map_location='Xp11.22-p11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53374149, end_position_on_the_genomic_accession=53422728, orientation='minus', exon_count=26, OMIM=300040, no_hdr0=''),
  8245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8245, CurrentID=0, Status='live', Symbol='IDDMX', Aliases='', description='Diabetes mellitus, insulin-dependent, X-linked, susceptibility to', other_designations='', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300136, no_hdr0=''),
  8246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8246, CurrentID=0, Status='live', Symbol='MRXS12', Aliases='', description='Mental retardation, X-linked, syndromic 12', other_designations='', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309545, no_hdr0=''),
  401463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401463, CurrentID=0, Status='live', Symbol='LOC401463', Aliases='', description='uncharacterized LOC401463', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64574308, end_position_on_the_genomic_accession=64577263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8248, CurrentID=0, Status='live', Symbol='DXS435E', Aliases='A11', description='A-11 gene', other_designations='', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300010, no_hdr0=''),
  8249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8249, CurrentID=0, Status='live', Symbol='MGR2', Aliases='MFTS', description='Migraine, familial typical, susceptibility to', other_designations='', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300125, no_hdr0=''),
  139322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139322, CurrentID=0, Status='live', Symbol='APOOL', Aliases='CXorf33, FAM121A, Mic27, UNQ8193', description='apolipoprotein O-like', other_designations='AAIR8193|family with sequence similarity 121A', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85003892, end_position_on_the_genomic_accession=85093317, orientation='plus', exon_count=12, OMIM=300955, no_hdr0=''),
  8251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8251, CurrentID=0, Status='live', Symbol='HNRNPDP1', Aliases='AUF1B, AUF1P, HNRNPDP, HNRPDP', description='heterogeneous nuclear ribonucleoprotein D pseudogene 1', other_designations='AU-rich element RNA-binding protein 1, type B|heterogeneous nuclear ribonucleoprotein D (AU-rich element RNA binding protein 1, 37kDa) pseudogene 1', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64043978, end_position_on_the_genomic_accession=64045649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139324, CurrentID=0, Status='live', Symbol='HDX', Aliases='CXorf43, D030011N01Rik', description='highly divergent homeobox', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84317874, end_position_on_the_genomic_accession=84502584, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  401470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401470, CurrentID=0, Status='live', Symbol='RPS26P10', Aliases='RPS26_16_925', description='ribosomal protein S26 pseudogene 10', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92144007, end_position_on_the_genomic_accession=92144459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8255, CurrentID=0, Status='live', Symbol='BZX', Aliases='BDCS', description='Bazex syndrome', other_designations='', map_location='Xq24-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301845, no_hdr0=''),
  401472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401472, CurrentID=0, Status='live', Symbol='FLJ45248', Aliases='', description='FLJ45248 protein', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=103819050, end_position_on_the_genomic_accession=103819818, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271456, CurrentID=0, Status='live', Symbol='RPL23AP27', Aliases='RPL23A_10_252', description='ribosomal protein L23a pseudogene 27', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104678232, end_position_on_the_genomic_accession=104678675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401474, CurrentID=0, Status='live', Symbol='SAMD12', Aliases='', description='sterile alpha motif domain containing 12', other_designations='SAM domain-containing protein 12', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118131828, end_position_on_the_genomic_accession=118621995, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  8259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8259, CurrentID=0, Status='live', Symbol='INDX', Aliases='', description='Immunoneurologic syndrome X-linked, of Wood, Black, and Norbury', other_designations='', map_location='Xq26-qter', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300076, no_hdr0=''),
  8260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8260, CurrentID=0, Status='live', Symbol='NAA10', Aliases='ARD1, ARD1A, ARD1P, DXS707, MCOPS1, NATD, OGDNS, TE2', description='N(alpha)-acetyltransferase 10, NatA catalytic subunit', other_designations='ARD1 homolog A, N-acetyltransferase|N-acetyltransferase ARD1, human homolog of|N-alpha-acetyltransferase 10, NatA catalytic subunit|N-terminal acetyltransferase complex ARD1 subunit homolog A|natA catalytic subunit Naa10', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153929827, end_position_on_the_genomic_accession=153935154, orientation='minus', exon_count=8, OMIM=300013, no_hdr0=''),
  401478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401478, CurrentID=0, Status='live', Symbol='LOC401478', Aliases='', description='uncharacterized LOC401478', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=137812896, end_position_on_the_genomic_accession=138088186, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  8263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8263, CurrentID=0, Status='live', Symbol='F8A1', Aliases='DXS522E, F8A, HAP40', description='coagulation factor VIII-associated 1', other_designations='coagulation factor VIII-associated (intronic transcript) 1|cpG island protein|factor VIII associated protein|huntingtin-associated protein 40', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154886360, end_position_on_the_genomic_accession=154888061, orientation='plus', exon_count=1, OMIM=305423, no_hdr0=''),
  401480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401480, CurrentID=0, Status='live', Symbol='LOC401480', Aliases='', description='uncharacterized LOC401480', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141389870, end_position_on_the_genomic_accession=141392590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8266, CurrentID=0, Status='live', Symbol='UBL4A', Aliases='DX254E, DXS254E, G6PD, GDX, GET5, MDY2, TMA24, UBL4', description='ubiquitin like 4A', other_designations='ubiquitin-like 4|ubiquitin-like 4A|ubiquitin-like protein GDX', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154483717, end_position_on_the_genomic_accession=154486670, orientation='minus', exon_count=4, OMIM=312070, no_hdr0=''),
  23223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23223, CurrentID=0, Status='live', Symbol='RRP12', Aliases='KIAA0690', description='ribosomal RNA processing 12 homolog', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97356701, end_position_on_the_genomic_accession=97401370, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  8268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8268, CurrentID=0, Status='live', Symbol='IPOX', Aliases='CIIP, CIIPX', description='Intestinal pseudoobstruction, neuronal, primary idiopathic', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300048, no_hdr0=''),
  139341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139341, CurrentID=0, Status='live', Symbol='FUNDC1', Aliases='', description='FUN14 domain containing 1', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44523639, end_position_on_the_genomic_accession=44542975, orientation='minus', exon_count=5, OMIM=300871, no_hdr0=''),
  8270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8270, CurrentID=0, Status='live', Symbol='LAGE3', Aliases='CVG5, DXS9879E, DXS9951E, ESO3, ITBA2', description='L antigen family member 3', other_designations='L antigen family, member 3|protein ESO-3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154477769, end_position_on_the_genomic_accession=154479257, orientation='minus', exon_count=3, OMIM=300060, no_hdr0=''),
  8273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8273, CurrentID=0, Status='live', Symbol='SLC10A3', Aliases='DXS253E, P3', description='solute carrier family 10 member 3', other_designations='Protein P3|solute carrier family 10 (sodium/bile acid cotransporter family), member 3|solute carrier family 10, member 3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154487306, end_position_on_the_genomic_accession=154490690, orientation='minus', exon_count=7, OMIM=312090, no_hdr0=''),
  401490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401490, CurrentID=0, Status='live', Symbol='LOC401490', Aliases='', description='uncharacterized LOC401490', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401491, CurrentID=0, Status='live', Symbol='VLDLR-AS1', Aliases='linc-VLDLR, lincRNA-VLDLR', description='VLDLR antisense RNA 1', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2535652, end_position_on_the_genomic_accession=2622373, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  401492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401492, CurrentID=0, Status='live', Symbol='LINC01235', Aliases='', description='long intergenic non-protein coding RNA 1235', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13406380, end_position_on_the_genomic_accession=13431329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8277, CurrentID=0, Status='live', Symbol='TKTL1', Aliases='TKR, TKT2', description='transketolase-like 1', other_designations='TK 2|transketolase-2|transketolase-related protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154295674, end_position_on_the_genomic_accession=154330363, orientation='plus', exon_count=13, OMIM=300044, no_hdr0=''),
  401494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401494, CurrentID=0, Status='live', Symbol='HACD4', Aliases='PTPLAD2', description='3-hydroxyacyl-CoA dehydratase 4', other_designations='protein tyrosine phosphatase-like A domain containing 2|protein tyrosine phosphatase-like protein PTPLAD2|protein-tyrosine phosphatase-like A domain-containing protein 2|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 4', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21003621, end_position_on_the_genomic_accession=21031670, orientation='minus', exon_count=12, OMIM=615941, no_hdr0=''),
  100271460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271460, CurrentID=0, Status='live', Symbol='RPL23AP31', Aliases='RPL23A_13_326', description='ribosomal protein L23a pseudogene 31', other_designations='', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218841605, end_position_on_the_genomic_accession=218842065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401498, CurrentID=0, Status='live', Symbol='TMEM215', Aliases='', description='transmembrane protein 215', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32783499, end_position_on_the_genomic_accession=32789201, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8284, CurrentID=0, Status='live', Symbol='KDM5D', Aliases='HY, HYA, JARID1D, SMCY', description='lysine (K)-specific demethylase 5D', other_designations='H-Y|Jumonji, AT rich interactive domain 1D (RBP2-like)|SMC homolog, Y chromosome|Smcy homolog, Y-linked|histocompatibility Y antigen|histone demethylase JARID1D|jumonji/ARID domain-containing protein 1D|protein SmcY|selected mouse cDNA on Y, human homolog of', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19705415, end_position_on_the_genomic_accession=19744939, orientation='minus', exon_count=29, OMIM=426000, no_hdr0=''),
  8287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8287, CurrentID=0, Status='live', Symbol='USP9Y', Aliases='DFFRY, SPGFY2', description='ubiquitin specific peptidase 9, Y-linked', other_designations='deubiquitinating enzyme FAF-Y|fat facets protein-related, Y-linked|ubiquitin specific peptidase 9, Y-linked (fat facets-like, Drosophila)|ubiquitin thioesterase FAF-Y|ubiquitin thiolesterase FAF-Y|ubiquitin-specific processing protease FAF-Y|ubiquitin-specific protease 9, Y chromosome|ubiquitin-specific-processing protease FAF-Y', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12701231, end_position_on_the_genomic_accession=12860844, orientation='plus', exon_count=49, OMIM=400005, no_hdr0=''),
  8288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8288, CurrentID=0, Status='live', Symbol='EPX', Aliases='EPO, EPP-PEN, EPXD, EPX', description='eosinophil peroxidase', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58192728, end_position_on_the_genomic_accession=58205174, orientation='plus', exon_count=13, OMIM=131399, no_hdr0=''),
  8289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8289, CurrentID=0, Status='live', Symbol='ARID1A', Aliases='B120, BAF250, BAF250a, BM029, C1orf4, ELD, MRD14, OSA1, P270, SMARCF1, hELD, hOSA1', description='AT-rich interaction domain 1A', other_designations='ARID domain-containing protein 1A|AT rich interactive domain 1A (SWI-like)|BRG1-associated factor 250a|OSA1 nuclear protein|SWI-like protein|SWI/SNF complex protein p270|SWI/SNF-related, matrix-associated, actin-dependent regulator of chromatin subfamily F member 1|brain protein 120|chromatin remodeling factor p250|osa homolog 1', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26696031, end_position_on_the_genomic_accession=26782110, orientation='plus', exon_count=20, OMIM=603024, no_hdr0=''),
  8290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8290, CurrentID=0, Status='live', Symbol='HIST3H3', Aliases='H3.4, H3/g, H3FT, H3t', description='histone cluster 3, H3', other_designations='H3 histone family, member T|H3/t|histone 3, H3', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228424845, end_position_on_the_genomic_accession=228425325, orientation='minus', exon_count=1, OMIM=602820, no_hdr0=''),
  8291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8291, CurrentID=0, Status='live', Symbol='DYSF', Aliases='FER1L1, LGMD2B, MMD1', description='dysferlin', other_designations='dystrophy-associated fer-1-like 1|fer-1-like family member 1|fer-1-like protein 1|limb girdle muscular dystrophy 2B (autosomal recessive)', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71453119, end_position_on_the_genomic_accession=71686763, orientation='plus', exon_count=58, OMIM=603009, no_hdr0=''),
  8292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8292, CurrentID=0, Status='live', Symbol='COLQ', Aliases='CMS5, EAD', description='collagen-like tail subunit (single strand of homotrimer) of asymmetric acetylcholinesterase', other_designations='AChE Q subunit|acetylcholinesterase-associated collagen|collagenic tail of endplate acetylcholinesterase|single strand of homotrimeric collagen-like tail subunit of asymmetric acetylcholinesterase', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15450133, end_position_on_the_genomic_accession=15521751, orientation='minus', exon_count=18, OMIM=603033, no_hdr0=''),
  8293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8293, CurrentID=0, Status='live', Symbol='SERF1A', Aliases='4F5, FAM2A, H4F5, SERF1, SMAM1', description='small EDRK-rich factor 1A (telomeric)', other_designations='SMA modifier 1|protein 4F5|spinal muscular atrophy-related gene H4F5', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70900663, end_position_on_the_genomic_accession=70918530, orientation='plus', exon_count=4, OMIM=603011, no_hdr0=''),
  8294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8294, CurrentID=0, Status='live', Symbol='HIST1H4I', Aliases='H4/m, H4FM, H4M', description='histone cluster 1, H4i', other_designations='H4 histone family, member M|Histone 4 family, member M|histone 1, H4i|histone family member', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27139309, end_position_on_the_genomic_accession=27139678, orientation='plus', exon_count=1, OMIM=602833, no_hdr0=''),
  8295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8295, CurrentID=0, Status='live', Symbol='TRRAP', Aliases='PAF350/400, PAF400, STAF40, TR-AP, Tra1', description='transformation/transcription domain-associated protein', other_designations='350/400 kDa PCAF-associated factor|tra1 homolog', map_location='7q21.2-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98878490, end_position_on_the_genomic_accession=99013243, orientation='plus', exon_count=72, OMIM=603015, no_hdr0=''),
  8301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8301, CurrentID=0, Status='live', Symbol='PICALM', Aliases='CALM, CLTH, LAP', description='phosphatidylinositol binding clathrin assembly protein', other_designations='clathrin assembly lymphoid myeloid leukemia protein', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85957171, end_position_on_the_genomic_accession=86069881, orientation='minus', exon_count=23, OMIM=603025, no_hdr0=''),
  8302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8302, CurrentID=0, Status='live', Symbol='KLRC4', Aliases='NKG2-F, NKG2F', description='killer cell lectin-like receptor subfamily C, member 4', other_designations='NK cell receptor F|NKG2-F-activating NK receptor|natual killer cell group 2-F', map_location='12p13.2-p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10407384, end_position_on_the_genomic_accession=10409757, orientation='minus', exon_count=4, OMIM=602893, no_hdr0=''),
  8303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8303, CurrentID=0, Status='live', Symbol='SNN', Aliases='', description='stannin', other_designations='AG8_1', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11668433, end_position_on_the_genomic_accession=11679159, orientation='plus', exon_count=2, OMIM=603032, no_hdr0=''),
  100271464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271464, CurrentID=0, Status='live', Symbol='RPL23AP51', Aliases='RPL23A_18_782', description='ribosomal protein L23a pseudogene 51', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7742106, end_position_on_the_genomic_accession=7742562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  139378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139378, CurrentID=0, Status='live', Symbol='ADGRG4', Aliases='GPR112, PGR17, RP1-299I16', description='adhesion G protein-coupled receptor G4', other_designations='G protein-coupled receptor 112|probable G-protein coupled receptor 112', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136300963, end_position_on_the_genomic_accession=136416907, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  100131101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131101, CurrentID=0, Status='live', Symbol='LOC100131101', Aliases='', description='uncharacterized LOC100131101', other_designations='', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=236563528, end_position_on_the_genomic_accession=236564106, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8309, CurrentID=0, Status='live', Symbol='ACOX2', Aliases='BCOX, BRCACOX, BRCOX, THCCox', description='acyl-CoA oxidase 2, branched chain', other_designations='3-alpha,7-alpha,12-alpha-trihydroxy-5-beta-cholestanoyl-CoA 24-hydroxylase|3-alpha,7-alpha,12-alpha-trihydroxy-5-beta-cholestanoyl-CoA oxidase|THCA-CoA oxidase|acyl-Coenzyme A oxidase 2, branched chain|peroxisomal branched chain acyl-CoA oxidase|trihydroxycoprostanoyl-CoA oxidase', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58505136, end_position_on_the_genomic_accession=58537348, orientation='minus', exon_count=16, OMIM=601641, no_hdr0=''),
  8310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8310, CurrentID=0, Status='live', Symbol='ACOX3', Aliases='', description='acyl-CoA oxidase 3, pristanoyl', other_designations='BRCACox|acyl-Coenzyme A oxidase 3, pristanoyl|branched-chain acyl-CoA oxidase|pristanoyl-CoA oxidase', map_location='4p15.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8366263, end_position_on_the_genomic_accession=8440725, orientation='minus', exon_count=22, OMIM=603402, no_hdr0=''),
  100271465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271465, CurrentID=0, Status='live', Symbol='RPL23AP52', Aliases='RPL23A_19_784', description='ribosomal protein L23a pseudogene 52', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11006399, end_position_on_the_genomic_accession=11006852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8312, CurrentID=0, Status='live', Symbol='AXIN1', Aliases='AXIN, PPP1R49', description='axin 1', other_designations='axis inhibition protein 1|axis inhibitor 1|fused, mouse, homolog of|protein phosphatase 1, regulatory subunit 49', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=287440, end_position_on_the_genomic_accession=361497, orientation='minus', exon_count=15, OMIM=603816, no_hdr0=''),
  8313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8313, CurrentID=0, Status='live', Symbol='AXIN2', Aliases='AXIL, ODCRCS', description='axin 2', other_designations='axin-like protein|axis inhibition protein 2|conductin', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65528563, end_position_on_the_genomic_accession=65561622, orientation='minus', exon_count=14, OMIM=604025, no_hdr0=''),
  8314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8314, CurrentID=0, Status='live', Symbol='BAP1', Aliases='HUCEP-13, UCHL2, hucep-6', description='BRCA1 associated protein 1', other_designations='BRCA1 associated protein-1 (ubiquitin carboxy-terminal hydrolase)|cerebral protein 6|cerebral protein-13', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52401004, end_position_on_the_genomic_accession=52410105, orientation='minus', exon_count=17, OMIM=603089, no_hdr0=''),
  8315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8315, CurrentID=0, Status='live', Symbol='BRAP', Aliases='BRAP2, IMP, RNF52', description='BRCA1 associated protein', other_designations='RING finger protein 52|galectin-2-binding protein|impedes mitogenic signal propagation|renal carcinoma antigen NY-REN-63', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111642146, end_position_on_the_genomic_accession=111685996, orientation='minus', exon_count=12, OMIM=604986, no_hdr0=''),
  8317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8317, CurrentID=0, Status='live', Symbol='CDC7', Aliases='CDC7L1, HsCDC7, Hsk1, huCDC7', description='cell division cycle 7', other_designations='CDC7 (cell division cycle 7, S. cerevisiae, homolog)-like 1|CDC7-related kinase|cell division cycle 7 homolog|cell division cycle 7-like protein 1', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91500814, end_position_on_the_genomic_accession=91525764, orientation='plus', exon_count=14, OMIM=603311, no_hdr0=''),
  8318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8318, CurrentID=0, Status='live', Symbol='CDC45', Aliases='CDC45LL2, PORC-PI-1, CDC45', description='cell division cycle 45', other_designations='CDC45 cell division cycle 45 homolog|CDC45-related protein|cell division cycle 45 homolog|cell division cycle 45-like 2|human CDC45', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19479385, end_position_on_the_genomic_accession=19520612, orientation='plus', exon_count=21, OMIM=603465, no_hdr0=''),
  401535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401535, CurrentID=0, Status='live', Symbol='C9orf170', Aliases='', description='chromosome 9 open reading frame 170', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87148644, end_position_on_the_genomic_accession=87159726, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8320, CurrentID=0, Status='live', Symbol='EOMES', Aliases='TBR2', description='eomesodermin', other_designations='T-box brain protein 2', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27715949, end_position_on_the_genomic_accession=27722715, orientation='minus', exon_count=6, OMIM=604615, no_hdr0=''),
  8321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8321, CurrentID=0, Status='live', Symbol='FZD1', Aliases='', description='frizzled class receptor 1', other_designations='Wnt receptor|frizzled 1, seven transmembrane spanning receptor|frizzled family receptor 1|frizzled homolog 1|frizzled, Drosophila, homolog of, 1|fz-1|fzE1|hFz1', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91264468, end_position_on_the_genomic_accession=91268817, orientation='plus', exon_count=1, OMIM=603408, no_hdr0=''),
  8322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8322, CurrentID=0, Status='live', Symbol='FZD4', Aliases='CD344, EVR1, FEVRS, Fz-4, Fz4, FzE4, GPCR, hFz4, FZD4', description='frizzled class receptor 4', other_designations='WNT receptor frizzled-4|frizzled 4, seven transmembrane spanning receptor|frizzled family receptor 4|frizzled homolog 4', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86945675, end_position_on_the_genomic_accession=86955398, orientation='minus', exon_count=2, OMIM=604579, no_hdr0=''),
  8323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8323, CurrentID=0, Status='live', Symbol='FZD6', Aliases='FZ-6, FZ6, HFZ6, NDNC10', description='frizzled class receptor 6', other_designations='frizzled 6, seven transmembrane spanning receptor|frizzled family receptor 6|frizzled homolog 6|seven transmembrane helix receptor', map_location='8q22.3-q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103298433, end_position_on_the_genomic_accession=103332866, orientation='plus', exon_count=8, OMIM=603409, no_hdr0=''),
  8324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8324, CurrentID=0, Status='live', Symbol='FZD7', Aliases='FzE3', description='frizzled class receptor 7', other_designations='Frizzled, drosophila, homolog of, 7|frizzled 7, seven transmembrane spanning receptor|frizzled family receptor 7|frizzled homolog 7|fz-7|hFz7', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202034587, end_position_on_the_genomic_accession=202038437, orientation='plus', exon_count=1, OMIM=603410, no_hdr0=''),
  8325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8325, CurrentID=0, Status='live', Symbol='FZD8', Aliases='FZ-8, hFZ8', description='frizzled class receptor 8', other_designations='frizzled 8, seven transmembrane spanning receptor|frizzled family receptor 8|frizzled homolog 8', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35638249, end_position_on_the_genomic_accession=35641434, orientation='minus', exon_count=1, OMIM=606146, no_hdr0=''),
  8326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8326, CurrentID=0, Status='live', Symbol='FZD9', Aliases='CD349, FZD3', description='frizzled class receptor 9', other_designations='frizzled 9, seven transmembrane spanning receptor|frizzled family receptor 9|frizzled homolog 9|fz-9|fzE6|hFz9', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73433779, end_position_on_the_genomic_accession=73436120, orientation='plus', exon_count=1, OMIM=601766, no_hdr0=''),
  8327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8327, CurrentID=0, Status='live', Symbol='GABPAP', Aliases='E4TF1, E4TF1B, GABPB1', description='GA binding protein transcription factor, alpha subunit pseudogene', other_designations='GA binding protein transcription factor, alpha subunit 60kDa pseudogene|GA binding protein transcription factor, beta subunit 1, 53kDa', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63893249, end_position_on_the_genomic_accession=63897910, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8328, CurrentID=0, Status='live', Symbol='GFI1B', Aliases='BDPLT17, ZNF163B', description='growth factor independent 1B transcription repressor', other_designations='growth factor independent 1B (potential regulator of CDKN1A, translocated in CML)|potential regulator of CDKN1A translocated in CML', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132945545, end_position_on_the_genomic_accession=132993434, orientation='plus', exon_count=15, OMIM=604383, no_hdr0=''),
  8329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8329, CurrentID=0, Status='live', Symbol='HIST1H2AI', Aliases='H2A/c, H2AFC', description='histone cluster 1, H2ai', other_designations='H2A histone family, member C|H2A.1|histone 1, H2ai', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27808199, end_position_on_the_genomic_accession=27808667, orientation='plus', exon_count=1, OMIM=602787, no_hdr0=''),
  8330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8330, CurrentID=0, Status='live', Symbol='HIST1H2AK', Aliases='H2A/d, H2AFD', description='histone cluster 1, H2ak', other_designations='H2A histone family, member D|H2A.1|histone 1, H2ak|histone H2A type 1-J', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27837880, end_position_on_the_genomic_accession=27838339, orientation='minus', exon_count=1, OMIM=602788, no_hdr0=''),
  8331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8331, CurrentID=0, Status='live', Symbol='HIST1H2AJ', Aliases='H2A/E, H2AFE, dJ160A22.4', description='histone cluster 1, H2aj', other_designations='H2A histone family, member E|histone 1, H2aj', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27814302, end_position_on_the_genomic_accession=27814740, orientation='minus', exon_count=1, OMIM=602791, no_hdr0=''),
  8332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8332, CurrentID=0, Status='live', Symbol='HIST1H2AL', Aliases='H2A.i, H2A/i, H2AFI, dJ193B12.9', description='histone cluster 1, H2al', other_designations='H2A histone family, member I|H2A.1|histone 1, H2al', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27865329, end_position_on_the_genomic_accession=27865798, orientation='plus', exon_count=1, OMIM=602793, no_hdr0=''),
  104472717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472717, CurrentID=0, Status='live', Symbol='LINC01224', Aliases='', description='long intergenic non-protein coding RNA 1224', other_designations='CTB-175P5.4', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23399233, end_position_on_the_genomic_accession=23416074, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  8334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8334, CurrentID=0, Status='live', Symbol='HIST1H2AC', Aliases='H2A/l, H2AFL, dJ221C16.4', description='histone cluster 1, H2ac', other_designations='H2A histone family, member L|histone 1, H2ac|histone H2A/l|histone H2AC', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26124145, end_position_on_the_genomic_accession=26124690, orientation='plus', exon_count=1, OMIM=602794, no_hdr0=''),
  8335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8335, CurrentID=0, Status='live', Symbol='HIST1H2AB', Aliases='H2A/m, H2AFM', description='histone cluster 1, H2ab', other_designations='H2A histone family, member M|histone 1, H2ab|histone H2A/m', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26033092, end_position_on_the_genomic_accession=26033568, orientation='minus', exon_count=1, OMIM=602795, no_hdr0=''),
  8336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8336, CurrentID=0, Status='live', Symbol='HIST1H2AM', Aliases='H2A.1, H2A/n, H2AFN, dJ193B12.1', description='histone cluster 1, H2am', other_designations='H2A histone family, member N|histone 1, H2am', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27892699, end_position_on_the_genomic_accession=27893185, orientation='minus', exon_count=1, OMIM=602796, no_hdr0=''),
  8337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8337, CurrentID=0, Status='live', Symbol='HIST2H2AA3', Aliases='H2A, H2A.2, H2A/O, H2A/q, H2AFO, H2a-615, HIST2H2AA', description='histone cluster 2, H2aa3', other_designations='H2A histone family, member O|histone 2, H2aa3', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149842218, end_position_on_the_genomic_accession=149842751, orientation='minus', exon_count=1, OMIM=142720, no_hdr0=''),
  8338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8338, CurrentID=0, Status='live', Symbol='HIST2H2AC', Aliases='H2A, H2A-GL101, H2A/q, H2AFQ', description='histone cluster 2, H2ac', other_designations='H2A histone family, member Q|histone 2, H2ac|histone H2A-GL101|histone H2A/q|histone IIa', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149886975, end_position_on_the_genomic_accession=149887411, orientation='plus', exon_count=1, OMIM=602797, no_hdr0=''),
  8339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8339, CurrentID=0, Status='live', Symbol='HIST1H2BG', Aliases='H2B.1A, H2B/a, H2BFA, dJ221C16.8', description='histone cluster 1, H2bg', other_designations='H2B histone family, member A|histone 1, H2bg|histone H2B.1 A|histone H2B.a', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26216200, end_position_on_the_genomic_accession=26216644, orientation='minus', exon_count=1, OMIM=602798, no_hdr0=''),
  8340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8340, CurrentID=0, Status='live', Symbol='HIST1H2BL', Aliases='H2B/c, H2BFC, dJ97D16.4', description='histone cluster 1, H2bl', other_designations='H2B histone family, member C|histone 1, H2bl|histone H2B.c', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27807479, end_position_on_the_genomic_accession=27807931, orientation='minus', exon_count=1, OMIM=602800, no_hdr0=''),
  8341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8341, CurrentID=0, Status='live', Symbol='HIST1H2BN', Aliases='H2B/d, H2BFD', description='histone cluster 1, H2bn', other_designations='H2B histone family, member D|histone 1, H2bn|histone H2B.d', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27838662, end_position_on_the_genomic_accession=27839110, orientation='plus', exon_count=1, OMIM=602801, no_hdr0=''),
  8342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8342, CurrentID=0, Status='live', Symbol='HIST1H2BM', Aliases='H2B/e, H2BFE, dJ160A22.3', description='histone cluster 1, H2bm', other_designations='H2B histone family, member E|histone 1, H2bm|histone H2B.e', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27815044, end_position_on_the_genomic_accession=27815489, orientation='plus', exon_count=1, OMIM=602802, no_hdr0=''),
  8343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8343, CurrentID=0, Status='live', Symbol='HIST1H2BF', Aliases='H2B/g, H2BFG', description='histone cluster 1, H2bf', other_designations='H2B histone family, member G|histone 1, H2bf|histone H2B.1 A|histone H2B.g', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26199559, end_position_on_the_genomic_accession=26199988, orientation='plus', exon_count=1, OMIM=602804, no_hdr0=''),
  8344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8344, CurrentID=0, Status='live', Symbol='HIST1H2BE', Aliases='H2B.h, H2B/h, H2BFH, dJ221C16.8', description='histone cluster 1, H2be', other_designations='H2B histone family, member H|histone 1, H2be|histone H2B.1 A|histone H2B.h', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26183796, end_position_on_the_genomic_accession=26184230, orientation='plus', exon_count=1, OMIM=602805, no_hdr0=''),
  8345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8345, CurrentID=0, Status='live', Symbol='HIST1H2BH', Aliases='H2B/j, H2BFJ', description='histone cluster 1, H2bh', other_designations='H2B histone family, member J|histone 1, H2bh|histone H2B.j', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26251651, end_position_on_the_genomic_accession=26252075, orientation='plus', exon_count=1, OMIM=602806, no_hdr0=''),
  8346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8346, CurrentID=0, Status='live', Symbol='HIST1H2BI', Aliases='H2B/k, H2BFK', description='histone cluster 1, H2bi', other_designations='H2B histone family, member K|histone 1, H2bi|histone H2B.1 A|histone H2B.k', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26272976, end_position_on_the_genomic_accession=26273412, orientation='plus', exon_count=1, OMIM=602807, no_hdr0=''),
  8347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8347, CurrentID=0, Status='live', Symbol='HIST1H2BC', Aliases='H2B.1, H2B/l, H2BFL, dJ221C16.3', description='histone cluster 1, H2bc', other_designations='H2B histone family, member L|histone 1, H2bc|histone H2B.1 A|histone H2B.l', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26123467, end_position_on_the_genomic_accession=26123904, orientation='minus', exon_count=1, OMIM=602847, no_hdr0=''),
  8348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8348, CurrentID=0, Status='live', Symbol='HIST1H2BO', Aliases='H2B.2, H2B/n, H2BFN, dJ193B12.2', description='histone cluster 1, H2bo', other_designations='H2B histone family, member N|histone 1, H2bo|histone H2B.2|histone H2B.n', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27893425, end_position_on_the_genomic_accession=27893891, orientation='plus', exon_count=1, OMIM=602808, no_hdr0=''),
  8349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8349, CurrentID=0, Status='live', Symbol='HIST2H2BE', Aliases='GL105, H2B, H2B.1, H2BFQ, H2BGL105, H2BQ', description='histone cluster 2, H2be', other_designations='H2B histone family, member Q|histone 2, H2be|histone H2B-GL105|histone H2B.q', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149884460, end_position_on_the_genomic_accession=149886682, orientation='minus', exon_count=1, OMIM=601831, no_hdr0=''),
  8350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8350, CurrentID=0, Status='live', Symbol='HIST1H3A', Aliases='H3/A, H3FA', description='histone cluster 1, H3a', other_designations='H3 histone family, member A|histone 1, H3a|histone H3/a', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26020490, end_position_on_the_genomic_accession=26020958, orientation='plus', exon_count=1, OMIM=602810, no_hdr0=''),
  8351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8351, CurrentID=0, Status='live', Symbol='HIST1H3D', Aliases='H3/b, H3FB', description='histone cluster 1, H3d', other_designations='H3 histone family, member B|histone 1, H3d|histone H3/b', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26196784, end_position_on_the_genomic_accession=26199293, orientation='minus', exon_count=2, OMIM=602811, no_hdr0=''),
  8352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8352, CurrentID=0, Status='live', Symbol='HIST1H3C', Aliases='H3.1, H3/c, H3FC', description='histone cluster 1, H3c', other_designations='H3 histone family, member C|histone 1, H3c|histone H3/c', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26045411, end_position_on_the_genomic_accession=26045869, orientation='plus', exon_count=1, OMIM=602812, no_hdr0=''),
  8353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8353, CurrentID=0, Status='live', Symbol='HIST1H3E', Aliases='H3.1, H3/d, H3FD', description='histone cluster 1, H3e', other_designations='H3 histone family, member D|histone 1, H3e|histone H3/d', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26225155, end_position_on_the_genomic_accession=26225616, orientation='plus', exon_count=1, OMIM=602813, no_hdr0=''),
  8354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8354, CurrentID=0, Status='live', Symbol='HIST1H3I', Aliases='H3.f, H3/f, H3FF', description='histone cluster 1, H3i', other_designations='H3 histone family, member F|histone 1, H3i|histone H3/f', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27871845, end_position_on_the_genomic_accession=27872321, orientation='minus', exon_count=1, OMIM=602814, no_hdr0=''),
  8355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8355, CurrentID=0, Status='live', Symbol='HIST1H3G', Aliases='H3/h, H3FH', description='histone cluster 1, H3g', other_designations='H3 histone family, member H|histone 1, H3g|histone H3/h', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26270918, end_position_on_the_genomic_accession=26271384, orientation='minus', exon_count=1, OMIM=602815, no_hdr0=''),
  8356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8356, CurrentID=0, Status='live', Symbol='HIST1H3J', Aliases='H3/j, H3FJ', description='histone cluster 1, H3j', other_designations='H3 histone family, member J|histone 1, H3j|histone H3/j', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27890315, end_position_on_the_genomic_accession=27890792, orientation='minus', exon_count=1, OMIM=602817, no_hdr0=''),
  8357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8357, CurrentID=0, Status='live', Symbol='HIST1H3H', Aliases='H3/k, H3F1K, H3FK', description='histone cluster 1, H3h', other_designations='H3 histone family, member K|histone 1, H3h|histone H3/k', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27810064, end_position_on_the_genomic_accession=27810536, orientation='plus', exon_count=1, OMIM=602818, no_hdr0=''),
  8358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8358, CurrentID=0, Status='live', Symbol='HIST1H3B', Aliases='H3/l, H3FL', description='histone cluster 1, H3b', other_designations='H3 histone family, member L|histone 1, H3b|histone H3/l', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26031589, end_position_on_the_genomic_accession=26032060, orientation='minus', exon_count=1, OMIM=602819, no_hdr0=''),
  8359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8359, CurrentID=0, Status='live', Symbol='HIST1H4A', Aliases='H4FA', description='histone cluster 1, H4a', other_designations='H4 histone family, member A|histone 1, H4a', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26021679, end_position_on_the_genomic_accession=26022050, orientation='plus', exon_count=1, OMIM=602822, no_hdr0=''),
  8360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8360, CurrentID=0, Status='live', Symbol='HIST1H4D', Aliases='H4/b, H4FB, dJ221C16.9', description='histone cluster 1, H4d', other_designations='H4 histone family, member B|histone 1, H4d', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26188710, end_position_on_the_genomic_accession=26189076, orientation='minus', exon_count=1, OMIM=602823, no_hdr0=''),
  401577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401577, CurrentID=0, Status='live', Symbol='CD99P1', Aliases='CD99L1, CXYorf12, MIC2R, NCRNA00103', description='CD99 molecule pseudogene 1', other_designations='CD99 antigen-like 1|CD99 molecule-like 1, pseudogene|MIC2-related', map_location='Xp22.33;Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2609265, end_position_on_the_genomic_accession=2657229, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  8362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8362, CurrentID=0, Status='live', Symbol='HIST1H4K', Aliases='H4/d, H4F2iii, H4FD, dJ160A22.1', description='histone cluster 1, H4k', other_designations='H4 histone family, member D|histone 1, H4k', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27831174, end_position_on_the_genomic_accession=27831527, orientation='minus', exon_count=1, OMIM=602825, no_hdr0=''),
  8363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8363, CurrentID=0, Status='live', Symbol='HIST1H4J', Aliases='H4/e, H4F2iv, H4FE, dJ160A22.2', description='histone cluster 1, H4j', other_designations='H4 histone family, member E|histone 1, H4j', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27824125, end_position_on_the_genomic_accession=27824480, orientation='plus', exon_count=1, OMIM=602826, no_hdr0=''),
  8364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8364, CurrentID=0, Status='live', Symbol='HIST1H4C', Aliases='H4/g, H4FG, dJ221C16.1', description='histone cluster 1, H4c', other_designations='H4 histone family, member G|histone 1, H4c', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26103948, end_position_on_the_genomic_accession=26104337, orientation='plus', exon_count=1, OMIM=602827, no_hdr0=''),
  8365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8365, CurrentID=0, Status='live', Symbol='HIST1H4H', Aliases='H4/h, H4FH', description='histone cluster 1, H4h', other_designations='H4 histone family, member H|histone 1, H4h', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26285126, end_position_on_the_genomic_accession=26285499, orientation='minus', exon_count=1, OMIM=602828, no_hdr0=''),
  8366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8366, CurrentID=0, Status='live', Symbol='HIST1H4B', Aliases='H4/I, H4FI', description='histone cluster 1, H4b', other_designations='H4 histone family, member I|histone 1, H4b', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26026896, end_position_on_the_genomic_accession=26027252, orientation='minus', exon_count=1, OMIM=602829, no_hdr0=''),
  8367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8367, CurrentID=0, Status='live', Symbol='HIST1H4E', Aliases='H4/j, H4FJ', description='histone cluster 1, H4e', other_designations='H4 histone family, member J|histone 1, H4e', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26204645, end_position_on_the_genomic_accession=26205021, orientation='plus', exon_count=1, OMIM=602830, no_hdr0=''),
  8368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8368, CurrentID=0, Status='live', Symbol='HIST1H4L', Aliases='H4.k, H4/k, H4FK', description='histone cluster 1, H4l', other_designations='H4 histone family, member K|histone 1, H4l', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27873148, end_position_on_the_genomic_accession=27873511, orientation='minus', exon_count=1, OMIM=602831, no_hdr0=''),
  8369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8369, CurrentID=0, Status='live', Symbol='HIST1H4G', Aliases='H4/l, H4FL', description='histone cluster 1, H4g', other_designations='H4 histone family, member L|histone 1, H4g', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26246611, end_position_on_the_genomic_accession=26246977, orientation='minus', exon_count=1, OMIM=602832, no_hdr0=''),
  8370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8370, CurrentID=0, Status='live', Symbol='HIST2H4A', Aliases='FO108, H4, H4/n, H4F2, H4FN, HIST2H4', description='histone cluster 2, H4a', other_designations='H4 histone family, member N|H4 histone, family 2|histone 2, H4a|histone IV, family 2', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149832657, end_position_on_the_genomic_accession=149833052, orientation='plus', exon_count=1, OMIM=142750, no_hdr0=''),
  100533619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533619, CurrentID=0, Status='live', Symbol='TMPOP2', Aliases='lncRNA-EBIC', description='thymopoietin pseudogene 2 (functional)', other_designations='EZH2-binding lncRNA in cervical cancer', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74667504, end_position_on_the_genomic_accession=74668903, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8372, CurrentID=0, Status='live', Symbol='HYAL3', Aliases='HYAL-3, LUCA-3, LUCA3', description='hyaluronoglucosaminidase 3', other_designations='lung carcinoma protein 3', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50292828, end_position_on_the_genomic_accession=50299468, orientation='minus', exon_count=5, OMIM=604038, no_hdr0=''),
  8373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8373, CurrentID=0, Status='live', Symbol='IFIT1P1', Aliases='G13P1, IFI56P, IFIT1P, II56P', description='interferon induced protein with tetratricopeptide repeats 1 pseudogene 1', other_designations='interferon-induced protein 56 pseudogene|interferon-induced protein with tetratricopeptide repeats 1 pseudogene 1|interferon-induced protein with tetratricopeptide repeats 1, pseudogene', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32384602, end_position_on_the_genomic_accession=32386642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23241, CurrentID=0, Status='live', Symbol='PACS2', Aliases='PACS-2, PACS1L', description='phosphofurin acidic cluster sorting protein 2', other_designations='PACS1-like protein', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105300811, end_position_on_the_genomic_accession=105398147, orientation='plus', exon_count=27, OMIM=610423, no_hdr0=''),
  8377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8377, CurrentID=0, Status='live', Symbol='LOH1CR1', Aliases='', description='loss of heterozygosity, 1, chromosomal region 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8378, CurrentID=0, Status='live', Symbol='LOH19CR1', Aliases='DBA', description='loss of heterozygosity, 19, chromosomal region 1', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8379, CurrentID=0, Status='live', Symbol='MAD1L1', Aliases='MAD1, PIG9, TP53I9, TXBP181', description='MAD1 mitotic arrest deficient-like 1 (yeast)', other_designations='MAD1-like protein 1|mitotic arrest deficient 1-like protein 1|mitotic checkpoint MAD1 protein homolog|mitotic-arrest deficient 1, yeast, homolog-like 1|tax-binding protein 181|tumor protein p53 inducible protein 9', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1815792, end_position_on_the_genomic_accession=2232948, orientation='minus', exon_count=25, OMIM=602686, no_hdr0=''),
  139452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139452, CurrentID=0, Status='live', Symbol='RPL6P30', Aliases='RPL6_11_1767', description='ribosomal protein L6 pseudogene 30', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16198402, end_position_on_the_genomic_accession=16199310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23242, CurrentID=0, Status='live', Symbol='COBL', Aliases='', description='cordon-bleu WH2 repeat protein', other_designations='cordon-bleu homolog', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51016212, end_position_on_the_genomic_accession=51316818, orientation='minus', exon_count=17, OMIM=610317, no_hdr0=''),
  8382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8382, CurrentID=0, Status='live', Symbol='NME5', Aliases='NM23-H5, NM23H5, RSPH23', description='NME/NM23 family member 5', other_designations='IPIA-beta|NDK-H 5|NDP kinase homolog 5|inhibitor of p53-induced apoptosis-beta|non-metastatic cells 5 protein expressed in|non-metastatic cells 5, protein expressed in (nucleoside-diphosphate kinase)|radial spoke 23 homolog|testis-specific nm23 homolog', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138115172, end_position_on_the_genomic_accession=138139443, orientation='minus', exon_count=8, OMIM=603575, no_hdr0=''),
  8383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8383, CurrentID=0, Status='live', Symbol='OR1A1', Aliases='OR17-7', description='olfactory receptor family 1 subfamily A member 1', other_designations='olfactory receptor 17-7|olfactory receptor OR17-11|olfactory receptor, family 1, subfamily A, member 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3215621, end_position_on_the_genomic_accession=3216550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8384, CurrentID=0, Status='live', Symbol='OR1D3P', Aliases='OR11-13, OR11-22, OR17-23, OR1D6P, OR1D7P', description='olfactory receptor family 1 subfamily D member 3 pseudogene', other_designations='olfactory receptor, family 1, subfamily D, member 3 pseudogene|olfactory receptor, family 1, subfamily D, member 6 pseudogene|olfactory receptor, family 1, subfamily D, member 7 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3265812, end_position_on_the_genomic_accession=3266454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8386, CurrentID=0, Status='live', Symbol='OR1D5', Aliases='C17orf2, OR17-2, OR17-30, OR17-31', description='olfactory receptor family 1 subfamily D member 5', other_designations='olfactory receptor 17-31|olfactory receptor OR17-2|olfactory receptor, family 1, subfamily D, member 5', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3062669, end_position_on_the_genomic_accession=3063607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8387, CurrentID=0, Status='live', Symbol='OR1E1', Aliases='HGM071, OR13-66, OR17-2, OR17-32, OR1E5, OR1E6, OR1E8P, OR1E9P, OST547', description='olfactory receptor family 1 subfamily E member 1', other_designations='OR5-85|olfactory receptor 13-66|olfactory receptor 17-2/17-32|olfactory receptor 1E5|olfactory receptor 1E6|olfactory receptor 5-85|olfactory receptor OR17-18|olfactory receptor OR17-4|olfactory receptor, family 1, subfamily E, member 1|olfactory receptor, family 1, subfamily E, member 5|olfactory receptor, family 1, subfamily E, member 6|olfactory receptor, family 1, subfamily E, member 8 pseudogene|olfactory receptor, family 1, subfamily E, member 9 pseudogene|olfactory receptor-like protein HGMP07I', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3397466, end_position_on_the_genomic_accession=3398410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8388, CurrentID=0, Status='live', Symbol='OR1E2', Aliases='OR17-135, OR17-136, OR17-93, OR1E4, OR1E7, OST529', description='olfactory receptor family 1 subfamily E member 2', other_designations='olfactory receptor 17-93/17-135/17-136|olfactory receptor 1E4|olfactory receptor OR17-17|olfactory receptor OR17-3|olfactory receptor, family 1, subfamily E, member 2|olfactory receptor, family 1, subfamily E, member 4|olfactory receptor, family 1, subfamily E, member 7', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3432870, end_position_on_the_genomic_accession=3433841, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8389, CurrentID=0, Status='live', Symbol='OR1E3', Aliases='OR17-210P, OR1E3', description='olfactory receptor family 1 subfamily E member 3 (gene/pseudogene)', other_designations='OLFR 17-210|olfactory receptor 17-210|olfactory receptor, family 1, subfamily E, member 3 pseudogene|seven transmembrane helices olfactory receptor', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3116427, end_position_on_the_genomic_accession=3117374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8390, CurrentID=0, Status='live', Symbol='OR1G1', Aliases='OR17-130, OR17-209, OR1G2', description='olfactory receptor family 1 subfamily G member 1', other_designations='olfactory receptor 17-209|olfactory receptor 1G2|olfactory receptor OR17-8|olfactory receptor, family 1, subfamily G, member 1|olfactory receptor, family 1, subfamily G, member 2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3126505, end_position_on_the_genomic_accession=3128253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8391, CurrentID=0, Status='live', Symbol='OR1P1', Aliases='OR17-208P, OR1P1', description='olfactory receptor family 1 subfamily P member 1 (gene/pseudogene)', other_designations='olfactory receptor, family 1, subfamily P, member 1 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3153890, end_position_on_the_genomic_accession=3154882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8392, CurrentID=0, Status='live', Symbol='OR3A3', Aliases='OR17-137, OR17-16, OR17-201, OR3A6, OR3A7, OR3A8P', description='olfactory receptor family 3 subfamily A member 3', other_designations='olfactory receptor 17-201|olfactory receptor 3A6|olfactory receptor 3A7|olfactory receptor 3A8|olfactory receptor OR17-22|olfactory receptor, family 3, subfamily A, member 3|olfactory receptor, family 3, subfamily A, member 6|olfactory receptor, family 3, subfamily A, member 7|olfactory receptor, family 3, subfamily A, member 8 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3420424, end_position_on_the_genomic_accession=3421666, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8394, CurrentID=0, Status='live', Symbol='PIP5K1A', Aliases='', description='phosphatidylinositol-4-phosphate 5-kinase, type I, alpha', other_designations='68 kDa type I phosphatidylinositol 4-phosphate 5-kinase alpha|68 kDa type I phosphatidylinositol-4-phosphate 5-kinase alpha|PIP5K1-alpha|PIP5KIalpha|phosphatidylinositol 4-phosphate 5-kinase type I alpha|phosphatidylinositol-4-phosphate 5-kinase type-1 alpha|ptdIns(4)P-5-kinase 1 alpha', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151197954, end_position_on_the_genomic_accession=151249535, orientation='plus', exon_count=21, OMIM=603275, no_hdr0=''),
  8395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8395, CurrentID=0, Status='live', Symbol='PIP5K1B', Aliases='MSS4, STM7', description='phosphatidylinositol-4-phosphate 5-kinase, type I, beta', other_designations='PIP5K1-beta|PIP5KIbeta|phosphatidylinositol 4-phosphate 5-kinase type I beta|phosphatidylinositol-4-phosphate 5-kinase type-1 beta|protein STM-7|ptdIns(4)P-5-kinase 1 beta|type I phosphatidylinositol 4-phosphate 5-kinase beta|type I phosphatidylinositol-4-phosphate 5-kinase beta', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68705217, end_position_on_the_genomic_accession=69009176, orientation='plus', exon_count=22, OMIM=602745, no_hdr0=''),
  401612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401612, CurrentID=0, Status='live', Symbol='SLC25A53', Aliases='MCART6', description='solute carrier family 25 member 53', other_designations='mitochondrial carrier triple repeat protein 6|solute carrier family 25, member 53', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104099214, end_position_on_the_genomic_accession=104157027, orientation='minus', exon_count=5, OMIM=300941, no_hdr0=''),
  401613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401613, CurrentID=0, Status='live', Symbol='LINC00890', Aliases='DKFZp686D0853', description='long intergenic non-protein coding RNA 890', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111511662, end_position_on_the_genomic_accession=111522399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8398, CurrentID=0, Status='live', Symbol='PLA2G6', Aliases='CaI-PLA2, GVI, INAD1, IPLA2-VIA, NBIA2, NBIA2A, NBIA2B, PARK14, PLA2, PNPLA9, iPLA2, iPLA2beta', description='phospholipase A2 group VI', other_designations='85 kDa calcium-independent phospholipase A2|GVI PLA2|iPLA2-beta|intracellular membrane-associated calcium-independent phospholipase A2 beta|neurodegeneration with brain iron accumulation 2|patatin-like phospholipase domain-containing protein 9|phospholipase A2, group VI (cytosolic, calcium-independent)', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38111495, end_position_on_the_genomic_accession=38192109, orientation='minus', exon_count=29, OMIM=603604, no_hdr0=''),
  8399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8399, CurrentID=0, Status='live', Symbol='PLA2G10', Aliases='GXPLA2, GXSPLA2, SPLA2, sPLA2-X', description='phospholipase A2 group X', other_designations='group X secretory phospholipase A2|phosphatidylcholine 2-acylhydrolase 10|phospholipase A2, group X', map_location='16p13.1-p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14672548, end_position_on_the_genomic_accession=14694669, orientation='minus', exon_count=8, OMIM=603603, no_hdr0=''),
  8400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8400, CurrentID=0, Status='live', Symbol='SCZD6', Aliases='', description='schizophrenia disorder 6', other_designations='Schizophrenia susceptibility locus, chromosome 8p-related', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603013, no_hdr0=''),
  8401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8401, CurrentID=0, Status='live', Symbol='SCZD7', Aliases='', description='schizophrenia disorder 7', other_designations='Schizophrenia susceptibility locus, chromosome 13q-related', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603176, no_hdr0=''),
  8402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8402, CurrentID=0, Status='live', Symbol='SLC25A11', Aliases='OGC, SLC20A4', description='solute carrier family 25 (mitochondrial carrier; oxoglutarate carrier), member 11', other_designations='OGCP|solute carrier family 20 (oxoglutarate carrier), member 4|solute carrier family 25 member 11', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4937130, end_position_on_the_genomic_accession=4940167, orientation='minus', exon_count=8, OMIM=604165, no_hdr0=''),
  8403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8403, CurrentID=0, Status='live', Symbol='SOX14', Aliases='SOX28', description='SRY-box 14', other_designations='HMG box transcription factor SOX-14|SRY (sex determining region Y)-box 14|SRY box 14', map_location='3q22-q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137764292, end_position_on_the_genomic_accession=137766334, orientation='plus', exon_count=1, OMIM=604747, no_hdr0=''),
  8404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8404, CurrentID=0, Status='live', Symbol='SPARCL1', Aliases='MAST 9, MAST9, PIG33, SC1', description='SPARC like 1', other_designations='SPARC-like 1 (hevin)|high endothelial venule protein|proliferation-inducing protein 33', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87473330, end_position_on_the_genomic_accession=87529503, orientation='minus', exon_count=14, OMIM=606041, no_hdr0=''),
  8405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8405, CurrentID=0, Status='live', Symbol='SPOP', Aliases='BTBD32, TEF2', description='speckle type POZ protein', other_designations='HIB homolog 1|roadkill homolog 1', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49598884, end_position_on_the_genomic_accession=49678163, orientation='minus', exon_count=14, OMIM=602650, no_hdr0=''),
  8406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8406, CurrentID=0, Status='live', Symbol='SRPX', Aliases='DRS, ETX1, HEL-S-83p1, SRPX', description='sushi-repeat containing protein, X-linked', other_designations='epididymis secretory sperm binding protein Li 83p|sushi-repeat-containing protein, X chromosome|sushi-repeat-containing protein, X-linked', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38149335, end_position_on_the_genomic_accession=38220924, orientation='minus', exon_count=10, OMIM=300187, no_hdr0=''),
  8407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8407, CurrentID=0, Status='live', Symbol='TAGLN2', Aliases='HA1756', description='transgelin 2', other_designations='SM22-alpha homolog|epididymis tissue protein Li 7e', map_location='1q21-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159918107, end_position_on_the_genomic_accession=159925542, orientation='minus', exon_count=7, OMIM=604634, no_hdr0=''),
  8408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8408, CurrentID=0, Status='live', Symbol='ULK1', Aliases='ATG1, ATG1A, UNC51, Unc51.1, hATG1', description='unc-51 like autophagy activating kinase 1', other_designations='ATG1 autophagy related 1 homolog|autophagy-related protein 1 homolog', map_location='12q24.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131894734, end_position_on_the_genomic_accession=131923162, orientation='plus', exon_count=28, OMIM=603168, no_hdr0=''),
  8409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8409, CurrentID=0, Status='live', Symbol='UXT', Aliases='ART-27, STAP1', description='ubiquitously expressed prefoldin like chaperone', other_designations='SKP2-associated alpha PFD 1|androgen receptor trapped clone 27 protein|ubiquitously expressed transcript protein|ubiquitously-expressed, prefoldin-like chaperone', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47651792, end_position_on_the_genomic_accession=47659180, orientation='minus', exon_count=8, OMIM=300234, no_hdr0=''),
  8410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8410, CurrentID=0, Status='live', Symbol='RPS9P1', Aliases='RPS9P, RPS9_2_1730', description='ribosomal protein S9 pseudogene 1', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36132394, end_position_on_the_genomic_accession=36133106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8411, CurrentID=0, Status='live', Symbol='EEA1', Aliases='MST105, MSTP105, ZFYVE2', description='early endosome antigen 1', other_designations='early endosome antigen 1, 162kD|early endosome-associated protein|endosome-associated protein p162|zinc finger FYVE domain-containing protein 2', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92772509, end_position_on_the_genomic_accession=92929331, orientation='minus', exon_count=32, OMIM=605070, no_hdr0=''),
  8412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8412, CurrentID=0, Status='live', Symbol='BCAR3', Aliases='NSP2, SH2D3B', description='breast cancer anti-estrogen resistance 3', other_designations='SH2 domain-containing protein 3B|breast cancer antiestrogen resistance 3 protein|dJ1033H22.2 (breast cancer anti-estrogen resistance 3)|novel SH2-containing protein 2', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93561786, end_position_on_the_genomic_accession=93847150, orientation='minus', exon_count=19, OMIM=604704, no_hdr0=''),
  401629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401629, CurrentID=0, Status='live', Symbol='FAM224B', Aliases='LINC00230B, NCRNA00230B', description='family with sequence similarity 224 member B (non-protein coding)', other_designations='family with sequence similarity 224, member B (non-protein coding)|long intergenic non-protein coding RNA 230B|non-protein coding RNA 230B', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17575161, end_position_on_the_genomic_accession=17579452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401630, CurrentID=0, Status='live', Symbol='FAM224A', Aliases='LINC00230A, NCRNA00230A', description='family with sequence similarity 224 member A (non-protein coding)', other_designations='family with sequence similarity 224, member A (non-protein coding)|long intergenic non-protein coding RNA 230A|non-protein coding RNA 230A', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18326533, end_position_on_the_genomic_accession=18330827, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8416, CurrentID=0, Status='live', Symbol='ANXA9', Aliases='ANX31', description='annexin A9', other_designations='annexin 31|annexin XXXI|annexin-31|annexin-9|pemphaxin', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150982023, end_position_on_the_genomic_accession=150996063, orientation='plus', exon_count=14, OMIM=603319, no_hdr0=''),
  8417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8417, CurrentID=0, Status='live', Symbol='STX7', Aliases='', description='syntaxin 7', other_designations='', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132457524, end_position_on_the_genomic_accession=132513198, orientation='minus', exon_count=11, OMIM=603217, no_hdr0=''),
  8418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8418, CurrentID=0, Status='live', Symbol='CMAHP', Aliases='CMAH, CSAH', description='cytidine monophospho-N-acetylneuraminic acid hydroxylase, pseudogene', other_designations='CMP-N-acetylneuraminic acid hydroxylase|CMP-Neu5Ac hydroxylase|CMP-NeuAc hydroxylase|CMP-sialic acid hydroxylase|cytidine monophosphate-N-acetylneuraminic acid hydroxylase (CMP-N-acetylneuraminate monooxygenase) pseudogene|cytidine monophosphate-N-acetylneuraminic acid hydroxylase (CMP-N-acetylneuraminate monooxygenase)(pseudogene)', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25081067, end_position_on_the_genomic_accession=25138392, orientation='minus', exon_count=14, OMIM=603209, no_hdr0=''),
  8419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8419, CurrentID=0, Status='live', Symbol='BFSP2', Aliases='CP47, CP49, CTRCT12, LIFL-L, PHAKOSIN', description='beaded filament structural protein 2', other_designations='49 kDa cytoskeletal protein|beaded filament protein CP49|beaded filament structural protein 2, phakinin|bfps2, Cytoskeletal protein, 49-kD|lens fiber cell beaded filament protein CP 47|lens fiber cell beaded filament protein CP 49|lens intermediate filament-like light', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133399995, end_position_on_the_genomic_accession=133475212, orientation='plus', exon_count=7, OMIM=603212, no_hdr0=''),
  8420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8420, CurrentID=0, Status='live', Symbol='SNHG3', Aliases='NCRNA00014, RNU17C, RNU17D, U17HG, U17HG-A, U17HG-AB', description='small nucleolar RNA host gene 3', other_designations='U17 small nucleolar RNA host|small nucleolar RNA host gene (non-protein coding) 3|small nucleolar RNA host gene 3 (non-protein coding)', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28505943, end_position_on_the_genomic_accession=28510892, orientation='plus', exon_count=4, OMIM=603238, no_hdr0=''),
  401637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401637, CurrentID=0, Status='live', Symbol='OR7E26P', Aliases='OR1-51, OR1-72, OR1-73, OR7E67P, OR7E68P, OR7E69P, OR7E70P, OR7E71P, OR7E72P, OR7E73P, OR7E74P, OR7E75P, OR912-108, OR912-109, OR912-110, OR912-46, OR912-95, hg523, hg674', description='olfactory receptor family 7 subfamily E member 26 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 26 pseudogene|olfactory receptor, family 7, subfamily E, member 67 pseudogene|olfactory receptor, family 7, subfamily E, member 68 pseudogene|olfactory receptor, family 7, subfamily E, member 69 pseudogene|olfactory receptor, family 7, subfamily E, member 70 pseudogene|olfactory receptor, family 7, subfamily E, member 71 pseudogene|olfactory receptor, family 7, subfamily E, member 72 pseudogene|olfactory receptor, family 7, subfamily E, member 73 pseudogene|olfactory receptor, family 7, subfamily E, member 74 pseudogene|olfactory receptor, family 7, subfamily E, member 75 pseudogene', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14999199, end_position_on_the_genomic_accession=14999852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8422, CurrentID=0, Status='live', Symbol='MEHMO', Aliases='', description='mental retardation, epileptic seizures, hypogonadism and -genitalism, microcephaly and obesity syndrome', other_designations='MEHMO syndrome (Mental retardation, epileptic seizures, hypogonadism and hypogenitalism, microcephaly, and obesity)', map_location='Xp22.13-p21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300148, no_hdr0=''),
  8424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8424, CurrentID=0, Status='live', Symbol='BBOX1', Aliases='BBH, BBOX, G-BBH, gamma-BBH', description='butyrobetaine (gamma), 2-oxoglutarate dioxygenase (gamma-butyrobetaine hydroxylase) 1', other_designations='gamma-butyrobetaine hydroxylase|gamma-butyrobetaine,2-oxoglutarate dioxygenase 1', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27040705, end_position_on_the_genomic_accession=27127807, orientation='plus', exon_count=10, OMIM=603312, no_hdr0=''),
  8425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8425, CurrentID=0, Status='live', Symbol='LTBP4', Aliases='ARCL1C, LTBP-4L, LTBP4S, LTBP4', description='latent transforming growth factor beta binding protein 4', other_designations='latent transforming growth factor-beta binding protein 4L', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40592888, end_position_on_the_genomic_accession=40629820, orientation='plus', exon_count=36, OMIM=604710, no_hdr0=''),
  8427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8427, CurrentID=0, Status='live', Symbol='ZNF282', Aliases='HUB1', description='zinc finger protein 282', other_designations='HTLV-I U5 repressive element-binding protein 1|HTLV-I U5RE-binding protein 1|HUB-1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149195462, end_position_on_the_genomic_accession=149226248, orientation='plus', exon_count=10, OMIM=603397, no_hdr0=''),
  8428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8428, CurrentID=0, Status='live', Symbol='STK24', Aliases='HEL-S-95, MST3, MST3B, STE20, STK3', description='serine/threonine kinase 24', other_designations='STE20-like kinase 3|STE20-like kinase MST3|epididymis secretory protein Li 95|mammalian STE20-like protein kinase 3|serine/threonine kinase 24 (STE20 homolog, yeast)|sterile 20-like kinase 3', map_location='13q31.2-q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98450199, end_position_on_the_genomic_accession=98577151, orientation='minus', exon_count=14, OMIM=604984, no_hdr0=''),
  401646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401646, CurrentID=0, Status='live', Symbol='GNAI2P2', Aliases='', description='guanine nucleotide binding protein (G protein), alpha inhibiting activity polypeptide 2 pseudogene 2', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78068631, end_position_on_the_genomic_accession=78069540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8431, CurrentID=0, Status='live', Symbol='NR0B2', Aliases='SHP, SHP1', description='nuclear receptor subfamily 0 group B member 2', other_designations='nuclear receptor SHP|nuclear receptor subfamily 0, group B, member 2|orphan nuclear receptor SHP|small heterodimer partner', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26911484, end_position_on_the_genomic_accession=26914076, orientation='minus', exon_count=3, OMIM=604630, no_hdr0=''),
  401648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401648, CurrentID=0, Status='live', Symbol='ST13P13', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 13', other_designations='', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103215709, end_position_on_the_genomic_accession=103216794, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8433, CurrentID=0, Status='live', Symbol='UTF1', Aliases='', description='undifferentiated embryonic cell transcription factor 1', other_designations='', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133230274, end_position_on_the_genomic_accession=133231558, orientation='plus', exon_count=2, OMIM=604130, no_hdr0=''),
  8434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8434, CurrentID=0, Status='live', Symbol='RECK', Aliases='ST15', description='reversion-inducing-cysteine-rich protein with kazal motifs', other_designations='membrane-anchored glycoprotein (metastasis and invasion)|suppression of tumorigenicity 15 (reversion-inducing-cysteine-rich protein with kazal motifs)|suppression of tumorigenicity 5 (reversion-inducing-cysteine-rich protein with kazal motifs)|suppressor of tumorigenicity 15 protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36036913, end_position_on_the_genomic_accession=36124455, orientation='plus', exon_count=23, OMIM=605227, no_hdr0=''),
  8435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8435, CurrentID=0, Status='live', Symbol='SOAT2', Aliases='ACACT2, ACAT2, ARGP2', description='sterol O-acyltransferase 2', other_designations='ACAT-2|acyl Co-A: cholesterol acyltransferase 2|acyl coenzyme A:cholesterol acyltransferase 2|acyl-CoA:cholesterol acyltransferase 2|acyl-coenzyme A:cholesterol acyltransferase 2|cholesterol acyltransferase 2', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53102504, end_position_on_the_genomic_accession=53124539, orientation='plus', exon_count=16, OMIM=601311, no_hdr0=''),
  8436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8436, CurrentID=0, Status='live', Symbol='SDPR', Aliases='CAVIN2, PS-p68, SDR, cavin-2', description='serum deprivation response', other_designations='phosphatidylserine binding protein|phosphatidylserine-binding protein', map_location='2q32-q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191834305, end_position_on_the_genomic_accession=191847280, orientation='minus', exon_count=2, OMIM=606728, no_hdr0=''),
  8437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8437, CurrentID=0, Status='live', Symbol='RASAL1', Aliases='RASAL', description='RAS protein activator like 1 (GAP1 like)', other_designations='GAP1 like protein|ras GTPase-activating-like protein', map_location='12q23-q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113098819, end_position_on_the_genomic_accession=113136239, orientation='minus', exon_count=24, OMIM=604118, no_hdr0=''),
  8438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8438, CurrentID=0, Status='live', Symbol='RAD54L', Aliases='HR54, RAD54A, hHR54, hRAD54', description='RAD54-like (S. cerevisiae)', other_designations='RAD54 homolog', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46247695, end_position_on_the_genomic_accession=46278473, orientation='plus', exon_count=18, OMIM=603615, no_hdr0=''),
  8439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8439, CurrentID=0, Status='live', Symbol='NSMAF', Aliases='FAN', description='neutral sphingomyelinase activation associated factor', other_designations='factor associated with N-SMase activation|factor associated with neutral sphingomyelinase activation|neutral sphingomyelinase (N-SMase) activation associated factor', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58583504, end_position_on_the_genomic_accession=58659845, orientation='minus', exon_count=32, OMIM=603043, no_hdr0=''),
  8440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8440, CurrentID=0, Status='live', Symbol='NCK2', Aliases='GRB4, NCKbeta', description='NCK adaptor protein 2', other_designations='SH2/SH3 adaptor protein NCK-beta|growth factor receptor-bound protein 4|noncatalytic region of tyrosine kinase, beta', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105744942, end_position_on_the_genomic_accession=105894274, orientation='plus', exon_count=7, OMIM=604930, no_hdr0=''),
  8443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8443, CurrentID=0, Status='live', Symbol='GNPAT', Aliases='DAP-AT, DAPAT, DHAPAT', description='glyceronephosphate O-acyltransferase', other_designations='DHAP-AT|acyl-CoA:dihydroxyacetonephosphateacyltransferase|glycerone-phosphate O-acyltransferase', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231241173, end_position_on_the_genomic_accession=231277973, orientation='plus', exon_count=17, OMIM=602744, no_hdr0=''),
  8444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8444, CurrentID=0, Status='live', Symbol='DYRK3', Aliases='DYRK5, RED, REDK, hYAK3-2', description='dual specificity tyrosine-(Y)-phosphorylation regulated kinase 3', other_designations='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 3|dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 5|protein kinase Dyrk3|regulatory erythroid kinase', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206635534, end_position_on_the_genomic_accession=206649197, orientation='plus', exon_count=7, OMIM=603497, no_hdr0=''),
  8445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8445, CurrentID=0, Status='live', Symbol='DYRK2', Aliases='', description='dual specificity tyrosine-(Y)-phosphorylation regulated kinase 2', other_designations='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 2', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67648732, end_position_on_the_genomic_accession=67662664, orientation='plus', exon_count=3, OMIM=603496, no_hdr0=''),
  8446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8446, CurrentID=0, Status='live', Symbol='DUSP11', Aliases='PIR1', description='dual specificity phosphatase 11', other_designations='RNA/RNP complex-interacting phosphatase|dual specificity phosphatase 11 (RNA/RNP complex 1-interacting)|dual specificity protein phosphatase 11|phosphatase that interacts with RNA/RNP complex 1|serine/threonine specific protein phosphatase', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73762198, end_position_on_the_genomic_accession=73780157, orientation='minus', exon_count=9, OMIM=603092, no_hdr0=''),
  8447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8447, CurrentID=0, Status='live', Symbol='DOC2B', Aliases='DOC2BL', description='double C2-like domains, beta', other_designations='doc2-beta', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=143688, end_position_on_the_genomic_accession=181630, orientation='minus', exon_count=12, OMIM=604568, no_hdr0=''),
  8448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8448, CurrentID=0, Status='live', Symbol='DOC2A', Aliases='Doc2', description='double C2-like domains, alpha', other_designations='doc2-alpha', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30005514, end_position_on_the_genomic_accession=30023280, orientation='minus', exon_count=18, OMIM=604567, no_hdr0=''),
  401665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401665, CurrentID=0, Status='live', Symbol='OR51T1', Aliases='OR11-26', description='olfactory receptor family 51 subfamily T member 1', other_designations='olfactory receptor OR11-26|olfactory receptor, family 51, subfamily T, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4881819, end_position_on_the_genomic_accession=4882883, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8450, CurrentID=0, Status='live', Symbol='CUL4B', Aliases='CUL-4B, MRXHF2, MRXS15, MRXSC, SFM2', description='cullin 4B', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120524589, end_position_on_the_genomic_accession=120575829, orientation='minus', exon_count=26, OMIM=300304, no_hdr0=''),
  8451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8451, CurrentID=0, Status='live', Symbol='CUL4A', Aliases='', description='cullin 4A', other_designations='CUL-4A', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113208193, end_position_on_the_genomic_accession=113265078, orientation='plus', exon_count=24, OMIM=603137, no_hdr0=''),
  8452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8452, CurrentID=0, Status='live', Symbol='CUL3', Aliases='CUL-3, PHA2E', description='cullin 3', other_designations='', map_location='2q36.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224470150, end_position_on_the_genomic_accession=224585397, orientation='minus', exon_count=21, OMIM=603136, no_hdr0=''),
  8453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8453, CurrentID=0, Status='live', Symbol='CUL2', Aliases='', description='cullin 2', other_designations='CUL-2', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35008551, end_position_on_the_genomic_accession=35126899, orientation='minus', exon_count=28, OMIM=603135, no_hdr0=''),
  8454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8454, CurrentID=0, Status='live', Symbol='CUL1', Aliases='', description='cullin 1', other_designations='CUL-1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148698178, end_position_on_the_genomic_accession=148801110, orientation='plus', exon_count=26, OMIM=603134, no_hdr0=''),
  8455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8455, CurrentID=0, Status='live', Symbol='ATRN', Aliases='DPPT-L, MGCA', description='attractin', other_designations='attractin-2|mahogany homolog|mahogany protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3471018, end_position_on_the_genomic_accession=3651122, orientation='plus', exon_count=32, OMIM=603130, no_hdr0=''),
  8456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8456, CurrentID=0, Status='live', Symbol='FOXN1', Aliases='FKHL20, RONU, WHN', description='forkhead box N1', other_designations='Rowett nude|winged helix nude|winged-helix nude|winged-helix transcription factor nude', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28506229, end_position_on_the_genomic_accession=28538900, orientation='plus', exon_count=13, OMIM=600838, no_hdr0=''),
  8458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8458, CurrentID=0, Status='live', Symbol='TTF2', Aliases='HuF2, ZGRF6', description='transcription termination factor, RNA polymerase II', other_designations='F2|RNA polymerase II termination factor|human factor 2|lodestar homolog|lodestar protein|transcription release factor 2|zinc finger, GRF-type containing 6', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117060288, end_position_on_the_genomic_accession=117102870, orientation='plus', exon_count=25, OMIM=604718, no_hdr0=''),
  8459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8459, CurrentID=0, Status='live', Symbol='TPST2', Aliases='TANGO13B', description='tyrosylprotein sulfotransferase 2', other_designations='TPST-2|transport and golgi organization 13 homolog B|tyrosylprotein phosphotransferase 2|tyrosylprotein sulfotransferase-2', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26525748, end_position_on_the_genomic_accession=26590125, orientation='minus', exon_count=9, OMIM=603126, no_hdr0=''),
  8460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8460, CurrentID=0, Status='live', Symbol='TPST1', Aliases='TANGO13A', description='tyrosylprotein sulfotransferase 1', other_designations='TPST-1|transport and golgi organization 13 homolog A|tyrosylprotein sulfotransferase-1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66205272, end_position_on_the_genomic_accession=66360451, orientation='plus', exon_count=8, OMIM=603125, no_hdr0=''),
  100271490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271490, CurrentID=0, Status='live', Symbol='RPL32P25', Aliases='RPL32_16_1167', description='ribosomal protein L32 pseudogene 25', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95963219, end_position_on_the_genomic_accession=95963624, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8462, CurrentID=0, Status='live', Symbol='KLF11', Aliases='FKLF, FKLF1, MODY7, TIEG2, Tieg3', description='Kruppel-like factor 11', other_designations='TGFB-inducible early growth response protein 2|TIEG-2|transforming growth factor-beta-inducible early growth response protein 2', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10042985, end_position_on_the_genomic_accession=10054836, orientation='plus', exon_count=7, OMIM=603301, no_hdr0=''),
  8463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8463, CurrentID=0, Status='live', Symbol='TEAD2', Aliases='ETF, TEAD-2, TEF-4, TEF4', description='TEA domain family member 2', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49340595, end_position_on_the_genomic_accession=49362457, orientation='minus', exon_count=17, OMIM=601729, no_hdr0=''),
  8464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8464, CurrentID=0, Status='live', Symbol='SUPT3H', Aliases='SPT3, SPT3L', description='SPT3 homolog, SAGA and STAGA complex component', other_designations='SPT3-like protein|suppressor of Ty 3 homolog', map_location='6p21.1-p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44807499, end_position_on_the_genomic_accession=45378051, orientation='minus', exon_count=20, OMIM=602947, no_hdr0=''),
  139538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139538, CurrentID=0, Status='live', Symbol='VENTXP1', Aliases='CT18, NA88A, VENTX2P1', description='VENT homeobox pseudogene 1', other_designations='VENT-like homeobox 2 pseudogene 1|cancer/testis antigen 18|tumor antigen NA88A', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26558337, end_position_on_the_genomic_accession=26561052, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8467, CurrentID=0, Status='live', Symbol='SMARCA5', Aliases='ISWI, SNF2H, WCRF135, hISWI, hSNF2H', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 5', other_designations='SWI/SNF-related matrix-associated actin-dependent regulator of chromatin A5|sucrose nonfermenting protein 2 homolog|sucrose nonfermenting-like 5', map_location='4q31.1-q31.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143513463, end_position_on_the_genomic_accession=143557489, orientation='plus', exon_count=24, OMIM=603375, no_hdr0=''),
  8468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8468, CurrentID=0, Status='live', Symbol='FKBP6', Aliases='FKBP36', description='FK506 binding protein 6', other_designations='36 kDa FKBP|FK506 binding protein 6, 36kDa|PPIase FKBP6|immunophilin FKBP36|inactive PPIase FKBP6|peptidyl-prolyl cis-trans isomerase FKBP6|rotamase', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73328152, end_position_on_the_genomic_accession=73358637, orientation='plus', exon_count=9, OMIM=604839, no_hdr0=''),
  8470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8470, CurrentID=0, Status='live', Symbol='SORBS2', Aliases='ARGBP2, PRO0618', description='sorbin and SH3 domain containing 2', other_designations='Arg binding protein 2|Arg/Abl-interacting protein 2|arg-binding protein 2', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185585444, end_position_on_the_genomic_accession=185956716, orientation='minus', exon_count=37, OMIM=616349, no_hdr0=''),
  8471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8471, CurrentID=0, Status='live', Symbol='IRS4', Aliases='IRS-4, PY160', description='insulin receptor substrate 4', other_designations='160 kDa phosphotyrosine protein|phosphoprotein of 160 kDa|pp160', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108719949, end_position_on_the_genomic_accession=108736584, orientation='minus', exon_count=4, OMIM=300904, no_hdr0=''),
  8473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8473, CurrentID=0, Status='live', Symbol='OGT', Aliases='HINCUT-1, HRNT1, O-GLCNAC', description='O-linked N-acetylglucosamine (GlcNAc) transferase', other_designations='O-GlcNAc transferase p110 subunit|O-GlcNAc transferase subunit p110|O-linked N-acetylglucosamine (GlcNAc) transferase (UDP-N-acetylglucosamine:polypeptide-N-acetylglucosaminyl transferase)|O-linked N-acetylglucosamine transferase 110 kDa subunit|UDP-N-acetylglucosamine:polypeptide-N-acetylglucosaminyl transferase|uridinediphospho-N-acetylglucosamine:polypeptide beta-N-acetylglucosaminyl transferase', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71533062, end_position_on_the_genomic_accession=71575897, orientation='plus', exon_count=22, OMIM=300255, no_hdr0=''),
  8476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8476, CurrentID=0, Status='live', Symbol='CDC42BPA', Aliases='MRCK, MRCKA, PK428', description='CDC42 binding protein kinase alpha (DMPK-like)', other_designations='CDC42 binidng protein kinase beta|CDC42-binding protein kinase alpha (DMPK-like)|DMPK-like alpha|MRCK alpha|myotonic dystrophy kinase-related CDC42-binding kinase alpha|myotonic dystrophy kinase-related CDC42-binding protein kinase alpha|myotonic dystrophy protein kinase-like alpha|ser-thr protein kinase PK428|ser-thr protein kinase related to the myotonic dystrophy protein kinase', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226989858, end_position_on_the_genomic_accession=227318492, orientation='minus', exon_count=40, OMIM=603412, no_hdr0=''),
  8477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8477, CurrentID=0, Status='live', Symbol='GPR65', Aliases='TDAG8, hTDAG8', description='G protein-coupled receptor 65', other_designations='G-protein coupled receptor 65|T-cell death-associated gene 8 protein', map_location='14q31-q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88005124, end_position_on_the_genomic_accession=88014811, orientation='plus', exon_count=2, OMIM=604620, no_hdr0=''),
  8479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8479, CurrentID=0, Status='live', Symbol='HIRIP3', Aliases='', description='HIRA interacting protein 3', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29992321, end_position_on_the_genomic_accession=29996096, orientation='minus', exon_count=7, OMIM=603365, no_hdr0=''),
  8480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8480, CurrentID=0, Status='live', Symbol='RAE1', Aliases='MIG14, MRNP41, Mnrp41, dJ481F12.3, dJ800J21.1', description='ribonucleic acid export 1', other_designations='RAE1 RNA export 1 homolog|homolog of yeast Rae1 (Bharathi) mRNA-associated protein of 41 kDa (Kraemer)|mRNA export protein|mRNA-associated protein MRNP 41|mRNA-binding protein, 41-kD|migration-inducing gene 14|rae1 protein homolog', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57351089, end_position_on_the_genomic_accession=57378463, orientation='plus', exon_count=15, OMIM=603343, no_hdr0=''),
  8481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8481, CurrentID=0, Status='live', Symbol='OFD1', Aliases='71-7A, CXorf5, JBTS10, RP23, SGBS2', description='oral-facial-digital syndrome 1', other_designations='protein 71-7A', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13715423, end_position_on_the_genomic_accession=13771402, orientation='plus', exon_count=28, OMIM=300170, no_hdr0=''),
  8482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8482, CurrentID=0, Status='live', Symbol='SEMA7A', Aliases='CD108, CDw108, H-SEMA-K1, H-Sema-L, JMH, SEMAK1, SEMAL', description='semaphorin 7A, GPI membrane anchor (John Milton Hagen blood group)', other_designations='JMH blood group antigen|John Milton Hagen blood group H-Sema K1|john-Milton-Hargen human blood group Ag|sema K1|sema L|sema domain, immunoglobulin domain (Ig), and GPI membrane anchor, (semaphorin) 7A (JMH blood group)|sema domain, immunoglobulin domain (Ig), and GPI membrane anchor, 7A|semaphorin K1|semaphorin L|semaphorin-K1|semaphorin-L', map_location='15q22.3-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74409289, end_position_on_the_genomic_accession=74433958, orientation='minus', exon_count=15, OMIM=607961, no_hdr0=''),
  8483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8483, CurrentID=0, Status='live', Symbol='CILP', Aliases='CILP-1, HsT18872', description='cartilage intermediate layer protein', other_designations='cartilage intermediate layer protein 1 C1|cartilage intermediate layer protein 1 C2|cartilage intermediate layer protein, nucleotide pyrophosphohydrolase|cartilage intermediate-layer protein', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65195999, end_position_on_the_genomic_accession=65211502, orientation='minus', exon_count=9, OMIM=603489, no_hdr0=''),
  8484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8484, CurrentID=0, Status='live', Symbol='GALR3', Aliases='', description='galanin receptor 3', other_designations='GAL3-R|GALR-3|galanin receptor, family member 3', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37822009, end_position_on_the_genomic_accession=37825865, orientation='plus', exon_count=3, OMIM=603692, no_hdr0=''),
  100271494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271494, CurrentID=0, Status='live', Symbol='RPL34P19', Aliases='RPL34_10_1028', description='ribosomal protein L34 pseudogene 19', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31984255, end_position_on_the_genomic_accession=31984555, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8487, CurrentID=0, Status='live', Symbol='GEMIN2', Aliases='SIP1, SIP1-delta', description='gem nuclear organelle associated protein 2', other_designations='SMN interacting protein 1-delta|component of gems 2|gem (nuclear organelle) associated protein 2|gemin-2|survival of motor neuron protein interacting protein 1', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39114284, end_position_on_the_genomic_accession=39136973, orientation='plus', exon_count=11, OMIM=602595, no_hdr0=''),
  8488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8488, CurrentID=0, Status='live', Symbol='RPL3P1', Aliases='RPL3P, RPL3_8_1731', description='ribosomal protein L3 pseudogene 1', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36168921, end_position_on_the_genomic_accession=36170206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8489, CurrentID=0, Status='live', Symbol='RPL23AP3', Aliases='RPL23A_37_1729', description='ribosomal protein L23a pseudogene 3', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36016042, end_position_on_the_genomic_accession=36016557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8490, CurrentID=0, Status='live', Symbol='RGS5', Aliases='MST092, MST106, MST129, MSTP032, MSTP092, MSTP106, MSTP129', description='regulator of G-protein signaling 5', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163142299, end_position_on_the_genomic_accession=163321791, orientation='minus', exon_count=9, OMIM=603276, no_hdr0=''),
  8491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8491, CurrentID=0, Status='live', Symbol='MAP4K3', Aliases='GLK, MAPKKKK3, MEKKK 3, MEKKK3, RAB8IPL1', description='mitogen-activated protein kinase kinase kinase kinase 3', other_designations='MAPK/ERK kinase kinase kinase 3|MEK kinase kinase 3|germinal center kinase-like kinase|germinal center kinase-related protein kinase', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39249266, end_position_on_the_genomic_accession=39437312, orientation='minus', exon_count=34, OMIM=604921, no_hdr0=''),
  8492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8492, CurrentID=0, Status='live', Symbol='PRSS12', Aliases='BSSP-3, BSSP3, MRT1', description='protease, serine 12', other_designations='brain-specific serine protease 3|leydin|protease, serine, 12 (neurotrypsin, motopsin)', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118280038, end_position_on_the_genomic_accession=118352815, orientation='minus', exon_count=15, OMIM=606709, no_hdr0=''),
  8493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8493, CurrentID=0, Status='live', Symbol='PPM1D', Aliases='PP2C-DELTA, WIP1', description='protein phosphatase, Mg2+/Mn2+ dependent 1D', other_designations='protein phosphatase 1D magnesium-dependent, delta isoform|protein phosphatase 2C delta isoform|protein phosphatase Mg2+/Mn2+ dependent 1D|protein phosphatase Wip1|protein phosphatase, Mg2+/Mn2+ dependent, 1D|wild-type p53-induced phosphatase 1', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60600183, end_position_on_the_genomic_accession=60666280, orientation='plus', exon_count=7, OMIM=605100, no_hdr0=''),
  8495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8495, CurrentID=0, Status='live', Symbol='PPFIBP2', Aliases='Cclp1', description='PTPRF interacting protein, binding protein 2 (liprin beta 2)', other_designations='protein tyrosine phosphatase receptor type f polypeptide-interacting protein-binding protein 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7513765, end_position_on_the_genomic_accession=7667344, orientation='plus', exon_count=36, OMIM=603142, no_hdr0=''),
  8496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8496, CurrentID=0, Status='live', Symbol='PPFIBP1', Aliases='L2, SGT2, hSGT2, hSgt2p', description='PTPRF interacting protein, binding protein 1 (liprin beta 1)', other_designations='PTPRF-interacting protein-binding protein 1|liprin related protein|liprin-beta 1|protein tyrosine phosphatase receptor type f polypeptide-interacting protein-binding protein 1|protein-tyrosine phosphatase receptor-type f polypeptide-interacting protein-binding protein 1', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27524112, end_position_on_the_genomic_accession=27695564, orientation='plus', exon_count=32, OMIM=603141, no_hdr0=''),
  8497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8497, CurrentID=0, Status='live', Symbol='PPFIA4', Aliases='', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 4', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203026521, end_position_on_the_genomic_accession=203078736, orientation='plus', exon_count=32, OMIM=603145, no_hdr0=''),
  8498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8498, CurrentID=0, Status='live', Symbol='RANBP3', Aliases='', description='RAN binding protein 3', other_designations='RAN-binding protein-3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5916139, end_position_on_the_genomic_accession=5978309, orientation='minus', exon_count=19, OMIM=603327, no_hdr0=''),
  8499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8499, CurrentID=0, Status='live', Symbol='PPFIA2', Aliases='', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 2', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81257975, end_position_on_the_genomic_accession=81759548, orientation='minus', exon_count=40, OMIM=603143, no_hdr0=''),
  8500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8500, CurrentID=0, Status='live', Symbol='PPFIA1', Aliases='LIP.1, LIP1, LIPRIN', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 1', other_designations='LAR-interacting protein 1|LIP-1|Liprin-alpha1|PTPRF-interacting protein alpha-1|protein tyrosine phosphatase receptor type f polypeptide-interacting protein alpha-1', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70270687, end_position_on_the_genomic_accession=70384501, orientation='plus', exon_count=35, OMIM=611054, no_hdr0=''),
  8501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8501, CurrentID=0, Status='live', Symbol='SLC43A1', Aliases='LAT3, PB39, POV1, R00504', description='solute carrier family 43 (amino acid system L transporter), member 1', other_designations='L-type amino acid transporter 3|prostate cancer overexpressed gene 1 protein|solute carrier family 43, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57484531, end_position_on_the_genomic_accession=57515719, orientation='minus', exon_count=17, OMIM=603733, no_hdr0=''),
  8502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8502, CurrentID=0, Status='live', Symbol='PKP4', Aliases='p0071', description='plakophilin 4', other_designations='catenin 4', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158456880, end_position_on_the_genomic_accession=158681429, orientation='plus', exon_count=26, OMIM=604276, no_hdr0=''),
  8503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8503, CurrentID=0, Status='live', Symbol='PIK3R3', Aliases='p55, p55-GAMMA, p55PIK', description='phosphoinositide-3-kinase regulatory subunit 3', other_designations='PI3-kinase regulatory subunit gamma|PI3-kinase subunit p55-gamma|PI3K regulatory subunit gamma|phosphatidylinositol 3-kinase 55 kDa regulatory subunit gamma|phosphatidylinositol 3-kinase, regulatory subunit, polypeptide 3 (p55, gamma)|phosphoinositide-3-kinase, regulatory subunit 3 (gamma)|phosphoinositide-3-kinase, regulatory subunit, polypeptide 3 (p55, gamma)|ptdIns-3-kinase regulatory subunit p55-gamma', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46040140, end_position_on_the_genomic_accession=46176495, orientation='minus', exon_count=12, OMIM=606076, no_hdr0=''),
  8504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8504, CurrentID=0, Status='live', Symbol='PEX3', Aliases='PBD10A, TRG18', description='peroxisomal biogenesis factor 3', other_designations='peroxin-3|peroxisomal assembly protein PEX3|transformation-related protein 18', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143450781, end_position_on_the_genomic_accession=143490616, orientation='plus', exon_count=12, OMIM=603164, no_hdr0=''),
  8505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8505, CurrentID=0, Status='live', Symbol='PARG', Aliases='PARG99', description='poly(ADP-ribose) glycohydrolase', other_designations='mitochondrial poly(ADP-ribose) glycohydrolase|poly (ADP-ribose) glycohydrolase|poly(ADP-ribose) glycohydrolase 60 kDa isoform', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49818275, end_position_on_the_genomic_accession=49942027, orientation='minus', exon_count=21, OMIM=603501, no_hdr0=''),
  8506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8506, CurrentID=0, Status='live', Symbol='CNTNAP1', Aliases='CASPR, CNTNAP, NRXN4, P190', description='contactin associated protein 1', other_designations='caspr1|neurexin IV|neurexin-4', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42682614, end_position_on_the_genomic_accession=42699993, orientation='plus', exon_count=25, OMIM=602346, no_hdr0=''),
  8507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8507, CurrentID=0, Status='live', Symbol='ENC1', Aliases='CCL28, ENC-1, KLHL35, KLHL37, NRPB, PIG10, TP53I10', description='ectodermal-neural cortex 1 (with BTB domain)', other_designations='ectodermal-neural cortex 1 (with BTB-like domain)|kelch-like 35|kelch-like family member 37|kelch-like protein 37|nuclear matrix protein NRP/B|nuclear restricted protein, BTB domain-like (brain)|p53-induced gene 10 protein|tumor protein p53 inducible protein 10', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74627406, end_position_on_the_genomic_accession=74641424, orientation='minus', exon_count=4, OMIM=605173, no_hdr0=''),
  8508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8508, CurrentID=0, Status='live', Symbol='NIPSNAP1', Aliases='', description='nipsnap homolog 1 (C. elegans)', other_designations='4-nitrophenylphosphatase domain and non-neuronal SNAP25-like 1', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29554808, end_position_on_the_genomic_accession=29581337, orientation='minus', exon_count=11, OMIM=603249, no_hdr0=''),
  8509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8509, CurrentID=0, Status='live', Symbol='NDST2', Aliases='HSST2, NST2', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 2', other_designations='N-HSST 2|N-heparan sulfate sulfotransferase 2|NDST-2|glucosaminyl N-deacetylase/N-sulfotransferase 2', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73801911, end_position_on_the_genomic_accession=73811831, orientation='minus', exon_count=16, OMIM=603268, no_hdr0=''),
  8510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8510, CurrentID=0, Status='live', Symbol='MMP23B', Aliases='MIFR, MIFR-1, MMP22, MMP23A', description='matrix metallopeptidase 23B', other_designations='MMP-21|MMP-22|MMP-23|femalysin|matrix metalloproteinase 22|matrix metalloproteinase 23B|matrix metalloproteinase in the female reproductive tract|matrix metalloproteinase-21|matrix metalloproteinase-22', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1632181, end_position_on_the_genomic_accession=1634654, orientation='plus', exon_count=8, OMIM=603321, no_hdr0=''),
  8511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8511, CurrentID=0, Status='live', Symbol='MMP23A', Aliases='MIFR, MIFR-1, MMP21', description='matrix metallopeptidase 23A (pseudogene)', other_designations='femalysin|matrix metalloproteinase 21|matrix metalloproteinase 23 pseudogene|matrix metalloproteinase 23A|matrix metalloproteinase in the female reproductive tract', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1699939, end_position_on_the_genomic_accession=1701808, orientation='plus', exon_count=6, OMIM=603320, no_hdr0=''),
  8512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8512, CurrentID=0, Status='live', Symbol='MBL1P', Aliases='COLEC3P1, MBL1P', description='mannose-binding lectin (protein A) 1, pseudogene', other_designations='mannose-binding lectin (protein A) 1, pseudogene 1', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79920178, end_position_on_the_genomic_accession=79923119, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8513, CurrentID=0, Status='live', Symbol='LIPF', Aliases='GL, HGL, HLAL', description='lipase, gastric', other_designations='gastric lipase', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88664389, end_position_on_the_genomic_accession=88678815, orientation='plus', exon_count=11, OMIM=601980, no_hdr0=''),
  8514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8514, CurrentID=0, Status='live', Symbol='KCNAB2', Aliases='AKR6A5, HKvbeta2, HKvbeta2.1, HKvbeta2.2, KCNA2B, KV-BETA-2', description='potassium channel, voltage gated subfamily A regulatory beta subunit 2', other_designations='K(+) channel subunit beta-2|K+ channel beta-2 subunit|potassium channel shaker chain beta 2|potassium voltage-gated channel, shaker-related subfamily, beta member 2', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5992298, end_position_on_the_genomic_accession=6101193, orientation='plus', exon_count=22, OMIM=601142, no_hdr0=''),
  8515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8515, CurrentID=0, Status='live', Symbol='ITGA10', Aliases='PRO827', description='integrin subunit alpha 10', other_designations='integrin alpha 10|integrin, alpha 10', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145891208, end_position_on_the_genomic_accession=145910411, orientation='minus', exon_count=30, OMIM=604042, no_hdr0=''),
  8516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8516, CurrentID=0, Status='live', Symbol='ITGA8', Aliases='', description='integrin subunit alpha 8', other_designations='integrin alpha 8|integrin, alpha 8', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15513949, end_position_on_the_genomic_accession=15720335, orientation='minus', exon_count=31, OMIM=604063, no_hdr0=''),
  8517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8517, CurrentID=0, Status='live', Symbol='IKBKG', Aliases='AMCBX1, FIP-3, FIP3, Fip3p, IKK-gamma, IMD33, IP, IP1, IP2, IPD2, NEMO, ZC2HC9', description='inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase gamma', other_designations='I-kappa-B kinase subunit gamma|IKKAP1|IKKG|IkB kinase gamma subunit|NF-kappa-B essential modifier|NFkappaB essential modulator|ikB kinase subunit gamma|ikB kinase-associated protein 1|incontinentia pigmenti|inhibitor of nuclear factor kappa-B kinase subunit gamma', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154541204, end_position_on_the_genomic_accession=154568573, orientation='plus', exon_count=15, OMIM=300248, no_hdr0=''),
  8518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8518, CurrentID=0, Status='live', Symbol='IKBKAP', Aliases='DYS, ELP1, FD, IKAP, IKI3, TOT1', description='inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase complex-associated protein', other_designations='IKK complex-associated protein|elongator acetyltransferase complex subunit 1|ikappaB kinase complex-associated protein|p150', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108867520, end_position_on_the_genomic_accession=108934328, orientation='minus', exon_count=37, OMIM=603722, no_hdr0=''),
  8519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8519, CurrentID=0, Status='live', Symbol='IFITM1', Aliases='9-27, CD225, DSPA2a, IFI17, LEU13', description='interferon induced transmembrane protein 1', other_designations='dispanin subfamily A member 2a|interferon-induced protein 17|interferon-inducible protein 9-27|leu-13 antigen', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=313991, end_position_on_the_genomic_accession=315272, orientation='plus', exon_count=2, OMIM=604456, no_hdr0=''),
  8520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8520, CurrentID=0, Status='live', Symbol='HAT1', Aliases='KAT1', description='histone acetyltransferase 1', other_designations='', map_location='2q31.2-q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171922425, end_position_on_the_genomic_accession=171983686, orientation='plus', exon_count=13, OMIM=603053, no_hdr0=''),
  8521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8521, CurrentID=0, Status='live', Symbol='GCM1', Aliases='GCMA, hGCMa', description='glial cells missing homolog 1 (Drosophila)', other_designations='GCM motif protein 1', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53126961, end_position_on_the_genomic_accession=53148826, orientation='minus', exon_count=6, OMIM=603715, no_hdr0=''),
  8522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8522, CurrentID=0, Status='live', Symbol='GAS7', Aliases='MLL/GAS7', description='growth arrest specific 7', other_designations='MLL/GAS7 fusion protein|growth arrest-specific 7', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9910609, end_position_on_the_genomic_accession=10198551, orientation='minus', exon_count=19, OMIM=603127, no_hdr0=''),
  139596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139596, CurrentID=0, Status='live', Symbol='UPRT', Aliases='FUR1, UPP', description='uracil phosphoribosyltransferase homolog', other_designations='RP11-311P8.3|UMP pyrophosphorylase|UPRTase|uracil phosphoribosyltransferase (FUR1) homolog', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75274059, end_position_on_the_genomic_accession=75304897, orientation='plus', exon_count=9, OMIM=300656, no_hdr0=''),
  8525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8525, CurrentID=0, Status='live', Symbol='DGKZ', Aliases='DAGK5, DAGK6, DGK-ZETA, hDGKzeta', description='diacylglycerol kinase, zeta', other_designations='DAG kinase zeta|diacylglycerol kinase, zeta 104kDa|diglyceride kinase zeta', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46332905, end_position_on_the_genomic_accession=46380554, orientation='plus', exon_count=36, OMIM=601441, no_hdr0=''),
  8526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8526, CurrentID=0, Status='live', Symbol='DGKE', Aliases='DAGK5, DAGK6, DGK, NPHS7', description='diacylglycerol kinase epsilon', other_designations='DAG kinase epsilon|DGK-epsilon|diacylglycerol kinase, epsilon 64kDa|diglyceride kinase epsilon', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56834099, end_position_on_the_genomic_accession=56869567, orientation='plus', exon_count=15, OMIM=601440, no_hdr0=''),
  8527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8527, CurrentID=0, Status='live', Symbol='DGKD', Aliases='DGKdelta, dgkd-2', description='diacylglycerol kinase delta', other_designations='DAG kinase delta|DGK-delta|diacylglycerol kinase, delta 130kDa|diglyceride kinase delta', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233354507, end_position_on_the_genomic_accession=233472098, orientation='plus', exon_count=44, OMIM=601826, no_hdr0=''),
  8528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8528, CurrentID=0, Status='live', Symbol='DDO', Aliases='DASOX-1, DDO-2, DDO', description='D-aspartate oxidase', other_designations='D-aspartate oxidase, DDO|aspartic oxidase', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110392180, end_position_on_the_genomic_accession=110415550, orientation='minus', exon_count=5, OMIM=124450, no_hdr0=''),
  8529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8529, CurrentID=0, Status='live', Symbol='CYP4F2', Aliases='CPF2', description='cytochrome P450 family 4 subfamily F member 2', other_designations='20-HETE synthase|20-hydroxyeicosatetraenoic acid synthase|CYPIVF2|arachidonic acid omega-hydroxylase|cytochrome P450 4F2|cytochrome P450, family 4, subfamily F, polypeptide 2|cytochrome P450, subfamily IVF, polypeptide 2|cytochrome P450-LTB-omega|leukotriene-B(4) 20-monooxygenase 1|leukotriene-B(4) omega-hydroxylase 1', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15878024, end_position_on_the_genomic_accession=15898120, orientation='minus', exon_count=13, OMIM=604426, no_hdr0=''),
  8530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8530, CurrentID=0, Status='live', Symbol='CST7', Aliases='CMAP', description='cystatin F', other_designations='cystatin F (leukocystatin)|cystatin-7|cystatin-like metastasis-associated protein|leukocystatin', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24949230, end_position_on_the_genomic_accession=24959928, orientation='plus', exon_count=4, OMIM=603253, no_hdr0=''),
  8531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8531, CurrentID=0, Status='live', Symbol='YBX3', Aliases='CSDA, CSDA1, DBPA, ZONAB', description='Y-box binding protein 3', other_designations='DNA-binding protein A|Y box binding protein 3|ZO-1-associated nucleic acid-binding protein|cold shock domain protein A|cold shock domain-containing protein A|cold-shock domain containing A1|cold-shock domain protein A|single-strand DNA-binding protein NF-GMB', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10699077, end_position_on_the_genomic_accession=10723448, orientation='minus', exon_count=11, OMIM=603437, no_hdr0=''),
  8532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8532, CurrentID=0, Status='live', Symbol='CPZ', Aliases='', description='carboxypeptidase Z', other_designations='metallocarboxypeptidase Z', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8592660, end_position_on_the_genomic_accession=8619761, orientation='plus', exon_count=11, OMIM=603105, no_hdr0=''),
  8533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8533, CurrentID=0, Status='live', Symbol='COPS3', Aliases='CSN3, SGN3', description='COP9 signalosome subunit 3', other_designations='COP9 complex subunit 3|COP9 constitutive photomorphogenic homolog subunit 3|JAB1-containing signalosome subunit 3|signalosome subunit 3', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17246624, end_position_on_the_genomic_accession=17281303, orientation='minus', exon_count=13, OMIM=604665, no_hdr0=''),
  8534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8534, CurrentID=0, Status='live', Symbol='CHST1', Aliases='C6ST, GST-1, KS6ST, KSGal6ST, KSST', description='carbohydrate (keratan sulfate Gal-6) sulfotransferase 1', other_designations='carbohydrate (chondroitin 6/keratan) sulfotransferase 1|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45647689, end_position_on_the_genomic_accession=45665656, orientation='minus', exon_count=4, OMIM=603797, no_hdr0=''),
  8535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8535, CurrentID=0, Status='live', Symbol='CBX4', Aliases='NBP16, PC2', description='chromobox 4', other_designations='NS5ATP1-binding protein 16|Pc class 2 homolog|chromobox homolog 4 (Pc class homolog, Drosophila)|chromobox protein homolog 4|chromobox-like protein 4|polycomb 2 homolog', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79833156, end_position_on_the_genomic_accession=79839414, orientation='minus', exon_count=6, OMIM=603079, no_hdr0=''),
  8536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8536, CurrentID=0, Status='live', Symbol='CAMK1', Aliases='CAMKI', description='calcium/calmodulin-dependent protein kinase I', other_designations='caM kinase I alpha|caM-KI|caMKI-alpha', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9757345, end_position_on_the_genomic_accession=9769987, orientation='minus', exon_count=11, OMIM=604998, no_hdr0=''),
  8537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8537, CurrentID=0, Status='live', Symbol='BCAS1', Aliases='AIBC1, NABC1', description='breast carcinoma amplified sequence 1', other_designations='amplified and overexpressed in breast cancer|novel amplified in breast cancer 1', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53943540, end_position_on_the_genomic_accession=54070765, orientation='minus', exon_count=14, OMIM=602968, no_hdr0=''),
  8538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8538, CurrentID=0, Status='live', Symbol='BARX2', Aliases='', description='BARX homeobox 2', other_designations='BarH-like homeobox 2', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129375238, end_position_on_the_genomic_accession=129452279, orientation='plus', exon_count=6, OMIM=604823, no_hdr0=''),
  8539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8539, CurrentID=0, Status='live', Symbol='API5', Aliases='AAC-11, AAC11', description='apoptosis inhibitor 5', other_designations='FIF|antiapoptosis clone 11 protein|cell migration-inducing gene 8 protein|fibroblast growth factor 2-interacting factor 2|migration-inducing protein MIG8', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43311955, end_position_on_the_genomic_accession=43344530, orientation='plus', exon_count=16, OMIM=609774, no_hdr0=''),
  8540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8540, CurrentID=0, Status='live', Symbol='AGPS', Aliases='ADAP-S, ADAS, ADHAPS, ADPS, ALDHPSY', description='alkylglycerone phosphate synthase', other_designations='aging-associated gene 5 protein|aging-associated protein 5|alkyl-DHAP synthase|alkylglycerone-phosphate synthase', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177392743, end_position_on_the_genomic_accession=177543836, orientation='plus', exon_count=23, OMIM=603051, no_hdr0=''),
  8541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8541, CurrentID=0, Status='live', Symbol='PPFIA3', Aliases='LPNA3', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 3', other_designations='PTPRF-interacting protein alpha-3|liprin|liprin-alpha 3|protein tyrosine phosphatase receptor type f polypeptide-interacting protein alpha-3|protein tyrosine phosphatase, receptor type, f polypeptide, alpha 3', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49119389, end_position_on_the_genomic_accession=49151030, orientation='plus', exon_count=30, OMIM=603144, no_hdr0=''),
  8542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8542, CurrentID=0, Status='live', Symbol='APOL1', Aliases='APO-L, APOL, APOL-I, FSGS4', description='apolipoprotein L1', other_designations='apolipoprotein L 1|apolipoprotein L, 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36253071, end_position_on_the_genomic_accession=36267531, orientation='plus', exon_count=8, OMIM=603743, no_hdr0=''),
  8543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8543, CurrentID=0, Status='live', Symbol='LMO4', Aliases='', description='LIM domain only 4', other_designations='LIM domain only protein 4|LIM-only 4 protein|LMO-4|breast tumor autoantigen', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87328468, end_position_on_the_genomic_accession=87348923, orientation='plus', exon_count=6, OMIM=603129, no_hdr0=''),
  8544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8544, CurrentID=0, Status='live', Symbol='PIR', Aliases='', description='pirin', other_designations='pirin (iron-binding nuclear protein)|probable quercetin 2,3-dioxygenase PIR|probable quercetinase', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15384799, end_position_on_the_genomic_accession=15493588, orientation='minus', exon_count=10, OMIM=300931, no_hdr0=''),
  8545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8545, CurrentID=0, Status='live', Symbol='CGGBP1', Aliases='CGGBP, p20-CGGBP', description='CGG triplet repeat binding protein 1', other_designations='20 kDa CGG-binding protein|CGG-binding protein 1|p20-CGG binding protein|p20-CGGBP DNA-binding protein', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88051950, end_position_on_the_genomic_accession=88149866, orientation='minus', exon_count=6, OMIM=603363, no_hdr0=''),
  8546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8546, CurrentID=0, Status='live', Symbol='AP3B1', Aliases='ADTB3, ADTB3A, HPS, HPS2, PE', description='adaptor related protein complex 3, beta 1 subunit', other_designations='AP-3 complex beta-3A subunit|adaptor protein complex AP-3 subunit beta-1|adaptor-related protein complex 3 subunit beta-1|adaptor-related protein complex 3, beta 1 subunit|beta-3A-adaptin|clathrin assembly protein complex 3 beta-1 large chain', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78002326, end_position_on_the_genomic_accession=78294755, orientation='minus', exon_count=29, OMIM=603401, no_hdr0=''),
  8547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8547, CurrentID=0, Status='live', Symbol='FCN3', Aliases='FCNH, HAKA1', description='ficolin 3', other_designations='H-ficolin|collagen/fibrinogen domain-containing lectin 3 p35|collagen/fibrinogen domain-containing protein 3|ficolin (collagen/fibrinogen domain containing) 3 (Hakata antigen)', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27369110, end_position_on_the_genomic_accession=27374825, orientation='minus', exon_count=9, OMIM=604973, no_hdr0=''),
  8548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8548, CurrentID=0, Status='live', Symbol='BLZF1', Aliases='GOLGIN-45, JEM-1, JEM-1s, JEM1', description='basic leucine zipper nuclear factor 1', other_designations='JEM-1short protein|cytoplasmic protein|p45 basic leucine-zipper nuclear factor', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169367956, end_position_on_the_genomic_accession=169396544, orientation='plus', exon_count=8, OMIM=608692, no_hdr0=''),
  8549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8549, CurrentID=0, Status='live', Symbol='LGR5', Aliases='FEX, GPR49, GPR67, GRP49, HG38', description='leucine-rich repeat containing G protein-coupled receptor 5', other_designations='G-protein coupled receptor 49|G-protein coupled receptor 67|G-protein coupled receptor HG38|leucine-rich repeat-containing G protein-coupled receptor 5|orphan G protein-coupled receptor HG38', map_location='12q22-q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71439162, end_position_on_the_genomic_accession=71586310, orientation='plus', exon_count=21, OMIM=606667, no_hdr0=''),
  8550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8550, CurrentID=0, Status='live', Symbol='MAPKAPK5', Aliases='MAPKAP-K5, MK-5, MK5, PRAK', description='mitogen-activated protein kinase-activated protein kinase 5', other_designations='MAPK-activated protein kinase 5|MAPKAP kinase 5|MAPKAPK-5|p38-regulated/activated protein kinase', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111842228, end_position_on_the_genomic_accession=111901619, orientation='plus', exon_count=19, OMIM=606723, no_hdr0=''),
  8551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8551, CurrentID=0, Status='live', Symbol='INE2', Aliases='NCRNA00011', description='inactivation escape 2 (non-protein coding)', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15785716, end_position_on_the_genomic_accession=15787589, orientation='minus', exon_count=1, OMIM=300165, no_hdr0=''),
  8552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8552, CurrentID=0, Status='live', Symbol='INE1', Aliases='NCRNA00010', description='inactivation escape 1 (non-protein coding)', other_designations='', map_location='Xp11.4-p11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47204848, end_position_on_the_genomic_accession=47205861, orientation='plus', exon_count=1, OMIM=300164, no_hdr0=''),
  8553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8553, CurrentID=0, Status='live', Symbol='BHLHE40', Aliases='BHLHB2, DEC1, HLHB2, SHARP-2, STRA13, Stra14', description='basic helix-loop-helix family member e40', other_designations='basic helix-loop-helix domain containing, class B, 2|basic helix-loop-helix family, member e40|class B basic helix-loop-helix protein 2|differentially expressed in chondrocytes 1|differentially expressed in chondrocytes protein 1|differentiated embryo chondrocyte expressed gene 1|enhancer-of-split and hairy-related protein 2|stimulated by retinoic acid gene 13 protein', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4979412, end_position_on_the_genomic_accession=4985181, orientation='plus', exon_count=5, OMIM=604256, no_hdr0=''),
  8554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8554, CurrentID=0, Status='live', Symbol='PIAS1', Aliases='DDXBP1, GBP, GU/RH-II, ZMIZ3', description='protein inhibitor of activated STAT 1', other_designations='AR interacting protein|DEAD/H (Asp-Glu-Ala-Asp/His) box binding protein 1|DEAD/H box-binding protein 1|RNA helicase II-binding protein|gu-binding protein|protein inhibitor of activated STAT protein 1|protein inhibitor of activated STAT, 1|protein inhibitor of activated STAT-1|zinc finger, MIZ-type containing 3', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68054234, end_position_on_the_genomic_accession=68191459, orientation='plus', exon_count=18, OMIM=603566, no_hdr0=''),
  8555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8555, CurrentID=0, Status='live', Symbol='CDC14B', Aliases='CDC14B3, Cdc14B1, Cdc14B2, hCDC14B', description='cell division cycle 14B', other_designations='CDC14 cell division cycle 14 homolog B', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96490218, end_position_on_the_genomic_accession=96619830, orientation='minus', exon_count=26, OMIM=603505, no_hdr0=''),
  8556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8556, CurrentID=0, Status='live', Symbol='CDC14A', Aliases='cdc14, hCDC14', description='cell division cycle 14A', other_designations='CDC10 (cell division cycle 10, S. cerevisiae, homolog)|CDC14 cell division cycle 14 homolog A', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100345042, end_position_on_the_genomic_accession=100520277, orientation='plus', exon_count=22, OMIM=603504, no_hdr0=''),
  8557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8557, CurrentID=0, Status='live', Symbol='TCAP', Aliases='CMD1N, CMH25, LGMD2G, T-cap, TELE, telethonin', description='titin-cap', other_designations='19 kDa sarcomeric protein|limb girdle muscular dystrophy 2G (autosomal recessive)|titin cap protein', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39665346, end_position_on_the_genomic_accession=39666554, orientation='plus', exon_count=2, OMIM=604488, no_hdr0=''),
  8558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8558, CurrentID=0, Status='live', Symbol='CDK10', Aliases='PISSLRE', description='cyclin-dependent kinase 10', other_designations='CDC2-related protein kinase|cell division protein kinase 10|cyclin-dependent kinase (CDC2-like) 10|cyclin-dependent kinase related protein|serine/threonine protein kinase PISSLRE|serine/threonine-protein kinase PISSLRE', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89672247, end_position_on_the_genomic_accession=89696364, orientation='plus', exon_count=21, OMIM=603464, no_hdr0=''),
  8559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8559, CurrentID=0, Status='live', Symbol='PRPF18', Aliases='PRP18, hPrp18', description='pre-mRNA processing factor 18', other_designations='PRP18 homolog|PRP18 pre-mRNA processing factor 18 homolog|PRP18 pre-mRNA processing factor 18 homolog(PRPF18)', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13586939, end_position_on_the_genomic_accession=13630868, orientation='plus', exon_count=12, OMIM=604993, no_hdr0=''),
  8560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8560, CurrentID=0, Status='live', Symbol='DEGS1', Aliases='DEGS, DEGS-1, DES1, Des-1, FADS7, MIG15, MLD', description='delta(4)-desaturase, sphingolipid 1', other_designations='cell migration-inducing gene 15 protein|degenerative spermatocyte homolog 1, lipid desaturase|degenerative spermatocyte homolog, lipid desaturase|dihydroceramide desaturase 1|membrane fatty acid (lipid) desaturase|membrane lipid desaturase|migration-inducing gene 15 protein|sphingolipid delta 4 desaturase|sphingolipid delta(4)-desaturase 1', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224183208, end_position_on_the_genomic_accession=224193441, orientation='plus', exon_count=5, OMIM=615843, no_hdr0=''),
  401777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401777, CurrentID=0, Status='live', Symbol='LOC401777', Aliases='', description='serine/arginine-rich splicing factor 10 pseudogene', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57629821, end_position_on_the_genomic_accession=57630492, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8562, CurrentID=0, Status='live', Symbol='DENR', Aliases='DRP, DRP1, SMAP-3', description='density-regulated protein', other_designations='smooth muscle cell associated protein-3|smooth muscle cell-associated protein 3', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122752821, end_position_on_the_genomic_accession=122771064, orientation='plus', exon_count=8, OMIM=604550, no_hdr0=''),
  8563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8563, CurrentID=0, Status='live', Symbol='THOC5', Aliases='C22orf19, Fmip, PK1.3, fSAP79', description='THO complex 5', other_designations='Fms-interacting protein|NF2/meningioma region protein pK1.3|PP39.2|functional spliceosome-associated protein 79|hTREX90|placental protein 39.2', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29508167, end_position_on_the_genomic_accession=29553760, orientation='minus', exon_count=24, OMIM=612733, no_hdr0=''),
  8564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8564, CurrentID=0, Status='live', Symbol='KMO', Aliases='dJ317G22.1', description='kynurenine 3-monooxygenase (kynurenine 3-hydroxylase)', other_designations='kynurenine 3-hydroxylase', map_location='1q42-q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241531889, end_position_on_the_genomic_accession=241595647, orientation='plus', exon_count=16, OMIM=603538, no_hdr0=''),
  8565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8565, CurrentID=0, Status='live', Symbol='YARS', Aliases='CMTDIC, TYRRS, YRS, YTS', description='tyrosyl-tRNA synthetase', other_designations='tyrosine tRNA ligase 1, cytoplasmic|tyrosyl--tRNA ligase|tyrosyl-tRNA synthetase, cytoplasmic', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32775239, end_position_on_the_genomic_accession=32818032, orientation='minus', exon_count=14, OMIM=603623, no_hdr0=''),
  8566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8566, CurrentID=0, Status='live', Symbol='PDXK', Aliases='C21orf124, C21orf97, HEL-S-1a, PKH, PNK, PRED79', description='pyridoxal (pyridoxine, vitamin B6) kinase', other_designations='epididymis secretory sperm binding protein Li 1a|pyridoxamine kinase|pyridoxine kinase|vitamin B6 kinase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43719097, end_position_on_the_genomic_accession=43762307, orientation='plus', exon_count=18, OMIM=179020, no_hdr0=''),
  8567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8567, CurrentID=0, Status='live', Symbol='MADD', Aliases='DENN, IG20, RAB3GEP', description='MAP-kinase activating death domain', other_designations='Rab3 GDP/GTP exchange factor|differentially expressed in normal and neoplastic cells|insulinoma glucagonoma clone 20', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47269376, end_position_on_the_genomic_accession=47330031, orientation='plus', exon_count=43, OMIM=603584, no_hdr0=''),
  8568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8568, CurrentID=0, Status='live', Symbol='RRP1', Aliases='D21S2056E, NNP-1, NOP52A, RRP1', description='ribosomal RNA processing 1', other_designations='Nnp1 homolog, nucleolar protein|RRP1-like protein|novel nuclear protein 1|nucleolar protein Nop52|protein NNP-1|ribosomal RNA processing 1 homolog', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43789535, end_position_on_the_genomic_accession=43804102, orientation='plus', exon_count=15, OMIM=610653, no_hdr0=''),
  8569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8569, CurrentID=0, Status='live', Symbol='MKNK1', Aliases='MNK1', description='MAP kinase interacting serine/threonine kinase 1', other_designations='MAP kinase signal-integrating kinase 1|MAPK signal-integrating kinase 1', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46557407, end_position_on_the_genomic_accession=46604294, orientation='minus', exon_count=20, OMIM=606724, no_hdr0=''),
  8570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8570, CurrentID=0, Status='live', Symbol='KHSRP', Aliases='FBP2, FUBP2, KSRP', description='KH-type splicing regulatory protein', other_designations='FUSE binding protein 2|FUSE-binding protein 2|KH type-splicing regulatory protein|p75', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6413104, end_position_on_the_genomic_accession=6424811, orientation='minus', exon_count=20, OMIM=603445, no_hdr0=''),
  8572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8572, CurrentID=0, Status='live', Symbol='PDLIM4', Aliases='RIL', description='PDZ and LIM domain 4', other_designations='LIM domain protein|LIM protein RIL|enigma homolog|reversion-induced LIM protein', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132257658, end_position_on_the_genomic_accession=132273454, orientation='plus', exon_count=7, OMIM=603422, no_hdr0=''),
  8573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8573, CurrentID=0, Status='live', Symbol='CASK', Aliases='CAGH39, CAMGUK, CMG, FGS4, LIN2, MICPCH, MRXSNA, TNRC8', description='calcium/calmodulin-dependent serine protein kinase (MAGUK family)', other_designations='calcium/calmodulin-dependent serin protein kinase|calcium/calmodulin-dependent serine protein kinase membrane-associated guanylate kinase|hCASK|protein lin-2 homolog|trinucleotide repeat containing 8', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41514934, end_position_on_the_genomic_accession=41923078, orientation='minus', exon_count=28, OMIM=300172, no_hdr0=''),
  8574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8574, CurrentID=0, Status='live', Symbol='AKR7A2', Aliases='AFAR, AFAR1, AFB1-AR1, AKR7', description='aldo-keto reductase family 7, member A2', other_designations='AFB1 aldehyde reductase 1|AFB1-AR 1|HEL-S-166mP|SSA reductase|aflatoxin aldehyde reductase|aflatoxin beta1 aldehyde reductase|aldo-keto reductase family 7, member A2 (aflatoxin aldehyde reductase)|aldoketoreductase 7|epididymis secretory sperm binding protein Li 166mP|succinic semialdehyde reductase', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19302708, end_position_on_the_genomic_accession=19312146, orientation='minus', exon_count=7, OMIM=603418, no_hdr0=''),
  8575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8575, CurrentID=0, Status='live', Symbol='PRKRA', Aliases='DYT16, HSD14, PACT, RAX', description='protein kinase, interferon-inducible double stranded RNA dependent activator', other_designations='PKR-associated protein X|PKR-associating protein X|interferon-inducible double stranded RNA-dependent protein kinase activator A|protein activator of the interferon-induced protein kinase|protein kinase, interferon-inducible double stranded RNA-dependent activator|protein kinase, interferon-inducible double-stranded RNA-dependent activator', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178431414, end_position_on_the_genomic_accession=178451231, orientation='minus', exon_count=10, OMIM=603424, no_hdr0=''),
  8576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8576, CurrentID=0, Status='live', Symbol='STK16', Aliases='KRCT, MPSK, PKL12, TSF1', description='serine/threonine kinase 16', other_designations='TGF-beta-stimulated factor 1|TSF-1|hPSK|myristoylated and palmitoylated serine/threonine-protein kinase|protein kinase PKL12|protein kinase expressed in day 12 fetal liver|tyrosine-protein kinase STK16', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219245469, end_position_on_the_genomic_accession=219250337, orientation='plus', exon_count=8, OMIM=604719, no_hdr0=''),
  8577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8577, CurrentID=0, Status='live', Symbol='TMEFF1', Aliases='C9orf2, CT120.1, H7365, TR-1', description='transmembrane protein with EGF-like and two follistatin-like domains 1', other_designations='cancer/testis antigen family 120, member 1|transmembrane protein with EGF-like and one follistatin-like domain', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100473238, end_position_on_the_genomic_accession=100577636, orientation='plus', exon_count=10, OMIM=603421, no_hdr0=''),
  8578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8578, CurrentID=0, Status='live', Symbol='SCARF1', Aliases='SREC, SREC-I, SREC1', description='scavenger receptor class F member 1', other_designations='acetyl LDL receptor|scavenger receptor class F, member 1|scavenger receptor expressed by endothelial cells 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1633858, end_position_on_the_genomic_accession=1645789, orientation='minus', exon_count=11, OMIM=607873, no_hdr0=''),
  285419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285419, CurrentID=0, Status='live', Symbol='LINC01091', Aliases='', description='long intergenic non-protein coding RNA 1091', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123652785, end_position_on_the_genomic_accession=123930363, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  8580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8580, CurrentID=0, Status='live', Symbol='OR8C1P', Aliases='OR11-175, OR8C3P, OR8C4P, OR912-106, OR912-45', description='olfactory receptor family 8 subfamily C member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily C, member 1 pseudogene|olfactory receptor, family 8, subfamily C, member 3 pseudogene|olfactory receptor, family 8, subfamily C, member 4 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124376941, end_position_on_the_genomic_accession=124377615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8581, CurrentID=0, Status='live', Symbol='LY6D', Aliases='E48, Ly-6D', description='lymphocyte antigen 6 complex, locus D', other_designations='e48 antigen', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142784882, end_position_on_the_genomic_accession=142786592, orientation='minus', exon_count=3, OMIM=606204, no_hdr0=''),
  8582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8582, CurrentID=0, Status='live', Symbol='OR8B1P', Aliases='OR11-561, OR8B11P', description='olfactory receptor family 8 subfamily B member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 1 pseudogene|olfactory receptor, family 8, subfamily B, member 11 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124365419, end_position_on_the_genomic_accession=124366391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8583, CurrentID=0, Status='live', Symbol='OR7E41P', Aliases='OR11-20, OR7E50P, OR7E95P, OR7F6P, OR8-126, hg84', description='olfactory receptor family 7 subfamily E member 41 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 41 pseudogene|olfactory receptor, family 7, subfamily E, member 50 pseudogene|olfactory receptor, family 7, subfamily E, member 95 pseudogene|olfactory receptor, family 7, subfamily F, member 6 pseudogene', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14987786, end_position_on_the_genomic_accession=14988428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8585, CurrentID=0, Status='live', Symbol='OR7E4P', Aliases='OR11-11a, OR7E9P, OR7F4P, OST289', description='olfactory receptor family 7 subfamily E member 4 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 4 pseudogene|olfactory receptor, family 7, subfamily E, member 9 pseudogene|olfactory receptor, family 7, subfamily F, member 4 pseudogene', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71620100, end_position_on_the_genomic_accession=71621022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8586, CurrentID=0, Status='live', Symbol='OR7E87P', Aliases='OR11-261, OR11-9, OR7E3P, OR7F3P', description='olfactory receptor family 7 subfamily E member 87 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 3 pseudogene|olfactory receptor, family 7, subfamily E, member 87 pseudogene', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71593658, end_position_on_the_genomic_accession=71594305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8587, CurrentID=0, Status='live', Symbol='OR7E2P', Aliases='OR11-6, OR7E143P, OR7E51P, OR7E52P, OR7F2P, OST245, hg94', description='olfactory receptor family 7 subfamily E member 2 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 143 pseudogene|olfactory receptor, family 7, subfamily E, member 2 pseudogene|olfactory receptor, family 7, subfamily E, member 51 pseudogene|olfactory receptor, family 7, subfamily E, member 52 pseudogene|olfactory receptor, family 7, subfamily F, member 2 pseudogene', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86856931, end_position_on_the_genomic_accession=86857954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8588, CurrentID=0, Status='live', Symbol='OR7E15P', Aliases='OR11-392, OR7E42P, OR7E80P, OST001', description='olfactory receptor family 7 subfamily E member 15 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 15 pseudogene|olfactory receptor, family 7, subfamily E, member 42 pseudogene|olfactory receptor, family 7, subfamily E, member 80 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12696370, end_position_on_the_genomic_accession=12697025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401805, CurrentID=0, Status='live', Symbol='LOC401805', Aliases='', description='ring finger protein 145 pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20626822, end_position_on_the_genomic_accession=20628653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8590, CurrentID=0, Status='live', Symbol='OR6A2', Aliases='I7, OR11-55, OR6A1P, OR6A2', description='olfactory receptor family 6 subfamily A member 2', other_designations='hI7 olfactory receptor|hP2 olfactory receptor|olfactory receptor 11-55|olfactory receptor 6A1|olfactory receptor OR11-83|olfactory receptor, family 6, subfamily A, member 1|olfactory receptor, family 6, subfamily A, member 2 pseudogene|seven transmembrane helix receptor', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6794525, end_position_on_the_genomic_accession=6795908, orientation='minus', exon_count=1, OMIM=608495, no_hdr0=''),
  8591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8591, CurrentID=0, Status='live', Symbol='OR5G1P', Aliases='OR11-104, OR5G2P, OR93, OR93HUM', description='olfactory receptor family 5 subfamily G member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily G, member 1 pseudogene|olfactory receptor, family 5, subfamily G, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56775284, end_position_on_the_genomic_accession=56776206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271512, CurrentID=0, Status='live', Symbol='RPL36AP32', Aliases='RPL36A_12_887', description='ribosomal protein L36a pseudogene 32', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28810103, end_position_on_the_genomic_accession=28810414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8594, CurrentID=0, Status='live', Symbol='OR5D3P', Aliases='OR11-8b, OR11-8c, OR5D3, OR5D4', description='olfactory receptor family 5 subfamily D member 3 pseudogene', other_designations='olfactory receptor OR11-140|olfactory receptor, family 5, subfamily D, member 3 pseudogene|olfactory receptor, family 5, subfamily D, member 4', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55726414, end_position_on_the_genomic_accession=55727062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8595, CurrentID=0, Status='live', Symbol='OR5D2P', Aliases='OR11-7a, OR18-17, OR18-42, OR18-43, OR18-44, OR18-79, OR5D10P, OR5D11, OR5D11P, OR5D12, OR5D12P, OR5D1P, OR5D5P, OR5D6P, OR5D7P, OR5D8P, OR5D9P, OR8-125, OR8-127, OR912-47, OR912-91, OR912-94, R5D9P', description='olfactory receptor family 5 subfamily D member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily D, member 1 pseudogene|olfactory receptor, family 5, subfamily D, member 10 pseudogene|olfactory receptor, family 5, subfamily D, member 11 pseudogene|olfactory receptor, family 5, subfamily D, member 12 pseudogene|olfactory receptor, family 5, subfamily D, member 2 pseudogene|olfactory receptor, family 5, subfamily D, member 5 pseudogene|olfactory receptor, family 5, subfamily D, member 6 pseudogene|olfactory receptor, family 5, subfamily D, member 7 pseudogene|olfactory receptor, family 5, subfamily D, member 8 pseudogene|olfactory receptor, family 5, subfamily D, member 9 pseudogene', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55714997, end_position_on_the_genomic_accession=55715633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8596, CurrentID=0, Status='live', Symbol='OR4A1P', Aliases='OR11-30, OR4A20P', description='olfactory receptor family 4 subfamily A member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 1 pseudogene|olfactory receptor, family 4, subfamily A, member 20 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49898281, end_position_on_the_genomic_accession=49899186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271513, CurrentID=0, Status='live', Symbol='RPL36AP34', Aliases='RPL36A_14_970', description='ribosomal protein L36a pseudogene 34', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27210327, end_position_on_the_genomic_accession=27210626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8600, CurrentID=0, Status='live', Symbol='TNFSF11', Aliases='CD254, ODF, OPGL, OPTB2, RANKL, TRANCE, hRANKL2, sOdf', description='tumor necrosis factor superfamily member 11', other_designations='TNF-related activation-induced cytokine|osteoclast differentiation factor|osteoprotegerin ligand|receptor activator of nuclear factor kappa B ligand|receptor activator of nuclear factor kappa-B ligand|tumor necrosis factor (ligand) superfamily member 11|tumor necrosis factor (ligand) superfamily, member 11', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42562736, end_position_on_the_genomic_accession=42608013, orientation='plus', exon_count=9, OMIM=602642, no_hdr0=''),
  8601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8601, CurrentID=0, Status='live', Symbol='RGS20', Aliases='RGSZ1, ZGAP1, g(z)GAP, gz-GAP', description='regulator of G-protein signaling 20', other_designations='gz-selective GTPase-activating protein|regulator of G-protein signaling 20 variant 2|regulator of G-protein signaling Z1|regulator of G-protein signalling 20|regulator of Gz-selective protein signaling 1', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53851808, end_position_on_the_genomic_accession=53959304, orientation='plus', exon_count=9, OMIM=607193, no_hdr0=''),
  8602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8602, CurrentID=0, Status='live', Symbol='NOP14', Aliases='C4orf9, NOL14, RES4-25, RES425, UTP2', description='NOP14 nucleolar protein', other_designations='NOP14 nucleolar protein homolog|probable nucleolar complex protein 14', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2937936, end_position_on_the_genomic_accession=2963506, orientation='minus', exon_count=19, OMIM=611526, no_hdr0=''),
  8603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8603, CurrentID=0, Status='live', Symbol='FAM193A', Aliases='C4orf8, RES4-22', description='family with sequence similarity 193 member A', other_designations='family with sequence similarity 193, member A', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2596288, end_position_on_the_genomic_accession=2732575, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  8604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8604, CurrentID=0, Status='live', Symbol='SLC25A12', Aliases='AGC1, ARALAR', description='solute carrier family 25 (aspartate/glutamate carrier), member 12', other_designations='araceli hiperlarga|calcium binding mitochondrial carrier superfamily member Aralar1|mitochondrial aspartate glutamate carrier 1|solute carrier family 25 (mitochondrial carrier, Aralar), member 12', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171783405, end_position_on_the_genomic_accession=171894306, orientation='minus', exon_count=20, OMIM=603667, no_hdr0=''),
  8605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8605, CurrentID=0, Status='live', Symbol='PLA2G4C', Aliases='CPLA2-gamma', description='phospholipase A2 group IVC', other_designations='phospholipase A2, group IVC (cytosolic, calcium-independent)', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48047843, end_position_on_the_genomic_accession=48110853, orientation='minus', exon_count=20, OMIM=603602, no_hdr0=''),
  401822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401822, CurrentID=0, Status='live', Symbol='FDPSP9', Aliases='', description='farnesyl diphosphate synthase pseudogene 9', other_designations='farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79944444, end_position_on_the_genomic_accession=79945632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8607, CurrentID=0, Status='live', Symbol='RUVBL1', Aliases='ECP54, INO80H, NMP238, PONTIN, Pontin52, RVB1, TIH1, TIP49, TIP49A', description='RuvB-like AAA ATPase 1', other_designations='49 kDa TATA box-binding protein-interacting protein|49 kDa TBP-interacting protein|54 kDa erythrocyte cytosolic protein|ECP-54|INO80 complex subunit H|NMP 238|RuvB (E coli homolog)-like 1|TAP54-alpha|TATA binding protein interacting protein 49 kDa|TIP60-associated protein 54-alpha|nuclear matrix protein 238|pontin 52', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128064877, end_position_on_the_genomic_accession=128153909, orientation='minus', exon_count=14, OMIM=603449, no_hdr0=''),
  8608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8608, CurrentID=0, Status='live', Symbol='RDH16', Aliases='RODH-4, SDR9C8', description='retinol dehydrogenase 16 (all-trans)', other_designations='microsomal NAD(+)-dependent retinol dehydrogenase 4|microsomal NAD+-dependent retinol dehydrogenase 4|retinol dehydrogenase 16 (all-trans and 13-cis)|short chain dehydrogenase/reductase family 9C member 8|short chain dehydrogenase/reductase family 9C, member 8|sterol/retinol dehydrogenase', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56951431, end_position_on_the_genomic_accession=56959374, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8609, CurrentID=0, Status='live', Symbol='KLF7', Aliases='UKLF', description='Kruppel-like factor 7 (ubiquitous)', other_designations='ubiquitous Kruppel-like transcription factor', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207074137, end_position_on_the_genomic_accession=207173851, orientation='minus', exon_count=13, OMIM=604865, no_hdr0=''),
  8611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8611, CurrentID=0, Status='live', Symbol='PLPP1', Aliases='LLP1a, LPP1, PAP-2a, PAP2, PPAP2A', description='phospholipid phosphatase 1', other_designations='lipid phosphate phosphohydrolase 1a|phosphatidate phosphohydrolase type 2a|phosphatidic acid phosphatase 2a|phosphatidic acid phosphatase type 2A|phosphatidic acid phosphohydrolase type 2a|type-2 phosphatidic acid phosphatase alpha', map_location='5q11', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55424842, end_position_on_the_genomic_accession=55535370, orientation='minus', exon_count=7, OMIM=607124, no_hdr0=''),
  8612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8612, CurrentID=0, Status='live', Symbol='PLPP2', Aliases='LPP2, PAP-2c, PAP2-g, PPAP2C', description='phospholipid phosphatase 2', other_designations='PAP2-gamma|PAP2c|phosphatidate phosphohydrolase type 2c|phosphatidic acid phosphatase 2c|phosphatidic acid phosphatase type 2C|phosphatidic acid phosphohydrolase type 2c|type-2 phosphatidic acid phosphatase-gamma', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=281040, end_position_on_the_genomic_accession=291435, orientation='minus', exon_count=7, OMIM=607126, no_hdr0=''),
  8613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8613, CurrentID=0, Status='live', Symbol='PLPP3', Aliases='Dri42, LPP3, PAP2B, PPAP2B, VCIP', description='phospholipid phosphatase 3', other_designations='PAP2 beta|phosphatidate phosphohydrolase type 2b|phosphatidic acid phosphatase type 2B|type-2 phosphatidic acid phosphatase-beta|vascular endothelial growth factor and type I collagen inducible', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56494747, end_position_on_the_genomic_accession=56579584, orientation='minus', exon_count=6, OMIM=607125, no_hdr0=''),
  8614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8614, CurrentID=0, Status='live', Symbol='STC2', Aliases='STC-2, STCRP', description='stanniocalcin 2', other_designations='STC-related protein|stanniocalcin-related protein', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173314723, end_position_on_the_genomic_accession=173329503, orientation='minus', exon_count=4, OMIM=603665, no_hdr0=''),
  8615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8615, CurrentID=0, Status='live', Symbol='USO1', Aliases='P115, TAP, VDP', description='USO1 vesicle transport factor', other_designations='USO1 vesicle docking protein homolog|transcytosis associated protein|vesicle docking protein p115', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75724522, end_position_on_the_genomic_accession=75814289, orientation='plus', exon_count=26, OMIM=603344, no_hdr0=''),
  100271516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271516, CurrentID=0, Status='live', Symbol='RPL37P21', Aliases='RPL37_11_1318', description='ribosomal protein L37 pseudogene 21', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68331332, end_position_on_the_genomic_accession=68331613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8618, CurrentID=0, Status='live', Symbol='CADPS', Aliases='CADPS1, CAPS, CAPS1', description='Ca++-dependent secretion activator', other_designations='CAPS-1|Ca2+-dependent activator protein for secretion|Ca2+-dependent secretion activator|Ca2+-regulated cytoskeletal protein|calcium-dependent activator protein for secretion 1', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62398346, end_position_on_the_genomic_accession=62875389, orientation='minus', exon_count=37, OMIM=604667, no_hdr0=''),
  8620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8620, CurrentID=0, Status='live', Symbol='NPFF', Aliases='FMRFAL', description='neuropeptide FF-amide peptide precursor', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53506690, end_position_on_the_genomic_accession=53507638, orientation='minus', exon_count=3, OMIM=604643, no_hdr0=''),
  8621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8621, CurrentID=0, Status='live', Symbol='CDK13', Aliases='CDC2L, CDC2L5, CHED, hCDK13', description='cyclin-dependent kinase 13', other_designations='CDC2-related protein kinase 5|cell division cycle 2-like protein kinase 5|cell division protein kinase 13|cholinesterase-related cell division controller', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39949838, end_position_on_the_genomic_accession=40097134, orientation='plus', exon_count=16, OMIM=603309, no_hdr0=''),
  8622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8622, CurrentID=0, Status='live', Symbol='PDE8B', Aliases='ADSD, PPNAD3', description='phosphodiesterase 8B', other_designations="3',5' cyclic nucleotide phosphodiesterase 8B|cell proliferation-inducing gene 22 protein|hsPDE8B", map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77180262, end_position_on_the_genomic_accession=77428256, orientation='plus', exon_count=26, OMIM=603390, no_hdr0=''),
  8623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8623, CurrentID=0, Status='live', Symbol='ASMTL', Aliases='ASMTLXY, ASTML, ASMTL', description='acetylserotonin O-methyltransferase-like', other_designations='acetylserotonin N-methyltransferase-like', map_location='Xp22.3; Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1403139, end_position_on_the_genomic_accession=1453795, orientation='minus', exon_count=15, OMIM='300162,400011', no_hdr0=''),
  8624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8624, CurrentID=0, Status='live', Symbol='PSMG1', Aliases='C21LRP, DSCR2, LRPC21, PAC-1, PAC1', description='proteasome (prosome, macropain) assembly chaperone 1', other_designations='Down syndrome critical region gene 2|Down syndrome critical region protein 2|chromosome 21 leucine-rich protein|leucine rich protein C21-LRP|proteasome assembling chaperone 1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39175446, end_position_on_the_genomic_accession=39183514, orientation='minus', exon_count=7, OMIM=605296, no_hdr0=''),
  8625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8625, CurrentID=0, Status='live', Symbol='RFXANK', Aliases='ANKRA1, BLS, F14150_1, RFX-B', description='regulatory factor X associated ankyrin containing protein', other_designations='RFX-Bdelta4|ankyrin repeat family A protein 1|ankyrin repeat-containing regulatory factor X-associated protein|regulatory factor X subunit B|regulatory factor X-associated ankyrin-containing protein', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19192199, end_position_on_the_genomic_accession=19201869, orientation='plus', exon_count=12, OMIM=603200, no_hdr0=''),
  8626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8626, CurrentID=0, Status='live', Symbol='TP63', Aliases='AIS, B(p51A), B(p51B), EEC3, KET, LMS, NBP, OFC8, RHS, SHFM4, TP53CP, TP53L, TP73L, p40, p51, p53CP, p63, p73H, p73L', description='tumor protein p63', other_designations='CUSP|amplified in squamous cell carcinoma|chronic ulcerative stomatitis protein|keratinocyte transcription factor KET|transformation-related protein 63|tumor protein p53-competing protein|tumor protein p63 deltaN isoform delta', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189566861, end_position_on_the_genomic_accession=189897279, orientation='plus', exon_count=20, OMIM=603273, no_hdr0=''),
  8629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8629, CurrentID=0, Status='live', Symbol='JRK', Aliases='JH8, jerky', description='Jrk helix-turn-helix protein', other_designations='Jrk homolog|jerky homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142651501, end_position_on_the_genomic_accession=142670132, orientation='minus', exon_count=7, OMIM=603210, no_hdr0=''),
  8630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8630, CurrentID=0, Status='live', Symbol='HSD17B6', Aliases='HSE, RODH, SDR9C6', description='hydroxysteroid (17-beta) dehydrogenase 6', other_designations='17-beta-HSD 6|3(alpha->beta)-hydroxysteroid epimerase|3(alpha->beta)-hydroxysteroid epimerasel|3-alpha->beta-HSE|3-alpha->beta-hydroxysteroid epimerase|3-hydroxysteroid epimerase|NAD+ -dependent 3 alpha-hydroxysteroid dehydrogenase 3-hydroxysteroid epimerase|hydroxysteroid (17-beta) dehydrogenase 6 homolog|oxidative 3-alpha hydroxysteroid dehydrogenase|oxidative 3-alpha-hydroxysteroid-dehydrogenase|oxidoreductase|retinol dehydrogenase|short chain dehydrogenase/reductase family 9C member 6|short chain dehydrogenase/reductase family 9C, member 6', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56752449, end_position_on_the_genomic_accession=56787790, orientation='plus', exon_count=7, OMIM=606623, no_hdr0=''),
  8631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8631, CurrentID=0, Status='live', Symbol='SKAP1', Aliases='HEL-S-81p, SCAP1, SKAP55', description='src kinase associated phosphoprotein 1', other_designations='SKAP-55|epididymis secretory sperm binding protein Li 81p|pp55|src family associated phosphoprotein 1|src family-associated phosphoprotein 1|src kinase-associated phosphoprotein of 55 kDa', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48133438, end_position_on_the_genomic_accession=48430235, orientation='minus', exon_count=19, OMIM=604969, no_hdr0=''),
  8632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8632, CurrentID=0, Status='live', Symbol='DNAH17', Aliases='DNAHL1, DNEL2', description='dynein, axonemal, heavy chain 17', other_designations='axonemal beta dynein heavy chain 17|axonemal dynein heavy chain|axonemal dynein heavy chain-like protein 1|ciliary dynein heavy chain 17|ciliary dynein heavy chain-like protein 1|dynein light chain 2, axonemal|dynein, axonemal, heavy chain like 1|dynein, axonemal, heavy like 1|dynein, axonemal, heavy polypeptide 17', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78423697, end_position_on_the_genomic_accession=78577432, orientation='minus', exon_count=82, OMIM=610063, no_hdr0=''),
  8633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8633, CurrentID=0, Status='live', Symbol='UNC5C', Aliases='UNC5H3', description='unc-5 netrin receptor C', other_designations='protein unc-5 homolog 3|protein unc-5 homolog C|unc-5 homolog 3|unc-5 homolog C|unc5 (C.elegans homolog) c', map_location='4q21-q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95162504, end_position_on_the_genomic_accession=95549210, orientation='minus', exon_count=17, OMIM=603610, no_hdr0=''),
  8634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8634, CurrentID=0, Status='live', Symbol='RTCA', Aliases='RPC, RTC1, RTCD1', description="RNA 3'-terminal phosphate cyclase", other_designations='RNA terminal phosphate cyclase domain 1|RNA terminal phosphate cyclase domain-containing protein 1|RTC domain-containing protein 1', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100266158, end_position_on_the_genomic_accession=100292769, orientation='plus', exon_count=12, OMIM=611286, no_hdr0=''),
  8635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8635, CurrentID=0, Status='live', Symbol='RNASET2', Aliases='RNASE6PL, bA514O12.3', description='ribonuclease T2', other_designations='ribonuclease 6', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166929516, end_position_on_the_genomic_accession=166956589, orientation='minus', exon_count=9, OMIM=612944, no_hdr0=''),
  8636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8636, CurrentID=0, Status='live', Symbol='SSNA1', Aliases='N14, NA-14, NA14', description='Sjogren syndrome nuclear autoantigen 1', other_designations="Sjogren's syndrome nuclear autoantigen 1|nuclear autoantigen of 14 kDa", map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137188602, end_position_on_the_genomic_accession=137190370, orientation='plus', exon_count=3, OMIM=610882, no_hdr0=''),
  8637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8637, CurrentID=0, Status='live', Symbol='EIF4EBP3', Aliases='4E-BP3, 4EBP3', description='eukaryotic translation initiation factor 4E binding protein 3', other_designations='eIF4E-binding protein 3|eukaryotic initiation factor 4E-binding protein 3', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140547666, end_position_on_the_genomic_accession=140549578, orientation='plus', exon_count=3, OMIM=603483, no_hdr0=''),
  8638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8638, CurrentID=0, Status='live', Symbol='OASL', Aliases='OASLd, TRIP-14, TRIP14, p59 OASL, p59-OASL, p59OASL', description="2'-5'-oligoadenylate synthetase-like", other_designations="2'-5'-OAS-RP|2'-5'-OAS-related protein|59 kDa 2'-5'-oligoadenylate synthase-like protein|59 kDa 2'-5'-oligoadenylate synthetase-like protein|TR-interacting protein 14|thyroid receptor-interacting protein 14", map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121020292, end_position_on_the_genomic_accession=121039242, orientation='minus', exon_count=6, OMIM=603281, no_hdr0=''),
  8639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8639, CurrentID=0, Status='live', Symbol='AOC3', Aliases='HPAO, SSAO, VAP-1, VAP1', description='amine oxidase, copper containing 3', other_designations='amine oxidase, copper containing 3 (vascular adhesion protein 1)|copper amine oxidase|placenta copper monamine oxidase|semicarbazide-sensitive amine oxidase', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42851123, end_position_on_the_genomic_accession=42858130, orientation='plus', exon_count=6, OMIM=603735, no_hdr0=''),
  8641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8641, CurrentID=0, Status='live', Symbol='PCDHGB4', Aliases='CDH20, FIB2, PCDH-GAMMA-B4', description='protocadherin gamma subfamily B, 4', other_designations='cadherin 20|cadherin-20|fibroblast cadherin FIB2|fibroblast cadherin-2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141387885, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=603058, no_hdr0=''),
  8642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8642, CurrentID=0, Status='live', Symbol='DCHS1', Aliases='CDH19, CDH25, CDHR6, FIB1, PCDH16, VMLDS1', description='dachsous cadherin-related 1', other_designations='cadherin-19|cadherin-25|cadherin-related family member 6|fibroblast cadherin 1|fibroblast cadherin FIB1|protein dachsous homolog 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6621324, end_position_on_the_genomic_accession=6655849, orientation='minus', exon_count=21, OMIM=603057, no_hdr0=''),
  8643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8643, CurrentID=0, Status='live', Symbol='PTCH2', Aliases='PTC2', description='patched 2', other_designations='patched homolog 2', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44819844, end_position_on_the_genomic_accession=44842944, orientation='minus', exon_count=23, OMIM=603673, no_hdr0=''),
  8644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8644, CurrentID=0, Status='live', Symbol='AKR1C3', Aliases='DD3, DDX, HA1753, HAKRB, HAKRe, HSD17B5, PGFS, hluPGFS', description='aldo-keto reductase family 1, member C3', other_designations='3-alpha hydroxysteroid dehydrogenase, type II|3-alpha-HSD type II, brain|aldo-keto reductase family 1, member C3 (3-alpha hydroxysteroid dehydrogenase, type II)|chlordecone reductase homolog HAKRb|dihydrodiol dehydrogenase 3|dihydrodiol dehydrogenase X|indanol dehydrogenase|prostaglandin F synthase|testosterone 17-beta-dehydrogenase 5|trans-1,2-dihydrobenzene-1,2-diol dehydrogenase|type IIb 3-alpha hydroxysteroid dehydrogenase', map_location='10p15-p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5048766, end_position_on_the_genomic_accession=5107686, orientation='plus', exon_count=10, OMIM=603966, no_hdr0=''),
  8645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8645, CurrentID=0, Status='live', Symbol='KCNK5', Aliases='K2p5.1b, TASK-2, TASK2, KCNK5', description='potassium channel, two pore domain subfamily K, member 5', other_designations='K2P5.1 potassium channel|TWIK-related acid-sensitive K(+) channel 2|TWIK-related acid-sensitive K+ channel 2|acid-sensitive potassium channel protein TASK-2|potassium channel, subfamily K, member 1 (TASK-2)|potassium channel, subfamily, member 5 (KCNK5)', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39188971, end_position_on_the_genomic_accession=39229475, orientation='minus', exon_count=5, OMIM=603493, no_hdr0=''),
  8646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8646, CurrentID=0, Status='live', Symbol='CHRD', Aliases='', description='chordin', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184380073, end_position_on_the_genomic_accession=184389835, orientation='plus', exon_count=25, OMIM=603475, no_hdr0=''),
  8647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8647, CurrentID=0, Status='live', Symbol='ABCB11', Aliases='ABC16, BRIC2, BSEP, PFIC-2, PFIC2, PGY4, SPGP', description='ATP binding cassette subfamily B member 11', other_designations='ABC member 16, MDR/TAP subfamily|ATP-binding cassette sub-family B member 11|ATP-binding cassette, sub-family B (MDR/TAP), member 11|progressive familial intrahepatic cholestasis 2|sister p-glycoprotein', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168922802, end_position_on_the_genomic_accession=169034890, orientation='minus', exon_count=29, OMIM=603201, no_hdr0=''),
  8648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8648, CurrentID=0, Status='live', Symbol='NCOA1', Aliases='F-SRC-1, KAT13A, RIP160, SRC1, bHLHe42, bHLHe74', description='nuclear receptor coactivator 1', other_designations='Hin-2 protein|PAX3/NCOA1 fusion protein|class E basic helix-loop-helix protein 74|renal carcinoma antigen NY-REN-52|steroid receptor coactivator-1', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24492050, end_position_on_the_genomic_accession=24770702, orientation='plus', exon_count=24, OMIM=602691, no_hdr0=''),
  8649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8649, CurrentID=0, Status='live', Symbol='LAMTOR3', Aliases='MAP2K1IP1, MAPBP, MAPKSP1, MP1, PRO0633, Ragulator3', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 3', other_designations='MAPK scaffold protein 1|MEK binding partner 1|late endosomal/lysosomal adaptor and MAPK and MTOR activator 3|mitogen-activated protein kinase kinase 1 interacting protein 1|mitogen-activated protein kinase scaffold protein 1', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99878338, end_position_on_the_genomic_accession=99894546, orientation='minus', exon_count=7, OMIM=603296, no_hdr0=''),
  8650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8650, CurrentID=0, Status='live', Symbol='NUMB', Aliases='C14orf41, S171, c14_5527', description='numb homolog (Drosophila)', other_designations='h-Numb', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73275154, end_position_on_the_genomic_accession=73458582, orientation='minus', exon_count=14, OMIM=603728, no_hdr0=''),
  8651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8651, CurrentID=0, Status='live', Symbol='SOCS1', Aliases='CIS1, CISH1, JAB, SOCS-1, SSI-1, SSI1, TIP3', description='suppressor of cytokine signaling 1', other_designations='JAK binding protein|JAK-binding protein|STAT induced SH3 protein 1|STAT-induced STAT inhibitor 1|TIP-3|Tec-interacting protein 3|cytokine-inducible SH2 protein 1', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11254417, end_position_on_the_genomic_accession=11256182, orientation='minus', exon_count=2, OMIM=603597, no_hdr0=''),
  8653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8653, CurrentID=0, Status='live', Symbol='DDX3Y', Aliases='DBY', description='DEAD (Asp-Glu-Ala-Asp) box helicase 3, Y-linked', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, Y-linked|DEAD box protein 3, Y-chromosomal|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide, Y chromosome', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12904108, end_position_on_the_genomic_accession=12920478, orientation='plus', exon_count=20, OMIM=400010, no_hdr0=''),
  8654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8654, CurrentID=0, Status='live', Symbol='PDE5A', Aliases='CGB-PDE, CN5A, PDE5', description='phosphodiesterase 5A', other_designations="cGMP-binding cGMP-specific 3',5'-cyclic nucleotide phosphodiesterase|cGMP-specific phosphodiesterase PDE5A2|cGMP-specific phosphodiesterase type 5A|phosphodiesterase 5A, cGMP-specific|phosphodiesterase isozyme 5", map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119494395, end_position_on_the_genomic_accession=119628826, orientation='minus', exon_count=23, OMIM=603310, no_hdr0=''),
  8655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8655, CurrentID=0, Status='live', Symbol='DYNLL1', Aliases='DLC1, DLC8, DNCL1, DNCLC1, LC8, LC8a, PIN, hdlc1', description='dynein, light chain, LC8-type 1', other_designations='8 kDa dynein light chain|cytoplasmic dynein light polypeptide|dynein, cytoplasmic, light polypeptide 1|protein inhibitor of neuronal nitric oxide synthase', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120469857, end_position_on_the_genomic_accession=120498495, orientation='plus', exon_count=4, OMIM=601562, no_hdr0=''),
  139728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139728, CurrentID=0, Status='live', Symbol='PNCK', Aliases='BSTK3, CaMK1b', description='pregnancy up-regulated nonubiquitous CaM kinase', other_designations='caM kinase I beta|caM kinase IB|caM-KI beta|caMKI-beta|pregnancy up-regulated non-ubiquitously expressed CaM kinase|pregnancy up-regulated non-ubiquitously-expressed CaM kinase|pregnancy upregulated non-ubiquitously expressed CaM kinase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153669733, end_position_on_the_genomic_accession=153687768, orientation='minus', exon_count=22, OMIM=300680, no_hdr0=''),
  8658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8658, CurrentID=0, Status='live', Symbol='TNKS', Aliases='ARTD5, PARP-5a, PARP5A, PARPL, TIN1, TINF11, pART5, TNKS', description='tankyrase, TRF1-interacting ankyrin-related ADP-ribose polymerase', other_designations='ADP-ribosyltransferase diphtheria toxin-like 5|TANK1|TNKS-1|TRF1-interacting ankyrin-related ADP-ribose polymerase|poly [ADP-ribose] polymerase 5A|tankyrase I', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9555214, end_position_on_the_genomic_accession=9782346, orientation='plus', exon_count=29, OMIM=603303, no_hdr0=''),
  8659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8659, CurrentID=0, Status='live', Symbol='ALDH4A1', Aliases='ALDH4, P5CD, P5CDh', description='aldehyde dehydrogenase 4 family member A1', other_designations='L-glutamate gamma-semialdehyde dehydrogenase|P5C dehydrogenase|aldehyde dehydrogenase 4 family, member A1|aldehyde dehydrogenase family 4 member A1|mitochondrial delta-1-pyrroline 5-carboxylate dehydrogenase', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18871430, end_position_on_the_genomic_accession=18902799, orientation='minus', exon_count=17, OMIM=606811, no_hdr0=''),
  8660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8660, CurrentID=0, Status='live', Symbol='IRS2', Aliases='IRS-2', description='insulin receptor substrate 2', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109753837, end_position_on_the_genomic_accession=109786567, orientation='minus', exon_count=2, OMIM=600797, no_hdr0=''),
  8661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8661, CurrentID=0, Status='live', Symbol='EIF3A', Aliases='EIF3, EIF3S10, P167, TIF32, eIF3-p170, eIF3-theta, p180, p185', description='eukaryotic translation initiation factor 3 subunit A', other_designations='EIF3, p180 subunit|centrosomin homolog|cytoplasmic protein p167|eIF-3-theta|eIF3 p167|eIF3 p180|eIF3 p185|eukaryotic translation initiation factor 3 subunit 10|eukaryotic translation initiation factor 3, subunit 10 (theta, 150/170kD)|eukaryotic translation initiation factor 3, subunit 10 (theta, 170kD)|eukaryotic translation initiation factor 3, subunit 10 theta, 150/170kDa|eukaryotic translation initiation factor 3, subunit 10, 170kD|eukaryotic translation initiation factor 3, subunit A', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119035029, end_position_on_the_genomic_accession=119080822, orientation='minus', exon_count=22, OMIM=602039, no_hdr0=''),
  8662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8662, CurrentID=0, Status='live', Symbol='EIF3B', Aliases='EIF3-ETA, EIF3-P110, EIF3-P116, EIF3S9, PRT1', description='eukaryotic translation initiation factor 3 subunit B', other_designations='eIF-3-eta|eIF3 p110|eIF3 p116|eukaryotic translation initiation factor 3 subunit 9|eukaryotic translation initiation factor 3, subunit 9 (eta, 116kD)|eukaryotic translation initiation factor 3, subunit 9 eta, 116kDa|eukaryotic translation initiation factor 3, subunit B|hPrt1|prt1 homolog', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2354540, end_position_on_the_genomic_accession=2380745, orientation='plus', exon_count=19, OMIM=603917, no_hdr0=''),
  8663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8663, CurrentID=0, Status='live', Symbol='EIF3C', Aliases='EIF3CL, EIF3S8, eIF3-p110', description='eukaryotic translation initiation factor 3 subunit C', other_designations='cell migration-inducing protein 17|eIF3 p110|eukaryotic translation initiation factor 3 subunit 8|eukaryotic translation initiation factor 3, subunit 8 (110kD)|eukaryotic translation initiation factor 3, subunit 8, 110kDa|eukaryotic translation initiation factor 3, subunit C', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28688558, end_position_on_the_genomic_accession=28735730, orientation='plus', exon_count=21, OMIM=603916, no_hdr0=''),
  8664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8664, CurrentID=0, Status='live', Symbol='EIF3D', Aliases='EIF3S7, eIF3-p66, eIF3-zeta', description='eukaryotic translation initiation factor 3 subunit D', other_designations='eIF3 p66|eukaryotic translation initiation factor 3, subunit 7 zeta, 66/67kDa|eukaryotic translation initiation factor 3, subunit D|translation initiation factor eIF3 p66 subunit', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36510850, end_position_on_the_genomic_accession=36529431, orientation='minus', exon_count=15, OMIM=603915, no_hdr0=''),
  8665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8665, CurrentID=0, Status='live', Symbol='EIF3F', Aliases='EIF3S5, eIF3-p47', description='eukaryotic translation initiation factor 3 subunit F', other_designations='deubiquitinating enzyme eIF3f|eIF-3-epsilon|eIF3-epsilon|eukaryotic translation initiation factor 3, subunit 5 (epsilon, 47kD)|eukaryotic translation initiation factor 3, subunit 5 epsilon, 47kDa|eukaryotic translation initiation factor 3, subunit F', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7987320, end_position_on_the_genomic_accession=7996171, orientation='plus', exon_count=8, OMIM=603914, no_hdr0=''),
  8666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8666, CurrentID=0, Status='live', Symbol='EIF3G', Aliases='EIF3-P42, EIF3S4, eIF3-delta, eIF3-p44', description='eukaryotic translation initiation factor 3 subunit G', other_designations='eIF-3 RNA-binding subunit|eIF-3-delta|eIF3 p42|eIF3 p44|eukaryotic translation initiation factor 3 RNA-binding subunit|eukaryotic translation initiation factor 3 subunit 4|eukaryotic translation initiation factor 3 subunit p42|eukaryotic translation initiation factor 3, subunit 4 (delta, 44kD)|eukaryotic translation initiation factor 3, subunit 4 delta, 44kDa|eukaryotic translation initiation factor 3, subunit G', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10115014, end_position_on_the_genomic_accession=10119923, orientation='minus', exon_count=11, OMIM=603913, no_hdr0=''),
  8667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8667, CurrentID=0, Status='live', Symbol='EIF3H', Aliases='EIF3S3, eIF3-gamma, eIF3-p40', description='eukaryotic translation initiation factor 3 subunit H', other_designations='eIF-3-gamma|eIF3 p40 subunit|eukaryotic translation initiation factor 3 subunit 3|eukaryotic translation initiation factor 3, subunit 2 (beta, 36kD)|eukaryotic translation initiation factor 3, subunit 3 (gamma, 40kD)|eukaryotic translation initiation factor 3, subunit 3 gamma, 40kDa|eukaryotic translation initiation factor 3, subunit H', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116644816, end_position_on_the_genomic_accession=116755823, orientation='minus', exon_count=8, OMIM=603912, no_hdr0=''),
  8668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8668, CurrentID=0, Status='live', Symbol='EIF3I', Aliases='EIF3S2, PRO2242, TRIP-1, TRIP1, eIF3-beta, eIF3-p36', description='eukaryotic translation initiation factor 3 subunit I', other_designations='TGF-beta receptor-interacting protein 1|TGFbeta receptor-interacting protein 1|eIF-3-beta|eIF3 p36|eukaryotic translation initiation factor 3 subunit 2|eukaryotic translation initiation factor 3, subunit 2 (beta, 36kD)|eukaryotic translation initiation factor 3, subunit 2 beta, 36kDa|eukaryotic translation initiation factor 3, subunit I|predicted protein of HQ2242', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32221867, end_position_on_the_genomic_accession=32231604, orientation='plus', exon_count=12, OMIM=603911, no_hdr0=''),
  8669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8669, CurrentID=0, Status='live', Symbol='EIF3J', Aliases='EIF3S1, eIF3-alpha, eIF3-p35', description='eukaryotic translation initiation factor 3 subunit J', other_designations='eukaryotic translation initiation factor 3, subunit 1 (alpha, 35kD)|eukaryotic translation initiation factor 3, subunit J', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44537068, end_position_on_the_genomic_accession=44562803, orientation='plus', exon_count=8, OMIM=603910, no_hdr0=''),
  8671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8671, CurrentID=0, Status='live', Symbol='SLC4A4', Aliases='HNBC1, KNBC, NBC1, NBC2, NBCe1-A, SLC4A5, hhNMC, pNBC', description='solute carrier family 4 (sodium bicarbonate cotransporter), member 4', other_designations='Na(+)/HCO3(-) cotransporter|kNBC1|sodium bicarbonate cotransporter 1 (sodium bicarbonate cotransporter, kidney; sodium bicarbonate cotransporter, pancreas)|solute carrier family 4, sodium bicarbonate cotransporter, member 4, brain type|solute carrier family 4, sodium bicarbonate cotransporter, member 5', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71187286, end_position_on_the_genomic_accession=71572087, orientation='plus', exon_count=27, OMIM=603345, no_hdr0=''),
  8672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8672, CurrentID=0, Status='live', Symbol='EIF4G3', Aliases='eIF-4G 3, eIF4G 3, eIF4GII', description='eukaryotic translation initiation factor 4 gamma, 3', other_designations='eIF-4-gamma 3|eIF-4-gamma II', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20806292, end_position_on_the_genomic_accession=21177320, orientation='minus', exon_count=45, OMIM=603929, no_hdr0=''),
  8673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8673, CurrentID=0, Status='live', Symbol='VAMP8', Aliases='EDB, VAMP-8', description='vesicle associated membrane protein 8', other_designations='vesicle-associated membrane protein 8 (endobrevin)', map_location='2p12-p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85577491, end_position_on_the_genomic_accession=85582033, orientation='plus', exon_count=3, OMIM=603177, no_hdr0=''),
  8674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8674, CurrentID=0, Status='live', Symbol='VAMP4', Aliases='VAMP-4, VAMP24', description='vesicle associated membrane protein 4', other_designations='', map_location='1q24-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171700156, end_position_on_the_genomic_accession=171742239, orientation='minus', exon_count=8, OMIM=606909, no_hdr0=''),
  8675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8675, CurrentID=0, Status='live', Symbol='STX16', Aliases='SYN16', description='syntaxin 16', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58651253, end_position_on_the_genomic_accession=58679526, orientation='plus', exon_count=9, OMIM=603666, no_hdr0=''),
  8676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8676, CurrentID=0, Status='live', Symbol='STX11', Aliases='FHL4, HLH4, HPLH4', description='syntaxin 11', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144140051, end_position_on_the_genomic_accession=144191939, orientation='plus', exon_count=9, OMIM=605014, no_hdr0=''),
  8677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8677, CurrentID=0, Status='live', Symbol='STX10', Aliases='SYN10, hsyn10', description='syntaxin 10', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13144058, end_position_on_the_genomic_accession=13151334, orientation='minus', exon_count=9, OMIM=603765, no_hdr0=''),
  8678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8678, CurrentID=0, Status='live', Symbol='BECN1', Aliases='ATG6, VPS30, beclin1', description='beclin 1, autophagy related', other_designations='ATG6 autophagy related 6 homolog|beclin 1 (coiled-coil, moesin-like BCL2-interacting protein)|coiled-coil myosin-like BCL2-interacting protein', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42810132, end_position_on_the_genomic_accession=42824292, orientation='minus', exon_count=13, OMIM=604378, no_hdr0=''),
  401895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401895, CurrentID=0, Status='live', Symbol='RPS15P9', Aliases='RPS15_5_1616', description='ribosomal protein S15 pseudogene 9', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1307405, end_position_on_the_genomic_accession=1307745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8681, CurrentID=0, Status='live', Symbol='JMJD7-PLA2G4B', Aliases='HsT16992, cPLA2-beta', description='JMJD7-PLA2G4B readthrough', other_designations='jumonji domain containing 7-phospholipase A2, group IVB (cytosolic) read-through', map_location='15q11.2-q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41828085, end_position_on_the_genomic_accession=41848148, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  8682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8682, CurrentID=0, Status='live', Symbol='PEA15', Aliases='HMAT1, HUMMAT1H, MAT1, MAT1H, PEA-15, PED', description='phosphoprotein enriched in astrocytes 15', other_designations='15 kDa phosphoprotein enriched in astrocytes|homolog of mouse MAT-1 oncogene|mammary transforming gene 1, mouse, homolog of|phosphoprotein enriched in diabetes', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160205319, end_position_on_the_genomic_accession=160215376, orientation='plus', exon_count=6, OMIM=603434, no_hdr0=''),
  8683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8683, CurrentID=0, Status='live', Symbol='SRSF9', Aliases='SFRS9, SRp30c', description='serine/arginine-rich splicing factor 9', other_designations='SR splicing factor 9|pre-mRNA-splicing factor SRp30C|splicing factor, arginine/serine-rich 9', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120461668, end_position_on_the_genomic_accession=120469755, orientation='minus', exon_count=4, OMIM=601943, no_hdr0=''),
  8685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8685, CurrentID=0, Status='live', Symbol='MARCO', Aliases='SCARA2', description='macrophage receptor with collagenous structure', other_designations='scavenger receptor class A, member 2', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118942166, end_position_on_the_genomic_accession=118994664, orientation='plus', exon_count=17, OMIM=604870, no_hdr0=''),
  8686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8686, CurrentID=0, Status='live', Symbol='KRT41P', Aliases='HHaA, KRTHAP1, phihHaA', description='keratin 41 pseudogene', other_designations='keratin 38 pseudogene|keratin, hair, acidic pseudogene 1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41406039, end_position_on_the_genomic_accession=41412699, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8687, CurrentID=0, Status='live', Symbol='KRT38', Aliases='HA8, KRTHA8, hHa8', description='keratin 38, type I', other_designations='K38|hair keratin, type I Ha8|keratin, hair, acidic, 8|keratin-38', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41436369, end_position_on_the_genomic_accession=41441344, orientation='minus', exon_count=7, OMIM=604542, no_hdr0=''),
  139760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139760, CurrentID=0, Status='live', Symbol='GPR119', Aliases='GPCR2', description='G protein-coupled receptor 119', other_designations='G-protein coupled receptor 2', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130384345, end_position_on_the_genomic_accession=130385537, orientation='minus', exon_count=1, OMIM=300513, no_hdr0=''),
  8689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8689, CurrentID=0, Status='live', Symbol='KRT36', Aliases='HA6, KRTHA6, hHa6', description='keratin 36, type I', other_designations='K36|hair keratin, type I Ha6|keratin, hair, acidic, 6|keratin-36', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41486136, end_position_on_the_genomic_accession=41494015, orientation='minus', exon_count=8, OMIM=604540, no_hdr0=''),
  8690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8690, CurrentID=0, Status='live', Symbol='JRKL', Aliases='HHMJG', description='JRK-like', other_designations='human homolog of mouse jerky gene protein', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96389994, end_position_on_the_genomic_accession=96393563, orientation='plus', exon_count=2, OMIM=603211, no_hdr0=''),
  8691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8691, CurrentID=0, Status='live', Symbol='IDDM17', Aliases='', description='insulin-dependent diabetes mellitus 17', other_designations='', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603266, no_hdr0=''),
  8692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8692, CurrentID=0, Status='live', Symbol='HYAL2', Aliases='LUCA2', description='hyaluronoglucosaminidase 2', other_designations='PH-20 homolog|PH20 homolog|hyal-2|hyaluronidase 2|hyaluronoglucosaminidase-2|lung carcinoma protein 2|lysosomal hyaluronidase', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50317790, end_position_on_the_genomic_accession=50322850, orientation='minus', exon_count=6, OMIM=603551, no_hdr0=''),
  8693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8693, CurrentID=0, Status='live', Symbol='GALNT4', Aliases='GALNAC-T4, GALNACT4', description='polypeptide N-acetylgalactosaminyltransferase 4', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 4|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 4 (GalNAc-T4)|polypeptide GalNAc transferase 4|pp-GaNTase 4|protein-UDP acetylgalactosaminyltransferase 4', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89519412, end_position_on_the_genomic_accession=89524806, orientation='minus', exon_count=1, OMIM=603565, no_hdr0=''),
  8694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8694, CurrentID=0, Status='live', Symbol='DGAT1', Aliases='ARAT, ARGP1, DGAT, DIAR7', description='diacylglycerol O-acyltransferase 1', other_designations='ACAT related gene product 1|acyl coenzyme A:cholesterol acyltransferase related gene 1|acyl-CoA retinol O-fatty-acyltransferase|acyl-CoA:diacylglycerol acyltransferase|diglyceride acyltransferase', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144314583, end_position_on_the_genomic_accession=144326919, orientation='minus', exon_count=18, OMIM=604900, no_hdr0=''),
  401911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401911, CurrentID=0, Status='live', Symbol='RPL29P33', Aliases='RPL29_22_1650', description='ribosomal protein L29 pseudogene 33', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34296166, end_position_on_the_genomic_accession=34296784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8697, CurrentID=0, Status='live', Symbol='CDC23', Aliases='ANAPC8, APC8, CUT23', description='cell division cycle 23', other_designations='anaphase promoting complex subunit 8|anaphase-promoting complex subunit 8|cell division cycle 23 homolog|cyclosome subunit 8', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138187648, end_position_on_the_genomic_accession=138213343, orientation='minus', exon_count=16, OMIM=603462, no_hdr0=''),
  8698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8698, CurrentID=0, Status='live', Symbol='S1PR4', Aliases='EDG6, LPC1, S1P4, SLP4', description='sphingosine-1-phosphate receptor 4', other_designations='S1P receptor 4|S1P receptor Edg-6|endothelial differentiation G-protein coupled receptor 6|endothelial differentiation, G protein coupled receptor 6|endothelial differentiation, G-protein-coupled receptor 6|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor, 6|sphingosine 1-phosphate receptor Edg-6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3178738, end_position_on_the_genomic_accession=3180332, orientation='plus', exon_count=1, OMIM=603751, no_hdr0=''),
  678655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=678655, CurrentID=0, Status='live', Symbol='CD27-AS1', Aliases='', description='CD27 antisense RNA 1', other_designations='CD27 antisense RNA 1 (non-protein coding)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6439001, end_position_on_the_genomic_accession=6451517, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  8701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8701, CurrentID=0, Status='live', Symbol='DNAH11', Aliases='CILD7, DNAHBL, DNAHC11, DNHBL, DPL11', description='dynein, axonemal, heavy chain 11', other_designations='axonemal beta dynein heavy chain 11|axonemal dynein heavy chain 11|ciliary dynein heavy chain 11|dynein, axonemal, heavy polypeptide 11|dynein, ciliary, heavy chain 11', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21543215, end_position_on_the_genomic_accession=21901568, orientation='plus', exon_count=82, OMIM=603339, no_hdr0=''),
  8702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8702, CurrentID=0, Status='live', Symbol='B4GALT4', Aliases='B4Gal-T4, beta4Gal-T4', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 4', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 4|UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase 4|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 4|beta-1,4-GalTase 4|beta-N-acetylglucosaminyl-glycolipid beta-1,4-galactosyltransferase 4', map_location='3q13.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119211742, end_position_on_the_genomic_accession=119240905, orientation='minus', exon_count=13, OMIM=604015, no_hdr0=''),
  8703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8703, CurrentID=0, Status='live', Symbol='B4GALT3', Aliases='beta4Gal-T3', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 3', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 3|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 3|b4Gal-T3|beta-1,4-GalTase 3|beta-N-acetylglucosaminyl-glycolipid beta-1,4-galactosyltransferase 3', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161171310, end_position_on_the_genomic_accession=161177968, orientation='minus', exon_count=9, OMIM=604014, no_hdr0=''),
  8704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8704, CurrentID=0, Status='live', Symbol='B4GALT2', Aliases='B4Gal-T2, B4Gal-T3, beta4Gal-T2', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 2', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 2|UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase 2|UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 2|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 2|beta-1,4-GalTase 2|beta-4-GalT2|beta-N-acetylglucosaminyl-glycolipid beta-1,4-galactosyltransferase 2', map_location='1p34-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43979202, end_position_on_the_genomic_accession=43991171, orientation='plus', exon_count=9, OMIM=604013, no_hdr0=''),
  8705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8705, CurrentID=0, Status='live', Symbol='B3GALT4', Aliases='BETA3GALT4, GALT2, GALT4', description='Beta-1,3-galactosyltransferase 4', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase 4|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 4|UDP-galactose:beta-N-acetyl-galactosamine-beta-1,3-galactosyltransferase|b3Gal-T4|beta-1,3-GalTase 4|beta-3-galactosyltransferase 4|gal-T2|ganglioside galactosyltransferase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33277140, end_position_on_the_genomic_accession=33278825, orientation='plus', exon_count=1, OMIM=603095, no_hdr0=''),
  8706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8706, CurrentID=0, Status='live', Symbol='B3GALNT1', Aliases='B3GALT3, GLCT3, GLOB, Gb4Cer, P, P1, beta3Gal-T3, galT3', description='beta-1,3-N-acetylgalactosaminyltransferase 1 (globoside blood group)', other_designations='P antigen synthase|P blood group globoside|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 3 (Globoside blood group)|UDP-GalNAc:betaGlcNAc beta-1,3-galactosaminyltransferase, polypeptide 1 (Globoside blood group)|UDP-N-acetylgalactosamine:globotriaosylceramide beta-1,3-N-acetylgalactosaminyltransferase|b3Gal-T3|beta-1,3-GalNAc-T1|beta-1,3-GalTase 3|beta-1,3-galactosyltransferase 3|beta-3-Gx-T3|beta3GalT3|brainiac1|galactosylgalactosylglucosylceramide beta-D-acetyl-galactosaminyltransferase|globoside synthase|globotriaosylceramide 3-beta-N-acetylgalactosaminyltransferase', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161083883, end_position_on_the_genomic_accession=161105630, orientation='minus', exon_count=9, OMIM=603094, no_hdr0=''),
  8707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8707, CurrentID=0, Status='live', Symbol='B3GALT2', Aliases='BETA3GALT2, GLCT2, beta3Gal-T2', description='Beta-1,3-galactosyltransferase 2', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase 2|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 2|UDP-galactose:2-acetamido-2-deoxy-D-glucose 3beta-galactosyltransferase 2|beta-1,3-GalTase 2|beta-3-galt2', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193178730, end_position_on_the_genomic_accession=193186613, orientation='minus', exon_count=2, OMIM=603018, no_hdr0=''),
  8708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8708, CurrentID=0, Status='live', Symbol='B3GALT1', Aliases='beta3Gal-T1', description='Beta-1,3-galactosyltransferase 1', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase 1|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 1|UDP-galactose:beta-N-acetyl-glucosamine-beta-1,3-galactosyltransferase 1|beta-1,3-GalTase 1|beta-3-galt1|beta3GalT1', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=167293061, end_position_on_the_genomic_accession=167870856, orientation='plus', exon_count=7, OMIM=603093, no_hdr0=''),
  8710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8710, CurrentID=0, Status='live', Symbol='SERPINB7', Aliases='MEGSIN, PPKN, TP55', description='serpin peptidase inhibitor, clade B (ovalbumin), member 7', other_designations='mesangium predominant gene, megsin|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 7', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63753047, end_position_on_the_genomic_accession=63805376, orientation='plus', exon_count=12, OMIM=603357, no_hdr0=''),
  8711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8711, CurrentID=0, Status='live', Symbol='TNK1', Aliases='', description='tyrosine kinase, non-receptor, 1', other_designations='CD38 negative kinase 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7380559, end_position_on_the_genomic_accession=7389774, orientation='plus', exon_count=14, OMIM=608076, no_hdr0=''),
  8712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8712, CurrentID=0, Status='live', Symbol='PAGE1', Aliases='AL5, CT16.3, GAGE-9, GAGEB1, PAGE-1', description='PAGE family member 1', other_designations='G antigen 9|G antigen family B member 1|G antigen, family B, 1 (prostate associated)|P antigen family, member 1 (prostate associated)|prostate associated gene 1|prostate-associated gene 1 protein', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49687450, end_position_on_the_genomic_accession=49696017, orientation='minus', exon_count=7, OMIM=300288, no_hdr0=''),
  100271532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271532, CurrentID=0, Status='live', Symbol='RPL39P41', Aliases='RPL39_22_1750', description='ribosomal protein L39 pseudogene 41', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37215546, end_position_on_the_genomic_accession=37215935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8714, CurrentID=0, Status='live', Symbol='ABCC3', Aliases='ABC31, EST90757, MLP2, MOAT-D, MRP3, cMOAT2', description='ATP binding cassette subfamily C member 3', other_designations='ATP-binding cassette sub-family C member 3|ATP-binding cassette, sub-family C (CFTR/MRP), member 3|canicular multispecific organic anion transporter|multi-specific organic anion transporter D|multidrug resistance associated protein|multidrug resistance-associated protein 3', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50634843, end_position_on_the_genomic_accession=50691704, orientation='plus', exon_count=32, OMIM=604323, no_hdr0=''),
  8715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8715, CurrentID=0, Status='live', Symbol='NOL4', Aliases='CT125, HRIHFB2255, NOLP', description='nucleolar protein 4', other_designations='cancer/testis antigen 125|nucleolar localized protein', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=33851100, end_position_on_the_genomic_accession=34224834, orientation='minus', exon_count=20, OMIM=603577, no_hdr0=''),
  8717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8717, CurrentID=0, Status='live', Symbol='TRADD', Aliases='Hs.89862', description='TNFRSF1A-associated via death domain', other_designations='TNFR1-associated death domain protein|tumor necrosis factor receptor type 1 associated death domain protein|tumor necrosis factor receptor-1-associated protein', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67154185, end_position_on_the_genomic_accession=67159909, orientation='minus', exon_count=5, OMIM=603500, no_hdr0=''),
  8718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8718, CurrentID=0, Status='live', Symbol='TNFRSF25', Aliases='APO-3, DDR3, DR3, LARD, TNFRSF12, TR3, TRAMP, WSL-1, WSL-LR', description='tumor necrosis factor receptor superfamily member 25', other_designations='apoptosis inducing receptor|apoptosis-inducing receptor AIR|apoptosis-mediating receptor DR3|apoptosis-mediating receptor TRAMP|death domain receptor 3 soluble form|death receptor beta|lymphocyte-associated receptor of death|protein WSL-1|tumor necrosis factor receptor superfamily, member 12 (translocating chain-association membrane protein)|tumor necrosis factor receptor superfamily, member 25', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6461151, end_position_on_the_genomic_accession=6466195, orientation='minus', exon_count=10, OMIM=603366, no_hdr0=''),
  100271533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271533, CurrentID=0, Status='live', Symbol='RPL7AP51', Aliases='RPL7A_17_1050', description='ribosomal protein L7a pseudogene 51', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67334123, end_position_on_the_genomic_accession=67334888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8720, CurrentID=0, Status='live', Symbol='MBTPS1', Aliases='PCSK8, S1P, SKI-1', description='membrane bound transcription factor peptidase, site 1', other_designations='endopeptidase S1P|membrane-bound transcription factor peptidase, site 1|proprotein convertase subtilisin/kexin type 8|site-1 protease|subtilisin/kexin isozyme-1', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84053761, end_position_on_the_genomic_accession=84116943, orientation='minus', exon_count=23, OMIM=603355, no_hdr0=''),
  8721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8721, CurrentID=0, Status='live', Symbol='EDF1', Aliases='EDF-1, MBF1', description='endothelial differentiation-related factor 1', other_designations='multiprotein bridging factor 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136862114, end_position_on_the_genomic_accession=136866336, orientation='minus', exon_count=5, OMIM=605107, no_hdr0=''),
  8722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8722, CurrentID=0, Status='live', Symbol='CTSF', Aliases='CATSF, CLN13', description='cathepsin F', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66563463, end_position_on_the_genomic_accession=66568576, orientation='minus', exon_count=14, OMIM=603539, no_hdr0=''),
  8723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8723, CurrentID=0, Status='live', Symbol='SNX4', Aliases='ATG24B', description='sorting nexin 4', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125446644, end_position_on_the_genomic_accession=125520214, orientation='minus', exon_count=14, OMIM=605931, no_hdr0=''),
  8724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8724, CurrentID=0, Status='live', Symbol='SNX3', Aliases='Grd19, MCOPS8, SDP3', description='sorting nexin 3', other_designations='sorting nexin 3A', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108211217, end_position_on_the_genomic_accession=108261260, orientation='minus', exon_count=4, OMIM=605930, no_hdr0=''),
  8725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8725, CurrentID=0, Status='live', Symbol='URI1', Aliases='C19orf2, NNX3, PPP1R19, RMP, URI', description='URI1 prefoldin-like chaperone', other_designations='RNA polymerase II subunit 5-mediating protein|RPB5-mediating protein|URI1, prefoldin-like chaperone|protein phosphatase 1 regulatory subunit 19|protein phosphatase 1, regulatory subunit 19', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29923644, end_position_on_the_genomic_accession=30016612, orientation='plus', exon_count=12, OMIM=603494, no_hdr0=''),
  8726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8726, CurrentID=0, Status='live', Symbol='EED', Aliases='HEED, WAIT1', description='embryonic ectoderm development', other_designations='WD protein associating with integrin cytoplasmic tails 1', map_location='11q14.2-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86244544, end_position_on_the_genomic_accession=86285420, orientation='plus', exon_count=15, OMIM=605984, no_hdr0=''),
  8727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8727, CurrentID=0, Status='live', Symbol='CTNNAL1', Aliases='ACRP, CLLP, alpha-CATU', description='catenin alpha-like 1', other_designations='alpha-catenin-related protein|alpha2-catulin|catenin (cadherin-associated protein), alpha-like 1|catenin alpha-like protein 1', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108942569, end_position_on_the_genomic_accession=109018802, orientation='minus', exon_count=21, OMIM=604785, no_hdr0=''),
  8728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8728, CurrentID=0, Status='live', Symbol='ADAM19', Aliases='FKSG34, MADDAM, MLTNB', description='ADAM metallopeptidase domain 19', other_designations='ADAM 19|a disintegrin and metalloproteinase domain 19 (meltrin beta)|meltrin-beta|metalloprotease and disintegrin dendritic antigen marker|metalloprotease-disintegrin meltrin beta', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157477304, end_position_on_the_genomic_accession=157575823, orientation='minus', exon_count=24, OMIM=603640, no_hdr0=''),
  8729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8729, CurrentID=0, Status='live', Symbol='GBF1', Aliases='ARF1GEF', description='golgi brefeldin A resistant guanine nucleotide exchange factor 1', other_designations='BFA-resistant GEF 1', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102245498, end_position_on_the_genomic_accession=102382899, orientation='plus', exon_count=41, OMIM=603698, no_hdr0=''),
  106480567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480567, CurrentID=0, Status='live', Symbol='RNU6-286P', Aliases='', description='RNA, U6 small nuclear 286, pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8731, CurrentID=0, Status='live', Symbol='RNMT', Aliases='CMT1, CMT1c, MET, Met, RG7MT1, cm1p, hCMT1, hMet', description='RNA (guanine-7-) methyltransferase', other_designations='hcm1p|mRNA (guanine-7-)methyltransferase|mRNA (guanine-N(7)-)-methyltransferase|mRNA cap methyltransferase', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13726647, end_position_on_the_genomic_accession=13764556, orientation='plus', exon_count=15, OMIM=603514, no_hdr0=''),
  8732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8732, CurrentID=0, Status='live', Symbol='RNGTT', Aliases='CAP1A, HCE, HCE1, hCAP', description="RNA guanylyltransferase and 5'-phosphatase", other_designations='HCAP1', map_location='6q16', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88609897, end_position_on_the_genomic_accession=88963716, orientation='minus', exon_count=18, OMIM=603512, no_hdr0=''),
  8733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8733, CurrentID=0, Status='live', Symbol='GPAA1', Aliases='GAA1, hGAA1', description='glycosylphosphatidylinositol anchor attachment 1', other_designations='GAA1 protein homolog|GPAA1P anchor attachment protein 1 homolog|GPI anchor attachment protein 1|GPI transamidase subunit|anchor attachment protein 1 (Gaa1p, yeast) homolog|glycophosphatidylinositol anchor attachment 1|glycosylphosphatidylinositol anchor attachment protein 1 homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144082621, end_position_on_the_genomic_accession=144086216, orientation='plus', exon_count=12, OMIM=603048, no_hdr0=''),
  8735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8735, CurrentID=0, Status='live', Symbol='MYH13', Aliases='MyHC-IIL, MyHC-eo', description='myosin, heavy chain 13, skeletal muscle', other_designations='extraocular muscle myosin heavy chain|extraocular myosin heavy chain|myosin heavy chain 13|myosin heavy chain, skeletal muscle, extraocular|myosin heavy chain, skeletal muscle, laryngeal|myosin, heavy polypeptide 13, skeletal muscle|superfast myosin', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10300866, end_position_on_the_genomic_accession=10373005, orientation='minus', exon_count=41, OMIM=603487, no_hdr0=''),
  8736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8736, CurrentID=0, Status='live', Symbol='MYOM1', Aliases='SKELEMIN', description='myomesin 1', other_designations='190 kDa connectin-associated protein|190 kDa titin-associated protein|EH-myomesin|myomesin (M-protein) 1 (190kD)|myomesin 1 (skelemin) 185kDa|myomesin 1, 185kDa|myomesin family member 1', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3066807, end_position_on_the_genomic_accession=3220108, orientation='minus', exon_count=38, OMIM=603508, no_hdr0=''),
  8737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8737, CurrentID=0, Status='live', Symbol='RIPK1', Aliases='RIP, RIP-1, RIP1', description='receptor interacting serine/threonine kinase 1', other_designations='cell death protein RIP|receptor (TNFRSF)-interacting serine-threonine kinase 1|receptor-interacting protein 1|serine/threonine-protein kinase RIP', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3063967, end_position_on_the_genomic_accession=3115187, orientation='plus', exon_count=13, OMIM=603453, no_hdr0=''),
  8738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8738, CurrentID=0, Status='live', Symbol='CRADD', Aliases='MRT34, RAIDD', description='CASP2 and RIPK1 domain containing adaptor with death domain', other_designations='RIP-associated ICH1/CED3-homologous protein with death domain|RIP-associated protein with a death domain|caspase and RIP adaptor with death domain|death adaptor molecule RAIDD', map_location='12q21.33-q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93677351, end_position_on_the_genomic_accession=93894742, orientation='plus', exon_count=4, OMIM=603454, no_hdr0=''),
  8739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8739, CurrentID=0, Status='live', Symbol='HRK', Aliases='DP5, HARAKIRI', description='harakiri, BCL2 interacting protein', other_designations='BCL2-interacting protein|BH3-interacting domain-containing protein 3|activator of apoptosis Hrk|death protein 5|harakiri, BCL2 interacting protein (contains only BH3 domain)|neuronal death protein DP5', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116860420, end_position_on_the_genomic_accession=116881427, orientation='minus', exon_count=5, OMIM=603447, no_hdr0=''),
  8740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8740, CurrentID=0, Status='live', Symbol='TNFSF14', Aliases='CD258, HVEML, LIGHT, LTg, TR2', description='tumor necrosis factor superfamily member 14', other_designations='delta transmembrane LIGHT|herpes virus entry mediator ligand|herpesvirus entry mediator A|herpesvirus entry mediator ligand|herpesvirus entry mediator-ligand|ligand for herpesvirus entry mediator|tumor necrosis factor (ligand) superfamily member 14|tumor necrosis factor (ligand) superfamily, member 14|tumor necrosis factor receptor-like 2|tumor necrosis factor superfamily member LIGHT', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6658135, end_position_on_the_genomic_accession=6670588, orientation='minus', exon_count=6, OMIM=604520, no_hdr0=''),
  8741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8741, CurrentID=0, Status='live', Symbol='TNFSF13', Aliases='APRIL, CD256, TALL-2, TALL2, TRDL-1, UNQ383/PRO715, ZTNF2', description='tumor necrosis factor superfamily member 13', other_designations='TNF- and APOL-related leukocyte expressed ligand 2|a proliferation-inducing ligand|tumor necrosis factor (ligand) superfamily, member 13|tumor necrosis factor-like protein ZTNF2|tumor necrosis factor-related death ligand-1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7558292, end_position_on_the_genomic_accession=7561608, orientation='plus', exon_count=6, OMIM=604472, no_hdr0=''),
  8742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8742, CurrentID=0, Status='live', Symbol='TNFSF12', Aliases='APO3L, DR3LG, TWEAK', description='tumor necrosis factor superfamily member 12', other_designations='APO3 ligand|APO3/DR3 ligand|TNF-related WEAK inducer of apoptosis|tumor necrosis factor (ligand) superfamily member 12|tumor necrosis factor (ligand) superfamily, member 12', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7549058, end_position_on_the_genomic_accession=7557890, orientation='plus', exon_count=8, OMIM=602695, no_hdr0=''),
  8743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8743, CurrentID=0, Status='live', Symbol='TNFSF10', Aliases='APO2L, Apo-2L, CD253, TL2, TRAIL', description='tumor necrosis factor superfamily member 10', other_designations='Apo-2 ligand|TNF-related apoptosis inducing ligand TRAIL|chemokine tumor necrosis factor ligand superfamily member 10|tumor necrosis factor (ligand) family, member 10|tumor necrosis factor (ligand) superfamily member 10|tumor necrosis factor (ligand) superfamily, member 10|tumor necrosis factor apoptosis-inducing ligand splice variant delta', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172505508, end_position_on_the_genomic_accession=172523507, orientation='minus', exon_count=6, OMIM=603598, no_hdr0=''),
  8744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8744, CurrentID=0, Status='live', Symbol='TNFSF9', Aliases='4-1BB-L, CD137L', description='tumor necrosis factor superfamily member 9', other_designations='4-1BB ligand|4-1BBL|homolog of mouse 4-1BB-L|receptor 4-1BB ligand|tumor necrosis factor (ligand) superfamily member 9|tumor necrosis factor (ligand) superfamily, member 9', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6528562, end_position_on_the_genomic_accession=6535928, orientation='plus', exon_count=4, OMIM=606182, no_hdr0=''),
  8745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8745, CurrentID=0, Status='live', Symbol='ADAM23', Aliases='MDC-3, MDC3', description='ADAM metallopeptidase domain 23', other_designations='metalloproteinase-like, disintegrin-like, and cysteine-rich protein 3', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206443544, end_position_on_the_genomic_accession=206621130, orientation='plus', exon_count=27, OMIM=603710, no_hdr0=''),
  139818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139818, CurrentID=0, Status='live', Symbol='DOCK11', Aliases='ACG, ZIZ2, bB128O4.1', description='dedicator of cytokinesis 11', other_designations='Activated Cdc42-associated guanine nucleotide exchange factor, ACG|bB128O4.1.1 (novel protein)|zizimin-2|zizimin2', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118495909, end_position_on_the_genomic_accession=118686160, orientation='plus', exon_count=58, OMIM=300681, no_hdr0=''),
  8747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8747, CurrentID=0, Status='live', Symbol='ADAM21', Aliases='ADAM 21, ADAM31', description='ADAM metallopeptidase domain 21', other_designations='ADAM metallopeptidase domain 21, preproprotein|a disintegrin and metalloproteinase domain 21', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70452157, end_position_on_the_genomic_accession=70459905, orientation='plus', exon_count=2, OMIM=603713, no_hdr0=''),
  8748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8748, CurrentID=0, Status='live', Symbol='ADAM20', Aliases='', description='ADAM metallopeptidase domain 20', other_designations='ADAM 20|a disintegrin and metalloproteinase domain 20', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70520877, end_position_on_the_genomic_accession=70557701, orientation='minus', exon_count=7, OMIM=603712, no_hdr0=''),
  8749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8749, CurrentID=0, Status='live', Symbol='ADAM18', Aliases='ADAM27, tMDCIII', description='ADAM metallopeptidase domain 18', other_designations='a disintegrin and metalloproteinase domain 18|putative ADAM18 gene product|transmembrane metalloproteinase-like, disintegrin-like, and cysteine-rich protein III', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39584568, end_position_on_the_genomic_accession=39730064, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100147758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147758, CurrentID=0, Status='live', Symbol='RNU7-10P', Aliases='U7.10', description='RNA, U7 small nuclear 10 pseudogene', other_designations='RNA, small nuclear U7.10 pseudogene', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17664967, end_position_on_the_genomic_accession=17665024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8751, CurrentID=0, Status='live', Symbol='ADAM15', Aliases='MDC15', description='ADAM metallopeptidase domain 15', other_designations='MDC-15|a disintegrin and metalloproteinase domain 15 (metargidin)|metalloprotease RGD disintegrin protein|metalloproteinase-like, disintegrin-like, and cysteine-rich protein 15', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155051272, end_position_on_the_genomic_accession=155062776, orientation='plus', exon_count=24, OMIM=605548, no_hdr0=''),
  100147760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147760, CurrentID=0, Status='live', Symbol='RNU7-12P', Aliases='RNU7-100P, U7.12', description='RNA, U7 small nuclear 12 pseudogene', other_designations='RNA, U7 small nuclear 100 pseudogene|RNA, small nuclear U7.12 pseudogene', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27232259, end_position_on_the_genomic_accession=27232316, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147761, CurrentID=0, Status='live', Symbol='RNU7-13P', Aliases='U7.13', description='RNA, U7 small nuclear 13 pseudogene', other_designations='RNA, small nuclear U7.13 pseudogene', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184821430, end_position_on_the_genomic_accession=184821489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8754, CurrentID=0, Status='live', Symbol='ADAM9', Aliases='CORD9, MCMP, MDC9, Mltng', description='ADAM metallopeptidase domain 9', other_designations='ADAM metallopeptidase domain 9 (meltrin gamma)|cellular disintegrin-related protein|cone rod dystrophy 9|metalloprotease/disintegrin/cysteine-rich protein 9|myeloma cell metalloproteinase', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38996965, end_position_on_the_genomic_accession=39105261, orientation='plus', exon_count=24, OMIM=602713, no_hdr0=''),
  8755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8755, CurrentID=0, Status='live', Symbol='ADAM6', Aliases='C14orf96, tMDCIV', description='ADAM metallopeptidase domain 6 (pseudogene)', other_designations='ADAM metallopeptidase domain 6, pseudogene|a disintegrin and metalloproteinase domain 6|metalloprotease-like, disintegrin-like protein pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105969729, end_position_on_the_genomic_accession=105972269, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8756, CurrentID=0, Status='live', Symbol='ADAM7', Aliases='ADAM 7, ADAM-7, EAPI, GP-83, GP83', description='ADAM metallopeptidase domain 7', other_designations='a disintegrin and metalloproteinase domain 7|epididymal apical protein I|sperm maturation-related glycoprotein GP-83', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24440996, end_position_on_the_genomic_accession=24512760, orientation='plus', exon_count=24, OMIM=607310, no_hdr0=''),
  100147765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147765, CurrentID=0, Status='live', Symbol='RNU7-17P', Aliases='U7.17', description='RNA, U7 small nuclear 17 pseudogene', other_designations='RNA, small nuclear U7.17 pseudogene', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22043357, end_position_on_the_genomic_accession=22043418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147766, CurrentID=0, Status='live', Symbol='RNU7-18P', Aliases='RNU7-142P, U7.18', description='RNA, U7 small nuclear 18 pseudogene', other_designations='RNA, U7 small nuclear 142 pseudogene|RNA, small nuclear U7.18 pseudogene', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196072820, end_position_on_the_genomic_accession=196072878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8759, CurrentID=0, Status='live', Symbol='ADAM1A', Aliases='ADAM1, ADAM1P, FTNAP, Ftna, PH-30a', description='ADAM metallopeptidase domain 1A (pseudogene)', other_designations='ADAM metallopeptidase domain 1 (fertilin alpha) pseudogene|ADAM metallopeptidase domain 1, pseudogene|ADAM metallopeptidase domain 1A, pseudogene|ADAM metallopeptidase domain 20 pseudogene|a disintegrin and metalloproteinase domain 1 ( fertilin alpha)|a disintegrin and metalloproteinase domain 1 (fertilin alpha) pseudogene|fertilin alpha', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111899063, end_position_on_the_genomic_accession=111901902, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8760, CurrentID=0, Status='live', Symbol='CDS2', Aliases='', description='CDP-diacylglycerol synthase 2', other_designations='CDP-DAG synthase 2|CDP-DG synthase 2|CDP-DG synthetase 2|CDP-diacylglycerol synthase (phosphatidate cytidylyltransferase) 2|CDP-diglyceride diphosphorylase 2|CDP-diglyceride pyrophosphorylase 2|CDP-diglyceride synthase 2|CDP-diglyceride synthetase 2|CDS 2|CTP:phosphatidate cytidylyltransferase 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5126761, end_position_on_the_genomic_accession=5197887, orientation='plus', exon_count=14, OMIM=603549, no_hdr0=''),
  8761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8761, CurrentID=0, Status='live', Symbol='PABPC4', Aliases='APP-1, APP1, PABP4, iPABP', description='poly(A) binding protein, cytoplasmic 4 (inducible form)', other_designations='PABP-4|activated-platelet protein 1|inducible poly(A)-binding protein|poly(A)-binding protein 4', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39560813, end_position_on_the_genomic_accession=39576849, orientation='minus', exon_count=16, OMIM=603407, no_hdr0=''),
  100147770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147770, CurrentID=0, Status='live', Symbol='RNU7-22P', Aliases='U7.22', description='RNA, U7 small nuclear 22 pseudogene', other_designations='RNA, small nuclear U7.22 pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32173685, end_position_on_the_genomic_accession=32173743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8763, CurrentID=0, Status='live', Symbol='CD164', Aliases='MGC-24, MUC-24, endolyn', description='CD164 molecule, sialomucin', other_designations='CD164 antigen, sialomucin|CD164 variant C|MGC-24v|multi-glycosylated core protein 24', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109366514, end_position_on_the_genomic_accession=109382559, orientation='minus', exon_count=6, OMIM=603356, no_hdr0=''),
  8764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8764, CurrentID=0, Status='live', Symbol='TNFRSF14', Aliases='ATAR, CD270, HVEA, HVEM, LIGHTR, TR2', description='tumor necrosis factor receptor superfamily member 14', other_designations='CD40-like protein|herpes virus entry mediator A|tumor necrosis factor receptor superfamily, member 14 (herpesvirus entry mediator)|tumor necrosis factor receptor-like gene2', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2556365, end_position_on_the_genomic_accession=2565622, orientation='plus', exon_count=8, OMIM=602746, no_hdr0=''),
  401981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401981, CurrentID=0, Status='live', Symbol='FDPSP8', Aliases='', description='farnesyl diphosphate synthase pseudogene 8', other_designations='farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211660133, end_position_on_the_genomic_accession=211661299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8766, CurrentID=0, Status='live', Symbol='RAB11A', Aliases='YL8', description='RAB11A, member RAS oncogene family', other_designations='RAB 11A, member oncogene family|rab-11', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65869459, end_position_on_the_genomic_accession=65891991, orientation='plus', exon_count=5, OMIM=605570, no_hdr0=''),
  8767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8767, CurrentID=0, Status='live', Symbol='RIPK2', Aliases='CARD3, CARDIAK, CCK, GIG30, RICK, RIP2', description='receptor interacting serine/threonine kinase 2', other_designations='CARD-carrying kinase|CARD-containing IL-1 beta ICE-kinase|CARD-containing interleukin-1 beta-converting enzyme (ICE)-associated kinase|RIP-2|growth-inhibiting gene 30|receptor-interacting protein (RIP)-like interacting caspase-like apoptosis regulatory protein (CLARP) kinase|receptor-interacting protein 2|receptor-interacting serine-threonine kinase 2|tyrosine-protein kinase RIPK2', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89757733, end_position_on_the_genomic_accession=89791064, orientation='plus', exon_count=12, OMIM=603455, no_hdr0=''),
  8771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8771, CurrentID=0, Status='live', Symbol='TNFRSF6B', Aliases='DCR3, DJ583P15.1.1, M68, M68E, TR6', description='tumor necrosis factor receptor superfamily member 6b', other_designations='decoy receptor 3 variant 1|decoy receptor 3 variant 2|decoy receptor for Fas ligand|tumor necrosis factor receptor superfamily member 6b, decoy|tumor necrosis factor receptor superfamily, member 6b, decoy', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63696651, end_position_on_the_genomic_accession=63698698, orientation='plus', exon_count=3, OMIM=603361, no_hdr0=''),
  8772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8772, CurrentID=0, Status='live', Symbol='FADD', Aliases='GIG3, MORT1', description='Fas associated via death domain', other_designations='Fas (TNFRSF6)-associated via death domain|Fas-associating death domain-containing protein|Fas-associating protein with death domain|growth-inhibiting gene 3 protein|mediator of receptor induced toxicity|mediator of receptor-induced toxicity', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70203163, end_position_on_the_genomic_accession=70207402, orientation='plus', exon_count=2, OMIM=602457, no_hdr0=''),
  8773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8773, CurrentID=0, Status='live', Symbol='SNAP23', Aliases='HsT17016, SNAP-23A, SNAP23B, SNAP23', description='synaptosome associated protein 23kDa', other_designations='synaptosomal-associated protein, 23kD|synaptosomal-associated protein, 23kDa|vesicle-membrane fusion protein SNAP-23', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42491244, end_position_on_the_genomic_accession=42533061, orientation='plus', exon_count=11, OMIM=602534, no_hdr0=''),
  8774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8774, CurrentID=0, Status='live', Symbol='NAPG', Aliases='GAMMASNAP', description='N-ethylmaleimide-sensitive factor attachment protein, gamma', other_designations='SNAP-gamma|gamma SNAP|soluble NSF attachment protein', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10525876, end_position_on_the_genomic_accession=10552769, orientation='plus', exon_count=14, OMIM=603216, no_hdr0=''),
  8775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8775, CurrentID=0, Status='live', Symbol='NAPA', Aliases='SNAPA', description='NSF attachment protein alpha', other_designations='N-ethylmaleimide-sensitive factor attachment protein, alpha|alpha soluble NSF attachment protein|alpha-SNAP', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47487634, end_position_on_the_genomic_accession=47515258, orientation='minus', exon_count=13, OMIM=603215, no_hdr0=''),
  8776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8776, CurrentID=0, Status='live', Symbol='MTMR1', Aliases='', description='myotubularin related protein 1', other_designations='phosphatidylinositol-3-phosphate phosphatase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150692983, end_position_on_the_genomic_accession=150765102, orientation='plus', exon_count=21, OMIM=300171, no_hdr0=''),
  8777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8777, CurrentID=0, Status='live', Symbol='MPDZ', Aliases='HYC2, MUPP1', description='multiple PDZ domain protein', other_designations='multi-PDZ domain protein 1', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13105296, end_position_on_the_genomic_accession=13279662, orientation='minus', exon_count=50, OMIM=603785, no_hdr0=''),
  8778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8778, CurrentID=0, Status='live', Symbol='SIGLEC5', Aliases='CD170, CD33L2, OB-BP2, OBBP2, SIGLEC-5', description='sialic acid binding Ig-like lectin 5', other_designations='CD33 antigen-like 2|OB-binding protein 2|obesity-binding protein 2|sialic acid-binding immunoglobulin-like lectin 5', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51610957, end_position_on_the_genomic_accession=51630474, orientation='minus', exon_count=9, OMIM=604200, no_hdr0=''),
  100271543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271543, CurrentID=0, Status='live', Symbol='RPL7AP65', Aliases='RPL7A_30_1522', description='ribosomal protein L7a pseudogene 65', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18311882, end_position_on_the_genomic_accession=18312693, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8780, CurrentID=0, Status='live', Symbol='RIOK3', Aliases='SUDD', description='RIO kinase 3', other_designations='homolog of the Aspergillus nidulans sudD gene product|sudD (suppressor of bimD6, Aspergillus nidulans) homolog|sudD homolog|sudD suppressor of Aspergillus nidulans bimD6 homolog|sudD suppressor of bimD6 homolog', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23452823, end_position_on_the_genomic_accession=23483140, orientation='plus', exon_count=13, OMIM=603579, no_hdr0=''),
  8781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8781, CurrentID=0, Status='live', Symbol='PSPHP1', Aliases='CO9, PSPHL', description='phosphoserine phosphatase pseudogene 1', other_designations='L-3-phosphoserine-phosphatase homolog', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55764797, end_position_on_the_genomic_accession=55773696, orientation='plus', exon_count=0, OMIM=604239, no_hdr0=''),
  8782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8782, CurrentID=0, Status='live', Symbol='MYP3', Aliases='', description='myopia 3 (high grade, autosomal dominant)', other_designations='Myopia, high grade, autosomal dominant 2', map_location='12q21-q23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603221, no_hdr0=''),
  105259599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105259599, CurrentID=0, Status='live', Symbol='H19-ICR', Aliases='H19-DMD, IC1, ICR1, ICR1-DMR', description='H19/IGF2 imprinting control region', other_designations='ICR1 differentially methylated region|imprinting center 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616186, no_hdr0=''),
  8784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8784, CurrentID=0, Status='live', Symbol='TNFRSF18', Aliases='AITR, CD357, GITR, GITR-D', description='tumor necrosis factor receptor superfamily member 18', other_designations='TNF receptor superfamily activation-inducible protein|activation-inducible TNFR family receptor|glucocorticoid-induced TNFR-related protein|tumor necrosis factor receptor superfamily, member 18', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1203508, end_position_on_the_genomic_accession=1206709, orientation='minus', exon_count=5, OMIM=603905, no_hdr0=''),
  8785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8785, CurrentID=0, Status='live', Symbol='MATN4', Aliases='', description='matrilin 4', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45293446, end_position_on_the_genomic_accession=45308566, orientation='minus', exon_count=10, OMIM=603897, no_hdr0=''),
  8786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8786, CurrentID=0, Status='live', Symbol='RGS11', Aliases='RS11', description='regulator of G-protein signaling 11', other_designations='regulator of G-protein signalling 11', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=268301, end_position_on_the_genomic_accession=275943, orientation='minus', exon_count=20, OMIM=603895, no_hdr0=''),
  8787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8787, CurrentID=0, Status='live', Symbol='RGS9', Aliases='PERRSL, RGS9', description='regulator of G-protein signaling 9', other_designations='regulator of G-protein signalling 9', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65137338, end_position_on_the_genomic_accession=65227703, orientation='plus', exon_count=20, OMIM=604067, no_hdr0=''),
  8788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8788, CurrentID=0, Status='live', Symbol='DLK1', Aliases='DLK, DLK-1, Delta1, FA1, PREF1, Pref-1, ZOG, pG2', description='delta-like 1 homolog (Drosophila)', other_designations='fetal antigen 1|preadipocyte factor 1|secredeltin', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100726865, end_position_on_the_genomic_accession=100735130, orientation='plus', exon_count=5, OMIM=176290, no_hdr0=''),
  8789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8789, CurrentID=0, Status='live', Symbol='FBP2', Aliases='', description='fructose-bisphosphatase 2', other_designations='D-fructose-1,6-bisphosphate 1-phosphohydrolase 2|FBPase 2|fructose-1,6-bisphosphatase 2|hexosediphosphatase|muscle FBPase|muscle fructose-bisphosphatase', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94558714, end_position_on_the_genomic_accession=94593832, orientation='minus', exon_count=8, OMIM=603027, no_hdr0=''),
  8790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8790, CurrentID=0, Status='live', Symbol='FPGT', Aliases='GFPP', description='fucose-1-phosphate guanylyltransferase', other_designations='GDP-L-fucose diphosphorylase|GDP-beta-L-fucose pyrophosphorylase|fucose-1-phosphate guanyltransferase', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74198212, end_position_on_the_genomic_accession=74208702, orientation='plus', exon_count=5, OMIM=603609, no_hdr0=''),
  100271545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271545, CurrentID=0, Status='live', Symbol='RPL7AP69', Aliases='RPL7A_31_1673', description='ribosomal protein L7a pseudogene 69', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57065346, end_position_on_the_genomic_accession=57066139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8792, CurrentID=0, Status='live', Symbol='TNFRSF11A', Aliases='CD265, FEO, LOH18CR1, ODFR, OFE, OPTB7, OSTS, PDB2, RANK, TRANCER', description='tumor necrosis factor receptor superfamily member 11a', other_designations='Paget disease of bone 2|loss of heterozygosity, 18, chromosomal region 1|osteoclast differentiation factor receptor|receptor activator of NF-KB|receptor activator of nuclear factor-kappa B|tumor necrosis factor receptor superfamily member 11a, NFKB activator|tumor necrosis factor receptor superfamily, member 11a, NFKB activator', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62325287, end_position_on_the_genomic_accession=62387710, orientation='plus', exon_count=12, OMIM=603499, no_hdr0=''),
  8793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8793, CurrentID=0, Status='live', Symbol='TNFRSF10D', Aliases='CD264, DCR2, TRAIL-R4, TRAILR4, TRUNDD', description='tumor necrosis factor receptor superfamily member 10d', other_designations='TNF receptor-related receptor for TRAIL|TNF-related apoptosis-inducing ligand receptor 4|TRAIL receptor 4|TRAIL receptor with a truncated death domain|decoy receptor 2|tumor necrosis factor receptor superfamily member 10d, decoy with truncated death domain|tumor necrosis factor receptor superfamily, member 10d, decoy with truncated death domain', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23135588, end_position_on_the_genomic_accession=23164030, orientation='minus', exon_count=9, OMIM=603614, no_hdr0=''),
  8794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8794, CurrentID=0, Status='live', Symbol='TNFRSF10C', Aliases='CD263, DCR1, DCR1-TNFR, LIT, TRAIL-R3, TRAILR3, TRID', description='tumor necrosis factor receptor superfamily member 10c', other_designations='TNF-related apoptosis-inducing ligand receptor 3|antagonist decoy receptor for TRAIL/Apo-2L|cytotoxic TRAIL receptor-3|decoy TRAIL receptor without death domain|decoy receptor 1|lymphocyte inhibitor of TRAIL|tumor necrosis factor receptor superfamily member 10c, decoy without an intracellular domain|tumor necrosis factor receptor superfamily, member 10c, decoy without an intracellular domain', map_location='8p22-p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23102814, end_position_on_the_genomic_accession=23117437, orientation='plus', exon_count=5, OMIM=603613, no_hdr0=''),
  8795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8795, CurrentID=0, Status='live', Symbol='TNFRSF10B', Aliases='CD262, DR5, KILLER, KILLER/DR5, TRAIL-R2, TRAILR2, TRICK2, TRICK2A, TRICK2B, TRICKB, ZTNFR9', description='tumor necrosis factor receptor superfamily member 10b', other_designations='Fas-like protein|TNF receptor superfamily member 10b|TNF-related apoptosis-inducing ligand receptor 2|apoptosis inducing protein TRICK2A/2B|apoptosis inducing receptor TRAIL-R2|cytotoxic TRAIL receptor-2|death domain containing receptor for TRAIL/Apo-2L|death receptor 5|p53-regulated DNA damage-inducible cell death receptor(killer)|tumor necrosis factor receptor superfamily, member 10b|tumor necrosis factor receptor-like protein ZTNFR9', map_location='8p22-p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23006383, end_position_on_the_genomic_accession=23069187, orientation='minus', exon_count=10, OMIM=603612, no_hdr0=''),
  8796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8796, CurrentID=0, Status='live', Symbol='SCEL', Aliases='', description='sciellin', other_designations='', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77535674, end_position_on_the_genomic_accession=77645263, orientation='plus', exon_count=33, OMIM=604112, no_hdr0=''),
  8797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8797, CurrentID=0, Status='live', Symbol='TNFRSF10A', Aliases='APO2, CD261, DR4, TRAILR-1, TRAILR1', description='tumor necrosis factor receptor superfamily member 10a', other_designations='TNF-related apoptosis-inducing ligand receptor 1|TRAIL receptor 1|TRAIL-R1|cytotoxic TRAIL receptor|death receptor 4|tumor necrosis factor receptor superfamily member 10a variant 2|tumor necrosis factor receptor superfamily, member 10a', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23191457, end_position_on_the_genomic_accession=23225167, orientation='minus', exon_count=10, OMIM=603611, no_hdr0=''),
  8798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8798, CurrentID=0, Status='live', Symbol='DYRK4', Aliases='', description='dual specificity tyrosine-(Y)-phosphorylation regulated kinase 4', other_designations='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 4', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4590072, end_position_on_the_genomic_accession=4613888, orientation='plus', exon_count=13, OMIM=609181, no_hdr0=''),
  8799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8799, CurrentID=0, Status='live', Symbol='PEX11B', Aliases='PEX11-BETA, PEX14B', description='peroxisomal biogenesis factor 11 beta', other_designations='peroxin-11B|peroxisomal biogenesis factor 11B|protein PEX11 homolog beta', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145911348, end_position_on_the_genomic_accession=145918924, orientation='minus', exon_count=5, OMIM=603867, no_hdr0=''),
  8800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8800, CurrentID=0, Status='live', Symbol='PEX11A', Aliases='PEX11-ALPHA, PMP28, hsPEX11p', description='peroxisomal biogenesis factor 11 alpha', other_designations='28 kDa peroxisomal integral membrane protein|peroxin-11A|peroxisomal biogenesis factor 11A|protein PEX11 homolog alpha', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89681531, end_position_on_the_genomic_accession=89690784, orientation='minus', exon_count=3, OMIM=603866, no_hdr0=''),
  8801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8801, CurrentID=0, Status='live', Symbol='SUCLG2', Aliases='GBETA', description='succinate-CoA ligase, GDP-forming, beta subunit', other_designations='GTP-specific succinyl-CoA synthetase beta subunit|GTP-specific succinyl-CoA synthetase subunit beta|SCS-betaG|succinyl-CoA ligase, GDP-forming, beta chain, mitochondrial|succinyl-CoA synthetase beta-G chain|succinyl-CoA synthetase, beta-G chain', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67360460, end_position_on_the_genomic_accession=67654614, orientation='minus', exon_count=14, OMIM=603922, no_hdr0=''),
  8802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8802, CurrentID=0, Status='live', Symbol='SUCLG1', Aliases='GALPHA, MTDPS9, SUCLA1', description='succinate-CoA ligase, alpha subunit', other_designations='SCS-alpha|succinyl-CoA ligase [GDP-forming] subunit alpha, mitochondrial|succinyl-CoA synthetase subunit alpha', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84423523, end_position_on_the_genomic_accession=84459462, orientation='minus', exon_count=9, OMIM=611224, no_hdr0=''),
  8803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8803, CurrentID=0, Status='live', Symbol='SUCLA2', Aliases='A-BETA, MTDPS5, SCS-betaA', description='succinate-CoA ligase, ADP-forming, beta subunit', other_designations='ATP-specific succinyl-CoA synthetase subunit beta|ATP-specific succinyl-CoA synthetase, beta subunit|mitochondrial succinyl-CoA ligase [ADP-forming] subunit beta|renal carcinoma antigen NY-REN-39|succinate-CoA ligase ADP-forming beta subunit|succinyl-CoA synthetase beta-A chain', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47942656, end_position_on_the_genomic_accession=48001326, orientation='minus', exon_count=13, OMIM=603921, no_hdr0=''),
  8804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8804, CurrentID=0, Status='live', Symbol='CREG1', Aliases='CREG', description='cellular repressor of E1A-stimulated genes 1', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167541013, end_position_on_the_genomic_accession=167553819, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8805, CurrentID=0, Status='live', Symbol='TRIM24', Aliases='PTC6, RNF82, TF1A, TIF1, TIF1A, TIF1ALPHA, hTIF1', description='tripartite motif containing 24', other_designations='E3 ubiquitin-protein ligase TRIM24|RING finger protein 82|TIF1-alpha|transcriptional intermediary factor 1|tripartite motif-containing 24', map_location='7q32-q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138460216, end_position_on_the_genomic_accession=138585588, orientation='plus', exon_count=20, OMIM=603406, no_hdr0=''),
  8806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8806, CurrentID=0, Status='live', Symbol='SCZD8', Aliases='', description='schizophrenia disorder 8', other_designations='Schizophrenia susceptibility locus, chromosome 18-related', map_location='18p', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603206, no_hdr0=''),
  8807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8807, CurrentID=0, Status='live', Symbol='IL18RAP', Aliases='ACPL, CD218b, CDw218b, IL-18R-beta, IL-18RAcP, IL-18Rbeta, IL-1R-7, IL-1R7, IL-1RAcPL, IL18RB', description='interleukin 18 receptor accessory protein', other_designations='CD218 antigen-like family member B|IL-18 receptor accessory protein|IL-18 receptor beta|cluster of differentiation w218b|interleukin-1 receptor 7|interleukin-18 receptor beta', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102418558, end_position_on_the_genomic_accession=102452568, orientation='plus', exon_count=14, OMIM=604509, no_hdr0=''),
  8808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8808, CurrentID=0, Status='live', Symbol='IL1RL2', Aliases='IL-1Rrp2, IL-36R, IL1R-rp2, IL1RRP2', description='interleukin 1 receptor-like 2', other_designations='IL-1 receptor-related protein 2|IL-36 receptor|interleukin-1 receptor-related protein 2', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102186157, end_position_on_the_genomic_accession=102240002, orientation='plus', exon_count=14, OMIM=604512, no_hdr0=''),
  8809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8809, CurrentID=0, Status='live', Symbol='IL18R1', Aliases='CD218a, CDw218a, IL-1Rrp, IL18RA, IL1RRP', description='interleukin 18 receptor 1', other_designations='CD218 antigen-like family member A|IL-18R-1|IL-18R1|IL1 receptor-related protein|IL18Ralpha2|IL1R-rp|cytokine receptor', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102356283, end_position_on_the_genomic_accession=102398777, orientation='plus', exon_count=13, OMIM=604494, no_hdr0=''),
  100147818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147818, CurrentID=0, Status='live', Symbol='RNU7-31P', Aliases='U7.31', description='RNA, U7 small nuclear 31 pseudogene', other_designations='RNA, small nuclear U7.31 pseudogene', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5514210, end_position_on_the_genomic_accession=5514270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8811, CurrentID=0, Status='live', Symbol='GALR2', Aliases='GAL2-R, GALNR2, GALR-2', description='galanin receptor 2', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76072088, end_position_on_the_genomic_accession=76077541, orientation='plus', exon_count=4, OMIM=603691, no_hdr0=''),
  8812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8812, CurrentID=0, Status='live', Symbol='CCNK', Aliases='CPR4', description='cyclin K', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99481402, end_position_on_the_genomic_accession=99511515, orientation='plus', exon_count=12, OMIM=603544, no_hdr0=''),
  8813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8813, CurrentID=0, Status='live', Symbol='DPM1', Aliases='CDGIE, MPDS', description='dolichyl-phosphate mannosyltransferase polypeptide 1, catalytic subunit', other_designations='DPM synthase complex, catalytic subunit|DPM synthase subunit 1|MPD synthase subunit 1|dolichol monophosphate mannose synthase|dolichol-phosphate mannose synthase subunit 1|dolichyl-phosphate beta-D-mannosyltransferase subunit 1|dolichyl-phosphate mannosyltransferase polypeptide 1 catalytic subunit|mannose-P-dolichol synthase subunit 1', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50934855, end_position_on_the_genomic_accession=50958561, orientation='minus', exon_count=10, OMIM=603503, no_hdr0=''),
  8814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8814, CurrentID=0, Status='live', Symbol='CDKL1', Aliases='KKIALRE, P42', description='cyclin-dependent kinase-like 1 (CDC2-related kinase)', other_designations='CDC2-related kinase 1|protein kinase p42 KKIALRE|serine/threonine protein kinase KKIALRE|serine/threonine-protein kinase KKIALRE', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50329404, end_position_on_the_genomic_accession=50397404, orientation='minus', exon_count=14, OMIM=603441, no_hdr0=''),
  8815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8815, CurrentID=0, Status='live', Symbol='BANF1', Aliases='BAF, BCRP1, D14S1460, NGPS', description='barrier to autointegration factor 1', other_designations='breakpoint cluster region protein 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66002079, end_position_on_the_genomic_accession=66004146, orientation='plus', exon_count=3, OMIM=603811, no_hdr0=''),
  8816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8816, CurrentID=0, Status='live', Symbol='DCAF5', Aliases='BCRG2, BCRP2, D14S1461E, WDR22', description='DDB1 and CUL4 associated factor 5', other_designations='WD repeat-containing protein 22|breakpoint cluster region protein, uterine leiomyoma, 2', map_location='14q23-q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69050920, end_position_on_the_genomic_accession=69153393, orientation='minus', exon_count=13, OMIM=603812, no_hdr0=''),
  8817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8817, CurrentID=0, Status='live', Symbol='FGF18', Aliases='FGF-18, ZFGF5', description='fibroblast growth factor 18', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171419663, end_position_on_the_genomic_accession=171457626, orientation='plus', exon_count=5, OMIM=603726, no_hdr0=''),
  8818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8818, CurrentID=0, Status='live', Symbol='DPM2', Aliases='CDG1U', description='dolichyl-phosphate mannosyltransferase polypeptide 2, regulatory subunit', other_designations='DPM synthase complex subunit|DPM synthase subunit 2|dolichol-phosphate mannose synthase subunit 2', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127935095, end_position_on_the_genomic_accession=127938484, orientation='minus', exon_count=4, OMIM=603564, no_hdr0=''),
  8819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8819, CurrentID=0, Status='live', Symbol='SAP30', Aliases='', description='Sin3A associated protein 30kDa', other_designations='30 kDa Sin3-associated polypeptide|Sin3 corepressor complex subunit SAP30|Sin3-associated polypeptide, 30kDa|Sin3A-associated protein, 30kDa|sin3-associated polypeptide p30|sin3-associated polypeptide, 30 kDa', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173370942, end_position_on_the_genomic_accession=173377532, orientation='plus', exon_count=4, OMIM=603378, no_hdr0=''),
  8820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8820, CurrentID=0, Status='live', Symbol='HESX1', Aliases='ANF, CPHD5, RPX', description='HESX homeobox 1', other_designations='Rathke pouch homeobox|hAnf|homeobox protein ANF', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57197838, end_position_on_the_genomic_accession=57227632, orientation='minus', exon_count=7, OMIM=601802, no_hdr0=''),
  8821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8821, CurrentID=0, Status='live', Symbol='INPP4B', Aliases='', description='inositol polyphosphate-4-phosphatase type II B', other_designations='inositol polyphosphate 4-phosphatase II; 4-phosphatase II|inositol polyphosphate 4-phosphatase type II|inositol polyphosphate-4-phosphatase, type II, 105kD|inositol polyphosphate-4-phosphatase, type II, 105kDa|type II inositol-3,4-bisphosphate 4-phosphatase', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=142028028, end_position_on_the_genomic_accession=142846535, orientation='minus', exon_count=39, OMIM=607494, no_hdr0=''),
  8822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8822, CurrentID=0, Status='live', Symbol='FGF17', Aliases='FGF-13, FGF-17, HH20', description='fibroblast growth factor 17', other_designations='', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22039709, end_position_on_the_genomic_accession=22048809, orientation='plus', exon_count=8, OMIM=603725, no_hdr0=''),
  8823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8823, CurrentID=0, Status='live', Symbol='FGF16', Aliases='FGF-16, MF4', description='fibroblast growth factor 16', other_designations='metacarpal 4-5 fusion', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77447675, end_position_on_the_genomic_accession=77456522, orientation='plus', exon_count=3, OMIM='300827,309630', no_hdr0=''),
  8824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8824, CurrentID=0, Status='live', Symbol='CES2', Aliases='CE-2A1, PCE-2, iCE, CES2', description='carboxylesterase 2', other_designations='carboxylesterase 2 (intestine, liver)|hCE-2|intestinal carboxylesterase; liver carboxylesterase-2|methylumbelliferyl-acetate deacetylase 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66934444, end_position_on_the_genomic_accession=66945096, orientation='plus', exon_count=12, OMIM=605278, no_hdr0=''),
  8825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8825, CurrentID=0, Status='live', Symbol='LIN7A', Aliases='LIN-7A, LIN7, MALS-1, TIP-33, VELI1', description='lin-7 homolog A (C. elegans)', other_designations='hLin-7|mammalian LIN-7 1|mammalian lin-seven protein 1|tax interaction protein 33|veli-1|vertebrate LIN7 homolog 1|vertebrate lin-7 homolog 1', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80792520, end_position_on_the_genomic_accession=80937923, orientation='minus', exon_count=6, OMIM=603380, no_hdr0=''),
  8826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8826, CurrentID=0, Status='live', Symbol='IQGAP1', Aliases='HUMORFA01, SAR1, p195', description='IQ motif containing GTPase activating protein 1', other_designations='RasGAP-like with IQ motifs', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90388241, end_position_on_the_genomic_accession=90502243, orientation='plus', exon_count=39, OMIM=603379, no_hdr0=''),
  100147835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147835, CurrentID=0, Status='live', Symbol='RNU7-48P', Aliases='U7.48', description='RNA, U7 small nuclear 48 pseudogene', other_designations='RNA, small nuclear U7.48 pseudogene', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8828, CurrentID=0, Status='live', Symbol='NRP2', Aliases='NP2, NPN2, PRO2714, VEGF165R2', description='neuropilin 2', other_designations='neuropilin-2a(17)|neuropilin-2a(22)|neuropilin-2b(0)|receptor for VEGF165 and semaphorins class3|vascular endothelial cell growth factor 165 receptor 2', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=205682449, end_position_on_the_genomic_accession=205798133, orientation='plus', exon_count=20, OMIM=602070, no_hdr0=''),
  8829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8829, CurrentID=0, Status='live', Symbol='NRP1', Aliases='BDCA4, CD304, NP1, NRP, VEGF165R', description='neuropilin 1', other_designations='transmembrane receptor|vascular endothelial cell growth factor 165 receptor', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33177491, end_position_on_the_genomic_accession=33334905, orientation='minus', exon_count=18, OMIM=602069, no_hdr0=''),
  8830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8830, CurrentID=0, Status='live', Symbol='PTLAH', Aliases='FPAH', description='patella aplasia-hypoplasia', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=168860, no_hdr0=''),
  8831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8831, CurrentID=0, Status='live', Symbol='SYNGAP1', Aliases='MRD5, RASA1, RASA5, SYNGAP', description='synaptic Ras GTPase activating protein 1', other_designations='Ras GTPase-activating protein SynGAP|neuronal RasGAP|synaptic Ras GTPase activating protein 1 homolog|synaptic Ras GTPase activating protein, 135kDa|synaptic Ras GTPase-activating protein 1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33420070, end_position_on_the_genomic_accession=33453689, orientation='plus', exon_count=19, OMIM=603384, no_hdr0=''),
  8832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8832, CurrentID=0, Status='live', Symbol='CD84', Aliases='LY9B, SLAMF5, hCD84, mCD84', description='CD84 molecule', other_designations='CD84 antigen (leukocyte antigen)|cell surface antigen MAX.3|hly9-beta|leucocyte differentiation antigen CD84|leukocyte antigen CD84|leukocyte differentiation antigen CD84|signaling lymphocytic activation molecule 5', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160541094, end_position_on_the_genomic_accession=160579516, orientation='minus', exon_count=9, OMIM=604513, no_hdr0=''),
  8833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8833, CurrentID=0, Status='live', Symbol='GMPS', Aliases='', description='guanine monophosphate synthase', other_designations="GMP synthase|GMP synthetase|MLL/GMPS fusion protein|glutamine amidotransferase|guanine monophosphate synthetase|guanosine 5'-monophosphate synthase", map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155869430, end_position_on_the_genomic_accession=155939608, orientation='plus', exon_count=18, OMIM=600358, no_hdr0=''),
  8834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8834, CurrentID=0, Status='live', Symbol='TMEM11', Aliases='C17orf35, PM1, PMI', description='transmembrane protein 11', other_designations='putative receptor protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21197950, end_position_on_the_genomic_accession=21214595, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8835, CurrentID=0, Status='live', Symbol='SOCS2', Aliases='CIS2, Cish2, SOCS-2, SSI-2, SSI2, STATI2', description='suppressor of cytokine signaling 2', other_designations='CIS-2|STAT induced STAT inhibitor-2|STAT-induced STAT inhibitor 2|STAT-induced STAT inhibitor-2|cytokine-inducible SH2 protein 2|suppressor of cytokine signaling-2', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93569822, end_position_on_the_genomic_accession=93626236, orientation='plus', exon_count=12, OMIM=605117, no_hdr0=''),
  8836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8836, CurrentID=0, Status='live', Symbol='GGH', Aliases='GH', description='gamma-glutamyl hydrolase (conjugase, folylpolygammaglutamyl hydrolase)', other_designations='gamma-Glu-X carboxypeptidase', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63015079, end_position_on_the_genomic_accession=63039445, orientation='minus', exon_count=10, OMIM=601509, no_hdr0=''),
  8837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8837, CurrentID=0, Status='live', Symbol='CFLAR', Aliases='CASH, CASP8AP1, CLARP, Casper, FLAME, FLAME-1, FLAME1, FLIP, I-FLICE, MRIT, c-FLIP, c-FLIPL, c-FLIPR, c-FLIPS', description='CASP8 and FADD like apoptosis regulator', other_designations='FADD-like antiapoptotic molecule 1|MACH-related inducer of toxicity|caspase homolog|caspase-eight-related protein|caspase-like apoptosis regulatory protein|caspase-related inducer of apoptosis|cellular FLICE-like inhibitory protein|inhibitor of FLICE|usurpin beta', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201116154, end_position_on_the_genomic_accession=201172688, orientation='plus', exon_count=14, OMIM=603599, no_hdr0=''),
  8838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8838, CurrentID=0, Status='live', Symbol='WISP3', Aliases='CCN6, LIBC, PPAC, PPD, WISP-3', description='WNT1 inducible signaling pathway protein 3', other_designations='CCN family member 6', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112053516, end_position_on_the_genomic_accession=112069686, orientation='plus', exon_count=10, OMIM=603400, no_hdr0=''),
  8839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8839, CurrentID=0, Status='live', Symbol='WISP2', Aliases='CCN5, CT58, CTGF-L', description='WNT1 inducible signaling pathway protein 2', other_designations='CCN family member 5|WISP-2|connective tissue growth factor-like protein|connective tissue growth factor-related protein 58|wnt-1 signaling pathway protein 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44714993, end_position_on_the_genomic_accession=44727812, orientation='plus', exon_count=5, OMIM=603399, no_hdr0=''),
  8840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8840, CurrentID=0, Status='live', Symbol='WISP1', Aliases='CCN4c, WISP1i, WISP1tc, WISP1', description='WNT1 inducible signaling pathway protein 1', other_designations='CCN family member 4|CTC-458A3.8|WNT1 induced secreted protein 1|Wnt-1 inducible signaling pathway protein 1', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133191039, end_position_on_the_genomic_accession=133231690, orientation='plus', exon_count=6, OMIM=603398, no_hdr0=''),
  8841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8841, CurrentID=0, Status='live', Symbol='HDAC3', Aliases='HD3, RPD3, RPD3-2', description='histone deacetylase 3', other_designations='SMAP45', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141620876, end_position_on_the_genomic_accession=141636856, orientation='minus', exon_count=15, OMIM=605166, no_hdr0=''),
  8842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8842, CurrentID=0, Status='live', Symbol='PROM1', Aliases='AC133, CD133, CORD12, MCDR2, MSTP061, PROML1, RP41, STGD4', description='prominin 1', other_designations='antigen AC133|hProminin|hematopoietic stem cell antigen|prominin-like protein 1', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15968226, end_position_on_the_genomic_accession=16119884, orientation='minus', exon_count=39, OMIM=604365, no_hdr0=''),
  8843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8843, CurrentID=0, Status='live', Symbol='HCAR3', Aliases='GPR109B, HCA3, HM74, PUMAG, Puma-g', description='hydroxycarboxylic acid receptor 3', other_designations='G protein-coupled receptor 109B|G-protein coupled receptor 109B|G-protein coupled receptor HM74|G-protein coupled receptor HM74B|GTP-binding protein|hydroxy-carboxylic acid receptor 3|niacin receptor 2|nicotinic acid receptor 2|putative chemokine receptor', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122714756, end_position_on_the_genomic_accession=122716892, orientation='minus', exon_count=1, OMIM=606039, no_hdr0=''),
  8844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8844, CurrentID=0, Status='live', Symbol='KSR1', Aliases='KSR, RSU2', description='kinase suppressor of ras 1', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27456464, end_position_on_the_genomic_accession=27626441, orientation='plus', exon_count=30, OMIM=601132, no_hdr0=''),
  8846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8846, CurrentID=0, Status='live', Symbol='ALKBH1', Aliases='ABH, ABH1, ALKBH, alkB, hABH', description='alkB homolog 1, histone H2A dioxygenase', other_designations='DNA lyase ABH1|alkB, alkylation repair homolog 1|alkylated DNA repair protein alkB homolog 1|alkylation repair, alkB homolog|alpha-ketoglutarate-dependent dioxygenase ABH1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77672406, end_position_on_the_genomic_accession=77708032, orientation='minus', exon_count=6, OMIM=605345, no_hdr0=''),
  8847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8847, CurrentID=0, Status='live', Symbol='DLEU2', Aliases='1B4, BCMSUN, DLB2, LEU2, LINC00022, MIR15AHG, NCRNA00022, RFP2OS, TRIM13OS', description='deleted in lymphocytic leukemia 2 (non-protein coding)', other_designations='deleted in lymphocytic leukemia, 2|leukemia associated gene 2|long intergenic non-protein coding RNA 22|mir-15a-16-1 cluster host gene (non-protein coding)|ret finger protein 2 opposite strand', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49982552, end_position_on_the_genomic_accession=50125541, orientation='minus', exon_count=11, OMIM=605766, no_hdr0=''),
  8848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8848, CurrentID=0, Status='live', Symbol='TSC22D1', Aliases='Ptg-2, TGFB1I4, TSC22', description='TSC22 domain family member 1', other_designations='TGFB-stimulated clone 22 homolog|TGFbeta-stimulated clone 22|TSC22 domain family, member 1|cerebral protein 2|regulatory protein TSC-22|transcriptional regulator TSC-22|transforming growth factor beta-1-induced transcript 4 protein|transforming growth factor beta-stimulated protein TSC-22', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44432143, end_position_on_the_genomic_accession=44576565, orientation='minus', exon_count=9, OMIM=607715, no_hdr0=''),
  8850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8850, CurrentID=0, Status='live', Symbol='KAT2B', Aliases='CAF, P/CAF, PCAF', description='K(lysine) acetyltransferase 2B', other_designations='CREBBP-associated factor|histone acetylase PCAF|histone acetyltransferase PCAF|lysine acetyltransferase 2B|p300/CBP-associated factor', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20040032, end_position_on_the_genomic_accession=20154404, orientation='plus', exon_count=20, OMIM=602303, no_hdr0=''),
  8851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8851, CurrentID=0, Status='live', Symbol='CDK5R1', Aliases='CDK5P35, CDK5R, NCK5A, p23, p25, p35, p35nck5a', description='cyclin-dependent kinase 5, regulatory subunit 1 (p35)', other_designations='CDK5 activator 1|TPKII regulatory subunit|cyclin-dependent kinase 5 regulatory subunit 1|neuronal CDK5 activator|regulatory partner for CDK5 kinase|tau protein kinase II 23kDa subunit', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32487087, end_position_on_the_genomic_accession=32491254, orientation='plus', exon_count=2, OMIM=603460, no_hdr0=''),
  8852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8852, CurrentID=0, Status='live', Symbol='AKAP4', Aliases='AKAP 82, AKAP-4, AKAP82, CT99, FSC1, HI, PRKA4, hAKAP82, p82', description='A-kinase anchoring protein 4', other_designations='A kinase (PRKA) anchor protein 4|A-kinase anchor protein 82 kDa|cancer/testis antigen 99|major sperm fibrous sheath protein|protein kinase A anchoring protein 4|protein kinase A-anchoring protein 4|testis-specific gene HI', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50190768, end_position_on_the_genomic_accession=50201013, orientation='minus', exon_count=7, OMIM=300185, no_hdr0=''),
  8853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8853, CurrentID=0, Status='live', Symbol='ASAP2', Aliases='AMAP2, CENTB3, DDEF2, PAG3, PAP, Pap-alpha, SHAG1', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain 2', other_designations='PYK2 C terminus-associated protein|centaurin, beta 3|development and differentiation enhancing factor 2|development and differentiation-enhancing factor 2|paxillin-associated protein with ARF GAP activity 3|pyk2 C-terminus-associated protein', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9206765, end_position_on_the_genomic_accession=9405683, orientation='plus', exon_count=32, OMIM=603817, no_hdr0=''),
  8854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8854, CurrentID=0, Status='live', Symbol='ALDH1A2', Aliases='RALDH(II), RALDH2, RALDH2-T', description='aldehyde dehydrogenase 1 family member A2', other_designations='RALDH 2|aldehyde dehydrogenase 1 family, member A2|retinaldehyde-specific dehydrogenase type 2', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57953424, end_position_on_the_genomic_accession=58065923, orientation='minus', exon_count=14, OMIM=603687, no_hdr0=''),
  8856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8856, CurrentID=0, Status='live', Symbol='NR1I2', Aliases='BXR, ONR1, PAR, PAR1, PAR2, PARq, PRR, PXR, SAR, SXR', description='nuclear receptor subfamily 1 group I member 2', other_designations='nuclear receptor subfamily 1, group I, member 2|orphan nuclear receptor PAR1|orphan nuclear receptor PXR|pregnane X nuclear receptor variant 2|pregnane X receptor|steroid and xenobiotic receptor', map_location='3q12-q13.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119780484, end_position_on_the_genomic_accession=119818485, orientation='plus', exon_count=10, OMIM=603065, no_hdr0=''),
  8857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8857, CurrentID=0, Status='live', Symbol='FCGBP', Aliases='FC(GAMMA)BP', description='Fc fragment of IgG binding protein', other_designations='Human Fc gamma BP|IgG Fc binding protein|fcgamma-binding protein antigen|fcgammaBP', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39863323, end_position_on_the_genomic_accession=39934626, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  8858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8858, CurrentID=0, Status='live', Symbol='PROZ', Aliases='PZ', description='protein Z, vitamin K-dependent plasma glycoprotein', other_designations='vitamin K-dependent protein Z precursor variant 1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113155061, end_position_on_the_genomic_accession=113172386, orientation='plus', exon_count=11, OMIM=176895, no_hdr0=''),
  8859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8859, CurrentID=0, Status='live', Symbol='STK19', Aliases='D6S60, D6S60E, G11, HLA-RP1, RP1', description='serine/threonine kinase 19', other_designations='MHC class III HLA-RP1|nuclear serine/threonine protein kinase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31971175, end_position_on_the_genomic_accession=31981446, orientation='plus', exon_count=9, OMIM=604977, no_hdr0=''),
  402076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402076, CurrentID=0, Status='live', Symbol='LOC402076', Aliases='', description='F-box and leucine-rich repeat protein 12 pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68023697, end_position_on_the_genomic_accession=68024575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8861, CurrentID=0, Status='live', Symbol='LDB1', Aliases='CLIM2, NLI', description='LIM domain binding 1', other_designations='CLIM-2|LDB-1|LIM domain-binding factor CLIM2|LIM domain-binding factor-1|carboxy terminal LIM domain protein 2|carboxyl-terminal LIM domain-binding protein 2|hLdb1|nuclear LIM interactor', map_location='10q24-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102106558, end_position_on_the_genomic_accession=102120453, orientation='minus', exon_count=12, OMIM=603451, no_hdr0=''),
  8862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8862, CurrentID=0, Status='live', Symbol='APLN', Aliases='APEL, XNPEP2', description='apelin', other_designations='AGTRL1 ligand|APJ endogenous ligand', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129645259, end_position_on_the_genomic_accession=129654956, orientation='minus', exon_count=3, OMIM=300297, no_hdr0=''),
  8863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8863, CurrentID=0, Status='live', Symbol='PER3', Aliases='GIG13', description='period circadian clock 3', other_designations='cell growth-inhibiting gene 13 protein|circadian clock protein PERIOD 3|growth-inhibiting protein 13|hPER3|period 3|period circadian protein 3|period homolog 3', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7784380, end_position_on_the_genomic_accession=7845181, orientation='plus', exon_count=24, OMIM=603427, no_hdr0=''),
  8864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8864, CurrentID=0, Status='live', Symbol='PER2', Aliases='FASPS, FASPS1', description='period circadian clock 2', other_designations='circadian clock protein PERIOD 2|hPER2|period 2|period circadian protein 2|period homolog 2', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238244038, end_position_on_the_genomic_accession=238290102, orientation='minus', exon_count=25, OMIM=603426, no_hdr0=''),
  8867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8867, CurrentID=0, Status='live', Symbol='SYNJ1', Aliases='INPP5G, PARK20', description='synaptojanin 1', other_designations="inositol 5'-phosphatase (synaptojanin 1)|synaptic inositol 1,4,5-trisphosphate 5-phosphatase 1|synaptic inositol-1,4,5-trisphosphate 5-phosphatase 1|synaptojanin-1, polyphosphoinositide phosphatase", map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32628759, end_position_on_the_genomic_accession=32728040, orientation='minus', exon_count=34, OMIM=604297, no_hdr0=''),
  8869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8869, CurrentID=0, Status='live', Symbol='ST3GAL5', Aliases='SATI, SIAT9, SIATGM3S, ST3GalV', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 5', other_designations='CMP-NeuAc:lactosylceramide alpha-2,3-sialyltransferase|GM3 synthase|ST3 beta-galactoside alpha-23-sialyltransferase 5|ST3Gal V|alpha 2,3-sialyltransferase V|ganglioside GM3 synthase|sialyltransferase 9 (CMP-NeuAc:lactosylceramide alpha-2,3-sialyltransferase; GM3 synthase)', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85839147, end_position_on_the_genomic_accession=85889034, orientation='minus', exon_count=12, OMIM=604402, no_hdr0=''),
  8870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8870, CurrentID=0, Status='live', Symbol='IER3', Aliases='DIF-2, DIF2, GLY96, IEX-1, IEX-1L, IEX1, PRG1', description='immediate early response 3', other_designations='PACAP-responsive gene 1 protein|anti-death protein|differentiation-dependent gene 2 protein|expressed in pancreatic carcinoma|gly96, mouse, homolog of|immediate early protein GLY96|immediate early response 3 protein|immediately early gene X-1|protein DIF-2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30743199, end_position_on_the_genomic_accession=30744550, orientation='minus', exon_count=2, OMIM=602996, no_hdr0=''),
  8871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8871, CurrentID=0, Status='live', Symbol='SYNJ2', Aliases='INPP5H', description='synaptojanin 2', other_designations="inositol phosphate 5'-phosphatase 2|synaptic inositol 1,4,5-trisphosphate 5-phosphatase 2|synaptic inositol-1,4,5-trisphosphate 5-phosphatase 2", map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157981856, end_position_on_the_genomic_accession=158099176, orientation='plus', exon_count=33, OMIM=609410, no_hdr0=''),
  8872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8872, CurrentID=0, Status='live', Symbol='CDC123', Aliases='C10orf7, D123', description='cell division cycle 123', other_designations='HT-1080|PZ32|cell division cycle 123 homolog', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12195962, end_position_on_the_genomic_accession=12250590, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  8874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8874, CurrentID=0, Status='live', Symbol='ARHGEF7', Aliases='BETA-PIX, COOL-1, COOL1, Nbla10314, P50, P50BP, P85, P85COOL1, P85SPR, PAK3, PIXB', description='Rho guanine nucleotide exchange factor 7', other_designations='PAK-interacting exchange factor beta|Rho guanine nucleotide exchange factor (GEF) 7|SH3 domain-containing proline-rich protein|guanine nucleotide exchange factor 7', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111115277, end_position_on_the_genomic_accession=111305734, orientation='plus', exon_count=29, OMIM=605477, no_hdr0=''),
  8875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8875, CurrentID=0, Status='live', Symbol='VNN2', Aliases='FOAP-4, GPI-80', description='vanin 2', other_designations='Vannin 2|glycosylphosphatidyl inositol-anchored protein GPI-80|pantetheinase|vanin-2', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132743870, end_position_on_the_genomic_accession=132763459, orientation='minus', exon_count=11, OMIM=603571, no_hdr0=''),
  8876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8876, CurrentID=0, Status='live', Symbol='VNN1', Aliases='HDLCQ8, Tiff66', description='vanin 1', other_designations='pantetheine hydrolase|vanin-1|vannin 1|vascular non-inflammatory molecule 1', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132680858, end_position_on_the_genomic_accession=132714055, orientation='minus', exon_count=7, OMIM=603570, no_hdr0=''),
  8877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8877, CurrentID=0, Status='live', Symbol='SPHK1', Aliases='SPHK', description='sphingosine kinase 1', other_designations='SK 1|SPK 1', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76381567, end_position_on_the_genomic_accession=76387860, orientation='plus', exon_count=8, OMIM=603730, no_hdr0=''),
  8878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8878, CurrentID=0, Status='live', Symbol='SQSTM1', Aliases='A170, FTDALS3, OSIL, PDB3, ZIP3, p60, p62, p62B', description='sequestosome 1', other_designations='EBI3-associated protein of 60 kDa|EBI3-associated protein p60|EBIAP|oxidative stress induced like|phosphotyrosine independent ligand for the Lck SH2 domain p62|phosphotyrosine-independent ligand for the Lck SH2 domain of 62 kDa|ubiquitin-binding protein p62', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179806388, end_position_on_the_genomic_accession=179838078, orientation='plus', exon_count=12, OMIM=601530, no_hdr0=''),
  8879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8879, CurrentID=0, Status='live', Symbol='SGPL1', Aliases='S1PL, SPL', description='sphingosine-1-phosphate lyase 1', other_designations='SP-lyase 1|SPL 1|hSPL|sphingosine-1-phosphate aldolase', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70815948, end_position_on_the_genomic_accession=70881189, orientation='plus', exon_count=17, OMIM=603729, no_hdr0=''),
  8880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8880, CurrentID=0, Status='live', Symbol='FUBP1', Aliases='FBP, FUBP, hDH V', description='far upstream element (FUSE) binding protein 1', other_designations='DNA helicase V', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77944055, end_position_on_the_genomic_accession=77979205, orientation='minus', exon_count=22, OMIM=603444, no_hdr0=''),
  8881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8881, CurrentID=0, Status='live', Symbol='CDC16', Aliases='ANAPC6, APC6, CUT9', description='cell division cycle 16', other_designations='CDC16Hs|anaphase-promoting complex subunit 6|anaphase-promoting complex, subunit 6|cell division cycle 16 homolog|cyclosome subunit 6', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114234845, end_position_on_the_genomic_accession=114272675, orientation='plus', exon_count=19, OMIM=603461, no_hdr0=''),
  8882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8882, CurrentID=0, Status='live', Symbol='ZPR1', Aliases='ZNF259', description='ZPR1 zinc finger', other_designations='zinc finger protein 259', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116778560, end_position_on_the_genomic_accession=116788023, orientation='minus', exon_count=14, OMIM=603901, no_hdr0=''),
  8883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8883, CurrentID=0, Status='live', Symbol='NAE1', Aliases='A-116A10.1, APPBP1, HPP1, ula-1', description='NEDD8 activating enzyme E1 subunit 1', other_designations='APP-BP1|NEDD8-activating enzyme E1 subunit|amyloid beta precursor protein binding protein 1, 59kDa|amyloid beta precursor protein-binding protein 1, 59 kDa|amyloid beta precursor protein-binding protein 1, 59kD|amyloid protein-binding protein 1|proto-oncogene protein 1|protooncogene protein 1', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66802878, end_position_on_the_genomic_accession=66830977, orientation='minus', exon_count=21, OMIM=603385, no_hdr0=''),
  8884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8884, CurrentID=0, Status='live', Symbol='SLC5A6', Aliases='SMVT', description='solute carrier family 5 (sodium/multivitamin and iodide cotransporter), member 6', other_designations='Na(+)-dependent multivitamin transporter|Na+-dependent multivitamin transporter|solute carrier family 5 (sodium-dependent vitamin transporter), member 6|solute carrier family 5 member 6', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27199587, end_position_on_the_genomic_accession=27212307, orientation='minus', exon_count=19, OMIM=604024, no_hdr0=''),
  139957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139957, CurrentID=0, Status='live', Symbol='LOC139957', Aliases='', description='platelet activating factor acetylhydrolase 1b catalytic subunit 2 pseudogene', other_designations='platelet-activating factor acetylhydrolase 1b, catalytic subunit 2 (30kDa) pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=25029465, end_position_on_the_genomic_accession=25030631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8886, CurrentID=0, Status='live', Symbol='DDX18', Aliases='MrDb', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18', other_designations='DEAD box protein 18|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 18 (Myc-regulated)|Myc-regulated DEAD box protein', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117814679, end_position_on_the_genomic_accession=117832377, orientation='plus', exon_count=14, OMIM=606355, no_hdr0=''),
  8887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8887, CurrentID=0, Status='live', Symbol='TAX1BP1', Aliases='CALCOCO3, T6BP, TXBP151', description='Tax1 (human T-cell leukemia virus type I) binding protein 1', other_designations='TRAF6-binding protein', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27739373, end_position_on_the_genomic_accession=27829767, orientation='plus', exon_count=18, OMIM=605326, no_hdr0=''),
  8888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8888, CurrentID=0, Status='live', Symbol='MCM3AP', Aliases='GANP, MAP80, SAC3', description='minichromosome maintenance complex component 3 associated protein', other_designations='80 kDa MCM3-associated protein|MCM3 acetylating protein|MCM3 acetyltransferase|MCM3 import protein|MCM3 minichromosome maintenance deficient 3 associated protein|germinal center-associated nuclear protein|germinal-centre associated nuclear protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46235125, end_position_on_the_genomic_accession=46286559, orientation='minus', exon_count=31, OMIM=603294, no_hdr0=''),
  8890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8890, CurrentID=0, Status='live', Symbol='EIF2B4', Aliases='EIF-2B, EIF2B, EIF2Bdelta', description='eukaryotic translation initiation factor 2B subunit delta', other_designations='eIF-2B GDP-GTP exchange factor subunit delta|eukaryotic translation initiation factor 2B subunit 4 delta|eukaryotic translation initiation factor 2B, subunit 4 delta, 67kDa|translation initiation factor eIF-2b delta subunit', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27364352, end_position_on_the_genomic_accession=27370457, orientation='minus', exon_count=13, OMIM=606687, no_hdr0=''),
  8891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8891, CurrentID=0, Status='live', Symbol='EIF2B3', Aliases='EIF-2B, EIF2Bgamma', description='eukaryotic translation initiation factor 2B subunit gamma', other_designations='eIF-2B GDP-GTP exchange factor subunit gamma|eukaryotic translation initiation factor 2B, subunit 3 gamma, 58kDa', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44850522, end_position_on_the_genomic_accession=44986722, orientation='minus', exon_count=12, OMIM=606273, no_hdr0=''),
  8892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8892, CurrentID=0, Status='live', Symbol='EIF2B2', Aliases='EIF-2Bbeta, EIF2B', description='eukaryotic translation initiation factor 2B subunit beta', other_designations='S20I15|S20III15|eIF-2B GDP-GTP exchange factor subunit beta|eukaryotic translation initiation factor 2B, subunit 2 beta, 39kDa', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75002909, end_position_on_the_genomic_accession=75009591, orientation='plus', exon_count=8, OMIM=606454, no_hdr0=''),
  8893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8893, CurrentID=0, Status='live', Symbol='EIF2B5', Aliases='CACH, CLE, EIF-2B, EIF2Bepsilon, LVWM', description='eukaryotic translation initiation factor 2B subunit epsilon', other_designations='eIF-2B GDP-GTP exchange factor subunit epsilon|eukaryotic translation initiation factor 2B, subunit 5 epsilon, 82kDa', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184134435, end_position_on_the_genomic_accession=184145311, orientation='plus', exon_count=17, OMIM=603945, no_hdr0=''),
  8894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8894, CurrentID=0, Status='live', Symbol='EIF2S2', Aliases='EIF2, EIF2B, EIF2beta, PPP1R67, eIF-2-beta', description='eukaryotic translation initiation factor 2 subunit beta', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta, 38kDa|protein phosphatase 1, regulatory subunit 67', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34088309, end_position_on_the_genomic_accession=34112279, orientation='minus', exon_count=9, OMIM=603908, no_hdr0=''),
  8895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8895, CurrentID=0, Status='live', Symbol='CPNE3', Aliases='CPN3, PRO1071', description='copine III', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86514427, end_position_on_the_genomic_accession=86561498, orientation='plus', exon_count=17, OMIM=604207, no_hdr0=''),
  8896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8896, CurrentID=0, Status='live', Symbol='BUD31', Aliases='Cwc14, EDG-2, EDG2, G10, YCR063W, fSAP17', description='BUD31 homolog', other_designations='G10 maternal transcript homolog|functional spliceosome-associated protein 17|maternal G10 transcript|protein EDG-2|protein G10 homolog', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99408641, end_position_on_the_genomic_accession=99419616, orientation='plus', exon_count=7, OMIM=603477, no_hdr0=''),
  8897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8897, CurrentID=0, Status='live', Symbol='MTMR3', Aliases='FYVE-DSP1, ZFYVE10', description='myotubularin related protein 3', other_designations='FYVE (Fab1 YGLO23 Vsp27 EEA1 domain) dual-specificity protein phosphatase|FYVE domain-containing dual specificity protein phosphatase 1|phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase|zinc finger FYVE domain-containing protein 10|zinc finger, FYVE domain containing 10', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29883165, end_position_on_the_genomic_accession=30030868, orientation='plus', exon_count=23, OMIM=603558, no_hdr0=''),
  8898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8898, CurrentID=0, Status='live', Symbol='MTMR2', Aliases='CMT4B, CMT4B1', description='myotubularin related protein 2', other_designations='phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95832880, end_position_on_the_genomic_accession=95924207, orientation='minus', exon_count=20, OMIM=603557, no_hdr0=''),
  8899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8899, CurrentID=0, Status='live', Symbol='PRPF4B', Aliases='PR4H, PRP4, PRP4H, PRP4K, dJ1013A10.1', description='pre-mRNA processing factor 4B', other_designations='PRP4 kinase|PRP4 pre-mRNA processing factor 4 homolog B|PRP4 pre-mRNA-processing factor 4 homolog|dJ1013A10.1 (PRP4 protein kinase homolog)|protein-serine/threonine kinase', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4021329, end_position_on_the_genomic_accession=4064983, orientation='plus', exon_count=20, OMIM=602338, no_hdr0=''),
  8900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8900, CurrentID=0, Status='live', Symbol='CCNA1', Aliases='CT146', description='cyclin A1', other_designations='', map_location='13q12.3-q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36431030, end_position_on_the_genomic_accession=36442882, orientation='plus', exon_count=11, OMIM=604036, no_hdr0=''),
  402117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402117, CurrentID=0, Status='live', Symbol='VWC2L', Aliases='', description='von Willebrand factor C domain containing protein 2-like', other_designations='brorin-like', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214411127, end_position_on_the_genomic_accession=214579037, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  8904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8904, CurrentID=0, Status='live', Symbol='CPNE1', Aliases='COPN1, CPN1', description='copine I', other_designations='chromobindin 17', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35626031, end_position_on_the_genomic_accession=35664956, orientation='minus', exon_count=19, OMIM=604205, no_hdr0=''),
  8905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8905, CurrentID=0, Status='live', Symbol='AP1S2', Aliases='DC22, MRX59, MRXS21, MRXS5, MRXSF, PGS, SIGMA1B', description='adaptor related protein complex 1 sigma 2 subunit', other_designations='adapter-related protein complex 1 sigma-1B subunit|adaptor protein complex AP-1 sigma-1B subunit|adaptor-related protein complex 1 sigma 2 subunit|adaptor-related protein complex 1 subunit sigma-1B|adaptor-related protein complex 1, sigma 2 subunit|clathrin adaptor complex AP1 sigma 1B subunit|clathrin assembly protein complex 1 sigma-1B small chain|golgi adaptor HA1/AP1 adaptin sigma-1B subunit|sigma1B-adaptin', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15825806, end_position_on_the_genomic_accession=15855014, orientation='minus', exon_count=10, OMIM=300629, no_hdr0=''),
  8906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8906, CurrentID=0, Status='live', Symbol='AP1G2', Aliases='G2AD', description='adaptor related protein complex 1 gamma 2 subunit', other_designations='adaptor-related protein complex 1 gamma 2 subunit|adaptor-related protein complex 1, gamma 2 subunit|clathrin-associated/assembly/adaptor protein, large, gamma-2|gamma2-adaptin', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23559563, end_position_on_the_genomic_accession=23569545, orientation='minus', exon_count=25, OMIM=603534, no_hdr0=''),
  8907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8907, CurrentID=0, Status='live', Symbol='AP1M1', Aliases='AP47, CLAPM2, CLTNM, MU-1A', description='adaptor related protein complex 1 mu 1 subunit', other_designations='AP-mu chain family member mu1A|HA1 47 kDa subunit|adapter-related protein complex 1 subunit mu-1|adaptor protein complex AP-1 mu-1 subunit|adaptor protein complex AP-1 subunit mu-1|adaptor-related protein complex 1 mu 1 subunit|adaptor-related protein complex 1 mu-1 subunit|adaptor-related protein complex 1 subunit mu-1|adaptor-related protein complex 1, mu 1 subunit|clathrin adaptor protein AP47|clathrin assembly protein complex 1 medium chain 1|clathrin assembly protein complex 1 mu-1 medium chain 1|clathrin assembly protein complex 1, medium chain|clathrin assembly protein complex AP1, mu subunit|clathrin coat assembly protein AP47|clathrin coat-associated protein AP47|golgi adaptor AP-1 47 kDa protein|golgi adaptor HA1/AP1 adaptin mu-1 subunit|mu-adaptin 1|mu1A-adaptin', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16197854, end_position_on_the_genomic_accession=16235345, orientation='plus', exon_count=15, OMIM=603535, no_hdr0=''),
  8908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8908, CurrentID=0, Status='live', Symbol='GYG2', Aliases='GN-2, GN2', description='glycogenin 2', other_designations='glycogenin glucosyltransferase', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2828822, end_position_on_the_genomic_accession=2882820, orientation='plus', exon_count=13, OMIM=300198, no_hdr0=''),
  8909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8909, CurrentID=0, Status='live', Symbol='ENDOU', Aliases='P11, PP11, PRSS26', description='endonuclease, poly(U) specific', other_designations='22 serine protease|26 serine protease|endonuclease, polyU-specific|placental protein 11|protein endoU|uridylate-specific endoribonuclease', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47709734, end_position_on_the_genomic_accession=47725572, orientation='minus', exon_count=10, OMIM=606720, no_hdr0=''),
  8910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8910, CurrentID=0, Status='live', Symbol='SGCE', Aliases='DYT11, ESG', description='sarcoglycan epsilon', other_designations='dystonia 11, myoclonic|epsilon-SG|sarcoglycan, epsilon', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94585224, end_position_on_the_genomic_accession=94656209, orientation='minus', exon_count=15, OMIM=604149, no_hdr0=''),
  8911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8911, CurrentID=0, Status='live', Symbol='CACNA1I', Aliases='Cav3.3, ca(v)3.3', description='calcium channel, voltage-dependent, T type, alpha 1I subunit', other_designations='calcium channel, voltage-dependent, alpha 1I subunit|voltage-gated calcium channel subunit alpha Cav3.3', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39557969, end_position_on_the_genomic_accession=39689739, orientation='plus', exon_count=38, OMIM=608230, no_hdr0=''),
  8912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8912, CurrentID=0, Status='live', Symbol='CACNA1H', Aliases='CACNA1HB, Cav3.2, ECA6, EIG6', description='calcium channel, voltage-dependent, T type, alpha 1H subunit', other_designations='calcium channel, voltage-dependent, T type, alpha 1Hb subunit|low-voltage-activated calcium channel alpha1 3.2 subunit|low-voltage-activated calcium channel alpha13.2 subunit|voltage dependent t-type calcium channel alpha-1H subunit|voltage-gated calcium channel alpha subunit Cav3.2|voltage-gated calcium channel alpha subunit CavT.2|voltage-gated calcium channel subunit alpha Cav3.2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1153241, end_position_on_the_genomic_accession=1221772, orientation='plus', exon_count=37, OMIM=607904, no_hdr0=''),
  8913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8913, CurrentID=0, Status='live', Symbol='CACNA1G', Aliases='Ca(V)T.1, Cav3.1, NBR13', description='calcium channel, voltage-dependent, T type, alpha 1G subunit', other_designations='cav3.1c|voltage-dependent T-type calcium channel alpha 1G subunit|voltage-dependent calcium channel alpha 1G subunit|voltage-gated calcium channel subunit alpha Cav3.1', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50561068, end_position_on_the_genomic_accession=50629625, orientation='plus', exon_count=38, OMIM=604065, no_hdr0=''),
  8914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8914, CurrentID=0, Status='live', Symbol='TIMELESS', Aliases='TIM, TIM1, hTIM', description='timeless circadian clock', other_designations='Tof1 homolog|timeless circadian clock 1|timeless homolog', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56416373, end_position_on_the_genomic_accession=56449416, orientation='minus', exon_count=29, OMIM=603887, no_hdr0=''),
  8915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8915, CurrentID=0, Status='live', Symbol='BCL10', Aliases='CARMEN, CIPER, CLAP, IMD37, c-E10, mE10', description='B-cell CLL/lymphoma 10', other_designations='CARD containing molecule enhancing NF-kB|CARD-containing apoptotic signaling protein|CARD-containing molecule enhancing NF-kappa-B|CARD-containing proapoptotic protein|CED-3/ICH-1 prodomain homologous E10-like regulator|cCARMEN|caspase-recruiting domain-containing protein|cellular homolog of vCARMEN|cellular-E10|hCLAP|mammalian CARD-containing adapter molecule E10', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85265776, end_position_on_the_genomic_accession=85276904, orientation='minus', exon_count=5, OMIM=603517, no_hdr0=''),
  8916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8916, CurrentID=0, Status='live', Symbol='HERC3', Aliases='', description='HECT and RLD domain containing E3 ubiquitin protein ligase 3', other_designations='HECT domain and RCC1-like domain-containing protein 3|hect domain and RLD 3', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88592423, end_position_on_the_genomic_accession=88709297, orientation='plus', exon_count=29, OMIM=605200, no_hdr0=''),
  8917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8917, CurrentID=0, Status='live', Symbol='HERC2P1', Aliases='D15F37S2', description='hect domain and RLD 2 pseudogene 1', other_designations='', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28328485, end_position_on_the_genomic_accession=28335177, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8918, CurrentID=0, Status='live', Symbol='HERC2P6', Aliases='', description='hect domain and RLD 2 pseudogene 6', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23430839, end_position_on_the_genomic_accession=23435210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402135, CurrentID=0, Status='live', Symbol='OR5K2', Aliases='OR3-9', description='olfactory receptor family 5 subfamily K member 2', other_designations='olfactory receptor OR3-9|olfactory receptor, family 5, subfamily K, member 2', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98497681, end_position_on_the_genomic_accession=98498631, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271567, CurrentID=0, Status='live', Symbol='RPS26P41', Aliases='RPS26_13_1064', description='ribosomal protein S26 pseudogene 41', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73538193, end_position_on_the_genomic_accession=73538453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8924, CurrentID=0, Status='live', Symbol='HERC2', Aliases='D15F37S1, MRT38, SHEP1, jdf2, p528', description='HECT and RLD domain containing E3 ubiquitin protein ligase 2', other_designations='hect domain and RCC1-like domain-containing protein 2|hect domain and RLD 2|probable E3 ubiquitin-protein ligase HERC2', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28111037, end_position_on_the_genomic_accession=28322172, orientation='minus', exon_count=98, OMIM=605837, no_hdr0=''),
  8925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8925, CurrentID=0, Status='live', Symbol='HERC1', Aliases='p532, p619', description='HECT and RLD domain containing E3 ubiquitin protein ligase family member 1', other_designations='HECT domain and RCC1-like domain-containing protein 1|guanine nucleotide exchange factor p532|hect (homologous to the E6-AP (UBE3A) carboxyl terminus) domain and RCC1 (CHC1)-like domain (RLD) 1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63608618, end_position_on_the_genomic_accession=63833949, orientation='minus', exon_count=83, OMIM=605109, no_hdr0=''),
  8926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8926, CurrentID=0, Status='live', Symbol='SNURF', Aliases='', description='SNRPN upstream reading frame', other_designations='', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24954923, end_position_on_the_genomic_accession=24978582, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  8927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8927, CurrentID=0, Status='live', Symbol='BSN', Aliases='ZNF231', description='bassoon presynaptic cytomatrix protein', other_designations='bassoon (presynaptic cytomatrix protein)|neuronal double zinc finger protein|zinc finger protein 231', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49554489, end_position_on_the_genomic_accession=49671549, orientation='plus', exon_count=12, OMIM=604020, no_hdr0=''),
  8928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8928, CurrentID=0, Status='live', Symbol='FOXH1', Aliases='FAST-1, FAST1', description='forkhead box H1', other_designations='TGF-beta/activin signal transducer|fast-2|forkhead activin signal transducer 2|forkhead activin signal transducer-1|hFAST-1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144473732, end_position_on_the_genomic_accession=144476335, orientation='minus', exon_count=3, OMIM=603621, no_hdr0=''),
  8929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8929, CurrentID=0, Status='live', Symbol='PHOX2B', Aliases='NBLST2, NBPhox, PMX2B', description='paired-like homeobox 2b', other_designations='PHOX2B homeodomain protein|neuroblastoma Phox|neuroblastoma paired-type homeobox protein|paired mesoderm homeobox 2b', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41744082, end_position_on_the_genomic_accession=41748970, orientation='minus', exon_count=3, OMIM=603851, no_hdr0=''),
  8930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8930, CurrentID=0, Status='live', Symbol='MBD4', Aliases='MED1', description='methyl-CpG binding domain 4 DNA glycosylase', other_designations='3,N(4)-ethenocytosine glycosylase|G/5-fluorouracil mismatch glycosylase with biphasic kinetics|G/T mismatch glycosylase|G/U mismatch glycosylase|methyl-CpG binding domain protein 4|methyl-CpG-binding endonuclease 1|methyl-CpG-binding protein MBD4|mismatch-specific DNA N-glycosylase|putative methyl-CpG binding protein', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129430944, end_position_on_the_genomic_accession=129440179, orientation='minus', exon_count=8, OMIM=603574, no_hdr0=''),
  8932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8932, CurrentID=0, Status='live', Symbol='MBD2', Aliases='DMTase, NY-CO-41', description='methyl-CpG binding domain protein 2', other_designations='demethylase', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54151601, end_position_on_the_genomic_accession=54224788, orientation='minus', exon_count=8, OMIM=603547, no_hdr0=''),
  8933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8933, CurrentID=0, Status='live', Symbol='FAM127A', Aliases='CXX1, MAR8C, MART8C, Mar8, Mart8', description='family with sequence similarity 127 member A', other_designations='CAAX box 1|family with sequence similarity 127, member A|mammalian retrotransposon derived protein 8C', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135032303, end_position_on_the_genomic_accession=135033546, orientation='plus', exon_count=1, OMIM=300213, no_hdr0=''),
  8934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8934, CurrentID=0, Status='live', Symbol='RAB29', Aliases='RAB7L, RAB7L1', description='RAB29, member RAS oncogene family', other_designations='RAB7, member RAS oncogene family-like 1|rab-7-like protein 1|ras-related protein Rab-29', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205767986, end_position_on_the_genomic_accession=205775487, orientation='minus', exon_count=7, OMIM=603949, no_hdr0=''),
  8935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8935, CurrentID=0, Status='live', Symbol='SKAP2', Aliases='PRAP, RA70, SAPS, SCAP2, SKAP-HOM, SKAP55R', description='src kinase associated phosphoprotein 2', other_designations='Fyn-associated phosphoprotein SKAP55 homologue|Pyk2/RAFTK-associated protein|SKAP-55HOM|SKAP55 homolog|retinoic acid-induced protein 70|src family-associated phosphoprotein 2|src kinase-associated phosphoprotein of 55-related protein|src-associated adapter protein with PH and SH3 domains', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26667062, end_position_on_the_genomic_accession=26864743, orientation='minus', exon_count=14, OMIM=605215, no_hdr0=''),
  8936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8936, CurrentID=0, Status='live', Symbol='WASF1', Aliases='SCAR1, WAVE, WAVE1', description='WAS protein family member 1', other_designations='WAS protein family, member 1|WASP family protein member 1|homology of dictyostelium scar 1|protein WAVE-1|verprolin homology domain-containing protein 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110099817, end_position_on_the_genomic_accession=110189874, orientation='minus', exon_count=13, OMIM=605035, no_hdr0=''),
  8938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8938, CurrentID=0, Status='live', Symbol='BAIAP3', Aliases='BAP3', description='BAI1-associated protein 3', other_designations='BAI-associated protein 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1333605, end_position_on_the_genomic_accession=1349441, orientation='plus', exon_count=36, OMIM=604009, no_hdr0=''),
  8939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8939, CurrentID=0, Status='live', Symbol='FUBP3', Aliases='FBP3', description='far upstream element (FUSE) binding protein 3', other_designations='FUSE-binding protein 3', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130578857, end_position_on_the_genomic_accession=130638352, orientation='plus', exon_count=24, OMIM=603536, no_hdr0=''),
  8940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8940, CurrentID=0, Status='live', Symbol='TOP3B', Aliases='TOP3B1', description='topoisomerase (DNA) III beta', other_designations='DNA topoisomerase III beta-1|topoisomerase III beta', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21957025, end_position_on_the_genomic_accession=21982843, orientation='minus', exon_count=21, OMIM=603582, no_hdr0=''),
  8941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8941, CurrentID=0, Status='live', Symbol='CDK5R2', Aliases='NCK5AI, P39, p39nck5ai', description='cyclin-dependent kinase 5, regulatory subunit 2 (p39)', other_designations='CDK5 activator 2|cyclin-dependent kinase 5 activator isoform p39i|cyclin-dependent kinase 5 regulatory subunit 2|neuronal CDK5 activator isoform|p39I', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218959628, end_position_on_the_genomic_accession=218962155, orientation='plus', exon_count=1, OMIM=603764, no_hdr0=''),
  8942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8942, CurrentID=0, Status='live', Symbol='KYNU', Aliases='KYNUU', description='kynureninase', other_designations='L-kynurenine hydrolase', map_location='2q22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142748928, end_position_on_the_genomic_accession=143042316, orientation='plus', exon_count=23, OMIM=605197, no_hdr0=''),
  8943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8943, CurrentID=0, Status='live', Symbol='AP3D1', Aliases='ADTD, hBLVR', description='adaptor related protein complex 3, delta 1 subunit', other_designations='AP-3 complex delta subunit, partial CDS|AP-3 complex subunit delta|adapter-related protein complex 3 subunit delta-1|adaptin, delta|adaptor-related protein complex 3 subunit delta-1|adaptor-related protein complex 3, delta 1 subunit|delta adaptin|delta-adaptin, partial CDS|subunit of putative vesicle coat adaptor complex AP-3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2100988, end_position_on_the_genomic_accession=2151620, orientation='minus', exon_count=32, OMIM=607246, no_hdr0=''),
  8944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8944, CurrentID=0, Status='live', Symbol='SNORD73A', Aliases='RNU73, RNU73A, U73, U73a', description='small nucleolar RNA, C/D box 73A', other_designations='RNA, U73 small nucleolar|U73A small nucleolar RNA|U73A snoRNA|snoRNA, U73', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151103827, end_position_on_the_genomic_accession=151103891, orientation='plus', exon_count=1, OMIM=603568, no_hdr0=''),
  8945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8945, CurrentID=0, Status='live', Symbol='BTRC', Aliases='BETA-TRCP, FBW1A, FBXW1, FBXW1A, FWD1, bTrCP, bTrCP1, betaTrCP', description='beta-transducin repeat containing E3 ubiquitin protein ligase', other_designations='E3RSIkappaB|F-box and WD repeats protein beta-TrCP|F-box and WD-repeat protein 1B|beta-TrCP1|epididymis tissue protein Li 2a|pIkappaBalpha-E3 receptor subunit', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101354033, end_position_on_the_genomic_accession=101557321, orientation='plus', exon_count=16, OMIM=603482, no_hdr0=''),
  100271571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271571, CurrentID=0, Status='live', Symbol='RPS26P43', Aliases='RPS26_22_1185', description='ribosomal protein S26 pseudogene 43', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123049734, end_position_on_the_genomic_accession=123049966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376352, CurrentID=0, Status='live', Symbol='LOC105376352', Aliases='', description='uncharacterized LOC105376352', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2984142, end_position_on_the_genomic_accession=2986444, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100533716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533716, CurrentID=0, Status='live', Symbol='LOC100533716', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 53 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133141475, end_position_on_the_genomic_accession=133144654, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140028, CurrentID=0, Status='live', Symbol='GAPDHP19', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 19', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19327499, end_position_on_the_genomic_accession=19328765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533717, CurrentID=0, Status='live', Symbol='BCAR1P1', Aliases='', description='breast cancer anti-estrogen resistance 1 pseudogene 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20073700, end_position_on_the_genomic_accession=20074476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140032, CurrentID=0, Status='live', Symbol='RPS4Y2', Aliases='RPS4Y2P', description='ribosomal protein S4, Y-linked 2', other_designations='40S ribosomal protein S4, Y|ribosomal protein S4, Y-linked 2 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20756068, end_position_on_the_genomic_accession=20781056, orientation='plus', exon_count=8, OMIM=400030, no_hdr0=''),
  100533718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533718, CurrentID=0, Status='live', Symbol='BCAR1P2', Aliases='', description='breast cancer anti-estrogen resistance 1 pseudogene 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21080813, end_position_on_the_genomic_accession=21081589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402182, CurrentID=0, Status='live', Symbol='LOC402182', Aliases='', description='family with sequence similarity 64, member A pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105525957, end_position_on_the_genomic_accession=105527478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402183, CurrentID=0, Status='live', Symbol='RAC1P5', Aliases='Psi1Rac1', description='ras-related C3 botulinum toxin substrate 1 pseudogene 5', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107202653, end_position_on_the_genomic_accession=107204757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8968, CurrentID=0, Status='live', Symbol='HIST1H3F', Aliases='H3/i, H3FI', description='histone cluster 1, H3f', other_designations='H3 histone family, member I|histone 1, H3f|histone H3/i', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26250142, end_position_on_the_genomic_accession=26250607, orientation='minus', exon_count=1, OMIM=602816, no_hdr0=''),
  8969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8969, CurrentID=0, Status='live', Symbol='HIST1H2AG', Aliases='H2A.1b, H2A/p, H2AFP, H2AG, pH2A/f', description='histone cluster 1, H2ag', other_designations='H2A histone family, member P|H2A.1|histone 1, H2ag|histone H2A/p', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27133038, end_position_on_the_genomic_accession=27133535, orientation='plus', exon_count=1, OMIM=615012, no_hdr0=''),
  8970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8970, CurrentID=0, Status='live', Symbol='HIST1H2BJ', Aliases='H2B/r, H2BFR, H2BJ', description='histone cluster 1, H2bj', other_designations='H2B histone family, member R|histone 1, H2bj|histone H2B.1|histone H2B.r', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27132316, end_position_on_the_genomic_accession=27132796, orientation='minus', exon_count=1, OMIM=615044, no_hdr0=''),
  8971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8971, CurrentID=0, Status='live', Symbol='H1FX', Aliases='H1.10, H1X', description='H1 histone family member X', other_designations='H1 histone family, member X', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129314771, end_position_on_the_genomic_accession=129316277, orientation='minus', exon_count=1, OMIM=602785, no_hdr0=''),
  8972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8972, CurrentID=0, Status='live', Symbol='MGAM', Aliases='MG, MGA', description='maltase-glucoamylase', other_designations='brush border hydrolase|maltase-glucoamylase (alpha-glucosidase)', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141988888, end_position_on_the_genomic_accession=142106747, orientation='plus', exon_count=74, OMIM=154360, no_hdr0=''),
  8973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8973, CurrentID=0, Status='live', Symbol='CHRNA6', Aliases='CHNRA6', description='cholinergic receptor, nicotinic alpha 6', other_designations='acetylcholine receptor, nicotinic, alpha 6 (neuronal)|cholinergic receptor, nicotinic, alpha 6 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 6', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42752620, end_position_on_the_genomic_accession=42768786, orientation='minus', exon_count=6, OMIM=606888, no_hdr0=''),
  8974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8974, CurrentID=0, Status='live', Symbol='P4HA2', Aliases='', description='prolyl 4-hydroxylase, alpha polypeptide II', other_designations='4-PH alpha 2|4-PH alpha-2|C-P4Halpha(II)|collagen prolyl 4-hydroxylase alpha(II)|procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha polypeptide II|procollagen-proline,2-oxoglutarate-4-dioxygenase subunit alpha-2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132192009, end_position_on_the_genomic_accession=132227863, orientation='minus', exon_count=18, OMIM=600608, no_hdr0=''),
  8975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8975, CurrentID=0, Status='live', Symbol='USP13', Aliases='ISOT3, IsoT-3', description='ubiquitin specific peptidase 13 (isopeptidase T-3)', other_designations='deubiquitinating enzyme 13|ubiquitin specific protease 13 (isopeptidase T-3)|ubiquitin thioesterase 13|ubiquitin thiolesterase 13|ubiquitin-specific-processing protease 13', map_location='3q26.2-q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179653145, end_position_on_the_genomic_accession=179789401, orientation='plus', exon_count=21, OMIM=603591, no_hdr0=''),
  8976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8976, CurrentID=0, Status='live', Symbol='WASL', Aliases='N-WASP, NWASP, WASPB', description='Wiskott-Aldrich syndrome-like', other_designations='', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123681927, end_position_on_the_genomic_accession=123749071, orientation='minus', exon_count=11, OMIM=605056, no_hdr0=''),
  100271576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271576, CurrentID=0, Status='live', Symbol='RPS26P51', Aliases='RPS26_28_1470', description='ribosomal protein S26 pseudogene 51', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=7126224, end_position_on_the_genomic_accession=7126660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271577, CurrentID=0, Status='live', Symbol='RPS26P52', Aliases='RPS26_29_1475', description='ribosomal protein S26 pseudogene 52', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=13922253, end_position_on_the_genomic_accession=13922701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8985, CurrentID=0, Status='live', Symbol='PLOD3', Aliases='LH3', description='procollagen-lysine, 2-oxoglutarate 5-dioxygenase 3', other_designations='lysine hydroxylase 3|lysyl hydroxlase 3|lysyl hydroxylase 3', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101205977, end_position_on_the_genomic_accession=101217730, orientation='minus', exon_count=19, OMIM=603066, no_hdr0=''),
  8986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8986, CurrentID=0, Status='live', Symbol='RPS6KA4', Aliases='MSK2, RSK-B', description='ribosomal protein S6 kinase, 90kDa, polypeptide 4', other_designations='90 kDa ribosomal protein S6 kinase 4|RSKB|S6K-alpha-4|mitogen- and stress-activated protein kinase 2|nuclear mitogen- and stress-activated protein kinase 2|ribosomal protein S6 kinase alpha 4|ribosomal protein kinase B', map_location='11q11-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64359153, end_position_on_the_genomic_accession=64372215, orientation='plus', exon_count=17, OMIM=603606, no_hdr0=''),
  8987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8987, CurrentID=0, Status='live', Symbol='STBD1', Aliases='GENEX3414, GENX-3414', description='starch binding domain 1', other_designations='genethonin 1|glycophagy cargo receptor STBD1', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76306026, end_position_on_the_genomic_accession=76311130, orientation='plus', exon_count=2, OMIM=607406, no_hdr0=''),
  8988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8988, CurrentID=0, Status='live', Symbol='HSPB3', Aliases='DHMN2C, HMN2C, HSPL27', description='heat shock protein family B (small) member 3', other_designations='HSP 17|heat shock 17 kDa protein|heat shock 27kDa protein 3|protein 3', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54455601, end_position_on_the_genomic_accession=54456384, orientation='plus', exon_count=1, OMIM=604624, no_hdr0=''),
  8989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8989, CurrentID=0, Status='live', Symbol='TRPA1', Aliases='ANKTM1, FEPS', description='transient receptor potential cation channel, subfamily A, member 1', other_designations='ankyrin-like with transmembrane domains 1|transformation-sensitive protein p120', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72021251, end_position_on_the_genomic_accession=72090009, orientation='minus', exon_count=29, OMIM=604775, no_hdr0=''),
  8991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8991, CurrentID=0, Status='live', Symbol='SELENBP1', Aliases='HEL-S-134P, LPSB, SBP56, SP56, hSBP', description='selenium binding protein 1', other_designations='56 kDa selenium-binding protein|epididymis secretory sperm binding protein Li 134P', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151364302, end_position_on_the_genomic_accession=151372734, orientation='minus', exon_count=13, OMIM=604188, no_hdr0=''),
  8992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8992, CurrentID=0, Status='live', Symbol='ATP6V0E1', Aliases='ATP6H, ATP6V0E, M9.2, Vma21, Vma21p', description='ATPase, H+ transporting, lysosomal 9kDa, V0 subunit e1', other_designations='H(+)-transporting two-sector ATPase, subunit H|V-ATPase 9.2 kDa membrane accessory protein|V-ATPase H subunit|V-ATPase M9.2 subunit|V-ATPase subunit e 1|vacuolar ATP synthase subunit H|vacuolar proton pump H subunit|vacuolar proton pump subunit e 1|vacuolar proton-ATPase subunit M9.2', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172983760, end_position_on_the_genomic_accession=173034897, orientation='plus', exon_count=4, OMIM=603931, no_hdr0=''),
  8993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8993, CurrentID=0, Status='live', Symbol='PGLYRP1', Aliases='PGLYRP, PGRP, PGRP-S, PGRPS, TAG7, TNFSF3L', description='peptidoglycan recognition protein 1', other_designations='TNF superfamily, member 3 (LTB)-like (peptidoglycan recognition protein)', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46019153, end_position_on_the_genomic_accession=46023298, orientation='minus', exon_count=3, OMIM=604963, no_hdr0=''),
  8994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8994, CurrentID=0, Status='live', Symbol='LIMD1', Aliases='', description='LIM domains containing 1', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45594188, end_position_on_the_genomic_accession=45681263, orientation='plus', exon_count=10, OMIM=604543, no_hdr0=''),
  8995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8995, CurrentID=0, Status='live', Symbol='TNFSF18', Aliases='AITRL, GITRL, TL6, hGITRL', description='tumor necrosis factor superfamily member 18', other_designations='AITR ligand|GITR ligand|activation-inducible TNF-related ligand|glucocorticoid-induced TNF-related ligand|glucocorticoid-induced TNFR-related protein ligand|tumor necrosis factor (ligand) superfamily member 18|tumor necrosis factor (ligand) superfamily, member 18', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173041220, end_position_on_the_genomic_accession=173050963, orientation='minus', exon_count=3, OMIM=603898, no_hdr0=''),
  8996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8996, CurrentID=0, Status='live', Symbol='NOL3', Aliases='ARC, FCM, MYP, NOP, NOP30', description='nucleolar protein 3', other_designations='muscle-enriched cytoplasmic protein|nucleolar protein 3 (apoptosis repressor with CARD domain)|nucleolar protein of 30 kDa', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67170497, end_position_on_the_genomic_accession=67175737, orientation='plus', exon_count=6, OMIM=605235, no_hdr0=''),
  8997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8997, CurrentID=0, Status='live', Symbol='KALRN', Aliases='ARHGEF24, CHD5, CHDS5, DUET, DUO, HAPIP, TRAD', description='kalirin, RhoGEF kinase', other_designations='huntingtin-associated protein interacting protein (duo)|serine/threonine kinase with Dbl- and pleckstrin homology domains|serine/threonine-protein kinase with Dbl- and pleckstrin homology domain', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124033339, end_position_on_the_genomic_accession=124726325, orientation='plus', exon_count=66, OMIM=604605, no_hdr0=''),
  402214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402214, CurrentID=0, Status='live', Symbol='RPL13AP13', Aliases='RPL13A_4_562', description='ribosomal protein L13a pseudogene 13', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53206563, end_position_on_the_genomic_accession=53207178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8999, CurrentID=0, Status='live', Symbol='CDKL2', Aliases='KKIAMRE, P56', description='cyclin-dependent kinase-like 2 (CDC2-related kinase)', other_designations='CDC2-related kinase|p56 KKIAMRE protein kinase|protein kinase p56 KKIAMRE|serine/threonine protein kinase KKIAMRE|serine/threonine-protein kinase KKIAMRE', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75576494, end_position_on_the_genomic_accession=75630537, orientation='minus', exon_count=14, OMIM=603442, no_hdr0=''),
  402216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402216, CurrentID=0, Status='live', Symbol='HNRNPH1P3', Aliases='', description='heterogeneous nuclear ribonucleoprotein H1 pseudogene 3', other_designations='heterogeneous nuclear ribonucleoprotein H1 (H) pseudogene', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55838752, end_position_on_the_genomic_accession=55840025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9001, CurrentID=0, Status='live', Symbol='HAP1', Aliases='HAP2, HIP5, HLP, hHLP1', description='huntingtin-associated protein 1', other_designations='HAP-1|huntingtin-associated protein 2|neuroan 1', map_location='17q21.2-q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41722639, end_position_on_the_genomic_accession=41734646, orientation='minus', exon_count=11, OMIM=600947, no_hdr0=''),
  9002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9002, CurrentID=0, Status='live', Symbol='F2RL3', Aliases='PAR4', description='coagulation factor II (thrombin) receptor-like 3', other_designations='PAR-4|protease-activated receptor-4|thrombin receptor-like 3', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16889015, end_position_on_the_genomic_accession=16892019, orientation='plus', exon_count=2, OMIM=602779, no_hdr0=''),
  402220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402220, CurrentID=0, Status='live', Symbol='YTHDF1P1', Aliases='YTHDF1P', description='YTH domain family member 1 pseudogene 1', other_designations='YTH domain family 1 pseudogene|YTH domain family, member 1 pseudogene 1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97368587, end_position_on_the_genomic_accession=97370619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402221, CurrentID=0, Status='live', Symbol='LOC402221', Aliases='', description='actin, beta pseudogene', other_designations='actin pseudogene', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98338660, end_position_on_the_genomic_accession=98340147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402222, CurrentID=0, Status='live', Symbol='DDX18P4', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 4', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98679302, end_position_on_the_genomic_accession=98681911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271581, CurrentID=0, Status='live', Symbol='RPS27P20', Aliases='RPS27_11_1188', description='ribosomal protein S27 pseudogene 20', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129537837, end_position_on_the_genomic_accession=129538173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402226, CurrentID=0, Status='live', Symbol='CCT5P1', Aliases='CCT5-3P', description='chaperonin containing TCP1, subunit 5 (epsilon) pseudogene 1', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115512077, end_position_on_the_genomic_accession=115513697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9013, CurrentID=0, Status='live', Symbol='TAF1C', Aliases='MGC:39976, SL1, TAFI110, TAFI95', description='TATA-box binding protein associated factor, RNA polymerase I, C', other_designations='RNA polymerase I-specific TBP-associated factor 110 kDa|SL1, 110kD subunit|TATA box binding protein (TBP)-associated factor, RNA polymerase I, C, 110kD|TATA box binding protein (TBP)-associated factor, RNA polymerase I, C, 110kDa|TATA box binding protein associated factor 1C|TATA box-binding protein-associated factor 1C|TBP-associated factor 1C|transcription factor SL1|transcription initiation factor SL1/TIF-IB subunit C', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84177847, end_position_on_the_genomic_accession=84187070, orientation='minus', exon_count=17, OMIM=604905, no_hdr0=''),
  9014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9014, CurrentID=0, Status='live', Symbol='TAF1B', Aliases='MGC:9349, RAF1B, RAFI63, SL1, TAFI63', description='TATA-box binding protein associated factor, RNA polymerase I, B', other_designations='RNA polymerase I-specific TBP-associated factor 63 kDa|SL1, 63kD subunit|TATA box binding protein (TBP)-associated factor, RNA polymerase I, B, 63kD|TATA box binding protein (TBP)-associated factor, RNA polymerase I, B, 63kDa|TATA box binding protein associated factor 1B|TATA box-binding protein-associated factor 1B|TBP-associated factor 1B|transcription initiation factor SL1/TIF-IB subunit B', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9843442, end_position_on_the_genomic_accession=9934416, orientation='plus', exon_count=17, OMIM=604904, no_hdr0=''),
  9015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9015, CurrentID=0, Status='live', Symbol='TAF1A', Aliases='MGC:17061, RAFI48, SL1, TAFI48', description='TATA-box binding protein associated factor, RNA polymerase I, A', other_designations='RNA polymerase I-specific TBP-associated factor 48 kDa|SL1, 48kD subunit|TATA box binding protein (TBP)-associated factor, RNA polymerase I, A, 48kD|TATA box binding protein (TBP)-associated factor, RNA polymerase I, A, 48kDa|TATA box binding protein associated factor 1A|TATA box-binding protein-associated factor 1A|TBP-associated factor 1A|transcription factor SL1|transcription initiation factor SL1/TIF-IB subunit A', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222557902, end_position_on_the_genomic_accession=222590096, orientation='minus', exon_count=14, OMIM=604903, no_hdr0=''),
  9016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9016, CurrentID=0, Status='live', Symbol='SLC25A14', Aliases='BMCP1, UCP5', description='solute carrier family 25 (mitochondrial carrier, brain), member 14', other_designations='mitochondrial uncoupling protein 5', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130339888, end_position_on_the_genomic_accession=130373361, orientation='plus', exon_count=12, OMIM=300242, no_hdr0=''),
  285492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285492, CurrentID=0, Status='live', Symbol='LINC00955', Aliases='', description='long intergenic non-protein coding RNA 955', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3576869, end_position_on_the_genomic_accession=3590985, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9019, CurrentID=0, Status='live', Symbol='MPZL1', Aliases='MPZL1b, PZR, PZR1b, PZRa, PZRb', description='myelin protein zero-like 1', other_designations='immunoglobulin family transmembrane protein|protein zero related|protein zero-related', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167721950, end_position_on_the_genomic_accession=167791919, orientation='plus', exon_count=6, OMIM=604376, no_hdr0=''),
  9020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9020, CurrentID=0, Status='live', Symbol='MAP3K14', Aliases='FTDCR1B, HS, HSNIK, NIK', description='mitogen-activated protein kinase kinase kinase 14', other_designations='NF-kappa-beta-inducing kinase|serine/threonine-protein kinase NIK', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45263119, end_position_on_the_genomic_accession=45317064, orientation='minus', exon_count=17, OMIM=604655, no_hdr0=''),
  9021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9021, CurrentID=0, Status='live', Symbol='SOCS3', Aliases='ATOD4, CIS3, Cish3, SOCS-3, SSI-3, SSI3', description='suppressor of cytokine signaling 3', other_designations='STAT-induced STAT inhibitor 3|cytokine-inducible SH2 protein 3', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78356777, end_position_on_the_genomic_accession=78360079, orientation='minus', exon_count=2, OMIM=604176, no_hdr0=''),
  9022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9022, CurrentID=0, Status='live', Symbol='CLIC3', Aliases='', description='chloride intracellular channel 3', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136994608, end_position_on_the_genomic_accession=136996572, orientation='minus', exon_count=6, OMIM=606533, no_hdr0=''),
  9023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9023, CurrentID=0, Status='live', Symbol='CH25H', Aliases='C25H', description='cholesterol 25-hydroxylase', other_designations='cholesterol 25-monooxygenase|h25OH', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89205937, end_position_on_the_genomic_accession=89207314, orientation='minus', exon_count=1, OMIM=604551, no_hdr0=''),
  9024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9024, CurrentID=0, Status='live', Symbol='BRSK2', Aliases='C11orf7, PEN11B, SAD1, SADA, STK29', description='BR serine/threonine kinase 2', other_designations='BR serine/threonine-protein kinase 2|brain-selective kinase 2|brain-specific serine/threonine-protein kinase 2|protein kinase SAD1B|serine/threonine-protein kinase 29|serine/threonine-protein kinase SAD-A', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1389899, end_position_on_the_genomic_accession=1462689, orientation='plus', exon_count=27, OMIM=609236, no_hdr0=''),
  9025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9025, CurrentID=0, Status='live', Symbol='RNF8', Aliases='hRNF8', description='ring finger protein 8, E3 ubiquitin protein ligase', other_designations='C3HC4-type zinc finger protein|UBC13/UEV-interacting ring finger protein|ring finger protein (C3HC4 type) 8', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37353972, end_position_on_the_genomic_accession=37394738, orientation='plus', exon_count=9, OMIM=611685, no_hdr0=''),
  9026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9026, CurrentID=0, Status='live', Symbol='HIP1R', Aliases='HIP12, HIP3, ILWEQ', description='huntingtin interacting protein 1 related', other_designations='HIP-12|HIP1-related protein|huntingtin interacting protein 12|huntingtin-interacting protein 12', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122834498, end_position_on_the_genomic_accession=122862961, orientation='plus', exon_count=34, OMIM=605613, no_hdr0=''),
  9027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9027, CurrentID=0, Status='live', Symbol='NAT8', Aliases='ATase2, CCNAT, CML1, GLA, Hcml1, TSC501, TSC510', description='N-acetyltransferase 8 (GCN5-related, putative)', other_designations='N-acetyltransferase 8 (camello like)|acetyltransferase 2|camello-like protein 1|cysteinyl-conjugate N-acetyltransferase|kidney- and liver-specific gene product|probable N-acetyltransferase 8', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73640723, end_position_on_the_genomic_accession=73642410, orientation='minus', exon_count=2, OMIM=606716, no_hdr0=''),
  9028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9028, CurrentID=0, Status='live', Symbol='RHBDL1', Aliases='RHBDL, RRP', description='rhomboid, veinlet-like 1 (Drosophila)', other_designations='rhomboid, drosophila, homolog of|rhomboid-like protein 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=675666, end_position_on_the_genomic_accession=678271, orientation='plus', exon_count=8, OMIM=603264, no_hdr0=''),
  140101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140101, CurrentID=0, Status='live', Symbol='RBMY2MP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member M pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8904218, end_position_on_the_genomic_accession=8916344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9030, CurrentID=0, Status='live', Symbol='ERDA1', Aliases='', description='expanded repeat domain, CAG/CTG 1', other_designations='', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603279, no_hdr0=''),
  9031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9031, CurrentID=0, Status='live', Symbol='BAZ1B', Aliases='WBSCR10, WBSCR9, WSTF', description='bromodomain adjacent to zinc finger domain 1B', other_designations='bromodomain adjacent to zinc finger domain, 1B|hWALp2|transcription factor WSTF|williams syndrome transcription factor|williams-Beuren syndrome chromosomal region 10 protein|williams-Beuren syndrome chromosomal region 9 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73440398, end_position_on_the_genomic_accession=73522298, orientation='minus', exon_count=21, OMIM=605681, no_hdr0=''),
  9032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9032, CurrentID=0, Status='live', Symbol='TM4SF5', Aliases='', description='transmembrane 4 L six family member 5', other_designations='tetraspan transmembrane protein L6H|transmembrane 4 superfamily member 5', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4771892, end_position_on_the_genomic_accession=4783211, orientation='plus', exon_count=5, OMIM=604657, no_hdr0=''),
  9033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9033, CurrentID=0, Status='live', Symbol='PKD2L1', Aliases='PCL, PKD2L, PKDL, TRPP3', description='polycystic kidney disease 2-like 1', other_designations='polycystin-2 homolog|polycystin-2L1|polycystin-L|polycystin-L1|transient receptor potential cation channel, subfamily P, member 3', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100288145, end_position_on_the_genomic_accession=100346524, orientation='minus', exon_count=20, OMIM=604532, no_hdr0=''),
  9034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9034, CurrentID=0, Status='live', Symbol='CCRL2', Aliases='ACKR5, CKRX, CRAM, CRAM-A, CRAM-B, HCR', description='chemokine (C-C motif) receptor-like 2', other_designations='atypical chemokine receptor 5|chemokine receptor CCR11|chemokine receptor X|putative MCP-1 chemokine receptor', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46407166, end_position_on_the_genomic_accession=46409523, orientation='plus', exon_count=5, OMIM=608379, no_hdr0=''),
  9037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9037, CurrentID=0, Status='live', Symbol='SEMA5A', Aliases='SEMAF, semF', description='sema domain, seven thrombospondin repeats (type 1 and type 1-like), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 5A', other_designations='sema F|semaphorin 5A|semaphorin F|semaphorin-F', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9035026, end_position_on_the_genomic_accession=9546121, orientation='minus', exon_count=28, OMIM=609297, no_hdr0=''),
  9038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9038, CurrentID=0, Status='live', Symbol='TAAR5', Aliases='PNR', description='trace amine associated receptor 5', other_designations='hTaar5|putative neurotransmitter receptor|taR-5|trace amine receptor 5', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132588592, end_position_on_the_genomic_accession=132589738, orientation='minus', exon_count=1, OMIM=607405, no_hdr0=''),
  9039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9039, CurrentID=0, Status='live', Symbol='UBA3', Aliases='NAE2, UBE1C, hUBA3', description='ubiquitin-like modifier activating enzyme 3', other_designations='NEDD8-activating enzyme E1 subunit 2|NEDD8-activating enzyme E1C|Nedd8-activating enzyme hUba3|UBA3, ubiquitin-activating enzyme E1 homolog|ubiquitin-activating enzyme 3|ubiquitin-activating enzyme E1C (UBA3 homolog, yeast)|ubiquitin-activating enzyme E1C (homologous to yeast UBA3)|ubiquitin-like modifier-activating enzyme 3', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69054730, end_position_on_the_genomic_accession=69080373, orientation='minus', exon_count=18, OMIM=603172, no_hdr0=''),
  9040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9040, CurrentID=0, Status='live', Symbol='UBE2M', Aliases='UBC-RS2, UBC12, hUbc12', description='ubiquitin conjugating enzyme E2M', other_designations='NEDD8 carrier protein|NEDD8 protein ligase|ubiquitin carrier protein M|ubiquitin-conjugating enzyme E2M (UBC12 homolog, yeast)|ubiquitin-protein ligase M|yeast UBC12 homolog', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58555712, end_position_on_the_genomic_accession=58558976, orientation='minus', exon_count=6, OMIM=603173, no_hdr0=''),
  9043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9043, CurrentID=0, Status='live', Symbol='SPAG9', Aliases='CT89, HLC-6, HLC4, HLC6, JIP-4, JIP4, JLP, PHET, PIG6', description='sperm associated antigen 9', other_designations='JNK interacting protein|JNK/SAPK-associated protein|Max-binding protein|c-Jun NH2-terminal kinase-associated leucine zipper protein|cancer/testis antigen 89|human lung cancer oncogene 6 protein|lung cancer oncogene 4|mitogen-activated protein kinase 8-interacting protein 4|proliferation-inducing gene 6|protein highly expressed in testis|sperm associated antigen 9 variant 5|sperm surface protein|sunday driver 1', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50962174, end_position_on_the_genomic_accession=51120865, orientation='minus', exon_count=34, OMIM=605430, no_hdr0=''),
  9044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9044, CurrentID=0, Status='live', Symbol='BTAF1', Aliases='MOT1, TAF(II)170, TAF172, TAFII170', description='B-TFIID TATA-box binding protein associated factor 1', other_designations='ATP-dependent helicase BTAF1|B-TFIID transcription factor-associated 170 kDa subunit|BTAF1 RNA polymerase II, B-TFIID transcription factor-associated, 170kDa (Mot1 homolog, S. cerevisiae)|Mot1 homolog|TBP-associated factor 172', map_location='10q22-q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91923769, end_position_on_the_genomic_accession=92030998, orientation='plus', exon_count=41, OMIM=605191, no_hdr0=''),
  9045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9045, CurrentID=0, Status='live', Symbol='RPL14', Aliases='CAG-ISL-7, CTG-B33, L14, RL14, hRL14', description='ribosomal protein L14', other_designations='', map_location='3p22-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40457292, end_position_on_the_genomic_accession=40462372, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  9046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9046, CurrentID=0, Status='live', Symbol='DOK2', Aliases='p56DOK, p56dok-2', description='docking protein 2', other_designations='docking protein 2, 56kD|docking protein 2, 56kDa|downstream of tyrosine kinase 2|p56(dok-2)', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21908873, end_position_on_the_genomic_accession=21913743, orientation='minus', exon_count=5, OMIM=604997, no_hdr0=''),
  9047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9047, CurrentID=0, Status='live', Symbol='SH2D2A', Aliases='F2771, SCAP, TSAD, VRAP', description='SH2 domain containing 2A', other_designations='SH2 domain protein 2A|T cell specific adapter protein TSAd|T lymphocyte specific adaptor protein|VEGF receptor-associated protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156806243, end_position_on_the_genomic_accession=156816862, orientation='minus', exon_count=9, OMIM=604514, no_hdr0=''),
  9048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9048, CurrentID=0, Status='live', Symbol='ARTN', Aliases='ENOVIN, EVN, NBN', description='artemin', other_designations='neublastin|neurotrophic factor', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43931897, end_position_on_the_genomic_accession=43937241, orientation='plus', exon_count=3, OMIM=603886, no_hdr0=''),
  9049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9049, CurrentID=0, Status='live', Symbol='AIP', Aliases='ARA9, FKBP16, FKBP37, SMTPHN, XAP-2, XAP2', description='aryl hydrocarbon receptor interacting protein', other_designations='HBV X-associated protein 2|immunophilin homolog ARA9', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67483029, end_position_on_the_genomic_accession=67491108, orientation='plus', exon_count=7, OMIM=605555, no_hdr0=''),
  9050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9050, CurrentID=0, Status='live', Symbol='PSTPIP2', Aliases='MAYP', description='proline-serine-threonine phosphatase interacting protein 2', other_designations='PEST phosphatase-interacting protein 2', map_location='tdb7990', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45983536, end_position_on_the_genomic_accession=46072297, orientation='minus', exon_count=16, OMIM=616046, no_hdr0=''),
  9051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9051, CurrentID=0, Status='live', Symbol='PSTPIP1', Aliases='CD2BP1, CD2BP1L, CD2BP1S, H-PIP, PAPAS, PSTPIP', description='proline-serine-threonine phosphatase interacting protein 1', other_designations='CD2 antigen-binding protein 1|CD2 cytoplasmic tail-binding protein|CD2-binding protein 1|PEST phosphatase-interacting protein 1', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76995124, end_position_on_the_genomic_accession=77037337, orientation='plus', exon_count=19, OMIM=606347, no_hdr0=''),
  9052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9052, CurrentID=0, Status='live', Symbol='GPRC5A', Aliases='GPCR5A, PEIG-1, RAI3, RAIG1, TIG1', description='G protein-coupled receptor, class C, group 5, member A', other_designations='G protein-coupled receptor, family C, group 5, member A|G-protein coupled receptor family C group 5 member A|TPA induced gene 1|orphan G-protein-coupling receptor PEIG-1|phorbol ester induced gene 1|phorbol ester induced protein-1|retinoic acid induced 3|retinoic acid responsive|retinoic acid-induced gene 1 protein', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12891022, end_position_on_the_genomic_accession=12913666, orientation='plus', exon_count=4, OMIM=604138, no_hdr0=''),
  9053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9053, CurrentID=0, Status='live', Symbol='MAP7', Aliases='E-MAP-115, EMAP115', description='microtubule associated protein 7', other_designations='MAP-7|dJ325F22.2 (microtubule-associated protein 7 (EMAP115, E-MAP-115))|epithelial microtubule-associated protein of 115 kDa|microtubule-associated protein 7', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136342281, end_position_on_the_genomic_accession=136550819, orientation='minus', exon_count=21, OMIM=604108, no_hdr0=''),
  9054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9054, CurrentID=0, Status='live', Symbol='NFS1', Aliases='HUSSY-08, IscS, NIFS', description='NFS1 cysteine desulfurase', other_designations='NFS1 nitrogen fixation 1 homolog|nitrogen fixation 1 (S. cerevisiae, homolog)|nitrogen-fixing bacteria S-like protein', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35668688, end_position_on_the_genomic_accession=35699365, orientation='minus', exon_count=14, OMIM=603485, no_hdr0=''),
  9055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9055, CurrentID=0, Status='live', Symbol='PRC1', Aliases='ASE1', description='protein regulator of cytokinesis 1', other_designations='anaphase spindle elongation 1 homolog|protein regulating cytokinesis 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90966038, end_position_on_the_genomic_accession=90994651, orientation='minus', exon_count=15, OMIM=603484, no_hdr0=''),
  9056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9056, CurrentID=0, Status='live', Symbol='SLC7A7', Aliases='LAT3, LPI, MOP-2, Y+LAT1, y+LAT-1', description='solute carrier family 7 (amino acid transporter light chain, y+L system), member 7', other_designations='monocyte amino acid permease 2|solute carrier family 7 (cationic amino acid transporter, y+ system), member 7|y(+)L-type amino acid transporter 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22773222, end_position_on_the_genomic_accession=22819811, orientation='minus', exon_count=12, OMIM=603593, no_hdr0=''),
  9057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9057, CurrentID=0, Status='live', Symbol='SLC7A6', Aliases='LAT-2, LAT3, y+LAT-2', description='solute carrier family 7 (amino acid transporter light chain, y+L system), member 6', other_designations='amino acid permease|cationic amino acid transporter, y+ system|solute carrier family 7 (cationic amino acid transporter, y+ system), member 6|solute carrier family 7 member 6|y(+)L-type amino acid transporter 2|y+LAT2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68263830, end_position_on_the_genomic_accession=68301823, orientation='plus', exon_count=17, OMIM=605641, no_hdr0=''),
  9058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9058, CurrentID=0, Status='live', Symbol='SLC13A2', Aliases='NADC1, NaCT, NaDC-1, SDCT1', description='solute carrier family 13 (sodium-dependent dicarboxylate transporter), member 2', other_designations='Na(+)-coupled citrate transporter|Na(+)/dicarboxylate cotransporter 1|renal sodium/dicarboxylate cotransporter', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28473357, end_position_on_the_genomic_accession=28497781, orientation='plus', exon_count=14, OMIM=604148, no_hdr0=''),
  9060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9060, CurrentID=0, Status='live', Symbol='PAPSS2', Aliases='ATPSK2, BCYM4, SK2', description="3'-phosphoadenosine 5'-phosphosulfate synthase 2", other_designations="3-prime-phosphoadenosine 5-prime-phosphosulfate synthase 2|ATP sulfurylase/APS kinase 2|ATP sulfurylase/adenosine 5'-phosphosulfate kinase|PAPS synthase 2|PAPS synthetase 2|SK 2|bifunctional 3'-phosphoadenosine 5'-phosphosulfate synthethase 2", map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87659719, end_position_on_the_genomic_accession=87747705, orientation='plus', exon_count=13, OMIM=603005, no_hdr0=''),
  9061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9061, CurrentID=0, Status='live', Symbol='PAPSS1', Aliases='ATPSK1, PAPSS, SK1', description="3'-phosphoadenosine 5'-phosphosulfate synthase 1", other_designations='3-prime-phosphoadenosine 5-prime-phosphosulfate synthase 1|PAPS synthase 1|PAPSS 1|SK 1|adenylyl-sulfate kinase|sulfate adenylyltransferase|sulfurylase kinase 1', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107613666, end_position_on_the_genomic_accession=107720263, orientation='minus', exon_count=13, OMIM=603262, no_hdr0=''),
  9063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9063, CurrentID=0, Status='live', Symbol='PIAS2', Aliases='ARIP3, DIP, MIZ, MIZ1, PIASX, PIASX-ALPHA, PIASX-BETA, SIZ2, ZMIZ4', description='protein inhibitor of activated STAT 2', other_designations='DAB2-interacting protein|androgen receptor-interacting protein 3|msx-interacting zinc finger protein|protein inhibitor of activated STAT X|protein inhibitor of activated STAT, 2|zinc finger, MIZ-type containing 4', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46808246, end_position_on_the_genomic_accession=46920167, orientation='minus', exon_count=20, OMIM=603567, no_hdr0=''),
  9064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9064, CurrentID=0, Status='live', Symbol='MAP3K6', Aliases='ASK2, MAPKKK6, MEKK6', description='mitogen-activated protein kinase kinase kinase 6', other_designations='apoptosis signal-regulating kinase 2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27355179, end_position_on_the_genomic_accession=27368014, orientation='minus', exon_count=29, OMIM=604468, no_hdr0=''),
  9066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9066, CurrentID=0, Status='live', Symbol='SYT7', Aliases='IPCA-7, IPCA7, PCANAP7, SYT-VII, SYTVII', description='synaptotagmin 7', other_designations='prostate cancer-associated protein 7|synaptotagmin VII', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61513716, end_position_on_the_genomic_accession=61588396, orientation='minus', exon_count=18, OMIM=604146, no_hdr0=''),
  100271591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271591, CurrentID=0, Status='live', Symbol='RPS29P26', Aliases='RPS29_20_1655', description='ribosomal protein S29 pseudogene 26', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39597835, end_position_on_the_genomic_accession=39598002, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9068, CurrentID=0, Status='live', Symbol='ANGPTL1', Aliases='ANG3, ANGPT3, ARP1, AngY, UNQ162, dJ595C2.2', description='angiopoietin like 1', other_designations='ANG-3|angioarrestin|angiopoietin 3|angiopoietin Y1|angiopoietin-3|angiopoietin-like 1|angiopoietin-like protein 1|dJ595C2.2 (angiopoietin Y1)', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178849535, end_position_on_the_genomic_accession=178871353, orientation='minus', exon_count=7, OMIM=603874, no_hdr0=''),
  9069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9069, CurrentID=0, Status='live', Symbol='CLDN12', Aliases='', description='claudin 12', other_designations='', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90403334, end_position_on_the_genomic_accession=90415954, orientation='plus', exon_count=4, OMIM=611232, no_hdr0=''),
  9070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9070, CurrentID=0, Status='live', Symbol='ASH2L', Aliases='ASH21, ASH2L2, Bre2, ASH2L', description='ash2 (absent, small, or homeotic)-like (Drosophila)', other_designations='ASH2-like protein', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38104965, end_position_on_the_genomic_accession=38140080, orientation='plus', exon_count=18, OMIM=604782, no_hdr0=''),
  9071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9071, CurrentID=0, Status='live', Symbol='CLDN10', Aliases='CPETRL3, OSP-L', description='claudin 10', other_designations='OSP-like protein', map_location='13q31-q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95433599, end_position_on_the_genomic_accession=95579756, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  9073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9073, CurrentID=0, Status='live', Symbol='CLDN8', Aliases='HEL-S-79', description='claudin 8', other_designations='epididymis secretory protein Li 79', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30214006, end_position_on_the_genomic_accession=30216151, orientation='minus', exon_count=1, OMIM=611231, no_hdr0=''),
  9074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9074, CurrentID=0, Status='live', Symbol='CLDN6', Aliases='', description='claudin 6', other_designations='skullin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3014712, end_position_on_the_genomic_accession=3018187, orientation='minus', exon_count=2, OMIM=615798, no_hdr0=''),
  9075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9075, CurrentID=0, Status='live', Symbol='CLDN2', Aliases='', description='claudin 2', other_designations='SP82', map_location='Xq22.3-q23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106900164, end_position_on_the_genomic_accession=106930861, orientation='plus', exon_count=4, OMIM=300520, no_hdr0=''),
  9076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9076, CurrentID=0, Status='live', Symbol='CLDN1', Aliases='CLD1, ILVASC, SEMP1', description='claudin 1', other_designations='senescence-associated epithelial membrane protein 1', map_location='3q28-q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190305701, end_position_on_the_genomic_accession=190322446, orientation='minus', exon_count=4, OMIM=603718, no_hdr0=''),
  9077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9077, CurrentID=0, Status='live', Symbol='DIRAS3', Aliases='ARHI, NOEY2', description='DIRAS family, GTP-binding RAS-like 3', other_designations='distinct subgroup of the Ras family member 3|ras homolog gene family, member I|rho-related GTP-binding protein RhoI', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68045962, end_position_on_the_genomic_accession=68051733, orientation='minus', exon_count=3, OMIM=605193, no_hdr0=''),
  101929309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929309, CurrentID=0, Status='live', Symbol='LOC101929309', Aliases='', description='uncharacterized LOC101929309', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35543393, end_position_on_the_genomic_accession=35593621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9079, CurrentID=0, Status='live', Symbol='LDB2', Aliases='CLIM1, LDB-2, LDB1', description='LIM domain binding 2', other_designations='LIM binding domain 2|LIM domain-binding factor CLIM1|LIM domain-binding factor-2|carboxyl-terminal LIM domain-binding protein 1', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16501534, end_position_on_the_genomic_accession=16898809, orientation='minus', exon_count=15, OMIM=603450, no_hdr0=''),
  9080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9080, CurrentID=0, Status='live', Symbol='CLDN9', Aliases='', description='claudin 9', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3012456, end_position_on_the_genomic_accession=3014505, orientation='plus', exon_count=1, OMIM=615799, no_hdr0=''),
  9081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9081, CurrentID=0, Status='live', Symbol='PRY', Aliases='PRY1, PTPN13LY', description='PTPN13-like, Y-linked', other_designations='testis-specific PTP-BL-related protein on Y', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22484280, end_position_on_the_genomic_accession=22515543, orientation='plus', exon_count=6, OMIM=400019, no_hdr0=''),
  9082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9082, CurrentID=0, Status='live', Symbol='XKRY', Aliases='XKRY1', description='X-linked Kx blood group related, Y-linked', other_designations='Testis-specific XK-related protein on Y|X Kell blood group precursor-related, Y-linked|XK, Kell blood group complex subunit-related, Y-linked', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17768980, end_position_on_the_genomic_accession=17770560, orientation='minus', exon_count=1, OMIM=400015, no_hdr0=''),
  9083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9083, CurrentID=0, Status='live', Symbol='BPY2', Aliases='BPY2A, VCY2, VCY2A', description='basic charge, Y-linked, 2', other_designations='basic protein on Y chromosome 2|testis-specific basic protein on Y, 2|variable charge, Y-linked, 2|variably charged protein Y 2', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22984263, end_position_on_the_genomic_accession=23005465, orientation='plus', exon_count=9, OMIM=400013, no_hdr0=''),
  9084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9084, CurrentID=0, Status='live', Symbol='VCY', Aliases='BPY11, VCY1A, VCY', description='variable charge, Y-linked', other_designations='basic charge, Y-linked 1|basic protein on Y chromosome, 1|testis-specific basic protein on Y, 1|variable charge, Y chromosome|variably charged protein Y', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13985772, end_position_on_the_genomic_accession=13986512, orientation='minus', exon_count=2, OMIM=400012, no_hdr0=''),
  9085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9085, CurrentID=0, Status='live', Symbol='CDY1', Aliases='CDYA, CDY1', description='chromodomain protein, Y-linked, 1', other_designations='chromodomain protein, Y chromosome, 1|testis-specific chromodomain protein on Y', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25622115, end_position_on_the_genomic_accession=25634745, orientation='plus', exon_count=4, OMIM=400016, no_hdr0=''),
  9086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9086, CurrentID=0, Status='live', Symbol='EIF1AY', Aliases='eIF-4C', description='eukaryotic translation initiation factor 1A, Y-linked', other_designations='eIF-1A Y isoform|eukaryotic translation initiation factor 1A, Y chromosome|eukaryotic translation initiation factor 4C', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20575711, end_position_on_the_genomic_accession=20593154, orientation='plus', exon_count=7, OMIM=400014, no_hdr0=''),
  9087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9087, CurrentID=0, Status='live', Symbol='TMSB4Y', Aliases='TB4Y', description='thymosin beta 4, Y-linked', other_designations='thymosin beta-4, Y isoform|thymosin, beta 4, Y chromosome', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13703567, end_position_on_the_genomic_accession=13706024, orientation='plus', exon_count=2, OMIM=400017, no_hdr0=''),
  9088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9088, CurrentID=0, Status='live', Symbol='PKMYT1', Aliases='MYT1, PPP1R126', description='protein kinase, membrane associated tyrosine/threonine 1', other_designations='myt1 kinase|protein phosphatase 1, regulatory subunit 126', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2972791, end_position_on_the_genomic_accession=2980539, orientation='minus', exon_count=10, OMIM=602474, no_hdr0=''),
  100131127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131127, CurrentID=0, Status='live', Symbol='LOC100131127', Aliases='', description='degenerative spermatocyte homolog 1, lipid desaturase (Drosophila) pseudogene', other_designations='delta(4)-desaturase, sphingolipid 1 pseudogene', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28225919, end_position_on_the_genomic_accession=28227559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9091, CurrentID=0, Status='live', Symbol='PIGQ', Aliases='GPI1, c407A10.1', description='phosphatidylinositol glycan anchor biosynthesis class Q', other_designations='N-acetylglucosaminyl transferase component Gpi1|N-acetylglucosamyl transferase component GPI1|PIG-Q|c407A10.1 (GPI1 (N-acetylglucosaminyl transferase component))|phosphatidylinositol glycan anchor biosynthesis, class Q|phosphatidylinositol glycan, class Q|phosphatidylinositol-glycan biosynthesis class Q protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=569968, end_position_on_the_genomic_accession=584136, orientation='plus', exon_count=11, OMIM=605754, no_hdr0=''),
  9092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9092, CurrentID=0, Status='live', Symbol='SART1', Aliases='Ara1, HOMS1259, SNRNP110, Snu66, SART1', description='squamous cell carcinoma antigen recognized by T-cells 1', other_designations='IgE autoantigen|SART-1|SART1(259) protein|SART1(800) protein|SNU66 homolog|U4/U6.U5 tri-snRNP-associated 110 kDa protein|hSART-1|hSnu66|small nuclear ribonucleoprotein 110kDa (U4/U6.U5)|squamous cell carcinoma antigen recognised by T cells|squamous cell carcinoma antigen recognized by T cells 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65961689, end_position_on_the_genomic_accession=65980136, orientation='plus', exon_count=20, OMIM=605941, no_hdr0=''),
  9093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9093, CurrentID=0, Status='live', Symbol='DNAJA3', Aliases='HCA57, TID1, hTID-1', description='DnaJ heat shock protein family (Hsp40) member A3', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 3|dnaJ protein Tid-1|hepatocellular carcinoma-associated antigen 57|tumorous imaginal discs protein Tid56 homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4425805, end_position_on_the_genomic_accession=4456775, orientation='plus', exon_count=12, OMIM=608382, no_hdr0=''),
  9094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9094, CurrentID=0, Status='live', Symbol='UNC119', Aliases='HRG4, IMD13, POC7, POC7A', description='unc-119 lipid binding chaperone', other_designations='POC7 centriolar protein homolog A|retinal protein 4|unc-119 homolog', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28546707, end_position_on_the_genomic_accession=28552628, orientation='minus', exon_count=5, OMIM=604011, no_hdr0=''),
  9095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9095, CurrentID=0, Status='live', Symbol='TBX19', Aliases='TBS19, TPIT, dJ747L4.1', description='T-box 19', other_designations='T-box factor, pituitary|T-box protein 19|TBS 19', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168281040, end_position_on_the_genomic_accession=168314426, orientation='plus', exon_count=8, OMIM=604614, no_hdr0=''),
  9096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9096, CurrentID=0, Status='live', Symbol='TBX18', Aliases='', description='T-box 18', other_designations='T-box protein 18', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84687351, end_position_on_the_genomic_accession=84764730, orientation='minus', exon_count=11, OMIM=604613, no_hdr0=''),
  9097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9097, CurrentID=0, Status='live', Symbol='USP14', Aliases='TGT', description='ubiquitin specific peptidase 14 (tRNA-guanine transglycosylase)', other_designations='deubiquitinating enzyme 14|tRNA-guanine transglycosylase, 60-kD subunit|ubiquitin specific protease 14 (tRNA-guanine transglycosylase)|ubiquitin thioesterase 14|ubiquitin thiolesterase 14|ubiquitin-specific processing protease 14|ubiquitin-specific-processing protease 14', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=158483, end_position_on_the_genomic_accession=213739, orientation='plus', exon_count=16, OMIM=607274, no_hdr0=''),
  9098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9098, CurrentID=0, Status='live', Symbol='USP6', Aliases='HRP1, TRE17, TRE2, TRESMCR, Tre-2-short, USP6', description='ubiquitin specific peptidase 6', other_designations='TBC1D3 and USP32 fusion|deubiquitinating enzyme 6|hyperpolymorphic gene 1|proto-oncogene TRE-2|ubiquitin specific peptidase 6 (Tre-2 oncogene)|ubiquitin specific protease 6 (Tre-2 oncogene)|ubiquitin thioesterase 6|ubiquitin thiolesterase 6|ubiquitin-specific protease USP6|ubiquitin-specific-processing protease 6', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5116438, end_position_on_the_genomic_accession=5175032, orientation='plus', exon_count=42, OMIM=604334, no_hdr0=''),
  9099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9099, CurrentID=0, Status='live', Symbol='USP2', Aliases='UBP41, USP9', description='ubiquitin specific peptidase 2', other_designations='41 kDa ubiquitin-specific protease|deubiquitinating enzyme 2|ubiquitin specific protease 12|ubiquitin specific protease 9|ubiquitin thioesterase 2|ubiquitin-specific-processing protease 2', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119355215, end_position_on_the_genomic_accession=119381726, orientation='minus', exon_count=16, OMIM=604725, no_hdr0=''),
  9100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9100, CurrentID=0, Status='live', Symbol='USP10', Aliases='UBPO', description='ubiquitin specific peptidase 10', other_designations='deubiquitinating enzyme 10|ubiquitin specific protease 10|ubiquitin thioesterase 10|ubiquitin thiolesterase 10|ubiquitin-specific-processing protease 10', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84699949, end_position_on_the_genomic_accession=84779922, orientation='plus', exon_count=18, OMIM=609818, no_hdr0=''),
  9101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9101, CurrentID=0, Status='live', Symbol='USP8', Aliases='HumORF8, SPG59, UBPY', description='ubiquitin specific peptidase 8', other_designations='deubiquitinating enzyme 8|ubiquitin isopeptidase Y|ubiquitin thiolesterase 8|ubiquitin-specific-processing protease 8', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50424377, end_position_on_the_genomic_accession=50501083, orientation='plus', exon_count=21, OMIM=603158, no_hdr0=''),
  9103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9103, CurrentID=0, Status='live', Symbol='FCGR2C', Aliases='CD32, CD32C, CDW32, FCG2, FCRIIC, IGFR2', description='Fc fragment of IgG, low affinity IIc, receptor for (CD32) (gene/pseudogene)', other_designations='Fc gamma receptor IIC|IgG Fc receptor II-c|fc-gamma-RIIc|immunoglobulin G Fc receptor II-c', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161581339, end_position_on_the_genomic_accession=161601220, orientation='plus', exon_count=7, OMIM=612169, no_hdr0=''),
  9104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9104, CurrentID=0, Status='live', Symbol='RGN', Aliases='GNL, HEL-S-41, RC, SMP30', description='regucalcin', other_designations='epididymis secretory protein Li 41|gluconolactonase|senescence marker protein 30', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47078355, end_position_on_the_genomic_accession=47093314, orientation='plus', exon_count=8, OMIM=300212, no_hdr0=''),
  303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=303, CurrentID=0, Status='live', Symbol='ANXA2P1', Aliases='ANX2L1, ANX2P1, LPC2A', description='annexin A2 pseudogene 1', other_designations='annexin II (lipocortin II) pseudogene 1', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153307469, end_position_on_the_genomic_accession=153308811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9107, CurrentID=0, Status='live', Symbol='MTMR6', Aliases='', description='myotubularin related protein 6', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25246201, end_position_on_the_genomic_accession=25287566, orientation='minus', exon_count=16, OMIM=603561, no_hdr0=''),
  9108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9108, CurrentID=0, Status='live', Symbol='MTMR7', Aliases='', description='myotubularin related protein 7', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17296797, end_position_on_the_genomic_accession=17413531, orientation='minus', exon_count=18, OMIM=603562, no_hdr0=''),
  9110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9110, CurrentID=0, Status='live', Symbol='MTMR4', Aliases='FYVE-DSP2, ZFYVE11', description='myotubularin related protein 4', other_designations='FYVE domain-containing dual specificity protein phosphatase 2|zinc finger FYVE domain-containing protein 11|zinc finger, FYVE domain containing 11', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58489532, end_position_on_the_genomic_accession=58518110, orientation='minus', exon_count=23, OMIM=603559, no_hdr0=''),
  9111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9111, CurrentID=0, Status='live', Symbol='NMI', Aliases='', description='N-myc and STAT interactor', other_designations='N-myc (and STAT) interactor|N-myc interactor', map_location='2q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151270465, end_position_on_the_genomic_accession=151289916, orientation='minus', exon_count=9, OMIM=603525, no_hdr0=''),
  9112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9112, CurrentID=0, Status='live', Symbol='MTA1', Aliases='', description='metastasis associated 1', other_designations='metastasis associated gene 1 protein', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105419849, end_position_on_the_genomic_accession=105470729, orientation='plus', exon_count=24, OMIM=603526, no_hdr0=''),
  9113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9113, CurrentID=0, Status='live', Symbol='LATS1', Aliases='WARTS, wts', description='large tumor suppressor kinase 1', other_designations='LATS (large tumor suppressor, Drosophila) homolog 1|LATS, large tumor suppressor, homolog 1|WARTS protein kinase|h-warts|large tumor suppressor homolog 1', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149658153, end_position_on_the_genomic_accession=149718256, orientation='minus', exon_count=13, OMIM=603473, no_hdr0=''),
  9114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9114, CurrentID=0, Status='live', Symbol='ATP6V0D1', Aliases='ATP6D, ATP6DV, P39, VATX, VMA6, VPATPD', description='ATPase, H+ transporting, lysosomal 38kDa, V0 subunit d1', other_designations='32 kDa accessory protein|ATPase, H+ transporting, lysosomal (vacuolar proton pump), member D|H(+)-transporting two-sector ATPase, subunit D|V-ATPase 40 KDa accessory protein|V-ATPase AC39 subunit|V-ATPase subunit d 1|V-ATPase, subunit D|vacuolar proton pump subunit d 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67438014, end_position_on_the_genomic_accession=67481186, orientation='minus', exon_count=11, OMIM=607028, no_hdr0=''),
  9117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9117, CurrentID=0, Status='live', Symbol='SEC22C', Aliases='SEC22L3', description='SEC22 homolog C, vesicle trafficking protein', other_designations='SEC22 vesicle trafficking protein homolog C|SEC22 vesicle trafficking protein-like 3|SEC22 vesicle trafficking protein-like protein C|secretion deficient 22C', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42547967, end_position_on_the_genomic_accession=42601080, orientation='minus', exon_count=12, OMIM=604028, no_hdr0=''),
  9118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9118, CurrentID=0, Status='live', Symbol='INA', Aliases='NEF5, NF-66, TXBP-1', description='internexin neuronal intermediate filament protein alpha', other_designations='66 kDa neurofilament protein|alpha-Inx|internexin neuronal intermediate filament protein, alpha|neurofilament 5 (66kD)|neurofilament-66, tax-binding protein', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103277163, end_position_on_the_genomic_accession=103290351, orientation='plus', exon_count=3, OMIM=605338, no_hdr0=''),
  9119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9119, CurrentID=0, Status='live', Symbol='KRT75', Aliases='K6HF, KB18, PFB', description='keratin 75, type II', other_designations='CK-75|K75|cytokeratin type II|cytokeratin-75|hK6hf|keratin-6 hair follicle|keratin-75|type II keratin-18|type II keratin-K6hf|type-II keratin Kb18', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52424070, end_position_on_the_genomic_accession=52434326, orientation='minus', exon_count=9, OMIM=609025, no_hdr0=''),
  9120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9120, CurrentID=0, Status='live', Symbol='SLC16A6', Aliases='MCT6, MCT7', description='solute carrier family 16 member 6', other_designations='breast cancer protein BCY5|monocarboxylate transporter 6|solute carrier family 16, member 6 (monocarboxylic acid transporter 7)', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68267026, end_position_on_the_genomic_accession=68291264, orientation='minus', exon_count=9, OMIM=603880, no_hdr0=''),
  9121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9121, CurrentID=0, Status='live', Symbol='SLC16A5', Aliases='MCT5, MCT6', description='solute carrier family 16 (monocarboxylate transporter), member 5', other_designations='monocarboxylate transporter 5|solute carrier family 16 (monocarboxylic acid transporters), member 5|solute carrier family 16, member 5 (monocarboxylic acid transporter 6)', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75087727, end_position_on_the_genomic_accession=75106162, orientation='plus', exon_count=9, OMIM=603879, no_hdr0=''),
  9122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9122, CurrentID=0, Status='live', Symbol='SLC16A4', Aliases='MCT4, MCT5', description='solute carrier family 16 member 4', other_designations='MCT 4|MCT 5|monocarboxylate transporter 4|solute carrier family 16 (monocarboxylic acid transporters), member 4|solute carrier family 16, member 4 (monocarboxylic acid transporter 5)', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110362851, end_position_on_the_genomic_accession=110391409, orientation='minus', exon_count=10, OMIM=603878, no_hdr0=''),
  9123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9123, CurrentID=0, Status='live', Symbol='SLC16A3', Aliases='MCT 3, MCT 4, MCT-3, MCT-4, MCT3, MCT4', description='solute carrier family 16 (monocarboxylate transporter), member 3', other_designations='monocarboxylate transporter 3|solute carrier family 16 (monocarboxylic acid transporters), member 3|solute carrier family 16, member 3 (monocarboxylic acid transporter 4)', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82228406, end_position_on_the_genomic_accession=82239499, orientation='plus', exon_count=9, OMIM=603877, no_hdr0=''),
  9124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9124, CurrentID=0, Status='live', Symbol='PDLIM1', Aliases='CLIM1, CLP-36, CLP36, HEL-S-112, hCLIM1', description='PDZ and LIM domain 1', other_designations='LIM domain protein CLP-36|carboxyl terminal LIM domain protein 1|elfin|epididymis secretory protein Li 112', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95237568, end_position_on_the_genomic_accession=95291148, orientation='minus', exon_count=7, OMIM=605900, no_hdr0=''),
  9125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9125, CurrentID=0, Status='live', Symbol='RQCD1', Aliases='CAF40, CNOT9, CT129, RCD-1, RCD1', description='RCD1 required for cell differentiation1 homolog (S. pombe)', other_designations='CCR4-NOT transcription complex subunit 9|cancer/testis antigen 129|protein involved in sexual development', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218568575, end_position_on_the_genomic_accession=218596435, orientation='plus', exon_count=11, OMIM=612054, no_hdr0=''),
  9126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9126, CurrentID=0, Status='live', Symbol='SMC3', Aliases='BAM, BMH, CDLS3, CSPG6, HCAPL1, SMC3', description='structural maintenance of chromosomes 3', other_designations='SMC protein 3|basement membrane-associated chondroitin proteoglycan|chondroitin sulfate proteoglycan 6 (bamacan)|chromosome-associated polypeptide', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110567691, end_position_on_the_genomic_accession=110604634, orientation='plus', exon_count=29, OMIM=606062, no_hdr0=''),
  9127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9127, CurrentID=0, Status='live', Symbol='P2RX6', Aliases='P2RXL1, P2X6, P2XM', description='purinergic receptor P2X, ligand gated ion channel, 6', other_designations='ATP receptor|purinergic receptor P2X, ligand-gated ion channel, 6|purinergic receptor P2X-like 1, orphan receptor|purinoceptor P2X6|purinoreceptor P2X6|skeletal muscle-expressed P2X', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21014451, end_position_on_the_genomic_accession=21028013, orientation='plus', exon_count=12, OMIM=608077, no_hdr0=''),
  9128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9128, CurrentID=0, Status='live', Symbol='PRPF4', Aliases='HPRP4, HPRP4P, PRP4, Prp4p, RP70, SNRNP60', description='pre-mRNA processing factor 4', other_designations='PRP4 homolog|PRP4 pre-mRNA processing factor 4 homolog|PRP4/STK/WD splicing factor|U4/U6 snRNP 60 kDa protein|WD splicing factor Prp4', map_location='9q31-q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113275634, end_position_on_the_genomic_accession=113293914, orientation='plus', exon_count=14, OMIM=607795, no_hdr0=''),
  9129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9129, CurrentID=0, Status='live', Symbol='PRPF3', Aliases='HPRP3, HPRP3P, PRP3, Prp3p, RP18, SNRNP90', description='pre-mRNA processing factor 3', other_designations='PRP3 pre-mRNA processing factor 3 homolog|U4/U6 snRNP 90 kDa protein|U4/U6-associated RNA splicing factor|pre-mRNA-splicing factor 3', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150321479, end_position_on_the_genomic_accession=150353228, orientation='plus', exon_count=19, OMIM=607301, no_hdr0=''),
  9130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9130, CurrentID=0, Status='live', Symbol='FAM50A', Aliases='9F, DXS9928E, HXC-26, HXC26, XAP5', description='family with sequence similarity 50 member A', other_designations='family with sequence similarity 50, member A|protein XAP-5', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154444126, end_position_on_the_genomic_accession=154450654, orientation='plus', exon_count=13, OMIM=300453, no_hdr0=''),
  9131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9131, CurrentID=0, Status='live', Symbol='AIFM1', Aliases='AIF, CMT2D, CMTX4, COWCK, COXPD6, NADMR, NAMSD, PDCD8', description='apoptosis inducing factor, mitochondria associated 1', other_designations='apoptosis-inducing factor, mitochondrion-associated, 1|programmed cell death 8 (apoptosis-inducing factor)|striatal apoptosis-inducing factor', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130129362, end_position_on_the_genomic_accession=130165887, orientation='minus', exon_count=18, OMIM='300169,310490', no_hdr0=''),
  9132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9132, CurrentID=0, Status='live', Symbol='KCNQ4', Aliases='DFNA2, DFNA2A, KV7.4', description='potassium channel, voltage gated KQT-like subfamily Q, member 4', other_designations='potassium channel KQT-like 4|potassium channel subunit alpha KvLQT4|potassium voltage-gated channel, KQT-like subfamily, member 4', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40784012, end_position_on_the_genomic_accession=40840452, orientation='plus', exon_count=18, OMIM=603537, no_hdr0=''),
  9133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9133, CurrentID=0, Status='live', Symbol='CCNB2', Aliases='HsT17299', description='cyclin B2', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59105085, end_position_on_the_genomic_accession=59125045, orientation='plus', exon_count=9, OMIM=602755, no_hdr0=''),
  9134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9134, CurrentID=0, Status='live', Symbol='CCNE2', Aliases='CYCE2', description='cyclin E2', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94880224, end_position_on_the_genomic_accession=94896653, orientation='minus', exon_count=14, OMIM=603775, no_hdr0=''),
  9135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9135, CurrentID=0, Status='live', Symbol='RABEP1', Aliases='RAB5EP, RABPT5', description='rabaptin, RAB GTPase binding effector protein 1', other_designations='neurocrescin|rabaptin-4|rabaptin-5|rabaptin-5alpha|renal carcinoma antigen NY-REN-17', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5282263, end_position_on_the_genomic_accession=5386340, orientation='plus', exon_count=20, OMIM=603616, no_hdr0=''),
  9136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9136, CurrentID=0, Status='live', Symbol='RRP9', Aliases='RNU3IP2, U3-55K', description='ribosomal RNA processing 9, small subunit (SSU) processome component, homolog (yeast)', other_designations='RNA, U3 small nucleolar interacting protein 2|RRP9 homolog|RRP9, small subunit (SSU) processome component, homolog|U3 small nucleolar ribonucleoprotein-associated 55 kDa protein|U3 snoRNP-associated 55-kDa protein', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51933426, end_position_on_the_genomic_accession=51941941, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23368, CurrentID=0, Status='live', Symbol='PPP1R13B', Aliases='ASPP1, p53BP2-like, p85', description='protein phosphatase 1 regulatory subunit 13B', other_designations='apoptosis-stimulating protein of p53, 1|protein phosphatase 1, regulatory (inhibitor) subunit 13B|protein phosphatase 1, regulatory subunit 13B', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103733751, end_position_on_the_genomic_accession=103848967, orientation='minus', exon_count=23, OMIM=606455, no_hdr0=''),
  9138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9138, CurrentID=0, Status='live', Symbol='ARHGEF1', Aliases='GEF1, LBCL2, LSC, P115-RHOGEF, SUB1.5', description='Rho guanine nucleotide exchange factor 1', other_designations='115 kDa guanine nucleotide exchange factor|115-kD protein|Lsc homolog|Rho guanine nucleotide exchange factor (GEF) 1|p115RhoGEF', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41883173, end_position_on_the_genomic_accession=41907454, orientation='plus', exon_count=32, OMIM=601855, no_hdr0=''),
  9139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9139, CurrentID=0, Status='live', Symbol='CBFA2T2', Aliases='EHT, MTGR1, ZMYND3, p85', description='core-binding factor, runt domain, alpha subunit 2; translocated to, 2', other_designations='ETO homolog on chromosome 20|ETO homologous on chromosome 20|MTG8-like protein|MTG8-related protein 1|myeloid translocation gene-related protein 1|myeloid translocation-related protein 1', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33490122, end_position_on_the_genomic_accession=33650031, orientation='plus', exon_count=19, OMIM=603672, no_hdr0=''),
  9140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9140, CurrentID=0, Status='live', Symbol='ATG12', Aliases='APG12, APG12L, FBR93, HAPG12', description='autophagy related 12', other_designations='ATG12 autophagy related 12 homolog|Apg12 (autophagy, yeast) homolog', map_location='5q21-q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115828196, end_position_on_the_genomic_accession=115841851, orientation='minus', exon_count=6, OMIM=609608, no_hdr0=''),
  9141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9141, CurrentID=0, Status='live', Symbol='PDCD5', Aliases='TFAR19', description='programmed cell death 5', other_designations='TF-1 cell apoptosis-related protein 19|TF1 cell apoptosis-related gene 19|TFAR19 novel apoptosis-related', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32581170, end_position_on_the_genomic_accession=32587452, orientation='plus', exon_count=6, OMIM=604583, no_hdr0=''),
  9142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9142, CurrentID=0, Status='live', Symbol='TMEM257', Aliases='CXorf1', description='transmembrane protein 257', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145827410, end_position_on_the_genomic_accession=145829856, orientation='plus', exon_count=1, OMIM=300565, no_hdr0=''),
  9143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9143, CurrentID=0, Status='live', Symbol='SYNGR3', Aliases='', description='synaptogyrin 3', other_designations='', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1989945, end_position_on_the_genomic_accession=1994275, orientation='plus', exon_count=4, OMIM=603927, no_hdr0=''),
  9144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9144, CurrentID=0, Status='live', Symbol='SYNGR2', Aliases='', description='synaptogyrin 2', other_designations='cellugyrin', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78168551, end_position_on_the_genomic_accession=78172928, orientation='plus', exon_count=4, OMIM=603926, no_hdr0=''),
  9145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9145, CurrentID=0, Status='live', Symbol='SYNGR1', Aliases='', description='synaptogyrin 1', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39349949, end_position_on_the_genomic_accession=39385588, orientation='plus', exon_count=6, OMIM=603925, no_hdr0=''),
  9146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9146, CurrentID=0, Status='live', Symbol='HGS', Aliases='HRS', description='hepatocyte growth factor-regulated tyrosine kinase substrate', other_designations='human growth factor-regulated tyrosine kinase substrate|protein pp110', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81683932, end_position_on_the_genomic_accession=81702121, orientation='plus', exon_count=24, OMIM=604375, no_hdr0=''),
  9147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9147, CurrentID=0, Status='live', Symbol='NEMF', Aliases='NY-CO-1, SDCCAG1', description='nuclear export mediator factor', other_designations='antigen NY-CO-1|serologically defined colon cancer antigen 1', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49783278, end_position_on_the_genomic_accession=49853073, orientation='minus', exon_count=37, OMIM=608378, no_hdr0=''),
  9148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9148, CurrentID=0, Status='live', Symbol='NEURL1', Aliases='NEUR1, NEURL, RNF67, bA416N2.1, neu, neu-1', description='neuralized E3 ubiquitin protein ligase 1', other_designations='RING finger protein 67|h-neuralized 1|neuralized homolog|neuralized-like protein 1A', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103493890, end_position_on_the_genomic_accession=103592552, orientation='plus', exon_count=12, OMIM=603804, no_hdr0=''),
  9149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9149, CurrentID=0, Status='live', Symbol='DYRK1B', Aliases='AOMS3, MIRK', description='dual specificity tyrosine-(Y)-phosphorylation regulated kinase 1B', other_designations='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 1B|minibrain-related kinase|mirk protein kinase', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39825347, end_position_on_the_genomic_accession=39834233, orientation='minus', exon_count=14, OMIM=604556, no_hdr0=''),
  9150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9150, CurrentID=0, Status='live', Symbol='CTDP1', Aliases='CCFDN, FCP1', description='CTD phosphatase subunit 1', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) phosphatase, subunit 1|CTD of POLR2A, phosphatase of, subunit 1|TFIIF-associating CTD phosphatase 1|serine phosphatase FCP1a|transcription factor IIF-associating CTD phosphatase 1', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79679801, end_position_on_the_genomic_accession=79754510, orientation='plus', exon_count=14, OMIM=604927, no_hdr0=''),
  100271605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271605, CurrentID=0, Status='live', Symbol='RPL27AP8', Aliases='RPL27A_5_1310', description='ribosomal protein L27a pseudogene 8', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47766640, end_position_on_the_genomic_accession=47767096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9152, CurrentID=0, Status='live', Symbol='SLC6A5', Aliases='GLYT-2, GLYT2, HKPX3, NET1', description='solute carrier family 6 (neurotransmitter transporter), member 5', other_designations='glycine transporter 2|solute carrier family 6 (neurotransmitter transporter, glycine), member 5', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20599400, end_position_on_the_genomic_accession=20659285, orientation='plus', exon_count=18, OMIM=604159, no_hdr0=''),
  9153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9153, CurrentID=0, Status='live', Symbol='SLC28A2', Aliases='CNT2, HCNT2, HsT17153, SPNT1', description='solute carrier family 28 (concentrative nucleoside transporter), member 2', other_designations='CNT 2|Na(+)/nucleoside cotransporter 2|SPNT|concentrative nucleoside transporter 2|sodium-coupled nucleoside transporter 2|sodium/purine nucleoside co-transporter|solute carrier family 28 (sodium-coupled nucleoside transporter), member 2|solute carrier family 28 member 2', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45252066, end_position_on_the_genomic_accession=45275934, orientation='plus', exon_count=18, OMIM=606208, no_hdr0=''),
  9154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9154, CurrentID=0, Status='live', Symbol='SLC28A1', Aliases='CNT1, HCNT1', description='solute carrier family 28 (concentrative nucleoside transporter), member 1', other_designations='Na(+)/nucleoside cotransporter 1|concentrative nucleoside transporter 1|sodium-coupled nucleoside transporter 1|solute carrier family 28 (sodium-coupled nucleoside transporter), member 1|solute carrier family 28 member 1', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84884577, end_position_on_the_genomic_accession=84975651, orientation='plus', exon_count=29, OMIM=606207, no_hdr0=''),
  9156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9156, CurrentID=0, Status='live', Symbol='EXO1', Aliases='HEX1, hExoI', description='exonuclease 1', other_designations='hExo1|rad2 nuclease family member, homolog of S. cerevisiae exonuclease 1', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241847967, end_position_on_the_genomic_accession=241909165, orientation='plus', exon_count=22, OMIM=606063, no_hdr0=''),
  100271606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271606, CurrentID=0, Status='live', Symbol='RPL37AP9', Aliases='RPL37A_4_1209', description='ribosomal protein L37a pseudogene 9', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12840804, end_position_on_the_genomic_accession=12841079, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9158, CurrentID=0, Status='live', Symbol='FIBP', Aliases='FGFIBP-1, FIBP', description='fibroblast growth factor (acidic) intracellular binding protein', other_designations='FGF-1 intracellular-binding protein|aFGF intracellular-binding protein', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65883740, end_position_on_the_genomic_accession=65888539, orientation='minus', exon_count=10, OMIM=608296, no_hdr0=''),
  9159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9159, CurrentID=0, Status='live', Symbol='PCSK7', Aliases='LPC, PC7, PC8, SPC7', description='proprotein convertase subtilisin/kexin type 7', other_designations='lymphoma proprotein convertase|prohormone convertase 7|prohormone convertase PC7|proprotein convertase 7|proprotein convertase 8|proprotein convertase PC7|proprotein convertase subtilisin/kexin type 7 precursor variant 2|subtilisin/kexin-like protease PC7', map_location='11q23-q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117205071, end_position_on_the_genomic_accession=117232525, orientation='minus', exon_count=20, OMIM=604872, no_hdr0=''),
  100270794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270794, CurrentID=0, Status='live', Symbol='STQTL10', Aliases='', description='Stature QTL 10', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612221, no_hdr0=''),
  9162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9162, CurrentID=0, Status='live', Symbol='DGKI', Aliases='DGK-IOTA', description='diacylglycerol kinase, iota', other_designations='DAG kinase iota|diglyceride kinase iota', map_location='7q32.3-q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137387654, end_position_on_the_genomic_accession=137846863, orientation='minus', exon_count=41, OMIM=604072, no_hdr0=''),
  9163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9163, CurrentID=0, Status='live', Symbol='AMLCR2', Aliases='AMLCR', description='acute myeloid leukemia chromosome region 2', other_designations='acute myelogenous leukemia - linkage', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602439, no_hdr0=''),
  402381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402381, CurrentID=0, Status='live', Symbol='SOHLH1', Aliases='C9orf157, NOHLH, SPATA27, TEB2, bA100C15.3, bHLHe80', description='spermatogenesis and oogenesis specific basic helix-loop-helix 1', other_designations='newborn ovary helix loop helix|spermatogenesis associated 27', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135693407, end_position_on_the_genomic_accession=135701316, orientation='minus', exon_count=11, OMIM=610224, no_hdr0=''),
  9166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9166, CurrentID=0, Status='live', Symbol='EBAG9', Aliases='EB9, PDAF', description='estrogen receptor binding site associated, antigen, 9', other_designations='cancer-associated surface antigen RCAS1|estrogen receptor-binding fragment-associated gene 9 protein', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109539700, end_position_on_the_genomic_accession=109565996, orientation='plus', exon_count=9, OMIM=605772, no_hdr0=''),
  9167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9167, CurrentID=0, Status='live', Symbol='COX7A2L', Aliases='COX7AR, COX7RP, EB1, SIG81', description='cytochrome c oxidase subunit VIIa polypeptide 2 like', other_designations='COX7a-related protein|cytochrome c oxidase subunit VII-related protein|cytochrome c oxidase subunit VIIa-related protein|estrogen receptor binding CpG island', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42350501, end_position_on_the_genomic_accession=42369010, orientation='minus', exon_count=4, OMIM=605771, no_hdr0=''),
  9168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9168, CurrentID=0, Status='live', Symbol='TMSB10', Aliases='MIG12, TB10', description='thymosin beta 10', other_designations='migration-inducing gene 12|migration-inducing protein 12', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84905639, end_position_on_the_genomic_accession=84906675, orientation='plus', exon_count=3, OMIM=188399, no_hdr0=''),
  9169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9169, CurrentID=0, Status='live', Symbol='SCAF11', Aliases='CASP11, SFRS2IP, SIP1, SRRP129, SRSF2IP', description='SR-related CTD-associated factor 11', other_designations='CTD-associated SR protein 11|SC35-interacting protein 1|SFRS2-interacting protein|SR-related and CTD-associated factor 11|SRSF2-interacting protein|renal carcinoma antigen NY-REN-40|serine/arginine-rich splicing factor 2, interacting protein|serine/arginine-rich splicing factor 2-interacting protein|splicing factor, arginine/serine-rich 2, interacting protein|splicing factor, arginine/serine-rich 2-interacting protein|splicing regulatory protein 129', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45919131, end_position_on_the_genomic_accession=45992040, orientation='minus', exon_count=20, OMIM=603668, no_hdr0=''),
  9170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9170, CurrentID=0, Status='live', Symbol='LPAR2', Aliases='EDG-4, EDG4, LPA-2, LPA2', description='lysophosphatidic acid receptor 2', other_designations='G protein-coupled receptor|LPA receptor 2|LPA receptor EDG4|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor, 4|lysophosphatidic acid receptor EDG4|lysophosphatidic acid receptor Edg-4', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19623655, end_position_on_the_genomic_accession=19628930, orientation='minus', exon_count=6, OMIM=605110, no_hdr0=''),
  9172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9172, CurrentID=0, Status='live', Symbol='MYOM2', Aliases='TTNAP', description='myomesin 2', other_designations='165 kDa connectin-associated protein|165 kDa titin-associated protein|M-band protein|M-protein|myomesin (M-protein) 2, 165kDa|myomesin family member 2|titin-associated protein, 165 kD', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2045043, end_position_on_the_genomic_accession=2145456, orientation='plus', exon_count=37, OMIM=603509, no_hdr0=''),
  9173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9173, CurrentID=0, Status='live', Symbol='IL1RL1', Aliases='DER4, FIT-1, IL33R, ST2, ST2L, ST2V, T1', description='interleukin 1 receptor-like 1', other_designations='growth stimulation-expressed|homolog of mouse growth stimulation-expressed|interleukin 1 receptor-related protein', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102311502, end_position_on_the_genomic_accession=102352366, orientation='plus', exon_count=13, OMIM=601203, no_hdr0=''),
  9175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9175, CurrentID=0, Status='live', Symbol='MAP3K13', Aliases='LZK, MEKK13, MLK', description='mitogen-activated protein kinase kinase kinase 13', other_designations='leucine zipper-bearing kinase|mixed lineage kinase', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185282941, end_position_on_the_genomic_accession=185489094, orientation='plus', exon_count=18, OMIM=604915, no_hdr0=''),
  9177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9177, CurrentID=0, Status='live', Symbol='HTR3B', Aliases='5-HT3B', description='5-hydroxytryptamine (serotonin) receptor 3B, ionotropic', other_designations='5-HT3-B|5-hydroxytryptamine 3 receptor B subunit|serotonin-gated ion channel subunit', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113898923, end_position_on_the_genomic_accession=113946561, orientation='plus', exon_count=12, OMIM=604654, no_hdr0=''),
  402394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402394, CurrentID=0, Status='live', Symbol='IMPDH1P2', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 2', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40358457, end_position_on_the_genomic_accession=40360809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9179, CurrentID=0, Status='live', Symbol='AP4M1', Aliases='CPSQ3, MU-4, MU-ARP2, SPG50', description='adaptor related protein complex 4, mu 1 subunit', other_designations='AP-4 adaptor complex mu subunit|adapter-related protein complex 4 mu-1 subunit|adapter-related protein complex 4 subunit mu-1|adaptor-related protein complex 4 subunit mu-1|adaptor-related protein complex 4, mu 1 subunit|adaptor-related protein complex AP-4 mu4 subunit|mu subunit of AP-4|mu-adaptin-related protein 2|mu-adaptin-related protein-2|mu4|mu4-adaptin', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100101412, end_position_on_the_genomic_accession=100107180, orientation='plus', exon_count=17, OMIM=602296, no_hdr0=''),
  9180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9180, CurrentID=0, Status='live', Symbol='OSMR', Aliases='OSMRB, PLCA1', description='oncostatin M receptor', other_designations='IL-31 receptor subunit beta|IL-31R subunit beta|IL-31R-beta|IL-31RB|interleukin-31 receptor subunit beta|oncostatin-M specific receptor beta subunit', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38845858, end_position_on_the_genomic_accession=38945579, orientation='plus', exon_count=21, OMIM=601743, no_hdr0=''),
  9181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9181, CurrentID=0, Status='live', Symbol='ARHGEF2', Aliases='GEF, GEF-H1, GEFH1, LFP40, P40', description='Rho/Rac guanine nucleotide exchange factor 2', other_designations='Rho/Rac guanine nucleotide exchange factor (GEF) 2|guanine nucleotide exchange factor H1|microtubule-regulated Rho-GEF|proliferating cell nucleolar antigen p40', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155946640, end_position_on_the_genomic_accession=156002635, orientation='minus', exon_count=36, OMIM=607560, no_hdr0=''),
  9182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9182, CurrentID=0, Status='live', Symbol='RASSF9', Aliases='P-CIP1, PAMCI, PCIP1', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 9', other_designations='PAM COOH-terminal interactor protein 1|peptidylglycine alpha-amidating monooxygenase COOH-terminal interactor protein-1', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85800703, end_position_on_the_genomic_accession=85836540, orientation='minus', exon_count=3, OMIM=610383, no_hdr0=''),
  9183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9183, CurrentID=0, Status='live', Symbol='ZW10', Aliases='HZW10, KNTC1AP', description='zw10 kinetochore protein', other_designations='ZW10 homolog, centromere/kinetochore protein|ZW10, kinetochore associated, homolog|zeste white 10 homolog', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113733183, end_position_on_the_genomic_accession=113773763, orientation='minus', exon_count=16, OMIM=603954, no_hdr0=''),
  9184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9184, CurrentID=0, Status='live', Symbol='BUB3', Aliases='BUB3L, hBUB3', description='BUB3 mitotic checkpoint protein', other_designations='BUB3 budding uninhibited by benzimidazoles 3 homolog|budding uninhibited by benomyl|budding uninhibited by benzimidazoles 3 homolog|mitotic checkpoint component', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123154244, end_position_on_the_genomic_accession=123165370, orientation='plus', exon_count=8, OMIM=603719, no_hdr0=''),
  9185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9185, CurrentID=0, Status='live', Symbol='REPS2', Aliases='POB1', description='RALBP1 associated Eps domain containing 2', other_designations='RALBP1-interacting protein 2|partner of Ral-binding protein 1|partner of RalBP1', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16946691, end_position_on_the_genomic_accession=17220732, orientation='plus', exon_count=22, OMIM=300317, no_hdr0=''),
  140258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140258, CurrentID=0, Status='live', Symbol='KRTAP13-1', Aliases='KAP13.1, KRTAP13.1', description='keratin associated protein 13-1', other_designations='high sulfur keratin associated protein 13.1|high sulfur keratin-associated protein 13.1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30396074, end_position_on_the_genomic_accession=30396822, orientation='plus', exon_count=1, OMIM=608718, no_hdr0=''),
  9187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9187, CurrentID=0, Status='live', Symbol='SLC24A1', Aliases='CSNB1D, HsT17412, NCKX, NCKX1, RODX', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 1', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 1|retinal rod Na+/Ca+/K+ exchanger|solute carrier family 24 member 1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65611405, end_position_on_the_genomic_accession=65661242, orientation='plus', exon_count=16, OMIM=603617, no_hdr0=''),
  9188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9188, CurrentID=0, Status='live', Symbol='DDX21', Aliases='GUA, GURDB, RH-II/GU, RH-II/GuA', description='DEAD (Asp-Glu-Ala-Asp) box helicase 21', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 21|DEAD box protein 21|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 21|Gu protein|RH II/Gu|RNA helicase II/Gu alpha|gu-alpha|nucleolar RNA helicase Gu|nucleolar RNA helicase II', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68956123, end_position_on_the_genomic_accession=68985069, orientation='plus', exon_count=16, OMIM=606357, no_hdr0=''),
  9189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9189, CurrentID=0, Status='live', Symbol='ZBED1', Aliases='ALTE, DREF, TRAMP, hDREF', description='zinc finger, BED-type containing 1', other_designations='Ac-like transposable element|BED-type zinc finger domain-containing protein 1|DNA replication-related element binding factor|dREF homolog', map_location='Xp22.33;Yp11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2486414, end_position_on_the_genomic_accession=2500967, orientation='minus', exon_count=3, OMIM=300178, no_hdr0=''),
  100270795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270795, CurrentID=0, Status='live', Symbol='STQTL11', Aliases='', description='Stature QTL 11', other_designations='', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612223, no_hdr0=''),
  9191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9191, CurrentID=0, Status='live', Symbol='DEDD', Aliases='CASP8IP11, DEFT, FLDED1, KE05, DEDD', description='death effector domain containing', other_designations='DEDPro1|FLDED-1|death effector domain-containing testicular molecule', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161120977, end_position_on_the_genomic_accession=161132783, orientation='minus', exon_count=8, OMIM=606841, no_hdr0=''),
  402408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402408, CurrentID=0, Status='live', Symbol='PKMP2', Aliases='', description='pyruvate kinase, muscle pseudogene 2', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66497711, end_position_on_the_genomic_accession=66499620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9193, CurrentID=0, Status='live', Symbol='SPG9', Aliases='', description='spastic paraplegia 9 (autosomal dominant)', other_designations='Spastic paraplegia-9 (spastic paraparesis with amyotrophy, cataracts and gastroesophageal reflux)', map_location='10q23.3-q24.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601162, no_hdr0=''),
  9194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9194, CurrentID=0, Status='live', Symbol='SLC16A7', Aliases='MCT2', description='solute carrier family 16 (monocarboxylate transporter), member 7', other_designations='solute carrier family 16, member 7 (monocarboxylic acid transporter 2)', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=59595934, end_position_on_the_genomic_accession=59789855, orientation='plus', exon_count=14, OMIM=603654, no_hdr0=''),
  9196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9196, CurrentID=0, Status='live', Symbol='KCNAB3', Aliases='AKR6A9, KCNA3.1B, KCNA3B, KV-BETA-3', description='potassium channel, voltage gated subfamily A regulatory beta subunit 3', other_designations='K(+) channel subunit beta-3|potassium channel, voltage-dependent, beta-3 subunit|potassium voltage-gated channel, shaker-related subfamily, beta member 3', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7921859, end_position_on_the_genomic_accession=7929435, orientation='minus', exon_count=14, OMIM=604111, no_hdr0=''),
  9197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9197, CurrentID=0, Status='live', Symbol='SLC33A1', Aliases='ACATN, AT-1, AT1, CCHLND, SPG42', description='solute carrier family 33 (acetyl-CoA transporter), member 1', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155826511, end_position_on_the_genomic_accession=155854459, orientation='minus', exon_count=7, OMIM=603690, no_hdr0=''),
  402415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402415, CurrentID=0, Status='live', Symbol='XKRX', Aliases='XKR2, XPLAC, XRG2', description='X-linked Kx blood group related, X-linked', other_designations='X Kell blood group precursor-related, X-linked|XK, Kell blood group complex subunit-related, X-linked|membrane protein XPLAC', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100886924, end_position_on_the_genomic_accession=100957401, orientation='minus', exon_count=6, OMIM=300684, no_hdr0=''),
  9200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9200, CurrentID=0, Status='live', Symbol='HACD1', Aliases='CAP, PTPLA', description='3-hydroxyacyl-CoA dehydratase 1', other_designations='cementum attachment protein|cementum-attachment protein|protein tyrosine phosphatase-like (proline instead of catalytic arginine), member A|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 1', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17589959, end_position_on_the_genomic_accession=17617390, orientation='minus', exon_count=7, OMIM=610467, no_hdr0=''),
  9201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9201, CurrentID=0, Status='live', Symbol='DCLK1', Aliases='CL1, CLICK1, DCAMKL1, DCDC3A, DCLK', description='doublecortin like kinase 1', other_designations='doublecortin domain-containing protein 3A|doublecortin-like and CAM kinase-like 1|doublecortin-like kinase 1', map_location='13q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35768652, end_position_on_the_genomic_accession=36132020, orientation='minus', exon_count=21, OMIM=604742, no_hdr0=''),
  9202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9202, CurrentID=0, Status='live', Symbol='ZMYM4', Aliases='CDIR, MYM, ZNF198L3, ZNF262', description='zinc finger, MYM-type 4', other_designations='cell death inhibiting RNA|zinc finger protein 262', map_location='1p32-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35268714, end_position_on_the_genomic_accession=35421944, orientation='plus', exon_count=36, OMIM=613568, no_hdr0=''),
  9203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9203, CurrentID=0, Status='live', Symbol='ZMYM3', Aliases='DXS6673E, MYM, XFIM, ZNF198L2, ZNF261', description='zinc finger, MYM-type 3', other_designations='zinc finger protein 261', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71239624, end_position_on_the_genomic_accession=71255261, orientation='minus', exon_count=27, OMIM=300061, no_hdr0=''),
  9204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9204, CurrentID=0, Status='live', Symbol='ZMYM6', Aliases='Buster2, MYM, ZBED7, ZNF198L4, ZNF258', description='zinc finger, MYM-type 6', other_designations='transposon-derived Buster2 transposase-like protein|zinc finger protein 258|zinc finger, BED-type containing 7|zinc finger, MYM-type containing 6', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34986165, end_position_on_the_genomic_accession=35031968, orientation='minus', exon_count=16, OMIM=613567, no_hdr0=''),
  9205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9205, CurrentID=0, Status='live', Symbol='ZMYM5', Aliases='HSPC050, MYM, ZNF198L1, ZNF237', description='zinc finger, MYM-type 5', other_designations='zinc finger protein 198-like 1|zinc finger protein 237', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19823482, end_position_on_the_genomic_accession=19863636, orientation='minus', exon_count=9, OMIM=616443, no_hdr0=''),
  402422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402422, CurrentID=0, Status='live', Symbol='TUBB4AP1', Aliases='', description='tubulin beta 4A class IVa pseudogene 1', other_designations='tubulin, beta 4 pseudogene|tubulin, beta 4A class IVa pseudogene 1', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123561548, end_position_on_the_genomic_accession=123563471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9207, CurrentID=0, Status='live', Symbol='RAD17P1', Aliases='HRAD17P1', description='RAD17 checkpoint clamp loader component pseudogene 1', other_designations='RAD17 homolog (S. pombe) pseudogene 1|RAD17 homolog, pseudogene 1', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16863865, end_position_on_the_genomic_accession=16866370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9208, CurrentID=0, Status='live', Symbol='LRRFIP1', Aliases='FLAP-1, FLAP1, FLIIAP1, GCF-2, GCF2, HUFI-1, TRIP', description='leucine rich repeat (in FLII) interacting protein 1', other_designations='GC-binding factor 2|LRR FLII-interacting protein 1|NEDD8-conjugating enzyme|TAR RNA-interacting protein', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237627581, end_position_on_the_genomic_accession=237781647, orientation='plus', exon_count=29, OMIM=603256, no_hdr0=''),
  9209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9209, CurrentID=0, Status='live', Symbol='LRRFIP2', Aliases='HUFI-2', description='leucine rich repeat (in FLII) interacting protein 2', other_designations='LRR FLII-interacting protein 2', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37052626, end_position_on_the_genomic_accession=37176817, orientation='minus', exon_count=33, OMIM=614043, no_hdr0=''),
  9210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9210, CurrentID=0, Status='live', Symbol='BMP15', Aliases='GDF9B, ODG2, POF4', description='bone morphogenetic protein 15', other_designations='BMP-15|GDF-9B|growth/differentiation factor 9B', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50910735, end_position_on_the_genomic_accession=50916641, orientation='plus', exon_count=2, OMIM=300247, no_hdr0=''),
  9211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9211, CurrentID=0, Status='live', Symbol='LGI1', Aliases='ADLTE, ADPAEF, ADPEAF, EPITEMPIN, EPT, ETL1, IB1099', description='leucine-rich, glioma inactivated 1', other_designations='epitempin-1|leucine-rich glioma inactivated 1', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93757809, end_position_on_the_genomic_accession=93798159, orientation='plus', exon_count=8, OMIM=604619, no_hdr0=''),
  9212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9212, CurrentID=0, Status='live', Symbol='AURKB', Aliases='AIK2, AIM-1, AIM1, ARK2, AurB, IPL1, PPP1R48, STK12, STK5, aurkb-sv1, aurkb-sv2', description='aurora kinase B', other_designations='ARK-2|STK-1|aurora 1|aurora kinase B-Sv1|aurora kinase B-Sv2|aurora- and Ipl1-like midbody-associated protein 1|aurora-1|aurora-B|aurora-related kinase 2|aurora/IPL1-related kinase 2|protein phosphatase 1, regulatory subunit 48|serine/threonine kinase 12|serine/threonine-protein kinase 12|serine/threonine-protein kinase 5|serine/threonine-protein kinase aurora-B', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8204731, end_position_on_the_genomic_accession=8210776, orientation='minus', exon_count=10, OMIM=604970, no_hdr0=''),
  9213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9213, CurrentID=0, Status='live', Symbol='XPR1', Aliases='IBGC6, SYG1, X3', description='xenotropic and polytropic retrovirus receptor 1', other_designations='X-receptor|protein SYG1 homolog|xenotropic and polytropic murine leukemia virus receptor X3', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180632010, end_position_on_the_genomic_accession=180890279, orientation='plus', exon_count=15, OMIM=605237, no_hdr0=''),
  9214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9214, CurrentID=0, Status='live', Symbol='FCMR', Aliases='FAIM3, TOSO', description='Fc fragment of IgM receptor', other_designations='Fc mu receptor|IgM Fc fragment receptor|IgM Fc receptor|immunoglobulin mu Fc receptor|regulator of Fas-induced apoptosis Toso', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206903285, end_position_on_the_genomic_accession=206922033, orientation='minus', exon_count=8, OMIM=606015, no_hdr0=''),
  9215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9215, CurrentID=0, Status='live', Symbol='LARGE', Aliases='MDC1D, MDDGA6, MDDGB6', description='like-glycosyltransferase', other_designations='acetylglucosaminyltransferase-like 1A|acetylglucosaminyltransferase-like protein', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33273076, end_position_on_the_genomic_accession=33922841, orientation='minus', exon_count=21, OMIM=603590, no_hdr0=''),
  9217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9217, CurrentID=0, Status='live', Symbol='VAPB', Aliases='ALS8, VAMP-B, VAP-B', description='VAMP (vesicle-associated membrane protein)-associated protein B and C', other_designations='VAMP-associated 33 kDa protein', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58389119, end_position_on_the_genomic_accession=58451101, orientation='plus', exon_count=6, OMIM=605704, no_hdr0=''),
  9218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9218, CurrentID=0, Status='live', Symbol='VAPA', Aliases='VAP-33, VAP-A, VAP33, hVAP-33', description='VAMP associated protein A', other_designations='33 kDa VAMP-associated protein|VAMP (vesicle-associated membrane protein)-associated protein A, 33kDa|VAMP-A|VAMP-associated protein A', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9913958, end_position_on_the_genomic_accession=9960021, orientation='plus', exon_count=8, OMIM=605703, no_hdr0=''),
  9219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9219, CurrentID=0, Status='live', Symbol='MTA2', Aliases='MTA1L1, PID', description='metastasis associated 1 family member 2', other_designations='MTA1-L1 protein|metastasis -associated gene 1-like 1|metastasis associated 1 family, member 2|metastasis associated gene family, member 2|metastasis-associated 1-like 1|metastasis-associated protein 2|p53 target protein in deacetylase complex', map_location='11q12-q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62593203, end_position_on_the_genomic_accession=62601840, orientation='minus', exon_count=18, OMIM=603947, no_hdr0=''),
  9220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9220, CurrentID=0, Status='live', Symbol='TIAF1', Aliases='MAJN, SPR210', description='TGFB1-induced anti-apoptotic factor 1', other_designations='12 kDa TGF-beta-1-induced antiapoptotic factor|TGF-beta-1-induced antiapoptotic factor 1|molecule associated with Jak-3 N-terminal', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29073521, end_position_on_the_genomic_accession=29075609, orientation='minus', exon_count=1, OMIM=609517, no_hdr0=''),
  9221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9221, CurrentID=0, Status='live', Symbol='NOLC1', Aliases='NOPP130, NOPP140, NS5ATP13, P130', description='nucleolar and coiled-body phosphoprotein 1', other_designations='140 kDa nucleolar phosphoprotein|HCV NS5A trans-regulated protein 13|HCV NS5A-transactivated protein 13|hepatitis C virus NS5A-transactivated protein 13|nucleolar 130 kDa protein|nucleolar and coiled-body phosphprotein 1|nucleolar phosphoprotein p130|nucleolar protein p130', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102152176, end_position_on_the_genomic_accession=102163871, orientation='plus', exon_count=13, OMIM=602394, no_hdr0=''),
  9223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9223, CurrentID=0, Status='live', Symbol='MAGI1', Aliases='AIP-3, AIP3, BAIAP1, BAP-1, BAP1, MAGI-1, Magi1d, TNRC19, WWP3', description='membrane associated guanylate kinase, WW and PDZ domain containing 1', other_designations='BAI1-associated protein 1|WW domain-containing protein 3|atrophin-1-interacting protein 3|membrane-associated guanylate kinase inverted 1|trinucleotide repeat-containing gene 19 protein', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65330764, end_position_on_the_genomic_accession=66038836, orientation='minus', exon_count=31, OMIM=602625, no_hdr0=''),
  9227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9227, CurrentID=0, Status='live', Symbol='LRAT', Aliases='LCA14', description='lecithin retinol acyltransferase (phosphatidylcholine--retinol O-acyltransferase)', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154626961, end_position_on_the_genomic_accession=154753119, orientation='plus', exon_count=8, OMIM=604863, no_hdr0=''),
  9228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9228, CurrentID=0, Status='live', Symbol='DLGAP2', Aliases='DAP2, SAPAP2', description='discs, large (Drosophila) homolog-associated protein 2', other_designations='PSD-95/SAP90-binding protein 2|SAP90/PSD-95-associated protein 2|discs large-associated protein 2', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1019561, end_position_on_the_genomic_accession=1708476, orientation='plus', exon_count=18, OMIM=605438, no_hdr0=''),
  9229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9229, CurrentID=0, Status='live', Symbol='DLGAP1', Aliases='DAP-1, DAP-1-ALPHA, DAP-1-BETA, DAP1A, DLGAP1B, GKAP, SAPAP1, DLGAP1', description='discs, large (Drosophila) homolog-associated protein 1', other_designations='PSD-95/SAP90 binding protein 1|SAP90/PSD-95-associated protein 1|guanylate kinase-associated protein', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3496032, end_position_on_the_genomic_accession=4455310, orientation='minus', exon_count=21, OMIM=605445, no_hdr0=''),
  9230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9230, CurrentID=0, Status='live', Symbol='RAB11B', Aliases='H-YPT3', description='RAB11B, member RAS oncogene family', other_designations='GTP-binding protein YPT3|RAB11B, member of RAS oncogene family', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8390321, end_position_on_the_genomic_accession=8404434, orientation='plus', exon_count=5, OMIM=604198, no_hdr0=''),
  9231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9231, CurrentID=0, Status='live', Symbol='DLG5', Aliases='LP-DLG, P-DLG5, PDLG', description='discs, large homolog 5 (Drosophila)', other_designations='discs large protein LP-DLG|discs large protein P-dlg|large type of P-DLG|placenta and prostate DLG', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77790787, end_position_on_the_genomic_accession=77926590, orientation='minus', exon_count=40, OMIM=604090, no_hdr0=''),
  9232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9232, CurrentID=0, Status='live', Symbol='PTTG1', Aliases='EAP1, HPTTG, PTTG, TUTR1', description='pituitary tumor-transforming 1', other_designations='ESP1-associated protein 1|pituitary tumor-transforming gene 1 protein|tumor-transforming protein 1', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160421807, end_position_on_the_genomic_accession=160428744, orientation='plus', exon_count=6, OMIM=604147, no_hdr0=''),
  9235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9235, CurrentID=0, Status='live', Symbol='IL32', Aliases='IL-32alpha, IL-32beta, IL-32delta, IL-32gamma, NK4, TAIF, TAIFa, TAIFb, TAIFc, TAIFd', description='interleukin 32', other_designations='interleukin-32 eta|interleukin-32 small|interleukin-32 theta|natural killer cell transcript 4|natural killer cells protein 4|tumor necrosis factor alpha-inducing factor', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3065312, end_position_on_the_genomic_accession=3069667, orientation='plus', exon_count=7, OMIM=606001, no_hdr0=''),
  9236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9236, CurrentID=0, Status='live', Symbol='CCPG1', Aliases='CPR8', description='cell cycle progression 1', other_designations='cell cycle progression restoration protein 8', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55355223, end_position_on_the_genomic_accession=55408510, orientation='minus', exon_count=9, OMIM=611326, no_hdr0=''),
  9238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9238, CurrentID=0, Status='live', Symbol='TBRG4', Aliases='CPR2, FASTKD4', description='transforming growth factor beta regulator 4', other_designations='FAST kinase domain-containing protein 4|FAST kinase domains 4|H_TD2522F11.8|cell cycle progression protein 2|cell cycle progression restoration protein 2', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45100100, end_position_on_the_genomic_accession=45111747, orientation='minus', exon_count=11, OMIM=611325, no_hdr0=''),
  103504729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504729, CurrentID=0, Status='live', Symbol='MIR3179-4', Aliases='hsa-mir-3179-4', description='microRNA 3179-4', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18494493, end_position_on_the_genomic_accession=18494576, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9240, CurrentID=0, Status='live', Symbol='PNMA1', Aliases='MA1', description='paraneoplastic Ma antigen 1', other_designations='37 kDa neuronal protein|neuron- and testis-specific protein 1|paraneoplastic neuronal antigen MA1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73711783, end_position_on_the_genomic_accession=73714425, orientation='minus', exon_count=1, OMIM=604010, no_hdr0=''),
  9241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9241, CurrentID=0, Status='live', Symbol='NOG', Aliases='SYM1, SYNS1', description='noggin', other_designations='symphalangism 1 (proximal)', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56593699, end_position_on_the_genomic_accession=56595590, orientation='plus', exon_count=1, OMIM=602991, no_hdr0=''),
  9242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9242, CurrentID=0, Status='live', Symbol='MSC', Aliases='ABF-1, ABF1, MYOR, bHLHa22', description='musculin', other_designations='activated B-cell factor 1, homolog of mouse musculin|activated B-cell factor-1|class A basic helix-loop-helix protein 22', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71841542, end_position_on_the_genomic_accession=71844496, orientation='minus', exon_count=2, OMIM=603628, no_hdr0=''),
  9244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9244, CurrentID=0, Status='live', Symbol='CRLF1', Aliases='CISS, CISS1, CLF, CLF-1, NR6, zcytor5', description='cytokine receptor-like factor 1', other_designations='class I cytokine receptor|cytokine type 1 receptor CRLP-1|cytokine-like factor 1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18593225, end_position_on_the_genomic_accession=18606850, orientation='minus', exon_count=11, OMIM=604237, no_hdr0=''),
  9245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9245, CurrentID=0, Status='live', Symbol='GCNT3', Aliases='C2/4GnT, C24GNT, C2GNT2, C2GNTM, GNTM', description='glucosaminyl (N-acetyl) transferase 3, mucin type', other_designations='C2GnT-mucin type|beta1,6-N-acetylglucosaminyltransferase-M|beta1,6GlcNAc-transferase|core 2/core 4 beta-1,6-N-acetylglucosaminyltransferase|hC2GnT-M|mucus-type core 2 beta-1,6-N-acetylglucosaminyltransferase', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59611783, end_position_on_the_genomic_accession=59620011, orientation='plus', exon_count=3, OMIM=606836, no_hdr0=''),
  9246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9246, CurrentID=0, Status='live', Symbol='UBE2L6', Aliases='RIG-B, UBCH8', description='ubiquitin conjugating enzyme E2L 6', other_designations='E2 ubiquitin-conjugating enzyme L6|retinoic acid induced gene B protein|retinoic acid-induced gene B protein|ubiquitin carrier protein L6|ubiquitin-conjugating enzyme E2L 6|ubiquitin-protein ligase L6', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57551655, end_position_on_the_genomic_accession=57568330, orientation='minus', exon_count=5, OMIM=603890, no_hdr0=''),
  9247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9247, CurrentID=0, Status='live', Symbol='GCM2', Aliases='GCMB, hGCMb', description='glial cells missing homolog 2 (Drosophila)', other_designations='GCM motif protein 2|glial cells missing 2|glial cells missing homolog b|glide/gcm protein homolog', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10872724, end_position_on_the_genomic_accession=10886430, orientation='minus', exon_count=6, OMIM=603716, no_hdr0=''),
  9248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9248, CurrentID=0, Status='live', Symbol='GPR50', Aliases='H9, Mel1c', description='G protein-coupled receptor 50', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151176584, end_position_on_the_genomic_accession=151182855, orientation='plus', exon_count=2, OMIM=300207, no_hdr0=''),
  9249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9249, CurrentID=0, Status='live', Symbol='DHRS3', Aliases='DD83.1, RDH17, Rsdr1, SDR1, SDR16C1, retSDR1', description='dehydrogenase/reductase (SDR family) member 3', other_designations='dehydrogenase/reductase member 3|retinal short-chain dehydrogenase/reductase 1|retinol dehydrogenase 17|short chain dehydrogenase/reductase family 16C member 1|short chain dehydrogenase/reductase family 16C, member 1|short-chain dehydrogenase/reductase 1', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12567910, end_position_on_the_genomic_accession=12618361, orientation='minus', exon_count=8, OMIM=612830, no_hdr0=''),
  100270797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270797, CurrentID=0, Status='live', Symbol='STQTL14', Aliases='', description='Stature QTL 14', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612228, no_hdr0=''),
  103504731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504731, CurrentID=0, Status='live', Symbol='MIR3648-2', Aliases='hsa-mir-3648-2', description='microRNA 3648-2', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8986999, end_position_on_the_genomic_accession=8987178, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  9252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9252, CurrentID=0, Status='live', Symbol='RPS6KA5', Aliases='MSK1, MSPK1, RLPK', description='ribosomal protein S6 kinase, 90kDa, polypeptide 5', other_designations='90 kDa ribosomal protein S6 kinase 5|RSK-like protein kinase|RSKL|S6K-alpha-5|nuclear mitogen- and stress-activated protein kinase 1', map_location='14q31-q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90868122, end_position_on_the_genomic_accession=91060649, orientation='minus', exon_count=18, OMIM=603607, no_hdr0=''),
  9253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9253, CurrentID=0, Status='live', Symbol='NUMBL', Aliases='CAG3A, CTG3a, NBL, NUMB-RIKE, NUMBR, TNRC23, NUMBL', description='numb homolog (Drosophila)-like', other_designations='numb homolog-like|numb-related protein', map_location='19q13.13-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40665905, end_position_on_the_genomic_accession=40691675, orientation='minus', exon_count=13, OMIM=604018, no_hdr0=''),
  9254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9254, CurrentID=0, Status='live', Symbol='CACNA2D2', Aliases='CACNA2D', description='calcium channel, voltage-dependent, alpha 2/delta subunit 2', other_designations='alpha 2 delta calcium channel subunit|gene 26|voltage-gated calcium channel subunit alpha-2/delta-2', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50362613, end_position_on_the_genomic_accession=50504244, orientation='minus', exon_count=40, OMIM=607082, no_hdr0=''),
  664615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664615, CurrentID=0, Status='live', Symbol='MIR376A2', Aliases='MIR376A-2, MIRN376A-2, MIRN376A2, mir-376a-2', description='microRNA 376a-2', other_designations='hsa-mir-376a-2', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101040069, end_position_on_the_genomic_accession=101040148, orientation='plus', exon_count=1, OMIM=610960, no_hdr0=''),
  664616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664616, CurrentID=0, Status='live', Symbol='MIR487B', Aliases='MIRN487B, mir-487b', description='microRNA 487b', other_designations='hsa-mir-487b', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101046455, end_position_on_the_genomic_accession=101046538, orientation='plus', exon_count=1, OMIM=615037, no_hdr0=''),
  664617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664617, CurrentID=0, Status='live', Symbol='MIR542', Aliases='MIRN542, hsa-mir-542, mir-542', description='microRNA 542', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134541341, end_position_on_the_genomic_accession=134541437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9258, CurrentID=0, Status='live', Symbol='MFHAS1', Aliases='LRRC65, MASL1', description='malignant fibrous histiocytoma amplified sequence 1', other_designations='MFH-amplified sequences with leucine-rich tandem repeats 1|leucine rich repeat containing 65|malignant fibrous histiocytoma-amplified sequence with leucine-rich tandem repeats 1|malignant fibrous histiocytoma-amplified sequences with leucine-rich tandem repeats 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8783354, end_position_on_the_genomic_accession=8893621, orientation='minus', exon_count=5, OMIM=605352, no_hdr0=''),
  100271623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271623, CurrentID=0, Status='live', Symbol='RPL23AP69', Aliases='RPL23A_32_1304', description='ribosomal protein L23a pseudogene 69', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25206052, end_position_on_the_genomic_accession=25206481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9260, CurrentID=0, Status='live', Symbol='PDLIM7', Aliases='LMP1, LMP3', description='PDZ and LIM domain 7', other_designations='1110003B01Rik|LIM domain protein|LMP|Lim mineralization protein 3|PDZ and LIM domain 7 (enigma)|protein enigma', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177483394, end_position_on_the_genomic_accession=177497619, orientation='minus', exon_count=15, OMIM=605903, no_hdr0=''),
  9261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9261, CurrentID=0, Status='live', Symbol='MAPKAPK2', Aliases='MAPKAP-K2, MK-2, MK2', description='mitogen-activated protein kinase-activated protein kinase 2', other_designations='MAPK-activated protein kinase 2|MAPKAP kinase 2', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206684912, end_position_on_the_genomic_accession=206734285, orientation='plus', exon_count=11, OMIM=602006, no_hdr0=''),
  9262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9262, CurrentID=0, Status='live', Symbol='STK17B', Aliases='DRAK2', description='serine/threonine kinase 17b', other_designations='DAP kinase-related apoptosis-inducing protein kinase 2|death-associated protein kinase-related 2', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196133583, end_position_on_the_genomic_accession=196176503, orientation='minus', exon_count=13, OMIM=604727, no_hdr0=''),
  9263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9263, CurrentID=0, Status='live', Symbol='STK17A', Aliases='DRAK1', description='serine/threonine kinase 17a', other_designations='DAP kinase-related apoptosis-inducing protein kinase 1|death-associated protein kinase-related 1|serine/threonine kinase 17a (apoptosis-inducing)', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43583093, end_position_on_the_genomic_accession=43627379, orientation='plus', exon_count=7, OMIM=604726, no_hdr0=''),
  9265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9265, CurrentID=0, Status='live', Symbol='CYTH3', Aliases='ARNO3, GRP1, PSCD3', description='cytohesin 3', other_designations='ARF nucleotide-binding site opener 3|PH, SEC7 and coiled-coil domain-containing protein 3|general receptor of phosphoinositides 1|pleckstrin homology, Sec7 and coiled-coil domains 3', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6161779, end_position_on_the_genomic_accession=6272611, orientation='minus', exon_count=13, OMIM=605081, no_hdr0=''),
  9266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9266, CurrentID=0, Status='live', Symbol='CYTH2', Aliases='ARNO, CTS18, CTS18.1, PSCD2, PSCD2L, SEC7L, Sec7p-L, Sec7p-like', description='cytohesin 2', other_designations='ARF exchange factor|ARF nucleotide-binding site opener|CTC-273B12.8|PH, SEC7 and coiled-coil domain-containing protein 2|pleckstrin homology, Sec7 and coiled-coil domains 2 (cytohesin-2)|pleckstrin homology, Sec7 and coiled/coil domains 2 (cytohesin-2)', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48469208, end_position_on_the_genomic_accession=48482314, orientation='plus', exon_count=13, OMIM=602488, no_hdr0=''),
  9267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9267, CurrentID=0, Status='live', Symbol='CYTH1', Aliases='B2-1, CYTOHESIN-1, D17S811E, PSCD1, SEC7', description='cytohesin 1', other_designations='PH, SEC7 and coiled-coil domain-containing protein 1|SEC7 homolog B2-1|cytoadhesin 1|homolog of secretory protein SEC7|pleckstrin homology, Sec7 and coiled-coil domains 1', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78674047, end_position_on_the_genomic_accession=78782342, orientation='minus', exon_count=18, OMIM=182115, no_hdr0=''),
  664629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664629, CurrentID=0, Status='live', Symbol='LOC664629', Aliases='', description='unknown transcript', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  9270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9270, CurrentID=0, Status='live', Symbol='ITGB1BP1', Aliases='ICAP-1A, ICAP-1B, ICAP-1alpha, ICAP1, ICAP1A, ICAP1B', description='integrin subunit beta 1 binding protein 1', other_designations='ICAP-1|bodenin|integrin beta 1 binding protein 1|integrin cytoplasmic domain-associated protein 1|integrin cytoplasmic domain-associated protein 1-alpha|integrin cytoplasmic domain-associated protein 1-beta', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9405684, end_position_on_the_genomic_accession=9423569, orientation='minus', exon_count=9, OMIM=607153, no_hdr0=''),
  9271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9271, CurrentID=0, Status='live', Symbol='PIWIL1', Aliases='CT80.1, HIWI, MIWI, PIWI', description='piwi-like RNA-mediated gene silencing 1', other_designations='piwi homolog', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130337887, end_position_on_the_genomic_accession=130426260, orientation='plus', exon_count=25, OMIM=605571, no_hdr0=''),
  9274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9274, CurrentID=0, Status='live', Symbol='BCL7C', Aliases='', description='B-cell CLL/lymphoma 7C', other_designations='', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30834041, end_position_on_the_genomic_accession=30894376, orientation='minus', exon_count=8, OMIM=605847, no_hdr0=''),
  9275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9275, CurrentID=0, Status='live', Symbol='BCL7B', Aliases='', description='B-cell CLL/lymphoma 7B', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73536353, end_position_on_the_genomic_accession=73557735, orientation='minus', exon_count=8, OMIM=605846, no_hdr0=''),
  9276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9276, CurrentID=0, Status='live', Symbol='COPB2', Aliases="beta'-COP", description='coatomer protein complex subunit beta 2 (beta prime)', other_designations="beta'-coat protein|betaprime-COP|coatomer binding complex, beta prime subunit|coatomer protein complex, subunit beta 2 (beta prime)|p102", map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139357587, end_position_on_the_genomic_accession=139389680, orientation='minus', exon_count=22, OMIM=606990, no_hdr0=''),
  9277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9277, CurrentID=0, Status='live', Symbol='WDR46', Aliases='BING4, C6orf11, FP221, UTP7', description='WD repeat domain 46', other_designations='WD repeat-containing protein BING4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33279103, end_position_on_the_genomic_accession=33289527, orientation='minus', exon_count=16, OMIM=611440, no_hdr0=''),
  9278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9278, CurrentID=0, Status='live', Symbol='ZBTB22', Aliases='BING1A, ZNF297, ZNF297A, fru, fruitless, ZBTB22', description='zinc finger and BTB domain containing 22', other_designations='zinc finger and BTB domain-containing protein 22A|zinc finger protein 297', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33314405, end_position_on_the_genomic_accession=33317942, orientation='minus', exon_count=3, OMIM=611439, no_hdr0=''),
  100270798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270798, CurrentID=0, Status='live', Symbol='IBD24', Aliases='', description='Inflammatory bowel disease-24', other_designations='', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612566, no_hdr0=''),
  9282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9282, CurrentID=0, Status='live', Symbol='MED14', Aliases='CRSP150, CRSP2, CSRP, CXorf4, DRIP150, EXLM1, RGR1, TRAP170', description='mediator complex subunit 14', other_designations='ARC150|CRSP complex subunit 2|RGR1 homolog|activator-recruited cofactor 150 kDa component|cofactor required for Sp1 transcriptional activation, subunit 2 (150kD)|hRGR1|human homolog of yeast RGR1|thyroid hormone receptor-associated protein complex 170 kDa component|thyroid hormone receptor-associated protein complex component TRAP170|transcriptional co-activator CRSP150|vitamin D receptor-interacting protein complex component DRIP150|vitamin D3 receptor-interacting protein complex 150 kDa component', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40649543, end_position_on_the_genomic_accession=40736122, orientation='minus', exon_count=34, OMIM=300182, no_hdr0=''),
  9283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9283, CurrentID=0, Status='live', Symbol='GPR37L1', Aliases='ET(B)R-LP-2, ETBR-LP-2', description='G protein-coupled receptor 37 like 1', other_designations='G-protein coupled receptor 37 like 1|G-protein coupled receptor 37-like 1|endothelin B receptor-like protein 2|endothelin type b receptor-like protein 2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202118705, end_position_on_the_genomic_accession=202131425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9284, CurrentID=0, Status='live', Symbol='NPIPA1', Aliases='NPIP, NPIPA, morpheus', description='nuclear pore complex interacting protein family member A1', other_designations='nuclear pore complex interacting protein family, member A1|nuclear pore complex-interacting protein A type', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14935310, end_position_on_the_genomic_accession=14952074, orientation='plus', exon_count=12, OMIM=606406, no_hdr0=''),
  9287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9287, CurrentID=0, Status='live', Symbol='TAAR2', Aliases='GPR58, taR-2', description='trace amine associated receptor 2', other_designations='G protein-coupled receptor 58', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132599627, end_position_on_the_genomic_accession=132624358, orientation='minus', exon_count=4, OMIM=604849, no_hdr0=''),
  9288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9288, CurrentID=0, Status='live', Symbol='TAAR3', Aliases='GPR57, GPR58, TAAR2', description='trace amine associated receptor 3 (gene/pseudogene)', other_designations='G protein-coupled receptor 57|trace amine associated receptor 2 pseudogene', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132608225, end_position_on_the_genomic_accession=132609302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9289, CurrentID=0, Status='live', Symbol='ADGRG1', Aliases='BFPP, BPPR, GPR56, TM7LN4, TM7XN1', description='adhesion G protein-coupled receptor G1', other_designations='7-transmembrane protein with no EGF-like N-terminal domains-1|G protein-coupled receptor 56', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57619535, end_position_on_the_genomic_accession=57665039, orientation='plus', exon_count=23, OMIM=604110, no_hdr0=''),
  9290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9290, CurrentID=0, Status='live', Symbol='GPR55', Aliases='LPIR1', description='G protein-coupled receptor 55', other_designations='', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230907318, end_position_on_the_genomic_accession=230961227, orientation='minus', exon_count=5, OMIM=604107, no_hdr0=''),
  9292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9292, CurrentID=0, Status='live', Symbol='GPR53P', Aliases='PsiGPR53, dJ271M21.8', description='G protein-coupled receptor 53, pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29537812, end_position_on_the_genomic_accession=29538651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9293, CurrentID=0, Status='live', Symbol='GPR52', Aliases='', description='G protein-coupled receptor 52', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174448074, end_position_on_the_genomic_accession=174449545, orientation='plus', exon_count=1, OMIM=604106, no_hdr0=''),
  9294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9294, CurrentID=0, Status='live', Symbol='S1PR2', Aliases='AGR16, EDG-5, EDG5, Gpcr13, H218, LPB2, S1P2', description='sphingosine-1-phosphate receptor 2', other_designations='CTD-2369P2.2|S1P receptor 2|S1P receptor EDG5|S1P receptor Edg-5|endothelial differentiation G-protein coupled receptor 5|endothelial differentiation, sphingolipid G-protein-coupled receptor, 5|sphingosine 1-phosphate receptor Edg-5', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10221433, end_position_on_the_genomic_accession=10231807, orientation='minus', exon_count=2, OMIM=605111, no_hdr0=''),
  9295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9295, CurrentID=0, Status='live', Symbol='SRSF11', Aliases='NET2, SFRS11, dJ677H15.2, p54', description='serine/arginine-rich splicing factor 11', other_designations='SR splicing factor 11|arginine-rich 54 kDa nuclear protein|splicing factor p54|splicing factor, arginine/serine-rich 11', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70205682, end_position_on_the_genomic_accession=70252020, orientation='plus', exon_count=15, OMIM=602010, no_hdr0=''),
  9296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9296, CurrentID=0, Status='live', Symbol='ATP6V1F', Aliases='ATP6S14, VATF, Vma7', description='ATPase, H+ transporting, lysosomal 14kDa, V1 subunit F', other_designations='ATPase, vacuolar, 14 kD|H(+)-transporting two-sector ATPase, 14kD subunit|V-ATPase 14 kDa subunit|V-ATPase F subunit|V-ATPase subunit F|adenosinetriphosphatase 14k chain|vacuolar ATP synthase subunit F|vacuolar proton pump F subunit|vacuolar proton pump subunit F', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128862803, end_position_on_the_genomic_accession=128865849, orientation='plus', exon_count=3, OMIM=607160, no_hdr0=''),
  9297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9297, CurrentID=0, Status='live', Symbol='SNORD29', Aliases='RNU29, U29', description='small nucleolar RNA, C/D box 29', other_designations='RNA, U29 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62853904, end_position_on_the_genomic_accession=62853968, orientation='minus', exon_count=1, OMIM=603228, no_hdr0=''),
  9298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9298, CurrentID=0, Status='live', Symbol='SNORD31', Aliases='RNU31, U31', description='small nucleolar RNA, C/D box 31', other_designations='RNA, U31 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62853326, end_position_on_the_genomic_accession=62853393, orientation='minus', exon_count=1, OMIM=603230, no_hdr0=''),
  9299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9299, CurrentID=0, Status='live', Symbol='SNORD30', Aliases='RNU30, U30', description='small nucleolar RNA, C/D box 30', other_designations='RNA, U30 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62853663, end_position_on_the_genomic_accession=62853732, orientation='minus', exon_count=1, OMIM=603229, no_hdr0=''),
  9300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9300, CurrentID=0, Status='live', Symbol='SNORD28', Aliases='RNU28, U28', description='small nucleolar RNA, C/D box 28', other_designations='RNA, U28 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62854621, end_position_on_the_genomic_accession=62854695, orientation='minus', exon_count=1, OMIM=603227, no_hdr0=''),
  9301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9301, CurrentID=0, Status='live', Symbol='SNORD27', Aliases='RNU27, U27', description='small nucleolar RNA, C/D box 27', other_designations='RNA, U27 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62855012, end_position_on_the_genomic_accession=62855083, orientation='minus', exon_count=1, OMIM=603226, no_hdr0=''),
  9302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9302, CurrentID=0, Status='live', Symbol='SNORD26', Aliases='RNU26, U26', description='small nucleolar RNA, C/D box 26', other_designations='RNA, U26 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62855292, end_position_on_the_genomic_accession=62855366, orientation='minus', exon_count=1, OMIM=603225, no_hdr0=''),
  9303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9303, CurrentID=0, Status='live', Symbol='SNORD25', Aliases='RNU25, U25', description='small nucleolar RNA, C/D box 25', other_designations='RNA, U25 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62855565, end_position_on_the_genomic_accession=62855631, orientation='minus', exon_count=1, OMIM=603224, no_hdr0=''),
  9304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9304, CurrentID=0, Status='live', Symbol='SNORD22', Aliases='RNU22, U22', description='small nucleolar RNA, C/D box 22', other_designations='RNA, U22 small nucleolar', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62852910, end_position_on_the_genomic_accession=62853035, orientation='minus', exon_count=1, OMIM=603223, no_hdr0=''),
  103504740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504740, CurrentID=0, Status='live', Symbol='MIR3670-4', Aliases='hsa-mir-3670-4', description='microRNA 3670-4', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18488301, end_position_on_the_genomic_accession=18488365, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9306, CurrentID=0, Status='live', Symbol='SOCS6', Aliases='CIS-4, CIS4, HSPC060, SOCS-4, SOCS-6, SOCS4, SSI4, STAI4, STATI4', description='suppressor of cytokine signaling 6', other_designations='STAT induced STAT inhibitor-4|cytokine-inducible SH2 protein 4|suppressor of cytokine signaling 4', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70288901, end_position_on_the_genomic_accession=70330199, orientation='plus', exon_count=3, OMIM=605118, no_hdr0=''),
  100271631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271631, CurrentID=0, Status='live', Symbol='RPL35AP24', Aliases='RPL35A_12_1042', description='ribosomal protein L35a pseudogene 24', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49909545, end_position_on_the_genomic_accession=49909987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9308, CurrentID=0, Status='live', Symbol='CD83', Aliases='BL11, HB15', description='CD83 molecule', other_designations='B-cell activation protein|CD83 antigen (activated B lymphocytes, immunoglobulin superfamily)|cell surface protein HB15|cell-surface glycoprotein|hCD83', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14117256, end_position_on_the_genomic_accession=14136918, orientation='plus', exon_count=6, OMIM=604534, no_hdr0=''),
  105371117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371117, CurrentID=0, Status='live', Symbol='LOC105371117', Aliases='', description='uncharacterized LOC105371117', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19976376, end_position_on_the_genomic_accession=19985078, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9310, CurrentID=0, Status='live', Symbol='ZNF235', Aliases='ANF270, HZF6, ZFP93, ZNF270', description='zinc finger protein 235', other_designations='zfp-93|zinc finger protein 270|zinc finger protein 6|zinc finger protein 93 homolog|zinc finger protein HZF6|zinc finger protein homologous to Zfp93 in mouse|zinc finger protein homologous to mouse Zfp93', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44286348, end_position_on_the_genomic_accession=44305707, orientation='minus', exon_count=5, OMIM=604749, no_hdr0=''),
  9311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9311, CurrentID=0, Status='live', Symbol='ASIC3', Aliases='ACCN3, DRASIC, SLNAC1, TNaC1', description='acid sensing ion channel subunit 3', other_designations='acid sensing (proton gated) ion channel 3|acid sensing ion channel 3|acid-sensing (proton-gated) ion channel 3|amiloride-sensitive cation channel 3, testis|modulatory subunit of ASIC2a|neuronal amiloride-sensitive cation channel 3|proton-gated cation channel subunit|testis sodium channel 1', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151048292, end_position_on_the_genomic_accession=151052756, orientation='plus', exon_count=11, OMIM=611741, no_hdr0=''),
  9312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9312, CurrentID=0, Status='live', Symbol='KCNB2', Aliases='KV2.2', description='potassium channel, voltage gated Shab related subfamily B, member 2', other_designations='delayed rectifier potassium channel protein|potassium channel Kv2.2|potassium voltage-gated channel, Shab-related subfamily, member 2|voltage-gated potassium channel subunit Kv2.2', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72537391, end_position_on_the_genomic_accession=72938349, orientation='plus', exon_count=3, OMIM=607738, no_hdr0=''),
  9313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9313, CurrentID=0, Status='live', Symbol='MMP20', Aliases='AI2A2, MMP-20', description='matrix metallopeptidase 20', other_designations='enamel metalloproteinase|matrix metalloproteinase 20 (enamelysin)', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102576835, end_position_on_the_genomic_accession=102625332, orientation='minus', exon_count=10, OMIM=604629, no_hdr0=''),
  9314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9314, CurrentID=0, Status='live', Symbol='KLF4', Aliases='EZF, GKLF', description='Kruppel-like factor 4 (gut)', other_designations='endothelial Kruppel-like zinc finger protein|epithelial zinc finger protein EZF|gut-enriched krueppel-like factor', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107484852, end_position_on_the_genomic_accession=107489766, orientation='minus', exon_count=5, OMIM=602253, no_hdr0=''),
  9315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9315, CurrentID=0, Status='live', Symbol='NREP', Aliases='C5orf13, D4S114, P311, PRO1873, PTZ17, SEZ17', description='neuronal regeneration related protein', other_designations='neuronal protein 3.1|neuronal regeneration related protein homolog|protein p311', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111729303, end_position_on_the_genomic_accession=111976931, orientation='minus', exon_count=9, OMIM=607332, no_hdr0=''),
  9317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9317, CurrentID=0, Status='live', Symbol='PTER', Aliases='HPHRP, RPR-1', description='phosphotriesterase related', other_designations='parathion hydrolase-related protein|resiniferatoxin-binding, phosphotriesterase-related', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16436943, end_position_on_the_genomic_accession=16513745, orientation='plus', exon_count=7, OMIM=604446, no_hdr0=''),
  9318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9318, CurrentID=0, Status='live', Symbol='COPS2', Aliases='ALIEN, CSN2, SGN2, TRIP15', description='COP9 signalosome subunit 2', other_designations='COP9 constitutive photomorphogenic homolog subunit 2|JAB1-containing signalosome subunit 2|TR-interacting protein 15|TRIP-15|alien homolog|signalosome subunit 2|thyroid receptor interacting protein 15|thyroid receptor-interacting protein 15', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49125274, end_position_on_the_genomic_accession=49155657, orientation='minus', exon_count=13, OMIM=604508, no_hdr0=''),
  132625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132625, CurrentID=0, Status='live', Symbol='ZFP42', Aliases='REX-1, REX1, ZNF754, zfp-42', description='ZFP42 zinc finger protein', other_designations='REX1 transcription factor|reduced expression protein 1|zinc finger protein 754', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187994054, end_position_on_the_genomic_accession=188005050, orientation='plus', exon_count=5, OMIM=614572, no_hdr0=''),
  9320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9320, CurrentID=0, Status='live', Symbol='TRIP12', Aliases='TRIP-12, ULF', description='thyroid hormone receptor interactor 12', other_designations='E3 ubiquitin-protein ligase for Arf|TR-interacting protein 12|probable E3 ubiquitin-protein ligase TRIP12|thyroid receptor interacting protein 12|thyroid receptor-interacting protein 12', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229763837, end_position_on_the_genomic_accession=229923234, orientation='minus', exon_count=47, OMIM=604506, no_hdr0=''),
  9321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9321, CurrentID=0, Status='live', Symbol='TRIP11', Aliases='ACG1A, CEV14, GMAP-210, GMAP210, TRIP-11, TRIP230', description='thyroid hormone receptor interactor 11', other_designations='Golgi-microtubule-associated protein of 210 kDa|TR-interacting protein 11|clonal evolution-related gene on chromosome 14 protein|golgi-associated microtubule-binding protein 210', map_location='14q31-q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91965991, end_position_on_the_genomic_accession=92040140, orientation='minus', exon_count=22, OMIM=604505, no_hdr0=''),
  9322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9322, CurrentID=0, Status='live', Symbol='TRIP10', Aliases='CIP4, HSTP, STOT, STP, TRIP-10', description='thyroid hormone receptor interactor 10', other_designations='TR-interacting protein 10|protein Felic|salt tolerant protein|salt tolerator|thyroid receptor-interacting protein 10', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6739682, end_position_on_the_genomic_accession=6751526, orientation='plus', exon_count=16, OMIM=604504, no_hdr0=''),
  9324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9324, CurrentID=0, Status='live', Symbol='HMGN3', Aliases='PNAS-24, PNAS-25, TRIP7', description='high mobility group nucleosomal binding domain 3', other_designations='TR-interacting protein 7|thyroid hormone receptor interacting protein 7', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79201245, end_position_on_the_genomic_accession=79234738, orientation='minus', exon_count=7, OMIM=604502, no_hdr0=''),
  9325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9325, CurrentID=0, Status='live', Symbol='TRIP4', Aliases='ASC-1, ASC1, HsT17391, ZC2HC5', description='thyroid hormone receptor interactor 4', other_designations='TR-interacting protein 4|TRIP-4|thyroid receptor interacting protein 4|thyroid receptor-interacting protein 4|zinc finger, C2HC5-type', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64387804, end_position_on_the_genomic_accession=64455303, orientation='plus', exon_count=13, OMIM=604501, no_hdr0=''),
  9326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9326, CurrentID=0, Status='live', Symbol='ZNHIT3', Aliases='TRIP3', description='zinc finger, HIT-type containing 3', other_designations='HNF-4a coactivator|TR-interacting protein 3|TRIP-3|thyroid hormone receptor interactor 3|thyroid receptor interacting protein 3|thyroid receptor-interacting protein 3|zinc finger, HIT domain containing 3|zinc finger, HIT type 3', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36486627, end_position_on_the_genomic_accession=36499310, orientation='plus', exon_count=6, OMIM=604500, no_hdr0=''),
  9328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9328, CurrentID=0, Status='live', Symbol='GTF3C5', Aliases='TFIIIC63, TFIIICepsilon, TFiiiC2-63', description='general transcription factor IIIC subunit 5', other_designations='TF3C-epsilon|TFIIIC 63 kDa subunit|general transcription factor IIIC, polypeptide 5, 63kDa|transcription factor IIIC 63 kDa subunit|transcription factor IIIC subunit epsilon|transcription factor IIIC, 63 kD', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133030675, end_position_on_the_genomic_accession=133058503, orientation='plus', exon_count=12, OMIM=604890, no_hdr0=''),
  9329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9329, CurrentID=0, Status='live', Symbol='GTF3C4', Aliases='KAT12, TF3C-delta, TFIII90, TFIIIC290, TFIIIC90, TFIIICDELTA', description='general transcription factor IIIC subunit 4', other_designations='TFIIIC 90 kDa subunit|general transcription factor 3C 4|general transcription factor IIIC, polypeptide 4, 90kDa|transcription factor IIIC 90 kDa subunit|transcription factor IIIC subunit delta', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132670035, end_position_on_the_genomic_accession=132690084, orientation='plus', exon_count=6, OMIM=604892, no_hdr0=''),
  9330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9330, CurrentID=0, Status='live', Symbol='GTF3C3', Aliases='TFIIIC102, TFIIICgamma, TFiiiC2-102', description='general transcription factor IIIC subunit 3', other_designations='TF3C-gamma|TFIIIC 102 kDa subunit|general transcription factor IIIC, polypeptide 3, 102kDa|transcription factor IIIC subunit gamma', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196763032, end_position_on_the_genomic_accession=196799768, orientation='minus', exon_count=19, OMIM=604888, no_hdr0=''),
  9331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9331, CurrentID=0, Status='live', Symbol='B4GALT6', Aliases='B4Gal-T6, beta4Gal-T6', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 6', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 6|UDP-Gal:glucosylceramide beta-1,4-galactosyltransferase|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 6|beta-1,4-GalTase 6|beta4GalT-VI', map_location='18q11', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31622246, end_position_on_the_genomic_accession=31725286, orientation='minus', exon_count=12, OMIM=604017, no_hdr0=''),
  9332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9332, CurrentID=0, Status='live', Symbol='CD163', Aliases='M130, MM130', description='CD163 molecule', other_designations='hemoglobin scavenger receptor|macrophage-associated antigen', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7470816, end_position_on_the_genomic_accession=7503818, orientation='minus', exon_count=17, OMIM=605545, no_hdr0=''),
  9333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9333, CurrentID=0, Status='live', Symbol='TGM5', Aliases='PSS2, TGASE5, TGASEX, TGM6, TGMX, TGX', description='transglutaminase 5', other_designations='TG(X)|TGase X|TGase-5|transglutaminase V|transglutaminase X|transglutaminase-5', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43232595, end_position_on_the_genomic_accession=43267023, orientation='minus', exon_count=15, OMIM=603805, no_hdr0=''),
  9334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9334, CurrentID=0, Status='live', Symbol='B4GALT5', Aliases='B4Gal-T5, BETA4-GALT-IV, beta4Gal-T5, beta4GalT-V, gt-V', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 5', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 5|UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 5|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 5|beta-1,4-GalT II|beta-1,4-GalT IV|beta-1,4-GalTase 5|beta-1.4-galactosyltransferase V|beta4-GalT IV', map_location='20q13.1-q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49632945, end_position_on_the_genomic_accession=49713884, orientation='minus', exon_count=11, OMIM=604016, no_hdr0=''),
  9337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9337, CurrentID=0, Status='live', Symbol='CNOT8', Aliases='CAF1, CALIF, Caf1b, POP2, hCAF1', description='CCR4-NOT transcription complex subunit 8', other_designations='CAF1-like protein|CAF2|CALIFp|CCR4-NOT transcription complex, subunit 8|CCR4-associated factor 8|PGK promoter directed over production', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154857975, end_position_on_the_genomic_accession=154876792, orientation='plus', exon_count=11, OMIM=603731, no_hdr0=''),
  9338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9338, CurrentID=0, Status='live', Symbol='TCEAL1', Aliases='SIIR, WEX9, p21, pp21', description='transcription elongation factor A (SII)-like 1', other_designations='TCEA-like protein 1|nuclear phosphoprotein p21/SIIR|transcription elongation factor S-II protein-like 1', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103628720, end_position_on_the_genomic_accession=103630953, orientation='plus', exon_count=4, OMIM=300237, no_hdr0=''),
  644261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644261, CurrentID=0, Status='live', Symbol='LOC644261', Aliases='', description='solute carrier family 25 member 53 pseudogene', other_designations='mitochondrial carrier triple repeat 6 pseudogene|solute carrier family 25, member 53 pseudogene', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103928097, end_position_on_the_genomic_accession=103929016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9340, CurrentID=0, Status='live', Symbol='GLP2R', Aliases='', description='glucagon-like peptide 2 receptor', other_designations='GLP-2 receptor', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9825543, end_position_on_the_genomic_accession=9892104, orientation='plus', exon_count=14, OMIM=603659, no_hdr0=''),
  9341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9341, CurrentID=0, Status='live', Symbol='VAMP3', Aliases='CEB', description='vesicle associated membrane protein 3', other_designations='VAMP-3|cellubrevin|synaptobrevin-3', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7771269, end_position_on_the_genomic_accession=7781432, orientation='plus', exon_count=5, OMIM=603657, no_hdr0=''),
  9342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9342, CurrentID=0, Status='live', Symbol='SNAP29', Aliases='CEDNIK, SNAP-29', description='synaptosome associated protein 29kDa', other_designations='soluble 29 kDa NSF attachment protein|synaptosomal-associated protein, 29kD|synaptosomal-associated protein, 29kDa|vesicle-membrane fusion protein SNAP-29', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20859004, end_position_on_the_genomic_accession=20891214, orientation='plus', exon_count=5, OMIM=604202, no_hdr0=''),
  9343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9343, CurrentID=0, Status='live', Symbol='EFTUD2', Aliases='MFDGA, MFDM, SNRNP116, Snrp116, Snu114, U5-116KD', description='elongation factor Tu GTP binding domain containing 2', other_designations='SNU114 homolog|U5 snRNP-specific protein, 116 kDa|elongation factor Tu GTP-binding domain-containing protein 2|hSNU114', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44850287, end_position_on_the_genomic_accession=44899625, orientation='minus', exon_count=29, OMIM=603892, no_hdr0=''),
  9344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9344, CurrentID=0, Status='live', Symbol='TAOK2', Aliases='MAP3K17, PSK, PSK1, PSK1-BETA, TAO1, TAO2', description='TAO kinase 2', other_designations='PSK-1|hKFC-C|kinase from chicken homolog C|prostate derived STE20-like kinase PSK|prostate-derived STE20-like kinase 1|prostate-derived sterile 20-like kinase 1|thousand and one amino acid protein kinase 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29973641, end_position_on_the_genomic_accession=29992261, orientation='plus', exon_count=22, OMIM=613199, no_hdr0=''),
  402562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402562, CurrentID=0, Status='live', Symbol='HNRNPA1P8', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 8', other_designations='hCG2023776', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84983495, end_position_on_the_genomic_accession=84984839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285547, CurrentID=0, Status='live', Symbol='LINC01097', Aliases='', description='long intergenic non-protein coding RNA 1097', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13526319, end_position_on_the_genomic_accession=13531417, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  9348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9348, CurrentID=0, Status='live', Symbol='NDST3', Aliases='HSST3', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 3', other_designations='GlcNAc N-deacetylase/ N-sulfotransferase 3|N-HSST 3|N-heparan sulfate sulfotransferase 3|NDST-3|glucosaminyl N-deacetylase/N-sulfotransferase 3|hNDST-3', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118033826, end_position_on_the_genomic_accession=118258634, orientation='plus', exon_count=17, OMIM=603950, no_hdr0=''),
  9349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9349, CurrentID=0, Status='live', Symbol='RPL23', Aliases='L23, rpL17', description='ribosomal protein L23', other_designations='60S ribosomal protein L17|ribosomal protein L17', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38850068, end_position_on_the_genomic_accession=38853800, orientation='minus', exon_count=5, OMIM=603662, no_hdr0=''),
  9350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9350, CurrentID=0, Status='live', Symbol='CER1', Aliases='DAND4', description='cerberus 1, DAN family BMP antagonist', other_designations='DAN domain family member 4|cerberus 1, cysteine knot superfamily, homolog|cerberus-related 1|cerberus-related protein', map_location='9p23-p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14719733, end_position_on_the_genomic_accession=14722717, orientation='minus', exon_count=2, OMIM=603777, no_hdr0=''),
  9351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9351, CurrentID=0, Status='live', Symbol='SLC9A3R2', Aliases='E3KARP, NHE3RF2, NHERF-2, NHERF2, OCTS2, SIP-1, SIP1, TKA-1', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 regulator 2', other_designations='NHE3 kinase A regulatory protein E3KARP|NHE3 regulatory factor 2|SRY-interacting protein 1|sodium/hydrogen exchanger 3 kinase A regulatory protein|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 regulator 2|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 regulatory factor 2|solute carrier family 9 (sodium/hydrogen exchanger), member 3 regulator 2|solute carrier family 9 isoform A3 regulatory factor 2|tyrosine kinase activator protein 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2026868, end_position_on_the_genomic_accession=2039026, orientation='plus', exon_count=10, OMIM=606553, no_hdr0=''),
  9352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9352, CurrentID=0, Status='live', Symbol='TXNL1', Aliases='HEL-S-114, TRP32, TXL-1, TXNL, Txl', description='thioredoxin like 1', other_designations='32 kDa thioredoxin-related protein|epididymis secretory protein Li 114|thioredoxin-like 1|thioredoxin-related 32 kDa protein|thioredoxin-related protein 1', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56602387, end_position_on_the_genomic_accession=56638689, orientation='minus', exon_count=11, OMIM=603049, no_hdr0=''),
  9353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9353, CurrentID=0, Status='live', Symbol='SLIT2', Aliases='SLIL3, Slit-2', description='slit guidance ligand 2', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20251925, end_position_on_the_genomic_accession=20620561, orientation='plus', exon_count=40, OMIM=603746, no_hdr0=''),
  9354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9354, CurrentID=0, Status='live', Symbol='UBE4A', Aliases='E4, UBOX2, UFD2', description='ubiquitination factor E4A', other_designations='ubiquitination factor E4A (UFD2 homolog, yeast)', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118359581, end_position_on_the_genomic_accession=118399211, orientation='plus', exon_count=20, OMIM=603753, no_hdr0=''),
  9355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9355, CurrentID=0, Status='live', Symbol='LHX2', Aliases='LH2, hLhx2', description='LIM homeobox 2', other_designations='LIM HOX gene 2|LIM homeobox protein 2|homeobox protein LH-2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124011610, end_position_on_the_genomic_accession=124033163, orientation='plus', exon_count=7, OMIM=603759, no_hdr0=''),
  9356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9356, CurrentID=0, Status='live', Symbol='SLC22A6', Aliases='HOAT1, OAT1, PAHT, ROAT1', description='solute carrier family 22 (organic anion transporter), member 6', other_designations='PAH transporter|hPAHT|hROAT1|organic anion transporter 1|para-aminohippurate transporter|renal organic anion transporter 1', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62976597, end_position_on_the_genomic_accession=62985023, orientation='minus', exon_count=10, OMIM=607582, no_hdr0=''),
  402573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402573, CurrentID=0, Status='live', Symbol='C7orf61', Aliases='', description='chromosome 7 open reading frame 61', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100456615, end_position_on_the_genomic_accession=100464271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  9358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9358, CurrentID=0, Status='live', Symbol='ITGBL1', Aliases='OSCP, TIED', description='integrin subunit beta like 1', other_designations='integrin beta like 1|integrin, beta-like 1 (with EGF-like repeat domains)|osteoblast-specific cysteine-rich protein|ten integrin EGF-like repeat domain-containing protein', map_location='13q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101452593, end_position_on_the_genomic_accession=101720857, orientation='plus', exon_count=13, OMIM=604234, no_hdr0=''),
  9360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9360, CurrentID=0, Status='live', Symbol='PPIG', Aliases='CARS-Cyp, CYP, SCAF10, SRCyp', description='peptidylprolyl isomerase G (cyclophilin G)', other_designations='CARS-cyclophilin|CASP10|Clk-associating RS-cyclophilin|PPIase G|SR-cyclophilin|SR-cyp|SR-related CTD-associated factor 10|cyclophilin G|peptidyl-prolyl isomerase G (cyclophilin G)|rotamase G', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169584340, end_position_on_the_genomic_accession=169641406, orientation='plus', exon_count=16, OMIM=606093, no_hdr0=''),
  9361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9361, CurrentID=0, Status='live', Symbol='LONP1', Aliases='CODASS, LON, LONP, LonHS, PIM1, PRSS15, hLON', description='lon peptidase 1, mitochondrial', other_designations='LONP|hLON ATP-dependent protease|lon protease-like protein|mitochondrial ATP-dependent protease Lon|mitochondrial lon protease-like protein|serine protease 15', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5691834, end_position_on_the_genomic_accession=5720452, orientation='minus', exon_count=21, OMIM=605490, no_hdr0=''),
  9362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9362, CurrentID=0, Status='live', Symbol='CPNE6', Aliases='', description='copine VI (neuronal)', other_designations='N-copine', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24070837, end_position_on_the_genomic_accession=24078100, orientation='plus', exon_count=18, OMIM=605688, no_hdr0=''),
  9363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9363, CurrentID=0, Status='live', Symbol='RAB33A', Aliases='RabS10', description='RAB33A, member RAS oncogene family', other_designations='Small GTP-binding protein S10', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130171799, end_position_on_the_genomic_accession=130184870, orientation='plus', exon_count=2, OMIM=300333, no_hdr0=''),
  9364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9364, CurrentID=0, Status='live', Symbol='RAB28', Aliases='CORD18', description='RAB28, member RAS oncogene family', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13296017, end_position_on_the_genomic_accession=13484365, orientation='minus', exon_count=17, OMIM=612994, no_hdr0=''),
  9365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9365, CurrentID=0, Status='live', Symbol='KL', Aliases='', description='klotho', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33016063, end_position_on_the_genomic_accession=33066145, orientation='plus', exon_count=6, OMIM=604824, no_hdr0=''),
  9366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9366, CurrentID=0, Status='live', Symbol='RAB9BP1', Aliases='RAB9P1', description='RAB9B, member RAS oncogene family pseudogene 1', other_designations='RAB9, member RAS oncogene family, pseudogene 1|RAB9A, member RAS oncogene family pseudogene', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=105099474, end_position_on_the_genomic_accession=105100098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  9367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9367, CurrentID=0, Status='live', Symbol='RAB9A', Aliases='RAB9', description='RAB9A, member RAS oncogene family', other_designations='RAB9, member RAS oncogene family', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13689121, end_position_on_the_genomic_accession=13709825, orientation='plus', exon_count=3, OMIM=300284, no_hdr0=''),
  9368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9368, CurrentID=0, Status='live', Symbol='SLC9A3R1', Aliases='EBP50, NHERF, NHERF-1, NHERF1, NPHLOP2', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 regulator 1', other_designations='Na+/H+ exchange regulatory co-factor|ezrin-radixin-moesin binding phosphoprotein-50|ezrin-radixin-moesin-binding phosphoprotein 50|regulatory cofactor of Na(+)/H(+) exchanger|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 regulatory factor 1|solute carrier family 9 isoform A3 regulatory factor 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74748612, end_position_on_the_genomic_accession=74769360, orientation='plus', exon_count=6, OMIM=604990, no_hdr0=''),
  9369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9369, CurrentID=0, Status='live', Symbol='NRXN3', Aliases='C14orf60', description='neurexin 3', other_designations='neurexin III|neurexin-3-alpha', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78021758, end_position_on_the_genomic_accession=79868290, orientation='plus', exon_count=31, OMIM=600567, no_hdr0=''),
  9370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9370, CurrentID=0, Status='live', Symbol='ADIPOQ', Aliases='ACDC, ACRP30, ADIPQTL1, ADPN, APM-1, APM1, GBP28', description='adiponectin, C1Q and collagen domain containing', other_designations='30 kDa adipocyte complement-related protein|adipocyte complement-related 30 kDa protein|adipose most abundant gene transcript 1 protein|adipose specific collagen-like factor|gelatin-binding protein 28', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186842668, end_position_on_the_genomic_accession=186858463, orientation='plus', exon_count=5, OMIM=605441, no_hdr0=''),
  9371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9371, CurrentID=0, Status='live', Symbol='KIF3B', Aliases='FLA8, HH0048, KLP-11', description='kinesin family member 3B', other_designations='microtubule plus end-directed kinesin motor 3B', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32277651, end_position_on_the_genomic_accession=32335011, orientation='plus', exon_count=9, OMIM=603754, no_hdr0=''),
  9372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9372, CurrentID=0, Status='live', Symbol='ZFYVE9', Aliases='MADHIP, NSP, PPP1R173, SARA, SMADIP', description='zinc finger, FYVE domain containing 9', other_designations='MAD, mothers against decapentaplegic homolog interacting protein, receptor activation anchor|MADH-interacting protein|novel serine protease|protein phosphatase 1, regulatory subunit 173|smad anchor for receptor activation', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52142094, end_position_on_the_genomic_accession=52346686, orientation='plus', exon_count=20, OMIM=603755, no_hdr0=''),
  9373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9373, CurrentID=0, Status='live', Symbol='PLAA', Aliases='DOA1, PLA2P, PLAP', description='phospholipase A2-activating protein', other_designations='DOA1 homolog|phospholipase A2 activating protein', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26903370, end_position_on_the_genomic_accession=26947470, orientation='minus', exon_count=15, OMIM=603873, no_hdr0=''),
  9374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9374, CurrentID=0, Status='live', Symbol='PPT2', Aliases='C6orf8, G14, PPT-2', description='palmitoyl-protein thioesterase 2', other_designations='S-thioesterase G14|VE-statin2|palmitoyl-protein hydrolase 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32153452, end_position_on_the_genomic_accession=32163681, orientation='plus', exon_count=10, OMIM=603298, no_hdr0=''),
  9375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9375, CurrentID=0, Status='live', Symbol='TM9SF2', Aliases='P76', description='transmembrane 9 superfamily member 2', other_designations='76 kDa membrane protein|dinucleotide oxidase disulfide thiol exchanger 3 superfamily member 2|transmembrane protein 9 superfamily member 2', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99501374, end_position_on_the_genomic_accession=99564048, orientation='plus', exon_count=17, OMIM=604678, no_hdr0=''),
  9376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9376, CurrentID=0, Status='live', Symbol='SLC22A8', Aliases='OAT3', description='solute carrier family 22 (organic anion transporter), member 8', other_designations='hOAT3|organic anion transporter 3', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62992824, end_position_on_the_genomic_accession=63015845, orientation='minus', exon_count=12, OMIM=607581, no_hdr0=''),
  9377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9377, CurrentID=0, Status='live', Symbol='COX5A', Aliases='COX, COX-VA, VA', description='cytochrome c oxidase subunit Va', other_designations='cytochrome c oxidase polypeptide Va|cytochrome c oxidase polypeptide, mitochondrial|mitochondrial cytochrome c oxidase subunit Va', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74920275, end_position_on_the_genomic_accession=74938154, orientation='minus', exon_count=5, OMIM=603773, no_hdr0=''),
  9378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9378, CurrentID=0, Status='live', Symbol='NRXN1', Aliases='Hs.22998, PTHSL2, SCZD17', description='neurexin 1', other_designations='neurexin I', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=49918505, end_position_on_the_genomic_accession=51032536, orientation='minus', exon_count=30, OMIM=600565, no_hdr0=''),
  140451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140451, CurrentID=0, Status='live', Symbol='SCA9', Aliases='', description='spinocerebellar ataxia 9', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612876, no_hdr0=''),
  9380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9380, CurrentID=0, Status='live', Symbol='GRHPR', Aliases='GLXR, GLYD, PH2', description='glyoxylate reductase/hydroxypyruvate reductase', other_designations='glycerate-2-dehydrogenase', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37422666, end_position_on_the_genomic_accession=37436993, orientation='plus', exon_count=9, OMIM=604296, no_hdr0=''),
  9381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9381, CurrentID=0, Status='live', Symbol='OTOF', Aliases='AUNB1, DFNB6, DFNB9, FER1L2, NSRD9', description='otoferlin', other_designations='fer-1-like family member 2|fer-1-like protein 2', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26457203, end_position_on_the_genomic_accession=26558698, orientation='minus', exon_count=49, OMIM=603681, no_hdr0=''),
  9382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9382, CurrentID=0, Status='live', Symbol='COG1', Aliases='CDG2G, LDLB', description='component of oligomeric golgi complex 1', other_designations='COG complex subunit 1|conserved oligomeric Golgi complex protein 1|low density lipoprotein receptor defect B complementing', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73193034, end_position_on_the_genomic_accession=73208507, orientation='plus', exon_count=14, OMIM=606973, no_hdr0=''),
  9383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9383, CurrentID=0, Status='live', Symbol='TSIX', Aliases='LINC00013, NCRNA00013, XIST-AS, XIST-AS1, XISTAS', description='TSIX transcript, XIST antisense RNA', other_designations='TSIX transcript, XIST antisense RNA (non-protein coding)|X (inactive)-specific transcript, antisense (non-protein coding)|XIST antisense RNA (non-protein coding)|long intergenic non-protein coding RNA 13|x-inactivation-specific transcript-antisense', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73792205, end_position_on_the_genomic_accession=73829231, orientation='plus', exon_count=1, OMIM=300181, no_hdr0=''),
  140456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140456, CurrentID=0, Status='live', Symbol='ASB11', Aliases='', description='ankyrin repeat and SOCS box containing 11', other_designations='ASB-11|ankyrin repeat and SOCS box containing 11, E3 ubiquitin protein ligase|ankyrin repeat domain-containing SOCS box protein ASB11', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15275644, end_position_on_the_genomic_accession=15315667, orientation='minus', exon_count=10, OMIM=300626, no_hdr0=''),
  9385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9385, CurrentID=0, Status='live', Symbol='COX5AP1', Aliases='', description='cytochrome c oxidase subunit Va pseudogene 1', other_designations='', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55240439, end_position_on_the_genomic_accession=55241157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140458, CurrentID=0, Status='live', Symbol='ASB5', Aliases='ASB-5', description='ankyrin repeat and SOCS box containing 5', other_designations='SOCS box protein ASB-5|ankyrin repeat and SOCS box-containing 5', map_location='4q34.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176213673, end_position_on_the_genomic_accession=176277571, orientation='minus', exon_count=11, OMIM=615050, no_hdr0=''),
  140459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140459, CurrentID=0, Status='live', Symbol='ASB6', Aliases='', description='ankyrin repeat and SOCS box containing 6', other_designations='ankyrin repeat and SOCS box-containing 6', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129634604, end_position_on_the_genomic_accession=129642169, orientation='minus', exon_count=6, OMIM=615051, no_hdr0=''),
  9388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9388, CurrentID=0, Status='live', Symbol='LIPG', Aliases='EDL, EL, PRO719', description='lipase, endothelial', other_designations='endothelial cell-derived lipase|lipoprotein lipase H', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49560699, end_position_on_the_genomic_accession=49592908, orientation='plus', exon_count=11, OMIM=603684, no_hdr0=''),
  140461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140461, CurrentID=0, Status='live', Symbol='ASB8', Aliases='PP14212', description='ankyrin repeat and SOCS box containing 8', other_designations='ASB-8|ankyrin repeat and SOCS box-containing 8', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48147788, end_position_on_the_genomic_accession=48157594, orientation='minus', exon_count=5, OMIM=615053, no_hdr0=''),
  140462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140462, CurrentID=0, Status='live', Symbol='ASB9', Aliases='', description='ankyrin repeat and SOCS box containing 9', other_designations='ASB-9|ankyrin repeat and SOCS box-containing 9|ankyrin repeat and suppressor of cytokine signaling box protein 9', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15243987, end_position_on_the_genomic_accession=15270467, orientation='minus', exon_count=8, OMIM=300890, no_hdr0=''),
  9391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9391, CurrentID=0, Status='live', Symbol='CIAO1', Aliases='CIA1, WDR39', description='cytosolic iron-sulfur assembly component 1', other_designations='WD repeat domain 39|WD repeat-containing protein 39|WD40 protein Ciao1|cytosolic iron-sulfur protein assembly 1 homolog', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96266146, end_position_on_the_genomic_accession=96274179, orientation='plus', exon_count=7, OMIM=604333, no_hdr0=''),
  9392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9392, CurrentID=0, Status='live', Symbol='TGFBRAP1', Aliases='TRAP-1, TRAP1', description='transforming growth factor beta receptor associated protein 1', other_designations='TGF beta receptor associated protein -1|TGF-beta receptor-associated protein 1|transforming growth factor, beta receptor associated protein 1', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105249404, end_position_on_the_genomic_accession=105329714, orientation='minus', exon_count=13, OMIM=606237, no_hdr0=''),
  140465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140465, CurrentID=0, Status='live', Symbol='MYL6B', Aliases='MLC1SA', description='myosin light chain 6B', other_designations='myosin alkali light chain 1 slow a|myosin light chain 1 slow a|myosin light chain 1 slow-twitch muscle A isoform|myosin light chain 1, slow-twitch muscle A isoform|myosin, light chain 6B, alkali, smooth muscle and non-muscle|myosin, light polypeptide 6B, alkali, smooth muscle and non-muscle|smooth muscle and non-muscle myosin alkali light chain 6B|smooth muscle and nonmuscle myosin light chain alkali 6B', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56152420, end_position_on_the_genomic_accession=56157987, orientation='plus', exon_count=8, OMIM=609930, no_hdr0=''),
  9394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9394, CurrentID=0, Status='live', Symbol='HS6ST1', Aliases='HH15, HS6ST', description='heparan sulfate 6-O-sulfotransferase 1', other_designations='HS6ST-1|heparan-sulfate 6-sulfotransferase', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128265480, end_position_on_the_genomic_accession=128318597, orientation='minus', exon_count=2, OMIM=604846, no_hdr0=''),
  140467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140467, CurrentID=0, Status='live', Symbol='ZNF358', Aliases='ZFEND', description='zinc finger protein 358', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7515282, end_position_on_the_genomic_accession=7521025, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140468, CurrentID=0, Status='live', Symbol='COX11P1', Aliases='COX11, COX11P', description='COX11 cytochrome c oxidase copper chaperone pseudogene 1', other_designations='COX11 cytochrome c oxidase assembly homolog pseudogene 1|COX11 homolog, cytochrome c oxidase assembly protein pseudogene 1|COX11 homolog, cytochrome c oxidase assembly protein, pseudogene|COX11 pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28446920, end_position_on_the_genomic_accession=28447630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9397, CurrentID=0, Status='live', Symbol='NMT2', Aliases='', description='N-myristoyltransferase 2', other_designations='NMT 2|glycylpeptide N-tetradecanoyltransferase 2 variant 3|glycylpeptide N-tetradecanoyltransferase 2 variant 4|myristoyl-CoA:protein N-myristoyltransferase 2|peptide N-myristoyltransferase 2|type II N-myristoyltransferase', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15105700, end_position_on_the_genomic_accession=15168705, orientation='minus', exon_count=14, OMIM=603801, no_hdr0=''),
  9398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9398, CurrentID=0, Status='live', Symbol='CD101', Aliases='EWI-101, IGSF2, V7', description='CD101 molecule', other_designations='cell surface glycoprotein V7|glu-Trp-Ile EWI motif-containing protein 101|leukocyte surface protein', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117001750, end_position_on_the_genomic_accession=117036551, orientation='plus', exon_count=11, OMIM=604516, no_hdr0=''),
  9399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9399, CurrentID=0, Status='live', Symbol='STOML1', Aliases='SLP-1, STORP, hUNC-24', description='stomatin like 1', other_designations='EPB72-like protein 1|protein unc-24 homolog|stomatin (EBP72)-like 1|stomatin (EPB72)-like 1|stomatin-related protein', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73983218, end_position_on_the_genomic_accession=73994622, orientation='minus', exon_count=9, OMIM=608326, no_hdr0=''),
  9400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9400, CurrentID=0, Status='live', Symbol='RECQL5', Aliases='RECQ5', description='RecQ helicase-like 5', other_designations='DNA helicase, RecQ-like type 5|RecQ protein-like 5', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75626845, end_position_on_the_genomic_accession=75667191, orientation='minus', exon_count=26, OMIM=603781, no_hdr0=''),
  9401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9401, CurrentID=0, Status='live', Symbol='RECQL4', Aliases='RECQ4', description='RecQ helicase-like 4', other_designations='DNA helicase, RecQ-like type 4|DNA helicase, RecQ-like, type 4|RecQ protein-like 4', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144511284, end_position_on_the_genomic_accession=144517827, orientation='minus', exon_count=21, OMIM=603780, no_hdr0=''),
  9402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9402, CurrentID=0, Status='live', Symbol='GRAP2', Aliases='GADS, GRAP-2, GRB2L, GRBLG, GRID, GRPL, GrbX, Grf40, Mona, P38', description='GRB2-related adaptor protein 2', other_designations='GRB-2-like protein|GRB2-related protein with insert domain|SH3-SH2-SH3 adapter Mona|SH3-SH2-SH3 adaptor molecule|adapter protein GRID|grf-40|grf40 adapter protein|growth factor receptor-binding protein|growth factor receptor-bound protein 2-related adaptor protein 2|hematopoietic cell-associated adapter protein GrpL|hematopoietic cell-associated adaptor protein GRPL', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39893808, end_position_on_the_genomic_accession=39973343, orientation='plus', exon_count=11, OMIM=604518, no_hdr0=''),
  9403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9403, CurrentID=0, Status='live', Symbol='SEP15', Aliases='', description='15 kDa selenoprotein', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86862445, end_position_on_the_genomic_accession=86914424, orientation='minus', exon_count=5, OMIM=606254, no_hdr0=''),
  9404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9404, CurrentID=0, Status='live', Symbol='LPXN', Aliases='LDPL', description='leupaxin', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58526871, end_position_on_the_genomic_accession=58578892, orientation='minus', exon_count=13, OMIM=605390, no_hdr0=''),
  9406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9406, CurrentID=0, Status='live', Symbol='ZRANB2', Aliases='ZIS, ZIS1, ZIS2, ZNF265', description='zinc finger, RAN-binding domain containing 2', other_designations='zinc finger protein 265|zinc finger, splicing|zinc-finger protein 265|zinc-finger, splicing', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71063291, end_position_on_the_genomic_accession=71081289, orientation='minus', exon_count=11, OMIM=604347, no_hdr0=''),
  9407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9407, CurrentID=0, Status='live', Symbol='TMPRSS11D', Aliases='HAT', description='transmembrane protease, serine 11D', other_designations='airway trypsin like protease|airway trypsin-like protease', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67820876, end_position_on_the_genomic_accession=67883998, orientation='minus', exon_count=10, OMIM=605369, no_hdr0=''),
  9409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9409, CurrentID=0, Status='live', Symbol='PEX16', Aliases='PBD8A, PBD8B', description='peroxisomal biogenesis factor 16', other_designations='peroxin 16|peroxisomal membrane protein PEX16|peroxisome biogenesis factor 16', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45909669, end_position_on_the_genomic_accession=45918123, orientation='minus', exon_count=12, OMIM=603360, no_hdr0=''),
  9410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9410, CurrentID=0, Status='live', Symbol='SNRNP40', Aliases='40K, HPRP8BP, PRP8BP, PRPF8BP, SPF38, WDR57', description='small nuclear ribonucleoprotein, U5 40kDa subunit', other_designations='38 kDa-splicing factor|Prp8-binding protein|U5 snRNP 40 kDa protein|U5 snRNP-specific 40 kDa protein (hPrp8-binding)|U5-40K|U5-40kD protein|WD repeat domain 57 (U5 snRNP specific)|WD repeat-containing protein 57|small nuclear ribonucleoprotein 40kDa (U5)', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31259568, end_position_on_the_genomic_accession=31296797, orientation='minus', exon_count=10, OMIM=607797, no_hdr0=''),
  9411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9411, CurrentID=0, Status='live', Symbol='ARHGAP29', Aliases='PARG1', description='Rho GTPase activating protein 29', other_designations='PTPL1-associated RhoGAP 1 (PARG1)|PTPL1-associated RhoGAP protein 1|rho-type GTPase-activating protein 29', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94168907, end_position_on_the_genomic_accession=94275068, orientation='minus', exon_count=25, OMIM=610496, no_hdr0=''),
  9412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9412, CurrentID=0, Status='live', Symbol='MED21', Aliases='SRB7, SURB7, hSrb7', description='mediator complex subunit 21', other_designations='RNA polymerase II holoenzyme component SRB7|RNAPII complex component SRB7|SRB7 suppressor of RNA polymerase B homolog', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27022522, end_position_on_the_genomic_accession=27030673, orientation='plus', exon_count=4, OMIM=603800, no_hdr0=''),
  9413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9413, CurrentID=0, Status='live', Symbol='FAM189A2', Aliases='C9orf61, X123', description='family with sequence similarity 189 member A2', other_designations='Friedreich ataxia region gene X123|family with sequence similarity 189, member A2', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69324572, end_position_on_the_genomic_accession=69392456, orientation='plus', exon_count=13, OMIM=607710, no_hdr0=''),
  9414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9414, CurrentID=0, Status='live', Symbol='TJP2', Aliases='C9DUPq21.11, DFNA51, DUP9q21.11, PFIC4, X104, ZO2', description='tight junction protein 2', other_designations='Friedreich ataxia region gene X104 (tight junction protein ZO-2)|zona occludens 2|zonula occludens protein 2', map_location='9q13-q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69121264, end_position_on_the_genomic_accession=69255208, orientation='plus', exon_count=31, OMIM=607709, no_hdr0=''),
  9415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9415, CurrentID=0, Status='live', Symbol='FADS2', Aliases='D6D, DES6, FADSD6, LLCDL2, SLL0262, TU13', description='fatty acid desaturase 2', other_designations='acyl-CoA 6-desaturase|delta(6) desaturase|delta(6) fatty acid desaturase|delta-6 desaturase|delta-6 fatty acid desaturase|delta-6-desaturase|linoleoyl-CoA desaturase (delta-6-desaturase)-like 2', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61816203, end_position_on_the_genomic_accession=61867354, orientation='plus', exon_count=14, OMIM=606149, no_hdr0=''),
  9416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9416, CurrentID=0, Status='live', Symbol='DDX23', Aliases='PRPF28, SNRNP100, U5-100K, U5-100KD, prp28', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 23', other_designations='100 kDa U5 snRNP-specific protein|DEAD box protein 23|PRP28 homolog, yeast|PRP28p homolog|U5 snRNP 100 kD protein', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48829756, end_position_on_the_genomic_accession=48852174, orientation='minus', exon_count=17, OMIM=612172, no_hdr0=''),
  402634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402634, CurrentID=0, Status='live', Symbol='LOC402634', Aliases='', description='neugrin, neurite outgrowth associated pseudogene', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2393094, end_position_on_the_genomic_accession=2394135, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9419, CurrentID=0, Status='live', Symbol='CRIPT', Aliases='HSPC139, SSMDF', description='cysteine-rich PDZ-binding protein', other_designations='cysteine-rich interactor of PDZ three|cysteine-rich interactor of PDZ3|postsynaptic protein CRIPT', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46617172, end_position_on_the_genomic_accession=46630176, orientation='plus', exon_count=5, OMIM=604594, no_hdr0=''),
  9420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9420, CurrentID=0, Status='live', Symbol='CYP7B1', Aliases='CBAS3, CP7B, SPG5A', description='cytochrome P450 family 7 subfamily B member 1', other_designations='cytochrome P450 7B1|cytochrome P450, family 7, subfamily B, polypeptide 1|cytochrome P450, subfamily VIIB (oxysterol 7 alpha-hydroxylase), polypeptide 1|oxysterol 7-alpha-hydroxylase|oxysterol 7alpha-hydroxylase', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64595972, end_position_on_the_genomic_accession=64798791, orientation='minus', exon_count=6, OMIM=603711, no_hdr0=''),
  9421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9421, CurrentID=0, Status='live', Symbol='HAND1', Aliases='Hxt, Thing1, bHLHa27, eHand', description='heart and neural crest derivatives expressed 1', other_designations='class A basic helix-loop-helix protein 27|extraembryonic tissues, heart, autonomic nervous system and neural crest derivatives-expressed protein 1', map_location='5q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154474972, end_position_on_the_genomic_accession=154478264, orientation='minus', exon_count=2, OMIM=602406, no_hdr0=''),
  9422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9422, CurrentID=0, Status='live', Symbol='ZNF264', Aliases='', description='zinc finger protein 264', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57191479, end_position_on_the_genomic_accession=57222846, orientation='plus', exon_count=8, OMIM=604668, no_hdr0=''),
  9423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9423, CurrentID=0, Status='live', Symbol='NTN1', Aliases='NTN1L', description='netrin 1', other_designations='epididymis tissue protein Li 131P|netrin 1, mouse, homolog of', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9021542, end_position_on_the_genomic_accession=9244000, orientation='plus', exon_count=8, OMIM=601614, no_hdr0=''),
  9424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9424, CurrentID=0, Status='live', Symbol='KCNK6', Aliases='K2p6.1, KCNK8, TOSS, TWIK-2, TWIK2', description='potassium channel, two pore domain subfamily K, member 6', other_designations='K2P6.1 potassium channel|TWIK-originated similarity sequence|TWIK-originated sodium similarity sequence|inward rectifying potassium channel protein TWIK-2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38319814, end_position_on_the_genomic_accession=38329014, orientation='plus', exon_count=5, OMIM=603939, no_hdr0=''),
  9425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9425, CurrentID=0, Status='live', Symbol='CDYL', Aliases='CDYL1', description='chromodomain protein, Y-like', other_designations='CDY-like, autosomal|testis-specific chromodomain Y-like protein', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4706159, end_position_on_the_genomic_accession=4955544, orientation='plus', exon_count=12, OMIM=603778, no_hdr0=''),
  9426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9426, CurrentID=0, Status='live', Symbol='CDY2A', Aliases='CDY, CDY2', description='chromodomain protein, Y-linked, 2A', other_designations='Y chromosome chromodomain protein 2A|chromodomain protein, Y chromosome, 2', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18025787, end_position_on_the_genomic_accession=18027746, orientation='plus', exon_count=1, OMIM=400018, no_hdr0=''),
  9427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9427, CurrentID=0, Status='live', Symbol='ECEL1', Aliases='DA5D, DINE, ECEX, XCE', description='endothelin converting enzyme-like 1', other_designations='X converting enzyme|damage induced neuronal endopeptidase', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232479827, end_position_on_the_genomic_accession=232487859, orientation='minus', exon_count=18, OMIM=605896, no_hdr0=''),
  402644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402644, CurrentID=0, Status='live', Symbol='LOC402644', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28279010, end_position_on_the_genomic_accession=28279742, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9429, CurrentID=0, Status='live', Symbol='ABCG2', Aliases='ABC15, ABCP, BCRP, BCRP1, BMDP, CD338, CDw338, EST157481, GOUT1, MRX, MXR, MXR-1, MXR1, UAQTL1', description='ATP binding cassette subfamily G member 2 (Junior blood group)', other_designations='ABC transporter|ATP-binding cassette transporter G2|ATP-binding cassette, sub-family G (WHITE), member 2 (Junior blood group)|breast cancer resistance protein|mitoxantrone resistance-associated protein|multi drug resistance efflux transport ATP-binding cassette sub-family G (WHITE) member 2|placenta specific MDR protein|placenta-specific ATP-binding cassette transporter|urate exporter', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88090264, end_position_on_the_genomic_accession=88231417, orientation='minus', exon_count=19, OMIM=603756, no_hdr0=''),
  9430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9430, CurrentID=0, Status='live', Symbol='MYLKP1', Aliases='MYLKP', description='myosin light chain kinase pseudogene 1 (functional)', other_designations='myosin, light polypeptide kinase pseudogene', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75328548, end_position_on_the_genomic_accession=75339072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9431, CurrentID=0, Status='live', Symbol='OR7E22P', Aliases='OR3.6, OR6DG', description='olfactory receptor family 7 subfamily E member 22 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 22 pseudogene', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75356485, end_position_on_the_genomic_accession=75357510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9432, CurrentID=0, Status='live', Symbol='OR7E21P', Aliases='OR4DG, OR7E127P, OR7E141P, OR7E142P, OR7E49P, OST035', description='olfactory receptor family 7 subfamily E member 21 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 127 pseudogene|olfactory receptor, family 7, subfamily E, member 141 pseudogene|olfactory receptor, family 7, subfamily E, member 142 pseudogene|olfactory receptor, family 7, subfamily E, member 21 pseudogene|olfactory receptor, family 7, subfamily E, member 49 pseudogene', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130034488, end_position_on_the_genomic_accession=130035541, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271652, CurrentID=0, Status='live', Symbol='RPS15AP31', Aliases='RPS15A_18_1130', description='ribosomal protein S15a pseudogene 31', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28287196, end_position_on_the_genomic_accession=28287643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9435, CurrentID=0, Status='live', Symbol='CHST2', Aliases='C6ST, GST-2, GST2, Gn6ST-1, HEL-S-75, glcNAc6ST-1', description='carbohydrate (N-acetylglucosamine-6-O) sulfotransferase 2', other_designations='N-acetylglucosamine 6-O-sulfotransferase 1|carbohydrate (chondroitin 6/keratan) sulfotransferase 2|epididymis secretory protein Li 75|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 2', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143119776, end_position_on_the_genomic_accession=143124014, orientation='plus', exon_count=2, OMIM=603798, no_hdr0=''),
  9436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9436, CurrentID=0, Status='live', Symbol='NCR2', Aliases='CD336, LY95, NK-p44, NKP44, dJ149M18.1', description='natural cytotoxicity triggering receptor 2', other_designations='NK cell activating receptor (NKp44)|NK cell-activating receptor|lymphocyte antigen 95 (activating NK-receptor; NK-p44)|lymphocyte antigen 95 homolog (activating NK-receptor; NK-p44)|natural killer cell p44-related protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41335790, end_position_on_the_genomic_accession=41350887, orientation='plus', exon_count=6, OMIM=604531, no_hdr0=''),
  9437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9437, CurrentID=0, Status='live', Symbol='NCR1', Aliases='CD335, LY94, NK-p46, NKP46', description='natural cytotoxicity triggering receptor 1', other_designations='NK cell-activating receptor|lymphocyte antigen 94 homolog (activating NK-receptor; NK-p46)|natural killer cell p46-related protein', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54905992, end_position_on_the_genomic_accession=54938208, orientation='plus', exon_count=9, OMIM=604530, no_hdr0=''),
  9439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9439, CurrentID=0, Status='live', Symbol='MED23', Aliases='ARC130, CRSP130, CRSP133, CRSP3, DRIP130, MRT18, SUR-2, SUR2', description='mediator complex subunit 23', other_designations='130 kDa transcriptional co-activator|133 kDa transcriptional co-activator|activator-recruited cofactor 130 kDa component|cofactor required for Sp1 transcriptional activation subunit 3|mediator complex subunit MED23 variant MED23_i6|mediator complex subunit MED23 variant MED23_i7|vitamin D3 receptor interacting protein', map_location='6q22.33-q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131573966, end_position_on_the_genomic_accession=131628239, orientation='minus', exon_count=33, OMIM=605042, no_hdr0=''),
  9440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9440, CurrentID=0, Status='live', Symbol='MED17', Aliases='CRSP6, CRSP77, DRIP80, TRAP80', description='mediator complex subunit 17', other_designations='ARC77|CRSP complex subunit 6|activator-recruited cofactor 77 kDa component|cofactor required for Sp1 transcriptional activation, subunit 6, 77kDa|thyroid hormone receptor-associated protein complex 80 kDa component|transcriptional coactivator CRSP77|vitamin D3 receptor-interacting protein complex 80 kDa component', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93784239, end_position_on_the_genomic_accession=93813330, orientation='plus', exon_count=12, OMIM=603810, no_hdr0=''),
  9441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9441, CurrentID=0, Status='live', Symbol='MED26', Aliases='CRSP7, CRSP70', description='mediator complex subunit 26', other_designations='ARC70|CRSP complex subunit 7|activator-recruited cofactor 70 kDa component|cofactor required for Sp1 transcriptional activation subunit 7|cofactor required for Sp1 transcriptional activation, subunit 7 (70kD)|cofactor required for Sp1 transcriptional activation, subunit 7, 70kDa|transcriptional coactivator CRSP70', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16574907, end_position_on_the_genomic_accession=16628204, orientation='minus', exon_count=3, OMIM=605043, no_hdr0=''),
  9442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9442, CurrentID=0, Status='live', Symbol='MED27', Aliases='CRAP34, CRSP34, CRSP8, TRAP37', description='mediator complex subunit 27', other_designations='CRSP complex subunit 8|cofactor required for Sp1 transcriptional activation, subunit 8, 34kDa|p37 TRAP/SMCC/PC2 subunit|transcriptional coactivator CRSP34', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131860110, end_position_on_the_genomic_accession=132079887, orientation='minus', exon_count=10, OMIM=605044, no_hdr0=''),
  9443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9443, CurrentID=0, Status='live', Symbol='MED7', Aliases='ARC34, CRSP33, CRSP9', description='mediator complex subunit 7', other_designations='CRSP complex subunit 9|RNA polymerase transcriptional regulation mediator subunit 7 homolog|activator-recruited cofactor 34 kDa component|cofactor required for Sp1 transcriptional activation subunit 9|cofactor required for Sp1 transcriptional activation, subunit 9 (33kD)|cofactor required for Sp1 transcriptional activation, subunit 9, 33kDa|transcriptional coactivator CRSP33', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157138440, end_position_on_the_genomic_accession=157142910, orientation='minus', exon_count=3, OMIM=605045, no_hdr0=''),
  9444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9444, CurrentID=0, Status='live', Symbol='QKI', Aliases='Hqk, QK, QK1, QK3, hqkI', description='QKI, KH domain containing, RNA binding', other_designations='RNA binding protein HQK|homolog of mouse quaking QKI (KH domain RNA binding protein)|quaking homolog, KH domain RNA binding', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163414643, end_position_on_the_genomic_accession=163578596, orientation='plus', exon_count=7, OMIM=609590, no_hdr0=''),
  9445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9445, CurrentID=0, Status='live', Symbol='ITM2B', Aliases='ABRI, BRI, BRI2, BRICD2B, E25B, E3-16, FBD, RDGCA, imBRI2', description='integral membrane protein 2B', other_designations='ABri/ADan amyloid peptide|BRICHOS domain containing 2B|immature BRI2|transmembrane protein BRI', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48233138, end_position_on_the_genomic_accession=48262096, orientation='plus', exon_count=6, OMIM=603904, no_hdr0=''),
  9446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9446, CurrentID=0, Status='live', Symbol='GSTO1', Aliases='GSTO 1-1, GSTTLp28, HEL-S-21, P28, SPG-R', description='glutathione S-transferase omega 1', other_designations='MMA(V) reductase|S-(Phenacyl)glutathione reductase|epididymis secretory protein Li 21|glutathione S-transferase omega 1-1|glutathione-S-transferase like|glutathione-dependent dehydroascorbate reductase|monomethylarsonic acid reductase', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104254194, end_position_on_the_genomic_accession=104267464, orientation='plus', exon_count=7, OMIM=605482, no_hdr0=''),
  9447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9447, CurrentID=0, Status='live', Symbol='AIM2', Aliases='PYHIN4', description='absent in melanoma 2', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159062479, end_position_on_the_genomic_accession=159076895, orientation='minus', exon_count=6, OMIM=604578, no_hdr0=''),
  9448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9448, CurrentID=0, Status='live', Symbol='MAP4K4', Aliases='FLH21957, HEL-S-31, HGK, MEKKK4, NIK', description='mitogen-activated protein kinase kinase kinase kinase 4', other_designations='HPK/GCK-like kinase HGK|MAPK/ERK kinase kinase kinase 4|MEK kinase kinase 4|Ste20 group protein kinase HGK|epididymis secretory protein Li 31|hepatocyte progenitor kinase-like/germinal center kinase-like kinase|nck-interacting kinase', map_location='2q11.2-q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101697703, end_position_on_the_genomic_accession=101894690, orientation='plus', exon_count=34, OMIM=604666, no_hdr0=''),
  402665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402665, CurrentID=0, Status='live', Symbol='IGLON5', Aliases='', description='IgLON family member 5', other_designations='Ig-like domain-containing protein ENSP00000270642|hCG1651476', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51311848, end_position_on_the_genomic_accession=51330848, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  9450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9450, CurrentID=0, Status='live', Symbol='LY86', Aliases='MD-1, MD1, MMD-1, dJ80N2.1', description='lymphocyte antigen 86', other_designations='MD-1, RP105-associated|ly-86|protein MD-1', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6588701, end_position_on_the_genomic_accession=6654983, orientation='plus', exon_count=5, OMIM=605241, no_hdr0=''),
  9451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9451, CurrentID=0, Status='live', Symbol='EIF2AK3', Aliases='PEK, PERK, WRS', description='eukaryotic translation initiation factor 2 alpha kinase 3', other_designations='PRKR-like endoplasmic reticulum kinase|pancreatic EIF2-alpha kinase', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88556741, end_position_on_the_genomic_accession=88627576, orientation='minus', exon_count=19, OMIM=604032, no_hdr0=''),
  9452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9452, CurrentID=0, Status='live', Symbol='ITM2A', Aliases='BRICD2A, E25A', description='integral membrane protein 2A', other_designations='BRICHOS domain containing 2A', map_location='Xq13.3-Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79360384, end_position_on_the_genomic_accession=79367552, orientation='minus', exon_count=6, OMIM=300222, no_hdr0=''),
  9453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9453, CurrentID=0, Status='live', Symbol='GGPS1', Aliases='GGPPS, GGPPS1', description='geranylgeranyl diphosphate synthase 1', other_designations='(2E,6E)-farnesyl diphosphate synthase|GGPP synthase|GGPPSase|dimethylallyltranstransferase|farnesyl diphosphate synthase|farnesyltranstransferase|geranyltranstransferase', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235328438, end_position_on_the_genomic_accession=235344532, orientation='plus', exon_count=5, OMIM=606982, no_hdr0=''),
  9454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9454, CurrentID=0, Status='live', Symbol='HOMER3', Aliases='HOMER-3, VESL3', description='homer scaffolding protein 3', other_designations='Homer, neuronal immediate early gene, 3|homer homolog 3', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18929201, end_position_on_the_genomic_accession=18941258, orientation='minus', exon_count=12, OMIM=604800, no_hdr0=''),
  9455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9455, CurrentID=0, Status='live', Symbol='HOMER2', Aliases='ACPD, CPD, HOMER-2, VESL-2', description='homer scaffolding protein 2', other_designations='cupidin|homer homolog 2|homer homolog 3|homer, neuronal immediate early gene, 2', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82836946, end_position_on_the_genomic_accession=82986161, orientation='minus', exon_count=14, OMIM=604799, no_hdr0=''),
  9456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9456, CurrentID=0, Status='live', Symbol='HOMER1', Aliases='HOMERA, HOMER1B, HOMER1C, SYN47, Ves-1, HOMER1', description='homer scaffolding protein 1', other_designations='homer homolog 1|homer, neuronal immediate early gene, 1|homer-1', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79373824, end_position_on_the_genomic_accession=79513836, orientation='minus', exon_count=9, OMIM=604798, no_hdr0=''),
  9457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9457, CurrentID=0, Status='live', Symbol='FHL5', Aliases='ACT, dJ393D12.2', description='four and a half LIM domains 5', other_designations='1700027G07Rik|FHL-5|LIM protein ACT|activator of CREM in testis|activator of cAMP-responsive element modulator (CREM) in testis|activator of cAMP-responsive element modulator in testis', map_location='6q16.1-q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96562548, end_position_on_the_genomic_accession=96618620, orientation='plus', exon_count=8, OMIM=605126, no_hdr0=''),
  9459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9459, CurrentID=0, Status='live', Symbol='ARHGEF6', Aliases='COOL2, Cool-2, MRX46, PIXA, alpha-PIX, alphaPIX', description='Rac/Cdc42 guanine nucleotide exchange factor 6', other_designations='PAK-interacting exchange factor alpha|PAK-interacting exchange factor, alpha|Rac/Cdc42 guanine exchange factor (GEF) 6|Rac/Cdc42 guanine nucleotide exchange factor (GEF) 6', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136665547, end_position_on_the_genomic_accession=136781656, orientation='minus', exon_count=25, OMIM=300267, no_hdr0=''),
  402676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402676, CurrentID=0, Status='live', Symbol='NDUFAF4P2', Aliases='', description='NADH:ubiquinone oxidoreductase complex assembly factor 4 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 4 pseudogene 2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93844789, end_position_on_the_genomic_accession=93845312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9462, CurrentID=0, Status='live', Symbol='RASAL2', Aliases='NGAP', description='RAS protein activator like 2', other_designations='RAS protein activator-like 2|Ras protein activator like 1', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178093729, end_position_on_the_genomic_accession=178479514, orientation='plus', exon_count=21, OMIM=606136, no_hdr0=''),
  9463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9463, CurrentID=0, Status='live', Symbol='PICK1', Aliases='PICK, PRKCABP', description='protein interacting with PRKCA 1', other_designations='protein interacting with C kinase 1|protein kinase C-alpha-binding protein', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38056420, end_position_on_the_genomic_accession=38075701, orientation='plus', exon_count=15, OMIM=605926, no_hdr0=''),
  9464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9464, CurrentID=0, Status='live', Symbol='HAND2', Aliases='DHAND2, Hed, Thing2, bHLHa26, dHand', description='heart and neural crest derivatives expressed 2', other_designations='basic helix-loop-helix transcription factor HAND2|class A basic helix-loop-helix protein 26|deciduum, heart, autonomic nervous system and neural crest derivatives-expressed protein 2', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173526501, end_position_on_the_genomic_accession=173530227, orientation='minus', exon_count=2, OMIM=602407, no_hdr0=''),
  9465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9465, CurrentID=0, Status='live', Symbol='AKAP7', Aliases='AKAP15, AKAP18', description='A-kinase anchoring protein 7', other_designations='A kinase (PRKA) anchor protein 7|A-kinase anchor protein 18 kDa|A-kinase anchor protein 7 isoforms alpha and beta|A-kinase anchor protein 9 kDa|AKAP 18|AKAP-7 isoform gamma|AKAP-7 isoforms alpha and beta|PRKA7 isoform gamma|PRKA7 isoforms alpha/beta', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131135686, end_position_on_the_genomic_accession=131283535, orientation='plus', exon_count=14, OMIM=604693, no_hdr0=''),
  9466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9466, CurrentID=0, Status='live', Symbol='IL27RA', Aliases='CRL1, IL-27RA, IL27R, TCCR, WSX1, zcytor1', description='interleukin 27 receptor subunit alpha', other_designations='IL-27 receptor subunit alpha|IL-27R subunit alpha|IL-27R-alpha|T-cell cytokine receptor type 1|class I cytokine receptor|cytokine receptor WSX-1|cytokine receptor-like 1|interleukin 27 receptor, alpha|type I T-cell cytokine receptor', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14031740, end_position_on_the_genomic_accession=14052919, orientation='plus', exon_count=14, OMIM=605350, no_hdr0=''),
  9467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9467, CurrentID=0, Status='live', Symbol='SH3BP5', Aliases='SAB, SH3BP-5', description='SH3-domain binding protein 5 (BTK-associated)', other_designations='SH3 domain-binding protein that preferentially associates with BTK', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15254353, end_position_on_the_genomic_accession=15341394, orientation='minus', exon_count=12, OMIM=605612, no_hdr0=''),
  9468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9468, CurrentID=0, Status='live', Symbol='PCYT1B', Aliases='CCTB, CTB', description='phosphate cytidylyltransferase 1, choline, beta', other_designations='CCT B|CCT-beta|CT B|CTP:phosphocholine cytidylyltransferase b|phosphorylcholine transferase B', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24558087, end_position_on_the_genomic_accession=24672862, orientation='minus', exon_count=10, OMIM=300948, no_hdr0=''),
  9469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9469, CurrentID=0, Status='live', Symbol='CHST3', Aliases='C6ST, C6ST1, HSD', description='carbohydrate (chondroitin 6) sulfotransferase 3', other_designations='C6ST-1|GST-0|chondroitin 6-O-sulfotransferase 1|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 0', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71964362, end_position_on_the_genomic_accession=72013564, orientation='plus', exon_count=5, OMIM=603799, no_hdr0=''),
  9470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9470, CurrentID=0, Status='live', Symbol='EIF4E2', Aliases='4E-LP, 4EHP, EIF4EL3, IF4e', description='eukaryotic translation initiation factor 4E family member 2', other_designations='eIF-4E type 2|eIF4E-like cap-binding protein|eIF4E-like protein 4E-LP|eukaryotic translation initiation factor 4E homologous protein|eukaryotic translation initiation factor 4E-like 3|mRNA cap-binding protein type 3', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232550057, end_position_on_the_genomic_accession=232583645, orientation='plus', exon_count=10, OMIM=605895, no_hdr0=''),
  284009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284009, CurrentID=0, Status='live', Symbol='LOC284009', Aliases='', description='uncharacterized LOC284009', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2406981, end_position_on_the_genomic_accession=2415436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  9472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9472, CurrentID=0, Status='live', Symbol='AKAP6', Aliases='ADAP100, ADAP6, AKAP100, PRKA6, mAKAP', description='A-kinase anchoring protein 6', other_designations='A kinase (PRKA) anchor protein 6|A-kinase anchor protein 100 kDa|AKAP 100|AKAP-6|protein kinase A anchoring protein 6|protein kinase A-anchoring protein 6', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32329273, end_position_on_the_genomic_accession=32833999, orientation='plus', exon_count=21, OMIM=604691, no_hdr0=''),
  9473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9473, CurrentID=0, Status='live', Symbol='THEMIS2', Aliases='C1orf38, ICB-1', description='thymocyte selection associated family member 2', other_designations='basement membrane-induced|induced by contact to basement membrane 1 protein|protein ICB-1|thymocyte-expressed molecule involved in selection protein 2', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27872543, end_position_on_the_genomic_accession=27886685, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  9474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9474, CurrentID=0, Status='live', Symbol='ATG5', Aliases='APG5, APG5-LIKE, APG5L, ASP, hAPG5', description='autophagy related 5', other_designations='ATG5 autophagy related 5 homolog|apoptosis-specific protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106184476, end_position_on_the_genomic_accession=106325820, orientation='minus', exon_count=10, OMIM=604261, no_hdr0=''),
  9475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9475, CurrentID=0, Status='live', Symbol='ROCK2', Aliases='ROCK-II', description='Rho-associated, coiled-coil containing protein kinase 2', other_designations='p164 ROCK-2|rho-associated, coiled-coil-containing protein kinase II', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11179764, end_position_on_the_genomic_accession=11345363, orientation='minus', exon_count=35, OMIM=604002, no_hdr0=''),
  9476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9476, CurrentID=0, Status='live', Symbol='NAPSA', Aliases='KAP, Kdap, NAP1, NAPA, SNAPA', description='napsin A aspartic peptidase', other_designations='ASP4|CTB-191K22.6|TA01/TA02|asp 4|aspartyl protease 4|kidney-derived aspartic protease-like protein|napsin-1|pronapsin A', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50358472, end_position_on_the_genomic_accession=50369331, orientation='minus', exon_count=10, OMIM=605631, no_hdr0=''),
  9477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9477, CurrentID=0, Status='live', Symbol='MED20', Aliases='PRO0213, TRFP', description='mediator complex subunit 20', other_designations='Trf (TATA binding protein-related factor)-proximal homolog', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41905354, end_position_on_the_genomic_accession=41921139, orientation='minus', exon_count=4, OMIM=612915, no_hdr0=''),
  9478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9478, CurrentID=0, Status='live', Symbol='CABP1', Aliases='CALBRAIN, HCALB_BR', description='calcium binding protein 1', other_designations='calcium binding protein 5|caldendrin', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120640619, end_position_on_the_genomic_accession=120667324, orientation='plus', exon_count=8, OMIM=605563, no_hdr0=''),
  9479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9479, CurrentID=0, Status='live', Symbol='MAPK8IP1', Aliases='IB1, JIP-1, JIP1, PRKM8IP', description='mitogen-activated protein kinase 8 interacting protein 1', other_designations='IB-1|JNK MAP kinase scaffold protein 1|JNK-interacting protein 1|PRKM8 interacting protein|islet-brain 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45885496, end_position_on_the_genomic_accession=45906465, orientation='plus', exon_count=13, OMIM=604641, no_hdr0=''),
  9480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9480, CurrentID=0, Status='live', Symbol='ONECUT2', Aliases='OC-2, OC2', description='one cut homeobox 2', other_designations='HNF-6-beta|ONECUT-2 homeodomain transcription factor|hepatocyte nuclear factor 6-beta|one cut domain, family member 2|onecut 2|transcription factor ONECUT-2', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57435685, end_position_on_the_genomic_accession=57491298, orientation='plus', exon_count=2, OMIM=604894, no_hdr0=''),
  9481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9481, CurrentID=0, Status='live', Symbol='SLC25A27', Aliases='UCP4', description='solute carrier family 25 member 27', other_designations='UCP 4|solute carrier family 25, member 27|uncoupling protein 4', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46652915, end_position_on_the_genomic_accession=46678190, orientation='plus', exon_count=9, OMIM=613725, no_hdr0=''),
  9482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9482, CurrentID=0, Status='live', Symbol='STX8', Aliases='CARB', description='syntaxin 8', other_designations='CIP-1-associated regulator of cyclin B', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9250471, end_position_on_the_genomic_accession=9575958, orientation='minus', exon_count=10, OMIM=604203, no_hdr0=''),
  9486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9486, CurrentID=0, Status='live', Symbol='CHST10', Aliases='HNK-1ST, HNK1ST', description='carbohydrate sulfotransferase 10', other_designations='HNK-1 sulfotransferase|huHNK-1ST', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100391860, end_position_on_the_genomic_accession=100417668, orientation='minus', exon_count=10, OMIM=606376, no_hdr0=''),
  9487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9487, CurrentID=0, Status='live', Symbol='PIGL', Aliases='CHIME', description='phosphatidylinositol glycan anchor biosynthesis class L', other_designations='N-acetylglucosaminylphosphatidylinositol deacetylase|PIG-L|phosphatidylinositol glycan anchor biosynthesis, class L|phosphatidylinositol glycan, class L|phosphatidylinositol-glycan biosynthesis class L protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16217195, end_position_on_the_genomic_accession=16349164, orientation='plus', exon_count=11, OMIM=605947, no_hdr0=''),
  9488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9488, CurrentID=0, Status='live', Symbol='PIGB', Aliases='GPI-MT-III, PIG-B', description='phosphatidylinositol glycan anchor biosynthesis class B', other_designations='GPI mannosyltransferase III|dol-P-Man dependent GPI mannosyltransferase|phosphatidylinositol glycan anchor biosynthesis, class B|phosphatidylinositol glycan, class B|phosphatidylinositol-glycan biosynthesis class B protein', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55318865, end_position_on_the_genomic_accession=55355648, orientation='plus', exon_count=13, OMIM=604122, no_hdr0=''),
  9489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9489, CurrentID=0, Status='live', Symbol='PGS1', Aliases='', description='phosphatidylglycerophosphate synthase 1', other_designations='PGP synthase 1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78378618, end_position_on_the_genomic_accession=78424659, orientation='plus', exon_count=12, OMIM=614942, no_hdr0=''),
  9491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9491, CurrentID=0, Status='live', Symbol='PSMF1', Aliases='PI31', description='proteasome inhibitor subunit 1', other_designations='hPI31|proteasome (prosome, macropain) inhibitor subunit 1 (PI31)|proteasome inhibitor hP131 subunit', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1113263, end_position_on_the_genomic_accession=1168378, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  140564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140564, CurrentID=0, Status='live', Symbol='APOBEC3D', Aliases='A3DE, APOBEC3E, ARP6, APOBEC3D', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3D', other_designations='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3E pseudogene|probable DNA dC->dU-editing enzyme APOBEC-3D', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39021113, end_position_on_the_genomic_accession=39033251, orientation='plus', exon_count=7, OMIM=609900, no_hdr0=''),
  9493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9493, CurrentID=0, Status='live', Symbol='KIF23', Aliases='CHO1, KNSL5, MKLP-1, MKLP1', description='kinesin family member 23', other_designations='kinesin-like 5 (mitotic kinesin-like protein 1)', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69414246, end_position_on_the_genomic_accession=69448427, orientation='plus', exon_count=26, OMIM=605064, no_hdr0=''),
  9495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9495, CurrentID=0, Status='live', Symbol='AKAP5', Aliases='AKAP75, AKAP79, H21', description='A-kinase anchoring protein 5', other_designations='A kinase (PRKA) anchor protein 5|A-kinase anchor protein 79 kDa|A-kinase anchoring protein 75/79|AKAP 79|cAMP-dependent protein kinase regulatory subunit II high affinity-binding protein', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64465499, end_position_on_the_genomic_accession=64474503, orientation='plus', exon_count=2, OMIM=604688, no_hdr0=''),
  9496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9496, CurrentID=0, Status='live', Symbol='TBX4', Aliases='SPS', description='T-box 4', other_designations='T-box protein 4', map_location='17q21-q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61444626, end_position_on_the_genomic_accession=61485110, orientation='plus', exon_count=15, OMIM=601719, no_hdr0=''),
  9497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9497, CurrentID=0, Status='live', Symbol='SLC4A7', Aliases='NBC2, NBC3, NBCN1, SBC2, SLC4A6', description='solute carrier family 4, sodium bicarbonate cotransporter, member 7', other_designations='bicarbonate transporter|electroneutral Na/HCO(3) cotransporter|sodium bicarbonate cotransporter 2|sodium bicarbonate cotransporter 2b|solute carrier family 4 member 7', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27372721, end_position_on_the_genomic_accession=27484420, orientation='minus', exon_count=27, OMIM=603353, no_hdr0=''),
  9498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9498, CurrentID=0, Status='live', Symbol='SLC4A8', Aliases='NBC3, NDCBE', description='solute carrier family 4, sodium bicarbonate cotransporter, member 8', other_designations='electroneutral Na(+)-driven Cl-HCO3 exchanger|k-NBC3', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51391317, end_position_on_the_genomic_accession=51515763, orientation='plus', exon_count=32, OMIM=605024, no_hdr0=''),
  9499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9499, CurrentID=0, Status='live', Symbol='MYOT', Aliases='LGMD1, LGMD1A, MFM3, TTID, TTOD', description='myotilin', other_designations='57 kDa cytoskeletal protein|myofibrillar titin-like Ig domains protein|titin immunoglobulin domain protein (myotilin)', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137867856, end_position_on_the_genomic_accession=137887851, orientation='plus', exon_count=10, OMIM=604103, no_hdr0=''),
  9500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9500, CurrentID=0, Status='live', Symbol='MAGED1', Aliases='DLXIN-1, NRAGE', description='MAGE family member D1', other_designations='MAGE tumor antigen CCF|MAGE-D1 antigen|melanoma antigen family D, 1|melanoma antigen family D1|neurotrophin receptor-interacting MAGE homolog', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51803059, end_position_on_the_genomic_accession=51902354, orientation='plus', exon_count=16, OMIM=300224, no_hdr0=''),
  9501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9501, CurrentID=0, Status='live', Symbol='RPH3AL', Aliases='NOC2', description='rabphilin 3A-like (without C2 domains)', other_designations='no C2 domains protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=212389, end_position_on_the_genomic_accession=386254, orientation='minus', exon_count=13, OMIM=604881, no_hdr0=''),
  9502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9502, CurrentID=0, Status='live', Symbol='XAGE2', Aliases='CT12.2, GAGED3, XAGE-2B, XAGE2', description='X antigen family member 2', other_designations='G antigen family D member 3|X antigen family, member 2|X antigen family, member 2B|cancer/testis antigen 12.2|cancer/testis antigen family 12, member 2', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52368811, end_position_on_the_genomic_accession=52375683, orientation='plus', exon_count=6, OMIM=300416, no_hdr0=''),
  101669765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669765, CurrentID=0, Status='live', Symbol='LINC00914', Aliases='RAGE', description='long intergenic non-protein coding RNA 914', other_designations='RNA for advanced-glycation end products', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  140576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140576, CurrentID=0, Status='live', Symbol='S100A16', Aliases='AAG13, DT1P1A7, S100F', description='S100 calcium binding protein A16', other_designations='S100 calcium-binding protein A16|aging-associated gene 13 protein|aging-associated protein 13|protein S100-F', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153606886, end_position_on_the_genomic_accession=153613038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  140578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140578, CurrentID=0, Status='live', Symbol='CHODL', Aliases='C21orf68, MT75, PRED12', description='chondrolectin', other_designations='transmembrane protein MT75', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17819595, end_position_on_the_genomic_accession=18267373, orientation='plus', exon_count=12, OMIM=607247, no_hdr0=''),
  9507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9507, CurrentID=0, Status='live', Symbol='ADAMTS4', Aliases='ADAMTS-2, ADAMTS-4, ADMP-1', description='ADAM metallopeptidase with thrombospondin type 1 motif 4', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 4|ADAM-TS 4|ADAM-TS4|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 4|aggrecanase-1', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161189748, end_position_on_the_genomic_accession=161199055, orientation='minus', exon_count=9, OMIM=603876, no_hdr0=''),
  9508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9508, CurrentID=0, Status='live', Symbol='ADAMTS3', Aliases='ADAMTS-4', description='ADAM metallopeptidase with thrombospondin type 1 motif 3', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 3|ADAM-TS 3|ADAM-TS3|ADAMTS-3|PC II-NP|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 3|procollagen II N-proteinase|procollagen II amino propeptide-processing enzyme|zinc metalloendopeptidase', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=72280969, end_position_on_the_genomic_accession=72568799, orientation='minus', exon_count=27, OMIM=605011, no_hdr0=''),
  9509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9509, CurrentID=0, Status='live', Symbol='ADAMTS2', Aliases='ADAM-TS2, ADAMTS-2, ADAMTS-3, NPI, PC I-NP, PCI-NP, PCINP, PCPNI, PNPI', description='ADAM metallopeptidase with thrombospondin type 1 motif 2', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 2|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 2|procollagen I N-proteinase|procollagen I/II amino propeptide-processing enzyme|procollagen N-endopeptidase', map_location='5qter', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179110851, end_position_on_the_genomic_accession=179345430, orientation='minus', exon_count=23, OMIM=604539, no_hdr0=''),
  9510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9510, CurrentID=0, Status='live', Symbol='ADAMTS1', Aliases='C3-C5, METH1', description='ADAM metallopeptidase with thrombospondin type 1 motif 1', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 1|ADAM-TS 1|ADAM-TS1|ADAMTS-1|METH-1|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 1|human metalloproteinase with thrombospondin type 1 motifs', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26836287, end_position_on_the_genomic_accession=26845409, orientation='minus', exon_count=9, OMIM=605174, no_hdr0=''),
  9512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9512, CurrentID=0, Status='live', Symbol='PMPCB', Aliases='Beta-MPP, MPP11, MPPB, MPPP52, P-52', description='peptidase (mitochondrial processing) beta', other_designations='mitochondrial processing peptidase beta subunit', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103297426, end_position_on_the_genomic_accession=103329902, orientation='plus', exon_count=14, OMIM=603131, no_hdr0=''),
  9513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9513, CurrentID=0, Status='live', Symbol='FXR2', Aliases='FMR1L2P, FXR2', description='fragile X mental retardation, autosomal homolog 2', other_designations='fragile X-mental retardation 1-like 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7591230, end_position_on_the_genomic_accession=7614897, orientation='minus', exon_count=18, OMIM=605339, no_hdr0=''),
  9514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9514, CurrentID=0, Status='live', Symbol='GAL3ST1', Aliases='CST', description='galactose-3-O-sulfotransferase 1', other_designations="3'-phosphoadenosine-5'-phosphosulfate:GalCer sulfotransferase|cerebroside (3'-phosphoadenylylsulfate:galactosylceramide 3') sulfotransferase", map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30554635, end_position_on_the_genomic_accession=30574588, orientation='minus', exon_count=11, OMIM=602300, no_hdr0=''),
  9515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9515, CurrentID=0, Status='live', Symbol='STXBP5L', Aliases='LLGL4', description='syntaxin binding protein 5 like', other_designations='lethal(2) giant larvae protein homolog 4|syntaxin binding protein 5-like|tomosyn-2|tomosyn-like', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120908203, end_position_on_the_genomic_accession=121424761, orientation='plus', exon_count=33, OMIM=609381, no_hdr0=''),
  9516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9516, CurrentID=0, Status='live', Symbol='LITAF', Aliases='PIG7, SIMPLE, TP53I7', description='lipopolysaccharide-induced TNF factor', other_designations='LPS-induced TNF-alpha factor|lipopolysaccharide-induced TNF-alpha factor|p53-induced gene 7 protein|small integral membrane protein of lysosome/late endosome|tumor protein p53 inducible protein 7', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11547722, end_position_on_the_genomic_accession=11636377, orientation='minus', exon_count=14, OMIM=603795, no_hdr0=''),
  9517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9517, CurrentID=0, Status='live', Symbol='SPTLC2', Aliases='HSN1C, LCB2, LCB2A, NSAN1C, SPT2, hLCB2a', description='serine palmitoyltransferase, long chain base subunit 2', other_designations='LCB 2|SPT 2|long chain base biosynthesis protein 2a|serine palmitoyltransferase, subunit II|serine-palmitoyl-CoA transferase 2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77505997, end_position_on_the_genomic_accession=77616768, orientation='minus', exon_count=13, OMIM=605713, no_hdr0=''),
  9518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9518, CurrentID=0, Status='live', Symbol='GDF15', Aliases='GDF-15, MIC-1, MIC1, NAG-1, PDF, PLAB, PTGFB', description='growth differentiation factor 15', other_designations='NRG-1|NSAID (nonsteroidal anti-inflammatory drug)-activated protein 1|NSAID-activated gene 1 protein|NSAID-regulated gene 1 protein|PTGF-beta|macrophage inhibitory cytokine 1|placental TGF-beta|placental bone morphogenetic protein|prostate differentiation factor', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18386158, end_position_on_the_genomic_accession=18389176, orientation='plus', exon_count=2, OMIM=605312, no_hdr0=''),
  9519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9519, CurrentID=0, Status='live', Symbol='TBPL1', Aliases='MGC:8389, MGC:9620, STUD, TLF, TLP, TRF2', description='TATA-box binding protein like 1', other_designations='21-kDA TBP-like protein|TATA box-binding protein-related factor 2|TBP-like 1|TBP-related factor 2|TBP-related protein|second TBP of unique DNA protein', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133952170, end_position_on_the_genomic_accession=133987500, orientation='plus', exon_count=8, OMIM=605521, no_hdr0=''),
  9520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9520, CurrentID=0, Status='live', Symbol='NPEPPS', Aliases='AAP-S, MP100, PSA', description='aminopeptidase puromycin sensitive', other_designations='cytosol alanyl aminopeptidase|metalloproteinase MP100', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47522950, end_position_on_the_genomic_accession=47623276, orientation='plus', exon_count=24, OMIM=606793, no_hdr0=''),
  9521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9521, CurrentID=0, Status='live', Symbol='EEF1E1', Aliases='AIMP3, P18', description='eukaryotic translation elongation factor 1 epsilon 1', other_designations='ARS-interacting multifunctional protein 3|aminoacyl tRNA synthetase complex-interacting multifunctional protein 3|multisynthase complex auxiliary component p18|p18 component of aminoacyl-tRNA synthetase complex', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8073360, end_position_on_the_genomic_accession=8102595, orientation='minus', exon_count=5, OMIM=609206, no_hdr0=''),
  9522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9522, CurrentID=0, Status='live', Symbol='SCAMP1', Aliases='SCAMP, SCAMP37', description='secretory carrier membrane protein 1', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78360503, end_position_on_the_genomic_accession=78480739, orientation='plus', exon_count=9, OMIM=606911, no_hdr0=''),
  140596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140596, CurrentID=0, Status='live', Symbol='DEFB104A', Aliases='BD-4, DEFB-4, DEFB104, DEFB4, hBD-4', description='defensin beta 104A', other_designations='defensin, beta 104A|defensin, beta 4', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7836471, end_position_on_the_genomic_accession=7841242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  9525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9525, CurrentID=0, Status='live', Symbol='VPS4B', Aliases='MIG1, SKD1, SKD1B, VPS4-2', description='vacuolar protein sorting 4 homolog B (S. cerevisiae)', other_designations='cell migration-inducing 1|cell migration-inducing gene 1 protein|suppressor of K(+) transport growth defect 1|suppressor of K+ transport defect 1|vacuolar protein sorting 4B', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63389190, end_position_on_the_genomic_accession=63422519, orientation='minus', exon_count=12, OMIM=609983, no_hdr0=''),
  9526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9526, CurrentID=0, Status='live', Symbol='MPDU1', Aliases='CDGIF, HBEBP2BPA, Lec35, My008, PP3958, PQLC5, SL15', description='mannose-P-dolichol utilization defect 1', other_designations='HBeAg-binding protein 2 binding protein A|suppressor of Lec15 and Lec35 glycosylation mutation homolog', map_location='17p13.1-p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7583647, end_position_on_the_genomic_accession=7588212, orientation='plus', exon_count=7, OMIM=604041, no_hdr0=''),
  9527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9527, CurrentID=0, Status='live', Symbol='GOSR1', Aliases='GOLIM2, GOS-28, GOS28, GOS28/P28, GS28, P28', description='golgi SNAP receptor complex member 1', other_designations='28 kDa Golgi SNARE protein|28 kDa cis-Golgi SNARE p28|Golgi SNARE 28 kDa|cis-golgi SNARE|golgi integral membrane protein 2', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30477408, end_position_on_the_genomic_accession=30526814, orientation='plus', exon_count=10, OMIM=604026, no_hdr0=''),
  9528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9528, CurrentID=0, Status='live', Symbol='TMEM59', Aliases='C1orf8, DCF1, HSPC001, PRO195, UNQ169', description='transmembrane protein 59', other_designations='dendritic cell factor 1|liver membrane-bound protein', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54031673, end_position_on_the_genomic_accession=54053462, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  9529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9529, CurrentID=0, Status='live', Symbol='BAG5', Aliases='BAG-5', description='BCL2 associated athanogene 5', other_designations='BAG-family molecular chaperone regulator-5|BCL2-associated athanogene 5|bcl-2-associated athanogene 5', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103556544, end_position_on_the_genomic_accession=103562814, orientation='minus', exon_count=3, OMIM=603885, no_hdr0=''),
  9530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9530, CurrentID=0, Status='live', Symbol='BAG4', Aliases='BAG-4, SODD', description='BCL2 associated athanogene 4', other_designations='BAG-family molecular chaperone regulator-4|BCL2-associated athanogene 4|bcl-2-associated athanogene 4|silencer of death domains', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38176588, end_position_on_the_genomic_accession=38213301, orientation='plus', exon_count=5, OMIM=603884, no_hdr0=''),
  9531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9531, CurrentID=0, Status='live', Symbol='BAG3', Aliases='BAG-3, BIS, CAIR-1, MFM6', description='BCL2 associated athanogene 3', other_designations='BCL2-associated athanogene 3|BCL2-binding athanogene 3|bcl-2-binding protein Bis|docking protein CAIR-1', map_location='10q25.2-q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119651347, end_position_on_the_genomic_accession=119677819, orientation='plus', exon_count=4, OMIM=603883, no_hdr0=''),
  9532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9532, CurrentID=0, Status='live', Symbol='BAG2', Aliases='BAG-2, dJ417I1.2', description='BCL2 associated athanogene 2', other_designations='BAG-family molecular chaperone regulator-2|BCL2-associated athanogene 2|bcl-2-associated athanogene 2|dJ417I1.2 (BAG-family molecular chaperone regulator 2)', map_location='6p12.1-p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57172306, end_position_on_the_genomic_accession=57185215, orientation='plus', exon_count=5, OMIM=603882, no_hdr0=''),
  9533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9533, CurrentID=0, Status='live', Symbol='POLR1C', Aliases='RPA39, RPA40, RPA5, RPAC1, TCS3', description='polymerase (RNA) I polypeptide C', other_designations='AC40|DNA-directed RNA polymerase I subunit C|DNA-directed RNA polymerases I and III 40 kDa polypeptide|RNA polymerases I and III subunit AC1|RPC40|polymerase (RNA) I polypeptide C, 30kDa', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43517039, end_position_on_the_genomic_accession=43550984, orientation='plus', exon_count=11, OMIM=610060, no_hdr0=''),
  140606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140606, CurrentID=0, Status='live', Symbol='SELM', Aliases='SEPM', description='selenoprotein M', other_designations='selenoprotein SelM', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31104777, end_position_on_the_genomic_accession=31107565, orientation='minus', exon_count=5, OMIM=610918, no_hdr0=''),
  9535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9535, CurrentID=0, Status='live', Symbol='GMFG', Aliases='GMF-GAMMA', description='glia maturation factor, gamma', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39328357, end_position_on_the_genomic_accession=39336086, orientation='minus', exon_count=7, OMIM=604104, no_hdr0=''),
  9536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9536, CurrentID=0, Status='live', Symbol='PTGES', Aliases='MGST-IV, MGST1-L1, MGST1L1, MPGES, PGES, PIG12, PP102, PP1294, TP53I12, mPGES-1', description='prostaglandin E synthase', other_designations='MGST1-like 1|glutathione S-transferase 1-like 1|microsomal glutathione S-transferase 1-like 1|microsomal prostaglandin E synthase 1|microsomal prostaglandin E synthase-1|p53-induced apoptosis protein 12|p53-induced gene 12 protein|tumor protein p53 inducible protein 12', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129738336, end_position_on_the_genomic_accession=129777327, orientation='minus', exon_count=4, OMIM=605172, no_hdr0=''),
  140609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140609, CurrentID=0, Status='live', Symbol='NEK7', Aliases='', description='NIMA-related kinase 7', other_designations='NIMA (never in mitosis gene a)-related kinase 7|never in mitosis A-related kinase 7|nimA-related protein kinase 7', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198156978, end_position_on_the_genomic_accession=198322420, orientation='plus', exon_count=10, OMIM=606848, no_hdr0=''),
  9538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9538, CurrentID=0, Status='live', Symbol='EI24', Aliases='EPG4, PIG8, TP53I8', description='etoposide induced 2.4', other_designations='ectopic P-granules autophagy protein 4 homolog|p53-induced gene 8 protein|tumor protein p53 inducible protein 8', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125569216, end_position_on_the_genomic_accession=125584689, orientation='plus', exon_count=13, OMIM=605170, no_hdr0=''),
  9540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9540, CurrentID=0, Status='live', Symbol='TP53I3', Aliases='PIG3', description='tumor protein p53 inducible protein 3', other_designations='p53-induced gene 3 protein|quinone oxidoreductase homolog', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24077428, end_position_on_the_genomic_accession=24085215, orientation='minus', exon_count=6, OMIM=605171, no_hdr0=''),
  9541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9541, CurrentID=0, Status='live', Symbol='CIR1', Aliases='CIR', description='corepressor interacting with RBPJ, 1', other_designations='CBF1 interacting corepressor|CBF1-interacting corepressor|recepin', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174348150, end_position_on_the_genomic_accession=174395715, orientation='minus', exon_count=10, OMIM=605228, no_hdr0=''),
  9542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9542, CurrentID=0, Status='live', Symbol='NRG2', Aliases='DON1, HRG2, NTAK', description='neuregulin 2', other_designations='divergent of neuregulin-1|neural- and thymus-derived activator for ErbB kinases|pro-NRG2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139846779, end_position_on_the_genomic_accession=140043437, orientation='minus', exon_count=17, OMIM=603818, no_hdr0=''),
  9543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9543, CurrentID=0, Status='live', Symbol='IGDCC3', Aliases='HsT18880, PUNC', description='immunoglobulin superfamily, DCC subclass, member 3', other_designations='putative neuronal cell adhesion molecule', map_location='15q22.3-q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65327127, end_position_on_the_genomic_accession=65378040, orientation='minus', exon_count=18, OMIM=604184, no_hdr0=''),
  9545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9545, CurrentID=0, Status='live', Symbol='RAB3D', Aliases='D2-2, GOV, RAB16, RAD3D', description='RAB3D, member RAS oncogene family', other_designations='Rab3D upregulated with myeloid differentiation|glioblastoma overexpressed', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11322046, end_position_on_the_genomic_accession=11339668, orientation='minus', exon_count=5, OMIM=604350, no_hdr0=''),
  9546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9546, CurrentID=0, Status='live', Symbol='APBA3', Aliases='MGC:15815, X11L2, mint3', description='amyloid beta (A4) precursor protein-binding, family A, member 3', other_designations='X11-like 2 protein|adapter protein X11gamma|amyloid beta (A4) precursor protein-binding, family A, member 3 (X11-like 2)|mint-3|neuron-specific X11L2 protein|neuronal munc18-1-interacting protein 3|phosphotyrosine-binding/-interacting domain (PTB)-bearing protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3750773, end_position_on_the_genomic_accession=3761698, orientation='minus', exon_count=10, OMIM=604262, no_hdr0=''),
  9547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9547, CurrentID=0, Status='live', Symbol='CXCL14', Aliases='BMAC, BRAK, KEC, KS1, MIP-2g, MIP2G, NJAC, SCYB14', description='chemokine (C-X-C motif) ligand 14', other_designations='CXC chemokine in breast and kidney|MIP-2 gamma|bolekine|breast and kidney|chemokine BRAK|small inducible cytokine subfamily B (Cys-X-Cys), member 14 (BRAK)|small-inducible cytokine B14|tumor-suppressing chemokine', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135570679, end_position_on_the_genomic_accession=135579279, orientation='minus', exon_count=4, OMIM=604186, no_hdr0=''),
  9550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9550, CurrentID=0, Status='live', Symbol='ATP6V1G1', Aliases='ATP6G, ATP6G1, ATP6GL, ATP6J, Vma10', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump), member J|V-ATPase 13 kDa subunit 1|V-ATPase subunit G 1|vacuolar ATP synthase subunit M16|vacuolar H(+)-ATPase subunit G 1|vacuolar proton pump subunit G 1|vacuolar proton pump subunit M16', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114587714, end_position_on_the_genomic_accession=114598872, orientation='plus', exon_count=3, OMIM=607296, no_hdr0=''),
  9551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9551, CurrentID=0, Status='live', Symbol='ATP5J2', Aliases='ATP5JL', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2', other_designations='ATP synthase f chain, mitochondrial|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit f|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2|F1F0-type ATPase subunit f|F1Fo-ATP synthase complex Fo membrane domain f subunit|F1Fo-ATPase synthase f subunit', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99458161, end_position_on_the_genomic_accession=99466201, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  9552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9552, CurrentID=0, Status='live', Symbol='SPAG7', Aliases='ACRP, FSA-1', description='sperm associated antigen 7', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4959226, end_position_on_the_genomic_accession=4967837, orientation='minus', exon_count=8, OMIM=610056, no_hdr0=''),
  9553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9553, CurrentID=0, Status='live', Symbol='MRPL33', Aliases='C2orf1, L33mt, MRP-L33, RPL33L', description='mitochondrial ribosomal protein L33', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27771717, end_position_on_the_genomic_accession=27779741, orientation='plus', exon_count=4, OMIM=610059, no_hdr0=''),
  9554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9554, CurrentID=0, Status='live', Symbol='SEC22B', Aliases='ERS-24, SEC22L1', description='SEC22 homolog B, vesicle trafficking protein (gene/pseudogene)', other_designations='ER-Golgi SNARE of 24 kDa|SEC22 vesicle trafficking protein homolog B|SEC22 vesicle trafficking protein-like 1|SEC22 vesicle-trafficking protein homolog B', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120155942, end_position_on_the_genomic_accession=120176521, orientation='minus', exon_count=5, OMIM=604029, no_hdr0=''),
  9555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9555, CurrentID=0, Status='live', Symbol='H2AFY', Aliases='H2A.y, H2A/y, H2AF12M, MACROH2A1.1, mH2A1, macroH2A1.2', description='H2A histone family member Y', other_designations='H2A histone family, member Y|histone H2A.y|histone macroH2A1|histone macroH2A1.1|histone macroH2A1.2|medulloblastoma antigen MU-MB-50.205', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135334381, end_position_on_the_genomic_accession=135400147, orientation='minus', exon_count=14, OMIM=610054, no_hdr0=''),
  140628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140628, CurrentID=0, Status='live', Symbol='GATA5', Aliases='GATAS, bB379O24.1', description='GATA binding protein 5', other_designations='GATA binding factor-5|GATA-binding factor 5|GATA-binding protein 5', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62463497, end_position_on_the_genomic_accession=62475970, orientation='minus', exon_count=8, OMIM=611496, no_hdr0=''),
  9557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9557, CurrentID=0, Status='live', Symbol='CHD1L', Aliases='ALC1, CHDL', description='chromodomain helicase DNA binding protein 1-like', other_designations='amplified in liver cancer 1|amplified in liver cancer protein 1', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147242641, end_position_on_the_genomic_accession=147295766, orientation='plus', exon_count=23, OMIM=613039, no_hdr0=''),
  9559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9559, CurrentID=0, Status='live', Symbol='VPS26A', Aliases='HB58, Hbeta58, PEP8A, VPS26', description='VPS26 retromer complex component A', other_designations='VPS26 retromer complex comonent A|vacuolar protein sorting 26 homolog A|vesicle protein sorting 26A', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69124152, end_position_on_the_genomic_accession=69172861, orientation='plus', exon_count=9, OMIM=605506, no_hdr0=''),
  9560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9560, CurrentID=0, Status='live', Symbol='CCL4L2', Aliases='AT744.2, CCL4L, SCYA4L, SCYQ4L2', description='chemokine (C-C motif) ligand 4-like 2', other_designations='MIP-1-beta|macrophage inflammatory protein 1-beta|macrophage inflammatory protein-1b2|monocyte adherence-induced protein 5-alpha|small inducible cytokine A4-like', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36211063, end_position_on_the_genomic_accession=36212878, orientation='plus', exon_count=3, OMIM=610757, no_hdr0=''),
  9562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9562, CurrentID=0, Status='live', Symbol='MINPP1', Aliases='HIPER1, MINPP2, MIPP', description='multiple inositol-polyphosphate phosphatase 1', other_designations='2,3-BPG phosphatase|2,3-bisphosphoglycerate 3-phosphatase|inositol (1,3,4,5)-tetrakisphosphate 3-phosphatase|ins(1,3,4,5)P(4) 3-phosphatase|multiple inositol polyphosphate histidine phosphatase, 1|multiple inositol polyphosphate phosphatase 2', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87502792, end_position_on_the_genomic_accession=87553461, orientation='plus', exon_count=6, OMIM=605391, no_hdr0=''),
  9563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9563, CurrentID=0, Status='live', Symbol='H6PD', Aliases='CORTRD1, G6PDH, GDH', description='hexose-6-phosphate dehydrogenase (glucose 1-dehydrogenase)', other_designations='6-phosphogluconolactonase|G6PD, H form|glucose 1- dehydrogenase|glucose dehydrogenase|glucose dehyrogenase|glucose-6-phosphate dehydrogenase, salivary', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9234776, end_position_on_the_genomic_accession=9271337, orientation='plus', exon_count=7, OMIM=138090, no_hdr0=''),
  9564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9564, CurrentID=0, Status='live', Symbol='BCAR1', Aliases='CAS, CAS1, CASS1, CRKAS, P130Cas', description='breast cancer anti-estrogen resistance 1', other_designations='Cas scaffolding protein family member 1|Crk-associated substrate p130Cas', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75229030, end_position_on_the_genomic_accession=75268053, orientation='minus', exon_count=17, OMIM=602941, no_hdr0=''),
  9566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9566, CurrentID=0, Status='live', Symbol='HPCX', Aliases='', description='hereditary prostate cancer, X-linked', other_designations='Prostate cancer, hereditary, X-linked', map_location='Xq27-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300147, no_hdr0=''),
  9567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9567, CurrentID=0, Status='live', Symbol='GTPBP1', Aliases='GP-1, GP1, HSPC018', description='GTP binding protein 1', other_designations='G-protein 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38704557, end_position_on_the_genomic_accession=38738173, orientation='plus', exon_count=16, OMIM=602245, no_hdr0=''),
  9568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9568, CurrentID=0, Status='live', Symbol='GABBR2', Aliases='GABABR2, GPR51, GPRC3B, HG20, HRIHFB2099', description='gamma-aminobutyric acid (GABA) B receptor, 2', other_designations='G protein-coupled receptor 51|G-protein coupled receptor 51|GABA-B receptor 2|GABA-B receptor, R2 subunit|GABA-B-R2|GABA-BR2|gamma-aminobutyric acid B receptor 2|gb2', map_location='9q22.1-q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98288082, end_position_on_the_genomic_accession=98709197, orientation='minus', exon_count=20, OMIM=607340, no_hdr0=''),
  9569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9569, CurrentID=0, Status='live', Symbol='GTF2IRD1', Aliases='BEN, CREAM1, GTF3, MUSTRD1, RBAP2, WBS, WBSCR11, WBSCR12, hMusTRD1alpha1', description='GTF2I repeat domain containing 1', other_designations='USE B1-binding protein|Williams-Beuren syndrome chromosome region 11|binding factor for early enhancer|general transcription factor 3|general transcription factor III|muscle TFII-I repeat domain-containing protein 1 alpha 1|slow-muscle-fiber enhancer-binding protein|williams-Beuren syndrome chromosomal region 12 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74453790, end_position_on_the_genomic_accession=74602605, orientation='plus', exon_count=28, OMIM=604318, no_hdr0=''),
  9570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9570, CurrentID=0, Status='live', Symbol='GOSR2', Aliases='Bos1, EPM6, GS27', description='golgi SNAP receptor complex member 2', other_designations='27 kDa Golgi SNARE protein|membrin', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46923114, end_position_on_the_genomic_accession=46967000, orientation='plus', exon_count=17, OMIM=604027, no_hdr0=''),
  9572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9572, CurrentID=0, Status='live', Symbol='NR1D1', Aliases='EAR1, THRA1, THRAL, ear-1, hRev', description='nuclear receptor subfamily 1 group D member 1', other_designations='Rev-ErbAalpha|V-erbA-related protein 1|nuclear receptor Rev-ErbA-alpha|nuclear receptor subfamily 1, group D, member 1|rev-erbA-alpha', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40092784, end_position_on_the_genomic_accession=40100725, orientation='minus', exon_count=8, OMIM=602408, no_hdr0=''),
  9573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9573, CurrentID=0, Status='live', Symbol='GDF3', Aliases='KFS3, MCOP7, MCOPCB6', description='growth differentiation factor 3', other_designations='GDF-3', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7689785, end_position_on_the_genomic_accession=7695764, orientation='minus', exon_count=2, OMIM=606522, no_hdr0=''),
  9575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9575, CurrentID=0, Status='live', Symbol='CLOCK', Aliases='KAT13D, bHLHe8', description='clock circadian regulator', other_designations='circadian locomoter output cycles kaput protein|class E basic helix-loop-helix protein 8|clock homolog', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55427901, end_position_on_the_genomic_accession=55546909, orientation='minus', exon_count=27, OMIM=601851, no_hdr0=''),
  9576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9576, CurrentID=0, Status='live', Symbol='SPAG6', Aliases='CT141, Repro-SA-1, pf16', description='sperm associated antigen 6', other_designations='axoneme central apparatus protein|protein PF16 homolog|sperm flagellar protein', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22345416, end_position_on_the_genomic_accession=22417610, orientation='plus', exon_count=15, OMIM=605730, no_hdr0=''),
  9577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9577, CurrentID=0, Status='live', Symbol='BRE', Aliases='BRCC4, BRCC45', description='brain and reproductive organ-expressed (TNFRSF1A modulator)', other_designations='BRCA1/BRCA2-containing complex subunit 45|BRCA1/BRCA2-containing complex, subunit 4|brain and reproductive organ-expressed protein', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27890615, end_position_on_the_genomic_accession=28338901, orientation='plus', exon_count=15, OMIM=610497, no_hdr0=''),
  9578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9578, CurrentID=0, Status='live', Symbol='CDC42BPB', Aliases='MRCKB', description='CDC42 binding protein kinase beta (DMPK-like)', other_designations='CDC42-binding protein kinase beta (DMPK-like)|CDC42BP-beta|DMPK-like beta|MRCK beta|myotonic dystrophy kinase-related CDC42-binding kinase beta|myotonic dystrophy protein kinase-like beta', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102932379, end_position_on_the_genomic_accession=103057405, orientation='minus', exon_count=39, OMIM=614062, no_hdr0=''),
  148430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148430, CurrentID=0, Status='live', Symbol='LOC148430', Aliases='', description='ribosomal protein S2 pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161406045, end_position_on_the_genomic_accession=161407121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9580, CurrentID=0, Status='live', Symbol='SOX13', Aliases='ICA12, Sox-13', description='SRY-box 13', other_designations='SRY (sex determining region Y)-box 13|SRY box 13|SRY-related HMG-box gene 13|islet cell antibody 12|islet cell antigen 12|type 1 diabetes autoantigen ICA12', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204073109, end_position_on_the_genomic_accession=204127743, orientation='plus', exon_count=14, OMIM=604748, no_hdr0=''),
  9581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9581, CurrentID=0, Status='live', Symbol='PREPL', Aliases='', description='prolyl endopeptidase-like', other_designations='putative prolyl oligopeptidase', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44317607, end_position_on_the_genomic_accession=44361862, orientation='minus', exon_count=16, OMIM=609557, no_hdr0=''),
  9582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9582, CurrentID=0, Status='live', Symbol='APOBEC3B', Aliases='A3B, APOBEC1L, ARCD3, ARP4, DJ742C19.2, PHRBNL, bK150C2.2', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3B', other_designations='apolipoprotein B mRNA editing enzyme catalytic polypeptide-like 3B|cytidine deaminase|phorbolin 2|phorbolin 3|phorbolin-1-related protein|phorbolin-2/3|probable DNA dC->dU-editing enzyme APOBEC-3B', map_location='22q13.1-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38982399, end_position_on_the_genomic_accession=38992779, orientation='plus', exon_count=8, OMIM=607110, no_hdr0=''),
  9583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9583, CurrentID=0, Status='live', Symbol='ENTPD4', Aliases='LALP70, LAP70, LYSAL1, NTPDase-4, UDPase', description='ectonucleoside triphosphate diphosphohydrolase 4', other_designations='golgi luminal UDPase|guanosine-diphosphatase like protein|lysosomal apyrase-like 1|lysosomal apyrase-like protein of 70 kDa|uridine-diphosphatase', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23429152, end_position_on_the_genomic_accession=23457731, orientation='minus', exon_count=13, OMIM=607577, no_hdr0=''),
  9584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9584, CurrentID=0, Status='live', Symbol='RBM39', Aliases='CAPER, CAPERalpha, FSAP59, HCC1, RNPC2', description='RNA binding motif protein 39', other_designations='CAPER alpha|RNA-binding region (RNP1, RRM) containing 2|coactivator of activating protein-1 and estrogen receptors|functional spliceosome-associated protein 59|hepatocellular carcinoma protein 1|splicing factor HCC1', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35703609, end_position_on_the_genomic_accession=35742336, orientation='minus', exon_count=19, OMIM=604739, no_hdr0=''),
  9585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9585, CurrentID=0, Status='live', Symbol='KIF20B', Aliases='CT90, KRMP1, MPHOSPH1, MPP-1, MPP1', description='kinesin family member 20B', other_designations='M-phase phosphoprotein 1|cancer/testis antigen 90|kinesin-related motor interacting with PIN1|mitotic kinesin-like protein|mitotic kinesin-related protein', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89701590, end_position_on_the_genomic_accession=89774943, orientation='plus', exon_count=33, OMIM=605498, no_hdr0=''),
  9586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9586, CurrentID=0, Status='live', Symbol='CREB5', Aliases='CRE-BPA, CREB-5', description='cAMP responsive element binding protein 5', other_designations='cAMP response element-binding protein CRE-BPa', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28299321, end_position_on_the_genomic_accession=28825894, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  9587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9587, CurrentID=0, Status='live', Symbol='MAD2L1BP', Aliases='CMT2', description='MAD2L1 binding protein', other_designations='caught by MAD2 protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43629542, end_position_on_the_genomic_accession=43640952, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  9588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9588, CurrentID=0, Status='live', Symbol='PRDX6', Aliases='1-Cys, AOP2, HEL-S-128m, NSGPx, PRX, aiPLA2, p29', description='peroxiredoxin 6', other_designations='1-Cys PRX|1-Cys peroxiredoxin|24 kDa protein|acidic calcium-independent phospholipase A2|antioxidant protein 2|epididymis secretory sperm binding protein Li 128m|liver 2D page spot 40|non-selenium glutathione peroxidase|red blood cells page spot 12', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173477347, end_position_on_the_genomic_accession=173488807, orientation='plus', exon_count=5, OMIM=602316, no_hdr0=''),
  9589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9589, CurrentID=0, Status='live', Symbol='WTAP', Aliases='Mum2', description='Wilms tumor 1 associated protein', other_designations="PNAS-132|WT1-associated protein|Wilms' tumour 1-associating protein|female-lethal(2)D homolog|hFL(2)D|putative pre-mRNA splicing regulator female-lethal(2D)|wilms tumor 1-associating protein", map_location='6q25-q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159726998, end_position_on_the_genomic_accession=159756319, orientation='plus', exon_count=10, OMIM=605442, no_hdr0=''),
  9590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9590, CurrentID=0, Status='live', Symbol='AKAP12', Aliases='AKAP250, SSeCKS', description='A-kinase anchoring protein 12', other_designations='A kinase (PRKA) anchor protein 12|A-kinase anchor protein, 250kDa|AKAP 250|Src-Suppressed C Kinase Substrate|kinase scaffold protein gravin|myasthenia gravis autoantigen gravin', map_location='6q24-q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151239999, end_position_on_the_genomic_accession=151358559, orientation='plus', exon_count=7, OMIM=604698, no_hdr0=''),
  9592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9592, CurrentID=0, Status='live', Symbol='IER2', Aliases='ETR101', description='immediate early response 2', other_designations='chxI protein homolog', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13150468, end_position_on_the_genomic_accession=13154908, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23444, CurrentID=0, Status='live', Symbol='CDW12', Aliases='p90-120', description='CDw12 antigen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  9595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9595, CurrentID=0, Status='live', Symbol='CYTIP', Aliases='B3-1, CASP, CYBR, CYTHIP, HE, PSCDBP', description='cytohesin 1 interacting protein', other_designations='cbp HE|cytohesin binder and regulator|cytohesin binding protein HE|cytohesin-1 interacting protein|cytohesin-associated scaffolding protein|cytohesin-binding protein HE|pleckstrin homology Sec7 and coiled-coil domains-binding protein|pleckstrin homology, Sec7 and coiled-coil domains, binding protein|pleckstrin homology, Sec7 and coiled/coil domains, binding protein', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157414619, end_position_on_the_genomic_accession=157444092, orientation='minus', exon_count=8, OMIM=604448, no_hdr0=''),
  9596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9596, CurrentID=0, Status='live', Symbol='OR1R1P', Aliases='OR17-01, OR17-1, OR1R2P, OR1R3P, OR20A1P', description='olfactory receptor family 1 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily R, member 1 pseudogene|olfactory receptor, family 1, subfamily R, member 2 pseudogene|olfactory receptor, family 1, subfamily R, member 3 pseudogene|olfactory receptor, family 20, subfamily A, member 1 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3385846, end_position_on_the_genomic_accession=3386874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9597, CurrentID=0, Status='live', Symbol='SMAD5-AS1', Aliases='DAMS, SMAD5AS, SMAD5OS', description='SMAD5 antisense RNA 1', other_designations='SMAD family member 5 opposite strand|SMAD family member 5, antisense|SMAD in the antisense orientation|SMAD5 antisense RNA 1 (non-protein coding)', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136129513, end_position_on_the_genomic_accession=136134890, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  9600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9600, CurrentID=0, Status='live', Symbol='PITPNM1', Aliases='DRES9, NIR2, PITPNM, RDGB, RDGB1, RDGBA, RDGBA1, Rd9', description='phosphatidylinositol transfer protein, membrane associated 1', other_designations='NIR-2|PITPnm 1|PYK2 N-terminal domain-interacting receptor 2|drosophila retinal degeneration B homolog|phosphatidylinositol transfer protein, membrane-associated 1|retinal degeneration B alpha 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67491767, end_position_on_the_genomic_accession=67505372, orientation='minus', exon_count=26, OMIM=608794, no_hdr0=''),
  9601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9601, CurrentID=0, Status='live', Symbol='PDIA4', Aliases='ERP70, ERP72, ERp-72', description='protein disulfide isomerase family A member 4', other_designations='ER protein 70|ER protein 72|endoplasmic reticulum resident protein 70|endoplasmic reticulum resident protein 72|protein disulfide isomerase family A, member 4|protein disulfide isomerase related protein (calcium-binding protein, intestinal-related)|protein disulfide isomerase-associated 4', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149003051, end_position_on_the_genomic_accession=149028690, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  9603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9603, CurrentID=0, Status='live', Symbol='NFE2L3', Aliases='NRF3', description='nuclear factor, erythroid 2 like 3', other_designations='NF-E2-related factor 3|NFE2-related factor 3|nuclear factor (erythroid-derived 2)-like 3|nuclear factor, erythroid 2-like 3|nuclear factor, erythroid derived 2, like 3|nuclear factor-erythroid 2 p45-related factor 3|nuclear factor-erythroid 2-related factor 3', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26152227, end_position_on_the_genomic_accession=26187137, orientation='plus', exon_count=4, OMIM=604135, no_hdr0=''),
  9604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9604, CurrentID=0, Status='live', Symbol='RNF14', Aliases='ARA54, HFB30, HRIHFB2038, TRIAD2', description='ring finger protein 14', other_designations='androgen receptor associated protein 54|androgen receptor-associated protein 54', map_location='5q23.3-q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141966837, end_position_on_the_genomic_accession=141990291, orientation='plus', exon_count=13, OMIM=605675, no_hdr0=''),
  285590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285590, CurrentID=0, Status='live', Symbol='SH3PXD2B', Aliases='FAD49, FTHS, HOFI, KIAA1295, TKS4, TSK4', description='SH3 and PX domains 2B', other_designations='adapter protein HOFI|adaptor protein HOFI|factor for adipocyte differentiation 49|tyrosine kinase substrate with four SH3 domains', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172325181, end_position_on_the_genomic_accession=172454523, orientation='minus', exon_count=14, OMIM=613293, no_hdr0=''),
  140678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140678, CurrentID=0, Status='live', Symbol='MLLT10P1', Aliases='MLLT10L, bA348I14.3', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 10 pseudogene 1', other_designations='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 10 pseudogene 1', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30402908, end_position_on_the_genomic_accession=30403462, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9607, CurrentID=0, Status='live', Symbol='CARTPT', Aliases='CART', description='CART prepropeptide', other_designations='cocaine and amphetamine regulated transcript', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71719163, end_position_on_the_genomic_accession=71721048, orientation='plus', exon_count=3, OMIM=602606, no_hdr0=''),
  140680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140680, CurrentID=0, Status='live', Symbol='C20orf96', Aliases='dJ1103G7.2', description='chromosome 20 open reading frame 96', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=270863, end_position_on_the_genomic_accession=290778, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  9609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9609, CurrentID=0, Status='live', Symbol='RAB36', Aliases='', description='RAB36, member RAS oncogene family', other_designations='small GTP-binding protein Rab36', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23145326, end_position_on_the_genomic_accession=23169285, orientation='plus', exon_count=15, OMIM=605662, no_hdr0=''),
  9610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9610, CurrentID=0, Status='live', Symbol='RIN1', Aliases='', description='Ras and Rab interactor 1', other_designations='ras inhibitor 1|ras inhibitor JC99|ras inhibitor RIN1|ras interaction/interference protein 1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66332064, end_position_on_the_genomic_accession=66337133, orientation='minus', exon_count=11, OMIM=605965, no_hdr0=''),
  9611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9611, CurrentID=0, Status='live', Symbol='NCOR1', Aliases='N-CoR, N-CoR1, PPP1R109, TRAC1, hN-CoR', description='nuclear receptor corepressor 1', other_designations='nuclear receptor co-repressor 1|protein phosphatase 1, regulatory subunit 109|thyroid hormone- and retinoic acid receptor-associated corepressor 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16030094, end_position_on_the_genomic_accession=16216746, orientation='minus', exon_count=51, OMIM=600849, no_hdr0=''),
  9612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9612, CurrentID=0, Status='live', Symbol='NCOR2', Aliases='CTG26, N-CoR2, SMAP270, SMRT, SMRTE, SMRTE-tau, TNRC14, TRAC, TRAC-1, TRAC1', description='nuclear receptor corepressor 2', other_designations='CTG repeat protein 26|T3 receptor-associating factor|silencing mediator for retinoid and thyroid hormone receptors|thyroid-, retinoic-acid-receptor-associated corepressor', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124324411, end_position_on_the_genomic_accession=124567464, orientation='minus', exon_count=49, OMIM=600848, no_hdr0=''),
  140685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140685, CurrentID=0, Status='live', Symbol='ZBTB46', Aliases='BTBD4, BZEL, RINZF, ZNF340, dJ583P15.7, dJ583P15.8', description='zinc finger and BTB domain containing 46', other_designations='BTB (POZ) domain containing 4|BTB-ZF protein expressed in effector lymphocytes|BTB/POZ domain-containing protein 4|zinc finger protein 340', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63743668, end_position_on_the_genomic_accession=63832344, orientation='minus', exon_count=15, OMIM=614639, no_hdr0=''),
  140686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140686, CurrentID=0, Status='live', Symbol='WFDC3', Aliases='WAP14, dJ447F3.3', description='WAP four-disulfide core domain 3', other_designations='protease inhibitor WAP14|putative protease inhibitor WAP14|whey acidic protein 14', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45774208, end_position_on_the_genomic_accession=45791908, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  9615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9615, CurrentID=0, Status='live', Symbol='GDA', Aliases='CYPIN, GUANASE, NEDASIN', description='guanine deaminase', other_designations='GAH|cytoplasmic PSD95 interactor|guanine aminase|guanine aminohydrolase|p51-nedasin', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72114595, end_position_on_the_genomic_accession=72259059, orientation='plus', exon_count=25, OMIM=139260, no_hdr0=''),
  9616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9616, CurrentID=0, Status='live', Symbol='RNF7', Aliases='CKBBP1, ROC2, SAG', description='ring finger protein 7', other_designations='CKII beta-binding protein 1|rbx2|regulator of cullins 2|sensitive to apoptosis gene protein|sensitive to apoptosis, zinc RING finger protein SAG, regulator of cullins 2|zinc RING finger protein SAG', map_location='3q22-q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141738209, end_position_on_the_genomic_accession=141746803, orientation='plus', exon_count=4, OMIM=603863, no_hdr0=''),
  140689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140689, CurrentID=0, Status='live', Symbol='CBLN4', Aliases='CBLNL1', description='cerebellin 4 precursor', other_designations='cerebellin precursor-like 1|cerebellin-like glycoprotein 1', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55997357, end_position_on_the_genomic_accession=56005474, orientation='minus', exon_count=4, OMIM=615029, no_hdr0=''),
  9618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9618, CurrentID=0, Status='live', Symbol='TRAF4', Aliases='CART1, MLN62, RNF83', description='TNF receptor associated factor 4', other_designations='MLN 62|RING finger protein 83|TRAF4 variant 6|cysteine-rich domain associated with RING and Traf domains protein 1|cysteine-rich domain associated with ring and TRAF domain|malignant 62|metastatic lymph node gene 62 protein|tumor necrosis receptor-associated factor 4A', map_location='17q11-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28744005, end_position_on_the_genomic_accession=28750958, orientation='plus', exon_count=8, OMIM=602464, no_hdr0=''),
  9619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9619, CurrentID=0, Status='live', Symbol='ABCG1', Aliases='ABC8, WHITE1', description='ATP binding cassette subfamily G member 1', other_designations='ABC transporter 8|ATP-binding cassette transporter 8|ATP-binding cassette transporter member 1 of subfamily G|ATP-binding cassette, sub-family G (WHITE), member 1|homolog of Drosophila white|white protein homolog (ATP-binding cassette transporter 8)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42199689, end_position_on_the_genomic_accession=42304387, orientation='plus', exon_count=20, OMIM=603076, no_hdr0=''),
  9620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9620, CurrentID=0, Status='live', Symbol='CELSR1', Aliases='ADGRC1, CDHF9, FMI2, HFMI2, ME2', description='cadherin, EGF LAG seven-pass G-type receptor 1', other_designations='adhesion G protein-coupled receptor C1|cadherin family member 9|cadherin, EGF LAG seven-pass G-type receptor 1 (flamingo homolog, Drosophila)|flamingo homolog 2|protocadherin flamingo 2', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46360834, end_position_on_the_genomic_accession=46537502, orientation='minus', exon_count=38, OMIM=604523, no_hdr0=''),
  9622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9622, CurrentID=0, Status='live', Symbol='KLK4', Aliases='AI2A1, ARM1, EMSP, EMSP1, KLK-L1, PRSS17, PSTS, kallikrein', description='kallikrein related peptidase 4', other_designations='androgen-regulated message 1|enamel matrix serine protease 1|enamel matrix serine proteinase 1|kallikrein-like protein 1|kallikrein-related peptidase 4|prostase|serine protease 17', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50906352, end_position_on_the_genomic_accession=50910749, orientation='minus', exon_count=6, OMIM=603767, no_hdr0=''),
  9623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9623, CurrentID=0, Status='live', Symbol='TCL1B', Aliases='SYN-1, TML1', description='T-cell leukemia/lymphoma 1B', other_designations='T-cell lymphoma/leukemia 1B|TCL1/ MTCP1-like 1|oncogene TCL-1B|syncytiotrophoblast-specific protein', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95686417, end_position_on_the_genomic_accession=95692643, orientation='plus', exon_count=4, OMIM=603769, no_hdr0=''),
  140696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140696, CurrentID=0, Status='live', Symbol='RPL37P1', Aliases='RPL37_9_1698, bA563A22B.2', description='ribosomal protein L37 pseudogene 1', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35588325, end_position_on_the_genomic_accession=35588608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9625, CurrentID=0, Status='live', Symbol='AATK', Aliases='AATYK, AATYK1, LMR1, LMTK1, PPP1R77, p35BP', description='apoptosis-associated tyrosine kinase', other_designations='CDK5-binding protein|brain apoptosis-associated tyrosine kinase|lemur tyrosine kinase 1|p35-binding protein|protein phosphatase 1, regulatory subunit 77', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81117295, end_position_on_the_genomic_accession=81166072, orientation='minus', exon_count=21, OMIM=605276, no_hdr0=''),
  9626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9626, CurrentID=0, Status='live', Symbol='GUCA1C', Aliases='GCAP3', description='guanylate cyclase activator 1C', other_designations='GCAP 3', map_location='3q13.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108907644, end_position_on_the_genomic_accession=108955188, orientation='minus', exon_count=5, OMIM=605128, no_hdr0=''),
  9627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9627, CurrentID=0, Status='live', Symbol='SNCAIP', Aliases='SYPH1, Sph1', description='synuclein alpha interacting protein', other_designations='synuclein, alpha interacting protein', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122311349, end_position_on_the_genomic_accession=122464099, orientation='plus', exon_count=19, OMIM=603779, no_hdr0=''),
  9628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9628, CurrentID=0, Status='live', Symbol='RGS6', Aliases='GAP', description='regulator of G-protein signaling 6', other_designations='GTPase activating protein|H_DJ1108A12.1|S914|WUGSC:H_DJ0283M22.1|WUGSC:H_DJ1108A12.1|regulator of G-protein signalling 6', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71932100, end_position_on_the_genomic_accession=72566530, orientation='plus', exon_count=33, OMIM=603894, no_hdr0=''),
  9629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9629, CurrentID=0, Status='live', Symbol='CLCA3P', Aliases='CLCA3', description='chloride channel accessory 3, pseudogene', other_designations='chloride channel accessory 3 (pseudogene)|chloride channel regulator 3 pseudogene|chloride channel, calcium activated, family member 3', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86634276, end_position_on_the_genomic_accession=86655376, orientation='plus', exon_count=15, OMIM=604337, no_hdr0=''),
  9630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9630, CurrentID=0, Status='live', Symbol='GNA14', Aliases='', description='guanine nucleotide binding protein (G protein), alpha 14', other_designations='G-protein subunit alpha-14|g alpha-14|guanine nucleotide-binding protein 14', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77423079, end_position_on_the_genomic_accession=77648316, orientation='minus', exon_count=8, OMIM=604397, no_hdr0=''),
  9631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9631, CurrentID=0, Status='live', Symbol='NUP155', Aliases='ATFB15, N155', description='nucleoporin 155kDa', other_designations='155 kDa nucleoporin|nucleoporin 155kD', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37291633, end_position_on_the_genomic_accession=37371455, orientation='minus', exon_count=36, OMIM=606694, no_hdr0=''),
  9632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9632, CurrentID=0, Status='live', Symbol='SEC24C', Aliases='', description='SEC24 homolog C, COPII coat complex component', other_designations='SEC24 family member C|SEC24 family, member C|SEC24 related gene family, member C|SEC24-related protein C', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73744281, end_position_on_the_genomic_accession=73772175, orientation='plus', exon_count=26, OMIM=607185, no_hdr0=''),
  9633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9633, CurrentID=0, Status='live', Symbol='MTL5', Aliases='CXCDC2, MTLT, TESMIN', description='metallothionein-like 5, testis-specific (tesmin)', other_designations='CXC domain containing 2|testis-specific metallothionein-like protein', map_location='11q13.2-q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68706674, end_position_on_the_genomic_accession=68751564, orientation='minus', exon_count=12, OMIM=604374, no_hdr0=''),
  140706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140706, CurrentID=0, Status='live', Symbol='CCM2L', Aliases='C20orf160, dJ310O13.5', description='cerebral cavernous malformation 2-like', other_designations='CCM2-like', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32010379, end_position_on_the_genomic_accession=32032182, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  9635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9635, CurrentID=0, Status='live', Symbol='CLCA2', Aliases='CACC, CACC3, CLCRG2, CaCC-3', description='chloride channel accessory 2', other_designations='CLCA family member 2, chloride channel regulator|calcium-activated chloride channel family member 2|calcium-activated chloride channel protein 3|calcium-activated chloride channel-2|chloride channel regulator 2|chloride channel, calcium activated, family member 2|hCLCA2|hCaCC-3', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86424086, end_position_on_the_genomic_accession=86456558, orientation='plus', exon_count=15, OMIM=604003, no_hdr0=''),
  9636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9636, CurrentID=0, Status='live', Symbol='ISG15', Aliases='G1P2, IFI15, IMD38, IP17, UCRP, hUCRP', description='ISG15 ubiquitin-like modifier', other_designations='interferon, alpha-inducible protein (clone IFI-15K)|interferon-induced 17-kDa/15-kDa protein|interferon-stimulated protein, 15 kDa|ubiquitin cross-reactive protein', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1013467, end_position_on_the_genomic_accession=1014540, orientation='plus', exon_count=2, OMIM=147571, no_hdr0=''),
  9637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9637, CurrentID=0, Status='live', Symbol='FEZ2', Aliases='HUM3CL', description='fasciculation and elongation protein zeta 2 (zygin II)', other_designations='pre-T/NK cell associated protein (3Cl)|zygin 2|zygin II|zygin-2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36552239, end_position_on_the_genomic_accession=36598190, orientation='minus', exon_count=10, OMIM=604826, no_hdr0=''),
  9638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9638, CurrentID=0, Status='live', Symbol='FEZ1', Aliases='', description='fasciculation and elongation protein zeta 1', other_designations='fasciculation and elongation protein zeta 1 (zygin I)|zygin I|zygin-1', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125445745, end_position_on_the_genomic_accession=125496310, orientation='minus', exon_count=12, OMIM=604825, no_hdr0=''),
  9639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9639, CurrentID=0, Status='live', Symbol='ARHGEF10', Aliases='GEF10, SNCV', description='Rho guanine nucleotide exchange factor 10', other_designations='Rho guanine nucleotide exchange factor (GEF) 10', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1823983, end_position_on_the_genomic_accession=1958641, orientation='plus', exon_count=31, OMIM=608136, no_hdr0=''),
  9640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9640, CurrentID=0, Status='live', Symbol='ZNF592', Aliases='CAMOS, SCAR5', description='zinc finger protein 592', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84748561, end_position_on_the_genomic_accession=84806432, orientation='plus', exon_count=13, OMIM=613624, no_hdr0=''),
  9641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9641, CurrentID=0, Status='live', Symbol='IKBKE', Aliases='IKK-E, IKK-i, IKKE, IKKI', description='inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase epsilon', other_designations='I-kappa-B kinase epsilon|IKK-epsilon|IKK-related kinase epsilon|inducible I kappa-B kinase|inducible IkappaB kinase', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206470243, end_position_on_the_genomic_accession=206496890, orientation='plus', exon_count=22, OMIM=605048, no_hdr0=''),
  140714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140714, CurrentID=0, Status='live', Symbol='RPS27AP3', Aliases='RPS27AP, RPS27A_8_1699, bA332A4.3', description='ribosomal protein S27a pseudogene 3', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37049238, end_position_on_the_genomic_accession=37049751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9643, CurrentID=0, Status='live', Symbol='MORF4L2', Aliases='MORFL2, MRGX', description='mortality factor 4 like 2', other_designations='MORF-related gene X protein|MSL3-2 protein|protein MSL3-2|transcription factor-like protein MRGX', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103675498, end_position_on_the_genomic_accession=103688158, orientation='minus', exon_count=6, OMIM=300409, no_hdr0=''),
  9644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9644, CurrentID=0, Status='live', Symbol='SH3PXD2A', Aliases='FISH, SH3MD1, TKS5', description='SH3 and PX domains 2A', other_designations='SH3 multiple domains 1|adapter protein TKS5|adaptor protein TKS5|five SH3 domain-containing protein|tyrosine kinase substrate with five SH3 domains', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103594027, end_position_on_the_genomic_accession=103855482, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  9645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9645, CurrentID=0, Status='live', Symbol='MICAL2', Aliases='MICAL-2PV1, MICAL2PV2, MICAL2', description='microtubule associated monooxygenase, calponin and LIM domain containing 2', other_designations='flavoprotein oxidoreductase MICAL2|microtubule associated monoxygenase, calponin and LIM domain containing 2|molecule interacting with CasL protein 2', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12110576, end_position_on_the_genomic_accession=12281016, orientation='plus', exon_count=32, OMIM=608881, no_hdr0=''),
  9646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9646, CurrentID=0, Status='live', Symbol='CTR9', Aliases='SH2BP1, TSBP, p150, p150TSP', description='CTR9 homolog, Paf1/RNA polymerase II complex component', other_designations='Ctr9, Paf1/RNA polymerase II complex component, homolog|SH2 domain binding protein 1 (tetratricopeptide repeat containing)|SH2 domain-binding protein 1|TPR-containing, SH2-binding phosphoprotein', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10750987, end_position_on_the_genomic_accession=10779755, orientation='plus', exon_count=25, OMIM=609366, no_hdr0=''),
  9647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9647, CurrentID=0, Status='live', Symbol='PPM1F', Aliases='CAMKP, CaMKPase, FEM-2, POPX2, hFEM-2', description='protein phosphatase, Mg2+/Mn2+ dependent 1F', other_designations='Ca(2+)/calmodulin-dependent protein kinase phosphatase|CaM-kinase phosphatase|PP2C phosphatase|partner of PIX 2|protein fem-2 homolog|protein phosphatase 1F (PP2C domain containing)|protein phosphatase, Mg2+/Mn2+ dependent, 1F', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21919419, end_position_on_the_genomic_accession=21952878, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  9648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9648, CurrentID=0, Status='live', Symbol='GCC2', Aliases='GCC185, RANBP2L4, REN53', description='GRIP and coiled-coil domain containing 2', other_designations='185 kDa Golgi coiled-coil protein|CLL-associated antigen KW-11|CTCL tumor antigen se1-1|GCC protein, 185-kD|Golgi coiled-coil protein GCC185|Ran-binding protein 2-like 4|renal carcinoma antigen NY-REN-53', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108449121, end_position_on_the_genomic_accession=108509398, orientation='plus', exon_count=25, OMIM=612711, no_hdr0=''),
  9649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9649, CurrentID=0, Status='live', Symbol='RALGPS1', Aliases='RALGEF2A, RALGPS1', description='Ral GEF with PH domain and SH3 binding motif 1', other_designations='Ral guanine nucleotide exchange factor RalGPS1A|RalGEF 2|ral GEF with PH domain and SH3-binding motif 1|ral guanine nucleotide exchange factor 2|ralA exchange factor RalGPS1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126914774, end_position_on_the_genomic_accession=127223166, orientation='plus', exon_count=26, OMIM=614444, no_hdr0=''),
  9650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9650, CurrentID=0, Status='live', Symbol='MTFR1', Aliases='CHPPR, FAM54A2', description='mitochondrial fission regulator 1', other_designations='chondrocyte protein with a poly-proline region', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65643873, end_position_on_the_genomic_accession=65771252, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  9651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9651, CurrentID=0, Status='live', Symbol='PLCH2', Aliases='PLC-L4, PLC-eta2, PLCL4, PLCeta2', description='phospholipase C eta 2', other_designations='phosphoinositide phospholipase C-eta-2|phosphoinositide phospholipase C-like 4|phospholipase C, eta 2|phospholipase C-eta2|phospholipase C-like 4|phospholipase C-like protein 4', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2415782, end_position_on_the_genomic_accession=2505530, orientation='plus', exon_count=29, OMIM=612836, no_hdr0=''),
  9652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9652, CurrentID=0, Status='live', Symbol='TTC37', Aliases='KIAA0372, Ski3, THES', description='tetratricopeptide repeat domain 37', other_designations='SKI3 homolog|TPR repeat protein 37|thespin|tricho-hepatic-enteric syndrome protein', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95463895, end_position_on_the_genomic_accession=95555005, orientation='minus', exon_count=43, OMIM=614589, no_hdr0=''),
  9653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9653, CurrentID=0, Status='live', Symbol='HS2ST1', Aliases='dJ604K5.2', description='heparan sulfate 2-O-sulfotransferase 1', other_designations='2-O-sulfotransferase|2OST', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86914652, end_position_on_the_genomic_accession=87109998, orientation='plus', exon_count=7, OMIM=604844, no_hdr0=''),
  9654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9654, CurrentID=0, Status='live', Symbol='TTLL4', Aliases='', description='tubulin tyrosine ligase like 4', other_designations='tubulin tyrosine ligase-like family member 4|tubulin tyrosine ligase-like family, member 4|tubulin--tyrosine ligase-like protein 4', map_location='2p24.3-p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218710835, end_position_on_the_genomic_accession=218755416, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  9655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9655, CurrentID=0, Status='live', Symbol='SOCS5', Aliases='CIS6, CISH6, Cish5, SOCS-5', description='suppressor of cytokine signaling 5', other_designations='CIS-6|cytokine-inducible SH2 protein 6|cytokine-inducible SH2-containing protein 5', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46698960, end_position_on_the_genomic_accession=46762788, orientation='plus', exon_count=3, OMIM=607094, no_hdr0=''),
  9656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9656, CurrentID=0, Status='live', Symbol='MDC1', Aliases='NFBD1', description='mediator of DNA damage checkpoint 1', other_designations='homologue to Drosophila photoreceptor protein calphotin|mediator of DNA-damage checkpoint 1|nuclear factor with BRCT domains 1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30699807, end_position_on_the_genomic_accession=30717966, orientation='minus', exon_count=18, OMIM=607593, no_hdr0=''),
  9657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9657, CurrentID=0, Status='live', Symbol='IQCB1', Aliases='NPHP5, PIQ, SLSN5', description='IQ motif containing B1', other_designations='IQ calmodulin-binding motif containing 1|nephrocystin 5|p53 and DNA damage-regulated IQ motif protein', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121769763, end_position_on_the_genomic_accession=121835079, orientation='minus', exon_count=15, OMIM=609237, no_hdr0=''),
  9658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9658, CurrentID=0, Status='live', Symbol='ZNF516', Aliases='HsT287', description='zinc finger protein 516', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76357682, end_position_on_the_genomic_accession=76495234, orientation='minus', exon_count=11, OMIM=615114, no_hdr0=''),
  9659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9659, CurrentID=0, Status='live', Symbol='PDE4DIP', Aliases='CMYA2, MMGL', description='phosphodiesterase 4D interacting protein', other_designations='cardiomyopathy-associated protein 2|myomegalin/phosphodiesterase 4D interacting protein variant 8', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148808167, end_position_on_the_genomic_accession=149033019, orientation='plus', exon_count=56, OMIM=608117, no_hdr0=''),
  140732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140732, CurrentID=0, Status='live', Symbol='SUN5', Aliases='SPAG4L, TSARG4, dJ726C3.1', description='Sad1 and UNC84 domain containing 5', other_designations='sad1 and UNC84 domain-containing protein 5|sperm-associated antigen 4-like protein|testis and spermatogenesis related gene 4|testis and spermatogenesis-related gene 4 protein', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32983773, end_position_on_the_genomic_accession=33004449, orientation='minus', exon_count=15, OMIM=613942, no_hdr0=''),
  140733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140733, CurrentID=0, Status='live', Symbol='MACROD2', Aliases='C20orf133', description='MACRO domain containing 2', other_designations='MACRO domain-containing protein 2|[Protein ADP-ribosylglutamate] hydrolase', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13995500, end_position_on_the_genomic_accession=16053197, orientation='plus', exon_count=20, OMIM=611567, no_hdr0=''),
  9662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9662, CurrentID=0, Status='live', Symbol='CEP135', Aliases='CEP4, KIAA0635, MCPH8', description='centrosomal protein 135kDa', other_designations='centrosomal protein 4|centrosome protein cep135', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55948808, end_position_on_the_genomic_accession=56033363, orientation='plus', exon_count=26, OMIM=611423, no_hdr0=''),
  9663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9663, CurrentID=0, Status='live', Symbol='LPIN2', Aliases='', description='lipin 2', other_designations='lipin-2', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2916994, end_position_on_the_genomic_accession=3013315, orientation='minus', exon_count=22, OMIM=605519, no_hdr0=''),
  285600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285600, CurrentID=0, Status='live', Symbol='KIAA0825', Aliases='C5orf36', description='KIAA0825', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94150851, end_position_on_the_genomic_accession=94618604, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  9666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9666, CurrentID=0, Status='live', Symbol='DZIP3', Aliases='PPP1R66, UURF2, hRUL138', description='DAZ interacting zinc finger protein 3', other_designations='DAZ interacting protein 3, zinc finger|DAZ-interacting protein 3|RNA-binding RING-H2 protein-ubiquitin ligase|RNA-binding ubiquitin ligase of 138 kDa|UURF2 ubiquitin ligase|human RNA-binding ubiquitin ligase of 138 kDa|protein phosphatase 1, regulatory subunit 66|zinc finger DAZ interacting protein 3', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108589490, end_position_on_the_genomic_accession=108694846, orientation='plus', exon_count=36, OMIM=608672, no_hdr0=''),
  9667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9667, CurrentID=0, Status='live', Symbol='SAFB2', Aliases='', description='scaffold attachment factor B2', other_designations='SAF-B2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5586993, end_position_on_the_genomic_accession=5622927, orientation='minus', exon_count=22, OMIM=608066, no_hdr0=''),
  9668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9668, CurrentID=0, Status='live', Symbol='ZNF432', Aliases='', description='zinc finger protein 432', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52031378, end_position_on_the_genomic_accession=52048820, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  9669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9669, CurrentID=0, Status='live', Symbol='EIF5B', Aliases='IF2', description='eukaryotic translation initiation factor 5B', other_designations='eIF-5B|translation initiation factor IF-2|translation initiation factor IF2', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99337371, end_position_on_the_genomic_accession=99400265, orientation='plus', exon_count=24, OMIM=606086, no_hdr0=''),
  9670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9670, CurrentID=0, Status='live', Symbol='IPO13', Aliases='IMP13, KAP13, LGL2, RANBP13', description='importin 13', other_designations='Ran binding protein 13|karyopherin 13|karyopherin-13|late gestation lung 2|ran-binding protein 13', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43946806, end_position_on_the_genomic_accession=43968022, orientation='plus', exon_count=21, OMIM=610411, no_hdr0=''),
  9671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9671, CurrentID=0, Status='live', Symbol='WSCD2', Aliases='', description='WSC domain containing 2', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108129343, end_position_on_the_genomic_accession=108250537, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  9672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9672, CurrentID=0, Status='live', Symbol='SDC3', Aliases='SDCN, SYND3', description='syndecan 3', other_designations='N-syndecan|syndecan neural type|syndecan proteoglycan 3', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30869466, end_position_on_the_genomic_accession=30909735, orientation='minus', exon_count=10, OMIM=186357, no_hdr0=''),
  9673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9673, CurrentID=0, Status='live', Symbol='SLC25A44', Aliases='', description='solute carrier family 25 member 44', other_designations='solute carrier family 25, member 44', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156193932, end_position_on_the_genomic_accession=156212796, orientation='plus', exon_count=6, OMIM=610824, no_hdr0=''),
  9674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9674, CurrentID=0, Status='live', Symbol='KIAA0040', Aliases='', description='KIAA0040', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175156987, end_position_on_the_genomic_accession=175193093, orientation='minus', exon_count=6, OMIM=616696, no_hdr0=''),
  9675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9675, CurrentID=0, Status='live', Symbol='TTI1', Aliases='KIAA0406, smg-10', description='TELO2 interacting protein 1', other_designations='TEL2-interacting protein 1 homolog|Tel two interacting protein 1|smg-10 homolog, nonsense mediated mRNA decay factor', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37983007, end_position_on_the_genomic_accession=38033468, orientation='minus', exon_count=11, OMIM=614425, no_hdr0=''),
  140748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140748, CurrentID=0, Status='live', Symbol='RPL17P1', Aliases='RPL17_18_1690, dJ1178H5.1', description='ribosomal protein L17 pseudogene 1', other_designations='', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20253303, end_position_on_the_genomic_accession=20253838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9677, CurrentID=0, Status='live', Symbol='PPIP5K1', Aliases='HISPPD2A, IP6K, IPS1, VIP1, hsVIP1', description='diphosphoinositol pentakisphosphate kinase 1', other_designations='IP6 kinase|VIP1 homolog|histidine acid phosphatase domain containing 2A|histidine acid phosphatase domain-containing protein 2A|inositol pyrophosphate synthase 1|insP6 and PP-IP5 kinase 1', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43533462, end_position_on_the_genomic_accession=43590272, orientation='minus', exon_count=34, OMIM=610979, no_hdr0=''),
  9678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9678, CurrentID=0, Status='live', Symbol='PHF14', Aliases='', description='PHD finger protein 14', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10957892, end_position_on_the_genomic_accession=11169828, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  9679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9679, CurrentID=0, Status='live', Symbol='FAM53B', Aliases='KIAA0140, bA12J10.2, smp', description='family with sequence similarity 53 member B', other_designations='family with sequence similarity 53, member B|simplet homolog', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124619294, end_position_on_the_genomic_accession=124744361, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  140752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140752, CurrentID=0, Status='live', Symbol='RPS27AP2', Aliases='RPS27A_7_1686, dJ531H16.4', description='ribosomal protein S27a pseudogene 2', other_designations='', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17516793, end_position_on_the_genomic_accession=17517342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9681, CurrentID=0, Status='live', Symbol='DEPDC5', Aliases='DEP.5, FFEVF', description='DEP domain containing 5', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31753951, end_position_on_the_genomic_accession=31907034, orientation='plus', exon_count=47, OMIM=614191, no_hdr0=''),
  9682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9682, CurrentID=0, Status='live', Symbol='KDM4A', Aliases='JHDM3A, JMJD2, JMJD2A, TDRD14A', description='lysine (K)-specific demethylase 4A', other_designations='jmjC domain-containing histone demethylation protein 3A|jumonji C domain-containing histone demethylase 3A|jumonji domain containing 2|jumonji domain containing 2A|jumonji domain-containing protein 2A|tudor domain containing 14A', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43650126, end_position_on_the_genomic_accession=43705518, orientation='plus', exon_count=25, OMIM=609764, no_hdr0=''),
  9683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9683, CurrentID=0, Status='live', Symbol='N4BP1', Aliases='', description='NEDD4 binding protein 1', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48538726, end_position_on_the_genomic_accession=48610214, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  9684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9684, CurrentID=0, Status='live', Symbol='LRRC14', Aliases='LRRC14A', description='leucine rich repeat containing 14', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144517965, end_position_on_the_genomic_accession=144525178, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  9685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9685, CurrentID=0, Status='live', Symbol='CLINT1', Aliases='CLINT, ENTH, EPN4, EPNR', description='clathrin interactor 1', other_designations='clathrin interacting protein localized in the trans-Golgi region|enthoprotin|epsin 4|epsin-related protein|epsinR', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157785743, end_position_on_the_genomic_accession=157859175, orientation='minus', exon_count=14, OMIM=607265, no_hdr0=''),
  9686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9686, CurrentID=0, Status='live', Symbol='VGLL4', Aliases='VGL-4', description='vestigial like family member 4', other_designations='Vestigial-like 4|vestigial like 4|vestigial-like family member 4', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11556067, end_position_on_the_genomic_accession=11720768, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  9687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9687, CurrentID=0, Status='live', Symbol='GREB1', Aliases='', description='growth regulation by estrogen in breast cancer 1', other_designations='gene regulated by estrogen in breast cancer|gene regulated in breast cancer 1 protein', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11534116, end_position_on_the_genomic_accession=11642788, orientation='plus', exon_count=39, OMIM=611736, no_hdr0=''),
  9688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9688, CurrentID=0, Status='live', Symbol='NUP93', Aliases='NIC96', description='nucleoporin 93kDa', other_designations='93 kDa nucleoporin|nucleoporin Nup93', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56730105, end_position_on_the_genomic_accession=56844950, orientation='plus', exon_count=25, OMIM=614351, no_hdr0=''),
  9689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9689, CurrentID=0, Status='live', Symbol='BZW1', Aliases='BZAP45, Nbla10236', description='basic leucine zipper and W2 domains 1', other_designations='basic leucine-zipper protein BZAP45|putative protein product of Nbla10236', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200811544, end_position_on_the_genomic_accession=200823846, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  9690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9690, CurrentID=0, Status='live', Symbol='UBE3C', Aliases='HECTH2', description='ubiquitin protein ligase E3C', other_designations='ubiquitin-protein isopeptide ligase (E3)', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157138959, end_position_on_the_genomic_accession=157269372, orientation='plus', exon_count=24, OMIM=614454, no_hdr0=''),
  100271695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271695, CurrentID=0, Status='live', Symbol='MS3', Aliases='', description='Multiple sclerosis, susceptibility to, 3', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612595, no_hdr0=''),
  9692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9692, CurrentID=0, Status='live', Symbol='KIAA0391', Aliases='MRPP3, PRORP', description='KIAA0391', other_designations='mitochondrial RNase P protein 3|mitochondrial RNase P subunit 3|proteinaceous RNase P', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35121283, end_position_on_the_genomic_accession=35274078, orientation='plus', exon_count=10, OMIM=609947, no_hdr0=''),
  9693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9693, CurrentID=0, Status='live', Symbol='RAPGEF2', Aliases='CNrasGEF, NRAPGEP, PDZ-GEF1, PDZGEF1, RA-GEF, RA-GEF-1, Rap-GEP, nRap GEP', description='Rap guanine nucleotide exchange factor 2', other_designations='PDZ domain containing guanine nucleotide exchange factor (GEF) 1|PDZ domain-containing guanine nucleotide exchange factor 1|RA(Ras/Rap1A-associating)-GEF|Rap guanine nucleotide exchange factor (GEF) 2|cyclic nucleotide ras GEF|neural RAP guanine nucleotide exchange protein|ras/Rap1-associating GEF-1', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159104071, end_position_on_the_genomic_accession=159360162, orientation='plus', exon_count=35, OMIM=609530, no_hdr0=''),
  140766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140766, CurrentID=0, Status='live', Symbol='ADAMTS14', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 14', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 14|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 14|metalloprotease-disintegrin protease', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70672539, end_position_on_the_genomic_accession=70762441, orientation='plus', exon_count=25, OMIM=607506, no_hdr0=''),
  9695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9695, CurrentID=0, Status='live', Symbol='EDEM1', Aliases='EDEM', description='ER degradation enhancer, mannosidase alpha-like 1', other_designations='ER degradation-enhancing alpha-mannosidase-like 1', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5187674, end_position_on_the_genomic_accession=5219965, orientation='plus', exon_count=13, OMIM=607673, no_hdr0=''),
  9696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9696, CurrentID=0, Status='live', Symbol='CROCC', Aliases='ROLT', description='ciliary rootlet coiled-coil, rootletin', other_designations='Tax1-binding protein 2|ciliary rootlet coiled-coil protein|rootletin, ciliary rootlet protein', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16895707, end_position_on_the_genomic_accession=16972979, orientation='plus', exon_count=41, OMIM=615776, no_hdr0=''),
  9697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9697, CurrentID=0, Status='live', Symbol='TRAM2', Aliases='', description='translocation associated membrane protein 2', other_designations='TRAM-like protein', map_location='6p21.1-p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52497402, end_position_on_the_genomic_accession=52577064, orientation='minus', exon_count=12, OMIM=608485, no_hdr0=''),
  9698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9698, CurrentID=0, Status='live', Symbol='PUM1', Aliases='HSPUM, PUMH, PUMH1, PUML1', description='pumilio RNA binding family member 1', other_designations='pumilio RNA-binding family member 1|pumilio-1', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30931506, end_position_on_the_genomic_accession=31065717, orientation='minus', exon_count=22, OMIM=607204, no_hdr0=''),
  9699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9699, CurrentID=0, Status='live', Symbol='RIMS2', Aliases='OBOE, RAB3IP3, RIM2', description='regulating synaptic membrane exocytosis 2', other_designations='RAB3 interacting protein 3|RIM 2|Rab3-interacting protein|non-small cell lung cancer RimL3a protein|non-small cell lung cancer RimL3c protein|nuclear protein|rab-3-interacting molecule 2|rab-3-interacting protein 3|rab3-interacting molecule 2', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103500642, end_position_on_the_genomic_accession=104254429, orientation='plus', exon_count=43, OMIM=606630, no_hdr0=''),
  9700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9700, CurrentID=0, Status='live', Symbol='ESPL1', Aliases='ESP1, SEPA', description='extra spindle pole bodies like 1, separase', other_designations='caspase-like protein ESPL1|extra spindle pole bodies 1, separase|extra spindle pole bodies homolog 1|extra spindle poles like 1|extra spindle poles-like 1 protein|separin, cysteine protease', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53268266, end_position_on_the_genomic_accession=53293643, orientation='plus', exon_count=32, OMIM=604143, no_hdr0=''),
  9701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9701, CurrentID=0, Status='live', Symbol='PPP6R2', Aliases='KIAA0685, PP6R2, SAP190, SAPS2', description='protein phosphatase 6 regulatory subunit 2', other_designations='SAPS domain family, member 2|protein phosphatase 6, regulatory subunit 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50343317, end_position_on_the_genomic_accession=50445090, orientation='plus', exon_count=28, OMIM=610877, no_hdr0=''),
  9702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9702, CurrentID=0, Status='live', Symbol='CEP57', Aliases='MVA2, PIG8, TSP57', description='centrosomal protein 57kDa', other_designations='FGF2-interacting protein|proliferation-inducing protein 8|testis-specific protein 57|translokin', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95790461, end_position_on_the_genomic_accession=95832693, orientation='plus', exon_count=12, OMIM=607951, no_hdr0=''),
  9703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9703, CurrentID=0, Status='live', Symbol='KIAA0100', Aliases='BCOX, BCOX1, CT101', description='KIAA0100', other_designations='U937-associated antigen|antigen MLAA-22|breast cancer overexpressed gene 1|breast cancer-overexpressed gene 1 protein|cancer/testis antigen 101', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28614440, end_position_on_the_genomic_accession=28645159, orientation='minus', exon_count=39, OMIM=610664, no_hdr0=''),
  9704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9704, CurrentID=0, Status='live', Symbol='DHX34', Aliases='DDX34, HRH1', description='DEAH (Asp-Glu-Ala-His) box polypeptide 34', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 34|DEAH box protein 34|probable ATP-dependent helicase DHX34', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47349281, end_position_on_the_genomic_accession=47382704, orientation='plus', exon_count=21, OMIM=615475, no_hdr0=''),
  9705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9705, CurrentID=0, Status='live', Symbol='ST18', Aliases='NZF3, ZC2HC10, ZNF387', description='suppression of tumorigenicity 18, zinc finger', other_designations='neural zinc finger transcription factor 3|suppression of tumorigenicity 18 (breast carcinoma) (zinc finger protein)|zinc finger protein 387', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52110832, end_position_on_the_genomic_accession=52409902, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  9706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9706, CurrentID=0, Status='live', Symbol='ULK2', Aliases='ATG1B, Unc51.2', description='unc-51 like autophagy activating kinase 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19770828, end_position_on_the_genomic_accession=19867926, orientation='minus', exon_count=29, OMIM=608650, no_hdr0=''),
  9708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9708, CurrentID=0, Status='live', Symbol='PCDHGA8', Aliases='PCDH-GAMMA-A8', description='protocadherin gamma subfamily A, 8', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141391916, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606295, no_hdr0=''),
  100271698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271698, CurrentID=0, Status='live', Symbol='ADIPQTL4', Aliases='', description='Adiponectin, serum level of, QTL4', other_designations='', map_location='11q23-q24', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612629, no_hdr0=''),
  9710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9710, CurrentID=0, Status='live', Symbol='KIAA0355', Aliases='', description='KIAA0355', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34254551, end_position_on_the_genomic_accession=34355566, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  9711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9711, CurrentID=0, Status='live', Symbol='RUBCN', Aliases='KIAA0226, RUBICON, SCAR15', description='RUN domain and cysteine-rich domain containing, Beclin 1-interacting protein', other_designations='RUN domain protein as Beclin 1-interacting and cysteine-rich containing|baron|beclin-1 associated RUN domain containing protein|run domain Beclin-1 interacting and cystein-rich containing protein|rundataxin', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197671393, end_position_on_the_genomic_accession=197749727, orientation='minus', exon_count=23, OMIM=613516, no_hdr0=''),
  9712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9712, CurrentID=0, Status='live', Symbol='USP6NL', Aliases='RNTRE, TRE2NL-IT1, USP6NL', description='USP6 N-terminal like', other_designations='related to the N-terminus of tre', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11460510, end_position_on_the_genomic_accession=11611789, orientation='minus', exon_count=22, OMIM=605405, no_hdr0=''),
  9715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9715, CurrentID=0, Status='live', Symbol='FAM131B', Aliases='', description='family with sequence similarity 131 member B', other_designations='family with sequence similarity 131, member B', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143353400, end_position_on_the_genomic_accession=143382304, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  9716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9716, CurrentID=0, Status='live', Symbol='AQR', Aliases='IBP160, fSAP164', description='aquarius intron-binding spliceosomal factor', other_designations='aquarius homolog|functional spliceosome-associated protein 164|intron-binding protein of 160 kDa', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34856351, end_position_on_the_genomic_accession=34969794, orientation='minus', exon_count=35, OMIM=610548, no_hdr0=''),
  9717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9717, CurrentID=0, Status='live', Symbol='SEC14L5', Aliases='PRELID4B', description='SEC14-like lipid binding 5', other_designations='SEC14-like 5', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4958317, end_position_on_the_genomic_accession=5019155, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  9718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9718, CurrentID=0, Status='live', Symbol='ECE2', Aliases='', description='endothelin converting enzyme 2', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184249657, end_position_on_the_genomic_accession=184293031, orientation='plus', exon_count=22, OMIM=610145, no_hdr0=''),
  9719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9719, CurrentID=0, Status='live', Symbol='ADAMTSL2', Aliases='GPHYSD1', description='ADAMTS like 2', other_designations='ADAMTS-like 2|ADAMTSL-2', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133532164, end_position_on_the_genomic_accession=133575519, orientation='plus', exon_count=21, OMIM=612277, no_hdr0=''),
  9720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9720, CurrentID=0, Status='live', Symbol='CCDC144A', Aliases='', description='coiled-coil domain containing 144A', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16690235, end_position_on_the_genomic_accession=16774999, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  9721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9721, CurrentID=0, Status='live', Symbol='GPRIN2', Aliases='GRIN2, KIAA0514', description='G protein regulated inducer of neurite outgrowth 2', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46543963, end_position_on_the_genomic_accession=46556947, orientation='minus', exon_count=9, OMIM=611240, no_hdr0=''),
  9722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9722, CurrentID=0, Status='live', Symbol='NOS1AP', Aliases='6330408P19Rik, CAPON', description='nitric oxide synthase 1 adaptor protein', other_designations='C-terminal PDZ domain ligand of neuronal nitric oxide synthase (CAPON)|C-terminal PDZ ligand of neuronal nitric oxide synthase protein|ligand of neuronal nitric oxide synthase with carboxyl-terminal PDZ domain|nitric oxide synthase 1 (neuronal) adaptor protein', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162069791, end_position_on_the_genomic_accession=162370023, orientation='plus', exon_count=10, OMIM=605551, no_hdr0=''),
  9723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9723, CurrentID=0, Status='live', Symbol='SEMA3E', Aliases='M-SEMAH, M-SemaK, SEMAH, coll-5', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3E', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83363906, end_position_on_the_genomic_accession=83649163, orientation='minus', exon_count=19, OMIM=608166, no_hdr0=''),
  9724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9724, CurrentID=0, Status='live', Symbol='UTP14C', Aliases='2700066J21Rik, KIAA0266, UTP14B', description='UTP14, U3 small nucleolar ribonucleoprotein, homolog C (yeast)', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52024691, end_position_on_the_genomic_accession=52033600, orientation='plus', exon_count=2, OMIM=608969, no_hdr0=''),
  9725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9725, CurrentID=0, Status='live', Symbol='TMEM63A', Aliases='KIAA0792', description='transmembrane protein 63A', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225840570, end_position_on_the_genomic_accession=225882720, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  9726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9726, CurrentID=0, Status='live', Symbol='ZNF646', Aliases='', description='zinc finger protein 646', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31073756, end_position_on_the_genomic_accession=31083512, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9727, CurrentID=0, Status='live', Symbol='RAB11FIP3', Aliases='CART1, Rab11-FIP3', description='RAB11 family interacting protein 3 (class II)', other_designations='EF hands-containing Rab-interacting protein|FIP3-Rab11|MU-MB-17.148|PAC196A12.1|arfophilin-1|cytoplasmic adaptor for RAR and TR|eferin|rab11-family interacting protein 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=425592, end_position_on_the_genomic_accession=522485, orientation='plus', exon_count=18, OMIM=608738, no_hdr0=''),
  9728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9728, CurrentID=0, Status='live', Symbol='SECISBP2L', Aliases='SBP2L, SLAN', description='SECIS binding protein 2-like', other_designations='SECIS-binding protein 2-like', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48988638, end_position_on_the_genomic_accession=49046563, orientation='minus', exon_count=18, OMIM=615756, no_hdr0=''),
  140801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140801, CurrentID=0, Status='live', Symbol='RPL10L', Aliases='RPL10_5_1358', description='ribosomal protein L10 like', other_designations='ribosomal protein L10-like protein', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46651013, end_position_on_the_genomic_accession=46651825, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  9730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9730, CurrentID=0, Status='live', Symbol='VPRBP', Aliases='DCAF1', description='Vpr (HIV-1) binding protein', other_designations='DDB1 and CUL4 associated factor 1|DDB1- and CUL4-associated factor 1|HIV-1 Vpr-binding protein|Vpr-binding protein|serine/threonine-protein kinase VPRBP|vpr-interacting protein', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51395867, end_position_on_the_genomic_accession=51505618, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  140803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140803, CurrentID=0, Status='live', Symbol='TRPM6', Aliases='CHAK2, HMGX, HOMG, HOMG1, HSH', description='transient receptor potential cation channel, subfamily M, member 6', other_designations='channel kinase 2|melastatin-related TRP cation channel 6', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74722495, end_position_on_the_genomic_accession=74888094, orientation='minus', exon_count=47, OMIM=607009, no_hdr0=''),
  9732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9732, CurrentID=0, Status='live', Symbol='DOCK4', Aliases='', description='dedicator of cytokinesis 4', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111726108, end_position_on_the_genomic_accession=112206407, orientation='minus', exon_count=53, OMIM=607679, no_hdr0=''),
  9733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9733, CurrentID=0, Status='live', Symbol='SART3', Aliases='DSAP1, P100, RP11-13G14, TIP110, p110, p110(nrb)', description='squamous cell carcinoma antigen recognized by T-cells 3', other_designations='SART-3|hSART-3|p110 nuclear RNA-binding protein|squamous cell carcinoma antigen recognized by T cells 3|tat-interacting protein of 110 kDa', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108522214, end_position_on_the_genomic_accession=108561389, orientation='minus', exon_count=20, OMIM=611684, no_hdr0=''),
  9734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9734, CurrentID=0, Status='live', Symbol='HDAC9', Aliases='HD7, HD7b, HD9, HDAC, HDAC7, HDAC7BB, HDAC9FL, HDRP, MITR, HDAC9', description='histone deacetylase 9', other_designations='MEF-2 interacting transcription repressor (MITR) protein|histone deacetylase 4/5-related protein|histone deacetylase 7B', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18086942, end_position_on_the_genomic_accession=18999521, orientation='plus', exon_count=38, OMIM=606543, no_hdr0=''),
  9735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9735, CurrentID=0, Status='live', Symbol='KNTC1', Aliases='ROD', description='kinetochore associated 1', other_designations='Rough Deal homolog, centromere/kinetochore protein', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122527249, end_position_on_the_genomic_accession=122627311, orientation='plus', exon_count=66, OMIM=607363, no_hdr0=''),
  9736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9736, CurrentID=0, Status='live', Symbol='USP34', Aliases='', description='ubiquitin specific peptidase 34', other_designations='deubiquitinating enzyme 34|ubiquitin specific protease 34|ubiquitin thioesterase 34|ubiquitin thiolesterase 34|ubiquitin-specific-processing protease 34', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61187455, end_position_on_the_genomic_accession=61470714, orientation='minus', exon_count=80, OMIM=615295, no_hdr0=''),
  140809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140809, CurrentID=0, Status='live', Symbol='SRXN1', Aliases='C20orf139, Npn3, SRX, SRX1', description='sulfiredoxin 1', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=646615, end_position_on_the_genomic_accession=653370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  9738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9738, CurrentID=0, Status='live', Symbol='CCP110', Aliases='CP110, Cep110', description='centriolar coiled-coil protein 110kDa', other_designations='centriolar coiled coil protein 110kDa|centrosomal protein CP110|centrosomal protein of 110 kDa', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19523857, end_position_on_the_genomic_accession=19553408, orientation='plus', exon_count=17, OMIM=609544, no_hdr0=''),
  9739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9739, CurrentID=0, Status='live', Symbol='SETD1A', Aliases='KMT2F, Set1, Set1A', description='SET domain containing 1A', other_designations='SET domain-containing protein 1A|hSET1A|lysine N-methyltransferase 2F|set1/Ash2 histone methyltransferase complex subunit SET1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30948505, end_position_on_the_genomic_accession=30984664, orientation='plus', exon_count=24, OMIM=611052, no_hdr0=''),
  9741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9741, CurrentID=0, Status='live', Symbol='LAPTM4A', Aliases='HUMORF13, LAPTM4, MBNT, Mtrp', description='lysosomal protein transmembrane 4 alpha', other_designations='golgi 4-transmembrane-spanning transporter MTP|lysosomal-associated protein transmembrane 4 alpha|membrane nucleoside transporter', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20032650, end_position_on_the_genomic_accession=20052028, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  9742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9742, CurrentID=0, Status='live', Symbol='IFT140', Aliases='MZSDS, SRTD9, WDTC2, c305C8.4, c380F5.1, gs114', description='intraflagellar transport 140', other_designations='WD and tetratricopeptide repeats protein 2|intraflagellar transport 140 homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1510427, end_position_on_the_genomic_accession=1612108, orientation='minus', exon_count=41, OMIM=614620, no_hdr0=''),
  9743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9743, CurrentID=0, Status='live', Symbol='ARHGAP32', Aliases='GC-GAP, GRIT, PX-RICS, RICS, p200RhoGAP, p250GAP', description='Rho GTPase activating protein 32', other_designations='GAB-associated CDC42|GAB-associated Cdc42/Rac GTPase-activating protein|GTPase regulator interacting with TrkA|GTPase-activating protein for Cdc42 and Rac1|RhoGAP involved in the -catenin-N-cadherin and NMDA receptor signaling|brain-specific Rho GTP-ase-activating protein|brain-specific Rho GTPase-activating protein|rac GTPase activating protein|rho-type GTPase-activating protein 32|rho/Cdc42/Rac GTPase-activating protein RICS|rhoGAP involved in the beta-catenin-N-cadherin and NMDA receptor signaling', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128965060, end_position_on_the_genomic_accession=129267015, orientation='minus', exon_count=27, OMIM=608541, no_hdr0=''),
  9744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9744, CurrentID=0, Status='live', Symbol='ACAP1', Aliases='CENTB1', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 1', other_designations='Arf GAP with coiled coil, ANK repeat and PH domains 1|centaurin, beta 1|centaurin-beta-1|cnt-b1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7336511, end_position_on_the_genomic_accession=7351478, orientation='plus', exon_count=22, OMIM=607763, no_hdr0=''),
  9745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9745, CurrentID=0, Status='live', Symbol='ZNF536', Aliases='', description='zinc finger protein 536', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=30225044, end_position_on_the_genomic_accession=30713538, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  9746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9746, CurrentID=0, Status='live', Symbol='CLSTN3', Aliases='CDHR14, CSTN3, alcbeta', description='calsyntenin 3', other_designations='alc-beta|alcadein beta|alcadein-beta|cadherin-related family member 14', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7130371, end_position_on_the_genomic_accession=7158945, orientation='plus', exon_count=18, OMIM=611324, no_hdr0=''),
  9747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9747, CurrentID=0, Status='live', Symbol='TCAF1', Aliases='FAM115A', description='TRPM8 channel-associated factor 1', other_designations='TRP channel-associated factor 1|family with sequence similarity 115, member A|protein FAM115A', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143851368, end_position_on_the_genomic_accession=143902198, orientation='minus', exon_count=12, OMIM=616251, no_hdr0=''),
  9748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9748, CurrentID=0, Status='live', Symbol='SLK', Aliases='LOSK, STK2, bA16H23.1, se20-9', description='STE20 like kinase', other_designations='CTCL tumor antigen se20-9|Long Ste20-like Kinase|SNF1 (sucrose nonfermenting, yeast, homolog)-like kinase, SNF1 sucrose nonfermenting like kinase|SNF1 sucrose nonfermenting like kinase|STE20-like kinase|STE20-related kinase|STE20-related serine/threonine-protein kinase|Ste20-related serine/threonine kinase|hSLK|serine/threonine-protein kinase 2', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103967182, end_position_on_the_genomic_accession=104029233, orientation='plus', exon_count=19, OMIM=616563, no_hdr0=''),
  9749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9749, CurrentID=0, Status='live', Symbol='PHACTR2', Aliases='C6orf56', description='phosphatase and actin regulator 2', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143608180, end_position_on_the_genomic_accession=143831185, orientation='plus', exon_count=14, OMIM=608724, no_hdr0=''),
  9750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9750, CurrentID=0, Status='live', Symbol='FAM65B', Aliases='C6orf32, DIFF40, DIFF48, MYONAP, PL48', description='family with sequence similarity 65 member B', other_designations='family with sequence similarity 65, member B|myogenesis-related and NCAM-associated protein homolog', map_location='6p22.3-p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24804281, end_position_on_the_genomic_accession=25042288, orientation='minus', exon_count=31, OMIM=611410, no_hdr0=''),
  140823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140823, CurrentID=0, Status='live', Symbol='ROMO1', Aliases='C20orf52, MTGM, MTGMP, bA353C18.2', description='reactive oxygen species modulator 1', other_designations='PCM19|ROS modulator 1|epididymis tissue protein Li 175|glyrichin|mitochondrial targeting GXXXG protein|mitochondrial targeting GxxxG motif protein|protein MGR2 homolog', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35699310, end_position_on_the_genomic_accession=35700980, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9752, CurrentID=0, Status='live', Symbol='PCDHA9', Aliases='PCDH-ALPHA9', description='protocadherin alpha 9', other_designations='KIAA0345-like 5|PCDH-alpha-9', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140847772, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606315, no_hdr0=''),
  140825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140825, CurrentID=0, Status='live', Symbol='NEURL2', Aliases='C20orf163, OZZ, OZZ-E3', description='neuralized E3 ubiquitin protein ligase 2', other_designations='neuralized homolog 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45888472, end_position_on_the_genomic_accession=45891287, orientation='minus', exon_count=2, OMIM=608597, no_hdr0=''),
  9754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9754, CurrentID=0, Status='live', Symbol='STARD8', Aliases='ARHGAP38, DLC3, STARTGAP3', description='StAR related lipid transfer domain containing 8', other_designations='START domain containing 8|StAR-related lipid transfer (START) domain containing 8|StAR-related lipid transfer domain containing 8|deleted in liver cancer 3 protein', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68647669, end_position_on_the_genomic_accession=68725842, orientation='plus', exon_count=17, OMIM=300689, no_hdr0=''),
  9755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9755, CurrentID=0, Status='live', Symbol='TBKBP1', Aliases='ProSAPiP2, SINTBAD', description='TBK1 binding protein 1', other_designations='TBK1-binding protein 1', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47694065, end_position_on_the_genomic_accession=47712063, orientation='plus', exon_count=13, OMIM=608476, no_hdr0=''),
  140828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140828, CurrentID=0, Status='live', Symbol='LINC00261', Aliases='ALIEN, C20orf56, DEANR1, HCCDR1, NCRNA00261, onco-lncRNA-17', description='long intergenic non-protein coding RNA 261', other_designations='TCONS_00027846|definitive endoderm-associated long non-coding RNA 1|hepatocellular carcinoma down-regulated RNA 1', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22560553, end_position_on_the_genomic_accession=22578642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  9757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9757, CurrentID=0, Status='live', Symbol='KMT2B', Aliases='CXXC10, HRX2, MLL1B, MLL2, MLL4, TRX2, WBP-7, WBP7', description='lysine (K)-specific methyltransferase 2B', other_designations='WW domain binding protein 7|histone-lysine N-methyltransferase MLL4|mixed lineage leukemia gene homolog 2|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila) 4|myeloid/lymphoid or mixed-lineage leukemia protein 4|trithorax homologue 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35717995, end_position_on_the_genomic_accession=35738880, orientation='plus', exon_count=38, OMIM=606834, no_hdr0=''),
  9758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9758, CurrentID=0, Status='live', Symbol='FRMPD4', Aliases='PDZD10, PDZK10', description='FERM and PDZ domain containing 4', other_designations='PDZ domain-containing protein 10|PSD-95-interacting FERM and PDZ domain protein|PSD-95-interacting regulator of spine morphogenesis|Preso', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12138466, end_position_on_the_genomic_accession=12724523, orientation='plus', exon_count=18, OMIM=300838, no_hdr0=''),
  9759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9759, CurrentID=0, Status='live', Symbol='HDAC4', Aliases='AHO3, BDMR, HA6116, HD4, HDAC-4, HDAC-A, HDACA', description='histone deacetylase 4', other_designations='histone deacetylase A', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239048168, end_position_on_the_genomic_accession=239401647, orientation='minus', exon_count=37, OMIM=605314, no_hdr0=''),
  9760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9760, CurrentID=0, Status='live', Symbol='TOX', Aliases='TOX1', description='thymocyte selection associated high mobility group box', other_designations='thymus high mobility group box protein TOX', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58805418, end_position_on_the_genomic_accession=59119208, orientation='minus', exon_count=9, OMIM=606863, no_hdr0=''),
  9761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9761, CurrentID=0, Status='live', Symbol='MLEC', Aliases='KIAA0152', description='malectin', other_designations='oligosaccharyltransferase complex subunit (non-catalytic)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120687124, end_position_on_the_genomic_accession=120701864, orientation='plus', exon_count=6, OMIM=613802, no_hdr0=''),
  9762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9762, CurrentID=0, Status='live', Symbol='LZTS3', Aliases='PROSAPIP1', description='leucine zipper, putative tumor suppressor family member 3', other_designations='ProSAP/Shank-interacting protein 1|proSAP-interacting protein 1|proline rich synapse associated protein interacting protein 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3162617, end_position_on_the_genomic_accession=3173592, orientation='minus', exon_count=8, OMIM=610484, no_hdr0=''),
  100533851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533851, CurrentID=0, Status='live', Symbol='ATP8A2P2', Aliases='', description='ATPase, aminophospholipid transporter, class I, type 8A, member 2 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32477174, end_position_on_the_genomic_accession=32478850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140836, CurrentID=0, Status='live', Symbol='BANF2', Aliases='BAF-L, BAF2, BAFL, C20orf179', description='barrier to autointegration factor 2', other_designations='barrier-to-autointegration factor 2', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17693675, end_position_on_the_genomic_accession=17735872, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  9765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9765, CurrentID=0, Status='live', Symbol='ZFYVE16', Aliases='PPP1R69', description='zinc finger, FYVE domain containing 16', other_designations='endosome-associated FYVE-domain protein|protein phosphatase 1, regulatory subunit 69', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80407592, end_position_on_the_genomic_accession=80480520, orientation='plus', exon_count=24, OMIM=608880, no_hdr0=''),
  9766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9766, CurrentID=0, Status='live', Symbol='SUSD6', Aliases='KIAA0247', description='sushi domain containing 6', other_designations='drug-activated gene overexpressed protein', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69611593, end_position_on_the_genomic_accession=69715144, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  9767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9767, CurrentID=0, Status='live', Symbol='JADE3', Aliases='JADE-3, PHF16', description='jade family PHD finger 3', other_designations='PHD finger protein 16|jade family PHD finger protein 3', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46912276, end_position_on_the_genomic_accession=47061242, orientation='plus', exon_count=12, OMIM=300618, no_hdr0=''),
  9768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9768, CurrentID=0, Status='live', Symbol='KIAA0101', Aliases='L5, NS5ATP9, OEATC, OEATC-1, OEATC1, PAF, PAF15, p15(PAF), p15/PAF, p15PAF', description='KIAA0101', other_designations='HCV NS5A-transactivated protein 9|PCNA-associated factor of 15 kDa|hepatitis C virus NS5A-transactivated protein 9|overexpressed in anaplastic thyroid carcinoma 1', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64364994, end_position_on_the_genomic_accession=64387687, orientation='minus', exon_count=5, OMIM=610696, no_hdr0=''),
  100533852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533852, CurrentID=0, Status='live', Symbol='LOC100533852', Aliases='', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells 2 (p49/p100) pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12091518, end_position_on_the_genomic_accession=12093091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9770, CurrentID=0, Status='live', Symbol='RASSF2', Aliases='CENP-34, RASFADIN', description='Ras association (RalGDS/AF-6) domain family member 2', other_designations='Ras association (RalGDS/AF-6) domain family 2|centromere protein 34', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4780023, end_position_on_the_genomic_accession=4823645, orientation='minus', exon_count=15, OMIM=609492, no_hdr0=''),
  9771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9771, CurrentID=0, Status='live', Symbol='RAPGEF5', Aliases='GFR, MR-GEF, REPAC', description='Rap guanine nucleotide exchange factor 5', other_designations='M-Ras-regulated GEF|M-Ras-regulated Rap GEF|Rap guanine nucleotide exchange factor (GEF) 5|guanine nucleotide exchange factor for Rap1|related to Epac', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22118290, end_position_on_the_genomic_accession=22357133, orientation='minus', exon_count=28, OMIM=609527, no_hdr0=''),
  9772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9772, CurrentID=0, Status='live', Symbol='TMEM94', Aliases='KIAA0195', description='transmembrane protein 94', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75441159, end_position_on_the_genomic_accession=75500452, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  9774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9774, CurrentID=0, Status='live', Symbol='BCLAF1', Aliases='BTF, bK211L9.1', description='BCL2-associated transcription factor 1', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136256863, end_position_on_the_genomic_accession=136290645, orientation='minus', exon_count=16, OMIM=612588, no_hdr0=''),
  9775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9775, CurrentID=0, Status='live', Symbol='EIF4A3', Aliases='DDX48, MUK34, NMP265, NUK34, RCPS, eIF4AIII', description='eukaryotic translation initiation factor 4A3', other_designations='ATP-dependent RNA helicase DDX48|ATP-dependent RNA helicase eIF4A-3|DEAD (Asp-Glu-Ala-Asp) box polypeptide 48|DEAD box protein 48|NMP 265|eIF-4A-III|eIF4A-III|eukaryotic initiation factor 4A-like NUK-34|eukaryotic translation initiation factor 4A|hNMP 265|nuclear matrix protein 265', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80135214, end_position_on_the_genomic_accession=80147183, orientation='minus', exon_count=12, OMIM=608546, no_hdr0=''),
  9776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9776, CurrentID=0, Status='live', Symbol='ATG13', Aliases='KIAA0652, PARATARG8', description='autophagy related 13', other_designations='ATG13 autophagy related 13 homolog', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46617276, end_position_on_the_genomic_accession=46676019, orientation='plus', exon_count=25, OMIM=615088, no_hdr0=''),
  9777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9777, CurrentID=0, Status='live', Symbol='TM9SF4', Aliases='dJ836N17.2', description='transmembrane 9 superfamily protein member 4', other_designations='dinucleotide oxidase disulfide thiol exchanger 3 superfamily member 4', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32109506, end_position_on_the_genomic_accession=32167258, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  140850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140850, CurrentID=0, Status='live', Symbol='DEFB127', Aliases='C20orf73, DEF-27, DEFB-27, DEFB27, bA530N10.2, hBD-27', description='defensin beta 127', other_designations='beta defensin 27|beta-defensin 27|defensin, beta 127', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=157470, end_position_on_the_genomic_accession=159163, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  9779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9779, CurrentID=0, Status='live', Symbol='TBC1D5', Aliases='', description='TBC1 domain family member 5', other_designations='TBC1 domain family, member 5', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17157162, end_position_on_the_genomic_accession=17742739, orientation='minus', exon_count=33, OMIM=615740, no_hdr0=''),
  9780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9780, CurrentID=0, Status='live', Symbol='PIEZO1', Aliases='DHS, FAM38A, Mib', description='piezo-type mechanosensitive ion channel component 1', other_designations='family with sequence similarity 38, member A|membrane protein induced by beta-amyloid treatment', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88715338, end_position_on_the_genomic_accession=88784964, orientation='minus', exon_count=51, OMIM=611184, no_hdr0=''),
  9781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9781, CurrentID=0, Status='live', Symbol='RNF144A', Aliases='RNF144, UBCE7IP4', description='ring finger protein 144A', other_designations='UbcM4-interacting protein 4|probable E3 ubiquitin-protein ligase RNF144A|ring finger protein 144|ubiquitin conjugating enzyme 7 interacting protein 4|ubiquitin-conjugating enzyme 7-interacting protein 4', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6917392, end_position_on_the_genomic_accession=7076197, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  9782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9782, CurrentID=0, Status='live', Symbol='MATR3', Aliases='ALS21, MPD2, VCPDM', description='matrin 3', other_designations='vocal cord and pharyngeal weakness with distal myopathy', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139273752, end_position_on_the_genomic_accession=139331677, orientation='plus', exon_count=19, OMIM=164015, no_hdr0=''),
  9783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9783, CurrentID=0, Status='live', Symbol='RIMS3', Aliases='NIM3, RIM3', description='regulating synaptic membrane exocytosis 3', other_designations='RIM 3|RIM3 gamma|Rab-3 interacting molecule 3|rab-3-interacting molecule 3', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40620680, end_position_on_the_genomic_accession=40692084, orientation='minus', exon_count=11, OMIM=611600, no_hdr0=''),
  9784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9784, CurrentID=0, Status='live', Symbol='SNX17', Aliases='', description='sorting nexin 17', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27370496, end_position_on_the_genomic_accession=27377533, orientation='plus', exon_count=16, OMIM=605963, no_hdr0=''),
  9785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9785, CurrentID=0, Status='live', Symbol='DHX38', Aliases='DDX38, PRP16, PRPF16', description='DEAH (Asp-Glu-Ala-His) box polypeptide 38', other_designations='ATP-dependent RNA helicase DHX38|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 38|DEAH box protein 38|PRP16 homolog of S.cerevisiae|pre-mRNA splicing factor ATP-dependent RNA helicase PRP16', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72093665, end_position_on_the_genomic_accession=72112912, orientation='plus', exon_count=27, OMIM=605584, no_hdr0=''),
  9786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9786, CurrentID=0, Status='live', Symbol='KIAA0586', Aliases='Talpid3', description='KIAA0586', other_designations='ortholog of chicken talpid{3}', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58427385, end_position_on_the_genomic_accession=58548831, orientation='plus', exon_count=36, OMIM=610178, no_hdr0=''),
  9787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9787, CurrentID=0, Status='live', Symbol='DLGAP5', Aliases='DLG7, HURP', description='discs, large (Drosophila) homolog-associated protein 5', other_designations='DAP-5|discs large homolog 7|discs, large homolog 7|disks large-associated protein DLG7|hepatoma up-regulated protein', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55148116, end_position_on_the_genomic_accession=55191678, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  9788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9788, CurrentID=0, Status='live', Symbol='MTSS1', Aliases='MIM, MIMA, MIMB', description='metastasis suppressor 1', other_designations='metastasis suppressor YGL-1|missing in metastasis protein', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124550770, end_position_on_the_genomic_accession=124728507, orientation='minus', exon_count=19, OMIM=608486, no_hdr0=''),
  9789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9789, CurrentID=0, Status='live', Symbol='SPCS2', Aliases='', description='signal peptidase complex subunit 2', other_designations='SPase 25 kDa subunit|microsomal signal peptidase 25 kDa subunit|signal peptidase 25kDa subunit|signal peptidase complex subunit 2 homolog', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74949247, end_position_on_the_genomic_accession=74979031, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  9790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9790, CurrentID=0, Status='live', Symbol='BMS1', Aliases='ACCL, BMS1', description='BMS1, ribosome biogenesis factor', other_designations='BMS1 homolog, ribosome assembly protein|BMS1 ribosome biogenesis factor|BMS1-like, ribosome assembly protein|ribosome assembly protein BMS1 homolog', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42782506, end_position_on_the_genomic_accession=42834937, orientation='plus', exon_count=24, OMIM=611448, no_hdr0=''),
  9791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9791, CurrentID=0, Status='live', Symbol='PTDSS1', Aliases='LMHD, PSS1, PSSA', description='phosphatidylserine synthase 1', other_designations='PSS-1|ptdSer synthase 1|serine-exchange enzyme I', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96261886, end_position_on_the_genomic_accession=96334552, orientation='plus', exon_count=13, OMIM=612792, no_hdr0=''),
  9792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9792, CurrentID=0, Status='live', Symbol='SERTAD2', Aliases='Sei-2, TRIP-Br2', description='SERTA domain containing 2', other_designations='transcriptional regulator interacting with the PHD-bromodomain 2|transcriptional regulator interacting with the PHS-bromodomain 2', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64631621, end_position_on_the_genomic_accession=64751091, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  9793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9793, CurrentID=0, Status='live', Symbol='CKAP5', Aliases='CHTOG, MSPS, TOG, TOGp, ch-TOG', description='cytoskeleton associated protein 5', other_designations='colonic and hepatic tumor over-expressed gene protein|colonic and hepatic tumor overexpressed gene protein', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46743534, end_position_on_the_genomic_accession=46846308, orientation='minus', exon_count=44, OMIM=611142, no_hdr0=''),
  9794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9794, CurrentID=0, Status='live', Symbol='MAML1', Aliases='Mam-1, Mam1', description='mastermind like transcriptional coactivator 1', other_designations='mastermind homolog|mastermind-like 1|mastermind-like transcriptional coactivator 1', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179732850, end_position_on_the_genomic_accession=179777286, orientation='plus', exon_count=5, OMIM=605424, no_hdr0=''),
  9796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9796, CurrentID=0, Status='live', Symbol='PHYHIP', Aliases='DYRK1AP3, PAHX-AP, PAHXAP1', description='phytanoyl-CoA 2-hydroxylase interacting protein', other_designations='DYRK1A interacting protein 3|PAHX-AP1|phytanoyl-CoA alpha-hydroxylase associated protein|phytanoyl-CoA hydroxylase interacting protein|phytanoyl-CoA hydroxylase-associated protein 1', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22219703, end_position_on_the_genomic_accession=22232338, orientation='minus', exon_count=6, OMIM=608511, no_hdr0=''),
  9797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9797, CurrentID=0, Status='live', Symbol='TATDN2', Aliases='', description='TatD DNase domain containing 2', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10248493, end_position_on_the_genomic_accession=10281222, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  9798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9798, CurrentID=0, Status='live', Symbol='IST1', Aliases='OLC1', description='increased sodium tolerance 1 homolog (yeast)', other_designations='overexpressed in lung cancer 1|putative MAPK-activating protein PM28', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71894408, end_position_on_the_genomic_accession=71930639, orientation='plus', exon_count=12, OMIM=616434, no_hdr0=''),
  140871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140871, CurrentID=0, Status='live', Symbol='CFTRP1', Aliases='C20orf189, dJ760C5.1', description='cystic fibrosis transmembrane conductance regulator pseudogene 1', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25919400, end_position_on_the_genomic_accession=25919782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140873, CurrentID=0, Status='live', Symbol='C20orf173', Aliases='dJ477O4.4', description='chromosome 20 open reading frame 173', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35520741, end_position_on_the_genomic_accession=35529652, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  9802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9802, CurrentID=0, Status='live', Symbol='DAZAP2', Aliases='PRTB', description='DAZ associated protein 2', other_designations='deleted in azoospermia associated protein 2|deleted in azoospermia-associated protein 2|proline-rich transcript in brain|proline-rich transcript, brain-expressed protein', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51238724, end_position_on_the_genomic_accession=51246717, orientation='plus', exon_count=5, OMIM=607431, no_hdr0=''),
  140875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140875, CurrentID=0, Status='live', Symbol='LINC00028', Aliases='C20orf93, NCRNA00028, dJ1093G12.6', description='long intergenic non-protein coding RNA 28', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31485778, end_position_on_the_genomic_accession=31487574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140876, CurrentID=0, Status='live', Symbol='FAM65C', Aliases='C20orf175, C20orf176', description='family with sequence similarity 65 member C', other_designations='family with sequence similarity 65, member C', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50586108, end_position_on_the_genomic_accession=50692004, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  9805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9805, CurrentID=0, Status='live', Symbol='SCRN1', Aliases='SES1', description='secernin 1', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29920103, end_position_on_the_genomic_accession=29990289, orientation='minus', exon_count=12, OMIM=614965, no_hdr0=''),
  9806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9806, CurrentID=0, Status='live', Symbol='SPOCK2', Aliases='testican-2', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan (testican) 2', other_designations='SPARC/osteonectin, CWCV, and Kazal-like domains proteoglycan 2', map_location='10pter-q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72059034, end_position_on_the_genomic_accession=72089032, orientation='minus', exon_count=13, OMIM=607988, no_hdr0=''),
  9807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9807, CurrentID=0, Status='live', Symbol='IP6K1', Aliases='IHPK1, PiUS', description='inositol hexakisphosphate kinase 1', other_designations='ATP:1D-myo-inositol-hexakisphosphate phosphotransferase|Pi uptake stimulator|inositol hexaphosphate kinase 1|insP6 kinase 1', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49724295, end_position_on_the_genomic_accession=49786540, orientation='minus', exon_count=7, OMIM=606991, no_hdr0=''),
  140880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140880, CurrentID=0, Status='live', Symbol='CST11', Aliases='CST8L, CTES2, SC13, dJ322G13.6', description='cystatin 11', other_designations='SC13delta|cystatin-related epididymal-specific protein', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23450404, end_position_on_the_genomic_accession=23452845, orientation='minus', exon_count=3, OMIM=609731, no_hdr0=''),
  140881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140881, CurrentID=0, Status='live', Symbol='DEFB129', Aliases='C20orf87, DEFB-29, DEFB29, bA530N10.3, hBD-29', description='defensin beta 129', other_designations='beta defensin 29|beta-defensin 29|defensin, beta 129', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=227258, end_position_on_the_genomic_accession=229886, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  9810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9810, CurrentID=0, Status='live', Symbol='RNF40', Aliases='BRE1B, RBP95, STARING', description='ring finger protein 40, E3 ubiquitin protein ligase', other_designations='95 kDa retinoblastoma protein binding protein|95 kDa retinoblastoma-associated protein|BRE1 E3 ubiquitin ligase homolog B|BRE1-B|Rb-associated protein', map_location='16p11.2-p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30761612, end_position_on_the_genomic_accession=30802340, orientation='plus', exon_count=24, OMIM=607700, no_hdr0=''),
  9811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9811, CurrentID=0, Status='live', Symbol='CTIF', Aliases='Gm672, KIAA0427', description='CBP80/20-dependent translation initiation factor', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48539056, end_position_on_the_genomic_accession=48863217, orientation='plus', exon_count=18, OMIM=613178, no_hdr0=''),
  9812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9812, CurrentID=0, Status='live', Symbol='KIAA0141', Aliases='DELE', description='KIAA0141', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141923785, end_position_on_the_genomic_accession=141951465, orientation='plus', exon_count=17, OMIM=615741, no_hdr0=''),
  140885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140885, CurrentID=0, Status='live', Symbol='SIRPA', Aliases='BIT, CD172A, MFR, MYD-1, P84, PTPNS1, SHPS1, SIRP', description='signal-regulatory protein alpha', other_designations='CD172 antigen-like family member A|brain-immunoglobulin-like molecule with tyrosine-based activation motifs|inhibitory receptor SHPS-1|macrophage fusion receptor|myd-1 antigen|tyrosine phosphatase SHP substrate 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1894117, end_position_on_the_genomic_accession=1939896, orientation='plus', exon_count=12, OMIM=602461, no_hdr0=''),
  9814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9814, CurrentID=0, Status='live', Symbol='SFI1', Aliases='PISD, PPP1R139, hSfi1p', description='SFI1 centrin binding protein', other_designations='Sfi1 homolog, spindle assembly associated|homolog of yeast Sfi1|protein phosphatase 1, regulatory subunit 139', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31488688, end_position_on_the_genomic_accession=31618551, orientation='plus', exon_count=37, OMIM=612765, no_hdr0=''),
  9815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9815, CurrentID=0, Status='live', Symbol='GIT2', Aliases='CAT-2, CAT2', description='G protein-coupled receptor kinase interacting ArfGAP 2', other_designations='ARF GAP GIT2|G protein-coupled receptor kinase-interactor 2|GRK-interacting protein 2|cool-associated, tyrosine phosphorylated protein 2|cool-interacting tyrosine-phosphorylated protein 2', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109929802, end_position_on_the_genomic_accession=109996389, orientation='minus', exon_count=22, OMIM=608564, no_hdr0=''),
  9816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9816, CurrentID=0, Status='live', Symbol='URB2', Aliases='KIAA0133, NET10, NPA2', description='URB2 ribosome biogenesis 2 homolog (S. cerevisiae)', other_designations='nucleolar preribosomal-associated protein 1', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229626205, end_position_on_the_genomic_accession=229660887, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  9817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9817, CurrentID=0, Status='live', Symbol='KEAP1', Aliases='INrf2, KLHL19', description='kelch like ECH associated protein 1', other_designations='cytosolic inhibitor of Nrf2|kelch-like family member 19|kelch-like protein 19', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10486120, end_position_on_the_genomic_accession=10503378, orientation='minus', exon_count=9, OMIM=606016, no_hdr0=''),
  140890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140890, CurrentID=0, Status='live', Symbol='SREK1', Aliases='SFRS12, SRrp508, SRrp86', description='splicing regulatory glutamine/lysine-rich protein 1', other_designations='serine-arginine-rich splicing regulatory protein 508|serine/arginine-rich-splicing regulatory protein 86|splicing factor, arginine/serine-rich 12|splicing regulatory protein 508', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66144218, end_position_on_the_genomic_accession=66183616, orientation='plus', exon_count=15, OMIM=609268, no_hdr0=''),
  9819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9819, CurrentID=0, Status='live', Symbol='TSC22D2', Aliases='TILZ4a, TILZ4b, TILZ4c', description='TSC22 domain family member 2', other_designations='TSC22 domain family 2|TSC22 domain family, member 2|TSC22-related-inducible leucine zipper protein 4', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150408335, end_position_on_the_genomic_accession=150460119, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  9820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9820, CurrentID=0, Status='live', Symbol='CUL7', Aliases='3M1, KIAA0076, dJ20C7.5', description='cullin 7', other_designations='CUL-7', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43037617, end_position_on_the_genomic_accession=43053950, orientation='minus', exon_count=27, OMIM=609577, no_hdr0=''),
  9821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9821, CurrentID=0, Status='live', Symbol='RB1CC1', Aliases='ATG17, CC1, FIP200, PPP1R131', description='RB1 inducible coiled-coil 1', other_designations='200 kDa FAK family kinase-interacting protein|FAK family kinase-interacting protein of 200 kDa|RB1-inducible coiled-coil 1|phosphatase 1, regulatory subunit 131', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52622458, end_position_on_the_genomic_accession=52714466, orientation='minus', exon_count=28, OMIM=606837, no_hdr0=''),
  140894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140894, CurrentID=0, Status='live', Symbol='CNBD2', Aliases='C20orf152, CNMPD1', description='cyclic nucleotide binding domain containing 2', other_designations='cyclic nucleotide (cNMP) binding domain containing 1', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35954562, end_position_on_the_genomic_accession=36030700, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  9823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9823, CurrentID=0, Status='live', Symbol='ARMCX2', Aliases='ALEX2, GASP9', description='armadillo repeat containing, X-linked 2', other_designations='ARM protein lost in epithelial cancers on chromosome X 2|armadillo repeat protein ALEX2', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101655281, end_position_on_the_genomic_accession=101659891, orientation='minus', exon_count=6, OMIM=300363, no_hdr0=''),
  9824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9824, CurrentID=0, Status='live', Symbol='ARHGAP11A', Aliases='GAP (1-12)', description='Rho GTPase activating protein 11A', other_designations='rho-type GTPase-activating protein 11A', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32613733, end_position_on_the_genomic_accession=32639949, orientation='plus', exon_count=16, OMIM=610589, no_hdr0=''),
  9825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9825, CurrentID=0, Status='live', Symbol='SPATA2', Aliases='PD1, PPP1R145, tamo', description='spermatogenesis associated 2', other_designations='protein phosphatase 1, regulatory subunit 145|spermatogenesis associated PD1|spermatogenesis-associated protein PD1', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49903391, end_position_on_the_genomic_accession=49915543, orientation='minus', exon_count=6, OMIM=607662, no_hdr0=''),
  9826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9826, CurrentID=0, Status='live', Symbol='ARHGEF11', Aliases='GTRAP48, PDZ-RHOGEF', description='Rho guanine nucleotide exchange factor 11', other_designations='Rho guanine exchange factor (GEF) 11|Rho guanine nucleotide exchange factor (GEF) 11|RhoA-specific guanine nucleotide exchange factor|RhoGEF glutamate transport modulator|glutamate transporter EAAT4-associated protein 48', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156934840, end_position_on_the_genomic_accession=157046640, orientation='minus', exon_count=48, OMIM=605708, no_hdr0=''),
  9827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9827, CurrentID=0, Status='live', Symbol='RGP1', Aliases='KIAA0258', description='RGP1 homolog, RAB6A GEF complex partner 1', other_designations='RGP1 retrograde golgi transport homolog|retrograde Golgi transport protein RGP1 homolog', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35749280, end_position_on_the_genomic_accession=35753267, orientation='plus', exon_count=9, OMIM=615742, no_hdr0=''),
  9828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9828, CurrentID=0, Status='live', Symbol='ARHGEF17', Aliases='P164RHOGEF, TEM4, p164-RhoGEF', description='Rho guanine nucleotide exchange factor 17', other_designations='164 kDa Rho-specific guanine-nucleotide exchange factor|Rho guanine nucleotide exchange factor (GEF) 17|Rho-specific guanine-nucleotide exchange factor 164 kDa|tumor endothelial marker 4', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73305939, end_position_on_the_genomic_accession=73369380, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  140901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140901, CurrentID=0, Status='live', Symbol='STK35', Aliases='CLIK1L1, STK35', description='serine/threonine kinase 35', other_designations='CLIK-1|CLP-36 interacting kinase|CLP-36-interacting kinase 1|PDLIM1-interacting kinase 1|serine threonine kinase 35 long form|serine/threonine-protein kinase 35 L1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2099799, end_position_on_the_genomic_accession=2148555, orientation='plus', exon_count=4, OMIM=609370, no_hdr0=''),
  9830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9830, CurrentID=0, Status='live', Symbol='TRIM14', Aliases='', description='tripartite motif containing 14', other_designations='tripartite motif protein TRIM14|tripartite motif-containing 14', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98083189, end_position_on_the_genomic_accession=98119408, orientation='minus', exon_count=12, OMIM=606556, no_hdr0=''),
  9831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9831, CurrentID=0, Status='live', Symbol='ZNF623', Aliases='', description='zinc finger protein 623', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143636013, end_position_on_the_genomic_accession=143653775, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9832, CurrentID=0, Status='live', Symbol='JAKMIP2', Aliases='JAMIP2, NECC1', description='janus kinase and microtubule interacting protein 2', other_designations='CTCL tumor antigen HD-CL-04|Jak and microtubule interacting protein 2|neuroendocrine long coiled-coil protein 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147585439, end_position_on_the_genomic_accession=147782848, orientation='minus', exon_count=22, OMIM=611197, no_hdr0=''),
  9833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9833, CurrentID=0, Status='live', Symbol='MELK', Aliases='HPK38', description='maternal embryonic leucine zipper kinase', other_designations='pEg3 kinase|protein kinase Eg3|protein kinase PK38|tyrosine-protein kinase MELK', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36572862, end_position_on_the_genomic_accession=36677683, orientation='plus', exon_count=22, OMIM=607025, no_hdr0=''),
  9834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9834, CurrentID=0, Status='live', Symbol='KIAA0125', Aliases='C14orf110, FAM30A, HSPC053', description='KIAA0125', other_designations='family with sequence similarity 30, member A', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105917979, end_position_on_the_genomic_accession=105932642, orientation='plus', exon_count=6, OMIM=616623, no_hdr0=''),
  140907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140907, CurrentID=0, Status='live', Symbol='SPG19', Aliases='', description='spastic paraplegia 19 (autosomal dominant)', other_designations='', map_location='9q', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607152, no_hdr0=''),
  9836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9836, CurrentID=0, Status='live', Symbol='LCMT2', Aliases='PPM2, TYW4', description='leucine carboxyl methyltransferase 2', other_designations='p21WAF1/CIP1 promoter-interacting protein|tRNA yW-synthesizing protein 4|tRNA(Phe) (7-(3-amino-3-(methoxycarbonyl)propyl)wyosine(37)-N)-methoxycarbonyltransferase|tRNA(Phe) (7-(3-amino-3-carboxypropyl)wyosine(37)-O)-methyltransferase|tRNA-Wybutosine-synthesizing protein 4|tRNA-yW synthesizing protein 4', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43327776, end_position_on_the_genomic_accession=43330622, orientation='minus', exon_count=1, OMIM=611246, no_hdr0=''),
  9837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9837, CurrentID=0, Status='live', Symbol='GINS1', Aliases='PSF1', description='GINS complex subunit 1 (Psf1 homolog)', other_designations='partner of sld five-1', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25407444, end_position_on_the_genomic_accession=25448555, orientation='plus', exon_count=11, OMIM=610608, no_hdr0=''),
  9839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9839, CurrentID=0, Status='live', Symbol='ZEB2', Aliases='HSPC082, SIP-1, SIP1, SMADIP1, ZFHX1B', description='zinc finger E-box binding homeobox 2', other_designations='SMAD interacting protein 1|Smad-interacting protein 1|zinc finger homeobox 1b', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144384375, end_position_on_the_genomic_accession=144520393, orientation='minus', exon_count=14, OMIM=605802, no_hdr0=''),
  9840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9840, CurrentID=0, Status='live', Symbol='TESPA1', Aliases='HSPC257, KIAA0748', description='thymocyte expressed, positive selection associated 1', other_designations='thymocyte-expressed positive selection-associated protein 1', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54948303, end_position_on_the_genomic_accession=54985161, orientation='minus', exon_count=17, OMIM=615664, no_hdr0=''),
  9841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9841, CurrentID=0, Status='live', Symbol='ZBTB24', Aliases='BIF1, ICF2, PATZ2, ZNF450', description='zinc finger and BTB domain containing 24', other_designations='POZ (BTB) and AT hook containing zinc finger 2|zinc finger protein 450', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109462516, end_position_on_the_genomic_accession=109483237, orientation='minus', exon_count=7, OMIM=614064, no_hdr0=''),
  9842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9842, CurrentID=0, Status='live', Symbol='PLEKHM1', Aliases='AP162, B2, OPTB6', description='pleckstrin homology and RUN domain containing M1', other_designations='162 kDa adapter protein|PH domain-containing family M member 1|pleckstrin homology domain containing, family M (with RUN domain) member 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45434247, end_position_on_the_genomic_accession=45490780, orientation='minus', exon_count=17, OMIM=611466, no_hdr0=''),
  9843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9843, CurrentID=0, Status='live', Symbol='HEPH', Aliases='CPL', description='hephaestin', other_designations='', map_location='Xq11-q12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66162519, end_position_on_the_genomic_accession=66267389, orientation='plus', exon_count=22, OMIM=300167, no_hdr0=''),
  9844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9844, CurrentID=0, Status='live', Symbol='ELMO1', Aliases='CED-12, CED12, ELMO-1', description='engulfment and cell motility 1', other_designations='ced-12 homolog 1', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36852906, end_position_on_the_genomic_accession=37449409, orientation='minus', exon_count=28, OMIM=606420, no_hdr0=''),
  9846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9846, CurrentID=0, Status='live', Symbol='GAB2', Aliases='', description='GRB2 associated binding protein 2', other_designations='GRB2-associated binding protein 2|Grb2-associated binder 2|growth factor receptor bound protein 2-associated protein 2|pp100', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78215290, end_position_on_the_genomic_accession=78417822, orientation='minus', exon_count=13, OMIM=606203, no_hdr0=''),
  9847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9847, CurrentID=0, Status='live', Symbol='C2CD5', Aliases='CDP138, KIAA0528', description='C2 calcium-dependent domain containing 5', other_designations='138 kDa C2 domain-containing phosphoprotein', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22448546, end_position_on_the_genomic_accession=22544546, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  9848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9848, CurrentID=0, Status='live', Symbol='MFAP3L', Aliases='NYD-sp9', description='microfibrillar associated protein 3 like', other_designations='testis development protein NYD-SP9', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169986597, end_position_on_the_genomic_accession=170033028, orientation='minus', exon_count=11, OMIM=616523, no_hdr0=''),
  9849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9849, CurrentID=0, Status='live', Symbol='ZNF518A', Aliases='ZNF518', description='zinc finger protein 518A', other_designations='zinc finger protein 518', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96129200, end_position_on_the_genomic_accession=96163793, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  9851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9851, CurrentID=0, Status='live', Symbol='KIAA0753', Aliases='MNR, OFIP', description='KIAA0753', other_designations='moonraker', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6578151, end_position_on_the_genomic_accession=6641744, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  9852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9852, CurrentID=0, Status='live', Symbol='EPM2AIP1', Aliases='', description='EPM2A (laforin) interacting protein 1', other_designations='EPM2A interacting protein 1|laforin-interacting protein', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36985866, end_position_on_the_genomic_accession=36993304, orientation='minus', exon_count=1, OMIM=607911, no_hdr0=''),
  9853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9853, CurrentID=0, Status='live', Symbol='RUSC2', Aliases='Iporin', description='RUN and SH3 domain containing 2', other_designations='RUN and SH3 domain-containing protein 2|interacting protein of Rab1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35489838, end_position_on_the_genomic_accession=35561898, orientation='plus', exon_count=13, OMIM=611053, no_hdr0=''),
  9854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9854, CurrentID=0, Status='live', Symbol='C2CD2L', Aliases='DLNB23, TMEM24', description='C2CD2-like', other_designations='transmembrane protein 24', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119107341, end_position_on_the_genomic_accession=119118347, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  9855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9855, CurrentID=0, Status='live', Symbol='FARP2', Aliases='FIR, FRG, PLEKHC3', description='FERM, ARH/RhoGEF and pleckstrin domain protein 2', other_designations='FERM domain including RhoGEF|FERM, RhoGEF and pleckstrin domain protein 2|FGD1-related Cdc42-GEF|PH domain-containing family C member 3|pleckstrin homology domain-containing family C member 3', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241356249, end_position_on_the_genomic_accession=241494842, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  9856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9856, CurrentID=0, Status='live', Symbol='KIAA0319', Aliases='DYLX2, DYX2, NMIG', description='KIAA0319', other_designations='dyslexia susceptibility 2|neuronal migration', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24544104, end_position_on_the_genomic_accession=24646172, orientation='minus', exon_count=25, OMIM=609269, no_hdr0=''),
  9857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9857, CurrentID=0, Status='live', Symbol='CEP350', Aliases='CAP350, GM133', description='centrosomal protein 350kDa', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179954773, end_position_on_the_genomic_accession=180114880, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  9858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9858, CurrentID=0, Status='live', Symbol='PPP1R26', Aliases='KIAA0649', description='protein phosphatase 1 regulatory subunit 26', other_designations='1A6/DRIM (down-regulated in metastasis) interacting protein|DRIM/UTP20 interacting protein|protein phosphatase 1, regulatory subunit 26', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135479699, end_position_on_the_genomic_accession=135488894, orientation='plus', exon_count=5, OMIM=614056, no_hdr0=''),
  9859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9859, CurrentID=0, Status='live', Symbol='CEP170', Aliases='FAM68A, KAB, KIAA0470', description='centrosomal protein 170kDa', other_designations='KARP-1-binding protein|XRCC5 binding protein', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243124428, end_position_on_the_genomic_accession=243255406, orientation='minus', exon_count=22, OMIM=613023, no_hdr0=''),
  9860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9860, CurrentID=0, Status='live', Symbol='LRIG2', Aliases='LIG-2, LIG2, UFS2', description='leucine-rich repeats and immunoglobulin-like domains 2', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113073170, end_position_on_the_genomic_accession=113125202, orientation='plus', exon_count=19, OMIM=608869, no_hdr0=''),
  9861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9861, CurrentID=0, Status='live', Symbol='PSMD6', Aliases='Rpn7, S10, SGA-113M, p42A, p44S10', description='proteasome 26S subunit, non-ATPase 6', other_designations='breast cancer-associated protein SGA-113M|phosphonoformate immuno-associated protein 4|proteasome (prosome, macropain) 26S subunit, non-ATPase, 6|proteasome regulatory particle subunit p44S10', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64010549, end_position_on_the_genomic_accession=64024010, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  9862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9862, CurrentID=0, Status='live', Symbol='MED24', Aliases='ARC100, CRSP100, CRSP4, DRIP100, THRAP4, TRAP100', description='mediator complex subunit 24', other_designations='CRSP complex subunit 4|activator-recruited cofactor 100 kDa component|cofactor required for Sp1 transcriptional activation subunit 4|cofactor required for Sp1 transcriptional activation, subunit 4, 100kDa|mediator of RNA polymerase II transcription, subunit 24 homolog|thyroid hormone receptor associated protein 4|thyroid hormone receptor-associated protein 4|thyroid hormone receptor-associated protein complex 100 kDa component|vitamin D3 receptor-interacting protein complex 100 kDa component|vitamin D3 receptor-interacting protein complex component DRIP100', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40019097, end_position_on_the_genomic_accession=40054636, orientation='minus', exon_count=28, OMIM=607000, no_hdr0=''),
  9863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9863, CurrentID=0, Status='live', Symbol='MAGI2', Aliases='ACVRIP1, AIP-1, AIP1, ARIP1, MAGI-2, SSCAM', description='membrane associated guanylate kinase, WW and PDZ domain containing 2', other_designations='activin receptor interacting protein 1|atrophin-1-interacting protein 1|atrophin-1-interacting protein A|membrane-associated guanylate kinase inverted 2', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=78017055, end_position_on_the_genomic_accession=79453805, orientation='minus', exon_count=30, OMIM=606382, no_hdr0=''),
  9865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9865, CurrentID=0, Status='live', Symbol='TRIL', Aliases='', description='TLR4 interactor with leucine-rich repeats', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28953358, end_position_on_the_genomic_accession=28958413, orientation='minus', exon_count=1, OMIM=613356, no_hdr0=''),
  9866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9866, CurrentID=0, Status='live', Symbol='TRIM66', Aliases='C11orf29, TIF1D, TIF1DELTA', description='tripartite motif containing 66', other_designations='transcriptional intermediary factor 1 delta|tripartite motif-containing 66', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8612037, end_position_on_the_genomic_accession=8682703, orientation='minus', exon_count=35, OMIM=612000, no_hdr0=''),
  9867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9867, CurrentID=0, Status='live', Symbol='PJA2', Aliases='Neurodap1, RNF131', description='praja ring finger 2, E3 ubiquitin protein ligase', other_designations='praja 2, RING-H2 motif containing|praja2|ring finger protein 131', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109334709, end_position_on_the_genomic_accession=109409974, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  9868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9868, CurrentID=0, Status='live', Symbol='TOMM70A', Aliases='', description='translocase of outer mitochondrial membrane 70 homolog A (S. cerevisiae)', other_designations='mitochondrial precursor proteins import receptor|translocase of outer membrane 70 kDa subunit', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100363459, end_position_on_the_genomic_accession=100401398, orientation='minus', exon_count=12, OMIM=606081, no_hdr0=''),
  9869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9869, CurrentID=0, Status='live', Symbol='SETDB1', Aliases='ESET, H3-K9-HMTase4, KG1T, KMT1E, TDRD21', description='SET domain, bifurcated 1', other_designations='ERG-associated protein with SET domain|ERG-associated protein with a SET domain, ESET|H3-K9-HMTase 4|histone H3-K9 methyltransferase 4|histone-lysine N-methyltransferase, H3lysine-9 specific 4|lysine N-methyltransferase 1E|tudor domain containing 21', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150926339, end_position_on_the_genomic_accession=150964901, orientation='plus', exon_count=22, OMIM=604396, no_hdr0=''),
  9870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9870, CurrentID=0, Status='live', Symbol='AREL1', Aliases='KIAA0317', description='apoptosis resistant E3 ubiquitin protein ligase 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74661252, end_position_on_the_genomic_accession=74713115, orientation='minus', exon_count=21, OMIM=615380, no_hdr0=''),
  9871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9871, CurrentID=0, Status='live', Symbol='SEC24D', Aliases='CLCRP2', description='SEC24 homolog D, COPII coat complex component', other_designations='SEC24 related gene family, member D|SEC24-related protein D', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118722823, end_position_on_the_genomic_accession=118838683, orientation='minus', exon_count=26, OMIM=607186, no_hdr0=''),
  9873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9873, CurrentID=0, Status='live', Symbol='FCHSD2', Aliases='NWK, SH3MD3', description='FCH and double SH3 domains 2', other_designations='FCH and double SH3 domains protein 2|SH3 multiple domains 3|SH3 multiple domains protein 3|carom|nervous wreck homolog', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72836743, end_position_on_the_genomic_accession=73142454, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  9874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9874, CurrentID=0, Status='live', Symbol='TLK1', Aliases='PKU-beta', description='tousled like kinase 1', other_designations='SNARE protein kinase SNAK|serine threonine protein kinase|tousled-like kinase 1', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170990823, end_position_on_the_genomic_accession=171231314, orientation='minus', exon_count=25, OMIM=608438, no_hdr0=''),
  140947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140947, CurrentID=0, Status='live', Symbol='DCANP1', Aliases='C5orf20, DCNP1', description='dendritic cell-associated nuclear protein', other_designations='dendritic cell nuclear protein-1', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135444214, end_position_on_the_genomic_accession=135447348, orientation='minus', exon_count=1, OMIM=609710, no_hdr0=''),
  9877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9877, CurrentID=0, Status='live', Symbol='ZC3H11A', Aliases='ZC3HDC11A', description='zinc finger CCCH-type containing 11A', other_designations='zinc finger CCCH-type domain containing 11A', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203795537, end_position_on_the_genomic_accession=203854128, orientation='plus', exon_count=20, OMIM=613513, no_hdr0=''),
  9878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9878, CurrentID=0, Status='live', Symbol='TOX4', Aliases='C14orf92, KIAA0737, LCP1, MIG7', description='TOX high mobility group box family member 4', other_designations='epidermal Langerhans cell protein LCP1|migration-inducing protein 7', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21477176, end_position_on_the_genomic_accession=21499177, orientation='plus', exon_count=9, OMIM=614032, no_hdr0=''),
  9879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9879, CurrentID=0, Status='live', Symbol='DDX46', Aliases='PRPF5, Prp5', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 46', other_designations='DEAD box protein 46|PRP5 homolog|Prp5-like DEAD-box protein|RNA helicase', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134758734, end_position_on_the_genomic_accession=134831138, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  9880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9880, CurrentID=0, Status='live', Symbol='ZBTB39', Aliases='ZNF922', description='zinc finger and BTB domain containing 39', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56998833, end_position_on_the_genomic_accession=57006513, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  9881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9881, CurrentID=0, Status='live', Symbol='TRANK1', Aliases='LBA1', description='tetratricopeptide repeat and ankyrin repeat containing 1', other_designations='lupus brain antigen 1 homolog', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36826817, end_position_on_the_genomic_accession=36990981, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  9882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9882, CurrentID=0, Status='live', Symbol='TBC1D4', Aliases='AS160, NIDDM5', description='TBC1 domain family member 4', other_designations='TBC (Tre-2, BUB2, CDC16) domain-containing protein|TBC1 domain family, member 4|akt substrate of 160 kDa', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75284664, end_position_on_the_genomic_accession=75482646, orientation='minus', exon_count=23, OMIM=612465, no_hdr0=''),
  9883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9883, CurrentID=0, Status='live', Symbol='POM121', Aliases='P145A, POM121', description='POM121 transmembrane nucleoporin', other_designations='POM121 membrane glycoprotein|nuclear envelope pore membrane protein POM 121A|nuclear pore membrane protein 121 kDa|nucleoporin Nup121', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72879335, end_position_on_the_genomic_accession=72951440, orientation='plus', exon_count=19, OMIM=615753, no_hdr0=''),
  9884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9884, CurrentID=0, Status='live', Symbol='LRRC37A', Aliases='LRRC37', description='leucine rich repeat containing 37A', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46239319, end_position_on_the_genomic_accession=46337794, orientation='plus', exon_count=19, OMIM=616555, no_hdr0=''),
  9885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9885, CurrentID=0, Status='live', Symbol='OSBPL2', Aliases='DFNA67, DNFA67, ORP-2, ORP2', description='oxysterol binding protein like 2', other_designations='OSBP-related protein 2|oxysterol binding protein-like 2', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62238485, end_position_on_the_genomic_accession=62296213, orientation='plus', exon_count=19, OMIM=606731, no_hdr0=''),
  9886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9886, CurrentID=0, Status='live', Symbol='RHOBTB1', Aliases='', description='Rho-related BTB domain containing 1', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60869438, end_position_on_the_genomic_accession=61001468, orientation='minus', exon_count=19, OMIM=607351, no_hdr0=''),
  9887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9887, CurrentID=0, Status='live', Symbol='SMG7', Aliases='C1orf16, EST1C, SGA56M', description='SMG7 nonsense mediated mRNA decay factor', other_designations='EST1 telomerase component homolog C|EST1-like protein C|breast cancer-associated antigen SGA-56M|ever shorter telomeres 1C|smg-7 homolog, nonsense mediated mRNA decay factor', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183472371, end_position_on_the_genomic_accession=183554193, orientation='plus', exon_count=28, OMIM=610964, no_hdr0=''),
  9889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9889, CurrentID=0, Status='live', Symbol='ZBED4', Aliases='', description='zinc finger, BED-type containing 4', other_designations='zinc finger, BED domain containing 4', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49853849, end_position_on_the_genomic_accession=49890078, orientation='plus', exon_count=4, OMIM=612552, no_hdr0=''),
  9890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9890, CurrentID=0, Status='live', Symbol='PLPPR4', Aliases='LPPR4, LPR4, PHP1, PRG-1, PRG1', description='phospholipid phosphatase related 4', other_designations='brain-specific phosphatidic acid phosphatase-like protein 1|plasticity related gene 1|plasticity-related gene 1 protein', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99262953, end_position_on_the_genomic_accession=99309584, orientation='plus', exon_count=8, OMIM=607813, no_hdr0=''),
  9891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9891, CurrentID=0, Status='live', Symbol='NUAK1', Aliases='ARK5', description='NUAK family, SNF1-like kinase, 1', other_designations='AMP-activated protein kinase family member 5|AMPK-related protein kinase 5|omphalocele kinase 1', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106063347, end_position_on_the_genomic_accession=106140033, orientation='minus', exon_count=7, OMIM=608130, no_hdr0=''),
  9892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9892, CurrentID=0, Status='live', Symbol='SNAP91', Aliases='AP180, CALM', description='synaptosome associated protein 91kDa', other_designations='91 kDa synaptosomal-associated protein|assembly protein, 180kDa|clathrin coat-associated protein AP180|phosphoprotein F1-20|synaptosomal-associated protein, 91kDa homolog', map_location='6q14.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83552885, end_position_on_the_genomic_accession=83709408, orientation='minus', exon_count=36, OMIM=607923, no_hdr0=''),
  9894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9894, CurrentID=0, Status='live', Symbol='TELO2', Aliases='CLK2, TEL2', description='telomere maintenance 2', other_designations='TEL2, telomere maintenance 2, homolog|protein clk-2 homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1493344, end_position_on_the_genomic_accession=1510459, orientation='plus', exon_count=23, OMIM=611140, no_hdr0=''),
  9895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9895, CurrentID=0, Status='live', Symbol='TECPR2', Aliases='KIAA0329, SPG49', description='tectonin beta-propeller repeat containing 2', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102362963, end_position_on_the_genomic_accession=102502481, orientation='plus', exon_count=21, OMIM=615000, no_hdr0=''),
  9896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9896, CurrentID=0, Status='live', Symbol='FIG4', Aliases='ALS11, BTOP, CMT4J, KIAA0274, SAC3, YVS, dJ249I4.1', description='FIG4 phosphoinositide 5-phosphatase', other_designations='FIG4 homolog, SAC domain containing lipid phosphatase|FIG4 homolog, SAC1 domain containing lipid phosphatase|FIG4 homolog, SAC1 lipid phosphatase domain containing|SAC domain-containing protein 3|Sac domain-containing inositol phosphatase 3|phosphatidylinositol 3,5-bisphosphate 5-phosphatase', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109691221, end_position_on_the_genomic_accession=109825431, orientation='plus', exon_count=24, OMIM=609390, no_hdr0=''),
  9897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9897, CurrentID=0, Status='live', Symbol='KIAA0196', Aliases='RTSC, SPG8', description='KIAA0196', other_designations='strumpellin', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125024260, end_position_on_the_genomic_accession=125091819, orientation='minus', exon_count=31, OMIM=610657, no_hdr0=''),
  9898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9898, CurrentID=0, Status='live', Symbol='UBAP2L', Aliases='NICE-4', description='ubiquitin associated protein 2 like', other_designations='protein NICE-4|ubiquitin associated protein 2-like', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154219951, end_position_on_the_genomic_accession=154271510, orientation='plus', exon_count=33, OMIM=616472, no_hdr0=''),
  9899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9899, CurrentID=0, Status='live', Symbol='SV2B', Aliases='HsT19680', description='synaptic vesicle glycoprotein 2B', other_designations='synaptic vesicle protein 2B homolog', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91099721, end_position_on_the_genomic_accession=91301309, orientation='plus', exon_count=17, OMIM=185861, no_hdr0=''),
  9900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9900, CurrentID=0, Status='live', Symbol='SV2A', Aliases='SV2', description='synaptic vesicle glycoprotein 2A', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149903318, end_position_on_the_genomic_accession=149917882, orientation='minus', exon_count=13, OMIM=185860, no_hdr0=''),
  9901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9901, CurrentID=0, Status='live', Symbol='SRGAP3', Aliases='ARHGAP14, MEGAP, SRGAP2, WRP', description='SLIT-ROBO Rho GTPase activating protein 3', other_designations='SLIT-ROBO Rho GTPase activating protein 2|WAVE-associated Rac GTPase activating protein|mental disorder-associated GAP|rho GTPase-activating protein 14', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8980591, end_position_on_the_genomic_accession=9362929, orientation='minus', exon_count=32, OMIM=606525, no_hdr0=''),
  9902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9902, CurrentID=0, Status='live', Symbol='MRC2', Aliases='CD280, CLEC13E, ENDO180, UPARAP', description='mannose receptor, C type 2', other_designations='C-type lectin domain family 13 member E|UPAR-associated protein|endocytic receptor (macrophage mannose receptor family)|endocytic receptor 180|macrophage mannose receptor 2|urokinase-type plasminogen activator receptor-associated protein', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62626515, end_position_on_the_genomic_accession=62693601, orientation='plus', exon_count=31, OMIM=612264, no_hdr0=''),
  9903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9903, CurrentID=0, Status='live', Symbol='KLHL21', Aliases='', description='kelch like family member 21', other_designations='kelch-like 21|kelch-like family member 21', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6590724, end_position_on_the_genomic_accession=6602898, orientation='minus', exon_count=5, OMIM=616262, no_hdr0=''),
  9904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9904, CurrentID=0, Status='live', Symbol='RBM19', Aliases='', description='RNA binding motif protein 19', other_designations='RNA-binding motif protein 19', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113816738, end_position_on_the_genomic_accession=113966371, orientation='minus', exon_count=27, OMIM=616444, no_hdr0=''),
  9905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9905, CurrentID=0, Status='live', Symbol='SGSM2', Aliases='RUTBC1', description='small G protein signaling modulator 2', other_designations='RUN and TBC1 domain containing 1|RUN and TBC1 domain-containing protein 1|small G protein signaling modulator 2 protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2337512, end_position_on_the_genomic_accession=2381074, orientation='plus', exon_count=28, OMIM=611418, no_hdr0=''),
  9906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9906, CurrentID=0, Status='live', Symbol='SLC35E2', Aliases='', description='solute carrier family 35 member E2', other_designations='solute carrier family 35, member E2', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1724838, end_position_on_the_genomic_accession=1745999, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  9907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9907, CurrentID=0, Status='live', Symbol='AP5Z1', Aliases='KIAA0415, SPG48, zeta', description='adaptor related protein complex 5, zeta 1 subunit', other_designations='adapter-related protein complex 5 zeta subunit|adaptor-related protein complex 5 zeta subunit|adaptor-related protein complex 5, zeta 1 subunit|zeta5', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4775631, end_position_on_the_genomic_accession=4794395, orientation='plus', exon_count=17, OMIM=613653, no_hdr0=''),
  9908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9908, CurrentID=0, Status='live', Symbol='G3BP2', Aliases='', description='GTPase activating protein (SH3 domain) binding protein 2', other_designations='G3BP-2|GAP SH3 domain-binding protein 2|Ras-GTPase activating protein SH3 domain-binding protein 2', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75642769, end_position_on_the_genomic_accession=75724525, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  9909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9909, CurrentID=0, Status='live', Symbol='DENND4B', Aliases='KIAA0476', description='DENN/MADD domain containing 4B', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153929501, end_position_on_the_genomic_accession=153946681, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  9910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9910, CurrentID=0, Status='live', Symbol='RABGAP1L', Aliases='HHL, TBC1D18', description='RAB GTPase activating protein 1-like', other_designations='TBC1 domain family, member 18|expressed in hematopoietic cells, heart, liver (HLL)', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174159414, end_position_on_the_genomic_accession=174995308, orientation='plus', exon_count=33, OMIM=609238, no_hdr0=''),
  9911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9911, CurrentID=0, Status='live', Symbol='TMCC2', Aliases='HUCEP11', description='transmembrane and coiled-coil domain family 2', other_designations='cerebral protein 11', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205227910, end_position_on_the_genomic_accession=205273343, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  9912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9912, CurrentID=0, Status='live', Symbol='ARHGAP44', Aliases='NPC-A-10, RICH2', description='Rho GTPase activating protein 44', other_designations='RICH-2|Rho-type GTPase-activating protein RICH2|RhoGAP interacting with CIP4 homologs protein 2|rho GTPase-activating protein RICH2', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12789512, end_position_on_the_genomic_accession=12991643, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  9913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9913, CurrentID=0, Status='live', Symbol='SUPT7L', Aliases='SPT7L, STAF65, STAF65(gamma), STAF65G, SUPT7H', description='SPT7-like STAGA complex gamma subunit', other_designations='SPTF-associated factor 65 gamma|STAF65gamma|STAGA complex 65 gamma subunit|adenocarcinoma antigen ART1|suppressor of Ty 7 (S. cerevisiae)-like|suppressor of Ty 7-like', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27642568, end_position_on_the_genomic_accession=27663973, orientation='minus', exon_count=8, OMIM=612762, no_hdr0=''),
  9914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9914, CurrentID=0, Status='live', Symbol='ATP2C2', Aliases='SPCA2', description='ATPase, Ca++ transporting, type 2C, member 2', other_designations='ATPase 2C2|secretory pathway Ca(2+)-ATPase 2|secretory pathway calcium ATPase 2', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84368523, end_position_on_the_genomic_accession=84464187, orientation='plus', exon_count=31, OMIM=613082, no_hdr0=''),
  9915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9915, CurrentID=0, Status='live', Symbol='ARNT2', Aliases='WEDAS, bHLHe1', description='aryl hydrocarbon receptor nuclear translocator 2', other_designations='ARNT protein 2|aryl-hydrocarbon receptor nuclear translocator 2|class E basic helix-loop-helix protein 1', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80404350, end_position_on_the_genomic_accession=80597937, orientation='plus', exon_count=19, OMIM=606036, no_hdr0=''),
  9917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9917, CurrentID=0, Status='live', Symbol='FAM20B', Aliases='gxk1', description='family with sequence similarity 20 member B', other_designations='family with sequence similarity 20, member B|xylose kinase', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179025809, end_position_on_the_genomic_accession=179076567, orientation='plus', exon_count=11, OMIM=611063, no_hdr0=''),
  9918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9918, CurrentID=0, Status='live', Symbol='NCAPD2', Aliases='CAP-D2, CNAP1, hCAP-D2', description='non-SMC condensin I complex subunit D2', other_designations='XCAP-D2 homolog|chromosome condensation related SMC associated protein 1|chromosome condensation-related SMC-associated protein 1|chromosome-associated protein D2|non-SMC condensin I complex, subunit D2', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6494132, end_position_on_the_genomic_accession=6531966, orientation='plus', exon_count=32, OMIM=615638, no_hdr0=''),
  9919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9919, CurrentID=0, Status='live', Symbol='SEC16A', Aliases='KIAA0310, SEC16L, p250', description='SEC16 homolog A, endoplasmic reticulum export factor', other_designations='SEC16 homolog A|protein SEC16 homolog A', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136440096, end_position_on_the_genomic_accession=136484549, orientation='minus', exon_count=36, OMIM=612854, no_hdr0=''),
  9920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9920, CurrentID=0, Status='live', Symbol='KBTBD11', Aliases='KLHDC7C', description='kelch repeat and BTB (POZ) domain containing 11', other_designations='chronic myelogenous leukemia-associated protein|kelch domain-containing protein 7B', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1970710, end_position_on_the_genomic_accession=2006943, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  9921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9921, CurrentID=0, Status='live', Symbol='RNF10', Aliases='RIE2', description='ring finger protein 10', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120534308, end_position_on_the_genomic_accession=120577594, orientation='plus', exon_count=19, OMIM=615998, no_hdr0=''),
  9922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9922, CurrentID=0, Status='live', Symbol='IQSEC1', Aliases='ARF-GEP100, ARFGEP100, BRAG2, GEP100', description='IQ motif and Sec7 domain 1', other_designations='ADP-ribosylation factors guanine nucleotide-exchange protein 100|ADP-ribosylation factors guanine nucleotide-exchange protein 2|brefeldin A-resistant ARF-GEF2|brefeldin-resistant Arf-GEF 2 protein', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12897043, end_position_on_the_genomic_accession=13282998, orientation='minus', exon_count=26, OMIM=610166, no_hdr0=''),
  9923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9923, CurrentID=0, Status='live', Symbol='ZBTB40', Aliases='ZNF923', description='zinc finger and BTB domain containing 40', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22428838, end_position_on_the_genomic_accession=22531157, orientation='plus', exon_count=21, OMIM=612106, no_hdr0=''),
  9924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9924, CurrentID=0, Status='live', Symbol='PAN2', Aliases='USP52', description='PAN2 poly(A) specific ribonuclease subunit', other_designations='PAB-dependent poly(A)-specific ribonuclease subunit 2|PAB1P-dependent poly(A)-nuclease|PABP-dependent poly(A) nuclease 2|PAN deadenylation complex catalytic subunit 2|PAN2 poly(A) specific ribonuclease subunit homolog|PAN2 polyA specific ribonuclease subunit|inactive ubiquitin carboxyl-terminal hydrolase 52|ubiquitin specific peptidase 52|ubiquitin specific protease 52', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56316223, end_position_on_the_genomic_accession=56334053, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  9925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9925, CurrentID=0, Status='live', Symbol='ZBTB5', Aliases='', description='zinc finger and BTB domain containing 5', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37438102, end_position_on_the_genomic_accession=37465410, orientation='minus', exon_count=3, OMIM=616590, no_hdr0=''),
  9926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9926, CurrentID=0, Status='live', Symbol='LPGAT1', Aliases='FAM34A, FAM34A1, NET8', description='lysophosphatidylglycerol acyltransferase 1', other_designations='family with sequence similarity 34, member A', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211743457, end_position_on_the_genomic_accession=211830832, orientation='minus', exon_count=12, OMIM=610473, no_hdr0=''),
  9927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9927, CurrentID=0, Status='live', Symbol='MFN2', Aliases='CMT2A, CMT2A2, CPRP1, HSG, MARF', description='mitofusin 2', other_designations='hyperplasia suppressor|mitochondrial assembly regulatory factor|transmembrane GTPase MFN2', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11980181, end_position_on_the_genomic_accession=12013515, orientation='plus', exon_count=20, OMIM=608507, no_hdr0=''),
  9928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9928, CurrentID=0, Status='live', Symbol='KIF14', Aliases='MKS12', description='kinesin family member 14', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200551497, end_position_on_the_genomic_accession=200620791, orientation='minus', exon_count=32, OMIM=611279, no_hdr0=''),
  9929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9929, CurrentID=0, Status='live', Symbol='JOSD1', Aliases='dJ508I15.2', description='Josephin domain containing 1', other_designations='josephin domain-containing 1|josephin domain-containing protein 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38685543, end_position_on_the_genomic_accession=38701418, orientation='minus', exon_count=7, OMIM=615323, no_hdr0=''),
  9931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9931, CurrentID=0, Status='live', Symbol='HELZ', Aliases='DHRC, DRHC, HUMORF5', description='helicase with zinc finger', other_designations='down-regulated in human cancers protein|helicase with zinc finger domain', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67070438, end_position_on_the_genomic_accession=67245963, orientation='minus', exon_count=37, OMIM=606699, no_hdr0=''),
  9933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9933, CurrentID=0, Status='live', Symbol='PUM3', Aliases='HA-8, HLA-HA8, KIAA0020, PEN, PUF-A, PUF6, XTP5', description='pumilio RNA binding family member 3', other_designations='HBV X-transactivated gene 5 protein|HBV XAg-transactivated protein 5|minor histocompatibility antigen HA-8|penguin homolog|protein 5 transactivated by hepatitis B virus X antigen (HBxAg)|pumilio RNA-binding family member 3', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2804152, end_position_on_the_genomic_accession=2844130, orientation='minus', exon_count=18, OMIM=609960, no_hdr0=''),
  9934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9934, CurrentID=0, Status='live', Symbol='P2RY14', Aliases='BPR105, GPR105, P2Y14', description='purinergic receptor P2Y, G-protein coupled, 14', other_designations='G protein coupled receptor for UDP-glucose|G protein-coupled receptor 105|G-protein coupled receptor 105|P2Y(14) receptor|P2Y14 receptor|UDP-glucose receptor', map_location='3q24-q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151212115, end_position_on_the_genomic_accession=151278628, orientation='minus', exon_count=5, OMIM=610116, no_hdr0=''),
  9935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9935, CurrentID=0, Status='live', Symbol='MAFB', Aliases='KRML, MCTO', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog B', other_designations='Kreisler maf-related leucine zipper homolog|MAFB/Kreisler basic region/leucine zipper transcription factor', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40685848, end_position_on_the_genomic_accession=40689240, orientation='minus', exon_count=1, OMIM=608968, no_hdr0=''),
  9936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9936, CurrentID=0, Status='live', Symbol='CD302', Aliases='BIMLEC, CLEC13A, DCL-1, DCL1', description='CD302 molecule', other_designations='C-type lectin domain family 13, member A|DEC205-associated C-type lectin 1|type I transmembrane C-type lectin receptor DCL-1', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159768628, end_position_on_the_genomic_accession=159798255, orientation='minus', exon_count=6, OMIM=612246, no_hdr0=''),
  9937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9937, CurrentID=0, Status='live', Symbol='DCLRE1A', Aliases='PSO2, SNM1, SNM1A', description='DNA cross-link repair 1A', other_designations='DNA-crosslink repair gene SNM1|SNM1 homolog A', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113834724, end_position_on_the_genomic_accession=113854404, orientation='minus', exon_count=10, OMIM=609682, no_hdr0=''),
  9938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9938, CurrentID=0, Status='live', Symbol='ARHGAP25', Aliases='HEL-S-308, KAIA0053', description='Rho GTPase activating protein 25', other_designations='epididymis secretory protein Li 308|rho-type GTPase-activating protein 25', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68710505, end_position_on_the_genomic_accession=68826833, orientation='plus', exon_count=15, OMIM=610587, no_hdr0=''),
  9939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9939, CurrentID=0, Status='live', Symbol='RBM8A', Aliases='BOV-1A, BOV-1B, BOV-1C, C1DELq21.1, DEL1q21.1, MDS014, RBM8, RBM8B, TAR, Y14, ZNRP, ZRNP1', description='RNA binding motif protein 8A', other_designations='BOV-1|RNA binding motif protein 8B|RNA-binding protein Y14|binder of OVCA1-1|ribonucleoprotein RBM8|ribonucleoprotein RBM8A', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145921556, end_position_on_the_genomic_accession=145927536, orientation='minus', exon_count=6, OMIM=605313, no_hdr0=''),
  9940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9940, CurrentID=0, Status='live', Symbol='DLEC1', Aliases='CFAP81, DLC1, F56', description='deleted in lung and esophageal cancer 1', other_designations='DLC-1|cilia and flagella associated protein 81|deleted in lung and esophageal cancer 1 transcript varient 2', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38038605, end_position_on_the_genomic_accession=38122737, orientation='plus', exon_count=38, OMIM=604050, no_hdr0=''),
  9941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9941, CurrentID=0, Status='live', Symbol='EXOG', Aliases='ENDOGL1, ENDOGL2, ENGL, ENGL-a, ENGL-b, ENGLA, ENGLB', description="endo/exonuclease (5'-3'), endonuclease G-like", other_designations='endo G-like 1|endonuclease G-like 1|endonuclease G-like 2', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38496127, end_position_on_the_genomic_accession=38526305, orientation='plus', exon_count=15, OMIM=604051, no_hdr0=''),
  9942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9942, CurrentID=0, Status='live', Symbol='XYLB', Aliases='', description='xylulokinase homolog (H. influenzae)', other_designations='', map_location='3p22-p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38346760, end_position_on_the_genomic_accession=38453041, orientation='plus', exon_count=23, OMIM=604049, no_hdr0=''),
  9943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9943, CurrentID=0, Status='live', Symbol='OXSR1', Aliases='OSR1', description='oxidative stress responsive 1', other_designations='oxidative stress-responsive 1 protein|oxidative-stress responsive 1', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38165260, end_position_on_the_genomic_accession=38255488, orientation='plus', exon_count=21, OMIM=604046, no_hdr0=''),
  9945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9945, CurrentID=0, Status='live', Symbol='GFPT2', Aliases='GFAT, GFAT 2, GFAT2', description='glutamine-fructose-6-phosphate transaminase 2', other_designations='D-fructose-6-phosphate amidotransferase 2|glucosamine--fructose-6-phosphate aminotransferase [isomerizing] 2|glutamine: fructose-6-phosphate aminotransferase 2|glutamine:fructose 6 phosphate amidotransferase 2|glutamine:fructose-6-phosphate amidotransferase 2|hexosephosphate aminotransferase 2', map_location='5q34-q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180300698, end_position_on_the_genomic_accession=180353336, orientation='minus', exon_count=20, OMIM=603865, no_hdr0=''),
  9946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9946, CurrentID=0, Status='live', Symbol='CRYZL1', Aliases='4P11, QOH-1', description='crystallin zeta like 1', other_designations='crystallin, zeta (quinone reductase)-like 1|protein 4P11|quinone oxidoreductase homolog 1|quinone reductase-like 1|zeta-crystallin homolog', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33589341, end_position_on_the_genomic_accession=33641854, orientation='minus', exon_count=13, OMIM=603920, no_hdr0=''),
  9947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9947, CurrentID=0, Status='live', Symbol='MAGEC1', Aliases='CT7, CT7.1', description='MAGE family member C1', other_designations='MAGE-C1 antigen|cancer/testis antigen 7.1|cancer/testis antigen family 7, member 1|melanoma antigen family C, 1|melanoma antigen family C1', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141903856, end_position_on_the_genomic_accession=141909401, orientation='plus', exon_count=4, OMIM=300223, no_hdr0=''),
  9948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9948, CurrentID=0, Status='live', Symbol='WDR1', Aliases='AIP1, HEL-S-52, NORI-1', description='WD repeat domain 1', other_designations='actin-interacting protein 1|epididymis secretory protein Li 52', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10074339, end_position_on_the_genomic_accession=10116949, orientation='minus', exon_count=16, OMIM=604734, no_hdr0=''),
  9949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9949, CurrentID=0, Status='live', Symbol='AMMECR1', Aliases='AMMERC1', description='Alport syndrome, mental retardation, midface hypoplasia and elliptocytosis chromosomal region gene 1', other_designations='Alport syndrome mental retardation midface hypoplasia and elliptocytosis chromosomal region protein 1', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110194186, end_position_on_the_genomic_accession=110440233, orientation='minus', exon_count=8, OMIM=300195, no_hdr0=''),
  9950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9950, CurrentID=0, Status='live', Symbol='GOLGA5', Aliases='GOLIM5, RFG5, ret-II', description='golgin A5', other_designations='RET-fused gene 5 protein|cell proliferation-inducing gene 31 protein|golgi autoantigen, golgin subfamily a, 5|golgi integral membrane protein 5|golgin-84', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92794231, end_position_on_the_genomic_accession=92839963, orientation='plus', exon_count=13, OMIM=606918, no_hdr0=''),
  9951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9951, CurrentID=0, Status='live', Symbol='HS3ST4', Aliases='3-OST-4, 30ST4, 3OST4, h3-OST-4', description='heparan sulfate-glucosamine 3-sulfotransferase 4', other_designations='heparan sulfate (glucosamine) 3-O-sulfotransferase 4|heparan sulfate 3-O-sulfotransferase 4|heparan sulfate 3-O-sulfotransferase-4|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 4', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25692026, end_position_on_the_genomic_accession=26137688, orientation='plus', exon_count=2, OMIM=604059, no_hdr0=''),
  9953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9953, CurrentID=0, Status='live', Symbol='HS3ST3B1', Aliases='3-OST-3B, 3OST3B1, h3-OST-3B', description='heparan sulfate-glucosamine 3-sulfotransferase 3B1', other_designations='heparan sulfate (glucosamine) 3-O-sulfotransferase 3B1|heparan sulfate 3-O-sulfotransferase 3B1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 3B1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14301050, end_position_on_the_genomic_accession=14349406, orientation='plus', exon_count=4, OMIM=604058, no_hdr0=''),
  9955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9955, CurrentID=0, Status='live', Symbol='HS3ST3A1', Aliases='3-OST-3A, 3OST3A1', description='heparan sulfate-glucosamine 3-sulfotransferase 3A1', other_designations='heparan sulfate (glucosamine) 3-O-sulfotransferase 3A1|heparan sulfate 3-O-sulfotransferase 3A1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 3A1|heparin-glucosamine 3-O-sulfotransferase', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13494032, end_position_on_the_genomic_accession=13601974, orientation='minus', exon_count=7, OMIM=604057, no_hdr0=''),
  9956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9956, CurrentID=0, Status='live', Symbol='HS3ST2', Aliases='30ST2, 3OST2', description='heparan sulfate-glucosamine 3-sulfotransferase 2', other_designations='3-OST-2|h3-OST-2|heparan sulfate (glucosamine) 3-O-sulfotransferase 2|heparan sulfate 3-O-sulfotransferase 2|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 2|heparin-glucosamine 3-O-sulfotransferase', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22814203, end_position_on_the_genomic_accession=22916338, orientation='plus', exon_count=4, OMIM=604056, no_hdr0=''),
  9957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9957, CurrentID=0, Status='live', Symbol='HS3ST1', Aliases='3OST, 3OST1', description='heparan sulfate-glucosamine 3-sulfotransferase 1', other_designations='3-OST-1|h3-OST-1|heparan sulfate (glucosamine) 3-O-sulfotransferase 1|heparan sulfate 3-O-sulfotransferase 1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 1|heparin-glucosamine 3-O-sulfotransferase', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11398364, end_position_on_the_genomic_accession=11429567, orientation='minus', exon_count=3, OMIM=603244, no_hdr0=''),
  9958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9958, CurrentID=0, Status='live', Symbol='USP15', Aliases='UNPH-2, UNPH4', description='ubiquitin specific peptidase 15', other_designations='deubiquitinating enzyme 15|ubiquitin thioesterase 15|ubiquitin thiolesterase 15|ubiquitin-specific-processing protease 15', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62260340, end_position_on_the_genomic_accession=62409721, orientation='plus', exon_count=25, OMIM=604731, no_hdr0=''),
  9959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9959, CurrentID=0, Status='live', Symbol='USP12P1', Aliases='UBH1, USP12', description='ubiquitin specific peptidase 12 pseudogene 1', other_designations='ubiquitin hydrolysing enzyme 1|ubiquitin specific protease 12 pseudogene 1', map_location='5q33-q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171251643, end_position_on_the_genomic_accession=171253952, orientation='plus', exon_count=0, OMIM=603091, no_hdr0=''),
  9960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9960, CurrentID=0, Status='live', Symbol='USP3', Aliases='SIH003, UBP', description='ubiquitin specific peptidase 3', other_designations='deubiquitinating enzyme 3|ubiquitin thioesterase 3|ubiquitin thiolesterase 3|ubiquitin-specific-processing protease 3', map_location='15q22.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63504511, end_position_on_the_genomic_accession=63594640, orientation='plus', exon_count=18, OMIM=604728, no_hdr0=''),
  9961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9961, CurrentID=0, Status='live', Symbol='MVP', Aliases='LRP, VAULT1', description='major vault protein', other_designations='lung resistance-related protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29820394, end_position_on_the_genomic_accession=29848039, orientation='plus', exon_count=15, OMIM=605088, no_hdr0=''),
  9962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9962, CurrentID=0, Status='live', Symbol='SLC23A2', Aliases='NBTL1, SLC23A1, SVCT2, YSPL2', description='solute carrier family 23 (ascorbic acid transporter), member 2', other_designations='Na(+)/L-ascorbic acid transporter 2|hSVCT2|nucleobase transporter-like 1 protein|sodium-dependent vitamin C transporter-2|solute carrier family 23 (nucleobase transporters), member 1|solute carrier family 23 (nucleobase transporters), member 2|yolk sac permease-like molecule 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4852356, end_position_on_the_genomic_accession=5010293, orientation='minus', exon_count=20, OMIM=603791, no_hdr0=''),
  9963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9963, CurrentID=0, Status='live', Symbol='SLC23A1', Aliases='SLC23A2, SVCT1, YSPL3', description='solute carrier family 23 (ascorbic acid transporter), member 1', other_designations='Na(+)/L-ascorbic acid transporter 1|hSVCT1|sodium-dependent vitamin C transporter 1|sodium-dependent vitamin C transporter-1|solute carrier family 23 (nucleobase transporters), member 1|solute carrier family 23 (nucleobase transporters), member 2|yolk sac permease-like molecule 3', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139367196, end_position_on_the_genomic_accession=139384553, orientation='minus', exon_count=17, OMIM=603790, no_hdr0=''),
  9964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9964, CurrentID=0, Status='live', Symbol='MYCLP3', Aliases='MYCL3', description='MYCL pseudogene 3', other_designations='v-myc avian myelocytomatosis viral oncogene homolog 3 (pseudogene)|v-myc myelocytomatosis viral oncogene homolog 1, lung carcinoma derived pseudogene|v-myc myelocytomatosis viral oncogene homolog 3 (pseudogene)|v-myc myelocytomatosis viral oncogene homolog 3 pseudogene', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143594846, end_position_on_the_genomic_accession=143596113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9965, CurrentID=0, Status='live', Symbol='FGF19', Aliases='', description='fibroblast growth factor 19', other_designations='FGF-19', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69698238, end_position_on_the_genomic_accession=69704338, orientation='minus', exon_count=3, OMIM=603891, no_hdr0=''),
  9966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9966, CurrentID=0, Status='live', Symbol='TNFSF15', Aliases='TL1, TL1A, VEGI, VEGI192A', description='tumor necrosis factor superfamily member 15', other_designations='TNF ligand-related molecule 1|TNF superfamily ligand TL1A|tumor necrosis factor (ligand) superfamily member 15|tumor necrosis factor (ligand) superfamily, member 15|vascular endothelial cell growth inhibitor|vascular endothelial growth inhibitor-192A', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114784635, end_position_on_the_genomic_accession=114806128, orientation='minus', exon_count=4, OMIM=604052, no_hdr0=''),
  9967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9967, CurrentID=0, Status='live', Symbol='THRAP3', Aliases='TRAP150', description='thyroid hormone receptor associated protein 3', other_designations='thyroid hormone receptor-associated protein complex 150 kDa component|thyroid hormone receptor-associated protein, 150 kDa subunit', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36224416, end_position_on_the_genomic_accession=36305357, orientation='plus', exon_count=14, OMIM=603809, no_hdr0=''),
  9968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9968, CurrentID=0, Status='live', Symbol='MED12', Aliases='ARC240, CAGH45, FGS1, HOPAS, OHDOX, OKS, OPA1, TNRC11, TRAP230, MED12', description='mediator complex subunit 12', other_designations='CAG repeat protein 45|OPA-containing protein|activator-recruited cofactor 240 kDa component|human opposite paired|mediator of RNA polymerase II transcription, subunit 12 homolog|putative mediator subunit 12|thyroid hormone receptor-associated protein complex 230 kDa component|thyroid hormone receptor-associated protein, 230 kDa subunit|trinucleotide repeat containing 11 (THR-associated protein, 230 kDa subunit)|trinucleotide repeat-containing gene 11 protein', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71118556, end_position_on_the_genomic_accession=71142454, orientation='plus', exon_count=45, OMIM=300188, no_hdr0=''),
  9969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9969, CurrentID=0, Status='live', Symbol='MED13', Aliases='ARC250, DRIP250, HSPC221, THRAP1, TRAP240', description='mediator complex subunit 13', other_designations='activator-recruited cofactor 250 kDa component|mediator of RNA polymerase II transcription, subunit 13 homolog|thyroid hormone receptor associated protein 1|thyroid hormone receptor-associated protein 1|thyroid hormone receptor-associated protein complex 240 kDa component|thyroid hormone receptor-associated protein complex component TRAP240|thyroid hormone receptor-associated protein, 240 kDa subunit|vitamin D3 receptor-interacting protein complex component DRIP250', map_location='17q22-q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61942605, end_position_on_the_genomic_accession=62065311, orientation='minus', exon_count=32, OMIM=603808, no_hdr0=''),
  9970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9970, CurrentID=0, Status='live', Symbol='NR1I3', Aliases='CAR, CAR1, MB67', description='nuclear receptor subfamily 1 group I member 3', other_designations='constitutive activator of retinoid response|constitutive active receptor|constitutive active response|constitutive androstane nuclear receptor variant 2|constitutive androstane nuclear receptor variant 3|constitutive androstane nuclear receptor variant 4|constitutive androstane nuclear receptor variant 5|constitutive androstane receptor|nuclear receptor subfamily 1, group I, member 3|orphan nuclear hormone receptor|orphan nuclear receptor MB67', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161229666, end_position_on_the_genomic_accession=161238598, orientation='minus', exon_count=9, OMIM=603881, no_hdr0=''),
  9971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9971, CurrentID=0, Status='live', Symbol='NR1H4', Aliases='BAR, FXR, HRR-1, HRR1, RIP14', description='nuclear receptor subfamily 1 group H member 4', other_designations='RXR-interacting protein 14|farnesoid X nuclear receptor|farnesoid X-activated receptor|farnesol receptor HRR-1|nuclear receptor subfamily 1, group H, member 4|retinoid X receptor-interacting protein 14', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100473759, end_position_on_the_genomic_accession=100563867, orientation='plus', exon_count=14, OMIM=603826, no_hdr0=''),
  9972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9972, CurrentID=0, Status='live', Symbol='NUP153', Aliases='HNUP153, N153', description='nucleoporin 153kDa', other_designations='153 kDa nucleoporin|nuclear pore complex protein hnup153|nucleoporin 153kD|nucleoporin Nup153', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17615035, end_position_on_the_genomic_accession=17706834, orientation='minus', exon_count=24, OMIM=603948, no_hdr0=''),
  9973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9973, CurrentID=0, Status='live', Symbol='CCS', Aliases='', description='copper chaperone for superoxide dismutase', other_designations='superoxide dismutase copper chaperone', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66593162, end_position_on_the_genomic_accession=66606019, orientation='plus', exon_count=8, OMIM=603864, no_hdr0=''),
  9975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9975, CurrentID=0, Status='live', Symbol='NR1D2', Aliases='BD73, EAR-1R, RVR', description='nuclear receptor subfamily 1 group D member 2', other_designations='V-erbA-related protein 1-related|nuclear receptor Rev-ErbA beta variant 1|nuclear receptor Rev-ErbA beta variant 2|nuclear receptor subfamily 1, group D, member 2|orphan nuclear hormone receptor BD73|rev-erb alpha-related receptor|rev-erb-beta|rev-erba-alpha-related receptor', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23945260, end_position_on_the_genomic_accession=23980618, orientation='plus', exon_count=10, OMIM=602304, no_hdr0=''),
  9976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9976, CurrentID=0, Status='live', Symbol='CLEC2B', Aliases='AICL, CLECSF2, HP10085, IFNRG1', description='C-type lectin domain family 2 member B', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 2 (activation-induced)|C-type lectin domain family 2, member B|C-type lectin superfamily member 2|IFN-alpha-2b-inducing-related protein 1|IFN-alpha2b-inducing related protein 1|activation-induced C-type lectin', map_location='12p13-p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9852369, end_position_on_the_genomic_accession=9869859, orientation='minus', exon_count=5, OMIM=603242, no_hdr0=''),
  9978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9978, CurrentID=0, Status='live', Symbol='RBX1', Aliases='BA554C12.1, RNF75, ROC1', description='ring-box 1, E3 ubiquitin protein ligase', other_designations='RING box protein 1|RING finger protein 75|RING-box protein 1|ZYP protein|regulator of cullins 1', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40951347, end_position_on_the_genomic_accession=40973015, orientation='plus', exon_count=5, OMIM=603814, no_hdr0=''),
  100533887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533887, CurrentID=0, Status='live', Symbol='LOC100533887', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  9980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9980, CurrentID=0, Status='live', Symbol='DOPEY2', Aliases='21orf5, C21orf5', description='dopey family member 2', other_designations='homolog of yeast DOP1', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36156783, end_position_on_the_genomic_accession=36294274, orientation='plus', exon_count=39, OMIM=604803, no_hdr0=''),
  9982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9982, CurrentID=0, Status='live', Symbol='FGFBP1', Aliases='FGF-BP, FGF-BP1, FGFBP, FGFBP-1, HBP17', description='fibroblast growth factor binding protein 1', other_designations='17 kDa HBGF-binding protein|17 kDa heparin-binding growth factor-binding protein|FGF-binding protein 1|heparin-binding growth factor binding protein', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15935569, end_position_on_the_genomic_accession=15938740, orientation='minus', exon_count=3, OMIM=607737, no_hdr0=''),
  9984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9984, CurrentID=0, Status='live', Symbol='THOC1', Aliases='HPR1, P84, P84N5', description='THO complex 1', other_designations='hTREX84|nuclear matrix protein p84|tho1', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=214520, end_position_on_the_genomic_accession=268059, orientation='minus', exon_count=23, OMIM=606930, no_hdr0=''),
  9985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9985, CurrentID=0, Status='live', Symbol='REC8', Aliases='HR21spBL1, Rec8p, REC8', description='REC8 meiotic recombination protein', other_designations='REC8 homolog|REC8-like 1|cohesin Rec8p|cohesion rec8p|human homolog of rad21, S. pombe|kleisin-alpha|meiotic recombination and sister chromatid cohesion phosphoprotein of the rad21p family|meiotic recombination protein REC8-like 1|recombination and sister chromatid cohesion protein homolog', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24171853, end_position_on_the_genomic_accession=24180408, orientation='plus', exon_count=20, OMIM=608193, no_hdr0=''),
  9986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9986, CurrentID=0, Status='live', Symbol='RCE1', Aliases='FACE2A, RCE1B, RCE1', description='Ras converting CAAX endopeptidase 1', other_designations='RCE1 homolog, prenyl protein peptidase|farnesylated protein-converting enzyme 2|farnesylated proteins-converting enzyme 2|prenyl protein-specific endoprotease 2', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66843412, end_position_on_the_genomic_accession=66846532, orientation='plus', exon_count=8, OMIM=605385, no_hdr0=''),
  9987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9987, CurrentID=0, Status='live', Symbol='HNRNPDL', Aliases='HNRNP, HNRPDL, JKTBP, JKTBP2, LGMD1G, laAUF1', description='heterogeneous nuclear ribonucleoprotein D like', other_designations='A+U-rich element RNA binding factor|AU-rich element RNA-binding factor|JKT41-binding protein|hnRNP D-like|hnRNP DL|protein laAUF1', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82422564, end_position_on_the_genomic_accession=82430225, orientation='minus', exon_count=9, OMIM=607137, no_hdr0=''),
  9988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9988, CurrentID=0, Status='live', Symbol='DMTF1', Aliases='DMP1, DMTF, MRUL, hDMP1', description='cyclin D binding myb-like transcription factor 1', other_designations='cyclin D-binding Myb-like protein|cyclin-D-interacting Myb-like protein 1|hDMTF1', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87152361, end_position_on_the_genomic_accession=87196332, orientation='plus', exon_count=22, OMIM=608491, no_hdr0=''),
  9989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9989, CurrentID=0, Status='live', Symbol='PPP4R1', Aliases='MEG1, PP4(Rmeg), PP4R1', description='protein phosphatase 4 regulatory subunit 1', other_designations='protein phosphatase 4, regulatory subunit 1|serine/threonine phosphatase 4', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9546791, end_position_on_the_genomic_accession=9615276, orientation='minus', exon_count=23, OMIM=604908, no_hdr0=''),
  9990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9990, CurrentID=0, Status='live', Symbol='SLC12A6', Aliases='ACCPN, KCC3, KCC3A, KCC3B', description='solute carrier family 12 (potassium/chloride transporter), member 6', other_designations='K-Cl cotransporter 3|electroneutral potassium-chloride cotransporter 3|potassium chloride cotransporter 3|potassium chloride cotransporter KCC3a-S3|potassium-chloride transporter-3a|potassium-chloride transporter-3b|solute carrier family 12 (potassium/chloride transporters), member 6', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34229996, end_position_on_the_genomic_accession=34338064, orientation='minus', exon_count=28, OMIM=604878, no_hdr0=''),
  9991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9991, CurrentID=0, Status='live', Symbol='PTBP3', Aliases='ROD1', description='polypyrimidine tract binding protein 3', other_designations='ROD1 regulator of differentiation 1|fission yeast differentiation regulator|regulator of differentiation 1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112217715, end_position_on_the_genomic_accession=112333664, orientation='minus', exon_count=17, OMIM=607527, no_hdr0=''),
  9992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9992, CurrentID=0, Status='live', Symbol='KCNE2', Aliases='ATFB4, LQT5, LQT6, MIRP1', description='potassium channel, voltage gated subfamily E regulatory beta subunit 2', other_designations='cardiac voltage-gated potassium channel accessory subunit 2|minK-related peptide 1|minK-related peptide-1|minimum potassium ion channel-related peptide 1|potassium channel subunit beta MiRP1|potassium channel subunit, MiRP1|potassium voltage-gated channel, Isk-related family, member 2|voltage-gated K+ channel subunit MIRP1', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34364024, end_position_on_the_genomic_accession=34371141, orientation='plus', exon_count=2, OMIM=603796, no_hdr0=''),
  9993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9993, CurrentID=0, Status='live', Symbol='DGCR2', Aliases='DGS-C, IDD, LAN, SEZ-12', description='DiGeorge syndrome critical region gene 2', other_designations='DiGeorge syndrome critical region protein 2|integral membrane protein deleted in DiGeorge syndrome', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19036282, end_position_on_the_genomic_accession=19122454, orientation='minus', exon_count=10, OMIM=600594, no_hdr0=''),
  9994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9994, CurrentID=0, Status='live', Symbol='CASP8AP2', Aliases='CED-4, FLASH, RIP25', description='caspase 8 associated protein 2', other_designations='CASP8 associated protein 2|FLASH homolog RIP25|FLICE associated huge|FLICE-associated huge protein|human FLASH', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89829900, end_position_on_the_genomic_accession=89874436, orientation='plus', exon_count=10, OMIM=606880, no_hdr0=''),
  9995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9995, CurrentID=0, Status='live', Symbol='ELK2BP', Aliases='ELK2.2, ELK2P2', description='ELK2B, member of ETS oncogene family, pseudogene', other_designations='ELK1, member of ETS oncogene family pseudogene|ELK2, member of ETS oncogene family, pseudogene 2', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105543992, end_position_on_the_genomic_accession=105550860, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9997, CurrentID=0, Status='live', Symbol='SCO2', Aliases='CEMCOX1, MYP6, SCO1L', description='SCO2 cytochrome c oxidase assembly protein', other_designations='SCO cytochrome oxidase deficient homolog 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50523568, end_position_on_the_genomic_accession=50526439, orientation='minus', exon_count=4, OMIM=604272, no_hdr0=''),
  10000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10000, CurrentID=0, Status='live', Symbol='AKT3', Aliases='MPPH, MPPH2, PKB-GAMMA, PKBG, PRKBG, RAC-PK-gamma, RAC-gamma, STK-2', description='v-akt murine thymoma viral oncogene homolog 3', other_designations='PKB gamma|RAC-gamma serine/threonine protein kinase|v-akt murine thymoma viral oncogene homolog 3 (protein kinase B, gamma)', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243488233, end_position_on_the_genomic_accession=243851079, orientation='minus', exon_count=23, OMIM=611223, no_hdr0=''),
  10001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10001, CurrentID=0, Status='live', Symbol='MED6', Aliases='ARC33, NY-REN-28', description='mediator complex subunit 6', other_designations='CTD-2540L5.5|activator-recruited cofactor 33 kDa component|renal carcinoma antigen NY-REN-28', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70583221, end_position_on_the_genomic_accession=70600690, orientation='minus', exon_count=9, OMIM=602984, no_hdr0=''),
  10002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10002, CurrentID=0, Status='live', Symbol='NR2E3', Aliases='ESCS, PNR, RNR, RP37, rd7', description='nuclear receptor subfamily 2 group E member 3', other_designations='nuclear receptor subfamily 2, group E, member 3|retina-specific nuclear receptor', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71807197, end_position_on_the_genomic_accession=71818666, orientation='plus', exon_count=9, OMIM=604485, no_hdr0=''),
  10003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10003, CurrentID=0, Status='live', Symbol='NAALAD2', Aliases='GCPIII, GPCIII, NAADALASE2, NAALADASE2', description='N-acetylated alpha-linked acidic dipeptidase 2', other_designations='N-acetylated alpha-linked acidic dipeptidase II|N-acetylated-alpha-linked acidic dipeptidase II|NAALADase II|glutamate carboxypeptidase III', map_location='11q14.3-q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90131573, end_position_on_the_genomic_accession=90199743, orientation='plus', exon_count=22, OMIM=611636, no_hdr0=''),
  10004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10004, CurrentID=0, Status='live', Symbol='NAALADL1', Aliases='I100, NAALADASEL', description='N-acetylated alpha-linked acidic dipeptidase-like 1', other_designations='100 kDa ileum brush border membrane protein|NAALADase L|ileal dipeptidylpeptidase|ileal peptidase I100', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65044818, end_position_on_the_genomic_accession=65059021, orientation='minus', exon_count=19, OMIM=602640, no_hdr0=''),
  10005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10005, CurrentID=0, Status='live', Symbol='ACOT8', Aliases='HNAACTE, PTE-1, PTE-2, PTE1, PTE2, hACTE-III, hTE', description='acyl-CoA thioesterase 8', other_designations='HIV-Nef associated acyl-CoA thioesterase|choloyl-CoA hydrolase|choloyl-coenzyme A thioesterase|long-chain fatty-acyl-CoA hydrolase|palmitoyl-CoA hydrolase|peroxisomal acyl-CoA thioesterase 1|peroxisomal acyl-coenzyme A thioester hydrolase 1|peroxisomal long-chain acyl-CoA thioesterase 1|thioesterase II|thioesterase III', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45841721, end_position_on_the_genomic_accession=45857409, orientation='minus', exon_count=7, OMIM=608123, no_hdr0=''),
  10006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10006, CurrentID=0, Status='live', Symbol='ABI1', Aliases='ABI-1, ABLBP4, E3B1, NAP1BP, SSH3BP, SSH3BP1', description='abl interactor 1', other_designations='Abelson interactor 1|Abl-interactor protein 1 long|abl-binding protein 4|abl-interactor 1|eps8 SH3 domain-binding protein|interactor protein AblBP4|nap1 binding protein|spectrin SH3 domain-binding protein 1', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26746596, end_position_on_the_genomic_accession=26861087, orientation='minus', exon_count=14, OMIM=603050, no_hdr0=''),
  10007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10007, CurrentID=0, Status='live', Symbol='GNPDA1', Aliases='GNP1, GNPDA, GNPI, GPI, HLN', description='glucosamine-6-phosphate deaminase 1', other_designations='GNPDA 1|glcN6P deaminase 1|oscillin', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142000669, end_position_on_the_genomic_accession=142013055, orientation='minus', exon_count=9, OMIM=601798, no_hdr0=''),
  10008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10008, CurrentID=0, Status='live', Symbol='KCNE3', Aliases='HOKPP, HYPP, MiRP2', description='potassium channel, voltage gated subfamily E regulatory beta subunit 3', other_designations='cardiac voltage-gated potassium channel accessory subunit|minK-related peptide 2|minimum potassium ion channel-related peptide 2|potassium channel subunit beta MiRP2|potassium voltage-gated channel, Isk-related family, member 3|voltage-gated K+ channel subunit MIRP2', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74454841, end_position_on_the_genomic_accession=74467557, orientation='minus', exon_count=3, OMIM=604433, no_hdr0=''),
  10009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10009, CurrentID=0, Status='live', Symbol='ZBTB33', Aliases='ZNF-kaiso, ZNF348', description='zinc finger and BTB domain containing 33', other_designations='WUGSC:H_DJ525N14.1|kaiso transcription factor|zinc finger and BTB domain-containing protein 33', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120250752, end_position_on_the_genomic_accession=120258396, orientation='plus', exon_count=3, OMIM=300329, no_hdr0=''),
  10010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10010, CurrentID=0, Status='live', Symbol='TANK', Aliases='I-TRAF, ITRAF, TRAF2', description='TRAF family member associated NFKB activator', other_designations='TRAF family member-associated NFKB activator|TRAF-interacting protein', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161136927, end_position_on_the_genomic_accession=161236176, orientation='plus', exon_count=12, OMIM=603893, no_hdr0=''),
  10011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10011, CurrentID=0, Status='live', Symbol='SRA1', Aliases='SRA, SRAP, STRAA1, pp7684', description='steroid receptor RNA activator 1', other_designations='steroid receptor RNA activator 1 (complexes with NCOA1)|steroid receptor RNA activator protein|steroid receptor coactivator', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140550067, end_position_on_the_genomic_accession=140558093, orientation='minus', exon_count=5, OMIM=603819, no_hdr0=''),
  403228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403228, CurrentID=0, Status='live', Symbol='OR13Z1P', Aliases='', description='olfactory receptor family 13 subfamily Z member 1 pseudogene', other_designations='olfactory receptor, family 13, subfamily Z, member 1 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147419006, end_position_on_the_genomic_accession=147419898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10013, CurrentID=0, Status='live', Symbol='HDAC6', Aliases='CPBHM, HD6, JM21, PPP1R90', description='histone deacetylase 6', other_designations='protein phosphatase 1, regulatory subunit 90', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48801896, end_position_on_the_genomic_accession=48824976, orientation='plus', exon_count=32, OMIM=300272, no_hdr0=''),
  10014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10014, CurrentID=0, Status='live', Symbol='HDAC5', Aliases='HD5, NY-CO-9', description='histone deacetylase 5', other_designations='antigen NY-CO-9', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44076753, end_position_on_the_genomic_accession=44123646, orientation='minus', exon_count=29, OMIM=605315, no_hdr0=''),
  10015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10015, CurrentID=0, Status='live', Symbol='PDCD6IP', Aliases='AIP1, ALIX, DRIP4, HP95', description='programmed cell death 6 interacting protein', other_designations='ALG-2 interacting protein 1|ALG-2-interacting protein X|PDCD6-interacting protein|apoptosis-linked gene 2-interacting protein X|dopamine receptor interacting protein 4', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33798571, end_position_on_the_genomic_accession=33869707, orientation='plus', exon_count=20, OMIM=608074, no_hdr0=''),
  10016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10016, CurrentID=0, Status='live', Symbol='PDCD6', Aliases='ALG-2, ALG2, PEF1B', description='programmed cell death 6', other_designations='apoptosis-linked gene 2 protein|probable calcium-binding protein ALG-2', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=271621, end_position_on_the_genomic_accession=314974, orientation='plus', exon_count=7, OMIM=601057, no_hdr0=''),
  10017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10017, CurrentID=0, Status='live', Symbol='BCL2L10', Aliases='BCL-B, Boo, Diva, bcl2-L-10', description='BCL2-like 10 (apoptosis facilitator)', other_designations='anti-apoptotic protein NrH|apoptosis regulator Bcl-B|death inducer binding to vBcl-2 and Apaf-1', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52109269, end_position_on_the_genomic_accession=52113042, orientation='minus', exon_count=2, OMIM=606910, no_hdr0=''),
  10018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10018, CurrentID=0, Status='live', Symbol='BCL2L11', Aliases='BAM, BIM, BOD', description='BCL2-like 11 (apoptosis facilitator)', other_designations='bcl-2 interacting mediator of cell death|bcl-2 interacting protein Bim|bcl-2-related ovarian death agonist', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111120618, end_position_on_the_genomic_accession=111168445, orientation='plus', exon_count=15, OMIM=603827, no_hdr0=''),
  10019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10019, CurrentID=0, Status='live', Symbol='SH2B3', Aliases='IDDM20, LNK', description='SH2B adaptor protein 3', other_designations='lymphocyte-specific adapter protein Lnk|signal transduction protein Lnk', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111405108, end_position_on_the_genomic_accession=111451624, orientation='plus', exon_count=12, OMIM=605093, no_hdr0=''),
  10020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10020, CurrentID=0, Status='live', Symbol='GNE', Aliases='DMRV, GLCNE, IBM2, NM, Uae1', description='glucosamine (UDP-N-acetyl)-2-epimerase/N-acetylmannosamine kinase', other_designations='N-acylmannosamine kinase|UDP-GlcNAc-2-epimerase/ManAc kinase|UDP-N-acetylglucosamine 2-epimerase/N-acetylmannosamine kinase|UDP-N-acetylglucosamine-2-epimerase/N-acetylmannosamine kinase', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36214441, end_position_on_the_genomic_accession=36277056, orientation='minus', exon_count=13, OMIM=603824, no_hdr0=''),
  10021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10021, CurrentID=0, Status='live', Symbol='HCN4', Aliases='SSS2', description='hyperpolarization activated cyclic nucleotide gated potassium channel 4', other_designations='hyperpolarization activated cyclic nucleotide-gated cation channel 4|hyperpolarization activated cyclic nucleotide-gated potassium channel 4', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73319859, end_position_on_the_genomic_accession=73369264, orientation='minus', exon_count=9, OMIM=605206, no_hdr0=''),
  10022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10022, CurrentID=0, Status='live', Symbol='INSL5', Aliases='PRO182, UNQ156', description='insulin like 5', other_designations='insulin-like 5|insulin-like peptide 5|prepro-INSL5', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66797741, end_position_on_the_genomic_accession=66801259, orientation='minus', exon_count=2, OMIM=606413, no_hdr0=''),
  10023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10023, CurrentID=0, Status='live', Symbol='FRAT1', Aliases='', description='frequently rearranged in advanced T-cell lymphomas 1', other_designations='FRAT-1', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97319265, end_position_on_the_genomic_accession=97321915, orientation='plus', exon_count=1, OMIM=602503, no_hdr0=''),
  10024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10024, CurrentID=0, Status='live', Symbol='TROAP', Aliases='TASTIN', description='trophinin associated protein', other_designations='trophinin assisting protein|trophinin-assisting protein (tastin)|trophinin-associated protein', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49322138, end_position_on_the_genomic_accession=49331733, orientation='plus', exon_count=17, OMIM=603872, no_hdr0=''),
  10025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10025, CurrentID=0, Status='live', Symbol='MED16', Aliases='DRIP92, THRAP5, TRAP95', description='mediator complex subunit 16', other_designations='thyroid hormone receptor-associated protein 5|thyroid hormone receptor-associated protein complex 95 kDa component|thyroid hormone receptor-associated protein, 95-kD subunit|vitamin D3 receptor-interacting protein complex 92 kDa component', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=867962, end_position_on_the_genomic_accession=893218, orientation='minus', exon_count=17, OMIM=604062, no_hdr0=''),
  10026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10026, CurrentID=0, Status='live', Symbol='PIGK', Aliases='GPI8', description='phosphatidylinositol glycan anchor biosynthesis class K', other_designations='GPI transamidase subunit|GPI8 homolog|PIG-K|phosphatidylinositol glycan anchor biosynthesis, class K|phosphatidylinositol glycan, class K|phosphatidylinositol-glycan biosynthesis class K protein', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77088981, end_position_on_the_genomic_accession=77219447, orientation='minus', exon_count=11, OMIM=605087, no_hdr0=''),
  100009607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009607, CurrentID=0, Status='live', Symbol='TRH-GTG1-9', Aliases='TRNAH5', description='transfer RNA-His (GTG) 1-9', other_designations='tRNA-His-GTG-1-9|transfer RNA histidine 5 (anticodon GUG)', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14433940, end_position_on_the_genomic_accession=14434011, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403244, CurrentID=0, Status='live', Symbol='OR2T35', Aliases='', description='olfactory receptor family 2 subfamily T member 35', other_designations='olfactory receptor OR1-66|olfactory receptor, family 2, subfamily T, member 35', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248638287, end_position_on_the_genomic_accession=248642735, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10029, CurrentID=0, Status='live', Symbol='CASKP1', Aliases='CASKP', description='calcium/calmodulin-dependent serine protein kinase (MAGUK family) pseudogene 1', other_designations='', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12929856, end_position_on_the_genomic_accession=12948185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10031, CurrentID=0, Status='live', Symbol='SHROOM2P1', Aliases='APXLP', description='shroom family member 2 pseudogene 1', other_designations='apical protein, Xenopus laevis-like, pseudogene|apical protein-like pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12541058, end_position_on_the_genomic_accession=12543220, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403248, CurrentID=0, Status='live', Symbol='OR4A42P', Aliases='', description='olfactory receptor family 4 subfamily A member 42 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 42 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48610064, end_position_on_the_genomic_accession=48610997, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10033, CurrentID=0, Status='live', Symbol='ARSEP1', Aliases='ARSEP', description='arylsulfatase E pseudogene 1', other_designations='arylsulfatase E (chondrodysplasia punctata 1) pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12332258, end_position_on_the_genomic_accession=12347515, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10034, CurrentID=0, Status='live', Symbol='ARSDP1', Aliases='ARSDP, ARSDY1', description='arylsulfatase D pseudogene 1', other_designations='arylsulfatase D, Y-linked, pseudogene 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12354031, end_position_on_the_genomic_accession=12387409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403251, CurrentID=0, Status='live', Symbol='OR4A45P', Aliases='', description='olfactory receptor family 4 subfamily A member 45 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 45 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48579383, end_position_on_the_genomic_accession=48580448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10036, CurrentID=0, Status='live', Symbol='CHAF1A', Aliases='CAF-1, CAF1, CAF1B, CAF1P150, P150', description='chromatin assembly factor 1 subunit A', other_designations='CAF-1 subunit A|CAF-I 150 kDa subunit|CAF-I p150|CTB-50L17.7|chromatin assembly factor 1, subunit A (p150)|chromatin assembly factor I (150 kDa)|chromatin assembly factor I p150 subunit|hp150', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4402612, end_position_on_the_genomic_accession=4448308, orientation='plus', exon_count=18, OMIM=601246, no_hdr0=''),
  403253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403253, CurrentID=0, Status='live', Symbol='OR4A47', Aliases='OR11-113', description='olfactory receptor family 4 subfamily A member 47', other_designations='olfactory receptor OR11-113|olfactory receptor, family 4, subfamily A, member 47', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48483466, end_position_on_the_genomic_accession=48489774, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10038, CurrentID=0, Status='live', Symbol='PARP2', Aliases='ADPRT2, ADPRTL2, ADPRTL3, ARTD2, PARP-2, pADPRT-2', description='poly(ADP-ribose) polymerase 2', other_designations='ADP-ribosyltransferase (NAD+; poly(ADP-ribose) polymerase)-like 2|ADP-ribosyltransferase diphtheria toxin-like 2|ADPRT-2|NAD(+) ADP-ribosyltransferase 2|hPARP-2|poly (ADP-ribose) polymerase 2|poly (ADP-ribose) polymerase family, member 2|poly (ADP-ribosyl) transferase-like 2|poly[ADP-ribose] synthase 2|poly[ADP-ribose] synthetase 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20343614, end_position_on_the_genomic_accession=20357910, orientation='plus', exon_count=15, OMIM=607725, no_hdr0=''),
  10039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10039, CurrentID=0, Status='live', Symbol='PARP3', Aliases='ADPRT3, ADPRTL2, ADPRTL3, ARTD3, IRT1, PADPRT-3', description='poly(ADP-ribose) polymerase family member 3', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 2|ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 3|ADP-ribosyltransferase diphtheria toxin-like 3|ADPRT-3|NAD(+) ADP-ribosyltransferase 3|NAD+ ADP-ribosyltransferase 3|poly (ADP-ribose) polymerase family, member 3|poly(ADP-ribose) synthetase-3|poly[ADP-ribose] synthase 3|poly[ADP-ribose] synthetase 3', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51942287, end_position_on_the_genomic_accession=51948867, orientation='plus', exon_count=12, OMIM=607726, no_hdr0=''),
  10040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10040, CurrentID=0, Status='live', Symbol='TOM1L1', Aliases='OK/KNS-CL.3, SRCASM', description='target of myb1 like 1 membrane trafficking protein', other_designations='Src activating and signaling molecule|src-activating and signaling molecule protein|target of Myb-like protein 1|target of myb 1-like 1|target of myb1 (chicken)-like 1|target of myb1-like 1', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54900691, end_position_on_the_genomic_accession=54961967, orientation='plus', exon_count=17, OMIM=604701, no_hdr0=''),
  403257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403257, CurrentID=0, Status='live', Symbol='OR4C45', Aliases='', description='olfactory receptor family 4 subfamily C member 45', other_designations='olfactory receptor, family 4, subfamily C, member 45', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50032874, end_position_on_the_genomic_accession=50033794, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10042, CurrentID=0, Status='live', Symbol='HMGXB4', Aliases='HMG2L1, HMGBCG, THC211630', description='HMG box domain containing 4', other_designations='HMG box-containing protein 4|high mobility group protein 2-like 1|high-mobility group protein 2-like 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35241535, end_position_on_the_genomic_accession=35295807, orientation='plus', exon_count=14, OMIM=604702, no_hdr0=''),
  10043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10043, CurrentID=0, Status='live', Symbol='TOM1', Aliases='', description='target of myb1 membrane trafficking protein', other_designations='target of myb 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35299275, end_position_on_the_genomic_accession=35347994, orientation='plus', exon_count=18, OMIM=604700, no_hdr0=''),
  10044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10044, CurrentID=0, Status='live', Symbol='SH2D3C', Aliases='CHAT, NSP3, PRO34088, SHEP1', description='SH2 domain containing 3C', other_designations='SH2 domain-containing Eph receptor-binding protein 1|novel SH2-containing protein 3', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127738317, end_position_on_the_genomic_accession=127785159, orientation='minus', exon_count=20, OMIM=604722, no_hdr0=''),
  10045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10045, CurrentID=0, Status='live', Symbol='SH2D3A', Aliases='NSP1', description='SH2 domain containing 3A', other_designations='SH2 domain-containing 3A|novel SH2-containing protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6752160, end_position_on_the_genomic_accession=6767512, orientation='minus', exon_count=10, OMIM=604721, no_hdr0=''),
  10046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10046, CurrentID=0, Status='live', Symbol='MAMLD1', Aliases='CG1, CXorf6, F18, HYSP2', description='mastermind like domain containing 1', other_designations='mastermind-like domain containing 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150361589, end_position_on_the_genomic_accession=150514178, orientation='plus', exon_count=11, OMIM=300120, no_hdr0=''),
  10047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10047, CurrentID=0, Status='live', Symbol='CST8', Aliases='CRES, CTES5', description='cystatin 8', other_designations='cystatin 8 (cystatin-related epididymal specific)|cystatin-related epididymal spermatogenic protein', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23491076, end_position_on_the_genomic_accession=23507128, orientation='plus', exon_count=7, OMIM=608683, no_hdr0=''),
  10048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10048, CurrentID=0, Status='live', Symbol='RANBP9', Aliases='BPM-L, BPM90, RANBPM, RanBP7', description='RAN binding protein 9', other_designations='Ran Binding Protein in the Microtubule organizing center|novel centrosomal protein RanBPM|ran binding protein, centrosomal|ran-binding protein M', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13621498, end_position_on_the_genomic_accession=13711580, orientation='minus', exon_count=15, OMIM=603854, no_hdr0=''),
  10049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10049, CurrentID=0, Status='live', Symbol='DNAJB6', Aliases='DJ4, DnaJ, HHDJ1, HSJ-2, HSJ2, LGMD1D, LGMD1E, MRJ, MSJ-1', description='DnaJ heat shock protein family (Hsp40) member B6', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 6|DnaJ-like 2 protein|heat shock protein J2', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157337016, end_position_on_the_genomic_accession=157417439, orientation='plus', exon_count=11, OMIM=611332, no_hdr0=''),
  10050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10050, CurrentID=0, Status='live', Symbol='SLC17A4', Aliases='KAIA2138', description='solute carrier family 17, member 4', other_designations='Na/PO4 cotransporter|putative small intestine sodium-dependent phosphate transport protein|solute carrier family 17 (sodium phosphate), member 4', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25753359, end_position_on_the_genomic_accession=25781175, orientation='plus', exon_count=18, OMIM=604216, no_hdr0=''),
  10051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10051, CurrentID=0, Status='live', Symbol='SMC4', Aliases='CAP-C, CAPC, SMC-4L1, hCAP-C, SMC4', description='structural maintenance of chromosomes 4', other_designations='SMC protein 4|SMC4 structural maintenance of chromosomes 4-like 1|XCAP-C homolog|chromosome-associated polypeptide C', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160399304, end_position_on_the_genomic_accession=160434962, orientation='plus', exon_count=24, OMIM=605575, no_hdr0=''),
  10052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10052, CurrentID=0, Status='live', Symbol='GJC1', Aliases='CX45, GJA7', description='gap junction protein gamma 1', other_designations='CTC-296K1.4|connexin 45|connexin-45|gap junction alpha-7 protein|gap junction protein, gamma 1, 45kDa', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44793954, end_position_on_the_genomic_accession=44830811, orientation='minus', exon_count=8, OMIM=608655, no_hdr0=''),
  10053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10053, CurrentID=0, Status='live', Symbol='AP1M2', Aliases='AP1-mu2, HSMU1B, MU-1B, MU1B, mu2', description='adaptor related protein complex 1 mu 2 subunit', other_designations='AP-mu chain family member mu1B|HA1 47 kDa subunit 2|adaptor protein complex AP-1 mu-2 subunit|adaptor-related protein complex 1 mu 2 subunit|adaptor-related protein complex 1, mu 2 subunit|clathrin assembly protein complex 1 mu-2 medium chain 2|clathrin coat assembly protein AP47 2|clathrin coat associated protein AP47 2|clathrin-associated adaptor medium chain mu2|golgi adaptor AP-1 47 kDa protein|golgi adaptor HA1/AP1 adaptin mu-2 subunit|mu-adaptin 2|mu1B-adaptin', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10572671, end_position_on_the_genomic_accession=10587315, orientation='minus', exon_count=13, OMIM=607309, no_hdr0=''),
  10054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10054, CurrentID=0, Status='live', Symbol='UBA2', Aliases='ARX, HRIHFB2115, SAE2', description='ubiquitin-like modifier activating enzyme 2', other_designations='SUMO-1 activating enzyme subunit 2|SUMO1 activating enzyme subunit 2|UBA2, ubiquitin-activating enzyme E1 homolog|anthracycline-associated resistance ARX|ubiquitin-like 1-activating enzyme E1B|ubiquitin-like modifier-activating enzyme 2', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34428176, end_position_on_the_genomic_accession=34469893, orientation='plus', exon_count=18, OMIM=613295, no_hdr0=''),
  10055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10055, CurrentID=0, Status='live', Symbol='SAE1', Aliases='AOS1, HSPC140, SUA1, UBLE1A', description='SUMO1 activating enzyme subunit 1', other_designations='SUMO-1 activating enzyme E1 N subunit|SUMO-1 activating enzyme subunit 1|activator of SUMO1|sentrin/SUMO-activating protein AOS1|ubiquitin-like 1-activating enzyme E1A|ubiquitin-like protein SUMO-1 activating enzyme', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47130823, end_position_on_the_genomic_accession=47210636, orientation='plus', exon_count=12, OMIM=613294, no_hdr0=''),
  10056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10056, CurrentID=0, Status='live', Symbol='FARSB', Aliases='FARSLB, FRSB, HSPC173, PheHB, PheRS', description='phenylalanyl-tRNA synthetase beta subunit', other_designations='phenylalanine tRNA ligase 1, beta, cytoplasmic|phenylalanine--tRNA ligase beta chain|phenylalanine-tRNA ligase beta chain|phenylalanine-tRNA synthetase-like, beta subunit|phenylalanyl-tRNA synthetase beta chain|phenylalanyl-tRNA synthetase beta-subunit|phenylalanyl-tRNA synthetase, beta subunit|phenylalanyl-tRNA synthetase-like, beta subunit', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222571443, end_position_on_the_genomic_accession=222656355, orientation='minus', exon_count=19, OMIM=609690, no_hdr0=''),
  10057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10057, CurrentID=0, Status='live', Symbol='ABCC5', Aliases='ABC33, EST277145, MOAT-C, MOATC, MRP5, SMRP, pABC11', description='ATP binding cassette subfamily C member 5', other_designations='ATP-binding cassette sub-family C member 5|ATP-binding cassette, sub-family C (CFTR/MRP), member 5|canalicular multispecific organic anion transporter C|multi-specific organic anion transporter C', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183919934, end_position_on_the_genomic_accession=184018010, orientation='minus', exon_count=36, OMIM=605251, no_hdr0=''),
  10058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10058, CurrentID=0, Status='live', Symbol='ABCB6', Aliases='ABC, ABC14, DUH3, LAN, MCOPCB7, MTABC3, PRP, umat', description='ATP binding cassette subfamily B member 6 (Langereis blood group)', other_designations='ATP-binding cassette half-transporter|ATP-binding cassette, sub-family B (MDR/TAP), member 6 (Langereis blood group)|P-glycoprotein-related protein|mitochondrial ABC transporter 3|mt-ABC transporter 3|ubiquitously-expressed mammalian ABC half transporter', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219209766, end_position_on_the_genomic_accession=219218990, orientation='minus', exon_count=19, OMIM=605452, no_hdr0=''),
  10059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10059, CurrentID=0, Status='live', Symbol='DNM1L', Aliases='DLP1, DRP1, DVLP, DYMPLE, EMPF, HDYNIV', description='dynamin 1-like', other_designations='Dnm1p/Vps1p-like protein|dynamin family member proline-rich carboxyl-terminal domain less|dynamin-like protein 4|dynamin-like protein IV|dynamin-related protein 1', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32679200, end_position_on_the_genomic_accession=32745650, orientation='plus', exon_count=22, OMIM=603850, no_hdr0=''),
  10060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10060, CurrentID=0, Status='live', Symbol='ABCC9', Aliases='ABC37, ATFB12, CANTU, CMD1O, SUR2', description='ATP binding cassette subfamily C member 9', other_designations='ATP-binding cassette transporter sub-family C member 9|ATP-binding cassette, sub-family C (CFTR/MRP), member 9|sulfonylurea receptor 2', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21797389, end_position_on_the_genomic_accession=21941863, orientation='minus', exon_count=42, OMIM=601439, no_hdr0=''),
  10061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10061, CurrentID=0, Status='live', Symbol='ABCF2', Aliases='ABC28, EST133090, HUSSY-18, HUSSY18', description='ATP binding cassette subfamily F member 2', other_designations='ABC-type transport protein|ATP-binding cassette, sub-family F (GCN20), member 2|iron-inhibited ABC transporter 2', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151207830, end_position_on_the_genomic_accession=151227374, orientation='minus', exon_count=17, OMIM=612510, no_hdr0=''),
  10062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10062, CurrentID=0, Status='live', Symbol='NR1H3', Aliases='LXR-a, LXRA, RLD-1', description='nuclear receptor subfamily 1 group H member 3', other_designations='liver X nuclear receptor alpha variant 1|nuclear receptor subfamily 1, group H, member 3', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47248300, end_position_on_the_genomic_accession=47269033, orientation='plus', exon_count=16, OMIM=602423, no_hdr0=''),
  10063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10063, CurrentID=0, Status='live', Symbol='COX17', Aliases='', description='COX17, cytochrome c oxidase copper chaperone', other_designations='COX17 cytochrome c oxidase assembly homolog|COX17 cytochrome c oxidase copper chaperone|cytochrome c oxidase 17 copper chaperone|cytochrome c oxidase assembly homolog 17|human homolog of yeast mitochondrial copper recruitment', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119669525, end_position_on_the_genomic_accession=119677396, orientation='minus', exon_count=3, OMIM=604813, no_hdr0=''),
  10064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10064, CurrentID=0, Status='live', Symbol='VDAC1P2', Aliases='VDAC1LP', description='voltage-dependent anion channel 1 pseudogene 2', other_designations='voltage-dependent anion channel 1-like pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49631635, end_position_on_the_genomic_accession=49633359, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403281, CurrentID=0, Status='live', Symbol='OR6C64P', Aliases='', description='olfactory receptor family 6 subfamily C member 64 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 64 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55522593, end_position_on_the_genomic_accession=55523471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10066, CurrentID=0, Status='live', Symbol='SCAMP2', Aliases='', description='secretory carrier membrane protein 2', other_designations='', map_location='15q23-q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74843730, end_position_on_the_genomic_accession=74873365, orientation='minus', exon_count=11, OMIM=606912, no_hdr0=''),
  10067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10067, CurrentID=0, Status='live', Symbol='SCAMP3', Aliases='C1orf3', description='secretory carrier membrane protein 3', other_designations='propin 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155255979, end_position_on_the_genomic_accession=155262434, orientation='minus', exon_count=9, OMIM=606913, no_hdr0=''),
  10068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10068, CurrentID=0, Status='live', Symbol='IL18BP', Aliases='IL18BPa', description='interleukin 18 binding protein', other_designations='IL-18BP|MC51L-53L-54L homolog gene product|tadekinig-alfa', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71997944, end_position_on_the_genomic_accession=72002919, orientation='plus', exon_count=8, OMIM=604113, no_hdr0=''),
  10069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10069, CurrentID=0, Status='live', Symbol='RWDD2B', Aliases='C21orf6, GL011', description='RWD domain containing 2B', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29004778, end_position_on_the_genomic_accession=29019371, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  10071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10071, CurrentID=0, Status='live', Symbol='MUC12', Aliases='MUC-11, MUC-12, MUC11', description='mucin 12, cell surface associated', other_designations='mucin 11|mucin-11', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100969623, end_position_on_the_genomic_accession=101018949, orientation='plus', exon_count=12, OMIM=604609, no_hdr0=''),
  10072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10072, CurrentID=0, Status='live', Symbol='DPP3', Aliases='DPPIII', description='dipeptidyl-peptidase 3', other_designations='DPP III|dipeptidyl aminopeptidase III|dipeptidyl arylamidase III|dipeptidyl peptidase III|enkephalinase B', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66480013, end_position_on_the_genomic_accession=66509659, orientation='plus', exon_count=18, OMIM=606818, no_hdr0=''),
  10073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10073, CurrentID=0, Status='live', Symbol='SNUPN', Aliases='KPNBL, RNUT1, Snurportin1', description='snurportin 1', other_designations='RNA U transporter 1|RNA, U transporter 1', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75598083, end_position_on_the_genomic_accession=75626378, orientation='minus', exon_count=10, OMIM=607902, no_hdr0=''),
  10075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10075, CurrentID=0, Status='live', Symbol='HUWE1', Aliases='ARF-BP1, HECTH9, HSPC272, Ib772, LASU1, MULE, URE-B1, UREB1', description='HECT, UBA and WWE domain containing 1, E3 ubiquitin protein ligase', other_designations='ARF-binding protein 1|BJ-HCC-24 tumor antigen|HECT domain protein LASU1|Mcl-1 ubiquitin ligase E3|URE-binding protein 1|homologous to E6AP carboxyl terminus homologous protein 9|large structure of UREB1|upstream regulatory element-binding protein 1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53532096, end_position_on_the_genomic_accession=53686729, orientation='minus', exon_count=90, OMIM=300697, no_hdr0=''),
  10076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10076, CurrentID=0, Status='live', Symbol='PTPRU', Aliases='FMI, PCP-2, PTP, PTP-J, PTP-PI, PTP-RO, PTPPSI, PTPRO, PTPU2, R-PTP-PSI, R-PTP-U, hPTP-J', description='protein tyrosine phosphatase, receptor type, U', other_designations='PTP pi|Receptor protein tyrosine phosphatase hPTP-J|pancreatic carcinoma phosphatase 2|pi R-PTP-Psi|protein-tyrosine phosphatase J|protein-tyrosine phosphatase pi|protein-tyrosine phosphatase receptor omicron|receptor-type protein-tyrosine phosphatase psi', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=29236516, end_position_on_the_genomic_accession=29326813, orientation='plus', exon_count=34, OMIM=602454, no_hdr0=''),
  10077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10077, CurrentID=0, Status='live', Symbol='TSPAN32', Aliases='ART1, PHEMX, PHMX, TSSC6', description='tetraspanin 32', other_designations='pan-hematopoietic expression protein|protein Phemx|tspan-32|tumor-suppressing STF cDNA 6|tumor-suppressing subchromosomal transferable fragment cDNA 6|tumor-suppressing subtransferable candidate 6', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2302013, end_position_on_the_genomic_accession=2318201, orientation='plus', exon_count=14, OMIM=603853, no_hdr0=''),
  10078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10078, CurrentID=0, Status='live', Symbol='TSSC4', Aliases='', description='tumor suppressing subtransferable candidate 4', other_designations='tumor-suppressing STF cDNA 4 protein|tumor-suppressing subchromosomal transferable fragment candidate gene 4 protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2400445, end_position_on_the_genomic_accession=2403878, orientation='plus', exon_count=5, OMIM=603852, no_hdr0=''),
  10079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10079, CurrentID=0, Status='live', Symbol='ATP9A', Aliases='ATPIIA', description='ATPase, class II, type 9A', other_designations='ATPase type IV, phospholipid-transporting (P-type),(putative)|phospholipid-transporting ATPase IIA', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51596775, end_position_on_the_genomic_accession=51768411, orientation='minus', exon_count=29, OMIM=609126, no_hdr0=''),
  403296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403296, CurrentID=0, Status='live', Symbol='OR7E154P', Aliases='', description='olfactory receptor family 7 subfamily E member 154 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 154 pseudogene', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7705162, end_position_on_the_genomic_accession=7706118, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10081, CurrentID=0, Status='live', Symbol='PDCD7', Aliases='ES18, HES18', description='programmed cell death 7', other_designations='U11/U12 snRNP 59K|apoptosis-related protein ES18', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65117379, end_position_on_the_genomic_accession=65133836, orientation='minus', exon_count=5, OMIM=608138, no_hdr0=''),
  10082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10082, CurrentID=0, Status='live', Symbol='GPC6', Aliases='OMIMD1', description='glypican 6', other_designations='glypican proteoglycan 6', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93226825, end_position_on_the_genomic_accession=94408020, orientation='plus', exon_count=10, OMIM=604404, no_hdr0=''),
  10083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10083, CurrentID=0, Status='live', Symbol='USH1C', Aliases='AIE-75, DFNB18, DFNB18A, NY-CO-37, NY-CO-38, PDZ-45, PDZ-73, PDZ-73/NY-CO-38, PDZ73, PDZD7C, ush1cpst', description='Usher syndrome 1C', other_designations='Usher syndrome 1C (autosomal recessive, severe)|antigen NY-CO-38/NY-CO-37|autoimmune enteropathy-related antigen AIE-75|renal carcinoma antigen NY-REN-3|usher syndrome type-1C protein', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17493895, end_position_on_the_genomic_accession=17544416, orientation='minus', exon_count=29, OMIM=605242, no_hdr0=''),
  10084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10084, CurrentID=0, Status='live', Symbol='PQBP1', Aliases='MRX2, MRX55, MRXS3, MRXS8, NPW38, RENS1, SHS', description='polyglutamine binding protein 1', other_designations='38 kDa nuclear protein containing a WW domain|polyglutamine tract-binding protein 1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48897912, end_position_on_the_genomic_accession=48903145, orientation='plus', exon_count=8, OMIM=300463, no_hdr0=''),
  10085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10085, CurrentID=0, Status='live', Symbol='EDIL3', Aliases='DEL1', description='EGF-like repeats and discoidin I-like domains 3', other_designations='developmental endothelial locus-1|developmentally-regulated endothelial cell locus 1 protein|integrin-binding protein DEL1', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83940595, end_position_on_the_genomic_accession=84384867, orientation='minus', exon_count=13, OMIM=606018, no_hdr0=''),
  10086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10086, CurrentID=0, Status='live', Symbol='HHLA1', Aliases='PLA2L', description='HERV-H LTR-associating 1', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132061486, end_position_on_the_genomic_accession=132105265, orientation='minus', exon_count=16, OMIM=604109, no_hdr0=''),
  10087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10087, CurrentID=0, Status='live', Symbol='COL4A3BP', Aliases='CERT, CERTL, GPBP, MRD34, STARD11', description='collagen, type IV, alpha 3 (Goodpasture antigen) binding protein', other_designations='StAR-related lipid transfer (START) domain containing 11|ceramide transfer protein|ceramide transporter|hCERT|lipid-transfer protein CERTL', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75368486, end_position_on_the_genomic_accession=75511981, orientation='minus', exon_count=20, OMIM=604677, no_hdr0=''),
  403304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403304, CurrentID=0, Status='live', Symbol='OR8S21P', Aliases='', description='olfactory receptor family 8 subfamily S member 21 pseudogene', other_designations='olfactory receptor, family 8, subfamily S, member 21 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48417150, end_position_on_the_genomic_accession=48417337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10089, CurrentID=0, Status='live', Symbol='KCNK7', Aliases='K2p7.1, TWIK3', description='potassium channel, two pore domain subfamily K, member 7', other_designations='potassium channel, subfamily K, member 7|two pore domain K+ channel', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65592855, end_position_on_the_genomic_accession=65595996, orientation='minus', exon_count=4, OMIM=603940, no_hdr0=''),
  10090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10090, CurrentID=0, Status='live', Symbol='UST', Aliases='2OST', description='uronyl-2-sulfotransferase', other_designations='dermatan/chondroitin sulfate 2-sulfotransferase', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148746962, end_position_on_the_genomic_accession=149076990, orientation='plus', exon_count=12, OMIM=610752, no_hdr0=''),
  285671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285671, CurrentID=0, Status='live', Symbol='RNF180', Aliases='RINES', description='ring finger protein 180', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64165311, end_position_on_the_genomic_accession=64374375, orientation='plus', exon_count=12, OMIM=616015, no_hdr0=''),
  10092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10092, CurrentID=0, Status='live', Symbol='ARPC5', Aliases='ARC16, dJ127C7.3, p16-Arc', description='actin related protein 2/3 complex subunit 5', other_designations='Arp2/3 protein complex subunit p16|actin related protein 2/3 complex subunit 5, 16kDa|actin related protein 2/3 complex, subunit 5, 16kDa|arp2/3 complex 16 kDa subunit', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183626193, end_position_on_the_genomic_accession=183635941, orientation='minus', exon_count=4, OMIM=604227, no_hdr0=''),
  10093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10093, CurrentID=0, Status='live', Symbol='ARPC4', Aliases='ARC20, P20-ARC', description='actin related protein 2/3 complex subunit 4', other_designations='actin related protein 2/3 complex subunit 4, 20kDa|actin related protein 2/3 complex, subunit 4, 20kDa|arp2/3 complex 20 kDa subunit|arp2/3 protein complex subunit p20', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9792495, end_position_on_the_genomic_accession=9807105, orientation='plus', exon_count=7, OMIM=604226, no_hdr0=''),
  10094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10094, CurrentID=0, Status='live', Symbol='ARPC3', Aliases='ARC21, p21-Arc', description='actin related protein 2/3 complex subunit 3', other_designations='ARP2/3 protein complex subunit p21|actin related protein 2/3 complex subunit 3, 21kDa|actin related protein 2/3 complex, subunit 3, 21kDa|arp2/3 complex 21 kDa subunit', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110434890, end_position_on_the_genomic_accession=110450411, orientation='minus', exon_count=7, OMIM=604225, no_hdr0=''),
  10095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10095, CurrentID=0, Status='live', Symbol='ARPC1B', Aliases='ARC41, p40-ARC, p41-ARC', description='actin related protein 2/3 complex subunit 1B', other_designations='ARP2/3 protein complex subunit p41|actin related protein 2/3 complex subunit 1B, 41kDa|actin related protein 2/3 complex, subunit 1B, 41kDa|arp2/3 complex 41 kDa subunit', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99374675, end_position_on_the_genomic_accession=99394801, orientation='plus', exon_count=11, OMIM=604223, no_hdr0=''),
  10096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10096, CurrentID=0, Status='live', Symbol='ACTR3', Aliases='ARP3', description='ARP3 actin-related protein 3 homolog (yeast)', other_designations='actin-like protein 3', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113889934, end_position_on_the_genomic_accession=113961554, orientation='plus', exon_count=13, OMIM=604222, no_hdr0=''),
  10097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10097, CurrentID=0, Status='live', Symbol='ACTR2', Aliases='ARP2', description='ARP2 actin-related protein 2 homolog (yeast)', other_designations='actin-like protein 2', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65227695, end_position_on_the_genomic_accession=65271253, orientation='plus', exon_count=10, OMIM=604221, no_hdr0=''),
  10098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10098, CurrentID=0, Status='live', Symbol='TSPAN5', Aliases='NET-4, NET4, TM4SF9, TSPAN-5', description='tetraspanin 5', other_designations='tetraspan 5|tetraspan NET-4|tetraspan TM4SF|transmembrane 4 superfamily member 9|transmembrane 4 superfamily, member 8', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98470367, end_position_on_the_genomic_accession=98658661, orientation='minus', exon_count=8, OMIM=613136, no_hdr0=''),
  10099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10099, CurrentID=0, Status='live', Symbol='TSPAN3', Aliases='TM4-A, TM4SF8, TSPAN-3', description='tetraspanin 3', other_designations='tetraspan 3|tetraspan TM4SF|tetraspanin TM4-A|transmembrane 4 superfamily member 8', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77044017, end_position_on_the_genomic_accession=77071228, orientation='minus', exon_count=7, OMIM=613134, no_hdr0=''),
  10100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10100, CurrentID=0, Status='live', Symbol='TSPAN2', Aliases='NET3, TSN2, TSPAN-2', description='tetraspanin 2', other_designations='new EST tetraspan 3|tetraspan 2|tetraspan NET-3|tetraspan TM4SF|tetraspanin 2 isoform', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115048011, end_position_on_the_genomic_accession=115089500, orientation='minus', exon_count=8, OMIM=613133, no_hdr0=''),
  10101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10101, CurrentID=0, Status='live', Symbol='NUBP2', Aliases='CFD1, NBP 2, NUBP1', description='nucleotide binding protein 2', other_designations='homolog of yeast cytosolic Fe-S cluster deficient 1|nucleotide binding protein 2 (MinD homolog, E. coli)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1782923, end_position_on_the_genomic_accession=1789191, orientation='plus', exon_count=10, OMIM=610779, no_hdr0=''),
  10102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10102, CurrentID=0, Status='live', Symbol='TSFM', Aliases='EFTS, EFTSMT', description='Ts translation elongation factor, mitochondrial', other_designations='EF-Ts|EF-TsMt|mitochondrial elongation factor Ts', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57782745, end_position_on_the_genomic_accession=57802856, orientation='plus', exon_count=8, OMIM=604723, no_hdr0=''),
  10103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10103, CurrentID=0, Status='live', Symbol='TSPAN1', Aliases='NET1, TM4C, TM4SF', description='tetraspanin 1', other_designations='tetraspan 1', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46175077, end_position_on_the_genomic_accession=46196489, orientation='plus', exon_count=11, OMIM=613170, no_hdr0=''),
  101928829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928829, CurrentID=0, Status='live', Symbol='LOC101928829', Aliases='', description='uncharacterized LOC101928829', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31508436, end_position_on_the_genomic_accession=31509569, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10105, CurrentID=0, Status='live', Symbol='PPIF', Aliases='CYP3, CyP-M, Cyp-D, CypD', description='peptidylprolyl isomerase F', other_designations='PPIase F|cyclophilin 3|cyclophilin D|cyclophilin F|mitochondrial cyclophilin|peptidyl-prolyl cis-trans isomerase, mitochondrial|rotamase F', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79347464, end_position_on_the_genomic_accession=79355334, orientation='plus', exon_count=6, OMIM=604486, no_hdr0=''),
  10106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10106, CurrentID=0, Status='live', Symbol='CTDSP2', Aliases='OS4, PSR2, SCP2', description='CTD small phosphatase 2', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 2|NLI-interacting factor 2|conserved gene amplified in osteosarcoma|nuclear LIM interactor-interacting factor 2|protein OS-4|small C-terminal domain phosphatase 2|small CTD phosphatase 2', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57819927, end_position_on_the_genomic_accession=57846964, orientation='minus', exon_count=8, OMIM=608711, no_hdr0=''),
  10107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10107, CurrentID=0, Status='live', Symbol='TRIM10', Aliases='HERF1, RFB30, RNF9', description='tripartite motif containing 10', other_designations='B30-RING finger protein|Zn-finger protein|hematopoietic RING finger 1|ring finger protein 9|tripartite motif protein 10|tripartite motif-containing 10', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30151939, end_position_on_the_genomic_accession=30163440, orientation='minus', exon_count=12, OMIM=605701, no_hdr0=''),
  10108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10108, CurrentID=0, Status='live', Symbol='MKRN3-AS1', Aliases='FNZ127, MKRN3-AS, MKRN3AS, NCRNA00009, ZNF127-AS, ZNF127AS', description='MKRN3 antisense RNA 1', other_designations='MKRN3 antisense RNA 1 (non-protein coding)|makorin, ring finger protein, 3 antisense|non-protein coding RNA 9|zinc finger protein 127 antisense', map_location='15q11-q13', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603857, no_hdr0=''),
  10109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10109, CurrentID=0, Status='live', Symbol='ARPC2', Aliases='ARC34, PNAS-139, PRO2446, p34-Arc', description='actin related protein 2/3 complex subunit 2', other_designations='ARP2/3 protein complex subunit 34|actin related protein 2/3 complex subunit 2, 34kDa|actin related protein 2/3 complex, subunit 2, 34kDa|arp2/3 complex 34 kDa subunit', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218217094, end_position_on_the_genomic_accession=218254348, orientation='plus', exon_count=12, OMIM=604224, no_hdr0=''),
  10110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10110, CurrentID=0, Status='live', Symbol='SGK2', Aliases='H-SGK2, dJ138B7.2', description='serum/glucocorticoid regulated kinase 2', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43558995, end_position_on_the_genomic_accession=43585633, orientation='plus', exon_count=14, OMIM=607589, no_hdr0=''),
  10111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10111, CurrentID=0, Status='live', Symbol='RAD50', Aliases='NBSLD2, hRad50, RAD50', description='RAD50 homolog, double strand break repair protein', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132556924, end_position_on_the_genomic_accession=132644621, orientation='plus', exon_count=25, OMIM=604040, no_hdr0=''),
  10112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10112, CurrentID=0, Status='live', Symbol='KIF20A', Aliases='MKLP2, RAB6KIFL', description='kinesin family member 20A', other_designations='GG10_2|RAB6 interacting, kinesin-like (rabkinesin6)|mitotic kinesin-like protein 2|rab6-interacting kinesin-like protein|rabkinesin-6', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138178728, end_position_on_the_genomic_accession=138187715, orientation='plus', exon_count=19, OMIM=605664, no_hdr0=''),
  10113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10113, CurrentID=0, Status='live', Symbol='PREB', Aliases='SEC12', description='prolactin regulatory element binding', other_designations='mammalian guanine nucleotide exchange factor mSec12|prolactin regulatory binding-element protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27130756, end_position_on_the_genomic_accession=27134674, orientation='minus', exon_count=9, OMIM=606395, no_hdr0=''),
  10114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10114, CurrentID=0, Status='live', Symbol='HIPK3', Aliases='DYRK6, FIST3, PKY, YAK1', description='homeodomain interacting protein kinase 3', other_designations='ANPK|FIST|androgen receptor-interacting nuclear protein kinase|fas-interacting serine/threonine-protein kinase|homolog of protein kinase YAK1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33256672, end_position_on_the_genomic_accession=33357023, orientation='plus', exon_count=19, OMIM=604424, no_hdr0=''),
  10116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10116, CurrentID=0, Status='live', Symbol='FEM1B', Aliases='F1A-ALPHA, F1AA, FEM1-beta', description='fem-1 homolog b (C. elegans)', other_designations='FEM-1-like death receptor binding protein|fem-1-like death receptor-binding protein alpha|fem-1-like in apoptotic pathway protein alpha', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68277756, end_position_on_the_genomic_accession=68295865, orientation='plus', exon_count=2, OMIM=613539, no_hdr0=''),
  10117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10117, CurrentID=0, Status='live', Symbol='ENAM', Aliases='ADAI, AI1C, AIH2', description='enamelin', other_designations='amelogenesis imperfecta 2, hypocalcification (autosomal dominant)', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70628744, end_position_on_the_genomic_accession=70646819, orientation='plus', exon_count=9, OMIM=606585, no_hdr0=''),
  10118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10118, CurrentID=0, Status='live', Symbol='EMWX', Aliases='', description='episodic muscle weakness, X-linked', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300211, no_hdr0=''),
  10120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10120, CurrentID=0, Status='live', Symbol='ACTR1B', Aliases='ARP1B, CTRN2, PC3', description='ARP1 actin-related protein 1 homolog B, centractin beta (yeast)', other_designations='actin-related protein 1B|centractin beta', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97655939, end_position_on_the_genomic_accession=97664098, orientation='minus', exon_count=11, OMIM=605144, no_hdr0=''),
  10121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10121, CurrentID=0, Status='live', Symbol='ACTR1A', Aliases='ARP1, Arp1A, CTRN1', description='ARP1 actin-related protein 1 homolog A, centractin alpha (yeast)', other_designations='actin-RPV|centractin|centrosome-associated actin homolog', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102479229, end_position_on_the_genomic_accession=102502755, orientation='minus', exon_count=11, OMIM=605143, no_hdr0=''),
  10123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10123, CurrentID=0, Status='live', Symbol='ARL4C', Aliases='ARL7, LAK', description='ADP ribosylation factor like GTPase 4C', other_designations='ADP ribosylation factor-like protein 7|ADP-ribosylation factor like GTPase 4C|ADP-ribosylation factor-like 4C|ADP-ribosylation factor-like 7|ADP-ribosylation factor-like protein 7|ADP-ribosylation factor-like protein LAK', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234493041, end_position_on_the_genomic_accession=234497053, orientation='minus', exon_count=1, OMIM=604787, no_hdr0=''),
  10124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10124, CurrentID=0, Status='live', Symbol='ARL4A', Aliases='ARL4', description='ADP ribosylation factor like GTPase 4A', other_designations='ADP-ribosylation factor like GTPase 4A|ADP-ribosylation factor-like 4|ADP-ribosylation factor-like 4A', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12686827, end_position_on_the_genomic_accession=12690934, orientation='plus', exon_count=3, OMIM=604786, no_hdr0=''),
  10125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10125, CurrentID=0, Status='live', Symbol='RASGRP1', Aliases='CALDAG-GEFI, CALDAG-GEFII, RASGRP, V, hRasGRP1', description='RAS guanyl releasing protein 1 (calcium and DAG-regulated)', other_designations='RAS guanyl nucleotide-releasing protein 1|calcium and DAG-regulated guanine nucleotide exchange factor II|guanine nucleotide exchange factor, calcium- and DAG-regulated, Rap1A|ras activator RasGRP', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38488101, end_position_on_the_genomic_accession=38564806, orientation='minus', exon_count=17, OMIM=603962, no_hdr0=''),
  10126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10126, CurrentID=0, Status='live', Symbol='DNAL4', Aliases='MRMV3, PIG27', description='dynein, axonemal, light chain 4', other_designations='dynein light chain, outer arm 4|dynein, axonemal, light polypeptide 4|proliferation-inducing gene 27|proliferation-inducing protein 27', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38778508, end_position_on_the_genomic_accession=38794156, orientation='minus', exon_count=4, OMIM=610565, no_hdr0=''),
  10127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10127, CurrentID=0, Status='live', Symbol='ZNF263', Aliases='FPM315, ZKSCAN12, ZSCAN44', description='zinc finger protein 263', other_designations='zinc finger protein FPM315|zinc finger protein with KRAB and SCAN domains 12', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3283380, end_position_on_the_genomic_accession=3301401, orientation='plus', exon_count=8, OMIM=604191, no_hdr0=''),
  10128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10128, CurrentID=0, Status='live', Symbol='LRPPRC', Aliases='CLONE-23970, GP130, LRP130, LSFC', description='leucine rich pentatricopeptide repeat containing', other_designations='130 kDa leucine-rich protein|LRP 130|leucine-rich PPR-motif containing|leucine-rich pentatricopeptide repeat containing|mitochondrial leucine-rich PPR motif-containing protein', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43886224, end_position_on_the_genomic_accession=43996005, orientation='minus', exon_count=39, OMIM=607544, no_hdr0=''),
  10129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10129, CurrentID=0, Status='live', Symbol='FRY', Aliases='13CDNA73, 214K23.2, C13orf14, CG003, bA207N4.2, bA37E23.1', description='FRY microtubule binding protein', other_designations='WUGSC:H_2G3A.1|furry homolog', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32031300, end_position_on_the_genomic_accession=32297498, orientation='plus', exon_count=64, OMIM=614818, no_hdr0=''),
  10130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10130, CurrentID=0, Status='live', Symbol='PDIA6', Aliases='ERP5, P5, TXNDC7', description='protein disulfide isomerase family A member 6', other_designations='ER protein 5|endoplasmic reticulum protein 5|protein disulfide isomerase P5|protein disulfide isomerase family A, member 6|protein disulfide isomerase-associated 6|protein disulfide isomerase-related protein|thioredoxin domain containing 7 (protein disulfide isomerase)|thioredoxin domain-containing protein 7', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10783391, end_position_on_the_genomic_accession=10837977, orientation='minus', exon_count=18, OMIM=611099, no_hdr0=''),
  10131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10131, CurrentID=0, Status='live', Symbol='TRAP1', Aliases='HSP 75, HSP75, HSP90L, TRAP-1', description='TNF receptor-associated protein 1', other_designations='TNFR-associated protein 1|tumor necrosis factor type 1 receptor-associated protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3658037, end_position_on_the_genomic_accession=3717597, orientation='minus', exon_count=19, OMIM=606219, no_hdr0=''),
  10133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10133, CurrentID=0, Status='live', Symbol='OPTN', Aliases='ALS12, FIP2, GLC1E, HIP7, HYPL, NRP, TFIIIA-INTP', description='optineurin', other_designations='E3-14.7K-interacting protein|FIP-2|HIP-7|Huntingtin interacting protein L|huntingtin yeast partner L|huntingtin-interacting protein 7|huntingtin-interacting protein L|nemo-related protein|optic neuropathy-inducing protein|transcription factor IIIA-interacting protein|transcrption factor IIIA-interacting protein|tumor necrosis factor alpha-inducible cellular protein containing leucine zipper domains', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13100075, end_position_on_the_genomic_accession=13138291, orientation='plus', exon_count=16, OMIM=602432, no_hdr0=''),
  10134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10134, CurrentID=0, Status='live', Symbol='BCAP31', Aliases='6C6-AG, BAP31, CDM, DDCH, DXS1357E', description='B-cell receptor-associated protein 31', other_designations='6C6-AG tumor-associated antigen|BCR-associated protein Bap31|p28 Bap31', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153700492, end_position_on_the_genomic_accession=153724746, orientation='minus', exon_count=9, OMIM=300398, no_hdr0=''),
  10135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10135, CurrentID=0, Status='live', Symbol='NAMPT', Aliases='1110035O14Rik, PBEF, PBEF1, VF, VISFATIN', description='nicotinamide phosphoribosyltransferase', other_designations='NAmPRTase|pre-B cell-enhancing factor|pre-B-cell colony enhancing factor 1|pre-B-cell colony-enhancing factor 1', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106248285, end_position_on_the_genomic_accession=106285732, orientation='minus', exon_count=13, OMIM=608764, no_hdr0=''),
  10136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10136, CurrentID=0, Status='live', Symbol='CELA3A', Aliases='ELA3, ELA3A', description='chymotrypsin like elastase family member 3A', other_designations='chymotrypsin-like elastase family, member 3A|elastase 1|elastase 3A, pancreatic|elastase IIIA|elastase-3A|protease E', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22001656, end_position_on_the_genomic_accession=22012542, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  10137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10137, CurrentID=0, Status='live', Symbol='RBM12', Aliases='HRIHFB2091, SWAN', description='RNA binding motif protein 12', other_designations='RNA-binding motif protein 12|SH3/WW domain anchor protein in the nucleus', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35648925, end_position_on_the_genomic_accession=35664956, orientation='minus', exon_count=3, OMIM=607179, no_hdr0=''),
  10138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10138, CurrentID=0, Status='live', Symbol='YAF2', Aliases='', description='YY1 associated factor 2', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42157104, end_position_on_the_genomic_accession=42238352, orientation='minus', exon_count=11, OMIM=607534, no_hdr0=''),
  10139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10139, CurrentID=0, Status='live', Symbol='ARFRP1', Aliases='ARL18, ARP, Arp1', description='ADP ribosylation factor related protein 1', other_designations='ADP-ribosylation factor related protein 1|ARF-related protein 1|SCG10 like-protein|helicase-like protein NHL', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63698642, end_position_on_the_genomic_accession=63708013, orientation='minus', exon_count=9, OMIM=604699, no_hdr0=''),
  10140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10140, CurrentID=0, Status='live', Symbol='TOB1', Aliases='APRO6, PIG49, TOB, TROB, TROB1', description='transducer of ERBB2, 1', other_designations='proliferation-inducing gene 49|transducer of erbB-2 1', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50862223, end_position_on_the_genomic_accession=50867978, orientation='minus', exon_count=4, OMIM=605523, no_hdr0=''),
  10141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10141, CurrentID=0, Status='live', Symbol='LINC01587', Aliases='C4orf6, aC1', description='long intergenic non-protein coding RNA 1587', other_designations='expressed in neuroblastoma', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5524569, end_position_on_the_genomic_accession=5527801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10142, CurrentID=0, Status='live', Symbol='AKAP9', Aliases='AKAP-9, AKAP350, AKAP450, CG-NAP, HYPERION, LQT11, MU-RMS-40.16A, PPP1R45, PRKA9, YOTIAO', description='A-kinase anchoring protein 9', other_designations='A kinase (PRKA) anchor protein (yotiao) 9|A kinase (PRKA) anchor protein 9|A-kinase anchor protein 350 kDa|A-kinase anchor protein 450 kDa|AKAP 120-like protein|AKAP9-BRAF fusion protein|centrosome- and Golgi-localized PKN-associated protein|centrosome- and golgi-localized protein kinase N-associated protein|kinase N-associated protein|protein hyperion|protein kinase A anchoring protein 9|protein phosphatase 1, regulatory subunit 45|protein yotiao', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91940867, end_position_on_the_genomic_accession=92110673, orientation='plus', exon_count=54, OMIM=604001, no_hdr0=''),
  10143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10143, CurrentID=0, Status='live', Symbol='CLEC3A', Aliases='CLECSF1', description='C-type lectin domain family 3 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 1 (cartilage-derived)|C-type lectin domain family 3, member A', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78022546, end_position_on_the_genomic_accession=78032106, orientation='plus', exon_count=4, OMIM=613588, no_hdr0=''),
  10144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10144, CurrentID=0, Status='live', Symbol='FAM13A', Aliases='ARHGAP481, FAM13A', description='family with sequence similarity 13 member A', other_designations='FAM13A1_v2 protein|family with sequence similarity 13, member A|family with sequence similarity 13, member A1', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88725954, end_position_on_the_genomic_accession=89111398, orientation='minus', exon_count=31, OMIM=613299, no_hdr0=''),
  10146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10146, CurrentID=0, Status='live', Symbol='G3BP1', Aliases='G3BP, HDH-VIII', description='GTPase activating protein (SH3 domain) binding protein 1', other_designations='ATP-dependent DNA helicase VIII|G3BP-1|GAP SH3 domain-binding protein 1|GAP binding protein|Ras-GTPase-activating protein SH3-domain-binding protein|RasGAP-associated endoribonuclease G3BP', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151771915, end_position_on_the_genomic_accession=151811356, orientation='plus', exon_count=12, OMIM=608431, no_hdr0=''),
  10147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10147, CurrentID=0, Status='live', Symbol='SUGP2', Aliases='SFRS14', description='SURP and G-patch domain containing 2', other_designations='SURP and G patch domain containing 2|arginine/serine-rich 14 splicing factor|arginine/serine-rich-splicing factor 14|putative splicing factor, arginine/serine-rich 14|splicing factor, arginine/serine-rich 14', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18990888, end_position_on_the_genomic_accession=19033571, orientation='minus', exon_count=14, OMIM=607993, no_hdr0=''),
  10148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10148, CurrentID=0, Status='live', Symbol='EBI3', Aliases='IL-27B, IL27B', description='Epstein-Barr virus induced 3', other_designations='EBV-induced gene 3 protein|Epstein-Barr virus induced gene 3|IL-27 subunit beta|IL27 subunit|IL35 subunit|cytokine receptor|epstein-Barr virus-induced gene 3 protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4229302, end_position_on_the_genomic_accession=4237542, orientation='plus', exon_count=6, OMIM=605816, no_hdr0=''),
  10149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10149, CurrentID=0, Status='live', Symbol='ADGRG2', Aliases='EDDM6, GPR64, HE6, TM7LN2', description='adhesion G protein-coupled receptor G2', other_designations='G protein-coupled receptor 64|G protein-coupled receptor, epididymis-specific (seven transmembrane family)|epididymal protein 6|human epididymis-specific protein 6', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18989307, end_position_on_the_genomic_accession=19122931, orientation='minus', exon_count=32, OMIM=300572, no_hdr0=''),
  10150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10150, CurrentID=0, Status='live', Symbol='MBNL2', Aliases='MBLL, MBLL39, PRO2032', description='muscleblind-like splicing regulator 2', other_designations='muscleblind-like 2|muscleblind-like protein 1|muscleblind-like protein-like 39', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97141820, end_position_on_the_genomic_accession=97394120, orientation='plus', exon_count=11, OMIM=607327, no_hdr0=''),
  10151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10151, CurrentID=0, Status='live', Symbol='HNRNPA3P1', Aliases='D10S102, FBRNP, HNRPA3, HNRPA3P1', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 1', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43787412, end_position_on_the_genomic_accession=43790417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10152, CurrentID=0, Status='live', Symbol='ABI2', Aliases='ABI-2B, AIP-1, AblBP3, SSH3BP2, argBP1, argBPIA, argBPIB, ABI2', description='abl-interactor 2', other_designations='abelson interactor 2|abl binding protein 3|abl-interacting protein 1 (SH3-containing protein)|abl-interactor protein 2b|arg protein tyrosine kinase-binding protein|arg-binding protein 1', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203327874, end_position_on_the_genomic_accession=203432173, orientation='plus', exon_count=15, OMIM=606442, no_hdr0=''),
  10153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10153, CurrentID=0, Status='live', Symbol='CEBPZ', Aliases='CBF, CBF2, HSP-CBF, NOC1', description='CCAAT/enhancer binding protein (C/EBP), zeta', other_designations='CCAAT-binding factor|CCAAT-box-binding transcription factor|CCAAT/enhancer binding protein zeta', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37201629, end_position_on_the_genomic_accession=37231597, orientation='minus', exon_count=16, OMIM=612828, no_hdr0=''),
  10154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10154, CurrentID=0, Status='live', Symbol='PLXNC1', Aliases='CD232, PLXN-C1, VESPR', description='plexin C1', other_designations='plexin (semaphorin receptor)|receptor for viral semaphorin protein|receptor for virally-encoded semaphorin|virus-encoded semaphorin protein receptor', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94148723, end_position_on_the_genomic_accession=94307675, orientation='plus', exon_count=34, OMIM=604259, no_hdr0=''),
  10155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10155, CurrentID=0, Status='live', Symbol='TRIM28', Aliases='KAP1, PPP1R157, RNF96, TF1B, TIF1B', description='tripartite motif containing 28', other_designations='E3 SUMO-protein ligase TRIM28|KAP-1|KRAB [Kruppel-associated box domain]-associated protein 1|KRAB-associated protein 1|KRAB-interacting protein 1|KRIP-1|RING finger protein 96|TIF1-beta|nuclear corepressor KAP-1|protein phosphatase 1, regulatory subunit 157|transcriptional intermediary factor 1-beta|tripartite motif-containing 28|tripartite motif-containing protein 28', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58544469, end_position_on_the_genomic_accession=58550715, orientation='plus', exon_count=17, OMIM=601742, no_hdr0=''),
  10156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10156, CurrentID=0, Status='live', Symbol='RASA4', Aliases='CAPRI, GAPL', description='RAS p21 protein activator 4', other_designations='Ca2+-promoted Ras inactivator|calcium-promoted Ras inactivator|rasGAP-activating-like protein 2', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102579646, end_position_on_the_genomic_accession=102616758, orientation='minus', exon_count=22, OMIM=607943, no_hdr0=''),
  10157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10157, CurrentID=0, Status='live', Symbol='AASS', Aliases='LKR/SDH, LKRSDH, LORSDH', description='aminoadipate-semialdehyde synthase', other_designations='alpha-aminoadipate semialdehyde synthase|aminoadipic semialdehyde synthase|lysine-2-oxoglutarate reductase|lysine-ketoglutarate reductase /saccharopine dehydrogenase', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122073544, end_position_on_the_genomic_accession=122144290, orientation='minus', exon_count=26, OMIM=605113, no_hdr0=''),
  10158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10158, CurrentID=0, Status='live', Symbol='PDZK1IP1', Aliases='DD96, MAP17, SPAP', description='PDZK1 interacting protein 1', other_designations='17 kDa membrane-associated protein|epithelial protein up-regulated in carcinoma, membrane associated protein 17|membrane-associated protein 17', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47183589, end_position_on_the_genomic_accession=47190099, orientation='minus', exon_count=4, OMIM=607178, no_hdr0=''),
  10159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10159, CurrentID=0, Status='live', Symbol='ATP6AP2', Aliases='APT6M8-9, ATP6IP2, ATP6M8-9, ELDF10, HT028, M8-9, MRXE, MRXSH, MSTP009, PRR, RENR, XMRE, XPDS', description='ATPase, H+ transporting, lysosomal accessory protein 2', other_designations='ATPase H(+)-transporting lysosomal-interacting protein 2|ATPase, H+ transporting, lysosomal (vacuolar proton pump) membrane sector associated protein M8-9|ATPase, H+ transporting, lysosomal interacting protein 2|ER-localized type I transmembrane adaptor|N14F|V-ATPase M8.9 subunit|embryonic liver differentiation factor 10|prorenin receptor|renin/prorenin receptor|vacuolar ATP synthase membrane sector-associated protein M8-9|vacuolar proton ATP synthase membrane sector associated protein M8-9', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40580964, end_position_on_the_genomic_accession=40606637, orientation='plus', exon_count=9, OMIM=300556, no_hdr0=''),
  10160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10160, CurrentID=0, Status='live', Symbol='FARP1', Aliases='CDEP-IT1, PLEKHC2, PPP1R75, FARP1', description='FERM, ARH/RhoGEF and pleckstrin domain protein 1', other_designations='FERM, RhoGEF (ARHGEF) and pleckstrin domain protein 1 (chondrocyte-derived)|PH domain-containing family C member 2|chondrocyte-derived ezrin-like protein|pleckstrin homology domain-containing family C member 2|protein phosphatase 1, regulatory subunit 75', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98142562, end_position_on_the_genomic_accession=98449773, orientation='plus', exon_count=30, OMIM=602654, no_hdr0=''),
  10161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10161, CurrentID=0, Status='live', Symbol='LPAR6', Aliases='ARWH1, HYPT8, LAH3, P2RY5, P2Y5', description='lysophosphatidic acid receptor 6', other_designations='G-protein coupled purinergic receptor P2Y5|LPA receptor 6|LPA-6|P2Y purinoceptor 5|RB intron encoded G-protein coupled receptor|oleoyl-L-alpha-lysophosphatidic acid receptor|purinergic receptor 5|purinergic receptor P2Y G protein-coupled protein 5|purinergic receptor P2Y, G-protein coupled, 5', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48411045, end_position_on_the_genomic_accession=48444704, orientation='minus', exon_count=8, OMIM=609239, no_hdr0=''),
  10162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10162, CurrentID=0, Status='live', Symbol='LPCAT3', Aliases='C3F, LPCAT, LPLAT 5, LPSAT, MBOAT5, OACT5, nessy', description='lysophosphatidylcholine acyltransferase 3', other_designations='1-acylglycerophosphocholine O-acyltransferase|1-acylglycerophosphoserine O-acyltransferase|O-acyltransferase (membrane bound) domain containing 5|O-acyltransferase domain-containing protein 5|lyso-PC acyltransferase 3|lyso-PS acyltransferase|lysophosphatidylserine acyltransferase|membrane bound O-acyltransferase domain containing 5|membrane-bound O-acyltransferase domain-containing protein 5|putative protein similar to nessy', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6976185, end_position_on_the_genomic_accession=7018538, orientation='minus', exon_count=13, OMIM=611950, no_hdr0=''),
  10163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10163, CurrentID=0, Status='live', Symbol='WASF2', Aliases='IMD2, SCAR2, WASF4, WAVE2, dJ393P12.2', description='WAS protein family member 2', other_designations='WAS protein family, member 2|WASP family Verprolin-homologous protein 2|WASP family protein member 2|WASP family protein member 4|protein WAVE-2|putative Wiskott-Aldrich syndrome protein family member 4|suppressor of cyclic-AMP receptor (WASP-family)|verprolin homology domain-containing protein 2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27404226, end_position_on_the_genomic_accession=27490187, orientation='minus', exon_count=9, OMIM=605875, no_hdr0=''),
  10164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10164, CurrentID=0, Status='live', Symbol='CHST4', Aliases='GST3, GlcNAc6ST2, HECGLCNAC6ST, LSST', description='carbohydrate (N-acetylglucosamine 6-O) sulfotransferase 4', other_designations='GST-3|HEC-GlcNAc6ST|L-selectin ligand sulfotransferase|N-acetylglucosamine 6-O-sulfotransferase 2|galactose/N-acetylglucosamine/N-acetylgalactosamine 6-O-sulfotransferase 3|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 3|glcNAc6ST-2|gn6st-2|high endothelial cells N-acetylglucosamine 6-O-sulfotransferase', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71526120, end_position_on_the_genomic_accession=71538590, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10165, CurrentID=0, Status='live', Symbol='SLC25A13', Aliases='ARALAR2, CITRIN, CTLN2', description='solute carrier family 25 (aspartate/glutamate carrier), member 13', other_designations='mitochondrial aspartate glutamate carrier 2|solute carrier family 25, member 13 (citrin)', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96120220, end_position_on_the_genomic_accession=96322147, orientation='minus', exon_count=20, OMIM=603859, no_hdr0=''),
  10166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10166, CurrentID=0, Status='live', Symbol='SLC25A15', Aliases='D13S327, HHH, ORC1, ORNT1', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15', other_designations='ornithine transporter 1|solute carrier family 25 member 15', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40789411, end_position_on_the_genomic_accession=40812460, orientation='plus', exon_count=7, OMIM=603861, no_hdr0=''),
  10167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10167, CurrentID=0, Status='live', Symbol='CFM1', Aliases='', description='cystic fibrosis modifier 1', other_designations='', map_location='19q13.2-q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603855, no_hdr0=''),
  10168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10168, CurrentID=0, Status='live', Symbol='ZNF197', Aliases='D3S1363E, P18, VHLaK, ZKSCAN9, ZNF166, ZNF20, ZSCAN41', description='zinc finger protein 197', other_designations='VHL-associated KRAB-A domain-containing protein|pVHL-associated KRAB domain-containing protein|zinc finger protein 166|zinc finger protein 20|zinc finger protein with KRAB and SCAN domains 9', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44625019, end_position_on_the_genomic_accession=44648471, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  10169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10169, CurrentID=0, Status='live', Symbol='SERF2', Aliases='4F5REL, FAM2C, H4F5REL, HsT17089', description='small EDRK-rich factor 2', other_designations='gastric cancer-related protein VRG107|protein 4F5-related', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43777096, end_position_on_the_genomic_accession=43796089, orientation='plus', exon_count=5, OMIM=605054, no_hdr0=''),
  10170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10170, CurrentID=0, Status='live', Symbol='DHRS9', Aliases='3-alpha-HSD, 3ALPHA-HSD, RDH-E2, RDH-TBE, RDH15, RDHL, RDHTBE, RETSDR8, SDR9C4', description='dehydrogenase/reductase (SDR family) member 9', other_designations='3-alpha hydroxysteroid dehydrogenase|NADP-dependent retinol dehydrogenase/reductase|retinol dehydrogenase 15|retinol dehydrogenase homolog|short chain dehydrogenase/reductase family 9C member 4|short chain dehydrogenase/reductase family 9C, member 4|short-chain dehydrogenase/reductase retSDR8|tracheobronchial epithelial cell-specific retinol dehydrogenase', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169067035, end_position_on_the_genomic_accession=169096167, orientation='plus', exon_count=9, OMIM=612131, no_hdr0=''),
  10171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10171, CurrentID=0, Status='live', Symbol='RCL1', Aliases='RNAC, RPCL1', description='RNA terminal phosphate cyclase-like 1', other_designations='RNA cyclase homolog', map_location='9p24.1-p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4792834, end_position_on_the_genomic_accession=4861077, orientation='plus', exon_count=10, OMIM=611405, no_hdr0=''),
  10172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10172, CurrentID=0, Status='live', Symbol='ZNF256', Aliases='BMZF-3, BMZF3', description='zinc finger protein 256', other_designations='bone marrow zinc finger 3', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57940833, end_position_on_the_genomic_accession=57947747, orientation='minus', exon_count=3, OMIM=606956, no_hdr0=''),
  10174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10174, CurrentID=0, Status='live', Symbol='SORBS3', Aliases='SCAM-1, SCAM1, SH3D4', description='sorbin and SH3 domain containing 3', other_designations='sorbin and SH3 domain-containing protein 3|vinexin beta (SH3-containing adaptor molecule-1)', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22544990, end_position_on_the_genomic_accession=22575495, orientation='plus', exon_count=24, OMIM=610795, no_hdr0=''),
  10175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10175, CurrentID=0, Status='live', Symbol='CNIH1', Aliases='CNIH, CNIH-1, CNIL, TGAM77', description='cornichon family AMPA receptor auxiliary protein 1', other_designations='T-cell growth-associated molecule 77|cornichon homolog', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54426929, end_position_on_the_genomic_accession=54441430, orientation='minus', exon_count=5, OMIM=611287, no_hdr0=''),
  100533920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533920, CurrentID=0, Status='live', Symbol='LOC100533920', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10178, CurrentID=0, Status='live', Symbol='TENM1', Aliases='ODZ1, ODZ3, TEN-M1, TNM, TNM1', description='teneurin transmembrane protein 1', other_designations='odz (odd Oz/ten-m, Drosophila) homolog 3|odz, odd Oz/ten-m homolog 1|odz, odd Oz/ten-m homolog 1(Drosophila)|protein Odd Oz/ten-m homolog 1|ten-1|tenascin M1|tenascin-M1|teneurin 1', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124375903, end_position_on_the_genomic_accession=125204358, orientation='minus', exon_count=40, OMIM=300588, no_hdr0=''),
  10179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10179, CurrentID=0, Status='live', Symbol='RBM7', Aliases='', description='RNA binding motif protein 7', other_designations='', map_location='11q23.1-q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114400529, end_position_on_the_genomic_accession=114410610, orientation='plus', exon_count=5, OMIM=612413, no_hdr0=''),
  10180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10180, CurrentID=0, Status='live', Symbol='RBM6', Aliases='3G2, DEF-3, DEF3, HLC-11, NY-LU-12, g16', description='RNA binding motif protein 6', other_designations='RNA-binding motif protein 6|RNA-binding protein DEF-3|lung cancer antigen NY-LU-12|lung cancer protooncogene 11', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49940038, end_position_on_the_genomic_accession=50077252, orientation='plus', exon_count=23, OMIM=606886, no_hdr0=''),
  10181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10181, CurrentID=0, Status='live', Symbol='RBM5', Aliases='G15, H37, LUCA15, RMB5', description='RNA binding motif protein 5', other_designations='putative tumor suppressor LUCA15|renal carcinoma antigen NY-REN-9', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50088908, end_position_on_the_genomic_accession=50118964, orientation='plus', exon_count=25, OMIM=606884, no_hdr0=''),
  100533921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533921, CurrentID=0, Status='live', Symbol='LOC100533921', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10184, CurrentID=0, Status='live', Symbol='LHFPL2', Aliases='', description='lipoma HMGIC fusion partner-like 2', other_designations='LHFP-like protein 2', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78485215, end_position_on_the_genomic_accession=78648895, orientation='minus', exon_count=8, OMIM=609718, no_hdr0=''),
  10186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10186, CurrentID=0, Status='live', Symbol='LHFP', Aliases='', description='lipoma HMGIC fusion partner', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39342892, end_position_on_the_genomic_accession=39603219, orientation='minus', exon_count=4, OMIM=606710, no_hdr0=''),
  10187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10187, CurrentID=0, Status='live', Symbol='VDAC1P5', Aliases='VDAC3, VDAC5P', description='voltage-dependent anion channel 1 pseudogene 5', other_designations='voltage-dependent anion channel 5, pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54801885, end_position_on_the_genomic_accession=54803623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10188, CurrentID=0, Status='live', Symbol='TNK2', Aliases='ACK, ACK-1, ACK1, p21cdc42Hs', description='tyrosine kinase, non-receptor, 2', other_designations='activated Cdc42-associated kinase 1|activated p21cdc42Hs kinase|tyrosine kinase non-receptor protein 2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195863364, end_position_on_the_genomic_accession=195909009, orientation='minus', exon_count=22, OMIM=606994, no_hdr0=''),
  10189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10189, CurrentID=0, Status='live', Symbol='ALYREF', Aliases='ALY, ALY/REF, BEF, REF, THOC4', description='Aly/REF export factor', other_designations='THO complex 4|ally of AML-1 and LEF-1|bZIP enhancing factor|bZIP-enhancing factor BEF|tho4|transcriptional coactivator Aly/REF', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81887835, end_position_on_the_genomic_accession=81891622, orientation='minus', exon_count=7, OMIM=604171, no_hdr0=''),
  10190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10190, CurrentID=0, Status='live', Symbol='TXNDC9', Aliases='APACD, PHLP3', description='thioredoxin domain containing 9', other_designations='ATP binding protein associated with cell differentiation|ATP-binding protein associated with cell differentiation|protein 1-4', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99319024, end_position_on_the_genomic_accession=99336397, orientation='minus', exon_count=5, OMIM=612564, no_hdr0=''),
  10193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10193, CurrentID=0, Status='live', Symbol='RNF41', Aliases='FLRF, NRDP1, SBBI03', description='ring finger protein 41, E3 ubiquitin protein ligase', other_designations='fetal liver ring finger|neuregulin receptor degradation protein-1', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56202504, end_position_on_the_genomic_accession=56221982, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  10194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10194, CurrentID=0, Status='live', Symbol='TSHZ1', Aliases='CAA, NY-CO-33, SDCCAG33, TSH1', description='teashirt zinc finger homeobox 1', other_designations='antigen NY-CO-33|serologically defined colon cancer antigen 33', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75210755, end_position_on_the_genomic_accession=75289950, orientation='plus', exon_count=4, OMIM=614427, no_hdr0=''),
  10195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10195, CurrentID=0, Status='live', Symbol='ALG3', Aliases='CDG1D, CDGS4, CDGS6, D16Ertd36e, NOT56L, Not56, not', description='ALG3, alpha-1,3- mannosyltransferase', other_designations='Not56-like protein|asparagine-linked glycosylation 3 homolog (S. cerevisiae, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation 3 homolog (yeast, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation 3, alpha-1,3- mannosyltransferase homolog|asparagine-linked glycosylation protein 3 homolog|carbohydrate deficient glycoprotein syndrome type IV|dol-P-Man dependent alpha(1-3)-mannosyltransferase|dol-P-Man dependent alpha-1,3- mannosyltransferase|dol-P-Man-dependent alpha(1-3)-mannosyltransferase|dolichyl-P-Man:Man(5)GlcNAc(2)-PP-dolichyl mannosyltransferase|dolichyl-phosphate-mannose--glycolipid alpha-mannosyltransferase', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184242329, end_position_on_the_genomic_accession=184249525, orientation='minus', exon_count=12, OMIM=608750, no_hdr0=''),
  10196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10196, CurrentID=0, Status='live', Symbol='PRMT3', Aliases='HRMT1L3', description='protein arginine methyltransferase 3', other_designations='HMT1 hnRNP methyltransferase-like 3|heterogeneous nuclear ribonucleoprotein methyltransferase-like protein 3', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20387530, end_position_on_the_genomic_accession=20509333, orientation='plus', exon_count=17, OMIM=603190, no_hdr0=''),
  10197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10197, CurrentID=0, Status='live', Symbol='PSME3', Aliases='HEL-S-283, Ki, PA28-gamma, PA28G, PA28gamma, REG-GAMMA', description='proteasome activator subunit 3', other_designations='11S regulator complex gamma subunit|11S regulator complex subunit gamma|Ki antigen|Ki nuclear autoantigen|PA28 gamma variant 5|REG gamma-3|activator of multicatalytic protease subunit 3|epididymis secretory protein Li 283|proteasome (prosome, macropain) activator subunit 3 (PA28 gamma; Ki)|proteasome activator 28 subunit gamma|proteasome activator 28-gamma', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42833142, end_position_on_the_genomic_accession=42843760, orientation='plus', exon_count=13, OMIM=605129, no_hdr0=''),
  10198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10198, CurrentID=0, Status='live', Symbol='MPHOSPH9', Aliases='MPP-9, MPP9', description='M-phase phosphoprotein 9', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123152324, end_position_on_the_genomic_accession=123244014, orientation='minus', exon_count=28, OMIM=605501, no_hdr0=''),
  10199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10199, CurrentID=0, Status='live', Symbol='MPHOSPH10', Aliases='CT90, MPP10, MPP10P, PPP1R106', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein)', other_designations='m phase phosphoprotein 10|protein phosphatase 1, regulatory subunit 106', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71130314, end_position_on_the_genomic_accession=71150102, orientation='plus', exon_count=11, OMIM=605503, no_hdr0=''),
  23636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23636, CurrentID=0, Status='live', Symbol='NUP62', Aliases='IBSN, SNDI, p62', description='nucleoporin 62kDa', other_designations='62 kDa nucleoporin|nucleoporin 62kD|nucleoporin Nup62', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49906825, end_position_on_the_genomic_accession=49929731, orientation='minus', exon_count=3, OMIM=605815, no_hdr0=''),
  10201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10201, CurrentID=0, Status='live', Symbol='NME6', Aliases='IPIA-ALPHA, NDK 6, NM23-H6', description='NME/NM23 nucleoside diphosphate kinase 6', other_designations='NDP kinase 6|inhibitor of p53-induced apoptosis-alpha|non-metastatic cells 6, protein expressed in (nucleoside-diphosphate kinase)', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48292317, end_position_on_the_genomic_accession=48302246, orientation='minus', exon_count=6, OMIM=608294, no_hdr0=''),
  10202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10202, CurrentID=0, Status='live', Symbol='DHRS2', Aliases='HEP27, SDR25C1', description='dehydrogenase/reductase (SDR family) member 2', other_designations='dehydrogenase/reductase member 2|dicarbonyl reductase HEP27|protein D|short chain dehydrogenase/reductase family 25C member 1|short chain dehydrogenase/reductase family 25C, member 1|short-chain alcohol dehydrogenase family member', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23630115, end_position_on_the_genomic_accession=23645639, orientation='plus', exon_count=13, OMIM=615194, no_hdr0=''),
  10203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10203, CurrentID=0, Status='live', Symbol='CALCRL', Aliases='CGRPR, CRLR', description='calcitonin receptor like receptor', other_designations='CGRP type 1 receptor|calcitonin receptor-like receptor', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187341963, end_position_on_the_genomic_accession=187448294, orientation='minus', exon_count=17, OMIM=114190, no_hdr0=''),
  10204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10204, CurrentID=0, Status='live', Symbol='NUTF2', Aliases='NTF2, PP15', description='nuclear transport factor 2', other_designations='NTF-2|placental protein 15', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67846916, end_position_on_the_genomic_accession=67872567, orientation='plus', exon_count=6, OMIM=605813, no_hdr0=''),
  10205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10205, CurrentID=0, Status='live', Symbol='MPZL2', Aliases='EVA, EVA1', description='myelin protein zero-like 2', other_designations='epithelial V-like antigen 1', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118253416, end_position_on_the_genomic_accession=118264536, orientation='minus', exon_count=6, OMIM=604873, no_hdr0=''),
  10206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10206, CurrentID=0, Status='live', Symbol='TRIM13', Aliases='CAR, DLEU5, LEU5, RFP2, RNF77', description='tripartite motif containing 13', other_designations='B-cell chronic lymphocytic leukemia tumor suppressor Leu5|CLL-associated RING finger|RING finger protein 77|leukemia-associated protein 5|putative tumor suppressor RFP2|ret finger protein 2|tripartite motif protein 13|tripartite motif-containing 13|tripartite motif-containing protein 13', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49997007, end_position_on_the_genomic_accession=50018467, orientation='plus', exon_count=4, OMIM=605661, no_hdr0=''),
  10207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10207, CurrentID=0, Status='live', Symbol='INADL', Aliases='Cipp, InaD-like, PATJ, hINADL', description='InaD-like (Drosophila)', other_designations='PALS1-associated tight junction protein|PDZ domain protein|channel-interacting PDZ domain protein|inactivation no after-potential D-like protein|inactivation-no-afterpotential D-like|protein associated to tight junctions', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61742477, end_position_on_the_genomic_accession=62163920, orientation='plus', exon_count=51, OMIM=603199, no_hdr0=''),
  10208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10208, CurrentID=0, Status='live', Symbol='USPL1', Aliases='C13orf22, D13S106E, bA121O19.1', description='ubiquitin specific peptidase like 1', other_designations='highly charged protein|ubiquitin-specific peptidase-like protein 1', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30617693, end_position_on_the_genomic_accession=30659557, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  10209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10209, CurrentID=0, Status='live', Symbol='EIF1', Aliases='A121, EIF-1A, ISO1, SUI1, EIF1', description='eukaryotic translation initiation factor 1', other_designations='protein translation factor SUI1 homolog|sui1iso1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41688875, end_position_on_the_genomic_accession=41691646, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10210, CurrentID=0, Status='live', Symbol='TOPORS', Aliases='LUN, P53BP3, RP31, TP53BPL', description='topoisomerase I binding, arginine/serine-rich, E3 ubiquitin protein ligase', other_designations='SUMO1-protein E3 ligase Topors|topoisomerase I-binding RING finger protein|tumor suppressor p53-binding protein 3', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32540544, end_position_on_the_genomic_accession=32552628, orientation='minus', exon_count=3, OMIM=609507, no_hdr0=''),
  10211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10211, CurrentID=0, Status='live', Symbol='FLOT1', Aliases='', description='flotillin 1', other_designations='integral membrane component of caveolae', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30727734, end_position_on_the_genomic_accession=30742685, orientation='minus', exon_count=14, OMIM=606998, no_hdr0=''),
  10212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10212, CurrentID=0, Status='live', Symbol='DDX39A', Aliases='BAT1, BAT1L, DDX39, DDXL, URH49', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39A', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39|DEAD box protein 39|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 39|UAP56-related helicase, 49 kDa|nuclear RNA helicase URH49|nuclear RNA helicase, DECD variant of DEAD box family', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14408798, end_position_on_the_genomic_accession=14419383, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  10213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10213, CurrentID=0, Status='live', Symbol='PSMD14', Aliases='PAD1, POH1, RPN11', description='proteasome 26S subunit, non-ATPase 14', other_designations='26S proteasome regulatory subunit rpn11|26S proteasome-associated PAD1 homolog 1|proteasome (prosome, macropain) 26S subunit, non-ATPase, 14', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161308275, end_position_on_the_genomic_accession=161411717, orientation='plus', exon_count=12, OMIM=607173, no_hdr0=''),
  10214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10214, CurrentID=0, Status='live', Symbol='SSX3', Aliases='CT5.3', description='synovial sarcoma, X breakpoint 3', other_designations='Kruppel-associated box containing gene product SSX3|cancer/testis antigen 5.3', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48346428, end_position_on_the_genomic_accession=48356753, orientation='minus', exon_count=9, OMIM=300325, no_hdr0=''),
  10215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10215, CurrentID=0, Status='live', Symbol='OLIG2', Aliases='BHLHB1, OLIGO2, PRKCBP2, RACK17, bHLHe19', description='oligodendrocyte lineage transcription factor 2', other_designations='basic domain, helix-loop-helix protein, class B, 1|class B basic helix-loop-helix protein 1|class E basic helix-loop-helix protein 19|human protein kinase C-binding protein RACK17|oligodendrocyte-specific bHLH transcription factor 2|protein kinase C-binding protein 2', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33025908, end_position_on_the_genomic_accession=33029196, orientation='plus', exon_count=2, OMIM=606386, no_hdr0=''),
  10216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10216, CurrentID=0, Status='live', Symbol='PRG4', Aliases='CACP, HAPO, JCAP, MSF, SZP', description='proteoglycan 4', other_designations='articular superficial zone protein|hemangiopoietin|lubricin|megakaryocyte stimulating factor|superficial zone proteoglycan', map_location='1q25-q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186296273, end_position_on_the_genomic_accession=186314562, orientation='plus', exon_count=13, OMIM=604283, no_hdr0=''),
  10217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10217, CurrentID=0, Status='live', Symbol='CTDSPL', Aliases='C3orf8, HYA22, PSR1, RBSP3, SCP3', description='CTD small phosphatase like', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase-like|CTDSP-like|NIF-like protein|NLI-interacting factor 1|RB protein serine phosphatase from chromosome 3|carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 3|nuclear LIM interactor-interacting factor 1|small C-terminal domain phosphatase 3|small CTD phosphatase 3', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37862178, end_position_on_the_genomic_accession=37984469, orientation='plus', exon_count=8, OMIM=608592, no_hdr0=''),
  10218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10218, CurrentID=0, Status='live', Symbol='ANGPTL7', Aliases='AngX, CDT6, dJ647M16.1', description='angiopoietin like 7', other_designations='angiopoietin-like 7|angiopoietin-like factor (CDT6)|angiopoietin-like protein 7|cornea-derived transcript 6 protein', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11189289, end_position_on_the_genomic_accession=11195981, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  10219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10219, CurrentID=0, Status='live', Symbol='KLRG1', Aliases='2F1, CLEC15A, MAFA, MAFA-2F1, MAFA-L, MAFA-LIKE', description='killer cell lectin-like receptor subfamily G, member 1', other_designations='C-type lectin domain family 15 member A|C-type lectin domain family 15, member A|ITIM-containing receptor MAFA-L|MAFA-like receptor|mast cell function-associated antigen (ITIM-containing)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8989625, end_position_on_the_genomic_accession=9010747, orientation='plus', exon_count=5, OMIM=604874, no_hdr0=''),
  10220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10220, CurrentID=0, Status='live', Symbol='GDF11', Aliases='BMP-11, BMP11', description='growth differentiation factor 11', other_designations='GDF-11|bone morphogenetic protein 11', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55743278, end_position_on_the_genomic_accession=55757466, orientation='plus', exon_count=4, OMIM=603936, no_hdr0=''),
  10221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10221, CurrentID=0, Status='live', Symbol='TRIB1', Aliases='C8FW, GIG-2, GIG2, SKIP1, TRB-1, TRB1', description='tribbles pseudokinase 1', other_designations='G-protein-coupled receptor induced protein|G-protein-coupled receptor-induced gene 2 protein|G-protein-coupled receptor-induced protein 2|phosphoprotein regulated by mitogenic pathways|tribbles-like protein 1', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125430321, end_position_on_the_genomic_accession=125438405, orientation='plus', exon_count=4, OMIM=609461, no_hdr0=''),
  284034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284034, CurrentID=0, Status='live', Symbol='LINC00670', Aliases='', description='long intergenic non-protein coding RNA 670', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12549968, end_position_on_the_genomic_accession=12637187, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10223, CurrentID=0, Status='live', Symbol='GPA33', Aliases='A33', description='glycoprotein A33 (transmembrane)', other_designations='transmembrane glycoprotein A33', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167052836, end_position_on_the_genomic_accession=167090631, orientation='minus', exon_count=7, OMIM=602171, no_hdr0=''),
  10224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10224, CurrentID=0, Status='live', Symbol='ZNF443', Aliases='ZK1', description='zinc finger protein 443', other_designations='Kruppel-type zinc finger (C2H2)|krueppel-type zinc finger protein ZK1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12429706, end_position_on_the_genomic_accession=12457521, orientation='minus', exon_count=6, OMIM=606697, no_hdr0=''),
  10225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10225, CurrentID=0, Status='live', Symbol='CD96', Aliases='TACTILE', description='CD96 molecule', other_designations='CD96 antigen|T cell activation, increased late expression|cell surface antigen CD96|t cell-activated increased late expression protein', map_location='3q13.13-q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111542079, end_position_on_the_genomic_accession=111665996, orientation='plus', exon_count=19, OMIM=606037, no_hdr0=''),
  10226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10226, CurrentID=0, Status='live', Symbol='PLIN3', Aliases='M6PRBP1, PP17, TIP47', description='perilipin 3', other_designations='47 kDa MPR-binding protein|cargo selection protein TIP47|mannose-6-phosphate receptor-binding protein 1|placental protein 17|tail-interacting protein, 47 kD', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4838334, end_position_on_the_genomic_accession=4867768, orientation='minus', exon_count=8, OMIM=602702, no_hdr0=''),
  10227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10227, CurrentID=0, Status='live', Symbol='MFSD10', Aliases='TETRAN, TETTRAN', description='major facilitator superfamily domain containing 10', other_designations='tetracycline transporter-like protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2930560, end_position_on_the_genomic_accession=2934868, orientation='minus', exon_count=13, OMIM=610977, no_hdr0=''),
  10228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10228, CurrentID=0, Status='live', Symbol='STX6', Aliases='', description='syntaxin 6', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180949690, end_position_on_the_genomic_accession=181023121, orientation='minus', exon_count=9, OMIM=603944, no_hdr0=''),
  10229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10229, CurrentID=0, Status='live', Symbol='COQ7', Aliases='CAT5, CLK-1, CLK1', description='coenzyme Q7 homolog, ubiquinone (yeast)', other_designations='COQ7 coenzyme Q, 7 homolog ubiquinone|DMQ hydroxylase|coenzyme Q biosynthesis protein 7 homolog|placental protein KG-20|timing protein clk-1 homolog|ubiquinone biosynthesis monooxygenase COQ7|ubiquinone biosynthesis protein COQ7 homolog', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19067595, end_position_on_the_genomic_accession=19080095, orientation='plus', exon_count=8, OMIM=601683, no_hdr0=''),
  10230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10230, CurrentID=0, Status='live', Symbol='NBR2', Aliases='NCRNA00192', description='neighbor of BRCA1 gene 2 (non-protein coding)', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43125583, end_position_on_the_genomic_accession=43145113, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  10231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10231, CurrentID=0, Status='live', Symbol='RCAN2', Aliases='CSP2, DSCR1L1, MCIP2, RCN2, ZAKI-4, ZAKI4', description='regulator of calcineurin 2', other_designations='Down syndrome candidate region 1-like 1|Down syndrome critical region gene 1-like 1|myocyte-enriched calcineurin-interacting protein 2|thyroid hormone-responsive (skin fibroblasts)|thyroid hormone-responsive protein ZAKI-4', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46220730, end_position_on_the_genomic_accession=46492067, orientation='minus', exon_count=8, OMIM=604876, no_hdr0=''),
  10232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10232, CurrentID=0, Status='live', Symbol='MSLN', Aliases='MPF, SMRP', description='mesothelin', other_designations='CAK1 antigen|megakaryocyte potentiating factor|pre-pro-megakaryocyte-potentiating factor|soluble MPF mesothelin related protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=760732, end_position_on_the_genomic_accession=768865, orientation='plus', exon_count=19, OMIM=601051, no_hdr0=''),
  10233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10233, CurrentID=0, Status='live', Symbol='LRRC23', Aliases='LRPB7', description='leucine rich repeat containing 23', other_designations='leucine-rich protein B7', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6904733, end_position_on_the_genomic_accession=6914242, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  10234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10234, CurrentID=0, Status='live', Symbol='LRRC17', Aliases='P37NB', description='leucine rich repeat containing 17', other_designations='37 kDa leucine-rich repeat (LRR) protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102912897, end_position_on_the_genomic_accession=102945111, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10235, CurrentID=0, Status='live', Symbol='RASGRP2', Aliases='CALDAG-GEFI, CDC25L', description='RAS guanyl releasing protein 2 (calcium and DAG-regulated)', other_designations='F25B3.3 kinase-like protein|RAS guanyl nucleotide-releasing protein 2|calcium and DAG-regulated guanine nucleotide exchange factor I|calcium and diacylglycerol-regulated guanine nucleotide exchange factor I|cdc25-like protein|guanine exchange factor MCG7|hCDC25L', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64726911, end_position_on_the_genomic_accession=64745456, orientation='minus', exon_count=21, OMIM=605577, no_hdr0=''),
  10236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10236, CurrentID=0, Status='live', Symbol='HNRNPR', Aliases='HNRPR, hnRNP-R', description='heterogeneous nuclear ribonucleoprotein R', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23304688, end_position_on_the_genomic_accession=23344650, orientation='minus', exon_count=11, OMIM=607201, no_hdr0=''),
  10237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10237, CurrentID=0, Status='live', Symbol='SLC35B1', Aliases='UGTREL1', description='solute carrier family 35 member B1', other_designations='UDP-galactose transporter related|UDP-galactose transporter-related protein 1|hUGTrel1|solute carrier family 35, member B1', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49700934, end_position_on_the_genomic_accession=49708135, orientation='minus', exon_count=11, OMIM=610790, no_hdr0=''),
  10238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10238, CurrentID=0, Status='live', Symbol='DCAF7', Aliases='AN11, HAN11, SWAN-1, WDR68', description='DDB1 and CUL4 associated factor 7', other_designations='WD repeat-containing protein 68|WD repeat-containing protein An11 homolog|human anthocyanin|seven-WD-repeat protein of the AN11 family-1', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63550435, end_position_on_the_genomic_accession=63594279, orientation='plus', exon_count=7, OMIM=605973, no_hdr0=''),
  10239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10239, CurrentID=0, Status='live', Symbol='AP3S2', Aliases='AP3S3, sigma3b', description='adaptor related protein complex 3, sigma 2 subunit', other_designations='AP-3 complex subunit sigma-3B|adapter-related protein complex 3 subunit sigma-2|adaptor complex sigma3B|adaptor-related protein complex 3 subunit sigma-2|adaptor-related protein complex 3, sigma 2 subunit|clathrin-associated/assembly/adaptor protein, small 4, 22-kD|sigma-3B-adaptin|sigma-adaptin 3b', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89830599, end_position_on_the_genomic_accession=89894385, orientation='minus', exon_count=7, OMIM=602416, no_hdr0=''),
  10240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10240, CurrentID=0, Status='live', Symbol='MRPS31', Aliases='IMOGN38, MRP-S31, S31mt', description='mitochondrial ribosomal protein S31', other_designations='imogen 38', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40729296, end_position_on_the_genomic_accession=40771223, orientation='minus', exon_count=8, OMIM=611992, no_hdr0=''),
  10241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10241, CurrentID=0, Status='live', Symbol='CALCOCO2', Aliases='NDP52', description='calcium binding and coiled-coil domain 2', other_designations='antigen nuclear dot 52 kDa protein|nuclear domain 10 protein 52|nuclear domain 10 protein NDP52|nuclear dot protein 52', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48830988, end_position_on_the_genomic_accession=48865245, orientation='plus', exon_count=16, OMIM=604587, no_hdr0=''),
  10242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10242, CurrentID=0, Status='live', Symbol='KCNMB2', Aliases='', description='potassium channel subfamily M regulatory beta subunit 2', other_designations='BK channel beta subunit 2|BK channel subunit beta-2|MaxiK channel beta 2 subunit|MaxiK channel beta-subunit 2|big potassium channel beta subunit 2|charybdotoxin receptor subunit beta-2|hCG1646471|hbeta2|k(VCA)beta-2|large conductance calcium-activated potassium channel beta 2 subunit|large-conductance Ca2+-activated K+ channel beta2 subunit|potassium large conductance calcium-activated channel, subfamily M, beta member 2|slo-beta-2', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178536298, end_position_on_the_genomic_accession=178844429, orientation='plus', exon_count=8, OMIM=605214, no_hdr0=''),
  10243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10243, CurrentID=0, Status='live', Symbol='GPHN', Aliases='GEPH, GPH, GPHRYN, HKPX1, MOCODC', description='gephyrin', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66507407, end_position_on_the_genomic_accession=67181808, orientation='plus', exon_count=28, OMIM=603930, no_hdr0=''),
  10244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10244, CurrentID=0, Status='live', Symbol='RABEPK', Aliases='RAB9P40, bA65N13.1, p40', description='Rab9 effector protein with kelch motifs', other_designations='40 kDa Rab9 effector protein|Rab9 effector p40', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125200542, end_position_on_the_genomic_accession=125234159, orientation='plus', exon_count=10, OMIM=605962, no_hdr0=''),
  10245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10245, CurrentID=0, Status='live', Symbol='TIMM17B', Aliases='DXS9822, JM3, TIM17B', description='translocase of inner mitochondrial membrane 17 homolog B (yeast)', other_designations='inner mitochondrial membrane preprotein translocase', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48893447, end_position_on_the_genomic_accession=48898143, orientation='minus', exon_count=8, OMIM=300249, no_hdr0=''),
  10246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10246, CurrentID=0, Status='live', Symbol='SLC17A2', Aliases='NPT3', description='solute carrier family 17, member 2', other_designations='Na(+)/PI cotransporter 3|sodium phosphate transporter 3|sodium/phosphate cotransporter 3|solute carrier family 17 (vesicular glutamate transporter), member 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25912754, end_position_on_the_genomic_accession=25930726, orientation='minus', exon_count=12, OMIM=611049, no_hdr0=''),
  10247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10247, CurrentID=0, Status='live', Symbol='HRSP12', Aliases='P14.5, PSP, UK114', description='heat-responsive protein 12', other_designations='14.5 kDa translational inhibitor protein|UK114 antigen homolog|perchloric acid-soluble protein|translational inhibitor p14.5|translational inhibitor protein p14.5', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98102339, end_position_on_the_genomic_accession=98117190, orientation='minus', exon_count=6, OMIM=602487, no_hdr0=''),
  10248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10248, CurrentID=0, Status='live', Symbol='POP7', Aliases='0610037N12Rik, RPP2, RPP20', description='POP7 homolog, ribonuclease P/MRP subunit', other_designations='POP7 (processing of precursor, S. cerevisiae) homolog|RNaseP protein p20|hPOP7|processing of precursor 7, ribonuclease P subunit|processing of precursor 7, ribonuclease P/MRP subunit|ribonucleases P/MRP protein subunit POP7 homolog', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100706053, end_position_on_the_genomic_accession=100707500, orientation='plus', exon_count=2, OMIM=606113, no_hdr0=''),
  10249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10249, CurrentID=0, Status='live', Symbol='GLYAT', Aliases='ACGNAT, CAT, GAT', description='glycine-N-acyltransferase', other_designations='AAc|HRP-1(CLP)|acyl-CoA:glycine N-acyltransferase|aralkyl acyl-CoA N-acyltransferase|aralkyl acyl-CoA:amino acid N-acyltransferase|aralkyl-CoA N-acyltransferase|benzoyl-coenzyme A:glycine N-acyltransferase|glycine N-benzoyltransferase', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58708757, end_position_on_the_genomic_accession=58731974, orientation='minus', exon_count=5, OMIM=607424, no_hdr0=''),
  10250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10250, CurrentID=0, Status='live', Symbol='SRRM1', Aliases='160-KD, POP101, SRM160', description='serine/arginine repetitive matrix 1', other_designations='SR-related nuclear matrix protein of 160 kDa|Ser/Arg-related nuclear matrix protein', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24643012, end_position_on_the_genomic_accession=24673281, orientation='plus', exon_count=20, OMIM=605975, no_hdr0=''),
  10251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10251, CurrentID=0, Status='live', Symbol='SPRY3', Aliases='spry-3', description='sprouty RTK signaling antagonist 3', other_designations='antagonist of FGF signaling|sprouty homolog 3', map_location='Xq28 and Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=56954255, end_position_on_the_genomic_accession=56968979, orientation='plus', exon_count=2, OMIM=300531, no_hdr0=''),
  10252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10252, CurrentID=0, Status='live', Symbol='SPRY1', Aliases='hSPRY1', description='sprouty RTK signaling antagonist 1', other_designations='sprouty homolog 1, antagonist of FGF signaling|sprouty, Drosophila, homolog of, 1 (antagonist of FGF signaling)|spry-1', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123396791, end_position_on_the_genomic_accession=123403760, orientation='plus', exon_count=5, OMIM=602465, no_hdr0=''),
  10253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10253, CurrentID=0, Status='live', Symbol='SPRY2', Aliases='hSPRY2', description='sprouty RTK signaling antagonist 2', other_designations='sprouty 2|sprouty homolog 2|spry-2', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80335976, end_position_on_the_genomic_accession=80340951, orientation='minus', exon_count=4, OMIM=602466, no_hdr0=''),
  10254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10254, CurrentID=0, Status='live', Symbol='STAM2', Aliases='HbpA, STAM2B, STAM2', description='signal transducing adaptor molecule 2', other_designations='HSE1 homolog|Hrs-binding protein|STAM-2|STAM-like protein containing SH3 and ITAM domains 2|signal transducing adaptor molecule (SH3 domain and ITAM motif) 2', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152116801, end_position_on_the_genomic_accession=152175992, orientation='minus', exon_count=15, OMIM=606244, no_hdr0=''),
  10255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10255, CurrentID=0, Status='live', Symbol='HCG9', Aliases='HCGIX, HCGIX4', description='HLA complex group 9 (non-protein coding)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29975115, end_position_on_the_genomic_accession=29978403, orientation='plus', exon_count=3, OMIM=615797, no_hdr0=''),
  10256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10256, CurrentID=0, Status='live', Symbol='CNKSR1', Aliases='CNK, CNK1', description='connector enhancer of kinase suppressor of Ras 1', other_designations='CNK homolog protein 1|connector enhancer of KSR 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26177490, end_position_on_the_genomic_accession=26189884, orientation='plus', exon_count=22, OMIM=603272, no_hdr0=''),
  10257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10257, CurrentID=0, Status='live', Symbol='ABCC4', Aliases='MOAT-B, MOATB, MRP4', description='ATP binding cassette subfamily C member 4', other_designations='ATP-binding cassette, sub-family C (CFTR/MRP), member 4|MRP/cMOAT-related ABC transporter|bA464I2.1 (ATP-binding cassette, sub-family C (CFTR/MRP), member 4)|canalicular multispecific organic anion transporter (ABC superfamily)|multi-specific organic anion transporter B', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95019829, end_position_on_the_genomic_accession=95301446, orientation='minus', exon_count=35, OMIM=605250, no_hdr0=''),
  10260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10260, CurrentID=0, Status='live', Symbol='DENND4A', Aliases='IRLB, MYCPBP', description='DENN/MADD domain containing 4A', other_designations='DENN domain-containing protein 4A|c-myc promoter binding protein', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65659123, end_position_on_the_genomic_accession=65792293, orientation='minus', exon_count=38, OMIM=600382, no_hdr0=''),
  10261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10261, CurrentID=0, Status='live', Symbol='IGSF6', Aliases='DORA', description='immunoglobulin superfamily member 6', other_designations='down-regulated by activation (immunoglobulin superfamily)|immunoglobulin superfamily, member 6', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21641284, end_position_on_the_genomic_accession=21652666, orientation='minus', exon_count=6, OMIM=606222, no_hdr0=''),
  10262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10262, CurrentID=0, Status='live', Symbol='SF3B4', Aliases='AFD1, Hsh49, SAP49, SF3b49', description='splicing factor 3b subunit 4', other_designations='SAP 49|SF3b50|pre-mRNA-splicing factor SF3b 49 kDa subunit|spliceosomal protein|spliceosome-associated protein (U2 snRNP)|spliceosome-associated protein 49|splicing factor 3b, subunit 4, 49kD|splicing factor 3b, subunit 4, 49kDa', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149923317, end_position_on_the_genomic_accession=149928252, orientation='minus', exon_count=6, OMIM=605593, no_hdr0=''),
  10263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10263, CurrentID=0, Status='live', Symbol='CDK2AP2', Aliases='DOC-1R, p14', description='cyclin-dependent kinase 2 associated protein 2', other_designations='CDK2-associated protein 2|DOC-1-related protein|tumor suppressor deleted in oral cancer related 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67506490, end_position_on_the_genomic_accession=67508728, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372955, CurrentID=0, Status='live', Symbol='LOC105372955', Aliases='', description='uncharacterized LOC105372955', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23510514, end_position_on_the_genomic_accession=23513334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10265, CurrentID=0, Status='live', Symbol='IRX5', Aliases='HMMS, IRX-2a, IRXB2', description='iroquois homeobox 5', other_designations='homeodomain protein IRX-2A|homeodomain protein IRXB2', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54931199, end_position_on_the_genomic_accession=54934485, orientation='plus', exon_count=4, OMIM=606195, no_hdr0=''),
  10266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10266, CurrentID=0, Status='live', Symbol='RAMP2', Aliases='', description='receptor (G protein-coupled) activity modifying protein 2', other_designations='CRLR activity-modifying protein 2|calcitonin receptor-like receptor activity modifying protein 2|calcitonin-receptor-like receptor activity-modifying protein 2|receptor (calcitonin) activity modifying protein 2|receptor activity modifying protein 2|receptor-activity-modifying protein 2', map_location='17q12-q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42761194, end_position_on_the_genomic_accession=42763041, orientation='plus', exon_count=4, OMIM=605154, no_hdr0=''),
  10267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10267, CurrentID=0, Status='live', Symbol='RAMP1', Aliases='', description='receptor (G protein-coupled) activity modifying protein 1', other_designations='CRLR activity-modifying protein 1|calcitonin receptor-like receptor activity modifying protein 1|calcitonin-receptor-like receptor activity-modifying protein 1|receptor (calcitonin) activity modifying protein 1|receptor activity modifying protein 1', map_location='2q36-q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237859544, end_position_on_the_genomic_accession=237912114, orientation='plus', exon_count=6, OMIM=605153, no_hdr0=''),
  10268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10268, CurrentID=0, Status='live', Symbol='RAMP3', Aliases='', description='receptor (G protein-coupled) activity modifying protein 3', other_designations='CRLR activity-modifying protein 3|calcitonin receptor-like receptor activity modifying protein 3|calcitonin-receptor-like receptor activity-modifying protein 3|receptor (calcitonin) activity modifying protein 3|receptor activity modifying protein 3|receptor-activity-modifying protein 3', map_location='7p13-p12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45157768, end_position_on_the_genomic_accession=45184250, orientation='plus', exon_count=5, OMIM=605155, no_hdr0=''),
  10269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10269, CurrentID=0, Status='live', Symbol='ZMPSTE24', Aliases='FACE-1, FACE1, HGPS, PRO1, STE24, Ste24p', description='zinc metallopeptidase STE24', other_designations='farnesylated proteins-converting enzyme 1|prenyl protein-specific endoprotease 1|zinc metallopeptidase STE24 homolog|zinc metalloproteinase Ste24 homolog', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40258050, end_position_on_the_genomic_accession=40294184, orientation='plus', exon_count=12, OMIM=606480, no_hdr0=''),
  10270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10270, CurrentID=0, Status='live', Symbol='AKAP8', Aliases='AKAP 95, AKAP-8, AKAP-95, AKAP95', description='A-kinase anchoring protein 8', other_designations='A kinase (PRKA) anchor protein 8|A-kinase anchor protein, 95kDa', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15353521, end_position_on_the_genomic_accession=15379828, orientation='minus', exon_count=15, OMIM=604692, no_hdr0=''),
  10272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10272, CurrentID=0, Status='live', Symbol='FSTL3', Aliases='FLRG, FSRP', description='follistatin-like 3 (secreted glycoprotein)', other_designations='follistatin-like protein 3|follistatin-related gene protein', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=676389, end_position_on_the_genomic_accession=683392, orientation='plus', exon_count=5, OMIM=605343, no_hdr0=''),
  10273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10273, CurrentID=0, Status='live', Symbol='STUB1', Aliases='CHIP, HSPABP2, NY-CO-7, SCAR16, SDCCAG7, UBOX1', description='STIP1 homology and U-box containing protein 1, E3 ubiquitin protein ligase', other_designations='CLL-associated antigen KW-8|antigen NY-CO-7|carboxy terminus of Hsp70-interacting protein|heat shock protein A binding protein 2 (c-terminal)|serologically defined colon cancer antigen 7', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=680111, end_position_on_the_genomic_accession=682768, orientation='plus', exon_count=7, OMIM=607207, no_hdr0=''),
  10274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10274, CurrentID=0, Status='live', Symbol='STAG1', Aliases='SA1, SCC3A', description='stromal antigen 1', other_designations='SCC3 homolog 1|nuclear protein stromal antigen 1', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136336234, end_position_on_the_genomic_accession=136752403, orientation='minus', exon_count=37, OMIM=604358, no_hdr0=''),
  10276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10276, CurrentID=0, Status='live', Symbol='NET1', Aliases='ARHGEF8A, NET1', description='neuroepithelial cell transforming 1', other_designations='Rho guanine nucleotide exchange factor (GEF) 8|guanine nucleotide regulatory protein (oncogene)|neuroepithelioma transforming gene 1|p65 Net1 proto-oncogene protein|small GTP-binding protein regulator', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5412551, end_position_on_the_genomic_accession=5459056, orientation='plus', exon_count=13, OMIM=606450, no_hdr0=''),
  10277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10277, CurrentID=0, Status='live', Symbol='UBE4B', Aliases='E4, HDNB1, UBOX3, UFD2, UFD2A', description='ubiquitination factor E4B', other_designations='UFD2A-III/UBE4B-III splice isoform|homologous to yeast UFD2|homozygously deleted in neuroblastoma 1|homozygously deleted in neuroblastoma-1|ubiquitin fusion degradation protein 2|ubiquitin-fusion degradation protein 2|ubiquitination factor E4B (UFD2 homolog, yeast)|ubiquitination factor E4B (homologous to yeast UFD2)', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10032983, end_position_on_the_genomic_accession=10181239, orientation='plus', exon_count=31, OMIM=613565, no_hdr0=''),
  10278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10278, CurrentID=0, Status='live', Symbol='EFS', Aliases='CAS3, CASS31, EFS2, HEFS, SIN, EFS', description='embryonal Fyn-associated substrate', other_designations='Cas scaffolding protein family member 3|signal transduction protein (SH3 containing)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23356400, end_position_on_the_genomic_accession=23365633, orientation='minus', exon_count=6, OMIM=609906, no_hdr0=''),
  10279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10279, CurrentID=0, Status='live', Symbol='PRSS16', Aliases='TSSP', description='protease, serine 16', other_designations='protease, serine, 16 (thymus)|serine protease 16|thymus specific serine peptidase', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27247723, end_position_on_the_genomic_accession=27256620, orientation='plus', exon_count=12, OMIM=607169, no_hdr0=''),
  10280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10280, CurrentID=0, Status='live', Symbol='SIGMAR1', Aliases='ALS16, OPRS1, SIG-1R, SR-BP, SR-BP1, SRBP, hSigmaR1, sigma1R', description='sigma non-opioid intracellular receptor 1', other_designations='SR31747 binding protein 1|aging-associated gene 8 protein|sigma 1-type opioid receptor', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34634722, end_position_on_the_genomic_accession=34637826, orientation='minus', exon_count=4, OMIM=601978, no_hdr0=''),
  10281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10281, CurrentID=0, Status='live', Symbol='DSCR4', Aliases='DCRB, DSCRB', description='Down syndrome critical region 4', other_designations='Down syndrome critical region gene 4|Down syndrome critical region protein B', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38006553, end_position_on_the_genomic_accession=38121360, orientation='minus', exon_count=5, OMIM=604829, no_hdr0=''),
  10282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10282, CurrentID=0, Status='live', Symbol='BET1', Aliases='HBET1', description='Bet1 golgi vesicular membrane trafficking protein', other_designations='Bet1p homolog|Golgi vesicular membrane trafficking protein p18|blocked early in transport 1 homolog|golgi vesicular membrane-trafficking protein p18', map_location='7q21.1-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93962770, end_position_on_the_genomic_accession=94004390, orientation='minus', exon_count=7, OMIM=605456, no_hdr0=''),
  10283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10283, CurrentID=0, Status='live', Symbol='CWC27', Aliases='NY-CO-10, SDCCAG-10, SDCCAG10', description='CWC27 spliceosome-associated protein homolog', other_designations='PPIase CWC27|PPIase SDCCAG10|antigen NY-CO-10|peptidyl-prolyl cis-trans isomerase SDCCAG10|serologically defined colon cancer antigen 10', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64768918, end_position_on_the_genomic_accession=65018763, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  10284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10284, CurrentID=0, Status='live', Symbol='SAP18', Aliases='2HOR0202P, SAP18', description='Sin3A associated protein 18kDa', other_designations='18 kDa Sin3-associated polypeptide|Sin3A-associated protein, 18kDa|cell growth inhibiting protein 38|cell growth-inhibiting gene 38 protein|cell growth-inhibiting protein 38|histone deacetlyase complex subunit SAP18|sin3-associated polypeptide p18|sin3-associated polypeptide, 18 kDa|sin3-associated polypeptide, p18', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21140514, end_position_on_the_genomic_accession=21149085, orientation='plus', exon_count=5, OMIM=602949, no_hdr0=''),
  10285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10285, CurrentID=0, Status='live', Symbol='SMNDC1', Aliases='SMNR, SPF30, TDRD16C', description='survival motor neuron domain containing 1', other_designations='30 kDa splicing factor SMNrp|SMN-related protein|splicing factor 30, survival of motor neuron-related|survival motor neuron domain-containing protein 1|tudor domain containing 16C', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110293040, end_position_on_the_genomic_accession=110304949, orientation='minus', exon_count=7, OMIM=603519, no_hdr0=''),
  10286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10286, CurrentID=0, Status='live', Symbol='BCAS2', Aliases='DAM1, SPF27, Snt309', description='breast carcinoma amplified sequence 2', other_designations='DNA amplified in mammary carcinoma 1 protein|breast carcinoma-amplified sequence 2|spliceosome associated protein, amplified in breast cancer|spliceosome-associated protein SPF 27', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114567557, end_position_on_the_genomic_accession=114581644, orientation='minus', exon_count=7, OMIM=605783, no_hdr0=''),
  10287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10287, CurrentID=0, Status='live', Symbol='RGS19', Aliases='GAIP, RGSGAIP', description='regulator of G-protein signaling 19', other_designations='G alpha interacting protein|G protein signalling regulator 19|G-alpha-interacting protein|guanine nucleotide binding protein alpha inhibiting activity polypeptide 3 interacting protein|regulator of G-protein signalling 19', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64073181, end_position_on_the_genomic_accession=64080003, orientation='minus', exon_count=9, OMIM=605071, no_hdr0=''),
  10288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10288, CurrentID=0, Status='live', Symbol='LILRB2', Aliases='CD85D, ILT-4, ILT4, LIR-2, LIR2, MIR-10, MIR10', description='leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 2', other_designations='CD85 antigen-like family member D|Ig-like transcript 4|monocyte/macrophage immunoglobulin-like receptor 10', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54273812, end_position_on_the_genomic_accession=54281178, orientation='minus', exon_count=16, OMIM=604815, no_hdr0=''),
  285704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285704, CurrentID=0, Status='live', Symbol='RGMB', Aliases='DRAGON', description='repulsive guidance molecule family member b', other_designations='DRG11-responsive axonal guidance and outgrowth of neurite|RGM domain family, member B|repulsive guidance molecule B', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98769295, end_position_on_the_genomic_accession=98796494, orientation='plus', exon_count=7, OMIM=612687, no_hdr0=''),
  10290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10290, CurrentID=0, Status='live', Symbol='SPEG', Aliases='APEG-1, APEG1, BPEG, CNM5alpha, SPEGbeta, SPEG', description='SPEG complex locus', other_designations='aortic preferentially expressed gene 1|aortic preferentially expressed protein 1|nuclear protein, marker for differentiated aortic smooth muscle and down-regulated with vascular injury', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219434192, end_position_on_the_genomic_accession=219493632, orientation='plus', exon_count=50, OMIM=615950, no_hdr0=''),
  10291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10291, CurrentID=0, Status='live', Symbol='SF3A1', Aliases='PRP21, PRPF21, SAP11420, SF3A1', description='splicing factor 3a subunit 1', other_designations='SAP 114|pre-mRNA processing 21|pre-mRNA splicing factor SF3a (120 kDa subunit)|spliceosome-associated protein 114|splicing factor 3 subunit 1|splicing factor 3a, subunit 1, 120kD|splicing factor 3a, subunit 1, 120kDa', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30331988, end_position_on_the_genomic_accession=30356947, orientation='minus', exon_count=16, OMIM=605595, no_hdr0=''),
  10293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10293, CurrentID=0, Status='live', Symbol='TRAIP', Aliases='RNF206, TRIP', description='TRAF interacting protein', other_designations='TRAF-interacting protein|ring finger protein 206', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49828595, end_position_on_the_genomic_accession=49856564, orientation='minus', exon_count=16, OMIM=605958, no_hdr0=''),
  10294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10294, CurrentID=0, Status='live', Symbol='DNAJA2', Aliases='CPR3, DJ3, DJA2, DNAJ, DNJ3, HIRIP4, PRO3015, RDJ2', description='DnaJ heat shock protein family (Hsp40) member A2', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 2|HIRA interacting protein 4|HIRA-interacting protein 4|cell cycle progression 3 protein|cell cycle progression restoration gene 3 protein|renal carcinoma antigen NY-REN-14', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46955362, end_position_on_the_genomic_accession=46973714, orientation='minus', exon_count=9, OMIM=611322, no_hdr0=''),
  10295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10295, CurrentID=0, Status='live', Symbol='BCKDK', Aliases='BCKDKD, BDK', description='branched chain ketoacid dehydrogenase kinase', other_designations='BCKD-kinase|BCKDHKIN|branched chain alpha-ketoacid dehydrogenase kinase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31108294, end_position_on_the_genomic_accession=31112791, orientation='plus', exon_count=11, OMIM=614901, no_hdr0=''),
  10296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10296, CurrentID=0, Status='live', Symbol='MAEA', Aliases='EMLP, EMP, GID9, HLC-10, PIG5', description='macrophage erythroblast attacher', other_designations='GID complex subunit 9, FYV10 homolog|cell proliferation-inducing gene 5 protein|erythroblast macrophage protein|human lung cancer oncogene 10 protein|lung cancer-related protein 10', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1289851, end_position_on_the_genomic_accession=1340148, orientation='plus', exon_count=14, OMIM=606801, no_hdr0=''),
  10297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10297, CurrentID=0, Status='live', Symbol='APC2', Aliases='APCL', description='adenomatosis polyposis coli 2', other_designations='adenomatous polyposis coli like', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1446268, end_position_on_the_genomic_accession=1473244, orientation='plus', exon_count=17, OMIM=612034, no_hdr0=''),
  10298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10298, CurrentID=0, Status='live', Symbol='PAK4', Aliases='', description='p21 protein (Cdc42/Rac)-activated kinase 4', other_designations='PAK-4|p21(CDKN1A)-activated kinase 4|p21-activated kinase 4|protein kinase related to S. cerevisiae STE20, effector for Cdc42Hs', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39125780, end_position_on_the_genomic_accession=39179406, orientation='plus', exon_count=14, OMIM=605451, no_hdr0=''),
  10299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10299, CurrentID=0, Status='live', Symbol='MARCH6', Aliases='DOA10, MARCH-VI, RNF176, TEB4', description='membrane associated ring-CH-type finger 6', other_designations='RING finger protein 176|doa10 homolog|membrane associated ring finger 6|membrane-associated RING finger protein 6|membrane-associated RING-CH finger protein 6|membrane-associated RING-CH protein VI|membrane-associated ring finger (C3HC4) 6, E3 ubiquitin protein ligase', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10353639, end_position_on_the_genomic_accession=10440388, orientation='plus', exon_count=30, OMIM=613297, no_hdr0=''),
  10300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10300, CurrentID=0, Status='live', Symbol='KATNB1', Aliases='KAT, LIS6', description='katanin p80 (WD repeat containing) subunit B 1', other_designations='katanin (80 kDa)|katanin p80 (WD40-containing) subunit B 1|katanin p80 WD40-containing subunit B1|katanin p80 subunit B1|p80 katanin', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57735720, end_position_on_the_genomic_accession=57757250, orientation='plus', exon_count=21, OMIM=602703, no_hdr0=''),
  10301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10301, CurrentID=0, Status='live', Symbol='DLEU1', Aliases='BCMS, BCMS1, DLB1, DLEU2, LEU1, LEU2, LINC00021, NCRNA00021, XTP6', description='deleted in lymphocytic leukemia 1 (non-protein coding)', other_designations='B-cell neoplasia-associated gene with multiple splicing|leukemia-associated protein 2|long intergenic non-protein coding RNA 21', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50082169, end_position_on_the_genomic_accession=50528643, orientation='plus', exon_count=8, OMIM=605765, no_hdr0=''),
  10302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10302, CurrentID=0, Status='live', Symbol='SNAPC5', Aliases='SNAP19', description='small nuclear RNA activating complex polypeptide 5', other_designations='CTD-3185P2.1|SNAPc 19 kDa subunit|SNAPc subunit 5|small nuclear RNA activating complex, polypeptide 5, 19kD|small nuclear RNA activating complex, polypeptide 5, 19kDa|small nuclear RNA-activating complex polypeptide 5|snRNA-activating protein complex 19 kDa subunit', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66490328, end_position_on_the_genomic_accession=66497808, orientation='minus', exon_count=4, OMIM=605979, no_hdr0=''),
  132789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132789, CurrentID=0, Status='live', Symbol='GNPDA2', Aliases='GNP2, SB52', description='glucosamine-6-phosphate deaminase 2', other_designations='glcN6P deaminase 2|glucosamine-6-phosphate isomerase SB52', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44682411, end_position_on_the_genomic_accession=44726634, orientation='minus', exon_count=10, OMIM=613222, no_hdr0=''),
  10305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10305, CurrentID=0, Status='live', Symbol='MMDK', Aliases='MDK', description='Mesomelic dysplasia, Kantaputra type', other_designations='', map_location='2q24-q32', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=156232, no_hdr0=''),
  10306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10306, CurrentID=0, Status='live', Symbol='CTPP', Aliases='ARCC2, CPP, CTPA, CTPP1, ECK', description='Cataract, posterior polar', other_designations='', map_location='1pter-p36.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=116600, no_hdr0=''),
  10307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10307, CurrentID=0, Status='live', Symbol='APBB3', Aliases='FE65L2, SRA', description='amyloid beta (A4) precursor protein-binding, family B, member 3', other_designations='FE65-like protein 2|amyloid precursor interacting protein|protein Fe65-like 2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140558268, end_position_on_the_genomic_accession=140564604, orientation='minus', exon_count=13, OMIM=602711, no_hdr0=''),
  10308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10308, CurrentID=0, Status='live', Symbol='ZNF267', Aliases='HZF2', description='zinc finger protein 267', other_designations='zinc finger (C2H2)|zinc finger protein 2|zinc finger protein HZF2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31873758, end_position_on_the_genomic_accession=31917308, orientation='plus', exon_count=6, OMIM=604752, no_hdr0=''),
  10309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10309, CurrentID=0, Status='live', Symbol='CCNO', Aliases='CCNU, CILD29, UDG2', description='cyclin O', other_designations='cyclin U|cyclin domain containing', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55231152, end_position_on_the_genomic_accession=55233717, orientation='minus', exon_count=3, OMIM=607752, no_hdr0=''),
  10311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10311, CurrentID=0, Status='live', Symbol='DSCR3', Aliases='DCRA, DSCRA', description='Down syndrome critical region 3', other_designations='Down syndrome critical region gene 3|Down syndrome critical region protein A', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37223425, end_position_on_the_genomic_accession=37268124, orientation='minus', exon_count=10, OMIM=605298, no_hdr0=''),
  10312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10312, CurrentID=0, Status='live', Symbol='TCIRG1', Aliases='ATP6N1C, ATP6V0A3, Atp6i, OC-116kDa, OC116, OPTB1, Stv1, TIRC7, Vph1, a3', description='T-cell, immune regulator 1, ATPase, H+ transporting, lysosomal V0 subunit A3', other_designations='ATPase, H+ transporting, 116kD|OC-116 kDa|T-cell immune response cDNA 7|T-cell, immune regulator 1, ATPase, H+ transporting, lysosomal V0 protein a|V-ATPase 116-kDa|V-type proton ATPase 116 kDa subunit a|osteoclastic proton pump 116 kDa subunit|specific 116-kDa vacuolar proton pump subunit|vacuolar proton translocating ATPase 116 kDa subunit A', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68038995, end_position_on_the_genomic_accession=68050899, orientation='plus', exon_count=22, OMIM=604592, no_hdr0=''),
  10313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10313, CurrentID=0, Status='live', Symbol='RTN3', Aliases='ASYIP, HAP, NSPL2, NSPLII-A1, RTN3', description='reticulon 3', other_designations='ASY interacting protein|NSP-like protein 2|NSP-like protein II|homolog of ASY protein|neuroendocrine-specific protein-like 2|neuroendocrine-specific protein-like II', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63681360, end_position_on_the_genomic_accession=63759891, orientation='plus', exon_count=9, OMIM=604249, no_hdr0=''),
  10314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10314, CurrentID=0, Status='live', Symbol='LANCL1', Aliases='GPR69A, p40', description='LanC lantibiotic synthetase component C-like 1 (bacterial)', other_designations='40 kDa erythrocyte membrane protein|G protein-coupled receptor 69A|LanC (bacterial lantibiotic synthetase component C)-like 1|LanC (bacterial lantibiotic synthetase component)', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210431249, end_position_on_the_genomic_accession=210477618, orientation='minus', exon_count=11, OMIM=604155, no_hdr0=''),
  100533943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533943, CurrentID=0, Status='live', Symbol='LOC100533943', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10316, CurrentID=0, Status='live', Symbol='NMUR1', Aliases='(FM-3), FM-3, FM3, GPC-R, GPR66, NMU1R', description='neuromedin U receptor 1', other_designations='G protein-coupled receptor 66|G-protein coupled receptor 66|G-protein coupled receptor FM-3|NMU-R1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231519907, end_position_on_the_genomic_accession=231530471, orientation='minus', exon_count=6, OMIM=604153, no_hdr0=''),
  10317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10317, CurrentID=0, Status='live', Symbol='B3GALT5', Aliases='B3GalT-V, B3GalTx, B3T5, GLCT5, beta-1,3-GalTase 5, beta-3-Gx-T5, beta3Gal-T5', description='Beta-1,3-galactosyltransferase 5', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase 5|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 5|UDP-galactose:beta-N-acetylglucosamine beta-1,3-galactosyltransferase 5|beta-3-galactosyltransferase 5|homolog of C. elegans Bt toxin resistance gene bre-5', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39642340, end_position_on_the_genomic_accession=39654265, orientation='plus', exon_count=9, OMIM=604066, no_hdr0=''),
  10318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10318, CurrentID=0, Status='live', Symbol='TNIP1', Aliases='ABIN-1, NAF1, VAN, nip40-1', description='TNFAIP3 interacting protein 1', other_designations='A20-binding inhibitor of NF-kappa-B activation 1|HIV-1 Nef-interacting protein|Nef-associated factor 1 SNP|virion-associated nuclear shuttling protein', map_location='5q32-q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151029943, end_position_on_the_genomic_accession=151087660, orientation='minus', exon_count=21, OMIM=607714, no_hdr0=''),
  10319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10319, CurrentID=0, Status='live', Symbol='LAMC3', Aliases='OCCM', description='laminin subunit gamma 3', other_designations='laminin, gamma 3|laminin-12 subunit gamma|laminin-14 subunit gamma|laminin-15 subunit gamma', map_location='9q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131009117, end_position_on_the_genomic_accession=131093059, orientation='plus', exon_count=29, OMIM=604349, no_hdr0=''),
  10320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10320, CurrentID=0, Status='live', Symbol='IKZF1', Aliases='Hs.54452, IK1, IKAROS, LYF1, LyF-1, PPP1R92, PRO0758, ZNFN1A1', description='IKAROS family zinc finger 1', other_designations='CLL-associated antigen KW-6|IKAROS family zinc finger 1 (Ikaros)|ikaros family zinc finger protein 1|lymphoid transcription factor LyF-1|protein phosphatase 1, regulatory subunit 92|zinc finger protein, subfamily 1A, 1 (Ikaros)', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50303465, end_position_on_the_genomic_accession=50405101, orientation='plus', exon_count=15, OMIM=603023, no_hdr0=''),
  10321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10321, CurrentID=0, Status='live', Symbol='CRISP3', Aliases='Aeg2, CRISP-3, CRS3, SGP28, dJ442L6.3', description='cysteine-rich secretory protein 3', other_designations='cysteine-rich secretory protein-3|specific granule protein (28 kDa)|specific granule protein of 28 kDa', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49727376, end_position_on_the_genomic_accession=49744455, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  10322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10322, CurrentID=0, Status='live', Symbol='SMYD5', Aliases='NN8-4AG, RAI15, RRG1, ZMYND23', description='SMYD family member 5', other_designations='retinoic acid induced 15|retinoic acid responsive|retinoic acid-induced protein 15', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73214238, end_position_on_the_genomic_accession=73227227, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  10324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10324, CurrentID=0, Status='live', Symbol='KLHL41', Aliases='KBTBD10, Krp1, SARCOSIN', description='kelch like family member 41', other_designations='kel-like protein 23|kelch repeat and BTB (POZ) domain containing 10|kelch-like family member 41|kelch-related protein 1|sarcomeric muscle protein', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169509702, end_position_on_the_genomic_accession=169526262, orientation='plus', exon_count=6, OMIM=607701, no_hdr0=''),
  10325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10325, CurrentID=0, Status='live', Symbol='RRAGB', Aliases='RAGB, bA465E19.1', description='Ras-related GTP binding B', other_designations='GTP-binding protein ragB|rag B', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55717677, end_position_on_the_genomic_accession=55759333, orientation='plus', exon_count=13, OMIM=300725, no_hdr0=''),
  10326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10326, CurrentID=0, Status='live', Symbol='SIRPB1', Aliases='CD172b, SIRP-BETA-1', description='signal-regulatory protein beta 1', other_designations='CD172 antigen-like family member B', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1564383, end_position_on_the_genomic_accession=1626497, orientation='minus', exon_count=16, OMIM=603889, no_hdr0=''),
  10327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10327, CurrentID=0, Status='live', Symbol='AKR1A1', Aliases='ALDR1, ALR, ARM, DD3, HEL-S-6', description='aldo-keto reductase family 1, member A1 (aldehyde reductase)', other_designations='HEL-S-165mP|alcohol dehydrogenase|aldehyde reductase|aldo-keto reductase family 1 member A1|dihydrodiol dehydrogenase 3|epididymis secretory protein Li 6|epididymis secretory sperm binding protein Li 165mP', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45550783, end_position_on_the_genomic_accession=45570051, orientation='plus', exon_count=10, OMIM=103830, no_hdr0=''),
  10328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10328, CurrentID=0, Status='live', Symbol='EMC8', Aliases='C16orf2, C16orf4, COX4NB, FAM158B, NOC4', description='ER membrane protein complex subunit 8', other_designations='COX4 neighbor|family with sequence similarity 158, member B|neighbor of COX4', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85778624, end_position_on_the_genomic_accession=85799555, orientation='minus', exon_count=7, OMIM=604886, no_hdr0=''),
  10329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10329, CurrentID=0, Status='live', Symbol='TMEM5', Aliases='HP10481, MDDGA10', description='transmembrane protein 5', other_designations='type II membrane protein', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63779803, end_position_on_the_genomic_accession=63809558, orientation='plus', exon_count=8, OMIM=605862, no_hdr0=''),
  10330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10330, CurrentID=0, Status='live', Symbol='CNPY2', Aliases='HP10390, MSAP, TMEM4, ZSIG9', description='canopy FGF signaling regulator 2', other_designations='MIR-interacting saposin-like protein|canopy 2 homolog|putative secreted protein Zsig9|transmembrane protein 4', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56310428, end_position_on_the_genomic_accession=56316344, orientation='minus', exon_count=6, OMIM=605861, no_hdr0=''),
  10331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10331, CurrentID=0, Status='live', Symbol='B3GNT3', Aliases='B3GAL-T8, B3GN-T3, B3GNT-3, HP10328, TMEM3, beta3Gn-T3', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 3', other_designations='UDP-Gal:beta-GlcNAc beta-1,3-galactosyltransferase 8|UDP-galactose:beta-N-acetylglucosamine beta-1,3-galactosyltransferase 8|beta-1,3-GalTase 8|beta-1,3-Gn-T3|beta-1,3-N-acetylglucosaminyltransferase bGnT-3|beta-1,3-galactosyl-O-glycosyl-glycoprotein beta-1,3-N-acetylglucosaminyltransferase|beta-3-Gx-T8|beta3GalT8|core 1 extending beta-1,3-N-acetylglucosaminyltransferase|core1-beta3GlcNAcT|putative type II membrane protein|transmembrane protein 3', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17794828, end_position_on_the_genomic_accession=17813576, orientation='plus', exon_count=5, OMIM=605863, no_hdr0=''),
  10332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10332, CurrentID=0, Status='live', Symbol='CLEC4M', Aliases='CD209L, CD299, DC-SIGN2, DC-SIGNR, DCSIGNR, HP10347, L-SIGN, LSIGN', description='C-type lectin domain family 4 member M', other_designations='C-type lectin domain family 4, member M|CD209 antigen-like protein 1|CD299 antigen|DC-SIGN-related protein|dendritic cell-specific ICAM-3-grabbing non-integrin 2|liver/lymph node-specific ICAM-3 grabbing non-integrin|liver/lymph node-specific ICAM-3-grabbing non-integrin|mannose binding C-type lectin DC-SIGNR', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7763054, end_position_on_the_genomic_accession=7770558, orientation='plus', exon_count=9, OMIM=605872, no_hdr0=''),
  10333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10333, CurrentID=0, Status='live', Symbol='TLR6', Aliases='CD286', description='toll-like receptor 6', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38823704, end_position_on_the_genomic_accession=38857711, orientation='minus', exon_count=6, OMIM=605403, no_hdr0=''),
  10335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10335, CurrentID=0, Status='live', Symbol='MRVI1', Aliases='IRAG, JAW1L', description='murine retrovirus integration site 1 homolog', other_designations='IP3R-associated cGMP kinase substrate|JAW1-related protein MRVI1|inositol 1,4,5-triphosphate receptor-associated cGMP kinase substrate|inositol 1,4,5-trisphosphate receptor-associated cGMP kinase substrate', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10573091, end_position_on_the_genomic_accession=10693988, orientation='minus', exon_count=22, OMIM=604673, no_hdr0=''),
  10336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10336, CurrentID=0, Status='live', Symbol='PCGF3', Aliases='DONG1, RNF3, RNF3A', description='polycomb group ring finger 3', other_designations='RING finger protein 3A|ring finger protein 3', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=705748, end_position_on_the_genomic_accession=770640, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  10337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10337, CurrentID=0, Status='live', Symbol='HIST1H4PS1', Aliases='H4FFP, dJ97D16.3, pH4/f', description='histone cluster 1, H4, pseudogene 1', other_designations='H4 histone family, member F, pseudogene|histone 1, H4, pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27807167, end_position_on_the_genomic_accession=27807337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10338, CurrentID=0, Status='live', Symbol='HIST1H1PS2', Aliases='H3FEP, pH3/e', description='histone cluster 1, H1, pseudogene 2', other_designations='H3 histone family, member E, pseudogene|histone 1, H1, pseudogene 2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26016107, end_position_on_the_genomic_accession=26016841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377322, CurrentID=0, Status='live', Symbol='LOC105377322', Aliases='', description='uncharacterized LOC105377322', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87520253, end_position_on_the_genomic_accession=87527531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10340, CurrentID=0, Status='live', Symbol='HIST1H2BPS2', Aliases='H2BFIP, pH2B.i, pH2B/i', description='histone cluster 1, H2b, pseudogene 2', other_designations='H2B histone family, member I, pseudogene|histone 1, H2b, pseudogene 2', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27864062, end_position_on_the_genomic_accession=27864401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10341, CurrentID=0, Status='live', Symbol='HIST1H2APS5', Aliases='H2AFFP, pH2A/f', description='histone cluster 1, H2a, pseudogene 5', other_designations='H2A histone family, member F, pseudogene|histone 1, H2a, pseudogene 5', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26043900, end_position_on_the_genomic_accession=26044550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10342, CurrentID=0, Status='live', Symbol='TFG', Aliases='HMSNP, SPG57, TF6, TRKT3', description='TRK-fused gene', other_designations='TRK-fused gene protein|TRKT3 oncogene', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100709284, end_position_on_the_genomic_accession=100748967, orientation='plus', exon_count=10, OMIM=602498, no_hdr0=''),
  10343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10343, CurrentID=0, Status='live', Symbol='PKDREJ', Aliases='', description='polycystin (PKD) family receptor for egg jelly', other_designations='PKD and REJ homolog|polycystic kidney disease (polycystin) and REJ homolog (sperm receptor for egg jelly homolog, sea urchin)', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46255663, end_position_on_the_genomic_accession=46263322, orientation='minus', exon_count=1, OMIM=604670, no_hdr0=''),
  10344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10344, CurrentID=0, Status='live', Symbol='CCL26', Aliases='IMAC, MIP-4a, MIP-4alpha, SCYA26, TSC-1', description='chemokine (C-C motif) ligand 26', other_designations='CC chemokine IMAC|MIP-4-alpha|chemokine N1|eotaxin-3|macrophage inflammatory protein 4-alpha|small inducible cytokine A26|small inducible cytokine subfamily A (Cys-Cys), member 26|small-inducible cytokine A26|thymic stroma chemokine-1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75769524, end_position_on_the_genomic_accession=75789746, orientation='minus', exon_count=4, OMIM=604697, no_hdr0=''),
  10345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10345, CurrentID=0, Status='live', Symbol='TRDN', Aliases='CPVT5, TDN, TRISK', description='triadin', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123216339, end_position_on_the_genomic_accession=123637283, orientation='minus', exon_count=42, OMIM=603283, no_hdr0=''),
  10346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10346, CurrentID=0, Status='live', Symbol='TRIM22', Aliases='GPSTAF50, RNF94, STAF50', description='tripartite motif containing 22', other_designations='50 kDa-stimulated trans-acting factor|RING finger protein 94|staf-50|stimulated trans-acting factor (50 kDa)|tripartite binding motif 22|tripartite motif protein TRIM22|tripartite motif-containing 22|tripartite motif-containing protein 22', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5689587, end_position_on_the_genomic_accession=5710863, orientation='plus', exon_count=8, OMIM=606559, no_hdr0=''),
  10347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10347, CurrentID=0, Status='live', Symbol='ABCA7', Aliases='ABCA-SSN, ABCX', description='ATP binding cassette subfamily A member 7', other_designations='ATP-binding cassette, sub-family A (ABC1), member 7|autoantigen SS-N|macrophage ABC transporter', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1040103, end_position_on_the_genomic_accession=1065572, orientation='plus', exon_count=47, OMIM=605414, no_hdr0=''),
  10349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10349, CurrentID=0, Status='live', Symbol='ABCA10', Aliases='EST698739', description='ATP binding cassette subfamily A member 10', other_designations='ATP-binding cassette A10|ATP-binding cassette, sub-family A (ABC1), member 10', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69148007, end_position_on_the_genomic_accession=69244815, orientation='minus', exon_count=40, OMIM=612508, no_hdr0=''),
  10350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10350, CurrentID=0, Status='live', Symbol='ABCA9', Aliases='EST640918', description='ATP binding cassette subfamily A member 9', other_designations='ATP-binding cassette A9|ATP-binding cassette, sub-family A (ABC1), member 9', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68974632, end_position_on_the_genomic_accession=69067443, orientation='minus', exon_count=42, OMIM=612507, no_hdr0=''),
  10351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10351, CurrentID=0, Status='live', Symbol='ABCA8', Aliases='', description='ATP binding cassette subfamily A member 8', other_designations='ATP-binding cassette, sub-family A (ABC1), member 8', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68867287, end_position_on_the_genomic_accession=68955408, orientation='minus', exon_count=43, OMIM=612505, no_hdr0=''),
  10352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10352, CurrentID=0, Status='live', Symbol='WARS2', Aliases='TrpRS', description='tryptophanyl tRNA synthetase 2, mitochondrial', other_designations='(Mt)TrpRS|tryptophan tRNA ligase 2, mitochondrial|tryptophan-tRNA ligase|tryptophanyl-tRNA synthetase, mitochondrial', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119031216, end_position_on_the_genomic_accession=119140788, orientation='minus', exon_count=8, OMIM=604733, no_hdr0=''),
  401602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401602, CurrentID=0, Status='live', Symbol='LOC401602', Aliases='', description='adaptor-related protein complex 2, beta 1 subunit pseudogene', other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93222221, end_position_on_the_genomic_accession=93225015, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10354, CurrentID=0, Status='live', Symbol='HMGB1P5', Aliases='HMG-1, HMG1L5, HMGB1L15, HMGB1L5, HMGB1P2', description='high mobility group box 1 pseudogene 5', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 5 pseudogene|high-mobility group box 1 pseudogene 2|high-mobility group box 1 pseudogene 5|high-mobility group box 1-like 15|high-mobility group box 1-like 5 pseudogene', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=22381670, end_position_on_the_genomic_accession=22382866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10355, CurrentID=0, Status='live', Symbol='HMGB1P4', Aliases='HMG1L4, HMG1L4P, HMGB1L4', description='high mobility group box 1 pseudogene 4', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 4 pseudogene|high-mobility group box 1 pseudogene 4|high-mobility group box 1-like 4', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170600138, end_position_on_the_genomic_accession=170602384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10356, CurrentID=0, Status='live', Symbol='HMGB1P3', Aliases='HMG1L3, HMGB1L3', description='high mobility group box 1 pseudogene 3', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 3|high-mobility group box 1 pseudogene 3|high-mobility group box 1-like 3', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230514861, end_position_on_the_genomic_accession=230516201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10357, CurrentID=0, Status='live', Symbol='HMGB1P1', Aliases='HMG1L1, HMG1L7, HMGB1L1, dJ579F20.1', description='high mobility group box 1 pseudogene 1', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 1|high-mobility group box 1 pseudogene 1|high-mobility group box 1-like 1', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57487941, end_position_on_the_genomic_accession=57489130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10360, CurrentID=0, Status='live', Symbol='NPM3', Aliases='PORMIN, TMEM123', description='nucleophosmin/nucleoplasmin 3', other_designations='nucleophosmin/nucleoplasmin family, member 3', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101781325, end_position_on_the_genomic_accession=101783413, orientation='minus', exon_count=6, OMIM=606456, no_hdr0=''),
  10361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10361, CurrentID=0, Status='live', Symbol='NPM2', Aliases='', description='nucleophosmin/nucleoplasmin 2', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22023708, end_position_on_the_genomic_accession=22036897, orientation='plus', exon_count=11, OMIM=608073, no_hdr0=''),
  10362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10362, CurrentID=0, Status='live', Symbol='HMG20B', Aliases='BRAF25, BRAF35, HMGX2, HMGXB2, PP7706, SMARCE1r, SOXL, pp8857', description='high mobility group 20B', other_designations='BRCA2-associated factor 35|HMG box domain containing 2|HMG box-containing protein 20B|HMG domain-containing protein 2|HMG domain-containing protein HMGX2|SMARCE1-related protein|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily E, member 1-related|Sox-like transcriptional factor|high-mobility group 20B|structural DNA-binding protein BRAF35', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3572945, end_position_on_the_genomic_accession=3579083, orientation='plus', exon_count=10, OMIM=605535, no_hdr0=''),
  10363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10363, CurrentID=0, Status='live', Symbol='HMG20A', Aliases='HMGX1, HMGXB1', description='high mobility group 20A', other_designations='HMG box domain containing 1|HMG box-containing protein 20A|HMG domain-containing protein 1|HMG domain-containing protein HMGX1', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77420651, end_position_on_the_genomic_accession=77485603, orientation='plus', exon_count=11, OMIM=605534, no_hdr0=''),
  10365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10365, CurrentID=0, Status='live', Symbol='KLF2', Aliases='LKLF', description='Kruppel-like factor 2', other_designations='Kruppel-like factor 2 (lung)|Kruppel-like factor LKLF|lung Kruppel-like zinc finger transcription factor|lung krueppel-like factor', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16324840, end_position_on_the_genomic_accession=16327534, orientation='plus', exon_count=3, OMIM=602016, no_hdr0=''),
  10367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10367, CurrentID=0, Status='live', Symbol='MICU1', Aliases='CALC, CBARA1, EFHA3, MPXPS', description='mitochondrial calcium uptake 1', other_designations='ara CALC|atopy-related autoantigen CALC|calcium-binding atopy-related autoantigen 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72367326, end_position_on_the_genomic_accession=72626191, orientation='minus', exon_count=17, OMIM=605084, no_hdr0=''),
  10368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10368, CurrentID=0, Status='live', Symbol='CACNG3', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 3', other_designations='TARP gamma-3|neuronal voltage-gated calcium channel gamma-3 subunit|transmembrane AMPAR regulatory protein gamma-3|voltage-gated calcium channel gamma subunit', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24255553, end_position_on_the_genomic_accession=24362416, orientation='plus', exon_count=4, OMIM=606403, no_hdr0=''),
  10369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10369, CurrentID=0, Status='live', Symbol='CACNG2', Aliases='MRD10', description='calcium channel, voltage-dependent, gamma subunit 2', other_designations='TARP gamma-2|neuronal voltage-gated calcium channel gamma-2 subunit|stargazin|transmembrane AMPAR regulatory protein gamma-2', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36560869, end_position_on_the_genomic_accession=36702645, orientation='minus', exon_count=4, OMIM=602911, no_hdr0=''),
  10370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10370, CurrentID=0, Status='live', Symbol='CITED2', Aliases='ASD8, MRG-1, MRG1, P35SRJ, VSD2', description='Cbp/p300-interacting transactivator, with Glu/Asp rich carboxy-terminal domain, 2', other_designations='Cbp/p300-interacting transactivator, with Glu/Asp-rich carboxy-terminal domain, 2|MSG-related protein 1|MSG1-related gene 1|melanocyte-specific gene 1-related gene 1', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139372255, end_position_on_the_genomic_accession=139374650, orientation='minus', exon_count=4, OMIM=602937, no_hdr0=''),
  10371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10371, CurrentID=0, Status='live', Symbol='SEMA3A', Aliases='COLL1, HH16, Hsema-I, Hsema-III, SEMA1, SEMAD, SEMAIII, SEMAL, SemD, coll-1', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3A', other_designations='collapsin 1|semaphorin D|semaphorin III|semaphorin L', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83956805, end_position_on_the_genomic_accession=84492724, orientation='minus', exon_count=24, OMIM=603961, no_hdr0=''),
  10376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10376, CurrentID=0, Status='live', Symbol='TUBA1B', Aliases='K-ALPHA-1', description='tubulin alpha 1b', other_designations='alpha tubulin|alpha-tubulin ubiquitous|tubulin K-alpha-1|tubulin alpha-ubiquitous chain|tubulin, alpha 1b|tubulin, alpha, ubiquitous', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49127782, end_position_on_the_genomic_accession=49131521, orientation='minus', exon_count=4, OMIM=602530, no_hdr0=''),
  10379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10379, CurrentID=0, Status='live', Symbol='IRF9', Aliases='IRF-9, ISGF3, ISGF3G, p48', description='interferon regulatory factor 9', other_designations='IFN-alpha-responsive transcription factor subunit|ISGF-3 gamma|ISGF3 p48 subunit|interferon-stimulated gene factor 3 gamma|interferon-stimulated transcription factor 3, gamma (48kD)|interferon-stimulated transcription factor 3, gamma 48kDa|transcriptional regulator ISGF3 subunit gamma', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24161213, end_position_on_the_genomic_accession=24166565, orientation='plus', exon_count=9, OMIM=147574, no_hdr0=''),
  10380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10380, CurrentID=0, Status='live', Symbol='BPNT1', Aliases='HEL20, PIP', description="3'(2'), 5'-bisphosphate nucleotidase 1", other_designations="BPntase|PAP-inositol 1,4-phosphatase|PAP-inositol-1,4-phosphatase|bisphosphate 3'-nucleotidase 1|epididymis luminal protein 20", map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220057482, end_position_on_the_genomic_accession=220089927, orientation='minus', exon_count=11, OMIM=604053, no_hdr0=''),
  10381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10381, CurrentID=0, Status='live', Symbol='TUBB3', Aliases='CDCBM, CDCBM1, CFEOM3, CFEOM3A, FEOM3, TUBB4, beta-4', description='tubulin beta 3 class III', other_designations='class III beta-tubulin|tubulin beta-4 chain|tubulin beta-III|tubulin, beta 3 class III', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89922009, end_position_on_the_genomic_accession=89936097, orientation='plus', exon_count=5, OMIM=602661, no_hdr0=''),
  10382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10382, CurrentID=0, Status='live', Symbol='TUBB4A', Aliases='DYT4, TUBB4, beta-5', description='tubulin beta 4A class IVa', other_designations='dystonia 4, torsion (autosomal dominant)|tubulin beta-4 chain|tubulin, beta 4|tubulin, beta 4A class IVa|tubulin, beta, 5', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6494319, end_position_on_the_genomic_accession=6502848, orientation='minus', exon_count=5, OMIM=602662, no_hdr0=''),
  10383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10383, CurrentID=0, Status='live', Symbol='TUBB4B', Aliases='Beta2, TUBB2, TUBB2C', description='tubulin beta 4B class IVb', other_designations='class IVb beta tubulin|class IVb beta-tubulin|tubulin beta-2 chain|tubulin beta-2C chain|tubulin, beta 2C|tubulin, beta 4B class IVb|tubulin, beta, 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137241259, end_position_on_the_genomic_accession=137243707, orientation='plus', exon_count=4, OMIM=602660, no_hdr0=''),
  10384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10384, CurrentID=0, Status='live', Symbol='BTN3A3', Aliases='BTF3, BTN3.3', description='butyrophilin subfamily 3 member A3', other_designations='butyrophilin 3|butyrophilin subfamily 3 member A3 secreted isoform|butyrophilin, subfamily 3, member A3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26440472, end_position_on_the_genomic_accession=26453415, orientation='plus', exon_count=11, OMIM=613595, no_hdr0=''),
  10385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10385, CurrentID=0, Status='live', Symbol='BTN2A2', Aliases='BT2.2, BTF2, BTN2.2', description='butyrophilin subfamily 2 member A2', other_designations='butyrophilin 2|butyrophilin, subfamily 2, member A2', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26382886, end_position_on_the_genomic_accession=26394874, orientation='plus', exon_count=8, OMIM=613591, no_hdr0=''),
  10386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10386, CurrentID=0, Status='live', Symbol='MHW1', Aliases='', description='Mental health wellness 1', other_designations='', map_location='4p', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603663, no_hdr0=''),
  100009667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009667, CurrentID=0, Status='live', Symbol='POU5F1P5', Aliases='Oct4-pg5', description='POU class 5 homeobox 1 pseudogene 5', other_designations='POU domain class 5, transcription factor 1 pseudogene 5', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68009947, end_position_on_the_genomic_accession=68010877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10388, CurrentID=0, Status='live', Symbol='SYCP2', Aliases='SCP-2, SCP2', description='synaptonemal complex protein 2', other_designations='synaptonemal complex lateral element protein', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59863557, end_position_on_the_genomic_accession=59933681, orientation='minus', exon_count=48, OMIM=604105, no_hdr0=''),
  10389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10389, CurrentID=0, Status='live', Symbol='SCML2', Aliases='', description='sex comb on midleg-like 2 (Drosophila)', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18239313, end_position_on_the_genomic_accession=18354724, orientation='minus', exon_count=17, OMIM=300208, no_hdr0=''),
  10390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10390, CurrentID=0, Status='live', Symbol='CEPT1', Aliases='', description='choline/ethanolamine phosphotransferase 1', other_designations='choline/ethanolaminephosphotransferase|hCEPT1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111139627, end_position_on_the_genomic_accession=111185102, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  10391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10391, CurrentID=0, Status='live', Symbol='CORO2B', Aliases='CLIPINC', description='coronin, actin binding protein, 2B', other_designations='clipin C|coronin-like protein C', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68559275, end_position_on_the_genomic_accession=68727806, orientation='plus', exon_count=17, OMIM=605002, no_hdr0=''),
  10392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10392, CurrentID=0, Status='live', Symbol='NOD1', Aliases='CARD4, CLR7.1, NLRC1', description='nucleotide binding oligomerization domain containing 1', other_designations='NLR family, CARD domain containing 1|caspase recruitment domain family, member 4|caspase recruitment domain-containing protein 4|nucleotide-binding oligomerization domain containing 1|nucleotide-binding oligomerization domain, leucine rich repeat and CARD domain containing 1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30424527, end_position_on_the_genomic_accession=30478797, orientation='minus', exon_count=19, OMIM=605980, no_hdr0=''),
  10393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10393, CurrentID=0, Status='live', Symbol='ANAPC10', Aliases='APC10, DOC1', description='anaphase promoting complex subunit 10', other_designations='cyclosome subunit 10', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144967117, end_position_on_the_genomic_accession=145098541, orientation='minus', exon_count=15, OMIM=613745, no_hdr0=''),
  10394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10394, CurrentID=0, Status='live', Symbol='PRG3', Aliases='MBP2, MBPH', description='proteoglycan 3', other_designations='eosinophil major basic protein homolog|prepro-MBPH|prepro-major basic protein homolog', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57376769, end_position_on_the_genomic_accession=57381150, orientation='minus', exon_count=6, OMIM=606814, no_hdr0=''),
  10395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10395, CurrentID=0, Status='live', Symbol='DLC1', Aliases='ARHGAP7, HP, STARD12, p122-RhoGAP', description='DLC1 Rho GTPase activating protein', other_designations='Rho-GTPase-activating protein 7|START domain-containing protein 12|StAR-related lipid transfer (START) domain containing 12|deleted in liver cancer 1 protein|deleted in liver cancer 1 variant 2|rho-type GTPase-activating protein 7', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13083361, end_position_on_the_genomic_accession=13515658, orientation='minus', exon_count=22, OMIM=604258, no_hdr0=''),
  10396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10396, CurrentID=0, Status='live', Symbol='ATP8A1', Aliases='ATPASEII, ATPIA, ATPP2', description='ATPase, aminophospholipid transporter (APLT), class I, type 8A, member 1', other_designations='ATPase II|ATPase class I type 8A member 1|P4-ATPase flippase complex alpha subunit ATP8A1|aminophospholipid translocase|chromaffin granule ATPase II|probable phospholipid-transporting ATPase IA', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42408373, end_position_on_the_genomic_accession=42657105, orientation='minus', exon_count=39, OMIM=609542, no_hdr0=''),
  10397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10397, CurrentID=0, Status='live', Symbol='NDRG1', Aliases='CAP43, CMT4D, DRG-1, DRG1, GC4, HMSNL, NDR1, NMSL, PROXY1, RIT42, RTP, TARG1, TDD5', description='N-myc downstream regulated 1', other_designations='N-myc downstream-regulated gene 1 protein|differentiation-related gene 1 protein|nickel-specific induction protein Cap43|protein regulated by oxygen-1|reducing agents and tunicamycin-responsive protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133237171, end_position_on_the_genomic_accession=133297587, orientation='minus', exon_count=17, OMIM=605262, no_hdr0=''),
  10398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10398, CurrentID=0, Status='live', Symbol='MYL9', Aliases='LC20, MLC-2C, MLC2, MRLC1, MYRL2', description='myosin light chain 9', other_designations='20 kDa myosin light chain|myosin RLC|myosin regulatory light chain 1|myosin regulatory light chain 2, smooth muscle isoform|myosin regulatory light chain 9|myosin regulatory light chain MRLC1|myosin, light chain 9, regulatory|myosin, light polypeptide 9, regulatory', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36541484, end_position_on_the_genomic_accession=36549823, orientation='plus', exon_count=4, OMIM=609905, no_hdr0=''),
  10399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10399, CurrentID=0, Status='live', Symbol='GNB2L1', Aliases='Gnb2-rs1, H12.3, HLC-7, PIG21, RACK1', description='guanine nucleotide binding protein (G protein), beta polypeptide 2-like 1', other_designations='cell proliferation-inducing gene 21 protein|guanine nucleotide binding protein beta polypeptide 2-like 1|guanine nucleotide-binding protein subunit beta-like protein 12.3|human lung cancer oncogene 7 protein|lung cancer oncogene 7|proliferation-inducing gene 21|protein homologous to chicken B complex protein, guanine nucleotide binding|receptor for activated C kinase 1|receptor of activated protein kinase C 1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181236928, end_position_on_the_genomic_accession=181243906, orientation='minus', exon_count=8, OMIM=176981, no_hdr0=''),
  10400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10400, CurrentID=0, Status='live', Symbol='PEMT', Aliases='PEAMT, PEMPT2, PNMT, PEMT', description='phosphatidylethanolamine N-methyltransferase', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17505563, end_position_on_the_genomic_accession=17591703, orientation='minus', exon_count=11, OMIM=602391, no_hdr0=''),
  10401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10401, CurrentID=0, Status='live', Symbol='PIAS3', Aliases='ZMIZ5', description='protein inhibitor of activated STAT 3', other_designations='protein inhibitor of activated STAT protein 3|protein inhibitor of activated STAT, 3|zinc finger, MIZ-type containing 5', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145848522, end_position_on_the_genomic_accession=145859081, orientation='minus', exon_count=14, OMIM=605987, no_hdr0=''),
  10402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10402, CurrentID=0, Status='live', Symbol='ST3GAL6', Aliases='SIAT10, ST3GALVI', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 6', other_designations='CMP-NeuAc:beta-galactoside alpha-2,3-sialyltransferase VI|alpha2,3-sialyltransferase ST3Gal VI|sialyltransferase 10 (alpha-2,3-sialyltransferase VI)', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98732236, end_position_on_the_genomic_accession=98797054, orientation='plus', exon_count=17, OMIM=607156, no_hdr0=''),
  10403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10403, CurrentID=0, Status='live', Symbol='NDC80', Aliases='HEC, HEC1, HsHec1, KNTC2, TID3, hsNDC80', description='NDC80 kinetochore complex component', other_designations='NDC80 homolog, kinetochore complex component|NDC80 kinetochore complex component homolog|highly expressed in cancer protein|highly expressed in cancer, rich in leucine heptad repeats|kinetochore associated 2|kinetochore protein Hec1|kinetochore-associated protein 2|retinoblastoma-associated protein HEC', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2571511, end_position_on_the_genomic_accession=2616635, orientation='plus', exon_count=17, OMIM=607272, no_hdr0=''),
  10404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10404, CurrentID=0, Status='live', Symbol='CPQ', Aliases='LDP, PGCP', description='carboxypeptidase Q', other_designations='Ser-Met dipeptidase|aminopeptidase|blood plasma glutamate carboxypeptidase|lysosomal dipeptidase', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96645227, end_position_on_the_genomic_accession=97143503, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  10406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10406, CurrentID=0, Status='live', Symbol='WFDC2', Aliases='EDDM4, HE4, WAP5, dJ461P17.6', description='WAP four-disulfide core domain 2', other_designations='WAP domain containing protein HE4-V4|epididymal protein 4|epididymal secretory protein E4|epididymis-specific, whey-acidic protein type, four-disulfide core|major epididymis-specific protein E4|putative protease inhibitor WAP5', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45448648, end_position_on_the_genomic_accession=45481532, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  10407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10407, CurrentID=0, Status='live', Symbol='SPAG11B', Aliases='EDDM2B, EP2, EP2C, EP2D, HE2, HE2C, SPAG11', description='sperm associated antigen 11B', other_designations='epididymal protein 2B|human epididymis-specific protein 2|sperm antigen HE2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7447754, end_position_on_the_genomic_accession=7463670, orientation='minus', exon_count=7, OMIM=606560, no_hdr0=''),
  10408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10408, CurrentID=0, Status='live', Symbol='MYCNOS', Aliases='MYCN-AS1, N-CYM, NCYM, NYCM', description='MYCN opposite strand', other_designations='MYCN antisense RNA 1|MYCN opposite strand/antisense RNA (non-protein coding)|v-myc myelocytomatosis viral related oncogene, neuroblastoma derived opposite strand', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15936265, end_position_on_the_genomic_accession=15941723, orientation='minus', exon_count=5, OMIM=605374, no_hdr0=''),
  10409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10409, CurrentID=0, Status='live', Symbol='BASP1', Aliases='CAP-23, CAP23, NAP-22, NAP22', description='brain abundant, membrane attached signal protein 1', other_designations='22 kDa neuronal tissue-enriched acidic protein|brain acid-soluble protein 1|neuronal axonal membrane protein NAP-22|neuronal tissue-enriched acidic protein', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17216823, end_position_on_the_genomic_accession=17276845, orientation='plus', exon_count=3, OMIM=605940, no_hdr0=''),
  10410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10410, CurrentID=0, Status='live', Symbol='IFITM3', Aliases='1-8U, DSPA2b, IP15', description='interferon induced transmembrane protein 3', other_designations='dispanin subfamily A member 2b|interferon-inducible protein 1-8U', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=319673, end_position_on_the_genomic_accession=320914, orientation='minus', exon_count=2, OMIM=605579, no_hdr0=''),
  10411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10411, CurrentID=0, Status='live', Symbol='RAPGEF3', Aliases='CAMP-GEFI, EPAC, EPAC1, HSU79275, bcm910', description='Rap guanine nucleotide exchange factor 3', other_designations='9330170P05Rik|EPAC 1|Rap guanine nucleotide exchange factor (GEF) 3|Rap1 guanine-nucleotide-exchange factor directly activated by cAMP|cAMP-regulated guanine nucleotide exchange factor I|exchange factor directly activated by cAMP 1|exchange protein directly activated by cAMP 1', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47734670, end_position_on_the_genomic_accession=47771035, orientation='minus', exon_count=31, OMIM=606057, no_hdr0=''),
  10412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10412, CurrentID=0, Status='live', Symbol='NSA2', Aliases='CDK105, HCL-G1, HCLG1, HUSSY-29, HUSSY29, TINP1', description='NSA2, ribosome biogenesis homolog', other_designations='NSA2 ribosome biogenesis homolog|TGF beta-inducible nuclear protein 1|TGF-beta inducible nuclear protein|TGF-beta-inducible nuclear protein 1|hairy cell leukemia protein 1', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74765429, end_position_on_the_genomic_accession=74776912, orientation='plus', exon_count=8, OMIM=612497, no_hdr0=''),
  10413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10413, CurrentID=0, Status='live', Symbol='YAP1', Aliases='COB1, YAP, YAP2, YAP65, YKI', description='Yes associated protein 1', other_designations='65 kDa Yes-associated protein|Yes-associated protein 1|protein yorkie homolog|yes-associated protein 2|yes-associated protein YAP65 homolog|yorkie homolog', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102110254, end_position_on_the_genomic_accession=102233423, orientation='plus', exon_count=10, OMIM=606608, no_hdr0=''),
  10415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10415, CurrentID=0, Status='live', Symbol='SNAI1P1', Aliases='SNAI1L, SNAI1L1, SNAI1P', description='snail family zinc finger 1 pseudogene 1', other_designations='snail 1, zinc finger protein, pseudogene|snail homolog 1 like 1|snail homolog 1 pseudogene 1|snail homolog 1-like 1', map_location='2q33-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209808722, end_position_on_the_genomic_accession=209810424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646014, CurrentID=0, Status='live', Symbol='LOC646014', Aliases='', description='uncharacterized LOC646014', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=60194372, end_position_on_the_genomic_accession=60195085, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10417, CurrentID=0, Status='live', Symbol='SPON2', Aliases='DIL-1, DIL1, M-SPONDIN, MINDIN', description='spondin 2, extracellular matrix protein', other_designations='differentially expressed in cancerous and non-cancerous lung cells 1', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1166932, end_position_on_the_genomic_accession=1208962, orientation='minus', exon_count=9, OMIM=605918, no_hdr0=''),
  10418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10418, CurrentID=0, Status='live', Symbol='SPON1', Aliases='VSGP/F-spondin, f-spondin', description='spondin 1, extracellular matrix protein', other_designations='spondin 1, (f-spondin) extracellular matrix protein|vascular smooth muscle cell growth-promoting factor', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13962637, end_position_on_the_genomic_accession=14268133, orientation='plus', exon_count=16, OMIM=604989, no_hdr0=''),
  10419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10419, CurrentID=0, Status='live', Symbol='PRMT5', Aliases='HRMT1L5, IBP72, JBP1, SKB1, SKB1Hs', description='protein arginine methyltransferase 5', other_designations='72 kDa ICln-binding protein|HMT1 hnRNP methyltransferase-like 5|SKB1 homolog|histone-arginine N-methyltransferase PRMT5|jak-binding protein 1|shk1 kinase-binding protein 1 homolog', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22920511, end_position_on_the_genomic_accession=22929585, orientation='minus', exon_count=17, OMIM=604045, no_hdr0=''),
  10420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10420, CurrentID=0, Status='live', Symbol='TESK2', Aliases='', description='testis-specific kinase 2', other_designations='testicular protein kinase 2', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45343883, end_position_on_the_genomic_accession=45491342, orientation='minus', exon_count=14, OMIM=604746, no_hdr0=''),
  10421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10421, CurrentID=0, Status='live', Symbol='CD2BP2', Aliases='FWP010, LIN1, PPP1R59, Snu40, U5-52K', description='CD2 (cytoplasmic tail) binding protein 2', other_designations='CD2 antigen (cytoplasmic tail) binding protein 2|CD2 cytoplasmic domain-binding protein 2|U5 snRNP 52K protein|protein phosphatase 1, regulatory subunit 59', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30350766, end_position_on_the_genomic_accession=30355361, orientation='minus', exon_count=7, OMIM=604470, no_hdr0=''),
  10422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10422, CurrentID=0, Status='live', Symbol='UBAC1', Aliases='GBDR1, UBADC1', description='UBA domain containing 1', other_designations='E3 ubiquitin-protein ligase subunit KPC2|UBA domain-containing protein 1|glialblastoma cell differentiation-related protein 1|kip1 ubiquitination-promoting complex protein 2|putative glialblastoma cell differentiation-related protein|ubiquitin associated domain containing 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135932969, end_position_on_the_genomic_accession=135961380, orientation='minus', exon_count=10, OMIM=608129, no_hdr0=''),
  10423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10423, CurrentID=0, Status='live', Symbol='CDIPT', Aliases='PIS, PIS1', description='CDP-diacylglycerol--inositol 3-phosphatidyltransferase', other_designations='PI synthase|PtdIns synthase|phosphatidylinositol synthase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29858356, end_position_on_the_genomic_accession=29863288, orientation='minus', exon_count=6, OMIM=605893, no_hdr0=''),
  10424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10424, CurrentID=0, Status='live', Symbol='PGRMC2', Aliases='DG6, PMBP', description='progesterone receptor membrane component 2', other_designations='progesterone membrane binding protein|steroid receptor protein DG6', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128269237, end_position_on_the_genomic_accession=128288829, orientation='minus', exon_count=4, OMIM=607735, no_hdr0=''),
  10425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10425, CurrentID=0, Status='live', Symbol='ARIH2', Aliases='ARI2, TRIAD1', description='ariadne RBR E3 ubiquitin protein ligase 2', other_designations='all-trans retinoic acid inducible RING finger|ariadne homolog 2|protein ariadne-2 homolog', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48918821, end_position_on_the_genomic_accession=48985541, orientation='plus', exon_count=21, OMIM=605615, no_hdr0=''),
  10426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10426, CurrentID=0, Status='live', Symbol='TUBGCP3', Aliases='104p, GCP3, Grip104, SPBC98, Spc98, Spc98p', description='tubulin gamma complex associated protein 3', other_designations='gamma-ring complex protein 104 kDa|spindle pole body protein Spc98 homolog|tubulin, gamma complex associated protein 3', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112485005, end_position_on_the_genomic_accession=112606442, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  10427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10427, CurrentID=0, Status='live', Symbol='SEC24B', Aliases='SEC24', description='SEC24 homolog B, COPII coat complex component', other_designations='SEC24 related gene family, member B|secretory protein 24', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109433772, end_position_on_the_genomic_accession=109540459, orientation='plus', exon_count=28, OMIM=607184, no_hdr0=''),
  10428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10428, CurrentID=0, Status='live', Symbol='CFDP1', Aliases='BCNT, BUCENTAUR, CENP-29, CP27, SWC5, Yeti, p97', description='craniofacial development protein 1', other_designations='centromere protein 29|phosphoprotein (Bucentaur)', map_location='16q22.2-q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75293710, end_position_on_the_genomic_accession=75433512, orientation='minus', exon_count=10, OMIM=608108, no_hdr0=''),
  10430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10430, CurrentID=0, Status='live', Symbol='TMEM147', Aliases='NIFIE14', description='transmembrane protein 147', other_designations='protein NIFIE 14|seven transmembrane domain protein', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35545600, end_position_on_the_genomic_accession=35547527, orientation='plus', exon_count=7, OMIM=613585, no_hdr0=''),
  10432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10432, CurrentID=0, Status='live', Symbol='RBM14', Aliases='COAA, PSP2, SIP, SYTIP1, TMEM137', description='RNA binding motif protein 14', other_designations='RRM-containing coactivator activator/modulator|SYT-interacting protein|paraspeckle protein 2|synaptotagmin-interacting protein|transmembrane protein 137', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66616582, end_position_on_the_genomic_accession=66629934, orientation='plus', exon_count=3, OMIM=612409, no_hdr0=''),
  106481450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481450, CurrentID=0, Status='live', Symbol='RNU6-854P', Aliases='', description='RNA, U6 small nuclear 854, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10434, CurrentID=0, Status='live', Symbol='LYPLA1', Aliases='APT-1, APT1, LPL-I, LPL1, hAPT1', description='lysophospholipase I', other_designations='lysoPLA I|lysophospholipid-specific lysophospholipase', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54046367, end_position_on_the_genomic_accession=54102017, orientation='minus', exon_count=10, OMIM=605599, no_hdr0=''),
  10435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10435, CurrentID=0, Status='live', Symbol='CDC42EP2', Aliases='BORG1, CEP2', description='CDC42 effector protein (Rho GTPase binding) 2', other_designations='CRIB-containing BOGR1 protein|binder of Rho GTPases 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65314818, end_position_on_the_genomic_accession=65322429, orientation='plus', exon_count=2, OMIM=606132, no_hdr0=''),
  10436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10436, CurrentID=0, Status='live', Symbol='EMG1', Aliases='C2F, Grcc2f, NEP1', description='EMG1 N1-specific pseudouridine methyltransferase', other_designations='18S rRNA (pseudouridine(1248)-N1)-methyltransferase|18S rRNA (pseudouridine-N1-)-methyltransferase NEP1|18S rRNA Psi1248 methyltransferase|EMG1 N1-specific pseudouridine methyl transferase|EMG1 nucleolar protein homolog|essential for mitotic growth 1|ribosome biogenesis protein NEP1', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6970781, end_position_on_the_genomic_accession=6976003, orientation='plus', exon_count=6, OMIM=611531, no_hdr0=''),
  10437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10437, CurrentID=0, Status='live', Symbol='IFI30', Aliases='GILT, IFI-30, IP-30, IP30', description='interferon, gamma-inducible protein 30', other_designations='gamma-interferon-inducible protein IP-30|interferon gamma-inducible protein 30 preproprotein|legumaturain', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18173780, end_position_on_the_genomic_accession=18178124, orientation='plus', exon_count=7, OMIM=604664, no_hdr0=''),
  10438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10438, CurrentID=0, Status='live', Symbol='C1D', Aliases='LRP1, Rrp47, SUN-CoR, SUNCOR, hC1D', description='C1D nuclear receptor corepressor', other_designations='C1D DNA-binding protein|C1D nuclear receptor co-repressor|nuclear DNA-binding protein|small unique nuclear receptor co-repressor|small unique nuclear receptor corepressor', map_location='2p13-p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68042200, end_position_on_the_genomic_accession=68063027, orientation='minus', exon_count=6, OMIM=606997, no_hdr0=''),
  10439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10439, CurrentID=0, Status='live', Symbol='OLFM1', Aliases='AMY, NOE1, NOELIN1, OlfA', description='olfactomedin 1', other_designations='neuroblastoma protein|neuronal olfactomedin-related ER localized protein|olfactomedin related ER localized protein|olfactomedin-1|pancortin 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135075243, end_position_on_the_genomic_accession=135121184, orientation='plus', exon_count=9, OMIM=605366, no_hdr0=''),
  10440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10440, CurrentID=0, Status='live', Symbol='TIMM17A', Aliases='TIM17, TIM17A', description='translocase of inner mitochondrial membrane 17 homolog A (yeast)', other_designations='inner membrane preprotein translocase Tim17a|mitochondrial inner membrane translocase|preprotein translocase', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201955491, end_position_on_the_genomic_accession=201970661, orientation='plus', exon_count=6, OMIM=605057, no_hdr0=''),
  100009676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009676, CurrentID=0, Status='live', Symbol='ZBTB11-AS1', Aliases='', description='ZBTB11 antisense RNA 1', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101676430, end_position_on_the_genomic_accession=101679217, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10443, CurrentID=0, Status='live', Symbol='N4BP2L2', Aliases='92M18.3, CG005, PFAAP5', description='NEDD4 binding protein 2-like 2', other_designations='92M18.3 (novel protein)|CG016|phosphonoformate immuno-associated protein 5|protein from BCRA2 region', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32432488, end_position_on_the_genomic_accession=32538885, orientation='minus', exon_count=16, OMIM=615788, no_hdr0=''),
  10444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10444, CurrentID=0, Status='live', Symbol='ZER1', Aliases='C9orf60, ZYG, ZYG11BL', description='zyg-11 related, cell cycle regulator', other_designations='ZYG homolog|zer-1 homolog|zyg-11 homolog B-like protein', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128729786, end_position_on_the_genomic_accession=128772422, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  10445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10445, CurrentID=0, Status='live', Symbol='MCRS1', Aliases='ICP22BP, INO80Q, MCRS2, MSP58, P78', description='microspherule protein 1', other_designations='58 kDa microspherule protein|INO80 complex subunit J|INO80 complex subunit Q|cell cycle-regulated factor (78 kDa)|cell cycle-regulated factor p78', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49558294, end_position_on_the_genomic_accession=49568225, orientation='minus', exon_count=17, OMIM=609504, no_hdr0=''),
  10446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10446, CurrentID=0, Status='live', Symbol='LRRN2', Aliases='FIGLER7, GAC1, LRANK1, LRRN5', description='leucine rich repeat neuronal 2', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domain 7|glioma amplified on chromosome 1 protein|leucine rich and ankyrin repeats 1|leucine rich repeat neuronal 5|leucine-rich repeat neuronal protein 5', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204617170, end_position_on_the_genomic_accession=204685733, orientation='minus', exon_count=5, OMIM=605492, no_hdr0=''),
  10447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10447, CurrentID=0, Status='live', Symbol='FAM3C', Aliases='GS3786, ILEI', description='family with sequence similarity 3 member C', other_designations='family with sequence similarity 3, member C|interleukin-like EMT inducer|interleukin-like epithelial-mesenchymal transition inducer|predicted osteoblast protein', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121348245, end_position_on_the_genomic_accession=121396369, orientation='minus', exon_count=12, OMIM=608618, no_hdr0=''),
  10449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10449, CurrentID=0, Status='live', Symbol='ACAA2', Aliases='DSAEC', description='acetyl-CoA acyltransferase 2', other_designations='T1|acetyl-Coenzyme A acyltransferase 2|beta ketothiolase|beta-ketothiolase|mitochondrial 3-oxoacyl-CoA thiolase|mitochondrial 3-oxoacyl-Coenzyme A thiolase', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49783504, end_position_on_the_genomic_accession=49813881, orientation='minus', exon_count=10, OMIM=604770, no_hdr0=''),
  10450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10450, CurrentID=0, Status='live', Symbol='PPIE', Aliases='CYP-33, CYP33', description='peptidylprolyl isomerase E (cyclophilin E)', other_designations='PPIase E|cyclophilin-33|rotamase E', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39738845, end_position_on_the_genomic_accession=39763914, orientation='plus', exon_count=13, OMIM=602435, no_hdr0=''),
  10451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10451, CurrentID=0, Status='live', Symbol='VAV3', Aliases='', description='vav guanine nucleotide exchange factor 3', other_designations='VAV-3|vav 3 guanine nucleotide exchange factor|vav 3 oncogene', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107571160, end_position_on_the_genomic_accession=107964923, orientation='minus', exon_count=32, OMIM=605541, no_hdr0=''),
  10452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10452, CurrentID=0, Status='live', Symbol='TOMM40', Aliases='C19orf1, D19S1177E, PER-EC1, PEREC1, TOM40', description='translocase of outer mitochondrial membrane 40 homolog (yeast)', other_designations='mitochondrial outer membrane protein|p38.5|protein Haymaker|translocase of outer membrane 40 kDa subunit homolog', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44891220, end_position_on_the_genomic_accession=44903689, orientation='plus', exon_count=10, OMIM=608061, no_hdr0=''),
  10454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10454, CurrentID=0, Status='live', Symbol='TAB1', Aliases="3'-Tab1, MAP3K7IP1", description='TGF-beta activated kinase 1/MAP3K7 binding protein 1', other_designations='TAK1-binding protein 1|mitogen-activated protein kinase kinase kinase 7-interacting protein 1|transforming growth factor beta-activated kinase-binding protein 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39399754, end_position_on_the_genomic_accession=39437127, orientation='plus', exon_count=12, OMIM=602615, no_hdr0=''),
  10455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10455, CurrentID=0, Status='live', Symbol='ECI2', Aliases='ACBD2, DRS-1, DRS1, HCA88, PECI, dJ1013A10.3', description='enoyl-CoA delta isomerase 2', other_designations='D3,D2-enoyl-CoA isomerase|DBI-related protein 1|acyl-Coenzyme A binding domain containing 2|delta(3),delta(2)-enoyl-CoA isomerase|diazepam-binding inhibitor-related protein 1|dodecenoyl-CoA isomerase|hepatocellular carcinoma-associated antigen 88|peroxisomal 3,2-trans-enoyl-CoA isomerase|peroxisomal D3,D2-enoyl-CoA isomerase|renal carcinoma antigen NY-REN-1', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4115693, end_position_on_the_genomic_accession=4135597, orientation='minus', exon_count=12, OMIM=608024, no_hdr0=''),
  10456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10456, CurrentID=0, Status='live', Symbol='HAX1', Aliases='HCLSBP1, HS1BP1, SCN3', description='HCLS1 associated protein X-1', other_designations='HAX-1|HCLS1 (and PKD2) associated protein|HS1 binding protein|HS1-associating protein X-1|HS1-binding protein 1|HSP1BP-1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154272563, end_position_on_the_genomic_accession=154275875, orientation='plus', exon_count=7, OMIM=605998, no_hdr0=''),
  10457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10457, CurrentID=0, Status='live', Symbol='GPNMB', Aliases='HGFIN, NMB', description='glycoprotein (transmembrane) nmb', other_designations='glycoprotein NMB|glycoprotein nmb-like protein|osteoactivin|transmembrane glycoprotein HGFIN', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23246686, end_position_on_the_genomic_accession=23275110, orientation='plus', exon_count=11, OMIM=604368, no_hdr0=''),
  10458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10458, CurrentID=0, Status='live', Symbol='BAIAP2', Aliases='BAP2, FLAF3, IRSP53', description='BAI1-associated protein 2', other_designations='IRS-58|IRSp53/58|fas ligand-associated factor 3|insulin receptor substrate p53/p58|insulin receptor substrate protein of 53 kDa', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81035144, end_position_on_the_genomic_accession=81117432, orientation='plus', exon_count=20, OMIM=605475, no_hdr0=''),
  10459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10459, CurrentID=0, Status='live', Symbol='MAD2L2', Aliases='MAD2B, POLZ2, REV7', description='MAD2 mitotic arrest deficient-like 2 (yeast)', other_designations='MAD2 (mitotic arrest deficient, yeast, homolog)-like 2|MAD2-like protein 2|REV7 homolog|hREV7|mitotic arrest deficient 2-like protein 2|mitotic arrest deficient homolog-like 2|polymerase (DNA-directed), zeta 2, accessory subunit', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11674480, end_position_on_the_genomic_accession=11691788, orientation='minus', exon_count=11, OMIM=604094, no_hdr0=''),
  10460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10460, CurrentID=0, Status='live', Symbol='TACC3', Aliases='ERIC-1, ERIC1', description='transforming, acidic coiled-coil containing protein 3', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1721490, end_position_on_the_genomic_accession=1745178, orientation='plus', exon_count=18, OMIM=605303, no_hdr0=''),
  10461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10461, CurrentID=0, Status='live', Symbol='MERTK', Aliases='MER, RP38, Tyro12, c-Eyk, c-mer', description='MER proto-oncogene, tyrosine kinase', other_designations='MER receptor tyrosine kinase|STK kinase|c-mer proto-oncogene tyrosine kinase|proto-oncogene c-Mer|receptor tyrosine kinase MerTK', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111898479, end_position_on_the_genomic_accession=112039953, orientation='plus', exon_count=22, OMIM=604705, no_hdr0=''),
  10462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10462, CurrentID=0, Status='live', Symbol='CLEC10A', Aliases='CD301, CLECSF13, CLECSF14, HML, HML2, MGL', description='C-type lectin domain family 10 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 13 (macrophage-derived)|C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 14 (macrophage-derived)|C-type lectin domain family 10, member A|macrophage lectin 2 (calcium dependent)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7074537, end_position_on_the_genomic_accession=7080601, orientation='minus', exon_count=10, OMIM=605999, no_hdr0=''),
  10463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10463, CurrentID=0, Status='live', Symbol='SLC30A9', Aliases='C4orf1, GAC63, HUEL, ZNT9', description='solute carrier family 30 (zinc transporter), member 9', other_designations='GRIP1-dependent nuclear receptor coactivator|expressed in human embryonic lung|human embryonic lung protein|solute carrier family 30 member 9|znT-9', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41990472, end_position_on_the_genomic_accession=42087534, orientation='plus', exon_count=18, OMIM=604604, no_hdr0=''),
  10464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10464, CurrentID=0, Status='live', Symbol='PIBF1', Aliases='C13orf24, CEP90, PIBF', description='progesterone immunomodulatory binding factor 1', other_designations='centrosomal protein of 90 kDa|progesterone-induced blocking factor 1', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72782092, end_position_on_the_genomic_accession=73016454, orientation='plus', exon_count=21, OMIM=607532, no_hdr0=''),
  10465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10465, CurrentID=0, Status='live', Symbol='PPIH', Aliases='CYP-20, CYPH, SnuCyp-20, USA-CYP', description='peptidylprolyl isomerase H (cyclophilin H)', other_designations='PPIase h|U-snRNP-associated cyclophilin SnuCyp-20|U-snRNP-associated cyclophilin SunCyp-20|USA-CyP SnuCyp-20|cyclophilin H|rotamase H|small nuclear ribonucleoprotein particle-specific cyclophilin H', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42657770, end_position_on_the_genomic_accession=42681654, orientation='plus', exon_count=16, OMIM=606095, no_hdr0=''),
  10466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10466, CurrentID=0, Status='live', Symbol='COG5', Aliases='CDG2I, GOLTC1, GTC90', description='component of oligomeric golgi complex 5', other_designations='13S golgi transport complex 1 90 kDa subunit|COG complex subunit 5|conserved oligomeric Golgi complex protein 5', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107201744, end_position_on_the_genomic_accession=107564514, orientation='minus', exon_count=23, OMIM=606821, no_hdr0=''),
  10467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10467, CurrentID=0, Status='live', Symbol='ZNHIT1', Aliases='CG1I, ZNFN4A1', description='zinc finger, HIT-type containing 1', other_designations='H_DJ0747G18.14|cyclin-G1-binding protein 1|p18 Hamlet|p18Hamlet|putative cyclin G1 interacting protein|zinc finger protein subfamily 4A member 1|zinc finger protein, subfamily 4A (HIT domain containing), member 1|zinc finger, HIT domain containing 1|zinc finger, HIT type 1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101217704, end_position_on_the_genomic_accession=101224192, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10468, CurrentID=0, Status='live', Symbol='FST', Aliases='FS', description='follistatin', other_designations='activin-binding protein|follistatin isoform FST317', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53480341, end_position_on_the_genomic_accession=53487134, orientation='plus', exon_count=7, OMIM=136470, no_hdr0=''),
  10469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10469, CurrentID=0, Status='live', Symbol='TIMM44', Aliases='TIM44', description='translocase of inner mitochondrial membrane 44 homolog (yeast)', other_designations='36/8-5 fusion protein with epitope for anti-lectin antibody|mitochondrial inner membrane translocase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7926718, end_position_on_the_genomic_accession=7943823, orientation='minus', exon_count=13, OMIM=605058, no_hdr0=''),
  10470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10470, CurrentID=0, Status='live', Symbol='D6S2723E', Aliases='BING5', description='DNA segment on chromosome 6 (unique, pseudogene) 2723 expressed sequence', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=33360001, end_position_on_the_genomic_accession=33360545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10471, CurrentID=0, Status='live', Symbol='PFDN6', Aliases='H2-KE2, HKE2, KE-2, PFD6', description='prefoldin subunit 6', other_designations='HLA class II region expressed gene KE2|KE2 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33289597, end_position_on_the_genomic_accession=33290934, orientation='plus', exon_count=5, OMIM=605660, no_hdr0=''),
  10472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10472, CurrentID=0, Status='live', Symbol='ZBTB18', Aliases='C2H2-171, MRD22, RP58, TAZ-1, ZNF238', description='zinc finger and BTB domain containing 18', other_designations='58 kDa repressor protein|transcriptional repressor RP58|translin-associated zinc finger protein 1|zinc finger protein 238|zinc finger protein C2H2-171', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244048939, end_position_on_the_genomic_accession=244057476, orientation='plus', exon_count=4, OMIM=608433, no_hdr0=''),
  10473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10473, CurrentID=0, Status='live', Symbol='HMGN4', Aliases='HMG17L3, NHC', description='high mobility group nucleosomal binding domain 4', other_designations='high mobility group protein N4|high-mobility group (nonhistone chromosomal) protein 17-like 3|high-mobility group protein 17-like 3|non-histone chromosomal protein HMG-17-like 3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26538344, end_position_on_the_genomic_accession=26546937, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10474, CurrentID=0, Status='live', Symbol='TADA3', Aliases='ADA3, NGG1, STAF54L, hADA3, TADA3', description='transcriptional adaptor 3', other_designations='ADA3 homolog|ADA3-like protein', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9779964, end_position_on_the_genomic_accession=9793011, orientation='minus', exon_count=10, OMIM=602945, no_hdr0=''),
  10475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10475, CurrentID=0, Status='live', Symbol='TRIM38', Aliases='RNF15, RORET', description='tripartite motif containing 38', other_designations='Ro/SSA ribonucleoprotein homolog|ring finger protein 15|tripartite motif-containing protein 38|zinc finger protein RoRet', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25962689, end_position_on_the_genomic_accession=25987329, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  10476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10476, CurrentID=0, Status='live', Symbol='ATP5H', Aliases='ATPQ', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit D', other_designations='ATP synthase D chain, mitochondrial|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit d|ATP synthase, H+ transporting, mitochondrial F1F0, subunit d|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d|ATPase subunit d|My032 protein', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75038860, end_position_on_the_genomic_accession=75046979, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  10477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10477, CurrentID=0, Status='live', Symbol='UBE2E3', Aliases='UBCH9, UbcM2', description='ubiquitin conjugating enzyme E2E 3', other_designations='E2 ubiquitin-conjugating enzyme E3|ubiquitin carrier protein E3|ubiquitin-conjugating enzyme E2-23 kDa|ubiquitin-conjugating enzyme E2E 3 (UBC4/5 homolog, yeast)|ubiquitin-conjugating enzyme E2E 3 (homologous to yeast UBC4/5)|ubiquitin-protein ligase E3', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180980385, end_position_on_the_genomic_accession=181063427, orientation='plus', exon_count=10, OMIM=604151, no_hdr0=''),
  10478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10478, CurrentID=0, Status='live', Symbol='SLC25A17', Aliases='PMP34', description='solute carrier family 25 (mitochondrial carrier; peroxisomal membrane protein, 34kDa), member 17', other_designations='solute carrier family 25 member 17', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40769630, end_position_on_the_genomic_accession=40819399, orientation='minus', exon_count=12, OMIM=606795, no_hdr0=''),
  10479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10479, CurrentID=0, Status='live', Symbol='SLC9A6', Aliases='MRSA, NHE6', description='solute carrier family 9, subfamily A (NHE6, cation proton antiporter 6), member 6', other_designations='Na(+)/H(+) exchanger 6|solute carrier family 9 (sodium/hydrogen exchanger), member 6', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135985424, end_position_on_the_genomic_accession=136047269, orientation='plus', exon_count=17, OMIM=300231, no_hdr0=''),
  10480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10480, CurrentID=0, Status='live', Symbol='EIF3M', Aliases='B5, GA17, PCID1, TANGO7, hfl-B5', description='eukaryotic translation initiation factor 3 subunit M', other_designations='B5 receptor|PCI domain containing 1 (herpesvirus entry mediator)|PCI domain-containing protein 1|dendritic cell protein|eukaryotic translation initiation factor 3, subunit M|fetal lung protein B5|transport and golgi organization 7 homolog', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32583767, end_position_on_the_genomic_accession=32604498, orientation='plus', exon_count=11, OMIM=609641, no_hdr0=''),
  10481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10481, CurrentID=0, Status='live', Symbol='HOXB13', Aliases='PSGD', description='homeobox B13', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48724763, end_position_on_the_genomic_accession=48728749, orientation='minus', exon_count=2, OMIM=604607, no_hdr0=''),
  10482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10482, CurrentID=0, Status='live', Symbol='NXF1', Aliases='MEX67, TAP', description='nuclear RNA export factor 1', other_designations='mRNA export factor TAP|tip associating protein|tip-associated protein|tip-associating protein', map_location='11q12-q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62792126, end_position_on_the_genomic_accession=62805492, orientation='minus', exon_count=20, OMIM=602647, no_hdr0=''),
  10483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10483, CurrentID=0, Status='live', Symbol='SEC23B', Aliases='CDA-II, CDAII, CDAN2, HEMPAS', description='Sec23 homolog B, COPII coat complex component', other_designations='SEC23-like protein B|SEC23-related protein B|Sec23 homolog B|transport protein SEC23B', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18507544, end_position_on_the_genomic_accession=18561415, orientation='plus', exon_count=20, OMIM=610512, no_hdr0=''),
  10484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10484, CurrentID=0, Status='live', Symbol='SEC23A', Aliases='CLSD', description='Sec23 homolog A, COPII coat complex component', other_designations='SEC23-related protein A|Sec23 homolog A', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39031919, end_position_on_the_genomic_accession=39103243, orientation='minus', exon_count=22, OMIM=610511, no_hdr0=''),
  10485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10485, CurrentID=0, Status='live', Symbol='C1orf61', Aliases='CROC4', description='chromosome 1 open reading frame 61', other_designations='contingent replication of cDNA 4|contingent replication of cDNA-4|transcriptional activator of the c fos promoter|transcriptional activator of the c-fos promoter', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156404256, end_position_on_the_genomic_accession=156434325, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  10486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10486, CurrentID=0, Status='live', Symbol='CAP2', Aliases='', description='CAP, adenylate cyclase-associated protein, 2 (yeast)', other_designations='2810452G09Rik|CAP 2', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17393216, end_position_on_the_genomic_accession=17557792, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  10487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10487, CurrentID=0, Status='live', Symbol='CAP1', Aliases='CAP-PEN, CAP1', description='CAP, adenylate cyclase-associated protein 1 (yeast)', other_designations='CAP 1', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40040240, end_position_on_the_genomic_accession=40072649, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  10488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10488, CurrentID=0, Status='live', Symbol='CREB3', Aliases='LUMAN, LZIP, sLZIP', description='cAMP responsive element binding protein 3', other_designations='CREB-3|basic leucine zipper protein|cAMP-responsive element-binding protein 3|cyclic AMP response element (CRE)-binding protein/activating transcription factor 1|leucin zipper proitein|leucine zipper protein|small leucine zipper protein|transcription factor LZIP-alpha', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35732320, end_position_on_the_genomic_accession=35737008, orientation='plus', exon_count=9, OMIM=606443, no_hdr0=''),
  10489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10489, CurrentID=0, Status='live', Symbol='LRRC41', Aliases='MUF1, PP7759', description='leucine rich repeat containing 41', other_designations='MUF1 protein (MUF1)|elongin BC-interacting leucine-rich repeat protein|protein Muf1', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46278400, end_position_on_the_genomic_accession=46303366, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  10490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10490, CurrentID=0, Status='live', Symbol='VTI1B', Aliases='VTI1, VTI1-LIKE, VTI1L, VTI2, v-SNARE, vti1-rp1', description='vesicle transport through interaction with t-SNAREs 1B', other_designations='vesicle transport v-SNARE protein Vti1-like 1|vesicle-associated soluble NSF attachment protein receptor', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67651149, end_position_on_the_genomic_accession=67674885, orientation='minus', exon_count=6, OMIM=603207, no_hdr0=''),
  10491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10491, CurrentID=0, Status='live', Symbol='CRTAP', Aliases='CASP, LEPREL3, OI7, P3H5', description='cartilage associated protein', other_designations='leprecan-like 3|prolyl 3-hydroxylase family member 5 (non-enzymatic)', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33113958, end_position_on_the_genomic_accession=33147773, orientation='plus', exon_count=7, OMIM=605497, no_hdr0=''),
  10492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10492, CurrentID=0, Status='live', Symbol='SYNCRIP', Aliases='GRY-RBP, GRYRBP, HNRNPQ, HNRPQ1, NSAP1, PP68, hnRNP-Q', description='synaptotagmin binding, cytoplasmic RNA interacting protein', other_designations='NS1-associated protein 1|glycine- and tyrosine-rich RNA-binding protein', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85607784, end_position_on_the_genomic_accession=85643850, orientation='minus', exon_count=13, OMIM=616686, no_hdr0=''),
  10493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10493, CurrentID=0, Status='live', Symbol='VAT1', Aliases='VATI', description='vesicle amine transport 1', other_designations='membrane protein of cholinergic synaptic vesicles|vesicle amine transport protein 1 homolog (T. californica)', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43014605, end_position_on_the_genomic_accession=43022442, orientation='minus', exon_count=6, OMIM=604631, no_hdr0=''),
  10494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10494, CurrentID=0, Status='live', Symbol='STK25', Aliases='SOK1, YSK1', description='serine/threonine kinase 25', other_designations='Ste20, yeast homolog|ste20-like kinase|ste20/oxidant stress response kinase 1|sterile 20/oxidant stress-response kinase 1', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241494707, end_position_on_the_genomic_accession=241509730, orientation='minus', exon_count=14, OMIM=602255, no_hdr0=''),
  10495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10495, CurrentID=0, Status='live', Symbol='ENOX2', Aliases='APK1, COVA1, tNOX', description='ecto-NOX disulfide-thiol exchanger 2', other_designations='APK1 antigen|cytosolic ovarian carcinoma antigen 1|tumor-associated hydroquinone oxidase', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130623379, end_position_on_the_genomic_accession=130903317, orientation='minus', exon_count=22, OMIM=300282, no_hdr0=''),
  10497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10497, CurrentID=0, Status='live', Symbol='UNC13B', Aliases='MUNC13, UNC13, Unc13h2', description='unc-13 homolog B (C. elegans)', other_designations='homolog of rat Munc13 (diacylglycerol-binding)|munc13-2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35161982, end_position_on_the_genomic_accession=35405337, orientation='plus', exon_count=43, OMIM=605836, no_hdr0=''),
  10498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10498, CurrentID=0, Status='live', Symbol='CARM1', Aliases='PRMT4', description='coactivator associated arginine methyltransferase 1', other_designations='coactivator-associated arginine methyltransferase 1|protein arginine N-methyltransferase 4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10871577, end_position_on_the_genomic_accession=10922775, orientation='plus', exon_count=18, OMIM=603934, no_hdr0=''),
  10499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10499, CurrentID=0, Status='live', Symbol='NCOA2', Aliases='GRIP1, KAT13C, NCoA-2, SRC2, TIF2, bHLHe75', description='nuclear receptor coactivator 2', other_designations='class E basic helix-loop-helix protein 75|glucocorticoid receptor-interacting protein-1|hTIF2|transcriptional intermediary factor 2', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70109770, end_position_on_the_genomic_accession=70405437, orientation='minus', exon_count=32, OMIM=601993, no_hdr0=''),
  10500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10500, CurrentID=0, Status='live', Symbol='SEMA6C', Aliases='SEMAY, m-SemaY, m-SemaY2', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6C', other_designations='sema Y|semaphorin-Y', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151131685, end_position_on_the_genomic_accession=151146664, orientation='minus', exon_count=20, OMIM=609294, no_hdr0=''),
  10501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10501, CurrentID=0, Status='live', Symbol='SEMA6B', Aliases='SEM-SEMA-Y, SEMA-VIB, SEMAN, semaZ', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6B', other_designations='semaphorin VIB|semaphorin Z|semaphorin-6Ba', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4542587, end_position_on_the_genomic_accession=4581602, orientation='minus', exon_count=18, OMIM=608873, no_hdr0=''),
  10505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10505, CurrentID=0, Status='live', Symbol='SEMA4F', Aliases='M-SEMA, PRO2353, S4F, SEMAM, SEMAW, m-Sema-M', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4F', other_designations='sema domain immunoglobulin domain transmembrane domain (TM) and short cytoplasmic domain (semaphorin) 4F|semaphorin M|semaphorin-M|semaphorin-W', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74654228, end_position_on_the_genomic_accession=74694897, orientation='plus', exon_count=16, OMIM=603706, no_hdr0=''),
  10507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10507, CurrentID=0, Status='live', Symbol='SEMA4D', Aliases='C9orf164, CD100, M-sema-G, SEMAJ, coll-4', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4D', other_designations='A8|BB18|GR3|sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, 4D', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89360791, end_position_on_the_genomic_accession=89498014, orientation='minus', exon_count=30, OMIM=601866, no_hdr0=''),
  10509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10509, CurrentID=0, Status='live', Symbol='SEMA4B', Aliases='SEMAC, SemC', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4B', other_designations='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, 4B', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90184920, end_position_on_the_genomic_accession=90229664, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100422604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422604, CurrentID=0, Status='live', Symbol='LOC100422604', Aliases='', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158336248, end_position_on_the_genomic_accession=158337060, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270839, CurrentID=0, Status='live', Symbol='RPL5P17', Aliases='RPL5_8_591', description='ribosomal protein L5 pseudogene 17', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85464359, end_position_on_the_genomic_accession=85465372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10512, CurrentID=0, Status='live', Symbol='SEMA3C', Aliases='SEMAE, SemE', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3C', other_designations='sema E|semaphorin E|semaphorin-E', map_location='7q21-q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80742538, end_position_on_the_genomic_accession=80919351, orientation='minus', exon_count=19, OMIM=602645, no_hdr0=''),
  10513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10513, CurrentID=0, Status='live', Symbol='APPBP2', Aliases='APP-BP2, HS.84084, PAT1', description='amyloid beta precursor protein (cytoplasmic tail) binding protein 2', other_designations='protein interacting with APP tail 1', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60443149, end_position_on_the_genomic_accession=60526240, orientation='minus', exon_count=15, OMIM=605324, no_hdr0=''),
  10514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10514, CurrentID=0, Status='live', Symbol='MYBBP1A', Aliases='P160, PAP2', description='MYB binding protein (P160) 1a', other_designations='p53-activated protein-2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4538896, end_position_on_the_genomic_accession=4555386, orientation='minus', exon_count=26, OMIM=604885, no_hdr0=''),
  10516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10516, CurrentID=0, Status='live', Symbol='FBLN5', Aliases='ADCL2, ARCL1A, ARMD3, DANCE, EVEC, FIBL-5, HNARMD, UP50', description='fibulin 5', other_designations='developmental arteries and neural crest EGF-like protein|urine p50 protein', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91869411, end_position_on_the_genomic_accession=91947702, orientation='minus', exon_count=14, OMIM=604580, no_hdr0=''),
  10517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10517, CurrentID=0, Status='live', Symbol='FBXW10', Aliases='Fbw10, HREP, SM25H2, SM2SH2', description='F-box and WD repeat domain containing 10', other_designations='F-box and WD-40 domain protein 10|ubiquitin ligase specificity factor', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18740223, end_position_on_the_genomic_accession=18779349, orientation='plus', exon_count=18, OMIM=611679, no_hdr0=''),
  10518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10518, CurrentID=0, Status='live', Symbol='CIB2', Aliases='DFNB48, KIP2, USH1J', description='calcium and integrin binding family member 2', other_designations='DNA-dependent protein kinase catalytic subunit-interacting protein 2', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78104606, end_position_on_the_genomic_accession=78131721, orientation='minus', exon_count=7, OMIM=605564, no_hdr0=''),
  10519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10519, CurrentID=0, Status='live', Symbol='CIB1', Aliases='CIB, CIBP, KIP1, PRKDCIP, SIP2-28', description='calcium and integrin binding 1', other_designations='DNA-PK interaction protein|DNA-PKcs-interacting protein|DNA-dependent protein kinase interacting protein|SNK-interacting protein 2-28|calcium and integrin binding 1 (calmyrin)', map_location='15q25.3-q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90230245, end_position_on_the_genomic_accession=90265759, orientation='minus', exon_count=8, OMIM=602293, no_hdr0=''),
  10520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10520, CurrentID=0, Status='live', Symbol='ZNF211', Aliases='C2H2-25, CH2H2-25, ZNF-25, ZNFC25', description='zinc finger protein 211', other_designations='zinc finger protein C2H2-25', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57633167, end_position_on_the_genomic_accession=57642779, orientation='plus', exon_count=6, OMIM=601856, no_hdr0=''),
  10521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10521, CurrentID=0, Status='live', Symbol='DDX17', Aliases='P72, RH70', description='DEAD (Asp-Glu-Ala-Asp) box helicase 17', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 17|DEAD box protein p72|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 17 (72kD)|RNA-dependent helicase p72', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38483438, end_position_on_the_genomic_accession=38506340, orientation='minus', exon_count=13, OMIM=608469, no_hdr0=''),
  10522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10522, CurrentID=0, Status='live', Symbol='DEAF1', Aliases='MRD24, NUDR, SPN, ZMYND5', description='DEAF1, transcription factor', other_designations='DEAF1 transcription factor|nuclear DEAF-1-related transcriptional regulator|suppressin|zinc finger MYND domain-containing protein 5', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=644220, end_position_on_the_genomic_accession=695754, orientation='minus', exon_count=17, OMIM=602635, no_hdr0=''),
  10523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10523, CurrentID=0, Status='live', Symbol='CHERP', Aliases='DAN16, SCAF6, SRA1', description='calcium homeostasis endoplasmic reticulum protein', other_designations='ERPROT 213-21|ERPROT213-21|SR-related CTD associated factor 6|SR-related CTD-associated factor 6|protein with polyglutamine repeat', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16517889, end_position_on_the_genomic_accession=16542452, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  10524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10524, CurrentID=0, Status='live', Symbol='KAT5', Aliases='ESA1, HTATIP, HTATIP1, PLIP, TIP, TIP60, ZC2HC5, cPLA2', description='K(lysine) acetyltransferase 5', other_designations='60 kDa Tat-interactive protein|HIV-1 Tat interactive protein, 60kDa|K-acetyltransferase 5|Tat interacting protein, 60kDa|cPLA(2)-interacting protein|cPLA2 interacting protein|histone acetyltransferase HTATIP', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65711996, end_position_on_the_genomic_accession=65719606, orientation='plus', exon_count=14, OMIM=601409, no_hdr0=''),
  10525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10525, CurrentID=0, Status='live', Symbol='HYOU1', Aliases='GRP-170, Grp170, HSP12A, ORP-150, ORP150', description='hypoxia up-regulated 1', other_designations='150 kDa oxygen-regulated protein|170 kDa glucose-regulated protein|oxygen regulated protein (150kD)', map_location='11q23.1-q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119044185, end_position_on_the_genomic_accession=119057246, orientation='minus', exon_count=28, OMIM=601746, no_hdr0=''),
  10526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10526, CurrentID=0, Status='live', Symbol='IPO8', Aliases='RANBP8', description='importin 8', other_designations='RAN binding protein 8|imp8|ran-binding protein 8', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30628981, end_position_on_the_genomic_accession=30695995, orientation='minus', exon_count=28, OMIM=605600, no_hdr0=''),
  10527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10527, CurrentID=0, Status='live', Symbol='IPO7', Aliases='Imp7, RANBP7', description='importin 7', other_designations='RAN-binding protein 7', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9384622, end_position_on_the_genomic_accession=9448127, orientation='plus', exon_count=25, OMIM=605586, no_hdr0=''),
  10528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10528, CurrentID=0, Status='live', Symbol='NOP56', Aliases='NOL5A, SCA36', description='NOP56 ribonucleoprotein', other_designations='NOP56 ribonucleoprotein homolog|nucleolar protein 5A (56kDa with KKE/D repeat)', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2652532, end_position_on_the_genomic_accession=2658393, orientation='plus', exon_count=12, OMIM=614154, no_hdr0=''),
  10529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10529, CurrentID=0, Status='live', Symbol='NEBL', Aliases='LASP2, LNEBL', description='nebulette', other_designations='LIM-nebulette|actin-binding Z-disk protein', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20779971, end_position_on_the_genomic_accession=21174187, orientation='minus', exon_count=37, OMIM=605491, no_hdr0=''),
  10531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10531, CurrentID=0, Status='live', Symbol='PITRM1', Aliases='MP1, PreP', description='pitrilysin metallopeptidase 1', other_designations='PreP peptidasome|metalloprotease 1 (pitrilysin family)|pitrilysin metalloproteinase 1', map_location='10p15.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3137727, end_position_on_the_genomic_accession=3172841, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  10533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10533, CurrentID=0, Status='live', Symbol='ATG7', Aliases='APG7-LIKE, APG7L, GSA7', description='autophagy related 7', other_designations='ATG12-activating enzyme E1 ATG7|hAGP7|ubiquitin-activating enzyme E1-like protein', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11272324, end_position_on_the_genomic_accession=11557665, orientation='plus', exon_count=27, OMIM=608760, no_hdr0=''),
  10534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10534, CurrentID=0, Status='live', Symbol='SSSCA1', Aliases='p27', description='Sjogren syndrome/scleroderma autoantigen 1', other_designations="Sjogren's syndrome/scleroderma autoantigen 1|autoantigen p27|centromeric autoantigen (27kD)", map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65570459, end_position_on_the_genomic_accession=65571888, orientation='plus', exon_count=4, OMIM=606044, no_hdr0=''),
  10535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10535, CurrentID=0, Status='live', Symbol='RNASEH2A', Aliases='AGS4, JUNB, RNASEHI, RNHIA, RNHL', description='ribonuclease H2 subunit A', other_designations='RNase H(35)|RNase H2 subunit A|RNase HI large subunit|aicardi-Goutieres syndrome 4 protein|ribonuclease H2, large subunit|ribonuclease H2, subunit A|ribonuclease HI large subunit|ribonuclease HI subunit A|ribonuclease HI, large subunit', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12802054, end_position_on_the_genomic_accession=12813648, orientation='plus', exon_count=10, OMIM=606034, no_hdr0=''),
  10536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10536, CurrentID=0, Status='live', Symbol='P3H3', Aliases='GRCB, HSU47926, LEPREL2', description='prolyl 3-hydroxylase 3', other_designations='gene rich cluster, B|leprecan-like 2|protein B', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6828373, end_position_on_the_genomic_accession=6839854, orientation='plus', exon_count=15, OMIM=610342, no_hdr0=''),
  10537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10537, CurrentID=0, Status='live', Symbol='UBD', Aliases='FAT10, GABBR1-3, UBD', description='ubiquitin D', other_designations='diubiquitin|ubiquitin-like protein FAT10', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29555612, end_position_on_the_genomic_accession=29559925, orientation='minus', exon_count=2, OMIM=606050, no_hdr0=''),
  10538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10538, CurrentID=0, Status='live', Symbol='BATF', Aliases='B-ATF1, SFA-2, SFA2, BATF', description='basic leucine zipper transcription factor, ATF-like', other_designations='B-cell-activating transcription factor|SF-HT-activated gene 2 protein|activating transcription factor B', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75522441, end_position_on_the_genomic_accession=75546992, orientation='plus', exon_count=3, OMIM=612476, no_hdr0=''),
  10539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10539, CurrentID=0, Status='live', Symbol='GLRX3', Aliases='GLRX4, GRX3, GRX4, PICOT, TXNL2, TXNL3', description='glutaredoxin 3', other_designations='PKC-interacting cousin of thioredoxin|PKC-theta-interacting protein|PKCq-interacting protein|glutaredoxin 4|thioredoxin-like protein 2', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130136375, end_position_on_the_genomic_accession=130180382, orientation='plus', exon_count=13, OMIM=612754, no_hdr0=''),
  10540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10540, CurrentID=0, Status='live', Symbol='DCTN2', Aliases='DCTN50, DYNAMITIN, HEL-S-77, RBP50', description='dynactin subunit 2', other_designations='50 kD dynein-associated polypeptide|50 kDa dynein-associated polypeptide|dynactin 2 (p50)|dynactin complex 50 kD subunit|dynactin complex 50 kDa subunit|epididymis secretory protein Li 77|p50 dynamitin', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57530050, end_position_on_the_genomic_accession=57547331, orientation='minus', exon_count=17, OMIM=607376, no_hdr0=''),
  10541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10541, CurrentID=0, Status='live', Symbol='ANP32B', Aliases='APRIL, PHAPI2, SSP29', description='acidic nuclear phosphoprotein 32 family member B', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member B|acidic protein rich in leucines|putative HLA-DR-associated protein I-2|silver-stainable protein SSP29', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97983207, end_position_on_the_genomic_accession=98015943, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  10542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10542, CurrentID=0, Status='live', Symbol='LAMTOR5', Aliases='HBXIP, XIP', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 5', other_designations='HBV X-interacting protein|HBx-interacting protein|hepatitis B virus x interacting protein|hepatitis B virus x-interacting protein (9.6kD)|late endosomal/lysosomal adaptor and MAPK and MTOR activator 5', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110401253, end_position_on_the_genomic_accession=110407924, orientation='minus', exon_count=4, OMIM=608521, no_hdr0=''),
  100271837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271837, CurrentID=0, Status='live', Symbol='USH1H', Aliases='', description='Usher syndrome 1H (autosomal recessive)', other_designations='', map_location='15q22-q23', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612632, no_hdr0=''),
  10544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10544, CurrentID=0, Status='live', Symbol='PROCR', Aliases='CCCA, CCD41, EPCR', description='protein C receptor', other_designations='APC receptor|CD201 antigen|activated protein C receptor|cell cycle, centrosome-associated protein|centrocyclin|protein C receptor, endothelial', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35171937, end_position_on_the_genomic_accession=35215989, orientation='plus', exon_count=5, OMIM=600646, no_hdr0=''),
  10547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10547, CurrentID=0, Status='live', Symbol='PSORS4', Aliases='', description='psoriasis susceptibility 4', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603935, no_hdr0=''),
  10548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10548, CurrentID=0, Status='live', Symbol='TM9SF1', Aliases='HMP70, MP70', description='transmembrane 9 superfamily member 1', other_designations='MP70 protein family member|multispanning membrane protein (70kD)|transmembrane protein 9 superfamily member 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24189143, end_position_on_the_genomic_accession=24195736, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  10549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10549, CurrentID=0, Status='live', Symbol='PRDX4', Aliases='AOE37-2, AOE372, HEL-S-97n, PRX-4', description='peroxiredoxin 4', other_designations='antioxidant enzyme AOE372|epididymis secretory sperm binding protein Li 97n|peroxiredoxin IV|prx-IV|thioredoxin peroxidase (antioxidant enzyme)|thioredoxin peroxidase AO372|thioredoxin-dependent peroxide reductase A0372', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23664260, end_position_on_the_genomic_accession=23686399, orientation='plus', exon_count=8, OMIM=300927, no_hdr0=''),
  10550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10550, CurrentID=0, Status='live', Symbol='ARL6IP5', Aliases='DERP11, GTRAP3-18, HSPC127, JWA, PRAF3, addicsin, hp22, jmx', description='ADP ribosylation factor like GTPase 6 interacting protein 5', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 5|ADP-ribosylation factor like GTPase 6 interacting protein 5|ADP-ribosylation factor-like 6 interacting protein 5|ADP-ribosylation factor-like protein 6-interacting protein 5|ADP-ribosylation-like factor 6 interacting protein 5|ARL-6-interacting protein 5|JM5|PRA1 domain family 3|aip-5|cytoskeleton related vitamin A responsive protein|cytoskeleton-related vitamin A-responsive protein|dermal papilla derived protein 11|dermal papilla-derived protein 11|glutamate transporter EAAC1-interacting protein|glutamate transporter EEAC1-associated protein|prenylated Rab acceptor protein 2|protein JWa|putative MAPK activating protein PM27|putative MAPK-activating protein PM27', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69084939, end_position_on_the_genomic_accession=69106088, orientation='plus', exon_count=3, OMIM=605709, no_hdr0=''),
  10551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10551, CurrentID=0, Status='live', Symbol='AGR2', Aliases='AG2, GOB-4, HAG-2, HEL-S-116, PDIA17, XAG-2', description='anterior gradient 2, protein disulphide isomerase family member', other_designations='AG-2|HPC8|anterior gradient homolog 2|epididymis secretory protein Li 116|protein disulfide isomerase family A, member 17|secreted cement gland homolog|secreted cement gland protein XAG-2 homolog', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16791811, end_position_on_the_genomic_accession=16805114, orientation='minus', exon_count=8, OMIM=606358, no_hdr0=''),
  10552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10552, CurrentID=0, Status='live', Symbol='ARPC1A', Aliases='Arc40, HEL-68, HEL-S-307, SOP2Hs, SOP2L', description='actin related protein 2/3 complex subunit 1A', other_designations='SOP2-like protein|actin binding protein (Schizosaccharomyces pombe sop2-like)|actin related protein 2/3 complex subunit 1A, 41kDa|actin related protein 2/3 complex, subunit 1A, 41kDa|epididymis luminal protein 68|epididymis secretory protein Li 307', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99325873, end_position_on_the_genomic_accession=99366262, orientation='plus', exon_count=10, OMIM=604220, no_hdr0=''),
  10553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10553, CurrentID=0, Status='live', Symbol='HTATIP2', Aliases='CC3, SDR44U1, TIP30', description='HIV-1 Tat interactive protein 2', other_designations='30 kDa HIV-1 TAT-interacting protein|HIV-1 TAT-interactive protein 2|HIV-1 Tat interactive protein 2, 30kDa|Tat-interacting protein (30kD)|short chain dehydrogenase/reductase family 44U, member 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20363685, end_position_on_the_genomic_accession=20383783, orientation='plus', exon_count=6, OMIM=605628, no_hdr0=''),
  10554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10554, CurrentID=0, Status='live', Symbol='AGPAT1', Aliases='1-AGPAT1, G15, LPAAT-alpha, LPAATA', description='1-acylglycerol-3-phosphate O-acyltransferase 1', other_designations='1-AGP acyltransferase 1|1-AGPAT 1|1-acylglycerol-3-phosphate O-acyltransferase 1 (acetoacetly Coenzyme A thiolase)|1-acylglycerol-3-phosphate O-acyltransferase 1 (lysophosphatidic acid acyltransferase, alpha)|lysophosphatidic acid acyltransferase alpha|lysophospholipid acyltransferase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32168206, end_position_on_the_genomic_accession=32178111, orientation='minus', exon_count=11, OMIM=603099, no_hdr0=''),
  10555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10555, CurrentID=0, Status='live', Symbol='AGPAT2', Aliases='1-AGPAT2, BSCL, BSCL1, LPAAB, LPAAT-beta', description='1-acylglycerol-3-phosphate O-acyltransferase 2', other_designations='1-AGP acyltransferase 2|1-AGPAT 2|1-acylglycerol-3-phosphate O-acyltransferase 2 (lysophosphatidic acid acyltransferase, beta)|lysophosphatidic acid acyltransferase beta|lysophosphatidic acid acyltransferase-beta', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136673143, end_position_on_the_genomic_accession=136687459, orientation='minus', exon_count=6, OMIM=603100, no_hdr0=''),
  10556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10556, CurrentID=0, Status='live', Symbol='RPP30', Aliases='TSG15', description='ribonuclease P/MRP 30kDa subunit', other_designations='RNase P subunit 2|RNaseP protein p30', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90871952, end_position_on_the_genomic_accession=90908558, orientation='plus', exon_count=14, OMIM=606115, no_hdr0=''),
  10557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10557, CurrentID=0, Status='live', Symbol='RPP38', Aliases='', description='ribonuclease P/MRP 38kDa subunit', other_designations='RNaseP protein p38', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15097180, end_position_on_the_genomic_accession=15104257, orientation='plus', exon_count=3, OMIM=606116, no_hdr0=''),
  10558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10558, CurrentID=0, Status='live', Symbol='SPTLC1', Aliases='HSAN1, HSN1, LBC1, LCB1, SPT1, SPTI', description='serine palmitoyltransferase, long chain base subunit 1', other_designations='LCB 1|SPT 1|long chain base biosynthesis protein 1|serine C-palmitoyltransferase|serine-palmitoyl-CoA transferase 1', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92031134, end_position_on_the_genomic_accession=92115474, orientation='minus', exon_count=17, OMIM=605712, no_hdr0=''),
  10559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10559, CurrentID=0, Status='live', Symbol='SLC35A1', Aliases='CDG2F, CMPST, CST, hCST', description='solute carrier family 35 (CMP-sialic acid transporter), member A1', other_designations='CMP-SA-Tr|CMP-Sia-Tr|mutated CMP-sialic acid transporter A1|solute carrier family 35 (CMP-sialic acid transporter), member 1|solute carrier family 35 (UDP-galactose transporter), member 1|solute carrier family 35 member A1', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87472925, end_position_on_the_genomic_accession=87512339, orientation='plus', exon_count=8, OMIM=605634, no_hdr0=''),
  10560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10560, CurrentID=0, Status='live', Symbol='SLC19A2', Aliases='TC1, THMD1, THT1, THTR1, TRMA', description='solute carrier family 19 (thiamine transporter), member 2', other_designations='high affinity thiamine transporter|reduced folate carrier protein (RFC) like|solute carrier family 19 member 2|thTr-1', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169463909, end_position_on_the_genomic_accession=169486079, orientation='minus', exon_count=7, OMIM=603941, no_hdr0=''),
  10561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10561, CurrentID=0, Status='live', Symbol='IFI44', Aliases='MTAP44, TLDC5, p44', description='interferon induced protein 44', other_designations='TBC/LysM-associated domain containing 5|interferon-induced, hepatitis C-associated microtubular aggregate protein (44kD)|microtubule-associated protein 44', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78648943, end_position_on_the_genomic_accession=78664078, orientation='plus', exon_count=11, OMIM=610468, no_hdr0=''),
  10562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10562, CurrentID=0, Status='live', Symbol='OLFM4', Aliases='GC1, GW112, OLM4, OlfD, UNQ362, bA209J19.1, hGC-1, hOLfD', description='olfactomedin 4', other_designations='G-CSF-stimulated clone 1 protein|antiapoptotic protein GW112', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53028741, end_position_on_the_genomic_accession=53052061, orientation='plus', exon_count=5, OMIM=614061, no_hdr0=''),
  10563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10563, CurrentID=0, Status='live', Symbol='CXCL13', Aliases='ANGIE, ANGIE2, BCA-1, BCA1, BLC, BLR1L, SCYB13', description='chemokine (C-X-C motif) ligand 13', other_designations="B-cell chemoattractant|B-cell-attracting chemokine 1|B-cell-homing chemokine (ligand for Burkitt's lymphoma receptor-1)|B-lymphocyte chemoattractant|CXC chemokine BLC|b cell-attracting chemokine 1|b lymphocyte chemoattractant|chemokine (C-X-C motif) ligand 13 (B-cell chemoattractant)|small inducible cytokine B subfamily (Cys-X-Cys motif), member 13 (B-cell chemoattractant)|small-inducible cytokine B13", map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77511753, end_position_on_the_genomic_accession=77611834, orientation='plus', exon_count=5, OMIM=605149, no_hdr0=''),
  10564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10564, CurrentID=0, Status='live', Symbol='ARFGEF2', Aliases='BIG2, PVNH2, dJ1164I10.1', description='ADP ribosylation factor guanine nucleotide exchange factor 2', other_designations='ADP-ribosylation factor guanine nucleotide-exchange factor 2 (brefeldin A-inhibited)|brefeldin A-inhibited GEP 2', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48921721, end_position_on_the_genomic_accession=49036693, orientation='plus', exon_count=39, OMIM=605371, no_hdr0=''),
  10565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10565, CurrentID=0, Status='live', Symbol='ARFGEF1', Aliases='ARFGEP1, BIG1, P200', description='ADP ribosylation factor guanine nucleotide exchange factor 1', other_designations='ADP-ribosylation factor guanine nucleotide-exchange factor 1 (brefeldin A-inhibited)|p200 ARF guanine nucleotide exchange factor', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67175296, end_position_on_the_genomic_accession=67343897, orientation='minus', exon_count=41, OMIM=604141, no_hdr0=''),
  10566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10566, CurrentID=0, Status='live', Symbol='AKAP3', Aliases='AKAP 110, AKAP110, CT82, FSP95, HEL159, PRKA3, SOB1', description='A-kinase anchoring protein 3', other_designations='A kinase (PRKA) anchor protein 3|A-kinase anchor protein, 110kDa|Fibrous Sheath Protein of 95 kDa|cancer/testis antigen 82|epididymis luminal protein 159|fibrousheathin I|fibrousheathin-1|protein kinase A binding protein AKAP 110|protein kinase A-anchoring protein 3|sperm oocyte-binding protein 1', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4615508, end_position_on_the_genomic_accession=4649047, orientation='minus', exon_count=7, OMIM=604689, no_hdr0=''),
  10567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10567, CurrentID=0, Status='live', Symbol='RABAC1', Aliases='PRA1, PRAF1, YIP3', description='Rab acceptor 1 (prenylated)', other_designations='PRA1 domain family 1|PRA1 family protein 1|prenylated Rab acceptor 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41956681, end_position_on_the_genomic_accession=41959376, orientation='minus', exon_count=5, OMIM=604925, no_hdr0=''),
  10568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10568, CurrentID=0, Status='live', Symbol='SLC34A2', Aliases='NAPI-3B, NAPI-IIb, NPTIIb', description='solute carrier family 34 (type II sodium/phosphate cotransporter), member 2', other_designations='sodium/phosphate cotransporter 2B|solute carrier family 34 (sodium phosphate), member 2|type II sodium-dependent phosphate transporter 3b', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25655813, end_position_on_the_genomic_accession=25678748, orientation='plus', exon_count=14, OMIM=604217, no_hdr0=''),
  10569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10569, CurrentID=0, Status='live', Symbol='SLU7', Aliases='9G8, hSlu7', description='SLU7 homolog, splicing factor', other_designations='SLU7 splicing factor homolog|splicing factor|step II splicing factor SLU7|zinc knuckle motif containing', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160401641, end_position_on_the_genomic_accession=160419161, orientation='minus', exon_count=17, OMIM=605974, no_hdr0=''),
  10570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10570, CurrentID=0, Status='live', Symbol='DPYSL4', Aliases='CRMP3, DRP-4, ULIP4', description='dihydropyrimidinase like 4', other_designations='CRMP-3|ULIP-4|UNC33-like phosphoprotein 4|collapsin response mediator protein 3|dihydropyrimidinase-like 4', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132184987, end_position_on_the_genomic_accession=132205776, orientation='plus', exon_count=16, OMIM=608407, no_hdr0=''),
  100270841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270841, CurrentID=0, Status='live', Symbol='RPL5P24', Aliases='RPL5_9_935', description='ribosomal protein L5 pseudogene 24', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102923313, end_position_on_the_genomic_accession=102924329, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10572, CurrentID=0, Status='live', Symbol='SIVA1', Aliases='CD27BP, SIVA, Siva-1, Siva-2', description='SIVA1, apoptosis-inducing factor', other_designations='CD27-binding (Siva) protein', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104752420, end_position_on_the_genomic_accession=104759659, orientation='plus', exon_count=5, OMIM=605567, no_hdr0=''),
  10573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10573, CurrentID=0, Status='live', Symbol='MRPL28', Aliases='MAAT1, p15', description='mitochondrial ribosomal protein L28', other_designations='L28mt|MRP-L28|melanoma antigen p15|melanoma-associated antigen recognised by cytotoxic T lymphocytes|melanoma-associated antigen recognized by T lymphocytes|melanoma-associated antigen recognized by T-lymphocytes', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=367384, end_position_on_the_genomic_accession=370569, orientation='minus', exon_count=6, OMIM=604853, no_hdr0=''),
  10574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10574, CurrentID=0, Status='live', Symbol='CCT7', Aliases='CCTETA, CCTH, NIP7-1, TCP1ETA', description='chaperonin containing TCP1, subunit 7 (eta)', other_designations='CCT-eta|HIV-1 Nef interacting protein|HIV-1 Nef-interacting protein|TCP-1-eta|chaperonin containing t-complex polypeptide 1, eta subunit', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73233425, end_position_on_the_genomic_accession=73253022, orientation='plus', exon_count=14, OMIM=605140, no_hdr0=''),
  10575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10575, CurrentID=0, Status='live', Symbol='CCT4', Aliases='CCT-DELTA, Cctd, SRB', description='chaperonin containing TCP1, subunit 4 (delta)', other_designations='TCP-1-delta|chaperonin containing t-complex polypeptide 1, delta subunit|stimulator of TAR RNA-binding', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61868127, end_position_on_the_genomic_accession=61888671, orientation='minus', exon_count=14, OMIM=605142, no_hdr0=''),
  10576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10576, CurrentID=0, Status='live', Symbol='CCT2', Aliases='99D8.1, CCT-beta, CCTB, HEL-S-100n, PRO1633, TCP-1-beta', description='chaperonin containing TCP1, subunit 2 (beta)', other_designations='T-complex protein 1, beta subunit|chaperonin containing t-complex polypeptide 1, beta subunit|chaperonin containing t-complex polypeptide 1, subunit 2|epididymis secretory sperm binding protein Li 100n', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69585428, end_position_on_the_genomic_accession=69601577, orientation='plus', exon_count=17, OMIM=605139, no_hdr0=''),
  10577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10577, CurrentID=0, Status='live', Symbol='NPC2', Aliases='EDDM1, HE1', description='Niemann-Pick disease, type C2', other_designations='Niemann-Pick disease type C2 protein|epididymal protein 1|human epididymis-specific protein 1|tissue-specific secretory protein', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74479940, end_position_on_the_genomic_accession=74493381, orientation='minus', exon_count=5, OMIM=601015, no_hdr0=''),
  10578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10578, CurrentID=0, Status='live', Symbol='GNLY', Aliases='D2S69E, LAG-2, LAG2, NKG5, TLA519', description='granulysin', other_designations='T-cell activation protein 519|T-lymphocyte activation gene 519|lymphocyte-activation gene 2|lymphokine LAG-2', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85694291, end_position_on_the_genomic_accession=85698854, orientation='plus', exon_count=8, OMIM=188855, no_hdr0=''),
  10579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10579, CurrentID=0, Status='live', Symbol='TACC2', Aliases='AZU-1, ECTACC', description='transforming, acidic coiled-coil containing protein 2', other_designations='anti zuai-1', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121989174, end_position_on_the_genomic_accession=122254545, orientation='plus', exon_count=30, OMIM=605302, no_hdr0=''),
  10580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10580, CurrentID=0, Status='live', Symbol='SORBS1', Aliases='CAP, FLAF2, R85FL, SH3D5, SH3P12, SORB1', description='sorbin and SH3 domain containing 1', other_designations='Fas-ligand associated factor 2|SH3 domain protein 5|c-Cbl associated protein|ponsin', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95311773, end_position_on_the_genomic_accession=95561420, orientation='minus', exon_count=41, OMIM=605264, no_hdr0=''),
  10581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10581, CurrentID=0, Status='live', Symbol='IFITM2', Aliases='1-8D, DSPA2c', description='interferon induced transmembrane protein 2', other_designations='dispanin subfamily A member 2c|interferon-inducible protein 1-8D', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=308107, end_position_on_the_genomic_accession=309410, orientation='plus', exon_count=2, OMIM=605578, no_hdr0=''),
  23609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23609, CurrentID=0, Status='live', Symbol='MKRN2', Aliases='HSPC070, RNF62', description='makorin ring finger protein 2', other_designations='RING finger protein 62|makorin RING zinc-finger protein 2', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12557014, end_position_on_the_genomic_accession=12583713, orientation='plus', exon_count=8, OMIM=608426, no_hdr0=''),
  10584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10584, CurrentID=0, Status='live', Symbol='COLEC10', Aliases='CLL1', description='collectin subfamily member 10', other_designations='CL-34|collectin 34|collectin liver 1|collectin liver protein 1|collectin sub-family member 10 (C-type lectin)|collectin-34', map_location='8q23-q24.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118952240, end_position_on_the_genomic_accession=119108455, orientation='plus', exon_count=9, OMIM=607620, no_hdr0=''),
  10585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10585, CurrentID=0, Status='live', Symbol='POMT1', Aliases='LGMD2K, MDDGA1, MDDGB1, MDDGC1, RT', description='protein-O-mannosyltransferase 1', other_designations='dolichyl-phosphate-mannose--protein mannosyltransferase 1', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131502895, end_position_on_the_genomic_accession=131523806, orientation='plus', exon_count=21, OMIM=607423, no_hdr0=''),
  10586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10586, CurrentID=0, Status='live', Symbol='MAB21L2', Aliases='MCOPS14', description='mab-21-like 2 (C. elegans)', other_designations='', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150581925, end_position_on_the_genomic_accession=150584693, orientation='plus', exon_count=1, OMIM=604357, no_hdr0=''),
  10587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10587, CurrentID=0, Status='live', Symbol='TXNRD2', Aliases='SELZ, TR, TR-BETA, TR3, TRXR2', description='thioredoxin reductase 2', other_designations='selenoprotein Z|thioredoxin reductase 3|thioredoxin reductase TR3|thioredoxin reductase beta', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19875518, end_position_on_the_genomic_accession=19941992, orientation='minus', exon_count=19, OMIM=606448, no_hdr0=''),
  10588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10588, CurrentID=0, Status='live', Symbol='MTHFS', Aliases='HsT19268', description='5,10-methenyltetrahydrofolate synthetase (5-formyltetrahydrofolate cyclo-ligase)', other_designations='methenyl-THF synthetase', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79843547, end_position_on_the_genomic_accession=79897285, orientation='minus', exon_count=4, OMIM=604197, no_hdr0=''),
  10589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10589, CurrentID=0, Status='live', Symbol='DRAP1', Aliases='NC2-alpha', description='DR1-associated protein 1 (negative cofactor 2 alpha)', other_designations='negative co-factor 2-alpha|negative cofactor 2 alpha|negative cofactor 2-alpha', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65919257, end_position_on_the_genomic_accession=65921577, orientation='plus', exon_count=7, OMIM=602289, no_hdr0=''),
  10590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10590, CurrentID=0, Status='live', Symbol='SCGN', Aliases='CALBL, DJ501N12.8, SECRET, SEGN, setagin', description='secretagogin, EF-hand calcium binding protein', other_designations='', map_location='6p22.3-p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25652201, end_position_on_the_genomic_accession=25701783, orientation='plus', exon_count=11, OMIM=609202, no_hdr0=''),
  10591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10591, CurrentID=0, Status='live', Symbol='DNPH1', Aliases='C6orf108, RCL, dJ330M21.3', description="2'-deoxynucleoside 5'-phosphate N-hydrolase 1", other_designations="c-Myc-responsive protein RCL|c-Myc-responsive protein Rcl|deoxyribonucleoside 5'-monophosphate N-glycosidase|putative c-Myc-responsive", map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43225629, end_position_on_the_genomic_accession=43229473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10592, CurrentID=0, Status='live', Symbol='SMC2', Aliases='CAP-E, CAPE, SMC-2L1, SMC2', description='structural maintenance of chromosomes 2', other_designations='SMC protein 2|SMC2 (structural maintenance of chromosomes 2, yeast)-like 1|XCAP-E homolog|chromosome-associated protein E|structural maintenance of chromosomes (SMC) family member, chromosome-associated protein E', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104093848, end_position_on_the_genomic_accession=104141419, orientation='plus', exon_count=29, OMIM=605576, no_hdr0=''),
  10594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10594, CurrentID=0, Status='live', Symbol='PRPF8', Aliases='HPRP8, PRP8, PRPC8, RP13, SNRNP220', description='pre-mRNA processing factor 8', other_designations='220 kDa U5 snRNP-specific protein|PRP8 homolog|PRP8 pre-mRNA processing factor 8 homolog|U5 snRNP-specific protein (220 kD), ortholog of S. cerevisiae Prp8p|apoptosis-regulated protein 1|apoptosis-regulated protein 2|p220|precursor mRNA processing protein|splicing factor Prp8', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1650629, end_position_on_the_genomic_accession=1684882, orientation='minus', exon_count=43, OMIM=607300, no_hdr0=''),
  10595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10595, CurrentID=0, Status='live', Symbol='ERN2', Aliases='IRE1-BETA, IRE1b, IRE2p, hIRE2p', description='endoplasmic reticulum to nucleus signaling 2', other_designations='ER to nucleus signalling 2|IRE1 beta|IRE1, S. cerevisiae, homolog of|endoplasmic reticulum to nucleus signalling 2|endoplasmic reticulum-to-nucleus signaling 2|inositol-requiring 1 beta|inositol-requiring protein 2', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23690304, end_position_on_the_genomic_accession=23713500, orientation='minus', exon_count=24, OMIM=604034, no_hdr0=''),
  10597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10597, CurrentID=0, Status='live', Symbol='TRAPPC2B', Aliases='MIP-2A, SEDLP, SEDLP1, TRAPPC2P1', description='trafficking protein particle complex 2B', other_designations='MBP-1 interacting protein-2A|spondyloepiphyseal dysplasia, late, pseudogene|trafficking protein particle complex 2 pseudogene 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57363511, end_position_on_the_genomic_accession=57365353, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10598, CurrentID=0, Status='live', Symbol='AHSA1', Aliases='AHA1, C14orf3, p38', description='AHA1, activator of heat shock 90kDa protein ATPase homolog 1 (yeast)', other_designations='', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77457875, end_position_on_the_genomic_accession=77469472, orientation='plus', exon_count=10, OMIM=608466, no_hdr0=''),
  10599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10599, CurrentID=0, Status='live', Symbol='SLCO1B1', Aliases='HBLRR, LST-1, LST1, OATP-C, OATP1B1, OATP2, OATPC, SLC21A6', description='solute carrier organic anion transporter family member 1B1', other_designations='OATP-2|liver-specific organic anion transporter 1|sodium-independent organic anion-transporting polypeptide 2|solute carrier family 21 (organic anion transporter), member 6|solute carrier organic anion transporter family, member 1B1', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21131194, end_position_on_the_genomic_accession=21239796, orientation='plus', exon_count=15, OMIM=604843, no_hdr0=''),
  10600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10600, CurrentID=0, Status='live', Symbol='USP16', Aliases='UBP-M, UBPM', description='ubiquitin specific peptidase 16', other_designations='deubiquitinating enzyme 16|ubiquitin specific protease 16|ubiquitin thioesterase 16|ubiquitin thiolesterase 16|ubiquitin-processing protease UBP-M|ubiquitin-specific processing protease 16|ubiquitin-specific-processing protease 16', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29024617, end_position_on_the_genomic_accession=29054488, orientation='plus', exon_count=19, OMIM=604735, no_hdr0=''),
  10602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10602, CurrentID=0, Status='live', Symbol='CDC42EP3', Aliases='BORG2, CEP3, UB1', description='CDC42 effector protein (Rho GTPase binding) 3', other_designations='CRIB-containing BORG2 protein|MSE55-related Cdc42-binding protein|MSE55-related protein|binder of Rho GTPases 2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37641882, end_position_on_the_genomic_accession=37672535, orientation='minus', exon_count=4, OMIM=606133, no_hdr0=''),
  10603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10603, CurrentID=0, Status='live', Symbol='SH2B2', Aliases='APS', description='SH2B adaptor protein 2', other_designations='SH2 and PH domain-containing adapter protein APS|adapter protein with pleckstrin homology and Src homology 2 domains', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102285059, end_position_on_the_genomic_accession=102321711, orientation='plus', exon_count=12, OMIM=605300, no_hdr0=''),
  10605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10605, CurrentID=0, Status='live', Symbol='PAIP1', Aliases='', description='poly(A) binding protein interacting protein 1', other_designations='PABC1-interacting protein 1|PABP-interacting protein 1|PAIP-1|poly(A)-binding protein-interacting protein 1|polyadenylate binding protein-interacting protein 1', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43526267, end_position_on_the_genomic_accession=43557419, orientation='minus', exon_count=13, OMIM=605184, no_hdr0=''),
  10606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10606, CurrentID=0, Status='live', Symbol='PAICS', Aliases='ADE2, ADE2H1, AIRC, PAIS', description='phosphoribosylaminoimidazole carboxylase; phosphoribosylaminoimidazolesuccinocarboxamide synthase', other_designations='AIR carboxylase|SAICAR synthetase|multifunctional protein ADE2H1|phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56435749, end_position_on_the_genomic_accession=56461368, orientation='plus', exon_count=11, OMIM=172439, no_hdr0=''),
  10607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10607, CurrentID=0, Status='live', Symbol='TBL3', Aliases='SAZD, UTP13', description='transducin (beta)-like 3', other_designations='WD repeat-containing protein SAZD|WD-repeat protein SAZD', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1972063, end_position_on_the_genomic_accession=1978750, orientation='plus', exon_count=22, OMIM=605915, no_hdr0=''),
  10608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10608, CurrentID=0, Status='live', Symbol='MXD4', Aliases='MAD4, MST149, MSTP149, bHLHc12', description='MAX dimerization protein 4', other_designations='Mad4 homolog|class C basic helix-loop-helix protein 12|max dimerizer 4|max-associated protein 4|max-interacting transcriptional repressor MAD4', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2247432, end_position_on_the_genomic_accession=2262143, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  10609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10609, CurrentID=0, Status='live', Symbol='P3H4', Aliases='LEPREL4, NO55, NOL55, SC65', description='prolyl 3-hydroxylase family member 4 (non-enzymatic)', other_designations='leprecan-like 4|leprecan-like protein 4|nucleolar autoantigen (55kD)|nucleolar autoantigen No55|nucleolar autoantigen, 55kDa', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41801952, end_position_on_the_genomic_accession=41812616, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  10610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10610, CurrentID=0, Status='live', Symbol='ST6GALNAC2', Aliases='SAITL1, SIAT7, SIAT7B, SIATL1, ST6GalNAII, STHM', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 2', other_designations='(alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide B|(alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase|SIAT7-B|ST6GalNAc II|ST6GalNAcII|galNAc alpha-2,6-sialyltransferase II|sialyltransferase 7 ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide B|sialyltransferase 7 ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) B|sialyltransferase 7B|sialyltransferase-like 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76565379, end_position_on_the_genomic_accession=76586066, orientation='minus', exon_count=11, OMIM=610137, no_hdr0=''),
  10611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10611, CurrentID=0, Status='live', Symbol='PDLIM5', Aliases='ENH, ENH1, L9, LIM', description='PDZ and LIM domain 5', other_designations='enigma homolog|enigma-like LIM domain protein|enigma-like PDZ and LIM domains protein', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94451857, end_position_on_the_genomic_accession=94668227, orientation='plus', exon_count=22, OMIM=605904, no_hdr0=''),
  10612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10612, CurrentID=0, Status='live', Symbol='TRIM3', Aliases='BERP, HAC1, RNF22, RNF97', description='tripartite motif containing 3', other_designations='RING finger protein 97|brain expressed ring finger|brain-expressed RING finger protein|ring finger protein 22|tripartite motif protein TRIM3|tripartite motif-containing 3', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6448613, end_position_on_the_genomic_accession=6474459, orientation='minus', exon_count=15, OMIM=605493, no_hdr0=''),
  10613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10613, CurrentID=0, Status='live', Symbol='ERLIN1', Aliases='C10orf69, Erlin-1, KE04, KEO4, SPFH1, SPG62', description='ER lipid raft associated 1', other_designations='Band_7 23-211 Keo4 (Interim) similar to C.elegans protein C42C1.9|SPFH domain family, member 1|SPFH domain-containing protein 1|endoplasmic reticulum lipid raft-associated protein 1|stomatin-prohibitin-flotillin-HflC/K domain-containing protein 1', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100150090, end_position_on_the_genomic_accession=100186057, orientation='minus', exon_count=12, OMIM=611604, no_hdr0=''),
  10614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10614, CurrentID=0, Status='live', Symbol='HEXIM1', Aliases='CLP1, EDG1, HIS1, MAQ1', description='hexamethylene bis-acetamide inducible 1', other_designations='HMBA-inducible|cardiac lineage protein 1|estrogen down-regulated gene 1 protein|hexamethylene bis-acetamide-inducible protein 1|hexamethylene bisacetamide-inducible protein|hexamethylene-bis-acetamide-inducible transcript 1|hexamthylene bis-acetamide inducible 1|menage a quatre 1|menage a quatre protein 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45147317, end_position_on_the_genomic_accession=45152101, orientation='plus', exon_count=1, OMIM=607328, no_hdr0=''),
  10615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10615, CurrentID=0, Status='live', Symbol='SPAG5', Aliases='DEEPEST, MAP126, hMAP126', description='sperm associated antigen 5', other_designations='astrin|mitotic spindle associated protein p126|mitotic spindle coiled-coil related protein|mitotic spindle-associated protein p126', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28577565, end_position_on_the_genomic_accession=28599038, orientation='minus', exon_count=24, OMIM=615562, no_hdr0=''),
  10616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10616, CurrentID=0, Status='live', Symbol='RBCK1', Aliases='C20orf18, HOIL-1, HOIL1, PBMEI, PGBM1, RBCK2, RNF54, UBCE7IP3, XAP3, XAP4, ZRANB4', description='RanBP-type and C3HC4-type zinc finger containing 1', other_designations='HBV associated factor 4|HBV-associated factor 4|RBCC protein interacting with PKC1|RING finger protein 54|heme-oxidized IRP2 ubiquitin ligase 1|hepatitis B virus X-associated protein 4|ubiquitin conjugating enzyme 7 interacting protein 3|ubiquitin-conjugating enzyme 7-interacting protein 3', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=408065, end_position_on_the_genomic_accession=430966, orientation='plus', exon_count=14, OMIM=610924, no_hdr0=''),
  10617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10617, CurrentID=0, Status='live', Symbol='STAMBP', Aliases='AMSH, MICCAP', description='STAM binding protein', other_designations='associated molecule with the SH3 domain of STAM|endosome-associated ubiquitin isopeptidase', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73828916, end_position_on_the_genomic_accession=73873656, orientation='plus', exon_count=16, OMIM=606247, no_hdr0=''),
  10618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10618, CurrentID=0, Status='live', Symbol='TGOLN2', Aliases='TGN38, TGN46, TGN48, TGN51, TTGN2', description='trans-golgi network protein 2', other_designations='TGN38 homolog|trans-Golgi network protein (46, 48, 51kD isoforms)|trans-Golgi network protein TGN51', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85318018, end_position_on_the_genomic_accession=85328296, orientation='minus', exon_count=5, OMIM=603062, no_hdr0=''),
  285759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285759, CurrentID=0, Status='live', Symbol='FLJ34503', Aliases='', description='uncharacterized FLJ34503', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113904387, end_position_on_the_genomic_accession=113921642, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10620, CurrentID=0, Status='live', Symbol='ARID3B', Aliases='BDP, DRIL2', description='AT-rich interaction domain 3B', other_designations='ARID domain-containing protein 3B|AT rich interactive domain 3B (BRIGHT- like)|AT rich interactive domain 3B (BRIGHT-like)|bright and dead ringer protein|bright-like protein', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74541177, end_position_on_the_genomic_accession=74598131, orientation='plus', exon_count=9, OMIM=612457, no_hdr0=''),
  10621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10621, CurrentID=0, Status='live', Symbol='POLR3F', Aliases='RPC39, RPC6', description='polymerase (RNA) III (DNA directed) polypeptide F, 39 kDa', other_designations='RNA polymerase III C39 subunit|RNA polymerase III subunit C6', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18467389, end_position_on_the_genomic_accession=18484648, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  10622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10622, CurrentID=0, Status='live', Symbol='POLR3G', Aliases='RPC32, RPC7', description='polymerase (RNA) III (DNA directed) polypeptide G (32kD)', other_designations='DNA-directed RNA polymerase III 32 kDa polypeptide|DNA-directed RNA polymerase III subunit G|RNA polymerase III 32 kDa subunit|RNA polymerase III subunit C7', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90474229, end_position_on_the_genomic_accession=90514677, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  10623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10623, CurrentID=0, Status='live', Symbol='POLR3C', Aliases='RPC3, RPC62', description='polymerase (RNA) III (DNA directed) polypeptide C (62kD)', other_designations='DNA-directed III 62 kDa polypeptide|DNA-directed RNA polymerase III subunit C|RNA polymerase III 62 kDa subunit|RNA polymerase III subunit C3', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145824069, end_position_on_the_genomic_accession=145844402, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  10625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10625, CurrentID=0, Status='live', Symbol='IVNS1ABP', Aliases='FLARA3, HSPC068, KLHL39, ND1, NS-1, NS1-BP, NS1BP', description='influenza virus NS1A binding protein', other_designations='NCX downstream gene 1|NS1-binding protein|aryl hydrocarbon receptor-associated 3|aryl hydrocarbon receptor-associated protein 3|kelch-like family member 39', map_location='1q25.1-q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185296388, end_position_on_the_genomic_accession=185317343, orientation='minus', exon_count=16, OMIM=609209, no_hdr0=''),
  10626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10626, CurrentID=0, Status='live', Symbol='TRIM16', Aliases='EBBP', description='tripartite motif containing 16', other_designations='estrogen-responsive B box protein|tripartite motif-containing 16', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15627966, end_position_on_the_genomic_accession=15682879, orientation='minus', exon_count=9, OMIM=609505, no_hdr0=''),
  10627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10627, CurrentID=0, Status='live', Symbol='MYL12A', Aliases='HEL-S-24, MLC-2B, MLCB, MRCL3, MRLC3, MYL2B', description='myosin light chain 12A', other_designations='epididymis secretory protein Li 24|myosin RLC|myosin regulatory light chain 2, nonsarcomeric|myosin regulatory light chain 3|myosin regulatory light chain MRLC3|myosin, light chain 12A, regulatory, non-sarcomeric|myosin, light polypeptide, regulatory, non-sarcomeric (20kD)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3247482, end_position_on_the_genomic_accession=3256237, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10628, CurrentID=0, Status='live', Symbol='TXNIP', Aliases='EST01027, HHCPA78, THIF, VDUP1', description='thioredoxin interacting protein', other_designations='thioredoxin binding protein 2|thioredoxin-binding protein 2|upregulated by 1,25-dihydroxyvitamin D-3|vitamin D3 up-regulated protein 1', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145992435, end_position_on_the_genomic_accession=145996631, orientation='minus', exon_count=8, OMIM=606599, no_hdr0=''),
  10629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10629, CurrentID=0, Status='live', Symbol='TAF6L', Aliases='PAF65A', description='TATA-box binding protein associated factor 6 like', other_designations='PAF65-alpha|PCAF-associated factor 65 alpha|PCAF-associated factor 65-alpha|TAF6-like RNA polymerase II, p300/CBP-associated factor (PCAF)-associated factor, 65kDa|TATA box binding protein associated factor 6 like|p300/CBP-associated factor (PCAF)-associated factor 65', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62771303, end_position_on_the_genomic_accession=62787342, orientation='plus', exon_count=11, OMIM=602946, no_hdr0=''),
  10630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10630, CurrentID=0, Status='live', Symbol='PDPN', Aliases='AGGRUS, GP36, GP40, Gp38, HT1A-1, OTS8, PA2.26, T1A, T1A-2, T1A2, TI1A', description='podoplanin', other_designations='PA2.26 antigen|T1-alpha|glycoprotein 36|glycoprotein, 36-KD|hT1alpha-1|hT1alpha-2|lung type I cell membrane associated glycoprotein|lung type-I cell membrane-associated glycoprotein (T1A-2)', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13583757, end_position_on_the_genomic_accession=13617957, orientation='plus', exon_count=8, OMIM=608863, no_hdr0=''),
  10631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10631, CurrentID=0, Status='live', Symbol='POSTN', Aliases='OSF-2, OSF2, PDLPOSTN, PN', description='periostin, osteoblast specific factor', other_designations='osteoblast specific factor 2 (fasciclin I-like)|periodontal ligament-specific periostin', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37562582, end_position_on_the_genomic_accession=37598844, orientation='minus', exon_count=23, OMIM=608777, no_hdr0=''),
  10632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10632, CurrentID=0, Status='live', Symbol='ATP5L', Aliases='ATP5JG', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G', other_designations='ATP synthase g chain, mitochondrial|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit G|ATP synthase, H+ transporting, mitochondrial F1F0, subunit g|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit G|ATPase subunit G|F1F0-type ATP synthase subunit g|F1Fo-ATP synthase complex Fo membrane domain g subunit', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118401389, end_position_on_the_genomic_accession=118409847, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10633, CurrentID=0, Status='live', Symbol='RASL10A', Aliases='RRP22', description='RAS-like family 10 member A', other_designations='RAS-like, family 10, member A|ras-like protein RRP22|ras-related protein on chromosome 22', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29312933, end_position_on_the_genomic_accession=29319618, orientation='minus', exon_count=6, OMIM=602220, no_hdr0=''),
  10634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10634, CurrentID=0, Status='live', Symbol='GAS2L1', Aliases='GAR22', description='growth arrest specific 2 like 1', other_designations='GAS2-related protein on chromosome 22|growth arrest-specific 2 like 1|growth arrest-specific protein 2-like 1', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29306618, end_position_on_the_genomic_accession=29312789, orientation='plus', exon_count=7, OMIM=602128, no_hdr0=''),
  10635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10635, CurrentID=0, Status='live', Symbol='RAD51AP1', Aliases='PIR51', description='RAD51 associated protein 1', other_designations='RAD51-interacting protein', map_location='12p13.2-p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4538746, end_position_on_the_genomic_accession=4560047, orientation='plus', exon_count=10, OMIM=603070, no_hdr0=''),
  10636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10636, CurrentID=0, Status='live', Symbol='RGS14', Aliases='', description='regulator of G-protein signaling 14', other_designations='regulator of G-protein signalling 14', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177357711, end_position_on_the_genomic_accession=177372601, orientation='plus', exon_count=15, OMIM=602513, no_hdr0=''),
  10637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10637, CurrentID=0, Status='live', Symbol='LEFTY1', Aliases='LEFTB, LEFTYB', description='left-right determination factor 1', other_designations='left-right determination factor B|left-right determination, factor B|protein lefty-1|protein lefty-B', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225886282, end_position_on_the_genomic_accession=225889146, orientation='minus', exon_count=4, OMIM=603037, no_hdr0=''),
  10638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10638, CurrentID=0, Status='live', Symbol='SPHAR', Aliases='', description='S-phase response (cyclin related)', other_designations='S-phase response (cyclin-related)|S-phase response protein', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229304382, end_position_on_the_genomic_accession=229305504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100533997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533997, CurrentID=0, Status='live', Symbol='MAGEA10-MAGEA5', Aliases='CT1.5', description='MAGEA10-MAGEA5 readthrough', other_designations='MAGE-5 antigen|cancer/testis antigen 1.5', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152114049, end_position_on_the_genomic_accession=152138578, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10640, CurrentID=0, Status='live', Symbol='EXOC5', Aliases='HSEC10, PRO1912, SEC10, SEC10L1, SEC10P', description='exocyst complex component 5', other_designations='SEC10-like 1|exocyst complex component Sec10', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57202476, end_position_on_the_genomic_accession=57268899, orientation='minus', exon_count=18, OMIM=604469, no_hdr0=''),
  10641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10641, CurrentID=0, Status='live', Symbol='NPRL2', Aliases='NPR2, NPR2L, TUSC4', description='NPR2-like, GATOR1 complex subunit', other_designations='2810446G01Rik|G21 protein|NPR2-like protein|gene 21 protein|homologous to yeast nitrogen permease (candidate tumor suppressor)|tumor suppressor candidate 4', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50347487, end_position_on_the_genomic_accession=50351055, orientation='minus', exon_count=12, OMIM=607072, no_hdr0=''),
  10642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10642, CurrentID=0, Status='live', Symbol='IGF2BP1', Aliases='CRD-BP, CRDBP, IMP-1, IMP1, VICKZ1, ZBP1', description='insulin like growth factor 2 mRNA binding protein 1', other_designations='IGF-II mRNA-binding protein 1|IGF2 mRNA-binding protein 1|VICKZ family member 1|ZBP-1|coding region determinant-binding protein|insulin-like growth factor 2 mRNA binding protein 1 deltaN CRDBP|zipcode-binding protein 1', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48997412, end_position_on_the_genomic_accession=49056145, orientation='plus', exon_count=16, OMIM=608288, no_hdr0=''),
  10643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10643, CurrentID=0, Status='live', Symbol='IGF2BP3', Aliases='CT98, IMP-3, IMP3, KOC, KOC1, VICKZ3', description='insulin like growth factor 2 mRNA binding protein 3', other_designations='IGF-II mRNA-binding protein 3|IGF2 mRNA-binding protein 3|KH domain containing protein overexpressed in cancer|VICKZ family member 3|cancer/testis antigen 98|insulin-like growth factor 2 mRNA binding protein 3', map_location='7p11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23310209, end_position_on_the_genomic_accession=23470527, orientation='minus', exon_count=18, OMIM=608259, no_hdr0=''),
  10644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10644, CurrentID=0, Status='live', Symbol='IGF2BP2', Aliases='IMP-2, IMP2, VICKZ2', description='insulin like growth factor 2 mRNA binding protein 2', other_designations='IGF-II mRNA-binding protein 2|IGF2 mRNA-binding protein 2|VICKZ family member 2|insulin-like growth factor 2 mRNA binding protein 2', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185643131, end_position_on_the_genomic_accession=185825039, orientation='minus', exon_count=23, OMIM=608289, no_hdr0=''),
  10645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10645, CurrentID=0, Status='live', Symbol='CAMKK2', Aliases='CAMKK, CAMKKB', description='calcium/calmodulin-dependent protein kinase kinase 2, beta', other_designations='CAMKK beta protein|caM-KK 2|caM-KK beta|caM-kinase kinase 2|caM-kinase kinase beta|calcium/calmodulin-dependent protein kinase beta', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121237692, end_position_on_the_genomic_accession=121298308, orientation='minus', exon_count=19, OMIM=615002, no_hdr0=''),
  10647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10647, CurrentID=0, Status='live', Symbol='SCGB1D2', Aliases='LIPB, LPHB, LPNB', description='secretoglobin, family 1D, member 2', other_designations='lipophilin-B|prostatein-like lipophilin B', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62241630, end_position_on_the_genomic_accession=62244812, orientation='plus', exon_count=4, OMIM=615061, no_hdr0=''),
  10648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10648, CurrentID=0, Status='live', Symbol='SCGB1D1', Aliases='LIPA, LPHA, LPNA', description='secretoglobin, family 1D, member 1', other_designations='lipophilin A (uteroglobin family member)|lipophilin-A|prostatein-like lipophilin A', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62190238, end_position_on_the_genomic_accession=62193537, orientation='plus', exon_count=3, OMIM=615060, no_hdr0=''),
  10650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10650, CurrentID=0, Status='live', Symbol='PRELID3A', Aliases='C18orf43, HFL-EDDG1, SLMO1', description='PRELI domain containing 3A', other_designations='erythroid differentiation and denucleation factor 1|protein slowmo homolog 1|slowmo homolog 1', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12407896, end_position_on_the_genomic_accession=12432238, orientation='plus', exon_count=11, OMIM=616545, no_hdr0=''),
  10651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10651, CurrentID=0, Status='live', Symbol='MTX2', Aliases='', description='metaxin 2', other_designations='mitochondrial outer membrane import complex protein 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176269395, end_position_on_the_genomic_accession=176338025, orientation='plus', exon_count=12, OMIM=608555, no_hdr0=''),
  10652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10652, CurrentID=0, Status='live', Symbol='YKT6', Aliases='', description='YKT6 v-SNARE homolog (S. cerevisiae)', other_designations='R-SNARE|SNARE protein Ykt6|YKT6 v-SNARE protein|YKT6, S. cerevisiae, homolog of', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44200968, end_position_on_the_genomic_accession=44214294, orientation='plus', exon_count=8, OMIM=606209, no_hdr0=''),
  10653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10653, CurrentID=0, Status='live', Symbol='SPINT2', Aliases='DIAR3, HAI-2, HAI2, Kop, PB', description='serine peptidase inhibitor, Kunitz type, 2', other_designations='hepatocyte growth factor activator inhibitor type 2|placental bikunin|serine protease inhibitor, Kunitz type, 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38264458, end_position_on_the_genomic_accession=38292614, orientation='plus', exon_count=7, OMIM=605124, no_hdr0=''),
  10654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10654, CurrentID=0, Status='live', Symbol='PMVK', Aliases='HUMPMKI, PMK, PMKA, PMKASE', description='phosphomevalonate kinase', other_designations='hPMK', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154924732, end_position_on_the_genomic_accession=154937310, orientation='minus', exon_count=6, OMIM=607622, no_hdr0=''),
  10655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10655, CurrentID=0, Status='live', Symbol='DMRT2', Aliases='', description='doublesex and mab-3 related transcription factor 2', other_designations='DSXL-2|doublesex-like 2 protein|terra-like protein', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1003242, end_position_on_the_genomic_accession=1057554, orientation='plus', exon_count=12, OMIM=604935, no_hdr0=''),
  10656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10656, CurrentID=0, Status='live', Symbol='KHDRBS3', Aliases='Etle, SALP, SLM-2, SLM2, T-STAR, TSTAR, etoile', description='KH domain containing, RNA binding, signal transduction associated 3', other_designations='RNA-binding protein T-Star|Sam68-like phosphotyrosine protein, T-STAR|sam68-like mammalian protein 2', map_location='8q24.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=135457461, end_position_on_the_genomic_accession=135656516, orientation='plus', exon_count=15, OMIM=610421, no_hdr0=''),
  10657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10657, CurrentID=0, Status='live', Symbol='KHDRBS1', Aliases='Sam68, p62, p68', description='KH domain containing, RNA binding, signal transduction associated 1', other_designations='GAP-associated tyrosine phosphoprotein p62 (Sam68)|p21 Ras GTPase-activating protein-associated p62|src-associated in mitosis 68 kDa protein', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32013694, end_position_on_the_genomic_accession=32060859, orientation='plus', exon_count=11, OMIM=602489, no_hdr0=''),
  10658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10658, CurrentID=0, Status='live', Symbol='CELF1', Aliases='BRUNOL2, CUG-BP, CUGBP, CUGBP1, EDEN-BP, NAB50, NAPOR, hNab50', description='CUGBP, Elav-like family member 1', other_designations='50 kDa nuclear polyadenylated RNA-binding protein|CELF-1|CUG RNA-binding protein|CUG triplet repeat RNA-binding protein 1|CUG triplet repeat, RNA binding protein 1|CUG triplet repeat, RNA-binding protein 1|CUG-BP- and ETR-3-like factor 1|CUG-BP1|EDEN-BP homolog|RNA-binding protein BRUNOL-2|bruno-like 2|bruno-like protein 2|deadenylation factor CUG-BP|embryo deadenylation element binding protein|embryo deadenylation element-binding protein homolog|nuclear polyadenylated RNA-binding protein, 50-kD', map_location='11p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47465937, end_position_on_the_genomic_accession=47565569, orientation='minus', exon_count=20, OMIM=601074, no_hdr0=''),
  10659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10659, CurrentID=0, Status='live', Symbol='CELF2', Aliases='BRUNOL3, CUGBP2, ETR-3, ETR3, NAPOR', description='CUGBP, Elav-like family member 2', other_designations='CELF-2|CUG triplet repeat RNA-binding protein 2|CUG-BP- and ETR-3-like factor 2|CUG-BP2|ELAV-type RNA-binding protein 3|RNA-binding protein BRUNOL-3|bruno-like protein 3|neuroblastoma apoptosis-related RNA-binding protein', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11004641, end_position_on_the_genomic_accession=11336675, orientation='plus', exon_count=17, OMIM=602538, no_hdr0=''),
  10660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10660, CurrentID=0, Status='live', Symbol='LBX1', Aliases='HPX-6, HPX6H, homeobox, LBX1', description='ladybird homeobox 1', other_designations='lady bird-like homeobox|ladybird homeobox homolog 1|ladybird homeobox protein homolog 1|transcription factor similar to D. melanogaster homeodomain protein lady bird late', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101226212, end_position_on_the_genomic_accession=101230327, orientation='minus', exon_count=3, OMIM=604255, no_hdr0=''),
  10661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10661, CurrentID=0, Status='live', Symbol='KLF1', Aliases='CDAN4, EKLF, HBFQTL6, INLU', description='Kruppel-like factor 1 (erythroid)', other_designations='erythroid Kruppel-like factor|erythroid krueppel-like transcription factor|erythroid-specific transcription factor EKLF|monoclonal antibody A3D8', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12884422, end_position_on_the_genomic_accession=12887203, orientation='minus', exon_count=3, OMIM=600599, no_hdr0=''),
  10663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10663, CurrentID=0, Status='live', Symbol='CXCR6', Aliases='BONZO, CD186, STRL33, TYMSTR', description='chemokine (C-X-C motif) receptor 6', other_designations='CDw186|CXC-R6|CXCR-6|G protein-coupled receptor|G-protein coupled receptor STRL33|G-protein coupled receptor bonzo', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45940399, end_position_on_the_genomic_accession=45948354, orientation='plus', exon_count=4, OMIM=605163, no_hdr0=''),
  10664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10664, CurrentID=0, Status='live', Symbol='CTCF', Aliases='MRD21', description='CCCTC-binding factor (zinc finger protein)', other_designations='11 zinc finger transcriptional repressor|11-zinc finger protein|CTCFL paralog', map_location='16q21-q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67562407, end_position_on_the_genomic_accession=67639185, orientation='plus', exon_count=12, OMIM=604167, no_hdr0=''),
  10665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10665, CurrentID=0, Status='live', Symbol='C6orf10', Aliases='TSBP', description='chromosome 6 open reading frame 10', other_designations='testis specific basic protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32288541, end_position_on_the_genomic_accession=32372114, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  10666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10666, CurrentID=0, Status='live', Symbol='CD226', Aliases='DNAM-1, DNAM1, PTA1, TLiSA1', description='CD226 molecule', other_designations='DNAX accessory molecule 1|DNAX accessory molecule-1|T lineage-specific activation antigen 1 antigen|adhesion glycoprotein|platelet and T cell activation antigen 1', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=69859459, end_position_on_the_genomic_accession=69961787, orientation='minus', exon_count=11, OMIM=605397, no_hdr0=''),
  10667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10667, CurrentID=0, Status='live', Symbol='FARS2', Aliases='COXPD14, FARS1, HSPC320, PheRS, dJ520B18.2', description='phenylalanyl-tRNA synthetase 2, mitochondrial', other_designations='dJ236A3.1 (phenylalanine-tRNA synthetase)|dJ520B18.2 (FARS1 (phenylalanine-tRNA synthetase))|phenylalanine tRNA ligase 2, mitochondrial|phenylalanine translase|phenylalanine-tRNA synthetase 1 (mitochondrial)', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5261018, end_position_on_the_genomic_accession=5771583, orientation='plus', exon_count=16, OMIM=611592, no_hdr0=''),
  10668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10668, CurrentID=0, Status='live', Symbol='CGRRF1', Aliases='CGR19, RNF197', description='cell growth regulator with ring finger domain 1', other_designations='RING finger protein 197|cell growth regulatory gene 19 protein', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54509869, end_position_on_the_genomic_accession=54539309, orientation='plus', exon_count=7, OMIM=606138, no_hdr0=''),
  10669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10669, CurrentID=0, Status='live', Symbol='CGREF1', Aliases='CGR11', description='cell growth regulator with EF-hand domain 1', other_designations='cell growth regulator with EF hand domain 1|cell growth regulatory gene 11 protein|hydrophobestin', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27099353, end_position_on_the_genomic_accession=27119127, orientation='minus', exon_count=8, OMIM=606137, no_hdr0=''),
  10670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10670, CurrentID=0, Status='live', Symbol='RRAGA', Aliases='FIP-1, FIP1, RAGA', description='Ras-related GTP binding A', other_designations='adenovirus E3 14.7 kDa-interacting protein 1|adenovirus E3-14.7K interacting protein 1|rag A', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19049374, end_position_on_the_genomic_accession=19051025, orientation='plus', exon_count=1, OMIM=612194, no_hdr0=''),
  10671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10671, CurrentID=0, Status='live', Symbol='DCTN6', Aliases='WS-3, WS3, p27', description='dynactin subunit 6', other_designations='dynactin 6|dynactin subunit p27|novel RGD-containing protein|protein WS-3', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30156297, end_position_on_the_genomic_accession=30183639, orientation='plus', exon_count=7, OMIM=612963, no_hdr0=''),
  10672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10672, CurrentID=0, Status='live', Symbol='GNA13', Aliases='G13', description='guanine nucleotide binding protein (G protein), alpha 13', other_designations='G-protein subunit alpha-13|g alpha-13', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65009289, end_position_on_the_genomic_accession=65056802, orientation='minus', exon_count=7, OMIM=604406, no_hdr0=''),
  10673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10673, CurrentID=0, Status='live', Symbol='TNFSF13B', Aliases='BAFF, BLYS, CD257, DTL, TALL-1, TALL1, THANK, TNFSF20, ZTNF4', description='tumor necrosis factor superfamily member 13b', other_designations='ApoL related ligand TALL-1|B-cell-activating factor|B-lymphocyte stimulator|Delta4 BAFF|TNF and ApoL-related leukocyte expressed ligand 1|TNF homolog that activates apoptosis|delta BAFF|dendritic cell-derived TNF-like molecule|tumor necrosis factor (ligand) superfamily member 13b|tumor necrosis factor (ligand) superfamily, member 13b|tumor necrosis factor (ligand) superfamily, member 20|tumor necrosis factor-like protein ZTNF4', map_location='13q32-q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108269420, end_position_on_the_genomic_accession=108308484, orientation='plus', exon_count=7, OMIM=603969, no_hdr0=''),
  10674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10674, CurrentID=0, Status='live', Symbol='DURS1', Aliases='DUS', description='Duane retraction syndrome 1', other_designations="Duane's retraction syndrome 1", map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=126800, no_hdr0=''),
  10675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10675, CurrentID=0, Status='live', Symbol='CSPG5', Aliases='NGC', description='chondroitin sulfate proteoglycan 5', other_designations='acidic leucine-rich EGF-like domain-containing brain protein|chondroitin sulfate proteoglycan 5 (neuroglycan C)', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47562238, end_position_on_the_genomic_accession=47580240, orientation='minus', exon_count=6, OMIM=606775, no_hdr0=''),
  10677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10677, CurrentID=0, Status='live', Symbol='AVIL', Aliases='ADVIL, DOC6, p92', description='advillin', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57797372, end_position_on_the_genomic_accession=57819824, orientation='minus', exon_count=21, OMIM=613397, no_hdr0=''),
  10678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10678, CurrentID=0, Status='live', Symbol='B3GNT2', Aliases='B3GN-T2, B3GNT, B3GNT-2, B3GNT1, BETA3GNT, BGNT2, BGnT-2', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 2', other_designations='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 1|beta-1,3-Gn-T1|beta-1,3-Gn-T2|beta-1,3-N-acetylglucosaminyltransferase bGnT-1|beta-1,3-N-acetylglucosaminyltransferase bGnT-2|beta3Gn-T1|beta3Gn-T2', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62196127, end_position_on_the_genomic_accession=62224731, orientation='plus', exon_count=3, OMIM=605581, no_hdr0=''),
  10681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10681, CurrentID=0, Status='live', Symbol='GNB5', Aliases='GB5', description='guanine nucleotide binding protein (G protein), beta 5', other_designations='G protein, beta subunit 5L|G protein, beta-5 subunit|gbeta5|guanine nucleotide-binding protein, beta subunit 5L|transducin beta chain 5', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52120926, end_position_on_the_genomic_accession=52191368, orientation='minus', exon_count=15, OMIM=604447, no_hdr0=''),
  10682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10682, CurrentID=0, Status='live', Symbol='EBP', Aliases='CDPX2, CHO2, CPX, CPXD, MEND', description='emopamil binding protein (sterol isomerase)', other_designations='3-beta-hydroxysteroid-delta-8,delta-7-isomerase|Chondrodysplasia punctata-2, X-linked dominant (Happle syndrome)|D8-D7 sterol isomerase|cholestenol Delta-isomerase|delta(8)-Delta(7) sterol isomerase|emopamil-binding protein|sterol 8-isomerase', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48521776, end_position_on_the_genomic_accession=48528716, orientation='plus', exon_count=5, OMIM=300205, no_hdr0=''),
  10683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10683, CurrentID=0, Status='live', Symbol='DLL3', Aliases='SCDO1', description='delta-like 3 (Drosophila)', other_designations='delta3|drosophila Delta homolog 3', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39498917, end_position_on_the_genomic_accession=39508481, orientation='plus', exon_count=8, OMIM=602768, no_hdr0=''),
  10686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10686, CurrentID=0, Status='live', Symbol='CLDN16', Aliases='HOMG3, PCLN1', description='claudin 16', other_designations='PCLN-1|hypomagnesemia 3, with hypercalciuria and nephrocalcinosis|paracellin-1', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190387872, end_position_on_the_genomic_accession=190412143, orientation='plus', exon_count=5, OMIM=603959, no_hdr0=''),
  10687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10687, CurrentID=0, Status='live', Symbol='PNMA2', Aliases='MA2, MM2, RGAG2', description='paraneoplastic Ma antigen 2', other_designations='40 kDa neuronal protein|onconeuronal antigen MA2|paraneoplastic neuronal antigen MM2|retrotransposon gag domain containing 2', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26504679, end_position_on_the_genomic_accession=26513967, orientation='minus', exon_count=4, OMIM=603970, no_hdr0=''),
  10688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10688, CurrentID=0, Status='live', Symbol='CRSP5', Aliases='CRSP85', description='cofactor required for Sp1 transcriptional activation, subunit 5', other_designations='cofactor required for Sp1 transcriptional activation, subunit 5, 85kDa', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10690, CurrentID=0, Status='live', Symbol='FUT9', Aliases='Fuc-TIX', description='fucosyltransferase 9 (alpha (1,3) fucosyltransferase)', other_designations='fucT-IX|fucosyltransferase IX|galactoside 3-L-fucosyltransferase', map_location='6q16', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96015969, end_position_on_the_genomic_accession=96215612, orientation='plus', exon_count=6, OMIM=606865, no_hdr0=''),
  10691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10691, CurrentID=0, Status='live', Symbol='GMEB1', Aliases='P96PIF, PIF96', description='glucocorticoid modulatory element binding protein 1', other_designations='DNA-binding protein p96PIF|GMEB-1|PIF p96|parvovirus initiation factor p96', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28668722, end_position_on_the_genomic_accession=28715603, orientation='plus', exon_count=11, OMIM=604409, no_hdr0=''),
  10692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10692, CurrentID=0, Status='live', Symbol='RRH', Aliases='', description='retinal pigment epithelium-derived rhodopsin homolog', other_designations='peropsin', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109827994, end_position_on_the_genomic_accession=109844707, orientation='plus', exon_count=7, OMIM=605224, no_hdr0=''),
  10693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10693, CurrentID=0, Status='live', Symbol='CCT6B', Aliases='CCT-zeta-2, CCTZ-2, Cctz2, TCP-1-zeta-2, TSA303', description='chaperonin containing TCP1, subunit 6B (zeta 2)', other_designations='chaperonin-containing T-complex polypeptide 1, subunit 6B|testis-specific Tcp20|testis-specific protein TSA303', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34927859, end_position_on_the_genomic_accession=34980919, orientation='minus', exon_count=15, OMIM=610730, no_hdr0=''),
  10694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10694, CurrentID=0, Status='live', Symbol='CCT8', Aliases='C21orf112, Cctq, D21S246, PRED71', description='chaperonin containing TCP1, subunit 8 (theta)', other_designations='CCT-theta|TCP-1-theta|renal carcinoma antigen NY-REN-15', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29056322, end_position_on_the_genomic_accession=29073797, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  10695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10695, CurrentID=0, Status='live', Symbol='CNPY3', Aliases='CAG4A, ERDA5, PRAT4A, TNRC5', description='canopy FGF signaling regulator 3', other_designations='CAG repeat containing|CTG repeat protein 4a|canopy 3 homolog|expanded repeat domain, CAG/CTG 5|expanded repeat-domain protein CAG/CTG 5|protein associated with TLR4|protein associated with Toll-like receptor 4A|trinucleotide repeat containing 5|trinucleotide repeat-containing gene 5 protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42929122, end_position_on_the_genomic_accession=42939270, orientation='plus', exon_count=8, OMIM=610774, no_hdr0=''),
  10699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10699, CurrentID=0, Status='live', Symbol='CORIN', Aliases='ATC2, CRN, Lrp4, PEE5, TMPRSS10', description='corin, serine peptidase', other_designations='heart-specific serine proteinase ATC2|pro-ANP-convertase|pro-ANP-converting enzyme|transmembrane protease serine 10', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47593998, end_position_on_the_genomic_accession=47838106, orientation='minus', exon_count=22, OMIM=605236, no_hdr0=''),
  101929363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929363, CurrentID=0, Status='live', Symbol='LOC101929363', Aliases='', description='uncharacterized LOC101929363', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207401103, end_position_on_the_genomic_accession=207416234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10712, CurrentID=0, Status='live', Symbol='FAM189B', Aliases='C1orf2, COTE1', description='family with sequence similarity 189 member B', other_designations='family with sequence similarity 189, member B', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155247205, end_position_on_the_genomic_accession=155255483, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  10713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10713, CurrentID=0, Status='live', Symbol='USP39', Aliases='65K, CGI-21, HSPC332, SAD1, SNRNP65', description='ubiquitin specific peptidase 39', other_designations='SAD1 homolog|SnRNP assembly defective 1 homolog|U4/U6.U5 tri-snRNP-associated 65 kDa protein|inactive ubiquitin-specific peptidase 39|small nuclear ribonucleoprotein 65kDa (U4/U6.U5)|ubiquitin specific protease 39', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85602842, end_position_on_the_genomic_accession=85649284, orientation='plus', exon_count=16, OMIM=611594, no_hdr0=''),
  10714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10714, CurrentID=0, Status='live', Symbol='POLD3', Aliases='P66, P68, PPP1R128', description='polymerase (DNA-directed), delta 3, accessory subunit', other_designations='DNA polymerase delta subunit p66|Pol delta C subunit (p66)|protein phosphatase 1, regulatory subunit 128', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74592530, end_position_on_the_genomic_accession=74669341, orientation='plus', exon_count=14, OMIM=611415, no_hdr0=''),
  10715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10715, CurrentID=0, Status='live', Symbol='CERS1', Aliases='EPM8, LAG1, LASS1, UOG1', description='ceramide synthase 1', other_designations='longevity assurance (LAG1, S. cerevisiae) homolog 1|longevity assurance gene 1 protein homolog 1|protein UOG-1|upstream of GDF1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18868546, end_position_on_the_genomic_accession=18896727, orientation='minus', exon_count=9, OMIM=606919, no_hdr0=''),
  10716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10716, CurrentID=0, Status='live', Symbol='TBR1', Aliases='TBR-1, TES-56', description='T-box, brain 1', other_designations='T-box, brain, 1|T-brain-1', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161416109, end_position_on_the_genomic_accession=161425062, orientation='plus', exon_count=6, OMIM=604616, no_hdr0=''),
  10717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10717, CurrentID=0, Status='live', Symbol='AP4B1', Aliases='BETA-4, CPSQ5, SPG47', description='adaptor related protein complex 4, beta 1 subunit', other_designations='AP-4 adaptor complex subunit beta|adaptor-related protein complex 4, beta 1 subunit|beta 4 subunit of AP-4|beta4-adaptin|spastic paraplegia 47', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113894748, end_position_on_the_genomic_accession=113905206, orientation='minus', exon_count=13, OMIM=607245, no_hdr0=''),
  10718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10718, CurrentID=0, Status='live', Symbol='NRG3', Aliases='HRG3, pro-NRG3', description='neuregulin 3', other_designations='neuregulin 3 variant 1|neuregulin 3 variant 10|neuregulin 3 variant 11|neuregulin 3 variant 12|neuregulin 3 variant 13|neuregulin 3 variant 14|neuregulin 3 variant 3|neuregulin 3 variant 4|neuregulin 3 variant 5|neuregulin 3 variant 6|neuregulin 3 variant 7|neuregulin 3 variant 8|neuregulin 3 variant 9|neuregulin-3-like polypeptide', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=81875309, end_position_on_the_genomic_accession=82987179, orientation='plus', exon_count=15, OMIM=605533, no_hdr0=''),
  10720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10720, CurrentID=0, Status='live', Symbol='UGT2B11', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B11', other_designations='UDP glycosyltransferase 2 family protein|UDP glycosyltransferase 2 family, polypeptide B11|UDPGT 2B11', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69199655, end_position_on_the_genomic_accession=69216768, orientation='minus', exon_count=6, OMIM=603064, no_hdr0=''),
  10721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10721, CurrentID=0, Status='live', Symbol='POLQ', Aliases='PRO0327', description='polymerase (DNA directed), theta', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121431420, end_position_on_the_genomic_accession=121546006, orientation='minus', exon_count=32, OMIM=604419, no_hdr0=''),
  10723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10723, CurrentID=0, Status='live', Symbol='SLC12A7', Aliases='KCC4', description='solute carrier family 12 (potassium/chloride transporter), member 7', other_designations='K-Cl cotransporter 4|electroneutral potassium-chloride cotransporter 4|potassium/chloride transporter KCC4|solute carrier family 12 (potassium/chloride transporters), member 7', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1050374, end_position_on_the_genomic_accession=1155747, orientation='minus', exon_count=30, OMIM=604879, no_hdr0=''),
  10724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10724, CurrentID=0, Status='live', Symbol='MGEA5', Aliases='MEA5, NCOAT, OGA', description='meningioma expressed antigen 5 (hyaluronidase)', other_designations='N-acetyl-beta-D-glucosaminidase|N-acetyl-beta-glucosaminidase|O-GlcNAcase|beta-N-acetylglucosaminidase|beta-N-acetylhexosaminidase|beta-hexosaminidase|bifunctional protein NCOAT|hyaluronidase in meningioma|meningioma-expressed antigen 5|nuclear cytoplasmic O-GlcNAcase and acetyltransferase', map_location='10q24.1-q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101784443, end_position_on_the_genomic_accession=101818709, orientation='minus', exon_count=18, OMIM=604039, no_hdr0=''),
  10725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10725, CurrentID=0, Status='live', Symbol='NFAT5', Aliases='NF-AT5, NFATL1, NFATZ, OREBP, TONEBP', description='nuclear factor of activated T-cells 5, tonicity-responsive', other_designations='NFAT-like protein 1|T-cell transcription factor NFAT5|TonE-binding protein|glutamine rich protein H65|osmotic response element-binding protein|tonicity-responsive enhancer-binding protein', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69565966, end_position_on_the_genomic_accession=69704666, orientation='plus', exon_count=19, OMIM=604708, no_hdr0=''),
  10726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10726, CurrentID=0, Status='live', Symbol='NUDC', Aliases='HNUDC, MNUDC, NPD011', description='nudC nuclear distribution protein', other_designations='nuclear distribution C homolog|nuclear distribution gene C homolog|nuclear distribution protein C homolog', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26911128, end_position_on_the_genomic_accession=26946871, orientation='plus', exon_count=11, OMIM=610325, no_hdr0=''),
  10728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10728, CurrentID=0, Status='live', Symbol='PTGES3', Aliases='P23, TEBP, cPGES', description='prostaglandin E synthase 3', other_designations='Hsp90 co-chaperone|cytosolic prostaglandin E synthase|cytosolic prostaglandin E2 synthase|progesterone receptor complex p23|prostaglandin E synthase 3 (cytosolic)|telomerase-binding protein p23|unactive progesterone receptor, 23 kD', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56663341, end_position_on_the_genomic_accession=56688408, orientation='minus', exon_count=10, OMIM=607061, no_hdr0=''),
  100271868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271868, CurrentID=0, Status='live', Symbol='MPVQTL2', Aliases='', description='Mean platelet volume QTL2', other_designations='', map_location='3p21-p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612574, no_hdr0=''),
  10730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10730, CurrentID=0, Status='live', Symbol='YME1L1', Aliases='FTSH, MEG4, PAMP, YME1L', description='YME1 like 1 ATPase', other_designations='ATP-dependent metalloprotease FtsH1 homolog|ATP-dependent metalloprotease YME1L1|YME1-like 1 ATPase|YME1-like protein 1|meg-4|presenilin-associated metalloprotease', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27110111, end_position_on_the_genomic_accession=27155278, orientation='minus', exon_count=20, OMIM=607472, no_hdr0=''),
  10732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10732, CurrentID=0, Status='live', Symbol='TCFL5', Aliases='CHA, E2BP-1, Figlb, bHLHe82', description='transcription factor-like 5 (basic helix-loop-helix)', other_designations='HPV-16 E2-binding protein 1|cha transcription factor', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62841014, end_position_on_the_genomic_accession=62861763, orientation='minus', exon_count=9, OMIM=604745, no_hdr0=''),
  10733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10733, CurrentID=0, Status='live', Symbol='PLK4', Aliases='MCCRP2, SAK, STK18', description='polo-like kinase 4', other_designations='Snk akin kinase|serine/threonine-protein kinase 18|serine/threonine-protein kinase Sak', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127880861, end_position_on_the_genomic_accession=127899224, orientation='plus', exon_count=17, OMIM=605031, no_hdr0=''),
  10734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10734, CurrentID=0, Status='live', Symbol='STAG3', Aliases='', description='stromal antigen 3', other_designations='SCC3 homolog 3|stromalin-3', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100177563, end_position_on_the_genomic_accession=100214387, orientation='plus', exon_count=35, OMIM=608489, no_hdr0=''),
  10735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10735, CurrentID=0, Status='live', Symbol='STAG2', Aliases='SA-2, SA2, SCC3B, bA517O1.1', description='stromal antigen 2', other_designations='SCC3 homolog 2', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123960560, end_position_on_the_genomic_accession=124102656, orientation='plus', exon_count=39, OMIM=300826, no_hdr0=''),
  10736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10736, CurrentID=0, Status='live', Symbol='SIX2', Aliases='', description='SIX homeobox 2', other_designations='sine oculis homeobox homolog 2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45005182, end_position_on_the_genomic_accession=45009610, orientation='minus', exon_count=2, OMIM=604994, no_hdr0=''),
  10737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10737, CurrentID=0, Status='live', Symbol='RFPL3S', Aliases='NCRNA00005, RFPL3-AS, RFPL3-AS1, RFPL3ANT, RFPL3AS', description='RFPL3 antisense', other_designations='RFPL3 antisense RNA 1 (non-protein coding)|non-protein coding RNA 5|ret finger protein-like 3 antisense', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32359906, end_position_on_the_genomic_accession=32371264, orientation='minus', exon_count=6, OMIM=605971, no_hdr0=''),
  10738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10738, CurrentID=0, Status='live', Symbol='RFPL3', Aliases='', description='ret finger protein-like 3', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32354885, end_position_on_the_genomic_accession=32361161, orientation='plus', exon_count=3, OMIM=605970, no_hdr0=''),
  10739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10739, CurrentID=0, Status='live', Symbol='RFPL2', Aliases='RNF79', description='ret finger protein-like 2', other_designations='RING finger protein 79', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32190435, end_position_on_the_genomic_accession=32207236, orientation='minus', exon_count=9, OMIM=605969, no_hdr0=''),
  10740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10740, CurrentID=0, Status='live', Symbol='RFPL1S', Aliases='NCRNA00006, RFPL1-AS, RFPL1-AS1, RFPL1AS', description='RFPL1 antisense RNA 1', other_designations='RFPL1 antisense RNA (non-protein coding)|RFPL1 antisense RNA 1 (non-protein coding)|RFPL1 antisense gene protein 1|Ret finger protein-like 1 antisense gene protein|non-protein coding RNA 6', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29437015, end_position_on_the_genomic_accession=29442129, orientation='minus', exon_count=1, OMIM=605972, no_hdr0=''),
  10741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10741, CurrentID=0, Status='live', Symbol='RBBP9', Aliases='BOG, RBBP10', description='retinoblastoma binding protein 9', other_designations='B5T overexpressed gene protein|B5T-overexpressed gene protein|RBBP-10|RBBP-9|retinoblastoma-binding protein 10|retinoblastoma-binding protein 9|retinoma-binding protein 9', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18486540, end_position_on_the_genomic_accession=18497255, orientation='minus', exon_count=5, OMIM=602908, no_hdr0=''),
  10742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10742, CurrentID=0, Status='live', Symbol='RAI2', Aliases='', description='retinoic acid induced 2', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17800049, end_position_on_the_genomic_accession=17861346, orientation='minus', exon_count=5, OMIM=300217, no_hdr0=''),
  10743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10743, CurrentID=0, Status='live', Symbol='RAI1', Aliases='SMCR, SMS', description='retinoic acid induced 1', other_designations='Smith-Magenis syndrome chromosome region', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17681473, end_position_on_the_genomic_accession=17811453, orientation='plus', exon_count=6, OMIM=607642, no_hdr0=''),
  10744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10744, CurrentID=0, Status='live', Symbol='PTTG2', Aliases='', description='pituitary tumor-transforming 2', other_designations='pituitary tumor-transforming gene 2 protein', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37960435, end_position_on_the_genomic_accession=37961010, orientation='plus', exon_count=1, OMIM=604231, no_hdr0=''),
  10745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10745, CurrentID=0, Status='live', Symbol='PHTF1', Aliases='PHTF', description='putative homeodomain transcription factor 1', other_designations='', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113697202, end_position_on_the_genomic_accession=113759892, orientation='minus', exon_count=23, OMIM=604950, no_hdr0=''),
  10746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10746, CurrentID=0, Status='live', Symbol='MAP3K2', Aliases='MEKK2, MEKK2B', description='mitogen-activated protein kinase kinase kinase 2', other_designations='MAP/ERK kinase kinase 2|MAPK/ERK kinase kinase 2|MEK kinase 2|MEKK 2', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127298669, end_position_on_the_genomic_accession=127343229, orientation='minus', exon_count=16, OMIM=609487, no_hdr0=''),
  10747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10747, CurrentID=0, Status='live', Symbol='MASP2', Aliases='MAP19, MASP-2, MASP1P1, sMAP', description='mannan-binding lectin serine peptidase 2', other_designations='MBL-associated plasma protein of 19 kD|MBL-associated serine protease 2|mannan-binding lectin serine peptidase 1 pseudogene 1|mannan-binding lectin serine protease 1 pseudogene 1|mannose-binding protein-associated serine protease 2|small MBL-associated protein', map_location='1p36.3-p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11026523, end_position_on_the_genomic_accession=11047239, orientation='minus', exon_count=12, OMIM=605102, no_hdr0=''),
  10748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10748, CurrentID=0, Status='live', Symbol='KLRAP1', Aliases='KLRA#, KLRA1, LY49L, Ly-49L, Ly49', description='killer cell lectin-like receptor subfamily A pseudogene 1', other_designations='killer cell lectin-like receptor subfamily A, member 1, pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10588478, end_position_on_the_genomic_accession=10599835, orientation='minus', exon_count=7, OMIM=604274, no_hdr0=''),
  10749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10749, CurrentID=0, Status='live', Symbol='KIF1C', Aliases='LTXS1, SATX2, SAX2, SPAX2, SPG58', description='kinesin family member 1C', other_designations='spastic ataxia 2 (autosomal recessive)', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4997948, end_position_on_the_genomic_accession=5028399, orientation='plus', exon_count=24, OMIM=603060, no_hdr0=''),
  10750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10750, CurrentID=0, Status='live', Symbol='GRAP', Aliases='', description='GRB2-related adaptor protein', other_designations='growth factor receptor-bound protein 2-related adaptor protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19020677, end_position_on_the_genomic_accession=19047023, orientation='minus', exon_count=6, OMIM=604330, no_hdr0=''),
  10751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10751, CurrentID=0, Status='live', Symbol='OFD1P18Y', Aliases='71-7A2, CYorf1, OFD1P18, OFD1PY18', description='OFD1 pseudogene 18, Y-linked', other_designations='OFD1 pseudogene, Y-linked 18', map_location='Yq11.22', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25897207, end_position_on_the_genomic_accession=25897581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10752, CurrentID=0, Status='live', Symbol='CHL1', Aliases='CALL, L1CAM2', description='cell adhesion molecule L1-like', other_designations='L1 cell adhesion molecule 2|cell adhesion molecule with homology to L1CAM (close homolog of L1)|cell adhesion molecule with homology to L1CAM (close homologue of L1)|close homolog of L1', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196596, end_position_on_the_genomic_accession=409417, orientation='plus', exon_count=33, OMIM=607416, no_hdr0=''),
  10753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10753, CurrentID=0, Status='live', Symbol='CAPN9', Aliases='GC36, nCL-4', description='calpain 9', other_designations='digestive tract-specific calpain|new calpain 4|novel calpain large subunit-4', map_location='1q42.11-q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230747334, end_position_on_the_genomic_accession=230802003, orientation='plus', exon_count=22, OMIM=606401, no_hdr0=''),
  10755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10755, CurrentID=0, Status='live', Symbol='GIPC1', Aliases='C19orf3, GIPC, GLUT1CBP, Hs.6454, IIP-1, NIP, RGS19IP1, SEMCAP, SYNECTIIN, SYNECTIN, TIP-2', description='GIPC PDZ domain containing family member 1', other_designations='GAIP C-terminus-interacting protein|GIPC PDZ domain containing family, member 1|GLUT1 C-terminal binding protein|IGF-1 receptor interacting protein 1|RGS-GAIP-interacting protein|RGS19-interacting protein 1|regulator of G-protein signalling 19 interacting protein 1|tax interaction protein 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14477759, end_position_on_the_genomic_accession=14496149, orientation='minus', exon_count=9, OMIM=605072, no_hdr0=''),
  10758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10758, CurrentID=0, Status='live', Symbol='TRAF3IP2', Aliases='ACT1, C6orf2, C6orf4, C6orf5, C6orf6, CANDF8, CIKS, PSORS13', description='TRAF3 interacting protein 2', other_designations='NFkB-activating protein ACT1|connection to IKK and SAPK/JNK|nuclear factor NF-kappa-B activator 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111555378, end_position_on_the_genomic_accession=111606874, orientation='minus', exon_count=12, OMIM=607043, no_hdr0=''),
  100271873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271873, CurrentID=0, Status='live', Symbol='AIRN', Aliases='AIR, IGF2R-AS, IGF2R-AS1, IGF2RAS, NCRNA00088', description='antisense of IGF2R non-protein coding RNA', other_designations='IGF2R antisense RNA 1 (non-protein coding)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160003291, end_position_on_the_genomic_accession=160007664, orientation='minus', exon_count=3, OMIM=604893, no_hdr0=''),
  10761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10761, CurrentID=0, Status='live', Symbol='PLAC1', Aliases='CT92, OOSP2L', description='placenta specific 1', other_designations='cancer/testis antigen 92|placenta-specific 1', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134565838, end_position_on_the_genomic_accession=134764312, orientation='minus', exon_count=6, OMIM=300296, no_hdr0=''),
  10762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10762, CurrentID=0, Status='live', Symbol='NUP50', Aliases='NPAP60, NPAP60L', description='nucleoporin 50kDa', other_designations='50 kDa nucleoporin|nuclear pore-associated protein 60L|nucleoporin 50kD|nucleoporin Nup50', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45163845, end_position_on_the_genomic_accession=45188011, orientation='plus', exon_count=12, OMIM=604646, no_hdr0=''),
  10763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10763, CurrentID=0, Status='live', Symbol='NES', Aliases='Nbla00170', description='nestin', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156668763, end_position_on_the_genomic_accession=156677397, orientation='minus', exon_count=4, OMIM=600915, no_hdr0=''),
  10765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10765, CurrentID=0, Status='live', Symbol='KDM5B', Aliases='CT31, JARID1B, PLU-1, PLU1, PPP1R98, PUT1, RBBP2H1A, RBP2-H1', description='lysine (K)-specific demethylase 5B', other_designations='cancer/testis antigen 31|histone demethylase JARID1B|jumonji, AT rich interactive domain 1B|jumonji/ARID domain-containing protein 1B|protein phosphatase 1, regulatory subunit 98|putative DNA/chromatin binding motif|retinoblastoma-binding protein 2 homolog 1|retinoblastoma-binding protein 2, homolog 1A', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202727403, end_position_on_the_genomic_accession=202808468, orientation='minus', exon_count=31, OMIM=605393, no_hdr0=''),
  10766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10766, CurrentID=0, Status='live', Symbol='TOB2', Aliases='TOB4, TOBL, TROB2', description='transducer of ERBB2, 2', other_designations='protein Tob4|transducer of erbB-2 2', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41433488, end_position_on_the_genomic_accession=41448230, orientation='minus', exon_count=3, OMIM=607396, no_hdr0=''),
  10767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10767, CurrentID=0, Status='live', Symbol='HBS1L', Aliases='EF-1a, ERFS, HBS1, HSPC276, eRF3c', description='HBS1 like translational GTPase', other_designations='ERF3-similar protein|HBS1-like translational GTPase|Hsp70 subfamily B suppressor 1-like protein|eRF3 family member', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134960378, end_position_on_the_genomic_accession=135054898, orientation='minus', exon_count=20, OMIM=612450, no_hdr0=''),
  10768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10768, CurrentID=0, Status='live', Symbol='AHCYL1', Aliases='DCAL, IRBIT, PPP1R78, PRO0233, XPVKONA', description='adenosylhomocysteinase like 1', other_designations='DC-expressed AHCY-like molecule|IP(3)Rs binding protein released with IP(3)|S-adenosyl homocysteine hydrolase homolog|S-adenosyl-L-homocysteine hydrolase 2|S-adenosylhomocysteine hydrolase-like protein 1|adenosylhomocysteinase-like 1|adoHcyase 2|dendritic cell expressed AHCY-like protein|inositol 1,4,5-trisphosphate receptor-binding protein|protein phosphatase 1, regulatory subunit 78|putative adenosylhomocysteinase 2', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109984741, end_position_on_the_genomic_accession=110023742, orientation='plus', exon_count=20, OMIM=607826, no_hdr0=''),
  10769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10769, CurrentID=0, Status='live', Symbol='PLK2', Aliases='SNK, hPlk2, hSNK', description='polo-like kinase 2', other_designations='PLK-2|serine/threonine-protein kinase SNK|serum-inducible kinase', map_location='5q12.1-q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58453982, end_position_on_the_genomic_accession=58460139, orientation='minus', exon_count=14, OMIM=607023, no_hdr0=''),
  10771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10771, CurrentID=0, Status='live', Symbol='ZMYND11', Aliases='BRAM1, BS69, MRD30', description='zinc finger, MYND-type containing 11', other_designations='adenovirus 5 E1A-binding protein|bone morphogenetic protein receptor-associated molecule 1', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=134465, end_position_on_the_genomic_accession=254637, orientation='plus', exon_count=18, OMIM=608668, no_hdr0=''),
  10772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10772, CurrentID=0, Status='live', Symbol='SRSF10', Aliases='FUSIP1, FUSIP2, NSSR, PPP1R149, SFRS13, SFRS13A, SRp38, SRrp40, TASR, TASR1, TASR2', description='serine/arginine-rich splicing factor 10', other_designations='40 kDa SR-repressor protein|FUS interacting protein (serine-arginine rich) 1|FUS-interacting protein (serine-arginine rich) 2|SR splicing factor 10|TLS-associated SR protein|TLS-associated protein TASR|TLS-associated protein with SR repeats|TLS-associated protein with Ser-Arg repeats|TLS-associated serine-arginine protein 1|TLS-associated serine-arginine protein 2|neural-salient SR protein|protein phosphatase 1, regulatory subunit 149|serine-arginine repressor protein (40 kDa)|splicing factor SRp38|splicing factor, arginine/serine-rich 13|splicing factor, arginine/serine-rich 13A', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23964347, end_position_on_the_genomic_accession=23980463, orientation='minus', exon_count=8, OMIM=605221, no_hdr0=''),
  10773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10773, CurrentID=0, Status='live', Symbol='ZBTB6', Aliases='ZID, ZNF482', description='zinc finger and BTB domain containing 6', other_designations='zinc finger protein 482|zinc finger protein with interaction domain', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122908056, end_position_on_the_genomic_accession=122913328, orientation='minus', exon_count=2, OMIM=605976, no_hdr0=''),
  10774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10774, CurrentID=0, Status='live', Symbol='DHS', Aliases='', description='dehydrated hereditary stomatocytosis', other_designations='', map_location='16q23-q24', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194380, no_hdr0=''),
  10775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10775, CurrentID=0, Status='live', Symbol='POP4', Aliases='RPP29', description='POP4 homolog, ribonuclease P/MRP subunit', other_designations='hPOP4', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29606263, end_position_on_the_genomic_accession=29617255, orientation='plus', exon_count=7, OMIM=606114, no_hdr0=''),
  10776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10776, CurrentID=0, Status='live', Symbol='ARPP19', Aliases='ARPP-16, ARPP-19, ARPP16, ENSAL', description='cAMP regulated phosphoprotein 19kDa', other_designations='cAMP-regulated phosphoprotein 19kDa|cAMP-regulated phosphoprotein, 19kDa|cyclic AMP phosphoprotein, 19 kD', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52547187, end_position_on_the_genomic_accession=52569221, orientation='minus', exon_count=5, OMIM=605487, no_hdr0=''),
  10777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10777, CurrentID=0, Status='live', Symbol='ARPP21', Aliases='ARPP-21, R3HDM3, RCS, TARPP', description='cAMP regulated phosphoprotein 21kDa', other_designations='R3H domain containing 3|cAMP-regulated phosphoprotein 21kDa|cAMP-regulated phosphoprotein, 21kDa|cyclic AMP-regulated phosphoprotein, 21 kD|thymocyte cAMP-regulated phosphoprotein', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35639174, end_position_on_the_genomic_accession=35794496, orientation='plus', exon_count=28, OMIM=605488, no_hdr0=''),
  10778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10778, CurrentID=0, Status='live', Symbol='ZNF271P', Aliases='CT-ZFP48, HZF7, ZNF-EB, ZNF-dp, ZNF271, ZNFEB', description='zinc finger protein 271, pseudogene', other_designations='C2H2 (Kruppel-type) zinc finger protein pseudogene|EBV-induced zinc finger protein|Zinc finger protein HZF7|Zinc finger protein ZNFphex133|Zinc finger protein dp|epstein-barr virus-induced zinc finger protein', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35290272, end_position_on_the_genomic_accession=35310766, orientation='plus', exon_count=3, OMIM=604754, no_hdr0=''),
  10780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10780, CurrentID=0, Status='live', Symbol='ZNF234', Aliases='HZF4, ZNF269', description='zinc finger protein 234', other_designations='C2-H2 type zinc finger protein|Homo sapiens zinc finger 234|zinc finger protein 234, partial sequence|zinc finger protein 269|zinc finger protein 4|zinc finger protein HZF4', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44141530, end_position_on_the_genomic_accession=44160309, orientation='plus', exon_count=7, OMIM=604750, no_hdr0=''),
  10781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10781, CurrentID=0, Status='live', Symbol='ZNF266', Aliases='HZF1', description='zinc finger protein 266', other_designations='zinc finger protein 1|zinc finger protein HZF1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9412426, end_position_on_the_genomic_accession=9435578, orientation='minus', exon_count=10, OMIM=604751, no_hdr0=''),
  10782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10782, CurrentID=0, Status='live', Symbol='ZNF274', Aliases='HFB101, ZF2, ZKSCAN19, ZSCAN51', description='zinc finger protein 274', other_designations='KRAB zinc finger protein HFB101|zinc finger protein HFB101|zinc finger protein with KRAB and SCAN domains 19|zinc finger protein zfp2', map_location='19qter', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58182989, end_position_on_the_genomic_accession=58213562, orientation='plus', exon_count=8, OMIM=605467, no_hdr0=''),
  10783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10783, CurrentID=0, Status='live', Symbol='NEK6', Aliases='SID6-1512', description='NIMA-related kinase 6', other_designations='NIMA (never in mitosis gene a)-related kinase 6|never in mitosis A-related kinase 6|nimA-related protein kinase 6|protein kinase SID6-1512|putative serine-threonine protein kinase', map_location='9q33.3-q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124257606, end_position_on_the_genomic_accession=124378316, orientation='plus', exon_count=18, OMIM=604884, no_hdr0=''),
  10785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10785, CurrentID=0, Status='live', Symbol='WDR4', Aliases='TRM82, TRMT82', description='WD repeat domain 4', other_designations='TRM82 tRNA methyltransferase 82 homolog|WD repeat-containing protein 4|tRNA (guanine-N(7)-)-methyltransferase subunit WDR4', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42843080, end_position_on_the_genomic_accession=42879589, orientation='minus', exon_count=14, OMIM=605924, no_hdr0=''),
  10786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10786, CurrentID=0, Status='live', Symbol='SLC17A3', Aliases='NPT4', description='solute carrier family 17 (organic anion transporter), member 3', other_designations='Na(+)/PI cotransporter 4|sodium/phosphate cotransporter 4|solute carrier family 17 (sodium phosphate), member 3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25845100, end_position_on_the_genomic_accession=25874243, orientation='minus', exon_count=13, OMIM=611034, no_hdr0=''),
  10787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10787, CurrentID=0, Status='live', Symbol='NCKAP1', Aliases='HEM2, NAP1, NAP125, p125Nap1', description='NCK-associated protein 1', other_designations='membrane-associated protein HEM-2', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182924851, end_position_on_the_genomic_accession=183038858, orientation='minus', exon_count=32, OMIM=604891, no_hdr0=''),
  10788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10788, CurrentID=0, Status='live', Symbol='IQGAP2', Aliases='', description='IQ motif containing GTPase activating protein 2', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76403255, end_position_on_the_genomic_accession=76708132, orientation='plus', exon_count=39, OMIM=605401, no_hdr0=''),
  10789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10789, CurrentID=0, Status='live', Symbol='NUDCP2', Aliases='', description='nudC nucelar distribution protein pseudogene 2', other_designations='nuclear distribution gene C homolog (A. nidulans) pseudogene 2|nuclear distribution gene C homolog, pseudogene 2', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178454716, end_position_on_the_genomic_accession=178455658, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10790, CurrentID=0, Status='live', Symbol='NUDCP1', Aliases='', description='nudC nucelar distribution protein pseudogene 1', other_designations='nuclear distribution gene C homolog (A. nidulans) pseudogene 1|nuclear distribution gene C homolog, pseudogene 1', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152389585, end_position_on_the_genomic_accession=152390849, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10791, CurrentID=0, Status='live', Symbol='VAMP5', Aliases='', description='vesicle associated membrane protein 5', other_designations='VAMP-5|myobrevin', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85584408, end_position_on_the_genomic_accession=85593388, orientation='plus', exon_count=3, OMIM=607029, no_hdr0=''),
  10793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10793, CurrentID=0, Status='live', Symbol='ZNF273', Aliases='HZF9', description='zinc finger protein 273', other_designations='zinc finger protein 9|zinc finger protein HZF9', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64870177, end_position_on_the_genomic_accession=64932239, orientation='plus', exon_count=13, OMIM=604756, no_hdr0=''),
  10794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10794, CurrentID=0, Status='live', Symbol='ZNF460', Aliases='HZF8, ZNF272', description='zinc finger protein 460', other_designations='zinc finger protein 272|zinc finger protein HZF8', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57279408, end_position_on_the_genomic_accession=57294075, orientation='plus', exon_count=4, OMIM=604755, no_hdr0=''),
  10795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10795, CurrentID=0, Status='live', Symbol='ZNF268', Aliases='HZF3', description='zinc finger protein 268', other_designations='zinc finger protein 3|zinc finger protein HZF3', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133181409, end_position_on_the_genomic_accession=133207112, orientation='plus', exon_count=7, OMIM=604753, no_hdr0=''),
  10797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10797, CurrentID=0, Status='live', Symbol='MTHFD2', Aliases='NMDMC', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase', other_designations='NAD-dependent methylene tetrahydrofolate dehydrogenase cyclohydrolase', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74198563, end_position_on_the_genomic_accession=74215298, orientation='plus', exon_count=9, OMIM=604887, no_hdr0=''),
  10798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10798, CurrentID=0, Status='live', Symbol='OR5I1', Aliases='HSOlf1, OLF1', description='olfactory receptor family 5 subfamily I member 1', other_designations='olfactory receptor OR11-159|olfactory receptor, family 5, subfamily I, member 1|olfactory receptor-like protein OLF1', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55935456, end_position_on_the_genomic_accession=55936400, orientation='minus', exon_count=1, OMIM=608496, no_hdr0=''),
  10799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10799, CurrentID=0, Status='live', Symbol='RPP40', Aliases='RNASEP1, bA428J1.3', description='ribonuclease P/MRP 40kDa subunit', other_designations='RNase P subunit 1|RNaseP protein p40|ribonuclease P (40 kD)|ribonuclease P, 40kD subunit|ribonuclease P1', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4992414, end_position_on_the_genomic_accession=5004063, orientation='minus', exon_count=10, OMIM=606117, no_hdr0=''),
  10800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10800, CurrentID=0, Status='live', Symbol='CYSLTR1', Aliases='CYSLT1, CYSLT1R, CYSLTR, HMTMF81', description='cysteinyl leukotriene receptor 1', other_designations='G-protein coupled receptor HG55|LTD4 receptor|cysteinyl leukotriene D4 receptor', map_location='Xq13.2-q21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78271468, end_position_on_the_genomic_accession=78327691, orientation='minus', exon_count=5, OMIM=300201, no_hdr0=''),
  10801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10801, CurrentID=0, Status='live', Symbol='SEPT9', Aliases='AF17q25, MSF, MSF1, NAPB, PNUTL4, SINT1, SeptD1', description='septin 9', other_designations='MLL septin-like fusion protein MSF-A|Ov/Br septin|ovarian/breast septin|septin D1', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77281410, end_position_on_the_genomic_accession=77500596, orientation='plus', exon_count=29, OMIM=604061, no_hdr0=''),
  10802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10802, CurrentID=0, Status='live', Symbol='SEC24A', Aliases='', description='SEC24 homolog A, COPII coat complex component', other_designations='SEC24 family member A|SEC24 family, member A|SEC24 related gene family, member A|SEC24-related protein A', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134648785, end_position_on_the_genomic_accession=134727911, orientation='plus', exon_count=24, OMIM=607183, no_hdr0=''),
  10803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10803, CurrentID=0, Status='live', Symbol='CCR9', Aliases='CC-CKR-9, CDw199, GPR-9-6, GPR28', description='chemokine (C-C motif) receptor 9', other_designations='G protein-coupled receptor 28', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45886504, end_position_on_the_genomic_accession=45903177, orientation='plus', exon_count=5, OMIM=604738, no_hdr0=''),
  10804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10804, CurrentID=0, Status='live', Symbol='GJB6', Aliases='CX30, DFNA3, DFNA3B, DFNB1B, ECTD2, ED2, EDH, HED, HED2', description='gap junction protein beta 6', other_designations='connexin 30|connexin-30|ectodermal dysplasia 2, hidrotic (Clouston syndrome)|gap junction protein, beta 6 (connexin 30)|gap junction protein, beta 6, 30kDa', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20221962, end_position_on_the_genomic_accession=20232395, orientation='minus', exon_count=6, OMIM=604418, no_hdr0=''),
  10806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10806, CurrentID=0, Status='live', Symbol='SDCCAG8', Aliases='BBS16, CCCAP, CCCAP SLSN7, HSPC085, NPHP10, NY-CO-8, SLSN7, hCCCAP', description='serologically defined colon cancer antigen 8', other_designations='antigen NY-CO-8|centrosomal colon cancer autoantigen protein', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243255419, end_position_on_the_genomic_accession=243500091, orientation='plus', exon_count=25, OMIM=613524, no_hdr0=''),
  10807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10807, CurrentID=0, Status='live', Symbol='SDCCAG3', Aliases='NY-CO-3', description='serologically defined colon cancer antigen 3', other_designations='antigen NY-CO-3', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136401922, end_position_on_the_genomic_accession=136410602, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  10808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10808, CurrentID=0, Status='live', Symbol='HSPH1', Aliases='HSP105, HSP105A, HSP105B, NY-CO-25', description='heat shock protein family H (Hsp110) member 1', other_designations='antigen NY-CO-25|heat shock 105kD alpha|heat shock 105kD beta|heat shock 105kDa/110kDa protein 1', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31134974, end_position_on_the_genomic_accession=31163109, orientation='minus', exon_count=19, OMIM=610703, no_hdr0=''),
  10809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10809, CurrentID=0, Status='live', Symbol='STARD10', Aliases='CGI-52, NY-CO-28, PCTP2, SDCCAG28', description='StAR related lipid transfer domain containing 10', other_designations='PCTP-L|START domain containing 10|START domain-containing protein 10|StAR-related lipid transfer (START) domain containing 10|StAR-related lipid transfer domain containing 10|antigen NY-CO-28|serologically defined colon cancer antigen 28|stAR-related lipid transfer protein 10', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72754729, end_position_on_the_genomic_accession=72793705, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  10810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10810, CurrentID=0, Status='live', Symbol='WASF3', Aliases='Brush-1, SCAR3, WAVE3', description='WAS protein family member 3', other_designations='WAS protein family, member 3|WASP family Verprolin-homologous protein 3|protein WAVE-3|verprolin homology domain-containing protein 3', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26557703, end_position_on_the_genomic_accession=26688945, orientation='plus', exon_count=13, OMIM=605068, no_hdr0=''),
  10811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10811, CurrentID=0, Status='live', Symbol='NOXA1', Aliases='NY-CO-31, SDCCAG31, p51NOX', description='NADPH oxidase activator 1', other_designations='NCF2-like protein|Nox activator 1|antigen NY-CO-31|inhibitory NADPH oxidase activator 1|p51-nox|p67phox-like factor|serologically defined colon cancer antigen 31', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137423395, end_position_on_the_genomic_accession=137434406, orientation='plus', exon_count=15, OMIM=611255, no_hdr0=''),
  10813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10813, CurrentID=0, Status='live', Symbol='UTP14A', Aliases='NYCO16, SDCCAG16, dJ537K23.3', description='UTP14A small subunit processome component', other_designations='UTP14, U3 small nucleolar ribonucleoprotein, homolog A|UTP14A small subunit (SSU) processome component|antigen NY-CO-16|serologically defined colon cancer antigen 16', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129906121, end_position_on_the_genomic_accession=129929762, orientation='plus', exon_count=18, OMIM=300508, no_hdr0=''),
  10814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10814, CurrentID=0, Status='live', Symbol='CPLX2', Aliases='921-L, CPX-2, CPX2, Hfb1', description='complexin 2', other_designations='CPX II|complexin II|synaphin 1|synaphin-1', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175796539, end_position_on_the_genomic_accession=175884021, orientation='plus', exon_count=7, OMIM=605033, no_hdr0=''),
  10815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10815, CurrentID=0, Status='live', Symbol='CPLX1', Aliases='CPX-I, CPX1', description='complexin 1', other_designations='CPX I|complexin I|synaphin 2|synaphin-2', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=784957, end_position_on_the_genomic_accession=826157, orientation='minus', exon_count=5, OMIM=605032, no_hdr0=''),
  10816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10816, CurrentID=0, Status='live', Symbol='SPINT3', Aliases='HKIB9', description='serine peptidase inhibitor, Kunitz type, 3', other_designations='serine protease inhibitor Kunitz type 3|serine protease inhibitor, Kunitz type, 3', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45512461, end_position_on_the_genomic_accession=45515624, orientation='minus', exon_count=2, OMIM=613941, no_hdr0=''),
  10817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10817, CurrentID=0, Status='live', Symbol='FRS3', Aliases='FRS2-beta, FRS2B, FRS2beta, SNT-2, SNT2', description='fibroblast growth factor receptor substrate 3', other_designations='FGFR substrate 3|FGFR-signaling adaptor SNT2|suc1-associated neurotrophic factor target 2 (FGFR signalling adaptor)', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41770176, end_position_on_the_genomic_accession=41779905, orientation='minus', exon_count=8, OMIM=607744, no_hdr0=''),
  10818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10818, CurrentID=0, Status='live', Symbol='FRS2', Aliases='FRS2Aalpha, SNT, SNT-1, SNT1, FRS2', description='fibroblast growth factor receptor substrate 2', other_designations='FGFR signalling adaptor|FGFR substrate 2|FGFR-signaling adaptor SNT|suc1-associated neurotrophic factor target 1', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69470349, end_position_on_the_genomic_accession=69579793, orientation='plus', exon_count=14, OMIM=607743, no_hdr0=''),
  10819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10819, CurrentID=0, Status='live', Symbol='OR7E14P', Aliases='OR11-5, OR7E151P', description='olfactory receptor family 7 subfamily E member 14 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 14 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17051886, end_position_on_the_genomic_accession=17053044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10820, CurrentID=0, Status='live', Symbol='OR7E13P', Aliases='OR11-4, OR7E44P, OST022, PJCG2', description='olfactory receptor family 7 subfamily E member 13 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 13 pseudogene|olfactory receptor, family 7, subfamily E, member 44 pseudogene|olfactory-like receptor PJCG2', map_location='11q13-q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86832606, end_position_on_the_genomic_accession=86833259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  404037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404037, CurrentID=0, Status='live', Symbol='HAPLN4', Aliases='BRAL2', description='hyaluronan and proteoglycan link protein 4', other_designations='brain link protein 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19255641, end_position_on_the_genomic_accession=19262787, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10822, CurrentID=0, Status='live', Symbol='OR7E11P', Aliases='OR11-2, OR7E144P', description='olfactory receptor family 7 subfamily E member 11 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 11 pseudogene|olfactory receptor, family 7, subfamily E, member 144 pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67735678, end_position_on_the_genomic_accession=67736341, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10823, CurrentID=0, Status='live', Symbol='OR7E10P', Aliases='OR11-1', description='olfactory receptor family 7 subfamily E member 10 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 10 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12703052, end_position_on_the_genomic_accession=12704038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10824, CurrentID=0, Status='live', Symbol='DIAPH2-AS1', Aliases='EPAG', description='DIAPH2 antisense RNA 1', other_designations='DIAPH2 antisense RNA 1 (non-protein coding)|early lymphoid activation protein', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=97528364, end_position_on_the_genomic_accession=97564535, orientation='minus', exon_count=3, OMIM=300347, no_hdr0=''),
  10825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10825, CurrentID=0, Status='live', Symbol='NEU3', Aliases='SIAL3', description='neuraminidase 3 (membrane sialidase)', other_designations='N-acetyl-alpha-neuraminidase 3|ganglioside sialidase|membrane sialidase|neuraminidase 3|sialidase 3 (membrane sialidase)', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74988905, end_position_on_the_genomic_accession=75007698, orientation='plus', exon_count=3, OMIM=604617, no_hdr0=''),
  10826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10826, CurrentID=0, Status='live', Symbol='FAXDC2', Aliases='C5orf4', description='fatty acid hydroxylase domain containing 2', other_designations='', map_location='5q31-q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154817034, end_position_on_the_genomic_accession=154850653, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  10827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10827, CurrentID=0, Status='live', Symbol='FAM114A2', Aliases='133K02, C5orf3', description='family with sequence similarity 114 member A2', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=153991599, end_position_on_the_genomic_accession=154038937, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  10828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10828, CurrentID=0, Status='live', Symbol='NPM1P19', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 19', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38977514, end_position_on_the_genomic_accession=38978514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10829, CurrentID=0, Status='live', Symbol='NPM1P18', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 18', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36818863, end_position_on_the_genomic_accession=36820016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10830, CurrentID=0, Status='live', Symbol='NPM1P17', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 17', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137723655, end_position_on_the_genomic_accession=137724568, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10831, CurrentID=0, Status='live', Symbol='NPM1P7', Aliases='NG7-6, NPM1P16, NPMP7', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 7', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 7|nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 16', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9695287, end_position_on_the_genomic_accession=9696549, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10833, CurrentID=0, Status='live', Symbol='NPM1P14', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 14', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112520396, end_position_on_the_genomic_accession=112521983, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10834, CurrentID=0, Status='live', Symbol='NPM1P13', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 13', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18948591, end_position_on_the_genomic_accession=18949862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10835, CurrentID=0, Status='live', Symbol='NPM1P12', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 12', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149333914, end_position_on_the_genomic_accession=149335170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10836, CurrentID=0, Status='live', Symbol='NPM1P11', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 11', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11257375, end_position_on_the_genomic_accession=11258640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10837, CurrentID=0, Status='live', Symbol='NPM1P10', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 10', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104024927, end_position_on_the_genomic_accession=104026859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  10838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10838, CurrentID=0, Status='live', Symbol='ZNF275', Aliases='', description='zinc finger protein 275', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153334155, end_position_on_the_genomic_accession=153352926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10840, CurrentID=0, Status='live', Symbol='ALDH1L1', Aliases='10-FTHFDH, 10-fTHF, FDH, FTHFD', description='aldehyde dehydrogenase 1 family member L1', other_designations='10-formyltetrahydrofolate dehydrogenase|MER57B-ALDH1L1|aldehyde dehydrogenase 1 family, member L1', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126103561, end_position_on_the_genomic_accession=126181526, orientation='minus', exon_count=25, OMIM=600249, no_hdr0=''),
  10841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10841, CurrentID=0, Status='live', Symbol='FTCD', Aliases='LCHC1', description='formimidoyltransferase cyclodeaminase', other_designations='formiminotransferase cyclodeaminase|formiminotransferase-cyclodeaminase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46136151, end_position_on_the_genomic_accession=46155856, orientation='minus', exon_count=16, OMIM=606806, no_hdr0=''),
  10842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10842, CurrentID=0, Status='live', Symbol='PPP1R17', Aliases='C7orf16, GSBS', description='protein phosphatase 1 regulatory subunit 17', other_designations='G-substrate|protein phosphatase 1, regulatory subunit 17', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31687017, end_position_on_the_genomic_accession=31709911, orientation='plus', exon_count=5, OMIM=604088, no_hdr0=''),
  10844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10844, CurrentID=0, Status='live', Symbol='TUBGCP2', Aliases='GCP-2, GCP2, Grip103, SPBC97, Spc97p, h103p, hGCP2, hSpc97', description='tubulin gamma complex associated protein 2', other_designations='gamma-ring complex protein 103 kDa|hGrip103|spindle pole body protein Spc97 homolog|tubulin, gamma complex associated protein 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133278630, end_position_on_the_genomic_accession=133312337, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  10845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10845, CurrentID=0, Status='live', Symbol='CLPX', Aliases='', description='caseinolytic mitochondrial matrix peptidase chaperone subunit', other_designations='ClpX caseinolytic peptidase X homolog|ClpX caseinolytic protease X homolog|energy-dependent regulator of proteolysis', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65150446, end_position_on_the_genomic_accession=65185404, orientation='minus', exon_count=14, OMIM=615611, no_hdr0=''),
  10846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10846, CurrentID=0, Status='live', Symbol='PDE10A', Aliases='HSPDE10A', description='phosphodiesterase 10A', other_designations='dJ416F21.1 (phosphodiesterase 10A)|phosphodiesterase 10A1 (PDE10A1)|phosphodiesterase 10A7 (PDE10A7)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165327287, end_position_on_the_genomic_accession=165711153, orientation='minus', exon_count=30, OMIM=610652, no_hdr0=''),
  10847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10847, CurrentID=0, Status='live', Symbol='SRCAP', Aliases='DOMO1, EAF1, FLHS, SWR1', description='Snf2-related CREBBP activator protein', other_designations='Snf2-related CBP activator protein|Swi2/Snf2-related ATPase homolog, domino homolog 1|domino homolog 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30699141, end_position_on_the_genomic_accession=30740129, orientation='plus', exon_count=34, OMIM=611421, no_hdr0=''),
  10848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10848, CurrentID=0, Status='live', Symbol='PPP1R13L', Aliases='IASPP, NKIP1, RAI, RAI4', description='protein phosphatase 1 regulatory subunit 13 like', other_designations='NFkB interacting protein 1|NFkB-interacting protein 1|PPP1R13B-like protein|inhibitor of ASPP protein|inhibitor of apoptosis stimulating protein of p53|protein iASPP|protein phosphatase 1, regulatory (inhibitor) subunit 13 like|protein phosphatase 1, regulatory subunit 13 like|retinoic acid induced 4', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45379634, end_position_on_the_genomic_accession=45406349, orientation='minus', exon_count=14, OMIM=607463, no_hdr0=''),
  10849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10849, CurrentID=0, Status='live', Symbol='CD3EAP', Aliases='ASE-1, ASE1, CAST, PAF49', description='CD3e molecule, epsilon associated protein', other_designations='CD3e antigen, epsilon polypeptide associated protein|RNA polymerase I-associated factor PAF49|antisense to ERCC-1 protein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45406209, end_position_on_the_genomic_accession=45410766, orientation='plus', exon_count=3, OMIM=107325, no_hdr0=''),
  10850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10850, CurrentID=0, Status='live', Symbol='CCL27', Aliases='ALP, CTACK, CTAK, ESKINE, ILC, PESKY, SCYA27', description='chemokine (C-C motif) ligand 27', other_designations='CC chemokine ILC|IL-11 R-alpha-locus chemokine|IL-11 Ralpha-locus chemokine|cutaneous T-cell attracting chemokine|cutaneous T-cell-attracting chemokine|skinkine|small inducible cytokine subfamily A (Cys-Cys), member 27|small-inducible cytokine A27', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34661883, end_position_on_the_genomic_accession=34662692, orientation='minus', exon_count=3, OMIM=604833, no_hdr0=''),
  10855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10855, CurrentID=0, Status='live', Symbol='HPSE', Aliases='HPA, HPA1, HPR11, HSE1, HPSE', description='heparanase', other_designations='endo-glucoronidase|heparanase-1', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83292461, end_position_on_the_genomic_accession=83335153, orientation='minus', exon_count=13, OMIM=604724, no_hdr0=''),
  10856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10856, CurrentID=0, Status='live', Symbol='RUVBL2', Aliases='CGI-46, ECP51, INO80J, REPTIN, RVB2, TIH2, TIP48, TIP49B', description='RuvB-like AAA ATPase 2', other_designations='48 kDa TATA box-binding protein-interacting protein|48 kDa TBP-interacting protein|51 kDa erythrocyte cytosolic protein|ECP-51|INO80 complex subunit J|RuvB (E coli homolog)-like 2|TAP54-beta|TIP60-associated protein 54-beta|erythrocyte cytosolic protein, 51-KD|repressing pontin 52|reptin52 protein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48993353, end_position_on_the_genomic_accession=49015926, orientation='plus', exon_count=16, OMIM=604788, no_hdr0=''),
  10857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10857, CurrentID=0, Status='live', Symbol='PGRMC1', Aliases='HPR6.6, MPR', description='progesterone receptor membrane component 1', other_designations='progesterone binding protein', map_location='Xq22-q24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119236245, end_position_on_the_genomic_accession=119244466, orientation='plus', exon_count=3, OMIM=300435, no_hdr0=''),
  10858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10858, CurrentID=0, Status='live', Symbol='CYP46A1', Aliases='CP46, CYP46', description='cytochrome P450 family 46 subfamily A member 1', other_designations='CH24H|cytochrome P450 46A1|cytochrome P450, family 46, subfamily A, polypeptide 1|cytochrome P450, subfamily 46 (cholesterol 24-hydroxylase)', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99684262, end_position_on_the_genomic_accession=99727321, orientation='plus', exon_count=17, OMIM=604087, no_hdr0=''),
  10859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10859, CurrentID=0, Status='live', Symbol='LILRB1', Aliases='CD85J, ILT-2, ILT2, LIR-1, LIR1, MIR-7, MIR7', description='leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 1', other_designations='CD85 antigen-like family member J|Ig-like transcript 2|leukocyte immunoglobulin-like receptor subfamily B member 1 soluble isoform|monocyte/macrophage immunoglobulin-like receptor 7', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54616930, end_position_on_the_genomic_accession=54638167, orientation='plus', exon_count=16, OMIM=604811, no_hdr0=''),
  10861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10861, CurrentID=0, Status='live', Symbol='SLC26A1', Aliases='EDM4, SAT-1, SAT1', description='solute carrier family 26 (anion exchanger), member 1', other_designations='solute carrier family 26 (sulfate transporter), member 1|sulfate anion tranporter AT1|sulfate/anion transporter SAT-1 protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=979073, end_position_on_the_genomic_accession=994217, orientation='minus', exon_count=5, OMIM=610130, no_hdr0=''),
  10863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10863, CurrentID=0, Status='live', Symbol='ADAM28', Aliases='ADAM 28, MDC-L, MDCL, eMDC II, eMDCII', description='ADAM metallopeptidase domain 28', other_designations='epididymal metalloproteinase-like, disintegrin-like, and cysteine-rich protein II|epididymial metalloproteinase-like, disintegrin-like, and cysteine-rich protein II|metalloproteinase-like, disintegrin-like, and cysteine-rich protein-L', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24294040, end_position_on_the_genomic_accession=24359018, orientation='plus', exon_count=28, OMIM=606188, no_hdr0=''),
  10864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10864, CurrentID=0, Status='live', Symbol='SLC22A7', Aliases='NLT, OAT2', description='solute carrier family 22 (organic anion transporter), member 7', other_designations='hOAT2|liver-specific transporter|novel liver transporter|organic anion transporter 2', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43295710, end_position_on_the_genomic_accession=43305538, orientation='plus', exon_count=15, OMIM=604995, no_hdr0=''),
  10865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10865, CurrentID=0, Status='live', Symbol='ARID5A', Aliases='MRF-1, MRF1, RP11-363D14', description='AT-rich interaction domain 5A', other_designations='ARID domain-containing protein 5A|AT rich interactive domain 5A (MRF1-like)|RFVG5814|modulator recognition factor 1|modulator recognition factor I', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96536727, end_position_on_the_genomic_accession=96552634, orientation='plus', exon_count=9, OMIM=611583, no_hdr0=''),
  10866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10866, CurrentID=0, Status='live', Symbol='HCP5', Aliases='6S2650E, D6S2650E, P5-1', description='HLA complex P5 (non-protein coding)', other_designations='HLA class I histocompatibility antigen protein P5|HLA complex protein P5|MHC class I region ORF', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31463180, end_position_on_the_genomic_accession=31465809, orientation='plus', exon_count=2, OMIM=604676, no_hdr0=''),
  10867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10867, CurrentID=0, Status='live', Symbol='TSPAN9', Aliases='NET-5, NET5, PP1057', description='tetraspanin 9', other_designations='new EST tetraspan 5|tetraspan NET-5|transmembrane 4 superfamily member tetraspan NET-5|tspan-9', map_location='12p13.33-p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3077355, end_position_on_the_genomic_accession=3286564, orientation='plus', exon_count=11, OMIM=613137, no_hdr0=''),
  10868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10868, CurrentID=0, Status='live', Symbol='USP20', Aliases='LSFR3A, VDU2, hVDU2', description='ubiquitin specific peptidase 20', other_designations='deubiquitinating enzyme 20|pVHL-interacting deubiquitinating enzyme 2|ubiquitin thioesterase 20|ubiquitin thiolesterase 20|ubiquitin-specific-processing protease 20', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129835417, end_position_on_the_genomic_accession=129881838, orientation='plus', exon_count=28, OMIM=615143, no_hdr0=''),
  10869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10869, CurrentID=0, Status='live', Symbol='USP19', Aliases='ZMYND9', description='ubiquitin specific peptidase 19', other_designations='deubiquitinating enzyme 19|ubiquitin specific protease 19|ubiquitin thioesterase 19|ubiquitin thiolesterase 19|ubiquitin-specific-processing protease 19|zinc finger MYND domain-containing protein 9', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49108046, end_position_on_the_genomic_accession=49120938, orientation='minus', exon_count=29, OMIM=614471, no_hdr0=''),
  10870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10870, CurrentID=0, Status='live', Symbol='HCST', Aliases='DAP10, KAP10, PIK3AP', description='hematopoietic cell signal transducer', other_designations='DNAX-activation protein 10|kinase assoc pro of ~10kDa|kinase assoc protein|membrane protein DAP10|phosphoinositide-3-kinase adaptor protein|transmembrane adapter protein KAP10', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35902480, end_position_on_the_genomic_accession=35904271, orientation='plus', exon_count=4, OMIM=604089, no_hdr0=''),
  10871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10871, CurrentID=0, Status='live', Symbol='CD300C', Aliases='CLM-6, CMRF-35, CMRF-35A, CMRF35, CMRF35-A1, CMRF35A, CMRF35A1, IGSF16, LIR', description='CD300c molecule', other_designations='CD300 antigen-like family member C|CD300c antigen|CMRF35 antigen|CMRF35 leukocyte immunoglobulin-like receptor|CMRF35A leukocyte immunoglobulin-like receptor|immunoglobulin superfamily member 16', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74541065, end_position_on_the_genomic_accession=74546171, orientation='minus', exon_count=4, OMIM=606786, no_hdr0=''),
  100270851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270851, CurrentID=0, Status='live', Symbol='RPL7P17', Aliases='RPL7_6_493', description='ribosomal protein L7 pseudogene 17', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77082399, end_position_on_the_genomic_accession=77083113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132884, CurrentID=0, Status='live', Symbol='EVC2', Aliases='LBN, WAD', description='Ellis van Creveld syndrome 2', other_designations='ellis-van Creveld syndrome protein 2', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5534397, end_position_on_the_genomic_accession=5709548, orientation='minus', exon_count=24, OMIM=607261, no_hdr0=''),
  10874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10874, CurrentID=0, Status='live', Symbol='NMU', Aliases='', description='neuromedin U', other_designations='prepro-NMU', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55595229, end_position_on_the_genomic_accession=55636698, orientation='minus', exon_count=13, OMIM=605103, no_hdr0=''),
  10875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10875, CurrentID=0, Status='live', Symbol='FGL2', Aliases='T49, pT49', description='fibrinogen-like 2', other_designations='fibrinogen-like protein 2', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77193371, end_position_on_the_genomic_accession=77199833, orientation='minus', exon_count=2, OMIM=605351, no_hdr0=''),
  10876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10876, CurrentID=0, Status='live', Symbol='EDDM3A', Aliases='EP3A, FAM12A, HE3-ALPHA, HE3A, HE3ALPHA, RAM1', description='epididymal protein 3A', other_designations='epididymal secretory protein E3 alpha|epididymis-specific 3 alpha|family with sequence similarity 12, member A|human epididymis-specific 3 alpha|human epididymis-specific protein 3-alpha|ribonuclease A M1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20745940, end_position_on_the_genomic_accession=20748380, orientation='plus', exon_count=2, OMIM=611580, no_hdr0=''),
  404093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404093, CurrentID=0, Status='live', Symbol='CUEDC1', Aliases='', description='CUE domain containing 1', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57861243, end_position_on_the_genomic_accession=57955323, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  10878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10878, CurrentID=0, Status='live', Symbol='CFHR3', Aliases='CFHL3, DOWN16, FHR-3, FHR3, HLF4', description='complement factor H-related 3', other_designations='H factor-like 4|H factor-like protein 3', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196774800, end_position_on_the_genomic_accession=196794073, orientation='plus', exon_count=8, OMIM=605336, no_hdr0=''),
  10879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10879, CurrentID=0, Status='live', Symbol='SMR3B', Aliases='P-B, PBII, PRL3, PROL3, SMR1B', description='submaxillary gland androgen regulated protein 3B', other_designations='proline rich 3|proline-rich peptide P-B|proline-rich protein 3|salivary proline-rich protein|submaxillary gland androgen regulated protein 3 homolog B', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70383078, end_position_on_the_genomic_accession=70390244, orientation='plus', exon_count=3, OMIM=611593, no_hdr0=''),
  10880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10880, CurrentID=0, Status='live', Symbol='ACTL7B', Aliases='Tact1', description='actin like 7B', other_designations='actin-like 7-beta|actin-like 7B|actin-like-7-beta', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108854588, end_position_on_the_genomic_accession=108855995, orientation='minus', exon_count=1, OMIM=604304, no_hdr0=''),
  10881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10881, CurrentID=0, Status='live', Symbol='ACTL7A', Aliases='', description='actin like 7A', other_designations='actin-like 7-alpha|actin-like 7A|actin-like-7-alpha', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108862228, end_position_on_the_genomic_accession=108863755, orientation='plus', exon_count=1, OMIM=604303, no_hdr0=''),
  10882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10882, CurrentID=0, Status='live', Symbol='C1QL1', Aliases='C1QRF, C1QTNF14, CRF', description='complement component 1, q subcomponent-like 1', other_designations='C1q and tumor necrosis factor-related protein 14|C1q/TNF-related protein 14', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44959693, end_position_on_the_genomic_accession=44968276, orientation='minus', exon_count=2, OMIM=611586, no_hdr0=''),
  10884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10884, CurrentID=0, Status='live', Symbol='MRPS30', Aliases='MRP-S30, PAP, PDCD9, S30mt', description='mitochondrial ribosomal protein S30', other_designations='programmed cell death 9', map_location='5q11', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=44808925, end_position_on_the_genomic_accession=44815514, orientation='plus', exon_count=5, OMIM=611991, no_hdr0=''),
  10885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10885, CurrentID=0, Status='live', Symbol='WDR3', Aliases='DIP2, UTP12', description='WD repeat domain 3', other_designations='dJ776P7.2 (WD repeat domain 3)', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117929749, end_position_on_the_genomic_accession=117960426, orientation='plus', exon_count=27, OMIM=604737, no_hdr0=''),
  10886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10886, CurrentID=0, Status='live', Symbol='NPFFR2', Aliases='GPR74, HLWAR77, NPFF2, NPGPR', description='neuropeptide FF receptor 2', other_designations='G protein-coupled receptor 74|G-protein coupled receptor 74|G-protein coupled receptor HLWAR77|neuropeptide FF 2|neuropeptide G-protein coupled receptor', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=72031804, end_position_on_the_genomic_accession=72148201, orientation='plus', exon_count=6, OMIM=607449, no_hdr0=''),
  10887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10887, CurrentID=0, Status='live', Symbol='PROKR1', Aliases='GPR73, GPR73a, PKR1, ZAQ', description='prokineticin receptor 1', other_designations='G protein-coupled receptor 73|G protein-coupled receptor ZAQ|G-protein coupled receptor 73|G-protein coupled receptor ZAQ|PK-R1', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68645822, end_position_on_the_genomic_accession=68655576, orientation='plus', exon_count=2, OMIM=607122, no_hdr0=''),
  10888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10888, CurrentID=0, Status='live', Symbol='GPR83', Aliases='GIR, GPR72', description='G protein-coupled receptor 83', other_designations='G protein-coupled receptor 72|G-protein coupled receptor 72|glucocorticoid induced receptor', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94377311, end_position_on_the_genomic_accession=94401419, orientation='minus', exon_count=4, OMIM=605569, no_hdr0=''),
  285804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285804, CurrentID=0, Status='live', Symbol='LOC285804', Aliases='', description='uncharacterized LOC285804', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170266669, end_position_on_the_genomic_accession=170276762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10890, CurrentID=0, Status='live', Symbol='RAB10', Aliases='', description='RAB10, member RAS oncogene family', other_designations='GTP-binding protein RAB10|ras-related GTP-binding protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26033860, end_position_on_the_genomic_accession=26137454, orientation='plus', exon_count=6, OMIM=612672, no_hdr0=''),
  10891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10891, CurrentID=0, Status='live', Symbol='PPARGC1A', Aliases='LEM6, PGC-1(alpha), PGC-1v, PGC1, PGC1A, PPARGC1', description='peroxisome proliferator-activated receptor gamma, coactivator 1 alpha', other_designations='L-PGC-1alpha|PGC-1-alpha|PPAR gamma coactivator variant form|PPARGC-1-alpha|ligand effect modulator-6|peroxisome proliferator-activated receptor gamma coactivator 1 alpha transcript variant B4|peroxisome proliferator-activated receptor gamma coactivator 1 alpha transcript variant B4-3ext|peroxisome proliferator-activated receptor gamma coactivator 1 alpha transcript variant B4-8a|peroxisome proliferator-activated receptor gamma coactivator 1 alpha transcript variant B5|peroxisome proliferator-activated receptor gamma coactivator 1 alpha transcript variant B5-NT', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23792021, end_position_on_the_genomic_accession=24472771, orientation='minus', exon_count=24, OMIM=604517, no_hdr0=''),
  10892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10892, CurrentID=0, Status='live', Symbol='MALT1', Aliases='IMD12, MLT, MLT1', description='MALT1 paracaspase', other_designations='MALT lymphoma-associated translocation|MALT1 protease|caspase-like protein|mucosa associated lymphoid tissue lymphoma translocation gene 1|paracaspase', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58671386, end_position_on_the_genomic_accession=58750139, orientation='plus', exon_count=19, OMIM=604860, no_hdr0=''),
  10893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10893, CurrentID=0, Status='live', Symbol='MMP24', Aliases='MMP-24, MMP25, MT-MMP 5, MT-MMP5, MT5-MMP, MT5MMP, MTMMP5', description='matrix metallopeptidase 24', other_designations='matrix metallopeptidase 24 (membrane-inserted)|matrix metalloproteinase 24 (membrane-inserted)|membrane-type 5 matrix metalloproteinase|membrane-type matrix metalloproteinase 5|membrane-type-5 matrix metalloproteinase', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35226736, end_position_on_the_genomic_accession=35277663, orientation='plus', exon_count=9, OMIM=604871, no_hdr0=''),
  10894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10894, CurrentID=0, Status='live', Symbol='LYVE1', Aliases='CRSBP-1, HAR, LYVE-1, XLKD1', description='lymphatic vessel endothelial hyaluronan receptor 1', other_designations='cell surface retention sequence binding protein-1|cell surface retention sequence-binding protein 1|extracellular link domain containing 1|extracellular link domain-containing 1|extracellular link domain-containing protein 1|hyaluronic acid receptor', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10557866, end_position_on_the_genomic_accession=10568818, orientation='minus', exon_count=6, OMIM=605702, no_hdr0=''),
  10895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10895, CurrentID=0, Status='live', Symbol='PPBPP2', Aliases='PPBPL2, SPBPBP', description='pro-platelet basic protein pseudogene 2', other_designations='DNA-binding protein SPBPBP|DNA-binding protein amplifying expression of surfactant protein B|Putative platelet basic protein-like 2|pro-platelet basic protein-like 2', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74054038, end_position_on_the_genomic_accession=74055399, orientation='minus', exon_count=1, OMIM=611591, no_hdr0=''),
  10896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10896, CurrentID=0, Status='live', Symbol='OCLM', Aliases='TISR', description='oculomedin', other_designations='trabecular meshwork inducible stretch response protein', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186400572, end_position_on_the_genomic_accession=186401455, orientation='plus', exon_count=1, OMIM=604301, no_hdr0=''),
  10897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10897, CurrentID=0, Status='live', Symbol='YIF1A', Aliases='54TM, FinGER7, YIF1, YIF1P', description='Yip1 interacting factor homolog A (S. cerevisiae)', other_designations='54TMp|YIP1-interacting factor homolog A|Yip1p-interacting factor|putative Rab5-interacting protein|putative transmembrane protein 54TMp', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66284580, end_position_on_the_genomic_accession=66289167, orientation='minus', exon_count=9, OMIM=611484, no_hdr0=''),
  10898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10898, CurrentID=0, Status='live', Symbol='CPSF4', Aliases='CPSF30, NAR, NEB-1, NEB1', description='cleavage and polyadenylation specific factor 4', other_designations='CPSF 30 kDa subunit|NS1 effector domain-binding protein 1|cleavage and polyadenylation specific factor 4, 30kDa|cleavage and polyadenylation specificity factor 30 kDa subunit|no arches homolog|no arches-like zinc finger protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99438940, end_position_on_the_genomic_accession=99457373, orientation='plus', exon_count=9, OMIM=603052, no_hdr0=''),
  10899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10899, CurrentID=0, Status='live', Symbol='JTB', Aliases='HJTB, HSPC222, PAR, hJT', description='jumping translocation breakpoint', other_designations='jumping translocation breakpoint protein|prostate androgen-regulated protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153974269, end_position_on_the_genomic_accession=153977975, orientation='minus', exon_count=5, OMIM=604671, no_hdr0=''),
  10900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10900, CurrentID=0, Status='live', Symbol='RUNDC3A', Aliases='RAP2IP, RPIP-8, RPIP8', description='RUN domain containing 3A', other_designations='RaP2 interacting protein 8|rap2-interacting protein 8', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44307918, end_position_on_the_genomic_accession=44318670, orientation='plus', exon_count=12, OMIM=605448, no_hdr0=''),
  10901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10901, CurrentID=0, Status='live', Symbol='DHRS4', Aliases='CR, NRDR, PHCR, PSCD, SCAD-SRL, SDR-SRL, SDR25C1, SDR25C2', description='dehydrogenase/reductase (SDR family) member 4', other_designations='NADPH-dependent carbonyl reductase/NADP-retinol dehydrogenase|NADPH-dependent retinol dehydrogenase/reductase|dehydrogenase/reductase (SDR family) member 4 like 2A|peroxisomal short-chain alcohol dehydrogenase|short chain dehydrogenase/reductase family 25C member 2|short chain dehydrogenase/reductase family 25C, member 1|short chain dehydrogenase/reductase family 25C, member 2|short-chain dehydrogenase/reductase family member 4', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23953735, end_position_on_the_genomic_accession=23969279, orientation='plus', exon_count=9, OMIM=611596, no_hdr0=''),
  10902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10902, CurrentID=0, Status='live', Symbol='BRD8', Aliases='SMAP, SMAP2, p120', description='bromodomain containing 8', other_designations='skeletal muscle abundant protein 2|thyroid hormone receptor coactivating protein of 120 kDa|trCP120', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138139766, end_position_on_the_genomic_accession=138178669, orientation='minus', exon_count=30, OMIM=602848, no_hdr0=''),
  10903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10903, CurrentID=0, Status='live', Symbol='MTMR11', Aliases='CRA', description='myotubularin related protein 11', other_designations='cisplatin resistance associated|cisplatin resistance-associated protein|hCRA', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149928650, end_position_on_the_genomic_accession=149936898, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  10904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10904, CurrentID=0, Status='live', Symbol='BLCAP', Aliases='BC10', description='bladder cancer associated protein', other_designations='bladder cancer related protein (10kD)', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37517417, end_position_on_the_genomic_accession=37527931, orientation='minus', exon_count=5, OMIM=613110, no_hdr0=''),
  10905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10905, CurrentID=0, Status='live', Symbol='MAN1A2', Aliases='MAN1B', description='mannosidase, alpha, class 1A, member 2', other_designations='alpha-1,2-mannosidase IB|alpha1,2-mannosidase|processing alpha-1,2-mannosidase IB', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117367393, end_position_on_the_genomic_accession=117528872, orientation='plus', exon_count=14, OMIM=604345, no_hdr0=''),
  10906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10906, CurrentID=0, Status='live', Symbol='TRAFD1', Aliases='FLN29', description='TRAF-type zinc finger domain containing 1', other_designations='FLN29 gene product', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112125545, end_position_on_the_genomic_accession=112153604, orientation='plus', exon_count=12, OMIM=613197, no_hdr0=''),
  10907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10907, CurrentID=0, Status='live', Symbol='TXNL4A', Aliases='BMKS, DIB1, DIM1, SNRNP15, TXNL4, U5-15kD', description='thioredoxin like 4A', other_designations='DIM1 protein homolog|spliceosomal U5 snRNP-specific 15 kDa protein|thioredoxin-like 4|thioredoxin-like 4A|thioredoxin-like U5 snRNP protein U5-15kD', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79972221, end_position_on_the_genomic_accession=79988592, orientation='minus', exon_count=4, OMIM=611595, no_hdr0=''),
  10908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10908, CurrentID=0, Status='live', Symbol='PNPLA6', Aliases='BNHS, LNMS, NTE, NTEMND, OMCS, SPG39, iPLA2delta, sws', description='patatin like phospholipase domain containing 6', other_designations='patatin-like phospholipase domain containing 6|patatin-like phospholipase domain-containing protein 6', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7534152, end_position_on_the_genomic_accession=7561767, orientation='plus', exon_count=37, OMIM=603197, no_hdr0=''),
  10910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10910, CurrentID=0, Status='live', Symbol='SUGT1', Aliases='SGT1', description='SGT1 homolog, MIS12 kinetochore complex assembly cochaperone', other_designations='SGT1, suppressor of G2 allele of SKP1|SGT1B protein|putative 40-6-3 protein|suppressor of G2 allele of SKP1, S. cerevisiae, homolog of', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52652696, end_position_on_the_genomic_accession=52688298, orientation='plus', exon_count=15, OMIM=604098, no_hdr0=''),
  10911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10911, CurrentID=0, Status='live', Symbol='UTS2', Aliases='PRO1068, U-II, UCN2, UII', description='urotensin 2', other_designations='prepro U-II|urotensin II', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7847612, end_position_on_the_genomic_accession=7913610, orientation='minus', exon_count=6, OMIM=604097, no_hdr0=''),
  10912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10912, CurrentID=0, Status='live', Symbol='GADD45G', Aliases='CR6, DDIT2, GADD45gamma, GRP17', description='growth arrest and DNA damage inducible gamma', other_designations='DDIT-2|DNA damage-inducible transcript 2 protein|GADD45-gamma|cytokine-responsive protein CR6|gadd-related protein, 17 kD|growth arrest and DNA-damage-inducible, gamma', map_location='9q22.1-q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89605012, end_position_on_the_genomic_accession=89606555, orientation='plus', exon_count=4, OMIM=604949, no_hdr0=''),
  10913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10913, CurrentID=0, Status='live', Symbol='EDAR', Aliases='DL, ECTD10A, ECTD10B, ED1R, ED3, ED5, EDA-A1R, EDA1R, EDA3, HRM1', description='ectodysplasin A receptor', other_designations='EDA-A1 receptor|anhidrotic ectodysplasin receptor 1|downless homolog|downless, mouse, homolog of|ectodermal dysplasia receptor|ectodysplasin 1, anhidrotic receptor|ectodysplasin-A receptor', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108894471, end_position_on_the_genomic_accession=108989372, orientation='minus', exon_count=14, OMIM=604095, no_hdr0=''),
  10914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10914, CurrentID=0, Status='live', Symbol='PAPOLA', Aliases='PAP', description='poly(A) polymerase alpha', other_designations='PAP-alpha|polynucleotide adenylyltransferase alpha', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96502376, end_position_on_the_genomic_accession=96567116, orientation='plus', exon_count=23, OMIM=605553, no_hdr0=''),
  10915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10915, CurrentID=0, Status='live', Symbol='TCERG1', Aliases='CA150, TAF2S, Urn1', description='transcription elongation regulator 1', other_designations='TATA box binding protein (TBP)-associated factor, RNA polymerase II, S, 150kD|TATA box-binding protein-associated factor 2S|co-activator of 150 kDa|transcription factor CA150', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146447310, end_position_on_the_genomic_accession=146511508, orientation='plus', exon_count=27, OMIM=605409, no_hdr0=''),
  10916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10916, CurrentID=0, Status='live', Symbol='MAGED2', Aliases='11B6, BCG-1, BCG1, HCA10, MAGE-D2', description='MAGE family member D2', other_designations='MAGE-D2 antigen|breast cancer-associated gene 1 protein|hepatocellular carcinoma-associated protein JCL-1|melanoma antigen family D, 2|melanoma antigen family D2', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54807599, end_position_on_the_genomic_accession=54816015, orientation='plus', exon_count=15, OMIM=300470, no_hdr0=''),
  10917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10917, CurrentID=0, Status='live', Symbol='BTNL3', Aliases='BTN9.1, BTNLR', description='butyrophilin-like 3', other_designations='butyrophilin-like receptor', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180988845, end_position_on_the_genomic_accession=181006727, orientation='plus', exon_count=8, OMIM=606192, no_hdr0=''),
  10919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10919, CurrentID=0, Status='live', Symbol='EHMT2', Aliases='BAT8, C6orf30, G9A, GAT8, KMT1C, NG36', description='euchromatic histone-lysine N-methyltransferase 2', other_designations='G9A histone methyltransferase|H3-K9-HMTase 3|HLA-B associated transcript 8|ankyrin repeat-containing protein|histone H3-K9 methyltransferase 3|histone-lysine N-methyltransferase, H3 lysine-9 specific 3|lysine N-methyltransferase 1C', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31879759, end_position_on_the_genomic_accession=31897707, orientation='minus', exon_count=28, OMIM=604599, no_hdr0=''),
  10920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10920, CurrentID=0, Status='live', Symbol='COPS8', Aliases='COP9, CSN8, SGN8', description='COP9 signalosome subunit 8', other_designations='COP9 constitutive photomorphogenic homolog subunit 8|COP9 homolog|JAB1-containing signalosome subunit 8|hCOP9|signalosome subunit 8', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237085441, end_position_on_the_genomic_accession=237098846, orientation='plus', exon_count=9, OMIM=616011, no_hdr0=''),
  10921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10921, CurrentID=0, Status='live', Symbol='RNPS1', Aliases='E5.1', description='RNA binding protein S1, serine-rich domain', other_designations='RNA-binding protein S1, serine-rich domain|SR protein|SR-related protein LDC2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2253116, end_position_on_the_genomic_accession=2268412, orientation='minus', exon_count=11, OMIM=606447, no_hdr0=''),
  10922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10922, CurrentID=0, Status='live', Symbol='FASTK', Aliases='FAST', description='Fas-activated serine/threonine kinase', other_designations='FAST kinase', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151076620, end_position_on_the_genomic_accession=151080883, orientation='minus', exon_count=11, OMIM=606965, no_hdr0=''),
  10923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10923, CurrentID=0, Status='live', Symbol='SUB1', Aliases='P15, PC4, p14', description='SUB1 homolog, transcriptional regulator', other_designations='SUB1 homolog|activated RNA polymerase II transcription cofactor 4|positive cofactor 4', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32585499, end_position_on_the_genomic_accession=32604079, orientation='plus', exon_count=6, OMIM=600503, no_hdr0=''),
  10924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10924, CurrentID=0, Status='live', Symbol='SMPDL3A', Aliases='ASM3A, ASML3a, yR36GH4.1', description='sphingomyelin phosphodiesterase acid-like 3A', other_designations='0610010C24Rik|ASM-like phosphodiesterase 3a|sphingomyelin phosphodiesterase, acid-like 3A', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122788456, end_position_on_the_genomic_accession=122809720, orientation='plus', exon_count=9, OMIM=610728, no_hdr0=''),
  105372977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372977, CurrentID=0, Status='live', Symbol='LOC105372977', Aliases='', description='uncharacterized LOC105372977', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27187511, end_position_on_the_genomic_accession=27189260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10926, CurrentID=0, Status='live', Symbol='DBF4', Aliases='ASK, CHIFA, ZDBF1, DBF4', description='DBF4 zinc finger', other_designations='DBF4 homolog|DBF4 zinc finger A|DBF4-type zinc finger-containing protein 1|activator of S phase kinase|chiffon homolog A|zinc finger, DBF-type containing 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87876229, end_position_on_the_genomic_accession=87909541, orientation='plus', exon_count=12, OMIM=604281, no_hdr0=''),
  10927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10927, CurrentID=0, Status='live', Symbol='SPIN1', Aliases='SPIN', description='spindlin 1', other_designations='ovarian cancer-related protein|spindlin1', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88388382, end_position_on_the_genomic_accession=88478708, orientation='plus', exon_count=6, OMIM=609936, no_hdr0=''),
  10928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10928, CurrentID=0, Status='live', Symbol='RALBP1', Aliases='RIP1, RLIP1, RLIP76', description='ralA binding protein 1', other_designations='76 kDa Ral-interacting protein|DNP-SG ATPase|dinitrophenyl S-glutathione ATPase|ral-interacting protein 1', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9475532, end_position_on_the_genomic_accession=9538108, orientation='plus', exon_count=10, OMIM=605801, no_hdr0=''),
  10929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10929, CurrentID=0, Status='live', Symbol='SRSF8', Aliases='DSM-1, SFRS2B, SRP46', description='serine/arginine-rich splicing factor 8', other_designations='SR splicing factor 8|pre-mRNA-splicing factor SRP46|splicing factor, arginine/serine-rich 2B|splicing factor, arginine/serine-rich, 46kD', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95066877, end_position_on_the_genomic_accession=95071227, orientation='plus', exon_count=1, OMIM=603269, no_hdr0=''),
  10930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10930, CurrentID=0, Status='live', Symbol='APOBEC2', Aliases='ARCD1, ARP1', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 2', other_designations='apolipoprotein B mRNA editing enzyme, catalytic polypeptide 2|mRNA(cytosine(6666)) deaminase 2|probable C->U-editing enzyme APOBEC-2', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41053201, end_position_on_the_genomic_accession=41064891, orientation='plus', exon_count=3, OMIM=604797, no_hdr0=''),
  10933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10933, CurrentID=0, Status='live', Symbol='MORF4L1', Aliases='Eaf3, FWP006, HsT17725, MEAF3, MORFRG15, MRG15, S863-6', description='mortality factor 4 like 1', other_designations='Esa1p-associated factor 3 homolog|MORF-related gene 15 protein|MORF-related gene on chromosome 15|protein MSL3-1|transcription factor-like protein MRG15', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78872781, end_position_on_the_genomic_accession=78897739, orientation='plus', exon_count=13, OMIM=607303, no_hdr0=''),
  10934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10934, CurrentID=0, Status='live', Symbol='MORF4', Aliases='CSR, CSRB, SEN, SEN1', description='mortality factor 4 (pseudogene)', other_designations='mortality factor 4 like 1 pseudogene|senescence (cellular)-related 1|senescence-related, cellular, 1', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173615936, end_position_on_the_genomic_accession=173616643, orientation='minus', exon_count=0, OMIM=116960, no_hdr0=''),
  10935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10935, CurrentID=0, Status='live', Symbol='PRDX3', Aliases='AOP-1, AOP1, HBC189, MER5, PRO1748, SP-22, prx-III', description='peroxiredoxin 3', other_designations='antioxidant protein 1|peroxiredoxin III|peroxiredoxin-3|protein MER5 homolog', map_location='10q25-q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119167699, end_position_on_the_genomic_accession=119178865, orientation='minus', exon_count=7, OMIM=604769, no_hdr0=''),
  10936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10936, CurrentID=0, Status='live', Symbol='GPR75', Aliases='GPRchr2, WI31133', description='G protein-coupled receptor 75', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53852913, end_position_on_the_genomic_accession=53860033, orientation='minus', exon_count=2, OMIM=606704, no_hdr0=''),
  285812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285812, CurrentID=0, Status='live', Symbol='LOC285812', Aliases='', description='uncharacterized LOC285812', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  10938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10938, CurrentID=0, Status='live', Symbol='EHD1', Aliases='H-PAST, HPAST1, PAST, PAST1', description='EH domain containing 1', other_designations='EH-domain containing 1|PAST homolog 1|testilin', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64852727, end_position_on_the_genomic_accession=64879713, orientation='minus', exon_count=7, OMIM=605888, no_hdr0=''),
  10939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10939, CurrentID=0, Status='live', Symbol='AFG3L2', Aliases='SCA28, SPAX5', description='AFG3-like AAA ATPase 2', other_designations='AFG3 ATPase family gene 3-like 2|AFG3 ATPase family member 3-like 2|ATPase family gene 3, yeast|paraplegin-like protein', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12328944, end_position_on_the_genomic_accession=12377309, orientation='minus', exon_count=17, OMIM=604581, no_hdr0=''),
  10940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10940, CurrentID=0, Status='live', Symbol='POP1', Aliases='', description='POP1 homolog, ribonuclease P/MRP subunit', other_designations='hPOP1|processing of precursor 1, ribonuclease P/MRP subunit|processing of precursors 1', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98117293, end_position_on_the_genomic_accession=98159841, orientation='plus', exon_count=19, OMIM=602486, no_hdr0=''),
  10941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10941, CurrentID=0, Status='live', Symbol='UGT2A1', Aliases='UDPGT2A1', description='UDP glucuronosyltransferase 2 family, polypeptide A1, complex locus', other_designations='UDP glycosyltransferase 2 family, polypeptide A1|uridine diphosphate glycosyltransferase 2 family, member A1', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69588417, end_position_on_the_genomic_accession=69653249, orientation='minus', exon_count=9, OMIM=604716, no_hdr0=''),
  10942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10942, CurrentID=0, Status='live', Symbol='PRSS21', Aliases='ESP-1, ESP1, TEST1, TESTISIN', description='protease, serine 21', other_designations='eosinophil serine protease 1|protease, serine, 21 (testisin)|serine protease from eosinophils', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2817163, end_position_on_the_genomic_accession=2821722, orientation='plus', exon_count=6, OMIM=608159, no_hdr0=''),
  10943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10943, CurrentID=0, Status='live', Symbol='MSL3', Aliases='MSL3L1', description='male-specific lethal 3 homolog (Drosophila)', other_designations='MSL3-like 1|male-specific lethal-3 protein-like 1', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11758159, end_position_on_the_genomic_accession=11775753, orientation='plus', exon_count=17, OMIM=300609, no_hdr0=''),
  10944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10944, CurrentID=0, Status='live', Symbol='C11orf58', Aliases='IMAGE145052, SMAP', description='chromosome 11 open reading frame 58', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16738601, end_position_on_the_genomic_accession=16758354, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  10945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10945, CurrentID=0, Status='live', Symbol='KDELR1', Aliases='ERD2, ERD2.1, HDEL, PM23', description='KDEL endoplasmic reticulum protein retention receptor 1', other_designations='ER lumen protein retaining receptor 1|KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 1|KDEL receptor 1|putative MAPK-activating protein PM23', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48382570, end_position_on_the_genomic_accession=48397689, orientation='minus', exon_count=7, OMIM=131235, no_hdr0=''),
  10946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10946, CurrentID=0, Status='live', Symbol='SF3A3', Aliases='PRP9, PRPF9, SAP61, SF3a60', description='splicing factor 3a subunit 3', other_designations='SAP 61|pre-mRNA splicing factor SF3a (60kD)|spliceosome associated protein 61|spliceosome-associated protein 61|splicing factor 3a, subunit 3, 60kD|splicing factor 3a, subunit 3, 60kDa', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37956975, end_position_on_the_genomic_accession=37990490, orientation='minus', exon_count=17, OMIM=605596, no_hdr0=''),
  10947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10947, CurrentID=0, Status='live', Symbol='AP3M2', Aliases='AP47B, CLA20, P47B', description='adaptor related protein complex 3, mu 2 subunit', other_designations='HA1 47 kDa subunit homolog 2|HA1 47kDA subunit homolog 2|adapter-related protein complex 3 mu-2 subunit|adapter-related protein complex 3 subunit mu-2|adaptor-related protein complex 3 subunit mu-2|adaptor-related protein complex 3, mu 2 subunit|clathrin assembly protein assembly protein complex 1 medium chain homolog 2|clathrin assembly protein assembly protein complex 3 mu-2 medium chain|clathrin coat assembly protein AP47 homolog 2|clathrin coat-associated protein AP47 homolog 2|clathrin-associated protein AP47 homolog 2|golgi adaptor AP-1 47 kDA protein homolog 2|mu3B-adaptin', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42152946, end_position_on_the_genomic_accession=42171183, orientation='plus', exon_count=10, OMIM=610469, no_hdr0=''),
  10948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10948, CurrentID=0, Status='live', Symbol='STARD3', Aliases='CAB1, MLN64, es64', description='StAR related lipid transfer domain containing 3', other_designations='MLN 64|START domain-containing protein 3|StAR-related lipid transfer (START) domain containing 3|StAR-related lipid transfer domain containing 3|metastatic lymph node gene 64 protein|metastatic lymph node protein 64|steroidogenic acute regulatory protein related', map_location='17q11-q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39637080, end_position_on_the_genomic_accession=39664201, orientation='plus', exon_count=16, OMIM=607048, no_hdr0=''),
  10949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10949, CurrentID=0, Status='live', Symbol='HNRNPA0', Aliases='HNRPA0', description='heterogeneous nuclear ribonucleoprotein A0', other_designations='hnRNA binding protein|hnRNP A0', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137751384, end_position_on_the_genomic_accession=137754350, orientation='minus', exon_count=1, OMIM=609409, no_hdr0=''),
  10950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10950, CurrentID=0, Status='live', Symbol='BTG3', Aliases='ANA, TOB5, TOB55, TOFA', description='BTG family member 3', other_designations='B-cell translocation gene 3|BTG family, member 3|abundant in neuroepithelium area protein|protein Tob5', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17593650, end_position_on_the_genomic_accession=17612950, orientation='minus', exon_count=8, OMIM=605674, no_hdr0=''),
  10951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10951, CurrentID=0, Status='live', Symbol='CBX1', Aliases='CBX, HP1-BETA, HP1Hs-beta, HP1Hsbeta, M31, MOD1, p25beta', description='chromobox 1', other_designations='HP1 beta homolog|chromobox homolog 1 (HP1 beta homolog Drosophila )|heterochromatin protein 1 homolog beta|heterochromatin protein 1-beta|heterochromatin protein p25 beta|modifier 1 protein', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48070052, end_position_on_the_genomic_accession=48101521, orientation='minus', exon_count=6, OMIM=604511, no_hdr0=''),
  10952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10952, CurrentID=0, Status='live', Symbol='SEC61B', Aliases='', description='Sec61 translocon beta subunit', other_designations='Sec61 beta subunit|Sec61 complex, beta subunit|protein translocation complex beta|protein transport protein SEC61 beta subunit', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99222288, end_position_on_the_genomic_accession=99230619, orientation='plus', exon_count=4, OMIM=609214, no_hdr0=''),
  10953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10953, CurrentID=0, Status='live', Symbol='TOMM34', Aliases='HTOM34P, TOM34, URCC3', description='translocase of outer mitochondrial membrane 34', other_designations='hTom34|outer mitochondrial membrane translocase (34kD)|translocase of outer membrane 34 kDa subunit', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44942130, end_position_on_the_genomic_accession=44960473, orientation='minus', exon_count=9, OMIM=616049, no_hdr0=''),
  10954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10954, CurrentID=0, Status='live', Symbol='PDIA5', Aliases='PDIR', description='protein disulfide isomerase family A member 5', other_designations='protein disulfide isomerase family A, member 5|protein disulfide isomerase-associated 5|protein disulfide isomerase-related protein', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123067009, end_position_on_the_genomic_accession=123162106, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  10955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10955, CurrentID=0, Status='live', Symbol='SERINC3', Aliases='AIGP1, DIFF33, SBBI99, TDE, TDE1, TMS-1', description='serine incorporator 3', other_designations='placental transmembrane protein (mouse testicular tumor differentially expressed)|transmembrane protein SBBI99|tumor differentially expressed 1|tumor differentially expressed protein 1|tumour differentially expressed 1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44496223, end_position_on_the_genomic_accession=44522085, orientation='minus', exon_count=11, OMIM=607165, no_hdr0=''),
  10956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10956, CurrentID=0, Status='live', Symbol='OS9', Aliases='ERLEC2, OS-9', description='osteosarcoma amplified 9, endoplasmic reticulum lectin', other_designations='amplified in osteosarcoma 9|endoplasmic reticulum lectin 2|erlectin 2|osteosarcoma amplified 9, endoplasmic reticulum associated protein', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57693955, end_position_on_the_genomic_accession=57721557, orientation='plus', exon_count=15, OMIM=609677, no_hdr0=''),
  10957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10957, CurrentID=0, Status='live', Symbol='PNRC1', Aliases='B4-2, PNAS-145, PROL2, PRR2', description='proline rich nuclear receptor coactivator 1', other_designations='proline rich 2 protein|proline-rich polypeptide 2|proline-rich protein 2|proline-rich protein with nuclear targeting signal|protein B4-2', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89080710, end_position_on_the_genomic_accession=89085160, orientation='plus', exon_count=2, OMIM=606714, no_hdr0=''),
  10959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10959, CurrentID=0, Status='live', Symbol='TMED2', Aliases='P24A, RNP24, p24', description='transmembrane p24 trafficking protein 2', other_designations='coated vesicle membrane protein|membrane protein p24A|p24 family protein beta-1|p24beta1|transmembrane emp24 domain trafficking protein 2', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123584529, end_position_on_the_genomic_accession=123598144, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  10960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10960, CurrentID=0, Status='live', Symbol='LMAN2', Aliases='C5orf8, GP36B, VIP36', description='lectin, mannose-binding 2', other_designations='glycoprotein GP36b|vesicular integral protein of 36 kDa|vesicular integral-membrane protein 36', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177331562, end_position_on_the_genomic_accession=177351884, orientation='minus', exon_count=8, OMIM=609551, no_hdr0=''),
  10961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10961, CurrentID=0, Status='live', Symbol='ERP29', Aliases='C12orf8, ERp28, ERp31, HEL-S-107, PDI-DB, PDIA9', description='endoplasmic reticulum protein 29', other_designations='endoplasmic reticulum lumenal protein ERp28|endoplasmic reticulum resident protein 28|endoplasmic reticulum resident protein 31|epididymis secretory protein Li 107|protein disulfide isomerase family A, member 9', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112013348, end_position_on_the_genomic_accession=112023220, orientation='plus', exon_count=3, OMIM=602287, no_hdr0=''),
  10962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10962, CurrentID=0, Status='live', Symbol='MLLT11', Aliases='AF1Q', description='myeloid/lymphoid or mixed-lineage leukemia; translocated to, 11', other_designations='ALL1 fused gene from chromosome 1q|ALL1-fused gene from chromosome 1q|myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 11', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151059675, end_position_on_the_genomic_accession=151068497, orientation='plus', exon_count=2, OMIM=604684, no_hdr0=''),
  10963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10963, CurrentID=0, Status='live', Symbol='STIP1', Aliases='HEL-S-94n, HOP, IEF-SSP-3521, P60, STI1, STI1L', description='stress induced phosphoprotein 1', other_designations='Hsp70/Hsp90-organizing protein|NY-REN-11 antigen|epididymis secretory sperm binding protein Li 94n|hsc70/Hsp90-organizing protein|renal carcinoma antigen NY-REN-11|stress-induced phosphoprotein 1|transformation-sensitive protein IEF SSP 3521', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64185272, end_position_on_the_genomic_accession=64204548, orientation='plus', exon_count=15, OMIM=605063, no_hdr0=''),
  10964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10964, CurrentID=0, Status='live', Symbol='IFI44L', Aliases='C1orf29, GS3686', description='interferon induced protein 44 like', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78620382, end_position_on_the_genomic_accession=78646145, orientation='plus', exon_count=10, OMIM=613975, no_hdr0=''),
  10965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10965, CurrentID=0, Status='live', Symbol='ACOT2', Aliases='CTE-IA, CTE1A, MTE1, PTE2, PTE2A, ZAP128', description='acyl-CoA thioesterase 2', other_designations='acyl-coenzyme A thioester hydrolase 2a|long-chain acyl-CoA thioesterase 2|mitochondrial acyl-CoA thioesterase 1|peroxisomal long-chain acyl-coA thioesterase 2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73567620, end_position_on_the_genomic_accession=73575658, orientation='plus', exon_count=4, OMIM=609972, no_hdr0=''),
  10966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10966, CurrentID=0, Status='live', Symbol='RAB40B', Aliases='RAR, SEC4L', description='RAB40B, member RAS oncogene family', other_designations='GTP-binding protein homologous to Saccharomyces cerevisiae SEC4|SOCS box-containing protein RAR|protein Rar', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82654959, end_position_on_the_genomic_accession=82698722, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  10969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10969, CurrentID=0, Status='live', Symbol='EBNA1BP2', Aliases='EBP2, NOBP, P40', description='EBNA1 binding protein 2', other_designations='EBNA1-binding protein 2|nuclear FGF3 binding protein|nucleolar protein p40|nucleolar protein p40; homolog of yeast EBNA1-binding protein', map_location='1p35-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43164174, end_position_on_the_genomic_accession=43172570, orientation='minus', exon_count=9, OMIM=614443, no_hdr0=''),
  10970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10970, CurrentID=0, Status='live', Symbol='CKAP4', Aliases='CLIMP-63, ERGIC-63, p63', description='cytoskeleton-associated protein 4', other_designations='63 kDa membrane protein|63-kDa cytoskeleton-linking membrane protein|transmembrane protein (63kD), endoplasmic reticulum/Golgi intermediate compartment|type-II transmembrane protein p63', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106237881, end_position_on_the_genomic_accession=106247935, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10971, CurrentID=0, Status='live', Symbol='YWHAQ', Aliases='14-3-3, 1C5, HS1', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, theta', other_designations='14-3-3 protein T-cell|14-3-3 protein tau|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, theta polypeptide', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9583967, end_position_on_the_genomic_accession=9631055, orientation='minus', exon_count=6, OMIM=609009, no_hdr0=''),
  10972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10972, CurrentID=0, Status='live', Symbol='TMED10', Aliases='P24(DELTA), S31I125, S31III125, TMP21, Tmp-21-I, p23', description='transmembrane p24 trafficking protein 10', other_designations='21 kDa transmembrane trafficking protein|21 kDa transmembrane-trafficking protein|p24 family protein delta-1|p24delta|p24delta1|transmembrane emp24-like trafficking protein 10|transmembrane protein Tmp21', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75131468, end_position_on_the_genomic_accession=75176646, orientation='minus', exon_count=5, OMIM=605406, no_hdr0=''),
  10973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10973, CurrentID=0, Status='live', Symbol='ASCC3', Aliases='ASC1p200, HELIC1, RNAH', description='activating signal cointegrator 1 complex subunit 3', other_designations='ASC-1 complex subunit P200|RNA helicase family|helicase, ATP binding 1|trip4 complex subunit p200', map_location='6q16', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100508194, end_position_on_the_genomic_accession=100881372, orientation='minus', exon_count=43, OMIM=614217, no_hdr0=''),
  10974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10974, CurrentID=0, Status='live', Symbol='ADIRF', Aliases='AFRO, APM2, C10orf116, apM-2', description='adipogenesis regulatory factor', other_designations='adipogenesis factor rich in obesity|adipose most abundant gene transcript 2 protein|adipose specific 2|adipose-specific protein 2', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86968431, end_position_on_the_genomic_accession=86970909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10975, CurrentID=0, Status='live', Symbol='UQCR11', Aliases='0710008D09Rik, QCR10, UQCR', description='ubiquinol-cytochrome c reductase, complex III subunit XI', other_designations='complex III subunit 10|complex III subunit XI|ubiquinol cytochrome c reductase, 6.4kDa subunit|ubiquinol-cytochrome c reductase complex 6.4 kDa protein|ubiquinol-cytochrome c reductase, 6.4kDa subunit', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1597155, end_position_on_the_genomic_accession=1605484, orientation='minus', exon_count=3, OMIM=609711, no_hdr0=''),
  10978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10978, CurrentID=0, Status='live', Symbol='CLP1', Aliases='HEAB, hClp1', description='cleavage and polyadenylation factor I subunit 1', other_designations='ATP/GTP-binding protein|homolog of yeast CFIA subunit Clp1p|polyadenylation factor Clp1|polynucleotide kinase Clp1|pre-mRNA cleavage complex II protein Clp1', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57657744, end_position_on_the_genomic_accession=57661865, orientation='plus', exon_count=3, OMIM=608757, no_hdr0=''),
  10979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10979, CurrentID=0, Status='live', Symbol='FERMT2', Aliases='KIND2, MIG2, PLEKHC1, UNC112, UNC112B, mig-2', description='fermitin family member 2', other_designations='PH domain-containing family C member 1|kindlin 2|kindlin-2|mitogen inducible gene 2 protein|mitogen-inducible gene 2 protein|pleckstrin homology domain containing, family C (with FERM domain) member 1|pleckstrin homology domain containing, family C member 1|pleckstrin homology domain-containing family C member 1', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52857271, end_position_on_the_genomic_accession=52951097, orientation='minus', exon_count=18, OMIM=607746, no_hdr0=''),
  10980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10980, CurrentID=0, Status='live', Symbol='COPS6', Aliases='CSN6, MOV34-34KD', description='COP9 signalosome subunit 6', other_designations='COP9 constitutive photomorphogenic homolog subunit 6|COP9 subunit 6 (MOV34 homolog, 34 kD)|H_NH0506M12.12|JAB1-containing signalosome subunit 6|MOV34 homolog, 34 kD|SGN6|hVIP|signalosome subunit 6|vpr-interacting protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100088960, end_position_on_the_genomic_accession=100092200, orientation='plus', exon_count=10, OMIM=614729, no_hdr0=''),
  10981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10981, CurrentID=0, Status='live', Symbol='RAB32', Aliases='', description='RAB32, member RAS oncogene family', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146543692, end_position_on_the_genomic_accession=146554950, orientation='plus', exon_count=3, OMIM=612906, no_hdr0=''),
  10982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10982, CurrentID=0, Status='live', Symbol='MAPRE2', Aliases='EB1, EB2, RP1', description='microtubule associated protein RP/EB family member 2', other_designations='APC-binding protein EB1|APC-binding protein EB2|T-cell activation protein, EB1 family|end-binding protein 2|microtubule-associated protein, RP/EB family, member 2', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34976928, end_position_on_the_genomic_accession=35143470, orientation='plus', exon_count=11, OMIM=605789, no_hdr0=''),
  10983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10983, CurrentID=0, Status='live', Symbol='CCNI', Aliases='CCNI1, CYC1, CYI', description='cyclin I', other_designations='cyclin ITI', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77047155, end_position_on_the_genomic_accession=77076013, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  10984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10984, CurrentID=0, Status='live', Symbol='KCNQ1OT1', Aliases='KCNQ1-AS2, KCNQ10T1, Kncq1, KvDMR1, KvLQT1-AS, LIT1, NCRNA00012', description='KCNQ1 opposite strand/antisense transcript 1 (non-protein coding)', other_designations='KCNQ1 antisense RNA 2 (non-protein coding)|KCNQ1 overlapping transcript 1 (non-protein coding)|long QT intronic transcript 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2608328, end_position_on_the_genomic_accession=2699998, orientation='minus', exon_count=1, OMIM=604115, no_hdr0=''),
  10985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10985, CurrentID=0, Status='live', Symbol='GCN1', Aliases='GCN1LL1, PRIC295, GCN1', description='GCN1, eIF2 alpha kinase activator homolog', other_designations='GCN1 (general control of amino-acid synthesis 1, yeast)-like 1|GCN1 eIF2 alpha kinase activator homolog|GCN1 general control of amino-acid synthesis 1-like 1|GCN1-like protein 1|hsGCN1|peroxisome proliferator activated receptor interacting complex protein', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120127210, end_position_on_the_genomic_accession=120194710, orientation='minus', exon_count=58, OMIM=605614, no_hdr0=''),
  10986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10986, CurrentID=0, Status='live', Symbol='IGAD1', Aliases='', description='immunoglobulin A (IgA) deficiency susceptibility 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=137100, no_hdr0=''),
  10987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10987, CurrentID=0, Status='live', Symbol='COPS5', Aliases='CSN5, JAB1, MOV-34, SGN5', description='COP9 signalosome subunit 5', other_designations='38 kDa Mov34 homolog|COP9 constitutive photomorphogenic homolog subunit 5|jun activation domain-binding protein 1|signalosome subunit 5', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67043079, end_position_on_the_genomic_accession=67062327, orientation='minus', exon_count=8, OMIM=604850, no_hdr0=''),
  10988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10988, CurrentID=0, Status='live', Symbol='METAP2', Aliases='MAP2, MNPEP, p67eIF2', description='methionyl aminopeptidase 2', other_designations='eIF-2-associated p67 homolog|initiation factor 2-associated 67 kDa glycoprotein|peptidase M 2', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95474046, end_position_on_the_genomic_accession=95515839, orientation='plus', exon_count=11, OMIM=601870, no_hdr0=''),
  10989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10989, CurrentID=0, Status='live', Symbol='IMMT', Aliases='HMP, MINOS2, Mic60, P87, P87/89, P89, PIG4, PIG52', description='inner membrane protein, mitochondrial', other_designations='cell proliferation-inducing gene 4/52 protein|cell proliferation-inducing protein 52|mitochondrial inner membrane organizing system 2|mitochondrial inner membrane protein|mitofilin|motor protein|proliferation-inducing gene 4', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86143932, end_position_on_the_genomic_accession=86195770, orientation='minus', exon_count=15, OMIM=600378, no_hdr0=''),
  10990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10990, CurrentID=0, Status='live', Symbol='LILRB5', Aliases='CD85C, LIR-8, LIR8', description='leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 5', other_designations='CD85 antigen-like family member C|leukocyte immunoglobulin-like receptor 8', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54250405, end_position_on_the_genomic_accession=54257308, orientation='minus', exon_count=13, OMIM=604814, no_hdr0=''),
  10991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10991, CurrentID=0, Status='live', Symbol='SLC38A3', Aliases='G17, NAT1, SN1, SNAT3', description='solute carrier family 38 member 3', other_designations='N-system amino acid transporter 1|Na(+)-coupled neutral amino acid transporter 3|solute carrier family 38, member 3|system N amino acid transporter 1|system N1 Na+ and H+-coupled glutamine transporter', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50205246, end_position_on_the_genomic_accession=50220979, orientation='plus', exon_count=17, OMIM=604437, no_hdr0=''),
  10992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10992, CurrentID=0, Status='live', Symbol='SF3B2', Aliases='Cus1, SAP145, SF3B145, SF3b1, SF3b150', description='splicing factor 3b subunit 2', other_designations='SAP 145|pre-mRNA splicing factor SF3b 145 kDa subunit|pre-mRNA-splicing factor SF3b 145 kDa subunit|spliceosome associated protein 145|spliceosome-associated protein 145|splicing factor 3b, subunit 2, 145kD|splicing factor 3b, subunit 2, 145kDa', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66051775, end_position_on_the_genomic_accession=66068911, orientation='plus', exon_count=22, OMIM=605591, no_hdr0=''),
  10993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10993, CurrentID=0, Status='live', Symbol='SDS', Aliases='SDH', description='serine dehydratase', other_designations='L-serine ammonia-lyase|L-serine deaminase|L-serine dehydratase|L-threonine dehydratase|TDH', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113392445, end_position_on_the_genomic_accession=113403887, orientation='minus', exon_count=8, OMIM=182128, no_hdr0=''),
  10994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10994, CurrentID=0, Status='live', Symbol='ILVBL', Aliases='209L8, AHAS, ILV2H', description='ilvB (bacterial acetolactate synthase)-like', other_designations='acetolactate synthase homolog|ilvB-like protein', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15114974, end_position_on_the_genomic_accession=15125799, orientation='minus', exon_count=17, OMIM=605770, no_hdr0=''),
  10998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10998, CurrentID=0, Status='live', Symbol='SLC27A5', Aliases='ACSB, ACSVL6, BACS, BAL, FACVL3, FATP-5, FATP5, VLACSR, VLCS-H2, VLCSH2', description='solute carrier family 27 (fatty acid transporter), member 5', other_designations='BA-CoA ligase|VLACS-related|bile acid-CoA ligase|cholate--CoA ligase|fatty acid transport protein 5|fatty-acid-Coenzyme A ligase, very long-chain 3|solute carrier family 27 member 5|very long-chain acyl-CoA synthetase homolog 2|very long-chain acyl-CoA synthetase-related protein', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58498333, end_position_on_the_genomic_accession=58512065, orientation='minus', exon_count=10, OMIM=603314, no_hdr0=''),
  10999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10999, CurrentID=0, Status='live', Symbol='SLC27A4', Aliases='ACSVL4, FATP4, IPS', description='solute carrier family 27 (fatty acid transporter), member 4', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128340560, end_position_on_the_genomic_accession=128361470, orientation='plus', exon_count=13, OMIM=604194, no_hdr0=''),
  11000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11000, CurrentID=0, Status='live', Symbol='SLC27A3', Aliases='ACSVL3, FATP3, VLCS-3', description='solute carrier family 27 (fatty acid transporter), member 3', other_designations='fatty acid transport protein 3|solute carrier family 27 member 3|very long-chain acyl-CoA synthetase homolog 3', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153775292, end_position_on_the_genomic_accession=153780157, orientation='plus', exon_count=10, OMIM=604193, no_hdr0=''),
  11001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11001, CurrentID=0, Status='live', Symbol='SLC27A2', Aliases='ACSVL1, FACVL1, FATP2, HsT17226, VLACS, VLCS, hFACVL1', description='solute carrier family 27 (fatty acid transporter), member 2', other_designations='FATP-2|THCA-CoA ligase|fatty acid transport protein 2|fatty-acid-coenzyme A ligase, very long-chain 1|long-chain-fatty-acid--CoA ligase|solute carrier family 27 member 2|very long-chain fatty-acid-coenzyme A ligase 1|very long-chain-fatty-acid-CoA ligase|very-long-chain acyl-CoA synthetase', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50182196, end_position_on_the_genomic_accession=50236392, orientation='plus', exon_count=10, OMIM=603247, no_hdr0=''),
  11004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11004, CurrentID=0, Status='live', Symbol='KIF2C', Aliases='CT139, KNSL6, MCAK', description='kinesin family member 2C', other_designations='kinesin-like protein 6|mitotic centromere-associated kinesin', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44739712, end_position_on_the_genomic_accession=44767767, orientation='plus', exon_count=22, OMIM=604538, no_hdr0=''),
  11005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11005, CurrentID=0, Status='live', Symbol='SPINK5', Aliases='LEKTI, LETKI, NETS, NS, VAKTI', description='serine peptidase inhibitor, Kazal type 5', other_designations='lympho-epithelial Kazal-type-related inhibitor|lymphoepithelial Kazal-type-related inhibitor', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148025688, end_position_on_the_genomic_accession=148137366, orientation='plus', exon_count=35, OMIM=605010, no_hdr0=''),
  11006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11006, CurrentID=0, Status='live', Symbol='LILRB4', Aliases='CD85K, ILT-3, ILT3, LIR-5, LIR5', description='leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 4', other_designations='CD85 antigen-like family member K|immunoglobulin-like transcript 3|leukocyte immunoglobulin-like receptor 5|monocyte inhibitory receptor HM18', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54643835, end_position_on_the_genomic_accession=54670359, orientation='plus', exon_count=13, OMIM=604821, no_hdr0=''),
  11007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11007, CurrentID=0, Status='live', Symbol='CCDC85B', Aliases='DIPA', description='coiled-coil domain containing 85B', other_designations='delta-interacting protein A|hepatitis delta antigen interacting protein A|hepatitis delta antigen-interacting protein A', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65890404, end_position_on_the_genomic_accession=65891635, orientation='plus', exon_count=1, OMIM=605360, no_hdr0=''),
  11009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11009, CurrentID=0, Status='live', Symbol='IL24', Aliases='C49A, FISP, IL10B, MDA7, MOB5, ST16', description='interleukin 24', other_designations='IL-4-induced secreted protein|melanocyte-associated Mda-7|melanoma differentiation-associated gene 7 protein|suppression of tumorigenicity 16 (melanoma differentiation)', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206897098, end_position_on_the_genomic_accession=206904139, orientation='plus', exon_count=7, OMIM=604136, no_hdr0=''),
  11010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11010, CurrentID=0, Status='live', Symbol='GLIPR1', Aliases='CRISP7, GLIPR, RTVP1', description='GLI pathogenesis-related 1', other_designations='GLI pathogenesis-related 1 (glioma)|gliPR 1|protein RTVP-1|related to testis-specific, vespid, and pathogenesis proteins 1|testes-specific vespid and pathogenesis protein 1', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75480733, end_position_on_the_genomic_accession=75501936, orientation='plus', exon_count=7, OMIM=602692, no_hdr0=''),
  11011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11011, CurrentID=0, Status='live', Symbol='TLK2', Aliases='HsHPK, PKU-ALPHA', description='tousled like kinase 2', other_designations='tousled-like kinase 2', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62477081, end_position_on_the_genomic_accession=62615481, orientation='plus', exon_count=30, OMIM=608439, no_hdr0=''),
  11012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11012, CurrentID=0, Status='live', Symbol='KLK11', Aliases='PRSS20, TLSP', description='kallikrein related peptidase 11', other_designations='hK11|hippostasin|kallikrein 11|kallikrein-related peptidase 11|serine protease 20|trypsin-like protease', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51022231, end_position_on_the_genomic_accession=51028034, orientation='minus', exon_count=10, OMIM=604434, no_hdr0=''),
  11013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11013, CurrentID=0, Status='live', Symbol='TMSB15A', Aliases='TMSB15, TMSB15B, TMSL8, TMSNB, Tb15, TbNB', description='thymosin beta 15a', other_designations='NB thymosin beta|Thymosin beta-15B|thymosin, beta, identified in neuroblastoma cells|thymosin-like 8|thymosin-like protein 8', map_location='Xq21.33-q22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102513682, end_position_on_the_genomic_accession=102516771, orientation='minus', exon_count=3, OMIM=300939, no_hdr0=''),
  11014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11014, CurrentID=0, Status='live', Symbol='KDELR2', Aliases='ELP-1, ERD2.2', description='KDEL endoplasmic reticulum protein retention receptor 2', other_designations='(Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 2|ER lumen protein retaining receptor 2|ERD-2-like protein|ERD2-like protein 1|KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 2|KDEL receptor 2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6461081, end_position_on_the_genomic_accession=6484218, orientation='minus', exon_count=5, OMIM=609024, no_hdr0=''),
  11015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11015, CurrentID=0, Status='live', Symbol='KDELR3', Aliases='ERD2L3', description='KDEL endoplasmic reticulum protein retention receptor 3', other_designations='ER lumen protein retaining receptor 3|KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 3|KDEL receptor 3', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38468062, end_position_on_the_genomic_accession=38483447, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11016, CurrentID=0, Status='live', Symbol='ATF7', Aliases='ATFA', description='activating transcription factor 7', other_designations='transcription factor ATF-A', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53507856, end_position_on_the_genomic_accession=53626415, orientation='minus', exon_count=15, OMIM=606371, no_hdr0=''),
  11017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11017, CurrentID=0, Status='live', Symbol='SNRNP27', Aliases='27K, RY1', description='small nuclear ribonucleoprotein, U4/U6.U5 27kDa subunit', other_designations='U4/U6.U5 tri-snRNP-associated 27 kDa protein|U4/U6.U5 tri-snRNP-associated protein 3|U4/U6.U5-27K|putative nucleic acid binding protein RY-1|small nuclear ribonucleoprotein 27kDa (U4/U6.U5)', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69893943, end_position_on_the_genomic_accession=69905236, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11018, CurrentID=0, Status='live', Symbol='TMED1', Aliases='IL1RL1LG, Il1rl1l, Tp24', description='transmembrane p24 trafficking protein 1', other_designations='IL1RL1-binding protein|interleukin 1 receptor-like 1 ligand|p24 family protein gamma-1|putative T1/ST2 receptor-binding protein|transmembrane emp24 protein transport domain containing 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10832067, end_position_on_the_genomic_accession=10836307, orientation='minus', exon_count=6, OMIM=605395, no_hdr0=''),
  11019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11019, CurrentID=0, Status='live', Symbol='LIAS', Aliases='HUSSY-01, LAS, LIP1, LS, PDHLD', description='lipoic acid synthetase', other_designations='LS|lip-syn|lipoate synthase', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39459024, end_position_on_the_genomic_accession=39477653, orientation='plus', exon_count=11, OMIM=607031, no_hdr0=''),
  11020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11020, CurrentID=0, Status='live', Symbol='IFT27', Aliases='BBS19, RABL4, RAYL', description='intraflagellar transport 27', other_designations='RAB, member of RAS oncogene family-like 4|rab-like protein 4', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36758202, end_position_on_the_genomic_accession=36776133, orientation='minus', exon_count=8, OMIM=615870, no_hdr0=''),
  11021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11021, CurrentID=0, Status='live', Symbol='RAB35', Aliases='H-ray, RAB1C, RAY', description='RAB35, member RAS oncogene family', other_designations='GTP-binding protein RAY|ras-related protein rab-1c (GTP-binding protein ray)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120095094, end_position_on_the_genomic_accession=120116839, orientation='minus', exon_count=6, OMIM=604199, no_hdr0=''),
  11022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11022, CurrentID=0, Status='live', Symbol='TDRKH', Aliases='TDRD2', description='tudor and KH domain containing', other_designations='putative RNA binding protein|tudor and KH domain containing protein|tudor domain containing 2|tudor domain-containing protein 2', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151770264, end_position_on_the_genomic_accession=151790534, orientation='minus', exon_count=16, OMIM=609501, no_hdr0=''),
  11023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11023, CurrentID=0, Status='live', Symbol='VAX1', Aliases='MCOPS11', description='ventral anterior homeobox 1', other_designations='', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117128521, end_position_on_the_genomic_accession=117138301, orientation='minus', exon_count=5, OMIM=604294, no_hdr0=''),
  11024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11024, CurrentID=0, Status='live', Symbol='LILRA1', Aliases='CD85I, LIR-6, LIR6', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 1', other_designations='CD85 antigen-like family member I|leukocyte immunoglobulin-like receptor 6', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54593576, end_position_on_the_genomic_accession=54602724, orientation='plus', exon_count=12, OMIM=604810, no_hdr0=''),
  11025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11025, CurrentID=0, Status='live', Symbol='LILRB3', Aliases='CD85A, HL9, ILT-5, ILT5, LILRA6, LIR-3, LIR3, PIRB', description='leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 3', other_designations='CD85 antigen-like family member A|immunoglobulin-like transcript 5|leukocyte immunoglobulin-like receptor 3|leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 6|monocyte inhibitory receptor HL9', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54216278, end_position_on_the_genomic_accession=54223087, orientation='minus', exon_count=14, OMIM=604820, no_hdr0=''),
  11026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11026, CurrentID=0, Status='live', Symbol='LILRA3', Aliases='CD85E, HM31, HM43, ILT-6, ILT6, LIR-4, LIR4', description='leukocyte immunoglobulin-like receptor, subfamily A (without TM domain), member 3', other_designations='CD85 antigen-like family member E|immunoglobulin-like transcript 6|leukocyte immunoglobulin-like receptor 4|leukocyte immunoglobulin-like receptor A3|monocyte inhibitory receptor HM43/HM31', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604818, no_hdr0=''),
  11027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11027, CurrentID=0, Status='live', Symbol='LILRA2', Aliases='CD85H, ILT1, LIR-7, LIR7', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 2', other_designations='CD85 antigen-like family member H|immunoglobulin-like transcript 1|leukocyte immunoglobulin-like receptor 7|leukocyte immunoglobulin-like receptor subfamily A member 2 soluble', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54572997, end_position_on_the_genomic_accession=54587560, orientation='plus', exon_count=10, OMIM=604812, no_hdr0=''),
  100271918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271918, CurrentID=0, Status='live', Symbol='MKI67P1', Aliases='', description='marker of proliferation Ki-67 pseudogene 1', other_designations='antigen identified by monoclonal antibody Ki-67 pseudogene 1', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46561300, end_position_on_the_genomic_accession=46562748, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  11030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11030, CurrentID=0, Status='live', Symbol='RBPMS', Aliases='HERMES', description='RNA binding protein with multiple splicing', other_designations='heart and RRM expressed sequence', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30384465, end_position_on_the_genomic_accession=30572261, orientation='plus', exon_count=17, OMIM=601558, no_hdr0=''),
  11031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11031, CurrentID=0, Status='live', Symbol='RAB31', Aliases='Rab22B', description='RAB31, member RAS oncogene family', other_designations='ras-related protein Rab-22B', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9708231, end_position_on_the_genomic_accession=9862556, orientation='plus', exon_count=7, OMIM=605694, no_hdr0=''),
  11033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11033, CurrentID=0, Status='live', Symbol='ADAP1', Aliases='CENTA1, GCS1L, p42IP4', description='ArfGAP with dual PH domains 1', other_designations='centaurin, alpha 1|centaurin-alpha|centaurin-alpha-1|cnt-a1|putative MAPK-activating protein PM25', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=897900, end_position_on_the_genomic_accession=955407, orientation='minus', exon_count=15, OMIM=608114, no_hdr0=''),
  11034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11034, CurrentID=0, Status='live', Symbol='DSTN', Aliases='ACTDP, ADF, HEL32, bA462D18.2', description='destrin (actin depolymerizing factor)', other_designations='actin-depolymerizing factor|bA462D18.2 (destrin (actin depolymerizing factor ADF) (ACTDP))|epididymis luminal protein 32', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17569173, end_position_on_the_genomic_accession=17608242, orientation='plus', exon_count=6, OMIM=609114, no_hdr0=''),
  11035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11035, CurrentID=0, Status='live', Symbol='RIPK3', Aliases='RIP3', description='receptor interacting serine/threonine kinase 3', other_designations='RIP-3|RIP-like protein kinase 3|receptor interacting protein 3|receptor-interacting protein 3|receptor-interacting serine-threonine kinase 3', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24336021, end_position_on_the_genomic_accession=24340036, orientation='minus', exon_count=10, OMIM=605817, no_hdr0=''),
  11036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11036, CurrentID=0, Status='live', Symbol='GTF2A1L', Aliases='ALF', description='general transcription factor IIA 1-like', other_designations='GTF2A1-like factor|TFIIA large subunit isoform ALF|TFIIA-alpha/beta-like factor|general transcription factor II A, 1-like factor|general transcription factor IIA, 1-like', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48617780, end_position_on_the_genomic_accession=48679609, orientation='plus', exon_count=9, OMIM=605358, no_hdr0=''),
  11037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11037, CurrentID=0, Status='live', Symbol='STON1', Aliases='SALF, SBLF, STN1, STNB1', description='stonin 1', other_designations='stoned B homolog 1|stoned B-like factor|stoned-b1', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48530169, end_position_on_the_genomic_accession=48598515, orientation='plus', exon_count=5, OMIM=605357, no_hdr0=''),
  11039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11039, CurrentID=0, Status='live', Symbol='SMA4', Aliases='SMA3, b55C20.2', description='glucuronidase, beta pseudogene', other_designations='Beta-glucuronidase-like protein SMA4|Putative beta-glucuronidase-like protein MGC156142', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70127462, end_position_on_the_genomic_accession=70290177, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11040, CurrentID=0, Status='live', Symbol='PIM2', Aliases='', description='Pim-2 proto-oncogene, serine/threonine kinase', other_designations='pim-2 oncogene|pim-2h|proto-oncogene Pim-2 (serine threonine kinase)|serine/threonine protein kinase pim-2', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48913182, end_position_on_the_genomic_accession=48919136, orientation='minus', exon_count=6, OMIM=300295, no_hdr0=''),
  11041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11041, CurrentID=0, Status='live', Symbol='B4GAT1', Aliases='B3GN-T1, B3GNT1, B3GNT6, BETA3GNTI, MDDGA13, iGAT, iGNT', description='beta-1,4-glucuronyltransferase 1', other_designations='N-acetyllactosaminide beta-1,3-N-acetylglucosaminyltransferase|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 1|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 6|beta-1,3-N-acetylglucosaminyltransferase bGnT-6|i-beta-1,3-N-acetylglucosaminyltransferase|poly-N-acetyllactosamine extension enzyme', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66345372, end_position_on_the_genomic_accession=66347690, orientation='minus', exon_count=2, OMIM=605517, no_hdr0=''),
  11042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11042, CurrentID=0, Status='live', Symbol='SMA5', Aliases='', description='glucuronidase, beta pseudogene', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70481043, end_position_on_the_genomic_accession=70585722, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  11043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11043, CurrentID=0, Status='live', Symbol='MID2', Aliases='FXY2, MRX101, RNF60, TRIM1', description='midline 2', other_designations='RING finger protein 60|midin 2|midin-2|midline defect 2|midline-2|tripartite motif protein 1|tripartite motif-containing protein 1', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107824647, end_position_on_the_genomic_accession=107931637, orientation='plus', exon_count=11, OMIM=300204, no_hdr0=''),
  11044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11044, CurrentID=0, Status='live', Symbol='PAPD7', Aliases='LAK-1, LAK1, POLK, POLS, TRF4, TRF4-1, TRF41, TUTASE5', description='PAP associated domain containing 7', other_designations='DNA polymerase kappa|DNA polymerase sigma|PAP-associated domain-containing protein 7|TRAMP-like complex polyadenylate polymerase|TUTase 5|polymerase (DNA directed) sigma|polymerase (DNA-directed) sigma|terminal uridylyltransferase 5|topoisomerase-related function protein 4-1', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6713953, end_position_on_the_genomic_accession=6757048, orientation='plus', exon_count=14, OMIM=605198, no_hdr0=''),
  11045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11045, CurrentID=0, Status='live', Symbol='UPK1A', Aliases='TSPAN21, UP1A, UPIA, UPKA', description='uroplakin 1A', other_designations='tetraspanin-21|tspan-21|uroplakin Ia', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35666516, end_position_on_the_genomic_accession=35678485, orientation='plus', exon_count=9, OMIM=611557, no_hdr0=''),
  11046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11046, CurrentID=0, Status='live', Symbol='SLC35D2', Aliases='HFRC1, SQV7L, UGTrel8, hfrc', description='solute carrier family 35 (UDP-GlcNAc/UDP-glucose transporter), member D2', other_designations='SQV7-like protein|UDP-N-acetylglucosamine transporter|UDP-galactose transporter-related protein 8|fringe connection|homolog of Fringe connection protein 1|solute carrier family 35, member D2', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96313437, end_position_on_the_genomic_accession=96383710, orientation='minus', exon_count=14, OMIM=609182, no_hdr0=''),
  11047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11047, CurrentID=0, Status='live', Symbol='ADRM1', Aliases='ARM-1, ARM1, GP110', description='adhesion regulating molecule 1', other_designations='110 kDa cell membrane glycoprotein|M(r) 110,000 surface antigen|proteasome regulatory particle non-ATPase 13|rpn13 homolog', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62302056, end_position_on_the_genomic_accession=62308862, orientation='plus', exon_count=10, OMIM=610650, no_hdr0=''),
  11049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11049, CurrentID=0, Status='live', Symbol='NUS1P3', Aliases='YDD19', description='NUS1 dehydrodolichyl diphosphate synthase subunit pseudogene 3', other_designations='nuclear undecaprenyl pyrophosphate synthase 1 homolog (S. cerevisiae) pseudogene 3|nuclear undecaprenyl pyrophosphate synthase 1 pseudogene', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24327708, end_position_on_the_genomic_accession=24332108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  404266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404266, CurrentID=0, Status='live', Symbol='HOXB-AS3', Aliases='', description='HOXB cluster antisense RNA 3', other_designations='HOXB cluster antisense RNA 3 (non-protein coding)', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48590420, end_position_on_the_genomic_accession=48606414, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11051, CurrentID=0, Status='live', Symbol='NUDT21', Aliases='CFIM25, CPSF5', description='nudix hydrolase 21', other_designations='CPSF 25 kDa subunit|cleavage and polyadenylation specific factor 5, 25 kD subunit|cleavage and polyadenylation specific factor 5, 25 kDa|cleavage and polyadenylation specificity factor 25 kDa subunit|cleavage factor Im complex 25 kDa subunit|nucleoside diphosphate-linked moiety X motif 21|nudix (nucleoside diphosphate linked moiety X)-type motif 21|nudix motif 21|pre-mRNA cleavage factor Im (25kD)|pre-mRNA cleavage factor Im 25 kDa subunit|pre-mRNA cleavage factor Im, 25kD subunit', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56429136, end_position_on_the_genomic_accession=56451349, orientation='minus', exon_count=7, OMIM=604978, no_hdr0=''),
  11052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11052, CurrentID=0, Status='live', Symbol='CPSF6', Aliases='CFIM, CFIM68, HPBRII-4, HPBRII-7', description='cleavage and polyadenylation specific factor 6', other_designations='CPSF 68 kDa subunit|cleavage and polyadenylation specific factor 6, 68kDa|cleavage and polyadenylation specificity factor 68 kDa subunit|cleavage factor Im complex 68 kDa subunit|pre-mRNA cleavage factor I, 68kD subunit|pre-mRNA cleavage factor Im (68kD)|pre-mRNA cleavage factor Im 68 kDa subunit|protein HPBRII-4/7', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69239537, end_position_on_the_genomic_accession=69274358, orientation='plus', exon_count=11, OMIM=604979, no_hdr0=''),
  100271922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271922, CurrentID=0, Status='live', Symbol='BFIS4', Aliases='BFIC4', description='Convulsions, benign familial infantile, 4', other_designations='', map_location='1p36.12-p35.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612627, no_hdr0=''),
  11054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11054, CurrentID=0, Status='live', Symbol='OGFR', Aliases='', description='opioid growth factor receptor', other_designations='protein 7-60|zeta-type opioid receptor', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62804825, end_position_on_the_genomic_accession=62814000, orientation='plus', exon_count=7, OMIM=606459, no_hdr0=''),
  11055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11055, CurrentID=0, Status='live', Symbol='ZPBP', Aliases='ZPBP1', description='zona pellucida binding protein', other_designations='zona pellucida binding protein 1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=49850322, end_position_on_the_genomic_accession=50093264, orientation='minus', exon_count=13, OMIM=608498, no_hdr0=''),
  11056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11056, CurrentID=0, Status='live', Symbol='DDX52', Aliases='HUSSY19, ROK1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 52', other_designations='DEAD box protein 52', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37609739, end_position_on_the_genomic_accession=37643464, orientation='minus', exon_count=17, OMIM=612500, no_hdr0=''),
  11057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11057, CurrentID=0, Status='live', Symbol='ABHD2', Aliases='HS1-2, LABH2, PHPS1-2', description='abhydrolase domain containing 2', other_designations='alpha/beta hydrolase domain containing protein 2|lung alpha/beta hydrolase 2|protein PHPS1-2', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89088150, end_position_on_the_genomic_accession=89202360, orientation='plus', exon_count=16, OMIM=612196, no_hdr0=''),
  11059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11059, CurrentID=0, Status='live', Symbol='WWP1', Aliases='AIP5, Tiul1, hSDRP1', description='WW domain containing E3 ubiquitin protein ligase 1', other_designations='Nedd-4-like ubiquitin-protein ligase|TGIF-interacting ubiquitin ligase 1|WW domain-containing protein 1|atrophin-1 interacting protein 5|atrophin-1-interacting protein 5', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86342736, end_position_on_the_genomic_accession=86468265, orientation='plus', exon_count=25, OMIM=602307, no_hdr0=''),
  11060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11060, CurrentID=0, Status='live', Symbol='WWP2', Aliases='AIP2, WWp2-like', description='WW domain containing E3 ubiquitin protein ligase 2', other_designations='atrophin-1 interacting protein 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69762284, end_position_on_the_genomic_accession=69941741, orientation='plus', exon_count=28, OMIM=602308, no_hdr0=''),
  11061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11061, CurrentID=0, Status='live', Symbol='LECT1', Aliases='BRICD3, CHM-I, CHM1, MYETS1', description='leukocyte cell derived chemotaxin 1', other_designations='BRICHOS domain containing 3|chondromodulin I|chondromodulin-1|chondromodulin-I|multiple myeloma tumor suppressor 1', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52703264, end_position_on_the_genomic_accession=52739812, orientation='minus', exon_count=7, OMIM=605147, no_hdr0=''),
  11062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11062, CurrentID=0, Status='live', Symbol='DUS4L', Aliases='DUS4, PP35', description='dihydrouridine synthase 4-like', other_designations='protein similar to E.coli yhdg and R. capsulatus nifR3', map_location='7q22-q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107563957, end_position_on_the_genomic_accession=107578523, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11063, CurrentID=0, Status='live', Symbol='SOX30', Aliases='', description='SRY-box 30', other_designations='SRY (sex determining region Y)-box 30|SRY box 30|Sox30 protein type II', map_location='5q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157624796, end_position_on_the_genomic_accession=157672756, orientation='minus', exon_count=8, OMIM=606698, no_hdr0=''),
  11064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11064, CurrentID=0, Status='live', Symbol='CNTRL', Aliases='CEP1, CEP110, FAN, bA165P4.1', description='centriolin', other_designations='110 kDa centrosomal protein|bA165P4.1 (ortholog of mouse Ma2a8)|bA165P4.2 (centrosomal protein 1)|centriole associated protein|centrosomal protein 110kDa|centrosomal protein of 110 kDa', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121074955, end_position_on_the_genomic_accession=121177610, orientation='plus', exon_count=47, OMIM=605496, no_hdr0=''),
  404281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404281, CurrentID=0, Status='live', Symbol='YY2', Aliases='ZNF631', description='YY2 transcription factor', other_designations='transcription factor yin yang 2|yin and yang 2|zinc finger protein 631', map_location='Xp22.2-p22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21855987, end_position_on_the_genomic_accession=21858727, orientation='plus', exon_count=1, OMIM=300570, no_hdr0=''),
  11066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11066, CurrentID=0, Status='live', Symbol='SNRNP35', Aliases='HM-1, U1SNRNPBP', description='small nuclear ribonucleoprotein, U11/U12 35kDa subunit', other_designations='U1 snRNP binding protein homolog|U1 snRNP-binding protein homolog|U1-snRNP binding protein homolog|U11/U12 snRNP 35 kDa protein|U11/U12 snRNP 35K|U11/U12-35K|protein HM-1|small nuclear ribonucleoprotein 35kDa (U11/U12)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123458026, end_position_on_the_genomic_accession=123472362, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11067, CurrentID=0, Status='live', Symbol='C10orf10', Aliases='DEPP, FIG, Fseg', description='chromosome 10 open reading frame 10', other_designations='decidual protein induced by progesterone|fasting induced|fasting-induced gene protein|fasting-induced protein|fat-specific expressed', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44976261, end_position_on_the_genomic_accession=44978882, orientation='minus', exon_count=2, OMIM=611309, no_hdr0=''),
  11068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11068, CurrentID=0, Status='live', Symbol='CYB561D2', Aliases='101F6, TSP10, XXcos-LUCA11.4', description='cytochrome b561 family member D2', other_designations='cytochrome b561 family, member D2|putative tumor suppressor protein 101F6', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50350695, end_position_on_the_genomic_accession=50368197, orientation='plus', exon_count=7, OMIM=607068, no_hdr0=''),
  11069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11069, CurrentID=0, Status='live', Symbol='RAPGEF4', Aliases='CAMP-GEFII, CGEF2, EPAC, EPAC 2, EPAC2, Nbla00496', description='Rap guanine nucleotide exchange factor 4', other_designations='RAP guanine-nucleotide-exchange factor (GEF) 4|Rap guanine nucleotide exchange factor (GEF) 4|cAMP-regulated guanine nucleotide exchange factor II|exchange factor directly activated by cAMP 2|exchange protein directly activated by cAMP 2|putative protein product of Nbla00496', map_location='2q31-q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172735279, end_position_on_the_genomic_accession=173052893, orientation='plus', exon_count=35, OMIM=606058, no_hdr0=''),
  11070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11070, CurrentID=0, Status='live', Symbol='TMEM115', Aliases='PL6', description='transmembrane protein 115', other_designations='PP6|placental protein 6', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50354749, end_position_on_the_genomic_accession=50359508, orientation='minus', exon_count=2, OMIM=607069, no_hdr0=''),
  11072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11072, CurrentID=0, Status='live', Symbol='DUSP14', Aliases='MKP-L, MKP6', description='dual specificity phosphatase 14', other_designations='MAP kinase phosphatase 6|MKP-1 like protein tyrosine phosphatase|MKP-1-like protein tyrosine phosphatase|MKP-6|mitogen-activated protein kinase phosphatase 6', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37488561, end_position_on_the_genomic_accession=37513501, orientation='plus', exon_count=5, OMIM=606618, no_hdr0=''),
  11073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11073, CurrentID=0, Status='live', Symbol='TOPBP1', Aliases='TOP2BP1', description='topoisomerase (DNA) II binding protein 1', other_designations='DNA topoisomerase II-beta-binding protein 1|DNA topoisomerase II-binding protein 1', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133600605, end_position_on_the_genomic_accession=133661918, orientation='minus', exon_count=28, OMIM=607760, no_hdr0=''),
  11074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11074, CurrentID=0, Status='live', Symbol='TRIM31', Aliases='C6orf13, HCG1, HCGI, RNF', description='tripartite motif containing 31', other_designations='ring finger protein|tripartite motif-containing 31 gamma|tripartite motif-containing protein 31', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30102892, end_position_on_the_genomic_accession=30113090, orientation='minus', exon_count=9, OMIM=609316, no_hdr0=''),
  11075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11075, CurrentID=0, Status='live', Symbol='STMN2', Aliases='SCG10, SCGN10', description='stathmin 2', other_designations='neuron-specific growth-associated protein|neuronal growth-associated protein (silencer element)|stathmin-like 2|superior cervical ganglia, neural specific 10|superior cervical ganglion-10 protein', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79610814, end_position_on_the_genomic_accession=79666175, orientation='plus', exon_count=8, OMIM=600621, no_hdr0=''),
  11076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11076, CurrentID=0, Status='live', Symbol='TPPP', Aliases='TPPP/p251, p24, p25, p25alpha, TPPP', description='tubulin polymerization promoting protein', other_designations='25 kDa brain-specific protein|brain specific protein p25 alpha|glycogen synthase kinase 3 (GSK3) inhibitor p24|p25-alpha', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=659862, end_position_on_the_genomic_accession=693395, orientation='minus', exon_count=5, OMIM=608773, no_hdr0=''),
  11077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11077, CurrentID=0, Status='live', Symbol='HSF2BP', Aliases='', description='heat shock transcription factor 2 binding protein', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43529186, end_position_on_the_genomic_accession=43659499, orientation='minus', exon_count=10, OMIM=604554, no_hdr0=''),
  11078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11078, CurrentID=0, Status='live', Symbol='TRIOBP', Aliases='DFNB28, HRIHFB2122, TAP68, TARA, dJ37E16.4', description='TRIO and F-actin binding protein', other_designations='protein Tara|tara-like protein|trio-associated repeat on actin', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37696988, end_position_on_the_genomic_accession=37776556, orientation='plus', exon_count=26, OMIM=609761, no_hdr0=''),
  11079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11079, CurrentID=0, Status='live', Symbol='RER1', Aliases='', description='retention in endoplasmic reticulum sorting receptor 1', other_designations='RER1 retention in endoplasmic reticulum 1 homolog', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2391314, end_position_on_the_genomic_accession=2405444, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11080, CurrentID=0, Status='live', Symbol='DNAJB4', Aliases='DNAJW, DjB4, HLJ1', description='DnaJ heat shock protein family (Hsp40) member B4', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 4|DnaJ-like heat shock protein 40|HSP40 homolog|heat shock 40 kDa protein 1 homolog|heat shock protein 40 homolog|human liver DnaJ-like protein', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77979217, end_position_on_the_genomic_accession=78018721, orientation='plus', exon_count=4, OMIM=611327, no_hdr0=''),
  11081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11081, CurrentID=0, Status='live', Symbol='KERA', Aliases='CNA2, KTN, SLRR2B', description='keratocan', other_designations='keratan sulfate proteoglycan keratocan', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91050491, end_position_on_the_genomic_accession=91058354, orientation='minus', exon_count=3, OMIM=603288, no_hdr0=''),
  11082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11082, CurrentID=0, Status='live', Symbol='ESM1', Aliases='endocan', description='endothelial cell-specific molecule 1', other_designations='ESM-1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54977867, end_position_on_the_genomic_accession=54985586, orientation='minus', exon_count=3, OMIM=601521, no_hdr0=''),
  11083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11083, CurrentID=0, Status='live', Symbol='DIDO1', Aliases='BYE1, C20orf158, DATF-1, DATF1, DIDO2, DIDO3, DIO-1, DIO1, dJ885L7.8', description='death inducer-obliterator 1', other_designations='death-associated transcription factor 1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62877734, end_position_on_the_genomic_accession=62937952, orientation='minus', exon_count=20, OMIM=604140, no_hdr0=''),
  11085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11085, CurrentID=0, Status='live', Symbol='ADAM30', Aliases='svph4', description='ADAM metallopeptidase domain 30', other_designations='a disintegrin and metalloproteinase domain 30', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119893533, end_position_on_the_genomic_accession=119896524, orientation='minus', exon_count=1, OMIM=604779, no_hdr0=''),
  11086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11086, CurrentID=0, Status='live', Symbol='ADAM29', Aliases='CT73, svph1', description='ADAM metallopeptidase domain 29', other_designations='a disintegrin and metalloproteinase domain 29|cancer/testis antigen 73|metallaproteinase-disintegrin (ADAM29)', map_location='4q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174918358, end_position_on_the_genomic_accession=174978180, orientation='plus', exon_count=9, OMIM=604778, no_hdr0=''),
  11091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11091, CurrentID=0, Status='live', Symbol='WDR5', Aliases='BIG-3, CFAP89, SWD3', description='WD repeat domain 5', other_designations='BMP2-induced 3-kb gene protein|SWD3, Set1c WD40 repeat protein, homolog|WD-repeat protein 5|cilia and flagella associated protein 89', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134135679, end_position_on_the_genomic_accession=134159972, orientation='plus', exon_count=15, OMIM=609012, no_hdr0=''),
  11092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11092, CurrentID=0, Status='live', Symbol='C9orf9', Aliases='Mast', description='chromosome 9 open reading frame 9', other_designations='Rsb66 homolog', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132878895, end_position_on_the_genomic_accession=132890034, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11093, CurrentID=0, Status='live', Symbol='ADAMTS13', Aliases='ADAM-TS13, ADAMTS-13, C9orf8, VWFCP, vWF-CP', description='ADAM metallopeptidase with thrombospondin type 1 motif 13', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 13|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 13|vWF-cleaving protease|von Willebrand factor-cleaving protease', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133414339, end_position_on_the_genomic_accession=133459403, orientation='plus', exon_count=31, OMIM=604134, no_hdr0=''),
  11094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11094, CurrentID=0, Status='live', Symbol='CACFD1', Aliases='C9orf7, D9S2135, FLOWER', description='calcium channel flower domain containing 1', other_designations='calcium channel flower domain-containing protein 1', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133459965, end_position_on_the_genomic_accession=133470787, orientation='plus', exon_count=9, OMIM=613104, no_hdr0=''),
  11095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11095, CurrentID=0, Status='live', Symbol='ADAMTS8', Aliases='ADAM-TS8, METH2', description='ADAM metallopeptidase with thrombospondin type 1 motif 8', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 8|ADAM-TS 8|ADAMTS-8|METH-2|METH-8|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 8', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130404923, end_position_on_the_genomic_accession=130428644, orientation='minus', exon_count=9, OMIM=605175, no_hdr0=''),
  11096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11096, CurrentID=0, Status='live', Symbol='ADAMTS5', Aliases='ADAM-TS 11, ADAM-TS 5, ADAM-TS5, ADAMTS-11, ADAMTS-5, ADAMTS11, ADMP-2', description='ADAM metallopeptidase with thrombospondin type 1 motif 5', other_designations='a disintegrin and metalloproteinase with thrombospondin motifs 11|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 5 (aggrecanase-2)|aggrecanase-2', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26917912, end_position_on_the_genomic_accession=26967120, orientation='minus', exon_count=8, OMIM=605007, no_hdr0=''),
  11097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11097, CurrentID=0, Status='live', Symbol='NUPL2', Aliases='CG1, NLP-1, NLP_1, hCG1', description='nucleoporin like 2', other_designations='H_RG271G13.9|NUP42 homolog|nucleoporin hCG1|nucleoporin-like protein 1', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23181827, end_position_on_the_genomic_accession=23201011, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11098, CurrentID=0, Status='live', Symbol='PRSS23', Aliases='SIG13, SPUVE, ZSIG13', description='protease, serine 23', other_designations='protease, serine, 23|putative secreted protein Zsig13|serine protease, umbilical endothelium', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86791059, end_position_on_the_genomic_accession=86952910, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11099, CurrentID=0, Status='live', Symbol='PTPN21', Aliases='PTPD1, PTPRL10', description='protein tyrosine phosphatase, non-receptor type 21', other_designations='dJ1175B15.2 (protein tyrosine phosphatase, non-receptor type 21)|protein tyrosine phosphatase D1|protein-tyrosine phosphatase D1', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88465778, end_position_on_the_genomic_accession=88554779, orientation='minus', exon_count=22, OMIM=603271, no_hdr0=''),
  11100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11100, CurrentID=0, Status='live', Symbol='HNRNPUL1', Aliases='E1B-AP5, E1BAP5, HNRPUL1', description='heterogeneous nuclear ribonucleoprotein U-like 1', other_designations='E1B 55kDa associated protein 5|E1B-55 kDa-associated protein 5|E1B-55kDa-associated protein 5|adenovirus early region 1B-associated protein 5', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41262476, end_position_on_the_genomic_accession=41307693, orientation='plus', exon_count=18, OMIM=605800, no_hdr0=''),
  11101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11101, CurrentID=0, Status='live', Symbol='ATE1', Aliases='', description='arginyltransferase 1', other_designations='R-transferase 1|arginine-tRNA--protein transferase 1|arginyl-tRNA-protein transferase', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121740421, end_position_on_the_genomic_accession=121928801, orientation='minus', exon_count=14, OMIM=607103, no_hdr0=''),
  11102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11102, CurrentID=0, Status='live', Symbol='RPP14', Aliases='P14', description='ribonuclease P/MRP 14kDa subunit', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58306245, end_position_on_the_genomic_accession=58320193, orientation='plus', exon_count=7, OMIM=606112, no_hdr0=''),
  11103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11103, CurrentID=0, Status='live', Symbol='KRR1', Aliases='HRB2, RIP-1', description='KRR1, small subunit (SSU) processome component, homolog (yeast)', other_designations='HIV-1 Rev binding protein 2|HIV-1 Rev-binding protein 2|KRR-R motif-containing protein 1|Rev interacting protein|rev-interacting protein 1', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75497639, end_position_on_the_genomic_accession=75511638, orientation='minus', exon_count=10, OMIM=612817, no_hdr0=''),
  11104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11104, CurrentID=0, Status='live', Symbol='KATNA1', Aliases='', description='katanin p60 (ATPase containing) subunit A 1', other_designations='katanin p60 (ATPase-containing) subunit A 1|katanin p60 subunit A1|p60 katanin', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149594873, end_position_on_the_genomic_accession=149649018, orientation='minus', exon_count=13, OMIM=606696, no_hdr0=''),
  11105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11105, CurrentID=0, Status='live', Symbol='PRDM7', Aliases='PFM4, ZNF910', description='PR domain containing 7', other_designations='PR domain zinc finger protein 7|PR-domain family protein 4', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90056566, end_position_on_the_genomic_accession=90077339, orientation='minus', exon_count=11, OMIM=609759, no_hdr0=''),
  11107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11107, CurrentID=0, Status='live', Symbol='PRDM5', Aliases='BCS2, PFM2', description='PR domain containing 5', other_designations='', map_location='4q25-q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120684928, end_position_on_the_genomic_accession=120922866, orientation='minus', exon_count=21, OMIM=614161, no_hdr0=''),
  11108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11108, CurrentID=0, Status='live', Symbol='PRDM4', Aliases='PFM1', description='PR domain containing 4', other_designations='PR domain-containing protein 4|PR-domain zinc-finger protein PFM1', map_location='12q23-q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107732866, end_position_on_the_genomic_accession=107761272, orientation='minus', exon_count=12, OMIM=605780, no_hdr0=''),
  11112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11112, CurrentID=0, Status='live', Symbol='HIBADH', Aliases='NS5ATP1', description='3-hydroxyisobutyrate dehydrogenase', other_designations="3'-hydroxyisobutyrate dehydrogenase", map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27525440, end_position_on_the_genomic_accession=27663001, orientation='minus', exon_count=9, OMIM=608475, no_hdr0=''),
  11113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11113, CurrentID=0, Status='live', Symbol='CIT', Aliases='CRIK, STK21', description='citron rho-interacting serine/threonine kinase', other_designations='citron (rho-interacting, serine/threonine kinase 21)|serine/threonine kinase 21|serine/threonine-protein kinase 21', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119685790, end_position_on_the_genomic_accession=119877318, orientation='minus', exon_count=51, OMIM=605629, no_hdr0=''),
  11116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11116, CurrentID=0, Status='live', Symbol='FGFR1OP', Aliases='FOP', description='FGFR1 oncogene partner', other_designations='fibroblast growth factor receptor 1 oncogene partner', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166999317, end_position_on_the_genomic_accession=167042418, orientation='plus', exon_count=13, OMIM=605392, no_hdr0=''),
  11117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11117, CurrentID=0, Status='live', Symbol='EMILIN1', Aliases='EMI, EMILIN, gp115', description='elastin microfibril interfacer 1', other_designations='elastin microfibril interface located protein|elastin microfibril interface-located protein 1', map_location='2p23.3-p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27078567, end_position_on_the_genomic_accession=27086404, orientation='plus', exon_count=8, OMIM=130660, no_hdr0=''),
  11118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11118, CurrentID=0, Status='live', Symbol='BTN3A2', Aliases='BT3.2, BTF4, BTN3.2, CD277', description='butyrophilin subfamily 3 member A2', other_designations='butyrophilin protein|butyrophilin, subfamily 3, member A2', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26365159, end_position_on_the_genomic_accession=26451653, orientation='plus', exon_count=12, OMIM=613594, no_hdr0=''),
  11119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11119, CurrentID=0, Status='live', Symbol='BTN3A1', Aliases='BT3.1, BTF5, BTN3.1, CD277', description='butyrophilin subfamily 3 member A1', other_designations='butyrophilin, subfamily 3, member A1|dJ45P21.3 (butyrophilin, subfamily 3, member A1)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26402237, end_position_on_the_genomic_accession=26415216, orientation='plus', exon_count=12, OMIM=613593, no_hdr0=''),
  11120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11120, CurrentID=0, Status='live', Symbol='BTN2A1', Aliases='BK14H9.1, BT2.1, BTF1, BTN2.1, DJ3E1.1', description='butyrophilin subfamily 2 member A1', other_designations='butyrophilin, subfamily 2, member A1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26457904, end_position_on_the_genomic_accession=26476621, orientation='plus', exon_count=9, OMIM=613590, no_hdr0=''),
  101929377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929377, CurrentID=0, Status='live', Symbol='LOC101929377', Aliases='', description='uncharacterized LOC101929377', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11122, CurrentID=0, Status='live', Symbol='PTPRT', Aliases='RPTPrho', description='protein tyrosine phosphatase, receptor type, T', other_designations='R-PTP-T|RPTP-rho|receptor protein tyrosine phosphatase|receptor-type tyrosine-protein phosphatase rho', map_location='20q12-q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42072752, end_position_on_the_genomic_accession=43189917, orientation='minus', exon_count=35, OMIM=608712, no_hdr0=''),
  11123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11123, CurrentID=0, Status='live', Symbol='RCAN3', Aliases='DSCR1L2, MCIP3, RCN3, hRCN3', description='RCAN family member 3', other_designations='Down syndrome candidate region 1-like 2|Down syndrome critical region gene 1-like 2|down syndrome candidate region 1-like protein 2|myocyte-enriched calcineurin-interacting protein 3|regulator of calcineurin 3 isoform 1b,2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24502351, end_position_on_the_genomic_accession=24537020, orientation='plus', exon_count=7, OMIM=605860, no_hdr0=''),
  11124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11124, CurrentID=0, Status='live', Symbol='FAF1', Aliases='CGI-03, HFAF1s, UBXD12, UBXN3A, hFAF1', description='Fas (TNFRSF6) associated factor 1', other_designations='TNFRSF6-associated factor 1|UBX domain protein 3A|UBX domain-containing protein 12|UBX domain-containing protein 3A', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50441263, end_position_on_the_genomic_accession=50960264, orientation='minus', exon_count=19, OMIM=604460, no_hdr0=''),
  11126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11126, CurrentID=0, Status='live', Symbol='CD160', Aliases='BY55, NK1, NK28', description='CD160 molecule', other_designations='CD160 transmembrane isoform|CD160-delta Ig|natural killer cell receptor BY55|natural killer cell receptor, immunoglobulin superfamily member', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145719433, end_position_on_the_genomic_accession=145739288, orientation='plus', exon_count=6, OMIM=604463, no_hdr0=''),
  11127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11127, CurrentID=0, Status='live', Symbol='KIF3A', Aliases='FLA10, KLP-20', description='kinesin family member 3A', other_designations='kinesin family protein 3A|microtubule plus end-directed kinesin motor 3A', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132689161, end_position_on_the_genomic_accession=132737633, orientation='minus', exon_count=20, OMIM=604683, no_hdr0=''),
  11128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11128, CurrentID=0, Status='live', Symbol='POLR3A', Aliases='ADDH, HLD7, RPC1, RPC155, hRPC155', description='polymerase (RNA) III (DNA directed) polypeptide A, 155kDa', other_designations='DNA-directed RNA polymerase III largest subunit|DNA-directed RNA polymerase III subunit A|RNA polymerase III 155 kDa subunit|RNA polymerase III subunit C1|RNA polymerase III subunit C160|RNA polymerase III subunit RPC155-D', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77975149, end_position_on_the_genomic_accession=78029540, orientation='minus', exon_count=31, OMIM=614258, no_hdr0=''),
  11129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11129, CurrentID=0, Status='live', Symbol='CLASRP', Aliases='CLASP, SFRS16, SWAP2', description='CLK4-associating serine/arginine rich protein', other_designations='Clk4 associating SR-related protein|splicing factor, arginine/serine-rich 16 (suppressor-of-white-apricot homolog, Drosophila)|suppressor of white apricot homolog 2|suppressor of white-apricot homolog 2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45039040, end_position_on_the_genomic_accession=45070956, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  11130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11130, CurrentID=0, Status='live', Symbol='ZWINT', Aliases='HZwint-1, KNTC2AP, SIP301, ZWINT', description='ZW10 interacting kinetochore protein', other_designations='SNAP25 interacting protein of 30 kDa|ZW10 interactor, kinetochore protein|human ZW10 interacting protein-1|zwint-1', map_location='10q21-q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=56357438, end_position_on_the_genomic_accession=56361302, orientation='minus', exon_count=9, OMIM=609177, no_hdr0=''),
  11131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11131, CurrentID=0, Status='live', Symbol='CAPN11', Aliases='calpain11', description='calpain 11', other_designations='CANP 11|calcium-activated neutral proteinase 11', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44158811, end_position_on_the_genomic_accession=44184402, orientation='plus', exon_count=25, OMIM=604822, no_hdr0=''),
  11132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11132, CurrentID=0, Status='live', Symbol='CAPN10', Aliases='CANP10, NIDDM1', description='calpain 10', other_designations='calcium-activated neutral proteinase 10|calpain-like protease CAPN10', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240586716, end_position_on_the_genomic_accession=240599109, orientation='plus', exon_count=12, OMIM=605286, no_hdr0=''),
  11133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11133, CurrentID=0, Status='live', Symbol='KPTN', Aliases='2E4, MRT41', description='kaptin (actin binding protein)', other_designations='actin-associated protein 2E4', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47475141, end_position_on_the_genomic_accession=47486791, orientation='minus', exon_count=17, OMIM=615620, no_hdr0=''),
  11135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11135, CurrentID=0, Status='live', Symbol='CDC42EP1', Aliases='BORG5, CEP1, MSE55', description='CDC42 effector protein (Rho GTPase binding) 1', other_designations='55 kDa bone marrow stromal/endothelial cell protein|binder of Rho GTPases 5|serum constituent protein|serum protein MSE55', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37559778, end_position_on_the_genomic_accession=37569405, orientation='plus', exon_count=6, OMIM=606084, no_hdr0=''),
  11136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11136, CurrentID=0, Status='live', Symbol='SLC7A9', Aliases='BAT1, CSNU3', description='solute carrier family 7 (amino acid transporter light chain, bo,+ system), member 9', other_designations='b(0,+)AT|glycoprotein-associated amino acid transporter b0,+AT1|solute carrier family 7 (cationic amino acid transporter, y+ system), member 9|solute carrier family 7 (glycoprotein-associated amino acid transporter light chain, bo,+ system), member 9', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32830511, end_position_on_the_genomic_accession=32870167, orientation='minus', exon_count=13, OMIM=604144, no_hdr0=''),
  11137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11137, CurrentID=0, Status='live', Symbol='PWP1', Aliases='IEF-SSP-9502', description='PWP1 homolog, endonuclein', other_designations='endonuclein|keratinocyte protein IEF SSP 9502|nuclear phosphoprotein similar to S. cerevisiae PWP1', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107685813, end_position_on_the_genomic_accession=107712480, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11138, CurrentID=0, Status='live', Symbol='TBC1D8', Aliases='AD3, HBLP1A, VRP, TBC1D8', description='TBC1 domain family member 8', other_designations='AD 3|BUB2-like protein 1|TBC1 domain family, member 8 (with GRAM domain)|TBC1 domain family, member 8A|vascular Rab-GAP/TBC-containing protein', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101007228, end_position_on_the_genomic_accession=101195326, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  11140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11140, CurrentID=0, Status='live', Symbol='CDC37', Aliases='P50CDC37', description='cell division cycle 37', other_designations='CDC37 (cell division cycle 37, S. cerevisiae, homolog)|CDC37 cell division cycle 37 homolog|cell division cycle 37 homolog|hsp90 chaperone protein kinase-targeting subunit', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10391131, end_position_on_the_genomic_accession=10403595, orientation='minus', exon_count=8, OMIM=605065, no_hdr0=''),
  11141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11141, CurrentID=0, Status='live', Symbol='IL1RAPL1', Aliases='IL1R8, IL1RAPL, MRX10, MRX21, MRX34, OPHN4, TIGIRR-2', description='interleukin 1 receptor accessory protein-like 1', other_designations='IL-1-RAPL-1|IL-1RAPL-1|IL1RAPL-1|X-linked interleukin-1 receptor accessory protein-like 1|interleukin 1 receptor-8|mental retardation, X-linked 10|oligophrenin-4|three immunoglobulin domain-containing IL-1 receptor-related 2', map_location='Xp22.1-p21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=28587564, end_position_on_the_genomic_accession=29956350, orientation='plus', exon_count=13, OMIM=300206, no_hdr0=''),
  11142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11142, CurrentID=0, Status='live', Symbol='PKIG', Aliases='PKI-gamma', description='protein kinase (cAMP-dependent, catalytic) inhibitor gamma', other_designations='', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44531781, end_position_on_the_genomic_accession=44619037, orientation='plus', exon_count=7, OMIM=604932, no_hdr0=''),
  11143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11143, CurrentID=0, Status='live', Symbol='KAT7', Aliases='HBO1, HBOA, MYST2, ZC2HC7', description='K(lysine) acetyltransferase 7', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 2|MYST histone acetyltransferase 2|MYST-2|histone acetyltransferase MYST2|histone acetyltransferase binding to ORC1|lysine acetyltransferase 7', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49788619, end_position_on_the_genomic_accession=49829096, orientation='plus', exon_count=16, OMIM=609880, no_hdr0=''),
  11144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11144, CurrentID=0, Status='live', Symbol='DMC1', Aliases='DMC1H, LIM15, dJ199H16.1', description='DNA meiotic recombinase 1', other_designations='DMC1 dosage suppressor of mck1 homolog, meiosis-specific homologous recombination|disrupted meiotic cDNA1, yeast, homolog of', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38509154, end_position_on_the_genomic_accession=38570286, orientation='minus', exon_count=16, OMIM=602721, no_hdr0=''),
  11145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11145, CurrentID=0, Status='live', Symbol='PLA2G16', Aliases='AdPLA, H-REV107-1, HRASLS3, HREV107, HREV107-1, HREV107-3, HRSL3', description='phospholipase A2 group XVI', other_designations='Ca-independent phospholipase A1/2|H-rev 107 protein homolog|HRAS-like suppressor 1|adipose-specific PLA2|adipose-specific phospholipase A2|group XVI phospholipase A1/A2|group XVI phospholipase A2|phospholipase A2, group XVI|renal carcinoma antigen NY-REN-65', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63574462, end_position_on_the_genomic_accession=63616926, orientation='minus', exon_count=5, OMIM=613867, no_hdr0=''),
  11146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11146, CurrentID=0, Status='live', Symbol='GLMN', Aliases='FAP, FAP48, FAP68, FKBPAP, GLML, GVM, VMGLOM', description='glomulin, FKBP associated protein', other_designations='FK506-binding protein-associated protein|FKBP-associated protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92246398, end_position_on_the_genomic_accession=92441946, orientation='minus', exon_count=24, OMIM=601749, no_hdr0=''),
  11147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11147, CurrentID=0, Status='live', Symbol='HHLA3', Aliases='', description='HERV-H LTR-associating 3', other_designations='human endogenous retrovirus-H long terminal repeat-associating 3|human endogenous retrovirus-H long terminal repeat-associating protein 3', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70354797, end_position_on_the_genomic_accession=70368023, orientation='plus', exon_count=5, OMIM=604372, no_hdr0=''),
  11148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11148, CurrentID=0, Status='live', Symbol='HHLA2', Aliases='B7-H7, B7H7', description='HERV-H LTR-associating 2', other_designations='human endogenous retrovirus-H long terminal repeat-associating protein 2', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108296485, end_position_on_the_genomic_accession=108388971, orientation='plus', exon_count=16, OMIM=604371, no_hdr0=''),
  11149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11149, CurrentID=0, Status='live', Symbol='BVES', Aliases='HBVES, POP1, POPDC1', description='blood vessel epicardial substance', other_designations='popeye domain containing 1|popeye domain-containing protein 1|popeye protein 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105096822, end_position_on_the_genomic_accession=105137174, orientation='minus', exon_count=11, OMIM=604577, no_hdr0=''),
  11151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11151, CurrentID=0, Status='live', Symbol='CORO1A', Aliases='CLABP, CLIPINA, HCORO1, IMD8, TACO, p57', description='coronin, actin binding protein, 1A', other_designations='clipin-A|coronin-1|coronin-like protein A|coronin-like protein p57|tryptophan aspartate-containing coat protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30183410, end_position_on_the_genomic_accession=30189076, orientation='plus', exon_count=12, OMIM=605000, no_hdr0=''),
  11152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11152, CurrentID=0, Status='live', Symbol='WDR45', Aliases='JM5, NBIA4, NBIA5, WDRX1, WIPI-4, WIPI4', description='WD repeat domain 45', other_designations='WD repeat domain, X-linked 1|WD repeat-containing protein 45|WD45 repeat protein interacting with phosphoinositides 4|neurodegeneration with brain iron accumulation 4|neurodegeneration with brain iron accumulation 5', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49074433, end_position_on_the_genomic_accession=49101121, orientation='minus', exon_count=13, OMIM=300526, no_hdr0=''),
  11153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11153, CurrentID=0, Status='live', Symbol='FICD', Aliases='HIP13, HYPE, UNQ3041', description='FIC domain containing', other_designations='AMPylator FICD|FIC domain-containing protein|HIP-13|Huntingtin interacting protein E|fic S-phase protein cell division homolog|huntingtin interacting protein 13|huntingtin interactor protein E|huntingtin yeast partner E|huntingtin-interacting protein 13|huntingtin-interacting protein E', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108515211, end_position_on_the_genomic_accession=108521210, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11154, CurrentID=0, Status='live', Symbol='AP4S1', Aliases='AP47B, CLA20, CLAPS4, CPSQ6, SPG52', description='adaptor related protein complex 4, sigma 1 subunit', other_designations='AP-4 adapter complex subunit sigma-1|AP-4 adaptor complex subunit sigma-1|adaptor-related protein complex 4 subunit sigma-1|adaptor-related protein complex 4, sigma 1 subunit|clathrin-associated/assembly/adaptor protein, sigma 4|sigma-4-adaptin', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31025106, end_position_on_the_genomic_accession=31096450, orientation='plus', exon_count=11, OMIM=607243, no_hdr0=''),
  11155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11155, CurrentID=0, Status='live', Symbol='LDB3', Aliases='CMD1C, CMH24, CMPD3, CYPHERZ1, LDB3Z4, LVNC3, MFM4, ORACLE, PDLIM6, ZASP, LDB3', description='LIM domain binding 3', other_designations='PDZ and LIM domain 6|Z-band alternatively spliced PDZ motif protein|Z-band alternatively spliced PDZ-motif protein|cardiomyopathy, dilated 1C (autosomal dominant)|protein cypher', map_location='10q22.3-q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86666717, end_position_on_the_genomic_accession=86736072, orientation='plus', exon_count=20, OMIM=605906, no_hdr0=''),
  11156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11156, CurrentID=0, Status='live', Symbol='PTP4A3', Aliases='PRL-3, PRL-R, PRL3', description='protein tyrosine phosphatase type IVA, member 3', other_designations='potentially prenylated protein tyrosine phosphatase|protein-tyrosine phosphatase 4a3|protein-tyrosine phosphatase of regenerating liver 3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141391993, end_position_on_the_genomic_accession=141432454, orientation='plus', exon_count=12, OMIM=606449, no_hdr0=''),
  11157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11157, CurrentID=0, Status='live', Symbol='LSM6', Aliases='YDR378C', description='LSM6 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM6 U6 small nuclear RNA and mRNA degradation associated|LSM6 homolog, U6 small nuclear RNA associated|Sm protein F', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146175683, end_position_on_the_genomic_accession=146190061, orientation='plus', exon_count=4, OMIM=607286, no_hdr0=''),
  11158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11158, CurrentID=0, Status='live', Symbol='RABL2B', Aliases='', description='RAB, member of RAS oncogene family-like 2B', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50767492, end_position_on_the_genomic_accession=50783675, orientation='minus', exon_count=11, OMIM=605413, no_hdr0=''),
  11159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11159, CurrentID=0, Status='live', Symbol='RABL2A', Aliases='', description='RAB, member of RAS oncogene family-like 2A', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113627237, end_position_on_the_genomic_accession=113643398, orientation='plus', exon_count=11, OMIM=605412, no_hdr0=''),
  11160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11160, CurrentID=0, Status='live', Symbol='ERLIN2', Aliases='C8orf2, Erlin-2, NET32, SPFH2, SPG18', description='ER lipid raft associated 2', other_designations='SPFH domain family, member 2|endoplasmic reticulum lipid raft-associated protein 2|stomatin-prohibitin-flotillin-HflC/K domain-containing protein 2', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37736579, end_position_on_the_genomic_accession=37757801, orientation='plus', exon_count=14, OMIM=611605, no_hdr0=''),
  11161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11161, CurrentID=0, Status='live', Symbol='C14orf1', Aliases='ERG28, NET51', description='chromosome 14 open reading frame 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75650890, end_position_on_the_genomic_accession=75661195, orientation='minus', exon_count=5, OMIM=604576, no_hdr0=''),
  11162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11162, CurrentID=0, Status='live', Symbol='NUDT6', Aliases='ASFGF2, FGF-AS, FGF2AS, GFG-1, GFG1', description='nudix hydrolase 6', other_designations='antisense basic fibroblast growth factor|nudix (nucleoside diphosphate linked moiety X)-type motif 6', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122892644, end_position_on_the_genomic_accession=122923004, orientation='minus', exon_count=6, OMIM=606261, no_hdr0=''),
  11163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11163, CurrentID=0, Status='live', Symbol='NUDT4', Aliases='DIPP2, DIPP2alpha, DIPP2beta, HDCMB47P', description='nudix hydrolase 4', other_designations="DIPP-2|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 2|diphosphoinositol polyphosphate phosphohydrolase type 2|nudix (nucleoside diphosphate linked moiety X)-type motif 4", map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93377923, end_position_on_the_genomic_accession=93403248, orientation='plus', exon_count=8, OMIM=609229, no_hdr0=''),
  11164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11164, CurrentID=0, Status='live', Symbol='NUDT5', Aliases='YSA1, YSA1H, YSAH1, hYSAH1', description='nudix hydrolase 5', other_designations='8-oxo-dGDP phosphatase|nudix (nucleoside diphosphate linked moiety X)-type motif 5', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12165330, end_position_on_the_genomic_accession=12196144, orientation='minus', exon_count=11, OMIM=609230, no_hdr0=''),
  11165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11165, CurrentID=0, Status='live', Symbol='NUDT3', Aliases='DIPP, DIPP-1, DIPP1', description='nudix hydrolase 3', other_designations="diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 1|nucleoside diphosphate-linked moiety X motif 3|nudix (nucleoside diphosphate linked moiety X)-type motif 3|nudix motif 3", map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34287194, end_position_on_the_genomic_accession=34392680, orientation='minus', exon_count=5, OMIM=609228, no_hdr0=''),
  11166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11166, CurrentID=0, Status='live', Symbol='SOX21', Aliases='SOX25', description='SRY-box 21', other_designations='SOX-A|SRY (sex determining region Y)-box 21|SRY box 21', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94709622, end_position_on_the_genomic_accession=94712543, orientation='minus', exon_count=1, OMIM=604974, no_hdr0=''),
  11167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11167, CurrentID=0, Status='live', Symbol='FSTL1', Aliases='FRP, FSL1, MIR198', description='follistatin like 1', other_designations='follistatin-like 1|follistatin-like protein 1', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120394214, end_position_on_the_genomic_accession=120451071, orientation='minus', exon_count=11, OMIM=605547, no_hdr0=''),
  11168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11168, CurrentID=0, Status='live', Symbol='PSIP1', Aliases='DFS70, LEDGF, PAIP, PSIP2, p52, p75', description='PC4 and SFRS1 interacting protein 1', other_designations='CLL-associated antigen KW-7|dense fine speckles 70 kDa protein|lens epithelium-derived growth factor|transcriptional coactivator p52/p75', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15464066, end_position_on_the_genomic_accession=15511005, orientation='minus', exon_count=18, OMIM=603620, no_hdr0=''),
  11169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11169, CurrentID=0, Status='live', Symbol='WDHD1', Aliases='AND-1, AND1, CHTF4, CTF4', description='WD repeat and HMG-box DNA binding protein 1', other_designations='Acidic nucleoplasmic DNA-binding protein 1|CTF4, chromosome transmission fidelity factor 4 homolog', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54938938, end_position_on_the_genomic_accession=55027101, orientation='minus', exon_count=26, OMIM=608126, no_hdr0=''),
  11170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11170, CurrentID=0, Status='live', Symbol='FAM107A', Aliases='DRR1, TU3A', description='family with sequence similarity 107 member A', other_designations='down-regulated in renal cell carcinoma 1|family with sequence similarity 107, member A', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58564112, end_position_on_the_genomic_accession=58627610, orientation='minus', exon_count=8, OMIM=608295, no_hdr0=''),
  11171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11171, CurrentID=0, Status='live', Symbol='STRAP', Aliases='MAWD, PT-WD, UNRIP', description='serine/threonine kinase receptor associated protein', other_designations='MAP activator with WD repeats|WD-40 repeat protein PT-WD|unr-interacting protein', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15882354, end_position_on_the_genomic_accession=15903478, orientation='plus', exon_count=10, OMIM=605986, no_hdr0=''),
  11172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11172, CurrentID=0, Status='live', Symbol='INSL6', Aliases='RIF1', description='insulin like 6', other_designations='insulin-like 6|insulin-like peptide 5|insulin-like peptide 6|relaxin/insulin-like factor 1', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4991949, end_position_on_the_genomic_accession=5185710, orientation='minus', exon_count=4, OMIM=606414, no_hdr0=''),
  11173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11173, CurrentID=0, Status='live', Symbol='ADAMTS7', Aliases='ADAM-TS 7, ADAM-TS7, ADAMTS-7', description='ADAM metallopeptidase with thrombospondin type 1 motif 7', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 7|COMPase|a disintegrin and metalloprotease with thrombospondin motifs-7 preproprotein|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 7', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78759203, end_position_on_the_genomic_accession=78811464, orientation='minus', exon_count=25, OMIM=605009, no_hdr0=''),
  11174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11174, CurrentID=0, Status='live', Symbol='ADAMTS6', Aliases='ADAM-TS 6, ADAM-TS6, ADAMTS-6', description='ADAM metallopeptidase with thrombospondin type 1 motif 6', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 6|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 6', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65148736, end_position_on_the_genomic_accession=65482027, orientation='minus', exon_count=34, OMIM=605008, no_hdr0=''),
  11176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11176, CurrentID=0, Status='live', Symbol='BAZ2A', Aliases='TIP5, WALp3', description='bromodomain adjacent to zinc finger domain 2A', other_designations='TTF-I interacting peptide 5|TTF-I-interacting protein 5|bromodomain adjacent to zinc finger domain, 2A|transcription termination factor I-interacting protein 5', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56595596, end_position_on_the_genomic_accession=56636816, orientation='minus', exon_count=35, OMIM=605682, no_hdr0=''),
  11177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11177, CurrentID=0, Status='live', Symbol='BAZ1A', Aliases='ACF1, WALp1, WCRF180, hACF1', description='bromodomain adjacent to zinc finger domain 1A', other_designations='ATP-dependent chromatin remodeling protein|ATP-dependent chromatin-remodeling protein|ATP-utilizing chromatin assembly and remodeling factor 1|CHRAC subunit ACF1|bromodomain adjacent to zinc finger domain, 1A|hWALp1|williams syndrome transcription factor-related chromatin-remodeling factor 180', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34752731, end_position_on_the_genomic_accession=34875647, orientation='minus', exon_count=29, OMIM=605680, no_hdr0=''),
  11178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11178, CurrentID=0, Status='live', Symbol='LZTS1', Aliases='F37, FEZ1', description='leucine zipper, putative tumor suppressor 1', other_designations='F37/Esophageal cancer-related gene-coding leucine-zipper motif', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20246165, end_position_on_the_genomic_accession=20303981, orientation='minus', exon_count=9, OMIM=606551, no_hdr0=''),
  11179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11179, CurrentID=0, Status='live', Symbol='ZNF277', Aliases='NRIF4P, ZNF277', description='zinc finger protein 277', other_designations='nuclear receptor-interacting factor 4|zinc finger protein (C2H2 type) 277|zinc finger protein 277 pseudogene', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112206588, end_position_on_the_genomic_accession=112343934, orientation='plus', exon_count=12, OMIM=605465, no_hdr0=''),
  11180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11180, CurrentID=0, Status='live', Symbol='WDR6', Aliases='', description='WD repeat domain 6', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49007204, end_position_on_the_genomic_accession=49015953, orientation='plus', exon_count=7, OMIM=606031, no_hdr0=''),
  11181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11181, CurrentID=0, Status='live', Symbol='TREH', Aliases='TRE, TREA', description='trehalase', other_designations='alpha,alpha-trehalose glucohydrolase|trehalase (brush-border membrane glycoprotein)', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118658232, end_position_on_the_genomic_accession=118679690, orientation='minus', exon_count=15, OMIM=275360, no_hdr0=''),
  11182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11182, CurrentID=0, Status='live', Symbol='SLC2A6', Aliases='GLUT6, GLUT9, HSA011372', description='solute carrier family 2 (facilitated glucose transporter), member 6', other_designations='GLUT-6|GLUT-9|glucose transporter type 6|glucose transporter type 9', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133471094, end_position_on_the_genomic_accession=133479154, orientation='minus', exon_count=10, OMIM=606813, no_hdr0=''),
  11183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11183, CurrentID=0, Status='live', Symbol='MAP4K5', Aliases='GCKR, KHS, KHS1, MAPKKKK5', description='mitogen-activated protein kinase kinase kinase kinase 5', other_designations='MAPK/ERK kinase kinase kinase 5|MEK kinase kinase 5|MEKKK 5|germinal center kinase-related|kinase homologous to SPS1/STE20', map_location='14q11.2-q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50418480, end_position_on_the_genomic_accession=50561126, orientation='minus', exon_count=37, OMIM=604923, no_hdr0=''),
  11184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11184, CurrentID=0, Status='live', Symbol='MAP4K1', Aliases='HPK1', description='mitogen-activated protein kinase kinase kinase kinase 1', other_designations='MAPK/ERK kinase kinase kinase 1|MEK kinase kinase 1|MEKKK 1|hematopoietic progenitor kinase 1', map_location='19q13.1-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38587640, end_position_on_the_genomic_accession=38618035, orientation='minus', exon_count=33, OMIM=601983, no_hdr0=''),
  11185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11185, CurrentID=0, Status='live', Symbol='INMT', Aliases='TEMT', description='indolethylamine N-methyltransferase', other_designations='amine N-methyltransferase|aromatic alkylamine N-methyltransferase|arylamine N-methyltransferase|indolamine N-methyltransferase|nicotine N-methyltransferase|thioether S-methyltransferase', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30752135, end_position_on_the_genomic_accession=30757602, orientation='plus', exon_count=3, OMIM=604854, no_hdr0=''),
  11186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11186, CurrentID=0, Status='live', Symbol='RASSF1', Aliases='123F2, NORE2AA, RDA32, REH3P21, RASSF1', description='Ras association (RalGDS/AF-6) domain family member 1', other_designations='WUGSC:H_LUCA12.5|cardiac-specific ras association domain family 1 protein|pancreas-specific ras association domain family 1 protein|tumor suppressor protein RDA32', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50329786, end_position_on_the_genomic_accession=50340936, orientation='minus', exon_count=10, OMIM=605082, no_hdr0=''),
  11187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11187, CurrentID=0, Status='live', Symbol='PKP3', Aliases='', description='plakophilin 3', other_designations='plakophilin 3b', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=392599, end_position_on_the_genomic_accession=404908, orientation='plus', exon_count=14, OMIM=605561, no_hdr0=''),
  11188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11188, CurrentID=0, Status='live', Symbol='NISCH', Aliases='I-1, IR1, IRAS, hIRAS', description='nischarin', other_designations='I-1 receptor candidate protein|I1R candidate protein|imidazoline receptor 1|imidazoline receptor antisera selected', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52455508, end_position_on_the_genomic_accession=52493072, orientation='plus', exon_count=23, OMIM=615507, no_hdr0=''),
  11189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11189, CurrentID=0, Status='live', Symbol='CELF3', Aliases='BRUNOL1, CAGH4, ERDA4, ETR-1, TNRC4', description='CUGBP, Elav-like family member 3', other_designations='CAG repeat domain|CAG repeat protein 4|CUG-BP and ETR-3 like factor 3|CUG-BP- and ETR-3-like factor 3|ELAV-type RNA-binding protein 1|RNA-binding protein BRUNOL-1|bruno-like protein 1|expanded repeat domain protein CAG/CTG 4|expanded repeat domain, CAG/CTG 4|trinucleotide repeat containing 4|trinucleotide repeat-containing gene 4 protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151700052, end_position_on_the_genomic_accession=151716821, orientation='minus', exon_count=16, OMIM=612678, no_hdr0=''),
  11190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11190, CurrentID=0, Status='live', Symbol='CEP250', Aliases='C-NAP1, CEP2, CNAP1', description='centrosomal protein 250kDa', other_designations='250 kDa centrosomal protein|centrosomal Nek2-associated protein 1|centrosome associated protein', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35455139, end_position_on_the_genomic_accession=35517531, orientation='plus', exon_count=38, OMIM=609689, no_hdr0=''),
  11191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11191, CurrentID=0, Status='live', Symbol='PTENP1', Aliases='PTEN-rs, PTEN2, PTENpg1, PTH2, psiPTEN', description='phosphatase and tensin homolog pseudogene 1 (functional)', other_designations='PTEN/MMAC1 pseudogene|phosphatase and tensin homolog (mutated in multiple advanced cancers 1), pseudogene 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33673504, end_position_on_the_genomic_accession=33677420, orientation='minus', exon_count=1, OMIM=613531, no_hdr0=''),
  11192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11192, CurrentID=0, Status='live', Symbol='DYX3', Aliases='', description='dyslexia susceptibility 3', other_designations='dyslexia 3', map_location='2p16-p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604254, no_hdr0=''),
  11193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11193, CurrentID=0, Status='live', Symbol='WBP4', Aliases='FBP21', description='WW domain binding protein 4', other_designations='WBP-4|WW domain-containing binding protein 4|WW domain-containing-binding protein 4|domain-containing binding protein 4|formin binding protein 21|formin-binding protein 21', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41061561, end_position_on_the_genomic_accession=41084006, orientation='plus', exon_count=11, OMIM=604981, no_hdr0=''),
  11194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11194, CurrentID=0, Status='live', Symbol='ABCB8', Aliases='EST328128, M-ABC1, MABC1', description='ATP binding cassette subfamily B member 8', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 8|mitochondrial ABC protein|mitochondrial ATP-binding cassette 1', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151028422, end_position_on_the_genomic_accession=151047782, orientation='plus', exon_count=17, OMIM=605464, no_hdr0=''),
  11196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11196, CurrentID=0, Status='live', Symbol='SEC23IP', Aliases='MSTP053, P125, P125A', description='SEC23 interacting protein', other_designations='', map_location='10q25-q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119892573, end_position_on_the_genomic_accession=119944658, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  11197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11197, CurrentID=0, Status='live', Symbol='WIF1', Aliases='WIF-1', description='WNT inhibitory factor 1', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65050624, end_position_on_the_genomic_accession=65121566, orientation='minus', exon_count=10, OMIM=605186, no_hdr0=''),
  11198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11198, CurrentID=0, Status='live', Symbol='SUPT16H', Aliases='CDC68, FACTP140, SPT16, SPT16/CDC68', description='SPT16 homolog, facilitates chromatin remodeling subunit', other_designations='FACT 140 kDa subunit|chromatin-specific transcription elongation factor 140 kDa subunit|facilitates chromatin remodeling 140 kDa subunit|facilitates chromatin transcription complex subunit SPT16|hSPT16|suppressor of Ty 16 homolog', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21351472, end_position_on_the_genomic_accession=21384639, orientation='minus', exon_count=27, OMIM=605012, no_hdr0=''),
  11199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11199, CurrentID=0, Status='live', Symbol='ANXA10', Aliases='ANX14', description='annexin A10', other_designations='annexin 14|annexin-10|annexin-14', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168092526, end_position_on_the_genomic_accession=168187742, orientation='plus', exon_count=13, OMIM=608008, no_hdr0=''),
  11200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11200, CurrentID=0, Status='live', Symbol='CHEK2', Aliases='CDS1, CHK2, HuCds1, LFS2, PP1425, RAD53, hCds1', description='checkpoint kinase 2', other_designations='CHK2 checkpoint homolog|cds1 homolog|checkpoint-like protein CHK2', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28687743, end_position_on_the_genomic_accession=28741905, orientation='minus', exon_count=22, OMIM=604373, no_hdr0=''),
  11201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11201, CurrentID=0, Status='live', Symbol='POLI', Aliases='RAD30B, RAD3OB', description='polymerase (DNA directed) iota', other_designations='RAD30 homolog B|eta2|polymerase (DNA-directed), iota', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54269479, end_position_on_the_genomic_accession=54321266, orientation='plus', exon_count=12, OMIM=605252, no_hdr0=''),
  11202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11202, CurrentID=0, Status='live', Symbol='KLK8', Aliases='HNP, NP, NRPN, PRSS19, TADG14', description='kallikrein related peptidase 8', other_designations='kallikrein-related peptidase 8|ovasin|serine protease 19|serine protease TADG-14|tumor-associated differentially expressed gene 14 protein', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50996008, end_position_on_the_genomic_accession=51001702, orientation='minus', exon_count=6, OMIM=605644, no_hdr0=''),
  285857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285857, CurrentID=0, Status='live', Symbol='LOC285857', Aliases='', description='uncharacterized LOC285857', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43845750, end_position_on_the_genomic_accession=43852341, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11209, CurrentID=0, Status='live', Symbol='MST1P2', Aliases='MSPL-1, MSPL-2, MSPL-3, MSPL1, MSPL2, MSPL3, MSTP1, MSTP2, MSTP3', description='macrophage stimulating 1 (hepatocyte growth factor-like) pseudogene 2', other_designations='macrophage stimulating, pseudogene 1|macrophage stimulating, pseudogene 2|macrophage stimulating, pseudogene 3', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16645574, end_position_on_the_genomic_accession=16650420, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  11211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11211, CurrentID=0, Status='live', Symbol='FZD10', Aliases='CD350, FZ-10, Fz10, FzE7, hFz10', description='frizzled class receptor 10', other_designations='frizzled 10, seven transmembrane spanning receptor|frizzled family receptor 10|frizzled homolog 10', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130162459, end_position_on_the_genomic_accession=130165740, orientation='plus', exon_count=1, OMIM=606147, no_hdr0=''),
  11212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11212, CurrentID=0, Status='live', Symbol='PROSC', Aliases='', description='proline synthetase co-transcribed homolog (bacterial)', other_designations='proline synthetase co-transcribed (bacterial homolog)|proline synthetase co-transcribed bacterial homolog protein', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37762583, end_position_on_the_genomic_accession=37779768, orientation='plus', exon_count=8, OMIM=604436, no_hdr0=''),
  11213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11213, CurrentID=0, Status='live', Symbol='IRAK3', Aliases='ASRT5, IRAKM', description='interleukin 1 receptor associated kinase 3', other_designations='IL-1 receptor-associated kinase M', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66189198, end_position_on_the_genomic_accession=66254622, orientation='plus', exon_count=12, OMIM=604459, no_hdr0=''),
  11214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11214, CurrentID=0, Status='live', Symbol='AKAP13', Aliases='AKAP-13, AKAP-Lbc, ARHGEF13, BRX, HA-3, Ht31, LBC, PRKA13, PROTO-LB, PROTO-LBC, c-lbc, p47', description='A-kinase anchoring protein 13', other_designations='A kinase (PRKA) anchor protein 13|A-kinase anchoring protein|LBC oncogene|breast cancer nuclear receptor-binding auxiliary protein|guanine nucleotide exchange factor Lbc|human thyroid-anchoring protein 31|lymphoid blast crisis oncogene|non-oncogenic Rho GTPase-specific GTP exchange factor|protein kinase A-anchoring protein 13', map_location='15q24-q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85380616, end_position_on_the_genomic_accession=85749358, orientation='plus', exon_count=39, OMIM=604686, no_hdr0=''),
  11215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11215, CurrentID=0, Status='live', Symbol='AKAP11', Aliases='AKAP-11, AKAP220, PPP1R44, PRKA11', description='A-kinase anchoring protein 11', other_designations='A kinase (PRKA) anchor protein 11|A-kinase anchor protein 220 kDa|A-kinase anchoring protein, 220kDa|a kinase anchor protein 220 kDa|protein kinase A anchoring protein 11|protein kinase A-anchoring protein 11|protein phosphatase 1, regulatory subunit 44', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42271481, end_position_on_the_genomic_accession=42323267, orientation='plus', exon_count=15, OMIM=604696, no_hdr0=''),
  11216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11216, CurrentID=0, Status='live', Symbol='AKAP10', Aliases='AKAP-10, D-AKAP-2, D-AKAP2, PRKA10', description='A-kinase anchoring protein 10', other_designations='A kinase (PRKA) anchor protein 10|A kinase anchor protein 10|dual specificity A kinase-anchoring protein 2|mitochondrial A kinase PPKA anchor protein 10|protein kinase A anchoring protein 10', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19904452, end_position_on_the_genomic_accession=19977856, orientation='minus', exon_count=15, OMIM=604694, no_hdr0=''),
  11217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11217, CurrentID=0, Status='live', Symbol='AKAP2', Aliases='AKAP-2, AKAPKL, PRKA2', description='A-kinase anchoring protein 2', other_designations='A kinase (PRKA) anchor protein 2|protein kinase A anchoring protein 2|protein kinase A2', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110048598, end_position_on_the_genomic_accession=110172512, orientation='plus', exon_count=6, OMIM=604582, no_hdr0=''),
  11218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11218, CurrentID=0, Status='live', Symbol='DDX20', Aliases='DP103, GEMIN3', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 20', other_designations='DEAD box protein 20|DEAD box protein DP 103|DEAD-box protein DP103|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 20, 103kD|SMN-interacting protein|component of gems 3|gemin-3', map_location='1p21.1-p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111755568, end_position_on_the_genomic_accession=111767577, orientation='plus', exon_count=11, OMIM=606168, no_hdr0=''),
  11219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11219, CurrentID=0, Status='live', Symbol='TREX2', Aliases='', description='three prime repair exonuclease 2', other_designations="3'-5' exonuclease TREX2 long form", map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153444720, end_position_on_the_genomic_accession=153446487, orientation='minus', exon_count=2, OMIM=300370, no_hdr0=''),
  11221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11221, CurrentID=0, Status='live', Symbol='DUSP10', Aliases='MKP-5, MKP5', description='dual specificity phosphatase 10', other_designations='dual specificity phosphatase MKP-5|map kinase phosphatase 5|mitogen-activated protein kinase phosphatase 5|serine/threonine specific protein phosphatase', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221701420, end_position_on_the_genomic_accession=221742176, orientation='minus', exon_count=7, OMIM=608867, no_hdr0=''),
  11222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11222, CurrentID=0, Status='live', Symbol='MRPL3', Aliases='COXPD9, MRL3, RPML3', description='mitochondrial ribosomal protein L3', other_designations='L3mt|MRP-L3|mitochondrial 60S ribosomal protein L3', map_location='3q21-q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131462201, end_position_on_the_genomic_accession=131503016, orientation='minus', exon_count=10, OMIM=607118, no_hdr0=''),
  11223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11223, CurrentID=0, Status='live', Symbol='MST1L', Aliases='BRF-1, D1F15S1A, MSPL-7, MSPL7, MST1P9, MSTP7, MSTP9', description='macrophage stimulating 1-like', other_designations='brain rescue factor 1|brain-rescue-factor-1|macrophage stimulating 1 (hepatocyte growth factor-like) pseudogene 9|macrophage stimulating, pseudogene 7|macrophage stimulating, pseudogene 9|putative macrophage-stimulating protein MSTP9', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16754634, end_position_on_the_genomic_accession=16764480, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11224, CurrentID=0, Status='live', Symbol='RPL35', Aliases='L35', description='ribosomal protein L35', other_designations='', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124857879, end_position_on_the_genomic_accession=124861961, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11226, CurrentID=0, Status='live', Symbol='GALNT6', Aliases='GALNAC-T6, GalNAcT6', description='polypeptide N-acetylgalactosaminyltransferase 6', other_designations='GalNAc transferase 6|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 6|UDP-N-acetyl-alpha-D-galactosamine: polypeptide N-acetylgalactosaminyltransferase 6|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 6 (GalNAc-T6)|polypeptide GalNAc transferase 6|pp-GaNTase 6|protein-UDP acetylgalactosaminyltransferase 6', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51351870, end_position_on_the_genomic_accession=51391955, orientation='minus', exon_count=16, OMIM=605148, no_hdr0=''),
  11227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11227, CurrentID=0, Status='live', Symbol='GALNT5', Aliases='GALNAC-T5, GALNACT5', description='polypeptide N-acetylgalactosaminyltransferase 5', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 5|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 5 (GalNAc-T5)|polypeptide GalNAc transferase 5|pp-GaNTase 5|protein-UDP acetylgalactosaminyltransferase 5', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157257828, end_position_on_the_genomic_accession=157311401, orientation='plus', exon_count=10, OMIM=615129, no_hdr0=''),
  11228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11228, CurrentID=0, Status='live', Symbol='RASSF8', Aliases='C12orf2, HOJ1', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 8', other_designations='carcinoma-associated protein HOJ-1', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25959031, end_position_on_the_genomic_accession=26079892, orientation='plus', exon_count=9, OMIM=608231, no_hdr0=''),
  11230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11230, CurrentID=0, Status='live', Symbol='PRAF2', Aliases='JM4', description='PRA1 domain family member 2', other_designations='Jena-Muenchen 4|PRA1 domain family, member 2', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49071156, end_position_on_the_genomic_accession=49074045, orientation='minus', exon_count=3, OMIM=300840, no_hdr0=''),
  11231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11231, CurrentID=0, Status='live', Symbol='SEC63', Aliases='DNAJC23, ERdj2, PRO2507L, SEC63', description='SEC63 homolog, protein translocation regulator', other_designations='SEC63 homolog|SEC63 protein translocation regulator|SEC63-like protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107867756, end_position_on_the_genomic_accession=107958278, orientation='minus', exon_count=22, OMIM=608648, no_hdr0=''),
  11232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11232, CurrentID=0, Status='live', Symbol='POLG2', Aliases='HP55, MTPOLB, PEOA4, POLB, POLG-BETA, POLGB', description='polymerase (DNA directed), gamma 2, accessory subunit', other_designations='DNA polymerase gamma accessory 55 kDa subunit|mitochondrial DNA polymerase accessory subunit|mitochondrial DNA polymerase subunit gamma-2|mitochondrial DNA polymerase, accessory subunit|p55', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64476994, end_position_on_the_genomic_accession=64497066, orientation='minus', exon_count=11, OMIM=604983, no_hdr0=''),
  11234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11234, CurrentID=0, Status='live', Symbol='HPS5', Aliases='AIBP63', description='Hermansky-Pudlak syndrome 5', other_designations='alpha integrin binding protein 63|alpha-integrin-binding protein 63|ruby-eye protein 2 homolog', map_location='11p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18278670, end_position_on_the_genomic_accession=18322498, orientation='minus', exon_count=24, OMIM=607521, no_hdr0=''),
  11235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11235, CurrentID=0, Status='live', Symbol='PDCD10', Aliases='CCM3, TFAR15', description='programmed cell death 10', other_designations='TF-1 cell apoptosis-related protein 15|apoptosis-related protein 15|cerebral cavernous malformations 3 protein', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167683735, end_position_on_the_genomic_accession=167735422, orientation='minus', exon_count=12, OMIM=609118, no_hdr0=''),
  11236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11236, CurrentID=0, Status='live', Symbol='RNF139', Aliases='HRCA1, RCA1, TRC8', description='ring finger protein 139', other_designations='multiple membrane spanning receptor TRC8|patched related protein translocated in renal cancer|translocation in renal carcinoma on chromosome 8 protein', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124474767, end_position_on_the_genomic_accession=124488618, orientation='plus', exon_count=2, OMIM=603046, no_hdr0=''),
  11237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11237, CurrentID=0, Status='live', Symbol='RNF24', Aliases='G1L', description='ring finger protein 24', other_designations='goliath-like protein (C3HC4 type)', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3929918, end_position_on_the_genomic_accession=4015569, orientation='minus', exon_count=10, OMIM=612489, no_hdr0=''),
  11238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11238, CurrentID=0, Status='live', Symbol='CA5B', Aliases='CA-VB', description='carbonic anhydrase VB, mitochondrial', other_designations='carbonate dehydratase VB|carbonic dehydratase', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15738250, end_position_on_the_genomic_accession=15787625, orientation='plus', exon_count=9, OMIM=300230, no_hdr0=''),
  11240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11240, CurrentID=0, Status='live', Symbol='PADI2', Aliases='PAD-H19, PAD2, PDI2', description='peptidyl arginine deiminase, type II', other_designations='protein-arginine deiminase type II', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17066761, end_position_on_the_genomic_accession=17119453, orientation='minus', exon_count=16, OMIM=607935, no_hdr0=''),
  11243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11243, CurrentID=0, Status='live', Symbol='PMF1', Aliases='', description='polyamine-modulated factor 1', other_designations='Est1p-like protein B (EST1B)', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156212988, end_position_on_the_genomic_accession=156240077, orientation='plus', exon_count=6, OMIM=609176, no_hdr0=''),
  11244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11244, CurrentID=0, Status='live', Symbol='ZHX1', Aliases='', description='zinc fingers and homeoboxes 1', other_designations='zinc finger and homeodomain protein 1|zinc fingers and homeobox 1', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123248450, end_position_on_the_genomic_accession=123275541, orientation='minus', exon_count=5, OMIM=604764, no_hdr0=''),
  11245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11245, CurrentID=0, Status='live', Symbol='GPR176', Aliases='HB-954', description='G protein-coupled receptor 176', other_designations='', map_location='15q14-q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39799022, end_position_on_the_genomic_accession=39920892, orientation='minus', exon_count=10, OMIM=612183, no_hdr0=''),
  11247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11247, CurrentID=0, Status='live', Symbol='NXPH4', Aliases='NPH4', description='neurexophilin 4', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57216795, end_position_on_the_genomic_accession=57226449, orientation='plus', exon_count=2, OMIM=604637, no_hdr0=''),
  11248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11248, CurrentID=0, Status='live', Symbol='NXPH3', Aliases='NPH3', description='neurexophilin 3', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49575936, end_position_on_the_genomic_accession=49583810, orientation='plus', exon_count=2, OMIM=604636, no_hdr0=''),
  11249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11249, CurrentID=0, Status='live', Symbol='NXPH2', Aliases='NPH2', description='neurexophilin 2', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138669157, end_position_on_the_genomic_accession=138780241, orientation='minus', exon_count=2, OMIM=604635, no_hdr0=''),
  11250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11250, CurrentID=0, Status='live', Symbol='GPR45', Aliases='PSP24, PSP24(ALPHA), PSP24A', description='G protein-coupled receptor 45', other_designations='PSP24-1|PSP24-alpha|high-affinity lysophosphatidic acid receptor', map_location='2q11.1-q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105241743, end_position_on_the_genomic_accession=105243467, orientation='plus', exon_count=1, OMIM=604838, no_hdr0=''),
  11251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11251, CurrentID=0, Status='live', Symbol='PTGDR2', Aliases='CD294, CRTH2, DL1R, DP2, GPR44', description='prostaglandin D2 receptor 2', other_designations='G protein-coupled receptor 44|G-protein coupled receptor 44|chemoattractant receptor homologous molecule expressed on T helper type 2 cells|chemoattractant receptor-homologous molecule expressed on TH2 cells|putative G-protein coupled receptor 44', map_location='11q12-q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60850925, end_position_on_the_genomic_accession=60855971, orientation='minus', exon_count=2, OMIM=604837, no_hdr0=''),
  11252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11252, CurrentID=0, Status='live', Symbol='PACSIN2', Aliases='SDPII', description='protein kinase C and casein kinase substrate in neurons 2', other_designations='cytoplasmic phosphoprotein PACSIN2|syndapin 2|syndapin II|syndapin-2|syndapin-II|syndapin2', map_location='22q13.2-q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42869766, end_position_on_the_genomic_accession=43015690, orientation='minus', exon_count=16, OMIM=604960, no_hdr0=''),
  11253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11253, CurrentID=0, Status='live', Symbol='MAN1B1', Aliases='ERMAN1, MANA-ER, MRT15', description='mannosidase, alpha, class 1B, member 1', other_designations='ER alpha 1,2-mannosidase|Man9GlcNAc2-specific processing alpha-mannosidase|endoplasmic reticulum mannosyl-oligosaccharide 1,2-alpha-mannosidase 1', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137086862, end_position_on_the_genomic_accession=137109187, orientation='plus', exon_count=14, OMIM=604346, no_hdr0=''),
  11254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11254, CurrentID=0, Status='live', Symbol='SLC6A14', Aliases='BMIQ11', description='solute carrier family 6 (amino acid transporter), member 14', other_designations='amino acid transporter ATB0+|amino acid transporter B0+|solute carrier family 6 (neurotransmitter transporter), member 14', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116436579, end_position_on_the_genomic_accession=116461458, orientation='plus', exon_count=14, OMIM=300444, no_hdr0=''),
  11255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11255, CurrentID=0, Status='live', Symbol='HRH3', Aliases='GPCR97, HH3R', description='histamine receptor H3', other_designations='G protein-coupled receptor 97|G-protein coupled receptor 97|H3R', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62214766, end_position_on_the_genomic_accession=62220278, orientation='minus', exon_count=4, OMIM=604525, no_hdr0=''),
  11257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11257, CurrentID=0, Status='live', Symbol='TP53TG1', Aliases='LINC00096, NCRNA00096, P53TG1, P53TG1-D, TP53AP1', description='TP53 target 1 (non-protein coding)', other_designations='H_RG012D21.9|TP53 activated protein 1|TP53 target gene 1|long intergenic non-protein coding RNA 96', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87325347, end_position_on_the_genomic_accession=87345492, orientation='minus', exon_count=3, OMIM=616403, no_hdr0=''),
  11258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11258, CurrentID=0, Status='live', Symbol='DCTN3', Aliases='DCTN-22, DCTN22', description='dynactin subunit 3', other_designations='dynactin 3 (p22)|dynactin complex subunit 22 kDa subunit|dynactin light chain', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34613545, end_position_on_the_genomic_accession=34620523, orientation='minus', exon_count=7, OMIM=607387, no_hdr0=''),
  11259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11259, CurrentID=0, Status='live', Symbol='FILIP1L', Aliases='DOC-1, DOC1, GIP130, GIP90', description='filamin A interacting protein 1-like', other_designations='130 kDa GPBP-interacting protein|90 kDa GPBP-interacting protein|protein down-regulated in ovarian cancer 1', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99833144, end_position_on_the_genomic_accession=100114513, orientation='minus', exon_count=10, OMIM=612993, no_hdr0=''),
  11260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11260, CurrentID=0, Status='live', Symbol='XPOT', Aliases='XPO3', description='exportin, tRNA', other_designations='exportin(tRNA)|exportin, tRNA (nuclear export receptor for tRNAs)|tRNA exportin', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64404350, end_position_on_the_genomic_accession=64451125, orientation='plus', exon_count=25, OMIM=603180, no_hdr0=''),
  11261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11261, CurrentID=0, Status='live', Symbol='CHP1', Aliases='CHP, SLC9A1BP, Sid470p, p22, p24', description='calcineurin-like EF-hand protein 1', other_designations='EF-hand calcium-binding domain-containing protein p22|SLC9A1 binding protein|calcineurin B homolog|calcineurin B-like protein|calcineurin homologous protein|calcineurin-like EF hand protein 1|calcium binding protein P22|calcium-binding protein CHP|calcium-binding protein p22', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41231149, end_position_on_the_genomic_accession=41281887, orientation='plus', exon_count=8, OMIM=606988, no_hdr0=''),
  11262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11262, CurrentID=0, Status='live', Symbol='SP140', Aliases='LYSP100, LYSP100-A, LYSP100-B', description='SP140 nuclear body protein', other_designations='lymphoid-restricted homolog of Sp100|lymphoid-specific SP100 homolog|nuclear autoantigen Sp-140|speckled 140 kDa', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230202742, end_position_on_the_genomic_accession=230316616, orientation='plus', exon_count=33, OMIM=608602, no_hdr0=''),
  132949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132949, CurrentID=0, Status='live', Symbol='AASDH', Aliases='ACSF4, LYS2, NRPS1098, NRPS998', description='aminoadipate-semialdehyde dehydrogenase', other_designations='2-aminoadipic 6-semialdehyde dehydrogenase|non-ribosomal peptide synthetase 1098|non-ribosomal peptide synthetase 998', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56337687, end_position_on_the_genomic_accession=56389418, orientation='minus', exon_count=17, OMIM=614365, no_hdr0=''),
  11264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11264, CurrentID=0, Status='live', Symbol='PXMP4', Aliases='PMP24', description='peroxisomal membrane protein 4', other_designations='24 kDa peroxisomal intrinsic membrane protein|peroxisomal membrane protein 4, 24kDa', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33702744, end_position_on_the_genomic_accession=33720330, orientation='minus', exon_count=4, OMIM=616397, no_hdr0=''),
  11266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11266, CurrentID=0, Status='live', Symbol='DUSP12', Aliases='DUSP1, YVH1', description='dual specificity phosphatase 12', other_designations='YVH1 protein-tyrosine phosphatase ortholog|dual specificity tyrosine phosphatase YVH1|serine/threonine specific protein phosphatase', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161749768, end_position_on_the_genomic_accession=161757164, orientation='plus', exon_count=6, OMIM=604835, no_hdr0=''),
  11267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11267, CurrentID=0, Status='live', Symbol='SNF8', Aliases='Dot3, EAP30, VPS22', description='SNF8, ESCRT-II complex subunit', other_designations='EAP30 subunit of ELL complex|ELL-associated protein of 30 kDa|ESCRT-II complex subunit VPS22|SNF8, ESCRT-II complex subunit, homolog', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48930096, end_position_on_the_genomic_accession=48945414, orientation='minus', exon_count=8, OMIM=610904, no_hdr0=''),
  11269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11269, CurrentID=0, Status='live', Symbol='DDX19B', Aliases='DBP5, DDX19, RNAh', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 19B', other_designations='ATP-dependent RNA helicase DDX19|DEAD (Asp-Glu-Ala-As) box polypeptide 19B|DEAD box RNA helicase DEAD5|DEAD box protein 19B|DEAD-box RNA helicase DEAD5|DEAD-box protein 5|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 19 (Dbp5, yeast, homolog)|yeast Dbp5 homolog', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70289767, end_position_on_the_genomic_accession=70333832, orientation='plus', exon_count=14, OMIM=605812, no_hdr0=''),
  11270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11270, CurrentID=0, Status='live', Symbol='NRM', Aliases='NRM29', description='nurim (nuclear envelope membrane protein)', other_designations='nuclear rim protein', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30688047, end_position_on_the_genomic_accession=30691420, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11272, CurrentID=0, Status='live', Symbol='PRR4', Aliases='LPRP, PROL4', description='proline rich 4 (lacrimal)', other_designations='lacrimal proline-rich protein|nasopharyngeal carcinoma-associated proline-rich protein 4|proline-rich polypeptide 4', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10845849, end_position_on_the_genomic_accession=10849476, orientation='minus', exon_count=4, OMIM=605359, no_hdr0=''),
  11273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11273, CurrentID=0, Status='live', Symbol='ATXN2L', Aliases='A2D, A2LG, A2LP, A2RP', description='ataxin 2-like', other_designations='ataxin 2 related protein|ataxin-2 domain protein', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28823048, end_position_on_the_genomic_accession=28837237, orientation='plus', exon_count=23, OMIM=607931, no_hdr0=''),
  11274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11274, CurrentID=0, Status='live', Symbol='USP18', Aliases='ISG43, UBP43', description='ubiquitin specific peptidase 18', other_designations='43 kDa ISG15-specific protease|ISG15-specific-processing protease|hUBP43|ubiquitin specific protease 18|ubl thioesterase 18|ubl thiolesterase 18', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18149959, end_position_on_the_genomic_accession=18177397, orientation='plus', exon_count=11, OMIM=607057, no_hdr0=''),
  11275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11275, CurrentID=0, Status='live', Symbol='KLHL2', Aliases='ABP-KELCH, MAV, MAYVEN', description='kelch like family member 2', other_designations='actin-binding protein Mayven|kelch|kelch-like 2, Mayven|kelch-like family member 2', map_location='4q21.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165207618, end_position_on_the_genomic_accession=165323156, orientation='plus', exon_count=20, OMIM=605774, no_hdr0=''),
  11276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11276, CurrentID=0, Status='live', Symbol='SYNRG', Aliases='AP1GBP1, SYNG', description='synergin, gamma', other_designations='AP1 gamma subunit binding protein 1|AP1 subunit gamma-binding protein 1|adaptor-related protein complex 1 gamma subunit-binding protein 1|gamma-synergin', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37514797, end_position_on_the_genomic_accession=37609496, orientation='minus', exon_count=27, OMIM=607291, no_hdr0=''),
  11277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11277, CurrentID=0, Status='live', Symbol='TREX1', Aliases='AGS1, CRV, DRN3, HERNS', description='three prime repair exonuclease 1', other_designations="3' repair exonuclease 1|3'-5' exonuclease TREX1|DNase III|deoxyribonuclease III", map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48465520, end_position_on_the_genomic_accession=48467645, orientation='plus', exon_count=1, OMIM=606609, no_hdr0=''),
  11278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11278, CurrentID=0, Status='live', Symbol='KLF12', Aliases='AP-2rep, AP2REP, HSPC122', description='Kruppel-like factor 12', other_designations='AP-2 repressor|AP-2rep transcription factor|KLF12 zinc finger transcriptional repressor|transcriptional repressor AP-2rep', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73686012, end_position_on_the_genomic_accession=74168169, orientation='minus', exon_count=16, OMIM=607531, no_hdr0=''),
  11279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11279, CurrentID=0, Status='live', Symbol='KLF8', Aliases='BKLF3, ZNF741', description='Kruppel-like factor 8', other_designations='basic krueppel-like factor 3|basic kruppel-like factor 3|zinc finger protein 741', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56232356, end_position_on_the_genomic_accession=56291527, orientation='plus', exon_count=7, OMIM=300286, no_hdr0=''),
  11280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11280, CurrentID=0, Status='live', Symbol='SCN11A', Aliases='FEPS3, HSAN7, NAV1.9, NaN, PN5, SCN12A, SNS-2', description='sodium channel, voltage gated, type XI alpha subunit', other_designations='peripheral nerve sodium channel 5|sensory neuron sodium channel 2|sodium channel protein type XI subunit alpha|sodium channel, voltage-gated, type XI, alpha polypeptide|sodium channel, voltage-gated, type XI, alpha subunit|sodium channel, voltage-gated, type XII, alpha polypeptide|voltage-gated sodium channel subunit alpha Nav1.9', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38845769, end_position_on_the_genomic_accession=38953651, orientation='minus', exon_count=28, OMIM=604385, no_hdr0=''),
  11281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11281, CurrentID=0, Status='live', Symbol='POU6F2', Aliases='RPF-1, WT5, WTSL', description='POU class 6 homeobox 2', other_designations='Wilms tumor suppressor locus|retina-derived POU domain factor 1|retina-derived POU-domain factor-1', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38977904, end_position_on_the_genomic_accession=39468601, orientation='plus', exon_count=12, OMIM=609062, no_hdr0=''),
  11282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11282, CurrentID=0, Status='live', Symbol='MGAT4B', Aliases='GNT-IV, GNT-IVB', description='mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isozyme B', other_designations='N-acetylglucosaminyltransferase IVb|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVb|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IV|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVb|alpha-1,3-mannosyl-glycoprotein beta-1,4-N-acetylglucosaminyltransferase|aminyltransferase|glcNAc-T IVb|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isoenzyme B', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179797597, end_position_on_the_genomic_accession=179806952, orientation='minus', exon_count=15, OMIM=604561, no_hdr0=''),
  11283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11283, CurrentID=0, Status='live', Symbol='CYP4F8', Aliases='CPF8, CYPIVF8', description='cytochrome P450 family 4 subfamily F member 8', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 8|cytochrome P450, subfamily IVF, polypeptide 8|flavoprotein-linked monooxygenase|microsomal monooxygenase', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15615218, end_position_on_the_genomic_accession=15631867, orientation='plus', exon_count=13, OMIM=611545, no_hdr0=''),
  11284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11284, CurrentID=0, Status='live', Symbol='PNKP', Aliases='AOA4, EIEE10, MCSZ, PNK', description="polynucleotide kinase 3'-phosphatase", other_designations="DNA 5'-kinase/3'-phosphatase|Homo sapiens polynucleotide kinase 3'-phosphatase (PNKP)", map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49861203, end_position_on_the_genomic_accession=49867565, orientation='minus', exon_count=17, OMIM=605610, no_hdr0=''),
  11285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11285, CurrentID=0, Status='live', Symbol='B4GALT7', Aliases='EDSP1, XGALT1, XGPT1', description='xylosylprotein beta 1,4-galactosyltransferase, polypeptide 7', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 7|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 7|beta-1,4-GalTase 7|beta-1,4-galactosyltransferase VII|beta4Gal-T7|beta4GalT-VII|galactosyltransferase I|proteoglycan UDP-galactose:beta-xylose beta1,4-galactosyltransferase I', map_location='5q35.2-q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177600118, end_position_on_the_genomic_accession=177610347, orientation='plus', exon_count=8, OMIM=604327, no_hdr0=''),
  100270865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270865, CurrentID=0, Status='live', Symbol='RPS6P10', Aliases='RPS6_4_967', description='ribosomal protein S6 pseudogene 10', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19200323, end_position_on_the_genomic_accession=19201055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873169, CurrentID=0, Status='live', Symbol='MTCYBP4', Aliases='', description='MT-CYB pseudogene 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133150547, end_position_on_the_genomic_accession=133150828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  11309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11309, CurrentID=0, Status='live', Symbol='SLCO2B1', Aliases='OATP-B, OATP2B1, OATPB, SLC21A9', description='solute carrier organic anion transporter family member 2B1', other_designations='OATP-RP2|organic anion transporter B|organic anion transporter polypeptide-related protein 2|solute carrier family 21 (organic anion transporter), member 9|solute carrier organic anion transporter family, member 2B1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75150987, end_position_on_the_genomic_accession=75206400, orientation='plus', exon_count=15, OMIM=604988, no_hdr0=''),
  11311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11311, CurrentID=0, Status='live', Symbol='VPS45', Aliases='H1, H1VPS45, SCN5A, VPS45B, VPS54A, VSP45, VSP45A, VPS45', description='vacuolar protein sorting 45 homolog (S. cerevisiae)', other_designations='hlVps45|leucocyte vacuolar protein sorting 45|vacuolar protein sorting 45A', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150067293, end_position_on_the_genomic_accession=150145327, orientation='plus', exon_count=18, OMIM=610035, no_hdr0=''),
  11313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11313, CurrentID=0, Status='live', Symbol='LYPLA2', Aliases='APT-2, APT2, DJ886K2.4', description='lysophospholipase II', other_designations='LPL-II|lysoPLA II', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23790785, end_position_on_the_genomic_accession=23795539, orientation='plus', exon_count=10, OMIM=616143, no_hdr0=''),
  11314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11314, CurrentID=0, Status='live', Symbol='CD300A', Aliases='CLM-8, CMRF-35-H9, CMRF-35H, CMRF35-H, CMRF35-H9, CMRF35H, CMRF35H9, IGSF12, IRC1, IRC1/IRC2, IRC2, IRp60', description='CD300a molecule', other_designations='CD300 antigen-like family member A|CD300a antigen|CMRF35H leukocyte immunoglobulin-like receptor|NK inhibitory receptor|immunoglobulin superfamily member 12|inhibitory receptor protein 60|leukocyte membrane antigen', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74466366, end_position_on_the_genomic_accession=74484798, orientation='plus', exon_count=8, OMIM=606790, no_hdr0=''),
  11315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11315, CurrentID=0, Status='live', Symbol='PARK7', Aliases='DJ-1, DJ1, HEL-S-67p', description='parkinson protein 7', other_designations='Parkinson disease (autosomal recessive, early onset) 7|epididymis secretory sperm binding protein Li 67p|oncogene DJ1|protein DJ-1', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7961654, end_position_on_the_genomic_accession=7985283, orientation='plus', exon_count=8, OMIM=602533, no_hdr0=''),
  11316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11316, CurrentID=0, Status='live', Symbol='COPE', Aliases='epsilon-COP', description='coatomer protein complex subunit epsilon', other_designations='coatomer epsilon subunit|coatomer protein complex, subunit epsilon|epsilon coat protein|epsilon-coat protein', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18899514, end_position_on_the_genomic_accession=18919464, orientation='minus', exon_count=11, OMIM=606942, no_hdr0=''),
  11317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11317, CurrentID=0, Status='live', Symbol='RBPJL', Aliases='RBPL, RBPSUHL, SUHL', description='recombination signal binding protein for immunoglobulin kappa J region-like', other_designations='transcription factor RBP-L', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45306824, end_position_on_the_genomic_accession=45319677, orientation='plus', exon_count=13, OMIM=616104, no_hdr0=''),
  11318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11318, CurrentID=0, Status='live', Symbol='GPR182', Aliases='7TMR, ADMR, AM-R, AMR, G10D, gamrh, hrhAMR', description='G protein-coupled receptor 182', other_designations='adrenomedullin receptor', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56994355, end_position_on_the_genomic_accession=57000007, orientation='plus', exon_count=5, OMIM=605307, no_hdr0=''),
  11319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11319, CurrentID=0, Status='live', Symbol='ECD', Aliases='GCR2, HSGT1', description='ecdysoneless homolog (Drosophila)', other_designations='human suppressor of GCR two|protein SGT1|suppressor of GCR2|suppressor of S. cerevisiae gcr2', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73134524, end_position_on_the_genomic_accession=73168095, orientation='minus', exon_count=15, OMIM=616464, no_hdr0=''),
  11320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11320, CurrentID=0, Status='live', Symbol='MGAT4A', Aliases='GNT-IV, GNT-IVA, GnT-4a', description='mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isozyme A', other_designations='N-acetylglucosaminyltransferase IVa|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVa|UDP-GlcNAc:a-1,3-D-mannoside b-1,4-acetylglucosaminyltransferase IV|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVa|UDP-N-acetylglucosamine:alpha1,3-d-mannoside beta1,4-N-acetylglucosaminyltransferase|alpha-1,3-mannosyl-glycoprotein beta-1,4-N-acetylglucosaminyltransferase|glcNAc-T IVa|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isoenzyme A', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98619106, end_position_on_the_genomic_accession=98731126, orientation='minus', exon_count=17, OMIM=604623, no_hdr0=''),
  11321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11321, CurrentID=0, Status='live', Symbol='GPN1', Aliases='ATPBD1A, MBDIN, NTPBP, RPAP4, XAB1', description='GPN-loop GTPase 1', other_designations='ATP(GTP)-binding protein|MBD2-interacting protein|RNA polymerase II associated protein 4|RNAPII-associated protein 4|XPA binding protein 1, GTPase', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27628648, end_position_on_the_genomic_accession=27650846, orientation='plus', exon_count=16, OMIM=611479, no_hdr0=''),
  11322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11322, CurrentID=0, Status='live', Symbol='TMC6', Aliases='EV1, EVER1, EVIN1, LAK-4P', description='transmembrane channel like 6', other_designations='epidermodysplasia verruciformis 1|epidermodysplasia verruciformis protein 1|expressed in activated T/LAK lymphocytes|protein LAK-4|transmembrane channel-like 6', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78112918, end_position_on_the_genomic_accession=78132407, orientation='minus', exon_count=21, OMIM=605828, no_hdr0=''),
  11325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11325, CurrentID=0, Status='live', Symbol='DDX42', Aliases='DDX42P, RHELP, RNAHP, SF3B8, SF3b125', description='DEAD (Asp-Glu-Ala-Asp) box helicase 42', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 42|SF3b DEAD box protein|splicing factor 3B-associated 125 kDa protein|splicing factor 3b, subunit 8', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63773884, end_position_on_the_genomic_accession=63819317, orientation='plus', exon_count=20, OMIM=613369, no_hdr0=''),
  11326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11326, CurrentID=0, Status='live', Symbol='VSIG4', Aliases='CRIg, Z39IG', description='V-set and immunoglobulin domain containing 4', other_designations='Ig superfamily protein|complement receptor of the immunoglobulin superfamily|protein Z39Ig', map_location='Xq12-q13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66021738, end_position_on_the_genomic_accession=66040125, orientation='minus', exon_count=8, OMIM=300353, no_hdr0=''),
  404544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404544, CurrentID=0, Status='live', Symbol='COX8BP', Aliases='COX8-1, COX8B, COX8H', description='cytochrome c oxidase subunit VIIIB, pseudogene', other_designations='cytochrome c oxidase, subunit 8B pseudogene', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=253417, end_position_on_the_genomic_accession=256972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  11329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11329, CurrentID=0, Status='live', Symbol='STK38', Aliases='NDR, NDR1', description='serine/threonine kinase 38', other_designations='NDR1 protein kinase|Ndr Ser/Thr kinase-like protein|nuclear Dbf2-related kinase 1', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36493889, end_position_on_the_genomic_accession=36547479, orientation='minus', exon_count=15, OMIM=606964, no_hdr0=''),
  11330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11330, CurrentID=0, Status='live', Symbol='CTRC', Aliases='CLCR, ELA4', description='chymotrypsin C', other_designations='caldecrin|chymotrypsin C (caldecrin)|elastase 4|elastase IV|serum calcium decreasing factor', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15438442, end_position_on_the_genomic_accession=15449247, orientation='plus', exon_count=8, OMIM=601405, no_hdr0=''),
  11331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11331, CurrentID=0, Status='live', Symbol='PHB2', Aliases='BAP, BCAP37, Bap37, PNAS-141, REA, hBAP, p22', description='prohibitin 2', other_designations='B-cell associated protein|B-cell receptor-associated protein BAP37|D-prohibitin|repressor of estrogen receptor activity', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6965352, end_position_on_the_genomic_accession=6970823, orientation='minus', exon_count=10, OMIM=610704, no_hdr0=''),
  11332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11332, CurrentID=0, Status='live', Symbol='ACOT7', Aliases='ACH1, ACT, BACH, CTE-II, LACH, LACH1, hBACH', description='acyl-CoA thioesterase 7', other_designations='CTE-IIa|acyl-CoA thioesterase 2|acyl-CoA thioesterase, long chain|brain acyl CoA hydrolase|brain acyl-CoA hydrolase|long chain acyl-CoA thioester hydrolase', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6264272, end_position_on_the_genomic_accession=6393766, orientation='minus', exon_count=12, OMIM=602587, no_hdr0=''),
  11333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11333, CurrentID=0, Status='live', Symbol='PDAP1', Aliases='HASPP28, PAP, PAP1', description='PDGFA associated protein 1', other_designations='PDGF-associated protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99394673, end_position_on_the_genomic_accession=99408682, orientation='minus', exon_count=6, OMIM=607075, no_hdr0=''),
  11334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11334, CurrentID=0, Status='live', Symbol='TUSC2', Aliases='C3orf11, FUS1, PAP, PDAP2', description='tumor suppressor candidate 2', other_designations='PDGFA associated protein 2|PDGFA-associated protein 2|fus-1 protein|fusion 1 protein', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50324909, end_position_on_the_genomic_accession=50328238, orientation='minus', exon_count=3, OMIM=607052, no_hdr0=''),
  11335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11335, CurrentID=0, Status='live', Symbol='CBX3', Aliases='HECH, HP1-GAMMA, HP1Hs-gamma', description='chromobox 3', other_designations='HP1 gamma homolog|chromobox homolog 3 (HP1 gamma homolog, Drosophila)|heterochromatin protein 1 homolog gamma|heterochromatin protein HP1 gamma|heterochromatin-like protein 1|modifier 2 protein', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26201211, end_position_on_the_genomic_accession=26213607, orientation='plus', exon_count=8, OMIM=604477, no_hdr0=''),
  11336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11336, CurrentID=0, Status='live', Symbol='EXOC3', Aliases='SEC6, SEC6L1, Sec6p', description='exocyst complex component 3', other_designations='SEC6-like 1|Sec 6 homolog|exocyst complex component Sec6', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=443209, end_position_on_the_genomic_accession=467305, orientation='plus', exon_count=14, OMIM=608186, no_hdr0=''),
  11337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11337, CurrentID=0, Status='live', Symbol='GABARAP', Aliases='ATG8A-a, MM46, GABARAP', description='GABA(A) receptor-associated protein', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7240419, end_position_on_the_genomic_accession=7242434, orientation='minus', exon_count=4, OMIM=605125, no_hdr0=''),
  11338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11338, CurrentID=0, Status='live', Symbol='U2AF2', Aliases='U2AF65', description='U2 small nuclear RNA auxiliary factor 2', other_designations='U2 (RNU2) small nuclear RNA auxiliary factor 2|U2 auxiliary factor 65 kDa subunit|U2 small nuclear ribonucleoprotein auxiliary factor (65kD)|U2 snRNP auxiliary factor large subunit|hU2AF65', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55654050, end_position_on_the_genomic_accession=55674716, orientation='plus', exon_count=14, OMIM=191318, no_hdr0=''),
  11339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11339, CurrentID=0, Status='live', Symbol='OIP5', Aliases='5730547N13Rik, CT86, LINT-25, MIS18B, MIS18beta, hMIS18beta', description='Opa interacting protein 5', other_designations='LAP2alpha interactor-25|MIS18 kinetochore protein homolog B|cancer/testis antigen 86|opa-interacting protein 5', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41309268, end_position_on_the_genomic_accession=41332729, orientation='minus', exon_count=5, OMIM=606020, no_hdr0=''),
  11340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11340, CurrentID=0, Status='live', Symbol='EXOSC8', Aliases='CIP3, EAP2, OIP2, PCH1C, RRP43, Rrp43p, bA421P11.3, p9', description='exosome component 8', other_designations='CBP-interacting protein 3|OIP-2|Opa interacting protein 2|exosome complex exonuclease RRP43|ribosomal RNA-processing protein 43', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37000373, end_position_on_the_genomic_accession=37009614, orientation='plus', exon_count=12, OMIM=606019, no_hdr0=''),
  11341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11341, CurrentID=0, Status='live', Symbol='SCRG1', Aliases='SCRG-1', description='stimulator of chondrogenesis 1', other_designations='scrapie responsive gene 1|scrapie responsive protein 1|scrapie-responsive gene 1 protein', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173388148, end_position_on_the_genomic_accession=173476691, orientation='minus', exon_count=5, OMIM=603163, no_hdr0=''),
  11342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11342, CurrentID=0, Status='live', Symbol='RNF13', Aliases='RZF', description='ring finger protein 13', other_designations='RING zinc finger protein', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149812484, end_position_on_the_genomic_accession=149962139, orientation='plus', exon_count=13, OMIM=609247, no_hdr0=''),
  11343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11343, CurrentID=0, Status='live', Symbol='MGLL', Aliases='HU-K5, HUK5, MAGL, MGL', description='monoglyceride lipase', other_designations='lysophospholipase homolog|monoacylglycerol lipase', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127689062, end_position_on_the_genomic_accession=127823250, orientation='minus', exon_count=12, OMIM=609699, no_hdr0=''),
  11344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11344, CurrentID=0, Status='live', Symbol='TWF2', Aliases='A6RP, A6r, MSTP011, PTK9L', description='twinfilin actin binding protein 2', other_designations='A6-related protein|PTK9L protein tyrosine kinase 9-like (A6-related protein)|protein tyrosine kinase 9-like (A6-related protein)|twinfilin actin-binding protein 2|twinfilin, actin-binding protein, homolog 2|twinfilin-1-like protein', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52228610, end_position_on_the_genomic_accession=52239167, orientation='minus', exon_count=9, OMIM=607433, no_hdr0=''),
  11345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11345, CurrentID=0, Status='live', Symbol='GABARAPL2', Aliases='ATG8, ATG8C, GATE-16, GATE16, GEF-2, GEF2', description='GABA(A) receptor-associated protein like 2', other_designations='GABA(A) receptor-associated protein-like 2|MAP1 light chain 3 related protein|MAP1 light chain 3-related protein|ganglioside expression factor 2|general protein transport factor p16|golgi-associated ATPase enhancer of 16 kDa', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75566351, end_position_on_the_genomic_accession=75577881, orientation='plus', exon_count=4, OMIM=607452, no_hdr0=''),
  11346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11346, CurrentID=0, Status='live', Symbol='SYNPO', Aliases='', description='synaptopodin', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150586037, end_position_on_the_genomic_accession=150659230, orientation='plus', exon_count=10, OMIM=608155, no_hdr0=''),
  100506551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506551, CurrentID=0, Status='live', Symbol='LOC100506551', Aliases='', description='uncharacterized LOC100506551', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116977440, end_position_on_the_genomic_accession=116987337, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23738, CurrentID=0, Status='live', Symbol='OFD1P17', Aliases='71-7A3, 71-7A5, C5orf1, OFD1P1', description='OFD1 pseudogene 17', other_designations='OFD1 pseudogene 1|oral-facial-digital syndrome 1 pseudogene|oral-facial-digital syndrome 1, pseudogene 1', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37208999, end_position_on_the_genomic_accession=37212497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873171, CurrentID=0, Status='live', Symbol='MTND1P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 4', other_designations='MT-ND1 pseudogene 4', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57185408, end_position_on_the_genomic_accession=57186201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140043, CurrentID=0, Status='live', Symbol='CDY10P', Aliases='', description='chromodomain protein, Y-linked 10 pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21636601, end_position_on_the_genomic_accession=21639468, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270868, CurrentID=0, Status='live', Symbol='RPS7P2', Aliases='RPS7_1_102', description='ribosomal protein S7 pseudogene 2', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154079422, end_position_on_the_genomic_accession=154079965, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873172, CurrentID=0, Status='live', Symbol='MTND1P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 7', other_designations='MT-ND1 pseudogene 7', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46828117, end_position_on_the_genomic_accession=46828413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377357, CurrentID=0, Status='live', Symbol='LOC105377357', Aliases='', description='uncharacterized LOC105377357', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107579145, end_position_on_the_genomic_accession=107600550, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23744, CurrentID=0, Status='live', Symbol='ATP5EP1', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F1 complex, epsilon subunit pseudogene 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105532388, end_position_on_the_genomic_accession=105532806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  404635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404635, CurrentID=0, Status='live', Symbol='NANOGP1', Aliases='NANOG2', description='Nanog homeobox pseudogene 1', other_designations='Nanog homeobox 2|nanog-like protein', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7892356, end_position_on_the_genomic_accession=7899246, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  404636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404636, CurrentID=0, Status='live', Symbol='FAM45A', Aliases='', description='family with sequence similarity 45 member A', other_designations='family with sequence similarity 45, member A', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119104065, end_position_on_the_genomic_accession=119137984, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23749, CurrentID=0, Status='live', Symbol='YME1L1P1', Aliases='YME1L2', description='YME1L1 pseudogene 1', other_designations='YME1-like 1 pseudogene 1', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15489984, end_position_on_the_genomic_accession=15492567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270870, CurrentID=0, Status='live', Symbol='RPS7P7', Aliases='RPS7_3_472', description='ribosomal protein S7 pseudogene 7', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42471775, end_position_on_the_genomic_accession=42472393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  404663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404663, CurrentID=0, Status='live', Symbol='LINC01194', Aliases='CT49, TAG', description='long intergenic non-protein coding RNA 1194', other_designations='cancer/testis antigen 49 (non-protein coding)|tumor antigen', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12574857, end_position_on_the_genomic_accession=12805183, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  404665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404665, CurrentID=0, Status='live', Symbol='CACTIN-AS1', Aliases='C19orf29-AS1, C19orf29OS', description='CACTIN antisense RNA 1', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3607247, end_position_on_the_genomic_accession=3613930, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  404672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404672, CurrentID=0, Status='live', Symbol='GTF2H5', Aliases='C6orf175, TFB5, TFIIH, TGF2H5, TTD, TTD-A, TTD3, TTDA, bA120J8.2', description='general transcription factor IIH subunit 5', other_designations='TFB5 ortholog|TFIIH basal transcription factor complex TTD-A subunit|TFIIH basal transcription factor complex TTDA subunit|general transcription factor IIH, polypeptide 5', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158168347, end_position_on_the_genomic_accession=158199344, orientation='plus', exon_count=3, OMIM=608780, no_hdr0=''),
  404677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404677, CurrentID=0, Status='live', Symbol='CIMT', Aliases='', description='Carotid intimal medial thickness', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608447, no_hdr0=''),
  285900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285900, CurrentID=0, Status='live', Symbol='RPL6P20', Aliases='RPL6_6_811', description='ribosomal protein L6 pseudogene 20', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64141493, end_position_on_the_genomic_accession=64142420, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  404682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404682, CurrentID=0, Status='live', Symbol='MYP5', Aliases='', description='myopia 5 (high grade, autosomal dominant)', other_designations='Myopia-5', map_location='17q21-q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608474, no_hdr0=''),
  404683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404683, CurrentID=0, Status='live', Symbol='BMIQ7', Aliases='OB4', description='Obesity, susceptibility to, on chromosome 4', other_designations='', map_location='4p15-p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608410, no_hdr0=''),
  404684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404684, CurrentID=0, Status='live', Symbol='PHOBS', Aliases='', description='Phobia, specific', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608251, no_hdr0=''),
  404685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404685, CurrentID=0, Status='live', Symbol='RTE1', Aliases='', description='repressor of telomerase expression 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608045, no_hdr0=''),
  404686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404686, CurrentID=0, Status='live', Symbol='SCZD11', Aliases='', description='Schizophrenia susceptibility locus, chromosome 10q-related', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608078, no_hdr0=''),
  285902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285902, CurrentID=0, Status='live', Symbol='LOC285902', Aliases='', description='uncharacterized LOC285902', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285905, CurrentID=0, Status='live', Symbol='INTS4P1', Aliases='INTS4L1', description='integrator complex subunit 4 pseudogene 1', other_designations='integrator complex subunit 4-like 1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65141193, end_position_on_the_genomic_accession=65233303, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  404713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404713, CurrentID=0, Status='live', Symbol='SSD', Aliases='', description='Speech-sound disorder', other_designations='', map_location='3p12-q13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608445, no_hdr0=''),
  404714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404714, CurrentID=0, Status='live', Symbol='SLEB4', Aliases='', description='systemic lupus erythematosus, susceptibility to, 4', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608437, no_hdr0=''),
  404715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404715, CurrentID=0, Status='live', Symbol='PAFC', Aliases='', description='Preauricular fistulae, congenital', other_designations='', map_location='8q11.1-q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=128700, no_hdr0=''),
  404716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404716, CurrentID=0, Status='live', Symbol='MTBS1', Aliases='', description='mycobacterium tuberculosis, susceptibility to infection by, 1', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607949, no_hdr0=''),
  404718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404718, CurrentID=0, Status='live', Symbol='GNRHR2P1', Aliases='', description='GNRHR2 pseudogene 1', other_designations='gonadotropin-releasing hormone (type 2) receptor 2 pseudogene 1|gonadotropin-releasing hormone receptor 2 pseudogene', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60398823, end_position_on_the_genomic_accession=60400013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132989, CurrentID=0, Status='live', Symbol='C4orf36', Aliases='', description='chromosome 4 open reading frame 36', other_designations='', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86876205, end_position_on_the_genomic_accession=86897123, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  404720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404720, CurrentID=0, Status='live', Symbol='HSCR5', Aliases='', description='Hirschsprung disease, susceptibility to, 5', other_designations='', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600156, no_hdr0=''),
  285908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285908, CurrentID=0, Status='live', Symbol='LINC00174', Aliases='NCRNA00174', description='long intergenic non-protein coding RNA 174', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66376044, end_position_on_the_genomic_accession=66400408, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  404734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404734, CurrentID=0, Status='live', Symbol='ANKHD1-EIF4EBP3', Aliases='MASK-BP3, MASK-BP3ARF', description='ANKHD1-EIF4EBP3 readthrough', other_designations='MASK-4E-BP3 alternate reading frame', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140401814, end_position_on_the_genomic_accession=140549578, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  404744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404744, CurrentID=0, Status='live', Symbol='NPSR1-AS1', Aliases='AAA1', description='NPSR1 antisense RNA 1', other_designations='NPSR1 antisense RNA 1 (non-protein coding)|asthma-associated alternatively spliced gene 1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34346512, end_position_on_the_genomic_accession=34834331, orientation='minus', exon_count=9, OMIM=608596, no_hdr0=''),
  100270873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270873, CurrentID=0, Status='live', Symbol='RPSAP17', Aliases='RPSA_3_103', description='ribosomal protein SA pseudogene 17', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154378126, end_position_on_the_genomic_accession=154379169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  404766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404766, CurrentID=0, Status='live', Symbol='FABP5P4', Aliases='FABP5L4', description='fatty acid binding protein 5 pseudogene 4', other_designations='fatty acid binding protein 5-like 4', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92180502, end_position_on_the_genomic_accession=92180885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  404770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404770, CurrentID=0, Status='live', Symbol='SPATA31B1P', Aliases='C9orf36B, FAM75B, SPATA31B1, bA388B24.2', description='SPATA31 subfamily B, member 1, pseudogene', other_designations='family with sequence similarity 75, member B', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82057542, end_position_on_the_genomic_accession=82063724, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371132, CurrentID=0, Status='live', Symbol='LOC105371132', Aliases='', description='uncharacterized LOC105371132', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22577466, end_position_on_the_genomic_accession=22608063, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  404785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404785, CurrentID=0, Status='live', Symbol='POTEG', Aliases='A26C2, ACTBL1, CT104.4, POTE-14, POTE14, POTE14alpha, POTE22', description='POTE ankyrin domain family member G', other_designations='ANKRD26-like family C member 2|ANKRD26-like family C, member 2|POTE ankyrin domain family, member G|cancer/testis antigen family 104, member 4|prostate, ovary, testis-expressed protein on chromosome 14|protein expressed in prostate, ovary, testis, and placenta 14', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19402764, end_position_on_the_genomic_accession=19434341, orientation='minus', exon_count=12, OMIM=608916, no_hdr0=''),
  266621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266621, CurrentID=0, Status='live', Symbol='DGCR7', Aliases='', description='DiGeorge syndrome critical region gene 7', other_designations='', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100532746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532746, CurrentID=0, Status='live', Symbol='PPT2-EGFL8', Aliases='', description='PPT2-EGFL8 readthrough (NMD candidate)', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32153999, end_position_on_the_genomic_accession=32168285, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100270875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270875, CurrentID=0, Status='live', Symbol='RPSAP23', Aliases='RPSA_5_268', description='ribosomal protein SA pseudogene 23', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115467985, end_position_on_the_genomic_accession=115468953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  142678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142678, CurrentID=0, Status='live', Symbol='MIB2', Aliases='ZZANK1, ZZZ5', description='mindbomb E3 ubiquitin protein ligase 2', other_designations='mind bomb homolog 2|mindbomb homolog 2|novel zinc finger protein|novelzin|putative NF-kappa-B-activating protein 002N|skeletrophin|zinc finger ZZ type with ankyrin repeat domain protein 1|zinc finger, ZZ type with ankyrin repeat domain 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1614975, end_position_on_the_genomic_accession=1630610, orientation='plus', exon_count=24, OMIM=611141, no_hdr0=''),
  142679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142679, CurrentID=0, Status='live', Symbol='DUSP19', Aliases='DUSP17, LMWDSP3, SKRP1, TS-DSP1', description='dual specificity phosphatase 19', other_designations='LMW-DSP3|SAPK pathway-regulating phosphatase 1|dual specificity phosphatase TS-DSP1|low molecular weight dual specificity phosphatase 3|protein phosphatase SKRP1|stress-activated protein kinase pathway-regulating phosphatase 1', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183078224, end_position_on_the_genomic_accession=183100008, orientation='plus', exon_count=4, OMIM=611437, no_hdr0=''),
  142680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142680, CurrentID=0, Status='live', Symbol='SLC34A3', Aliases='HHRH, NPTIIc', description='solute carrier family 34 (type II sodium/phosphate cotransporter), member 3', other_designations='Na(+)-dependent phosphate cotransporter 2C|Na(+)/Pi cotransporter 2C|naPi-2c|sodium-phosphate transport protein 2C|sodium/inorganic phosphate cotransporter IIC|solute carrier family 34 (sodium phosphate), member 3|type IIc Na+/Pi cotransporter', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137230116, end_position_on_the_genomic_accession=137236554, orientation='plus', exon_count=15, OMIM=609826, no_hdr0=''),
  100359396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359396, CurrentID=0, Status='live', Symbol='HSPD1P11', Aliases='HSPD1-11P', description='heat shock protein family D (Hsp60) member 1 pseudogene 11', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 11', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95769697, end_position_on_the_genomic_accession=95771187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  142683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142683, CurrentID=0, Status='live', Symbol='ITLN2', Aliases='HL-2, HL2', description='intelectin 2', other_designations='endothelial lectin HL-2', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160945020, end_position_on_the_genomic_accession=160957378, orientation='minus', exon_count=10, OMIM=609874, no_hdr0=''),
  142684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142684, CurrentID=0, Status='live', Symbol='RAB40A', Aliases='RAR2, RAR2A', description='RAB40A, member RAS oncogene family', other_designations='SOCS box containing protein RAR2A|SOCS box-containing protein RAR2A|protein Rar-2', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103499136, end_position_on_the_genomic_accession=103519489, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  142685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142685, CurrentID=0, Status='live', Symbol='ASB15', Aliases='', description='ankyrin repeat and SOCS box containing 15', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123566999, end_position_on_the_genomic_accession=123637900, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  142686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142686, CurrentID=0, Status='live', Symbol='ASB14', Aliases='', description='ankyrin repeat and SOCS box containing 14', other_designations='ankyrin repeat and SOCS box-containing 14|ankyrin repeat domain-containing SOCS box protein Asb-14', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57268351, end_position_on_the_genomic_accession=57292720, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  142689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142689, CurrentID=0, Status='live', Symbol='ASB12', Aliases='', description='ankyrin repeat and SOCS box containing 12', other_designations='ankyrin repeat and SOCS box-containing 12|ankyrin repeat domain-containing SOCS box protein Asb-12', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64224192, end_position_on_the_genomic_accession=64230631, orientation='minus', exon_count=3, OMIM=300891, no_hdr0=''),
  100270877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270877, CurrentID=0, Status='live', Symbol='RPL21P21', Aliases='RPL21_1_11', description='ribosomal protein L21 pseudogene 21', other_designations='', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10054440, end_position_on_the_genomic_accession=10054782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  154860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154860, CurrentID=0, Status='live', Symbol='FEZF1-AS1', Aliases='', description='FEZF1 antisense RNA 1', other_designations='FEZF1 antisense RNA 1 (non-protein coding)', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122303658, end_position_on_the_genomic_accession=122310077, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  133015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133015, CurrentID=0, Status='live', Symbol='PACRGL', Aliases='C4orf28', description='PARK2 co-regulated like', other_designations='PARK2 co-regulated-like', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20696287, end_position_on_the_genomic_accession=20766315, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  391136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391136, CurrentID=0, Status='live', Symbol='LOC391136', Aliases='', description='RNA binding motif protein 3 pseudogene', other_designations='RNA binding motif (RNP1, RRM) protein 3 pseudogene', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171761994, end_position_on_the_genomic_accession=171762966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100532749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532749, CurrentID=0, Status='live', Symbol='LOC100532749', Aliases='', description='ubiquitination factor E4A (UFD2 homolog, yeast) pseudogene', other_designations='', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112245957, end_position_on_the_genomic_accession=112247866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270878, CurrentID=0, Status='live', Symbol='RPL21P22', Aliases='RPL21_2_29', description='ribosomal protein L21 pseudogene 22', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31108101, end_position_on_the_genomic_accession=31108663, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  133022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133022, CurrentID=0, Status='live', Symbol='TRAM1L1', Aliases='', description='translocation associated membrane protein 1-like 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117083554, end_position_on_the_genomic_accession=117085580, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131124, CurrentID=0, Status='live', Symbol='LOC100131124', Aliases='', description='adaptor-related protein complex 4, sigma 1 subunit pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89059628, end_position_on_the_genomic_accession=89060061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270879, CurrentID=0, Status='live', Symbol='RPL21P23', Aliases='RPL21_2_54', description='ribosomal protein L21 pseudogene 23', other_designations='', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56538412, end_position_on_the_genomic_accession=56538957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373004, CurrentID=0, Status='live', Symbol='LL22NC03-104C7.1', Aliases='', description='uncharacterized LOC105373004', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32582644, end_position_on_the_genomic_accession=32587010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100270880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270880, CurrentID=0, Status='live', Symbol='RPL21P24', Aliases='RPL21_4_45', description='ribosomal protein L21 pseudogene 24', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47497862, end_position_on_the_genomic_accession=47498381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  142827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142827, CurrentID=0, Status='live', Symbol='ACSM6', Aliases='C10orf129, bA310E22.3', description='acyl-CoA synthetase medium-chain family member 6', other_designations='AMP-binding enzyme', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95194200, end_position_on_the_genomic_accession=95228929, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100131216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131216, CurrentID=0, Status='live', Symbol='LOC100131216', Aliases='', description='transmembrane protein 38B pseudogene', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176113700, end_position_on_the_genomic_accession=176114376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288339, CurrentID=0, Status='live', Symbol='NDUFA5P6', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 6', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 6|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126524342, end_position_on_the_genomic_accession=126525705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270881, CurrentID=0, Status='live', Symbol='RPL21P25', Aliases='RPL21_5_46', description='ribosomal protein L21 pseudogene 25', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48497257, end_position_on_the_genomic_accession=48497736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377370, CurrentID=0, Status='live', Symbol='LOC105377370', Aliases='', description='uncharacterized LOC105377370', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112120843, end_position_on_the_genomic_accession=112131077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271237, CurrentID=0, Status='live', Symbol='RPS26P27', Aliases='RPS26_12_587', description='ribosomal protein S26 pseudogene 27', other_designations='', map_location='5q11-q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80855505, end_position_on_the_genomic_accession=80855839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481594, CurrentID=0, Status='live', Symbol='RNU6-1312P', Aliases='', description='RNA, U6 small nuclear 1312, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270882, CurrentID=0, Status='live', Symbol='RPL21P26', Aliases='RPL21_6_74', description='ribosomal protein L21 pseudogene 26', other_designations='', map_location='1p22-p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94406387, end_position_on_the_genomic_accession=94406922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  142891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142891, CurrentID=0, Status='live', Symbol='SAMD8', Aliases='HEL-177, SMSr', description='sterile alpha motif domain containing 8', other_designations='CPE synthase|HEL-S-181mP|SAM domain-containing protein 8|ceramide phosphoethanolamine synthase|epididymis luminal protein 177|epididymis secretory sperm binding protein Li 181mP|sphingomyelin synthase related|sterile alpha motif domain-containing protein 8', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75099592, end_position_on_the_genomic_accession=75182123, orientation='plus', exon_count=7, OMIM=611575, no_hdr0=''),
  142910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142910, CurrentID=0, Status='live', Symbol='LIPJ', Aliases='LIPL1, bA425M17.2', description='lipase, family member J', other_designations='lipase J|lipase-like abhydrolase domain-containing protein 1|lipase-like, ab-hydrolase domain containing 1', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88583277, end_position_on_the_genomic_accession=88606976, orientation='plus', exon_count=15, OMIM=613921, no_hdr0=''),
  142913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142913, CurrentID=0, Status='live', Symbol='CFL1P1', Aliases='CFLP1', description='cofilin 1 (non-muscle) pseudogene 1', other_designations='cofilin pseudogene 1', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87818313, end_position_on_the_genomic_accession=87845612, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373008, CurrentID=0, Status='live', Symbol='LOC105373008', Aliases='', description='uncharacterized LOC105373008', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33926364, end_position_on_the_genomic_accession=33936950, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371134, CurrentID=0, Status='live', Symbol='LOC105371134', Aliases='', description='uncharacterized LOC105371134', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270884, CurrentID=0, Status='live', Symbol='RPL26P8', Aliases='RPL26_1_22', description='ribosomal protein L26 pseudogene 8', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24563597, end_position_on_the_genomic_accession=24563971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  142937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142937, CurrentID=0, Status='live', Symbol='LOC142937', Aliases='', description='uncharacterized protein BC008131', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=6253747, end_position_on_the_genomic_accession=6254800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  142940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=142940, CurrentID=0, Status='live', Symbol='TRUB1', Aliases='PUS4', description='TruB pseudouridine (psi) synthase family member 1', other_designations='TruB pseudouridine (psi) synthase homolog 1', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114938193, end_position_on_the_genomic_accession=114977680, orientation='plus', exon_count=9, OMIM=610726, no_hdr0=''),
  105377373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377373, CurrentID=0, Status='live', Symbol='LOC105377373', Aliases='', description='uncharacterized LOC105377373', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113029508, end_position_on_the_genomic_accession=113045225, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373684, CurrentID=0, Status='live', Symbol='LOC105373684', Aliases='', description='uncharacterized LOC105373684', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150169491, end_position_on_the_genomic_accession=150353345, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103689843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689843, CurrentID=0, Status='live', Symbol='LINC01046', Aliases='', description='long intergenic non-protein coding RNA 1046', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21234568, end_position_on_the_genomic_accession=21235204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103689844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689844, CurrentID=0, Status='live', Symbol='LINC01288', Aliases='TCONS_00014671', description='long intergenic non-protein coding RNA 1288', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34783921, end_position_on_the_genomic_accession=34864798, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100270885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270885, CurrentID=0, Status='live', Symbol='RPL26P9', Aliases='RPL26_2_76', description='ribosomal protein L26 pseudogene 9', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97585866, end_position_on_the_genomic_accession=97586300, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103689846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689846, CurrentID=0, Status='live', Symbol='ADAMTS19-AS1', Aliases='', description='ADAMTS19 antisense RNA 1', other_designations='RP11-346A9.1', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129459559, end_position_on_the_genomic_accession=129460689, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100270886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270886, CurrentID=0, Status='live', Symbol='RPL26P10', Aliases='RPL26_3_90', description='ribosomal protein L26 pseudogene 10', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114459869, end_position_on_the_genomic_accession=114460380, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270887, CurrentID=0, Status='live', Symbol='RPL31P12', Aliases='RPL31_1_62', description='ribosomal protein L31 pseudogene 12', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72301445, end_position_on_the_genomic_accession=72301871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646347, CurrentID=0, Status='live', Symbol='LOC646347', Aliases='', description='spermine synthase pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160894889, end_position_on_the_genomic_accession=160896275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103689911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689911, CurrentID=0, Status='live', Symbol='CATIP-AS2', Aliases='', description='CATIP antisense RNA 2', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218326241, end_position_on_the_genomic_accession=218357966, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  103689912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689912, CurrentID=0, Status='live', Symbol='TBX2-AS1', Aliases='', description='TBX2 antisense RNA 1', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61393456, end_position_on_the_genomic_accession=61399606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103689913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689913, CurrentID=0, Status='live', Symbol='LINC00383', Aliases='LINC00401', description='long intergenic non-protein coding RNA 383', other_designations='long intergenic non-protein coding RNA 401', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69222346, end_position_on_the_genomic_accession=69322101, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103689914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689914, CurrentID=0, Status='live', Symbol='LOC103689914', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 10 pseudogene', other_designations='', map_location='7p', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38311308, end_position_on_the_genomic_accession=38311978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103689915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689915, CurrentID=0, Status='live', Symbol='DLEU1-AS1', Aliases='LINC01308', description='DLEU1 antisense RNA 1', other_designations='long intergenic non-protein coding RNA 1308', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50520933, end_position_on_the_genomic_accession=50527449, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103689916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689916, CurrentID=0, Status='live', Symbol='LOC103689916', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 10 pseudogene', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10216989, end_position_on_the_genomic_accession=10217349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103689917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689917, CurrentID=0, Status='live', Symbol='LINC01337', Aliases='TCONS_00009713', description='long intergenic non-protein coding RNA 1337', other_designations='CTC-325J23.3', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80608620, end_position_on_the_genomic_accession=80622524, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103689918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689918, CurrentID=0, Status='live', Symbol='LINC01378', Aliases='', description='long intergenic non-protein coding RNA 1378', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117428398, end_position_on_the_genomic_accession=117689103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131223, CurrentID=0, Status='live', Symbol='LOC100131223', Aliases='', description='ADP ribosylation factor like GTPase 8B pseudogene', other_designations='ADP-ribosylation factor like GTPase 8B pseudogene|ADP-ribosylation factor-like 8B pseudogene', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28120835, end_position_on_the_genomic_accession=28121462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270888, CurrentID=0, Status='live', Symbol='RPL32P6', Aliases='RPL32_1_26', description='ribosomal protein L32 pseudogene 6', other_designations='', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26983599, end_position_on_the_genomic_accession=26984024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646348, CurrentID=0, Status='live', Symbol='ARL6IP1P2', Aliases='', description='ADP ribosylation factor like GTPase 6 interacting protein 1 pseudogene 2', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 1 pseudogene 2|ADP-ribosylation factor like GTPase 6 interacting protein 1 pseudogene 2|ADP-ribosylation factor-like 6 interacting protein 1 pseudogene 2', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36995023, end_position_on_the_genomic_accession=36995587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270889, CurrentID=0, Status='live', Symbol='RPL34P4', Aliases='RPL34_1_24', description='ribosomal protein L34 pseudogene 4', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26326687, end_position_on_the_genomic_accession=26327039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729275, CurrentID=0, Status='live', Symbol='LOC729275', Aliases='', description='uncharacterized LOC729275', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=48431482, end_position_on_the_genomic_accession=48434797, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  143098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143098, CurrentID=0, Status='live', Symbol='MPP7', Aliases='', description='membrane protein, palmitoylated 7', other_designations='membrane protein, palmitoylated 7 (MAGUK p55 subfamily member 7)', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28050993, end_position_on_the_genomic_accession=28303061, orientation='minus', exon_count=26, OMIM=610973, no_hdr0=''),
  105373014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373014, CurrentID=0, Status='live', Symbol='LOC105373014', Aliases='', description='uncharacterized LOC105373014', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34764128, end_position_on_the_genomic_accession=34767698, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372881, CurrentID=0, Status='live', Symbol='LOC105372881', Aliases='', description='uncharacterized LOC105372881', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207368580, end_position_on_the_genomic_accession=207373246, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100270890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270890, CurrentID=0, Status='live', Symbol='RPL37P7', Aliases='RPL37_1_52', description='ribosomal protein L37 pseudogene 7', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53841545, end_position_on_the_genomic_accession=53841834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270790, CurrentID=0, Status='live', Symbol='GNL2P1', Aliases='', description='guanine nucleotide binding protein-like 2 (nucleolar) pseudogene 1', other_designations='', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=49942210, end_position_on_the_genomic_accession=49944528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  133083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133083, CurrentID=0, Status='live', Symbol='PMPCAP1', Aliases='', description='peptidase (mitochondrial processing) alpha pseudogene 1', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92181956, end_position_on_the_genomic_accession=92184078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270891, CurrentID=0, Status='live', Symbol='RPL39P6', Aliases='RPL39_1_12', description='ribosomal protein L39 pseudogene 6', other_designations='', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11232910, end_position_on_the_genomic_accession=11233310, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79514, CurrentID=0, Status='live', Symbol='OR10T1P', Aliases='', description='olfactory receptor family 10 subfamily T member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily T, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158445068, end_position_on_the_genomic_accession=158446014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143162, CurrentID=0, Status='live', Symbol='FRMPD2', Aliases='PDZD5C, PDZK4, PDZK5C', description='FERM and PDZ domain containing 2', other_designations='PDZ domain containing 5C|PDZ domain-containing protein 4', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48150001, end_position_on_the_genomic_accession=48274898, orientation='minus', exon_count=33, OMIM=613323, no_hdr0=''),
  100128508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128508, CurrentID=0, Status='live', Symbol='LOC100128508', Aliases='', description='PP12100', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=14775189, end_position_on_the_genomic_accession=14777141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  143187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143187, CurrentID=0, Status='live', Symbol='VTI1A', Aliases='MMDS3, MVti1, VTI1RP2, Vti1-rp2', description='vesicle transport through interaction with t-SNAREs 1A', other_designations='SNARE Vti1a-beta protein|vesicle transport v-SNARE protein Vti1-like 2', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112446998, end_position_on_the_genomic_accession=112855364, orientation='plus', exon_count=15, OMIM=614316, no_hdr0=''),
  143188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143188, CurrentID=0, Status='live', Symbol='LOC143188', Aliases='', description='uncharacterized LOC143188', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=114610673, end_position_on_the_genomic_accession=114615127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100270893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270893, CurrentID=0, Status='live', Symbol='RPL39P8', Aliases='RPL39_2_85', description='ribosomal protein L39 pseudogene 8', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112890592, end_position_on_the_genomic_accession=112890971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929411, CurrentID=0, Status='live', Symbol='LOC101929411', Aliases='', description='uncharacterized LOC101929411', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102163638, end_position_on_the_genomic_accession=102164879, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102216341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102216341, CurrentID=0, Status='live', Symbol='LINC01109', Aliases='lncRNA-N1', description='long intergenic non-protein coding RNA 1109', other_designations='lncRNA neuronal 1', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270894, CurrentID=0, Status='live', Symbol='RPL39P9', Aliases='RPL39_4_77', description='ribosomal protein L39 pseudogene 9', other_designations='', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99842561, end_position_on_the_genomic_accession=99842962, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143241, CurrentID=0, Status='live', Symbol='DYDC1', Aliases='DPY30D1', description='DPY30 domain containing 1', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80336102, end_position_on_the_genomic_accession=80356787, orientation='minus', exon_count=8, OMIM=615154, no_hdr0=''),
  105377383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377383, CurrentID=0, Status='live', Symbol='LOC105377383', Aliases='', description='uncharacterized LOC105377383', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116282074, end_position_on_the_genomic_accession=116297174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  143244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143244, CurrentID=0, Status='live', Symbol='EIF5AL1', Aliases='EIF5AP1, bA342M3.3', description='eukaryotic translation initiation factor 5A-like 1', other_designations='eukaryotic translation initiation factor 5A pseudogene 1', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79512601, end_position_on_the_genomic_accession=79516440, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100270895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270895, CurrentID=0, Status='live', Symbol='RPL7AP15', Aliases='RPL7A_1_95', description='ribosomal protein L7a pseudogene 15', other_designations='', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147223533, end_position_on_the_genomic_accession=147224403, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481463, CurrentID=0, Status='live', Symbol='RNU6-900P', Aliases='', description='RNA, U6 small nuclear 900, pseudogene', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  143279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143279, CurrentID=0, Status='live', Symbol='HECTD2', Aliases='', description='HECT domain containing E3 ubiquitin protein ligase 2', other_designations='HECT domain containing 2|HECT domain-containing protein 2', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91409280, end_position_on_the_genomic_accession=91514829, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  143282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143282, CurrentID=0, Status='live', Symbol='FGFBP3', Aliases='C10orf13, FGF-BP3', description='fibroblast growth factor binding protein 3', other_designations='FGF-binding protein 3|FGFBP-3', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91906588, end_position_on_the_genomic_accession=91909501, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  143286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143286, CurrentID=0, Status='live', Symbol='LOC143286', Aliases='', description='uncharacterized LOC143286', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270896, CurrentID=0, Status='live', Symbol='RPL7AP16', Aliases='RPL7A_2_43', description='ribosomal protein L7a pseudogene 16', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45650978, end_position_on_the_genomic_accession=45651850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79515, CurrentID=0, Status='live', Symbol='OR10G5P', Aliases='', description='olfactory receptor family 10 subfamily G member 5 pseudogene', other_designations='olfactory receptor, family 10, subfamily G, member 5 pseudogene', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124009512, end_position_on_the_genomic_accession=124010416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266643, CurrentID=0, Status='live', Symbol='EDDM3CP', Aliases='FAM12CP, HE3-GAMMA', description='epididymal protein 3C, pseudogene', other_designations='epididymal protein 3C (pseudogene)|family with sequence similarity 12, member C (pseudogene)', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134319554, end_position_on_the_genomic_accession=134320175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143341, CurrentID=0, Status='live', Symbol='EIF3LP2', Aliases='EIF3EIPP2, EIF3S6IPP2', description='eukaryotic translation initiation factor 3 subunit L pseudogene 2', other_designations='eukaryotic translation initiation factor 3, subunit 6 interacting protein pseudogene 2|eukaryotic translation initiation factor 3, subunit 6 interacting protein, putative pseudogene|eukaryotic translation initiation factor 3, subunit E interacting protein pseudogene 2|eukaryotic translation initiation factor 3, subunit L pseudogene 2', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42566676, end_position_on_the_genomic_accession=42567666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131233, CurrentID=0, Status='live', Symbol='LOC100131233', Aliases='', description='AKT interacting protein pseudogene', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96642687, end_position_on_the_genomic_accession=96643225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270898, CurrentID=0, Status='live', Symbol='RPS11P3', Aliases='RPS11_1_99', description='ribosomal protein S11 pseudogene 3', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151758081, end_position_on_the_genomic_accession=151758432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  133121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133121, CurrentID=0, Status='live', Symbol='ENPP6', Aliases='NPP6', description='ectonucleotide pyrophosphatase/phosphodiesterase 6', other_designations='B830047L21Rik|E-NPP 6|GPC-Cpde|NPP-6|choline-specific glycerophosphodiester phosphodiesterase|glycerophosphocholine cholinephosphodiesterase', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184088706, end_position_on_the_genomic_accession=184217961, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  143371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143371, CurrentID=0, Status='live', Symbol='YWHAZP3', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 3', other_designations='', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23136910, end_position_on_the_genomic_accession=23139778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143379, CurrentID=0, Status='live', Symbol='C10orf82', Aliases='', description='chromosome 10 open reading frame 82', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116662211, end_position_on_the_genomic_accession=116670269, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  143384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143384, CurrentID=0, Status='live', Symbol='CACUL1', Aliases='C10orf46, CAC1', description='CDK2-associated, cullin domain 1', other_designations='Cdk-Associated Cullin1|cullin', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118672993, end_position_on_the_genomic_accession=118755246, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100270899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270899, CurrentID=0, Status='live', Symbol='RPS20P5', Aliases='RPS20_1_55', description='ribosomal protein S20 pseudogene 5', other_designations='', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=57605793, end_position_on_the_genomic_accession=57606306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270900, CurrentID=0, Status='live', Symbol='RPS20P6', Aliases='RPS20_1_80', description='ribosomal protein S20 pseudogene 6', other_designations='', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101256276, end_position_on_the_genomic_accession=101256623, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873204, CurrentID=0, Status='live', Symbol='MTCYBP3', Aliases='', description='MT-CYB pseudogene 3', other_designations='MTCYB pseudogene 3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95694581, end_position_on_the_genomic_accession=95695702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143425, CurrentID=0, Status='live', Symbol='SYT9', Aliases='', description='synaptotagmin 9', other_designations='synaptotagmin IX|sytIX', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7238796, end_position_on_the_genomic_accession=7469045, orientation='plus', exon_count=14, OMIM=613528, no_hdr0=''),
  101928727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928727, CurrentID=0, Status='live', Symbol='LOC101928727', Aliases='', description='uncharacterized LOC101928727', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175556988, end_position_on_the_genomic_accession=175568248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100270901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270901, CurrentID=0, Status='live', Symbol='RPS20P7', Aliases='RPS20_2_68', description='ribosomal protein S20 pseudogene 7', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81596083, end_position_on_the_genomic_accession=81596606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143458, CurrentID=0, Status='live', Symbol='LDLRAD3', Aliases='LRAD3', description='low density lipoprotein receptor class A domain containing 3', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35943981, end_position_on_the_genomic_accession=36232136, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100272146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272146, CurrentID=0, Status='live', Symbol='NFE2L3P2', Aliases='', description='nuclear factor, erythroid 2 like 3 pseudogene 2', other_designations='nuclear factor (erythroid-derived 2)-like 3 pseudogene 2|nuclear factor, erythroid 2-like 3 pseudogene 2', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47423477, end_position_on_the_genomic_accession=47426692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143471, CurrentID=0, Status='live', Symbol='PSMA8', Aliases='PSMA7L', description='proteasome subunit alpha 8', other_designations='proteasome (prosome, macropain) subunit, alpha type, 8|proteasome subunit alpha type 7-like protein variant 1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26133836, end_position_on_the_genomic_accession=26193355, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376375, CurrentID=0, Status='live', Symbol='LOC105376375', Aliases='', description='uncharacterized LOC105376375', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4706029, end_position_on_the_genomic_accession=4766552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100272147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272147, CurrentID=0, Status='live', Symbol='CMC4', Aliases='C6.1B, MTCP1, MTCP1B, MTCP1NB, p8, p8MTCP1', description='C-x(9)-C motif containing 4', other_designations='C-x(9)-C motif containing 4 homolog|MTCP-1 type A|mature T-cell proliferation 1|mature T-cell proliferation 1 neighbor protein|mature T-cell proliferation-1 type A|protein p8 MTCP-1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155061622, end_position_on_the_genomic_accession=155071272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929143, CurrentID=0, Status='live', Symbol='LOC101929143', Aliases='', description='uncharacterized LOC101929143', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17458613, end_position_on_the_genomic_accession=17459963, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377248, CurrentID=0, Status='live', Symbol='LOC105377248', Aliases='', description='uncharacterized LOC105377248', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=60750622, end_position_on_the_genomic_accession=60795710, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  143496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143496, CurrentID=0, Status='live', Symbol='OR52B4', Aliases='OR11-3', description='olfactory receptor family 52 subfamily B member 4 (gene/pseudogene)', other_designations='olfactory receptor OR11-3|olfactory receptor, family 52, subfamily B, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4367263, end_position_on_the_genomic_accession=4368386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  143501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143501, CurrentID=0, Status='live', Symbol='C11orf40', Aliases='NOV1', description='chromosome 11 open reading frame 40', other_designations='ro/SSA1-related protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4571423, end_position_on_the_genomic_accession=4577820, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  143502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143502, CurrentID=0, Status='live', Symbol='OR52I2', Aliases='OR11-12', description='olfactory receptor family 52 subfamily I member 2', other_designations='olfactory receptor OR11-12|olfactory receptor, family 52, subfamily I, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4586791, end_position_on_the_genomic_accession=4587905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  143503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143503, CurrentID=0, Status='live', Symbol='OR51E1', Aliases='D-GPCR, DGPCR, GPR136, GPR164P, OR52A3P, POGR, PSGR2, OR51E1', description='olfactory receptor family 51 subfamily E member 1', other_designations='Dresden-G-protein-coupled receptor|G-protein coupled receptor 164|olfactory receptor 52A3|olfactory receptor OR11-15|olfactory receptor, family 51, subfamily E, member 1 pseudogene|olfactory receptor, family 52, subfamily A, member 3 pseudogene|prostate overexpressed G protein-coupled receptor|prostate-overexpressed G protein-coupled receptor|prostate-specific G protein-coupled receptor 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4643926, end_position_on_the_genomic_accession=4655488, orientation='plus', exon_count=3, OMIM=611267, no_hdr0=''),
  100270903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270903, CurrentID=0, Status='live', Symbol='RPS29P6', Aliases='RPS29_1_32', description='ribosomal protein S29 pseudogene 6', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37330852, end_position_on_the_genomic_accession=37330979, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143506, CurrentID=0, Status='live', Symbol='SSU72P3', Aliases='', description='SSU72 pseudogene 3', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4329864, end_position_on_the_genomic_accession=4330449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286065, CurrentID=0, Status='live', Symbol='MAPK6PS4', Aliases='Erk3ps4', description='mitogen-activated protein kinase 6 pseudogene 4', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46971328, end_position_on_the_genomic_accession=46975397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131129, CurrentID=0, Status='live', Symbol='LOC100131129', Aliases='', description='uncharacterized LOC100131129', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=84450707, end_position_on_the_genomic_accession=84451144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100270904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270904, CurrentID=0, Status='live', Symbol='RPS3AP11', Aliases='RPS3A_1_40', description='ribosomal protein S3a pseudogene 11', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42491715, end_position_on_the_genomic_accession=42492513, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2078, CurrentID=0, Status='live', Symbol='ERG', Aliases='erg-3, p55', description='v-ets avian erythroblastosis virus E26 oncogene homolog', other_designations='TMPRSS2/ERG fusion|erythroblast transformation-specific transcription factor ERG variant 10|ets-related|transcriptional regulator ERG (transforming protein ERG)|v-ets avian erythroblastosis virus E26 oncogene related|v-ets erythroblastosis virus E26 oncogene homolog|v-ets erythroblastosis virus E26 oncogene like', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38367261, end_position_on_the_genomic_accession=38661833, orientation='minus', exon_count=18, OMIM=165080, no_hdr0=''),
  143543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143543, CurrentID=0, Status='live', Symbol='RBMXP3', Aliases='', description='RNA binding motif protein, X-linked pseudogene 3', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82805187, end_position_on_the_genomic_accession=82807211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286071, CurrentID=0, Status='live', Symbol='LOC286071', Aliases='', description='uncharacterized LOC286071', other_designations='', map_location='8q11.22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270905, CurrentID=0, Status='live', Symbol='RPS3AP12', Aliases='RPS3A_1_92', description='ribosomal protein S3a pseudogene 12', other_designations='', map_location='1p13.3-p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119126495, end_position_on_the_genomic_accession=119127311, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143570, CurrentID=0, Status='live', Symbol='XRRA1', Aliases='', description='X-ray radiation resistance associated 1', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74840910, end_position_on_the_genomic_accession=74949187, orientation='minus', exon_count=20, OMIM=609788, no_hdr0=''),
  441233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441233, CurrentID=0, Status='live', Symbol='LOC441233', Aliases='', description='uncharacterized LOC441233', other_designations='', map_location='7p11.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=57475797, end_position_on_the_genomic_accession=57476359, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131241, CurrentID=0, Status='live', Symbol='LOC100131241', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=71012559, end_position_on_the_genomic_accession=71013589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270906, CurrentID=0, Status='live', Symbol='RPL5P19', Aliases='RPL5_10_724', description='ribosomal protein L5 pseudogene 19', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92013834, end_position_on_the_genomic_accession=92014638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  405753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=405753, CurrentID=0, Status='live', Symbol='DUOXA2', Aliases='SIMNIPHOM, TDH5', description='dual oxidase maturation factor 2', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45114325, end_position_on_the_genomic_accession=45118106, orientation='plus', exon_count=6, OMIM=612772, no_hdr0=''),
  405754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=405754, CurrentID=0, Status='live', Symbol='ERVFRD-1', Aliases='ERVFRDE1, GLLL6191, HERV-FRD, HERV-W/FRD, UNQ6191, envFRD', description='endogenous retrovirus group FRD member 1', other_designations='HERV-FRD provirus ancestral Env polyprotein|HERV-FRD_6p24.1 provirus ancestral Env polyprotein|endogenous retrovirus group FRD, member 1|envelope polyprotein|syncytin 2', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11102489, end_position_on_the_genomic_accession=11111838, orientation='minus', exon_count=2, OMIM=610524, no_hdr0=''),
  100131242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131242, CurrentID=0, Status='live', Symbol='LOC100131242', Aliases='', description='uncharacterized LOC100131242', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=76851342, end_position_on_the_genomic_accession=76851851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101593348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101593348, CurrentID=0, Status='live', Symbol='LOC101593348', Aliases='', description='uncharacterized LOC101593348', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122226442, end_position_on_the_genomic_accession=122228792, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100270907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270907, CurrentID=0, Status='live', Symbol='RPL6P12', Aliases='RPL6_11_535', description='ribosomal protein L6 pseudogene 12', other_designations='', map_location='4q32.3-q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169166254, end_position_on_the_genomic_accession=169167146, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533626, CurrentID=0, Status='live', Symbol='LOC100533626', Aliases='', description='tubulin, alpha 1c pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47074778, end_position_on_the_genomic_accession=47076082, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143630, CurrentID=0, Status='live', Symbol='UBQLNL', Aliases='', description='ubiquilin-like', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5514393, end_position_on_the_genomic_accession=5516726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100533814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533814, CurrentID=0, Status='live', Symbol='LOC100533814', Aliases='', description='TRPM8 channel-associated factor 2 pseudogene', other_designations='family with sequence similarity 115, member C pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143642254, end_position_on_the_genomic_accession=143647649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270908, CurrentID=0, Status='live', Symbol='RPL7P30', Aliases='RPL7_13_818', description='ribosomal protein L7 pseudogene 30', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84528117, end_position_on_the_genomic_accession=84528830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  143662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143662, CurrentID=0, Status='live', Symbol='MUC15', Aliases='MUC-15, PAS3, PASIII', description='mucin 15, cell surface associated', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26559032, end_position_on_the_genomic_accession=26572268, orientation='minus', exon_count=5, OMIM=608566, no_hdr0=''),
  143666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143666, CurrentID=0, Status='live', Symbol='LOC143666', Aliases='', description='uncharacterized LOC143666', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=573808, end_position_on_the_genomic_accession=575885, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  266655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266655, CurrentID=0, Status='live', Symbol='LINC00094', Aliases='LP2477, NCRNA00094, bA374P20.3', description='long intergenic non-protein coding RNA 94', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134025439, end_position_on_the_genomic_accession=134031597, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  143678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143678, CurrentID=0, Status='live', Symbol='C11orf94', Aliases='', description='chromosome 11 open reading frame 94', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45906534, end_position_on_the_genomic_accession=45907282, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  143684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143684, CurrentID=0, Status='live', Symbol='FAM76B', Aliases='', description='family with sequence similarity 76 member B', other_designations='family with sequence similarity 76, member B', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95768942, end_position_on_the_genomic_accession=95790396, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100270909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270909, CurrentID=0, Status='live', Symbol='RPL7P31', Aliases='RPL7_14_803', description='ribosomal protein L7 pseudogene 31', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34711103, end_position_on_the_genomic_accession=34711482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143686, CurrentID=0, Status='live', Symbol='SESN3', Aliases='SEST3', description='sestrin 3', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95165513, end_position_on_the_genomic_accession=95232541, orientation='minus', exon_count=12, OMIM=607768, no_hdr0=''),
  143689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143689, CurrentID=0, Status='live', Symbol='PIWIL4', Aliases='HIWI2, MIWI2', description='piwi-like RNA-mediated gene silencing 4', other_designations='piwi-like 4', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94567308, end_position_on_the_genomic_accession=94621421, orientation='plus', exon_count=20, OMIM=610315, no_hdr0=''),
  101954275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954275, CurrentID=0, Status='live', Symbol='RNU6-7', Aliases='U6-7', description='RNA, U6 small nuclear 7', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32202045, end_position_on_the_genomic_accession=32202151, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100151640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151640, CurrentID=0, Status='live', Symbol='EXOSC3P1', Aliases='', description='exosome component 3 pseudogene 1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32496812, end_position_on_the_genomic_accession=32497568, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151641, CurrentID=0, Status='live', Symbol='PHKG1P4', Aliases='', description='phosphorylase kinase, gamma 1 pseudogene 4', other_designations='phosphorylase kinase, gamma 1 (muscle) pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57060512, end_position_on_the_genomic_accession=57061301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151642, CurrentID=0, Status='live', Symbol='ATG12P1', Aliases='ATG12P', description='autophagy related 12 pseudogene 1', other_designations='ATG12 autophagy related 12 homolog pseudogene 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73067545, end_position_on_the_genomic_accession=73068303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151643, CurrentID=0, Status='live', Symbol='KRTAP20-4', Aliases='KAP20.4', description='keratin associated protein 20-4', other_designations='Putative keratin-associated protein 20-4|keratin associated protein 20-1 pseudogene', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30620627, end_position_on_the_genomic_accession=30620850, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100151644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151644, CurrentID=0, Status='live', Symbol='OTSC8', Aliases='', description='otosclerosis 8', other_designations='', map_location='9p13.1-q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612096, no_hdr0=''),
  100151645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151645, CurrentID=0, Status='live', Symbol='USF1P1', Aliases='USF1P', description='upstream transcription factor 1 pseudogene 1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33334613, end_position_on_the_genomic_accession=33335100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151646, CurrentID=0, Status='live', Symbol='LETM1P3', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 1, pseudogene 3', other_designations='leucine zipper-EF-hand containing transmembrane protein 1 pseudogene', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99695932, end_position_on_the_genomic_accession=99698520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151647, CurrentID=0, Status='live', Symbol='RNU7-51P', Aliases='U7.51', description='RNA, U7 small nuclear 51 pseudogene', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=83048580, end_position_on_the_genomic_accession=83048638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151648, CurrentID=0, Status='live', Symbol='RNU7-52P', Aliases='U7.52', description='RNA, U7 small nuclear 52 pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62383104, end_position_on_the_genomic_accession=62383165, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151649, CurrentID=0, Status='live', Symbol='RNU7-53P', Aliases='U7.53', description='RNA, U7 small nuclear 53 pseudogene', other_designations='', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130386622, end_position_on_the_genomic_accession=130386679, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151650, CurrentID=0, Status='live', Symbol='RNU7-54P', Aliases='U7.54', description='RNA, U7 small nuclear 54 pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128443450, end_position_on_the_genomic_accession=128443510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151651, CurrentID=0, Status='live', Symbol='RNU7-56P', Aliases='U7.56', description='RNA, U7 small nuclear 56 pseudogene', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16767845, end_position_on_the_genomic_accession=16767903, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151652, CurrentID=0, Status='live', Symbol='RNU7-57P', Aliases='U7.57', description='RNA, U7 small nuclear 57 pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154338744, end_position_on_the_genomic_accession=154338802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151653, CurrentID=0, Status='live', Symbol='RNU7-58P', Aliases='U7.58', description='RNA, U7 small nuclear 58 pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59558800, end_position_on_the_genomic_accession=59558855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151654, CurrentID=0, Status='live', Symbol='RNU7-59P', Aliases='U7.59', description='RNA, U7 small nuclear 59 pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77566938, end_position_on_the_genomic_accession=77566995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151655, CurrentID=0, Status='live', Symbol='RNU7-60P', Aliases='U7.60', description='RNA, U7 small nuclear 60 pseudogene', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11546478, end_position_on_the_genomic_accession=11546537, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151656, CurrentID=0, Status='live', Symbol='RNU7-61P', Aliases='U7.61', description='RNA, U7 small nuclear 61 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30483138, end_position_on_the_genomic_accession=30483194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151657, CurrentID=0, Status='live', Symbol='RNU7-62P', Aliases='U7.62', description='RNA, U7 small nuclear 62 pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64384300, end_position_on_the_genomic_accession=64384559, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151658, CurrentID=0, Status='live', Symbol='RNU7-63P', Aliases='U7.63', description='RNA, U7 small nuclear 63 pseudogene', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8705457, end_position_on_the_genomic_accession=8705518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151661, CurrentID=0, Status='live', Symbol='CBX5P1', Aliases='', description='chromobox 5 pseudogene 1', other_designations='chromobox homolog 5 (HP1 alpha homolog, Drosophila) pseudogene|chromobox homolog 5 pseudogene 1', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88095282, end_position_on_the_genomic_accession=88096004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151662, CurrentID=0, Status='live', Symbol='RNU7-64P', Aliases='U7.64', description='RNA, U7 small nuclear 64 pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85441918, end_position_on_the_genomic_accession=85441975, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151663, CurrentID=0, Status='live', Symbol='RNU7-65P', Aliases='U7.65', description='RNA, U7 small nuclear 65 pseudogene', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49344802, end_position_on_the_genomic_accession=49344861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151664, CurrentID=0, Status='live', Symbol='RNU7-66P', Aliases='U7.66', description='RNA, U7 small nuclear 66 pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66728842, end_position_on_the_genomic_accession=66728905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266657, CurrentID=0, Status='live', Symbol='TAS2R15P', Aliases='PS8, T2R15, TAS2R15', description='taste 2 receptor member 15 pseudogene', other_designations='bitter taste receptor pseudogene 8|taste receptor, type 2, member 15 pseudogene|taste receptor, type 2, member 15, pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10964403, end_position_on_the_genomic_accession=10965386, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151666, CurrentID=0, Status='live', Symbol='RNU7-69P', Aliases='U7.69', description='RNA, U7 small nuclear 69 pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123772080, end_position_on_the_genomic_accession=123772137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151667, CurrentID=0, Status='live', Symbol='RNU7-70P', Aliases='U7.70', description='RNA, U7 small nuclear 70 pseudogene', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112634729, end_position_on_the_genomic_accession=112634781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151668, CurrentID=0, Status='live', Symbol='RNU7-71P', Aliases='U7.71', description='RNA, U7 small nuclear 71 pseudogene', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72821935, end_position_on_the_genomic_accession=72821992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151669, CurrentID=0, Status='live', Symbol='RNU7-72P', Aliases='U7.72', description='RNA, U7 small nuclear 72 pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49827996, end_position_on_the_genomic_accession=49828052, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151670, CurrentID=0, Status='live', Symbol='RNU7-73P', Aliases='U7.73', description='RNA, U7 small nuclear 73 pseudogene', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37573159, end_position_on_the_genomic_accession=37573213, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151671, CurrentID=0, Status='live', Symbol='RNU7-74P', Aliases='U7.74', description='RNA, U7 small nuclear 74 pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40377453, end_position_on_the_genomic_accession=40377511, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151672, CurrentID=0, Status='live', Symbol='RNU7-75P', Aliases='U7.75', description='RNA, U7 small nuclear 75 pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37327135, end_position_on_the_genomic_accession=37327195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151673, CurrentID=0, Status='live', Symbol='RNU7-76P', Aliases='U7.76', description='RNA, U7 small nuclear 76 pseudogene', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45975382, end_position_on_the_genomic_accession=45975438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151674, CurrentID=0, Status='live', Symbol='RNU7-77P', Aliases='U7.77', description='RNA, U7 small nuclear 77 pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35135444, end_position_on_the_genomic_accession=35135500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151675, CurrentID=0, Status='live', Symbol='RNU7-78P', Aliases='U7.78', description='RNA, U7 small nuclear 78 pseudogene', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174095243, end_position_on_the_genomic_accession=174095298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151676, CurrentID=0, Status='live', Symbol='RNU7-79P', Aliases='U7.79', description='RNA, U7 small nuclear 79 pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85389830, end_position_on_the_genomic_accession=85389883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151677, CurrentID=0, Status='live', Symbol='RNU7-80P', Aliases='U7.80', description='RNA, U7 small nuclear 80 pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67772600, end_position_on_the_genomic_accession=67772652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151678, CurrentID=0, Status='live', Symbol='RNU7-81P', Aliases='U7.81', description='RNA, U7 small nuclear 81 pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54850290, end_position_on_the_genomic_accession=54850349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151679, CurrentID=0, Status='live', Symbol='RNU7-82P', Aliases='U7.82', description='RNA, U7 small nuclear 82 pseudogene', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=163470206, end_position_on_the_genomic_accession=163470265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151680, CurrentID=0, Status='live', Symbol='RNU7-83P', Aliases='U7.83', description='RNA, U7 small nuclear 83 pseudogene', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108271001, end_position_on_the_genomic_accession=108271060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151681, CurrentID=0, Status='live', Symbol='RNU7-84P', Aliases='U7.84', description='RNA, U7 small nuclear 84 pseudogene', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106729796, end_position_on_the_genomic_accession=106729857, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151682, CurrentID=0, Status='live', Symbol='RNU7-85P', Aliases='U7.85', description='RNA, U7 small nuclear 85 pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79800036, end_position_on_the_genomic_accession=79800294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151683, CurrentID=0, Status='live', Symbol='RNU4ATAC', Aliases='MOPD1, RFMN1, TALS, U4ATAC, RNU4ATAC', description='RNA, U4atac small nuclear (U12-dependent splicing)', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121530880, end_position_on_the_genomic_accession=121531009, orientation='plus', exon_count=1, OMIM=601428, no_hdr0=''),
  100151684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151684, CurrentID=0, Status='live', Symbol='RNU6ATAC', Aliases='RNU6ATAC1, U6ATAC', description='RNA, U6atac small nuclear (U12-dependent splicing)', other_designations='RNA, U6atac small nuclear 1', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134164440, end_position_on_the_genomic_accession=134164564, orientation='minus', exon_count=1, OMIM=601429, no_hdr0=''),
  100151685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151685, CurrentID=0, Status='live', Symbol='RNU6ATAC2P', Aliases='', description='RNA, U6atac small nuclear 2, pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56297565, end_position_on_the_genomic_accession=56297689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151686, CurrentID=0, Status='live', Symbol='RNU6ATAC3P', Aliases='', description='RNA, U6atac small nuclear 3, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46921124, end_position_on_the_genomic_accession=46921248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100151687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151687, CurrentID=0, Status='live', Symbol='RNU6ATAC4P', Aliases='', description='RNA, U6atac small nuclear 4, pseudogene', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36968192, end_position_on_the_genomic_accession=36968316, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151688, CurrentID=0, Status='live', Symbol='RNU6ATAC5P', Aliases='', description='RNA, U6atac small nuclear 5, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73026748, end_position_on_the_genomic_accession=73026872, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373036, CurrentID=0, Status='live', Symbol='LOC105373036', Aliases='', description='uncharacterized LOC105373036', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40260185, end_position_on_the_genomic_accession=40277067, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270912, CurrentID=0, Status='live', Symbol='RPS2P24', Aliases='RPS2_11_579', description='ribosomal protein S2 pseudogene 24', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77545371, end_position_on_the_genomic_accession=77546313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401550, CurrentID=0, Status='live', Symbol='TTLL11-IT1', Aliases='C9orf148', description='TTLL11 intronic transcript 1', other_designations='TTLL11 intronic transcript 1 (non-protein coding)', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=124646890, end_position_on_the_genomic_accession=124725998, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100270913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270913, CurrentID=0, Status='live', Symbol='RPS2P25', Aliases='RPS2_12_592', description='ribosomal protein S2 pseudogene 25', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85762499, end_position_on_the_genomic_accession=85763266, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645529, CurrentID=0, Status='live', Symbol='LOC645529', Aliases='', description='acrosin pseudogene', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113667387, end_position_on_the_genomic_accession=113668627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929431, CurrentID=0, Status='live', Symbol='LOC101929431', Aliases='', description='uncharacterized LOC101929431', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32434354, end_position_on_the_genomic_accession=32446056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401960, CurrentID=0, Status='live', Symbol='GAPDHP27', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 27', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119558552, end_position_on_the_genomic_accession=119559845, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370267, CurrentID=0, Status='live', Symbol='LOC105370267', Aliases='', description='uncharacterized LOC105370267', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76668059, end_position_on_the_genomic_accession=76689510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  148818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148818, CurrentID=0, Status='live', Symbol='HSD17B7P1', Aliases='', description='hydroxysteroid (17-beta) dehydrogenase 7 pseudogene 1', other_designations='hydroxysteroid dehydrogenase 17 beta type 7 pseudogene 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247827464, end_position_on_the_genomic_accession=247828745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100272205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272205, CurrentID=0, Status='live', Symbol='TATDN2P2', Aliases='', description='TatD DNase domain containing 2 pseudogene 2', other_designations='', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158609328, end_position_on_the_genomic_accession=158611712, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270914, CurrentID=0, Status='live', Symbol='RPS7P12', Aliases='RPS7_6_1816', description='ribosomal protein S7 pseudogene 12', other_designations='', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134650783, end_position_on_the_genomic_accession=134651448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441242, CurrentID=0, Status='live', Symbol='LOC441242', Aliases='', description='uncharacterized LOC441242', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65647010, end_position_on_the_genomic_accession=65770810, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106050037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106050037, CurrentID=0, Status='live', Symbol='LOC106050037', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B9 pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 9, 22kDa pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372886, CurrentID=0, Status='live', Symbol='LOC105372886', Aliases='', description='uncharacterized LOC105372886', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208112152, end_position_on_the_genomic_accession=208121442, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  143872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143872, CurrentID=0, Status='live', Symbol='ARHGAP42', Aliases='GRAF3', description='Rho GTPase activating protein 42', other_designations='rho-type GTPase-activating protein 42', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100687676, end_position_on_the_genomic_accession=100990925, orientation='plus', exon_count=27, OMIM=615936, no_hdr0=''),
  143879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143879, CurrentID=0, Status='live', Symbol='KBTBD3', Aliases='BKLHD3', description='kelch repeat and BTB (POZ) domain containing 3', other_designations='BTB and kelch domain containing 3|BTB and kelch domain-containing protein 3', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106046806, end_position_on_the_genomic_accession=106077765, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  143884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143884, CurrentID=0, Status='live', Symbol='CWF19L2', Aliases='', description='CWF19-like 2, cell cycle control (S. pombe)', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107326345, end_position_on_the_genomic_accession=107457846, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  143888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143888, CurrentID=0, Status='live', Symbol='KDELC2', Aliases='', description='KDEL motif containing 2', other_designations='KDEL (Lys-Asp-Glu-Leu) containing 2', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108472105, end_position_on_the_genomic_accession=108498519, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100272217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272217, CurrentID=0, Status='live', Symbol='LOC100272217', Aliases='', description='uncharacterized LOC100272217', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130577350, end_position_on_the_genomic_accession=130579494, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100270916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270916, CurrentID=0, Status='live', Symbol='RPSAP35', Aliases='RPSA_13_513', description='ribosomal protein SA pseudogene 35', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117413462, end_position_on_the_genomic_accession=117414281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  143903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143903, CurrentID=0, Status='live', Symbol='LAYN', Aliases='', description='layilin', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111538985, end_position_on_the_genomic_accession=111561745, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100272220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272220, CurrentID=0, Status='live', Symbol='RPAP2P1', Aliases='', description='RNA polymerase II associated protein 2 pseudogene 1', other_designations='', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86854695, end_position_on_the_genomic_accession=86856408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131252, CurrentID=0, Status='live', Symbol='LOC100131252', Aliases='', description='family with sequence similarity 213, member A pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73277362, end_position_on_the_genomic_accession=73278046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376378, CurrentID=0, Status='live', Symbol='LOC105376378', Aliases='', description='uncharacterized LOC105376378', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5429190, end_position_on_the_genomic_accession=5446520, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270917, CurrentID=0, Status='live', Symbol='RPSAP40', Aliases='RPSA_13_766', description='ribosomal protein SA pseudogene 40', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148647624, end_position_on_the_genomic_accession=148648671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286133, CurrentID=0, Status='live', Symbol='SCARA5', Aliases='NET33, Tesr', description='scavenger receptor class A member 5', other_designations='scavenger receptor class A, member 5 (putative)|scavenger receptor hlg|testis expressed scavenger receptor', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27869882, end_position_on_the_genomic_accession=27992852, orientation='minus', exon_count=9, OMIM=611306, no_hdr0=''),
  143941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=143941, CurrentID=0, Status='live', Symbol='TTC36', Aliases='HBP21', description='tetratricopeptide repeat domain 36', other_designations='HSP70 binding protein 21|HSP70-binding protein 21|TPR repeat protein 36', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118527495, end_position_on_the_genomic_accession=118531025, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100272225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100272225, CurrentID=0, Status='live', Symbol='UAQTL4', Aliases='GOUT4', description='Uric acid concentration, serum, QTL4', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612671, no_hdr0=''),
  100270918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270918, CurrentID=0, Status='live', Symbol='RPSAP41', Aliases='RPSA_17_711', description='ribosomal protein SA pseudogene 41', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73290550, end_position_on_the_genomic_accession=73291335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266665, CurrentID=0, Status='live', Symbol='TAS2R33', Aliases='T2R33', description='taste 2 receptor member 33', other_designations='taste receptor, type 2, member 33', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131254, CurrentID=0, Status='live', Symbol='LOC100131254', Aliases='', description='ESF1, nucleolar pre-rRNA processing protein, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53971598, end_position_on_the_genomic_accession=53973892, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131132, CurrentID=0, Status='live', Symbol='LOC100131132', Aliases='', description='uncharacterized LOC100131132', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77866902, end_position_on_the_genomic_accession=77871535, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  266666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266666, CurrentID=0, Status='live', Symbol='TAS2R36', Aliases='T2R36', description='taste 2 receptor member 36', other_designations='taste receptor, type 2, member 36', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372455, CurrentID=0, Status='live', Symbol='LOC105372455', Aliases='', description='uncharacterized LOC105372455', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53739000, end_position_on_the_genomic_accession=53746389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100270796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270796, CurrentID=0, Status='live', Symbol='STQTL13', Aliases='', description='Stature QTL 13', other_designations='', map_location='4p15.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612226, no_hdr0=''),
  266667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266667, CurrentID=0, Status='live', Symbol='TAS2R37', Aliases='T2R37', description='taste 2 receptor member 37', other_designations='taste receptor, type 2, member 37', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286149, CurrentID=0, Status='live', Symbol='LOC286149', Aliases='', description='uncharacterized LOC286149', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=96079036, end_position_on_the_genomic_accession=96085843, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419870, CurrentID=0, Status='live', Symbol='LOC100419870', Aliases='', description='zinc finger protein 717 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130283229, end_position_on_the_genomic_accession=130284245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288379, CurrentID=0, Status='live', Symbol='LOC100288379', Aliases='', description='ribosomal protein S7 pseudogene', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1673283, end_position_on_the_genomic_accession=1674435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270921, CurrentID=0, Status='live', Symbol='RPL12P21', Aliases='RPL12_3_619', description='ribosomal protein L12 pseudogene 21', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139035107, end_position_on_the_genomic_accession=139035645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379797, CurrentID=0, Status='live', Symbol='LOC105379797', Aliases='', description='uncharacterized LOC105379797', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729307, CurrentID=0, Status='live', Symbol='LOC729307', Aliases='', description='uncharacterized LOC729307', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137193756, end_position_on_the_genomic_accession=137198367, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375506, CurrentID=0, Status='live', Symbol='LOC105375506', Aliases='', description='uncharacterized LOC105375506', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130660023, end_position_on_the_genomic_accession=130662627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131257, CurrentID=0, Status='live', Symbol='LOC100131257', Aliases='', description='zinc finger protein 655 pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7075770, end_position_on_the_genomic_accession=7096786, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374441, CurrentID=0, Status='live', Symbol='LOC105374441', Aliases='', description='uncharacterized LOC105374441', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44897079, end_position_on_the_genomic_accession=44924587, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  406214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406214, CurrentID=0, Status='live', Symbol='LRSL', Aliases='', description='Larsen-like syndrome', other_designations='', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608545, no_hdr0=''),
  406215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406215, CurrentID=0, Status='live', Symbol='KTCN3', Aliases='', description='Keratoconus 3', other_designations='', map_location='3p14-q13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608586, no_hdr0=''),
  406216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406216, CurrentID=0, Status='live', Symbol='ATFB1', Aliases='', description='Atrial fibrillation, familial 1', other_designations='', map_location='10q22-q24', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608583, no_hdr0=''),
  105376379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376379, CurrentID=0, Status='live', Symbol='LOC105376379', Aliases='', description='uncharacterized LOC105376379', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5483489, end_position_on_the_genomic_accession=5493864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270922, CurrentID=0, Status='live', Symbol='RPL12P22', Aliases='RPL12_4_640', description='ribosomal protein L12 pseudogene 22', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174039751, end_position_on_the_genomic_accession=174040222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  406237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406237, CurrentID=0, Status='live', Symbol='PDON2', Aliases='LAP, PDLA', description='Periodontitis, localized aggressive', other_designations='', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608526, no_hdr0=''),
  406238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406238, CurrentID=0, Status='live', Symbol='ALS7', Aliases='', description='amyotrophic lateral sclerosis 7', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608031, no_hdr0=''),
  144097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144097, CurrentID=0, Status='live', Symbol='C11orf84', Aliases='', description='chromosome 11 open reading frame 84', other_designations='hypothetical protein BC007540', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63813374, end_position_on_the_genomic_accession=63827718, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  144100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144100, CurrentID=0, Status='live', Symbol='PLEKHA7', Aliases='', description='pleckstrin homology domain containing A7', other_designations='PH domain-containing family A member 7|pleckstrin homology domain containing, family A member 2|pleckstrin homology domain containing, family A member 7', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16777297, end_position_on_the_genomic_accession=17014417, orientation='minus', exon_count=35, OMIM=612686, no_hdr0=''),
  144106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144106, CurrentID=0, Status='live', Symbol='ST13P5', Aliases='FAM10A5', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 5', other_designations='family with sequence similarity 10, member A5 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18261885, end_position_on_the_genomic_accession=18263501, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144108, CurrentID=0, Status='live', Symbol='SPTY2D1', Aliases='', description='SPT2 chromatin protein domain containing 1', other_designations='SPT2 domain-containing protein 1|SPT2, Suppressor of Ty, domain containing 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18606401, end_position_on_the_genomic_accession=18634786, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  144110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144110, CurrentID=0, Status='live', Symbol='TMEM86A', Aliases='', description='transmembrane protein 86A', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18698735, end_position_on_the_genomic_accession=18704785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  144124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144124, CurrentID=0, Status='live', Symbol='OR10A5', Aliases='JCG6, OR10A1, OR11-403', description='olfactory receptor family 10 subfamily A member 5', other_designations='HP3|olfactory receptor 10A1|olfactory receptor 11-403|olfactory receptor OR11-85|olfactory receptor, family 10, subfamily A, member 1|olfactory receptor, family 10, subfamily A, member 5|olfactory receptor-like protein JCG6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6845683, end_position_on_the_genomic_accession=6846636, orientation='plus', exon_count=1, OMIM=608493, no_hdr0=''),
  144125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144125, CurrentID=0, Status='live', Symbol='OR2AG1', Aliases='OR11-79, OR2AG3', description='olfactory receptor family 2 subfamily AG member 1 (gene/pseudogene)', other_designations='hT3 olfactory receptor|olfactory receptor 2AG3|olfactory receptor OR11-79|olfactory receptor, family 2, subfamily AG, member 1|olfactory receptor, family 2, subfamily AG, member 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6785017, end_position_on_the_genomic_accession=6786014, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  144132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144132, CurrentID=0, Status='live', Symbol='DNHD1', Aliases='C11orf47, CCDC35, DHCD1L, DNHD1', description='dynein heavy chain domain 1', other_designations='DNHD1 variant protein|coiled-coil domain containing 35|dynein heavy chain domain 1-like protein', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6497296, end_position_on_the_genomic_accession=6572025, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  144165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144165, CurrentID=0, Status='live', Symbol='PRICKLE1', Aliases='EPM1B, RILP', description='prickle homolog 1', other_designations='REST (RE-1 silencing transcription factor)/NRSF (neuron-restrictive silencer factor)-interacting LIM domain protein|REST/NRSF interacting LIM domain protein|REST/NRSF-interacting LIM domain protein 1|prickle-like 1', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42458338, end_position_on_the_genomic_accession=42589770, orientation='minus', exon_count=11, OMIM=608500, no_hdr0=''),
  100270925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270925, CurrentID=0, Status='live', Symbol='RPL12P23', Aliases='RPL12_7_777', description='ribosomal protein L12 pseudogene 23', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170199790, end_position_on_the_genomic_accession=170200235, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144193, CurrentID=0, Status='live', Symbol='AMDHD1', Aliases='HMFT1272', description='amidohydrolase domain containing 1', other_designations='amidohydrolase domain-containing protein 1', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95943293, end_position_on_the_genomic_accession=95968592, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  144195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144195, CurrentID=0, Status='live', Symbol='SLC2A14', Aliases='GLUT14, SLC2A3P3', description='solute carrier family 2 (facilitated glucose transporter), member 14', other_designations='glucose transporter type 14|solute carrier family 2 (facilitated glucose transporter), member 3 pseudogene 3', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7812512, end_position_on_the_genomic_accession=7891196, orientation='minus', exon_count=20, OMIM=611039, no_hdr0=''),
  100270926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270926, CurrentID=0, Status='live', Symbol='RPL12P25', Aliases='RPL12_9_954', description='ribosomal protein L12 pseudogene 25', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=477778, end_position_on_the_genomic_accession=478396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372814, CurrentID=0, Status='live', Symbol='LOC105372814', Aliases='', description='uncharacterized LOC105372814', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42209428, end_position_on_the_genomic_accession=42222893, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144203, CurrentID=0, Status='live', Symbol='OVOS2', Aliases='', description='ovostatin 2', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31111657, end_position_on_the_genomic_accession=31232532, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  100131262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131262, CurrentID=0, Status='live', Symbol='LOC100131262', Aliases='', description='uncharacterized LOC100131262', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270927, CurrentID=0, Status='live', Symbol='RPL13P9', Aliases='RPL13_1_589', description='ribosomal protein L13 pseudogene 9', other_designations='', map_location='5q13-q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83201229, end_position_on_the_genomic_accession=83201761, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144233, CurrentID=0, Status='live', Symbol='BCDIN3D', Aliases='', description='BCDIN3 domain containing RNA methyltransfease', other_designations='BCDIN3 domain-containing protein|probable methyltransferase BCDIN3D', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49836043, end_position_on_the_genomic_accession=49843129, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729313, CurrentID=0, Status='live', Symbol='LOC729313', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81475499, end_position_on_the_genomic_accession=81476566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144245, CurrentID=0, Status='live', Symbol='ALG10B', Aliases='ALG10, KCR1', description='ALG10B, alpha-1,2-glucosyltransferase', other_designations='alpha-1,2-glucosyltransferase ALG10-A|alpha-2-glucosyltransferase ALG10-B|asparagine-linked glycosylation 10 homolog B (yeast, alpha-1,2-glucosyltransferase)|asparagine-linked glycosylation 10, alpha-1,2-glucosyltransferase homolog B|asparagine-linked glycosylation protein 10 homolog B|dolichyl-P-Glc:Glc(2)Man(9)GlcNAc(2)-PP-dolichol alpha-1,2- glucosyltransferase|modifier of the HERG potassium channel|potassium channel regulator 1|putative alpha-1,2-glucosyltransferase ALG10-B', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38316686, end_position_on_the_genomic_accession=38329726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100270928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270928, CurrentID=0, Status='live', Symbol='RPL13P8', Aliases='RPL13_2_391', description='ribosomal protein L13 pseudogene 8', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108543374, end_position_on_the_genomic_accession=108543875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401975, CurrentID=0, Status='live', Symbol='RPS3AP10', Aliases='RPS3A_2_114', description='ribosomal protein S3a pseudogene 10', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166022189, end_position_on_the_genomic_accession=166023075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131264, CurrentID=0, Status='live', Symbol='LOC100131264', Aliases='', description='prefoldin subunit 4 pseudogene', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2835621, end_position_on_the_genomic_accession=2836022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144321, CurrentID=0, Status='live', Symbol='GLIPR1L2', Aliases='', description='GLI pathogenesis-related 1 like 2', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75391070, end_position_on_the_genomic_accession=75432688, orientation='plus', exon_count=8, OMIM=610394, no_hdr0=''),
  144347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144347, CurrentID=0, Status='live', Symbol='FAM101A', Aliases='CFM2', description='family with sequence similarity 101 member A', other_designations='family with sequence similarity 101, member A|protein FAM101A|refilinA|regulator of filamin protein A', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124289164, end_position_on_the_genomic_accession=124316024, orientation='plus', exon_count=3, OMIM=615927, no_hdr0=''),
  144348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144348, CurrentID=0, Status='live', Symbol='ZNF664', Aliases='ZFOC1, ZNF176', description='zinc finger protein 664', other_designations='zinc finger Organ of Corti 1|zinc finger protein 176|zinc finger protein from organ of Corti', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123973215, end_position_on_the_genomic_accession=124015439, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  144360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144360, CurrentID=0, Status='live', Symbol='LINC00477', Aliases='C12orf67, FAM191B', description='long intergenic non-protein coding RNA 477', other_designations='family with sequence similarity 191, member B', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24566964, end_position_on_the_genomic_accession=24584168, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  144363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144363, CurrentID=0, Status='live', Symbol='LYRM5', Aliases='', description='LYR motif containing 5', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25195216, end_position_on_the_genomic_accession=25205015, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376381, CurrentID=0, Status='live', Symbol='LOC105376381', Aliases='', description='uncharacterized LOC105376381', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5564942, end_position_on_the_genomic_accession=5566713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144383, CurrentID=0, Status='live', Symbol='DSPA2D', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31754629, end_position_on_the_genomic_accession=31755283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144402, CurrentID=0, Status='live', Symbol='CPNE8', Aliases='', description='copine VIII', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38652200, end_position_on_the_genomic_accession=38906135, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  105369398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369398, CurrentID=0, Status='live', Symbol='LOC105369398', Aliases='', description='uncharacterized LOC105369398', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76812141, end_position_on_the_genomic_accession=76819752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144404, CurrentID=0, Status='live', Symbol='TMEM120B', Aliases='', description='transmembrane protein 120B', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121712752, end_position_on_the_genomic_accession=121782924, orientation='plus', exon_count=13, OMIM=616551, no_hdr0=''),
  144406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144406, CurrentID=0, Status='live', Symbol='WDR66', Aliases='CaM-IP4', description='WD repeat domain 66', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121918557, end_position_on_the_genomic_accession=122003927, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  144423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144423, CurrentID=0, Status='live', Symbol='GLT1D1', Aliases='', description='glycosyltransferase 1 domain containing 1', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128853440, end_position_on_the_genomic_accession=128984968, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105373058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373058, CurrentID=0, Status='live', Symbol='LOC105373058', Aliases='', description='uncharacterized LOC105373058', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43818559, end_position_on_the_genomic_accession=43824421, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371144, CurrentID=0, Status='live', Symbol='LOC105371144', Aliases='', description='uncharacterized LOC105371144', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25075795, end_position_on_the_genomic_accession=25078786, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  144448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144448, CurrentID=0, Status='live', Symbol='TSPAN19', Aliases='', description='tetraspanin 19', other_designations='tspan-19', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85014311, end_position_on_the_genomic_accession=85036278, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  144453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144453, CurrentID=0, Status='live', Symbol='BEST3', Aliases='VMD2L3', description='bestrophin 3', other_designations='vitelliform macular dystrophy 2-like 3|vitelliform macular dystrophy 2-like protein 3', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69643508, end_position_on_the_genomic_accession=69699416, orientation='minus', exon_count=16, OMIM=607337, no_hdr0=''),
  144455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144455, CurrentID=0, Status='live', Symbol='E2F7', Aliases='', description='E2F transcription factor 7', other_designations='E2F-7', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77021246, end_position_on_the_genomic_accession=77065580, orientation='minus', exon_count=14, OMIM=612046, no_hdr0=''),
  105372890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372890, CurrentID=0, Status='live', Symbol='LOC105372890', Aliases='', description='uncharacterized LOC105372890', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208266202, end_position_on_the_genomic_accession=208270568, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929453, CurrentID=0, Status='live', Symbol='LOC101929453', Aliases='', description='uncharacterized LOC101929453', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=103146806, end_position_on_the_genomic_accession=103152394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  144481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144481, CurrentID=0, Status='live', Symbol='SOCS2-AS1', Aliases='', description='SOCS2 antisense RNA 1', other_designations='SOCS2 antisense RNA 1 (non-protein coding)', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93565628, end_position_on_the_genomic_accession=93571398, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  144486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144486, CurrentID=0, Status='live', Symbol='CEP83-AS1', Aliases='CCDC41-AS1', description='CEP83 antisense RNA 1 (head to head)', other_designations='CCDC41 antisense RNA 1 (head to head)', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94460003, end_position_on_the_genomic_accession=94462570, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2236, CurrentID=0, Status='live', Symbol='FECHP1', Aliases='FECHP', description='ferrochelatase pseudogene 1', other_designations='ferrochelatase (protoporphyria) pseudogene', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34872724, end_position_on_the_genomic_accession=34875218, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100291628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100291628, CurrentID=0, Status='live', Symbol='LOC100291628', Aliases='', description='penta-EF-hand domain containing 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204528093, end_position_on_the_genomic_accession=204529727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270936, CurrentID=0, Status='live', Symbol='RPL17P10', Aliases='RPL17_2_309', description='ribosomal protein L17 pseudogene 10', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201102493, end_position_on_the_genomic_accession=201102990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144501, CurrentID=0, Status='live', Symbol='KRT80', Aliases='KB20', description='keratin 80, type II', other_designations='CK-80|K80|cytokeratin-80|keratin b20|keratin-80|type II keratin|type-II keratin Kb20', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52168996, end_position_on_the_genomic_accession=52192000, orientation='minus', exon_count=9, OMIM=611161, no_hdr0=''),
  79523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79523, CurrentID=0, Status='live', Symbol='OR5M2P', Aliases='OR11-193', description='olfactory receptor family 5 subfamily M member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56479491, end_position_on_the_genomic_accession=56480426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266683, CurrentID=0, Status='live', Symbol='LOC266683', Aliases='', description='eukaryotic translation initiation factor 3 subunit M pseudogene', other_designations='dendritic cell protein pseudogene|eukaryotic translation initiation factor 3, subunit M pseudogene', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=82875623, end_position_on_the_genomic_accession=82875974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376624, CurrentID=0, Status='live', Symbol='LOC105376624', Aliases='', description='uncharacterized LOC105376624', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34853602, end_position_on_the_genomic_accession=34875283, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100270937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270937, CurrentID=0, Status='live', Symbol='RPL17P11', Aliases='RPL17_3_336', description='ribosomal protein L17 pseudogene 11', other_designations='', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233721492, end_position_on_the_genomic_accession=233722110, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144535, CurrentID=0, Status='live', Symbol='CFAP54', Aliases='C12orf55, C12orf63', description='cilia and flagella associated 54', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96489546, end_position_on_the_genomic_accession=96875555, orientation='plus', exon_count=76, OMIM='', no_hdr0=''),
  100270938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270938, CurrentID=0, Status='live', Symbol='RPL17P12', Aliases='RPL17_6_281', description='ribosomal protein L17 pseudogene 12', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146194296, end_position_on_the_genomic_accession=146194821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144568, CurrentID=0, Status='live', Symbol='A2ML1', Aliases='CPAMD9', description='alpha-2-macroglobulin-like 1', other_designations='C3 and PZP-like, alpha-2-macroglobulin domain containing 9', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8822554, end_position_on_the_genomic_accession=8886918, orientation='plus', exon_count=40, OMIM=610627, no_hdr0=''),
  144571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144571, CurrentID=0, Status='live', Symbol='A2M-AS1', Aliases='', description='A2M antisense RNA 1 (head to head)', other_designations='A2M antisense RNA 1 (non-protein coding)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9065177, end_position_on_the_genomic_accession=9068055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  144577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144577, CurrentID=0, Status='live', Symbol='C12orf66', Aliases='', description='chromosome 12 open reading frame 66', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64186312, end_position_on_the_genomic_accession=64237341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373063, CurrentID=0, Status='live', Symbol='LOC105373063', Aliases='', description='uncharacterized LOC105373063', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44949344, end_position_on_the_genomic_accession=44974146, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144581, CurrentID=0, Status='live', Symbol='RPL14P1', Aliases='RPL14L, RPL14_1_1246, bcm1298', description='ribosomal protein L14 pseudogene 1', other_designations='60S ribosomal protein L14', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62965302, end_position_on_the_genomic_accession=62966024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270939, CurrentID=0, Status='live', Symbol='RPL17P13', Aliases='RPL17_7_285', description='ribosomal protein L17 pseudogene 13', other_designations='', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149439897, end_position_on_the_genomic_accession=149440427, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873243, CurrentID=0, Status='live', Symbol='MTND4P21', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 21', other_designations='MT-ND4 pseudogene 21', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131372309, end_position_on_the_genomic_accession=131373650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144608, CurrentID=0, Status='live', Symbol='C12orf60', Aliases='', description='chromosome 12 open reading frame 60', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14785217, end_position_on_the_genomic_accession=14823858, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100270940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270940, CurrentID=0, Status='live', Symbol='RPL17P28', Aliases='RPL17_7_858', description='ribosomal protein L17 pseudogene 28', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139049456, end_position_on_the_genomic_accession=139050006, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270800, CurrentID=0, Status='live', Symbol='SHFLD3', Aliases='C17DUPp13.3, DUP17p13.3', description='Split-hand/foot malformation with long bone deficiency 3', other_designations='', map_location='17p13.3-p13.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612576, no_hdr0=''),
  100873244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873244, CurrentID=0, Status='live', Symbol='MTND4P22', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 22', other_designations='MT-ND4 pseudogene 22', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143096762, end_position_on_the_genomic_accession=143098125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060173, CurrentID=0, Status='live', Symbol='LOC101060173', Aliases='', description='spidroin-2-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  133332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133332, CurrentID=0, Status='live', Symbol='LOC133332', Aliases='', description='mitochondrial ribosomal protein S5 pseudogene', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163265427, end_position_on_the_genomic_accession=163266583, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144638, CurrentID=0, Status='live', Symbol='ST13P3', Aliases='FAM10A3, FAM10A3P', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 3', other_designations='family with sequence similarity 10, member A3', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106015813, end_position_on_the_genomic_accession=106018409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270941, CurrentID=0, Status='live', Symbol='RPL17P14', Aliases='RPL17_8_331', description='ribosomal protein L17 pseudogene 14', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229003886, end_position_on_the_genomic_accession=229004382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929459, CurrentID=0, Status='live', Symbol='LOC101929459', Aliases='', description='uncharacterized LOC101929459', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205233777, end_position_on_the_genomic_accession=205237316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  155184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155184, CurrentID=0, Status='live', Symbol='SLC2A7', Aliases='GLUT7', description='solute carrier family 2 (facilitated glucose transporter), member 7', other_designations='GLUT-7|glucose transporter type 7|intestinal facilitative glucose transporter 7', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8960853, end_position_on_the_genomic_accession=9026384, orientation='minus', exon_count=15, OMIM=610371, no_hdr0=''),
  100270942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270942, CurrentID=0, Status='live', Symbol='RPL17P32', Aliases='RPL17_8_939', description='ribosomal protein L17 pseudogene 32', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105287841, end_position_on_the_genomic_accession=105288365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286257, CurrentID=0, Status='live', Symbol='C9orf142', Aliases='PAXX', description='chromosome 9 open reading frame 142', other_designations='paralog of XRCC4 and XLF', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136992418, end_position_on_the_genomic_accession=136993976, orientation='plus', exon_count=7, OMIM=616315, no_hdr0=''),
  105377431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377431, CurrentID=0, Status='live', Symbol='LOC105377431', Aliases='', description='uncharacterized LOC105377431', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=133012761, end_position_on_the_genomic_accession=133029697, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  144699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144699, CurrentID=0, Status='live', Symbol='FBXL14', Aliases='Fbl14', description='F-box and leucine-rich repeat protein 14', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1565993, end_position_on_the_genomic_accession=1594180, orientation='minus', exon_count=4, OMIM=609081, no_hdr0=''),
  105369400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369400, CurrentID=0, Status='live', Symbol='LOC105369400', Aliases='', description='uncharacterized LOC105369400', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77058401, end_position_on_the_genomic_accession=77063723, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270943, CurrentID=0, Status='live', Symbol='RPL17P17', Aliases='RPL17_9_379', description='ribosomal protein L17 pseudogene 17', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65843862, end_position_on_the_genomic_accession=65844474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144712, CurrentID=0, Status='live', Symbol='HMGA1P3', Aliases='HMGA1L3, HMGIYL3', description='high mobility group AT-hook 1 pseudogene 3', other_designations='HMGA1L3 retropseudogene|high mobility group AT-hook 1-like 3', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110428598, end_position_on_the_genomic_accession=110430393, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144715, CurrentID=0, Status='live', Symbol='RAD9B', Aliases='', description='RAD9 checkpoint clamp component B', other_designations='DNA repair exonuclease rad9 homolog B|RAD9 homolog B', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110502200, end_position_on_the_genomic_accession=110532086, orientation='plus', exon_count=13, OMIM=608368, no_hdr0=''),
  144717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144717, CurrentID=0, Status='live', Symbol='FAM109A', Aliases='IPIP27A, SES1', description='family with sequence similarity 109 member A', other_designations='27 kDa inositol polyphosphate phosphatase-interacting protein A|family with sequence similarity 109, member A|protein FAM109A', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111360651, end_position_on_the_genomic_accession=111369121, orientation='minus', exon_count=6, OMIM=614239, no_hdr0=''),
  100131279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131279, CurrentID=0, Status='live', Symbol='C2orf69P3', Aliases='', description='chromosome 2 open reading frame 69 pseudogene 3', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35425227, end_position_on_the_genomic_accession=35427625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  406874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406874, CurrentID=0, Status='live', Symbol='DYX8', Aliases='', description='dyslexia susceptibility 8', other_designations='', map_location='1p36-p34', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608995, no_hdr0=''),
  406875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406875, CurrentID=0, Status='live', Symbol='FXYD6P3', Aliases='FXYD8', description='FXYD domain containing ion transport regulator 6 pseudogene 3', other_designations='FXYD domain containing ion transport regulator 8|FXYD domain-containing ion transport regulator 6 pseudogene 3|FXYD domain-containing ion transport regulator 6, pseudogene 3', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73875509, end_position_on_the_genomic_accession=73876960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  406881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406881, CurrentID=0, Status='live', Symbol='MIRLET7A1', Aliases='LET7A1, MIRNLET7A1, let-7a-1', description='microRNA let-7a-1', other_designations='hsa-let-7a-1', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94175957, end_position_on_the_genomic_accession=94176036, orientation='plus', exon_count=1, OMIM=605386, no_hdr0=''),
  406882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406882, CurrentID=0, Status='live', Symbol='MIRLET7A2', Aliases='LET7A2, MIRNLET7A2, let-7a-2', description='microRNA let-7a-2', other_designations='hsa-let-7a-2', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122146522, end_position_on_the_genomic_accession=122146593, orientation='minus', exon_count=1, OMIM=612142, no_hdr0=''),
  406883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406883, CurrentID=0, Status='live', Symbol='MIRLET7A3', Aliases='LET7A3, MIRNLET7A3, let-7a-3', description='microRNA let-7a-3', other_designations='hsa-let-7a-3', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46112749, end_position_on_the_genomic_accession=46112822, orientation='plus', exon_count=1, OMIM=612143, no_hdr0=''),
  406884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406884, CurrentID=0, Status='live', Symbol='MIRLET7B', Aliases='LET7B, MIRNLET7B, hsa-let-7b, let-7b', description='microRNA let-7b', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46113686, end_position_on_the_genomic_accession=46113768, orientation='plus', exon_count=1, OMIM=611249, no_hdr0=''),
  406885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406885, CurrentID=0, Status='live', Symbol='MIRLET7C', Aliases='LET7C, MIRNLET7C, hsa-let-7c, let-7c', description='microRNA let-7c', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16539828, end_position_on_the_genomic_accession=16539911, orientation='plus', exon_count=1, OMIM=612144, no_hdr0=''),
  406886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406886, CurrentID=0, Status='live', Symbol='MIRLET7D', Aliases='LET7D, MIRNLET7D, hsa-let-7d, let-7d', description='microRNA let-7d', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94178834, end_position_on_the_genomic_accession=94178920, orientation='plus', exon_count=1, OMIM=612145, no_hdr0=''),
  406887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406887, CurrentID=0, Status='live', Symbol='MIRLET7E', Aliases='LET7E, MIRNLET7E, hsa-let-7e, let-7e', description='microRNA let-7e', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51692786, end_position_on_the_genomic_accession=51692864, orientation='plus', exon_count=1, OMIM=611250, no_hdr0=''),
  406888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406888, CurrentID=0, Status='live', Symbol='MIRLET7F1', Aliases='LET7F1, MIRNLET7F1, let-7f-1', description='microRNA let-7f-1', other_designations='hsa-let-7f-1', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94176347, end_position_on_the_genomic_accession=94176433, orientation='plus', exon_count=1, OMIM=612146, no_hdr0=''),
  406889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406889, CurrentID=0, Status='live', Symbol='MIRLET7F2', Aliases='LET7F2, MIRNLET7F2, let-7f-2', description='microRNA let-7f-2', other_designations='hsa-let-7f-2', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53557192, end_position_on_the_genomic_accession=53557274, orientation='minus', exon_count=1, OMIM=300721, no_hdr0=''),
  406890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406890, CurrentID=0, Status='live', Symbol='MIRLET7G', Aliases='LET7G, MIRNLET7G, hsa-let-7g, let-7g', description='microRNA let-7g', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52268278, end_position_on_the_genomic_accession=52268361, orientation='minus', exon_count=1, OMIM=612102, no_hdr0=''),
  406891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406891, CurrentID=0, Status='live', Symbol='MIRLET7I', Aliases='LET7I, MIRNLET7I, hsa-let-7i, let-7i', description='microRNA let-7i', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62603686, end_position_on_the_genomic_accession=62603769, orientation='plus', exon_count=1, OMIM=612148, no_hdr0=''),
  406892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406892, CurrentID=0, Status='live', Symbol='MIR100', Aliases='MIRN100, miR-100', description='microRNA 100', other_designations='hsa-mir-100', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122152229, end_position_on_the_genomic_accession=122152308, orientation='minus', exon_count=1, OMIM=613186, no_hdr0=''),
  406893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406893, CurrentID=0, Status='live', Symbol='MIR101-1', Aliases='MIRN101-1, mir-101-1', description='microRNA 101-1', other_designations='hsa-mir-101-1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65058434, end_position_on_the_genomic_accession=65058508, orientation='minus', exon_count=1, OMIM=612511, no_hdr0=''),
  406894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406894, CurrentID=0, Status='live', Symbol='MIR101-2', Aliases='MIRN101-2, mir-101-2', description='microRNA 101-2', other_designations='hsa-mir-101-2', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4850297, end_position_on_the_genomic_accession=4850375, orientation='plus', exon_count=1, OMIM=612512, no_hdr0=''),
  406895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406895, CurrentID=0, Status='live', Symbol='MIR103A1', Aliases='MIR103-1, MIRN103-1, mir-103a-1', description='microRNA 103a-1', other_designations='hsa-mir-103-1|hsa-mir-103a-1|microRNA 103-1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168560896, end_position_on_the_genomic_accession=168560973, orientation='minus', exon_count=1, OMIM=613187, no_hdr0=''),
  406896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406896, CurrentID=0, Status='live', Symbol='MIR103A2', Aliases='MIR103-2, MIRN103-2, mir-103a-2', description='microRNA 103a-2', other_designations='hsa-mir-103-2|hsa-mir-103a-2|microRNA 103-2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3917494, end_position_on_the_genomic_accession=3917571, orientation='plus', exon_count=1, OMIM=613188, no_hdr0=''),
  406897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406897, CurrentID=0, Status='live', Symbol='MIR105-1', Aliases='MIRN105-1, mir-105-1', description='microRNA 105-1', other_designations='hsa-mir-105-1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152392219, end_position_on_the_genomic_accession=152392299, orientation='minus', exon_count=1, OMIM=300811, no_hdr0=''),
  406898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406898, CurrentID=0, Status='live', Symbol='MIR105-2', Aliases='MIRN105-2, mir-105-2', description='microRNA 105-2', other_designations='hsa-mir-105-2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152394412, end_position_on_the_genomic_accession=152394492, orientation='minus', exon_count=1, OMIM=300812, no_hdr0=''),
  406899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406899, CurrentID=0, Status='live', Symbol='MIR106A', Aliases='MIRN106A, mir-106, mir-106a', description='microRNA 106a', other_designations='hsa-mir-106a', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134170198, end_position_on_the_genomic_accession=134170278, orientation='minus', exon_count=1, OMIM=300792, no_hdr0=''),
  406900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406900, CurrentID=0, Status='live', Symbol='MIR106B', Aliases='MIRN106B, mir-106b', description='microRNA 106b', other_designations='hsa-mir-106b', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100093993, end_position_on_the_genomic_accession=100094074, orientation='minus', exon_count=1, OMIM=612983, no_hdr0=''),
  406901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406901, CurrentID=0, Status='live', Symbol='MIR107', Aliases='MIRN107, miR-107', description='microRNA 107', other_designations='hsa-mir-107', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89592747, end_position_on_the_genomic_accession=89592827, orientation='minus', exon_count=1, OMIM=613189, no_hdr0=''),
  406902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406902, CurrentID=0, Status='live', Symbol='MIR10A', Aliases='MIRN10A, hsa-mir-10a, miRNA10A, mir-10a', description='microRNA 10a', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48579838, end_position_on_the_genomic_accession=48579947, orientation='minus', exon_count=1, OMIM=610173, no_hdr0=''),
  406903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406903, CurrentID=0, Status='live', Symbol='MIR10B', Aliases='MIRN10B, hsa-mir-10b, miRNA10B, mir-10b', description='microRNA 10b', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176150303, end_position_on_the_genomic_accession=176150412, orientation='plus', exon_count=1, OMIM=611576, no_hdr0=''),
  406904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406904, CurrentID=0, Status='live', Symbol='MIR1-1', Aliases='MIRN1-1, hsa-mir-1-1, miRNA1-1, mir-1-1', description='microRNA 1-1', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62554306, end_position_on_the_genomic_accession=62554376, orientation='plus', exon_count=1, OMIM=609326, no_hdr0=''),
  406905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406905, CurrentID=0, Status='live', Symbol='MIR1-2', Aliases='MIRN1-2, hsa-mir-1-2, miRNA1-2, mir-1-2', description='microRNA 1-2', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21829004, end_position_on_the_genomic_accession=21829088, orientation='minus', exon_count=1, OMIM=610252, no_hdr0=''),
  406906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406906, CurrentID=0, Status='live', Symbol='MIR122', Aliases='MIR122A, MIRN122, MIRN122A, hsa-mir-122, miRNA122, miRNA122A, mir-122', description='microRNA 122', other_designations='hsa-mir-122a|microRNA 122a', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58451074, end_position_on_the_genomic_accession=58451158, orientation='plus', exon_count=1, OMIM=609582, no_hdr0=''),
  406907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406907, CurrentID=0, Status='live', Symbol='MIR124-1', Aliases='MIR124A, MIR124A1, MIRN124-1, MIRN124A1, mir-124-1', description='microRNA 124-1', other_designations='hsa-mir-124-1|hsa-mir-124a-1|microRNA 124a-1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9903388, end_position_on_the_genomic_accession=9903472, orientation='minus', exon_count=1, OMIM=609327, no_hdr0=''),
  406908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406908, CurrentID=0, Status='live', Symbol='MIR124-2', Aliases='MIRN124-2, MIRN124A2, mir-124-2', description='microRNA 124-2', other_designations='hsa-mir-124-2|hsa-mir-124a-2|microRNA 124a-2', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64379149, end_position_on_the_genomic_accession=64379257, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406909, CurrentID=0, Status='live', Symbol='MIR124-3', Aliases='MIRN124-3, MIRN124A3, mir-124-3', description='microRNA 124-3', other_designations='hsa-mir-124-3|hsa-mir-124a-3|microRNA 124a-3', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63178500, end_position_on_the_genomic_accession=63178586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406910, CurrentID=0, Status='live', Symbol='MIR125A', Aliases='MIRN125A, miRNA125A, mir-125a', description='microRNA 125a', other_designations='hsa-mir-125a', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51693254, end_position_on_the_genomic_accession=51693339, orientation='plus', exon_count=1, OMIM=611191, no_hdr0=''),
  406911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406911, CurrentID=0, Status='live', Symbol='MIR125B1', Aliases='MIRN125B1, mir-125b-1', description='microRNA 125b-1', other_designations='hsa-mir-125b-1', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122099757, end_position_on_the_genomic_accession=122099844, orientation='minus', exon_count=1, OMIM=610104, no_hdr0=''),
  406912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406912, CurrentID=0, Status='live', Symbol='MIR125B2', Aliases='MIRN125B2, mir-125b-2', description='microRNA 125b-2', other_designations='hsa-mir-125b-2', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16590237, end_position_on_the_genomic_accession=16590325, orientation='plus', exon_count=1, OMIM=610105, no_hdr0=''),
  406913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406913, CurrentID=0, Status='live', Symbol='MIR126', Aliases='MIRN126, miRNA126, mir-126', description='microRNA 126', other_designations='hsa-mir-126', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136670602, end_position_on_the_genomic_accession=136670686, orientation='plus', exon_count=1, OMIM=611767, no_hdr0=''),
  406914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406914, CurrentID=0, Status='live', Symbol='MIR127', Aliases='MIRN127, miRNA127, mir-127', description='microRNA 127', other_designations='hsa-mir-127', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100882979, end_position_on_the_genomic_accession=100883075, orientation='plus', exon_count=1, OMIM=611709, no_hdr0=''),
  406915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406915, CurrentID=0, Status='live', Symbol='MIR128-1', Aliases='MIR128A, MIRN128-1, MIRN128A, mir-128-1', description='microRNA 128-1', other_designations='hsa-mir-128-1|hsa-mir-128a|microRNA 128a', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135665397, end_position_on_the_genomic_accession=135665478, orientation='plus', exon_count=1, OMIM=611774, no_hdr0=''),
  406916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406916, CurrentID=0, Status='live', Symbol='MIR128-2', Aliases='MIR128B, MIRN128-2, MIRN128B, mir-128-2, mir-128b', description='microRNA 128-2', other_designations='hsa-mir-128-2|hsa-mir-128b|microRNA 128b', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35744476, end_position_on_the_genomic_accession=35744559, orientation='plus', exon_count=1, OMIM=611769, no_hdr0=''),
  406917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406917, CurrentID=0, Status='live', Symbol='MIR129-1', Aliases='MIR-129b, MIRN129-1, mir-129-1', description='microRNA 129-1', other_designations='hsa-mir-129-1', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128207872, end_position_on_the_genomic_accession=128207943, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406918, CurrentID=0, Status='live', Symbol='MIR129-2', Aliases='MIR-129b, MIRN129-2, mir-129-2', description='microRNA 129-2', other_designations='hsa-mir-129-2', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43581394, end_position_on_the_genomic_accession=43581483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406919, CurrentID=0, Status='live', Symbol='MIR130A', Aliases='MIRN130A, miRNA130A, mir-130a', description='microRNA 130a', other_designations='hsa-mir-130a', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57641198, end_position_on_the_genomic_accession=57641286, orientation='plus', exon_count=1, OMIM=610175, no_hdr0=''),
  406920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406920, CurrentID=0, Status='live', Symbol='MIR130B', Aliases='MIRN130B, mir-130b', description='microRNA 130b', other_designations='hsa-mir-130b', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21653304, end_position_on_the_genomic_accession=21653385, orientation='plus', exon_count=1, OMIM=613682, no_hdr0=''),
  406921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406921, CurrentID=0, Status='live', Symbol='MIR132', Aliases='MIRN132, miRNA132, mir-132', description='microRNA 132', other_designations='hsa-mir-132', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2049908, end_position_on_the_genomic_accession=2050008, orientation='minus', exon_count=1, OMIM=610016, no_hdr0=''),
  406922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406922, CurrentID=0, Status='live', Symbol='MIR133A1', Aliases='MIRN133A1, mir-133a-1', description='microRNA 133a-1', other_designations='hsa-mir-133a-1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21825698, end_position_on_the_genomic_accession=21825785, orientation='minus', exon_count=1, OMIM=610254, no_hdr0=''),
  406923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406923, CurrentID=0, Status='live', Symbol='MIR133A2', Aliases='MIRN133A2, mir-133a-2', description='microRNA 133a-2', other_designations='hsa-mir-133a-2', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62564912, end_position_on_the_genomic_accession=62565013, orientation='plus', exon_count=1, OMIM=610255, no_hdr0=''),
  406924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406924, CurrentID=0, Status='live', Symbol='MIR134', Aliases='MIRN134, mir-134', description='microRNA 134', other_designations='hsa-mir-134', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101054687, end_position_on_the_genomic_accession=101054759, orientation='plus', exon_count=1, OMIM=610164, no_hdr0=''),
  406925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406925, CurrentID=0, Status='live', Symbol='MIR135A1', Aliases='MIRN135-1, MIRN135A1, mir-135a-1', description='microRNA 135a-1', other_designations='hsa-mir-135-1|hsa-mir-135a-1|microRNA 135-1', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52294219, end_position_on_the_genomic_accession=52294308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406926, CurrentID=0, Status='live', Symbol='MIR135A2', Aliases='MIRN135-2, MIRN135A2, mir-135a-2', description='microRNA 135a-2', other_designations='hsa-mir-135-2|hsa-mir-135a-2|microRNA 135-2', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97563812, end_position_on_the_genomic_accession=97563911, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406927, CurrentID=0, Status='live', Symbol='MIR136', Aliases='MIRN136, miRNA136, mir-136', description='microRNA 136', other_designations='hsa-mir-136', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100884702, end_position_on_the_genomic_accession=100884783, orientation='plus', exon_count=1, OMIM=611710, no_hdr0=''),
  406928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406928, CurrentID=0, Status='live', Symbol='MIR137', Aliases='MIRN137, miR-137', description='microRNA 137', other_designations='hsa-mir-137', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98046070, end_position_on_the_genomic_accession=98046171, orientation='minus', exon_count=1, OMIM=614304, no_hdr0=''),
  406929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406929, CurrentID=0, Status='live', Symbol='MIR138-1', Aliases='MIRN138-1, mir-138-1', description='microRNA 138-1', other_designations='hsa-mir-138-1', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44114212, end_position_on_the_genomic_accession=44114310, orientation='plus', exon_count=1, OMIM=613394, no_hdr0=''),
  406930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406930, CurrentID=0, Status='live', Symbol='MIR138-2', Aliases='MIRN138-2, mir-138-2', description='microRNA 138-2', other_designations='hsa-mir-138-2', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56858518, end_position_on_the_genomic_accession=56858601, orientation='plus', exon_count=1, OMIM=613395, no_hdr0=''),
  406931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406931, CurrentID=0, Status='live', Symbol='MIR139', Aliases='MIR139-3p, MIRN139, mir-139', description='microRNA 139', other_designations='hsa-mir-139', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72615063, end_position_on_the_genomic_accession=72615130, orientation='minus', exon_count=1, OMIM=615017, no_hdr0=''),
  406932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406932, CurrentID=0, Status='live', Symbol='MIR140', Aliases='MIRN140, miRNA140, mir-140', description='microRNA 140', other_designations='hsa-mir-140', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69933081, end_position_on_the_genomic_accession=69933180, orientation='plus', exon_count=1, OMIM=611894, no_hdr0=''),
  406933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406933, CurrentID=0, Status='live', Symbol='MIR141', Aliases='MIRN141, mir-141', description='microRNA 141', other_designations='hsa-mir-141', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6964097, end_position_on_the_genomic_accession=6964191, orientation='plus', exon_count=1, OMIM=612093, no_hdr0=''),
  406934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406934, CurrentID=0, Status='live', Symbol='MIR142', Aliases='MIRN142, mir-142', description='microRNA 142', other_designations='hsa-mir-142', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58331232, end_position_on_the_genomic_accession=58331318, orientation='minus', exon_count=1, OMIM=615657, no_hdr0=''),
  406935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406935, CurrentID=0, Status='live', Symbol='MIR143', Aliases='MIRN143, mir-143', description='microRNA 143', other_designations='hsa-mir-143', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149428918, end_position_on_the_genomic_accession=149429023, orientation='plus', exon_count=1, OMIM=612117, no_hdr0=''),
  406936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406936, CurrentID=0, Status='live', Symbol='MIR144', Aliases='MIRN144, mir-144', description='microRNA 144', other_designations='hsa-mir-144', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28861533, end_position_on_the_genomic_accession=28861618, orientation='minus', exon_count=1, OMIM=612070, no_hdr0=''),
  406937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406937, CurrentID=0, Status='live', Symbol='MIR145', Aliases='MIRN145, miR-145, miRNA145', description='microRNA 145', other_designations='hsa-mir-145', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149430646, end_position_on_the_genomic_accession=149430733, orientation='plus', exon_count=1, OMIM=611795, no_hdr0=''),
  406938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406938, CurrentID=0, Status='live', Symbol='MIR146A', Aliases='MIRN146, MIRN146A, miR-146a, miRNA146A', description='microRNA 146a', other_designations='hsa-mir-146|hsa-mir-146a', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160485352, end_position_on_the_genomic_accession=160485450, orientation='plus', exon_count=1, OMIM=610566, no_hdr0=''),
  406939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406939, CurrentID=0, Status='live', Symbol='MIR147A', Aliases='MIR147, MIRN147, hsa-mir-147a', description='microRNA 147a', other_designations='hsa-mir-147|microRNA 147', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120244979, end_position_on_the_genomic_accession=120245050, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406940, CurrentID=0, Status='live', Symbol='MIR148A', Aliases='MIRN148, MIRN148A, hsa-mir-148, mir-148a', description='microRNA 148a', other_designations='hsa-mir-148a|microRNA 148', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25949919, end_position_on_the_genomic_accession=25949986, orientation='minus', exon_count=1, OMIM=613786, no_hdr0=''),
  406941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406941, CurrentID=0, Status='live', Symbol='MIR149', Aliases='MIRN149, mir-149', description='microRNA 149', other_designations='hsa-mir-149', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240456001, end_position_on_the_genomic_accession=240456089, orientation='plus', exon_count=1, OMIM=615209, no_hdr0=''),
  406942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406942, CurrentID=0, Status='live', Symbol='MIR150', Aliases='MIRN150, miRNA150, mir-150', description='microRNA 150', other_designations='hsa-mir-150', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49500785, end_position_on_the_genomic_accession=49500868, orientation='minus', exon_count=1, OMIM=611114, no_hdr0=''),
  406943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406943, CurrentID=0, Status='live', Symbol='MIR152', Aliases='MIRN152, mir-152', description='microRNA 152', other_designations='hsa-mir-152', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48037161, end_position_on_the_genomic_accession=48037247, orientation='minus', exon_count=1, OMIM=613788, no_hdr0=''),
  406944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406944, CurrentID=0, Status='live', Symbol='MIR153-1', Aliases='MIRN153-1, mir-153-1', description='microRNA 153-1', other_designations='hsa-mir-153-1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219294111, end_position_on_the_genomic_accession=219294200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406945, CurrentID=0, Status='live', Symbol='MIR153-2', Aliases='MIRN153-2, mir-153-2', description='microRNA 153-2', other_designations='hsa-mir-153-2', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157574336, end_position_on_the_genomic_accession=157574422, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406946, CurrentID=0, Status='live', Symbol='MIR154', Aliases='MIRN154, mir-154', description='microRNA 154', other_designations='hsa-mir-154', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101059755, end_position_on_the_genomic_accession=101059838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406947, CurrentID=0, Status='live', Symbol='MIR155', Aliases='MIRN155, miRNA155, mir-155', description='microRNA 155', other_designations='hsa-mir-155', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25573980, end_position_on_the_genomic_accession=25574044, orientation='plus', exon_count=1, OMIM=609337, no_hdr0=''),
  406948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406948, CurrentID=0, Status='live', Symbol='MIR15A', Aliases='MIRN15A, hsa-mir-15a, miRNA15A, mir-15a', description='microRNA 15a', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50049119, end_position_on_the_genomic_accession=50049201, orientation='minus', exon_count=1, OMIM=609703, no_hdr0=''),
  406949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406949, CurrentID=0, Status='live', Symbol='MIR15B', Aliases='MIRN15B, hsa-mir-15b, miR-15b', description='microRNA 15b', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160404588, end_position_on_the_genomic_accession=160404685, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406950, CurrentID=0, Status='live', Symbol='MIR16-1', Aliases='MIRN16-1, miRNA16-1, mir-16-1', description='microRNA 16-1', other_designations='hsa-mir-16-1', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50048973, end_position_on_the_genomic_accession=50049061, orientation='minus', exon_count=1, OMIM=609704, no_hdr0=''),
  406951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406951, CurrentID=0, Status='live', Symbol='MIR16-2', Aliases='MIRN16-2, mir-16-2, mir-16-3', description='microRNA 16-2', other_designations='hsa-mir-16-2', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160404745, end_position_on_the_genomic_accession=160404825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406952, CurrentID=0, Status='live', Symbol='MIR17', Aliases='MIR17-5p, MIR91, MIRN17, MIRN91, hsa-mir-17, miR-17, miR17-3p, miRNA17, miRNA91', description='microRNA 17', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91350605, end_position_on_the_genomic_accession=91350688, orientation='plus', exon_count=1, OMIM=609416, no_hdr0=''),
  406953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406953, CurrentID=0, Status='live', Symbol='MIR18A', Aliases='MIR18, MIRN18, MIRN18A, hsa-mir-18, hsa-mir-18a, miR-18, miRNA18A, mir-18a', description='microRNA 18a', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91350751, end_position_on_the_genomic_accession=91350821, orientation='plus', exon_count=1, OMIM=609417, no_hdr0=''),
  406954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406954, CurrentID=0, Status='live', Symbol='MIR181A2', Aliases='MIRN181A, MIRN181A2, hsa-mir-181a-2, mir-181a-2', description='microRNA 181a-2', other_designations='hsa-mir-181a', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124692442, end_position_on_the_genomic_accession=124692551, orientation='plus', exon_count=1, OMIM=612743, no_hdr0=''),
  406955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406955, CurrentID=0, Status='live', Symbol='MIR181B1', Aliases='MIRN181B1, mir-181b-1', description='microRNA 181b-1', other_designations='hsa-mir-181b-1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198858873, end_position_on_the_genomic_accession=198858982, orientation='minus', exon_count=1, OMIM=612744, no_hdr0=''),
  406956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406956, CurrentID=0, Status='live', Symbol='MIR181B2', Aliases='MIRN181B2, mir-181b-2', description='microRNA 181b-2', other_designations='hsa-mir-181b-2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124693710, end_position_on_the_genomic_accession=124693798, orientation='plus', exon_count=1, OMIM=612745, no_hdr0=''),
  406957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406957, CurrentID=0, Status='live', Symbol='MIR181C', Aliases='MIRN181C, mir-181c', description='microRNA 181c', other_designations='hsa-mir-181c', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13874699, end_position_on_the_genomic_accession=13874808, orientation='plus', exon_count=1, OMIM=612746, no_hdr0=''),
  406958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406958, CurrentID=0, Status='live', Symbol='MIR182', Aliases='MIRN182, miRNA182, mir-182', description='microRNA 182', other_designations='hsa-mir-182', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129770383, end_position_on_the_genomic_accession=129770492, orientation='minus', exon_count=1, OMIM=611607, no_hdr0=''),
  406959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406959, CurrentID=0, Status='live', Symbol='MIR183', Aliases='MIRN183, miR-183, miRNA183', description='microRNA 183', other_designations='hsa-mir-183', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129774905, end_position_on_the_genomic_accession=129775014, orientation='minus', exon_count=1, OMIM=611608, no_hdr0=''),
  406960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406960, CurrentID=0, Status='live', Symbol='MIR184', Aliases='EDICT, MIRN184, miR-184', description='microRNA 184', other_designations='hsa-mir-184', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79209788, end_position_on_the_genomic_accession=79209871, orientation='plus', exon_count=1, OMIM=613146, no_hdr0=''),
  406961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406961, CurrentID=0, Status='live', Symbol='MIR185', Aliases='MIRN185, miR-185', description='microRNA 185', other_designations='hsa-mir-185', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20033139, end_position_on_the_genomic_accession=20033220, orientation='plus', exon_count=1, OMIM=615576, no_hdr0=''),
  406962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406962, CurrentID=0, Status='live', Symbol='MIR186', Aliases='MIRN186, miR-186', description='microRNA 186', other_designations='hsa-mir-186', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71067631, end_position_on_the_genomic_accession=71067716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406963, CurrentID=0, Status='live', Symbol='MIR187', Aliases='MIRN187, miR-187, miRNA187', description='microRNA 187', other_designations='hsa-mir-187', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35904818, end_position_on_the_genomic_accession=35904926, orientation='minus', exon_count=1, OMIM=612698, no_hdr0=''),
  406964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406964, CurrentID=0, Status='live', Symbol='MIR188', Aliases='MIRN188, miR-188', description='microRNA 188', other_designations='hsa-mir-188', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50003503, end_position_on_the_genomic_accession=50003588, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  406965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406965, CurrentID=0, Status='live', Symbol='MIR190A', Aliases='MIR190, MIRN190, hsa-mir-190a, miR-190, mir-190a', description='microRNA 190a', other_designations='hsa-mir-190|microRNA 190', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62823957, end_position_on_the_genomic_accession=62824041, orientation='plus', exon_count=1, OMIM=615845, no_hdr0=''),
  406966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406966, CurrentID=0, Status='live', Symbol='MIR191', Aliases='MIRN191, miR-191', description='microRNA 191', other_designations='hsa-mir-191', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49020618, end_position_on_the_genomic_accession=49020709, orientation='minus', exon_count=1, OMIM=615150, no_hdr0=''),
  406967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406967, CurrentID=0, Status='live', Symbol='MIR192', Aliases='MIRN192, miR-192, miRNA192', description='microRNA 192', other_designations='hsa-mir-192', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64891137, end_position_on_the_genomic_accession=64891246, orientation='minus', exon_count=1, OMIM=610939, no_hdr0=''),
  406968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406968, CurrentID=0, Status='live', Symbol='MIR193A', Aliases='MIRN193, MIRN193A, mir-193a', description='microRNA 193a', other_designations='hsa-mir-193|hsa-mir-193a', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31559996, end_position_on_the_genomic_accession=31560083, orientation='plus', exon_count=1, OMIM=614733, no_hdr0=''),
  406969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406969, CurrentID=0, Status='live', Symbol='MIR194-1', Aliases='MIRN194-1, mir-194-1', description='microRNA 194-1', other_designations='hsa-mir-194-1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220118157, end_position_on_the_genomic_accession=220118241, orientation='minus', exon_count=1, OMIM=610940, no_hdr0=''),
  406970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406970, CurrentID=0, Status='live', Symbol='MIR194-2', Aliases='MIRN194-2, mir-194-2', description='microRNA 194-2', other_designations='hsa-mir-194-2', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64891355, end_position_on_the_genomic_accession=64891439, orientation='minus', exon_count=1, OMIM=610941, no_hdr0=''),
  406971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406971, CurrentID=0, Status='live', Symbol='MIR195', Aliases='MIRN195, miRNA195, mir-195', description='microRNA 195', other_designations='hsa-mir-195', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7017615, end_position_on_the_genomic_accession=7017701, orientation='minus', exon_count=1, OMIM=610718, no_hdr0=''),
  406972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406972, CurrentID=0, Status='live', Symbol='MIR196A1', Aliases='MIRN196-1, MIRN196A1, mir-196a-1', description='microRNA 196a-1', other_designations='hsa-mir-196-1|hsa-mir-196a-1|microRNA 196-1|microRNA 196a|microRNA miR-196', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48632490, end_position_on_the_genomic_accession=48632559, orientation='minus', exon_count=1, OMIM=608632, no_hdr0=''),
  406973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406973, CurrentID=0, Status='live', Symbol='MIR196A2', Aliases='MIRN196-2, MIRN196A2, mir-196a-2', description='microRNA 196a-2', other_designations='hsa-mir-196-2|hsa-mir-196a-2|microRNA 196-2', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53991738, end_position_on_the_genomic_accession=53991847, orientation='plus', exon_count=1, OMIM=609687, no_hdr0=''),
  406974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406974, CurrentID=0, Status='live', Symbol='MIR197', Aliases='MIRN197, miR-197, miRNA197', description='microRNA 197', other_designations='hsa-mir-197', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109598893, end_position_on_the_genomic_accession=109598967, orientation='plus', exon_count=1, OMIM=611189, no_hdr0=''),
  406975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406975, CurrentID=0, Status='live', Symbol='MIR198', Aliases='MIRN198, miR-198', description='microRNA 198', other_designations='hsa-mir-198', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120395668, end_position_on_the_genomic_accession=120395729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  406976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406976, CurrentID=0, Status='live', Symbol='MIR199A1', Aliases='MIR-199-s, MIRN199A1, mir-199a-1', description='microRNA 199a-1', other_designations='hsa-mir-199a-1|miR-199a-3p', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10817426, end_position_on_the_genomic_accession=10817496, orientation='minus', exon_count=1, OMIM=610719, no_hdr0=''),
  406977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406977, CurrentID=0, Status='live', Symbol='MIR199A2', Aliases='MIR-199-s, MIRN199A2, mir-199a-2', description='microRNA 199a-2', other_designations='hsa-mir-199a-2|miR-199a-3p', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172144535, end_position_on_the_genomic_accession=172144644, orientation='minus', exon_count=1, OMIM=610720, no_hdr0=''),
  406978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406978, CurrentID=0, Status='live', Symbol='MIR199B', Aliases='MIRN199B, mir-199b', description='microRNA 199b', other_designations='hsa-mir-199b', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128244721, end_position_on_the_genomic_accession=128244830, orientation='minus', exon_count=1, OMIM=614791, no_hdr0=''),
  406979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406979, CurrentID=0, Status='live', Symbol='MIR19A', Aliases='MIRN19A, hsa-mir-19a, miR-19a, miRNA19A', description='microRNA 19a', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91350891, end_position_on_the_genomic_accession=91350972, orientation='plus', exon_count=1, OMIM=609418, no_hdr0=''),
  406980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406980, CurrentID=0, Status='live', Symbol='MIR19B1', Aliases='MIR19B, MIRN19B1, miR-19b-1, miRNA19B1', description='microRNA 19b-1', other_designations='MicroRNA-19b|hsa-mir-19b-1|miR-19b', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91351192, end_position_on_the_genomic_accession=91351278, orientation='plus', exon_count=1, OMIM=609419, no_hdr0=''),
  406981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406981, CurrentID=0, Status='live', Symbol='MIR19B2', Aliases='MIRN19B2, miR-19b-2', description='microRNA 19b-2', other_designations='hsa-mir-19b-2', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134169671, end_position_on_the_genomic_accession=134169766, orientation='minus', exon_count=1, OMIM=300722, no_hdr0=''),
  406982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406982, CurrentID=0, Status='live', Symbol='MIR20A', Aliases='MIR20, MIRN20, MIRN20A, hsa-mir-20, hsa-mir-20a, miR-20, miRNA20A, mir-20a', description='microRNA 20a', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91351065, end_position_on_the_genomic_accession=91351135, orientation='plus', exon_count=1, OMIM=609420, no_hdr0=''),
  406983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406983, CurrentID=0, Status='live', Symbol='MIR200A', Aliases='MIRN200A, mir-200a', description='microRNA 200a', other_designations='hsa-mir-200a', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1167863, end_position_on_the_genomic_accession=1167952, orientation='plus', exon_count=1, OMIM=612090, no_hdr0=''),
  406984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406984, CurrentID=0, Status='live', Symbol='MIR200B', Aliases='MIRN200B, mir-200b', description='microRNA 200b', other_designations='hsa-mir-200b', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1167104, end_position_on_the_genomic_accession=1167198, orientation='plus', exon_count=1, OMIM=612091, no_hdr0=''),
  406985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406985, CurrentID=0, Status='live', Symbol='MIR200C', Aliases='MIRN200C, mir-200c', description='microRNA 200c', other_designations='hsa-mir-200c', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6963699, end_position_on_the_genomic_accession=6963766, orientation='plus', exon_count=1, OMIM=612092, no_hdr0=''),
  406986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406986, CurrentID=0, Status='live', Symbol='MIR203A', Aliases='MIR203, MIRN203, hsa-mir-203a, miR-203, miRNA203, mir-203a', description='microRNA 203a', other_designations='hsa-mir-203|microRNA 203', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104117405, end_position_on_the_genomic_accession=104117514, orientation='plus', exon_count=1, OMIM=611899, no_hdr0=''),
  406987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406987, CurrentID=0, Status='live', Symbol='MIR204', Aliases='MIRN204, RDICC, miRNA204, mir-204', description='microRNA 204', other_designations='hsa-mir-204', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70809975, end_position_on_the_genomic_accession=70810084, orientation='minus', exon_count=1, OMIM=610942, no_hdr0=''),
  406988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406988, CurrentID=0, Status='live', Symbol='MIR205', Aliases='MIRN205, mir-205', description='microRNA 205', other_designations='hsa-mir-205', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209432133, end_position_on_the_genomic_accession=209432242, orientation='plus', exon_count=1, OMIM=613147, no_hdr0=''),
  406989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406989, CurrentID=0, Status='live', Symbol='MIR206', Aliases='MIRN206, miRNA206, mir-206', description='microRNA 206', other_designations='hsa-mir-206', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52144349, end_position_on_the_genomic_accession=52144434, orientation='plus', exon_count=1, OMIM=611599, no_hdr0=''),
  406990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406990, CurrentID=0, Status='live', Symbol='MIR208A', Aliases='MIR208, MIRN208, MIRN208A, hsa-mir-208a, miRNA208', description='microRNA 208a', other_designations='hsa-mir-208|microRNA 208', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23388596, end_position_on_the_genomic_accession=23388666, orientation='minus', exon_count=1, OMIM=611116, no_hdr0=''),
  144847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144847, CurrentID=0, Status='live', Symbol='ESRRAP1', Aliases='ESRRAP, bA76K19.2', description='estrogen-related receptor alpha pseudogene 1', other_designations='', map_location='13q12.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19559417, end_position_on_the_genomic_accession=19561498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  406992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406992, CurrentID=0, Status='live', Symbol='MIR210', Aliases='MIRN210, mir-210', description='microRNA 210', other_designations='hsa-mir-210', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=568089, end_position_on_the_genomic_accession=568198, orientation='minus', exon_count=1, OMIM=612982, no_hdr0=''),
  406993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406993, CurrentID=0, Status='live', Symbol='MIR211', Aliases='MIRN211, mir-211', description='microRNA 211', other_designations='hsa-mir-211', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31065032, end_position_on_the_genomic_accession=31065141, orientation='minus', exon_count=1, OMIM=613753, no_hdr0=''),
  406994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406994, CurrentID=0, Status='live', Symbol='MIR212', Aliases='MIRN212, mir-212', description='microRNA 212', other_designations='hsa-mir-212', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2050271, end_position_on_the_genomic_accession=2050380, orientation='minus', exon_count=1, OMIM=613487, no_hdr0=''),
  406995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406995, CurrentID=0, Status='live', Symbol='MIR181A1', Aliases='MIR213, MIRN181A1, MIRN213, hsa-mir-181a-1, mir-181a-1, mir-213', description='microRNA 181a-1', other_designations='hsa-mir-213|microRNA 213', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198859044, end_position_on_the_genomic_accession=198859153, orientation='minus', exon_count=1, OMIM=612742, no_hdr0=''),
  406996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406996, CurrentID=0, Status='live', Symbol='MIR214', Aliases='MIRN214, miRNA214, mir-214', description='microRNA 214', other_designations='hsa-mir-214', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172138798, end_position_on_the_genomic_accession=172138907, orientation='minus', exon_count=1, OMIM=610721, no_hdr0=''),
  406997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406997, CurrentID=0, Status='live', Symbol='MIR215', Aliases='MIRN215, miRNA215, mir-215', description='microRNA 215', other_designations='hsa-mir-215', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220117853, end_position_on_the_genomic_accession=220117962, orientation='minus', exon_count=1, OMIM=610943, no_hdr0=''),
  406998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406998, CurrentID=0, Status='live', Symbol='MIR216A', Aliases='MIR216, MIRN216, MIRN216A, miRNA216, mir-216, mir-216a', description='microRNA 216a', other_designations='hsa-mir-216|hsa-mir-216a|microRNA 216', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55988950, end_position_on_the_genomic_accession=55989059, orientation='minus', exon_count=1, OMIM=610944, no_hdr0=''),
  406999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406999, CurrentID=0, Status='live', Symbol='MIR217', Aliases='MIRN217, mir-217', description='microRNA 217', other_designations='hsa-mir-217', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55982967, end_position_on_the_genomic_accession=55983076, orientation='minus', exon_count=1, OMIM=615096, no_hdr0=''),
  407000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407000, CurrentID=0, Status='live', Symbol='MIR218-1', Aliases='MIRN218-1, mir-218-1', description='microRNA 218-1', other_designations='hsa-mir-218-1', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20528275, end_position_on_the_genomic_accession=20528384, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407001, CurrentID=0, Status='live', Symbol='MIR218-2', Aliases='MIRN218-2, mir-218-2', description='microRNA 218-2', other_designations='hsa-mir-218-2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168768146, end_position_on_the_genomic_accession=168768255, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407002, CurrentID=0, Status='live', Symbol='MIR219A1', Aliases='MIR219-1, MIRN219-1, MRI219-1, hsa-mir-219a-1, mir-219, mir-219a-1', description='microRNA 219a-1', other_designations='hsa-mir-219-1|microRNA 219-1', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33207835, end_position_on_the_genomic_accession=33207944, orientation='plus', exon_count=1, OMIM=611500, no_hdr0=''),
  407003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407003, CurrentID=0, Status='live', Symbol='MIR219A2', Aliases='MIR219-2, MIRN219-2, hsa-mir-219a-2, mir-219a-2', description='microRNA 219a-2', other_designations='hsa-mir-219-2|microRNA 219-2', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128392618, end_position_on_the_genomic_accession=128392714, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407004, CurrentID=0, Status='live', Symbol='MIR22', Aliases='MIRN22, hsa-mir-22, miR-22', description='microRNA 22', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1713903, end_position_on_the_genomic_accession=1713987, orientation='minus', exon_count=1, OMIM=612077, no_hdr0=''),
  407006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407006, CurrentID=0, Status='live', Symbol='MIR221', Aliases='MIRN221, miRNA221, mir-221', description='microRNA 221', other_designations='hsa-mir-221', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45746157, end_position_on_the_genomic_accession=45746266, orientation='minus', exon_count=1, OMIM=300568, no_hdr0=''),
  407007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407007, CurrentID=0, Status='live', Symbol='MIR222', Aliases='MIRN222, miRNA222, mir-222', description='microRNA 222', other_designations='hsa-mir-222', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45747015, end_position_on_the_genomic_accession=45747124, orientation='minus', exon_count=1, OMIM=300569, no_hdr0=''),
  407008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407008, CurrentID=0, Status='live', Symbol='MIR223', Aliases='MIRN223, miRNA223, mir-223', description='microRNA 223', other_designations='hsa-mir-223', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66018870, end_position_on_the_genomic_accession=66018979, orientation='plus', exon_count=1, OMIM=300694, no_hdr0=''),
  407009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407009, CurrentID=0, Status='live', Symbol='MIR224', Aliases='MIRN224, miRNA224', description='microRNA 224', other_designations='hsa-mir-224', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151958578, end_position_on_the_genomic_accession=151958658, orientation='minus', exon_count=1, OMIM=300769, no_hdr0=''),
  407010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407010, CurrentID=0, Status='live', Symbol='MIR23A', Aliases='MIRN23A, hsa-mir-23a, miRNA23A, mir-23a', description='microRNA 23a', other_designations='', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13836587, end_position_on_the_genomic_accession=13836659, orientation='minus', exon_count=1, OMIM=607962, no_hdr0=''),
  407011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407011, CurrentID=0, Status='live', Symbol='MIR23B', Aliases='MIRN23B, hsa-mir-23b, miRNA23B, mir-23b', description='microRNA 23b', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95085208, end_position_on_the_genomic_accession=95085304, orientation='plus', exon_count=1, OMIM=610723, no_hdr0=''),
  407012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407012, CurrentID=0, Status='live', Symbol='MIR24-1', Aliases='MIR189, MIRN24-1, miR-24-1, miRNA24-1', description='microRNA 24-1', other_designations='hsa-mir-24-1|microRNA-24', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95086021, end_position_on_the_genomic_accession=95086088, orientation='plus', exon_count=1, OMIM=609705, no_hdr0=''),
  407013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407013, CurrentID=0, Status='live', Symbol='MIR24-2', Aliases='MIRN24-2, miR-24-2, miRNA24-2', description='microRNA 24-2', other_designations='hsa-mir-24-2', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13836287, end_position_on_the_genomic_accession=13836359, orientation='minus', exon_count=1, OMIM=610724, no_hdr0=''),
  407014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407014, CurrentID=0, Status='live', Symbol='MIR25', Aliases='MIRN25, hsa-mir-25, miR-25', description='microRNA 25', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100093560, end_position_on_the_genomic_accession=100093643, orientation='minus', exon_count=1, OMIM=612150, no_hdr0=''),
  407015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407015, CurrentID=0, Status='live', Symbol='MIR26A1', Aliases='MIR26A, MIRN26A1, mir-26a-1', description='microRNA 26a-1', other_designations='hsa-mir-26a-1', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37969404, end_position_on_the_genomic_accession=37969480, orientation='plus', exon_count=1, OMIM=612151, no_hdr0=''),
  407016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407016, CurrentID=0, Status='live', Symbol='MIR26A2', Aliases='MIRN26A2, mir-26a-2', description='microRNA 26a-2', other_designations='hsa-mir-26a-2', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57824609, end_position_on_the_genomic_accession=57824692, orientation='minus', exon_count=1, OMIM=613057, no_hdr0=''),
  407017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407017, CurrentID=0, Status='live', Symbol='MIR26B', Aliases='MIRN26B, hsa-mir-26b, miR-26b', description='microRNA 26b', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218402646, end_position_on_the_genomic_accession=218402722, orientation='plus', exon_count=1, OMIM=612152, no_hdr0=''),
  407018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407018, CurrentID=0, Status='live', Symbol='MIR27A', Aliases='MIR27, MIRN27A, mir-27a', description='microRNA 27a', other_designations='hsa-mir-27a', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13836440, end_position_on_the_genomic_accession=13836517, orientation='minus', exon_count=1, OMIM=612153, no_hdr0=''),
  407019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407019, CurrentID=0, Status='live', Symbol='MIR27B', Aliases='MIR-27b, MIRN27B, miRNA27B', description='microRNA 27b', other_designations='hsa-mir-27b', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95085445, end_position_on_the_genomic_accession=95085541, orientation='plus', exon_count=1, OMIM=610636, no_hdr0=''),
  407020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407020, CurrentID=0, Status='live', Symbol='MIR28', Aliases='MIRN28, hsa-mir-28, miR-28', description='microRNA 28', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188688781, end_position_on_the_genomic_accession=188688866, orientation='plus', exon_count=1, OMIM=612154, no_hdr0=''),
  407021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407021, CurrentID=0, Status='live', Symbol='MIR29A', Aliases='MIRN29, MIRN29A, hsa-mir-29, hsa-mir-29a, miRNA29A, mir-29a', description='microRNA 29a', other_designations='microRNA 29', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130876747, end_position_on_the_genomic_accession=130876810, orientation='minus', exon_count=1, OMIM=610782, no_hdr0=''),
  407022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407022, CurrentID=0, Status='live', Symbol='MIR296', Aliases='MIRN296, miRNA296, mir-296', description='microRNA 296', other_designations='hsa-mir-296', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58817615, end_position_on_the_genomic_accession=58817694, orientation='minus', exon_count=1, OMIM=610945, no_hdr0=''),
  407023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407023, CurrentID=0, Status='live', Symbol='MIR299', Aliases='MIRN299, mir-299', description='microRNA 299', other_designations='hsa-mir-299', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101023794, end_position_on_the_genomic_accession=101023856, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407024, CurrentID=0, Status='live', Symbol='MIR29B1', Aliases='MIRN29B1, miR-29b, miRNA29B1, mir-29b-1', description='microRNA 29b-1', other_designations='hsa-mir-29b-1', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130877459, end_position_on_the_genomic_accession=130877539, orientation='minus', exon_count=1, OMIM=610783, no_hdr0=''),
  407025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407025, CurrentID=0, Status='live', Symbol='MIR29B2', Aliases='MIRN29B2', description='microRNA 29b-2', other_designations='hsa-mir-29b-2|miR-29b|mir-29b-2', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207802443, end_position_on_the_genomic_accession=207802523, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407026, CurrentID=0, Status='live', Symbol='MIR29C', Aliases='MIRN29C, miRNA29C, mir-29c', description='microRNA 29c', other_designations='hsa-mir-29c', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207801852, end_position_on_the_genomic_accession=207801939, orientation='minus', exon_count=1, OMIM=610784, no_hdr0=''),
  407027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407027, CurrentID=0, Status='live', Symbol='MIR301A', Aliases='MIR301, MIRN301, MIRN301A, mir-301a', description='microRNA 301a', other_designations='hsa-mir-301|hsa-mir-301a|microRNA 301', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59151136, end_position_on_the_genomic_accession=59151221, orientation='minus', exon_count=1, OMIM=615675, no_hdr0=''),
  407028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407028, CurrentID=0, Status='live', Symbol='MIR302A', Aliases='MIRN302, MIRN302A, hsa-mir-302, mir-302a', description='microRNA 302a', other_designations='hsa-mir-302a|microRNA 302', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112648183, end_position_on_the_genomic_accession=112648251, orientation='minus', exon_count=1, OMIM=614596, no_hdr0=''),
  407029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407029, CurrentID=0, Status='live', Symbol='MIR30A', Aliases='MIRN30A, mir-30a', description='microRNA 30a', other_designations='hsa-mir-30a', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71403551, end_position_on_the_genomic_accession=71403621, orientation='minus', exon_count=1, OMIM=612329, no_hdr0=''),
  407030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407030, CurrentID=0, Status='live', Symbol='MIR30B', Aliases='MIRN30B, mir-30b', description='microRNA 30b', other_designations='hsa-mir-30b', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134800520, end_position_on_the_genomic_accession=134800607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407031, CurrentID=0, Status='live', Symbol='MIR30C1', Aliases='MIRN30C1, mir-30c-1', description='microRNA 30c-1', other_designations='hsa-mir-30c-1|miR-30c', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40757284, end_position_on_the_genomic_accession=40757372, orientation='plus', exon_count=1, OMIM=615151, no_hdr0=''),
  407032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407032, CurrentID=0, Status='live', Symbol='MIR30C2', Aliases='MIRN30C2, mir-30c-2', description='microRNA 30c-2', other_designations='hsa-mir-30c-2|miR-30c', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71376960, end_position_on_the_genomic_accession=71377031, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407033, CurrentID=0, Status='live', Symbol='MIR30D', Aliases='MIRN30D, mir-30d', description='microRNA 30d', other_designations='hsa-mir-30d', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134804876, end_position_on_the_genomic_accession=134804945, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407034, CurrentID=0, Status='live', Symbol='MIR30E', Aliases='MIRN30E, mir-30e', description='microRNA 30e', other_designations='hsa-mir-30e', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40754355, end_position_on_the_genomic_accession=40754446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407035, CurrentID=0, Status='live', Symbol='MIR31', Aliases='MIRN31, hsa-mir-31, miR-31', description='microRNA 31', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21512115, end_position_on_the_genomic_accession=21512185, orientation='minus', exon_count=1, OMIM=612155, no_hdr0=''),
  407036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407036, CurrentID=0, Status='live', Symbol='MIR32', Aliases='MIRN32, hsa-mir-32, miR-32, miRNA32', description='microRNA 32', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109046229, end_position_on_the_genomic_accession=109046298, orientation='minus', exon_count=1, OMIM=609355, no_hdr0=''),
  407037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407037, CurrentID=0, Status='live', Symbol='MIR320A', Aliases='MIRN320, MIRN320A, hsa-mir-320a, mir-320a', description='microRNA 320a', other_designations='hsa-mir-320|microRNA 320', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22244962, end_position_on_the_genomic_accession=22245043, orientation='minus', exon_count=1, OMIM=614112, no_hdr0=''),
  407039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407039, CurrentID=0, Status='live', Symbol='MIR33A', Aliases='MIR33, MIRN33, MIRN33A, hsa-mir-33, hsa-mir-33a, miR-33, miRNA33A, mir-33a', description='microRNA 33a', other_designations='microRNA 33', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41900944, end_position_on_the_genomic_accession=41901012, orientation='plus', exon_count=1, OMIM=612156, no_hdr0=''),
  407040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407040, CurrentID=0, Status='live', Symbol='MIR34A', Aliases='MIRN34A, miRNA34A, mir-34, mir-34a', description='microRNA 34a', other_designations='hsa-mir-34a', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9151668, end_position_on_the_genomic_accession=9151777, orientation='minus', exon_count=1, OMIM=611172, no_hdr0=''),
  407041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407041, CurrentID=0, Status='live', Symbol='MIR34B', Aliases='MIRN34B, miRNA34B, mir-34b', description='microRNA 34b', other_designations='hsa-mir-34b', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111512938, end_position_on_the_genomic_accession=111513021, orientation='plus', exon_count=1, OMIM=611374, no_hdr0=''),
  407042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407042, CurrentID=0, Status='live', Symbol='MIR34C', Aliases='MIRN34C, miRNA34C, mir-34c', description='microRNA 34c', other_designations='hsa-mir-34c', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111513439, end_position_on_the_genomic_accession=111513515, orientation='plus', exon_count=1, OMIM=611375, no_hdr0=''),
  407043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407043, CurrentID=0, Status='live', Symbol='MIR7-1', Aliases='MIRN7-1, hsa-mir-7-1, mir-7-1', description='microRNA 7-1', other_designations='miR-7-5p', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83969748, end_position_on_the_genomic_accession=83969857, orientation='minus', exon_count=1, OMIM=615239, no_hdr0=''),
  407044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407044, CurrentID=0, Status='live', Symbol='MIR7-2', Aliases='MIRN7-2, hsa-mir-7-2, mir-7-2', description='microRNA 7-2', other_designations='miR-7-5p', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88611825, end_position_on_the_genomic_accession=88611934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407045, CurrentID=0, Status='live', Symbol='MIR7-3', Aliases='MIRN7-3, hsa-mir-7-3, mir-7-3', description='microRNA 7-3', other_designations='miR-7-5p', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4770670, end_position_on_the_genomic_accession=4770779, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407046, CurrentID=0, Status='live', Symbol='MIR9-1', Aliases='MIRN9-1, hsa-mir-9-1, miRNA9-1, mir-9-1', description='microRNA 9-1', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156420341, end_position_on_the_genomic_accession=156420429, orientation='minus', exon_count=1, OMIM=611186, no_hdr0=''),
  407047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407047, CurrentID=0, Status='live', Symbol='MIR9-2', Aliases='MIRN9-2, hsa-mir-9-2, miRNA9-2, mir-9-2', description='microRNA 9-2', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88666853, end_position_on_the_genomic_accession=88666939, orientation='minus', exon_count=1, OMIM=611187, no_hdr0=''),
  407048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407048, CurrentID=0, Status='live', Symbol='MIR92A1', Aliases='MIR92-1, MIRN92-1, MIRN92A-1, MIRN92A1, miRNA92-1, mir-92a-1', description='microRNA 92a-1', other_designations='hsa-mir-92-1|hsa-mir-92a-1|microRNA 92-1', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91351314, end_position_on_the_genomic_accession=91351391, orientation='plus', exon_count=1, OMIM=609422, no_hdr0=''),
  407049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407049, CurrentID=0, Status='live', Symbol='MIR92A2', Aliases='MIRN92-2, MIRN92A-2, MIRN92A2, mir-92a-2', description='microRNA 92a-2', other_designations='hsa-mir-92-2|hsa-mir-92a-2|microRNA 92-2', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134169538, end_position_on_the_genomic_accession=134169612, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  407050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407050, CurrentID=0, Status='live', Symbol='MIR93', Aliases='MIRN9, MIRN93, hsa-mir-93, miR-93', description='microRNA 93', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100093768, end_position_on_the_genomic_accession=100093847, orientation='minus', exon_count=1, OMIM=612984, no_hdr0=''),
  407051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407051, CurrentID=0, Status='live', Symbol='MIR9-3', Aliases='MIRN9-3, hsa-mir-9-3, miRNA9-3, mir-9-3', description='microRNA 9-3', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89368017, end_position_on_the_genomic_accession=89368106, orientation='plus', exon_count=1, OMIM=611188, no_hdr0=''),
  407052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407052, CurrentID=0, Status='live', Symbol='MIR95', Aliases='MIRN95, hsa-mir-95, miR-95', description='microRNA 95', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8005301, end_position_on_the_genomic_accession=8005381, orientation='minus', exon_count=1, OMIM=613185, no_hdr0=''),
  407053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407053, CurrentID=0, Status='live', Symbol='MIR96', Aliases='DFNA50, MIRN96, hsa-mir-96, miR-96, miRNA96', description='microRNA 96', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129774692, end_position_on_the_genomic_accession=129774769, orientation='minus', exon_count=1, OMIM=611606, no_hdr0=''),
  407054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407054, CurrentID=0, Status='live', Symbol='MIR98', Aliases='MIRN98, hsa-mir-98, miR-98', description='microRNA 98', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53556223, end_position_on_the_genomic_accession=53556341, orientation='minus', exon_count=1, OMIM=300810, no_hdr0=''),
  407055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407055, CurrentID=0, Status='live', Symbol='MIR99A', Aliases='MIRN99A, mir-99a', description='microRNA 99a', other_designations='hsa-mir-99a', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16539089, end_position_on_the_genomic_accession=16539169, orientation='plus', exon_count=1, OMIM=614509, no_hdr0=''),
  407056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407056, CurrentID=0, Status='live', Symbol='MIR99B', Aliases='MIRN99B, mir-99b', description='microRNA 99b', other_designations='hsa-mir-99b', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51692612, end_position_on_the_genomic_accession=51692681, orientation='plus', exon_count=1, OMIM=614510, no_hdr0=''),
  144920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144920, CurrentID=0, Status='live', Symbol='LINC00343', Aliases='LINC00344, NCRNA00343, NCRNA00344', description='long intergenic non-protein coding RNA 343', other_designations='long intergenic non-protein coding RNA 344', map_location='13q33.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105706830, end_position_on_the_genomic_accession=105761796, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370720, CurrentID=0, Status='live', Symbol='LOC105370720', Aliases='', description='uncharacterized LOC105370720', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266697, CurrentID=0, Status='live', Symbol='POM121L4P', Aliases='', description='POM121 transmembrane nucleoporin like 4, pseudogene', other_designations='POM121 membrane glycoprotein-like 4 pseudogene|POM121 transmembrane nucleoporin-like 4 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20689555, end_position_on_the_genomic_accession=20691721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2311, CurrentID=0, Status='live', Symbol='FOXO1B', Aliases='FKHRP1', description='forkhead box O1B pseudogene', other_designations='forkhead homolog (rhabdomyosarcoma) pseudogene 1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181098777, end_position_on_the_genomic_accession=181100878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372815, CurrentID=0, Status='live', Symbol='LOC105372815', Aliases='', description='uncharacterized LOC105372815', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42351223, end_position_on_the_genomic_accession=42363359, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  144983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144983, CurrentID=0, Status='live', Symbol='HNRNPA1L2', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1-like 2', other_designations='hnRNP A1-like 2|hnRNP core protein A1-like 2', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52617470, end_position_on_the_genomic_accession=52643784, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  266699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266699, CurrentID=0, Status='live', Symbol='FABP5P11', Aliases='FABP5L11, FABP5P1', description='fatty acid binding protein 5 pseudogene 11', other_designations='fatty acid binding protein 5, pseudogene 1|fatty acid binding protein 5-like 11 (pseudogene)', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16594871, end_position_on_the_genomic_accession=16595548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373077, CurrentID=0, Status='live', Symbol='LOC105373077', Aliases='', description='uncharacterized LOC105373077', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46734612, end_position_on_the_genomic_accession=46737726, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106182249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106182249, CurrentID=0, Status='live', Symbol='LUADT1', Aliases='', description='lung adenocarcinoma associated transcript 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270953, CurrentID=0, Status='live', Symbol='RPL21P33', Aliases='RPL21_5_222', description='ribosomal protein L21 pseudogene 33', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60852220, end_position_on_the_genomic_accession=60852785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377442, CurrentID=0, Status='live', Symbol='LOC105377442', Aliases='', description='uncharacterized LOC105377442', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137483552, end_position_on_the_genomic_accession=137486633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  133396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133396, CurrentID=0, Status='live', Symbol='IL31RA', Aliases='CRL, CRL3, GLM-R, GLMR, GPL, IL-31RA, PLCA2, PRO21384, hGLM-R', description='interleukin 31 receptor A', other_designations='IL-31 receptor subunit alpha|IL-31R subunit alpha|class I cytokine receptor|cytokine receptor-like 3|gp130-like monocyte receptor|soluble type I cytokine receptor CRL3|zcytoR17', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55840334, end_position_on_the_genomic_accession=55922854, orientation='plus', exon_count=21, OMIM=609510, no_hdr0=''),
  100616392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616392, CurrentID=0, Status='live', Symbol='MIR4773-1', Aliases='', description='microRNA 4773-1', other_designations='hsa-mir-4773-1', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151368334, end_position_on_the_genomic_accession=151368411, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131289, CurrentID=0, Status='live', Symbol='LOC100131289', Aliases='', description='uncharacterized LOC100131289', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27761744, end_position_on_the_genomic_accession=27763187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100270954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270954, CurrentID=0, Status='live', Symbol='RPL21P34', Aliases='RPL21_6_274', description='ribosomal protein L21 pseudogene 34', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128217553, end_position_on_the_genomic_accession=128218109, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132719, CurrentID=0, Status='live', Symbol='TUBB8P3', Aliases='', description='tubulin beta 8 class VIII pseudogene 3', other_designations='tubulin, beta 2C pseudogene|tubulin, beta 8 class VIII pseudogene', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112077599, end_position_on_the_genomic_accession=112078376, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372894, CurrentID=0, Status='live', Symbol='LOC105372894', Aliases='', description='uncharacterized LOC105372894', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208972453, end_position_on_the_genomic_accession=208975307, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100270955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270955, CurrentID=0, Status='live', Symbol='RPL21P35', Aliases='RPL21_8_333', description='ribosomal protein L21 pseudogene 35', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231559848, end_position_on_the_genomic_accession=231560397, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131291, CurrentID=0, Status='live', Symbol='ACTG1P21', Aliases='', description='actin gamma 1 pseudogene 21', other_designations='actin, gamma 1 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77773868, end_position_on_the_genomic_accession=77774879, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374640, CurrentID=0, Status='live', Symbol='LOC105374640', Aliases='', description='uncharacterized LOC105374640', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54515826, end_position_on_the_genomic_accession=54558211, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  220793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220793, CurrentID=0, Status='live', Symbol='FUNDC2P1', Aliases='FUNDC2P', description='FUN14 domain containing 2 pseudogene 1', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73508609, end_position_on_the_genomic_accession=73509715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646417, CurrentID=0, Status='live', Symbol='RPL29P5', Aliases='RPL29_2_63', description='ribosomal protein L29 pseudogene 5', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75582051, end_position_on_the_genomic_accession=75582672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  133418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133418, CurrentID=0, Status='live', Symbol='EMB', Aliases='GP70', description='embigin', other_designations='embigin homolog', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=50396192, end_position_on_the_genomic_accession=50443307, orientation='minus', exon_count=11, OMIM=615669, no_hdr0=''),
  145165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145165, CurrentID=0, Status='live', Symbol='ST13P4', Aliases='FAM10A4, FAM10A4P', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 4', other_designations='ST13-like tumor suppressor|family with sequence similarity 10, member A4 pseudogene', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50172018, end_position_on_the_genomic_accession=50173616, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729344, CurrentID=0, Status='live', Symbol='KRT8P24', Aliases='', description='keratin 8 pseudogene 24', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48971725, end_position_on_the_genomic_accession=48973564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145173, CurrentID=0, Status='live', Symbol='B3GLCT', Aliases='B3GALTL, B3GTL, B3Glc-T, Gal-T, beta3Glc-T', description='beta 3-glucosyltransferase', other_designations='UDP-GAL:beta-GlcNAc beta-1,3-galactosyltransferase-like|beta 1,3-galactosyltransferase-like|beta 3-glycosyltransferase-like|beta-3-glycosyltransferase-like', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31199975, end_position_on_the_genomic_accession=31332276, orientation='plus', exon_count=18, OMIM=610308, no_hdr0=''),
  131405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131405, CurrentID=0, Status='live', Symbol='TRIM71', Aliases='LIN-41, LIN41', description='tripartite motif containing 71, E3 ubiquitin protein ligase', other_designations='abnormal cell LINeage LIN-41|homolog of C. elegans Lin-41|protein lin-41 homolog|tripartite motif-containing 71|tripartite motif-containing protein 71', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32818018, end_position_on_the_genomic_accession=32892279, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723819, CurrentID=0, Status='live', Symbol='LOC102723819', Aliases='', description='uncharacterized LOC102723819', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38029294, end_position_on_the_genomic_accession=38041724, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131294, CurrentID=0, Status='live', Symbol='LOC100131294', Aliases='', description='RAB13, member RAS oncogene family pseudogene', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55979974, end_position_on_the_genomic_accession=55981067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373083, CurrentID=0, Status='live', Symbol='LOC105373083', Aliases='', description='uncharacterized LOC105373083', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48554618, end_position_on_the_genomic_accession=48559446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  145195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145195, CurrentID=0, Status='live', Symbol='C14orf144', Aliases='', description='chromosome 14 open reading frame 144', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=103775312, end_position_on_the_genomic_accession=103780535, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  145200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145200, CurrentID=0, Status='live', Symbol='LINC00239', Aliases='C14orf72, NCRNA00239', description='long intergenic non-protein coding RNA 239', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101730437, end_position_on_the_genomic_accession=101732525, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373084, CurrentID=0, Status='live', Symbol='LOC105373084', Aliases='', description='uncharacterized LOC105373084', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48518063, end_position_on_the_genomic_accession=48521265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372895, CurrentID=0, Status='live', Symbol='LOC105372895', Aliases='', description='uncharacterized LOC105372895', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209105592, end_position_on_the_genomic_accession=209114464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  145216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145216, CurrentID=0, Status='live', Symbol='LINC00637', Aliases='', description='long intergenic non-protein coding RNA 637', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103847721, end_position_on_the_genomic_accession=103858049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100270804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270804, CurrentID=0, Status='live', Symbol='LOC100270804', Aliases='', description='uncharacterized LOC100270804', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18794049, end_position_on_the_genomic_accession=18796065, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  145226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145226, CurrentID=0, Status='live', Symbol='RDH12', Aliases='LCA13, RP53, SDR7C2', description='retinol dehydrogenase 12 (all-trans/9-cis/11-cis)', other_designations='all-trans and 9-cis retinol dehydrogenase|retinol dehydrogenase 12, all-trans and 9-cis|short chain dehydrogenase/reductase family 7C member 2|short chain dehydrogenase/reductase family 7C, member 2', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67701886, end_position_on_the_genomic_accession=67734451, orientation='plus', exon_count=9, OMIM=608830, no_hdr0=''),
  145235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145235, CurrentID=0, Status='live', Symbol='RPL24P3', Aliases='RPL24_2_1385', description='ribosomal protein L24 pseudogene 3', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69561000, end_position_on_the_genomic_accession=69561513, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145241, CurrentID=0, Status='live', Symbol='ADAM21P1', Aliases='ADAM21P', description='ADAM metallopeptidase domain 21 pseudogene 1', other_designations='a disintegrin and metalloproteinase domain 21 pseudogene', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70245753, end_position_on_the_genomic_accession=70247801, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100270961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270961, CurrentID=0, Status='live', Symbol='RPL23P4', Aliases='RPL23_2_327', description='ribosomal protein L23 pseudogene 4', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219697813, end_position_on_the_genomic_accession=219698214, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376307, CurrentID=0, Status='live', Symbol='LOC105376307', Aliases='', description='uncharacterized LOC105376307', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133140690, end_position_on_the_genomic_accession=133145462, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  145249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145249, CurrentID=0, Status='live', Symbol='LINC00618', Aliases='C14orf63', description='long intergenic non-protein coding RNA 618', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96943579, end_position_on_the_genomic_accession=96945394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372990, CurrentID=0, Status='live', Symbol='LOC105372990', Aliases='', description='uncharacterized LOC105372990', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30431981, end_position_on_the_genomic_accession=30436462, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128093, CurrentID=0, Status='live', Symbol='LOC100128093', Aliases='', description='magnesium transporter 1 pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36080320, end_position_on_the_genomic_accession=36081334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79528, CurrentID=0, Status='live', Symbol='OR4R1P', Aliases='', description='olfactory receptor family 4 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily R, member 1 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48486187, end_position_on_the_genomic_accession=48487032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481451, CurrentID=0, Status='live', Symbol='RNU6-856P', Aliases='', description='RNA, U6 small nuclear 856, pseudogene', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145258, CurrentID=0, Status='live', Symbol='GSC', Aliases='SAMS', description='goosecoid homeobox', other_designations='', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94768223, end_position_on_the_genomic_accession=94770162, orientation='minus', exon_count=3, OMIM=138890, no_hdr0=''),
  145259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145259, CurrentID=0, Status='live', Symbol='RPSAP4', Aliases='LAMR1P4, RPSA_27_1403', description='ribosomal protein SA pseudogene 4', other_designations='laminin receptor 1 pseudogene 4', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94726212, end_position_on_the_genomic_accession=94727245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2365, CurrentID=0, Status='live', Symbol='FRA1F', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(q21)', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145264, CurrentID=0, Status='live', Symbol='SERPINA12', Aliases='OL-64', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 12', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 12|vaspin|visceral adipose tissue-derived serine protease inhibitor|visceral adipose-specific SERPIN', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94481651, end_position_on_the_genomic_accession=94517844, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  145270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145270, CurrentID=0, Status='live', Symbol='PRIMA1', Aliases='PRIMA', description='proline rich membrane anchor 1', other_designations='acetylcholinesterase membrane anchor precursor PRiMA', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93718298, end_position_on_the_genomic_accession=93789035, orientation='minus', exon_count=6, OMIM=613851, no_hdr0=''),
  100270962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270962, CurrentID=0, Status='live', Symbol='RPL23P10', Aliases='RPL23_2_893', description='ribosomal protein L23 pseudogene 10', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36581197, end_position_on_the_genomic_accession=36581588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2368, CurrentID=0, Status='live', Symbol='FRA1I', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(q44)', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145282, CurrentID=0, Status='live', Symbol='MIPOL1', Aliases='', description='mirror-image polydactyly 1', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=37197913, end_position_on_the_genomic_accession=37578857, orientation='plus', exon_count=25, OMIM=606850, no_hdr0=''),
  729348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729348, CurrentID=0, Status='live', Symbol='LOC729348', Aliases='', description='uncharacterized LOC729348', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66383306, end_position_on_the_genomic_accession=66392450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2369, CurrentID=0, Status='live', Symbol='FRA1J', Aliases='', description='fragile site, 5-azacytidine type, common, fra(1)(q12)', other_designations='', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286359, CurrentID=0, Status='live', Symbol='LOC286359', Aliases='', description='uncharacterized LOC286359', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97390836, end_position_on_the_genomic_accession=97396691, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  2371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2371, CurrentID=0, Status='live', Symbol='FRA1L', Aliases='', description='fragile site, aphidicolin type, common, fra(1)(p31)', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129395, CurrentID=0, Status='live', Symbol='LOC100129395', Aliases='', description='NS5ATP13TP1', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=36881761, end_position_on_the_genomic_accession=36886056, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100270963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270963, CurrentID=0, Status='live', Symbol='RPL24P7', Aliases='RPL24_2_348', description='ribosomal protein L24 pseudogene 7', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23134319, end_position_on_the_genomic_accession=23134867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2375, CurrentID=0, Status='live', Symbol='FRA2C', Aliases='', description='fragile site, aphidicolin type, common, fra(2)(p24.2)', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  319085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319085, CurrentID=0, Status='live', Symbol='ITPK1-AS1', Aliases='C14orf85, ITPK1-AS, ITPK1AS, NCRNA00203', description='ITPK1 antisense RNA 1', other_designations='ITPK1 antisense RNA (non-protein coding)|ITPK1 antisense RNA 1 (non-protein coding)', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93067452, end_position_on_the_genomic_accession=93072152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  220832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220832, CurrentID=0, Status='live', Symbol='FABP5P3', Aliases='FABP5L3', description='fatty acid binding protein 5 pseudogene 3', other_designations='Putative fatty acid-binding protein 5-like protein 3|TCAG_1781704|fatty acid binding protein 5-like 3 (pseudogene)', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152436895, end_position_on_the_genomic_accession=152443016, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100301516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301516, CurrentID=0, Status='live', Symbol='LOC100301516', Aliases='', description='acyl-CoA thioesterase 13 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148649476, end_position_on_the_genomic_accession=148650343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  89761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89761, CurrentID=0, Status='live', Symbol='NUFIP1P', Aliases='', description='nuclear fragile X mental retardation protein interacting protein 1 pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66091475, end_position_on_the_genomic_accession=66094945, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402425, CurrentID=0, Status='live', Symbol='LOC402425', Aliases='', description="5'-nucleotidase domain containing 1 pseudogene", other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134272287, end_position_on_the_genomic_accession=134273641, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270965, CurrentID=0, Status='live', Symbol='RPL26P15', Aliases='RPL26_2_213', description='ribosomal protein L26 pseudogene 15', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46003913, end_position_on_the_genomic_accession=46004421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145370, CurrentID=0, Status='live', Symbol='RPL21P7', Aliases='', description='ribosomal protein L21 pseudogene 7', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65266603, end_position_on_the_genomic_accession=65267732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145375, CurrentID=0, Status='live', Symbol='NUP50P1', Aliases='NUP50P', description='nucleoporin 50 pseudogene 1', other_designations='nucleoporin 50kDa pseudogene', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64596892, end_position_on_the_genomic_accession=64601951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145376, CurrentID=0, Status='live', Symbol='PPP1R36', Aliases='C14orf50', description='protein phosphatase 1 regulatory subunit 36', other_designations='protein phosphatase 1, regulatory subunit 36', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64549902, end_position_on_the_genomic_accession=64589386, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  145389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145389, CurrentID=0, Status='live', Symbol='SLC38A6', Aliases='NAT-1, SNAT6', description='solute carrier family 38 member 6', other_designations='N system amino acid transporter NAT-1|N-system amino acid transporter 1|Na(+)-coupled neutral amino acid transporter 6|solute carrier family 38, member 6', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60981114, end_position_on_the_genomic_accession=61083733, orientation='plus', exon_count=22, OMIM=616518, no_hdr0=''),
  2387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2387, CurrentID=0, Status='live', Symbol='FRA3D', Aliases='', description='fragile site, aphidicolin type, common, fra(3)(q25)', other_designations='', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285533, CurrentID=0, Status='live', Symbol='RNF175', Aliases='', description='ring finger protein 175', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153707672, end_position_on_the_genomic_accession=153760235, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  9206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9206, CurrentID=0, Status='live', Symbol='RAD17P2', Aliases='HRAD17P2', description='RAD17 checkpoint clamp loader component pseudogene 2', other_designations='RAD17 homolog (S. pombe) pseudogene 2|RAD17 homolog, pseudogene 2', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49043684, end_position_on_the_genomic_accession=49046869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319087, CurrentID=0, Status='live', Symbol='COX5AP2', Aliases='', description='cytochrome c oxidase subunit Va pseudogene 2', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52117516, end_position_on_the_genomic_accession=52118177, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145407, CurrentID=0, Status='live', Symbol='C14orf37', Aliases='c14_5376', description='chromosome 14 open reading frame 37', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57993531, end_position_on_the_genomic_accession=58152239, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100294316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294316, CurrentID=0, Status='live', Symbol='LOC100294316', Aliases='', description='protein FAM136A-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288425, CurrentID=0, Status='live', Symbol='EIF1P2', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 2', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38958278, end_position_on_the_genomic_accession=38958925, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270967, CurrentID=0, Status='live', Symbol='RPL26P26', Aliases='RPL26_7_907', description='ribosomal protein L26 pseudogene 26', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58113141, end_position_on_the_genomic_accession=58113665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729353, CurrentID=0, Status='live', Symbol='LOC729353', Aliases='', description='uncharacterized LOC729353', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145438, CurrentID=0, Status='live', Symbol='FRMD6-AS1', Aliases='C14orf82', description='FRMD6 antisense RNA 1', other_designations='FRMD6 antisense RNA 1 (non-protein coding)', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51649516, end_position_on_the_genomic_accession=51651744, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  145447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145447, CurrentID=0, Status='live', Symbol='ABHD12B', Aliases='BEM46L3, C14orf29, c14_5314', description='abhydrolase domain containing 12B', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50872090, end_position_on_the_genomic_accession=50904970, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  2400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2400, CurrentID=0, Status='live', Symbol='FRA5G', Aliases='', description='fragile site, folic acid type, rare, fra(5)(q35)', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145474, CurrentID=0, Status='live', Symbol='LOC145474', Aliases='', description='uncharacterized LOC145474', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71487861, end_position_on_the_genomic_accession=71489714, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  145482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145482, CurrentID=0, Status='live', Symbol='PTGR2', Aliases='HEL-S-298, PGR2, ZADH1', description='prostaglandin reductase 2', other_designations='15-oxoprostaglandin 13-reductase|15-oxoprostaglandin-delta13-reductase|PRG-2|epididymis secretory protein Li 298|zinc binding alcohol dehydrogenase, domain containing 1|zinc-binding alcohol dehydrogenase domain-containing protein 1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73851831, end_position_on_the_genomic_accession=73885465, orientation='plus', exon_count=10, OMIM=608642, no_hdr0=''),
  145483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145483, CurrentID=0, Status='live', Symbol='FAM161B', Aliases='C14orf44, c14_5547', description='family with sequence similarity 161 member B', other_designations='family with sequence similarity 161, member B', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73932031, end_position_on_the_genomic_accession=73950414, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100270969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270969, CurrentID=0, Status='live', Symbol='RPL27P5', Aliases='RPL27_2_231', description='ribosomal protein L27 pseudogene 5', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63108096, end_position_on_the_genomic_accession=63108429, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319090, CurrentID=0, Status='live', Symbol='OTSC6', Aliases='', description='otosclerosis 6', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369204, CurrentID=0, Status='live', Symbol='LOC105369204', Aliases='', description='uncharacterized LOC105369204', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71251239, end_position_on_the_genomic_accession=71328267, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  145497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145497, CurrentID=0, Status='live', Symbol='LRRC74A', Aliases='C14orf166B, LRRC74', description='leucine rich repeat containing 74A', other_designations='leucine rich repeat containing 74|leucine-rich repeat-containing protein 74', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76826376, end_position_on_the_genomic_accession=76870302, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  145499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145499, CurrentID=0, Status='live', Symbol='RPLP1P1', Aliases='RPLP1_4_1394', description='ribosomal protein, large, P1 pseudogene 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76968555, end_position_on_the_genomic_accession=76969183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145501, CurrentID=0, Status='live', Symbol='ISM2', Aliases='TAIL1, THSD3', description='isthmin 2', other_designations='isthmin 2 homolog|thrombospondin and AMOP containing isthmin-like 1|thrombospondin and AMOP domain-containing isthmin-like protein 1|thrombospondin type-1 domain-containing protein 3|thrombospondin, type I domain-containing 3|thrombospondin, type I, domain containing 3', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77474394, end_position_on_the_genomic_accession=77498867, orientation='minus', exon_count=8, OMIM=612684, no_hdr0=''),
  145508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145508, CurrentID=0, Status='live', Symbol='CEP128', Aliases='C14orf145, C14orf61, LEDP/132', description='centrosomal protein 128kDa', other_designations='lens epithelium-derived protein', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80476988, end_position_on_the_genomic_accession=80959501, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  100288428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288428, CurrentID=0, Status='live', Symbol='LMCD1-AS1', Aliases='', description='LMCD1 antisense RNA 1 (head to head)', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8221147, end_position_on_the_genomic_accession=8501658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644295, CurrentID=0, Status='live', Symbol='LOC644295', Aliases='', description='KIAA1143 pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32271545, end_position_on_the_genomic_accession=32272303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270970, CurrentID=0, Status='live', Symbol='RPL27P9', Aliases='RPL27_4_375', description='ribosomal protein L27 pseudogene 9', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58511310, end_position_on_the_genomic_accession=58511687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646430, CurrentID=0, Status='live', Symbol='LOC646430', Aliases='', description='amyloid P component, serum pseudogene', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159557585, end_position_on_the_genomic_accession=159558181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729356, CurrentID=0, Status='live', Symbol='PRAMEF29P', Aliases='', description='PRAME family member 29, pseudogene', other_designations='PRAME family member 1 pseudogene', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12926162, end_position_on_the_genomic_accession=12928882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2410, CurrentID=0, Status='live', Symbol='FRA7C', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(p14.2)', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270971, CurrentID=0, Status='live', Symbol='RPL28P2', Aliases='RPL28_1_335', description='ribosomal protein L28 pseudogene 2', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231789486, end_position_on_the_genomic_accession=231789886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2413, CurrentID=0, Status='live', Symbol='FRA7F', Aliases='', description='fragile site, aphidicolin type, common, fra(7)(q22)', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145553, CurrentID=0, Status='live', Symbol='MDP1', Aliases='FN6PASE, MDP-1', description='magnesium-dependent phosphatase 1', other_designations='fructosamine-6-phosphatase', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24213937, end_position_on_the_genomic_accession=24216070, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100301523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301523, CurrentID=0, Status='live', Symbol='STQTL12', Aliases='', description='Stature QTL 12', other_designations='', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612224, no_hdr0=''),
  145567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145567, CurrentID=0, Status='live', Symbol='TTC7B', Aliases='TTC7L1, c14_5685', description='tetratricopeptide repeat domain 7B', other_designations='TPR repeat protein 7-like-1|TPR repeat protein 7B|tetratricopeptide repeat domain 7 like 1|tetratricopeptide repeat protein 7-like-1', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90540588, end_position_on_the_genomic_accession=90816479, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  100421911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421911, CurrentID=0, Status='live', Symbol='LOC100421911', Aliases='', description='olfactory receptor, family 1, subfamily M, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376482, CurrentID=0, Status='live', Symbol='LOC105376482', Aliases='', description='uncharacterized LOC105376482', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31129939, end_position_on_the_genomic_accession=31134228, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  145581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145581, CurrentID=0, Status='live', Symbol='LRFN5', Aliases='C14orf146, FIGLER8, SALM5', description='leucine rich repeat and fibronectin type III domain containing 5', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 8', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=41607561, end_position_on_the_genomic_accession=41904549, orientation='plus', exon_count=7, OMIM=612811, no_hdr0=''),
  319093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319093, CurrentID=0, Status='live', Symbol='RHOQP1', Aliases='ARHQP, RHOQP', description='ras homolog family member Q pseudogene 1', other_designations='ras homolog gene family, member Q pseudogene 1', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49599673, end_position_on_the_genomic_accession=49600913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377461, CurrentID=0, Status='live', Symbol='LOC105377461', Aliases='', description='uncharacterized LOC105377461', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144201486, end_position_on_the_genomic_accession=144216119, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2419, CurrentID=0, Status='live', Symbol='FRA8B', Aliases='', description='fragile site, aphidicolin type, common, fra(8)(q22.1)', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533642, CurrentID=0, Status='live', Symbol='FNBP1P1', Aliases='', description='formin binding protein 1 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74120683, end_position_on_the_genomic_accession=74123216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100884169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100884169, CurrentID=0, Status='live', Symbol='LOC100884169', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A12 pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 12 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179784376, end_position_on_the_genomic_accession=179784901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  407738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407738, CurrentID=0, Status='live', Symbol='FAM19A1', Aliases='TAFA-1, TAFA1', description='family with sequence similarity 19 (chemokine (C-C motif)-like), member A1', other_designations='chemokine-like protein TAFA-1', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67991592, end_position_on_the_genomic_accession=68545621, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  107001063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=107001063, CurrentID=0, Status='live', Symbol='LOC107001063', Aliases='', description='adaptor related protein complex 3, sigma 1 subunit pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369406, CurrentID=0, Status='live', Symbol='LOC105369406', Aliases='', description='uncharacterized LOC105369406', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80137126, end_position_on_the_genomic_accession=80162669, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100301525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301525, CurrentID=0, Status='live', Symbol='DUH2', Aliases='', description='Dyschromatosis universalis hereditaria 2', other_designations='', map_location='12q21-q23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612715, no_hdr0=''),
  145624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145624, CurrentID=0, Status='live', Symbol='PWAR1', Aliases='D15S227E, PAR-1, PAR1', description='Prader Willi/Angelman region RNA 1', other_designations='Prader Willi/Angelman region 1', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25135642, end_position_on_the_genomic_accession=25138053, orientation='plus', exon_count=1, OMIM=600161, no_hdr0=''),
  100131143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131143, CurrentID=0, Status='live', Symbol='RPL23AP19', Aliases='RPL23A_2_8', description='ribosomal protein L23a pseudogene 19', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8831007, end_position_on_the_genomic_accession=8831378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270974, CurrentID=0, Status='live', Symbol='RPL29P14', Aliases='RPL29_7_626', description='ribosomal protein L29 pseudogene 14', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149545270, end_position_on_the_genomic_accession=149545844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106707175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106707175, CurrentID=0, Status='live', Symbol='ERVK-28', Aliases='', description='endogenous retrovirus group K member 28', other_designations='endogenous retrovirus group K, member 28', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145645, CurrentID=0, Status='live', Symbol='C15orf43', Aliases='', description='chromosome 15 open reading frame 43', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44956633, end_position_on_the_genomic_accession=44979223, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  276719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=276719, CurrentID=0, Status='live', Symbol='SOCS2P1', Aliases='', description='suppressor of cytokine signaling 2 pseudogene 1', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33187392, end_position_on_the_genomic_accession=33188293, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131310, CurrentID=0, Status='live', Symbol='LOC100131310', Aliases='', description='uncharacterized LOC100131310', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=39562417, end_position_on_the_genomic_accession=39610121, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373099, CurrentID=0, Status='live', Symbol='LOC105373099', Aliases='', description='hornerin-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50606999, end_position_on_the_genomic_accession=50610205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100270975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270975, CurrentID=0, Status='live', Symbol='RPL30P1', Aliases='RPL30_1_127', description='ribosomal protein L30 pseudogene 1', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174090134, end_position_on_the_genomic_accession=174090472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  2434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2434, CurrentID=0, Status='live', Symbol='FRA10D', Aliases='', description='fragile site, aphidicolin type, common, fra(10)(q22.1)', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145678, CurrentID=0, Status='live', Symbol='LOC145678', Aliases='', description='uncharacterized LOC145678', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  407835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407835, CurrentID=0, Status='live', Symbol='LOC407835', Aliases='', description='mitogen-activated protein kinase kinase 2 pseudogene', other_designations='mitogen-activated protein kinase kinase 2; protein kinase, mitogen-activated, kinase 2, p45 (MAP kinase kinase 2)', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129126271, end_position_on_the_genomic_accession=129128007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  145694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145694, CurrentID=0, Status='live', Symbol='LOC145694', Aliases='', description='uncharacterized LOC145694', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69396904, end_position_on_the_genomic_accession=69414476, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106707243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106707243, CurrentID=0, Status='live', Symbol='PRH1-TAS2R14', Aliases='', description='PRH1-TAS2R14 readthrough', other_designations='', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270977, CurrentID=0, Status='live', Symbol='RPL30P3', Aliases='RPL30_2_193', description='ribosomal protein L30 pseudogene 3', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9081393, end_position_on_the_genomic_accession=9081795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145741, CurrentID=0, Status='live', Symbol='C2CD4A', Aliases='FAM148A, NLF1', description='C2 calcium-dependent domain containing 4A', other_designations='family with sequence similarity 148, member A|nuclear localized factor 1|nuclear-localized factor 1', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62066977, end_position_on_the_genomic_accession=62070917, orientation='plus', exon_count=2, OMIM=610343, no_hdr0=''),
  145748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145748, CurrentID=0, Status='live', Symbol='LYSMD4', Aliases='', description='LysM, putative peptidoglycan-binding, domain containing 4', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99716804, end_position_on_the_genomic_accession=99733458, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100288436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288436, CurrentID=0, Status='live', Symbol='SLC4A1APP1', Aliases='', description='solute carrier family 4 (anion exchanger), member 1, adaptor protein pseudogene 1', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30558883, end_position_on_the_genomic_accession=30559481, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319099, CurrentID=0, Status='live', Symbol='DDX18P1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 1', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 18 pseudogene 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69083345, end_position_on_the_genomic_accession=69085714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145767, CurrentID=0, Status='live', Symbol='RPS3AP6', Aliases='RPS3A-like, RPS3A_20_1440', description='ribosomal protein S3A pseudogene 6', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59768330, end_position_on_the_genomic_accession=59769188, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145773, CurrentID=0, Status='live', Symbol='FAM81A', Aliases='', description='family with sequence similarity 81 member A', other_designations='family with sequence similarity 81, member A', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59402276, end_position_on_the_genomic_accession=59523552, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  145781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145781, CurrentID=0, Status='live', Symbol='GCOM1', Aliases='GRINL1A, Gcom2, MYZAP, MYZAP-POLR2M, gcom', description='GRINL1A complex locus 1', other_designations='GRINL1A combined protein Gcom12|MYZAP-POLR2M protein|MYZAP-POLR2M readthrough|NMDAR1 subunit-interacting protein|glutamate receptor, ionotropic, N-methyl D-aspartate-like 1A combined protein|myocardial zonula adherens protein', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57591904, end_position_on_the_genomic_accession=57717557, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  145783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145783, CurrentID=0, Status='live', Symbol='LOC145783', Aliases='', description='uncharacterized LOC145783', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56886170, end_position_on_the_genomic_accession=56918499, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288437, CurrentID=0, Status='live', Symbol='LOC100288437', Aliases='', description='ring finger and WD repeat domain 2, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106722350, end_position_on_the_genomic_accession=106722927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145788, CurrentID=0, Status='live', Symbol='C15orf65', Aliases='', description='chromosome 15 open reading frame 65', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55408525, end_position_on_the_genomic_accession=55418712, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  286442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286442, CurrentID=0, Status='live', Symbol='LINC01281', Aliases='', description='long intergenic non-protein coding RNA 1281', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39304956, end_position_on_the_genomic_accession=39327362, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377468, CurrentID=0, Status='live', Symbol='LOC105377468', Aliases='', description='uncharacterized LOC105377468', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146040415, end_position_on_the_genomic_accession=146059634, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  139334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139334, CurrentID=0, Status='live', Symbol='IMPDH1P4', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 4', other_designations="IMP (inosine 5'-monophosphate) dehydrogenase 2 pseudogene|inosine monophosphate dehydrogenase 1 pseudogene", map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43277376, end_position_on_the_genomic_accession=43279762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131315, CurrentID=0, Status='live', Symbol='LOC100131315', Aliases='', description='uncharacterized LOC100131315', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32718101, end_position_on_the_genomic_accession=32718865, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100415893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415893, CurrentID=0, Status='live', Symbol='DUPXQ28', Aliases='CXq28', description='Chromosome Xq28 duplication syndrome', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300815, no_hdr0=''),
  145814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145814, CurrentID=0, Status='live', Symbol='PGPEP1L', Aliases='', description='pyroglutamyl-peptidase I-like', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98968230, end_position_on_the_genomic_accession=99007795, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100415895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415895, CurrentID=0, Status='live', Symbol='GLM6', Aliases='', description='Glioma susceptibility 6', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613031, no_hdr0=''),
  100415896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415896, CurrentID=0, Status='live', Symbol='ASRT8', Aliases='', description='Asthma-related traits, susceptibility to, 8', other_designations='', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613207, no_hdr0=''),
  145820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145820, CurrentID=0, Status='live', Symbol='LINC00924', Aliases='', description='long intergenic non-protein coding RNA 924', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95433093, end_position_on_the_genomic_accession=95507847, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100415903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415903, CurrentID=0, Status='live', Symbol='HWE1', Aliases='', description='Epilepsy, hot water, 1', other_designations='', map_location='10q21.3-q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613339, no_hdr0=''),
  100415904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415904, CurrentID=0, Status='live', Symbol='HWE2', Aliases='', description='Epilepsy, hot water, 2', other_designations='', map_location='4q24-q28', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613340, no_hdr0=''),
  729366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729366, CurrentID=0, Status='live', Symbol='HNRNPA1P51', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 51', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206418974, end_position_on_the_genomic_accession=206420331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  407973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407973, CurrentID=0, Status='live', Symbol='SCA20', Aliases='C11DUPq12, DUP11q12', description='spinocerebellar ataxia 20', other_designations='', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608687, no_hdr0=''),
  407975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407975, CurrentID=0, Status='live', Symbol='MIR17HG', Aliases='C13orf25, FGLDS2, LINC00048, MIHG1, MIRH1, MIRHG1, NCRNA00048, miR-17-92', description='miR-17-92 cluster host gene', other_designations='MIR17 host gene (non-protein coding)|long intergenic non-protein coding RNA 48|miR-17-92 cluster host gene (non-protein coding)|microRNA host gene (non-protein coding) 1|microRNA host gene 1 (non-protein coding)|mir-17-92 microRNA cluster', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91347820, end_position_on_the_genomic_accession=91354575, orientation='plus', exon_count=2, OMIM=609415, no_hdr0=''),
  407977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=407977, CurrentID=0, Status='live', Symbol='TNFSF12-TNFSF13', Aliases='TWE-PRIL', description='TNFSF12-TNFSF13 readthrough', other_designations='tumor necrosis factor (ligand) superfamily, member 12-member 13', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7549058, end_position_on_the_genomic_accession=7561608, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929315, CurrentID=0, Status='live', Symbol='LOC101929315', Aliases='', description='uncharacterized LOC101929315', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24490312, end_position_on_the_genomic_accession=24514868, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  145837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145837, CurrentID=0, Status='live', Symbol='DRAIC', Aliases='', description='downregulated RNA in cancer, inhibitor of cell invasion and migration', other_designations='downregulated RNA in androgen independent cells', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69561720, end_position_on_the_genomic_accession=69571440, orientation='plus', exon_count=5, OMIM=616387, no_hdr0=''),
  105373105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373105, CurrentID=0, Status='live', Symbol='LOC105373105', Aliases='', description='uncharacterized LOC105373105', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=1732556, end_position_on_the_genomic_accession=1744502, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  145845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145845, CurrentID=0, Status='live', Symbol='LOC145845', Aliases='', description='uncharacterized LOC145845', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36864443, end_position_on_the_genomic_accession=36886533, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100270981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270981, CurrentID=0, Status='live', Symbol='RPL30P9', Aliases='RPL30_5_882', description='ribosomal protein L30 pseudogene 9', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20113336, end_position_on_the_genomic_accession=20113657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145853, CurrentID=0, Status='live', Symbol='C15orf61', Aliases='', description='chromosome 15 open reading frame 61', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67521184, end_position_on_the_genomic_accession=67527303, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  145858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145858, CurrentID=0, Status='live', Symbol='C15orf32', Aliases='', description='chromosome 15 open reading frame 32', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92471654, end_position_on_the_genomic_accession=92501117, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100415941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415941, CurrentID=0, Status='live', Symbol='DEL17Q23.1Q23.2', Aliases='C17DELq23.1q23.2', description='Chromosome 17q23.1-q23.2 deletion syndrome', other_designations='', map_location='17q23.1-q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613355, no_hdr0=''),
  100415942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415942, CurrentID=0, Status='live', Symbol='DEL2P21', Aliases='C2DELp21', description='Hypotonia-cystinuria syndrome', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606407, no_hdr0=''),
  100415943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415943, CurrentID=0, Status='live', Symbol='MGR9', Aliases='', description='Migraine with aura, susceptibility to, 9', other_designations='', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609670, no_hdr0=''),
  145864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145864, CurrentID=0, Status='live', Symbol='HAPLN3', Aliases='EXLD1, HsT19883', description='hyaluronan and proteoglycan link protein 3', other_designations='extracellular link domain containing, 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88877288, end_position_on_the_genomic_accession=88895626, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  548599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548599, CurrentID=0, Status='live', Symbol='PHBP7', Aliases='', description='prohibitin pseudogene 7', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85898831, end_position_on_the_genomic_accession=85900199, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289580, CurrentID=0, Status='live', Symbol='LOC100289580', Aliases='', description='uncharacterized LOC100289580', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88731180, end_position_on_the_genomic_accession=88741425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  145873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145873, CurrentID=0, Status='live', Symbol='MESP2', Aliases='SCDO2, bHLHc6', description='mesoderm posterior bHLH transcription factor 2', other_designations='class C basic helix-loop-helix protein 6|mesoderm posterior 2 homolog|mesoderm posterior basic helix-loop-helix transcription factor 2', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89776358, end_position_on_the_genomic_accession=89778754, orientation='plus', exon_count=2, OMIM=605195, no_hdr0=''),
  253635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253635, CurrentID=0, Status='live', Symbol='GPATCH11', Aliases='CCDC75, CENP-Y, CENPY', description='G-patch domain containing 11', other_designations='G patch domain containing 11|centromere protein Y|coiled-coil domain containing 75|coiled-coil domain-containing protein 75|g patch domain-containing protein 11', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37084451, end_position_on_the_genomic_accession=37099244, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  408029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408029, CurrentID=0, Status='live', Symbol='C2orf27B', Aliases='', description='chromosome 2 open reading frame 27B', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131794961, end_position_on_the_genomic_accession=131801661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  408050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408050, CurrentID=0, Status='live', Symbol='NOMO3', Aliases='Nomo', description='NODAL modulator 3', other_designations='pM5 protein 3|pM5 protein, middle copy', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16232532, end_position_on_the_genomic_accession=16294811, orientation='plus', exon_count=32, OMIM=609159, no_hdr0=''),
  26681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26681, CurrentID=0, Status='live', Symbol='OR4G1P', Aliases='OLB, OR4G8P', description='olfactory receptor family 4 subfamily G member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily G, member 1 pseudogene|olfactory receptor, family 4, subfamily G, member 8 pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=104602, end_position_on_the_genomic_accession=105470, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729369, CurrentID=0, Status='live', Symbol='LOC729369', Aliases='', description='ankyrin repeat domain-containing protein 36B-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  220929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220929, CurrentID=0, Status='live', Symbol='ZNF438', Aliases='bA330O11.1', description='zinc finger protein 438', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30844634, end_position_on_the_genomic_accession=31032407, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  89858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89858, CurrentID=0, Status='live', Symbol='SIGLEC12', Aliases='S2V, SIGLECL1, SLG, Siglec-XII', description='sialic acid binding Ig-like lectin 12 (gene/pseudogene)', other_designations='SIGLEC-like 1|sialic acid-binding Ig-like lectin-like 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51491227, end_position_on_the_genomic_accession=51501789, orientation='minus', exon_count=8, OMIM=606094, no_hdr0=''),
  100270984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270984, CurrentID=0, Status='live', Symbol='RPL31P15', Aliases='RPL31_3_303', description='ribosomal protein L31 pseudogene 15', other_designations='', map_location='2qter', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182788073, end_position_on_the_genomic_accession=182788475, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  145942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145942, CurrentID=0, Status='live', Symbol='TMCO5A', Aliases='TMCO5', description='transmembrane and coiled-coil domains 5A', other_designations='transmembrane and coiled-coil domains 5', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=37934604, end_position_on_the_genomic_accession=37995033, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  145945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145945, CurrentID=0, Status='live', Symbol='LOC145945', Aliases='', description='uncharacterized LOC145945', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145946, CurrentID=0, Status='live', Symbol='SPATA8', Aliases='SRG8', description='spermatogenesis associated 8', other_designations='spermatogenesis-related protein 8', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96783407, end_position_on_the_genomic_accession=96785615, orientation='plus', exon_count=4, OMIM=613948, no_hdr0=''),
  145957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145957, CurrentID=0, Status='live', Symbol='NRG4', Aliases='HRG4', description='neuregulin 4', other_designations='heregulin 4|pro-NRG4', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75941973, end_position_on_the_genomic_accession=76012444, orientation='minus', exon_count=6, OMIM=610894, no_hdr0=''),
  105373109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373109, CurrentID=0, Status='live', Symbol='LOC105373109', Aliases='', description='uncharacterized LOC105373109', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224883788, end_position_on_the_genomic_accession=224929574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270985, CurrentID=0, Status='live', Symbol='RPL31P19', Aliases='RPL31_5_346', description='ribosomal protein L31 pseudogene 19', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17465131, end_position_on_the_genomic_accession=17465548, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286472, CurrentID=0, Status='live', Symbol='RAC1P4', Aliases='Psi4Rac1, RAC4', description='ras-related C3 botulinum toxin substrate 1 pseudogene 4', other_designations='ras-related C3 botulinum toxin substrate family member 4', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=137441047, end_position_on_the_genomic_accession=137442031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  145978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145978, CurrentID=0, Status='live', Symbol='LINC00052', Aliases='NCRNA00052, TMEM83', description='long intergenic non-protein coding RNA 52', other_designations='transmembrane protein 83', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=87576929, end_position_on_the_genomic_accession=87579686, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481698, CurrentID=0, Status='live', Symbol='ECM1P2', Aliases='', description='extracellular matrix protein 1 pseudogene 2', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  145989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145989, CurrentID=0, Status='live', Symbol='FSCN1P1', Aliases='', description='fascin actin-bundling protein 1 pseudogene 1', other_designations='fascin homolog 1, actin-bundling protein pseudogene|fascin pseudogene', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34494296, end_position_on_the_genomic_accession=34497078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376308, CurrentID=0, Status='live', Symbol='LOC105376308', Aliases='', description='uncharacterized LOC105376308', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133663964, end_position_on_the_genomic_accession=133664749, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372991, CurrentID=0, Status='live', Symbol='LOC105372991', Aliases='', description='uncharacterized LOC105372991', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30447952, end_position_on_the_genomic_accession=30452774, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79533, CurrentID=0, Status='live', Symbol='OR52E3P', Aliases='', description='olfactory receptor family 52 subfamily E member 3 pseudogene', other_designations='olfactory receptor, family 52, subfamily E, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5092676, end_position_on_the_genomic_accession=5093650, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390729, CurrentID=0, Status='live', Symbol='RPL10P14', Aliases='', description='ribosomal protein L10 pseudogene 14', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50151219, end_position_on_the_genomic_accession=50151585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421737, CurrentID=0, Status='live', Symbol='LOC100421737', Aliases='', description='ubiquitin specific peptidase 4 (proto-oncogene) pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46851578, end_position_on_the_genomic_accession=46854788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373111, CurrentID=0, Status='live', Symbol='LOC105373111', Aliases='', description='uncharacterized LOC105373111', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225914990, end_position_on_the_genomic_accession=225917992, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100270987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270987, CurrentID=0, Status='live', Symbol='RPL31P25', Aliases='RPL31_6_464', description='ribosomal protein L31 pseudogene 25', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21656523, end_position_on_the_genomic_accession=21656848, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  408186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408186, CurrentID=0, Status='live', Symbol='OVOS', Aliases='', description='ovostatin', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9517803, end_position_on_the_genomic_accession=9573297, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  408187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408187, CurrentID=0, Status='live', Symbol='SPINK14', Aliases='SPINK5L2', description='serine peptidase inhibitor, Kazal type 14 (putative)', other_designations='Kazal type serine protease inhibitor 5-like 2', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148168546, end_position_on_the_genomic_accession=148175619, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  408188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408188, CurrentID=0, Status='live', Symbol='VTLGL1', Aliases='', description='vitellogenin-like 1', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  146050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146050, CurrentID=0, Status='live', Symbol='ZSCAN29', Aliases='ZNF690, Zfp690', description='zinc finger and SCAN domain containing 29', other_designations='KOX31-like zinc finger protein|zinc finger protein 690', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43358172, end_position_on_the_genomic_accession=43370648, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  146053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146053, CurrentID=0, Status='live', Symbol='RPS3AP47', Aliases='RPS3A_22_1428', description='ribosomal protein S3a pseudogene 47', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43115655, end_position_on_the_genomic_accession=43116517, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  133569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133569, CurrentID=0, Status='live', Symbol='RPS10P12', Aliases='RPS10_5_559', description='ribosomal protein S10 pseudogene 12', other_designations='', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51275217, end_position_on_the_genomic_accession=51275780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146057, CurrentID=0, Status='live', Symbol='TTBK2', Aliases='SCA11, TTBK', description='tau tubulin kinase 2', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42738734, end_position_on_the_genomic_accession=42921007, orientation='minus', exon_count=19, OMIM=611695, no_hdr0=''),
  100270988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270988, CurrentID=0, Status='live', Symbol='RPL31P21', Aliases='RPL31_8_410', description='ribosomal protein L31 pseudogene 21', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141258418, end_position_on_the_genomic_accession=141258779, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146059, CurrentID=0, Status='live', Symbol='CDAN1', Aliases='CDA1, CDAIA, DLT, PRO1295, CDAN1', description='codanin 1', other_designations='congenital dyserythropoietic anemia, type I|discs lost homolog', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42723557, end_position_on_the_genomic_accession=42737219, orientation='minus', exon_count=28, OMIM=607465, no_hdr0=''),
  408221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408221, CurrentID=0, Status='live', Symbol='SCA26', Aliases='', description='spinocerebellar ataxia 26', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609306, no_hdr0=''),
  388942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388942, CurrentID=0, Status='live', Symbol='LOC388942', Aliases='', description='uncharacterized LOC388942', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41877555, end_position_on_the_genomic_accession=41894046, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131146, CurrentID=0, Status='live', Symbol='LOC100131146', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C8 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8 pseudogene', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141968094, end_position_on_the_genomic_accession=141968841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371155, CurrentID=0, Status='live', Symbol='LOC105371155', Aliases='', description='uncharacterized LOC105371155', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27151310, end_position_on_the_genomic_accession=27153864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270989, CurrentID=0, Status='live', Symbol='RPL31P22', Aliases='RPL31_9_449', description='ribosomal protein L31 pseudogene 22', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195349675, end_position_on_the_genomic_accession=195350030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370547, CurrentID=0, Status='live', Symbol='LOC105370547', Aliases='', description='uncharacterized LOC105370547', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929507, CurrentID=0, Status='live', Symbol='LOC101929507', Aliases='', description='uncharacterized LOC101929507', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13808298, end_position_on_the_genomic_accession=14031280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  220959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220959, CurrentID=0, Status='live', Symbol='KRT8P16', Aliases='', description='keratin 8 pseudogene 16', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8536736, end_position_on_the_genomic_accession=8542601, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146110, CurrentID=0, Status='live', Symbol='RPL7P47', Aliases='RPL7_27_1478', description='ribosomal protein L7 pseudogene 47', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17932717, end_position_on_the_genomic_accession=17933345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  408257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408257, CurrentID=0, Status='live', Symbol='CILD4', Aliases='', description='Ciliary dyskinesia, primary, 4', other_designations='', map_location='15q13.1-q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608646, no_hdr0=''),
  408259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408259, CurrentID=0, Status='live', Symbol='HPC3', Aliases='HPC20', description='Prostate cancer, hereditary, 3', other_designations='', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608656, no_hdr0=''),
  408260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408260, CurrentID=0, Status='live', Symbol='HPC4', Aliases='HPC7', description='Prostate cancer, hereditary, 4', other_designations='', map_location='7p11-q21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608658, no_hdr0=''),
  408263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=408263, CurrentID=0, Status='live', Symbol='FNDC9', Aliases='C5orf40', description='fibronectin type III domain containing 9', other_designations='fibronectin type-III domain-containing protein C5orf40', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157341599, end_position_on_the_genomic_accession=157345721, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319111, CurrentID=0, Status='live', Symbol='DNAJC8P1', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C8 pseudogene 1', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8 pseudogene 1', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35429750, end_position_on_the_genomic_accession=35431326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  220963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220963, CurrentID=0, Status='live', Symbol='SLC16A9', Aliases='C10orf36, MCT9', description='solute carrier family 16 member 9', other_designations='MCT 9|monocarboxylic acid transporter 9|solute carrier family 16 (monocarboxylic acid transporters), member 9|solute carrier family 16, member 9 (monocarboxylic acid transporter 9)', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59650764, end_position_on_the_genomic_accession=59709891, orientation='minus', exon_count=7, OMIM=614242, no_hdr0=''),
  133584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133584, CurrentID=0, Status='live', Symbol='EGFLAM', Aliases='AGRINL, AGRNL, PIKA', description='EGF-like, fibronectin type III and laminin G domains', other_designations='EGF-like, fibronectin type-III and laminin G-like domain-containing protein|agrin-like protein', map_location='5p13.2-p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38258409, end_position_on_the_genomic_accession=38465480, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  100270991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270991, CurrentID=0, Status='live', Symbol='RPL32P8', Aliases='RPL32_2_419', description='ribosomal protein L32 pseudogene 8', other_designations='', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149433043, end_position_on_the_genomic_accession=149433440, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79534, CurrentID=0, Status='live', Symbol='OR4V1P', Aliases='', description='olfactory receptor family 4 subfamily V member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily V, member 1 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55673537, end_position_on_the_genomic_accession=55674676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873295, CurrentID=0, Status='live', Symbol='RNA5SP60', Aliases='RN5S60', description='RNA, 5S ribosomal pseudogene 60', other_designations='RNA, 5S ribosomal 60', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159178473, end_position_on_the_genomic_accession=159178559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146167, CurrentID=0, Status='live', Symbol='SLC38A8', Aliases='FVH2', description='solute carrier family 38 member 8', other_designations='solute carrier family 38, member 8', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84009667, end_position_on_the_genomic_accession=84043097, orientation='minus', exon_count=12, OMIM=615585, no_hdr0=''),
  79064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79064, CurrentID=0, Status='live', Symbol='TMEM223', Aliases='', description='transmembrane protein 223', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62790315, end_position_on_the_genomic_accession=62792014, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101202686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101202686, CurrentID=0, Status='live', Symbol='ARCI7', Aliases='', description='Ichthyosis, congenital, autosomal recessive 7', other_designations='', map_location='12p11.2-q13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615022, no_hdr0=''),
  101202687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101202687, CurrentID=0, Status='live', Symbol='ARCI9', Aliases='', description='Ichthyosis, congenital, autosomal recessive 9', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615023, no_hdr0=''),
  146177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146177, CurrentID=0, Status='live', Symbol='VWA3A', Aliases='', description='von Willebrand factor A domain containing 3A', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22091914, end_position_on_the_genomic_accession=22156966, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  100270992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270992, CurrentID=0, Status='live', Symbol='RPL32P9', Aliases='RPL32_3_420', description='ribosomal protein L32 pseudogene 9', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150150538, end_position_on_the_genomic_accession=150150947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146183, CurrentID=0, Status='live', Symbol='OTOA', Aliases='CT108, DFNB22', description='otoancorin', other_designations='cancer/testis antigen 108', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21664155, end_position_on_the_genomic_accession=21760729, orientation='plus', exon_count=31, OMIM=607038, no_hdr0=''),
  102775564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102775564, CurrentID=0, Status='live', Symbol='MRX89', Aliases='', description='Mental retardation, X-linked 89', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300848, no_hdr0=''),
  102775565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102775565, CurrentID=0, Status='live', Symbol='MRX92', Aliases='', description='Mental retardation, X-linked 92', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300851, no_hdr0=''),
  102775567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102775567, CurrentID=0, Status='live', Symbol='DDH2', Aliases='', description='Developmental dysplasia of the hip 2', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615612, no_hdr0=''),
  102775568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102775568, CurrentID=0, Status='live', Symbol='MRX95', Aliases='', description='Mental retardation, X-linked 95', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300716, no_hdr0=''),
  146198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146198, CurrentID=0, Status='live', Symbol='ZFP90', Aliases='FIK, NK10, ZNF756, zfp-90', description='ZFP90 zinc finger protein', other_designations='FOXP3-interacting KRAB domain-containing protein|zinc finger protein 476|zinc finger protein 756', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68539289, end_position_on_the_genomic_accession=68576072, orientation='plus', exon_count=6, OMIM=609451, no_hdr0=''),
  146206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146206, CurrentID=0, Status='live', Symbol='RLTPR', Aliases='CARMIL2, CARMIL2b, LRRC16C', description='RGD motif, leucine rich repeats, tropomodulin domain and proline-rich containing', other_designations='RGD, leucine-rich repeat, tropomodulin and proline-rich containing protein|RGD, leucine-rich repeat, tropomodulin and proline-rich-containing protein|leucine rich repeat containing 16C', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67645004, end_position_on_the_genomic_accession=67657569, orientation='plus', exon_count=39, OMIM=610859, no_hdr0=''),
  100270993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270993, CurrentID=0, Status='live', Symbol='RPL32P13', Aliases='RPL32_4_510', description='ribosomal protein L32 pseudogene 13', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112738675, end_position_on_the_genomic_accession=112739106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146212, CurrentID=0, Status='live', Symbol='KCTD19', Aliases='', description='potassium channel tetramerization domain containing 19', other_designations='potassium channel tetramerisation domain containing 19', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67289432, end_position_on_the_genomic_accession=67326761, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  319114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319114, CurrentID=0, Status='live', Symbol='ENSAP2', Aliases='', description='endosulfine alpha pseudogene 2', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=82692507, end_position_on_the_genomic_accession=82693198, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146223, CurrentID=0, Status='live', Symbol='CMTM4', Aliases='CKLFSF4', description='CKLF-like MARVEL transmembrane domain containing 4', other_designations='chemokine-like factor super family 4|chemokine-like factor superfamily member 4', map_location='16q21-q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66598214, end_position_on_the_genomic_accession=66696707, orientation='minus', exon_count=7, OMIM=607887, no_hdr0=''),
  146225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146225, CurrentID=0, Status='live', Symbol='CMTM2', Aliases='CKLFSF2', description='CKLF-like MARVEL transmembrane domain containing 2', other_designations='chemokine-like factor superfamily 2|chemokine-like factor superfamily member 2', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66579448, end_position_on_the_genomic_accession=66588275, orientation='plus', exon_count=4, OMIM=607885, no_hdr0=''),
  146227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146227, CurrentID=0, Status='live', Symbol='BEAN1', Aliases='BEAN, SCA31', description='brain expressed, associated with NEDD4, 1', other_designations='brain-expressed protein associating with Nedd4 homolog', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66427297, end_position_on_the_genomic_accession=66493529, orientation='plus', exon_count=16, OMIM=612051, no_hdr0=''),
  401634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401634, CurrentID=0, Status='live', Symbol='GOLGA2P3Y', Aliases='GOLGA2LY2, GOLGA2P3', description='golgin A2 pseudogene 3, Y-linked', other_designations='golgi autoantigen, golgin subfamily a, 2-like, Y-linked 2|golgi autoantigen, golgin subfamily a, 2-like, Y-linked, telomeric|golgin A2-like, Y-linked 2 (pseudogene)', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25455311, end_position_on_the_genomic_accession=25460175, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100270994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270994, CurrentID=0, Status='live', Symbol='RPL32P14', Aliases='RPL32_5_546', description='ribosomal protein L32 pseudogene 14', other_designations='', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=19041036, end_position_on_the_genomic_accession=19041368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319115, CurrentID=0, Status='live', Symbol='EIF2S2P1', Aliases='EIF2S2P', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 1', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 1|eukaryotic translation initiation factor 2, subunit 2 beta, 38kDa pseudogene|eukaryotic translation initiation factor 2, subunit 2 beta, pseudogene', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63734675, end_position_on_the_genomic_accession=63736219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377483, CurrentID=0, Status='live', Symbol='LOC105377483', Aliases='', description='uncharacterized LOC105377483', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148615116, end_position_on_the_genomic_accession=148680767, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  146253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146253, CurrentID=0, Status='live', Symbol='LOC146253', Aliases='', description='tropomyosin 3 pseudogene', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52655233, end_position_on_the_genomic_accession=52657320, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4870, CurrentID=0, Status='live', Symbol='NPM1P1', Aliases='NG1-1.6, NPMP1', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 1', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 1', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41788907, end_position_on_the_genomic_accession=41789396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270995, CurrentID=0, Status='live', Symbol='RPL34P5', Aliases='RPL34_2_101', description='ribosomal protein L34 pseudogene 5', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153995632, end_position_on_the_genomic_accession=153995963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319116, CurrentID=0, Status='live', Symbol='ATP6V1G1P1', Aliases='ATP6V1GP1', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 1', other_designations='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G pseudogene 1', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105784496, end_position_on_the_genomic_accession=105785972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377705, CurrentID=0, Status='live', Symbol='LOC105377705', Aliases='', description='uncharacterized LOC105377705', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166230129, end_position_on_the_genomic_accession=166233375, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  146279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146279, CurrentID=0, Status='live', Symbol='TEKT5', Aliases='CT149', description='tektin 5', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10581369, end_position_on_the_genomic_accession=10742416, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105371071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371071, CurrentID=0, Status='live', Symbol='LOC105371071', Aliases='', description='uncharacterized LOC105371071', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8296494, end_position_on_the_genomic_accession=8299377, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373120, CurrentID=0, Status='live', Symbol='LOC105373120', Aliases='', description='uncharacterized LOC105373120', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226896888, end_position_on_the_genomic_accession=226909737, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  133609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133609, CurrentID=0, Status='live', Symbol='RPLP1P7', Aliases='RPLP1_4_611', description='ribosomal protein, large, P1 pseudogene 7', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126627979, end_position_on_the_genomic_accession=126628448, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060159, CurrentID=0, Status='live', Symbol='LOC101060159', Aliases='', description='putative uncharacterized protein C3orf56', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127220325, end_position_on_the_genomic_accession=127221985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377485, CurrentID=0, Status='live', Symbol='LOC105377485', Aliases='', description='uncharacterized LOC105377485', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151168034, end_position_on_the_genomic_accession=151171934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  146310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146310, CurrentID=0, Status='live', Symbol='RNF151', Aliases='', description='ring finger protein 151', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1966823, end_position_on_the_genomic_accession=1968975, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  146325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146325, CurrentID=0, Status='live', Symbol='PRR35', Aliases='C16orf11, LA16c-366D1.2', description='proline rich 35', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=560422, end_position_on_the_genomic_accession=565529, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  146330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146330, CurrentID=0, Status='live', Symbol='FBXL16', Aliases='C16orf22, Fbl16, c380A1.1', description='F-box and leucine-rich repeat protein 16', other_designations='c380A1.1 (novel protein)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=692500, end_position_on_the_genomic_accession=705825, orientation='minus', exon_count=6, OMIM=609082, no_hdr0=''),
  133615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133615, CurrentID=0, Status='live', Symbol='MRPS5P3', Aliases='', description='mitochondrial ribosomal protein S5 pseudogene 3', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127142479, end_position_on_the_genomic_accession=127143959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146336, CurrentID=0, Status='live', Symbol='SSTR5-AS1', Aliases='', description='SSTR5 antisense RNA 1', other_designations='SSTR5 antisense RNA 1 (non-protein coding)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1064081, end_position_on_the_genomic_accession=1078731, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  133619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133619, CurrentID=0, Status='live', Symbol='PRRC1', Aliases='', description='proline-rich coiled-coil 1', other_designations='proline-rich and coiled-coil-containing protein 1', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127517609, end_position_on_the_genomic_accession=127555089, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  319119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319119, CurrentID=0, Status='live', Symbol='ATP5G2P2', Aliases='ATP5GP2', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9) pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C2 (subunit 9) pseudogene 2|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9) pseudogene 2|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9) pseudogene 2|pseudogene 2 for mitochondrial ATP synthase c subunit (P2 form)', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49373912, end_position_on_the_genomic_accession=49374534, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146378, CurrentID=0, Status='live', Symbol='C16orf92', Aliases='', description='chromosome 16 open reading frame 92', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30023334, end_position_on_the_genomic_accession=30024702, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100147822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147822, CurrentID=0, Status='live', Symbol='RNU7-35P', Aliases='U7.35', description='RNA, U7 small nuclear 35 pseudogene', other_designations='RNA, small nuclear U7.35 pseudogene', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43194422, end_position_on_the_genomic_accession=43194482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270999, CurrentID=0, Status='live', Symbol='RPL34P8', Aliases='RPL34_5_265', description='ribosomal protein L34 pseudogene 8', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111675026, end_position_on_the_genomic_accession=111675343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146395, CurrentID=0, Status='live', Symbol='GSG1L', Aliases='PRO19651', description='GSG1-like', other_designations='GSG1-like protein|KTSR5831', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27787529, end_position_on_the_genomic_accession=28063509, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  133629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133629, CurrentID=0, Status='live', Symbol='RPL11P2', Aliases='RPL11_1_612', description='ribosomal protein L11 pseudogene 2', other_designations='', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130989879, end_position_on_the_genomic_accession=130990471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271000, CurrentID=0, Status='live', Symbol='RPL34P15', Aliases='RPL34_5_775', description='ribosomal protein L34 pseudogene 15', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162957249, end_position_on_the_genomic_accession=162957548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319121, CurrentID=0, Status='live', Symbol='YWHAQP1', Aliases='YWHAQP', description='YWHAQ pseudogene 1', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, theta polypeptide pseudogene|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, theta pseudogene', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=42807117, end_position_on_the_genomic_accession=42808021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146429, CurrentID=0, Status='live', Symbol='SLC22A31', Aliases='', description='solute carrier family 22 member 31', other_designations='putative solute carrier family 22 member ENSG00000182157|solute carrier family 22, member 31', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89195756, end_position_on_the_genomic_accession=89201671, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  146433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146433, CurrentID=0, Status='live', Symbol='IL34', Aliases='C16orf77, IL-34', description='interleukin 34', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70579895, end_position_on_the_genomic_accession=70660684, orientation='plus', exon_count=11, OMIM=612081, no_hdr0=''),
  146434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146434, CurrentID=0, Status='live', Symbol='ZNF597', Aliases='HIT-4', description='zinc finger protein 597', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3432422, end_position_on_the_genomic_accession=3443537, orientation='minus', exon_count=4, OMIM=614685, no_hdr0=''),
  146439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146439, CurrentID=0, Status='live', Symbol='CCDC64B', Aliases='BICDR-2', description='coiled-coil domain containing 64B', other_designations='BICD-related protein 2|coiled-coil domain-containing protein 64B', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3027710, end_position_on_the_genomic_accession=3036963, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105373125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373125, CurrentID=0, Status='live', Symbol='LOC105373125', Aliases='', description='uncharacterized LOC105373125', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8923781, end_position_on_the_genomic_accession=8927194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271001, CurrentID=0, Status='live', Symbol='RPL34P17', Aliases='RPL34_6_902', description='ribosomal protein L34 pseudogene 17', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52313153, end_position_on_the_genomic_accession=52313494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285540, CurrentID=0, Status='live', Symbol='SEPSECS-AS1', Aliases='', description='SEPSECS antisense RNA 1 (head to head)', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25160672, end_position_on_the_genomic_accession=25198506, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319122, CurrentID=0, Status='live', Symbol='UNGP3', Aliases='', description='uracil-DNA glycosylase pseudogene 3', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81258328, end_position_on_the_genomic_accession=81260234, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  221017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221017, CurrentID=0, Status='live', Symbol='CEP57L1P1', Aliases='C10orf37, C6orf182P', description='centrosomal protein 57kDa-like 1 pseudogene 1', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70389366, end_position_on_the_genomic_accession=70390620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146456, CurrentID=0, Status='live', Symbol='TMED6', Aliases='PRO34237, SPLL9146', description='transmembrane p24 trafficking protein 6', other_designations='p24 family protein gamma-5|p24gamma5|transmembrane emp24 protein transport domain containing 6', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69343246, end_position_on_the_genomic_accession=69351809, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929519, CurrentID=0, Status='live', Symbol='LOC101929519', Aliases='', description='mucin-2-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93667342, end_position_on_the_genomic_accession=93673250, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  388955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388955, CurrentID=0, Status='live', Symbol='LOC388955', Aliases='', description='PRELI domain-containing protein 1, mitochondrial pseudogene', other_designations='PRELI domain containing 1 pseudogene|PX19 protein pseudogene', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63622084, end_position_on_the_genomic_accession=63623027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271002, CurrentID=0, Status='live', Symbol='RPL34P9', Aliases='RPL34_7_396', description='ribosomal protein L34 pseudogene 9', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120512854, end_position_on_the_genomic_accession=120513189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146481, CurrentID=0, Status='live', Symbol='FRG2DP', Aliases='', description='FSHD region gene 2 family member D, pseudogene', other_designations='FSHD region gene 2 family, member C pseudogene|FSHD region gene 2 family, member D, pseudogene', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35477414, end_position_on_the_genomic_accession=35480596, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  319123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319123, CurrentID=0, Status='live', Symbol='ZFP64P1', Aliases='ZNF405P', description='ZFP64 zinc finger protein pseudogene 1', other_designations='zinc finger protein 405, pseudogene|zinc finger protein 64 homolog pseudogene 1', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50703994, end_position_on_the_genomic_accession=50705511, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373127, CurrentID=0, Status='live', Symbol='LOC105373127', Aliases='', description='uncharacterized LOC105373127', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=10949875, end_position_on_the_genomic_accession=10964637, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271003, CurrentID=0, Status='live', Symbol='RPL34P18', Aliases='RPL34_7_926', description='ribosomal protein L34 pseudogene 18', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93957338, end_position_on_the_genomic_accession=93957691, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146512, CurrentID=0, Status='live', Symbol='FLJ30679', Aliases='', description='uncharacterized protein FLJ30679', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86555320, end_position_on_the_genomic_accession=86557299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  146513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146513, CurrentID=0, Status='live', Symbol='LOC146513', Aliases='', description='uncharacterized LOC146513', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86286431, end_position_on_the_genomic_accession=86293389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340947, CurrentID=0, Status='live', Symbol='EIF3LP3', Aliases='', description='eukaryotic translation initiation factor 3 subunit L pseudogene 3', other_designations='eukaryotic translation initiation factor 3, subunit L pseudogene 3', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38080059, end_position_on_the_genomic_accession=38080871, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271004, CurrentID=0, Status='live', Symbol='RPL34P10', Aliases='RPL34_8_445', description='ribosomal protein L34 pseudogene 10', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185606575, end_position_on_the_genomic_accession=185606874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146540, CurrentID=0, Status='live', Symbol='ZNF785', Aliases='', description='zinc finger protein 785', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30573219, end_position_on_the_genomic_accession=30586421, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  146542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146542, CurrentID=0, Status='live', Symbol='ZNF688', Aliases='', description='zinc finger protein 688', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30569677, end_position_on_the_genomic_accession=30572495, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  146547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146547, CurrentID=0, Status='live', Symbol='PRSS36', Aliases='', description='protease, serine 36', other_designations='polyserase 2|polyserine protease 2|polyserine protease-2|protease, serine, 36|serine protease 36', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31138923, end_position_on_the_genomic_accession=31150094, orientation='minus', exon_count=15, OMIM=610560, no_hdr0=''),
  100288708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288708, CurrentID=0, Status='live', Symbol='LOC100288708', Aliases='', description='myomegalin-like', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143543950, end_position_on_the_genomic_accession=143573021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146556, CurrentID=0, Status='live', Symbol='C16orf89', Aliases='', description='chromosome 16 open reading frame 89', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5042772, end_position_on_the_genomic_accession=5066145, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100131340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131340, CurrentID=0, Status='live', Symbol='LOC100131340', Aliases='', description="5'-nucleotidase domain containing 1 pseudogene", other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93388084, end_position_on_the_genomic_accession=93389498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146562, CurrentID=0, Status='live', Symbol='C16orf71', Aliases='', description='chromosome 16 open reading frame 71', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4734288, end_position_on_the_genomic_accession=4749396, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101928658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928658, CurrentID=0, Status='live', Symbol='LOC101928658', Aliases='', description='uncharacterized LOC101928658', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52930959, end_position_on_the_genomic_accession=52951019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271006, CurrentID=0, Status='live', Symbol='RPL35P4', Aliases='RPL35_3_799', description='ribosomal protein L35 pseudogene 4', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27269315, end_position_on_the_genomic_accession=27269753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319127, CurrentID=0, Status='live', Symbol='NARSP1', Aliases='NARSP', description='asparaginyl-tRNA synthetase pseudogene 1', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31286116, end_position_on_the_genomic_accession=31289371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373131, CurrentID=0, Status='live', Symbol='LOC105373131', Aliases='', description='uncharacterized LOC105373131', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11732698, end_position_on_the_genomic_accession=11735266, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271007, CurrentID=0, Status='live', Symbol='RPL35P6', Aliases='RPL35_5_881', description='ribosomal protein L35 pseudogene 6', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18774547, end_position_on_the_genomic_accession=18774966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271008, CurrentID=0, Status='live', Symbol='RPL36P6', Aliases='RPL36_2_186', description='ribosomal protein L36 pseudogene 6', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241305603, end_position_on_the_genomic_accession=241305889, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146664, CurrentID=0, Status='live', Symbol='MGAT5B', Aliases='GnT-IX, GnT-VB', description='mannosyl (alpha-1,6-)-glycoprotein beta-1,6-N-acetyl-glucosaminyltransferase, isozyme B', other_designations='N-acetylglucosaminyl-transferase Vb|N-acetylglucosaminyltransferase IX|alpha-mannoside beta-1,6-N-acetylglucosaminyltransferase B|beta(1,6)-N-acetylglucosaminyltransferase V|glcNAc-T Vb|hGnTVb|mannoside acetylglucosaminyltransferase 5B|mannosyl (alpha-1,6-)-glycoprotein beta-1,6-N-acetyl-glucosaminyltransferase, isoenzyme B', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76850503, end_position_on_the_genomic_accession=76950393, orientation='plus', exon_count=27, OMIM=612441, no_hdr0=''),
  100271009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271009, CurrentID=0, Status='live', Symbol='RPL36P9', Aliases='RPL36_2_687', description='ribosomal protein L36 pseudogene 9', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35607593, end_position_on_the_genomic_accession=35607980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146691, CurrentID=0, Status='live', Symbol='TOM1L2', Aliases='', description='target of myb1 like 2 membrane trafficking protein', other_designations='target of Myb-like protein 2|target of myb1-like 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17843508, end_position_on_the_genomic_accession=17972470, orientation='minus', exon_count=17, OMIM=615519, no_hdr0=''),
  101929527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929527, CurrentID=0, Status='live', Symbol='LOC101929527', Aliases='', description='serine/arginine-rich splicing factor 3 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117873680, end_position_on_the_genomic_accession=117883723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146705, CurrentID=0, Status='live', Symbol='ENTHD2', Aliases='C17orf56', description='ENTH domain containing 2', other_designations='ENTH domain-containing protein 2|epsin for AP-4|tetra-epsin', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81228277, end_position_on_the_genomic_accession=81239091, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  146712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146712, CurrentID=0, Status='live', Symbol='B3GNTL1', Aliases='B3GNT8', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like 1', other_designations='BGnT-8|BGnT-like protein 1|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 8|beta-1,3-Gn-T8|beta-1,3-N-acetylglucosaminyltransferase 8|beta1,3-N-acetylglucosaminyltransferase-like protein 1|beta3Gn-T-like protein 1|beta3Gn-T8|beta3GnTL1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82942149, end_position_on_the_genomic_accession=83051845, orientation='minus', exon_count=22, OMIM=615337, no_hdr0=''),
  146713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146713, CurrentID=0, Status='live', Symbol='RBFOX3', Aliases='FOX-3, FOX3, HRNBP3, NEUN', description='RNA binding protein, fox-1 homolog (C. elegans) 3', other_designations='fox-1 homolog C|hexaribonucleotide binding protein 3|neuronal nuclei', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79089345, end_position_on_the_genomic_accession=79611101, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100271010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271010, CurrentID=0, Status='live', Symbol='RPL36P7', Aliases='RPL36_3_453', description='ribosomal protein L36 pseudogene 7', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197580350, end_position_on_the_genomic_accession=197580629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419643, CurrentID=0, Status='live', Symbol='LOC100419643', Aliases='', description='ring finger protein 14 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98998536, end_position_on_the_genomic_accession=98999928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146722, CurrentID=0, Status='live', Symbol='CD300LF', Aliases='CD300f, CLM-1, CLM1, IREM-1, IREM1, IgSF13, LMIR3, NKIR', description='CD300 molecule like family member f', other_designations='CD300 antigen-like family member F|CD300 molecule-like family member f|NK inhibitory receptor|immune receptor expressed on myeloid cells 1|immunoglobin superfamily member 13|immunoglobulin superfamily member 13|inhibitory receptor IREM1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74694308, end_position_on_the_genomic_accession=74713059, orientation='minus', exon_count=9, OMIM=609807, no_hdr0=''),
  146723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146723, CurrentID=0, Status='live', Symbol='C17orf77', Aliases='', description='chromosome 17 open reading frame 77', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74584679, end_position_on_the_genomic_accession=74594209, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929321, CurrentID=0, Status='live', Symbol='LOC101929321', Aliases='', description='uncharacterized LOC101929321', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85291405, end_position_on_the_genomic_accession=85304187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131346, CurrentID=0, Status='live', Symbol='LOC100131346', Aliases='', description='uncharacterized LOC100131346', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=49440508, end_position_on_the_genomic_accession=49441279, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376309, CurrentID=0, Status='live', Symbol='LOC105376309', Aliases='', description='uncharacterized LOC105376309', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134179347, end_position_on_the_genomic_accession=134207375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271011, CurrentID=0, Status='live', Symbol='RPL36P10', Aliases='RPL36_3_697', description='ribosomal protein L36 pseudogene 10', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46438310, end_position_on_the_genomic_accession=46438613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146754, CurrentID=0, Status='live', Symbol='DNAH2', Aliases='DNAHC2, DNHD3', description='dynein, axonemal, heavy chain 2', other_designations='axonemal beta dynein heavy chain 2|ciliary dynein heavy chain 2|dynein heavy chain domain-containing protein 3|dynein, axonemal, heavy polypeptide 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7717354, end_position_on_the_genomic_accession=7833744, orientation='plus', exon_count=91, OMIM=603333, no_hdr0=''),
  146760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146760, CurrentID=0, Status='live', Symbol='RTN4RL1', Aliases='NGRH2, NgR3', description='reticulon 4 receptor-like 1', other_designations='Nogo-66 receptor homolog 2|Nogo-66 receptor-related protein 3|nogo receptor-like 2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1934677, end_position_on_the_genomic_accession=2025345, orientation='minus', exon_count=14, OMIM=610461, no_hdr0=''),
  2616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2616, CurrentID=0, Status='live', Symbol='GAPDHP1', Aliases='GAPDHL2, GAPDL2, GAPDP1', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 1', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 2', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39786931, end_position_on_the_genomic_accession=39787959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146771, CurrentID=0, Status='live', Symbol='TCAM1P', Aliases='TCAM1', description='testicular cell adhesion molecule 1, pseudogene', other_designations='testicular cell adhesion molecule 1 homolog, pseudogene', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63857016, end_position_on_the_genomic_accession=63864379, orientation='plus', exon_count=6, OMIM=612756, no_hdr0=''),
  646120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646120, CurrentID=0, Status='live', Symbol='LOC646120', Aliases='', description='Spi-C transcription factor (Spi-1/PU.1 related) pseudogene', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33786826, end_position_on_the_genomic_accession=33787821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146779, CurrentID=0, Status='live', Symbol='EFCAB3', Aliases='', description='EF-hand calcium binding domain 3', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62370218, end_position_on_the_genomic_accession=62416479, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  133690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133690, CurrentID=0, Status='live', Symbol='CAPSL', Aliases='', description='calcyphosine-like', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35904295, end_position_on_the_genomic_accession=35938779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  146795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146795, CurrentID=0, Status='live', Symbol='LOC146795', Aliases='', description='uncharacterized LOC146795', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72342811, end_position_on_the_genomic_accession=72355309, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  146802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146802, CurrentID=0, Status='live', Symbol='SLC47A2', Aliases='MATE2, MATE2-B, MATE2-K, MATE2K', description='solute carrier family 47 (multidrug and toxin extrusion), member 2', other_designations='H+/organic cation antiporter|MATE-2|hMATE-2|kidney-specific H(+)/organic cation antiporter|multidrug and toxin extrusion 2|solute carrier family 47, member 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19678315, end_position_on_the_genomic_accession=19718842, orientation='minus', exon_count=24, OMIM=609833, no_hdr0=''),
  100271013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271013, CurrentID=0, Status='live', Symbol='RPL36P13', Aliases='RPL36_5_837', description='ribosomal protein L36 pseudogene 13', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=114297067, end_position_on_the_genomic_accession=114297462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319134, CurrentID=0, Status='live', Symbol='HNRNPUP1', Aliases='HNRPUP', description='heterogeneous nuclear ribonucleoprotein U pseudogene 1', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43300260, end_position_on_the_genomic_accession=43300910, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377502, CurrentID=0, Status='live', Symbol='LOC105377502', Aliases='', description='uncharacterized LOC105377502', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155286218, end_position_on_the_genomic_accession=155304895, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729399, CurrentID=0, Status='live', Symbol='SEPHS1P6', Aliases='', description='selenophosphate synthetase 1 pseudogene 6', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199659551, end_position_on_the_genomic_accession=199660721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  146822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146822, CurrentID=0, Status='live', Symbol='CDRT15', Aliases='', description='CMT1A duplicated region transcript 15', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14235856, end_position_on_the_genomic_accession=14236833, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90007, CurrentID=0, Status='live', Symbol='MIDN', Aliases='', description='midnolin', other_designations='midbrain nucleolar protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1248553, end_position_on_the_genomic_accession=1259143, orientation='plus', exon_count=9, OMIM=606700, no_hdr0=''),
  146845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146845, CurrentID=0, Status='live', Symbol='CFAP52', Aliases='WDR16, WDRPUH', description='cilia and flagella associated protein 52', other_designations='WD repeat domain 16|WD repeat-containing protein 16|WD40-repeat protein up-regulated in HCC|WD40-repeat protein upregulated in HCC', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9576627, end_position_on_the_genomic_accession=9643459, orientation='plus', exon_count=14, OMIM=609804, no_hdr0=''),
  101929532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929532, CurrentID=0, Status='live', Symbol='LOC101929532', Aliases='', description='uncharacterized LOC101929532', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162114441, end_position_on_the_genomic_accession=162172739, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  146849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146849, CurrentID=0, Status='live', Symbol='CCDC42', Aliases='CCDC42A', description='coiled-coil domain containing 42', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8729928, end_position_on_the_genomic_accession=8744836, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  146850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146850, CurrentID=0, Status='live', Symbol='PIK3R6', Aliases='C17orf38, HsT41028, p84 PIKAP, p87(PIKAP), p87PIKAP', description='phosphoinositide-3-kinase regulatory subunit 6', other_designations='PI3Kgamma adapter protein of 87 kDa|p84 PI3K adapter protein|p87 phosphoinositide 3-kinase gamma (PI3Kg) adapter protein|phosphoinositide-3-kinase, regulatory subunit 6', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8802722, end_position_on_the_genomic_accession=8867697, orientation='minus', exon_count=24, OMIM=611462, no_hdr0=''),
  146852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146852, CurrentID=0, Status='live', Symbol='ODF4', Aliases='CT134, CT136, OPPO1', description='outer dense fiber of sperm tails 4', other_designations='cancer/testis antigen 134|cancer/testis antigen 136|hOPPO1|outer dense fiber 4|outer dense fiber of sperm tails protein 4|testis-specific protein oppo 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8339837, end_position_on_the_genomic_accession=8346048, orientation='plus', exon_count=5, OMIM=610097, no_hdr0=''),
  146853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146853, CurrentID=0, Status='live', Symbol='C17orf50', Aliases='', description='chromosome 17 open reading frame 50', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35760593, end_position_on_the_genomic_accession=35765079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  146857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146857, CurrentID=0, Status='live', Symbol='SLFN13', Aliases='SLFN10', description='schlafen family member 13', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35435096, end_position_on_the_genomic_accession=35449008, orientation='minus', exon_count=9, OMIM=614957, no_hdr0=''),
  146861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146861, CurrentID=0, Status='live', Symbol='SLC35G3', Aliases='AMAC1, TMEM21A', description='solute carrier family 35 member G3', other_designations='acyl-malonyl condensing enzyme 1|acyl-malonyl-condensing enzyme 1|protein AMAC1|solute carrier family 35, member G3|transmembrane protein 21A', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35192520, end_position_on_the_genomic_accession=35194393, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  146862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146862, CurrentID=0, Status='live', Symbol='UNC45B', Aliases='CMYA4, CTRCT43, SMUNC45, UNC-45B, UNC45', description='unc-45 myosin chaperone B', other_designations='cardiomyopathy associated 4|striated muscle UNC45|unc-45 homolog B', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35147798, end_position_on_the_genomic_accession=35189345, orientation='plus', exon_count=20, OMIM=611220, no_hdr0=''),
  100271015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271015, CurrentID=0, Status='live', Symbol='RPL37P10', Aliases='RPL37_2_245', description='ribosomal protein L37 pseudogene 10', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=83594983, end_position_on_the_genomic_accession=83595246, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060176, CurrentID=0, Status='live', Symbol='TPRKBP1', Aliases='', description='TP53RK binding protein pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28630761, end_position_on_the_genomic_accession=28659470, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319136, CurrentID=0, Status='live', Symbol='ATP5G1P1', Aliases='ATP5GP3', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C1 (subunit 9) pseudogene 1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9) pseudogene 3|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 1|pseudogene 3 for mitochondrial ATP synthase c subunit (P1 form)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105536746, end_position_on_the_genomic_accession=105537292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146880, CurrentID=0, Status='live', Symbol='LOC146880', Aliases='', description='Rho GTPase activating protein 27 pseudogene', other_designations='SH3 domain containing 20 pseudogene', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64749662, end_position_on_the_genomic_accession=64781999, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105373140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373140, CurrentID=0, Status='live', Symbol='LOC105373140', Aliases='', description='prothymosin alpha-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15653191, end_position_on_the_genomic_accession=15653711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146894, CurrentID=0, Status='live', Symbol='CD300LG', Aliases='CLM-9, CLM9, NEPMUCIN, TREM-4, TREM4', description='CD300 molecule like family member g', other_designations='CD300 antigen like family member G|CD300 antigen-like family member G|CD300 molecule-like family member g|triggering receptor expressed on myeloid cells 4', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43846777, end_position_on_the_genomic_accession=43863639, orientation='plus', exon_count=10, OMIM=610520, no_hdr0=''),
  90019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90019, CurrentID=0, Status='live', Symbol='SYT8', Aliases='', description='synaptotagmin 8', other_designations='synaptotagmin VIII|sytVIII', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1827479, end_position_on_the_genomic_accession=1837521, orientation='plus', exon_count=13, OMIM=607719, no_hdr0=''),
  319137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319137, CurrentID=0, Status='live', Symbol='ATP5G1P2', Aliases='ATP5GP4', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C1 (subunit 9) pseudogene 2|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9) pseudogene 4|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 2|pseudogene 4 for mitochondrial ATP synthase c subunit (P1 form)', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31320121, end_position_on_the_genomic_accession=31320688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  146909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146909, CurrentID=0, Status='live', Symbol='KIF18B', Aliases='', description='kinesin family member 18B', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44924711, end_position_on_the_genomic_accession=44947745, orientation='minus', exon_count=17, OMIM=614570, no_hdr0=''),
  146923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146923, CurrentID=0, Status='live', Symbol='RUNDC1', Aliases='', description='RUN domain containing 1', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42980510, end_position_on_the_genomic_accession=42993696, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100271017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271017, CurrentID=0, Status='live', Symbol='RPL37P12', Aliases='RPL37_3_259', description='ribosomal protein L37 pseudogene 12', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109552537, end_position_on_the_genomic_accession=109552846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319138, CurrentID=0, Status='live', Symbol='BNIP3P1', Aliases='BNIP3P', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 1', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28264410, end_position_on_the_genomic_accession=28265974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369201, CurrentID=0, Status='live', Symbol='LOC105369201', Aliases='', description='uncharacterized LOC105369201', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100855307, end_position_on_the_genomic_accession=100877011, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  146956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=146956, CurrentID=0, Status='live', Symbol='EME1', Aliases='MMS4L, SLX2A', description='essential meiotic structure-specific endonuclease 1', other_designations='MMS4 homolog|SLX2 structure-specific endonuclease subunit homolog A|essential meiotic endonuclease 1 homolog 1|essential meiotic endonuclease 1 homolog 2|hMMS4|homolog of yeast EME1 endonuclease', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50373220, end_position_on_the_genomic_accession=50381752, orientation='plus', exon_count=10, OMIM=610885, no_hdr0=''),
  100271018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271018, CurrentID=0, Status='live', Symbol='RPL37P13', Aliases='RPL37_4_228', description='ribosomal protein L37 pseudogene 13', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62507542, end_position_on_the_genomic_accession=62507823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319139, CurrentID=0, Status='live', Symbol='SNORD56B', Aliases='RNU56B', description='small nucleolar RNA, C/D box 56B', other_designations='RNA, U56B small nuclear', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71398337, end_position_on_the_genomic_accession=71398407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271019, CurrentID=0, Status='live', Symbol='RPL37P14', Aliases='RPL37_4_470', description='ribosomal protein L37 pseudogene 14', other_designations='', map_location='4p13-p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40491735, end_position_on_the_genomic_accession=40492025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319140, CurrentID=0, Status='live', Symbol='GCATP1', Aliases='GCATP', description='glycine C-acetyltransferase pseudogene 1', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63593852, end_position_on_the_genomic_accession=63595326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147004, CurrentID=0, Status='live', Symbol='LOC147004', Aliases='', description='uncharacterized LOC147004', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=33533833, end_position_on_the_genomic_accession=33627260, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  147007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147007, CurrentID=0, Status='live', Symbol='TMEM199', Aliases='C17orf32', description='transmembrane protein 199', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28357581, end_position_on_the_genomic_accession=28363683, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100131355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131355, CurrentID=0, Status='live', Symbol='LOC100131355', Aliases='', description='uncharacterized LOC100131355', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=129705115, end_position_on_the_genomic_accession=129705775, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  147011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147011, CurrentID=0, Status='live', Symbol='PROCA1', Aliases='', description='protein interacting with cyclin A1', other_designations='proline-rich cyclin A1-interacting protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28703197, end_position_on_the_genomic_accession=28712071, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  147015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147015, CurrentID=0, Status='live', Symbol='DHRS13', Aliases='SDR7C5', description='dehydrogenase/reductase (SDR family) member 13', other_designations='short chain dehydrogenase/reductase family 7C member 5|short chain dehydrogenase/reductase family 7C, member 5', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28897781, end_position_on_the_genomic_accession=28903071, orientation='minus', exon_count=5, OMIM=616157, no_hdr0=''),
  100271020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271020, CurrentID=0, Status='live', Symbol='RPL37P15', Aliases='RPL37_5_710', description='ribosomal protein L37 pseudogene 15', other_designations='', map_location='6q12-q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70098389, end_position_on_the_genomic_accession=70098682, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283713, CurrentID=0, Status='live', Symbol='LOC283713', Aliases='', description='uncharacterized LOC283713', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  147040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147040, CurrentID=0, Status='live', Symbol='KCTD11', Aliases='C17orf36, KCASH1, REN, REN/KCTD11', description='potassium channel tetramerization domain containing 11', other_designations='potassium channel tetramerisation domain containing 11|retinoic acid, EGF, NGF induced gene protein|retinoic acid, EGF, NGF induced gene/potassium channel tetramerization domain containing 11', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7351889, end_position_on_the_genomic_accession=7354944, orientation='plus', exon_count=1, OMIM=609848, no_hdr0=''),
  319142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319142, CurrentID=0, Status='live', Symbol='RPL12P5', Aliases='RPL12_15_1346', description='ribosomal protein L12 pseudogene 5', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30791098, end_position_on_the_genomic_accession=30791489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376484, CurrentID=0, Status='live', Symbol='LOC105376484', Aliases='', description='uncharacterized LOC105376484', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31609272, end_position_on_the_genomic_accession=31620873, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271022, CurrentID=0, Status='live', Symbol='RPL39P11', Aliases='RPL39_3_134', description='ribosomal protein L39 pseudogene 11', other_designations='', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179271116, end_position_on_the_genomic_accession=179271266, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147081, CurrentID=0, Status='live', Symbol='CRHR1-IT1', Aliases='C17orf69', description='CRHR1 intronic transcript 1', other_designations='CRHR1 intronic transcript 1 (non-protein coding)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45638975, end_position_on_the_genomic_accession=45646229, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  221122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221122, CurrentID=0, Status='live', Symbol='LOC221122', Aliases='', description='uncharacterized LOC221122', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44973902, end_position_on_the_genomic_accession=44978028, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377511, CurrentID=0, Status='live', Symbol='LOC105377511', Aliases='', description='uncharacterized LOC105377511', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159040769, end_position_on_the_genomic_accession=159052408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  147093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147093, CurrentID=0, Status='live', Symbol='LINC00974', Aliases='', description='long intergenic non-protein coding RNA 974', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41549606, end_position_on_the_genomic_accession=41554495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723883, CurrentID=0, Status='live', Symbol='LOC102723883', Aliases='', description='uncharacterized LOC102723883', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68055351, end_position_on_the_genomic_accession=68060502, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  147111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147111, CurrentID=0, Status='live', Symbol='NOTUM', Aliases='hNOTUM', description='notum pectinacetylesterase homolog (Drosophila)', other_designations='protein notum homolog', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81952507, end_position_on_the_genomic_accession=81961823, orientation='minus', exon_count=12, OMIM=609847, no_hdr0=''),
  319144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319144, CurrentID=0, Status='live', Symbol='RPL21P13', Aliases='RPL21_52_1410', description='ribosomal protein L21 pseudogene 13', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103241452, end_position_on_the_genomic_accession=103242005, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  133746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133746, CurrentID=0, Status='live', Symbol='JMY', Aliases='WHAMM2, WHDC1L3', description='junction mediating and regulatory protein, p53 cofactor', other_designations='WAS protein homology region 2 domain containing 1-like 3|junction-mediating and regulatory protein', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79236102, end_position_on_the_genomic_accession=79327215, orientation='plus', exon_count=11, OMIM=604279, no_hdr0=''),
  157657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157657, CurrentID=0, Status='live', Symbol='C8orf37', Aliases='CORD16, RP64, smalltalk', description='chromosome 8 open reading frame 37', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95244913, end_position_on_the_genomic_accession=95270337, orientation='minus', exon_count=6, OMIM=614477, no_hdr0=''),
  133748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133748, CurrentID=0, Status='live', Symbol='RPL7AP32', Aliases='RPL7A_8_583', description='ribosomal protein L7a pseudogene 32', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79529281, end_position_on_the_genomic_accession=79530151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373148, CurrentID=0, Status='live', Symbol='LOC105373148', Aliases='', description='uncharacterized LOC105373148', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23756276, end_position_on_the_genomic_accession=23782901, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100288482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288482, CurrentID=0, Status='live', Symbol='LOC100288482', Aliases='', description='uncharacterized LOC100288482', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30413540, end_position_on_the_genomic_accession=30414063, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147138, CurrentID=0, Status='live', Symbol='TMC8', Aliases='EV2, EVER2, EVIN2', description='transmembrane channel like 8', other_designations='epidermodysplasia verruciformis 2|epidermodysplasia verruciformis protein 2|transmembrane channel-like 8', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78128851, end_position_on_the_genomic_accession=78142968, orientation='plus', exon_count=19, OMIM=605829, no_hdr0=''),
  319145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319145, CurrentID=0, Status='live', Symbol='RPL21P8', Aliases='RPL21_48_1381', description='ribosomal protein L21 pseudogene 8', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65467774, end_position_on_the_genomic_accession=65468384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377513, CurrentID=0, Status='live', Symbol='LOC105377513', Aliases='', description='uncharacterized LOC105377513', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=160919296, end_position_on_the_genomic_accession=160930793, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  147150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147150, CurrentID=0, Status='live', Symbol='UPF3AP2', Aliases='', description='UPF3A pseudogene 2', other_designations='UPF3 regulator of nonsense transcripts homolog A pseudogene 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20374816, end_position_on_the_genomic_accession=20376838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480616, CurrentID=0, Status='live', Symbol='RNU6-869P', Aliases='', description='RNA, U6 small nuclear 869, pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  147166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147166, CurrentID=0, Status='live', Symbol='TRIM16L', Aliases='TRIM70', description='tripartite motif containing 16-like', other_designations='tripartite motif protein 70|tripartite motif-containing 16-like|tripartite motif-containing protein 70', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18697998, end_position_on_the_genomic_accession=18736118, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100271025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271025, CurrentID=0, Status='live', Symbol='RPL39P18', Aliases='RPL39_7_347', description='ribosomal protein L39 pseudogene 18', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20012403, end_position_on_the_genomic_accession=20012801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147172, CurrentID=0, Status='live', Symbol='LRRC37BP1', Aliases='LRRC37B2', description='leucine rich repeat containing 37B pseudogene 1', other_designations='LRRC37B pseudogene 1|leucine rich repeat containing 37, member B2|putative LRRC37B-like protein 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30576465, end_position_on_the_genomic_accession=30637466, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  147179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147179, CurrentID=0, Status='live', Symbol='WIPF2', Aliases='WICH, WIRE', description='WAS/WASL interacting protein family member 2', other_designations='WAS/WASL interacting protein family, member 2|WASP-binding protein|WASP-interacting protein-related protein|WIP- and CR16-homologous protein|WIP-related protein', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40219304, end_position_on_the_genomic_accession=40284136, orientation='plus', exon_count=9, OMIM=609692, no_hdr0=''),
  284230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284230, CurrentID=0, Status='live', Symbol='RPL36AP49', Aliases='RPL36A_23_1584', description='ribosomal protein L36a pseudogene 49', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13919573, end_position_on_the_genomic_accession=13919980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147183, CurrentID=0, Status='live', Symbol='KRT25', Aliases='KRT25A', description='keratin 25, type I', other_designations='CK-25|K25|K25A|cytokeratin-25|keratin 25A|keratin-25|keratin-25A|type I inner root sheath specific keratin 25 irs1|type I inner root sheath-specific keratin-K25irs1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40748021, end_position_on_the_genomic_accession=40755542, orientation='minus', exon_count=9, OMIM=616646, no_hdr0=''),
  147184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147184, CurrentID=0, Status='live', Symbol='TMEM99', Aliases='', description='transmembrane protein 99', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40819092, end_position_on_the_genomic_accession=40836274, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  147199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147199, CurrentID=0, Status='live', Symbol='SCGB1C1', Aliases='RYD5', description='secretoglobin, family 1C, member 1', other_designations='ligand binding protein RYD5|secretoglobin RYD5', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=187899, end_position_on_the_genomic_accession=194575, orientation='plus', exon_count=4, OMIM=610176, no_hdr0=''),
  147226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147226, CurrentID=0, Status='live', Symbol='UPF3AP1', Aliases='', description='UPF3A pseudogene 1', other_designations='UPF3 regulator of nonsense transcripts homolog A pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16744740, end_position_on_the_genomic_accession=16746762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147228, CurrentID=0, Status='live', Symbol='KRT17P1', Aliases='', description='keratin 17 pseudogene 1', other_designations='keratin pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16841963, end_position_on_the_genomic_accession=16845936, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271027, CurrentID=0, Status='live', Symbol='RPL39P19', Aliases='RPL39_8_448', description='ribosomal protein L39 pseudogene 19', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187120954, end_position_on_the_genomic_accession=187121349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271028, CurrentID=0, Status='live', Symbol='RPL39P22', Aliases='RPL39_8_554', description='ribosomal protein L39 pseudogene 22', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36588506, end_position_on_the_genomic_accession=36588661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271251, CurrentID=0, Status='live', Symbol='RPS2P49', Aliases='RPS2_24_1542', description='ribosomal protein S2 pseudogene 49', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37140571, end_position_on_the_genomic_accession=37141413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147276, CurrentID=0, Status='live', Symbol='C18orf15', Aliases='HsT3231', description='chromosome 18 open reading frame 15', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=13229542, end_position_on_the_genomic_accession=13230722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379784, CurrentID=0, Status='live', Symbol='LOC105379784', Aliases='', description='uncharacterized LOC105379784', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373154, CurrentID=0, Status='live', Symbol='LOC105373154', Aliases='', description='myosin heavy chain IB-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36816330, end_position_on_the_genomic_accession=36959103, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271030, CurrentID=0, Status='live', Symbol='UBA52P4', Aliases='RPL40_1_352', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 4', other_designations='ribosomal protein L40 pseudogene', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27486250, end_position_on_the_genomic_accession=27486617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060002, CurrentID=0, Status='live', Symbol='LOC101060002', Aliases='', description='uncharacterized LOC101060002', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108645646, end_position_on_the_genomic_accession=108653521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147323, CurrentID=0, Status='live', Symbol='STARD6', Aliases='', description='StAR related lipid transfer domain containing 6', other_designations='START domain containing 6|START domain-containing protein 6|StAR-related lipid transfer (START) domain containing 6|StAR-related lipid transfer domain containing 6', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54324592, end_position_on_the_genomic_accession=54357890, orientation='minus', exon_count=9, OMIM=607051, no_hdr0=''),
  645180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645180, CurrentID=0, Status='live', Symbol='LOC645180', Aliases='', description='family with sequence similarity 214 member B pseudogene', other_designations='family with sequence similarity 214, member B pseudogene', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114231401, end_position_on_the_genomic_accession=114233021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147339, CurrentID=0, Status='live', Symbol='C18orf25', Aliases='ARKL1, RNF111L1', description='chromosome 18 open reading frame 25', other_designations='ARKadia-like 1', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46173553, end_position_on_the_genomic_accession=46266992, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373155, CurrentID=0, Status='live', Symbol='LOC105373155', Aliases='', description='uncharacterized LOC105373155', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36719765, end_position_on_the_genomic_accession=36744117, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271031, CurrentID=0, Status='live', Symbol='UBA52P5', Aliases='RPL40_2_948', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 5', other_designations='ribosomal protein L40 pseudogene', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123236852, end_position_on_the_genomic_accession=123237216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929549, CurrentID=0, Status='live', Symbol='LOC101929549', Aliases='', description='uncharacterized LOC101929549', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4700417, end_position_on_the_genomic_accession=4720102, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  147372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147372, CurrentID=0, Status='live', Symbol='CCBE1', Aliases='HKLLS1', description='collagen and calcium binding EGF domains 1', other_designations='full of fluid protein homolog', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59430939, end_position_on_the_genomic_accession=59697643, orientation='minus', exon_count=12, OMIM=612753, no_hdr0=''),
  133789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133789, CurrentID=0, Status='live', Symbol='RAB5CP2', Aliases='', description='RAB5C, member RAS oncogene family pseudogene 2', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91476411, end_position_on_the_genomic_accession=91477020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288490, CurrentID=0, Status='live', Symbol='LOC100288490', Aliases='', description='uncharacterized LOC100288490', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=58520198, end_position_on_the_genomic_accession=58522600, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100533627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533627, CurrentID=0, Status='live', Symbol='LOC100533627', Aliases='', description='tubulin, alpha 1c pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46891526, end_position_on_the_genomic_accession=46892830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147381, CurrentID=0, Status='live', Symbol='CBLN2', Aliases='', description='cerebellin 2 precursor', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72535898, end_position_on_the_genomic_accession=72638477, orientation='minus', exon_count=8, OMIM=600433, no_hdr0=''),
  105373342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373342, CurrentID=0, Status='live', Symbol='LOC105373342', Aliases='', description='uncharacterized LOC105373342', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135326903, end_position_on_the_genomic_accession=135330459, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377521, CurrentID=0, Status='live', Symbol='LOC105377521', Aliases='', description='uncharacterized LOC105377521', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165845607, end_position_on_the_genomic_accession=165855704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373157, CurrentID=0, Status='live', Symbol='LOC105373157', Aliases='', description='uncharacterized LOC105373157', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  147407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147407, CurrentID=0, Status='live', Symbol='SLC25A52', Aliases='MCART2', description='solute carrier family 25 member 52', other_designations='mitochondrial carrier triple repeat 2|mitochondrial carrier triple repeat protein 2|solute carrier family 25, member 52', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31759696, end_position_on_the_genomic_accession=31760880, orientation='minus', exon_count=1, OMIM=616153, no_hdr0=''),
  147409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147409, CurrentID=0, Status='live', Symbol='DSG4', Aliases='CDGF13, CDHF13, HYPT6, LAH', description='desmoglein 4', other_designations='CDH family member 13|cadherin family member 13', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31376777, end_position_on_the_genomic_accession=31415791, orientation='plus', exon_count=16, OMIM=607892, no_hdr0=''),
  101928536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928536, CurrentID=0, Status='live', Symbol='LINC01493', Aliases='', description='long intergenic non-protein coding RNA 1493', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38648858, end_position_on_the_genomic_accession=38655249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544, CurrentID=0, Status='live', Symbol='ALDH7A1P4', Aliases='ATQL4', description='aldehyde dehydrogenase 7 family member A1 pseudogene 4', other_designations='aldehyde dehydrogenase 7 family, member A1 pseudogene 4|antiquitin-like 4', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62740990, end_position_on_the_genomic_accession=62741394, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157667, CurrentID=0, Status='live', Symbol='MYL12AP1', Aliases='', description='myosin light chain 12A pseudogene 1', other_designations='myosin, light chain 12A pseudogene 1|myosin, light chain 12A, regulatory, non-sarcomeric pseudogene', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93849548, end_position_on_the_genomic_accession=93850277, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147429, CurrentID=0, Status='live', Symbol='AQP4-AS1', Aliases='C18orf16, CHST9-AS1', description='AQP4 antisense RNA 1', other_designations='CHST9 antisense RNA 1 (non-protein coding)', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26865308, end_position_on_the_genomic_accession=26935946, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  401904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401904, CurrentID=0, Status='live', Symbol='RPL23AP2', Aliases='RPL23A_38_1631', description='ribosomal protein L23a pseudogene 2', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15611641, end_position_on_the_genomic_accession=15612609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373158, CurrentID=0, Status='live', Symbol='LOC105373158', Aliases='', description='uncharacterized LOC105373158', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229087100, end_position_on_the_genomic_accession=229184277, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100271034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271034, CurrentID=0, Status='live', Symbol='RPL6P24', Aliases='RPL6_8_1023', description='ribosomal protein L6 pseudogene 24', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13591425, end_position_on_the_genomic_accession=13592350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147463, CurrentID=0, Status='live', Symbol='ANKRD29', Aliases='', description='ankyrin repeat domain 29', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23598926, end_position_on_the_genomic_accession=23662892, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  147468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147468, CurrentID=0, Status='live', Symbol='WBP2P1', Aliases='', description='WW domain binding protein 2 pseudogene 1', other_designations='hCG1643692', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=25016433, end_position_on_the_genomic_accession=25017203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271035, CurrentID=0, Status='live', Symbol='RPL6P26', Aliases='RPL6_9_1550', description='ribosomal protein L6 pseudogene 26', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44719737, end_position_on_the_genomic_accession=44720615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147495, CurrentID=0, Status='live', Symbol='APCDD1', Aliases='B7323, DRAPC1, FP7019, HHS, HTS, HYPT1', description='adenomatosis polyposis coli down-regulated 1', other_designations='adenomatosis polyposis coli down-regulated 1 protein|hypoptrichosis simplex', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10454628, end_position_on_the_genomic_accession=10488701, orientation='plus', exon_count=6, OMIM=607479, no_hdr0=''),
  100271036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271036, CurrentID=0, Status='live', Symbol='RPL7AP19', Aliases='RPL7A_2_119', description='ribosomal protein L7a pseudogene 19', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168542706, end_position_on_the_genomic_accession=168543356, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284241, CurrentID=0, Status='live', Symbol='LOC284241', Aliases='', description='synaptojanin-1', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79641551, end_position_on_the_genomic_accession=79679775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131372, CurrentID=0, Status='live', Symbol='LOC100131372', Aliases='', description='uncharacterized LOC100131372', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=427056, end_position_on_the_genomic_accession=433613, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  147525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147525, CurrentID=0, Status='live', Symbol='LINC00526', Aliases='C18orf18, HsT959', description='long intergenic non-protein coding RNA 526', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5236724, end_position_on_the_genomic_accession=5238029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271037, CurrentID=0, Status='live', Symbol='RPL7AP20', Aliases='RPL7A_3_155', description='ribosomal protein L7a pseudogene 20', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206528893, end_position_on_the_genomic_accession=206529765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929555, CurrentID=0, Status='live', Symbol='LOC101929555', Aliases='', description='uncharacterized LOC101929555', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40878886, end_position_on_the_genomic_accession=41023280, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  103724387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103724387, CurrentID=0, Status='live', Symbol='LINC01058', Aliases='', description='long intergenic non-protein coding RNA 1058', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30419439, end_position_on_the_genomic_accession=30422244, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377275, CurrentID=0, Status='live', Symbol='LOC105377275', Aliases='', description='uncharacterized LOC105377275', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74038762, end_position_on_the_genomic_accession=74042622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131373, CurrentID=0, Status='live', Symbol='LOC100131373', Aliases='', description='uncharacterized LOC100131373', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=20425035, end_position_on_the_genomic_accession=20427748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271038, CurrentID=0, Status='live', Symbol='RPL7AP21', Aliases='RPL7A_4_120', description='ribosomal protein L7a pseudogene 21', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168578601, end_position_on_the_genomic_accession=168579463, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271039, CurrentID=0, Status='live', Symbol='RPL7AP22', Aliases='RPL7A_4_289', description='ribosomal protein L7a pseudogene 22', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158548404, end_position_on_the_genomic_accession=158549274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996333, CurrentID=0, Status='live', Symbol='LOC100996333', Aliases='', description='G protein-coupled receptor 125 pseudogene', other_designations='adhesion G protein-coupled receptor A3 pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12075206, end_position_on_the_genomic_accession=12076850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372911, CurrentID=0, Status='live', Symbol='LOC105372911', Aliases='', description='uncharacterized LOC105372911', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214019269, end_position_on_the_genomic_accession=214112021, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100271040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271040, CurrentID=0, Status='live', Symbol='RPL7AP23', Aliases='RPL7A_5_385', description='ribosomal protein L7a pseudogene 23', other_designations='', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=82319122, end_position_on_the_genomic_accession=82319808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147645, CurrentID=0, Status='live', Symbol='VSIG10L', Aliases='', description='V-set and immunoglobulin domain containing 10 like', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51331541, end_position_on_the_genomic_accession=51342124, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  147646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147646, CurrentID=0, Status='live', Symbol='C19orf84', Aliases='', description='chromosome 19 open reading frame 84', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51388289, end_position_on_the_genomic_accession=51390574, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  147650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147650, CurrentID=0, Status='live', Symbol='SPACA6', Aliases='LET7EH, LINC00085, NCRNA00085P, SPACA6', description='sperm acrosome associated 6', other_designations='BACHELOR-like protein|Let-7e host|long intergenic non-protein coding RNA 85|sperm acrosome associated 6, pseudogene', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51693340, end_position_on_the_genomic_accession=51705190, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  285548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285548, CurrentID=0, Status='live', Symbol='LINC01096', Aliases='', description='long intergenic non-protein coding RNA 1096', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13546076, end_position_on_the_genomic_accession=13547824, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  147657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147657, CurrentID=0, Status='live', Symbol='ZNF480', Aliases='', description='zinc finger protein 480', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52297169, end_position_on_the_genomic_accession=52325927, orientation='plus', exon_count=6, OMIM=613910, no_hdr0=''),
  147658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147658, CurrentID=0, Status='live', Symbol='ZNF534', Aliases='KRBO3', description='zinc finger protein 534', other_designations='KRAB domain only 3|KRAB domain only protein 3', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52429321, end_position_on_the_genomic_accession=52451939, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  147660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147660, CurrentID=0, Status='live', Symbol='ZNF578', Aliases='', description='zinc finger protein 578', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52453506, end_position_on_the_genomic_accession=52516878, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  147664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147664, CurrentID=0, Status='live', Symbol='ERVV-1', Aliases='ENVV1, HERV-V1', description='endogenous retrovirus group V member 1', other_designations='HERV-V_19q13.41 provirus ancestral Env polyprotein 1|endogenous retrovirus group V, member 1|endogenous retrovirus group V, member 2|envelope glycoprotein ENVV1|envelope protein ENVV1', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53014091, end_position_on_the_genomic_accession=53016580, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  650866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650866, CurrentID=0, Status='live', Symbol='LOC650866', Aliases='', description='transient receptor potential cation channel, subfamily C, member 6 pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145729005, end_position_on_the_genomic_accession=145729310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147670, CurrentID=0, Status='live', Symbol='SMIM17', Aliases='', description='small integral membrane protein 17', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56643159, end_position_on_the_genomic_accession=56657246, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  147685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147685, CurrentID=0, Status='live', Symbol='C19orf18', Aliases='', description='chromosome 19 open reading frame 18', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57958437, end_position_on_the_genomic_accession=57974534, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  147686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147686, CurrentID=0, Status='live', Symbol='ZNF418', Aliases='', description='zinc finger protein 418', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57921884, end_position_on_the_genomic_accession=57935408, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  147687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147687, CurrentID=0, Status='live', Symbol='ZNF417', Aliases='', description='zinc finger protein 417', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57905774, end_position_on_the_genomic_accession=57916616, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  147694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147694, CurrentID=0, Status='live', Symbol='ZNF548', Aliases='', description='zinc finger protein 548', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57389850, end_position_on_the_genomic_accession=57402551, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  147699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147699, CurrentID=0, Status='live', Symbol='PPM1N', Aliases='', description='protein phosphatase, Mg2+/Mn2+ dependent 1N (putative)', other_designations='protein phosphatase, Mg2+/Mn2+ dependent, 1N (putative)', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45497975, end_position_on_the_genomic_accession=45502510, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  147700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147700, CurrentID=0, Status='live', Symbol='KLC3', Aliases='KLC2, KLC2L, KLCt, KNS2B', description='kinesin light chain 3', other_designations='kinesin light chain 2', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45340740, end_position_on_the_genomic_accession=45351559, orientation='plus', exon_count=13, OMIM=601334, no_hdr0=''),
  147710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147710, CurrentID=0, Status='live', Symbol='IGSF23', Aliases='', description='immunoglobulin superfamily member 23', other_designations='immunoglobulin superfamily, member 23', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44613590, end_position_on_the_genomic_accession=44636781, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  147711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147711, CurrentID=0, Status='live', Symbol='ZNF285B', Aliases='', description='zinc finger protein 285B, pseudogene', other_designations='zinc finger protein 285 pseudogene', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44464182, end_position_on_the_genomic_accession=44473358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147719, CurrentID=0, Status='live', Symbol='LYPD4', Aliases='SMR', description='LY6/PLAUR domain containing 4', other_designations='sperm membrane receptor', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41833686, end_position_on_the_genomic_accession=41844862, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  147727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147727, CurrentID=0, Status='live', Symbol='ILF3-AS1', Aliases='', description='ILF3 antisense RNA 1 (head to head)', other_designations='CTC-539A10.4|ILF3 antisense RNA 1 (non-protein coding)', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10651862, end_position_on_the_genomic_accession=10653872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100286746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286746, CurrentID=0, Status='live', Symbol='LOC100286746', Aliases='', description='GABA(A) receptor-associated protein like 2 pseudogene', other_designations='GABA(A) receptor-associated protein-like 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118982010, end_position_on_the_genomic_accession=118982941, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928711, CurrentID=0, Status='live', Symbol='LOC101928711', Aliases='', description='uncharacterized LOC101928711', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177765619, end_position_on_the_genomic_accession=177767316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  147741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147741, CurrentID=0, Status='live', Symbol='ZNF560', Aliases='', description='zinc finger protein 560', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9445826, end_position_on_the_genomic_accession=9506689, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  147744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147744, CurrentID=0, Status='live', Symbol='TMEM190', Aliases='MDAC1', description='transmembrane protein 190', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55376836, end_position_on_the_genomic_accession=55378244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  147746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147746, CurrentID=0, Status='live', Symbol='HIPK4', Aliases='', description='homeodomain interacting protein kinase 4', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40379271, end_position_on_the_genomic_accession=40390221, orientation='minus', exon_count=4, OMIM=611712, no_hdr0=''),
  106480620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480620, CurrentID=0, Status='live', Symbol='RNU6-907P', Aliases='', description='RNA, U6 small nuclear 907, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373169, CurrentID=0, Status='live', Symbol='LOC105373169', Aliases='', description='uncharacterized LOC105373169', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231268889, end_position_on_the_genomic_accession=231270660, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644310, CurrentID=0, Status='live', Symbol='LOC644310', Aliases='', description='ubiquinol-cytochrome c reductase, complex III subunit X pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51923544, end_position_on_the_genomic_accession=51923970, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271045, CurrentID=0, Status='live', Symbol='RPL9P24', Aliases='RPL9_7_1272', description='ribosomal protein L9 pseudogene 24', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101936216, end_position_on_the_genomic_accession=101936674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147791, CurrentID=0, Status='live', Symbol='LOC147791', Aliases='', description='uncharacterized LOC147791', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  147798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147798, CurrentID=0, Status='live', Symbol='TMC4', Aliases='', description='transmembrane channel like 4', other_designations='transmembrane channel-like 4', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54160095, end_position_on_the_genomic_accession=54173250, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  147804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147804, CurrentID=0, Status='live', Symbol='TPM3P9', Aliases='', description='tropomyosin 3 pseudogene 9', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53431974, end_position_on_the_genomic_accession=53444672, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  147807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147807, CurrentID=0, Status='live', Symbol='ZNF524', Aliases='', description='zinc finger protein 524', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55599755, end_position_on_the_genomic_accession=55603138, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  147808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147808, CurrentID=0, Status='live', Symbol='ZNF784', Aliases='', description='zinc finger protein 784', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55620741, end_position_on_the_genomic_accession=55624575, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  157680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157680, CurrentID=0, Status='live', Symbol='VPS13B', Aliases='CHS1, COH1', description='vacuolar protein sorting 13 homolog B (yeast)', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99013071, end_position_on_the_genomic_accession=99877586, orientation='plus', exon_count=80, OMIM=607817, no_hdr0=''),
  105373171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373171, CurrentID=0, Status='live', Symbol='LOC105373171', Aliases='', description='uncharacterized LOC105373171', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232061404, end_position_on_the_genomic_accession=232092497, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100286836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286836, CurrentID=0, Status='live', Symbol='OS6', Aliases='', description='Osteoarthritis susceptibility 6', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612401, no_hdr0=''),
  100271047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271047, CurrentID=0, Status='live', Symbol='RPLP0P7', Aliases='RPLP0_2_288', description='ribosomal protein, large, P0 pseudogene 7', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156777692, end_position_on_the_genomic_accession=156778378, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100286844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286844, CurrentID=0, Status='live', Symbol='BCDIN3D-AS1', Aliases='', description='BCDIN3D antisense RNA 1', other_designations='BCDIN3D antisense RNA 1 (non-protein coding)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49828543, end_position_on_the_genomic_accession=49841154, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  147837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147837, CurrentID=0, Status='live', Symbol='ZNF563', Aliases='', description='zinc finger protein 563', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12317477, end_position_on_the_genomic_accession=12348130, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  147841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147841, CurrentID=0, Status='live', Symbol='SPC24', Aliases='SPBC24', description='SPC24, NDC80 kinetochore complex component', other_designations='SPC24, NDC80 kinetochore complex component, homolog|hSpc24|spindle pole body component 24 homolog', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11145498, end_position_on_the_genomic_accession=11156002, orientation='minus', exon_count=6, OMIM=609394, no_hdr0=''),
  147872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147872, CurrentID=0, Status='live', Symbol='CCDC155', Aliases='KASH5', description='coiled-coil domain containing 155', other_designations='KASH domain-containing protein 5', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49388218, end_position_on_the_genomic_accession=49417999, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105371946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371946, CurrentID=0, Status='live', Symbol='LOC105371946', Aliases='', description='uncharacterized LOC105371946', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  133874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133874, CurrentID=0, Status='live', Symbol='C5orf58', Aliases='', description='chromosome 5 open reading frame 58', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170232552, end_position_on_the_genomic_accession=170246231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100286895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286895, CurrentID=0, Status='live', Symbol='LOC100286895', Aliases='', description='cell division cycle 27 homolog pseudogene', other_designations='cell division cycle 27 homolog (S. cerevisiae) pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100286906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286906, CurrentID=0, Status='live', Symbol='LOC100286906', Aliases='', description='uncharacterized LOC100286906', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155382076, end_position_on_the_genomic_accession=155395383, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  147906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147906, CurrentID=0, Status='live', Symbol='DACT3', Aliases='DAPPER3, RRR1', description='dishevelled-binding antagonist of beta-catenin 3', other_designations='antagonist of beta-catenin Dapper homolog 3|arginine-rich region 1 protein|dapper, antagonist of beta-catenin, homolog 3', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46647564, end_position_on_the_genomic_accession=46661138, orientation='minus', exon_count=8, OMIM=611112, no_hdr0=''),
  100286918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286918, CurrentID=0, Status='live', Symbol='NDUFS5P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 2', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96584407, end_position_on_the_genomic_accession=96584787, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147912, CurrentID=0, Status='live', Symbol='SIX5', Aliases='BOR2, DMAHP', description='SIX homeobox 5', other_designations='DM locus-associated homeodomain protein|dystrophia myotonica-associated homeodomain protein|sine oculis homeobox homolog 5', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45764785, end_position_on_the_genomic_accession=45769239, orientation='minus', exon_count=3, OMIM=600963, no_hdr0=''),
  100286922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286922, CurrentID=0, Status='live', Symbol='LOC100286922', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B3 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 3 pseudogene', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233754798, end_position_on_the_genomic_accession=233757288, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100286925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286925, CurrentID=0, Status='live', Symbol='LOC100286925', Aliases='', description='uncharacterized LOC100286925', other_designations='', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=22292609, end_position_on_the_genomic_accession=22297806, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  147920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147920, CurrentID=0, Status='live', Symbol='IGFL2', Aliases='UNQ645, VPRI645', description='IGF like family member 2', other_designations='IGF-like family member 2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46078528, end_position_on_the_genomic_accession=46195097, orientation='plus', exon_count=10, OMIM=610545, no_hdr0=''),
  100286929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286929, CurrentID=0, Status='live', Symbol='NMD3P2', Aliases='', description='NMD3 ribosome export adaptor pseudogene 2', other_designations='NMD3 homolog pseudogene 2|nonsense-mediated mRNA 3 homolog pseudogene 2', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56502284, end_position_on_the_genomic_accession=56503044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147923, CurrentID=0, Status='live', Symbol='ZNF420', Aliases='APAK', description='zinc finger protein 420', other_designations='ATM and p53-associated KZNF protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37007813, end_position_on_the_genomic_accession=37130324, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100873354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873354, CurrentID=0, Status='live', Symbol='MTND5P21', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 21', other_designations='MT-ND5 pseudogene 21', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11239855, end_position_on_the_genomic_accession=11240175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147929, CurrentID=0, Status='live', Symbol='ZNF565', Aliases='', description='zinc finger protein 565', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36182060, end_position_on_the_genomic_accession=36215084, orientation='minus', exon_count=9, OMIM=614275, no_hdr0=''),
  100286938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286938, CurrentID=0, Status='live', Symbol='VPS13A-AS1', Aliases='', description='VPS13A antisense RNA 1', other_designations='VPS13A antisense RNA 1 (non-protein coding)', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77176756, end_position_on_the_genomic_accession=77177917, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100286942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286942, CurrentID=0, Status='live', Symbol='LOC100286942', Aliases='', description='ribosomal protein S15a pseudogene', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67973817, end_position_on_the_genomic_accession=67974627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100286945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286945, CurrentID=0, Status='live', Symbol='LOC100286945', Aliases='', description='X antigen family member 3 pseudogene', other_designations='X antigen family, member 3 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3312421, end_position_on_the_genomic_accession=3312929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286946, CurrentID=0, Status='live', Symbol='LOC100286946', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9030262, end_position_on_the_genomic_accession=9031105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286948, CurrentID=0, Status='live', Symbol='APOOP1', Aliases='', description='apolipoprotein O pseudogene 1', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156924089, end_position_on_the_genomic_accession=156925003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147945, CurrentID=0, Status='live', Symbol='NLRP4', Aliases='CLR19.5, CT58, NALP4, PAN2, PYPAF4, RNH2', description='NLR family, pyrin domain containing 4', other_designations='NACHT, LRR and PYD containing protein 4|NACHT, leucine rich repeat and PYD containing 4|PAAD and NACHT-containing protein 2|PYRIN and NACHT-containing protein 2|PYRIN-containing APAF1-like protein 4|cancer/testis antigen 58|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 4|ribonuclease inhibitor 2', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55836086, end_position_on_the_genomic_accession=55881855, orientation='plus', exon_count=11, OMIM=609645, no_hdr0=''),
  100286954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286954, CurrentID=0, Status='live', Symbol='HSPE1P12', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 12', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 12', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10770307, end_position_on_the_genomic_accession=10770606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147947, CurrentID=0, Status='live', Symbol='ZNF542P', Aliases='ZNF542', description='zinc finger protein 542, pseudogene', other_designations='zinc finger protein pseudogene', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56368099, end_position_on_the_genomic_accession=56379828, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  147948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147948, CurrentID=0, Status='live', Symbol='ZNF582', Aliases='', description='zinc finger protein 582', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56374832, end_position_on_the_genomic_accession=56393585, orientation='minus', exon_count=6, OMIM=615600, no_hdr0=''),
  147949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147949, CurrentID=0, Status='live', Symbol='ZNF583', Aliases='', description='zinc finger protein 583', other_designations='zinc finger protein L3-5', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56404014, end_position_on_the_genomic_accession=56425031, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100286958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286958, CurrentID=0, Status='live', Symbol='LOC100286958', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112546973, end_position_on_the_genomic_accession=112547505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286962, CurrentID=0, Status='live', Symbol='LOC100286962', Aliases='', description='ras homolog family member G pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17599297, end_position_on_the_genomic_accession=17599807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100286963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286963, CurrentID=0, Status='live', Symbol='DEFB109', Aliases='', description='defensin, beta 109', other_designations='beta-defensin 109', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=7847937, end_position_on_the_genomic_accession=7855043, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100286964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286964, CurrentID=0, Status='live', Symbol='TSPY23P', Aliases='', description='testis specific protein, Y-linked 23, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22183850, end_position_on_the_genomic_accession=22186021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286965, CurrentID=0, Status='live', Symbol='CBX3P8', Aliases='', description='chromobox 3 pseudogene 8', other_designations='chromobox homolog 3 pseudogene 8', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49406190, end_position_on_the_genomic_accession=49406714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286971, CurrentID=0, Status='live', Symbol='UBDP1', Aliases='dJ994E9.3', description='ubiquitin D pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29464675, end_position_on_the_genomic_accession=29464945, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  147965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147965, CurrentID=0, Status='live', Symbol='FAM98C', Aliases='', description='family with sequence similarity 98 member C', other_designations='family with sequence similarity 98, member C', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38403125, end_position_on_the_genomic_accession=38409088, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  147968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147968, CurrentID=0, Status='live', Symbol='CAPN12', Aliases='', description='calpain 12', other_designations='CANP 12|calcium-activated neutral proteinase 12', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38730192, end_position_on_the_genomic_accession=38744474, orientation='minus', exon_count=21, OMIM=608839, no_hdr0=''),
  100286979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286979, CurrentID=0, Status='live', Symbol='ANAPC1P1', Aliases='', description='anaphase promoting complex subunit 1 pseudogene 1', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86861825, end_position_on_the_genomic_accession=86895042, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100286986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286986, CurrentID=0, Status='live', Symbol='LOC100286986', Aliases='', description='uncharacterized LOC100286986', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.9', start_position_on_the_genomic_accession=5592974, end_position_on_the_genomic_accession=5630929, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271052, CurrentID=0, Status='live', Symbol='RPS11P4', Aliases='RPS11_2_972', description='ribosomal protein S11 pseudogene 4', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32727066, end_position_on_the_genomic_accession=32727464, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100286989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286989, CurrentID=0, Status='live', Symbol='HSPC081', Aliases='', description='uncharacterized LOC100286989', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=159879800, end_position_on_the_genomic_accession=159880159, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100286992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286992, CurrentID=0, Status='live', Symbol='LOC100286992', Aliases='', description='signal recognition particle 19kDa pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59168474, end_position_on_the_genomic_accession=59169343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873356, CurrentID=0, Status='live', Symbol='RNA5SP24', Aliases='RN5S24', description='RNA, 5S ribosomal pseudogene 24', other_designations='RNA, 5S ribosomal 24', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19088487, end_position_on_the_genomic_accession=19088581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100286996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286996, CurrentID=0, Status='live', Symbol='TSPY22P', Aliases='', description='testis specific protein, Y-linked 22, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22304428, end_position_on_the_genomic_accession=22307893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100286997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286997, CurrentID=0, Status='live', Symbol='LOC100286997', Aliases='', description='ribosomal protein L31 pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94565036, end_position_on_the_genomic_accession=94565708, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  147991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=147991, CurrentID=0, Status='live', Symbol='DPY19L3', Aliases='', description='dpy-19-like 3 (C. elegans)', other_designations='dpy-19-like protein 3|protein dpy-19 homolog 3', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32405749, end_position_on_the_genomic_accession=32485893, orientation='plus', exon_count=22, OMIM=613894, no_hdr0=''),
  100287010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287010, CurrentID=0, Status='live', Symbol='LOC100287010', Aliases='', description='uncharacterized LOC100287010', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104378850, end_position_on_the_genomic_accession=104408332, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  148003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148003, CurrentID=0, Status='live', Symbol='LGALS16', Aliases='', description='lectin, galactoside-binding, soluble, 16', other_designations='beta-galactoside-binding lectin', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39655894, end_position_on_the_genomic_accession=39660647, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287013, CurrentID=0, Status='live', Symbol='FAM86KP', Aliases='', description='family with sequence similarity 86 member K, pseudogene', other_designations='family with sequence similarity 86, member K, pseudogene|protein FAM86A-like', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9153296, end_position_on_the_genomic_accession=9161952, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100287014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287014, CurrentID=0, Status='live', Symbol='LOC100287014', Aliases='', description='LysM, putative peptidoglycan-binding, domain containing 4 pseudogene', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143045127, end_position_on_the_genomic_accession=143045875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287015, CurrentID=0, Status='live', Symbol='LOC100287015', Aliases='', description='uncharacterized LOC100287015', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6403556, end_position_on_the_genomic_accession=6406548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287016, CurrentID=0, Status='live', Symbol='DSTNP3', Aliases='', description='destrin (actin depolymerizing factor) pseudogene 3', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73984394, end_position_on_the_genomic_accession=73984884, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148014, CurrentID=0, Status='live', Symbol='TTC9B', Aliases='', description='tetratricopeptide repeat domain 9B', other_designations='TPR repeat protein 9B', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40216058, end_position_on_the_genomic_accession=40218399, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287026, CurrentID=0, Status='live', Symbol='TTTY25P', Aliases='', description='testis-specific transcript, Y-linked 25, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22331532, end_position_on_the_genomic_accession=22332002, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287028, CurrentID=0, Status='live', Symbol='LOC100287028', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49891839, end_position_on_the_genomic_accession=49898841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287029, CurrentID=0, Status='live', Symbol='DDX11L10', Aliases='DDX11L1, DDX11P', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 10', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 10|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 pseudogene|DEAD/H box polypeptide 11 like 10', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11555, end_position_on_the_genomic_accession=14090, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  148022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148022, CurrentID=0, Status='live', Symbol='TICAM1', Aliases='IIAE6, MyD88-3, PRVTIRB, TICAM-1, TRIF', description='toll-like receptor adaptor molecule 1', other_designations='TIR domain containing adaptor inducing interferon-beta|TIR domain-containing adapter protein inducing IFN-beta|proline-rich, vinculin and TIR domain-containing protein B|putative NF-kappa-B-activating protein 502H|toll-interleukin-1 receptor domain-containing adapter protein inducing interferon beta', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4815924, end_position_on_the_genomic_accession=4831742, orientation='minus', exon_count=2, OMIM=607601, no_hdr0=''),
  100287033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287033, CurrentID=0, Status='live', Symbol='LOC100287033', Aliases='', description='TSC22 domain family member 3 pseudogene', other_designations='TSC22 domain family, member 3 pseudogene', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91422887, end_position_on_the_genomic_accession=91424416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287036, CurrentID=0, Status='live', Symbol='LOC100287036', Aliases='', description='uncharacterized LOC100287036', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89321133, end_position_on_the_genomic_accession=89325110, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373178, CurrentID=0, Status='live', Symbol='LOC105373178', Aliases='', description='uncharacterized LOC105373178', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39759837, end_position_on_the_genomic_accession=39766745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287042, CurrentID=0, Status='live', Symbol='LOC100287042', Aliases='', description='uncharacterized LOC100287042', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75271299, end_position_on_the_genomic_accession=75273895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131159, CurrentID=0, Status='live', Symbol='LOC100131159', Aliases='', description='PERP, TP53 apoptosis effector pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16160934, end_position_on_the_genomic_accession=16162667, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287045, CurrentID=0, Status='live', Symbol='FAM90A26', Aliases='FAM90A26P', description='family with sequence similarity 90 member A26', other_designations='family with sequence similarity 90, member A26', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9170388, end_position_on_the_genomic_accession=9176730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287046, CurrentID=0, Status='live', Symbol='LOC100287046', Aliases='', description='glucuronidase, beta pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69910374, end_position_on_the_genomic_accession=69920867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287049, CurrentID=0, Status='live', Symbol='LOC100287049', Aliases='', description='uncharacterized LOC100287049', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=160231518, end_position_on_the_genomic_accession=160232568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  148046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148046, CurrentID=0, Status='live', Symbol='CIRBP-AS1', Aliases='C19orf23', description='CIRBP antisense RNA 1', other_designations='CIRBP antisense RNA 1 (non-protein coding)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1267471, end_position_on_the_genomic_accession=1270260, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287055, CurrentID=0, Status='live', Symbol='TSPY21P', Aliases='', description='testis specific protein, Y-linked 21, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22394574, end_position_on_the_genomic_accession=22402820, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372801, CurrentID=0, Status='live', Symbol='LOC105372801', Aliases='', description='uncharacterized LOC105372801', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38235677, end_position_on_the_genomic_accession=38260085, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287063, CurrentID=0, Status='live', Symbol='GEMIN2P2', Aliases='', description='gem nuclear organelle associated protein 2 pseudogene 2', other_designations='gem (nuclear organelle) associated protein 2 pseudogene 2', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41977028, end_position_on_the_genomic_accession=41977860, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927584, CurrentID=0, Status='live', Symbol='LOC101927584', Aliases='', description='uncharacterized LOC101927584', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46717521, end_position_on_the_genomic_accession=46719561, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287067, CurrentID=0, Status='live', Symbol='MTCO1P51', Aliases='', description='MT-CO1 pseudogene 51', other_designations='mitochondrially encoded cytochrome c oxidase I pseudogene', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80563554, end_position_on_the_genomic_accession=80564631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287072, CurrentID=0, Status='live', Symbol='LOC100287072', Aliases='', description='ribosomal protein S6 kinase, 70kDa, polypeptide 1 pseudogene', other_designations='ribosomal protein S6 kinase beta-1-like', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20712446, end_position_on_the_genomic_accession=20737865, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  148066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148066, CurrentID=0, Status='live', Symbol='ZNRF4', Aliases='RNF204, SPERIZIN, Ssrzf1, spzn', description='zinc and ring finger 4', other_designations='RING finger protein 204|nixin', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5455415, end_position_on_the_genomic_accession=5456856, orientation='plus', exon_count=1, OMIM=612063, no_hdr0=''),
  100287082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287082, CurrentID=0, Status='live', Symbol='GACAT2', Aliases='MTCL1-AS1, MTCL1AS1', description='gastric cancer associated transcript 2 (non-protein coding)', other_designations='HMlincRNA717|MTCL1 antisense RNA 1', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8695856, end_position_on_the_genomic_accession=8707621, orientation='minus', exon_count=2, OMIM=616131, no_hdr0=''),
  100287083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287083, CurrentID=0, Status='live', Symbol='DEFT1P2', Aliases='', description='defensin, theta 1 pseudogene 2', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7006281, end_position_on_the_genomic_accession=7008824, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287084, CurrentID=0, Status='live', Symbol='ZNF72P', Aliases='Cos8, ZNF72', description='zinc finger protein 72, pseudogene', other_designations='zinc finger protein 72 (Cos8)', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=16634990, end_position_on_the_genomic_accession=16652137, orientation='plus', exon_count=0, OMIM=194546, no_hdr0=''),
  100287085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287085, CurrentID=0, Status='live', Symbol='OCIAD2P1', Aliases='', description='OCIA domain containing 2 pseudogene 1', other_designations='OCIAD2 pseudogene 1', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80567965, end_position_on_the_genomic_accession=80568578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287087, CurrentID=0, Status='live', Symbol='H3F3BP1', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene 1', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71675067, end_position_on_the_genomic_accession=71675471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287091, CurrentID=0, Status='live', Symbol='ATP5LP5', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 5', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 5|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 5', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152800434, end_position_on_the_genomic_accession=152826545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287098, CurrentID=0, Status='live', Symbol='TONSL-AS1', Aliases='', description='TONSL antisense RNA 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144435168, end_position_on_the_genomic_accession=144439893, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373180, CurrentID=0, Status='live', Symbol='LOC105373180', Aliases='', description='uncharacterized LOC105373180', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39928981, end_position_on_the_genomic_accession=39939216, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100287102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287102, CurrentID=0, Status='live', Symbol='DDX11L1', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 1', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 1|DEAD/H box polypeptide 11 like 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11874, end_position_on_the_genomic_accession=14409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  148103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148103, CurrentID=0, Status='live', Symbol='ZNF599', Aliases='', description='zinc finger protein 599', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34758074, end_position_on_the_genomic_accession=34773229, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287114, CurrentID=0, Status='live', Symbol='LINC00421', Aliases='', description='long intergenic non-protein coding RNA 421', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19345049, end_position_on_the_genomic_accession=19346749, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  148109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148109, CurrentID=0, Status='live', Symbol='FAM187B', Aliases='TMEM162', description='family with sequence similarity 187 member B', other_designations='family with sequence similarity 187, member B|transmembrane protein 162', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35224801, end_position_on_the_genomic_accession=35228725, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  148113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148113, CurrentID=0, Status='live', Symbol='CILP2', Aliases='CLIP-2', description='cartilage intermediate layer protein 2', other_designations='CILP-2|cartilage intermediate layer protein-like protein 2', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19538265, end_position_on_the_genomic_accession=19546659, orientation='plus', exon_count=8, OMIM=612419, no_hdr0=''),
  100287128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287128, CurrentID=0, Status='live', Symbol='IGHV1OR15-2', Aliases='IGHV1OR152', description='immunoglobulin heavy variable 1/OR15-2 (pseudogene)', other_designations='IGHV1/OR15-2|immunoglobulin heavy variable 1/OR15-2 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=19972757, end_position_on_the_genomic_accession=19973271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373181, CurrentID=0, Status='live', Symbol='LOC105373181', Aliases='', description='uncharacterized LOC105373181', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40170647, end_position_on_the_genomic_accession=40176262, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873552, CurrentID=0, Status='live', Symbol='RNA5SP501', Aliases='RN5S501', description='RNA, 5S ribosomal pseudogene 501', other_designations='RNA, 5S ribosomal 501', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=32102065, end_position_on_the_genomic_accession=32102176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287136, CurrentID=0, Status='live', Symbol='CICP28', Aliases='', description='capicua transcriptional repressor pseudogene 28', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56805339, end_position_on_the_genomic_accession=56807653, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271057, CurrentID=0, Status='live', Symbol='RPS12P13', Aliases='RPS12_5_735', description='ribosomal protein S12 pseudogene 13', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108856353, end_position_on_the_genomic_accession=108856737, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287140, CurrentID=0, Status='live', Symbol='HSPD1P14', Aliases='HSPD1-17P', description='heat shock protein family D (Hsp60) member 1 pseudogene 14', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 14', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33838521, end_position_on_the_genomic_accession=33840137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287144, CurrentID=0, Status='live', Symbol='USP17L10', Aliases='', description='ubiquitin specific peptidase 17-like family member 10', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9210657, end_position_on_the_genomic_accession=9212249, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  148137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148137, CurrentID=0, Status='live', Symbol='PROSER3', Aliases='C19orf55', description='proline and serine rich 3', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35758143, end_position_on_the_genomic_accession=35769371, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100287148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287148, CurrentID=0, Status='live', Symbol='LOC100287148', Aliases='', description='peroxiredoxin 6 pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95100531, end_position_on_the_genomic_accession=95101048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148145, CurrentID=0, Status='live', Symbol='LINC00906', Aliases='', description='long intergenic non-protein coding RNA 906', other_designations='CTD-2081K17.1', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28965131, end_position_on_the_genomic_accession=28969148, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287157, CurrentID=0, Status='live', Symbol='LOC100287157', Aliases='', description='cell division cycle associated 8 pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48191959, end_position_on_the_genomic_accession=48194594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131393, CurrentID=0, Status='live', Symbol='UBQLN4P2', Aliases='', description='ubiquilin 4 pseudogene 2', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152876744, end_position_on_the_genomic_accession=152878730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287160, CurrentID=0, Status='live', Symbol='LOC100287160', Aliases='', description='uncharacterized LOC100287160', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=19855407, end_position_on_the_genomic_accession=19856269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373182, CurrentID=0, Status='live', Symbol='LOC105373182', Aliases='', description='uncharacterized LOC105373182', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40318229, end_position_on_the_genomic_accession=40322408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  148156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148156, CurrentID=0, Status='live', Symbol='ZNF558', Aliases='', description='zinc finger protein 558', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8809574, end_position_on_the_genomic_accession=8832304, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100287166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287166, CurrentID=0, Status='live', Symbol='LOC100287166', Aliases='', description='uncharacterized LOC100287166', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19756390, end_position_on_the_genomic_accession=19758561, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287167, CurrentID=0, Status='live', Symbol='TTTY29P', Aliases='', description='testis-specific transcript, Y-linked 29, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=25672802, end_position_on_the_genomic_accession=25673201, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271058, CurrentID=0, Status='live', Symbol='RPS12P14', Aliases='RPS12_6_698', description='ribosomal protein S12 pseudogene 14', other_designations='', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47562535, end_position_on_the_genomic_accession=47563043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287171, CurrentID=0, Status='live', Symbol='WASH1', Aliases='FAM39E, WASH', description='WAS protein family homolog 1', other_designations='CXYorf1-like protein on chromosome 9|FLJ00075 protein|family with sequence similarity 39, member E', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14511, end_position_on_the_genomic_accession=30487, orientation='minus', exon_count=16, OMIM=613632, no_hdr0=''),
  100287173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287173, CurrentID=0, Status='live', Symbol='UBE2V1P5', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 5', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 5|ubiquitin-conjugating enzyme E2 variant 2 pseudogene', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114117147, end_position_on_the_genomic_accession=114117553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287175, CurrentID=0, Status='live', Symbol='LOC100287175', Aliases='', description='uncharacterized LOC100287175', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=636443, end_position_on_the_genomic_accession=638620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287177, CurrentID=0, Status='live', Symbol='EML2-AS1', Aliases='C19orf83', description='EML2 antisense RNA 1', other_designations='putative uncharacterized protein C19orf83', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45641494, end_position_on_the_genomic_accession=45642840, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287178, CurrentID=0, Status='live', Symbol='USP17L11', Aliases='', description='ubiquitin specific peptidase 17-like family member 11', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9215405, end_position_on_the_genomic_accession=9216997, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  221303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221303, CurrentID=0, Status='live', Symbol='FAM162B', Aliases='C6orf189, bA86F4.2', description='family with sequence similarity 162 member B', other_designations='family with sequence similarity 162, member B', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116752197, end_position_on_the_genomic_accession=116765723, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287183, CurrentID=0, Status='live', Symbol='NUTF2P8', Aliases='', description='nuclear transport factor 2 pseudogene 8', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21361899, end_position_on_the_genomic_accession=21364120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287188, CurrentID=0, Status='live', Symbol='LOC100287188', Aliases='', description='family with sequence similarity 104, member B pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287189, CurrentID=0, Status='live', Symbol='LOC100287189', Aliases='', description='Sec61 translocon beta subunit pseudogene', other_designations='Sec61 beta subunit pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47905128, end_position_on_the_genomic_accession=47905683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287191, CurrentID=0, Status='live', Symbol='HNRNPA1P47', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 47', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194187298, end_position_on_the_genomic_accession=194188509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287193, CurrentID=0, Status='live', Symbol='ISCA2P1', Aliases='', description='iron-sulfur cluster assembly 2 pseudogene 1', other_designations='iron-sulfur cluster assembly 2 homolog pseudogene 1', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26644887, end_position_on_the_genomic_accession=26645336, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148189, CurrentID=0, Status='live', Symbol='LINC00662', Aliases='', description='long intergenic non-protein coding RNA 662', other_designations='CTC-459F4.2', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27790493, end_position_on_the_genomic_accession=27793940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287200, CurrentID=0, Status='live', Symbol='RAB9AP1', Aliases='', description='RAB9A, member RAS oncogene family pseudogene 1', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23751188, end_position_on_the_genomic_accession=23751754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873363, CurrentID=0, Status='live', Symbol='RNA5SP31', Aliases='RN5S31', description='RNA, 5S ribosomal pseudogene 31', other_designations='RNA, 5S ribosomal 31', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60613608, end_position_on_the_genomic_accession=60613734, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287205, CurrentID=0, Status='live', Symbol='USP17L12', Aliases='', description='ubiquitin specific peptidase 17-like family member 12', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9220152, end_position_on_the_genomic_accession=9221744, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  148198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148198, CurrentID=0, Status='live', Symbol='ZNF98', Aliases='F7175, ZNF739', description='zinc finger protein 98', other_designations='zinc finger protein 739|zinc finger protein F7175', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22391011, end_position_on_the_genomic_accession=22422362, orientation='minus', exon_count=4, OMIM=603980, no_hdr0=''),
  100287210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287210, CurrentID=0, Status='live', Symbol='LOC100287210', Aliases='', description='uncharacterized LOC100287210', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=155014613, end_position_on_the_genomic_accession=155015561, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  148203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148203, CurrentID=0, Status='live', Symbol='ZNF738', Aliases='', description='zinc finger protein 738', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21358933, end_position_on_the_genomic_accession=21388582, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101059924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059924, CurrentID=0, Status='live', Symbol='LOC101059924', Aliases='', description='AKT-interacting protein-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52753461, end_position_on_the_genomic_accession=52754694, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148206, CurrentID=0, Status='live', Symbol='ZNF714', Aliases='', description='zinc finger protein 714', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21082147, end_position_on_the_genomic_accession=21125094, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100287215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287215, CurrentID=0, Status='live', Symbol='UBL5P2', Aliases='', description='ubiquitin like 5 pseudogene 2', other_designations='ubiquitin-like 5 pseudogene 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32227777, end_position_on_the_genomic_accession=32228186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287216, CurrentID=0, Status='live', Symbol='SH3RF3-AS1', Aliases='', description='SH3RF3 antisense RNA 1', other_designations='SH3RF3 antisense RNA 1 (non-protein coding)', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109127327, end_position_on_the_genomic_accession=109130119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131395, CurrentID=0, Status='live', Symbol='LOC100131395', Aliases='', description='uncharacterized LOC100131395', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1934391, end_position_on_the_genomic_accession=1938216, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480623, CurrentID=0, Status='live', Symbol='RNU6-943P', Aliases='', description='RNA, U6 small nuclear 943, pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148213, CurrentID=0, Status='live', Symbol='ZNF681', Aliases='', description='zinc finger protein 681', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23739195, end_position_on_the_genomic_accession=23758891, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287225, CurrentID=0, Status='live', Symbol='LOC100287225', Aliases='', description='uncharacterized LOC100287225', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51392042, end_position_on_the_genomic_accession=51562469, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287226, CurrentID=0, Status='live', Symbol='ZNF729', Aliases='', description='zinc finger protein 729', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22286450, end_position_on_the_genomic_accession=22317176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287227, CurrentID=0, Status='live', Symbol='TIPARP-AS1', Aliases='', description='TIPARP antisense RNA 1', other_designations='TIPARP antisense RNA 1 (non-protein coding)', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156673171, end_position_on_the_genomic_accession=156675713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100287229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287229, CurrentID=0, Status='live', Symbol='BRD7P4', Aliases='', description='bromodomain containing 7 pseudogene 4', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111429709, end_position_on_the_genomic_accession=111431954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148223, CurrentID=0, Status='live', Symbol='C19orf25', Aliases='', description='chromosome 19 open reading frame 25', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1473201, end_position_on_the_genomic_accession=1479229, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873364, CurrentID=0, Status='live', Symbol='RNA5SP32', Aliases='RN5S32', description='RNA, 5S ribosomal pseudogene 32', other_designations='RNA, 5S ribosomal 32', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71978088, end_position_on_the_genomic_accession=71978206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148229, CurrentID=0, Status='live', Symbol='ATP8B3', Aliases='ATPIK', description='ATPase, aminophospholipid transporter, class I, type 8B, member 3', other_designations='ATPase, class I, type 8B, member 3|aminophospholipid translocase ATP8B3|potential phospholipid-transporting ATPase IK|probable phospholipid-transporting ATPase IK', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1782070, end_position_on_the_genomic_accession=1812276, orientation='minus', exon_count=30, OMIM=605866, no_hdr0=''),
  100287238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287238, CurrentID=0, Status='live', Symbol='USP17L13', Aliases='', description='ubiquitin specific peptidase 17-like family member 13', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9224896, end_position_on_the_genomic_accession=9226488, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  148231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148231, CurrentID=0, Status='live', Symbol='LINC00905', Aliases='', description='long intergenic non-protein coding RNA 905', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16033713, end_position_on_the_genomic_accession=16042135, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100287243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287243, CurrentID=0, Status='live', Symbol='LOC100287243', Aliases='', description='nuclease-sensitive element-binding protein 1 pseudogene', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62953150, end_position_on_the_genomic_accession=62954265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  541461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541461, CurrentID=0, Status='live', Symbol='PSHK2', Aliases='', description='Pseudohyperkalemia, familial, 2, due to red cell leak', other_designations='', map_location='2q35-q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609153, no_hdr0=''),
  541465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541465, CurrentID=0, Status='live', Symbol='CT45A6', Aliases='CT45-6, CT45.6', description='cancer/testis antigen family 45, member A6', other_designations='cancer/testis antigen 45-4/6|cancer/testis antigen 45-6|cancer/testis antigen 45A4/45A6|cancer/testis antigen 45A6|cancer/testis antigen CT45-6|cancer/testis antigen family 45 member A4/A6', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135794687, end_position_on_the_genomic_accession=135812082, orientation='minus', exon_count=7, OMIM=300797, no_hdr0=''),
  541466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541466, CurrentID=0, Status='live', Symbol='CT45A1', Aliases='CT45, CT45-1, CT45.1', description='cancer/testis antigen family 45, member A1', other_designations='cancer/testis antigen 45-1|cancer/testis antigen 45A1|cancer/testis antigen CT45-1', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135708398, end_position_on_the_genomic_accession=135723318, orientation='plus', exon_count=6, OMIM=300648, no_hdr0=''),
  541467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541467, CurrentID=0, Status='live', Symbol='LOC541467', Aliases='', description='uncharacterized LOC541467', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  541468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541468, CurrentID=0, Status='live', Symbol='LURAP1', Aliases='C1orf190, LRAP35a, LRP35A', description='leucine rich adaptor protein 1', other_designations='NF-kappa-B activator C1orf190|leucine repeat adapter protein 35A|leucine repeat adaptor protein 35a', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46203334, end_position_on_the_genomic_accession=46221256, orientation='plus', exon_count=2, OMIM=616129, no_hdr0=''),
  100287261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287261, CurrentID=0, Status='live', Symbol='CSPG4P3Y', Aliases='CSPG4P3, CSPG4P6Y', description='chondroitin sulfate proteoglycan 4 pseudogene 3, Y-linked', other_designations='chondroitin sulfate proteoglycan 4 pseudogene 6, Y-linked', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24187532, end_position_on_the_genomic_accession=24188843, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148254, CurrentID=0, Status='live', Symbol='ZNF555', Aliases='', description='zinc finger protein 555', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2841435, end_position_on_the_genomic_accession=2860474, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  541471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541471, CurrentID=0, Status='live', Symbol='MIR4435-2HG', Aliases='AGD2, AK001796, LINC00978, MIR4435-1HG', description='MIR4435-2 host gene', other_designations='MIR4435-1 host gene (non-protein coding)|adipogenesis down-regulated transcript 2|long intergenic non-protein coding RNA 978', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111367002, end_position_on_the_genomic_accession=111495115, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873365, CurrentID=0, Status='live', Symbol='RNA5SP33', Aliases='RN5S33', description='RNA, 5S ribosomal pseudogene 33', other_designations='RNA, 5S ribosomal 33', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79389534, end_position_on_the_genomic_accession=79389631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  541473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541473, CurrentID=0, Status='live', Symbol='LOC541473', Aliases='', description='FK506 binding protein 6, 36kDa pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75391949, end_position_on_the_genomic_accession=75395461, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100287266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287266, CurrentID=0, Status='live', Symbol='HNRNPA1P42', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 42', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240919629, end_position_on_the_genomic_accession=240920809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287270, CurrentID=0, Status='live', Symbol='USP17L14P', Aliases='', description='ubiquitin specific peptidase 17-like family member 14, pseudogene', other_designations='ubiquitin specific peptidase 17 pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9229641, end_position_on_the_genomic_accession=9231233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287272, CurrentID=0, Status='live', Symbol='USP8P1', Aliases='D84394.7, USP8P', description='ubiquitin specific peptidase 8 pseudogene 1', other_designations='ubiquitin specific protease 8 pseudogene', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31275574, end_position_on_the_genomic_accession=31278751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148266, CurrentID=0, Status='live', Symbol='ZNF569', Aliases='ZAP1, ZNF', description='zinc finger protein 569', other_designations='DNA-binding protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37410876, end_position_on_the_genomic_accession=37469064, orientation='minus', exon_count=10, OMIM=613904, no_hdr0=''),
  100287275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287275, CurrentID=0, Status='live', Symbol='LOC100287275', Aliases='', description='C1QTNF9B antisense RNA 1 pseudogene', other_designations='protein PCOTH pseudogene', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24320954, end_position_on_the_genomic_accession=24321672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148268, CurrentID=0, Status='live', Symbol='ZNF570', Aliases='', description='zinc finger protein 570', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37467465, end_position_on_the_genomic_accession=37485358, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100287279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287279, CurrentID=0, Status='live', Symbol='UBE2V1P6', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 6', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 6|ubiquitin-conjugating enzyme E2 variant 2 pseudogene', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169115195, end_position_on_the_genomic_accession=169115674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373186, CurrentID=0, Status='live', Symbol='LOC105373186', Aliases='', description='uncharacterized LOC105373186', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=42695961, end_position_on_the_genomic_accession=42727047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287284, CurrentID=0, Status='live', Symbol='MANSC4', Aliases='', description='MANSC domain containing 4', other_designations='MANSC domain-containing protein ENSP00000370673', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27762641, end_position_on_the_genomic_accession=27773083, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271062, CurrentID=0, Status='live', Symbol='RPS13P6', Aliases='RPS13_4_614', description='ribosomal protein S13 pseudogene 6', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134267386, end_position_on_the_genomic_accession=134267897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148281, CurrentID=0, Status='live', Symbol='SYT6', Aliases='sytVI', description='synaptotagmin 6', other_designations='synaptotagmin VI', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114089292, end_position_on_the_genomic_accession=114153850, orientation='minus', exon_count=10, OMIM=607718, no_hdr0=''),
  100287290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287290, CurrentID=0, Status='live', Symbol='LOC100287290', Aliases='', description='cytokine receptor CRL2', other_designations='', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158732441, end_position_on_the_genomic_accession=158784039, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100287294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287294, CurrentID=0, Status='live', Symbol='LOC100287294', Aliases='', description='equilibrative nucleoside transporter 4-like', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56936027, end_position_on_the_genomic_accession=56945473, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100287296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287296, CurrentID=0, Status='live', Symbol='LINC00265-2P', Aliases='NCRNA00265-2P, NCRNA00265B', description='long intergenic non-protein coding RNA 265-2, pseudogene', other_designations='non-protein coding RNA 265-2, pseudogene|non-protein coding RNA 265B', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24275987, end_position_on_the_genomic_accession=24277367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287301, CurrentID=0, Status='live', Symbol='GS52', Aliases='', description='uncharacterized LOC100287301', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=757313, end_position_on_the_genomic_accession=759437, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100287302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287302, CurrentID=0, Status='live', Symbol='USP17L16P', Aliases='', description='ubiquitin specific peptidase 17-like family member 16, pseudogene', other_designations='ubiquitin carboxyl-terminal hydrolase 17-like pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9239232, end_position_on_the_genomic_accession=9240254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148300, CurrentID=0, Status='live', Symbol='SONP1', Aliases='', description='SON pseudogene 1', other_designations='SON DNA binding protein pseudogene|son-pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157708996, end_position_on_the_genomic_accession=157711485, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287309, CurrentID=0, Status='live', Symbol='LOC100287309', Aliases='', description='abhydrolase domain containing 15 pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22044295, end_position_on_the_genomic_accession=22045218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148304, CurrentID=0, Status='live', Symbol='C1orf74', Aliases='URLC4', description='chromosome 1 open reading frame 74', other_designations='up-regulated in lung cancer 4', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209782317, end_position_on_the_genomic_accession=209784545, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287314, CurrentID=0, Status='live', Symbol='LINC00941', Aliases='', description='long intergenic non-protein coding RNA 941', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30795681, end_position_on_the_genomic_accession=30802711, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271063, CurrentID=0, Status='live', Symbol='RPS14P1', Aliases='RPS14_1_156', description='ribosomal protein S14 pseudogene 1', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206695782, end_position_on_the_genomic_accession=206696302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374208, CurrentID=0, Status='live', Symbol='LOC105374208', Aliases='', description='uncharacterized LOC105374208', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169775637, end_position_on_the_genomic_accession=169776715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287325, CurrentID=0, Status='live', Symbol='KRT8P37', Aliases='', description='keratin 8 pseudogene 37', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8513499, end_position_on_the_genomic_accession=8515207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287327, CurrentID=0, Status='live', Symbol='USP17L17', Aliases='', description='ubiquitin specific peptidase 17-like family member 17', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9243879, end_position_on_the_genomic_accession=9245471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100287329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287329, CurrentID=0, Status='live', Symbol='LOC100287329', Aliases='', description='uncharacterized LOC100287329', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31548402, end_position_on_the_genomic_accession=31572698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  148327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148327, CurrentID=0, Status='live', Symbol='CREB3L4', Aliases='AIBZIP, ATCE1, CREB3, CREB4, JAL, hJAL', description='cAMP responsive element binding protein 3-like 4', other_designations='CREB-4|androgen-induced basic leucine zipper protein|attaching to CRE-like 1|cAMP responsive element binding protein 1|cAMP-responsive element-binding protein 3-like protein 4|cAMP-responsive element-binding protein 4|cyclic AMP-responsive element-binding protein 4|tisp40|transcript induced in spermiogenesis protein 40', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153967839, end_position_on_the_genomic_accession=153974364, orientation='plus', exon_count=11, OMIM=607138, no_hdr0=''),
  148330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148330, CurrentID=0, Status='live', Symbol='MRPS33P1', Aliases='', description='mitochondrial ribosomal protein S33 pseudogene 1', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154377852, end_position_on_the_genomic_accession=154378218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373188, CurrentID=0, Status='live', Symbol='LOC105373188', Aliases='', description='uncharacterized LOC105373188', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=42920389, end_position_on_the_genomic_accession=42945056, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100287345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287345, CurrentID=0, Status='live', Symbol='VENTXP2', Aliases='', description='VENT homeobox pseudogene 2', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=83907304, end_position_on_the_genomic_accession=83907839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271064, CurrentID=0, Status='live', Symbol='RPS14P2', Aliases='RPS14_2_132', description='ribosomal protein S14 pseudogene 2', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178411584, end_position_on_the_genomic_accession=178412063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148345, CurrentID=0, Status='live', Symbol='C1orf127', Aliases='', description='chromosome 1 open reading frame 127', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10946473, end_position_on_the_genomic_accession=10982037, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100287355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287355, CurrentID=0, Status='live', Symbol='LOC100287355', Aliases='', description='microtubule associated protein 1 light chain 3 beta pseudogene', other_designations='microtubule-associated protein 1 light chain 3 beta pseudogene', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89027689, end_position_on_the_genomic_accession=89028535, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  541565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541565, CurrentID=0, Status='live', Symbol='C8orf58', Aliases='', description='chromosome 8 open reading frame 58', other_designations='', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22599601, end_position_on_the_genomic_accession=22604150, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100287361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287361, CurrentID=0, Status='live', Symbol='LOC100287361', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68125215, end_position_on_the_genomic_accession=68125942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287362, CurrentID=0, Status='live', Symbol='PRSS46', Aliases='TESSP6', description='protease, serine 46', other_designations='protease, serine, 46|testis serine protease 6', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46719583, end_position_on_the_genomic_accession=46736431, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287364, CurrentID=0, Status='live', Symbol='USP17L18', Aliases='', description='ubiquitin specific peptidase 17-like family member 18', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9248630, end_position_on_the_genomic_accession=9250222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100287366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287366, CurrentID=0, Status='live', Symbol='OSTM1-AS1', Aliases='', description='OSTM1 antisense RNA 1', other_designations='OSTM1 antisense RNA 1 (non-protein coding)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108123636, end_position_on_the_genomic_accession=108159395, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287369, CurrentID=0, Status='live', Symbol='HSPE1P4', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 4', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 4', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101873319, end_position_on_the_genomic_accession=101873623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148362, CurrentID=0, Status='live', Symbol='BROX', Aliases='C1orf58', description='BRO1 domain and CAAX motif containing', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222712553, end_position_on_the_genomic_accession=222735196, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100287372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287372, CurrentID=0, Status='live', Symbol='IGHV3OR16-13', Aliases='IGHV3OR1613', description='immunoglobulin heavy variable 3/OR16-13 (non-functional)', other_designations='IGHV3/OR16-13', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33827366, end_position_on_the_genomic_accession=33827659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402429, CurrentID=0, Status='live', Symbol='KRT8P8', Aliases='', description='keratin 8 pseudogene 8', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152479501, end_position_on_the_genomic_accession=152481224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287387, CurrentID=0, Status='live', Symbol='LOC100287387', Aliases='', description='uncharacterized LOC100287387', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238844546, end_position_on_the_genomic_accession=238847679, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287388, CurrentID=0, Status='live', Symbol='LIN28AP2', Aliases='', description='LIN28A pseudogene 2', other_designations='lin-28 homolog A pseudogene 2', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87379466, end_position_on_the_genomic_accession=87380684, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  133957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=133957, CurrentID=0, Status='live', Symbol='CCDC127', Aliases='', description='coiled-coil domain containing 127', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=204760, end_position_on_the_genomic_accession=218182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287399, CurrentID=0, Status='live', Symbol='POTEB2', Aliases='', description='POTE ankyrin domain family member B2', other_designations='POTE ankyrin domain family, member B2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20835372, end_position_on_the_genomic_accession=20866431, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105373190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373190, CurrentID=0, Status='live', Symbol='LOC105373190', Aliases='', description='uncharacterized LOC105373190', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44866863, end_position_on_the_genomic_accession=44870825, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  541611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541611, CurrentID=0, Status='live', Symbol='HSP90AB6P', Aliases='HSP90Bf', description='heat shock protein 90kDa alpha family class B member 6, pseudogene', other_designations='heat shock protein 90Bf|heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class B member 6, pseudogene', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96883419, end_position_on_the_genomic_accession=96885868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287404, CurrentID=0, Status='live', Symbol='USP17L19', Aliases='', description='ubiquitin specific peptidase 17-like family member 19', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9253378, end_position_on_the_genomic_accession=9254970, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  148398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148398, CurrentID=0, Status='live', Symbol='SAMD11', Aliases='', description='sterile alpha motif domain containing 11', other_designations='SAM domain-containing protein 11|sterile alpha motif domain containing 11 splice variant ASV1|sterile alpha motif domain containing 11 splice variant ASV10|sterile alpha motif domain containing 11 splice variant ASV11|sterile alpha motif domain containing 11 splice variant ASV12|sterile alpha motif domain containing 11 splice variant ASV13|sterile alpha motif domain containing 11 splice variant ASV14|sterile alpha motif domain containing 11 splice variant ASV15|sterile alpha motif domain containing 11 splice variant ASV16|sterile alpha motif domain containing 11 splice variant ASV17|sterile alpha motif domain containing 11 splice variant ASV18|sterile alpha motif domain containing 11 splice variant ASV19|sterile alpha motif domain containing 11 splice variant ASV2|sterile alpha motif domain containing 11 splice variant ASV20|sterile alpha motif domain containing 11 splice variant ASV21|sterile alpha motif domain containing 11 splice variant ASV22|sterile alpha motif domain containing 11 splice variant ASV23|sterile alpha motif domain containing 11 splice variant ASV24|sterile alpha motif domain containing 11 splice variant ASV25|sterile alpha motif domain containing 11 splice variant ASV26|sterile alpha motif domain containing 11 splice variant ASV27|sterile alpha motif domain containing 11 splice variant ASV28|sterile alpha motif domain containing 11 splice variant ASV29|sterile alpha motif domain containing 11 splice variant ASV3|sterile alpha motif domain containing 11 splice variant ASV30|sterile alpha motif domain containing 11 splice variant ASV31|sterile alpha motif domain containing 11 splice variant ASV32|sterile alpha motif domain containing 11 splice variant ASV33|sterile alpha motif domain containing 11 splice variant ASV34|sterile alpha motif domain containing 11 splice variant ASV35|sterile alpha motif domain containing 11 splice variant ASV36|sterile alpha motif domain containing 11 splice variant ASV37|sterile alpha motif domain containing 11 splice variant ASV38|sterile alpha motif domain containing 11 splice variant ASV39|sterile alpha motif domain containing 11 splice variant ASV4|sterile alpha motif domain containing 11 splice variant ASV40|sterile alpha motif domain containing 11 splice variant ASV41|sterile alpha motif domain containing 11 splice variant ASV42|sterile alpha motif domain containing 11 splice variant ASV43|sterile alpha motif domain containing 11 splice variant ASV44|sterile alpha motif domain containing 11 splice variant ASV45|sterile alpha motif domain containing 11 splice variant ASV5|sterile alpha motif domain containing 11 splice variant ASV6|sterile alpha motif domain containing 11 splice variant ASV7|sterile alpha motif domain containing 11 splice variant ASV8|sterile alpha motif domain containing 11 splice variant ASV9', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=925741, end_position_on_the_genomic_accession=944581, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  284064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284064, CurrentID=0, Status='live', Symbol='RPL29P31', Aliases='RPL29_11_1549', description='ribosomal protein L29 pseudogene 31', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43476154, end_position_on_the_genomic_accession=43476769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418484, CurrentID=0, Status='live', Symbol='LOC100418484', Aliases='', description='KAT8 regulatory NSL complex subunit 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418485, CurrentID=0, Status='live', Symbol='LOC100418485', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418486, CurrentID=0, Status='live', Symbol='LOC100418486', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418487, CurrentID=0, Status='live', Symbol='LOC100418487', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418488, CurrentID=0, Status='live', Symbol='LOC100418488', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418489, CurrentID=0, Status='live', Symbol='LOC100418489', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418490, CurrentID=0, Status='live', Symbol='LOC100418490', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418491, CurrentID=0, Status='live', Symbol='LOC100418491', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418492, CurrentID=0, Status='live', Symbol='LOC100418492', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148413, CurrentID=0, Status='live', Symbol='LOC148413', Aliases='', description='uncharacterized LOC148413', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1399530, end_position_on_the_genomic_accession=1402046, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418494, CurrentID=0, Status='live', Symbol='LOC100418494', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418495, CurrentID=0, Status='live', Symbol='LOC100418495', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418496, CurrentID=0, Status='live', Symbol='LOC100418496', Aliases='', description='fibrosin-like 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418497, CurrentID=0, Status='live', Symbol='LOC100418497', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148418, CurrentID=0, Status='live', Symbol='SAMD13', Aliases='HSD-41, HSD-42', description='sterile alpha motif domain containing 13', other_designations='SAM domain-containing protein 13|dnaj-like protein (LOC148418)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84274952, end_position_on_the_genomic_accession=84350798, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100418499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418499, CurrentID=0, Status='live', Symbol='LOC100418499', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418500, CurrentID=0, Status='live', Symbol='LOC100418500', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418501, CurrentID=0, Status='live', Symbol='LOC100418501', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418502, CurrentID=0, Status='live', Symbol='LOC100418502', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148423, CurrentID=0, Status='live', Symbol='C1orf52', Aliases='gm117', description='chromosome 1 open reading frame 52', other_designations='BCL10-associated gene protein', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85249953, end_position_on_the_genomic_accession=85259672, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373191, CurrentID=0, Status='live', Symbol='LOC105373191', Aliases='', description='uncharacterized LOC105373191', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45324908, end_position_on_the_genomic_accession=45333307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100418505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418505, CurrentID=0, Status='live', Symbol='LOC100418505', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418506, CurrentID=0, Status='live', Symbol='LOC100418506', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418507, CurrentID=0, Status='live', Symbol='LOC100418507', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418508, CurrentID=0, Status='live', Symbol='LOC100418508', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418509, CurrentID=0, Status='live', Symbol='LOC100418509', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271067, CurrentID=0, Status='live', Symbol='RPS16P3', Aliases='RPS16_2_279', description='ribosomal protein S16 pseudogene 3', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=141208082, end_position_on_the_genomic_accession=141208435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418511, CurrentID=0, Status='live', Symbol='LOC100418511', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418512, CurrentID=0, Status='live', Symbol='LOC100418512', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287441, CurrentID=0, Status='live', Symbol='USP17L20', Aliases='', description='ubiquitin specific peptidase 17-like family member 20', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9258124, end_position_on_the_genomic_accession=9259716, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418514, CurrentID=0, Status='live', Symbol='LOC100418514', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418515, CurrentID=0, Status='live', Symbol='LOC100418515', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418516, CurrentID=0, Status='live', Symbol='LOC100418516', Aliases='', description='lysophospholipase II pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418517, CurrentID=0, Status='live', Symbol='LOC100418517', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418518, CurrentID=0, Status='live', Symbol='LOC100418518', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418519, CurrentID=0, Status='live', Symbol='LOC100418519', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418520, CurrentID=0, Status='live', Symbol='LOC100418520', Aliases='', description='lysophospholipase II pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418521, CurrentID=0, Status='live', Symbol='LOC100418521', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418522, CurrentID=0, Status='live', Symbol='LOC100418522', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418523, CurrentID=0, Status='live', Symbol='LOC100418523', Aliases='', description='lysophospholipase II pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418524, CurrentID=0, Status='live', Symbol='LOC100418524', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418525, CurrentID=0, Status='live', Symbol='LOC100418525', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418526, CurrentID=0, Status='live', Symbol='LOC100418526', Aliases='', description='ribosomal protein SA pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418527, CurrentID=0, Status='live', Symbol='LOC100418527', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418528, CurrentID=0, Status='live', Symbol='LOC100418528', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418529, CurrentID=0, Status='live', Symbol='LOC100418529', Aliases='', description='ribosomal protein L13 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418530, CurrentID=0, Status='live', Symbol='LOC100418530', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418531, CurrentID=0, Status='live', Symbol='LOC100418531', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418532, CurrentID=0, Status='live', Symbol='LOC100418532', Aliases='', description='ribosomal protein L13 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287461, CurrentID=0, Status='live', Symbol='MTND1P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 8', other_designations='MT-ND1 pseudogene 8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3369950, end_position_on_the_genomic_accession=3370905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418534, CurrentID=0, Status='live', Symbol='LOC100418534', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418535, CurrentID=0, Status='live', Symbol='LOC100418535', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418536, CurrentID=0, Status='live', Symbol='LOC100418536', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418537, CurrentID=0, Status='live', Symbol='LOC100418537', Aliases='', description='ribosomal protein L13 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418538, CurrentID=0, Status='live', Symbol='LOC100418538', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271078, CurrentID=0, Status='live', Symbol='RPS19P6', Aliases='RPS19_2_988', description='ribosomal protein S19 pseudogene 6', other_designations='', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80782332, end_position_on_the_genomic_accession=80782770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271068, CurrentID=0, Status='live', Symbol='RPS16P4', Aliases='RPS16_2_363', description='ribosomal protein S16 pseudogene 4', other_designations='', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37196234, end_position_on_the_genomic_accession=37196669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418541, CurrentID=0, Status='live', Symbol='LOC100418541', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287470, CurrentID=0, Status='live', Symbol='LINC00265-3P', Aliases='NCRNA00265-3P, NCRNA00265C', description='long intergenic non-protein coding RNA 265-3, pseudogene', other_designations='non-protein coding RNA 265-3, pseudogene|non-protein coding RNA 265C', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25392781, end_position_on_the_genomic_accession=25394160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418543, CurrentID=0, Status='live', Symbol='LOC100418543', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287472, CurrentID=0, Status='live', Symbol='NPM1P31', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 31', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124867417, end_position_on_the_genomic_accession=124868531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418545, CurrentID=0, Status='live', Symbol='LOC100418545', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873372, CurrentID=0, Status='live', Symbol='RNA5SP101', Aliases='RN5S101', description='RNA, 5S ribosomal pseudogene 101', other_designations='RNA, 5S ribosomal 101', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96956708, end_position_on_the_genomic_accession=96956815, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418547, CurrentID=0, Status='live', Symbol='LOC100418547', Aliases='', description='zinc finger protein 93 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418548, CurrentID=0, Status='live', Symbol='LOC100418548', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287477, CurrentID=0, Status='live', Symbol='LOC100287477', Aliases='', description='zinc finger protein 420-like', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50058266, end_position_on_the_genomic_accession=50060497, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287478, CurrentID=0, Status='live', Symbol='USP17L21', Aliases='', description='ubiquitin specific peptidase 17-like family member 21', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9262872, end_position_on_the_genomic_accession=9264464, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418551, CurrentID=0, Status='live', Symbol='LOC100418551', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418552, CurrentID=0, Status='live', Symbol='LOC100418552', Aliases='', description='zinc finger protein 93 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418553, CurrentID=0, Status='live', Symbol='LOC100418553', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287482, CurrentID=0, Status='live', Symbol='SMKR1', Aliases='', description='small lysine-rich protein 1', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129502479, end_position_on_the_genomic_accession=129512932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287483, CurrentID=0, Status='live', Symbol='TCEB1P28', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 28', other_designations='transcription elongation factor B (SIII), polypeptide 1 (15kDa, elongin C) pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2537735, end_position_on_the_genomic_accession=2539309, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418556, CurrentID=0, Status='live', Symbol='LOC100418556', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287485, CurrentID=0, Status='live', Symbol='LOC100287485', Aliases='', description='glycine-N-acyltransferase like 2 pseudogene', other_designations='glycine-N-acyltransferase-like 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58995766, end_position_on_the_genomic_accession=58999689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418558, CurrentID=0, Status='live', Symbol='LOC100418558', Aliases='', description='dihydrofolate reductase pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148479, CurrentID=0, Status='live', Symbol='PHF13', Aliases='PHF5, SPOC1', description='PHD finger protein 13', other_designations='PHD zinc finger protein PHF5|survival time-associated PHD finger protein in ovarian cancer 1|survival time-associated PHD protein in ovarian cancer', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6613696, end_position_on_the_genomic_accession=6624033, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418560, CurrentID=0, Status='live', Symbol='LOC100418560', Aliases='', description='dihydrofolate reductase pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418561, CurrentID=0, Status='live', Symbol='LOC100418561', Aliases='', description='ADP ribosylation factor 3 pseudogene', other_designations='ADP-ribosylation factor 3 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103695363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695363, CurrentID=0, Status='live', Symbol='FOXG1-AS1', Aliases='FOXG1-AS', description='FOXG1 antisense RNA 1 (head to head)', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28725242, end_position_on_the_genomic_accession=28765319, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  103695364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695364, CurrentID=0, Status='live', Symbol='PKNOX2-AS1', Aliases='', description='PKNOX2 antisense RNA 1 (head to head)', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125158461, end_position_on_the_genomic_accession=125164609, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103695365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695365, CurrentID=0, Status='live', Symbol='LINC01435', Aliases='TCONS_00018040', description='long intergenic non-protein coding RNA 1435', other_designations='RP11-215N21.1', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=107871576, end_position_on_the_genomic_accession=108069293, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  103695366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695366, CurrentID=0, Status='live', Symbol='LINC01461', Aliases='TCONS_00001552', description='long intergenic non-protein coding RNA 1461', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84597964, end_position_on_the_genomic_accession=84621031, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287495, CurrentID=0, Status='live', Symbol='GCSHP4', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene 4', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8140500, end_position_on_the_genomic_accession=8142591, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287497, CurrentID=0, Status='live', Symbol='LOC100287497', Aliases='', description='uncharacterized LOC100287497', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224008590, end_position_on_the_genomic_accession=224009969, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271069, CurrentID=0, Status='live', Symbol='RPS17P7', Aliases='RPS17_1_271', description='ribosomal protein S17 pseudogene 7', other_designations='', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121309996, end_position_on_the_genomic_accession=121310416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418571, CurrentID=0, Status='live', Symbol='LOC100418571', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418572, CurrentID=0, Status='live', Symbol='LOC100418572', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418573, CurrentID=0, Status='live', Symbol='LOC100418573', Aliases='', description='dihydrofolate reductase like 1 pseudogene', other_designations='dihydrofolate reductase-like 1 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287502, CurrentID=0, Status='live', Symbol='CSPG4P4Y', Aliases='CSPG4P4, CSPG4P7Y', description='chondroitin sulfate proteoglycan 4 pseudogene 4, Y-linked', other_designations='chondroitin sulfate proteoglycan 4 pseudogene 7, Y-linked', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25481296, end_position_on_the_genomic_accession=25482607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418575, CurrentID=0, Status='live', Symbol='LOC100418575', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418576, CurrentID=0, Status='live', Symbol='LOC100418576', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287505, CurrentID=0, Status='live', Symbol='BRWD1P2', Aliases='', description='bromodomain and WD repeat domain containing 1 pseudogene 2', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89919421, end_position_on_the_genomic_accession=89919774, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418578, CurrentID=0, Status='live', Symbol='LOC100418578', Aliases='', description='RAN, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418579, CurrentID=0, Status='live', Symbol='LOC100418579', Aliases='', description='RAN, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418580, CurrentID=0, Status='live', Symbol='LOC100418580', Aliases='', description='RAN, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418581, CurrentID=0, Status='live', Symbol='LOC100418581', Aliases='', description='RAN, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418582, CurrentID=0, Status='live', Symbol='LOC100418582', Aliases='', description='RAN, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418583, CurrentID=0, Status='live', Symbol='LOC100418583', Aliases='', description='mucin 4, cell surface associated pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287512, CurrentID=0, Status='live', Symbol='RPS3AP53', Aliases='', description='ribosomal protein S3a pseudogene 53', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15141356, end_position_on_the_genomic_accession=15142204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287513, CurrentID=0, Status='live', Symbol='USP17L22', Aliases='', description='ubiquitin specific peptidase 17-like family member 22', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9267619, end_position_on_the_genomic_accession=9269211, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418586, CurrentID=0, Status='live', Symbol='LOC100418586', Aliases='', description='N-ethylmaleimide-sensitive factor pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418587, CurrentID=0, Status='live', Symbol='LOC100418587', Aliases='', description='myosin light chain 12B pseudogene', other_designations='myosin, light chain 12B, regulatory pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418588, CurrentID=0, Status='live', Symbol='LOC100418588', Aliases='', description='myosin light chain 12B pseudogene', other_designations='myosin, light chain 12B, regulatory pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418589, CurrentID=0, Status='live', Symbol='LOC100418589', Aliases='', description='myosin light chain 12B pseudogene', other_designations='myosin, light chain 12B, regulatory pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418590, CurrentID=0, Status='live', Symbol='LOC100418590', Aliases='', description='myosin light chain 12B pseudogene', other_designations='myosin, light chain 12B, regulatory pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418591, CurrentID=0, Status='live', Symbol='LOC100418591', Aliases='', description='tigger transposable element derived 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480625, CurrentID=0, Status='live', Symbol='RNU6-966P', Aliases='', description='RNA, U6 small nuclear 966, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418593, CurrentID=0, Status='live', Symbol='LOC100418593', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418594, CurrentID=0, Status='live', Symbol='LOC100418594', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418595, CurrentID=0, Status='live', Symbol='LOC100418595', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418596, CurrentID=0, Status='live', Symbol='LOC100418596', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287525, CurrentID=0, Status='live', Symbol='LOC100287525', Aliases='', description='uncharacterized LOC100287525', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=38517744, end_position_on_the_genomic_accession=38521704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418598, CurrentID=0, Status='live', Symbol='LOC100418598', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418599, CurrentID=0, Status='live', Symbol='LOC100418599', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287528, CurrentID=0, Status='live', Symbol='IGLL4P', Aliases='IGLL4', description='immunoglobulin lambda-like polypeptide 4, pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20632590, end_position_on_the_genomic_accession=20632826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418601, CurrentID=0, Status='live', Symbol='LOC100418601', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418602, CurrentID=0, Status='live', Symbol='LOC100418602', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148523, CurrentID=0, Status='live', Symbol='CIART', Aliases='C1orf51, CHRONO, GM129', description='circadian associated repressor of transcription', other_designations='chIP-derived repressor of network oscillator|computationally highlighted repressor of the network oscillator', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150282533, end_position_on_the_genomic_accession=150287093, orientation='plus', exon_count=7, OMIM=615782, no_hdr0=''),
  100418604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418604, CurrentID=0, Status='live', Symbol='LOC100418604', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418605, CurrentID=0, Status='live', Symbol='LOC100418605', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270826, CurrentID=0, Status='live', Symbol='RPS2P13', Aliases='RPS2_1_34', description='ribosomal protein S2 pseudogene 13', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37840845, end_position_on_the_genomic_accession=37841680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418607, CurrentID=0, Status='live', Symbol='LOC100418607', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418608, CurrentID=0, Status='live', Symbol='LOC100418608', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418609, CurrentID=0, Status='live', Symbol='LOC100418609', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418610, CurrentID=0, Status='live', Symbol='LOC100418610', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418611, CurrentID=0, Status='live', Symbol='LOC100418611', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418612, CurrentID=0, Status='live', Symbol='LOC100418612', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418613, CurrentID=0, Status='live', Symbol='LOC100418613', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148534, CurrentID=0, Status='live', Symbol='TMEM56', Aliases='', description='transmembrane protein 56', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95092517, end_position_on_the_genomic_accession=95197607, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100418615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418615, CurrentID=0, Status='live', Symbol='LOC100418615', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929333, CurrentID=0, Status='live', Symbol='LOC101929333', Aliases='', description='uncharacterized LOC101929333', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74478068, end_position_on_the_genomic_accession=74490944, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418617, CurrentID=0, Status='live', Symbol='LOC100418617', Aliases='', description='major histocompatibility complex, class I, B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418618, CurrentID=0, Status='live', Symbol='LOC100418618', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418619, CurrentID=0, Status='live', Symbol='LOC100418619', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418620, CurrentID=0, Status='live', Symbol='LOC100418620', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418621, CurrentID=0, Status='live', Symbol='LOC100418621', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418622, CurrentID=0, Status='live', Symbol='LOC100418622', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287551, CurrentID=0, Status='live', Symbol='HSPA8P8', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 8', other_designations='heat shock 70kDa protein 8 pseudogene 8', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10451236, end_position_on_the_genomic_accession=10453500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418624, CurrentID=0, Status='live', Symbol='LOC100418624', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148545, CurrentID=0, Status='live', Symbol='NBPF4', Aliases='', description='neuroblastoma breakpoint family member 4', other_designations='neuroblastoma breakpoint family, member 4', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108222464, end_position_on_the_genomic_accession=108244081, orientation='minus', exon_count=16, OMIM=613994, no_hdr0=''),
  100418626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418626, CurrentID=0, Status='live', Symbol='LOC100418626', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418627, CurrentID=0, Status='live', Symbol='LOC100418627', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379803, CurrentID=0, Status='live', Symbol='LOC105379803', Aliases='', description='uncharacterized LOC105379803', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418629, CurrentID=0, Status='live', Symbol='LOC100418629', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271071, CurrentID=0, Status='live', Symbol='RPS17P10', Aliases='RPS17_2_414', description='ribosomal protein S17 pseudogene 10', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143855692, end_position_on_the_genomic_accession=143856165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287559, CurrentID=0, Status='live', Symbol='ADPGK-AS1', Aliases='', description='ADPGK antisense RNA 1', other_designations='ADPGK antisense RNA 1 (non-protein coding)', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72782835, end_position_on_the_genomic_accession=72798199, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103695432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695432, CurrentID=0, Status='live', Symbol='LINC01167', Aliases='TCONS_00018075', description='long intergenic non-protein coding RNA 1167', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132961340, end_position_on_the_genomic_accession=132962237, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103695433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695433, CurrentID=0, Status='live', Symbol='LINC01322', Aliases='', description='long intergenic non-protein coding RNA 1322', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=165206960, end_position_on_the_genomic_accession=165513728, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  103695434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695434, CurrentID=0, Status='live', Symbol='LINC01396', Aliases='', description='long intergenic non-protein coding RNA 1396', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4844437, end_position_on_the_genomic_accession=4850827, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103695435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695435, CurrentID=0, Status='live', Symbol='BBOX1-AS1', Aliases='', description='BBOX1 antisense RNA 1', other_designations='', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27047186, end_position_on_the_genomic_accession=27220113, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  103695436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695436, CurrentID=0, Status='live', Symbol='LINC01269', Aliases='', description='long intergenic non-protein coding RNA 1269', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70698698, end_position_on_the_genomic_accession=70712153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100418637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418637, CurrentID=0, Status='live', Symbol='LOC100418637', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418638, CurrentID=0, Status='live', Symbol='LOC100418638', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287567, CurrentID=0, Status='live', Symbol='TRMT112P7', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 7', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54479393, end_position_on_the_genomic_accession=54479993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418640, CurrentID=0, Status='live', Symbol='LOC100418640', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287569, CurrentID=0, Status='live', Symbol='LINC00173', Aliases='NCRNA00173', description='long intergenic non-protein coding RNA 173', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116533422, end_position_on_the_genomic_accession=116536513, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100418642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418642, CurrentID=0, Status='live', Symbol='LOC100418642', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418643, CurrentID=0, Status='live', Symbol='LOC100418643', Aliases='', description='olfactory receptor, family 2, subfamily G, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418644, CurrentID=0, Status='live', Symbol='LOC100418644', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287573, CurrentID=0, Status='live', Symbol='TOMM22P6', Aliases='', description='TOMM22 pseudogene 6', other_designations='translocase of outer mitochondrial membrane 22 homolog pseudogene 6', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=162486519, end_position_on_the_genomic_accession=162486949, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418646, CurrentID=0, Status='live', Symbol='LOC100418646', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418647, CurrentID=0, Status='live', Symbol='LOC100418647', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287576, CurrentID=0, Status='live', Symbol='LOC100287576', Aliases='', description='uncharacterized LOC100287576', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=18984352, end_position_on_the_genomic_accession=18987264, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100287577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287577, CurrentID=0, Status='live', Symbol='RAB9AP3', Aliases='', description='RAB9A, member RAS oncogene family pseudogene 3', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25918329, end_position_on_the_genomic_accession=25918895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418650, CurrentID=0, Status='live', Symbol='LOC100418650', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287579, CurrentID=0, Status='live', Symbol='TCEB1P22', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 22', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120619922, end_position_on_the_genomic_accession=120620270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418652, CurrentID=0, Status='live', Symbol='LOC100418652', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418653, CurrentID=0, Status='live', Symbol='LOC100418653', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418654, CurrentID=0, Status='live', Symbol='LOC100418654', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418655, CurrentID=0, Status='live', Symbol='LOC100418655', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418656, CurrentID=0, Status='live', Symbol='LOC100418656', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418657, CurrentID=0, Status='live', Symbol='LOC100418657', Aliases='', description='olfactory receptor, family 2, subfamily G, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418658, CurrentID=0, Status='live', Symbol='LOC100418658', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418659, CurrentID=0, Status='live', Symbol='LOC100418659', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418660, CurrentID=0, Status='live', Symbol='LOC100418660', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  148581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148581, CurrentID=0, Status='live', Symbol='UBE2U', Aliases='', description='ubiquitin conjugating enzyme E2U (putative)', other_designations='E2 ubiquitin-conjugating enzyme U|ubiquitin carrier protein U|ubiquitin-conjugating enzyme E2U (putative)|ubiquitin-protein ligase U', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64186828, end_position_on_the_genomic_accession=64267368, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100287590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287590, CurrentID=0, Status='live', Symbol='LOC100287590', Aliases='', description='uncharacterized LOC100287590', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=38800482, end_position_on_the_genomic_accession=38805980, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287591, CurrentID=0, Status='live', Symbol='MTND1P22', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 22', other_designations='MT-ND1 pseudogene 22', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155462873, end_position_on_the_genomic_accession=155463823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287592, CurrentID=0, Status='live', Symbol='LOC100287592', Aliases='', description='uncharacterized LOC100287592', other_designations='CTD-2089N3.2', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=50969217, end_position_on_the_genomic_accession=50970187, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418665, CurrentID=0, Status='live', Symbol='LOC100418665', Aliases='', description='olfactory receptor, family 2, subfamily G, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418666, CurrentID=0, Status='live', Symbol='LOC100418666', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873376, CurrentID=0, Status='live', Symbol='RNA5SP105', Aliases='RN5S105', description='RNA, 5S ribosomal pseudogene 105', other_designations='RNA, 5S ribosomal 105', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=137512098, end_position_on_the_genomic_accession=137512210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287596, CurrentID=0, Status='live', Symbol='DDX11L5', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 5', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 5', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11987, end_position_on_the_genomic_accession=14525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418669, CurrentID=0, Status='live', Symbol='LOC100418669', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418670, CurrentID=0, Status='live', Symbol='LOC100418670', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418671, CurrentID=0, Status='live', Symbol='LOC100418671', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418672, CurrentID=0, Status='live', Symbol='LOC100418672', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418673, CurrentID=0, Status='live', Symbol='LOC100418673', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418674, CurrentID=0, Status='live', Symbol='LOC100418674', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287603, CurrentID=0, Status='live', Symbol='SSXP10', Aliases='psiSSX10', description='SSX family pseudogene 10', other_designations='SSX10 pseudogene', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118589052, end_position_on_the_genomic_accession=118590241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418676, CurrentID=0, Status='live', Symbol='LOC100418676', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418677, CurrentID=0, Status='live', Symbol='LOC100418677', Aliases='', description='olfactory receptor, family 2, subfamily G, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287606, CurrentID=0, Status='live', Symbol='TTTY28P', Aliases='', description='testis-specific transcript, Y-linked 28, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=28289709, end_position_on_the_genomic_accession=28290108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418679, CurrentID=0, Status='live', Symbol='LOC100418679', Aliases='', description='olfactory receptor, family 2, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418680, CurrentID=0, Status='live', Symbol='LOC100418680', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418681, CurrentID=0, Status='live', Symbol='LOC100418681', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418682, CurrentID=0, Status='live', Symbol='LOC100418682', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418683, CurrentID=0, Status='live', Symbol='LOC100418683', Aliases='', description='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418684, CurrentID=0, Status='live', Symbol='LOC100418684', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418685, CurrentID=0, Status='live', Symbol='LOC100418685', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418686, CurrentID=0, Status='live', Symbol='LOC100418686', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418687, CurrentID=0, Status='live', Symbol='LOC100418687', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287616, CurrentID=0, Status='live', Symbol='LOXL1-AS1', Aliases='', description='LOXL1 antisense RNA 1', other_designations='LOXL1 antisense RNA 1 (non-protein coding)', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73917468, end_position_on_the_genomic_accession=73928248, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100418689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418689, CurrentID=0, Status='live', Symbol='LOC100418689', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418690, CurrentID=0, Status='live', Symbol='LOC100418690', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271073, CurrentID=0, Status='live', Symbol='RPS17P13', Aliases='RPS17_6_804', description='ribosomal protein S17 pseudogene 13', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37116808, end_position_on_the_genomic_accession=37117185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418692, CurrentID=0, Status='live', Symbol='LOC100418692', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418693, CurrentID=0, Status='live', Symbol='LOC100418693', Aliases='', description='protein kinase, interferon-inducible double stranded RNA dependent activator pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418694, CurrentID=0, Status='live', Symbol='LOC100418694', Aliases='', description='protein kinase, interferon-inducible double stranded RNA dependent activator pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418696, CurrentID=0, Status='live', Symbol='LOC100418696', Aliases='', description='axin 2 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112346045, end_position_on_the_genomic_accession=112346905, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287625, CurrentID=0, Status='live', Symbol='SUPT4H1P2', Aliases='', description='SPT4 homolog, DSIF elongation factor subunit pseudogene 2', other_designations='suppressor of Ty 4 homolog 1 (S. cerevisiae) pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8512372, end_position_on_the_genomic_accession=8512724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418699, CurrentID=0, Status='live', Symbol='LOC100418699', Aliases='', description='nebulin pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418700, CurrentID=0, Status='live', Symbol='LOC100418700', Aliases='', description='plectin pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29406820, end_position_on_the_genomic_accession=29408001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418701, CurrentID=0, Status='live', Symbol='LOC100418701', Aliases='', description='talin 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=96212279, end_position_on_the_genomic_accession=96213106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287630, CurrentID=0, Status='live', Symbol='PSAT1P4', Aliases='', description='phosphoserine aminotransferase 1 pseudogene 4', other_designations='phosphoserine aminotransferase 1 pseuodogene', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166691188, end_position_on_the_genomic_accession=166691714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418703, CurrentID=0, Status='live', Symbol='LOC100418703', Aliases='', description='repetin pseudogene', other_designations='', map_location='X;Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1127409, end_position_on_the_genomic_accession=1128785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287632, CurrentID=0, Status='live', Symbol='LOC100287632', Aliases='', description='selenoprotein K pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118782708, end_position_on_the_genomic_accession=118783420, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418705, CurrentID=0, Status='live', Symbol='LOC100418705', Aliases='', description='KIAA1586 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113837528, end_position_on_the_genomic_accession=113839144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418706, CurrentID=0, Status='live', Symbol='FXNP2', Aliases='', description='frataxin pseuodgene 2', other_designations='frataxin pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124968297, end_position_on_the_genomic_accession=124968857, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287635, CurrentID=0, Status='live', Symbol='KRT18P62', Aliases='', description='keratin 18 pseudogene 62', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19257518, end_position_on_the_genomic_accession=19258861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418708, CurrentID=0, Status='live', Symbol='CLUHP1', Aliases='KIAA0664P1', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17623863, end_position_on_the_genomic_accession=17624424, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418709, CurrentID=0, Status='live', Symbol='CLUHP2', Aliases='KIAA0664P2', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 2', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18281584, end_position_on_the_genomic_accession=18282145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418710, CurrentID=0, Status='live', Symbol='LOC100418710', Aliases='', description='KIAA1191 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121369433, end_position_on_the_genomic_accession=121370288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287639, CurrentID=0, Status='live', Symbol='MTHFD2P1', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 1', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=95654423, end_position_on_the_genomic_accession=95683193, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100418712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418712, CurrentID=0, Status='live', Symbol='LOC100418712', Aliases='', description='KIAA1328 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153129065, end_position_on_the_genomic_accession=153129640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418713, CurrentID=0, Status='live', Symbol='LOC100418713', Aliases='', description='stomatin pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79406149, end_position_on_the_genomic_accession=79406850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373198, CurrentID=0, Status='live', Symbol='LOC105373198', Aliases='', description='uncharacterized LOC105373198', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232374409, end_position_on_the_genomic_accession=232376900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100418715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418715, CurrentID=0, Status='live', Symbol='SEPT7P4', Aliases='', description='septin 7 pseudogene 4', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63241590, end_position_on_the_genomic_accession=63248321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418716, CurrentID=0, Status='live', Symbol='LOC100418716', Aliases='', description='KIAA1468 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116873451, end_position_on_the_genomic_accession=116875542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418718, CurrentID=0, Status='live', Symbol='SEPT7P5', Aliases='', description='septin 7 pseudogene 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63495799, end_position_on_the_genomic_accession=63502536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418719, CurrentID=0, Status='live', Symbol='CLUHP7', Aliases='', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65262394, end_position_on_the_genomic_accession=65262982, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418720, CurrentID=0, Status='live', Symbol='SEPT7P6', Aliases='', description='septin 7 pseudogene 6', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152367507, end_position_on_the_genomic_accession=152368487, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148641, CurrentID=0, Status='live', Symbol='SLC35F3', Aliases='', description='solute carrier family 35 member F3', other_designations='solute carrier family 35, member F3', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233904712, end_position_on_the_genomic_accession=234324518, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100418722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418722, CurrentID=0, Status='live', Symbol='LOC100418722', Aliases='', description='cilia and flagella associated protein 97 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35873270, end_position_on_the_genomic_accession=35874041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418723, CurrentID=0, Status='live', Symbol='LOC100418723', Aliases='', description='KIAA1429 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41542068, end_position_on_the_genomic_accession=41544148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418724, CurrentID=0, Status='live', Symbol='CCNJP2', Aliases='', description='cyclin J pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92755814, end_position_on_the_genomic_accession=92756955, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148645, CurrentID=0, Status='live', Symbol='LINC00337', Aliases='C1orf211, NCRNA00337', description='long intergenic non-protein coding RNA 337', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6236240, end_position_on_the_genomic_accession=6239444, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418728, CurrentID=0, Status='live', Symbol='KRT8P40', Aliases='', description='keratin 8 pseudogene 40', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177197981, end_position_on_the_genomic_accession=177198550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418730, CurrentID=0, Status='live', Symbol='LOC100418730', Aliases='', description='T-box 20 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63684746, end_position_on_the_genomic_accession=63724931, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418731, CurrentID=0, Status='live', Symbol='AKIRIN1P1', Aliases='', description='akirin 1 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80561013, end_position_on_the_genomic_accession=80561544, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418732, CurrentID=0, Status='live', Symbol='LOC100418732', Aliases='', description='KIAA1586 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82576542, end_position_on_the_genomic_accession=82577322, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418733, CurrentID=0, Status='live', Symbol='CLUHP8', Aliases='', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 8', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38082046, end_position_on_the_genomic_accession=38084082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418734, CurrentID=0, Status='live', Symbol='LOC100418734', Aliases='', description='KIAA1586 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50287732, end_position_on_the_genomic_accession=50288274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418735, CurrentID=0, Status='live', Symbol='SEPT7P8', Aliases='', description='septin 7 pseudogene 8', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53772695, end_position_on_the_genomic_accession=53773286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418736, CurrentID=0, Status='live', Symbol='GOLGA5P1', Aliases='', description='golgin A5 pseudogene 1', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39168934, end_position_on_the_genomic_accession=39170362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418737, CurrentID=0, Status='live', Symbol='KRT8P33', Aliases='', description='keratin 8 pseudogene 33', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123400672, end_position_on_the_genomic_accession=123402418, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418739, CurrentID=0, Status='live', Symbol='DNM1P48', Aliases='', description='dynamin 1 pseudogene 48', other_designations='DNM1 pseudogene 48', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25486528, end_position_on_the_genomic_accession=25487227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418740, CurrentID=0, Status='live', Symbol='PPP1R26P1', Aliases='KIAA0649P1', description='protein phosphatase 1 regulatory subunit 26 pseudogene 1', other_designations='protein phosphatase 1, regulatory subunit 26 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48316689, end_position_on_the_genomic_accession=48320555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418742, CurrentID=0, Status='live', Symbol='CLUHP9', Aliases='', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 9', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24742670, end_position_on_the_genomic_accession=24743226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418743, CurrentID=0, Status='live', Symbol='CCNL2P1', Aliases='', description='cyclin L2 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42561880, end_position_on_the_genomic_accession=42562422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418744, CurrentID=0, Status='live', Symbol='KRT8P46', Aliases='', description='keratin 8 pseudogene 46', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102728522, end_position_on_the_genomic_accession=102730247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418745, CurrentID=0, Status='live', Symbol='CLUHP5', Aliases='KIAA0664P5', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 5', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133660279, end_position_on_the_genomic_accession=133662199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418747, CurrentID=0, Status='live', Symbol='KRT8P35', Aliases='', description='keratin 8 pseudogene 35', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63096246, end_position_on_the_genomic_accession=63097755, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418748, CurrentID=0, Status='live', Symbol='KRT8P36', Aliases='', description='keratin 8 pseudogene 36', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138101236, end_position_on_the_genomic_accession=138102993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372918, CurrentID=0, Status='live', Symbol='LOC105372918', Aliases='', description='uncharacterized LOC105372918', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=216194059, end_position_on_the_genomic_accession=216238039, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100418750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418750, CurrentID=0, Status='live', Symbol='KRT8P42', Aliases='', description='keratin 8 pseudogene 42', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134296744, end_position_on_the_genomic_accession=134298781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271075, CurrentID=0, Status='live', Symbol='RPS18P4', Aliases='RPS18_1_117', description='ribosomal protein S18 pseudogene 4', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167591349, end_position_on_the_genomic_accession=167591876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287680, CurrentID=0, Status='live', Symbol='SUB1P2', Aliases='', description='SUB1 homolog, transcriptional regulator pseudogene 2', other_designations='SUB1 homolog pseudogene 2|SUB1 pseudogene 2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74305423, end_position_on_the_genomic_accession=74306063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418753, CurrentID=0, Status='live', Symbol='LOC100418753', Aliases='', description='septin 7 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60256689, end_position_on_the_genomic_accession=60258408, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287682, CurrentID=0, Status='live', Symbol='EXTL2P1', Aliases='EXTL2P', description='exostoses (multiple)-like 2 pseudogene 1', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175842898, end_position_on_the_genomic_accession=175843302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418756, CurrentID=0, Status='live', Symbol='CLUHP11', Aliases='', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 11', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35083397, end_position_on_the_genomic_accession=35084316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287685, CurrentID=0, Status='live', Symbol='LOC100287685', Aliases='', description='uncharacterized LOC100287685', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=1604131, end_position_on_the_genomic_accession=1604782, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418759, CurrentID=0, Status='live', Symbol='LOC100418759', Aliases='', description='myoneurin pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30822681, end_position_on_the_genomic_accession=30823432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418763, CurrentID=0, Status='live', Symbol='CCNT2P1', Aliases='', description='cyclin T2 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111007693, end_position_on_the_genomic_accession=111015519, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287692, CurrentID=0, Status='live', Symbol='TCEB1P24', Aliases='TCEB1P25', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 24', other_designations='transcription elongation factor B (SIII), polypeptide 1 pseudogene 25', map_location='Xq28 and Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57165512, end_position_on_the_genomic_accession=57165845, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418768, CurrentID=0, Status='live', Symbol='LOC100418768', Aliases='', description='ataxin 3 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48045481, end_position_on_the_genomic_accession=48046219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134008, CurrentID=0, Status='live', Symbol='MRPL11P2', Aliases='', description='mitochondrial ribosomal protein L11 pseudogene 2', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142036707, end_position_on_the_genomic_accession=142037125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418770, CurrentID=0, Status='live', Symbol='KRT18P26', Aliases='', description='keratin 18 pseudogene 26', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74306670, end_position_on_the_genomic_accession=74308058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418774, CurrentID=0, Status='live', Symbol='PHBP4', Aliases='', description='prohibitin pseudogene 4', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154286589, end_position_on_the_genomic_accession=154287422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148696, CurrentID=0, Status='live', Symbol='LOC148696', Aliases='', description='uncharacterized LOC148696', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207818379, end_position_on_the_genomic_accession=207822596, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480528, CurrentID=0, Status='live', Symbol='RN7SL720P', Aliases='', description='RNA, 7SL, cytoplasmic 720, pseudogene', other_designations='', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372820, CurrentID=0, Status='live', Symbol='LOC105372820', Aliases='', description='uncharacterized LOC105372820', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418779, CurrentID=0, Status='live', Symbol='KRT127P', Aliases='KRT3P1', description='keratin 127 pseudogene', other_designations='keratin 3 pseudogene 1', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52753311, end_position_on_the_genomic_accession=52754048, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287708, CurrentID=0, Status='live', Symbol='CCDC58P2', Aliases='', description='coiled-coil domain containing 58 pseudogene 2', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50962744, end_position_on_the_genomic_accession=50963430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271076, CurrentID=0, Status='live', Symbol='RPS18P6', Aliases='RPS18_3_386', description='ribosomal protein S18 pseudogene 6', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=95171909, end_position_on_the_genomic_accession=95172404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287715, CurrentID=0, Status='live', Symbol='KRTAP9-12P', Aliases='', description='keratin associated protein 9-12, pseudogene', other_designations='keratin associated protein 16-1 pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41212715, end_position_on_the_genomic_accession=41213416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148709, CurrentID=0, Status='live', Symbol='LOC148709', Aliases='', description='actin pseudogene', other_designations='actin gamma 1 pseudogene|actin, gamma 1 pseudogene', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202861754, end_position_on_the_genomic_accession=202875241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287718, CurrentID=0, Status='live', Symbol='ANKRD66', Aliases='', description='ankyrin repeat domain 66', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46746917, end_position_on_the_genomic_accession=46759218, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90321, CurrentID=0, Status='live', Symbol='ZNF766', Aliases='', description='zinc finger protein 766', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52269571, end_position_on_the_genomic_accession=52292726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418792, CurrentID=0, Status='live', Symbol='KRT18P54', Aliases='', description='keratin 18 pseudogene 54', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135371180, end_position_on_the_genomic_accession=135371922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148713, CurrentID=0, Status='live', Symbol='PTPRVP', Aliases='ESP, OST-PTP, PTPRV', description='protein tyrosine phosphatase, receptor type, V, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202168051, end_position_on_the_genomic_accession=202189449, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100287722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287722, CurrentID=0, Status='live', Symbol='AP4B1-AS1', Aliases='', description='AP4B1 antisense RNA 1', other_designations='AP4B1 antisense RNA 1 (non-protein coding)', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113812612, end_position_on_the_genomic_accession=113901237, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100287725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287725, CurrentID=0, Status='live', Symbol='GAPDHP76', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 76', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11585274, end_position_on_the_genomic_accession=11587311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418799, CurrentID=0, Status='live', Symbol='KRT8P38', Aliases='', description='keratin 8 pseudogene 38', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88727571, end_position_on_the_genomic_accession=88729112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287728, CurrentID=0, Status='live', Symbol='LOC100287728', Aliases='', description='uncharacterized LOC100287728', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135120621, end_position_on_the_genomic_accession=135123601, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418804, CurrentID=0, Status='live', Symbol='PHBP8', Aliases='', description='prohibitin pseudogene 8', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119791829, end_position_on_the_genomic_accession=119792550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287735, CurrentID=0, Status='live', Symbol='TTTY13B', Aliases='pFIV', description='testis-specific transcript, Y-linked 13B', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9107322, end_position_on_the_genomic_accession=9109922, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271077, CurrentID=0, Status='live', Symbol='RPS18P11', Aliases='RPS18_3_913', description='ribosomal protein S18 pseudogene 11', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70424088, end_position_on_the_genomic_accession=70424619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418814, CurrentID=0, Status='live', Symbol='LOC100418814', Aliases='', description='metadherin pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64942163, end_position_on_the_genomic_accession=64942911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418815, CurrentID=0, Status='live', Symbol='PHBP10', Aliases='', description='prohibitin pseudogene 10', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104647219, end_position_on_the_genomic_accession=104648000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287745, CurrentID=0, Status='live', Symbol='MTATP6P2', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 2', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100051917, end_position_on_the_genomic_accession=100052597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148738, CurrentID=0, Status='live', Symbol='HFE2', Aliases='HFE2A, HJV, JH, RGMC', description='hemochromatosis type 2 (juvenile)', other_designations='RGM domain family member C|haemojuvelin|hemochromatosis type 2 protein|repulsive guidance molecule c', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146017468, end_position_on_the_genomic_accession=146021822, orientation='minus', exon_count=4, OMIM=608374, no_hdr0=''),
  100418819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418819, CurrentID=0, Status='live', Symbol='IPO8P1', Aliases='', description='importin 8 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210859177, end_position_on_the_genomic_accession=210862283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418820, CurrentID=0, Status='live', Symbol='PHBP12', Aliases='', description='prohibitin pseudogene 12', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50780340, end_position_on_the_genomic_accession=50781149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148741, CurrentID=0, Status='live', Symbol='ANKRD35', Aliases='', description='ankyrin repeat domain 35', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145866560, end_position_on_the_genomic_accession=145885866, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100418822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418822, CurrentID=0, Status='live', Symbol='LOC100418822', Aliases='', description='atrophin 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235360365, end_position_on_the_genomic_accession=235362564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287754, CurrentID=0, Status='live', Symbol='LOC100287754', Aliases='', description='zinc finger protein 44 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11756511, end_position_on_the_genomic_accession=11757644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287755, CurrentID=0, Status='live', Symbol='ATP5G1P4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 4', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172828633, end_position_on_the_genomic_accession=172829215, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418828, CurrentID=0, Status='live', Symbol='KRT125P', Aliases='', description='keratin 125 pseudogene', other_designations='keratin 79 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52735287, end_position_on_the_genomic_accession=52735971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418829, CurrentID=0, Status='live', Symbol='KRT18P60', Aliases='', description='keratin 18 pseudogene 60', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65418667, end_position_on_the_genomic_accession=65420025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418832, CurrentID=0, Status='live', Symbol='LOC100418832', Aliases='', description='metadherin pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21677550, end_position_on_the_genomic_accession=21678221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148753, CurrentID=0, Status='live', Symbol='FAM163A', Aliases='C1orf76, NDSP', description='family with sequence similarity 163 member A', other_designations='cebelin|family with sequence similarity 163, member A|neuroblastoma derived secretory protein|neuroblastoma-derived secretory protein', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179727763, end_position_on_the_genomic_accession=179816198, orientation='plus', exon_count=14, OMIM=611727, no_hdr0=''),
  105373202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373202, CurrentID=0, Status='live', Symbol='LOC105373202', Aliases='', description='uncharacterized LOC105373202', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50979159, end_position_on_the_genomic_accession=50984221, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418835, CurrentID=0, Status='live', Symbol='LOC100418835', Aliases='', description='metadherin pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21940820, end_position_on_the_genomic_accession=21941533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148756, CurrentID=0, Status='live', Symbol='OVAAL', Aliases='LINC01131, OVAL', description='ovarian adenocarcinoma amplified long non-coding RNA', other_designations='long intergenic non-protein coding RNA 1131|ovarian adenocarcinoma amplified lncRNA', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180558974, end_position_on_the_genomic_accession=180566518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418837, CurrentID=0, Status='live', Symbol='PHBP2', Aliases='', description='prohibitin pseudogene 2', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43733759, end_position_on_the_genomic_accession=43734564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287768, CurrentID=0, Status='live', Symbol='COX6A1P3', Aliases='', description='cytochrome c oxidase subunit VIa polypeptide 1 pseudogene 3', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=120781188, end_position_on_the_genomic_accession=120781636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287769, CurrentID=0, Status='live', Symbol='RBMY1GP', Aliases='RBMY1G', description='RNA binding motif protein, Y-linked, family 1, member G, pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 1, member G', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9325490, end_position_on_the_genomic_accession=9330587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418843, CurrentID=0, Status='live', Symbol='PHBP13', Aliases='', description='prohibitin pseudogene 13', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35251063, end_position_on_the_genomic_accession=35251833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271253, CurrentID=0, Status='live', Symbol='RPS3AP24', Aliases='RPS3A_10_762', description='ribosomal protein S3a pseudogene 24', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140761540, end_position_on_the_genomic_accession=140762262, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418848, CurrentID=0, Status='live', Symbol='LOC100418848', Aliases='', description='calponin 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57979519, end_position_on_the_genomic_accession=57980683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287778, CurrentID=0, Status='live', Symbol='TCEB1P20', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 20', other_designations='', map_location='1p13.2-p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115556676, end_position_on_the_genomic_accession=115557052, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287785, CurrentID=0, Status='live', Symbol='LINC00266-2P', Aliases='NCRNA00266-2P, NCRNA00266B, NCRNA00266Y, NCRNA00266Y1', description='long intergenic non-protein coding RNA 266-2, pseudogene', other_designations='non-protein coding RNA 266-2, pseudogene|non-protein coding RNA 266B', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24278538, end_position_on_the_genomic_accession=24291476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418861, CurrentID=0, Status='live', Symbol='KRT43P', Aliases='', description='keratin 43 pseudogene', other_designations='keratin 37 pseudogene|keratin 43, pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41448462, end_position_on_the_genomic_accession=41449404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418862, CurrentID=0, Status='live', Symbol='LOC100418862', Aliases='', description='sperm antigen with calponin homology and coiled-coil domains 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15829356, end_position_on_the_genomic_accession=15834267, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287792, CurrentID=0, Status='live', Symbol='LOC100287792', Aliases='', description='uncharacterized LOC100287792', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37676910, end_position_on_the_genomic_accession=37683234, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100418865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418865, CurrentID=0, Status='live', Symbol='KRT18P61', Aliases='', description='keratin 18 pseudogene 61', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60810115, end_position_on_the_genomic_accession=60811518, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287794, CurrentID=0, Status='live', Symbol='LOC100287794', Aliases='', description='double homeobox protein 4-like protein 4', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418868, CurrentID=0, Status='live', Symbol='LOC100418868', Aliases='', description='calponin 2 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68650831, end_position_on_the_genomic_accession=68651986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148789, CurrentID=0, Status='live', Symbol='B3GALNT2', Aliases='B3GalNAc-T2, MDDGA11', description='beta-1,3-N-acetylgalactosaminyltransferase 2', other_designations='UDP-GalNAc:betaGlcNAc beta-1,3-galactosaminyltransferase, polypeptide 2|beta-1,3-GalNAc-T2', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235440660, end_position_on_the_genomic_accession=235504481, orientation='minus', exon_count=15, OMIM=610194, no_hdr0=''),
  100287798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287798, CurrentID=0, Status='live', Symbol='FAM197Y8', Aliases='', description='family with sequence similarity 197 Y-linked member 8', other_designations='family with sequence similarity 197, Y-linked, member 8', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9190859, end_position_on_the_genomic_accession=9193055, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271079, CurrentID=0, Status='live', Symbol='RPS20P8', Aliases='RPS20_2_149', description='ribosomal protein S20 pseudogene 8', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203000055, end_position_on_the_genomic_accession=203000339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418874, CurrentID=0, Status='live', Symbol='LOC100418874', Aliases='', description='cathepsin B pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227234191, end_position_on_the_genomic_accession=227235354, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418877, CurrentID=0, Status='live', Symbol='KRT18P7', Aliases='', description='keratin 18 pseudogene 7', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70594502, end_position_on_the_genomic_accession=70595892, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287808, CurrentID=0, Status='live', Symbol='LOC100287808', Aliases='', description='uncharacterized LOC100287808', other_designations='CTB-58E17.5', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38748915, end_position_on_the_genomic_accession=38751457, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418881, CurrentID=0, Status='live', Symbol='NBEAP2', Aliases='BCL8D', description='neurobeachin pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131722418, end_position_on_the_genomic_accession=131751585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418882, CurrentID=0, Status='live', Symbol='LOC100418882', Aliases='', description='reticulon 3 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102230030, end_position_on_the_genomic_accession=102230675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418883, CurrentID=0, Status='live', Symbol='LOC100418883', Aliases='', description='actin, beta pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9518150, end_position_on_the_genomic_accession=9519321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418884, CurrentID=0, Status='live', Symbol='LOC100418884', Aliases='', description='TSPY-like 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106783413, end_position_on_the_genomic_accession=106784478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418885, CurrentID=0, Status='live', Symbol='LOC100418885', Aliases='', description='amphiphysin pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5354692, end_position_on_the_genomic_accession=5355621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287814, CurrentID=0, Status='live', Symbol='LINC00582', Aliases='', description='long intergenic non-protein coding RNA 582', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231591292, end_position_on_the_genomic_accession=231612090, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418887, CurrentID=0, Status='live', Symbol='TAGLN2P1', Aliases='', description='transgelin 2 pseudogene 1', other_designations='transgelin 3 pseudogene', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106697427, end_position_on_the_genomic_accession=106698040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148808, CurrentID=0, Status='live', Symbol='MFSD4', Aliases='UNQ3064', description='major facilitator superfamily domain containing 4', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205568984, end_position_on_the_genomic_accession=205602918, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100287817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287817, CurrentID=0, Status='live', Symbol='NDUFS5P4', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 4', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 4', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98976659, end_position_on_the_genomic_accession=98977143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287818, CurrentID=0, Status='live', Symbol='TTTY26P', Aliases='', description='testis-specific transcript, Y-linked 26, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=7528211, end_position_on_the_genomic_accession=7528789, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148811, CurrentID=0, Status='live', Symbol='PM20D1', Aliases='Cps1', description='peptidase M20 domain containing 1', other_designations='peptidase M20 domain-containing protein 1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205828022, end_position_on_the_genomic_accession=205850148, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100418892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418892, CurrentID=0, Status='live', Symbol='GYG2P2', Aliases='', description='glycogenin 2 pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205091163, end_position_on_the_genomic_accession=205092030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287822, CurrentID=0, Status='live', Symbol='SETP10', Aliases='', description='SET pseudogene 10', other_designations='SET nuclear oncogene pseudogene', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179183952, end_position_on_the_genomic_accession=179184692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287823, CurrentID=0, Status='live', Symbol='LOC100287823', Aliases='', description='double homeobox protein 4-like protein 4', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418897, CurrentID=0, Status='live', Symbol='NBEAP4', Aliases='', description='neurobeachin pseudogene 4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21250168, end_position_on_the_genomic_accession=21260147, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287826, CurrentID=0, Status='live', Symbol='FAM197Y7P', Aliases='FAM197Y7', description='family with sequence similarity 197 Y-linked member 7, pseudogene', other_designations='family with sequence similarity 197, Y-linked, member 7|family with sequence similarity 197, Y-linked, member 7, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9207466, end_position_on_the_genomic_accession=9213071, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372919, CurrentID=0, Status='live', Symbol='LOC105372919', Aliases='', description='uncharacterized LOC105372919', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215566642, end_position_on_the_genomic_accession=215567357, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271080, CurrentID=0, Status='live', Symbol='RPS20P9', Aliases='RPS20_3_230', description='ribosomal protein S20 pseudogene 9', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62939873, end_position_on_the_genomic_accession=62940322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418902, CurrentID=0, Status='live', Symbol='TUBBP10', Aliases='', description='tubulin beta class I pseudogene 10', other_designations='tubulin, beta class I pseudogene 10', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52994560, end_position_on_the_genomic_accession=52996217, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148823, CurrentID=0, Status='live', Symbol='GCSAML', Aliases='C1orf150', description='germinal center-associated, signaling and motility-like', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247507058, end_position_on_the_genomic_accession=247577690, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  148824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148824, CurrentID=0, Status='live', Symbol='GCSAML-AS1', Aliases='', description='GCSAML antisense RNA 1', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247524679, end_position_on_the_genomic_accession=247530804, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100418905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418905, CurrentID=0, Status='live', Symbol='NBEAP3', Aliases='BCL8C', description='neurobeachin pseudogene 3', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15852990, end_position_on_the_genomic_accession=15876684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287834, CurrentID=0, Status='live', Symbol='LOC100287834', Aliases='', description='uncharacterized LOC100287834', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63396340, end_position_on_the_genomic_accession=63399041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418908, CurrentID=0, Status='live', Symbol='GLRX3P1', Aliases='', description='glutaredoxin 3 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19967198, end_position_on_the_genomic_accession=19968010, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287837, CurrentID=0, Status='live', Symbol='LOC100287837', Aliases='', description='uncharacterized LOC100287837', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=73018280, end_position_on_the_genomic_accession=73020406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418910, CurrentID=0, Status='live', Symbol='ACTG1P14', Aliases='', description='actin gamma 1 pseudogene 14', other_designations='actin, gamma 1 pseudogene 14', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6834408, end_position_on_the_genomic_accession=6835584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287840, CurrentID=0, Status='live', Symbol='LOC100287840', Aliases='', description='tetraspanin 12 pseudogene', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115577226, end_position_on_the_genomic_accession=115577785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287844, CurrentID=0, Status='live', Symbol='TTTY30P', Aliases='', description='testis-specific transcript, Y-linked 30, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8253772, end_position_on_the_genomic_accession=8254266, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287845, CurrentID=0, Status='live', Symbol='GOLGA6L16P', Aliases='', description='golgin A6 family-like 16, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25497476, end_position_on_the_genomic_accession=25500233, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287846, CurrentID=0, Status='live', Symbol='LOC100287846', Aliases='', description='patched 1 pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47189307, end_position_on_the_genomic_accession=47192816, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418919, CurrentID=0, Status='live', Symbol='TNPO1P1', Aliases='', description='transportin 1 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84390734, end_position_on_the_genomic_accession=84392046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418920, CurrentID=0, Status='live', Symbol='ACTG1P13', Aliases='', description='actin gamma 1 pseudogene 13', other_designations='actin, gamma 1 pseudogene 13', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99215226, end_position_on_the_genomic_accession=99216033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418921, CurrentID=0, Status='live', Symbol='ACTG1P12', Aliases='', description='actin gamma 1 pseudogene 12', other_designations='actin, gamma 1 pseudogene 12', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12070155, end_position_on_the_genomic_accession=12071560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418923, CurrentID=0, Status='live', Symbol='LOC100418923', Aliases='', description='TWIST neighbor pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87173356, end_position_on_the_genomic_accession=87174252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418924, CurrentID=0, Status='live', Symbol='ACTG1P18', Aliases='', description='actin gamma 1 pseudogene 18', other_designations='actin, gamma 1 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=101430411, end_position_on_the_genomic_accession=101431319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418928, CurrentID=0, Status='live', Symbol='LOC100418928', Aliases='', description='RAD52 motif containing 1 pseudogene', other_designations='RAD52 motif 1 pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68156497, end_position_on_the_genomic_accession=68159060, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271081, CurrentID=0, Status='live', Symbol='RPS20P10', Aliases='RPS20_3_240', description='ribosomal protein S20 pseudogene 10', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71984177, end_position_on_the_genomic_accession=71984434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285556, CurrentID=0, Status='live', Symbol='LOC285556', Aliases='', description='uncharacterized LOC285556', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99636529, end_position_on_the_genomic_accession=99822027, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100418937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418937, CurrentID=0, Status='live', Symbol='TUBB8P1', Aliases='', description='tubulin beta 8 class VIII pseudogene 1', other_designations='tubulin, beta 8 class VIII pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21811651, end_position_on_the_genomic_accession=21812347, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418938, CurrentID=0, Status='live', Symbol='LOC100418938', Aliases='', description='forkhead box H1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137787668, end_position_on_the_genomic_accession=137788385, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418939, CurrentID=0, Status='live', Symbol='LOC100418939', Aliases='', description='tetraspanin 11 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3323324, end_position_on_the_genomic_accession=3337202, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287868, CurrentID=0, Status='live', Symbol='NUTF2P7', Aliases='', description='nuclear transport factor 2 pseudogene 7', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71016331, end_position_on_the_genomic_accession=71016970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287869, CurrentID=0, Status='live', Symbol='LOC100287869', Aliases='', description='uncharacterized LOC100287869', other_designations='', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=52746874, end_position_on_the_genomic_accession=52748971, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418942, CurrentID=0, Status='live', Symbol='PRDX4P2', Aliases='', description='peroxiredoxin 4 pseudogene 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38197617, end_position_on_the_genomic_accession=38198145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418943, CurrentID=0, Status='live', Symbol='PRDX2P3', Aliases='', description='peroxiredoxin 2 pseudogene 3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179572087, end_position_on_the_genomic_accession=179572873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287874, CurrentID=0, Status='live', Symbol='TTTY27P', Aliases='', description='testis-specific transcript, Y-linked 27, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=8382203, end_position_on_the_genomic_accession=8382598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148867, CurrentID=0, Status='live', Symbol='SLC30A7', Aliases='ZNT7, ZnT-7, ZnTL2', description='solute carrier family 30 (zinc transporter), member 7', other_designations='solute carrier family 30 member 7|zinc transporter ZnT-7|zinc transporter like 2|znt-like transporter 2', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100896073, end_position_on_the_genomic_accession=100996913, orientation='plus', exon_count=15, OMIM=611149, no_hdr0=''),
  100418949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418949, CurrentID=0, Status='live', Symbol='PRDX4P1', Aliases='', description='peroxiredoxin 4 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44944439, end_position_on_the_genomic_accession=44944999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148870, CurrentID=0, Status='live', Symbol='CCDC27', Aliases='', description='coiled-coil domain containing 27', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3752398, end_position_on_the_genomic_accession=3771645, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100418951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418951, CurrentID=0, Status='live', Symbol='ACTG1P15', Aliases='', description='actin gamma 1 pseudogene 15', other_designations='actin, gamma 1 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34373534, end_position_on_the_genomic_accession=34374577, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287880, CurrentID=0, Status='live', Symbol='NDUFA4P2', Aliases='NDUFA4L, bcm1723', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa, pseudogene 2', other_designations='', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=103240989, end_position_on_the_genomic_accession=103241234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418955, CurrentID=0, Status='live', Symbol='PRDX3P4', Aliases='', description='peroxiredoxin 3 pseudogene 4', other_designations='', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64174378, end_position_on_the_genomic_accession=64175347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418956, CurrentID=0, Status='live', Symbol='LOC100418956', Aliases='', description='proline rich 11 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=48952069, end_position_on_the_genomic_accession=48952753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418957, CurrentID=0, Status='live', Symbol='LOC100418957', Aliases='', description='forkhead box H1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72316599, end_position_on_the_genomic_accession=72317396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287887, CurrentID=0, Status='live', Symbol='RPL13AP25', Aliases='FWP004, RPL13A_11_1314, S863-4', description='ribosomal protein L13a pseudogene 25', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54440678, end_position_on_the_genomic_accession=54441337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418960, CurrentID=0, Status='live', Symbol='ACTG1P16', Aliases='', description='actin gamma 1 pseudogene 16', other_designations='actin, gamma 1 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50057956, end_position_on_the_genomic_accession=50059301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271082, CurrentID=0, Status='live', Symbol='RPS20P11', Aliases='RPS20_4_266', description='ribosomal protein S20 pseudogene 11', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112621692, end_position_on_the_genomic_accession=112622281, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418964, CurrentID=0, Status='live', Symbol='TUBB6P1', Aliases='', description='tubulin beta 6 class V pseudogene 1', other_designations='tubulin, beta 6 class V pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38383847, end_position_on_the_genomic_accession=38384430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418965, CurrentID=0, Status='live', Symbol='LOC100418965', Aliases='', description='peroxiredoxin 5 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76353580, end_position_on_the_genomic_accession=76354239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287895, CurrentID=0, Status='live', Symbol='NARSP2', Aliases='', description='asparaginyl-tRNA synthetase pseudogene 2', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62110526, end_position_on_the_genomic_accession=62111150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418968, CurrentID=0, Status='live', Symbol='SNX2P1', Aliases='', description='sorting nexin 2 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220207648, end_position_on_the_genomic_accession=220208282, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287898, CurrentID=0, Status='live', Symbol='TTC34', Aliases='', description='tetratricopeptide repeat domain 34', other_designations='TPR repeat protein 34|TPR repeat-containing protein ENSP00000383873', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2636988, end_position_on_the_genomic_accession=2789665, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100418972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418972, CurrentID=0, Status='live', Symbol='SNX18P14', Aliases='', description='sorting nexin 18 pseudogene 14', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94786472, end_position_on_the_genomic_accession=94787944, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418973, CurrentID=0, Status='live', Symbol='LOC100418973', Aliases='', description='citrate synthase pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63717124, end_position_on_the_genomic_accession=63717887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287902, CurrentID=0, Status='live', Symbol='LGALS8-AS1', Aliases='', description='LGALS8 antisense RNA 1', other_designations='LGALS8 antisense RNA 1 (non-protein coding)', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236523069, end_position_on_the_genomic_accession=236524508, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100418975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418975, CurrentID=0, Status='live', Symbol='SNX18P3', Aliases='', description='sorting nexin 18 pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38567401, end_position_on_the_genomic_accession=38569182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418976, CurrentID=0, Status='live', Symbol='SNX18P4', Aliases='', description='sorting nexin 18 pseudogene 4', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66082385, end_position_on_the_genomic_accession=66083772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90352, CurrentID=0, Status='live', Symbol='MAGOH3P', Aliases='MAGOHP, MAGOHP1', description='mago homolog 3, pseudogene', other_designations='Putative protein mago nashi homolog 3|mago-nashi homolog 2, proliferation-associated, pseudogene|mago-nashi homolog 3, pseudogene|mago-nashi homolog, proliferation-associated pseudogene', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68861723, end_position_on_the_genomic_accession=68862793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  148898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148898, CurrentID=0, Status='live', Symbol='ZNF436-AS1', Aliases='C1orf213', description='ZNF436 antisense RNA 1', other_designations='putative uncharacterized protein C1orf213', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23368971, end_position_on_the_genomic_accession=23371839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418980, CurrentID=0, Status='live', Symbol='SNX18P7', Aliases='', description='sorting nexin 18 pseudogene 7', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33575502, end_position_on_the_genomic_accession=33577014, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418981, CurrentID=0, Status='live', Symbol='SNX18P8', Aliases='', description='sorting nexin 18 pseudogene 8', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40292534, end_position_on_the_genomic_accession=40293921, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418983, CurrentID=0, Status='live', Symbol='NF1P12', Aliases='', description='neurofibromin 1 pseudogene 12', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38206231, end_position_on_the_genomic_accession=38207192, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373207, CurrentID=0, Status='live', Symbol='LOC105373207', Aliases='', description='uncharacterized LOC105373207', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234646900, end_position_on_the_genomic_accession=234649404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418986, CurrentID=0, Status='live', Symbol='LOC100418986', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20871355, end_position_on_the_genomic_accession=20872367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418987, CurrentID=0, Status='live', Symbol='LOC100418987', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20878669, end_position_on_the_genomic_accession=20879633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418988, CurrentID=0, Status='live', Symbol='LOC100418988', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20884202, end_position_on_the_genomic_accession=20885036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418989, CurrentID=0, Status='live', Symbol='LOC100418989', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20887192, end_position_on_the_genomic_accession=20888065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418990, CurrentID=0, Status='live', Symbol='LOC100418990', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20899153, end_position_on_the_genomic_accession=20900069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271083, CurrentID=0, Status='live', Symbol='RPS20P12', Aliases='RPS20_4_337', description='ribosomal protein S20 pseudogene 12', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234329402, end_position_on_the_genomic_accession=234329873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418992, CurrentID=0, Status='live', Symbol='LOC100418992', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20910931, end_position_on_the_genomic_accession=20911498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418993, CurrentID=0, Status='live', Symbol='LOC100418993', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20915472, end_position_on_the_genomic_accession=20916237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287922, CurrentID=0, Status='live', Symbol='ANKRD18CP', Aliases='', description='ankyrin repeat domain 18C, pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97188583, end_position_on_the_genomic_accession=97221235, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100418997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418997, CurrentID=0, Status='live', Symbol='LOC100418997', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101955792, end_position_on_the_genomic_accession=101956839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418999, CurrentID=0, Status='live', Symbol='SNX18P1Y', Aliases='', description='sorting nexin 18 pseudogene 1, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11178288, end_position_on_the_genomic_accession=11179040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419477, CurrentID=0, Status='live', Symbol='RPL12P45', Aliases='', description='ribosomal protein L12 pseudogene 45', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35053471, end_position_on_the_genomic_accession=35059268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419003, CurrentID=0, Status='live', Symbol='LOC100419003', Aliases='', description='sorting nexin 18 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49202947, end_position_on_the_genomic_accession=49203700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287932, CurrentID=0, Status='live', Symbol='TIMM23', Aliases='TIM23', description='translocase of inner mitochondrial membrane 23 homolog (yeast)', other_designations='translocase of the inner mitochondrial membrane', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45972453, end_position_on_the_genomic_accession=46003742, orientation='plus', exon_count=10, OMIM=605034, no_hdr0=''),
  100419005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419005, CurrentID=0, Status='live', Symbol='SNX18P25', Aliases='', description='sorting nexin 18 pseudogene 25', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49588776, end_position_on_the_genomic_accession=49589529, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419006, CurrentID=0, Status='live', Symbol='NF1P1', Aliases='NF1HHS', description='neurofibromin 1 pseudogene 1', other_designations='Albright osteodystrophy', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20916683, end_position_on_the_genomic_accession=20935719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419007, CurrentID=0, Status='live', Symbol='TUBAP3', Aliases='TUBA8P1', description='tubulin alpha pseudogene 3', other_designations='tubulin, alpha 8 pseudogene 1|tubulin, alpha pseudogene 3', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43687282, end_position_on_the_genomic_accession=43688150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419008, CurrentID=0, Status='live', Symbol='LOC100419008', Aliases='', description='proline rich 20B pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127284081, end_position_on_the_genomic_accession=127284641, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148930, CurrentID=0, Status='live', Symbol='KNCN', Aliases='Kino, L5', description='kinocilin', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46545638, end_position_on_the_genomic_accession=46551657, orientation='minus', exon_count=4, OMIM=611455, no_hdr0=''),
  148932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148932, CurrentID=0, Status='live', Symbol='MOB3C', Aliases='MOB1E, MOBKL2C', description='MOB kinase activator 3C', other_designations='MOB1, Mps One Binder kinase activator-like 2C|mob1 homolog 2C', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46607715, end_position_on_the_genomic_accession=46616891, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100287941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287941, CurrentID=0, Status='live', Symbol='TTTY24P', Aliases='', description='testis-specific transcript, Y-linked 24, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8970055, end_position_on_the_genomic_accession=8970444, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419014, CurrentID=0, Status='live', Symbol='LOC100419014', Aliases='', description='tubulin, beta 6 class V pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53439047, end_position_on_the_genomic_accession=53439632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287944, CurrentID=0, Status='live', Symbol='LOC100287944', Aliases='', description='uncharacterized LOC100287944', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106496410, end_position_on_the_genomic_accession=106774831, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419017, CurrentID=0, Status='live', Symbol='LOC100419017', Aliases='', description='cyclin Y like 1 pseudogene', other_designations='cyclin Y-like 1 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23500839, end_position_on_the_genomic_accession=23501665, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288542, CurrentID=0, Status='live', Symbol='LOC100288542', Aliases='', description='chromosome 14 open reading frame 119 pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113758997, end_position_on_the_genomic_accession=113759419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419019, CurrentID=0, Status='live', Symbol='SNX18P15', Aliases='', description='sorting nexin 18 pseudogene 15', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287948, CurrentID=0, Status='live', Symbol='GM140', Aliases='', description='uncharacterized LOC100287948', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181236388, end_position_on_the_genomic_accession=181238604, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271084, CurrentID=0, Status='live', Symbol='RPS20P13', Aliases='RPS20_5_284', description='ribosomal protein S20 pseudogene 13', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148869942, end_position_on_the_genomic_accession=148870467, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419022, CurrentID=0, Status='live', Symbol='SNX18P19', Aliases='', description='sorting nexin 18 pseudogene 19', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419023, CurrentID=0, Status='live', Symbol='PDCL3P6', Aliases='', description='phosducin-like 3 pseudogene 6', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52179815, end_position_on_the_genomic_accession=52180694, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419024, CurrentID=0, Status='live', Symbol='SNX18P17', Aliases='', description='sorting nexin 18 pseudogene 17', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419025, CurrentID=0, Status='live', Symbol='SNX18P21', Aliases='', description='sorting nexin 18 pseudogene 21', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419026, CurrentID=0, Status='live', Symbol='SNX18P20', Aliases='', description='sorting nexin 18 pseudogene 20', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419027, CurrentID=0, Status='live', Symbol='SNX18P22', Aliases='', description='sorting nexin 18 pseudogene 22', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419028, CurrentID=0, Status='live', Symbol='SNX25P1', Aliases='', description='sorting nexin 25 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211417025, end_position_on_the_genomic_accession=211417937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419029, CurrentID=0, Status='live', Symbol='NF1P4', Aliases='NF1L4, NF1L6', description='neurofibromin 1 pseudogene 4', other_designations='Putative neurofibromin 1-like protein 4/6|neurofibromin 1-like 4', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19492279, end_position_on_the_genomic_accession=19503252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419031, CurrentID=0, Status='live', Symbol='PRDX3P1', Aliases='', description='peroxiredoxin 3 pseudogene 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38722544, end_position_on_the_genomic_accession=38723522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287960, CurrentID=0, Status='live', Symbol='HIGD1AP10', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 10', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 10', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65144965, end_position_on_the_genomic_accession=65145995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419033, CurrentID=0, Status='live', Symbol='LOC100419033', Aliases='', description='apolipoprotein F pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56376748, end_position_on_the_genomic_accession=56377450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287964, CurrentID=0, Status='live', Symbol='ATP5LP7', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 7', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 7|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 7', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42647838, end_position_on_the_genomic_accession=42648147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419037, CurrentID=0, Status='live', Symbol='LOC100419037', Aliases='', description='spermine synthase pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43206601, end_position_on_the_genomic_accession=43208008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419038, CurrentID=0, Status='live', Symbol='SNX18P26', Aliases='', description='sorting nexin 18 pseudogene 26', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18807276, end_position_on_the_genomic_accession=18808855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419040, CurrentID=0, Status='live', Symbol='SNX18P10', Aliases='', description='sorting nexin 18 pseudogene 10', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8801140, end_position_on_the_genomic_accession=8801893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419041, CurrentID=0, Status='live', Symbol='SNX18P12', Aliases='', description='sorting nexin 18 pseudogene 12', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9364458, end_position_on_the_genomic_accession=9366038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419042, CurrentID=0, Status='live', Symbol='SNX18P13', Aliases='', description='sorting nexin 18 pseudogene 13', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13905738, end_position_on_the_genomic_accession=13907320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419043, CurrentID=0, Status='live', Symbol='SNX18P11', Aliases='', description='sorting nexin 18 pseudogene 11', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9092868, end_position_on_the_genomic_accession=9093609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419045, CurrentID=0, Status='live', Symbol='LOC100419045', Aliases='', description='tubulin, alpha 4a pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44840429, end_position_on_the_genomic_accession=44841580, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419046, CurrentID=0, Status='live', Symbol='LOC100419046', Aliases='', description='tubulin, alpha 4a pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67725226, end_position_on_the_genomic_accession=67726399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419047, CurrentID=0, Status='live', Symbol='SNX19P4', Aliases='', description='sorting nexin 19 pseudogene 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72483227, end_position_on_the_genomic_accession=72483855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372920, CurrentID=0, Status='live', Symbol='LOC105372920', Aliases='', description='uncharacterized LOC105372920', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217163169, end_position_on_the_genomic_accession=217190918, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100271085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271085, CurrentID=0, Status='live', Symbol='RPS20P15', Aliases='RPS20_5_351', description='ribosomal protein S20 pseudogene 15', other_designations='', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27462506, end_position_on_the_genomic_accession=27462819, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419052, CurrentID=0, Status='live', Symbol='LOC100419052', Aliases='', description='sorting nexin 18 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14252711, end_position_on_the_genomic_accession=14254293, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419053, CurrentID=0, Status='live', Symbol='LOC100419053', Aliases='', description='spermine synthase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47717807, end_position_on_the_genomic_accession=47719180, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419054, CurrentID=0, Status='live', Symbol='SAGE3P', Aliases='', description='sarcoma antigen 3, pseudogene', other_designations='sarcoma antigen 1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135637624, end_position_on_the_genomic_accession=135638932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419056, CurrentID=0, Status='live', Symbol='CLCN3P1', Aliases='', description='chloride channel, voltage-sensitive 3 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15034170, end_position_on_the_genomic_accession=15048066, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287985, CurrentID=0, Status='live', Symbol='LOC100287985', Aliases='', description='ring-box 1, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82609638, end_position_on_the_genomic_accession=82610070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419058, CurrentID=0, Status='live', Symbol='LOC100419058', Aliases='', description='GPN-loop GTPase 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131770783, end_position_on_the_genomic_accession=131771546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  148979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148979, CurrentID=0, Status='live', Symbol='GLIS1', Aliases='', description='GLIS family zinc finger 1', other_designations='GLI-similar 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53506233, end_position_on_the_genomic_accession=53734204, orientation='minus', exon_count=11, OMIM=610378, no_hdr0=''),
  103171572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103171572, CurrentID=0, Status='live', Symbol='DUP8Q22.1', Aliases='C8DUPq22.1', description='Leri pleonosteosis chromosome duplication syndrome', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151200, no_hdr0=''),
  101056699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101056699, CurrentID=0, Status='live', Symbol='TRK-CTT7-1', Aliases='TRNAL48, TRNAQ49P', description='transfer RNA-Lys (CTT) 7-1', other_designations='tRNA-Lys-CTT-7-1|transfer RNA glutamine 49 (anticodon CUG) pseudogene', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16553462, end_position_on_the_genomic_accession=16553533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419062, CurrentID=0, Status='live', Symbol='FOSL1P1', Aliases='', description='FOS-like antigen 1 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128155340, end_position_on_the_genomic_accession=128156399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419066, CurrentID=0, Status='live', Symbol='LOC100419066', Aliases='', description='nucleoporin 35kDa pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24157824, end_position_on_the_genomic_accession=24159343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419068, CurrentID=0, Status='live', Symbol='PNLIPP1', Aliases='', description='pancreatic lipase pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116578454, end_position_on_the_genomic_accession=116582958, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419069, CurrentID=0, Status='live', Symbol='ENO1P3', Aliases='', description='enolase 1, (alpha) pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124862097, end_position_on_the_genomic_accession=124863316, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419070, CurrentID=0, Status='live', Symbol='NUP210P3', Aliases='', description='nucleoporin 210kDa pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129323038, end_position_on_the_genomic_accession=129323976, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419073, CurrentID=0, Status='live', Symbol='LOC100419073', Aliases='', description='nucleoporin 155kDa pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79067114, end_position_on_the_genomic_accession=79068195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419074, CurrentID=0, Status='live', Symbol='AK3P2', Aliases='HsT26053', description='adenylate kinase 3 pseudogene 2', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143057063, end_position_on_the_genomic_accession=143057816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419075, CurrentID=0, Status='live', Symbol='AK3P3', Aliases='', description='adenylate kinase 3 pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23129178, end_position_on_the_genomic_accession=23129838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419077, CurrentID=0, Status='live', Symbol='FCRL6P1', Aliases='', description='Fc receptor-like 6 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159972551, end_position_on_the_genomic_accession=159974068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419079, CurrentID=0, Status='live', Symbol='LOC100419079', Aliases='', description='tubulin, alpha 1c pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62150780, end_position_on_the_genomic_accession=62151286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288008, CurrentID=0, Status='live', Symbol='PRR13P4', Aliases='', description='proline rich 13 pseudogene 4', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49469740, end_position_on_the_genomic_accession=49470416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271086, CurrentID=0, Status='live', Symbol='RPS20P16', Aliases='RPS20_7_597', description='ribosomal protein S20 pseudogene 16', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96942894, end_position_on_the_genomic_accession=96943189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288010, CurrentID=0, Status='live', Symbol='LOC100288010', Aliases='', description='cytochrome c oxidase subunit VIIa polypeptide 2 like pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154994172, end_position_on_the_genomic_accession=154994441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288013, CurrentID=0, Status='live', Symbol='NDUFA5P12', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 12', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 12|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47164849, end_position_on_the_genomic_accession=47166613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288016, CurrentID=0, Status='live', Symbol='LOC100288016', Aliases='', description='serine/threonine-protein phosphatase 4 regulatory subunit 2-like', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195650007, end_position_on_the_genomic_accession=195651708, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419092, CurrentID=0, Status='live', Symbol='LOC100419092', Aliases='', description='REST corepressor 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85336062, end_position_on_the_genomic_accession=85336571, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149013, CurrentID=0, Status='live', Symbol='NBPF12', Aliases='COAS1, KIAA1245', description='neuroblastoma breakpoint family member 12', other_designations='chomosome one amplified sequence 1 cyclophilin|chromosome 1 amplified sequence 1|neuroblastoma breakpoint family, member 12', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146938545, end_position_on_the_genomic_accession=146996093, orientation='plus', exon_count=37, OMIM=608607, no_hdr0=''),
  100288022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288022, CurrentID=0, Status='live', Symbol='VN1R33P', Aliases='', description='vomeronasal 1 receptor 33 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63401379, end_position_on_the_genomic_accession=63402271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419095, CurrentID=0, Status='live', Symbol='TXNL1P1', Aliases='', description='thioredoxin like 1 pseudogene 1', other_designations='thioredoxin-like 1 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=86725538, end_position_on_the_genomic_accession=86726056, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419096, CurrentID=0, Status='live', Symbol='LOC100419096', Aliases='', description='WD repeat domain 12 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102751247, end_position_on_the_genomic_accession=102752835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288025, CurrentID=0, Status='live', Symbol='DYZ1L2', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58823186, end_position_on_the_genomic_accession=58825242, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149018, CurrentID=0, Status='live', Symbol='LELP1', Aliases='', description='late cornified envelope-like proline-rich 1', other_designations='novel small proline rich protein|novel small proline-rich protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153203430, end_position_on_the_genomic_accession=153205125, orientation='plus', exon_count=2, OMIM=611042, no_hdr0=''),
  100288028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288028, CurrentID=0, Status='live', Symbol='LOC100288028', Aliases='', description='coiled-coil domain containing 59 pseudogene', other_designations='', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94675277, end_position_on_the_genomic_accession=94675827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419102, CurrentID=0, Status='live', Symbol='LOC100419102', Aliases='', description='calponin 3, acidic pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65797096, end_position_on_the_genomic_accession=65797905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288031, CurrentID=0, Status='live', Symbol='LOC100288031', Aliases='', description='BRICK1, SCAR/WAVE actin-nucleating complex subunit pseudogene', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9179385, end_position_on_the_genomic_accession=9179700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288035, CurrentID=0, Status='live', Symbol='NDUFA5P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 3|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19383220, end_position_on_the_genomic_accession=19383887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419108, CurrentID=0, Status='live', Symbol='LOC100419108', Aliases='', description='centromere protein L pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160437429, end_position_on_the_genomic_accession=160438261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288038, CurrentID=0, Status='live', Symbol='HNRNPA1P28', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 28', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119220721, end_position_on_the_genomic_accession=119221941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271087, CurrentID=0, Status='live', Symbol='RPS20P22', Aliases='RPS20_8_895', description='ribosomal protein S20 pseudogene 22', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38434337, end_position_on_the_genomic_accession=38434840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288042, CurrentID=0, Status='live', Symbol='ZNF519P3', Aliases='', description='zinc finger protein 519 pseudogene 3', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21911520, end_position_on_the_genomic_accession=21913041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149041, CurrentID=0, Status='live', Symbol='RC3H1', Aliases='RNF198, ROQUIN', description='ring finger and CCCH-type domains 1', other_designations='RING finger and C3H zinc finger protein 1|RING finger and CCCH-type zinc finger domain-containing protein 1|RING finger protein 198|probable E3 ubiquitin-protein ligase Roquin', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173931084, end_position_on_the_genomic_accession=173993072, orientation='minus', exon_count=19, OMIM=609424, no_hdr0=''),
  100288053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288053, CurrentID=0, Status='live', Symbol='DYZ1L3', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58826744, end_position_on_the_genomic_accession=58828686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149047, CurrentID=0, Status='live', Symbol='MGC27382', Aliases='', description='uncharacterized MGC27382', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78229599, end_position_on_the_genomic_accession=78369464, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  401557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401557, CurrentID=0, Status='live', Symbol='LOC401557', Aliases='', description='uncharacterized LOC401557', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135181609, end_position_on_the_genomic_accession=135187306, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419140, CurrentID=0, Status='live', Symbol='LOC100419140', Aliases='', description='ribosomal protein L5 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65209921, end_position_on_the_genomic_accession=65210945, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271088, CurrentID=0, Status='live', Symbol='RPS20P17', Aliases='RPS20_9_642', description='ribosomal protein S20 pseudogene 17', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177346078, end_position_on_the_genomic_accession=177346425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  767585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767585, CurrentID=0, Status='live', Symbol='SNORD114-9', Aliases='14q(II-9)', description='small nucleolar RNA, C/D box 114-9', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100966029, end_position_on_the_genomic_accession=100966099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100288072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288072, CurrentID=0, Status='live', Symbol='SDR42E2', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 2', other_designations='Putative 3-beta-HSD family protein ENSP00000330812', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22162438, end_position_on_the_genomic_accession=22191856, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100288073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288073, CurrentID=0, Status='live', Symbol='LOC100288073', Aliases='', description='heat shock protein family A (Hsp70) member 5 pseudogene', other_designations='heat shock 70kDa protein 5 (glucose-regulated protein, 78kDa) pseudogene', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164188384, end_position_on_the_genomic_accession=164190333, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646548, CurrentID=0, Status='live', Symbol='LOC646548', Aliases='', description='ADAM metallopeptidase domain 20 pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70230325, end_position_on_the_genomic_accession=70232294, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149069, CurrentID=0, Status='live', Symbol='DCDC2B', Aliases='', description='doublecortin domain containing 2B', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32209094, end_position_on_the_genomic_accession=32216196, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101056702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101056702, CurrentID=0, Status='live', Symbol='HHV8S', Aliases='', description='Human herpesvirus 8, susceptibility to', other_designations='', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614836, no_hdr0=''),
  100419154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419154, CurrentID=0, Status='live', Symbol='LOC100419154', Aliases='', description='ribosomal protein L4 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=45414340, end_position_on_the_genomic_accession=45414909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288083, CurrentID=0, Status='live', Symbol='CICP22', Aliases='', description='capicua transcriptional repressor pseudogene 22', other_designations='protein capicua homolog', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39796595, end_position_on_the_genomic_accession=39798437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149076, CurrentID=0, Status='live', Symbol='ZNF362', Aliases='RN, lin-29', description='zinc finger protein 362', other_designations='rotund homolog', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33256307, end_position_on_the_genomic_accession=33300720, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100288085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288085, CurrentID=0, Status='live', Symbol='DYZ1L5', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58836194, end_position_on_the_genomic_accession=58837606, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419158, CurrentID=0, Status='live', Symbol='SETP12', Aliases='', description='SET pseudogene 12', other_designations='SET nuclear oncogene pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120896118, end_position_on_the_genomic_accession=120896649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419161, CurrentID=0, Status='live', Symbol='LOC100419161', Aliases='', description='nei-like DNA glycosylase 2 pseudogene', other_designations='nei endonuclease VIII-like 2 (E. coli) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150422651, end_position_on_the_genomic_accession=150423642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373213, CurrentID=0, Status='live', Symbol='LOC105373213', Aliases='', description='uncharacterized LOC105373213', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288093, CurrentID=0, Status='live', Symbol='LOC100288093', Aliases='', description='nuclear RNA export factor 3 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53657765, end_position_on_the_genomic_accession=53658592, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149086, CurrentID=0, Status='live', Symbol='LINC01225', Aliases='', description='long intergenic non-protein coding RNA 1225', other_designations='ornithine decarboxylase pseudogene', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31500085, end_position_on_the_genomic_accession=31508566, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100288097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288097, CurrentID=0, Status='live', Symbol='LOC100288097', Aliases='', description='BCL2 associated athanogene 4 pseudogene', other_designations='BCL2-associated athanogene 4 pseudogene', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69859270, end_position_on_the_genomic_accession=69860540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419170, CurrentID=0, Status='live', Symbol='LOC100419170', Aliases='', description='toll-like receptor 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153678172, end_position_on_the_genomic_accession=153680085, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271089, CurrentID=0, Status='live', Symbol='RPS20P23', Aliases='RPS20_9_932', description='ribosomal protein S20 pseudogene 23', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100776512, end_position_on_the_genomic_accession=100776868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419174, CurrentID=0, Status='live', Symbol='LOC100419174', Aliases='', description='WD repeat domain 12 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100645233, end_position_on_the_genomic_accession=100646819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149095, CurrentID=0, Status='live', Symbol='DCST1', Aliases='', description='DC-STAMP domain containing 1', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155033806, end_position_on_the_genomic_accession=155050930, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=600, CurrentID=0, Status='live', Symbol='BAK1P1', Aliases='BAK2, BCL2L7P1, Bak-2', description='BCL2-antagonist/killer 1 pseudogene 1', other_designations='BCL2-like 7 pseudogene 1', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32688930, end_position_on_the_genomic_accession=32691054, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873393, CurrentID=0, Status='live', Symbol='RNA5SP123', Aliases='RN5S123', description='RNA, 5S ribosomal pseudogene 123', other_designations='RNA, 5S ribosomal 123', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12511095, end_position_on_the_genomic_accession=12511213, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419178, CurrentID=0, Status='live', Symbol='SETP6', Aliases='', description='SET pseudogene 6', other_designations='SET nuclear oncogene pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=81612238, end_position_on_the_genomic_accession=81612805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377578, CurrentID=0, Status='live', Symbol='LOC105377578', Aliases='', description='uncharacterized LOC105377578', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183073369, end_position_on_the_genomic_accession=183092283, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929607, CurrentID=0, Status='live', Symbol='LOC101929607', Aliases='', description='uncharacterized LOC101929607', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107430930, end_position_on_the_genomic_accession=107463912, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284294, CurrentID=0, Status='live', Symbol='LOC284294', Aliases='', description='uncharacterized LOC284294', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64104091, end_position_on_the_genomic_accession=64423601, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100288111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288111, CurrentID=0, Status='live', Symbol='BOLA3P1', Aliases='', description='bolA family member 3 pseudogene 1', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102880484, end_position_on_the_genomic_accession=102881037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288112, CurrentID=0, Status='live', Symbol='OR11J7P', Aliases='', description='olfactory receptor family 11 subfamily J member 7 pseudogene', other_designations='olfactory receptor, family 11, subfamily J, member 7 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21971719, end_position_on_the_genomic_accession=21972625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419185, CurrentID=0, Status='live', Symbol='PDLIM1P4', Aliases='', description='PDZ and LIM domain 1 pseudogene 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98782209, end_position_on_the_genomic_accession=98783190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288114, CurrentID=0, Status='live', Symbol='LOC100288114', Aliases='', description='uncharacterized LOC100288114', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=56988366, end_position_on_the_genomic_accession=57005676, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100419187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419187, CurrentID=0, Status='live', Symbol='SETP11', Aliases='', description='SET pseudogene 11', other_designations='SET nuclear oncogene pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151269438, end_position_on_the_genomic_accession=151270131, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149111, CurrentID=0, Status='live', Symbol='CNIH3', Aliases='CNIH-3', description='cornichon family AMPA receptor auxiliary protein 3', other_designations='cornichon homolog 3', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224434790, end_position_on_the_genomic_accession=224740567, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100288122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288122, CurrentID=0, Status='live', Symbol='LDLRAD4-AS1', Aliases='', description='LDLRAD4 antisense RNA 1', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13419499, end_position_on_the_genomic_accession=13427534, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288123, CurrentID=0, Status='live', Symbol='LOC100288123', Aliases='', description='uncharacterized LOC100288123', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1822089, end_position_on_the_genomic_accession=1824543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288124, CurrentID=0, Status='live', Symbol='DYZ1L8', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58844485, end_position_on_the_genomic_accession=58846452, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  767594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767594, CurrentID=0, Status='live', Symbol='SNORD114-16', Aliases='14q(II-16)', description='small nucleolar RNA, C/D box 114-16', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100973595, end_position_on_the_genomic_accession=100973663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271090, CurrentID=0, Status='live', Symbol='RPS21P1', Aliases='RPS21_1_179', description='ribosomal protein S21 pseudogene 1', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235432947, end_position_on_the_genomic_accession=235433232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288130, CurrentID=0, Status='live', Symbol='MTCL1P1', Aliases='CCDC165P1, SOGA2P1', description='MTCL1 pseudogene 1', other_designations='SOGA family member 2 pseudogene 1|coiled-coil domain containing 165 pseudogene 1', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=70923659, end_position_on_the_genomic_accession=70925610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419209, CurrentID=0, Status='live', Symbol='PRSS52P', Aliases='', description='protease, serine 52, pseudogene', other_designations='protease, serine, 52, pseudogene|protease, serine, 55 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10474882, end_position_on_the_genomic_accession=10498737, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929608, CurrentID=0, Status='live', Symbol='LOC101929608', Aliases='', description='uncharacterized LOC101929608', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21610787, end_position_on_the_genomic_accession=21703585, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100288139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288139, CurrentID=0, Status='live', Symbol='PIH2', Aliases='', description='pregnancy-induced hypertension syndrome-related protein', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=70846744, end_position_on_the_genomic_accession=70850618, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  134082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134082, CurrentID=0, Status='live', Symbol='OR2AI1P', Aliases='', description='olfactory receptor family 2 subfamily AI member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AI, member 1 pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180692835, end_position_on_the_genomic_accession=180693763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288142, CurrentID=0, Status='live', Symbol='NBPF20', Aliases='', description='neuroblastoma breakpoint family member 20', other_designations='neuroblastoma breakpoint family, member 20', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145289900, end_position_on_the_genomic_accession=145405778, orientation='minus', exon_count=138, OMIM=614007, no_hdr0=''),
  100288146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288146, CurrentID=0, Status='live', Symbol='LOC100288146', Aliases='', description='nucleoporin Nup43-like', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105102892, end_position_on_the_genomic_accession=105103994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288148, CurrentID=0, Status='live', Symbol='OR11J6P', Aliases='', description='olfactory receptor family 11 subfamily J member 6 pseudogene', other_designations='olfactory receptor, family 11, subfamily J, member 6 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21980313, end_position_on_the_genomic_accession=21981209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288149, CurrentID=0, Status='live', Symbol='NDUFS5P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 3', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52709081, end_position_on_the_genomic_accession=52709557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288151, CurrentID=0, Status='live', Symbol='LOC100288151', Aliases='', description='SREK1-interacting protein 1 pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19739709, end_position_on_the_genomic_accession=19740147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288152, CurrentID=0, Status='live', Symbol='LOC100288152', Aliases='', description='uncharacterized LOC100288152', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=473236, end_position_on_the_genomic_accession=480892, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  767599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767599, CurrentID=0, Status='live', Symbol='SNORD114-21', Aliases='14q(II-21)', description='small nucleolar RNA, C/D box 114-21', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100981975, end_position_on_the_genomic_accession=100982045, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100288157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288157, CurrentID=0, Status='live', Symbol='DYZ1L9', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58848059, end_position_on_the_genomic_accession=58850136, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419232, CurrentID=0, Status='live', Symbol='LOC100419232', Aliases='', description='UBX domain protein 4 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47389961, end_position_on_the_genomic_accession=47391336, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288162, CurrentID=0, Status='live', Symbol='LOC100288162', Aliases='', description='uncharacterized LOC100288162', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14925750, end_position_on_the_genomic_accession=14927283, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100288165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288165, CurrentID=0, Status='live', Symbol='HMGN2P15', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 15', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 15', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42752731, end_position_on_the_genomic_accession=42753906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419238, CurrentID=0, Status='live', Symbol='LOC100419238', Aliases='', description='centromere protein V pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51396095, end_position_on_the_genomic_accession=51396670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929609, CurrentID=0, Status='live', Symbol='LOC101929609', Aliases='', description='uncharacterized LOC101929609', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43944370, end_position_on_the_genomic_accession=43946757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100288170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288170, CurrentID=0, Status='live', Symbol='LOC100288170', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75465129, end_position_on_the_genomic_accession=75465570, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288172, CurrentID=0, Status='live', Symbol='LOC100288172', Aliases='', description='WEE1 homolog (S. pombe) pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=201420, end_position_on_the_genomic_accession=205027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288173, CurrentID=0, Status='live', Symbol='TRMT112P4', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 4', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101356298, end_position_on_the_genomic_accession=101356844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419246, CurrentID=0, Status='live', Symbol='BTG3P1', Aliases='', description='BTG family member 3 pseudogene 1', other_designations='BTG family, member 3 pseudogene 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128122817, end_position_on_the_genomic_accession=128123475, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288175, CurrentID=0, Status='live', Symbol='LOC100288175', Aliases='', description='uncharacterized LOC100288175', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1059713, end_position_on_the_genomic_accession=1061058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  767603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767603, CurrentID=0, Status='live', Symbol='SNORD114-23', Aliases='14q(II-23)', description='small nucleolar RNA, C/D box 114-23', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100983876, end_position_on_the_genomic_accession=100983946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100288181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288181, CurrentID=0, Status='live', Symbol='LOC100288181', Aliases='', description='uncharacterized LOC100288181', other_designations='CTD-2292P10.4', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142702252, end_position_on_the_genomic_accession=142726973, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  149175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149175, CurrentID=0, Status='live', Symbol='MANEAL', Aliases='', description='mannosidase, endo-alpha-like', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37793820, end_position_on_the_genomic_accession=37801606, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100288185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288185, CurrentID=0, Status='live', Symbol='PPIAP1', Aliases='PPIAP', description='peptidylprolyl isomerase A (cyclophilin A), pseudogene 1', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20827918, end_position_on_the_genomic_accession=20828664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288188, CurrentID=0, Status='live', Symbol='DYZ1L10', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58851653, end_position_on_the_genomic_accession=58853715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271092, CurrentID=0, Status='live', Symbol='RPS23P4', Aliases='RPS23_1_526', description='ribosomal protein S23 pseudogene 4', other_designations='', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145269810, end_position_on_the_genomic_accession=145270131, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288192, CurrentID=0, Status='live', Symbol='LOC100288192', Aliases='', description='archaelysin family metallopeptidase 2 pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127930871, end_position_on_the_genomic_accession=127931587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390306, CurrentID=0, Status='live', Symbol='LOC390306', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39466389, end_position_on_the_genomic_accession=39467362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288195, CurrentID=0, Status='live', Symbol='VN1R104P', Aliases='', description='vomeronasal 1 receptor 104 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54320234, end_position_on_the_genomic_accession=54321151, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288198, CurrentID=0, Status='live', Symbol='HMGN3-AS1', Aliases='', description='HMGN3 antisense RNA 1', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79233674, end_position_on_the_genomic_accession=79236800, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  149194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149194, CurrentID=0, Status='live', Symbol='LOC149194', Aliases='', description='uncharacterized LOC149194', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288203, CurrentID=0, Status='live', Symbol='LOC100288203', Aliases='', description='WAS protein homolog associated with actin, golgi membranes and microtubules pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30470682, end_position_on_the_genomic_accession=30484687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288208, CurrentID=0, Status='live', Symbol='LOC100288208', Aliases='', description='uncharacterized LOC100288208', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=74987094, end_position_on_the_genomic_accession=74993252, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100288211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288211, CurrentID=0, Status='live', Symbol='LOC100288211', Aliases='', description='PPPDE peptidase domain containing 1 pseudogene', other_designations='desumoylating isopeptidase 2 pseudogene', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57317215, end_position_on_the_genomic_accession=57317825, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419284, CurrentID=0, Status='live', Symbol='LOC100419284', Aliases='', description='RAN binding protein 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124598558, end_position_on_the_genomic_accession=124599676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  767609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767609, CurrentID=0, Status='live', Symbol='SNORD114-28', Aliases='14q(II-28)', description='small nucleolar RNA, C/D box 114-28', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100989130, end_position_on_the_genomic_accession=100989200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271093, CurrentID=0, Status='live', Symbol='RPS23P5', Aliases='RPS23_2_542', description='ribosomal protein S23 pseudogene 5', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=13250235, end_position_on_the_genomic_accession=13250662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288220, CurrentID=0, Status='live', Symbol='DYZ1L11', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58855197, end_position_on_the_genomic_accession=58857154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419293, CurrentID=0, Status='live', Symbol='LOC100419293', Aliases='', description='carbonyl reductase 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45200885, end_position_on_the_genomic_accession=45201530, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288222, CurrentID=0, Status='live', Symbol='ATP5G1P8', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 8', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 8', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=76836346, end_position_on_the_genomic_accession=76836957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  767611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767611, CurrentID=0, Status='live', Symbol='SNORD114-30', Aliases='14q(II-30)', description='small nucleolar RNA, C/D box 114-30', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100991919, end_position_on_the_genomic_accession=100991989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419301, CurrentID=0, Status='live', Symbol='LOC100419301', Aliases='', description='ribosomal protein S8 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20079859, end_position_on_the_genomic_accession=20080539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149224, CurrentID=0, Status='live', Symbol='RPS7P4', Aliases='RPS7_1_61', description='ribosomal protein S7 pseudogene 4', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68242429, end_position_on_the_genomic_accession=68243099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288233, CurrentID=0, Status='live', Symbol='LOC100288233', Aliases='', description='ribosomal protein L23 pseudogene', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24681936, end_position_on_the_genomic_accession=24682634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288235, CurrentID=0, Status='live', Symbol='HSPE1P10', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 10', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 10', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126737771, end_position_on_the_genomic_accession=126738031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288238, CurrentID=0, Status='live', Symbol='LOC100288238', Aliases='', description='uncharacterized LOC100288238', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=21394887, end_position_on_the_genomic_accession=21402975, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  149233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149233, CurrentID=0, Status='live', Symbol='IL23R', Aliases='', description='interleukin 23 receptor', other_designations='IL-23 receptor', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67138640, end_position_on_the_genomic_accession=67259979, orientation='plus', exon_count=14, OMIM=607562, no_hdr0=''),
  105373218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373218, CurrentID=0, Status='live', Symbol='LOC105373218', Aliases='', description='uncharacterized LOC105373218', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236906694, end_position_on_the_genomic_accession=236921553, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100419318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419318, CurrentID=0, Status='live', Symbol='LOC100419318', Aliases='', description='FAST kinase domains 3 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8455946, end_position_on_the_genomic_accession=8456558, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419319, CurrentID=0, Status='live', Symbol='LOC100419319', Aliases='', description='UBX domain protein 2A pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20830166, end_position_on_the_genomic_accession=20830947, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288248, CurrentID=0, Status='live', Symbol='ZNHIT1P1', Aliases='', description='zinc finger, HIT-type containing 1 pseudogene 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142858354, end_position_on_the_genomic_accession=142858794, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271094, CurrentID=0, Status='live', Symbol='RPS24P5', Aliases='RPS24_1_136', description='ribosomal protein S24 pseudogene 5', other_designations='', map_location='1p36.13-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180000440, end_position_on_the_genomic_accession=180000808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288250, CurrentID=0, Status='live', Symbol='ISCA1P5', Aliases='', description='iron-sulfur cluster assembly 1 pseudogene 5', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 5', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19827014, end_position_on_the_genomic_accession=19827476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288252, CurrentID=0, Status='live', Symbol='RPL10P1', Aliases='RPL10P, RPL10_7_1723', description='ribosomal protein L10 pseudogene 1', other_designations='60S ribosomal protein L10 pseudogene', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27420318, end_position_on_the_genomic_accession=27421044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288254, CurrentID=0, Status='live', Symbol='LOC100288254', Aliases='', description='uncharacterized LOC100288254', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172285597, end_position_on_the_genomic_accession=172298376, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288259, CurrentID=0, Status='live', Symbol='DYZ1L12', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58857539, end_position_on_the_genomic_accession=58860728, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288264, CurrentID=0, Status='live', Symbol='VN1R105P', Aliases='', description='vomeronasal 1 receptor 105 pseudogene', other_designations='vomeronasal 1 receptor 1 pseudogene', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54647911, end_position_on_the_genomic_accession=54648865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419337, CurrentID=0, Status='live', Symbol='CENPIP1', Aliases='', description='centromere protein I pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19126477, end_position_on_the_genomic_accession=19131640, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928465, CurrentID=0, Status='live', Symbol='LOC101928465', Aliases='', description='uncharacterized LOC101928465', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62648961, end_position_on_the_genomic_accession=62656352, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419353, CurrentID=0, Status='live', Symbol='LOC100419353', Aliases='', description='histone cluster 3, H3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126106142, end_position_on_the_genomic_accession=126106923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288283, CurrentID=0, Status='live', Symbol='HNRNPA1P61', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 61', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33636442, end_position_on_the_genomic_accession=33637655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419356, CurrentID=0, Status='live', Symbol='LOC100419356', Aliases='', description='ethanolamine kinase 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143760533, end_position_on_the_genomic_accession=143762085, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873399, CurrentID=0, Status='live', Symbol='RNA5SP129', Aliases='RN5S129', description='RNA, 5S ribosomal pseudogene 129', other_designations='RNA, 5S ribosomal 129', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37341649, end_position_on_the_genomic_accession=37341766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288287, CurrentID=0, Status='live', Symbol='KRTAP22-2', Aliases='KAP22.2', description='keratin associated protein 22-2', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30590105, end_position_on_the_genomic_accession=30590397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149281, CurrentID=0, Status='live', Symbol='METTL11B', Aliases='C1orf184, HOMT1B, NTM1B', description='methyltransferase like 11B', other_designations='X-Pro-Lys N-terminal protein methyltransferase 1B|methyltransferase-like protein 11B', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170146047, end_position_on_the_genomic_accession=170168923, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100288293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288293, CurrentID=0, Status='live', Symbol='DYZ1L13', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58861113, end_position_on_the_genomic_accession=58862528, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419366, CurrentID=0, Status='live', Symbol='LOC100419366', Aliases='', description='phosphodiesterase 8B pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=46099203, end_position_on_the_genomic_accession=46100374, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288298, CurrentID=0, Status='live', Symbol='GCSHP2', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene 2', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103184322, end_position_on_the_genomic_accession=103185071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149297, CurrentID=0, Status='live', Symbol='FAM78B', Aliases='', description='family with sequence similarity 78 member B', other_designations='family with sequence similarity 78, member B', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166057426, end_position_on_the_genomic_accession=166166755, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100288306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288306, CurrentID=0, Status='live', Symbol='LOC100288306', Aliases='', description='uncharacterized LOC100288306', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=126372994, end_position_on_the_genomic_accession=126373107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100288307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288307, CurrentID=0, Status='live', Symbol='USMG5P1', Aliases='', description='USMG5 pseudogene 1', other_designations='up-regulated during skeletal muscle growth 5 homolog pseudogene 1|upregulated during skeletal muscle growth 5 pseudogene 1', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74173861, end_position_on_the_genomic_accession=74174190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288310, CurrentID=0, Status='live', Symbol='LOC100288310', Aliases='', description='uncharacterized LOC100288310', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=73856383, end_position_on_the_genomic_accession=73860870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  134111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134111, CurrentID=0, Status='live', Symbol='UBE2QL1', Aliases='', description='ubiquitin conjugating enzyme E2Q family-like 1', other_designations='E2Q-like ubiquitin-conjugating enzyme 1|ubiquitin-conjugating enzyme E2Q family-like 1', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6437347, end_position_on_the_genomic_accession=6496721, orientation='plus', exon_count=3, OMIM=615832, no_hdr0=''),
  100419391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419391, CurrentID=0, Status='live', Symbol='RPL21P135', Aliases='', description='ribosomal protein L21 pseudogene 135', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37748071, end_position_on_the_genomic_accession=37748935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419392, CurrentID=0, Status='live', Symbol='RPL12P44', Aliases='', description='ribosomal protein L12 pseudogene 44', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44937585, end_position_on_the_genomic_accession=44938525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288323, CurrentID=0, Status='live', Symbol='KRTAP21-3', Aliases='', description='keratin associated protein 21-3', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30718525, end_position_on_the_genomic_accession=30718777, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288329, CurrentID=0, Status='live', Symbol='DYZ1L14', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58865932, end_position_on_the_genomic_accession=58868009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288330, CurrentID=0, Status='live', Symbol='H2AFZP5', Aliases='', description='H2A histone family member Z pseudogene 5', other_designations='H2A histone family, member Z pseudogene 5', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77953495, end_position_on_the_genomic_accession=77953882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288332, CurrentID=0, Status='live', Symbol='NPIPA5', Aliases='NPIP', description='nuclear pore complex interacting protein family member A5', other_designations='nuclear pore complex interacting protein family, member A5', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15363624, end_position_on_the_genomic_accession=15381972, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105373221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373221, CurrentID=0, Status='live', Symbol='LOC105373221', Aliases='', description='uncharacterized LOC105373221', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100033403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033403, CurrentID=0, Status='live', Symbol='VN2R15P', Aliases='', description='vomeronasal 2 receptor 15 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11915326, end_position_on_the_genomic_accession=11915561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288337, CurrentID=0, Status='live', Symbol='NDUFB5P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B5 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 5, 16kDa pseudogene', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=180573467, end_position_on_the_genomic_accession=180574027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271097, CurrentID=0, Status='live', Symbol='RPS24P10', Aliases='RPS24_3_344', description='ribosomal protein S24 pseudogene 10', other_designations='', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15131836, end_position_on_the_genomic_accession=15132203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288343, CurrentID=0, Status='live', Symbol='SLC25A14P1', Aliases='', description='solute carrier family 25 (mitochondrial carrier, brain), member 14 pseudogene 1', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83477438, end_position_on_the_genomic_accession=83478621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419416, CurrentID=0, Status='live', Symbol='SRPK2P', Aliases='SRPK2ps', description='SRSF protein kinase 2 pseudogene', other_designations='SFRS protein kinase 2 pseudogene', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62863082, end_position_on_the_genomic_accession=62864005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873401, CurrentID=0, Status='live', Symbol='RNA5SP132', Aliases='RN5S132', description='RNA, 5S ribosomal pseudogene 132', other_designations='RNA, 5S ribosomal 132', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51694465, end_position_on_the_genomic_accession=51694582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288346, CurrentID=0, Status='live', Symbol='NCAM1-AS1', Aliases='', description='NCAM1 antisense RNA1', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113269532, end_position_on_the_genomic_accession=113273901, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929615, CurrentID=0, Status='live', Symbol='LOC101929615', Aliases='', description='uncharacterized LOC101929615', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164653383, end_position_on_the_genomic_accession=164667896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419423, CurrentID=0, Status='live', Symbol='ASNSP4', Aliases='', description='asparagine synthetase pseudogene 4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46697630, end_position_on_the_genomic_accession=46698685, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149345, CurrentID=0, Status='live', Symbol='SHISA4', Aliases='C1orf40, TMEM58', description='shisa family member 4', other_designations='shisa homolog 4|transmembrane protein 58', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201888669, end_position_on_the_genomic_accession=201892587, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100419426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419426, CurrentID=0, Status='live', Symbol='PDLIM1P2', Aliases='', description='PDZ and LIM domain 1 pseudogene 2', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21444831, end_position_on_the_genomic_accession=21445435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419427, CurrentID=0, Status='live', Symbol='PDLIM1P3', Aliases='', description='PDZ and LIM domain 1 pseudogene 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27002653, end_position_on_the_genomic_accession=27003259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  131544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131544, CurrentID=0, Status='live', Symbol='CRYBG3', Aliases='DKFZp667G2110, vlAKAP', description='crystallin beta-gamma domain containing 3', other_designations='beta-gamma crystallin domain containing 3|beta/gamma crystallin domain-containing protein 3', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97821937, end_position_on_the_genomic_accession=97944986, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100288358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288358, CurrentID=0, Status='live', Symbol='LOC100288358', Aliases='', description='uncharacterized LOC100288358', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=75681602, end_position_on_the_genomic_accession=75688250, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  439953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439953, CurrentID=0, Status='live', Symbol='PPIAP31', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 31', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32188125, end_position_on_the_genomic_accession=32188855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288363, CurrentID=0, Status='live', Symbol='DYZ1L15', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58868289, end_position_on_the_genomic_accession=58869710, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419436, CurrentID=0, Status='live', Symbol='LOC100419436', Aliases='', description='ribosomal protein S2 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18010644, end_position_on_the_genomic_accession=18011826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288365, CurrentID=0, Status='live', Symbol='LOC100288365', Aliases='', description='activating transcription factor 7 interacting protein 2 pseudogene', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11135482, end_position_on_the_genomic_accession=11136745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288367, CurrentID=0, Status='live', Symbol='LOC100288367', Aliases='', description='golgin A2 pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84391915, end_position_on_the_genomic_accession=84395345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134121, CurrentID=0, Status='live', Symbol='C5orf49', Aliases='', description='chromosome 5 open reading frame 49', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7830378, end_position_on_the_genomic_accession=7851490, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288378, CurrentID=0, Status='live', Symbol='LRRC34P1', Aliases='', description='leucine rich repeat containing 34 pseudogene 1', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22573996, end_position_on_the_genomic_accession=22575105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149371, CurrentID=0, Status='live', Symbol='EXOC8', Aliases='EXO84, Exo84p, SEC84', description='exocyst complex component 8', other_designations='exocyst complex 84 kDa subunit', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231332731, end_position_on_the_genomic_accession=231337872, orientation='minus', exon_count=1, OMIM=615283, no_hdr0=''),
  100288380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288380, CurrentID=0, Status='live', Symbol='ULK4P2', Aliases='D-X, FAM7A2', description='ULK4 pseudogene 2', other_designations='family with sequence similarity 7, member A2|unc-51-like kinase 4 (ULK4) pseudogene|unc-51-like kinase 4 pseudogene 2', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30572555, end_position_on_the_genomic_accession=30600708, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  149373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149373, CurrentID=0, Status='live', Symbol='LOC149373', Aliases='', description='uncharacterized LOC149373', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231184099, end_position_on_the_genomic_accession=231187627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379033, CurrentID=0, Status='live', Symbol='LOC105379033', Aliases='', description='uncharacterized LOC105379033', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73337790, end_position_on_the_genomic_accession=73338765, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288385, CurrentID=0, Status='live', Symbol='LOC100288385', Aliases='', description='cytochrome c oxidase subunit VIIa polypeptide 2 like pseudogene', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113769660, end_position_on_the_genomic_accession=113769996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419458, CurrentID=0, Status='live', Symbol='LOC100419458', Aliases='', description='ribosomal protein L21 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=68640758, end_position_on_the_genomic_accession=68641607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288392, CurrentID=0, Status='live', Symbol='LOC100288392', Aliases='', description='olfactory receptor pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8981409, end_position_on_the_genomic_accession=8982415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373223, CurrentID=0, Status='live', Symbol='LOC105373223', Aliases='', description='uncharacterized LOC105373223', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239248069, end_position_on_the_genomic_accession=239254929, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100288397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288397, CurrentID=0, Status='live', Symbol='RAB9AP4', Aliases='', description='RAB9A, member RAS oncogene family pseudogene 4', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18649646, end_position_on_the_genomic_accession=18650768, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419471, CurrentID=0, Status='live', Symbol='RPS23P10', Aliases='', description='ribosomal protein S23 pseudogene 10', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161536571, end_position_on_the_genomic_accession=161538160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419472, CurrentID=0, Status='live', Symbol='RPS23P9', Aliases='', description='ribosomal protein S23 pseudogene 9', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161617995, end_position_on_the_genomic_accession=161619585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288402, CurrentID=0, Status='live', Symbol='DYZ1L17', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58876644, end_position_on_the_genomic_accession=58878716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288405, CurrentID=0, Status='live', Symbol='SFTPA3P', Aliases='COLEC6, SFTPAP1, SFTPP1', description='surfactant protein A3, pseudogene', other_designations='surfactant protein A pseudogene 1|surfactant, pulmonary-associated protein A pseudogene 1|surfactant-associated protein pseudogene 1', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79593947, end_position_on_the_genomic_accession=79595657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374043, CurrentID=0, Status='live', Symbol='LOC105374043', Aliases='', description='uncharacterized LOC105374043', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112142794, end_position_on_the_genomic_accession=112200721, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929617, CurrentID=0, Status='live', Symbol='LOC101929617', Aliases='', description='uncharacterized LOC101929617', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=22754448, end_position_on_the_genomic_accession=22760403, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  149401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149401, CurrentID=0, Status='live', Symbol='LOC149401', Aliases='', description='uncharacterized LOC149401', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288413, CurrentID=0, Status='live', Symbol='ERVMER34-1', Aliases='', description='endogenous retrovirus group MER34 member 1', other_designations='HERV-MER_4q12 provirus ancestral Env polyprotein|LP9056 protein|endogenous retrovirus group MER34, member 1', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52738952, end_position_on_the_genomic_accession=52751640, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100288416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288416, CurrentID=0, Status='live', Symbol='LOC100288416', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108836333, end_position_on_the_genomic_accession=108836594, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419489, CurrentID=0, Status='live', Symbol='HDAC1P2', Aliases='', description='histone deacetylase 1 pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220625706, end_position_on_the_genomic_accession=220627642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288420, CurrentID=0, Status='live', Symbol='NDUFB4P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B4 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 4, 15kDa pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19840096, end_position_on_the_genomic_accession=19840567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101730217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101730217, CurrentID=0, Status='live', Symbol='SPECC1L-ADORA2A', Aliases='', description='SPECC1L-ADORA2A readthrough (NMD candidate)', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24270817, end_position_on_the_genomic_accession=24442360, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  105372923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372923, CurrentID=0, Status='live', Symbol='LOC105372923', Aliases='', description='uncharacterized LOC105372923', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218165056, end_position_on_the_genomic_accession=218198611, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  149420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149420, CurrentID=0, Status='live', Symbol='PDIK1L', Aliases='CLIK1L, STK35L2', description='PDLIM1 interacting kinase 1 like', other_designations='PDLIM1-interacting kinase 1-like|casein kinase', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26111165, end_position_on_the_genomic_accession=26125548, orientation='plus', exon_count=5, OMIM=610785, no_hdr0=''),
  100271100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271100, CurrentID=0, Status='live', Symbol='RPS25P2', Aliases='RPS25_1_159', description='ribosomal protein S25 pseudogene 2', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211173418, end_position_on_the_genomic_accession=211173898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288430, CurrentID=0, Status='live', Symbol='LOC100288430', Aliases='', description='uncharacterized LOC100288430', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9000458, end_position_on_the_genomic_accession=9074790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419503, CurrentID=0, Status='live', Symbol='LOC100419503', Aliases='', description='ribosomal protein S2 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74658042, end_position_on_the_genomic_accession=74658601, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288432, CurrentID=0, Status='live', Symbol='IL10RB-AS1', Aliases='IFNAR2-AS1', description='IL10RB antisense RNA 1 (head to head)', other_designations='IFNAR2 antisense RNA 1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33265632, end_position_on_the_genomic_accession=33266260, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100270832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270832, CurrentID=0, Status='live', Symbol='RPL5P9', Aliases='RPL5_3_263', description='ribosomal protein L5 pseudogene 9', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110967776, end_position_on_the_genomic_accession=110968780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288435, CurrentID=0, Status='live', Symbol='DYZ1L18', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58878996, end_position_on_the_genomic_accession=58880408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149428, CurrentID=0, Status='live', Symbol='BNIPL', Aliases='BNIP-S-1, BNIPL-2, BNIPL1, BNIPL2, BNIPS, PP753, BNIPL', description='BCL2/adenovirus E1B 19kD interacting protein like', other_designations='BNIP-2 similar', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151036553, end_position_on_the_genomic_accession=151047600, orientation='plus', exon_count=11, OMIM=611275, no_hdr0=''),
  105377589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377589, CurrentID=0, Status='live', Symbol='LOC105377589', Aliases='', description='uncharacterized LOC105377589', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185323810, end_position_on_the_genomic_accession=185341987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419511, CurrentID=0, Status='live', Symbol='LOC100419511', Aliases='', description='zinc finger protein 79 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222945935, end_position_on_the_genomic_accession=222946640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288440, CurrentID=0, Status='live', Symbol='COX5BP5', Aliases='COX5BL5', description='cytochrome c oxidase subunit Vb pseudogene 5', other_designations='cytochrome c oxidase subunit Vb-like 5', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49709002, end_position_on_the_genomic_accession=49709464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419513, CurrentID=0, Status='live', Symbol='LOC100419513', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199108584, end_position_on_the_genomic_accession=199110088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288443, CurrentID=0, Status='live', Symbol='LOC100288443', Aliases='', description='uncharacterized LOC100288443', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=124769622, end_position_on_the_genomic_accession=124771990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419516, CurrentID=0, Status='live', Symbol='LOC100419516', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93493722, end_position_on_the_genomic_accession=93494992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419520, CurrentID=0, Status='live', Symbol='LOC100419520', Aliases='', description='RAN binding protein 6 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9506152, end_position_on_the_genomic_accession=9508607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419521, CurrentID=0, Status='live', Symbol='RPL21P136', Aliases='', description='ribosomal protein L21 pseudogene 136', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12614654, end_position_on_the_genomic_accession=12615909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288451, CurrentID=0, Status='live', Symbol='RBBP4P5', Aliases='', description='retinoblastoma binding protein 4 pseudogene 5', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21563818, end_position_on_the_genomic_accession=21566329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373225, CurrentID=0, Status='live', Symbol='LOC105373225', Aliases='', description='uncharacterized LOC105373225', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239587932, end_position_on_the_genomic_accession=239633648, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100288463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288463, CurrentID=0, Status='live', Symbol='CUL1P1', Aliases='', description='cullin 1 pseudogene 1', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127784165, end_position_on_the_genomic_accession=127787386, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873405, CurrentID=0, Status='live', Symbol='RNA5SP137', Aliases='RN5S137', description='RNA, 5S ribosomal pseudogene 137', other_designations='RNA, 5S ribosomal 137', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125058304, end_position_on_the_genomic_accession=125058438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288468, CurrentID=0, Status='live', Symbol='DYZ1L21', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58890900, end_position_on_the_genomic_accession=58892962, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149461, CurrentID=0, Status='live', Symbol='CLDN19', Aliases='HOMG5', description='claudin 19', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42733093, end_position_on_the_genomic_accession=42740254, orientation='minus', exon_count=4, OMIM=610036, no_hdr0=''),
  100288470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288470, CurrentID=0, Status='live', Symbol='REXO1L11P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 11, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 11, pseudogene|REX1, RNA exonuclease 1 homolog-like 1 pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85735314, end_position_on_the_genomic_accession=85735508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370881, CurrentID=0, Status='live', Symbol='LOC105370881', Aliases='', description='uncharacterized LOC105370881', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70692816, end_position_on_the_genomic_accession=70695010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  149465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149465, CurrentID=0, Status='live', Symbol='CFAP57', Aliases='VWS2, WDR65', description='cilia and flagella associated protein 57', other_designations='WD repeat domain 65|WD repeat-containing protein 65', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43172330, end_position_on_the_genomic_accession=43254358, orientation='plus', exon_count=28, OMIM=614259, no_hdr0=''),
  149466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149466, CurrentID=0, Status='live', Symbol='C1orf210', Aliases='TEMP', description='chromosome 1 open reading frame 210', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43281865, end_position_on_the_genomic_accession=43285840, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419548, CurrentID=0, Status='live', Symbol='LOC100419548', Aliases='', description='ribosomal protein S27a pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138266999, end_position_on_the_genomic_accession=138267757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419549, CurrentID=0, Status='live', Symbol='LOC100419549', Aliases='', description='PRAME family member 17 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65733094, end_position_on_the_genomic_accession=65733767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419552, CurrentID=0, Status='live', Symbol='LOC100419552', Aliases='', description='zinc finger protein 57 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141259235, end_position_on_the_genomic_accession=141260301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149473, CurrentID=0, Status='live', Symbol='CCDC24', Aliases='', description='coiled-coil domain containing 24', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43991608, end_position_on_the_genomic_accession=43996528, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100419554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419554, CurrentID=0, Status='live', Symbol='ZNF92P1Y', Aliases='', description='zinc finger protein 92 pseudogene 1, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7853710, end_position_on_the_genomic_accession=7854843, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288483, CurrentID=0, Status='live', Symbol='HSBP1P2', Aliases='', description='heat shock factor binding protein 1 pseudogene 2', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110251622, end_position_on_the_genomic_accession=110252126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288484, CurrentID=0, Status='live', Symbol='LOC100288484', Aliases='', description='formyl peptide receptor 1 pseudogene', other_designations='formyl peptide receptor pseudogene', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147851656, end_position_on_the_genomic_accession=147852549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288485, CurrentID=0, Status='live', Symbol='MTRNR2L7', Aliases='HN7', description='MT-RNR2-like 7', other_designations='MTRNR2-like 7|humanin-like protein 7', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37601438, end_position_on_the_genomic_accession=37602931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149478, CurrentID=0, Status='live', Symbol='BTBD19', Aliases='', description='BTB (POZ) domain containing 19', other_designations='novel BTB domain containing protein', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44807457, end_position_on_the_genomic_accession=44814199, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100288560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288560, CurrentID=0, Status='live', Symbol='LOC100288560', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53825867, end_position_on_the_genomic_accession=53826349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419560, CurrentID=0, Status='live', Symbol='KARSP2', Aliases='', description='lysyl-tRNA synthetase pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113130422, end_position_on_the_genomic_accession=113132040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271102, CurrentID=0, Status='live', Symbol='RPS25P5', Aliases='RPS25_3_418', description='ribosomal protein S25 pseudogene 5', other_designations='', map_location='3q21-q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149357122, end_position_on_the_genomic_accession=149357604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419562, CurrentID=0, Status='live', Symbol='LOC100419562', Aliases='', description='ribosomal protein L13 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37165782, end_position_on_the_genomic_accession=37166319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149483, CurrentID=0, Status='live', Symbol='CCDC17', Aliases='', description='coiled-coil domain containing 17', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45620044, end_position_on_the_genomic_accession=45624787, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100288493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288493, CurrentID=0, Status='live', Symbol='TCEB1P5', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 5', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21631679, end_position_on_the_genomic_accession=21632015, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873406, CurrentID=0, Status='live', Symbol='RNA5SP138', Aliases='RN5S138', description='RNA, 5S ribosomal pseudogene 138', other_designations='RNA, 5S ribosomal 138', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126564565, end_position_on_the_genomic_accession=126564678, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419570, CurrentID=0, Status='live', Symbol='LOC100419570', Aliases='', description='ribosomal protein L29 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51734540, end_position_on_the_genomic_accession=51735064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288499, CurrentID=0, Status='live', Symbol='DYZ1L22', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58894474, end_position_on_the_genomic_accession=58896541, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288500, CurrentID=0, Status='live', Symbol='REXO1L10P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 10, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 10, pseudogene|REX1, RNA exonuclease 1 homolog-like 1 pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85746322, end_position_on_the_genomic_accession=85746516, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419574, CurrentID=0, Status='live', Symbol='LOC100419574', Aliases='', description='HECT and RLD domain containing E3 ubiquitin protein ligase 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28679404, end_position_on_the_genomic_accession=28682251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288503, CurrentID=0, Status='live', Symbol='SMIM12P1', Aliases='', description='small integral membrane protein 12 pseudogene 1', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108423587, end_position_on_the_genomic_accession=108424327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288504, CurrentID=0, Status='live', Symbol='LOC100288504', Aliases='', description='uncharacterized LOC100288504', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288506, CurrentID=0, Status='live', Symbol='PNPT1P1', Aliases='', description='polyribonucleotide nucleotidyltransferase 1 pseudogene 1', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3982723, end_position_on_the_genomic_accession=3984295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149499, CurrentID=0, Status='live', Symbol='LRRC71', Aliases='C1orf92', description='leucine rich repeat containing 71', other_designations='leucine-rich repeat-containing protein C10orf92', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156920608, end_position_on_the_genomic_accession=156942683, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  149501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149501, CurrentID=0, Status='live', Symbol='KRT8P45', Aliases='', description='keratin 8 pseudogene 45', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157073181, end_position_on_the_genomic_accession=157074898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288510, CurrentID=0, Status='live', Symbol='LOC100288510', Aliases='', description='TSR1, 20S rRNA accumulation, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57317987, end_position_on_the_genomic_accession=57320020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419583, CurrentID=0, Status='live', Symbol='LOC100419583', Aliases='', description='ring finger protein 4 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44728988, end_position_on_the_genomic_accession=44732851, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288513, CurrentID=0, Status='live', Symbol='TUBB3P1', Aliases='', description='tubulin beta 3 class III pseudogene 1', other_designations='tubulin, beta 3 class III pseudogene 1|tubulin, beta 3 pseudogene', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89301726, end_position_on_the_genomic_accession=89303308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288517, CurrentID=0, Status='live', Symbol='ISCA1P2', Aliases='', description='iron-sulfur cluster assembly 1 pseudogene 2', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229042092, end_position_on_the_genomic_accession=229042589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134145, CurrentID=0, Status='live', Symbol='FAM173B', Aliases='JS-2', description='family with sequence similarity 173 member B', other_designations='family with sequence similarity 173, member B', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10225507, end_position_on_the_genomic_accession=10250036, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100288520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288520, CurrentID=0, Status='live', Symbol='USP17L15', Aliases='', description='ubiquitin specific peptidase 17-like family member 15', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9234385, end_position_on_the_genomic_accession=9236334, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928675, CurrentID=0, Status='live', Symbol='LOC101928675', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51386082, end_position_on_the_genomic_accession=51387506, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419595, CurrentID=0, Status='live', Symbol='HDAC1P1', Aliases='', description='histone deacetylase 1 pseudogene 1', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94137750, end_position_on_the_genomic_accession=94139001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288524, CurrentID=0, Status='live', Symbol='WI2-2373I1.2', Aliases='', description='forkhead box L1-like', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=290170, end_position_on_the_genomic_accession=294422, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100873407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873407, CurrentID=0, Status='live', Symbol='RNA5SP139', Aliases='RN5S139', description='RNA, 5S ribosomal pseudogene 139', other_designations='RNA, 5S ribosomal 139', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127963608, end_position_on_the_genomic_accession=127963724, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288526, CurrentID=0, Status='live', Symbol='DYZ1L24', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58901617, end_position_on_the_genomic_accession=58903679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288527, CurrentID=0, Status='live', Symbol='REXO1L2P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 2, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 2, pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85775016, end_position_on_the_genomic_accession=85776093, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929621, CurrentID=0, Status='live', Symbol='LOC101929621', Aliases='', description='uncharacterized LOC101929621', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108557233, end_position_on_the_genomic_accession=108560392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  134147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134147, CurrentID=0, Status='live', Symbol='CMBL', Aliases='JS-1', description='carboxymethylenebutenolidase homolog (Pseudomonas)', other_designations='carboxymethylenebutenolidase-like (Pseudomonas)', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10277595, end_position_on_the_genomic_accession=10308056, orientation='minus', exon_count=6, OMIM=613379, no_hdr0=''),
  100288533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288533, CurrentID=0, Status='live', Symbol='LOC100288533', Aliases='', description='ribosomal protein L14 pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128203361, end_position_on_the_genomic_accession=128203731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419606, CurrentID=0, Status='live', Symbol='ZNF90P2', Aliases='ZNF463P, dJ111M5.3', description='zinc finger protein 90 pseudogene 2', other_designations='zinc finger protein 463, pseudogene|zinc finger protein 93 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28888981, end_position_on_the_genomic_accession=28889634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419608, CurrentID=0, Status='live', Symbol='ZNF602P', Aliases='dJ265C24.5', description='zinc finger protein 602, pseudogene', other_designations='zinc finger protein 34 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28115649, end_position_on_the_genomic_accession=28116220, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419609, CurrentID=0, Status='live', Symbol='ZNF70P1', Aliases='ZNF314P, bCX269C15.3, dJ1033B10.7', description='zinc finger protein 70 pseudogene 1', other_designations='zinc finger protein 314, pseudogene', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33215574, end_position_on_the_genomic_accession=33216183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419610, CurrentID=0, Status='live', Symbol='LOC100419610', Aliases='', description='thymidylate synthetase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96653352, end_position_on_the_genomic_accession=96654126, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288540, CurrentID=0, Status='live', Symbol='POM121B', Aliases='', description='POM121 transmembrane nucleoporin B (pseudogene)', other_designations='POM121 membrane glycoprotein B (pseudogene)|Putative nuclear envelope pore membrane protein POM 121B', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73293495, end_position_on_the_genomic_accession=73303235, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419613, CurrentID=0, Status='live', Symbol='LOC100419613', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72660230, end_position_on_the_genomic_accession=72661832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419614, CurrentID=0, Status='live', Symbol='MIOXP1', Aliases='', description='myo-inositol oxygenase pseudogene 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86593566, end_position_on_the_genomic_accession=86594624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288543, CurrentID=0, Status='live', Symbol='FOLR3P1', Aliases='', description='folate receptor 3 (gamma) pseudogene 1', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72172440, end_position_on_the_genomic_accession=72178042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419616, CurrentID=0, Status='live', Symbol='LOC100419616', Aliases='', description='impact RWD domain protein pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56776307, end_position_on_the_genomic_accession=56776854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419617, CurrentID=0, Status='live', Symbol='LOC100419617', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133946560, end_position_on_the_genomic_accession=133947423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419620, CurrentID=0, Status='live', Symbol='LOC100419620', Aliases='', description='RAN binding protein 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18908173, end_position_on_the_genomic_accession=18908999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419621, CurrentID=0, Status='live', Symbol='LOC100419621', Aliases='', description='ribosomal protein L24 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37454156, end_position_on_the_genomic_accession=37454973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271104, CurrentID=0, Status='live', Symbol='RPS26P12', Aliases='RPS26_2_128', description='ribosomal protein S26 pseudogene 12', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174922101, end_position_on_the_genomic_accession=174922434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419624, CurrentID=0, Status='live', Symbol='LOC100419624', Aliases='', description='RAN binding protein 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78522247, end_position_on_the_genomic_accession=78522828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288558, CurrentID=0, Status='live', Symbol='LOC100288558', Aliases='', description='chromosome 4 open reading frame 46 pseudogene', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143386511, end_position_on_the_genomic_accession=143389538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  542767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=542767, CurrentID=0, Status='live', Symbol='C1QTNF9B-AS1', Aliases='PCOTH', description='C1QTNF9B antisense RNA 1', other_designations='C1QTNF9B antisense RNA 1 (non-protein coding)|C1QTNF9B antisense gene protein 1|protein PCOTH', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23888889, end_position_on_the_genomic_accession=23892103, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100288712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288712, CurrentID=0, Status='live', Symbol='LOC100288712', Aliases='', description='S100 calcium binding protein A11 pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5560717, end_position_on_the_genomic_accession=5561219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288561, CurrentID=0, Status='live', Symbol='DYZ1L25', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58907517, end_position_on_the_genomic_accession=58910716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288562, CurrentID=0, Status='live', Symbol='LOC100288562', Aliases='', description='putative exonuclease GOR', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85790325, end_position_on_the_genomic_accession=85792352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288563, CurrentID=0, Status='live', Symbol='HMGB3P23', Aliases='', description='high mobility group box 3 pseudogene 23', other_designations='high-mobility group box 3 pseudogene 23', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31643345, end_position_on_the_genomic_accession=31645184, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419639, CurrentID=0, Status='live', Symbol='LOC100419639', Aliases='', description='ribosomal protein L21 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82434789, end_position_on_the_genomic_accession=82435413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419641, CurrentID=0, Status='live', Symbol='LOC100419641', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12439419, end_position_on_the_genomic_accession=12440410, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419642, CurrentID=0, Status='live', Symbol='LOC100419642', Aliases='', description='zinc finger protein 85 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56525428, end_position_on_the_genomic_accession=56526113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149563, CurrentID=0, Status='live', Symbol='C1orf64', Aliases='ERRF', description='chromosome 1 open reading frame 64', other_designations='ER-related factor', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16004236, end_position_on_the_genomic_accession=16006695, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  280636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280636, CurrentID=0, Status='live', Symbol='C11orf31', Aliases='C17orf10, SELH', description='chromosome 11 open reading frame 31', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57741250, end_position_on_the_genomic_accession=57743411, orientation='plus', exon_count=4, OMIM=607914, no_hdr0=''),
  280637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280637, CurrentID=0, Status='live', Symbol='PSMD7P1', Aliases='MOV34L, PSMD7L', description='proteasome 26S subunit, non-ATPase 7 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 like|proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 pseudogene 1|proteasome 26S non-ATPase subunit 7 pseudogene', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66197589, end_position_on_the_genomic_accession=66198707, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419649, CurrentID=0, Status='live', Symbol='PRAMEF35P', Aliases='', description='PRAME family member 35, pseudogene', other_designations='PRAME family member 10 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13104375, end_position_on_the_genomic_accession=13105689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419650, CurrentID=0, Status='live', Symbol='RPL23AP85', Aliases='', description='ribosomal protein L23a pseudogene 85', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56585615, end_position_on_the_genomic_accession=56586242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  280644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280644, CurrentID=0, Status='live', Symbol='PSMD10P2', Aliases='', description='proteasome 26S subunit, non-ATPase, 10 pseudogene 2', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 10 pseudogene|proteasome 26S non-ATPase subunit 10 pseudogene', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186761111, end_position_on_the_genomic_accession=186763825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419654, CurrentID=0, Status='live', Symbol='LOC100419654', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97855575, end_position_on_the_genomic_accession=97857016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270833, CurrentID=0, Status='live', Symbol='RPL5P12', Aliases='RPL5_4_502', description='ribosomal protein L5 pseudogene 12', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98025340, end_position_on_the_genomic_accession=98026324, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873409, CurrentID=0, Status='live', Symbol='RNA5SP142', Aliases='RN5S142', description='RNA, 5S ribosomal pseudogene 142', other_designations='RNA, 5S ribosomal 142', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137518134, end_position_on_the_genomic_accession=137518255, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288590, CurrentID=0, Status='live', Symbol='LOC100288590', Aliases='', description='cutaneous T-cell lymphoma-associated antigen 1 pseudogene', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38204544, end_position_on_the_genomic_accession=38205606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  280655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280655, CurrentID=0, Status='live', Symbol='IGBP1P1', Aliases='C14orf19', description='immunoglobulin (CD79A) binding protein 1 pseudogene 1', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34939922, end_position_on_the_genomic_accession=34940496, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  280657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280657, CurrentID=0, Status='live', Symbol='SSX6', Aliases='SSXP2, dJ54B20.1, psiSSX2', description='synovial sarcoma, X breakpoint 6 (pseudogene)', other_designations='SSX family pseudogene 2|SSX2 pseudogene|synovial sarcoma X breakpoint 6 protein', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48107982, end_position_on_the_genomic_accession=48120686, orientation='plus', exon_count=8, OMIM=300541, no_hdr0=''),
  280658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280658, CurrentID=0, Status='live', Symbol='SSX7', Aliases='', description='synovial sarcoma, X breakpoint 7', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52644061, end_position_on_the_genomic_accession=52654900, orientation='minus', exon_count=8, OMIM=300542, no_hdr0=''),
  280659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280659, CurrentID=0, Status='live', Symbol='SSX8', Aliases='', description='synovial sarcoma, X breakpoint 8', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52622935, end_position_on_the_genomic_accession=52633948, orientation='plus', exon_count=8, OMIM=300543, no_hdr0=''),
  280660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280660, CurrentID=0, Status='live', Symbol='SSX9', Aliases='', description='synovial sarcoma, X breakpoint 9', other_designations='synovial sarcoma, X breakpoint 7, pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48301550, end_position_on_the_genomic_accession=48306179, orientation='minus', exon_count=5, OMIM=300544, no_hdr0=''),
  280661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280661, CurrentID=0, Status='live', Symbol='EIF1AXP1', Aliases='EIF1AP1', description='eukaryotic translation initiation factor 1A, X-linked pseudogene 1', other_designations='eukaryotic translation initiation factor 1A pseudogene 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16683091, end_position_on_the_genomic_accession=16686249, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  280664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280664, CurrentID=0, Status='live', Symbol='WFDC10B', Aliases='WAP12', description='WAP four-disulfide core domain 10B', other_designations='protease inhibitor WAP12', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45684651, end_position_on_the_genomic_accession=45705019, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  280665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=280665, CurrentID=0, Status='live', Symbol='LOC280665', Aliases='', description='anti-CNG alpha 1 cation channel translation product-like', other_designations='putative anti-CNG alpha 1 cation channel translation product', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373230, CurrentID=0, Status='live', Symbol='LOC105373230', Aliases='', description='uncharacterized LOC105373230', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241424221, end_position_on_the_genomic_accession=241433741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  439994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439994, CurrentID=0, Status='live', Symbol='LOC439994', Aliases='', description='uncharacterized LOC439994', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87342411, end_position_on_the_genomic_accession=87343574, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419678, CurrentID=0, Status='live', Symbol='LOC100419678', Aliases='', description='zinc finger, MYM-type 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72901977, end_position_on_the_genomic_accession=72902931, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419679, CurrentID=0, Status='live', Symbol='LOC100419679', Aliases='', description='zinc finger protein 670 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203061321, end_position_on_the_genomic_accession=203062403, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419680, CurrentID=0, Status='live', Symbol='LOC100419680', Aliases='', description='zinc finger protein 430 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38015583, end_position_on_the_genomic_accession=38016636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288610, CurrentID=0, Status='live', Symbol='LINC00268-2P', Aliases='NCRNA00268-2P, NCRNA00268B', description='long intergenic non-protein coding RNA 268-2, pseudogene', other_designations='non-protein coding RNA 268-2, pseudogene|non-protein coding RNA 268B', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11170731, end_position_on_the_genomic_accession=11172105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149603, CurrentID=0, Status='live', Symbol='RNF187', Aliases='RACO-1, RACO1', description='ring finger protein 187', other_designations='RING domain AP1 coactivator 1|protein RNF187', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228487367, end_position_on_the_genomic_accession=228496188, orientation='plus', exon_count=4, OMIM=613754, no_hdr0=''),
  100419684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419684, CurrentID=0, Status='live', Symbol='LOC100419684', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94750579, end_position_on_the_genomic_accession=94751295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288613, CurrentID=0, Status='live', Symbol='UBE2NP1', Aliases='', description='ubiquitin conjugating enzyme E2N pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2N pseudogene 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21611836, end_position_on_the_genomic_accession=21612293, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419686, CurrentID=0, Status='live', Symbol='LOC100419686', Aliases='', description='zinc finger protein 256 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238454522, end_position_on_the_genomic_accession=238455196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288615, CurrentID=0, Status='live', Symbol='WHAMMP1', Aliases='WHDC1P1', description='WAS protein homolog associated with actin, golgi membranes and microtubules pseudogene 1', other_designations='WAS protein homology region 2 domain containing 1 pseudogene 1', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32519848, end_position_on_the_genomic_accession=32533741, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100873410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873410, CurrentID=0, Status='live', Symbol='RNA5SP143', Aliases='RN5S143', description='RNA, 5S ribosomal pseudogene 143', other_designations='RNA, 5S ribosomal 143', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142591677, end_position_on_the_genomic_accession=142591791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419690, CurrentID=0, Status='live', Symbol='LOC100419690', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40332986, end_position_on_the_genomic_accession=40333694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419691, CurrentID=0, Status='live', Symbol='LOC100419691', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64473037, end_position_on_the_genomic_accession=64473746, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419692, CurrentID=0, Status='live', Symbol='LOC100419692', Aliases='', description='zinc finger protein 791 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23955760, end_position_on_the_genomic_accession=23956444, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419693, CurrentID=0, Status='live', Symbol='LOC100419693', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42959278, end_position_on_the_genomic_accession=42959988, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157740, CurrentID=0, Status='live', Symbol='LOC157740', Aliases='', description='uncharacterized protein C8orf9', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=11178400, end_position_on_the_genomic_accession=11179989, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149620, CurrentID=0, Status='live', Symbol='CHIAP2', Aliases='', description='chitinase, acidic pseudogene 2', other_designations='CHIA-like pseudogene|novel protein containing a Glycosyl hydrolases family 18 domain', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111280524, end_position_on_the_genomic_accession=111286108, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100419701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419701, CurrentID=0, Status='live', Symbol='LOC100419701', Aliases='', description='CTAGE family, member 9 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125900318, end_position_on_the_genomic_accession=125901485, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419702, CurrentID=0, Status='live', Symbol='LOC100419702', Aliases='', description='necdin, MAGE family member pseudogene', other_designations='necdin, melanoma antigen (MAGE) family member pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75305452, end_position_on_the_genomic_accession=75306359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419704, CurrentID=0, Status='live', Symbol='LOC100419704', Aliases='', description='zinc finger protein 57 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2896596, end_position_on_the_genomic_accession=2897226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419705, CurrentID=0, Status='live', Symbol='ZNF92P2', Aliases='', description='zinc finger protein 92 pseudogene 2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22129802, end_position_on_the_genomic_accession=22130320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419706, CurrentID=0, Status='live', Symbol='LOC100419706', Aliases='', description='zinc finger protein 84 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51861659, end_position_on_the_genomic_accession=51863403, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419707, CurrentID=0, Status='live', Symbol='LOC100419707', Aliases='', description='zinc finger protein 41 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51914667, end_position_on_the_genomic_accession=51916076, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149628, CurrentID=0, Status='live', Symbol='PYHIN1', Aliases='IFIX', description='pyrin and HIN domain family member 1', other_designations='interferon-inducible protein X|pyrin and HIN domain family, member 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158931547, end_position_on_the_genomic_accession=158988872, orientation='plus', exon_count=12, OMIM=612677, no_hdr0=''),
  100419709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419709, CurrentID=0, Status='live', Symbol='ZNF887P', Aliases='ZNF887', description='zinc finger protein 887, pseudogene', other_designations='zinc finger protein 20 pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11648348, end_position_on_the_genomic_accession=11652160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419711, CurrentID=0, Status='live', Symbol='ZNF92P3', Aliases='', description='zinc finger protein 92 pseudogene 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23934379, end_position_on_the_genomic_accession=23935836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271107, CurrentID=0, Status='live', Symbol='RPS26P23', Aliases='RPS26_4_515', description='ribosomal protein S26 pseudogene 23', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122840349, end_position_on_the_genomic_accession=122840690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419713, CurrentID=0, Status='live', Symbol='LOC100419713', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73640482, end_position_on_the_genomic_accession=73641037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419714, CurrentID=0, Status='live', Symbol='AHCYP6', Aliases='', description='adenosylhomocysteinase pseudogene 6', other_designations='', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85857001, end_position_on_the_genomic_accession=85859157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419715, CurrentID=0, Status='live', Symbol='LOC100419715', Aliases='', description='zinc finger protein 626 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41920723, end_position_on_the_genomic_accession=41921227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419716, CurrentID=0, Status='live', Symbol='LOC100419716', Aliases='', description='zinc finger protein 709 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163855131, end_position_on_the_genomic_accession=163856604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419717, CurrentID=0, Status='live', Symbol='SALL4P1', Aliases='', description='spalt-like transcription factor 4 pseudogene 1', other_designations='sal-like 4 pseudogene 1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57422466, end_position_on_the_genomic_accession=57424265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929625, CurrentID=0, Status='live', Symbol='LOC101929625', Aliases='', description='uncharacterized LOC101929625', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21570024, end_position_on_the_genomic_accession=21615722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100419721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419721, CurrentID=0, Status='live', Symbol='LOC100419721', Aliases='', description='zinc finger protein 670 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180257680, end_position_on_the_genomic_accession=180258711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419722, CurrentID=0, Status='live', Symbol='RNF19BPY', Aliases='', description='ring finger protein 19B pseudogene, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3778537, end_position_on_the_genomic_accession=3779542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149643, CurrentID=0, Status='live', Symbol='SPATA45', Aliases='C1orf227', description='spermatogenesis associated 45', other_designations='UPF0732 protein C1orf227', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212830141, end_position_on_the_genomic_accession=212847649, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419724, CurrentID=0, Status='live', Symbol='ZNF736P8Y', Aliases='', description='zinc finger protein 736 pseudogene 8, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7913234, end_position_on_the_genomic_accession=7914140, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419725, CurrentID=0, Status='live', Symbol='ZNF736P7Y', Aliases='', description='zinc finger protein 736 pseudogene 7, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7990743, end_position_on_the_genomic_accession=7991930, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419726, CurrentID=0, Status='live', Symbol='ZNF736P9Y', Aliases='', description='zinc finger protein 736 pseudogene 9, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8068896, end_position_on_the_genomic_accession=8070120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149647, CurrentID=0, Status='live', Symbol='FAM71A', Aliases='', description='family with sequence similarity 71 member A', other_designations='family with sequence similarity 71, member A', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212624447, end_position_on_the_genomic_accession=212626778, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419728, CurrentID=0, Status='live', Symbol='ZNF886P', Aliases='', description='zinc finger protein 886, pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20001188, end_position_on_the_genomic_accession=20001711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419729, CurrentID=0, Status='live', Symbol='ZNF884P', Aliases='', description='zinc finger protein 884, pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20357614, end_position_on_the_genomic_accession=20358416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149650, CurrentID=0, Status='live', Symbol='FLJ32154', Aliases='', description='uncharacterized protein FLJ32154', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419731, CurrentID=0, Status='live', Symbol='ZNF736P12Y', Aliases='', description='zinc finger protein 736 pseudogene 12, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24500890, end_position_on_the_genomic_accession=24501503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419732, CurrentID=0, Status='live', Symbol='ZNF736P6Y', Aliases='', description='zinc finger protein 736 pseudogene 6, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8350917, end_position_on_the_genomic_accession=8352143, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419733, CurrentID=0, Status='live', Symbol='ZNF839P1', Aliases='', description='zinc finger protein 839 pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18985110, end_position_on_the_genomic_accession=18986385, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419734, CurrentID=0, Status='live', Symbol='ZNF885P', Aliases='', description='zinc finger protein 885, pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20396168, end_position_on_the_genomic_accession=20396796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419735, CurrentID=0, Status='live', Symbol='ZNF736P4Y', Aliases='', description='zinc finger protein 736 pseudogene 4, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25169212, end_position_on_the_genomic_accession=25170193, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373345, CurrentID=0, Status='live', Symbol='LOC105373345', Aliases='', description='uncharacterized LOC105373345', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142546405, end_position_on_the_genomic_accession=142624656, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288666, CurrentID=0, Status='live', Symbol='LOC100288666', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42725547, end_position_on_the_genomic_accession=42726228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288667, CurrentID=0, Status='live', Symbol='CICP7', Aliases='', description='capicua transcriptional repressor pseudogene 7', other_designations='capicua homolog pseudogene 7', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=487055, end_position_on_the_genomic_accession=490819, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419741, CurrentID=0, Status='live', Symbol='LOC100419741', Aliases='', description='zinc finger protein 772 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174539213, end_position_on_the_genomic_accession=174540183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271108, CurrentID=0, Status='live', Symbol='RPS26P18', Aliases='RPS26_5_195', description='ribosomal protein S26 pseudogene 18', other_designations='', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15397369, end_position_on_the_genomic_accession=15397808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419743, CurrentID=0, Status='live', Symbol='DBET', Aliases='DBE-T, DUX4L30', description='D4Z4 binding element transcript (non-protein coding)', other_designations='chromatin-associated long non-coding RNA|double homeobox 4 like 4 pseudogene', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190064502, end_position_on_the_genomic_accession=190067864, orientation='plus', exon_count=1, OMIM=614865, no_hdr0=''),
  100288675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288675, CurrentID=0, Status='live', Symbol='LOC100288675', Aliases='', description='uncharacterized LOC100288675', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=131334297, end_position_on_the_genomic_accession=131339909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419748, CurrentID=0, Status='live', Symbol='LOC100419748', Aliases='', description='zinc finger protein 852 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44528746, end_position_on_the_genomic_accession=44529501, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419750, CurrentID=0, Status='live', Symbol='LOC100419750', Aliases='', description='DTW domain containing 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155402415, end_position_on_the_genomic_accession=155402985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419751, CurrentID=0, Status='live', Symbol='ZNF587P1', Aliases='', description='zinc finger protein 587 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31794479, end_position_on_the_genomic_accession=31796172, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419752, CurrentID=0, Status='live', Symbol='LOC100419752', Aliases='', description='zinc finger protein 251 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44675722, end_position_on_the_genomic_accession=44676709, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419755, CurrentID=0, Status='live', Symbol='LOC100419755', Aliases='', description='zinc finger protein 426 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121099108, end_position_on_the_genomic_accession=121099907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419756, CurrentID=0, Status='live', Symbol='ZNF968P', Aliases='', description='zinc finger protein 968, pseudogene', other_designations='zinc finger protein 182 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28408225, end_position_on_the_genomic_accession=28409250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419758, CurrentID=0, Status='live', Symbol='PKMP3', Aliases='', description='pyruvate kinase, muscle pseudogene 3', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85659866, end_position_on_the_genomic_accession=85660633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288687, CurrentID=0, Status='live', Symbol='DUX4', Aliases='DUX4L', description='double homeobox 4', other_designations='double homeobox 4-like|double homeobox protein 10|double homeobox protein 4/10', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190173669, end_position_on_the_genomic_accession=190175845, orientation='plus', exon_count=2, OMIM=606009, no_hdr0=''),
  100419761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419761, CurrentID=0, Status='live', Symbol='LOC100419761', Aliases='', description='zinc finger protein 654 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13741810, end_position_on_the_genomic_accession=13742780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288690, CurrentID=0, Status='live', Symbol='FAM21FP', Aliases='', description='family with sequence similarity 21 member F, pseudogene', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45705254, end_position_on_the_genomic_accession=45727163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288691, CurrentID=0, Status='live', Symbol='HMGN2P8', Aliases='HMGN2L8', description='high mobility group nucleosomal binding domain 2 pseudogene 8', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 8|high-mobility group nucleosomal binding domain 2-like 8', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84081331, end_position_on_the_genomic_accession=84082811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419764, CurrentID=0, Status='live', Symbol='GARSP1', Aliases='', description='glycyl-tRNA synthetase pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55550954, end_position_on_the_genomic_accession=55551650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149685, CurrentID=0, Status='live', Symbol='ADIG', Aliases='SMAF1', description='adipogenin', other_designations='adipogenesis associated|small adipocyte factor 1 (SMAF1)', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38581195, end_position_on_the_genomic_accession=38588463, orientation='plus', exon_count=3, OMIM=611396, no_hdr0=''),
  100288695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288695, CurrentID=0, Status='live', Symbol='LIMS3L', Aliases='', description='LIM and senescent cell antigen-like domains 3-like', other_designations='particularly interesting new Cys-His protein 3', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110445883, end_position_on_the_genomic_accession=110473075, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100419768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419768, CurrentID=0, Status='live', Symbol='LOC100419768', Aliases='', description='ribosomal protein S15a pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56951648, end_position_on_the_genomic_accession=56952443, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271109, CurrentID=0, Status='live', Symbol='RPS26P26', Aliases='RPS26_5_578', description='ribosomal protein S26 pseudogene 26', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75374437, end_position_on_the_genomic_accession=75374878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419773, CurrentID=0, Status='live', Symbol='LOC100419773', Aliases='', description='zinc finger protein 33B pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6673225, end_position_on_the_genomic_accession=6675293, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288702, CurrentID=0, Status='live', Symbol='ARL6IP1P1', Aliases='ARL6IPP1', description='ADP ribosylation factor like GTPase 6 interacting protein 1 pseudogene 1', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 1 pseudogene 1|ADP-ribosylation factor like GTPase 6 interacting protein 1 pseudogene 1|ADP-ribosylation factor-like 6 interacting protein 1 pseudogene 1|ADP-ribosylation factor-like 6 interacting protein pseudogene 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21647717, end_position_on_the_genomic_accession=21650405, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419775, CurrentID=0, Status='live', Symbol='LOC100419775', Aliases='', description='zinc finger protein 519 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45303620, end_position_on_the_genomic_accession=45304362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419776, CurrentID=0, Status='live', Symbol='ZNF619P1', Aliases='', description='zinc finger protein 619 pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46144574, end_position_on_the_genomic_accession=46145768, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419777, CurrentID=0, Status='live', Symbol='RNF138P2', Aliases='', description='ring finger protein 138, E3 ubiquitin protein ligase pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53316205, end_position_on_the_genomic_accession=53316719, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149699, CurrentID=0, Status='live', Symbol='GTSF1L', Aliases='C20orf65, FAM112A, dJ1028D15.4', description='gametocyte specific factor 1-like', other_designations='family with sequence similarity 112, member A', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43726161, end_position_on_the_genomic_accession=43727012, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419780, CurrentID=0, Status='live', Symbol='LOC100419780', Aliases='', description='zinc finger protein 479 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64005613, end_position_on_the_genomic_accession=64016139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419781, CurrentID=0, Status='live', Symbol='LOC100419781', Aliases='', description='zinc finger protein 267 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65291623, end_position_on_the_genomic_accession=65292776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419782, CurrentID=0, Status='live', Symbol='LOC100419782', Aliases='', description='zinc finger protein 717 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110215899, end_position_on_the_genomic_accession=110216550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288711, CurrentID=0, Status='live', Symbol='DUX4L9', Aliases='DUX4C', description='double homeobox 4 like 9', other_designations='double homeobox 4 like 9 (pseudogene)|double homeobox 4 pseudogene|double homeobox 4c', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190021169, end_position_on_the_genomic_accession=190022665, orientation='minus', exon_count=0, OMIM=615581, no_hdr0=''),
  100419784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419784, CurrentID=0, Status='live', Symbol='LOC100419784', Aliases='', description='zinc finger protein 484 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47776872, end_position_on_the_genomic_accession=47778657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288713, CurrentID=0, Status='live', Symbol='PMS2P10', Aliases='PMS2L10, PMSR4', description='PMS1 homolog 2, mismatch repair system component pseudogene 10', other_designations='postmeiotic segregation increased 2 pseudogene 10|postmeiotic segregation increased 2-like 10', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75323753, end_position_on_the_genomic_accession=75327765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419786, CurrentID=0, Status='live', Symbol='LOC100419786', Aliases='', description='zinc finger protein 670 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64144909, end_position_on_the_genomic_accession=64147052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149708, CurrentID=0, Status='live', Symbol='WFDC5', Aliases='PRG5, WAP1, dJ211D12.5', description='WAP four-disulfide core domain 5', other_designations='p53-responsive gene 5 protein|protease inhibitor WAP1|putative protease inhibitor WAP1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45109452, end_position_on_the_genomic_accession=45116321, orientation='minus', exon_count=6, OMIM=605161, no_hdr0=''),
  100419790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419790, CurrentID=0, Status='live', Symbol='LOC100419790', Aliases='', description='zinc finger protein 670 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115063933, end_position_on_the_genomic_accession=115064591, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440013, CurrentID=0, Status='live', Symbol='DUX4L10', Aliases='DUX10', description='double homeobox 4 like 10', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133743330, end_position_on_the_genomic_accession=133744608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419792, CurrentID=0, Status='live', Symbol='LOC100419792', Aliases='', description='zinc finger protein 622 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154919569, end_position_on_the_genomic_accession=154920194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419793, CurrentID=0, Status='live', Symbol='BEND7P1', Aliases='', description='BEN domain containing 7 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102705778, end_position_on_the_genomic_accession=102706366, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419794, CurrentID=0, Status='live', Symbol='LOC100419794', Aliases='', description='zinc finger protein 449 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135236790, end_position_on_the_genomic_accession=135237348, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373234, CurrentID=0, Status='live', Symbol='LOC105373234', Aliases='', description='uncharacterized LOC105373234', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242489553, end_position_on_the_genomic_accession=242500055, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419796, CurrentID=0, Status='live', Symbol='LOC100419796', Aliases='', description='zinc finger protein 146 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44150801, end_position_on_the_genomic_accession=44151645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288726, CurrentID=0, Status='live', Symbol='HSPE1P20', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 20', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 20', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126912036, end_position_on_the_genomic_accession=126912323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419799, CurrentID=0, Status='live', Symbol='LOC100419799', Aliases='', description='gon-4-like (C. elegans) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155649700, end_position_on_the_genomic_accession=155650267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288728, CurrentID=0, Status='live', Symbol='LOC100288728', Aliases='', description='uncharacterized LOC100288728', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=3038263, end_position_on_the_genomic_accession=3080223, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419801, CurrentID=0, Status='live', Symbol='LOC100419801', Aliases='', description='zinc finger protein 532 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242376925, end_position_on_the_genomic_accession=242377470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288730, CurrentID=0, Status='live', Symbol='PAN3-AS1', Aliases='', description='PAN3 antisense RNA 1', other_designations='PAN3 antisense RNA 1 (non-protein coding)', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28136843, end_position_on_the_genomic_accession=28139174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419803, CurrentID=0, Status='live', Symbol='LOC100419803', Aliases='', description='zinc finger protein 510 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92203354, end_position_on_the_genomic_accession=92203978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419806, CurrentID=0, Status='live', Symbol='LOC100419806', Aliases='', description='zinc finger protein 519 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247015190, end_position_on_the_genomic_accession=247017233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129321, CurrentID=0, Status='live', Symbol='LOC100129321', Aliases='', description='single stranded DNA binding protein 3 pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39833571, end_position_on_the_genomic_accession=39834715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288739, CurrentID=0, Status='live', Symbol='SPCS2P3', Aliases='', description='signal peptidase complex subunit 2 homolog (S. cerevisiae) pseudogene 3', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=816346, end_position_on_the_genomic_accession=816998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419812, CurrentID=0, Status='live', Symbol='LOC100419812', Aliases='', description='COBW domain containing 7 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194454538, end_position_on_the_genomic_accession=194455127, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288742, CurrentID=0, Status='live', Symbol='HIST1H2BPS1', Aliases='H2BFVP, bA317E16.4', description='histone cluster 1, H2b, pseudogene 1', other_designations='histone 1, H2b, pseudogene 1', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25731728, end_position_on_the_genomic_accession=25732166, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419815, CurrentID=0, Status='live', Symbol='LOC100419815', Aliases='', description='ankyrin repeat domain 57 pseudogene', other_designations='sosondowah ankyrin repeat domain family member C pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94861362, end_position_on_the_genomic_accession=94862661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419816, CurrentID=0, Status='live', Symbol='LOC100419816', Aliases='', description='orthodenticle homeobox 2 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146477355, end_position_on_the_genomic_accession=146478110, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288748, CurrentID=0, Status='live', Symbol='LOC100288748', Aliases='', description='uncharacterized LOC100288748', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94637285, end_position_on_the_genomic_accession=94639467, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419821, CurrentID=0, Status='live', Symbol='CSNK1G2P1', Aliases='', description='casein kinase 1, gamma 2 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5040945, end_position_on_the_genomic_accession=5041940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419824, CurrentID=0, Status='live', Symbol='LOC100419824', Aliases='', description='kinesin family member 27 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85753300, end_position_on_the_genomic_accession=85786577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419825, CurrentID=0, Status='live', Symbol='LOC100419825', Aliases='', description='zinc finger protein 655 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7041131, end_position_on_the_genomic_accession=7041641, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419827, CurrentID=0, Status='live', Symbol='ZKSCAN7P1', Aliases='', description='zinc finger with KRAB and SCAN domains 7 pseudogene 1', other_designations='zinc finger protein 167 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18493075, end_position_on_the_genomic_accession=18493785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419829, CurrentID=0, Status='live', Symbol='RARSP1', Aliases='', description='arginyl-tRNA synthetase pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27496925, end_position_on_the_genomic_accession=27498683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419830, CurrentID=0, Status='live', Symbol='LOC100419830', Aliases='', description='zinc finger protein 337 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53574586, end_position_on_the_genomic_accession=53576106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419831, CurrentID=0, Status='live', Symbol='LOC100419831', Aliases='', description='zinc finger protein 420 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9756152, end_position_on_the_genomic_accession=9756863, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271111, CurrentID=0, Status='live', Symbol='RPS26P30', Aliases='RPS26_7_791', description='ribosomal protein S26 pseudogene 30', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20843630, end_position_on_the_genomic_accession=20843967, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419833, CurrentID=0, Status='live', Symbol='LOC100419833', Aliases='', description='zinc finger protein 709 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27982742, end_position_on_the_genomic_accession=27983338, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419834, CurrentID=0, Status='live', Symbol='LOC100419834', Aliases='', description='zinc finger protein 134 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34801295, end_position_on_the_genomic_accession=34802095, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419835, CurrentID=0, Status='live', Symbol='LOC100419835', Aliases='', description='zinc finger protein 160 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52063189, end_position_on_the_genomic_accession=52064026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419836, CurrentID=0, Status='live', Symbol='LOC100419836', Aliases='', description='zinc finger protein 347 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52473143, end_position_on_the_genomic_accession=52475254, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873415, CurrentID=0, Status='live', Symbol='RNA5SP148', Aliases='RN5S148', description='RNA, 5S ribosomal pseudogene 148', other_designations='RNA, 5S ribosomal 148', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178429197, end_position_on_the_genomic_accession=178429290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419839, CurrentID=0, Status='live', Symbol='LOC100419839', Aliases='', description='zinc finger protein 300 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57111120, end_position_on_the_genomic_accession=57112027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419840, CurrentID=0, Status='live', Symbol='LOC100419840', Aliases='', description='zinc finger protein 446 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58358049, end_position_on_the_genomic_accession=58359239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419841, CurrentID=0, Status='live', Symbol='LOC100419841', Aliases='', description='zinc finger protein 107 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20198433, end_position_on_the_genomic_accession=20199423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419842, CurrentID=0, Status='live', Symbol='LOC100419842', Aliases='', description='zinc finger protein 396 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34825372, end_position_on_the_genomic_accession=34825998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134187, CurrentID=0, Status='live', Symbol='POU5F2', Aliases='SPRM-1', description='POU domain class 5, transcription factor 2', other_designations='sperm 1 POU domain transcription factor', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93740309, end_position_on_the_genomic_accession=93741603, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288772, CurrentID=0, Status='live', Symbol='CCT7P2', Aliases='CCT7-1P', description='chaperonin containing TCP1, subunit 7 (eta) pseudogene 2', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=92889372, end_position_on_the_genomic_accession=92890919, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419845, CurrentID=0, Status='live', Symbol='LOC100419845', Aliases='', description='zinc finger protein 415 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53210980, end_position_on_the_genomic_accession=53212503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288774, CurrentID=0, Status='live', Symbol='AGGF1P10', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 10', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 10', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29427376, end_position_on_the_genomic_accession=29432989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419847, CurrentID=0, Status='live', Symbol='LOC100419847', Aliases='', description='zinc finger protein 436 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58224506, end_position_on_the_genomic_accession=58225357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419848, CurrentID=0, Status='live', Symbol='LOC100419848', Aliases='', description='zinc finger protein 250 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58444928, end_position_on_the_genomic_accession=58445479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288778, CurrentID=0, Status='live', Symbol='LOC100288778', Aliases='', description='WAS protein family homolog 1 pseudogene', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14477, end_position_on_the_genomic_accession=32015, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100419851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419851, CurrentID=0, Status='live', Symbol='LOC100419851', Aliases='', description='protein serine kinase H1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55356307, end_position_on_the_genomic_accession=55357525, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419852, CurrentID=0, Status='live', Symbol='LOC100419852', Aliases='', description='sterol carrier protein 2 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71475758, end_position_on_the_genomic_accession=71476327, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149773, CurrentID=0, Status='live', Symbol='APCDD1L-AS1', Aliases='', description='APCDD1L antisense RNA 1 (head to head)', other_designations='APCDD1L antisense RNA 1 (non-protein coding)', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58515379, end_position_on_the_genomic_accession=58619892, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100419854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419854, CurrentID=0, Status='live', Symbol='ZNF646P1', Aliases='', description='zinc finger protein 646 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53408765, end_position_on_the_genomic_accession=53409876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149775, CurrentID=0, Status='live', Symbol='GNAS-AS1', Aliases='GNAS-AS, GNAS1AS, GNASAS, NCRNA00075, NESP-AS, NESPAS, SANG', description='GNAS antisense RNA 1', other_designations='GNAS antisense RNA 1 (non-protein coding)|GNAS1 antisense', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58818918, end_position_on_the_genomic_accession=58850903, orientation='minus', exon_count=5, OMIM=610540, no_hdr0=''),
  100288785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288785, CurrentID=0, Status='live', Symbol='UQCRBP3', Aliases='', description='ubiquinol-cytochrome c reductase binding protein pseudogene 3', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14874401, end_position_on_the_genomic_accession=14874711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419859, CurrentID=0, Status='live', Symbol='LOC100419859', Aliases='', description='zinc finger, MYM-type 1 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41382359, end_position_on_the_genomic_accession=41383083, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288788, CurrentID=0, Status='live', Symbol='LOC100288788', Aliases='', description='RNA binding motif protein, X-linked-like 3 pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65955623, end_position_on_the_genomic_accession=65957343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419861, CurrentID=0, Status='live', Symbol='LOC100419861', Aliases='', description='metallophosphoesterase 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52440509, end_position_on_the_genomic_accession=52441535, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271112, CurrentID=0, Status='live', Symbol='RPS26P22', Aliases='RPS26_8_399', description='ribosomal protein S26 pseudogene 22', other_designations='', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124733418, end_position_on_the_genomic_accession=124733754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419863, CurrentID=0, Status='live', Symbol='LOC100419863', Aliases='', description='zinc finger protein 705A pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102461250, end_position_on_the_genomic_accession=102462150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419866, CurrentID=0, Status='live', Symbol='ZNF444P1', Aliases='ZNF864P', description='zinc finger protein 444 pseudogene 1', other_designations='zinc finger protein 864, pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58865006, end_position_on_the_genomic_accession=58865672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419867, CurrentID=0, Status='live', Symbol='ZNF101P1', Aliases='', description='zinc finger protein 101 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28339277, end_position_on_the_genomic_accession=28341065, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873416, CurrentID=0, Status='live', Symbol='RNA5SP150', Aliases='RN5S150', description='RNA, 5S ribosomal pseudogene 150', other_designations='RNA, 5S ribosomal 150', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181822872, end_position_on_the_genomic_accession=181822990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288797, CurrentID=0, Status='live', Symbol='TMEM239', Aliases='', description='transmembrane protein 239', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2815874, end_position_on_the_genomic_accession=2818062, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288798, CurrentID=0, Status='live', Symbol='LOC100288798', Aliases='', description='uncharacterized LOC100288798', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46383676, end_position_on_the_genomic_accession=46652579, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100419506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419506, CurrentID=0, Status='live', Symbol='LOC100419506', Aliases='', description='ribosomal protein L4 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43759215, end_position_on_the_genomic_accession=43759922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419872, CurrentID=0, Status='live', Symbol='LOC100419872', Aliases='', description='zinc finger, MYM-type 4 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=55746308, end_position_on_the_genomic_accession=55747810, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288801, CurrentID=0, Status='live', Symbol='FRG2C', Aliases='', description='FSHD region gene 2 family member C', other_designations='FSHD region gene 2 family, member C|FSHD region gene 2 protein family member C|HSA3-FRG2', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75646783, end_position_on_the_genomic_accession=75667217, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100288802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288802, CurrentID=0, Status='live', Symbol='RPL23AP93', Aliases='', description='ribosomal protein L23a pseudogene 93', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194562929, end_position_on_the_genomic_accession=194564011, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419875, CurrentID=0, Status='live', Symbol='FNBP1P2', Aliases='', description='formin binding protein 1 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30808678, end_position_on_the_genomic_accession=30809696, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288805, CurrentID=0, Status='live', Symbol='HYDIN2', Aliases='HYDINP1', description='HYDIN2, axonemal central pair apparatus protein (pseudogene)', other_designations='HYDIN, axonemal central pair apparatus protein pseudogene 1|hydrocephalus inducing homolog 2', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146547489, end_position_on_the_genomic_accession=146822034, orientation='plus', exon_count=46, OMIM=610813, no_hdr0=''),
  100288806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288806, CurrentID=0, Status='live', Symbol='CDC27P8', Aliases='', description='cell division cycle 27 pseudogene 8', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11623826, end_position_on_the_genomic_accession=11625744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419880, CurrentID=0, Status='live', Symbol='LOC100419880', Aliases='', description='kinesin family member 3A pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195843988, end_position_on_the_genomic_accession=195844596, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419882, CurrentID=0, Status='live', Symbol='LOC100419882', Aliases='', description='transmembrane protein 98 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=64733199, end_position_on_the_genomic_accession=64733837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419883, CurrentID=0, Status='live', Symbol='YAP1P3', Aliases='', description='Yes associated protein 1 pseudogene 3', other_designations='Yes-associated protein 1 pseudogene 3', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126627477, end_position_on_the_genomic_accession=126628655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373237, CurrentID=0, Status='live', Symbol='LOC105373237', Aliases='', description='uncharacterized LOC105373237', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64205970, end_position_on_the_genomic_accession=64308111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100288814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288814, CurrentID=0, Status='live', Symbol='CLDN34', Aliases='Claudin-34', description='claudin 34', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9961359, end_position_on_the_genomic_accession=9968373, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100419888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419888, CurrentID=0, Status='live', Symbol='LOC100419888', Aliases='', description='zinc finger protein pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63088064, end_position_on_the_genomic_accession=63088912, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419890, CurrentID=0, Status='live', Symbol='LOC100419890', Aliases='', description='zinc finger, MYM-type 4 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22606440, end_position_on_the_genomic_accession=22607044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419891, CurrentID=0, Status='live', Symbol='LOC100419891', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14346214, end_position_on_the_genomic_accession=14346945, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271113, CurrentID=0, Status='live', Symbol='RPS26P24', Aliases='RPS26_9_482', description='ribosomal protein S26 pseudogene 24', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57352999, end_position_on_the_genomic_accession=57353329, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135924, CurrentID=0, Status='live', Symbol='OR9A2', Aliases='', description='olfactory receptor family 9 subfamily A member 2', other_designations='olfactory receptor OR7-2|olfactory receptor, family 9, subfamily A, member 2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142997851, end_position_on_the_genomic_accession=143027284, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100419894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419894, CurrentID=0, Status='live', Symbol='LOC100419894', Aliases='', description='zinc finger protein 347 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26348335, end_position_on_the_genomic_accession=26349078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419895, CurrentID=0, Status='live', Symbol='ZNF971P', Aliases='', description='zinc finger protein 971, pseudogene', other_designations='zinc finger protein 670 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35447331, end_position_on_the_genomic_accession=35448158, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419896, CurrentID=0, Status='live', Symbol='LOC100419896', Aliases='', description='zinc finger protein 267 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31754227, end_position_on_the_genomic_accession=31755800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288825, CurrentID=0, Status='live', Symbol='TTC39CP1', Aliases='', description='tetratricopeptide repeat domain 39C pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116194031, end_position_on_the_genomic_accession=116194749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288827, CurrentID=0, Status='live', Symbol='CYP4F25P', Aliases='', description='cytochrome P450 family 4 subfamily F member 25, pseudogene', other_designations='CYP4F-se1[6:8]|cytochrome P450, family 4, subfamily F, polypeptide 25, pseudogene', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64465476, end_position_on_the_genomic_accession=64465758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419901, CurrentID=0, Status='live', Symbol='LOC100419901', Aliases='', description='casein kinase 1, alpha 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18578661, end_position_on_the_genomic_accession=18579316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419903, CurrentID=0, Status='live', Symbol='RWDD4P1', Aliases='FAM28BP', description='RWD domain containing 4 pseudogene 1', other_designations='RWD domain containing 4A pseudogene|family with sequence similarity 28, member B pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105301522, end_position_on_the_genomic_accession=105302083, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419906, CurrentID=0, Status='live', Symbol='LOC100419906', Aliases='', description='ankyrin repeat domain 11 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146619590, end_position_on_the_genomic_accession=146620762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419907, CurrentID=0, Status='live', Symbol='LOC100419907', Aliases='', description='kinesin family member 23 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146963938, end_position_on_the_genomic_accession=146964494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419908, CurrentID=0, Status='live', Symbol='LOC100419908', Aliases='', description='zinc finger protein 585A pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47743076, end_position_on_the_genomic_accession=47744166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419909, CurrentID=0, Status='live', Symbol='YAP1P2', Aliases='', description='Yes associated protein 1 pseudogene 2', other_designations='Yes-associated protein 1 pseudogene 2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115069076, end_position_on_the_genomic_accession=115069861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149830, CurrentID=0, Status='live', Symbol='PRNT', Aliases='M8', description='prion protein (testis specific)', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4731282, end_position_on_the_genomic_accession=4740668, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419911, CurrentID=0, Status='live', Symbol='LOC100419911', Aliases='', description='prenylcysteine oxidase 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20038930, end_position_on_the_genomic_accession=20039444, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419912, CurrentID=0, Status='live', Symbol='LOC100419912', Aliases='', description='spalt-like transcription factor 4 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22011910, end_position_on_the_genomic_accession=22014962, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288841, CurrentID=0, Status='live', Symbol='RBPJP5', Aliases='', description='RBPJ pseudogene 5', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39739096, end_position_on_the_genomic_accession=39741367, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288842, CurrentID=0, Status='live', Symbol='LOC100288842', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 5 pseudogene', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120793497, end_position_on_the_genomic_accession=120798731, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288843, CurrentID=0, Status='live', Symbol='NAPGP1', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88699681, end_position_on_the_genomic_accession=88700401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419916, CurrentID=0, Status='live', Symbol='LOC100419916', Aliases='', description='cyclin-dependent kinase 8 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11665326, end_position_on_the_genomic_accession=11665969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149837, CurrentID=0, Status='live', Symbol='LINC00654', Aliases='', description='long intergenic non-protein coding RNA 654', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5498572, end_position_on_the_genomic_accession=5504596, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100288846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288846, CurrentID=0, Status='live', Symbol='LOC100288846', Aliases='', description='uncharacterized LOC100288846', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39265703, end_position_on_the_genomic_accession=39267061, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  149840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149840, CurrentID=0, Status='live', Symbol='C20orf196', Aliases='', description='chromosome 20 open reading frame 196', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5750335, end_position_on_the_genomic_accession=5864407, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100419921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419921, CurrentID=0, Status='live', Symbol='LOC100419921', Aliases='', description='uncharacterized LOC100419921', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131760214, end_position_on_the_genomic_accession=131760767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419923, CurrentID=0, Status='live', Symbol='LOC100419923', Aliases='', description='density-regulated protein pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127300004, end_position_on_the_genomic_accession=127301038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149844, CurrentID=0, Status='live', Symbol='LOC149844', Aliases='', description='synaptotagmin binding, cytoplasmic RNA interacting protein pseudogene', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6213116, end_position_on_the_genomic_accession=6215339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419925, CurrentID=0, Status='live', Symbol='LOC100419925', Aliases='', description='uncharacterized LOC100419925', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63711660, end_position_on_the_genomic_accession=63712163, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419926, CurrentID=0, Status='live', Symbol='LOC100419926', Aliases='', description='uncharacterized LOC100419926', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66261693, end_position_on_the_genomic_accession=66262211, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873418, CurrentID=0, Status='live', Symbol='RNA5SP152', Aliases='RN5S152', description='RNA, 5S ribosomal pseudogene 152', other_designations='RNA, 5S ribosomal 152', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8393015, end_position_on_the_genomic_accession=8393124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419929, CurrentID=0, Status='live', Symbol='LOC100419929', Aliases='', description='acidic repeat containing pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9817348, end_position_on_the_genomic_accession=9818025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419930, CurrentID=0, Status='live', Symbol='CENPUP2', Aliases='', description='centromere protein U pseudogene 2', other_designations='MLF1 interacting protein pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25092666, end_position_on_the_genomic_accession=25093861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419932, CurrentID=0, Status='live', Symbol='LOC100419932', Aliases='', description='FGFR1 oncogene partner 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127876999, end_position_on_the_genomic_accession=127877516, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288861, CurrentID=0, Status='live', Symbol='TRMT112P1', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116378884, end_position_on_the_genomic_accession=116379454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419935, CurrentID=0, Status='live', Symbol='LOC100419935', Aliases='', description='tumor protein D52-like 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132745157, end_position_on_the_genomic_accession=132745750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288864, CurrentID=0, Status='live', Symbol='LOC100288864', Aliases='', description='uncharacterized LOC100288864', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=89902072, end_position_on_the_genomic_accession=89906668, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288866, CurrentID=0, Status='live', Symbol='LOC100288866', Aliases='', description='uncharacterized LOC100288866', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49505657, end_position_on_the_genomic_accession=49574064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100288868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288868, CurrentID=0, Status='live', Symbol='KRT18P63', Aliases='', description='keratin 18 pseudogene 63', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17911624, end_position_on_the_genomic_accession=17913044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288869, CurrentID=0, Status='live', Symbol='LOC100288869', Aliases='', description='uncharacterized LOC100288869', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=35718954, end_position_on_the_genomic_accession=35725581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288870, CurrentID=0, Status='live', Symbol='ERLEC1P1', Aliases='', description='endoplasmic reticulum lectin 1 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14078974, end_position_on_the_genomic_accession=14144467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419945, CurrentID=0, Status='live', Symbol='TRIM51DP', Aliases='', description='tripartite motif-containing 51D, pseudogene', other_designations='SPRY domain containing 5 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49876133, end_position_on_the_genomic_accession=49876638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419947, CurrentID=0, Status='live', Symbol='HMGB1P21', Aliases='', description='high mobility group box 1 pseudogene 21', other_designations='high-mobility group box 1 pseudogene 21', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78644242, end_position_on_the_genomic_accession=78644845, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419948, CurrentID=0, Status='live', Symbol='HMGB3P17', Aliases='', description='high mobility group box 3 pseudogene 17', other_designations='high-mobility group box 3 pseudogene 17', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123468598, end_position_on_the_genomic_accession=123469216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419949, CurrentID=0, Status='live', Symbol='HMGB3P3', Aliases='', description='high mobility group box 3 pseudogene 3', other_designations='high-mobility group box 3 pseudogene 3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6289807, end_position_on_the_genomic_accession=6290613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419950, CurrentID=0, Status='live', Symbol='HMGB1P35', Aliases='', description='high mobility group box 1 pseudogene 35', other_designations='high-mobility group box 1 pseudogene 35|high-mobility group box 2 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77146568, end_position_on_the_genomic_accession=77147157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419951, CurrentID=0, Status='live', Symbol='LOC100419951', Aliases='', description='uncharacterized LOC100419951', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177797092, end_position_on_the_genomic_accession=177797604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419952, CurrentID=0, Status='live', Symbol='LOC100419952', Aliases='', description='uncharacterized LOC100419952', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11094968, end_position_on_the_genomic_accession=11095487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440040, CurrentID=0, Status='live', Symbol='LOC440040', Aliases='', description='glutamate receptor, metabotropic 5 pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49558528, end_position_on_the_genomic_accession=49810419, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100419955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419955, CurrentID=0, Status='live', Symbol='SDAD1P4', Aliases='', description='SDA1 domain containing 1 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23267236, end_position_on_the_genomic_accession=23269141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270835, CurrentID=0, Status='live', Symbol='RPL5P14', Aliases='RPL5_5_550', description='ribosomal protein L5 pseudogene 14', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31847230, end_position_on_the_genomic_accession=31848217, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288885, CurrentID=0, Status='live', Symbol='DYNLL1P6', Aliases='', description='dynein, light chain, LC8-type 1 pseudogene 6', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100041844, end_position_on_the_genomic_accession=100042104, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873419, CurrentID=0, Status='live', Symbol='RNA5SP153', Aliases='RN5S153', description='RNA, 5S ribosomal pseudogene 153', other_designations='RNA, 5S ribosomal 153', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9411178, end_position_on_the_genomic_accession=9411253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419959, CurrentID=0, Status='live', Symbol='LOC100419959', Aliases='', description='HBS1 like translational GTPase pseudogene', other_designations='HBS1-like translational GTPase pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2088115, end_position_on_the_genomic_accession=2090704, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419960, CurrentID=0, Status='live', Symbol='LOC100419960', Aliases='', description='kelch domain containing 4 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154373789, end_position_on_the_genomic_accession=154374349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419962, CurrentID=0, Status='live', Symbol='ANKRD54P1', Aliases='', description='ankyrin repeat domain 54 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45151440, end_position_on_the_genomic_accession=45152165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419963, CurrentID=0, Status='live', Symbol='YY1P1', Aliases='', description='YY1 transcription factor pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70485717, end_position_on_the_genomic_accession=70486334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419964, CurrentID=0, Status='live', Symbol='RSU1P3', Aliases='', description='Ras suppressor protein 1 pseudogene 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51319848, end_position_on_the_genomic_accession=51320645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288893, CurrentID=0, Status='live', Symbol='LOC100288893', Aliases='', description='uncharacterized LOC100288893', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3481676, end_position_on_the_genomic_accession=3485094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419966, CurrentID=0, Status='live', Symbol='LOC100419966', Aliases='', description='uncharacterized LOC100419966', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125680587, end_position_on_the_genomic_accession=125681101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419967, CurrentID=0, Status='live', Symbol='LOC100419967', Aliases='', description='phospholipid scramblase 4 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146342983, end_position_on_the_genomic_accession=146348477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419968, CurrentID=0, Status='live', Symbol='HMGB1P36', Aliases='', description='high mobility group box 1 pseudogene 36', other_designations='high-mobility group box 1 pseudogene 36|high-mobility group box 2 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70753854, end_position_on_the_genomic_accession=70754573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288897, CurrentID=0, Status='live', Symbol='FAR2P3', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene 3', other_designations='fatty acyl CoA reductase 2 pseudogene 3', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419971, CurrentID=0, Status='live', Symbol='PRELID1P2', Aliases='', description='PRELI domain containing 1 pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25678779, end_position_on_the_genomic_accession=25679423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419972, CurrentID=0, Status='live', Symbol='HMGB1P39', Aliases='', description='high mobility group box 1 pseudogene 39', other_designations='high-mobility group box 1 pseudogene 39|high-mobility group box 2 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75319032, end_position_on_the_genomic_accession=75319980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419974, CurrentID=0, Status='live', Symbol='HMGB1P17', Aliases='', description='high mobility group box 1 pseudogene 17', other_designations='high-mobility group box 1 pseudogene 17', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135636110, end_position_on_the_genomic_accession=135636713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  199745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199745, CurrentID=0, Status='live', Symbol='THAP8', Aliases='', description='THAP domain containing 8', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36034984, end_position_on_the_genomic_accession=36054762, orientation='minus', exon_count=4, OMIM=612536, no_hdr0=''),
  100288904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288904, CurrentID=0, Status='live', Symbol='CICP11', Aliases='', description='capicua transcriptional repressor pseudogene 11', other_designations='capicua homolog pseudogene 11', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55736740, end_position_on_the_genomic_accession=55740503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419977, CurrentID=0, Status='live', Symbol='LOC100419977', Aliases='', description='myeloid leukemia factor 2 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101223941, end_position_on_the_genomic_accession=101224573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419978, CurrentID=0, Status='live', Symbol='HMGB1P46', Aliases='', description='high mobility group box 1 pseudogene 46', other_designations='high-mobility group box 1 pseudogene 46|high-mobility group box 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107173199, end_position_on_the_genomic_accession=107173809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419979, CurrentID=0, Status='live', Symbol='HMGB1P19', Aliases='', description='high mobility group box 1 pseudogene 19', other_designations='high-mobility group box 1 pseudogene 19', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123123110, end_position_on_the_genomic_accession=123123684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419980, CurrentID=0, Status='live', Symbol='LOC100419980', Aliases='', description='golgi membrane protein 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36118111, end_position_on_the_genomic_accession=36118716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  724034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724034, CurrentID=0, Status='live', Symbol='CEND1P2', Aliases='', description='cell cycle exit and neuronal differentiation 1 pseudogene 2', other_designations='CEND1 pseudogene', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24322958, end_position_on_the_genomic_accession=24323864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271116, CurrentID=0, Status='live', Symbol='RPS27P5', Aliases='RPS27_4_170', description='ribosomal protein S27 pseudogene 5', other_designations='ribosomal protein S27-like pseudogene', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226781501, end_position_on_the_genomic_accession=226781745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288911, CurrentID=0, Status='live', Symbol='LOC100288911', Aliases='', description='uncharacterized LOC100288911', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36354749, end_position_on_the_genomic_accession=36355570, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419984, CurrentID=0, Status='live', Symbol='LOC100419984', Aliases='', description='uncharacterized LOC100419984', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55742377, end_position_on_the_genomic_accession=55743528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419985, CurrentID=0, Status='live', Symbol='LOC100419985', Aliases='', description='uncharacterized LOC100419985', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56370638, end_position_on_the_genomic_accession=56371751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419986, CurrentID=0, Status='live', Symbol='LOC100419986', Aliases='', description='uncharacterized LOC100419986', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56373231, end_position_on_the_genomic_accession=56373885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419987, CurrentID=0, Status='live', Symbol='LOC100419987', Aliases='', description='uncharacterized LOC100419987', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56815552, end_position_on_the_genomic_accession=56816753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873420, CurrentID=0, Status='live', Symbol='RNA5SP154', Aliases='RN5S154', description='RNA, 5S ribosomal pseudogene 154', other_designations='RNA, 5S ribosomal 154', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9792751, end_position_on_the_genomic_accession=9792861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419989, CurrentID=0, Status='live', Symbol='THUMPD3P1', Aliases='', description='THUMP domain containing 3 pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40152426, end_position_on_the_genomic_accession=40154364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377605, CurrentID=0, Status='live', Symbol='LOC105377605', Aliases='', description='uncharacterized LOC105377605', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187945822, end_position_on_the_genomic_accession=187992886, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  157247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157247, CurrentID=0, Status='live', Symbol='MGC27345', Aliases='', description='uncharacterized protein MGC27345', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128297685, end_position_on_the_genomic_accession=128307763, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419995, CurrentID=0, Status='live', Symbol='HMGB1P32', Aliases='', description='high mobility group box 1 pseudogene 32', other_designations='high-mobility group box 1 pseudogene 32', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=98422135, end_position_on_the_genomic_accession=98422742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419996, CurrentID=0, Status='live', Symbol='HMGB1P11', Aliases='HMGB1L11', description='high mobility group box 1 pseudogene 11', other_designations='high-mobility group box 1 pseudogene 11|high-mobility group box 1-like 11', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171270963, end_position_on_the_genomic_accession=171271588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288929, CurrentID=0, Status='live', Symbol='LOC100288929', Aliases='', description='coxsackievirus and adenovirus receptor-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420002, CurrentID=0, Status='live', Symbol='LOC100420002', Aliases='', description='PDGFA associated protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171972742, end_position_on_the_genomic_accession=171973250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420003, CurrentID=0, Status='live', Symbol='E2F3P2', Aliases='', description='E2F transcription factor 3 pseudogene 2', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195990999, end_position_on_the_genomic_accession=195995694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420004, CurrentID=0, Status='live', Symbol='HSPA8P10', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 10', other_designations='heat shock 70kDa protein 8 pseudogene 10', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234744709, end_position_on_the_genomic_accession=234745579, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420005, CurrentID=0, Status='live', Symbol='FAR2P4', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene 4', other_designations='fatty acyl CoA reductase 2 pseudogene 4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131296824, end_position_on_the_genomic_accession=131309069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420006, CurrentID=0, Status='live', Symbol='LOC100420006', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene', other_designations='fatty acyl CoA reductase 2 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131335500, end_position_on_the_genomic_accession=131355769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288935, CurrentID=0, Status='live', Symbol='LOC100288935', Aliases='', description='transmembrane protein 258 pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5362574, end_position_on_the_genomic_accession=5362966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420008, CurrentID=0, Status='live', Symbol='LOC100420008', Aliases='', description='transmembrane protein 126B pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73871250, end_position_on_the_genomic_accession=73873321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420011, CurrentID=0, Status='live', Symbol='LOC100420011', Aliases='', description='transmembrane protein 38A pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88057800, end_position_on_the_genomic_accession=88058576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420012, CurrentID=0, Status='live', Symbol='ARF1P2', Aliases='', description='ADP ribosylation factor 1 pseudogene 2', other_designations='ADP-ribosylation factor 1 pseudogene 2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120845112, end_position_on_the_genomic_accession=120845647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420013, CurrentID=0, Status='live', Symbol='HMGB1P49', Aliases='', description='high mobility group box 1 pseudogene 49', other_designations='high-mobility group box 1 pseudogene 49|high-mobility group box 2 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27702338, end_position_on_the_genomic_accession=27702906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149934, CurrentID=0, Status='live', Symbol='NCOR1P1', Aliases='C20orf191, bB329D4.2', description='nuclear receptor corepressor 1 pseudogene 1', other_designations='nuclear receptor co-repressor 1 pseudogene', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=26103416, end_position_on_the_genomic_accession=26114041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420015, CurrentID=0, Status='live', Symbol='HMGB3P29', Aliases='', description='high mobility group box 3 pseudogene 29', other_designations='high-mobility group box 3 pseudogene 29', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17417031, end_position_on_the_genomic_accession=17418288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420018, CurrentID=0, Status='live', Symbol='LOC100420018', Aliases='', description='WEE1 homolog 2 (S. pombe) pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35967010, end_position_on_the_genomic_accession=35968676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420019, CurrentID=0, Status='live', Symbol='LOC100420019', Aliases='', description='glycine-N-acyltransferase pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58748997, end_position_on_the_genomic_accession=58758657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288948, CurrentID=0, Status='live', Symbol='BRD7P1', Aliases='BRD7P', description='bromodomain containing 7 pseudogene 1', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23749106, end_position_on_the_genomic_accession=23751329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929635, CurrentID=0, Status='live', Symbol='LOC101929635', Aliases='', description='uncharacterized LOC101929635', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288950, CurrentID=0, Status='live', Symbol='MTND1P25', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 25', other_designations='MT-ND1 pseudogene 25', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237942702, end_position_on_the_genomic_accession=237943641, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420024, CurrentID=0, Status='live', Symbol='LOC100420024', Aliases='', description='ankyrin repeat domain 33B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89785946, end_position_on_the_genomic_accession=89787837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420025, CurrentID=0, Status='live', Symbol='LOC100420025', Aliases='', description='ankyrin repeat domain 33B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89978913, end_position_on_the_genomic_accession=89980662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420027, CurrentID=0, Status='live', Symbol='LOC100420027', Aliases='', description='dedicator of cytokinesis 9 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=63301523, end_position_on_the_genomic_accession=63302056, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420028, CurrentID=0, Status='live', Symbol='HSPA8P4', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 4', other_designations='heat shock 70kDa protein 8 pseudogene 4', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130140008, end_position_on_the_genomic_accession=130142147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134218, CurrentID=0, Status='live', Symbol='DNAJC21', Aliases='DNAJA5, GS3, JJJ1', description='DnaJ heat shock protein family (Hsp40) member C21', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 21|DnaJ homology subfamily A member 5|JJJ1 DnaJ domain protein homolog|dnaJ homolog subfamily A member 5', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34929593, end_position_on_the_genomic_accession=34958964, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100420030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420030, CurrentID=0, Status='live', Symbol='MORF4L1P4', Aliases='', description='mortality factor 4 like 1 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41199808, end_position_on_the_genomic_accession=41200676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149951, CurrentID=0, Status='live', Symbol='COMMD7', Aliases='C20orf92, dJ1085F17.3', description='COMM domain containing 7', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32702691, end_position_on_the_genomic_accession=32744008, orientation='minus', exon_count=10, OMIM=616703, no_hdr0=''),
  100288960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288960, CurrentID=0, Status='live', Symbol='PRSS43', Aliases='TESSP3', description='protease, serine 43', other_designations='protease, serine, 43', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46753043, end_position_on_the_genomic_accession=46812420, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100420033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420033, CurrentID=0, Status='live', Symbol='HMGB3P7', Aliases='', description='high mobility group box 3 pseudogene 7', other_designations='high-mobility group box 3 pseudogene 7', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101801780, end_position_on_the_genomic_accession=101802609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149954, CurrentID=0, Status='live', Symbol='BPIFB4', Aliases='C20orf186, LPLUNC4, RY2G5, dJ726C3.5', description='BPI fold containing family B member 4', other_designations='BPI fold containing family B, member 4|ligand-binding protein RY2G5|likely ortholog of rat probable ligand-binding protein RY2G5|long palate, lung and nasal epithelium carcinoma associated 4|long palate, lung and nasal epithelium carcinoma-associated protein 4', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33081512, end_position_on_the_genomic_accession=33111751, orientation='plus', exon_count=16, OMIM=615718, no_hdr0=''),
  100420035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420035, CurrentID=0, Status='live', Symbol='LOC100420035', Aliases='', description='phosphoglycerate kinase 1 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=22135575, end_position_on_the_genomic_accession=22136135, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420036, CurrentID=0, Status='live', Symbol='RFPL4AP4', Aliases='', description='ret finger protein-like 4A pseudogene 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147916536, end_position_on_the_genomic_accession=147917364, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420038, CurrentID=0, Status='live', Symbol='HSPA8P12', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 12', other_designations='heat shock 70kDa protein 8 pseudogene 12', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188713644, end_position_on_the_genomic_accession=188715028, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420041, CurrentID=0, Status='live', Symbol='HMGB1P43', Aliases='', description='high mobility group box 1 pseudogene 43', other_designations='high-mobility group box 1 pseudogene 43|high-mobility group box 2 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84940507, end_position_on_the_genomic_accession=84941132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271118, CurrentID=0, Status='live', Symbol='RPS27P15', Aliases='RPS27_8_714', description='ribosomal protein S27 pseudogene 15', other_designations='', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73618312, end_position_on_the_genomic_accession=73618654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420043, CurrentID=0, Status='live', Symbol='CACYBPP1', Aliases='', description='calcyclin binding protein pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84454689, end_position_on_the_genomic_accession=84456052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420044, CurrentID=0, Status='live', Symbol='HMGB1P7', Aliases='', description='high mobility group box 1 pseudogene 7', other_designations='high-mobility group box 1 pseudogene 7', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31913149, end_position_on_the_genomic_accession=31913754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288974, CurrentID=0, Status='live', Symbol='BMS1P21', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 21', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 21|BMS1 ribosome biogenesis factor pseudogene 21', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79904898, end_position_on_the_genomic_accession=79931801, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100420047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420047, CurrentID=0, Status='live', Symbol='RAD51AP1P1', Aliases='', description='RAD51 associated protein 1 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136899076, end_position_on_the_genomic_accession=136900076, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420048, CurrentID=0, Status='live', Symbol='LOC100420048', Aliases='', description='poly(rC) binding protein 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157230359, end_position_on_the_genomic_accession=157231195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420049, CurrentID=0, Status='live', Symbol='HSPA8P19', Aliases='', description='heat shock protein family A (Hsp70) member pseudogene 19', other_designations='heat shock 70kDa protein 8 pseudogene 19', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132647121, end_position_on_the_genomic_accession=132650394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420050, CurrentID=0, Status='live', Symbol='LOC100420050', Aliases='', description='downstream neighbor of SON pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136055184, end_position_on_the_genomic_accession=136055764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420051, CurrentID=0, Status='live', Symbol='HSPA8P15', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 15', other_designations='heat shock 70kDa protein 8 pseudogene 15', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151411262, end_position_on_the_genomic_accession=151413957, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420052, CurrentID=0, Status='live', Symbol='LOC100420052', Aliases='', description='poly(rC) binding protein 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45158846, end_position_on_the_genomic_accession=45159531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420053, CurrentID=0, Status='live', Symbol='LOC100420053', Aliases='', description='transmembrane protein 126A pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10336783, end_position_on_the_genomic_accession=10337656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420054, CurrentID=0, Status='live', Symbol='LOC100420054', Aliases='', description='NSFL1 (p97) cofactor (p47) pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47510800, end_position_on_the_genomic_accession=47512161, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420055, CurrentID=0, Status='live', Symbol='RFPL4AP7', Aliases='', description='ret finger protein-like 4A pseudogene 7', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49352472, end_position_on_the_genomic_accession=49353209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288984, CurrentID=0, Status='live', Symbol='MTND1P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 3', other_designations='MT-ND1 pseudogene 3', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141803695, end_position_on_the_genomic_accession=141804632, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420057, CurrentID=0, Status='live', Symbol='LOC100420057', Aliases='', description='aconitase 2, mitochondrial pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99346068, end_position_on_the_genomic_accession=99348300, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420058, CurrentID=0, Status='live', Symbol='RFPL4AP5', Aliases='', description='ret finger protein-like 4A pseudogene 5', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126174186, end_position_on_the_genomic_accession=126175089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420059, CurrentID=0, Status='live', Symbol='RFPL4AP2', Aliases='', description='ret finger protein-like 4A pseudogene 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68500908, end_position_on_the_genomic_accession=68501796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420060, CurrentID=0, Status='live', Symbol='LOC100420060', Aliases='', description='stem-loop binding protein pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4200367, end_position_on_the_genomic_accession=4200960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420061, CurrentID=0, Status='live', Symbol='LOC100420061', Aliases='', description='adenylosuccinate synthase pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35459480, end_position_on_the_genomic_accession=35460549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420062, CurrentID=0, Status='live', Symbol='LOC100420062', Aliases='', description='transmembrane protein 160 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35596520, end_position_on_the_genomic_accession=35597130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420063, CurrentID=0, Status='live', Symbol='HMGB1P24', Aliases='', description='high mobility group box 1 pseudogene 24', other_designations='high-mobility group box 1 pseudogene 24', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37143655, end_position_on_the_genomic_accession=37144291, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420064, CurrentID=0, Status='live', Symbol='LOC100420064', Aliases='', description='density-regulated protein pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63331785, end_position_on_the_genomic_accession=63332646, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149986, CurrentID=0, Status='live', Symbol='LSM14B', Aliases='C20orf40, FAM61B, FT005, LSM13, RAP55B, bA11M20.3', description='LSM family member 14B', other_designations='LSM14 homolog B|LSM14B, SCD6 homolog B|RNA-associated protein 55B|family with sequence similarity 61, member B|hRAP55B', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62122459, end_position_on_the_genomic_accession=62135378, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100420067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420067, CurrentID=0, Status='live', Symbol='LOC100420067', Aliases='', description='transmembrane protein 231 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75499788, end_position_on_the_genomic_accession=75505283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288998, CurrentID=0, Status='live', Symbol='MTND1P15', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 15', other_designations='MT-ND1 pseudogene 15', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22524736, end_position_on_the_genomic_accession=22525696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149992, CurrentID=0, Status='live', Symbol='ANKRD30BP2', Aliases='C21orf99, CT85, CTSP-1, CTSP1', description='ankyrin repeat domain 30B pseudogene 2', other_designations='ankyrin repeat domain 30A pseudogene|cancer-testis SP-1|cancer/testis antigen 85', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13038166, end_position_on_the_genomic_accession=13118251, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100420073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420073, CurrentID=0, Status='live', Symbol='LOC100420073', Aliases='', description='dpy-19-like 4 (C. elegans) pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145583201, end_position_on_the_genomic_accession=145584822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420074, CurrentID=0, Status='live', Symbol='LOC100420074', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=146592018, end_position_on_the_genomic_accession=146592660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420075, CurrentID=0, Status='live', Symbol='PLCE1P1', Aliases='', description='phospholipase C epsilon 1 pseudogene 1', other_designations='phospholipase C, epsilon 1 pseudogene 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30289853, end_position_on_the_genomic_accession=30290358, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289004, CurrentID=0, Status='live', Symbol='LOC100289004', Aliases='', description='actin-related protein 3B-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11124337, end_position_on_the_genomic_accession=11125712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149998, CurrentID=0, Status='live', Symbol='LIPI', Aliases='CT17, LPDL, PLA1C, PRED5, mPA-PLA1 beta', description='lipase, member I', other_designations='LPD lipase|cancer/testis antigen 17|membrane-associated phosphatidic acid-selective phospholipase A1-beta', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14108463, end_position_on_the_genomic_accession=14211040, orientation='minus', exon_count=12, OMIM=609252, no_hdr0=''),
  150000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150000, CurrentID=0, Status='live', Symbol='ABCC13', Aliases='C21orf73, PRED6', description='ATP binding cassette subfamily C member 13, pseudogene', other_designations='ATP-binding cassette protein C13|ATP-binding cassette transporter C13|ATP-binding cassette, sub-family C (CFTR/MRP), member 13, pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14273799, end_position_on_the_genomic_accession=14301386, orientation='plus', exon_count=7, OMIM=608835, no_hdr0=''),
  100289009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289009, CurrentID=0, Status='live', Symbol='LOC100289009', Aliases='', description='uncharacterized LOC100289009', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=32112581, end_position_on_the_genomic_accession=32122062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100420083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420083, CurrentID=0, Status='live', Symbol='LOC100420083', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48197221, end_position_on_the_genomic_accession=48199074, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420084, CurrentID=0, Status='live', Symbol='LOC100420084', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48244879, end_position_on_the_genomic_accession=48245746, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150005, CurrentID=0, Status='live', Symbol='LOC150005', Aliases='', description='uncharacterized LOC150005', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289017, CurrentID=0, Status='live', Symbol='MOXD2P', Aliases='MOXD2', description='monooxygenase, DBH-like 2, pseudogene', other_designations='DBH-like monooxygenase protein 2|dopamine-beta-hydroxylase-like pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142240737, end_position_on_the_genomic_accession=142247067, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100289019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289019, CurrentID=0, Status='live', Symbol='SLC25A25-AS1', Aliases='', description='SLC25A25 antisense RNA 1', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128111171, end_position_on_the_genomic_accession=128118734, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100420092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420092, CurrentID=0, Status='live', Symbol='LOC100420092', Aliases='', description='poly(rC) binding protein 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109103446, end_position_on_the_genomic_accession=109104776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420095, CurrentID=0, Status='live', Symbol='LOC100420095', Aliases='', description='uncharacterized LOC100420095', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19160556, end_position_on_the_genomic_accession=19161071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420096, CurrentID=0, Status='live', Symbol='LOC100420096', Aliases='', description='IGF-like family receptor 1 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65234515, end_position_on_the_genomic_accession=65235101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420097, CurrentID=0, Status='live', Symbol='LOC100420097', Aliases='', description='uncharacterized LOC100420097', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19238523, end_position_on_the_genomic_accession=19239041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420098, CurrentID=0, Status='live', Symbol='LOC100420098', Aliases='', description='kinesin family member 20B pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=47794443, end_position_on_the_genomic_accession=47795754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289027, CurrentID=0, Status='live', Symbol='CDK2AP2P1', Aliases='', description='cyclin-dependent kinase 2 associated protein 2 pseudogene 1', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41280458, end_position_on_the_genomic_accession=41280826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420100, CurrentID=0, Status='live', Symbol='HMGB1P34', Aliases='', description='high mobility group box 1 pseudogene 34', other_designations='high-mobility group box 1 pseudogene 34|high-mobility group box 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67615484, end_position_on_the_genomic_accession=67616094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271120, CurrentID=0, Status='live', Symbol='RPS28P2', Aliases='RPS28_1_163', description='ribosomal protein S28 pseudogene 2', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212855169, end_position_on_the_genomic_accession=212855366, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420103, CurrentID=0, Status='live', Symbol='LOC100420103', Aliases='', description='uncharacterized LOC100420103', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19659258, end_position_on_the_genomic_accession=19659814, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289034, CurrentID=0, Status='live', Symbol='RPL23AP88', Aliases='', description='ribosomal protein L23a pseudogene 88', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242175172, end_position_on_the_genomic_accession=242175633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873424, CurrentID=0, Status='live', Symbol='RNA5SP158', Aliases='RN5S158', description='RNA, 5S ribosomal pseudogene 158', other_designations='RNA, 5S ribosomal 158', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38758791, end_position_on_the_genomic_accession=38758892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420109, CurrentID=0, Status='live', Symbol='LOC100420109', Aliases='', description='WW domain binding protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86929716, end_position_on_the_genomic_accession=86930909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420110, CurrentID=0, Status='live', Symbol='LOC100420110', Aliases='', description='SCY1-like, kinase-like 2 pseudogene', other_designations='SCY1-like 2 (S. cerevisiae) pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201410320, end_position_on_the_genomic_accession=201413337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729506, CurrentID=0, Status='live', Symbol='LOC729506', Aliases='', description='uncharacterized LOC729506', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8333483, end_position_on_the_genomic_accession=8457564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100420113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420113, CurrentID=0, Status='live', Symbol='KLF4P1', Aliases='', description='Kruppel-like factor 4 (gut) pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4803791, end_position_on_the_genomic_accession=4805034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420114, CurrentID=0, Status='live', Symbol='LOC100420114', Aliases='', description='SPATA31 subfamily D, member 1 pseudogene', other_designations='family with sequence similarity 75, member D1 pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35001342, end_position_on_the_genomic_accession=35002994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100682260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100682260, CurrentID=0, Status='live', Symbol='HPPD', Aliases='', description='Hypertelorism, preauricular sinus, punctal pits, and deafness', other_designations='', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614187, no_hdr0=''),
  100289045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289045, CurrentID=0, Status='live', Symbol='LOC100289045', Aliases='', description='uncharacterized LOC100289045', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=73240861, end_position_on_the_genomic_accession=73249005, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420118, CurrentID=0, Status='live', Symbol='LOC100420118', Aliases='', description='transmembrane protein 132C pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126195019, end_position_on_the_genomic_accession=126195657, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420119, CurrentID=0, Status='live', Symbol='LOC100420119', Aliases='', description='transmembrane protein 132B pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127624153, end_position_on_the_genomic_accession=127626336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420121, CurrentID=0, Status='live', Symbol='LOC100420121', Aliases='', description='MAS-related GPR, member X3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19029691, end_position_on_the_genomic_accession=19030435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420122, CurrentID=0, Status='live', Symbol='LOC100420122', Aliases='', description='dihydropyrimidinase-like 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58353990, end_position_on_the_genomic_accession=58354499, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289051, CurrentID=0, Status='live', Symbol='CYB5AP5', Aliases='CYB5P5', description='cytochrome b5 type A (microsomal) pseudogene 5', other_designations='cytochrome b-5 pseudogene 5', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26207962, end_position_on_the_genomic_accession=26208728, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420127, CurrentID=0, Status='live', Symbol='LOC100420127', Aliases='', description='mediator complex subunit 13 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151377948, end_position_on_the_genomic_accession=151378640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289058, CurrentID=0, Status='live', Symbol='LOC100289058', Aliases='', description='uncharacterized LOC100289058', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=41110607, end_position_on_the_genomic_accession=41120625, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100289059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289059, CurrentID=0, Status='live', Symbol='TMEM261P1', Aliases='', description='transmembrane protein 261 pseudogene 1', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25167283, end_position_on_the_genomic_accession=25169694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271121, CurrentID=0, Status='live', Symbol='RPS28P5', Aliases='RPS28_3_243', description='ribosomal protein S28 pseudogene 5', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74754670, end_position_on_the_genomic_accession=74754819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135932, CurrentID=0, Status='live', Symbol='TMEM139', Aliases='', description='transmembrane protein 139', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143284899, end_position_on_the_genomic_accession=143288049, orientation='plus', exon_count=5, OMIM=616524, no_hdr0=''),
  100873425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873425, CurrentID=0, Status='live', Symbol='RNA5SP159', Aliases='RN5S159', description='RNA, 5S ribosomal pseudogene 159', other_designations='RNA, 5S ribosomal 159', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39936753, end_position_on_the_genomic_accession=39936855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377610, CurrentID=0, Status='live', Symbol='LOC105377610', Aliases='', description='uncharacterized LOC105377610', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188776564, end_position_on_the_genomic_accession=188816899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100419515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419515, CurrentID=0, Status='live', Symbol='LOC100419515', Aliases='', description='zinc finger protein 41 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92840955, end_position_on_the_genomic_accession=92841688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420143, CurrentID=0, Status='live', Symbol='HSPA8P3', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 3', other_designations='heat shock 70kDa protein 8 pseudogene 3', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26714632, end_position_on_the_genomic_accession=26716055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420144, CurrentID=0, Status='live', Symbol='LOC100420144', Aliases='', description='mediator of cell motility 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124621710, end_position_on_the_genomic_accession=124622532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420149, CurrentID=0, Status='live', Symbol='BECN1P2', Aliases='', description='beclin 1, autophagy related, pseudogene 2', other_designations='beclin 1, autophagy related pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71075261, end_position_on_the_genomic_accession=71076315, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420155, CurrentID=0, Status='live', Symbol='RPL7L1P2', Aliases='', description='ribosomal protein L7 like 1 pseudogene 2', other_designations='ribosomal protein L7-like 1 pseudogene 2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51259311, end_position_on_the_genomic_accession=51260172, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289085, CurrentID=0, Status='live', Symbol='LOC100289085', Aliases='', description='protein fem-1 homolog A-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6894721, end_position_on_the_genomic_accession=6901509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420158, CurrentID=0, Status='live', Symbol='FTLP16', Aliases='', description='ferritin, light polypeptide pseudogene 16', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38484761, end_position_on_the_genomic_accession=38485264, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289087, CurrentID=0, Status='live', Symbol='TSPY10', Aliases='', description='testis specific protein, Y-linked 10', other_designations='Testis-specific Y-encoded protein 10', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9527899, end_position_on_the_genomic_accession=9530683, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100288580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288580, CurrentID=0, Status='live', Symbol='ACTR3P1', Aliases='', description='ACTR3 pseudogene 1', other_designations='ARP3 actin-related protein 3 homolog pseudogene 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21585289, end_position_on_the_genomic_accession=21586280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150082, CurrentID=0, Status='live', Symbol='LCA5L', Aliases='C21orf13', description='Leber congenital amaurosis 5-like', other_designations='leber congenital amaurosis 5-like protein', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39405844, end_position_on_the_genomic_accession=39445793, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  106480879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480879, CurrentID=0, Status='live', Symbol='RN7SKP147', Aliases='', description='RNA, 7SK small nuclear pseudogene 147', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  150084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150084, CurrentID=0, Status='live', Symbol='IGSF5', Aliases='GSF5, JAM4', description='immunoglobulin superfamily member 5', other_designations='JAM-4|immunoglobulin superfamily 5 like|immunoglobulin superfamily, member 5|junctional adhesion molecule 4', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39725838, end_position_on_the_genomic_accession=39802096, orientation='plus', exon_count=11, OMIM=610638, no_hdr0=''),
  100420166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420166, CurrentID=0, Status='live', Symbol='LOC100420166', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86404179, end_position_on_the_genomic_accession=86404902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289095, CurrentID=0, Status='live', Symbol='BCL2L12P1', Aliases='', description='BCL2-like 12 (proline rich) pseudogene 1', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131526468, end_position_on_the_genomic_accession=131527141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289097, CurrentID=0, Status='live', Symbol='FRG1CP', Aliases='', description='FSHD region gene 1 family member C, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=28557090, end_position_on_the_genomic_accession=28602838, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100289098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289098, CurrentID=0, Status='live', Symbol='GS1-124K5.4', Aliases='', description='uncharacterized LOC100289098', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66493684, end_position_on_the_genomic_accession=66495474, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100289099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289099, CurrentID=0, Status='live', Symbol='RPL30P16', Aliases='', description='ribosomal protein L30 pseudogene 16', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=112538889, end_position_on_the_genomic_accession=112539314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420172, CurrentID=0, Status='live', Symbol='LOC100420172', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene', other_designations='fatty acyl CoA reductase 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18631388, end_position_on_the_genomic_accession=18648749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150094, CurrentID=0, Status='live', Symbol='SIK1', Aliases='EIEE30, MSK, SIK, SNF1LK', description='salt inducible kinase 1', other_designations='SIK-1|SNF1-like kinase|myocardial SNF1-like kinase|salt-inducible kinase 1|salt-inducible protein kinase 1|serine/threonine protein kinase|serine/threonine-protein kinase SNF1-like kinase 1|serine/threonine-protein kinase SNF1LK', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43414509, end_position_on_the_genomic_accession=43427126, orientation='minus', exon_count=14, OMIM=605705, no_hdr0=''),
  100420175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420175, CurrentID=0, Status='live', Symbol='LOC100420175', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene', other_designations='fatty acyl CoA reductase 2 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15558448, end_position_on_the_genomic_accession=15575790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420177, CurrentID=0, Status='live', Symbol='LOC100420177', Aliases='', description='kelch like family member 12 pseudogene', other_designations='kelch-like family member 12 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20449901, end_position_on_the_genomic_accession=20453240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420178, CurrentID=0, Status='live', Symbol='E2F6P3', Aliases='', description='E2F transcription factor 6 pseudogene 3', other_designations='E2F6 pseudogene 3', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21267884, end_position_on_the_genomic_accession=21270091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420180, CurrentID=0, Status='live', Symbol='DBF4P2', Aliases='', description='DBF4 zinc finger pseudogene 2', other_designations='DBF4 homolog pseudogene 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111558400, end_position_on_the_genomic_accession=111559937, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100682325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100682325, CurrentID=0, Status='live', Symbol='NDNC9', Aliases='', description='Nail disorder, nonsyndromic congenital, 9', other_designations='', map_location='17q25.1-q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614149, no_hdr0=''),
  100682326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100682326, CurrentID=0, Status='live', Symbol='MYP20', Aliases='', description='Myopia 20, autosomal dominant', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614166, no_hdr0=''),
  105373247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373247, CurrentID=0, Status='live', Symbol='LOC105373247', Aliases='', description='uncharacterized LOC105373247', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71254913, end_position_on_the_genomic_accession=71272322, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420186, CurrentID=0, Status='live', Symbol='LOC100420186', Aliases='', description='WW domain binding protein 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32227424, end_position_on_the_genomic_accession=32228154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420187, CurrentID=0, Status='live', Symbol='LOC100420187', Aliases='', description='YME1 like 1 ATPase pseudogene', other_designations='YME1-like 1 ATPase pseudogene|YME1-like 1 protein pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54268005, end_position_on_the_genomic_accession=54269042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420189, CurrentID=0, Status='live', Symbol='SEPHS1P3', Aliases='', description='selenophosphate synthetase 1 pseudogene 3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176129713, end_position_on_the_genomic_accession=176130294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420190, CurrentID=0, Status='live', Symbol='SEPHS2P1', Aliases='SEPHS1P5', description='selenophosphate synthetase 2 pseudogene 1', other_designations='selenophosphate synthetase 1 pseudogene 5', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14960030, end_position_on_the_genomic_accession=14962182, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271123, CurrentID=0, Status='live', Symbol='RPS29P4', Aliases='RPS29_3_129', description='ribosomal protein S29 pseudogene 4', other_designations='', map_location='1q24-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175297080, end_position_on_the_genomic_accession=175297241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289124, CurrentID=0, Status='live', Symbol='FAM27E2', Aliases='FAM27E1', description='family with sequence similarity 27 member E2', other_designations='family with sequence similarity 27, member E1|family with sequence similarity 27, member E2', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41647674, end_position_on_the_genomic_accession=41649440, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100420198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420198, CurrentID=0, Status='live', Symbol='LUC7L3P1', Aliases='', description='LUC7-like 3 (S. cerevisiae) pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=70090939, end_position_on_the_genomic_accession=70091883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420200, CurrentID=0, Status='live', Symbol='RPL7L1P12', Aliases='', description='ribosomal protein L7 like 1 pseudogene 12', other_designations='ribosomal protein L7-like 1 pseudogene 12', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98949722, end_position_on_the_genomic_accession=98950447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420201, CurrentID=0, Status='live', Symbol='RPL7L1P6', Aliases='', description='ribosomal protein L7 like 1 pseudogene 6', other_designations='ribosomal protein L7-like 1 pseudogene 6', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100469221, end_position_on_the_genomic_accession=100470196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420203, CurrentID=0, Status='live', Symbol='CICP16', Aliases='', description='capicua transcriptional repressor pseudogene 16', other_designations='capicua homolog pseudogene 16', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118636994, end_position_on_the_genomic_accession=118638826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289133, CurrentID=0, Status='live', Symbol='KRT8P32', Aliases='', description='keratin 8 pseudogene 32', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98391816, end_position_on_the_genomic_accession=98393610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420206, CurrentID=0, Status='live', Symbol='RPL36AP55', Aliases='', description='ribosomal protein L36a pseudogene 55', other_designations='ribosomal protein L36a-like pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27904212, end_position_on_the_genomic_accession=27905162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420207, CurrentID=0, Status='live', Symbol='MED28P1', Aliases='', description='mediator complex subunit 28 pseudogene 1', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68991128, end_position_on_the_genomic_accession=68991868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289137, CurrentID=0, Status='live', Symbol='FAM95C', Aliases='', description='family with sequence similarity 95 member C', other_designations='family with sequence similarity 95, member C', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38540567, end_position_on_the_genomic_accession=38545372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420211, CurrentID=0, Status='live', Symbol='LOC100420211', Aliases='', description='SH3-domain binding protein 4 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67456350, end_position_on_the_genomic_accession=67457260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420213, CurrentID=0, Status='live', Symbol='RBM11P1', Aliases='', description='RNA binding motif protein 11 pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132764660, end_position_on_the_genomic_accession=132765457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420214, CurrentID=0, Status='live', Symbol='LOC100420214', Aliases='', description='rhomboid domain containing 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144397959, end_position_on_the_genomic_accession=144398749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150135, CurrentID=0, Status='live', Symbol='LINC00479', Aliases='C21orf129, PRED76', description='long intergenic non-protein coding RNA 479', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41711520, end_position_on_the_genomic_accession=41715775, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100420216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420216, CurrentID=0, Status='live', Symbol='ALOXE3P1', Aliases='', description='arachidonate lipoxygenase 3 pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8071386, end_position_on_the_genomic_accession=8072530, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  724074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724074, CurrentID=0, Status='live', Symbol='DFNY1', Aliases='', description='deafness, Y-linked 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=400043, no_hdr0=''),
  150142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150142, CurrentID=0, Status='live', Symbol='ZNF295-AS1', Aliases='C21orf121, NCRNA00318, PRED87', description='ZNF295 antisense RNA 1', other_designations='ZNF295 antisense RNA 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42009194, end_position_on_the_genomic_accession=42024920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420223, CurrentID=0, Status='live', Symbol='RNMTL1P2', Aliases='', description='RNA methyltransferase like 1 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17962079, end_position_on_the_genomic_accession=17963033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420224, CurrentID=0, Status='live', Symbol='CICP12', Aliases='', description='capicua transcriptional repressor pseudogene 12', other_designations='capicua homolog pseudogene 12', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55797996, end_position_on_the_genomic_accession=55798826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420226, CurrentID=0, Status='live', Symbol='LOC100420226', Aliases='', description='argininosuccinate synthase 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111411319, end_position_on_the_genomic_accession=111411880, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150147, CurrentID=0, Status='live', Symbol='UMODL1-AS1', Aliases='C21orf128', description='UMODL1 antisense RNA 1', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42102134, end_position_on_the_genomic_accession=42108534, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420228, CurrentID=0, Status='live', Symbol='LOC100420228', Aliases='', description='MAGE family member C3 pseudogene', other_designations='melanoma antigen family C, 3 pseudogene|melanoma antigen family C3 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142174829, end_position_on_the_genomic_accession=142196001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374044, CurrentID=0, Status='live', Symbol='LOC105374044', Aliases='', description='uncharacterized LOC105374044', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112429730, end_position_on_the_genomic_accession=112462050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100289158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289158, CurrentID=0, Status='live', Symbol='HMGB3P13', Aliases='', description='high mobility group box 3 pseudogene 13', other_designations='high-mobility group box 3 pseudogene 13', map_location='3q22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134437502, end_position_on_the_genomic_accession=134438388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150159, CurrentID=0, Status='live', Symbol='SLC9B1', Aliases='NHA1, NHEDC1', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1', other_designations='NHE domain-containing protein 1|Na(+)/H(+) exchanger-like domain-containing protein 1|sodium/hydrogen exchanger-like domain-containing protein 1|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102885048, end_position_on_the_genomic_accession=103019739, orientation='minus', exon_count=18, OMIM=611527, no_hdr0=''),
  150160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150160, CurrentID=0, Status='live', Symbol='CCT8L2', Aliases='CESK1', description='chaperonin containing TCP1, subunit 8 (theta)-like 2', other_designations='putative T-complex protein 1 subunit theta-like 2', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16590758, end_position_on_the_genomic_accession=16592810, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100289170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289170, CurrentID=0, Status='live', Symbol='COX20P2', Aliases='', description='COX20 cytochrome c oxidase assembly factor pseudogene 2', other_designations='COX20 Cox2 chaperone homolog pseudogene 2|COX20 pseudogene 2', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230957504, end_position_on_the_genomic_accession=230958079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150165, CurrentID=0, Status='live', Symbol='XKR3', Aliases='XRG3, XTES', description='X-linked Kx blood group related 3', other_designations='X Kell blood group precursor-related family, member 3|XK, Kell blood group complex subunit-related family, member 3|expressed in testis|x Kell blood group-related 3', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16783412, end_position_on_the_genomic_accession=16821694, orientation='minus', exon_count=4, OMIM=611674, no_hdr0=''),
  105373249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373249, CurrentID=0, Status='live', Symbol='LOC105373249', Aliases='', description='uncharacterized LOC105373249', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420247, CurrentID=0, Status='live', Symbol='LOC100420247', Aliases='', description='mitochondrial fission factor pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101683793, end_position_on_the_genomic_accession=101684450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420249, CurrentID=0, Status='live', Symbol='LOC100420249', Aliases='', description='melanoma antigen family C, 3 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=140990916, end_position_on_the_genomic_accession=140992049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289178, CurrentID=0, Status='live', Symbol='GNG12-AS1', Aliases='', description='GNG12 antisense RNA 1', other_designations='GNG12 antisense RNA 1 (non-protein coding)', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67832288, end_position_on_the_genomic_accession=68202987, orientation='plus', exon_count=10, OMIM=615406, no_hdr0=''),
  100682395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100682395, CurrentID=0, Status='live', Symbol='LNCR5', Aliases='', description='Lung cancer susceptibility 5', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614210, no_hdr0=''),
  100271125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271125, CurrentID=0, Status='live', Symbol='FAUP2', Aliases='RPS30_1_440', description='FBR-MuSV-associated ubiquitously expressed (fox derived) pseudogene 2', other_designations='ribosomal protein S30 pseudogene', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181242605, end_position_on_the_genomic_accession=181242901, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420253, CurrentID=0, Status='live', Symbol='CTBP2P8', Aliases='', description='C-terminal binding protein 2 pseudogene 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68161757, end_position_on_the_genomic_accession=68163033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420254, CurrentID=0, Status='live', Symbol='LOC100420254', Aliases='', description='T, brachyury homolog (mouse) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183709305, end_position_on_the_genomic_accession=183710217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420255, CurrentID=0, Status='live', Symbol='LOC100420255', Aliases='', description='collagen, type XXII, alpha 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200147531, end_position_on_the_genomic_accession=200148479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270837, CurrentID=0, Status='live', Symbol='RPL5P16', Aliases='RPL5_7_588', description='ribosomal protein L5 pseudogene 16', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82777797, end_position_on_the_genomic_accession=82778593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420257, CurrentID=0, Status='live', Symbol='CICP27', Aliases='', description='capicua transcriptional repressor pseudogene 27', other_designations='capicua homolog pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=131068, end_position_on_the_genomic_accession=134836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289186, CurrentID=0, Status='live', Symbol='SDHDP1', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 1', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 1', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12101994, end_position_on_the_genomic_accession=12102471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289187, CurrentID=0, Status='live', Symbol='GS1-259H13.2', Aliases='', description='transmembrane protein 225-like', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99597956, end_position_on_the_genomic_accession=99610833, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100420260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420260, CurrentID=0, Status='live', Symbol='NOTCH2P1', Aliases='', description='notch 2 pseudogene 1', other_designations='Notch homolog 2 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119886306, end_position_on_the_genomic_accession=119886927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420262, CurrentID=0, Status='live', Symbol='LOC100420262', Aliases='', description='mitochondrial fission factor pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179306839, end_position_on_the_genomic_accession=179307418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289191, CurrentID=0, Status='live', Symbol='HMGN3P1', Aliases='', description='high mobility group nucleosomal binding domain 3 pseudogene 1', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152399463, end_position_on_the_genomic_accession=152403683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150185, CurrentID=0, Status='live', Symbol='LINC00895', Aliases='', description='long intergenic non-protein coding RNA 895', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19566130, end_position_on_the_genomic_accession=19566839, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100289194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289194, CurrentID=0, Status='live', Symbol='LOC100289194', Aliases='', description='fibrous sheath CABYR-binding protein-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10858995, end_position_on_the_genomic_accession=10864479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420267, CurrentID=0, Status='live', Symbol='LOC100420267', Aliases='', description='l(3)mbt-like 4 (Drosophila) pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23163370, end_position_on_the_genomic_accession=23164245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289196, CurrentID=0, Status='live', Symbol='LOC100289196', Aliases='', description='chromosome 7 open reading frame 73 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56603312, end_position_on_the_genomic_accession=56604025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131572, CurrentID=0, Status='live', Symbol='MESTP4', Aliases='', description='mesoderm specific transcript pseudogene 4', other_designations='mesoderm specific transcript homolog pseudogene 4', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29087346, end_position_on_the_genomic_accession=29089616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929344, CurrentID=0, Status='live', Symbol='LINC01198', Aliases='', description='long intergenic non-protein coding RNA 1198', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46455203, end_position_on_the_genomic_accession=46466776, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100420273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420273, CurrentID=0, Status='live', Symbol='LOC100420273', Aliases='', description='RAD54-like 2 (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22609608, end_position_on_the_genomic_accession=22610598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420275, CurrentID=0, Status='live', Symbol='PRELID3BP4', Aliases='', description='PRELI domain containing 3B pseudogene 4', other_designations='slowmo homolog 2 (Drosophila) pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42897267, end_position_on_the_genomic_accession=42897830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150197, CurrentID=0, Status='live', Symbol='LINC00896', Aliases='', description='long intergenic non-protein coding RNA 896', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20206332, end_position_on_the_genomic_accession=20208537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100289206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289206, CurrentID=0, Status='live', Symbol='LOC100289206', Aliases='', description='uncharacterized protein CXorf49-like', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71869518, end_position_on_the_genomic_accession=71873439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420281, CurrentID=0, Status='live', Symbol='PLA2G12AP2', Aliases='', description='phospholipase A2 group XIIA pseudogene 2', other_designations='phospholipase A2, group XIIA pseudogene 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38902292, end_position_on_the_genomic_accession=38902801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271126, CurrentID=0, Status='live', Symbol='FAUP3', Aliases='RPS30_1_539', description='FBR-MuSV-associated ubiquitously expressed (fox derived) pseudogene 3', other_designations='ribosomal protein S30 pseudogene', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188106558, end_position_on_the_genomic_accession=188106914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289211, CurrentID=0, Status='live', Symbol='LINC00624', Aliases='', description='long intergenic non-protein coding RNA 624', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147382192, end_position_on_the_genomic_accession=147517875, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100420287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420287, CurrentID=0, Status='live', Symbol='SEC63P2', Aliases='', description='SEC63 homolog, protein translocation regulator pseudogene 2', other_designations='SEC63 homolog (S. cerevisiae) pseudogene 2|SEC63 protein translocation regulator pseudogene 2', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=35487812, end_position_on_the_genomic_accession=35489720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150209, CurrentID=0, Status='live', Symbol='AIFM3', Aliases='AIFL', description='apoptosis inducing factor, mitochondria associated 3', other_designations='apoptosis-inducing factor like|apoptosis-inducing factor, mitochondrion-associated, 3', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20965130, end_position_on_the_genomic_accession=20981360, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100420292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420292, CurrentID=0, Status='live', Symbol='SEPHS1P2', Aliases='', description='selenophosphate synthetase 1 pseudogene 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93430657, end_position_on_the_genomic_accession=93431217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420293, CurrentID=0, Status='live', Symbol='CICP9', Aliases='', description='capicua transcriptional repressor pseudogene 9', other_designations='capicua homolog pseudogene 9', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38454007, end_position_on_the_genomic_accession=38455839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420295, CurrentID=0, Status='live', Symbol='RNMTL1P1', Aliases='', description='RNA methyltransferase like 1 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21394033, end_position_on_the_genomic_accession=21394592, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420298, CurrentID=0, Status='live', Symbol='LOC100420298', Aliases='', description='makorin ring finger protein 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130927750, end_position_on_the_genomic_accession=130928285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420300, CurrentID=0, Status='live', Symbol='FTH1P23', Aliases='FTHL23', description='ferritin, heavy polypeptide 1 pseudogene 23', other_designations='ferritin, heavy polypeptide-like 23', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72928843, end_position_on_the_genomic_accession=72929707, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150221, CurrentID=0, Status='live', Symbol='RIMBP3C', Aliases='RIM-BP3.3, RIMBP3.3', description='RIMS binding protein 3C', other_designations='RIM-BP3.C|RIMS binding protein 3.3|RIMS-binding protein 3.3', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21545666, end_position_on_the_genomic_accession=21551461, orientation='minus', exon_count=1, OMIM=612701, no_hdr0=''),
  100289230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289230, CurrentID=0, Status='live', Symbol='LOC100289230', Aliases='', description='uncharacterized LOC100289230', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98929134, end_position_on_the_genomic_accession=98931009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  150223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150223, CurrentID=0, Status='live', Symbol='YDJC', Aliases='', description='YdjC homolog (bacterial)', other_designations='UPF0249 protein ydjC homolog', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21628089, end_position_on_the_genomic_accession=21630325, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100420304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420304, CurrentID=0, Status='live', Symbol='LOC100420304', Aliases='', description='alpha-methylacyl-CoA racemase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150509600, end_position_on_the_genomic_accession=150510300, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420305, CurrentID=0, Status='live', Symbol='LOC100420305', Aliases='', description='unc-93 homolog A (C. elegans) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167208634, end_position_on_the_genomic_accession=167209561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373251, CurrentID=0, Status='live', Symbol='LOC105373251', Aliases='', description='uncharacterized LOC105373251', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420307, CurrentID=0, Status='live', Symbol='NTAN1P2', Aliases='', description='N-terminal asparagine amidase pseudogene 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86481561, end_position_on_the_genomic_accession=86483028, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420308, CurrentID=0, Status='live', Symbol='ITM2BP1', Aliases='HsT33912, ITM2BP', description='integral membrane protein 2B pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27612573, end_position_on_the_genomic_accession=27613251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289237, CurrentID=0, Status='live', Symbol='CEP164P1', Aliases='', description='centrosomal protein 164kDa pseudogene 1', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45070764, end_position_on_the_genomic_accession=45071209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289238, CurrentID=0, Status='live', Symbol='LOC100289238', Aliases='', description='zinc finger RNA binding protein pseudogene', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89139044, end_position_on_the_genomic_accession=89139447, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3193, CurrentID=0, Status='live', Symbol='HOAC', Aliases='ULG4', description='hypoacusis 2 (autosomal recessive)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271127, CurrentID=0, Status='live', Symbol='RPS3AP8', Aliases='RPS3A_3_138', description='ribosomal protein S3a pseudogene 8', other_designations='', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183266500, end_position_on_the_genomic_accession=183267271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289244, CurrentID=0, Status='live', Symbol='LOC100289244', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84999618, end_position_on_the_genomic_accession=85000312, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420317, CurrentID=0, Status='live', Symbol='LOC100420317', Aliases='', description='charged multivesicular body protein 5 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46725162, end_position_on_the_genomic_accession=46725818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420318, CurrentID=0, Status='live', Symbol='LOC100420318', Aliases='', description='melanoma antigen family A, 12 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51030622, end_position_on_the_genomic_accession=51031369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420320, CurrentID=0, Status='live', Symbol='LOC100420320', Aliases='', description='kinesin heavy chain member 2A pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=127728721, end_position_on_the_genomic_accession=127730940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929645, CurrentID=0, Status='live', Symbol='LOC101929645', Aliases='', description='uncharacterized LOC101929645', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29065244, end_position_on_the_genomic_accession=29073223, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420322, CurrentID=0, Status='live', Symbol='LOC100420322', Aliases='', description='melanoma antigen family A10 pseudogene', other_designations='melanoma antigen family A, 10 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149758009, end_position_on_the_genomic_accession=149758661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420323, CurrentID=0, Status='live', Symbol='LOC100420323', Aliases='', description='melanoma antigen family B18 pseudogene', other_designations='melanoma antigen family B, 18 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27811884, end_position_on_the_genomic_accession=27812915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150244, CurrentID=0, Status='live', Symbol='ZDHHC8P1', Aliases='ZDHHC8P', description='zinc finger, DHHC-type containing 8 pseudogene 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23390605, end_position_on_the_genomic_accession=23402612, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100420325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420325, CurrentID=0, Status='live', Symbol='LOC100420325', Aliases='', description='melanoma antigen family B16 pseudogene', other_designations='melanoma antigen family B, 16 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35710950, end_position_on_the_genomic_accession=35711858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420326, CurrentID=0, Status='live', Symbol='LOC100420326', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene', other_designations='ferritin, heavy polypeptide-like 17 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37441441, end_position_on_the_genomic_accession=37442236, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289255, CurrentID=0, Status='live', Symbol='LINC00675', Aliases='', description='long intergenic non-protein coding RNA 675', other_designations='', map_location='17p13.1-p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10794913, end_position_on_the_genomic_accession=10804099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  150248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150248, CurrentID=0, Status='live', Symbol='C22orf15', Aliases='N27C7-3', description='chromosome 22 open reading frame 15', other_designations='protein N27C7-3', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23760033, end_position_on_the_genomic_accession=23765863, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100289259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289259, CurrentID=0, Status='live', Symbol='ZFYVE9P1', Aliases='', description='zinc finger, FYVE domain containing 9 pseudogene 1', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=137660519, end_position_on_the_genomic_accession=137662242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420334, CurrentID=0, Status='live', Symbol='LOC100420334', Aliases='', description='MAGE family member A11 pseudogene', other_designations='melanoma antigen family A, 11 pseudogene|melanoma antigen family A11 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149760770, end_position_on_the_genomic_accession=149761298, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289264, CurrentID=0, Status='live', Symbol='RPL7P60', Aliases='', description='ribosomal protein L7 pseudogene 60', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100139565, end_position_on_the_genomic_accession=100140698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289265, CurrentID=0, Status='live', Symbol='FAM197Y1P', Aliases='FAM197Y1', description='family with sequence similarity 197 Y-linked member 1, pseudogene', other_designations='family with sequence similarity 197 Y-linked member 1|family with sequence similarity 197, Y-linked, member 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9378306, end_position_on_the_genomic_accession=9384726, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101926983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926983, CurrentID=0, Status='live', Symbol='LOC101926983', Aliases='', description='uncharacterized LOC101926983', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120812181, end_position_on_the_genomic_accession=120836386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420339, CurrentID=0, Status='live', Symbol='LOC100420339', Aliases='', description='opioid growth factor receptor pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3069658, end_position_on_the_genomic_accession=3070697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420341, CurrentID=0, Status='live', Symbol='SEC63P1', Aliases='', description='SEC63 homolog, protein translocation regulator pseudogene 1', other_designations='SEC63 homolog (S. cerevisiae) pseudogene 1|SEC63 protein translocation regulator pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97545701, end_position_on_the_genomic_accession=97546960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420342, CurrentID=0, Status='live', Symbol='CHIAP1', Aliases='', description='chitinase, acidic pseudogene 1', other_designations='chitinase 1 (chitotriosidase) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111251012, end_position_on_the_genomic_accession=111256716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420343, CurrentID=0, Status='live', Symbol='LOC100420343', Aliases='', description='translin-associated factor X pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22332925, end_position_on_the_genomic_accession=22333435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289274, CurrentID=0, Status='live', Symbol='DNAJC3-AS1', Aliases='', description='DNAJC3 antisense RNA 1 (head to head)', other_designations='DNAJC3 antisense RNA 1 (non-protein coding)', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95672633, end_position_on_the_genomic_accession=95676960, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271255, CurrentID=0, Status='live', Symbol='RPS3AP27', Aliases='RPS3A_12_850', description='ribosomal protein S3a pseudogene 27', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133732493, end_position_on_the_genomic_accession=133733220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873432, CurrentID=0, Status='live', Symbol='RNA5SP167', Aliases='RN5S167', description='RNA, 5S ribosomal pseudogene 167', other_designations='RNA, 5S ribosomal 167', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=149907560, end_position_on_the_genomic_accession=149907675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289279, CurrentID=0, Status='live', Symbol='LOC100289279', Aliases='', description='contactin associated protein-like 3-like', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61330592, end_position_on_the_genomic_accession=61461342, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100289280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289280, CurrentID=0, Status='live', Symbol='CUBNP2', Aliases='', description='cubilin (intrinsic factor-cobalamin receptor) pseudogene 2', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45232018, end_position_on_the_genomic_accession=45247306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420353, CurrentID=0, Status='live', Symbol='LOC100420353', Aliases='', description='tripartite motif containing 60 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222645916, end_position_on_the_genomic_accession=222647320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150274, CurrentID=0, Status='live', Symbol='HSCB', Aliases='DNAJC20, HSC20, JAC1', description='HscB mitochondrial iron-sulfur cluster co-chaperone', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 20|HscB iron-sulfur cluster co-chaperone homolog|J-type co-chaperone HSC20', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28742001, end_position_on_the_genomic_accession=28757518, orientation='plus', exon_count=6, OMIM=608142, no_hdr0=''),
  150275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150275, CurrentID=0, Status='live', Symbol='CCDC117', Aliases='dJ366L4.1', description='coiled-coil domain containing 117', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28772674, end_position_on_the_genomic_accession=28789301, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100420356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420356, CurrentID=0, Status='live', Symbol='NTAN1P3', Aliases='', description='N-terminal asparagine amidase pseudogene 3', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67424255, end_position_on_the_genomic_accession=67424985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147759, CurrentID=0, Status='live', Symbol='RNU7-11P', Aliases='U7.11', description='RNA, U7 small nuclear 11 pseudogene', other_designations='RNA, small nuclear U7.11 pseudogene', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39621012, end_position_on_the_genomic_accession=39621072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420359, CurrentID=0, Status='live', Symbol='RPA2P3', Aliases='', description='replication protein A2 pseudogene 3', other_designations='replication protein A2, 32kDa pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100336869, end_position_on_the_genomic_accession=100337622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150280, CurrentID=0, Status='live', Symbol='HORMAD2', Aliases='CT46.2', description='HORMA domain containing 2', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30080069, end_position_on_the_genomic_accession=30210882, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100420362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420362, CurrentID=0, Status='live', Symbol='TRIM60P1Y', Aliases='TRIM60P1, TRIM60PY1', description='tripartite motif containing 60 pseudogene 1, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 1, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7845483, end_position_on_the_genomic_accession=7846463, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420363, CurrentID=0, Status='live', Symbol='TRIM60P2Y', Aliases='TRIM60P2, TRIM60PY2', description='tripartite motif containing 60 pseudogene 2, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 2, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 2', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7905536, end_position_on_the_genomic_accession=7907317, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420364, CurrentID=0, Status='live', Symbol='TRIM60P3Y', Aliases='TRIM60P3, TRIM60PY3', description='tripartite motif containing 60 pseudogene 3, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 3, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 3', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8413046, end_position_on_the_genomic_accession=8414054, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373253, CurrentID=0, Status='live', Symbol='LOC105373253', Aliases='', description='uncharacterized LOC105373253', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75215719, end_position_on_the_genomic_accession=75226003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420367, CurrentID=0, Status='live', Symbol='TRIM60P4Y', Aliases='TRIM60P4, TRIM60PY4', description='tripartite motif containing 60 pseudogene 4, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 4, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 4', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22860980, end_position_on_the_genomic_accession=22861844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420368, CurrentID=0, Status='live', Symbol='TRIM60P5Y', Aliases='TRIM60P5, TRIM60PY5', description='tripartite motif containing 60 pseudogene 5, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 5, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 5', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24494728, end_position_on_the_genomic_accession=24495592, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420369, CurrentID=0, Status='live', Symbol='TRIM60P6Y', Aliases='TRIM60P6, TRIM60PY6', description='tripartite motif containing 60 pseudogene 6, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 6, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 6', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24977179, end_position_on_the_genomic_accession=24978326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150290, CurrentID=0, Status='live', Symbol='DUSP18', Aliases='DSP18, DUSP20, LMWDSP20', description='dual specificity phosphatase 18', other_designations='low molecular weight dual specificity phosphatase 20', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30652051, end_position_on_the_genomic_accession=30667890, orientation='minus', exon_count=6, OMIM=611446, no_hdr0=''),
  150291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150291, CurrentID=0, Status='live', Symbol='MORC2-AS1', Aliases='C22orf27, NCRNA00325', description='MORC2 antisense RNA 1', other_designations='MORC2 antisense RNA 1 (non-protein coding)', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30922308, end_position_on_the_genomic_accession=30926654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100289300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289300, CurrentID=0, Status='live', Symbol='HIGD1AP5', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 5', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 5', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18106612, end_position_on_the_genomic_accession=18106843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420373, CurrentID=0, Status='live', Symbol='TRIM60P8Y', Aliases='TRIM60P8, TRIM60PY8', description='tripartite motif containing 60 pseudogene 8, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 8, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 8', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23025594, end_position_on_the_genomic_accession=23026660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420374, CurrentID=0, Status='live', Symbol='TRIM60P9Y', Aliases='TRIM60P9, TRIM60PY9', description='tripartite motif containing 60 pseudogene 9, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 9, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 9', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23058036, end_position_on_the_genomic_accession=23059183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289303, CurrentID=0, Status='live', Symbol='LOC100289303', Aliases='', description='tubulin, alpha 1b pseudogene', other_designations='tubulin alpha-1B chain pseudogene', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96388213, end_position_on_the_genomic_accession=96389056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420376, CurrentID=0, Status='live', Symbol='TRIM60P10Y', Aliases='TRIM60P10, TRIM60PY10', description='tripartite motif containing 60 pseudogene 10, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 10, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 10', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24659337, end_position_on_the_genomic_accession=24660403, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150297, CurrentID=0, Status='live', Symbol='C22orf42', Aliases='dJ90G24.6', description='chromosome 22 open reading frame 42', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32149006, end_position_on_the_genomic_accession=32160432, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100420378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420378, CurrentID=0, Status='live', Symbol='TRIM60P12Y', Aliases='TRIM60P12, TRIM60PY12', description='tripartite motif containing 60 pseudogene 12, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 12, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 12', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25174511, end_position_on_the_genomic_accession=25175386, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377618, CurrentID=0, Status='live', Symbol='LOC105377618', Aliases='', description='uncharacterized LOC105377618', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189821615, end_position_on_the_genomic_accession=189829110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420385, CurrentID=0, Status='live', Symbol='LOC100420385', Aliases='', description='tripartite motif-containing 75 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165026348, end_position_on_the_genomic_accession=165026950, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420386, CurrentID=0, Status='live', Symbol='LOC100420386', Aliases='', description='tripartite motif-containing 75 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165016889, end_position_on_the_genomic_accession=165017503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289315, CurrentID=0, Status='live', Symbol='LOC100289315', Aliases='', description='NME/NM23 nucleoside diphosphate kinase 2 pseudogene', other_designations='non-metastatic cells 2, protein (NM23B) expressed in pseudogene', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47705468, end_position_on_the_genomic_accession=47706036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420389, CurrentID=0, Status='live', Symbol='NTAN1P1', Aliases='', description='N-terminal asparagine amidase pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=115819869, end_position_on_the_genomic_accession=115820514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289319, CurrentID=0, Status='live', Symbol='KRT8P43', Aliases='', description='keratin 8 pseudogene 43', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24597952, end_position_on_the_genomic_accession=24599986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480637, CurrentID=0, Status='live', Symbol='RNU6-1091P', Aliases='', description='RNA, U6 small nuclear 1091, pseudogene', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420393, CurrentID=0, Status='live', Symbol='TPMTP2', Aliases='', description='thiopurine S-methyltransferase pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42145508, end_position_on_the_genomic_accession=42146427, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420394, CurrentID=0, Status='live', Symbol='LOC100420394', Aliases='', description="5'-nucleotidase, cytosolic IIIA pseudogene", other_designations="5'-nucleotidase, cytosolic III pseudogene", map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111633943, end_position_on_the_genomic_accession=111634490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373254, CurrentID=0, Status='live', Symbol='LOC105373254', Aliases='', description='uncharacterized LOC105373254', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77070603, end_position_on_the_genomic_accession=77081473, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81205, CurrentID=0, Status='live', Symbol='OR5BL1P', Aliases='', description='olfactory receptor family 5 subfamily BL member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BL, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58171090, end_position_on_the_genomic_accession=58171722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420401, CurrentID=0, Status='live', Symbol='BTBD10P2', Aliases='', description='BTB (POZ) domain containing 10 pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149795207, end_position_on_the_genomic_accession=149795749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271130, CurrentID=0, Status='live', Symbol='RPS3AP17', Aliases='RPS3A_4_467', description='ribosomal protein S3a pseudogene 17', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29963039, end_position_on_the_genomic_accession=29963756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420403, CurrentID=0, Status='live', Symbol='LOC100420403', Aliases='', description='K(lysine) acetyltransferase 2B pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39309909, end_position_on_the_genomic_accession=39310443, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420404, CurrentID=0, Status='live', Symbol='LOC100420404', Aliases='', description='DENN/MADD domain containing 4A pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9454294, end_position_on_the_genomic_accession=9456826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420405, CurrentID=0, Status='live', Symbol='TRIM60P15', Aliases='', description='tripartite motif containing 60 pseudogene 15', other_designations='tripartite motif-containing 60 pseudogene 15', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46792045, end_position_on_the_genomic_accession=46793119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100158262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100158262, CurrentID=0, Status='live', Symbol='SCARNA9L', Aliases='', description='small Cajal body-specific RNA 9-like', other_designations='small Cajal body-specific RNA 9-like (retrotransposed)', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20136066, end_position_on_the_genomic_accession=20136413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  724105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724105, CurrentID=0, Status='live', Symbol='LOC724105', Aliases='', description='cysteine-rich transmembrane module containing 1 pseudogene', other_designations='putative nuclear protein ORF1-FL49 pseudogene', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174479936, end_position_on_the_genomic_accession=174480740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420408, CurrentID=0, Status='live', Symbol='LOC100420408', Aliases='', description='TBC1 domain family, member 29 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27667168, end_position_on_the_genomic_accession=27668277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420410, CurrentID=0, Status='live', Symbol='H3F3BP2', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene 2', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28416911, end_position_on_the_genomic_accession=28417489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420412, CurrentID=0, Status='live', Symbol='TRIM60P16', Aliases='', description='tripartite motif containing 60 pseudogene 16', other_designations='tripartite motif-containing 60 pseudogene 16', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56631736, end_position_on_the_genomic_accession=56632727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289341, CurrentID=0, Status='live', Symbol='MAN1B1-AS1', Aliases='', description='MAN1B1 antisense RNA 1 (head to head)', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137084946, end_position_on_the_genomic_accession=137086817, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420414, CurrentID=0, Status='live', Symbol='TRIM60P17', Aliases='', description='tripartite motif containing 60 pseudogene 17', other_designations='tripartite motif-containing 60 pseudogene 17', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64085560, end_position_on_the_genomic_accession=64086576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420415, CurrentID=0, Status='live', Symbol='TRIM60P18', Aliases='', description='tripartite motif containing 60 pseudogene 18', other_designations='tripartite motif-containing 60 pseudogene 18', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64355018, end_position_on_the_genomic_accession=64356199, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420418, CurrentID=0, Status='live', Symbol='LOC100420418', Aliases='', description='tripartite motif containing 60 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204206103, end_position_on_the_genomic_accession=204207289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289347, CurrentID=0, Status='live', Symbol='LOC100289347', Aliases='', description='uncharacterized LOC100289347', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=87434931, end_position_on_the_genomic_accession=87441180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420420, CurrentID=0, Status='live', Symbol='LOC100420420', Aliases='', description='tripartite motif containing 39 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242138594, end_position_on_the_genomic_accession=242139882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289350, CurrentID=0, Status='live', Symbol='LOC100289350', Aliases='', description='immunoglobulin superfamily, member 3 pseudogene', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91736711, end_position_on_the_genomic_accession=91767476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420423, CurrentID=0, Status='live', Symbol='LOC100420423', Aliases='', description='FK506 binding protein 8, 38kDa pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202438328, end_position_on_the_genomic_accession=202439765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420424, CurrentID=0, Status='live', Symbol='LOC100420424', Aliases='', description='Vac14 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28857384, end_position_on_the_genomic_accession=28858701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100551499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100551499, CurrentID=0, Status='live', Symbol='LINC00159', Aliases='C21orf44, NCRNA00159', description='long intergenic non-protein coding RNA 159', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32080481, end_position_on_the_genomic_accession=32086295, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420428, CurrentID=0, Status='live', Symbol='LOC100420428', Aliases='', description='ubiquitin specific peptidase 21 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76260754, end_position_on_the_genomic_accession=76261676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420429, CurrentID=0, Status='live', Symbol='LOC100420429', Aliases='', description='ubiquitin specific peptidase 15 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101013664, end_position_on_the_genomic_accession=101014978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150350, CurrentID=0, Status='live', Symbol='ENTHD1', Aliases='CACNA1I, dJ370M22.3', description='ENTH domain containing 1', other_designations='dJ370M22.3 (EPSIN 2B)|epsin-2B', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39743044, end_position_on_the_genomic_accession=39893806, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  134285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134285, CurrentID=0, Status='live', Symbol='TMEM171', Aliases='PRP2', description='transmembrane protein 171', other_designations='proline-rich protein PRP2', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73120561, end_position_on_the_genomic_accession=73131817, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271131, CurrentID=0, Status='live', Symbol='RPS3AP23', Aliases='RPS3A_8_764', description='ribosomal protein S3a pseudogene 23', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=141635720, end_position_on_the_genomic_accession=141636503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150353, CurrentID=0, Status='live', Symbol='DNAJB7', Aliases='DJ5, HSC3', description='DnaJ heat shock protein family (Hsp40) member B7', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7|novel DnaJ domain protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40859549, end_position_on_the_genomic_accession=40862126, orientation='minus', exon_count=1, OMIM=611336, no_hdr0=''),
  135942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135942, CurrentID=0, Status='live', Symbol='OR2A15P', Aliases='OR2A23P, OR2A28P', description='olfactory receptor family 2 subfamily A member 15 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 15 pseudogene|olfactory receptor, family 2, subfamily A, member 23 pseudogene|olfactory receptor, family 2, subfamily A, member 28 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144118464, end_position_on_the_genomic_accession=144119360, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420435, CurrentID=0, Status='live', Symbol='LOC100420435', Aliases='', description='ubiquitin specific peptidase 21 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224170717, end_position_on_the_genomic_accession=224171538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150356, CurrentID=0, Status='live', Symbol='CHADL', Aliases='SLRR4B', description='chondroadherin like', other_designations='chondroadherin-like', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41209869, end_position_on_the_genomic_accession=41241762, orientation='minus', exon_count=8, OMIM=616236, no_hdr0=''),
  100420437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420437, CurrentID=0, Status='live', Symbol='LOC100420437', Aliases='', description='ubiquitin specific peptidase 12 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39261300, end_position_on_the_genomic_accession=39261842, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420438, CurrentID=0, Status='live', Symbol='LOC100420438', Aliases='', description='ubiquitin specific peptidase 12 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61357187, end_position_on_the_genomic_accession=61357729, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420439, CurrentID=0, Status='live', Symbol='LOC100420439', Aliases='', description='ubiquitin specific peptidase 12 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67087746, end_position_on_the_genomic_accession=67088288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420440, CurrentID=0, Status='live', Symbol='LOC100420440', Aliases='', description='ubiquitin specific peptidase 12 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42099146, end_position_on_the_genomic_accession=42099688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420442, CurrentID=0, Status='live', Symbol='LOC100420442', Aliases='', description='tripartite motif-containing 75 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80860325, end_position_on_the_genomic_accession=80861484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289371, CurrentID=0, Status='live', Symbol='MTCO3P22', Aliases='', description='MT-CO3 pseudogene 22', other_designations='mitochondrially encoded cytochrome c oxidase III pseudogene', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100051134, end_position_on_the_genomic_accession=100051917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150365, CurrentID=0, Status='live', Symbol='MEI1', Aliases='SPATA38', description='meiotic double-stranded break formation protein 1', other_designations='meiosis defective 1|meiosis defective protein 1|meiosis inhibitor 1|spermatogenesis associated 38', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41699501, end_position_on_the_genomic_accession=41799456, orientation='plus', exon_count=34, OMIM=608797, no_hdr0=''),
  100420446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420446, CurrentID=0, Status='live', Symbol='LOC100420446', Aliases='', description='tripartite motif containing 48 pseudogene', other_designations='tripartite motif-containing protein LOC642612-like pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55302466, end_position_on_the_genomic_accession=55304160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150368, CurrentID=0, Status='live', Symbol='FAM109B', Aliases='IPIP27B, Ses2', description='family with sequence similarity 109 member B', other_designations='27 kDa inositol polyphosphate phosphatase interacting protein B|family with sequence similarity 109, member B', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42074251, end_position_on_the_genomic_accession=42079438, orientation='plus', exon_count=4, OMIM=614240, no_hdr0=''),
  3216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3216, CurrentID=0, Status='live', Symbol='HOXB6', Aliases='HOX2, HOX2B, HU-2, Hox-2.2', description='homeobox B6', other_designations='homeo box 2B|homeo box B6|homeobox protein Hox-2.2|homeobox protein Hox-2B|homeobox protein Hu-2', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48594095, end_position_on_the_genomic_accession=48604992, orientation='minus', exon_count=5, OMIM=142961, no_hdr0=''),
  100420450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420450, CurrentID=0, Status='live', Symbol='SERBP1P6', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene 6', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42465397, end_position_on_the_genomic_accession=42468866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150372, CurrentID=0, Status='live', Symbol='NFAM1', Aliases='CNAIP', description='NFAT activating protein with ITAM motif 1', other_designations='NFAT-activating protein with ITAM motif 1|calcineurin/NFAT-activating ITAM-containing protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42380407, end_position_on_the_genomic_accession=42437083, orientation='minus', exon_count=8, OMIM=608740, no_hdr0=''),
  100289381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289381, CurrentID=0, Status='live', Symbol='LOC100289381', Aliases='', description='adaptor related protein complex 3, sigma 1 subunit pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24750675, end_position_on_the_genomic_accession=24751624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289383, CurrentID=0, Status='live', Symbol='CICP25', Aliases='', description='capicua transcriptional repressor pseudogene 25', other_designations='capicua homolog pseudogene', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90172104, end_position_on_the_genomic_accession=90175872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373256, CurrentID=0, Status='live', Symbol='LOC105373256', Aliases='', description='uncharacterized LOC105373256', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76149152, end_position_on_the_genomic_accession=76150413, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420457, CurrentID=0, Status='live', Symbol='NT5CP3', Aliases='', description="5',3'-nucleotidase, cytosolic pseudogene 3", other_designations="5', 3'-nucleotidase, cytosolic pseudogene", map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69912864, end_position_on_the_genomic_accession=69913384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150378, CurrentID=0, Status='live', Symbol='FLJ30901', Aliases='', description='uncharacterized protein FLJ30901', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  150379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150379, CurrentID=0, Status='live', Symbol='PNPLA5', Aliases='GS2L, dJ388M5, dJ388M5.4', description='patatin like phospholipase domain containing 5', other_designations='GS2-like protein|patatin-like phospholipase domain containing 5', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43879678, end_position_on_the_genomic_accession=43892013, orientation='minus', exon_count=9, OMIM=611589, no_hdr0=''),
  100289388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289388, CurrentID=0, Status='live', Symbol='KCTD21-AS1', Aliases='', description='KCTD21 antisense RNA 1', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78139793, end_position_on_the_genomic_accession=78173977, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  150381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150381, CurrentID=0, Status='live', Symbol='PRR34-AS1', Aliases='', description='PRR34 antisense RNA 1', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46053846, end_position_on_the_genomic_accession=46058522, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420462, CurrentID=0, Status='live', Symbol='LDHAL6EP', Aliases='', description='lactate dehydrogenase A-like 6E, pseudogene', other_designations='lactate dehydrogenase A-like 6B pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71434408, end_position_on_the_genomic_accession=71437882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150383, CurrentID=0, Status='live', Symbol='CDPF1', Aliases='C22orf40', description='cysteine rich, DPF motif domain containing 1', other_designations='UPF0595 protein C22orf40|cysteine-rich, DPF motif domain containing 1', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46244013, end_position_on_the_genomic_accession=46250679, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  150384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150384, CurrentID=0, Status='live', Symbol='GTSE1-AS1', Aliases='CN5H6.4, LL22NC03-5H6.6', description='GTSE1 antisense RNA 1 (head to head)', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46295143, end_position_on_the_genomic_accession=46296660, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420466, CurrentID=0, Status='live', Symbol='LOC100420466', Aliases='', description='dual specificity phosphatase 3 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27208089, end_position_on_the_genomic_accession=27208629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420467, CurrentID=0, Status='live', Symbol='LOC100420467', Aliases='', description='glyoxalase domain containing 4 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21394016, end_position_on_the_genomic_accession=21394528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873436, CurrentID=0, Status='live', Symbol='RNA5SP171', Aliases='RN5S171', description='RNA, 5S ribosomal pseudogene 171', other_designations='RNA, 5S ribosomal 171', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=166918364, end_position_on_the_genomic_accession=166918478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420471, CurrentID=0, Status='live', Symbol='LOC100420471', Aliases='', description='ADP ribosylation factor like GTPase 4A pseudogene', other_designations='ADP-ribosylation factor like GTPase 4A pseudogene|ADP-ribosylation factor-like 4A pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23878978, end_position_on_the_genomic_accession=23879498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420473, CurrentID=0, Status='live', Symbol='DUSP12P1', Aliases='', description='dual specificity phosphatase 12 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100221283, end_position_on_the_genomic_accession=100223858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420476, CurrentID=0, Status='live', Symbol='SSR1P1', Aliases='', description='signal sequence receptor, alpha pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150384334, end_position_on_the_genomic_accession=150385100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420477, CurrentID=0, Status='live', Symbol='LDHAL6FP', Aliases='', description='lactate dehydrogenase A-like 6F, pseudogene', other_designations='lactate dehydrogenase A-like 6B pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157298627, end_position_on_the_genomic_accession=157300251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420480, CurrentID=0, Status='live', Symbol='LOC100420480', Aliases='', description='polo-like kinase 4 (Drosophila) pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39934848, end_position_on_the_genomic_accession=39937387, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289409, CurrentID=0, Status='live', Symbol='LOC100289409', Aliases='', description='uncharacterized LOC100289409', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=140475648, end_position_on_the_genomic_accession=140476299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100289410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289410, CurrentID=0, Status='live', Symbol='MCF2L-AS1', Aliases='', description='MCF2L antisense RNA 1', other_designations='MCF2L antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112967484, end_position_on_the_genomic_accession=112968638, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420484, CurrentID=0, Status='live', Symbol='THEMIS3P', Aliases='', description='thymocyte selection associated family member 3, pseudogene', other_designations='thymocyte selection associated pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8512827, end_position_on_the_genomic_accession=8513841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289414, CurrentID=0, Status='live', Symbol='PHBP9', Aliases='', description='prohibitin pseudogene 9', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100248195, end_position_on_the_genomic_accession=100249224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289416, CurrentID=0, Status='live', Symbol='LOC100289416', Aliases='', description='zinc finger protein 793', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=98130149, end_position_on_the_genomic_accession=98131200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420489, CurrentID=0, Status='live', Symbol='LOC100420489', Aliases='', description='tripartite motif containing 13 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71834321, end_position_on_the_genomic_accession=71835944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289419, CurrentID=0, Status='live', Symbol='CTF2P', Aliases='NP', description='cardiotrophin 2, pseudogene', other_designations='neuropoietin|ortholog of mouse neuropoietin (pseudogene in humans)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30904248, end_position_on_the_genomic_accession=30910308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271133, CurrentID=0, Status='live', Symbol='RPS4XP10', Aliases='RPS4P10, RPS4X_6_927', description='ribosomal protein S4X pseudogene 10', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94352921, end_position_on_the_genomic_accession=94353715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420494, CurrentID=0, Status='live', Symbol='LOC100420494', Aliases='', description='ADP-ribosylation factor like GTPase 16 pseudogene', other_designations='ADP-ribosylation factor-like 16 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26219086, end_position_on_the_genomic_accession=26219693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150417, CurrentID=0, Status='live', Symbol='LOC150417', Aliases='', description="small nuclear ribonucleoprotein polypeptide A' pseudogene", other_designations="u2 small nuclear ribonucleoprotein A' pseudogene", map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50754449, end_position_on_the_genomic_accession=50755468, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420499, CurrentID=0, Status='live', Symbol='LOC100420499', Aliases='', description='BAI1-associated protein 2-like 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63086310, end_position_on_the_genomic_accession=63087322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420500, CurrentID=0, Status='live', Symbol='LOC100420500', Aliases='', description='galactose-3-O-sulfotransferase 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240672403, end_position_on_the_genomic_accession=240673097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420501, CurrentID=0, Status='live', Symbol='YRDCP1', Aliases='', description='yrdC N(6)-threonylcarbamoyltransferase domain containing pseudogene 1', other_designations='yrdC domain containing pseudogene 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94432634, end_position_on_the_genomic_accession=94433256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420502, CurrentID=0, Status='live', Symbol='CDCA4P1', Aliases='', description='cell division cycle associated 4 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14593578, end_position_on_the_genomic_accession=14594185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420503, CurrentID=0, Status='live', Symbol='EGLN3P1', Aliases='', description='egl-9 family hypoxia-inducible factor 3 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15971893, end_position_on_the_genomic_accession=15972609, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420505, CurrentID=0, Status='live', Symbol='ADAM1B', Aliases='', description='ADAM metallopeptidase domain 1B (pseudogene)', other_designations='ADAM metallopeptidase domain 1B, pseudogene|ADAM metallopeptidase domain 21 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111927194, end_position_on_the_genomic_accession=111928934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420508, CurrentID=0, Status='live', Symbol='LDHAL6DP', Aliases='', description='lactate dehydrogenase A-like 6D, pseudogene', other_designations='lactate dehydrogenase A-like 6B pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83814459, end_position_on_the_genomic_accession=83815561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420509, CurrentID=0, Status='live', Symbol='LOC100420509', Aliases='', description='X-linked inhibitor of apoptosis, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87094673, end_position_on_the_genomic_accession=87096306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420510, CurrentID=0, Status='live', Symbol='LOC100420510', Aliases='', description='myotubularin related protein 12 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101208914, end_position_on_the_genomic_accession=101209647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420512, CurrentID=0, Status='live', Symbol='RPEP1', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65251129, end_position_on_the_genomic_accession=65251781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420513, CurrentID=0, Status='live', Symbol='LOC100420513', Aliases='', description='TNF receptor associated factor 4 pseudogene', other_designations='TNF receptor-associated factor 4 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143752098, end_position_on_the_genomic_accession=143752749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420514, CurrentID=0, Status='live', Symbol='LOC100420514', Aliases='', description='coiled-coil domain containing 86 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180671894, end_position_on_the_genomic_accession=180672600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420515, CurrentID=0, Status='live', Symbol='LOC100420515', Aliases='', description='CTD nuclear envelope phosphatase 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72996920, end_position_on_the_genomic_accession=72997642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373258, CurrentID=0, Status='live', Symbol='LOC105373258', Aliases='', description='uncharacterized LOC105373258', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243960035, end_position_on_the_genomic_accession=243979514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420519, CurrentID=0, Status='live', Symbol='RANP8', Aliases='', description='RAN, member RAS oncogene family pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20747765, end_position_on_the_genomic_accession=20748102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420520, CurrentID=0, Status='live', Symbol='MOCS3P2', Aliases='', description='molybdenum cofactor synthesis 3 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43179500, end_position_on_the_genomic_accession=43180008, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289450, CurrentID=0, Status='live', Symbol='SLC7A15P', Aliases='ARPAT', description='solute carrier family 7 member 15, pseudogene', other_designations='aromatic-preferring amino acid transporter pseudogene|solute carrier family 7 (cationic amino acid transporter, y+ system), member 15, pseudogene|solute carrier family 7, member 15, pseudogene', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20386386, end_position_on_the_genomic_accession=20396585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420526, CurrentID=0, Status='live', Symbol='PAWRP1', Aliases='', description='PRKC, apoptosis, WT1, regulator pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95183648, end_position_on_the_genomic_accession=95184219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420527, CurrentID=0, Status='live', Symbol='RBBP4P2', Aliases='', description='retinoblastoma binding protein 4 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94075911, end_position_on_the_genomic_accession=94077174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420528, CurrentID=0, Status='live', Symbol='LOC100420528', Aliases='', description='CTD nuclear envelope phosphatase 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113885298, end_position_on_the_genomic_accession=113886028, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420530, CurrentID=0, Status='live', Symbol='LOC100420530', Aliases='', description='mucin 4, cell surface associated pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30964215, end_position_on_the_genomic_accession=30966160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420531, CurrentID=0, Status='live', Symbol='LOC100420531', Aliases='', description='abhydrolase domain containing 10 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110598104, end_position_on_the_genomic_accession=110598689, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420532, CurrentID=0, Status='live', Symbol='LOC100420532', Aliases='', description='ribosomal protein S4, Y-linked 2 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47546808, end_position_on_the_genomic_accession=47547925, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289462, CurrentID=0, Status='live', Symbol='DEFB4B', Aliases='DEFB4P', description='defensin beta 4B', other_designations='beta defensin 2|beta defensin-2|beta-defensin 2|defensin, beta 4, pseudogene|defensin, beta 4B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7414863, end_position_on_the_genomic_accession=7416832, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420535, CurrentID=0, Status='live', Symbol='PAWRP2', Aliases='', description='PRKC, apoptosis, WT1, regulator pseudogene 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28749962, end_position_on_the_genomic_accession=28750595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420536, CurrentID=0, Status='live', Symbol='LOC100420536', Aliases='', description='islet cell autoantigen 1, 69kDa pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21998803, end_position_on_the_genomic_accession=21999656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420539, CurrentID=0, Status='live', Symbol='LOC100420539', Aliases='', description='FK506 binding protein 10, 65 kDa pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55711304, end_position_on_the_genomic_accession=55713228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420540, CurrentID=0, Status='live', Symbol='LOC100420540', Aliases='', description='cell division cycle associated 8 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55810784, end_position_on_the_genomic_accession=55811461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420541, CurrentID=0, Status='live', Symbol='LOC100420541', Aliases='', description='abhydrolase domain containing 15 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57835336, end_position_on_the_genomic_accession=57836252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289470, CurrentID=0, Status='live', Symbol='LOC100289470', Aliases='', description='chromosome 5 open reading frame 60 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179522457, end_position_on_the_genomic_accession=179525077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150465, CurrentID=0, Status='live', Symbol='TTL', Aliases='', description='tubulin tyrosine ligase', other_designations='2410003M22Rik|tubulin-tyrosine ligase', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112482166, end_position_on_the_genomic_accession=112532770, orientation='plus', exon_count=8, OMIM=608291, no_hdr0=''),
  100420546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420546, CurrentID=0, Status='live', Symbol='GTF2IP5', Aliases='', description='general transcription factor IIi pseudogene 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65773620, end_position_on_the_genomic_accession=65802092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289475, CurrentID=0, Status='live', Symbol='LDHBP1', Aliases='LDHBL1', description='lactate dehydrogenase B pseudogene 1', other_designations='lactate dehydrogenase B-like 1|lactate dehydrogenase pseudogene', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113276674, end_position_on_the_genomic_accession=113277731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150468, CurrentID=0, Status='live', Symbol='CKAP2L', Aliases='', description='cytoskeleton associated protein 2 like', other_designations='cytoskeleton associated protein 2-like|radial fiber and mitotic spindle protein|radmis', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112736349, end_position_on_the_genomic_accession=112764677, orientation='minus', exon_count=9, OMIM=616174, no_hdr0=''),
  100420549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420549, CurrentID=0, Status='live', Symbol='LOC100420549', Aliases='', description='centrosomal protein 83kDa pseudogene', other_designations='coiled-coil domain containing 41 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149164428, end_position_on_the_genomic_accession=149165421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420551, CurrentID=0, Status='live', Symbol='LOC100420551', Aliases='', description='synovial sarcoma, X breakpoint 5 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48276703, end_position_on_the_genomic_accession=48277578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150472, CurrentID=0, Status='live', Symbol='CBWD2', Aliases='', description='COBW domain containing 2', other_designations='cobalamin synthase W domain-containing protein 2|cobalamin synthetase W domain-containing protein 2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113437691, end_position_on_the_genomic_accession=113496204, orientation='plus', exon_count=20, OMIM=611079, no_hdr0=''),
  100420555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420555, CurrentID=0, Status='live', Symbol='RPEP3', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27739954, end_position_on_the_genomic_accession=27740626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420556, CurrentID=0, Status='live', Symbol='CPT2P1', Aliases='', description='carnitine palmitoyltransferase 2 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53186295, end_position_on_the_genomic_accession=53186871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420557, CurrentID=0, Status='live', Symbol='NSRP1P1', Aliases='CCDC55P1', description='nuclear speckle splicing regulatory protein 1 pseudogene 1', other_designations='coiled-coil domain containing 55 pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77847080, end_position_on_the_genomic_accession=77848999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420558, CurrentID=0, Status='live', Symbol='CDCA4P3', Aliases='', description='cell division cycle associated 4 pseudogene 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207658450, end_position_on_the_genomic_accession=207659139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289489, CurrentID=0, Status='live', Symbol='MINOS1P2', Aliases='', description='mitochondrial inner membrane organizing system 1 pseudogene 2', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64747035, end_position_on_the_genomic_accession=64747540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150483, CurrentID=0, Status='live', Symbol='TEKT4', Aliases='', description='tektin 4', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94867767, end_position_on_the_genomic_accession=94876829, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100420567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420567, CurrentID=0, Status='live', Symbol='PGAM1P6', Aliases='', description='phosphoglycerate mutase 1 pseudogene 6', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23872265, end_position_on_the_genomic_accession=23873014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420568, CurrentID=0, Status='live', Symbol='VDAC1P13', Aliases='', description='voltage-dependent anion channel 1 pseudogene 13', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42463139, end_position_on_the_genomic_accession=42463997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420569, CurrentID=0, Status='live', Symbol='LOC100420569', Aliases='', description='ribosomal L24 domain containing 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97000445, end_position_on_the_genomic_accession=97001584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420571, CurrentID=0, Status='live', Symbol='LOC100420571', Aliases='', description='RAB1A, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190992642, end_position_on_the_genomic_accession=190993567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420572, CurrentID=0, Status='live', Symbol='LOC100420572', Aliases='', description='glutathione S-transferase alpha 5 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196063346, end_position_on_the_genomic_accession=196063963, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289501, CurrentID=0, Status='live', Symbol='FTLP11', Aliases='', description='ferritin, light polypeptide pseudogene 11', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28857786, end_position_on_the_genomic_accession=28858498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420574, CurrentID=0, Status='live', Symbol='VDAC2P4', Aliases='', description='voltage-dependent anion channel 2 pseudogene 4', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134797169, end_position_on_the_genomic_accession=134797879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420575, CurrentID=0, Status='live', Symbol='LOC100420575', Aliases='', description='leucine rich repeat containing 57 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140898423, end_position_on_the_genomic_accession=140899106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420576, CurrentID=0, Status='live', Symbol='LOC100420576', Aliases='', description='death-associated protein kinase 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87799501, end_position_on_the_genomic_accession=87821803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420580, CurrentID=0, Status='live', Symbol='LOC100420580', Aliases='', description='taste receptor, type 2, member 7 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10831086, end_position_on_the_genomic_accession=10831904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289509, CurrentID=0, Status='live', Symbol='KCNIP2-AS1', Aliases='', description='KCNIP2 antisense RNA 1', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101819068, end_position_on_the_genomic_accession=101828889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271136, CurrentID=0, Status='live', Symbol='RPS6P19', Aliases='RPS6_9_1270', description='ribosomal protein S6 pseudogene 19', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101228100, end_position_on_the_genomic_accession=101228917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420584, CurrentID=0, Status='live', Symbol='LOC100420584', Aliases='', description='microtubule associated protein 6 pseudogene', other_designations='microtubule-associated protein 6 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40479153, end_position_on_the_genomic_accession=40480109, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420586, CurrentID=0, Status='live', Symbol='LOC100420586', Aliases='', description='Rho GTPase activating protein 33 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=590303, end_position_on_the_genomic_accession=591615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420587, CurrentID=0, Status='live', Symbol='LOC100420587', Aliases='', description='SHC SH2-domain binding protein 1 pseudogene', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28435388, end_position_on_the_genomic_accession=28727694, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100420588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420588, CurrentID=0, Status='live', Symbol='CCDC58P5', Aliases='', description='coiled-coil domain containing 58 pseudogene 5', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73850265, end_position_on_the_genomic_accession=73851210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420589, CurrentID=0, Status='live', Symbol='LRRC6P1', Aliases='', description='leucine rich repeat containing 6 pseudogene 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38211389, end_position_on_the_genomic_accession=38212799, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420590, CurrentID=0, Status='live', Symbol='RPEP6', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 6', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72131282, end_position_on_the_genomic_accession=72131909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420591, CurrentID=0, Status='live', Symbol='OSBPL9P3', Aliases='', description='oxysterol binding protein like 9 pseudogene 3', other_designations='oxysterol binding protein-like 9 pseudogene 3', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=91115962, end_position_on_the_genomic_accession=91116554, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420593, CurrentID=0, Status='live', Symbol='GTF3AP4', Aliases='', description='general transcription factor IIIA pseudogene 4', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=101271737, end_position_on_the_genomic_accession=101272624, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150519, CurrentID=0, Status='live', Symbol='LOC150519', Aliases='', description='uncharacterized LOC150519', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=132057771, end_position_on_the_genomic_accession=132068935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289533, CurrentID=0, Status='live', Symbol='LOC100289533', Aliases='', description='uncharacterized LOC100289533', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24396906, end_position_on_the_genomic_accession=24397815, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  150527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150527, CurrentID=0, Status='live', Symbol='TISP43', Aliases='', description='uncharacterized LOC150527', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130570846, end_position_on_the_genomic_accession=130584161, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420608, CurrentID=0, Status='live', Symbol='RPP40P2', Aliases='', description='ribonuclease P/MRP 40kDa subunit pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59262845, end_position_on_the_genomic_accession=59263393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420611, CurrentID=0, Status='live', Symbol='LOC100420611', Aliases='', description='coiled-coil domain containing 84 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13073665, end_position_on_the_genomic_accession=13074312, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420612, CurrentID=0, Status='live', Symbol='ERP29P1', Aliases='ERP28P', description='endoplasmic reticulum lumenal protein 29 pseudogene 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52110147, end_position_on_the_genomic_accession=52111196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271137, CurrentID=0, Status='live', Symbol='RPL13AP9', Aliases='RPL13A_1_58', description='ribosomal protein L13a pseudogene 9', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62641117, end_position_on_the_genomic_accession=62641693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420615, CurrentID=0, Status='live', Symbol='LOC100420615', Aliases='', description='signal peptide peptidase like 2A pseudogene', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49364865, end_position_on_the_genomic_accession=49366827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289545, CurrentID=0, Status='live', Symbol='HIST1H3PS1', Aliases='H3F3AP1, dJ45P21.6', description='histone cluster 1, H3, pseudogene 1', other_designations='histone 1, H3, pseudogene 1', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26321876, end_position_on_the_genomic_accession=26322292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150538, CurrentID=0, Status='live', Symbol='SATB2-AS1', Aliases='', description='SATB2 antisense RNA 1', other_designations='SATB2 antisense RNA 1 (non-protein coding)', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199468098, end_position_on_the_genomic_accession=199472758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420620, CurrentID=0, Status='live', Symbol='LOC100420620', Aliases='', description='mitochondrial calcium uniporter regulator 1 pseudogene', other_designations='coiled-coil domain containing 90A pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166181311, end_position_on_the_genomic_accession=166181970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420621, CurrentID=0, Status='live', Symbol='MTAPP2', Aliases='', description='methylthioadenosine phosphorylase pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189969030, end_position_on_the_genomic_accession=189969861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420627, CurrentID=0, Status='live', Symbol='GSTA8P', Aliases='GSTAP7', description='glutathione S-transferase alpha 8, pseudogene', other_designations='glutathione S-transferase alpha 5 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52687945, end_position_on_the_genomic_accession=52697115, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420631, CurrentID=0, Status='live', Symbol='BTRCP1', Aliases='', description='beta-transducin repeat containing E3 ubiquitin protein ligase pseudogene 1', other_designations='beta-transducin repeat containing pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135330633, end_position_on_the_genomic_accession=135331568, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289561, CurrentID=0, Status='live', Symbol='LOC100289561', Aliases='', description='uncharacterized LOC100289561', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102363861, end_position_on_the_genomic_accession=102380633, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  150554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150554, CurrentID=0, Status='live', Symbol='YWHAZP2', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 2', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126555301, end_position_on_the_genomic_accession=126558179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289564, CurrentID=0, Status='live', Symbol='SNRPCP18', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 18', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101562488, end_position_on_the_genomic_accession=101563212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420638, CurrentID=0, Status='live', Symbol='LOC100420638', Aliases='', description='chloride intracellular channel 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65642960, end_position_on_the_genomic_accession=65644230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288596, CurrentID=0, Status='live', Symbol='SNRPCP1', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 1', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32675843, end_position_on_the_genomic_accession=32676529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420641, CurrentID=0, Status='live', Symbol='LOC100420641', Aliases='', description='TNF receptor associated factor 7 pseudogene', other_designations='TNF receptor-associated factor 7, E3 ubiquitin protein ligase pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25825400, end_position_on_the_genomic_accession=25826332, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420642, CurrentID=0, Status='live', Symbol='LOC100420642', Aliases='', description='Rho GTPase activating protein 20 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47561391, end_position_on_the_genomic_accession=47562566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271138, CurrentID=0, Status='live', Symbol='RPL23AP17', Aliases='RPL23A_3_37', description='ribosomal protein L23a pseudogene 17', other_designations='', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41098638, end_position_on_the_genomic_accession=41099133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420644, CurrentID=0, Status='live', Symbol='LOC100420644', Aliases='', description='cell division cycle associated 8 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21980095, end_position_on_the_genomic_accession=21980478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289574, CurrentID=0, Status='live', Symbol='HERC2P4', Aliases='D16F37S5', description='hect domain and RLD 2 pseudogene 4', other_designations='HECT and RLD domain containing E3 ubiquitin protein ligase 2 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32170043, end_position_on_the_genomic_accession=32188113, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100420647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420647, CurrentID=0, Status='live', Symbol='LOC100420647', Aliases='', description='voltage-dependent anion channel 3 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81191361, end_position_on_the_genomic_accession=81192009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150568, CurrentID=0, Status='live', Symbol='LINC01102', Aliases='', description='long intergenic non-protein coding RNA 1102', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104434347, end_position_on_the_genomic_accession=104512757, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420650, CurrentID=0, Status='live', Symbol='PGAM1P7', Aliases='', description='phosphoglycerate mutase 1 pseudogene 7', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46646597, end_position_on_the_genomic_accession=46647330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289579, CurrentID=0, Status='live', Symbol='NDUFA5P9', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 9', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 9|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86213701, end_position_on_the_genomic_accession=86214576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150572, CurrentID=0, Status='live', Symbol='SMYD1', Aliases='BOP, KMT3D, ZMYND18, ZMYND22', description='SET and MYND domain containing 1', other_designations='CD8 beta opposite|SET and MYND domain-containing protein 1|zinc finger, MYND domain containing 18', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88067780, end_position_on_the_genomic_accession=88113384, orientation='plus', exon_count=10, OMIM=606846, no_hdr0=''),
  100289581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289581, CurrentID=0, Status='live', Symbol='DUX4L13', Aliases='', description='double homeobox 4 like 13', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133753248, end_position_on_the_genomic_accession=133754526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420656, CurrentID=0, Status='live', Symbol='QTRT1P1', Aliases='', description='queuine tRNA-ribosyltransferase 1 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=113882647, end_position_on_the_genomic_accession=113883666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150577, CurrentID=0, Status='live', Symbol='LINC01104', Aliases='', description='long intergenic non-protein coding RNA 1104', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100208254, end_position_on_the_genomic_accession=100251484, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420658, CurrentID=0, Status='live', Symbol='LOC100420658', Aliases='', description='post-GPI attachment to proteins 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165941238, end_position_on_the_genomic_accession=165941768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420660, CurrentID=0, Status='live', Symbol='PGAM1P11', Aliases='', description='phosphoglycerate mutase 1 pseudogene 11', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10058673, end_position_on_the_genomic_accession=10059654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420665, CurrentID=0, Status='live', Symbol='RPEP4', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 4', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29608900, end_position_on_the_genomic_accession=29609498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373263, CurrentID=0, Status='live', Symbol='LOC105373263', Aliases='', description='uncharacterized LOC105373263', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244363905, end_position_on_the_genomic_accession=244376806, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420667, CurrentID=0, Status='live', Symbol='LOC100420667', Aliases='', description='solute carrier family 19 (thiamine transporter), member 3 pseudogene', other_designations='solute carrier family 19, member 3 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227642074, end_position_on_the_genomic_accession=227648612, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420668, CurrentID=0, Status='live', Symbol='LOC100420668', Aliases='', description='chromosome 2 open reading frame 16 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27569817, end_position_on_the_genomic_accession=27571966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420669, CurrentID=0, Status='live', Symbol='LOC100420669', Aliases='', description='solute carrier family 44, member 3 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=193256482, end_position_on_the_genomic_accession=193258323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150590, CurrentID=0, Status='live', Symbol='C2orf15', Aliases='', description='chromosome 2 open reading frame 15', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99141722, end_position_on_the_genomic_accession=99151487, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420671, CurrentID=0, Status='live', Symbol='NPAP1P5', Aliases='', description='nuclear pore associated protein 1 pseudogene 5', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87855580, end_position_on_the_genomic_accession=87856586, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271139, CurrentID=0, Status='live', Symbol='RPL36AP9', Aliases='RPL36A_1_38', description='ribosomal protein L36a pseudogene 9', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41264501, end_position_on_the_genomic_accession=41264879, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150596, CurrentID=0, Status='live', Symbol='LOC150596', Aliases='', description='uncharacterized LOC150596', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149803513, end_position_on_the_genomic_accession=149858207, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100289605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289605, CurrentID=0, Status='live', Symbol='HIGD1AP15', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 15', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 15', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10334354, end_position_on_the_genomic_accession=10335218, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420678, CurrentID=0, Status='live', Symbol='CCDC150P1', Aliases='', description='coiled-coil domain containing 150 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122322174, end_position_on_the_genomic_accession=122323847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873443, CurrentID=0, Status='live', Symbol='RNA5SP181', Aliases='RN5S181', description='RNA, 5S ribosomal pseudogene 181', other_designations='RNA, 5S ribosomal 181', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36485343, end_position_on_the_genomic_accession=36485440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420680, CurrentID=0, Status='live', Symbol='LOC100420680', Aliases='', description='coiled-coil domain containing 90B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87323707, end_position_on_the_genomic_accession=87324361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420681, CurrentID=0, Status='live', Symbol='PGAM1P9', Aliases='', description='phosphoglycerate mutase 1 pseudogene 9', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92366499, end_position_on_the_genomic_accession=92367250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420683, CurrentID=0, Status='live', Symbol='LOC100420683', Aliases='', description='chromosome X open reading frame 67 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12794731, end_position_on_the_genomic_accession=12795536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420689, CurrentID=0, Status='live', Symbol='CASC4P1', Aliases='', description='cancer susceptibility candidate 4 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19557832, end_position_on_the_genomic_accession=19565196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289619, CurrentID=0, Status='live', Symbol='TRIAP1P1', Aliases='C10orf135, bA748L13.4', description='TP53 regulated inhibitor of apoptosis 1 pseudogene 1', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27376948, end_position_on_the_genomic_accession=27378399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420693, CurrentID=0, Status='live', Symbol='LOC100420693', Aliases='', description='solute carrier family 47 (multidrug and toxin extrusion), member 2 pseudogene', other_designations='solute carrier family 47, member 2 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68076321, end_position_on_the_genomic_accession=68077854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420695, CurrentID=0, Status='live', Symbol='ABT1P1', Aliases='', description='activator of basal transcription 1 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99022314, end_position_on_the_genomic_accession=99023128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373264, CurrentID=0, Status='live', Symbol='LOC105373264', Aliases='', description='uncharacterized LOC105373264', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244431547, end_position_on_the_genomic_accession=244432532, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  90639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90639, CurrentID=0, Status='live', Symbol='COX19', Aliases='', description='COX19 cytochrome c oxidase assembly factor', other_designations='COX19 cytochrome c oxidase assembly homolog|cytochrome c oxidase assembly homolog 19|hCOX19', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=964850, end_position_on_the_genomic_accession=975599, orientation='minus', exon_count=3, OMIM=610429, no_hdr0=''),
  150622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150622, CurrentID=0, Status='live', Symbol='LINC01105', Aliases='', description='long intergenic non-protein coding RNA 1105', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5932687, end_position_on_the_genomic_accession=5980218, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271140, CurrentID=0, Status='live', Symbol='RPL36AP10', Aliases='RPL36A_1_70', description='ribosomal protein L36a pseudogene 10', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88577845, end_position_on_the_genomic_accession=88578251, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289635, CurrentID=0, Status='live', Symbol='ZNF605', Aliases='', description='zinc finger protein 605', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132921433, end_position_on_the_genomic_accession=132956306, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100420710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420710, CurrentID=0, Status='live', Symbol='TPST2P1', Aliases='', description='tyrosylprotein sulfotransferase 2 pseudogene 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36626677, end_position_on_the_genomic_accession=36627776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420711, CurrentID=0, Status='live', Symbol='LOC100420711', Aliases='', description='thioesterase superfamily member 4 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55442635, end_position_on_the_genomic_accession=55443357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289640, CurrentID=0, Status='live', Symbol='PPATP1', Aliases='', description='phosphoribosyl pyrophosphate amidotransferase pseudogene 1', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87051195, end_position_on_the_genomic_accession=87052738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289643, CurrentID=0, Status='live', Symbol='HSPA8P16', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 16', other_designations='heat shock 70kDa protein 8 pseudogene 16', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=85027748, end_position_on_the_genomic_accession=85029935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289644, CurrentID=0, Status='live', Symbol='PRR13P7', Aliases='', description='proline rich 13 pseudogene 7', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91909536, end_position_on_the_genomic_accession=91910253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289650, CurrentID=0, Status='live', Symbol='LOC100289650', Aliases='', description='uncharacterized LOC100289650', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42821863, end_position_on_the_genomic_accession=42826878, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373265, CurrentID=0, Status='live', Symbol='LOC105373265', Aliases='', description='uncharacterized LOC105373265', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245693823, end_position_on_the_genomic_accession=245698908, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100289656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289656, CurrentID=0, Status='live', Symbol='LOC100289656', Aliases='', description='Dexi homolog (mouse) pseudogene', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28788243, end_position_on_the_genomic_accession=28789392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271141, CurrentID=0, Status='live', Symbol='RPS15AP6', Aliases='RPS15A_1_18', description='ribosomal protein S15a pseudogene 6', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21003511, end_position_on_the_genomic_accession=21003964, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420734, CurrentID=0, Status='live', Symbol='RAB5CP1', Aliases='', description='RAB5C, member RAS oncogene family pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74423433, end_position_on_the_genomic_accession=74424019, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289663, CurrentID=0, Status='live', Symbol='STMN1P2', Aliases='', description='stathmin 1 pseudogene 2', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137425253, end_position_on_the_genomic_accession=137425929, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420737, CurrentID=0, Status='live', Symbol='NEFMP1', Aliases='', description='neurofilament, medium polypeptide pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=54598946, end_position_on_the_genomic_accession=54600680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420738, CurrentID=0, Status='live', Symbol='LOC100420738', Aliases='', description='chromosome 9 open reading frame 78 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159721523, end_position_on_the_genomic_accession=159722367, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289667, CurrentID=0, Status='live', Symbol='POLR3KP1', Aliases='', description='polymerase (RNA) III (DNA directed) polypeptide K, 12.3 kDa pseudogene 1', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59071104, end_position_on_the_genomic_accession=59071840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420741, CurrentID=0, Status='live', Symbol='LOC100420741', Aliases='', description='adenosylmethionine decarboxylase 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87441171, end_position_on_the_genomic_accession=87442146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420742, CurrentID=0, Status='live', Symbol='LOC100420742', Aliases='', description='tetratricopeptide repeat domain 22 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100093187, end_position_on_the_genomic_accession=100093768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420743, CurrentID=0, Status='live', Symbol='LOC100420743', Aliases='', description='chromosome 6 open reading frame 58 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127624292, end_position_on_the_genomic_accession=127634190, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289673, CurrentID=0, Status='live', Symbol='LOC100289673', Aliases='', description='PGAM family member 5, serine/threonine protein phosphatase, mitochondrial pseudogene', other_designations='phosphoglycerate mutase family member 5 pseudogene', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109883182, end_position_on_the_genomic_accession=109885499, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420746, CurrentID=0, Status='live', Symbol='LOC100420746', Aliases='', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 7 pseudogene', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=114318400, end_position_on_the_genomic_accession=114318927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420748, CurrentID=0, Status='live', Symbol='LOC100420748', Aliases='', description='suppressor of cytokine signaling 5 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109416026, end_position_on_the_genomic_accession=109417054, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289678, CurrentID=0, Status='live', Symbol='ZNF783', Aliases='', description='zinc finger family member 783', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149262171, end_position_on_the_genomic_accession=149284994, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100420752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420752, CurrentID=0, Status='live', Symbol='SRP72P1', Aliases='', description='signal recognition particle 72kDa pseudogene 1', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49495146, end_position_on_the_genomic_accession=49496897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289682, CurrentID=0, Status='live', Symbol='HMGN2P13', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 13', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 13', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153067169, end_position_on_the_genomic_accession=153068361, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289683, CurrentID=0, Status='live', Symbol='ERVH-3', Aliases='HERV-H/F', description='endogenous retrovirus group H member 3', other_designations='TCONS_l2_00025412|endogenous retrovirus group H, member 3', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=68591214, end_position_on_the_genomic_accession=68599231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373266, CurrentID=0, Status='live', Symbol='LOC105373266', Aliases='', description='uncharacterized LOC105373266', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245448829, end_position_on_the_genomic_accession=245450551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  150677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150677, CurrentID=0, Status='live', Symbol='OTOS', Aliases='OTOSP', description='otospiralin', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240139029, end_position_on_the_genomic_accession=240140656, orientation='minus', exon_count=4, OMIM=607877, no_hdr0=''),
  150678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150678, CurrentID=0, Status='live', Symbol='MYEOV2', Aliases='', description='myeloma overexpressed 2', other_designations='helicase/primase complex protein', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240126563, end_position_on_the_genomic_accession=240136347, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100420759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420759, CurrentID=0, Status='live', Symbol='LOC100420759', Aliases='', description='suppressor of cytokine signaling 6 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71527621, end_position_on_the_genomic_accession=71528635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420760, CurrentID=0, Status='live', Symbol='LOC100420760', Aliases='', description='coronin, actin binding protein, 1C pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79369385, end_position_on_the_genomic_accession=79370176, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150681, CurrentID=0, Status='live', Symbol='OR6B3', Aliases='OR6B3PQ, OR6B3', description='olfactory receptor family 6 subfamily B member 3', other_designations='olfactory receptor OR2-2|olfactory receptor, family 6, subfamily B, member 3 pseudogene', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240044636, end_position_on_the_genomic_accession=240053711, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271142, CurrentID=0, Status='live', Symbol='RPS15AP7', Aliases='RPS15A_1_57', description='ribosomal protein S15a pseudogene 7', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62190492, end_position_on_the_genomic_accession=62190971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150684, CurrentID=0, Status='live', Symbol='COMMD1', Aliases='C2orf5, MURR1', description='copper metabolism (Murr1) domain containing 1', other_designations='copper metabolism gene MURR1|protein Murr1', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61888728, end_position_on_the_genomic_accession=62136070, orientation='plus', exon_count=6, OMIM=607238, no_hdr0=''),
  100420765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420765, CurrentID=0, Status='live', Symbol='LOC100420765', Aliases='', description='leucine rich pentatricopeptide repeat containing pseudogene', other_designations='leucine-rich pentatricopeptide repeat containing pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=42608868, end_position_on_the_genomic_accession=42609565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420769, CurrentID=0, Status='live', Symbol='LOC100420769', Aliases='', description='glutathione S-transferase theta 1 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420770, CurrentID=0, Status='live', Symbol='LOC100420770', Aliases='', description='chromosome 12 open reading frame 10 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27896126, end_position_on_the_genomic_accession=27897195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420772, CurrentID=0, Status='live', Symbol='LOC100420772', Aliases='', description='coordinator of PRMT5, differentiation stimulator pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140683162, end_position_on_the_genomic_accession=140683951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420774, CurrentID=0, Status='live', Symbol='LOC100420774', Aliases='', description='zinc finger, DHHC-type containing 6 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73877182, end_position_on_the_genomic_accession=73877819, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420775, CurrentID=0, Status='live', Symbol='LOC100420775', Aliases='', description="5'-nucleotidase domain containing 2 pseudogene", other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=211104927, end_position_on_the_genomic_accession=211105730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150696, CurrentID=0, Status='live', Symbol='PROM2', Aliases='PROML2', description='prominin 2', other_designations='hPROML2|prominin-like protein 2|prominin-related protein', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95274416, end_position_on_the_genomic_accession=95291320, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  100420777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420777, CurrentID=0, Status='live', Symbol='CNTNAP3P7', Aliases='', description='contactin associated protein-like 3 pseudogene 7', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42752743, end_position_on_the_genomic_accession=42756964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420778, CurrentID=0, Status='live', Symbol='LOC100420778', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41341033, end_position_on_the_genomic_accession=41341618, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420779, CurrentID=0, Status='live', Symbol='LOC100420779', Aliases='', description='chromosome 16 open reading frame 70 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92339670, end_position_on_the_genomic_accession=92340510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420784, CurrentID=0, Status='live', Symbol='C2orf27AP3', Aliases='', description='chromosome 2 open reading frame 27A pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40121958, end_position_on_the_genomic_accession=40122740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420785, CurrentID=0, Status='live', Symbol='CNTNAP3P8', Aliases='', description='contactin associated protein-like 3 pseudogene 8', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42537581, end_position_on_the_genomic_accession=42541908, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420786, CurrentID=0, Status='live', Symbol='CNTNAP3P5', Aliases='', description='contactin associated protein-like 3 pseudogene 5', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62559838, end_position_on_the_genomic_accession=62564169, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420787, CurrentID=0, Status='live', Symbol='CNTNAP3P1', Aliases='', description='contactin associated protein-like 3 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62924910, end_position_on_the_genomic_accession=62929215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420788, CurrentID=0, Status='live', Symbol='LOC100420788', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63719857, end_position_on_the_genomic_accession=63720442, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150709, CurrentID=0, Status='live', Symbol='ANKAR', Aliases='', description='ankyrin and armadillo repeat containing', other_designations='', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189674292, end_position_on_the_genomic_accession=189763058, orientation='plus', exon_count=30, OMIM=609803, no_hdr0=''),
  100420790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420790, CurrentID=0, Status='live', Symbol='LOC100420790', Aliases='', description='glucosamine-6-phosphate deaminase 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=71804283, end_position_on_the_genomic_accession=71804806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420792, CurrentID=0, Status='live', Symbol='NDFIP1P1', Aliases='', description='Nedd4 family interacting protein 1 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18358443, end_position_on_the_genomic_accession=18358967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420794, CurrentID=0, Status='live', Symbol='LOC100420794', Aliases='', description='RE1-silencing transcription factor pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79307687, end_position_on_the_genomic_accession=79308867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420797, CurrentID=0, Status='live', Symbol='LOC100420797', Aliases='', description='Fc fragment of IgG binding protein pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35079868, end_position_on_the_genomic_accession=35088832, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420798, CurrentID=0, Status='live', Symbol='LOC100420798', Aliases='', description='poly(A) binding protein, nuclear 1 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52809805, end_position_on_the_genomic_accession=52814041, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420800, CurrentID=0, Status='live', Symbol='LOC100420800', Aliases='', description='suppressor of cytokine signaling 6 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83789973, end_position_on_the_genomic_accession=83791578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420801, CurrentID=0, Status='live', Symbol='LOC100420801', Aliases='', description='lysine (K)-specific demethylase 4D pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95011858, end_position_on_the_genomic_accession=95012532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420802, CurrentID=0, Status='live', Symbol='LOC100420802', Aliases='', description='lysine (K)-specific demethylase 4D pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95030543, end_position_on_the_genomic_accession=95031246, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420803, CurrentID=0, Status='live', Symbol='LOC100420803', Aliases='', description='lysine (K)-specific demethylase 4D pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95057273, end_position_on_the_genomic_accession=95057986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420804, CurrentID=0, Status='live', Symbol='LOC100420804', Aliases='', description='zinc finger CCCH-type containing 14 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23262572, end_position_on_the_genomic_accession=23263165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150726, CurrentID=0, Status='live', Symbol='FBXO41', Aliases='FBX41', description='F-box protein 41', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73254682, end_position_on_the_genomic_accession=73284453, orientation='minus', exon_count=17, OMIM=609108, no_hdr0=''),
  100420811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420811, CurrentID=0, Status='live', Symbol='C2orf27AP1', Aliases='', description='chromosome 2 open reading frame 27A pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11085476, end_position_on_the_genomic_accession=11086061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131479, CurrentID=0, Status='live', Symbol='LOC100131479', Aliases='', description='zinc finger protein 430 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23596127, end_position_on_the_genomic_accession=23596898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373268, CurrentID=0, Status='live', Symbol='LOC105373268', Aliases='', description='uncharacterized LOC105373268', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246194292, end_position_on_the_genomic_accession=246201730, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  150737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150737, CurrentID=0, Status='live', Symbol='TTC30B', Aliases='IFT70, fleer', description='tetratricopeptide repeat domain 30B', other_designations='TPR repeat protein 30B', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177550152, end_position_on_the_genomic_accession=177552796, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420818, CurrentID=0, Status='live', Symbol='LOC100420818', Aliases='', description='eva-1 homolog C (C. elegans) pseudogene', other_designations='family with sequence similarity 176, member C pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3921631, end_position_on_the_genomic_accession=3924274, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420821, CurrentID=0, Status='live', Symbol='LOC100420821', Aliases='', description='eva-1 homolog C (C. elegans) pseudogene', other_designations='family with sequence similarity 176, member C pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9733359, end_position_on_the_genomic_accession=9742374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420823, CurrentID=0, Status='live', Symbol='LOC100420823', Aliases='', description='family with sequence similarity 208, member A pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28416449, end_position_on_the_genomic_accession=28422374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420827, CurrentID=0, Status='live', Symbol='NPAP1P1', Aliases='', description='nuclear pore associated protein 1 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50696267, end_position_on_the_genomic_accession=50697149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420828, CurrentID=0, Status='live', Symbol='LOC100420828', Aliases='', description='nuclear pore associated protein 1 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63973722, end_position_on_the_genomic_accession=63976617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420829, CurrentID=0, Status='live', Symbol='NPAP1P2', Aliases='', description='nuclear pore associated protein 1 pseudogene 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79861262, end_position_on_the_genomic_accession=79862137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420830, CurrentID=0, Status='live', Symbol='NPAP1P3', Aliases='', description='nuclear pore associated protein 1 pseudogene 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87377573, end_position_on_the_genomic_accession=87378448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420831, CurrentID=0, Status='live', Symbol='LOC100420831', Aliases='', description='chromosome 1 open reading frame 106 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112185483, end_position_on_the_genomic_accession=112186007, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420832, CurrentID=0, Status='live', Symbol='LOC100420832', Aliases='', description='cysteine and glycine rich protein 2 pseudogene', other_designations='cysteine and glycine-rich protein 2 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130917976, end_position_on_the_genomic_accession=130919199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420833, CurrentID=0, Status='live', Symbol='FGFR3P4', Aliases='', description='fibroblast growth factor receptor 3 pseudogene 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177700301, end_position_on_the_genomic_accession=177701072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420834, CurrentID=0, Status='live', Symbol='SRRM1P2', Aliases='', description='serine/arginine repetitive matrix 1 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83937056, end_position_on_the_genomic_accession=83939038, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134353, CurrentID=0, Status='live', Symbol='LSM11', Aliases='', description='LSM11, U7 small nuclear RNA associated', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157743695, end_position_on_the_genomic_accession=157760709, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420841, CurrentID=0, Status='live', Symbol='KATNBL1P5', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25248189, end_position_on_the_genomic_accession=25248780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150763, CurrentID=0, Status='live', Symbol='GPAT2', Aliases='CT123', description='glycerol-3-phosphate acyltransferase 2, mitochondrial', other_designations='GPAT-2|cancer/testis antigen 123|xGPAT1', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96021946, end_position_on_the_genomic_accession=96035245, orientation='minus', exon_count=22, OMIM=616431, no_hdr0=''),
  100420845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420845, CurrentID=0, Status='live', Symbol='LOC100420845', Aliases='', description='solute carrier family 25 member 43 pseudogene', other_designations='solute carrier family 25, member 43 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=83099976, end_position_on_the_genomic_accession=83100846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420848, CurrentID=0, Status='live', Symbol='LOC100420848', Aliases='', description='Myb/SANT-like DNA-binding domain containing 3 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27526112, end_position_on_the_genomic_accession=27527267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420850, CurrentID=0, Status='live', Symbol='LOC100420850', Aliases='', description='leucine rich repeat containing 37B pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30617666, end_position_on_the_genomic_accession=30672150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150771, CurrentID=0, Status='live', Symbol='ITPRIPL1', Aliases='KIAA1754L', description='inositol 1,4,5-trisphosphate receptor interacting protein-like 1', other_designations='inositol 1,4,5-triphosphate receptor interacting protein-like 1|inositol 1,4,5-triphosphate receptor-interacting protein-like 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96325324, end_position_on_the_genomic_accession=96328353, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928475, CurrentID=0, Status='live', Symbol='LOC101928475', Aliases='', description='uncharacterized LOC101928475', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15014790, end_position_on_the_genomic_accession=15052276, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420853, CurrentID=0, Status='live', Symbol='LOC100420853', Aliases='', description='nitric oxide synthase 2, inducible pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37973983, end_position_on_the_genomic_accession=37974855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420854, CurrentID=0, Status='live', Symbol='SPDYE22P', Aliases='', description='speedy/RINGO cell cycle regulator family member E22, pseudogene', other_designations='speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59174734, end_position_on_the_genomic_accession=59180055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150776, CurrentID=0, Status='live', Symbol='LOC150776', Aliases='', description='sphingomyelin phosphodiesterase 4, neutral membrane (neutral sphingomyelinase-3) pseudogene', other_designations='neutral sphingomyelinase-3 pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131492813, end_position_on_the_genomic_accession=131521576, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100420859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420859, CurrentID=0, Status='live', Symbol='C2orf69P4', Aliases='', description='chromosome 2 open reading frame 69 pseudogene 4', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35352483, end_position_on_the_genomic_accession=35353022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420863, CurrentID=0, Status='live', Symbol='LOC100420863', Aliases='', description='eva-1 homolog C (C. elegans) pseudogene', other_designations='family with sequence similarity 176, member C pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6895595, end_position_on_the_genomic_accession=6900624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420864, CurrentID=0, Status='live', Symbol='LOC100420864', Aliases='', description='chromosome 11 open reading frame 58 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125159973, end_position_on_the_genomic_accession=125160545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150786, CurrentID=0, Status='live', Symbol='WTH3DI', Aliases='', description='RAB6C-like', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131360492, end_position_on_the_genomic_accession=131364158, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420869, CurrentID=0, Status='live', Symbol='LOC100420869', Aliases='', description='acetyl-CoA acyltransferase 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48775996, end_position_on_the_genomic_accession=48776970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420872, CurrentID=0, Status='live', Symbol='LOC100420872', Aliases='', description='ribosome binding factor A (putative) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95720243, end_position_on_the_genomic_accession=95721144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420875, CurrentID=0, Status='live', Symbol='FAM204BP', Aliases='', description='family with sequence similarity 204 member B, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197746755, end_position_on_the_genomic_accession=197747451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420877, CurrentID=0, Status='live', Symbol='SLC25A39P1', Aliases='', description='solute carrier family 25, member 39 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202795952, end_position_on_the_genomic_accession=202797291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420878, CurrentID=0, Status='live', Symbol='LOC100420878', Aliases='', description='GABA(A) receptor-associated protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205625483, end_position_on_the_genomic_accession=205626150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420879, CurrentID=0, Status='live', Symbol='LOC100420879', Aliases='', description='chromosome 2 open reading frame 27B pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242890191, end_position_on_the_genomic_accession=242890803, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420880, CurrentID=0, Status='live', Symbol='LOC100420880', Aliases='', description='primase, DNA, polypeptide 1 (49kDa) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247042758, end_position_on_the_genomic_accession=247044167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420881, CurrentID=0, Status='live', Symbol='FGFR3P6', Aliases='', description='fibroblast growth factor receptor 3 pseudogene 6', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247119734, end_position_on_the_genomic_accession=247120282, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271146, CurrentID=0, Status='live', Symbol='RPL7P54', Aliases='RPL7_31_1789', description='ribosomal protein L7 pseudogene 54', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78763142, end_position_on_the_genomic_accession=78763974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26844, CurrentID=0, Status='live', Symbol='RNU3P2', Aliases='U3, U3.2', description='RNA, U3 small nucleolar pseudogene 2', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20614856, end_position_on_the_genomic_accession=20614926, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420885, CurrentID=0, Status='live', Symbol='LOC100420885', Aliases='', description='suppressor of cytokine signaling 2 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49817168, end_position_on_the_genomic_accession=49817712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420886, CurrentID=0, Status='live', Symbol='RNFT1P1', Aliases='', description='ring finger protein, transmembrane 1 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19984123, end_position_on_the_genomic_accession=19985245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420887, CurrentID=0, Status='live', Symbol='LOC100420887', Aliases='', description='UPF0607 protein ENSP00000381418-like pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68528214, end_position_on_the_genomic_accession=68529439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420888, CurrentID=0, Status='live', Symbol='LOC100420888', Aliases='', description='torsin family 1, member B (torsin B) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74618954, end_position_on_the_genomic_accession=74620859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420889, CurrentID=0, Status='live', Symbol='LOC100420889', Aliases='', description='anaphase promoting complex subunit 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87118534, end_position_on_the_genomic_accession=87120980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420890, CurrentID=0, Status='live', Symbol='LOC100420890', Aliases='', description='anaphase promoting complex subunit 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87985727, end_position_on_the_genomic_accession=87992864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420892, CurrentID=0, Status='live', Symbol='LOC100420892', Aliases='', description='chromosome 14 open reading frame 166 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111940118, end_position_on_the_genomic_accession=111941055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420893, CurrentID=0, Status='live', Symbol='LOC100420893', Aliases='', description='Down syndrome critical region 3 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206686517, end_position_on_the_genomic_accession=206687363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420895, CurrentID=0, Status='live', Symbol='LOC100420895', Aliases='', description='fibrous sheath-interacting protein 2 pseudogene', other_designations='fibrous sheath interacting protein 2 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185833589, end_position_on_the_genomic_accession=185835116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420896, CurrentID=0, Status='live', Symbol='LOC100420896', Aliases='', description='UPF0607 protein ENSP00000381418-like pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92104225, end_position_on_the_genomic_accession=92105451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420897, CurrentID=0, Status='live', Symbol='LOC100420897', Aliases='', description="5'-aminolevulinate synthase 1 pseudogene", other_designations='aminolevulinate, delta-, synthase 1 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42929851, end_position_on_the_genomic_accession=42930404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420898, CurrentID=0, Status='live', Symbol='KNOP1P2', Aliases='', description='lysine-rich nucleolar protein 1 pseudogene 2', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45880950, end_position_on_the_genomic_accession=45882268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420899, CurrentID=0, Status='live', Symbol='LOC100420899', Aliases='', description='mitochondrial ribosomal protein S25 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64442512, end_position_on_the_genomic_accession=64443319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420900, CurrentID=0, Status='live', Symbol='LOC100420900', Aliases='', description='family with sequence similarity 213, member A pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78130333, end_position_on_the_genomic_accession=78130995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420901, CurrentID=0, Status='live', Symbol='HAUS8P1', Aliases='', description='HAUS augmin like complex subunit 8 pseudogene 1', other_designations='HAUS augmin-like complex, subunit 8 pseudogene 1', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114097652, end_position_on_the_genomic_accession=114098223, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420902, CurrentID=0, Status='live', Symbol='LOC100420902', Aliases='', description='exocyst complex component 3-like 4 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45234195, end_position_on_the_genomic_accession=45234809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420904, CurrentID=0, Status='live', Symbol='LOC100420904', Aliases='', description='SIGLEC family like 1 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51282478, end_position_on_the_genomic_accession=51283025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373271, CurrentID=0, Status='live', Symbol='LOC105373271', Aliases='', description='uncharacterized LOC105373271', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246582936, end_position_on_the_genomic_accession=246609340, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100420907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420907, CurrentID=0, Status='live', Symbol='LOC100420907', Aliases='', description='eva-1 homolog C (C. elegans) pseudogene', other_designations='family with sequence similarity 176, member C pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67751730, end_position_on_the_genomic_accession=67756979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420909, CurrentID=0, Status='live', Symbol='LOC100420909', Aliases='', description='transmembrane protein 251 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91030990, end_position_on_the_genomic_accession=91031491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420910, CurrentID=0, Status='live', Symbol='PRMT1P1', Aliases='', description='protein arginine methyltransferase 1 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177265365, end_position_on_the_genomic_accession=177266638, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420911, CurrentID=0, Status='live', Symbol='LOC100420911', Aliases='', description='chromosome 14 open reading frame 166 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95440056, end_position_on_the_genomic_accession=95440848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420916, CurrentID=0, Status='live', Symbol='FAM204CP', Aliases='', description='family with sequence similarity 204 member C, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75688554, end_position_on_the_genomic_accession=75689197, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420919, CurrentID=0, Status='live', Symbol='KATNBL1P1', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28067587, end_position_on_the_genomic_accession=28068343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420920, CurrentID=0, Status='live', Symbol='ZDHHC4P1', Aliases='', description='zinc finger, DHHC-type containing 4 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42933677, end_position_on_the_genomic_accession=42934932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420921, CurrentID=0, Status='live', Symbol='AMMECR1LP1', Aliases='', description='AMMECR1 like pseudogene 1', other_designations='AMME chromosomal region gene 1-like pseudogene 1|AMMECR1-like pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96573642, end_position_on_the_genomic_accession=96574503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420924, CurrentID=0, Status='live', Symbol='RNF6P1', Aliases='', description='ring finger protein (C3H2C3 type) 6 pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39373763, end_position_on_the_genomic_accession=39374421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420925, CurrentID=0, Status='live', Symbol='LOC100420925', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10136419, end_position_on_the_genomic_accession=10137004, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420926, CurrentID=0, Status='live', Symbol='ZEB2P1', Aliases='ZEB1P1', description='zinc finger E-box binding homeobox 2 pseudogene 1', other_designations='zinc finger E-box binding homeobox 1 pseudogene 1', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16360691, end_position_on_the_genomic_accession=16361826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420927, CurrentID=0, Status='live', Symbol='PRMT5P1', Aliases='', description='protein arginine methyltransferase 5 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147458727, end_position_on_the_genomic_accession=147459321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420928, CurrentID=0, Status='live', Symbol='LOC100420928', Aliases='', description='chromosome 15 open reading frame 43 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44914799, end_position_on_the_genomic_accession=44930622, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420929, CurrentID=0, Status='live', Symbol='LOC100420929', Aliases='', description='ribosome binding factor A (putative) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58966782, end_position_on_the_genomic_accession=58967637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420930, CurrentID=0, Status='live', Symbol='LOC100420930', Aliases='', description='SWI5 recombination repair homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72604353, end_position_on_the_genomic_accession=72604992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373272, CurrentID=0, Status='live', Symbol='LOC105373272', Aliases='', description='uncharacterized LOC105373272', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247172043, end_position_on_the_genomic_accession=247174480, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420939, CurrentID=0, Status='live', Symbol='LOC100420939', Aliases='', description='nucleosome assembly protein 1-like 5 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172680714, end_position_on_the_genomic_accession=172681221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420940, CurrentID=0, Status='live', Symbol='LOC100420940', Aliases='', description='transcription elongation regulator 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66710242, end_position_on_the_genomic_accession=66711537, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420941, CurrentID=0, Status='live', Symbol='BTN1A1P1', Aliases='', description='butyrophilin subfamily 1 member A1 pseudogene 1', other_designations='butyrophilin, subfamily 1, member A1 pseudogene 1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26477865, end_position_on_the_genomic_accession=26478521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271148, CurrentID=0, Status='live', Symbol='RPSAP59', Aliases='RPSA_27_1799', description='ribosomal protein SA pseudogene 59', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103206009, end_position_on_the_genomic_accession=103206899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150864, CurrentID=0, Status='live', Symbol='FAM117B', Aliases='ALS2CR13', description='family with sequence similarity 117 member B', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 13|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 13 protein|family with sequence similarity 117, member B', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202635178, end_position_on_the_genomic_accession=202769757, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100420946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420946, CurrentID=0, Status='live', Symbol='KATNBL1P3', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 3', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2593674, end_position_on_the_genomic_accession=2594913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420948, CurrentID=0, Status='live', Symbol='LOC100420948', Aliases='', description='ADP-ribosylation factor like GTPase 14 effector protein pseudogene', other_designations='ADP-ribosylation factor-like 14 effector protein pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60673513, end_position_on_the_genomic_accession=60674109, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420949, CurrentID=0, Status='live', Symbol='LOC100420949', Aliases='', description='neuroepithelial cell transforming 1 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51433243, end_position_on_the_genomic_accession=51434939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420950, CurrentID=0, Status='live', Symbol='MRPL37P1', Aliases='HsT4307', description='mitochondrial ribosomal protein L37 pseudogene 1', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58690356, end_position_on_the_genomic_accession=58691602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420951, CurrentID=0, Status='live', Symbol='LOC100420951', Aliases='', description='mitochondrial ribosomal protein L18 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11380859, end_position_on_the_genomic_accession=11381361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420952, CurrentID=0, Status='live', Symbol='FGFR3P5', Aliases='', description='fibroblast growth factor receptor 3 pseudogene 5', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35232458, end_position_on_the_genomic_accession=35233049, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420954, CurrentID=0, Status='live', Symbol='LOC100420954', Aliases='', description='GrpE-like 1, mitochondrial (E. coli) pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66789815, end_position_on_the_genomic_accession=66790343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420955, CurrentID=0, Status='live', Symbol='LOC100420955', Aliases='', description='scavenger receptor class B, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=98348500, end_position_on_the_genomic_accession=98349001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420957, CurrentID=0, Status='live', Symbol='LOC100420957', Aliases='', description='transcription elongation regulator 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=117399467, end_position_on_the_genomic_accession=117400081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420958, CurrentID=0, Status='live', Symbol='LOC100420958', Aliases='', description='transmembrane 9 superfamily member 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136961917, end_position_on_the_genomic_accession=137002495, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420963, CurrentID=0, Status='live', Symbol='FAM204DP', Aliases='', description='family with sequence similarity 204 member D, pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77173075, end_position_on_the_genomic_accession=77173652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420966, CurrentID=0, Status='live', Symbol='LOC100420966', Aliases='', description='chromosome 17 open reading frame 98 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72960595, end_position_on_the_genomic_accession=72961995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420967, CurrentID=0, Status='live', Symbol='LOC100420967', Aliases='', description='chromosome 2 open reading frame 27B pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22785668, end_position_on_the_genomic_accession=22786233, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420968, CurrentID=0, Status='live', Symbol='LYARP1', Aliases='', description='Ly1 antibody reactive pseudogene 1', other_designations='Ly1 antibody reactive homolog pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82268738, end_position_on_the_genomic_accession=82269680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420972, CurrentID=0, Status='live', Symbol='ZFYVE9P2', Aliases='', description='zinc finger, FYVE domain containing 9 pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17284285, end_position_on_the_genomic_accession=17284897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271149, CurrentID=0, Status='live', Symbol='RPSAP60', Aliases='RPSA_28_1820', description='ribosomal protein SA pseudogene 60', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152045244, end_position_on_the_genomic_accession=152046270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420975, CurrentID=0, Status='live', Symbol='GMCL1P2', Aliases='', description='germ cell-less, spermatogenesis associated 1 pseudogene 2', other_designations='germ cell-less homolog 1 pseudogene 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108367116, end_position_on_the_genomic_accession=108367940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420981, CurrentID=0, Status='live', Symbol='LOC100420981', Aliases='', description='nucleosome assembly protein 1-like 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32726383, end_position_on_the_genomic_accession=32727357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420982, CurrentID=0, Status='live', Symbol='LOC100420982', Aliases='', description='retinol dehydrogenase 16 (all-trans) pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56938450, end_position_on_the_genomic_accession=56941482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420983, CurrentID=0, Status='live', Symbol='LOC100420983', Aliases='', description='acyl-CoA synthetase medium-chain family member 6 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7237411, end_position_on_the_genomic_accession=7240988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420990, CurrentID=0, Status='live', Symbol='LOC100420990', Aliases='', description='aurora kinase A interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29515984, end_position_on_the_genomic_accession=29516504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  122867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122867, CurrentID=0, Status='live', Symbol='COX4I1P1', Aliases='COX4L1, COX4P1', description='cytochrome c oxidase subunit IV isoform 1 pseudogene 1', other_designations='cytochrome c oxidase subunit IV pseudogene 1', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61831739, end_position_on_the_genomic_accession=61832422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150921, CurrentID=0, Status='live', Symbol='TCF23', Aliases='OUT, TCF-23, bHLHa24', description='transcription factor 23', other_designations='class A basic helix-loop-helix protein 24|class II basic helix-loop-helix protein TCF23', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27148933, end_position_on_the_genomic_accession=27153576, orientation='plus', exon_count=4, OMIM=609635, no_hdr0=''),
  100421002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421002, CurrentID=0, Status='live', Symbol='HAVCR1P2', Aliases='', description='hepatitis A virus cellular receptor 1 pseudogene 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116858096, end_position_on_the_genomic_accession=116858969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271150, CurrentID=0, Status='live', Symbol='RPL17P19', Aliases='RPL17_11_483', description='ribosomal protein L17 pseudogene 19', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=61211652, end_position_on_the_genomic_accession=61212200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372933, CurrentID=0, Status='live', Symbol='LOC105372933', Aliases='', description='uncharacterized LOC105372933', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421006, CurrentID=0, Status='live', Symbol='LOC100421006', Aliases='', description='very low density lipoprotein receptor pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184072422, end_position_on_the_genomic_accession=184073036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150928, CurrentID=0, Status='live', Symbol='PTMAP5', Aliases='', description='prothymosin, alpha pseudogene 5', other_designations='gene sequence 150|novel protein similar to prothymosin, alpha (PTMA)', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81689896, end_position_on_the_genomic_accession=81691080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873454, CurrentID=0, Status='live', Symbol='RNA5SP192', Aliases='RN5S192', description='RNA, 5S ribosomal pseudogene 192', other_designations='RNA, 5S ribosomal 192', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132848577, end_position_on_the_genomic_accession=132848696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421010, CurrentID=0, Status='live', Symbol='LOC100421010', Aliases='', description='UPF0607 protein ENSP00000381418-like pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79860007, end_position_on_the_genomic_accession=79860995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  150935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150935, CurrentID=0, Status='live', Symbol='LOC150935', Aliases='', description='uncharacterized LOC150935', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239762860, end_position_on_the_genomic_accession=239800664, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421018, CurrentID=0, Status='live', Symbol='AMZ2P2', Aliases='', description='archaelysin family metallopeptidase 2 pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158725690, end_position_on_the_genomic_accession=158726808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481863, CurrentID=0, Status='live', Symbol='RN7SL845P', Aliases='', description='RNA, 7SL, cytoplasmic 845, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421023, CurrentID=0, Status='live', Symbol='LOC100421023', Aliases='', description='nuclear transcription factor Y subunit gamma pseudogene', other_designations='nuclear transcription factor Y, gamma pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13583844, end_position_on_the_genomic_accession=13584536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150946, CurrentID=0, Status='live', Symbol='GAREM2', Aliases='FAM59B, GAREML', description='GRB2 associated regulator of MAPK1 subtype 2', other_designations='GRB2 associated regulator of MAPK1 2|GRB2 associated, regulator of MAPK1-like|GRB2-associated and regulator of MAPK1-like|family with sequence similarity 59, member B|protein FAM59B', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26173091, end_position_on_the_genomic_accession=26202529, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  150947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150947, CurrentID=0, Status='live', Symbol='PPIL1P1', Aliases='', description='peptidylprolyl isomerase like 1 pseudogene 1', other_designations='peptidylprolyl isomerase (cyclophilin)-like 1 pseudogene 1', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26161828, end_position_on_the_genomic_accession=26163859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421028, CurrentID=0, Status='live', Symbol='LOC100421028', Aliases='', description='gap junction protein alpha 4 pseudogene', other_designations='gap junction protein, alpha 4, 37kDa pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29197074, end_position_on_the_genomic_accession=29197789, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421029, CurrentID=0, Status='live', Symbol='LOC100421029', Aliases='', description='transcription elongation regulator 1 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17094353, end_position_on_the_genomic_accession=17094927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421031, CurrentID=0, Status='live', Symbol='LOC100421031', Aliases='', description='carbonic anhydrase VA, mitochondrial pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30336400, end_position_on_the_genomic_accession=30343359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271151, CurrentID=0, Status='live', Symbol='RPL17P21', Aliases='RPL17_12_561', description='ribosomal protein L17 pseudogene 21', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52903908, end_position_on_the_genomic_accession=52904478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421039, CurrentID=0, Status='live', Symbol='LOC100421039', Aliases='', description='NLR family, pyrin domain containing 7 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131492282, end_position_on_the_genomic_accession=131492887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150962, CurrentID=0, Status='live', Symbol='PUS10', Aliases='CCDC139, DOBI', description='pseudouridylate synthase 10', other_designations='coiled-coil domain containing 139|coiled-coil domain-containing protein 139|pseudouridine synthase 10|psi55 synthase|tRNA pseudouridine 55 synthase|tRNA pseudouridylate synthase|tRNA-uridine isomerase', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60938352, end_position_on_the_genomic_accession=61020031, orientation='minus', exon_count=27, OMIM=612787, no_hdr0=''),
  100421043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421043, CurrentID=0, Status='live', Symbol='LOC100421043', Aliases='', description='tigger transposable element derived 7 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135297054, end_position_on_the_genomic_accession=135297792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421045, CurrentID=0, Status='live', Symbol='BCAS2P2', Aliases='', description='breast carcinoma amplified sequence 2 pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100393036, end_position_on_the_genomic_accession=100393658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421046, CurrentID=0, Status='live', Symbol='SOD2P1', Aliases='', description='superoxide dismutase 2, mitochondrial pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103100140, end_position_on_the_genomic_accession=103100707, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  150967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150967, CurrentID=0, Status='live', Symbol='PKI55', Aliases='', description='DKFZp434H1419', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216216889, end_position_on_the_genomic_accession=216220192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  450208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450208, CurrentID=0, Status='live', Symbol='LCEP1', Aliases='', description='late cornified envelope pseudogene 1', other_designations='small proline rich-like (epidermal differentiation complex) 1B pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152744299, end_position_on_the_genomic_accession=152744606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421053, CurrentID=0, Status='live', Symbol='RIPPLY3P1', Aliases='DSCR6P1', description='RIPPLY3 pseudogene 1', other_designations='DSCR6 pseudogene 1|Down syndrome critical region gene 6 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38973528, end_position_on_the_genomic_accession=38974030, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421055, CurrentID=0, Status='live', Symbol='ZMYND19P1', Aliases='', description='zinc finger, MYND-type containing 19 pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77696769, end_position_on_the_genomic_accession=77697364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373276, CurrentID=0, Status='live', Symbol='LOC105373276', Aliases='', description='uncharacterized LOC105373276', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421057, CurrentID=0, Status='live', Symbol='GSTM3P2', Aliases='', description='glutathione S-transferase mu 3 pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159055995, end_position_on_the_genomic_accession=159056988, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421058, CurrentID=0, Status='live', Symbol='PABPC1P11', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 11', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17589463, end_position_on_the_genomic_accession=17590696, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421061, CurrentID=0, Status='live', Symbol='LOC100421061', Aliases='', description='hepatitis A virus cellular receptor 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82468458, end_position_on_the_genomic_accession=82469221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271152, CurrentID=0, Status='live', Symbol='RPL17P23', Aliases='RPL17_15_751', description='ribosomal protein L17 pseudogene 23', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127435614, end_position_on_the_genomic_accession=127436220, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421064, CurrentID=0, Status='live', Symbol='LOC100421064', Aliases='', description='sushi, nidogen and EGF-like domains 1 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35075798, end_position_on_the_genomic_accession=35076483, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421071, CurrentID=0, Status='live', Symbol='LOC100421071', Aliases='', description='zinc finger, CCHC domain containing 10 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29848481, end_position_on_the_genomic_accession=29849222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421072, CurrentID=0, Status='live', Symbol='COQ10BP2', Aliases='', description='coenzyme Q10B pseudogene 2', other_designations='coenzyme Q10 homolog B pseudogene 2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83279920, end_position_on_the_genomic_accession=83280643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421074, CurrentID=0, Status='live', Symbol='LOC100421074', Aliases='', description='ZFP14 zinc finger protein pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140764141, end_position_on_the_genomic_accession=140764955, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421075, CurrentID=0, Status='live', Symbol='RCC2P2', Aliases='', description='regulator of chromosome condensation 2 pseudogene 2', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19986588, end_position_on_the_genomic_accession=19987873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  450209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450209, CurrentID=0, Status='live', Symbol='LCEP2', Aliases='', description='late cornified envelope pseudogene 2', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152737518, end_position_on_the_genomic_accession=152737868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421087, CurrentID=0, Status='live', Symbol='LOC100421087', Aliases='', description='zinc finger, CCHC domain containing 9 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98466835, end_position_on_the_genomic_accession=98467412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151009, CurrentID=0, Status='live', Symbol='LINC01106', Aliases='', description='long intergenic non-protein coding RNA 1106', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110375109, end_position_on_the_genomic_accession=110384536, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421090, CurrentID=0, Status='live', Symbol='RCC2P4', Aliases='', description='regulator of chromosome condensation 2 pseudogene 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126766437, end_position_on_the_genomic_accession=126767475, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151011, CurrentID=0, Status='live', Symbol='SEPT10', Aliases='', description='septin 10', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109542797, end_position_on_the_genomic_accession=109614206, orientation='minus', exon_count=14, OMIM=611737, no_hdr0=''),
  100421092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421092, CurrentID=0, Status='live', Symbol='LOC100421092', Aliases='', description='esterase D/formylglutathione hydrolase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17381487, end_position_on_the_genomic_accession=17382027, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271153, CurrentID=0, Status='live', Symbol='RPL17P24', Aliases='RPL17_16_771', description='ribosomal protein L17 pseudogene 24', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154216894, end_position_on_the_genomic_accession=154217524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421094, CurrentID=0, Status='live', Symbol='LOC100421094', Aliases='', description='ubiquitin specific peptidase 17-like family member 2 pseudogene', other_designations='ubiquitin carboxyl-terminal hydrolase 17-like pseudogene|ubiquitin specific peptidase 17-like 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12374354, end_position_on_the_genomic_accession=12375546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421096, CurrentID=0, Status='live', Symbol='NCAPGP1', Aliases='', description='non-SMC condensin I complex subunit G pseudogene 1', other_designations='non-SMC condensin I complex, subunit G pseudogene 1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120635570, end_position_on_the_genomic_accession=120638165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421097, CurrentID=0, Status='live', Symbol='ALG1L12P', Aliases='', description='asparagine-linked glycosylation 1-like 12, pseudogene', other_designations='asparagine-linked glycosylation 1, beta-1,4-mannosyltransferase homolog pseudogene|asparagine-linked glycosylation 1-like pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12421019, end_position_on_the_genomic_accession=12426274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421100, CurrentID=0, Status='live', Symbol='LOC100421100', Aliases='', description='RNA terminal phosphate cyclase-like 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29775718, end_position_on_the_genomic_accession=29776692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377642, CurrentID=0, Status='live', Symbol='LOC105377642', Aliases='', description='uncharacterized LOC105377642', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37203929, end_position_on_the_genomic_accession=37209547, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100290036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100290036, CurrentID=0, Status='live', Symbol='LOC100290036', Aliases='', description='immunglobulin heavy chain variable region', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421109, CurrentID=0, Status='live', Symbol='LOC100421109', Aliases='', description='karyopherin alpha 4 (importin alpha 3) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64305046, end_position_on_the_genomic_accession=64306740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421110, CurrentID=0, Status='live', Symbol='LOC100421110', Aliases='', description='VENT homeobox pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26557675, end_position_on_the_genomic_accession=26558316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  450210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450210, CurrentID=0, Status='live', Symbol='LCEP4', Aliases='', description='late cornified envelope pseudogene 4', other_designations='small proline rich-like (epidermal differentiation complex) 1B pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152644393, end_position_on_the_genomic_accession=152644717, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421113, CurrentID=0, Status='live', Symbol='CROCCP4', Aliases='', description='ciliary rootlet coiled-coil, rootletin pseudogene 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16750235, end_position_on_the_genomic_accession=16750964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421114, CurrentID=0, Status='live', Symbol='CROCCP5', Aliases='', description='ciliary rootlet coiled-coil, rootletin pseudogene 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21434320, end_position_on_the_genomic_accession=21436826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421116, CurrentID=0, Status='live', Symbol='LOC100421116', Aliases='', description='trafficking protein, kinesin binding 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114353986, end_position_on_the_genomic_accession=114355159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421121, CurrentID=0, Status='live', Symbol='LOC100421121', Aliases='', description='small G protein signaling modulator 1 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20639072, end_position_on_the_genomic_accession=20650483, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421122, CurrentID=0, Status='live', Symbol='LOC100421122', Aliases='', description='zinc finger protein 106 homolog (mouse) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43680558, end_position_on_the_genomic_accession=43681607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421123, CurrentID=0, Status='live', Symbol='PGBD4P5', Aliases='', description='piggyBac transposable element derived 4 pseudogene 5', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88922108, end_position_on_the_genomic_accession=88922753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421125, CurrentID=0, Status='live', Symbol='SMC4P1', Aliases='', description='structural maintenance of chromosomes 4 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136544656, end_position_on_the_genomic_accession=136545539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421126, CurrentID=0, Status='live', Symbol='LOC100421126', Aliases='', description='sucrase-isomaltase (alpha-glucosidase) pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38994832, end_position_on_the_genomic_accession=38996984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421129, CurrentID=0, Status='live', Symbol='IFNL3P1', Aliases='IL28BP1', description='interferon, lambda 3 pseudogene 1', other_designations='interleukin 28B (interferon, lambda 3) pseudogene 1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39236604, end_position_on_the_genomic_accession=39237299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151050, CurrentID=0, Status='live', Symbol='KANSL1L', Aliases='C2orf67, MSL1v2', description='KAT8 regulatory NSL complex subunit 1 like', other_designations='male-specific lethal 1 homolog', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210020711, end_position_on_the_genomic_accession=210171344, orientation='minus', exon_count=23, OMIM=613833, no_hdr0=''),
  100421131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421131, CurrentID=0, Status='live', Symbol='RCC2P6', Aliases='', description='regulator of chromosome condensation 2 pseudogene 6', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62371966, end_position_on_the_genomic_accession=62373172, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421133, CurrentID=0, Status='live', Symbol='PGAM5P1', Aliases='', description='phosphoglycerate mutase family member 5 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109884609, end_position_on_the_genomic_accession=109885437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151056, CurrentID=0, Status='live', Symbol='PLB1', Aliases='PLB, PLB/LIP, hPLB', description='phospholipase B1', other_designations='lysophospholipase|phospholipase A2|phospholipase B|phospholipase B/lipase', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28488719, end_position_on_the_genomic_accession=28643788, orientation='plus', exon_count=78, OMIM=610179, no_hdr0=''),
  100421137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421137, CurrentID=0, Status='live', Symbol='DPPA2P1', Aliases='', description='developmental pluripotency associated 2 pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13230884, end_position_on_the_genomic_accession=13231448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131601, CurrentID=0, Status='live', Symbol='TPRA1', Aliases='GPR175, TMEM227, TPRA40', description='transmembrane protein, adipocyte asscociated 1', other_designations='G protein-coupled receptor 175|integral membrane protein GPR175|seven transmembrane domain orphan receptor|transmembrane domain protein regulated in adipocytes 40 kDa|transmembrane protein 227|transmembrane protein, adipocyte associated 1', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127573064, end_position_on_the_genomic_accession=127595059, orientation='minus', exon_count=16, OMIM=608336, no_hdr0=''),
  450211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450211, CurrentID=0, Status='live', Symbol='LCEP3', Aliases='', description='late cornified envelope pseudogene 3', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152656332, end_position_on_the_genomic_accession=152656511, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421142, CurrentID=0, Status='live', Symbol='LOC100421142', Aliases='', description='developmental pluripotency associated 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77350392, end_position_on_the_genomic_accession=77351178, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421143, CurrentID=0, Status='live', Symbol='ELL2P2', Aliases='', description='elongation factor, RNA polymerase II, 2 pseudogene 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132367769, end_position_on_the_genomic_accession=132368413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375791, CurrentID=0, Status='live', Symbol='LOC105375791', Aliases='', description='uncharacterized LOC105375791', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141512609, end_position_on_the_genomic_accession=141514215, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421146, CurrentID=0, Status='live', Symbol='PGBD4P4', Aliases='', description='piggyBac transposable element derived 4 pseudogene 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130494673, end_position_on_the_genomic_accession=130495279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421148, CurrentID=0, Status='live', Symbol='NCAPGP2', Aliases='', description='non-SMC condensin I complex subunit G pseudogene 2', other_designations='non-SMC condensin I complex, subunit G pseudogene 2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30005190, end_position_on_the_genomic_accession=30006439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421150, CurrentID=0, Status='live', Symbol='COQ10BP1', Aliases='', description='coenzyme Q10B pseudogene 1', other_designations='coenzyme Q10 homolog B pseudogene 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24660403, end_position_on_the_genomic_accession=24660991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271155, CurrentID=0, Status='live', Symbol='RPL17P30', Aliases='RPL17_20_898', description='ribosomal protein L17 pseudogene 30', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42101318, end_position_on_the_genomic_accession=42101840, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421154, CurrentID=0, Status='live', Symbol='LOC100421154', Aliases='', description='developmental pluripotency associated 4 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134510527, end_position_on_the_genomic_accession=134511401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270843, CurrentID=0, Status='live', Symbol='RPL6P15', Aliases='RPL6_3_613', description='ribosomal protein L6 pseudogene 15', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133172218, end_position_on_the_genomic_accession=133173042, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421158, CurrentID=0, Status='live', Symbol='LOC100421158', Aliases='', description='piggyBac transposable element derived 4 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49115535, end_position_on_the_genomic_accession=49116115, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421160, CurrentID=0, Status='live', Symbol='LOC100421160', Aliases='', description='structural maintenance of chromosomes 5 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14022684, end_position_on_the_genomic_accession=14023418, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421161, CurrentID=0, Status='live', Symbol='LOC100421161', Aliases='', description='ependymin related 1 pseudogene', other_designations='ependymin related protein 1 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25141397, end_position_on_the_genomic_accession=25141922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421165, CurrentID=0, Status='live', Symbol='LOC100421165', Aliases='', description='fermitin family member 3 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22277874, end_position_on_the_genomic_accession=22279406, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421166, CurrentID=0, Status='live', Symbol='LOC100421166', Aliases='', description='SEC24 homolog A, COPII coat complex component pseudogene', other_designations='SEC24 family member A pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69094256, end_position_on_the_genomic_accession=69096117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421167, CurrentID=0, Status='live', Symbol='SCML2P1', Aliases='', description='SCML2 pseudogene 1', other_designations='sex comb on midleg-like 2 pseudogene 1', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6926974, end_position_on_the_genomic_accession=6928215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103157000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103157000, CurrentID=0, Status='live', Symbol='PAUPAR', Aliases='', description='PAX6 upstream antisense RNA', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31826567, end_position_on_the_genomic_accession=31829303, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421171, CurrentID=0, Status='live', Symbol='LOC100421171', Aliases='', description='thyroid hormone receptor interactor 11 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25015978, end_position_on_the_genomic_accession=25022380, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421172, CurrentID=0, Status='live', Symbol='SCML2P2', Aliases='', description='SCML2 pseudogene 2', other_designations='sex comb on midleg-like 2 pseudogene 2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25069572, end_position_on_the_genomic_accession=25070301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421173, CurrentID=0, Status='live', Symbol='LOC100421173', Aliases='', description='thyroid hormone receptor interactor 11 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25093411, end_position_on_the_genomic_accession=25095180, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421174, CurrentID=0, Status='live', Symbol='LOC100421174', Aliases='', description='thyroid hormone receptor interactor 13 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53298688, end_position_on_the_genomic_accession=53299794, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421175, CurrentID=0, Status='live', Symbol='NCAPD2P1', Aliases='', description='non-SMC condensin I complex subunit D2 pseudogene 1', other_designations='non-SMC condensin I complex, subunit D2 pseudogene 1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34750253, end_position_on_the_genomic_accession=34752044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421177, CurrentID=0, Status='live', Symbol='TDGF1P1', Aliases='', description='teratocarcinoma-derived growth factor 1 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38395369, end_position_on_the_genomic_accession=38395895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421178, CurrentID=0, Status='live', Symbol='PGAM4P1', Aliases='', description='phosphoglycerate mutase family member 4 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54671986, end_position_on_the_genomic_accession=54672635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  544315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544315, CurrentID=0, Status='live', Symbol='COPD7', Aliases='', description='Chronic obstructive pulmonary disease QTL 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544316, CurrentID=0, Status='live', Symbol='COPD23', Aliases='', description='Chronic obstructive pulmonary disease QTL 23', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544317, CurrentID=0, Status='live', Symbol='AASTH1', Aliases='', description='Allergic/atopic asthma related QTL 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544318, CurrentID=0, Status='live', Symbol='OA19', Aliases='', description='Osteoarthritis QTL 19', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544319, CurrentID=0, Status='live', Symbol='BP7', Aliases='', description='Blood pressure QTL 7', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544320, CurrentID=0, Status='live', Symbol='OA18', Aliases='', description='Osteoarthritis QTL 18', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544321, CurrentID=0, Status='live', Symbol='OA13', Aliases='', description='Osteoarthritis QTL 13', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544322, CurrentID=0, Status='live', Symbol='OA28', Aliases='', description='Osteoarthritis QTL 28', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544323, CurrentID=0, Status='live', Symbol='OA20', Aliases='', description='Osteoarthritis QTL 20', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544324, CurrentID=0, Status='live', Symbol='OA23', Aliases='', description='Osteoarthritis QTL 23', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544325, CurrentID=0, Status='live', Symbol='OA21', Aliases='', description='Osteoarthritis QTL 21', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544326, CurrentID=0, Status='live', Symbol='BP8', Aliases='', description='Blood pressure QTL 8', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544327, CurrentID=0, Status='live', Symbol='OA10', Aliases='', description='Osteoarthritis QTL 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151112, CurrentID=0, Status='live', Symbol='ZSWIM2', Aliases='MEX, ZZZ2', description='zinc finger, SWIM-type containing 2', other_designations='MEKK1-related protein X|ZZ-type zinc finger-containing protein 2|zinc finger SWIM domain-containing protein 2|zinc finger, SWIM domain containing 2', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186827480, end_position_on_the_genomic_accession=186849170, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  544329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544329, CurrentID=0, Status='live', Symbol='OA17', Aliases='', description='Osteoarthritis QTL 17', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544330, CurrentID=0, Status='live', Symbol='BW1', Aliases='', description='Body weight QTL 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544331, CurrentID=0, Status='live', Symbol='OA14', Aliases='', description='Osteoarthritis QTL 14', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544332, CurrentID=0, Status='live', Symbol='OA22', Aliases='', description='Osteoarthritis QTL 22', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544333, CurrentID=0, Status='live', Symbol='OA12', Aliases='', description='Osteoarthritis QTL 12', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544334, CurrentID=0, Status='live', Symbol='OA27', Aliases='', description='Osteoarthritis QTL 27', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544335, CurrentID=0, Status='live', Symbol='OA15', Aliases='', description='Osteoarthritis QTL 15', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544336, CurrentID=0, Status='live', Symbol='BW17', Aliases='', description='Body weight QTL 17', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544337, CurrentID=0, Status='live', Symbol='BW16', Aliases='', description='Body weight QTL 16', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544338, CurrentID=0, Status='live', Symbol='BP29', Aliases='', description='Blood pressure QTL 29', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421204, CurrentID=0, Status='live', Symbol='LOC100421204', Aliases='', description='cyclic nucleotide gated channel alpha 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72551060, end_position_on_the_genomic_accession=72552383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151126, CurrentID=0, Status='live', Symbol='ZNF385B', Aliases='ZNF533', description='zinc finger protein 385B', other_designations='zinc finger protein 533', map_location='2q31.2-q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179441982, end_position_on_the_genomic_accession=179862321, orientation='minus', exon_count=22, OMIM=612344, no_hdr0=''),
  105373281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373281, CurrentID=0, Status='live', Symbol='LOC105373281', Aliases='', description='uncharacterized LOC105373281', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223677818, end_position_on_the_genomic_accession=223708362, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100533628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533628, CurrentID=0, Status='live', Symbol='LOC100533628', Aliases='', description='WD repeat domain 75 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=41296410, end_position_on_the_genomic_accession=41298686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421217, CurrentID=0, Status='live', Symbol='USP9YP5', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 5', other_designations='USP9Y pseudogene 5', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18490914, end_position_on_the_genomic_accession=18494295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421219, CurrentID=0, Status='live', Symbol='USP9YP8', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 8', other_designations='USP9Y pseudogene 8', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21677246, end_position_on_the_genomic_accession=21679223, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421220, CurrentID=0, Status='live', Symbol='USP9YP9', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 9', other_designations='USP9Y pseudogene 9', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25730010, end_position_on_the_genomic_accession=25733452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421222, CurrentID=0, Status='live', Symbol='USP9YP10', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 10', other_designations='USP9Y pseudogene 10', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18826107, end_position_on_the_genomic_accession=18829488, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421224, CurrentID=0, Status='live', Symbol='USP9YP11', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 11', other_designations='USP9Y pseudogene 11', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23936663, end_position_on_the_genomic_accession=23940102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421226, CurrentID=0, Status='live', Symbol='RABEPKP1', Aliases='', description='Rab9 effector protein with kelch motifs pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71296225, end_position_on_the_genomic_accession=71297232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421228, CurrentID=0, Status='live', Symbol='PGBD4P2', Aliases='', description='piggyBac transposable element derived 4 pseudogene 2', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44841925, end_position_on_the_genomic_accession=44842548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421232, CurrentID=0, Status='live', Symbol='GAPDHP56', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 56', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130503366, end_position_on_the_genomic_accession=130504341, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421235, CurrentID=0, Status='live', Symbol='LOC100421235', Aliases='', description='serine/arginine-rich splicing factor 9 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82691526, end_position_on_the_genomic_accession=82692452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421236, CurrentID=0, Status='live', Symbol='GAPDHP50', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 50', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=89047445, end_position_on_the_genomic_accession=89048432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373282, CurrentID=0, Status='live', Symbol='LOC105373282', Aliases='', description='uncharacterized LOC105373282', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79450726, end_position_on_the_genomic_accession=79453953, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421238, CurrentID=0, Status='live', Symbol='LOC100421238', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 5 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175675083, end_position_on_the_genomic_accession=175676795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421239, CurrentID=0, Status='live', Symbol='FANCD2P1', Aliases='', description='Fanconi anemia complementation group D2 pseudogene 1', other_designations='Fanconi anemia, complementation group D2 pseudogene 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9992874, end_position_on_the_genomic_accession=9994323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100380271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380271, CurrentID=0, Status='live', Symbol='RNY5P4', Aliases='', description='RNA, Ro-associated Y5 pseudogene 4', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129860420, end_position_on_the_genomic_accession=129860500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421244, CurrentID=0, Status='live', Symbol='LOC100421244', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421245, CurrentID=0, Status='live', Symbol='NIP7P3', Aliases='', description='NIP7, nucleolar pre-rRNA processing protein pseudogene 3', other_designations='nuclear import 7 homolog pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116137061, end_position_on_the_genomic_accession=116137572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421246, CurrentID=0, Status='live', Symbol='LOC100421246', Aliases='', description='D4, zinc and double PHD fingers family 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131184328, end_position_on_the_genomic_accession=131185486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421247, CurrentID=0, Status='live', Symbol='LOC100421247', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421248, CurrentID=0, Status='live', Symbol='LOC100421248', Aliases='', description='siah E3 ubiquitin protein ligase family member 3 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11015446, end_position_on_the_genomic_accession=11015983, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421250, CurrentID=0, Status='live', Symbol='LOC100421250', Aliases='', description='epithelial splicing regulatory protein 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134009008, end_position_on_the_genomic_accession=134009547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151171, CurrentID=0, Status='live', Symbol='LINC01107', Aliases='', description='long intergenic non-protein coding RNA 1107', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238510690, end_position_on_the_genomic_accession=238555499, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421253, CurrentID=0, Status='live', Symbol='LOC100421253', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151174, CurrentID=0, Status='live', Symbol='LOC151174', Aliases='', description='uncharacterized LOC151174', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238225113, end_position_on_the_genomic_accession=238231677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  151176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151176, CurrentID=0, Status='live', Symbol='FAM132B', Aliases='C1QTNF15, CTRP15', description='family with sequence similarity 132 member B', other_designations='complement C1q tumor necrosis factor-related protein 15|family with sequence similarity 132, member B|myonectin|protein FAM132B', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238159008, end_position_on_the_genomic_accession=238168890, orientation='plus', exon_count=8, OMIM=615099, no_hdr0=''),
  100421257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421257, CurrentID=0, Status='live', Symbol='LOC100421257', Aliases='', description='lysosomal-associated membrane protein 1 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71028781, end_position_on_the_genomic_accession=71033053, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375442, CurrentID=0, Status='live', Symbol='LOC105375442', Aliases='', description='uncharacterized LOC105375442', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107221776, end_position_on_the_genomic_accession=107235472, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421259, CurrentID=0, Status='live', Symbol='LOC100421259', Aliases='', description='zinc finger and BTB domain containing 14 pseudogene', other_designations='ZFP161 zinc finger protein pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46573664, end_position_on_the_genomic_accession=46575094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421261, CurrentID=0, Status='live', Symbol='LOC100421261', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421264, CurrentID=0, Status='live', Symbol='LOC100421264', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421265, CurrentID=0, Status='live', Symbol='LOC100421265', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79231, CurrentID=0, Status='live', Symbol='ZNF33CP', Aliases='BA393J16.4, ZNF33AP1', description='zinc finger protein 33C, pseudogene', other_designations='zinc finger protein 33A pseudogene 1|zinc finger pseudogene', map_location='10p11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37894593, end_position_on_the_genomic_accession=37896936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151188, CurrentID=0, Status='live', Symbol='ARL6IP6', Aliases='AIP-6, PFAAP1', description='ADP ribosylation factor like GTPase 6 interacting protein 6', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 6|ADP-ribosylation factor like GTPase 6 interacting protein 6|ADP-ribosylation factor-like 6 interacting protein 6|ADP-ribosylation-like factor 6 interacting protein 6|ARL-6-interacting protein 6|phosphonoformate immuno-associated protein 1|regulated by phosphonoformate', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152717893, end_position_on_the_genomic_accession=152761253, orientation='plus', exon_count=7, OMIM=616495, no_hdr0=''),
  100421269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421269, CurrentID=0, Status='live', Symbol='LOC100421269', Aliases='', description='phosphodiesterase 4D interacting protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21411462, end_position_on_the_genomic_accession=21411990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131180, CurrentID=0, Status='live', Symbol='LOC100131180', Aliases='', description='uncharacterized LOC100131180', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=32954489, end_position_on_the_genomic_accession=32958868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421271, CurrentID=0, Status='live', Symbol='GAPDHP51', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 51', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51707138, end_position_on_the_genomic_accession=51708146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421273, CurrentID=0, Status='live', Symbol='LOC100421273', Aliases='', description='F-box and WD repeat domain containing 11 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93384487, end_position_on_the_genomic_accession=93384998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151194, CurrentID=0, Status='live', Symbol='METTL21A', Aliases='FAM119A, HCA557b, HSPA-KMT', description='methyltransferase like 21A', other_designations='HSPA lysine methyltransferase|family with sequence similarity 119, member A|heat shock protein 70kDa lysine (K) methyltransferase|hepatocellular carcinoma-associated antigen 557b|methyltransferase-like protein 21A|protein-lysine methyltransferase METTL21A', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207580635, end_position_on_the_genomic_accession=207625928, orientation='minus', exon_count=12, OMIM=615257, no_hdr0=''),
  151195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151195, CurrentID=0, Status='live', Symbol='CCNYL1', Aliases='', description='cyclin Y like 1', other_designations='cyclin Y-like 1', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207711540, end_position_on_the_genomic_accession=207756174, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100421279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421279, CurrentID=0, Status='live', Symbol='LOC100421279', Aliases='', description='phosphodiesterase 4D interacting protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16856711, end_position_on_the_genomic_accession=16859283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  544416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544416, CurrentID=0, Status='live', Symbol='COPD19', Aliases='', description='Chronic obstructive pulmonary disease QTL 19', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544417, CurrentID=0, Status='live', Symbol='COPD13', Aliases='', description='Chronic obstructive pulmonary disease QTL 13', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544418, CurrentID=0, Status='live', Symbol='COPD2', Aliases='', description='Chronic obstructive pulmonary disease QTL 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544419, CurrentID=0, Status='live', Symbol='COPD14', Aliases='', description='Chronic obstructive pulmonary disease QTL 14', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544420, CurrentID=0, Status='live', Symbol='COPD3', Aliases='', description='Chronic obstructive pulmonary disease QTL 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544421, CurrentID=0, Status='live', Symbol='COPD11', Aliases='', description='Chronic obstructive pulmonary disease QTL 11', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544422, CurrentID=0, Status='live', Symbol='COPD25', Aliases='', description='Chronic obstructive pulmonary disease QTL 25', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544423, CurrentID=0, Status='live', Symbol='AASTH6', Aliases='', description='Allergic/atopic asthma related QTL 6', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544424, CurrentID=0, Status='live', Symbol='COPD16', Aliases='', description='Chronic obstructive pulmonary disease QTL 16', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544425, CurrentID=0, Status='live', Symbol='AASTH7', Aliases='', description='Allergic/atopic asthma related QTL 7', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544426, CurrentID=0, Status='live', Symbol='COPD18', Aliases='', description='Chronic obstructive pulmonary disease QTL 18', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544427, CurrentID=0, Status='live', Symbol='COPD21', Aliases='', description='Chronic obstructive pulmonary disease QTL 21', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544428, CurrentID=0, Status='live', Symbol='COPD4', Aliases='', description='Chronic obstructive pulmonary disease QTL 4', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544429, CurrentID=0, Status='live', Symbol='COPD8', Aliases='', description='Chronic obstructive pulmonary disease QTL 8', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544430, CurrentID=0, Status='live', Symbol='COPD22', Aliases='', description='Chronic obstructive pulmonary disease QTL 22', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  134429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134429, CurrentID=0, Status='live', Symbol='STARD4', Aliases='', description='StAR related lipid transfer domain containing 4', other_designations='START domain containing 4, sterol regulated|START domain-containing protein 4|StAR-related lipid transfer (START) domain containing 4|StAR-related lipid transfer domain containing 4', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111496033, end_position_on_the_genomic_accession=111512536, orientation='minus', exon_count=7, OMIM=607049, no_hdr0=''),
  544432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544432, CurrentID=0, Status='live', Symbol='COPD12', Aliases='', description='Chronic obstructive pulmonary disease QTL 12', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544433, CurrentID=0, Status='live', Symbol='COPD17', Aliases='', description='Chronic obstructive pulmonary disease QTL 17', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544434, CurrentID=0, Status='live', Symbol='AASTH5', Aliases='', description='Allergic/atopic asthma related QTL 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544435, CurrentID=0, Status='live', Symbol='COPD28', Aliases='', description='Chronic obstructive pulmonary disease QTL 28', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544436, CurrentID=0, Status='live', Symbol='COPD20', Aliases='', description='Chronic obstructive pulmonary disease QTL 20', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  134430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134430, CurrentID=0, Status='live', Symbol='WDR36', Aliases='GLC1G, TA-WDRP, TAWDRP, UTP21', description='WD repeat domain 36', other_designations='T-cell activation WD repeat protein|T-cell activation WD repeat-containing protein', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111092172, end_position_on_the_genomic_accession=111130502, orientation='plus', exon_count=23, OMIM=609669, no_hdr0=''),
  544438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544438, CurrentID=0, Status='live', Symbol='COPD9', Aliases='', description='Chronic obstructive pulmonary disease QTL 9', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544439, CurrentID=0, Status='live', Symbol='COPD27', Aliases='', description='Chronic obstructive pulmonary disease QTL 27', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544440, CurrentID=0, Status='live', Symbol='AASTH4', Aliases='', description='Allergic/atopic asthma related QTL 4', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544441, CurrentID=0, Status='live', Symbol='AASTH3', Aliases='', description='Allergic/atopic asthma related QTL 3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544442, CurrentID=0, Status='live', Symbol='COPD10', Aliases='', description='Chronic obstructive pulmonary disease QTL 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544443, CurrentID=0, Status='live', Symbol='COPD29', Aliases='', description='Chronic obstructive pulmonary disease QTL 29', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544444, CurrentID=0, Status='live', Symbol='COPD1', Aliases='', description='Chronic obstructive pulmonary disease QTL 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544445, CurrentID=0, Status='live', Symbol='COPD6', Aliases='', description='Chronic obstructive pulmonary disease QTL 6', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151230, CurrentID=0, Status='live', Symbol='KLHL23', Aliases='DITHP', description='kelch like family member 23', other_designations='kelch-like 23|kelch-like family member 23', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169733846, end_position_on_the_genomic_accession=169751886, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  544447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544447, CurrentID=0, Status='live', Symbol='COPD5', Aliases='', description='Chronic obstructive pulmonary disease QTL 5', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544448, CurrentID=0, Status='live', Symbol='COPD24', Aliases='', description='Chronic obstructive pulmonary disease QTL 24', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421313, CurrentID=0, Status='live', Symbol='GAPDHP52', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 52', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21358376, end_position_on_the_genomic_accession=21359311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151234, CurrentID=0, Status='live', Symbol='SULT1C2P1', Aliases='SULT1C1P', description='sulfotransferase family 1C member 2 pseudogene 1', other_designations='sulfotransferase family, cytosolic, 1C, member 2 pseudogene 1', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108322238, end_position_on_the_genomic_accession=108353799, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421317, CurrentID=0, Status='live', Symbol='LOC100421317', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421318, CurrentID=0, Status='live', Symbol='GAPDHP53', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 53', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24445412, end_position_on_the_genomic_accession=24446314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421319, CurrentID=0, Status='live', Symbol='GAPDHP54', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 54', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46997711, end_position_on_the_genomic_accession=46998477, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421321, CurrentID=0, Status='live', Symbol='LOC100421321', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151242, CurrentID=0, Status='live', Symbol='PPP1R1C', Aliases='IPP5', description='protein phosphatase 1 regulatory inhibitor subunit 1C', other_designations='PP1 inhibitor protein 5|inhibitor-5 of protein phosphatase 1|protein phosphatase 1 regulatory subunit 1A|protein phosphatase 1, regulatory (inhibitor) subunit 1C', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181954241, end_position_on_the_genomic_accession=182131382, orientation='plus', exon_count=8, OMIM=613240, no_hdr0=''),
  100421323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421323, CurrentID=0, Status='live', Symbol='HNRNPRP3', Aliases='', description='heterogeneous nuclear ribonucleoprotein R pseudogene 3', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37868294, end_position_on_the_genomic_accession=37869071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421324, CurrentID=0, Status='live', Symbol='GAPDHP55', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 55', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44063511, end_position_on_the_genomic_accession=44064313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271717, CurrentID=0, Status='live', Symbol='LOC100271717', Aliases='', description='natural cytotoxicity triggering receptor 1 pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182955190, end_position_on_the_genomic_accession=182955798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151246, CurrentID=0, Status='live', Symbol='SGOL2', Aliases='SGO2, TRIPIN', description='shugoshin-like 2 (S. pombe)', other_designations='shugoshin-2', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200524802, end_position_on_the_genomic_accession=200584095, orientation='plus', exon_count=13, OMIM=612425, no_hdr0=''),
  105373285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373285, CurrentID=0, Status='live', Symbol='LOC105373285', Aliases='', description='uncharacterized LOC105373285', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81476861, end_position_on_the_genomic_accession=81498238, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421328, CurrentID=0, Status='live', Symbol='DDX50P2', Aliases='RH_II/GuBp2', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 50 pseudogene 2', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 50 pseudogene|RH_II/GuB pseudogene 2', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154351916, end_position_on_the_genomic_accession=154354874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421330, CurrentID=0, Status='live', Symbol='LOC100421330', Aliases='', description='lysophosphatidylcholine acyltransferase 3 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155004792, end_position_on_the_genomic_accession=155005328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421331, CurrentID=0, Status='live', Symbol='SAMD11P1', Aliases='', description='sterile alpha motif domain containing 11 pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2613282, end_position_on_the_genomic_accession=2613969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421332, CurrentID=0, Status='live', Symbol='LOC100421332', Aliases='', description='zinc finger with KRAB and SCAN domains 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5061768, end_position_on_the_genomic_accession=5062305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271161, CurrentID=0, Status='live', Symbol='RPL21P32', Aliases='RPL21_13_305', description='ribosomal protein L21 pseudogene 32', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185956587, end_position_on_the_genomic_accession=185957052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151254, CurrentID=0, Status='live', Symbol='ALS2CR11', Aliases='', description='amyotrophic lateral sclerosis 2 chromosome region candidate 11', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 11|amyotrophic lateral sclerosis 2 chromosome region, candidate 11', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201487421, end_position_on_the_genomic_accession=201619183, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100421335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421335, CurrentID=0, Status='live', Symbol='LOC100421335', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421336, CurrentID=0, Status='live', Symbol='LOC100421336', Aliases='', description='RAB3A interacting protein (rabin3)-like 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33379949, end_position_on_the_genomic_accession=33380866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151258, CurrentID=0, Status='live', Symbol='SLC38A11', Aliases='AVT2', description='solute carrier family 38 member 11', other_designations='solute carrier family 38, member 11', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164897745, end_position_on_the_genomic_accession=164955560, orientation='minus', exon_count=16, OMIM=616526, no_hdr0=''),
  100421340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421340, CurrentID=0, Status='live', Symbol='FNDC3CP', Aliases='', description='fibronectin type III domain containing 3C, pseudogene', other_designations='fibronectin type III domain containing 3A pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78165696, end_position_on_the_genomic_accession=78170137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421341, CurrentID=0, Status='live', Symbol='UBE2V2P3', Aliases='', description='ubiquitin conjugating enzyme E2 variant 2 pseudogene 3', other_designations='ubiquitin-conjugating enzyme E2 variant 2 pseudogene 3', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11278414, end_position_on_the_genomic_accession=11279275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421343, CurrentID=0, Status='live', Symbol='LOC100421343', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 19B pseudogene', other_designations='DEAD (Asp-Glu-Ala-As) box polypeptide 19B pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=188242087, end_position_on_the_genomic_accession=188242705, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421344, CurrentID=0, Status='live', Symbol='LOC100421344', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 55 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242383345, end_position_on_the_genomic_accession=242384796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421347, CurrentID=0, Status='live', Symbol='LOC100421347', Aliases='', description='golgi reassembly stacking protein 2, 55kDa pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69251818, end_position_on_the_genomic_accession=69252374, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421349, CurrentID=0, Status='live', Symbol='LOC100421349', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216175047, end_position_on_the_genomic_accession=216175938, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421350, CurrentID=0, Status='live', Symbol='TCEA1P4', Aliases='', description='transcription elongation factor A (SII), 1 pseudogene 4', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32975723, end_position_on_the_genomic_accession=32980441, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421351, CurrentID=0, Status='live', Symbol='HNRNPA1P14', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 14', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88410332, end_position_on_the_genomic_accession=88411520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421353, CurrentID=0, Status='live', Symbol='HNRNPA1P15', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 15', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125595593, end_position_on_the_genomic_accession=125596825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421354, CurrentID=0, Status='live', Symbol='HNRNPRP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein R pseudogene 2', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29668886, end_position_on_the_genomic_accession=29669628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151276, CurrentID=0, Status='live', Symbol='TXNP5', Aliases='', description='thioredoxin pseudogene 5', other_designations='', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149068468, end_position_on_the_genomic_accession=149068927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421357, CurrentID=0, Status='live', Symbol='HNRNPA3P7', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 7', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154337547, end_position_on_the_genomic_accession=154338431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151278, CurrentID=0, Status='live', Symbol='CCDC140', Aliases='', description='coiled-coil domain containing 140', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222298147, end_position_on_the_genomic_accession=222305217, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421361, CurrentID=0, Status='live', Symbol='LOC100421361', Aliases='', description='isopentenyl-diphosphate delta isomerase 1 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4735669, end_position_on_the_genomic_accession=4736298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421362, CurrentID=0, Status='live', Symbol='LOC100421362', Aliases='', description='hydroxysteroid (17-beta) dehydrogenase 12 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21488215, end_position_on_the_genomic_accession=21488736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271162, CurrentID=0, Status='live', Symbol='RPL21P42', Aliases='RPL21_13_421', description='ribosomal protein L21 pseudogene 42', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154023796, end_position_on_the_genomic_accession=154024337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421364, CurrentID=0, Status='live', Symbol='LOC100421364', Aliases='', description='fibroblast growth factor binding protein 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16307457, end_position_on_the_genomic_accession=16308063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421368, CurrentID=0, Status='live', Symbol='LOC100421368', Aliases='', description='zinc finger and SCAN domain containing 5A pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89936483, end_position_on_the_genomic_accession=89937006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421369, CurrentID=0, Status='live', Symbol='DDX20P1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 20 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=147830, end_position_on_the_genomic_accession=148796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421371, CurrentID=0, Status='live', Symbol='DDX10P1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 10 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30919008, end_position_on_the_genomic_accession=30921130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421372, CurrentID=0, Status='live', Symbol='LOC100421372', Aliases='', description='zinc finger and SCAN domain containing 29 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14842306, end_position_on_the_genomic_accession=14842828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421374, CurrentID=0, Status='live', Symbol='MPRIPP1', Aliases='', description='myosin phosphatase Rho interacting protein pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44577730, end_position_on_the_genomic_accession=44581070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151295, CurrentID=0, Status='live', Symbol='SLC23A3', Aliases='E2BP3, SVCT3, Yspl1', description='solute carrier family 23 member 3', other_designations='E2-binding protein 3|HPC E2-binding protein 3|Na(+)/L-ascorbic acid transporter 3|sodium-dependent vitamin C transporter 3|solute carrier family 23 (nucleobase transporters), member 3|solute carrier family 23, member 3', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219161459, end_position_on_the_genomic_accession=219170095, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100421376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421376, CurrentID=0, Status='live', Symbol='HNRNPA1P17', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 17', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108325508, end_position_on_the_genomic_accession=108326814, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421378, CurrentID=0, Status='live', Symbol='HNRNPA1P77', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 77', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44844647, end_position_on_the_genomic_accession=44845602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421379, CurrentID=0, Status='live', Symbol='HNRNPA1P37', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 37', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5608830, end_position_on_the_genomic_accession=5610282, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151300, CurrentID=0, Status='live', Symbol='LINC00608', Aliases='', description='long intergenic non-protein coding RNA 608', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218976284, end_position_on_the_genomic_accession=218977922, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421384, CurrentID=0, Status='live', Symbol='HNRNPA1P38', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1pseudogene 38', other_designations='heterogeneous nuclear ribonucleoprotein A1 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141459017, end_position_on_the_genomic_accession=141460228, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421385, CurrentID=0, Status='live', Symbol='LOC100421385', Aliases='', description='testin LIM domain protein pseudogene', other_designations='testis derived transcript (3 LIM domains) pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60525625, end_position_on_the_genomic_accession=60526174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151306, CurrentID=0, Status='live', Symbol='GPBAR1', Aliases='BG37, GPCR19, GPR131, M-BAR, TGR5', description='G protein-coupled bile acid receptor 1', other_designations='G-protein coupled bile acid receptor BG37|G-protein coupled receptor GPCR19|hBG37|hGPCR19|membrane bile acid receptor|membrane-type receptor for bile acids', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218259496, end_position_on_the_genomic_accession=218263861, orientation='plus', exon_count=3, OMIM=610147, no_hdr0=''),
  105373287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373287, CurrentID=0, Status='live', Symbol='LOC105373287', Aliases='', description='uncharacterized LOC105373287', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=82741217, end_position_on_the_genomic_accession=82765809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421389, CurrentID=0, Status='live', Symbol='HNRNPA1P9', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 9', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87521393, end_position_on_the_genomic_accession=87522612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421390, CurrentID=0, Status='live', Symbol='EIF5P1', Aliases='', description='eukaryotic translation initiation factor 5 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9818039, end_position_on_the_genomic_accession=9819669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421391, CurrentID=0, Status='live', Symbol='EIF5P2', Aliases='', description='eukaryotic translation initiation factor 5 pseudogene 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19496366, end_position_on_the_genomic_accession=19497170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723689, CurrentID=0, Status='live', Symbol='LOC102723689', Aliases='', description='actin-related protein 3B-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151313, CurrentID=0, Status='live', Symbol='FAHD2B', Aliases='', description='fumarylacetoacetate hydrolase domain containing 2B', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97082404, end_position_on_the_genomic_accession=97094882, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100421395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421395, CurrentID=0, Status='live', Symbol='HNRNPA3P12', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 12', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53975178, end_position_on_the_genomic_accession=53976031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421397, CurrentID=0, Status='live', Symbol='HNRNPA1P68', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 68', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100941022, end_position_on_the_genomic_accession=100941987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421398, CurrentID=0, Status='live', Symbol='HNRNPA1P54', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 54', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179455920, end_position_on_the_genomic_accession=179457398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421399, CurrentID=0, Status='live', Symbol='HNRNPA1P46', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 46', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=191145623, end_position_on_the_genomic_accession=191147021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421400, CurrentID=0, Status='live', Symbol='LOC100421400', Aliases='', description='alanyl-tRNA synthetase domain containing 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77194818, end_position_on_the_genomic_accession=77195399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421401, CurrentID=0, Status='live', Symbol='LOC100421401', Aliases='', description='guanylate binding protein family member 6 pseudogene', other_designations='guanylate binding protein family, member 6 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89237424, end_position_on_the_genomic_accession=89302958, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929681, CurrentID=0, Status='live', Symbol='LOC101929681', Aliases='', description='uncharacterized LOC101929681', other_designations='CTD-2010I22.2', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29380070, end_position_on_the_genomic_accession=29395976, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421404, CurrentID=0, Status='live', Symbol='LOC100421404', Aliases='', description='retinitis pigmentosa 9 (autosomal dominant) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20291539, end_position_on_the_genomic_accession=20292262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151325, CurrentID=0, Status='live', Symbol='MYADML', Aliases='', description='myeloid-associated differentiation marker-like (pseudogene)', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33726061, end_position_on_the_genomic_accession=33728217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421409, CurrentID=0, Status='live', Symbol='LOC100421409', Aliases='', description='glucosamine-phosphate N-acetyltransferase 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189822294, end_position_on_the_genomic_accession=189822830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421411, CurrentID=0, Status='live', Symbol='DNAJB5P1', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B5 pseudogene 1', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 5 pseudogene 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76931440, end_position_on_the_genomic_accession=76932054, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421413, CurrentID=0, Status='live', Symbol='HNRNPA3P10', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 10', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51712687, end_position_on_the_genomic_accession=51713606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421414, CurrentID=0, Status='live', Symbol='LOC100421414', Aliases='', description='glial cells missing homolog 1 (Drosophila) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32823436, end_position_on_the_genomic_accession=32824505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421415, CurrentID=0, Status='live', Symbol='HNRNPA3P9', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 9', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32591863, end_position_on_the_genomic_accession=32592766, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421417, CurrentID=0, Status='live', Symbol='LOC100421417', Aliases='', description='ATPase, H+/K+ exchanging, beta polypeptide pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134316771, end_position_on_the_genomic_accession=134317272, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421419, CurrentID=0, Status='live', Symbol='LOC100421419', Aliases='', description='protein phosphatase 4 regulatory subunit 2 pseudogene', other_designations='protein phosphatase 4, regulatory subunit 2 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=288693, end_position_on_the_genomic_accession=290391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421420, CurrentID=0, Status='live', Symbol='TPT1P5', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31908361, end_position_on_the_genomic_accession=31909183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421422, CurrentID=0, Status='live', Symbol='ANKRD20A10P', Aliases='ANKRD20A4P1', description='ankyrin repeat domain 20 family member A10, pseudogene', other_designations='ankyrin repeat domain 20 family, member A10, pseudogene|ankyrin repeat domain 20 family, member A4 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24751766, end_position_on_the_genomic_accession=24752333, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271164, CurrentID=0, Status='live', Symbol='RPL21P47', Aliases='RPL21_16_486', description='ribosomal protein L21 pseudogene 47', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62248294, end_position_on_the_genomic_accession=62248770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421424, CurrentID=0, Status='live', Symbol='HNRNPA1P29', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 29', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93394976, end_position_on_the_genomic_accession=93395833, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  544561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544561, CurrentID=0, Status='live', Symbol='AASTH31', Aliases='', description='Allergic/atopic asthma related QTL 31', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544562, CurrentID=0, Status='live', Symbol='AASTH52', Aliases='', description='Allergic/atopic asthma related QTL 52', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544563, CurrentID=0, Status='live', Symbol='AASTH48', Aliases='', description='Allergic/atopic asthma related QTL 48', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544564, CurrentID=0, Status='live', Symbol='AASTH34', Aliases='', description='Allergic/atopic asthma related QTL 34', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544565, CurrentID=0, Status='live', Symbol='AASTH20', Aliases='', description='Allergic/atopic asthma related QTL 20', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421430, CurrentID=0, Status='live', Symbol='HNRNPA1P74', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 74', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=54320730, end_position_on_the_genomic_accession=54321660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  352905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352905, CurrentID=0, Status='live', Symbol='CTBP2P1', Aliases='', description='C-terminal binding protein 2 pseudogene 1', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=56855243, end_position_on_the_genomic_accession=56856170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  544568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544568, CurrentID=0, Status='live', Symbol='AASTH9', Aliases='', description='Allergic/atopic asthma related QTL 9', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544569, CurrentID=0, Status='live', Symbol='AASTH21', Aliases='', description='Allergic/atopic asthma related QTL 21', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151354, CurrentID=0, Status='live', Symbol='FAM84A', Aliases='NSE1, PP11517', description='family with sequence similarity 84 member A', other_designations='family with sequence similarity 84, member A|neurologic sensory protein 1|neurological/sensory 1', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=14632686, end_position_on_the_genomic_accession=14640046, orientation='plus', exon_count=2, OMIM=611234, no_hdr0=''),
  544571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544571, CurrentID=0, Status='live', Symbol='AASTH47', Aliases='', description='Allergic/atopic asthma related QTL 47', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544572, CurrentID=0, Status='live', Symbol='AASTH39', Aliases='', description='Allergic/atopic asthma related QTL 39', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544573, CurrentID=0, Status='live', Symbol='AASTH17', Aliases='', description='Allergic/atopic asthma related QTL 17', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544574, CurrentID=0, Status='live', Symbol='AASTH35', Aliases='', description='Allergic/atopic asthma related QTL 35', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544575, CurrentID=0, Status='live', Symbol='AASTH38', Aliases='', description='Allergic/atopic asthma related QTL 38', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544576, CurrentID=0, Status='live', Symbol='AASTH26', Aliases='', description='Allergic/atopic asthma related QTL 25', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544577, CurrentID=0, Status='live', Symbol='AASTH49', Aliases='', description='Allergic/atopic asthma related QTL 49', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544578, CurrentID=0, Status='live', Symbol='AASTH50', Aliases='', description='Allergic/atopic asthma related QTL 50', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544579, CurrentID=0, Status='live', Symbol='AASTH12', Aliases='', description='Allergic/atopic asthma related QTL 12', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544580, CurrentID=0, Status='live', Symbol='AASTH23', Aliases='', description='Allergic/atopic asthma related QTL 23', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544581, CurrentID=0, Status='live', Symbol='AASTH55', Aliases='', description='Allergic/atopic asthma related QTL 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544582, CurrentID=0, Status='live', Symbol='AASTH42', Aliases='', description='Allergic/atopic asthma related QTL 42', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544583, CurrentID=0, Status='live', Symbol='AASTH13', Aliases='', description='Allergic/atopic asthma related QTL 13', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544584, CurrentID=0, Status='live', Symbol='AASTH11', Aliases='', description='Allergic/atopic asthma related QTL 11', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544585, CurrentID=0, Status='live', Symbol='AASTH36', Aliases='', description='Allergic/atopic asthma related QTL 36', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544586, CurrentID=0, Status='live', Symbol='AASTH54', Aliases='', description='Allergic/atopic asthma related QTL 54', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544587, CurrentID=0, Status='live', Symbol='AASTH51', Aliases='', description='Allergic/atopic asthma related QTL 51', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544588, CurrentID=0, Status='live', Symbol='AASTH33', Aliases='', description='Allergic/atopic asthma related QTL 33', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544589, CurrentID=0, Status='live', Symbol='AASTH19', Aliases='', description='Allergic/atopic asthma related QTL 19', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544590, CurrentID=0, Status='live', Symbol='AASTH43', Aliases='', description='Allergic/atopic asthma related QTL 43', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544591, CurrentID=0, Status='live', Symbol='AASTH8', Aliases='', description='Allergic/atopic asthma related QTL 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544592, CurrentID=0, Status='live', Symbol='AASTH27', Aliases='', description='Allergic/atopic asthma related QTL 27', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544593, CurrentID=0, Status='live', Symbol='AASTH53', Aliases='', description='Allergic/atopic asthma related QTL 53', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544594, CurrentID=0, Status='live', Symbol='AASTH44', Aliases='', description='Allergic/atopic asthma related QTL 44', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544595, CurrentID=0, Status='live', Symbol='AASTH18', Aliases='', description='Allergic/atopic asthma related QTL 18', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544596, CurrentID=0, Status='live', Symbol='AASTH16', Aliases='', description='Allergic/atopic asthma related QTL 16', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544597, CurrentID=0, Status='live', Symbol='AASTH15', Aliases='', description='Allergic/atopic asthma related QTL 15', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544598, CurrentID=0, Status='live', Symbol='AASTH37', Aliases='', description='Allergic/atopic asthma related QTL 37', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544599, CurrentID=0, Status='live', Symbol='AASTH45', Aliases='', description='Allergic/atopic asthma related QTL 45', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544600, CurrentID=0, Status='live', Symbol='AASTH24', Aliases='', description='Allergic/atopic asthma related QTL 24', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544601, CurrentID=0, Status='live', Symbol='AASTH46', Aliases='', description='Allergic/atopic asthma related QTL 46', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544602, CurrentID=0, Status='live', Symbol='AASTH22', Aliases='', description='Allergic/atopic asthma related QTL 22', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544603, CurrentID=0, Status='live', Symbol='AASTH10', Aliases='', description='Allergic/atopic asthma related QTL 10', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421468, CurrentID=0, Status='live', Symbol='LOC100421468', Aliases='', description='ubiquitin-like modifier activating enzyme 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87045877, end_position_on_the_genomic_accession=87046717, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421471, CurrentID=0, Status='live', Symbol='LOC100421471', Aliases='', description='NSE4 homolog A, SMC5-SMC6 complex component pseudogene', other_designations='NSE4 homolog, SMC5-SMC6 complex component A pseudogene|non-SMC element 4 homolog A (S. cerevisiae) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182734229, end_position_on_the_genomic_accession=182734751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421472, CurrentID=0, Status='live', Symbol='LOC100421472', Aliases='', description='mitochondrial fission regulator 2 pseudogene', other_designations='family with sequence similarity 54, member A pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63232427, end_position_on_the_genomic_accession=63233627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151393, CurrentID=0, Status='live', Symbol='RMDN2', Aliases='BLOCK18, FAM82A, FAM82A1, PRO34163, PYST9371, RMD-2, RMD2, RMD4', description='regulator of microtubule dynamics 2', other_designations='family with sequence similarity 82, member A|family with sequence similarity 82, member A1|microtubule-associated protein', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37920845, end_position_on_the_genomic_accession=38069970, orientation='plus', exon_count=21, OMIM=611872, no_hdr0=''),
  100421474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421474, CurrentID=0, Status='live', Symbol='LOC100421474', Aliases='', description='family with sequence similarity 49, member B pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169658721, end_position_on_the_genomic_accession=169661356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421476, CurrentID=0, Status='live', Symbol='LOC100421476', Aliases='', description='family with sequence similarity 71, member A pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84850646, end_position_on_the_genomic_accession=84851155, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373290, CurrentID=0, Status='live', Symbol='LOC105373290', Aliases='', description='uncharacterized LOC105373290', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86161044, end_position_on_the_genomic_accession=86176516, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  544614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544614, CurrentID=0, Status='live', Symbol='BW59', Aliases='', description='Body weight QTL 59', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421479, CurrentID=0, Status='live', Symbol='HTR1DP1', Aliases='', description='5-hydroxytryptamine (serotonin) receptor 1D pseudogene 1', other_designations='5-hydroxytryptamine (serotonin) receptor 1D, G protein-coupled pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1152257, end_position_on_the_genomic_accession=1153037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  544616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544616, CurrentID=0, Status='live', Symbol='COHEN2', Aliases='', description='Cohen syndrome QTL 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544617, CurrentID=0, Status='live', Symbol='BW35', Aliases='', description='Body weight QTL 35', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544618, CurrentID=0, Status='live', Symbol='COHEN1', Aliases='', description='Cohen syndrome QTL 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544619, CurrentID=0, Status='live', Symbol='BW37', Aliases='', description='Body weight QTL 37', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544620, CurrentID=0, Status='live', Symbol='BW54', Aliases='', description='Body weight QTL 54', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544621, CurrentID=0, Status='live', Symbol='BW55', Aliases='', description='Body weight QTL 55', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544622, CurrentID=0, Status='live', Symbol='BW36', Aliases='', description='Body weight QTL 36', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544623, CurrentID=0, Status='live', Symbol='HRTRT1', Aliases='', description='Heart rate QTL 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544624, CurrentID=0, Status='live', Symbol='BW69', Aliases='', description='Body weight QTL 69', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544625, CurrentID=0, Status='live', Symbol='BW53', Aliases='', description='Body weight QTL 53', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544626, CurrentID=0, Status='live', Symbol='BW52', Aliases='', description='Body weight QTL 52', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544627, CurrentID=0, Status='live', Symbol='BW72', Aliases='', description='Body weight QTL 72', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544628, CurrentID=0, Status='live', Symbol='BW70', Aliases='', description='Body weight QTL 70', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544629, CurrentID=0, Status='live', Symbol='BW64', Aliases='', description='Body weight QTL 64', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544630, CurrentID=0, Status='live', Symbol='BW57', Aliases='', description='Body weight QTL 57', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544631, CurrentID=0, Status='live', Symbol='SLEP2', Aliases='', description='Serum leptin concentration QTL 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544632, CurrentID=0, Status='live', Symbol='BW61', Aliases='', description='Body weight QTL 61', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544633, CurrentID=0, Status='live', Symbol='BW71', Aliases='', description='Body weight QTL 71', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544634, CurrentID=0, Status='live', Symbol='BW60', Aliases='', description='Body weight QTL 60', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544635, CurrentID=0, Status='live', Symbol='BW66', Aliases='', description='Body weight QTL 66', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544636, CurrentID=0, Status='live', Symbol='BW58', Aliases='', description='Body weight QTL 58', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544637, CurrentID=0, Status='live', Symbol='BW62', Aliases='', description='Body weight QTL 62', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544638, CurrentID=0, Status='live', Symbol='BW68', Aliases='', description='Body weight QTL 68', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544639, CurrentID=0, Status='live', Symbol='BW67', Aliases='', description='Body weight QTL 67', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544640, CurrentID=0, Status='live', Symbol='BW63', Aliases='', description='Body weight QTL 63', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421505, CurrentID=0, Status='live', Symbol='LOC100421505', Aliases='', description='MAP/microtubule affinity-regulating kinase 4 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12862417, end_position_on_the_genomic_accession=12863604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421506, CurrentID=0, Status='live', Symbol='ILF2P1', Aliases='', description='interleukin enhancer binding factor 2 pseudogene 1', other_designations='interleukin enhancer binding factor 2, 45kDa pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65687256, end_position_on_the_genomic_accession=65688657, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421507, CurrentID=0, Status='live', Symbol='MARK2P3', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129334530, end_position_on_the_genomic_accession=129335305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421508, CurrentID=0, Status='live', Symbol='LOC100421508', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129345410, end_position_on_the_genomic_accession=129346129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421509, CurrentID=0, Status='live', Symbol='LARP7P4', Aliases='', description='La ribonucleoprotein domain family member 7 pseudogene 4', other_designations='La ribonucleoprotein domain family, member 7 pseudogene 4', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145523526, end_position_on_the_genomic_accession=145524407, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421510, CurrentID=0, Status='live', Symbol='LOC100421510', Aliases='', description='family with sequence similarity 20, member B pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169931998, end_position_on_the_genomic_accession=169933232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421511, CurrentID=0, Status='live', Symbol='LOC100421511', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=669080, end_position_on_the_genomic_accession=670307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421512, CurrentID=0, Status='live', Symbol='RBPMS2P1', Aliases='', description='RBPMS2 pseudogene 1', other_designations='RNA binding protein with multiple splicing 2 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73157320, end_position_on_the_genomic_accession=73158225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271167, CurrentID=0, Status='live', Symbol='RPL21P50', Aliases='RPL21_18_516', description='ribosomal protein L21 pseudogene 50', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123745713, end_position_on_the_genomic_accession=123746267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  544650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544650, CurrentID=0, Status='live', Symbol='SLEP1', Aliases='', description='Serum leptin concentration QTL 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421515, CurrentID=0, Status='live', Symbol='BMPR1APS1', Aliases='', description='bone morphogenetic protein receptor type IA pseudogene 1', other_designations='bone morphogenetic protein receptor, type IA pseudogene 1', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129157526, end_position_on_the_genomic_accession=129159120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  134466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134466, CurrentID=0, Status='live', Symbol='ZNF300P1', Aliases='', description='zinc finger protein 300 pseudogene 1 (functional)', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150930436, end_position_on_the_genomic_accession=150946584, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421518, CurrentID=0, Status='live', Symbol='LOC100421518', Aliases='', description='MAD2 mitotic arrest deficient-like 1 (yeast) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167347472, end_position_on_the_genomic_accession=167348134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421523, CurrentID=0, Status='live', Symbol='LOC100421523', Aliases='', description='UDP-N-acetylglucosamine pyrophosphorylase 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47367404, end_position_on_the_genomic_accession=47369308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105664404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105664404, CurrentID=0, Status='live', Symbol='WSPAR', Aliases='LncTCF7, TCONS_00009511', description='WNT signaling pathway activating non-coding RNA', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616333, no_hdr0=''),
  100421525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421525, CurrentID=0, Status='live', Symbol='LOC100421525', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7 pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63182166, end_position_on_the_genomic_accession=63182817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421527, CurrentID=0, Status='live', Symbol='LOC100421527', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C8 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8 pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77632887, end_position_on_the_genomic_accession=77633426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421528, CurrentID=0, Status='live', Symbol='EIF4BP5', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 5', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47564976, end_position_on_the_genomic_accession=47566954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151449, CurrentID=0, Status='live', Symbol='GDF7', Aliases='BMP12', description='growth differentiation factor 7', other_designations='GDF-7', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20666664, end_position_on_the_genomic_accession=20671490, orientation='plus', exon_count=2, OMIM=604651, no_hdr0=''),
  100421531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421531, CurrentID=0, Status='live', Symbol='MARK2P7', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36146834, end_position_on_the_genomic_accession=36148025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421533, CurrentID=0, Status='live', Symbol='EIF4A1P10', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 10', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=92113098, end_position_on_the_genomic_accession=92114477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421535, CurrentID=0, Status='live', Symbol='GEMIN8P3', Aliases='', description='gem nuclear organelle associated protein 8 pseudogene 3', other_designations='gem (nuclear organelle) associated protein 8 pseudogene 3', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86304367, end_position_on_the_genomic_accession=86305087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421536, CurrentID=0, Status='live', Symbol='LOC100421536', Aliases='', description='diphosphoinositol pentakisphosphate kinase 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75641338, end_position_on_the_genomic_accession=75642087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421537, CurrentID=0, Status='live', Symbol='LOC100421537', Aliases='', description='inositol monophosphatase domain containing 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210462345, end_position_on_the_genomic_accession=210463012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421539, CurrentID=0, Status='live', Symbol='SUZ12P2', Aliases='', description='SUZ12 polycomb repressive complex 2 subunit pseudogene 2', other_designations='suppressor of zeste 12 homolog pseudogene 2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23768200, end_position_on_the_genomic_accession=23769022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421541, CurrentID=0, Status='live', Symbol='LOC100421541', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C9 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 9 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67799004, end_position_on_the_genomic_accession=67799806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421542, CurrentID=0, Status='live', Symbol='DNAJA1P6', Aliases='', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 6', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 6', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17517835, end_position_on_the_genomic_accession=17519277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271168, CurrentID=0, Status='live', Symbol='RPL21P54', Aliases='RPL21_19_553', description='ribosomal protein L21 pseudogene 54', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34883077, end_position_on_the_genomic_accession=34883519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421545, CurrentID=0, Status='live', Symbol='EIF4HP2', Aliases='', description='eukaryotic translation initiation factor 4H pseudogene 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30902154, end_position_on_the_genomic_accession=30902778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421548, CurrentID=0, Status='live', Symbol='LOC100421548', Aliases='', description='family with sequence similarity 178 member B pseudogene', other_designations='family with sequence similarity 178, member B pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95607073, end_position_on_the_genomic_accession=95616589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421549, CurrentID=0, Status='live', Symbol='RFC5P1', Aliases='', description='replication factor C 5 pseudogene 1', other_designations='replication factor C (activator 1) 5, 36.5kDa pseudogene 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77072678, end_position_on_the_genomic_accession=77073292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421550, CurrentID=0, Status='live', Symbol='MARK3P1', Aliases='', description='MAP/microtubule affinity-regulating kinase 3 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46681991, end_position_on_the_genomic_accession=46682734, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100290481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100290481, CurrentID=0, Status='live', Symbol='CYAT1', Aliases='', description='immunoglobulin lambda light chain-like', other_designations='cyclosporin A transporter 1', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151475, CurrentID=0, Status='live', Symbol='LOC151475', Aliases='', description='uncharacterized LOC151475', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230690921, end_position_on_the_genomic_accession=230700529, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  282549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282549, CurrentID=0, Status='live', Symbol='STQTL3', Aliases='', description='stature quantitative trait locus 3', other_designations='', map_location='12p11.2-q14', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606257, no_hdr0=''),
  282550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282550, CurrentID=0, Status='live', Symbol='STQTL4', Aliases='', description='stature quantitative trait locus 4', other_designations='', map_location='13q32-q33', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606258, no_hdr0=''),
  282551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282551, CurrentID=0, Status='live', Symbol='ITS', Aliases='', description='Insulinoma tumor suppressor gene locus', other_designations='', map_location='22q12.1-q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606960, no_hdr0=''),
  282552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282552, CurrentID=0, Status='live', Symbol='IS2', Aliases='AIS2', description='Scoliosis, idiopathic 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607354, no_hdr0=''),
  282553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282553, CurrentID=0, Status='live', Symbol='AUTS8', Aliases='AUTS2', description='Autism, susceptibility to, 8', other_designations='Autism, susceptibility to, 2, formerly', map_location='3q25-q27', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607373, no_hdr0=''),
  151484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151484, CurrentID=0, Status='live', Symbol='LOC151484', Aliases='', description='uncharacterized LOC151484', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230886546, end_position_on_the_genomic_accession=230904517, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421565, CurrentID=0, Status='live', Symbol='MARK2P12', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 12', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73407789, end_position_on_the_genomic_accession=73408325, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421566, CurrentID=0, Status='live', Symbol='FAM58DP', Aliases='', description='family with sequence similarity 58 member D, pseudogene', other_designations='family with sequence similarity 58, member D, pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19382259, end_position_on_the_genomic_accession=19383308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421569, CurrentID=0, Status='live', Symbol='LOC100421569', Aliases='', description='glyoxylate reductase 1 homolog (Arabidopsis) pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49829125, end_position_on_the_genomic_accession=49831082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059915, CurrentID=0, Status='live', Symbol='LOC101059915', Aliases='', description='uncharacterized protein CXorf49-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71667779, end_position_on_the_genomic_accession=71720811, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100421572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421572, CurrentID=0, Status='live', Symbol='MESTP2', Aliases='', description='mesoderm specific transcript pseudogene 2', other_designations='mesoderm specific transcript homolog pseudogene 2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60053207, end_position_on_the_genomic_accession=60054065, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271169, CurrentID=0, Status='live', Symbol='RPL21P51', Aliases='RPL21_20_533', description='ribosomal protein L21 pseudogene 51', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165012326, end_position_on_the_genomic_accession=165012885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421577, CurrentID=0, Status='live', Symbol='FAM21EP', Aliases='', description='family with sequence similarity 21 member E, pseudogene', other_designations='FAM21B pseudogene|family with sequence similarity 21, member E, pseudogene', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50021182, end_position_on_the_genomic_accession=50067803, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421579, CurrentID=0, Status='live', Symbol='EIF4A2P2', Aliases='', description='eukaryotic translation initiation factor 4A2 pseudogene 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73200363, end_position_on_the_genomic_accession=73201225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421580, CurrentID=0, Status='live', Symbol='LOC100421580', Aliases='', description='family with sequence similarity 175, member A pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100910977, end_position_on_the_genomic_accession=100912163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421581, CurrentID=0, Status='live', Symbol='GCNT1P3', Aliases='', description='glucosaminyl (N-acetyl) transferase 1, core 2 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190624599, end_position_on_the_genomic_accession=190625166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421582, CurrentID=0, Status='live', Symbol='PAIP1P1', Aliases='bCX300M20.3, bPG254M11.3, bQB44E15.5', description='poly(A) binding protein interacting protein 1 pseudogene 1', other_designations='bPG254M11.3 bCX300M20.3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30186798, end_position_on_the_genomic_accession=30188606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421583, CurrentID=0, Status='live', Symbol='LOC100421583', Aliases='', description='hormonally up-regulated Neu-associated kinase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85156679, end_position_on_the_genomic_accession=85161809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151507, CurrentID=0, Status='live', Symbol='MSL3P1', Aliases='MSL3L2', description='male-specific lethal 3 homolog (Drosophila) pseudogene 1', other_designations='MSL3-like 2|male-specific lethal 3 homolog pseudogene|male-specific lethal 3-like 2|male-specific lethal-3 homolog 2|putative male-specific lethal-3 protein-like 2', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233865437, end_position_on_the_genomic_accession=233868409, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421588, CurrentID=0, Status='live', Symbol='FAM60BP', Aliases='HsT2323', description='family with sequence similarity 60 member B, pseudogene', other_designations='family with sequence similarity 60, member A pseudogene', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26255698, end_position_on_the_genomic_accession=26256357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059918, CurrentID=0, Status='live', Symbol='GOLGA8R', Aliases='', description='golgin A8 family member R', other_designations='golgin A8 family, member R', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30400562, end_position_on_the_genomic_accession=30414260, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  100421591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421591, CurrentID=0, Status='live', Symbol='LOC100421591', Aliases='', description='family with sequence similarity 76, member A pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31883260, end_position_on_the_genomic_accession=31883873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421593, CurrentID=0, Status='live', Symbol='LOC100421593', Aliases='', description='family with sequence similarity 71, member A pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=60927968, end_position_on_the_genomic_accession=60928541, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421594, CurrentID=0, Status='live', Symbol='LOC100421594', Aliases='', description='mitochondrial carrier 2 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57713782, end_position_on_the_genomic_accession=57714954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421595, CurrentID=0, Status='live', Symbol='LOC100421595', Aliases='', description='thyroid hormone receptor associated protein 3 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11820320, end_position_on_the_genomic_accession=11821300, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151516, CurrentID=0, Status='live', Symbol='ASPRV1', Aliases='MUNO, SASP, SASPase, Taps', description='aspartic peptidase, retroviral-like 1', other_designations='TPA-inducible aspartic proteinase-like protein|skin aspartic protease|skin-specific retroviral-like aspartic protease', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69960089, end_position_on_the_genomic_accession=69962265, orientation='minus', exon_count=1, OMIM=611765, no_hdr0=''),
  105373294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373294, CurrentID=0, Status='live', Symbol='LOC105373294', Aliases='', description='uncharacterized LOC105373294', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96363955, end_position_on_the_genomic_accession=96391856, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421598, CurrentID=0, Status='live', Symbol='GCNT1P5', Aliases='', description='glucosaminyl (N-acetyl) transferase 1, core 2 pseudogene 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77461156, end_position_on_the_genomic_accession=77462139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421599, CurrentID=0, Status='live', Symbol='TVP23CP1', Aliases='', description='TVP23C pseudogene 1', other_designations='family with sequence similarity 18, member B2 pseudogene|trans-golgi network vesicle protein 23 homolog C pseudogene 1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90631906, end_position_on_the_genomic_accession=90632568, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059920, CurrentID=0, Status='live', Symbol='FRG2HP', Aliases='', description='FSHD region gene 2 family member H, pseudogene', other_designations='protein FRG2-like-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35334582, end_position_on_the_genomic_accession=35344332, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421602, CurrentID=0, Status='live', Symbol='LOC100421602', Aliases='', description='CDC42 binding protein kinase beta (DMPK-like) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35741995, end_position_on_the_genomic_accession=35744871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421603, CurrentID=0, Status='live', Symbol='LOC100421603', Aliases='', description='family with sequence similarity 156, member A pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52326166, end_position_on_the_genomic_accession=52326690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271170, CurrentID=0, Status='live', Symbol='RPL21P55', Aliases='RPL21_20_571', description='ribosomal protein L21 pseudogene 55', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67608613, end_position_on_the_genomic_accession=67609160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151525, CurrentID=0, Status='live', Symbol='WDSUB1', Aliases='UBOX6, WDSAM1', description='WD repeat, sterile alpha motif and U-box domain containing 1', other_designations='2610014F08Rik|WD repeat and SAM domain containing 1|WD repeat, SAM and U-box domain containing 1', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159235793, end_position_on_the_genomic_accession=159286793, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100270846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270846, CurrentID=0, Status='live', Symbol='RPL6P21', Aliases='RPL6_5_786', description='ribosomal protein L6 pseudogene 21', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14070506, end_position_on_the_genomic_accession=14071399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421610, CurrentID=0, Status='live', Symbol='C4BPAP3', Aliases='', description='complement component 4 binding protein, alpha pseudogene 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207194789, end_position_on_the_genomic_accession=207195307, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151531, CurrentID=0, Status='live', Symbol='UPP2', Aliases='UDRPASE2, UP2, UPASE2', description='uridine phosphorylase 2', other_designations='UPase 2|liver-specific uridine phosphorylase|urdPase 2|uridine phosphorylase-2', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157995179, end_position_on_the_genomic_accession=158136154, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  151534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151534, CurrentID=0, Status='live', Symbol='LBX2-AS1', Aliases='', description='LBX2 antisense RNA 1', other_designations='LBX2 antisense RNA 1 (non-protein coding)', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74502617, end_position_on_the_genomic_accession=74505065, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  266917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266917, CurrentID=0, Status='live', Symbol='D21S2088E', Aliases='', description='D21S2088E', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23361104, end_position_on_the_genomic_accession=23384835, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421618, CurrentID=0, Status='live', Symbol='LOC100421618', Aliases='', description='family with sequence similarity 166, member A pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38205385, end_position_on_the_genomic_accession=38206012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421620, CurrentID=0, Status='live', Symbol='LOC100421620', Aliases='', description='polymerase (RNA) mitochondrial (DNA directed) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15902326, end_position_on_the_genomic_accession=15903239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421621, CurrentID=0, Status='live', Symbol='FAM187B2P', Aliases='', description='family with sequence similarity 187 member B2, pseudogene', other_designations='family with sequence similarity 187, member B2, pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35232290, end_position_on_the_genomic_accession=35233003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421622, CurrentID=0, Status='live', Symbol='LOC100421622', Aliases='', description='family with sequence similarity 91 member A1 pseudogene', other_designations='family with sequence similarity 91, member A1 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73994977, end_position_on_the_genomic_accession=73996181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421623, CurrentID=0, Status='live', Symbol='LOC100421623', Aliases='', description='zona pellucida glycoprotein 3 (sperm receptor) pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83875190, end_position_on_the_genomic_accession=83875754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282616, CurrentID=0, Status='live', Symbol='IFNL2', Aliases='IL-28A, IL28A', description='interferon, lambda 2', other_designations='IFN-lambda-2|cytokine Zcyto20|interferon lambda 2|interferon-lambda 2|interleukin 28A (interferon, lambda 2)|interleukin-28A', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39265883, end_position_on_the_genomic_accession=39270092, orientation='plus', exon_count=7, OMIM=607401, no_hdr0=''),
  282617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282617, CurrentID=0, Status='live', Symbol='IFNL3', Aliases='IL-28B, IL28B, IL28C', description='interferon, lambda 3', other_designations='IFN-lambda-3|IFN-lambda-4|IL-28C|cytokine Zcyto22|interferon, lambda 4|interleukin-28B|interleukin-28C', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39243606, end_position_on_the_genomic_accession=39247034, orientation='minus', exon_count=7, OMIM=607402, no_hdr0=''),
  282618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282618, CurrentID=0, Status='live', Symbol='IFNL1', Aliases='IL-29, IL29', description='interferon, lambda 1', other_designations='IFN-lambda-1|cytokine Zcyto21|interferon-lambda-1|interleukin 29 (interferon, lambda 1)|interleukin-29', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39296325, end_position_on_the_genomic_accession=39298672, orientation='plus', exon_count=5, OMIM=607403, no_hdr0=''),
  105379807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379807, CurrentID=0, Status='live', Symbol='LOC105379807', Aliases='', description='uncharacterized LOC105379807', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421629, CurrentID=0, Status='live', Symbol='METTL21AP1', Aliases='', description='methyltransferase like 21A pseudogene 1', other_designations='family with sequence similarity 119, member A pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39235401, end_position_on_the_genomic_accession=39236020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421630, CurrentID=0, Status='live', Symbol='LOC100421630', Aliases='', description='cutaneous T-cell lymphoma-associated antigen 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176631598, end_position_on_the_genomic_accession=176632529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421632, CurrentID=0, Status='live', Symbol='LOC100421632', Aliases='', description='thyroid hormone receptor associated protein 3 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91742812, end_position_on_the_genomic_accession=91743990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421633, CurrentID=0, Status='live', Symbol='WARS2P1', Aliases='', description='tryptophanyl tRNA synthetase 2 pseudogene 1', other_designations='tryptophanyl tRNA synthetase 2, mitochondrial pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80940644, end_position_on_the_genomic_accession=80941207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271171, CurrentID=0, Status='live', Symbol='RPL21P56', Aliases='RPL21_21_549', description='ribosomal protein L21 pseudogene 56', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31840457, end_position_on_the_genomic_accession=31841001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151556, CurrentID=0, Status='live', Symbol='GPR155', Aliases='DEP.7, DEPDC3, PGR22', description='G protein-coupled receptor 155', other_designations='G-protein coupled receptor PGR22', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174431571, end_position_on_the_genomic_accession=174487088, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  101059926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059926, CurrentID=0, Status='live', Symbol='FRG2JP', Aliases='', description='FSHD region gene 2 family member J, pseudogene', other_designations='protein FRG2-like-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35245693, end_position_on_the_genomic_accession=35247634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421638, CurrentID=0, Status='live', Symbol='CTAGE13P', Aliases='', description='CTAGE family member 13, pseudogene', other_designations='CTAGE family, member 13, pseudogene|cutaneous T-cell lymphoma-associated antigen 1 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168286997, end_position_on_the_genomic_accession=168289221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282631, CurrentID=0, Status='live', Symbol='GCCD2', Aliases='GCCD3', description='Glucocorticoid deficiency 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607398, no_hdr0=''),
  100421641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421641, CurrentID=0, Status='live', Symbol='LOC100421641', Aliases='', description='zinc finger CCHC-type and RNA binding motif 1 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69264303, end_position_on_the_genomic_accession=69264938, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421642, CurrentID=0, Status='live', Symbol='LOC100421642', Aliases='', description='zona pellucida glycoprotein 3 (sperm receptor) pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136486039, end_position_on_the_genomic_accession=136486565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874019, CurrentID=0, Status='live', Symbol='KLHL6-AS1', Aliases='', description='KLHL6 antisense RNA 1', other_designations='KLHL6 antisense RNA 1 (non-protein coding)', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183548735, end_position_on_the_genomic_accession=183552326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421646, CurrentID=0, Status='live', Symbol='LOC100421646', Aliases='', description='von Willebrand factor A domain containing 5B2 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22723476, end_position_on_the_genomic_accession=22726683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421648, CurrentID=0, Status='live', Symbol='SUCLA2P3', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25079706, end_position_on_the_genomic_accession=25081736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421651, CurrentID=0, Status='live', Symbol='LOC100421651', Aliases='', description='caseinolytic mitochondrial matrix peptidase chaperone subunit pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78437864, end_position_on_the_genomic_accession=78438916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421652, CurrentID=0, Status='live', Symbol='PTP4A1P1', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158065545, end_position_on_the_genomic_accession=158066059, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421655, CurrentID=0, Status='live', Symbol='LOC100421655', Aliases='', description='adaptor-related protein complex 3, mu 2 subunit pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68788564, end_position_on_the_genomic_accession=68789670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421657, CurrentID=0, Status='live', Symbol='HNRNPDLP4', Aliases='', description='heterogeneous nuclear ribonucleoprotein D-like pseudogene 4', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58049046, end_position_on_the_genomic_accession=58049547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421658, CurrentID=0, Status='live', Symbol='LOC100421658', Aliases='', description='sacsin molecular chaperone pseudogene', other_designations='spastic ataxia of Charlevoix-Saguenay (sacsin) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102766801, end_position_on_the_genomic_accession=102767927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151579, CurrentID=0, Status='live', Symbol='BZW1P2', Aliases='BZW1L1', description='basic leucine zipper and W2 domains 1 pseudogene 2', other_designations='basic leucine zipper and W2 domains 1 like 1', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116644238, end_position_on_the_genomic_accession=116646724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  134490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134490, CurrentID=0, Status='live', Symbol='MRPL57P6', Aliases='MRP63P6', description='mitochondrial ribosomal protein L57 pseudogene 6', other_designations='mitochondrial ribosomal protein 63 pseudogene 6', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163039281, end_position_on_the_genomic_accession=163039606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421663, CurrentID=0, Status='live', Symbol='LOC100421663', Aliases='', description='cutaneous T-cell lymphoma-associated antigen 1 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19809088, end_position_on_the_genomic_accession=19810021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271172, CurrentID=0, Status='live', Symbol='RPL21P57', Aliases='RPL21_21_631', description='ribosomal protein L21 pseudogene 57', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154729189, end_position_on_the_genomic_accession=154729711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421667, CurrentID=0, Status='live', Symbol='LOC100421667', Aliases='', description='abhydrolase domain containing 17C pseudogene', other_designations='family with sequence similarity 108, member C1 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23225888, end_position_on_the_genomic_accession=23226392, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132817, CurrentID=0, Status='live', Symbol='RBM48P1', Aliases='', description='RNA binding motif protein 48 pseudogene 1', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126563490, end_position_on_the_genomic_accession=126564936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421669, CurrentID=0, Status='live', Symbol='LOC100421669', Aliases='', description='ribosome binding protein 1 homolog 180kDa (dog) pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=22811636, end_position_on_the_genomic_accession=22812456, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421670, CurrentID=0, Status='live', Symbol='LOC100421670', Aliases='', description='developmentally regulated GTP binding protein 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117674315, end_position_on_the_genomic_accession=117675257, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421671, CurrentID=0, Status='live', Symbol='SUCLG2P1', Aliases='', description='succinate-CoA ligase, GDP-forming, beta subunit pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142532247, end_position_on_the_genomic_accession=142532968, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421672, CurrentID=0, Status='live', Symbol='LOC100421672', Aliases='', description='small nuclear ribonucleoprotein polypeptide B pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59535863, end_position_on_the_genomic_accession=59536415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134492, CurrentID=0, Status='live', Symbol='NUDCD2', Aliases='', description='NudC domain containing 2', other_designations='NudC-like protein 2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163453488, end_position_on_the_genomic_accession=163460137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421674, CurrentID=0, Status='live', Symbol='LOC100421674', Aliases='', description='kelch repeat and BTB (POZ) domain containing 2 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40380666, end_position_on_the_genomic_accession=40381838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421675, CurrentID=0, Status='live', Symbol='ARL2BPP9', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 9', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 9|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 9|ADP-ribosylation factor-like 2 binding protein pseudogene 9', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82576614, end_position_on_the_genomic_accession=82577392, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421681, CurrentID=0, Status='live', Symbol='PTP4A1P7', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176616271, end_position_on_the_genomic_accession=176616783, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104355138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355138, CurrentID=0, Status='live', Symbol='LINC01162', Aliases='', description='long intergenic non-protein coding RNA 1162', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20835431, end_position_on_the_genomic_accession=21023148, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100421684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421684, CurrentID=0, Status='live', Symbol='LOC100421684', Aliases='', description='tetratricopeptide repeat domain 34 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19199763, end_position_on_the_genomic_accession=19200279, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421685, CurrentID=0, Status='live', Symbol='LOC100421685', Aliases='', description='tetratricopeptide repeat domain 34 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15922428, end_position_on_the_genomic_accession=15922944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282679, CurrentID=0, Status='live', Symbol='AQP11', Aliases='AQPX1', description='aquaporin 11', other_designations='AQP-11', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77589635, end_position_on_the_genomic_accession=77610356, orientation='plus', exon_count=5, OMIM=609914, no_hdr0=''),
  100421689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421689, CurrentID=0, Status='live', Symbol='LOC100421689', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170077063, end_position_on_the_genomic_accession=170078387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421690, CurrentID=0, Status='live', Symbol='LOC100421690', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67192587, end_position_on_the_genomic_accession=67193408, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059935, CurrentID=0, Status='live', Symbol='LOC101059935', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421692, CurrentID=0, Status='live', Symbol='LOC100421692', Aliases='', description='leishmanolysin-like (metallopeptidase M8 family) pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94496272, end_position_on_the_genomic_accession=94509584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151613, CurrentID=0, Status='live', Symbol='TTC14', Aliases='DRDL5813, PRO19630', description='tetratricopeptide repeat domain 14', other_designations='TPR repeat protein 14', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180602130, end_position_on_the_genomic_accession=180618347, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100271173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271173, CurrentID=0, Status='live', Symbol='RPL21P58', Aliases='RPL21_22_606', description='ribosomal protein L21 pseudogene 58', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119547099, end_position_on_the_genomic_accession=119547645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421695, CurrentID=0, Status='live', Symbol='BNIP3P10', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 10', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19870724, end_position_on_the_genomic_accession=19871283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421696, CurrentID=0, Status='live', Symbol='BNIP3P14', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 14', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20087016, end_position_on_the_genomic_accession=20087576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421697, CurrentID=0, Status='live', Symbol='BNIP3P23', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 23', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20443213, end_position_on_the_genomic_accession=20443747, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421698, CurrentID=0, Status='live', Symbol='BNIP3P24', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 24', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20784599, end_position_on_the_genomic_accession=20785152, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421699, CurrentID=0, Status='live', Symbol='BNIP3P26', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 26', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21521346, end_position_on_the_genomic_accession=21521896, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873477, CurrentID=0, Status='live', Symbol='RNA5SP217', Aliases='RN5S217', description='RNA, 5S ribosomal pseudogene 217', other_designations='RNA, 5S ribosomal 217', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127317883, end_position_on_the_genomic_accession=127317973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421701, CurrentID=0, Status='live', Symbol='BNIP3P39', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 39', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23798928, end_position_on_the_genomic_accession=23799496, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282694, CurrentID=0, Status='live', Symbol='NASPP1', Aliases='', description='nuclear autoantigenic sperm protein pseudogene 1', other_designations='nuclear autoantigenic sperm protein (histone-binding) pseudogene', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60936948, end_position_on_the_genomic_accession=60939978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421703, CurrentID=0, Status='live', Symbol='BNIP3P12', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 12', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19943703, end_position_on_the_genomic_accession=20009237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421704, CurrentID=0, Status='live', Symbol='BNIP3P17', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 17', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20221878, end_position_on_the_genomic_accession=20222461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282697, CurrentID=0, Status='live', Symbol='DIS3L2P1', Aliases='DIS3L2P, FAM6B', description="DIS3 like 3'-5' exoribonuclease 2 pseudogene 1", other_designations='DIS3 mitotic control homolog-like 2 pseudogene 1|DIS3L2 pseudogene 1|family with sequence similarity 6, member B pseudogene', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232442039, end_position_on_the_genomic_accession=232447966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421706, CurrentID=0, Status='live', Symbol='BNIP3P22', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 22', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20416514, end_position_on_the_genomic_accession=20417066, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421707, CurrentID=0, Status='live', Symbol='BNIP3P25', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 25', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21318161, end_position_on_the_genomic_accession=21318688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421708, CurrentID=0, Status='live', Symbol='BNIP3P27', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 27', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21826762, end_position_on_the_genomic_accession=21827279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059938, CurrentID=0, Status='live', Symbol='NPIPA7', Aliases='', description='nuclear pore complex interacting protein family member A7', other_designations='nuclear pore complex interacting protein family, member A7', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16378794, end_position_on_the_genomic_accession=16393972, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100421710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421710, CurrentID=0, Status='live', Symbol='BNIP3P8', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 8', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23380317, end_position_on_the_genomic_accession=23380873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421712, CurrentID=0, Status='live', Symbol='SUCLG2P3', Aliases='', description='succinate-CoA ligase, GDP-forming, beta subunit pseudogene 3', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107585900, end_position_on_the_genomic_accession=107586698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282706, CurrentID=0, Status='live', Symbol='DAOA-AS1', Aliases='DAOA-AS, DAOAAS, G30', description='DAOA antisense RNA 1', other_designations='DAOA antisense RNA (non-protein coding)|DAOA antisense RNA 1 (non-protein coding)|G30 transcript|putative protein LG30', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105459055, end_position_on_the_genomic_accession=105505681, orientation='minus', exon_count=7, OMIM=607415, no_hdr0=''),
  151636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151636, CurrentID=0, Status='live', Symbol='DTX3L', Aliases='BBAP', description='deltex 3 like, E3 ubiquitin ligase', other_designations='B-lymphoma- and BAL-associated protein|deltex 3-like|rhysin 2|rhysin-2|rhysin2', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122564338, end_position_on_the_genomic_accession=122575203, orientation='plus', exon_count=5, OMIM=613143, no_hdr0=''),
  100421717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421717, CurrentID=0, Status='live', Symbol='UBE2Q2P5Y', Aliases='UBE2Q2PY2', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 5, Y-linked', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 5, Y-linked|ubiquitin-conjugating enzyme E2Q family member 2 pseudogene, Y-linked 2', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25431147, end_position_on_the_genomic_accession=25437315, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421718, CurrentID=0, Status='live', Symbol='NEK4P1', Aliases='', description='NIMA-related kinase 4 pseudogene 1', other_designations='NIMA (never in mitosis gene a)-related kinase 4 pseudogene 1', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17210468, end_position_on_the_genomic_accession=17211356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421721, CurrentID=0, Status='live', Symbol='LOC100421721', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152795924, end_position_on_the_genomic_accession=152796887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271174, CurrentID=0, Status='live', Symbol='RPL21P59', Aliases='RPL21_22_635', description='ribosomal protein L21 pseudogene 59', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=165905084, end_position_on_the_genomic_accession=165905637, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151645, CurrentID=0, Status='live', Symbol='PSMC1P1', Aliases='PSMC1P, bcm3415', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 1', other_designations='protease (prosome, macropain) 26S subunit, ATPase, 1 pseudogene', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68635685, end_position_on_the_genomic_accession=68637242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  151647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151647, CurrentID=0, Status='live', Symbol='FAM19A4', Aliases='TAFA-4, TAFA4', description='family with sequence similarity 19 (chemokine (C-C motif)-like), member A4', other_designations='chemokine-like protein TAFA-4', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68731764, end_position_on_the_genomic_accession=68932610, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  151648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151648, CurrentID=0, Status='live', Symbol='SGOL1', Aliases='CAID, NY-BR-85, SGO, Sgo1', description='shugoshin-like 1 (S. pombe)', other_designations='serologically defined breast cancer antigen NY-BR-85|shugoshin 1AB protein|shugoshin 1CD protein|shugoshin 1EF protein|shugoshin 1GH protein|shugoshin 1KL protein', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20159046, end_position_on_the_genomic_accession=20188131, orientation='minus', exon_count=13, OMIM=609168, no_hdr0=''),
  151649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151649, CurrentID=0, Status='live', Symbol='PP2D1', Aliases='C3orf48', description='protein phosphatase 2C-like domain containing 1', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19979961, end_position_on_the_genomic_accession=20012780, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100421730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421730, CurrentID=0, Status='live', Symbol='LOC100421730', Aliases='', description='transmembrane p24 trafficking protein 2 pseudogene', other_designations='transmembrane emp24 domain trafficking protein 2 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60730053, end_position_on_the_genomic_accession=60730642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151651, CurrentID=0, Status='live', Symbol='EFHB', Aliases='CFAP21', description='EF-hand domain family member B', other_designations='EF-hand domain family, member B|cilia and flagella associated protein 21', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19879472, end_position_on_the_genomic_accession=19947456, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100421733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421733, CurrentID=0, Status='live', Symbol='EEF1GP6', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96750800, end_position_on_the_genomic_accession=96751728, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421734, CurrentID=0, Status='live', Symbol='NEK4P2', Aliases='', description='NIMA-related kinase 4 pseudogene 2', other_designations='NIMA (never in mitosis gene a)-related kinase 4 pseudogene 2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16666105, end_position_on_the_genomic_accession=16667113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421736, CurrentID=0, Status='live', Symbol='BNIP3P3', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 3', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15247456, end_position_on_the_genomic_accession=15247984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151657, CurrentID=0, Status='live', Symbol='LOC151657', Aliases='', description='uncharacterized LOC151657', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151658, CurrentID=0, Status='live', Symbol='LINC00635', Aliases='', description='long intergenic non-protein coding RNA 635', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107841662, end_position_on_the_genomic_accession=107878068, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100421741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421741, CurrentID=0, Status='live', Symbol='ARAFP1', Aliases='', description='ARAF pseudogene 1', other_designations='A-Raf proto-oncogene, serine/threonine kinase pseudogene|v-raf murine sarcoma 3611 viral oncogene homolog pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63937707, end_position_on_the_genomic_accession=63938512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421742, CurrentID=0, Status='live', Symbol='LOC100421742', Aliases='', description='NIF3 NGG1 interacting factor 3-like 1 (S. cerevisiae)', other_designations='NIF3 NGG1 interacting factor 3-like 1 (S. pombe) pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3465515, end_position_on_the_genomic_accession=3466442, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101059944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059944, CurrentID=0, Status='live', Symbol='LOC101059944', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421746, CurrentID=0, Status='live', Symbol='LOC100421746', Aliases='', description='PAGE family member 5 pseudogene', other_designations='P antigen family, member 5 (prostate associated) pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55280088, end_position_on_the_genomic_accession=55288786, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421749, CurrentID=0, Status='live', Symbol='LOC100421749', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203656733, end_position_on_the_genomic_accession=203657875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  134505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134505, CurrentID=0, Status='live', Symbol='EIF3KP1', Aliases='', description='eukaryotic translation initiation factor 3 subunit K pseudogene 1', other_designations='eukaryotic translation initiation factor 3, subunit K pseudogene 1', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103032313, end_position_on_the_genomic_accession=103033061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271175, CurrentID=0, Status='live', Symbol='RPL21P60', Aliases='RPL21_23_641', description='ribosomal protein L21 pseudogene 60', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177264726, end_position_on_the_genomic_accession=177265274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270847, CurrentID=0, Status='live', Symbol='RPL6P9', Aliases='RPL6_6_412', description='ribosomal protein L6 pseudogene 9', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142581289, end_position_on_the_genomic_accession=142581978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282754, CurrentID=0, Status='live', Symbol='OR4D12P', Aliases='OR7E103P', description='olfactory receptor family 4 subfamily D member 12 pseudogene', other_designations='olfactory receptor, family 4, subfamily D, member 12 pseudogene|olfactory receptor, family 7, subfamily E, member 103 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4126606, end_position_on_the_genomic_accession=4127630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282756, CurrentID=0, Status='live', Symbol='OR7E149P', Aliases='', description='olfactory receptor family 7 subfamily E member 149 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 149 pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8437402, end_position_on_the_genomic_accession=8438382, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421765, CurrentID=0, Status='live', Symbol='LOC100421765', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59401080, end_position_on_the_genomic_accession=59402112, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421766, CurrentID=0, Status='live', Symbol='UBE2Q2P7', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 7', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 7', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421767, CurrentID=0, Status='live', Symbol='UBE2Q2P11', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 11', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 11', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84412768, end_position_on_the_genomic_accession=84420775, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421768, CurrentID=0, Status='live', Symbol='UBE2Q2P12', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 12', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 12', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84491474, end_position_on_the_genomic_accession=84499537, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421769, CurrentID=0, Status='live', Symbol='UBE2Q2P6', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 6', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 6', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82446712, end_position_on_the_genomic_accession=82454702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282763, CurrentID=0, Status='live', Symbol='OR51B5', Aliases="HOR5'Beta5, OR11-37", description='olfactory receptor family 51 subfamily B member 5', other_designations="odorant receptor HOR5'beta5|olfactory receptor OR11-37|olfactory receptor, family 51, subfamily B, member 5", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5340479, end_position_on_the_genomic_accession=5505652, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421774, CurrentID=0, Status='live', Symbol='EEF1B2P8', Aliases='', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 8', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175059315, end_position_on_the_genomic_accession=175060093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059949, CurrentID=0, Status='live', Symbol='LOC101059949', Aliases='', description='putative ankyrin repeat domain-containing protein 20A12 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421776, CurrentID=0, Status='live', Symbol='NIPA2P4', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2 pseudogene 4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81617147, end_position_on_the_genomic_accession=81618245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282769, CurrentID=0, Status='live', Symbol='OR10AF1P', Aliases='', description='olfactory receptor family 10 subfamily AF member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily AF, member 1 pseudogene', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55948119, end_position_on_the_genomic_accession=55948923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282770, CurrentID=0, Status='live', Symbol='OR10AG1', Aliases='OR11-160', description='olfactory receptor family 10 subfamily AG member 1', other_designations='olfactory receptor OR11-160|olfactory receptor, family 10, subfamily AG, member 1', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55967558, end_position_on_the_genomic_accession=55968463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421779, CurrentID=0, Status='live', Symbol='ACSM5P1', Aliases='', description='acyl-CoA synthetase medium-chain family member 5 pseudogene 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20587965, end_position_on_the_genomic_accession=20607169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  134510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134510, CurrentID=0, Status='live', Symbol='UBLCP1', Aliases='CPUB1', description='ubiquitin like domain containing CTD phosphatase 1', other_designations='CTD phosphatase-like with ubiquitin domain 1|CTD-like phosphatase domain-containing protein|nuclear proteasome inhibitor UBLCP1|ubiquitin-like domain containing CTD phosphatase 1', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159263081, end_position_on_the_genomic_accession=159286040, orientation='plus', exon_count=11, OMIM=609867, no_hdr0=''),
  100421782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421782, CurrentID=0, Status='live', Symbol='LOC100421782', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1-like 2 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117262818, end_position_on_the_genomic_accession=117264892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282775, CurrentID=0, Status='live', Symbol='OR5J2', Aliases='OR11-266', description='olfactory receptor family 5 subfamily J member 2', other_designations='olfactory receptor OR11-266|olfactory receptor, family 5, subfamily J, member 2', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56176618, end_position_on_the_genomic_accession=56177556, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421784, CurrentID=0, Status='live', Symbol='STK24P1', Aliases='', description='serine/threonine kinase 24 pseudogene 1', other_designations='serine/threonine kinase 24 (STE20 homolog, yeast) pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136295493, end_position_on_the_genomic_accession=136296933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421787, CurrentID=0, Status='live', Symbol='LOC100421787', Aliases='', description='sulfotransferase family 1C member 2 pseudogene', other_designations='sulfotransferase family, cytosolic, 1C, member 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34416767, end_position_on_the_genomic_accession=34417566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421788, CurrentID=0, Status='live', Symbol='LOC100421788', Aliases='', description='lysine (K)-specific methyltransferase 5A pseudogene', other_designations='SET domain containing (lysine methyltransferase) 8 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91747943, end_position_on_the_genomic_accession=91748986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421790, CurrentID=0, Status='live', Symbol='PUS7P1', Aliases='', description='pseudouridylate synthase 7 pseudogene 1', other_designations='pseudouridylate synthase 7 homolog pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18153568, end_position_on_the_genomic_accession=18154216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421793, CurrentID=0, Status='live', Symbol='LOC100421793', Aliases='', description='tRNA methyltransferase 11 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58072381, end_position_on_the_genomic_accession=58073628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421794, CurrentID=0, Status='live', Symbol='HNRNPCP6', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 6', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85020781, end_position_on_the_genomic_accession=85021653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421795, CurrentID=0, Status='live', Symbol='ST3GAL5P1', Aliases='', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 5 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33656110, end_position_on_the_genomic_accession=33656676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421796, CurrentID=0, Status='live', Symbol='EEF1A1P19', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 19', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43495076, end_position_on_the_genomic_accession=43496454, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421797, CurrentID=0, Status='live', Symbol='EEF1A1P20', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 20', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=99996547, end_position_on_the_genomic_accession=99997931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282790, CurrentID=0, Status='live', Symbol='OR5BT1P', Aliases='', description='olfactory receptor family 5 subfamily BT member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BT, member 1 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48385350, end_position_on_the_genomic_accession=48386304, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059953, CurrentID=0, Status='live', Symbol='NPIPA8', Aliases='LCR16a9', description='nuclear pore complex interacting protein family member A8', other_designations='morpheus gene family member 9|nuclear pore complex interacting protein family, member A8', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18317919, end_position_on_the_genomic_accession=18333112, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100421802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421802, CurrentID=0, Status='live', Symbol='LOC100421802', Aliases='', description='protein phosphatase 2 regulatory subunit B, alpha pseudogene', other_designations='protein phosphatase 2, regulatory subunit B, alpha pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1100044, end_position_on_the_genomic_accession=1101702, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  352967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352967, CurrentID=0, Status='live', Symbol='MICG', Aliases='', description='MHC class I polypeptide-related sequence G (pseudogene)', other_designations='MHC class I polypeptide-related sequence G pseudogene|MHC class I polypeptide-related sequence G, pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29812390, end_position_on_the_genomic_accession=29812692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421804, CurrentID=0, Status='live', Symbol='ST3GAL1P1', Aliases='', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 1 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67716378, end_position_on_the_genomic_accession=67717388, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421805, CurrentID=0, Status='live', Symbol='ENPP7P10', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 10', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9078958, end_position_on_the_genomic_accession=9141608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421806, CurrentID=0, Status='live', Symbol='ENPP7P11', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 11', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9677308, end_position_on_the_genomic_accession=9677934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421807, CurrentID=0, Status='live', Symbol='EEF1A1P21', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 21', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29748759, end_position_on_the_genomic_accession=29750097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421808, CurrentID=0, Status='live', Symbol='LOC100421808', Aliases='', description='protein phosphatase 2 regulatory subunit B, gamma pseudogene', other_designations='protein phosphatase 2, regulatory subunit B, gamma pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=60784153, end_position_on_the_genomic_accession=60784696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  282801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282801, CurrentID=0, Status='live', Symbol='OR7E148P', Aliases='OR7E150P', description='olfactory receptor family 7 subfamily E member 148 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 148 pseudogene|olfactory receptor, family 7, subfamily E, member 150 pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8427661, end_position_on_the_genomic_accession=8428613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421810, CurrentID=0, Status='live', Symbol='LOC100421810', Aliases='', description='LON peptidase N-terminal domain and ring finger 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21344549, end_position_on_the_genomic_accession=21345203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421811, CurrentID=0, Status='live', Symbol='LOC100421811', Aliases='', description='fumarylacetoacetate hydrolase domain containing 1 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69434467, end_position_on_the_genomic_accession=69435120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421813, CurrentID=0, Status='live', Symbol='ENPP7P2', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75446765, end_position_on_the_genomic_accession=75491203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271177, CurrentID=0, Status='live', Symbol='RPL21P64', Aliases='RPL21_25_752', description='ribosomal protein L21 pseudogene 64', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128639188, end_position_on_the_genomic_accession=128639653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421815, CurrentID=0, Status='live', Symbol='EEF1A1P23', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 23', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193578815, end_position_on_the_genomic_accession=193579413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282808, CurrentID=0, Status='live', Symbol='RAB40AL', Aliases='MRXSMP, RAR2, RLGP', description='RAB40A, member RAS oncogene family-like', other_designations='Ras like GTPase|SOCS box containing protein RAR2|ras-like GTPase', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102937272, end_position_on_the_genomic_accession=102938300, orientation='plus', exon_count=1, OMIM=300405, no_hdr0=''),
  282809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282809, CurrentID=0, Status='live', Symbol='POC1B', Aliases='CORD20, PIX1, TUWD12, WDR51B', description='POC1 centriolar protein B', other_designations='WD repeat-containing protein 51B|proteome of centriole protein 1B', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89419718, end_position_on_the_genomic_accession=89526262, orientation='minus', exon_count=13, OMIM=614784, no_hdr0=''),
  100421819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421819, CurrentID=0, Status='live', Symbol='LOC100421819', Aliases='', description='lysine (K)-specific methyltransferase 5A pseudogene', other_designations='SET domain containing (lysine methyltransferase) 8 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170062340, end_position_on_the_genomic_accession=170062953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421820, CurrentID=0, Status='live', Symbol='EEF1A1P36', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 36', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132271982, end_position_on_the_genomic_accession=132273329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421821, CurrentID=0, Status='live', Symbol='LOC100421821', Aliases='', description='protein phosphatase 4 regulatory subunit 3B pseudogene', other_designations='SMEK homolog 2, suppressor of mek1 (Dictyostelium) pseudogene|protein phosphatase 4, regulatory subunit 3B pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26987126, end_position_on_the_genomic_accession=26988031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151742, CurrentID=0, Status='live', Symbol='PPM1L', Aliases='PP2C-epsilon, PP2CE, PPM1-LIKE', description='protein phosphatase, Mg2+/Mn2+ dependent 1L', other_designations='PP2C epsilon|Protein phosphatase 2C epsilon isoform|protein phosphatase 2C isoform epsilon|protein phosphatase, Mg2+/Mn2+ dependent, 1L', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160755532, end_position_on_the_genomic_accession=161078902, orientation='plus', exon_count=6, OMIM=611931, no_hdr0=''),
  100421823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421823, CurrentID=0, Status='live', Symbol='ENPP7P1', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8154365, end_position_on_the_genomic_accession=8216916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421824, CurrentID=0, Status='live', Symbol='LOC100421824', Aliases='', description='polymerase (DNA directed), epsilon 2, accessory subunit pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53394235, end_position_on_the_genomic_accession=53395371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421825, CurrentID=0, Status='live', Symbol='EEF1A1P37', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 37', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111225636, end_position_on_the_genomic_accession=111226192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421827, CurrentID=0, Status='live', Symbol='LONRF2P1', Aliases='', description='LON peptidase N-terminal domain and ring finger 2 pseudogene 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14450219, end_position_on_the_genomic_accession=14450883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421828, CurrentID=0, Status='live', Symbol='HNRNPCP4', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 4', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11242653, end_position_on_the_genomic_accession=11243549, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421829, CurrentID=0, Status='live', Symbol='EEF1A1P10', Aliases='EEF1AL8', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 10', other_designations='eukaryotic translation elongation factor 1 alpha-like 8', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144647189, end_position_on_the_genomic_accession=144648575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104355139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355139, CurrentID=0, Status='live', Symbol='LINC01397', Aliases='', description='long intergenic non-protein coding RNA 1397', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110082688, end_position_on_the_genomic_accession=110109719, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  151760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151760, CurrentID=0, Status='live', Symbol='LOC151760', Aliases='', description='putative uncharacterized protein LOC151760', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=110892123, end_position_on_the_genomic_accession=110970917, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421842, CurrentID=0, Status='live', Symbol='LOC100421842', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228776312, end_position_on_the_genomic_accession=228776865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421845, CurrentID=0, Status='live', Symbol='LOC100421845', Aliases='', description='olfactory receptor, family 1, subfamily S, member 2 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421846, CurrentID=0, Status='live', Symbol='LOC100421846', Aliases='', description='olfactory receptor, family 9, subfamily G, member 4 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421848, CurrentID=0, Status='live', Symbol='LOC100421848', Aliases='', description='olfactory receptor, family 2, subfamily S, member 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421849, CurrentID=0, Status='live', Symbol='LOC100421849', Aliases='', description='olfactory receptor, family 2, subfamily S, member 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421850, CurrentID=0, Status='live', Symbol='LOC100421850', Aliases='', description='olfactory receptor, family 2, subfamily S, member 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421852, CurrentID=0, Status='live', Symbol='LOC100421852', Aliases='', description='olfactory receptor, family 2, subfamily K, member 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421853, CurrentID=0, Status='live', Symbol='LOC100421853', Aliases='', description='olfactory receptor, family 1, subfamily N, member 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421854, CurrentID=0, Status='live', Symbol='EEF1A1P4', Aliases='EEF1AL2', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 4', other_designations='eukaryotic translation elongation factor 1 alpha 2 pseudogene|eukaryotic translation elongation factor 1 alpha-like 2', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19456251, end_position_on_the_genomic_accession=19457648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421855, CurrentID=0, Status='live', Symbol='LOC100421855', Aliases='', description='SCO1 cytochrome c oxidase assembly protein pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48160249, end_position_on_the_genomic_accession=48161104, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282849, CurrentID=0, Status='live', Symbol='SLEN3', Aliases='', description='systemic lupus erythematosus with nephritis 3', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607967, no_hdr0=''),
  100421859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421859, CurrentID=0, Status='live', Symbol='ENPP7P7', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 7', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67812542, end_position_on_the_genomic_accession=67873371, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421860, CurrentID=0, Status='live', Symbol='ENPP7P8', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 8', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71722143, end_position_on_the_genomic_accession=71780314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421861, CurrentID=0, Status='live', Symbol='CBX3P3', Aliases='', description='chromobox 3 pseudogene 3', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112777643, end_position_on_the_genomic_accession=112778189, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421863, CurrentID=0, Status='live', Symbol='LOC100421863', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila) pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86402329, end_position_on_the_genomic_accession=86402942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421864, CurrentID=0, Status='live', Symbol='LOC100421864', Aliases='', description='olfactory receptor, family 7, subfamily D, member 2 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421865, CurrentID=0, Status='live', Symbol='EEF1A1P3', Aliases='EEF1AL, EEF1AL1', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 3', other_designations='EEF1A1 (Eukaryotic translation elongation factor 1 alpha 1) pseudogene|eukaryotic translation elongation factor 1 alpha 2 pseudogene|eukaryotic translation elongation factor 1 alpha-like 1 pseudogene', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28214522, end_position_on_the_genomic_accession=28215891, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373303, CurrentID=0, Status='live', Symbol='LOC105373303', Aliases='', description='uncharacterized LOC105373303', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=105663131, end_position_on_the_genomic_accession=105773377, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100421868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421868, CurrentID=0, Status='live', Symbol='LOC100421868', Aliases='', description='glucose-fructose oxidoreductase domain containing 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139426843, end_position_on_the_genomic_accession=139430884, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151790, CurrentID=0, Status='live', Symbol='WDR49', Aliases='', description='WD repeat domain 49', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167478684, end_position_on_the_genomic_accession=167653501, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100421874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421874, CurrentID=0, Status='live', Symbol='LOC100421874', Aliases='', description='olfactory receptor, family 5, subfamily H, member 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  134526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134526, CurrentID=0, Status='live', Symbol='ACOT12', Aliases='CACH-1, Cach, STARD15, THEAL', description='acyl-CoA thioesterase 12', other_designations='START domain-containing protein 15|StAR-related lipid transfer (START) domain containing 15|acyl-CoA thioester hydrolase 12|cytoplasmic acetyl-CoA hydrolase 1|cytosolic acetyl-CoA hydrolase|hCACH-1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81329991, end_position_on_the_genomic_accession=81394169, orientation='minus', exon_count=16, OMIM=614315, no_hdr0=''),
  100421880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421880, CurrentID=0, Status='live', Symbol='LOC100421880', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421885, CurrentID=0, Status='live', Symbol='LOC100421885', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421886, CurrentID=0, Status='live', Symbol='LOC100421886', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421887, CurrentID=0, Status='live', Symbol='LOC100421887', Aliases='', description='olfactory receptor, family 2, subfamily W, member 3 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421893, CurrentID=0, Status='live', Symbol='LOC100421893', Aliases='', description='insulin like growth factor 2 mRNA binding protein 2 pseudogene', other_designations='insulin-like growth factor 2 mRNA binding protein 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99233052, end_position_on_the_genomic_accession=99234459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421894, CurrentID=0, Status='live', Symbol='LOC100421894', Aliases='', description='exportin, tRNA (nuclear export receptor for tRNAs) pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76885082, end_position_on_the_genomic_accession=76888714, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282890, CurrentID=0, Status='live', Symbol='ZNF311', Aliases='zf31', description='zinc finger protein 311', other_designations='zinc finger protein zfp-31|zinc finger protein zfp31', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28994785, end_position_on_the_genomic_accession=29005628, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421901, CurrentID=0, Status='live', Symbol='LOC100421901', Aliases='', description='immunoglobulin (CD79A) binding protein 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=109763681, end_position_on_the_genomic_accession=109764377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421902, CurrentID=0, Status='live', Symbol='LOC100421902', Aliases='', description='olfactory receptor, family 9, subfamily A, member 2 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421904, CurrentID=0, Status='live', Symbol='LOC100421904', Aliases='', description='olfactory receptor, family 2, subfamily F, member 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151825, CurrentID=0, Status='live', Symbol='KRT18P43', Aliases='', description='keratin 18 pseudogene 43', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169902913, end_position_on_the_genomic_accession=169904258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151827, CurrentID=0, Status='live', Symbol='LRRC34', Aliases='', description='leucine rich repeat containing 34', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169793006, end_position_on_the_genomic_accession=169812866, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100421909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421909, CurrentID=0, Status='live', Symbol='LOC100421909', Aliases='', description='olfactory receptor, family 8, subfamily B, member 4 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421910, CurrentID=0, Status='live', Symbol='LOC100421910', Aliases='', description='olfactory receptor, family 5, subfamily R, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131437064, end_position_on_the_genomic_accession=131437690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270848, CurrentID=0, Status='live', Symbol='RPL6P7', Aliases='RPL6_7_423', description='ribosomal protein L6 pseudogene 7', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155657743, end_position_on_the_genomic_accession=155658644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421912, CurrentID=0, Status='live', Symbol='LOC100421912', Aliases='', description='olfactory receptor, family 1, subfamily N, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101059972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059972, CurrentID=0, Status='live', Symbol='LOC101059972', Aliases='', description='myosin light chain 9 pseudogene', other_designations='myosin, light chain 9, regulatory pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42128158, end_position_on_the_genomic_accession=42129337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151835, CurrentID=0, Status='live', Symbol='CPNE9', Aliases='', description='copine family member IX', other_designations='copine IX|copine-like protein', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9703812, end_position_on_the_genomic_accession=9729908, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  101059973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059973, CurrentID=0, Status='live', Symbol='NACAP7', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 7', other_designations='nascent-polypeptide-associated complex alpha polypeptide pseudogene 7', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89422099, end_position_on_the_genomic_accession=89428245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  151842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151842, CurrentID=0, Status='live', Symbol='PP1P', Aliases='', description='pyrophosphatase (inorganic) 1 pseudogene', other_designations='inorganic pyrophosphatase pseudogene', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17871417, end_position_on_the_genomic_accession=17872701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421925, CurrentID=0, Status='live', Symbol='LOC100421925', Aliases='', description='MOB family member 4, phocein pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70810801, end_position_on_the_genomic_accession=70811474, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421926, CurrentID=0, Status='live', Symbol='NACAP4', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 4', other_designations='nascent polypeptide-associated complex alpha subunit pseudogene|nascent-polypeptide-associated complex alpha polypeptide pseudogene 4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13583688, end_position_on_the_genomic_accession=13584294, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421927, CurrentID=0, Status='live', Symbol='LOC100421927', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421929, CurrentID=0, Status='live', Symbol='LOC100421929', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421933, CurrentID=0, Status='live', Symbol='LOC100421933', Aliases='', description='olfactory receptor, family 13, subfamily C, member 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271181, CurrentID=0, Status='live', Symbol='RPL21P72', Aliases='RPL21_28_779', description='ribosomal protein L21 pseudogene 72', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3338085, end_position_on_the_genomic_accession=3338626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421935, CurrentID=0, Status='live', Symbol='LOC100421935', Aliases='', description='olfactory receptor, family 13, subfamily C, member 8 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421937, CurrentID=0, Status='live', Symbol='LOC100421937', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421938, CurrentID=0, Status='live', Symbol='LOC100421938', Aliases='', description='olfactory receptor, family 13, subfamily D, member 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421939, CurrentID=0, Status='live', Symbol='LOC100421939', Aliases='', description='olfactory receptor, family 5, subfamily M, member 8 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421941, CurrentID=0, Status='live', Symbol='LOC100421941', Aliases='', description='olfactory receptor, family 6, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421942, CurrentID=0, Status='live', Symbol='LOC100421942', Aliases='', description='olfactory receptor, family 6, subfamily C, member 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421943, CurrentID=0, Status='live', Symbol='CBX3P5', Aliases='', description='chromobox 3 pseudogene 5', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 5', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95286320, end_position_on_the_genomic_accession=95286838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421944, CurrentID=0, Status='live', Symbol='LOC100421944', Aliases='', description='olfactory receptor, family 9, subfamily K, member 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421945, CurrentID=0, Status='live', Symbol='LOC100421945', Aliases='', description='olfactory receptor, family 6, subfamily C, member 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421946, CurrentID=0, Status='live', Symbol='LOC100421946', Aliases='', description='olfactory receptor, family 6, subfamily C, member 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421947, CurrentID=0, Status='live', Symbol='LOC100421947', Aliases='', description='olfactory receptor, family 6, subfamily C, member 3 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421948, CurrentID=0, Status='live', Symbol='LOC100421948', Aliases='', description='olfactory receptor, family 6, subfamily C, member 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421950, CurrentID=0, Status='live', Symbol='LOC100421950', Aliases='', description='olfactory receptor, family 7, subfamily C, member 1 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151871, CurrentID=0, Status='live', Symbol='DPPA2', Aliases='CT100, ECAT15-2, PESCRG1', description='developmental pluripotency associated 2', other_designations='cancer/testis antigen 100|embryonic stem cell (ESC) associated transcript 15-2|pluripotent embryonic stem cell-related gene 1 protein', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109293788, end_position_on_the_genomic_accession=109316535, orientation='minus', exon_count=12, OMIM=614445, no_hdr0=''),
  106481894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481894, CurrentID=0, Status='live', Symbol='RNU6-428P', Aliases='', description='RNA, U6 small nuclear 428, pseudogene', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421954, CurrentID=0, Status='live', Symbol='LOC100421954', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421955, CurrentID=0, Status='live', Symbol='LOC100421955', Aliases='', description='olfactory receptor, family 8, subfamily U, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421956, CurrentID=0, Status='live', Symbol='LOC100421956', Aliases='', description='olfactory receptor, family 5, subfamily F, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151877, CurrentID=0, Status='live', Symbol='MAGI1-IT1', Aliases='', description='MAGI1 intronic transcript 1', other_designations='MAGI1 intronic transcript 1 (non-protein coding)', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65872814, end_position_on_the_genomic_accession=65954611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421958, CurrentID=0, Status='live', Symbol='LOC100421958', Aliases='', description='olfactory receptor, family 8, subfamily K, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421960, CurrentID=0, Status='live', Symbol='LOC100421960', Aliases='', description='olfactory receptor, family 8, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421961, CurrentID=0, Status='live', Symbol='LOC100421961', Aliases='', description='olfactory receptor, family 8, subfamily I, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421962, CurrentID=0, Status='live', Symbol='LOC100421962', Aliases='', description='olfactory receptor, family 5, subfamily M, member 8 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421963, CurrentID=0, Status='live', Symbol='LOC100421963', Aliases='', description='olfactory receptor, family 5, subfamily M, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271182, CurrentID=0, Status='live', Symbol='RPL21P73', Aliases='RPL21_28_856', description='ribosomal protein L21 pseudogene 73', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138646097, end_position_on_the_genomic_accession=138646490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421965, CurrentID=0, Status='live', Symbol='LOC100421965', Aliases='', description='olfactory receptor, family 5, subfamily R, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151887, CurrentID=0, Status='live', Symbol='CCDC80', Aliases='DRO1, SSG1, URB, okuribin', description='coiled-coil domain containing 80', other_designations='down-regulated by oncogenes protein 1|nuclear envelope protein okuribin|steroid sensitive gene 1|up-regulated in BRS-3 deficient mouse homolog', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112604386, end_position_on_the_genomic_accession=112641143, orientation='minus', exon_count=8, OMIM=608298, no_hdr0=''),
  151888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151888, CurrentID=0, Status='live', Symbol='BTLA', Aliases='BTLA1, CD272', description='B and T lymphocyte associated', other_designations='B and T lymphocyte attenuator|B- and T-lymphocyte-associated protein', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112463966, end_position_on_the_genomic_accession=112499702, orientation='minus', exon_count=5, OMIM=607925, no_hdr0=''),
  100421971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421971, CurrentID=0, Status='live', Symbol='LOC100421971', Aliases='', description='olfactory receptor, family 5, subfamily B, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421972, CurrentID=0, Status='live', Symbol='LOC100421972', Aliases='', description='olfactory receptor, family 5, subfamily A, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421973, CurrentID=0, Status='live', Symbol='ARPC3P4', Aliases='', description='actin related protein 2/3 complex subunit 3 pseudogene 4', other_designations='actin related protein 2/3 complex, subunit 3 pseudogene 4|actin related protein 2/3 complex, subunit 3, 21kDa pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73921404, end_position_on_the_genomic_accession=73922227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282966, CurrentID=0, Status='live', Symbol='C10orf53', Aliases='', description='chromosome 10 open reading frame 53', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49679638, end_position_on_the_genomic_accession=49708910, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  282967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282967, CurrentID=0, Status='live', Symbol='MAPK6PS6', Aliases='', description='mitogen-activated protein kinase 6 pseudogene 6', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49771063, end_position_on_the_genomic_accession=49774418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421976, CurrentID=0, Status='live', Symbol='LOC100421976', Aliases='', description='interferon stimulated exonuclease gene 20kDa like 2 pseudogene', other_designations='interferon stimulated exonuclease gene 20kDa-like 2 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113820414, end_position_on_the_genomic_accession=113821902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282969, CurrentID=0, Status='live', Symbol='FUOM', Aliases='C10orf125, FUCU, FucM', description='fucose mutarotase', other_designations='protein fucU homolog', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133352775, end_position_on_the_genomic_accession=133358025, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421978, CurrentID=0, Status='live', Symbol='LOC100421978', Aliases='', description='olfactory receptor, family 6, subfamily M, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  282973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282973, CurrentID=0, Status='live', Symbol='JAKMIP3', Aliases='C10orf14, C10orf39, Jamip3, NECC2, bA140A10.5', description='Janus kinase and microtubule interacting protein 3', other_designations='neuroendocrine long coiled-coil 2|neuroendocrine long coiled-coil protein 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132036107, end_position_on_the_genomic_accession=132184809, orientation='plus', exon_count=29, OMIM=611198, no_hdr0=''),
  282974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282974, CurrentID=0, Status='live', Symbol='STK32C', Aliases='PKE, YANK3', description='serine/threonine kinase 32C', other_designations='PKE protein kinase|protein kinase (PKE)|yet another novel kinase 3', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132207489, end_position_on_the_genomic_accession=132332256, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  151903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151903, CurrentID=0, Status='live', Symbol='CCDC12', Aliases='', description='coiled-coil domain containing 12', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46921730, end_position_on_the_genomic_accession=46982018, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100421984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421984, CurrentID=0, Status='live', Symbol='LOC100421984', Aliases='', description='olfactory receptor, family 8, subfamily B, member 8 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421985, CurrentID=0, Status='live', Symbol='LOC100421985', Aliases='', description='olfactory receptor, family 8, subfamily B, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124538542, end_position_on_the_genomic_accession=124539318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421986, CurrentID=0, Status='live', Symbol='LOC100421986', Aliases='', description='olfactory receptor, family 5, subfamily P, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  282980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282980, CurrentID=0, Status='live', Symbol='LINC00700', Aliases='', description='long intergenic non-protein coding RNA 700', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2005473, end_position_on_the_genomic_accession=2014348, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421989, CurrentID=0, Status='live', Symbol='LOC100421989', Aliases='', description='olfactory receptor, family 2, subfamily V, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421990, CurrentID=0, Status='live', Symbol='LOC100421990', Aliases='', description='olfactory receptor, family 4, subfamily B, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421991, CurrentID=0, Status='live', Symbol='LOC100421991', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421992, CurrentID=0, Status='live', Symbol='LOC100421992', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421993, CurrentID=0, Status='live', Symbol='LOC100421993', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414058, CurrentID=0, Status='live', Symbol='ACRPV', Aliases='', description='Acropectorovertebral dysplasia (F syndrome)', other_designations='', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=102510, no_hdr0=''),
  414059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414059, CurrentID=0, Status='live', Symbol='TBC1D3B', Aliases='PRC17, TBC1D3I, TBC1D3L', description='TBC1 domain family member 3B', other_designations='Rab GTPase-activating protein PRC17 duplicate|TBC1 domain family member 3B/I|TBC1 domain family, member 3-like|TBC1 domain family, member 3B|TBC1 domain family, member 3I|prostate cancer gene 17 duplicate', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36165681, end_position_on_the_genomic_accession=36176624, orientation='minus', exon_count=14, OMIM=610144, no_hdr0=''),
  414060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414060, CurrentID=0, Status='live', Symbol='TBC1D3C', Aliases='TBC1D3D', description='TBC1 domain family member 3C', other_designations='Rab GTPase-activating protein PRC17|TBC1 domain family member 3C-like protein ENSP00000341742|TBC1 domain family member 3C/3D|TBC1 domain family, member 3C', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38057693, end_position_on_the_genomic_accession=38068634, orientation='minus', exon_count=15, OMIM=610806, no_hdr0=''),
  414061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414061, CurrentID=0, Status='live', Symbol='DNAJB3', Aliases='HCG3', description='DnaJ heat shock protein family (Hsp40) member B3', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 3', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233742750, end_position_on_the_genomic_accession=233744015, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  414062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414062, CurrentID=0, Status='live', Symbol='CCL3L3', Aliases='464.2, D17S1718, G0S19-2, LD78, LD78BETA, SCYA3L, SCYA3L1', description='chemokine (C-C motif) ligand 3-like 3', other_designations='G0/G1 switch regulatory protein 19-2|LD78-beta(1-70)|chemokine (C-C motif) ligand 3-like, centromeric|small inducible cytokine A3-like 1|tonsillar lymphocyte LD78 beta protein', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36194869, end_position_on_the_genomic_accession=36196748, orientation='minus', exon_count=3, OMIM=609468, no_hdr0=''),
  282991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282991, CurrentID=0, Status='live', Symbol='BLOC1S2', Aliases='BLOS2, CEAP, CEAP11', description='biogenesis of lysosomal organelles complex-1, subunit 2', other_designations='11 kDa centrosome associated protein|BLOC-1 subunit 2|centrosomal 10 kDa protein|centrosome protein oncogene', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100273278, end_position_on_the_genomic_accession=100286712, orientation='minus', exon_count=7, OMIM=609768, no_hdr0=''),
  100422000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422000, CurrentID=0, Status='live', Symbol='LOC100422000', Aliases='', description='olfactory receptor, family 5, subfamily J, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422001, CurrentID=0, Status='live', Symbol='LOC100422001', Aliases='', description='olfactory receptor, family 5, subfamily J, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377672, CurrentID=0, Status='live', Symbol='LOC105377672', Aliases='', description='uncharacterized LOC105377672', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=58776422, end_position_on_the_genomic_accession=58786610, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422003, CurrentID=0, Status='live', Symbol='LOC100422003', Aliases='', description='olfactory receptor, family 8, subfamily K, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  282996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282996, CurrentID=0, Status='live', Symbol='RBM20', Aliases='', description='RNA binding motif protein 20', other_designations='probable RNA-binding protein 20', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110644397, end_position_on_the_genomic_accession=110839471, orientation='plus', exon_count=15, OMIM=613171, no_hdr0=''),
  282997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282997, CurrentID=0, Status='live', Symbol='PDCD4-AS1', Aliases='', description='PDCD4 antisense RNA 1', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110868890, end_position_on_the_genomic_accession=110870904, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422006, CurrentID=0, Status='live', Symbol='LOC100422006', Aliases='', description='olfactory receptor, family 8, subfamily H, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422007, CurrentID=0, Status='live', Symbol='LOC100422007', Aliases='', description='olfactory receptor, family 9, subfamily G, member 9 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  134548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134548, CurrentID=0, Status='live', Symbol='SOWAHA', Aliases='ANKRD43', description='sosondowah ankyrin repeat domain family member A', other_designations='ankyrin repeat domain-containing protein 43|protein sosondowah homolog A', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132813324, end_position_on_the_genomic_accession=132816797, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422010, CurrentID=0, Status='live', Symbol='LOC100422010', Aliases='', description='olfactory receptor, family 8, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422011, CurrentID=0, Status='live', Symbol='LOC100422011', Aliases='', description='olfactory receptor, family 8, subfamily A, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422012, CurrentID=0, Status='live', Symbol='LOC100422012', Aliases='', description='olfactory receptor, family 8, subfamily A, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422013, CurrentID=0, Status='live', Symbol='LOC100422013', Aliases='', description='catenin beta 1 pseudogene', other_designations='catenin (cadherin-associated protein), beta 1, 88kDa pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71754381, end_position_on_the_genomic_accession=71756311, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283008, CurrentID=0, Status='live', Symbol='NUTM2E', Aliases='FAM22E', description='NUT family member 2E', other_designations='family with sequence similarity 22, member E', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79841358, end_position_on_the_genomic_accession=79850876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373308, CurrentID=0, Status='live', Symbol='LOC105373308', Aliases='', description='uncharacterized LOC105373308', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107589567, end_position_on_the_genomic_accession=107597286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422018, CurrentID=0, Status='live', Symbol='LOC100422018', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  353003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353003, CurrentID=0, Status='live', Symbol='HCG4P5', Aliases='HCGIV-05, HCGIV-5, HCGIV.4', description='HLA complex group 4 pseudogene 5', other_designations='HCGIV-05 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29940912, end_position_on_the_genomic_accession=29941801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422020, CurrentID=0, Status='live', Symbol='LOC100422020', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B17 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69021656, end_position_on_the_genomic_accession=69022340, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059990, CurrentID=0, Status='live', Symbol='LOC101059990', Aliases='', description='dnaJ homolog subfamily B member 6-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219683835, end_position_on_the_genomic_accession=219684839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283014, CurrentID=0, Status='live', Symbol='EIF2S2P3', Aliases='', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 3', other_designations='eukaryotic initiation factor 2-beta|eukaryotic translation initiation factor 2, subunit 2 (beta, 38kD )|eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 3|eukaryotic translation initiation factor 2, subunit 2 beta, 38kDa pseudogene', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92668456, end_position_on_the_genomic_accession=92669847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422023, CurrentID=0, Status='live', Symbol='UGT2A3P7', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene 7', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69517728, end_position_on_the_genomic_accession=69518379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271184, CurrentID=0, Status='live', Symbol='RPL21P74', Aliases='RPL21_30_821', description='ribosomal protein L21 pseudogene 74', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96283313, end_position_on_the_genomic_accession=96283853, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422026, CurrentID=0, Status='live', Symbol='LOC100422026', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68784419, end_position_on_the_genomic_accession=68787476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422029, CurrentID=0, Status='live', Symbol='LOC100422029', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B7 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69406977, end_position_on_the_genomic_accession=69407907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353005, CurrentID=0, Status='live', Symbol='HCG4P8', Aliases='HCGIV-08, HCGIV-8, HCGIV.7', description='HLA complex group 4 pseudogene 8', other_designations='HCGIV-08 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29826129, end_position_on_the_genomic_accession=29827114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422032, CurrentID=0, Status='live', Symbol='LOC100422032', Aliases='', description='intraflagellar transport 80 homolog (Chlamydomonas) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=54048179, end_position_on_the_genomic_accession=54050877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283025, CurrentID=0, Status='live', Symbol='LINC01553', Aliases='C10orf40', description='long intergenic non-protein coding RNA 1553', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59958217, end_position_on_the_genomic_accession=59960913, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422034, CurrentID=0, Status='live', Symbol='LOC100422034', Aliases='', description='olfactory receptor, family 6, subfamily A, member 2 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422035, CurrentID=0, Status='live', Symbol='LOC100422035', Aliases='', description='olfactory receptor, family 6, subfamily A, member 2 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283028, CurrentID=0, Status='live', Symbol='LOC283028', Aliases='', description='putative SCAN domain-containing protein SCAND2P', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42752908, end_position_on_the_genomic_accession=42755383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422037, CurrentID=0, Status='live', Symbol='LOC100422037', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422038, CurrentID=0, Status='live', Symbol='LOC100422038', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  134553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134553, CurrentID=0, Status='live', Symbol='C5orf24', Aliases='', description='chromosome 5 open reading frame 24', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134845680, end_position_on_the_genomic_accession=134859735, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422040, CurrentID=0, Status='live', Symbol='LOC100422040', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283033, CurrentID=0, Status='live', Symbol='LINC00841', Aliases='', description='long intergenic non-protein coding RNA 841', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43909304, end_position_on_the_genomic_accession=43969907, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  151963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151963, CurrentID=0, Status='live', Symbol='MB21D2', Aliases='C3orf59', description='Mab-21 domain containing 2', other_designations='mab-21 domain-containing protein 2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192796815, end_position_on_the_genomic_accession=192918161, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422044, CurrentID=0, Status='live', Symbol='LOC100422044', Aliases='', description='actin related protein 2/3 complex subunit 1A pseudogene', other_designations='actin related protein 2/3 complex, subunit 1A, 41kDa pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121356991, end_position_on_the_genomic_accession=121357944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422045, CurrentID=0, Status='live', Symbol='LOC100422045', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283038, CurrentID=0, Status='live', Symbol='LOC283038', Aliases='', description='uncharacterized LOC283038', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125683243, end_position_on_the_genomic_accession=125709677, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  353008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353008, CurrentID=0, Status='live', Symbol='3.8-1.3', Aliases='', description='HLA complex group 26 (non-protein coding) pseudogene', other_designations='3.8-1.3 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29910235, end_position_on_the_genomic_accession=29911400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283045, CurrentID=0, Status='live', Symbol='LOC283045', Aliases='', description='uncharacterized LOC283045', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62339588, end_position_on_the_genomic_accession=62375127, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271185, CurrentID=0, Status='live', Symbol='RPL21P78', Aliases='RPL21_30_950', description='ribosomal protein L21 pseudogene 78', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133002577, end_position_on_the_genomic_accession=133003115, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90865, CurrentID=0, Status='live', Symbol='IL33', Aliases='C9orf26, DVS27, IL1F11, NF-HEV, NFEHEV', description='interleukin 33', other_designations='DVS27-related protein|interleukin-1 family member 11|nuclear factor for high endothelial venules|nuclear factor from high endothelial venules', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6215149, end_position_on_the_genomic_accession=6257983, orientation='plus', exon_count=11, OMIM=608678, no_hdr0=''),
  100422056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422056, CurrentID=0, Status='live', Symbol='LOC100422056', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101059996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059996, CurrentID=0, Status='live', Symbol='LOC101059996', Aliases='', description='circumsporozoite protein-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283050, CurrentID=0, Status='live', Symbol='ZMIZ1-AS1', Aliases='', description='ZMIZ1 antisense RNA 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78943326, end_position_on_the_genomic_accession=79067448, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100422059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422059, CurrentID=0, Status='live', Symbol='ABCF2P2', Aliases='', description='ATP binding cassette subfamily F member 2 pseudogene 2', other_designations='ATP-binding cassette, sub-family F (GCN20), member 2 pseudogene 2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72103172, end_position_on_the_genomic_accession=72104209, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353010, CurrentID=0, Status='live', Symbol='3.8-1.5', Aliases='', description='HLA complex group 26 (non-protein coding) pseudogene', other_designations='3.8-1.5 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29765116, end_position_on_the_genomic_accession=29766296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422062, CurrentID=0, Status='live', Symbol='LOC100422062', Aliases='', description='olfactory receptor, family 11, subfamily A, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101059997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059997, CurrentID=0, Status='live', Symbol='LOC101059997', Aliases='', description='alpha/beta hydrolase domain-containing protein 17A-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414130, CurrentID=0, Status='live', Symbol='NANOGP7', Aliases='NANOGP3', description='Nanog homeobox pseudogene 7', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91604182, end_position_on_the_genomic_accession=91605522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414131, CurrentID=0, Status='live', Symbol='NANOGP2', Aliases='NANOGP4', description='Nanog homeobox pseudogene 2', other_designations='', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222452042, end_position_on_the_genomic_accession=222453486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414132, CurrentID=0, Status='live', Symbol='NANOGP4', Aliases='NANOGP2', description='Nanog homeobox pseudogene 4', other_designations='', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29212113, end_position_on_the_genomic_accession=29213887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414133, CurrentID=0, Status='live', Symbol='NANOGP5', Aliases='', description='Nanog homeobox pseudogene 5', other_designations='', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100174999, end_position_on_the_genomic_accession=100176996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414134, CurrentID=0, Status='live', Symbol='NANOGP6', Aliases='', description='Nanog homeobox pseudogene 6', other_designations='NANOG pseudogene 6', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99788474, end_position_on_the_genomic_accession=99789414, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414135, CurrentID=0, Status='live', Symbol='NANOGP11', Aliases='', description='Nanog homeobox pseudogene 11', other_designations='', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152546597, end_position_on_the_genomic_accession=152546965, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422076, CurrentID=0, Status='live', Symbol='LOC100422076', Aliases='', description='phosphatidylinositol transfer protein, cytoplasmic 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199908909, end_position_on_the_genomic_accession=199909649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060000, CurrentID=0, Status='live', Symbol='LOC101060000', Aliases='', description='dnaJ homolog subfamily B member 6-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108572645, end_position_on_the_genomic_accession=108617457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152002, CurrentID=0, Status='live', Symbol='XXYLT1', Aliases='C3orf21', description='xyloside xylosyltransferase 1', other_designations='UDP-xylose:alpha-xyloside alpha-1,3-xylosyltransferase', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195068276, end_position_on_the_genomic_accession=195271166, orientation='minus', exon_count=16, OMIM=614552, no_hdr0=''),
  283075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283075, CurrentID=0, Status='live', Symbol='LOC283075', Aliases='', description='uncharacterized LOC283075', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271186, CurrentID=0, Status='live', Symbol='RPL21P83', Aliases='RPL21_31_978', description='ribosomal protein L21 pseudogene 83', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37490389, end_position_on_the_genomic_accession=37490933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414149, CurrentID=0, Status='live', Symbol='ACBD7', Aliases='bA455B2.2', description='acyl-CoA binding domain containing 7', other_designations='acyl-Coenzyme A binding domain containing 7', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15075475, end_position_on_the_genomic_accession=15088776, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  152006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152006, CurrentID=0, Status='live', Symbol='RNF38', Aliases='', description='ring finger protein 38', other_designations='', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36336396, end_position_on_the_genomic_accession=36487182, orientation='minus', exon_count=17, OMIM=612488, no_hdr0=''),
  152007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152007, CurrentID=0, Status='live', Symbol='GLIPR2', Aliases='C9orf19, GAPR-1, GAPR1', description='GLI pathogenesis-related 2', other_designations='17kD fetal brain protein|gliPR 2|glioma pathogenesis-related protein 2|golgi-associated PR-1 protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36136536, end_position_on_the_genomic_accession=36163913, orientation='plus', exon_count=9, OMIM=607141, no_hdr0=''),
  283080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283080, CurrentID=0, Status='live', Symbol='C10orf126', Aliases='bA492M23.1', description='chromosome 10 open reading frame 126', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28846408, end_position_on_the_genomic_accession=28881898, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  414153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414153, CurrentID=0, Status='live', Symbol='SNRPEP2', Aliases='SNRPEL1, bA390F4.4', description='small nuclear ribonucleoprotein polypeptide E pseudogene 2', other_designations='small nuclear ribonucleoprotein polypeptide E-like 1', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6748553, end_position_on_the_genomic_accession=6749014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422091, CurrentID=0, Status='live', Symbol='LOC100422091', Aliases='', description='solute carrier family 30 (zinc transporter), member 6 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111952171, end_position_on_the_genomic_accession=111952671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414157, CurrentID=0, Status='live', Symbol='C10orf62', Aliases='bA548K23.1', description='chromosome 10 open reading frame 62', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97589693, end_position_on_the_genomic_accession=97590934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422094, CurrentID=0, Status='live', Symbol='LOC100422094', Aliases='', description='potassium channel tetramerization domain containing 5 pseudogene', other_designations='potassium channel tetramerisation domain containing 5 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241478311, end_position_on_the_genomic_accession=241478990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152015, CurrentID=0, Status='live', Symbol='ROPN1B', Aliases='', description='rhophilin associated tail protein 1B', other_designations='AKAP-binding sperm protein ropporin|rhophilin-associated protein 1B|ropporin, rhophilin associated protein 1B', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125969144, end_position_on_the_genomic_accession=125983454, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  283089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283089, CurrentID=0, Status='live', Symbol='WDR11-AS1', Aliases='', description='WDR11 antisense RNA 1', other_designations='WDR11 antisense RNA 1 (non-protein coding)', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120761812, end_position_on_the_genomic_accession=120851179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422098, CurrentID=0, Status='live', Symbol='LOC100422098', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422099, CurrentID=0, Status='live', Symbol='LOC100422099', Aliases='', description='olfactory receptor, family 5, subfamily B, member 21 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283092, CurrentID=0, Status='live', Symbol='OR4C13', Aliases='', description='olfactory receptor family 4 subfamily C member 13', other_designations='olfactory receptor OR11-260|olfactory receptor, family 4, subfamily C, member 13', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49952391, end_position_on_the_genomic_accession=49953419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283093, CurrentID=0, Status='live', Symbol='OR4C12', Aliases='OR11-259', description='olfactory receptor family 4 subfamily C member 12', other_designations='olfactory receptor OR11-259|olfactory receptor, family 4, subfamily C, member 12|seven transmembrane helix receptor', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49981473, end_position_on_the_genomic_accession=49982535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422102, CurrentID=0, Status='live', Symbol='LOC100422102', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  152024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152024, CurrentID=0, Status='live', Symbol='LINC00691', Aliases='', description='long intergenic non-protein coding RNA 691', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24099974, end_position_on_the_genomic_accession=24103247, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101060004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060004, CurrentID=0, Status='live', Symbol='LINC01523', Aliases='', description='long intergenic non-protein coding RNA 1523', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47983226, end_position_on_the_genomic_accession=47984313, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422106, CurrentID=0, Status='live', Symbol='LOC100422106', Aliases='', description='staufen double-stranded RNA binding protein 1 pseudogene', other_designations='staufen, RNA binding protein, homolog 1 pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15515343, end_position_on_the_genomic_accession=15516837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422108, CurrentID=0, Status='live', Symbol='LOC100422108', Aliases='', description='olfactory receptor, family 4, subfamily F, member 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  221946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221946, CurrentID=0, Status='live', Symbol='LOC221946', Aliases='', description='uncharacterized LOC221946', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5513854, end_position_on_the_genomic_accession=5525545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  283102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283102, CurrentID=0, Status='live', Symbol='KRT8P41', Aliases='', description='keratin 8 pseudogene 41', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9094363, end_position_on_the_genomic_accession=9096190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060005, CurrentID=0, Status='live', Symbol='SOD1P3', Aliases='', description='superoxide dismutase 1, soluble pseudogene 3', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=126964098, end_position_on_the_genomic_accession=126964724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283104, CurrentID=0, Status='live', Symbol='SBF2-AS1', Aliases='', description='SBF2 antisense RNA 1', other_designations='SBF2 antisense RNA 1 (non-protein coding)', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9758293, end_position_on_the_genomic_accession=9811319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283106, CurrentID=0, Status='live', Symbol='CSNK2A3', Aliases='CSNK2A1P', description='casein kinase 2, alpha 3 polypeptide', other_designations='casein kinase 2, alpha 1 polypeptide pseudogene', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11351942, end_position_on_the_genomic_accession=11353357, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422115, CurrentID=0, Status='live', Symbol='LOC100422115', Aliases='', description='olfactory receptor, family 52, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422116, CurrentID=0, Status='live', Symbol='LOC100422116', Aliases='', description='olfactory receptor, family 52, subfamily M, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060006, CurrentID=0, Status='live', Symbol='LOC101060006', Aliases='', description='NANOG neighbor homeobox-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132972158, end_position_on_the_genomic_accession=132982657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283110, CurrentID=0, Status='live', Symbol='OR52Z1', Aliases='OR52Z1P', description='olfactory receptor family 52 subfamily Z member 1 (gene/pseudogene)', other_designations='Olfactory receptor 52Z1|olfactory receptor, family 52, subfamily Z, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5177721, end_position_on_the_genomic_accession=5178611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283111, CurrentID=0, Status='live', Symbol='OR51V1', Aliases='OR11-36, OR51A12', description='olfactory receptor family 51 subfamily V member 1', other_designations="odorant receptor HOR3'beta1|olfactory receptor 51A12|olfactory receptor OR11-36|olfactory receptor, family 51, subfamily V, member 1", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5199735, end_position_on_the_genomic_accession=5200700, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283112, CurrentID=0, Status='live', Symbol='LOC283112', Aliases='', description='uncharacterized LOC283112', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283114, CurrentID=0, Status='live', Symbol='RPS25P1', Aliases='RPS25_2_1129', description='ribosomal protein S25 pseudogene 1', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27581619, end_position_on_the_genomic_accession=27582100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422123, CurrentID=0, Status='live', Symbol='LOC100422123', Aliases='', description='olfactory receptor, family 51, subfamily I, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283116, CurrentID=0, Status='live', Symbol='TRIM49B', Aliases='', description='tripartite motif containing 49B', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49031600, end_position_on_the_genomic_accession=49037977, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  414189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414189, CurrentID=0, Status='live', Symbol='AGAP6', Aliases='AGAP-6, CTGLF3, bA324H6.1', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 6', other_designations='centaurin, gamma-like family, member 3|centaurin-gamma-like family member 3', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49973060, end_position_on_the_genomic_accession=50010499, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100422126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422126, CurrentID=0, Status='live', Symbol='LOC100422126', Aliases='', description='olfactory receptor, family 51, subfamily I, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422127, CurrentID=0, Status='live', Symbol='LOC100422127', Aliases='', description='olfactory receptor, family 52, subfamily H, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283120, CurrentID=0, Status='live', Symbol='H19', Aliases='ASM, ASM1, BWS, D11S813E, LINC00008, NCRNA00008, WT2', description='H19, imprinted maternally expressed transcript (non-protein coding)', other_designations='H19, imprinted maternally expressed untranslated mRNA|long intergenic non-protein coding RNA 8', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1995176, end_position_on_the_genomic_accession=1997835, orientation='minus', exon_count=5, OMIM=103280, no_hdr0=''),
  100422129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422129, CurrentID=0, Status='live', Symbol='LOC100422129', Aliases='', description='olfactory receptor, family 56, subfamily A, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401561, CurrentID=0, Status='live', Symbol='LINC01451', Aliases='HCCAT4', description='long intergenic non-protein coding RNA 1451', other_designations='FP7915 protein|hepatocellular carcinoma associated transcript 4 (non-protein coding)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136612028, end_position_on_the_genomic_accession=136617065, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422131, CurrentID=0, Status='live', Symbol='LOC100422131', Aliases='', description='olfactory receptor, family 52, subfamily N, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414196, CurrentID=0, Status='live', Symbol='CELF2-AS1', Aliases='C10orf31', description='CELF2 antisense RNA 1', other_designations='CELF2 antisense RNA 1 (non-protein coding)', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11316833, end_position_on_the_genomic_accession=11344674, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  414197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414197, CurrentID=0, Status='live', Symbol='ZNF32-AS1', Aliases='ZNF32OS1, bA402L1.8', description='ZNF32 antisense RNA 1', other_designations='ZNF32 antisense RNA 1 (non-protein coding)|zinc finger protein 32 opposite strand 1', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43643872, end_position_on_the_genomic_accession=43645047, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422135, CurrentID=0, Status='live', Symbol='LOC100422135', Aliases='', description='olfactory receptor, family 4, subfamily C, member 46 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414200, CurrentID=0, Status='live', Symbol='NHP2P1', Aliases='NOLA2P1, bA360G10.4', description='NHP2 ribonucleoprotein pseudogene 1', other_designations='NHP2 ribonucleoprotein homolog pseudogene 1|nucleolar protein family A, member 2 pseudogene 1', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92216292, end_position_on_the_genomic_accession=92217055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112724, CurrentID=0, Status='live', Symbol='RDH13', Aliases='SDR7C3', description='retinol dehydrogenase 13 (all-trans/9-cis)', other_designations='retinol dehydrogenase 13 (all-trans and 9-cis)|short chain dehydrogenase/reductase family 7C member 3|short chain dehydrogenase/reductase family 7C, member 3', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55041319, end_position_on_the_genomic_accession=55069774, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  283130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283130, CurrentID=0, Status='live', Symbol='SLC25A45', Aliases='', description='solute carrier family 25 member 45', other_designations='solute carrier family 25, member 45', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65373053, end_position_on_the_genomic_accession=65383701, orientation='minus', exon_count=14, OMIM=610825, no_hdr0=''),
  283131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283131, CurrentID=0, Status='live', Symbol='NEAT1', Aliases='LINC00084, NCRNA00084, TncRNA, VINC', description='nuclear paraspeckle assembly transcript 1 (non-protein coding)', other_designations='MENepsilon/beta|long intergenic non-protein coding RNA 84|nuclear enriched abundant transcript 1|trophoblast MHC class II suppressor|trophoblast-derived noncoding RNA|virus inducible non-coding RNA', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65422798, end_position_on_the_genomic_accession=65426532, orientation='plus', exon_count=1, OMIM=612769, no_hdr0=''),
  100422140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422140, CurrentID=0, Status='live', Symbol='LOC100422140', Aliases='', description='olfactory receptor, family 2, subfamily A, member 42 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422141, CurrentID=0, Status='live', Symbol='LOC100422141', Aliases='', description='olfactory receptor, family 10, subfamily Q, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422142, CurrentID=0, Status='live', Symbol='LOC100422142', Aliases='', description='olfactory receptor, family 5, subfamily B, member 12 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271188, CurrentID=0, Status='live', Symbol='RPL21P79', Aliases='RPL21_33_903', description='ribosomal protein L21 pseudogene 79', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53533285, end_position_on_the_genomic_accession=53533738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152065, CurrentID=0, Status='live', Symbol='C3orf22', Aliases='', description='chromosome 3 open reading frame 22', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126521997, end_position_on_the_genomic_accession=126558965, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  414212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414212, CurrentID=0, Status='live', Symbol='GLUD1P7', Aliases='GLUD1P2, GLUDP2, GLUDP7, bA342C24.3', description='glutamate dehydrogenase 1 pseudogene 7', other_designations='glutamate dehydrogenase 1 pseudogene 2|glutamate dehydrogenase pseudogene 2|glutamate dehydrogenase pseudogene 7', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46759023, end_position_on_the_genomic_accession=46786770, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  414213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414213, CurrentID=0, Status='live', Symbol='GLUD1P8', Aliases='GLUD1P6, GLUDP6, GLUDP8, bA144G6.5, bK2173J9.1', description='glutamate dehydrogenase 1 pseudogene 8', other_designations='glutamate dehydrogenase 1 pseudogene 6|glutamate dehydrogenase pseudogene 6', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47563606, end_position_on_the_genomic_accession=47572216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422150, CurrentID=0, Status='live', Symbol='LOC100422150', Aliases='', description='olfactory receptor, family 52, subfamily K, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283143, CurrentID=0, Status='live', Symbol='LINC00900', Aliases='', description='long intergenic non-protein coding RNA 900', other_designations='CTC-774J1.2', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115755331, end_position_on_the_genomic_accession=115760200, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422152, CurrentID=0, Status='live', Symbol='LOC100422152', Aliases='', description='olfactory receptor, family 51, subfamily L, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414217, CurrentID=0, Status='live', Symbol='BMS1P3', Aliases='BMS1LP3', description='BMS1, ribosome biogenesis factor pseudogene 3', other_designations='BMS1 pseudogene 3|BMS1 ribosome biogenesis factor pseudogene 3|BMS1L pseudogene 3', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86992253, end_position_on_the_genomic_accession=87000435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283149, CurrentID=0, Status='live', Symbol='BCL9L', Aliases='BCL9-2, DLNB11', description='B-cell CLL/lymphoma 9-like', other_designations='B-cell lymphoma 9-like protein|BCL9-like protein|nuclear co-factor of beta-catenin signalling|protein BCL9-2', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118896140, end_position_on_the_genomic_accession=118925618, orientation='minus', exon_count=11, OMIM=609004, no_hdr0=''),
  283150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283150, CurrentID=0, Status='live', Symbol='FOXR1', Aliases='DLNB13, FOXN5', description='forkhead box R1', other_designations='forkhead box R1 variant 1|forkhead box R1 variant 2|forkhead box protein N5', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118971543, end_position_on_the_genomic_accession=118981285, orientation='plus', exon_count=7, OMIM=615755, no_hdr0=''),
  100422159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422159, CurrentID=0, Status='live', Symbol='LOC100422159', Aliases='', description='olfactory receptor, family 52, subfamily N, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283152, CurrentID=0, Status='live', Symbol='CCDC153', Aliases='', description='coiled-coil domain containing 153', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119190226, end_position_on_the_genomic_accession=119206286, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100422161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422161, CurrentID=0, Status='live', Symbol='LOC100422161', Aliases='', description='olfactory receptor, family 52, subfamily N, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422162, CurrentID=0, Status='live', Symbol='LOC100422162', Aliases='', description='olfactory receptor, family 52, subfamily E, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422163, CurrentID=0, Status='live', Symbol='LOC100422163', Aliases='', description='olfactory receptor, family 52, subfamily E, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414228, CurrentID=0, Status='live', Symbol='RPS27P1', Aliases='RPS27_10_1079, bA108M11.3', description='ribosomal protein S27 pseudogene 1', other_designations='', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91544858, end_position_on_the_genomic_accession=91545088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422165, CurrentID=0, Status='live', Symbol='LOC100422165', Aliases='', description='olfactory receptor, family 4, subfamily A, member 15 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414230, CurrentID=0, Status='live', Symbol='RPS15AP5', Aliases='RPS15A_9_1106', description='ribosomal protein S15a pseudogene 5', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121713577, end_position_on_the_genomic_accession=121714037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283159, CurrentID=0, Status='live', Symbol='OR8D1', Aliases='JCG9, OR8D3, OST004, PDJ9J14', description='olfactory receptor family 8 subfamily D member 1', other_designations='olfactory receptor 8D3|olfactory receptor OR11-301|olfactory receptor, family 8, subfamily D, member 1|olfactory receptor, family 8, subfamily D, member 3|olfactory receptor-like protein JCG9|olfactory-like receptor JCG9', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124309840, end_position_on_the_genomic_accession=124310766, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283160, CurrentID=0, Status='live', Symbol='OR8D2', Aliases='JCG2', description='olfactory receptor family 8 subfamily D member 2 (gene/pseudogene)', other_designations='olfactory receptor OR11-303|olfactory receptor, family 8, subfamily D, member 2|olfactory receptor-like protein JCG2', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124319262, end_position_on_the_genomic_accession=124320197, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283162, CurrentID=0, Status='live', Symbol='OR8B4', Aliases='OR11-315P, OR8B4', description='olfactory receptor family 8 subfamily B member 4 (gene/pseudogene)', other_designations='olfactory receptor OR11-315|olfactory receptor, family 8, subfamily B, member 4 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124423942, end_position_on_the_genomic_accession=124424871, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  414235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414235, CurrentID=0, Status='live', Symbol='PRR26', Aliases='C10orf108', description='proline rich 26', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=649948, end_position_on_the_genomic_accession=665169, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  414236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414236, CurrentID=0, Status='live', Symbol='C10orf55', Aliases='bA417O11.3', description='chromosome 10 open reading frame 55', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73909969, end_position_on_the_genomic_accession=73922777, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  283165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283165, CurrentID=0, Status='live', Symbol='KIRREL3-AS3', Aliases='NCRNA00288, PRR10', description='KIRREL3 antisense RNA 3', other_designations='KIRREL3 antisense RNA 3 (non-protein coding)|proline rich 10', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=127002910, end_position_on_the_genomic_accession=127006058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271189, CurrentID=0, Status='live', Symbol='RPL21P85', Aliases='RPL21_33_996', description='ribosomal protein L21 pseudogene 85', other_designations='', map_location='9q22.1-q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88484852, end_position_on_the_genomic_accession=88485321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422176, CurrentID=0, Status='live', Symbol='LOC100422176', Aliases='', description='olfactory receptor, family 4, subfamily D, member 11 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414241, CurrentID=0, Status='live', Symbol='FAM35BP', Aliases='FAM35B, FAM35EP, bA38L15.1', description='family with sequence similarity 35 member B, pseudogene', other_designations='family with sequence similarity 35, member B, pseudogene|family with sequence similarity 35, member E, pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46610472, end_position_on_the_genomic_accession=46651976, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  152098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152098, CurrentID=0, Status='live', Symbol='ZCWPW2', Aliases='ZCW2', description='zinc finger, CW type with PWWP domain 2', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28349149, end_position_on_the_genomic_accession=28525142, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  283171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283171, CurrentID=0, Status='live', Symbol='C11orf44', Aliases='', description='chromosome 11 open reading frame 44', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130672956, end_position_on_the_genomic_accession=130717352, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  152100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152100, CurrentID=0, Status='live', Symbol='CMC1', Aliases='C3orf68', description='C-x(9)-C motif containing 1', other_designations='COX assembly mitochondrial protein 1 homolog|CX9C mitochondrial protein required for full expression of COX 1|cmc1p|mitochondrial metallochaperone-like protein', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28241589, end_position_on_the_genomic_accession=28319776, orientation='plus', exon_count=11, OMIM=615166, no_hdr0=''),
  152101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152101, CurrentID=0, Status='live', Symbol='VENTXP4', Aliases='VENTX2P4', description='VENT homeobox pseudogene 4', other_designations='VENT homeobox homolog pseudogene|VENT-like homeobox 2 pseudogene 4', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26346263, end_position_on_the_genomic_accession=26347725, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283174, CurrentID=0, Status='live', Symbol='MIR4697HG', Aliases='LINC00947', description='MIR4697 host gene', other_designations='MIR4697 host gene (non-protein coding)|long intergenic non-protein coding RNA 947', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133896435, end_position_on_the_genomic_accession=133901740, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101060017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060017, CurrentID=0, Status='live', Symbol='LOC101060017', Aliases='', description='50 kDa spicule matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90359288, end_position_on_the_genomic_accession=90362907, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  283177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283177, CurrentID=0, Status='live', Symbol='LOC283177', Aliases='', description='uncharacterized LOC283177', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134436482, end_position_on_the_genomic_accession=134505661, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100422186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422186, CurrentID=0, Status='live', Symbol='LOC100422186', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422188, CurrentID=0, Status='live', Symbol='LOC100422188', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B11 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68300322, end_position_on_the_genomic_accession=68302459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422189, CurrentID=0, Status='live', Symbol='LOC100422189', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B17 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68813995, end_position_on_the_genomic_accession=68814727, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152110, CurrentID=0, Status='live', Symbol='NEK10', Aliases='', description='NIMA-related kinase 10', other_designations='NIMA (never in mitosis gene a)- related kinase 10|never in mitosis A-related kinase 10|nimA-related protein kinase 10', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27107336, end_position_on_the_genomic_accession=27369460, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  100422194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422194, CurrentID=0, Status='live', Symbol='LOC100422194', Aliases='', description='ATP binding cassette subfamily G member 2 (Junior blood group) pseudogene', other_designations='ATP-binding cassette, sub-family G (WHITE), member 2 (Junior blood group) pseudogene|ATP-binding cassette, sub-family G (WHITE), member 2 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69022209, end_position_on_the_genomic_accession=69023679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060019, CurrentID=0, Status='live', Symbol='LOC101060019', Aliases='', description='uncharacterized LOC101060019', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66904436, end_position_on_the_genomic_accession=66971466, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  414260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414260, CurrentID=0, Status='live', Symbol='LINC00619', Aliases='C10orf136, bA168P8.1', description='long intergenic non-protein coding RNA 619', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43845306, end_position_on_the_genomic_accession=43850622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283189, CurrentID=0, Status='live', Symbol='OR9G4', Aliases='OR11-216', description='olfactory receptor family 9 subfamily G member 4', other_designations='olfactory receptor OR11-216|olfactory receptor, family 9, subfamily G, member 4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56742828, end_position_on_the_genomic_accession=56743811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  152118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152118, CurrentID=0, Status='live', Symbol='C3orf79', Aliases='', description='chromosome 3 open reading frame 79', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153484495, end_position_on_the_genomic_accession=153502697, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422200, CurrentID=0, Status='live', Symbol='EIF3EP4', Aliases='', description='eukaryotic translation initiation factor 3 subunit E pseudogene 4', other_designations='eukaryotic translation initiation factor 3, subunit E pseudogene 4', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182843573, end_position_on_the_genomic_accession=182844620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283193, CurrentID=0, Status='live', Symbol='OR5AZ1P', Aliases='', description='olfactory receptor family 5 subfamily AZ member 1 pseudogene', other_designations='olfactory receptor OR11-223|olfactory receptor, family 5, subfamily AZ, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57917302, end_position_on_the_genomic_accession=57918230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283194, CurrentID=0, Status='live', Symbol='LOC283194', Aliases='', description='uncharacterized LOC283194', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58933643, end_position_on_the_genomic_accession=59058452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100422204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422204, CurrentID=0, Status='live', Symbol='LOC100422204', Aliases='', description='regulatory solute carrier protein, family 1, member 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12958407, end_position_on_the_genomic_accession=12962206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283197, CurrentID=0, Status='live', Symbol='LINC00301', Aliases='C11orf64, NCRNA00301', description='long intergenic non-protein coding RNA 301', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60615751, end_position_on_the_genomic_accession=60687149, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422209, CurrentID=0, Status='live', Symbol='LOC100422209', Aliases='', description='phenazine biosynthesis-like protein domain containing pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26345335, end_position_on_the_genomic_accession=26346111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270850, CurrentID=0, Status='live', Symbol='RPL7P15', Aliases='RPL7_4_398', description='ribosomal protein L7 pseudogene 15', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124151895, end_position_on_the_genomic_accession=124152562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422212, CurrentID=0, Status='live', Symbol='LOC100422212', Aliases='', description='eukaryotic translation initiation factor 3 subunit J pseudogene', other_designations='eukaryotic translation initiation factor 3, subunit J pseudogene', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163421070, end_position_on_the_genomic_accession=163423191, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422213, CurrentID=0, Status='live', Symbol='LOC100422213', Aliases='', description='guanine nucleotide binding protein-like 3 (nucleolar) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75521365, end_position_on_the_genomic_accession=75522201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283208, CurrentID=0, Status='live', Symbol='P4HA3', Aliases='', description='prolyl 4-hydroxylase, alpha polypeptide III', other_designations='4-PH alpha-3|C-P4H alpha III|collagen prolyl 4-hydroxylase alpha(III)|procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha polypeptide III', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74247863, end_position_on_the_genomic_accession=74311676, orientation='minus', exon_count=15, OMIM=608987, no_hdr0=''),
  283209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283209, CurrentID=0, Status='live', Symbol='PGM2L1', Aliases='BM32A, PMMLP', description='phosphoglucomutase 2-like 1', other_designations='phosphoglucomutase-2-like 1', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74330312, end_position_on_the_genomic_accession=74398465, orientation='minus', exon_count=15, OMIM=611610, no_hdr0=''),
  152138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152138, CurrentID=0, Status='live', Symbol='PYDC2', Aliases='POP2, cPOP2', description='pyrin domain containing 2', other_designations='cellular POP2|pyrin-only protein 2', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191461163, end_position_on_the_genomic_accession=191461456, orientation='plus', exon_count=1, OMIM=615701, no_hdr0=''),
  283212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283212, CurrentID=0, Status='live', Symbol='KLHL35', Aliases='', description='kelch like family member 35', other_designations='kelch-like 35|kelch-like family member 35', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75422394, end_position_on_the_genomic_accession=75433444, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  283214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283214, CurrentID=0, Status='live', Symbol='LOC283214', Aliases='', description='uncharacterized LOC283214', other_designations='CTD-2530H12.1', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75758455, end_position_on_the_genomic_accession=75768647, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101060024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060024, CurrentID=0, Status='live', Symbol='LOC101060024', Aliases='', description='uncharacterized protein C7orf73-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134081431, end_position_on_the_genomic_accession=134082822, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283218, CurrentID=0, Status='live', Symbol='DKFZp434E1119', Aliases='', description='uncharacterized DKFZp434E1119', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=77183485, end_position_on_the_genomic_accession=77184868, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  283219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283219, CurrentID=0, Status='live', Symbol='KCTD21', Aliases='KCASH2', description='potassium channel tetramerization domain containing 21', other_designations='potassium channel tetramerisation domain containing 21', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78171243, end_position_on_the_genomic_accession=78188817, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105373315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373315, CurrentID=0, Status='live', Symbol='LOC105373315', Aliases='', description='uncharacterized LOC105373315', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115694842, end_position_on_the_genomic_accession=115713201, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422229, CurrentID=0, Status='live', Symbol='MAPRE1P3', Aliases='', description='MAPRE1 pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165285253, end_position_on_the_genomic_accession=165286045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422230, CurrentID=0, Status='live', Symbol='KCTD10P1', Aliases='', description='potassium channel tetramerization domain containing 10 pseudogene 1', other_designations='potassium channel tetramerisation domain containing 10 pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28598664, end_position_on_the_genomic_accession=28599280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422232, CurrentID=0, Status='live', Symbol='LOC100422232', Aliases='', description='parkinson protein 7 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95918075, end_position_on_the_genomic_accession=95918873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422233, CurrentID=0, Status='live', Symbol='LOC100422233', Aliases='', description='iron-sulfur cluster assembly 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136809279, end_position_on_the_genomic_accession=136809874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283226, CurrentID=0, Status='live', Symbol='FNTAP1', Aliases='FNTAL1', description='farnesyltransferase, CAAX box, alpha pseudogene 1', other_designations='farnesyltransferase, CAAX box, alpha-like 1', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86195131, end_position_on_the_genomic_accession=86196778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422235, CurrentID=0, Status='live', Symbol='CYP4F36P', Aliases='', description='cytochrome P450 family 4 subfamily F member 36, pseudogene', other_designations='CYP4F2-de12b|cytochrome P450, family 4, subfamily F, polypeptide 3 pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 36, pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15871667, end_position_on_the_genomic_accession=15871852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414300, CurrentID=0, Status='live', Symbol='LOC414300', Aliases='', description='uncharacterized LOC414300', other_designations='', map_location='12q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105102472, end_position_on_the_genomic_accession=105107642, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283229, CurrentID=0, Status='live', Symbol='CRACR2B', Aliases='EFCAB4A', description='calcium release activated channel regulator 2B', other_designations='CRAC channel regulator 2B|CRAC regulator 2B|Ca2+ release-activated Ca2+ (CRAC) channel regulator 2B|EF-hand calcium binding domain 4A|calcium release-activated calcium channel regulator 2B|calcium release-activated channel regulator 2B', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=826144, end_position_on_the_genomic_accession=831991, orientation='plus', exon_count=12, OMIM=614177, no_hdr0=''),
  100422238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422238, CurrentID=0, Status='live', Symbol='CYP4F23P', Aliases='', description='cytochrome P450 family 4 subfamily F member 23, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 23, pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 3 pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15564073, end_position_on_the_genomic_accession=15584721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283232, CurrentID=0, Status='live', Symbol='TMEM80', Aliases='', description='transmembrane protein 80', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=695591, end_position_on_the_genomic_accession=705028, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422241, CurrentID=0, Status='live', Symbol='LOC100422241', Aliases='', description='transforming, acidic coiled-coil containing protein 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36696317, end_position_on_the_genomic_accession=36697090, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283234, CurrentID=0, Status='live', Symbol='CCDC88B', Aliases='BRLZ, CCDC88, HKRP3, gipie', description='coiled-coil domain containing 88B', other_designations='78 kDa glucose-regulated protein [GRP78]-interacting protein induced by ER stress|GRP78-interacting protein induced by ER stress|brain leucine zipper domain-containing protein|brain leucine zipper protein|coiled-coil domain containing 88|hook-related protein 3', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64340197, end_position_on_the_genomic_accession=64357534, orientation='plus', exon_count=29, OMIM=611205, no_hdr0=''),
  100874023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874023, CurrentID=0, Status='live', Symbol='WDFY3-AS1', Aliases='', description='WDFY3 antisense RNA 1', other_designations='WDFY3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283236, CurrentID=0, Status='live', Symbol='EEF1DP8', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 8', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62169219, end_position_on_the_genomic_accession=62169827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283237, CurrentID=0, Status='live', Symbol='TTC9C', Aliases='', description='tetratricopeptide repeat domain 9C', other_designations='TPR repeat protein 9C', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62728078, end_position_on_the_genomic_accession=62738638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283238, CurrentID=0, Status='live', Symbol='SLC22A24', Aliases='NET46', description='solute carrier family 22 member 24', other_designations='solute carrier family 22, member 24', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63079255, end_position_on_the_genomic_accession=63144221, orientation='minus', exon_count=11, OMIM=611698, no_hdr0=''),
  100422247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422247, CurrentID=0, Status='live', Symbol='CYP4F34P', Aliases='', description='cytochrome P450 family 4 subfamily F member 34, pseudogene', other_designations='CYP4F-se7[6:7:8]|cytochrome P450, family 4, subfamily F, polypeptide 3 pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 34, pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18741912, end_position_on_the_genomic_accession=18742701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414312, CurrentID=0, Status='live', Symbol='C9orf118', Aliases='bA342C23.1', description='chromosome 9 open reading frame 118', other_designations='Putative uncharacterized protein C9orf118', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422249, CurrentID=0, Status='live', Symbol='NPM1P41', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 41', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82010671, end_position_on_the_genomic_accession=82011239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422250, CurrentID=0, Status='live', Symbol='SUMO2P11', Aliases='', description='SUMO2 pseudogene 11', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene|small ubiquitin-like modifier 2 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76668169, end_position_on_the_genomic_accession=76668795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414316, CurrentID=0, Status='live', Symbol='COL5A1-AS1', Aliases='C9orf104, bA54A22.4', description='COL5A1 antisense RNA 1', other_designations='COL5A1 antisense RNA 1 (non-protein coding)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134649385, end_position_on_the_genomic_accession=134653029, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422253, CurrentID=0, Status='live', Symbol='LOC100422253', Aliases='', description='olfactory receptor, family 11, subfamily H, member 12 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414318, CurrentID=0, Status='live', Symbol='C9orf106', Aliases='bA65J3.5', description='chromosome 9 open reading frame 106', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129321016, end_position_on_the_genomic_accession=129322603, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  414319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414319, CurrentID=0, Status='live', Symbol='LINC00587', Aliases='C9orf107, bA785H23.1', description='long intergenic non-protein coding RNA 587', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102519637, end_position_on_the_genomic_accession=102657509, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283248, CurrentID=0, Status='live', Symbol='RCOR2', Aliases='', description='REST corepressor 2', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63911221, end_position_on_the_genomic_accession=63916844, orientation='minus', exon_count=12, OMIM=616019, no_hdr0=''),
  100422259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422259, CurrentID=0, Status='live', Symbol='NPM1P29', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 29', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149938793, end_position_on_the_genomic_accession=149939893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414325, CurrentID=0, Status='live', Symbol='DEFB103A', Aliases='BD-3, DEFB-3, DEFB103, DEFB3, HBD3, HBP-3, HBP3, hBD-3', description='defensin beta 103A', other_designations='beta defensin 3|beta defensin-3|beta-defensin 3|defensin, beta 103|defensin, beta 103A|defensin, beta 3|defensin-like protein', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7881204, end_position_on_the_genomic_accession=7882664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  283254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283254, CurrentID=0, Status='live', Symbol='HARBI1', Aliases='C11orf77', description='harbinger transposase derived 1', other_designations='harbinger transposase-derived nuclease', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46603306, end_position_on_the_genomic_accession=46617894, orientation='minus', exon_count=7, OMIM=615086, no_hdr0=''),
  414327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414327, CurrentID=0, Status='live', Symbol='PS1TP4', Aliases='', description='HBV preS1-transactivated protein 4', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414328, CurrentID=0, Status='live', Symbol='IDNK', Aliases='C9orf103, bA522I20.2', description='idnK, gluconokinase homolog (E. coli)', other_designations='glucokinase-like protein|gluconate kinase|gluconokinase-like protein', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83623049, end_position_on_the_genomic_accession=83644130, orientation='plus', exon_count=8, OMIM=611343, no_hdr0=''),
  152185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152185, CurrentID=0, Status='live', Symbol='SPICE1', Aliases='CCDC52, SPICE', description='spindle and centriole associated protein 1', other_designations='coiled-coil domain containing 52|coiled-coil domain-containing protein 52|spindle and centriole protein', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113442718, end_position_on_the_genomic_accession=113515187, orientation='minus', exon_count=19, OMIM=613447, no_hdr0=''),
  100422267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422267, CurrentID=0, Status='live', Symbol='LOC100422267', Aliases='', description='coagulation factor III (thromboplastin, tissue factor) pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50490795, end_position_on_the_genomic_accession=50491372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414332, CurrentID=0, Status='live', Symbol='LCN10', Aliases='', description='lipocalin 10', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136738167, end_position_on_the_genomic_accession=136742959, orientation='minus', exon_count=6, OMIM=612904, no_hdr0=''),
  152189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152189, CurrentID=0, Status='live', Symbol='CMTM8', Aliases='CKLFSF8, CKLFSF8-V2', description='CKLF-like MARVEL transmembrane domain containing 8', other_designations='chemokine-like factor super family 8|chemokine-like factor superfamily 8|chemokine-like factor superfamily member 8', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32238679, end_position_on_the_genomic_accession=32370321, orientation='plus', exon_count=5, OMIM=607891, no_hdr0=''),
  100422270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422270, CurrentID=0, Status='live', Symbol='NPM1P44', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 44', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67185165, end_position_on_the_genomic_accession=67186303, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422271, CurrentID=0, Status='live', Symbol='LOC100422271', Aliases='', description='proteasome 26S subunit, ATPase 2 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84678798, end_position_on_the_genomic_accession=84679842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422272, CurrentID=0, Status='live', Symbol='NPM1P52', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 52', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103016652, end_position_on_the_genomic_accession=103017607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422273, CurrentID=0, Status='live', Symbol='KCTD9P1', Aliases='HsT38306', description='potassium channel tetramerization domain containing 9 pseudogene 1', other_designations='potassium channel tetramerisation domain containing 9 pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20111824, end_position_on_the_genomic_accession=20112987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422274, CurrentID=0, Status='live', Symbol='LOC100422274', Aliases='', description='transforming, acidic coiled-coil containing protein 3 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5240263, end_position_on_the_genomic_accession=5241146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283267, CurrentID=0, Status='live', Symbol='LINC00294', Aliases='NCRNA00294', description='long intergenic non-protein coding RNA 294', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33076150, end_position_on_the_genomic_accession=33079454, orientation='plus', exon_count=1, OMIM=612296, no_hdr0=''),
  283270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283270, CurrentID=0, Status='live', Symbol='LOC283270', Aliases='', description='uncharacterized LOC283270', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724053, CurrentID=0, Status='live', Symbol='LOC102724053', Aliases='', description='uncharacterized LOC102724053', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159383531, end_position_on_the_genomic_accession=159396443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422284, CurrentID=0, Status='live', Symbol='NPM1P49', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 49', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47438736, end_position_on_the_genomic_accession=47439772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422285, CurrentID=0, Status='live', Symbol='LOC100422285', Aliases='', description='cleavage and polyadenylation specific factor 1, 160kDa pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49911589, end_position_on_the_genomic_accession=49912381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152206, CurrentID=0, Status='live', Symbol='CCDC13', Aliases='', description='coiled-coil domain containing 13', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42704118, end_position_on_the_genomic_accession=42773253, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  414351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414351, CurrentID=0, Status='live', Symbol='LINC00229', Aliases='C22orf10, NCRNA00229, dJ474I12.C22.2', description='long intergenic non-protein coding RNA 229', other_designations='', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44606328, end_position_on_the_genomic_accession=44625419, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422288, CurrentID=0, Status='live', Symbol='CYP4A26P', Aliases='', description='cytochrome P450 family 4 subfamily A member 26, pseudogene', other_designations='cytochrome P450, family 4, subfamily A, polypeptide 11 pseudogene|cytochrome P450, family 4, subfamily A, polypeptide 26, pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46967677, end_position_on_the_genomic_accession=47001094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422289, CurrentID=0, Status='live', Symbol='PSMB7P1', Aliases='', description='PSMB7 pseudogene 1', other_designations='proteasome (prosome, macropain) subunit, beta type, 7 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20149232, end_position_on_the_genomic_accession=20149747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414355, CurrentID=0, Status='live', Symbol='VKORC1P1', Aliases='', description='VKORC1 pseudogene 1', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27863832, end_position_on_the_genomic_accession=27864662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283284, CurrentID=0, Status='live', Symbol='IGSF22', Aliases='IGFN2', description='immunoglobulin superfamily member 22', other_designations='immunoglobulin superfamily, member 22', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18704305, end_position_on_the_genomic_accession=18726230, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  414357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414357, CurrentID=0, Status='live', Symbol='VKORC1P2', Aliases='', description='VKORC1 pseudogene 2', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58228405, end_position_on_the_genomic_accession=58229001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422294, CurrentID=0, Status='live', Symbol='LOC100422294', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159462396, end_position_on_the_genomic_accession=159463256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422296, CurrentID=0, Status='live', Symbol='LOC100422296', Aliases='', description='parkinson protein 7 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49595148, end_position_on_the_genomic_accession=49595688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152217, CurrentID=0, Status='live', Symbol='NCBP2-AS2', Aliases='', description='NCBP2 antisense RNA 2 (head to head)', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196942623, end_position_on_the_genomic_accession=196944013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422298, CurrentID=0, Status='live', Symbol='LOC100422298', Aliases='', description='iron-sulfur cluster assembly 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21381223, end_position_on_the_genomic_accession=21381921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422299, CurrentID=0, Status='live', Symbol='C1QBPP2', Aliases='', description='complement component 1, q subcomponent binding protein, pseudogene 2', other_designations='complement component 1, q subcomponent binding protein pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66761403, end_position_on_the_genomic_accession=66762456, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422300, CurrentID=0, Status='live', Symbol='LOC100422300', Aliases='', description='pyridine nucleotide-disulphide oxidoreductase domain 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106826395, end_position_on_the_genomic_accession=106827890, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402240, CurrentID=0, Status='live', Symbol='OR1X5P', Aliases='', description='olfactory receptor family 1 subfamily X member 5 pseudogene', other_designations='olfactory receptor, family 1, subfamily X, member 5 pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177836412, end_position_on_the_genomic_accession=177837663, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283297, CurrentID=0, Status='live', Symbol='OR10A4', Aliases='JCG5P, OR10A4', description='olfactory receptor family 10 subfamily A member 4', other_designations='hP2 olfactory receptor|olfactory receptor OR11-87|olfactory receptor, family 10, subfamily A, member 4 pseudogene|olfactory receptor-like protein JCG5', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6876625, end_position_on_the_genomic_accession=6877619, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283298, CurrentID=0, Status='live', Symbol='OLFML1', Aliases='UNQ564', description='olfactomedin like 1', other_designations='MVAL564|olfactomedin-like 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7485369, end_position_on_the_genomic_accession=7511897, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  283299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283299, CurrentID=0, Status='live', Symbol='LOC283299', Aliases='', description='uncharacterized LOC283299', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7850751, end_position_on_the_genomic_accession=7905955, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100422309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422309, CurrentID=0, Status='live', Symbol='NPM1P43', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 43', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73161776, end_position_on_the_genomic_accession=73162930, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422310, CurrentID=0, Status='live', Symbol='C1QBPP3', Aliases='', description='complement component 1, q subcomponent binding protein, pseudogene 3', other_designations='complement component 1, q subcomponent binding protein pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33693211, end_position_on_the_genomic_accession=33694251, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283303, CurrentID=0, Status='live', Symbol='MRGPRG-AS1', Aliases='C11orf36, HSD-40', description='MRGPRG antisense RNA 1', other_designations='MRGPRG antisense RNA 1 (non-protein coding)', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3218332, end_position_on_the_genomic_accession=3223131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422312, CurrentID=0, Status='live', Symbol='NPM1P24', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 24', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72917318, end_position_on_the_genomic_accession=72918630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422313, CurrentID=0, Status='live', Symbol='NPM1P26', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 26', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96560079, end_position_on_the_genomic_accession=96561122, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131529, CurrentID=0, Status='live', Symbol='LOC100131529', Aliases='', description='family with sequence similarity 60, member A pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98317741, end_position_on_the_genomic_accession=98318349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  3527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3527, CurrentID=0, Status='live', Symbol='IGKV3OR22-2', Aliases='IGKV3OR222, IGKVP3', description='immunoglobulin kappa variable 3/OR22-2 (pseudogene)', other_designations='IGKV3/OR22-2|immunoglobulin kappa variable 3/OR22-2 pseudogene|immunoglobulin kappa variable region pseudogene 3', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16926050, end_position_on_the_genomic_accession=16926444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422316, CurrentID=0, Status='live', Symbol='NPM1P45', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 45', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9852741, end_position_on_the_genomic_accession=9853308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422317, CurrentID=0, Status='live', Symbol='LOC100422317', Aliases='', description='proteasome 26S subunit, ATPase 5 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 5 pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63415272, end_position_on_the_genomic_accession=63416488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283310, CurrentID=0, Status='live', Symbol='OTOGL', Aliases='C12orf64, DFNB84B', description='otogelin-like', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80130502, end_position_on_the_genomic_accession=80379090, orientation='plus', exon_count=65, OMIM=614925, no_hdr0=''),
  100422319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422319, CurrentID=0, Status='live', Symbol='LOC100422319', Aliases='', description='iron-sulfur cluster assembly 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85814425, end_position_on_the_genomic_accession=85815062, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283314, CurrentID=0, Status='live', Symbol='C1RL-AS1', Aliases='MATL2963', description='C1RL antisense RNA 1', other_designations='ABC12-49244600F4.3|C1RL antisense RNA 1 (non-protein coding)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7108308, end_position_on_the_genomic_accession=7121851, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  374899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374899, CurrentID=0, Status='live', Symbol='ZNF829', Aliases='', description='zinc finger protein 829', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36888124, end_position_on_the_genomic_accession=36916387, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  283316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283316, CurrentID=0, Status='live', Symbol='CD163L1', Aliases='CD163B, M160', description='CD163 molecule like 1', other_designations='CD163 antigen B|CD163 antigen-like 1|CD163 molecule-like 1', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7326484, end_position_on_the_genomic_accession=7444185, orientation='minus', exon_count=25, OMIM=606079, no_hdr0=''),
  100422325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422325, CurrentID=0, Status='live', Symbol='LOC100422325', Aliases='', description='poly(A) binding protein, cytoplasmic 4 (inducible form) pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=126511138, end_position_on_the_genomic_accession=126512796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283320, CurrentID=0, Status='live', Symbol='HSPD1P12', Aliases='HSPD1-10P', description='heat shock protein family D (Hsp60) member 1 pseudogene 12', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 12', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8014960, end_position_on_the_genomic_accession=8016711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422330, CurrentID=0, Status='live', Symbol='QRSL1P2', Aliases='', description='glutaminyl-tRNA synthase (glutamine-hydrolyzing)-like 1 pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222261833, end_position_on_the_genomic_accession=222262795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060042, CurrentID=0, Status='live', Symbol='LOC101060042', Aliases='', description='X antigen family member 3-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55172717, end_position_on_the_genomic_accession=55185238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422334, CurrentID=0, Status='live', Symbol='LOC100422334', Aliases='', description='general transcription factor IIH subunit 3 pseudogene', other_designations='general transcription factor IIH, polypeptide 3, 34kDa pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38605265, end_position_on_the_genomic_accession=38606101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422337, CurrentID=0, Status='live', Symbol='LOC100422337', Aliases='', description='guanine nucleotide binding protein (G protein), alpha 13 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79573715, end_position_on_the_genomic_accession=79575678, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422338, CurrentID=0, Status='live', Symbol='LOC100422338', Aliases='', description='GTPase activating protein (SH3 domain) binding protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135510549, end_position_on_the_genomic_accession=135511943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422339, CurrentID=0, Status='live', Symbol='SDR42E1P5', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 1 pseudogene 5', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102411243, end_position_on_the_genomic_accession=102412111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422340, CurrentID=0, Status='live', Symbol='SDR42E1P1', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 1 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39651725, end_position_on_the_genomic_accession=39652477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422342, CurrentID=0, Status='live', Symbol='PIGUP1', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class U pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class U pseudogene 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70916766, end_position_on_the_genomic_accession=70917508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422343, CurrentID=0, Status='live', Symbol='UPF3AP3', Aliases='', description='UPF3A pseudogene 3', other_designations='UPF3 regulator of nonsense transcripts homolog A pseudogene 3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99998304, end_position_on_the_genomic_accession=99999133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283337, CurrentID=0, Status='live', Symbol='ZNF740', Aliases='Zfp740', description='zinc finger protein 740', other_designations='oriLyt TD-element-binding protein 7', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53180751, end_position_on_the_genomic_accession=53190870, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100422346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422346, CurrentID=0, Status='live', Symbol='SDR42E1P2', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 1 pseudogene 2', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42408452, end_position_on_the_genomic_accession=42409004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422347, CurrentID=0, Status='live', Symbol='SDR42E1P3', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 1 pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62690061, end_position_on_the_genomic_accession=62690613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422348, CurrentID=0, Status='live', Symbol='SDR42E1P4', Aliases='', description='short chain dehydrogenase/reductase family 42E, member 1 pseudogene 4', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66724586, end_position_on_the_genomic_accession=66725144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422349, CurrentID=0, Status='live', Symbol='PIGFP2', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class F pseudogene 2', other_designations='phosphatidylinositol glycan anchor biosynthesis, class F pseudogene|phosphatidylinositol glycan anchor biosynthesis, class F, pseudogene 2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123842933, end_position_on_the_genomic_accession=123843880, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422352, CurrentID=0, Status='live', Symbol='LOC100422352', Aliases='', description='transmembrane and tetratricopeptide repeat containing 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30454919, end_position_on_the_genomic_accession=30520453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152273, CurrentID=0, Status='live', Symbol='FGD5', Aliases='ZFYVE23', description='FYVE, RhoGEF and PH domain containing 5', other_designations='zinc finger FYVE domain-containing protein 23', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14818962, end_position_on_the_genomic_accession=14934565, orientation='plus', exon_count=20, OMIM=614788, no_hdr0=''),
  100422354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422354, CurrentID=0, Status='live', Symbol='CHCHD3P2', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 3 pseudogene 2', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=72610678, end_position_on_the_genomic_accession=72611318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283349, CurrentID=0, Status='live', Symbol='RASSF3', Aliases='RASSF5', description='Ras association (RalGDS/AF-6) domain family member 3', other_designations='Ras association (RalGDS/AF-6) domain family 3', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64610513, end_position_on_the_genomic_accession=64697567, orientation='plus', exon_count=8, OMIM=607019, no_hdr0=''),
  283352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283352, CurrentID=0, Status='live', Symbol='LOC283352', Aliases='', description='uncharacterized LOC283352', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129109690, end_position_on_the_genomic_accession=129113298, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377684, CurrentID=0, Status='live', Symbol='LOC105377684', Aliases='', description='uncharacterized LOC105377684', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159515731, end_position_on_the_genomic_accession=159629601, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100422363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422363, CurrentID=0, Status='live', Symbol='BUB1P1', Aliases='', description='BUB1, mitotic checkpoint serine/threonine kinase pseudogene 1', other_designations='BUB1 mitotic checkpoint serine/threonine kinase pseudogene 1|budding uninhibited by benzimidazoles 1 homolog pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127867918, end_position_on_the_genomic_accession=127868622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283357, CurrentID=0, Status='live', Symbol='LOC283357', Aliases='', description='uncharacterized LOC283357', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283358, CurrentID=0, Status='live', Symbol='B4GALNT3', Aliases='', description='beta-1,4-N-acetyl-galactosaminyl transferase 3', other_designations='B4GalNAcT3|Beta4GalNAc-T3|N-acetyl-beta-glucosaminyl-glycoprotein 4-beta-N- acetylgalactosaminyltransferase|N-acetyl-beta-glucosaminyl-glycoprotein 4-beta-N-acetylgalactosaminyltransferase 2|NGalNAc-T2|beta-1,4-N-acetylgalactosaminyltransferase III|beta4GalNAcT3', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=460377, end_position_on_the_genomic_accession=561892, orientation='plus', exon_count=20, OMIM=612220, no_hdr0=''),
  100422368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422368, CurrentID=0, Status='live', Symbol='LOC100422368', Aliases='', description='UPF3 regulator of nonsense transcripts homolog B (yeast) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119815165, end_position_on_the_genomic_accession=119823046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422369, CurrentID=0, Status='live', Symbol='LOC100422369', Aliases='', description='par-6 family cell polarity regulator beta pseudogene', other_designations='par-6 partitioning defective 6 homolog beta pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123878593, end_position_on_the_genomic_accession=123879299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422371, CurrentID=0, Status='live', Symbol='RIMKLBP2', Aliases='', description='ribosomal modification protein rimK-like family member B pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219199852, end_position_on_the_genomic_accession=219200753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422372, CurrentID=0, Status='live', Symbol='NDE1P1', Aliases='', description='nudE neurodevelopment protein 1 pseudogene 1', other_designations='nudE nuclear distribution E homolog 1 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106817651, end_position_on_the_genomic_accession=106818339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283365, CurrentID=0, Status='live', Symbol='OR6C6', Aliases='', description='olfactory receptor family 6 subfamily C member 6', other_designations='olfactory receptor, family 6, subfamily C, member 6', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55294288, end_position_on_the_genomic_accession=55295232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422374, CurrentID=0, Status='live', Symbol='QRSL1P3', Aliases='', description='glutaminyl-tRNA synthase (glutamine-hydrolyzing)-like 1 pseudogene 3', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35765811, end_position_on_the_genomic_accession=35767110, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060049, CurrentID=0, Status='live', Symbol='LOC101060049', Aliases='', description='ornithine aminotransferase, mitochondrial-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48379351, end_position_on_the_genomic_accession=48383153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422376, CurrentID=0, Status='live', Symbol='LOC100422376', Aliases='', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73648891, end_position_on_the_genomic_accession=73649860, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283372, CurrentID=0, Status='live', Symbol='FLJ25613', Aliases='', description='uncharacterized protein FLJ25613', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283373, CurrentID=0, Status='live', Symbol='ANKRD52', Aliases='ANKRD33', description='ankyrin repeat domain 52', other_designations='CVWG5837|PP6-ARS-C|ankyrin repeat domain 33|ankyrin repeat domain-containing protein 52|protein phosphatase 6 ankyrin repeat subunit C|serine/threonine-protein phosphatase 6 regulatory subunit ARS-C', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56237807, end_position_on_the_genomic_accession=56258359, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  152302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152302, CurrentID=0, Status='live', Symbol='CIDECP', Aliases='CICE', description='cell death-inducing DFFA-like effector c pseudogene', other_designations='cell death-inducing CIDE-like effector pseudogene', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10017553, end_position_on_the_genomic_accession=10026136, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283375, CurrentID=0, Status='live', Symbol='SLC39A5', Aliases='LZT-Hs7, MYP24, ZIP5', description='solute carrier family 39 (zinc transporter), member 5', other_designations='solute carrier family 39 (metal ion transporter), member 5|solute carrier family 39 member 5|zrt- and Irt-like protein 5', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56230036, end_position_on_the_genomic_accession=56237846, orientation='plus', exon_count=13, OMIM=608730, no_hdr0=''),
  100271196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271196, CurrentID=0, Status='live', Symbol='RPL31P40', Aliases='RPL31_12_911', description='ribosomal protein L31 pseudogene 40', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68293537, end_position_on_the_genomic_accession=68293816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283377, CurrentID=0, Status='live', Symbol='SPRYD4', Aliases='', description='SPRY domain containing 4', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56468517, end_position_on_the_genomic_accession=56470985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422386, CurrentID=0, Status='live', Symbol='ADAD1P1', Aliases='', description='adenosine deaminase domain containing 1 pseudogene 1', other_designations='adenosine deaminase domain containing 1 (testis-specific) pseudogene 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39286734, end_position_on_the_genomic_accession=39287898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060051, CurrentID=0, Status='live', Symbol='LOC101060051', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71155919, end_position_on_the_genomic_accession=71156759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873500, CurrentID=0, Status='live', Symbol='RNA5SP242', Aliases='RN5S242', description='RNA, 5S ribosomal pseudogene 242', other_designations='RNA, 5S ribosomal 242', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128697439, end_position_on_the_genomic_accession=128697556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283383, CurrentID=0, Status='live', Symbol='ADGRD1', Aliases='GPR133, PGR25', description='adhesion G protein-coupled receptor D1', other_designations='G protein-coupled receptor 133|G-protein coupled receptor 133|G-protein coupled receptor PGR25|probable G-protein coupled receptor 133', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130953907, end_position_on_the_genomic_accession=131141465, orientation='plus', exon_count=30, OMIM=613639, no_hdr0=''),
  283385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283385, CurrentID=0, Status='live', Symbol='MORN3', Aliases='', description='MORN repeat containing 3', other_designations='membrane occupation and recognition nexus repeat containing protein', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121651118, end_position_on_the_genomic_accession=121672639, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100422395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422395, CurrentID=0, Status='live', Symbol='POLR3GP1', Aliases='', description='polymerase (RNA) III (DNA directed) polypeptide G (32kD) pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91602568, end_position_on_the_genomic_accession=91603563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103568124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103568124, CurrentID=0, Status='live', Symbol='LOC103568124', Aliases='', description='coiled-coil domain containing 181 pseudogene', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4602189, end_position_on_the_genomic_accession=4603636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422398, CurrentID=0, Status='live', Symbol='LOC100422398', Aliases='', description='protein kinase, cAMP-dependent, regulatory subunit type II beta pseudogene', other_designations='protein kinase, cAMP-dependent, regulatory, type II, beta pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58967596, end_position_on_the_genomic_accession=58968291, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422399, CurrentID=0, Status='live', Symbol='LOC100422399', Aliases='', description='protein kinase, cAMP-dependent, regulatory subunit type II beta pseudogene', other_designations='protein kinase, cAMP-dependent, regulatory, type II, beta pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59142087, end_position_on_the_genomic_accession=59142630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283392, CurrentID=0, Status='live', Symbol='TRHDE-AS1', Aliases='', description='TRHDE antisense RNA 1', other_designations='TRHDE antisense RNA 1 (non-protein coding)', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=72253507, end_position_on_the_genomic_accession=72273509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422402, CurrentID=0, Status='live', Symbol='LOC100422402', Aliases='', description='RAS guanyl releasing protein 2 (calcium and DAG-regulated) pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68614341, end_position_on_the_genomic_accession=68615007, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283398, CurrentID=0, Status='live', Symbol='SUCLG2P2', Aliases='', description='succinate-CoA ligase, GDP-forming, beta subunit pseudogene 2', other_designations='succinyl-CoA ligase [GDP-forming] subunit beta, mitochondrial pseudogene', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94548221, end_position_on_the_genomic_accession=94550555, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422407, CurrentID=0, Status='live', Symbol='ADAD1P2', Aliases='', description='adenosine deaminase domain containing 1 pseudogene 2', other_designations='adenosine deaminase domain containing 1 (testis-specific) pseudogene 2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50063792, end_position_on_the_genomic_accession=50064486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422408, CurrentID=0, Status='live', Symbol='LOC100422408', Aliases='', description='LanC lantibiotic synthetase component C-like 3 (bacterial) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36985346, end_position_on_the_genomic_accession=36986093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152330, CurrentID=0, Status='live', Symbol='CNTN4', Aliases='AXCAM, BIG-2', description='contactin 4', other_designations='axonal-associated cell adhesion molecule|brain-derived immunoglobulin superfamily protein 2|neural cell adhesion protein BIG-2', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=2098866, end_position_on_the_genomic_accession=3057961, orientation='plus', exon_count=31, OMIM=607280, no_hdr0=''),
  283403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283403, CurrentID=0, Status='live', Symbol='C12orf80', Aliases='', description='chromosome 12 open reading frame 80', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52205581, end_position_on_the_genomic_accession=52210855, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  283404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283404, CurrentID=0, Status='live', Symbol='LINC00592', Aliases='', description='long intergenic non-protein coding RNA 592', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52210930, end_position_on_the_genomic_accession=52223813, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100422413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422413, CurrentID=0, Status='live', Symbol='LOC100422413', Aliases='', description='proteasome 26S subunit, non-ATPase 7 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51372202, end_position_on_the_genomic_accession=51373400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271197, CurrentID=0, Status='live', Symbol='RPL31P42', Aliases='RPL31_13_966', description='ribosomal protein L31 pseudogene 42', other_designations='', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17108301, end_position_on_the_genomic_accession=17108634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422416, CurrentID=0, Status='live', Symbol='LOC100422416', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 7 pseudogene', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 7 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42791640, end_position_on_the_genomic_accession=42793703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060056, CurrentID=0, Status='live', Symbol='LOC101060056', Aliases='', description='mpv17-like protein 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53340978, end_position_on_the_genomic_accession=53341519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422418, CurrentID=0, Status='live', Symbol='LOC100422418', Aliases='', description='proteasome 26S subunit, non-ATPase 12 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 12 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61575768, end_position_on_the_genomic_accession=61576395, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283412, CurrentID=0, Status='live', Symbol='RPL29P12', Aliases='RPL29_6_558', description='ribosomal protein L29 pseudogene 12', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43666634, end_position_on_the_genomic_accession=43667272, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283416, CurrentID=0, Status='live', Symbol='LINC01465', Aliases='C12orf61', description='long intergenic non-protein coding RNA 1465', other_designations='putative uncharacterized protein C12orf61', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62601751, end_position_on_the_genomic_accession=62603434, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283417, CurrentID=0, Status='live', Symbol='DPY19L2', Aliases='SPATA34, SPGF9', description='dpy-19-like 2 (C. elegans)', other_designations='protein dpy-19 homolog 2|spermatogenesis associated 34', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63558913, end_position_on_the_genomic_accession=63669201, orientation='minus', exon_count=27, OMIM=613893, no_hdr0=''),
  100422426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422426, CurrentID=0, Status='live', Symbol='LOC100422426', Aliases='', description='CASP2 and RIPK1 domain containing adaptor with death domain pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156988120, end_position_on_the_genomic_accession=156988712, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283420, CurrentID=0, Status='live', Symbol='CLEC9A', Aliases='CD370, DNGR-1, DNGR1, UNQ9341', description='C-type lectin domain family 9 member A', other_designations='C-type lectin domain family 9, member A|HEEE9341', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10030677, end_position_on_the_genomic_accession=10066030, orientation='plus', exon_count=9, OMIM=612252, no_hdr0=''),
  100422429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422429, CurrentID=0, Status='live', Symbol='NAPGP2', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30961401, end_position_on_the_genomic_accession=30961978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283422, CurrentID=0, Status='live', Symbol='LINC01559', Aliases='C12orf36', description='long intergenic non-protein coding RNA 1559', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13370671, end_position_on_the_genomic_accession=13376745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422431, CurrentID=0, Status='live', Symbol='NDUFB5P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B5 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 5, 16kDa pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96953639, end_position_on_the_genomic_accession=96954195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422432, CurrentID=0, Status='live', Symbol='LOC100422432', Aliases='', description='solute carrier family 6 (amino acid transporter), member 14 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116669814, end_position_on_the_genomic_accession=116678273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422434, CurrentID=0, Status='live', Symbol='LOC100422434', Aliases='', description='complement component (3b/4b) receptor 1 (Knops blood group) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207710373, end_position_on_the_genomic_accession=207712867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422437, CurrentID=0, Status='live', Symbol='MYL6BP1', Aliases='', description='myosin light chain 6B pseudogene 1', other_designations='myosin, light chain 6B pseudogene 1|myosin, light chain 6B, alkali, smooth muscle and non-muscle pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208269228, end_position_on_the_genomic_accession=208270024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422438, CurrentID=0, Status='live', Symbol='LOC100422438', Aliases='', description='mitogen-activated protein kinase-activated protein kinase 2 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68776938, end_position_on_the_genomic_accession=68777759, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283431, CurrentID=0, Status='live', Symbol='GAS2L3', Aliases='G2L3', description='growth arrest specific 2 like 3', other_designations='growth arrest-specific 2 like 3|growth arrest-specific protein 2-like 3', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100573646, end_position_on_the_genomic_accession=100628288, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  283432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283432, CurrentID=0, Status='live', Symbol='LINC00485', Aliases='', description='long intergenic non-protein coding RNA 485', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102809283, end_position_on_the_genomic_accession=102824399, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101060060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060060, CurrentID=0, Status='live', Symbol='LOC101060060', Aliases='', description='putative glycosyltransferase ALG1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288656, CurrentID=0, Status='live', Symbol='LOC100288656', Aliases='', description='ankyrin repeat domain 20 family, member A2 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178007218, end_position_on_the_genomic_accession=178007508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283435, CurrentID=0, Status='live', Symbol='LOC283435', Aliases='', description='uncharacterized LOC283435', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271198, CurrentID=0, Status='live', Symbol='RPL31P27', Aliases='RPL31_15_765', description='ribosomal protein L31 pseudogene 27', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143327234, end_position_on_the_genomic_accession=143327695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422445, CurrentID=0, Status='live', Symbol='NUDT19P5', Aliases='', description='nudix hydrolase 19 pseudogene 5', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 19 pseudogene 5', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158182825, end_position_on_the_genomic_accession=158183393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283440, CurrentID=0, Status='live', Symbol='LOC283440', Aliases='', description='uncharacterized LOC283440', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2761200, end_position_on_the_genomic_accession=2771669, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422449, CurrentID=0, Status='live', Symbol='LOC100422449', Aliases='', description='TIA1 cytotoxic granule-associated RNA binding protein-like 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51385301, end_position_on_the_genomic_accession=51385826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422450, CurrentID=0, Status='live', Symbol='NUDT19P3', Aliases='', description='nudix hydrolase 19 pseudogene 3', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 19 pseudogene 3', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114020049, end_position_on_the_genomic_accession=114020640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422452, CurrentID=0, Status='live', Symbol='NUDT19P4', Aliases='', description='nudix hydrolase 19 pseudogene 4', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 19 pseudogene 4', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44898633, end_position_on_the_genomic_accession=44899295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422453, CurrentID=0, Status='live', Symbol='LOC100422453', Aliases='', description='polymerase (DNA directed), delta 1, catalytic subunit 125kDa pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72522553, end_position_on_the_genomic_accession=72523215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283446, CurrentID=0, Status='live', Symbol='MYO1H', Aliases='', description='myosin IH', other_designations='myosin-1H|myosin-Ih', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109388277, end_position_on_the_genomic_accession=109449004, orientation='plus', exon_count=32, OMIM=614636, no_hdr0=''),
  100422456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422456, CurrentID=0, Status='live', Symbol='LOC100422456', Aliases='', description='v-rel avian reticuloendotheliosis viral oncogene homolog A pseudogene', other_designations='v-rel reticuloendotheliosis viral oncogene homolog A pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122849770, end_position_on_the_genomic_accession=122850360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283450, CurrentID=0, Status='live', Symbol='HECTD4', Aliases='C12orf51, POTAGE', description='HECT domain containing E3 ubiquitin protein ligase 4', other_designations='AF-1 specific protein phosphatase|HECT domain-containing protein 4|probable E3 ubiquitin-protein ligase C12orf51|transmembrane protein C12orf51', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112160188, end_position_on_the_genomic_accession=112382092, orientation='minus', exon_count=76, OMIM='', no_hdr0=''),
  283454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283454, CurrentID=0, Status='live', Symbol='LOC283454', Aliases='', description='uncharacterized LOC283454', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283455, CurrentID=0, Status='live', Symbol='KSR2', Aliases='', description='kinase suppressor of ras 2', other_designations='', map_location='12q24.22-q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117453012, end_position_on_the_genomic_accession=117969103, orientation='minus', exon_count=29, OMIM=610737, no_hdr0=''),
  401675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401675, CurrentID=0, Status='live', Symbol='OR56B3P', Aliases='', description='olfactory receptor family 56 subfamily B member 3 pseudogene', other_designations='olfactory receptor, family 56, subfamily B, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6128610, end_position_on_the_genomic_accession=6129553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283458, CurrentID=0, Status='live', Symbol='NME2P1', Aliases='', description='NME/NM23 nucleoside diphosphate kinase 2 pseudogene 1', other_designations='non-metastatic cells 2, protein (NM23B) expressed in, pseudogene 1|nucleoside diphosphate kinase B-like pseudogene', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120282253, end_position_on_the_genomic_accession=120282865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283459, CurrentID=0, Status='live', Symbol='GATC', Aliases='15E1.2', description='glutamyl-tRNA(Gln) amidotransferase, subunit C', other_designations='glu-AdT subunit C|glutamyl-tRNA(Gln) amidotransferase, subunit C homolog', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120446438, end_position_on_the_genomic_accession=120463753, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283460, CurrentID=0, Status='live', Symbol='HNF1A-AS1', Aliases='C12orf27, NCRNA00262', description='HNF1A antisense RNA 1', other_designations='HNF1A antisense RNA 1 (non-protein coding)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120969838, end_position_on_the_genomic_accession=120972292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283461, CurrentID=0, Status='live', Symbol='C12orf40', Aliases='HEL-206', description='chromosome 12 open reading frame 40', other_designations='epididymis luminal protein 206', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39626170, end_position_on_the_genomic_accession=39721914, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  283463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283463, CurrentID=0, Status='live', Symbol='MUC19', Aliases='MUC-19', description='mucin 19, oligomeric', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40393394, end_position_on_the_genomic_accession=40570757, orientation='plus', exon_count=171, OMIM=612170, no_hdr0=''),
  283464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283464, CurrentID=0, Status='live', Symbol='GXYLT1', Aliases='GLT8D3', description='glucoside xylosyltransferase 1', other_designations='glycosyltransferase 8 domain containing 3|glycosyltransferase 8 domain-containing protein 3', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42081845, end_position_on_the_genomic_accession=42144871, orientation='minus', exon_count=8, OMIM=613321, no_hdr0=''),
  100422473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422473, CurrentID=0, Status='live', Symbol='LOC100422473', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131723117, end_position_on_the_genomic_accession=131724378, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271199, CurrentID=0, Status='live', Symbol='RPL31P35', Aliases='RPL31_16_810', description='ribosomal protein L31 pseudogene 35', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54656315, end_position_on_the_genomic_accession=54656775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060066, CurrentID=0, Status='live', Symbol='LOC101060066', Aliases='', description='dnaJ homolog subfamily B member 6-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179939376, end_position_on_the_genomic_accession=179940131, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422478, CurrentID=0, Status='live', Symbol='LOC100422478', Aliases='', description='potassium channel, inwardly rectifying subfamily J, member 12 pseudogene', other_designations='potassium inwardly-rectifying channel, subfamily J, member 12 pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=26981692, end_position_on_the_genomic_accession=26982282, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283471, CurrentID=0, Status='live', Symbol='TMPRSS12', Aliases='CT151', description='transmembrane (C-terminal) protease, serine 12', other_designations='transmembrane protease, serine 12', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50842918, end_position_on_the_genomic_accession=50887884, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283475, CurrentID=0, Status='live', Symbol='LOC283475', Aliases='', description='uncharacterized LOC283475', other_designations='', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  152404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152404, CurrentID=0, Status='live', Symbol='IGSF11', Aliases='BT-IgSF, CT119, CXADRL1, Igsf13, VSIG3', description='immunoglobulin superfamily member 11', other_designations='CXADR like 1|V-set and immunoglobulin domain containing 3|V-set and immunoglobulin domain-containing protein 3|brain and testis-specific immunoglobin superfamily protein|brain and testis-specific immunoglobulin superfamily protein|cancer/testis antigen 119|immunoglobulin superfamily, member 11', map_location='3q13.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118900630, end_position_on_the_genomic_accession=119146068, orientation='minus', exon_count=14, OMIM=608351, no_hdr0=''),
  152405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152405, CurrentID=0, Status='live', Symbol='C3orf30', Aliases='TSCPA', description='chromosome 3 open reading frame 30', other_designations='', map_location='3q13.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119146150, end_position_on_the_genomic_accession=119151455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422486, CurrentID=0, Status='live', Symbol='KRR1P1', Aliases='', description='KRR1, small subunit (SSU) processome component, homolog (yeast) pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20069987, end_position_on_the_genomic_accession=20072136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283480, CurrentID=0, Status='live', Symbol='FGF14-IT1', Aliases='', description='FGF14 intronic transcript 1', other_designations='FGF14 intronic transcript 1 (non-protein coding)', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102292320, end_position_on_the_genomic_accession=102394519, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  283481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283481, CurrentID=0, Status='live', Symbol='FGF14-AS2', Aliases='', description='FGF14 antisense RNA 2', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102394578, end_position_on_the_genomic_accession=102395705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283482, CurrentID=0, Status='live', Symbol='LOC283482', Aliases='', description='uncharacterized LOC283482', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102697739, end_position_on_the_genomic_accession=102704300, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283483, CurrentID=0, Status='live', Symbol='LINC00551', Aliases='', description='long intergenic non-protein coding RNA 551', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106617810, end_position_on_the_genomic_accession=106631658, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  283484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283484, CurrentID=0, Status='live', Symbol='LOC283484', Aliases='', description='uncharacterized LOC283484', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283485, CurrentID=0, Status='live', Symbol='LOC283485', Aliases='', description='uncharacterized LOC283485', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283486, CurrentID=0, Status='live', Symbol='LINC00567', Aliases='', description='long intergenic non-protein coding RNA 567', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=111462023, end_position_on_the_genomic_accession=111465435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283487, CurrentID=0, Status='live', Symbol='LINC00346', Aliases='C13orf29, NCRNA00346', description='long intergenic non-protein coding RNA 346', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110863987, end_position_on_the_genomic_accession=110870308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422496, CurrentID=0, Status='live', Symbol='PRPF19P1', Aliases='HsT1682, PRP19P, PRPF19P', description='PRPF19 pseudogene 1', other_designations='PRP19 homolog pseudogene|PRP19 pre-mRNA processing factor 19 homolog pseudogene|PRP19/PSO4 pre-mRNA processing factor 19 homolog pseudogene|PRP19/PSO4 pre-mRNA processing factor 19 pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=66068727, end_position_on_the_genomic_accession=66069645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283489, CurrentID=0, Status='live', Symbol='CHAMP1', Aliases='C13orf8, CAMP, CHAMP, ZNF828', description='chromosome alignment maintaining phosphoprotein 1', other_designations='zinc finger protein 828', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114314490, end_position_on_the_genomic_accession=114327328, orientation='plus', exon_count=3, OMIM=616327, no_hdr0=''),
  283491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283491, CurrentID=0, Status='live', Symbol='OR7E156P', Aliases='', description='olfactory receptor family 7 subfamily E member 156 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 156 pseudogene', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63737435, end_position_on_the_genomic_accession=63742568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422500, CurrentID=0, Status='live', Symbol='LOC100422500', Aliases='', description='transient receptor potential cation channel, subfamily C, member 6 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101125056, end_position_on_the_genomic_accession=101125902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422501, CurrentID=0, Status='live', Symbol='LOC100422501', Aliases='', description='COP9 signalosome subunit 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123062262, end_position_on_the_genomic_accession=123063306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422502, CurrentID=0, Status='live', Symbol='LOC100422502', Aliases='', description='transient receptor potential cation channel, subfamily C, member 6 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24749289, end_position_on_the_genomic_accession=24749895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  153469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153469, CurrentID=0, Status='live', Symbol='JAKMIP2-AS1', Aliases='', description='JAKMIP2 antisense RNA 1', other_designations='JAKMIP2 antisense RNA 1 (non-protein coding)', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147559994, end_position_on_the_genomic_accession=147662009, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271200, CurrentID=0, Status='live', Symbol='RPL31P36', Aliases='RPL31_19_846', description='ribosomal protein L31 pseudogene 36', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131520114, end_position_on_the_genomic_accession=131520546, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422510, CurrentID=0, Status='live', Symbol='LOC100422510', Aliases='', description='solute carrier family 12 (sodium/chloride transporter), member 3 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20438504, end_position_on_the_genomic_accession=20439449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270852, CurrentID=0, Status='live', Symbol='RPL7P18', Aliases='RPL7_7_596', description='ribosomal protein L7 pseudogene 18', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94825890, end_position_on_the_genomic_accession=94826716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422513, CurrentID=0, Status='live', Symbol='LOC100422513', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129992605, end_position_on_the_genomic_accession=129993105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422514, CurrentID=0, Status='live', Symbol='ATP5F1P4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153158938, end_position_on_the_genomic_accession=153159695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283507, CurrentID=0, Status='live', Symbol='SUGT1P3', Aliases='SUGT1L1', description='SGT1 homolog, MIS12 kinetochore complex assembly cochaperone pseudogene 3', other_designations='SGT1, suppressor of G2 allele of SKP1 like 1|SUGT1 pseudogene 3|suppressor of G2 allele of SKP1 pseudogene 3', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40911887, end_position_on_the_genomic_accession=40921774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422517, CurrentID=0, Status='live', Symbol='BSNDP2', Aliases='', description='barttin CLCNK-type chloride channel accessory beta subunit pseudogene 2', other_designations='Bartter syndrome, infantile, with sensorineural deafness (Barttin) pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25845221, end_position_on_the_genomic_accession=25845865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422519, CurrentID=0, Status='live', Symbol='LOC100422519', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39833286, end_position_on_the_genomic_accession=39833793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929359, CurrentID=0, Status='live', Symbol='LOC101929359', Aliases='', description='uncharacterized LOC101929359', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9594748, end_position_on_the_genomic_accession=9600782, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  283514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283514, CurrentID=0, Status='live', Symbol='SIAH3', Aliases='', description='siah E3 ubiquitin protein ligase family member 3', other_designations='siah-3', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45780270, end_position_on_the_genomic_accession=45851711, orientation='minus', exon_count=2, OMIM=615609, no_hdr0=''),
  100422523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422523, CurrentID=0, Status='live', Symbol='ATP5F1P1', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51598440, end_position_on_the_genomic_accession=51599460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283516, CurrentID=0, Status='live', Symbol='LOC283516', Aliases='', description='uncharacterized LOC283516', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283518, CurrentID=0, Status='live', Symbol='KCNRG', Aliases='DLTET', description='potassium channel regulator', other_designations='putative potassium channel regulatory protein', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50015254, end_position_on_the_genomic_accession=50020922, orientation='plus', exon_count=3, OMIM=607947, no_hdr0=''),
  100422527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422527, CurrentID=0, Status='live', Symbol='SLC4A1APP2', Aliases='', description='solute carrier family 4 (anion exchanger), member 1, adaptor protein pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187706553, end_position_on_the_genomic_accession=187708482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373325, CurrentID=0, Status='live', Symbol='LOC105373325', Aliases='', description='uncharacterized LOC105373325', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283521, CurrentID=0, Status='live', Symbol='LINC00282', Aliases='NCRNA00282', description='long intergenic non-protein coding RNA 282', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51813347, end_position_on_the_genomic_accession=51845150, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422530, CurrentID=0, Status='live', Symbol='LOC100422530', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15573033, end_position_on_the_genomic_accession=15573547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422531, CurrentID=0, Status='live', Symbol='FCF1P8', Aliases='', description='FCF1 pseudogene 8', other_designations='FCF1 small subunit processome component homolog pseudogene 8', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55351815, end_position_on_the_genomic_accession=55352408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422532, CurrentID=0, Status='live', Symbol='NDUFAF4P3', Aliases='', description='NADH:ubiquinone oxidoreductase complex assembly factor 4 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 4 pseudogene 3', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37789783, end_position_on_the_genomic_accession=37790532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422533, CurrentID=0, Status='live', Symbol='FCF1P1', Aliases='', description='FCF1 pseudogene 1', other_designations='FCF1 small subunit processome component homolog pseudogene 1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23619333, end_position_on_the_genomic_accession=23622823, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271201, CurrentID=0, Status='live', Symbol='RPL31P41', Aliases='RPL31_20_909', description='ribosomal protein L31 pseudogene 41', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65155407, end_position_on_the_genomic_accession=65155778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422536, CurrentID=0, Status='live', Symbol='FCF1P7', Aliases='', description='FCF1 pseudogene 7', other_designations='FCF1 small subunit processome component homolog pseudogene 7', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243267618, end_position_on_the_genomic_accession=243268119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422537, CurrentID=0, Status='live', Symbol='LOC100422537', Aliases='', description='signal sequence receptor, gamma (translocon-associated protein gamma) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78091499, end_position_on_the_genomic_accession=78092053, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422538, CurrentID=0, Status='live', Symbol='LOC100422538', Aliases='', description='recombination signal binding protein for immunoglobulin kappa J region pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422540, CurrentID=0, Status='live', Symbol='VTI1BP4', Aliases='', description='vesicle transport through interaction with t-SNAREs 1B pseudogene 4', other_designations='vesicle transport through interaction with t-SNAREs homolog 1B pseudogene 4', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128650998, end_position_on_the_genomic_accession=128652150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873505, CurrentID=0, Status='live', Symbol='RNA5SP249', Aliases='RN5S249', description='RNA, 5S ribosomal pseudogene 249', other_designations='RNA, 5S ribosomal 249', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147849884, end_position_on_the_genomic_accession=147849982, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134637, CurrentID=0, Status='live', Symbol='ADAT2', Aliases='DEADC1, TAD2, dJ20N2, dJ20N2.1', description='adenosine deaminase, tRNA-specific 2', other_designations='adenosine deaminase, tRNA-specific 2, TAD2 homolog|deaminase domain containing 1|deaminase domain-containing protein 1|tRNA-specific adenosine deaminase 2 homolog|tRNA-specific adenosine-34 deaminase subunit ADAT2', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143422832, end_position_on_the_genomic_accession=143450750, orientation='minus', exon_count=11, OMIM=615388, no_hdr0=''),
  283537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283537, CurrentID=0, Status='live', Symbol='SLC46A3', Aliases='FKSG16', description='solute carrier family 46 member 3', other_designations='solute carrier family 46, member 3', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28700080, end_position_on_the_genomic_accession=28719013, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100422546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422546, CurrentID=0, Status='live', Symbol='XRCC6P4', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 6 pseudogene 4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62855085, end_position_on_the_genomic_accession=62857100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422547, CurrentID=0, Status='live', Symbol='SLC16A1P1', Aliases='', description='SLC16A1 pseudogene 1', other_designations='solute carrier family 16, member 1 (monocarboxylic acid transporter 1) pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143985142, end_position_on_the_genomic_accession=143986517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422548, CurrentID=0, Status='live', Symbol='LOC100422548', Aliases='', description='solute carrier family 5 (glucose activated ion channel), member 4 pseudogene', other_designations='solute carrier family 5 (low affinity glucose cotransporter), member 4 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171751436, end_position_on_the_genomic_accession=171752028, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422549, CurrentID=0, Status='live', Symbol='LOC100422549', Aliases='', description='solute carrier family 5 (glucose activated ion channel), member 4 pseudogene', other_designations='solute carrier family 5 (low affinity glucose cotransporter), member 4 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171755803, end_position_on_the_genomic_accession=171756423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422550, CurrentID=0, Status='live', Symbol='SLC5A4P1', Aliases='', description='solute carrier family 5 (glucose activated ion channel), member 4 pseudogene 1', other_designations='solute carrier family 5 (low affinity glucose cotransporter), member 4 pseudogene 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124273970, end_position_on_the_genomic_accession=124274578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283547, CurrentID=0, Status='live', Symbol='LINC00639', Aliases='', description='long intergenic non-protein coding RNA 639', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38749339, end_position_on_the_genomic_accession=38916882, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422556, CurrentID=0, Status='live', Symbol='LOC100422556', Aliases='', description='LEO1 homolog, Paf1/RNA polymerase II complex component pseudogene', other_designations='Leo1 Paf1/RNA polymerase II complex component pseudogene|Leo1, Paf1/RNA polymerase II complex component, homolog (S. cerevisiae) pseudogene', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51918963, end_position_on_the_genomic_accession=51924120, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373326, CurrentID=0, Status='live', Symbol='LOC105373326', Aliases='', description='uncharacterized LOC105373326', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120431504, end_position_on_the_genomic_accession=120435086, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283551, CurrentID=0, Status='live', Symbol='LINC01588', Aliases='C14orf182', description='long intergenic non-protein coding RNA 1588', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49981669, end_position_on_the_genomic_accession=50007520, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  283553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283553, CurrentID=0, Status='live', Symbol='LINC00640', Aliases='', description='long intergenic non-protein coding RNA 640', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51333393, end_position_on_the_genomic_accession=51365557, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283554, CurrentID=0, Status='live', Symbol='GPR137C', Aliases='TM7SF1L2', description='G protein-coupled receptor 137C', other_designations='G protein-coupled receptor TM7SF1L2|transmembrane 7 superfamily member 1-like 2 protein', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52553148, end_position_on_the_genomic_accession=52637713, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  283556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283556, CurrentID=0, Status='live', Symbol='UBE2L1', Aliases='L-UBC, UBC4, UBCH7N3, UBCL, UBE2L7, UBE2L7P, UBEL1', description='ubiquitin conjugating enzyme E2L 1 (pseudogene)', other_designations='ubiquitin-conjugating enzyme E2L 1 (pseudogene)|ubiquitin-conjugating enzyme E2L 1 pseudogene|ubiquitin-conjugating enzyme E2L 7 pseudogene|ubiquitin-conjugating enzyme-like UBCH7N3', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55229193, end_position_on_the_genomic_accession=55229936, orientation='plus', exon_count=0, OMIM=600012, no_hdr0=''),
  152485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152485, CurrentID=0, Status='live', Symbol='ZNF827', Aliases='', description='zinc finger protein 827', other_designations='', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145757627, end_position_on_the_genomic_accession=145938455, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  283562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283562, CurrentID=0, Status='live', Symbol='RPL21P9', Aliases='', description='ribosomal protein L21 pseudogene 9', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67714505, end_position_on_the_genomic_accession=67715225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377691, CurrentID=0, Status='live', Symbol='LOC105377691', Aliases='', description='uncharacterized LOC105377691', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160263528, end_position_on_the_genomic_accession=160271065, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  283571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283571, CurrentID=0, Status='live', Symbol='PROX2', Aliases='PROX-2', description='prospero homeobox 2', other_designations='homeobox prospero-like protein PROX2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74852806, end_position_on_the_genomic_accession=74876419, orientation='minus', exon_count=7, OMIM=615094, no_hdr0=''),
  100422580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422580, CurrentID=0, Status='live', Symbol='LOC100422580', Aliases='', description='transcription elongation factor B (SIII), polypeptide 3 (110kDa, elongin A) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=123695320, end_position_on_the_genomic_accession=123697393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422581, CurrentID=0, Status='live', Symbol='DFFBP1', Aliases='', description='DNA fragmentation factor, 40kDa, beta polypeptide pseudogene 1', other_designations='DNA fragmentation factor, 40kDa, beta polypeptide (caspase-activated DNase) pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32566379, end_position_on_the_genomic_accession=32567126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422582, CurrentID=0, Status='live', Symbol='TCEB1P31', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 31', other_designations='transcription elongation factor B (SIII), polypeptide 1 (15kDa, elongin C) pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21447586, end_position_on_the_genomic_accession=21448222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152503, CurrentID=0, Status='live', Symbol='SH3D19', Aliases='EBP, EVE1, Kryn, SH3P19', description='SH3 domain containing 19', other_designations='ADAM-binding protein Eve-1|EEN-binding protein|SH3 domain protein D19', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151120281, end_position_on_the_genomic_accession=151228030, orientation='minus', exon_count=23, OMIM=608674, no_hdr0=''),
  283576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283576, CurrentID=0, Status='live', Symbol='ZDHHC22', Aliases='C14orf59', description='zinc finger, DHHC-type containing 22', other_designations='DHHC-22|putative palmitoyltransferase ZDHHC22|zinc finger DHHC domain-containing protein 22|zinc finger, DHHC domain containing 22', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77131270, end_position_on_the_genomic_accession=77144268, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101060084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060084, CurrentID=0, Status='live', Symbol='LOC101060084', Aliases='', description='uncharacterized LOC101060084', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93863980, end_position_on_the_genomic_accession=93870926, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283578, CurrentID=0, Status='live', Symbol='TMED8', Aliases='FAM15B', description='transmembrane p24 trafficking protein family member 8', other_designations='family with sequence similarity 15, member B|transmembrane emp24 domain containing 8|transmembrane emp24 protein transport domain containing 8', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77335021, end_position_on_the_genomic_accession=77377053, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  283579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283579, CurrentID=0, Status='live', Symbol='C14orf178', Aliases='', description='chromosome 14 open reading frame 178', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77760830, end_position_on_the_genomic_accession=77769742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422588, CurrentID=0, Status='live', Symbol='MGAT2P1', Aliases='', description='MGAT2 pseudogene 1', other_designations='mannosyl (alpha-1,6-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135730533, end_position_on_the_genomic_accession=135731130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422589, CurrentID=0, Status='live', Symbol='MGAT2P2', Aliases='', description='MGAT2 pseudogene 2', other_designations='mannosyl (alpha-1,6-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135752187, end_position_on_the_genomic_accession=135752784, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422591, CurrentID=0, Status='live', Symbol='SDHCP3', Aliases='', description='succinate dehydrogenase complex subunit C pseudogene 3', other_designations='succinate dehydrogenase complex, subunit C pseudogene 3|succinate dehydrogenase complex, subunit C, integral membrane protein, 15kDa pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23943823, end_position_on_the_genomic_accession=23945078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283585, CurrentID=0, Status='live', Symbol='LOC283585', Aliases='', description='uncharacterized LOC283585', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=86905778, end_position_on_the_genomic_accession=86922755, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283587, CurrentID=0, Status='live', Symbol='LINC01146', Aliases='', description='long intergenic non-protein coding RNA 1146', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88024550, end_position_on_the_genomic_accession=88087346, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  283588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283588, CurrentID=0, Status='live', Symbol='LOC283588', Aliases='', description='uncharacterized LOC283588', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=91314206, end_position_on_the_genomic_accession=91316995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101060086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060086, CurrentID=0, Status='live', Symbol='LOC101060086', Aliases='', description='cell cycle exit and neuronal differentiation protein 1-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24322606, end_position_on_the_genomic_accession=24323460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152518, CurrentID=0, Status='live', Symbol='NFXL1', Aliases='CDZFP, HOZFP, OZFP, URCC5', description='nuclear transcription factor, X-box binding-like 1', other_designations='cytoplasm-distribution zinc finger protein|ovarian zinc finger protein|up-regulated in colon cancer 5', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47847233, end_position_on_the_genomic_accession=47914667, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  152519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152519, CurrentID=0, Status='live', Symbol='NIPAL1', Aliases='NPAL1', description='NIPA-like domain containing 1', other_designations='NIPA-like protein 1|non-imprinted in Prader-Willi/Angelman syndrome region protein 3', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48016774, end_position_on_the_genomic_accession=48040173, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  283592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283592, CurrentID=0, Status='live', Symbol='FAM181A-AS1', Aliases='C14orf86', description='FAM181A antisense RNA 1', other_designations='FAM181A antisense RNA 1 (non-protein coding)', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93904730, end_position_on_the_genomic_accession=93926372, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100873507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873507, CurrentID=0, Status='live', Symbol='RNA5SP251', Aliases='RN5S251', description='RNA, 5S ribosomal pseudogene 251', other_designations='RNA, 5S ribosomal 251', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=3700492, end_position_on_the_genomic_accession=3700628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422602, CurrentID=0, Status='live', Symbol='ST13P14', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 14', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50877065, end_position_on_the_genomic_accession=50877661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422603, CurrentID=0, Status='live', Symbol='ST13P15', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 15', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143517123, end_position_on_the_genomic_accession=143518229, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283596, CurrentID=0, Status='live', Symbol='SNHG10', Aliases='C14orf62, LINC00063, NCRNA00063', description='small nucleolar RNA host gene 10', other_designations='long intergenic non-protein coding RNA 63|small nucleolar RNA host gene 10 (non-protein coding)', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95532912, end_position_on_the_genomic_accession=95534872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  374946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374946, CurrentID=0, Status='live', Symbol='DRAXIN', Aliases='AGPA3119, C1orf187, UNQ3119, neucrin', description='dorsal inhibitory axon guidance protein', other_designations='dorsal repulsive axon guidance protein|neural tissue-specific cysteine-rich protein', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11686647, end_position_on_the_genomic_accession=11720612, orientation='plus', exon_count=10, OMIM=612682, no_hdr0=''),
  283598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283598, CurrentID=0, Status='live', Symbol='C14orf177', Aliases='', description='chromosome 14 open reading frame 177', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98711613, end_position_on_the_genomic_accession=98717766, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  222029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222029, CurrentID=0, Status='live', Symbol='DKFZp434L192', Aliases='', description='uncharacterized protein DKFZp434L192', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56496223, end_position_on_the_genomic_accession=56497284, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283600, CurrentID=0, Status='live', Symbol='SLC25A47', Aliases='C14orf68, HDMCP, HMFN1655', description='solute carrier family 25 member 47', other_designations='HCC-down-regulated mitochondrial carrier protein|hepatocellular carcinoma down-regulated mitochondrial carrier protein|hepatocellular carcinoma-downregulated mitochondrial carrier protein|solute carrier family 25, member 47', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100323342, end_position_on_the_genomic_accession=100330378, orientation='plus', exon_count=6, OMIM=609911, no_hdr0=''),
  283601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283601, CurrentID=0, Status='live', Symbol='LINC00523', Aliases='C14orf70', description='long intergenic non-protein coding RNA 523', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100657268, end_position_on_the_genomic_accession=100672744, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422610, CurrentID=0, Status='live', Symbol='LOC100422610', Aliases='', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 3 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53606134, end_position_on_the_genomic_accession=53607056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283603, CurrentID=0, Status='live', Symbol='RPL26P4', Aliases='RPL26_18_1407', description='ribosomal protein L26 pseudogene 4', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101677937, end_position_on_the_genomic_accession=101678423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422614, CurrentID=0, Status='live', Symbol='LOC100422614', Aliases='', description='solute carrier family 7 (anionic amino acid transporter), member 13 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85086615, end_position_on_the_genomic_accession=85087186, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060089, CurrentID=0, Status='live', Symbol='LOC101060089', Aliases='', description='cleavage and polyadenylation specificity factor subunit 5-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117117731, end_position_on_the_genomic_accession=117126384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373328, CurrentID=0, Status='live', Symbol='LOC105373328', Aliases='', description='uncharacterized LOC105373328', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121019062, end_position_on_the_genomic_accession=121047453, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131189, CurrentID=0, Status='live', Symbol='SSR1P2', Aliases='', description='signal sequence receptor, alpha pseudogene 2', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75136178, end_position_on_the_genomic_accession=75136937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060090, CurrentID=0, Status='live', Symbol='LOC101060090', Aliases='', description='uncharacterized LOC101060090', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422622, CurrentID=0, Status='live', Symbol='LOC100422622', Aliases='', description='TAF10 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 30kDa pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31465062, end_position_on_the_genomic_accession=31465906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422623, CurrentID=0, Status='live', Symbol='LOC100422623', Aliases='', description='carcinoembryonic antigen-related cell adhesion molecule 1 (biliary glycoprotein) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41803756, end_position_on_the_genomic_accession=41808067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422624, CurrentID=0, Status='live', Symbol='SLC25A15P1', Aliases='SLC25A15P, SLC25A15PY1', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 pseudogene 1', other_designations='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 Y-linked pseudogene|solute carrier family 25 member 15 pseudogene 1|solute carrier family 25 member 15 pseudogene, Y-linked 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26586645, end_position_on_the_genomic_accession=26587564, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283617, CurrentID=0, Status='live', Symbol='OR4K3', Aliases='OR14-10P, OR4K3', description='olfactory receptor family 4 subfamily K member 3 (gene/pseudogene)', other_designations='olfactory receptor, family 4, subfamily K, member 3 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19867997, end_position_on_the_genomic_accession=19869343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060091, CurrentID=0, Status='live', Symbol='LOC101060091', Aliases='', description='uncharacterized LOC101060091', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113831183, end_position_on_the_genomic_accession=113890992, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422628, CurrentID=0, Status='live', Symbol='LOC100422628', Aliases='', description='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11334233, end_position_on_the_genomic_accession=11335901, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283624, CurrentID=0, Status='live', Symbol='LINC00641', Aliases='', description='long intergenic non-protein coding RNA 641', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21200079, end_position_on_the_genomic_accession=21206900, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422633, CurrentID=0, Status='live', Symbol='LOC100422633', Aliases='', description='ALG11, alpha-1,2-mannosyltransferase pseudogene', other_designations='asparagine-linked glycosylation 11, alpha-1,2-mannosyltransferase homolog pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8173238, end_position_on_the_genomic_accession=8175557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422634, CurrentID=0, Status='live', Symbol='CASP1P3', Aliases='', description='caspase 1 pseudogene 3', other_designations='caspase 1, apoptosis-related cysteine peptidase (interleukin 1, beta, convertase) pseudogene|caspase 1, apoptosis-related cysteine peptidase pseudogene 3', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105095438, end_position_on_the_genomic_accession=105097589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422635, CurrentID=0, Status='live', Symbol='SLC25A15P3', Aliases='', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24943575, end_position_on_the_genomic_accession=24944497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422636, CurrentID=0, Status='live', Symbol='SLC25A15P4', Aliases='', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 pseudogene 4', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10612455, end_position_on_the_genomic_accession=10613379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283629, CurrentID=0, Status='live', Symbol='TSSK4', Aliases='C14orf20, STK22E, TSK-4, TSK4, TSSK5', description='testis-specific serine kinase 4', other_designations='serine/threonine kinase 22E|serine/threonine-protein kinase 22E', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24205530, end_position_on_the_genomic_accession=24208248, orientation='plus', exon_count=4, OMIM=610711, no_hdr0=''),
  100422638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422638, CurrentID=0, Status='live', Symbol='LOC100422638', Aliases='', description='MX dynamin-like GTPase 1 pseudogene', other_designations='myxovirus (influenza virus) resistance 1, interferon-inducible protein p78 (mouse) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10292191, end_position_on_the_genomic_accession=10292772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152559, CurrentID=0, Status='live', Symbol='PAQR3', Aliases='RKTG', description='progestin and adipoQ receptor family member III', other_designations='Raf kinase trapping to Golgi', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78887223, end_position_on_the_genomic_accession=78939428, orientation='minus', exon_count=15, OMIM=614577, no_hdr0=''),
  100422640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422640, CurrentID=0, Status='live', Symbol='LOC100422640', Aliases='', description='5-aminoimidazole-4-carboxamide ribonucleotide formyltransferase/IMP cyclohydrolase pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154555210, end_position_on_the_genomic_accession=154555997, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422641, CurrentID=0, Status='live', Symbol='LOC100422641', Aliases='', description='STT3A, subunit of the oligosaccharyltransferase complex (catalytic) pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48398309, end_position_on_the_genomic_accession=48399375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422642, CurrentID=0, Status='live', Symbol='SLC25A15P5', Aliases='', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 pseudogene 5', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16645692, end_position_on_the_genomic_accession=16646614, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283635, CurrentID=0, Status='live', Symbol='FAM177A1', Aliases='C14orf24', description='family with sequence similarity 177 member A1', other_designations='family with sequence similarity 177, member A1', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35044907, end_position_on_the_genomic_accession=35083383, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100422644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422644, CurrentID=0, Status='live', Symbol='SERPINA7P1', Aliases='', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 7 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106562774, end_position_on_the_genomic_accession=106563404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422645, CurrentID=0, Status='live', Symbol='LOC100422645', Aliases='', description='CTD small phosphatase 2 pseudogene', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 2 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38458757, end_position_on_the_genomic_accession=38459540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283638, CurrentID=0, Status='live', Symbol='CEP170B', Aliases='CEP170R, FAM68C, KIAA0284', description='centrosomal protein 170B', other_designations='Cep170-related', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104865313, end_position_on_the_genomic_accession=104896770, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105373329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373329, CurrentID=0, Status='live', Symbol='LOC105373329', Aliases='', description='uncharacterized LOC105373329', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123367723, end_position_on_the_genomic_accession=123471972, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  283643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283643, CurrentID=0, Status='live', Symbol='C14orf80', Aliases='', description='chromosome 14 open reading frame 80', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105489855, end_position_on_the_genomic_accession=105499248, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  152573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152573, CurrentID=0, Status='live', Symbol='SHISA3', Aliases='hShisa3', description='shisa family member 3', other_designations='shisa homolog 3', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42397839, end_position_on_the_genomic_accession=42402487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422656, CurrentID=0, Status='live', Symbol='EEF1DP7', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 7', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63636601, end_position_on_the_genomic_accession=63637110, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422657, CurrentID=0, Status='live', Symbol='RAC1P6', Aliases='', description='ras-related C3 botulinum toxin substrate 1 pseudogene 6', other_designations='ras-related C3 botulinum toxin substrate 1 (rho family, small GTP binding protein Rac1) pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115136576, end_position_on_the_genomic_accession=115137162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152578, CurrentID=0, Status='live', Symbol='LINC01618', Aliases='', description='long intergenic non-protein coding RNA 1618', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52789994, end_position_on_the_genomic_accession=52815464, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283651, CurrentID=0, Status='live', Symbol='HMGN2P46', Aliases='C15orf21, D-PCa-2', description='high mobility group nucleosomal binding domain 2 pseudogene 46', other_designations='Dresden prostate cancer 2|Dresden prostate carcinoma 2|high-mobility group nucleosomal binding domain 2 pseudogene 46', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45511136, end_position_on_the_genomic_accession=45556730, orientation='plus', exon_count=6, OMIM=611314, no_hdr0=''),
  283652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283652, CurrentID=0, Status='live', Symbol='SLC24A5', Aliases='JSX, NCKX5, OCA6, SHEP4', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 5', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 5|ion transporter JSX|oculocutaneous albinism 6 (autosomal recessive)|solute carrier family 24, member 5', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48120972, end_position_on_the_genomic_accession=48142392, orientation='plus', exon_count=9, OMIM=609802, no_hdr0=''),
  101929723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929723, CurrentID=0, Status='live', Symbol='LOC101929723', Aliases='', description='uncharacterized LOC101929723', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41931599, end_position_on_the_genomic_accession=41933723, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419599, CurrentID=0, Status='live', Symbol='RNF10P1', Aliases='', description='ring finger protein 10 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5068627, end_position_on_the_genomic_accession=5069241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152586, CurrentID=0, Status='live', Symbol='MGAT4D', Aliases='GnT1IP', description='MGAT4 family member D', other_designations='GlcNAcT-I Inhibitory Protein|MGAT4 family, member D|glycosyltransferase 54 domain containing 1|glycosyltransferase 54 domain-containing protein|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase family, member D', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140442258, end_position_on_the_genomic_accession=140499829, orientation='minus', exon_count=14, OMIM=610310, no_hdr0=''),
  283659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283659, CurrentID=0, Status='live', Symbol='PRTG', Aliases='IGDCC5', description='protogenin', other_designations='immunoglobulin superfamily, DCC subclass, member 5|protogenin homolog (Gallus gallus)|shen-dan', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55611540, end_position_on_the_genomic_accession=55743649, orientation='minus', exon_count=22, OMIM=613261, no_hdr0=''),
  101060098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060098, CurrentID=0, Status='live', Symbol='LOC101060098', Aliases='', description="deoxyuridine 5'-triphosphate nucleotidohydrolase, mitochondrial-like", other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69025855, end_position_on_the_genomic_accession=69026675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283663, CurrentID=0, Status='live', Symbol='LINC00926', Aliases='', description='long intergenic non-protein coding RNA 926', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57300365, end_position_on_the_genomic_accession=57307769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283665, CurrentID=0, Status='live', Symbol='LOC283665', Aliases='', description='uncharacterized LOC283665', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422674, CurrentID=0, Status='live', Symbol='YWHAQP7', Aliases='', description='YWHAQ pseudogene 7', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76619380, end_position_on_the_genomic_accession=76620045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060099, CurrentID=0, Status='live', Symbol='LOC101060099', Aliases='', description='CST complex subunit TEN1-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81487816, end_position_on_the_genomic_accession=81488758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373330, CurrentID=0, Status='live', Symbol='LOC105373330', Aliases='', description='uncharacterized LOC105373330', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123187596, end_position_on_the_genomic_accession=123226632, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283673, CurrentID=0, Status='live', Symbol='EWSAT1', Aliases='LINC00277, NCRNA00277, TMEM84', description='Ewing sarcoma associated transcript 1', other_designations='long intergenic non-protein coding RNA 277|transmembrane protein 84', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69080850, end_position_on_the_genomic_accession=69095824, orientation='plus', exon_count=4, OMIM=616492, no_hdr0=''),
  283674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283674, CurrentID=0, Status='live', Symbol='LOC283674', Aliases='', description='uncharacterized LOC283674', other_designations='', map_location='15q22.32', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283677, CurrentID=0, Status='live', Symbol='REC114', Aliases='C15orf60, CT147', description='REC114 meiotic recombination protein', other_designations='meiotic recombination protein REC114-like', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73376209, end_position_on_the_genomic_accession=73560013, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100422686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422686, CurrentID=0, Status='live', Symbol='SLC25A6P3', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51709083, end_position_on_the_genomic_accession=51709939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414751, CurrentID=0, Status='live', Symbol='TRAPPC2P8', Aliases='TRAPPC2P3B', description='trafficking protein particle complex 2 pseudogene 8', other_designations='spondyloepiphyseal dysplasia, late, pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18105192, end_position_on_the_genomic_accession=18109825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414752, CurrentID=0, Status='live', Symbol='TRAPPC2P10', Aliases='TRAPPC2P5B', description='trafficking protein particle complex 2 pseudogene 10', other_designations='spondyloepiphyseal dysplasia, late, pseudogene, centromeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23963031, end_position_on_the_genomic_accession=23970797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414753, CurrentID=0, Status='live', Symbol='TRAPPC2P9', Aliases='TRAPPC2P4B', description='trafficking protein particle complex 2 pseudogene 9', other_designations='spondyloepiphyseal dysplasia, late, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23754870, end_position_on_the_genomic_accession=23765685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283682, CurrentID=0, Status='live', Symbol='LINC01579', Aliases='', description='long intergenic non-protein coding RNA 1579', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93718542, end_position_on_the_genomic_accession=93900777, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100422691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422691, CurrentID=0, Status='live', Symbol='ITGB1P1', Aliases='FNRBL, ITGB1L', description='integrin subunit beta 1 pseudogene 1', other_designations='Fibronectin receptor, beta subunit, pseudogene (insertion polymorphism)|integrin beta 1 pseudogene 1|integrin, beta 1 (fibronectin receptor, beta polypeptide, antigen CD29 includes MDF2, MSK12) pseudogene', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14622221, end_position_on_the_genomic_accession=14623288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377695, CurrentID=0, Status='live', Symbol='LOC105377695', Aliases='', description='uncharacterized LOC105377695', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161735747, end_position_on_the_genomic_accession=161766913, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283685, CurrentID=0, Status='live', Symbol='GOLGA6L2', Aliases='CT105', description='golgin A6 family-like 2', other_designations='cancer/testis antigen 105|golgi autoantigen, golgin subfamily a, 6-like 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23439386, end_position_on_the_genomic_accession=23447243, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100422694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422694, CurrentID=0, Status='live', Symbol='LOC100422694', Aliases='', description='polypeptide N-acetylgalactosaminyltransferase 11 pseudogene', other_designations='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 11 (GalNAc-T11) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65575954, end_position_on_the_genomic_accession=65578499, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283687, CurrentID=0, Status='live', Symbol='ST20-AS1', Aliases='C15orf37', description='ST20 antisense RNA 1', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79922771, end_position_on_the_genomic_accession=79924854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283688, CurrentID=0, Status='live', Symbol='LINC00927', Aliases='', description='long intergenic non-protein coding RNA 927', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80263068, end_position_on_the_genomic_accession=80341805, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  414761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414761, CurrentID=0, Status='live', Symbol='HCG15', Aliases='dJ25J6.6', description='HLA complex group 15 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414763, CurrentID=0, Status='live', Symbol='BMS1P18', Aliases='C14orf17, LINC00516', description='BMS1, ribosome biogenesis factor pseudogene 18', other_designations='BMS1 pseudogene 18|BMS1 ribosome biogenesis factor pseudogene 18|long intergenic non-protein coding RNA 516|ribosome biogenesis protein BMS1 homolog pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19083134, end_position_on_the_genomic_accession=19093337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  414764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414764, CurrentID=0, Status='live', Symbol='HCG23', Aliases='dJ1077I5.3', description='HLA complex group 23 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32390510, end_position_on_the_genomic_accession=32393691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283693, CurrentID=0, Status='live', Symbol='ACTG1P17', Aliases='', description='actin gamma 1 pseudogene 17', other_designations='actin, alpha 2, smooth muscle, aorta pseudogene|actin, gamma 1 pseudogene', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82725880, end_position_on_the_genomic_accession=82739780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283694, CurrentID=0, Status='live', Symbol='OR4N4', Aliases='OR15-1, OR15-5', description='olfactory receptor family 4 subfamily N member 4', other_designations='olfactory receptor OR15-1|olfactory receptor OR15-5|olfactory receptor, family 4, subfamily N, member 4', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22094431, end_position_on_the_genomic_accession=22095864, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  414767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414767, CurrentID=0, Status='live', Symbol='LINC00596', Aliases='C14orf165, CNSLT1I7G', description='long intergenic non-protein coding RNA 596', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=24391456, end_position_on_the_genomic_accession=24403777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  414768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414768, CurrentID=0, Status='live', Symbol='HCG24', Aliases='dJ1033B10.15', description='HLA complex group 24 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33143388, end_position_on_the_genomic_accession=33147298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060104, CurrentID=0, Status='live', Symbol='LOC101060104', Aliases='', description='spidroin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32115943, end_position_on_the_genomic_accession=32117539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283698, CurrentID=0, Status='live', Symbol='SYNGR2P1', Aliases='SYNGR2P', description='synaptogyrin 2 pseudogene 1', other_designations='cellugyrin pseudogene', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30044440, end_position_on_the_genomic_accession=30046080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414771, CurrentID=0, Status='live', Symbol='DDR1-AS1', Aliases='TIGD1L, bCX111D4.7, bPG70P20.2, bQB10J12.1', description='DDR1 antisense RNA 1 (head to head)', other_designations='DDR1 antisense RNA 1 (non-protein coding)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283701, CurrentID=0, Status='live', Symbol='LOC283701', Aliases='', description='uncharacterized LOC283701', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=30701113, end_position_on_the_genomic_accession=30701623, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  414774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414774, CurrentID=0, Status='live', Symbol='HCG20', Aliases='NCRNA00149', description='HLA complex group 20 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060105, CurrentID=0, Status='live', Symbol='FAM183DP', Aliases='', description='family with sequence similarity 183 member D, pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102249767, end_position_on_the_genomic_accession=102250794, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422712, CurrentID=0, Status='live', Symbol='PRKRIRP7', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 7', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78759445, end_position_on_the_genomic_accession=78762560, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414777, CurrentID=0, Status='live', Symbol='HCG18', Aliases='', description='HLA complex group 18 (non-protein coding)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30287397, end_position_on_the_genomic_accession=30327156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  414778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414778, CurrentID=0, Status='live', Symbol='HCG17', Aliases='HCG18, LINC00046, NCRNA00046', description='HLA complex group 17 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 46', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30234039, end_position_on_the_genomic_accession=30326134, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422715, CurrentID=0, Status='live', Symbol='PRKRIRP10', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 10', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49232874, end_position_on_the_genomic_accession=49234399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422716, CurrentID=0, Status='live', Symbol='ST6GALNAC2P1', Aliases='', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 2 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84039912, end_position_on_the_genomic_accession=84040763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422717, CurrentID=0, Status='live', Symbol='LOC100422717', Aliases='', description='pterin-4 alpha-carbinolamine dehydratase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1) 2 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3619256, end_position_on_the_genomic_accession=3619913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283710, CurrentID=0, Status='live', Symbol='LOC283710', Aliases='', description='uncharacterized LOC283710', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31221813, end_position_on_the_genomic_accession=31230847, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  283711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283711, CurrentID=0, Status='live', Symbol='UBE2CP4', Aliases='', description='ubiquitin conjugating enzyme E2C pseudogene 4', other_designations='ubiquitin-conjugating enzyme E2C pseudogene 4', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31420117, end_position_on_the_genomic_accession=31421168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152641, CurrentID=0, Status='live', Symbol='WWC2-AS2', Aliases='C4orf38', description='WWC2 antisense RNA 2', other_designations='WWC2 antisense RNA 2 (non-protein coding)', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183097021, end_position_on_the_genomic_accession=183099199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377696, CurrentID=0, Status='live', Symbol='LOC105377696', Aliases='', description='uncharacterized LOC105377696', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161809046, end_position_on_the_genomic_accession=161819517, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  266954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266954, CurrentID=0, Status='live', Symbol='CABYRP1', Aliases='CABYRP', description='calcium binding tyrosine-(Y)-phosphorylation regulated pseudogene 1', other_designations='calcium binding tyrosine-(Y)-phosphorylation regulated (fibrousheathin 2) pseudogene|calcium-binding tyrosine phosphorylation-regulated protein pseudogene|calcium-binding tyrosine-(Y)-phosphorylation regulated (fibrousheathin 2) pseudogene', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54047660, end_position_on_the_genomic_accession=54048819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060108, CurrentID=0, Status='live', Symbol='LOC101060108', Aliases='', description='uncharacterized LOC101060108', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422730, CurrentID=0, Status='live', Symbol='LOC100422730', Aliases='', description='sema domain, seven thrombospondin repeats (type 1 and type 1-like), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 5A pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7927016, end_position_on_the_genomic_accession=7927603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283726, CurrentID=0, Status='live', Symbol='SAXO2', Aliases='FAM154B', description='stabilizer of axonemal microtubules 2', other_designations='family with sequence similarity 154, member B|protein FAM154B', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82262811, end_position_on_the_genomic_accession=82284927, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  283728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283728, CurrentID=0, Status='live', Symbol='LOC283728', Aliases='', description='uncharacterized LOC283728', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422737, CurrentID=0, Status='live', Symbol='LOC100422737', Aliases='', description='uncharacterized LOC100422737', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106717452, end_position_on_the_genomic_accession=106787425, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100291666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100291666, CurrentID=0, Status='live', Symbol='LOC100291666', Aliases='', description='serologically defined breast cancer antigen NY-BR-40', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283731, CurrentID=0, Status='live', Symbol='LOC283731', Aliases='', description='uncharacterized LOC283731', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74126373, end_position_on_the_genomic_accession=74129278, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  152663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152663, CurrentID=0, Status='live', Symbol='RPL7AP27', Aliases='RPL7A_11_540', description='ribosomal protein L7a pseudogene 27', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188349318, end_position_on_the_genomic_accession=188350047, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  222052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222052, CurrentID=0, Status='live', Symbol='TTC4P1', Aliases='TTC4P', description='tetratricopeptide repeat domain 4 pseudogene 1', other_designations='tetratricopeptide repeat domain 4, pseudogene', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45999713, end_position_on_the_genomic_accession=46000778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101340250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101340250, CurrentID=0, Status='live', Symbol='SNORA11E', Aliases='', description='small nucleolar RNA, H/ACA box 11E', other_designations='SNORA11E snoRNA', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52063347, end_position_on_the_genomic_accession=52063473, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101340251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101340251, CurrentID=0, Status='live', Symbol='SNORD124', Aliases='', description='small nucleolar RNA, C/D box 124', other_designations='', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40027542, end_position_on_the_genomic_accession=40027645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101340252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101340252, CurrentID=0, Status='live', Symbol='SNORD121B', Aliases='', description='small nucleolar RNA, C/D box 121B', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33934297, end_position_on_the_genomic_accession=33934375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283742, CurrentID=0, Status='live', Symbol='FAM98B', Aliases='', description='family with sequence similarity 98 member B', other_designations='family with sequence similarity 98, member B', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38454127, end_position_on_the_genomic_accession=38487710, orientation='plus', exon_count=8, OMIM=616142, no_hdr0=''),
  283745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283745, CurrentID=0, Status='live', Symbol='LOC283745', Aliases='', description='uncharacterized LOC283745', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283747, CurrentID=0, Status='live', Symbol='TCEB1P2', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 2', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41556977, end_position_on_the_genomic_accession=41557676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283748, CurrentID=0, Status='live', Symbol='PLA2G4D', Aliases='cPLA2delta', description='phospholipase A2 group IVD', other_designations='cPLA2-delta|phospholipase A2 delta, cytosolic|phospholipase A2, group IVD (cytosolic)', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42067565, end_position_on_the_genomic_accession=42094559, orientation='minus', exon_count=21, OMIM=612864, no_hdr0=''),
  101060113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060113, CurrentID=0, Status='live', Symbol='LOC101060113', Aliases='', description='proteasome maturation protein pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=83559829, end_position_on_the_genomic_accession=83560312, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283755, CurrentID=0, Status='live', Symbol='HERC2P3', Aliases='D15F37S4', description='hect domain and RLD 2 pseudogene 3', other_designations='', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20408396, end_position_on_the_genomic_accession=20506180, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  152687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152687, CurrentID=0, Status='live', Symbol='ZNF595', Aliases='', description='zinc finger protein 595', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53285, end_position_on_the_genomic_accession=88211, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  283761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283761, CurrentID=0, Status='live', Symbol='LINC00928', Aliases='', description='long intergenic non-protein coding RNA 928', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89504930, end_position_on_the_genomic_accession=89524034, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131190, CurrentID=0, Status='live', Symbol='LOC100131190', Aliases='', description='uncharacterized LOC100131190', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.7', start_position_on_the_genomic_accession=41657429, end_position_on_the_genomic_accession=41657848, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283767, CurrentID=0, Status='live', Symbol='GOLGA6L1', Aliases='', description='golgin A6 family-like 1', other_designations='golgi autoantigen, golgin subfamily a, 6-like 1', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23127066, end_position_on_the_genomic_accession=23136822, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  283768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283768, CurrentID=0, Status='live', Symbol='GOLGA8G', Aliases='', description='golgin A8 family member G', other_designations='golgi autoantigen, golgin subfamily a, 8G|golgin A8 family, member G', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28519611, end_position_on_the_genomic_accession=28532997, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100422781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422781, CurrentID=0, Status='live', Symbol='LOC100422781', Aliases='', description='uncharacterized LOC100422781', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283777, CurrentID=0, Status='live', Symbol='FAM169B', Aliases='', description='family with sequence similarity 169 member B', other_designations='family with sequence similarity 169, member B', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98437162, end_position_on_the_genomic_accession=98514382, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  90987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90987, CurrentID=0, Status='live', Symbol='ZNF251', Aliases='', description='zinc finger protein 251', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144720909, end_position_on_the_genomic_accession=144755585, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  152709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152709, CurrentID=0, Status='live', Symbol='ARL4AP2', Aliases='ARL4P2', description='ADP ribosylation factor like GTPase 4A pseudogene 2', other_designations='ADP-ribosylation factor like GTPase 4A pseudogene 2|ADP-ribosylation factor-like 4 pseudogene 2|ADP-ribosylation factor-like 4A pseudogene 2', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40785888, end_position_on_the_genomic_accession=40787036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372072, CurrentID=0, Status='live', Symbol='LOC105372072', Aliases='', description='uncharacterized LOC105372072', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37753615, end_position_on_the_genomic_accession=37776383, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101060119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060119, CurrentID=0, Status='live', Symbol='LOC101060119', Aliases='', description='RING-box protein 2-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35337353, end_position_on_the_genomic_accession=35344574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283788, CurrentID=0, Status='live', Symbol='LOC283788', Aliases='', description='FSHD region gene 1 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283796, CurrentID=0, Status='live', Symbol='GOLGA8IP', Aliases='GOLGA8I, GOLGA9P', description='golgin A8 family member I, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 9 pseudogene|golgin A8 family, member I (pseudogene)|golgin A8 family, member I, pseudogene|golgin A9 (pseudogene)', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22610353, end_position_on_the_genomic_accession=22617854, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  90990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90990, CurrentID=0, Status='live', Symbol='KIFC2', Aliases='', description='kinesin family member C2', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144466149, end_position_on_the_genomic_accession=144474116, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100270854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270854, CurrentID=0, Status='live', Symbol='RPL8P3', Aliases='RPL8_1_413', description='ribosomal protein L8 pseudogene 3', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142827054, end_position_on_the_genomic_accession=142827728, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283804, CurrentID=0, Status='live', Symbol='LOC283804', Aliases='', description='disintegrin and metalloproteinase domain-containing protein 21-like', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21053002, end_position_on_the_genomic_accession=21055005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377699, CurrentID=0, Status='live', Symbol='LOC105377699', Aliases='', description='uncharacterized LOC105377699', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=162795811, end_position_on_the_genomic_accession=163012064, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  283807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283807, CurrentID=0, Status='live', Symbol='FBXL22', Aliases='Fbl22', description='F-box and leucine-rich repeat protein 22', other_designations='F-box/LRR-repeat protein 22', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63594830, end_position_on_the_genomic_accession=63602428, orientation='plus', exon_count=3, OMIM=609088, no_hdr0=''),
  100422821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422821, CurrentID=0, Status='live', Symbol='MIR1273C', Aliases='', description='microRNA 1273c', other_designations='hsa-mir-1273c', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154853360, end_position_on_the_genomic_accession=154853436, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  414886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414886, CurrentID=0, Status='live', Symbol='SPG27', Aliases='', description='spastic paraplegia 27 (autosomal recessive)', other_designations='', map_location='10q22.1-q24.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609041, no_hdr0=''),
  90993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90993, CurrentID=0, Status='live', Symbol='CREB3L1', Aliases='OASIS', description='cAMP responsive element binding protein 3-like 1', other_designations='BBF-2 homolog|cAMP-responsive element-binding protein 3-like protein 1|old astrocyte specifically-induced substance', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46277638, end_position_on_the_genomic_accession=46321422, orientation='plus', exon_count=13, OMIM=616215, no_hdr0=''),
  100422824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422824, CurrentID=0, Status='live', Symbol='MIR3128', Aliases='mir-3128', description='microRNA 3128', other_designations='hsa-mir-3128', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177255945, end_position_on_the_genomic_accession=177256010, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422825, CurrentID=0, Status='live', Symbol='MIR3160-2', Aliases='mir-3160-2', description='microRNA 3160-2', other_designations='hsa-mir-3160-2', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46451807, end_position_on_the_genomic_accession=46451887, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422826, CurrentID=0, Status='live', Symbol='MIR4274', Aliases='', description='microRNA 4274', other_designations='hsa-mir-4274', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7460028, end_position_on_the_genomic_accession=7460118, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422827, CurrentID=0, Status='live', Symbol='MIR3160-1', Aliases='mir-3160-1', description='microRNA 3160-1', other_designations='hsa-mir-3160-1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46451805, end_position_on_the_genomic_accession=46451889, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283820, CurrentID=0, Status='live', Symbol='NOMO2', Aliases='Nomo, PM5', description='NODAL modulator 2', other_designations='pM5 protein 2|pM5 protein, centromeric copy', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18499924, end_position_on_the_genomic_accession=18562112, orientation='minus', exon_count=31, OMIM=609158, no_hdr0=''),
  100422829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422829, CurrentID=0, Status='live', Symbol='MIR4319', Aliases='', description='microRNA 4319', other_designations='hsa-mir-4319', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44970082, end_position_on_the_genomic_accession=44970166, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422830, CurrentID=0, Status='live', Symbol='MIR3171', Aliases='', description='microRNA 3171', other_designations='hsa-mir-3171', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27633205, end_position_on_the_genomic_accession=27633278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422831, CurrentID=0, Status='live', Symbol='MIR1302-9', Aliases='', description='microRNA 1302-9', other_designations='hsa-mir-1302-9', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30144, end_position_on_the_genomic_accession=30281, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422832, CurrentID=0, Status='live', Symbol='MIR3191', Aliases='mir-3191', description='microRNA 3191', other_designations='hsa-mir-3191', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47226944, end_position_on_the_genomic_accession=47227019, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422833, CurrentID=0, Status='live', Symbol='MIR3188', Aliases='mir-3188', description='microRNA 3188', other_designations='hsa-mir-3188', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18282077, end_position_on_the_genomic_accession=18282161, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422834, CurrentID=0, Status='live', Symbol='MIR1302-10', Aliases='', description='microRNA 1302-10', other_designations='hsa-mir-1302-10', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101960459, end_position_on_the_genomic_accession=101960596, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  414899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414899, CurrentID=0, Status='live', Symbol='BLID', Aliases='BRCC2', description='BH3-like motif containing, cell death inducer', other_designations='breast cancer cell 2|breast cancer cell protein 2', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122115354, end_position_on_the_genomic_accession=122116215, orientation='minus', exon_count=1, OMIM=608853, no_hdr0=''),
  152756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152756, CurrentID=0, Status='live', Symbol='FAM218A', Aliases='C4orf39', description='family with sequence similarity 218 member A', other_designations='family with sequence similarity 218, member A|uncharacterized protein C4orf39', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164956948, end_position_on_the_genomic_accession=164959121, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422837, CurrentID=0, Status='live', Symbol='MIR1193', Aliases='', description='microRNA 1193', other_designations='hsa-mir-1193', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101030052, end_position_on_the_genomic_accession=101030129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422838, CurrentID=0, Status='live', Symbol='MIR3195', Aliases='', description='microRNA 3195', other_designations='hsa-mir-3195', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62064802, end_position_on_the_genomic_accession=62064885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422839, CurrentID=0, Status='live', Symbol='MIR3119-1', Aliases='', description='microRNA 3119-1', other_designations='hsa-mir-3119-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170151378, end_position_on_the_genomic_accession=170151462, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422840, CurrentID=0, Status='live', Symbol='MIR4317', Aliases='', description='microRNA 4317', other_designations='hsa-mir-4317', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6374361, end_position_on_the_genomic_accession=6374425, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  222068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222068, CurrentID=0, Status='live', Symbol='TMED4', Aliases='ERS25, GMP25iso, HNLF, p24alpha3', description='transmembrane p24 trafficking protein 4', other_designations='endoplasmic reticulum stress-response protein 25 kDa|p24 family protein alpha-3|putative NF-kappa-B-activating protein 156|putative NFkB activating protein HNLF|transmembrane emp24 protein transport domain containing 4', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44577894, end_position_on_the_genomic_accession=44582295, orientation='minus', exon_count=5, OMIM=612038, no_hdr0=''),
  100422842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422842, CurrentID=0, Status='live', Symbol='MIR3074', Aliases='mir-3074', description='microRNA 3074', other_designations='hsa-mir-3074', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95086014, end_position_on_the_genomic_accession=95086094, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422843, CurrentID=0, Status='live', Symbol='MIR4293', Aliases='', description='microRNA 4293', other_designations='hsa-mir-4293', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14383200, end_position_on_the_genomic_accession=14383277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422844, CurrentID=0, Status='live', Symbol='MIR3118-3', Aliases='', description='microRNA 3118-3', other_designations='hsa-mir-3118-3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21406385, end_position_on_the_genomic_accession=21406459, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422845, CurrentID=0, Status='live', Symbol='MIR1233-2', Aliases='', description='microRNA 1233-2', other_designations='hsa-mir-1233-2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34528290, end_position_on_the_genomic_accession=34528371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422846, CurrentID=0, Status='live', Symbol='MIR3164', Aliases='', description='microRNA 3164', other_designations='hsa-mir-3164', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69083176, end_position_on_the_genomic_accession=69083258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422847, CurrentID=0, Status='live', Symbol='MIR514B', Aliases='mir-514b', description='microRNA 514b', other_designations='hsa-mir-514b', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147250151, end_position_on_the_genomic_accession=147250230, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422848, CurrentID=0, Status='live', Symbol='MIR4283-2', Aliases='', description='microRNA 4283-2', other_designations='hsa-mir-4283-2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63621090, end_position_on_the_genomic_accession=63621169, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422849, CurrentID=0, Status='live', Symbol='MIR548T', Aliases='', description='microRNA 548t', other_designations='hsa-mir-548t', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173268160, end_position_on_the_genomic_accession=173268233, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422850, CurrentID=0, Status='live', Symbol='MIR548V', Aliases='', description='microRNA 548v', other_designations='hsa-mir-548v', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17681578, end_position_on_the_genomic_accession=17681657, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422851, CurrentID=0, Status='live', Symbol='MIR4316', Aliases='', description='microRNA 4316', other_designations='hsa-mir-4316', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77396984, end_position_on_the_genomic_accession=77397054, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422852, CurrentID=0, Status='live', Symbol='MIR4259', Aliases='', description='microRNA 4259', other_designations='hsa-mir-4259', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159899979, end_position_on_the_genomic_accession=159900079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422853, CurrentID=0, Status='live', Symbol='MIR3182', Aliases='mir-3182', description='microRNA 3182', other_designations='hsa-mir-3182', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83508346, end_position_on_the_genomic_accession=83508408, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  414918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414918, CurrentID=0, Status='live', Symbol='DENND6B', Aliases='AFI1B, FAM116B', description='DENN/MADD domain containing 6B', other_designations='DENN domain-containing protein 6B|family with sequence similarity 116, member B|protein FAM116B', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50311963, end_position_on_the_genomic_accession=50327089, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  283847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283847, CurrentID=0, Status='live', Symbol='CCDC79', Aliases='', description='coiled-coil domain containing 79', other_designations='coiled-coil domain-containing protein 79', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66754411, end_position_on_the_genomic_accession=66801620, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  283848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283848, CurrentID=0, Status='live', Symbol='CES4A', Aliases='CES6, CES8', description='carboxylesterase 4A', other_designations='carboxylesterase 6|carboxylesterase 8 (putative)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66988589, end_position_on_the_genomic_accession=67009758, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  283849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283849, CurrentID=0, Status='live', Symbol='EXOC3L1', Aliases='EXOC3L', description='exocyst complex component 3-like 1', other_designations='protein Jiangli', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67184379, end_position_on_the_genomic_accession=67190204, orientation='minus', exon_count=14, OMIM=614117, no_hdr0=''),
  100422858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422858, CurrentID=0, Status='live', Symbol='MIR4285', Aliases='', description='microRNA 4285', other_designations='hsa-mir-4285', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102293103, end_position_on_the_genomic_accession=102293187, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422859, CurrentID=0, Status='live', Symbol='MIR3136', Aliases='mir-3136', description='microRNA 3136', other_designations='hsa-mir-3136', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69048958, end_position_on_the_genomic_accession=69049035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422860, CurrentID=0, Status='live', Symbol='MIR4292', Aliases='', description='microRNA 4292', other_designations='hsa-mir-4292', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136830957, end_position_on_the_genomic_accession=136831023, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422861, CurrentID=0, Status='live', Symbol='MIR4306', Aliases='mir-4306', description='microRNA 4306', other_designations='hsa-mir-4306', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99643059, end_position_on_the_genomic_accession=99643149, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283854, CurrentID=0, Status='live', Symbol='CASC22', Aliases='LINC01373', description='cancer susceptibility candidate 22 (non-protein coding)', other_designations='LincRNA-ENST00000515084|TCONS_00024290|long intergenic non-protein coding RNA 1373', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52271423, end_position_on_the_genomic_accession=52280299, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422863, CurrentID=0, Status='live', Symbol='MIR4265', Aliases='', description='microRNA 4265', other_designations='hsa-mir-4265', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109141490, end_position_on_the_genomic_accession=109141588, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422864, CurrentID=0, Status='live', Symbol='MIR544B', Aliases='', description='microRNA 544b', other_designations='hsa-mir-544b', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124732439, end_position_on_the_genomic_accession=124732516, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422865, CurrentID=0, Status='live', Symbol='MIR4320', Aliases='', description='microRNA 4320', other_designations='hsa-mir-4320', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50126499, end_position_on_the_genomic_accession=50126563, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422866, CurrentID=0, Status='live', Symbol='MIR3115', Aliases='mir-3115', description='microRNA 3115', other_designations='hsa-mir-3115', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23044305, end_position_on_the_genomic_accession=23044372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422867, CurrentID=0, Status='live', Symbol='MIR378C', Aliases='mir-378c', description='microRNA 378c', other_designations='hsa-mir-378c', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130962588, end_position_on_the_genomic_accession=130962668, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422868, CurrentID=0, Status='live', Symbol='MIR4270', Aliases='', description='microRNA 4270', other_designations='hsa-mir-4270', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15496239, end_position_on_the_genomic_accession=15496308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  152789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152789, CurrentID=0, Status='live', Symbol='JAKMIP1', Aliases='Gababrbp, JAMIP1, MARLIN1', description='janus kinase and microtubule interacting protein 1', other_designations='GABA-B receptor-binding protein|Jak and microtubule interacting protein 1|marlin-1|multiple alpha-helices and RNA-linker protein 1|multiple coiled-coil GABABR1-binding protein', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6026199, end_position_on_the_genomic_accession=6200591, orientation='minus', exon_count=22, OMIM=611195, no_hdr0=''),
  100422870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422870, CurrentID=0, Status='live', Symbol='MIR3180-1', Aliases='', description='microRNA 3180-1', other_designations='hsa-mir-3180-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14911220, end_position_on_the_genomic_accession=14911313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422871, CurrentID=0, Status='live', Symbol='MIR3117', Aliases='mir-3117', description='microRNA 3117', other_designations='hsa-mir-3117', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66628440, end_position_on_the_genomic_accession=66628517, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422872, CurrentID=0, Status='live', Symbol='MIR1244-3', Aliases='', description='microRNA 1244-3', other_designations='hsa-mir-1244-3', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9239467, end_position_on_the_genomic_accession=9239551, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422873, CurrentID=0, Status='live', Symbol='MIR4297', Aliases='', description='microRNA 4297', other_designations='hsa-mir-4297', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129843299, end_position_on_the_genomic_accession=129843374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422874, CurrentID=0, Status='live', Symbol='MIR4279', Aliases='', description='microRNA 4279', other_designations='hsa-mir-4279', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31936102, end_position_on_the_genomic_accession=31936159, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422875, CurrentID=0, Status='live', Symbol='MIR3192', Aliases='mir-3192', description='microRNA 3192', other_designations='hsa-mir-3192', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18470615, end_position_on_the_genomic_accession=18470691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422876, CurrentID=0, Status='live', Symbol='MIR3148', Aliases='mir-3148', description='microRNA 3148', other_designations='hsa-mir-3148', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29957272, end_position_on_the_genomic_accession=29957348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283869, CurrentID=0, Status='live', Symbol='NPW', Aliases='L8, L8C, PPL8, PPNPW', description='neuropeptide W', other_designations='hPPL8|prepro-neuropeptide W|preproprotein L8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2019520, end_position_on_the_genomic_accession=2020755, orientation='plus', exon_count=2, OMIM=607997, no_hdr0=''),
  283870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283870, CurrentID=0, Status='live', Symbol='BRICD5', Aliases='C16orf79', description='BRICHOS domain containing 5', other_designations='BRICHOS domain-containing protein C16orf79', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2209253, end_position_on_the_genomic_accession=2213540, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  283871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283871, CurrentID=0, Status='live', Symbol='PGP', Aliases='PGPase', description='phosphoglycolate phosphatase', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2211602, end_position_on_the_genomic_accession=2214821, orientation='minus', exon_count=3, OMIM=172280, no_hdr0=''),
  100422880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422880, CurrentID=0, Status='live', Symbol='MIR3162', Aliases='mir-3162', description='microRNA 3162', other_designations='hsa-mir-3162', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59595077, end_position_on_the_genomic_accession=59595158, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422881, CurrentID=0, Status='live', Symbol='MIR3170', Aliases='mir-3170', description='microRNA 3170', other_designations='hsa-mir-3170', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98208524, end_position_on_the_genomic_accession=98208600, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422882, CurrentID=0, Status='live', Symbol='MIR3120', Aliases='mir-3120', description='microRNA 3120', other_designations='hsa-mir-3120', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172138808, end_position_on_the_genomic_accession=172138888, orientation='plus', exon_count=1, OMIM=614722, no_hdr0=''),
  100422883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422883, CurrentID=0, Status='live', Symbol='MIR4325', Aliases='', description='microRNA 4325', other_designations='hsa-mir-4325', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57321502, end_position_on_the_genomic_accession=57321591, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283876, CurrentID=0, Status='live', Symbol='LINC00921', Aliases='', description='long intergenic non-protein coding RNA 921', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3263768, end_position_on_the_genomic_accession=3267567, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422885, CurrentID=0, Status='live', Symbol='MIR1244-2', Aliases='', description='microRNA 1244-2', other_designations='hsa-mir-1244-2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118974586, end_position_on_the_genomic_accession=118974670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422886, CurrentID=0, Status='live', Symbol='MIR3179-2', Aliases='', description='microRNA 3179-2', other_designations='hsa-mir-3179-2', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16300159, end_position_on_the_genomic_accession=16300242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422887, CurrentID=0, Status='live', Symbol='MIR4280', Aliases='', description='microRNA 4280', other_designations='hsa-mir-4280', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87114879, end_position_on_the_genomic_accession=87114954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422888, CurrentID=0, Status='live', Symbol='MIR4264', Aliases='', description='microRNA 4264', other_designations='hsa-mir-4264', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79649294, end_position_on_the_genomic_accession=79649359, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422889, CurrentID=0, Status='live', Symbol='MIR3194', Aliases='mir-3194', description='microRNA 3194', other_designations='hsa-mir-3194', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51452905, end_position_on_the_genomic_accession=51452977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422891, CurrentID=0, Status='live', Symbol='MIR4327', Aliases='', description='microRNA 4327', other_designations='hsa-mir-4327', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30375294, end_position_on_the_genomic_accession=30375378, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422892, CurrentID=0, Status='live', Symbol='MIR3157', Aliases='mir-3157', description='microRNA 3157', other_designations='hsa-mir-3157', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96064315, end_position_on_the_genomic_accession=96064399, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422893, CurrentID=0, Status='live', Symbol='MIR3154', Aliases='mir-3154', description='microRNA 3154', other_designations='hsa-mir-3154', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128244947, end_position_on_the_genomic_accession=128245030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422894, CurrentID=0, Status='live', Symbol='MIR4260', Aliases='', description='microRNA 4260', other_designations='hsa-mir-4260', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209623444, end_position_on_the_genomic_accession=209623510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  152815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152815, CurrentID=0, Status='live', Symbol='THAP6', Aliases='', description='THAP domain containing 6', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75509860, end_position_on_the_genomic_accession=75544102, orientation='plus', exon_count=8, OMIM=612535, no_hdr0=''),
  152816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152816, CurrentID=0, Status='live', Symbol='C4orf26', Aliases='AI2A4', description='chromosome 4 open reading frame 26', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75555510, end_position_on_the_genomic_accession=75565893, orientation='plus', exon_count=5, OMIM=614829, no_hdr0=''),
  100422897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422897, CurrentID=0, Status='live', Symbol='MIR4302', Aliases='', description='microRNA 4302', other_designations='hsa-mir-4302', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25874020, end_position_on_the_genomic_accession=25874079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422898, CurrentID=0, Status='live', Symbol='MIR4255', Aliases='', description='microRNA 4255', other_designations='hsa-mir-4255', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37161563, end_position_on_the_genomic_accession=37161634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422899, CurrentID=0, Status='live', Symbol='MIR3190', Aliases='mir-3190', description='microRNA 3190', other_designations='hsa-mir-3190', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47226942, end_position_on_the_genomic_accession=47227021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422900, CurrentID=0, Status='live', Symbol='MIR3158-1', Aliases='mir-3158-1', description='microRNA 3158-1', other_designations='hsa-mir-3158-1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101601417, end_position_on_the_genomic_accession=101601497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422901, CurrentID=0, Status='live', Symbol='MIR3135A', Aliases='MIR3135, hsa-mir-3135a', description='microRNA 3135a', other_designations='hsa-mir-3135|microRNA 3135', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20137565, end_position_on_the_genomic_accession=20137641, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422902, CurrentID=0, Status='live', Symbol='MIR3116-1', Aliases='', description='microRNA 3116-1', other_designations='hsa-mir-3116-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62078786, end_position_on_the_genomic_accession=62078859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422903, CurrentID=0, Status='live', Symbol='MIR4288', Aliases='', description='microRNA 4288', other_designations='hsa-mir-4288', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28505116, end_position_on_the_genomic_accession=28505182, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422904, CurrentID=0, Status='live', Symbol='MIR3193', Aliases='', description='microRNA 3193', other_designations='hsa-mir-3193', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31607186, end_position_on_the_genomic_accession=31607240, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283897, CurrentID=0, Status='live', Symbol='C16orf54', Aliases='', description='chromosome 16 open reading frame 54', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29742463, end_position_on_the_genomic_accession=29746329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283899, CurrentID=0, Status='live', Symbol='INO80E', Aliases='CCDC95', description='INO80 complex subunit E', other_designations='coiled-coil domain containing 95|coiled-coil domain-containing protein 95', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29995690, end_position_on_the_genomic_accession=30005790, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100422908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422908, CurrentID=0, Status='live', Symbol='MIR3129', Aliases='mir-3129', description='microRNA 3129', other_designations='hsa-mir-3129', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189133036, end_position_on_the_genomic_accession=189133111, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422909, CurrentID=0, Status='live', Symbol='MIR4295', Aliases='', description='microRNA 4295', other_designations='hsa-mir-4295', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112634170, end_position_on_the_genomic_accession=112634254, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283902, CurrentID=0, Status='live', Symbol='HCCAT5', Aliases='FJ222407, HTA', description='hepatocellular carcinoma associated transcript 5 (non-protein coding)', other_designations='hepatoma associated protein', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73092349, end_position_on_the_genomic_accession=73093774, orientation='plus', exon_count=3, OMIM=615613, no_hdr0=''),
  152831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152831, CurrentID=0, Status='live', Symbol='KLB', Aliases='BKL', description='klotho beta', other_designations='betaKlotho|klotho beta like|klotho beta-like protein', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39406853, end_position_on_the_genomic_accession=39451533, orientation='plus', exon_count=5, OMIM=611135, no_hdr0=''),
  100422912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422912, CurrentID=0, Status='live', Symbol='MIR3200', Aliases='mir-3200', description='microRNA 3200', other_designations='hsa-mir-3200', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30731557, end_position_on_the_genomic_accession=30731641, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422913, CurrentID=0, Status='live', Symbol='MIR320E', Aliases='mir-320e', description='microRNA 320e', other_designations='hsa-mir-320e', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46709293, end_position_on_the_genomic_accession=46709345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422914, CurrentID=0, Status='live', Symbol='MIR4253', Aliases='', description='microRNA 4253', other_designations='hsa-mir-4253', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22863159, end_position_on_the_genomic_accession=22863226, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422915, CurrentID=0, Status='live', Symbol='MIR3065', Aliases='mir-3065', description='microRNA 3065', other_designations='hsa-mir-3065', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81125877, end_position_on_the_genomic_accession=81125955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422916, CurrentID=0, Status='live', Symbol='MIR3201', Aliases='', description='microRNA 3201', other_designations='hsa-mir-3201', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48274364, end_position_on_the_genomic_accession=48274415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422917, CurrentID=0, Status='live', Symbol='MIR4283-1', Aliases='', description='microRNA 4283-1', other_designations='hsa-mir-4283-1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56955785, end_position_on_the_genomic_accession=56955864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422918, CurrentID=0, Status='live', Symbol='MIR3167', Aliases='', description='microRNA 3167', other_designations='hsa-mir-3167', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126988458, end_position_on_the_genomic_accession=126988542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422919, CurrentID=0, Status='live', Symbol='MIR1302-11', Aliases='', description='microRNA 1302-11', other_designations='hsa-mir-1302-11', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=71973, end_position_on_the_genomic_accession=72110, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422920, CurrentID=0, Status='live', Symbol='MIR548X', Aliases='', description='microRNA 548x', other_designations='hsa-mir-548x', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=18686090, end_position_on_the_genomic_accession=18686164, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422921, CurrentID=0, Status='live', Symbol='MIR3149', Aliases='', description='microRNA 3149', other_designations='hsa-mir-3149', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76966768, end_position_on_the_genomic_accession=76966850, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422922, CurrentID=0, Status='live', Symbol='MIR1972-2', Aliases='', description='microRNA 1972-2', other_designations='hsa-mir-1972-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70030346, end_position_on_the_genomic_accession=70030422, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422923, CurrentID=0, Status='live', Symbol='MIR548W', Aliases='', description='microRNA 548w', other_designations='hsa-mir-548w', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26025237, end_position_on_the_genomic_accession=26025310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422924, CurrentID=0, Status='live', Symbol='MIR4303', Aliases='', description='microRNA 4303', other_designations='hsa-mir-4303', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97995383, end_position_on_the_genomic_accession=97995448, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422925, CurrentID=0, Status='live', Symbol='MIR4322', Aliases='', description='microRNA 4322', other_designations='hsa-mir-4322', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10230413, end_position_on_the_genomic_accession=10230485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422926, CurrentID=0, Status='live', Symbol='MIR3137', Aliases='', description='microRNA 3137', other_designations='hsa-mir-3137', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195134506, end_position_on_the_genomic_accession=195134580, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422927, CurrentID=0, Status='live', Symbol='MIR4291', Aliases='', description='microRNA 4291', other_designations='hsa-mir-4291', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93819357, end_position_on_the_genomic_accession=93819421, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422928, CurrentID=0, Status='live', Symbol='MIR3127', Aliases='mir-3127', description='microRNA 3127', other_designations='hsa-mir-3127', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96798278, end_position_on_the_genomic_accession=96798353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422929, CurrentID=0, Status='live', Symbol='MIR4261', Aliases='', description='microRNA 4261', other_designations='hsa-mir-4261', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10192614, end_position_on_the_genomic_accession=10192671, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422930, CurrentID=0, Status='live', Symbol='MIR4330', Aliases='', description='microRNA 4330', other_designations='hsa-mir-4330', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151168222, end_position_on_the_genomic_accession=151168326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422931, CurrentID=0, Status='live', Symbol='MIR4304', Aliases='', description='microRNA 4304', other_designations='hsa-mir-4304', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123010667, end_position_on_the_genomic_accession=123010728, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422932, CurrentID=0, Status='live', Symbol='MIR4328', Aliases='', description='microRNA 4328', other_designations='hsa-mir-4328', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78901194, end_position_on_the_genomic_accession=78901249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422933, CurrentID=0, Status='live', Symbol='MIR378B', Aliases='', description='microRNA 378b', other_designations='hsa-mir-378b', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10330229, end_position_on_the_genomic_accession=10330285, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422934, CurrentID=0, Status='live', Symbol='MIR3143', Aliases='mir-3143', description='microRNA 3143', other_designations='hsa-mir-3143', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27147626, end_position_on_the_genomic_accession=27147688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283927, CurrentID=0, Status='live', Symbol='NUDT7', Aliases='', description='nudix hydrolase 7', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 7', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77722492, end_position_on_the_genomic_accession=77742260, orientation='plus', exon_count=7, OMIM=609231, no_hdr0=''),
  100422936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422936, CurrentID=0, Status='live', Symbol='MIR3153', Aliases='mir-3153', description='microRNA 3153', other_designations='hsa-mir-3153', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89312225, end_position_on_the_genomic_accession=89312306, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422937, CurrentID=0, Status='live', Symbol='MIR4275', Aliases='', description='microRNA 4275', other_designations='hsa-mir-4275', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28819582, end_position_on_the_genomic_accession=28819668, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422938, CurrentID=0, Status='live', Symbol='MIR3142', Aliases='', description='microRNA 3142', other_designations='hsa-mir-3142', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160474444, end_position_on_the_genomic_accession=160474525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422939, CurrentID=0, Status='live', Symbol='MIR3147', Aliases='mir-3147', description='microRNA 3147', other_designations='hsa-mir-3147', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57405025, end_position_on_the_genomic_accession=57405090, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422940, CurrentID=0, Status='live', Symbol='MIR4305', Aliases='', description='microRNA 4305', other_designations='hsa-mir-4305', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39664034, end_position_on_the_genomic_accession=39664135, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  768218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768218, CurrentID=0, Status='live', Symbol='MIR766', Aliases='MIRN766, hsa-mir-766, mir-766', description='microRNA 766', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119646738, end_position_on_the_genomic_accession=119646848, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422942, CurrentID=0, Status='live', Symbol='MIR3133', Aliases='', description='microRNA 3133', other_designations='hsa-mir-3133', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241477905, end_position_on_the_genomic_accession=241477982, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422943, CurrentID=0, Status='live', Symbol='MIR3189', Aliases='mir-3189', description='microRNA 3189', other_designations='hsa-mir-3189', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18386562, end_position_on_the_genomic_accession=18386634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422944, CurrentID=0, Status='live', Symbol='MIR3186', Aliases='mir-3186', description='microRNA 3186', other_designations='hsa-mir-3186', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81451104, end_position_on_the_genomic_accession=81451188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422945, CurrentID=0, Status='live', Symbol='MIR4326', Aliases='mir-4326', description='microRNA 4326', other_designations='hsa-mir-4326', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63286808, end_position_on_the_genomic_accession=63286866, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422946, CurrentID=0, Status='live', Symbol='MIR3116-2', Aliases='', description='microRNA 3116-2', other_designations='hsa-mir-3116-2', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62078789, end_position_on_the_genomic_accession=62078856, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422947, CurrentID=0, Status='live', Symbol='MIR3122', Aliases='mir-3122', description='microRNA 3122', other_designations='hsa-mir-3122', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212077613, end_position_on_the_genomic_accession=212077685, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422948, CurrentID=0, Status='live', Symbol='MIR4284', Aliases='mir-4284', description='microRNA 4284', other_designations='hsa-mir-4284', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73711317, end_position_on_the_genomic_accession=73711397, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422949, CurrentID=0, Status='live', Symbol='MIR3118-2', Aliases='', description='microRNA 3118-2', other_designations='hsa-mir-3118-2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20832795, end_position_on_the_genomic_accession=20832869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422950, CurrentID=0, Status='live', Symbol='MIR3141', Aliases='mir-3141', description='microRNA 3141', other_designations='hsa-mir-3141', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154596012, end_position_on_the_genomic_accession=154596072, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422951, CurrentID=0, Status='live', Symbol='MIR3144', Aliases='mir-3144', description='microRNA 3144', other_designations='hsa-mir-3144', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=120015179, end_position_on_the_genomic_accession=120015257, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422952, CurrentID=0, Status='live', Symbol='MIR4271', Aliases='', description='microRNA 4271', other_designations='hsa-mir-4271', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49274120, end_position_on_the_genomic_accession=49274186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422953, CurrentID=0, Status='live', Symbol='MIR3165', Aliases='mir-3165', description='microRNA 3165', other_designations='hsa-mir-3165', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72072228, end_position_on_the_genomic_accession=72072302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422954, CurrentID=0, Status='live', Symbol='MIR4309', Aliases='', description='microRNA 4309', other_designations='hsa-mir-4309', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102539644, end_position_on_the_genomic_accession=102539726, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422955, CurrentID=0, Status='live', Symbol='MIR4273', Aliases='', description='microRNA 4273', other_designations='hsa-mir-4273', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75738280, end_position_on_the_genomic_accession=75738363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283948, CurrentID=0, Status='live', Symbol='NHLRC4', Aliases='', description='NHL repeat containing 4', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=566995, end_position_on_the_genomic_accession=569495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  152877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152877, CurrentID=0, Status='live', Symbol='FAM53A', Aliases='DNTNP', description='family with sequence similarity 53 member A', other_designations='dorsal neural-tube nuclear protein|family with sequence similarity 53, member A', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1574062, end_position_on_the_genomic_accession=1684313, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  283951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283951, CurrentID=0, Status='live', Symbol='C16orf91', Aliases='CCSMST1, URLC5, gs103', description='chromosome 16 open reading frame 91', other_designations='cattle cerebrum and skeletal muscle-specific protein 1 family member|up-regulated in lung cancer 5', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1419744, end_position_on_the_genomic_accession=1420800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422960, CurrentID=0, Status='live', Symbol='MIR3179-1', Aliases='', description='microRNA 3179-1', other_designations='hsa-mir-3179-1', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14901508, end_position_on_the_genomic_accession=14901591, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283953, CurrentID=0, Status='live', Symbol='TMEM114', Aliases='', description='transmembrane protein 114', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8526549, end_position_on_the_genomic_accession=8590193, orientation='minus', exon_count=8, OMIM=611579, no_hdr0=''),
  100422962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422962, CurrentID=0, Status='live', Symbol='MIR4281', Aliases='', description='microRNA 4281', other_designations='hsa-mir-4281', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176629439, end_position_on_the_genomic_accession=176629500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283955, CurrentID=0, Status='live', Symbol='PKD1P2', Aliases='HG2', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 2', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16356417, end_position_on_the_genomic_accession=16377507, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422964, CurrentID=0, Status='live', Symbol='MIR3150A', Aliases='MIR3150, mir-3150a', description='microRNA 3150a', other_designations='hsa-mir-3150|hsa-mir-3150a|microRNA 3150', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95072914, end_position_on_the_genomic_accession=95072993, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422965, CurrentID=0, Status='live', Symbol='MIR4263', Aliases='', description='microRNA 4263', other_designations='hsa-mir-4263', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27996367, end_position_on_the_genomic_accession=27996449, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422966, CurrentID=0, Status='live', Symbol='MIR4277', Aliases='', description='microRNA 4277', other_designations='hsa-mir-4277', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1708785, end_position_on_the_genomic_accession=1708868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422967, CurrentID=0, Status='live', Symbol='MIR3146', Aliases='mir-3146', description='microRNA 3146', other_designations='hsa-mir-3146', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19705358, end_position_on_the_genomic_accession=19705436, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422968, CurrentID=0, Status='live', Symbol='MIR4251', Aliases='', description='microRNA 4251', other_designations='hsa-mir-4251', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3127975, end_position_on_the_genomic_accession=3128035, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422969, CurrentID=0, Status='live', Symbol='MIR2909', Aliases='', description='microRNA 2909', other_designations='hsa-mir-2909', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37033745, end_position_on_the_genomic_accession=37033813, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422970, CurrentID=0, Status='live', Symbol='MIR1273D', Aliases='', description='microRNA 1273d', other_designations='hsa-mir-1273d', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10227718, end_position_on_the_genomic_accession=10227803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422971, CurrentID=0, Status='live', Symbol='MIR4312', Aliases='', description='microRNA 4312', other_designations='hsa-mir-4312', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68801850, end_position_on_the_genomic_accession=68801925, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422972, CurrentID=0, Status='live', Symbol='MIR3181', Aliases='', description='microRNA 3181', other_designations='hsa-mir-3181', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50742305, end_position_on_the_genomic_accession=50742377, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104355133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355133, CurrentID=0, Status='live', Symbol='ABHD15-AS1', Aliases='linc-TP53I13, lnc-TP53I13', description='ABHD15 antisense RNA 1', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422974, CurrentID=0, Status='live', Symbol='MIR3178', Aliases='mir-3178', description='microRNA 3178', other_designations='hsa-mir-3178', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2531922, end_position_on_the_genomic_accession=2532005, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422975, CurrentID=0, Status='live', Symbol='MIR4252', Aliases='', description='microRNA 4252', other_designations='hsa-mir-4252', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6429834, end_position_on_the_genomic_accession=6429896, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422976, CurrentID=0, Status='live', Symbol='MIR4256', Aliases='', description='microRNA 4256', other_designations='hsa-mir-4256', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112461770, end_position_on_the_genomic_accession=112461833, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422977, CurrentID=0, Status='live', Symbol='MIR1184-3', Aliases='', description='microRNA 1184-3', other_designations='hsa-mir-1184-3', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155457517, end_position_on_the_genomic_accession=155457615, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283970, CurrentID=0, Status='live', Symbol='PDXDC2P', Aliases='NPIP, PDXDC2', description='pyridoxal-dependent decarboxylase domain containing 2, pseudogene', other_designations='nuclear pore complex-interacting protein', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69976297, end_position_on_the_genomic_accession=70065948, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  283971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283971, CurrentID=0, Status='live', Symbol='CLEC18C', Aliases='MRCL, MRCL3', description='C-type lectin domain family 18 member C', other_designations='C-type lectin domain family 18, member C|mannose receptor-like 3|mannose receptor-like protein 3', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70173322, end_position_on_the_genomic_accession=70187147, orientation='plus', exon_count=14, OMIM=616573, no_hdr0=''),
  100422980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422980, CurrentID=0, Status='live', Symbol='MIR4323', Aliases='', description='microRNA 4323', other_designations='hsa-mir-4323', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42133445, end_position_on_the_genomic_accession=42133513, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422981, CurrentID=0, Status='live', Symbol='MIR3173', Aliases='mir-3173', description='microRNA 3173', other_designations='hsa-mir-3173', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95137919, end_position_on_the_genomic_accession=95137986, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422982, CurrentID=0, Status='live', Symbol='MIR4286', Aliases='mir-4286', description='microRNA 4286', other_designations='hsa-mir-4286', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10666978, end_position_on_the_genomic_accession=10667070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422983, CurrentID=0, Status='live', Symbol='MIR4314', Aliases='', description='microRNA 4314', other_designations='hsa-mir-4314', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8088056, end_position_on_the_genomic_accession=8088147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422984, CurrentID=0, Status='live', Symbol='MIR4308', Aliases='', description='microRNA 4308', other_designations='hsa-mir-4308', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54878113, end_position_on_the_genomic_accession=54878193, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  152905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152905, CurrentID=0, Status='live', Symbol='RTN3P1', Aliases='NSPLII, RTN3PS', description='reticulon 3 pseudogene 1', other_designations='2NSP-like protein II', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145375380, end_position_on_the_genomic_accession=145377977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422986, CurrentID=0, Status='live', Symbol='MIR3125', Aliases='mir-3125', description='microRNA 3125', other_designations='hsa-mir-3125', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12737367, end_position_on_the_genomic_accession=12737444, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422987, CurrentID=0, Status='live', Symbol='MIR3202-1', Aliases='mir-3202-1', description='microRNA 3202-1', other_designations='hsa-mir-3202-1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153981097, end_position_on_the_genomic_accession=153981177, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422988, CurrentID=0, Status='live', Symbol='MIR3156-1', Aliases='mir-3156-1', description='microRNA 3156-1', other_designations='hsa-mir-3156-1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45164014, end_position_on_the_genomic_accession=45164088, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422989, CurrentID=0, Status='live', Symbol='MIR3155A', Aliases='MIR3155, hsa-mir-3155a, mir-3155a', description='microRNA 3155a', other_designations='hsa-mir-3155|microRNA 3155', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6152196, end_position_on_the_genomic_accession=6152277, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422990, CurrentID=0, Status='live', Symbol='MIR3134', Aliases='', description='microRNA 3134', other_designations='hsa-mir-3134', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15697298, end_position_on_the_genomic_accession=15697371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422991, CurrentID=0, Status='live', Symbol='MIR1260B', Aliases='mir-1260b', description='microRNA 1260b', other_designations='hsa-mir-1260b', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96341438, end_position_on_the_genomic_accession=96341526, orientation='plus', exon_count=1, OMIM=615372, no_hdr0=''),
  100422992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422992, CurrentID=0, Status='live', Symbol='MIR3151', Aliases='mir-3151', description='microRNA 3151', other_designations='hsa-mir-3151', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103154614, end_position_on_the_genomic_accession=103154689, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283985, CurrentID=0, Status='live', Symbol='FADS6', Aliases='FP18279', description='fatty acid desaturase 6', other_designations='fatty acid desaturase domain family, member 6', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74876776, end_position_on_the_genomic_accession=74893811, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100422994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422994, CurrentID=0, Status='live', Symbol='MIR4267', Aliases='', description='microRNA 4267', other_designations='hsa-mir-4267', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110069961, end_position_on_the_genomic_accession=110070042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283987, CurrentID=0, Status='live', Symbol='HID1', Aliases='C17orf28, DMC1, HID-1', description='HID1 domain containing', other_designations='HID1 domain-containing protein|UPF0663 transmembrane protein C17orf28|down-regulated in multiple cancers 1|downregulated in multiple cancer 1|protein hid-1 homolog', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74950742, end_position_on_the_genomic_accession=74972805, orientation='minus', exon_count=20, OMIM=605752, no_hdr0=''),
  100422996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422996, CurrentID=0, Status='live', Symbol='MIR4262', Aliases='', description='microRNA 4262', other_designations='hsa-mir-4262', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11836933, end_position_on_the_genomic_accession=11836986, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  283989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283989, CurrentID=0, Status='live', Symbol='TSEN54', Aliases='PCH2A, PCH4, PCH5, SEN54L, sen54', description='tRNA splicing endonuclease subunit 54', other_designations='SEN54 homolog|TSEN54 tRNA splicing endonuclease subunit|hsSen54|tRNA-intron endonuclease Sen54', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75515707, end_position_on_the_genomic_accession=75524739, orientation='plus', exon_count=12, OMIM=608755, no_hdr0=''),
  100422998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422998, CurrentID=0, Status='live', Symbol='MIR3199-2', Aliases='mir-3199-2', description='microRNA 3199-2', other_designations='hsa-mir-3199-2', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27920526, end_position_on_the_genomic_accession=27920611, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283991, CurrentID=0, Status='live', Symbol='UBALD2', Aliases='FAM100B', description='UBA-like domain containing 2', other_designations='family with sequence similarity 100, member B|protein FAM100B', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76265205, end_position_on_the_genomic_accession=76271298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100423000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423000, CurrentID=0, Status='live', Symbol='MIR3161', Aliases='', description='microRNA 3161', other_designations='hsa-mir-3161', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48096782, end_position_on_the_genomic_accession=48096858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423001, CurrentID=0, Status='live', Symbol='MIR3145', Aliases='mir-3145', description='microRNA 3145', other_designations='hsa-mir-3145', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138435213, end_position_on_the_genomic_accession=138435294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423002, CurrentID=0, Status='live', Symbol='MIR3130-2', Aliases='MIR3130-4, mir-3130-2', description='microRNA 3130-2', other_designations='hsa-mir-3130-2|hsa-mir-3130-4|microRNA 3130-4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206783234, end_position_on_the_genomic_accession=206783308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423003, CurrentID=0, Status='live', Symbol='MIR3184', Aliases='', description='microRNA 3184', other_designations='hsa-mir-3184', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30117086, end_position_on_the_genomic_accession=30117160, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423004, CurrentID=0, Status='live', Symbol='MIR4315-1', Aliases='', description='microRNA 4315-1', other_designations='hsa-mir-4315-1', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45475363, end_position_on_the_genomic_accession=45475435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423005, CurrentID=0, Status='live', Symbol='MIR4282', Aliases='', description='microRNA 4282', other_designations='hsa-mir-4282', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72967687, end_position_on_the_genomic_accession=72967753, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  152926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152926, CurrentID=0, Status='live', Symbol='PPM1K', Aliases='BDP, MSUDMV, PP2Ckappa, PP2Cm, PTMP, UG0882E07', description='protein phosphatase, Mg2+/Mn2+ dependent 1K', other_designations='PP2C domain-containing protein phosphatase 1K|PP2C-kappa|PP2C-type mitochondrial phosphoprotein phosphatase|branched-chain alpha-ketoacid dehydrogenase phosphatase|protein phosphatase 2C kappa|protein phosphatase, Mg2+/Mn2+ dependent, 1K', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88257609, end_position_on_the_genomic_accession=88284831, orientation='minus', exon_count=11, OMIM=611065, no_hdr0=''),
  283999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283999, CurrentID=0, Status='live', Symbol='TMEM235', Aliases='ARGM1', description='transmembrane protein 235', other_designations='transmembrane protein ENSP00000364084', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78231310, end_position_on_the_genomic_accession=78240987, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100423008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423008, CurrentID=0, Status='live', Symbol='MIR3118-1', Aliases='mir-3118-1', description='microRNA 3118-1', other_designations='hsa-mir-3118-1', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13644775, end_position_on_the_genomic_accession=13644850, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284001, CurrentID=0, Status='live', Symbol='CCDC57', Aliases='', description='coiled-coil domain containing 57', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82101470, end_position_on_the_genomic_accession=82213520, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  100423010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423010, CurrentID=0, Status='live', Symbol='MIR3119-2', Aliases='', description='microRNA 3119-2', other_designations='hsa-mir-3119-2', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170151378, end_position_on_the_genomic_accession=170151462, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423011, CurrentID=0, Status='live', Symbol='MIR3138', Aliases='mir-3138', description='microRNA 3138', other_designations='hsa-mir-3138', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10078611, end_position_on_the_genomic_accession=10078692, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284004, CurrentID=0, Status='live', Symbol='HEXDC', Aliases='', description='hexosaminidase (glycosyl hydrolase family 20, catalytic domain) containing', other_designations='N-acetyl-beta-galactosaminidase|beta-N-acetylhexosaminidase|beta-hexosaminidase D|hexosaminidase domain-containing protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82418376, end_position_on_the_genomic_accession=82442640, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100423013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423013, CurrentID=0, Status='live', Symbol='MIR4310', Aliases='', description='microRNA 4310', other_designations='hsa-mir-4310', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41866495, end_position_on_the_genomic_accession=41866551, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423014, CurrentID=0, Status='live', Symbol='MIR3196', Aliases='mir-3196', description='microRNA 3196', other_designations='hsa-mir-3196', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63238779, end_position_on_the_genomic_accession=63238842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423015, CurrentID=0, Status='live', Symbol='MIR4289', Aliases='', description='microRNA 4289', other_designations='hsa-mir-4289', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88745836, end_position_on_the_genomic_accession=88745905, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423016, CurrentID=0, Status='live', Symbol='MIR3159', Aliases='', description='microRNA 3159', other_designations='hsa-mir-3159', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18387787, end_position_on_the_genomic_accession=18387860, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423017, CurrentID=0, Status='live', Symbol='MIR3139', Aliases='', description='microRNA 3139', other_designations='hsa-mir-3139', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143343460, end_position_on_the_genomic_accession=143343535, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423018, CurrentID=0, Status='live', Symbol='MIR3156-3', Aliases='mir-3156-3', description='microRNA 3156-3', other_designations='hsa-mir-3156-3', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13406384, end_position_on_the_genomic_accession=13406460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423019, CurrentID=0, Status='live', Symbol='MIR4307', Aliases='', description='microRNA 4307', other_designations='hsa-mir-4307', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26908642, end_position_on_the_genomic_accession=26908725, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  152940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152940, CurrentID=0, Status='live', Symbol='C4orf45', Aliases='', description='chromosome 4 open reading frame 45', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158893532, end_position_on_the_genomic_accession=159082867, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  284013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284013, CurrentID=0, Status='live', Symbol='VMO1', Aliases='ERGA6350, PRO21055', description='vitelline membrane outer layer 1 homolog (chicken)', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4785285, end_position_on_the_genomic_accession=4786434, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284014, CurrentID=0, Status='live', Symbol='LOC284014', Aliases='', description='uncharacterized LOC284014', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060157, CurrentID=0, Status='live', Symbol='LOC101060157', Aliases='', description='uncharacterized LOC101060157', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18232812, end_position_on_the_genomic_accession=18237840, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100423025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423025, CurrentID=0, Status='live', Symbol='MIR3198-1', Aliases='MIR3198, mir-3198-1', description='microRNA 3198-1', other_designations='hsa-mir-3198|hsa-mir-3198-1|microRNA 3198', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17764180, end_position_on_the_genomic_accession=17764259, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284018, CurrentID=0, Status='live', Symbol='C17orf58', Aliases='', description='chromosome 17 open reading frame 58', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67991099, end_position_on_the_genomic_accession=67996784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100423027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423027, CurrentID=0, Status='live', Symbol='MIR4266', Aliases='', description='microRNA 4266', other_designations='hsa-mir-4266', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109313571, end_position_on_the_genomic_accession=109313625, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423028, CurrentID=0, Status='live', Symbol='MIR4254', Aliases='', description='microRNA 4254', other_designations='hsa-mir-4254', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31758660, end_position_on_the_genomic_accession=31758735, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284021, CurrentID=0, Status='live', Symbol='MILR1', Aliases='Allergin-1, C17orf60, MCA-32, MCA32', description='mast cell immunoglobulin-like receptor 1', other_designations='allergy inhibitory receptor 1|mast cell antigen 32|probable mast cell antigen 32 homolog', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64447249, end_position_on_the_genomic_accession=64479383, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100423030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423030, CurrentID=0, Status='live', Symbol='MIR3126', Aliases='mir-3126', description='microRNA 3126', other_designations='hsa-mir-3126', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69103682, end_position_on_the_genomic_accession=69103755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423031, CurrentID=0, Status='live', Symbol='MIR4321', Aliases='', description='microRNA 4321', other_designations='hsa-mir-4321', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2250639, end_position_on_the_genomic_accession=2250718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423032, CurrentID=0, Status='live', Symbol='MIR3121', Aliases='mir-3121', description='microRNA 3121', other_designations='hsa-mir-3121', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180438314, end_position_on_the_genomic_accession=180438390, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423033, CurrentID=0, Status='live', Symbol='MIR3158-2', Aliases='mir-3158-2', description='microRNA 3158-2', other_designations='hsa-mir-3158-2', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101601417, end_position_on_the_genomic_accession=101601497, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423034, CurrentID=0, Status='live', Symbol='MIR3199-1', Aliases='mir-3199-1', description='microRNA 3199-1', other_designations='hsa-mir-3199-1', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27920525, end_position_on_the_genomic_accession=27920612, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423035, CurrentID=0, Status='live', Symbol='MIR4313', Aliases='', description='microRNA 4313', other_designations='hsa-mir-4313', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75762215, end_position_on_the_genomic_accession=75762315, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423036, CurrentID=0, Status='live', Symbol='MIR2355', Aliases='mir-2355', description='microRNA 2355', other_designations='hsa-mir-2355', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207109987, end_position_on_the_genomic_accession=207110073, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284029, CurrentID=0, Status='live', Symbol='LINC00324', Aliases='C17orf44, NCRNA00324', description='long intergenic non-protein coding RNA 324', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8220630, end_position_on_the_genomic_accession=8224043, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100423038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423038, CurrentID=0, Status='live', Symbol='MIR466', Aliases='hsa-mir-466', description='microRNA 466', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31161704, end_position_on_the_genomic_accession=31161787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423039, CurrentID=0, Status='live', Symbol='MIR3132', Aliases='', description='microRNA 3132', other_designations='hsa-mir-3132', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219549073, end_position_on_the_genomic_accession=219549147, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423040, CurrentID=0, Status='live', Symbol='MIR3166', Aliases='', description='microRNA 3166', other_designations='hsa-mir-3166', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88176502, end_position_on_the_genomic_accession=88176593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423041, CurrentID=0, Status='live', Symbol='MIR4296', Aliases='', description='microRNA 4296', other_designations='hsa-mir-4296', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125032783, end_position_on_the_genomic_accession=125032870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100423042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423042, CurrentID=0, Status='live', Symbol='MIR4276', Aliases='', description='microRNA 4276', other_designations='hsa-mir-4276', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174423795, end_position_on_the_genomic_accession=174423864, orientation='plus', exon_count=1, OMIM=616274, no_hdr0=''),
  100423043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423043, CurrentID=0, Status='live', Symbol='MIR4269', Aliases='', description='microRNA 4269', other_designations='hsa-mir-4269', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239305462, end_position_on_the_genomic_accession=239305545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423044, CurrentID=0, Status='live', Symbol='LOC100423044', Aliases='', description='solute carrier family 34 (type II sodium/phosphate cotransporter), member 2 pseudogene', other_designations='sodium-dependent phosphate transport protein 2B pseudogene|solute carrier family 34 (sodium phosphate), member 2 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14053009, end_position_on_the_genomic_accession=14054039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271218, CurrentID=0, Status='live', Symbol='RPL7P36', Aliases='RPL7_18_1089', description='ribosomal protein L7 pseudogene 36', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99047623, end_position_on_the_genomic_accession=99048369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100423046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423046, CurrentID=0, Status='live', Symbol='STX8P1', Aliases='', description='syntaxin 8 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12771236, end_position_on_the_genomic_accession=12771701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284040, CurrentID=0, Status='live', Symbol='CDRT4', Aliases='', description='CMT1A duplicated region transcript 4', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15436015, end_position_on_the_genomic_accession=15467611, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  415116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=415116, CurrentID=0, Status='live', Symbol='PIM3', Aliases='pim-3', description='Pim-3 proto-oncogene, serine/threonine kinase', other_designations='pim-3 oncogene|serine/threonine kinase Pim-3', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49960495, end_position_on_the_genomic_accession=49964072, orientation='plus', exon_count=6, OMIM=610580, no_hdr0=''),
  415117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=415117, CurrentID=0, Status='live', Symbol='STX19', Aliases='', description='syntaxin 19', other_designations='', map_location='3q11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94014365, end_position_on_the_genomic_accession=94028776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284047, CurrentID=0, Status='live', Symbol='CCDC144B', Aliases='', description='coiled-coil domain containing 144B (pseudogene)', other_designations='coiled-coil domain-containing protein 144B', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18537800, end_position_on_the_genomic_accession=18625617, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  284048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284048, CurrentID=0, Status='live', Symbol='LOC284048', Aliases='', description='FLJ11800 related gene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104355217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355217, CurrentID=0, Status='live', Symbol='ERICD', Aliases='ERIC, LINC01130, TCONS_00014875', description='E2F1-regulated inhibitor of cell death (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1130', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104355218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355218, CurrentID=0, Status='live', Symbol='LINC01177', Aliases='', description='long intergenic non-protein coding RNA 1177', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9441294, end_position_on_the_genomic_accession=9444985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100142659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100142659, CurrentID=0, Status='live', Symbol='CTAGE8', Aliases='', description='CTAGE family member 8', other_designations='CTAGE family, member 4 pseudogene|CTAGE family, member 8|Cutaneous T-cell lymphoma-associated antigen 8|Protein cTAGE-8', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144266674, end_position_on_the_genomic_accession=144269784, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104355220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355220, CurrentID=0, Status='live', Symbol='LINC01219', Aliases='', description='long intergenic non-protein coding RNA 1219', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1991096, end_position_on_the_genomic_accession=1993469, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100423061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423061, CurrentID=0, Status='live', Symbol='ASNSP3', Aliases='', description='asparagine synthetase pseudogene 3', other_designations='', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100062297, end_position_on_the_genomic_accession=100063617, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100423062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423062, CurrentID=0, Status='live', Symbol='IGLL5', Aliases='IGL, IGLV, VL-MAR', description='immunoglobulin lambda-like polypeptide 5', other_designations='G lambda-1|anti HBs antibody light-chain Fab fragment|germline immunoglobulin lambda 1|immunoglobin anti-granzymeB light chain variable region|immunoglobulin lambda-3 surrogate light chain|omega light chain protein 14.1', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22887780, end_position_on_the_genomic_accession=22895834, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060164, CurrentID=0, Status='live', Symbol='LOC101060164', Aliases='', description='secretory carrier membrane protein 1 pseudogene', other_designations='secretory carrier-associated membrane protein 1-like', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96390651, end_position_on_the_genomic_accession=96391371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284058, CurrentID=0, Status='live', Symbol='KANSL1', Aliases='CENP-36, KDVS, KIAA1267, MSL1v1, NSL1, hMSL1v1', description='KAT8 regulatory NSL complex subunit 1', other_designations='MLL1/MLL complex subunit KANSL1|MSL1 homolog 1|NSL complex protein NSL1|centromere protein 36|male-specific lethal 1 homolog|non-specific lethal 1 homolog', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46029916, end_position_on_the_genomic_accession=46225374, orientation='minus', exon_count=20, OMIM=612452, no_hdr0=''),
  152992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152992, CurrentID=0, Status='live', Symbol='TRMT44', Aliases='C4orf23, METTL19, TRM44', description='tRNA methyltransferase 44 homolog (S. cerevisiae)', other_designations='methyltransferase like 19|methyltransferase-like protein 19', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8440778, end_position_on_the_genomic_accession=8516547, orientation='plus', exon_count=20, OMIM=614309, no_hdr0=''),
  284067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284067, CurrentID=0, Status='live', Symbol='C17orf105', Aliases='', description='chromosome 17 open reading frame 105', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43778391, end_position_on_the_genomic_accession=43784842, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284069, CurrentID=0, Status='live', Symbol='FAM171A2', Aliases='', description='family with sequence similarity 171 member A2', other_designations='family with sequence similarity 171, member A2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44353733, end_position_on_the_genomic_accession=44363867, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  284071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284071, CurrentID=0, Status='live', Symbol='MEIOC', Aliases='C17orf104', description='meiosis specific with coiled-coil domain', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44656394, end_position_on_the_genomic_accession=44678306, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  284072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284072, CurrentID=0, Status='live', Symbol='LOC284072', Aliases='', description='uncharacterized LOC284072', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=70587537, end_position_on_the_genomic_accession=70590215, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284076, CurrentID=0, Status='live', Symbol='TTLL6', Aliases='TTL.6', description='tubulin tyrosine ligase like 6', other_designations='tubulin tyrosine ligase-like family member 6|tubulin--tyrosine ligase-like protein 6', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48762231, end_position_on_the_genomic_accession=48817107, orientation='minus', exon_count=17, OMIM=610849, no_hdr0=''),
  284080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284080, CurrentID=0, Status='live', Symbol='LOC284080', Aliases='', description='uncharacterized LOC284080', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50050349, end_position_on_the_genomic_accession=50055739, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  134728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134728, CurrentID=0, Status='live', Symbol='IRAK1BP1', Aliases='AIP70, SIMPL', description='interleukin 1 receptor associated kinase 1 binding protein 1', other_designations='ActA binding protein 3|IRAK1-binding protein 1|interleukin-1 receptor-associated kinase 1 binding protein 1', map_location='6q14-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=78867085, end_position_on_the_genomic_accession=78936160, orientation='plus', exon_count=7, OMIM=615375, no_hdr0=''),
  284083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284083, CurrentID=0, Status='live', Symbol='C17orf47', Aliases='', description='chromosome 17 open reading frame 47', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58540681, end_position_on_the_genomic_accession=58544372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284085, CurrentID=0, Status='live', Symbol='KRT18P55', Aliases='', description='keratin 18 pseudogene 55', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28275986, end_position_on_the_genomic_accession=28307382, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284086, CurrentID=0, Status='live', Symbol='NEK8', Aliases='JCK, NEK12A, NPHP9, RHPD2', description='NIMA-related kinase 8', other_designations='NIMA (never in mitosis gene a)- related kinase 8|NIMA-family kinase NEK8|NIMA-related kinase 12a|nephrocystin 9|never in mitosis A-related kinase 8|nimA-related protein kinase 8|nima-related protein kinase 12a|serine/thrionine-protein kinase NEK8', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28725897, end_position_on_the_genomic_accession=28742766, orientation='plus', exon_count=19, OMIM=609799, no_hdr0=''),
  153020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153020, CurrentID=0, Status='live', Symbol='RASGEF1B', Aliases='GPIG4', description='RasGEF domain family member 1B', other_designations='GPI gamma-4|GPI-gamma 4|RasGEF domain family, member 1B', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81426393, end_position_on_the_genomic_accession=81471928, orientation='minus', exon_count=14, OMIM=614532, no_hdr0=''),
  101060170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060170, CurrentID=0, Status='live', Symbol='LOC101060170', Aliases='', description='uncharacterized LOC101060170', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145282607, end_position_on_the_genomic_accession=145289041, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284098, CurrentID=0, Status='live', Symbol='PIGW', Aliases='Gwt1, HPMRS5', description='phosphatidylinositol glycan anchor biosynthesis class W', other_designations='PIG-W|phosphatidylinositol glycan anchor biosynthesis, class W|phosphatidylinositol glycan, class W', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36535002, end_position_on_the_genomic_accession=36539368, orientation='plus', exon_count=3, OMIM=610275, no_hdr0=''),
  284099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284099, CurrentID=0, Status='live', Symbol='C17orf78', Aliases='', description='chromosome 17 open reading frame 78', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37375935, end_position_on_the_genomic_accession=37392708, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284100, CurrentID=0, Status='live', Symbol='YWHAEP7', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 7', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37842949, end_position_on_the_genomic_accession=37884743, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100270856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270856, CurrentID=0, Status='live', Symbol='RPL9P13', Aliases='RPL9_2_278', description='ribosomal protein L9 pseudogene 13', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139732708, end_position_on_the_genomic_accession=139733288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060172, CurrentID=0, Status='live', Symbol='LOC101060172', Aliases='', description='uncharacterized LOC101060172', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115353883, end_position_on_the_genomic_accession=115366020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284106, CurrentID=0, Status='live', Symbol='CISD3', Aliases='Miner2', description='CDGSH iron sulfur domain 3', other_designations='CDGSH iron sulfur domain-containing protein 3, mitochondrial|mitoNEET related 2|mitoNEET-related protein 2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38730257, end_position_on_the_genomic_accession=38735605, orientation='plus', exon_count=4, OMIM=611933, no_hdr0=''),
  284108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284108, CurrentID=0, Status='live', Symbol='LOC284108', Aliases='', description='uncharacterized LOC284108', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284110, CurrentID=0, Status='live', Symbol='GSDMA', Aliases='FKSG9, GSDM, GSDM1', description='gasdermin A', other_designations='gasdermin 1|gasdermin-1', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39962973, end_position_on_the_genomic_accession=39977880, orientation='plus', exon_count=14, OMIM=611218, no_hdr0=''),
  284111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284111, CurrentID=0, Status='live', Symbol='SLC13A5', Aliases='EIEE25, NACT, mIndy', description='solute carrier family 13 (sodium-dependent citrate transporter), member 5', other_designations='Na(+)/citrate cotransporter|Na+-coupled citrate transporter protein|sodium-dependent dicarboxylate transporter', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6684713, end_position_on_the_genomic_accession=6713421, orientation='minus', exon_count=13, OMIM=608305, no_hdr0=''),
  284112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284112, CurrentID=0, Status='live', Symbol='LOC284112', Aliases='', description='uncharacterized LOC284112', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284114, CurrentID=0, Status='live', Symbol='TMEM102', Aliases='CBAP', description='transmembrane protein 102', other_designations='common beta-chain associated protein|common beta-chain-associated protein', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7428825, end_position_on_the_genomic_accession=7437679, orientation='plus', exon_count=6, OMIM=613936, no_hdr0=''),
  284116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284116, CurrentID=0, Status='live', Symbol='KRT42P', Aliases='', description='keratin 42 pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41626327, end_position_on_the_genomic_accession=41640199, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  104355285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355285, CurrentID=0, Status='live', Symbol='LINC01256', Aliases='TCONS_l2_00021715', description='long intergenic non-protein coding RNA 1256', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132591089, end_position_on_the_genomic_accession=132678503, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104355286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355286, CurrentID=0, Status='live', Symbol='LINC01307', Aliases='', description='long intergenic non-protein coding RNA 1307', other_designations='RP11-157N3.1', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101323837, end_position_on_the_genomic_accession=101377313, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284119, CurrentID=0, Status='live', Symbol='PTRF', Aliases='CAVIN, CAVIN1, CGL4, FKSG13, cavin-1', description='polymerase I and transcript release factor', other_designations='RNA polymerase I and transcript release factor|TTF-I interacting peptide 12', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42402449, end_position_on_the_genomic_accession=42423320, orientation='minus', exon_count=3, OMIM=603198, no_hdr0=''),
  104355288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355288, CurrentID=0, Status='live', Symbol='LINC01320', Aliases='', description='long intergenic non-protein coding RNA 1320', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34677557, end_position_on_the_genomic_accession=34722563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104355289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355289, CurrentID=0, Status='live', Symbol='LINC01324', Aliases='', description='long intergenic non-protein coding RNA 1324', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164714095, end_position_on_the_genomic_accession=164831480, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104355290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355290, CurrentID=0, Status='live', Symbol='MIR217HG', Aliases='MIR216AHG, MIR216BHG', description='MIR217 host gene', other_designations='MIR217 host gene (non-protein coding)', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55963406, end_position_on_the_genomic_accession=56047326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284123, CurrentID=0, Status='live', Symbol='FAM27L', Aliases='', description='family with sequence similarity 27-like', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22298764, end_position_on_the_genomic_accession=22299893, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  284124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284124, CurrentID=0, Status='live', Symbol='FLJ36000', Aliases='', description='uncharacterized FLJ36000', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22406019, end_position_on_the_genomic_accession=22413741, orientation='plus', exon_count=2, OMIM=605160, no_hdr0=''),
  104355293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355293, CurrentID=0, Status='live', Symbol='LINC00376', Aliases='', description='long intergenic non-protein coding RNA 376', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63183101, end_position_on_the_genomic_accession=63328094, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  104355294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355294, CurrentID=0, Status='live', Symbol='LINC01386', Aliases='', description='long intergenic non-protein coding RNA 1386', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73454190, end_position_on_the_genomic_accession=73472966, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  104355295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355295, CurrentID=0, Status='live', Symbol='SGMS1-AS1', Aliases='', description='SGMS1 antisense RNA 1', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50624951, end_position_on_the_genomic_accession=50631636, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104355296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355296, CurrentID=0, Status='live', Symbol='SCEL-AS1', Aliases='', description='SCEL antisense RNA 1', other_designations='SCEL antisense RNA 1 (non-protein coding)', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77599755, end_position_on_the_genomic_accession=77606551, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284129, CurrentID=0, Status='live', Symbol='SLC26A11', Aliases='', description='solute carrier family 26 (anion exchanger), member 11', other_designations='solute carrier family 26, member 11', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80220401, end_position_on_the_genomic_accession=80253509, orientation='plus', exon_count=18, OMIM=610117, no_hdr0=''),
  284131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284131, CurrentID=0, Status='live', Symbol='ENDOV', Aliases='', description='endonuclease V', other_designations='hEndoV|inosine-specific endoribonuclease', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80415117, end_position_on_the_genomic_accession=80438086, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  101060177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060177, CurrentID=0, Status='live', Symbol='LOC101060177', Aliases='', description='uncharacterized LOC101060177', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52588526, end_position_on_the_genomic_accession=52621724, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100419615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419615, CurrentID=0, Status='live', Symbol='LOC100419615', Aliases='', description='complement component 7 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54354239, end_position_on_the_genomic_accession=54355190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100292073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100292073, CurrentID=0, Status='live', Symbol='LOC100292073', Aliases='', description='uncharacterized LOC100292073', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060179, CurrentID=0, Status='live', Symbol='LOC101060179', Aliases='', description='uncharacterized LOC101060179', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45213066, end_position_on_the_genomic_accession=45214744, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373346, CurrentID=0, Status='live', Symbol='LOC105373346', Aliases='', description='uncharacterized LOC105373346', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=305121, end_position_on_the_genomic_accession=314395, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  353194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353194, CurrentID=0, Status='live', Symbol='KRT16P6', Aliases='', description='keratin 16 pseudogene 6', other_designations='keratin 14 pseudogene|keratin pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16817983, end_position_on_the_genomic_accession=16822553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060181, CurrentID=0, Status='live', Symbol='LOC101060181', Aliases='', description='zinc finger protein ZnFP12', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12247244, end_position_on_the_genomic_accession=12267038, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284161, CurrentID=0, Status='live', Symbol='GDPD1', Aliases='GDE4', description='glycerophosphodiester phosphodiesterase domain containing 1', other_designations='glycerophosphodiester phosphodiesterase 4', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59220467, end_position_on_the_genomic_accession=59275970, orientation='plus', exon_count=10, OMIM=616317, no_hdr0=''),
  153090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153090, CurrentID=0, Status='live', Symbol='DAB2IP', Aliases='AF9Q34, AIP-1, AIP1, DIP1/2', description='DAB2 interacting protein', other_designations='ASK-interacting protein 1|ASK1-interacting protein 1|DAB2 interaction protein|DOC-2/DAB2 interactive protein|nGAP-like protein', map_location='9q33.1-q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121566883, end_position_on_the_genomic_accession=121785530, orientation='plus', exon_count=26, OMIM=609205, no_hdr0=''),
  284167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284167, CurrentID=0, Status='live', Symbol='POLRMTP1', Aliases='POLRMTL', description='polymerase (RNA) mitochondrial (DNA directed) pseudogene 1', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62136923, end_position_on_the_genomic_accession=62140639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373347, CurrentID=0, Status='live', Symbol='LOC105373347', Aliases='', description='periphilin-1-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147180941, end_position_on_the_genomic_accession=147271895, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  284184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284184, CurrentID=0, Status='live', Symbol='C17orf89', Aliases='', description='chromosome 17 open reading frame 89', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81239311, end_position_on_the_genomic_accession=81241298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284185, CurrentID=0, Status='live', Symbol='LINC00482', Aliases='C17orf55', description='long intergenic non-protein coding RNA 482', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81302824, end_position_on_the_genomic_accession=81309248, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284186, CurrentID=0, Status='live', Symbol='TMEM105', Aliases='', description='transmembrane protein 105', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81305814, end_position_on_the_genomic_accession=81330674, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  284191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284191, CurrentID=0, Status='live', Symbol='LOC284191', Aliases='', description='uncharacterized LOC284191', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16975235, end_position_on_the_genomic_accession=16981353, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284194, CurrentID=0, Status='live', Symbol='LGALS9B', Aliases='', description='lectin, galactoside-binding, soluble, 9B', other_designations='gal-9B|galectin-9 like|galectin-9 pseudogene|galectin-9-like protein A', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20449395, end_position_on_the_genomic_accession=20467535, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101060187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060187, CurrentID=0, Status='live', Symbol='LOC101060187', Aliases='', description='putative uncharacterized protein FLJ25328', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19745229, end_position_on_the_genomic_accession=19753407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284196, CurrentID=0, Status='live', Symbol='KRT17P6', Aliases='', description='keratin 17 pseudogene 6', other_designations='keratin pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20513494, end_position_on_the_genomic_accession=20517479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  153129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153129, CurrentID=0, Status='live', Symbol='SLC38A9', Aliases='URLC11', description='solute carrier family 38 member 9', other_designations='putative sodium-coupled neutral amino acid transporter 9|solute carrier family 38, member 9|up-regulated in lung cancer 11', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55625845, end_position_on_the_genomic_accession=55712335, orientation='minus', exon_count=21, OMIM=616203, no_hdr0=''),
  284203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284203, CurrentID=0, Status='live', Symbol='RNASEH1P1', Aliases='RNASEH1, RNASEH1P, RNH1', description='ribonuclease H1 pseudogene 1', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20901976, end_position_on_the_genomic_accession=20902776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284207, CurrentID=0, Status='live', Symbol='METRNL', Aliases='', description='meteorin, glial cell differentiation regulator-like', other_designations='subfatin', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83079691, end_position_on_the_genomic_accession=83095129, orientation='plus', exon_count=5, OMIM=616241, no_hdr0=''),
  105373348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373348, CurrentID=0, Status='live', Symbol='LOC105373348', Aliases='', description='uncharacterized LOC105373348', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147323980, end_position_on_the_genomic_accession=147332136, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79296, CurrentID=0, Status='live', Symbol='OR52E1', Aliases='OR52E1P', description='olfactory receptor family 52 subfamily E member 1 (gene/pseudogene)', other_designations='Olfactory receptor 52E1|olfactory receptor, family 52, subfamily E, member 1 pseudogene|seven transmembrane helix receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5069472, end_position_on_the_genomic_accession=5070596, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284215, CurrentID=0, Status='live', Symbol='DLGAP1-AS5', Aliases='', description='DLGAP1 antisense RNA 5', other_designations='DLGAP1 antisense RNA 5 (non-protein coding)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=4264602, end_position_on_the_genomic_accession=4296000, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284217, CurrentID=0, Status='live', Symbol='LAMA1', Aliases='LAMA, PTBHS, S-LAM-alpha', description='laminin subunit alpha 1', other_designations='S-LAM alpha|S-laminin subunit alpha|laminin A chain|laminin, alpha 1|laminin-3 subunit alpha', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6941744, end_position_on_the_genomic_accession=7117814, orientation='minus', exon_count=64, OMIM=150320, no_hdr0=''),
  284219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284219, CurrentID=0, Status='live', Symbol='LOC284219', Aliases='', description='uncharacterized LOC284219', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060192, CurrentID=0, Status='live', Symbol='BNIP3P13', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 13', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20059569, end_position_on_the_genomic_accession=20063382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266971, CurrentID=0, Status='live', Symbol='PIPSL', Aliases='PIP5K1L1, PIP5K1P3, PSMD4P2, bA429H9.1', description='PIP5K1A and PSMD4-like, pseudogene', other_designations='PIP5K1A-PSMD4|phosphatidylinositol-4-phosphate 5-kinase, type I, alpha pseudogene|phosphatidylinositol-4-phosphate 5-kinase, type I-like 1|proteasome (prosome, macropain) 26S subunit, non-ATPase, 4, pseudogene 2|proteasome 26S non-ATPase subunit 4 pseudogene|putative PIP5K1A and PSMD4-like protein', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93958140, end_position_on_the_genomic_accession=93961915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284232, CurrentID=0, Status='live', Symbol='ANKRD20A9P', Aliases='', description='ankyrin repeat domain 20 family member A9, pseudogene', other_designations='ankyrin repeat domain 20 family, member A2 pseudogene|ankyrin repeat domain 20 family, member A9, pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18834403, end_position_on_the_genomic_accession=18871969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284233, CurrentID=0, Status='live', Symbol='CYP4F35P', Aliases='', description='cytochrome P450 family 4 subfamily F member 35, pseudogene', other_designations='CYP4F-se8[6:7:8]|cytochrome P450, family 4, subfamily F, polypeptide 35, pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14337423, end_position_on_the_genomic_accession=14342524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  153163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153163, CurrentID=0, Status='live', Symbol='MGC32805', Aliases='', description='uncharacterized LOC153163', other_designations='CTC-210G5.1', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122436497, end_position_on_the_genomic_accession=122479087, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  284240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284240, CurrentID=0, Status='live', Symbol='LOC284240', Aliases='', description='uncharacterized LOC284240', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79574635, end_position_on_the_genomic_accession=79591858, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105373349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373349, CurrentID=0, Status='live', Symbol='LOC105373349', Aliases='', description='uncharacterized LOC105373349', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147877985, end_position_on_the_genomic_accession=147885470, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  284242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284242, CurrentID=0, Status='live', Symbol='LOC284242', Aliases='', description='uncharacterized LOC284242', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.9', start_position_on_the_genomic_accession=29522538, end_position_on_the_genomic_accession=29524119, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060195, CurrentID=0, Status='live', Symbol='LOC101060195', Aliases='', description='rRNA-processing protein FCF1 homolog pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48290279, end_position_on_the_genomic_accession=48291375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284244, CurrentID=0, Status='live', Symbol='LOC284244', Aliases='', description='uncharacterized LOC284244', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288683, CurrentID=0, Status='live', Symbol='LOC100288683', Aliases='', description='ribosomal protein L21 pseudogene', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13631785, end_position_on_the_genomic_accession=13632641, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271225, CurrentID=0, Status='live', Symbol='RPL7P45', Aliases='RPL7_25_1334', description='ribosomal protein L7 pseudogene 45', other_designations='', map_location='13q33.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=104814311, end_position_on_the_genomic_accession=104815135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134756, CurrentID=0, Status='live', Symbol='RPL29P1', Aliases='RPL29_5_654, dJ336K20B.2', description='ribosomal protein L29 pseudogene 1', other_designations='RPL29P1 ribosomal protein L29 pseudogene 1', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7620423, end_position_on_the_genomic_accession=7620885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284252, CurrentID=0, Status='live', Symbol='KCTD1', Aliases='C18orf5, SENS', description='potassium channel tetramerization domain containing 1', other_designations='potassium channel tetramerisation domain containing 1|potassium channel tetramerization domain-containing protein 1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26454910, end_position_on_the_genomic_accession=26657401, orientation='minus', exon_count=10, OMIM=613420, no_hdr0=''),
  284254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284254, CurrentID=0, Status='live', Symbol='DYNAP', Aliases='C18orf26', description='dynactin associated protein', other_designations='full', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54532713, end_position_on_the_genomic_accession=54599493, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284257, CurrentID=0, Status='live', Symbol='BOD1L2', Aliases='BOD1P, FAM44C', description='biorientation of chromosomes in cell division 1-like 2', other_designations='biorientation of chromosomes in cell division 1 pseudogene|biorientation of chromosomes in cell division protein 1 pseudogene|family with sequence similarity 44, member C|putative biorientation of chromosomes in cell division protein 1 pseudogene|putative biorientation of chromosomes in cell division protein 1-like 2|putative protein FAM44C', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57147062, end_position_on_the_genomic_accession=57150408, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104355426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355426, CurrentID=0, Status='live', Symbol='LENG8-AS1', Aliases='', description='LENG8 antisense RNA 1', other_designations='LENG8 antisense RNA 1 (non-protein coding)', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54444816, end_position_on_the_genomic_accession=54449038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284260, CurrentID=0, Status='live', Symbol='LINC00907', Aliases='', description='long intergenic non-protein coding RNA 907', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=42186668, end_position_on_the_genomic_accession=42691426, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  284263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284263, CurrentID=0, Status='live', Symbol='LOC284263', Aliases='', description='uncharacterized LOC284263', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284266, CurrentID=0, Status='live', Symbol='SIGLEC15', Aliases='CD33L3, HsT1361, SIGLEC-15', description='sialic acid binding Ig-like lectin 15', other_designations='CD33 antigen-like 3|CD33 molecule-like 3', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45825580, end_position_on_the_genomic_accession=45842556, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101060199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060199, CurrentID=0, Status='live', Symbol='LOC101060199', Aliases='', description='acyl-coenzyme A synthetase ACSM6, mitochondrial-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51082787, end_position_on_the_genomic_accession=51091237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284269, CurrentID=0, Status='live', Symbol='LOC284269', Aliases='', description='chromosome 9 open reading frame 86 pseudogene 1', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15003208, end_position_on_the_genomic_accession=15006188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  153201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153201, CurrentID=0, Status='live', Symbol='SLC36A2', Aliases='PAT2, TRAMD1', description='solute carrier family 36 (proton/amino acid symporter), member 2', other_designations='tramdorin-1', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151314978, end_position_on_the_genomic_accession=151347590, orientation='minus', exon_count=12, OMIM=608331, no_hdr0=''),
  284274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284274, CurrentID=0, Status='live', Symbol='SMIM21', Aliases='C18orf62', description='small integral membrane protein 21', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75409476, end_position_on_the_genomic_accession=75455653, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  284276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284276, CurrentID=0, Status='live', Symbol='LINC00908', Aliases='', description='long intergenic non-protein coding RNA 908', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76528655, end_position_on_the_genomic_accession=76559827, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271226, CurrentID=0, Status='live', Symbol='RPL7P51', Aliases='RPL7_29_1664', description='ribosomal protein L7 pseudogene 51', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51648838, end_position_on_the_genomic_accession=51649677, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284288, CurrentID=0, Status='live', Symbol='RSL24D1P11', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 11', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57838476, end_position_on_the_genomic_accession=57839039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  153218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153218, CurrentID=0, Status='live', Symbol='SPINK13', Aliases='HBVDNAPTP1, HESPINTOR, LiESP6, SPINK5L3', description='serine peptidase inhibitor, Kazal type 13 (putative)', other_designations='Hepatitis B Virus (HBV) DNA polymerase transactivated protein 1|hepatitis B virus DNA polymerase transactivated human serine protease inhibitor|hepatitis B virus DNA polymerase transactivated serine protease inhibitor|serine protease inhibitor Kazal-type 5-like 3', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148268860, end_position_on_the_genomic_accession=148286210, orientation='plus', exon_count=5, OMIM=615205, no_hdr0=''),
  284293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284293, CurrentID=0, Status='live', Symbol='HMSD', Aliases='ACC-6, ACC6, C18orf53, HSMD-v', description='histocompatibility (minor) serpin domain containing', other_designations='minor histocompatibility protein HMSD', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63949250, end_position_on_the_genomic_accession=63961821, orientation='plus', exon_count=5, OMIM=612086, no_hdr0=''),
  153222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153222, CurrentID=0, Status='live', Symbol='CREBRF', Aliases='C5orf41, LRF', description='CREB3 regulatory factor', other_designations='UPF0474 protein C5orf41|adult retina protein|luman recruitment factor|luman-recruiting factor|luman/CREB3 recruitment factor', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173056349, end_position_on_the_genomic_accession=173139293, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  284297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284297, CurrentID=0, Status='live', Symbol='SSC5D', Aliases='', description='scavenger receptor cysteine rich family, 5 domains', other_designations='scavenger receptor cysteine rich domain containing (5 domains)|scavenger receptor cysteine-rich glycoprotein|soluble scavenger protein with 5 SRCR domains|soluble scavenger with 5 domains', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55488497, end_position_on_the_genomic_accession=55519099, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  284306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284306, CurrentID=0, Status='live', Symbol='ZNF547', Aliases='', description='zinc finger protein 547', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57363435, end_position_on_the_genomic_accession=57379559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  284307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284307, CurrentID=0, Status='live', Symbol='ZIK1', Aliases='ZNF762', description='zinc finger protein interacting with K protein 1', other_designations='zinc finger protein 762|zinc finger protein interacting with K protein 1 homolog', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57583642, end_position_on_the_genomic_accession=57593888, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284309, CurrentID=0, Status='live', Symbol='ZNF776', Aliases='', description='zinc finger protein 776', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57746796, end_position_on_the_genomic_accession=57758159, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284312, CurrentID=0, Status='live', Symbol='ZSCAN1', Aliases='MZF-1, ZNF915', description='zinc finger and SCAN domain containing 1', other_designations='zinc finger with SCAN domain 1', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58033989, end_position_on_the_genomic_accession=58058907, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  153241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153241, CurrentID=0, Status='live', Symbol='CEP120', Aliases='CCDC100, SRTD13', description='centrosomal protein 120kDa', other_designations='coiled-coil domain-containing protein 100', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123344885, end_position_on_the_genomic_accession=123423592, orientation='minus', exon_count=23, OMIM=613446, no_hdr0=''),
  101060207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060207, CurrentID=0, Status='live', Symbol='LOC101060207', Aliases='', description='uncharacterized LOC101060207', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144414068, end_position_on_the_genomic_accession=144419448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284323, CurrentID=0, Status='live', Symbol='ZNF780A', Aliases='ZNF780', description='zinc finger protein 780A', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40069152, end_position_on_the_genomic_accession=40091736, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  284325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284325, CurrentID=0, Status='live', Symbol='C19orf54', Aliases='', description='chromosome 19 open reading frame 54', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40740856, end_position_on_the_genomic_accession=40750713, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79300, CurrentID=0, Status='live', Symbol='OR51P1P', Aliases='OR51P2P', description='olfactory receptor family 51 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily P, member 1 pseudogene|olfactory receptor, family 51, subfamily P, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5015030, end_position_on_the_genomic_accession=5016181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284336, CurrentID=0, Status='live', Symbol='LOC284336', Aliases='', description='uncharacterized LOC284336', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284338, CurrentID=0, Status='live', Symbol='PRR19', Aliases='', description='proline rich 19', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42302132, end_position_on_the_genomic_accession=42310821, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  284339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284339, CurrentID=0, Status='live', Symbol='TMEM145', Aliases='', description='transmembrane protein 145', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42313325, end_position_on_the_genomic_accession=42325064, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  284340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284340, CurrentID=0, Status='live', Symbol='CXCL17', Aliases='DMC, Dcip1, UNQ473, VCC-1, VCC1', description='chemokine (C-X-C motif) ligand 17', other_designations='C-X-C motif chemokine 17|VEGF co-regulated chemokine 1|dendritic cell and monocyte chemokine-like protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42428543, end_position_on_the_genomic_accession=42442984, orientation='minus', exon_count=4, OMIM=611387, no_hdr0=''),
  284344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284344, CurrentID=0, Status='live', Symbol='LOC284344', Aliases='', description='uncharacterized LOC284344', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43211791, end_position_on_the_genomic_accession=43248646, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101060212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060212, CurrentID=0, Status='live', Symbol='LOC101060212', Aliases='', description='puromycin-sensitive aminopeptidase-like protein', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36274176, end_position_on_the_genomic_accession=36320688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284346, CurrentID=0, Status='live', Symbol='ZNF575', Aliases='', description='zinc finger protein 575', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43531805, end_position_on_the_genomic_accession=43536138, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100419622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419622, CurrentID=0, Status='live', Symbol='LOC100419622', Aliases='', description='ribosomal protein L37 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48899131, end_position_on_the_genomic_accession=48899748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284348, CurrentID=0, Status='live', Symbol='LYPD5', Aliases='PRO4356', description='LY6/PLAUR domain containing 5', other_designations='metastasis-associated protein', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43795927, end_position_on_the_genomic_accession=43827199, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  284349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284349, CurrentID=0, Status='live', Symbol='ZNF283', Aliases='HZF19, HZF41', description='zinc finger protein 283', other_designations='zinc finger protein 41|zinc finger protein HZF19', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43827292, end_position_on_the_genomic_accession=43848898, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  284352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284352, CurrentID=0, Status='live', Symbol='PPP1R37', Aliases='LRRC68', description='protein phosphatase 1 regulatory subunit 37', other_designations='leucine rich repeat containing 68|leucine-rich repeat-containing protein 68|protein phosphatase 1, regulatory subunit 37', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45093173, end_position_on_the_genomic_accession=45147285, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  284353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284353, CurrentID=0, Status='live', Symbol='NKPD1', Aliases='', description='NTPase, KAP family P-loop domain containing 1', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45149111, end_position_on_the_genomic_accession=45162841, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100292290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100292290, CurrentID=0, Status='live', Symbol='HSPE1P8', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 8', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 8', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39303907, end_position_on_the_genomic_accession=39304919, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284355, CurrentID=0, Status='live', Symbol='TPRX1', Aliases='TPRX', description='tetra-peptide repeat homeobox 1', other_designations='CRX like homeobox 2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47801243, end_position_on_the_genomic_accession=47803604, orientation='minus', exon_count=2, OMIM=611166, no_hdr0=''),
  284358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284358, CurrentID=0, Status='live', Symbol='MAMSTR', Aliases='MASTR', description='MEF2 activating motif and SAP domain containing transcriptional regulator', other_designations='MEF2-activating SAP transcriptional regulatory protein|likely ortholog of MEF2-activating SAP transcriptional regulator', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48712722, end_position_on_the_genomic_accession=48719778, orientation='minus', exon_count=10, OMIM=610349, no_hdr0=''),
  284359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284359, CurrentID=0, Status='live', Symbol='IZUMO1', Aliases='IZUMO', description='izumo sperm-egg fusion 1', other_designations='OBF|oocyte binding/fusion factor|sperm-specific protein Izumo', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48740888, end_position_on_the_genomic_accession=48747650, orientation='minus', exon_count=14, OMIM=609278, no_hdr0=''),
  284361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284361, CurrentID=0, Status='live', Symbol='EMC10', Aliases='C19orf63, HSM1, HSS1', description='ER membrane protein complex subunit 10', other_designations='UPF0510 protein INM02|hematopoietic signal peptide-containing membrane domain-containing 1|hematopoietic signal peptide-containing secreted 1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50476457, end_position_on_the_genomic_accession=50505909, orientation='plus', exon_count=11, OMIM=614545, no_hdr0=''),
  284365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284365, CurrentID=0, Status='live', Symbol='MGC45922', Aliases='', description='uncharacterized LOC284365', other_designations='CTD-2568A17.5', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50817681, end_position_on_the_genomic_accession=50818878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284366, CurrentID=0, Status='live', Symbol='KLK9', Aliases='KLK-L3, KLKL3', description='kallikrein related peptidase 9', other_designations='kallikrein-like protein 3|kallikrein-related peptidase 9', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51002513, end_position_on_the_genomic_accession=51009634, orientation='minus', exon_count=5, OMIM=605504, no_hdr0=''),
  284367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284367, CurrentID=0, Status='live', Symbol='SIGLEC17P', Aliases='HSPC078, SIGLECP2, SIGLECP3', description='sialic acid binding Ig-like lectin 17, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 3|silec pseudogene-2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51167328, end_position_on_the_genomic_accession=51173524, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284369, CurrentID=0, Status='live', Symbol='SIGLECL1', Aliases='C19orf75, SIGLEC23P, SIGLECP7', description='SIGLEC family like 1', other_designations='sialic acid binding Ig-like lectin, pseudogene 7', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51251026, end_position_on_the_genomic_accession=51269330, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  284370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284370, CurrentID=0, Status='live', Symbol='ZNF615', Aliases='', description='zinc finger protein 615', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51991332, end_position_on_the_genomic_accession=52008238, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  284371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284371, CurrentID=0, Status='live', Symbol='ZNF841', Aliases='', description='zinc finger protein 841', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52064363, end_position_on_the_genomic_accession=52095773, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  284379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284379, CurrentID=0, Status='live', Symbol='LOC284379', Aliases='', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 3 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53599631, end_position_on_the_genomic_accession=53603497, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284382, CurrentID=0, Status='live', Symbol='ACTL9', Aliases='HSD21', description='actin like 9', other_designations='actin-like 9', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8697397, end_position_on_the_genomic_accession=8698830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284383, CurrentID=0, Status='live', Symbol='OR2Z1', Aliases='OR19-4, OR2Z2', description='olfactory receptor family 2 subfamily Z member 1', other_designations='olfactory receptor 2Z2|olfactory receptor OR19-4|olfactory receptor, family 2, subfamily Z, member 1|olfactory receptor, family 2, subfamily Z, member 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8730954, end_position_on_the_genomic_accession=8732009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284385, CurrentID=0, Status='live', Symbol='ZNF561-AS1', Aliases='C19orf82', description='ZNF561 antisense RNA 1 (head to head)', other_designations='uncharacterized protein C19orf82', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9621482, end_position_on_the_genomic_accession=9634862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481942, CurrentID=0, Status='live', Symbol='RNU6-1201P', Aliases='', description='RNA, U6 small nuclear 1201, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284390, CurrentID=0, Status='live', Symbol='ZNF763', Aliases='ZNF, ZNF440L', description='zinc finger protein 763', other_designations='DNA-binding protein|zinc finger protein 440 like', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11965054, end_position_on_the_genomic_accession=11980399, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  284391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284391, CurrentID=0, Status='live', Symbol='ZNF844', Aliases='', description='zinc finger protein 844', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12064731, end_position_on_the_genomic_accession=12077812, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79302, CurrentID=0, Status='live', Symbol='OR52J1P', Aliases='', description='olfactory receptor family 52 subfamily J member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily J, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5104224, end_position_on_the_genomic_accession=5105093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284393, CurrentID=0, Status='live', Symbol='RPL10P16', Aliases='RPL10_6_1628', description='ribosomal protein L10 pseudogene 16', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12643212, end_position_on_the_genomic_accession=12643942, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284395, CurrentID=0, Status='live', Symbol='LOC284395', Aliases='', description='uncharacterized LOC284395', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29287011, end_position_on_the_genomic_accession=29525752, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  153328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153328, CurrentID=0, Status='live', Symbol='SLC25A48', Aliases='', description='solute carrier family 25 member 48', other_designations='solute carrier family 25, member 48', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135834676, end_position_on_the_genomic_accession=135888637, orientation='plus', exon_count=14, OMIM=616150, no_hdr0=''),
  284402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284402, CurrentID=0, Status='live', Symbol='SCGB2B2', Aliases='SCGB4A2, SCGBL', description='secretoglobin, family 2B, member 2', other_designations='secretoglobin-like protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34576740, end_position_on_the_genomic_accession=34675731, orientation='minus', exon_count=15, OMIM=615063, no_hdr0=''),
  284403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284403, CurrentID=0, Status='live', Symbol='WDR62', Aliases='C19orf14, MCPH2', description='WD repeat domain 62', other_designations='microcephaly, primary autosomal recessive 2|truncated WDR62', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36054881, end_position_on_the_genomic_accession=36105111, orientation='plus', exon_count=34, OMIM=613583, no_hdr0=''),
  284406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284406, CurrentID=0, Status='live', Symbol='ZFP82', Aliases='ZNF545', description='ZFP82 zinc finger protein', other_designations='zfp-82|zinc finger protein 545|zinc finger protein 85', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36390332, end_position_on_the_genomic_accession=36418658, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  284408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284408, CurrentID=0, Status='live', Symbol='ZNF790-AS1', Aliases='', description='ZNF790 antisense RNA 1', other_designations='CTD-2162K18.5|ZNF790 antisense RNA 1 (non-protein coding)', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36797550, end_position_on_the_genomic_accession=36828111, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  153339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153339, CurrentID=0, Status='live', Symbol='TMEM167A', Aliases='TMEM167', description='transmembrane protein 167A', other_designations='transmembrane protein 167', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83052846, end_position_on_the_genomic_accession=83077453, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284412, CurrentID=0, Status='live', Symbol='LOC284412', Aliases='', description='uncharacterized LOC284412', other_designations='CTD-3220F14.1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37265939, end_position_on_the_genomic_accession=37269010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  284415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284415, CurrentID=0, Status='live', Symbol='VSTM1', Aliases='SIRL-1, SIRL1, UNQ3033', description='V-set and transmembrane domain containing 1', other_designations='LAIR homolog|OSCAR-like transcript-1|signal inhibitory receptor on leukocytes-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54040825, end_position_on_the_genomic_accession=54063966, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  284417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284417, CurrentID=0, Status='live', Symbol='TMEM150B', Aliases='TMEM224', description='transmembrane protein 150B', other_designations='hCG1651476|transmembrane protein 224|transmembrane protein LOC284417', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55308886, end_position_on_the_genomic_accession=55334043, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  284418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284418, CurrentID=0, Status='live', Symbol='FAM71E2', Aliases='C19orf16', description='family with sequence similarity 71 member E2', other_designations='family with sequence similarity 71, member E2|putative protein FAM71E2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55354908, end_position_on_the_genomic_accession=55363252, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  353238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353238, CurrentID=0, Status='live', Symbol='PADI6', Aliases='hPADVI', description='peptidyl arginine deiminase, type VI', other_designations='peptidyl arginine deiminase-like protein|peptidylarginine deiminase type 6', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17372196, end_position_on_the_genomic_accession=17401699, orientation='plus', exon_count=16, OMIM=610363, no_hdr0=''),
  284422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284422, CurrentID=0, Status='live', Symbol='SMIM24', Aliases='C19orf77, HSPC323', description='small integral membrane protein 24', other_designations='transmembrane protein C19orf77|transmembrane protein HSPC323', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3474407, end_position_on_the_genomic_accession=3480542, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284424, CurrentID=0, Status='live', Symbol='MIR7-3HG', Aliases='C19orf30, Huh7, LINC00306, NCRNA00306, PGSF1, uc002mbe.2', description='MIR7-3 host gene', other_designations='MIR7-3 host gene (non-protein coding)|long intergenic non-protein coding RNA 306|pituitary gland specific factor 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4769105, end_position_on_the_genomic_accession=4772556, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284427, CurrentID=0, Status='live', Symbol='SLC25A41', Aliases='APC4', description='solute carrier family 25 member 41', other_designations='solute carrier family 25, member 41', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6426037, end_position_on_the_genomic_accession=6436267, orientation='minus', exon_count=10, OMIM=610822, no_hdr0=''),
  284428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284428, CurrentID=0, Status='live', Symbol='MBD3L5', Aliases='', description='methyl-CpG binding domain protein 3-like 5', other_designations='MBD3-like 5|MBD3-like protein 5', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7029930, end_position_on_the_genomic_accession=7033011, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060226, CurrentID=0, Status='live', Symbol='NBPF19', Aliases='', description='neuroblastoma breakpoint family member 19', other_designations='neuroblastoma breakpoint family, member 19', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149463737, end_position_on_the_genomic_accession=149556256, orientation='plus', exon_count=16, OMIM=614006, no_hdr0=''),
  284433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284433, CurrentID=0, Status='live', Symbol='OR10H5', Aliases='OR19-25, OR19-26', description='olfactory receptor family 10 subfamily H member 5', other_designations='olfactory receptor OR19-25|olfactory receptor OR19-26|olfactory receptor, family 10, subfamily H, member 5', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15794049, end_position_on_the_genomic_accession=15794996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284434, CurrentID=0, Status='live', Symbol='NWD1', Aliases='', description='NACHT and WD repeat domain containing 1', other_designations='NACHT and WD repeat domain-containing protein 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16719853, end_position_on_the_genomic_accession=16817963, orientation='plus', exon_count=22, OMIM=616250, no_hdr0=''),
  153364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153364, CurrentID=0, Status='live', Symbol='MBLAC2', Aliases='', description='metallo-beta-lactamase domain containing 2', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90458203, end_position_on_the_genomic_accession=90474768, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  284439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284439, CurrentID=0, Status='live', Symbol='SLC25A42', Aliases='', description='solute carrier family 25 member 42', other_designations='solute carrier family 25, member 42', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19063966, end_position_on_the_genomic_accession=19113030, orientation='plus', exon_count=10, OMIM=610823, no_hdr0=''),
  284440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284440, CurrentID=0, Status='live', Symbol='LINC00663', Aliases='', description='long intergenic non-protein coding RNA 663', other_designations='CTC-559E9.1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19756371, end_position_on_the_genomic_accession=19776413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284441, CurrentID=0, Status='live', Symbol='LOC284441', Aliases='', description='actin-related protein 2 pseudogene', other_designations='ARP2 actin-related protein 2 homolog (yeast) pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20257647, end_position_on_the_genomic_accession=20262883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284443, CurrentID=0, Status='live', Symbol='ZNF493', Aliases='', description='zinc finger protein 493', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21397119, end_position_on_the_genomic_accession=21427495, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  677662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677662, CurrentID=0, Status='live', Symbol='PARK12', Aliases='', description='Parkinson disease 12 (susceptibility)', other_designations='Parkinson disease (X-linked) 12', map_location='Xq21-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300557, no_hdr0=''),
  677663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677663, CurrentID=0, Status='live', Symbol='PAND3', Aliases='', description='panic disorder 3', other_designations='', map_location='4q31-q34', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609985, no_hdr0=''),
  284451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284451, CurrentID=0, Status='live', Symbol='ODF3L2', Aliases='C19orf19', description='outer dense fiber of sperm tails 3-like 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=463346, end_position_on_the_genomic_accession=474983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  284454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284454, CurrentID=0, Status='live', Symbol='LOC284454', Aliases='', description='uncharacterized LOC284454', other_designations='CTD-3252C9.4', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13834516, end_position_on_the_genomic_accession=13836289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271232, CurrentID=0, Status='live', Symbol='RPS20P24', Aliases='RPS20_10_984', description='ribosomal protein S20 pseudogene 24', other_designations='', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72655782, end_position_on_the_genomic_accession=72656292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284459, CurrentID=0, Status='live', Symbol='HKR1', Aliases='ZNF875', description='HKR1, GLI-Kruppel zinc finger family member', other_designations='GLI-Kruppel family member HKR1|oncogene HKR1|zinc finger protein 875', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37334678, end_position_on_the_genomic_accession=37364455, orientation='plus', exon_count=6, OMIM=165250, no_hdr0=''),
  677679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677679, CurrentID=0, Status='live', Symbol='SCARNA3', Aliases='HBI-100', description='small Cajal body-specific RNA 3', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175968397, end_position_on_the_genomic_accession=175968540, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873536, CurrentID=0, Status='live', Symbol='RNA5SP283', Aliases='RN5S283', description='RNA, 5S ribosomal pseudogene 283', other_designations='RNA, 5S ribosomal 283', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62802320, end_position_on_the_genomic_accession=62802438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  677681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677681, CurrentID=0, Status='live', Symbol='SCARNA20', Aliases='ACA66', description='small Cajal body-specific RNA 20', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60231516, end_position_on_the_genomic_accession=60231645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284467, CurrentID=0, Status='live', Symbol='FAM19A3', Aliases='TAFA-3, TAFA3', description='family with sequence similarity 19 (chemokine (C-C motif)-like), member A3', other_designations='chemokine-like protein TAFA-3', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112718994, end_position_on_the_genomic_accession=112727235, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  153396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153396, CurrentID=0, Status='live', Symbol='TMEM161B', Aliases='FLB3342, PRO1313', description='transmembrane protein 161B', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88185575, end_position_on_the_genomic_accession=88268879, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100129449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129449, CurrentID=0, Status='live', Symbol='LOC100129449', Aliases='', description='PRO2055', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=153150200, end_position_on_the_genomic_accession=153150680, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284485, CurrentID=0, Status='live', Symbol='RIIAD1', Aliases='C1orf230, NCRNA00166', description='regulatory subunit of type II PKA R-subunit (RIIa) domain containing 1', other_designations='RIIa domain-containing protein C1orf230|RIIa domain-containing protein ENSP00000357824', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151721537, end_position_on_the_genomic_accession=151729606, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284486, CurrentID=0, Status='live', Symbol='THEM5', Aliases='ACOT15', description='thioesterase superfamily member 5', other_designations='acyl-CoA thioesterase THEM5|acyl-coenzyme A thioesterase 15', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151847101, end_position_on_the_genomic_accession=151853697, orientation='minus', exon_count=6, OMIM=615653, no_hdr0=''),
  100271233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271233, CurrentID=0, Status='live', Symbol='RPS20P20', Aliases='RPS20_11_915', description='ribosomal protein S20 pseudogene 20', other_designations='', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71895995, end_position_on_the_genomic_accession=71896332, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421870, CurrentID=0, Status='live', Symbol='LOC100421870', Aliases='', description='WW domain containing E3 ubiquitin protein ligase 2 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65248046, end_position_on_the_genomic_accession=65248750, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284498, CurrentID=0, Status='live', Symbol='C1orf167', Aliases='', description='chromosome 1 open reading frame 167', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11762183, end_position_on_the_genomic_accession=11789585, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  284513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284513, CurrentID=0, Status='live', Symbol='LOC284513', Aliases='', description='uncharacterized LOC284513', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  153443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153443, CurrentID=0, Status='live', Symbol='SRFBP1', Aliases='BUD22, P49, Rlb1, STRAP, p49/STRAP', description='serum response factor binding protein 1', other_designations='BUD22 homolog|SRF-dependent transcription regulation associated protein|SRF-dependent transcription regulation-associated protein', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121961961, end_position_on_the_genomic_accession=122028600, orientation='plus', exon_count=8, OMIM=610479, no_hdr0=''),
  100271234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271234, CurrentID=0, Status='live', Symbol='RPS20P21', Aliases='RPS20_12_916', description='ribosomal protein S20 pseudogene 21', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73981989, end_position_on_the_genomic_accession=73982514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284521, CurrentID=0, Status='live', Symbol='OR2L13', Aliases='OR2L14', description='olfactory receptor family 2 subfamily L member 13', other_designations='olfactory receptor 2L14|olfactory receptor, family 2, subfamily L, member 13|olfactory receptor, family 2, subfamily L, member 14', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247937029, end_position_on_the_genomic_accession=248100922, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  284525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284525, CurrentID=0, Status='live', Symbol='SLC9C2', Aliases='SLC9A11', description='solute carrier family 9, member C2 (putative)', other_designations='NHE-11|Na(+)/H(+) exchanger 11|solute carrier family 9|solute carrier family 9 member C2|solute carrier family 9, member 11', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173500465, end_position_on_the_genomic_accession=173603094, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  284532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284532, CurrentID=0, Status='live', Symbol='OR14A16', Aliases='OR5AT1', description='olfactory receptor family 14 subfamily A member 16', other_designations='olfactory receptor 5AT1|olfactory receptor OR1-45|olfactory receptor, family 14, subfamily A, member 16|olfactory receptor, family 5, subfamily AT, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247814800, end_position_on_the_genomic_accession=247834858, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284541, CurrentID=0, Status='live', Symbol='CYP4A22', Aliases='', description='cytochrome P450 family 4 subfamily A member 22', other_designations='CYPIVA22|cytochrome P450 4A22K|cytochrome P450, family 4, subfamily A, polypeptide 22|fatty acid omega-hydroxylase|lauric acid omega-hydroxylase|long-chain fatty acid omega-monooxygenase', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47137435, end_position_on_the_genomic_accession=47149738, orientation='plus', exon_count=12, OMIM=615341, no_hdr0=''),
  677759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677759, CurrentID=0, Status='live', Symbol='LINC00863', Aliases='', description='long intergenic non-protein coding RNA 863', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284546, CurrentID=0, Status='live', Symbol='C1orf185', Aliases='', description='chromosome 1 open reading frame 185', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51102234, end_position_on_the_genomic_accession=51148082, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  677763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677763, CurrentID=0, Status='live', Symbol='SCARNA21', Aliases='ACA68', description='small Cajal body-specific RNA 21', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7906123, end_position_on_the_genomic_accession=7906260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677764, CurrentID=0, Status='live', Symbol='MYP13', Aliases='', description='myopia 13', other_designations='', map_location='Xq23-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300613, no_hdr0=''),
  677765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677765, CurrentID=0, Status='live', Symbol='SCARNA18', Aliases='U109', description='small Cajal body-specific RNA 18', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83064204, end_position_on_the_genomic_accession=83064337, orientation='minus', exon_count=1, OMIM=611329, no_hdr0=''),
  677766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677766, CurrentID=0, Status='live', Symbol='SCARNA2', Aliases='HBII-382, mgU2-25/61', description='small Cajal body-specific RNA 2', other_designations='', map_location='1q13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109100193, end_position_on_the_genomic_accession=109100612, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677767, CurrentID=0, Status='live', Symbol='SCARNA7', Aliases='U90', description='small Cajal body-specific RNA 7', other_designations='', map_location='3q25.22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160514907, end_position_on_the_genomic_accession=160515236, orientation='minus', exon_count=1, OMIM=615644, no_hdr0=''),
  677768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677768, CurrentID=0, Status='live', Symbol='SCARNA13', Aliases='U93', description='small Cajal body-specific RNA 13', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95533355, end_position_on_the_genomic_accession=95533629, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677769, CurrentID=0, Status='live', Symbol='SCARNA17', Aliases='U91', description='small Cajal body-specific RNA 17', other_designations='mgU12-22/U4-8', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49814023, end_position_on_the_genomic_accession=49814443, orientation='plus', exon_count=1, OMIM=615645, no_hdr0=''),
  677770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677770, CurrentID=0, Status='live', Symbol='SCARNA22', Aliases='ACA11', description='small Cajal body-specific RNA 22', other_designations='ACA11 snoRNA', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1974636, end_position_on_the_genomic_accession=1974760, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677771, CurrentID=0, Status='live', Symbol='SCARNA4', Aliases='ACA26', description='small Cajal body-specific RNA 4', other_designations='ACA26 scaRNA', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155925958, end_position_on_the_genomic_accession=155926086, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677772, CurrentID=0, Status='live', Symbol='SCARNA6', Aliases='U88', description='small Cajal body-specific RNA 6', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233288676, end_position_on_the_genomic_accession=233288941, orientation='plus', exon_count=1, OMIM=615641, no_hdr0=''),
  677773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677773, CurrentID=0, Status='live', Symbol='SCARNA23', Aliases='ACA12', description='small Cajal body-specific RNA 23', other_designations='ACA12 scaRNA', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24744441, end_position_on_the_genomic_accession=24744570, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677774, CurrentID=0, Status='live', Symbol='SCARNA1', Aliases='ACA35', description='small Cajal body-specific RNA 1', other_designations='ACA35 scaRNA', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27834401, end_position_on_the_genomic_accession=27834566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677775, CurrentID=0, Status='live', Symbol='SCARNA5', Aliases='U87', description='small Cajal body-specific RNA 5', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233275726, end_position_on_the_genomic_accession=233276003, orientation='plus', exon_count=1, OMIM=615640, no_hdr0=''),
  677776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677776, CurrentID=0, Status='live', Symbol='SCARNA8', Aliases='U92', description='small Cajal body-specific RNA 8', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19063656, end_position_on_the_genomic_accession=19063786, orientation='minus', exon_count=1, OMIM=615646, no_hdr0=''),
  677777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677777, CurrentID=0, Status='live', Symbol='SCARNA12', Aliases='U89', description='small Cajal body-specific RNA 12', other_designations='U89 small nuclear RNA', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6967337, end_position_on_the_genomic_accession=6967606, orientation='minus', exon_count=1, OMIM=615642, no_hdr0=''),
  677778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677778, CurrentID=0, Status='live', Symbol='SCARNA15', Aliases='ACA45', description='small Cajal body-specific RNA 15', other_designations='ACA45 scaRNA', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82755945, end_position_on_the_genomic_accession=82756071, orientation='plus', exon_count=1, OMIM=612675, no_hdr0=''),
  677779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677779, CurrentID=0, Status='live', Symbol='LINC00488', Aliases='C3orf66', description='long intergenic non-protein coding RNA 488', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109178165, end_position_on_the_genomic_accession=109185261, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  677780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677780, CurrentID=0, Status='live', Symbol='SCARNA11', Aliases='ACA57', description='small Cajal body-specific RNA 11', other_designations='ACA57 scaRNA', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6581473, end_position_on_the_genomic_accession=6581609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284565, CurrentID=0, Status='live', Symbol='NBPF15', Aliases='AB14, AG3, NBPF16', description='neuroblastoma breakpoint family member 15', other_designations='neuroblastoma breakpoint family member 16|neuroblastoma breakpoint family, member 15|neuroblastoma breakpoint family, member 16', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144421386, end_position_on_the_genomic_accession=144461667, orientation='minus', exon_count=24, OMIM='610414,614005', no_hdr0=''),
  677784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677784, CurrentID=0, Status='live', Symbol='FAM138D', Aliases='F379', description='family with sequence similarity 138 member D', other_designations='F379 retina-specific protein|family with sequence similarity 138, member D', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=36661, end_position_on_the_genomic_accession=38133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284570, CurrentID=0, Status='live', Symbol='LOC284570', Aliases='', description='uncharacterized LOC284570', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533107, CurrentID=0, Status='live', Symbol='RTEL1-TNFRSF6B', Aliases='', description='RTEL1-TNFRSF6B readthrough (NMD candidate)', other_designations='RTEL1-TNFRSF6B readthrough (non-protein coding)', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63657810, end_position_on_the_genomic_accession=63698698, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  284573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284573, CurrentID=0, Status='live', Symbol='LINC00303', Aliases='C1orf157, NCRNA00303', description='long intergenic non-protein coding RNA 303', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204032447, end_position_on_the_genomic_accession=204041264, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  677790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677790, CurrentID=0, Status='live', Symbol='TMEM78', Aliases='', description='transmembrane protein 78', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=227452006, end_position_on_the_genomic_accession=227454180, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677792, CurrentID=0, Status='live', Symbol='SNORA1', Aliases='ACA1', description='small nucleolar RNA, H/ACA box 1', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93732004, end_position_on_the_genomic_accession=93732133, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677793, CurrentID=0, Status='live', Symbol='SNORA2A', Aliases='ACA2A', description='small nucleolar RNA, H/ACA box 2A', other_designations='ACA2a snoRNA', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48656648, end_position_on_the_genomic_accession=48656782, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677794, CurrentID=0, Status='live', Symbol='SNORA2B', Aliases='ACA2b', description='small nucleolar RNA, H/ACA box 2B', other_designations='ACA2b snoRNA', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48667457, end_position_on_the_genomic_accession=48667593, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677795, CurrentID=0, Status='live', Symbol='SNORA5B', Aliases='ACA5b', description='small nucleolar RNA, H/ACA box 5B', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45105968, end_position_on_the_genomic_accession=45106099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677796, CurrentID=0, Status='live', Symbol='SNORA5C', Aliases='ACA5c', description='small nucleolar RNA, H/ACA box 5C', other_designations='HBI-80', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45104906, end_position_on_the_genomic_accession=45105042, orientation='minus', exon_count=1, OMIM=611335, no_hdr0=''),
  677797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677797, CurrentID=0, Status='live', Symbol='SNORA7B', Aliases='ACA7B', description='small nucleolar RNA, H/ACA box 7B', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129397210, end_position_on_the_genomic_accession=129397348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677798, CurrentID=0, Status='live', Symbol='SNORA9', Aliases='ACA9', description='small nucleolar RNA, H/ACA box 9', other_designations='ACA9 snoRNA', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44985378, end_position_on_the_genomic_accession=44985510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677799, CurrentID=0, Status='live', Symbol='SNORA11', Aliases='SNORA11A, U107', description='small nucleolar RNA, H/ACA box 11', other_designations='U107 snoRNA|small nucleolar RNA, H/ACA box 11A', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54814370, end_position_on_the_genomic_accession=54814500, orientation='plus', exon_count=1, OMIM=300662, no_hdr0=''),
  677800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677800, CurrentID=0, Status='live', Symbol='SNORA12', Aliases='U108', description='small nucleolar RNA, H/ACA box 12', other_designations='U108 snoRNA', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100237156, end_position_on_the_genomic_accession=100237302, orientation='minus', exon_count=1, OMIM=611330, no_hdr0=''),
  677801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677801, CurrentID=0, Status='live', Symbol='SNORA14A', Aliases='ACA14a', description='small nucleolar RNA, H/ACA box 14A', other_designations='ACA14a snoRNA', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75943783, end_position_on_the_genomic_accession=75943916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677802, CurrentID=0, Status='live', Symbol='SNORA14B', Aliases='ACA14b', description='small nucleolar RNA, H/ACA box 14B', other_designations='ACA14b snoRNA', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235127803, end_position_on_the_genomic_accession=235127937, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677803, CurrentID=0, Status='live', Symbol='SNORA15', Aliases='ACA15', description='small nucleolar RNA, H/ACA box 15', other_designations='ACA15 snoRNA', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56060470, end_position_on_the_genomic_accession=56060602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677804, CurrentID=0, Status='live', Symbol='SNORA17A', Aliases='ACA17, SNORA17', description='small nucleolar RNA, H/ACA box 17A', other_designations='ACA17 snoRNA|small nucleolar RNA, H/ACA box 17', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136726747, end_position_on_the_genomic_accession=136726879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677805, CurrentID=0, Status='live', Symbol='SNORA18', Aliases='ACA18', description='small nucleolar RNA, H/ACA box 18', other_designations='ACA18 snoRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93733466, end_position_on_the_genomic_accession=93733597, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677806, CurrentID=0, Status='live', Symbol='SNORA20', Aliases='ACA20', description='small nucleolar RNA, H/ACA box 20', other_designations='ACA20 snoRNA', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159780250, end_position_on_the_genomic_accession=159780381, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677807, CurrentID=0, Status='live', Symbol='SNORA22', Aliases='ACA22', description='small nucleolar RNA, H/ACA box 22', other_designations='ACA22 snoRNA', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65755526, end_position_on_the_genomic_accession=65755659, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677808, CurrentID=0, Status='live', Symbol='SNORA23', Aliases='ACA23', description='small nucleolar RNA, H/ACA box 23', other_designations='ACA23 snoRNA', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9428766, end_position_on_the_genomic_accession=9428954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677809, CurrentID=0, Status='live', Symbol='SNORA24', Aliases='ACA24', description='small nucleolar RNA, H/ACA box 24', other_designations='ACA24 snoRNA|ACA53 snoRNA', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118279190, end_position_on_the_genomic_accession=118279320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677810, CurrentID=0, Status='live', Symbol='SNORA26', Aliases='HBI-6', description='small nucleolar RNA, H/ACA box 26', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52713249, end_position_on_the_genomic_accession=52713370, orientation='plus', exon_count=1, OMIM=614626, no_hdr0=''),
  677811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677811, CurrentID=0, Status='live', Symbol='SNORA28', Aliases='ACA28', description='small nucleolar RNA, H/ACA box 28', other_designations='ACA28 snoRNA', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103337849, end_position_on_the_genomic_accession=103337974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677812, CurrentID=0, Status='live', Symbol='SNORA29', Aliases='ACA29', description='small nucleolar RNA, H/ACA box 29', other_designations='ACA29 snoRNA', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159785594, end_position_on_the_genomic_accession=159785733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677813, CurrentID=0, Status='live', Symbol='SNORA30', Aliases='ACA30', description='small nucleolar RNA, H/ACA box 30', other_designations='ACA30 snoRNA', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30710537, end_position_on_the_genomic_accession=30710665, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677814, CurrentID=0, Status='live', Symbol='SNORA31', Aliases='ACA31', description='small nucleolar RNA, H/ACA box 31', other_designations='ACA31 snoRNA', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45337480, end_position_on_the_genomic_accession=45337609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677815, CurrentID=0, Status='live', Symbol='SNORA2C', Aliases='ACA34, MIR1291, SNORA34', description='small nucleolar RNA, H/ACA box 2C', other_designations='ACA34 snoRNA|small nucleolar RNA, H/ACA box 34', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48654382, end_position_on_the_genomic_accession=48654518, orientation='minus', exon_count=1, OMIM=615487, no_hdr0=''),
  677816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677816, CurrentID=0, Status='live', Symbol='SNORA35', Aliases='HBI-36', description='small nucleolar RNA, H/ACA box 35', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114630797, end_position_on_the_genomic_accession=114630924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677817, CurrentID=0, Status='live', Symbol='SNORA36A', Aliases='ACA36', description='small nucleolar RNA, H/ACA box 36A', other_designations='ACA36 snoRNA', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154768528, end_position_on_the_genomic_accession=154768659, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677818, CurrentID=0, Status='live', Symbol='SNORA36B', Aliases='ACA36b', description='small nucleolar RNA, H/ACA box 36B', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220200546, end_position_on_the_genomic_accession=220200676, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677819, CurrentID=0, Status='live', Symbol='SNORA37', Aliases='ACA37', description='small nucleolar RNA, H/ACA box 37', other_designations='ACA37 snoRNA', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54222284, end_position_on_the_genomic_accession=54222412, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677820, CurrentID=0, Status='live', Symbol='SNORA38', Aliases='ACA38', description='small nucleolar RNA, H/ACA box 38', other_designations='ACA38 snoRNA', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31623079, end_position_on_the_genomic_accession=31623210, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677821, CurrentID=0, Status='live', Symbol='SNORA71E', Aliases='ACA39, SNORA39', description='small nucleolar RNA, H/ACA box 71E', other_designations='ACA39 snoRNA|small nucleolar RNA, H/ACA box 39', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38448083, end_position_on_the_genomic_accession=38448218, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677822, CurrentID=0, Status='live', Symbol='SNORA40', Aliases='ACA40', description='small nucleolar RNA, H/ACA box 40', other_designations='ACA40 snoRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93735110, end_position_on_the_genomic_accession=93735236, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677823, CurrentID=0, Status='live', Symbol='SNORA80E', Aliases='ACA42, SNORA42', description='small nucleolar RNA, H/ACA box 80E', other_designations='ACA42 snoRNA|small nucleolar RNA, H/ACA box 42', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155919909, end_position_on_the_genomic_accession=155920042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677824, CurrentID=0, Status='live', Symbol='SNORA17B', Aliases='ACA43, SNORA43', description='small nucleolar RNA, H/ACA box 17B', other_designations='ACA43 snoRNA|small nucleolar RNA, H/ACA box 43', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136726104, end_position_on_the_genomic_accession=136726237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677825, CurrentID=0, Status='live', Symbol='SNORA44', Aliases='ACA44', description='small nucleolar RNA, H/ACA box 44', other_designations='ACA44 snoRNA', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28580381, end_position_on_the_genomic_accession=28580512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677826, CurrentID=0, Status='live', Symbol='SNORA3B', Aliases='ACA3-2, SNORA45, SNORA45B', description='small nucleolar RNA, H/ACA box 3B', other_designations='ACA3-2 snoRNA|small nucleolar RNA, H/ACA box 45|small nucleolar RNA, H/ACA box 45B', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8685439, end_position_on_the_genomic_accession=8685569, orientation='plus', exon_count=1, OMIM=611333, no_hdr0=''),
  284611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284611, CurrentID=0, Status='live', Symbol='FAM102B', Aliases='SYM-3B', description='family with sequence similarity 102 member B', other_designations='family with sequence similarity 102, member B|sym-3 homolog B', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108560349, end_position_on_the_genomic_accession=108639327, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  284612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284612, CurrentID=0, Status='live', Symbol='SYPL2', Aliases='MG29', description='synaptophysin-like 2', other_designations='mitsugumin 29|mitsugumin-29', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109466478, end_position_on_the_genomic_accession=109482142, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  284613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284613, CurrentID=0, Status='live', Symbol='CYB561D1', Aliases='', description='cytochrome b561 family member D1', other_designations='cytochrome b-561 domain containing 1|cytochrome b561 family, member D1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109494036, end_position_on_the_genomic_accession=109500441, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  677830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677830, CurrentID=0, Status='live', Symbol='SNORA50A', Aliases='ACA50, SNORA50, SNORA76A', description='small nucleolar RNA, H/ACA box 50A', other_designations='ACA50 snoRNA|small nucleolar RNA, H/ACA box 50|small nucleolar RNA, H/ACA box 76A', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58559796, end_position_on_the_genomic_accession=58559931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284615, CurrentID=0, Status='live', Symbol='ANKRD34A', Aliases='ANKRD34', description='ankyrin repeat domain 34A', other_designations='ankyrin repeat domain 34', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145959441, end_position_on_the_genomic_accession=145964591, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  677832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677832, CurrentID=0, Status='live', Symbol='SNORA53', Aliases='ACA53', description='small nucleolar RNA, H/ACA box 53', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98599635, end_position_on_the_genomic_accession=98599884, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677833, CurrentID=0, Status='live', Symbol='SNORA54', Aliases='ACA54', description='small nucleolar RNA, H/ACA box 54', other_designations='ACA54 snoRNA', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2963771, end_position_on_the_genomic_accession=2963893, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677834, CurrentID=0, Status='live', Symbol='SNORA55', Aliases='ACA55', description='small nucleolar RNA, H/ACA box 55', other_designations='ACA55 snoRNA', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39567374, end_position_on_the_genomic_accession=39567510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677835, CurrentID=0, Status='live', Symbol='SNORA56', Aliases='ACA56', description='small nucleolar RNA, H/ACA box 56', other_designations='ACA56 snoRNA', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154774998, end_position_on_the_genomic_accession=154775126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677836, CurrentID=0, Status='live', Symbol='SNORA58', Aliases='ACA58', description='small nucleolar RNA, H/ACA box 58', other_designations='ACA58 snoRNA', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131478998, end_position_on_the_genomic_accession=131479233, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677837, CurrentID=0, Status='live', Symbol='SNORA60', Aliases='ACA60', description='small nucleolar RNA, H/ACA box 60', other_designations='ACA60 snoRNA', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38449369, end_position_on_the_genomic_accession=38449504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677838, CurrentID=0, Status='live', Symbol='SNORA61', Aliases='ACA61', description='small nucleolar RNA, H/ACA box 61', other_designations='ACA61 snoRNA', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28579764, end_position_on_the_genomic_accession=28579893, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677839, CurrentID=0, Status='live', Symbol='SNORA71C', Aliases='U71c', description='small nucleolar RNA, H/ACA box 71C', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38429667, end_position_on_the_genomic_accession=38429804, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677840, CurrentID=0, Status='live', Symbol='SNORA71D', Aliases='U71d', description='small nucleolar RNA, H/ACA box 71D', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38433862, end_position_on_the_genomic_accession=38433999, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677841, CurrentID=0, Status='live', Symbol='SNORA74B', Aliases='U19-2', description='small nucleolar RNA, H/ACA box 74B', other_designations='U19-2 snoRNA', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173020726, end_position_on_the_genomic_accession=173020929, orientation='plus', exon_count=1, OMIM=611331, no_hdr0=''),
  677842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677842, CurrentID=0, Status='live', Symbol='SNORA50C', Aliases='ACA62, SNORA76, SNORA76C', description='small nucleolar RNA, H/ACA box 50C', other_designations='small nucleolar RNA, H/ACA box 76|small nucleolar RNA, H/ACA box 76C', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64146339, end_position_on_the_genomic_accession=64146471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677843, CurrentID=0, Status='live', Symbol='SNORA77', Aliases='ACA63', description='small nucleolar RNA, H/ACA box 77', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203729581, end_position_on_the_genomic_accession=203729705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677844, CurrentID=0, Status='live', Symbol='SNORA78', Aliases='ACA64', description='small nucleolar RNA, H/ACA box 78', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1965184, end_position_on_the_genomic_accession=1965310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677845, CurrentID=0, Status='live', Symbol='SNORA79', Aliases='ACA65', description='small nucleolar RNA, H/ACA box 79', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81202695, end_position_on_the_genomic_accession=81202834, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677846, CurrentID=0, Status='live', Symbol='SNORA80A', Aliases='ACA67, SNORA80', description='small nucleolar RNA, H/ACA box 80A', other_designations='small nucleolar RNA, H/ACA box 80', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32377187, end_position_on_the_genomic_accession=32377322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  677847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677847, CurrentID=0, Status='live', Symbol='SNORA81', Aliases='HBI-61', description='small nucleolar RNA, H/ACA box 81', other_designations='HBI-61 snoRNA', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186786675, end_position_on_the_genomic_accession=186786852, orientation='plus', exon_count=1, OMIM=611334, no_hdr0=''),
  677848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677848, CurrentID=0, Status='live', Symbol='SNORD1A', Aliases='R38A, snR38A', description='small nucleolar RNA, C/D box 1A', other_designations='R38a snoRNA', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76561633, end_position_on_the_genomic_accession=76561706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677849, CurrentID=0, Status='live', Symbol='SNORD1B', Aliases='R38B, snR38B', description='small nucleolar RNA, C/D box 1B', other_designations='R38b snoRNA', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76561108, end_position_on_the_genomic_accession=76561193, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  153562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153562, CurrentID=0, Status='live', Symbol='MARVELD2', Aliases='DFNB49, MARVD2, MRVLDC2, Tric', description='MARVEL domain containing 2', other_designations='MARVEL (membrane-associating) domain containing 2|tricellulin', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69415112, end_position_on_the_genomic_accession=69442063, orientation='plus', exon_count=9, OMIM=610572, no_hdr0=''),
  100271238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271238, CurrentID=0, Status='live', Symbol='RPS26P37', Aliases='RPS26_12_999', description='ribosomal protein S26 pseudogene 37', other_designations='', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95123501, end_position_on_the_genomic_accession=95123937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  153571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153571, CurrentID=0, Status='live', Symbol='C5orf38', Aliases='CEI, IRX2NB', description='chromosome 5 open reading frame 38', other_designations='IRX2 neighbor|coordinated expression to IRX2|coordinated expression to IRXA2 homeobox|coordinated expression to IRXA2 protein', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2751943, end_position_on_the_genomic_accession=2755397, orientation='plus', exon_count=4, OMIM=610522, no_hdr0=''),
  153572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153572, CurrentID=0, Status='live', Symbol='IRX2', Aliases='IRXA2', description='iroquois homeobox 2', other_designations='homeodomain protein IRXA2|iroquois homeobox protein 2', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2741863, end_position_on_the_genomic_accession=2751655, orientation='minus', exon_count=6, OMIM=606198, no_hdr0=''),
  284648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284648, CurrentID=0, Status='live', Symbol='LOC284648', Aliases='', description='uncharacterized LOC284648', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182615140, end_position_on_the_genomic_accession=182616629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284649, CurrentID=0, Status='live', Symbol='SMG7-AS1', Aliases='DKFZP564C196', description='SMG7 antisense RNA 1', other_designations='SMG7 antisense RNA 1 (non-protein coding)', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183460874, end_position_on_the_genomic_accession=183471982, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  153579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153579, CurrentID=0, Status='live', Symbol='BTNL9', Aliases='BTN3, BTN8, VDLS1900', description='butyrophilin-like 9', other_designations='butyrophilin 3', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181040225, end_position_on_the_genomic_accession=181061523, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  284654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284654, CurrentID=0, Status='live', Symbol='RSPO1', Aliases='CRISTIN3, RSPO', description='R-spondin 1', other_designations='R-spondin homolog|roof plate-specific spondin-1', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37611350, end_position_on_the_genomic_accession=37634923, orientation='minus', exon_count=8, OMIM=609595, no_hdr0=''),
  284656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284656, CurrentID=0, Status='live', Symbol='EPHA10', Aliases='', description='EPH receptor A10', other_designations='EphA10s protein', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37713886, end_position_on_the_genomic_accession=37765152, orientation='minus', exon_count=21, OMIM=611123, no_hdr0=''),
  101060264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060264, CurrentID=0, Status='live', Symbol='FOXP4-AS1', Aliases='', description='FOXP4 antisense RNA 1', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41523895, end_position_on_the_genomic_accession=41548621, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284661, CurrentID=0, Status='live', Symbol='LOC284661', Aliases='', description='uncharacterized LOC284661', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4412051, end_position_on_the_genomic_accession=4424684, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  677880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677880, CurrentID=0, Status='live', Symbol='THEM4P1', Aliases='', description='thioesterase superfamily member 4 pseudogene 1', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22008658, end_position_on_the_genomic_accession=22009390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  677882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677882, CurrentID=0, Status='live', Symbol='SNORA59B', Aliases='ACA59', description='small nucleolar RNA, H/ACA box 59B', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19557560, end_position_on_the_genomic_accession=19557711, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  677883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677883, CurrentID=0, Status='live', Symbol='CHTF8P1', Aliases='CHTF8P, CTF8P', description='chromosome transmission fidelity factor 8 pseudogene 1', other_designations='CHTF8 pseudogene 1|CTF8, chromosome transmission fidelity factor 8 homolog pseudogene 1', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44633330, end_position_on_the_genomic_accession=44636065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271239, CurrentID=0, Status='live', Symbol='RPS26P33', Aliases='RPS26_14_825', description='ribosomal protein S26 pseudogene 33', other_designations='', map_location='7q21.3-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98478551, end_position_on_the_genomic_accession=98478796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  677885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677885, CurrentID=0, Status='live', Symbol='SNORA59A', Aliases='ACA59', description='small nucleolar RNA, H/ACA box 59A', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12507246, end_position_on_the_genomic_accession=12507397, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284672, CurrentID=0, Status='live', Symbol='PTGES3P1', Aliases='PTGES3P', description='prostaglandin E synthase 3 (cytosolic) pseudogene 1', other_designations='hCG2028557', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89103119, end_position_on_the_genomic_accession=89104954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284677, CurrentID=0, Status='live', Symbol='C1orf204', Aliases='', description='chromosome 1 open reading frame 204', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159834474, end_position_on_the_genomic_accession=159855347, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  284680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284680, CurrentID=0, Status='live', Symbol='C1orf111', Aliases='HSD20', description='chromosome 1 open reading frame 111', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162373725, end_position_on_the_genomic_accession=162376925, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284685, CurrentID=0, Status='live', Symbol='LOC284685', Aliases='', description='EWS RNA binding protein 1 pseudogene', other_designations='EWS RNA-binding protein 1 pseudogene|Ewing sarcoma breakpoint region 1 pseudogene', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166275454, end_position_on_the_genomic_accession=166277636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  134829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134829, CurrentID=0, Status='live', Symbol='CLVS2', Aliases='C6orf212, C6orf213, RLBP1L2, bA160A10.4', description='clavesin 2', other_designations='retinaldehyde binding protein 1-like 2', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122995971, end_position_on_the_genomic_accession=123072920, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  284688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284688, CurrentID=0, Status='live', Symbol='LINC01142', Aliases='', description='long intergenic non-protein coding RNA 1142', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170271405, end_position_on_the_genomic_accession=170284208, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480659, CurrentID=0, Status='live', Symbol='RNU6-1320P', Aliases='', description='RNA, U6 small nuclear 1320, pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373364, CurrentID=0, Status='live', Symbol='LOC105373364', Aliases='', description='uncharacterized LOC105373364', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1558527, end_position_on_the_genomic_accession=1564260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284695, CurrentID=0, Status='live', Symbol='ZNF326', Aliases='ZAN75, ZIRD, Zfp326, dJ871E2.1', description='zinc finger protein 326', other_designations='ZNF-protein interacting with nuclear mRNPs and DBC1|dJ871E2.1 (novel protein (ortholog of mouse zinc finger protein Zan75))|zinc finger protein interacting with mRNPs and DBC1', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89995119, end_position_on_the_genomic_accession=90028538, orientation='plus', exon_count=12, OMIM=614601, no_hdr0=''),
  284697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284697, CurrentID=0, Status='live', Symbol='BTBD8', Aliases='', description='BTB (POZ) domain containing 8', other_designations='double BTB/POZ domain containing protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92080305, end_position_on_the_genomic_accession=92147840, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100271240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271240, CurrentID=0, Status='live', Symbol='RPS26P36', Aliases='RPS26_18_953', description='ribosomal protein S26 pseudogene 36', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143065915, end_position_on_the_genomic_accession=143066163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270860, CurrentID=0, Status='live', Symbol='RPS2P16', Aliases='RPS2_8_310', description='ribosomal protein S2 pseudogene 16', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201762714, end_position_on_the_genomic_accession=201763610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929367, CurrentID=0, Status='live', Symbol='LOC101929367', Aliases='', description='uncharacterized LOC101929367', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201429195, end_position_on_the_genomic_accession=201430233, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  153642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153642, CurrentID=0, Status='live', Symbol='ARSK', Aliases='TSULF', description='arylsulfatase family member K', other_designations='ASK|arylsulfatase family, member K|telethon sulfatase', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95555111, end_position_on_the_genomic_accession=95605102, orientation='plus', exon_count=9, OMIM=610011, no_hdr0=''),
  153643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153643, CurrentID=0, Status='live', Symbol='FAM81B', Aliases='', description='family with sequence similarity 81 member B', other_designations='family with sequence similarity 81, member B', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95384802, end_position_on_the_genomic_accession=95450441, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  284716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284716, CurrentID=0, Status='live', Symbol='RIMKLA', Aliases='FAM80A, NAAGS, NAAGS-II', description='ribosomal modification protein rimK-like family member A', other_designations='N-acetylaspartyl-glutamate synthetase A|N-acetylaspartylglutamate synthetase II|N-acetylaspartylglutamylglutamate synthase A|NAAG synthetase A|RP11-157D18.1|family with sequence similarity 80, member A|ribosomal protein S6 modification-like protein A', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42380797, end_position_on_the_genomic_accession=42424229, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101060274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060274, CurrentID=0, Status='live', Symbol='LOC101060274', Aliases='', description='uncharacterized protein CXorf49-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373365, CurrentID=0, Status='live', Symbol='LOC105373365', Aliases='', description='uncharacterized LOC105373365', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1750737, end_position_on_the_genomic_accession=1753954, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79313, CurrentID=0, Status='live', Symbol='OR2AD1P', Aliases='OR2AD1, hs6M1-8P', description='olfactory receptor family 2 subfamily AD member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AD, member 1 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29026680, end_position_on_the_genomic_accession=29027607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284723, CurrentID=0, Status='live', Symbol='SLC25A34', Aliases='', description='solute carrier family 25 member 34', other_designations='solute carrier family 25, member 34', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15735088, end_position_on_the_genomic_accession=15741392, orientation='plus', exon_count=6, OMIM=610817, no_hdr0=''),
  100271241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271241, CurrentID=0, Status='live', Symbol='RPS29P15', Aliases='RPS29_10_826', description='ribosomal protein S29 pseudogene 15', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100928339, end_position_on_the_genomic_accession=100928595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  153657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153657, CurrentID=0, Status='live', Symbol='TTC23L', Aliases='MC25-1', description='tetratricopeptide repeat domain 23-like', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34839164, end_position_on_the_genomic_accession=34899650, orientation='plus', exon_count=18, OMIM=616344, no_hdr0=''),
  100873545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873545, CurrentID=0, Status='live', Symbol='RNA5SP292', Aliases='RN5S292', description='RNA, 5S ribosomal pseudogene 292', other_designations='RNA, 5S ribosomal 292', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106843749, end_position_on_the_genomic_accession=106843866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929759, CurrentID=0, Status='live', Symbol='LOC101929759', Aliases='', description='uncharacterized LOC101929759', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69834111, end_position_on_the_genomic_accession=69837829, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  284739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284739, CurrentID=0, Status='live', Symbol='LINC00176', Aliases='NCRNA00176, PRR17', description='long intergenic non-protein coding RNA 176', other_designations='proline rich 17', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64034344, end_position_on_the_genomic_accession=64039962, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100292680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100292680, CurrentID=0, Status='live', Symbol='LINC00942', Aliases='', description='long intergenic non-protein coding RNA 942', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1500491, end_position_on_the_genomic_accession=1504424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  284749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284749, CurrentID=0, Status='live', Symbol='LINC00494', Aliases='', description='long intergenic non-protein coding RNA 494', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48359911, end_position_on_the_genomic_accession=48370638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  91149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91149, CurrentID=0, Status='live', Symbol='RAPGEF4-AS1', Aliases='', description='RAPGEF4 antisense RNA 1', other_designations='RAPGEF4 antisense RNA 1 (non-protein coding)', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172723189, end_position_on_the_genomic_accession=172736206, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79314, CurrentID=0, Status='live', Symbol='OR4K6P', Aliases='', description='olfactory receptor family 4 subfamily K member 6 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 6 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19847267, end_position_on_the_genomic_accession=19848329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060280, CurrentID=0, Status='live', Symbol='SPDYE20P', Aliases='', description='speedy/RINGO cell cycle regulator family member E20, pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6715630, end_position_on_the_genomic_accession=6722048, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  153682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153682, CurrentID=0, Status='live', Symbol='LOC153682', Aliases='', description='uncharacterized LOC153682', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284756, CurrentID=0, Status='live', Symbol='C20orf197', Aliases='', description='chromosome 20 open reading frame 197', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60055925, end_position_on_the_genomic_accession=60072953, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284757, CurrentID=0, Status='live', Symbol='MIR646HG', Aliases='', description='MIR646 host gene', other_designations='MIR646 host gene (non-protein coding)', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60138492, end_position_on_the_genomic_accession=60322256, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284759, CurrentID=0, Status='live', Symbol='SIRPB2', Aliases='PTPN1L, PTPNS1L3, dJ776F14.2', description='signal-regulatory protein beta 2', other_designations='SIRP-beta-2|protein tyrosine phosphatase non-receptor type substrate 1-like 3|protein tyrosine phosphatase non-receptor type substrate protein|protein tyrosine phosphatase, non-receptor type substrate 1-like 3', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1470688, end_position_on_the_genomic_accession=1491587, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  91151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91151, CurrentID=0, Status='live', Symbol='TIGD7', Aliases='Sancho', description='tigger transposable element derived 7', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3298808, end_position_on_the_genomic_accession=3305439, orientation='minus', exon_count=2, OMIM=612969, no_hdr0=''),
  284764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284764, CurrentID=0, Status='live', Symbol='FTLP3', Aliases='FTLL1, dJ681N20.2', description='ferritin, light polypeptide pseudogene 3', other_designations='ferritin light polypeptide pseudogene|ferritin, light polypeptide-like 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4023718, end_position_on_the_genomic_accession=4024606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373367, CurrentID=0, Status='live', Symbol='LOC105373367', Aliases='', description='uncharacterized LOC105373367', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151512703, end_position_on_the_genomic_accession=151538489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271243, CurrentID=0, Status='live', Symbol='RPS2P36', Aliases='RPS2_13_1085', description='ribosomal protein S2 pseudogene 36', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96620877, end_position_on_the_genomic_accession=96621707, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284798, CurrentID=0, Status='live', Symbol='LOC284798', Aliases='', description='uncharacterized LOC284798', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25140798, end_position_on_the_genomic_accession=25148790, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  284800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284800, CurrentID=0, Status='live', Symbol='FAM182A', Aliases='C20orf91, bB329D4.1', description='family with sequence similarity 182 member A', other_designations='family with sequence similarity 182, member A', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=26054614, end_position_on_the_genomic_accession=26086917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  284801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284801, CurrentID=0, Status='live', Symbol='MIR663AHG', Aliases='', description='MIR663A host gene', other_designations='MIR663A host gene (non-protein coding)', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=26187019, end_position_on_the_genomic_accession=26209233, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  284802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284802, CurrentID=0, Status='live', Symbol='FRG1BP', Aliases='C20orf80, FRG1B, bA348I14.2', description='FSHD region gene 1 family member B, pseudogene', other_designations='FSHD region gene 1 family, member B', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30377203, end_position_on_the_genomic_accession=30399331, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  153733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153733, CurrentID=0, Status='live', Symbol='CCDC112', Aliases='MBC1', description='coiled-coil domain containing 112', other_designations='mutated in bladder cancer 1|mutated in bladder cancer protein 1', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115267188, end_position_on_the_genomic_accession=115296761, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  284808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284808, CurrentID=0, Status='live', Symbol='FLJ38773', Aliases='', description='uncharacterized protein FLJ38773', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=32565542, end_position_on_the_genomic_accession=32567957, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373368, CurrentID=0, Status='live', Symbol='LOC105373368', Aliases='', description='uncharacterized LOC105373368', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151904431, end_position_on_the_genomic_accession=151912859, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  153743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153743, CurrentID=0, Status='live', Symbol='PPP1R2P3', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 3', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 3|protein phosphatase inhibitor 2-like protein', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156850538, end_position_on_the_genomic_accession=156852528, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  153745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153745, CurrentID=0, Status='live', Symbol='FAM71B', Aliases='', description='family with sequence similarity 71 member B', other_designations='family with sequence similarity 71, member B', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157162332, end_position_on_the_genomic_accession=157166268, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271244, CurrentID=0, Status='live', Symbol='RPS2P39', Aliases='RPS2_15_1171', description='ribosomal protein S2 pseudogene 39', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108690289, end_position_on_the_genomic_accession=108690956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284821, CurrentID=0, Status='live', Symbol='RPL13AP7', Aliases='RPL13A_12_1722', description='ribosomal protein L13a pseudogene 7', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25361795, end_position_on_the_genomic_accession=25362924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873548, CurrentID=0, Status='live', Symbol='RNA5SP295', Aliases='RN5S295', description='RNA, 5S ribosomal pseudogene 295', other_designations='RNA, 5S ribosomal 295', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112252258, end_position_on_the_genomic_accession=112252378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284825, CurrentID=0, Status='live', Symbol='LOC284825', Aliases='', description='uncharacterized LOC284825', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28048414, end_position_on_the_genomic_accession=28137609, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284827, CurrentID=0, Status='live', Symbol='KRTAP13-4', Aliases='KAP13.4', description='keratin associated protein 13-4', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30430276, end_position_on_the_genomic_accession=30430758, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284835, CurrentID=0, Status='live', Symbol='LINC00323', Aliases='C21orf130, NCRNA00323, PRED42', description='long intergenic non-protein coding RNA 323', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41141500, end_position_on_the_genomic_accession=41148064, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  284836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284836, CurrentID=0, Status='live', Symbol='LINC00319', Aliases='C21orf125, NCRNA00319, PRED49', description='long intergenic non-protein coding RNA 319', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43450024, end_position_on_the_genomic_accession=43453893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284837, CurrentID=0, Status='live', Symbol='AATBC', Aliases='', description='apoptosis associated transcript in bladder cancer', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43805758, end_position_on_the_genomic_accession=43812567, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  153768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153768, CurrentID=0, Status='live', Symbol='PRELID2', Aliases='', description='PRELI domain containing 2', other_designations='CTC-806A22.1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145337932, end_position_on_the_genomic_accession=145835383, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  153769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153769, CurrentID=0, Status='live', Symbol='SH3RF2', Aliases='HEPP1, POSHER, PPP1R39, RNF158', description='SH3 domain containing ring finger 2', other_designations='POSH-eliminating RING protein|RING finger protein 158|SH3 domain-containing RING finger protein 2|heart protein phosphatase 1-binding protein|phosphatase 1 binding protein|protein phosphatase 1 regulatory subunit 39|protein phosphatase 1, regulatory subunit 39', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145936563, end_position_on_the_genomic_accession=146081773, orientation='plus', exon_count=12, OMIM=613377, no_hdr0=''),
  153770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153770, CurrentID=0, Status='live', Symbol='PLAC8L1', Aliases='', description='PLAC8-like 1', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146084149, end_position_on_the_genomic_accession=146104383, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100288703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288703, CurrentID=0, Status='live', Symbol='LOC100288703', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A)-like 4C pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231117640, end_position_on_the_genomic_accession=231118372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271245, CurrentID=0, Status='live', Symbol='RPS2P41', Aliases='RPS2_16_1278', description='ribosomal protein S2 pseudogene 41', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111879680, end_position_on_the_genomic_accession=111880253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284865, CurrentID=0, Status='live', Symbol='LOC284865', Aliases='', description='uncharacterized LOC284865', other_designations='XXbac-B444P24.8', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20198730, end_position_on_the_genomic_accession=20204918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3788, CurrentID=0, Status='live', Symbol='KCNS2', Aliases='KV9.2', description='potassium voltage-gated channel, modifier subfamily S, member 2', other_designations='delayed-rectifier K(+) channel alpha subunit 2|potassium voltage-gated channel, delayed-rectifier, subfamily S, member 2|voltage-gated potassium channel subunit Kv9.2', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98427022, end_position_on_the_genomic_accession=98430797, orientation='plus', exon_count=2, OMIM=602906, no_hdr0=''),
  153811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153811, CurrentID=0, Status='live', Symbol='LOC153811', Aliases='', description='uncharacterized LOC153811', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284889, CurrentID=0, Status='live', Symbol='MIF-AS1', Aliases='', description='MIF antisense RNA 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23893710, end_position_on_the_genomic_accession=23898930, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  134864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134864, CurrentID=0, Status='live', Symbol='TAAR1', Aliases='TA1, TAR1, TRAR1', description='trace amine associated receptor 1', other_designations='trace amine receptor 1', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132644984, end_position_on_the_genomic_accession=132646003, orientation='minus', exon_count=1, OMIM=609333, no_hdr0=''),
  284898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284898, CurrentID=0, Status='live', Symbol='LOC284898', Aliases='', description='uncharacterized LOC284898', other_designations='CTA-992D9.6', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27048144, end_position_on_the_genomic_accession=27060505, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  284900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284900, CurrentID=0, Status='live', Symbol='TTC28-AS1', Aliases='TTC28-AS, TTC28AS, dJ353E16.2', description='TTC28 antisense RNA 1', other_designations='TTC28 antisense RNA (non-protein coding)|TTC28 antisense RNA 1 (non-protein coding)', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27919376, end_position_on_the_genomic_accession=28002679, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  153830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153830, CurrentID=0, Status='live', Symbol='RNF145', Aliases='', description='ring finger protein 145', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159157409, end_position_on_the_genomic_accession=159210053, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  284904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284904, CurrentID=0, Status='live', Symbol='SEC14L4', Aliases='TAP3', description='SEC14-like lipid binding 4', other_designations='SEC14-like 4|SEC14p-like protein TAP3|tocopherol-associated protein 3', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30488890, end_position_on_the_genomic_accession=30505711, orientation='minus', exon_count=13, OMIM=612825, no_hdr0=''),
  100271247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271247, CurrentID=0, Status='live', Symbol='RPS2P42', Aliases='RPS2_19_1214', description='ribosomal protein S2 pseudogene 42', other_designations='', map_location='12p13.3-p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15439504, end_position_on_the_genomic_accession=15440450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284912, CurrentID=0, Status='live', Symbol='LOC284912', Aliases='', description='uncharacterized LOC284912', other_designations='CITF22-62D4.1', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35626988, end_position_on_the_genomic_accession=35637850, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  153842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153842, CurrentID=0, Status='live', Symbol='SHFM1P1', Aliases='DSS1P1, SHFM1P', description='split hand/foot malformation (ectrodactyly) type 1 pseudogene 1', other_designations='deleted in split hand/split foot 1 pseudogene|deleted in split-hand/split-foot 1pseudogene', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81892438, end_position_on_the_genomic_accession=81892891, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130210, CurrentID=0, Status='live', Symbol='LOC100130210', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56567909, end_position_on_the_genomic_accession=56568947, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100532731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532731, CurrentID=0, Status='live', Symbol='COMMD3-BMI1', Aliases='', description='COMMD3-BMI1 readthrough', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22316383, end_position_on_the_genomic_accession=22331485, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  284926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284926, CurrentID=0, Status='live', Symbol='LOC284926', Aliases='', description='uncharacterized LOC284926', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481960, CurrentID=0, Status='live', Symbol='RNU2-34P', Aliases='', description='RNA, U2 small nuclear 34, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284930, CurrentID=0, Status='live', Symbol='LOC284930', Aliases='', description='uncharacterized LOC284930', other_designations='', map_location='22q13.32', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47631674, end_position_on_the_genomic_accession=47855600, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  353323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353323, CurrentID=0, Status='live', Symbol='KRTAP12-2', Aliases='KAP12.2, KRTAP12.2', description='keratin associated protein 12-2', other_designations='high sulfur keratin-associated protein 12.2|keratin-associated protein 12.2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44666189, end_position_on_the_genomic_accession=44666927, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  284933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284933, CurrentID=0, Status='live', Symbol='LOC284933', Aliases='', description='uncharacterized LOC284933', other_designations='', map_location='22q13.32', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48538900, end_position_on_the_genomic_accession=48547387, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22794, CurrentID=0, Status='live', Symbol='CASC3', Aliases='BTZ, MLN51', description='cancer susceptibility candidate 3', other_designations='MLN 51|barentsz|cancer susceptibility candidate gene 3 protein|metastatic lymph node 51|metastatic lymph node gene 51 protein|protein barentsz', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40140254, end_position_on_the_genomic_accession=40172178, orientation='plus', exon_count=14, OMIM=606504, no_hdr0=''),
  22795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22795, CurrentID=0, Status='live', Symbol='NID2', Aliases='NID-2', description='nidogen 2', other_designations='nidogen 2 (osteonidogen)|osteonidogen', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52004802, end_position_on_the_genomic_accession=52069231, orientation='minus', exon_count=21, OMIM=605399, no_hdr0=''),
  22796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22796, CurrentID=0, Status='live', Symbol='COG2', Aliases='LDLC', description='component of oligomeric golgi complex 2', other_designations='COG complex subunit 2|brefeldin A-sensitive, peripheral Golgi protein|conserved oligomeric Golgi complex protein 2|low density lipoprotein receptor defect C complementing|low density lipoprotein receptor defect C-complementing protein', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230642456, end_position_on_the_genomic_accession=230693985, orientation='plus', exon_count=18, OMIM=606974, no_hdr0=''),
  22797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22797, CurrentID=0, Status='live', Symbol='TFEC', Aliases='TCFEC, TFE-C-L, TFECL, bHLHe34, hTFEC-L, TFEC', description='transcription factor EC', other_designations='class E basic helix-loop-helix protein 34', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115935148, end_position_on_the_genomic_accession=116159891, orientation='minus', exon_count=17, OMIM=604732, no_hdr0=''),
  22798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22798, CurrentID=0, Status='live', Symbol='LAMB4', Aliases='', description='laminin subunit beta 4', other_designations='laminin beta-1-related protein', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108003986, end_position_on_the_genomic_accession=108130356, orientation='minus', exon_count=42, OMIM=616380, no_hdr0=''),
  646708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646708, CurrentID=0, Status='live', Symbol='LOC646708', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C5 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5 pseudogene', map_location='8p21.3-p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23481185, end_position_on_the_genomic_accession=23483685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22800, CurrentID=0, Status='live', Symbol='RRAS2', Aliases='TC21', description='related RAS viral (r-ras) oncogene homolog 2', other_designations='ras-like protein TC21|teratocarcinoma oncogene', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14277920, end_position_on_the_genomic_accession=14364506, orientation='minus', exon_count=8, OMIM=600098, no_hdr0=''),
  22801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22801, CurrentID=0, Status='live', Symbol='ITGA11', Aliases='HsT18964', description='integrin subunit alpha 11', other_designations='integrin alpha 11|integrin, alpha 11', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68297428, end_position_on_the_genomic_accession=68432160, orientation='minus', exon_count=31, OMIM=604789, no_hdr0=''),
  22802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22802, CurrentID=0, Status='live', Symbol='CLCA4', Aliases='CaCC, CaCC2', description='chloride channel accessory 4', other_designations='caCC-2|calcium-activated chloride channel family member 4|calcium-activated chloride channel protein 2|chloride channel regulator 4|chloride channel, calcium activated, family member 4|hCLCA4|hCaCC-2', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86547076, end_position_on_the_genomic_accession=86580758, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22803, CurrentID=0, Status='live', Symbol='XRN2', Aliases='', description="5'-3' exoribonuclease 2", other_designations='DHP protein|Dhm1-like protein', map_location='20p11.2-p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21303304, end_position_on_the_genomic_accession=21389825, orientation='plus', exon_count=31, OMIM=608851, no_hdr0=''),
  284948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284948, CurrentID=0, Status='live', Symbol='SH2D6', Aliases='', description='SH2 domain containing 6', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85418732, end_position_on_the_genomic_accession=85437029, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  22806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22806, CurrentID=0, Status='live', Symbol='IKZF3', Aliases='AIO, AIOLOS, ZNFN1A3', description='IKAROS family zinc finger 3', other_designations='IKAROS family zinc finger 3 (Aiolos)|zinc finger DNA binding protein Aiolos|zinc finger protein, subfamily 1A, 3 (Aiolos)', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39757715, end_position_on_the_genomic_accession=39864188, orientation='minus', exon_count=9, OMIM=606221, no_hdr0=''),
  22807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22807, CurrentID=0, Status='live', Symbol='IKZF2', Aliases='ANF1A2, HELIOS, ZNF1A2, ZNFN1A2', description='IKAROS family zinc finger 2', other_designations='IKAROS family zinc finger 2 (Helios)|ikaros family zinc finger protein 2|zinc finger DNA binding protein Helios|zinc finger protein, subfamily 1A, 2 (Helios)', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212999684, end_position_on_the_genomic_accession=213154266, orientation='minus', exon_count=13, OMIM=606234, no_hdr0=''),
  22808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22808, CurrentID=0, Status='live', Symbol='MRAS', Aliases='M-RAs, R-RAS3, RRAS3', description='muscle RAS oncogene homolog', other_designations='muscle and microspikes RAS|ras-related protein R-Ras3', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138347648, end_position_on_the_genomic_accession=138405535, orientation='plus', exon_count=9, OMIM=608435, no_hdr0=''),
  22809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22809, CurrentID=0, Status='live', Symbol='ATF5', Aliases='ATFX, HMFN0395', description='activating transcription factor 5', other_designations='cAMP-dependent transcription factor ATF-5|transcription factor ATFx', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49928702, end_position_on_the_genomic_accession=49933936, orientation='plus', exon_count=5, OMIM=606398, no_hdr0=''),
  22811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22811, CurrentID=0, Status='live', Symbol='MBS3', Aliases='HCFP2', description='Moebius syndrome 3', other_designations='hereditary congenital facial paresis 2', map_location='10q21.3-q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604185, no_hdr0=''),
  22813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22813, CurrentID=0, Status='live', Symbol='TDGF1P6', Aliases='CR-6, CRIPTO-6, TDGF6', description='teratocarcinoma-derived growth factor 1 pseudogene 6', other_designations='teratocarcinoma-derived growth factor 6, pseudogene', map_location='3q22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136155548, end_position_on_the_genomic_accession=136157044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22814, CurrentID=0, Status='live', Symbol='TDGF1P5', Aliases='CR-5, CRIPTO-5, TDGF5', description='teratocarcinoma-derived growth factor 1 pseudogene 5', other_designations='teratocarcinoma-derived growth factor 5, pseudogene', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54072991, end_position_on_the_genomic_accession=54073937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22815, CurrentID=0, Status='live', Symbol='TDGF1P4', Aliases='CR-4, CRIPTO-4, TDGF4', description='teratocarcinoma-derived growth factor 1 pseudogene 4', other_designations='teratocarcinoma-derived growth factor 4 (pseudogene)|teratocarcinoma-derived growth factor 4, pseudogene', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3941278, end_position_on_the_genomic_accession=3942020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22816, CurrentID=0, Status='live', Symbol='TDGF1P2', Aliases='CR-2, CRIPTO-2, TDGF2', description='teratocarcinoma-derived growth factor 1 pseudogene 2', other_designations='teratocarcinoma-derived growth factor 2, pseudogene', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227869170, end_position_on_the_genomic_accession=227870585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22817, CurrentID=0, Status='live', Symbol='NR1I4', Aliases='CAR2', description='nuclear receptor subfamily 1 group I member 4', other_designations='nuclear receptor subfamily 1, group I, member 4', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22818, CurrentID=0, Status='live', Symbol='COPZ1', Aliases='CGI-120, COPZ, HSPC181, zeta-COP, zeta1-COP', description='coatomer protein complex subunit zeta 1', other_designations='coatomer protein complex, subunit zeta 1|zeta-1 COP|zeta-1-coat protein', map_location='12q13.2-q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54325090, end_position_on_the_genomic_accession=54351851, orientation='plus', exon_count=9, OMIM=615472, no_hdr0=''),
  22820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22820, CurrentID=0, Status='live', Symbol='COPG1', Aliases='COPG', description='coatomer protein complex subunit gamma 1', other_designations='coat protein gamma-cop|coatomer protein complex, subunit gamma 1|coatomer subunit gamma|gamma-1-COP|gamma-1-coat protein|gamma-COP|gamma-coat protein', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129249575, end_position_on_the_genomic_accession=129277773, orientation='plus', exon_count=25, OMIM=615525, no_hdr0=''),
  22821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22821, CurrentID=0, Status='live', Symbol='RASA3', Aliases='GAP1IP4BP, GAPIII', description='RAS p21 protein activator 3', other_designations='GAP1(IP4BP)|GTPase activating protein III|Ins(1,3,4,5)P4-binding protein|inositol 1,3,4,5-tetrakisphosphate-binding protein|ins P4-binding protein', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113977783, end_position_on_the_genomic_accession=114132620, orientation='minus', exon_count=28, OMIM=605182, no_hdr0=''),
  22822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22822, CurrentID=0, Status='live', Symbol='PHLDA1', Aliases='DT1P1B11, PHRIP, TDAG51', description='pleckstrin homology-like domain, family A, member 1', other_designations='PQ-rich protein|PQR protein|T-cell death-associated gene 51 protein|apoptosis-associated nuclear protein|proline- and glutamine-rich protein|proline- and histidine-rich protein|proline-histidine rich protein', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76025447, end_position_on_the_genomic_accession=76031776, orientation='minus', exon_count=2, OMIM=605335, no_hdr0=''),
  22823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22823, CurrentID=0, Status='live', Symbol='MTF2', Aliases='M96, PCL2, TDRD19A, dJ976O13.2', description='metal response element binding transcription factor 2', other_designations='hPCl2|metal regulatory transcription factor 2|metal response element-binding transcription factor 2|metal-response element DNA-binding protein M96|polycomb-like 2|polycomb-like protein 2|putative DNA binding protein|tudor domain containing 19A', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93079235, end_position_on_the_genomic_accession=93139081, orientation='plus', exon_count=16, OMIM=609882, no_hdr0=''),
  22824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22824, CurrentID=0, Status='live', Symbol='HSPA4L', Aliases='APG-1, HSPH3, Osp94', description='heat shock protein family A (Hsp70) member 4 like', other_designations='heat shock 70 kDa protein 4-like protein|heat shock 70-related protein APG-1|heat shock 70kDa protein 4-like|heat shock protein (hsp110 family)|osmotic stress protein 94', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127782209, end_position_on_the_genomic_accession=127834073, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  22826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22826, CurrentID=0, Status='live', Symbol='DNAJC8', Aliases='HSPC331, SPF31', description='DnaJ heat shock protein family (Hsp40) member C8', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8|splicing protein spf31', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28200278, end_position_on_the_genomic_accession=28233031, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22827, CurrentID=0, Status='live', Symbol='PUF60', Aliases='FIR, RoBPI, SIAHBP1, VRJS', description='poly(U) binding splicing factor 60KDa', other_designations='FBP interacting repressor|FUSE-binding protein-interacting repressor|Ro ribonucleoprotein-binding protein 1|Siah binding protein 1|poly-U binding splicing factor 60KDa|pyrimidine tract binding splicing factor', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143816344, end_position_on_the_genomic_accession=143829386, orientation='minus', exon_count=14, OMIM=604819, no_hdr0=''),
  22828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22828, CurrentID=0, Status='live', Symbol='SCAF8', Aliases='RBM16', description='SR-related CTD-associated factor 8', other_designations='CDC5L complex-associated protein 7|RNA-binding motif protein 16|SR-like CTD-associated factor 8|SR-related and CTD-associated factor 8', map_location='6q25.1-q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154733378, end_position_on_the_genomic_accession=154834244, orientation='plus', exon_count=22, OMIM=616024, no_hdr0=''),
  22829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22829, CurrentID=0, Status='live', Symbol='NLGN4Y', Aliases='', description='neuroligin 4, Y-linked', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14522608, end_position_on_the_genomic_accession=14844945, orientation='plus', exon_count=16, OMIM=400028, no_hdr0=''),
  22832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22832, CurrentID=0, Status='live', Symbol='CEP162', Aliases='C6orf84, KIAA1009, QN1', description='centrosomal protein 162kDa', other_designations='protein QN1 homolog', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84124241, end_position_on_the_genomic_accession=84227890, orientation='minus', exon_count=32, OMIM=610201, no_hdr0=''),
  22833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22833, CurrentID=0, Status='live', Symbol='KIAA0894', Aliases='', description='KIAA0894 protein', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22834, CurrentID=0, Status='live', Symbol='ZNF652', Aliases='', description='zinc finger protein 652', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49289206, end_position_on_the_genomic_accession=49362473, orientation='minus', exon_count=8, OMIM=613907, no_hdr0=''),
  22835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22835, CurrentID=0, Status='live', Symbol='ZFP30', Aliases='ZNF745', description='ZFP30 zinc finger protein', other_designations='zfp-30|zinc finger protein 745', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37632487, end_position_on_the_genomic_accession=37657285, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22836, CurrentID=0, Status='live', Symbol='RHOBTB3', Aliases='', description='Rho-related BTB domain containing 3', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95731066, end_position_on_the_genomic_accession=95796367, orientation='plus', exon_count=12, OMIM=607353, no_hdr0=''),
  22837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22837, CurrentID=0, Status='live', Symbol='COBLL1', Aliases='COBLR1', description='cordon-bleu WH2 repeat protein like 1', other_designations='COBL-like 1|cordon-bleu WH2 repeat protein-like 1', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164680185, end_position_on_the_genomic_accession=164842168, orientation='minus', exon_count=17, OMIM=610318, no_hdr0=''),
  22838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22838, CurrentID=0, Status='live', Symbol='RNF44', Aliases='', description='ring finger protein 44', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176526699, end_position_on_the_genomic_accession=176543413, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  22839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22839, CurrentID=0, Status='live', Symbol='DLGAP4', Aliases='DAP-4, DAP4, DLP4, SAPAP-4, SAPAP4', description='discs, large (Drosophila) homolog-associated protein 4', other_designations='PSD-95/SAP90-binding protein 4|SAP90/PSD-95-associated protein 4', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36306323, end_position_on_the_genomic_accession=36528637, orientation='plus', exon_count=17, OMIM=616191, no_hdr0=''),
  22841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22841, CurrentID=0, Status='live', Symbol='RAB11FIP2', Aliases='Rab11-FIP2, nRip11', description='RAB11 family interacting protein 2 (class I)', other_designations='RAB11-FIP2 long isoform', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118004916, end_position_on_the_genomic_accession=118046877, orientation='minus', exon_count=7, OMIM=608599, no_hdr0=''),
  100292922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100292922, CurrentID=0, Status='live', Symbol='LOC100292922', Aliases='', description='putative ankyrin repeat domain-containing protein ENSP00000330211', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16132469, end_position_on_the_genomic_accession=16154854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22843, CurrentID=0, Status='live', Symbol='PPM1E', Aliases='CaMKP-N, POPX1, PP2CH, caMKN', description='protein phosphatase, Mg2+/Mn2+ dependent 1E', other_designations='ca(2+)/calmodulin-dependent protein kinase phosphatase N|caMKP-nucleus|nuclear calmodulin-dependent protein kinase phosphatase|partner of PIX 1|partner of PIX-alpha|partner of PIXA|protein phosphatase 1E (PP2C domain containing)|protein phosphatase, Mg2+/Mn2+ dependent, 1E', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58755869, end_position_on_the_genomic_accession=58985179, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22844, CurrentID=0, Status='live', Symbol='FRMPD1', Aliases='FRMD2', description='FERM and PDZ domain containing 1', other_designations='FERM domain-containing protein 2', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37602978, end_position_on_the_genomic_accession=37746904, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  22845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22845, CurrentID=0, Status='live', Symbol='DOLK', Aliases='CDG1M, DK, DK1, SEC59, TMEM15', description='dolichol kinase', other_designations='SEC59 homolog|dolichol kinase 1|transmembrane protein 15', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128945530, end_position_on_the_genomic_accession=128947733, orientation='minus', exon_count=1, OMIM=610746, no_hdr0=''),
  153918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153918, CurrentID=0, Status='live', Symbol='ZC2HC1B', Aliases='C6orf94, FAM164B, dJ468K18.5', description='zinc finger, C2HC-type containing 1B', other_designations='family with sequence similarity 164, member B|protein FAM164B', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143864436, end_position_on_the_genomic_accession=143938346, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22847, CurrentID=0, Status='live', Symbol='ZNF507', Aliases='', description='zinc finger protein 507', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32345608, end_position_on_the_genomic_accession=32387667, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  284992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284992, CurrentID=0, Status='live', Symbol='CCDC150', Aliases='', description='coiled-coil domain containing 150', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196639632, end_position_on_the_genomic_accession=196732806, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  22849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22849, CurrentID=0, Status='live', Symbol='CPEB3', Aliases='', description='cytoplasmic polyadenylation element binding protein 3', other_designations='CPE-BP3|CPE-binding protein 3|hCPEB-3', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92048638, end_position_on_the_genomic_accession=92291118, orientation='minus', exon_count=14, OMIM=610606, no_hdr0=''),
  22850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22850, CurrentID=0, Status='live', Symbol='ADNP2', Aliases='ZNF508', description='ADNP homeobox 2', other_designations='zinc finger protein 508', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80109031, end_position_on_the_genomic_accession=80140348, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22852, CurrentID=0, Status='live', Symbol='ANKRD26', Aliases='THC2, bA145E8.1', description='ankyrin repeat domain 26', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26966584, end_position_on_the_genomic_accession=27100498, orientation='minus', exon_count=51, OMIM=610855, no_hdr0=''),
  22853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22853, CurrentID=0, Status='live', Symbol='LMTK2', Aliases='AATYK2, BREK, KPI-2, KPI2, LMR2, PPP1R100, cprk', description='lemur tyrosine kinase 2', other_designations='CDK5/p35-regulated kinase|apoptosis-associated tyrosine kinase 2|brain-enriched kinase|cyclin-dependent kinase 5/p35-regulated kinase|hBREK|kinase phosphatase inhibitor 2|kinase/phosphatase/inhibitor 2|protein phosphatase 1, regulatory subunit 100|serine/threonine-protein kinase KPI-2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98106875, end_position_on_the_genomic_accession=98209636, orientation='plus', exon_count=14, OMIM=610989, no_hdr0=''),
  22854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22854, CurrentID=0, Status='live', Symbol='NTNG1', Aliases='Lmnt1', description='netrin G1', other_designations='axon guidance molecule|laminet 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107139918, end_position_on_the_genomic_accession=107484899, orientation='plus', exon_count=11, OMIM=608818, no_hdr0=''),
  101060321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060321, CurrentID=0, Status='live', Symbol='TBC1D3G', Aliases='', description='TBC1 domain family member 3G', other_designations='TBC1 domain family member-like|TBC1 domain family, member 3G', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36322242, end_position_on_the_genomic_accession=36334759, orientation='plus', exon_count=14, OMIM=610810, no_hdr0=''),
  22856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22856, CurrentID=0, Status='live', Symbol='CHSY1', Aliases='CHSY, CSS1, ChSy-1, TPBS', description='chondroitin sulfate synthase 1', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase 1|N-acetylgalactosaminyltransferase II|carbohydrate synthase 1|chondroitin glucuronyltransferase 1|chondroitin glucuronyltransferase II|chondroitin synthase 1|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase 1', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101175723, end_position_on_the_genomic_accession=101251932, orientation='minus', exon_count=6, OMIM=608183, no_hdr0=''),
  22858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22858, CurrentID=0, Status='live', Symbol='ICK', Aliases='ECO, LCK2, MRK', description='intestinal cell (MAK-like) kinase', other_designations='MAK-related kinase|hICK|laryngeal cancer kinase 2|serine/threonine protein kinase', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53001299, end_position_on_the_genomic_accession=53061833, orientation='minus', exon_count=16, OMIM=612325, no_hdr0=''),
  22859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22859, CurrentID=0, Status='live', Symbol='ADGRL1', Aliases='CIRL1, CL1, LEC2, LPHN1', description='adhesion G protein-coupled receptor L1', other_designations='CIRL-1|calcium-independent alpha-latrotoxin receptor 1|latrophilin 1|latrophilin-1|lectomedin-2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14147737, end_position_on_the_genomic_accession=14206248, orientation='minus', exon_count=27, OMIM=616416, no_hdr0=''),
  153932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153932, CurrentID=0, Status='live', Symbol='SRP72P2', Aliases='dJ274L14.2', description='signal recognition particle 72kDa pseudogene 2', other_designations='signal recognition particle 72kD pseudogene', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158237136, end_position_on_the_genomic_accession=158242784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22861, CurrentID=0, Status='live', Symbol='NLRP1', Aliases='CARD7, CIDED, CLR17.1, DEFCAP, DEFCAP-L/S, NAC, NALP1, PP1044, SLEV1, VAMAS1', description='NLR family, pyrin domain containing 1', other_designations='NACHT, LRR and PYD containing protein 1|NACHT, leucine rich repeat and PYD (pyrin domain) containing 1|NACHT, leucine rich repeat and PYD containing 1|caspase recruitment domain protein 7|caspase recruitment domain-containing protein 7|death effector filament-forming Ced-4-like apoptosis protein|nucleotide-binding domain and caspase recruitment domain|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5501399, end_position_on_the_genomic_accession=5584512, orientation='minus', exon_count=18, OMIM=606636, no_hdr0=''),
  22862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22862, CurrentID=0, Status='live', Symbol='FNDC3A', Aliases='FNDC3, HUGO, bA203I16.1, bA203I16.5', description='fibronectin type III domain containing 3A', other_designations='human gene expressed in odontoblasts', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48975912, end_position_on_the_genomic_accession=49209779, orientation='plus', exon_count=30, OMIM=615794, no_hdr0=''),
  22863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22863, CurrentID=0, Status='live', Symbol='ATG14', Aliases='ATG14L, BARKOR, KIAA0831', description='autophagy related 14', other_designations='ATG14 autophagy related 14 homolog|Beclin 1-Interacting protein|autophagy-related protein 14-like protein', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55366391, end_position_on_the_genomic_accession=55411880, orientation='minus', exon_count=12, OMIM=613515, no_hdr0=''),
  22864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22864, CurrentID=0, Status='live', Symbol='R3HDM2', Aliases='CAG6, PR01365', description='R3H domain containing 2', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57253764, end_position_on_the_genomic_accession=57431000, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  22865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22865, CurrentID=0, Status='live', Symbol='SLITRK3', Aliases='', description='SLIT and NTRK like family member 3', other_designations='SLIT and NTRK-like family, member 3|slit and trk like gene 3', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=165186720, end_position_on_the_genomic_accession=165196681, orientation='minus', exon_count=4, OMIM=609679, no_hdr0=''),
  22866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22866, CurrentID=0, Status='live', Symbol='CNKSR2', Aliases='CNK2, KSR2, MAGUIN', description='connector enhancer of kinase suppressor of Ras 2', other_designations='CNK homolog protein 2|connector enhancer of KSR 2|connector enhancer of KSR2|membrane-associated guanylate kinase-interacting protein', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21374414, end_position_on_the_genomic_accession=21654695, orientation='plus', exon_count=23, OMIM=300724, no_hdr0=''),
  22868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22868, CurrentID=0, Status='live', Symbol='FASTKD2', Aliases='KIAA0971', description='FAST kinase domains 2', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206765388, end_position_on_the_genomic_accession=206796189, orientation='plus', exon_count=12, OMIM=612322, no_hdr0=''),
  22869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22869, CurrentID=0, Status='live', Symbol='ZNF510', Aliases='', description='zinc finger protein 510', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96754553, end_position_on_the_genomic_accession=96778119, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22870, CurrentID=0, Status='live', Symbol='PPP6R1', Aliases='KIAA1115, PP6R1, SAP190, SAPS1', description='protein phosphatase 6 regulatory subunit 1', other_designations='SAPS domain family, member 1|protein phosphatase 6, regulatory subunit 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55229779, end_position_on_the_genomic_accession=55258670, orientation='minus', exon_count=24, OMIM=610875, no_hdr0=''),
  22871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22871, CurrentID=0, Status='live', Symbol='NLGN1', Aliases='NL1', description='neuroligin 1', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173396490, end_position_on_the_genomic_accession=174286644, orientation='plus', exon_count=15, OMIM=600568, no_hdr0=''),
  285016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285016, CurrentID=0, Status='live', Symbol='FAM150B', Aliases='PRO1097, RGPG542', description='family with sequence similarity 150 member B', other_designations='family with sequence similarity 150, member B', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=279561, end_position_on_the_genomic_accession=288880, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22873, CurrentID=0, Status='live', Symbol='DZIP1', Aliases='DZIP, DZIPt1', description='DAZ interacting zinc finger protein 1', other_designations='DAZ interacting protein 1|DAZ interacting protein testis1|DAZ-interacting protein 1/2|zinc finger DAZ interacting protein 1|zinc-finger protein DZIPt1', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95578202, end_position_on_the_genomic_accession=95644706, orientation='minus', exon_count=25, OMIM=608671, no_hdr0=''),
  22874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22874, CurrentID=0, Status='live', Symbol='PLEKHA6', Aliases='PEPP-3, PEPP3', description='pleckstrin homology domain containing A6', other_designations='PH domain-containing family A member 6|phosphoinositol 3-phosphate-binding protein 3|phosphoinositol 3-phosphate-binding protein-3|pleckstrin homology domain containing, family A member 6', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204218851, end_position_on_the_genomic_accession=204377763, orientation='minus', exon_count=39, OMIM=607771, no_hdr0=''),
  22875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22875, CurrentID=0, Status='live', Symbol='ENPP4', Aliases='NPP4', description='ectonucleotide pyrophosphatase/phosphodiesterase 4 (putative)', other_designations='AP3A hydrolase|AP3Aase|E-NPP 4|NPP-4|ectonucleotide pyrophosphatase/phosphodiesterase 4 (putative function)|ectonucleotide pyrophosphatase/phosphodiesterase family member 4', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46129964, end_position_on_the_genomic_accession=46146699, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22876, CurrentID=0, Status='live', Symbol='INPP5F', Aliases='MSTP007, MSTPO47, SAC2, hSAC2', description='inositol polyphosphate-5-phosphatase F', other_designations='Sac domain-containing inositol phosphatase 2|sac domain-containing phosphoinositide 5-phosphatase 2', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119704395, end_position_on_the_genomic_accession=119829278, orientation='plus', exon_count=26, OMIM=609389, no_hdr0=''),
  22877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22877, CurrentID=0, Status='live', Symbol='MLXIP', Aliases='MIR, MONDOA, bHLHe36', description='MLX interacting protein', other_designations='Mlx interactor|class E basic helix-loop-helix protein 36|transcriptional activator MondoA', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122078728, end_position_on_the_genomic_accession=122147347, orientation='plus', exon_count=19, OMIM=608090, no_hdr0=''),
  22878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22878, CurrentID=0, Status='live', Symbol='TRAPPC8', Aliases='GSG1, HsT2706, KIAA1012, TRS85', description='trafficking protein particle complex 8', other_designations='general sporulation gene 1 homolog|protein TRS85 homolog', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31829173, end_position_on_the_genomic_accession=31943212, orientation='minus', exon_count=31, OMIM=614136, no_hdr0=''),
  22879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22879, CurrentID=0, Status='live', Symbol='MON1B', Aliases='HSRG1, SAND2, SRG1', description='MON1 homolog B, secretory trafficking associated', other_designations='HSV-1 stimulation-related gene 1 protein|HSV-I stimulating-related protein|MON1 homolog B|MON1 secretory trafficking family member B', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77190919, end_position_on_the_genomic_accession=77199646, orientation='plus', exon_count=6, OMIM=608954, no_hdr0=''),
  22880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22880, CurrentID=0, Status='live', Symbol='MORC2', Aliases='ZCW3, ZCWCC1', description='MORC family CW-type zinc finger 2', other_designations='zinc finger CW-type coiled-coil domain protein 1|zinc finger, CW type with coiled-coil domain 1|zinc finger, CW-type with coiled-coil domain 1', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30925138, end_position_on_the_genomic_accession=30968497, orientation='minus', exon_count=27, OMIM=616661, no_hdr0=''),
  285025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285025, CurrentID=0, Status='live', Symbol='CCDC141', Aliases='CAMDI', description='coiled-coil domain containing 141', other_designations='coiled-coil protein associated with myosin II and DISC1', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178829757, end_position_on_the_genomic_accession=179050118, orientation='minus', exon_count=27, OMIM=616031, no_hdr0=''),
  22882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22882, CurrentID=0, Status='live', Symbol='ZHX2', Aliases='AFR1, RAF', description='zinc fingers and homeoboxes 2', other_designations='AFP regulator 1|alpha-fetoprotein regulator 1|regulator of AFP|transcription factor ZHX2|zinc finger and homeodomain protein 2|zinc-fingers and homeoboxes 2', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122781349, end_position_on_the_genomic_accession=122974515, orientation='plus', exon_count=7, OMIM=609185, no_hdr0=''),
  22883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22883, CurrentID=0, Status='live', Symbol='CLSTN1', Aliases='ALC-ALPHA, CDHR12, CST-1, CSTN1, PIK3CD, XB31alpha, alcalpha1, alcalpha2', description='calsyntenin 1', other_designations='alcadein-alpha|alzheimer-related cadherin-like protein|cadherin-related family member 12|non-classical cadherin XB31alpha', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9729021, end_position_on_the_genomic_accession=9824526, orientation='minus', exon_count=19, OMIM=611321, no_hdr0=''),
  22884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22884, CurrentID=0, Status='live', Symbol='WDR37', Aliases='', description='WD repeat domain 37', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1056836, end_position_on_the_genomic_accession=1132297, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22885, CurrentID=0, Status='live', Symbol='ABLIM3', Aliases='HMFN1661', description='actin binding LIM protein family member 3', other_designations='abLIM-3|actin binding LIM protein family, member 3|actin-binding LIM protein family member 3', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149141483, end_position_on_the_genomic_accession=149260439, orientation='plus', exon_count=25, OMIM=611305, no_hdr0=''),
  22887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22887, CurrentID=0, Status='live', Symbol='FOXJ3', Aliases='', description='forkhead box J3', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42176539, end_position_on_the_genomic_accession=42335877, orientation='minus', exon_count=18, OMIM=616035, no_hdr0=''),
  22888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22888, CurrentID=0, Status='live', Symbol='UBOX5', Aliases='RNF37, UBCE7IP5, UIP5, hUIP5', description='U-box domain containing 5', other_designations='U-box domain-containing protein 5|ubcM4-interacting protein 5|ubiquitin conjugating enzyme 7 interacting protein 5|ubiquitin-conjugating enzyme 7-interacting protein 5', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3107573, end_position_on_the_genomic_accession=3159910, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22889, CurrentID=0, Status='live', Symbol='KIAA0907', Aliases='BLOM7', description='KIAA0907', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155913043, end_position_on_the_genomic_accession=155934442, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  22890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22890, CurrentID=0, Status='live', Symbol='ZBTB1', Aliases='ZNF909', description='zinc finger and BTB domain containing 1', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64503950, end_position_on_the_genomic_accession=64533690, orientation='plus', exon_count=7, OMIM=616578, no_hdr0=''),
  22891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22891, CurrentID=0, Status='live', Symbol='ZNF365', Aliases='Su48, UAND, ZNF365', description='zinc finger protein 365', other_designations='protein su48|talanin', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62374157, end_position_on_the_genomic_accession=62672011, orientation='plus', exon_count=18, OMIM=607818, no_hdr0=''),
  22893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22893, CurrentID=0, Status='live', Symbol='BAHD1', Aliases='', description='bromo adjacent homology domain containing 1', other_designations='BAH domain-containing protein 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40439721, end_position_on_the_genomic_accession=40468242, orientation='plus', exon_count=9, OMIM=613880, no_hdr0=''),
  22894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22894, CurrentID=0, Status='live', Symbol='DIS3', Aliases='2810028N01Rik, EXOSC11, KIAA1008, RRP44, dis3p', description="DIS3 homolog, exosome endoribonuclease and 3'-5' exoribonuclease", other_designations="DIS3 exosome endoribonuclease and 3'-5' exoribonuclease|DIS3 mitotic control homolog|exosome component 11|mitotic control protein dis3 homolog|ribosomal RNA-processing protein 44", map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72755402, end_position_on_the_genomic_accession=72782153, orientation='minus', exon_count=22, OMIM=607533, no_hdr0=''),
  22895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22895, CurrentID=0, Status='live', Symbol='RPH3A', Aliases='', description='rabphilin 3A', other_designations='exophilin-1|rabphilin 3A homolog', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112791538, end_position_on_the_genomic_accession=112898881, orientation='plus', exon_count=23, OMIM=612159, no_hdr0=''),
  22897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22897, CurrentID=0, Status='live', Symbol='CEP164', Aliases='NPHP15', description='centrosomal protein 164kDa', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117321778, end_position_on_the_genomic_accession=117413266, orientation='plus', exon_count=41, OMIM=614848, no_hdr0=''),
  22898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22898, CurrentID=0, Status='live', Symbol='DENND3', Aliases='', description='DENN/MADD domain containing 3', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141128621, end_position_on_the_genomic_accession=141195804, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  22899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22899, CurrentID=0, Status='live', Symbol='ARHGEF15', Aliases='ARGEF15, E5, Ephexin5, Vsm-RhoGEF', description='Rho guanine nucleotide exchange factor 15', other_designations='Rho guanine nucleotide exchange factor (GEF) 15|ephexin-5', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8310238, end_position_on_the_genomic_accession=8322516, orientation='plus', exon_count=18, OMIM=608504, no_hdr0=''),
  22900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22900, CurrentID=0, Status='live', Symbol='CARD8', Aliases='CARDINAL, DACAR, DAKAR, NDPP, NDPP1, TUCAN', description='caspase recruitment domain family member 8', other_designations='CARD inhibitor of NF-kappaB-activating ligands|apoptotic protein NDPP1|caspase recruitment domain family, member 8|tumor up-regulated CARD-containing antagonist of CASP9|tumor up-regulated CARD-containing antagonist of caspase nine', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48190371, end_position_on_the_genomic_accession=48256277, orientation='minus', exon_count=26, OMIM=609051, no_hdr0=''),
  22901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22901, CurrentID=0, Status='live', Symbol='ARSG', Aliases='', description='arylsulfatase G', other_designations='ASG', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68257159, end_position_on_the_genomic_accession=68450773, orientation='plus', exon_count=23, OMIM=610008, no_hdr0=''),
  22902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22902, CurrentID=0, Status='live', Symbol='RUFY3', Aliases='RIPX, SINGAR1, ZFYVE30', description='RUN and FYVE domain containing 3', other_designations='RUN and FYVE domain-containing protein 3|rap2 interacting protein x|single axon-regulated protein|single axon-related 1', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70704686, end_position_on_the_genomic_accession=70808622, orientation='plus', exon_count=23, OMIM=611194, no_hdr0=''),
  22903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22903, CurrentID=0, Status='live', Symbol='BTBD3', Aliases='dJ742J24.1', description='BTB (POZ) domain containing 3', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11890723, end_position_on_the_genomic_accession=11926595, orientation='plus', exon_count=7, OMIM=615566, no_hdr0=''),
  22904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22904, CurrentID=0, Status='live', Symbol='SBNO2', Aliases='KIAA0963, SNO, STNO', description='strawberry notch homolog 2 (Drosophila)', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1107634, end_position_on_the_genomic_accession=1174283, orientation='minus', exon_count=36, OMIM=615729, no_hdr0=''),
  22905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22905, CurrentID=0, Status='live', Symbol='EPN2', Aliases='EHB21', description='epsin 2', other_designations='EPS-15-interacting protein 2|Eps15 binding protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19237377, end_position_on_the_genomic_accession=19336715, orientation='plus', exon_count=11, OMIM=607263, no_hdr0=''),
  22906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22906, CurrentID=0, Status='live', Symbol='TRAK1', Aliases='MILT1, OIP106', description='trafficking protein, kinesin binding 1', other_designations='106 kDa O-GlcNAc transferase-interacting protein|O-linked N-acetylglucosamine transferase interacting protein 106|OGT(O-Glc-NAc transferase)-interacting protein 106 KDa|milton homolog 1', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42013802, end_position_on_the_genomic_accession=42225890, orientation='plus', exon_count=24, OMIM=608112, no_hdr0=''),
  285051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285051, CurrentID=0, Status='live', Symbol='C2orf61', Aliases='', description='chromosome 2 open reading frame 61', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47086991, end_position_on_the_genomic_accession=47155378, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22908, CurrentID=0, Status='live', Symbol='SACM1L', Aliases='SAC1', description='SAC1 suppressor of actin mutations 1-like (yeast)', other_designations='suppressor of actin 1|suppressor of actin mutations 1-like protein', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45689262, end_position_on_the_genomic_accession=45745409, orientation='plus', exon_count=21, OMIM=606569, no_hdr0=''),
  22909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22909, CurrentID=0, Status='live', Symbol='FAN1', Aliases='KIAA1018, KMIN, MTMR15, hFAN1', description='FANCD2/FANCI-associated nuclease 1', other_designations='coiled-coil domain-containing protein MTMR15|fanconi anemia associated nuclease 1|myotubularin-related protein 15', map_location='15q13.2-q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30903852, end_position_on_the_genomic_accession=30943108, orientation='plus', exon_count=18, OMIM=613534, no_hdr0=''),
  22911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22911, CurrentID=0, Status='live', Symbol='WDR47', Aliases='', description='WD repeat domain 47', other_designations='nemitin|neuronal enriched MAP interacting protein|neuronal enriched MAP-interacting protein', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108970211, end_position_on_the_genomic_accession=109042228, orientation='minus', exon_count=16, OMIM=615734, no_hdr0=''),
  22913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22913, CurrentID=0, Status='live', Symbol='RALY', Aliases='HNRPCL2, P542', description='RALY heterogeneous nuclear ribonucleoprotein', other_designations='RNA binding protein, autoantigenic (hnRNP-associated with lethal yellow homolog)|RNA-binding protein (autoantigenic)|RNA-binding protein (autoantigenic, hnRNP-associated with lethal yellow)|autoantigen p542|heterogeneous nuclear ribonucleoprotein C-like 2|hnRNP associated with lethal yellow protein homolog|hnRNP core protein C-like 2', map_location='20q11.21-q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33993652, end_position_on_the_genomic_accession=34083185, orientation='plus', exon_count=12, OMIM=614663, no_hdr0=''),
  22914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22914, CurrentID=0, Status='live', Symbol='KLRK1', Aliases='CD314, D12S2489E, KLR, NKG2-D, NKG2D', description='killer cell lectin-like receptor subfamily K, member 1', other_designations='NK cell receptor D|NKG2-D-activating NK receptor', map_location='12p13.2-p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10372353, end_position_on_the_genomic_accession=10390054, orientation='minus', exon_count=8, OMIM=611817, no_hdr0=''),
  22915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22915, CurrentID=0, Status='live', Symbol='MMRN1', Aliases='ECM, EMILIN4, GPIa*, MMRN', description='multimerin 1', other_designations='EMILIN-4|elastin microfibril interface located protein 4|elastin microfibril interfacer 4|endothelial cell multimerin|glycoprotein Ia*', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89894876, end_position_on_the_genomic_accession=89954629, orientation='plus', exon_count=8, OMIM=601456, no_hdr0=''),
  22916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22916, CurrentID=0, Status='live', Symbol='NCBP2', Aliases='CBC2, CBP20, NIP1, PIG55', description='nuclear cap binding protein subunit 2', other_designations='20 kDa nuclear cap-binding protein|NCBP 20 kDa subunit|NCBP interacting protein 1|NCBP-interacting protein 1|cell proliferation-inducing gene 55 protein|nuclear cap binding protein subunit 2, 20kD|nuclear cap binding protein subunit 2, 20kDa', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196935402, end_position_on_the_genomic_accession=196942593, orientation='minus', exon_count=5, OMIM=605133, no_hdr0=''),
  22917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22917, CurrentID=0, Status='live', Symbol='ZP1', Aliases='HEL163, OOMD', description='zona pellucida glycoprotein 1', other_designations='epididymis luminal protein 163|zona pellucida glycoprotein 1 (sperm receptor)', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60867542, end_position_on_the_genomic_accession=60875693, orientation='plus', exon_count=13, OMIM=195000, no_hdr0=''),
  22918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22918, CurrentID=0, Status='live', Symbol='CD93', Aliases='C1QR1, C1qR(P), C1qRP, CDw93, ECSM3, MXRA4, dJ737E23.1', description='CD93 molecule', other_designations='C1q receptor 1|C1q/MBL/SPA receptor|C1qR|CD93 antigen|complement component 1 q subcomponent receptor 1|complement component 1, q subcomponent, receptor 1|matrix-remodeling-associated protein 4|matrix-remodelling associated 4', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23079356, end_position_on_the_genomic_accession=23086340, orientation='minus', exon_count=2, OMIM=120577, no_hdr0=''),
  22919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22919, CurrentID=0, Status='live', Symbol='MAPRE1', Aliases='EB1', description='microtubule associated protein RP/EB family member 1', other_designations='APC-binding protein EB1|adenomatous polyposis coli-binding protein EB1|end-binding protein 1|microtubule-associated protein, RP/EB family, member 1', map_location='20q11.1-q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32819780, end_position_on_the_genomic_accession=32850405, orientation='plus', exon_count=7, OMIM=603108, no_hdr0=''),
  22920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22920, CurrentID=0, Status='live', Symbol='KIFAP3', Aliases='FLA3, KAP-1, KAP-3, KAP3, SMAP, Smg-GDS, dJ190I16.1', description='kinesin-associated protein 3', other_designations='small G protein GDP dissociation stimulator|smg GDS-associated protein', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169921329, end_position_on_the_genomic_accession=170074741, orientation='minus', exon_count=24, OMIM=601836, no_hdr0=''),
  22921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22921, CurrentID=0, Status='live', Symbol='MSRB2', Aliases='CBS-1, CBS1, CGI-131, MSRB, PILB', description='methionine sulfoxide reductase B2', other_designations='pilin-like transcription factor', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23095498, end_position_on_the_genomic_accession=23122013, orientation='plus', exon_count=6, OMIM=613782, no_hdr0=''),
  22924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22924, CurrentID=0, Status='live', Symbol='MAPRE3', Aliases='EB3, EBF3, EBF3-S, RP3', description='microtubule associated protein RP/EB family member 3', other_designations='APC binding protein|EB1 protein family member 3|end-binding protein 3|microtubule-associated protein, RP/EB family, member 3', map_location='2p23.3-p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26970371, end_position_on_the_genomic_accession=27027219, orientation='plus', exon_count=8, OMIM=605788, no_hdr0=''),
  22925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22925, CurrentID=0, Status='live', Symbol='PLA2R1', Aliases='CLEC13C, PLA2-R, PLA2G1R, PLA2IR, PLA2R', description='phospholipase A2 receptor 1', other_designations='180 kDa secretory phospholipase A2 receptor|C-type lectin domain family 13 member C|M-type receptor|phospholipase A2 receptor 1, 180kD|phospholipase A2 receptor 1, 180kDa', map_location='2q23-q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159932002, end_position_on_the_genomic_accession=160062902, orientation='minus', exon_count=34, OMIM=604939, no_hdr0=''),
  22926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22926, CurrentID=0, Status='live', Symbol='ATF6', Aliases='ATF6A', description='activating transcription factor 6', other_designations='cAMP-dependent transcription factor ATF-6 alpha', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161766244, end_position_on_the_genomic_accession=161964070, orientation='plus', exon_count=17, OMIM=605537, no_hdr0=''),
  22927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22927, CurrentID=0, Status='live', Symbol='HABP4', Aliases='IHABP-4, IHABP4, Ki-1/57, SERBP1L', description='hyaluronan binding protein 4', other_designations='chromodomain helicase DNA binding protein 3 interacting protein|intracellular antigen detected by monoclonal antibody Ki-1|ki-1/57 intracellular antigen', map_location='9q22.3-q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96450155, end_position_on_the_genomic_accession=96491336, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22928, CurrentID=0, Status='live', Symbol='SEPHS2', Aliases='SPS2, SPS2b', description='selenophosphate synthetase 2', other_designations='selenium donor protein 2|selenophosphate synthase 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30443625, end_position_on_the_genomic_accession=30445975, orientation='minus', exon_count=1, OMIM=606218, no_hdr0=''),
  22929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22929, CurrentID=0, Status='live', Symbol='SEPHS1', Aliases='SELD, SPS, SPS1', description='selenophosphate synthetase 1', other_designations='selenium donor protein 1|selenophosphate synthase 1|selenophosphate synthetase 1 +E9|selenophosphate synthetase 1 +E9a|selenophosphate synthetase 1 delta E2|selenophosphate synthetase 1 delta E8|selenophosphate synthetase 1 major', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13317428, end_position_on_the_genomic_accession=13348298, orientation='minus', exon_count=9, OMIM=600902, no_hdr0=''),
  22930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22930, CurrentID=0, Status='live', Symbol='RAB3GAP1', Aliases='P130, RAB3GAP30, WARBM1, RAB3GAP1', description='RAB3 GTPase activating protein catalytic subunit 1', other_designations='RAB3 GTPase activating protein subunit 1 (catalytic)|RAB3 GTPase-activating protein 130 kDa subunit|rab3-GAP p130', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135052265, end_position_on_the_genomic_accession=135175584, orientation='plus', exon_count=27, OMIM=602536, no_hdr0=''),
  22931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22931, CurrentID=0, Status='live', Symbol='RAB18', Aliases='RAB18LI1, WARBM3', description='RAB18, member RAS oncogene family', other_designations='RAB18 small GTPase', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27504174, end_position_on_the_genomic_accession=27542237, orientation='plus', exon_count=8, OMIM=602207, no_hdr0=''),
  22932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22932, CurrentID=0, Status='live', Symbol='POMZP3', Aliases='POM-ZP3, POM121', description='POM121 and ZP3 fusion', other_designations='POM (POM121 homolog, rat) and ZP3 fusion|POM (POM121 rat homolog) and ZP3 fusion|POM-ZP3 fusion protein|POM121/ZP3 fusion protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76609986, end_position_on_the_genomic_accession=76627303, orientation='minus', exon_count=7, OMIM=600587, no_hdr0=''),
  22933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22933, CurrentID=0, Status='live', Symbol='SIRT2', Aliases='SIR2, SIR2L, SIR2L2', description='sirtuin 2', other_designations='NAD-dependent deacetylase sirtuin-2|SIR2-like protein 2|regulatory protein SIR2 homolog 2|silent information regulator 2|sir2-related protein type 2|sirtuin type 2|sirtuin-2', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38878555, end_position_on_the_genomic_accession=38899862, orientation='minus', exon_count=18, OMIM=604480, no_hdr0=''),
  22934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22934, CurrentID=0, Status='live', Symbol='RPIA', Aliases='RPID, RPIA', description='ribose 5-phosphate isomerase A', other_designations='phosphoriboisomerase|ribose 5-phosphate epimerase', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88691648, end_position_on_the_genomic_accession=88750935, orientation='plus', exon_count=10, OMIM=180430, no_hdr0=''),
  154007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154007, CurrentID=0, Status='live', Symbol='SNRNP48', Aliases='C6orf151, dJ336K20B.1, dJ512B11.2', description='small nuclear ribonucleoprotein, U11/U12 48KDa subunit', other_designations='U11/U12 snRNP 48 kDa protein|U11/U12 snRNP 48K|U11/U12-48K|small nuclear ribonucleoprotein 48kDa (U11/U12)', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7590199, end_position_on_the_genomic_accession=7611967, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22936, CurrentID=0, Status='live', Symbol='ELL2', Aliases='', description='elongation factor, RNA polymerase II, 2', other_designations='ELL-related RNA polymerase II, elongation factor', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95885098, end_position_on_the_genomic_accession=95962071, orientation='minus', exon_count=13, OMIM=601874, no_hdr0=''),
  22937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22937, CurrentID=0, Status='live', Symbol='SCAP', Aliases='', description='SREBF chaperone', other_designations='SREBP cleavage-activating protein|sterol regulatory element binding protein cleavage-activating protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47413694, end_position_on_the_genomic_accession=47477126, orientation='minus', exon_count=24, OMIM=601510, no_hdr0=''),
  22938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22938, CurrentID=0, Status='live', Symbol='SNW1', Aliases='Bx42, NCOA-62, PRPF45, Prp45, SKIIP, SKIP', description='SNW domain containing 1', other_designations='SKI interacting protein|SKI-interacting protein|homolog of Drosophila BX42|nuclear protein SkiP|nuclear receptor coactivator NCoA-62|nuclear receptor coactivator, 62-kD', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77717599, end_position_on_the_genomic_accession=77761218, orientation='minus', exon_count=13, OMIM=603055, no_hdr0=''),
  285084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285084, CurrentID=0, Status='live', Symbol='LINC01305', Aliases='', description='long intergenic non-protein coding RNA 1305', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174326027, end_position_on_the_genomic_accession=174330643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22941, CurrentID=0, Status='live', Symbol='SHANK2', Aliases='AUTS17, CORTBP1, CTTNBP1, ProSAP1, SHANK, SPANK-3', description='SH3 and multiple ankyrin repeat domains 2', other_designations='GKAP/SAPAP interacting protein|cortactin SH3 domain-binding protein|cortactin-binding protein 1|proline-rich synapse associated protein 1|proline-rich synapse-associated protein 1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70467856, end_position_on_the_genomic_accession=71252738, orientation='minus', exon_count=28, OMIM=603290, no_hdr0=''),
  22943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22943, CurrentID=0, Status='live', Symbol='DKK1', Aliases='DKK-1, SK', description='dickkopf WNT signaling pathway inhibitor 1', other_designations='dickkopf 1 homolog|dickkopf related protein-1|dickkopf-1 like|dickkopf-like protein 1|hDkk-1', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52314281, end_position_on_the_genomic_accession=52317657, orientation='plus', exon_count=4, OMIM=605189, no_hdr0=''),
  22944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22944, CurrentID=0, Status='live', Symbol='KIN', Aliases='BTCD17, Rts2, KIN', description='Kin17 DNA and RNA binding protein', other_designations='KIN, antigenic determinant of recA protein homolog|binding to curved DNA', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7750962, end_position_on_the_genomic_accession=7788027, orientation='minus', exon_count=14, OMIM=601720, no_hdr0=''),
  22945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22945, CurrentID=0, Status='live', Symbol='GPAA1P1', Aliases='', description='glycosylphosphatidylinositol anchor attachment 1 pseudogene 1', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109983968, end_position_on_the_genomic_accession=109985187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22947, CurrentID=0, Status='live', Symbol='DUX4L1', Aliases='DUX10, DUX4', description='double homeobox 4 like 1', other_designations='double homeobox protein 10|double homeobox protein 4/10', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190084412, end_position_on_the_genomic_accession=190085696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22948, CurrentID=0, Status='live', Symbol='CCT5', Aliases='CCT-epsilon, CCTE, HEL-S-69, PNAS-102, TCP-1-epsilon', description='chaperonin containing TCP1, subunit 5 (epsilon)', other_designations='epididymis secretory protein Li 69', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10250170, end_position_on_the_genomic_accession=10266389, orientation='plus', exon_count=12, OMIM=610150, no_hdr0=''),
  285093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285093, CurrentID=0, Status='live', Symbol='RTP5', Aliases='C2orf85, CXXC11, Z3CXXC5', description='receptor (chemosensory) transporter protein 5 (putative)', other_designations='3CxxC-type zinc finger protein 5|CXXC finger protein 11|CXXC-type zinc finger protein 11|zinc finger, 3CxxC-type 5', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241868319, end_position_on_the_genomic_accession=241873823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22950, CurrentID=0, Status='live', Symbol='SLC4A1AP', Aliases='HLC3', description='solute carrier family 4 (anion exchanger), member 1, adaptor protein', other_designations='HLC-3|human lung cancer oncogene 3 protein|kidney anion exchanger adapter protein|kidney anion exchanger adaptor protein|lung cancer oncogene 3|solute carrier family 4 (anion exchanger), member 1, adapter protein|solute carrier family 4 anion exchanger member 1 adapter protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27663471, end_position_on_the_genomic_accession=27694980, orientation='plus', exon_count=14, OMIM=602655, no_hdr0=''),
  285095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285095, CurrentID=0, Status='live', Symbol='LOC285095', Aliases='', description='uncharacterized LOC285095', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241893988, end_position_on_the_genomic_accession=241903279, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22952, CurrentID=0, Status='live', Symbol='CYP2G1P', Aliases='CYP2G1, CYP2G2P, CYP2GP1', description='cytochrome P450 family 2 subfamily G member 1, pseudogene', other_designations='cytochrome P450, family 2, subfamily G, polypeptide 1 pseudogene|cytochrome P450, subfamily IIG, olfactory-specific|cytochrome P450, subfamily IIG, polypeptide 1, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40890826, end_position_on_the_genomic_accession=40900508, orientation='plus', exon_count=5, OMIM=601133, no_hdr0=''),
  22953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22953, CurrentID=0, Status='live', Symbol='P2RX2', Aliases='DFNA41, P2X2', description='purinergic receptor P2X, ligand gated ion channel, 2', other_designations='ATP receptor|P2X Receptor, subunit 2|purinergic receptor P2X, ligand-gated ion channel, 2', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132618780, end_position_on_the_genomic_accession=132622462, orientation='plus', exon_count=10, OMIM=600844, no_hdr0=''),
  22954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22954, CurrentID=0, Status='live', Symbol='TRIM32', Aliases='BBS11, HT2A, LGMD2H, TATIP', description='tripartite motif containing 32', other_designations='72 kDa Tat-interacting protein|TAT-interactive protein, 72-KD|tripartite motif-containing 32|tripartite motif-containing protein 32|zinc finger protein HT2A|zinc-finger protein HT2A', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116687302, end_position_on_the_genomic_accession=116701300, orientation='plus', exon_count=4, OMIM=602290, no_hdr0=''),
  22955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22955, CurrentID=0, Status='live', Symbol='SCMH1', Aliases='Scml3', description='sex comb on midleg homolog 1 (Drosophila)', other_designations='', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41027199, end_position_on_the_genomic_accession=41242159, orientation='minus', exon_count=24, OMIM=616396, no_hdr0=''),
  285103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285103, CurrentID=0, Status='live', Symbol='MED15P9', Aliases='CCDC74B-AS1, MED15P2', description='mediator complex subunit 15 pseudogene 9', other_designations='CCDC74B antisense RNA 1 (non-protein coding)|mediator complex subunit 15 pseudogene 2', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130129623, end_position_on_the_genomic_accession=130139413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100293044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293044, CurrentID=0, Status='live', Symbol='LOC100293044', Aliases='', description='uncharacterized LOC100293044', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=35167187, end_position_on_the_genomic_accession=35168486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22969, CurrentID=0, Status='live', Symbol='RPL41P3', Aliases='', description='ribosomal protein L41 pseudogene 3', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35838275, end_position_on_the_genomic_accession=35838713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22970, CurrentID=0, Status='live', Symbol='RPL41P2', Aliases='', description='ribosomal protein L41 pseudogene 2', other_designations='', map_location='15q11-q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28315653, end_position_on_the_genomic_accession=28316071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  154043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154043, CurrentID=0, Status='live', Symbol='CNKSR3', Aliases='MAGI1', description='CNKSR family member 3', other_designations='CNK homolog protein 3|CNK3|connector enhancer of KSR 3|maguin-like protein|membrane associated guanylate kinase interacting protein-like 1|membrane associated guanylate kinase, WW and PDZ domain containing 1|membrane-associated guanylate kinase-interacting protein-like 1', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154405299, end_position_on_the_genomic_accession=154510619, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  285116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285116, CurrentID=0, Status='live', Symbol='AHCTF1P1', Aliases='AHCTF1P, ELYS-like', description='AT-hook containing transcription factor 1 pseudogene 1', other_designations='AT hook containing transcription factor 1 pseudogene 1|ELYS transcription factor-like protein TMBS62 pseudogene', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161499067, end_position_on_the_genomic_accession=161507902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22973, CurrentID=0, Status='live', Symbol='LAMB2P1', Aliases='LAMB2L', description='laminin subunit beta 2 pseudogene 1', other_designations='laminin, beta 2 pseudogene 1', map_location='3p21.3-p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49152859, end_position_on_the_genomic_accession=49154401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22974, CurrentID=0, Status='live', Symbol='TPX2', Aliases='C20orf1, C20orf2, DIL-2, DIL2, FLS353, GD:C20orf1, HCA519, HCTP4, REPP86, p100', description='TPX2, microtubule-associated', other_designations='TPX2, microtubule-associated protein homolog|TPX2, microtubule-associated, homolog|differentially expressed in cancerous and non-cancerous lung cells 2|differentially expressed in lung cells|hepatocellular carcinoma-associated antigen 519|hepatocellular carcinoma-associated antigen 90|preferentially expressed in colorectal cancer|protein fls353|restricted expression proliferation associated protein 100|restricted expression proliferation-associated protein 100', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31733819, end_position_on_the_genomic_accession=31801802, orientation='plus', exon_count=19, OMIM=605917, no_hdr0=''),
  22976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22976, CurrentID=0, Status='live', Symbol='PAXIP1', Aliases='CAGF28, CAGF29, PACIP1L, PTIP, TNRC2, PAXIP1', description='PAX interacting (with transcription-activation domain) protein 1', other_designations='PAX transcription activation domain interacting protein 1 like|protein encoded by CAG trinucleotide repeats', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154943687, end_position_on_the_genomic_accession=155003178, orientation='minus', exon_count=24, OMIM=608254, no_hdr0=''),
  22977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22977, CurrentID=0, Status='live', Symbol='AKR7A3', Aliases='AFAR2', description='aldo-keto reductase family 7, member A3 (aflatoxin aldehyde reductase)', other_designations='AFB1 aldehyde reductase 2|AFB1-AR 2|aflatoxin B1 aldehyde reductase 2', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19274325, end_position_on_the_genomic_accession=19289346, orientation='minus', exon_count=8, OMIM=608477, no_hdr0=''),
  22978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22978, CurrentID=0, Status='live', Symbol='NT5C2', Aliases='GMP, NT5B, PNT5, SPG45, SPG65, cN-II', description="5'-nucleotidase, cytosolic II", other_designations="5'-nucleotidase (purine), cytosolic type B|IMP-specific 5'-NT", map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103088017, end_position_on_the_genomic_accession=103193306, orientation='minus', exon_count=26, OMIM=600417, no_hdr0=''),
  22979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22979, CurrentID=0, Status='live', Symbol='EFR3B', Aliases='KIAA0953', description='EFR3 homolog B', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25042104, end_position_on_the_genomic_accession=25159135, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  22980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22980, CurrentID=0, Status='live', Symbol='TCF25', Aliases='FKSG26, Hulp1, NULP1, PRO2620, hKIAA1049', description='transcription factor 25 (basic helix-loop-helix)', other_designations='NULP1|TCF-25|nuclear localized protein 1', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89873586, end_position_on_the_genomic_accession=89911387, orientation='plus', exon_count=20, OMIM=612326, no_hdr0=''),
  22981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22981, CurrentID=0, Status='live', Symbol='NINL', Aliases='NLP', description='ninein like', other_designations='', map_location='20p11.22-p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25452697, end_position_on_the_genomic_accession=25585531, orientation='minus', exon_count=32, OMIM=609580, no_hdr0=''),
  22982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22982, CurrentID=0, Status='live', Symbol='DIP2C', Aliases='KIAA0934', description='disco interacting protein 2 homolog C', other_designations='DIP2 disco-interacting protein 2 homolog C|DIP2 homolog C', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=274190, end_position_on_the_genomic_accession=689689, orientation='minus', exon_count=43, OMIM=611380, no_hdr0=''),
  22983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22983, CurrentID=0, Status='live', Symbol='MAST1', Aliases='SAST, SAST170', description='microtubule associated serine/threonine kinase 1', other_designations='syntrophin associated serine/threonine kinase|syntrophin-associated serine/threonine-protein kinase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12833934, end_position_on_the_genomic_accession=12874953, orientation='plus', exon_count=30, OMIM=612256, no_hdr0=''),
  22984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22984, CurrentID=0, Status='live', Symbol='PDCD11', Aliases='ALG-4, ALG4, NFBP, RRP5', description='programmed cell death 11', other_designations='NF-kappa-B-binding protein|apoptosis-linked gene 4|programmed cell death protein 11', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103396624, end_position_on_the_genomic_accession=103446295, orientation='plus', exon_count=37, OMIM=612333, no_hdr0=''),
  22985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22985, CurrentID=0, Status='live', Symbol='ACIN1', Aliases='ACINUS, ACN, fSAP152', description='apoptotic chromatin condensation inducer 1', other_designations='functional spliceosome-associated protein 152', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23058564, end_position_on_the_genomic_accession=23095614, orientation='minus', exon_count=23, OMIM=604562, no_hdr0=''),
  22986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22986, CurrentID=0, Status='live', Symbol='SORCS3', Aliases='SORCS', description='sortilin-related VPS10 domain containing receptor 3', other_designations='VPS10 domain receptor protein SORCS 3 (SORCS3)', map_location='10q23-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104640837, end_position_on_the_genomic_accession=105265250, orientation='plus', exon_count=29, OMIM=606285, no_hdr0=''),
  22987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22987, CurrentID=0, Status='live', Symbol='SV2C', Aliases='', description='synaptic vesicle glycoprotein 2C', other_designations='synaptic vesicle protein 2C', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76083414, end_position_on_the_genomic_accession=76353939, orientation='plus', exon_count=15, OMIM=610291, no_hdr0=''),
  105372081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372081, CurrentID=0, Status='live', Symbol='LOC105372081', Aliases='', description='uncharacterized LOC105372081', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=40064402, end_position_on_the_genomic_accession=40066465, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22989, CurrentID=0, Status='live', Symbol='MYH15', Aliases='', description='myosin, heavy chain 15', other_designations='myosin, heavy polypeptide 15', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108380368, end_position_on_the_genomic_accession=108549465, orientation='minus', exon_count=43, OMIM=609929, no_hdr0=''),
  22990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22990, CurrentID=0, Status='live', Symbol='PCNX', Aliases='PCNXL1, pecanex', description='pecanex homolog (Drosophila)', other_designations='pecanex-like 1', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70907405, end_position_on_the_genomic_accession=71115382, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  22992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22992, CurrentID=0, Status='live', Symbol='KDM2A', Aliases='CXXC8, FBL11, FBL7, FBXL11, JHDM1A, LILINA', description='lysine (K)-specific demethylase 2A', other_designations='CXXC-type zinc finger protein 8|F-box and leucine-rich repeat protein 11|F-box/LRR-repeat protein 11|[Histone-H3]-lysine-36 demethylase 1A|jmjC domain-containing histone demethylation protein 1A|jumonji C domain-containing histone demethylase 1A', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67119268, end_position_on_the_genomic_accession=67258082, orientation='plus', exon_count=24, OMIM=605657, no_hdr0=''),
  22993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22993, CurrentID=0, Status='live', Symbol='HMGXB3', Aliases='HMGX3, SMF', description='HMG box domain containing 3', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150000606, end_position_on_the_genomic_accession=150053143, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22994, CurrentID=0, Status='live', Symbol='CEP131', Aliases='AZ1, AZI1, ZA1', description='centrosomal protein 131kDa', other_designations='5-azacytidine induced 1|5-azacytidine-induced protein 1|centrosomal protein 131 kDa|pre-acrosome localization protein 1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81189593, end_position_on_the_genomic_accession=81222994, orientation='minus', exon_count=27, OMIM=613479, no_hdr0=''),
  22995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22995, CurrentID=0, Status='live', Symbol='CEP152', Aliases='MCPH4, MCPH9, SCKL5', description='centrosomal protein 152kDa', other_designations='asterless|microcephaly, primary autosomal recessive 4', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48662534, end_position_on_the_genomic_accession=48811918, orientation='minus', exon_count=38, OMIM=613529, no_hdr0=''),
  22996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22996, CurrentID=0, Status='live', Symbol='TTC39A', Aliases='C1orf34, DEME-6', description='tetratricopeptide repeat domain 39A', other_designations='TPR repeat protein 39A|differentially expressed in MCF7 with estradiol protein 6', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51287258, end_position_on_the_genomic_accession=51345116, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  285141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285141, CurrentID=0, Status='live', Symbol='ERICH2', Aliases='', description='glutamate rich 2', other_designations='glutamate-rich 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170770541, end_position_on_the_genomic_accession=170798971, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22998, CurrentID=0, Status='live', Symbol='LIMCH1', Aliases='LIMCH1A, LMO7B', description='LIM and calponin homology domains 1', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41359607, end_position_on_the_genomic_accession=41700044, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  22999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22999, CurrentID=0, Status='live', Symbol='RIMS1', Aliases='CORD7, RAB3IP2, RIM, RIM1', description='regulating synaptic membrane exocytosis 1', other_designations='RAB3-interacting protein 2|rab-3-interacting protein 2|rab3-interacting molecule 1', map_location='6q12-q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71886703, end_position_on_the_genomic_accession=72403143, orientation='plus', exon_count=40, OMIM=606629, no_hdr0=''),
  23001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23001, CurrentID=0, Status='live', Symbol='WDFY3', Aliases='ALFY, ZFYVE25', description='WD repeat and FYVE domain containing 3', other_designations='autophagy-linked FYVE protein', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84669537, end_position_on_the_genomic_accession=84966391, orientation='minus', exon_count=72, OMIM='', no_hdr0=''),
  23002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23002, CurrentID=0, Status='live', Symbol='DAAM1', Aliases='', description='dishevelled associated activator of morphogenesis 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59188611, end_position_on_the_genomic_accession=59371405, orientation='plus', exon_count=26, OMIM=606626, no_hdr0=''),
  154075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154075, CurrentID=0, Status='live', Symbol='SAMD3', Aliases='', description='sterile alpha motif domain containing 3', other_designations='SAM domain-containing protein 3', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130133203, end_position_on_the_genomic_accession=130365425, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  285148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285148, CurrentID=0, Status='live', Symbol='IAH1', Aliases='', description='isoamyl acetate-hydrolyzing esterase 1 homolog', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9474532, end_position_on_the_genomic_accession=9517208, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23005, CurrentID=0, Status='live', Symbol='MAPKBP1', Aliases='JNKBP-1', description='mitogen-activated protein kinase binding protein 1', other_designations='JNK-binding protein 1|mitogen activated protein kinase binding protein 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41774434, end_position_on_the_genomic_accession=41827855, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  285150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285150, CurrentID=0, Status='live', Symbol='FLJ33534', Aliases='', description='uncharacterized LOC285150', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11099851, end_position_on_the_genomic_accession=11132176, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23007, CurrentID=0, Status='live', Symbol='PLCH1', Aliases='PLCL3', description='phospholipase C eta 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase eta-1|PLC eta 1|PLC-L3|PLC-eta-1|phosphoinositide phospholipase C-eta-1|phospholipase C, eta 1|phospholipase C-eta-1|phospholipase C-eta1a|phospholipase C-eta1b|phospholipase C-like 3|phospholipase C-like protein 3', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155479881, end_position_on_the_genomic_accession=155745089, orientation='minus', exon_count=31, OMIM=612835, no_hdr0=''),
  23008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23008, CurrentID=0, Status='live', Symbol='KLHDC10', Aliases='slim', description='kelch domain containing 10', other_designations='PNAS-119|scruin like at the midline homolog', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130070509, end_position_on_the_genomic_accession=130135720, orientation='plus', exon_count=10, OMIM=615152, no_hdr0=''),
  285154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285154, CurrentID=0, Status='live', Symbol='CYP1B1-AS1', Aliases='C2orf58', description='CYP1B1 antisense RNA 1', other_designations='CYP1B1 antisense RNA 1 (non-protein coding)', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38131105, end_position_on_the_genomic_accession=38181855, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23011, CurrentID=0, Status='live', Symbol='RAB21', Aliases='', description='RAB21, member RAS oncogene family', other_designations='GTP-binding protein RAB21', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71754850, end_position_on_the_genomic_accession=71787370, orientation='plus', exon_count=7, OMIM=612398, no_hdr0=''),
  23012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23012, CurrentID=0, Status='live', Symbol='STK38L', Aliases='NDR2', description='serine/threonine kinase 38 like', other_designations='NDR2 protein kinase|nuclear Dbf2-related 2|nuclear Dbf2-related kinase 2', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27244145, end_position_on_the_genomic_accession=27325959, orientation='plus', exon_count=16, OMIM=615836, no_hdr0=''),
  23013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23013, CurrentID=0, Status='live', Symbol='SPEN', Aliases='HIAA0929, MINT, RBM15C, SHARP', description='spen family transcriptional repressor', other_designations='Msx2 interacting nuclear target (MINT) homolog|SMART/HDAC1 associated repressor protein|SMART/HDAC1-associated repressor protein|nuclear receptor transcription cofactor|spen homolog, transcriptional regulator', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15847864, end_position_on_the_genomic_accession=15940455, orientation='plus', exon_count=15, OMIM=613484, no_hdr0=''),
  23014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23014, CurrentID=0, Status='live', Symbol='FBXO21', Aliases='FBX21', description='F-box protein 21', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117143780, end_position_on_the_genomic_accession=117190495, orientation='minus', exon_count=12, OMIM=609095, no_hdr0=''),
  23015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23015, CurrentID=0, Status='live', Symbol='GOLGA8A', Aliases='GM88', description='golgin A8 family member A', other_designations='88 kDa Golgi matrix protein|88-kDa golgi protein|GM88 autoantigen|golgi autoantigen, golgin subfamily a, 8A|golgin A8 family, member A|golgin-67', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34379068, end_position_on_the_genomic_accession=34437466, orientation='minus', exon_count=25, OMIM=616180, no_hdr0=''),
  23016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23016, CurrentID=0, Status='live', Symbol='EXOSC7', Aliases='EAP1, RRP42, Rrp42p, hRrp42p, p8', description='exosome component 7', other_designations='exosome complex exonuclease RRP42|ribosomal RNA-processing protein 42', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44976230, end_position_on_the_genomic_accession=45012668, orientation='plus', exon_count=10, OMIM=606488, no_hdr0=''),
  23017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23017, CurrentID=0, Status='live', Symbol='FAIM2', Aliases='LFG, LFG2, NGP35, NMP35, TMBIM2', description='Fas apoptotic inhibitory molecule 2', other_designations='neural membrane protein 35|transmembrane BAX inhibitor motif-containing protein 2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49866896, end_position_on_the_genomic_accession=49903977, orientation='minus', exon_count=13, OMIM=604306, no_hdr0=''),
  154091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154091, CurrentID=0, Status='live', Symbol='SLC2A12', Aliases='GLUT12, GLUT8', description='solute carrier family 2 (facilitated glucose transporter), member 12', other_designations='GLUT-12|glucose transporter type 12', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133987581, end_position_on_the_genomic_accession=134052668, orientation='minus', exon_count=6, OMIM=610372, no_hdr0=''),
  23020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23020, CurrentID=0, Status='live', Symbol='SNRNP200', Aliases='ASCC3L1, BRR2, HELIC2, RP33, U5-200KD', description='small nuclear ribonucleoprotein, U5 200kDa subunit', other_designations='BRR2 homolog|U5 snRNP-specific 200 kDa protein|activating signal cointegrator 1 complex subunit 3-like 1|small nuclear ribonucleoprotein 200kDa (U5)', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96274336, end_position_on_the_genomic_accession=96305569, orientation='minus', exon_count=45, OMIM=601664, no_hdr0=''),
  23022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23022, CurrentID=0, Status='live', Symbol='PALLD', Aliases='CGI-151, CGI151, MYN, PNCA1, SIH002', description='palladin, cytoskeletal associated protein', other_designations='myoneurin|sarcoma antigen NY-SAR-77', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168497064, end_position_on_the_genomic_accession=168928457, orientation='plus', exon_count=28, OMIM=608092, no_hdr0=''),
  23023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23023, CurrentID=0, Status='live', Symbol='TMCC1', Aliases='', description='transmembrane and coiled-coil domain family 1', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129647792, end_position_on_the_genomic_accession=129893597, orientation='minus', exon_count=16, OMIM=616242, no_hdr0=''),
  23024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23024, CurrentID=0, Status='live', Symbol='PDZRN3', Aliases='LNX3, SEMACAP3, SEMCAP3', description='PDZ domain containing ring finger 3', other_designations='PDZ domain-containing RING finger protein 3|ligand of Numb protein X 3|likely ortholog of mouse semaF cytoplasmic domain associated protein 3|semaphorin cytoplasmic domain-associated protein 3', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73382430, end_position_on_the_genomic_accession=73624921, orientation='minus', exon_count=14, OMIM=609729, no_hdr0=''),
  23025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23025, CurrentID=0, Status='live', Symbol='UNC13A', Aliases='Munc13-1', description='unc-13 homolog A (C. elegans)', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17601328, end_position_on_the_genomic_accession=17688353, orientation='minus', exon_count=46, OMIM=609894, no_hdr0=''),
  23026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23026, CurrentID=0, Status='live', Symbol='MYO16', Aliases='MYAP3, MYR8, Myo16b, NYAP3, PPP1R107', description='myosin XVI', other_designations='MYO16 variant protein|myosin heavy chain Myr 8|myosin-XVI|neuronal tyrosine-phosphorylated phosphoinositide-3-kinase adapter 3|neuronal tyrosine-phosphorylated phosphoinositide-3-kinase adaptor 3|protein phosphatase 1, regulatory subunit 107|unconventional myosin-16', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108495729, end_position_on_the_genomic_accession=109208007, orientation='plus', exon_count=37, OMIM=615479, no_hdr0=''),
  23028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23028, CurrentID=0, Status='live', Symbol='KDM1A', Aliases='AOF2, BHC110, KDM1, LSD1', description='lysine (K)-specific demethylase 1A', other_designations='BRAF35-HDAC complex protein BHC110|FAD-binding protein BRAF35-HDAC complex, 110 kDa subunit|amine oxidase (flavin containing) domain 2|flavin-containing amine oxidase domain-containing protein 2|lysine (K)-specific demethylase 1|lysine-specific histone demethylase 1', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23019443, end_position_on_the_genomic_accession=23083691, orientation='plus', exon_count=22, OMIM=609132, no_hdr0=''),
  23029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23029, CurrentID=0, Status='live', Symbol='RBM34', Aliases='', description='RNA binding motif protein 34', other_designations='RNA-binding motif protein 34', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235131183, end_position_on_the_genomic_accession=235161616, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  23030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23030, CurrentID=0, Status='live', Symbol='KDM4B', Aliases='JMJD2B, TDRD14B', description='lysine (K)-specific demethylase 4B', other_designations='jmjC domain-containing histone demethylation protein 3B|jumonji domain containing 2B|jumonji domain-containing protein 2B|tudor domain containing 14B', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4969112, end_position_on_the_genomic_accession=5153598, orientation='plus', exon_count=32, OMIM=609765, no_hdr0=''),
  23031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23031, CurrentID=0, Status='live', Symbol='MAST3', Aliases='', description='microtubule associated serine/threonine kinase 3', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18097764, end_position_on_the_genomic_accession=18151692, orientation='plus', exon_count=33, OMIM=612258, no_hdr0=''),
  23032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23032, CurrentID=0, Status='live', Symbol='USP33', Aliases='VDU1', description='ubiquitin specific peptidase 33', other_designations='VHL-interacting deubiquitinating enzyme 1|deubiquitinating enzyme 33|pVHL-interacting deubiquitinating enzyme 1|ubiquitin thioesterase 33|ubiquitin thiolesterase 33|ubiquitin-specific-processing protease 33', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77695987, end_position_on_the_genomic_accession=77759879, orientation='minus', exon_count=26, OMIM=615146, no_hdr0=''),
  23033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23033, CurrentID=0, Status='live', Symbol='DOPEY1', Aliases='KIAA1117, dJ202D23.2', description='dopey family member 1', other_designations='homolog of yeast DOP1', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83067660, end_position_on_the_genomic_accession=83168471, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  23034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23034, CurrentID=0, Status='live', Symbol='SAMD4A', Aliases='SAMD4, SMAUG, SMAUG1, SMG, SMGA', description='sterile alpha motif domain containing 4A', other_designations='SAM domain-containing protein 4A', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54567612, end_position_on_the_genomic_accession=54793315, orientation='plus', exon_count=14, OMIM=610747, no_hdr0=''),
  23035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23035, CurrentID=0, Status='live', Symbol='PHLPP2', Aliases='PHLPPL', description='PH domain and leucine rich repeat protein phosphatase 2', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71644924, end_position_on_the_genomic_accession=71724701, orientation='minus', exon_count=20, OMIM=611066, no_hdr0=''),
  23036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23036, CurrentID=0, Status='live', Symbol='ZNF292', Aliases='Nbla00365, ZFP292, ZN-16, Zn-15, bA393I2.3', description='zinc finger protein 292', other_designations='16 zinc-finger domain protein|putative protein product of Nbla00365', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87155551, end_position_on_the_genomic_accession=87263690, orientation='plus', exon_count=10, OMIM=616213, no_hdr0=''),
  23037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23037, CurrentID=0, Status='live', Symbol='PDZD2', Aliases='AIPC, PAPIN, PDZK3, PIN1', description='PDZ domain containing 2', other_designations='PDZ domain-containing protein 3|activated in prostate cancer protein|plakophilin-related armadillo repeat protein-interactive PDZ protein', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31639363, end_position_on_the_genomic_accession=32110932, orientation='plus', exon_count=30, OMIM=610697, no_hdr0=''),
  23038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23038, CurrentID=0, Status='live', Symbol='WDTC1', Aliases='ADP, DCAF9', description='WD and tetratricopeptide repeats 1', other_designations='DDB1 and CUL4 associated factor 9|adipose homolog', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27234516, end_position_on_the_genomic_accession=27308633, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  23039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23039, CurrentID=0, Status='live', Symbol='XPO7', Aliases='EXP7, RANBP16', description='exportin 7', other_designations='RAN binding protein 16', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21919669, end_position_on_the_genomic_accession=22006585, orientation='plus', exon_count=28, OMIM=606140, no_hdr0=''),
  23040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23040, CurrentID=0, Status='live', Symbol='MYT1L', Aliases='NZF1, ZC2HC4B', description='myelin transcription factor 1-like', other_designations='myT1-L|neural zinc finger transcription factor 1', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1789113, end_position_on_the_genomic_accession=2331375, orientation='minus', exon_count=34, OMIM=613084, no_hdr0=''),
  23041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23041, CurrentID=0, Status='live', Symbol='MON2', Aliases='', description='MON2 homolog, regulator of endosome-to-Golgi trafficking', other_designations='MON2 regulator of endosome-to-Golgi trafficking', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62466817, end_position_on_the_genomic_accession=62597583, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  23042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23042, CurrentID=0, Status='live', Symbol='PDXDC1', Aliases='LP8165', description='pyridoxal-dependent decarboxylase domain containing 1', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14974591, end_position_on_the_genomic_accession=15139349, orientation='plus', exon_count=26, OMIM=614244, no_hdr0=''),
  23043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23043, CurrentID=0, Status='live', Symbol='TNIK', Aliases='', description='TRAF2 and NCK interacting kinase', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171062503, end_position_on_the_genomic_accession=171460408, orientation='minus', exon_count=34, OMIM=610005, no_hdr0=''),
  285189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285189, CurrentID=0, Status='live', Symbol='PLGLA', Aliases='PLGLA1, PLGP2, PRGA', description='plasminogen-like A (pseudogene)', other_designations='plasminogen pseudogene 2|plasminogen-like A1|plasminogen-related protein A', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106382114, end_position_on_the_genomic_accession=106391395, orientation='plus', exon_count=4, OMIM=612212, no_hdr0=''),
  285190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285190, CurrentID=0, Status='live', Symbol='RGPD4', Aliases='RGP4', description='RANBP2-like and GRIP domain containing 4', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107826932, end_position_on_the_genomic_accession=107892544, orientation='plus', exon_count=30, OMIM=612707, no_hdr0=''),
  23047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23047, CurrentID=0, Status='live', Symbol='PDS5B', Aliases='APRIN, AS3, CG008', description='PDS5 cohesin associated factor B', other_designations='androgen induced inhibitor of proliferation|androgen-induced proliferation inhibitor|androgen-induced prostate proliferative shutoff-associated protein AS3|androgen-induced shutoff 3', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32586427, end_position_on_the_genomic_accession=32778020, orientation='plus', exon_count=36, OMIM=605333, no_hdr0=''),
  23048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23048, CurrentID=0, Status='live', Symbol='FNBP1', Aliases='FBP17', description='formin binding protein 1', other_designations='formin-binding protein 17', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129887187, end_position_on_the_genomic_accession=130043194, orientation='minus', exon_count=18, OMIM=606191, no_hdr0=''),
  285193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285193, CurrentID=0, Status='live', Symbol='DUSP28', Aliases='DUSP26, VHP', description='dual specificity phosphatase 28', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240560054, end_position_on_the_genomic_accession=240564014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  285194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285194, CurrentID=0, Status='live', Symbol='TUSC7', Aliases='LINC00902, LSAMP-AS1, LSAMP-AS3, LSAMPAS3, NCRNA00295', description='tumor suppressor candidate 7 (non-protein coding)', other_designations='LSAMP antisense RNA 3 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116709788, end_position_on_the_genomic_accession=116717040, orientation='plus', exon_count=4, OMIM=616057, no_hdr0=''),
  23051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23051, CurrentID=0, Status='live', Symbol='ZHX3', Aliases='TIX1', description='zinc fingers and homeoboxes 3', other_designations='triple homeobox 1|triple homeobox protein 1|zinc finger and homeodomain protein 3', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41178449, end_position_on_the_genomic_accession=41317618, orientation='minus', exon_count=17, OMIM=609598, no_hdr0=''),
  23052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23052, CurrentID=0, Status='live', Symbol='ENDOD1', Aliases='', description='endonuclease domain containing 1', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95089810, end_position_on_the_genomic_accession=95132651, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23053, CurrentID=0, Status='live', Symbol='ZSWIM8', Aliases='KIAA0913', description='zinc finger, SWIM-type containing 8', other_designations='4832404P21Rik', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73785577, end_position_on_the_genomic_accession=73801798, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  23054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23054, CurrentID=0, Status='live', Symbol='NCOA6', Aliases='AIB3, ASC2, NRC, PRIP, RAP250, TRBP', description='nuclear receptor coactivator 6', other_designations='NRC RAP250|PPAR-interacting protein|activating signal cointegrator 2|activating signal cointegrator-2|amplified in breast cancer protein 3|amplified in breast cancer-3 protein|cancer-amplified transcriptional coactivator ASC-2|nuclear receptor coactivator RAP250|nuclear receptor-activating protein, 250 kDa|peroxisome proliferator-activated receptor interacting protein|peroxisome proliferator-activated receptor-interacting protein|thyroid hormone receptor binding protein|thyroid hormone receptor-binding protein', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34714774, end_position_on_the_genomic_accession=34825630, orientation='minus', exon_count=21, OMIM=605299, no_hdr0=''),
  23057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23057, CurrentID=0, Status='live', Symbol='NMNAT2', Aliases='C1orf15, PNAT2', description='nicotinamide nucleotide adenylyltransferase 2', other_designations='NMN adenylyltransferase 2|NMN/NaMN adenylyltransferase 2|NaMN adenylyltransferase 2|nicotinamide mononucleotide adenylyltransferase 2|nicotinate-nucleotide adenylyltransferase 2|pyridine nucleotide adenylyltransferase 2', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183248237, end_position_on_the_genomic_accession=183418499, orientation='minus', exon_count=12, OMIM=608701, no_hdr0=''),
  23059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23059, CurrentID=0, Status='live', Symbol='CLUAP1', Aliases='CFAP22, FAP22', description='clusterin associated protein 1', other_designations='cilia and flagella associated protein 22|flagellar associated protein 22, qilin-like protein, homolog|qilin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3500945, end_position_on_the_genomic_accession=3539048, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  23060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23060, CurrentID=0, Status='live', Symbol='ZNF609', Aliases='', description='zinc finger protein 609', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64460743, end_position_on_the_genomic_accession=64686068, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23061, CurrentID=0, Status='live', Symbol='TBC1D9B', Aliases='', description='TBC1 domain family member 9B', other_designations='TBC1 domain family, member 9B (with GRAM domain)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179862071, end_position_on_the_genomic_accession=179907856, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  23062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23062, CurrentID=0, Status='live', Symbol='GGA2', Aliases='VEAR', description='golgi-associated, gamma adaptin ear containing, ARF binding protein 2', other_designations='VHS domain and ear domain of gamma-adaptin|VHS domain and ear domain-containing protein|gamma-adaptin-related protein 2|golgi associated, gamma adaptin ear containing, ARF binding protein 2|golgi-localized, gamma ear-containing, ARF-binding protein 2', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23463542, end_position_on_the_genomic_accession=23510494, orientation='minus', exon_count=18, OMIM=606005, no_hdr0=''),
  23063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23063, CurrentID=0, Status='live', Symbol='WAPL', Aliases='FOE, KIAA0261, WAPAL', description='WAPL cohesin release factor', other_designations='friend of EBNA2 (Epstein-Barr virus nuclear protein 2)|friend of EBNA2 protein', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86435256, end_position_on_the_genomic_accession=86521816, orientation='minus', exon_count=19, OMIM=610754, no_hdr0=''),
  23064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23064, CurrentID=0, Status='live', Symbol='SETX', Aliases='ALS4, AOA2, SCAR1, bA479K20.2', description='senataxin', other_designations='SEN1 homolog|amyotrophic lateral sclerosis 4 protein', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132261440, end_position_on_the_genomic_accession=132356726, orientation='minus', exon_count=33, OMIM=608465, no_hdr0=''),
  23065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23065, CurrentID=0, Status='live', Symbol='EMC1', Aliases='KIAA0090', description='ER membrane protein complex subunit 1', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19215664, end_position_on_the_genomic_accession=19251559, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  23066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23066, CurrentID=0, Status='live', Symbol='CAND2', Aliases='TIP120B, Tp120b', description='cullin-associated and neddylation-dissociated 2 (putative)', other_designations='TBP interacting protein|TBP-interacting protein 120B|TBP-interacting protein of 120 kDa B|cullin-associated and neddylation-dissociated protein 2|epididymis tissue protein Li 169|p120 CAND2', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12796653, end_position_on_the_genomic_accession=12834818, orientation='plus', exon_count=17, OMIM=610403, no_hdr0=''),
  23067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23067, CurrentID=0, Status='live', Symbol='SETD1B', Aliases='KMT2G, Set1B', description='SET domain containing 1B', other_designations='SET domain-containing protein 1B|hSET1B|lysine N-methyltransferase 2G', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121803530, end_position_on_the_genomic_accession=121832656, orientation='plus', exon_count=19, OMIM=611055, no_hdr0=''),
  154141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154141, CurrentID=0, Status='live', Symbol='MBOAT1', Aliases='LPEAT1, LPLAT, LPLAT 1, LPSAT, OACT1, dJ434O11.1', description='membrane bound O-acyltransferase domain containing 1', other_designations='1-acylglycerophosphoserine O-acyltransferase|O-acyltransferase (membrane bound) domain containing 1|O-acyltransferase domain-containing protein 1|lyso-PS acyltransferase|lysophosphatidylethanolamine acyltransferase 1|lysophosphatidylserine acyltransferase|membrane-bound O-acyltransferase domain-containing protein 1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20099684, end_position_on_the_genomic_accession=20212464, orientation='minus', exon_count=15, OMIM=611732, no_hdr0=''),
  23070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23070, CurrentID=0, Status='live', Symbol='CMTR1', Aliases='FTSJD2, KIAA0082, MTr1, hMTr1', description='cap methyltransferase 1', other_designations="FtsJ methyltransferase domain containing 2|ISG95|S-adenosyl-L-methionine-dependent methyltransferase FTSJD2|cap1 2'O-ribose methyltransferase 1|ftsJ methyltransferase domain-containing protein 2|interferon-stimulated gene 95 kDa protein", map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37433131, end_position_on_the_genomic_accession=37482827, orientation='plus', exon_count=25, OMIM=616189, no_hdr0=''),
  23071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23071, CurrentID=0, Status='live', Symbol='ERP44', Aliases='PDIA10, TXNDC4', description='endoplasmic reticulum protein 44', other_designations='ER protein 44|endoplasmic reticulum resident protein 44 kDa|protein disulfide isomerase family A, member 10|thioredoxin domain containing 4 (endoplasmic reticulum)|thioredoxin domain-containing protein 4', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99979179, end_position_on_the_genomic_accession=100099052, orientation='minus', exon_count=12, OMIM=609170, no_hdr0=''),
  23072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23072, CurrentID=0, Status='live', Symbol='HECW1', Aliases='NEDL1', description='HECT, C2 and WW domain containing E3 ubiquitin protein ligase 1', other_designations='HECT type E3 ubiquitin ligase|NEDD4-like ubiquitin-protein ligase 1', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43112599, end_position_on_the_genomic_accession=43563654, orientation='plus', exon_count=35, OMIM=610384, no_hdr0=''),
  23074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23074, CurrentID=0, Status='live', Symbol='UHRF1BP1L', Aliases='SHIP164', description='UHRF1 binding protein 1-like', other_designations='UHRF1 (ICBP90) binding protein 1-like|syntaxin 6Habc-interacting protein of 164 kDa', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100036794, end_position_on_the_genomic_accession=100142864, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  23075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23075, CurrentID=0, Status='live', Symbol='SWAP70', Aliases='HSPC321, SWAP-70', description='SWAP switching B-cell complex 70kDa subunit', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9664077, end_position_on_the_genomic_accession=9752991, orientation='plus', exon_count=13, OMIM=604762, no_hdr0=''),
  285220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285220, CurrentID=0, Status='live', Symbol='EPHA6', Aliases='EHK-2, EHK2, EK12, EPA6, HEK12, PRO57066', description='EPH receptor A6', other_designations='EPH homology kinase 2|EPH-like kinase 12|ephrin receptor EphA6', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96814581, end_position_on_the_genomic_accession=97764977, orientation='plus', exon_count=32, OMIM=600066, no_hdr0=''),
  23077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23077, CurrentID=0, Status='live', Symbol='MYCBP2', Aliases='PAM', description='MYC binding protein 2, E3 ubiquitin protein ligase', other_designations='myc-binding protein 2|pam/highwire/rpm-1 protein|probable E3 ubiquitin-protein ligase MYCBP2|protein associated with Myc', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77044657, end_position_on_the_genomic_accession=77327042, orientation='minus', exon_count=89, OMIM=610392, no_hdr0=''),
  154150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154150, CurrentID=0, Status='live', Symbol='HDGFL1', Aliases='PWWP1, dJ309H15.1', description='hepatoma derived growth factor-like 1', other_designations='HDGF (hepatoma-derived growth factor) like|PWWP domain containing 1|PWWP domain-containing protein 1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22569449, end_position_on_the_genomic_accession=22570521, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442713, CurrentID=0, Status='live', Symbol='LOC442713', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98453905, end_position_on_the_genomic_accession=98454648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23080, CurrentID=0, Status='live', Symbol='AVL9', Aliases='KIAA0241', description='AVL9 homolog (S. cerevisiase)', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32495426, end_position_on_the_genomic_accession=32588741, orientation='plus', exon_count=18, OMIM=612927, no_hdr0=''),
  23081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23081, CurrentID=0, Status='live', Symbol='KDM4C', Aliases='GASC1, JHDM3C, JMJD2C, TDRD14C', description='lysine (K)-specific demethylase 4C', other_designations='JmjC domain-containing histone demethylation protein 3C|gene amplified in squamous cell carcinoma 1 protein|jumonji domain-containing protein 2C|tudor domain containing 14C', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6720863, end_position_on_the_genomic_accession=7175648, orientation='plus', exon_count=32, OMIM=605469, no_hdr0=''),
  23082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23082, CurrentID=0, Status='live', Symbol='PPRC1', Aliases='PRC', description='peroxisome proliferator-activated receptor gamma, coactivator-related 1', other_designations='PGC-1-related coactivator', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102132994, end_position_on_the_genomic_accession=102150333, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  23085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23085, CurrentID=0, Status='live', Symbol='ERC1', Aliases='Cast2, ELKS, ERC-1, RAB6IP2', description='ELKS/RAB6-interacting/CAST family member 1', other_designations='RAB6 interacting protein 2', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=991208, end_position_on_the_genomic_accession=1495933, orientation='plus', exon_count=25, OMIM=607127, no_hdr0=''),
  23086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23086, CurrentID=0, Status='live', Symbol='EXPH5', Aliases='SLAC2-B, SLAC2B', description='exophilin 5', other_designations='slp homolog lacking C2 domains b|synaptotagmin-like homologue lacking C2 domains b|synaptotagmin-like protein homolog lacking C2 domains b', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108505431, end_position_on_the_genomic_accession=108593765, orientation='minus', exon_count=11, OMIM=612878, no_hdr0=''),
  23087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23087, CurrentID=0, Status='live', Symbol='TRIM35', Aliases='HLS5, MAIR', description='tripartite motif containing 35', other_designations='hemopoietic lineage switch protein 5|tripartite motif-containing 35', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27284886, end_position_on_the_genomic_accession=27311319, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  285232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285232, CurrentID=0, Status='live', Symbol='LOC285232', Aliases='', description='phosphoribosyl pyrophosphate amidotransferase pseudogene', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87049198, end_position_on_the_genomic_accession=87052869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23089, CurrentID=0, Status='live', Symbol='PEG10', Aliases='EDR, HB-1, MEF3L, Mar2, Mart2, RGAG3', description='paternally expressed 10', other_designations='MEF3 like 1|embryonal carcinoma differentiation regulated|mammalian retrotransposon-derived protein 2|myelin expression factor 3-like protein 1|paternally expressed gene 10 protein|retrotransposon gag domain-containing protein 3|retrotransposon-derived gag-like polyprotein|ty3/Gypsy-like protein', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94656325, end_position_on_the_genomic_accession=94669695, orientation='plus', exon_count=2, OMIM=609810, no_hdr0=''),
  23090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23090, CurrentID=0, Status='live', Symbol='ZNF423', Aliases='Ebfaz, JBTS19, NPHP14, OAZ, Roaz, ZFP423, Zfp104, hOAZ', description='zinc finger protein 423', other_designations='OLF-1/EBF associated zinc finger|Smad- and Olf-interacting zinc finger protein|early B-cell factor associated zinc finger protein|olf1/EBF-associated zinc finger protein', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49487531, end_position_on_the_genomic_accession=49857919, orientation='minus', exon_count=15, OMIM=604557, no_hdr0=''),
  23091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23091, CurrentID=0, Status='live', Symbol='ZC3H13', Aliases='KIAA0853', description='zinc finger CCCH-type containing 13', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45955684, end_position_on_the_genomic_accession=46052761, orientation='minus', exon_count=21, OMIM=616453, no_hdr0=''),
  23092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23092, CurrentID=0, Status='live', Symbol='ARHGAP26', Aliases='GRAF, GRAF1, OPHN1L, OPHN1L1', description='Rho GTPase activating protein 26', other_designations='GTPase regulator associated with focal adhesion kinase pp125(FAK)|oligophrenin-1-like protein', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142770380, end_position_on_the_genomic_accession=143229007, orientation='plus', exon_count=37, OMIM=605370, no_hdr0=''),
  285237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285237, CurrentID=0, Status='live', Symbol='C3orf38', Aliases='', description='chromosome 3 open reading frame 38', other_designations='', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88149743, end_position_on_the_genomic_accession=88157965, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23094, CurrentID=0, Status='live', Symbol='SIPA1L3', Aliases='SPAL3', description='signal-induced proliferation-associated 1 like 3', other_designations='SIPA1-like protein 3|SPA-1-like 3|SPA-1-like protein 3', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37907221, end_position_on_the_genomic_accession=38208372, orientation='plus', exon_count=25, OMIM=616655, no_hdr0=''),
  23095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23095, CurrentID=0, Status='live', Symbol='KIF1B', Aliases='CMT2, CMT2A, CMT2A1, HMSNII, KLP, NBLST1', description='kinesin family member 1B', other_designations='kinesin superfamily protein KIF1B', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10210706, end_position_on_the_genomic_accession=10381603, orientation='plus', exon_count=48, OMIM=605995, no_hdr0=''),
  23096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23096, CurrentID=0, Status='live', Symbol='IQSEC2', Aliases='BRAG1, MRX1', description='IQ motif and Sec7 domain 2', other_designations='brefeldin A resistant Arf-guanine nucleotide exchange factor 1|brefeldin A resistant Arf-guanine nucleotide exchange factor 1b|brefeldin A resistant Arf-guanine nucleotide exchange factor 1c', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53227454, end_position_on_the_genomic_accession=53321350, orientation='minus', exon_count=23, OMIM=300522, no_hdr0=''),
  23097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23097, CurrentID=0, Status='live', Symbol='CDK19', Aliases='CDC2L6, CDK11, bA346C16.3', description='cyclin-dependent kinase 19', other_designations='CDC2-related protein kinase 6|CDK8-like cyclin-dependent kinase|cell division cycle 2-like 6 (CDK8-like)|cell division cycle 2-like protein kinase 6|cell division protein kinase 19|cyclin dependent kinase 19 variant 2|cyclin-dependent kinase (CDC2-like) 11|cyclin-dependent kinase 11|death-preventing kinase', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110609978, end_position_on_the_genomic_accession=110816516, orientation='minus', exon_count=18, OMIM=614720, no_hdr0=''),
  285242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285242, CurrentID=0, Status='live', Symbol='HTR3E', Aliases='5-HT3-E, 5-HT3E, 5-HT3c1', description='5-hydroxytryptamine (serotonin) receptor 3E, ionotropic', other_designations='5-hydroxytryptamine (serotonin) receptor 3, family member E|5-hydroxytryptamine receptor 3 subunit E', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184097064, end_position_on_the_genomic_accession=184106995, orientation='plus', exon_count=9, OMIM=610123, no_hdr0=''),
  23099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23099, CurrentID=0, Status='live', Symbol='ZBTB43', Aliases='ZBTB22B, ZNF-X, ZNF297B', description='zinc finger and BTB domain containing 43', other_designations='zinc finger and BTB domain-containing protein 22B|zinc finger protein 297B', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126804072, end_position_on_the_genomic_accession=126838210, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  23101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23101, CurrentID=0, Status='live', Symbol='MCF2L2', Aliases='ARHGEF22', description='MCF.2 cell line derived transforming sequence-like 2', other_designations='MCF2-transforming sequence-like protein 2|dbs-related Rho family guanine nucleotide exchange factor', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183178004, end_position_on_the_genomic_accession=183428619, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  23102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23102, CurrentID=0, Status='live', Symbol='TBC1D2B', Aliases='', description='TBC1 domain family member 2B', other_designations='TBC1 domain family, member 2B', map_location='15q24.3-q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77994985, end_position_on_the_genomic_accession=78077667, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  23105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23105, CurrentID=0, Status='live', Symbol='FSTL4', Aliases='', description='follistatin-like 4', other_designations='follistatin-like protein 4', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133194827, end_position_on_the_genomic_accession=133842066, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  23107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23107, CurrentID=0, Status='live', Symbol='MRPS27', Aliases='MRP-S27, S27mt', description='mitochondrial ribosomal protein S27', other_designations='mitochondrial 28S ribosomal protein S27', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72219409, end_position_on_the_genomic_accession=72320257, orientation='minus', exon_count=12, OMIM=611989, no_hdr0=''),
  23108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23108, CurrentID=0, Status='live', Symbol='RAP1GAP2', Aliases='GARNL4, RAP1GA3', description='RAP1 GTPase activating protein 2', other_designations='GTPase activating RANGAP domain-like 4|GTPase activating Rap/RanGAP domain-like 4|GTPase-activating Rap/Ran-GAP domain-like protein 4', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2755698, end_position_on_the_genomic_accession=3037741, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  23109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23109, CurrentID=0, Status='live', Symbol='DDN', Aliases='', description='dendrin', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48995149, end_position_on_the_genomic_accession=49001091, orientation='minus', exon_count=3, OMIM=610588, no_hdr0=''),
  23111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23111, CurrentID=0, Status='live', Symbol='SPG20', Aliases='SPARTIN, TAHCCP1', description='spastic paraplegia 20 (Troyer syndrome)', other_designations='trans-activated by hepatitis C virus core protein 1', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36301638, end_position_on_the_genomic_accession=36370180, orientation='minus', exon_count=14, OMIM=607111, no_hdr0=''),
  23112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23112, CurrentID=0, Status='live', Symbol='TNRC6B', Aliases='', description='trinucleotide repeat containing 6B', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40044817, end_position_on_the_genomic_accession=40335808, orientation='plus', exon_count=27, OMIM=610740, no_hdr0=''),
  23113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23113, CurrentID=0, Status='live', Symbol='CUL9', Aliases='H7AP1, PARC', description='cullin 9', other_designations='CUL-9|UbcH7-associated protein 1|p53-associated parkin-like cytoplasmic protein|parkin-like cytoplasmic p53 binding protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43182174, end_position_on_the_genomic_accession=43224587, orientation='plus', exon_count=42, OMIM=607489, no_hdr0=''),
  23114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23114, CurrentID=0, Status='live', Symbol='NFASC', Aliases='NF, NRCAML', description='neurofascin', other_designations='neurofascin homolog', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204828654, end_position_on_the_genomic_accession=205022822, orientation='plus', exon_count=35, OMIM=609145, no_hdr0=''),
  23116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23116, CurrentID=0, Status='live', Symbol='FAM179B', Aliases='KIAA0423', description='family with sequence similarity 179 member B', other_designations='family with sequence similarity 179, member B', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44962213, end_position_on_the_genomic_accession=45074431, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  23117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23117, CurrentID=0, Status='live', Symbol='NPIPB3', Aliases='KIAA0220L, NPIP, NPIPB, NPIPB5, NPIPL3', description='nuclear pore complex interacting protein family, member B3', other_designations='KIAA0220-like protein|PI-3-kinase-related kinase SMG-1 isoform 1 homolog|nuclear pore complex-interacting protein B type|protein pps22-1', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21402134, end_position_on_the_genomic_accession=21425337, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23118, CurrentID=0, Status='live', Symbol='TAB2', Aliases='CHTD2, MAP3K7IP2, TAB-2', description='TGF-beta activated kinase 1/MAP3K7 binding protein 2', other_designations='TAK1-binding protein 2|mitogen-activated protein kinase kinase kinase 7-interacting protein 2', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149217924, end_position_on_the_genomic_accession=149411613, orientation='plus', exon_count=15, OMIM=605101, no_hdr0=''),
  23119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23119, CurrentID=0, Status='live', Symbol='HIC2', Aliases='HRG22, ZBTB30, ZNF907', description='hypermethylated in cancer 2', other_designations='HIC1-related gene on chromosome 22 protein|hic-2|hic-3|zinc finger and BTB domain-containing protein 30', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21417404, end_position_on_the_genomic_accession=21451463, orientation='plus', exon_count=6, OMIM=607712, no_hdr0=''),
  23120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23120, CurrentID=0, Status='live', Symbol='ATP10B', Aliases='ATPVB', description='ATPase, class V, type 10B', other_designations='P4-ATPase flippase complex alpha subunit ATP10B', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160563120, end_position_on_the_genomic_accession=160929368, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  23122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23122, CurrentID=0, Status='live', Symbol='CLASP2', Aliases='', description='cytoplasmic linker associated protein 2', other_designations='CLIP-associating protein CLASP2|cytoplasmic linker-associated protein 2|multiple asters (Mast)-like homolog 2|protein Orbit homolog 2', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33496245, end_position_on_the_genomic_accession=33718356, orientation='minus', exon_count=46, OMIM=605853, no_hdr0=''),
  285267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285267, CurrentID=0, Status='live', Symbol='ZNF619', Aliases='', description='zinc finger protein 619', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40477113, end_position_on_the_genomic_accession=40490237, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  285268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285268, CurrentID=0, Status='live', Symbol='ZNF621', Aliases='', description='zinc finger protein 621', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40524316, end_position_on_the_genomic_accession=40539794, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  23125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23125, CurrentID=0, Status='live', Symbol='CAMTA2', Aliases='', description='calmodulin binding transcription activator 2', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4967992, end_position_on_the_genomic_accession=4987673, orientation='minus', exon_count=24, OMIM=611508, no_hdr0=''),
  23126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23126, CurrentID=0, Status='live', Symbol='POGZ', Aliases='MRD37, ZNF280E, ZNF635, ZNF635m', description='pogo transposable element with ZNF domain', other_designations='putative protein product of Nbla00003|zinc finger protein 280E|zinc finger protein 635', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151402724, end_position_on_the_genomic_accession=151459465, orientation='minus', exon_count=24, OMIM=614787, no_hdr0=''),
  23127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23127, CurrentID=0, Status='live', Symbol='COLGALT2', Aliases='C1orf17, GLT25D2', description='collagen beta(1-O)galactosyltransferase 2', other_designations='glycosyltransferase 25 domain containing 2|glycosyltransferase 25 family member 2|hydroxylysine galactosyltransferase 2', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183929662, end_position_on_the_genomic_accession=184051611, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  23129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23129, CurrentID=0, Status='live', Symbol='PLXND1', Aliases='PLEXD1', description='plexin D1', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129555213, end_position_on_the_genomic_accession=129606739, orientation='minus', exon_count=37, OMIM=604282, no_hdr0=''),
  23130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23130, CurrentID=0, Status='live', Symbol='ATG2A', Aliases='', description='autophagy related 2A', other_designations='ATG2 autophagy related 2 homolog A', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64894532, end_position_on_the_genomic_accession=64917250, orientation='minus', exon_count=41, OMIM=616225, no_hdr0=''),
  23131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23131, CurrentID=0, Status='live', Symbol='GPATCH8', Aliases='GPATC8, KIAA0553', description='G-patch domain containing 8', other_designations='G patch domain containing 8', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44395277, end_position_on_the_genomic_accession=44503602, orientation='minus', exon_count=13, OMIM=614396, no_hdr0=''),
  23132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23132, CurrentID=0, Status='live', Symbol='RAD54L2', Aliases='ARIP4, HSPC325, SRISNF2L', description='RAD54-like 2 (S. cerevisiae)', other_designations='AR interacting protein 4|RAD54-like protein 2|androgen receptor-interacting protein 4', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51538683, end_position_on_the_genomic_accession=51668660, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  23133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23133, CurrentID=0, Status='live', Symbol='PHF8', Aliases='JHDM1F, MRXSSD, ZNF422', description='PHD finger protein 8', other_designations='jumonji C domain-containing histone demethylase 1F', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53936680, end_position_on_the_genomic_accession=54048935, orientation='minus', exon_count=28, OMIM=300560, no_hdr0=''),
  23135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23135, CurrentID=0, Status='live', Symbol='KDM6B', Aliases='JMJD3', description='lysine (K)-specific demethylase 6B', other_designations='jmjC domain-containing protein 3|jumonji domain containing 3, histone lysine demethylase|jumonji domain-containing protein 3|lysine demethylase 6B', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7834200, end_position_on_the_genomic_accession=7854800, orientation='plus', exon_count=30, OMIM=611577, no_hdr0=''),
  23136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23136, CurrentID=0, Status='live', Symbol='EPB41L3', Aliases='4.1B, DAL-1, DAL1', description='erythrocyte membrane protein band 4.1-like 3', other_designations='differentially expressed in adenocarcinoma of the lung protein 1', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5392381, end_position_on_the_genomic_accession=5630666, orientation='minus', exon_count=33, OMIM=605331, no_hdr0=''),
  23137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23137, CurrentID=0, Status='live', Symbol='SMC5', Aliases='SMC5L1', description='structural maintenance of chromosomes 5', other_designations='SMC protein 5|SMC-5|SMC5 structural maintenance of chromosomes 5-like 1|hSMC5', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70258962, end_position_on_the_genomic_accession=70354873, orientation='plus', exon_count=25, OMIM=609386, no_hdr0=''),
  285282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285282, CurrentID=0, Status='live', Symbol='RABL3', Aliases='', description='RAB, member of RAS oncogene family-like 3', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120686681, end_position_on_the_genomic_accession=120742537, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23139, CurrentID=0, Status='live', Symbol='MAST2', Aliases='MAST205, MTSSK', description='microtubule associated serine/threonine kinase 2', other_designations='microtubule associated testis specific serine/threonine protein kinase', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45786992, end_position_on_the_genomic_accession=46036124, orientation='plus', exon_count=37, OMIM=612257, no_hdr0=''),
  23140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23140, CurrentID=0, Status='live', Symbol='ZZEF1', Aliases='ZZZ4', description='zinc finger, ZZ-type with EF-hand domain 1', other_designations='zinc finger, ZZ-type with EF hand domain 1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4004445, end_position_on_the_genomic_accession=4143020, orientation='minus', exon_count=56, OMIM='', no_hdr0=''),
  23141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23141, CurrentID=0, Status='live', Symbol='ANKLE2', Aliases='KIAA0692, LEMD7, Lem4', description='ankyrin repeat and LEM domain containing 2', other_designations='LEM domain containing 7|LEM domain-containing protein 4', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132725667, end_position_on_the_genomic_accession=132761888, orientation='minus', exon_count=17, OMIM=616062, no_hdr0=''),
  154214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154214, CurrentID=0, Status='live', Symbol='RNF217', Aliases='C6orf172, IBRDC1, dJ84N20.1', description='ring finger protein 217', other_designations='IBR domain containing 1', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=124962486, end_position_on_the_genomic_accession=125092633, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  154215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154215, CurrentID=0, Status='live', Symbol='NKAIN2', Aliases='FAM77B, NKAIP2, TCBA, TCBA1', description='Na+/K+ transporting ATPase interacting 2', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 2|T-cell lymphoma breakpoint-associated target protein 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123803837, end_position_on_the_genomic_accession=124825640, orientation='plus', exon_count=12, OMIM=609758, no_hdr0=''),
  23144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23144, CurrentID=0, Status='live', Symbol='ZC3H3', Aliases='ZC3HDC3', description='zinc finger CCCH-type containing 3', other_designations='smad-interacting CPSF-like factor|zinc finger CCCH type domain containing 3|zinc finger CCCH-type domain containing 3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143437655, end_position_on_the_genomic_accession=143542245, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  23145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23145, CurrentID=0, Status='live', Symbol='SSPO', Aliases='', description='SCO-spondin', other_designations='SCO protein, thrombospondin domain containing|SCO-spondin homolog (Bos taurus)|subcommissural organ spondin', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149776042, end_position_on_the_genomic_accession=149833965, orientation='plus', exon_count=103, OMIM='', no_hdr0=''),
  106480663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480663, CurrentID=0, Status='live', Symbol='RNU1-46P', Aliases='', description='RNA, U1 small nuclear 46, pseudogene', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23148, CurrentID=0, Status='live', Symbol='NACAD', Aliases='', description='NAC alpha domain containing', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45080437, end_position_on_the_genomic_accession=45088961, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23149, CurrentID=0, Status='live', Symbol='FCHO1', Aliases='', description='FCH domain only 1', other_designations='FCH domain only protein 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17747718, end_position_on_the_genomic_accession=17788568, orientation='plus', exon_count=30, OMIM=613437, no_hdr0=''),
  23150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23150, CurrentID=0, Status='live', Symbol='FRMD4B', Aliases='6030440G05Rik, GRSP1', description='FERM domain containing 4B', other_designations='GRP1-binding protein GRSP1', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69168782, end_position_on_the_genomic_accession=69541340, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  23151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23151, CurrentID=0, Status='live', Symbol='GRAMD4', Aliases='DIP, dA59H18.1, dJ439F8.1', description='GRAM domain containing 4', other_designations='death-inducing protein|death-inducing-protein', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46576017, end_position_on_the_genomic_accession=46682758, orientation='plus', exon_count=24, OMIM=613691, no_hdr0=''),
  23152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23152, CurrentID=0, Status='live', Symbol='CIC', Aliases='', description='capicua transcriptional repressor', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42268537, end_position_on_the_genomic_accession=42295796, orientation='plus', exon_count=23, OMIM=612082, no_hdr0=''),
  23154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23154, CurrentID=0, Status='live', Symbol='NCDN', Aliases='', description='neurochondrin', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35557792, end_position_on_the_genomic_accession=35566779, orientation='plus', exon_count=8, OMIM=608458, no_hdr0=''),
  23155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23155, CurrentID=0, Status='live', Symbol='CLCC1', Aliases='MCLC', description='chloride channel CLIC-like 1', other_designations='Mid-1-related chloride channel protein 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108929508, end_position_on_the_genomic_accession=108963499, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  285300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285300, CurrentID=0, Status='live', Symbol='LOC285300', Aliases='', description='uncharacterized LOC285300', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23157, CurrentID=0, Status='live', Symbol='SEPT6', Aliases='SEP2, SEPT2', description='septin 6', other_designations='septin 2', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119615724, end_position_on_the_genomic_accession=119693370, orientation='minus', exon_count=13, OMIM=300683, no_hdr0=''),
  23158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23158, CurrentID=0, Status='live', Symbol='TBC1D9', Aliases='MDR1', description='TBC1 domain family member 9', other_designations='TBC1 domain family member 9A|TBC1 domain family, member 9 (with GRAM domain)', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140620782, end_position_on_the_genomic_accession=140756317, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100270864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270864, CurrentID=0, Status='live', Symbol='RPS6P2', Aliases='RPS6_1_301', description='ribosomal protein S6 pseudogene 2', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179002118, end_position_on_the_genomic_accession=179002772, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23160, CurrentID=0, Status='live', Symbol='WDR43', Aliases='NET12, UTP5', description='WD repeat domain 43', other_designations='UTP5, small subunit (SSU) processome component, homolog', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28894643, end_position_on_the_genomic_accession=28948220, orientation='plus', exon_count=18, OMIM=616195, no_hdr0=''),
  23161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23161, CurrentID=0, Status='live', Symbol='SNX13', Aliases='RGS-PX1', description='sorting nexin 13', other_designations='rgs domain- and phox domain-containing protein', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17790761, end_position_on_the_genomic_accession=17940518, orientation='minus', exon_count=33, OMIM=606589, no_hdr0=''),
  23162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23162, CurrentID=0, Status='live', Symbol='MAPK8IP3', Aliases='JIP-3, JIP3, JSAP1, SYD2, syd', description='mitogen-activated protein kinase 8 interacting protein 3', other_designations='C-jun-amino-terminal kinase interacting protein 3|JNK MAP kinase scaffold protein 3|JNK-interacting protein 3|JNK/SAPK-associated protein-1|JNK/stress-activated protein kinase-associated protein 1|homolog of Drosophila Sunday driver 2|mitogen-activated protein kinase 8-interacting protein 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1706183, end_position_on_the_genomic_accession=1770317, orientation='plus', exon_count=34, OMIM=605431, no_hdr0=''),
  23163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23163, CurrentID=0, Status='live', Symbol='GGA3', Aliases='', description='golgi-associated, gamma adaptin ear containing, ARF binding protein 3', other_designations='Golgi-localized, gamma ear-containing, ARF-binding protein 3', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75225480, end_position_on_the_genomic_accession=75262393, orientation='minus', exon_count=21, OMIM=606006, no_hdr0=''),
  23164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23164, CurrentID=0, Status='live', Symbol='MPRIP', Aliases='M-RIP, MRIP, RHOIP3, RIP3, p116Rip', description='myosin phosphatase Rho interacting protein', other_designations='Rho interacting protein 3', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17042468, end_position_on_the_genomic_accession=17192648, orientation='plus', exon_count=25, OMIM=612935, no_hdr0=''),
  23165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23165, CurrentID=0, Status='live', Symbol='NUP205', Aliases='C7orf14', description='nucleoporin 205kDa', other_designations='205 kDa nucleoporin|nucleoporin Nup205', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135557914, end_position_on_the_genomic_accession=135648753, orientation='plus', exon_count=43, OMIM=614352, no_hdr0=''),
  23166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23166, CurrentID=0, Status='live', Symbol='STAB1', Aliases='CLEVER-1, FEEL-1, FELE-1, FEX1, STAB-1', description='stabilin 1', other_designations='MS-1 antigen|common lymphatic endothelial and vascular endothelial receptor-1|fasciclin egf-like, laminin-type egf-like, and link domain-containing scavenger receptor-1|fasciclin, EGF-like, laminin-type EGF-like and link domain-containing scavenger receptor 1', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52495338, end_position_on_the_genomic_accession=52524495, orientation='plus', exon_count=69, OMIM=608560, no_hdr0=''),
  285311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285311, CurrentID=0, Status='live', Symbol='C3orf56', Aliases='', description='chromosome 3 open reading frame 56', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127193131, end_position_on_the_genomic_accession=127198185, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23168, CurrentID=0, Status='live', Symbol='RTF1', Aliases='GTL7, KIAA0252', description='RTF1 homolog, Paf1/RNA polymerase II complex component', other_designations='Rtf1, Paf1/RNA polymerase II complex component, homolog|ortholog of mouse gene trap locus 7', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41417104, end_position_on_the_genomic_accession=41483563, orientation='plus', exon_count=18, OMIM=611633, no_hdr0=''),
  23169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23169, CurrentID=0, Status='live', Symbol='SLC35D1', Aliases='UGTREL7', description='solute carrier family 35 (UDP-GlcA/UDP-GalNAc transporter), member D1', other_designations='UDP-GlcA/UDP-GalNAc transporter|UDP-galactose transporter-related 7|UDP-galactose transporter-related protein 7|solute carrier family 35 (UDP-glucuronic acid/UDP-N-acetylgalactosamine dual transporter), member D1|solute carrier family 35 member D1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66999332, end_position_on_the_genomic_accession=67054397, orientation='minus', exon_count=13, OMIM=610804, no_hdr0=''),
  285314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285314, CurrentID=0, Status='live', Symbol='PABPC1P10', Aliases='D3S3192, PAB1, PABPL1', description='poly(A) binding protein, cytoplasmic 1 pseudogene 10', other_designations='poly(A) binding protein-like 1', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155309427, end_position_on_the_genomic_accession=155309807, orientation='minus', exon_count=0, OMIM=173865, no_hdr0=''),
  285315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285315, CurrentID=0, Status='live', Symbol='C3orf33', Aliases='AC3-33', description='chromosome 3 open reading frame 33', other_designations='AP-1 activity suppressor', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155762612, end_position_on_the_genomic_accession=155806287, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23172, CurrentID=0, Status='live', Symbol='FAM175B', Aliases='ABRO1, KIAA0157', description='family with sequence similarity 175 member B', other_designations='abraxas brother 1|abraxas brother protein 1|family with sequence similarity 175, member B', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124801785, end_position_on_the_genomic_accession=124836670, orientation='plus', exon_count=9, OMIM=611144, no_hdr0=''),
  23173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23173, CurrentID=0, Status='live', Symbol='METAP1', Aliases='MAP1A, MetAP1A', description='methionyl aminopeptidase 1', other_designations='MAP 1|metAP 1|peptidase M 1', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98995637, end_position_on_the_genomic_accession=99062809, orientation='plus', exon_count=15, OMIM=610151, no_hdr0=''),
  23174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23174, CurrentID=0, Status='live', Symbol='ZCCHC14', Aliases='BDG-29, BDG29', description='zinc finger, CCHC domain containing 14', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87406246, end_position_on_the_genomic_accession=87492269, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23175, CurrentID=0, Status='live', Symbol='LPIN1', Aliases='PAP1', description='lipin 1', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11677545, end_position_on_the_genomic_accession=11827409, orientation='plus', exon_count=29, OMIM=605518, no_hdr0=''),
  23176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23176, CurrentID=0, Status='live', Symbol='SEPT8', Aliases='SEP2', description='septin 8', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132750817, end_position_on_the_genomic_accession=132778284, orientation='minus', exon_count=13, OMIM=608418, no_hdr0=''),
  23177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23177, CurrentID=0, Status='live', Symbol='CEP68', Aliases='KIAA0582', description='centrosomal protein 68kDa', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65056361, end_position_on_the_genomic_accession=65087008, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  23178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23178, CurrentID=0, Status='live', Symbol='PASK', Aliases='PASKIN, STK37', description='PAS domain containing serine/threonine kinase', other_designations='per-arnt-sim (PAS) domain kinase', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241106099, end_position_on_the_genomic_accession=241150364, orientation='minus', exon_count=26, OMIM=607505, no_hdr0=''),
  23179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23179, CurrentID=0, Status='live', Symbol='RGL1', Aliases='RGL', description='ral guanine nucleotide dissociation stimulator-like 1', other_designations='ralGDS-like 1', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183636047, end_position_on_the_genomic_accession=183928551, orientation='plus', exon_count=23, OMIM=605667, no_hdr0=''),
  23180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23180, CurrentID=0, Status='live', Symbol='RFTN1', Aliases='MIG2, PIB10, PIG9, RAFTLIN', description='raftlin, lipid raft linker 1', other_designations='cell migration-inducing gene 2 protein|proliferation-inducing protein 10|proliferation-inducing protein 9|raft-linking protein', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16315845, end_position_on_the_genomic_accession=16513715, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23181, CurrentID=0, Status='live', Symbol='DIP2A', Aliases='C21orf106, DIP2', description='disco interacting protein 2 homolog A', other_designations='DIP2 disco-interacting protein 2 homolog A|DIP2 homolog A|disco-interacting protein 2A', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46458942, end_position_on_the_genomic_accession=46584263, orientation='plus', exon_count=45, OMIM=607711, no_hdr0=''),
  285326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285326, CurrentID=0, Status='live', Symbol='LINC00692', Aliases='', description='long intergenic non-protein coding RNA 692', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25858532, end_position_on_the_genomic_accession=25873695, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23184, CurrentID=0, Status='live', Symbol='MESDC2', Aliases='BOCA, MESD', description='mesoderm development candidate 2', other_designations='mesoderm development protein|renal carcinoma antigen NY-REN-61', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80947326, end_position_on_the_genomic_accession=80989878, orientation='minus', exon_count=5, OMIM=607783, no_hdr0=''),
  23185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23185, CurrentID=0, Status='live', Symbol='LARP4B', Aliases='KIAA0217, LARP5', description='La ribonucleoprotein domain family member 4B', other_designations='La ribonucleoprotein domain family, member 4B|La ribonucleoprotein domain family, member 5|la-related protein 5', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=806914, end_position_on_the_genomic_accession=988323, orientation='minus', exon_count=26, OMIM=616513, no_hdr0=''),
  23186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23186, CurrentID=0, Status='live', Symbol='RCOR1', Aliases='COREST, RCOR', description='REST corepressor 1', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102592639, end_position_on_the_genomic_accession=102730576, orientation='plus', exon_count=13, OMIM=607675, no_hdr0=''),
  285331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285331, CurrentID=0, Status='live', Symbol='CCDC66', Aliases='', description='coiled-coil domain containing 66', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56557111, end_position_on_the_genomic_accession=56621836, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  23189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23189, CurrentID=0, Status='live', Symbol='KANK1', Aliases='ANKRD15, CPSQ2, KANK', description='KN motif and ankyrin repeat domains 1', other_designations='ankyrin repeat domain-containing protein 15|kidney ankyrin repeat-containing protein', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=470294, end_position_on_the_genomic_accession=746106, orientation='plus', exon_count=29, OMIM=607704, no_hdr0=''),
  23190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23190, CurrentID=0, Status='live', Symbol='UBXN4', Aliases='UBXD2, UBXDC1, erasin', description='UBX domain protein 4', other_designations='UBX domain containing 2|UBX domain-containing 1|UBX domain-containing protein 2', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135741619, end_position_on_the_genomic_accession=135785063, orientation='plus', exon_count=13, OMIM=611216, no_hdr0=''),
  23191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23191, CurrentID=0, Status='live', Symbol='CYFIP1', Aliases='P140SRA-1, SHYC, SRA-1, SRA1', description='cytoplasmic FMR1 interacting protein 1', other_designations='cytoplasmic FMRP interacting protein 1|selective hybridizing clone|specifically Rac1-associated protein 1', map_location='15q11', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22869465, end_position_on_the_genomic_accession=22980419, orientation='minus', exon_count=33, OMIM=606322, no_hdr0=''),
  23192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23192, CurrentID=0, Status='live', Symbol='ATG4B', Aliases='APG4B, AUTL1', description='autophagy related 4B, cysteine peptidase', other_designations='APG4 autophagy 4 homolog B|ATG4 autophagy related 4 homolog B|AUT-like 1 cysteine endopeptidase|autophagin-1|autophagy-related cysteine endopeptidase 1|autophagy-related protein 4 homolog B|hAPG4B', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241637612, end_position_on_the_genomic_accession=241673857, orientation='plus', exon_count=16, OMIM=611338, no_hdr0=''),
  23193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23193, CurrentID=0, Status='live', Symbol='GANAB', Aliases='G2AN, GLUII', description='glucosidase, alpha; neutral AB', other_designations='HEL-S-164nA|alpha glucosidase II alpha subunit|alpha-glucosidase 2|epididymis secretory sperm binding protein Li 164nA|glucosidase II subunit alpha', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62624826, end_position_on_the_genomic_accession=62646726, orientation='minus', exon_count=25, OMIM=104160, no_hdr0=''),
  23194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23194, CurrentID=0, Status='live', Symbol='FBXL7', Aliases='FBL6, FBL7', description='F-box and leucine-rich repeat protein 7', other_designations='F-box protein Fbl7', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15500165, end_position_on_the_genomic_accession=15939796, orientation='plus', exon_count=7, OMIM=605656, no_hdr0=''),
  23195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23195, CurrentID=0, Status='live', Symbol='MDN1', Aliases='Rea1', description='midasin AAA ATPase 1', other_designations='MDN1, midasin homolog|MIDAS-containing protein', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89642775, end_position_on_the_genomic_accession=89827961, orientation='minus', exon_count=105, OMIM='', no_hdr0=''),
  23196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23196, CurrentID=0, Status='live', Symbol='FAM120A', Aliases='C9orf10, DNAPTP1, DNAPTP5, OSSA', description='family with sequence similarity 120A', other_designations='DNA polymerase-transactivated protein 1|DNA polymerase-transactivated protein 5|oxidative stess-associated Src activator|oxidative stress-associated Src activator', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93451696, end_position_on_the_genomic_accession=93566115, orientation='plus', exon_count=21, OMIM=612265, no_hdr0=''),
  23197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23197, CurrentID=0, Status='live', Symbol='FAF2', Aliases='ETEA, UBXD8, UBXN3B', description='Fas associated factor family member 2', other_designations='UBX domain containing 8|UBX domain protein 3B|UBX domain-containing protein 3B|UBX domain-containing protein 8|expressed in T-cells and eosinophils in atopic dermatitis', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176448306, end_position_on_the_genomic_accession=176510074, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23198, CurrentID=0, Status='live', Symbol='PSME4', Aliases='PA200', description='proteasome activator subunit 4', other_designations='proteasome (prosome, macropain) activator subunit 4|proteasome activator 200 kDa|proteasome activator PA200', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53864067, end_position_on_the_genomic_accession=53970840, orientation='minus', exon_count=49, OMIM=607705, no_hdr0=''),
  285343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285343, CurrentID=0, Status='live', Symbol='TCAIM', Aliases='C3orf23, TOAG-1, TOAG1', description='T-cell activation inhibitor, mitochondrial', other_designations='T cell activation inhibitor, mitochondrial|tolerance associated gene-1 protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44338119, end_position_on_the_genomic_accession=44409448, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  23200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23200, CurrentID=0, Status='live', Symbol='ATP11B', Aliases='ATPIF, ATPIR', description='ATPase, class VI, type 11B', other_designations='ATPase IR|P4-ATPase flippase complex alpha subunit ATP11B|truncated ATPase 11B protein', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182793503, end_position_on_the_genomic_accession=182921635, orientation='plus', exon_count=36, OMIM=605869, no_hdr0=''),
  23201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23201, CurrentID=0, Status='live', Symbol='FAM168A', Aliases='KIAA0280, TCRP1', description='family with sequence similarity 168 member A', other_designations='family with sequence similarity 168, member A|tongue cancer chemotherapy resistance-associated protein 1', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73400478, end_position_on_the_genomic_accession=73598189, orientation='minus', exon_count=9, OMIM=616316, no_hdr0=''),
  285346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285346, CurrentID=0, Status='live', Symbol='ZNF852', Aliases='', description='zinc finger protein 852', other_designations='putative zinc finger protein 852', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44498970, end_position_on_the_genomic_accession=44510640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23203, CurrentID=0, Status='live', Symbol='PMPCA', Aliases='Alpha-MPP, INPP5E, P-55', description='peptidase (mitochondrial processing) alpha', other_designations='inositol polyphosphate-5-phosphatase, 72 kD|mitochondrial matrix processing protease, alpha subunit', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136410573, end_position_on_the_genomic_accession=136423761, orientation='plus', exon_count=13, OMIM=613036, no_hdr0=''),
  23204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23204, CurrentID=0, Status='live', Symbol='ARL6IP1', Aliases='AIP1, ARL6IP, ARMER, SPG61', description='ADP ribosylation factor like GTPase 6 interacting protein 1', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 1|ADP-ribosylation factor like GTPase 6 interacting protein 1|ADP-ribosylation factor-like 6 interacting protein 1|ARL-6-interacting protein 1|aip-1|apoptotic regulator in the membrane of the endoplasmic reticulum', map_location='16p12-p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18791669, end_position_on_the_genomic_accession=18801535, orientation='minus', exon_count=6, OMIM=607669, no_hdr0=''),
  285349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285349, CurrentID=0, Status='live', Symbol='ZNF660', Aliases='', description='zinc finger protein 660', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44584849, end_position_on_the_genomic_accession=44599699, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23207, CurrentID=0, Status='live', Symbol='PLEKHM2', Aliases='SKIP', description='pleckstrin homology and RUN domain containing M2', other_designations='PH domain-containing family M member 2|SifA (Salmonella-induced filaments A) and kinesin-interacting protein|novel RUN and PH domain-containing protein|pleckstrin homology domain containing, family M (with RUN domain) member 2|salmonella-induced filaments A and kinesin-interacting protein|sifA and kinesin-interacting protein', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15684267, end_position_on_the_genomic_accession=15734769, orientation='plus', exon_count=21, OMIM=609613, no_hdr0=''),
  23208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23208, CurrentID=0, Status='live', Symbol='SYT11', Aliases='SYT12, sytXI', description='synaptotagmin 11', other_designations='synaptotagmin 12|synaptotagmin XI', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155859469, end_position_on_the_genomic_accession=155885199, orientation='plus', exon_count=4, OMIM=608741, no_hdr0=''),
  23209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23209, CurrentID=0, Status='live', Symbol='MLC1', Aliases='LVM, MLC, VL', description='megalencephalic leukoencephalopathy with subcortical cysts 1', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50059391, end_position_on_the_genomic_accession=50085929, orientation='minus', exon_count=13, OMIM=605908, no_hdr0=''),
  23210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23210, CurrentID=0, Status='live', Symbol='JMJD6', Aliases='PSR, PTDSR, PTDSR1', description='jumonji domain containing 6', other_designations='histone arginine demethylase JMJD6|jmjC domain-containing protein 6|jumonji domain-containing protein 6|lysyl-hydroxylase JMJD6|peptide-lysine 5-dioxygenase JMJD6|phosphatidylserine receptor', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76712832, end_position_on_the_genomic_accession=76726799, orientation='minus', exon_count=8, OMIM=604914, no_hdr0=''),
  23211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23211, CurrentID=0, Status='live', Symbol='ZC3H4', Aliases='C19orf7', description='zinc finger CCCH-type containing 4', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47064187, end_position_on_the_genomic_accession=47113752, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  23212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23212, CurrentID=0, Status='live', Symbol='RRS1', Aliases='', description='ribosome biogenesis regulator homolog', other_designations='RRS1 ribosome biogenesis regulator homolog|homolog of yeast ribosome biogenesis regulator|homolog of yeast ribosome biogenesis regulatory protein RRS1|ribosome biogenesis regulatory protein RRS1 homolog', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66429028, end_position_on_the_genomic_accession=66430733, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23213, CurrentID=0, Status='live', Symbol='SULF1', Aliases='SULF-1', description='sulfatase 1', other_designations='sulfatase FP', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69466624, end_position_on_the_genomic_accession=69660912, orientation='plus', exon_count=25, OMIM=610012, no_hdr0=''),
  23214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23214, CurrentID=0, Status='live', Symbol='XPO6', Aliases='EXP6, RANBP20', description='exportin 6', other_designations='ran-binding protein 20', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28097976, end_position_on_the_genomic_accession=28211960, orientation='minus', exon_count=28, OMIM=608411, no_hdr0=''),
  23215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23215, CurrentID=0, Status='live', Symbol='PRRC2C', Aliases='BAT2-iso, BAT2D1, BAT2L2, XTP2', description='proline-rich coiled-coil 2C', other_designations='BAT2 domain containing 1|BAT2 domain-containing protein 1|HBV X-transactivated gene 2 protein|HBV XAg-transactivated protein 2|HBxAg transactivated protein 2|HLA-B associated transcript 2-like 2|HLA-B-associated transcript 2-like 2|proline-rich and coiled-coil-containing protein 2C|protein BAT2-like 2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171485513, end_position_on_the_genomic_accession=171593511, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  154288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154288, CurrentID=0, Status='live', Symbol='KHDC3L', Aliases='C6orf221, ECAT1, HYDM2', description='KH domain containing 3-like, subcortical maternal complex member', other_designations='ES cell-associated transcript 1 protein', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73362211, end_position_on_the_genomic_accession=73364175, orientation='plus', exon_count=4, OMIM=611687, no_hdr0=''),
  23217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23217, CurrentID=0, Status='live', Symbol='ZFR2', Aliases='KIAA1086', description='zinc finger RNA binding protein 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3804024, end_position_on_the_genomic_accession=3869029, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  23218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23218, CurrentID=0, Status='live', Symbol='NBEAL2', Aliases='BDPLT4, GPS', description='neurobeachin like 2', other_designations='neurobeachin-like 2', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46979683, end_position_on_the_genomic_accession=47009704, orientation='plus', exon_count=57, OMIM=614169, no_hdr0=''),
  23219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23219, CurrentID=0, Status='live', Symbol='FBXO28', Aliases='CENP-30, Fbx28', description='F-box protein 28', other_designations='centromere protein 30', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224114087, end_position_on_the_genomic_accession=224162047, orientation='plus', exon_count=5, OMIM=609100, no_hdr0=''),
  23220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23220, CurrentID=0, Status='live', Symbol='DTX4', Aliases='RNF155', description='deltex 4, E3 ubiquitin ligase', other_designations='RING finger protein 155|deltex 4 homolog|deltex homolog 4|deltex4|protein deltex-4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59171430, end_position_on_the_genomic_accession=59208587, orientation='plus', exon_count=11, OMIM=616110, no_hdr0=''),
  23221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23221, CurrentID=0, Status='live', Symbol='RHOBTB2', Aliases='DBC2', description='Rho-related BTB domain containing 2', other_designations='deleted in breast cancer 2 gene protein|p83', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22987258, end_position_on_the_genomic_accession=23020197, orientation='plus', exon_count=15, OMIM=607352, no_hdr0=''),
  285367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285367, CurrentID=0, Status='live', Symbol='RPUSD3', Aliases='', description='RNA pseudouridylate synthase domain containing 3', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9837849, end_position_on_the_genomic_accession=9844018, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  285368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285368, CurrentID=0, Status='live', Symbol='PRRT3', Aliases='', description='proline-rich transmembrane protein 3', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9945542, end_position_on_the_genomic_accession=9952394, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23225, CurrentID=0, Status='live', Symbol='NUP210', Aliases='GP210, POM210', description='nucleoporin 210kDa', other_designations='nuclear envelope pore membrane protein POM 210|nuclear pore protein gp210|nucleoporin Nup210|pore membrane protein of 210 kDa', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13316230, end_position_on_the_genomic_accession=13420319, orientation='minus', exon_count=43, OMIM=607703, no_hdr0=''),
  285370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285370, CurrentID=0, Status='live', Symbol='LINC00606', Aliases='', description='long intergenic non-protein coding RNA 606', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10759484, end_position_on_the_genomic_accession=10764192, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23228, CurrentID=0, Status='live', Symbol='PLCL2', Aliases='PLCE2', description='phospholipase C like 2', other_designations='PLC-L(2)|PLC-L2|PLC-epsilon-2|phospholipase C, epsilon 2|phospholipase C-L2|phospholipase C-epsilon-2|phospholipase C-like 2', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16884959, end_position_on_the_genomic_accession=17090606, orientation='plus', exon_count=8, OMIM=614276, no_hdr0=''),
  23229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23229, CurrentID=0, Status='live', Symbol='ARHGEF9', Aliases='COLLYBISTIN, EIEE8, HPEM-2, PEM-2, PEM2', description='Cdc42 guanine nucleotide exchange factor 9', other_designations='Cdc42 guanine nucleotide exchange factor (GEF) 9|PEM-2 homolog|hPEM-2 collybistin|rac/Cdc42 guanine nucleotide exchange factor 9', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63634967, end_position_on_the_genomic_accession=63785546, orientation='minus', exon_count=13, OMIM=300429, no_hdr0=''),
  23230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23230, CurrentID=0, Status='live', Symbol='VPS13A', Aliases='CHAC, CHOREIN', description='vacuolar protein sorting 13 homolog A (S. cerevisiae)', other_designations='chorea-acanthocytosis protein|vacuolar protein sorting 13A', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77177353, end_position_on_the_genomic_accession=77417483, orientation='plus', exon_count=75, OMIM=605978, no_hdr0=''),
  23231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23231, CurrentID=0, Status='live', Symbol='SEL1L3', Aliases='Sel-1L3', description='sel-1 suppressor of lin-12-like 3 (C. elegans)', other_designations='suppressor of lin-12-like protein 3', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25721867, end_position_on_the_genomic_accession=25863595, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  23232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23232, CurrentID=0, Status='live', Symbol='TBC1D12', Aliases='', description='TBC1 domain family member 12', other_designations='TBC1 domain family, member 12', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94402429, end_position_on_the_genomic_accession=94536332, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23233, CurrentID=0, Status='live', Symbol='EXOC6B', Aliases='SEC15B, SEC15L2', description='exocyst complex component 6B', other_designations='SEC15 homolog B|SEC15-like 2|SEC15-like protein 2|WUGSC:H_NH0343N14.1|exocyst complex component Sec15B', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72175984, end_position_on_the_genomic_accession=72826048, orientation='minus', exon_count=26, OMIM=607880, no_hdr0=''),
  23234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23234, CurrentID=0, Status='live', Symbol='DNAJC9', Aliases='HDJC9, JDD1, SB73', description='DnaJ heat shock protein family (Hsp40) member C9', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 9|DnaJ protein SB73', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73241954, end_position_on_the_genomic_accession=73247331, orientation='minus', exon_count=5, OMIM=611206, no_hdr0=''),
  23235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23235, CurrentID=0, Status='live', Symbol='SIK2', Aliases='LOH11CR1I, QIK, SIK-2, SNF1LK2', description='salt inducible kinase 2', other_designations='SNF1-like kinase 2|qin-induced kinase|salt-inducible kinase 2|salt-inducible protein kinase 2|salt-inducible serine/threonine kinase 2|serine/threonine-protein kinase SNF1-like kinase 2', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111602391, end_position_on_the_genomic_accession=111732388, orientation='plus', exon_count=17, OMIM=608973, no_hdr0=''),
  23236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23236, CurrentID=0, Status='live', Symbol='PLCB1', Aliases='EIEE12, PI-PLC, PLC-154, PLC-I, PLC154A, PLCB1B, PLCB1', description='phospholipase C beta 1', other_designations='1-phosphatidyl-D-myo-inositol-4,5-bisphosphate|PLC-beta-1|inositoltrisphosphohydrolase|monophosphatidylinositol phosphodiesterase|phosphoinositidase C|phospholipase C, beta 1 (phosphoinositide-specific)|phospholipase C-I|triphosphoinositide phosphodiesterase', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8132254, end_position_on_the_genomic_accession=8884900, orientation='plus', exon_count=36, OMIM=607120, no_hdr0=''),
  23237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23237, CurrentID=0, Status='live', Symbol='ARC', Aliases='Arg3.1', description='activity-regulated cytoskeleton-associated protein', other_designations='ARC/ARG3.1|activity-regulated gene 3.1 protein homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142611044, end_position_on_the_genomic_accession=142614472, orientation='minus', exon_count=3, OMIM=612461, no_hdr0=''),
  285382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285382, CurrentID=0, Status='live', Symbol='C3orf70', Aliases='', description='chromosome 3 open reading frame 70', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185078050, end_position_on_the_genomic_accession=185153014, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23239, CurrentID=0, Status='live', Symbol='PHLPP1', Aliases='PHLPP, PLEKHE1, SCOP', description='PH domain and leucine rich repeat protein phosphatase 1', other_designations='PH domain-containing family E member 1|SCN circadian oscillatory protein|pleckstrin homology domain containing, family E (with leucine rich repeats) member 1|suprachiasmatic nucleus circadian oscillatory protein', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62715439, end_position_on_the_genomic_accession=62980443, orientation='plus', exon_count=18, OMIM=609396, no_hdr0=''),
  23240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23240, CurrentID=0, Status='live', Symbol='KIAA0922', Aliases='TMEM131L', description='KIAA0922', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153466346, end_position_on_the_genomic_accession=153637608, orientation='plus', exon_count=37, OMIM=616243, no_hdr0=''),
  154313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154313, CurrentID=0, Status='live', Symbol='CFAP206', Aliases='C6orf165, dJ382I10.1', description='cilia and flagella associated protein 206', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87407972, end_position_on_the_genomic_accession=87464473, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  285386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285386, CurrentID=0, Status='live', Symbol='TPRG1', Aliases='FAM79B', description='tumor protein p63 regulated 1', other_designations='family with sequence similarity 79, member B', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189099459, end_position_on_the_genomic_accession=189323482, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  23243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23243, CurrentID=0, Status='live', Symbol='ANKRD28', Aliases='PITK, PPP1R65', description='ankyrin repeat domain 28', other_designations='PP6-ARS-A|ankyrin repeat domain-containing protein 28|phosphatase interactor targeting K protein|phosphatase interactor targeting protein hnRNP K|protein phosphatase 1, regulatory subunit 65|protein phosphatase 6 ankyrin repeat subunit A|serine/threonine-protein phosphatase 6 regulatory subunit ARS-A', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15667236, end_position_on_the_genomic_accession=15859555, orientation='minus', exon_count=32, OMIM=611122, no_hdr0=''),
  23244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23244, CurrentID=0, Status='live', Symbol='PDS5A', Aliases='PIG54, SCC-112, SCC112', description='PDS5 cohesin associated factor A', other_designations='PDS5, regulator of cohesion maintenance, homolog A|cell proliferation-inducing gene 54 protein|sister chromatid cohesion protein 112', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39822863, end_position_on_the_genomic_accession=39977956, orientation='minus', exon_count=35, OMIM=613200, no_hdr0=''),
  23245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23245, CurrentID=0, Status='live', Symbol='ASTN2', Aliases='bA67K19.1', description='astrotactin 2', other_designations='bA264C15.1', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116425225, end_position_on_the_genomic_accession=117415039, orientation='minus', exon_count=24, OMIM=612856, no_hdr0=''),
  23246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23246, CurrentID=0, Status='live', Symbol='BOP1', Aliases='', description='block of proliferation 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144262046, end_position_on_the_genomic_accession=144291438, orientation='minus', exon_count=16, OMIM=610596, no_hdr0=''),
  23247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23247, CurrentID=0, Status='live', Symbol='KIAA0556', Aliases='', description='KIAA0556', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27550119, end_position_on_the_genomic_accession=27780371, orientation='plus', exon_count=31, OMIM=616650, no_hdr0=''),
  23248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23248, CurrentID=0, Status='live', Symbol='RPRD2', Aliases='HSPC099, KIAA0460', description='regulation of nuclear pre-mRNA domain containing 2', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150364111, end_position_on_the_genomic_accession=150476566, orientation='plus', exon_count=11, OMIM=614695, no_hdr0=''),
  23250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23250, CurrentID=0, Status='live', Symbol='ATP11A', Aliases='ATPIH, ATPIS', description='ATPase, class VI, type 11A', other_designations='P4-ATPase flippase complex alpha subunit ATP11A|phospholipid-translocating ATPase|potential phospholipid-transporting ATPase IH', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112690049, end_position_on_the_genomic_accession=112887168, orientation='plus', exon_count=34, OMIM=605868, no_hdr0=''),
  23251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23251, CurrentID=0, Status='live', Symbol='KIAA1024', Aliases='', description='KIAA1024', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79413494, end_position_on_the_genomic_accession=79472304, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23252, CurrentID=0, Status='live', Symbol='OTUD3', Aliases='DUBA4', description='OTU deubiquitinase 3', other_designations='OTU domain containing 3', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19881702, end_position_on_the_genomic_accession=19912945, orientation='plus', exon_count=10, OMIM=611758, no_hdr0=''),
  23253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23253, CurrentID=0, Status='live', Symbol='ANKRD12', Aliases='ANCO-2, ANCO1, GAC-1, Nbla00144', description='ankyrin repeat domain 12', other_designations='ankyrin repeat-containing cofactor 2', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9136753, end_position_on_the_genomic_accession=9285985, orientation='plus', exon_count=16, OMIM=610616, no_hdr0=''),
  23254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23254, CurrentID=0, Status='live', Symbol='KAZN', Aliases='KAZ', description='kazrin, periplakin interacting protein', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13893515, end_position_on_the_genomic_accession=15118048, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  23255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23255, CurrentID=0, Status='live', Symbol='MTCL1', Aliases='CCDC165, KIAA0802, SOGA2', description='microtubule crosslinking factor 1', other_designations='PAR-1-interacting protein|SOGA family member 2|coiled-coil domain containing 165|coiled-coil domain-containing protein 165|protein SOGA2', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8705578, end_position_on_the_genomic_accession=8832778, orientation='plus', exon_count=20, OMIM=615766, no_hdr0=''),
  23256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23256, CurrentID=0, Status='live', Symbol='SCFD1', Aliases='C14orf163, RA410, SLY1, SLY1P, STXBP1L2', description='sec1 family domain containing 1', other_designations='syntaxin-binding protein 1-like 2|vesicle transport-related protein', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30622254, end_position_on_the_genomic_accession=30735828, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  285401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285401, CurrentID=0, Status='live', Symbol='LINC00698', Aliases='', description='long intergenic non-protein coding RNA 698', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63102688, end_position_on_the_genomic_accession=63125062, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  23258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23258, CurrentID=0, Status='live', Symbol='DENND5A', Aliases='RAB6IP1', description='DENN/MADD domain containing 5A', other_designations='RAB6 interacting protein 1|rab6-interacting protein 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9138825, end_position_on_the_genomic_accession=9265335, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  23259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23259, CurrentID=0, Status='live', Symbol='DDHD2', Aliases='SAMWD1, SPG54, iPLA(1)gamma', description='DDHD domain containing 2', other_designations='PA-PLA1 like|SAM, WWE and DDHD domain-containing protein 1|intracellular phospholipase A1gamma|sec23p-interacting protein p125-like phosphatidic acid-preferring phospholipase A1', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38231343, end_position_on_the_genomic_accession=38273885, orientation='plus', exon_count=22, OMIM=615003, no_hdr0=''),
  23261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23261, CurrentID=0, Status='live', Symbol='CAMTA1', Aliases='CANPMR', description='calmodulin binding transcription activator 1', other_designations='', map_location='1p36.31-p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6785324, end_position_on_the_genomic_accession=7769706, orientation='plus', exon_count=29, OMIM=611501, no_hdr0=''),
  23262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23262, CurrentID=0, Status='live', Symbol='PPIP5K2', Aliases='HISPPD1, IP7K2, VIP2', description='diphosphoinositol pentakisphosphate kinase 2', other_designations='VIP1 homolog 2|histidine acid phosphatase domain-containing protein 1|inositol heptaphosphate kinase 2|insP6 and PP-IP5 kinase 2', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103120254, end_position_on_the_genomic_accession=103203523, orientation='plus', exon_count=34, OMIM=611648, no_hdr0=''),
  23263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23263, CurrentID=0, Status='live', Symbol='MCF2L', Aliases='ARHGEF14, DBS, OST', description='MCF.2 cell line derived transforming sequence like', other_designations="DBL's big sister|MCF.2 cell line derived transforming sequence-like|MCF2 transforming sequence-like protein", map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112894378, end_position_on_the_genomic_accession=113099739, orientation='plus', exon_count=41, OMIM=609499, no_hdr0=''),
  23264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23264, CurrentID=0, Status='live', Symbol='ZC3H7B', Aliases='RoXaN', description='zinc finger CCCH-type containing 7B', other_designations="Rotavirus 'X' associated non-structural protein|rotavirus 'X'-associated non-structural protein|rotavirus X protein associated with NSP3|ubiquitous tetratricopeptide containing protein RoXaN", map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41301503, end_position_on_the_genomic_accession=41360147, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  23265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23265, CurrentID=0, Status='live', Symbol='EXOC7', Aliases='2-5-3p, EX070, EXO70, EXOC1, Exo70p, YJL085W', description='exocyst complex component 7', other_designations='exocyst complex component Exo70', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76080993, end_position_on_the_genomic_accession=76103787, orientation='minus', exon_count=22, OMIM=608163, no_hdr0=''),
  23266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23266, CurrentID=0, Status='live', Symbol='ADGRL2', Aliases='CIRL2, CL2, LEC1, LPHH1, LPHN2', description='adhesion G protein-coupled receptor L2', other_designations='calcium-independent alpha-latrotoxin receptor 2|latrophilin 2|latrophilin homolog 1|latrophilin-2|lectomedin-1', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81306199, end_position_on_the_genomic_accession=81992738, orientation='plus', exon_count=37, OMIM=607018, no_hdr0=''),
  23268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23268, CurrentID=0, Status='live', Symbol='DNMBP', Aliases='ARHGEF36, TUBA', description='dynamin binding protein', other_designations='scaffold protein TUBA', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99875576, end_position_on_the_genomic_accession=100009948, orientation='minus', exon_count=19, OMIM=611282, no_hdr0=''),
  23269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23269, CurrentID=0, Status='live', Symbol='MGA', Aliases='MAD5, MXD5', description='MGA, MAX dimerization protein', other_designations='MAX dimerization protein 5', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41621232, end_position_on_the_genomic_accession=41769943, orientation='plus', exon_count=30, OMIM=616061, no_hdr0=''),
  23270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23270, CurrentID=0, Status='live', Symbol='TSPYL4', Aliases='dJ486I3.2', description='TSPY-like 4', other_designations='TSPY-like protein 4', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116249964, end_position_on_the_genomic_accession=116254098, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23271, CurrentID=0, Status='live', Symbol='CAMSAP2', Aliases='CAMSAP1L1', description='calmodulin regulated spectrin associated protein family member 2', other_designations='calmodulin regulated spectrin-associated protein family, member 2|calmodulin-regulated spectrin-associated protein 1-like protein 1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200739558, end_position_on_the_genomic_accession=200860707, orientation='plus', exon_count=18, OMIM=613775, no_hdr0=''),
  23272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23272, CurrentID=0, Status='live', Symbol='FAM208A', Aliases='C3orf63, RAP140, se89-1', description='family with sequence similarity 208 member A', other_designations='CTCL tumor antigen se89-1|family with sequence similarity 208, member A|protein FAM208A|retinoblastoma-associated protein 140|retinoblastoma-associated protein RAP140|transgene activation suppressor protein', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56620132, end_position_on_the_genomic_accession=56683242, orientation='minus', exon_count=24, OMIM=616493, no_hdr0=''),
  23274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23274, CurrentID=0, Status='live', Symbol='CLEC16A', Aliases='Gop-1, KIAA0350', description='C-type lectin domain family 16 member A', other_designations='C-type lectin domain family 16, member A', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10944488, end_position_on_the_genomic_accession=11193278, orientation='plus', exon_count=33, OMIM=611303, no_hdr0=''),
  23275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23275, CurrentID=0, Status='live', Symbol='POFUT2', Aliases='C21orf80, FUT13', description='protein O-fucosyltransferase 2', other_designations='O-FucT-2|peptide-O-fucosyltransferase 2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45263928, end_position_on_the_genomic_accession=45287899, orientation='minus', exon_count=15, OMIM=610249, no_hdr0=''),
  23276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23276, CurrentID=0, Status='live', Symbol='KLHL18', Aliases='', description='kelch like family member 18', other_designations='kelch-like family member 18', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47282840, end_position_on_the_genomic_accession=47346816, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23277, CurrentID=0, Status='live', Symbol='CLUH', Aliases='CLU1, KIAA0664', description='clustered mitochondria (cluA/CLU1) homolog', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2689386, end_position_on_the_genomic_accession=2711633, orientation='minus', exon_count=26, OMIM=616184, no_hdr0=''),
  285422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285422, CurrentID=0, Status='live', Symbol='LOC285422', Aliases='', description='uncharacterized LOC285422', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145581465, end_position_on_the_genomic_accession=145588021, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23279, CurrentID=0, Status='live', Symbol='NUP160', Aliases='', description='nucleoporin 160kDa', other_designations='160 kDa nucleoporin|nucleoporin 160kD|nucleoporin Nup160', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47778113, end_position_on_the_genomic_accession=47848505, orientation='minus', exon_count=36, OMIM=607614, no_hdr0=''),
  23281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23281, CurrentID=0, Status='live', Symbol='MTUS2', Aliases='CAZIP, ICIS, KIAA0774, TIP150', description='microtubule associated tumor suppressor candidate 2', other_designations='+TIP of 150 kDa|cardiac zipper protein|microtubule plus-end tracking protein TIP150|tracking protein of 150 kDa', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28820348, end_position_on_the_genomic_accession=29505947, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  23283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23283, CurrentID=0, Status='live', Symbol='CSTF2T', Aliases='CstF-64T', description="cleavage stimulation factor, 3' pre-RNA, subunit 2, tau variant", other_designations="CF-1 64 kDa subunit tau|CSTF 64 kDa subunit tau|cleavage stimulation factor 64 kDa subunit tau|cleavage stimulation factor subunit 2 tau|cleavage stimulation factor, 3' pre-RNA, subunit 2, 64kDa, tau|tauCstF-64", map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51695486, end_position_on_the_genomic_accession=51699595, orientation='minus', exon_count=1, OMIM=611968, no_hdr0=''),
  23284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23284, CurrentID=0, Status='live', Symbol='ADGRL3', Aliases='CIRL3, CL3, LEC3, LPHN3', description='adhesion G protein-coupled receptor L3', other_designations='CIRL-3|calcium-independent alpha-latrotoxin receptor 3|latrophilin 3|latrophilin homolog 3 (cow)|latrophilin-3|lectomedin 3', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=61383124, end_position_on_the_genomic_accession=62078335, orientation='plus', exon_count=27, OMIM=616417, no_hdr0=''),
  23285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23285, CurrentID=0, Status='live', Symbol='KIAA1107', Aliases='', description='KIAA1107', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92167052, end_position_on_the_genomic_accession=92184723, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  23286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23286, CurrentID=0, Status='live', Symbol='WWC1', Aliases='HBEBP3, HBEBP36, KIBRA, MEMRYQTL, PPP1R168', description='WW and C2 domain containing 1', other_designations='HBeAg-binding protein 3|WW, C2 and coiled-coil domain containing 1|kidney and brain protein|protein WWC1|protein phosphatase 1, regulatory subunit 168', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168292060, end_position_on_the_genomic_accession=168472303, orientation='plus', exon_count=25, OMIM=610533, no_hdr0=''),
  23287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23287, CurrentID=0, Status='live', Symbol='AGTPBP1', Aliases='CCP1, NNA1', description='ATP/GTP binding protein 1', other_designations='ATP/GTP-binding protein 1|carboxypeptidase-tubulin|nervous system nuclear protein induced by axotomy protein 1 homolog|soluble carboxypeptidase|tubulinyl-Tyr carboxypeptidase|tyrosine carboxypeptidase', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85546539, end_position_on_the_genomic_accession=85742029, orientation='minus', exon_count=30, OMIM=606830, no_hdr0=''),
  23288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23288, CurrentID=0, Status='live', Symbol='IQCE', Aliases='1700028P05Rik', description='IQ motif containing E', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2558972, end_position_on_the_genomic_accession=2614734, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  23291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23291, CurrentID=0, Status='live', Symbol='FBXW11', Aliases='BTRC2, BTRCP2, FBW1B, FBXW1B, Fbw11, Hos', description='F-box and WD repeat domain containing 11', other_designations='F-box and WD repeats protein beta-TrCP2|F-box and WD-40 domain protein 11|F-box and WD-40 domain protein 1B|F-box protein Fbw1b|F-box/WD repeat-containing protein 1B|beta-transducin repeat-containing protein 2|homologous to Slimb protein', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171861549, end_position_on_the_genomic_accession=172006873, orientation='minus', exon_count=17, OMIM=605651, no_hdr0=''),
  23293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23293, CurrentID=0, Status='live', Symbol='SMG6', Aliases='C17orf31, EST1A, SMG-6, hSMG5/7a', description='SMG6 nonsense mediated mRNA decay factor', other_designations='EST1 telomerase component homolog A|ever shorter telomeres 1A|smg-6 homolog, nonsense mediated mRNA decay factor|telomerase subunit EST1A', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2059839, end_position_on_the_genomic_accession=2303813, orientation='minus', exon_count=31, OMIM=610963, no_hdr0=''),
  23294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23294, CurrentID=0, Status='live', Symbol='ANKS1A', Aliases='ANKS1', description='ankyrin repeat and sterile alpha motif domain containing 1A', other_designations='Odin|ankyrin repeat and SAM domain containing 1', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34889261, end_position_on_the_genomic_accession=35098010, orientation='plus', exon_count=28, OMIM=608994, no_hdr0=''),
  23295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23295, CurrentID=0, Status='live', Symbol='MGRN1', Aliases='RNF156', description='mahogunin ring finger 1, E3 ubiquitin protein ligase', other_designations='RING finger protein 156|mahogunin RING finger protein 1|mahogunin, ring finger 1|probable E3 ubiquitin-protein ligase MGRN1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4616493, end_position_on_the_genomic_accession=4690974, orientation='plus', exon_count=20, OMIM=607559, no_hdr0=''),
  285440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285440, CurrentID=0, Status='live', Symbol='CYP4V2', Aliases='BCD, CYP4AH1', description='cytochrome P450 family 4 subfamily V member 2', other_designations='cytochrome P450, family 4, subfamily V, polypeptide 2|docosahexaenoic acid omega-hydroxylase CYP4V2', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186191411, end_position_on_the_genomic_accession=186213463, orientation='plus', exon_count=11, OMIM=608614, no_hdr0=''),
  285441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285441, CurrentID=0, Status='live', Symbol='F11-AS1', Aliases='', description='F11 antisense RNA 1', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186286098, end_position_on_the_genomic_accession=186501058, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  285442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285442, CurrentID=0, Status='live', Symbol='LOC285442', Aliases='', description='mitochondrial translational release factor 1 like pseudogene', other_designations='mitochondrial translational release factor 1-like pseudogene', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188738371, end_position_on_the_genomic_accession=188741073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23299, CurrentID=0, Status='live', Symbol='BICD2', Aliases='SMALED2, bA526D8.1', description='bicaudal D homolog 2 (Drosophila)', other_designations='bic-D 2|coiled-coil protein BICD2|cytoskeleton-like bicaudal D protein homolog 2|homolog of Drosophila bicaudal D', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92711363, end_position_on_the_genomic_accession=92764801, orientation='minus', exon_count=8, OMIM=609797, no_hdr0=''),
  23300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23300, CurrentID=0, Status='live', Symbol='ATMIN', Aliases='ASCIZ, ZNF822', description='ATM interactor', other_designations='ATM INteracting protein|ATM/ATR-Substrate Chk2-Interacting Zn++-finger protein|ATM/ATR-substrate CHEK2-interacting zinc finger protein|ATM/ATR-substrate CHK2-interacting zinc finger protein|zinc finger protein 822', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81035853, end_position_on_the_genomic_accession=81047346, orientation='plus', exon_count=5, OMIM=614693, no_hdr0=''),
  23301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23301, CurrentID=0, Status='live', Symbol='EHBP1', Aliases='HPC12, NACSIN', description='EH domain binding protein 1', other_designations='NPF calponin-like protein', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62673851, end_position_on_the_genomic_accession=63046487, orientation='plus', exon_count=29, OMIM=609922, no_hdr0=''),
  23302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23302, CurrentID=0, Status='live', Symbol='WSCD1', Aliases='', description='WSC domain containing 1', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6069106, end_position_on_the_genomic_accession=6124427, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  23303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23303, CurrentID=0, Status='live', Symbol='KIF13B', Aliases='GAKIN', description='kinesin family member 13B', other_designations='guanylate kinase associated kinesin|kinesin 13B|kinesin-like protein GAKIN', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29067278, end_position_on_the_genomic_accession=29263304, orientation='minus', exon_count=42, OMIM=607350, no_hdr0=''),
  23304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23304, CurrentID=0, Status='live', Symbol='UBR2', Aliases='C6orf133, bA49A4.1, dJ242G1.1, dJ392M17.3', description='ubiquitin protein ligase E3 component n-recognin 2', other_designations='ubiquitin-protein ligase E3-alpha-2|ubiquitin-protein ligase E3-alpha-II', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42563952, end_position_on_the_genomic_accession=42693505, orientation='plus', exon_count=52, OMIM=609134, no_hdr0=''),
  23305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23305, CurrentID=0, Status='live', Symbol='ACSL6', Aliases='ACS2, FACL6, LACS 6, LACS2, LACS5', description='acyl-CoA synthetase long-chain family member 6', other_designations='fatty-acid-Coenzyme A ligase, long-chain 6|long fatty acyl-CoA synthetase 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131949973, end_position_on_the_genomic_accession=132012068, orientation='minus', exon_count=23, OMIM=604443, no_hdr0=''),
  23306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23306, CurrentID=0, Status='live', Symbol='NEMP1', Aliases='TMEM194, TMEM194A', description='nuclear envelope integral membrane protein 1', other_designations='transmembrane protein 194|transmembrane protein 194A', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57055643, end_position_on_the_genomic_accession=57088857, orientation='minus', exon_count=12, OMIM=616496, no_hdr0=''),
  23307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23307, CurrentID=0, Status='live', Symbol='FKBP15', Aliases='FKBP133, KIAA0674, PPP1R76', description='FK506 binding protein 15', other_designations='133 kDa FK506-binding protein|133 kDa FKBP|FK506 binding protein 15, 133kDa|FK506-binding protein 133kDa|FKBP-133|FKBP-15|WAFL|WASP and FKBP-like protein|protein phosphatase 1, regulatory subunit 76', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113165519, end_position_on_the_genomic_accession=113221361, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  23308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23308, CurrentID=0, Status='live', Symbol='ICOSLG', Aliases='B7-H2, B7H2, B7RP-1, B7RP1, CD275, GL50, ICOS-L, ICOSL, LICOS', description='inducible T-cell co-stimulator ligand', other_designations='B7 homolog 2|B7 homologue 2|B7-like protein Gl50|B7-related protein 1|transmembrane protein B7-H2 ICOS ligand', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44216971, end_position_on_the_genomic_accession=44241447, orientation='minus', exon_count=10, OMIM=605717, no_hdr0=''),
  23309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23309, CurrentID=0, Status='live', Symbol='SIN3B', Aliases='', description='SIN3 transcription regulator family member B', other_designations='SIN3 homolog B, transcriptional regulator|SIN3 transcription regulator homolog B|histone deacetylase complex subunit Sin3b|transcriptional corepressor Sin3b', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16829387, end_position_on_the_genomic_accession=16880355, orientation='plus', exon_count=21, OMIM=607777, no_hdr0=''),
  23310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23310, CurrentID=0, Status='live', Symbol='NCAPD3', Aliases='CAP-D3, CAPD3, hCAP-D3, hHCP-6, hcp-6', description='non-SMC condensin II complex subunit D3', other_designations='non-SMC condensin II complex, subunit D3', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134152442, end_position_on_the_genomic_accession=134225345, orientation='minus', exon_count=38, OMIM=609276, no_hdr0=''),
  3885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3885, CurrentID=0, Status='live', Symbol='KRT34', Aliases='HA4, Ha-4, KRTHA4, hHa4', description='keratin 34, type I', other_designations='K34|hair keratin, type I Ha4|hard keratin, type I, 4|keratin, hair, acidic, 4|keratin-34', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41377650, end_position_on_the_genomic_accession=41383816, orientation='minus', exon_count=8, OMIM=602763, no_hdr0=''),
  23312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23312, CurrentID=0, Status='live', Symbol='DMXL2', Aliases='PEPNS, RC3', description='Dmx-like 2', other_designations='rabconnectin-3', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51447724, end_position_on_the_genomic_accession=51622833, orientation='minus', exon_count=46, OMIM=612186, no_hdr0=''),
  23313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23313, CurrentID=0, Status='live', Symbol='KIAA0930', Aliases='C22orf9, LSC3', description='KIAA0930', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45192242, end_position_on_the_genomic_accession=45240769, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  154386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154386, CurrentID=0, Status='live', Symbol='LINC01600', Aliases='C6orf195, bA145H9.2', description='long intergenic non-protein coding RNA 1600', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2622738, end_position_on_the_genomic_accession=2635064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23315, CurrentID=0, Status='live', Symbol='SLC9A8', Aliases='NHE-8, NHE8', description='solute carrier family 9, subfamily A (NHE8, cation proton antiporter 8), member 8', other_designations='Na(+)/H(+) exchanger 8|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 (sodium/hydrogen exchanger), member 8', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49812713, end_position_on_the_genomic_accession=49892242, orientation='plus', exon_count=20, OMIM=612730, no_hdr0=''),
  23316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23316, CurrentID=0, Status='live', Symbol='CUX2', Aliases='CDP2, CUTL2', description='cut-like homeobox 2', other_designations='homeobox protein cux-2', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111034024, end_position_on_the_genomic_accession=111350554, orientation='plus', exon_count=32, OMIM=610648, no_hdr0=''),
  23317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23317, CurrentID=0, Status='live', Symbol='DNAJC13', Aliases='PARK21, RME8', description='DnaJ heat shock protein family (Hsp40) member C13', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 13|DnaJ domain-containing protein RME-8|RME-8|required for receptor-mediated endocytosis 8', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132417532, end_position_on_the_genomic_accession=132539032, orientation='plus', exon_count=57, OMIM=614334, no_hdr0=''),
  23318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23318, CurrentID=0, Status='live', Symbol='ZCCHC11', Aliases='PAPD3, TUT4', description='zinc finger, CCHC domain containing 11', other_designations='PAP associated domain containing 3|TUTase 4|zinc finger CCHC domain-containing protein 11', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52423275, end_position_on_the_genomic_accession=52553463, orientation='minus', exon_count=36, OMIM=613692, no_hdr0=''),
  285463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285463, CurrentID=0, Status='live', Symbol='CTBP1-AS', Aliases='PCAT10', description='CTBP1 antisense RNA', other_designations='prostate cancer associated transcript 10', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1210120, end_position_on_the_genomic_accession=1218591, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  285464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285464, CurrentID=0, Status='live', Symbol='CRIPAK', Aliases='', description='cysteine-rich PAK1 inhibitor', other_designations='cysteine-rich inhibitor of PAK1', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1391552, end_position_on_the_genomic_accession=1395994, orientation='plus', exon_count=1, OMIM=610203, no_hdr0=''),
  23321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23321, CurrentID=0, Status='live', Symbol='TRIM2', Aliases='CMT2R, RNF86', description='tripartite motif containing 2', other_designations='E3 ubiquitin-protein ligase TRIM2|RING finger protein 86|tripartite motif protein TRIM2', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153152469, end_position_on_the_genomic_accession=153339322, orientation='plus', exon_count=19, OMIM=614141, no_hdr0=''),
  23322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23322, CurrentID=0, Status='live', Symbol='RPGRIP1L', Aliases='CORS3, FTM, JBTS7, MKS5, NPHP8, PPP1R134', description='RPGRIP1-like', other_designations='RPGR-interacting protein 1-like protein|fantom homolog|nephrocystin-8|protein phosphatase 1, regulatory subunit 134', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53599906, end_position_on_the_genomic_accession=53703859, orientation='minus', exon_count=35, OMIM=610937, no_hdr0=''),
  101060399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060399, CurrentID=0, Status='live', Symbol='LOC101060399', Aliases='', description='cerebellar degeneration-related protein 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23324, CurrentID=0, Status='live', Symbol='MAN2B2', Aliases='', description='mannosidase, alpha, class 2B, member 2', other_designations='core-specific lysosomal alpha-1,6-Mannosidase', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6575174, end_position_on_the_genomic_accession=6622403, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  23325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23325, CurrentID=0, Status='live', Symbol='KIAA1033', Aliases='MRT43, SWIP', description='KIAA1033', other_designations='strumpellin and WASH-interacting protein', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105107714, end_position_on_the_genomic_accession=105169135, orientation='plus', exon_count=37, OMIM=615748, no_hdr0=''),
  23326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23326, CurrentID=0, Status='live', Symbol='USP22', Aliases='USP3L', description='ubiquitin specific peptidase 22', other_designations='deubiquitinating enzyme 22|ubiquitin specific protease 22|ubiquitin thioesterase 22|ubiquitin thiolesterase 22|ubiquitin-specific processing protease 22|ubiquitin-specific-processing protease 22', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20999593, end_position_on_the_genomic_accession=21043409, orientation='minus', exon_count=14, OMIM=612116, no_hdr0=''),
  23327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23327, CurrentID=0, Status='live', Symbol='NEDD4L', Aliases='NEDD4-2, NEDD4.2, RSP5, hNEDD4-2', description='neural precursor cell expressed, developmentally down-regulated 4-like, E3 ubiquitin protein ligase', other_designations='ubiquitin-protein ligase Rsp5', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58044233, end_position_on_the_genomic_accession=58401541, orientation='plus', exon_count=38, OMIM=606384, no_hdr0=''),
  23328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23328, CurrentID=0, Status='live', Symbol='SASH1', Aliases='SH3D6A, dJ323M4, dJ323M4.1', description='SAM and SH3 domain containing 1', other_designations='2500002E12Rik|proline-glutamate repeat-containing protein', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148193440, end_position_on_the_genomic_accession=148552049, orientation='plus', exon_count=25, OMIM=607955, no_hdr0=''),
  23329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23329, CurrentID=0, Status='live', Symbol='TBC1D30', Aliases='', description='TBC1 domain family member 30', other_designations='TBC1 domain family, member 30', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64824572, end_position_on_the_genomic_accession=64881022, orientation='plus', exon_count=13, OMIM=615077, no_hdr0=''),
  23331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23331, CurrentID=0, Status='live', Symbol='TTC28', Aliases='TPRBK', description='tetratricopeptide repeat domain 28', other_designations='TPR repeat protein 28|TPR repeat-containing big gene cloned at Keio|TPR-containing big gene cloned at Keio', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27978014, end_position_on_the_genomic_accession=28679865, orientation='minus', exon_count=28, OMIM=615098, no_hdr0=''),
  23332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23332, CurrentID=0, Status='live', Symbol='CLASP1', Aliases='MAST1', description='cytoplasmic linker associated protein 1', other_designations='cytoplasmic linker-associated protein 1|multiple asters 1|multiple asters homolog 1|protein Orbit homolog 1', map_location='2q14.2-q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121337776, end_position_on_the_genomic_accession=121649476, orientation='minus', exon_count=47, OMIM=605852, no_hdr0=''),
  23333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23333, CurrentID=0, Status='live', Symbol='DPY19L1', Aliases='', description='dpy-19-like 1 (C. elegans)', other_designations='DPY-19-like protein 1|protein dpy-19 homolog 1', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34928879, end_position_on_the_genomic_accession=35038041, orientation='minus', exon_count=23, OMIM=613892, no_hdr0=''),
  23334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23334, CurrentID=0, Status='live', Symbol='SZT2', Aliases='C1orf84, EIEE18, KIAA0467A, SZT2B, SZT2', description='seizure threshold 2 homolog (mouse)', other_designations='RP11-506B15.1|seizure threshold 2 protein homolog', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43384830, end_position_on_the_genomic_accession=43454247, orientation='plus', exon_count=76, OMIM=615463, no_hdr0=''),
  23335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23335, CurrentID=0, Status='live', Symbol='WDR7', Aliases='TRAG', description='WD repeat domain 7', other_designations='TGF-beta resistance associated|TGF-beta resistance-associated protein TRAG|rabconnectin-3 beta', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56651343, end_position_on_the_genomic_accession=57029805, orientation='plus', exon_count=31, OMIM=613473, no_hdr0=''),
  23336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23336, CurrentID=0, Status='live', Symbol='SYNM', Aliases='DMN, SYN', description='synemin', other_designations='desmuslin|synemin alpha|synemin beta|synemin, intermediate filament protein', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99105080, end_position_on_the_genomic_accession=99135595, orientation='plus', exon_count=4, OMIM=606087, no_hdr0=''),
  23338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23338, CurrentID=0, Status='live', Symbol='JADE2', Aliases='JADE-2, PHF15', description='jade family PHD finger 2', other_designations='PHD finger protein 15|jade family PHD finger protein 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134524162, end_position_on_the_genomic_accession=134583230, orientation='plus', exon_count=16, OMIM=610515, no_hdr0=''),
  23339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23339, CurrentID=0, Status='live', Symbol='VPS39', Aliases='TLP, VAM6, hVam6p', description='VPS39 HOPS complex subunit', other_designations='TRAP1-like protein|vacuolar protein sorting 39 homolog', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42158701, end_position_on_the_genomic_accession=42208326, orientation='minus', exon_count=27, OMIM=612188, no_hdr0=''),
  285484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285484, CurrentID=0, Status='live', Symbol='LOC285484', Aliases='', description='uncharacterized LOC285484', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6200733, end_position_on_the_genomic_accession=6233936, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  23341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23341, CurrentID=0, Status='live', Symbol='DNAJC16', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C16', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 16|novel DnaJ domain-containing protein', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15526813, end_position_on_the_genomic_accession=15571733, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23344, CurrentID=0, Status='live', Symbol='ESYT1', Aliases='FAM62A, MBC2', description='extended synaptotagmin-like protein 1', other_designations='family with sequence similarity 62 (C2 domain containing), member A|membrane-bound C2 domain-containing protein', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56128202, end_position_on_the_genomic_accession=56144676, orientation='plus', exon_count=31, OMIM=616670, no_hdr0=''),
  23345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23345, CurrentID=0, Status='live', Symbol='SYNE1', Aliases='8B, ARCA1, C6orf98, CPG2, EDMD4, MYNE1, Nesp1, SCAR8, dJ45H2.2', description='spectrin repeat containing, nuclear envelope 1', other_designations='enaptin|myocyte nuclear envelope protein 1|nesprin 1|synaptic nuclear envelope protein 1|synaptic nuclei expressed gene 1', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152121684, end_position_on_the_genomic_accession=152637399, orientation='minus', exon_count=152, OMIM=608441, no_hdr0=''),
  23347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23347, CurrentID=0, Status='live', Symbol='SMCHD1', Aliases='', description='structural maintenance of chromosomes flexible hinge domain containing 1', other_designations='SMC hinge domain-containing protein 1', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2655887, end_position_on_the_genomic_accession=2805017, orientation='plus', exon_count=52, OMIM=614982, no_hdr0=''),
  23348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23348, CurrentID=0, Status='live', Symbol='DOCK9', Aliases='ZIZ1, ZIZIMIN1', description='dedicator of cytokinesis 9', other_designations='cdc42 guanine nucleotide exchange factor zizimin-1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98793487, end_position_on_the_genomic_accession=99088618, orientation='minus', exon_count=62, OMIM=607325, no_hdr0=''),
  23349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23349, CurrentID=0, Status='live', Symbol='PHF24', Aliases='KIAA1045', description='PHD finger protein 24', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34810040, end_position_on_the_genomic_accession=34982544, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23350, CurrentID=0, Status='live', Symbol='U2SURP', Aliases='SR140, fSAPa', description='U2 snRNP-associated SURP domain containing', other_designations='140 kDa Ser/Arg-rich domain protein|Ser/Arg-rich domain protein, 140 kDa|U2 associated SR140 protein|U2-associated SR140 protein|U2-associated protein SR140|functional spliceosome-associated protein a', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143001524, end_position_on_the_genomic_accession=143060725, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  23351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23351, CurrentID=0, Status='live', Symbol='KHNYN', Aliases='KIAA0323', description='KH and NYN domain containing', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24423171, end_position_on_the_genomic_accession=24441342, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23352, CurrentID=0, Status='live', Symbol='UBR4', Aliases='RBAF600, ZUBR1, p600', description='ubiquitin protein ligase E3 component n-recognin 4', other_designations='N-recognin-4|retinoblastoma-associated factor 600-like protein|retinoblastoma-associated factor of 600 kDa|zinc finger UBR1-type protein 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19074506, end_position_on_the_genomic_accession=19210319, orientation='minus', exon_count=108, OMIM=609890, no_hdr0=''),
  23353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23353, CurrentID=0, Status='live', Symbol='SUN1', Aliases='UNC84A', description='Sad1 and UNC84 domain containing 1', other_designations='Sad1 unc-84 domain protein 1|protein unc-84 homolog A|sad1/unc-84 protein-like 1|unc-84 homolog A', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=815557, end_position_on_the_genomic_accession=874920, orientation='plus', exon_count=22, OMIM=607723, no_hdr0=''),
  23354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23354, CurrentID=0, Status='live', Symbol='HAUS5', Aliases='KIAA0841, dgt5', description='HAUS augmin like complex subunit 5', other_designations='HAUS augmin-like complex, subunit 5', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35612744, end_position_on_the_genomic_accession=35625349, orientation='plus', exon_count=19, OMIM=613432, no_hdr0=''),
  23355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23355, CurrentID=0, Status='live', Symbol='VPS8', Aliases='KIAA0804', description='VPS8, CORVET complex subunit', other_designations='VPS8 CORVET complex subunit|vacuolar protein sorting 8 homolog', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184812137, end_position_on_the_genomic_accession=185052614, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  285500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285500, CurrentID=0, Status='live', Symbol='LOC285500', Aliases='', description='uncharacterized LOC285500', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177444513, end_position_on_the_genomic_accession=177677123, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  23357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23357, CurrentID=0, Status='live', Symbol='ANGEL1', Aliases='Ccr4e, KIAA0759', description='angel homolog 1 (Drosophila)', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76787243, end_position_on_the_genomic_accession=76828473, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  23358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23358, CurrentID=0, Status='live', Symbol='USP24', Aliases='', description='ubiquitin specific peptidase 24', other_designations='deubiquitinating enzyme 24|ubiquitin specific protease 24|ubiquitin thioesterase 24|ubiquitin thiolesterase 24|ubiquitin-specific processing protease 24|ubiquitin-specific-processing protease 24', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55066359, end_position_on_the_genomic_accession=55215368, orientation='minus', exon_count=72, OMIM=610569, no_hdr0=''),
  23359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23359, CurrentID=0, Status='live', Symbol='FAM189A1', Aliases='TMEM228', description='family with sequence similarity 189 member A1', other_designations='family with sequence similarity 189, member A1|transmembrane protein 228', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29120252, end_position_on_the_genomic_accession=29570723, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  23360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23360, CurrentID=0, Status='live', Symbol='FNBP4', Aliases='FBP30', description='formin binding protein 4', other_designations='formin-binding protein 30', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47716517, end_position_on_the_genomic_accession=47767460, orientation='minus', exon_count=19, OMIM=615265, no_hdr0=''),
  23361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23361, CurrentID=0, Status='live', Symbol='ZNF629', Aliases='ZNF, ZNF65', description='zinc finger protein 629', other_designations='DNA-binding protein|zinc finger protein 65', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30778449, end_position_on_the_genomic_accession=30787277, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23362, CurrentID=0, Status='live', Symbol='PSD3', Aliases='EFA6D, EFA6R, HCA67', description='pleckstrin and Sec7 domain containing 3', other_designations='ADP-ribosylation factor guanine nucleotide factor 6|epididymis tissue protein Li 20mP|exchange factor for ADP-ribosylation factor guanine nucleotide factor 6|hepatocellular carcinoma-associated antigen 67|pleckstrin homology and SEC7 domain-containing protein 3', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18527303, end_position_on_the_genomic_accession=19086921, orientation='minus', exon_count=33, OMIM=614440, no_hdr0=''),
  23363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23363, CurrentID=0, Status='live', Symbol='OBSL1', Aliases='', description='obscurin-like 1', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219549403, end_position_on_the_genomic_accession=219571577, orientation='minus', exon_count=24, OMIM=610991, no_hdr0=''),
  23365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23365, CurrentID=0, Status='live', Symbol='ARHGEF12', Aliases='LARG, PRO2792', description='Rho guanine nucleotide exchange factor 12', other_designations='Rho guanine nucleotide exchange factor (GEF) 12|leukemia-associated RhoGEF|leukemia-associated rho guanine nucleotide exchange factor', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120336406, end_position_on_the_genomic_accession=120489936, orientation='plus', exon_count=43, OMIM=604763, no_hdr0=''),
  23366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23366, CurrentID=0, Status='live', Symbol='KIAA0895', Aliases='', description='KIAA0895', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36324150, end_position_on_the_genomic_accession=36390125, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  23367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23367, CurrentID=0, Status='live', Symbol='LARP1', Aliases='LARP', description='La ribonucleoprotein domain family member 1', other_designations='La ribonucleoprotein domain family, member 1', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154682663, end_position_on_the_genomic_accession=154817607, orientation='plus', exon_count=27, OMIM=612059, no_hdr0=''),
  285512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285512, CurrentID=0, Status='live', Symbol='FAM13A-AS1', Aliases='FAM13A1OS, FAM13AOS, NCRNA00039', description='FAM13A antisense RNA 1', other_designations='FAM13A antisense RNA 1 (non-protein coding)|FAM13A opposite strand (non-protein coding)|FAM13A1 opposite strand (non-protein coding)|family with sequence similarity 13, member A1 opposite strand', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88709789, end_position_on_the_genomic_accession=88730103, orientation='plus', exon_count=5, OMIM=613300, no_hdr0=''),
  285513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285513, CurrentID=0, Status='live', Symbol='GPRIN3', Aliases='GRIN3', description='GPRIN family member 3', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89241276, end_position_on_the_genomic_accession=89308010, orientation='minus', exon_count=3, OMIM=611241, no_hdr0=''),
  23370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23370, CurrentID=0, Status='live', Symbol='ARHGEF18', Aliases='P114-RhoGEF', description='Rho/Rac guanine nucleotide exchange factor 18', other_designations='114 kDa Rho-specific guanine nucleotide exchange factor|P114-RHO-GEF|Rho-specific guanine nucleotide exchange factor p114|Rho/Rac guanine nucleotide exchange factor (GEF) 18|SA-RhoGEF|p114RhoGEF|septin-associated RhoGEF', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7348911, end_position_on_the_genomic_accession=7476990, orientation='plus', exon_count=35, OMIM=616432, no_hdr0=''),
  23371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23371, CurrentID=0, Status='live', Symbol='TNS2', Aliases='C1-TEN, C1TEN, TENC1', description='tensin 2', other_designations='C1 domain-containing phosphatase and tensin homolog|tensin like C1 domain containing phosphatase (tensin 2)|tensin like C1 domain-containing phosphatase|tensin-like C1 domain-containing phosphatase', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53046980, end_position_on_the_genomic_accession=53064379, orientation='plus', exon_count=31, OMIM=607717, no_hdr0=''),
  105377756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377756, CurrentID=0, Status='live', Symbol='LOC105377756', Aliases='', description='uncharacterized LOC105377756', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178477750, end_position_on_the_genomic_accession=178480563, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23373, CurrentID=0, Status='live', Symbol='CRTC1', Aliases='MECT1, TORC-1, TORC1, WAMTP1', description='CREB regulated transcription coactivator 1', other_designations='mucoepidermoid carcinoma translocated protein 1|transducer of regulated cAMP response element-binding protein (CREB) 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18683615, end_position_on_the_genomic_accession=18782333, orientation='plus', exon_count=15, OMIM=607536, no_hdr0=''),
  23376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23376, CurrentID=0, Status='live', Symbol='UFL1', Aliases='KIAA0776, Maxer, NLBP, RCAD', description='UFM1-specific ligase 1', other_designations='LZAP-binding protein|Regulator of CDK5RAP3 and DDRGK1|novel LZAP-binding protein|regulator of C53/LZAP and DDRGK1', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96521826, end_position_on_the_genomic_accession=96555276, orientation='plus', exon_count=19, OMIM=613372, no_hdr0=''),
  285521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285521, CurrentID=0, Status='live', Symbol='COX18', Aliases='COX18HS', description='COX18 cytochrome c oxidase assembly factor', other_designations='cytochrome c oxidase assembly homolog 18|cytochrome c oxidase assembly protein 18', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73054696, end_position_on_the_genomic_accession=73069759, orientation='minus', exon_count=6, OMIM=610428, no_hdr0=''),
  23378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23378, CurrentID=0, Status='live', Symbol='RRP8', Aliases='KIAA0409, NML', description='ribosomal RNA processing 8, methyltransferase, homolog (yeast)', other_designations='RRP8 methyltransferase homolog|cerebral protein 1|nucleomethylin', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6514830, end_position_on_the_genomic_accession=6603678, orientation='minus', exon_count=10, OMIM=615818, no_hdr0=''),
  23379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23379, CurrentID=0, Status='live', Symbol='ICE1', Aliases='KIAA0947', description='interactor of little elongation complex ELL subunit 1', other_designations='interactor of little elongator complex ELL subunit 1', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5422673, end_position_on_the_genomic_accession=5490234, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  23380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23380, CurrentID=0, Status='live', Symbol='SRGAP2', Aliases='ARHGAP34, FNBP2A, SRGAP3, SRGAP2', description='SLIT-ROBO Rho GTPase activating protein 2', other_designations='SLIT-ROBO GAP2|formin-binding protein 2|rho GTPase-activating protein 34', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206203464, end_position_on_the_genomic_accession=206464443, orientation='plus', exon_count=28, OMIM=606524, no_hdr0=''),
  285525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285525, CurrentID=0, Status='live', Symbol='YIPF7', Aliases='FinGER9', description='Yip1 domain family member 7', other_designations='YIP1 family member 7|Yip1 domain family, member 7|five-pass transmembrane protein localizing in the Golgi apparatus and the endoplasmic reticulum 9', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44621970, end_position_on_the_genomic_accession=44651641, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23382, CurrentID=0, Status='live', Symbol='AHCYL2', Aliases='ADOHCYASE3, IRBIT2', description='adenosylhomocysteinase-like 2', other_designations='IP(3)Rs binding protein released with IP(3) 2|S-adenosyl-L-homocysteine hydrolase 3|S-adenosylhomocysteine hydrolase-like 2|long-IRBIT|putative adenosylhomocysteinase 3', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129225014, end_position_on_the_genomic_accession=129430211, orientation='plus', exon_count=21, OMIM=616520, no_hdr0=''),
  285527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285527, CurrentID=0, Status='live', Symbol='FRYL', Aliases='AF4p12, KIAA0826', description='FRY like transcription coactivator', other_designations='ALL1-fused gene from chromosome 4p12 protein|FRY-like|furry homolog-like|furry-like', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48497363, end_position_on_the_genomic_accession=48780312, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  23384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23384, CurrentID=0, Status='live', Symbol='SPECC1L', Aliases='CYTSA, GBBB2, OBLFC1', description='sperm antigen with calponin homology and coiled-coil domains 1-like', other_designations='SPECC1-like protein|cytokinesis and spindle organization A|renal carcinoma antigen NY-REN-22', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24270817, end_position_on_the_genomic_accession=24417740, orientation='plus', exon_count=19, OMIM=614140, no_hdr0=''),
  23385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23385, CurrentID=0, Status='live', Symbol='NCSTN', Aliases='ATAG1874', description='nicastrin', other_designations='anterior pharynx-defective 2', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160343273, end_position_on_the_genomic_accession=160358952, orientation='plus', exon_count=20, OMIM=605254, no_hdr0=''),
  23386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23386, CurrentID=0, Status='live', Symbol='NUDCD3', Aliases='NudCL', description='NudC domain containing 3', other_designations='NudC-like protein', map_location='7p13-p12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44382366, end_position_on_the_genomic_accession=44490878, orientation='minus', exon_count=7, OMIM=610296, no_hdr0=''),
  23387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23387, CurrentID=0, Status='live', Symbol='SIK3', Aliases='L19, QSK, SIK-3', description='SIK family kinase 3', other_designations='salt-inducible kinase 3|serine/threonine-protein kinase QSK', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116843402, end_position_on_the_genomic_accession=117098453, orientation='minus', exon_count=29, OMIM=614776, no_hdr0=''),
  105373392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373392, CurrentID=0, Status='live', Symbol='LOC105373392', Aliases='', description='uncharacterized LOC105373392', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23389, CurrentID=0, Status='live', Symbol='MED13L', Aliases='PROSIT240, THRAP2, TRAP240L', description='mediator complex subunit 13-like', other_designations='thyroid hormone receptor-associated protein 2|thyroid hormone receptor-associated protein complex 240 kDa component-like', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115955771, end_position_on_the_genomic_accession=116277203, orientation='minus', exon_count=32, OMIM=608771, no_hdr0=''),
  23390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23390, CurrentID=0, Status='live', Symbol='ZDHHC17', Aliases='HIP14, HIP3, HSPC294, HYPH', description='zinc finger, DHHC-type containing 17', other_designations='DHHC-17|HIP-14|HIP-3|Huntingtin interacting protein H|huntingtin interacting protein 14|huntingtin interacting protein 3|huntingtin yeast partner H|huntingtin-interacting protein 14|huntingtin-interacting protein 3|huntingtin-interacting protein H|putative MAPK-activating protein PM11|putative NF-kappa-B-activating protein 205|zinc finger DHHC domain-containing protein 17|zinc finger, DHHC domain containing 17', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76764074, end_position_on_the_genomic_accession=76853701, orientation='plus', exon_count=20, OMIM=607799, no_hdr0=''),
  23392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23392, CurrentID=0, Status='live', Symbol='KIAA0368', Aliases='ECM29', description='KIAA0368', other_designations='ECM29 homolog, proteasome accessory protein|homolog of yeast Ecm29', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111360693, end_position_on_the_genomic_accession=111484745, orientation='minus', exon_count=54, OMIM=616694, no_hdr0=''),
  23394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23394, CurrentID=0, Status='live', Symbol='ADNP', Aliases='ADNP1, HVDAS, MRD28', description='activity-dependent neuroprotector homeobox', other_designations='ADNP homeobox 1|activity-dependent neuroprotective protein', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50888918, end_position_on_the_genomic_accession=50930990, orientation='minus', exon_count=8, OMIM=611386, no_hdr0=''),
  23395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23395, CurrentID=0, Status='live', Symbol='LARS2', Aliases='LEURS, PRLTS4, mtLeuRS', description='leucyl-tRNA synthetase 2', other_designations='leucine tRNA ligase 2, mitochondrial|leucine tRNA ligase 2, mitocondrial|leucine translase|leucyl-tRNA synthetase 2, mitochondrial|probable leucyl-tRNA synthetase, mitochondrial', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45388561, end_position_on_the_genomic_accession=45548842, orientation='plus', exon_count=23, OMIM=604544, no_hdr0=''),
  23396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23396, CurrentID=0, Status='live', Symbol='PIP5K1C', Aliases='LCCS3, PIP5K-GAMMA, PIP5K1-gamma, PIP5Kgamma', description='phosphatidylinositol-4-phosphate 5-kinase, type I, gamma', other_designations='diphosphoinositide kinase|ptdIns(4)P-5-kinase 1 gamma|type I PIP kinase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3630181, end_position_on_the_genomic_accession=3700492, orientation='minus', exon_count=25, OMIM=606102, no_hdr0=''),
  23397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23397, CurrentID=0, Status='live', Symbol='NCAPH', Aliases='BRRN1, CAP-H', description='non-SMC condensin I complex subunit H', other_designations='barren homolog 1|chromosome-associated protein H|non-SMC condensin I complex, subunit H', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96335741, end_position_on_the_genomic_accession=96377095, orientation='plus', exon_count=18, OMIM=602332, no_hdr0=''),
  23398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23398, CurrentID=0, Status='live', Symbol='PPWD1', Aliases='', description='peptidylprolyl isomerase domain and WD repeat containing 1', other_designations='spliceosome-associated cyclophilin', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65563236, end_position_on_the_genomic_accession=65587549, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  23399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23399, CurrentID=0, Status='live', Symbol='CTDNEP1', Aliases='DULLARD, HSA011916, NET56', description='CTD nuclear envelope phosphatase 1', other_designations='C-terminal domain nuclear envelope phosphatase 1|dullard homolog|serine/threonine-protein phosphatase dullard', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7243587, end_position_on_the_genomic_accession=7251940, orientation='minus', exon_count=9, OMIM=610684, no_hdr0=''),
  23400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23400, CurrentID=0, Status='live', Symbol='ATP13A2', Aliases='CLN12, HSA9947, KRPPD, PARK9', description='ATPase type 13A2', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16985958, end_position_on_the_genomic_accession=17011972, orientation='minus', exon_count=29, OMIM=610513, no_hdr0=''),
  23401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23401, CurrentID=0, Status='live', Symbol='FRAT2', Aliases='', description='frequently rearranged in advanced T-cell lymphomas 2', other_designations='FRAT-2|GSK-3 binding protein FRAT2', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97332497, end_position_on_the_genomic_accession=97334701, orientation='minus', exon_count=1, OMIM=605006, no_hdr0=''),
  23403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23403, CurrentID=0, Status='live', Symbol='FBXO46', Aliases='20D7-FC4, FBXO34L, Fbx46', description='F-box protein 46', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45710629, end_position_on_the_genomic_accession=45730893, orientation='minus', exon_count=2, OMIM=609117, no_hdr0=''),
  23404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23404, CurrentID=0, Status='live', Symbol='EXOSC2', Aliases='RRP4, Rrp4p, hRrp4p, p7', description='exosome component 2', other_designations="exosome complex exonuclease RRP4|homolog of yeast RRP4 (ribosomal RNA processing 4), 3' 5' exoribonuclease (RRP4)|homolog of yeast RRP4 (ribosomal RNA processing 4), 3'-5'-exoribonuclease|ribosomal RNA-processing protein 4", map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130693760, end_position_on_the_genomic_accession=130704894, orientation='plus', exon_count=10, OMIM=602238, no_hdr0=''),
  23405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23405, CurrentID=0, Status='live', Symbol='DICER1', Aliases='DCR1, Dicer, Dicer1e, HERNA, K12H4.8-LIKE, MNG1, RMSE2', description='dicer 1, ribonuclease type III', other_designations='Dicer1, Dcr-1 homolog|dicer 1, double-stranded RNA-specific endoribonuclease|helicase MOI|helicase with RNAse motif', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95086228, end_position_on_the_genomic_accession=95158010, orientation='minus', exon_count=34, OMIM=606241, no_hdr0=''),
  285550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285550, CurrentID=0, Status='live', Symbol='FAM200B', Aliases='', description='family with sequence similarity 200 member B', other_designations='Putative protein FAM200B|family with sequence similarity 200, member B', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15681729, end_position_on_the_genomic_accession=15690447, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23408, CurrentID=0, Status='live', Symbol='SIRT5', Aliases='SIR2L5', description='sirtuin 5', other_designations='NAD-dependent deacetylase sirtuin-5|NAD-dependent lysine demalonylase and desuccinylase sirtuin-5, mitochondrial|silent mating type information regulation 2, S.cerevisiae, homolog 5|sir2-like 5|sirtuin type 5', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13574274, end_position_on_the_genomic_accession=13615158, orientation='plus', exon_count=15, OMIM=604483, no_hdr0=''),
  23409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23409, CurrentID=0, Status='live', Symbol='SIRT4', Aliases='SIR2L4', description='sirtuin 4', other_designations='NAD-dependent ADP-ribosyltransferase sirtuin-4|NAD-dependent protein deacetylase sirtuin-4|SIR2-like protein 4|regulatory protein SIR2 homolog 4|sir2-like 4|sirtuin type 4', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120292199, end_position_on_the_genomic_accession=120313242, orientation='plus', exon_count=6, OMIM=604482, no_hdr0=''),
  23410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23410, CurrentID=0, Status='live', Symbol='SIRT3', Aliases='SIR2L3', description='sirtuin 3', other_designations='NAD-dependent deacetylase sirtuin-3, mitochondrial|SIR2-like protein 3|mitochondrial nicotinamide adenine dinucleotide-dependent deacetylase|regulatory protein SIR2 homolog 3|silent mating type information regulation 2, S.cerevisiae, homolog 3|sir2-like 3|sirtuin type 3', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=215030, end_position_on_the_genomic_accession=236362, orientation='minus', exon_count=8, OMIM=604481, no_hdr0=''),
  285555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285555, CurrentID=0, Status='live', Symbol='STPG2', Aliases='C4orf37', description='sperm-tail PG-rich repeat containing 2', other_designations='', map_location='4q22.3-q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=97441146, end_position_on_the_genomic_accession=98143470, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  23412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23412, CurrentID=0, Status='live', Symbol='COMMD3', Aliases='BUP, C10orf8', description='COMM domain containing 3', other_designations='protein Bup', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22316383, end_position_on_the_genomic_accession=22320317, orientation='plus', exon_count=8, OMIM=616700, no_hdr0=''),
  23413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23413, CurrentID=0, Status='live', Symbol='NCS1', Aliases='FLUP, FREQ', description='neuronal calcium sensor 1', other_designations='frequenin homolog|frequenin-like protein|frequenin-like ubiquitous protein', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130172578, end_position_on_the_genomic_accession=130237304, orientation='plus', exon_count=9, OMIM=603315, no_hdr0=''),
  23414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23414, CurrentID=0, Status='live', Symbol='ZFPM2', Aliases='DIH3, FOG2, SRXY9, ZC2HC11B, ZNF89B, hFOG-2', description='zinc finger protein, FOG family member 2', other_designations='FOG-2|Friend of GATA2|friend of GATA 2|friend of GATA protein 2|transcription factor GATA4, modulator of|zinc finger protein 89B|zinc finger protein, multitype 2', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105246436, end_position_on_the_genomic_accession=105804539, orientation='plus', exon_count=10, OMIM=603693, no_hdr0=''),
  23415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23415, CurrentID=0, Status='live', Symbol='KCNH4', Aliases='BEC2, ELK1, Kv12.3', description='potassium channel, voltage gated eag related subfamily H, member 4', other_designations='ELK channel 1|brain-specific eag-like channel 2|ether-a-go-go K(+) channel family member|ether-a-go-go-like potassium channel 1|potassium voltage-gated channel, subfamily H (eag-related), member 4|voltage-gated potassium channel subunit Kv12.3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42156892, end_position_on_the_genomic_accession=42181278, orientation='minus', exon_count=17, OMIM=604528, no_hdr0=''),
  23416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23416, CurrentID=0, Status='live', Symbol='KCNH3', Aliases='BEC1, ELK2, Kv12.2', description='potassium channel, voltage gated eag related subfamily H, member 3', other_designations='ELK channel 2|brain-specific eag-like channel 1|ether-a-go-go K(+) channel family member|ether-a-go-go-like potassium channel 2|potassium voltage-gated channel, subfamily H (eag-related), member 3|voltage-gated potassium channel subunit Kv12.2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49539041, end_position_on_the_genomic_accession=49558308, orientation='plus', exon_count=16, OMIM=604527, no_hdr0=''),
  23417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23417, CurrentID=0, Status='live', Symbol='MLYCD', Aliases='MCD', description='malonyl-CoA decarboxylase', other_designations='malonyl coenzyme A decarboxylase', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83899125, end_position_on_the_genomic_accession=83916182, orientation='plus', exon_count=5, OMIM=606761, no_hdr0=''),
  23418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23418, CurrentID=0, Status='live', Symbol='CRB1', Aliases='LCA8, RP12', description='crumbs family member 1, photoreceptor morphogenesis associated', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197201462, end_position_on_the_genomic_accession=197478455, orientation='plus', exon_count=18, OMIM=604210, no_hdr0=''),
  23420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23420, CurrentID=0, Status='live', Symbol='NOMO1', Aliases='Nomo, PM5', description='NODAL modulator 1', other_designations='nodal modulator 3|pM5 protein 3|pM5 protein, telomeric copy', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14833786, end_position_on_the_genomic_accession=14896157, orientation='plus', exon_count=31, OMIM=609157, no_hdr0=''),
  23421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23421, CurrentID=0, Status='live', Symbol='ITGB3BP', Aliases='CENP-R, CENPR, HSU37139, NRIF3, TAP20', description='integrin subunit beta 3 binding protein', other_designations='beta 3 endonexin|beta-3-endonexin|integrin beta 3 binding protein (beta3-endonexin)|integrin beta-3-binding protein|nuclear receptor-interacting factor 3', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63440770, end_position_on_the_genomic_accession=63529159, orientation='minus', exon_count=14, OMIM=605494, no_hdr0=''),
  23423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23423, CurrentID=0, Status='live', Symbol='TMED3', Aliases='C15orf22, P24B, p26', description='transmembrane p24 trafficking protein 3', other_designations='integral type I protein|membrane protein p24B|p24 family protein gamma-4|p24gamma4|transmembrane emp24 protein transport domain containing 3', map_location='15q24-q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79311062, end_position_on_the_genomic_accession=79384721, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23424, CurrentID=0, Status='live', Symbol='TDRD7', Aliases='CATC4, PCTAIRE2BP, TRAP', description='tudor domain containing 7', other_designations='PCTAIRE2-binding protein|tudor repeat associator with PCTAIRE 2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97411989, end_position_on_the_genomic_accession=97496125, orientation='plus', exon_count=17, OMIM=611258, no_hdr0=''),
  23426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23426, CurrentID=0, Status='live', Symbol='GRIP1', Aliases='GRIP', description='glutamate receptor interacting protein 1', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66347398, end_position_on_the_genomic_accession=67069267, orientation='minus', exon_count=28, OMIM=604597, no_hdr0=''),
  23428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23428, CurrentID=0, Status='live', Symbol='SLC7A8', Aliases='LAT2, LPI-PC1', description='solute carrier family 7 (amino acid transporter light chain, L system), member 8', other_designations='L-type amino acid transporter 2|integral membrane protein E16H|solute carrier family 7 (amino acid transporter, L-type), member 8|solute carrier family 7 (cationic amino acid transporter, y+ system), member 8', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23125295, end_position_on_the_genomic_accession=23183660, orientation='minus', exon_count=13, OMIM=604235, no_hdr0=''),
  23429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23429, CurrentID=0, Status='live', Symbol='RYBP', Aliases='AAP1, APAP-1, DEDAF, YEAF1', description='RING1 and YY1 binding protein', other_designations='DED-associated factor|YY1 and E4TF1 associated factor 1|YY1 and E4TF1-associated factor 1|apoptin-associating protein 1|death effector domain-associated factor|ring1 interactor RYBP', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72374593, end_position_on_the_genomic_accession=72446623, orientation='minus', exon_count=4, OMIM=607535, no_hdr0=''),
  23430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23430, CurrentID=0, Status='live', Symbol='TPSD1', Aliases='MCP7-LIKE, MCP7L1, MMCP-7L', description='tryptase delta 1', other_designations='delta-tryptase|hmMCP-3-like tryptase III|mMCP-7-like delta II tryptase|mMCP-7-like-1|mMCP-7-like-2|mast cell tryptase|tryptase-3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1256272, end_position_on_the_genomic_accession=1258493, orientation='plus', exon_count=5, OMIM=609272, no_hdr0=''),
  23431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23431, CurrentID=0, Status='live', Symbol='AP4E1', Aliases='CPSQ4, SPG51', description='adaptor related protein complex 4, epsilon 1 subunit', other_designations='AP-4 adaptor complex subunit epsilon|adaptor-related protein complex 4 subunit epsilon-1|adaptor-related protein complex 4, epsilon 1 subunit|adaptor-related protein complex AP-4 epsilon|epsilon-adaptin', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50908625, end_position_on_the_genomic_accession=51005900, orientation='plus', exon_count=24, OMIM=607244, no_hdr0=''),
  23432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23432, CurrentID=0, Status='live', Symbol='GPR161', Aliases='RE2', description='G protein-coupled receptor 161', other_designations='G-protein coupled receptor RE2', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168079542, end_position_on_the_genomic_accession=168137667, orientation='minus', exon_count=13, OMIM=612250, no_hdr0=''),
  23433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23433, CurrentID=0, Status='live', Symbol='RHOQ', Aliases='ARHQ, HEL-S-42, RASL7A, TC10, TC10A', description='ras homolog family member Q', other_designations='RAS-like, family 7, member A|epididymis secretory protein Li 42|ras homolog gene family, member Q|ras-like protein TC10|ras-like protein family member 7A', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46542728, end_position_on_the_genomic_accession=46584688, orientation='plus', exon_count=7, OMIM=605857, no_hdr0=''),
  23434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23434, CurrentID=0, Status='live', Symbol='LINC01565', Aliases='C3orf27, GR6', description='long intergenic non-protein coding RNA 1565', other_designations='putative GR6 protein', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128572000, end_position_on_the_genomic_accession=128576086, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23435, CurrentID=0, Status='live', Symbol='TARDBP', Aliases='ALS10, TDP-43', description='TAR DNA binding protein', other_designations='TAR DNA-binding protein-43', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11012622, end_position_on_the_genomic_accession=11025492, orientation='plus', exon_count=6, OMIM=605078, no_hdr0=''),
  23436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23436, CurrentID=0, Status='live', Symbol='CELA3B', Aliases='CBPP, E1, EL-1, ELA3B', description='chymotrypsin like elastase family member 3B', other_designations='cholesterol-binding pancreatic protease|chymotrypsin-like elastase family, member 3B|elastase 1|elastase 3B, pancreatic|elastase IIIB|elastase-3B|fecal elastase 1|pancreatic elastase 1|pancreatic endopeptidase E|protease E|proteinase E', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21976925, end_position_on_the_genomic_accession=21998642, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23437, CurrentID=0, Status='live', Symbol='TRS-AGA2-4', Aliases='TRNAS2, TRNS, TRS2, tRNA-SER', description='transfer RNA-Ser (AGA) 2-4', other_designations='tRNA serine (AGA) 2|tRNA serine 2 (anticodon AGA)|tRNA, serine (AGA)|tRNA-Ser-AGA-2-4|transfer RNA serine 2 (anticodon AGA)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27495814, end_position_on_the_genomic_accession=27495895, orientation='plus', exon_count=0, OMIM=189910, no_hdr0=''),
  23438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23438, CurrentID=0, Status='live', Symbol='HARS2', Aliases='HARSL, HARSR, HO3, PRLTS2', description='histidyl-tRNA synthetase 2', other_designations='HARS-related|hisRS|histidine tRNA ligase 2, mitochondrial (putative)|histidine translase|histidine-tRNA ligase homolog|histidyl-tRNA synthetase 2, mitochondrial (putative)|probable histidyl-tRNA synthetase, mitochondrial', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140691426, end_position_on_the_genomic_accession=140699318, orientation='plus', exon_count=13, OMIM=600783, no_hdr0=''),
  23439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23439, CurrentID=0, Status='live', Symbol='ATP1B4', Aliases='', description='ATPase, Na+/K+ transporting, beta 4 polypeptide', other_designations='Na,K-ATPase beta m-subunit|X,K-ATPase beta-m subunit|x/potassium-transporting ATPase subunit beta-m', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120362085, end_position_on_the_genomic_accession=120383249, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  23440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23440, CurrentID=0, Status='live', Symbol='OTP', Aliases='', description='orthopedia homeobox', other_designations='orthopedia homolog', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77628712, end_position_on_the_genomic_accession=77638697, orientation='minus', exon_count=4, OMIM=604529, no_hdr0=''),
  23443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23443, CurrentID=0, Status='live', Symbol='SLC35A3', Aliases='AMRS', description='solute carrier family 35 (UDP-N-acetylglucosamine (UDP-GlcNAc) transporter), member A3', other_designations='golgi UDP-GlcNAc transporter|solute carrier family 35 (UDP-N-acetylglucosamine (UDP-GlcNAc) transporter), member 3', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99968887, end_position_on_the_genomic_accession=100026979, orientation='plus', exon_count=11, OMIM=605632, no_hdr0=''),
  285588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285588, CurrentID=0, Status='live', Symbol='EFCAB9', Aliases='', description='EF-hand calcium binding domain 9', other_designations='EF-hand domain-containing protein', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172188430, end_position_on_the_genomic_accession=172203454, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  23446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23446, CurrentID=0, Status='live', Symbol='SLC44A1', Aliases='CD92, CDW92, CHTL1, CTL1', description='solute carrier family 44 (choline transporter), member 1', other_designations='CDW92 antigen|solute carrier family 44, member 1', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105244608, end_position_on_the_genomic_accession=105438504, orientation='plus', exon_count=19, OMIM=606105, no_hdr0=''),
  23447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23447, CurrentID=0, Status='live', Symbol='CDW93', Aliases='', description='CDw93 antigen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23448, CurrentID=0, Status='live', Symbol='CD139', Aliases='', description='CD139 antigen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23449, CurrentID=0, Status='live', Symbol='CD165', Aliases='', description='CD165 antigen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23450, CurrentID=0, Status='live', Symbol='SF3B3', Aliases='RSE1, SAP130, SF3b130, STAF130', description='splicing factor 3b subunit 3', other_designations='SAP 130|pre-mRNA splicing factor SF3b, 130 kDa subunit|pre-mRNA-splicing factor SF3b 130 kDa subunit|spliceosome-associated protein 130|splicing factor 3b, subunit 3, 130kD|splicing factor 3b, subunit 3, 130kDa', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70523788, end_position_on_the_genomic_accession=70577668, orientation='plus', exon_count=26, OMIM=605592, no_hdr0=''),
  23451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23451, CurrentID=0, Status='live', Symbol='SF3B1', Aliases='Hsh155, MDS, PRP10, PRPF10, SAP155, SF3b155', description='splicing factor 3b subunit 1', other_designations='pre-mRNA processing 10|pre-mRNA splicing factor SF3b, 155 kDa subunit|spliceosome-associated protein 155|splicing factor 3b, subunit 1, 155kDa', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197391974, end_position_on_the_genomic_accession=197435091, orientation='minus', exon_count=27, OMIM=605590, no_hdr0=''),
  285596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285596, CurrentID=0, Status='live', Symbol='FAM153A', Aliases='NY-REN-7', description='family with sequence similarity 153 member A', other_designations='NY REN 7 antigen|family with sequence similarity 153, member A|renal carcinoma antigen NY-REN-7', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177722380, end_position_on_the_genomic_accession=177783398, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  285598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285598, CurrentID=0, Status='live', Symbol='ARL10', Aliases='ARL10A', description='ADP ribosylation factor like GTPase 10', other_designations='ADP-ribosylation factor like GTPase 10|ADP-ribosylation factor-like 10A|ADP-ribosylation factor-like membrane-associated protein', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176365474, end_position_on_the_genomic_accession=176415463, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23456, CurrentID=0, Status='live', Symbol='ABCB10', Aliases='EST20237, M-ABC2, MTABC2', description='ATP binding cassette subfamily B member 10', other_designations='ABC transporter 10 protein|ATP-binding cassette transporter 10|ATP-binding cassette, sub-family B (MDR/TAP), member 10|mitochondrial ATP-binding cassette 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229516582, end_position_on_the_genomic_accession=229558695, orientation='minus', exon_count=14, OMIM=605454, no_hdr0=''),
  23457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23457, CurrentID=0, Status='live', Symbol='ABCB9', Aliases='EST122234, TAPL', description='ATP binding cassette subfamily B member 9', other_designations='ABC transporter 9 protein|ATP-binding cassette, sub-family B (MDR/TAP), member 9|TAP-like protein', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122918708, end_position_on_the_genomic_accession=122975125, orientation='minus', exon_count=19, OMIM=605453, no_hdr0=''),
  23459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23459, CurrentID=0, Status='live', Symbol='CMD1H', Aliases='', description='cardiomyopathy, dilated 1H (autosomal dominant)', other_designations='', map_location='2q14-q22', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604288, no_hdr0=''),
  23460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23460, CurrentID=0, Status='live', Symbol='ABCA6', Aliases='EST155051', description='ATP binding cassette subfamily A member 6', other_designations='ABC transporter ABCA6|ATP-binding cassette A6|ATP-binding cassette, sub-family A (ABC1), member 6', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69078089, end_position_on_the_genomic_accession=69141992, orientation='minus', exon_count=40, OMIM=612504, no_hdr0=''),
  23461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23461, CurrentID=0, Status='live', Symbol='ABCA5', Aliases='ABC13, EST90625', description='ATP binding cassette subfamily A member 5', other_designations='ATP-binding cassette A5|ATP-binding cassette, sub-family A (ABC1), member 5', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69244435, end_position_on_the_genomic_accession=69327182, orientation='minus', exon_count=39, OMIM=612503, no_hdr0=''),
  23462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23462, CurrentID=0, Status='live', Symbol='HEY1', Aliases='BHLHb31, CHF2, HERP2, HESR1, HRT-1, OAF1, hHRT1', description='hes-related family bHLH transcription factor with YRPW motif 1', other_designations='HES-related repressor protein 1|HES-related repressor protein 2|basic helix-loop-helix protein OAF1|cardiovascular helix-loop-helix factor 2|class B basic helix-loop-helix protein 31|hairy and enhancer of split-related protein 1|hairy-related transcription factor 1|hairy/enhancer-of-split related with YRPW motif 1', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79764010, end_position_on_the_genomic_accession=79767863, orientation='minus', exon_count=6, OMIM=602953, no_hdr0=''),
  23463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23463, CurrentID=0, Status='live', Symbol='ICMT', Aliases='HSTE14, MST098, MSTP098, PCCMT, PCMT, PPMT', description='isoprenylcysteine carboxyl methyltransferase', other_designations='prenylated protein carboxyl methyltransferase|prenylcysteine carboxyl methyltransferase|protein-S isoprenylcysteine O-methyltransferase', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6221193, end_position_on_the_genomic_accession=6235984, orientation='minus', exon_count=6, OMIM=605851, no_hdr0=''),
  23464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23464, CurrentID=0, Status='live', Symbol='GCAT', Aliases='KBL', description='glycine C-acetyltransferase', other_designations='2-amino-3-ketobutyrate-CoA ligase|AKB ligase|aminoacetone synthase|glycine acetyltransferase', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37807900, end_position_on_the_genomic_accession=37817327, orientation='plus', exon_count=10, OMIM=607422, no_hdr0=''),
  23466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23466, CurrentID=0, Status='live', Symbol='CBX6', Aliases='', description='chromobox 6', other_designations='chromobox homolog 6', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38861216, end_position_on_the_genomic_accession=38872334, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23467, CurrentID=0, Status='live', Symbol='NPTXR', Aliases='NPR', description='neuronal pentraxin receptor', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38818451, end_position_on_the_genomic_accession=38844012, orientation='minus', exon_count=5, OMIM=609474, no_hdr0=''),
  23468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23468, CurrentID=0, Status='live', Symbol='CBX5', Aliases='HEL25, HP1, HP1A', description='chromobox 5', other_designations='HP1 alpha homolog|HP1-ALPHA|HP1Hs alpha|antigen p25|chromobox homolog 5 (HP1 alpha homolog, Drosophila)|epididymis luminal protein 25|heterochromatin protein 1 homolog alpha|heterochromatin protein 1-alpha', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54230946, end_position_on_the_genomic_accession=54280131, orientation='minus', exon_count=6, OMIM=604478, no_hdr0=''),
  23469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23469, CurrentID=0, Status='live', Symbol='PHF3', Aliases='', description='PHD finger protein 3', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63635802, end_position_on_the_genomic_accession=63715522, orientation='plus', exon_count=18, OMIM=607789, no_hdr0=''),
  23471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23471, CurrentID=0, Status='live', Symbol='TRAM1', Aliases='PNAS8, TRAM, TRAMP', description='translocation associated membrane protein 1', other_designations='translocating chain-associating membrane protein|translocation-associating membrane protein 1', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70573218, end_position_on_the_genomic_accession=70608459, orientation='minus', exon_count=12, OMIM=605190, no_hdr0=''),
  23473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23473, CurrentID=0, Status='live', Symbol='CAPN7', Aliases='CALPAIN7, PALBH', description='calpain 7', other_designations='calpain like protease|homolog of Aspergillus Nidulans PALB|palB homolog', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15206226, end_position_on_the_genomic_accession=15252916, orientation='plus', exon_count=23, OMIM=606400, no_hdr0=''),
  23474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23474, CurrentID=0, Status='live', Symbol='ETHE1', Aliases='HSCO, YF13H12', description='ethylmalonic encephalopathy 1', other_designations='ethylmalonic encephalopathy protein 1|hepatoma subtracted clone one protein|protein ETHE1, mitochondrial|sulfur dioxygenase ETHE1', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43506719, end_position_on_the_genomic_accession=43527244, orientation='minus', exon_count=7, OMIM=608451, no_hdr0=''),
  23475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23475, CurrentID=0, Status='live', Symbol='QPRT', Aliases='HEL-S-90nase, QPRT', description='quinolinate phosphoribosyltransferase', other_designations='epididymis secretory sperm binding protein Li 90n|nicotinate-nucleotide pyrophosphorylase (carboxylating)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29669840, end_position_on_the_genomic_accession=29697994, orientation='plus', exon_count=5, OMIM=606248, no_hdr0=''),
  23476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23476, CurrentID=0, Status='live', Symbol='BRD4', Aliases='CAP, HUNK1, HUNKI, MCAP', description='bromodomain containing 4', other_designations='bromodomain-containing 4|chromosome-associated protein', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15236836, end_position_on_the_genomic_accession=15332543, orientation='minus', exon_count=23, OMIM=608749, no_hdr0=''),
  23478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23478, CurrentID=0, Status='live', Symbol='SEC11A', Aliases='1810012E07Rik, SEC11L1, SPC18, SPCS4A, sid2895', description='SEC11 homolog A, signal peptidase complex subunit', other_designations='SEC11 homolog A|SEC11-like protein 1|SPase 18 kDa subunit|endopeptidase SP18|microsomal signal peptidase 18 kDa subunit|signal peptidase complex (18kD)', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84669537, end_position_on_the_genomic_accession=84716460, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23479, CurrentID=0, Status='live', Symbol='ISCU', Aliases='2310020H20Rik, HML, ISU2, NIFU, NIFUN, hnifU', description='iron-sulfur cluster assembly enzyme', other_designations='IscU iron-sulfur cluster scaffold homolog|nifU-like N-terminal domain-containing protein', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108561463, end_position_on_the_genomic_accession=108569384, orientation='plus', exon_count=8, OMIM=611911, no_hdr0=''),
  23480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23480, CurrentID=0, Status='live', Symbol='SEC61G', Aliases='SSS1', description='Sec61 translocon gamma subunit', other_designations='Sec61 gamma subunit|protein transport protein SEC61 gamma subunit', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54752247, end_position_on_the_genomic_accession=54759246, orientation='minus', exon_count=4, OMIM=609215, no_hdr0=''),
  23481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23481, CurrentID=0, Status='live', Symbol='PES1', Aliases='PES', description='pescadillo ribosomal biogenesis factor 1', other_designations='pescadillo homolog 1, containing BRCT domain', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30576625, end_position_on_the_genomic_accession=30607013, orientation='minus', exon_count=18, OMIM=605819, no_hdr0=''),
  285626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285626, CurrentID=0, Status='live', Symbol='LOC285626', Aliases='', description='uncharacterized LOC285626', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159331518, end_position_on_the_genomic_accession=159362834, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23483, CurrentID=0, Status='live', Symbol='TGDS', Aliases='CATMANS, SDR2E1, TDPGD', description='TDP-glucose 4,6-dehydratase', other_designations='growth-inhibiting protein 21|short chain dehydrogenase/reductase family 2E, member 1', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94574054, end_position_on_the_genomic_accession=94596275, orientation='minus', exon_count=12, OMIM=616146, no_hdr0=''),
  23484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23484, CurrentID=0, Status='live', Symbol='LEPROTL1', Aliases='HSPC112, Vps55, my047', description='leptin receptor overlapping transcript-like 1', other_designations='endospanin-2', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30095406, end_position_on_the_genomic_accession=30137706, orientation='plus', exon_count=5, OMIM=607338, no_hdr0=''),
  285629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285629, CurrentID=0, Status='live', Symbol='LOC285629', Aliases='', description='uncharacterized LOC285629', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160931778, end_position_on_the_genomic_accession=160938626, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105274304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105274304, CurrentID=0, Status='live', Symbol='LOC105274304', Aliases='', description='uncharacterized LOC105274304', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76671928, end_position_on_the_genomic_accession=76673658, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  285634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285634, CurrentID=0, Status='live', Symbol='INTS6P1', Aliases='', description='integrator complex subunit 6 pseudogene 1', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39718984, end_position_on_the_genomic_accession=39721512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23491, CurrentID=0, Status='live', Symbol='CES3', Aliases='ES31', description='carboxylesterase 3', other_designations='carboxylesterase 3 (brain)|esterase 31|liver carboxylesterase 31 homolog', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66961229, end_position_on_the_genomic_accession=66975149, orientation='plus', exon_count=15, OMIM=605279, no_hdr0=''),
  23492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23492, CurrentID=0, Status='live', Symbol='CBX7', Aliases='', description='chromobox 7', other_designations='chromobox homolog 7', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39130772, end_position_on_the_genomic_accession=39152662, orientation='minus', exon_count=6, OMIM=608457, no_hdr0=''),
  23493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23493, CurrentID=0, Status='live', Symbol='HEY2', Aliases='CHF1, GRIDLOCK, GRL, HERP1, HESR2, HRT2, bHLHb32', description='hes-related family bHLH transcription factor with YRPW motif 2', other_designations='HES-related repressor protein 1|HES-related repressor protein 2|HESR-2|HRT-2|cardiovascular basic helix-loop-helix factor 1|cardiovascular helix-loop-helix factor 1|class B basic helix-loop-helix protein 32|hCHF1|hHRT2|hairy and enhancer of split-related protein 2|hairy-related transcription factor 2|hairy/enhancer-of-split related with YRPW motif 2|protein gridlock homolog', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125749586, end_position_on_the_genomic_accession=125761271, orientation='plus', exon_count=8, OMIM=604674, no_hdr0=''),
  285638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285638, CurrentID=0, Status='live', Symbol='LOC285638', Aliases='', description='uncharacterized LOC285638', other_designations='CTD-2587M2.1', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109237134, end_position_on_the_genomic_accession=109330452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23495, CurrentID=0, Status='live', Symbol='TNFRSF13B', Aliases='CD267, CVID, CVID2, IGAD2, RYZN, TACI, TNFRSF14B', description='tumor necrosis factor receptor superfamily member 13B', other_designations='transmembrane activator and CAML interactor|tumor necrosis factor receptor 13B|tumor necrosis factor receptor superfamily, member 13B', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16939084, end_position_on_the_genomic_accession=16972088, orientation='minus', exon_count=5, OMIM=604907, no_hdr0=''),
  285641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285641, CurrentID=0, Status='live', Symbol='SLC36A3', Aliases='PAT3, TRAMD2, tramdorin2', description='solute carrier family 36 member 3', other_designations='proton/amino acid transporter 3|solute carrier family 36 (proton/amino acid symporter), member 3|solute carrier family 36, member 3|tramdorin 2|tramdorin-2', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151276365, end_position_on_the_genomic_accession=151312981, orientation='minus', exon_count=12, OMIM=608332, no_hdr0=''),
  23498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23498, CurrentID=0, Status='live', Symbol='HAAO', Aliases='3-HAO, HAO', description='3-hydroxyanthranilate 3,4-dioxygenase', other_designations='3-HAO|3-hydroxyanthranilate oxygenase|3-hydroxyanthranilic acid dioxygenase|HAD', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42767089, end_position_on_the_genomic_accession=42793138, orientation='minus', exon_count=11, OMIM=604521, no_hdr0=''),
  23499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23499, CurrentID=0, Status='live', Symbol='MACF1', Aliases='ABP620, ACF7, MACF, OFC4', description='microtubule-actin crosslinking factor 1', other_designations='620 kDa actin binding protein|actin cross-linking family protein 7|macrophin 1|trabeculin-alpha', map_location='1p32-p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39081417, end_position_on_the_genomic_accession=39487138, orientation='plus', exon_count=109, OMIM=608271, no_hdr0=''),
  23500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23500, CurrentID=0, Status='live', Symbol='DAAM2', Aliases='dJ90A20A.1', description='dishevelled associated activator of morphogenesis 2', other_designations='dishevelled-associated activator of morphogenesis 2', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39792366, end_position_on_the_genomic_accession=39904877, orientation='plus', exon_count=32, OMIM=606627, no_hdr0=''),
  105379811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379811, CurrentID=0, Status='live', Symbol='LOC105379811', Aliases='', description='uncharacterized LOC105379811', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23503, CurrentID=0, Status='live', Symbol='ZFYVE26', Aliases='FYVE-CENT, SPG15', description='zinc finger, FYVE domain containing 26', other_designations='FYVE domain-containing centrosomal protein|spastizin', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67729569, end_position_on_the_genomic_accession=67816611, orientation='minus', exon_count=45, OMIM=612012, no_hdr0=''),
  23504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23504, CurrentID=0, Status='live', Symbol='RIMBP2', Aliases='PPP1R133, RBP2, RIM-BP2', description='RIMS binding protein 2', other_designations='RIM binding protein 2|protein phosphatase 1, regulatory subunit 133', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130396133, end_position_on_the_genomic_accession=130716351, orientation='minus', exon_count=25, OMIM=611602, no_hdr0=''),
  23505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23505, CurrentID=0, Status='live', Symbol='TMEM131', Aliases='CC28, PRO1048, RW1, YR-23', description='transmembrane protein 131', other_designations='2610524E03Rik', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97756336, end_position_on_the_genomic_accession=97995932, orientation='minus', exon_count=42, OMIM=615659, no_hdr0=''),
  23506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23506, CurrentID=0, Status='live', Symbol='GLTSCR1L', Aliases='KIAA0240', description='GLTSCR1 like', other_designations='GLTSCR1-like|glioma tumor suppressor candidate region gene 1 protein-like', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42782052, end_position_on_the_genomic_accession=42868560, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  23507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23507, CurrentID=0, Status='live', Symbol='LRRC8B', Aliases='TA-LRRP, TALRRP', description='leucine-rich repeat containing 8 family member B', other_designations='T cell activation leucine repeat rich protein|T-cell activation leucine repeat-rich protein|leucine rich repeat containing 8 family, member B|leucine-rich repeat-containing protein 8B', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89524144, end_position_on_the_genomic_accession=89601539, orientation='plus', exon_count=12, OMIM=612888, no_hdr0=''),
  23508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23508, CurrentID=0, Status='live', Symbol='TTC9', Aliases='TTC9A', description='tetratricopeptide repeat domain 9', other_designations='TPR repeat protein 9A', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70641787, end_position_on_the_genomic_accession=70675360, orientation='plus', exon_count=3, OMIM=610488, no_hdr0=''),
  23509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23509, CurrentID=0, Status='live', Symbol='POFUT1', Aliases='DDD2, FUT12, O-FUT, O-Fuc-T, O-FucT-1, OFUCT1', description='protein O-fucosyltransferase 1', other_designations='o-fucosyltransferase protein|peptide-O-fucosyltransferase 1', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32207822, end_position_on_the_genomic_accession=32238664, orientation='plus', exon_count=9, OMIM=607491, no_hdr0=''),
  23510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23510, CurrentID=0, Status='live', Symbol='KCTD2', Aliases='', description='potassium channel tetramerization domain containing 2', other_designations='potassium channel tetramerisation domain containing 2|potassium channel tetramerization domain-containing protein 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75032567, end_position_on_the_genomic_accession=75065889, orientation='plus', exon_count=10, OMIM=613422, no_hdr0=''),
  23511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23511, CurrentID=0, Status='live', Symbol='NUP188', Aliases='KIAA0169, hNup188', description='nucleoporin 188kDa', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128947693, end_position_on_the_genomic_accession=129007096, orientation='plus', exon_count=44, OMIM=615587, no_hdr0=''),
  23512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23512, CurrentID=0, Status='live', Symbol='SUZ12', Aliases='CHET9, JJAZ1', description='SUZ12 polycomb repressive complex 2 subunit', other_designations='chET 9 protein|chromatin precipitated E2F target 9 protein|joined to JAZF1 protein|suppressor of zeste 12 homolog|suppressor of zeste 12 protein homolog', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31937025, end_position_on_the_genomic_accession=32001045, orientation='plus', exon_count=18, OMIM=606245, no_hdr0=''),
  23513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23513, CurrentID=0, Status='live', Symbol='SCRIB', Aliases='CRIB1, SCRB11, Vartul, SCRIB', description='scribbled planar cell polarity protein', other_designations='scribbled homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143790920, end_position_on_the_genomic_accession=143815379, orientation='minus', exon_count=37, OMIM=607733, no_hdr0=''),
  23514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23514, CurrentID=0, Status='live', Symbol='SPIDR', Aliases='KIAA0146', description='scaffolding protein involved in DNA repair', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47260878, end_position_on_the_genomic_accession=47736001, orientation='plus', exon_count=31, OMIM=615384, no_hdr0=''),
  285659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285659, CurrentID=0, Status='live', Symbol='OR2V2', Aliases='OR2V3, OST713', description='olfactory receptor family 2 subfamily V member 2', other_designations='olfactory receptor 2V3|olfactory receptor OR5-3|olfactory receptor, family 2, subfamily V, member 2|olfactory receptor, family 2, subfamily V, member 3', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181154943, end_position_on_the_genomic_accession=181155890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23516, CurrentID=0, Status='live', Symbol='SLC39A14', Aliases='LZT-Hs4, NET34, ZIP14, cig19', description='solute carrier family 39 (zinc transporter), member 14', other_designations='LIV-1 subfamily of ZIP zinc transporter 4|ZIP-14|Zrt-, Irt-like protein 14|solute carrier family 39 (metal ion transporter), member 14|solute carrier family 39 member 14|zrt- and Irt-like protein 14', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22367249, end_position_on_the_genomic_accession=22434129, orientation='plus', exon_count=16, OMIM=608736, no_hdr0=''),
  23517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23517, CurrentID=0, Status='live', Symbol='SKIV2L2', Aliases='Dob1, KIAA0052, Mtr4, fSAP118', description='superkiller viralicidic activity 2-like 2 (S. cerevisiae)', other_designations='ATP-dependent RNA helicase DOB1|ATP-dependent RNA helicase SKIV2L2|ATP-dependent helicase SKIV2L2|TRAMP-like complex helicase|functional spliceosome-associated protein 118', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55307748, end_position_on_the_genomic_accession=55425581, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  23518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23518, CurrentID=0, Status='live', Symbol='R3HDM1', Aliases='R3HDM', description='R3H domain containing 1', other_designations='R3H domain (binds single-stranded nucleic acids) containing', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135531455, end_position_on_the_genomic_accession=135725270, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  23519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23519, CurrentID=0, Status='live', Symbol='ANP32D', Aliases='PP32R2', description='acidic nuclear phosphoprotein 32 family member D', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member D|acidic nuclear phosphoprotein 32D|phosphoprotein 32-related protein 2|pp32 related 2|tumorigenic protein pp32r2', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48472665, end_position_on_the_genomic_accession=48473060, orientation='plus', exon_count=1, OMIM=606878, no_hdr0=''),
  23520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23520, CurrentID=0, Status='live', Symbol='ANP32C', Aliases='PP32R1', description='acidic nuclear phosphoprotein 32 family member C', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member C|phosphoprotein 32-related protein 1|pp32 related 1|tumorigenic protein pp32r1', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164197007, end_position_on_the_genomic_accession=164197711, orientation='minus', exon_count=1, OMIM=606877, no_hdr0=''),
  23521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23521, CurrentID=0, Status='live', Symbol='RPL13A', Aliases='L13A, TSTA1', description='ribosomal protein L13a', other_designations='23 kDa highly basic protein|tissue specific transplantation antigen 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49487554, end_position_on_the_genomic_accession=49492308, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  23522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23522, CurrentID=0, Status='live', Symbol='KAT6B', Aliases='GTPTS, MORF, MOZ2, MYST4, ZC2HC6B, qkf, querkopf', description='K(lysine) acetyltransferase 6B', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 4|MOZ-related factor|MYST histone acetyltransferase (monocytic leukemia) 4|MYST-4|histone acetyltransferase MORF|histone acetyltransferase MOZ2|histone acetyltransferase MYST4|monocytic leukemia zinc finger protein-related factor', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74824927, end_position_on_the_genomic_accession=75032623, orientation='plus', exon_count=19, OMIM=605880, no_hdr0=''),
  23523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23523, CurrentID=0, Status='live', Symbol='CABIN1', Aliases='CAIN, KB-318B8.7, PPP3IN', description='calcineurin binding protein 1', other_designations='calcineurin binding protein cabin 1|calcineurin inhibitor', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24011192, end_position_on_the_genomic_accession=24178628, orientation='plus', exon_count=44, OMIM=604251, no_hdr0=''),
  285668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285668, CurrentID=0, Status='live', Symbol='C5orf64', Aliases='', description='chromosome 5 open reading frame 64', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61637781, end_position_on_the_genomic_accession=61730997, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23526, CurrentID=0, Status='live', Symbol='HMHA1', Aliases='ARHGAP45, HA-1, HLA-HA1', description='histocompatibility (minor) HA-1', other_designations='minor histocompatibility antigen HA-1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1065923, end_position_on_the_genomic_accession=1087831, orientation='plus', exon_count=26, OMIM=601155, no_hdr0=''),
  23527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23527, CurrentID=0, Status='live', Symbol='ACAP2', Aliases='CENTB2, CNT-B2', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 2', other_designations='Arf GAP with coiled coil, ANK repeat and PH domains 2|centaurin, beta 2|centaurin-beta-2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195274736, end_position_on_the_genomic_accession=195443176, orientation='minus', exon_count=28, OMIM=607766, no_hdr0=''),
  285672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285672, CurrentID=0, Status='live', Symbol='SREK1IP1', Aliases='P18SRP, SFRS12IP1', description='SREK1-interacting protein 1', other_designations='SFRS12-interacting protein 1|p18 splicing regulatory protein|protein SFRS12IP1|splicing regulatory protein of 18 kDa', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64718148, end_position_on_the_genomic_accession=64768731, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  23529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23529, CurrentID=0, Status='live', Symbol='CLCF1', Aliases='BSF-3, BSF3, CISS2, CLC, NNT-1, NNT1, NR6', description='cardiotrophin-like cytokine factor 1', other_designations='B-cell stimulating factor 3|B-cell stimulatory factor 3|B-cell-stimulating factor 3|CRLF1 associated cytokine-like factor 1|neurotrophin-1/B-cell stimulating factor-3|novel neurotrophin-1', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67364168, end_position_on_the_genomic_accession=67374177, orientation='minus', exon_count=4, OMIM=607672, no_hdr0=''),
  23530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23530, CurrentID=0, Status='live', Symbol='NNT', Aliases='GCCD4', description='nicotinamide nucleotide transhydrogenase', other_designations='pyridine nucleotide transhydrogenase', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43602689, end_position_on_the_genomic_accession=43705566, orientation='plus', exon_count=26, OMIM=607878, no_hdr0=''),
  23531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23531, CurrentID=0, Status='live', Symbol='MMD', Aliases='MMA1, PAQR11, MMD', description='monocyte to macrophage differentiation-associated', other_designations='macrophage maturation-associated|monocyte to macrophage differentiation protein|progestin and adipoQ receptor family member 11|progestin and adipoQ receptor family member XI', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55392613, end_position_on_the_genomic_accession=55421993, orientation='minus', exon_count=9, OMIM=604467, no_hdr0=''),
  23532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23532, CurrentID=0, Status='live', Symbol='PRAME', Aliases='CT130, MAPE, OIP-4, OIP4', description='preferentially expressed antigen in melanoma', other_designations='Opa-interacting protein OIP4|cancer/testis antigen 130|opa-interacting protein 4|preferentially expressed antigen of melanoma', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22547696, end_position_on_the_genomic_accession=22559361, orientation='minus', exon_count=7, OMIM=606021, no_hdr0=''),
  23533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23533, CurrentID=0, Status='live', Symbol='PIK3R5', Aliases='F730038I15Rik, FOAP-2, P101-PI3K, p101', description='phosphoinositide-3-kinase regulatory subunit 5', other_designations='PI3-kinase p101 subunit|phosphatidylinositol 4,5-bisphosphate 3-kinase regulatory subunit|phosphatidylinositol-4,5-bisphosphate 3-kinase regulatory subunit|phosphoinositide-3-kinase, regulatory subunit 5|protein FOAP-2|ptdIns-3-kinase p101', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8878916, end_position_on_the_genomic_accession=8965712, orientation='minus', exon_count=21, OMIM=611317, no_hdr0=''),
  23534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23534, CurrentID=0, Status='live', Symbol='TNPO3', Aliases='IPO12, LGMD1F, MTR10A, TRN-SR, TRN-SR2, TRNSR', description='transportin 3', other_designations='imp12|importin 12|limb girdle muscular dystrophy 1F (autosomal dominant)|transportin-SR', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128954180, end_position_on_the_genomic_accession=129055173, orientation='minus', exon_count=25, OMIM=610032, no_hdr0=''),
  285679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285679, CurrentID=0, Status='live', Symbol='C5orf60', Aliases='', description='chromosome 5 open reading frame 60', other_designations='putative uncharacterized protein C5orf60', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179641542, end_position_on_the_genomic_accession=179645167, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23536, CurrentID=0, Status='live', Symbol='ADAT1', Aliases='HADAT1', description='adenosine deaminase, tRNA-specific 1', other_designations='adenosine deaminase acting on tRNA|tRNA-specific adenosine-37 deaminase', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75598349, end_position_on_the_genomic_accession=75623323, orientation='minus', exon_count=11, OMIM=604230, no_hdr0=''),
  23538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23538, CurrentID=0, Status='live', Symbol='OR52A1', Aliases='HPFH1OR', description='olfactory receptor family 52 subfamily A member 1', other_designations="odorant receptor HOR3'beta4|olfactory receptor OR11-319|olfactory receptor, family 52, subfamily A, member 1", map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5151431, end_position_on_the_genomic_accession=5152369, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23539, CurrentID=0, Status='live', Symbol='SLC16A8', Aliases='MCT3, REMP', description='solute carrier family 16 (monocarboxylate transporter), member 8', other_designations='MCT 3|solute carrier 16 (monocarboxylic acid transporters), member 8|solute carrier family 16 member 8|solute carrier family 16, member 8 (monocarboxylic acid transporter 3)', map_location='22q12.3-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38078137, end_position_on_the_genomic_accession=38083163, orientation='minus', exon_count=5, OMIM=610409, no_hdr0=''),
  23540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23540, CurrentID=0, Status='live', Symbol='TRS-TGA4-1', Aliases='TRNAS3, TRS3', description='transfer RNA-Ser (TGA) 4-1', other_designations='tRNA serine (UGA) 3|tRNA serine 3 (anticodon UGA)|tRNA-Ser-TGA-4-1|transfer RNA serine 3 (anticodon UGA)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27505828, end_position_on_the_genomic_accession=27505909, orientation='minus', exon_count=0, OMIM=606172, no_hdr0=''),
  23541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23541, CurrentID=0, Status='live', Symbol='SEC14L2', Aliases='C22orf6, SPF, TAP, TAP1', description='SEC14-like lipid binding 2', other_designations='alpha-tocopherol-associated protein|squalene transfer protein|supernatant protein factor|tocopherol-associated protein', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30396941, end_position_on_the_genomic_accession=30425303, orientation='plus', exon_count=12, OMIM=607558, no_hdr0=''),
  23542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23542, CurrentID=0, Status='live', Symbol='MAPK8IP2', Aliases='IB-2, IB2, JIP2, PRKM8IPL', description='mitogen-activated protein kinase 8 interacting protein 2', other_designations='JNK MAP kinase scaffold protein 2|JNK MAP kinase scaffold protein JIP2|JNK-interacting protein 2|homologous to mouse JIP-1|islet-brain 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50600685, end_position_on_the_genomic_accession=50611551, orientation='plus', exon_count=12, OMIM=607755, no_hdr0=''),
  23543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23543, CurrentID=0, Status='live', Symbol='RBFOX2', Aliases='FOX2, Fox-2, HNRBP2, HRNBP2, RBM9, RTA, dJ106I20.3, fxh', description='RNA binding protein, fox-1 homolog (C. elegans) 2', other_designations='RNA-binding motif protein 9|fox-1 homolog B|fox-1 homologue|hexaribonucleotide-binding protein 2|repressor of tamoxifen transcriptional activity', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35738736, end_position_on_the_genomic_accession=36028892, orientation='minus', exon_count=21, OMIM=612149, no_hdr0=''),
  23544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23544, CurrentID=0, Status='live', Symbol='SEZ6L', Aliases='', description='seizure related 6 homolog (mouse)-like', other_designations='seizure related gene 6-like', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26169474, end_position_on_the_genomic_accession=26383597, orientation='plus', exon_count=17, OMIM=607021, no_hdr0=''),
  23545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23545, CurrentID=0, Status='live', Symbol='ATP6V0A2', Aliases='A2, ARCL, ARCL2A, ATP6A2, ATP6N1D, J6B7, RTF, STV1, TJ6, TJ6M, TJ6S, VPH1, WSS', description='ATPase, H+ transporting, lysosomal V0 subunit a2', other_designations='A2V-ATPase|lysosomal H(+)-transporting ATPase V0 subunit a2|regeneration and tolerance factor|v-ATPase 116 kDa|v-type proton ATPase 116 kDa subunit a|vacuolar proton translocating ATPase 116 kDa subunit a', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123712318, end_position_on_the_genomic_accession=123761755, orientation='plus', exon_count=21, OMIM=611716, no_hdr0=''),
  23546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23546, CurrentID=0, Status='live', Symbol='SYNGR4', Aliases='', description='synaptogyrin 4', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48364281, end_position_on_the_genomic_accession=48376377, orientation='plus', exon_count=7, OMIM=608373, no_hdr0=''),
  23547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23547, CurrentID=0, Status='live', Symbol='LILRA4', Aliases='CD85g, ILT7', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 4', other_designations='CD85 antigen-like family member G|immunoglobulin-like transcript 7|leukocyte immunoglobulin-like receptor, subfamily A (without TM domain), member 4', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54333185, end_position_on_the_genomic_accession=54339169, orientation='minus', exon_count=8, OMIM=607517, no_hdr0=''),
  23548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23548, CurrentID=0, Status='live', Symbol='TTC33', Aliases='OSRF', description='tetratricopeptide repeat domain 33', other_designations='TPR repeat protein 33|osmosis responsive factor|osmosis-responsive factor', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40711576, end_position_on_the_genomic_accession=40755970, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23549, CurrentID=0, Status='live', Symbol='DNPEP', Aliases='ASPEP, DAP', description='aspartyl aminopeptidase', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219373458, end_position_on_the_genomic_accession=219400017, orientation='minus', exon_count=17, OMIM=611367, no_hdr0=''),
  23550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23550, CurrentID=0, Status='live', Symbol='PSD4', Aliases='EFA6B, TIC', description='pleckstrin and Sec7 domain containing 4', other_designations='ADP-ribosylation factor guanine nucleotide-exchange factor 6|SEC7 homolog|exchange factor for ADP-ribosylation factor guanine nucleotide factor 6 B|exchange factor for ADP-ribosylation factor guanine nucleotide factor 6B|exchange factor for ARF6 B|pleckstrin homology and SEC7 domain-containing protein 4|telomeric of interleukin-1 cluster protein', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113157399, end_position_on_the_genomic_accession=113203229, orientation='plus', exon_count=18, OMIM=614442, no_hdr0=''),
  23551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23551, CurrentID=0, Status='live', Symbol='RASD2', Aliases='MGC:4834, Rhes, TEM2', description='RASD family member 2', other_designations='RASD family, member 2|Ras homolog enriched in striatum|tumor endothelial marker 2', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35540518, end_position_on_the_genomic_accession=35554002, orientation='plus', exon_count=6, OMIM=612842, no_hdr0=''),
  23552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23552, CurrentID=0, Status='live', Symbol='CDK20', Aliases='CCRK, CDCH, P42, PNQALRE', description='cyclin-dependent kinase 20', other_designations='CAK-kinase p42|CDK-activating kinase p42|cell cycle-related kinase|cell division protein kinase 20|cyclin-dependent protein kinase H|cyclin-kinase-activating kinase p42', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87966441, end_position_on_the_genomic_accession=87974780, orientation='minus', exon_count=9, OMIM=610076, no_hdr0=''),
  23553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23553, CurrentID=0, Status='live', Symbol='HYAL4', Aliases='CSHY', description='hyaluronoglucosaminidase 4', other_designations='chondroitin sulfate endo-beta-N-acetylgalactosaminidase|chondroitin sulfate hydrolase|hyal-4|hyaluronidase 4|hyaluronoglucosaminidase-4', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123845169, end_position_on_the_genomic_accession=123877481, orientation='plus', exon_count=8, OMIM=604510, no_hdr0=''),
  23554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23554, CurrentID=0, Status='live', Symbol='TSPAN12', Aliases='EVR5, NET-2, NET2, TM4SF12', description='tetraspanin 12', other_designations='tetraspan NET-2|transmembrane 4 superfamily member 12|tspan-12', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120787320, end_position_on_the_genomic_accession=120858366, orientation='minus', exon_count=11, OMIM=613138, no_hdr0=''),
  23555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23555, CurrentID=0, Status='live', Symbol='TSPAN15', Aliases='2700063A19Rik, NET-7, NET7, TM4SF15', description='tetraspanin 15', other_designations='tetraspan NET-7|transmembrane 4 superfamily member 15|transmembrane 4 superfamily member tetraspan NET-7|tspan-15', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69425218, end_position_on_the_genomic_accession=69507669, orientation='plus', exon_count=11, OMIM=613140, no_hdr0=''),
  23556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23556, CurrentID=0, Status='live', Symbol='PIGN', Aliases='MCAHS, MCAHS1, MCD4, MDC4, PIG-N', description='phosphatidylinositol glycan anchor biosynthesis class N', other_designations='MCD4 homolog|phosphatidylinositol glycan anchor biosynthesis, class N|phosphatidylinositol-glycan biosynthesis class N protein', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62044224, end_position_on_the_genomic_accession=62187118, orientation='minus', exon_count=33, OMIM=606097, no_hdr0=''),
  23557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23557, CurrentID=0, Status='live', Symbol='SNAPIN', Aliases='BLOC1S7, BLOS7, SNAPAP', description='SNAP-associated protein', other_designations='BLOC-1 subunit 7|SNAP-25-binding protein|SNARE associated protein snapin|biogenesis of lysosomal organelles complex-1, subunit 7|biogenesis of lysosome-related organelles complex 1 subunit 7|synaptosomal-associated protein 25-binding protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153658654, end_position_on_the_genomic_accession=153661852, orientation='plus', exon_count=4, OMIM=607007, no_hdr0=''),
  23558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23558, CurrentID=0, Status='live', Symbol='WBP2', Aliases='WBP-2', description='WW domain binding protein 2', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75845699, end_position_on_the_genomic_accession=75855420, orientation='minus', exon_count=9, OMIM=606962, no_hdr0=''),
  23559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23559, CurrentID=0, Status='live', Symbol='WBP1', Aliases='WBP-1', description='WW domain binding protein 1', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74458400, end_position_on_the_genomic_accession=74460891, orientation='plus', exon_count=4, OMIM=606961, no_hdr0=''),
  23560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23560, CurrentID=0, Status='live', Symbol='GTPBP4', Aliases='CRFG, NGB, NOG1', description='GTP binding protein 4', other_designations='G protein-binding protein CRFG|GTP-binding protein NGB|chronic renal failure gene protein', map_location='10p15-p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=988409, end_position_on_the_genomic_accession=1017768, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  23562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23562, CurrentID=0, Status='live', Symbol='CLDN14', Aliases='DFNB29', description='claudin 14', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36460621, end_position_on_the_genomic_accession=36580085, orientation='minus', exon_count=9, OMIM=605608, no_hdr0=''),
  23563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23563, CurrentID=0, Status='live', Symbol='CHST5', Aliases='I-GlcNAc-6-ST, I-GlcNAc6ST, glcNAc6ST-3, gn6st-3, hIGn6ST', description='carbohydrate (N-acetylglucosamine 6-O) sulfotransferase 5', other_designations='GST4-alpha|N-acetylglucosamine 6-O-sulfotransferase 3|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 4-alpha|intestinal GlcNAc-6-sulfotransferase', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75528530, end_position_on_the_genomic_accession=75535170, orientation='minus', exon_count=3, OMIM=604817, no_hdr0=''),
  285708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285708, CurrentID=0, Status='live', Symbol='LINC00491', Aliases='', description='long intergenic non-protein coding RNA 491', other_designations='CTD-2340E1.3', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102608492, end_position_on_the_genomic_accession=102671464, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  23566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23566, CurrentID=0, Status='live', Symbol='LPAR3', Aliases='EDG7, Edg-7, GPCR, HOFNH30, LP-A3, LPA3, RP4-678I3', description='lysophosphatidic acid receptor 3', other_designations='LPA receptor 3|LPA receptor EDG7|LPA-3|calcium-mobilizing lysophosphatidic acid receptor LP-A3|endothelial cell differentiation gene 7|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor, 7|lysophosphatidic acid receptor Edg-7', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84811602, end_position_on_the_genomic_accession=84893213, orientation='minus', exon_count=4, OMIM=605106, no_hdr0=''),
  23567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23567, CurrentID=0, Status='live', Symbol='ZNF346', Aliases='JAZ, Zfp346', description='zinc finger protein 346', other_designations='double-stranded RNA-binding zinc finger protein JAZ|just another zinc finger protein', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177022680, end_position_on_the_genomic_accession=177081188, orientation='plus', exon_count=10, OMIM=605308, no_hdr0=''),
  23568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23568, CurrentID=0, Status='live', Symbol='ARL2BP', Aliases='BART, BART1, RP66', description='ADP ribosylation factor like GTPase 2 binding protein', other_designations='ADP-ribosylation factor like 2 binding protein|ADP-ribosylation factor like GTPase 2 binding protein|ADP-ribosylation factor-like 2 binding protein|ARF-like 2-binding protein|ARL2-binding protein|Arf-like 2 binding protein BART1|binder of ARF2 protein 1|binder of Arl Two|binder of Arl2|retinitis pigmentosa 66 (autosomal recessive)', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57245126, end_position_on_the_genomic_accession=57253635, orientation='plus', exon_count=7, OMIM=615407, no_hdr0=''),
  23569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23569, CurrentID=0, Status='live', Symbol='PADI4', Aliases='PAD, PAD4, PADI5, PDI4, PDI5', description='peptidyl arginine deiminase, type IV', other_designations='HL-60 PAD|PADI-H protein|peptidyl arginine deiminase, type V|protein-arginine deiminase type IV', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17308195, end_position_on_the_genomic_accession=17364004, orientation='plus', exon_count=18, OMIM=605347, no_hdr0=''),
  23574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23574, CurrentID=0, Status='live', Symbol='PRG1', Aliases='', description='p53-responsive gene 1', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43349056, end_position_on_the_genomic_accession=43349548, orientation='plus', exon_count=1, OMIM=605157, no_hdr0=''),
  23576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23576, CurrentID=0, Status='live', Symbol='DDAH1', Aliases='DDAH, HEL-S-16', description='dimethylarginine dimethylaminohydrolase 1', other_designations='DDAH-1|DDAHI|NG, NG-dimethylarginine dimethylaminohydrolase|dimethylargininase-1|epididymis secretory protein Li 16', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85318485, end_position_on_the_genomic_accession=85578363, orientation='minus', exon_count=11, OMIM=604743, no_hdr0=''),
  23580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23580, CurrentID=0, Status='live', Symbol='CDC42EP4', Aliases='BORG4, CEP4, KAIA1777', description='CDC42 effector protein (Rho GTPase binding) 4', other_designations='binder of Rho GTPases 4', map_location='17q24-q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73283624, end_position_on_the_genomic_accession=73312004, orientation='minus', exon_count=3, OMIM=605468, no_hdr0=''),
  23581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23581, CurrentID=0, Status='live', Symbol='CASP14', Aliases='', description='caspase 14', other_designations='CASP-14|apoptosis-related cysteine protease|caspase 14, apoptosis-related cysteine peptidase|caspase 14, apoptosis-related cysteine protease', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15049480, end_position_on_the_genomic_accession=15058293, orientation='plus', exon_count=7, OMIM=605848, no_hdr0=''),
  23582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23582, CurrentID=0, Status='live', Symbol='CCNDBP1', Aliases='DIP1, GCIP, HHM', description='cyclin D-type binding-protein 1', other_designations='D-type cyclin-interacting protein 1|grap2 and cyclin-D-interacting protein|human homolog of Maid', map_location='15q14-q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43185118, end_position_on_the_genomic_accession=43197177, orientation='plus', exon_count=12, OMIM=607089, no_hdr0=''),
  23583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23583, CurrentID=0, Status='live', Symbol='SMUG1', Aliases='FDG, HMUDG, UNG3', description='single-strand-selective monofunctional uracil-DNA glycosylase 1', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54172031, end_position_on_the_genomic_accession=54188994, orientation='minus', exon_count=8, OMIM=607753, no_hdr0=''),
  23584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23584, CurrentID=0, Status='live', Symbol='VSIG2', Aliases='2210413P10Rik, CTH, CTXL', description='V-set and immunoglobulin domain containing 2', other_designations='CT-like protein|cortical thymocyte receptor (X. laevis CTX) like|cortical thymocyte-like protein', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124747472, end_position_on_the_genomic_accession=124753025, orientation='minus', exon_count=8, OMIM=606011, no_hdr0=''),
  23585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23585, CurrentID=0, Status='live', Symbol='TMEM50A', Aliases='IFNRC, SMP1', description='transmembrane protein 50A', other_designations='cervical cancer oncogene 9|small membrane protein 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25338298, end_position_on_the_genomic_accession=25362361, orientation='plus', exon_count=8, OMIM=605348, no_hdr0=''),
  23586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23586, CurrentID=0, Status='live', Symbol='DDX58', Aliases='RIG-I, RIGI, RLR-1, SGMRT2', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 58', other_designations='DEAD box protein 58|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide|RNA helicase RIG-I|retinoic acid-inducible gene 1 protein|retinoic acid-inducible gene I protein', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32455302, end_position_on_the_genomic_accession=32526324, orientation='minus', exon_count=18, OMIM=609631, no_hdr0=''),
  23587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23587, CurrentID=0, Status='live', Symbol='ELP5', Aliases='C17orf81, DERP6, HSPC002, MST071, MSTP071', description='elongator acetyltransferase complex subunit 5', other_designations='S-phase 2 protein|dermal papilla derived protein 6', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7252053, end_position_on_the_genomic_accession=7259940, orientation='plus', exon_count=9, OMIM=615019, no_hdr0=''),
  23588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23588, CurrentID=0, Status='live', Symbol='KLHDC2', Aliases='HCLP-1, HCLP1, LCP', description='kelch domain containing 2', other_designations='hepatocellular carcinoma-associated antigen 33|host cell factor homolog LCP|host cell factor-like protein 1', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49767589, end_position_on_the_genomic_accession=49785660, orientation='plus', exon_count=15, OMIM=611280, no_hdr0=''),
  154661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154661, CurrentID=0, Status='live', Symbol='RUNDC3B', Aliases='RPIB9, RPIP9', description='RUN domain containing 3B', other_designations='RPIP-9|Rap2 binding protein 9|Rap2-binding protein 9|Rap2-interacting protein 9', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87627675, end_position_on_the_genomic_accession=87832298, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23590, CurrentID=0, Status='live', Symbol='PDSS1', Aliases='COQ1, COQ10D2, DPS, SPS, TPRT, TPT, TPT 1, hDPS1', description='prenyl (decaprenyl) diphosphate synthase, subunit 1', other_designations='all-trans-decaprenyl-diphosphate synthase subunit 1|coenzyme Q1 homolog|decaprenyl pyrophosphate synthase subunit 1|polyprenyl pyrophosphate synthetase|subunit 1 of decaprenyl diphosphate synthase|trans-prenyltransferase (TPT)|trans-prenyltransferase 1', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26697424, end_position_on_the_genomic_accession=26748564, orientation='plus', exon_count=14, OMIM=607429, no_hdr0=''),
  23591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23591, CurrentID=0, Status='live', Symbol='FAM215A', Aliases='APR-2, C17orf88, LINC00530', description='family with sequence similarity 215 member A (non-protein coding)', other_designations='apoptosis related protein|family with sequence similarity 215, member A (non-protein coding)|long intergenic non-protein coding RNA 530', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43917208, end_position_on_the_genomic_accession=43917987, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23592, CurrentID=0, Status='live', Symbol='LEMD3', Aliases='MAN1', description='LEM domain containing 3', other_designations='LEM domain-containing protein 3|integral inner nuclear membrane protein', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65169571, end_position_on_the_genomic_accession=65248361, orientation='plus', exon_count=13, OMIM=607844, no_hdr0=''),
  23593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23593, CurrentID=0, Status='live', Symbol='HEBP2', Aliases='C6ORF34B, C6orf34, PP23, SOUL', description='heme binding protein 2', other_designations='placental protein 23', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138404199, end_position_on_the_genomic_accession=138421545, orientation='plus', exon_count=5, OMIM=605825, no_hdr0=''),
  23594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23594, CurrentID=0, Status='live', Symbol='ORC6', Aliases='ORC6L', description='origin recognition complex subunit 6', other_designations='', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46689646, end_position_on_the_genomic_accession=46698394, orientation='plus', exon_count=8, OMIM=607213, no_hdr0=''),
  23595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23595, CurrentID=0, Status='live', Symbol='ORC3', Aliases='LAT, LATHEOL, ORC3', description='origin recognition complex subunit 3', other_designations='homolog of latheo, Drosophila|origin recognition complex subunit Latheo|origin recognition complex, subunit 3 honolog', map_location='6q14.3-q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87590067, end_position_on_the_genomic_accession=87667454, orientation='plus', exon_count=23, OMIM=604972, no_hdr0=''),
  23596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23596, CurrentID=0, Status='live', Symbol='OPN3', Aliases='ECPN, PPP1R116', description='opsin 3', other_designations='encephalopsin|opsin 3 (encephalopsin, panopsin)|protein phosphatase 1, regulatory subunit 116', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241593150, end_position_on_the_genomic_accession=241640399, orientation='minus', exon_count=4, OMIM=606695, no_hdr0=''),
  23597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23597, CurrentID=0, Status='live', Symbol='ACOT9', Aliases='ACATE2, CGI-16, MT-ACT48, MTACT48', description='acyl-CoA thioesterase 9', other_designations='acyl-CoA thioester hydrolase 9|acyl-Coenzyme A thioesterase 2, mitochondrial|mitochondrial Acyl-CoA Thioesterase', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23703660, end_position_on_the_genomic_accession=23743290, orientation='minus', exon_count=16, OMIM=300862, no_hdr0=''),
  23598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23598, CurrentID=0, Status='live', Symbol='PATZ1', Aliases='MAZR, PATZ, RIAZ, ZBTB19, ZNF278, ZSG, dJ400N23', description='POZ (BTB) and AT hook containing zinc finger 1', other_designations='BTB-POZ domain zinc finger transcription factor|BTB/POZ domain zinc finger transcription factor|MAZ-related factor|POZ-AT hook-zinc finger protein|protein kinase A RI subunit alpha-associated protein|zinc finger and BTB domain-containing protein 19|zinc finger protein 278|zinc finger sarcoma gene protein', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31325804, end_position_on_the_genomic_accession=31346263, orientation='minus', exon_count=6, OMIM=605165, no_hdr0=''),
  101060445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060445, CurrentID=0, Status='live', Symbol='LOC101060445', Aliases='', description='uncharacterized LOC101060445', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32459521, end_position_on_the_genomic_accession=32474910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23600, CurrentID=0, Status='live', Symbol='AMACR', Aliases='AMACRD, CBAS4, RACE, RM', description='alpha-methylacyl-CoA racemase', other_designations='2-methylacyl-CoA racemase', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33986986, end_position_on_the_genomic_accession=34008115, orientation='minus', exon_count=5, OMIM=604489, no_hdr0=''),
  23601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23601, CurrentID=0, Status='live', Symbol='CLEC5A', Aliases='CLECSF5, MDL-1, MDL1', description='C-type lectin domain family 5 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 5|C-type lectin domain family 5, member A|C-type lectin superfamily member 5|myeloid DAP12-associating lectin 1|myeloid DAP12-associating lectin-1', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141927357, end_position_on_the_genomic_accession=141947007, orientation='minus', exon_count=8, OMIM=604987, no_hdr0=''),
  23603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23603, CurrentID=0, Status='live', Symbol='CORO1C', Aliases='HCRNN4', description='coronin, actin binding protein, 1C', other_designations='coronin 1C|coronin, actin-binding protein, 1C|coronin-3', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108645109, end_position_on_the_genomic_accession=108731550, orientation='minus', exon_count=16, OMIM=605269, no_hdr0=''),
  23604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23604, CurrentID=0, Status='live', Symbol='DAPK2', Aliases='DRP-1, DRP1', description='death-associated protein kinase 2', other_designations='DAP kinase 2|DAP-kinase-related protein 1 beta isoform', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63907036, end_position_on_the_genomic_accession=64046455, orientation='minus', exon_count=18, OMIM=616567, no_hdr0=''),
  23605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23605, CurrentID=0, Status='live', Symbol='HMGN2P9', Aliases='HMG17L1, HMGN2L9, dJ388M5, dJ388M5.2', description='high mobility group nucleosomal binding domain 2 pseudogene 9', other_designations='high-mobility group (nonhistone chromosomal) protein 17-like 1|high-mobility group nucleosomal binding domain 2 pseudogene 9|high-mobility group nucleosomal binding domain 2-like 9', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43802109, end_position_on_the_genomic_accession=43803568, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23606, CurrentID=0, Status='live', Symbol='HMGN2P10', Aliases='HMG17L2, HMGN2L10, bK216E10', description='high mobility group nucleosomal binding domain 2 pseudogene10', other_designations='high-mobility group (nonhistone chromosomal) protein 17-like 2|high-mobility group nucleosomal binding domain 2 pseudogene10|high-mobility group nucleosomal binding domain 2-like 10', map_location='22q13.2-q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41709223, end_position_on_the_genomic_accession=41709487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23607, CurrentID=0, Status='live', Symbol='CD2AP', Aliases='CMS', description='CD2-associated protein', other_designations='Cas ligand with multiple Src homology 3 (SH3) domains|adapter protein CMS|cas ligand with multiple SH3 domains', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47477746, end_position_on_the_genomic_accession=47627263, orientation='plus', exon_count=19, OMIM=604241, no_hdr0=''),
  23608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23608, CurrentID=0, Status='live', Symbol='MKRN1', Aliases='RNF61', description='makorin ring finger protein 1', other_designations='RING finger protein 61', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140453040, end_position_on_the_genomic_accession=140479569, orientation='minus', exon_count=11, OMIM=607754, no_hdr0=''),
  285753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285753, CurrentID=0, Status='live', Symbol='CEP57L1', Aliases='C6orf182, bA487F23.2, cep57R', description='centrosomal protein 57kDa-like 1', other_designations='centrosomal protein 57kDa-like protein 1|centrosomal protein of 57 kDa-related protein|cep57-related protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109095153, end_position_on_the_genomic_accession=109163912, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  23610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23610, CurrentID=0, Status='live', Symbol='MKRN6P', Aliases='MKRNP2', description='makorin ring finger protein 6, pseudogene', other_designations='makorin ring finger protein pseudogene 2', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35446210, end_position_on_the_genomic_accession=35449865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285755, CurrentID=0, Status='live', Symbol='PPIL6', Aliases='PPIase, RSPH12, bA425D10.6, dJ919F19.1', description='peptidylprolyl isomerase like 6', other_designations='cyclophilin-like protein PPIL6|peptidyl-prolyl cis-trans isomerase-like variant a|peptidylprolyl isomerase (cyclophilin)-like 6|radial spoke 12 homolog|rotamase PPIL6', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109390215, end_position_on_the_genomic_accession=109441171, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  23612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23612, CurrentID=0, Status='live', Symbol='PHLDA3', Aliases='TIH1', description='pleckstrin homology-like domain, family A, member 3', other_designations='TDAG51/Ipl homolog 1|pleckstrin homology-like domain, family A, member 2', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201465479, end_position_on_the_genomic_accession=201469171, orientation='minus', exon_count=3, OMIM=607054, no_hdr0=''),
  23613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23613, CurrentID=0, Status='live', Symbol='ZMYND8', Aliases='PRKCBP1, PRO2893, RACK7', description='zinc finger, MYND-type containing 8', other_designations='CTCL tumor antigen se14-3|cutaneous T-cell lymphoma-associated antigen se14-3|predicted protein of HQ2893|zinc finger MYND domain-containing protein 8', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47209214, end_position_on_the_genomic_accession=47356889, orientation='minus', exon_count=32, OMIM=615713, no_hdr0=''),
  285758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285758, CurrentID=0, Status='live', Symbol='LINC01268', Aliases='', description='long intergenic non-protein coding RNA 1268', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113868013, end_position_on_the_genomic_accession=113873347, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23615, CurrentID=0, Status='live', Symbol='PYY2', Aliases='', description='peptide YY, 2 (pseudogene)', other_designations='peptide YY, 2 (seminalplasmin)|seminalplasmin', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28226563, end_position_on_the_genomic_accession=28228060, orientation='plus', exon_count=2, OMIM=606637, no_hdr0=''),
  23616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23616, CurrentID=0, Status='live', Symbol='SH3BP1', Aliases='ARHGAP43', description='SH3-domain binding protein 1', other_designations='3BP-1|SH3 domain binding protein 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37639677, end_position_on_the_genomic_accession=37656043, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  23617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23617, CurrentID=0, Status='live', Symbol='TSSK2', Aliases='DGS-G, SPOGA2, STK22B, TSK2', description='testis-specific serine kinase 2', other_designations='TSK-2|TSSK-2|diGeorge syndrome protein G|serine/threonine kinase 22B (spermiogenesis associated)|serine/threonine-protein kinase 22B|spermiogenesis associated 2|testis specific serine threonine kinase 2|testis-specific kinase 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19130808, end_position_on_the_genomic_accession=19132623, orientation='plus', exon_count=1, OMIM=610710, no_hdr0=''),
  285762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285762, CurrentID=0, Status='live', Symbol='LOC285762', Aliases='', description='uncharacterized LOC285762', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119452547, end_position_on_the_genomic_accession=119491302, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23619, CurrentID=0, Status='live', Symbol='ZIM2', Aliases='ZNF656', description='zinc finger, imprinted 2', other_designations='zinc finger protein 656', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56774547, end_position_on_the_genomic_accession=56840729, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  23620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23620, CurrentID=0, Status='live', Symbol='NTSR2', Aliases='NTR2', description='neurotensin receptor 2', other_designations='NT-R-2|levocabastine-sensitive neurotensin receptor|neurotensin receptor, type 2', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11658178, end_position_on_the_genomic_accession=11670203, orientation='minus', exon_count=4, OMIM=605538, no_hdr0=''),
  23621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23621, CurrentID=0, Status='live', Symbol='BACE1', Aliases='ASP2, BACE, HSPC104', description='beta-site APP-cleaving enzyme 1', other_designations='APP beta-secretase|asp 2|aspartyl protease 2|beta-secretase 1 precursor variant 1|beta-site APP cleaving enzyme 1|beta-site amyloid beta A4 precursor protein-cleaving enzyme|memapsin-2|membrane-associated aspartic protease 2|transmembrane aspartic proteinase Asp2', map_location='11q23.2-q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117285686, end_position_on_the_genomic_accession=117316256, orientation='minus', exon_count=10, OMIM=604252, no_hdr0=''),
  285766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285766, CurrentID=0, Status='live', Symbol='LOC285766', Aliases='', description='uncharacterized LOC285766', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=181466, end_position_on_the_genomic_accession=205484, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23623, CurrentID=0, Status='live', Symbol='RUSC1', Aliases='NESCA', description='RUN and SH3 domain containing 1', other_designations='new molecule containing SH3 at the carboxy-terminus', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155320849, end_position_on_the_genomic_accession=155331118, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23624, CurrentID=0, Status='live', Symbol='CBLC', Aliases='CBL-3, CBL-SL, RNF57', description='Cbl proto-oncogene C, E3 ubiquitin protein ligase', other_designations='Cas-Br-M (murine) ecotropic retroviral transforming sequence c|Cas-Br-M (murine) ectropic retroviral transforming sequence c|Cbl proto-oncogene, E3 ubiquitin protein ligase C|RING finger protein 57|SH3-binding protein CBL-3|SH3-binding protein CBL-C|signal transduction protein CBL-C', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44777861, end_position_on_the_genomic_accession=44800646, orientation='plus', exon_count=13, OMIM=608453, no_hdr0=''),
  23625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23625, CurrentID=0, Status='live', Symbol='FAM89B', Aliases='LRAP25, MTVR, MTVR1', description='family with sequence similarity 89 member B', other_designations='family with sequence similarity 89, member B|leucine repeat adaptor protein 25|mammary tumor virus receptor homolog 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65572349, end_position_on_the_genomic_accession=65574198, orientation='plus', exon_count=2, OMIM=616128, no_hdr0=''),
  23626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23626, CurrentID=0, Status='live', Symbol='SPO11', Aliases='CT35, SPATA43, TOPVIA', description='SPO11 meiotic protein covalently bound to DSB', other_designations='SPO11 meiotic protein covalently bound to DSB homolog|cancer/testis antigen 35|spermatogenesis associated 43', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57329759, end_position_on_the_genomic_accession=57343993, orientation='plus', exon_count=14, OMIM=605114, no_hdr0=''),
  23627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23627, CurrentID=0, Status='live', Symbol='PRND', Aliases='DOPPEL, DPL, PrPLP, dJ1068H6.4', description='prion protein 2 (dublet)', other_designations='prion gene complex, downstream', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4721854, end_position_on_the_genomic_accession=4728462, orientation='plus', exon_count=2, OMIM=604263, no_hdr0=''),
  105373400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373400, CurrentID=0, Status='live', Symbol='LOC105373400', Aliases='', description='uncharacterized LOC105373400', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5811127, end_position_on_the_genomic_accession=5812618, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23629, CurrentID=0, Status='live', Symbol='BRD7P3', Aliases='BP75', description='bromodomain containing 7 pseudogene 3', other_designations='bromodomain containing protein 75 kDa pseudogene', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118501373, end_position_on_the_genomic_accession=118503846, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23630, CurrentID=0, Status='live', Symbol='KCNE5', Aliases='KCNE1L', description='potassium channel, voltage gated subfamily E regulatory beta subunit 5', other_designations='AMME syndrome candidate gene 2 protein|AMMECR2 protein|KCNE1-like|cardiac voltage-gated potassium channel accessory subunit 5|potassium channel subunit beta MiRP4|potassium voltage-gated channel subfamily E member 1-like protein|potassium voltage-gated channel, Isk-related family, member 1-like', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109623700, end_position_on_the_genomic_accession=109625164, orientation='minus', exon_count=1, OMIM=300328, no_hdr0=''),
  23632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23632, CurrentID=0, Status='live', Symbol='CA14', Aliases='CAXiV', description='carbonic anhydrase XIV', other_designations='CA-XIV|carbonate dehydratase XIV|carbonic dehydratase', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150257159, end_position_on_the_genomic_accession=150265078, orientation='plus', exon_count=11, OMIM=604832, no_hdr0=''),
  23633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23633, CurrentID=0, Status='live', Symbol='KPNA6', Aliases='IPOA7, KPNA7', description='karyopherin alpha 6 (importin alpha 7)', other_designations='importin alpha 7 subunit|importin-alpha-S2|karyopherin subunit alpha-6', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32108043, end_position_on_the_genomic_accession=32176568, orientation='plus', exon_count=16, OMIM=610563, no_hdr0=''),
  23635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23635, CurrentID=0, Status='live', Symbol='SSBP2', Aliases='HSPC116, SOSS-B2', description='single-stranded DNA binding protein 2', other_designations='sequence-specific single-stranded-DNA-binding protein 2', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81417360, end_position_on_the_genomic_accession=81751253, orientation='minus', exon_count=17, OMIM=607389, no_hdr0=''),
  285780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285780, CurrentID=0, Status='live', Symbol='LY86-AS1', Aliases='LY86-AS, LY86AS', description='LY86 antisense RNA 1', other_designations='LY86 antisense RNA (non-protein coding)|LY86 antisense RNA 1 (non-protein coding)', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6346465, end_position_on_the_genomic_accession=6622826, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23637, CurrentID=0, Status='live', Symbol='RABGAP1', Aliases='GAPCENA, TBC1D11', description='RAB GTPase activating protein 1', other_designations='GAP and centrosome-associated protein|Rab6 GTPase activating protein, GAPCenA|TBC1 domain family, member 11|rab6 GTPase activating protein (GAP and centrosome-associated)|rab6 GTPase-activating protein GAPCenA', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122941009, end_position_on_the_genomic_accession=123104868, orientation='plus', exon_count=31, OMIM=615882, no_hdr0=''),
  285782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285782, CurrentID=0, Status='live', Symbol='CAGE1', Aliases='CT3, CT95, CTAG3, bA69L16.7', description='cancer antigen 1', other_designations='cancer/testis antigen 3|cancer/testis antigen 95|cancer/testis antigen gene 1', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7326654, end_position_on_the_genomic_accession=7389709, orientation='minus', exon_count=15, OMIM=608304, no_hdr0=''),
  23639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23639, CurrentID=0, Status='live', Symbol='LRRC6', Aliases='CILD19, LRTP, TSLRP', description='leucine rich repeat containing 6', other_designations='protein TILB homolog|testis-specific leucine-rich repeat protein', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132570416, end_position_on_the_genomic_accession=132685097, orientation='minus', exon_count=17, OMIM=614930, no_hdr0=''),
  23640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23640, CurrentID=0, Status='live', Symbol='HSPBP1', Aliases='FES1', description='HSPA (heat shock 70kDa) binding protein, cytoplasmic cochaperone 1', other_designations='heat shock protein-binding protein 1|hsp70 interacting protein', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55262222, end_position_on_the_genomic_accession=55280383, orientation='minus', exon_count=9, OMIM=612939, no_hdr0=''),
  23641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23641, CurrentID=0, Status='live', Symbol='LDOC1', Aliases='BCUR1, Mar7, Mart7', description='leucine zipper, down-regulated in cancer 1', other_designations='breast cancer, up-regulated 1|leucine zipper downregulated in cancer|leucine zipper protein down-regulated in cancer cells', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141175745, end_position_on_the_genomic_accession=141177214, orientation='minus', exon_count=1, OMIM=300402, no_hdr0=''),
  23642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23642, CurrentID=0, Status='live', Symbol='SNHG1', Aliases='LINC00057, NCRNA00057, U22HG, UHG', description='small nucleolar RNA host gene 1', other_designations='U22 snoRNA host|long intergenic non-protein coding RNA 57|small nucleolar RNA host gene 1 (non-protein coding)', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62851988, end_position_on_the_genomic_accession=62855888, orientation='minus', exon_count=11, OMIM=603222, no_hdr0=''),
  23643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23643, CurrentID=0, Status='live', Symbol='LY96', Aliases='ESOP-1, MD-2, MD2, ly-96', description='lymphocyte antigen 96', other_designations='myeloid differentiation protein-2|protein MD-2', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73991329, end_position_on_the_genomic_accession=74032562, orientation='plus', exon_count=6, OMIM=605243, no_hdr0=''),
  23644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23644, CurrentID=0, Status='live', Symbol='EDC4', Aliases='GE1, Ge-1, HEDL5, HEDLS, RCD-8, RCD8', description='enhancer of mRNA decapping 4', other_designations='autoantigen Ge-1|autoantigen RCD-8|human enhancer of decapping large subunit', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67873023, end_position_on_the_genomic_accession=67884514, orientation='plus', exon_count=29, OMIM=606030, no_hdr0=''),
  23645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23645, CurrentID=0, Status='live', Symbol='PPP1R15A', Aliases='GADD34', description='protein phosphatase 1 regulatory subunit 15A', other_designations='growth arrest and DNA damage-inducible protein GADD34|growth arrest and DNA-damage-inducible 34|myeloid differentiation primary response protein MyD116 homolog|protein phosphatase 1, regulatory (inhibitor) subunit 15A|protein phosphatase 1, regulatory subunit 15A', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48872392, end_position_on_the_genomic_accession=48876062, orientation='plus', exon_count=3, OMIM=611048, no_hdr0=''),
  23646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23646, CurrentID=0, Status='live', Symbol='PLD3', Aliases='AD19, HU-K4, HUK4', description='phospholipase D family member 3', other_designations='choline phosphatase 3|hindIII K4L homolog|phosphatidylcholine-hydrolyzing phospholipase D3|phospholipase D family, member 3', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40348425, end_position_on_the_genomic_accession=40378490, orientation='plus', exon_count=15, OMIM=615698, no_hdr0=''),
  23647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23647, CurrentID=0, Status='live', Symbol='ARFIP2', Aliases='POR1', description='ADP ribosylation factor interacting protein 2', other_designations='partner of RAC1 (arfaptin 2)', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6474683, end_position_on_the_genomic_accession=6481479, orientation='minus', exon_count=9, OMIM=601638, no_hdr0=''),
  23648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23648, CurrentID=0, Status='live', Symbol='SSBP3', Aliases='CSDP, SSDP, SSDP1', description='single stranded DNA binding protein 3', other_designations='sequence-specific single-stranded-DNA-binding protein', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54225431, end_position_on_the_genomic_accession=54413485, orientation='minus', exon_count=20, OMIM=607390, no_hdr0=''),
  23649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23649, CurrentID=0, Status='live', Symbol='POLA2', Aliases='', description='polymerase (DNA directed), alpha 2, accessory subunit', other_designations='DNA polymerase alpha 70 kDa subunit|polymerase (DNA directed), alpha 2 (70kD subunit)|polymerase (DNA-directed), alpha (70kD)', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65261852, end_position_on_the_genomic_accession=65305598, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  23650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23650, CurrentID=0, Status='live', Symbol='TRIM29', Aliases='ATDC', description='tripartite motif containing 29', other_designations='ataxia telangiectasia group D-associated protein|ataxia-telangiectasia group D-associated protein|tripartite motif protein TRIM29|tripartite motif-containing 29', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120111286, end_position_on_the_genomic_accession=120185423, orientation='minus', exon_count=13, OMIM=610658, no_hdr0=''),
  285796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285796, CurrentID=0, Status='live', Symbol='PACRG-AS1', Aliases='', description='PACRG antisense RNA 1', other_designations='PACRG antisense RNA 1 (non-protein coding)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163309985, end_position_on_the_genomic_accession=163324473, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23654, CurrentID=0, Status='live', Symbol='PLXNB2', Aliases='MM1, Nbla00445, PLEXB2, dJ402G11.3', description='plexin B2', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50274979, end_position_on_the_genomic_accession=50307642, orientation='minus', exon_count=40, OMIM=604293, no_hdr0=''),
  285800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285800, CurrentID=0, Status='live', Symbol='PRR18', Aliases='', description='proline rich 18', other_designations='proline rich region 18', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166305680, end_position_on_the_genomic_accession=166308383, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23657, CurrentID=0, Status='live', Symbol='SLC7A11', Aliases='CCBR1, xCT', description='solute carrier family 7 (anionic amino acid transporter light chain, xc- system), member 11', other_designations='amino acid transport system xc-|calcium channel blocker resistance protein CCBR1|solute carrier family 7 member 11|solute carrier family 7, (cationic amino acid transporter, y+ system) member 11', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138164094, end_position_on_the_genomic_accession=138312660, orientation='minus', exon_count=22, OMIM=607933, no_hdr0=''),
  23658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23658, CurrentID=0, Status='live', Symbol='LSM5', Aliases='YER146W', description='LSM5 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM5 U6 small nuclear RNA and mRNA degradation associated|LSM5 homolog, U6 small nuclear RNA associated', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32485333, end_position_on_the_genomic_accession=32495258, orientation='minus', exon_count=6, OMIM=607285, no_hdr0=''),
  23659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23659, CurrentID=0, Status='live', Symbol='PLA2G15', Aliases='ACS, GXVPLA2, LLPL, LPLA2, LYPLA3', description='phospholipase A2 group XV', other_designations='1-O-acylceramide synthase|LCAT-like lysophospholipase|lysophospholipase 3 (lysosomal phospholipase A2)|lysosomal phospholipase A2|phospholipase A2, group XV', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68245343, end_position_on_the_genomic_accession=68261062, orientation='plus', exon_count=7, OMIM=609362, no_hdr0=''),
  23660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23660, CurrentID=0, Status='live', Symbol='ZKSCAN5', Aliases='ZFP-95, ZFP95, ZNF914, ZSCAN37', description='zinc finger with KRAB and SCAN domains 5', other_designations='zinc finger protein 95 homolog|zinc finger protein homologous to Zfp95 in mouse', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99501297, end_position_on_the_genomic_accession=99534107, orientation='plus', exon_count=11, OMIM=611272, no_hdr0=''),
  23662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23662, CurrentID=0, Status='live', Symbol='USH3B', Aliases='', description='Usher syndrome 3B', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  153546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153546, CurrentID=0, Status='live', Symbol='LOC153546', Aliases='', description='uncharacterized LOC153546', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271277, CurrentID=0, Status='live', Symbol='RPL13AP12', Aliases='RPL13A_3_312', description='ribosomal protein L13a pseudogene 12', other_designations='', map_location='2q33-q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202520393, end_position_on_the_genomic_accession=202520992, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23666, CurrentID=0, Status='live', Symbol='UBBP4', Aliases='', description='ubiquitin B pseudogene 4', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22203650, end_position_on_the_genomic_accession=22205156, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23667, CurrentID=0, Status='live', Symbol='UBBP3', Aliases='', description='ubiquitin B pseudogene 3', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148863426, end_position_on_the_genomic_accession=148863801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23668, CurrentID=0, Status='live', Symbol='UBBP2', Aliases='', description='ubiquitin B pseudogene 2', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217850259, end_position_on_the_genomic_accession=217850683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23670, CurrentID=0, Status='live', Symbol='TMEM2', Aliases='', description='transmembrane protein 2', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=71683366, end_position_on_the_genomic_accession=71769466, orientation='minus', exon_count=25, OMIM=605835, no_hdr0=''),
  23671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23671, CurrentID=0, Status='live', Symbol='TMEFF2', Aliases='CT120.2, HPP1, TENB2, TPEF, TR, TR-2', description='transmembrane protein with EGF-like and two follistatin-like domains 2', other_designations='cancer/testis antigen family 120, member 2|hyperplastic polyposis protein 1|tomoregulin|transmembrane protein TENB2', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191949046, end_position_on_the_genomic_accession=192194940, orientation='minus', exon_count=10, OMIM=605734, no_hdr0=''),
  23673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23673, CurrentID=0, Status='live', Symbol='STX12', Aliases='STX13, STX14', description='syntaxin 12', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27773183, end_position_on_the_genomic_accession=27824452, orientation='plus', exon_count=9, OMIM=606892, no_hdr0=''),
  285819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285819, CurrentID=0, Status='live', Symbol='LOC285819', Aliases='', description='uncharacterized LOC285819', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26471944, end_position_on_the_genomic_accession=26482509, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23676, CurrentID=0, Status='live', Symbol='SMPX', Aliases='DFN6, DFNX4', description='small muscle protein, X-linked', other_designations='deafness, X-linked 6, sensorineural|stretch-responsive skeletal muscle protein', map_location='Xp22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21705972, end_position_on_the_genomic_accession=21758160, orientation='minus', exon_count=6, OMIM='300066,300226', no_hdr0=''),
  23677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23677, CurrentID=0, Status='live', Symbol='SH3BP4', Aliases='BOG25, TTP', description='SH3-domain binding protein 4', other_designations='EH-binding protein 10|transferrin receptor-trafficking protein', map_location='2q37.1-q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234951984, end_position_on_the_genomic_accession=235055714, orientation='plus', exon_count=10, OMIM=605611, no_hdr0=''),
  23678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23678, CurrentID=0, Status='live', Symbol='SGK3', Aliases='CISK, SGK2, SGKL', description='serum/glucocorticoid regulated kinase family member 3', other_designations='cytokine-independent survival kinase|serum/glucocorticoid regulated kinase family, member 3', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66712418, end_position_on_the_genomic_accession=66862022, orientation='plus', exon_count=18, OMIM=607591, no_hdr0=''),
  23679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23679, CurrentID=0, Status='live', Symbol='SEPW1P', Aliases='', description='selenoprotein W, 1 pseudogene', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31094571, end_position_on_the_genomic_accession=31095638, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140453, CurrentID=0, Status='live', Symbol='MUC17', Aliases='MUC-17, MUC-3, MUC3', description='mucin 17, cell surface associated', other_designations='membrane mucin MUC17|secreted mucin MUC17|small intestinal mucin MUC3|small intestinal mucin-3', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101020076, end_position_on_the_genomic_accession=101058859, orientation='plus', exon_count=13, OMIM=608424, no_hdr0=''),
  23682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23682, CurrentID=0, Status='live', Symbol='RAB38', Aliases='NY-MEL-1, rrGTPbp', description='RAB38, member RAS oncogene family', other_designations='Rab-related GTP-binding protein|melanoma antigen NY-MEL-1', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88113247, end_position_on_the_genomic_accession=88175467, orientation='minus', exon_count=3, OMIM=606281, no_hdr0=''),
  23683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23683, CurrentID=0, Status='live', Symbol='PRKD3', Aliases='EPK2, PKC-NU, PKD3, PRKCN, nPKC-NU', description='protein kinase D3', other_designations='protein kinase C nu type|protein kinase EPK2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37250502, end_position_on_the_genomic_accession=37324772, orientation='minus', exon_count=21, OMIM=607077, no_hdr0=''),
  375318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375318, CurrentID=0, Status='live', Symbol='AQP12A', Aliases='AQP-12, AQP12, AQPX2', description='aquaporin 12A', other_designations='aquaporin 12|aquaporin X2', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240691845, end_position_on_the_genomic_accession=240698483, orientation='plus', exon_count=4, OMIM=609789, no_hdr0=''),
  285830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285830, CurrentID=0, Status='live', Symbol='HLA-F-AS1', Aliases='', description='HLA-F antisense RNA 1', other_designations='HLA-F antisense RNA 1 (non-protein coding)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29726601, end_position_on_the_genomic_accession=29749049, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373402, CurrentID=0, Status='live', Symbol='LOC105373402', Aliases='', description='uncharacterized LOC105373402', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6398177, end_position_on_the_genomic_accession=6421122, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  154761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154761, CurrentID=0, Status='live', Symbol='LOC154761', Aliases='', description='family with sequence similarity 115, member C pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143811968, end_position_on_the_genomic_accession=143836717, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  285834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285834, CurrentID=0, Status='live', Symbol='HCG22', Aliases='PBMUCL2', description='HLA complex group 22', other_designations='HLA complex group 22 (non-protein coding)|panbronchiolitis related mucin-like 2', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31053450, end_position_on_the_genomic_accession=31059890, orientation='plus', exon_count=4, OMIM=613918, no_hdr0=''),
  23691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23691, CurrentID=0, Status='live', Symbol='NNO1', Aliases='', description='nanophthalmos 1', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600165, no_hdr0=''),
  23694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23694, CurrentID=0, Status='live', Symbol='MRX73', Aliases='', description='mental retardation, X-linked 73', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300355, no_hdr0=''),
  100271278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271278, CurrentID=0, Status='live', Symbol='RPL13AP14', Aliases='RPL13A_3_590', description='ribosomal protein L13a pseudogene 14', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83746388, end_position_on_the_genomic_accession=83746989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23696, CurrentID=0, Status='live', Symbol='MRX70', Aliases='', description='mental retardation, X-linked 70', other_designations='', map_location='Xq23-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23697, CurrentID=0, Status='live', Symbol='MRX69', Aliases='', description='mental retardation, X-linked 69', other_designations='', map_location='Xp11.21-q22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23699, CurrentID=0, Status='live', Symbol='MRX67', Aliases='', description='mental retardation, X-linked 67', other_designations='', map_location='Xq13.1-q21.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23700, CurrentID=0, Status='live', Symbol='MRX66', Aliases='', description='mental retardation, X-linked 66', other_designations='', map_location='Xq21.33-q23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285847, CurrentID=0, Status='live', Symbol='LOC285847', Aliases='', description='uncharacterized LOC285847', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35726762, end_position_on_the_genomic_accession=35736947, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  285848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285848, CurrentID=0, Status='live', Symbol='PNPLA1', Aliases='ARCI10, dJ50J22.1', description='patatin like phospholipase domain containing 1', other_designations='patatin-like phospholipase domain containing 1', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36239766, end_position_on_the_genomic_accession=36313955, orientation='plus', exon_count=13, OMIM=612121, no_hdr0=''),
  23705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23705, CurrentID=0, Status='live', Symbol='CADM1', Aliases='BL2, IGSF4, IGSF4A, NECL2, Necl-2, RA175, ST17, SYNCAM, TSLC1, sTSLC-1, sgIGSF, synCAM1', description='cell adhesion molecule 1', other_designations='TSLC-1|TSLC1/Nectin-like 2/IGSF4|immunoglobulin superfamily member 4|immunoglobulin superfamily, member 4|immunoglobulin superfamily, member 4D variant 1|immunoglobulin superfamily, member 4D variant 2|nectin-like 2|nectin-like protein 2|spermatogenic immunoglobulin superfamily|synaptic cell adhesion molecule|truncated CADM1 protein|tumor suppressor in lung cancer 1', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115173625, end_position_on_the_genomic_accession=115504523, orientation='minus', exon_count=12, OMIM=605686, no_hdr0=''),
  23706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23706, CurrentID=0, Status='live', Symbol='IBGC1', Aliases='BGCI, IBGC', description='idiopathic basal ganglia calcification 1', other_designations='Basal ganglia calcification, idiopathic (Fahr disease)', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=213600, no_hdr0=''),
  23708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23708, CurrentID=0, Status='live', Symbol='GSPT2', Aliases='ERF3B, GST2', description='G1 to S phase transition 2', other_designations='eukaryotic peptide chain release factor subunit 3b', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51743385, end_position_on_the_genomic_accession=51746232, orientation='plus', exon_count=1, OMIM=300418, no_hdr0=''),
  23710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23710, CurrentID=0, Status='live', Symbol='GABARAPL1', Aliases='APG8-LIKE, APG8L, ATG8, ATG8B, ATG8L, GEC1', description='GABA(A) receptor-associated protein like 1', other_designations='GABA(A) receptor-associated protein-like 1|GABARAPL1-a|GEC-1|early estrogen-regulated protein|glandular epithelial cell protein 1', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10212841, end_position_on_the_genomic_accession=10223128, orientation='plus', exon_count=4, OMIM=607420, no_hdr0=''),
  285855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285855, CurrentID=0, Status='live', Symbol='RPL7L1', Aliases='dJ475N16.4', description='ribosomal protein L7 like 1', other_designations='ribosomal protein L7-like 1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42879917, end_position_on_the_genomic_accession=42889889, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  23713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23713, CurrentID=0, Status='live', Symbol='DFNB27', Aliases='', description='deafness, autosomal recessive 27', other_designations='', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605818, no_hdr0=''),
  23714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23714, CurrentID=0, Status='live', Symbol='DFNB26', Aliases='', description='deafness, autosomal recessive 26', other_designations='', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605428, no_hdr0=''),
  375323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375323, CurrentID=0, Status='live', Symbol='LHFPL4', Aliases='', description='lipoma HMGIC fusion partner-like 4', other_designations='LHFP-like protein 4', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9498361, end_position_on_the_genomic_accession=9553802, orientation='minus', exon_count=4, OMIM=610240, no_hdr0=''),
  154790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154790, CurrentID=0, Status='live', Symbol='CLEC2L', Aliases='', description='C-type lectin domain family 2 member L', other_designations='C-type lectin domain family 2, member L', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139523928, end_position_on_the_genomic_accession=139544985, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23719, CurrentID=0, Status='live', Symbol='DFNA30', Aliases='', description='deafness, autosomal dominant 30', other_designations='', map_location='15q25-q26', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606451, no_hdr0=''),
  23720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23720, CurrentID=0, Status='live', Symbol='DFNA29', Aliases='', description='deafness, autosomal dominant 29', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23722, CurrentID=0, Status='live', Symbol='DFNA27', Aliases='', description='deafness, autosomal dominant 27', other_designations='', map_location='4q12-q13.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612431, no_hdr0=''),
  23723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23723, CurrentID=0, Status='live', Symbol='DFNA24', Aliases='', description='deafness, autosomal dominant 24', other_designations='', map_location='4q35-qter', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606282, no_hdr0=''),
  154796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154796, CurrentID=0, Status='live', Symbol='AMOT', Aliases='', description='angiomotin', other_designations='angiomotin p130 isoform|angiomotin p80 isoform', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112774877, end_position_on_the_genomic_accession=112840961, orientation='minus', exon_count=17, OMIM=300410, no_hdr0=''),
  23727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23727, CurrentID=0, Status='live', Symbol='CFM2', Aliases='', description='cystic fibrosis modifier 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285872, CurrentID=0, Status='live', Symbol='THAP5P1', Aliases='', description='THAP domain containing 5 pseudogene 1', other_designations='THAP5 pseudogene', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158622173, end_position_on_the_genomic_accession=158623381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23729, CurrentID=0, Status='live', Symbol='SHPK', Aliases='CARKL, SHK', description='sedoheptulokinase', other_designations='carbohydrate kinase-like protein', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3608262, end_position_on_the_genomic_accession=3636322, orientation='minus', exon_count=7, OMIM=605060, no_hdr0=''),
  23731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23731, CurrentID=0, Status='live', Symbol='TMEM245', Aliases='C9orf5, CG-2, CG2', description='transmembrane protein 245', other_designations='protein CG-2|transmembrane protein C9orf5', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109015135, end_position_on_the_genomic_accession=109119945, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  23732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23732, CurrentID=0, Status='live', Symbol='FRRS1L', Aliases='C9orf4, CG-6, CG6', description='ferric-chelate reductase 1-like', other_designations='brain protein CG-6|ferric-chelate reductase 1-like protein', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109136188, end_position_on_the_genomic_accession=109167291, orientation='minus', exon_count=7, OMIM=604574, no_hdr0=''),
  285877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285877, CurrentID=0, Status='live', Symbol='POM121L12', Aliases='', description='POM121 transmembrane nucleoporin like 12', other_designations='POM121 membrane glycoprotein-like 12|POM121 transmembrane nucleoporin-like 12', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53035656, end_position_on_the_genomic_accession=53036925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  285878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285878, CurrentID=0, Status='live', Symbol='VSTM2A-OT1', Aliases='GS1-18A18.1', description='VSTM2A overlapping transcript 1', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54556970, end_position_on_the_genomic_accession=54571726, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  154807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154807, CurrentID=0, Status='live', Symbol='VKORC1L1', Aliases='', description='vitamin K epoxide reductase complex subunit 1 like 1', other_designations='VKORC1-like protein 1|vitamin K epoxide reductase complex, subunit 1-like 1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65836336, end_position_on_the_genomic_accession=65959563, orientation='plus', exon_count=6, OMIM=608838, no_hdr0=''),
  154810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154810, CurrentID=0, Status='live', Symbol='AMOTL1', Aliases='JEAP', description='angiomotin like 1', other_designations='junction-enriched and associated protein', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94706436, end_position_on_the_genomic_accession=94876753, orientation='plus', exon_count=20, OMIM=614657, no_hdr0=''),
  23739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23739, CurrentID=0, Status='live', Symbol='C3CER1', Aliases='CER1', description='chromosome 3 common eliminated region 1', other_designations='', map_location='3p22-p21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23741, CurrentID=0, Status='live', Symbol='EID1', Aliases='C15orf3, CRI1, EID-1, IRO45620, PNAS-22, PTD014, RBP21', description='EP300 interacting inhibitor of differentiation 1', other_designations='21 kDa pRb-associated protein|CREBBP/EP300 inhibitor 1|CREBBP/EP300 inhibitory protein 1|E1A-like inhibitor of differentiation 1|NB4 apoptosis related protein|Rb- and p300-binding protein EID-1|retinoblastoma protein-associated protein', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48878093, end_position_on_the_genomic_accession=48880183, orientation='plus', exon_count=1, OMIM=605894, no_hdr0=''),
  23742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23742, CurrentID=0, Status='live', Symbol='NPAP1', Aliases='C15orf2', description='nuclear pore associated protein 1', other_designations='protein C15orf2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24675394, end_position_on_the_genomic_accession=24683446, orientation='plus', exon_count=1, OMIM=610922, no_hdr0=''),
  23743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23743, CurrentID=0, Status='live', Symbol='BHMT2', Aliases='', description='betaine--homocysteine S-methyltransferase 2', other_designations='SMM-hcy methyltransferase|betaine-homocysteine methyltransferase 2', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79069724, end_position_on_the_genomic_accession=79090074, orientation='plus', exon_count=8, OMIM=605932, no_hdr0=''),
  285888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285888, CurrentID=0, Status='live', Symbol='CNPY1', Aliases='', description='canopy FGF signaling regulator 1', other_designations='canopy 1 homolog', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155501258, end_position_on_the_genomic_accession=155533844, orientation='minus', exon_count=4, OMIM=612493, no_hdr0=''),
  285889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285889, CurrentID=0, Status='live', Symbol='LOC285889', Aliases='', description='uncharacterized LOC285889', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156437789, end_position_on_the_genomic_accession=156445588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23746, CurrentID=0, Status='live', Symbol='AIPL1', Aliases='AIPL2, LCA4', description='aryl hydrocarbon receptor interacting protein-like 1', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6423737, end_position_on_the_genomic_accession=6435199, orientation='minus', exon_count=6, OMIM=604392, no_hdr0=''),
  105373404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373404, CurrentID=0, Status='live', Symbol='LOC105373404', Aliases='', description='uncharacterized LOC105373404', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6790929, end_position_on_the_genomic_accession=6796822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  285893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285893, CurrentID=0, Status='live', Symbol='ARAFP3', Aliases='ARAF3P, ARAFPS2', description='ARAF pseudogene 3', other_designations='v-raf murine sarcoma 3611 viral oncogene homolog 3, pseudogene|v-raf murine sarcoma 3611 viral oncogene homolog pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63342713, end_position_on_the_genomic_accession=63343733, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  154822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154822, CurrentID=0, Status='live', Symbol='LINC00689', Aliases='', description='long intergenic non-protein coding RNA 689', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=159008354, end_position_on_the_genomic_accession=159026238, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23751, CurrentID=0, Status='live', Symbol='TOP3BP1', Aliases='TOP3B2', description='topoisomerase (DNA) III beta pseudogene 1', other_designations='topoisomerase (DNA) III beta, 2', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22223186, end_position_on_the_genomic_accession=22223665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23752, CurrentID=0, Status='live', Symbol='TSSK1A', Aliases='SPOGA1, STK22A, TSSK1, TSSK1b, TSSK7P', description='testis-specific serine kinase 1A pseudogene', other_designations='serine/threonine kinase 22A (spermiogenesis associated)|spermiogenesis associated 1|testis-specific serine kinase 7 pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19124918, end_position_on_the_genomic_accession=19125867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23753, CurrentID=0, Status='live', Symbol='SDF2L1', Aliases='', description='stromal cell-derived factor 2-like 1', other_designations='OTTHUMT00000075032|PWP1-interacting protein 8|SDF2-like 1|SDF2-like protein 1|dihydropyrimidinase-like 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21642253, end_position_on_the_genomic_accession=21644299, orientation='plus', exon_count=3, OMIM=607551, no_hdr0=''),
  23754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23754, CurrentID=0, Status='live', Symbol='RPL32P5', Aliases='RPL32L2, RPL32_23_1737', description='ribosomal protein L32 pseudogene 5', other_designations='ribosomal protein L32-like 2', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17192848, end_position_on_the_genomic_accession=17193358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23755, CurrentID=0, Status='live', Symbol='SIRPAP1', Aliases='PTPNS1L, SIRPA2P', description='signal-regulatory protein alpha pseudogene 1', other_designations='WUGSC:H_DJ0539M06.5', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30542215, end_position_on_the_genomic_accession=30544935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23756, CurrentID=0, Status='live', Symbol='RPS24P1', Aliases='RPS24P', description='ribosomal protein S24 pseudogene 1', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12244745, end_position_on_the_genomic_accession=12252124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23757, CurrentID=0, Status='live', Symbol='ARSFP1', Aliases='ARSFP', description='arylsulfatase F pseudogene 1', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12252300, end_position_on_the_genomic_accession=12258035, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23758, CurrentID=0, Status='live', Symbol='SFPQP1', Aliases='SFPQP', description='splicing factor proline/glutamine-rich pseudogene 1', other_designations='splicing factor proline/glutamine rich (polypyrimidine tract binding protein associated) pseudogene|splicing factor proline/glutamine rich (polypyrimidine tract-binding protein-associated) pseudogene|splicing factor proline/glutamine-rich (polypyrimidine tract binding protein associated) pseudogene 1', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13094804, end_position_on_the_genomic_accession=13097721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23759, CurrentID=0, Status='live', Symbol='PPIL2', Aliases='CYC4, CYP60, Cyp-60, UBOX7, hCyP-60', description='peptidylprolyl isomerase like 2', other_designations='PPIase|U-box domain containing 7|cyclophilin, 60kDa|cyclophilin-60|cyclophilin-like protein CyP-60|peptidylprolyl cis-trans isomerase|peptidylprolyl isomerase (cyclophilin)-like 2|rotamase PPIL2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21665984, end_position_on_the_genomic_accession=21697913, orientation='plus', exon_count=24, OMIM=607588, no_hdr0=''),
  23760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23760, CurrentID=0, Status='live', Symbol='PITPNB', Aliases='PI-TP-beta, PtdInsTP, VIB1B', description='phosphatidylinositol transfer protein beta', other_designations='PtdIns transfer protein beta|phosphatidylinositol transfer protein, beta|phosphotidylinositol transfer protein, beta', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27851669, end_position_on_the_genomic_accession=27919306, orientation='minus', exon_count=12, OMIM=606876, no_hdr0=''),
  23761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23761, CurrentID=0, Status='live', Symbol='PISD', Aliases='DJ858B16, PSD, PSDC, PSSC, dJ858B16.2', description='phosphatidylserine decarboxylase', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31618491, end_position_on_the_genomic_accession=31662564, orientation='minus', exon_count=18, OMIM=612770, no_hdr0=''),
  23762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23762, CurrentID=0, Status='live', Symbol='OSBP2', Aliases='HLM, ORP-4, ORP4, OSBPL1, OSBPL4', description='oxysterol binding protein 2', other_designations='OSBP-related protein 4|oxysterol binding protein-like 1|oxysterol-binding protein-related protein 4', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30693782, end_position_on_the_genomic_accession=30907824, orientation='plus', exon_count=20, OMIM=606729, no_hdr0=''),
  23764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23764, CurrentID=0, Status='live', Symbol='MAFF', Aliases='U-MAF, hMafF', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog F', other_designations='v-maf avian musculoaponeurotic fibrosarcoma oncogene family protein F', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38201932, end_position_on_the_genomic_accession=38216511, orientation='plus', exon_count=4, OMIM=604877, no_hdr0=''),
  23765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23765, CurrentID=0, Status='live', Symbol='IL17RA', Aliases='CANDF5, CD217, CDw217, IL-17RA, IL17R, hIL-17R', description='interleukin 17 receptor A', other_designations='IL-17 receptor A', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17084959, end_position_on_the_genomic_accession=17115694, orientation='plus', exon_count=13, OMIM=605461, no_hdr0=''),
  23766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23766, CurrentID=0, Status='live', Symbol='GABARAPL3', Aliases='ATG8D', description='GABA(A) receptors associated protein like 3, pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90346527, end_position_on_the_genomic_accession=90349447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23767, CurrentID=0, Status='live', Symbol='FLRT3', Aliases='HH21', description='fibronectin leucine rich transmembrane protein 3', other_designations='fibronectin-like domain-containing leucine-rich transmembrane protein 3', map_location='20p11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14323993, end_position_on_the_genomic_accession=14337667, orientation='minus', exon_count=4, OMIM=604808, no_hdr0=''),
  23768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23768, CurrentID=0, Status='live', Symbol='FLRT2', Aliases='', description='fibronectin leucine rich transmembrane protein 2', other_designations='fibronectin-like domain-containing leucine-rich transmembrane protein 2', map_location='14q24-q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85530105, end_position_on_the_genomic_accession=85628696, orientation='plus', exon_count=8, OMIM=604807, no_hdr0=''),
  23769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23769, CurrentID=0, Status='live', Symbol='FLRT1', Aliases='SPG68', description='fibronectin leucine rich transmembrane protein 1', other_designations='fibronectin-like domain-containing leucine-rich transmembrane protein 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64035970, end_position_on_the_genomic_accession=64119183, orientation='plus', exon_count=5, OMIM=604806, no_hdr0=''),
  23770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23770, CurrentID=0, Status='live', Symbol='FKBP8', Aliases='FKBP38, FKBPr38', description='FK506 binding protein 8', other_designations='38 kDa FK506-binding protein|38 kDa FKBP|FK506 binding protein 8, 38kDa|FK506-binding protein 8 (38kD)|FKBP-38|FKBP-8|PPIase FKBP8|hFKBP38|rotamase', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18531758, end_position_on_the_genomic_accession=18544077, orientation='minus', exon_count=10, OMIM=604840, no_hdr0=''),
  23771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23771, CurrentID=0, Status='live', Symbol='DIANPH', Aliases='', description='diabetic nephropathy', other_designations='', map_location='18q22.3-q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603933, no_hdr0=''),
  23772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23772, CurrentID=0, Status='live', Symbol='CSF2RBP1', Aliases='CSF2RB2, fF45C1.2', description='colony stimulating factor 2 receptor, beta, low-affinity (granulocyte-macrophage) pseudogene 1', other_designations='colony stimulating factor 2 receptor, beta, 2 pseudogene', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36950530, end_position_on_the_genomic_accession=36953724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23774, CurrentID=0, Status='live', Symbol='BRD1', Aliases='BRL, BRPF1, BRPF2', description='bromodomain containing 1', other_designations='BR140-like protein|bromodomain and PHD finger-containing protein 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49773278, end_position_on_the_genomic_accession=49827548, orientation='minus', exon_count=19, OMIM=604589, no_hdr0=''),
  105373405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373405, CurrentID=0, Status='live', Symbol='LOC105373405', Aliases='', description='uncharacterized LOC105373405', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7314117, end_position_on_the_genomic_accession=7332485, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23779, CurrentID=0, Status='live', Symbol='ARHGAP8', Aliases='BPGAP1, PP610', description='Rho GTPase activating protein 8', other_designations='BCH domain-containing Cdc42GAP-like protein|BNIP-2 and Cdc42GAP homology domain-containing, proline-rich and Cdc42GAP-like protein subtype-1|rho-type GTPase-activating protein 8', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44752558, end_position_on_the_genomic_accession=44862784, orientation='plus', exon_count=13, OMIM=609405, no_hdr0=''),
  23780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23780, CurrentID=0, Status='live', Symbol='APOL2', Aliases='APOL-II, APOL3', description='apolipoprotein L2', other_designations='apolipoprotein L, 2|apolipoprotein L-II', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36226209, end_position_on_the_genomic_accession=36239954, orientation='minus', exon_count=10, OMIM=607252, no_hdr0=''),
  23781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23781, CurrentID=0, Status='live', Symbol='AP1B1P2', Aliases='ADTB1L2, dJ127L4.3', description='adaptor related protein complex 1 beta 1 subunit pseudogene 2', other_designations='adaptin, beta 1-like 2|adaptor-related protein complex 1 beta 1 subunit pseudogene 2|adaptor-related protein complex 1, beta 1 subunit pseudogene 2', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32134028, end_position_on_the_genomic_accession=32134129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23782, CurrentID=0, Status='live', Symbol='AP1B1P1', Aliases='ADTB1L1, dJ127L4.2', description='adaptor related protein complex 1 beta 1 subunit pseudogene 1', other_designations='adaptin, beta 1-like 1|adaptor-related protein complex 1 beta 1 subunit pseudogene 1|adaptor-related protein complex 1, beta 1 subunit pseudogene 1', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32121977, end_position_on_the_genomic_accession=32133469, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  23783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23783, CurrentID=0, Status='live', Symbol='ANKRD62P1-PARP4P3', Aliases='ADPRTL4, PARP4P3', description='ANKRD62P1-PARP4P3 readthrough, transcribed pseudogene', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 4|ANKRD62P1-PARP4P3 readthrough (non-protein coding)|ANKRD62P1-PARP4P3 readthrough transcribed pseudogene|VWFP1-ANKRD62P1-PARP4P3|poly (ADP-ribose) polymerase family, member 4 pseudogene 3', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16653708, end_position_on_the_genomic_accession=16675540, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23784, CurrentID=0, Status='live', Symbol='POTEH', Aliases='A26C3, ACTBL1, CT104.7, POTE22', description='POTE ankyrin domain family member H', other_designations='ANKRD26-like family C member 3|ANKRD26-like family C, member 3|POTE ankyrin domain family, member H|POTE-22|actin, beta-like 1|cancer/testis antigen family 104, member 7|prostate, ovary, testis-expressed protein on chromosome 22|protein expressed in prostate, ovary, testis, and placenta 22|protein expressed in prostate, ovary, testis, and placenta POTE14 like', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15690026, end_position_on_the_genomic_accession=15721631, orientation='plus', exon_count=11, OMIM=608913, no_hdr0=''),
  23785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23785, CurrentID=0, Status='live', Symbol='ABCD1P1', Aliases='', description='ATP binding cassette subfamily D member 1 pseudogene 1', other_designations='ATP-binding cassette, sub-family D (ALD), member 1, pseudogene 1', map_location='2p11', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89763173, end_position_on_the_genomic_accession=89767368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23786, CurrentID=0, Status='live', Symbol='BCL2L13', Aliases='BCL-RAMBO, Bcl2-L-13, MIL1', description='BCL2-like 13 (apoptosis facilitator)', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17628855, end_position_on_the_genomic_accession=17730855, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23787, CurrentID=0, Status='live', Symbol='MTCH1', Aliases='CGI-64, PIG60, PSAP, SLC25A49', description='mitochondrial carrier 1', other_designations='cell proliferation-inducing protein 60|presenilin-associated protein|solute carrier family 25, member 49', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36968135, end_position_on_the_genomic_accession=36986551, orientation='minus', exon_count=12, OMIM=610449, no_hdr0=''),
  23788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23788, CurrentID=0, Status='live', Symbol='MTCH2', Aliases='HSPC032, MIMP, SLC25A50', description='mitochondrial carrier 2', other_designations='2310034D24Rik|met-induced mitochondrial protein|solute carrier family 25, member 50', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47605204, end_position_on_the_genomic_accession=47642654, orientation='minus', exon_count=15, OMIM=613221, no_hdr0=''),
  154865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154865, CurrentID=0, Status='live', Symbol='IQUB', Aliases='', description='IQ motif and ubiquitin domain containing', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123452181, end_position_on_the_genomic_accession=123534664, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  285941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285941, CurrentID=0, Status='live', Symbol='C7orf71', Aliases='', description='chromosome 7 open reading frame 71', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26637488, end_position_on_the_genomic_accession=26651531, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  285943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285943, CurrentID=0, Status='live', Symbol='HOXA-AS2', Aliases='HOXA3as', description='HOXA cluster antisense RNA 2', other_designations='HOXA cluster antisense RNA 2 (non-protein coding)', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27121919, end_position_on_the_genomic_accession=27128760, orientation='plus', exon_count=4, OMIM=616068, no_hdr0=''),
  154872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154872, CurrentID=0, Status='live', Symbol='C7orf77', Aliases='', description='chromosome 7 open reading frame 77', other_designations='putative uncharacterized protein LOC154872', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124777292, end_position_on_the_genomic_accession=124790810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  154881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154881, CurrentID=0, Status='live', Symbol='KCTD7', Aliases='CLN14, EPM3', description='potassium channel tetramerization domain containing 7', other_designations='potassium channel tetramerisation domain containing 7', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66628881, end_position_on_the_genomic_accession=66643229, orientation='plus', exon_count=4, OMIM=611725, no_hdr0=''),
  285954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285954, CurrentID=0, Status='live', Symbol='INHBA-AS1', Aliases='', description='INHBA antisense RNA 1', other_designations='INHBA antisense RNA 1 (non-protein coding)', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41693916, end_position_on_the_genomic_accession=41779378, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  285955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285955, CurrentID=0, Status='live', Symbol='SPDYE1', Aliases='Ringo1, SPDYE, WBSCR19', description='speedy/RINGO cell cycle regulator family member E1', other_designations='Speedy E|Williams Beuren syndrome chromosome region 19 protein|speedy homolog E1|speedy protein E1|williams-Beuren syndrome chromosomal region 19 protein', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43997440, end_position_on_the_genomic_accession=44010124, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  285957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285957, CurrentID=0, Status='live', Symbol='LOC285957', Aliases='', description='uncharacterized LOC285957', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285958, CurrentID=0, Status='live', Symbol='SNHG15', Aliases='C7orf40, Linc-Myo1g, MYO1GUT', description='small nucleolar RNA host gene 15', other_designations='MYO1G upstream transcript|Putative uncharacterized protein C7orf40|small nucleolar RNA host gene 15 (non-protein coding)', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44983028, end_position_on_the_genomic_accession=44986660, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  285961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285961, CurrentID=0, Status='live', Symbol='SEPT7P9', Aliases='CDC10L, SEPT7L, bA291L22.2', description='septin 7 pseudogene 9', other_designations='CDC10 cell division cycle 10 homolog-like|CDC10-like|septin 7-like', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38383069, end_position_on_the_genomic_accession=38402852, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  285962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285962, CurrentID=0, Status='live', Symbol='WEE2-AS1', Aliases='', description='WEE2 antisense RNA 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141704338, end_position_on_the_genomic_accession=141738230, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  285965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285965, CurrentID=0, Status='live', Symbol='EPHA1-AS1', Aliases='', description='EPHA1 antisense RNA 1', other_designations='EPHA1 antisense RNA 1 (non-protein coding)', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143407813, end_position_on_the_genomic_accession=143523449, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  285966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285966, CurrentID=0, Status='live', Symbol='TCAF2', Aliases='FAM115C, FAM139A', description='TRPM8 channel-associated factor 2', other_designations='TRP channel-associated factor 2|family with sequence similarity 115, member C|family with sequence similarity 139, member A|protein FAM115C', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143620952, end_position_on_the_genomic_accession=143730080, orientation='plus', exon_count=11, OMIM=616252, no_hdr0=''),
  285971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285971, CurrentID=0, Status='live', Symbol='ZNF775', Aliases='', description='zinc finger protein 775', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150379301, end_position_on_the_genomic_accession=150398635, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  285972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285972, CurrentID=0, Status='live', Symbol='LINC00996', Aliases='', description='long intergenic non-protein coding RNA 996', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150433654, end_position_on_the_genomic_accession=150448140, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  285973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285973, CurrentID=0, Status='live', Symbol='ATG9B', Aliases='APG9L2, NOS3AS, SONE', description='autophagy related 9B', other_designations='APG9-like 2|ATG9 autophagy related 9 homolog B|autophagy 9-like 2 protein|endothelial nitric oxide synthase antisense|nitric oxide synthase 3-overlapping antisense gene protein', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151012209, end_position_on_the_genomic_accession=151024499, orientation='minus', exon_count=19, OMIM=612205, no_hdr0=''),
  353497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353497, CurrentID=0, Status='live', Symbol='POLN', Aliases='POL4P', description='polymerase (DNA directed) nu', other_designations='DNA polymerase N|DNA polymerase POL4P', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2071958, end_position_on_the_genomic_accession=2242164, orientation='minus', exon_count=26, OMIM=610887, no_hdr0=''),
  154907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154907, CurrentID=0, Status='live', Symbol='C7orf66', Aliases='', description='chromosome 7 open reading frame 66', other_designations='putative uncharacterized protein LOC154907', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108883975, end_position_on_the_genomic_accession=108896211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  285987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285987, CurrentID=0, Status='live', Symbol='DLX6-AS1', Aliases='DLX6-AS, DLX6AS, Evf-2, NCRNA00212', description='DLX6 antisense RNA 1', other_designations='DLX6 antisense RNA (non-protein coding)|DLX6 antisense RNA 1 (non-protein coding)', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96968515, end_position_on_the_genomic_accession=97014065, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  285989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285989, CurrentID=0, Status='live', Symbol='ZNF789', Aliases='', description='zinc finger protein 789', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99472829, end_position_on_the_genomic_accession=99487600, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  646743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646743, CurrentID=0, Status='live', Symbol='LOC646743', Aliases='', description='uncharacterized LOC646743', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130536570, end_position_on_the_genomic_accession=130549915, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130550, CurrentID=0, Status='live', Symbol='LOC100130550', Aliases='', description='g2/M phase-specific E3 ubiquitin-protein ligase-like', other_designations='G2/M-phase specific E3 ubiquitin protein ligase pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132386656, end_position_on_the_genomic_accession=132389037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286002, CurrentID=0, Status='live', Symbol='SLC26A4-AS1', Aliases='', description='SLC26A4 antisense RNA 1', other_designations='SLC26A4 antisense RNA 1 (non-protein coding)', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107656516, end_position_on_the_genomic_accession=107661798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  286006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286006, CurrentID=0, Status='live', Symbol='LSMEM1', Aliases='C7orf53', description='leucine-rich single-pass membrane protein 1', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112480238, end_position_on_the_genomic_accession=112491062, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  154937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154937, CurrentID=0, Status='live', Symbol='LOC154937', Aliases='', description="protein phosphatase 2 regulatory subunit B', epsilon pseudogene", other_designations="protein phosphatase 2, regulatory subunit B', epsilon isoform pseudogene|protein phosphatase 2, regulatory subunit B', gamma pseudogene", map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56302924, end_position_on_the_genomic_accession=56305532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100293948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293948, CurrentID=0, Status='live', Symbol='ATG12P2', Aliases='', description='autophagy related 12 pseudogene 2', other_designations='ATG12 autophagy related 12 homolog pseudogene 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222901664, end_position_on_the_genomic_accession=222905059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373408, CurrentID=0, Status='live', Symbol='LOC105373408', Aliases='', description='uncharacterized LOC105373408', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7797098, end_position_on_the_genomic_accession=7863440, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  286016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286016, CurrentID=0, Status='live', Symbol='TPI1P2', Aliases='', description='triosephosphate isomerase 1 pseudogene 2', other_designations='Putative triosephosphate isomerase-like protein LOC286016', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129055223, end_position_on_the_genomic_accession=129057239, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  550112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550112, CurrentID=0, Status='live', Symbol='UBA6-AS1', Aliases='', description='UBA6 antisense RNA 1 (head to head)', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67701278, end_position_on_the_genomic_accession=67722505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271284, CurrentID=0, Status='live', Symbol='RPL17P35', Aliases='RPL17_9_1049', description='ribosomal protein L17 pseudogene 35', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64620380, end_position_on_the_genomic_accession=64620922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286023, CurrentID=0, Status='live', Symbol='FLJ40288', Aliases='', description='uncharacterized FLJ40288', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132648794, end_position_on_the_genomic_accession=132728769, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480668, CurrentID=0, Status='live', Symbol='RNU1-119P', Aliases='', description='RNA, U1 small nuclear 119, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373571, CurrentID=0, Status='live', Symbol='LOC105373571', Aliases='', description='uncharacterized LOC105373571', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113963216, end_position_on_the_genomic_accession=113970753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286042, CurrentID=0, Status='live', Symbol='FAM86B3P', Aliases='', description='family with sequence similarity 86 member B3, pseudogene', other_designations='family with sequence similarity 86, member B3, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8228570, end_position_on_the_genomic_accession=8244865, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  286043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286043, CurrentID=0, Status='live', Symbol='MRPS18CP2', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene 2', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8933678, end_position_on_the_genomic_accession=8934091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286046, CurrentID=0, Status='live', Symbol='XKR6', Aliases='C8orf21, C8orf5, C8orf7, XRG6', description='X-linked Kx blood group related 6', other_designations='Transmembrane protein C8orf5|X Kell blood group precursor-related family, member 6|XK, Kell blood group complex subunit-related family, member 6', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10896045, end_position_on_the_genomic_accession=11201381, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  286052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286052, CurrentID=0, Status='live', Symbol='LOC286052', Aliases='', description='uncharacterized LOC286052', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286053, CurrentID=0, Status='live', Symbol='NSMCE2', Aliases='C8orf36, MMS21, NSE2, ZMIZ7', description='NSE2/MMS21 homolog, SMC5-SMC6 complex SUMO ligase', other_designations='MMS21 homolog|hMMS21|methyl methanesulfonate sensitivity gene 21|non-SMC element 2 homolog|non-SMC element 2, MMS21 homolog|non-structural maintenance of chromosomes element 2 homolog|zinc finger, MIZ-type containing 7', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125091818, end_position_on_the_genomic_accession=125367125, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  646745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646745, CurrentID=0, Status='live', Symbol='LOC646745', Aliases='', description='general transcription factor IIIC subunit 6 pseudogene', other_designations='general transcription factor IIIC, polypeptide 6, alpha 35kDa pseudogene', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29551510, end_position_on_the_genomic_accession=29552244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286058, CurrentID=0, Status='live', Symbol='LOC286058', Aliases='', description='uncharacterized LOC286058', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286059, CurrentID=0, Status='live', Symbol='LOC286059', Aliases='', description='tumor necrosis factor receptor superfamily member 10d, decoy with truncated death domain pseudogene', other_designations='tumor necrosis factor receptor superfamily, member 10d, decoy with truncated death domain pseudogene', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23068229, end_position_on_the_genomic_accession=23083619, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286063, CurrentID=0, Status='live', Symbol='LOC286063', Aliases='', description='uncharacterized LOC286063', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  353512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353512, CurrentID=0, Status='live', Symbol='PKD1P4', Aliases='HG4', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 4', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18334400, end_position_on_the_genomic_accession=18352283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286068, CurrentID=0, Status='live', Symbol='LOC286068', Aliases='', description='uncharacterized LOC286068', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  353513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353513, CurrentID=0, Status='live', Symbol='VCY1B', Aliases='BPY1B', description='variable charge, Y-linked 1B', other_designations='basic charge, Y-linked 1|basic protein on Y chromosome, 1|testis-specific basic protein on Y, 1|variably charged protein Y', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14056222, end_position_on_the_genomic_accession=14056958, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  286075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286075, CurrentID=0, Status='live', Symbol='ZNF707', Aliases='', description='zinc finger protein 707', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143684452, end_position_on_the_genomic_accession=143695385, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  286076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286076, CurrentID=0, Status='live', Symbol='BREA2', Aliases='', description='breast cancer estrogen-induced apoptosis 2', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143697115, end_position_on_the_genomic_accession=143698413, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  286077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286077, CurrentID=0, Status='live', Symbol='FAM83H', Aliases='AI3', description='family with sequence similarity 83 member H', other_designations='FAM83H variant 1|FAM83H variant 2|family with sequence similarity 83, member H', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143723933, end_position_on_the_genomic_accession=143738030, orientation='minus', exon_count=7, OMIM=611927, no_hdr0=''),
  155006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155006, CurrentID=0, Status='live', Symbol='TMEM213', Aliases='', description='transmembrane protein 213', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138797915, end_position_on_the_genomic_accession=138806024, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060501, CurrentID=0, Status='live', Symbol='LOC101060501', Aliases='', description='lupus La protein pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286083, CurrentID=0, Status='live', Symbol='LOC286083', Aliases='', description='uncharacterized LOC286083', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1296034, end_position_on_the_genomic_accession=1302607, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  286087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286087, CurrentID=0, Status='live', Symbol='LOC286087', Aliases='', description='uncharacterized LOC286087', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  155019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155019, CurrentID=0, Status='live', Symbol='ST13P7', Aliases='FAM10A7, FAM10A7P', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 7', other_designations='family with sequence similarity 10, member A7 (pseudogene)', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133168002, end_position_on_the_genomic_accession=133170598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286094, CurrentID=0, Status='live', Symbol='LINC01591', Aliases='', description='long intergenic non-protein coding RNA 1591', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=135234131, end_position_on_the_genomic_accession=135299719, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  286097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286097, CurrentID=0, Status='live', Symbol='MICU3', Aliases='EFHA2', description='mitochondrial calcium uptake family member 3', other_designations='EF hand domain family A2|EF hand domain family, member A2|EF-hand domain family, member A2|EF-hand domain-containing family member A2|mitochondrial calcium uptake family, member 3|mitochondrial uptake family, member 3', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17027238, end_position_on_the_genomic_accession=17122639, orientation='plus', exon_count=17, OMIM=610633, no_hdr0=''),
  286101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286101, CurrentID=0, Status='live', Symbol='ZNF252P', Aliases='ZNF252', description='zinc finger protein 252, pseudogene', other_designations='zinc finger protein pseudogene', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144973589, end_position_on_the_genomic_accession=145002899, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  286102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286102, CurrentID=0, Status='live', Symbol='TMED10P1', Aliases='TMED10P, Tmp21-II', description='transmembrane emp24-like trafficking protein 10 (yeast) pseudogene 1', other_designations='Tmp21-II , transcribed pseudogene', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144994865, end_position_on_the_genomic_accession=144998897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  286103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286103, CurrentID=0, Status='live', Symbol='ZNF252P-AS1', Aliases='C8orf77', description='ZNF252P antisense RNA 1', other_designations='ZNF252P antisense RNA 1 (non-protein coding)', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=145002811, end_position_on_the_genomic_accession=145006046, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  286109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286109, CurrentID=0, Status='live', Symbol='CASC7', Aliases='LINC00980', description='cancer susceptibility candidate 7 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 980', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  155038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155038, CurrentID=0, Status='live', Symbol='GIMAP8', Aliases='IAN-9, IAN6, IAN9, IANT', description='GTPase, IMAP family member 8', other_designations='human immune associated nucleotide 6|immune-associated nucleotide-binding protein 9', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150450630, end_position_on_the_genomic_accession=150479395, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  286114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286114, CurrentID=0, Status='live', Symbol='LOC286114', Aliases='', description='uncharacterized LOC286114', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20973986, end_position_on_the_genomic_accession=20995119, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  286121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286121, CurrentID=0, Status='live', Symbol='LOC286121', Aliases='', description='uncharacterized LOC286121', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286122, CurrentID=0, Status='live', Symbol='C8orf31', Aliases='', description='chromosome 8 open reading frame 31', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143039209, end_position_on_the_genomic_accession=143054303, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  155051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155051, CurrentID=0, Status='live', Symbol='CRYGN', Aliases='', description='crystallin gamma N', other_designations='crystallin, gamma N|gamma-N-crystallin|gammaN-crystallin', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151429970, end_position_on_the_genomic_accession=151448419, orientation='minus', exon_count=6, OMIM=609603, no_hdr0=''),
  155054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155054, CurrentID=0, Status='live', Symbol='ZNF425', Aliases='', description='zinc finger protein 425', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149102784, end_position_on_the_genomic_accession=149126346, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  286128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286128, CurrentID=0, Status='live', Symbol='ZFP41', Aliases='ZNF753, zfp-41', description='ZFP41 zinc finger protein', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143246821, end_position_on_the_genomic_accession=143262705, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  155060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155060, CurrentID=0, Status='live', Symbol='LOC155060', Aliases='', description='AI894139 pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149285281, end_position_on_the_genomic_accession=149297312, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  155061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155061, CurrentID=0, Status='live', Symbol='ZNF746', Aliases='PARIS', description='zinc finger protein 746', other_designations='parkin interacting substrate|parkin-interacting substrate|parkin-interacting sustrate', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149472793, end_position_on_the_genomic_accession=149497837, orientation='minus', exon_count=8, OMIM=613914, no_hdr0=''),
  286135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286135, CurrentID=0, Status='live', Symbol='FAM183CP', Aliases='', description='family with sequence similarity 183 member C, pseudogene', other_designations='family with sequence similarity 183, member C, pseudogene', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29921513, end_position_on_the_genomic_accession=29953609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  155066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155066, CurrentID=0, Status='live', Symbol='ATP6V0E2', Aliases='ATP6V0E2L, C7orf32', description='ATPase, H+ transporting V0 subunit e2', other_designations='H+-ATPase e2 subunit|V-ATPase subunit e 2|lysosomal 9 kDa H(+)-transporting ATPase V0 subunit e2|vacuolar proton pump subunit e 2|vacuolar proton-ATPase subunit', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149872968, end_position_on_the_genomic_accession=149880713, orientation='plus', exon_count=5, OMIM=611019, no_hdr0=''),
  100271288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271288, CurrentID=0, Status='live', Symbol='RPL19P18', Aliases='RPL19_7_1512', description='ribosomal protein L19 pseudogene 18', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9589820, end_position_on_the_genomic_accession=9590515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286140, CurrentID=0, Status='live', Symbol='RNF5P1', Aliases='', description='ring finger protein 5, E3 ubiquitin protein ligase pseudogene 1', other_designations='ring finger protein 5 pseudogene 1', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38600173, end_position_on_the_genomic_accession=38601257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286144, CurrentID=0, Status='live', Symbol='TRIQK', Aliases='C8orf83, PRO0845, UPF0599', description='triple QxxK/R motif containing', other_designations='protein TRIQK|triple repetitive-sequence of QXXK/R protein homolog', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92883530, end_position_on_the_genomic_accession=92966144, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377777, CurrentID=0, Status='live', Symbol='LOC105377777', Aliases='', description='uncharacterized LOC105377777', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=862574, end_position_on_the_genomic_accession=882930, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  286148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286148, CurrentID=0, Status='live', Symbol='DPY19L4', Aliases='', description='dpy-19-like 4 (C. elegans)', other_designations='dpy-19-like protein 4|protein dpy-19 homolog 4', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94719875, end_position_on_the_genomic_accession=94793848, orientation='plus', exon_count=20, OMIM=613895, no_hdr0=''),
  100419682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419682, CurrentID=0, Status='live', Symbol='ZNF863P', Aliases='', description='zinc finger protein 863, pseudogene', other_designations='zinc finger protein 791 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52076098, end_position_on_the_genomic_accession=52077440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286150, CurrentID=0, Status='live', Symbol='RPS2P33', Aliases='RPS2_11_929', description='ribosomal protein S2 pseudogene 33', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97624166, end_position_on_the_genomic_accession=97625102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286151, CurrentID=0, Status='live', Symbol='FBXO43', Aliases='EMI2, ERP1, FBX43', description='F-box protein 43', other_designations='early mitotic inhibitor 2|endogenous meiotic inhibitor 2', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100133360, end_position_on_the_genomic_accession=100150569, orientation='minus', exon_count=9, OMIM=609110, no_hdr0=''),
  286154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286154, CurrentID=0, Status='live', Symbol='LOC286154', Aliases='', description='uncharacterized LOC286154', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286156, CurrentID=0, Status='live', Symbol='MAPK6PS5', Aliases='', description='mitogen-activated protein kinase 6 pseudogene 5', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109470138, end_position_on_the_genomic_accession=109472254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286157, CurrentID=0, Status='live', Symbol='PCBP2P2', Aliases='', description='poly(rC) binding protein 2 pseudogene 2', other_designations='', map_location='8q21-q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74602877, end_position_on_the_genomic_accession=74604528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286161, CurrentID=0, Status='live', Symbol='LOC286161', Aliases='', description='uncharacterized LOC286161', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  548313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548313, CurrentID=0, Status='live', Symbol='SSX4B', Aliases='CT5.4', description='synovial sarcoma, X breakpoint 4B', other_designations='OTTHUMT00000056510|cancer/testis antigen 5.4', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48402082, end_position_on_the_genomic_accession=48411910, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  155100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155100, CurrentID=0, Status='live', Symbol='CCT8L1P', Aliases='CCT8L1', description='chaperonin containing TCP1, subunit 8 (theta)-like 1, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152445217, end_position_on_the_genomic_accession=152447269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286177, CurrentID=0, Status='live', Symbol='LOC286177', Aliases='', description='uncharacterized LOC286177', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57261226, end_position_on_the_genomic_accession=57266611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286178, CurrentID=0, Status='live', Symbol='LOC286178', Aliases='', description='uncharacterized LOC286178', other_designations='CTD-2339F6.1', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57746224, end_position_on_the_genomic_accession=57750197, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  286183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286183, CurrentID=0, Status='live', Symbol='NKAIN3', Aliases='FAM77D', description='Na+/K+ transporting ATPase interacting 3', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 3|family with sequence similarity 77, member D', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62248310, end_position_on_the_genomic_accession=62999652, orientation='plus', exon_count=12, OMIM=612872, no_hdr0=''),
  286184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286184, CurrentID=0, Status='live', Symbol='LINC01289', Aliases='', description='long intergenic non-protein coding RNA 1289', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63769430, end_position_on_the_genomic_accession=63785501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927631, CurrentID=0, Status='live', Symbol='ADNP-AS1', Aliases='', description='ADNP antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50930984, end_position_on_the_genomic_accession=50945134, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  286186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286186, CurrentID=0, Status='live', Symbol='LINC01299', Aliases='', description='long intergenic non-protein coding RNA 1299', other_designations='CTD-3025N20.2', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65527008, end_position_on_the_genomic_accession=65562666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  286187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286187, CurrentID=0, Status='live', Symbol='PPP1R42', Aliases='LRRC67, TLLR, dtr', description='protein phosphatase 1 regulatory subunit 42', other_designations='leucine rich repeat containing 67|leucine-rich repeat-containing protein 67|protein phosphatase 1, regulatory subunit 42|testis leucine-rich repeat', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66988131, end_position_on_the_genomic_accession=67028551, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  286189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286189, CurrentID=0, Status='live', Symbol='C8orf34-AS1', Aliases='', description='C8orf34 antisense RNA 1', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68303468, end_position_on_the_genomic_accession=68331491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  286190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286190, CurrentID=0, Status='live', Symbol='LACTB2-AS1', Aliases='', description='LACTB2 antisense RNA 1', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70608577, end_position_on_the_genomic_accession=70663279, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  286191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286191, CurrentID=0, Status='live', Symbol='LOC286191', Aliases='', description='uncharacterized LOC286191', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060521, CurrentID=0, Status='live', Symbol='LOC101060521', Aliases='', description='DNA-directed RNA polymerase III subunit RPC5', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372961, CurrentID=0, Status='live', Symbol='LOC105372961', Aliases='', description='uncharacterized LOC105372961', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24576065, end_position_on_the_genomic_accession=24580246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  286204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286204, CurrentID=0, Status='live', Symbol='CRB2', Aliases='FSGS9, VMCKD', description='crumbs family member 2', other_designations='crumbs homolog 2|crumbs-like protein 2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123354459, end_position_on_the_genomic_accession=123380335, orientation='plus', exon_count=17, OMIM=609720, no_hdr0=''),
  286205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286205, CurrentID=0, Status='live', Symbol='SCAI', Aliases='C9orf126, NET40', description='suppressor of cancer cell invasion', other_designations='suppressor of cancer cell invasion protein', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124942608, end_position_on_the_genomic_accession=125143559, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  286207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286207, CurrentID=0, Status='live', Symbol='CFAP157', Aliases='C9orf117', description='cilia and flagella associated protein 157', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127706988, end_position_on_the_genomic_accession=127716002, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100294145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294145, CurrentID=0, Status='live', Symbol='LOC100294145', Aliases='', description='uncharacterized LOC100294145', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32894176, end_position_on_the_genomic_accession=32903758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101060524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060524, CurrentID=0, Status='live', Symbol='LOC101060524', Aliases='', description='D(1B) dopamine receptor-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144274534, end_position_on_the_genomic_accession=144276762, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286223, CurrentID=0, Status='live', Symbol='C9orf47', Aliases='C9orf108, bA791O21.3', description='chromosome 9 open reading frame 47', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88990863, end_position_on_the_genomic_accession=88996142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100294160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294160, CurrentID=0, Status='live', Symbol='LOC100294160', Aliases='', description='protein FAM136A-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271291, CurrentID=0, Status='live', Symbol='RPL23AP22', Aliases='RPL23A_2_142', description='ribosomal protein L23a pseudogene 22', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193756813, end_position_on_the_genomic_accession=193757267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286234, CurrentID=0, Status='live', Symbol='SPATA31E1', Aliases='C9orf79, FAM75E1', description='SPATA31 subfamily E, member 1', other_designations='FAM75-like protein C9orf79|XXyac-YM21GA2.5|family with sequence similarity 75, member E1|protein FAM75E1', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87882857, end_position_on_the_genomic_accession=87888899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  286238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286238, CurrentID=0, Status='live', Symbol='LOC286238', Aliases='', description='uncharacterized LOC286238', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88627068, end_position_on_the_genomic_accession=88652179, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  286239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286239, CurrentID=0, Status='live', Symbol='ME2P1', Aliases='', description='malic enzyme 2, NAD(+)-dependent, mitochondrial pseudogene 1', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=29823711, end_position_on_the_genomic_accession=29826718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100294188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294188, CurrentID=0, Status='live', Symbol='LOC100294188', Aliases='', description='protein FAM136A-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286254, CurrentID=0, Status='live', Symbol='LOC286254', Aliases='', description='uncharacterized LOC286254', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286256, CurrentID=0, Status='live', Symbol='LCN12', Aliases='', description='lipocalin 12', other_designations='lipocalcin 12', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136949195, end_position_on_the_genomic_accession=136955498, orientation='plus', exon_count=8, OMIM=612905, no_hdr0=''),
  155185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155185, CurrentID=0, Status='live', Symbol='AMZ1', Aliases='', description='archaelysin family metallopeptidase 1', other_designations='archeobacterial metalloproteinase-like protein 1|metalloproteinase-like protein', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2679522, end_position_on_the_genomic_accession=2765679, orientation='plus', exon_count=9, OMIM=615168, no_hdr0=''),
  100271292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271292, CurrentID=0, Status='live', Symbol='RPL23AP23', Aliases='RPL23A_5_178', description='ribosomal protein L23a pseudogene 23', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235295867, end_position_on_the_genomic_accession=235296337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286262, CurrentID=0, Status='live', Symbol='TPRN', Aliases='C9orf75, DFNB79', description='taperin', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137191617, end_position_on_the_genomic_accession=137200711, orientation='minus', exon_count=4, OMIM=613354, no_hdr0=''),
  25865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25865, CurrentID=0, Status='live', Symbol='PRKD2', Aliases='HSPC187, PKD2, nPKC-D2', description='protein kinase D2', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46674316, end_position_on_the_genomic_accession=46717127, orientation='minus', exon_count=19, OMIM=607074, no_hdr0=''),
  286272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286272, CurrentID=0, Status='live', Symbol='LOC286272', Aliases='', description='uncharacterized LOC286272', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421009, CurrentID=0, Status='live', Symbol='LOC100421009', Aliases='', description='UPF0607 protein ENSP00000381418-like pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50695014, end_position_on_the_genomic_accession=50696002, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  24137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24137, CurrentID=0, Status='live', Symbol='KIF4A', Aliases='KIF4, KIF4G1, MRX100', description='kinesin family member 4A', other_designations='chromokinesin-A', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70290029, end_position_on_the_genomic_accession=70420924, orientation='plus', exon_count=31, OMIM=300521, no_hdr0=''),
  24138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24138, CurrentID=0, Status='live', Symbol='IFIT5', Aliases='ISG58, P58, RI58', description='interferon induced protein with tetratricopeptide repeats 5', other_designations='IFIT-5|interferon-induced 58 kDa protein|retinoic acid- and interferon-inducible 58 kDa protein|retinoic acid- and interferon-inducible protein (58kD)', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89414568, end_position_on_the_genomic_accession=89421002, orientation='plus', exon_count=2, OMIM=616135, no_hdr0=''),
  24139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24139, CurrentID=0, Status='live', Symbol='EML2', Aliases='ELP70, EMAP-2, EMAP2', description='echinoderm microtubule associated protein like 2', other_designations='echinoderm MT-associated protein (EMAP)-like protein 70|microtubule-associated protein like echinoderm EMAP', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45609400, end_position_on_the_genomic_accession=45645601, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  24140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24140, CurrentID=0, Status='live', Symbol='FTSJ1', Aliases='CDLIV, JM23, MRX44, MRX9, SPB1, TRMT7', description='FtsJ RNA methyltransferase homolog 1 (E. coli)', other_designations="2'-O-ribose RNA methyltransferase TRM7 homolog|FtsJ-like protein 1|cell division protein|putative ribosomal RNA methyltransferase 1|rRNA (uridine-2'-O-)-methyltransferase|tRNA methyltransferase 7 homolog", map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48476021, end_position_on_the_genomic_accession=48486364, orientation='plus', exon_count=15, OMIM=300499, no_hdr0=''),
  24141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24141, CurrentID=0, Status='live', Symbol='LAMP5', Aliases='BAD-LAMP, BADLAMP, C20orf103, LAMP-5, UNC-43', description='lysosomal associated membrane protein family member 5', other_designations='LAMP family protein C20orf103|brain and dendritic cell associated LAMP|brain and dendritic cell-associated LAMP|brain-associated LAMP-like protein|lysosomal-associated membrane protein family, member 5|lysosome-associated membrane protein 5', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9514318, end_position_on_the_genomic_accession=9530524, orientation='plus', exon_count=6, OMIM=614641, no_hdr0=''),
  24142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24142, CurrentID=0, Status='live', Symbol='NAT6', Aliases='FUS-2, FUS2', description='N-acetyltransferase 6 (GCN5-related)', other_designations='protein fusion-2', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50296402, end_position_on_the_genomic_accession=50299421, orientation='minus', exon_count=3, OMIM=607073, no_hdr0=''),
  91405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91405, CurrentID=0, Status='live', Symbol='PLEKHA3P1', Aliases='PLEKHA3P', description='pleckstrin homology domain containing A3 pseudogene 1', other_designations='pleckstrin homology domain containing, family A member 3 pseudogene 1|pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 3 pseudogene|pleckstrin homology domain-containing, family A member 3 pseudogene 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41520234, end_position_on_the_genomic_accession=41521953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  24144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24144, CurrentID=0, Status='live', Symbol='TFIP11', Aliases='NTR1, Spp382, TIP39, bK445C9.6, hNtr1', description='tuftelin interacting protein 11', other_designations='STIP-1|septin and tuftelin-interacting protein 1', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26491927, end_position_on_the_genomic_accession=26512496, orientation='minus', exon_count=16, OMIM=612747, no_hdr0=''),
  24145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24145, CurrentID=0, Status='live', Symbol='PANX1', Aliases='MRS1, PX1, UNQ2529', description='pannexin 1', other_designations='innexin', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94128928, end_position_on_the_genomic_accession=94185596, orientation='plus', exon_count=7, OMIM=608420, no_hdr0=''),
  24146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24146, CurrentID=0, Status='live', Symbol='CLDN15', Aliases='', description='claudin 15', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101232092, end_position_on_the_genomic_accession=101238820, orientation='minus', exon_count=6, OMIM=615778, no_hdr0=''),
  24147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24147, CurrentID=0, Status='live', Symbol='FJX1', Aliases='', description='four jointed box 1', other_designations='four-jointed protein homolog|putative secreted ligand homologous to fjx1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35618187, end_position_on_the_genomic_accession=35620873, orientation='plus', exon_count=1, OMIM=612206, no_hdr0=''),
  24148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24148, CurrentID=0, Status='live', Symbol='PRPF6', Aliases='ANT-1, ANT1, C20orf14, Prp6, RP60, SNRNP102, TOM, U5-102K, hPrp6', description='pre-mRNA processing factor 6', other_designations='PRP6 homolog|PRP6 pre-mRNA processing factor 6 homolog|U5 snRNP-associated 102 kDa protein|U5-102 kDa protein|androgen receptor N-terminal domain transactivating protein-1|androgen receptor N-terminal domain-transactivating protein 1|p102 U5 small nuclear ribonucleoprotein particle-binding protein|putative mitochondrial outer membrane protein import receptor', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63981069, end_position_on_the_genomic_accession=64033100, orientation='plus', exon_count=21, OMIM=613979, no_hdr0=''),
  24149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24149, CurrentID=0, Status='live', Symbol='ZNF318', Aliases='HRIHFB2436, TZF, ZFP318', description='zinc finger protein 318', other_designations='endocrine regulator|endocrine regulatory protein|testicular zinc finger', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43327165, end_position_on_the_genomic_accession=43369459, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  24150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=24150, CurrentID=0, Status='live', Symbol='TP53TG3', Aliases='P53TG3A, TP53TG3E, TP53TG3F, TP53TG3', description='TP53 target 3', other_designations='TP53-inducible gene 3 protein|p53 target gene 3', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32673528, end_position_on_the_genomic_accession=32676129, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  286297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286297, CurrentID=0, Status='live', Symbol='LOC286297', Aliases='', description='uncharacterized LOC286297', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63112403, end_position_on_the_genomic_accession=63127100, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  25871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25871, CurrentID=0, Status='live', Symbol='C3orf17', Aliases='NEPRO, NET17', description='chromosome 3 open reading frame 17', other_designations='uncharacterized protein C3orf17', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113002444, end_position_on_the_genomic_accession=113019851, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100885776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885776, CurrentID=0, Status='live', Symbol='UGDH-AS1', Aliases='', description='UGDH antisense RNA 1', other_designations='UGDH antisense RNA 1 (non-protein coding)', map_location='4p', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39527839, end_position_on_the_genomic_accession=39594707, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  286310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286310, CurrentID=0, Status='live', Symbol='LCN1P1', Aliases='LCN1L1, bA430N14.2', description='lipocalin 1 pseudogene 1', other_designations='lipocalin 1 (tear prealbumin) pseudogene|lipocalin 1-like 1', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133224905, end_position_on_the_genomic_accession=133228651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25873, CurrentID=0, Status='live', Symbol='RPL36', Aliases='L36', description='ribosomal protein L36', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5690261, end_position_on_the_genomic_accession=5691667, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  25874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25874, CurrentID=0, Status='live', Symbol='MPC2', Aliases='BRP44', description='mitochondrial pyruvate carrier 2', other_designations='brain protein 44', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167916675, end_position_on_the_genomic_accession=167937069, orientation='minus', exon_count=7, OMIM=614737, no_hdr0=''),
  286319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286319, CurrentID=0, Status='live', Symbol='TUSC1', Aliases='TSG-9, TSG9', description='tumor suppressor candidate 1', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25676389, end_position_on_the_genomic_accession=25678858, orientation='minus', exon_count=1, OMIM=610529, no_hdr0=''),
  25875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25875, CurrentID=0, Status='live', Symbol='LETMD1', Aliases='1110019O13Rik, HCCR-1, HCCR-2, HCCR1, HCRR-2', description='LETM1 domain containing 1', other_designations='cervical cancer 1 proto-oncogene protein p40|cervical cancer proto-oncogene 2 protein|regulator of TP53', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51048230, end_position_on_the_genomic_accession=51067369, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100873598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873598, CurrentID=0, Status='live', Symbol='RNA5SP332', Aliases='RN5S332', description='RNA, 5S ribosomal pseudogene 332', other_designations='RNA, 5S ribosomal 332', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14134986, end_position_on_the_genomic_accession=14135100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  25876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25876, CurrentID=0, Status='live', Symbol='SPEF1', Aliases='C20orf28, CLAMPA, SPEF1', description='sperm flagellar 1', other_designations='calponin-homology and microtubule-associated protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3777504, end_position_on_the_genomic_accession=3781466, orientation='minus', exon_count=7, OMIM=610674, no_hdr0=''),
  286333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286333, CurrentID=0, Status='live', Symbol='FAM225A', Aliases='C9orf109, LINC00256A, NCRNA00256A', description='family with sequence similarity 225 member A (non-protein coding)', other_designations='family with sequence similarity 225, member A (non-protein coding)|long intergenic non-protein coding RNA 256A|non-protein coding RNA 256A', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113112896, end_position_on_the_genomic_accession=113119846, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286336, CurrentID=0, Status='live', Symbol='FAM78A', Aliases='C9orf59', description='family with sequence similarity 78 member A', other_designations='family with sequence similarity 78, member A', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131258078, end_position_on_the_genomic_accession=131281920, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  25878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25878, CurrentID=0, Status='live', Symbol='MXRA5', Aliases='', description='matrix-remodelling associated 5', other_designations='adhesion protein with leucine-rich repeats and Immunoglobulin domains related to perlecan|adlican', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3308565, end_position_on_the_genomic_accession=3347092, orientation='minus', exon_count=7, OMIM=300938, no_hdr0=''),
  286343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286343, CurrentID=0, Status='live', Symbol='LURAP1L', Aliases='C9orf150, HYST0841, LRAP35b, bA3L8.2', description='leucine rich adaptor protein 1-like', other_designations='leucine repeat adaptor protein 35b', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12774990, end_position_on_the_genomic_accession=12823060, orientation='plus', exon_count=3, OMIM=616130, no_hdr0=''),
  100033815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033815, CurrentID=0, Status='live', Symbol='SNORD115-41', Aliases='HBII-52-41', description='small nucleolar RNA, C/D box 115-41', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25245478, end_position_on_the_genomic_accession=25245559, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  286348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286348, CurrentID=0, Status='live', Symbol='RPL7P33', Aliases='RPL7_16_965', description='ribosomal protein L7 pseudogene 33', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15361374, end_position_on_the_genomic_accession=15362204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271295, CurrentID=0, Status='live', Symbol='RPL23AP35', Aliases='RPL23A_8_306', description='ribosomal protein L23a pseudogene 35', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186261417, end_position_on_the_genomic_accession=186261915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25880, CurrentID=0, Status='live', Symbol='TMEM186', Aliases='C16orf51', description='transmembrane protein 186', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8795180, end_position_on_the_genomic_accession=8797648, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  402342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402342, CurrentID=0, Status='live', Symbol='RPL13AP18', Aliases='RPL13A_5_919', description='ribosomal protein L13a pseudogene 18', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80265505, end_position_on_the_genomic_accession=80268774, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286362, CurrentID=0, Status='live', Symbol='OR13C9', Aliases='OR37L, OR9-13', description='olfactory receptor family 13 subfamily C member 9', other_designations='olfactory receptor OR9-13|olfactory receptor, family 13, subfamily C, member 9', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104617248, end_position_on_the_genomic_accession=104618204, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286365, CurrentID=0, Status='live', Symbol='OR13D1', Aliases='OR9-15', description='olfactory receptor family 13 subfamily D member 1', other_designations='olfactory receptor OR9-15|olfactory receptor, family 13, subfamily D, member 1', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104694422, end_position_on_the_genomic_accession=104695462, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441623, CurrentID=0, Status='live', Symbol='FTLP7', Aliases='', description='ferritin, light polypeptide pseudogene 7', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115538718, end_position_on_the_genomic_accession=115539583, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286370, CurrentID=0, Status='live', Symbol='MIR4290HG', Aliases='', description='MIR4290 host gene', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90020654, end_position_on_the_genomic_accession=90041499, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  91419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91419, CurrentID=0, Status='live', Symbol='XRCC6BP1', Aliases='KUB3', description='XRCC6 binding protein 1', other_designations='Ku70 binding protein 3|Ku70-binding protein 3|XRCC6-binding protein 1', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57941541, end_position_on_the_genomic_accession=57957269, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100033820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033820, CurrentID=0, Status='live', Symbol='SNORD116-28', Aliases='HBII-85-28', description='small nucleolar RNA, C/D box 116-28', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25104641, end_position_on_the_genomic_accession=25104733, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  286380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286380, CurrentID=0, Status='live', Symbol='FOXD4L3', Aliases='FOXD4L2, FOXD6', description='forkhead box D4-like 3', other_designations='FOXD4-like 2|FOXD4-like 3|Forkhead box protein D4-like 3|forkhead box D4 like 3', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68302867, end_position_on_the_genomic_accession=68305084, orientation='plus', exon_count=1, OMIM=611086, no_hdr0=''),
  286382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286382, CurrentID=0, Status='live', Symbol='LOC286382', Aliases='', description='uncharacterized LOC286382', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100033821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033821, CurrentID=0, Status='live', Symbol='SNORD116-29', Aliases='HBII-85-29', description='small nucleolar RNA, C/D box 116-29', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25106520, end_position_on_the_genomic_accession=25106604, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  135112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135112, CurrentID=0, Status='live', Symbol='NCOA7', Aliases='ERAP140, ESNA1-AS, Nbla00052, Nbla10993, TLDC4, dJ187J11.3, NCOA7', description='nuclear receptor coactivator 7', other_designations='140 kDa estrogen receptor-associated protein|TBC/LysM-associated domain containing 4|estrogen nuclear receptor coactivator 1|estrogen receptor associated protein 140 kDa|putative protein product of Nbla00052|putative protein product of Nbla10993', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125781131, end_position_on_the_genomic_accession=125932030, orientation='plus', exon_count=23, OMIM=609752, no_hdr0=''),
  105377785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377785, CurrentID=0, Status='live', Symbol='LOC105377785', Aliases='', description='uncharacterized LOC105377785', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2727412, end_position_on_the_genomic_accession=2834810, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  729682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729682, CurrentID=0, Status='live', Symbol='KRT17P3', Aliases='', description='keratin 17 pseudogene 3', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30567760, end_position_on_the_genomic_accession=30571841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033822, CurrentID=0, Status='live', Symbol='SNORD115-48', Aliases='HBII-52-48', description='small nucleolar RNA, C/D box 115-48', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25269783, end_position_on_the_genomic_accession=25269858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060553, CurrentID=0, Status='live', Symbol='LOC101060553', Aliases='', description='uncharacterized LOC101060553', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111148197, end_position_on_the_genomic_accession=111153577, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100294336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294336, CurrentID=0, Status='live', Symbol='LOC100294336', Aliases='', description='dead end protein homolog 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100294341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294341, CurrentID=0, Status='live', Symbol='LOC100294341', Aliases='', description='ADP-ribosylation factor-like protein 17', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286410, CurrentID=0, Status='live', Symbol='ATP11C', Aliases='ATPIG, ATPIQ', description='ATPase, class VI, type 11C', other_designations='P4-ATPase flippase complex alpha subunit ATP11C|probable phospholipid-transporting ATPase IG', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139726346, end_position_on_the_genomic_accession=139945234, orientation='minus', exon_count=38, OMIM=300516, no_hdr0=''),
  286411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286411, CurrentID=0, Status='live', Symbol='LINC00632', Aliases='', description='long intergenic non-protein coding RNA 632', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140709759, end_position_on_the_genomic_accession=140772677, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  155340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155340, CurrentID=0, Status='live', Symbol='OR7E136P', Aliases='OR7E139P, OR7E147P', description='olfactory receptor family 7 subfamily E member 136 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 136 pseudogene|olfactory receptor, family 7, subfamily E, member 139 pseudogene|olfactory receptor, family 7, subfamily E, member 147 pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6866656, end_position_on_the_genomic_accession=6867622, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9537, CurrentID=0, Status='live', Symbol='TP53I11', Aliases='PIG11', description='tumor protein p53 inducible protein 11', other_designations='p53-induced gene 11 protein', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44932348, end_position_on_the_genomic_accession=44951306, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105377786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377786, CurrentID=0, Status='live', Symbol='LOC105377786', Aliases='', description='uncharacterized LOC105377786', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2914186, end_position_on_the_genomic_accession=2919605, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929815, CurrentID=0, Status='live', Symbol='LOC101929815', Aliases='', description='D(1B) dopamine receptor-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143450185, end_position_on_the_genomic_accession=143479561, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  286423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286423, CurrentID=0, Status='live', Symbol='MRRFP1', Aliases='MRRFP', description='mitochondrial ribosome recycling factor pseudogene 1', other_designations='MRRF pseudogene|novel protein similar to mitochondrial ribosome recycling factor isoform 1 (LOC286423)', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123116107, end_position_on_the_genomic_accession=123117784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100294362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294362, CurrentID=0, Status='live', Symbol='LOC100294362', Aliases='', description='uncharacterized LOC100294362', other_designations='CTD-2047H16.4', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80351830, end_position_on_the_genomic_accession=80415168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  286430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286430, CurrentID=0, Status='live', Symbol='NLRP2P', Aliases='CLRX.1, NALP2P, NOD24, POP4', description='NLR family, pyrin domain containing 2 pseudogene', other_designations='NLR family and pyrin domain-containing 2 pseudogene|NOD24 pseudogene|Pyrin domain-containing protein 2-like protein POP4|Pyrin-only protein 4', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57676818, end_position_on_the_genomic_accession=57680280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100885797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885797, CurrentID=0, Status='live', Symbol='BPY2DP', Aliases='', description='basic charge, Y-linked, 2D, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7926089, end_position_on_the_genomic_accession=7948999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286436, CurrentID=0, Status='live', Symbol='H2BFM', Aliases='', description='H2B histone family member M', other_designations='H2B histone family, member M|H2B/s|histone H2B.s', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104039909, end_position_on_the_genomic_accession=104042454, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  286437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286437, CurrentID=0, Status='live', Symbol='LOC286437', Aliases='', description='uncharacterized LOC286437', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104122305, end_position_on_the_genomic_accession=104124632, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375282, CurrentID=0, Status='live', Symbol='LOC105375282', Aliases='', description='uncharacterized LOC105375282', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53348024, end_position_on_the_genomic_accession=53418890, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  155368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155368, CurrentID=0, Status='live', Symbol='WBSCR27', Aliases='', description='Williams Beuren syndrome chromosome region 27', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73834590, end_position_on_the_genomic_accession=73842525, orientation='minus', exon_count=7, OMIM=612546, no_hdr0=''),
  155370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155370, CurrentID=0, Status='live', Symbol='SBDSP1', Aliases='SBDSP', description='Shwachman-Bodian-Diamond syndrome pseudogene 1', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72829373, end_position_on_the_genomic_accession=72837399, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  286444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286444, CurrentID=0, Status='live', Symbol='RPS2P55', Aliases='RPS2_29_1773', description='ribosomal protein S2 pseudogene 55', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40934939, end_position_on_the_genomic_accession=40935887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  548593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548593, CurrentID=0, Status='live', Symbol='SLX1A', Aliases='GIYD1', description='SLX1 homolog A, structure-specific endonuclease subunit', other_designations='GIY-YIG domain-containing protein 1|SLX1 structure-specific endonuclease subunit homolog A|SLX1B', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30193843, end_position_on_the_genomic_accession=30197566, orientation='plus', exon_count=6, OMIM=615822, no_hdr0=''),
  548594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548594, CurrentID=0, Status='live', Symbol='KIR3DP1', Aliases='CD158c, KIR2DS6, KIR3DS2P, KIR48, KIRX', description='killer cell immunoglobulin-like receptor, three domains, pseudogene 1', other_designations='killer-cell Ig-like receptor pseudogene', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54786362, end_position_on_the_genomic_accession=54790418, orientation='plus', exon_count=0, OMIM=610604, no_hdr0=''),
  286451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286451, CurrentID=0, Status='live', Symbol='YIPF6', Aliases='FinGER6', description='Yip1 domain family member 6', other_designations='YIP1 family member 6|Yip1 domain family, member 6', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68498374, end_position_on_the_genomic_accession=68572330, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  548596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548596, CurrentID=0, Status='live', Symbol='CKMT1A', Aliases='CKMT1', description='creatine kinase, mitochondrial 1A', other_designations='U-MtCK|acidic-type mitochondrial creatine kinase|creatine kinase, mitochondrial 1 (ubiquitous)|mia-CK|ubiquitous mitochondrial creatine kinase', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43692886, end_position_on_the_genomic_accession=43699222, orientation='plus', exon_count=10, OMIM=613415, no_hdr0=''),
  286453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286453, CurrentID=0, Status='live', Symbol='LOC286453', Aliases='', description='chondroitin sulfate N-acetylgalactosaminyltransferase 2 pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100956593, end_position_on_the_genomic_accession=100957536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  155382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155382, CurrentID=0, Status='live', Symbol='VPS37D', Aliases='WBSCR24', description='vacuolar protein sorting 37 homolog D (S. cerevisiae)', other_designations='ESCRT-I complex subunit VPS37D|Williams Beuren syndrome chromosome region 24|Williams-Beuren syndrome critical region protein 24|vacuolar protein sorting 37D|williams-Beuren syndrome chromosomal region 24 protein|williams-Beuren syndrome region protein 24', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73667844, end_position_on_the_genomic_accession=73672110, orientation='plus', exon_count=4, OMIM=610039, no_hdr0=''),
  100885801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885801, CurrentID=0, Status='live', Symbol='KTCN5', Aliases='', description='Keratoconus 5', other_designations='', map_location='5q14.1-q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614622, no_hdr0=''),
  286456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286456, CurrentID=0, Status='live', Symbol='LOC286456', Aliases='', description='NGFI-A binding protein 1 (EGR1 binding protein 1) pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151226232, end_position_on_the_genomic_accession=151227715, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286460, CurrentID=0, Status='live', Symbol='FAM3C2', Aliases='', description='family with sequence similarity 3 member C2 (pseudogene)', other_designations='family with sequence similarity 3, member C2 (pseudogene)', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23075586, end_position_on_the_genomic_accession=23077412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286463, CurrentID=0, Status='live', Symbol='LOC286463', Aliases='', description='ferritin, light polypeptide pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  286464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286464, CurrentID=0, Status='live', Symbol='CFAP47', Aliases='CHDC2, CXorf22, CXorf30, CXorf59', description='cilia and flagella associated protein 47', other_designations='calponin homology domain-containing protein 2', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35919734, end_position_on_the_genomic_accession=36385319, orientation='plus', exon_count=64, OMIM='', no_hdr0=''),
  286467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286467, CurrentID=0, Status='live', Symbol='FIRRE', Aliases='LINC01200', description='firre intergenic repeating RNA element', other_designations='functional intergenic repeating RNA element', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131702650, end_position_on_the_genomic_accession=131830643, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  155400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155400, CurrentID=0, Status='live', Symbol='NSUN5P1', Aliases='NSUN5B, WBSCR20B', description='NOP2/Sun RNA methyltransferase family member 5 pseudogene 1', other_designations='NOL1/NOP2/Sun domain family, member 5B|NOP2/Sun domain family, member 5 pseudogene 1|NOP2/Sun domain family, member 5B, pseudogene|Williams Beuren syndrome chromosome region 20B|Williams-Beuren Syndrome critical region protein 20 copy B', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75410322, end_position_on_the_genomic_accession=75416792, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929817, CurrentID=0, Status='live', Symbol='LOC101929817', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61546, end_position_on_the_genomic_accession=63558, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286480, CurrentID=0, Status='live', Symbol='UBE2E4P', Aliases='UbcM2', description='ubiquitin conjugating enzyme E2E 4 pseudogene', other_designations='ubiquitin-conjugating enzyme E2E 3 (UBC4/5 homolog, yeast) pseudogene|ubiquitin-conjugating enzyme E2E 4 pseudogene|ubiquitin-conjugating enzyme UbcM2', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14244265, end_position_on_the_genomic_accession=14245423, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060569, CurrentID=0, Status='live', Symbol='LOC101060569', Aliases='', description='golgin subfamily A member 6-like protein 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30515862, end_position_on_the_genomic_accession=30518467, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  548637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548637, CurrentID=0, Status='live', Symbol='DUTP1', Aliases='DUTL', description='deoxyuridine triphosphatase pseudogene 1', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125310898, end_position_on_the_genomic_accession=125311344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286494, CurrentID=0, Status='live', Symbol='BRAFP1', Aliases='BRAF2, BRAFPS2', description='BRAF pseudogene 1', other_designations='murine sarcoma viral (v-raf) oncogene homolog B1 oncogene|v-raf murine sarcoma viral oncogene homolog B pseudogene 1|v-raf murine sarcoma viral oncogene homolog B pseudogene 2|v-raf murine sarcoma viral oncogene homolog B1 pseudogene 1|v-raf murine sarcoma viral oncogene homolog B1 pseudogene 2', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75582654, end_position_on_the_genomic_accession=75586009, orientation='plus', exon_count=0, OMIM=300956, no_hdr0=''),
  286495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286495, CurrentID=0, Status='live', Symbol='TTC3P1', Aliases='RNF105L, TTC3L', description='tetratricopeptide repeat domain 3 pseudogene 1', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75740538, end_position_on_the_genomic_accession=75743079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286499, CurrentID=0, Status='live', Symbol='FAM133A', Aliases='CT115', description='family with sequence similarity 133 member A', other_designations='cancer/testis antigen 115|family with sequence similarity 133, member A', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93673736, end_position_on_the_genomic_accession=93712274, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  548644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548644, CurrentID=0, Status='live', Symbol='POLR2J3', Aliases='POLR2J2, RPB11b1, RPB11b2', description='polymerase (RNA) II (DNA directed) polypeptide J3', other_designations='DNA-directed RNA polymerase II subunit 11|DNA-directed RNA polymerase II subunit J3|RNA polymerase II subunit B11-b2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102521171, end_position_on_the_genomic_accession=102572660, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  548645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=548645, CurrentID=0, Status='live', Symbol='DNAJC25', Aliases='bA16L21.2.1', description='DnaJ heat shock protein family (Hsp40) member C25', other_designations='DnaJ (Hsp40) homolog, subfamily C , member 25|DnaJ-like protein', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111631352, end_position_on_the_genomic_accession=111654351, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  155435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155435, CurrentID=0, Status='live', Symbol='RBM33', Aliases='PRR8', description='RNA binding motif protein 33', other_designations='RNA-binding motif protein 33|hypothetical protein MGC20460|proline rich 8|proline-rich protein 8', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155644494, end_position_on_the_genomic_accession=155781485, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  286512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286512, CurrentID=0, Status='live', Symbol='EIF4A2P4', Aliases='', description='eukaryotic translation initiation factor 4A2 pseudogene 4', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52832643, end_position_on_the_genomic_accession=52833933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286514, CurrentID=0, Status='live', Symbol='MAGEB18', Aliases='', description='MAGE family member B18', other_designations='MAGE-B18 antigen|melanoma antigen family B, 18|melanoma antigen family B18', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26138343, end_position_on_the_genomic_accession=26140736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286527, CurrentID=0, Status='live', Symbol='TMSB15B', Aliases='Tbeta15b', description='thymosin beta 15B', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103962627, end_position_on_the_genomic_accession=103965990, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  286528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286528, CurrentID=0, Status='live', Symbol='SSU72P1', Aliases='', description='SSU72 pseudogene 1', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 1', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114390674, end_position_on_the_genomic_accession=114391255, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286530, CurrentID=0, Status='live', Symbol='P2RY8', Aliases='P2Y8', description='purinergic receptor P2Y, G-protein coupled, 8', other_designations='G-protein coupled purinergic receptor P2Y8', map_location='Xp22.33; Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1462572, end_position_on_the_genomic_accession=1537506, orientation='minus', exon_count=6, OMIM=300525, no_hdr0=''),
  155465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=155465, CurrentID=0, Status='live', Symbol='AGR3', Aliases='AG-3, AG3, BCMP11, HAG3, PDIA18, hAG-3', description='anterior gradient 3, protein disulphide isomerase family member', other_designations='anterior gradient homolog 3|anterior gradient protein 3 homolog|breast cancer membrane protein 11|protein disulfide isomerase family A, member 18', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16854728, end_position_on_the_genomic_accession=16881989, orientation='minus', exon_count=10, OMIM=609482, no_hdr0=''),
  4066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4066, CurrentID=0, Status='live', Symbol='LYL1', Aliases='bHLHa18', description='lymphoblastic leukemia associated hematopoiesis regulator 1', other_designations='class A basic helix-loop-helix protein 18|lymphoblastic leukemia derived sequence 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13097188, end_position_on_the_genomic_accession=13103160, orientation='minus', exon_count=6, OMIM=151440, no_hdr0=''),
  286544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286544, CurrentID=0, Status='live', Symbol='MXRA5Y', Aliases='ADLICANP, MXRA5P, MXRA5P1', description='matrix-remodelling associated 5, Y-linked, pseudogene', other_designations='adlican pseudogene|matrix-remodelling associated 5 pseudogene 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11957177, end_position_on_the_genomic_accession=11988583, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286550, CurrentID=0, Status='live', Symbol='CYCSP46', Aliases='HCP46, HCP47', description='cytochrome c, somatic pseudogene 46', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14940689, end_position_on_the_genomic_accession=14941749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373426, CurrentID=0, Status='live', Symbol='LOC105373426', Aliases='', description='uncharacterized LOC105373426', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10534485, end_position_on_the_genomic_accession=10548676, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  286554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286554, CurrentID=0, Status='live', Symbol='BCORP1', Aliases='BCORL2', description='BCL6 corepressor pseudogene 1', other_designations='BCL-6 corepressor pseudogene 1|BCL-6 corepressor-like protein 2|BCL6 co-repressor pseudogene|BCL6 co-repressor-like 2|BCL6 corepressor-like 2|BCoR-like protein 2|putative BCoR-like protein 2', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19455431, end_position_on_the_genomic_accession=19503153, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  286555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286555, CurrentID=0, Status='live', Symbol='TBL1YP1', Aliases='TBL1YP', description='transducin (beta)-like 1, Y-linked pseudogene 1', other_designations='transducin (beta)-like 1Y-linked pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20727255, end_position_on_the_genomic_accession=20742901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286556, CurrentID=0, Status='live', Symbol='CDY3P', Aliases='', description='chromodomain protein, Y-linked 3 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9165228, end_position_on_the_genomic_accession=9168027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286557, CurrentID=0, Status='live', Symbol='RBMY1A3P', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member A3 pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 1, member A3 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9317061, end_position_on_the_genomic_accession=9322874, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101060581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060581, CurrentID=0, Status='live', Symbol='LOC101060581', Aliases='', description='arf-GAP with GTPase, ANK repeat and PH domain-containing protein 5-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271302, CurrentID=0, Status='live', Symbol='RPL27P12', Aliases='RPL27_4_1293', description='ribosomal protein L27 pseudogene 12', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123721218, end_position_on_the_genomic_accession=123721642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100689211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689211, CurrentID=0, Status='live', Symbol='HYSP3', Aliases='', description='Hypospadias 3, autosomal', other_designations='', map_location='7q32.2-q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146450, no_hdr0=''),
  286568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286568, CurrentID=0, Status='live', Symbol='RPL41P6', Aliases='', description='ribosomal protein L41 pseudogene 6', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25307660, end_position_on_the_genomic_accession=25308202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286570, CurrentID=0, Status='live', Symbol='RPL41P7', Aliases='', description='ribosomal protein L41 pseudogene 7', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24361936, end_position_on_the_genomic_accession=24362478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286573, CurrentID=0, Status='live', Symbol='TPTE2P4', Aliases='TPTEP', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 4', other_designations='transmembrane phosphatase with tensin homology pseudogene', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26563992, end_position_on_the_genomic_accession=26579868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100689213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689213, CurrentID=0, Status='live', Symbol='GFND1', Aliases='', description='Glomerulopathy with fibronectin deposits 1', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=137950, no_hdr0=''),
  101060584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060584, CurrentID=0, Status='live', Symbol='LOC101060584', Aliases='', description='histone H1.10-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100532742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532742, CurrentID=0, Status='live', Symbol='PIR-FIGF', Aliases='', description='PIR-FIGF readthrough', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15345591, end_position_on_the_genomic_accession=15491309, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105373427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373427, CurrentID=0, Status='live', Symbol='LOC105373427', Aliases='', description='uncharacterized LOC105373427', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10709200, end_position_on_the_genomic_accession=10715339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100689216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689216, CurrentID=0, Status='live', Symbol='NRCLP6', Aliases='', description='Narcolepsy 6', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614223, no_hdr0=''),
  101060588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060588, CurrentID=0, Status='live', Symbol='LOC101060588', Aliases='', description='uncharacterized LOC101060588', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373428, CurrentID=0, Status='live', Symbol='LOC105373428', Aliases='', description='uncharacterized LOC105373428', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10768709, end_position_on_the_genomic_accession=10772525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4080, CurrentID=0, Status='live', Symbol='MAA', Aliases='LZMS, MCOPS1', description='microphthalmia or anophthalmia and associated anomalies', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309800, no_hdr0=''),
  135154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135154, CurrentID=0, Status='live', Symbol='SDHAF4', Aliases='C6orf57, Sdh8', description='succinate dehydrogenase complex assembly factor 4', other_designations='SDH assembly factor 4|UPF0369 protein C6orf57', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70566922, end_position_on_the_genomic_accession=70588903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373429, CurrentID=0, Status='live', Symbol='LOC105373429', Aliases='', description='uncharacterized LOC105373429', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11406207, end_position_on_the_genomic_accession=11426386, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100288763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288763, CurrentID=0, Status='live', Symbol='LOC100288763', Aliases='', description='KIAA1191 pseudogene', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52942139, end_position_on_the_genomic_accession=52943006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060596, CurrentID=0, Status='live', Symbol='LOC101060596', Aliases='', description='putative uncharacterized protein LOC641298', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533177, CurrentID=0, Status='live', Symbol='KRTAP29-1', Aliases='KAP29.2', description='keratin associated protein 29-1', other_designations='Keratin-associated protein 29.2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41301826, end_position_on_the_genomic_accession=41302851, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286676, CurrentID=0, Status='live', Symbol='ILDR1', Aliases="DFNB42alpha, ILDR1alpha', ILDR1beta, ILDR1", description='immunoglobulin-like domain containing receptor 1', other_designations='immunoglobulin-like domain-containing receptor 1 alpha|immunoglobulin-like domain-containing receptor 1 beta', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121987323, end_position_on_the_genomic_accession=122060516, orientation='minus', exon_count=9, OMIM=609739, no_hdr0=''),
  100419700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419700, CurrentID=0, Status='live', Symbol='LOC100419700', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63887379, end_position_on_the_genomic_accession=63888598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060604, CurrentID=0, Status='live', Symbol='LOC101060604', Aliases='', description='putative L-type amino acid transporter 1-like protein IMAA', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533178, CurrentID=0, Status='live', Symbol='PRORY', Aliases='CYorf17', description='proline rich, Y-linked', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21382961, end_position_on_the_genomic_accession=21454622, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646767, CurrentID=0, Status='live', Symbol='LOC646767', Aliases='', description='uncharacterized LOC646767', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=29689971, end_position_on_the_genomic_accession=29694093, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128591, CurrentID=0, Status='live', Symbol='LOC100128591', Aliases='', description='uncharacterized LOC100128591', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=44393987, end_position_on_the_genomic_accession=44394907, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100533179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533179, CurrentID=0, Status='live', Symbol='UBE2F-SCLY', Aliases='', description='UBE2F-SCLY readthrough (NMD candidate)', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237966945, end_position_on_the_genomic_accession=238099413, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  101867536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101867536, CurrentID=0, Status='live', Symbol='PRNCR1', Aliases='CARLo-3, PCAT8', description='prostate cancer associated non-coding RNA 1', other_designations='prostate cancer associated transcript 8 (non-protein coding)|prostate cancer non-coding RNA 1', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127079874, end_position_on_the_genomic_accession=127092595, orientation='plus', exon_count=1, OMIM=615452, no_hdr0=''),
  153577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153577, CurrentID=0, Status='live', Symbol='LOC153577', Aliases='', description='uncharacterized LOC153577', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377797, CurrentID=0, Status='live', Symbol='LOC105377797', Aliases='', description='uncharacterized LOC105377797', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6058166, end_position_on_the_genomic_accession=6066137, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  286749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286749, CurrentID=0, Status='live', Symbol='STON1-GTF2A1L', Aliases='SALF', description='STON1-GTF2A1L readthrough', other_designations='stoned B/TFIIA-alpha/beta-like factor', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48529925, end_position_on_the_genomic_accession=48776517, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  286751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286751, CurrentID=0, Status='live', Symbol='OTSC4', Aliases='', description='otosclerosis 4', other_designations='', map_location='16q21-q23.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611571, no_hdr0=''),
  286753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286753, CurrentID=0, Status='live', Symbol='TUSC5', Aliases='DSPB1, IFITMD3, LOST1', description='tumor suppressor candidate 5', other_designations='dispanin subfamily B member 1|interferon induced transmembrane protein domain containing 3|interferon-induced transmembrane domain-containing protein D3|located at seventeen p thirteen point three 1|protein located at seventeen-p-thirteen point three 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1279663, end_position_on_the_genomic_accession=1300987, orientation='plus', exon_count=3, OMIM=612211, no_hdr0=''),
  286754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286754, CurrentID=0, Status='live', Symbol='LEFTY3', Aliases='', description='left-right determination factor 2 pseudogene', other_designations='', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225803157, end_position_on_the_genomic_accession=225803453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  286756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286756, CurrentID=0, Status='live', Symbol='SCAMPER', Aliases='', description='sphingolipid Ca2+ release mediating protein of endoplasmic reticulum', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100294713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294713, CurrentID=0, Status='live', Symbol='LOC100294713', Aliases='', description='IQ motif containing F1 pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=44248381, end_position_on_the_genomic_accession=44249013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100294715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294715, CurrentID=0, Status='live', Symbol='BOP', Aliases='', description='Polymicrogyria, bilateral occipital', other_designations='', map_location='6q16.3-q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612691, no_hdr0=''),
  100294716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294716, CurrentID=0, Status='live', Symbol='MYP15', Aliases='', description='myopia 15', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612717, no_hdr0=''),
  100294717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294717, CurrentID=0, Status='live', Symbol='BMND13', Aliases='', description='Bone mineral density QTL 13', other_designations='', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612727, no_hdr0=''),
  100294718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294718, CurrentID=0, Status='live', Symbol='BMND14', Aliases='', description='Bone mineral density QTL 14', other_designations='', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612728, no_hdr0=''),
  100294719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294719, CurrentID=0, Status='live', Symbol='LBMQTL1', Aliases='', description='Lean body mass QTL 1', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612729, no_hdr0=''),
  100294720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294720, CurrentID=0, Status='live', Symbol='NHEG1', Aliases='', description='neuroblastoma highly expressed 1', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136982159, end_position_on_the_genomic_accession=136993231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100533181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533181, CurrentID=0, Status='live', Symbol='FXYD6-FXYD2', Aliases='', description='FXYD6-FXYD2 readthrough', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117820075, end_position_on_the_genomic_accession=117877031, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106480673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480673, CurrentID=0, Status='live', Symbol='RNU2-61P', Aliases='', description='RNA, U2 small nuclear 61, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271310, CurrentID=0, Status='live', Symbol='RPL35AP8', Aliases='RPL35A_1_367', description='ribosomal protein L35a pseudogene 8', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42469640, end_position_on_the_genomic_accession=42469882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101867602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101867602, CurrentID=0, Status='live', Symbol='CTRCT29', Aliases='', description='Cataract 29, coralliform', other_designations='', map_location='2pter-p24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=115800, no_hdr0=''),
  646770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646770, CurrentID=0, Status='live', Symbol='TRIM77BP', Aliases='', description='tripartite motif containing 77B, pseudogene', other_designations='tripartite motif containing 43 pseudogene|tripartite motif containing 77, pseudogene pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49117010, end_position_on_the_genomic_accession=49124570, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25957, CurrentID=0, Status='live', Symbol='PNISR', Aliases='C6orf111, HSPC306, SFRS18, SRrp130, bA98I9.2', description='PNN-interacting serine/arginine-rich protein', other_designations='SR rich protein|SR-related protein|SR-rich protein|serine-arginine-rich-splicing regulatory protein 130|serine/arginine-rich-splicing regulatory protein 130|splicing factor, arginine/serine-rich 130|splicing factor, arginine/serine-rich 18', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99398655, end_position_on_the_genomic_accession=99425342, orientation='minus', exon_count=13, OMIM=616653, no_hdr0=''),
  286826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286826, CurrentID=0, Status='live', Symbol='LIN9', Aliases='BARA, BARPsv, Lin-9, TGS, TGS1, TGS2', description='lin-9 DREAM MuvB core complex component', other_designations='TUDOR gene similar protein|beta subunit-associated regulator of apoptosis|lin-9 homolog|pRB-associated protein|rb related pathway actor|type I interferon receptor beta chain-associated protein', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226231149, end_position_on_the_genomic_accession=226310099, orientation='minus', exon_count=19, OMIM=609375, no_hdr0=''),
  286827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286827, CurrentID=0, Status='live', Symbol='TRIM59', Aliases='IFT80L, MRF1, RNF104, TRIM57, TSBF1', description='tripartite motif containing 59', other_designations='RING finger protein 104|tripartite motif-containing 57|tripartite motif-containing 59|tumor suppressor TSBF-1|tumor suppressor TSBF1', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160435503, end_position_on_the_genomic_accession=160449838, orientation='minus', exon_count=3, OMIM=616148, no_hdr0=''),
  286828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286828, CurrentID=0, Status='live', Symbol='CSN1S2AP', Aliases='CSN1S2A', description='casein alpha s2-like A, pseudogene', other_designations='alpha-S2-like casein A', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70067386, end_position_on_the_genomic_accession=70085273, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  402359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402359, CurrentID=0, Status='live', Symbol='KHSRPP1', Aliases='', description='KH-type splicing regulatory protein pseudogene 1', other_designations='KH-type splicing regulatory protein (FUSE binding protein 2) pseudogene', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21695174, end_position_on_the_genomic_accession=21700073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373437, CurrentID=0, Status='live', Symbol='LOC105373437', Aliases='', description='uncharacterized LOC105373437', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13357301, end_position_on_the_genomic_accession=13403233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533184, CurrentID=0, Status='live', Symbol='ARHGAP19-SLIT1', Aliases='', description='ARHGAP19-SLIT1 readthrough (NMD candidate)', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97153042, end_position_on_the_genomic_accession=97292673, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  286887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286887, CurrentID=0, Status='live', Symbol='KRT6C', Aliases='K6E, KRT6E, PPKNEFD', description='keratin 6C, type II', other_designations='CK-6C|CK-6E|K6C|cytokeratin-6C|cytokeratin-6E|keratin 6E|keratin K6h|keratin-6C|type-II keratin Kb12', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52468516, end_position_on_the_genomic_accession=52473785, orientation='minus', exon_count=9, OMIM=612315, no_hdr0=''),
  100288771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288771, CurrentID=0, Status='live', Symbol='LOC100288771', Aliases='', description='prefoldin subunit 1 pseudogene', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15743232, end_position_on_the_genomic_accession=15743582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271313, CurrentID=0, Status='live', Symbol='RPL35AP10', Aliases='RPL35A_3_429', description='ribosomal protein L35a pseudogene 10', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160342687, end_position_on_the_genomic_accession=160343015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060637, CurrentID=0, Status='live', Symbol='LOC101060637', Aliases='', description='nuclear ubiquitous casein and cyclin-dependent kinase substrate 1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402360, CurrentID=0, Status='live', Symbol='NOP56P2', Aliases='', description='NOP56 ribonucleoprotein pseudogene 2', other_designations='NOP56 ribonucleoprotein homolog pseudogene', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23681871, end_position_on_the_genomic_accession=23682762, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130338, CurrentID=0, Status='live', Symbol='PABPC1P9', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 9', other_designations='poly(A) binding protein, cytoplasmic, pseudogene', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16424530, end_position_on_the_genomic_accession=16426477, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373438, CurrentID=0, Status='live', Symbol='LOC105373438', Aliases='', description='uncharacterized LOC105373438', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13609062, end_position_on_the_genomic_accession=13758396, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271314, CurrentID=0, Status='live', Symbol='RPL35AP11', Aliases='RPL35A_4_500', description='ribosomal protein L35a pseudogene 11', other_designations='', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94369771, end_position_on_the_genomic_accession=94370200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271315, CurrentID=0, Status='live', Symbol='RPL35AP14', Aliases='RPL35A_4_569', description='ribosomal protein L35a pseudogene 14', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62575975, end_position_on_the_genomic_accession=62576271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480841, CurrentID=0, Status='live', Symbol='RN7SKP27', Aliases='', description='RNA, 7SK small nuclear pseudogene 27', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271236, CurrentID=0, Status='live', Symbol='RPS21P5', Aliases='RPS21_3_1008', description='ribosomal protein S21 pseudogene 5', other_designations='', map_location='9q31.3-q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110727449, end_position_on_the_genomic_accession=110727679, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  286967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=286967, CurrentID=0, Status='live', Symbol='FAM223B', Aliases='CXorf52, CXorf52B, LINC00204B, NCRNA00204B', description='family with sequence similarity 223 member B (non-protein coding)', other_designations='family with sequence similarity 223, member B (non-protein coding)|long intergenic non-protein coding RNA 204B|non-protein coding RNA 204B', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154632470, end_position_on_the_genomic_accession=154633182, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271316, CurrentID=0, Status='live', Symbol='RPL35AP12', Aliases='RPL35A_5_531', description='ribosomal protein L35a pseudogene 12', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163382179, end_position_on_the_genomic_accession=163382516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9556, CurrentID=0, Status='live', Symbol='C14orf2', Aliases='MP68, PLPM', description='chromosome 14 open reading frame 2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103912288, end_position_on_the_genomic_accession=103921566, orientation='minus', exon_count=5, OMIM=604573, no_hdr0=''),
  105373441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373441, CurrentID=0, Status='live', Symbol='LOC105373441', Aliases='', description='uncharacterized LOC105373441', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15781176, end_position_on_the_genomic_accession=15786575, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271317, CurrentID=0, Status='live', Symbol='RPL35AP15', Aliases='RPL35A_6_605', description='ribosomal protein L35a pseudogene 15', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117301107, end_position_on_the_genomic_accession=117301442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  287015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=287015, CurrentID=0, Status='live', Symbol='TRIM42', Aliases='PPP1R40, T4A1', description='tripartite motif containing 42', other_designations='protein phosphatase 1, regulatory subunit 40|tripartite motif-containing 42', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140678024, end_position_on_the_genomic_accession=140701150, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100128593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128593, CurrentID=0, Status='live', Symbol='LOC100128593', Aliases='', description='uncharacterized LOC100128593', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136746161, end_position_on_the_genomic_accession=136749911, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271318, CurrentID=0, Status='live', Symbol='RPL35AP16', Aliases='RPL35A_6_624', description='ribosomal protein L35a pseudogene 16', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146138134, end_position_on_the_genomic_accession=146138381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533618, CurrentID=0, Status='live', Symbol='LOC100533618', Aliases='', description='ubinuclein 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44467449, end_position_on_the_genomic_accession=44471620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419712, CurrentID=0, Status='live', Symbol='LOC100419712', Aliases='', description='zinc finger protein 12 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=40445304, end_position_on_the_genomic_accession=40447141, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131654, CurrentID=0, Status='live', Symbol='LOC100131654', Aliases='', description='uncharacterized LOC100131654', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=87736900, end_position_on_the_genomic_accession=87737541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373443, CurrentID=0, Status='live', Symbol='LOC105373443', Aliases='', description='uncharacterized LOC105373443', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16201590, end_position_on_the_genomic_accession=16204940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100689296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689296, CurrentID=0, Status='live', Symbol='CNN2P11', Aliases='', description='calponin 2 pseudogene 11', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94725476, end_position_on_the_genomic_accession=94725912, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4156, CurrentID=0, Status='live', Symbol='MBS1', Aliases='MBS', description='Moebius syndrome 1', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=157900, no_hdr0=''),
  105372094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372094, CurrentID=0, Status='live', Symbol='LOC105372094', Aliases='', description='uncharacterized LOC105372094', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46407756, end_position_on_the_genomic_accession=46415426, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  389273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389273, CurrentID=0, Status='live', Symbol='LOC389273', Aliases='', description='uncharacterized LOC389273', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10504996, end_position_on_the_genomic_accession=10509256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271321, CurrentID=0, Status='live', Symbol='RPL35AP18', Aliases='RPL35A_8_717', description='ribosomal protein L35a pseudogene 18', other_designations='', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79964215, end_position_on_the_genomic_accession=79964522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376496, CurrentID=0, Status='live', Symbol='LOC105376496', Aliases='', description='uncharacterized LOC105376496', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36917276, end_position_on_the_genomic_accession=36943132, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4531, CurrentID=0, Status='live', Symbol='MTL3P', Aliases='MTL3', description='metallothionein-like 3, pseudogene', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68528030, end_position_on_the_genomic_accession=68528119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060684, CurrentID=0, Status='live', Symbol='NBPF26', Aliases='', description='neuroblastoma breakpoint family member 26', other_designations='neuroblastoma breakpoint family, member 26', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120805277, end_position_on_the_genomic_accession=120842217, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  100271323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271323, CurrentID=0, Status='live', Symbol='RPL36AP13', Aliases='RPL36A_3_200', description='ribosomal protein L36a pseudogene 13', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24334512, end_position_on_the_genomic_accession=24334783, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533195, CurrentID=0, Status='live', Symbol='FAM24B-CUZD1', Aliases='', description='FAM24B-CUZD1 readthrough', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122832155, end_position_on_the_genomic_accession=122879641, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101060691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060691, CurrentID=0, Status='live', Symbol='NUTM2B-AS1', Aliases='', description='NUTM2B antisense RNA 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79691500, end_position_on_the_genomic_accession=79826594, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271324, CurrentID=0, Status='live', Symbol='RPL36AP14', Aliases='RPL36A_3_214', description='ribosomal protein L36a pseudogene 14', other_designations='', map_location='2p21-p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46256860, end_position_on_the_genomic_accession=46257175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441087, CurrentID=0, Status='live', Symbol='LOC441087', Aliases='', description='uncharacterized LOC441087', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75717506, end_position_on_the_genomic_accession=75733324, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101060696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060696, CurrentID=0, Status='live', Symbol='LOC101060696', Aliases='', description='cTAGE family member 4-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419720, CurrentID=0, Status='live', Symbol='LOC100419720', Aliases='', description='RUN domain containing 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151380340, end_position_on_the_genomic_accession=151382382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131662, CurrentID=0, Status='live', Symbol='LOC100131662', Aliases='', description='uncharacterized LOC100131662', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=95691153, end_position_on_the_genomic_accession=95692215, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128595, CurrentID=0, Status='live', Symbol='HNRNPDLP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein D-like pseudogene 1', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95638357, end_position_on_the_genomic_accession=95638908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271328, CurrentID=0, Status='live', Symbol='RPL36AP18', Aliases='RPL36A_5_491', description='ribosomal protein L36a pseudogene 18', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76401224, end_position_on_the_genomic_accession=76401619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533620, CurrentID=0, Status='live', Symbol='TUBBP7', Aliases='', description='tubulin beta class I pseudogene 7', other_designations='tubulin, beta class I pseudogene 7|tubulin, beta pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96126614, end_position_on_the_genomic_accession=96127802, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402375, CurrentID=0, Status='live', Symbol='YBX1P6', Aliases='', description='Y-box binding protein 1 pseudogene 6', other_designations='Y box binding protein 1 pseudogene 6', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109532414, end_position_on_the_genomic_accession=109534471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271329, CurrentID=0, Status='live', Symbol='RPL36AP20', Aliases='RPL36A_6_630', description='ribosomal protein L36a pseudogene 20', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152495718, end_position_on_the_genomic_accession=152496017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377818, CurrentID=0, Status='live', Symbol='LOC105377818', Aliases='', description='uncharacterized LOC105377818', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80736070, end_position_on_the_genomic_accession=80764232, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060719, CurrentID=0, Status='live', Symbol='LOC101060719', Aliases='', description='uncharacterized LOC101060719', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874051, CurrentID=0, Status='live', Symbol='LZTS1-AS1', Aliases='', description='LZTS1 antisense RNA 1', other_designations='LZTS1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20275773, end_position_on_the_genomic_accession=20290458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103047893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103047893, CurrentID=0, Status='live', Symbol='LOC103047893', Aliases='', description='vomeronasal 1 receptor 49 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40715977, end_position_on_the_genomic_accession=40716711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  135293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135293, CurrentID=0, Status='live', Symbol='PM20D2', Aliases='ACY1L2, bA63L7.3', description='peptidase M20 domain containing 2', other_designations='&#946;-alanyl-lysine dipeptidase|aminoacylase 1-like 2|aminoacylase 1-like protein 2|aminoacylase-1-like protein 2', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89096753, end_position_on_the_genomic_accession=89165569, orientation='plus', exon_count=13, OMIM=615913, no_hdr0=''),
  135295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135295, CurrentID=0, Status='live', Symbol='SRSF12', Aliases='SFRS13B, SFRS19, SRrp35', description='serine/arginine-rich splicing factor 12', other_designations='35 kDa SR repressor protein|SR splicing factor 12|serine-arginine repressor protein (35 kDa)|splicing factor, arginine/serine-rich 13B|splicing factor, arginine/serine-rich 19', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89095957, end_position_on_the_genomic_accession=89118081, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100533621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533621, CurrentID=0, Status='live', Symbol='LOC100533621', Aliases='', description='fibrosin-like 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125958361, end_position_on_the_genomic_accession=125961270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929851, CurrentID=0, Status='live', Symbol='LOC101929851', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419727, CurrentID=0, Status='live', Symbol='ZNF736P10Y', Aliases='', description='zinc finger protein 736 pseudogene 10, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8418851, end_position_on_the_genomic_accession=8419923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131669, CurrentID=0, Status='live', Symbol='LOC100131669', Aliases='', description='uncharacterized LOC100131669', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.9', start_position_on_the_genomic_accession=42641802, end_position_on_the_genomic_accession=42649308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373458, CurrentID=0, Status='live', Symbol='LOC105373458', Aliases='', description='uncharacterized LOC105373458', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19431396, end_position_on_the_genomic_accession=19436404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271334, CurrentID=0, Status='live', Symbol='RPL36P15', Aliases='RPL36_6_1283', description='ribosomal protein L36 pseudogene 15', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116908312, end_position_on_the_genomic_accession=116908631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646794, CurrentID=0, Status='live', Symbol='LOC646794', Aliases='', description='SNF8, ESCRT-II complex subunit, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=228352120, end_position_on_the_genomic_accession=228352855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  375601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375601, CurrentID=0, Status='live', Symbol='OR7E31P', Aliases='OR7E114P, OR7E32P, OR7E45P, OR9-4, OST016, OST049, OST205', description='olfactory receptor family 7 subfamily E member 31 pseudogene', other_designations='olfactory receptor OR9-4 pseudogene|olfactory receptor, family 7, subfamily E, member 114 pseudogene|olfactory receptor, family 7, subfamily E, member 31 pseudogene|olfactory receptor, family 7, subfamily E, member 32 pseudogene|olfactory receptor, family 7, subfamily E, member 45 pseudogene', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90216226, end_position_on_the_genomic_accession=90216882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373459, CurrentID=0, Status='live', Symbol='LOC105373459', Aliases='', description='uncharacterized LOC105373459', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19590264, end_position_on_the_genomic_accession=19605464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480678, CurrentID=0, Status='live', Symbol='KLF2P2', Aliases='', description='Kruppel-like factor 2 pseudogene 2', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372970, CurrentID=0, Status='live', Symbol='LOC105372970', Aliases='', description='uncharacterized LOC105372970', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222501469, end_position_on_the_genomic_accession=222510315, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  283861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283861, CurrentID=0, Status='live', Symbol='LOC283861', Aliases='', description='uncharacterized LOC283861', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873640, CurrentID=0, Status='live', Symbol='RNA5SP385', Aliases='RN5S385', description='RNA, 5S ribosomal pseudogene 385', other_designations='RNA, 5S ribosomal 385', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51692752, end_position_on_the_genomic_accession=51692861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874052, CurrentID=0, Status='live', Symbol='LINC00534', Aliases='', description='long intergenic non-protein coding RNA 534', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90221488, end_position_on_the_genomic_accession=90387959, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100033391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033391, CurrentID=0, Status='live', Symbol='VN2R2P', Aliases='', description='vomeronasal 2 receptor 2 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178662168, end_position_on_the_genomic_accession=178662853, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100033392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033392, CurrentID=0, Status='live', Symbol='VN2R3P', Aliases='VN2R4P, VN2R5P, VN2R8P', description='vomeronasal 2 receptor 3 pseudogene', other_designations='vomeronasal 2 receptor 4 pseudogene|vomeronasal 2 receptor 5 pseudogene|vomeronasal 2 receptor 8 pseudogene', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39026656, end_position_on_the_genomic_accession=39039604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91624, CurrentID=0, Status='live', Symbol='NEXN', Aliases='CMH20, NELIN', description='nexilin (F actin binding protein)', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77888515, end_position_on_the_genomic_accession=77948643, orientation='plus', exon_count=13, OMIM=613121, no_hdr0=''),
  100033395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033395, CurrentID=0, Status='live', Symbol='VN2R6P', Aliases='', description='vomeronasal 2 receptor 6 pseudogene', other_designations='calcium-sensing receptor pseudogene', map_location='9p11.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41848318, end_position_on_the_genomic_accession=41859347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100033396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033396, CurrentID=0, Status='live', Symbol='VN2R7P', Aliases='', description='vomeronasal 2 receptor 7 pseudogene', other_designations='calcium-sensing receptor pseudogene', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67328114, end_position_on_the_genomic_accession=67339143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033398, CurrentID=0, Status='live', Symbol='VN2R10P', Aliases='', description='vomeronasal 2 receptor 10 pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77220394, end_position_on_the_genomic_accession=77221264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100033399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033399, CurrentID=0, Status='live', Symbol='VN2R11P', Aliases='', description='vomeronasal 2 receptor 11 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=279554, end_position_on_the_genomic_accession=280243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100033400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033400, CurrentID=0, Status='live', Symbol='VN2R12P', Aliases='', description='vomeronasal 2 receptor 12 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11620902, end_position_on_the_genomic_accession=11621224, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033401, CurrentID=0, Status='live', Symbol='VN2R13P', Aliases='', description='vomeronasal 2 receptor 13 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11836436, end_position_on_the_genomic_accession=11836671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033402, CurrentID=0, Status='live', Symbol='VN2R14P', Aliases='', description='vomeronasal 2 receptor 14 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11871622, end_position_on_the_genomic_accession=11871856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271337, CurrentID=0, Status='live', Symbol='RPL37AP2', Aliases='RPL37A_1_476', description='ribosomal protein L37a pseudogene 2', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=51978082, end_position_on_the_genomic_accession=51978372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033404, CurrentID=0, Status='live', Symbol='VN2R16P', Aliases='', description='vomeronasal 2 receptor 16 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12607832, end_position_on_the_genomic_accession=12607978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033406, CurrentID=0, Status='live', Symbol='VN2R18P', Aliases='', description='vomeronasal 2 receptor 18 pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56338218, end_position_on_the_genomic_accession=56338905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033407, CurrentID=0, Status='live', Symbol='VN2R19P', Aliases='', description='vomeronasal 2 receptor 19 pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58012586, end_position_on_the_genomic_accession=58013430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033408, CurrentID=0, Status='live', Symbol='VN2R20P', Aliases='', description='vomeronasal 2 receptor 20 pseudogene', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23910126, end_position_on_the_genomic_accession=23910356, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100033409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033409, CurrentID=0, Status='live', Symbol='OTX2P1', Aliases='OTX2P', description='orthodenticle homeobox 2 pseudogene 1', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75724171, end_position_on_the_genomic_accession=75724811, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100033411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033411, CurrentID=0, Status='live', Symbol='DUXB', Aliases='', description='double homeobox B', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75693729, end_position_on_the_genomic_accession=75701418, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100033413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033413, CurrentID=0, Status='live', Symbol='SNORD116-1', Aliases='HBII-85-1, PWCR1', description='small nucleolar RNA, C/D box 116-1', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25051476, end_position_on_the_genomic_accession=25051572, orientation='plus', exon_count=1, OMIM=605436, no_hdr0=''),
  100033414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033414, CurrentID=0, Status='live', Symbol='SNORD116-2', Aliases='HBII-85-2', description='small nucleolar RNA, C/D box 116-2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25054209, end_position_on_the_genomic_accession=25054305, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033415, CurrentID=0, Status='live', Symbol='SNORD116-3', Aliases='HBII-85-3', description='small nucleolar RNA, C/D box 116-3', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25056859, end_position_on_the_genomic_accession=25056955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033416, CurrentID=0, Status='live', Symbol='SNORD116-4', Aliases='HBII-85-4', description='small nucleolar RNA, C/D box 116-4', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25059537, end_position_on_the_genomic_accession=25059634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033417, CurrentID=0, Status='live', Symbol='SNORD116-5', Aliases='HBII-85-5', description='small nucleolar RNA, C/D box 116-5', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25062332, end_position_on_the_genomic_accession=25062428, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033418, CurrentID=0, Status='live', Symbol='SNORD116-6', Aliases='HBII-85-6', description='small nucleolar RNA, C/D box 116-6', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25065025, end_position_on_the_genomic_accession=25065122, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033419, CurrentID=0, Status='live', Symbol='SNORD116-7', Aliases='HBII-85-7', description='small nucleolar RNA, C/D box 116-7', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25067787, end_position_on_the_genomic_accession=25067883, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033420, CurrentID=0, Status='live', Symbol='SNORD116-8', Aliases='HBII-85-8', description='small nucleolar RNA, C/D box 116-8', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25070431, end_position_on_the_genomic_accession=25070527, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033421, CurrentID=0, Status='live', Symbol='SNORD116-9', Aliases='HBII-85-9', description='small nucleolar RNA, C/D box 116-9', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25073106, end_position_on_the_genomic_accession=25073202, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033422, CurrentID=0, Status='live', Symbol='SNORD116-10', Aliases='HBII-85-10', description='small nucleolar RNA, C/D box 116-10', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25074113, end_position_on_the_genomic_accession=25074216, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033423, CurrentID=0, Status='live', Symbol='SNORD116-11', Aliases='HBII-85-11', description='small nucleolar RNA, C/D box 116-11', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25075928, end_position_on_the_genomic_accession=25076021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033424, CurrentID=0, Status='live', Symbol='SNORD116-12', Aliases='HBII-85-12', description='small nucleolar RNA, C/D box 116-12', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25077050, end_position_on_the_genomic_accession=25077143, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033425, CurrentID=0, Status='live', Symbol='SNORD116-13', Aliases='HBII-85-13', description='small nucleolar RNA, C/D box 116-13', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25079057, end_position_on_the_genomic_accession=25079150, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033426, CurrentID=0, Status='live', Symbol='SNORD116-14', Aliases='HBII-85-14', description='small nucleolar RNA, C/D box 116-14', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25080141, end_position_on_the_genomic_accession=25080234, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033427, CurrentID=0, Status='live', Symbol='SNORD116-15', Aliases='HBII-85-15', description='small nucleolar RNA, C/D box 116-15', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25081286, end_position_on_the_genomic_accession=25081379, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033428, CurrentID=0, Status='live', Symbol='SNORD116-16', Aliases='HBII-85-16', description='small nucleolar RNA, C/D box 116-16', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25082767, end_position_on_the_genomic_accession=25082860, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033429, CurrentID=0, Status='live', Symbol='SNORD116-17', Aliases='HBII-85-17', description='small nucleolar RNA, C/D box 116-17', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25083587, end_position_on_the_genomic_accession=25083680, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033430, CurrentID=0, Status='live', Symbol='SNORD116-18', Aliases='HBII-85-18', description='small nucleolar RNA, C/D box 116-18', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25085384, end_position_on_the_genomic_accession=25085477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033431, CurrentID=0, Status='live', Symbol='SNORD116-20', Aliases='HBII-85-20', description='small nucleolar RNA, C/D box 116-20', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25087661, end_position_on_the_genomic_accession=25087754, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033432, CurrentID=0, Status='live', Symbol='SNORD116-21', Aliases='HBII-85-21', description='small nucleolar RNA, C/D box 116-21', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25088803, end_position_on_the_genomic_accession=25088896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033433, CurrentID=0, Status='live', Symbol='SNORD116-22', Aliases='HBII-85-22', description='small nucleolar RNA, C/D box 116-22', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25089922, end_position_on_the_genomic_accession=25090015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033434, CurrentID=0, Status='live', Symbol='SNORD116-23', Aliases='HBII-85-23', description='small nucleolar RNA, C/D box 116-23', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25091785, end_position_on_the_genomic_accession=25091878, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033435, CurrentID=0, Status='live', Symbol='SNORD116-24', Aliases='HBII-85-24', description='small nucleolar RNA, C/D box 116-24', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25094036, end_position_on_the_genomic_accession=25094129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033436, CurrentID=0, Status='live', Symbol='SNORD116-25', Aliases='HBII-85-25', description='small nucleolar RNA, C/D box 116-25', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25097662, end_position_on_the_genomic_accession=25097755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033437, CurrentID=0, Status='live', Symbol='SNORD115-2', Aliases='HBII-52-2', description='small nucleolar RNA, C/D box 115-2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25172635, end_position_on_the_genomic_accession=25172716, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033438, CurrentID=0, Status='live', Symbol='SNORD116-26', Aliases='HBII-85-26', description='small nucleolar RNA, C/D box 116-26', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25099498, end_position_on_the_genomic_accession=25099595, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033439, CurrentID=0, Status='live', Symbol='SNORD116-27', Aliases='HBII-85-27', description='small nucleolar RNA, C/D box 116-27', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25101574, end_position_on_the_genomic_accession=25101667, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033440, CurrentID=0, Status='live', Symbol='SNORD115-3', Aliases='HBII-52-3', description='small nucleolar RNA, C/D box 115-3', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25174927, end_position_on_the_genomic_accession=25175008, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033441, CurrentID=0, Status='live', Symbol='SNORD115-4', Aliases='HBII-52-4', description='small nucleolar RNA, C/D box 115-4', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25176832, end_position_on_the_genomic_accession=25176913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033442, CurrentID=0, Status='live', Symbol='SNORD115-5', Aliases='HBII-52-5', description='small nucleolar RNA, C/D box 115-5', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25178738, end_position_on_the_genomic_accession=25178819, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033443, CurrentID=0, Status='live', Symbol='SNORD115-6', Aliases='HBII-52-6', description='small nucleolar RNA, C/D box 115-6', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25180497, end_position_on_the_genomic_accession=25180578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033444, CurrentID=0, Status='live', Symbol='SNORD115-7', Aliases='HBII-52-7', description='small nucleolar RNA, C/D box 115-7', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25182385, end_position_on_the_genomic_accession=25182466, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033445, CurrentID=0, Status='live', Symbol='SNORD115-8', Aliases='HBII-52-8', description='small nucleolar RNA, C/D box 115-8', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25184306, end_position_on_the_genomic_accession=25184387, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033446, CurrentID=0, Status='live', Symbol='SNORD115-9', Aliases='HBII-52-9', description='small nucleolar RNA, C/D box 115-9', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25185631, end_position_on_the_genomic_accession=25185712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033447, CurrentID=0, Status='live', Symbol='SNORD115-10', Aliases='HBII-52-10', description='small nucleolar RNA, C/D box 115-10', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25187536, end_position_on_the_genomic_accession=25187616, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033448, CurrentID=0, Status='live', Symbol='SNORD115-11', Aliases='HBII-52-11', description='small nucleolar RNA, C/D box 115-11', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25189414, end_position_on_the_genomic_accession=25189495, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033449, CurrentID=0, Status='live', Symbol='SNORD115-12', Aliases='HBII-52-12', description='small nucleolar RNA, C/D box 115-12', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25191416, end_position_on_the_genomic_accession=25191497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033450, CurrentID=0, Status='live', Symbol='SNORD115-13', Aliases='HBII-52-13', description='small nucleolar RNA, C/D box 115-13', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25193321, end_position_on_the_genomic_accession=25193402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033451, CurrentID=0, Status='live', Symbol='SNORD115-14', Aliases='HBII-52-14', description='small nucleolar RNA, C/D box 115-14', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25194921, end_position_on_the_genomic_accession=25195001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441666, CurrentID=0, Status='live', Symbol='LOC441666', Aliases='', description='zinc finger protein 91 pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42331866, end_position_on_the_genomic_accession=42368045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100033453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033453, CurrentID=0, Status='live', Symbol='SNORD115-15', Aliases='HBII-52-15', description='small nucleolar RNA, C/D box 115-15', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25197576, end_position_on_the_genomic_accession=25197656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033454, CurrentID=0, Status='live', Symbol='SNORD115-16', Aliases='HBII-52-16', description='small nucleolar RNA, C/D box 115-16', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25199448, end_position_on_the_genomic_accession=25199529, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033455, CurrentID=0, Status='live', Symbol='SNORD115-17', Aliases='HBII-52-17', description='small nucleolar RNA, C/D box 115-17', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25201323, end_position_on_the_genomic_accession=25201404, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033456, CurrentID=0, Status='live', Symbol='SNORD115-18', Aliases='HBII-52-18', description='small nucleolar RNA, C/D box 115-18', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25203227, end_position_on_the_genomic_accession=25203308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033458, CurrentID=0, Status='live', Symbol='SNORD115-19', Aliases='HBII-52-19', description='small nucleolar RNA, C/D box 115-19', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25204357, end_position_on_the_genomic_accession=25204438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033460, CurrentID=0, Status='live', Symbol='SNORD115-20', Aliases='HBII-52-20', description='small nucleolar RNA, C/D box 115-20', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25206262, end_position_on_the_genomic_accession=25206343, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271339, CurrentID=0, Status='live', Symbol='RPL37AP6', Aliases='RPL37A_3_833', description='ribosomal protein L37a pseudogene 6', other_designations='', map_location='7q22-q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107450090, end_position_on_the_genomic_accession=107450362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373464, CurrentID=0, Status='live', Symbol='LOC105373464', Aliases='', description='uncharacterized LOC105373464', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20182376, end_position_on_the_genomic_accession=20186047, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271340, CurrentID=0, Status='live', Symbol='RPL37P20', Aliases='RPL37_6_1200', description='ribosomal protein L37 pseudogene 20', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7032108, end_position_on_the_genomic_accession=7032398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271341, CurrentID=0, Status='live', Symbol='RPL37P22', Aliases='RPL37_7_1538', description='ribosomal protein L37 pseudogene 22', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35076792, end_position_on_the_genomic_accession=35077047, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646801, CurrentID=0, Status='live', Symbol='LOC646801', Aliases='', description='pantothenate kinase 3 pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50268161, end_position_on_the_genomic_accession=50269272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377830, CurrentID=0, Status='live', Symbol='LOC105377830', Aliases='', description='uncharacterized LOC105377830', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874053, CurrentID=0, Status='live', Symbol='JRKL-AS1', Aliases='', description='JRKL antisense RNA 1', other_designations='JRKL antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96443281, end_position_on_the_genomic_accession=96506876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377336, CurrentID=0, Status='live', Symbol='LOC105377336', Aliases='', description='uncharacterized LOC105377336', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95400641, end_position_on_the_genomic_accession=95431773, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129471, CurrentID=0, Status='live', Symbol='LOC100129471', Aliases='', description='plastin 1 pseudogene', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101765347, end_position_on_the_genomic_accession=101770002, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271343, CurrentID=0, Status='live', Symbol='UBA52P8', Aliases='RPL40_3_1481', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 8', other_designations='ribosomal protein L40 pseudogene', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48334471, end_position_on_the_genomic_accession=48334854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929861, CurrentID=0, Status='live', Symbol='LOC101929861', Aliases='', description='uncharacterized LOC101929861', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7867404, end_position_on_the_genomic_accession=7884666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100033603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033603, CurrentID=0, Status='live', Symbol='SNORD115-21', Aliases='HBII-52-21', description='small nucleolar RNA, C/D box 115-21', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25208083, end_position_on_the_genomic_accession=25208163, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646804, CurrentID=0, Status='live', Symbol='LOC646804', Aliases='', description='alkylated DNA repair protein alkB homolog 8-like', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171083561, end_position_on_the_genomic_accession=171085205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372307, CurrentID=0, Status='live', Symbol='LOC105372307', Aliases='', description='uncharacterized LOC105372307', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19938672, end_position_on_the_genomic_accession=19941223, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101060796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060796, CurrentID=0, Status='live', Symbol='LOC101060796', Aliases='', description='histone H1, gonadal-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=60991902, end_position_on_the_genomic_accession=60993220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419739, CurrentID=0, Status='live', Symbol='LOC100419739', Aliases='', description='zinc finger protein 341 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80385584, end_position_on_the_genomic_accession=80388725, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100689013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689013, CurrentID=0, Status='live', Symbol='MRT17', Aliases='MRT21', description='mental retardation, non-syndromic, autosomal recessive, 17', other_designations='', map_location='4q26-q28', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614207, no_hdr0=''),
  100689014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689014, CurrentID=0, Status='live', Symbol='MRT16', Aliases='', description='Mental retardation, autosomal recessive 16', other_designations='', map_location='9p23-p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614208, no_hdr0=''),
  26134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26134, CurrentID=0, Status='live', Symbol='RALGAPA1P', Aliases='GARNL2, GARNL2P, bA235C23.1', description='Ral GTPase activating protein, alpha subunit 1 (catalytic) pseudogene', other_designations='GTPase activating RANGAP domain-like 2 pseudogene|GTPase activating Rap/RanGAP domain-like 2 pseudogene', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105519824, end_position_on_the_genomic_accession=105527319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271346, CurrentID=0, Status='live', Symbol='RPLP1P10', Aliases='RPLP1_6_1047', description='ribosomal protein, large, P1 pseudogene 10', other_designations='', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59088661, end_position_on_the_genomic_accession=59089023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401585, CurrentID=0, Status='live', Symbol='LOC401585', Aliases='', description='uncharacterized LOC401585', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45848074, end_position_on_the_genomic_accession=45851490, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284551, CurrentID=0, Status='live', Symbol='LINC01226', Aliases='', description='long intergenic non-protein coding RNA 1226', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31518435, end_position_on_the_genomic_accession=31524245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  25758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25758, CurrentID=0, Status='live', Symbol='KIAA1549L', Aliases='C11orf41, C11orf69, G2', description='KIAA1549-like', other_designations='UPF0606 protein C11orf41|Uncharacterized protein C11orf69', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33376568, end_position_on_the_genomic_accession=33674102, orientation='plus', exon_count=22, OMIM=612297, no_hdr0=''),
  25759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25759, CurrentID=0, Status='live', Symbol='SHC2', Aliases='SCK, SHCB, SLI', description='SHC (Src homology 2 domain containing) transforming protein 2', other_designations='SH2 domain protein C2|SHC-transforming protein B|neuronal Shc adaptor homolog|protein Sck|src homology 2 domain-containing-transforming protein C2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=416583, end_position_on_the_genomic_accession=461024, orientation='minus', exon_count=16, OMIM=605217, no_hdr0=''),
  25763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25763, CurrentID=0, Status='live', Symbol='HYPM', Aliases='CXorf27, HIP17', description='huntingtin interacting protein M', other_designations='huntingtin yeast partner M', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37990817, end_position_on_the_genomic_accession=37991317, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  25764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25764, CurrentID=0, Status='live', Symbol='HYPK', Aliases='C15orf63, HSPC136', description='huntingtin interacting protein K', other_designations='huntingtin yeast partner K', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43800421, end_position_on_the_genomic_accession=43802589, orientation='plus', exon_count=4, OMIM=612784, no_hdr0=''),
  25766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25766, CurrentID=0, Status='live', Symbol='PRPF40B', Aliases='HYPC', description='pre-mRNA processing factor 40 homolog B', other_designations='Huntingtin interacting protein C|PRP40 homolog, pre-mRNA processing factor B|PRP40 pre-mRNA processing factor 40 homolog B|huntingtin yeast partner C', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49623414, end_position_on_the_genomic_accession=49644669, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  25769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25769, CurrentID=0, Status='live', Symbol='SLC24A2', Aliases='NCKX2', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 2', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 2|retinal cone Na-Ca+K exchanger', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19507452, end_position_on_the_genomic_accession=19789695, orientation='minus', exon_count=12, OMIM=609838, no_hdr0=''),
  25770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25770, CurrentID=0, Status='live', Symbol='C22orf31', Aliases='HS747E2A, bK747E2.1', description='chromosome 22 open reading frame 31', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29058672, end_position_on_the_genomic_accession=29073966, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  25771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25771, CurrentID=0, Status='live', Symbol='TBC1D22A', Aliases='C22orf4, HSC79E021', description='TBC1 domain family member 22A', other_designations='TBC1 domain family, member 22A|putative GTPase activator', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46762617, end_position_on_the_genomic_accession=47203479, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  25774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25774, CurrentID=0, Status='live', Symbol='GSTTP1', Aliases='GSTT4, HS322B1A', description='glutathione S-transferase theta pseudogene 1', other_designations='GST class-theta-4|Glutathione S-transferase theta-4|glutathione S-transferase theta 1 pseudogene', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23998401, end_position_on_the_genomic_accession=24005361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  25775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25775, CurrentID=0, Status='live', Symbol='C22orf24', Aliases='HSN44A4A', description='chromosome 22 open reading frame 24', other_designations='', map_location='22q12.1-q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31925316, end_position_on_the_genomic_accession=31945362, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  25776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25776, CurrentID=0, Status='live', Symbol='CBY1', Aliases='C22orf2, CBY, HS508I15A, PGEA1, PIGEA-14, PIGEA14, arb1', description='chibby homolog 1 (Drosophila)', other_designations='ARPP-binding protein|PKD2 interactor, Golgi and endoplasmic reticulum-associated 1|PKD2 interactor, golgi and endoplasmic reticulum associated 1|chibby CTNNB1-mediated transcription inhibitor|coiled-coil protein PIGEA-14|cytosolic leucine-rich protein|polycystin-2 interactor, Golgi- and endoplasmic reticulum-associated protein, 14 kDa', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38656653, end_position_on_the_genomic_accession=38673850, orientation='plus', exon_count=6, OMIM=607757, no_hdr0=''),
  25777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25777, CurrentID=0, Status='live', Symbol='SUN2', Aliases='UNC84B', description='Sad1 and UNC84 domain containing 2', other_designations='Sad1 unc-84 domain protein 2|nuclear envelope protein|protein unc-84 homolog B|rab5-interacting protein|rab5IP|sad1/unc-84 protein-like 2|unc-84 homolog B', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38734714, end_position_on_the_genomic_accession=38756019, orientation='minus', exon_count=21, OMIM=613569, no_hdr0=''),
  25778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25778, CurrentID=0, Status='live', Symbol='DSTYK', Aliases='CAKUT1, DustyPK, HDCMD38P, RIP5, RIPK5', description='dual serine/threonine and tyrosine protein kinase', other_designations='RIP-homologous kinase|dusty PK|dusty protein kinase|receptor-interacting serine/threonine-protein kinase 5|sgK496|sugen kinase 496', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205142503, end_position_on_the_genomic_accession=205211599, orientation='minus', exon_count=14, OMIM=612666, no_hdr0=''),
  25780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25780, CurrentID=0, Status='live', Symbol='RASGRP3', Aliases='GRP3', description='RAS guanyl releasing protein 3 (calcium and DAG-regulated)', other_designations='calcium and DAG-regulated guanine nucleotide exchange factor III|guanine nucleotide exchange factor for Rap1', map_location='2p25.1-p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33436349, end_position_on_the_genomic_accession=33564731, orientation='plus', exon_count=21, OMIM=609531, no_hdr0=''),
  25782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25782, CurrentID=0, Status='live', Symbol='RAB3GAP2', Aliases='RAB3-GAP150, RAB3GAP150, SPG69, WARBM2, p150', description='RAB3 GTPase activating non-catalytic protein subunit 2', other_designations='RAB3 GTPase activating protein subunit 2 (non-catalytic)|RGAP-iso|rab3 GTPase-activating protein 150 kDa subunit|rab3-GAP p150|rab3-GAP regulatory subunit', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220148268, end_position_on_the_genomic_accession=220272501, orientation='minus', exon_count=35, OMIM=609275, no_hdr0=''),
  25784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25784, CurrentID=0, Status='live', Symbol='DGCR12', Aliases='DGS-E', description='DiGeorge syndrome critical region gene 12 (non-protein coding)', other_designations='DiGeorge syndrome gene E', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  25786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25786, CurrentID=0, Status='live', Symbol='DGCR11', Aliases='DGS-D', description='DiGeorge syndrome critical region gene 11 (non-protein coding)', other_designations='DiGeorge syndrome gene D', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19046162, end_position_on_the_genomic_accession=19048375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  25787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25787, CurrentID=0, Status='live', Symbol='DGCR9', Aliases='DGS-A', description='DiGeorge syndrome critical region gene 9 (non-protein coding)', other_designations='DiGeorge syndrome gene A', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19017834, end_position_on_the_genomic_accession=19020248, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  25788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25788, CurrentID=0, Status='live', Symbol='RAD54B', Aliases='RDH54', description='RAD54 homolog B (S. cerevisiae)', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94371960, end_position_on_the_genomic_accession=94475115, orientation='minus', exon_count=17, OMIM=604289, no_hdr0=''),
  25789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25789, CurrentID=0, Status='live', Symbol='TMEM59L', Aliases='BSMAP, C19orf4', description='transmembrane protein 59 like', other_designations='brain-specific membrane-anchored protein', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18607435, end_position_on_the_genomic_accession=18621039, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  25790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25790, CurrentID=0, Status='live', Symbol='CFAP45', Aliases='CCDC19, NESG1', description='cilia and flagella associated protein 45', other_designations='coiled-coil domain containing 19|coiled-coil domain-containing protein 19, mitochondrial|nasopharyngeal epithelium specific protein 1 (NESG1)|nasopharyngeal epithelium-specific protein 1|protein CFAP45, mitochondrial', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159872364, end_position_on_the_genomic_accession=159900116, orientation='minus', exon_count=12, OMIM=605152, no_hdr0=''),
  25791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25791, CurrentID=0, Status='live', Symbol='NGEF', Aliases='ARHGEF27, EPHEXIN', description='neuronal guanine nucleotide exchange factor', other_designations='eph-interacting exchange protein', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232878686, end_position_on_the_genomic_accession=233013241, orientation='minus', exon_count=18, OMIM=605991, no_hdr0=''),
  25792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25792, CurrentID=0, Status='live', Symbol='CIZ1', Aliases='LSFR1, NP94, ZNF356', description='CDKN1A interacting zinc finger protein 1', other_designations='nuclear protein NP94|zinc finger protein 356', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128166065, end_position_on_the_genomic_accession=128204383, orientation='minus', exon_count=21, OMIM=611420, no_hdr0=''),
  25793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25793, CurrentID=0, Status='live', Symbol='FBXO7', Aliases='FBX, FBX07, FBX7, PARK15, PKPS', description='F-box protein 7', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32474682, end_position_on_the_genomic_accession=32498831, orientation='plus', exon_count=10, OMIM=605648, no_hdr0=''),
  25794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25794, CurrentID=0, Status='live', Symbol='FSCN2', Aliases='RFSN, RP30', description='fascin actin-bundling protein 2, retinal', other_designations='fascin homolog 2, actin-bundling protein, retinal', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81513051, end_position_on_the_genomic_accession=81539790, orientation='plus', exon_count=12, OMIM=607643, no_hdr0=''),
  25796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25796, CurrentID=0, Status='live', Symbol='PGLS', Aliases='6PGL, HEL-S-304', description='6-phosphogluconolactonase', other_designations='epididymis secretory protein Li 304', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17511623, end_position_on_the_genomic_accession=17521288, orientation='plus', exon_count=6, OMIM=604951, no_hdr0=''),
  25797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25797, CurrentID=0, Status='live', Symbol='QPCT', Aliases='GCT, QC, sQC', description='glutaminyl-peptide cyclotransferase', other_designations='EC|glutaminyl cyclase|glutaminyl-tRNA cyclotransferase|glutamyl cyclase', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37344610, end_position_on_the_genomic_accession=37373322, orientation='plus', exon_count=7, OMIM=607065, no_hdr0=''),
  25798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25798, CurrentID=0, Status='live', Symbol='BRI3', Aliases='I3', description='brain protein I3', other_designations='pRGR2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98281667, end_position_on_the_genomic_accession=98323430, orientation='plus', exon_count=7, OMIM=615628, no_hdr0=''),
  25799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25799, CurrentID=0, Status='live', Symbol='ZNF324', Aliases='ZF5128A, ZNF324', description='zinc finger protein 324', other_designations='zinc finger protein ZF5128', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58467045, end_position_on_the_genomic_accession=58473578, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25800, CurrentID=0, Status='live', Symbol='SLC39A6', Aliases='LIV-1, ZIP6', description='solute carrier family 39 (zinc transporter), member 6', other_designations='LIV-1 protein, estrogen regulated|ZIP-6|estrogen-regulated protein LIV-1|solute carrier family 39 (metal ion transporter), member 6|solute carrier family 39 member 6|zrt- and Irt-like protein 6', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36108531, end_position_on_the_genomic_accession=36129814, orientation='minus', exon_count=11, OMIM=608731, no_hdr0=''),
  25801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25801, CurrentID=0, Status='live', Symbol='GCA', Aliases='GCL', description='grancalcin', other_designations='grancalcin, EF-hand calcium binding protein|grancalcin, EF-hand calcium-binding protein|grancalcin, penta-EF-hand protein', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162318793, end_position_on_the_genomic_accession=162379989, orientation='plus', exon_count=14, OMIM=607030, no_hdr0=''),
  25802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25802, CurrentID=0, Status='live', Symbol='LMOD1', Aliases='1D, 64kD, D1, SM-LMOD, SMLMOD', description='leiomodin 1', other_designations='64 kDa autoantigen 1D|64 kDa autoantigen 1D3|64 kDa autoantigen D1|leimodin 1 (smooth muscle)|leiomodin 1 (smooth muscle)|leiomodin, muscle form|smooth muscle leiomodin|thyroid and eye muscle autoantigen D1 (64kD)|thyroid-associated ophthalmopathy autoantigen', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201896456, end_position_on_the_genomic_accession=201946588, orientation='minus', exon_count=3, OMIM=602715, no_hdr0=''),
  25803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25803, CurrentID=0, Status='live', Symbol='SPDEF', Aliases='PDEF, bA375E1.3', description='SAM pointed domain containing ETS transcription factor', other_designations='prostate epithelium-specific Ets transcription factor|prostate-derived Ets factor|prostate-specific Ets', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34537802, end_position_on_the_genomic_accession=34556333, orientation='minus', exon_count=8, OMIM=608144, no_hdr0=''),
  25804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25804, CurrentID=0, Status='live', Symbol='LSM4', Aliases='GRP, YER112W', description='LSM4 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM4 U6 small nuclear RNA and mRNA degradation associated|LSM4 homolog, U6 small nuclear RNA associated|glycine-rich protein', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18306230, end_position_on_the_genomic_accession=18323191, orientation='minus', exon_count=5, OMIM=607284, no_hdr0=''),
  25805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25805, CurrentID=0, Status='live', Symbol='BAMBI', Aliases='NMA', description='BMP and activin membrane-bound inhibitor', other_designations='non-metastatic gene A protein|putative transmembrane protein NMA', map_location='10p12.3-p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28677495, end_position_on_the_genomic_accession=28682939, orientation='plus', exon_count=3, OMIM=604444, no_hdr0=''),
  25806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25806, CurrentID=0, Status='live', Symbol='VAX2', Aliases='DRES93', description='ventral anterior homeobox 2', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70900590, end_position_on_the_genomic_accession=70933446, orientation='plus', exon_count=6, OMIM=604295, no_hdr0=''),
  25807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25807, CurrentID=0, Status='live', Symbol='RHBDD3', Aliases='C22orf3, HS984G1A, PTAG', description='rhomboid domain containing 3', other_designations='pituitary tumor apoptosis', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29259855, end_position_on_the_genomic_accession=29268185, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  25809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25809, CurrentID=0, Status='live', Symbol='TTLL1', Aliases='C22orf7, HS323M22B', description='tubulin tyrosine ligase like 1', other_designations='PGs3|catalytic subunit of neural tubulin polyglutamylase|tubulin polyglutamylase complex subunit 3|tubulin tyrosine ligase-like family member 1|tubulin tyrosine ligase-like family, member 1|tubulin--tyrosine ligase-like protein 1|tubulin-tyrosine ligase', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43039516, end_position_on_the_genomic_accession=43089428, orientation='minus', exon_count=13, OMIM=608955, no_hdr0=''),
  25812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25812, CurrentID=0, Status='live', Symbol='POM121L1P', Aliases='POM121L1', description='POM121 transmembrane nucleoporin like 1, pseudogene', other_designations='POM121 membrane glycoprotein-like 1, pseudogene|POM121 transmembrane nucleoporin-like 1, pseudogene|POM121-like 2', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22631557, end_position_on_the_genomic_accession=22644541, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  25813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25813, CurrentID=0, Status='live', Symbol='SAMM50', Aliases='CGI-51, OMP85, SAM50, TOB55, TRG-3, YNL026W', description='SAMM50 sorting and assembly machinery component', other_designations='sorting and assembly machinery 50kDa|transformation-related gene 3 protein', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43955381, end_position_on_the_genomic_accession=43996532, orientation='plus', exon_count=15, OMIM=612058, no_hdr0=''),
  25814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25814, CurrentID=0, Status='live', Symbol='ATXN10', Aliases='E46L, HUMEEP, SCA10', description='ataxin 10', other_designations='brain protein E46 homolog|spinocerebellar ataxia type 10 protein', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45671798, end_position_on_the_genomic_accession=45845307, orientation='plus', exon_count=12, OMIM=611150, no_hdr0=''),
  25816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25816, CurrentID=0, Status='live', Symbol='TNFAIP8', Aliases='GG2-1, MDC-3.13, NDED, SCC-S2, SCCS2', description='TNF alpha induced protein 8', other_designations='NF-kappa-B-inducible DED-containing protein|TNF alpha-induced protein 8|TNF-induced protein GG2-1|head and neck tumor and metastasis-related protein|tumor necrosis factor, alpha induced protein 8|tumor necrosis factor, alpha-induced protein 8', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119268692, end_position_on_the_genomic_accession=119394604, orientation='plus', exon_count=6, OMIM=612111, no_hdr0=''),
  25817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25817, CurrentID=0, Status='live', Symbol='FAM19A5', Aliases='QLLK5208, TAFA-5, TAFA5, UNQ5208', description='family with sequence similarity 19 (chemokine (C-C motif)-like), member A5', other_designations='TAFA protein 5|chemokine-like protein TAFA-5', map_location='22q13.32', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48489460, end_position_on_the_genomic_accession=48751935, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25818, CurrentID=0, Status='live', Symbol='KLK5', Aliases='KLK-L2, KLKL2, SCTE', description='kallikrein related peptidase 5', other_designations='kallikrein 5|kallikrein-like protein 2|kallikrein-related peptidase 5|stratum corneum tryptic enzyme', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50943299, end_position_on_the_genomic_accession=50953088, orientation='minus', exon_count=7, OMIM=605643, no_hdr0=''),
  25819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25819, CurrentID=0, Status='live', Symbol='NOCT', Aliases='CCR4L, CCRN4L, Ccr4c, NOC', description='nocturnin', other_designations='CCR4 carbon catabolite repression 4-like|CCR4 protein homolog|CCR4-like (carbon catabolite repression 4, S.cerevisiae)|carbon catabolite repression 4-like protein|circadian deadenylase NOC', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139015759, end_position_on_the_genomic_accession=139045939, orientation='plus', exon_count=3, OMIM=608468, no_hdr0=''),
  25820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25820, CurrentID=0, Status='live', Symbol='ARIH1', Aliases='ARI, HARI, HHARI, UBCH7BP', description='ariadne RBR E3 ubiquitin protein ligase 1', other_designations='ARI-1|H7-AP2|MOP-6|ariadne homolog, ubiquitin-conjugating enzyme E2 binding protein, 1|ariadne, Drosophila, homolog of|monocyte protein 6|protein ariadne-1 homolog|ubcH7-binding protein|ubcM4-interacting protein|ubiquitin-conjugating enzyme E2-binding protein 1', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72474282, end_position_on_the_genomic_accession=72586560, orientation='plus', exon_count=15, OMIM=605624, no_hdr0=''),
  25821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25821, CurrentID=0, Status='live', Symbol='MTO1', Aliases='CGI-02, COXPD10', description='mitochondrial tRNA translation optimization 1', other_designations='homolog of yeast Mto1|mitochondrial MTO1-3|mitochondrial translation optimization 1 homolog', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73461731, end_position_on_the_genomic_accession=73501456, orientation='plus', exon_count=15, OMIM=614667, no_hdr0=''),
  25822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25822, CurrentID=0, Status='live', Symbol='DNAJB5', Aliases='Hsc40', description='DnaJ heat shock protein family (Hsp40) member B5', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 5|heat shock cognate 40|heat shock protein Hsp40-2|heat shock protein Hsp40-3|heat shock protein cognate 40', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34989369, end_position_on_the_genomic_accession=34998433, orientation='plus', exon_count=7, OMIM=611328, no_hdr0=''),
  25823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25823, CurrentID=0, Status='live', Symbol='TPSG1', Aliases='PRSS31, TMT, trpA', description='tryptase gamma 1', other_designations='gamma I|gamma II|lung tryptase|mast cell protease II|mast cell tryptase|pituitary tryptase|serine protease 31|skin tryptase|transmembrane tryptase|tryptase gamma I|tryptase gamma II', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1221488, end_position_on_the_genomic_accession=1226384, orientation='minus', exon_count=7, OMIM=609341, no_hdr0=''),
  25824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25824, CurrentID=0, Status='live', Symbol='PRDX5', Aliases='ACR1, AOEB166, B166, HEL-S-55, PLP, PMP20, PRDX6, PRXV, SBBI10, prx-V', description='peroxiredoxin 5', other_designations='Alu co-repressor 1|TPx type VI|antioxidant enzyme B166|epididymis secretory protein Li 55|liver tissue 2D-page spot 71B|peroxiredoxin V|peroxisomal antioxidant enzyme|thioredoxin peroxidase PMP20|thioredoxin reductase', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64318088, end_position_on_the_genomic_accession=64321823, orientation='plus', exon_count=6, OMIM=606583, no_hdr0=''),
  25825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25825, CurrentID=0, Status='live', Symbol='BACE2', Aliases='AEPLC, ALP56, ASP1, ASP21, BAE2, CDA13, CEAP1, DRAP', description='beta-site APP-cleaving enzyme 2', other_designations='56 kDa aspartic-like protease|Down syndrome region aspartic protease|aspartyl protease 1|beta-site amyloid beta A4 precursor protein-cleaving enzyme 2|memapsin-1|membrane-associated aspartic protease 1|theta-secretase|transmembrane aspartic proteinase Asp1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41167801, end_position_on_the_genomic_accession=41282534, orientation='plus', exon_count=9, OMIM=605668, no_hdr0=''),
  25826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25826, CurrentID=0, Status='live', Symbol='SNORD82', Aliases='RNU82, U82, Z25', description='small nucleolar RNA, C/D box 82', other_designations='RNA, U82 small nucleolar', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231460368, end_position_on_the_genomic_accession=231460442, orientation='minus', exon_count=1, OMIM=611133, no_hdr0=''),
  25827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25827, CurrentID=0, Status='live', Symbol='FBXL2', Aliases='FBL2, FBL3', description='F-box and leucine-rich repeat protein 2', other_designations='F-box protein FBL2/FBL3|F-box protein containing leucine-rich repeats', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33277442, end_position_on_the_genomic_accession=33403554, orientation='plus', exon_count=18, OMIM=605652, no_hdr0=''),
  25828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25828, CurrentID=0, Status='live', Symbol='TXN2', Aliases='MT-TRX, MTRX, TRX2', description='thioredoxin 2', other_designations='mitochondrial thioredoxin|thioredoxin-2', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36467036, end_position_on_the_genomic_accession=36482025, orientation='minus', exon_count=5, OMIM=609063, no_hdr0=''),
  25829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25829, CurrentID=0, Status='live', Symbol='TMEM184B', Aliases='C22orf5, FM08, HS5O6A, HSPC256', description='transmembrane protein 184B', other_designations='putative MAPK-activating protein FM08', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38216377, end_position_on_the_genomic_accession=38273034, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  25830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25830, CurrentID=0, Status='live', Symbol='SULT4A1', Aliases='BR-STL-1, BRSTL1, DJ388M5.3, NST, SULTX3, hBR-STL-1', description='sulfotransferase family 4A member 1', other_designations='ST4A1|brain sulfotransferase-like protein|hBR-STL|nervous system cytosolic sulfotransferase|nervous system sulfotransferase|sulfotransferase family 4A, member 1|sulfotransferase-related protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43824507, end_position_on_the_genomic_accession=43862498, orientation='minus', exon_count=9, OMIM=608359, no_hdr0=''),
  25831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25831, CurrentID=0, Status='live', Symbol='HECTD1', Aliases='EULIR', description='HECT domain containing E3 ubiquitin protein ligase 1', other_designations='E3 ligase for inhibin receptor|HECT domain-containing protein 1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31100115, end_position_on_the_genomic_accession=31208060, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  25832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25832, CurrentID=0, Status='live', Symbol='NBPF14', Aliases='DJ328E19.C1.1, NBPF', description='neuroblastoma breakpoint family member 14', other_designations='AE5|neuroblastoma breakpoint family, member 14', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148531385, end_position_on_the_genomic_accession=148595947, orientation='minus', exon_count=67, OMIM=614003, no_hdr0=''),
  25833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25833, CurrentID=0, Status='live', Symbol='POU2F3', Aliases='Epoc-1, OCT-11, OCT11, OTF-11, PLA-1, PLA1, Skn-1a', description='POU class 2 homeobox 3', other_designations='POU domain transcription factor OCT11a|octamer-binding protein 11|octamer-binding transcription factor 11|transcription factor PLA-1|transcription factor Skn-1', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120236640, end_position_on_the_genomic_accession=120319945, orientation='plus', exon_count=18, OMIM=607394, no_hdr0=''),
  25834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25834, CurrentID=0, Status='live', Symbol='MGAT4C', Aliases='GNTIVH, HGNT-IV-H', description='MGAT4 family member C', other_designations='MGAT4 family, member C|N-acetylglucosaminyltransferase IV homolog|N-acetylglucosaminyltransferase IVc|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVc|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVc|UDP-N-acetylglucosamine:a-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IV|UDP-N-acetylglucosamine:a-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IV-homolog|glcNAc-T IVc|gnT-IVc|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isozyme C (putative)', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85976235, end_position_on_the_genomic_accession=86838904, orientation='minus', exon_count=15, OMIM=607385, no_hdr0=''),
  25836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25836, CurrentID=0, Status='live', Symbol='NIPBL', Aliases='CDLS, CDLS1, IDN3, IDN3-B, Scc2', description='Nipped-B homolog (Drosophila)', other_designations='SCC2 homolog|delangin|sister chromatid cohesion 2 homolog', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36876759, end_position_on_the_genomic_accession=37065819, orientation='plus', exon_count=48, OMIM=608667, no_hdr0=''),
  25837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25837, CurrentID=0, Status='live', Symbol='RAB26', Aliases='V46133', description='RAB26, member RAS oncogene family', other_designations='Ras-related oncogene protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2148144, end_position_on_the_genomic_accession=2154140, orientation='plus', exon_count=11, OMIM=605455, no_hdr0=''),
  25839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25839, CurrentID=0, Status='live', Symbol='COG4', Aliases='CDG2J, COD1', description='component of oligomeric golgi complex 4', other_designations='COG complex subunit 4|complexed with Dor1p|conserved oligomeric Golgi complex protein 4', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70480569, end_position_on_the_genomic_accession=70523554, orientation='minus', exon_count=20, OMIM=606976, no_hdr0=''),
  25840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25840, CurrentID=0, Status='live', Symbol='METTL7A', Aliases='AAM-B', description='methyltransferase like 7A', other_designations='protein AAM-B', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50923811, end_position_on_the_genomic_accession=50932517, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  25841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25841, CurrentID=0, Status='live', Symbol='ABTB2', Aliases='ABTB2A, BTBD22', description='ankyrin repeat and BTB (POZ) domain containing 2', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34150987, end_position_on_the_genomic_accession=34358008, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  25842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25842, CurrentID=0, Status='live', Symbol='ASF1A', Aliases='CGI-98, CIA, HSPC146', description='anti-silencing function 1A histone chaperone', other_designations='ASF1 anti-silencing function 1 homolog A|CCG1-interacting factor A|anti-silencing function protein 1 homolog A|hAsf1|hAsf1a|hCIA', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118894077, end_position_on_the_genomic_accession=118909171, orientation='plus', exon_count=4, OMIM=609189, no_hdr0=''),
  25843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25843, CurrentID=0, Status='live', Symbol='MOB4', Aliases='2C4D, CGI-95, MOB1, MOB3, MOBKL3, PHOCN, PREI3', description='MOB family member 4, phocein', other_designations='MOB1, Mps One Binder kinase activator-like 3|Mps One Binder kinase activator-like 3|class II mMOB1|mob1 homolog 3|phocein, Mob-like protein|preimplantation protein 3', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197515571, end_position_on_the_genomic_accession=197553699, orientation='plus', exon_count=10, OMIM=609361, no_hdr0=''),
  25844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25844, CurrentID=0, Status='live', Symbol='YIPF3', Aliases='C6orf109, FinGER3, KLIP1, dJ337H4.3', description='Yip1 domain family member 3', other_designations='Yip1 domain family, member 3|killer lineage protein 1|natural killer cell-specific antigen KLIP1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43511827, end_position_on_the_genomic_accession=43516990, orientation='minus', exon_count=10, OMIM=609775, no_hdr0=''),
  25845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25845, CurrentID=0, Status='live', Symbol='PP7080', Aliases='', description='uncharacterized LOC25845', other_designations='CTD-2228K2.5', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=470510, end_position_on_the_genomic_accession=472965, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  25847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25847, CurrentID=0, Status='live', Symbol='ANAPC13', Aliases='APC13, SWM1', description='anaphase promoting complex subunit 13', other_designations='cyclosome subunit 13', map_location='3q22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134477704, end_position_on_the_genomic_accession=134486023, orientation='minus', exon_count=3, OMIM=614484, no_hdr0=''),
  25849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25849, CurrentID=0, Status='live', Symbol='PARM1', Aliases='Cipar1, DKFZP564O0823, PARM-1, WSC4', description='prostate androgen-regulated mucin-like protein 1', other_designations='WSC4, cell wall integrity and stress response component 4 homolog|castration-induced prostatic apoptosis-related protein 1|prostatic androgen-regulated mucin-like protein 1|prostatic androgen-repressed message 1|prostatic androgen-repressed message-1', map_location='4q13.3-q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74933075, end_position_on_the_genomic_accession=75050115, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25850, CurrentID=0, Status='live', Symbol='ZNF345', Aliases='HZF10', description='zinc finger protein 345', other_designations='zinc finger protein HZF10', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36850358, end_position_on_the_genomic_accession=36893218, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  25851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25851, CurrentID=0, Status='live', Symbol='TECPR1', Aliases='', description='tectonin beta-propeller repeat containing 1', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98215443, end_position_on_the_genomic_accession=98252251, orientation='minus', exon_count=27, OMIM=614781, no_hdr0=''),
  25852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25852, CurrentID=0, Status='live', Symbol='ARMC8', Aliases='GID5, HSPC056, S863-2, VID28', description='armadillo repeat containing 8', other_designations='GID complex subunit 5, VID28 homolog', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138187248, end_position_on_the_genomic_accession=138298389, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  25853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25853, CurrentID=0, Status='live', Symbol='DCAF12', Aliases='CT102, KIAA1892, TCC52, WDR40A', description='DDB1 and CUL4 associated factor 12', other_designations='WD repeat domain 40A|WD repeat-containing protein 40A|cancer/testis antigen 102|centrosome-related protein TCC52|testis cancer centrosome-related protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34086383, end_position_on_the_genomic_accession=34126773, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  25854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25854, CurrentID=0, Status='live', Symbol='FAM149A', Aliases='MST119, MSTP119', description='family with sequence similarity 149 member A', other_designations='family with sequence similarity 149, member A', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186144287, end_position_on_the_genomic_accession=186172663, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  25855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25855, CurrentID=0, Status='live', Symbol='BRMS1', Aliases='', description='breast cancer metastasis suppressor 1', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66337333, end_position_on_the_genomic_accession=66345111, orientation='minus', exon_count=10, OMIM=606259, no_hdr0=''),
  100271350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271350, CurrentID=0, Status='live', Symbol='RPS10P19', Aliases='RPS10_7_1142', description='ribosomal protein S10 pseudogene 19', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46428593, end_position_on_the_genomic_accession=46429381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25858, CurrentID=0, Status='live', Symbol='TEX40', Aliases='C11orf20', description='testis expressed 40', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64300391, end_position_on_the_genomic_accession=64304767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25859, CurrentID=0, Status='live', Symbol='PART1', Aliases='NCRNA00206', description='prostate androgen-regulated transcript 1 (non-protein coding)', other_designations='prostate-specific and androgen-regulated cDNA 14D7 protein', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60487713, end_position_on_the_genomic_accession=60547657, orientation='plus', exon_count=6, OMIM=604991, no_hdr0=''),
  25861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25861, CurrentID=0, Status='live', Symbol='DFNB31', Aliases='CIP98, PDZD7B, USH2D, WHRN, WI', description='deafness, autosomal recessive 31', other_designations='CASK-interacting protein CIP98|autosomal recessive deafness type 31 protein', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114402078, end_position_on_the_genomic_accession=114505509, orientation='minus', exon_count=20, OMIM=607928, no_hdr0=''),
  25862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25862, CurrentID=0, Status='live', Symbol='USP49', Aliases='', description='ubiquitin specific peptidase 49', other_designations='deubiquitinating enzyme 49|ubiquitin specific protease 49|ubiquitin thioesterase 49|ubiquitin thiolesterase 49|ubiquitin-specific-processing protease 49', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41789896, end_position_on_the_genomic_accession=41895361, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100033799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033799, CurrentID=0, Status='live', Symbol='SNORD115-22', Aliases='HBII-52-22', description='small nucleolar RNA, C/D box 115-22', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25209918, end_position_on_the_genomic_accession=25209999, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033800, CurrentID=0, Status='live', Symbol='SNORD115-23', Aliases='HBII-52-23', description='small nucleolar RNA, C/D box 115-23', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25211796, end_position_on_the_genomic_accession=25211877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033801, CurrentID=0, Status='live', Symbol='SNORD115-25', Aliases='HBII-52-25', description='small nucleolar RNA, C/D box 115-25', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25215541, end_position_on_the_genomic_accession=25215622, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033802, CurrentID=0, Status='live', Symbol='SNORD115-26', Aliases='HBII-52-26', description='small nucleolar RNA, C/D box 115-26', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25218617, end_position_on_the_genomic_accession=25218698, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033803, CurrentID=0, Status='live', Symbol='SNORD115-29', Aliases='HBII-52-29', description='small nucleolar RNA, C/D box 115-29', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25223246, end_position_on_the_genomic_accession=25223327, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033804, CurrentID=0, Status='live', Symbol='SNORD115-30', Aliases='HBII-52-30', description='small nucleolar RNA, C/D box 115-30', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25225203, end_position_on_the_genomic_accession=25225284, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033805, CurrentID=0, Status='live', Symbol='SNORD115-31', Aliases='HBII-52-31', description='small nucleolar RNA, C/D box 115-31', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25227109, end_position_on_the_genomic_accession=25227190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033806, CurrentID=0, Status='live', Symbol='SNORD115-32', Aliases='HBII-52-32', description='small nucleolar RNA, C/D box 115-32', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25228967, end_position_on_the_genomic_accession=25229048, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033807, CurrentID=0, Status='live', Symbol='SNORD115-33', Aliases='HBII-52-33', description='small nucleolar RNA, C/D box 115-33', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25230838, end_position_on_the_genomic_accession=25230919, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033808, CurrentID=0, Status='live', Symbol='SNORD115-34', Aliases='HBII-52-34', description='small nucleolar RNA, C/D box 115-34', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25232387, end_position_on_the_genomic_accession=25232468, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033809, CurrentID=0, Status='live', Symbol='SNORD115-35', Aliases='HBII-52-35', description='small nucleolar RNA, C/D box 115-35', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25234247, end_position_on_the_genomic_accession=25234328, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033810, CurrentID=0, Status='live', Symbol='SNORD115-36', Aliases='HBII-52-36', description='small nucleolar RNA, C/D box 115-36', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25236085, end_position_on_the_genomic_accession=25236166, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033811, CurrentID=0, Status='live', Symbol='SNORD115-37', Aliases='HBII-52-37', description='small nucleolar RNA, C/D box 115-37', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25237986, end_position_on_the_genomic_accession=25238067, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033812, CurrentID=0, Status='live', Symbol='SNORD115-38', Aliases='HBII-52-38', description='small nucleolar RNA, C/D box 115-38', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25239838, end_position_on_the_genomic_accession=25239919, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033813, CurrentID=0, Status='live', Symbol='SNORD115-39', Aliases='HBII-52-39', description='small nucleolar RNA, C/D box 115-39', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25241746, end_position_on_the_genomic_accession=25241827, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033814, CurrentID=0, Status='live', Symbol='SNORD115-40', Aliases='HBII-52-40', description='small nucleolar RNA, C/D box 115-40', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25243614, end_position_on_the_genomic_accession=25243695, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  25879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25879, CurrentID=0, Status='live', Symbol='DCAF13', Aliases='GM83, HSPC064, WDSOF1', description='DDB1 and CUL4 associated factor 13', other_designations='WD repeat and SOF domain-containing protein 1|WD repeats and SOF1 domain containing', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103414714, end_position_on_the_genomic_accession=103443453, orientation='plus', exon_count=11, OMIM=616196, no_hdr0=''),
  100033816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033816, CurrentID=0, Status='live', Symbol='SNORD115-42', Aliases='HBII-52-42', description='small nucleolar RNA, C/D box 115-42', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25247345, end_position_on_the_genomic_accession=25247426, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033817, CurrentID=0, Status='live', Symbol='SNORD115-43', Aliases='HBII-52-43', description='small nucleolar RNA, C/D box 115-43', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25249198, end_position_on_the_genomic_accession=25249279, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033818, CurrentID=0, Status='live', Symbol='SNORD115-44', Aliases='HBII-52-44', description='small nucleolar RNA, C/D box 115-44', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25250859, end_position_on_the_genomic_accession=25250940, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100033819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100033819, CurrentID=0, Status='live', Symbol='MIR675', Aliases='MIRN675, hsa-mir-675', description='microRNA 675', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1996759, end_position_on_the_genomic_accession=1996831, orientation='minus', exon_count=1, OMIM=615509, no_hdr0=''),
  25884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25884, CurrentID=0, Status='live', Symbol='CHRDL2', Aliases='BNF1, CHL2', description='chordin-like 2', other_designations='breast tumor novel factor 1|chordin-related protein 2', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74696428, end_position_on_the_genomic_accession=74731385, orientation='minus', exon_count=12, OMIM=613127, no_hdr0=''),
  25885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25885, CurrentID=0, Status='live', Symbol='POLR1A', Aliases='A190, AFDCIN, RPA1, RPA194, RPO1-4, RPO14', description='polymerase (RNA) I polypeptide A', other_designations='DNA-directed RNA polymerase I largest subunit|DNA-directed RNA polymerase I subunit A|DNA-directed RNA polymerase I subunit A1|RNA polymerase I 194 kDa subunit|RNA polymerase I subunit A1|polymerase (RNA) I polypeptide A, 194kDa', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86025168, end_position_on_the_genomic_accession=86106155, orientation='minus', exon_count=35, OMIM=616404, no_hdr0=''),
  25886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25886, CurrentID=0, Status='live', Symbol='POC1A', Aliases='PIX2, SOFT, WDR51A', description='POC1 centriolar protein A', other_designations='WD repeat domain 51A|WD repeat-containing protein 51A|proteome of centriole protein 1A', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52075233, end_position_on_the_genomic_accession=52154690, orientation='minus', exon_count=13, OMIM=614783, no_hdr0=''),
  100271351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271351, CurrentID=0, Status='live', Symbol='RPS10P21', Aliases='RPS10_8_1322', description='ribosomal protein S10 pseudogene 21', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71913702, end_position_on_the_genomic_accession=71914176, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25888, CurrentID=0, Status='live', Symbol='ZNF473', Aliases='HZFP100, ZN473', description='zinc finger protein 473', other_designations='zfp-100|zinc finger protein 100 homolog|zinc finger protein ZFP100', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50025383, end_position_on_the_genomic_accession=50048776, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25890, CurrentID=0, Status='live', Symbol='ABI3BP', Aliases='NESHBP, TARSH', description='ABI family member 3 binding protein', other_designations='ABI family, member 3 (NESH) binding protein|ABI gene family member 3-binding protein|ABI gene family, member 3 (NESH) binding protein|nesh-binding protein', map_location='3q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100749335, end_position_on_the_genomic_accession=100993493, orientation='minus', exon_count=68, OMIM=606279, no_hdr0=''),
  25891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25891, CurrentID=0, Status='live', Symbol='PAMR1', Aliases='DKFZP586H2123, FP938, RAMP', description='peptidase domain containing associated with muscle regeneration 1', other_designations='peptidase domain-containing protein associated with muscle regeneration 1|regeneration associated muscle protease|regeneration-associated muscle protease homolog', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35431827, end_position_on_the_genomic_accession=35530300, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  25893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25893, CurrentID=0, Status='live', Symbol='TRIM58', Aliases='BIA2', description='tripartite motif containing 58', other_designations='tripartite motif-containing 58|tripartite motif-containing protein 58', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247857199, end_position_on_the_genomic_accession=247880138, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  25894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25894, CurrentID=0, Status='live', Symbol='PLEKHG4', Aliases='ARHGEF44, PRTPHN1, SCA4', description='pleckstrin homology and RhoGEF domain containing G4', other_designations='PH domain-containing family G member 4|Purkinje cell atrophy associated protein 1|pleckstrin homology domain containing, family G (with RhoGef domain) member 4', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67277820, end_position_on_the_genomic_accession=67289500, orientation='plus', exon_count=27, OMIM=609526, no_hdr0=''),
  25895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25895, CurrentID=0, Status='live', Symbol='METTL21B', Aliases='FAM119B', description='methyltransferase like 21B', other_designations='family with sequence similarity 119, member B|hepatocellular carcinoma-associated antigen 557a|hepatocellularcarcinoma-associated antigen HCA557a|methyltransferase-like protein 21B', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57772600, end_position_on_the_genomic_accession=57782541, orientation='plus', exon_count=4, OMIM=615258, no_hdr0=''),
  25896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25896, CurrentID=0, Status='live', Symbol='INTS7', Aliases='C1orf73, INT7', description='integrator complex subunit 7', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211940399, end_position_on_the_genomic_accession=212035677, orientation='minus', exon_count=22, OMIM=611350, no_hdr0=''),
  25897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25897, CurrentID=0, Status='live', Symbol='RNF19A', Aliases='RNF19', description='ring finger protein 19A, RBR E3 ubiquitin protein ligase', other_designations='double ring-finger protein|protein p38 interacting with transcription factor Sp1|ring finger protein 19|ring finger protein 19A, E3 ubiquitin protein ligase|ring-IBR-ring domain containing protein Dorfin', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100257059, end_position_on_the_genomic_accession=100336218, orientation='minus', exon_count=15, OMIM=607119, no_hdr0=''),
  25898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25898, CurrentID=0, Status='live', Symbol='RCHY1', Aliases='ARNIP, CHIMP, PIRH2, PRO1996, RNF199, ZCHY, ZNF363', description='ring finger and CHY zinc finger domain containing 1, E3 ubiquitin protein ligase', other_designations='CH-rich interacting match with PLAG1|E3 ubiquitin-protein ligase Pirh2|RING finger protein 199|androgen-receptor N-terminal-interacting protein|p53-induced protein with a RING-H2 domain|zinc finger protein 363|zinc finger, CHY-type', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75479037, end_position_on_the_genomic_accession=75514744, orientation='minus', exon_count=10, OMIM=607680, no_hdr0=''),
  25900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25900, CurrentID=0, Status='live', Symbol='IFFO1', Aliases='HOM-TES-103, IFFO', description='intermediate filament family orphan 1', other_designations='intermediate filament-like MGC:2625', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6539528, end_position_on_the_genomic_accession=6556083, orientation='minus', exon_count=13, OMIM=610495, no_hdr0=''),
  25901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25901, CurrentID=0, Status='live', Symbol='CCDC28A', Aliases='C6orf80, CCRL1AP', description='coiled-coil domain containing 28A', other_designations='chemokine C-C motif receptor-like 1 adjacent', map_location='6q23.1-q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138773520, end_position_on_the_genomic_accession=138808327, orientation='plus', exon_count=9, OMIM=615353, no_hdr0=''),
  25902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25902, CurrentID=0, Status='live', Symbol='MTHFD1L', Aliases='FTHFSDC1, MTC1THFS, dJ292B18.2', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1-like', other_designations='10-formyl-THF synthetase|formyltetrahydrofolate synthetase domain containing 1', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150865679, end_position_on_the_genomic_accession=151101887, orientation='plus', exon_count=38, OMIM=611427, no_hdr0=''),
  25903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25903, CurrentID=0, Status='live', Symbol='OLFML2B', Aliases='', description='olfactomedin like 2B', other_designations='olfactomedin-like 2B|photomedin-2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161983192, end_position_on_the_genomic_accession=162024465, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  25904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25904, CurrentID=0, Status='live', Symbol='CNOT10', Aliases='', description='CCR4-NOT transcription complex subunit 10', other_designations='CCR4-NOT transcription complex, subunit 10', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32685145, end_position_on_the_genomic_accession=32773875, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  25906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25906, CurrentID=0, Status='live', Symbol='ANAPC15', Aliases='APC15, C11orf51, HSPC020', description='anaphase promoting complex subunit 15', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72109587, end_position_on_the_genomic_accession=72112776, orientation='minus', exon_count=9, OMIM=614717, no_hdr0=''),
  25907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25907, CurrentID=0, Status='live', Symbol='TMEM158', Aliases='BBP, RIS1, p40BBP', description='transmembrane protein 158 (gene/pseudogene)', other_designations='40 kDa BINP-binding protein|BINP receptor|Ras induced senescence 1|brain injury-derived neurotrophic peptide (BINP) binding protein|brain specific binding protein', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45224464, end_position_on_the_genomic_accession=45226322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  25909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25909, CurrentID=0, Status='live', Symbol='AHCTF1', Aliases='ELYS, MST108, MSTP108, TMBS62', description='AT-hook containing transcription factor 1', other_designations='AT hook containing transcription factor 1|ELYS transcription factor like protein TMBS62|ELYS transcription factor-like protein TMBS62|embryonic large molecule derived from yolk sac|putative AT-hook-containing transcription factor 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246839098, end_position_on_the_genomic_accession=246931978, orientation='minus', exon_count=39, OMIM=610853, no_hdr0=''),
  25911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25911, CurrentID=0, Status='live', Symbol='DPCD', Aliases='', description='deleted in primary ciliary dyskinesia homolog (mouse)', other_designations='RP11-529I10.4|deleted in a mouse model of primary ciliary dyskinesia', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101588332, end_position_on_the_genomic_accession=101609653, orientation='plus', exon_count=6, OMIM=616467, no_hdr0=''),
  25912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25912, CurrentID=0, Status='live', Symbol='C1orf43', Aliases='HSPC012, NICE-3, NS5ATP4, S863-3', description='chromosome 1 open reading frame 43', other_designations='HCV NS5A-transactivated protein 4|hepatitis C virus NS5A-transactivated protein 4', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154206701, end_position_on_the_genomic_accession=154220797, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  25913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25913, CurrentID=0, Status='live', Symbol='POT1', Aliases='CMM10, HPOT1', description='protection of telomeres 1', other_designations='POT1-like telomere end-binding protein|protection of telomeres 1 homolog', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124822386, end_position_on_the_genomic_accession=124929983, orientation='minus', exon_count=22, OMIM=606478, no_hdr0=''),
  25914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25914, CurrentID=0, Status='live', Symbol='RTTN', Aliases='', description='rotatin', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70003804, end_position_on_the_genomic_accession=70205726, orientation='minus', exon_count=53, OMIM=610436, no_hdr0=''),
  25915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25915, CurrentID=0, Status='live', Symbol='NDUFAF3', Aliases='2P1, C3orf60, E3-3', description='NADH:ubiquinone oxidoreductase complex assembly factor 3', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 3|NADH dehydrogenase (ubiquinone) complex I, assembly factor 3|nuclear protein E3-3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49020475, end_position_on_the_genomic_accession=49023495, orientation='plus', exon_count=7, OMIM=612911, no_hdr0=''),
  100689212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689212, CurrentID=0, Status='live', Symbol='PBC5', Aliases='', description='Biliary cirrhosis, primary, 5', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614221, no_hdr0=''),
  25917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25917, CurrentID=0, Status='live', Symbol='THUMPD3', Aliases='', description='THUMP domain containing 3', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9362976, end_position_on_the_genomic_accession=9386791, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100689214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689214, CurrentID=0, Status='live', Symbol='HYSP4', Aliases='', description='Hypospadias 4, X-linked, susceptibilty to', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300856, no_hdr0=''),
  100689215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689215, CurrentID=0, Status='live', Symbol='PBC4', Aliases='', description='Biliary cirrhosis, primary, 4', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614220, no_hdr0=''),
  25920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25920, CurrentID=0, Status='live', Symbol='NELFB', Aliases='COBRA1, NELF-B', description='negative elongation factor complex member B', other_designations='cofactor of BRCA1|negative elongation factor protein B', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137255307, end_position_on_the_genomic_accession=137273548, orientation='plus', exon_count=13, OMIM=611180, no_hdr0=''),
  25921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25921, CurrentID=0, Status='live', Symbol='ZDHHC5', Aliases='DHHC5, ZNF375', description='zinc finger, DHHC-type containing 5', other_designations='DHHC-5|membrane-associated DHHC5 zinc finger protein|probable palmitoyltransferase ZDHHC5|zinc finger DHHC domain-containing protein 5|zinc finger protein 375|zinc finger, DHHC domain containing 5', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57668002, end_position_on_the_genomic_accession=57701187, orientation='plus', exon_count=15, OMIM=614586, no_hdr0=''),
  25923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25923, CurrentID=0, Status='live', Symbol='ATL3', Aliases='HSN1F', description='atlastin GTPase 3', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63624082, end_position_on_the_genomic_accession=63671974, orientation='minus', exon_count=16, OMIM=609369, no_hdr0=''),
  25924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25924, CurrentID=0, Status='live', Symbol='MYRIP', Aliases='SLAC2-C, SLAC2C', description='myosin VIIA and Rab interacting protein', other_designations='Slp homologue lacking C2 domains|exophilin-8|myosin-VIIa- and Rab-interacting protein|slp homolog lacking C2 domains c|synaptotagmin-like protein homologue lacking C2 domains-c|synaptotagmin-like protein lacking C2 domains C', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39808914, end_position_on_the_genomic_accession=40260321, orientation='plus', exon_count=20, OMIM=611790, no_hdr0=''),
  25925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25925, CurrentID=0, Status='live', Symbol='ZNF521', Aliases='EHZF, Evi3', description='zinc finger protein 521', other_designations='LYST-interacting protein 3|early hematopoietic zinc finger protein', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=25061924, end_position_on_the_genomic_accession=25352250, orientation='minus', exon_count=11, OMIM=610974, no_hdr0=''),
  25926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25926, CurrentID=0, Status='live', Symbol='NOL11', Aliases='', description='nucleolar protein 11', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67717833, end_position_on_the_genomic_accession=67744208, orientation='plus', exon_count=18, OMIM=615366, no_hdr0=''),
  25927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25927, CurrentID=0, Status='live', Symbol='CNRIP1', Aliases='C2orf32, CRIP1', description='cannabinoid receptor interacting protein 1', other_designations='CRIP-1|cannabinoid receptor CB1-interacting protein 1', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68284171, end_position_on_the_genomic_accession=68320051, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  25928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25928, CurrentID=0, Status='live', Symbol='SOSTDC1', Aliases='CDA019, ECTODIN, USAG1', description='sclerostin domain containing 1', other_designations='cystine-knot containing secreted protein|ectodermal BMP inhibitor|uterine sensitization-associated protein-1', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16461481, end_position_on_the_genomic_accession=16465849, orientation='minus', exon_count=2, OMIM=609675, no_hdr0=''),
  25929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25929, CurrentID=0, Status='live', Symbol='GEMIN5', Aliases='GEMIN-5', description='gem nuclear organelle associated protein 5', other_designations='gem (nuclear organelle) associated protein 5', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154887416, end_position_on_the_genomic_accession=154938216, orientation='minus', exon_count=28, OMIM=607005, no_hdr0=''),
  25930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25930, CurrentID=0, Status='live', Symbol='PTPN23', Aliases='HD-PTP, HDPTP, PTP-TD14', description='protein tyrosine phosphatase, non-receptor type 23', other_designations='his domain-containing protein tyrosine phosphatase|protein tyrosine phosphatase TD14', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47380982, end_position_on_the_genomic_accession=47413441, orientation='plus', exon_count=25, OMIM=606584, no_hdr0=''),
  25932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25932, CurrentID=0, Status='live', Symbol='CLIC4', Aliases='CLIC4L, H1, MTCLIC, huH1, p64H1', description='chloride intracellular channel 4', other_designations='chloride intracellular channel 4 like|intracellular chloride ion channel protein p64H1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24745269, end_position_on_the_genomic_accession=24844324, orientation='plus', exon_count=7, OMIM=606536, no_hdr0=''),
  100689229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689229, CurrentID=0, Status='live', Symbol='H3F3AP3', Aliases='p18', description='H3 histone, family 3A pseudogene 3', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109408516, end_position_on_the_genomic_accession=109410202, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25934, CurrentID=0, Status='live', Symbol='NIPSNAP3A', Aliases='HSPC299, NIPSNAP4, TASSC', description='nipsnap homolog 3A (C. elegans)', other_designations='protein NipSnap homolog 4|target for Salmonella secreted protein C', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104747688, end_position_on_the_genomic_accession=104760122, orientation='plus', exon_count=6, OMIM=608871, no_hdr0=''),
  25936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25936, CurrentID=0, Status='live', Symbol='NSL1', Aliases='C1orf48, DC8, MIS14', description='NSL1, MIS12 kinetochore complex component', other_designations='NSL1, MIND kinetochore complex component, homolog', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212726153, end_position_on_the_genomic_accession=212791797, orientation='minus', exon_count=7, OMIM=609174, no_hdr0=''),
  25937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25937, CurrentID=0, Status='live', Symbol='WWTR1', Aliases='TAZ', description='WW domain containing transcription regulator 1', other_designations='transcriptional co-activator with PDZ-binding motif|transcriptional coactivator with PDZ-binding motif', map_location='3q23-q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149517235, end_position_on_the_genomic_accession=149703273, orientation='minus', exon_count=9, OMIM=607392, no_hdr0=''),
  25938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25938, CurrentID=0, Status='live', Symbol='HEATR5A', Aliases='C14orf125', description='HEAT repeat containing 5A', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31291788, end_position_on_the_genomic_accession=31420582, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  25939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25939, CurrentID=0, Status='live', Symbol='SAMHD1', Aliases='CHBL2, DCIP, HDDC1, MOP-5, SBBI88', description='SAM domain and HD domain 1', other_designations='SAM domain and HD domain-containing protein 1|dNTPase|dendritic cell-derived IFNG-induced protein|monocyte protein 5', map_location='20pter-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36891824, end_position_on_the_genomic_accession=36951843, orientation='minus', exon_count=17, OMIM=606754, no_hdr0=''),
  25940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25940, CurrentID=0, Status='live', Symbol='FAM98A', Aliases='', description='family with sequence similarity 98 member A', other_designations='family with sequence similarity 98, member A', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33583661, end_position_on_the_genomic_accession=33599362, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  25941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25941, CurrentID=0, Status='live', Symbol='TPGS2', Aliases='C18orf10, HMFN0601, L17, PGs2', description='tubulin polyglutamylase complex subunit 2', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36780024, end_position_on_the_genomic_accession=36829393, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  25942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25942, CurrentID=0, Status='live', Symbol='SIN3A', Aliases='', description='SIN3 transcription regulator family member A', other_designations='SIN3 homolog A, transcription regulator|SIN3 transcription regulator homolog A|histone deacetylase complex subunit Sin3a|transcriptional co-repressor Sin3A|transcriptional corepressor Sin3a|transcriptional regulator, SIN3A', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75369379, end_position_on_the_genomic_accession=75455815, orientation='minus', exon_count=26, OMIM=607776, no_hdr0=''),
  25943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25943, CurrentID=0, Status='live', Symbol='C20orf194', Aliases='', description='chromosome 20 open reading frame 194', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3249302, end_position_on_the_genomic_accession=3407662, orientation='minus', exon_count=39, OMIM=614146, no_hdr0=''),
  25945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25945, CurrentID=0, Status='live', Symbol='PVRL3', Aliases='CD113, CDW113, NECTIN-3, PPR3, PRR3, PVRR3', description='poliovirus receptor-related 3', other_designations='nectin 3|poliovirus receptor-related protein 3', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111071759, end_position_on_the_genomic_accession=111269283, orientation='plus', exon_count=15, OMIM=607147, no_hdr0=''),
  25946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25946, CurrentID=0, Status='live', Symbol='ZNF385A', Aliases='HZF, RZF, ZFP385, ZNF385', description='zinc finger protein 385A', other_designations='hematopoietic zinc finger protein|retinal zinc finger protein|zinc finger protein 385', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54369129, end_position_on_the_genomic_accession=54391305, orientation='minus', exon_count=10, OMIM=609124, no_hdr0=''),
  25948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25948, CurrentID=0, Status='live', Symbol='KBTBD2', Aliases='BKLHD1', description='kelch repeat and BTB (POZ) domain containing 2', other_designations='BTB and kelch domain containing 1|BTB and kelch domain-containing protein 1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32868166, end_position_on_the_genomic_accession=32892144, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  25949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25949, CurrentID=0, Status='live', Symbol='SYF2', Aliases='CBPIN, NTC31, P29, fSAP29', description='SYF2 pre-mRNA-splicing factor', other_designations='CCNDBP1 interactor|CCNDBP1-interactor|GCIP-interacting protein p29|SYF2 homolog, RNA splicing factor|functional spliceosome-associated protein 29', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25222276, end_position_on_the_genomic_accession=25232522, orientation='minus', exon_count=7, OMIM=607090, no_hdr0=''),
  25950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25950, CurrentID=0, Status='live', Symbol='RWDD3', Aliases='RSUME', description='RWD domain containing 3', other_designations='RWD domain-containing sumoylation enhancer|RWD-containing sumoylation enhancer', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95234155, end_position_on_the_genomic_accession=95247225, orientation='plus', exon_count=5, OMIM=615875, no_hdr0=''),
  646813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646813, CurrentID=0, Status='live', Symbol='LOC646813', Aliases='', description='DEAH (Asp-Glu-Ala-His) box helicase 9 pseudogene', other_designations='DDX9 pseudogene|DEAH (Asp-Glu-Ala-His) box polypeptide 9 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50409147, end_position_on_the_genomic_accession=50420631, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  25953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25953, CurrentID=0, Status='live', Symbol='PNKD', Aliases='BRP17, DYT8, FKSG19, FPD1, KIPP1184, MR-1, MR1, PDC, PKND1, TAHCCP2', description='paroxysmal nonkinesigenic dyskinesia', other_designations='brain protein 17|myofibrillogenesis regulator 1|trans-activated by hepatitis C virus core protein 2', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218270392, end_position_on_the_genomic_accession=218346793, orientation='plus', exon_count=12, OMIM=609023, no_hdr0=''),
  25956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25956, CurrentID=0, Status='live', Symbol='SEC31B', Aliases='SEC31B-1, SEC31L2', description='SEC31 homolog B, COPII coat complex component', other_designations='SEC31 homolog B|SEC31 homolog B, COPII coating complex component|SEC31-like 2|SEC31-like protein 2|SEC31-related protein B|secretory pathway component Sec31B-1', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100486646, end_position_on_the_genomic_accession=100519838, orientation='minus', exon_count=26, OMIM=610258, no_hdr0=''),
  135398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135398, CurrentID=0, Status='live', Symbol='C6orf141', Aliases='', description='chromosome 6 open reading frame 141', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49550400, end_position_on_the_genomic_accession=49560572, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25959, CurrentID=0, Status='live', Symbol='KANK2', Aliases='ANKRD25, MXRA3, PPKWH, SIP', description='KN motif and ankyrin repeat domains 2', other_designations='SRC-1 interacting protein|SRC-1-interacting protein|SRC-interacting protein|SRC1-interacting protein|ankyrin repeat domain 25|ankyrin repeat domain-containing protein 25|kidney ankyrin repeat-containing protein 2|matrix-remodeling-associated protein 3|matrix-remodelling associated 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11164267, end_position_on_the_genomic_accession=11197567, orientation='minus', exon_count=13, OMIM=614610, no_hdr0=''),
  25960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25960, CurrentID=0, Status='live', Symbol='ADGRA2', Aliases='GPR124, TEM5', description='adhesion G protein-coupled receptor A2', other_designations='G protein-coupled receptor 124|tumor endothelial marker 5', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37796883, end_position_on_the_genomic_accession=37843986, orientation='plus', exon_count=20, OMIM=606823, no_hdr0=''),
  25961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25961, CurrentID=0, Status='live', Symbol='NUDT13', Aliases='', description='nudix hydrolase 13', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 13|nudix motif 13', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73110375, end_position_on_the_genomic_accession=73131828, orientation='plus', exon_count=10, OMIM=609233, no_hdr0=''),
  25962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25962, CurrentID=0, Status='live', Symbol='KIAA1429', Aliases='MSTP054, fSAP121', description='KIAA1429', other_designations='functional spliceosome-associated protein 121', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94487777, end_position_on_the_genomic_accession=94553518, orientation='minus', exon_count=24, OMIM=616447, no_hdr0=''),
  25963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25963, CurrentID=0, Status='live', Symbol='TMEM87A', Aliases='', description='transmembrane protein 87A', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42210452, end_position_on_the_genomic_accession=42274526, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  25966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25966, CurrentID=0, Status='live', Symbol='C2CD2', Aliases='C21orf25, C21orf258, TMEM24L', description='C2 calcium-dependent domain containing 2', other_designations='TMEM24 related geme', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41885110, end_position_on_the_genomic_accession=41953957, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105370387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370387, CurrentID=0, Status='live', Symbol='LOC105370387', Aliases='', description='uncharacterized LOC105370387', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  25970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25970, CurrentID=0, Status='live', Symbol='SH2B1', Aliases='PSM, SH2B', description='SH2B adaptor protein 1', other_designations='SH2 domain-containing protein 1B|SH2 domain-containing putative adapter SH2-B|SH2-B signaling protein|pro-rich, PH and SH2 domain-containing signaling mediator', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28846600, end_position_on_the_genomic_accession=28874213, orientation='plus', exon_count=14, OMIM=608937, no_hdr0=''),
  25972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25972, CurrentID=0, Status='live', Symbol='UNC50', Aliases='GMH1, HSD23, UNCL, URP', description='unc-50 homolog (C. elegans)', other_designations='PDLs22|geal-6 membrane-associated high-copy suppressor 1|hGMH1|periodontal ligament-specific protein 22|protein GMH1 homolog|unc-50 related|unc-50-like protein|uncoordinated-like protein', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98608579, end_position_on_the_genomic_accession=98618515, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  25973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25973, CurrentID=0, Status='live', Symbol='PARS2', Aliases='MT-PRORS, proRS', description='prolyl-tRNA synthetase 2, mitochondrial (putative)', other_designations='proline tRNA ligase 2, mitochondrial (putative)', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54755519, end_position_on_the_genomic_accession=54764553, orientation='minus', exon_count=3, OMIM=612036, no_hdr0=''),
  25974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25974, CurrentID=0, Status='live', Symbol='MMACHC', Aliases='cblC', description='methylmalonic aciduria (cobalamin deficiency) cblC type, with homocystinuria', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45500184, end_position_on_the_genomic_accession=45511067, orientation='plus', exon_count=4, OMIM=609831, no_hdr0=''),
  25975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25975, CurrentID=0, Status='live', Symbol='EGFL6', Aliases='MAEG, W80', description='EGF-like-domain, multiple 6', other_designations='EGF repeat-containing protein 6|EGF-like protein 6|MAM and EGF domain containing|MAM and EGF domains-containing gene protein', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13569575, end_position_on_the_genomic_accession=13633575, orientation='plus', exon_count=12, OMIM=300239, no_hdr0=''),
  25976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25976, CurrentID=0, Status='live', Symbol='TIPARP', Aliases='ARTD14, PARP7, pART14', description='TCDD-inducible poly(ADP-ribose) polymerase', other_designations='ADP-ribosyltransferase diphtheria toxin-like 14|poly [ADP-ribose] polymerase 7', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156674416, end_position_on_the_genomic_accession=156706770, orientation='plus', exon_count=11, OMIM=612480, no_hdr0=''),
  25977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25977, CurrentID=0, Status='live', Symbol='NECAP1', Aliases='EIEE21', description='NECAP endocytosis associated 1', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8082211, end_position_on_the_genomic_accession=8097777, orientation='plus', exon_count=8, OMIM=611623, no_hdr0=''),
  25978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25978, CurrentID=0, Status='live', Symbol='CHMP2B', Aliases='ALS17, CHMP2.5, DMT1, VPS2-2, VPS2B', description='charged multivesicular body protein 2B', other_designations='VPS2 homolog B|chromatin modifying protein 2B|vacuolar protein-sorting-associated protein 2-2', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87227263, end_position_on_the_genomic_accession=87255548, orientation='plus', exon_count=7, OMIM=609512, no_hdr0=''),
  25979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25979, CurrentID=0, Status='live', Symbol='DHRS7B', Aliases='CGI-93, SDR32C1', description='dehydrogenase/reductase (SDR family) member 7B', other_designations='short chain dehydrogenase/reductase family 32C, member 1|short-chain dehydrogenase/reductase family 32C member 1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21126275, end_position_on_the_genomic_accession=21191972, orientation='plus', exon_count=9, OMIM=616160, no_hdr0=''),
  25980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25980, CurrentID=0, Status='live', Symbol='AAR2', Aliases='C20orf4, CGI-23', description='AAR2 splicing factor homolog', other_designations='', map_location='20pter-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36236417, end_position_on_the_genomic_accession=36256941, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  25981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25981, CurrentID=0, Status='live', Symbol='DNAH1', Aliases='DNAHC1, HDHC7, HL-11, HL11, HSRF-1, XLHSRF-1', description='dynein, axonemal, heavy chain 1', other_designations='DNAH1 variant protein|axonemal beta dynein heavy chain 1|ciliary dynein heavy chain 1|dynein, axonemal, heavy polypeptide 1|heat shock regulated protein 1', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52316319, end_position_on_the_genomic_accession=52400497, orientation='plus', exon_count=81, OMIM=603332, no_hdr0=''),
  25983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25983, CurrentID=0, Status='live', Symbol='NGDN', Aliases='C14orf120, CANu1, LCP5, NGD, lpd-2', description='neuroguidin', other_designations='EIF4E-binding protein|centromere accumulated nuclear protein 1|eukaryotic initiation factor 4E and CPEB binding protein|neuroguidin, EIF4E binding protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23469628, end_position_on_the_genomic_accession=23478193, orientation='plus', exon_count=10, OMIM=610777, no_hdr0=''),
  25984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25984, CurrentID=0, Status='live', Symbol='KRT23', Aliases='CK23, HAIK1, K23', description='keratin 23, type I', other_designations='CK-23|cytokeratin 23|cytokeratin-23|histone deacetylase inducible keratin 23|hyperacetylation-inducible type I keratin|keratin 23 (histone deacetylase inducible)|keratin-23|type I intermediate filament cytokeratin', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40922696, end_position_on_the_genomic_accession=40937643, orientation='minus', exon_count=10, OMIM=606194, no_hdr0=''),
  25987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25987, CurrentID=0, Status='live', Symbol='TSKU', Aliases='E2IG4, LRRC54, TSK', description='tsukushi, small leucine rich proteoglycan', other_designations='E2-induced gene 4 protein|leucine rich repeat containing 54|leucine-rich repeat-containing protein 54|tsukushi homolog|tsukushi small leucine rich proteoglycan homolog', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76782313, end_position_on_the_genomic_accession=76798154, orientation='plus', exon_count=4, OMIM=608015, no_hdr0=''),
  25988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25988, CurrentID=0, Status='live', Symbol='HINFP', Aliases='HiNF-P, MIZF, ZNF743', description='histone H4 transcription factor', other_designations='MBD2 (methyl-CpG-binding protein)-interacting zinc finger protein|MBD2-interacting zinc finger 1|MBD2-interacting zinc finger protein|histone H4 gene-specific protein HiNF-P|histone nuclear factor P|methyl-CpG-binding protein 2-interacting zinc finger protein', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119121523, end_position_on_the_genomic_accession=119135055, orientation='plus', exon_count=13, OMIM=607099, no_hdr0=''),
  25989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25989, CurrentID=0, Status='live', Symbol='ULK3', Aliases='', description='unc-51 like kinase 3', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74836116, end_position_on_the_genomic_accession=74843211, orientation='minus', exon_count=17, OMIM=613472, no_hdr0=''),
  25992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25992, CurrentID=0, Status='live', Symbol='SNED1', Aliases='SST3, Snep', description='sushi, nidogen and EGF-like domains 1', other_designations='6720455I24Rik homolog|IRE-BP1|insulin responsive sequence DNA binding protein-1|insulin-responsive sequence DNA-binding protein 1', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240997597, end_position_on_the_genomic_accession=241094228, orientation='plus', exon_count=37, OMIM=616634, no_hdr0=''),
  25994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25994, CurrentID=0, Status='live', Symbol='HIGD1A', Aliases='HIG1, RCF1a', description='HIG1 hypoxia inducible domain family member 1A', other_designations='HIG1 domain family, member 1A|HIG1 hypoxia inducible domain family, member 1A|RCF1 homolog A|hypoxia inducible gene 1|hypoxia-inducible gene 1 protein', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42782908, end_position_on_the_genomic_accession=42804535, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  25996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25996, CurrentID=0, Status='live', Symbol='REXO2', Aliases='CGI-114, REX2, RFN, SFN', description='RNA exonuclease 2', other_designations='REX2, RNA exonuclease 2 homolog|RNA exonuclease 2 homolog|small fragment nuclease', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114439386, end_position_on_the_genomic_accession=114450279, orientation='plus', exon_count=7, OMIM=607149, no_hdr0=''),
  25998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25998, CurrentID=0, Status='live', Symbol='IBTK', Aliases='BTBD26, BTKI', description='inhibitor of Bruton tyrosine kinase', other_designations="BTK-binding protein|Bruton agammaglobulinemia tyrosine kinase inhibitor|inhibitor of Bruton agammaglobulinemia tyrosine kinase|inhibitor of Bruton's tyrosine kinase|inhibitor of Bruton's tyrosine kinase-alpha|inhibitor of Bruton's tyrosine kinase-beta|inhibitor of Bruton's tyrosine kinase-gamma", map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82169987, end_position_on_the_genomic_accession=82247754, orientation='minus', exon_count=30, OMIM=606457, no_hdr0=''),
  25999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25999, CurrentID=0, Status='live', Symbol='CLIP3', Aliases='CLIPR-59, CLIPR59, RSNL1', description='CAP-Gly domain containing linker protein 3', other_designations='CLIP-170-related 59 kDa protein|cytoplasmic linker protein 170-related 59 kDa protein|restin-like 1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36014660, end_position_on_the_genomic_accession=36032895, orientation='minus', exon_count=14, OMIM=607382, no_hdr0=''),
  26000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26000, CurrentID=0, Status='live', Symbol='TBC1D10B', Aliases='EPI64B, FP2461', description='TBC1 domain family member 10B', other_designations='Rab27A-GAPbeta|TBC1 domain family, member 10B|rab27A-GAP-beta', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30357101, end_position_on_the_genomic_accession=30370258, orientation='minus', exon_count=9, OMIM=613620, no_hdr0=''),
  26001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26001, CurrentID=0, Status='live', Symbol='RNF167', Aliases='5730408C10Rik, LP2254, RING105', description='ring finger protein 167', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4940033, end_position_on_the_genomic_accession=4945222, orientation='plus', exon_count=11, OMIM=610431, no_hdr0=''),
  26002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26002, CurrentID=0, Status='live', Symbol='MOXD1', Aliases='MOX, PRO5780, dJ248E1.1', description='monooxygenase, DBH-like 1', other_designations='dopamine-oxygenase|monooxygenase X', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132296055, end_position_on_the_genomic_accession=132401534, orientation='minus', exon_count=12, OMIM=609000, no_hdr0=''),
  26003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26003, CurrentID=0, Status='live', Symbol='GORASP2', Aliases='GOLPH6, GRASP55, GRS2, p59', description='golgi reassembly stacking protein 2', other_designations='golgi phosphoprotein 6|golgi reassembly stacking protein 2, 55kDa', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170928438, end_position_on_the_genomic_accession=170967135, orientation='plus', exon_count=11, OMIM=608693, no_hdr0=''),
  26005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26005, CurrentID=0, Status='live', Symbol='C2CD3', Aliases='OFD14', description='C2 calcium-dependent domain containing 3', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74012714, end_position_on_the_genomic_accession=74171210, orientation='minus', exon_count=40, OMIM=615944, no_hdr0=''),
  26007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26007, CurrentID=0, Status='live', Symbol='TKFC', Aliases='DAK, NET45', description='triokinase/FMN cyclase', other_designations='ATP-dependent dihydroxyacetone kinase|DHA kinase|Dha kinase/FMN cyclase|FAD-AMP lyase cyclic FMN forming|FAD-AMP lyase cyclizing|bifunctional ATP-dependent dihydroxyacetone kinase/FAD-AMP lyase (cyclizing)|dihydroxyacetone kinase 2 homolog|glycerone kinase', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61333182, end_position_on_the_genomic_accession=61353139, orientation='plus', exon_count=23, OMIM=615844, no_hdr0=''),
  26009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26009, CurrentID=0, Status='live', Symbol='ZZZ3', Aliases='ATAC1', description='zinc finger, ZZ-type containing 3', other_designations='ATAC component 1 homolog|zinc finger, ZZ domain containing 3', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77561521, end_position_on_the_genomic_accession=77683426, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26010, CurrentID=0, Status='live', Symbol='SPATS2L', Aliases='DNAPTP6, SGNP', description='spermatogenesis associated, serine rich 2 like', other_designations='DNA polymerase transactivated protein 6|DNA polymerase-transactivated protein 6|spermatogenesis associated, serine-rich 2-like|stress granule and nucleolar protein', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200305881, end_position_on_the_genomic_accession=200482263, orientation='plus', exon_count=19, OMIM=613817, no_hdr0=''),
  26011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26011, CurrentID=0, Status='live', Symbol='TENM4', Aliases='Doc4, ODZ4, TNM4, Ten-M4', description='teneurin transmembrane protein 4', other_designations='odd Oz/ten-m homolog 4|odz, odd Oz/ten-m homolog 4|protein Odd Oz/ten-m homolog 4|ten-4|tenascin-M4', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78653283, end_position_on_the_genomic_accession=79440989, orientation='minus', exon_count=40, OMIM=610084, no_hdr0=''),
  26012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26012, CurrentID=0, Status='live', Symbol='NSMF', Aliases='HH9, NELF', description='NMDA receptor synaptonuclear signaling and neuronal migration factor', other_designations='nasal embryonic LHRH factor|nasal embryonic luteinizing hormone-releasing hormone factor', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137447570, end_position_on_the_genomic_accession=137459357, orientation='minus', exon_count=17, OMIM=608137, no_hdr0=''),
  26013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26013, CurrentID=0, Status='live', Symbol='L3MBTL1', Aliases='H-L(3)MBT, L3MBTL, ZC2HC3, dJ138B7.3', description='l(3)mbt-like 1 (Drosophila)', other_designations='l(3)mbt protein homolog|lethal (3) malignant brain tumor l(3)', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43507680, end_position_on_the_genomic_accession=43541895, orientation='plus', exon_count=22, OMIM=608802, no_hdr0=''),
  26015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26015, CurrentID=0, Status='live', Symbol='RPAP1', Aliases='', description='RNA polymerase II associated protein 1', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41517176, end_position_on_the_genomic_accession=41544277, orientation='minus', exon_count=26, OMIM=611475, no_hdr0=''),
  26017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26017, CurrentID=0, Status='live', Symbol='FAM32A', Aliases='OTAG-12, OTAG12', description='family with sequence similarity 32 member A', other_designations='family with sequence similarity 32, member A|ovarian tumor associated gene-12|ovarian tumor-associated gene 12', map_location='19pter-p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16185184, end_position_on_the_genomic_accession=16192046, orientation='plus', exon_count=4, OMIM=614554, no_hdr0=''),
  26018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26018, CurrentID=0, Status='live', Symbol='LRIG1', Aliases='LIG-1, LIG1', description='leucine-rich repeats and immunoglobulin-like domains 1', other_designations='leucine-rich repeat protein LRIG1|ortholog of mouse integral membrane glycoprotein LIG-1', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66378797, end_position_on_the_genomic_accession=66501263, orientation='minus', exon_count=20, OMIM=608868, no_hdr0=''),
  26019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26019, CurrentID=0, Status='live', Symbol='UPF2', Aliases='HUPF2, RENT2, smg-3', description='UPF2 regulator of nonsense transcripts homolog (yeast)', other_designations='nonsense mRNA reducing factor 2|smg-3 homolog, nonsense mediated mRNA decay factor|up-frameshift suppressor 2 homolog|yeast Upf2p homolog', map_location='10p14-p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11920022, end_position_on_the_genomic_accession=12043172, orientation='minus', exon_count=24, OMIM=605529, no_hdr0=''),
  26020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26020, CurrentID=0, Status='live', Symbol='LRP10', Aliases='LRP9, MST087, MSTP087', description='LDL receptor related protein 10', other_designations='LRP-10|low density lipoprotein receptor-related protein 10', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22871613, end_position_on_the_genomic_accession=22881580, orientation='plus', exon_count=7, OMIM=609921, no_hdr0=''),
  26022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26022, CurrentID=0, Status='live', Symbol='TMEM98', Aliases='TADA1', description='transmembrane protein 98', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32927910, end_position_on_the_genomic_accession=32941649, orientation='plus', exon_count=9, OMIM=615949, no_hdr0=''),
  26024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26024, CurrentID=0, Status='live', Symbol='PTCD1', Aliases='', description='pentatricopeptide repeat domain 1', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99416739, end_position_on_the_genomic_accession=99438839, orientation='minus', exon_count=8, OMIM=614774, no_hdr0=''),
  26025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26025, CurrentID=0, Status='live', Symbol='PCDHGA12', Aliases='CDH21, FIB3, PCDH-GAMMA-A12', description='protocadherin gamma subfamily A, 12', other_designations='cadherin 21|cadherin-21|fibroblast cadherin FIB3|fibroblast cadherin-3', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141430591, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=603059, no_hdr0=''),
  26027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26027, CurrentID=0, Status='live', Symbol='ACOT11', Aliases='BFIT, STARD14, THEA, THEM1', description='acyl-CoA thioesterase 11', other_designations='START domain containing 14|StAR-related lipid transfer (START) domain containing 14|acyl-CoA thioester hydrolase 11|adipose-associated thioesterase|brown fat-inducible thioesterase|thioesterase superfamily member 1|thioesterase, adipose associated', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54548134, end_position_on_the_genomic_accession=54634744, orientation='plus', exon_count=18, OMIM=606803, no_hdr0=''),
  26030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26030, CurrentID=0, Status='live', Symbol='PLEKHG3', Aliases='ARHGEF43, KIAA0599', description='pleckstrin homology and RhoGEF domain containing G3', other_designations='PH domain-containing family G member 3|pleckstrin homology domain containing, family G (with RhoGef domain) member 3|pleckstrin homology domain containing, family G, member 3', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64704413, end_position_on_the_genomic_accession=64746905, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  26031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26031, CurrentID=0, Status='live', Symbol='OSBPL3', Aliases='ORP-3, ORP3, OSBP3', description='oxysterol binding protein like 3', other_designations='OSBP-related protein 3|oxysterol binding protein-like 3', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24796537, end_position_on_the_genomic_accession=24980212, orientation='minus', exon_count=26, OMIM=606732, no_hdr0=''),
  26032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26032, CurrentID=0, Status='live', Symbol='SUSD5', Aliases='', description='sushi domain containing 5', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33150045, end_position_on_the_genomic_accession=33219215, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26033, CurrentID=0, Status='live', Symbol='ATRNL1', Aliases='ALP, bA338L11.1, bA454H24.1', description='attractin like 1', other_designations='attractin-like 1', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=115093365, end_position_on_the_genomic_accession=115948999, orientation='plus', exon_count=35, OMIM=612869, no_hdr0=''),
  26034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26034, CurrentID=0, Status='live', Symbol='IPCEF1', Aliases='PIP3-E', description='interaction protein for cytohesin exchange factors 1', other_designations='interaction protein for cytohesin exchange factors 1 Interaction protein for cytohesin exchange factors 1|phosphoinositide binding protein PIP3-E|phosphoinositide-binding protein PIP3-E', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154154483, end_position_on_the_genomic_accession=154356802, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  26035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26035, CurrentID=0, Status='live', Symbol='GLCE', Aliases='HSEPI', description='glucuronic acid epimerase', other_designations='UDP-glucuronic acid epimerase|glucuronyl C5-epimerase|heparan sulfate C5-epimerase|heparan sulfate epimerase|heparin/heparan sulfate-glucuronic acid C5-epimerase|heparin/heparan sulfate:glucuronic acid C5-epimerase|heparosan-N-sulfate-glucuronate 5-epimerase', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69160560, end_position_on_the_genomic_accession=69272207, orientation='plus', exon_count=7, OMIM=612134, no_hdr0=''),
  26036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26036, CurrentID=0, Status='live', Symbol='ZNF451', Aliases='COASTER, dJ417I1.1', description='zinc finger protein 451', other_designations='coactivator for steroid receptors', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57090010, end_position_on_the_genomic_accession=57170305, orientation='plus', exon_count=18, OMIM=615708, no_hdr0=''),
  26037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26037, CurrentID=0, Status='live', Symbol='SIPA1L1', Aliases='E6TP1', description='signal-induced proliferation-associated 1 like 1', other_designations='SIPA1-like protein 1|high-risk human papilloma viruses E6 oncoproteins targeted protein 1', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71321396, end_position_on_the_genomic_accession=71741229, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  26038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26038, CurrentID=0, Status='live', Symbol='CHD5', Aliases='CHD-5', description='chromodomain helicase DNA binding protein 5', other_designations='ATP-dependent helicase CHD5', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6101787, end_position_on_the_genomic_accession=6180134, orientation='minus', exon_count=42, OMIM=610771, no_hdr0=''),
  26039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26039, CurrentID=0, Status='live', Symbol='SS18L1', Aliases='CREST, LP2261', description='synovial sarcoma translocation gene on chromosome 18-like 1', other_designations='SS18-like protein 1|SYT homolog 1', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62143720, end_position_on_the_genomic_accession=62182510, orientation='plus', exon_count=16, OMIM=606472, no_hdr0=''),
  26040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26040, CurrentID=0, Status='live', Symbol='SETBP1', Aliases='MRD29, SEB', description='SET binding protein 1', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44680173, end_position_on_the_genomic_accession=45068510, orientation='plus', exon_count=8, OMIM=611060, no_hdr0=''),
  101060852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060852, CurrentID=0, Status='live', Symbol='LOC101060852', Aliases='', description='protein phosphatase 1 regulatory subunit 26-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18220547, end_position_on_the_genomic_accession=18224162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26043, CurrentID=0, Status='live', Symbol='UBXN7', Aliases='UBXD7', description='UBX domain protein 7', other_designations='UBX domain containing 7', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196353490, end_position_on_the_genomic_accession=196454318, orientation='minus', exon_count=12, OMIM=616379, no_hdr0=''),
  26045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26045, CurrentID=0, Status='live', Symbol='LRRTM2', Aliases='', description='leucine rich repeat transmembrane neuronal 2', other_designations='leucine-rich repeat neuronal 2 protein|leucine-rich repeat transmembrane neuronal 2 protein', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138869346, end_position_on_the_genomic_accession=138875368, orientation='minus', exon_count=2, OMIM=610868, no_hdr0=''),
  26046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26046, CurrentID=0, Status='live', Symbol='LTN1', Aliases='C21orf10, C21orf98, RNF160, ZNF294', description='listerin E3 ubiquitin protein ligase 1', other_designations='ring finger protein 160|zinc finger protein 294', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28928144, end_position_on_the_genomic_accession=28992956, orientation='minus', exon_count=30, OMIM=613083, no_hdr0=''),
  26047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26047, CurrentID=0, Status='live', Symbol='CNTNAP2', Aliases='AUTS15, CASPR2, CDFE, NRXN4, PTHSL1', description='contactin associated protein-like 2', other_designations='cell recognition molecule Caspr2|homolog of Drosophila neurexin IV', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=146116361, end_position_on_the_genomic_accession=148420998, orientation='plus', exon_count=25, OMIM=604569, no_hdr0=''),
  26048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26048, CurrentID=0, Status='live', Symbol='ZNF500', Aliases='ZKSCAN18, ZSCAN50', description='zinc finger protein 500', other_designations='zinc finger protein with KRAB and SCAN domains 18', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4745987, end_position_on_the_genomic_accession=4767218, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26049, CurrentID=0, Status='live', Symbol='FAM169A', Aliases='SLAP75', description='family with sequence similarity 169 member A', other_designations='family with sequence similarity 169, member A|protein FAM169A|soluble lamina-associated protein of 75 kD', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74777574, end_position_on_the_genomic_accession=74866838, orientation='minus', exon_count=15, OMIM=615769, no_hdr0=''),
  26050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26050, CurrentID=0, Status='live', Symbol='SLITRK5', Aliases='LRRC11, bA364G4.2', description='SLIT and NTRK like family member 5', other_designations='SLIT and NTRK-like family, member 5|leucine rich repeat containing 11|leucine-rich repeat-containing protein 11|slit and trk like gene 5', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87671232, end_position_on_the_genomic_accession=87679616, orientation='plus', exon_count=4, OMIM=609680, no_hdr0=''),
  26051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26051, CurrentID=0, Status='live', Symbol='PPP1R16B', Aliases='ANKRD4, TIMAP', description='protein phosphatase 1 regulatory subunit 16B', other_designations='CAAX box protein TIMAP|TGF-beta-inhibited membrane-associated protein|ankyrin repeat domain protein 4|ankyrin repeat domain-containing protein 4|hTIMAP|protein phosphatase 1, regulatory (inhibitor) subunit 16B|protein phosphatase 1, regulatory subunit 16B', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38805705, end_position_on_the_genomic_accession=38923024, orientation='plus', exon_count=12, OMIM=613275, no_hdr0=''),
  26052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26052, CurrentID=0, Status='live', Symbol='DNM3', Aliases='Dyna III', description='dynamin 3', other_designations='T-dynamin|dynamin family member|dynamin, testicular', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171841478, end_position_on_the_genomic_accession=172418466, orientation='plus', exon_count=28, OMIM=611445, no_hdr0=''),
  26053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26053, CurrentID=0, Status='live', Symbol='AUTS2', Aliases='FBRSL2, MRD26', description='autism susceptibility candidate 2', other_designations='autism-related protein 1', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69598475, end_position_on_the_genomic_accession=70793068, orientation='plus', exon_count=23, OMIM=607270, no_hdr0=''),
  26054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26054, CurrentID=0, Status='live', Symbol='SENP6', Aliases='SSP1, SUSP1', description='SUMO1/sentrin specific peptidase 6', other_designations='2810017C20Rik|SUMO-1-specific protease 1|SUMO1/sentrin specific protease 6|sentrin/SUMO-specific protease SENP6', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75601509, end_position_on_the_genomic_accession=75718285, orientation='plus', exon_count=26, OMIM=605003, no_hdr0=''),
  26056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26056, CurrentID=0, Status='live', Symbol='RAB11FIP5', Aliases='GAF1, RIP11, pp75', description='RAB11 family interacting protein 5 (class I)', other_designations='gaf-1|gamma-SNAP-associated factor 1|phosphoprotein pp75|rab11-FIP5|rab11-interacting protein Rip11', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73073382, end_position_on_the_genomic_accession=73113018, orientation='minus', exon_count=7, OMIM=605536, no_hdr0=''),
  26057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26057, CurrentID=0, Status='live', Symbol='ANKRD17', Aliases='GTAR, MASK2, NY-BR-16', description='ankyrin repeat domain 17', other_designations='gene trap ankyrin repeat protein|serologically defined breast cancer antigen NY-BR-16', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73074785, end_position_on_the_genomic_accession=73259142, orientation='minus', exon_count=37, OMIM=615929, no_hdr0=''),
  26058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26058, CurrentID=0, Status='live', Symbol='GIGYF2', Aliases='GYF2, PARK11, PERQ2, PERQ3, TNRC15', description='GRB10 interacting GYF protein 2', other_designations='PERQ amino acid rich, with GYF domain 3|Parkinson disease (autosomal recessive, early onset) 11|trinucleotide repeat-containing gene 15 protein', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232697305, end_position_on_the_genomic_accession=232860577, orientation='plus', exon_count=35, OMIM=612003, no_hdr0=''),
  26059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26059, CurrentID=0, Status='live', Symbol='ERC2', Aliases='CAST, CAST1, ELKSL, SPBC110, Spc110', description='ELKS/RAB6-interacting/CAST family member 2', other_designations='CAZ-associated structural protein|cytomatrix protein p110', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55508308, end_position_on_the_genomic_accession=56468366, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  26060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26060, CurrentID=0, Status='live', Symbol='APPL1', Aliases='APPL, DIP13alpha', description='adaptor protein, phosphotyrosine interaction, PH domain and leucine zipper containing 1', other_designations='AKT2 interactor|adapter protein containing PH domain, PTB domain and leucine zipper motif 1|adaptor protein containing pH domain, PTB domain and leucine zipper motif 1|dip13-alpha|signaling adaptor protein DIP13alpha', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57227737, end_position_on_the_genomic_accession=57273471, orientation='plus', exon_count=23, OMIM=604299, no_hdr0=''),
  26061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26061, CurrentID=0, Status='live', Symbol='HACL1', Aliases='2-HPCL, HPCL, HPCL2, PHYH2', description='2-hydroxyacyl-CoA lyase 1', other_designations='1600020H07Rik|2-hydroxyphytanol-CoA lyase|2-hydroxyphytanoyl-CoA lyase|phytanoyl-CoA 2-hydroxylase 2|phytanoyl-CoA hydroxylase 2', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15560704, end_position_on_the_genomic_accession=15601852, orientation='minus', exon_count=17, OMIM=604300, no_hdr0=''),
  26062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26062, CurrentID=0, Status='live', Symbol='HYALP1', Aliases='HYAL6', description='hyaluronoglucosaminidase pseudogene 1', other_designations='', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123814139, end_position_on_the_genomic_accession=123819430, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26063, CurrentID=0, Status='live', Symbol='DECR2', Aliases='PDCR, SDR17C1', description='2,4-dienoyl-CoA reductase 2, peroxisomal', other_designations='2,4-dienoyl CoA reductase 2, peroxisomal|2,4-dienoyl-CoA reductase 2|short chain dehydrogenase/reductase family 17C member 1|short chain dehydrogenase/reductase family 17C, member 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=401858, end_position_on_the_genomic_accession=412487, orientation='plus', exon_count=9, OMIM=615839, no_hdr0=''),
  26064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26064, CurrentID=0, Status='live', Symbol='RAI14', Aliases='NORPEG, RAI13', description='retinoic acid induced 14', other_designations='ankyrin repeat and coiled-coil structure-containing protein|novel retinal pigment epithelial cell protein|retinoic acid-induced protein 14', map_location='5p13.3-p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34656328, end_position_on_the_genomic_accession=34832612, orientation='plus', exon_count=28, OMIM=606586, no_hdr0=''),
  26065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26065, CurrentID=0, Status='live', Symbol='LSM14A', Aliases='C19orf13, FAM61A, RAP55, RAP55A', description='LSM14A mRNA processing body assembly factor', other_designations='LSM14 homolog A|LSM14A, SCD6 homolog A|RNA-associated protein 55|RNA-associated protein 55A|alphaSNBP|family with sequence similarity 61, member A|hRAP55|hRAP55A|protein SCD6 homolog|putative alpha-synuclein-binding protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34172447, end_position_on_the_genomic_accession=34229515, orientation='plus', exon_count=11, OMIM=610677, no_hdr0=''),
  26069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26069, CurrentID=0, Status='live', Symbol='DKFZP586B0319', Aliases='', description='DKFZP586B0319 protein', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26070, CurrentID=0, Status='live', Symbol='DKFZP434K028', Aliases='', description='uncharacterized LOC26070', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61754029, end_position_on_the_genomic_accession=61757664, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26071, CurrentID=0, Status='live', Symbol='FAM127B', Aliases='CXX1b, MAR8A', description='family with sequence similarity 127 member B', other_designations='family with sequence similarity 127, member B|mammalian retrotransposon derived protein 8A', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135050932, end_position_on_the_genomic_accession=135052191, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26073, CurrentID=0, Status='live', Symbol='POLDIP2', Aliases='PDIP38, POLD4, p38', description='polymerase (DNA-directed), delta interacting protein 2', other_designations='38 kDa DNA polymerase delta interaction protein|polymerase delta interacting protein 38', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28346628, end_position_on_the_genomic_accession=28357589, orientation='minus', exon_count=11, OMIM=611519, no_hdr0=''),
  26074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26074, CurrentID=0, Status='live', Symbol='CFAP61', Aliases='C20orf26, CaM-IP3, dJ1002M8.3, dJ1178H5.4', description='cilia and flagella associated protein 61', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20052514, end_position_on_the_genomic_accession=20360714, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  101929875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929875, CurrentID=0, Status='live', Symbol='LOC101929875', Aliases='', description='uncharacterized LOC101929875', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26077, CurrentID=0, Status='live', Symbol='DKFZP434H168', Aliases='', description='uncharacterized LOC26077', other_designations='CTD-2050B12.2', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56192617, end_position_on_the_genomic_accession=56194525, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26080, CurrentID=0, Status='live', Symbol='FAM230C', Aliases='LINC00281, NCRNA00281', description='family with sequence similarity 230 member C', other_designations='family with sequence similarity 230, member C|long intergenic non-protein coding RNA 281', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18195297, end_position_on_the_genomic_accession=18232016, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  26082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26082, CurrentID=0, Status='live', Symbol='DKFZP434L187', Aliases='', description='uncharacterized LOC26082', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30196036, end_position_on_the_genomic_accession=30214540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26083, CurrentID=0, Status='live', Symbol='TBC1D29', Aliases='', description='TBC1 domain family member 29', other_designations='TBC1 domain family, member 29', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30559566, end_position_on_the_genomic_accession=30565667, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26084, CurrentID=0, Status='live', Symbol='ARHGEF26', Aliases='CSGEF, HMFN1864, SGEF', description='Rho guanine nucleotide exchange factor 26', other_designations='Rho guanine nucleotide exchange factor (GEF) 26|SH3 domain-containing guanine exchange factor|Src homology 3 domain-containing guanine nucleotide exchange factor', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154121003, end_position_on_the_genomic_accession=154257827, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26085, CurrentID=0, Status='live', Symbol='KLK13', Aliases='KLK-L4, KLKL4', description='kallikrein related peptidase 13', other_designations='kallikrein 13|kallikrein-like gene 4|kallikrein-like protein 4|kallikrein-related peptidase 13', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51056206, end_position_on_the_genomic_accession=51065110, orientation='minus', exon_count=5, OMIM=605505, no_hdr0=''),
  26086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26086, CurrentID=0, Status='live', Symbol='GPSM1', Aliases='AGS3', description='G-protein signaling modulator 1', other_designations='G-protein signalling modulator 1 (AGS3-like, C. elegans)|activator of G-protein signaling 3', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136327476, end_position_on_the_genomic_accession=136359605, orientation='plus', exon_count=17, OMIM=609491, no_hdr0=''),
  26088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26088, CurrentID=0, Status='live', Symbol='GGA1', Aliases='', description='golgi-associated, gamma adaptin ear containing, ARF binding protein 1', other_designations='ADP-ribosylation factor binding protein 1|gamma-adaptin related protein 1|gamma-adaptin-related protein 1|golgi associated, gamma adaptin ear containing, ARF binding protein 1|golgi-localized, gamma ear-containing, ARF-binding protein 1', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37607235, end_position_on_the_genomic_accession=37633564, orientation='plus', exon_count=21, OMIM=606004, no_hdr0=''),
  26090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26090, CurrentID=0, Status='live', Symbol='ABHD12', Aliases='ABHD12A, BEM46L2, C20orf22, PHARC, dJ965G21.2', description='abhydrolase domain containing 12', other_designations='2-arachidonoylglycerol hydrolase|abhydrolase domain-containing protein 12', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25294743, end_position_on_the_genomic_accession=25390982, orientation='minus', exon_count=17, OMIM=613599, no_hdr0=''),
  26091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26091, CurrentID=0, Status='live', Symbol='HERC4', Aliases='', description='HECT and RLD domain containing E3 ubiquitin protein ligase 4', other_designations='HECT domain and RCC1-like domain-containing protein 4|hect domain and RLD 4', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67921899, end_position_on_the_genomic_accession=68075348, orientation='minus', exon_count=31, OMIM=609248, no_hdr0=''),
  26092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26092, CurrentID=0, Status='live', Symbol='TOR1AIP1', Aliases='LAP1, LAP1B', description='torsin A interacting protein 1', other_designations='lamin-associated protein 1B|lamina-associated polypeptide 1B', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179882042, end_position_on_the_genomic_accession=179920077, orientation='plus', exon_count=11, OMIM=614512, no_hdr0=''),
  26093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26093, CurrentID=0, Status='live', Symbol='CCDC9', Aliases='', description='coiled-coil domain containing 9', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47256474, end_position_on_the_genomic_accession=47271953, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  26094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26094, CurrentID=0, Status='live', Symbol='DCAF4', Aliases='WDR21, WDR21A', description='DDB1 and CUL4 associated factor 4', other_designations='WD repeat domain 21A|WD repeat-containing protein 21A', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72926332, end_position_on_the_genomic_accession=72962098, orientation='plus', exon_count=19, OMIM=616372, no_hdr0=''),
  26095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26095, CurrentID=0, Status='live', Symbol='PTPN20', Aliases='CT126A, PTPN20B, bA142I17.1, bA42B19.1, hPTPN20, PTPN20', description='protein tyrosine phosphatase, non-receptor type 20', other_designations='cancer/testis antigen 126|protein tyrosine phosphatase, non-receptor type 20A|protein tyrosine phosphatase, non-receptor type 20B', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46911413, end_position_on_the_genomic_accession=47031830, orientation='plus', exon_count=21, OMIM='610630,610631', no_hdr0=''),
  26097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26097, CurrentID=0, Status='live', Symbol='CHTOP', Aliases='C1orf77, FL-SRAG, FOP, SRAG, SRAG-3, SRAG-5, pp7704', description='chromatin target of PRMT1', other_designations='friend of PRMT1 protein|small protein rich in arginine and glycine', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153633982, end_position_on_the_genomic_accession=153646306, orientation='plus', exon_count=7, OMIM=614206, no_hdr0=''),
  26098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26098, CurrentID=0, Status='live', Symbol='EDRF1', Aliases='C10orf137', description='erythroid differentiation regulatory factor 1', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125719515, end_position_on_the_genomic_accession=125764143, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  26099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26099, CurrentID=0, Status='live', Symbol='SZRD1', Aliases='C1orf144', description='SUZ RNA binding domain containing 1', other_designations='UPF0485 protein C1orf144|putative MAPK activating protein PM20,PM21|putative MAPK-activating protein PM18/PM20/PM22', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16367030, end_position_on_the_genomic_accession=16398148, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26100, CurrentID=0, Status='live', Symbol='WIPI2', Aliases='ATG18B, Atg21, CGI-50, WIPI-2', description='WD repeat domain, phosphoinositide interacting 2', other_designations='WD40 repeat protein interacting with phosphoinositides 2|WIPI49-like protein 2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5190204, end_position_on_the_genomic_accession=5233855, orientation='plus', exon_count=13, OMIM=609225, no_hdr0=''),
  26101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26101, CurrentID=0, Status='live', Symbol='DKFZP564M1462', Aliases='', description='DKFZP564M1462 protein', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26102, CurrentID=0, Status='live', Symbol='DKFZP434A062', Aliases='', description='uncharacterized LOC26102', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136322302, end_position_on_the_genomic_accession=136327323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26103, CurrentID=0, Status='live', Symbol='LRIT1', Aliases='FIGLER9, LRRC21, PAL', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 1', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 9|leucine rich repeat containing 21|leucine-rich repeat-containing protein 21|photoreceptor-associated LRR superfamily protein|retina-specific protein PAL', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84231520, end_position_on_the_genomic_accession=84246493, orientation='minus', exon_count=7, OMIM=616103, no_hdr0=''),
  26105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26105, CurrentID=0, Status='live', Symbol='DKFZP434C153', Aliases='', description='DKFZP434C153 protein', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26106, CurrentID=0, Status='live', Symbol='DKFZP586K1520', Aliases='', description='DKFZP586K1520 protein', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26107, CurrentID=0, Status='live', Symbol='DKFZP547J0410', Aliases='', description='DKFZP547J0410 protein', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26108, CurrentID=0, Status='live', Symbol='PYGO1', Aliases='', description='pygopus family PHD finger 1', other_designations='pygopus-like protein 1', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55538884, end_position_on_the_genomic_accession=55588852, orientation='minus', exon_count=4, OMIM=606902, no_hdr0=''),
  106456574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106456574, CurrentID=0, Status='live', Symbol='LINC01616', Aliases='n341006', description='long intergenic non-protein coding RNA 1616', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26112, CurrentID=0, Status='live', Symbol='CCDC69', Aliases='', description='coiled-coil domain containing 69', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151181052, end_position_on_the_genomic_accession=151224093, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26115, CurrentID=0, Status='live', Symbol='TANC2', Aliases='ROLSA, rols', description='tetratricopeptide repeat, ankyrin repeat and coiled-coil containing 2', other_designations='putative ankyrin-repeat containing protein|rolling pebbles homolog B|tetratricopeptide repeat, ankyrin repeat and coiled-coil domain-containing protein 2', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62966664, end_position_on_the_genomic_accession=63427706, orientation='plus', exon_count=31, OMIM=615047, no_hdr0=''),
  26118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26118, CurrentID=0, Status='live', Symbol='WSB1', Aliases='SWIP1, WSB-1', description='WD repeat and SOCS box containing 1', other_designations='SOCS box-containing WD protein SWiP-1|WD repeat and SOCS box-containing 1', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27294080, end_position_on_the_genomic_accession=27313844, orientation='plus', exon_count=10, OMIM=610091, no_hdr0=''),
  26119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26119, CurrentID=0, Status='live', Symbol='LDLRAP1', Aliases='ARH, ARH1, ARH2, FHCB1, FHCB2', description='low density lipoprotein receptor adaptor protein 1', other_designations='LDL receptor adaptor protein|autosomal recessive hypercholesterolemia protein', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25543580, end_position_on_the_genomic_accession=25591287, orientation='plus', exon_count=14, OMIM=605747, no_hdr0=''),
  26120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26120, CurrentID=0, Status='live', Symbol='DKFZP564C152', Aliases='', description='DKFZP564C152 protein', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26121, CurrentID=0, Status='live', Symbol='PRPF31', Aliases='NY-BR-99, PRP31, RP11, SNRNP61', description='pre-mRNA processing factor 31', other_designations='PRP31 pre-mRNA processing factor 31 homolog|U4/U6 snRNP 61 kDa protein|hPrp31|pre-mRNA-processing factor 31|protein 61K|serologically defined breast cancer antigen NY-BR-99', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54115376, end_position_on_the_genomic_accession=54131719, orientation='plus', exon_count=14, OMIM=606419, no_hdr0=''),
  26122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26122, CurrentID=0, Status='live', Symbol='EPC2', Aliases='EPC-LIKE', description='enhancer of polycomb homolog 2 (Drosophila)', other_designations='', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148644811, end_position_on_the_genomic_accession=148788491, orientation='plus', exon_count=16, OMIM=611000, no_hdr0=''),
  26123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26123, CurrentID=0, Status='live', Symbol='TCTN3', Aliases='C10orf61, JBTS18, OFD4, TECT3', description='tectonic family member 3', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95663396, end_position_on_the_genomic_accession=95694143, orientation='minus', exon_count=14, OMIM=613847, no_hdr0=''),
  26127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26127, CurrentID=0, Status='live', Symbol='FGFR1OP2', Aliases='HSPC123-like, WIT3.0', description='FGFR1 oncogene partner 2', other_designations='fibroblast growth factor receptor 1 oncogene partner 2|wound inducible transcript 3.0', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26938372, end_position_on_the_genomic_accession=26966648, orientation='plus', exon_count=7, OMIM=608858, no_hdr0=''),
  26128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26128, CurrentID=0, Status='live', Symbol='KIF1BP', Aliases='KBP, KIAA1279, TTC20', description='KIF1 binding protein', other_designations='kinesin binding protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68988721, end_position_on_the_genomic_accession=69016983, orientation='plus', exon_count=7, OMIM=609367, no_hdr0=''),
  26130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26130, CurrentID=0, Status='live', Symbol='GAPVD1', Aliases='GAPEX5, GAPex-5, RAP6', description='GTPase activating protein and VPS9 domains 1', other_designations='rab5-activating protein 6', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125241663, end_position_on_the_genomic_accession=125365011, orientation='plus', exon_count=35, OMIM=611714, no_hdr0=''),
  26133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26133, CurrentID=0, Status='live', Symbol='TRPC4AP', Aliases='C20orf188, PPP1R158, TRRP4AP, TRUSS', description='transient receptor potential cation channel, subfamily C, member 4 associated protein', other_designations='TNF-receptor ubiquitous scaffolding/signaling protein|TRP4-associated protein|protein phosphatase 1, regulatory subunit 158|short transient receptor potential channel 4 associated protein|trpc4-associated protein|tumor necrosis factor receptor-associated ubiquitous scaffolding and signaling protein', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35002404, end_position_on_the_genomic_accession=35092837, orientation='minus', exon_count=22, OMIM=608430, no_hdr0=''),
  157206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157206, CurrentID=0, Status='live', Symbol='RPL21P11', Aliases='RPL21_50_1399', description='ribosomal protein L21 pseudogene 11', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90366635, end_position_on_the_genomic_accession=90367222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26135, CurrentID=0, Status='live', Symbol='SERBP1', Aliases='CGI-55, CHD3IP, HABP4L, PAI-RBP1, PAIRBP1', description='SERPINE1 mRNA binding protein 1', other_designations='PAI-1 mRNA binding protein|PAI1 RNA-binding protein 1|SERPINE1 mRNA-binding protein 1|chromodomain helicase DNA binding protein 3 interacting protein|plasminogen activator inhibitor 1 RNA binding protein', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67407810, end_position_on_the_genomic_accession=67430440, orientation='minus', exon_count=8, OMIM=607378, no_hdr0=''),
  26136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26136, CurrentID=0, Status='live', Symbol='TES', Aliases='TESSS-2, TES', description='testin LIM domain protein', other_designations='testis derived transcript (3 LIM domains)', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116210493, end_position_on_the_genomic_accession=116258786, orientation='plus', exon_count=8, OMIM=606085, no_hdr0=''),
  26137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26137, CurrentID=0, Status='live', Symbol='ZBTB20', Aliases='DPZF, HOF, ODA-8S, PRIMS, ZNF288', description='zinc finger and BTB domain containing 20', other_designations='BTB/POZ zinc finger protein DPZF|dendritic cell-derived BTB/POZ zinc finger|zinc finger protein 288', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114314500, end_position_on_the_genomic_accession=115147280, orientation='minus', exon_count=21, OMIM=606025, no_hdr0=''),
  26138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26138, CurrentID=0, Status='live', Symbol='LINC00588', Aliases='C8orf71', description='long intergenic non-protein coding RNA 588', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57279543, end_position_on_the_genomic_accession=57284731, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26140, CurrentID=0, Status='live', Symbol='TTLL3', Aliases='HOTTL', description='tubulin tyrosine ligase like 3', other_designations='tubulin tyrosine ligase-like family member 3|tubulin tyrosine ligase-like family, member 3|tubulin--tyrosine ligase-like protein 3', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9809960, end_position_on_the_genomic_accession=9836356, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  101060869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060869, CurrentID=0, Status='live', Symbol='LVSKS', Aliases='', description='Levy-Shanske syndrome', other_designations='', map_location='15q26-qter', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614846, no_hdr0=''),
  26145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26145, CurrentID=0, Status='live', Symbol='IRF2BP1', Aliases='', description='interferon regulatory factor 2 binding protein 1', other_designations='IRF-2-binding protein 1|probable E3 ubiquitin-protein ligase IRF2BP1', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45883607, end_position_on_the_genomic_accession=45886170, orientation='minus', exon_count=1, OMIM=615331, no_hdr0=''),
  26146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26146, CurrentID=0, Status='live', Symbol='TRAF3IP1', Aliases='IFT54, MIP-T3, MIPT3', description='TRAF3 interacting protein 1', other_designations='TNF receptor-associated factor 3 interacting protein 1|interleukin 13 receptor alpha 1-binding protein-1|interleukin-13 receptor alpha 1-binding protein 1|intraflagellar transport protein 54 homolog|microtubule interacting protein that associates with TRAF3|microtubule-interacting protein associated with TRAF3|microtubule-interacting protein that associates with TRAF3', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238320544, end_position_on_the_genomic_accession=238400900, orientation='plus', exon_count=22, OMIM=607380, no_hdr0=''),
  26147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26147, CurrentID=0, Status='live', Symbol='PHF19', Aliases='MTF2L1, PCL3, TDRD19B', description='PHD finger protein 19', other_designations='polycomb like 3|polycomb-like protein 3|tudor domain containing 19B', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120855651, end_position_on_the_genomic_accession=120894896, orientation='minus', exon_count=22, OMIM=609740, no_hdr0=''),
  26148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26148, CurrentID=0, Status='live', Symbol='C10orf12', Aliases='', description='chromosome 10 open reading frame 12', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96981284, end_position_on_the_genomic_accession=96985828, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26149, CurrentID=0, Status='live', Symbol='ZNF658', Aliases='', description='zinc finger protein 658', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66900729, end_position_on_the_genomic_accession=66921653, orientation='plus', exon_count=7, OMIM=616290, no_hdr0=''),
  26150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26150, CurrentID=0, Status='live', Symbol='RIBC2', Aliases='C22orf11, TRIB', description='RIB43A domain with coiled-coils 2', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45413691, end_position_on_the_genomic_accession=45432422, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26151, CurrentID=0, Status='live', Symbol='NAT9', Aliases='EBSP, hNATL', description='N-acetyltransferase 9 (GCN5-related, putative)', other_designations='embryo brain specific protein', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74770547, end_position_on_the_genomic_accession=74776395, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26152, CurrentID=0, Status='live', Symbol='ZNF337', Aliases='', description='zinc finger protein 337', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25673195, end_position_on_the_genomic_accession=25696904, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26153, CurrentID=0, Status='live', Symbol='KIF26A', Aliases='', description='kinesin family member 26A', other_designations='KIF26A variant protein', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104138723, end_position_on_the_genomic_accession=104180898, orientation='plus', exon_count=15, OMIM=613231, no_hdr0=''),
  26154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26154, CurrentID=0, Status='live', Symbol='ABCA12', Aliases='ARCI4A, ARCI4B, ICR2B, LI2', description='ATP binding cassette subfamily A member 12', other_designations='ATP-binding cassette transporter 12|ATP-binding cassette, sub-family A (ABC1), member 12', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214931542, end_position_on_the_genomic_accession=215138580, orientation='minus', exon_count=55, OMIM=607800, no_hdr0=''),
  26155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26155, CurrentID=0, Status='live', Symbol='NOC2L', Aliases='NET15, NET7, NIR, PPP1R112', description='NOC2-like nucleolar associated transcriptional repressor', other_designations='NOC2-like protein|NOC2L nucleolar associated transcriptional repressor|novel INHAT (inhibitor of histone acetyltransferase) repressor|novel INHAT repressor|nucleolar complex associated 2 homolog|protein NOC2 homolog|protein phosphatase 1, regulatory subunit 12', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=944203, end_position_on_the_genomic_accession=959299, orientation='minus', exon_count=19, OMIM=610770, no_hdr0=''),
  26156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26156, CurrentID=0, Status='live', Symbol='RSL1D1', Aliases='CSIG, L12, PBK1, UTP30', description='ribosomal L1 domain containing 1', other_designations='CATX-11|cellular senescence inhibited gene protein|cellular senescence-inhibited gene protein', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11834196, end_position_on_the_genomic_accession=11851585, orientation='minus', exon_count=9, OMIM=615874, no_hdr0=''),
  26157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26157, CurrentID=0, Status='live', Symbol='GIMAP2', Aliases='HIMAP2, IAN12, IMAP2', description='GTPase, IMAP family member 2', other_designations='immune-associated nucleotide-binding protein 12|immunity associated protein 2|immunity-associated protein 2', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150685706, end_position_on_the_genomic_accession=150693641, orientation='plus', exon_count=3, OMIM=608085, no_hdr0=''),
  100271360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271360, CurrentID=0, Status='live', Symbol='RPS15AP18', Aliases='RPS15A_9_644', description='ribosomal protein S15a pseudogene 18', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179933775, end_position_on_the_genomic_accession=179934226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26160, CurrentID=0, Status='live', Symbol='IFT172', Aliases='NPHP17, RP71, SLB, SRTD10, osm-1, wim', description='intraflagellar transport 172', other_designations='intraflagellar transport 172 homolog|selective LIM binding factor homolog|wimple homolog', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27444373, end_position_on_the_genomic_accession=27489811, orientation='minus', exon_count=53, OMIM=607386, no_hdr0=''),
  26164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26164, CurrentID=0, Status='live', Symbol='MTG2', Aliases='GTPBP5, ObgH1, dJ1005F21.2', description='mitochondrial ribosome-associated GTPase 2', other_designations='GTP binding protein 5 (putative)|GTP-binding protein 5 (putative)|protein obg homolog 1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62183025, end_position_on_the_genomic_accession=62203568, orientation='plus', exon_count=11, OMIM=610919, no_hdr0=''),
  26165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26165, CurrentID=0, Status='live', Symbol='SPATA31A7', Aliases='C9orf36A, FAM75A4, FAM75A7, SPATA31A4', description='SPATA31 subfamily A, member 7', other_designations='SPATA31 subfamily A, member 4|family with sequence similarity 75, member A4|family with sequence similarity 75, member A7|protein FAM75A4|protein FAM75A7|spermatogenesis-associated protein 31A4', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61190036, end_position_on_the_genomic_accession=61196280, orientation='plus', exon_count=4, OMIM=616584, no_hdr0=''),
  26166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26166, CurrentID=0, Status='live', Symbol='RGS22', Aliases='CT145, PRTD-NY2', description='regulator of G-protein signaling 22', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99960936, end_position_on_the_genomic_accession=100106116, orientation='minus', exon_count=31, OMIM=615650, no_hdr0=''),
  26167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26167, CurrentID=0, Status='live', Symbol='PCDHB5', Aliases='PCDH-BETA5', description='protocadherin beta 5', other_designations='PCDH-beta-5', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141135206, end_position_on_the_genomic_accession=141138127, orientation='plus', exon_count=1, OMIM=606331, no_hdr0=''),
  26168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26168, CurrentID=0, Status='live', Symbol='SENP3', Aliases='SMT3IP1, SSP3, Ulp1', description='SUMO1/sentrin/SMT3 specific peptidase 3', other_designations='SUMO-1-specific protease 3|sentrin/SUMO-specific protease 3', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7561992, end_position_on_the_genomic_accession=7571969, orientation='plus', exon_count=11, OMIM=612844, no_hdr0=''),
  26173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26173, CurrentID=0, Status='live', Symbol='INTS1', Aliases='INT1, NET28', description='integrator complex subunit 1', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1470277, end_position_on_the_genomic_accession=1504382, orientation='minus', exon_count=48, OMIM=611345, no_hdr0=''),
  26175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26175, CurrentID=0, Status='live', Symbol='TMEM251', Aliases='C14orf109', description='transmembrane protein 251', other_designations='UPF0694 transmembrane protein C14orf109', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93184951, end_position_on_the_genomic_accession=93187086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26184, CurrentID=0, Status='live', Symbol='OR1F2P', Aliases='OLFMF2, OR16-3, OR1F11, OR1F2, OR1F3P, hg91', description='olfactory receptor family 1 subfamily F member 2 pseudogene', other_designations='olfactory receptor OR16-3 pseudogene|olfactory receptor, family 1, subfamily F, member 11|olfactory receptor, family 1, subfamily F, member 2 pseudogene|olfactory receptor, family 1, subfamily F, member 2, pseudogene|olfactory receptor, family 1, subfamily F, member 3 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3215562, end_position_on_the_genomic_accession=3216546, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26188, CurrentID=0, Status='live', Symbol='OR1C1', Aliases='HSTPCR27, OR1-42, OR1.5.10, ORL211, TPCR27', description='olfactory receptor family 1 subfamily C member 1', other_designations='olfactory receptor OR1-42|olfactory receptor TPCR27|olfactory receptor, family 1, subfamily C, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247757462, end_position_on_the_genomic_accession=247758406, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26189, CurrentID=0, Status='live', Symbol='OR1A2', Aliases='OR17-6', description='olfactory receptor family 1 subfamily A member 2', other_designations='olfactory receptor 17-6|olfactory receptor OR17-10|olfactory receptor, family 1, subfamily A, member 2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3197517, end_position_on_the_genomic_accession=3198448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26190, CurrentID=0, Status='live', Symbol='FBXW2', Aliases='FBW2, Fwd2, Md6', description='F-box and WD repeat domain containing 2', other_designations='F-box and WD-40 domain protein 2|F-box and WD-40 domain-containing protein 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120756976, end_position_on_the_genomic_accession=120793462, orientation='minus', exon_count=9, OMIM=609071, no_hdr0=''),
  26191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26191, CurrentID=0, Status='live', Symbol='PTPN22', Aliases='LYP, LYP1, LYP2, PEP, PTPN8', description='protein tyrosine phosphatase, non-receptor type 22 (lymphoid)', other_designations='PEST-domain phosphatase|hematopoietic cell protein-tyrosine phosphatase 70Z-PEP|lymphoid-specific protein tyrosine phosphatase|protein tyrosine phosphatase, non-receptor type 8', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113813811, end_position_on_the_genomic_accession=113871757, orientation='minus', exon_count=22, OMIM=600716, no_hdr0=''),
  100689490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689490, CurrentID=0, Status='live', Symbol='HNFJ3', Aliases='', description='Hyperuricemic nephropathy, familial juvenile, 3', other_designations='', map_location='2p22.1-p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614227, no_hdr0=''),
  100689491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689491, CurrentID=0, Status='live', Symbol='DEL8Q21.11', Aliases='C8DELq21.11', description='Chromosome 8q21.11 deletion syndrome', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614230, no_hdr0=''),
  157273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157273, CurrentID=0, Status='live', Symbol='LOC157273', Aliases='', description='uncharacterized LOC157273', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9325051, end_position_on_the_genomic_accession=9335080, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26205, CurrentID=0, Status='live', Symbol='GMEB2', Aliases='GMEB-2, P79PIF, PIF79', description='glucocorticoid modulatory element binding protein 2', other_designations='DNA-binding protein p79PIF|PIF p79|parvovirus initiation factor p79|parvovirus initiation factor, p79', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63587602, end_position_on_the_genomic_accession=63627028, orientation='minus', exon_count=11, OMIM=607451, no_hdr0=''),
  26206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26206, CurrentID=0, Status='live', Symbol='SPAG8', Aliases='BS-84, CILD28, CT142, HSD-1, SMP1, SPAG3, hSMP-1', description='sperm associated antigen 8', other_designations='sperm membrane protein 1|sperm membrane protein BS-84', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35807785, end_position_on_the_genomic_accession=35812272, orientation='minus', exon_count=9, OMIM=605731, no_hdr0=''),
  26207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26207, CurrentID=0, Status='live', Symbol='PITPNC1', Aliases='M-RDGB-beta, MRDGBbeta, RDGB-BETA, RDGBB, RDGBB1', description='phosphatidylinositol transfer protein, cytoplasmic 1', other_designations='M-rdgB beta|mammalian rdgB homolog beta|retinal degeneration B beta 1|retinal degeneration B homolog beta', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67377281, end_position_on_the_genomic_accession=67697263, orientation='plus', exon_count=11, OMIM=605134, no_hdr0=''),
  26211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26211, CurrentID=0, Status='live', Symbol='OR2F1', Aliases='7M1-2, OLF3, OR14-60, OR2F3, OR2F3P, OR2F4, OR2F5, OR7-139, OR7-140', description='olfactory receptor family 2 subfamily F member 1 (gene/pseudogene)', other_designations='olfactory receptor 2F3|olfactory receptor 2F4|olfactory receptor 2F5|olfactory receptor OR7-7|olfactory receptor, family 2, subfamily F, member 1|olfactory receptor, family 2, subfamily F, member 3|olfactory receptor, family 2, subfamily F, member 4|olfactory receptor, family 2, subfamily F, member 5|olfactory receptor-like protein OLF3', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143907126, end_position_on_the_genomic_accession=143961015, orientation='plus', exon_count=5, OMIM=608497, no_hdr0=''),
  26212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26212, CurrentID=0, Status='live', Symbol='OR2B6', Aliases='OR2B1, OR2B1P, OR2B5P, OR5-40, OR5-41, OR6-31, dJ408B20.2, OR2B6', description='olfactory receptor family 2 subfamily B member 6', other_designations='hs6M1-32|olfactory receptor 2B1|olfactory receptor 2B5|olfactory receptor 5-40|olfactory receptor 6-31|olfactory receptor OR6-4|olfactory receptor, family 2, subfamily B, member 1 pseudogene|olfactory receptor, family 2, subfamily B, member 5|olfactory receptor, family 2, subfamily B, member 6 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27957241, end_position_on_the_genomic_accession=27958182, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157285, CurrentID=0, Status='live', Symbol='SGK223', Aliases='PRAGMIN', description='homolog of rat pragma of Rnd2', other_designations='sugen kinase 223', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8317736, end_position_on_the_genomic_accession=8386443, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100616159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616159, CurrentID=0, Status='live', Symbol='MIR4779', Aliases='', description='microRNA 4779', other_designations='hsa-mir-4779', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86193026, end_position_on_the_genomic_accession=86193108, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271362, CurrentID=0, Status='live', Symbol='RPS16P6', Aliases='RPS16_3_1149', description='ribosomal protein S16 pseudogene 6', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64881535, end_position_on_the_genomic_accession=64881945, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26219, CurrentID=0, Status='live', Symbol='OR1J4', Aliases='HSHTPCRX01, HTPCRX01, OR9-21', description='olfactory receptor family 1 subfamily J member 4', other_designations='olfactory receptor OR9-21|olfactory receptor, family 1, subfamily J, member 4', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122519141, end_position_on_the_genomic_accession=122520082, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26220, CurrentID=0, Status='live', Symbol='DGCR5', Aliases='LINC00037, NCRNA00037', description='DiGeorge syndrome critical region gene 5 (non-protein coding)', other_designations='cadherin EGF LAG seven-pass G-type receptor 1 pseudogene|long intergenic non-protein coding RNA 37', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18970498, end_position_on_the_genomic_accession=19031242, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26222, CurrentID=0, Status='live', Symbol='DGCR10', Aliases='DGS-B', description='DiGeorge syndrome critical region gene 10 (non-protein coding)', other_designations='DiGeorge syndrome gene B', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19022624, end_position_on_the_genomic_accession=19023550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26223, CurrentID=0, Status='live', Symbol='FBXL21', Aliases='FBL3B, FBXL3B, FBXL3P, Fbl21', description='F-box and leucine-rich repeat protein 21 (gene/pseudogene)', other_designations='F-box and leucine-rich repeat protein 3 pseudogene|F-box and leucine-rich repeat protein 3B|F-box protein Fbl3b|F-box/LRR-repeat protein 3B', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135930317, end_position_on_the_genomic_accession=135941678, orientation='plus', exon_count=8, OMIM=609087, no_hdr0=''),
  26224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26224, CurrentID=0, Status='live', Symbol='FBXL3', Aliases='FBL3, FBL3AA, FBXL3', description='F-box and leucine-rich repeat protein 3', other_designations='F-box and leucine-rich repeat protein 3A|F-box protein Fbl3a|F-box/LRR-repeat protein 3A', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77001339, end_position_on_the_genomic_accession=77027196, orientation='minus', exon_count=7, OMIM=605653, no_hdr0=''),
  26225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26225, CurrentID=0, Status='live', Symbol='ARL5A', Aliases='ARFLP5, ARL5', description='ADP ribosylation factor like GTPase 5A', other_designations='ADP-ribosylation factor like GTPase 5A|ADP-ribosylation factor-like 5|ADP-ribosylation factor-like 5A|ADP-ribosylation factor-like protein 5', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151800966, end_position_on_the_genomic_accession=151828495, orientation='minus', exon_count=6, OMIM=608960, no_hdr0=''),
  26226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26226, CurrentID=0, Status='live', Symbol='FBXW4P1', Aliases='FBW3, FBXW3, SHFM3P1', description='F-box and WD repeat domain containing 4 pseudogene 1', other_designations='F-box and WD-40 domain protein 3|split hand/foot malformation (ectrodactyly) type 3 pseudogene 1', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23262767, end_position_on_the_genomic_accession=23265005, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26227, CurrentID=0, Status='live', Symbol='PHGDH', Aliases='3-PGDH, 3PGDH, HEL-S-113, NLS, NLS1, PDG, PGAD, PGD, PGDHD, SERA, PHGDH', description='phosphoglycerate dehydrogenase', other_designations='3-phosphoglycerate dehydrogenase|epididymis secretory protein Li 113', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119711737, end_position_on_the_genomic_accession=119744226, orientation='plus', exon_count=17, OMIM=606879, no_hdr0=''),
  26228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26228, CurrentID=0, Status='live', Symbol='STAP1', Aliases='BRDG1, STAP-1', description='signal transducing adaptor family member 1', other_designations='BCR downstream-signaling protein 1|docking protein BRDG1|stem cell adaptor protein 1', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67558728, end_position_on_the_genomic_accession=67607337, orientation='plus', exon_count=10, OMIM=604298, no_hdr0=''),
  26229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26229, CurrentID=0, Status='live', Symbol='B3GAT3', Aliases='GLCATI, glcUAT-I', description='beta-1,3-glucuronyltransferase 3', other_designations='Sqv-8-like protein|UDP-GlcUA:Gal beta-1,3-Gal-R glucuronyltransferase|beta-1,3-glucuronyltransferase 3 (glucuronosyltransferase I)', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62615296, end_position_on_the_genomic_accession=62622175, orientation='minus', exon_count=6, OMIM=606374, no_hdr0=''),
  26230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26230, CurrentID=0, Status='live', Symbol='TIAM2', Aliases='STEF, TIAM-2', description='T-cell lymphoma invasion and metastasis 2', other_designations='SIF and TIAM1-like exchange factor', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155090289, end_position_on_the_genomic_accession=155257723, orientation='plus', exon_count=27, OMIM=604709, no_hdr0=''),
  26231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26231, CurrentID=0, Status='live', Symbol='LRRC29', Aliases='FBL9, FBXL9', description='leucine rich repeat containing 29', other_designations='F-box and leucine-rich repeat protein 9|F-box protein FBL9|F-box/LRR-repeat protein 9', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67207139, end_position_on_the_genomic_accession=67227059, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26232, CurrentID=0, Status='live', Symbol='FBXO2', Aliases='FBG1, FBX2, Fbs1, NFB42, OCP1', description='F-box protein 2', other_designations='F-box gene 1|organ of Corti protein 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11648361, end_position_on_the_genomic_accession=11654831, orientation='minus', exon_count=6, OMIM=607112, no_hdr0=''),
  26233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26233, CurrentID=0, Status='live', Symbol='FBXL6', Aliases='FBL6, FBL6A, PP14630', description='F-box and leucine-rich repeat protein 6', other_designations='F-box protein Fbl6', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144355428, end_position_on_the_genomic_accession=144358523, orientation='minus', exon_count=9, OMIM=609076, no_hdr0=''),
  26234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26234, CurrentID=0, Status='live', Symbol='FBXL5', Aliases='FBL4, FBL5, FLR1', description='F-box and leucine-rich repeat protein 5', other_designations='F-box protein FBL4/FBL5|F-box protein FBL5|p45SKP2-like protein', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15604384, end_position_on_the_genomic_accession=15659917, orientation='minus', exon_count=14, OMIM=605655, no_hdr0=''),
  26235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26235, CurrentID=0, Status='live', Symbol='FBXL4', Aliases='FBL4, FBL5, MTDPS13', description='F-box and leucine-rich repeat protein 4', other_designations='', map_location='6q16.1-q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=98873725, end_position_on_the_genomic_accession=98948006, orientation='minus', exon_count=11, OMIM=605654, no_hdr0=''),
  222826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222826, CurrentID=0, Status='live', Symbol='FAM217A', Aliases='C6orf146', description='family with sequence similarity 217 member A', other_designations='family with sequence similarity 217, member A|uncharacterized protein C6orf146', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4068352, end_position_on_the_genomic_accession=4087344, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  157310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157310, CurrentID=0, Status='live', Symbol='PEBP4', Aliases='CORK-1, CORK1, GWTM1933, HEL-S-300, PEBP-4, PRO4408, hPEBP4', description='phosphatidylethanolamine binding protein 4', other_designations='cousin-of-RKIP 1 protein|epididymis secretory protein Li 300|protein cousin-of-RKIP 1', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22713251, end_position_on_the_genomic_accession=22999754, orientation='minus', exon_count=9, OMIM=612473, no_hdr0=''),
  26239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26239, CurrentID=0, Status='live', Symbol='LCE2B', Aliases='LEP10, SPRL1B, XP5', description='late cornified envelope 2B', other_designations='late envelope protein 10|skin-specific protein Xp5|small proline rich-like (epidermal differentiation complex) 1B|small proline-rich-like epidermal differentiation complex protein 1B', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152686123, end_position_on_the_genomic_accession=152687401, orientation='plus', exon_count=2, OMIM=612610, no_hdr0=''),
  26240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26240, CurrentID=0, Status='live', Symbol='FAM50B', Aliases='D6S2654E, X5L', description='family with sequence similarity 50 member B', other_designations='XAP5-like protein|family with sequence similarity 50, member B', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3849398, end_position_on_the_genomic_accession=3851317, orientation='plus', exon_count=2, OMIM=614686, no_hdr0=''),
  157313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157313, CurrentID=0, Status='live', Symbol='CDCA2', Aliases='PPP1R81, Repo-Man', description='cell division cycle associated 2', other_designations='protein phosphatase 1, regulatory subunit 81|recruits PP1 onto mitotic chromatin at anaphase protein', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25458574, end_position_on_the_genomic_accession=25511308, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  26242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26242, CurrentID=0, Status='live', Symbol='OR4C1P', Aliases='HSHTPCRX11, HTPCRX11, OR4C1', description='olfactory receptor family 4 subfamily C member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 1 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55509638, end_position_on_the_genomic_accession=55510762, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26245, CurrentID=0, Status='live', Symbol='OR2M4', Aliases='HSHTPCRX18, HTPCRX18, OR1-55, OST710, TPCR100', description='olfactory receptor family 2 subfamily M member 4', other_designations='novel 7 transmembrane receptor (rhodopsin family) protein|olfactory receptor OR1-55|olfactory receptor TPCR100|olfactory receptor, family 2, subfamily M, member 4', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248238929, end_position_on_the_genomic_accession=248239864, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26246, CurrentID=0, Status='live', Symbol='OR2L2', Aliases='HSHTPCRH07, HTPCRH07, OR1-48, OR2L12, OR2L4P', description='olfactory receptor family 2 subfamily L member 2', other_designations='olfactory receptor 2L12|olfactory receptor 2L4|olfactory receptor OR1-48 pseudogene|olfactory receptor, family 2, subfamily L, member 12|olfactory receptor, family 2, subfamily L, member 2|olfactory receptor, family 2, subfamily L, member 4 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248035510, end_position_on_the_genomic_accession=248039332, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26247, CurrentID=0, Status='live', Symbol='OR2L1P', Aliases='HSHTPCRX02, HTPCRX02, OR2L1, OR2L7P', description='olfactory receptor family 2 subfamily L member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily L, member 1 pseudogene|olfactory receptor, family 2, subfamily L, member 7 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247990267, end_position_on_the_genomic_accession=247991191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26248, CurrentID=0, Status='live', Symbol='OR2K2', Aliases='HSHTPCRH06, HTPCRH06, OR2AN1P, OR2AR1P', description='olfactory receptor family 2 subfamily K member 2', other_designations='olfactory receptor OR9-17|olfactory receptor, family 2, subfamily AN, member 1 pseudogene|olfactory receptor, family 2, subfamily AR, member 1 pseudogene|olfactory receptor, family 2, subfamily K, member 2', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111327370, end_position_on_the_genomic_accession=111330173, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26249, CurrentID=0, Status='live', Symbol='KLHL3', Aliases='PHA2D', description='kelch like family member 3', other_designations='kelch-like family member 3', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137617500, end_position_on_the_genomic_accession=137736090, orientation='minus', exon_count=17, OMIM=605775, no_hdr0=''),
  26251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26251, CurrentID=0, Status='live', Symbol='KCNG2', Aliases='KCNF2, KV6.2', description='potassium channel, voltage gated modifier subfamily G, member 2', other_designations='cardiac potassium channel subunit|potassium voltage-gated channel, subfamily G, member 2|voltage-gated potassium channel subunit Kv6.2', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79797963, end_position_on_the_genomic_accession=79903195, orientation='plus', exon_count=5, OMIM=605696, no_hdr0=''),
  26253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26253, CurrentID=0, Status='live', Symbol='CLEC4E', Aliases='CLECSF9, MINCLE', description='C-type lectin domain family 4 member E', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 9|C-type lectin domain family 4, member E|C-type lectin superfamily member 9|macrophage-inducible C-type lectin', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8533305, end_position_on_the_genomic_accession=8540962, orientation='minus', exon_count=6, OMIM=609962, no_hdr0=''),
  26254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26254, CurrentID=0, Status='live', Symbol='OPTC', Aliases='OPT', description='opticin', other_designations='oculoglycan', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203494143, end_position_on_the_genomic_accession=203508949, orientation='plus', exon_count=8, OMIM=605127, no_hdr0=''),
  26255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26255, CurrentID=0, Status='live', Symbol='PTTG3P', Aliases='PTTG3, rcPTTG1', description='pituitary tumor-transforming 3, pseudogene', other_designations='pituitary tumor-transforming 3 (pseudogene)', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66767397, end_position_on_the_genomic_accession=66768005, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26256, CurrentID=0, Status='live', Symbol='CABYR', Aliases='CABYRac, CABYRc/d, CABYRe, CBP86, CT88, FSP-2, FSP2, CABYR', description='calcium binding tyrosine-(Y)-phosphorylation regulated', other_designations='calcium binding tyrosine-(Y)-phosphorylation regulated (fibrousheathin 2)|calcium-binding protein 86|calcium-binding tyrosine-(Y)-phosphorylation regulated (fibrousheathin 2)|cancer/testis antigen 88|fibrousheathin 2|fibrousheathin II|fibrousheathin-2|testis-specific calcium-binding protein CBP86', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24138956, end_position_on_the_genomic_accession=24161600, orientation='plus', exon_count=10, OMIM=612135, no_hdr0=''),
  26257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26257, CurrentID=0, Status='live', Symbol='NKX2-8', Aliases='NKX2.8, NKX2H, Nkx2-9', description='NK2 homeobox 8', other_designations='NK-2 homolog 8|NK-2 homolog H|NK2 transcription factor related, locus 8|homeobox protein NK-2 homolog H', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36580011, end_position_on_the_genomic_accession=36582581, orientation='minus', exon_count=2, OMIM=603245, no_hdr0=''),
  26258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26258, CurrentID=0, Status='live', Symbol='BLOC1S6', Aliases='BLOS6, HPS9, PA, PALLID, PLDN', description='biogenesis of lysosomal organelles complex-1, subunit 6, pallidin', other_designations='BLOC-1 subunit 6|biogenesis of lysosomal organelles complex-1, subunit 5, pallidin|pallid protein homolog|syntaxin 13 binding protein 1|syntaxin 13-interacting protein pallid', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45587219, end_position_on_the_genomic_accession=45609711, orientation='plus', exon_count=5, OMIM=604310, no_hdr0=''),
  26259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26259, CurrentID=0, Status='live', Symbol='FBXW8', Aliases='FBW6, FBW8, FBX29, FBXO29, FBXW6', description='F-box and WD repeat domain containing 8', other_designations='F-box and WD-40 domain protein 8|F-box and WD-40 domain-containing protein 8|F-box only protein 29', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116910951, end_position_on_the_genomic_accession=117032498, orientation='plus', exon_count=14, OMIM=609073, no_hdr0=''),
  26260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26260, CurrentID=0, Status='live', Symbol='FBXO25', Aliases='FBX25', description='F-box protein 25', other_designations='F-box protein Fbx25', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=406808, end_position_on_the_genomic_accession=469876, orientation='plus', exon_count=14, OMIM=609098, no_hdr0=''),
  26261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26261, CurrentID=0, Status='live', Symbol='FBXO24', Aliases='FBX24', description='F-box protein 24', other_designations='F-box protein Fbx24', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100586333, end_position_on_the_genomic_accession=100601117, orientation='plus', exon_count=12, OMIM=609097, no_hdr0=''),
  26262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26262, CurrentID=0, Status='live', Symbol='TSPAN17', Aliases='FBX23, FBXO23, TM4SF17', description='tetraspanin 17', other_designations='F-box only protein 23, transmembrane 4 superfamily member 17|tetraspan protein SB134', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176647387, end_position_on_the_genomic_accession=176659058, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26263, CurrentID=0, Status='live', Symbol='FBXO22', Aliases='FBX22, FISTC1', description='F-box protein 22', other_designations='F-box protein FBX22p44|FIST domain containing 1', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75903859, end_position_on_the_genomic_accession=75935268, orientation='plus', exon_count=8, OMIM=609096, no_hdr0=''),
  100820632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820632, CurrentID=0, Status='live', Symbol='DEL15Q25', Aliases='C15DELq25', description='Chromosome 15q25 deletion syndrome', other_designations='', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614294, no_hdr0=''),
  100820633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820633, CurrentID=0, Status='live', Symbol='DEL2Q23.1', Aliases='C2DELq23.1', description='Chromosome 2q23.1 deletion syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=156200, no_hdr0=''),
  26266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26266, CurrentID=0, Status='live', Symbol='SLC13A4', Aliases='NAS2, SUT-1, SUT1', description='solute carrier family 13 (sodium/sulfate symporter), member 4', other_designations='Na(+)/sulfate cotransporter SUT-1|solute carrier family 13 (sodium/sulfate symporters), member 4|solute carrier family 13 (sodium/sulphate symporters), member 4|sulphate transporter 1', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135681231, end_position_on_the_genomic_accession=135728185, orientation='minus', exon_count=16, OMIM=604309, no_hdr0=''),
  26267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26267, CurrentID=0, Status='live', Symbol='FBXO10', Aliases='FBX10, PRMT11', description='F-box protein 10', other_designations='F-box protein Fbx10', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37510892, end_position_on_the_genomic_accession=37576311, orientation='minus', exon_count=14, OMIM=609092, no_hdr0=''),
  26268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26268, CurrentID=0, Status='live', Symbol='FBXO9', Aliases='FBX9, NY-REN-57, VCIA1, dJ341E18.2', description='F-box protein 9', other_designations='F-box protein Fbx9|cross-immune reaction antigen 1|renal carcinoma antigen NY-REN-57', map_location='6p12.3-p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53057783, end_position_on_the_genomic_accession=53100873, orientation='plus', exon_count=17, OMIM=609091, no_hdr0=''),
  26269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26269, CurrentID=0, Status='live', Symbol='FBXO8', Aliases='DC10, FBS, FBX8', description='F-box protein 8', other_designations='F-box protein Fbx8|F-box/SEC7 protein FBS', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174236658, end_position_on_the_genomic_accession=174284263, orientation='minus', exon_count=7, OMIM=605649, no_hdr0=''),
  26270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26270, CurrentID=0, Status='live', Symbol='FBXO6', Aliases='FBG2, FBS2, FBX6, Fbx6b', description='F-box protein 6', other_designations='F-box protein FBG2|F-box protein Fbx6|F-box protein that recognizes sugar chains 2|F-box/G-domain protein 2', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11664093, end_position_on_the_genomic_accession=11674354, orientation='plus', exon_count=11, OMIM=605647, no_hdr0=''),
  26271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26271, CurrentID=0, Status='live', Symbol='FBXO5', Aliases='EMI1, FBX5, Fbxo31', description='F-box protein 5', other_designations='F-box protein Fbx5|early mitotic inhibitor 1', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152970519, end_position_on_the_genomic_accession=152983605, orientation='minus', exon_count=7, OMIM=606013, no_hdr0=''),
  26272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26272, CurrentID=0, Status='live', Symbol='FBXO4', Aliases='FBX4', description='F-box protein 4', other_designations='F-box protein Fbx4', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41925252, end_position_on_the_genomic_accession=42040228, orientation='plus', exon_count=9, OMIM=609090, no_hdr0=''),
  26273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26273, CurrentID=0, Status='live', Symbol='FBXO3', Aliases='FBA, FBX3', description='F-box protein 3', other_designations='F-box protein FBX3', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33740944, end_position_on_the_genomic_accession=33774531, orientation='minus', exon_count=14, OMIM=609089, no_hdr0=''),
  26275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26275, CurrentID=0, Status='live', Symbol='HIBCH', Aliases='HIBYLCOAH', description='3-hydroxyisobutyryl-CoA hydrolase', other_designations='3-hydroxyisobutyryl-Coenzyme A hydrolase|HIB-CoA hydrolase|HIBYL-CoA-H', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190204634, end_position_on_the_genomic_accession=190320045, orientation='minus', exon_count=15, OMIM=610690, no_hdr0=''),
  26276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26276, CurrentID=0, Status='live', Symbol='VPS33B', Aliases='', description='vacuolar protein sorting 33 homolog B (yeast)', other_designations='vacuolar protein sorting 33-like protein B', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90998416, end_position_on_the_genomic_accession=91022621, orientation='minus', exon_count=25, OMIM=608552, no_hdr0=''),
  26277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26277, CurrentID=0, Status='live', Symbol='TINF2', Aliases='DKCA3, TIN2', description='TERF1 (TRF1)-interacting nuclear factor 2', other_designations='TRF1-interacting nuclear protein 2', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24234403, end_position_on_the_genomic_accession=24243027, orientation='minus', exon_count=10, OMIM=604319, no_hdr0=''),
  26278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26278, CurrentID=0, Status='live', Symbol='SACS', Aliases='ARSACS, DNAJC29, PPP1R138, SPAX6', description='sacsin molecular chaperone', other_designations='dnaJ homolog subfamily C member 29|protein phosphatase 1, regulatory subunit 138|spastic ataxia of Charlevoix-Saguenay (sacsin)', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23328823, end_position_on_the_genomic_accession=23433728, orientation='minus', exon_count=12, OMIM=604490, no_hdr0=''),
  26279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26279, CurrentID=0, Status='live', Symbol='PLA2G2D', Aliases='PLA2IID, SPLASH, sPLA2-IID, sPLA2S', description='phospholipase A2 group IID', other_designations='GIID sPLA2|phosphatidylcholine 2-acylhydrolase 2D|phospholipase A2, group IID|secretory phospholipase A2s|secretory-type PLA, stroma-associated homolog', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20112650, end_position_on_the_genomic_accession=20119566, orientation='minus', exon_count=4, OMIM=605630, no_hdr0=''),
  26280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26280, CurrentID=0, Status='live', Symbol='IL1RAPL2', Aliases='IL-1R9, IL1R9, IL1RAPL-2, TIGIRR-1', description='interleukin 1 receptor accessory protein-like 2', other_designations='IL-1 receptor accessory protein-like 2|IL-1R-9|IL1RAPL-2-related protein|interleukin 1 receptor 9|three immunoglobulin domain-containing IL-1 receptor-related 1', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104566199, end_position_on_the_genomic_accession=105769729, orientation='plus', exon_count=13, OMIM=300277, no_hdr0=''),
  26281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26281, CurrentID=0, Status='live', Symbol='FGF20', Aliases='FGF-20, RHDA2', description='fibroblast growth factor 20', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16992825, end_position_on_the_genomic_accession=17002165, orientation='minus', exon_count=3, OMIM=605558, no_hdr0=''),
  26284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26284, CurrentID=0, Status='live', Symbol='ERAL1', Aliases='CEGA, ERA, ERA-WA, H-ERA, HERA-A, HERA-B, ERAL1', description='Era-like 12S mitochondrial rRNA chaperone 1', other_designations='ERA-like protein 1|Era G-protein-like 1|GTP-binding protein era homolog|GTPase, human homolog of E. coli essential cell cycle protein Era|conserved ERA-like GTPase|era (E. coli G-protein homolog)-like 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28855025, end_position_on_the_genomic_accession=28861054, orientation='plus', exon_count=10, OMIM=607435, no_hdr0=''),
  26285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26285, CurrentID=0, Status='live', Symbol='CLDN17', Aliases='', description='claudin 17', other_designations='human CLDN17 gene for claudin-17', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30165923, end_position_on_the_genomic_accession=30166653, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26286, CurrentID=0, Status='live', Symbol='ARFGAP3', Aliases='ARFGAP1', description='ADP ribosylation factor GTPase activating protein 3', other_designations='ADP-ribosylation factor GTPase activating protein 1|ADP-ribosylation factor GTPase activating protein 3|ARF GAP 3', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42796524, end_position_on_the_genomic_accession=42857402, orientation='minus', exon_count=16, OMIM=612439, no_hdr0=''),
  26287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26287, CurrentID=0, Status='live', Symbol='ANKRD2', Aliases='ARPP', description='ankyrin repeat domain 2 (stretch responsive muscle)', other_designations='skeletal muscle ankyrin repeat protein', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97572441, end_position_on_the_genomic_accession=97583884, orientation='plus', exon_count=9, OMIM=610734, no_hdr0=''),
  26289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26289, CurrentID=0, Status='live', Symbol='AK5', Aliases='AK6', description='adenylate kinase 5', other_designations='AK 5|ATP-AMP transphosphorylase 5|adenylate kinase 6', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77281977, end_position_on_the_genomic_accession=77559969, orientation='plus', exon_count=15, OMIM=608009, no_hdr0=''),
  26290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26290, CurrentID=0, Status='live', Symbol='GALNT8', Aliases='GALNAC-T8', description='polypeptide N-acetylgalactosaminyltransferase 8', other_designations='GalNAc transferase 8|UDP-GalNAc: polypeptide N-acetylgalactosaminyltransferase 8|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 8|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 8 (GalNAc-T8)|polypeptide GalNAc transferase 8|pp-GaNTase 8|protein-UDP acetylgalactosaminyltransferase 8', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4720586, end_position_on_the_genomic_accession=4772726, orientation='plus', exon_count=11, OMIM=606250, no_hdr0=''),
  26291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26291, CurrentID=0, Status='live', Symbol='FGF21', Aliases='', description='fibroblast growth factor 21', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48755559, end_position_on_the_genomic_accession=48758330, orientation='plus', exon_count=4, OMIM=609436, no_hdr0=''),
  26292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26292, CurrentID=0, Status='live', Symbol='MYCBP', Aliases='AMY-1', description='MYC binding protein', other_designations='associate of myc-1|c-myc binding protein', map_location='1p33-p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38862490, end_position_on_the_genomic_accession=38873378, orientation='minus', exon_count=5, OMIM=606535, no_hdr0=''),
  26297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26297, CurrentID=0, Status='live', Symbol='SERGEF', Aliases='DELGEF, Gnefr', description='secretion regulating guanine nucleotide exchange factor', other_designations='deafness locus associated putative guanine nucleotide exchange factor|deafness locus-associated putative guanine nucleotide exchange factor|guanine nucleotide exchange factor-related protein', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17788048, end_position_on_the_genomic_accession=18013162, orientation='minus', exon_count=12, OMIM=606051, no_hdr0=''),
  26298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26298, CurrentID=0, Status='live', Symbol='EHF', Aliases='ESE3, ESE3B, ESEJ', description='ets homologous factor', other_designations='ESE3 transcription factor|ETS domain-containing transcription factor|epithelium-specific Ets transcription factor 3|hEHF', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34621041, end_position_on_the_genomic_accession=34663288, orientation='plus', exon_count=14, OMIM=605439, no_hdr0=''),
  26300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26300, CurrentID=0, Status='live', Symbol='DYX4', Aliases='', description='dyslexia susceptibility 4', other_designations='', map_location='6q13-q16.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26301, CurrentID=0, Status='live', Symbol='GBGT1', Aliases='A3GALNT, FS, UNQ2513', description='globoside alpha-1,3-N-acetylgalactosaminyltransferase 1', other_designations='Forssman glycolipid synthetase (FS)|forssman glycolipid synthase-like protein', map_location='9q34.13-q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133152948, end_position_on_the_genomic_accession=133163945, orientation='minus', exon_count=7, OMIM=606074, no_hdr0=''),
  157376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157376, CurrentID=0, Status='live', Symbol='FER1L6-AS2', Aliases='C8orf78', description='FER1L6 antisense RNA 2', other_designations='FER1L6 antisense RNA 2 (non-protein coding)', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124046071, end_position_on_the_genomic_accession=124171522, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  157378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157378, CurrentID=0, Status='live', Symbol='TMEM65', Aliases='', description='transmembrane protein 65', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124306189, end_position_on_the_genomic_accession=124372699, orientation='minus', exon_count=8, OMIM=616609, no_hdr0=''),
  26307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26307, CurrentID=0, Status='live', Symbol='PCDHA14', Aliases='CNR3, CNRN3, CNRS3, CRNR3, PCDH-PSI4', description='protocadherin alpha 14 pseudogene', other_designations='cadherin-related neuronal receptor 3|protocadherin gamma subfamily A 14 pseudogene', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140861188, end_position_on_the_genomic_accession=140863550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271365, CurrentID=0, Status='live', Symbol='RPS21P6', Aliases='RPS21_3_1599', description='ribosomal protein S21 pseudogene 6', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46722975, end_position_on_the_genomic_accession=46723203, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157381, CurrentID=0, Status='live', Symbol='LINC00964', Aliases='', description='long intergenic non-protein coding RNA 964', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124942008, end_position_on_the_genomic_accession=124951095, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9605, CurrentID=0, Status='live', Symbol='VPS9D1', Aliases='ATP-BL, C16orf7', description='VPS9 domain containing 1', other_designations='', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89707133, end_position_on_the_genomic_accession=89721272, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  26322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26322, CurrentID=0, Status='live', Symbol='OR4H6P', Aliases='OR14-58, OR15-69, OR15-71, OR15-80, OR15-81, OR15-82, OR4-114, OR4-115, OR4-119, OR4H10, OR4H10P, OR4H11, OR4H11P, OR4H1P, OR4H2, OR4H2P, OR4H3, OR4H3P, OR4H4, OR4H4P, OR4H5, OR4H5P, OR4H6, OR4H7, OR4H7P, OR4H8, OR4H8P, OR4H9, OR4H9P, OR5-39, OR5-84', description='olfactory receptor family 4 subfamily H member 6 pseudogene', other_designations='OR5-39OR5-84|olfactory receptor, family 4, subfamily H, member 6 pseudogene|olfactory receptor, family 4, subfamily H, member 8 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22056903, end_position_on_the_genomic_accession=22057538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102524628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102524628, CurrentID=0, Status='live', Symbol='RASGRF2-AS1', Aliases='', description='RASGRF2 antisense RNA 1', other_designations='CTC-459I6.1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80947693, end_position_on_the_genomic_accession=80960907, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26330, CurrentID=0, Status='live', Symbol='GAPDHS', Aliases='GAPD2, GAPDH-2, GAPDS, HEL-S-278, HSD-35', description='glyceraldehyde-3-phosphate dehydrogenase, spermatogenic', other_designations='epididymis secretory protein Li 278|spermatogenic cell-specific glyceraldehyde 3-phosphate dehydrogenase 2|spermatogenic glyceraldehyde-3-phosphate dehydrogenase', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35533412, end_position_on_the_genomic_accession=35545319, orientation='plus', exon_count=11, OMIM=609169, no_hdr0=''),
  26333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26333, CurrentID=0, Status='live', Symbol='OR7A17', Aliases='BC85395_4, HTPCRX19', description='olfactory receptor family 7 subfamily A member 17', other_designations='olfactory receptor OR19-20|olfactory receptor, family 7, subfamily A, member 17', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14880426, end_position_on_the_genomic_accession=14881355, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26335, CurrentID=0, Status='live', Symbol='OR7A15P', Aliases='OR19-1, OR19-134, OR19-146, OR7A16P, OR7A20P, OR7A4P', description='olfactory receptor family 7 subfamily A member 15 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 15 pseudogene|olfactory receptor, family 7, subfamily A, member 16 pseudogene|olfactory receptor, family 7, subfamily A, member 4 pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14927312, end_position_on_the_genomic_accession=14927961, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  550624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550624, CurrentID=0, Status='live', Symbol='KTCN4', Aliases='', description='Keratoconus 4', other_designations='', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609271, no_hdr0=''),
  550625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550625, CurrentID=0, Status='live', Symbol='ANIC', Aliases='', description='Anosmia, isolated congenital', other_designations='', map_location='18p11.23-q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=107200, no_hdr0=''),
  26338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26338, CurrentID=0, Status='live', Symbol='OR5L2', Aliases='HSHTPCRX16, HTPCRX16, OR11-153', description='olfactory receptor family 5 subfamily L member 2', other_designations='olfactory receptor OR11-153|olfactory receptor, family 5, subfamily L, member 1|olfactory receptor, family 5, subfamily L, member 2', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55827219, end_position_on_the_genomic_accession=55828154, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26339, CurrentID=0, Status='live', Symbol='OR5K1', Aliases='HSHTPCRX10, HTPCRX10, OR3-8', description='olfactory receptor family 5 subfamily K member 1', other_designations='olfactory receptor OR3-8|olfactory receptor, family 5, subfamily K, member 1', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98469577, end_position_on_the_genomic_accession=98470528, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26341, CurrentID=0, Status='live', Symbol='OR5H1', Aliases='HSHTPCRX14, HTPCRX14', description='olfactory receptor family 5 subfamily H member 1', other_designations='olfactory receptor, family 5, subfamily H, member 1', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98132698, end_position_on_the_genomic_accession=98133639, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  550631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550631, CurrentID=0, Status='live', Symbol='CCDC157', Aliases='', description='coiled-coil domain containing 157', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30356638, end_position_on_the_genomic_accession=30378383, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26353, CurrentID=0, Status='live', Symbol='HSPB8', Aliases='CMT2L, DHMN2, E2IG1, H11, HMN2, HMN2A, HSP22', description='heat shock protein family B (small) member 8', other_designations='E2-induced gene 1 protein|alpha-crystallin C chain|heat shock 22kDa protein 8|heat shock 27kDa protein 8|protein kinase H11|small stress protein-like protein HSP22', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119178790, end_position_on_the_genomic_accession=119194746, orientation='plus', exon_count=3, OMIM=608014, no_hdr0=''),
  26354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26354, CurrentID=0, Status='live', Symbol='GNL3', Aliases='C77032, E2IG3, NNP47, NS', description='guanine nucleotide binding protein-like 3 (nucleolar)', other_designations='E2-induced gene 3 protein|estradiol-induced nucleotide binding protein|novel nucleolar protein 47|nucleolar GTP-binding protein 3|nucleostemin', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52685920, end_position_on_the_genomic_accession=52694497, orientation='plus', exon_count=15, OMIM=608011, no_hdr0=''),
  26355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26355, CurrentID=0, Status='live', Symbol='FAM162A', Aliases='C3orf28, E2IG5, HGTD-P', description='family with sequence similarity 162 member A', other_designations='E2-induced gene 5 protein|HIF-1 alpha-responsive proapoptotic molecule|family with sequence similarity 162, member A|growth and transformation-dependent protein', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122384176, end_position_on_the_genomic_accession=122410114, orientation='plus', exon_count=5, OMIM=608017, no_hdr0=''),
  550644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550644, CurrentID=0, Status='live', Symbol='LOC550644', Aliases='', description='structural maintenance of chromosomes 3 pseudogene', other_designations='CSPG6 pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99101979, end_position_on_the_genomic_accession=99103101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26238, CurrentID=0, Status='live', Symbol='LINC01558', Aliases='C6orf123, HGC6.2, LINC01557, dJ431P23.4', description='long intergenic non-protein coding RNA 1558', other_designations='long intergenic non-protein coding RNA 1557', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167784537, end_position_on_the_genomic_accession=167796859, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100820732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820732, CurrentID=0, Status='live', Symbol='LSM1P1', Aliases='', description='LSM1 homolog, mRNA degradation associated pseudogene 1', other_designations='LSM1 homolog, U6 small nuclear RNA associated pseudogene|LSM1 mRNA degradation associated pseudogene 1|LSM1, U6 small nuclear RNA associated pseudogene 1', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=16775230, end_position_on_the_genomic_accession=16775743, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100820733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820733, CurrentID=0, Status='live', Symbol='LOC100820733', Aliases='', description='LSM2 homolog, U6 small nuclear RNA associated (S. cerevisiae) pseudogene', other_designations='', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108893990, end_position_on_the_genomic_accession=108894670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100820734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820734, CurrentID=0, Status='live', Symbol='LOC100820734', Aliases='', description='LSM2 homolog, U6 small nuclear RNA associated (S. cerevisiae) pseudogene', other_designations='', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10597999, end_position_on_the_genomic_accession=10598980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100820735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820735, CurrentID=0, Status='live', Symbol='HIST1H2BPS3', Aliases='', description='histone cluster 1, H2b, pseudogene 3', other_designations='HIST1H2B pseudogene 3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21483834, end_position_on_the_genomic_accession=21484174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100820741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820741, CurrentID=0, Status='live', Symbol='RBBP4P6', Aliases='', description='retinoblastoma binding protein 4 pseudogene 6', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85190627, end_position_on_the_genomic_accession=85190974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422178, CurrentID=0, Status='live', Symbol='LOC100422178', Aliases='', description='olfactory receptor, family 10, subfamily Q, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100820757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820757, CurrentID=0, Status='live', Symbol='VUR4', Aliases='', description='Vesicoureteral reflux 4', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614317, no_hdr0=''),
  100820758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820758, CurrentID=0, Status='live', Symbol='CCCSX', Aliases='', description='Cerebral-cerebellar-coloboma syndrome, X-linked', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300864, no_hdr0=''),
  100820759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820759, CurrentID=0, Status='live', Symbol='SPGFX2', Aliases='', description='Spermatogenic failure, X-linked, 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309120, no_hdr0=''),
  100820760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820760, CurrentID=0, Status='live', Symbol='VUR5', Aliases='', description='Vesicoureteral reflux 5', other_designations='', map_location='13q33.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614318, no_hdr0=''),
  100820761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820761, CurrentID=0, Status='live', Symbol='MRXSCS', Aliases='', description='Mental retardation, X-linked, syndromic, Chudley-Schwartz type', other_designations='', map_location='Xq21.33-q23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300861, no_hdr0=''),
  100820762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820762, CurrentID=0, Status='live', Symbol='CPBHM', Aliases='', description='Chondrodysplasia with platyspondyly, distinctive brachydactyly, hydrocephaly, and microphthalmia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300863, no_hdr0=''),
  100820763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820763, CurrentID=0, Status='live', Symbol='VUR6', Aliases='', description='Vesicoureteral reflux 6', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614319, no_hdr0=''),
  157317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157317, CurrentID=0, Status='live', Symbol='CYCSP55', Aliases='CYCSL1, HCP15, HS11, bA513I15.3', description='cytochrome c, somatic pseudogene 55', other_designations='cytochrome c, somatic-like 1 (pseudogene)', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34219372, end_position_on_the_genomic_accession=34220937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419762, CurrentID=0, Status='live', Symbol='LOC100419762', Aliases='', description='zinc finger protein 773 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87867610, end_position_on_the_genomic_accession=87868429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157489, CurrentID=0, Status='live', Symbol='SDAD1P1', Aliases='', description='SDA1 domain containing 1 pseudogene 1', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26378730, end_position_on_the_genomic_accession=26381636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616166, CurrentID=0, Status='live', Symbol='MIR4796', Aliases='mir-4796', description='microRNA 4796', other_designations='hsa-mir-4796', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114743445, end_position_on_the_genomic_accession=114743525, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  157503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157503, CurrentID=0, Status='live', Symbol='LOC157503', Aliases='', description='uncharacterized LOC157503', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  157506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157506, CurrentID=0, Status='live', Symbol='RDH10', Aliases='SDR16C4', description='retinol dehydrogenase 10 (all-trans)', other_designations='short chain dehydrogenase/reductase family 16C member 4|short chain dehydrogenase/reductase family 16C, member 4', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73294602, end_position_on_the_genomic_accession=73325285, orientation='plus', exon_count=7, OMIM=607599, no_hdr0=''),
  105373494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373494, CurrentID=0, Status='live', Symbol='LOC105373494', Aliases='', description='uncharacterized LOC105373494', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96049531, end_position_on_the_genomic_accession=96057639, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  157530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157530, CurrentID=0, Status='live', Symbol='ST13P6', Aliases='FAM10A6, FAM10A6P', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 6', other_designations='family with sequence similarity 10, member A6', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133407664, end_position_on_the_genomic_accession=133409525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100820829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820829, CurrentID=0, Status='live', Symbol='MYZAP', Aliases='GCOM1, Gup, MYOZAP', description='myocardial zonula adherens protein', other_designations='GRINL1A complex locus upstream|GRINL1A upstream protein|myocardial intercalated disc protein|myocardium-enriched ZO1-associated protein|myocardium-enriched zonula adherens protein|myocardium-enriched zonula occludens-1-associated protein', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57591904, end_position_on_the_genomic_accession=57685364, orientation='plus', exon_count=13, OMIM=614071, no_hdr0=''),
  881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=881, CurrentID=0, Status='live', Symbol='CCIN', Aliases='BTBD20, KBTBD14', description='calicin', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36169392, end_position_on_the_genomic_accession=36171334, orientation='plus', exon_count=1, OMIM=603960, no_hdr0=''),
  26468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26468, CurrentID=0, Status='live', Symbol='LHX6', Aliases='LHX6.1', description='LIM homeobox 6', other_designations='LIM homeodomain protein 6.1|LIM/homeobox protein Lhx6.1', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122185652, end_position_on_the_genomic_accession=122228893, orientation='minus', exon_count=15, OMIM=608215, no_hdr0=''),
  26469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26469, CurrentID=0, Status='live', Symbol='PTPN18', Aliases='BDP1, PTP-HSCF', description='protein tyrosine phosphatase, non-receptor type 18 (brain-derived)', other_designations='brain-derived phosphatase', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130356007, end_position_on_the_genomic_accession=130375409, orientation='plus', exon_count=15, OMIM=606587, no_hdr0=''),
  26470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26470, CurrentID=0, Status='live', Symbol='SEZ6L2', Aliases='BSRPA, PSK-1', description='seizure related 6 homolog (mouse)-like 2', other_designations='seizure related 6-like protein 2|type I transmembrane receptor (seizure-related protein)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29871159, end_position_on_the_genomic_accession=29899264, orientation='minus', exon_count=18, OMIM=616667, no_hdr0=''),
  26471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26471, CurrentID=0, Status='live', Symbol='NUPR1', Aliases='COM1, P8', description='nuclear protein 1, transcriptional regulator', other_designations='candidate of metastasis 1|nuclear protein, transcriptional regulator, 1|protein p8', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28537341, end_position_on_the_genomic_accession=28539174, orientation='minus', exon_count=3, OMIM=614812, no_hdr0=''),
  26472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26472, CurrentID=0, Status='live', Symbol='PPP1R14B', Aliases='PHI-1, PLCB3N, PNG, SOM172', description='protein phosphatase 1 regulatory inhibitor subunit 14B', other_designations='phospholipase C beta-3 neighboring gene protein|phospholipase C-beta-3 neighbouring gene protein|protein phosphatase 1, regulatory (inhibitor) subunit 14B', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64244479, end_position_on_the_genomic_accession=64246941, orientation='minus', exon_count=4, OMIM=601140, no_hdr0=''),
  26473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26473, CurrentID=0, Status='live', Symbol='OR7E16P', Aliases='OR19-133, OR19-9, OR7E17P, OR7E60P', description='olfactory receptor family 7 subfamily E member 16 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 16 pseudogene|olfactory receptor, family 7, subfamily E, member 17 pseudogene|olfactory receptor, family 7, subfamily E, member 60 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9191429, end_position_on_the_genomic_accession=9192077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26475, CurrentID=0, Status='live', Symbol='OR7E43P', Aliases='OR4-116, OR7E124P, OST712', description='olfactory receptor family 7 subfamily E member 43 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 124 pseudogene|olfactory receptor, family 7, subfamily E, member 43 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4174399, end_position_on_the_genomic_accession=4175051, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26476, CurrentID=0, Status='live', Symbol='OR10J1', Aliases='HGMP07J, HSHGMP07J', description='olfactory receptor family 10 subfamily J member 1', other_designations='olfactory receptor OR1-26|olfactory receptor, family 10, subfamily J, member 1|olfactory receptor-like protein HGMP07J', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159397494, end_position_on_the_genomic_accession=159440967, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26478, CurrentID=0, Status='live', Symbol='OR7E53P', Aliases='OR3-142, OR3-143, OR7E132P, OR7E78, OR7E78P', description='olfactory receptor family 7 subfamily E member 53 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 132 pseudogene|olfactory receptor, family 7, subfamily E, member 53 pseudogene|olfactory receptor, family 7, subfamily E, member 78 pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125734493, end_position_on_the_genomic_accession=125735148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26479, CurrentID=0, Status='live', Symbol='OR7E62P', Aliases='OR2-4, OR2-52, OR2-53, OR2-75, OR7E62, OR7E63, OR7E63P, OR7E64, OR7E64P, OR7E82P, PJCG8, PJCG9', description='olfactory receptor family 7 subfamily E member 62 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 62 pseudogene|olfactory receptor, family 7, subfamily E, member 63 pseudogene|olfactory receptor, family 7, subfamily E, member 64 pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71055392, end_position_on_the_genomic_accession=71056038, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157556, CurrentID=0, Status='live', Symbol='BAALC-AS2', Aliases='BAALCOS, C8orf56', description='BAALC antisense RNA 2', other_designations='BAALC opposite strand', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103132963, end_position_on_the_genomic_accession=103141342, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929045, CurrentID=0, Status='live', Symbol='LOC101929045', Aliases='', description='uncharacterized LOC101929045', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  157562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157562, CurrentID=0, Status='live', Symbol='LOC157562', Aliases='', description='uncharacterized LOC157562', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26492, CurrentID=0, Status='live', Symbol='OR8G2', Aliases='HSTPCR120P, OR8G4, ORL206, ORL486, TPCR120, OR8G2', description='olfactory receptor family 8 subfamily G member 2', other_designations='olfactory receptor 8G4|olfactory receptor OR11-292|olfactory receptor OR11-297|olfactory receptor TPCR120|olfactory receptor, family 8, subfamily G, member 4|seven transmembrane helix receptor', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124224639, end_position_on_the_genomic_accession=124225663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26493, CurrentID=0, Status='live', Symbol='OR8B8', Aliases='TPCR85', description='olfactory receptor family 8 subfamily B member 8', other_designations='olfactory receptor OR11-316|olfactory receptor TPCR85|olfactory receptor, family 8, subfamily B, member 8|olfactory-like receptor JCG8', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124440150, end_position_on_the_genomic_accession=124441085, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26494, CurrentID=0, Status='live', Symbol='OR8G1', Aliases='HSTPCR25P, TPCR25, OR8G1', description='olfactory receptor family 8 subfamily G member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-281|olfactory receptor TPCR25|olfactory receptor, family 8, subfamily G, member 1 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124249676, end_position_on_the_genomic_accession=124250611, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157567, CurrentID=0, Status='live', Symbol='ANKRD46', Aliases='ANK-S, GENX-115279', description='ankyrin repeat domain 46', other_designations='ankyrin repeat small protein', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100509752, end_position_on_the_genomic_accession=100559786, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26496, CurrentID=0, Status='live', Symbol='OR10A3', Aliases='HSHTPCRX12, HTPCRX12', description='olfactory receptor family 10 subfamily A member 3', other_designations='olfactory receptor OR11-97|olfactory receptor, family 10, subfamily A, member 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7938576, end_position_on_the_genomic_accession=7939520, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26497, CurrentID=0, Status='live', Symbol='OR10D3', Aliases='HTPCRX09P, OR10D3', description='olfactory receptor family 10 subfamily D member 3 (non-functional)', other_designations='Olfactory receptor OR11-293|Putative olfactory receptor 10D3|olfactory receptor, family 10, subfamily D, member 3 (non-functional)|olfactory receptor, family 10, subfamily D, member 3 pseudogene|seven transmembrane helix receptor', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124185271, end_position_on_the_genomic_accession=124186206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157570, CurrentID=0, Status='live', Symbol='ESCO2', Aliases='2410004I17Rik, EFO2, RBS', description='establishment of sister chromatid cohesion N-acetyltransferase 2', other_designations='ECO1 homolog 2', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27771477, end_position_on_the_genomic_accession=27812396, orientation='plus', exon_count=14, OMIM=609353, no_hdr0=''),
  26499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26499, CurrentID=0, Status='live', Symbol='PLEK2', Aliases='', description='pleckstrin 2', other_designations='pleckstrin 2 homolog', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67385659, end_position_on_the_genomic_accession=67412174, orientation='minus', exon_count=10, OMIM=608007, no_hdr0=''),
  157574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157574, CurrentID=0, Status='live', Symbol='FBXO16', Aliases='FBX16', description='F-box protein 16', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28428408, end_position_on_the_genomic_accession=28490318, orientation='minus', exon_count=9, OMIM=608519, no_hdr0=''),
  26503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26503, CurrentID=0, Status='live', Symbol='SLC17A5', Aliases='AST, ISSD, NSD, SD, SIALIN, SIASD, SLD', description='solute carrier family 17 (acidic sugar transporter), member 5', other_designations='H(+)/nitrate cotransporter|H(+)/sialic acid cotransporter|membrane glycoprotein HP59|sialic acid storage disease|sodium/sialic acid cotransporter|solute carrier family 17 (anion/sugar transporter), member 5|vesicular H(+)/Aspartate-glutamate cotransporter', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73593379, end_position_on_the_genomic_accession=73654014, orientation='minus', exon_count=14, OMIM=604322, no_hdr0=''),
  26504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26504, CurrentID=0, Status='live', Symbol='CNNM4', Aliases='ACDP4', description='cyclin and CBS domain divalent metal cation transport mediator 4', other_designations='ancient conserved domain protein 4|ancient conserved domain-containing protein 4|cyclin M4|cyclin-M4', map_location='2q11', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96760902, end_position_on_the_genomic_accession=96811891, orientation='plus', exon_count=10, OMIM=607805, no_hdr0=''),
  26505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26505, CurrentID=0, Status='live', Symbol='CNNM3', Aliases='ACDP3', description='cyclin and CBS domain divalent metal cation transport mediator 3', other_designations='ancient conserved domain protein 3|ancient conserved domain-containing protein 3|cyclin M3|cyclin-M3', map_location='2p12-p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96814952, end_position_on_the_genomic_accession=96837466, orientation='plus', exon_count=10, OMIM=607804, no_hdr0=''),
  26507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26507, CurrentID=0, Status='live', Symbol='CNNM1', Aliases='ACDP1, CLP-1', description='cyclin and CBS domain divalent metal cation transport mediator 1', other_designations='ancient conserved domain protein 1|ancient conserved domain-containing protein 1|cyclin M1|cyclin-M1', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99329099, end_position_on_the_genomic_accession=99394330, orientation='plus', exon_count=12, OMIM=607802, no_hdr0=''),
  26508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26508, CurrentID=0, Status='live', Symbol='HEYL', Aliases='HESR3, HEY3, HRT3, bHLHb33', description='hes-related family bHLH transcription factor with YRPW motif-like', other_designations='HEY-like protein|HRT-3|class B basic helix-loop-helix protein 33|hHRT3|hHeyL|hairy-related transcription factor 3|hairy/enhancer-of-split related with YRPW motif 3', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39623431, end_position_on_the_genomic_accession=39639676, orientation='minus', exon_count=6, OMIM=609034, no_hdr0=''),
  26509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26509, CurrentID=0, Status='live', Symbol='MYOF', Aliases='FER1L3', description='myoferlin', other_designations='fer-1-like 3, myoferlin|fer-1-like family member 3|fer-1-like protein 3', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93306429, end_position_on_the_genomic_accession=93482486, orientation='minus', exon_count=57, OMIM=604603, no_hdr0=''),
  26511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26511, CurrentID=0, Status='live', Symbol='CHIC2', Aliases='BTL', description='cysteine rich hydrophobic domain 2', other_designations='BRX-like translocated in leukemia|cystein-rich hydrophobic domain 2|cysteine-rich hydrophobic domain 2 protein', map_location='4q11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54008373, end_position_on_the_genomic_accession=54065427, orientation='minus', exon_count=8, OMIM=604332, no_hdr0=''),
  26512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26512, CurrentID=0, Status='live', Symbol='INTS6', Aliases='DBI-1, DDX26, DDX26A, DICE1, HDB, INT6, Notchl2', description='integrator complex subunit 6', other_designations='DEAD box protein|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 26|RNA helicase HDB|protein deleted in cancer 1', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51353862, end_position_on_the_genomic_accession=51453236, orientation='minus', exon_count=21, OMIM=604331, no_hdr0=''),
  26513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26513, CurrentID=0, Status='live', Symbol='RPL34P2', Aliases='RPL34_11_1133', description='ribosomal protein L34 pseudogene 2', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32758267, end_position_on_the_genomic_accession=32758619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26514, CurrentID=0, Status='live', Symbol='RPL34P1', Aliases='RPL34_3_118, dJ69E11.2', description='ribosomal protein L34 pseudogene 1', other_designations='60S ribosomal protein L34-like pseudogene', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168210572, end_position_on_the_genomic_accession=168210983, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26515, CurrentID=0, Status='live', Symbol='TIMM10B', Aliases='FXC1, TIM10B, Tim9b', description='translocase of inner mitochondrial membrane 10 homolog B (yeast)', other_designations='fracture callus 1 homolog|fracture callus protein 1|mitochondrial import inner membrane translocase subunit Tim9 B', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6481447, end_position_on_the_genomic_accession=6484681, orientation='plus', exon_count=3, OMIM=607388, no_hdr0=''),
  26516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26516, CurrentID=0, Status='live', Symbol='RPS5P1', Aliases='RPS5_1_743', description='ribosomal protein S5 pseudogene 1', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116258445, end_position_on_the_genomic_accession=116259187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26517, CurrentID=0, Status='live', Symbol='TIMM13', Aliases='TIM13, TIM13BA, TIMM13B, ppv1, TIMM13', description='translocase of inner mitochondrial membrane 13 homolog (yeast)', other_designations='mitochondrial import inner membrane translocase subunit Tim13B', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2425624, end_position_on_the_genomic_accession=2427916, orientation='minus', exon_count=3, OMIM=607383, no_hdr0=''),
  100271372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271372, CurrentID=0, Status='live', Symbol='RPS25P10', Aliases='RPS25_3_1755', description='ribosomal protein S25 pseudogene 10', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43092134, end_position_on_the_genomic_accession=43092588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26519, CurrentID=0, Status='live', Symbol='TIMM10', Aliases='TIM10, TIM10A', description='translocase of inner mitochondrial membrane 10 homolog (yeast)', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57528463, end_position_on_the_genomic_accession=57530759, orientation='minus', exon_count=3, OMIM=602251, no_hdr0=''),
  26520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26520, CurrentID=0, Status='live', Symbol='TIMM9', Aliases='TIM9, TIM9A', description='translocase of inner mitochondrial membrane 9 homolog (yeast)', other_designations='', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58408494, end_position_on_the_genomic_accession=58427614, orientation='minus', exon_count=6, OMIM=607384, no_hdr0=''),
  26521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26521, CurrentID=0, Status='live', Symbol='TIMM8B', Aliases='DDP2, TIM8B', description='translocase of inner mitochondrial membrane 8 homolog B (yeast)', other_designations='DDP-like protein|deafness dystonia protein 2', map_location='11q23.1-q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112084800, end_position_on_the_genomic_accession=112086798, orientation='minus', exon_count=3, OMIM=606659, no_hdr0=''),
  26523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26523, CurrentID=0, Status='live', Symbol='AGO1', Aliases='EIF2C, EIF2C1, GERP95, Q99, hAgo1', description='argonaute 1, RISC catalytic component', other_designations='Golgi Endoplasmic Reticulum protein 95 kDa|argonaute 1|argonaute RISC catalytic component 1|argonaute1|eIF-2C 1|eIF2C 1|eukaryotic translation initiation factor 2C, 1|putative RNA-binding protein Q99', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35869791, end_position_on_the_genomic_accession=35924298, orientation='plus', exon_count=20, OMIM=606228, no_hdr0=''),
  26524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26524, CurrentID=0, Status='live', Symbol='LATS2', Aliases='KPM', description='large tumor suppressor kinase 2', other_designations='LATS (large tumor suppressor, Drosophila) homolog 2|LATS, large tumor suppressor, homolog 2|kinase phosphorylated during mitosis protein|large tumor suppressor homolog 2|serine/threonine kinase KPM|serine/threonine-protein kinase kpm|warts-like kinase', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20973036, end_position_on_the_genomic_accession=21061583, orientation='minus', exon_count=10, OMIM=604861, no_hdr0=''),
  26525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26525, CurrentID=0, Status='live', Symbol='IL36RN', Aliases='FIL1, FIL1(DELTA), FIL1D, IL-36Ra, IL1F5, IL1HY1, IL1L1, IL1RP3, IL36RA, PSORP, PSORS14', description='interleukin 36 receptor antagonist', other_designations='IL-1 related protein 3|IL-1F5 (IL-1HY1, FIL1-delta, IL-1RP3, IL-1L1, IL-1-delta)|IL-1ra homolog 1|IL1F5 (Canonical product IL-1F5a)|interleukin 1 family, member 5 (delta)|interleukin-1 HY1|interleukin-1 receptor antagonist homolog 1|interleukin-1-like protein 1', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113058638, end_position_on_the_genomic_accession=113064744, orientation='plus', exon_count=6, OMIM=605507, no_hdr0=''),
  26526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26526, CurrentID=0, Status='live', Symbol='TSPAN16', Aliases='TM-8, TM4-B, TM4SF16', description='tetraspanin 16', other_designations='tetraspanin TM4-B|transmembrane 4 superfamily member 16|tspan-16', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11296139, end_position_on_the_genomic_accession=11326996, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26528, CurrentID=0, Status='live', Symbol='DAZAP1', Aliases='', description='DAZ associated protein 1', other_designations='deleted in azoospermia-associated protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1407569, end_position_on_the_genomic_accession=1435687, orientation='plus', exon_count=15, OMIM=607430, no_hdr0=''),
  26529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26529, CurrentID=0, Status='live', Symbol='OR12D2', Aliases='DJ994E9.8, HS6M1-20', description='olfactory receptor family 12 subfamily D member 2 (gene/pseudogene)', other_designations='olfactory receptor OR6-28|olfactory receptor, family 12, subfamily D, member 2', map_location='6p22.2-p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29396639, end_position_on_the_genomic_accession=29397671, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26530, CurrentID=0, Status='live', Symbol='OR12D1', Aliases='OR12D1P, hs6M1-19', description='olfactory receptor family 12 subfamily D member 1 (gene/pseudogene)', other_designations='olfactory receptor, family 12, subfamily D, member 1 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29417165, end_position_on_the_genomic_accession=29418296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26531, CurrentID=0, Status='live', Symbol='OR11A1', Aliases='6M1-18, OR11A2, dJ994E9.6, hs6M1-18', description='olfactory receptor family 11 subfamily A member 1', other_designations='olfactory receptor 11A2|olfactory receptor OR6-30|olfactory receptor, family 11, subfamily A, member 1|olfactory receptor, family 11, subfamily A, member 2', map_location='6p22.2-p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29425504, end_position_on_the_genomic_accession=29427732, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26532, CurrentID=0, Status='live', Symbol='OR10H3', Aliases='', description='olfactory receptor family 10 subfamily H member 3', other_designations='olfactory receptor OR19-24|olfactory receptor, family 10, subfamily H, member 3', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15741393, end_position_on_the_genomic_accession=15742343, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26533, CurrentID=0, Status='live', Symbol='OR10G3', Aliases='OR14-40', description='olfactory receptor family 10 subfamily G member 3', other_designations='olfactory receptor OR14-40|olfactory receptor, family 10, subfamily G, member 3', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21569803, end_position_on_the_genomic_accession=21570744, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26534, CurrentID=0, Status='live', Symbol='OR10G2', Aliases='OR14-41', description='olfactory receptor family 10 subfamily G member 2', other_designations='olfactory receptor OR14-41 pseudogene|olfactory receptor, family 10, subfamily G, member 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21633836, end_position_on_the_genomic_accession=21634940, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26537, CurrentID=0, Status='live', Symbol='OR51A1P', Aliases='HPFH6OR', description='olfactory receptor family 51 subfamily A member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 1 pseudogene|olfactory receptor, family 51, subfamily I, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5170057, end_position_on_the_genomic_accession=5171323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26538, CurrentID=0, Status='live', Symbol='OR10H2', Aliases='', description='olfactory receptor family 10 subfamily H member 2', other_designations='olfactory receptor OR19-23|olfactory receptor, family 10, subfamily H, member 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15728024, end_position_on_the_genomic_accession=15729052, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26539, CurrentID=0, Status='live', Symbol='OR10H1', Aliases='', description='olfactory receptor family 10 subfamily H member 1', other_designations='olfactory receptor OR19-27|olfactory receptor, family 10, subfamily H, member 1', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15806957, end_position_on_the_genomic_accession=15808759, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26540, CurrentID=0, Status='live', Symbol='OR10G1P', Aliases='', description='olfactory receptor family 10 subfamily G member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily G, member 1 pseudogene|olfactory receptor, family 10, subfamily G, member 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21601720, end_position_on_the_genomic_accession=21602914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26541, CurrentID=0, Status='live', Symbol='OR10D1P', Aliases='HTPCRX03, OR10D2P, OST074', description='olfactory receptor family 10 subfamily D member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily D, member 1 pseudogene|olfactory receptor, family 10, subfamily D, member 2 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124158422, end_position_on_the_genomic_accession=124159354, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26548, CurrentID=0, Status='live', Symbol='ITGB1BP2', Aliases='CHORDC3, ITGB1BP, MELUSIN, MSTP015', description='integrin subunit beta 1 binding protein 2', other_designations='integrin beta 1 binding protein (melusin) 2|integrin beta 1 binding protein 2', map_location='Xq12-q13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71301748, end_position_on_the_genomic_accession=71305371, orientation='plus', exon_count=11, OMIM=300332, no_hdr0=''),
  100533629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533629, CurrentID=0, Status='live', Symbol='LOC100533629', Aliases='', description='FAST kinase domains 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94592386, end_position_on_the_genomic_accession=94594865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157627, CurrentID=0, Status='live', Symbol='LINC00599', Aliases='Rncr3', description='long intergenic non-protein coding RNA 599', other_designations='retinal non-coding RNA 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9900064, end_position_on_the_genomic_accession=9903329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  402420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402420, CurrentID=0, Status='live', Symbol='EEF1A1P30', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 30', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120210456, end_position_on_the_genomic_accession=120220034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  284578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284578, CurrentID=0, Status='live', Symbol='LOC284578', Aliases='', description='uncharacterized LOC284578', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205554273, end_position_on_the_genomic_accession=205556635, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  157638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157638, CurrentID=0, Status='live', Symbol='FAM84B', Aliases='BCMP101, NSE2', description='family with sequence similarity 84 member B', other_designations='breast cancer membrane protein 101|breast cancer membrane-associated protein 101|family with sequence similarity 84, member B|neurological/sensory 2', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126552438, end_position_on_the_genomic_accession=126558478, orientation='minus', exon_count=2, OMIM=609483, no_hdr0=''),
  26574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26574, CurrentID=0, Status='live', Symbol='AATF', Aliases='BFR2, CHE-1, CHE1, DED', description='apoptosis antagonizing transcription factor', other_designations='apoptosis-antagonizing transcription factor|rb-binding protein Che-1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36948875, end_position_on_the_genomic_accession=37056871, orientation='plus', exon_count=12, OMIM=608463, no_hdr0=''),
  26575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26575, CurrentID=0, Status='live', Symbol='RGS17', Aliases='RGS-17, RGSZ2, hRGS17', description='regulator of G-protein signaling 17', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153010823, end_position_on_the_genomic_accession=153131254, orientation='minus', exon_count=8, OMIM=607191, no_hdr0=''),
  26576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26576, CurrentID=0, Status='live', Symbol='SRPK3', Aliases='MSSK-1, MSSK1, STK23', description='SRSF protein kinase 3', other_designations='SFRS protein kinase 3|SR-protein-specific kinase 3|muscle-specific serine kinase 1|serine arginine rich protein-specific kinase 3|serine/arginine-rich protein-specific kinase 3|serine/threonine kinase 23|serine/threonine-protein kinase 23|serine/threonine-protein kinase SRPK3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153781001, end_position_on_the_genomic_accession=153785732, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  26577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26577, CurrentID=0, Status='live', Symbol='PCOLCE2', Aliases='PCPE2', description='procollagen C-endopeptidase enhancer 2', other_designations='procollagen C-proteinase enhancer 2|procollagen COOH-terminal proteinase enhancer 2', map_location='3q21-q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142817860, end_position_on_the_genomic_accession=142889203, orientation='minus', exon_count=9, OMIM=607064, no_hdr0=''),
  26578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26578, CurrentID=0, Status='live', Symbol='OSTF1', Aliases='OSF, SH3P2, bA235O14.1', description='osteoclast stimulating factor 1', other_designations='', map_location='9q13-q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75088482, end_position_on_the_genomic_accession=75148137, orientation='plus', exon_count=11, OMIM=610180, no_hdr0=''),
  26579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26579, CurrentID=0, Status='live', Symbol='MYEOV', Aliases='OCIM', description='myeloma overexpressed', other_designations='myeloma overexpressed (in a subset of t(11;14) positive multiple myelomas)|myeloma overexpressed gene (in a subset of t(11;14) positive multiple myelomas)|oncogene in multiple myeloma', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69294138, end_position_on_the_genomic_accession=69297287, orientation='plus', exon_count=3, OMIM=605625, no_hdr0=''),
  26580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26580, CurrentID=0, Status='live', Symbol='BSCL2', Aliases='GNG3LG, HMN5, PELD, SPG17', description='Berardinelli-Seip congenital lipodystrophy 2 (seipin)', other_designations='Bernardinelli-Seip congenital lipodystrophy type 2 protein', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62690262, end_position_on_the_genomic_accession=62709619, orientation='minus', exon_count=13, OMIM=606158, no_hdr0=''),
  26581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26581, CurrentID=0, Status='live', Symbol='DUX5', Aliases='DUX1', description='double homeobox 5', other_designations='double homeobox protein 1|double homeobox, 5', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611444, no_hdr0=''),
  26582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26582, CurrentID=0, Status='live', Symbol='DUX3', Aliases='', description='double homeobox 3', other_designations='double homeobox, 3', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611443, no_hdr0=''),
  26583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26583, CurrentID=0, Status='live', Symbol='DUX4L8', Aliases='DUX2', description='double homeobox 4 like 8', other_designations='double homeobox 2|double homeobox 4 like 8 (pseudogene)|putative double homeobox protein 2', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190067935, end_position_on_the_genomic_accession=190069219, orientation='plus', exon_count=0, OMIM=611442, no_hdr0=''),
  26584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26584, CurrentID=0, Status='live', Symbol='DUX1', Aliases='', description='double homeobox 1', other_designations='double homeobox, 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611441, no_hdr0=''),
  26585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26585, CurrentID=0, Status='live', Symbol='GREM1', Aliases='C15DUPq, CKTSF1B1, CRAC1, CRCS4, DAND2, DRM, DUP15q, GREMLIN, HMPS, HMPS1, IHG-2, MPSH, PIG2', description='gremlin 1, DAN family BMP antagonist', other_designations='DAN domain family member 2|cell proliferation-inducing gene 2 protein|colorectal adenoma and carcinoma 1|cysteine knot superfamily 1, BMP antagonist 1|down-regulated in Mos-transformed cells protein|gremlin 1, cysteine knot superfamily, homolog|gremlin 1-like protein|increased in high glucose-2', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32717974, end_position_on_the_genomic_accession=32734669, orientation='plus', exon_count=2, OMIM=603054, no_hdr0=''),
  26586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26586, CurrentID=0, Status='live', Symbol='CKAP2', Aliases='LB1, TMAP, se20-10', description='cytoskeleton associated protein 2', other_designations='CTCL tumor antigen se20-10|tumor- and microtubule-associated protein|tumor-associated microtubule-associated protein', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52455360, end_position_on_the_genomic_accession=52476631, orientation='plus', exon_count=10, OMIM=611569, no_hdr0=''),
  26589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26589, CurrentID=0, Status='live', Symbol='MRPL46', Aliases='C15orf4, LIECG2, P2ECSL', description='mitochondrial ribosomal protein L46', other_designations='L46mt|MRP-L46', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88459478, end_position_on_the_genomic_accession=88467402, orientation='minus', exon_count=4, OMIM=611851, no_hdr0=''),
  26590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26590, CurrentID=0, Status='live', Symbol='OR8B7P', Aliases='OR8B13P', description='olfactory receptor family 8 subfamily B member 7 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 7 pseudogene', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124324789, end_position_on_the_genomic_accession=124325407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26591, CurrentID=0, Status='live', Symbol='OR8B6P', Aliases='', description='olfactory receptor family 8 subfamily B member 6 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 6 pseudogene', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124338812, end_position_on_the_genomic_accession=124339712, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26592, CurrentID=0, Status='live', Symbol='OR8B5P', Aliases='', description='olfactory receptor family 8 subfamily B member 5 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 5 pseudogene', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124357609, end_position_on_the_genomic_accession=124358411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26595, CurrentID=0, Status='live', Symbol='OR8B2', Aliases='OR11-309, OR11-310', description='olfactory receptor family 8 subfamily B member 2', other_designations='olfactory receptor OR11-309|olfactory receptor OR11-310|olfactory receptor, family 8, subfamily B, member 2', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124382402, end_position_on_the_genomic_accession=124383343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26608, CurrentID=0, Status='live', Symbol='TBL2', Aliases='WBSCR13, WS-betaTRP', description='transducin (beta)-like 2', other_designations='WS beta-transducin repeats protein|Williams-Beuren syndrome chromosome region 13|transducin -like 2', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73568944, end_position_on_the_genomic_accession=73578683, orientation='minus', exon_count=9, OMIM=605842, no_hdr0=''),
  26609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26609, CurrentID=0, Status='live', Symbol='VCX', Aliases='VCX-10r-B1, VCX1, VCX10R, VCXB1, VCX', description='variable charge, X-linked', other_designations='variable charge protein on X with ten repeats|variable charge, X chromosome|variably charged protein X-B1', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=7842262, end_position_on_the_genomic_accession=7844143, orientation='plus', exon_count=3, OMIM=300229, no_hdr0=''),
  26610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26610, CurrentID=0, Status='live', Symbol='ELP4', Aliases='AN, C11orf19, PAX6NEB, PAXNEB, dJ68P15A.1, hELP4', description='elongator acetyltransferase complex subunit 4', other_designations='PAX6 neighbor gene protein|elongation protein 4 homolog', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31509729, end_position_on_the_genomic_accession=31785252, orientation='plus', exon_count=15, OMIM=606985, no_hdr0=''),
  100533106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533106, CurrentID=0, Status='live', Symbol='ZHX1-C8orf76', Aliases='', description='ZHX1-C8orf76 readthrough', other_designations='ZHX1-C8orf76 readthrough transcript protein', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123226189, end_position_on_the_genomic_accession=123274487, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26614, CurrentID=0, Status='live', Symbol='OR7E66P', Aliases='OR3.3, OR3DG, OR7E20P, OR7E6P, hg630', description='olfactory receptor family 7 subfamily E member 66 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 20 pseudogene|olfactory receptor, family 7, subfamily E, member 6 pseudogene|olfactory receptor, family 7, subfamily E, member 66 pseudogene', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75348243, end_position_on_the_genomic_accession=75348764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157693, CurrentID=0, Status='live', Symbol='FAM87A', Aliases='', description='family with sequence similarity 87 member A', other_designations='family with sequence similarity 87, member A', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=375931, end_position_on_the_genomic_accession=383174, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  157695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157695, CurrentID=0, Status='live', Symbol='TDRP', Aliases='C8orf42, Inm011, TDRP2, TDRP', description='testis development related protein', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=457018, end_position_on_the_genomic_accession=545781, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  157697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157697, CurrentID=0, Status='live', Symbol='ERICH1', Aliases='HSPC319', description='glutamate rich 1', other_designations='glutamate-rich 1', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=614737, end_position_on_the_genomic_accession=731239, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26628, CurrentID=0, Status='live', Symbol='OR7E47P', Aliases='OR7E141', description='olfactory receptor family 7 subfamily E member 47 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 47 pseudogene', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52079696, end_position_on_the_genomic_accession=52108319, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  157708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157708, CurrentID=0, Status='live', Symbol='SEC11B', Aliases='SEC11L2, SPCS4B', description='SEC11 homolog B, signal peptidase complex subunit, pseudogene', other_designations='SEC11 homolog A pseudogene|SEC11 homolog B, pseudogene|SEC11-like 2', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54522490, end_position_on_the_genomic_accession=54523381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26637, CurrentID=0, Status='live', Symbol='OR7E36P', Aliases='OR7E119P, OST024', description='olfactory receptor family 7 subfamily E member 36 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 119 pseudogene|olfactory receptor, family 7, subfamily E, member 36 pseudogene', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41431264, end_position_on_the_genomic_accession=41432258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26640, CurrentID=0, Status='live', Symbol='OR7E33P', Aliases='hg688', description='olfactory receptor family 7 subfamily E member 33 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 33 pseudogene', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67910910, end_position_on_the_genomic_accession=67912117, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157713, CurrentID=0, Status='live', Symbol='LYPLA1P3', Aliases='', description='lysophospholipase I pseudogene 3', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71164980, end_position_on_the_genomic_accession=71167456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646836, CurrentID=0, Status='live', Symbol='LOC646836', Aliases='', description='uncharacterized LOC646836', other_designations='Putative uncharacterized protein LOC646836', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=132227710, end_position_on_the_genomic_accession=132229299, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26644, CurrentID=0, Status='live', Symbol='OR7E29P', Aliases='OR7E30P, OST032, OST339', description='olfactory receptor family 7 subfamily E member 29 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 29 pseudogene|olfactory receptor, family 7, subfamily E, member 30 pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125712135, end_position_on_the_genomic_accession=125713079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26645, CurrentID=0, Status='live', Symbol='OR7E28P', Aliases='OR7E107P, OR7E133P, OR7E27P, OST128, hg616', description='olfactory receptor family 7 subfamily E member 28 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 107 pseudogene|olfactory receptor, family 7, subfamily E, member 133 pseudogene|olfactory receptor, family 7, subfamily E, member 27 pseudogene|olfactory receptor, family 7, subfamily E, member 28 pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158862346, end_position_on_the_genomic_accession=158863292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26647, CurrentID=0, Status='live', Symbol='OR7E25P', Aliases='CIT-B-440L2, OR19-12, OR19-C', description='olfactory receptor family 7 subfamily E member 25 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 25 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9204252, end_position_on_the_genomic_accession=9205232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26648, CurrentID=0, Status='live', Symbol='OR7E24', Aliases='HSHT2, OR19-8P, OR7E24Q, OR7E24', description='olfactory receptor family 7 subfamily E member 24', other_designations='olfactory receptor OR19-14', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9251044, end_position_on_the_genomic_accession=9252063, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  284581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284581, CurrentID=0, Status='live', Symbol='LOC284581', Aliases='', description='uncharacterized LOC284581', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205862079, end_position_on_the_genomic_accession=205896087, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26651, CurrentID=0, Status='live', Symbol='OR7E19P', Aliases='HSCIT-B-440L2, OR19-7, OR7E65, TPCR11, tpcr110', description='olfactory receptor family 7 subfamily E member 19 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 19 pseudogene|olfactory receptor, family 7, subfamily E, member 65', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9265393, end_position_on_the_genomic_accession=9266369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157724, CurrentID=0, Status='live', Symbol='SLC7A13', Aliases='AGT-1, AGT1, XAT2', description='solute carrier family 7 (anionic amino acid transporter), member 13', other_designations='X-amino acid transporter 2|amino acid transporter XAT2|sodium-independent aspartate/glutamate transporter 1|solute carrier family 7, (cationic amino acid transporter, y+ system) member 13', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86201763, end_position_on_the_genomic_accession=86230380, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26656, CurrentID=0, Status='live', Symbol='OR7D1P', Aliases='OR19-131, OR19-A, OR7D3, OR7D3P', description='olfactory receptor family 7 subfamily D member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily D, member 1 pseudogene|olfactory receptor, family 7, subfamily D, member 3 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9235344, end_position_on_the_genomic_accession=9235992, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26658, CurrentID=0, Status='live', Symbol='OR7C2', Aliases='CIT-HSP-87M17, OR19-18, OR7C3', description='olfactory receptor family 7 subfamily C member 2', other_designations='olfactory receptor 19-18|olfactory receptor 7C3|olfactory receptor OR19-22|olfactory receptor, family 7, subfamily C, member 2|olfactory receptor, family 7, subfamily C, member 3', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14941489, end_position_on_the_genomic_accession=14942448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26659, CurrentID=0, Status='live', Symbol='OR7A5', Aliases='HTPCR2', description='olfactory receptor family 7 subfamily A member 5', other_designations='olfactory receptor OR19-17|olfactory receptor TPCR92|olfactory receptor, family 7, subfamily A, member 5 pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14825742, end_position_on_the_genomic_accession=14833245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26661, CurrentID=0, Status='live', Symbol='OR7A8P', Aliases='HG83, OR19-11, OR7A9P, OST042', description='olfactory receptor family 7 subfamily A member 8 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 8 pseudogene|olfactory receptor, family 7, subfamily A, member 9 pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14852111, end_position_on_the_genomic_accession=14852743, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26664, CurrentID=0, Status='live', Symbol='OR7C1', Aliases='CIT-HSP-146E8, HSTPCR86P, OR19-5, OR7C4, TPCR86', description='olfactory receptor family 7 subfamily C member 1', other_designations='olfactory receptor 7C4|olfactory receptor OR19-16|olfactory receptor TPCR86|olfactory receptor, family 7, subfamily C, member 1|olfactory receptor, family 7, subfamily C, member 4', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14797568, end_position_on_the_genomic_accession=14835376, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  157739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157739, CurrentID=0, Status='live', Symbol='TDH', Aliases='SDR14E1P', description='L-threonine dehydrogenase (pseudogene)', other_designations='short chain dehydrogenase/reductase family 14E, member 1 (pseudogene)', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11339637, end_position_on_the_genomic_accession=11368452, orientation='plus', exon_count=9, OMIM=615174, no_hdr0=''),
  26668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26668, CurrentID=0, Status='live', Symbol='OR7A3P', Aliases='OR11-7b, OR14-11, OR14-59, OR19-12, OR7A12P, OR7A13P, OR7A14, OR7A14P', description='olfactory receptor family 7 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 12 pseudogene|olfactory receptor, family 7, subfamily A, member 13 pseudogene|olfactory receptor, family 7, subfamily A, member 14 pseudogene|olfactory receptor, family 7, subfamily A, member 3 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14903327, end_position_on_the_genomic_accession=14903973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26669, CurrentID=0, Status='live', Symbol='OR7A11P', Aliases='OR7A11, OST527', description='olfactory receptor family 7 subfamily A member 11 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 11 pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14916414, end_position_on_the_genomic_accession=14917402, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646853, CurrentID=0, Status='live', Symbol='LOC646853', Aliases='', description='non-histone chromosomal protein HMG-17 pseudogene', other_designations='high mobility group nucleosomal binding domain 2 pseudogene', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77278379, end_position_on_the_genomic_accession=77279850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402424, CurrentID=0, Status='live', Symbol='OR13K1P', Aliases='', description='olfactory receptor family 13 subfamily K member 1 pseudogene', other_designations='olfactory receptor, family 13, subfamily K, member 1 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131562730, end_position_on_the_genomic_accession=131563813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26679, CurrentID=0, Status='live', Symbol='OR4G3P', Aliases='OLC, OLC-7501, OR4G3, OR4G5P', description='olfactory receptor family 4 subfamily G member 3 pseudogene', other_designations='olfactory receptor, family 4, subfamily G, member 3 pseudogene|olfactory receptor, family 4, subfamily G, member 5 pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=94077, end_position_on_the_genomic_accession=94940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26680, CurrentID=0, Status='live', Symbol='OR4G2P', Aliases='OR4G7P', description='olfactory receptor family 4 subfamily G member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily G, member 2 pseudogene|olfactory receptor, family 4, subfamily G, member 7 pseudogene', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101926627, end_position_on_the_genomic_accession=101927303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157753, CurrentID=0, Status='live', Symbol='TMEM74', Aliases='NET36', description='transmembrane protein 74', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108783117, end_position_on_the_genomic_accession=108787541, orientation='minus', exon_count=2, OMIM=613935, no_hdr0=''),
  26682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26682, CurrentID=0, Status='live', Symbol='OR4F4', Aliases='OLA-7501, OR4F18', description='olfactory receptor family 4 subfamily F member 4', other_designations='HS14a-1-A|OR4F17|olfactory receptor OR19-3|olfactory receptor, family 4, subfamily F, member 18|olfactory receptor, family 4, subfamily F, member 4', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101922142, end_position_on_the_genomic_accession=101923059, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26683, CurrentID=0, Status='live', Symbol='OR4F3', Aliases='', description='olfactory receptor family 4 subfamily F member 3', other_designations='olfactory receptor OR1-1|olfactory receptor OR7-21|olfactory receptor, family 4, subfamily F, member 3', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181367287, end_position_on_the_genomic_accession=181368225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26685, CurrentID=0, Status='live', Symbol='OR4F1P', Aliases='HSDJ0609N19, OR4F1', description='olfactory receptor family 4 subfamily F member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 1 pseudogene', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105919, end_position_on_the_genomic_accession=106856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26686, CurrentID=0, Status='live', Symbol='OR4E2', Aliases='OR14-42', description='olfactory receptor family 4 subfamily E member 2', other_designations='olfactory receptor OR14-42|olfactory receptor, family 4, subfamily E, member 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21665083, end_position_on_the_genomic_accession=21666024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26687, CurrentID=0, Status='live', Symbol='OR4E1', Aliases='OR4E1P', description='olfactory receptor family 4 subfamily E member 1 (gene/pseudogene)', other_designations='olfactory receptor OR14-43|olfactory receptor, family 4, subfamily E, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21669891, end_position_on_the_genomic_accession=21671035, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26689, CurrentID=0, Status='live', Symbol='OR4D1', Aliases='OR17-23, OR4D3, OR4D4P, TPCR16', description='olfactory receptor family 4 subfamily D member 1', other_designations='olfactory receptor 4D3|olfactory receptor OR17-23 pseudogene|olfactory receptor TPCR16|olfactory receptor, family 4, subfamily D, member 1|olfactory receptor, family 4, subfamily D, member 3|olfactory receptor, family 4, subfamily D, member 4 pseudogene|seven transmembrane helix receptor', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58155154, end_position_on_the_genomic_accession=58156086, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26690, CurrentID=0, Status='live', Symbol='OR4B2P', Aliases='hg449', description='olfactory receptor family 4 subfamily B member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily B, member 2 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48227430, end_position_on_the_genomic_accession=48228332, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26692, CurrentID=0, Status='live', Symbol='OR2W1', Aliases='hs6M1-15', description='olfactory receptor family 2 subfamily W member 1', other_designations='olfactory receptor OR6-13|olfactory receptor, family 2, subfamily W, member 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29044213, end_position_on_the_genomic_accession=29045175, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26693, CurrentID=0, Status='live', Symbol='OR2V1', Aliases='OR2V1P, OST265', description='olfactory receptor family 2 subfamily V member 1', other_designations='olfactory receptor, family 2, subfamily V, member 1 pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181124357, end_position_on_the_genomic_accession=181125304, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26694, CurrentID=0, Status='live', Symbol='OR2U2P', Aliases='OR6-24, hs6M1-23', description='olfactory receptor family 2 subfamily U member 2 pseudogene', other_designations='olfactory receptor OR6-24 pseudogene|olfactory receptor, family 2, subfamily U, member 2 pseudogene', map_location='6p22.2-p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29268465, end_position_on_the_genomic_accession=29269421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26695, CurrentID=0, Status='live', Symbol='OR2U1P', Aliases='OR2AU1P, hs6M1-24', description='olfactory receptor family 2 subfamily U member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AU, member 1 pseudogene|olfactory receptor, family 2, subfamily U, member 1 pseudogene|seven transmembrane helix receptor', map_location='6p22.2-p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29262659, end_position_on_the_genomic_accession=29264079, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26696, CurrentID=0, Status='live', Symbol='OR2T1', Aliases='OR1-25', description='olfactory receptor family 2 subfamily T member 1', other_designations='olfactory receptor 1-25|olfactory receptor OR1-61|olfactory receptor, family 2, subfamily T, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248405995, end_position_on_the_genomic_accession=248407104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157769, CurrentID=0, Status='live', Symbol='FAM91A1', Aliases='', description='family with sequence similarity 91 member A1', other_designations='skeletal muscle cells re-entry induced', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123768439, end_position_on_the_genomic_accession=123815452, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  100271378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271378, CurrentID=0, Status='live', Symbol='RPS27AP14', Aliases='RPS27A_5_955', description='ribosomal protein S27a pseudogene 14', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1164414, end_position_on_the_genomic_accession=1164867, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533630, CurrentID=0, Status='live', Symbol='LOC100533630', Aliases='', description='chimerin 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183380441, end_position_on_the_genomic_accession=183381702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157773, CurrentID=0, Status='live', Symbol='C8orf48', Aliases='', description='chromosome 8 open reading frame 48', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13566843, end_position_on_the_genomic_accession=13568288, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157777, CurrentID=0, Status='live', Symbol='MCMDC2', Aliases='C8orf45', description='minichromosome maintenance domain containing 2', other_designations='MCM domain-containing protein C8orf45', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66870749, end_position_on_the_genomic_accession=66926386, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  26707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26707, CurrentID=0, Status='live', Symbol='OR2J2', Aliases='OR6-19, OR6-8, OR6.3.8, ORL684, dJ80I19.4, hs6M1-6', description='olfactory receptor family 2 subfamily J member 2', other_designations='olfactory receptor 6-8|olfactory receptor OR6-19|olfactory receptor, family 2, subfamily J, member 2', map_location='6p22.2-p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29173534, end_position_on_the_genomic_accession=29174574, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26713, CurrentID=0, Status='live', Symbol='OR2H5P', Aliases='HS271M21, OLFR42B, OR2H5, OR2H7, OR6-33, OR6-35, hs6M1-13', description='olfactory receptor family 2 subfamily H member 5 pseudogene', other_designations='olfactory receptor, family 2, subfamily H, member 5 pseudogene|olfactory receptor, family 2, subfamily H, member 7', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29574073, end_position_on_the_genomic_accession=29574651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26716, CurrentID=0, Status='live', Symbol='OR2H1', Aliases='6M1-16, HS6M1-16, OLFR42A-9004-14, OLFR42A-9004.14/9026.2, OR2H6, OR2H8, OR6-2, dJ994E9.4', description='olfactory receptor family 2 subfamily H member 1', other_designations='olfactory receptor 2H6|olfactory receptor 2H8|olfactory receptor 6-2|olfactory receptor OR6-32|olfactory receptor, family 2, subfamily H, member 1|olfactory receptor, family 2, subfamily H, member 6|olfactory receptor, family 2, subfamily H, member 8', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29457133, end_position_on_the_genomic_accession=29464322, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26717, CurrentID=0, Status='live', Symbol='OR2G1P', Aliases='OR6-22, OST619, hs6M1-25', description='olfactory receptor family 2 subfamily G member 1 pseudogene', other_designations='olfactory receptor OR6-22 pseudogene|olfactory receptor, family 2, subfamily G, member 1 pseudogene', map_location='6p22.2-p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29229230, end_position_on_the_genomic_accession=29230220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26719, CurrentID=0, Status='live', Symbol='OR2E1P', Aliases='HS29K1, HSNH0569I24, OR2E1, OR2E2, hs6M1-9, hs6M1-9p', description='olfactory receptor family 2 subfamily E member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily E, member 1 pseudogene|olfactory receptor, family 2, subfamily E, member 2', map_location='6p22-p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28455530, end_position_on_the_genomic_accession=28456157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271379, CurrentID=0, Status='live', Symbol='RPS27AP15', Aliases='RPS27A_6_985', description='ribosomal protein S27a pseudogene 15', other_designations='', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72741152, end_position_on_the_genomic_accession=72741604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26735, CurrentID=0, Status='live', Symbol='OR1L3', Aliases='OR9-28, OR9-D', description='olfactory receptor family 1 subfamily L member 3', other_designations='olfactory receptor 9-D|olfactory receptor OR9-28|olfactory receptor, family 1, subfamily L, member 3', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122675130, end_position_on_the_genomic_accession=122676104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26737, CurrentID=0, Status='live', Symbol='OR1L1', Aliases='HG23, OR1L2, OR9-27, OR9-C', description='olfactory receptor family 1 subfamily L member 1', other_designations='olfactory receptor 1L2|olfactory receptor 9-C|olfactory receptor OR9-27|olfactory receptor, family 1, subfamily L, member 1|olfactory receptor, family 1, subfamily L, member 2', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122661716, end_position_on_the_genomic_accession=122662648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441722, CurrentID=0, Status='live', Symbol='LOC441722', Aliases='', description='U2 small nuclear RNA auxiliary factor 1-like 4 pseudogene', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35092441, end_position_on_the_genomic_accession=35093365, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26740, CurrentID=0, Status='live', Symbol='OR1J2', Aliases='HG152, HSA5, OR1J3, OR1J5, OR9-19, OST044', description='olfactory receptor family 1 subfamily J member 2', other_designations='HTPCRX15|olfactory receptor 1J3|olfactory receptor 1J5|olfactory receptor OR9-19|olfactory receptor, family 1, subfamily J, member 2|olfactory receptor, family 1, subfamily J, member 3|olfactory receptor, family 1, subfamily J, member 5', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122448953, end_position_on_the_genomic_accession=122580751, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  26742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26742, CurrentID=0, Status='live', Symbol='OR1H1P', Aliases='OR1H1, OST26', description='olfactory receptor family 1 subfamily H member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily H, member 1 pseudogene', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122607865, end_position_on_the_genomic_accession=122608798, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26747, CurrentID=0, Status='live', Symbol='NUFIP1', Aliases='NUFIP, bA540M5.1', description='nuclear fragile X mental retardation protein interacting protein 1', other_designations='nuclear FMRP-interacting protein 1', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44939249, end_position_on_the_genomic_accession=44989478, orientation='minus', exon_count=11, OMIM=604354, no_hdr0=''),
  26748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26748, CurrentID=0, Status='live', Symbol='GAGE12I', Aliases='AL4, CT4.7, GAGE-7B, GAGE7B', description='G antigen 12I', other_designations='G antigen 7B|cancer/testis antigen 4.7|g antigen 12F/G/I', map_location='Xp11.23 not on reference assembly', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49341210, end_position_on_the_genomic_accession=49529985, orientation='plus', exon_count=5, OMIM=300637, no_hdr0=''),
  26749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26749, CurrentID=0, Status='live', Symbol='GAGE2E', Aliases='GAGE-2E, GAGE8', description='G antigen 2E', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300736, no_hdr0=''),
  26750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26750, CurrentID=0, Status='live', Symbol='RPS6KC1', Aliases='RPK118, RSKL1, S6K-delta-1, S6PKh1, humS6PKh1', description='ribosomal protein S6 kinase, 52kDa, polypeptide 1', other_designations='52 kDa ribosomal protein S6 kinase|SPHK1-binding protein|ribosomal S6 kinase-like protein with two PSK domains 118 kDa protein', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213051233, end_position_on_the_genomic_accession=213273465, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  26751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26751, CurrentID=0, Status='live', Symbol='SH3YL1', Aliases='RAY', description='SH3 and SYLF domain containing 1', other_designations='SH3 domain containing, Ysc84-like 1', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218136, end_position_on_the_genomic_accession=264866, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105373504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373504, CurrentID=0, Status='live', Symbol='LOC105373504', Aliases='', description='uncharacterized LOC105373504', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99697612, end_position_on_the_genomic_accession=99701828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26762, CurrentID=0, Status='live', Symbol='HAVCR1', Aliases='CD365, HAVCR, HAVCR-1, KIM-1, KIM1, TIM, TIM-1, TIM1, TIMD-1, TIMD1', description='hepatitis A virus cellular receptor 1', other_designations='T cell immunoglobin domain and mucin domain protein 1|T-cell immunoglobulin mucin family member 1|T-cell immunoglobulin mucin receptor 1|T-cell membrane protein 1|kidney injury molecule 1', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156979480, end_position_on_the_genomic_accession=157069527, orientation='minus', exon_count=18, OMIM=606518, no_hdr0=''),
  100034700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034700, CurrentID=0, Status='live', Symbol='AA1', Aliases='', description='Alopecia areata 1', other_designations='', map_location='18p11.3-p11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=104000, no_hdr0=''),
  26765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26765, CurrentID=0, Status='live', Symbol='SNORD12C', Aliases='E2, E2-1, E3, RNU106, SNORD106, U106', description='small nucleolar RNA, C/D box 12C', other_designations='RNA, U106 small nucleolar|small nucleolar RNA, C/D box 106', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49278945, end_position_on_the_genomic_accession=49279023, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26766, CurrentID=0, Status='live', Symbol='RNU105C', Aliases='E1-5, E1-6', description='RNA, U105C small nucleolar', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54330686, end_position_on_the_genomic_accession=54330895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100034703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034703, CurrentID=0, Status='live', Symbol='AA2', Aliases='', description='Alopecia areata 2', other_designations='', map_location='16q11-q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610753, no_hdr0=''),
  26768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26768, CurrentID=0, Status='live', Symbol='SNORA73B', Aliases='105A, E1c, RNU105A, RNU17B, U17B', description='small nucleolar RNA, H/ACA box 73B', other_designations='RNA, U105A small nucleolar|RNA, U17b small nuclear|RNA, U17b small nucleolar', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28508558, end_position_on_the_genomic_accession=28508715, orientation='plus', exon_count=1, OMIM=603239, no_hdr0=''),
  100034705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034705, CurrentID=0, Status='live', Symbol='MMVP3', Aliases='', description='Mitral valve prolapse, myxomatous 3', other_designations='', map_location='13q31.3-q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610840, no_hdr0=''),
  26770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26770, CurrentID=0, Status='live', Symbol='SNORD79', Aliases='RNU103, U79, Z22', description='small nucleolar RNA, C/D box 79', other_designations='RNA, U103 small nucleolar', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173865350, end_position_on_the_genomic_accession=173865430, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26771, CurrentID=0, Status='live', Symbol='SNORD102', Aliases='RNU102, Z18', description='small nucleolar RNA, C/D box 102', other_designations='C/D box snoRNA U102|RNA, U102 small nucleolar', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27255064, end_position_on_the_genomic_accession=27255135, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26772, CurrentID=0, Status='live', Symbol='SNORD4B', Aliases='RNU101B, Z17B', description='small nucleolar RNA, C/D box 4B', other_designations='RNA, U101B small nucleolar|U101B small nucleolar RNA|U101B snoRNA|Z17B small nucleolar RNA|Z17B snoRNA', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28723681, end_position_on_the_genomic_accession=28723754, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26773, CurrentID=0, Status='live', Symbol='SNORD4A', Aliases='RNU101A, Z17A, mgh18S-121', description='small nucleolar RNA, C/D box 4A', other_designations='RNA, U101A small nucleolar|U101A small nucleolar RNA|U101A snoRNA|Z17A small nucleolar RNA|Z17A snoRNA', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28722582, end_position_on_the_genomic_accession=28722653, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26774, CurrentID=0, Status='live', Symbol='SNORD80', Aliases='RNU100, U80, Z15', description='small nucleolar RNA, C/D box 80', other_designations='RNA, U100 small nucleolar', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173864833, end_position_on_the_genomic_accession=173864903, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26775, CurrentID=0, Status='live', Symbol='SNORA72', Aliases='RNU72, U72', description='small nucleolar RNA, H/ACA box 72', other_designations='RNA, U72 small nucleolar|U72 small nucleolar RNA|U72 snoRNA', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98042086, end_position_on_the_genomic_accession=98042217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  157848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157848, CurrentID=0, Status='live', Symbol='NKX6-3', Aliases='NKX6.3', description='NK6 homeobox 3', other_designations='NK6 transcription factor related, locus 3', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41646310, end_position_on_the_genomic_accession=41647359, orientation='minus', exon_count=2, OMIM=610772, no_hdr0=''),
  26777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26777, CurrentID=0, Status='live', Symbol='SNORA71A', Aliases='RNU71A, U71a', description='small nucleolar RNA, H/ACA box 71A', other_designations='RNA, U71A small nucleolar', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38427306, end_position_on_the_genomic_accession=38427443, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26778, CurrentID=0, Status='live', Symbol='SNORA70', Aliases='DXS648E, RNU70, U70', description='small nucleolar RNA, H/ACA box 70', other_designations='RNA, U70 small nucleolar|U70 small nucleolar RNA|U70 snoRNA', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154400281, end_position_on_the_genomic_accession=154400415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26779, CurrentID=0, Status='live', Symbol='SNORA69', Aliases='RNU69, U69, U69A', description='small nucleolar RNA, H/ACA box 69', other_designations='U69 small nucleolar RNA|U69 snoRNA|U69a small nucleolar RNA', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119787353, end_position_on_the_genomic_accession=119787484, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26780, CurrentID=0, Status='live', Symbol='SNORA68', Aliases='RNU68, U68', description='small nucleolar RNA, H/ACA box 68', other_designations='RNA, U68 small nucleolar|U68 small nucleolar RNA|U68 snoRNA', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17862588, end_position_on_the_genomic_accession=17862720, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26781, CurrentID=0, Status='live', Symbol='SNORA67', Aliases='RNU67, U67', description='small nucleolar RNA, H/ACA box 67', other_designations='RNA, U67 small nucleolar', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7577955, end_position_on_the_genomic_accession=7578091, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26782, CurrentID=0, Status='live', Symbol='SNORA66', Aliases='RNU66, U66', description='small nucleolar RNA, H/ACA box 66', other_designations='RNA, U66 small nucleolar', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92840719, end_position_on_the_genomic_accession=92840851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157855, CurrentID=0, Status='live', Symbol='KCNU1', Aliases='KCNMC1, KCa5, KCa5.1, Kcnma3, Slo3', description='potassium channel, subfamily U, member 1', other_designations='Calcium-activated potassium channel subunit alpha-3|Calcium-activated potassium channel, subfamily M subunit alpha-3|Slowpoke homolog 3', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36784324, end_position_on_the_genomic_accession=36936125, orientation='plus', exon_count=33, OMIM=615215, no_hdr0=''),
  26784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26784, CurrentID=0, Status='live', Symbol='SNORA64', Aliases='RNU64, U64', description='small nucleolar RNA, H/ACA box 64', other_designations='RNA, U64 small nucleolar|U64 small nucleolar RNA|U64 snoRNA', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1962973, end_position_on_the_genomic_accession=1963106, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26785, CurrentID=0, Status='live', Symbol='SNORD63', Aliases='RNU63, U63', description='small nucleolar RNA, C/D box 63', other_designations='RNA, U63 small nucleolar', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138561043, end_position_on_the_genomic_accession=138561110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26786, CurrentID=0, Status='live', Symbol='SNORD62A', Aliases='RNU62, U62, U62A', description='small nucleolar RNA, C/D box 62A', other_designations='RNA, U62 small nucleolar', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131485665, end_position_on_the_genomic_accession=131485750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26787, CurrentID=0, Status='live', Symbol='SNORD61', Aliases='HBII-342, RNU61, U61', description='small nucleolar RNA, C/D box 61', other_designations='RNA, U61 small nuclear', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136879199, end_position_on_the_genomic_accession=136879271, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26788, CurrentID=0, Status='live', Symbol='SNORD60', Aliases='RNU60, U60', description='small nucleolar RNA, C/D box 60', other_designations='RNA, U60 small nuclear', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2155023, end_position_on_the_genomic_accession=2155105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26789, CurrentID=0, Status='live', Symbol='SNORD59A', Aliases='RNU59, U59', description='small nucleolar RNA, C/D box 59A', other_designations='RNA, U59 small nuclear', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56645027, end_position_on_the_genomic_accession=56645101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26790, CurrentID=0, Status='live', Symbol='SNORD58B', Aliases='RNU58B, U58b', description='small nucleolar RNA, C/D box 58B', other_designations='RNA, U58B small nuclear', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49491664, end_position_on_the_genomic_accession=49491729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26791, CurrentID=0, Status='live', Symbol='SNORD58A', Aliases='RNU58A, U58a', description='small nucleolar RNA, C/D box 58A', other_designations='RNA, U58A small nuclear', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49491283, end_position_on_the_genomic_accession=49491347, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26792, CurrentID=0, Status='live', Symbol='SNORD57', Aliases='RNU57, U57', description='small nucleolar RNA, C/D box 57', other_designations='RNA, U57 small nuclear', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2656939, end_position_on_the_genomic_accession=2657010, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26793, CurrentID=0, Status='live', Symbol='SNORD56', Aliases='RNU56, U56', description='small nucleolar RNA, C/D box 56', other_designations='RNA, U56 small nuclear', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2656624, end_position_on_the_genomic_accession=2656694, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26795, CurrentID=0, Status='live', Symbol='SNORD54', Aliases='RNU54, U54', description='small nucleolar RNA, C/D box 54', other_designations='RNA, U54 small nuclear|U54 small nuclear RNA|U54 snoRNA', map_location='8q12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56073839, end_position_on_the_genomic_accession=56073901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26796, CurrentID=0, Status='live', Symbol='SNORD53', Aliases='RNU53, U53', description='small nucleolar RNA, C/D box 53', other_designations='RNA, U53 small nuclear', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28927067, end_position_on_the_genomic_accession=28927142, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  157869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157869, CurrentID=0, Status='live', Symbol='SBSPON', Aliases='C8orf84, RPESP', description='somatomedin B and thrombospondin type 1 domain containing', other_designations='RPE spondin|RPE-spondin|somatomedin B and thrombospondin, type 1 domain containing', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73064543, end_position_on_the_genomic_accession=73093272, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26798, CurrentID=0, Status='live', Symbol='SNORD51', Aliases='RNU51, U51', description='small nucleolar RNA, C/D box 51', other_designations='C/D box snoRNA U51|RNA, U51 small nuclear', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206161881, end_position_on_the_genomic_accession=206161950, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26799, CurrentID=0, Status='live', Symbol='SNORD50A', Aliases='RNU50, U50', description='small nucleolar RNA, C/D box 50A', other_designations='RNA, U50 small nuclear', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85677294, end_position_on_the_genomic_accession=85677368, orientation='minus', exon_count=1, OMIM=613117, no_hdr0=''),
  26800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26800, CurrentID=0, Status='live', Symbol='SNORD49A', Aliases='RNU49, U49, U49A', description='small nucleolar RNA, C/D box 49A', other_designations='RNA, U49 small nuclear', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16440036, end_position_on_the_genomic_accession=16440106, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26801, CurrentID=0, Status='live', Symbol='SNORD48', Aliases='RNU48, U48', description='small nucleolar RNA, C/D box 48', other_designations='RNA, U48 small nuclear', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31835263, end_position_on_the_genomic_accession=31835326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26802, CurrentID=0, Status='live', Symbol='SNORD47', Aliases='RNU47, U47', description='small nucleolar RNA, C/D box 47', other_designations='RNA, U47 small nuclear', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173864369, end_position_on_the_genomic_accession=173864445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26804, CurrentID=0, Status='live', Symbol='SNORD45B', Aliases='RNU45B, U45b', description='small nucleolar RNA, C/D box 45B', other_designations='RNA, U45B small nuclear', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75789477, end_position_on_the_genomic_accession=75789547, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26805, CurrentID=0, Status='live', Symbol='SNORD45A', Aliases='RNU45A, U45a', description='small nucleolar RNA, C/D box 45A', other_designations='RNA, U45A small nuclear', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75787889, end_position_on_the_genomic_accession=75787972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26806, CurrentID=0, Status='live', Symbol='SNORD44', Aliases='RNU44, U44', description='small nucleolar RNA, C/D box 44', other_designations='RNA, U44 small nuclear', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173865968, end_position_on_the_genomic_accession=173866028, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26807, CurrentID=0, Status='live', Symbol='SNORD43', Aliases='RNU43, U43', description='small nucleolar RNA, C/D box 43', other_designations='RNA, U43 small nuclear|U43 small nuclear RNA|U43 small nucleolar RNA|U43 snoRNA', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39319052, end_position_on_the_genomic_accession=39319113, orientation='minus', exon_count=1, OMIM=611068, no_hdr0=''),
  26808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26808, CurrentID=0, Status='live', Symbol='SNORD42B', Aliases='RNU42B, U42B', description='small nucleolar RNA, C/D box 42B', other_designations='RNA, U42B small nucleolar|U42B small nucleolar RNA|U42B snoRNA', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28720550, end_position_on_the_genomic_accession=28720616, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26809, CurrentID=0, Status='live', Symbol='SNORD42A', Aliases='RNU42A, U42, U42A', description='small nucleolar RNA, C/D box 42A', other_designations='RNA, U42A small nuclear|U42 small nuclear RNA|U42A snoRNA', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28723430, end_position_on_the_genomic_accession=28723491, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26810, CurrentID=0, Status='live', Symbol='SNORD41', Aliases='RNU41, U41', description='small nucleolar RNA, C/D box 41', other_designations='RNA, U41 small nuclear', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12706449, end_position_on_the_genomic_accession=12706518, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26811, CurrentID=0, Status='live', Symbol='SNORD55', Aliases='RNU39, RNU55, SNORD39, U39, U55', description='small nucleolar RNA, C/D box 55', other_designations='RNA, U39 small nucleolar|RNA, U55 small nuclear|U39 small nucleolar RNA|U39 snoRNA|small nucleolar RNA, C/D box 39', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44775865, end_position_on_the_genomic_accession=44775938, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26812, CurrentID=0, Status='live', Symbol='SNORD37', Aliases='RNU37, U37', description='small nucleolar RNA, C/D box 37', other_designations='RNA, U37 small nucleolar', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3982507, end_position_on_the_genomic_accession=3982572, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26813, CurrentID=0, Status='live', Symbol='SNORD36C', Aliases='RNU36C, U36c', description='small nucleolar RNA, C/D box 36C', other_designations='RNA, U36C small nucleolar|U36c small nucleolar RNA|U36c snoRNA', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133350846, end_position_on_the_genomic_accession=133350913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26814, CurrentID=0, Status='live', Symbol='SNORD36B', Aliases='RNU36B, U36b', description='small nucleolar RNA, C/D box 36B', other_designations='RNA, U36B small nucleolar|U36b small nucleolar RNA|U36b snoRNA', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133350094, end_position_on_the_genomic_accession=133350164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26815, CurrentID=0, Status='live', Symbol='SNORD36A', Aliases='RNU36A, U36a', description='small nucleolar RNA, C/D box 36A', other_designations='RNA, U36A small nucleolar|U36a small nucleolar RNA|U36a snoRNA', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133350456, end_position_on_the_genomic_accession=133350527, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26816, CurrentID=0, Status='live', Symbol='SNORD35A', Aliases='RNU35, RNU35A, U35', description='small nucleolar RNA, C/D box 35A', other_designations='RNA, U35 small nucleolar|U35 small nucleolar RNA|U35 snoRNA', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49491175, end_position_on_the_genomic_accession=49491260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26817, CurrentID=0, Status='live', Symbol='SNORD34', Aliases='RNU34, U34', description='small nucleolar RNA, C/D box 34', other_designations='RNA, U34 small nucleolar|U34 small nucleolar RNA|U34 snoRNA', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49490907, end_position_on_the_genomic_accession=49490972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26818, CurrentID=0, Status='live', Symbol='SNORD33', Aliases='RNU33, U33', description='small nucleolar RNA, C/D box 33', other_designations='RNA, U33 small nucleolar|U33 small nucleolar RNA|U33 snoRNA', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49490617, end_position_on_the_genomic_accession=49490699, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26819, CurrentID=0, Status='live', Symbol='SNORD32A', Aliases='RNU32, U32, U32A', description='small nucleolar RNA, C/D box 32A', other_designations='RNA, U32 small nucleolar|U32 small nucleolar RNA|U32 snoRNA', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49489966, end_position_on_the_genomic_accession=49490047, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26820, CurrentID=0, Status='live', Symbol='SNORD24', Aliases='RNU24, U24', description='small nucleolar RNA, C/D box 24', other_designations='RNA, U24 small nucleolar|U24 small nucleolar RNA|U24 snoRNA', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133349396, end_position_on_the_genomic_accession=133349470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26821, CurrentID=0, Status='live', Symbol='SNORA74A', Aliases='RNU19, U19', description='small nucleolar RNA, H/ACA box 74A', other_designations='RNA, U19 small nucleolar', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139278780, end_position_on_the_genomic_accession=139278979, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26822, CurrentID=0, Status='live', Symbol='SNORD14A', Aliases='RNU14, RNU14A, U14, U14-S13-5', description='small nucleolar RNA, C/D box 14A', other_designations='RNA, U14 small nucleolar|U14 small nucleolar RNA|U14 snoRNA', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17074653, end_position_on_the_genomic_accession=17074744, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26823, CurrentID=0, Status='live', Symbol='RNU12-2P', Aliases='RNU12, RNU12P, U12', description='RNA, U12 small nuclear 2, pseudogene', other_designations='RNA, U12 small nuclear 2 (pseudogene)|RNA, U12 small nuclear pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47132672, end_position_on_the_genomic_accession=47132824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26824, CurrentID=0, Status='live', Symbol='RNU11', Aliases='RNU11-1, U11', description='RNA, U11 small nuclear', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28648600, end_position_on_the_genomic_accession=28648734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26825, CurrentID=0, Status='live', Symbol='RNU6-50P', Aliases='RNU6-50, RNU6P1, U6', description='RNA, U6 small nuclear 50, pseudogene', other_designations='RNA, U6 small nuclear 50 (pseudogene)|RNA, U6 small nuclear pseudogene 1', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46517523, end_position_on_the_genomic_accession=46518000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26826, CurrentID=0, Status='live', Symbol='RNU6-6P', Aliases='RNU6-6, RNU6B, U6-6', description='RNA, U6 small nuclear 6, pseudogene', other_designations='RNA, U6B small nuclear', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13217269, end_position_on_the_genomic_accession=13217375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26827, CurrentID=0, Status='live', Symbol='RNU6-1', Aliases='RNU6, RNU6A, U6, U6-1', description='RNA, U6 small nuclear 1', other_designations='RNA, U6A small nuclear', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67839940, end_position_on_the_genomic_accession=67840045, orientation='minus', exon_count=1, OMIM=180692, no_hdr0=''),
  26828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26828, CurrentID=0, Status='live', Symbol='RNU5F-1', Aliases='RNU5F, U5F', description='RNA, U5F small nuclear 1', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44721786, end_position_on_the_genomic_accession=44721901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26829, CurrentID=0, Status='live', Symbol='RNU5E-1', Aliases='RNU5E, U5E', description='RNA, U5E small nuclear 1', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11908154, end_position_on_the_genomic_accession=11908272, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26830, CurrentID=0, Status='live', Symbol='RNU5D-1', Aliases='RNU5D, U5DL, U5DS', description='RNA, U5D small nuclear 1', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44731069, end_position_on_the_genomic_accession=44731168, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26831, CurrentID=0, Status='live', Symbol='RNU5A-1', Aliases='RNU5, RNU5A, RNU5C, U5A, U5B1', description='RNA, U5A small nuclear 1', other_designations='RNA, U5C small nuclear', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65296051, end_position_on_the_genomic_accession=65296167, orientation='plus', exon_count=1, OMIM=180691, no_hdr0=''),
  26832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26832, CurrentID=0, Status='live', Symbol='RNU5B-1', Aliases='RNU5BP, U5B1, RNU5B-1', description='RNA, U5B small nuclear 1', other_designations='RNA, U5B small nuclear pseudogene 1', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65304677, end_position_on_the_genomic_accession=65304793, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26834, CurrentID=0, Status='live', Symbol='RNU4-2', Aliases='RNU4-1B, RNU4B1, RNU4C, U4A, U4b, U4c', description='RNA, U4 small nuclear 2', other_designations='RNA, U4 small nuclear 1B|RNA, U4B1 small nuclear|RNA, U4C small nuclear', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120291763, end_position_on_the_genomic_accession=120291903, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26835, CurrentID=0, Status='live', Symbol='RNU4-1', Aliases='RNU4A, RNU4B2, U4, U4BL', description='RNA, U4 small nuclear 1', other_designations='RNA, U4A small nuclear|RNA, U4B2 small nuclear', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120293094, end_position_on_the_genomic_accession=120293237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26838, CurrentID=0, Status='live', Symbol='RNU4-6P', Aliases='RNU4P6, U4, U4/8', description='RNA, U4 small nuclear 6, pseudogene', other_designations='RNA, U4 small nuclear 6 (pseudogene)|RNA, U4 small nuclear pseudogene 6', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16875117, end_position_on_the_genomic_accession=16875320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26839, CurrentID=0, Status='live', Symbol='RNU4-5P', Aliases='RNU4P5, U4, U4/6', description='RNA, U4 small nuclear 5, pseudogene', other_designations='RNA, U4 small nuclear 5 (pseudogene)|RNA, U4 small nuclear pseudogene 5', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109869806, end_position_on_the_genomic_accession=109870008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26840, CurrentID=0, Status='live', Symbol='RNU4-4P', Aliases='RNU4P4, U4, U4/5', description='RNA, U4 small nuclear 4, pseudogene', other_designations='RNA, U4 small nuclear 4 (pseudogene)|RNA, U4 small nuclear pseudogene 4', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172796308, end_position_on_the_genomic_accession=172796710, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26841, CurrentID=0, Status='live', Symbol='RNU4-3P', Aliases='RNU4P3, U4, U4/4', description='RNA, U4 small nuclear 3, pseudogene', other_designations='RNA, U4 small nuclear 3 (pseudogene)|RNA, U4 small nuclear pseudogene 3', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36264737, end_position_on_the_genomic_accession=36264939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26842, CurrentID=0, Status='live', Symbol='RNU3P4', Aliases='U3, U3.6', description='RNA, U3 small nucleolar pseudogene 4', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183174937, end_position_on_the_genomic_accession=183175007, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26843, CurrentID=0, Status='live', Symbol='RNU3P3', Aliases='U3, U3.5, u3.7', description='RNA, U3 small nucleolar pseudogene 3', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85271905, end_position_on_the_genomic_accession=85272049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157916, CurrentID=0, Status='live', Symbol='IL6RP1', Aliases='IL6RL1', description='interleukin 6 receptor pseudogene 1', other_designations='interleukin 6 receptor-like 1', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90081181, end_position_on_the_genomic_accession=90082736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533631, CurrentID=0, Status='live', Symbol='LOC100533631', Aliases='', description='5-azacytidine induced 2 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7735774, end_position_on_the_genomic_accession=7736828, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157922, CurrentID=0, Status='live', Symbol='CAMSAP1', Aliases='', description='calmodulin regulated spectrin associated protein 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135808487, end_position_on_the_genomic_accession=135907251, orientation='minus', exon_count=18, OMIM=613774, no_hdr0=''),
  26851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26851, CurrentID=0, Status='live', Symbol='SNORD3B-1', Aliases='RNU3A1, U3a, U3b1, U3b2', description='small nucleolar RNA, C/D box 3B-1', other_designations='RNA, U3A1 small nucleolar, RNA, U3A1 small nucleolar', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19061912, end_position_on_the_genomic_accession=19062128, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26852, CurrentID=0, Status='live', Symbol='RNU2-5P', Aliases='RNU2P3', description='RNA, U2 small nuclear 5, pseudogene', other_designations='RNA, U2 small nuclear pseudogene 3', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70188574, end_position_on_the_genomic_accession=70188750, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26853, CurrentID=0, Status='live', Symbol='RNU2-4P', Aliases='RNU2P2, U2, U2/7', description='RNA, U2 small nuclear 4, pseudogene', other_designations='RNA, U2 small nuclear 4 (pseudogene)|RNA, U2 small nuclear pseudogene 2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43387229, end_position_on_the_genomic_accession=43387517, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26854, CurrentID=0, Status='live', Symbol='RNU2-3P', Aliases='RNU2P1, U2', description='RNA, U2 small nuclear 3, pseudogene', other_designations='RNA, U2 small nuclear 3 (pseudogene)|RNA, U2 small nuclear pseudogene 1', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95745804, end_position_on_the_genomic_accession=95745991, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157927, CurrentID=0, Status='live', Symbol='C9orf62', Aliases='', description='chromosome 9 open reading frame 62', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135343249, end_position_on_the_genomic_accession=135346562, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26858, CurrentID=0, Status='live', Symbol='RNU1-15P', Aliases='RNU1P8, U1P11', description='RNA, U1 small nuclear 15, pseudogene', other_designations='RNA, U1 small nuclear 15 (pseudogene)|RNA, U1 small nuclear pseudogene 8', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21381363, end_position_on_the_genomic_accession=21381494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26859, CurrentID=0, Status='live', Symbol='RNU1-14P', Aliases='RNU1P7, U1P17', description='RNA, U1 small nuclear 14, pseudogene', other_designations='RNA, U1 small nuclear 14 (pseudogene)|RNA, U1 small nuclear pseudogene 7', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53366058, end_position_on_the_genomic_accession=53366211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26860, CurrentID=0, Status='live', Symbol='RNU1-13P', Aliases='RNU1-71, RNU1P6, RNVU1-2, U1.1, U1P1, vU1.2', description='RNA, U1 small nuclear 13, pseudogene', other_designations='RNA, U1 small nuclear pseudogene 6|RNA, variant U1 small nuclear 2', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148388490, end_position_on_the_genomic_accession=148388651, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26861, CurrentID=0, Status='live', Symbol='RNU1-11P', Aliases='RNU1P1, RNU1P5, U1P1A', description='RNA, U1 small nuclear 11, pseudogene', other_designations='RNA, U1 small nuclear 11 (pseudogene)|RNA, U1 small nuclear pseudogene 5', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13214040, end_position_on_the_genomic_accession=13214255, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26863, CurrentID=0, Status='live', Symbol='RNVU1-18', Aliases='RNU1-25, RNU1-25P, RNU1-5, RNU1P1, RNU1P9, U1.15, U1P101, U1P15, vU1.18', description='RNA, variant U1 small nuclear 18', other_designations='RNA, U1 small nuclear 25, pseudogene|RNA, U1 small nuclear 5|RNA, U1 small nuclear pseudogene 1|RNA, U1 small nuclear pseudogene 9', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143729407, end_position_on_the_genomic_accession=143729570, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26864, CurrentID=0, Status='live', Symbol='RNVU1-7', Aliases='RNU1-26P, RNU1-6, RNU1-6P, RNU1-9, RNU1-9P, RNVU1-9, vU1.7, vU1.9', description='RNA, variant U1 small nuclear 7', other_designations='RNA, U1 small nuclear 26, pseudogene|RNA, U1 small nuclear 6, pseudogene|RNA, U1 small nuclear 9, pseudogene|RNA, variant U1 small nuclear 9', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148038753, end_position_on_the_genomic_accession=148038916, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26865, CurrentID=0, Status='live', Symbol='RNU1-27P', Aliases='RNU1-7, RNU1-7P', description='RNA, U1 small nuclear 27, pseudogene', other_designations='RNA, U1 small nuclear 7, pseudogene', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34546714, end_position_on_the_genomic_accession=34546877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26866, CurrentID=0, Status='live', Symbol='RNU1-28P', Aliases='RNU1-8, RNU1-8P', description='RNA, U1 small nuclear 28, pseudogene', other_designations='RNA, U1 small nuclear 8, pseudogene', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34556226, end_position_on_the_genomic_accession=34556389, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441711, CurrentID=0, Status='live', Symbol='LOC441711', Aliases='', description='RING finger protein 145-like', other_designations='ring finger protein 145 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22371736, end_position_on_the_genomic_accession=22373567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26869, CurrentID=0, Status='live', Symbol='RNU1-3', Aliases='HSD4, RNU1G3', description='RNA, U1 small nuclear 3', other_designations='RNA, U1G3 small nuclear', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16666785, end_position_on_the_genomic_accession=16666948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26870, CurrentID=0, Status='live', Symbol='RNU1-2', Aliases='RNU1C1, RNU1C2, U1C1, U1C21', description='RNA, U1 small nuclear 2', other_designations='RNA, U1C1 small nuclear|RNA, U1C2 small nuclear', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16895980, end_position_on_the_genomic_accession=16896143, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26871, CurrentID=0, Status='live', Symbol='RNU1-1', Aliases='HSD1, HU1-1, RNU1, RNU1A, RNU1A3, RNU1G4, Rnu1a1, U1, U1A1', description='RNA, U1 small nuclear 1', other_designations='RNA, U1A small nuclear|RNA, U1A3 small nuclear|RNA, U1G4 small nuclear', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16514122, end_position_on_the_genomic_accession=16514285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26872, CurrentID=0, Status='live', Symbol='STEAP1', Aliases='PRSS24, STEAP', description='six transmembrane epithelial antigen of the prostate 1', other_designations='six-transmembrane epithelial antigen of prostate 1', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90154375, end_position_on_the_genomic_accession=90164827, orientation='plus', exon_count=5, OMIM=604415, no_hdr0=''),
  26873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26873, CurrentID=0, Status='live', Symbol='OPLAH', Aliases='5-Opase, OPLAD, OPLAH', description='5-oxoprolinase (ATP-hydrolysing)', other_designations='5-oxo-L-prolinase|pyroglutamase', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144051266, end_position_on_the_genomic_accession=144064026, orientation='minus', exon_count=28, OMIM=614243, no_hdr0=''),
  266623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266623, CurrentID=0, Status='live', Symbol='SLC25A5P1', Aliases='ANTP3, bK250D10.4', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 1', other_designations='adenine nucleotide translocator pseudogene 3 (clone 13)', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42000433, end_position_on_the_genomic_accession=42002033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  157956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157956, CurrentID=0, Status='live', Symbol='CDC20P1', Aliases='CDC20P', description='cell division cycle 20 pseudogene 1', other_designations='CDC20 cell division cycle 20 homolog pseudogene|cell division cycle 20 homolog pseudogene 1', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87011574, end_position_on_the_genomic_accession=87013210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266625, CurrentID=0, Status='live', Symbol='NDUFA9P1', Aliases='dJ106I20.2', description='NADH:ubiquinone oxidoreductase subunit A9 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 9, 39kDa pseudogene|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 9, pseudogene 1', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35897132, end_position_on_the_genomic_accession=35898419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266626, CurrentID=0, Status='live', Symbol='CLCP1', Aliases='', description='Charcot-Leyden crystal protein pseudogene 1', other_designations='Charot-Leyden crystal protein pseudogene 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17424244, end_position_on_the_genomic_accession=17424851, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140560, CurrentID=0, Status='live', Symbol='LI5', Aliases='', description='Lamellar ichthyosis 5', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606545, no_hdr0=''),
  106480688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480688, CurrentID=0, Status='live', Symbol='BRWD1P3', Aliases='', description='bromodomain and WD repeat domain containing 1 pseudogene 3', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266629, CurrentID=0, Status='live', Symbol='SEC14L3', Aliases='TAP2', description='SEC14-like lipid binding 3', other_designations='SEC14-like 3|SEC14p-like protein TAP2|tocopherol-associated protein 2', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30447959, end_position_on_the_genomic_accession=30472047, orientation='minus', exon_count=15, OMIM=612824, no_hdr0=''),
  100419779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419779, CurrentID=0, Status='live', Symbol='LOC100419779', Aliases='', description='zinc finger protein 195 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57274253, end_position_on_the_genomic_accession=57274854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91869, CurrentID=0, Status='live', Symbol='RFT1', Aliases='CDG1N', description='RFT1 homolog', other_designations='RFT1, requiring fifty three 1 homolog|congenital disorder of glycosylation 1N|putative endoplasmic reticulum multispan transmembrane protein', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53071152, end_position_on_the_genomic_accession=53130459, orientation='minus', exon_count=19, OMIM=611908, no_hdr0=''),
  105373510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373510, CurrentID=0, Status='live', Symbol='LOC105373510', Aliases='', description='uncharacterized LOC105373510', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101023411, end_position_on_the_genomic_accession=101024307, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288844, CurrentID=0, Status='live', Symbol='LOC100288844', Aliases='', description='zinc finger protein 133-like', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23789256, end_position_on_the_genomic_accession=23790712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271386, CurrentID=0, Status='live', Symbol='RPS4XP20', Aliases='RPS4P20, RPS4X_9_1648', description='ribosomal protein S4X pseudogene 20', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34022565, end_position_on_the_genomic_accession=34023450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158011, CurrentID=0, Status='live', Symbol='IFNA12P', Aliases='IFNP12, IFNWP12', description='interferon, alpha 12, pseudogene', other_designations='interferon pseudogene 12', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21403246, end_position_on_the_genomic_accession=21403555, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377875, CurrentID=0, Status='live', Symbol='LOC105377875', Aliases='', description='uncharacterized LOC105377875', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82352779, end_position_on_the_genomic_accession=82363656, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26952, CurrentID=0, Status='live', Symbol='SMR3A', Aliases='P-B1, PBI, PRL5, PROL5', description='submaxillary gland androgen regulated protein 3A', other_designations='proline rich 5 (salivary)|proline- rich protein 5|proline-rich protein 5|proline-rich protein PBI|protein homologous to salivary proline-rich protein P-B|submaxillary gland androgen regulated protein 3 homolog A', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70360776, end_position_on_the_genomic_accession=70367106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26953, CurrentID=0, Status='live', Symbol='RANBP6', Aliases='', description='RAN binding protein 6', other_designations='Ran-GTP binding protein', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6011019, end_position_on_the_genomic_accession=6015640, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  550626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550626, CurrentID=0, Status='live', Symbol='FHASD', Aliases='', description='Foveal hypoplasia and anterior segment dysgenesis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609218, no_hdr0=''),
  26958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26958, CurrentID=0, Status='live', Symbol='COPG2', Aliases='2-COP, gamma-2-COP', description='coatomer protein complex subunit gamma 2', other_designations='coat protein, nonclathrin, gamma-2-cop|coatomer protein complex, subunit gamma 2|gamma-2-coat protein', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130506238, end_position_on_the_genomic_accession=130668758, orientation='minus', exon_count=24, OMIM=604355, no_hdr0=''),
  26959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26959, CurrentID=0, Status='live', Symbol='HBP1', Aliases='', description='HMG-box transcription factor 1', other_designations='HMG box transcription factor 1|high mobility group box transcription factor 1', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107168961, end_position_on_the_genomic_accession=107202529, orientation='plus', exon_count=13, OMIM=616714, no_hdr0=''),
  26960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26960, CurrentID=0, Status='live', Symbol='NBEA', Aliases='BCL8B, LYST2', description='neurobeachin', other_designations='lysosomal-trafficking regulator 2', map_location='13q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34942287, end_position_on_the_genomic_accession=35672737, orientation='plus', exon_count=63, OMIM=604889, no_hdr0=''),
  105373511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373511, CurrentID=0, Status='live', Symbol='LOC105373511', Aliases='', description='uncharacterized LOC105373511', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101260816, end_position_on_the_genomic_accession=101266670, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  158035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158035, CurrentID=0, Status='live', Symbol='LINC00032', Aliases='C9orf14, NCRNA00032', description='long intergenic non-protein coding RNA 32', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27245684, end_position_on_the_genomic_accession=27282793, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  158038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158038, CurrentID=0, Status='live', Symbol='LINGO2', Aliases='LERN3, LRRN6C', description='leucine rich repeat and Ig domain containing 2', other_designations='leucine rich repeat neuronal 6C|leucine-rich repeat neuronal protein 3|leucine-rich repeat neuronal protein 6C', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27937617, end_position_on_the_genomic_accession=29213981, orientation='minus', exon_count=22, OMIM=609793, no_hdr0=''),
  26973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26973, CurrentID=0, Status='live', Symbol='CHORDC1', Aliases='CHP1', description='cysteine and histidine rich domain containing 1', other_designations='CHORD-containing protein 1|CHP-1|chord domain-containing protein 1|cysteine and histidine-rich domain (CHORD) containing 1|cysteine and histidine-rich domain (CHORD)-containing 1|cysteine and histidine-rich domain (CHORD)-containing, zinc-binding protein 1|morgana|protein morgana', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90193947, end_position_on_the_genomic_accession=90223364, orientation='minus', exon_count=14, OMIM=604353, no_hdr0=''),
  158046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158046, CurrentID=0, Status='live', Symbol='NXNL2', Aliases='C9orf121, RDCVF2', description='nucleoredoxin-like 2', other_designations='rod-derived cone viability factor 2', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88535073, end_position_on_the_genomic_accession=88584466, orientation='plus', exon_count=6, OMIM=615299, no_hdr0=''),
  100820709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100820709, CurrentID=0, Status='live', Symbol='LINC00501', Aliases='', description='long intergenic non-protein coding RNA 501', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177294442, end_position_on_the_genomic_accession=177323435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26978, CurrentID=0, Status='live', Symbol='PABPC1P3', Aliases='PABP3, PABPCP3', description='poly(A) binding protein, cytoplasmic 1 pseudogene 3', other_designations='poly(A) binding protein, cytoplasmic, pseudogene 3', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74582584, end_position_on_the_genomic_accession=74583547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26980, CurrentID=0, Status='live', Symbol='PABPC1P1', Aliases='PABP1, PABPCP1', description='poly(A) binding protein, cytoplasmic 1 pseudogene 1', other_designations='poly(A) binding protein, cytoplasmic, pseudogene 1', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39973436, end_position_on_the_genomic_accession=39974377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101410534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410534, CurrentID=0, Status='live', Symbol='DLGAP1-AS4', Aliases='', description='DLGAP1 antisense RNA 4', other_designations='DLGAP1 antisense RNA 4 (non-protein coding)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3962353, end_position_on_the_genomic_accession=4013943, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26982, CurrentID=0, Status='live', Symbol='ABCD1P3', Aliases='', description='ATP binding cassette subfamily D member 1 pseudogene 3', other_designations='ATP-binding cassette, sub-family D (ALD), member 1, pseudogene 3', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32475023, end_position_on_the_genomic_accession=32475815, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26983, CurrentID=0, Status='live', Symbol='ABCD1P2', Aliases='bA453N3.6', description='ATP binding cassette subfamily D member 1 pseudogene 2', other_designations='ATP-binding cassette, sub-family D (ALD), member 1 pseudogene|ATP-binding cassette, sub-family D (ALD), member 1, pseudogene 2', map_location='10p11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38601416, end_position_on_the_genomic_accession=38605609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  26984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26984, CurrentID=0, Status='live', Symbol='SEC22A', Aliases='SEC22L2', description='SEC22 homolog A, vesicle trafficking protein', other_designations='SEC22 vesicle trafficking protein homolog A|SEC22 vesicle trafficking protein-like 2|SEC22 vesicle-trafficking protein homolog A|SEC22 vesicle-trafficking protein-like 2|sec22 homolog', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123201927, end_position_on_the_genomic_accession=123274136, orientation='plus', exon_count=9, OMIM=612442, no_hdr0=''),
  26985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26985, CurrentID=0, Status='live', Symbol='AP3M1', Aliases='', description='adaptor related protein complex 3, mu 1 subunit', other_designations='AP-3 adapter complex mu3A subunit|AP-3 adaptor complex mu3A subunit|adapter-related protein complex 3 mu-1 subunit|adapter-related protein complex 3 subunit mu-1|adaptor-related protein complex 3 subunit mu-1|adaptor-related protein complex 3, mu 1 subunit|clathrin adaptor complex AP3, mu-3A subunit|mu-adaptin 3A|mu3A-adaptin', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74120257, end_position_on_the_genomic_accession=74151068, orientation='minus', exon_count=12, OMIM=610366, no_hdr0=''),
  26986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26986, CurrentID=0, Status='live', Symbol='PABPC1', Aliases='PAB1, PABP, PABP1, PABPC2, PABPL1', description='poly(A) binding protein, cytoplasmic 1', other_designations='poly(A) binding protein, cytoplasmic 2', map_location='8q22.2-q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100702916, end_position_on_the_genomic_accession=100722998, orientation='minus', exon_count=15, OMIM=604679, no_hdr0=''),
  26343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26343, CurrentID=0, Status='live', Symbol='OR5E1P', Aliases='HSTPCR24, OR5E1, TPCR24', description='olfactory receptor family 5 subfamily E member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily E, member 1 pseudogene|olfactory receptor, family 5, subfamily R, member 1 pseudogene|seven transmembrane helix receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7849051, end_position_on_the_genomic_accession=7849571, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  158062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158062, CurrentID=0, Status='live', Symbol='LCN6', Aliases='LCN5, UNQ643, hLcn5', description='lipocalin 6', other_designations='epididymal-specific lipocalin LCN6|lipocalin 5|lipocalin-5', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136744017, end_position_on_the_genomic_accession=136748528, orientation='minus', exon_count=7, OMIM=609379, no_hdr0=''),
  26993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26993, CurrentID=0, Status='live', Symbol='AKAP8L', Aliases='HA95, HAP95, NAKAP, NAKAP95', description='A-kinase anchoring protein 8 like', other_designations='A kinase (PRKA) anchor protein 8-like|AKAP8-like protein|helicase A-binding protein 95 kDa|homologous to AKAP95 protein|neighbor of A-kinase anchoring protein 95|neighbor of AKAP95', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15380048, end_position_on_the_genomic_accession=15419121, orientation='minus', exon_count=17, OMIM=609475, no_hdr0=''),
  26994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26994, CurrentID=0, Status='live', Symbol='RNF11', Aliases='CGI-123, SID1669', description='ring finger protein 11', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51236273, end_position_on_the_genomic_accession=51273447, orientation='plus', exon_count=3, OMIM=612598, no_hdr0=''),
  158067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158067, CurrentID=0, Status='live', Symbol='AK8', Aliases='AK 8, C9orf98, DDX31', description='adenylate kinase 8', other_designations='ATP-AMP transphosphorylase 8|putative adenylate kinase-like protein C9orf98', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132725574, end_position_on_the_genomic_accession=132878811, orientation='minus', exon_count=16, OMIM=615365, no_hdr0=''),
  26996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26996, CurrentID=0, Status='live', Symbol='GPR160', Aliases='GPCR1, GPCR150', description='G protein-coupled receptor 160', other_designations='G-protein coupled receptor GPCR1|hGPCR1|putative G protein-coupled receptor', map_location='3q26.2-q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170037543, end_position_on_the_genomic_accession=170085418, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26998, CurrentID=0, Status='live', Symbol='FETUB', Aliases='16G2, Gugu, IRL685', description='fetuin B', other_designations='fetuin-like protein IRL685', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186635832, end_position_on_the_genomic_accession=186653008, orientation='plus', exon_count=8, OMIM=605954, no_hdr0=''),
  26999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26999, CurrentID=0, Status='live', Symbol='CYFIP2', Aliases='PIR121', description='cytoplasmic FMR1 interacting protein 2', other_designations='p53-inducible protein 121', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157266080, end_position_on_the_genomic_accession=157395598, orientation='plus', exon_count=35, OMIM=606323, no_hdr0=''),
  27000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27000, CurrentID=0, Status='live', Symbol='DNAJC2', Aliases='MPHOSPH11, MPP11, ZRF1, ZUO1', description='DnaJ heat shock protein family (Hsp40) member C2', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 2|M-phase phosphoprotein 11|zuotin related factor 1|zuotin-related factor 1', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103312474, end_position_on_the_genomic_accession=103344873, orientation='minus', exon_count=18, OMIM=605502, no_hdr0=''),
  27002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27002, CurrentID=0, Status='live', Symbol='PAFAH1B1P2', Aliases='LIS2P, PAFAH1P2', description='platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 pseudogene 2', other_designations='platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 (45kDa) pseudogene|platelet-activating factor acetylhydrolase, isoform Ib, pseudogene 2|platelet-activating factor acetylhydrolase, isoform Ib, subunit 1 pseudogene 2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111383562, end_position_on_the_genomic_accession=111385143, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27004, CurrentID=0, Status='live', Symbol='TCL6', Aliases='TNG1, TNG2', description='T-cell leukemia/lymphoma 6 (non-protein coding)', other_designations='TCL1-neighboring gene 1|TCL1-neighboring gene 2', map_location='14q32.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95651178, end_position_on_the_genomic_accession=95673452, orientation='plus', exon_count=8, OMIM=604412, no_hdr0=''),
  27005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27005, CurrentID=0, Status='live', Symbol='USP21', Aliases='USP16, USP23', description='ubiquitin specific peptidase 21', other_designations='NEDD8-specific protease|deubiquitinating enzyme 21|ubiquitin specific protease 21|ubiquitin thioesterase 21|ubiquitin thiolesterase 21|ubiquitin-specific processing protease 21|ubiquitin-specific protease 16|ubiquitin-specific-processing protease 21', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161159464, end_position_on_the_genomic_accession=161165726, orientation='plus', exon_count=14, OMIM=604729, no_hdr0=''),
  27006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27006, CurrentID=0, Status='live', Symbol='FGF22', Aliases='', description='fibroblast growth factor 22', other_designations='FGF-22', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=639895, end_position_on_the_genomic_accession=643703, orientation='plus', exon_count=3, OMIM=605831, no_hdr0=''),
  284593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284593, CurrentID=0, Status='live', Symbol='FAM41C', Aliases='', description='family with sequence similarity 41 member C', other_designations='Putative protein FAM41C|family with sequence similarity 41, member C', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=868071, end_position_on_the_genomic_accession=876802, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27010, CurrentID=0, Status='live', Symbol='TPK1', Aliases='HTPK1, PP20, THMD5', description='thiamin pyrophosphokinase 1', other_designations='placental protein 20|thiamine diphosphokinase|thiamine kinase', map_location='7q34-q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144451941, end_position_on_the_genomic_accession=144836411, orientation='minus', exon_count=27, OMIM=606370, no_hdr0=''),
  27012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27012, CurrentID=0, Status='live', Symbol='KCNV1', Aliases='HNKA, KCNB3, KV2.3, KV8.1', description='potassium channel, voltage gated modifier subfamily V, member 1', other_designations='neuronal potassium channel alpha subunit HNKA|potassium channel Kv8.1|potassium channel, subfamily V, member 1|voltage-gated potassium channel subunit Kv8.1', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109963636, end_position_on_the_genomic_accession=109975734, orientation='minus', exon_count=4, OMIM=608164, no_hdr0=''),
  27013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27013, CurrentID=0, Status='live', Symbol='CNPPD1', Aliases='C2orf24, CGI-57', description='cyclin Pas1/PHO80 domain containing 1', other_designations='cyclin Pas1/PHO80 domain-containing protein 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219171897, end_position_on_the_genomic_accession=219178164, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27018, CurrentID=0, Status='live', Symbol='NGFRAP1', Aliases='BEX3, Bex, DXS6984E, HGR74, NADE', description='nerve growth factor receptor (TNFRSF16) associated protein 1', other_designations='brain-expressed X-linked protein 3|ovarian granulosa cell 13.0 kDa protein HGR74|p75NTR-associated cell death executor', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103376323, end_position_on_the_genomic_accession=103378164, orientation='plus', exon_count=3, OMIM=300361, no_hdr0=''),
  27019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27019, CurrentID=0, Status='live', Symbol='DNAI1', Aliases='CILD1, DIC1, ICS1, PCD', description='dynein, axonemal, intermediate chain 1', other_designations='dynein, axonemal, intermediate polypeptide 1|immotile cilia syndrome 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34458752, end_position_on_the_genomic_accession=34520989, orientation='plus', exon_count=24, OMIM=604366, no_hdr0=''),
  27020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27020, CurrentID=0, Status='live', Symbol='NPTN', Aliases='GP55, GP65, SDFR1, SDR1, np55, np65', description='neuroplastin', other_designations='SDR-1|stromal cell derived factor receptor 1|stromal cell-derived receptor 1', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73560003, end_position_on_the_genomic_accession=73633412, orientation='minus', exon_count=12, OMIM=612820, no_hdr0=''),
  105377197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377197, CurrentID=0, Status='live', Symbol='LOC105377197', Aliases='', description='uncharacterized LOC105377197', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35314113, end_position_on_the_genomic_accession=35314890, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  27022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27022, CurrentID=0, Status='live', Symbol='FOXD3', Aliases='AIS1, Genesis, HFH2, VAMAS2', description='forkhead box D3', other_designations='HNF3/FH transcription factor genesis', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63323059, end_position_on_the_genomic_accession=63325126, orientation='plus', exon_count=1, OMIM=611539, no_hdr0=''),
  27023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27023, CurrentID=0, Status='live', Symbol='FOXB1', Aliases='FKH5, HFKH-5', description='forkhead box B1', other_designations='transcription factor FKH-5', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60004222, end_position_on_the_genomic_accession=60005943, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  27030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27030, CurrentID=0, Status='live', Symbol='MLH3', Aliases='HNPCC7', description='mutL homolog 3', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75013764, end_position_on_the_genomic_accession=75051532, orientation='minus', exon_count=14, OMIM=604395, no_hdr0=''),
  27031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27031, CurrentID=0, Status='live', Symbol='NPHP3', Aliases='CFAP31, MKS7, NPH3, RHPD, RHPD1, SLSN3', description='nephronophthisis 3 (adolescent)', other_designations='Meckel syndrome, type 7|cilia and flagella associated protein 31', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132680609, end_position_on_the_genomic_accession=132722459, orientation='minus', exon_count=27, OMIM=608002, no_hdr0=''),
  27032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27032, CurrentID=0, Status='live', Symbol='ATP2C1', Aliases='ATP2C1A, BCPM, HHD, PMR1, SPCA1, hSPCA1', description='ATPase, Ca++ transporting, type 2C, member 1', other_designations='ATP-dependent Ca(2+) pump PMR1|ATPase 2C1|ATPase, Ca(2+)-sequestering|HUSSY-28|secretory pathway Ca2+/Mn2+ ATPase 1', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130850483, end_position_on_the_genomic_accession=131016712, orientation='plus', exon_count=32, OMIM=604384, no_hdr0=''),
  27033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27033, CurrentID=0, Status='live', Symbol='ZBTB32', Aliases='FAXF, FAZF, Rog, TZFP, ZNF538', description='zinc finger and BTB domain containing 32', other_designations='FANCC-interacting protein|fanconi anemia zinc finger protein|repressor of GATA|testis zinc finger protein|zinc finger protein 538', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35704492, end_position_on_the_genomic_accession=35717048, orientation='plus', exon_count=8, OMIM=605859, no_hdr0=''),
  27034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27034, CurrentID=0, Status='live', Symbol='ACAD8', Aliases='ACAD-8, ARC42', description='acyl-CoA dehydrogenase family member 8', other_designations='activator-recruited cofactor 42 kDa component|acyl-CoA dehydrogenase family, member 8|acyl-Coenzyme A dehydrogenase family, member 8', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134253537, end_position_on_the_genomic_accession=134265944, orientation='plus', exon_count=12, OMIM=604773, no_hdr0=''),
  27035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27035, CurrentID=0, Status='live', Symbol='NOX1', Aliases='GP91-2, MOX1, NOH-1, NOH1', description='NADPH oxidase 1', other_designations='NADH/NADPH mitogenic oxidase subunit P65-MOX|NADPH oxidase homolog-1|mitogenic oxidase (pyridine nucleotide-dependent superoxide-generating)|mitogenic oxidase 1', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100843324, end_position_on_the_genomic_accession=100874345, orientation='minus', exon_count=13, OMIM=300225, no_hdr0=''),
  27036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27036, CurrentID=0, Status='live', Symbol='SIGLEC7', Aliases='AIRM1, CD328, CDw328, D-siglec, QA79, SIGLEC-7, SIGLEC19P, SIGLECP2, p75, p75/AIRM1', description='sialic acid binding Ig-like lectin 7', other_designations='AIRM-1|QA79 membrane protein|adhesion inhibitory receptor molecule 1, siglec-7|sialic acid binding immunoglobulin-like lectin 7', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51142294, end_position_on_the_genomic_accession=51153526, orientation='plus', exon_count=7, OMIM=604410, no_hdr0=''),
  27037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27037, CurrentID=0, Status='live', Symbol='TRMT2A', Aliases='HTF9C', description='tRNA methyltransferase 2 homolog A', other_designations='TRM2 tRNA methyltransferase 2 homolog A|hpaII tiny fragments locus 9c protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20111866, end_position_on_the_genomic_accession=20117295, orientation='minus', exon_count=13, OMIM=611151, no_hdr0=''),
  27038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27038, CurrentID=0, Status='live', Symbol='ARVD6', Aliases='', description='arrhythmogenic right ventricular dysplasia 6', other_designations='', map_location='10p14-p12', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604401, no_hdr0=''),
  27039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27039, CurrentID=0, Status='live', Symbol='PKD2L2', Aliases='TRPP5', description='polycystic kidney disease 2-like 2', other_designations='polycystin-2L2|polycystin-L2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137889211, end_position_on_the_genomic_accession=137942747, orientation='plus', exon_count=18, OMIM=604669, no_hdr0=''),
  27040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27040, CurrentID=0, Status='live', Symbol='LAT', Aliases='LAT1, pp36', description='linker for activation of T-cells', other_designations='36 kDa phospho-tyrosine adapter protein|36 kDa phospho-tyrosine adaptor protein|linker for activation of T cells|linker for activation of T cells, transmembrane adaptor|p36-38', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28984826, end_position_on_the_genomic_accession=28990783, orientation='plus', exon_count=12, OMIM=602354, no_hdr0=''),
  27042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27042, CurrentID=0, Status='live', Symbol='DIEXF', Aliases='C1orf107, DEF, DJ434O14.5, UTP25', description='digestive organ expansion factor homolog (zebrafish)', other_designations='digestive-organ expansion factor homolog', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209827967, end_position_on_the_genomic_accession=209857565, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  27043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27043, CurrentID=0, Status='live', Symbol='PELP1', Aliases='MNAR, P160', description='proline, glutamate and leucine rich protein 1', other_designations='modulator of non-genomic activity of estrogen receptor|proline and glutamic acid rich nuclear protein|transcription factor HMX3', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4671384, end_position_on_the_genomic_accession=4704337, orientation='minus', exon_count=19, OMIM=609455, no_hdr0=''),
  27044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27044, CurrentID=0, Status='live', Symbol='SND1', Aliases='TDRD11, p100', description='staphylococcal nuclease and tudor domain containing 1', other_designations='100 kDa coactivator|EBNA-2 co-activator (100kD)|EBNA2 coactivator p100|SND1-BRAF fusion|Tudor-SN|p100 EBNA2 co-activator|p100 co-activator|staphylococcal nuclease domain containing 1|tudor domain-containing protein 11', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127652119, end_position_on_the_genomic_accession=128092607, orientation='plus', exon_count=26, OMIM=602181, no_hdr0=''),
  550641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550641, CurrentID=0, Status='live', Symbol='TELM', Aliases='', description='telomere length, mean leukocyte', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609113, no_hdr0=''),
  100131725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131725, CurrentID=0, Status='live', Symbol='LOC100131725', Aliases='', description='solute carrier family 39 (zinc transporter), member 14 pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235614682, end_position_on_the_genomic_accession=235616396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060191, CurrentID=0, Status='live', Symbol='LOC101060191', Aliases='', description='E3 ubiquitin-protein ligase RING2-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52975030, end_position_on_the_genomic_accession=52976116, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101410546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410546, CurrentID=0, Status='live', Symbol='CST2P1', Aliases='', description='cystatin SA pseudogene 1', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23711058, end_position_on_the_genomic_accession=23711399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158131, CurrentID=0, Status='live', Symbol='OR1Q1', Aliases='HSTPCR106, OR1Q2, OR1Q3, OR9-25, OR9-A, OST226, OST226OR9-A, TPCR106', description='olfactory receptor family 1 subfamily Q member 1', other_designations='olfactory receptor 1Q2|olfactory receptor 1Q3|olfactory receptor 9-A|olfactory receptor OR9-25|olfactory receptor TPCR106|olfactory receptor, family 1, subfamily Q, member 1|olfactory receptor, family 1, subfamily Q, member 2|olfactory receptor, family 1, subfamily Q, member 3', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122568380, end_position_on_the_genomic_accession=122615830, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  27063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27063, CurrentID=0, Status='live', Symbol='ANKRD1', Aliases='ALRP, C-193, CARP, CVARP, MCARP, bA320F15.2', description='ankyrin repeat domain 1 (cardiac muscle)', other_designations='cardiac ankyrin repeat protein|cytokine-inducible gene C-193 protein|cytokine-inducible nuclear protein|liver ankyrin repeat domain 1', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90912100, end_position_on_the_genomic_accession=90921275, orientation='minus', exon_count=9, OMIM=609599, no_hdr0=''),
  27065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27065, CurrentID=0, Status='live', Symbol='NSG1', Aliases='D4S234, D4S234E, NEEP21, P21', description='neuron specific gene family member 1', other_designations='brain neuron cytoplasmic protein 1|carboxyterminally EE-tagged neuron-enriched endosomal 21 kDa protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4386256, end_position_on_the_genomic_accession=4419058, orientation='plus', exon_count=6, OMIM=607645, no_hdr0=''),
  27067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27067, CurrentID=0, Status='live', Symbol='STAU2', Aliases='39K2, 39K3', description='staufen double-stranded RNA binding protein 2', other_designations='staufen homolog 2|staufen, RNA binding protein, homolog 2', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73420369, end_position_on_the_genomic_accession=73747708, orientation='minus', exon_count=17, OMIM=605920, no_hdr0=''),
  27068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27068, CurrentID=0, Status='live', Symbol='PPA2', Aliases='HSPC124, SID6-306', description='pyrophosphatase (inorganic) 2', other_designations='PPase 2|pyrophosphatase SID6-306|pyrophosphate phospho-hydrolase 2', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105369077, end_position_on_the_genomic_accession=105474070, orientation='minus', exon_count=12, OMIM=609988, no_hdr0=''),
  27069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27069, CurrentID=0, Status='live', Symbol='GHITM', Aliases='DERP2, HSPC282, MICS1, My021, PTD010, TMBIM5', description='growth hormone inducible transmembrane protein', other_designations='dermal papilla-derived protein 2|mitochondrial morphology and cristae structure 1|transmembrane BAX inhibitor motif containing 5|transmembrane BAX inhibitor motif-containing protein 5', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84139429, end_position_on_the_genomic_accession=84153555, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  27071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27071, CurrentID=0, Status='live', Symbol='DAPP1', Aliases='BAM32', description='dual adaptor of phosphotyrosine and 3-phosphoinositides', other_designations='B-cell adapter molecule of 32 kDa|b lymphocyte adapter protein Bam32|hDAPP1', map_location='4q25-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99816824, end_position_on_the_genomic_accession=99872249, orientation='plus', exon_count=13, OMIM=605768, no_hdr0=''),
  27072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27072, CurrentID=0, Status='live', Symbol='VPS41', Aliases='HVPS41, HVSP41, hVps41p', description='VPS41 HOPS complex subunit', other_designations='S53|vacuolar assembly protein 41|vacuolar protein sorting 41 homolog', map_location='7p14-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38723943, end_position_on_the_genomic_accession=38909200, orientation='minus', exon_count=30, OMIM=605485, no_hdr0=''),
  266656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266656, CurrentID=0, Status='live', Symbol='TAS2R12P', Aliases='PS10, T2R12, TAS2R12, TAS2R26', description='taste 2 receptor member 12 pseudogene', other_designations='taste receptor Ps10 pseudogene|taste receptor, type 2, member 12 pseudogene|taste receptor, type 2, member 12, pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10894934, end_position_on_the_genomic_accession=10895881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27074, CurrentID=0, Status='live', Symbol='LAMP3', Aliases='CD208, DC LAMP, DC-LAMP, DCLAMP, LAMP, LAMP-3, TSC403', description='lysosomal-associated membrane protein 3', other_designations='DC-lysosome-associated membrane glycoprotein', map_location='3q26.3-q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183122215, end_position_on_the_genomic_accession=183162879, orientation='minus', exon_count=10, OMIM=605883, no_hdr0=''),
  27075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27075, CurrentID=0, Status='live', Symbol='TSPAN13', Aliases='NET-6, NET6, TM4SF13', description='tetraspanin 13', other_designations='tetraspan NET-6|transmembrane 4 superfamily member 13|transmembrane 4 superfamily member tetraspan NET-6|tspan-13', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16753726, end_position_on_the_genomic_accession=16784536, orientation='plus', exon_count=6, OMIM=613139, no_hdr0=''),
  27076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27076, CurrentID=0, Status='live', Symbol='LYPD3', Aliases='C4.4A', description='LY6/PLAUR domain containing 3', other_designations='2310061G07Rik|GPI-anchored metastasis-associated protein C4.4A homolog|GPI-anchored metastasis-associated protein homolog|MIG-C4|matrigel-induced gene C4 protein', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43460794, end_position_on_the_genomic_accession=43465679, orientation='minus', exon_count=5, OMIM=609484, no_hdr0=''),
  27077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27077, CurrentID=0, Status='live', Symbol='B9D1', Aliases='B9, EPPB9, MKS9, MKSR1', description='B9 protein domain 1', other_designations='MKS1-related protein 1|endothelial precursor protein B9', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19335263, end_position_on_the_genomic_accession=19378182, orientation='minus', exon_count=11, OMIM=614144, no_hdr0=''),
  27079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27079, CurrentID=0, Status='live', Symbol='RPUSD2', Aliases='C15orf19, C18B11', description='RNA pseudouridylate synthase domain containing 2', other_designations='C18B11 homolog (44.9kD)', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40569293, end_position_on_the_genomic_accession=40574706, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  27085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27085, CurrentID=0, Status='live', Symbol='MTBP', Aliases='MDM2BP', description='MDM2 binding protein', other_designations='MDM2 (mouse double minute 2)-binding protein, 104kD|Mdm2, transformed 3T3 cell double minute 2, p53 binding protein binding protein, 104kDa', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120445398, end_position_on_the_genomic_accession=120523635, orientation='plus', exon_count=25, OMIM=605927, no_hdr0=''),
  27086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27086, CurrentID=0, Status='live', Symbol='FOXP1', Aliases='12CC4, HSPC215, MFH, QRF1, hFKH1B', description='forkhead box P1', other_designations='fork head-related protein like B|glutamine-rich factor 1|mac-1-regulated forkhead', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70954714, end_position_on_the_genomic_accession=71583989, orientation='minus', exon_count=36, OMIM=605515, no_hdr0=''),
  27087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27087, CurrentID=0, Status='live', Symbol='B3GAT1', Aliases='CD57, GLCATP, GLCUATP, HNK1, LEU7, NK-1, NK1', description='beta-1,3-glucuronyltransferase 1', other_designations='LEU7 antigen|UDP-GlcUA:glycoprotein beta-1,3-glucuronyltransferase|glcUAT-P|glucuronosyltransferase P', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134378504, end_position_on_the_genomic_accession=134411987, orientation='minus', exon_count=11, OMIM=151290, no_hdr0=''),
  158160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158160, CurrentID=0, Status='live', Symbol='HSD17B7P2', Aliases='HSD17B7, Hsd17b_2, bA291L22.1', description='hydroxysteroid (17-beta) dehydrogenase 7 pseudogene 2', other_designations='17beta-hydroxysteroid dehydrogenase type 7 form 2', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38356380, end_position_on_the_genomic_accession=38378505, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  27089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27089, CurrentID=0, Status='live', Symbol='UQCRQ', Aliases='MC3DN4, QCR8, QP-C, QPC, UQCR7', description='ubiquinol-cytochrome c reductase complex III subunit VII', other_designations='complex III subunit 8|complex III subunit VIII|low molecular mass ubiquinone-binding protein (9.5kD)|ubiquinol-cytochrome c reductase complex 9.5 kDa protein|ubiquinol-cytochrome c reductase complex ubiquinone-binding protein QP-C|ubiquinol-cytochrome c reductase, complex III subunit VII, 9.5kDa', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132866627, end_position_on_the_genomic_accession=132868844, orientation='plus', exon_count=3, OMIM=612080, no_hdr0=''),
  27090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27090, CurrentID=0, Status='live', Symbol='ST6GALNAC4', Aliases='IV, SIAT3-C, SIAT3C, SIAT7-D, SIAT7D, ST6GALNACIV, ST6GalNAc', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 4', other_designations='NeuAc-alpha-2,3-Gal-beta-1,3-GalNAc-alpha-2, 6-sialyltransferase alpha2,6-sialyltransferase|NeuAc-alpha-2,3-Gal-beta-1,3-GalNAc-alpha-2,6-sialyltransferase IV|ST6 neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminidealpha-2,6-sialyltransferase 4|ST6GalNAc IV|sialyltransferase 3C|sialyltransferase 7D ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase)', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127907886, end_position_on_the_genomic_accession=127917051, orientation='minus', exon_count=8, OMIM=606378, no_hdr0=''),
  27091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27091, CurrentID=0, Status='live', Symbol='CACNG5', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 5', other_designations='TARP gamma-5|neuronal voltage-gated calcium channel gamma-5 subunit|transmembrane AMPAR regulatory protein gamma-5', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66835117, end_position_on_the_genomic_accession=66887296, orientation='plus', exon_count=6, OMIM=606405, no_hdr0=''),
  27092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27092, CurrentID=0, Status='live', Symbol='CACNG4', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 4', other_designations='TARP gamma-4|neuronal voltage-gated calcium channel gamma-4 subunit|transmembrane AMPAR regulatory protein gamma-4', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66964864, end_position_on_the_genomic_accession=67033402, orientation='plus', exon_count=4, OMIM=606404, no_hdr0=''),
  27093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27093, CurrentID=0, Status='live', Symbol='KCNMB3P1', Aliases='KCNMB2L, KCNMB3L, KCNMB3L1, KCNMBLP', description='potassium channel subfamily M regulatory beta subunit 3 pseudogene 1', other_designations='potassium large conductance calcium-activated channel, subfamily M, beta member 3 pseudogene 1', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16580766, end_position_on_the_genomic_accession=16582707, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27094, CurrentID=0, Status='live', Symbol='KCNMB3', Aliases='BKBETA3, HBETA3, K(VCA)BETA-3, KCNMB2, KCNMBL, SLO-BETA-3, SLOBETA3', description='potassium channel subfamily M regulatory beta subunit 3', other_designations='BK channel beta subunit 3|BK channel subunit beta-3|MaxiK channel beta-subunit 3|big potassium channel beta subunit 3|calcium-activated potassium channel regulatory subunit|calcium-activated potassium channel, subfamily M subunit beta-3|charybdotoxin receptor subunit beta-3|large conductance, voltage and Ca2+ activated potassium channel Maxi K beta 3 subunit|maxi K channel subunit beta-3|potassium large conductance calcium-activated channel, subfamily M beta member 3', map_location='3q26.3-q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179239749, end_position_on_the_genomic_accession=179267050, orientation='minus', exon_count=7, OMIM=605222, no_hdr0=''),
  27095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27095, CurrentID=0, Status='live', Symbol='TRAPPC3', Aliases='BET3', description='trafficking protein particle complex 3', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36136569, end_position_on_the_genomic_accession=36156053, orientation='minus', exon_count=6, OMIM=610955, no_hdr0=''),
  105377880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377880, CurrentID=0, Status='live', Symbol='LINC01611', Aliases='TCONS_l2_00025430', description='long intergenic non-protein coding RNA 1611', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84421043, end_position_on_the_genomic_accession=84458249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27097, CurrentID=0, Status='live', Symbol='TAF5L', Aliases='PAF65B', description='TATA-box binding protein associated factor 5 like', other_designations='PAF65-beta|PCAF associated factor 65 beta|PCAF-associated factor 65 beta|TAF5-like RNA polymerase II, p300/CBP-associated factor (PCAF)-associated factor, 65kDa|TATA box binding protein associated factor 5 like', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229593111, end_position_on_the_genomic_accession=229626047, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27098, CurrentID=0, Status='live', Symbol='CLUL1', Aliases='RA337M', description='clusterin like 1', other_designations='clusterin-like 1 (retinal)|retinal clusterin-like protein|retinal-specific clusterin-like protein', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=596998, end_position_on_the_genomic_accession=650298, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  27099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27099, CurrentID=0, Status='live', Symbol='SND1-IT1', Aliases='C7orf54, NAG8, NSG-X', description='SND1 intronic transcript 1', other_designations='SND1 intronic transcript 1 (non-protein coding)|brain and nasopharyngeal carcinoma susceptibility|nasopharyngeal carcinoma associated gene 8', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127997509, end_position_on_the_genomic_accession=128000077, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  158172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158172, CurrentID=0, Status='live', Symbol='SERPINH1P1', Aliases='SERPINHP1, bA571F15.6, pshsp47', description='serpin peptidase inhibitor, clade H1, pseudogene 1', other_designations='CBP2 (human hsp47) pseudogene|serine (or cysteine) proteinase inhibitor, clade H (heat shock protein 47), member 2 pseudogene|serine (or cysteine) proteinase inhibitor, clade H, pseudogene 1|serpin peptidase inhibitor, clade H (heat shock protein 47), member 1, (collagen binding protein 1) pseudogene|serpin peptidase inhibitor, clade H, pseudogene 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34318266, end_position_on_the_genomic_accession=34320331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27101, CurrentID=0, Status='live', Symbol='CACYBP', Aliases='GIG5, PNAS-107, S100A6BP, SIP', description='calcyclin binding protein', other_designations='S100A6-binding protein|Siah-interacting protein (SIP)|growth-inhibiting gene 5 protein|hCacyBP', map_location='1q24-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174999435, end_position_on_the_genomic_accession=175012027, orientation='plus', exon_count=7, OMIM=606186, no_hdr0=''),
  27102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27102, CurrentID=0, Status='live', Symbol='EIF2AK1', Aliases='HCR, HRI', description='eukaryotic translation initiation factor 2 alpha kinase 1', other_designations='heme regulated initiation factor 2 alpha kinase|heme sensitive initiation factor 2a kinase|heme-controlled repressor|heme-regulated eukaryotic initiation factor eIF-2-alpha kinase|heme-regulated inhibitor|heme-regulated initiation factor 2-alpha kinase|heme-regulated repressor|hemin-sensitive initiation factor 2-alpha kinase', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6022247, end_position_on_the_genomic_accession=6059229, orientation='minus', exon_count=15, OMIM=613635, no_hdr0=''),
  27106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27106, CurrentID=0, Status='live', Symbol='ARRDC2', Aliases='CLONE24945, PP2703', description='arrestin domain containing 2', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18001132, end_position_on_the_genomic_accession=18014102, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  27107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27107, CurrentID=0, Status='live', Symbol='ZBTB11', Aliases='ZNF-U69274, ZNF913', description='zinc finger and BTB domain containing 11', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101649439, end_position_on_the_genomic_accession=101677144, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  27109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27109, CurrentID=0, Status='live', Symbol='ATP5S', Aliases='ATPW, FB, HSU79253', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit s (factor B)', other_designations='ATP synthase coupling factor B, mitochondrial|ATP synthase coupling factor B-like 1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit s (factor B)|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit s (factor B)|ATP synthase-coupling factor B|mitochondrial ATP synthase regulatory component factor B', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50312329, end_position_on_the_genomic_accession=50326715, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27111, CurrentID=0, Status='live', Symbol='SDCBP2', Aliases='SITAC, SITAC18, ST-2', description='syndecan binding protein 2', other_designations='syndecan binding protein (syntenin) 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1309909, end_position_on_the_genomic_accession=1329235, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27112, CurrentID=0, Status='live', Symbol='FAM155B', Aliases='CXorf63, TED, TMEM28, bB57D9.1', description='family with sequence similarity 155 member B', other_designations='bB57D9.1 (TED protein)|family with sequence similarity 155, member B|transmembrane protein 28', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69505201, end_position_on_the_genomic_accession=69532508, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27113, CurrentID=0, Status='live', Symbol='BBC3', Aliases='JFY-1, JFY1, PUMA', description='BCL2 binding component 3', other_designations='p53 up-regulated modulator of apoptosis', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47220822, end_position_on_the_genomic_accession=47232996, orientation='minus', exon_count=7, OMIM=605854, no_hdr0=''),
  27115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27115, CurrentID=0, Status='live', Symbol='PDE7B', Aliases='bA472E5.1', description='phosphodiesterase 7B', other_designations="high-affinity cAMP-specific 3',5'-cyclic phosphodiesterase|rolipram-insensitive phosphodiesterase type 7", map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135808752, end_position_on_the_genomic_accession=136195571, orientation='plus', exon_count=16, OMIM=604645, no_hdr0=''),
  100271392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271392, CurrentID=0, Status='live', Symbol='RPL31P63', Aliases='RPL31_36_1782', description='ribosomal protein L31 pseudogene 63', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68600072, end_position_on_the_genomic_accession=68600517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27120, CurrentID=0, Status='live', Symbol='DKKL1', Aliases='CT34, SGY, SGY-1, SGY1', description='dickkopf-like 1', other_designations='cancer/testis antigen 34|protein soggy-1|soggy|soggy-1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49360495, end_position_on_the_genomic_accession=49375119, orientation='plus', exon_count=10, OMIM=605418, no_hdr0=''),
  27121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27121, CurrentID=0, Status='live', Symbol='DKK4', Aliases='DKK-4', description='dickkopf WNT signaling pathway inhibitor 4', other_designations='dickkopf-4|hDkk-4', map_location='8p11.2-p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42373194, end_position_on_the_genomic_accession=42391754, orientation='minus', exon_count=7, OMIM=605417, no_hdr0=''),
  27122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27122, CurrentID=0, Status='live', Symbol='DKK3', Aliases='REIC, RIG', description='dickkopf WNT signaling pathway inhibitor 3', other_designations='RIG-like 5-6|RIG-like 7-1|dickkopf 3 homolog|dickkopf homolog 3|dickkopf-3|dkk-3|hDkk-3|regulated in glioma', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11962996, end_position_on_the_genomic_accession=12009370, orientation='minus', exon_count=9, OMIM=605416, no_hdr0=''),
  27123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27123, CurrentID=0, Status='live', Symbol='DKK2', Aliases='DKK-2', description='dickkopf WNT signaling pathway inhibitor 2', other_designations='dickkopf 2 homolog|dickkopf homolog 2|dickkopf related protein-2|dickkopf-2|hDkk-2', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106921802, end_position_on_the_genomic_accession=107036296, orientation='minus', exon_count=4, OMIM=605415, no_hdr0=''),
  27124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27124, CurrentID=0, Status='live', Symbol='INPP5J', Aliases='INPP5, PIB5PA, PIPP', description='inositol polyphosphate-5-phosphatase J', other_designations='inositol polyphosphate 5-phosphatase J|phosphatidylinositol (4,5) bisphosphate 5-phosphatase, A', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31107736, end_position_on_the_genomic_accession=31134697, orientation='plus', exon_count=17, OMIM=606481, no_hdr0=''),
  27125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27125, CurrentID=0, Status='live', Symbol='AFF4', Aliases='AF5Q31, CHOPS, MCEF', description='AF4/FMR2 family member 4', other_designations='AF4/FMR2 family, member 4|ALL1-fused gene from chromosome 5q31 protein|major CDK9 elongation factor-associated protein', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132875379, end_position_on_the_genomic_accession=132963662, orientation='minus', exon_count=23, OMIM=604417, no_hdr0=''),
  27127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27127, CurrentID=0, Status='live', Symbol='SMC1B', Aliases='SMC1BETA, SMC1L2', description='structural maintenance of chromosomes 1B', other_designations='SMC protein 1B|SMC1 (structural maintenance of chromosomes 1, yeast)-like 1|SMC1 structural maintenance of chromosomes 1-like 2|mitosis-specific chromosome segregation protein like protein beta', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45344063, end_position_on_the_genomic_accession=45413629, orientation='minus', exon_count=26, OMIM=608685, no_hdr0=''),
  27128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27128, CurrentID=0, Status='live', Symbol='CYTH4', Aliases='CYT4, DJ63G5.1, PSCD4', description='cytohesin 4', other_designations='PH, SEC7 and coiled-coil domain-containing protein 4|pleckstrin homology, Sec7 and coiled-coil domains 4|pleckstrin homology, Sec7 and coiled/coil domains 4', map_location='22q12.3-q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37282454, end_position_on_the_genomic_accession=37315348, orientation='plus', exon_count=14, OMIM=606514, no_hdr0=''),
  27129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27129, CurrentID=0, Status='live', Symbol='HSPB7', Aliases='cvHSP', description='heat shock protein family B (small) member 7', other_designations='cardiovascular heat shock protein|heat shock 27kD protein family, member 7 (cardiovascular)|heat shock 27kDa protein family member 7 (cardiovascular)|heat shock 27kDa protein family, member 7 (cardiovascular)', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16014028, end_position_on_the_genomic_accession=16018796, orientation='minus', exon_count=3, OMIM=610692, no_hdr0=''),
  27130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27130, CurrentID=0, Status='live', Symbol='INVS', Aliases='INV, NPH2, NPHP2', description='inversin', other_designations='inversion of embryo turning homolog|inversion of embryonic turning|nephrocystin-2', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100099185, end_position_on_the_genomic_accession=100338617, orientation='plus', exon_count=25, OMIM=243305, no_hdr0=''),
  27131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27131, CurrentID=0, Status='live', Symbol='SNX5', Aliases='', description='sorting nexin 5', other_designations='', map_location='20p11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17941596, end_position_on_the_genomic_accession=17968991, orientation='minus', exon_count=14, OMIM=605937, no_hdr0=''),
  27132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27132, CurrentID=0, Status='live', Symbol='CPNE7', Aliases='', description='copine VII', other_designations='copine 7', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89575768, end_position_on_the_genomic_accession=89597246, orientation='plus', exon_count=19, OMIM=605689, no_hdr0=''),
  27133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27133, CurrentID=0, Status='live', Symbol='KCNH5', Aliases='EAG2, H-EAG2, Kv10.2, hEAG2', description='potassium channel, voltage gated eag related subfamily H, member 5', other_designations='ether-a-go-go-related potassium channel 2|potassium voltage-gated channel, subfamily H (eag-related), member 5|voltage-gated potassium channel subunit Kv10.2', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62706573, end_position_on_the_genomic_accession=63045256, orientation='minus', exon_count=11, OMIM=605716, no_hdr0=''),
  27134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27134, CurrentID=0, Status='live', Symbol='TJP3', Aliases='ZO-3, ZO3', description='tight junction protein 3', other_designations='tight junction protein 3 (zona occludens 3)|zona occludens protein 3|zonula occludens protein 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3708337, end_position_on_the_genomic_accession=3750813, orientation='plus', exon_count=22, OMIM=612689, no_hdr0=''),
  27136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27136, CurrentID=0, Status='live', Symbol='MORC1', Aliases='CT33, MORC, ZCW6', description='MORC family CW-type zinc finger 1', other_designations='cancer/testis antigen 33|microrchidia, mouse, homolog of', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108958239, end_position_on_the_genomic_accession=109118146, orientation='minus', exon_count=31, OMIM=603205, no_hdr0=''),
  4523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4523, CurrentID=0, Status='live', Symbol='MTHFD1P1', Aliases='MTHFDL1, MTHFDP1', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1 pseudogene 1', other_designations='5,10-methylenetetrahydrofolate dehydrogenase, 5,10-methylenetetrahydrofolate cyclohydrolase, 10-formyltetrahydrofolate synthetas|methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1, methenyltetrahydrofolate cyclohydrolase, formyltetrahydrofolate synthetase pseudogene|methylenetetrahydrofolate dehydrogenase (NADP+ dependent), methenyltetrahydrofolate cyclohydrolase, formyltetrahydrofolate synthetase pseudogene 1', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57392393, end_position_on_the_genomic_accession=57395452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27141, CurrentID=0, Status='live', Symbol='CIDEB', Aliases='', description='cell death-inducing DFFA-like effector b', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24305187, end_position_on_the_genomic_accession=24311377, orientation='minus', exon_count=8, OMIM=604441, no_hdr0=''),
  27143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27143, CurrentID=0, Status='live', Symbol='PALD1', Aliases='KIAA1274, PALD', description='phosphatase domain containing, paladin 1', other_designations='palladin', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70458545, end_position_on_the_genomic_accession=70568450, orientation='plus', exon_count=24, OMIM=614656, no_hdr0=''),
  27145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27145, CurrentID=0, Status='live', Symbol='FILIP1', Aliases='FILIP', description='filamin A interacting protein 1', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75294948, end_position_on_the_genomic_accession=75493829, orientation='minus', exon_count=10, OMIM=607307, no_hdr0=''),
  27146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27146, CurrentID=0, Status='live', Symbol='FAM184B', Aliases='', description='family with sequence similarity 184 member B', other_designations='family with sequence similarity 184, member B', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17632073, end_position_on_the_genomic_accession=17781527, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  158219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158219, CurrentID=0, Status='live', Symbol='TTC39B', Aliases='C9orf52', description='tetratricopeptide repeat domain 39B', other_designations='TPR repeat protein 39B', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15170844, end_position_on_the_genomic_accession=15307360, orientation='minus', exon_count=23, OMIM=613574, no_hdr0=''),
  27148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27148, CurrentID=0, Status='live', Symbol='STK36', Aliases='FU', description='serine/threonine kinase 36', other_designations='fused homolog', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218672026, end_position_on_the_genomic_accession=218702717, orientation='plus', exon_count=29, OMIM=607652, no_hdr0=''),
  158222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158222, CurrentID=0, Status='live', Symbol='LDHAP4', Aliases='LDHAL4', description='lactate dehydrogenase A pseudogene 4', other_designations='lactate dehydrogenase A-like 4|lactate dehydrogenase pseudogene', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14920689, end_position_on_the_genomic_accession=14922433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27151, CurrentID=0, Status='live', Symbol='CPAMD8', Aliases='K-CAP, VIP', description='C3 and PZP-like, alpha-2-macroglobulin domain containing 8', other_designations='alpha-2 macroglobulin family protein VIP', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16892951, end_position_on_the_genomic_accession=17026818, orientation='minus', exon_count=43, OMIM=608841, no_hdr0=''),
  27152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27152, CurrentID=0, Status='live', Symbol='INTU', Aliases='INT, PDZD6, PDZK6', description='inturned planar cell polarity protein', other_designations='PDZ domain containing 6|PDZ domain-containing protein 6|homolog of inturned|inturned planar cell polarity effector homolog', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127623271, end_position_on_the_genomic_accession=127716779, orientation='plus', exon_count=20, OMIM=610621, no_hdr0=''),
  27153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27153, CurrentID=0, Status='live', Symbol='ZNF777', Aliases='', description='zinc finger protein 777', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149431363, end_position_on_the_genomic_accession=149460962, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  27154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27154, CurrentID=0, Status='live', Symbol='BRPF3', Aliases='', description='bromodomain and PHD finger containing 3', other_designations='bromodomain and PHD finger containing, 3', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36196773, end_position_on_the_genomic_accession=36232790, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27156, CurrentID=0, Status='live', Symbol='RSPH14', Aliases='RTDR1', description='radial spoke head 14 homolog (Chlamydomonas)', other_designations='rhabdoid tumor deletion region gene 1|rhabdoid tumor deletion region protein 1', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23059408, end_position_on_the_genomic_accession=23145230, orientation='minus', exon_count=13, OMIM=605663, no_hdr0=''),
  27158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27158, CurrentID=0, Status='live', Symbol='NDOR1', Aliases='NR1, bA350O14.9', description='NADPH dependent diflavin oxidoreductase 1', other_designations='NADPH-dependent FMN and FAD-containing oxidoreductase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137205667, end_position_on_the_genomic_accession=137219361, orientation='plus', exon_count=15, OMIM=606073, no_hdr0=''),
  27159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27159, CurrentID=0, Status='live', Symbol='CHIA', Aliases='AMCASE, CHIT2, TSA1902', description='chitinase, acidic', other_designations='lung-specific protein TSA1902', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111290852, end_position_on_the_genomic_accession=111320566, orientation='plus', exon_count=13, OMIM=606080, no_hdr0=''),
  27160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27160, CurrentID=0, Status='live', Symbol='INGX', Aliases='ING1-like, ING2', description='inhibitor of growth family, X-linked, pseudogene', other_designations='ING1-like tumor suppressor protein|inhibitor of growth family, member 2', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71491681, end_position_on_the_genomic_accession=71492449, orientation='minus', exon_count=1, OMIM=300452, no_hdr0=''),
  27161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27161, CurrentID=0, Status='live', Symbol='AGO2', Aliases='EIF2C2, Q10', description='argonaute 2, RISC catalytic component', other_designations='CTA-204B4.6|PAZ Piwi domain protein|PPD|argonaute 2|argonaute RISC catalytic component 2|argonaute2|eIF-2C 2|eIF2C 2|eukaryotic translation initiation factor 2C, 2|hAgo2|protein slicer', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140520156, end_position_on_the_genomic_accession=140642326, orientation='minus', exon_count=24, OMIM=606229, no_hdr0=''),
  158234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158234, CurrentID=0, Status='live', Symbol='TRMT10B', Aliases='RG9MTD3, bA3J10.9', description='tRNA methyltransferase 10B', other_designations='RNA (guanine-9-) methyltransferase domain containing 3|RNA (guanine-9-)-methyltransferase domain-containing protein 3', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37753204, end_position_on_the_genomic_accession=37778972, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  27163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27163, CurrentID=0, Status='live', Symbol='NAAA', Aliases='ASAHL, PLT', description='N-acylethanolamine acid amidase', other_designations='ASAH-like protein|acid ceramidase-like protein', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75910655, end_position_on_the_genomic_accession=75941013, orientation='minus', exon_count=15, OMIM=607469, no_hdr0=''),
  27164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27164, CurrentID=0, Status='live', Symbol='SALL3', Aliases='ZNF796', description='spalt-like transcription factor 3', other_designations='C2H2 zinc finger protein SALL3|hSALL3|spalt-like zinc finger protein|zinc finger protein 796', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78980275, end_position_on_the_genomic_accession=78998969, orientation='plus', exon_count=3, OMIM=605079, no_hdr0=''),
  27165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27165, CurrentID=0, Status='live', Symbol='GLS2', Aliases='GA, GLS, LGA, hLGA', description='glutaminase 2', other_designations='L-glutamine amidohydrolase|breast cell glutaminase|glutaminase 2 (liver, mitochondrial)|glutaminase I|phosphate-activated glutaminase|phosphate-dependent glutaminase', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56470944, end_position_on_the_genomic_accession=56488414, orientation='minus', exon_count=19, OMIM=606365, no_hdr0=''),
  27166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27166, CurrentID=0, Status='live', Symbol='PRELID1', Aliases='CGI-106, PRELI, PX19, SBBI12', description='PRELI domain containing 1', other_designations='25 kDa protein of relevant evolutionary and lymphoid interest|px19-like protein', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177303762, end_position_on_the_genomic_accession=177306959, orientation='plus', exon_count=5, OMIM=605733, no_hdr0=''),
  27170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27170, CurrentID=0, Status='live', Symbol='COTL1P1', Aliases='CLPSMCR', description='coactosin-like F-actin binding protein 1 pseudogene 1', other_designations='COTL1 pseudogene 1|coactosin-like 1 pseudogene 1|coactosin-like, Smith Magenis syndrome chromosome region', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16853862, end_position_on_the_genomic_accession=16856116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373518, CurrentID=0, Status='live', Symbol='LOC105373518', Aliases='', description='uncharacterized LOC105373518', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102873283, end_position_on_the_genomic_accession=102895751, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27173, CurrentID=0, Status='live', Symbol='SLC39A1', Aliases='ZIP1, ZIRTL', description='solute carrier family 39 (zinc transporter), member 1', other_designations='solute carrier family 39 (zinc transporter), member 3|zrt- and Irt-like protein 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153959099, end_position_on_the_genomic_accession=153968184, orientation='minus', exon_count=7, OMIM=604740, no_hdr0=''),
  27175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27175, CurrentID=0, Status='live', Symbol='TUBG2', Aliases='', description='tubulin gamma 2', other_designations='gamma-2-tubulin|tubulin, gamma 2', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42659248, end_position_on_the_genomic_accession=42667006, orientation='plus', exon_count=12, OMIM=605785, no_hdr0=''),
  158248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158248, CurrentID=0, Status='live', Symbol='TTC16', Aliases='', description='tetratricopeptide repeat domain 16', other_designations='TPR repeat protein 16', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127716064, end_position_on_the_genomic_accession=127731600, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  27177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27177, CurrentID=0, Status='live', Symbol='IL36B', Aliases='FIL1, FIL1-(ETA), FIL1H, FILI-(ETA), IL-1F8, IL-1H2, IL1-ETA, IL1F8, IL1H2', description='interleukin 36, beta', other_designations='FIL1 eta|IL-1 eta|IL-1F8 (FIL1-eta)|IL1F8 (Canonical product IL-1F8a)|Interleukin-1 Superfamily e|family of interleukin 1-eta|interleukin 1 family, member 8 (eta)|interleukin 1, eta|interleukin-1 eta|interleukin-1 family member 8|interleukin-1 homolog 2', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113022091, end_position_on_the_genomic_accession=113052867, orientation='minus', exon_count=8, OMIM=605508, no_hdr0=''),
  27178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27178, CurrentID=0, Status='live', Symbol='IL37', Aliases='FIL1, FIL1(ZETA), FIL1Z, IL-1F7, IL-1H, IL-1H4, IL-1RP1, IL-37, IL1F7, IL1H4, IL1RP1', description='interleukin 37', other_designations='FIL1 zeta|IL-1 zeta|IL-1F7b (IL-1H4, IL-1H, IL-1RP1)|IL-1X protein|IL1F7 (canonical product IL-1F7b)|interleukin 1 family member 7|interleukin 1, zeta|interleukin-1 homolog 4|interleukin-1 superfamily z|interleukin-1-related protein|interleukin-23', map_location='2q12-q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112911064, end_position_on_the_genomic_accession=112918882, orientation='plus', exon_count=7, OMIM=605510, no_hdr0=''),
  27179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27179, CurrentID=0, Status='live', Symbol='IL36A', Aliases='FIL1, FIL1(EPSILON), FIL1E, IL-1F6, IL1(EPSILON), IL1F6', description='interleukin 36, alpha', other_designations='FIL1 epsilon|IL-1 epsilon|IL-1F6 (FIL-1-epsilon)|interleukin 1 family, member 6 (epsilon)|interleukin 1, epsilon|interleukin-1 epsilon|interleukin-1 family member 6', map_location='2q12-q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113005461, end_position_on_the_genomic_accession=113011071, orientation='plus', exon_count=6, OMIM=605509, no_hdr0=''),
  27180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27180, CurrentID=0, Status='live', Symbol='SIGLEC9', Aliases='CD329, CDw329, FOAP-9, OBBP-LIKE, siglec-9', description='sialic acid binding Ig-like lectin 9', other_designations='protein FOAP-9', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51124880, end_position_on_the_genomic_accession=51141020, orientation='plus', exon_count=13, OMIM=605640, no_hdr0=''),
  27181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27181, CurrentID=0, Status='live', Symbol='SIGLEC8', Aliases='SAF2, SIGLEC-8L, SIGLEC8', description='sialic acid binding Ig-like lectin 8', other_designations='CDw329|SAF-2|sialoadhesin family member 2', map_location='19q13.33-q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51450997, end_position_on_the_genomic_accession=51458454, orientation='minus', exon_count=7, OMIM=605639, no_hdr0=''),
  27183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27183, CurrentID=0, Status='live', Symbol='VPS4A', Aliases='SKD1, SKD1A, SKD2, VPS4, VPS4-1', description='vacuolar protein sorting 4 homolog A (S. cerevisiae)', other_designations='SKD1-homolog|hVPS4|vacuolar protein sorting factor 4A|vacuolar sorting protein 4', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69311384, end_position_on_the_genomic_accession=69325043, orientation='plus', exon_count=11, OMIM=609982, no_hdr0=''),
  27184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27184, CurrentID=0, Status='live', Symbol='DISC2', Aliases='DISC1-AS1, DISC1OS, NCRNA00015', description='disrupted in schizophrenia 2 (non-protein coding)', other_designations='DISC1 antisense RNA 1|DISC1 opposite strand', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231814626, end_position_on_the_genomic_accession=231818517, orientation='minus', exon_count=1, OMIM=606271, no_hdr0=''),
  27185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27185, CurrentID=0, Status='live', Symbol='DISC1', Aliases='C1orf136, SCZD9', description='disrupted in schizophrenia 1', other_designations='', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231626815, end_position_on_the_genomic_accession=232041272, orientation='plus', exon_count=19, OMIM=605210, no_hdr0=''),
  27187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27187, CurrentID=0, Status='live', Symbol='MOCS1P1', Aliases='MOCS1P2', description='molybdenum cofactor synthesis 1 pseudogene 1', other_designations='molybdenum cofactor synthesis 1 pseudogene 2', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48496606, end_position_on_the_genomic_accession=48499426, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419788, CurrentID=0, Status='live', Symbol='RNF19BPX', Aliases='RNF19BX', description='ring finger protein 19B pseudogene, X-linked', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=90209417, end_position_on_the_genomic_accession=90210234, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27189, CurrentID=0, Status='live', Symbol='IL17C', Aliases='CX2, IL-17C', description='interleukin 17C', other_designations='cytokine CX2', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88638593, end_position_on_the_genomic_accession=88640474, orientation='plus', exon_count=3, OMIM=604628, no_hdr0=''),
  27190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27190, CurrentID=0, Status='live', Symbol='IL17B', Aliases='IL-17B, IL-20, NIRF, ZCYTO7', description='interleukin 17B', other_designations='cytokine-like protein ZCYTO7|interleukin 20|interleukin-17 beta|neuronal interleukin-17 related factor', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149374267, end_position_on_the_genomic_accession=149404244, orientation='minus', exon_count=7, OMIM=604627, no_hdr0=''),
  158263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158263, CurrentID=0, Status='live', Symbol='FLJ31713', Aliases='', description='uncharacterized protein FLJ31713', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=115418339, end_position_on_the_genomic_accession=115418830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27192, CurrentID=0, Status='live', Symbol='TRAPPC2P6', Aliases='SEDLP6', description='trafficking protein particle complex 2 pseudogene 6', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 6', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27193, CurrentID=0, Status='live', Symbol='TRAPPC2P5', Aliases='SEDLP5', description='trafficking protein particle complex 2 pseudogene 5', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 5', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25699329, end_position_on_the_genomic_accession=25707095, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27194, CurrentID=0, Status='live', Symbol='TRAPPC2P4', Aliases='SEDLP4', description='trafficking protein particle complex 2 pseudogene 4', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 4', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25904397, end_position_on_the_genomic_accession=25915212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27195, CurrentID=0, Status='live', Symbol='TRAPPC2P3', Aliases='SEDLP3', description='trafficking protein particle complex 2 pseudogene 3', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 3', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17796185, end_position_on_the_genomic_accession=17800818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27196, CurrentID=0, Status='live', Symbol='TRAPPC2P2', Aliases='SEDLP2', description='trafficking protein particle complex 2 pseudogene 2', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 2', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71448711, end_position_on_the_genomic_accession=71449240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27197, CurrentID=0, Status='live', Symbol='GPR82', Aliases='', description='G protein-coupled receptor 82', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41724155, end_position_on_the_genomic_accession=41730135, orientation='plus', exon_count=3, OMIM=300748, no_hdr0=''),
  27198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27198, CurrentID=0, Status='live', Symbol='HCAR1', Aliases='FKSG80, GPR104, GPR81, HCA1, LACR1, TA-GPCR, TAGPCR', description='hydroxycarboxylic acid receptor 1', other_designations='G protein-coupled receptor 104|G protein-coupled receptor 81|G-protein coupled receptor 104|G-protein coupled receptor 81|T-cell activation G protein-coupled receptor|hydroxy-carboxylic acid receptor 1|lactate receptor 1', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122727606, end_position_on_the_genomic_accession=122730582, orientation='minus', exon_count=1, OMIM=606923, no_hdr0=''),
  27199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27199, CurrentID=0, Status='live', Symbol='OXGR1', Aliases='GPR80, GPR99, P2RY15, P2Y15, aKGR', description='oxoglutarate (alpha-ketoglutarate) receptor 1', other_designations='G protein-coupled receptor 80|G protein-coupled receptor 99|G-protein coupled receptor 80|G-protein coupled receptor 99|P2Y purinoceptor 15|P2Y-like GPCR|P2Y-like nucleotide receptor|alpha-ketoglutarate receptor 1|oxoglutarate receptor 1|seven transmembrane helix receptor', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96985719, end_position_on_the_genomic_accession=96995826, orientation='minus', exon_count=5, OMIM=606922, no_hdr0=''),
  27200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27200, CurrentID=0, Status='live', Symbol='GPR79', Aliases='GPR79P', description='G protein-coupled receptor 79, pseudogene', other_designations='G protein-coupled receptor 79 pseudogene', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158841067, end_position_on_the_genomic_accession=158842084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27201, CurrentID=0, Status='live', Symbol='GPR78', Aliases='', description='G protein-coupled receptor 78', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8580237, end_position_on_the_genomic_accession=8590023, orientation='plus', exon_count=3, OMIM=606921, no_hdr0=''),
  27202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27202, CurrentID=0, Status='live', Symbol='C5AR2', Aliases='C5L2, GPF77, GPR77', description='complement component 5a receptor 2', other_designations='C5a anaphylatoxin chemotactic receptor C5L2|G protein-coupled receptor 77|G protein-coupled receptor C5L2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47331841, end_position_on_the_genomic_accession=47342080, orientation='plus', exon_count=5, OMIM=609949, no_hdr0=''),
  266678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266678, CurrentID=0, Status='live', Symbol='MRX81', Aliases='', description='mental retardation, X-linked 81', other_designations='', map_location='Xp11.2-q12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300433, no_hdr0=''),
  100288853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288853, CurrentID=0, Status='live', Symbol='LOC100288853', Aliases='', description='poly(A) binding protein, nuclear 1 pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=67533154, end_position_on_the_genomic_accession=67533943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369155, CurrentID=0, Status='live', Symbol='LOC105369155', Aliases='', description='uncharacterized LOC105369155', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67210434, end_position_on_the_genomic_accession=67226998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  284600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284600, CurrentID=0, Status='live', Symbol='LOC284600', Aliases='', description='uncharacterized LOC284600', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=905115, end_position_on_the_genomic_accession=920484, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  158293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158293, CurrentID=0, Status='live', Symbol='FAM120AOS', Aliases='C9orf10OS', description='family with sequence similarity 120A opposite strand', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93443332, end_position_on_the_genomic_accession=93453592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  790969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790969, CurrentID=0, Status='live', Symbol='ZNF705F', Aliases='', description='zinc finger protein 705F', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158297, CurrentID=0, Status='live', Symbol='SAXO1', Aliases='C9orf138, FAM154A', description='stabilizer of axonemal microtubules 1', other_designations='4930500O09Rik|family with sequence similarity 154, member A|protein FAM154A', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18927660, end_position_on_the_genomic_accession=19049354, orientation='minus', exon_count=8, OMIM=616292, no_hdr0=''),
  27229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27229, CurrentID=0, Status='live', Symbol='TUBGCP4', Aliases='76P, GCP-4, GCP4, Grip76, MCCRP3', description='tubulin gamma complex associated protein 4', other_designations='gamma tubulin ring complex protein (76p gene)|gamma-ring complex protein 76 kDa|tubulin, gamma complex associated protein 4', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43371059, end_position_on_the_genomic_accession=43407095, orientation='plus', exon_count=19, OMIM=609610, no_hdr0=''),
  27230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27230, CurrentID=0, Status='live', Symbol='SERP1', Aliases='RAMP4', description='stress-associated endoplasmic reticulum protein 1', other_designations='ribosome associated membrane protein 4|ribosome-attached membrane protein 4', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150541993, end_position_on_the_genomic_accession=150603177, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27231, CurrentID=0, Status='live', Symbol='NMRK2', Aliases='ITGB1BP3, MIBP, NRK2', description='nicotinamide riboside kinase 2', other_designations='NRK 2|RNK 2|integrin beta-1-binding protein 3|muscle-specific beta 1 integrin binding protein|nicotinic acid riboside kinase 2|nmR-K 2|ribosylnicotinamide kinase 2|ribosylnicotinic acid kinase 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3933103, end_position_on_the_genomic_accession=3942416, orientation='plus', exon_count=8, OMIM=608705, no_hdr0=''),
  27232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27232, CurrentID=0, Status='live', Symbol='GNMT', Aliases='', description='glycine N-methyltransferase', other_designations='HEL-S-182mP|epididymis secretory sperm binding protein Li 182mP', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42939415, end_position_on_the_genomic_accession=42963880, orientation='plus', exon_count=9, OMIM=606628, no_hdr0=''),
  27233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27233, CurrentID=0, Status='live', Symbol='SULT1C4', Aliases='SULT1C, SULT1C2', description='sulfotransferase family 1C member 4', other_designations='ST1C4|SULT1C#2|sulfotransferase 1C2|sulfotransferase family, cytosolic, 1C, member 2|sulfotransferase family, cytosolic, 1C, member 4|sulfotransferase family, cytosolic, 1C, member C2', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108377957, end_position_on_the_genomic_accession=108388989, orientation='plus', exon_count=7, OMIM=608357, no_hdr0=''),
  27235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27235, CurrentID=0, Status='live', Symbol='COQ2', Aliases='CL640, COQ10D1, MSA1, PHB:PPT', description='coenzyme Q2 4-hydroxybenzoate polyprenyltransferase', other_designations='4-HB polyprenyltransferase|4-hydroxybenzoate decaprenyltransferase|PHB:polyprenyltransferase|coenzyme Q2 homolog, prenyltransferase|para-hydroxybenzoate-polyprenyltransferase, mitochondrial', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83138601, end_position_on_the_genomic_accession=83285089, orientation='minus', exon_count=22, OMIM=609825, no_hdr0=''),
  27236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27236, CurrentID=0, Status='live', Symbol='ARFIP1', Aliases='HSU52521', description='ADP ribosylation factor interacting protein 1', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152779937, end_position_on_the_genomic_accession=152911912, orientation='plus', exon_count=14, OMIM=605928, no_hdr0=''),
  27237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27237, CurrentID=0, Status='live', Symbol='ARHGEF16', Aliases='GEF16, NBR', description='Rho guanine nucleotide exchange factor 16', other_designations='Rho guanine exchange factor (GEF) 16|Rho guanine nucleotide exchange factor (GEF) 16|ephexin-4', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3454583, end_position_on_the_genomic_accession=3481115, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  27238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27238, CurrentID=0, Status='live', Symbol='GPKOW', Aliases='GPATC5, GPATCH5, Spp2, T54', description='G-patch domain and KOW motifs', other_designations='G patch domain and KOW motifs|protein MOS2 homolog', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49113385, end_position_on_the_genomic_accession=49123801, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27239, CurrentID=0, Status='live', Symbol='GPR162', Aliases='A-2, GRCA', description='G protein-coupled receptor 162', other_designations='gene rich cluster, A|gene-rich cluster gene A protein', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6821797, end_position_on_the_genomic_accession=6827418, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27240, CurrentID=0, Status='live', Symbol='SIT1', Aliases='SIT, SIT-R', description='signaling threshold regulating transmembrane adaptor 1', other_designations='SHP-2 interacting transmembrane adaptor protein|SHP2 interacting transmembrane adaptor|SHP2-interacting transmembrane adapter protein|SHP2-interacting transmembrane adaptor protein|gp30/40|suppression inducing transmembrane adaptor 1|suppression-inducing transmembrane adapter 1', map_location='9p13-p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35649298, end_position_on_the_genomic_accession=35650950, orientation='minus', exon_count=5, OMIM=604964, no_hdr0=''),
  27241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27241, CurrentID=0, Status='live', Symbol='BBS9', Aliases='B1, C18, D1, PTHB1', description='Bardet-Biedl syndrome 9', other_designations='PTH-responsive osteosarcoma B1 protein|bardet-Biedl syndrome 9 protein|parathyroid hormone-responsive B1 gene protein', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33129244, end_position_on_the_genomic_accession=33635769, orientation='plus', exon_count=26, OMIM=607968, no_hdr0=''),
  27242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27242, CurrentID=0, Status='live', Symbol='TNFRSF21', Aliases='BM-018, CD358, DR6', description='tumor necrosis factor receptor superfamily member 21', other_designations='TNFR-related death receptor 6|death receptor 6|tumor necrosis factor receptor superfamily, member 21', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47231527, end_position_on_the_genomic_accession=47309976, orientation='minus', exon_count=7, OMIM=605732, no_hdr0=''),
  27243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27243, CurrentID=0, Status='live', Symbol='CHMP2A', Aliases='BC-2, BC2, CHMP2, VPS2, VPS2A', description='charged multivesicular body protein 2A', other_designations='VPS2 homolog A|chromatin modifying protein 2A|putative breast adenocarcinoma marker (32kD)|putative breast adenocarcinoma marker BC-2|vacuolar protein sorting-associated protein 2-1|vps2-1', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58551566, end_position_on_the_genomic_accession=58555128, orientation='minus', exon_count=5, OMIM=610893, no_hdr0=''),
  27244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27244, CurrentID=0, Status='live', Symbol='SESN1', Aliases='PA26, SEST1', description='sestrin 1', other_designations='p53 activated gene 26|p53 regulated PA26 nuclear protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108986437, end_position_on_the_genomic_accession=109094505, orientation='minus', exon_count=12, OMIM=606103, no_hdr0=''),
  27245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27245, CurrentID=0, Status='live', Symbol='AHDC1', Aliases='MRD25', description='AT-hook DNA binding motif containing 1', other_designations='AT hook, DNA binding motif, containing 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27534245, end_position_on_the_genomic_accession=27604178, orientation='minus', exon_count=12, OMIM=615790, no_hdr0=''),
  27246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27246, CurrentID=0, Status='live', Symbol='RNF115', Aliases='BCA2, ZNF364', description='ring finger protein 115', other_designations='rabring 7|zinc finger protein 364', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145746082, end_position_on_the_genomic_accession=145824123, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27247, CurrentID=0, Status='live', Symbol='NFU1', Aliases='CGI-33, HIRIP, HIRIP5, MMDS1, NIFUC, Nfu, NifU', description='NFU1 iron-sulfur cluster scaffold', other_designations='HIRA-interacting protein 5|NifU-like C-terminal domain containing|iron-sulfur cluster scaffold protein', map_location='2p15-p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69396113, end_position_on_the_genomic_accession=69437628, orientation='minus', exon_count=8, OMIM=608100, no_hdr0=''),
  27248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27248, CurrentID=0, Status='live', Symbol='ERLEC1', Aliases='C2orf30, CIM, CL24936, CL25084, HEL117, XTP3-B, XTP3TPB', description='endoplasmic reticulum lectin 1', other_designations='ER lectin|XTP3-transactivated gene B protein|XTP3-transactivated protein B|cancer invasion and metastasis-related|epididymis luminal protein 117|erlectin 1', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53786931, end_position_on_the_genomic_accession=53826928, orientation='plus', exon_count=15, OMIM=611229, no_hdr0=''),
  27249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27249, CurrentID=0, Status='live', Symbol='MMADHC', Aliases='C2orf25, CL25022, cblD', description='methylmalonic aciduria (cobalamin deficiency) cblD type, with homocystinuria', other_designations='protein C2orf25, mitochondrial', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149569633, end_position_on_the_genomic_accession=149587816, orientation='minus', exon_count=8, OMIM=611935, no_hdr0=''),
  27250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27250, CurrentID=0, Status='live', Symbol='PDCD4', Aliases='H731', description='programmed cell death 4 (neoplastic transformation inhibitor)', other_designations='neoplastic transformation inhibitor protein|nuclear antigen H731|protein 197/15a', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110871795, end_position_on_the_genomic_accession=110900006, orientation='plus', exon_count=13, OMIM=608610, no_hdr0=''),
  105379816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379816, CurrentID=0, Status='live', Symbol='LOC105379816', Aliases='', description='uncharacterized LOC105379816', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27252, CurrentID=0, Status='live', Symbol='KLHL20', Aliases='KHLHX, KLEIP, KLHLX', description='kelch like family member 20', other_designations='Kelch motif containing protein|kelch-like 20|kelch-like ECT2 interacting protein|kelch-like ECT2-interacting protein|kelch-like family member 20|kelch-like protein X', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173714912, end_position_on_the_genomic_accession=173786822, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  27253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27253, CurrentID=0, Status='live', Symbol='PCDH17', Aliases='PCDH68, PCH68', description='protocadherin 17', other_designations='protocadherin 68|protocadherin-68', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57630104, end_position_on_the_genomic_accession=57729311, orientation='plus', exon_count=4, OMIM=611760, no_hdr0=''),
  158326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158326, CurrentID=0, Status='live', Symbol='FREM1', Aliases='BNAR, C9orf143, C9orf145, C9orf154, MOTA, TILRR, TRIGNO2', description='FRAS1 related extracellular matrix 1', other_designations='extracellular matrix protein QBRICK', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14734666, end_position_on_the_genomic_accession=14911647, orientation='minus', exon_count=46, OMIM=608944, no_hdr0=''),
  27255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27255, CurrentID=0, Status='live', Symbol='CNTN6', Aliases='NB3', description='contactin 6', other_designations='neural adhesion molecule|neural recognition molecule NB-3', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1008135, end_position_on_the_genomic_accession=1403610, orientation='plus', exon_count=30, OMIM=607220, no_hdr0=''),
  27257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27257, CurrentID=0, Status='live', Symbol='LSM1', Aliases='CASM, YJL124C', description='LSM1 homolog, mRNA degradation associated', other_designations='LSM1 homolog, U6 small nuclear RNA associated|LSM1 mRNA degradation associated|LSM1, U6 small nuclear RNA associated|LSM1-like protein U6 small nuclear RNA associated|cancer&#8208;associated Sm protein|cancer-associated Sm protein|cancer-associated Sm-like protein|small nuclear ribonuclear CaSm', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38163321, end_position_on_the_genomic_accession=38176730, orientation='minus', exon_count=4, OMIM=607281, no_hdr0=''),
  27258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27258, CurrentID=0, Status='live', Symbol='LSM3', Aliases='SMX4, USS2, YLR438C', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated|LSM3 homolog, U6 small nuclear RNA associated', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14178728, end_position_on_the_genomic_accession=14198369, orientation='plus', exon_count=4, OMIM=607283, no_hdr0=''),
  27259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27259, CurrentID=0, Status='live', Symbol='HPLH1', Aliases='FHL1, HLH1', description='hemophagocytic lymphohistiocytosis 1', other_designations='', map_location='9q21.3-q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=267700, no_hdr0=''),
  105373521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373521, CurrentID=0, Status='live', Symbol='LOC105373521', Aliases='', description='uncharacterized LOC105373521', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104093618, end_position_on_the_genomic_accession=104111966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  158345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158345, CurrentID=0, Status='live', Symbol='RPL4P5', Aliases='RPL4_3_960', description='ribosomal protein L4 pseudogene 5', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7476951, end_position_on_the_genomic_accession=7478376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101410537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410537, CurrentID=0, Status='live', Symbol='MAGEA8-AS1', Aliases='', description='MAGEA8 antisense RNA 1 (head to head)', other_designations='RP5-869M20.2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149878763, end_position_on_the_genomic_accession=149897006, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100419791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419791, CurrentID=0, Status='live', Symbol='LOC100419791', Aliases='', description='zinc finger protein 449 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135135188, end_position_on_the_genomic_accession=135135746, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266691, CurrentID=0, Status='live', Symbol='DYX6', Aliases='DYXQTL18', description='dyslexia susceptibility 6', other_designations='Dyslexia QTL on 18', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606616, no_hdr0=''),
  27284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27284, CurrentID=0, Status='live', Symbol='SULT1B1', Aliases='ST1B1, ST1B2, SULT1B2', description='sulfotransferase family 1B member 1', other_designations='sulfotransferase 1B1|sulfotransferase 1B2|sulfotransferase family, cytosolic, 1B, member 1|thyroid hormone sulfotransferase', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69726120, end_position_on_the_genomic_accession=69760712, orientation='minus', exon_count=10, OMIM=608436, no_hdr0=''),
  27285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27285, CurrentID=0, Status='live', Symbol='TEKT2', Aliases='TEKTB1, TEKTIN-T, h-tektin-t', description='tektin 2 (testicular)', other_designations='tektin-B1|testicular tektin B1-like protein', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36084075, end_position_on_the_genomic_accession=36088276, orientation='plus', exon_count=10, OMIM=608953, no_hdr0=''),
  158358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158358, CurrentID=0, Status='live', Symbol='KIAA2026', Aliases='', description='KIAA2026', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5859595, end_position_on_the_genomic_accession=6008479, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  27287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27287, CurrentID=0, Status='live', Symbol='VENTX', Aliases='HPX42B, NA88A2, VENTX', description='VENT homeobox', other_designations='VENT homeobox homolog|VENT-like homeobox 2|VENT-like homeobox protein 2|hemopoietic progenitor homeobox protein VENTX2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133237904, end_position_on_the_genomic_accession=133241930, orientation='plus', exon_count=3, OMIM=607158, no_hdr0=''),
  27288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27288, CurrentID=0, Status='live', Symbol='RBMXL2', Aliases='HNRNPG-T, HNRNPGT, HNRPGT', description='RNA binding motif protein, X-linked-like 2', other_designations='heterogeneous nuclear ribonucleoprotein G T|hnRNP G-T|testes specific heterogenous nuclear ribonucleoprotein G T|testes-specific heterogenous nuclear ribonucleoprotein G-T|testis-specific heterogeneous nuclear ribonucleoprotein G-T', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7088934, end_position_on_the_genomic_accession=7091148, orientation='plus', exon_count=1, OMIM=605444, no_hdr0=''),
  27289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27289, CurrentID=0, Status='live', Symbol='RND1', Aliases='ARHS, RHO6, RHOS', description='Rho family GTPase 1', other_designations='ras homolog gene family, member S', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48857133, end_position_on_the_genomic_accession=48865870, orientation='minus', exon_count=5, OMIM=609038, no_hdr0=''),
  27290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27290, CurrentID=0, Status='live', Symbol='SPINK4', Aliases='HEL136, PEC-60, PEC60', description='serine peptidase inhibitor, Kazal type 4', other_designations='epididymis luminal protein 136|gastrointestinal peptide|peptide PEC-60 homolog|serine protease inhibitor, Kazal type 4', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33240198, end_position_on_the_genomic_accession=33248567, orientation='plus', exon_count=4, OMIM=613929, no_hdr0=''),
  27291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27291, CurrentID=0, Status='live', Symbol='R3HCC1L', Aliases='C10orf28, GIDRP86, GIDRP88, PSORT', description='R3H domain and coiled-coil containing 1 like', other_designations='R3H and coiled-coil domain-containing protein 1-like|R3H domain and coiled-coil containing 1-like|growth inhibition and differentiation related protein 86|growth inhibition and differentiation-related protein 88|putative mitochondrial space protein 32.1', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98134624, end_position_on_the_genomic_accession=98244897, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  27292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27292, CurrentID=0, Status='live', Symbol='DIMT1', Aliases='DIM1L, HSA9761, HUSSY5, DIMT1', description='DIM1 dimethyladenosine transferase 1 homolog', other_designations="18S rRNA (adenine(1779)-N(6)/adenine(1780)-N(6))-dimethyltransferase|18S rRNA dimethylase|S-adenosylmethionine-6-N',N'-adenosyl(rRNA) dimethyltransferase|dimethyladenosine transferase|probable 18S rRNA (adenine(1779)-N(6)/adenine(1780)-N(6))-dimethyltransferase|probable 18S rRNA dimethylase|probable S-adenosylmethionine-6-N',N'-adenosyl(rRNA) dimethyltransferase", map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62388524, end_position_on_the_genomic_accession=62403901, orientation='minus', exon_count=12, OMIM=612499, no_hdr0=''),
  27293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27293, CurrentID=0, Status='live', Symbol='SMPDL3B', Aliases='ASML3B', description='sphingomyelin phosphodiesterase acid-like 3B', other_designations='ASM-like phosphodiesterase 3b|sphingomyelin phosphodiesterase, acid-like 3B', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27934948, end_position_on_the_genomic_accession=27959157, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27294, CurrentID=0, Status='live', Symbol='DHDH', Aliases='2DD, HUM2DD', description='dihydrodiol dehydrogenase (dimeric)', other_designations='3-deoxyglucosone reductase|D-xylose 1-dehydrogenase|D-xylose-NADP dehydrogenase', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48933639, end_position_on_the_genomic_accession=48944969, orientation='plus', exon_count=7, OMIM=606377, no_hdr0=''),
  266693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266693, CurrentID=0, Status='live', Symbol='EZH2P1', Aliases='', description='enhancer of zeste 2 polycomb repressive complex 2 subunit pseudogene 1', other_designations='enhancer of zeste homolog 2 pseudogene 1', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35599679, end_position_on_the_genomic_accession=35600255, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27296, CurrentID=0, Status='live', Symbol='TP53TG5', Aliases='C20orf10, CLG01', description='TP53 target 5', other_designations='TP53-inducible gene 5 protein', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45373880, end_position_on_the_genomic_accession=45407838, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27297, CurrentID=0, Status='live', Symbol='CRCP', Aliases='CGRP-RCP, CGRPRCP, RCP, RCP9', description='CGRP receptor component', other_designations='CGRP-receptor component protein|RNA polymerase III subunit C9|calcitonin gene-related peptide-receptor component protein', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66114818, end_position_on_the_genomic_accession=66154568, orientation='plus', exon_count=7, OMIM=606121, no_hdr0=''),
  27299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27299, CurrentID=0, Status='live', Symbol='ADAMDEC1', Aliases='M12.219', description='ADAM-like, decysin 1', other_designations='ADAM-like protein decysin-1|a disintegrin and metalloproteinase domain-like protein decysin-1|decysin|disintegrin protease', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24384285, end_position_on_the_genomic_accession=24406013, orientation='plus', exon_count=15, OMIM=606393, no_hdr0=''),
  27300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27300, CurrentID=0, Status='live', Symbol='ZNF544', Aliases='', description='zinc finger protein 544', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58227783, end_position_on_the_genomic_accession=58263644, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27301, CurrentID=0, Status='live', Symbol='APEX2', Aliases='APE2, APEXL2, XTH2, ZGRF2', description='APEX nuclease (apurinic/apyrimidinic endonuclease) 2', other_designations='AP endonuclease 2|AP endonuclease XTH2|apurinic/apyrimidinic endonuclease-like 2|zinc finger, GRF-type containing 2', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55000323, end_position_on_the_genomic_accession=55007873, orientation='plus', exon_count=6, OMIM=300773, no_hdr0=''),
  27302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27302, CurrentID=0, Status='live', Symbol='BMP10', Aliases='', description='bone morphogenetic protein 10', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68865481, end_position_on_the_genomic_accession=68871517, orientation='minus', exon_count=2, OMIM=608748, no_hdr0=''),
  27303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27303, CurrentID=0, Status='live', Symbol='RBMS3', Aliases='', description='RNA binding motif, single stranded interacting protein 3', other_designations='RNA-binding protein', map_location='3p24-p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29280860, end_position_on_the_genomic_accession=30010395, orientation='plus', exon_count=22, OMIM=605786, no_hdr0=''),
  27304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27304, CurrentID=0, Status='live', Symbol='MOCS3', Aliases='UBA4', description='molybdenum cofactor synthesis 3', other_designations='MPT synthase sulfurylase|UBA4, ubiquitin-activating enzyme E1 homolog|molybdenum cofactor synthesis protein 3|molybdopterin synthase sulfurylase|ubiquitin-like modifier activating enzyme 4', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50958814, end_position_on_the_genomic_accession=50961863, orientation='plus', exon_count=1, OMIM=609277, no_hdr0=''),
  100873702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873702, CurrentID=0, Status='live', Symbol='RNA5SP457', Aliases='RN5S457', description='RNA, 5S ribosomal pseudogene 457', other_designations='RNA, 5S ribosomal 457', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49946266, end_position_on_the_genomic_accession=49946389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27306, CurrentID=0, Status='live', Symbol='HPGDS', Aliases='GSTS, GSTS1, GSTS1-1, PGD2, PGDS', description='hematopoietic prostaglandin D synthase', other_designations='GST class-sigma|glutathione S-transferase sigma|glutathione-dependent PGD synthase|glutathione-dependent PGD synthetase|glutathione-requiring prostaglandin D synthase|hematopoietic prostaglandin D2 synthase|prostaglandin-H2 D-isomerase', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94298556, end_position_on_the_genomic_accession=94342876, orientation='minus', exon_count=6, OMIM=602598, no_hdr0=''),
  27309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27309, CurrentID=0, Status='live', Symbol='ZNF330', Aliases='HSA6591, NOA36', description='zinc finger protein 330', other_designations='nucleolar autoantigen 36|nucleolar cysteine-rich protein|zinc finger autoantigen 330', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=141220294, end_position_on_the_genomic_accession=141234697, orientation='plus', exon_count=11, OMIM=609550, no_hdr0=''),
  158383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158383, CurrentID=0, Status='live', Symbol='YWHAZP6', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 6', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34921187, end_position_on_the_genomic_accession=34923019, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27314, CurrentID=0, Status='live', Symbol='RAB30', Aliases='', description='RAB30, member RAS oncogene family', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82973133, end_position_on_the_genomic_accession=83071923, orientation='minus', exon_count=7, OMIM=605693, no_hdr0=''),
  27315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27315, CurrentID=0, Status='live', Symbol='PGAP2', Aliases='CWH43-N, FRAG1, HPMRS3, MRT17, MRT21', description='post-GPI attachment to proteins 2', other_designations='FGF receptor activating protein 1|FGF receptor-activating protein 1|cell wall biogenesis 43 N-terminal homolog|mental retardation, non-syndromic, autosomal recessive, 21', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3797724, end_position_on_the_genomic_accession=3826371, orientation='plus', exon_count=15, OMIM=615187, no_hdr0=''),
  27316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27316, CurrentID=0, Status='live', Symbol='RBMX', Aliases='HNRNPG, HNRPGP1, RBMXRT, RNMX, hnRNP-G, RBMX', description='RNA binding motif protein, X-linked', other_designations='RNA binding motif protein, X chromosome|glycoprotein p43|heterogeneous nuclear ribonucleoprotein G|hnRNP G', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136869194, end_position_on_the_genomic_accession=136880780, orientation='minus', exon_count=11, OMIM=300199, no_hdr0=''),
  441726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441726, CurrentID=0, Status='live', Symbol='RPL28P4', Aliases='RPL28_1_1435', description='ribosomal protein L28 pseudogene 4', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58200606, end_position_on_the_genomic_accession=58201018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27319, CurrentID=0, Status='live', Symbol='BHLHE22', Aliases='BHLHB5, Beta3, CAGL85, TNRC20', description='basic helix-loop-helix family member e22', other_designations='basic helix-loop-helix domain containing, class B, 5|basic helix-loop-helix family, member e22|class B basic helix-loop-helix protein 5|trinucleotide repeat containing 20|trinucleotide repeat-containing gene 20 protein', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64580238, end_position_on_the_genomic_accession=64583634, orientation='plus', exon_count=1, OMIM=613483, no_hdr0=''),
  27320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27320, CurrentID=0, Status='live', Symbol='TNRC18P2', Aliases='TNRC18, TNRC18B', description='trinucleotide repeat containing 18 pseudogene 2', other_designations='trinucleotide repeat containing 18B', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63566373, end_position_on_the_genomic_accession=63585074, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27321, CurrentID=0, Status='live', Symbol='TNRC17', Aliases='CAGL234', description='trinucleotide repeat containing 17', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27324, CurrentID=0, Status='live', Symbol='TOX3', Aliases='CAGF9, TNRC9', description='TOX high mobility group box family member 3', other_designations='CAG trinucleotide repeat-containing gene F9 protein|trinucleotide repeat containing 9|trinucleotide repeat-containing gene 9 protein', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52436425, end_position_on_the_genomic_accession=52547802, orientation='minus', exon_count=10, OMIM=611416, no_hdr0=''),
  27327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27327, CurrentID=0, Status='live', Symbol='TNRC6A', Aliases='CAGH26, GW1, GW182, TNRC6', description='trinucleotide repeat containing 6A', other_designations='CAG repeat protein 26|EDIE|EMSY interactor protein|GW182 autoantigen|glycine-tryptophan protein of 182 kDa', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24729714, end_position_on_the_genomic_accession=24826227, orientation='plus', exon_count=27, OMIM=610739, no_hdr0=''),
  27328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27328, CurrentID=0, Status='live', Symbol='PCDH11X', Aliases='PCDH-X, PCDH11, PCDHX, PPP1R119', description='protocadherin 11 X-linked', other_designations='protein phosphatase 1, regulatory subunit 119|protocadherin 11X|protocadherin on the X chromosome|protocadherin-S', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91779311, end_position_on_the_genomic_accession=92623230, orientation='plus', exon_count=14, OMIM=300246, no_hdr0=''),
  27329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27329, CurrentID=0, Status='live', Symbol='ANGPTL3', Aliases='ANG-5, ANGPT5, ANL3, FHBL2', description='angiopoietin like 3', other_designations='angiopoietin 5|angiopoietin-like 3', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62597487, end_position_on_the_genomic_accession=62606305, orientation='plus', exon_count=7, OMIM=604774, no_hdr0=''),
  27330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27330, CurrentID=0, Status='live', Symbol='RPS6KA6', Aliases='PP90RSK4, RSK4', description='ribosomal protein S6 kinase, 90kDa, polypeptide 6', other_designations='90 kDa ribosomal protein S6 kinase 6|RSK-4|S6K-alpha 6|S6K-alpha-6|p90-RSK 6|p90RSK6|ribosomal S6 kinase 4', map_location='Xq21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84058346, end_position_on_the_genomic_accession=84187964, orientation='minus', exon_count=26, OMIM=300303, no_hdr0=''),
  4555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4555, CurrentID=0, Status='live', Symbol='MT-TD', Aliases='MTTD, TRND', description='mitochondrially encoded tRNA aspartic acid', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=7518, end_position_on_the_genomic_accession=7585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27332, CurrentID=0, Status='live', Symbol='ZNF638', Aliases='NP220, ZFML, Zfp638', description='zinc finger protein 638', other_designations='CTCL tumor antigen se33-1|CTCL-associated antigen se33-1|NP220 nuclear protein|cutaneous T-cell lymphoma-associated antigen se33-1|nuclear protein 220|zinc finger matrin-like protein', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71331755, end_position_on_the_genomic_accession=71435061, orientation='plus', exon_count=29, OMIM=614349, no_hdr0=''),
  158405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158405, CurrentID=0, Status='live', Symbol='KIAA1958', Aliases='', description='KIAA1958', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112486968, end_position_on_the_genomic_accession=112665311, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27334, CurrentID=0, Status='live', Symbol='P2RY10', Aliases='LYPSR2, P2Y10', description='purinergic receptor P2Y, G-protein coupled, 10', other_designations='G-protein coupled purinergic receptor P2Y10|P2Y purinoceptor 10|P2Y-like receptor', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78945332, end_position_on_the_genomic_accession=78963727, orientation='plus', exon_count=4, OMIM=300529, no_hdr0=''),
  27335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27335, CurrentID=0, Status='live', Symbol='EIF3K', Aliases='ARG134, EIF3-p28, EIF3S12, HSPC029, M9, MSTP001, PLAC-24, PLAC24, PRO1474, PTD001', description='eukaryotic translation initiation factor 3 subunit K', other_designations='eIF-3 p28|eukaryotic translation initiation factor 3, subunit 12|eukaryotic translation initiation factor 3, subunit K|muscle specific|muscle-specific gene M9 protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38618419, end_position_on_the_genomic_accession=38636959, orientation='plus', exon_count=8, OMIM=609596, no_hdr0=''),
  27336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27336, CurrentID=0, Status='live', Symbol='HTATSF1', Aliases='TAT-SF1, TATSF1, dJ196E23.2', description='HIV-1 Tat specific factor 1', other_designations='HIV TAT specific factor 1|cofactor required for Tat activation of HIV-1 transcription', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136496932, end_position_on_the_genomic_accession=136512346, orientation='plus', exon_count=10, OMIM=300346, no_hdr0=''),
  266700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266700, CurrentID=0, Status='live', Symbol='BMP6P1', Aliases='TGF-beta1P, TGFB1P', description='bone morphogenetic protein 6 pseudogene 1', other_designations='transforming growth factor beta 1 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22185470, end_position_on_the_genomic_accession=22187116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27338, CurrentID=0, Status='live', Symbol='UBE2S', Aliases='E2-EPF, E2EPF, EPF5', description='ubiquitin conjugating enzyme E2S', other_designations='E2 ubiquitin-conjugating enzyme S|E2-EPF5|ubiquitin carrier protein S|ubiquitin-conjugating enzyme E2-24 kD|ubiquitin-conjugating enzyme E2-24 kDa|ubiquitin-conjugating enzyme E2-EPF5|ubiquitin-conjugating enzyme E2S|ubiquitin-protein ligase S', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55401282, end_position_on_the_genomic_accession=55408146, orientation='minus', exon_count=6, OMIM=610309, no_hdr0=''),
  27339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27339, CurrentID=0, Status='live', Symbol='PRPF19', Aliases='NMP200, PRP19, PSO4, SNEV, UBOX4, hPSO4', description='pre-mRNA processing factor 19', other_designations='PRP19/PSO4 homolog|PRP19/PSO4 pre-mRNA processing factor 19 homolog|nuclear matrix protein 200|nuclear matrix protein NMP200 related to splicing factor PRP19|psoralen 4|senescence evasion factor', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60890547, end_position_on_the_genomic_accession=60906589, orientation='minus', exon_count=16, OMIM=608330, no_hdr0=''),
  27340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27340, CurrentID=0, Status='live', Symbol='UTP20', Aliases='DRIM', description='UTP20 small subunit processome component', other_designations='NNP73|UTP20 small subunit (SSU) processome component|UTP20, small subunit (SSU) processome component, homolog|down regulated in metastasis|down-regulated in metastasis protein|novel nucleolar protein 73', map_location='12q23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101280127, end_position_on_the_genomic_accession=101386619, orientation='plus', exon_count=62, OMIM=612822, no_hdr0=''),
  27341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27341, CurrentID=0, Status='live', Symbol='RRP7A', Aliases='BK126B4.3, CGI-96', description='ribosomal RNA processing 7 homolog A', other_designations='CTA-126B4.5|gastric cancer antigen Zg14', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42508335, end_position_on_the_genomic_accession=42519823, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27342, CurrentID=0, Status='live', Symbol='RABGEF1', Aliases='RABEX5, RAP1, rabex-5', description='RAB guanine nucleotide exchange factor (GEF) 1', other_designations='rabaptin-5-associated exchange factor for Rab5', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66682091, end_position_on_the_genomic_accession=66811464, orientation='plus', exon_count=14, OMIM=609700, no_hdr0=''),
  27343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27343, CurrentID=0, Status='live', Symbol='POLL', Aliases='BETAN, POLKAPPA', description='polymerase (DNA directed), lambda', other_designations='DNA polymerase beta-2|DNA polymerase beta-N|DNA polymerase kappa', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101578882, end_position_on_the_genomic_accession=101588270, orientation='minus', exon_count=11, OMIM=606343, no_hdr0=''),
  27344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27344, CurrentID=0, Status='live', Symbol='PCSK1N', Aliases='PROSAAS, SAAS, SCG8, SgVIII', description='proprotein convertase subtilisin/kexin type 1 inhibitor', other_designations='granin-like neuroendocrine peptide|pro-SAAS|proprotein convertase 1 inhibitor', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48831092, end_position_on_the_genomic_accession=48835638, orientation='minus', exon_count=3, OMIM=300399, no_hdr0=''),
  27345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27345, CurrentID=0, Status='live', Symbol='KCNMB4', Aliases='', description='potassium channel subfamily M regulatory beta subunit 4', other_designations='BK channel beta subunit 4|BK channel subunit beta-4|BKbeta4|MaxiK channel beta-subunit 4|big potassium channel beta subunit 4|calcium-activated potassium channel, subfamily M subunit beta-4|charybdotoxin receptor subunit beta-4|hbeta4|k(VCA)beta-4|large conductance calcium-dependent potassium ion channel beta 4 subunit|maxi K channel subunit beta-4|potassium large conductance calcium-activated channel, subfamily M, beta member 4|slo-beta-4', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70366223, end_position_on_the_genomic_accession=70434292, orientation='plus', exon_count=5, OMIM=605223, no_hdr0=''),
  27346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27346, CurrentID=0, Status='live', Symbol='TMEM97', Aliases='MAC30', description='transmembrane protein 97', other_designations='meningioma-associated protein 30', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28319095, end_position_on_the_genomic_accession=28328685, orientation='plus', exon_count=3, OMIM=612912, no_hdr0=''),
  27347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27347, CurrentID=0, Status='live', Symbol='STK39', Aliases='DCHT, PASK, SPAK', description='serine/threonine kinase 39', other_designations='STE20/SPS1 homolog|Ste20-like protein kinase|proline-alanine-rich STE20-related kinase|serine threonine kinase 39 (STE20/SPS1 homolog, yeast)|serine/threonine-protein kinase 39|small intestine SPAK-like kinase|ste-20-related kinase', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=167954020, end_position_on_the_genomic_accession=168247595, orientation='minus', exon_count=21, OMIM=607648, no_hdr0=''),
  27348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27348, CurrentID=0, Status='live', Symbol='TOR1B', Aliases='DQ1', description='torsin family 1, member B (torsin B)', other_designations='torsin ATPase-1B|torsin family 1 member B', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129803148, end_position_on_the_genomic_accession=129811284, orientation='plus', exon_count=5, OMIM=608050, no_hdr0=''),
  27349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27349, CurrentID=0, Status='live', Symbol='MCAT', Aliases='FASN2C, MCT, MT, NET62, fabD', description='malonyl-CoA-acyl carrier protein transacylase', other_designations='[Acyl-carrier-protein] malonyltransferase|[acyl-carrier-protein] S-malonyltransferase|malonyl CoA:ACP acyltransferase (mitochondrial)|malonyl-CoA:acyl carrier protein transacylase, mitochondrial|mitochondrial malonyl CoA:ACP acyltransferase|mitochondrial malonyltransferase', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43132206, end_position_on_the_genomic_accession=43143397, orientation='minus', exon_count=4, OMIM=614479, no_hdr0=''),
  27350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27350, CurrentID=0, Status='live', Symbol='APOBEC3C', Aliases='A3C, APOBEC1L, ARDC2, ARDC4, ARP5, PBI, bK150C2.3', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3C', other_designations='apolipoprotein B editing enzyme catalytic polypeptide-like 3C|phorbolin I|probable DNA dC->dU-editing enzyme APOBEC-3C', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39014260, end_position_on_the_genomic_accession=39018838, orientation='plus', exon_count=4, OMIM=607750, no_hdr0=''),
  27351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27351, CurrentID=0, Status='live', Symbol='DESI1', Aliases='D15Wsu75e, DESI2, DJ347H13.4, DeSI-1, FAM152B, PPPDE2', description='desumoylating isopeptidase 1', other_designations='PPPDE peptidase domain containing 2|PPPDE peptidase domain-containing protein 2|desumoylating isopeptidase 2|family with sequence similarity 152, member B', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41598028, end_position_on_the_genomic_accession=41621061, orientation='minus', exon_count=8, OMIM=614637, no_hdr0=''),
  27352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27352, CurrentID=0, Status='live', Symbol='SGSM3', Aliases='MAP, RABGAP5, RUSC3, RUTBC3, RabGAP-5, rabGAPLP', description='small G protein signaling modulator 3', other_designations='RUN and SH3 containing 3|RUN and TBC1 domain containing 3|RUN and TBC1 domain-containing protein 3|merlin binding protein|merlin-associated protein|rab-GTPase-activating protein-like protein|small G protein signaling modulator 3 protein', map_location='22q13.1-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40370562, end_position_on_the_genomic_accession=40410289, orientation='plus', exon_count=22, OMIM=610440, no_hdr0=''),
  158427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158427, CurrentID=0, Status='live', Symbol='TSTD2', Aliases='C9orf97', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 2', other_designations='PP4189|rhodanese domain-containing protein 2|thiosulfate sulfurtransferase -like domain containing 2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97600080, end_position_on_the_genomic_accession=97633680, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  158431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158431, CurrentID=0, Status='live', Symbol='ZNF782', Aliases='', description='zinc finger protein 782', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96816699, end_position_on_the_genomic_accession=97013673, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  158434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158434, CurrentID=0, Status='live', Symbol='LOC158434', Aliases='', description='uncharacterized LOC158434', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96106661, end_position_on_the_genomic_accession=96116411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  158435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158435, CurrentID=0, Status='live', Symbol='LOC158435', Aliases='', description='uncharacterized LOC158435', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96065839, end_position_on_the_genomic_accession=96101912, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  158441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158441, CurrentID=0, Status='live', Symbol='PSMA7P', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 7 pseudogene', other_designations='proteasome alpha 7 subunit pseudogene', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95652860, end_position_on_the_genomic_accession=95653265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441728, CurrentID=0, Status='live', Symbol='LOC441728', Aliases='', description='golgin-like', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75775330, end_position_on_the_genomic_accession=75789719, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  158452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158452, CurrentID=0, Status='live', Symbol='ASS1P3', Aliases='ASSP3', description='argininosuccinate synthetase 1 pseudogene 3', other_designations='argininosuccinate synthetase pseudogene 3', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78178162, end_position_on_the_genomic_accession=78179741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646127, CurrentID=0, Status='live', Symbol='LOC646127', Aliases='', description='telomeric repeat-binding factor 1 pseudogene', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75325136, end_position_on_the_genomic_accession=75359455, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271401, CurrentID=0, Status='live', Symbol='RPS26P57', Aliases='RPS26_34_1808', description='ribosomal protein S26 pseudogene 57', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124688504, end_position_on_the_genomic_accession=124688838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266710, CurrentID=0, Status='live', Symbol='COMA', Aliases='', description='Cogan-type congential oculomotor apraxia', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=257550, no_hdr0=''),
  100419795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419795, CurrentID=0, Status='live', Symbol='LOC100419795', Aliases='', description='zinc finger, MYM-type 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142372278, end_position_on_the_genomic_accession=142373201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  158471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158471, CurrentID=0, Status='live', Symbol='PRUNE2', Aliases='BMCC1, BNIPXL, C9orf65, KIAA0367', description='prune homolog 2 (Drosophila)', other_designations='BCH motif-containing molecule at the carboxyl terminal region 1|BNIP2 motif containing molecule at the carboxyl terminal region 1|BNIP2 motif-containing molecule at the C-terminal region 1|olfaxin|truncated PRUNE2', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76611376, end_position_on_the_genomic_accession=76906220, orientation='minus', exon_count=22, OMIM=610691, no_hdr0=''),
  105373526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373526, CurrentID=0, Status='live', Symbol='LOC105373526', Aliases='', description='uncharacterized LOC105373526', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104659337, end_position_on_the_genomic_accession=104667649, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100288860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288860, CurrentID=0, Status='live', Symbol='LOC100288860', Aliases='', description='uncharacterized LOC100288860', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=34378039, end_position_on_the_genomic_accession=34399296, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27429, CurrentID=0, Status='live', Symbol='HTRA2', Aliases='OMI, PARK13, PRSS25', description='HtrA serine peptidase 2', other_designations='HtrA-like serine protease|Omi stress-regulated endoprotease|high temperature requirement protein A2|protease, serine, 25|serine protease 25|serine proteinase OMI', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74529405, end_position_on_the_genomic_accession=74533556, orientation='plus', exon_count=8, OMIM=606441, no_hdr0=''),
  27430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27430, CurrentID=0, Status='live', Symbol='MAT2B', Aliases='MAT-II, MATIIbeta, Nbla02999, SDR23E1, TGR', description='methionine adenosyltransferase II, beta', other_designations='MAT II beta|beta regulatory subunit of methionine adenosyltransferase|dTDP-4-keto-6-deoxy-D-glucose 4-reductase|putative dTDP-4-keto-6-deoxy-D-glucose 4-reductase|putative protein product of Nbla02999|short chain dehydrogenase/reductase family 23E, member 1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163503064, end_position_on_the_genomic_accession=163519353, orientation='plus', exon_count=8, OMIM=605527, no_hdr0=''),
  27433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27433, CurrentID=0, Status='live', Symbol='TOR2A', Aliases='TORP1', description='torsin family 2, member A', other_designations='salusin-beta|torsin-2A|torsin-related protein 1', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127731524, end_position_on_the_genomic_accession=127736377, orientation='minus', exon_count=6, OMIM=608052, no_hdr0=''),
  158506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158506, CurrentID=0, Status='live', Symbol='ZNF645', Aliases='CT138, HAKAIL', description='zinc finger protein 645', other_designations='E3 ubiquitin ligase', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22272913, end_position_on_the_genomic_accession=22274461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  27436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27436, CurrentID=0, Status='live', Symbol='EML4', Aliases='C2orf2, ELP120, EMAP-4, EMAPL4, ROPP120', description='echinoderm microtubule associated protein like 4', other_designations='restrictedly overexpressed proliferation-associated protein|ropp 120', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42169338, end_position_on_the_genomic_accession=42332548, orientation='plus', exon_count=26, OMIM=607442, no_hdr0=''),
  27437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27437, CurrentID=0, Status='live', Symbol='HSFY1P1', Aliases='CECR8, HSFYL1, HSFYP1, NCRNA00016', description='heat shock transcription factor, Y-linked 1 pseudogene 1', other_designations='cat eye syndrome chromosome region, candidate 8 (non-protein coding)|heat shock transcription factor, Y-linked-like 1, pseudogene|heat shock transcription factor, Y-linked-like pseudogene 1', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16827474, end_position_on_the_genomic_accession=16829335, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  27439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27439, CurrentID=0, Status='live', Symbol='CECR6', Aliases='', description='cat eye syndrome chromosome region, candidate 6', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17116299, end_position_on_the_genomic_accession=17121367, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27440, CurrentID=0, Status='live', Symbol='CECR5', Aliases='', description='cat eye syndrome chromosome region, candidate 5', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17137520, end_position_on_the_genomic_accession=17165287, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27442, CurrentID=0, Status='live', Symbol='CECR3', Aliases='', description='cat eye syndrome chromosome region, candidate 3 (non-protein coding)', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17256859, end_position_on_the_genomic_accession=17266733, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27443, CurrentID=0, Status='live', Symbol='CECR2', Aliases='', description='cat eye syndrome chromosome region, candidate 2', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17359949, end_position_on_the_genomic_accession=17558155, orientation='plus', exon_count=24, OMIM=607576, no_hdr0=''),
  27445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27445, CurrentID=0, Status='live', Symbol='PCLO', Aliases='ACZ, PCH3', description='piccolo presynaptic cytomatrix protein', other_designations='aczonin|piccolo (presynaptic cytomatrix protein)', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82754005, end_position_on_the_genomic_accession=83162881, orientation='minus', exon_count=25, OMIM=604918, no_hdr0=''),
  158521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158521, CurrentID=0, Status='live', Symbol='FMR1NB', Aliases='CT37, NY-SAR-35, NYSAR35', description='fragile X mental retardation 1 neighbor', other_designations='cancer/testis antigen 37|sarcoma antigen NY-SAR-35', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147981329, end_position_on_the_genomic_accession=148026667, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  266719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266719, CurrentID=0, Status='live', Symbol='RRDX', Aliases='', description='Radial ray deficiency', other_designations='', map_location='Xq24-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300378, no_hdr0=''),
  4578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4578, CurrentID=0, Status='live', Symbol='MT-TW', Aliases='MTTW, TRNW', description='mitochondrially encoded tRNA tryptophan', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_012920.1', start_position_on_the_genomic_accession=5512, end_position_on_the_genomic_accession=5579, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79476, CurrentID=0, Status='live', Symbol='OR5AL2P', Aliases='', description='olfactory receptor family 5 subfamily AL member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily AL, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56393730, end_position_on_the_genomic_accession=56394649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266723, CurrentID=0, Status='live', Symbol='PSMC6P1', Aliases='PSMC6P', description='proteasome 26S subunit, ATPase, 6 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 6 pseudogene|proteasome 26S ATPase subunit 6 pseudogene', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56049681, end_position_on_the_genomic_accession=56051216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288724, CurrentID=0, Status='live', Symbol='LOC100288724', Aliases='', description='transient receptor potential cation channel, subfamily C, member 6 pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150614070, end_position_on_the_genomic_accession=150614785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266724, CurrentID=0, Status='live', Symbol='HSPA9P1', Aliases='HSPA9BP, HSPA9P', description='heat shock protein family A (Hsp70) member 9 pseudogene 1', other_designations='heat shock 70kDa protein 9 (mortalin) pseudogene|heat shock 70kDa protein 9 pseudogene 1', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221961670, end_position_on_the_genomic_accession=221964457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441487, CurrentID=0, Status='live', Symbol='LOC441487', Aliases='', description='methyltransferase like 1 pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=25194023, end_position_on_the_genomic_accession=25195240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369164, CurrentID=0, Status='live', Symbol='LOC105369164', Aliases='', description='uncharacterized LOC105369164', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24790748, end_position_on_the_genomic_accession=24793118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  266726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266726, CurrentID=0, Status='live', Symbol='DYX7', Aliases='', description='dyslexia susceptibility 7', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266727, CurrentID=0, Status='live', Symbol='MDGA1', Aliases='GPIM, MAMDC3', description='MAM domain containing glycosylphosphatidylinositol anchor 1', other_designations='GPI and MAM protein|MAM domain-containing protein 3|glycosyl-phosphatidyl-inositol-MAM|glycosylphosphatidylinositol-MAM', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37632508, end_position_on_the_genomic_accession=37697990, orientation='minus', exon_count=19, OMIM=609626, no_hdr0=''),
  158572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158572, CurrentID=0, Status='live', Symbol='USP27X-AS1', Aliases='', description='USP27X antisense RNA 1 (head to head)', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49876724, end_position_on_the_genomic_accession=49879356, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  135656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135656, CurrentID=0, Status='live', Symbol='DPCR1', Aliases='PBLT', description='diffuse panbronchiolitis critical region 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30941000, end_position_on_the_genomic_accession=30954221, orientation='plus', exon_count=3, OMIM=613928, no_hdr0=''),
  100271405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271405, CurrentID=0, Status='live', Symbol='RPL12P30', Aliases='RPL12_11_1132', description='ribosomal protein L12 pseudogene 30', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30368060, end_position_on_the_genomic_accession=30368688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  158584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158584, CurrentID=0, Status='live', Symbol='FAAH2', Aliases='AMDD', description='fatty acid amide hydrolase 2', other_designations='amidase domain containing|amidase domain-containing protein|anandamide amidohydrolase 2|oleamide hydrolase 2', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57121879, end_position_on_the_genomic_accession=57489196, orientation='plus', exon_count=17, OMIM=300654, no_hdr0=''),
  158586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158586, CurrentID=0, Status='live', Symbol='ZXDB', Aliases='ZNF905, dJ83L6.1', description='zinc finger, X-linked, duplicated B', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57591836, end_position_on_the_genomic_accession=57597477, orientation='plus', exon_count=1, OMIM=300236, no_hdr0=''),
  84847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84847, CurrentID=0, Status='live', Symbol='LINC00525', Aliases='', description='long intergenic non-protein coding RNA 525', other_designations='hypothetical protein MGC16075', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47761476, end_position_on_the_genomic_accession=47766772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271406, CurrentID=0, Status='live', Symbol='RPL12P31', Aliases='RPL12_11_1136', description='ribosomal protein L12 pseudogene 31', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35997148, end_position_on_the_genomic_accession=35997766, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929924, CurrentID=0, Status='live', Symbol='LOC101929924', Aliases='', description='uncharacterized LOC101929924', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271407, CurrentID=0, Status='live', Symbol='RPL12P34', Aliases='RPL12_14_1319', description='ribosomal protein L12 pseudogene 34', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68654572, end_position_on_the_genomic_accession=68655048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100384885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100384885, CurrentID=0, Status='live', Symbol='LOC100384885', Aliases='', description='related RAS viral (r-ras) oncogene homolog 2 pseudogene', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222676774, end_position_on_the_genomic_accession=222678245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441505, CurrentID=0, Status='live', Symbol='STIP1P3', Aliases='', description='stress induced phosphoprotein 1 pseudogene 3', other_designations='stress-induced phosphoprotein 1 pseudogene 3|stress-induced-phosphoprotein 1 pseudogene', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86084296, end_position_on_the_genomic_accession=86086398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266743, CurrentID=0, Status='live', Symbol='NPAS4', Aliases='Le-PAS, NXF, PASD10, bHLHe79', description='neuronal PAS domain protein 4', other_designations='HLH-PAS transcription factor NXF|PAS domain-containing protein 10|class E basic helix-loop-helix protein 79|neuronal PAS4', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66421004, end_position_on_the_genomic_accession=66426706, orientation='plus', exon_count=8, OMIM=608554, no_hdr0=''),
  158674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158674, CurrentID=0, Status='live', Symbol='RPSAP3', Aliases='LAMR1P3, RPSA_21_1392', description='ribosomal protein SA pseudogene 3', other_designations='laminin receptor 1 pseudogene 3', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76634705, end_position_on_the_genomic_accession=76635740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369168, CurrentID=0, Status='live', Symbol='LOC105369168', Aliases='', description='uncharacterized LOC105369168', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65921949, end_position_on_the_genomic_accession=66192020, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  401717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401717, CurrentID=0, Status='live', Symbol='EEF1A1P17', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 17', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43646926, end_position_on_the_genomic_accession=43661101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266747, CurrentID=0, Status='live', Symbol='RGL4', Aliases='Rgr', description='ral guanine nucleotide dissociation stimulator-like 4', other_designations='Ral-GDS related protein Rgr|RalGDS related oncogene|hRGR|ralGDS-like 4', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23690861, end_position_on_the_genomic_accession=23699705, orientation='plus', exon_count=12, OMIM=612214, no_hdr0=''),
  158696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158696, CurrentID=0, Status='live', Symbol='LINC00889', Aliases='', description='long intergenic non-protein coding RNA 889', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138614731, end_position_on_the_genomic_accession=138617638, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369169, CurrentID=0, Status='live', Symbol='LOC105369169', Aliases='', description='uncharacterized LOC105369169', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74407553, end_position_on_the_genomic_accession=74414171, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  158724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158724, CurrentID=0, Status='live', Symbol='FAM47A', Aliases='', description='family with sequence similarity 47 member A', other_designations='family with sequence similarity 47, member A', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=34129752, end_position_on_the_genomic_accession=34132330, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928347, CurrentID=0, Status='live', Symbol='LOC101928347', Aliases='', description='uncharacterized LOC101928347', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88137206, end_position_on_the_genomic_accession=88170560, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644383, CurrentID=0, Status='live', Symbol='LOC644383', Aliases='', description='heat shock transcription factor 2 pseudogene', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55123440, end_position_on_the_genomic_accession=55124185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372985, CurrentID=0, Status='live', Symbol='LOC105372985', Aliases='', description='uncharacterized LOC105372985', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29223408, end_position_on_the_genomic_accession=29226482, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646870, CurrentID=0, Status='live', Symbol='LOC646870', Aliases='', description='centrosomal protein 57kDa pseudogene', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173362378, end_position_on_the_genomic_accession=173363890, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929401, CurrentID=0, Status='live', Symbol='LOC101929401', Aliases='', description='uncharacterized LOC101929401', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102560368, end_position_on_the_genomic_accession=102578849, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  158747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158747, CurrentID=0, Status='live', Symbol='MOSPD2', Aliases='', description='motile sperm domain containing 2', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14873405, end_position_on_the_genomic_accession=14922166, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100873194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873194, CurrentID=0, Status='live', Symbol='MTND4P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 8', other_designations='MT-ND4 pseudogene 8', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155455047, end_position_on_the_genomic_accession=155456411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285622, CurrentID=0, Status='live', Symbol='NBPF22P', Aliases='', description='neuroblastoma breakpoint family member 22, pseudogene', other_designations='neuroblastoma breakpoint family, member 22, pseudogene', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86282444, end_position_on_the_genomic_accession=86297547, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  158763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158763, CurrentID=0, Status='live', Symbol='ARHGAP36', Aliases='', description='Rho GTPase activating protein 36', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131056486, end_position_on_the_genomic_accession=131089885, orientation='plus', exon_count=14, OMIM=300937, no_hdr0=''),
  100421754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421754, CurrentID=0, Status='live', Symbol='BNIP3P2', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15349488, end_position_on_the_genomic_accession=15350019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929929, CurrentID=0, Status='live', Symbol='LOC101929929', Aliases='', description='uncharacterized LOC101929929', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56477809, end_position_on_the_genomic_accession=56483448, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131747, CurrentID=0, Status='live', Symbol='LOC100131747', Aliases='', description='small EDRK-rich factor 2 pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203353283, end_position_on_the_genomic_accession=203353787, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158787, CurrentID=0, Status='live', Symbol='RIBC1', Aliases='2610028I09Rik', description='RIB43A domain with coiled-coils 1', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53422782, end_position_on_the_genomic_accession=53431120, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  158798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158798, CurrentID=0, Status='live', Symbol='AKAP14', Aliases='AKAP28, PRKA14', description='A-kinase anchoring protein 14', other_designations='A kinase (PRKA) anchor protein 14|A-kinase anchor protein 28 kDa|A-kinase anchoring protein 28|AKAP 28|AKAP-14|protein kinase A-anchoring protein 14', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119895973, end_position_on_the_genomic_accession=119920716, orientation='plus', exon_count=7, OMIM=300462, no_hdr0=''),
  158800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158800, CurrentID=0, Status='live', Symbol='RHOXF1', Aliases='OTEX, PEPP1', description='Rhox homeobox family member 1', other_designations='PEPP subfamily gene 1|Rhox homeobox family, member 1|ovary-, testis- and epididymis-expressed gene protein|paired-like homeobox protein PEPP-1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120109049, end_position_on_the_genomic_accession=120117709, orientation='minus', exon_count=4, OMIM=300446, no_hdr0=''),
  158801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158801, CurrentID=0, Status='live', Symbol='NKAPP1', Aliases='CXorf42', description='NFKB activating protein pseudogene 1', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120236452, end_position_on_the_genomic_accession=120245267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  158809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158809, CurrentID=0, Status='live', Symbol='MAGEB6', Aliases='CT3.4, MAGE-B6A, MAGEB6', description='MAGE family member B6', other_designations='MAGE-B6 antigen|cancer/testis antigen 3.4|cancer/testis antigen family 3, member 4|melanoma antigen family B, 6|melanoma antigen family B6', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26192440, end_position_on_the_genomic_accession=26195646, orientation='plus', exon_count=2, OMIM=300467, no_hdr0=''),
  158830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158830, CurrentID=0, Status='live', Symbol='CXorf65', Aliases='', description='chromosome X open reading frame 65', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71103889, end_position_on_the_genomic_accession=71106972, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  284618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284618, CurrentID=0, Status='live', Symbol='RUSC1-AS1', Aliases='C1orf104', description='RUSC1 antisense RNA 1', other_designations='RUSC1 antisense RNA 1 (non-protein coding)|RUSC1 antisense gene protein 1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155320460, end_position_on_the_genomic_accession=155324147, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  158833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158833, CurrentID=0, Status='live', Symbol='AWAT1', Aliases='DGA2, DGAT2L3', description='acyl-CoA wax alcohol acyltransferase 1', other_designations='diacyl-glycerol acyltransferase 2|diacylglycerol O-acyltransferase 2-like 3|diacylglycerol O-acyltransferase 2-like protein 3|diacylglycerol acyltransferase 2|long-chain-alcohol O-fatty-acyltransferase 1', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70234655, end_position_on_the_genomic_accession=70240661, orientation='plus', exon_count=7, OMIM=300924, no_hdr0=''),
  158835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158835, CurrentID=0, Status='live', Symbol='AWAT2', Aliases='ARAT, DC4, DGAT2L4, MFAT, WS', description='acyl-CoA wax alcohol acyltransferase 2', other_designations='acyl-CoA retinol O-fatty-acyltransferase|diacylglycerol O-acyltransferase 2-like protein 4|diacylglycerol O-acyltransferase candidate 4|long-chain-alcohol O-fatty-acyltransferase 2|multifunctional O-acyltransferase|retinol O-fatty-acyltransferase', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70040542, end_position_on_the_genomic_accession=70049965, orientation='minus', exon_count=8, OMIM=300925, no_hdr0=''),
  401718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401718, CurrentID=0, Status='live', Symbol='OR5BK1P', Aliases='', description='olfactory receptor family 5 subfamily BK member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BK, member 1 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48355659, end_position_on_the_genomic_accession=48356545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158863, CurrentID=0, Status='live', Symbol='LOC158863', Aliases='', description='uncharacterized LOC158863', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158866, CurrentID=0, Status='live', Symbol='ZDHHC15', Aliases='MRX91', description='zinc finger, DHHC-type containing 15', other_designations='DHHC-15|zinc finger DHHC domain-containing protein 15|zinc finger, DHHC domain containing 15', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75368427, end_position_on_the_genomic_accession=75523626, orientation='minus', exon_count=13, OMIM=300576, no_hdr0=''),
  158880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158880, CurrentID=0, Status='live', Symbol='USP51', Aliases='', description='ubiquitin specific peptidase 51', other_designations='deubiquitinating enzyme 51|ubiquitin thioesterase 51|ubiquitin thiolesterase 51|ubiquitin-specific-processing protease 51', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55484616, end_position_on_the_genomic_accession=55489868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646875, CurrentID=0, Status='live', Symbol='RPL12P2', Aliases='RPL12_6_688, dJ441G21.1', description='ribosomal protein L12 pseudogene 2', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37091221, end_position_on_the_genomic_accession=37091845, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284620, CurrentID=0, Status='live', Symbol='LOC284620', Aliases='', description='claudin 6 pseudogene', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155590404, end_position_on_the_genomic_accession=155591318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158905, CurrentID=0, Status='live', Symbol='PHF10P1', Aliases='XAP135P', description='PHD finger protein 10 pseudogene 1', other_designations='XAP135 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155291939, end_position_on_the_genomic_accession=155293441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266783, CurrentID=0, Status='live', Symbol='PSMD2P1', Aliases='', description='proteasome 26S subunit, non-ATPase, 2 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 2 pseudogene|proteasome 26S non-ATPase subunit 2 pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240142510, end_position_on_the_genomic_accession=240145019, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266785, CurrentID=0, Status='live', Symbol='NUP50P2', Aliases='', description='nucleoporin 50 pseudogene 2', other_designations='nucleoporin 50kDa pseudogene', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25139844, end_position_on_the_genomic_accession=25141539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369176, CurrentID=0, Status='live', Symbol='LOC105369176', Aliases='', description='putative tripartite motif-containing protein 64B', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89821830, end_position_on_the_genomic_accession=89826934, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  266786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266786, CurrentID=0, Status='live', Symbol='NUP50P3', Aliases='', description='nucleoporin 50 pseudogene 3', other_designations='nucleoporin 50 kDa pseudogene|nucleoporin 50kDa pseudogene', map_location='5p14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20302004, end_position_on_the_genomic_accession=20305825, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158931, CurrentID=0, Status='live', Symbol='TCEAL6', Aliases='Tceal3, WEX2', description='transcription elongation factor A (SII)-like 6', other_designations='TCEA-like protein 6|transcription elongation factor A (SII)-like 3|transcription elongation factor S-II protein-like 6', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102139961, end_position_on_the_genomic_accession=102142416, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373541, CurrentID=0, Status='live', Symbol='LOC105373541', Aliases='', description='uncharacterized LOC105373541', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108594692, end_position_on_the_genomic_accession=108613386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271417, CurrentID=0, Status='live', Symbol='RPL17P38', Aliases='RPL17_22_1275', description='ribosomal protein L17 pseudogene 38', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107719564, end_position_on_the_genomic_accession=107720163, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266789, CurrentID=0, Status='live', Symbol='EPPS', Aliases='', description='Epilepsy, partial, with pericentral spikes', other_designations='', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607221, no_hdr0=''),
  158948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158948, CurrentID=0, Status='live', Symbol='LOC158948', Aliases='', description='protein kinase C, iota pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101534636, end_position_on_the_genomic_accession=101539446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  266790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266790, CurrentID=0, Status='live', Symbol='HCA1', Aliases='AH, HCA', description='Hypercalciuria, absorptive, 1', other_designations='', map_location='4q33-qter', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607258, no_hdr0=''),
  105369177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369177, CurrentID=0, Status='live', Symbol='LOC105369177', Aliases='', description='uncharacterized LOC105369177', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111517642, end_position_on_the_genomic_accession=111525432, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100419811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419811, CurrentID=0, Status='live', Symbol='LOC100419811', Aliases='', description='interleukin 9 receptor pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17012404, end_position_on_the_genomic_accession=17012932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  158960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158960, CurrentID=0, Status='live', Symbol='CH17-340M24.3', Aliases='', description='uncharacterized protein BC009467', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154424380, end_position_on_the_genomic_accession=154428479, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26495, CurrentID=0, Status='live', Symbol='OR9A1P', Aliases='HSHTPCRX06, HTPCRX06, OR9A1, OR9A5P', description='olfactory receptor family 9 subfamily A member 1 pseudogene', other_designations='olfactory receptor, family 9, subfamily A, member 1 pseudogene|olfactory receptor, family 9, subfamily A, member 5 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141887148, end_position_on_the_genomic_accession=141887936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791082, CurrentID=0, Status='live', Symbol='DBIP3', Aliases='', description='diazepam binding inhibitor (GABA receptor modulator, acyl-CoA binding protein) pseudogene 3', other_designations='diazepam binding inhibitor-like pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4130592, end_position_on_the_genomic_accession=4130995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369178, CurrentID=0, Status='live', Symbol='LOC105369178', Aliases='', description='uncharacterized LOC105369178', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125980474, end_position_on_the_genomic_accession=126368825, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  158983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158983, CurrentID=0, Status='live', Symbol='H2BFWT', Aliases='', description='H2B histone family member W, testis specific', other_designations='H2B histone family, member W, testis-specific', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104011147, end_position_on_the_genomic_accession=104013687, orientation='minus', exon_count=3, OMIM=300507, no_hdr0=''),
  791085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791085, CurrentID=0, Status='live', Symbol='DEL17Q21.31', Aliases='C17DELq21.31', description='Microdeletion 17q21.31 syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610443, no_hdr0=''),
  105373543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373543, CurrentID=0, Status='live', Symbol='LOC105373543', Aliases='', description='uncharacterized LOC105373543', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109065023, end_position_on_the_genomic_accession=109069925, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  791086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791086, CurrentID=0, Status='live', Symbol='CTPP5', Aliases='', description='Cataract, posterior polar, 5', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610634, no_hdr0=''),
  100271419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271419, CurrentID=0, Status='live', Symbol='RPL17P46', Aliases='RPL17_27_1601', description='ribosomal protein L17 pseudogene 46', other_designations='', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50823577, end_position_on_the_genomic_accession=50824104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  4655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4655, CurrentID=0, Status='live', Symbol='MYSA', Aliases='', description='myasthenic (Lambert-Eaton) syndrome antigen A', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  159007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159007, CurrentID=0, Status='live', Symbol='LOC159007', Aliases='', description='aldo-keto reductase family 7, member A2 (aflatoxin aldehyde reductase) pseudogene', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118877883, end_position_on_the_genomic_accession=118878408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369179, CurrentID=0, Status='live', Symbol='LOC105369179', Aliases='', description='uncharacterized LOC105369179', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20731055, end_position_on_the_genomic_accession=20754369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  159013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159013, CurrentID=0, Status='live', Symbol='CXorf38', Aliases='', description='chromosome X open reading frame 38', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40626921, end_position_on_the_genomic_accession=40647567, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105373544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373544, CurrentID=0, Status='live', Symbol='LOC105373544', Aliases='', description='uncharacterized LOC105373544', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109045835, end_position_on_the_genomic_accession=109061999, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79492, CurrentID=0, Status='live', Symbol='OR7E97P', Aliases='', description='olfactory receptor family 7 subfamily E member 97 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 97 pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125747129, end_position_on_the_genomic_accession=125748099, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791091, CurrentID=0, Status='live', Symbol='LOC791091', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48279390, end_position_on_the_genomic_accession=48282138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271420, CurrentID=0, Status='live', Symbol='RPL18AP14', Aliases='RPL18A_9_1800', description='ribosomal protein L18a pseudogene 14', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104803869, end_position_on_the_genomic_accession=104804497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791092, CurrentID=0, Status='live', Symbol='LOC791092', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48308191, end_position_on_the_genomic_accession=48309690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791093, CurrentID=0, Status='live', Symbol='LOC791093', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48356989, end_position_on_the_genomic_accession=48360459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369180, CurrentID=0, Status='live', Symbol='LOC105369180', Aliases='', description='uncharacterized LOC105369180', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22656181, end_position_on_the_genomic_accession=22664431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419814, CurrentID=0, Status='live', Symbol='LOC100419814', Aliases='', description='ankyrin repeat domain 49 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224499380, end_position_on_the_genomic_accession=224500100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  791095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791095, CurrentID=0, Status='live', Symbol='LOC791095', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48412273, end_position_on_the_genomic_accession=48416084, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791096, CurrentID=0, Status='live', Symbol='LOC791096', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48446260, end_position_on_the_genomic_accession=48448909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129129, CurrentID=0, Status='live', Symbol='LOC100129129', Aliases='', description='uncharacterized LOC100129129', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  791097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791097, CurrentID=0, Status='live', Symbol='LOC791097', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52655177, end_position_on_the_genomic_accession=52657894, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791098, CurrentID=0, Status='live', Symbol='LOC791098', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52709073, end_position_on_the_genomic_accession=52710908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369181, CurrentID=0, Status='live', Symbol='LOC105369181', Aliases='', description='uncharacterized LOC105369181', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61577286, end_position_on_the_genomic_accession=61580198, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  158055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158055, CurrentID=0, Status='live', Symbol='C9orf163', Aliases='', description='chromosome 9 open reading frame 163', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136483495, end_position_on_the_genomic_accession=136486067, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  140597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140597, CurrentID=0, Status='live', Symbol='TCEAL2', Aliases='MY0876G05, WEX1, my048', description='transcription elongation factor A (SII)-like 2', other_designations='TCEA-like protein 2|transcription elongation factor S-II protein-like 2', map_location='Xq22.1-q22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102125688, end_position_on_the_genomic_accession=102127712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  266812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266812, CurrentID=0, Status='live', Symbol='NAP1L5', Aliases='DRLM', description='nucleosome assembly protein 1-like 5', other_designations='down-regulated in liver malignancy', map_location='4q21-q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88695915, end_position_on_the_genomic_accession=88697872, orientation='minus', exon_count=1, OMIM=612203, no_hdr0=''),
  159090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159090, CurrentID=0, Status='live', Symbol='FAM122B', Aliases='SPACIA2', description='family with sequence similarity 122B', other_designations='RP11-308B5.5|synoviocyte proliferation associated in collagen-induced arthritis 2', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134769566, end_position_on_the_genomic_accession=134797232, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  159091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159091, CurrentID=0, Status='live', Symbol='FAM122C', Aliases='', description='family with sequence similarity 122C', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134796413, end_position_on_the_genomic_accession=134855135, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  402469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402469, CurrentID=0, Status='live', Symbol='LOC402469', Aliases='FKSG50, FKSG53, FKSG54', description='uncharacterized LOC402469', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441750, CurrentID=0, Status='live', Symbol='LOC441750', Aliases='', description='chromosome 12 open reading frame 49 pseudogene', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16952503, end_position_on_the_genomic_accession=16953468, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  159119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159119, CurrentID=0, Status='live', Symbol='HSFY2', Aliases='HSF2L, HSFY', description='heat shock transcription factor, Y-linked 2', other_designations='heat shock transcription factor 2-like protein', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18731440, end_position_on_the_genomic_accession=18773735, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  159125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159125, CurrentID=0, Status='live', Symbol='RBMY2EP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member E pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21395148, end_position_on_the_genomic_accession=21401562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369183, CurrentID=0, Status='live', Symbol='LOC105369183', Aliases='', description='uncharacterized LOC105369183', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33966104, end_position_on_the_genomic_accession=33975873, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  284628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284628, CurrentID=0, Status='live', Symbol='LOC284628', Aliases='', description='uncharacterized LOC284628', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401720, CurrentID=0, Status='live', Symbol='FIGNL2', Aliases='', description='fidgetin-like 2', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51817892, end_position_on_the_genomic_accession=51848760, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105373548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373548, CurrentID=0, Status='live', Symbol='LOC105373548', Aliases='', description='uncharacterized LOC105373548', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110035839, end_position_on_the_genomic_accession=110039614, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369184, CurrentID=0, Status='live', Symbol='LOC105369184', Aliases='', description='uncharacterized LOC105369184', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32290341, end_position_on_the_genomic_accession=32310965, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  159162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159162, CurrentID=0, Status='live', Symbol='RBMY2FP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member F pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22308859, end_position_on_the_genomic_accession=22316205, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  159163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159163, CurrentID=0, Status='live', Symbol='RBMY1F', Aliases='YRRM2', description='RNA binding motif protein, Y-linked, family 1, member F', other_designations='y chromosome RNA recognition motif 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22168181, end_position_on_the_genomic_accession=22183008, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  4682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4682, CurrentID=0, Status='live', Symbol='NUBP1', Aliases='NBP, NBP1, NBP35', description='nucleotide binding protein 1', other_designations='NBP 1|nucleotide binding protein (e.coli MinD like)|nucleotide binding protein 1 (E.coli MinD like)|nucleotide binding protein 1 (MinD homolog, E. coli)|nucleotide-binding protein 1', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10743321, end_position_on_the_genomic_accession=10769351, orientation='plus', exon_count=11, OMIM=600280, no_hdr0=''),
  791115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791115, CurrentID=0, Status='live', Symbol='PWRN2', Aliases='NCRNA00199', description='Prader-Willi region non-protein coding RNA 2', other_designations='Prader-Willi region non-protein-coding RNA 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24164779, end_position_on_the_genomic_accession=24169906, orientation='minus', exon_count=1, OMIM=611217, no_hdr0=''),
  105373549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373549, CurrentID=0, Status='live', Symbol='LOC105373549', Aliases='', description='uncharacterized LOC105373549', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110039635, end_position_on_the_genomic_accession=110041438, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  550113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550113, CurrentID=0, Status='live', Symbol='LOC550113', Aliases='', description='uncharacterized LOC550113', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68053508, end_position_on_the_genomic_accession=68073233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284630, CurrentID=0, Status='live', Symbol='LOC284630', Aliases='', description='uncharacterized LOC284630', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  159195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159195, CurrentID=0, Status='live', Symbol='USP54', Aliases='C10orf29, bA137L10.3, bA137L10.4', description='ubiquitin specific peptidase 54', other_designations='inactive ubiquitin-specific peptidase 54|ubiquitin specific protease 54|ubiquitin specific proteinase 54', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73497538, end_position_on_the_genomic_accession=73591342, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  642521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642521, CurrentID=0, Status='live', Symbol='C1DP3', Aliases='bA369J21.2', description='C1D nuclear receptor corepressor pseudogene 3', other_designations='C1D nuclear receptor co-repressor pseudogene 3', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80024379, end_position_on_the_genomic_accession=80025241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  791121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791121, CurrentID=0, Status='live', Symbol='SHFL1', Aliases='SHFLD', description='Split-hand/foot malformation with long bone deficiency 1', other_designations='', map_location='1q42.2-q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=119100, no_hdr0=''),
  100288884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288884, CurrentID=0, Status='live', Symbol='LOC100288884', Aliases='', description='uncharacterized LOC100288884', other_designations='', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=10600301, end_position_on_the_genomic_accession=10607057, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  791122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791122, CurrentID=0, Status='live', Symbol='KTWS', Aliases='KTS', description='Klippel-Trenaunay-Weber syndrome', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=149000, no_hdr0=''),
  791123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791123, CurrentID=0, Status='live', Symbol='SHFLD2', Aliases='', description='Split-hand/foot malformation with long deficiency 2', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610685, no_hdr0=''),
  105369186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369186, CurrentID=0, Status='live', Symbol='LOC105369186', Aliases='', description='beta-glucuronidase-like protein SMA4', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373551, CurrentID=0, Status='live', Symbol='LOC105373551', Aliases='', description='uncharacterized LOC105373551', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110316888, end_position_on_the_genomic_accession=110347069, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  791126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791126, CurrentID=0, Status='live', Symbol='LOC791126', Aliases='', description='WDR45-like pseudogene', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36273542, end_position_on_the_genomic_accession=36274800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271427, CurrentID=0, Status='live', Symbol='RPL21P95', Aliases='RPL21_40_1159', description='ribosomal protein L21 pseudogene 95', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77868696, end_position_on_the_genomic_accession=77869235, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369187, CurrentID=0, Status='live', Symbol='LOC105369187', Aliases='', description='uncharacterized LOC105369187', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65544238, end_position_on_the_genomic_accession=65557311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  284632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284632, CurrentID=0, Status='live', Symbol='LOC284632', Aliases='', description='uncharacterized LOC284632', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24200240, end_position_on_the_genomic_accession=24211690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140603, CurrentID=0, Status='live', Symbol='COX6B1P1', Aliases='COX6BP-1, COX6BP1', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 1', other_designations='cytochrome c oxidase subunit VIb pseudogene 1 (HCOX52)', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149053867, end_position_on_the_genomic_accession=149054307, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373552, CurrentID=0, Status='live', Symbol='LOC105373552', Aliases='', description='uncharacterized LOC105373552', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110448572, end_position_on_the_genomic_accession=110452647, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271428, CurrentID=0, Status='live', Symbol='RPL21P96', Aliases='RPL21_41_1169', description='ribosomal protein L21 pseudogene 96', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103045203, end_position_on_the_genomic_accession=103045739, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369188, CurrentID=0, Status='live', Symbol='LOC105369188', Aliases='', description='uncharacterized LOC105369188', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37986323, end_position_on_the_genomic_accession=37988098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101978719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101978719, CurrentID=0, Status='live', Symbol='LINC00970', Aliases='', description='long intergenic non-protein coding RNA 970', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168903905, end_position_on_the_genomic_accession=169087005, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  159296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159296, CurrentID=0, Status='live', Symbol='NKX2-3', Aliases='CSX3, NK2.3, NKX2.3, NKX2C, NKX4-3', description='NK2 homeobox 3', other_designations='NK2 transcription factor related, locus 3|homeobox protein NK-2 homolog C', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99532933, end_position_on_the_genomic_accession=99536524, orientation='plus', exon_count=3, OMIM=606727, no_hdr0=''),
  28227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28227, CurrentID=0, Status='live', Symbol='PPP2R3B', Aliases='NYREN8, PPP2R3L, PPP2R3LY, PR48', description="protein phosphatase 2 regulatory subunit B'', beta", other_designations="NY-REN-8 antigen|PP2A, subunit B, PR48 isoform|protein phosphatase 2 (formerly 2A), regulatory subunit B'', beta|protein phosphatase 2, regulatory subunit B'', beta|serine/threonine protein phosphatase 2A, 48kDa regulatory subunit B", map_location='Xp22.33; Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=333933, end_position_on_the_genomic_accession=386955, orientation='minus', exon_count=14, OMIM=300339, no_hdr0=''),
  28231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28231, CurrentID=0, Status='live', Symbol='SLCO4A1', Aliases='OATP-E, OATP1, OATP4A1, OATPE, OATPRP1, POAT, SLC21A12', description='solute carrier organic anion transporter family member 4A1', other_designations='OATP-RP1|colon organic anion transporter|organic anion transporter polypeptide-related protein 1|organic anion transporting polypeptide E|organic anion-transporting polypeptide E|sodium-independent organic anion transporter E|solute carrier family 21 (organic anion transporter), member 12|solute carrier family 21 member 12|solute carrier organic anion transporter family, member 4A1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62642445, end_position_on_the_genomic_accession=62685830, orientation='plus', exon_count=15, OMIM=612436, no_hdr0=''),
  28232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28232, CurrentID=0, Status='live', Symbol='SLCO3A1', Aliases='OATP-D, OATP3A1, OATPD, SLC21A11', description='solute carrier organic anion transporter family member 3A1', other_designations='OATP-RP3|OATPRP3|PGE1 transporter|organic anion transporter polypeptide-related protein 3|organic anion-transporting polypeptide D|sodium-independent organic anion transporter D|solute carrier family 21 (organic anion transporter), member 11|solute carrier family 21 member 11|solute carrier organic anion transporter family, member 3A1', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91853708, end_position_on_the_genomic_accession=92172435, orientation='plus', exon_count=15, OMIM=612435, no_hdr0=''),
  28234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28234, CurrentID=0, Status='live', Symbol='SLCO1B3', Aliases='HBLRR, LST-2, LST-3TM13, LST3, OATP-8, OATP1B3, OATP8, SLC21A8', description='solute carrier organic anion transporter family member 1B3', other_designations='liver-specific organic anion transporter 2|liver-specific organic anion transporter 3TM13|organic anion transporter 8|organic anion transporter LST-3c|organic anion-transporting polypeptide 8|solute carrier family 21 (organic anion transporter), member 8|solute carrier organic anion transporter family, member 1B3', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20810704, end_position_on_the_genomic_accession=20916911, orientation='plus', exon_count=16, OMIM=605495, no_hdr0=''),
  105369189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369189, CurrentID=0, Status='live', Symbol='LOC105369189', Aliases='', description='uncharacterized LOC105369189', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271430, CurrentID=0, Status='live', Symbol='RPL21P100', Aliases='RPL21_42_1206', description='ribosomal protein L21 pseudogene 100', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12155223, end_position_on_the_genomic_accession=12155804, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369190, CurrentID=0, Status='live', Symbol='LOC105369190', Aliases='', description='uncharacterized LOC105369190', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=779359, end_position_on_the_genomic_accession=780898, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373555, CurrentID=0, Status='live', Symbol='LOC105373555', Aliases='', description='uncharacterized LOC105373555', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110730384, end_position_on_the_genomic_accession=110732444, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  159371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159371, CurrentID=0, Status='live', Symbol='SLC35G1', Aliases='C10orf60, TMEM20', description='solute carrier family 35 member G1', other_designations='partner of STIM1|solute carrier family 35, member G1|transmembrane protein 20', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93893973, end_position_on_the_genomic_accession=93914255, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  28300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28300, CurrentID=0, Status='live', Symbol='IGHV3OR16-16', Aliases='IGHV3OR1616', description='immunoglobulin heavy variable 3/OR16-16 (pseudogene)', other_designations='IGHV3/OR16-16|immunoglobulin heavy variable 3/OR16-16 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28301, CurrentID=0, Status='live', Symbol='IGHV3OR16-15', Aliases='IGHV3OR1615', description='immunoglobulin heavy variable 3/OR16-15 (pseudogene)', other_designations='IGHV3/OR16-15|immunoglobulin heavy variable 3/OR16-15 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28304, CurrentID=0, Status='live', Symbol='IGHV3OR16-12', Aliases='IGHV3OR1612', description='immunoglobulin heavy variable 3/OR16-12 (non-functional)', other_designations='IGHV3/OR16-12', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33802728, end_position_on_the_genomic_accession=33803311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  28305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28305, CurrentID=0, Status='live', Symbol='IGHV3OR16-11', Aliases='IGHV3OR1611', description='immunoglobulin heavy variable 3/OR16-11 (pseudogene)', other_designations='IGHV3/OR16-11|immunoglobulin heavy variable 3/OR16-11 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33858802, end_position_on_the_genomic_accession=33859352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28306, CurrentID=0, Status='live', Symbol='IGHV3OR16-10', Aliases='IGHV3OR1610', description='immunoglobulin heavy variable 3/OR16-10 (non-functional)', other_designations='IGHV3/OR16-10', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=33006369, end_position_on_the_genomic_accession=33006920, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28307, CurrentID=0, Status='live', Symbol='IGHV3OR16-9', Aliases='IGHV3OR169', description='immunoglobulin heavy variable 3/OR16-9 (non-functional)', other_designations='IGHV3/OR16-9', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28309, CurrentID=0, Status='live', Symbol='IGHV3OR16-7', Aliases='IGHV3/OR16-7, IGHV3OR167', description='immunoglobulin heavy variable 3/OR16-7 (pseudogene)', other_designations='immunoglobulin heavy variable 3/OR16-7 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33938339, end_position_on_the_genomic_accession=33938639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28312, CurrentID=0, Status='live', Symbol='IGHV1OR16-4', Aliases='IGHV1OR164', description='immunoglobulin heavy variable 1/OR16-4 (pseudogene)', other_designations='IGHV1/OR16-4|immunoglobulin heavy variable 1/OR16-4 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28313, CurrentID=0, Status='live', Symbol='IGHV1OR16-3', Aliases='IGHV1OR163', description='immunoglobulin heavy variable 1/OR16-3 (pseudogene)', other_designations='IGHV1/OR16-3|immunoglobulin heavy variable 1/OR16-3 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28314, CurrentID=0, Status='live', Symbol='IGHV1OR16-2', Aliases='IGHV1OR162', description='immunoglobulin heavy variable 1/OR16-2 (pseudogene)', other_designations='IGHV1/OR16-2|immunoglobulin heavy variable 1/OR16-2 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28315, CurrentID=0, Status='live', Symbol='IGHV1OR16-1', Aliases='IGHV1OR161', description='immunoglobulin heavy variable 1/OR16-1 (pseudogene)', other_designations='IGHV1/OR16-1|immunoglobulin heavy variable 1/OR16-1 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32035070, end_position_on_the_genomic_accession=32035326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28316, CurrentID=0, Status='live', Symbol='CDH20', Aliases='CDH7L3, Cdh7', description='cadherin 20, type 2', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61333582, end_position_on_the_genomic_accession=61555773, orientation='plus', exon_count=12, OMIM=605807, no_hdr0=''),
  28317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28317, CurrentID=0, Status='live', Symbol='IGHV4OR15-8', Aliases='IGHV4OR158, VSIG6', description='immunoglobulin heavy variable 4/OR15-8 (non-functional)', other_designations='IGHV4/OR15-8|V-set and immunoglobulin domain containing 6', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28318, CurrentID=0, Status='live', Symbol='IGHV3OR15-7', Aliases='IGHV3/OR15-7', description='immunoglobulin heavy variable 3/OR15-7 (pseudogene)', other_designations='immunoglobulin heavy variable 3/OR15-7 pseudogene', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28320, CurrentID=0, Status='live', Symbol='IGHV1OR15-6', Aliases='IGHV1OR156', description='immunoglobulin heavy variable 1/OR15-6 (pseudogene)', other_designations='IGHV1/OR15-6|immunoglobulin heavy variable 1/OR15-6 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28326, CurrentID=0, Status='live', Symbol='IGHD5OR15-5B', Aliases='IGHD5OR155B', description='immunoglobulin heavy diversity 5/OR15-5B (non-functional)', other_designations='IGHD5/OR15-5B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28327, CurrentID=0, Status='live', Symbol='IGHD5OR15-5A', Aliases='IGHD5OR155A', description='immunoglobulin heavy diversity 5/OR15-5A (non-functional)', other_designations='IGHD5/OR15-5A', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28328, CurrentID=0, Status='live', Symbol='IGHD4OR15-4B', Aliases='IGHD4OR154B', description='immunoglobulin heavy diversity 4/OR15-4B (non-functional)', other_designations='IGHD4/OR15-4B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28329, CurrentID=0, Status='live', Symbol='IGHD4OR15-4A', Aliases='IGHD4OR154A', description='immunoglobulin heavy diversity 4/OR15-4A (non-functional)', other_designations='IGHD4/OR15-4A', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28330, CurrentID=0, Status='live', Symbol='IGHD3OR15-3B', Aliases='IGHD3OR153B', description='immunoglobulin heavy diversity 3/OR15-3B (non-functional)', other_designations='IGHD3/OR15-3B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28331, CurrentID=0, Status='live', Symbol='IGHD3OR15-3A', Aliases='IGHD3OR153A', description='immunoglobulin heavy diversity 3/OR15-3A (non-functional)', other_designations='IGHD3/OR15-3A', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28332, CurrentID=0, Status='live', Symbol='IGHD2OR15-2B', Aliases='IGHD2OR152B', description='immunoglobulin heavy diversity 2/OR15-2B (non-functional)', other_designations='IGHD2/OR15-2B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28333, CurrentID=0, Status='live', Symbol='IGHD2OR15-2A', Aliases='IGHD2OR152A', description='immunoglobulin heavy diversity 2/OR15-2A (non-functional)', other_designations='IGHD2/OR15-2A', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28334, CurrentID=0, Status='live', Symbol='IGHD1OR15-1B', Aliases='IGHD1OR151B', description='immunoglobulin heavy diversity 1/OR15-1B (non-functional)', other_designations='IGHD1/OR15-1B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28335, CurrentID=0, Status='live', Symbol='IGHD1OR15-1A', Aliases='IGHD1OR151A', description='immunoglobulin heavy diversity 1/OR15-1A (non-functional)', other_designations='IGHD1/OR15-1A', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28337, CurrentID=0, Status='live', Symbol='IGHVIV-44-1', Aliases='(IV)-44, 4-44.1P, IGHVIV441', description='immunoglobulin heavy variable (IV)-44-1 (pseudogene)', other_designations='IGHV(IV)-44-1|immunoglobulin heavy variable (IV)-44-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106489342, end_position_on_the_genomic_accession=106489770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28338, CurrentID=0, Status='live', Symbol='IGHVIII-82', Aliases='3-82P, IGHVIII82', description='immunoglobulin heavy variable (III)-82 (pseudogene)', other_designations='IGHV(III)-82|immunoglobulin heavy variable (III)-82 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106879553, end_position_on_the_genomic_accession=106879844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28339, CurrentID=0, Status='live', Symbol='IGHVIII-76-1', Aliases='3-76.1P, IGHVIII761', description='immunoglobulin heavy variable (III)-76-1 (pseudogene)', other_designations='IGHV(III)-76-1|immunoglobulin heavy variable (III)-76-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106831763, end_position_on_the_genomic_accession=106832069, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28340, CurrentID=0, Status='live', Symbol='IGHVIII-67-4', Aliases='3-67.4P, IGHVIII674', description='immunoglobulin heavy variable (III)-67-4 (pseudogene)', other_designations='IGHV(III)-67-4|immunoglobulin heavy variable (III)-67-4 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106695092, end_position_on_the_genomic_accession=106695397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28341, CurrentID=0, Status='live', Symbol='IGHVIII-67-3', Aliases='3-67.3P, IGHV(III)-6, IGHVIII673', description='immunoglobulin heavy variable (III)-67-3 (pseudogene)', other_designations='IGHV(III)-67-3|immunoglobulin heavy variable (III)-67-3 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106692614, end_position_on_the_genomic_accession=106692888, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28342, CurrentID=0, Status='live', Symbol='IGHVIII-67-2', Aliases='3-67.2P, IGHVIII672', description='immunoglobulin heavy variable (III)-67-2 (pseudogene)', other_designations='IGHV(III)-67-2|immunoglobulin heavy variable (III)-67-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106687102, end_position_on_the_genomic_accession=106687200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28343, CurrentID=0, Status='live', Symbol='IGHVIII-51-1', Aliases='3-51.1P, IGHVIII511', description='immunoglobulin heavy variable (III)-51-1 (pseudogene)', other_designations='IGHV(III)-51-1|immunoglobulin heavy variable (III)-51-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106583501, end_position_on_the_genomic_accession=106583807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28344, CurrentID=0, Status='live', Symbol='IGHVIII-47-1', Aliases='3-47.1P, IGHVIII471', description='immunoglobulin heavy variable (III)-47-1 (pseudogene)', other_designations='IGHV(III)-47-1|immunoglobulin heavy variable (III)-47-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106531133, end_position_on_the_genomic_accession=106531436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28345, CurrentID=0, Status='live', Symbol='IGHVIII-44', Aliases='3-44P, IGHVIII44', description='immunoglobulin heavy variable (III)-44 (pseudogene)', other_designations='IGHV(III)-44|immunoglobulin heavy variable (III)-44 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106478217, end_position_on_the_genomic_accession=106478393, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28346, CurrentID=0, Status='live', Symbol='IGHVIII-38-1', Aliases='3-38.1P, 38-1, IGHVIII381', description='immunoglobulin heavy variable (III)-38-1 (pseudogene)', other_designations='IGHV(III)-38-1|immunoglobulin heavy variable (III)-38-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106418018, end_position_on_the_genomic_accession=106418310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28347, CurrentID=0, Status='live', Symbol='IGHVIII-26-1', Aliases='3-26.1P, IGHVIII261', description='immunoglobulin heavy variable (III)-26-1 (pseudogene)', other_designations='IGHV(III)-26-1|immunoglobulin heavy variable (III)-26-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106309406, end_position_on_the_genomic_accession=106309712, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28348, CurrentID=0, Status='live', Symbol='IGHVIII-25-1', Aliases='3-25.1P, IGHVIII251', description='immunoglobulin heavy variable (III)-25-1 (pseudogene)', other_designations='IGHV(III)-25-1|immunoglobulin heavy variable (III)-25-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106293566, end_position_on_the_genomic_accession=106293807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28349, CurrentID=0, Status='live', Symbol='IGHVIII-22-2', Aliases='3-22.2P, IGHVIII222', description='immunoglobulin heavy variable (III)-22-2 (pseudogene)', other_designations='IGHV(III)-22-2|immunoglobulin heavy variable (III)-22-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106264686, end_position_on_the_genomic_accession=106264873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28350, CurrentID=0, Status='live', Symbol='IGHVIII-16-1', Aliases='3-16.1P, IGHVIII161', description='immunoglobulin heavy variable (III)-16-1 (pseudogene)', other_designations='IGHV(III)-16-1|immunoglobulin heavy variable (III)-16-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106170746, end_position_on_the_genomic_accession=106171051, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28351, CurrentID=0, Status='live', Symbol='IGHVIII-13-1', Aliases='3-13.1P, IGHVIII131', description='immunoglobulin heavy variable (III)-13-1 (pseudogene)', other_designations='IGHV(III)-13-1|immunoglobulin heavy variable (III)-13-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106142286, end_position_on_the_genomic_accession=106142584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28352, CurrentID=0, Status='live', Symbol='IGHVIII-11-1', Aliases='3-11.1P, IGHVIII111', description='immunoglobulin heavy variable (III)-11-1 (pseudogene)', other_designations='IGHV(III)-11-1|immunoglobulin heavy variable (III)-11-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106120289, end_position_on_the_genomic_accession=106120475, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28353, CurrentID=0, Status='live', Symbol='IGHVIII-5-2', Aliases='3-05.2P, IGHVIII52', description='immunoglobulin heavy variable (III)-5-2 (pseudogene)', other_designations='IGHV(III)-5-2|immunoglobulin heavy variable (III)-5-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106045600, end_position_on_the_genomic_accession=106045860, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28354, CurrentID=0, Status='live', Symbol='IGHVIII-5-1', Aliases='3-05.1P, IGHVIII51', description='immunoglobulin heavy variable (III)-5-1 (pseudogene)', other_designations='IGHV(III)-5-1|immunoglobulin heavy variable (III)-5-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106039666, end_position_on_the_genomic_accession=106039764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28355, CurrentID=0, Status='live', Symbol='IGHVIII-2-1', Aliases='3-02.1P, IGHVIII21', description='immunoglobulin heavy variable (III)-2-1 (pseudogene)', other_designations='IGHV(III)-2-1|immunoglobulin heavy variable (III)-2-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106001534, end_position_on_the_genomic_accession=106001835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28356, CurrentID=0, Status='live', Symbol='IGHVII-78-1', Aliases='4-78.1P, IGHVII781', description='immunoglobulin heavy variable (II)-78-1 (pseudogene)', other_designations='IGHV(II)-78-1|immunoglobulin heavy variable (II)-78-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106865621, end_position_on_the_genomic_accession=106865895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28357, CurrentID=0, Status='live', Symbol='IGHVII-74-1', Aliases='4-74.1P, IGHVII741', description='immunoglobulin heavy variable (II)-74-1 (pseudogene)', other_designations='IGHV(II)-74-1|immunoglobulin heavy variable (II)-74-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106821242, end_position_on_the_genomic_accession=106821410, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28358, CurrentID=0, Status='live', Symbol='IGHVII-67-1', Aliases='4-67.1P, IGHVII671', description='immunoglobulin heavy variable (II)-67-1 (pseudogene)', other_designations='IGHV(II)-67-1|immunoglobulin heavy variable (II)-67-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106686393, end_position_on_the_genomic_accession=106686542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28359, CurrentID=0, Status='live', Symbol='IGHVII-65-1', Aliases='4-65.1P, IGHVII651', description='immunoglobulin heavy variable (II)-65-1 (pseudogene)', other_designations='IGHV(II)-65-1|immunoglobulin heavy variable (II)-65-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106671772, end_position_on_the_genomic_accession=106672044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28360, CurrentID=0, Status='live', Symbol='IGHVII-62-1', Aliases='4-62.1P, IGHVII621', description='immunoglobulin heavy variable (II)-62-1 (pseudogene)', other_designations='IGHV(II)-62-1|immunoglobulin heavy variable (II)-62-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106650516, end_position_on_the_genomic_accession=106650788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28361, CurrentID=0, Status='live', Symbol='IGHVII-60-1', Aliases='4-60.1P, IGHVII601', description='immunoglobulin heavy variable (II)-60-1 (pseudogene)', other_designations='IGHV(II)-60-1|immunoglobulin heavy variable (II)-60-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106637705, end_position_on_the_genomic_accession=106637973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28362, CurrentID=0, Status='live', Symbol='IGHVII-53-1', Aliases='4-53.1P, IGHVII531', description='immunoglobulin heavy variable (II)-53-1 (pseudogene)', other_designations='IGHV(II)-53-1|immunoglobulin heavy variable (II)-53-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106599655, end_position_on_the_genomic_accession=106599924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28363, CurrentID=0, Status='live', Symbol='IGHVII-51-2', Aliases='4-51.2P, IGHVII512', description='immunoglobulin heavy variable (II)-51-2 (pseudogene)', other_designations='IGHV(II)-51-2|immunoglobulin heavy variable (II)-51-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106584714, end_position_on_the_genomic_accession=106584976, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28364, CurrentID=0, Status='live', Symbol='IGHVII-49-1', Aliases='4-49.1P, IGHVII491', description='immunoglobulin heavy variable (II)-49-1 (pseudogene)', other_designations='IGHV(II)-49-1|immunoglobulin heavy variable (II)-49-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106564328, end_position_on_the_genomic_accession=106564598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28365, CurrentID=0, Status='live', Symbol='IGHVII-46-1', Aliases='4-46.1P, IGHVII461', description='immunoglobulin heavy variable (II)-46-1 (pseudogene)', other_designations='IGHV(II)-46-1|immunoglobulin heavy variable (II)-46-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106515836, end_position_on_the_genomic_accession=106516124, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28366, CurrentID=0, Status='live', Symbol='IGHVII-44-2', Aliases='4-44.2P, IGHVII442', description='immunoglobulin heavy variable (II)-44-2 (pseudogene)', other_designations='IGHV(II)-44-2|immunoglobulin heavy variable (II)-44-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106494145, end_position_on_the_genomic_accession=106494381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28367, CurrentID=0, Status='live', Symbol='IGHVII-43-1', Aliases='4-43.1P, IGHVII431', description='immunoglobulin heavy variable (II)-43-1 (pseudogene)', other_designations='IGHV(II)-43-1|immunoglobulin heavy variable (II)-43-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106472791, end_position_on_the_genomic_accession=106473000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28368, CurrentID=0, Status='live', Symbol='IGHVII-40-1', Aliases='4-40.1P, IGHVII401', description='immunoglobulin heavy variable (II)-40-1 (pseudogene)', other_designations='IGHV(II)-40-1|immunoglobulin heavy variable (II)-40-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106440934, end_position_on_the_genomic_accession=106441010, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28369, CurrentID=0, Status='live', Symbol='IGHVII-33-1', Aliases='4-33.1P, IGHVII331', description='immunoglobulin heavy variable (II)-33-1 (pseudogene)', other_designations='IGHV(II)-33-1|immunoglobulin heavy variable (II)-33-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106367391, end_position_on_the_genomic_accession=106367664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28370, CurrentID=0, Status='live', Symbol='IGHVII-31-1', Aliases='4-31.1P, IGHVII311', description='immunoglobulin heavy variable (II)-31-1 (pseudogene)', other_designations='IGHV(II)-31-1|immunoglobulin heavy variable (II)-31-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106354616, end_position_on_the_genomic_accession=106354718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28371, CurrentID=0, Status='live', Symbol='IGHVII-30-1', Aliases='4-30.1P, IGHVII301', description='immunoglobulin heavy variable (II)-30-1 (pseudogene)', other_designations='IGHV(II)-30-1|immunoglobulin heavy variable (II)-30-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106342680, end_position_on_the_genomic_accession=106342953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28372, CurrentID=0, Status='live', Symbol='IGHVII-28-1', Aliases='4-28.1P, IGHVII281', description='immunoglobulin heavy variable (II)-28-1 (pseudogene)', other_designations='IGHV(II)-28-1|immunoglobulin heavy variable (II)-28-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106329432, end_position_on_the_genomic_accession=106329686, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28373, CurrentID=0, Status='live', Symbol='IGHVII-26-2', Aliases='4-26.2P, IGHVII262', description='immunoglobulin heavy variable (II)-26-2 (pseudogene)', other_designations='IGHV(II)-26-2|immunoglobulin heavy variable (II)-26-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106314648, end_position_on_the_genomic_accession=106314960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28374, CurrentID=0, Status='live', Symbol='IGHVII-22-1', Aliases='4-22.1P, IGHVII221', description='immunoglobulin heavy variable (II)-22-1 (pseudogene)', other_designations='IGHV(II)-22-1|immunoglobulin heavy variable (II)-22-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106263360, end_position_on_the_genomic_accession=106263628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28375, CurrentID=0, Status='live', Symbol='IGHVII-20-1', Aliases='4-20.1P, IGHVII201', description='immunoglobulin heavy variable (II)-20-1 (pseudogene)', other_designations='IGHV(II)-20-1|immunoglobulin heavy variable (II)-20-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106212825, end_position_on_the_genomic_accession=106213323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28376, CurrentID=0, Status='live', Symbol='IGHVII-15-1', Aliases='4-15.1P, IGHVII151', description='immunoglobulin heavy variable (II)-15-1 (pseudogene)', other_designations='IGHV(II)-15-1|immunoglobulin heavy variable (II)-15-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106163534, end_position_on_the_genomic_accession=106163804, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28377, CurrentID=0, Status='live', Symbol='IGHVII-1-1', Aliases='4-01.1P, IGHVII11', description='immunoglobulin heavy variable (II)-1-1 (pseudogene)', other_designations='IGHV(II)-1-1|immunoglobulin heavy variable (II)-1-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105944938, end_position_on_the_genomic_accession=105945391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28378, CurrentID=0, Status='live', Symbol='IGHV7-81', Aliases='IGHV781', description='immunoglobulin heavy variable 7-81 (non-functional)', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106874581, end_position_on_the_genomic_accession=106875017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371244, CurrentID=0, Status='live', Symbol='LOC105371244', Aliases='', description='uncharacterized LOC105371244', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49374122, end_position_on_the_genomic_accession=49454091, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  28380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28380, CurrentID=0, Status='live', Symbol='IGHV7-56', Aliases='7-56P, IGHV756', description='immunoglobulin heavy variable 7-56 (pseudogene)', other_designations='immunoglobulin heavy variable 7-56 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106609762, end_position_on_the_genomic_accession=106610196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28381, CurrentID=0, Status='live', Symbol='IGHV7-40', Aliases='7-40P, IGHV740', description='immunoglobulin heavy variable 7-40 (pseudogene)', other_designations='immunoglobulin heavy variable 7-40 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106425359, end_position_on_the_genomic_accession=106425561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28382, CurrentID=0, Status='live', Symbol='IGHV7-34-1', Aliases='7-34.1P, IGHV7341', description='immunoglobulin heavy variable 7-34-1 (pseudogene)', other_designations='immunoglobulin heavy variable 7-34-1 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106377300, end_position_on_the_genomic_accession=106377733, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28383, CurrentID=0, Status='live', Symbol='IGHV7-27', Aliases='7-27P, IGHV727', description='immunoglobulin heavy variable 7-27 (pseudogene)', other_designations='immunoglobulin heavy variable 7-27 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106317823, end_position_on_the_genomic_accession=106318236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28385, CurrentID=0, Status='live', Symbol='IGHV6-1', Aliases='IGHV61, VH', description='immunoglobulin heavy variable 6-1', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105939754, end_position_on_the_genomic_accession=105940201, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28386, CurrentID=0, Status='live', Symbol='IGHV5-10-1', Aliases='IGHV5-A, IGHV5A', description='immunoglobulin heavy variable 5-10-1', other_designations='immunoglobulin heavy variable 5-A (provisional)|immunoglobulin heavy variable 5-A (provisional, gene/pseudogene)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28387, CurrentID=0, Status='live', Symbol='IGHV5-78', Aliases='5-78P, IGHV578, VH5', description='immunoglobulin heavy variable 5-78 (pseudogene)', other_designations='immunoglobulin heavy variable 5-78 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106851112, end_position_on_the_genomic_accession=106851546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28388, CurrentID=0, Status='live', Symbol='IGHV5-51', Aliases='IGHV551, VH', description='immunoglobulin heavy variable 5-51', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106578742, end_position_on_the_genomic_accession=106579177, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28389, CurrentID=0, Status='live', Symbol='IGHV4-38-2', Aliases='IGHV4-B, IGHV4B', description='immunoglobulin heavy variable 4-38-2', other_designations='immunoglobulin heavy variable 4-B (provisional)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28390, CurrentID=0, Status='live', Symbol='IGHV4-80', Aliases='4-80P, IGHV480', description='immunoglobulin heavy variable 4-80 (pseudogene)', other_designations='immunoglobulin heavy variable 4-80 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106872783, end_position_on_the_genomic_accession=106873186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28391, CurrentID=0, Status='live', Symbol='IGHV4-61', Aliases='IGHV461, VH', description='immunoglobulin heavy variable 4-61', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106639119, end_position_on_the_genomic_accession=106639556, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28392, CurrentID=0, Status='live', Symbol='IGHV4-59', Aliases='IGHV459, VH', description='immunoglobulin heavy variable 4-59', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106627249, end_position_on_the_genomic_accession=106627680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28393, CurrentID=0, Status='live', Symbol='IGHV4-55', Aliases='4-55P, IGHV455', description='immunoglobulin heavy variable 4-55 (pseudogene)', other_designations='immunoglobulin heavy variable 4-55 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106606116, end_position_on_the_genomic_accession=106606551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28394, CurrentID=0, Status='live', Symbol='IGHV4-39', Aliases='IGHV439, VH', description='immunoglobulin heavy variable 4-39', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106421709, end_position_on_the_genomic_accession=106422147, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28395, CurrentID=0, Status='live', Symbol='IGHV4-34', Aliases='IGHV434, VH', description='immunoglobulin heavy variable 4-34', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106373661, end_position_on_the_genomic_accession=106374093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28396, CurrentID=0, Status='live', Symbol='IGHV4-31', Aliases='IGHV431', description='immunoglobulin heavy variable 4-31', other_designations='anti-RhD monoclonal T125 gamma1 heavy chain', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106349283, end_position_on_the_genomic_accession=106349720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28397, CurrentID=0, Status='live', Symbol='IGHV4-30-4', Aliases='IGHV4-3, IGHV4304', description='immunoglobulin heavy variable 4-30-4', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28398, CurrentID=0, Status='live', Symbol='IGHV4-30-2', Aliases='IGHV4-3, IGHV4302', description='immunoglobulin heavy variable 4-30-2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28399, CurrentID=0, Status='live', Symbol='IGHV4-30-1', Aliases='IGHV4-3', description='immunoglobulin heavy variable 4-30-1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28400, CurrentID=0, Status='live', Symbol='IGHV4-28', Aliases='IGHV428, VH', description='immunoglobulin heavy variable 4-28', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106324254, end_position_on_the_genomic_accession=106324688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28401, CurrentID=0, Status='live', Symbol='IGHV4-4', Aliases='IGHV44, VH', description='immunoglobulin heavy variable 4-4', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106011922, end_position_on_the_genomic_accession=106012356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28402, CurrentID=0, Status='live', Symbol='IGHV3-69-1', Aliases='IGHV3-H, IGHV3H', description='immunoglobulin heavy variable 3-69-1 (pseudogene)', other_designations='immunoglobulin heavy variable 3-H (provisional) pseudogene|immunoglobulin heavy variable 3-H (pseudogene)|immunoglobulin heavy variable 3-H pseudogene (provisional)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373559, CurrentID=0, Status='live', Symbol='LOC105373559', Aliases='', description='uncharacterized LOC105373559', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111607744, end_position_on_the_genomic_accession=111612541, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  28404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28404, CurrentID=0, Status='live', Symbol='IGHV3-38-3', Aliases='IGHV3-D, IGHV3D', description='immunoglobulin heavy variable 3-38-3 (non-functional)', other_designations='immunoglobulin heavy variable 3-D (provisional)', map_location='14q32.33 not on reference assembly', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28405, CurrentID=0, Status='live', Symbol='IGHV3-79', Aliases='3-79P, IGHV379', description='immunoglobulin heavy variable 3-79 (pseudogene)', other_designations='immunoglobulin heavy variable 3-79 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106867640, end_position_on_the_genomic_accession=106868092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28406, CurrentID=0, Status='live', Symbol='IGHV3-76', Aliases='3-76P, IGHV376', description='immunoglobulin heavy variable 3-76 (pseudogene)', other_designations='immunoglobulin heavy variable 3-76 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106827861, end_position_on_the_genomic_accession=106828310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28407, CurrentID=0, Status='live', Symbol='IGHV3-75', Aliases='3-75P, IGHV375, VH3', description='immunoglobulin heavy variable 3-75 (pseudogene)', other_designations='immunoglobulin heavy variable 3-75 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106823678, end_position_on_the_genomic_accession=106824147, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28408, CurrentID=0, Status='live', Symbol='IGHV3-74', Aliases='IGHV374, VH', description='immunoglobulin heavy variable 3-74', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106810440, end_position_on_the_genomic_accession=106810895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28409, CurrentID=0, Status='live', Symbol='IGHV3-73', Aliases='IGHV373, VH', description='immunoglobulin heavy variable 3-73', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106802692, end_position_on_the_genomic_accession=106803153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28410, CurrentID=0, Status='live', Symbol='IGHV3-72', Aliases='IGHV372, VH', description='immunoglobulin heavy variable 3-72', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106790692, end_position_on_the_genomic_accession=106791153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28411, CurrentID=0, Status='live', Symbol='IGHV3-71', Aliases='3-71P, IGHV3-G, IGHV371, IGHV3G', description='immunoglobulin heavy variable 3-71 (pseudogene)', other_designations='immunoglobulin heavy variable 3-71 pseudogene|immunoglobulin heavy variable 3-G (provisional)|immunoglobulin heavy variable 3-G pseudogene (provisional)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106775157, end_position_on_the_genomic_accession=106775618, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28412, CurrentID=0, Status='live', Symbol='IGHV3-66', Aliases='IGHV366, VH', description='immunoglobulin heavy variable 3-66', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106675015, end_position_on_the_genomic_accession=106675465, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28413, CurrentID=0, Status='live', Symbol='IGHV3-65', Aliases='3-65P, IGHV365', description='immunoglobulin heavy variable 3-65 (pseudogene)', other_designations='immunoglobulin heavy variable 3-65 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106666038, end_position_on_the_genomic_accession=106666532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28414, CurrentID=0, Status='live', Symbol='IGHV3-64', Aliases='IGHV364, VH', description='immunoglobulin heavy variable 3-64', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106657723, end_position_on_the_genomic_accession=106658178, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28415, CurrentID=0, Status='live', Symbol='IGHV3-63', Aliases='IGHV363, V3-63P', description='immunoglobulin heavy variable 3-63 (pseudogene)', other_designations='immunoglobulin heavy variable 3-63 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106652213, end_position_on_the_genomic_accession=106652680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28416, CurrentID=0, Status='live', Symbol='IGHV3-62', Aliases='3-62P, IGHV362', description='immunoglobulin heavy variable 3-62 (pseudogene)', other_designations='immunoglobulin heavy variable 3-62 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106643130, end_position_on_the_genomic_accession=106643585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28417, CurrentID=0, Status='live', Symbol='IGHV3-60', Aliases='3-60P, IGHV360', description='immunoglobulin heavy variable 3-60 (pseudogene)', other_designations='immunoglobulin heavy variable 3-60 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106631197, end_position_on_the_genomic_accession=106631653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28418, CurrentID=0, Status='live', Symbol='IGHV3-57', Aliases='3-57P, IGHV357', description='immunoglobulin heavy variable 3-57 (pseudogene)', other_designations='immunoglobulin heavy variable 3-57 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106618909, end_position_on_the_genomic_accession=106619214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28419, CurrentID=0, Status='live', Symbol='IGHV3-54', Aliases='3-54P, IGHV354', description='immunoglobulin heavy variable 3-54 (pseudogene)', other_designations='immunoglobulin heavy variable 3-54 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106601336, end_position_on_the_genomic_accession=106601792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28420, CurrentID=0, Status='live', Symbol='IGHV3-53', Aliases='IGHV353, VH', description='immunoglobulin heavy variable 3-53', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106592676, end_position_on_the_genomic_accession=106593126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28421, CurrentID=0, Status='live', Symbol='IGHV3-52', Aliases='3-52P, IGHV352', description='immunoglobulin heavy variable 3-52 (pseudogene)', other_designations='immunoglobulin heavy variable 3-52 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106586376, end_position_on_the_genomic_accession=106586826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28422, CurrentID=0, Status='live', Symbol='IGHV3-50', Aliases='3-50P, IGHV350', description='immunoglobulin heavy variable 3-50 (pseudogene)', other_designations='immunoglobulin heavy variable 3-50 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106566100, end_position_on_the_genomic_accession=106566555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28423, CurrentID=0, Status='live', Symbol='IGHV3-49', Aliases='IGHV349, VH', description='immunoglobulin heavy variable 3-49', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106556936, end_position_on_the_genomic_accession=106557397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28424, CurrentID=0, Status='live', Symbol='IGHV3-48', Aliases='IGHV348, VH', description='immunoglobulin heavy variable 3-48', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106537810, end_position_on_the_genomic_accession=106538265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28425, CurrentID=0, Status='live', Symbol='IGHV3-47', Aliases='3-47P, IGHV347', description='immunoglobulin heavy variable 3-47 (pseudogene)', other_designations='immunoglobulin heavy variable 3-47 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106518579, end_position_on_the_genomic_accession=106519034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28426, CurrentID=0, Status='live', Symbol='IGHV3-43', Aliases='IGHV343, VH', description='immunoglobulin heavy variable 3-43', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106470263, end_position_on_the_genomic_accession=106470720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28427, CurrentID=0, Status='live', Symbol='IGHV3-42', Aliases='3-42P, IGHV342', description='immunoglobulin heavy variable 3-42 (pseudogene)', other_designations='immunoglobulin heavy variable 3-42 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106463254, end_position_on_the_genomic_accession=106463688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28428, CurrentID=0, Status='live', Symbol='IGHV3-41', Aliases='3-41P, IGHV341', description='immunoglobulin heavy variable 3-41 (pseudogene)', other_designations='immunoglobulin heavy variable 3-41 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106443129, end_position_on_the_genomic_accession=106443584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28429, CurrentID=0, Status='live', Symbol='IGHV3-38', Aliases='IGHV338, VH', description='immunoglobulin heavy variable 3-38 (non-functional)', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106410494, end_position_on_the_genomic_accession=106410943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28430, CurrentID=0, Status='live', Symbol='IGHV3-37', Aliases='3-37P, IGHV337', description='immunoglobulin heavy variable 3-37 (pseudogene)', other_designations='immunoglobulin heavy variable 3-37 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106396665, end_position_on_the_genomic_accession=106397114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28431, CurrentID=0, Status='live', Symbol='IGHV3-36', Aliases='3-36P, IGHV336', description='immunoglobulin heavy variable 3-36 (pseudogene)', other_designations='immunoglobulin heavy variable 3-36 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106392769, end_position_on_the_genomic_accession=106393231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28432, CurrentID=0, Status='live', Symbol='IGHV3-35', Aliases='IGHV335, VH', description='immunoglobulin heavy variable 3-35 (non-functional)', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106389390, end_position_on_the_genomic_accession=106389843, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28433, CurrentID=0, Status='live', Symbol='IGHV3-33-2', Aliases='3-33.2P, IGHV3332', description='immunoglobulin heavy variable 3-33-2 (pseudogene)', other_designations='immunoglobulin heavy variable 3-33-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106369098, end_position_on_the_genomic_accession=106369546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28434, CurrentID=0, Status='live', Symbol='IGHV3-33', Aliases='IGHV333, VH', description='immunoglobulin heavy variable 3-33', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106359791, end_position_on_the_genomic_accession=106360244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28435, CurrentID=0, Status='live', Symbol='IGHV3-32', Aliases='3-32P, IGHV332', description='immunoglobulin heavy variable 3-32 (pseudogene)', other_designations='immunoglobulin heavy variable 3-32 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106356145, end_position_on_the_genomic_accession=106356602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28438, CurrentID=0, Status='live', Symbol='IGHV3-30-2', Aliases='3-30.2P, IGHV3302', description='immunoglobulin heavy variable 3-30-2 (pseudogene)', other_designations='immunoglobulin heavy variable 3-30-2 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106344385, end_position_on_the_genomic_accession=106344833, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28439, CurrentID=0, Status='live', Symbol='IGHV3-30', Aliases='IGHV330, VH', description='immunoglobulin heavy variable 3-30', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106335080, end_position_on_the_genomic_accession=106335533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28440, CurrentID=0, Status='live', Symbol='IGHV3-29', Aliases='3-29P, IGHV329', description='immunoglobulin heavy variable 3-29 (pseudogene)', other_designations='immunoglobulin heavy variable 3-29 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106331106, end_position_on_the_genomic_accession=106331563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28441, CurrentID=0, Status='live', Symbol='IGHV3-25', Aliases='3-25P, IGHV325', description='immunoglobulin heavy variable 3-25 (pseudogene)', other_designations='immunoglobulin heavy variable 3-25 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106289026, end_position_on_the_genomic_accession=106289479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28442, CurrentID=0, Status='live', Symbol='IGHV3-23', Aliases='DP47, IGHV323, V3-23, VH26', description='immunoglobulin heavy variable 3-23', other_designations='mu-chain precursor (AA -19 to 142)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106268606, end_position_on_the_genomic_accession=106269061, orientation='minus', exon_count=0, OMIM=611939, no_hdr0=''),
  28443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28443, CurrentID=0, Status='live', Symbol='IGHV3-22', Aliases='3-22P, IGHV322', description='immunoglobulin heavy variable 3-22 (pseudogene)', other_designations='immunoglobulin heavy variable 3-22 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106257762, end_position_on_the_genomic_accession=106258223, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28444, CurrentID=0, Status='live', Symbol='IGHV3-21', Aliases='IGHV321, VH', description='immunoglobulin heavy variable 3-21', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106235062, end_position_on_the_genomic_accession=106235515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28445, CurrentID=0, Status='live', Symbol='IGHV3-20', Aliases='IGHV320, VH', description='immunoglobulin heavy variable 3-20', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106210936, end_position_on_the_genomic_accession=106211391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28446, CurrentID=0, Status='live', Symbol='IGHV3-19', Aliases='3-19P, IGHV319', description='immunoglobulin heavy variable 3-19 (pseudogene)', other_designations='immunoglobulin heavy variable 3-19 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106196698, end_position_on_the_genomic_accession=106196990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28447, CurrentID=0, Status='live', Symbol='IGHV3-16', Aliases='IGHV316, VH', description='immunoglobulin heavy variable 3-16 (non-functional)', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106165203, end_position_on_the_genomic_accession=106165656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28448, CurrentID=0, Status='live', Symbol='IGHV3-15', Aliases='IGHV315, VH', description='immunoglobulin heavy variable 3-15', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106153622, end_position_on_the_genomic_accession=106154083, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402483, CurrentID=0, Status='live', Symbol='LINC01000', Aliases='', description='long intergenic non-protein coding RNA 1000', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128641241, end_position_on_the_genomic_accession=128660998, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  28450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28450, CurrentID=0, Status='live', Symbol='IGHV3-11', Aliases='IGHV311, VH', description='immunoglobulin heavy variable 3-11 (gene/pseudogene)', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106116635, end_position_on_the_genomic_accession=106117084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28451, CurrentID=0, Status='live', Symbol='IGHV3-9', Aliases='IGHV39, VH', description='immunoglobulin heavy variable 3-9', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28452, CurrentID=0, Status='live', Symbol='IGHV3-7', Aliases='IGHV37, VH', description='immunoglobulin heavy variable 3-7', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106062149, end_position_on_the_genomic_accession=106062604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28453, CurrentID=0, Status='live', Symbol='IGHV3-6', Aliases='3-06P, IGHV36', description='immunoglobulin heavy variable 3-6 (pseudogene)', other_designations='immunoglobulin heavy variable 3-6 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106055544, end_position_on_the_genomic_accession=106055998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28454, CurrentID=0, Status='live', Symbol='IGHV2-70', Aliases='IGHV270, VH', description='immunoglobulin heavy variable 2-70', other_designations='Ig heavy chain V-II region|immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106770577, end_position_on_the_genomic_accession=106771020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28455, CurrentID=0, Status='live', Symbol='IGHV2-26', Aliases='IGHV226, VH', description='immunoglobulin heavy variable 2-26', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106301395, end_position_on_the_genomic_accession=106301838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28456, CurrentID=0, Status='live', Symbol='IGHV2-10', Aliases='2-10P, IGHV210', description='immunoglobulin heavy variable 2-10 (pseudogene)', other_designations='immunoglobulin heavy variable 2-10 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28457, CurrentID=0, Status='live', Symbol='IGHV2-5', Aliases='IGHV25, VH', description='immunoglobulin heavy variable 2-5', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106037902, end_position_on_the_genomic_accession=106038345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28458, CurrentID=0, Status='live', Symbol='IGHV1-69-2', Aliases='IGHV1-F, IGHV1F', description='immunoglobulin heavy variable 1-69-2', other_designations='immunoglobulin heavy variable 1-F (provisional)', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28460, CurrentID=0, Status='live', Symbol='IGHV1-38-4', Aliases='IGHV1-C, IGHV1C', description='immunoglobulin heavy variable 1-38-4 (non-functional)', other_designations='immunoglobulin heavy variable 1-C (provisional)|immunoglobulin heavy variable 1-C (provisional, non-functional)', map_location='14q32.33 not on reference assembly', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28461, CurrentID=0, Status='live', Symbol='IGHV1-69', Aliases='IGHV1-E, IGHV169, IGHV1E', description='immunoglobulin heavy variable 1-69', other_designations='immunoglobulin heavy variable 1-e', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106714682, end_position_on_the_genomic_accession=106715120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28462, CurrentID=0, Status='live', Symbol='IGHV1-68', Aliases='1-68P, IGHV168', description='immunoglobulin heavy variable 1-68 (pseudogene)', other_designations='immunoglobulin heavy variable 1-68 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106703807, end_position_on_the_genomic_accession=106704286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28463, CurrentID=0, Status='live', Symbol='IGHV1-67', Aliases='1-67P, IGHV167', description='immunoglobulin heavy variable 1-67 (pseudogene)', other_designations='immunoglobulin heavy variable 1-67 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106680603, end_position_on_the_genomic_accession=106681042, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28464, CurrentID=0, Status='live', Symbol='IGHV1-58', Aliases='IGHV158, VH', description='immunoglobulin heavy variable 1-58', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106622357, end_position_on_the_genomic_accession=106622794, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28465, CurrentID=0, Status='live', Symbol='IGHV1-46', Aliases='IGHV146', description='immunoglobulin heavy variable 1-46', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106511115, end_position_on_the_genomic_accession=106511552, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28466, CurrentID=0, Status='live', Symbol='IGHV1-45', Aliases='IGHV145, VH', description='immunoglobulin heavy variable 1-45', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106506996, end_position_on_the_genomic_accession=106507433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28467, CurrentID=0, Status='live', Symbol='IGHV1-24', Aliases='IGHV124, VH', description='immunoglobulin heavy variable 1-24', other_designations='immunogloblin heavy chain variable region', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106276546, end_position_on_the_genomic_accession=106276983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28468, CurrentID=0, Status='live', Symbol='IGHV1-18', Aliases='IGHV118', description='immunoglobulin heavy variable 1-18', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106184899, end_position_on_the_genomic_accession=106185335, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28469, CurrentID=0, Status='live', Symbol='IGHV1-17', Aliases='1-17P, IGHV117', description='immunoglobulin heavy variable 1-17 (pseudogene)', other_designations='immunoglobulin heavy variable 1-17 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106174342, end_position_on_the_genomic_accession=106174778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28470, CurrentID=0, Status='live', Symbol='IGHV1-14', Aliases='1-14P, IGHV114', description='immunoglobulin heavy variable 1-14 (pseudogene)', other_designations='immunoglobulin heavy variable 1-14 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106145882, end_position_on_the_genomic_accession=106146131, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28471, CurrentID=0, Status='live', Symbol='IGHV1-12', Aliases='1-12P, IGHV112', description='immunoglobulin heavy variable 1-12 (pseudogene)', other_designations='immunoglobulin heavy variable 1-12 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106122420, end_position_on_the_genomic_accession=106122709, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28472, CurrentID=0, Status='live', Symbol='IGHV1-8', Aliases='IGHV18', description='immunoglobulin heavy variable 1-8', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28473, CurrentID=0, Status='live', Symbol='IGHV1-3', Aliases='IGHV13, VI-3B', description='immunoglobulin heavy variable 1-3', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106005095, end_position_on_the_genomic_accession=106005532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28474, CurrentID=0, Status='live', Symbol='IGHV1-2', Aliases='IGHV12, V35', description='immunoglobulin heavy variable 1-2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105986582, end_position_on_the_genomic_accession=105987019, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28475, CurrentID=0, Status='live', Symbol='IGHJ6', Aliases='JH6b', description='immunoglobulin heavy joining 6', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105863196, end_position_on_the_genomic_accession=105863260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28476, CurrentID=0, Status='live', Symbol='IGHJ5', Aliases='JH5b', description='immunoglobulin heavy joining 5', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105863812, end_position_on_the_genomic_accession=105863864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28477, CurrentID=0, Status='live', Symbol='IGHJ4', Aliases='JH4b', description='immunoglobulin heavy joining 4', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105864213, end_position_on_the_genomic_accession=105864262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28478, CurrentID=0, Status='live', Symbol='IGHJ3P', Aliases='Jpsi3', description='immunoglobulin heavy joining 3P (pseudogene)', other_designations='immunoglobulin heavy joining 3P pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105863415, end_position_on_the_genomic_accession=105863465, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28479, CurrentID=0, Status='live', Symbol='IGHJ3', Aliases='JH3b', description='immunoglobulin heavy joining 3', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105864585, end_position_on_the_genomic_accession=105864636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28480, CurrentID=0, Status='live', Symbol='IGHJ2P', Aliases='Jpsi2', description='immunoglobulin heavy joining 2P (pseudogene)', other_designations='immunoglobulin heavy joining 2P pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105864793, end_position_on_the_genomic_accession=105864852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28481, CurrentID=0, Status='live', Symbol='IGHJ2', Aliases='JH2', description='immunoglobulin heavy joining 2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105865197, end_position_on_the_genomic_accession=105865251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28482, CurrentID=0, Status='live', Symbol='IGHJ1P', Aliases='J-psi-1, Jpsi1', description='immunoglobulin heavy joining 1P (pseudogene)', other_designations='immunoglobulin heavy joining 1P pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105865624, end_position_on_the_genomic_accession=105865677, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28483, CurrentID=0, Status='live', Symbol='IGHJ1', Aliases='JH1', description='immunoglobulin heavy joining 1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105865405, end_position_on_the_genomic_accession=105865458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28484, CurrentID=0, Status='live', Symbol='IGHD7-27', Aliases='DHQ52, IGHD727', description='immunoglobulin heavy diversity 7-27', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105865551, end_position_on_the_genomic_accession=105865561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28485, CurrentID=0, Status='live', Symbol='IGHD6-25', Aliases='IGHD625', description='immunoglobulin heavy diversity 6-25', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105881539, end_position_on_the_genomic_accession=105881556, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28486, CurrentID=0, Status='live', Symbol='IGHD6-19', Aliases='IGHD619', description='immunoglobulin heavy diversity 6-19', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105891699, end_position_on_the_genomic_accession=105891719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28487, CurrentID=0, Status='live', Symbol='IGHD6-13', Aliases='DN1, IGHD613', description='immunoglobulin heavy diversity 6-13', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105901142, end_position_on_the_genomic_accession=105901162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28488, CurrentID=0, Status='live', Symbol='IGHD6-6', Aliases='D(N4), IGHD66', description='immunoglobulin heavy diversity 6-6', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105910410, end_position_on_the_genomic_accession=105910427, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28489, CurrentID=0, Status='live', Symbol='IGHD5-24', Aliases='IGHD524', description='immunoglobulin heavy diversity 5-24 (non-functional)', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105883903, end_position_on_the_genomic_accession=105883922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28490, CurrentID=0, Status='live', Symbol='IGHD5-18', Aliases='IGHD518', description='immunoglobulin heavy diversity 5-18', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105893542, end_position_on_the_genomic_accession=105893561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28491, CurrentID=0, Status='live', Symbol='IGHD5-12', Aliases='DK1, IGHD512', description='immunoglobulin heavy diversity 5-12', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105902649, end_position_on_the_genomic_accession=105902671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28492, CurrentID=0, Status='live', Symbol='IGHD5-5', Aliases='DK4, IGHD55', description='immunoglobulin heavy diversity 5-5', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105912257, end_position_on_the_genomic_accession=105912276, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28493, CurrentID=0, Status='live', Symbol='IGHD4-23', Aliases='IGHD423', description='immunoglobulin heavy diversity 4-23 (non-functional)', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105884870, end_position_on_the_genomic_accession=105884888, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28494, CurrentID=0, Status='live', Symbol='IGHD4-17', Aliases='IGHD417', description='immunoglobulin heavy diversity 4-17', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105894508, end_position_on_the_genomic_accession=105894523, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28495, CurrentID=0, Status='live', Symbol='IGHD4-11', Aliases='DA1, IGHD411', description='immunoglobulin heavy diversity 4-11 (non-functional)', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105903616, end_position_on_the_genomic_accession=105903631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28496, CurrentID=0, Status='live', Symbol='IGHD4-4', Aliases='DA4, IGHD44', description='immunoglobulin heavy diversity 4-4', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105913222, end_position_on_the_genomic_accession=105913237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28497, CurrentID=0, Status='live', Symbol='IGHD3-22', Aliases='IGHD322', description='immunoglobulin heavy diversity 3-22', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105886031, end_position_on_the_genomic_accession=105886061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28498, CurrentID=0, Status='live', Symbol='IGHD3-16', Aliases='IGHD316', description='immunoglobulin heavy diversity 3-16', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105895634, end_position_on_the_genomic_accession=105895670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28499, CurrentID=0, Status='live', Symbol='IGHD3-10', Aliases="DXP'1, IGHD310", description='immunoglobulin heavy diversity 3-10', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105904497, end_position_on_the_genomic_accession=105904527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28500, CurrentID=0, Status='live', Symbol='IGHD3-9', Aliases='DXP1, IGHD39', description='immunoglobulin heavy diversity 3-9', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105904681, end_position_on_the_genomic_accession=105904711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28501, CurrentID=0, Status='live', Symbol='IGHD3-3', Aliases='DXP4, IGHD33', description='immunoglobulin heavy diversity 3-3', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105914359, end_position_on_the_genomic_accession=105914389, orientation='minus', exon_count=0, OMIM=611937, no_hdr0=''),
  28502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28502, CurrentID=0, Status='live', Symbol='IGHD2-21', Aliases='IGHD221', description='immunoglobulin heavy diversity 2-21', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105888551, end_position_on_the_genomic_accession=105888578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28503, CurrentID=0, Status='live', Symbol='IGHD2-15', Aliases='D2, IGHD215', description='immunoglobulin heavy diversity 2-15', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105897957, end_position_on_the_genomic_accession=105897987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28504, CurrentID=0, Status='live', Symbol='IGHD2-8', Aliases='DLR1, IGHD28', description='immunoglobulin heavy diversity 2-8', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105907211, end_position_on_the_genomic_accession=105907241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28505, CurrentID=0, Status='live', Symbol='IGHD2-2', Aliases='IGHD22', description='immunoglobulin heavy diversity 2-2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105916826, end_position_on_the_genomic_accession=105916856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28506, CurrentID=0, Status='live', Symbol='IGHD1-26', Aliases='IGHD126', description='immunoglobulin heavy diversity 1-26', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105881034, end_position_on_the_genomic_accession=105881053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28507, CurrentID=0, Status='live', Symbol='IGHD1-20', Aliases='IGHD120', description='immunoglobulin heavy diversity 1-20', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105891191, end_position_on_the_genomic_accession=105891207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28508, CurrentID=0, Status='live', Symbol='IGHD1-14', Aliases='DM2, IGHD114', description='immunoglobulin heavy diversity 1-14 (non-functional)', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105900638, end_position_on_the_genomic_accession=105900654, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28509, CurrentID=0, Status='live', Symbol='IGHD1-7', Aliases='DM1, IGHD17', description='immunoglobulin heavy diversity 1-7', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105909907, end_position_on_the_genomic_accession=105909923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28510, CurrentID=0, Status='live', Symbol='IGHD1-1', Aliases='IGHD11', description='immunoglobulin heavy diversity 1-1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105919502, end_position_on_the_genomic_accession=105919518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28511, CurrentID=0, Status='live', Symbol='NKIRAS2', Aliases='KBRAS2, kappaB-Ras2', description='NFKB inhibitor interacting Ras-like 2', other_designations='I-kappa-B-interacting Ras-like protein 2|NFKB inhibitor interacting Ras-like protein 2|kappa B-Ras protein 2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42017561, end_position_on_the_genomic_accession=42025641, orientation='plus', exon_count=6, OMIM=604497, no_hdr0=''),
  28512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28512, CurrentID=0, Status='live', Symbol='NKIRAS1', Aliases='KBRAS1, kappaB-Ras1', description='NFKB inhibitor interacting Ras-like 1', other_designations='I-kappa-B-interacting Ras-like protein 1|NFKB inhibitor interacting Ras-like protein 1|kappa B-Ras protein 1|kappa B-ras 1', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23892081, end_position_on_the_genomic_accession=23917046, orientation='minus', exon_count=5, OMIM=604496, no_hdr0=''),
  28513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28513, CurrentID=0, Status='live', Symbol='CDH19', Aliases='CDH7, CDH7L2', description='cadherin 19, type 2', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=66501187, end_position_on_the_genomic_accession=66604138, orientation='minus', exon_count=13, OMIM=603016, no_hdr0=''),
  28514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28514, CurrentID=0, Status='live', Symbol='DLL1', Aliases='DELTA1, DL1, Delta', description='delta-like 1 (Drosophila)', other_designations='H-Delta-1|drosophila Delta homolog 1', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170282200, end_position_on_the_genomic_accession=170306555, orientation='minus', exon_count=12, OMIM=606582, no_hdr0=''),
  28515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28515, CurrentID=0, Status='live', Symbol='HYPLIP2', Aliases='FCHL2', description='hyperlipidemia, combined, 2', other_designations='', map_location='11p', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604499, no_hdr0=''),
  28516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28516, CurrentID=0, Status='live', Symbol='TRDV3', Aliases='hDV103S1', description='T cell receptor delta variable 3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22469041, end_position_on_the_genomic_accession=22469614, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28517, CurrentID=0, Status='live', Symbol='TRDV2', Aliases='hDV102S1', description='T cell receptor delta variable 2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22422546, end_position_on_the_genomic_accession=22423042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28518, CurrentID=0, Status='live', Symbol='TRDV1', Aliases='hDV101S1', description='T cell receptor delta variable 1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22096050, end_position_on_the_genomic_accession=22096619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28519, CurrentID=0, Status='live', Symbol='TRDJ4', Aliases='', description='T cell receptor delta joining 4', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22455249, end_position_on_the_genomic_accession=22455296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28520, CurrentID=0, Status='live', Symbol='TRDJ3', Aliases='', description='T cell receptor delta joining 3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22459098, end_position_on_the_genomic_accession=22459156, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28521, CurrentID=0, Status='live', Symbol='TRDJ2', Aliases='', description='T cell receptor delta joining 2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22456689, end_position_on_the_genomic_accession=22456742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28522, CurrentID=0, Status='live', Symbol='TRDJ1', Aliases='TCRD', description='T cell receptor delta joining 1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22450089, end_position_on_the_genomic_accession=22450139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28523, CurrentID=0, Status='live', Symbol='TRDD3', Aliases='TCRD', description='T cell receptor delta diversity 3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22449113, end_position_on_the_genomic_accession=22449125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28524, CurrentID=0, Status='live', Symbol='TRDD2', Aliases='', description='T cell receptor delta diversity 2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22439007, end_position_on_the_genomic_accession=22439015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28525, CurrentID=0, Status='live', Symbol='TRDD1', Aliases='', description='T cell receptor delta diversity 1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22438547, end_position_on_the_genomic_accession=22438554, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28526, CurrentID=0, Status='live', Symbol='TRDC', Aliases='TCRD', description='T cell receptor delta constant', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22462932, end_position_on_the_genomic_accession=22466577, orientation='plus', exon_count=0, OMIM=186810, no_hdr0=''),
  106480547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480547, CurrentID=0, Status='live', Symbol='RNU4-56P', Aliases='', description='RNA, U4 small nuclear 56, pseudogene', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441662, CurrentID=0, Status='live', Symbol='TET1P1', Aliases='CXXC6P1', description='tet methylcytosine dioxygenase 1 pseudogene 1', other_designations='CXXC finger 6 pseudogene 1|CXXC zinc finger 6 pseudogene 1|tet oncogene 1 pseudogene 1', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87889943, end_position_on_the_genomic_accession=87893397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369199, CurrentID=0, Status='live', Symbol='LOC105369199', Aliases='', description='putative neuroblastoma breakpoint family member 6-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120963238, end_position_on_the_genomic_accession=121002782, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370565, CurrentID=0, Status='live', Symbol='LOC105370565', Aliases='', description='uncharacterized LOC105370565', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74450034, end_position_on_the_genomic_accession=74476547, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724300, CurrentID=0, Status='live', Symbol='LOC102724300', Aliases='', description='uncharacterized LOC102724300', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54785869, end_position_on_the_genomic_accession=54817653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28552, CurrentID=0, Status='live', Symbol='TRBV23OR9-2', Aliases='TCRBV19S2O, TCRBV23S2, TRBV23OR92', description='T cell receptor beta variable 23/OR9-2 (non-functional)', other_designations='T cell receptor beta variable 23/OR9-2 pseudogene|TRBV23/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33638037, end_position_on_the_genomic_accession=33638508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373564, CurrentID=0, Status='live', Symbol='LOC105373564', Aliases='', description='uncharacterized LOC105373564', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113209949, end_position_on_the_genomic_accession=113213332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79512, CurrentID=0, Status='live', Symbol='OR5M6P', Aliases='', description='olfactory receptor family 5 subfamily M member 6 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 6 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56512234, end_position_on_the_genomic_accession=56513170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28555, CurrentID=0, Status='live', Symbol='TRBVB', Aliases='TCRBV34S1', description='T cell receptor beta variable B (pseudogene)', other_designations='T cell receptor beta variable B pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142711384, end_position_on_the_genomic_accession=142711924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28556, CurrentID=0, Status='live', Symbol='TRBVA', Aliases='TCRBV33S1', description='T cell receptor beta variable A (pseudogene)', other_designations='T cell receptor beta variable A pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142681415, end_position_on_the_genomic_accession=142681869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28557, CurrentID=0, Status='live', Symbol='TRBV30', Aliases='TCRBV20S1A1N2, TCRBV30S1', description='T cell receptor beta variable 30 (gene/pseudogene)', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142812586, end_position_on_the_genomic_accession=142813287, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28558, CurrentID=0, Status='live', Symbol='TRBV29-1', Aliases='TCRBV29S1, TCRBV4S1A1T, TRBV291', description='T cell receptor beta variable 29-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142740273, end_position_on_the_genomic_accession=142740894, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28559, CurrentID=0, Status='live', Symbol='TRBV28', Aliases='TCRBV28S1, TCRBV3S1', description='T cell receptor beta variable 28', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142720680, end_position_on_the_genomic_accession=142721160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28560, CurrentID=0, Status='live', Symbol='TRBV27', Aliases='TCRBV14S1, TCRBV27S1', description='T cell receptor beta variable 27', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142715389, end_position_on_the_genomic_accession=142715861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28561, CurrentID=0, Status='live', Symbol='TRBV26', Aliases='TCRBV26S1, TCRBV28S1P', description='T cell receptor beta variable 26 (pseudogene)', other_designations='T cell receptor beta variable 26 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142695699, end_position_on_the_genomic_accession=142696183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28562, CurrentID=0, Status='live', Symbol='TRBV25-1', Aliases='TCRBV11S1A1T, TCRBV25S1, TRBV251', description='T cell receptor beta variable 25-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142670777, end_position_on_the_genomic_accession=142671244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28563, CurrentID=0, Status='live', Symbol='TRBV24-1', Aliases='TCRBV15S1, TCRBV24S1, TRBV241', description='T cell receptor beta variable 24-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142656737, end_position_on_the_genomic_accession=142657213, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28564, CurrentID=0, Status='live', Symbol='TRBV23-1', Aliases='TCRBV19S1P, TCRBV23S1, TRBV231', description='T cell receptor beta variable 23-1 (non-functional)', other_designations='T cell receptor beta variable 23-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142645961, end_position_on_the_genomic_accession=142646467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28565, CurrentID=0, Status='live', Symbol='TRBV22-1', Aliases='TCRBV22S1, TCRBV29S1P, TRBV22', description='T cell receptor beta variable 22-1 (pseudogene)', other_designations='T cell receptor beta variable 22 pseudogene|T cell receptor beta variable 22-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142641746, end_position_on_the_genomic_accession=142642196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28566, CurrentID=0, Status='live', Symbol='TRBV21-1', Aliases='TCRBV10S1P, TCRBV21S1, TRBV211', description='T cell receptor beta variable 21-1 (pseudogene)', other_designations='T cell receptor beta variable 21-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142636924, end_position_on_the_genomic_accession=142637384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28567, CurrentID=0, Status='live', Symbol='TRBV20-1', Aliases='TCRBV20S1, TCRBV2S1, TRBV201', description='T cell receptor beta variable 20-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142626727, end_position_on_the_genomic_accession=142627399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28568, CurrentID=0, Status='live', Symbol='TRBV19', Aliases='TCRBV17S1A1T, TCRBV19S1', description='T cell receptor beta variable 19', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142619057, end_position_on_the_genomic_accession=142619532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28569, CurrentID=0, Status='live', Symbol='TRBV18', Aliases='TCRBV18S1', description='T cell receptor beta variable 18', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142615797, end_position_on_the_genomic_accession=142616415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28570, CurrentID=0, Status='live', Symbol='TRBV17', Aliases='TCRBV17S1, TCRBV26S1P', description='T cell receptor beta variable 17 (non-functional)', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142601628, end_position_on_the_genomic_accession=142602360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28571, CurrentID=0, Status='live', Symbol='TRBV16', Aliases='BV25S1J1.2, TCRB, TCRBV16S1, TCRBV25S1, TCRBV25S1A2PT', description='T cell receptor beta variable 16 (gene/pseudogene)', other_designations='T-cell receptor beta chain V region', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142598016, end_position_on_the_genomic_accession=142598469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28572, CurrentID=0, Status='live', Symbol='TRBV15', Aliases='TCRBV15S1, TCRBV24S1A3T', description='T cell receptor beta variable 15', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142593005, end_position_on_the_genomic_accession=142593473, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28573, CurrentID=0, Status='live', Symbol='TRBV14', Aliases='TCRBV14S1, TCRBV16S1A1N1', description='T cell receptor beta variable 14', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142587927, end_position_on_the_genomic_accession=142588359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28574, CurrentID=0, Status='live', Symbol='TRBV13', Aliases='TCRBV13S1, TCRBV23S1A2T', description='T cell receptor beta variable 13', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142535809, end_position_on_the_genomic_accession=142536292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28575, CurrentID=0, Status='live', Symbol='TRBV12-5', Aliases='TCRBV12S5, TCRBV8S3, TRBV125', description='T cell receptor beta variable 12-5', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142580981, end_position_on_the_genomic_accession=142581427, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28576, CurrentID=0, Status='live', Symbol='TRBV12-4', Aliases='TCRBV12S4, TCRBV8S2A1T, TRBV124', description='T cell receptor beta variable 12-4', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142563799, end_position_on_the_genomic_accession=142564245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28577, CurrentID=0, Status='live', Symbol='TRBV12-3', Aliases='TCRBV12S3, TCRBV8S1, TRBV123', description='T cell receptor beta variable 12-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142560485, end_position_on_the_genomic_accession=142560931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28578, CurrentID=0, Status='live', Symbol='TRBV12-2', Aliases='TCRBV12S2, TCRBV8S5P, TRBV122', description='T cell receptor beta variable 12-2 (pseudogene)', other_designations='T cell receptor beta variable 12-2 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142440883, end_position_on_the_genomic_accession=142441325, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28579, CurrentID=0, Status='live', Symbol='TRBV12-1', Aliases='TCRBV12S1, TCRBV8S4P, TRBV121', description='T cell receptor beta variable 12-1 (pseudogene)', other_designations='T cell receptor beta variable 12-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142415224, end_position_on_the_genomic_accession=142415666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28580, CurrentID=0, Status='live', Symbol='TRBV11-3', Aliases='TCRBV11S3, TCRBV21S2A2, TRBV113', description='T cell receptor beta variable 11-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142554881, end_position_on_the_genomic_accession=142555318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28581, CurrentID=0, Status='live', Symbol='TRBV11-2', Aliases='TCRBV11S2, TCRBV21S3A2N2T, TRBV112', description='T cell receptor beta variable 11-2', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142433956, end_position_on_the_genomic_accession=142434394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28582, CurrentID=0, Status='live', Symbol='TRBV11-1', Aliases='TCRBV11S1, TCRBV21S1, TRBV111', description='T cell receptor beta variable 11-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142407689, end_position_on_the_genomic_accession=142408136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28583, CurrentID=0, Status='live', Symbol='TRBV10-3', Aliases='TCRBV10S3, TCRBV12S1A1N2, TRBV103', description='T cell receptor beta variable 10-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142544236, end_position_on_the_genomic_accession=142544685, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28584, CurrentID=0, Status='live', Symbol='TRBV10-2', Aliases='TCRBV10S2, TCRBV12S3, TRBV102', description='T cell receptor beta variable 10-2', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142425016, end_position_on_the_genomic_accession=142425465, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28585, CurrentID=0, Status='live', Symbol='TRBV10-1', Aliases='TCRBV10S1, TCRBV12S2, TCRBV12S2A1T, TRBV101', description='T cell receptor beta variable 10-1(gene/pseudogene)', other_designations='T cell receptor beta variable 10-1|V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142399928, end_position_on_the_genomic_accession=142400377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28586, CurrentID=0, Status='live', Symbol='TRBV9', Aliases='TCRBV1S1A1N1, TCRBV9S1', description='T cell receptor beta variable 9', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142391938, end_position_on_the_genomic_accession=142392412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28587, CurrentID=0, Status='live', Symbol='TRBV8-2', Aliases='TCRBV32S1P, TRBV82', description='T cell receptor beta variable 8-2 (pseudogene)', other_designations='T cell receptor beta variable 8-2 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142386187, end_position_on_the_genomic_accession=142386647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28588, CurrentID=0, Status='live', Symbol='TRBV8-1', Aliases='TCRBV30S1P, TCRBV8S1P, TRBV81', description='T cell receptor beta variable 8-1 (pseudogene)', other_designations='T cell receptor beta variable 8-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142358631, end_position_on_the_genomic_accession=142358917, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28589, CurrentID=0, Status='live', Symbol='TRBV7-9', Aliases='TCRBV6S4A1, TCRBV7S9, TRBV79', description='T cell receptor beta variable 7-9', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142529290, end_position_on_the_genomic_accession=142529762, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28590, CurrentID=0, Status='live', Symbol='TRBV7-8', Aliases='TCRBV6S2A1N1T, TCRBV7S8, TRBV78', description='T cell receptor beta variable 7-8', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28591, CurrentID=0, Status='live', Symbol='TRBV7-7', Aliases='TCRBV6S6A2T, TCRBV7S7, TRBV77', description='T cell receptor beta variable 7-7', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142511626, end_position_on_the_genomic_accession=142512127, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28592, CurrentID=0, Status='live', Symbol='TRBV7-6', Aliases='TCRBV6S3A1N1T, TCRBV7S6, TRBV76', description='T cell receptor beta variable 7-6', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142492177, end_position_on_the_genomic_accession=142492673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28593, CurrentID=0, Status='live', Symbol='TRBV7-5', Aliases='TCRBV6S9P, TCRBV7S5, TRBV75', description='T cell receptor beta variable 7-5 (pseudogene)', other_designations='T cell receptor beta variable 7-5 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142474096, end_position_on_the_genomic_accession=142474567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28594, CurrentID=0, Status='live', Symbol='TRBV7-4', Aliases='TCRBV6S8A2T, TCRBV7S4, TRBV74', description='T cell receptor beta variable 7-4 (gene/pseudogene)', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142455174, end_position_on_the_genomic_accession=142455635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28595, CurrentID=0, Status='live', Symbol='TRBV7-3', Aliases='TCRBV6S1A1N1, TCRBV7S3, TRBV73', description='T cell receptor beta variable 7-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142384379, end_position_on_the_genomic_accession=142384841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28596, CurrentID=0, Status='live', Symbol='TRBV7-2', Aliases='TCRBV6S5A1N1, TCRBV6S5A2, TCRBV7S2, TRBV72', description='T cell receptor beta variable 7-2', other_designations='T-cell receptor beta chain V region|V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142352866, end_position_on_the_genomic_accession=142353358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28597, CurrentID=0, Status='live', Symbol='TRBV7-1', Aliases='TCRBV6S7P, TCRBV7S1, TRBV71', description='T cell receptor beta variable 7-1 (non-functional)', other_designations='T cell receptor beta variable 7-1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142332215, end_position_on_the_genomic_accession=142332701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28598, CurrentID=0, Status='live', Symbol='TRBV6-9', Aliases='TCRBV13S4, TCRBV6S9, TRBV69', description='T cell receptor beta variable 6-9', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28599, CurrentID=0, Status='live', Symbol='TRBV6-8', Aliases='TCRBV13S7P, TCRBV6S8, TRBV68', description='T cell receptor beta variable 6-8', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142507382, end_position_on_the_genomic_accession=142507810, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28600, CurrentID=0, Status='live', Symbol='TRBV6-7', Aliases='TCRBV13S8P, TCRBV6S7, TRBV67', description='T cell receptor beta variable 6-7 (non-functional)', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142487863, end_position_on_the_genomic_accession=142488295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  552889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552889, CurrentID=0, Status='live', Symbol='ATXN7L3B', Aliases='lnc-SCA7', description='ataxin 7-like 3B', other_designations='', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74537771, end_position_on_the_genomic_accession=74541452, orientation='plus', exon_count=1, OMIM=615579, no_hdr0=''),
  28602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28602, CurrentID=0, Status='live', Symbol='TRBV6-5', Aliases='TCRBV13S1, TCRBV6S5, TRBV65', description='T cell receptor beta variable 6-5', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142451013, end_position_on_the_genomic_accession=142451448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  552891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552891, CurrentID=0, Status='live', Symbol='DNAJC25-GNG10', Aliases='', description='DNAJC25-GNG10 readthrough', other_designations='DNAJC25-GNG10 readthrough transcript', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111631352, end_position_on_the_genomic_accession=111670246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  28604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28604, CurrentID=0, Status='live', Symbol='TRBV6-3', Aliases='TCRBV13S9/13S2A1T, TCRBV6S3, TRBV63', description='T cell receptor beta variable 6-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28605, CurrentID=0, Status='live', Symbol='TRBV6-2', Aliases='TCRBV13S2, TCRBV13S2A1T, TCRBV6S2, TRBV62', description='T cell receptor beta variable 6-2 (gene/pseudogene)', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142349232, end_position_on_the_genomic_accession=142349664, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28606, CurrentID=0, Status='live', Symbol='TRBV6-1', Aliases='TCRBV13S3, TCRBV6S1, TRBV61', description='T cell receptor beta variable 6-1', other_designations='T-cell receptor beta chain V region C5 -like, V_segment translation product|V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142328354, end_position_on_the_genomic_accession=142328786, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28607, CurrentID=0, Status='live', Symbol='TRBV5-8', Aliases='TCRBV5S4A2T, TCRBV5S8, TRBV58', description='T cell receptor beta variable 5-8', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28608, CurrentID=0, Status='live', Symbol='TRBV5-7', Aliases='TCRBV5S7, TCRBV5S7P, TRBV57', description='T cell receptor beta variable 5-7 (non-functional)', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142520090, end_position_on_the_genomic_accession=142520556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28609, CurrentID=0, Status='live', Symbol='TRBV5-6', Aliases='TCRBV5S2, TCRBV5S6, TRBV56', description='T cell receptor beta variable 5-6', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142500069, end_position_on_the_genomic_accession=142500534, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28610, CurrentID=0, Status='live', Symbol='TRBV5-5', Aliases='TCRBV5S3A2T, TCRBV5S5, TRBV55', description='T cell receptor beta variable 5-5', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142482555, end_position_on_the_genomic_accession=142483019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28611, CurrentID=0, Status='live', Symbol='TRBV5-4', Aliases='TCRBV5S4, TCRBV5S6A3N2T, TRBV54', description='T cell receptor beta variable 5-4', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142463117, end_position_on_the_genomic_accession=142463581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  552900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552900, CurrentID=0, Status='live', Symbol='BOLA2', Aliases='BOLA2A, My016', description='bolA family member 2', other_designations='BolA-like protein 2 member A|bolA homolog 2|bolA-like 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29443249, end_position_on_the_genomic_accession=29455128, orientation='minus', exon_count=7, OMIM=613182, no_hdr0=''),
  28613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28613, CurrentID=0, Status='live', Symbol='TRBV5-2', Aliases='TCRBV31S1, TCRBV5S2P, TRBV52', description='T cell receptor beta variable 5-2 (pseudogene)', other_designations='T cell receptor beta variable 5-2 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142372473, end_position_on_the_genomic_accession=142372913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  159686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159686, CurrentID=0, Status='live', Symbol='CFAP58', Aliases='C10orf80, CCDC147, bA127L20.4, bA127L20.5, bA554P13.1', description='cilia and flagella associated protein 58', other_designations='bA127L20.4 (novel protein)|bA127L20.5 (novel protein)|coiled-coil domain containing 147|coiled-coil domain-containing protein 147', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104338558, end_position_on_the_genomic_accession=104455090, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  28615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28615, CurrentID=0, Status='live', Symbol='TRBV4-3', Aliases='TCRBV4S3, TCRBV7S2A1N4T, TRBV43', description='T cell receptor beta variable 4-3', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28616, CurrentID=0, Status='live', Symbol='TRBV4-2', Aliases='TCRBV4S2, TCRBV7S3A2, TCRBV7S3A2T, TRBV42', description='T cell receptor beta variable 4-2', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142345532, end_position_on_the_genomic_accession=142345985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28617, CurrentID=0, Status='live', Symbol='TRBV4-1', Aliases='BV07S1J2.7, TCRBV4S1, TCRBV7S1A1N2T, TRBV41', description='T cell receptor beta variable 4-1', other_designations='T-cell receptor beta chain V region 86T1|V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142313213, end_position_on_the_genomic_accession=142313666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28618, CurrentID=0, Status='live', Symbol='TRBV3-2', Aliases='TCRBV3S2, TCRBV9S2A2PT, TRBV32', description='T cell receptor beta variable 3-2 (pseudogene)', other_designations='T cell receptor beta variable 3-2 pseudogene|TCR Vbeta 9.2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28619, CurrentID=0, Status='live', Symbol='TRBV3-1', Aliases='TCRBV3S1, TCRBV9S1A1T, TRBV31', description='T cell receptor beta variable 3-1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142308589, end_position_on_the_genomic_accession=142309048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28620, CurrentID=0, Status='live', Symbol='TRBV2', Aliases='TCRBV22S1A2N1T, TCRBV2S1', description='T cell receptor beta variable 2', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142300998, end_position_on_the_genomic_accession=142301432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28621, CurrentID=0, Status='live', Symbol='TRBV1', Aliases='TCRBV1S1P, TCRBV27S1P', description='T cell receptor beta variable 1 (pseudogene)', other_designations='T cell receptor beta variable 1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142299011, end_position_on_the_genomic_accession=142299460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28622, CurrentID=0, Status='live', Symbol='TRBJ2-7', Aliases='TCRBJ2S7, TRBJ27', description='T cell receptor beta joining 2-7', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142797456, end_position_on_the_genomic_accession=142797502, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28623, CurrentID=0, Status='live', Symbol='TRBJ2-6', Aliases='TCRBJ2S6, TRBJ26', description='T cell receptor beta joining 2-6', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142797239, end_position_on_the_genomic_accession=142797291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28624, CurrentID=0, Status='live', Symbol='TRBJ2-5', Aliases='TCRBJ2S5, TRBJ25', description='T cell receptor beta joining 2-5', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142797119, end_position_on_the_genomic_accession=142797166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28625, CurrentID=0, Status='live', Symbol='TRBJ2-4', Aliases='TCRBJ2S4, TRBJ24', description='T cell receptor beta joining 2-4', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142796998, end_position_on_the_genomic_accession=142797047, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28626, CurrentID=0, Status='live', Symbol='TRBJ2-3', Aliases='TCRBJ2S3, TRBJ23', description='T cell receptor beta joining 2-3', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142796847, end_position_on_the_genomic_accession=142796895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28627, CurrentID=0, Status='live', Symbol='TRBJ2-2P', Aliases='TRBJ22P', description='T cell receptor beta joining 2-2P (non-functional)', other_designations='T cell receptor beta joining 2-2P pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142796697, end_position_on_the_genomic_accession=142796742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28628, CurrentID=0, Status='live', Symbol='TRBJ2-2', Aliases='TCRBJ2S2, TRBJ22', description='T cell receptor beta joining 2-2', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142796560, end_position_on_the_genomic_accession=142796610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28629, CurrentID=0, Status='live', Symbol='TRBJ2-1', Aliases='TCRBJ2S1, TRBJ21', description='T cell receptor beta joining 2-1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142796365, end_position_on_the_genomic_accession=142796414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28630, CurrentID=0, Status='live', Symbol='TRBJ1-6', Aliases='TCRBJ1S6, TRBJ16', description='T cell receptor beta joining 1-6', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142788988, end_position_on_the_genomic_accession=142789040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28631, CurrentID=0, Status='live', Symbol='TRBJ1-5', Aliases='TCRBJ1S5, TRBJ15', description='T cell receptor beta joining 1-5', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142788498, end_position_on_the_genomic_accession=142788547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28632, CurrentID=0, Status='live', Symbol='TRBJ1-4', Aliases='TCRBJ1S4, TRBJ14', description='T cell receptor beta joining 1-4', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142788225, end_position_on_the_genomic_accession=142788275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28633, CurrentID=0, Status='live', Symbol='TRBJ1-3', Aliases='TCRBJ1S3, TRBJ13', description='T cell receptor beta joining 1-3', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142787630, end_position_on_the_genomic_accession=142787679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28634, CurrentID=0, Status='live', Symbol='TRBJ1-2', Aliases='TCRBJ1S2, TRBJ12', description='T cell receptor beta joining 1-2', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142787017, end_position_on_the_genomic_accession=142787064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28635, CurrentID=0, Status='live', Symbol='TRBJ1-1', Aliases='TCRBJ1S1, TRBJ11', description='T cell receptor beta joining 1-1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142786880, end_position_on_the_genomic_accession=142786927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28636, CurrentID=0, Status='live', Symbol='TRBD2', Aliases='TCRBD2', description='T cell receptor beta diversity 2', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142795705, end_position_on_the_genomic_accession=142795720, orientation='plus', exon_count=0, OMIM=615448, no_hdr0=''),
  28637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28637, CurrentID=0, Status='live', Symbol='TRBD1', Aliases='TCRBD1', description='T cell receptor beta diversity 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142786213, end_position_on_the_genomic_accession=142786224, orientation='plus', exon_count=0, OMIM=615447, no_hdr0=''),
  28638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28638, CurrentID=0, Status='live', Symbol='TRBC2', Aliases='TCRBC2', description='T cell receptor beta constant 2', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142801041, end_position_on_the_genomic_accession=142802529, orientation='plus', exon_count=0, OMIM=615445, no_hdr0=''),
  28639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28639, CurrentID=0, Status='live', Symbol='TRBC1', Aliases='BV05S1J2.2, TCRB, TCRBC1', description='T cell receptor beta constant 1', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142791694, end_position_on_the_genomic_accession=142793141, orientation='plus', exon_count=0, OMIM=186930, no_hdr0=''),
  28640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28640, CurrentID=0, Status='live', Symbol='TRAV41', Aliases='TCRAV19S1, TCRAV41S1', description='T cell receptor alpha variable 41', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22320188, end_position_on_the_genomic_accession=22320691, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28641, CurrentID=0, Status='live', Symbol='TRAV40', Aliases='TCRAV31S1, TCRAV40S1', description='T cell receptor alpha variable 40', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22314490, end_position_on_the_genomic_accession=22314919, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28642, CurrentID=0, Status='live', Symbol='TRAV39', Aliases='TCRAV27S1, TCRAV39S1', description='T cell receptor alpha variable 39', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22304054, end_position_on_the_genomic_accession=22304553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28643, CurrentID=0, Status='live', Symbol='TRAV38-2DV8', Aliases='TCRAV14S1, TRAV382DV8, hADV38S2', description='T cell receptor alpha variable 38-2/delta variable 8', other_designations='T cell receptor alpha/delta variable 38-2/DV8|TRAV38-2/DV8', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22281151, end_position_on_the_genomic_accession=22281748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28644, CurrentID=0, Status='live', Symbol='TRAV38-1', Aliases='TCRAV14S2, TCRAV38S1, TRAV381', description='T cell receptor alpha variable 38-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22271971, end_position_on_the_genomic_accession=22272563, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266918, CurrentID=0, Status='live', Symbol='D21S2090E', Aliases='', description='uncharacterized D21S2090E', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28646, CurrentID=0, Status='live', Symbol='TRAV36DV7', Aliases='TCRAV28S1, TRAV36/DV7, hADV36S1', description='T cell receptor alpha variable 36/delta variable 7', other_designations='T cell receptor alpha/delta variable 36/DV7', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22226762, end_position_on_the_genomic_accession=22227254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28647, CurrentID=0, Status='live', Symbol='TRAV35', Aliases='TCRAV25S1, TCRAV35S1', description='T cell receptor alpha variable 35', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22221896, end_position_on_the_genomic_accession=22222475, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28648, CurrentID=0, Status='live', Symbol='TRAV34', Aliases='TCRAV26S1, TCRAV34S1', description='T cell receptor alpha variable 34', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22207534, end_position_on_the_genomic_accession=22208129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28649, CurrentID=0, Status='live', Symbol='TRAV33', Aliases='C14orf12, TCRAV33S1', description='T cell receptor alpha variable 33 (pseudogene)', other_designations='T cell receptor alpha variable 33 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22190158, end_position_on_the_genomic_accession=22190714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28650, CurrentID=0, Status='live', Symbol='TRAV32', Aliases='TCRAV32S1', description='T cell receptor alpha variable 32 (pseudogene)', other_designations='T cell receptor alpha variable 32 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22185562, end_position_on_the_genomic_accession=22186061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266919, CurrentID=0, Status='live', Symbol='LINC00307', Aliases='D21S2091E, NCRNA00307', description='long intergenic non-protein coding RNA 307', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30209151, end_position_on_the_genomic_accession=30211783, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28652, CurrentID=0, Status='live', Symbol='TRAV30', Aliases='TCRAV29S1, TCRAV30S1', description='T cell receptor alpha variable 30', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22168429, end_position_on_the_genomic_accession=22168988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28653, CurrentID=0, Status='live', Symbol='TRAV29DV5', Aliases='TCRAV21S1, TRAV29/DV5, hADV29S1', description='T cell receptor alpha variable 29/delta variable 5 (gene/pseudogene)', other_designations='T cell receptor alpha/delta variable 29/DV5', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22163349, end_position_on_the_genomic_accession=22163870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28654, CurrentID=0, Status='live', Symbol='TRAV28', Aliases='TCRAV28S1', description='T cell receptor alpha variable 28 (pseudogene)', other_designations='T cell receptor alpha variable 28 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22155079, end_position_on_the_genomic_accession=22155638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28655, CurrentID=0, Status='live', Symbol='TRAV27', Aliases='TCRAV10S1, TCRAV27S1', description='T cell receptor alpha variable 27', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22148092, end_position_on_the_genomic_accession=22148633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28656, CurrentID=0, Status='live', Symbol='TRAV26-2', Aliases='TCRAV26S2, TCRAV4S1, TRAV262', description='T cell receptor alpha variable 26-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22202583, end_position_on_the_genomic_accession=22203368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266920, CurrentID=0, Status='live', Symbol='AK2P1', Aliases='AK2B', description='adenylate kinase 2 pseudogene 1', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58630846, end_position_on_the_genomic_accession=58631572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28658, CurrentID=0, Status='live', Symbol='TRAV25', Aliases='TCRAV25S1, TCRAV32S1', description='T cell receptor alpha variable 25', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22112423, end_position_on_the_genomic_accession=22113031, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28659, CurrentID=0, Status='live', Symbol='TRAV24', Aliases='TCRAV18S1, TCRAV24S1', description='T cell receptor alpha variable 24', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22105343, end_position_on_the_genomic_accession=22105846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28660, CurrentID=0, Status='live', Symbol='TRAV23DV6', Aliases='TCRAV17S1, TRAV23/DV6, hADV23S1', description='T cell receptor alpha variable 23/delta variable 6', other_designations='T cell receptor alpha/delta variable 23/DV6', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22086452, end_position_on_the_genomic_accession=22086961, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28661, CurrentID=0, Status='live', Symbol='TRAV22', Aliases='TCRAV13S1, TCRAV22S1', description='T cell receptor alpha variable 22', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22070690, end_position_on_the_genomic_accession=22071208, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28662, CurrentID=0, Status='live', Symbol='TRAV21', Aliases='TCRAV21S1, TCRAV23S1', description='T cell receptor alpha variable 21', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22052521, end_position_on_the_genomic_accession=22053056, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28663, CurrentID=0, Status='live', Symbol='TRAV20', Aliases='TCRAV20S1, TCRAV30S1', description='T cell receptor alpha variable 20', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22040663, end_position_on_the_genomic_accession=22041153, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28664, CurrentID=0, Status='live', Symbol='TRAV19', Aliases='TCRAV12S1, TCRAV19S1', description='T cell receptor alpha variable 19', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22007629, end_position_on_the_genomic_accession=22008181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28665, CurrentID=0, Status='live', Symbol='TRAV18', Aliases='TCRAV18S1', description='T cell receptor alpha variable 18', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22003211, end_position_on_the_genomic_accession=22003673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28666, CurrentID=0, Status='live', Symbol='TRAV17', Aliases='TCRAV17S1, TCRAV3S1', description='T cell receptor alpha variable 17', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21997686, end_position_on_the_genomic_accession=21998168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131890, CurrentID=0, Status='live', Symbol='GRK7', Aliases='GPRK7', description='G protein-coupled receptor kinase 7', other_designations='g protein-coupled receptor kinase GRK7', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141749984, end_position_on_the_genomic_accession=141817370, orientation='plus', exon_count=7, OMIM=606987, no_hdr0=''),
  28668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28668, CurrentID=0, Status='live', Symbol='TRAV15', Aliases='TCRAV15S1', description='T cell receptor alpha variable 15 (pseudogene)', other_designations='T cell receptor alpha variable 15 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21950127, end_position_on_the_genomic_accession=21950654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28669, CurrentID=0, Status='live', Symbol='TRAV14DV4', Aliases='TCRAV6S1-hDV104S1, TRAV14/DV4, hADV14S1', description='T cell receptor alpha variable 14/delta variable 4', other_designations='T cell receptor alpha/delta variable 14/DV4', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21924138, end_position_on_the_genomic_accession=21924651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28670, CurrentID=0, Status='live', Symbol='TRAV13-2', Aliases='TCRAV13S2, TCRAV8S2, TRAV132', description='T cell receptor alpha variable 13-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21918258, end_position_on_the_genomic_accession=21918756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28671, CurrentID=0, Status='live', Symbol='TRAV13-1', Aliases='TCRAV13S1, TCRAV8S1, TRAV131', description='T cell receptor alpha variable 13-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21868862, end_position_on_the_genomic_accession=21869365, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28672, CurrentID=0, Status='live', Symbol='TRAV12-3', Aliases='TCRAV12S3, TCRAV2S2, TRAV123', description='T cell receptor alpha variable 12-3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21965507, end_position_on_the_genomic_accession=21966061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28673, CurrentID=0, Status='live', Symbol='TRAV12-2', Aliases='TCRAV12S2, TCRAV2S1, TRAV122', description='T cell receptor alpha variable 12-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21887959, end_position_on_the_genomic_accession=21888502, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28674, CurrentID=0, Status='live', Symbol='TRAV12-1', Aliases='TCRAV12S1, TCRAV2S3, TRAV121', description='T cell receptor alpha variable 12-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21841243, end_position_on_the_genomic_accession=21841774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28675, CurrentID=0, Status='live', Symbol='TRAV11', Aliases='TCRAV11S1', description='T cell receptor alpha variable 11 (pseudogene)', other_designations='T cell receptor alpha variable 11 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21829539, end_position_on_the_genomic_accession=21830070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28676, CurrentID=0, Status='live', Symbol='TRAV10', Aliases='TCRAV10S1, TCRAV24S1', description='T cell receptor alpha variable 10', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21825508, end_position_on_the_genomic_accession=21826075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28677, CurrentID=0, Status='live', Symbol='TRAV9-2', Aliases='TCRAV22S1, TCRAV9S2, TRAV92', description='T cell receptor alpha variable 9-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21941184, end_position_on_the_genomic_accession=21941657, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28678, CurrentID=0, Status='live', Symbol='TRAV9-1', Aliases='TCRAV9S1, TRAV91', description='T cell receptor alpha variable 9-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21811502, end_position_on_the_genomic_accession=21811977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28679, CurrentID=0, Status='live', Symbol='TRAV8-7', Aliases='TCRAV8S7, TRAV87', description='T cell receptor alpha variable 8-7 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22132553, end_position_on_the_genomic_accession=22133034, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28680, CurrentID=0, Status='live', Symbol='TRAV8-6', Aliases='TCRAV1S3, TCRAV8S6, TRAV86', description='T cell receptor alpha variable 8-6', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21978679, end_position_on_the_genomic_accession=21979120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28681, CurrentID=0, Status='live', Symbol='TRAV8-5', Aliases='TCRAV8S5, TRAV85', description='T cell receptor alpha variable 8-5 (pseudogene)', other_designations='T cell receptor alpha variable 8-5 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21903077, end_position_on_the_genomic_accession=21904598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28682, CurrentID=0, Status='live', Symbol='TRAV8-4', Aliases='TCRAV1S2, TCRAV8S4, TRAV84', description='T cell receptor alpha variable 8-4', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21894561, end_position_on_the_genomic_accession=21895030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28683, CurrentID=0, Status='live', Symbol='TRAV8-3', Aliases='TCRAV1S4, TCRAV8S3, TRAV83', description='T cell receptor alpha variable 8-3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21852558, end_position_on_the_genomic_accession=21853006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28684, CurrentID=0, Status='live', Symbol='TRAV8-2', Aliases='TCRAV1S5, TCRAV8S2, TRAV82', description='T cell receptor alpha variable 8-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21846762, end_position_on_the_genomic_accession=21847221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28685, CurrentID=0, Status='live', Symbol='TRAV8-1', Aliases='TCRAV1S1, TCRAV8S1, TRAV81', description='T cell receptor alpha variable 8-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21797419, end_position_on_the_genomic_accession=21797886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28686, CurrentID=0, Status='live', Symbol='TRAV7', Aliases='TCRAV7S1', description='T cell receptor alpha variable 7', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21782993, end_position_on_the_genomic_accession=21783503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28688, CurrentID=0, Status='live', Symbol='TRAV5', Aliases='TCRAV15S1, TCRAV5S1', description='T cell receptor alpha variable 5', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21749190, end_position_on_the_genomic_accession=21749705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28689, CurrentID=0, Status='live', Symbol='TRAV4', Aliases='TCRAV20S1, TCRAV4S1', description='T cell receptor alpha variable 4', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21736219, end_position_on_the_genomic_accession=21736982, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28690, CurrentID=0, Status='live', Symbol='TRAV3', Aliases='TCRAV16S1, TCRAV3S1', description='T cell receptor alpha variable 3 (gene/pseudogene)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21723889, end_position_on_the_genomic_accession=21724321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28691, CurrentID=0, Status='live', Symbol='TRAV2', Aliases='TCRAV11S1, TCRAV2S1', description='T cell receptor alpha variable 2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21712331, end_position_on_the_genomic_accession=21712843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28692, CurrentID=0, Status='live', Symbol='TRAV1-2', Aliases='TCRAV1S2, TCRAV7S2, TRAV12', description='T cell receptor alpha variable 1-2', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21642973, end_position_on_the_genomic_accession=21643586, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28693, CurrentID=0, Status='live', Symbol='TRAV1-1', Aliases='TCRAV1S1, TCRAV7S1, TRAV11', description='T cell receptor alpha variable 1-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21621904, end_position_on_the_genomic_accession=21622567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28694, CurrentID=0, Status='live', Symbol='TRAJ61', Aliases='', description='T cell receptor alpha joining 61 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22475316, end_position_on_the_genomic_accession=22475375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28695, CurrentID=0, Status='live', Symbol='TRAJ60', Aliases='TCRA', description='T cell receptor alpha joining 60 (pseudogene)', other_designations='T cell receptor alpha joining 60 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22476306, end_position_on_the_genomic_accession=22476362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28696, CurrentID=0, Status='live', Symbol='TRAJ59', Aliases='', description='T cell receptor alpha joining 59 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22476553, end_position_on_the_genomic_accession=22476606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28697, CurrentID=0, Status='live', Symbol='TRAJ58', Aliases='', description='T cell receptor alpha joining 58 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22477707, end_position_on_the_genomic_accession=22477769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  159770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159770, CurrentID=0, Status='live', Symbol='CBX3P1', Aliases='', description='chromobox 3 pseudogene 1', other_designations='HP1-Hs-gamma pseudogene|chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 gamma pseudogene|chromobox homolog 3 pseudogene 1', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27806317, end_position_on_the_genomic_accession=27808558, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28699, CurrentID=0, Status='live', Symbol='TRAJ56', Aliases='', description='T cell receptor alpha joining 56', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22479521, end_position_on_the_genomic_accession=22479582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28700, CurrentID=0, Status='live', Symbol='TRAJ55', Aliases='', description='T cell receptor alpha joining 55 (pseudogene)', other_designations='T cell receptor alpha joining 55 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22481697, end_position_on_the_genomic_accession=22481753, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28701, CurrentID=0, Status='live', Symbol='TRAJ54', Aliases='', description='T cell receptor alpha joining 54', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22482287, end_position_on_the_genomic_accession=22482346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28702, CurrentID=0, Status='live', Symbol='TRAJ53', Aliases='', description='T cell receptor alpha joining 53', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22483004, end_position_on_the_genomic_accession=22483069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28703, CurrentID=0, Status='live', Symbol='TRAJ52', Aliases='', description='T cell receptor alpha joining 52', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22486228, end_position_on_the_genomic_accession=22486296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28704, CurrentID=0, Status='live', Symbol='TRAJ51', Aliases='', description='T cell receptor alpha joining 51 (pseudogene)', other_designations='T cell receptor alpha joining 51 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22487183, end_position_on_the_genomic_accession=22487245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28705, CurrentID=0, Status='live', Symbol='TRAJ50', Aliases='', description='T cell receptor alpha joining 50', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22488593, end_position_on_the_genomic_accession=22488652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28706, CurrentID=0, Status='live', Symbol='TRAJ49', Aliases='', description='T cell receptor alpha joining 49', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22489488, end_position_on_the_genomic_accession=22489543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28707, CurrentID=0, Status='live', Symbol='TRAJ48', Aliases='', description='T cell receptor alpha joining 48', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22490491, end_position_on_the_genomic_accession=22490553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28708, CurrentID=0, Status='live', Symbol='TRAJ47', Aliases='', description='T cell receptor alpha joining 47', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22492851, end_position_on_the_genomic_accession=22492907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28709, CurrentID=0, Status='live', Symbol='TRAJ46', Aliases='', description='T cell receptor alpha joining 46', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22493403, end_position_on_the_genomic_accession=22493465, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28710, CurrentID=0, Status='live', Symbol='TRAJ45', Aliases='', description='T cell receptor alpha joining 45', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22493925, end_position_on_the_genomic_accession=22493990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28711, CurrentID=0, Status='live', Symbol='TRAJ44', Aliases='', description='T cell receptor alpha joining 44', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22494821, end_position_on_the_genomic_accession=22494883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28712, CurrentID=0, Status='live', Symbol='TRAJ43', Aliases='', description='T cell receptor alpha joining 43', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22495913, end_position_on_the_genomic_accession=22495966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28713, CurrentID=0, Status='live', Symbol='TRAJ42', Aliases='', description='T cell receptor alpha joining 42', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22496887, end_position_on_the_genomic_accession=22496952, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28714, CurrentID=0, Status='live', Symbol='TRAJ41', Aliases='', description='T cell receptor alpha joining 41', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22497657, end_position_on_the_genomic_accession=22497718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28715, CurrentID=0, Status='live', Symbol='TRAJ40', Aliases='', description='T cell receptor alpha joining 40', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22499689, end_position_on_the_genomic_accession=22499749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28716, CurrentID=0, Status='live', Symbol='TRAJ39', Aliases='', description='T cell receptor alpha joining 39', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22501601, end_position_on_the_genomic_accession=22501663, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28717, CurrentID=0, Status='live', Symbol='TRAJ38', Aliases='', description='T cell receptor alpha joining 38', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22502231, end_position_on_the_genomic_accession=22502292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28718, CurrentID=0, Status='live', Symbol='TRAJ37', Aliases='', description='T cell receptor alpha joining 37', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22503750, end_position_on_the_genomic_accession=22503811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28719, CurrentID=0, Status='live', Symbol='TRAJ36', Aliases='', description='T cell receptor alpha joining 36', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22505110, end_position_on_the_genomic_accession=22505168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28720, CurrentID=0, Status='live', Symbol='TRAJ35', Aliases='', description='T cell receptor alpha joining 35 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22506644, end_position_on_the_genomic_accession=22506702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28721, CurrentID=0, Status='live', Symbol='TRAJ34', Aliases='', description='T cell receptor alpha joining 34', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22507666, end_position_on_the_genomic_accession=22507723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28722, CurrentID=0, Status='live', Symbol='TRAJ33', Aliases='', description='T cell receptor alpha joining 33', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22508602, end_position_on_the_genomic_accession=22508658, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28723, CurrentID=0, Status='live', Symbol='TRAJ32', Aliases='', description='T cell receptor alpha joining 32', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22509341, end_position_on_the_genomic_accession=22509406, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28724, CurrentID=0, Status='live', Symbol='TRAJ31', Aliases='', description='T cell receptor alpha joining 31', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22510968, end_position_on_the_genomic_accession=22511024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28725, CurrentID=0, Status='live', Symbol='TRAJ30', Aliases='', description='T cell receptor alpha joining 30', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22512852, end_position_on_the_genomic_accession=22512908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28726, CurrentID=0, Status='live', Symbol='TRAJ29', Aliases='', description='T cell receptor alpha joining 29', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22513939, end_position_on_the_genomic_accession=22513998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28727, CurrentID=0, Status='live', Symbol='TRAJ28', Aliases='', description='T cell receptor alpha joining 28', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22515623, end_position_on_the_genomic_accession=22515688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28728, CurrentID=0, Status='live', Symbol='TRAJ27', Aliases='', description='T cell receptor alpha joining 27', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22516273, end_position_on_the_genomic_accession=22516331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28729, CurrentID=0, Status='live', Symbol='TRAJ26', Aliases='', description='T cell receptor alpha joining 26', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22518446, end_position_on_the_genomic_accession=22518505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28730, CurrentID=0, Status='live', Symbol='TRAJ25', Aliases='', description='T cell receptor alpha joining 25 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22518812, end_position_on_the_genomic_accession=22518871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28731, CurrentID=0, Status='live', Symbol='TRAJ24', Aliases='', description='T cell receptor alpha joining 24', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22519969, end_position_on_the_genomic_accession=22520031, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28732, CurrentID=0, Status='live', Symbol='TRAJ23', Aliases='', description='T cell receptor alpha joining 23', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22520416, end_position_on_the_genomic_accession=22520478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28733, CurrentID=0, Status='live', Symbol='TRAJ22', Aliases='', description='T cell receptor alpha joining 22', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22522040, end_position_on_the_genomic_accession=22522102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28734, CurrentID=0, Status='live', Symbol='TRAJ21', Aliases='', description='T cell receptor alpha joining 21', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22523600, end_position_on_the_genomic_accession=22523654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28735, CurrentID=0, Status='live', Symbol='TRAJ20', Aliases='', description='T cell receptor alpha joining 20', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22524325, end_position_on_the_genomic_accession=22524381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28736, CurrentID=0, Status='live', Symbol='TRAJ19', Aliases='', description='T cell receptor alpha joining 19 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22525263, end_position_on_the_genomic_accession=22525322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28737, CurrentID=0, Status='live', Symbol='TRAJ18', Aliases='', description='T cell receptor alpha joining 18', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22525650, end_position_on_the_genomic_accession=22525715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28738, CurrentID=0, Status='live', Symbol='TRAJ17', Aliases='', description='T cell receptor alpha joining 17', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22526844, end_position_on_the_genomic_accession=22526906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28739, CurrentID=0, Status='live', Symbol='TRAJ16', Aliases='', description='T cell receptor alpha joining 16', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22528527, end_position_on_the_genomic_accession=22528586, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28740, CurrentID=0, Status='live', Symbol='TRAJ15', Aliases='', description='T cell receptor alpha joining 15', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22529629, end_position_on_the_genomic_accession=22529688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28741, CurrentID=0, Status='live', Symbol='TRAJ14', Aliases='', description='T cell receptor alpha joining 14', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22530327, end_position_on_the_genomic_accession=22530378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28742, CurrentID=0, Status='live', Symbol='TRAJ13', Aliases='', description='T cell receptor alpha joining 13', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22531076, end_position_on_the_genomic_accession=22531138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28743, CurrentID=0, Status='live', Symbol='TRAJ12', Aliases='', description='T cell receptor alpha joining 12', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22531939, end_position_on_the_genomic_accession=22531998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28744, CurrentID=0, Status='live', Symbol='TRAJ11', Aliases='', description='T cell receptor alpha joining 11', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22532502, end_position_on_the_genomic_accession=22532561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28745, CurrentID=0, Status='live', Symbol='TRAJ10', Aliases='', description='T cell receptor alpha joining 10', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22533497, end_position_on_the_genomic_accession=22533560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28746, CurrentID=0, Status='live', Symbol='TRAJ9', Aliases='', description='T cell receptor alpha joining 9', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22535554, end_position_on_the_genomic_accession=22535614, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28747, CurrentID=0, Status='live', Symbol='TRAJ8', Aliases='', description='T cell receptor alpha joining 8', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22536145, end_position_on_the_genomic_accession=22536204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28748, CurrentID=0, Status='live', Symbol='TRAJ7', Aliases='', description='T cell receptor alpha joining 7', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22537622, end_position_on_the_genomic_accession=22537680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28749, CurrentID=0, Status='live', Symbol='TRAJ6', Aliases='', description='T cell receptor alpha joining 6', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22539073, end_position_on_the_genomic_accession=22539134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28750, CurrentID=0, Status='live', Symbol='TRAJ5', Aliases='', description='T cell receptor alpha joining 5', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22540247, end_position_on_the_genomic_accession=22540306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28751, CurrentID=0, Status='live', Symbol='TRAJ4', Aliases='', description='T cell receptor alpha joining 4', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22542199, end_position_on_the_genomic_accession=22542261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28752, CurrentID=0, Status='live', Symbol='TRAJ3', Aliases='', description='T cell receptor alpha joining 3', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22543179, end_position_on_the_genomic_accession=22543240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28753, CurrentID=0, Status='live', Symbol='TRAJ2', Aliases='', description='T cell receptor alpha joining 2 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22544071, end_position_on_the_genomic_accession=22544136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28754, CurrentID=0, Status='live', Symbol='TRAJ1', Aliases='', description='T cell receptor alpha joining 1 (non-functional)', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22545037, end_position_on_the_genomic_accession=22545098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28755, CurrentID=0, Status='live', Symbol='TRAC', Aliases='IMD7, TCRA, TRA, TRCA', description='T-cell receptor alpha constant', other_designations='T cell receptor alpha constant', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22547506, end_position_on_the_genomic_accession=22552132, orientation='plus', exon_count=0, OMIM=186880, no_hdr0=''),
  28756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28756, CurrentID=0, Status='live', Symbol='IGLV8OR8-1', Aliases='IGLV8/OR8-1, IGLV8OR81', description='immunoglobulin lambda variable 8/OR8-1 (pseudogene)', other_designations='immunoglobulin lambda variable 8/OR8-1 pseudogene', map_location='8q11.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28757, CurrentID=0, Status='live', Symbol='IGLVV-66', Aliases='IGLV(V)-66, IGLVV66, V5-5P', description='immunoglobulin lambda variable (V)-66 (pseudogene)', other_designations='immunoglobulin lambda variable (V)-66 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22061246, end_position_on_the_genomic_accession=22061557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28758, CurrentID=0, Status='live', Symbol='IGLVV-58', Aliases='IGLV(V)-58, IGLVV58, V5-3P', description='immunoglobulin lambda variable (V)-58 (pseudogene)', other_designations='immunoglobulin lambda variable (V)-58 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22182583, end_position_on_the_genomic_accession=22182890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28759, CurrentID=0, Status='live', Symbol='IGLVIV-65', Aliases='IGLV(IV)-65, IGLVIV65, V4-9P', description='immunoglobulin lambda variable (IV)-65 (pseudogene)', other_designations='immunoglobulin lambda variable (IV)-65 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22070225, end_position_on_the_genomic_accession=22070711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28760, CurrentID=0, Status='live', Symbol='IGLVIV-64', Aliases='IGLV(IV)-64, IGLVIV64, V4-8P', description='immunoglobulin lambda variable (IV)-64 (pseudogene)', other_designations='immunoglobulin lambda variable (IV)-64 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22075186, end_position_on_the_genomic_accession=22075682, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28761, CurrentID=0, Status='live', Symbol='IGLVIV-59', Aliases='IGLV(IV)-59, IGLVIV59, V4-7P', description='immunoglobulin lambda variable (IV)-59 (pseudogene)', other_designations='immunoglobulin lambda variable (IV)-59 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22179165, end_position_on_the_genomic_accession=22179488, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28762, CurrentID=0, Status='live', Symbol='IGLVIV-53', Aliases='IGLV(IV)-53, IGLVIV53, V4-5P', description='immunoglobulin lambda variable (IV)-53 (pseudogene)', other_designations='immunoglobulin lambda variable (IV)-53 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22219647, end_position_on_the_genomic_accession=22220133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28763, CurrentID=0, Status='live', Symbol='IGLVI-70', Aliases='IGLV(I)-70, IGLVI70, V1-27P', description='immunoglobulin lambda variable (I)-70 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-70 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22026076, end_position_on_the_genomic_accession=22026553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28764, CurrentID=0, Status='live', Symbol='IGLVI-68', Aliases='IGLV(I)-68, IGLVI68, V1-26P', description='immunoglobulin lambda variable (I)-68 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-68 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22032902, end_position_on_the_genomic_accession=22033205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791226, CurrentID=0, Status='live', Symbol='TRN-GTT4-1', Aliases='TRN2, TRNAN2', description='transfer RNA-Asn (GTT) 4-1', other_designations='tRNA-Asn-GTT-4-1|transfer RNA asparagine 2 (anticodon GUU)', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16889677, end_position_on_the_genomic_accession=16889750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28766, CurrentID=0, Status='live', Symbol='IGLVI-56', Aliases='IGLV(I)-56, IGLVI56, V1-21P', description='immunoglobulin lambda variable (I)-56 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-56 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22198724, end_position_on_the_genomic_accession=22199019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28767, CurrentID=0, Status='live', Symbol='IGLVI-42', Aliases='IGLV(I)-42, IGLVI42, V1-15P', description='immunoglobulin lambda variable (I)-42 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-42 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22397032, end_position_on_the_genomic_accession=22397490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28768, CurrentID=0, Status='live', Symbol='IGLVI-38', Aliases='IGLV(I)-38, IGLVI38, V1-12P', description='immunoglobulin lambda variable (I)-38 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-38 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22425988, end_position_on_the_genomic_accession=22426278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28769, CurrentID=0, Status='live', Symbol='IGLVI-20', Aliases='IGLV(I)-20, IGLVI20, V1-6P', description='immunoglobulin lambda variable (I)-20 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-20 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22715158, end_position_on_the_genomic_accession=22715443, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28770, CurrentID=0, Status='live', Symbol='IGLV11-55', Aliases='IGLV1155, V4-6', description='immunoglobulin lambda variable 11-55 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22201675, end_position_on_the_genomic_accession=22202161, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791227, CurrentID=0, Status='live', Symbol='TRN-GTT6-1', Aliases='TRN3, TRNAN3', description='transfer RNA-Asn (GTT) 6-1', other_designations='tRNA-Asn-GTT-6-1|transfer RNA asparagine 3 (anticodon GUU)', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143735920, end_position_on_the_genomic_accession=143735993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28772, CurrentID=0, Status='live', Symbol='IGLV10-54', Aliases='IGLV1054, V1-20', description='immunoglobulin lambda variable 10-54', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22214807, end_position_on_the_genomic_accession=22215271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28773, CurrentID=0, Status='live', Symbol='IGLV9-49', Aliases='IGLV949, V5-2', description='immunoglobulin lambda variable 9-49', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22343224, end_position_on_the_genomic_accession=22343734, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28774, CurrentID=0, Status='live', Symbol='IGLV8-61', Aliases='IGLV861, V3-4', description='immunoglobulin lambda variable 8-61', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22098761, end_position_on_the_genomic_accession=22099212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28775, CurrentID=0, Status='live', Symbol='IGLV7-46', Aliases='IGLV746, V3-3', description='immunoglobulin lambda variable 7-46 (gene/pseudogene)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22369648, end_position_on_the_genomic_accession=22370087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28776, CurrentID=0, Status='live', Symbol='IGLV7-43', Aliases='IGLV743, V3-2', description='immunoglobulin lambda variable 7-43', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22395052, end_position_on_the_genomic_accession=22395489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791228, CurrentID=0, Status='live', Symbol='SPG36', Aliases='', description='spastic paraplegia 36 (autosomal dominant)', other_designations='', map_location='12q23-q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613096, no_hdr0=''),
  28778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28778, CurrentID=0, Status='live', Symbol='IGLV6-57', Aliases='IGLV657, V1-22', description='immunoglobulin lambda variable 6-57', other_designations='amyloid lambda 6 light chain variable region PIP|amyloid lambda 6 light chain variable region SAR', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22195799, end_position_on_the_genomic_accession=22196276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28779, CurrentID=0, Status='live', Symbol='IGLV5-52', Aliases='IGLV552, V4-4', description='immunoglobulin lambda variable 5-52', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22318666, end_position_on_the_genomic_accession=22319226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28780, CurrentID=0, Status='live', Symbol='IGLV5-48', Aliases='IGLV548, V4-3', description='immunoglobulin lambda variable 5-48 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22352940, end_position_on_the_genomic_accession=22353433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28781, CurrentID=0, Status='live', Symbol='IGLV5-45', Aliases='IGLV545, V4-2', description='immunoglobulin lambda variable 5-45', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22376014, end_position_on_the_genomic_accession=22376505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28782, CurrentID=0, Status='live', Symbol='IGLV5-39', Aliases='IGLV539', description='immunoglobulin lambda variable 5-39', other_designations='', map_location='22q11.2 alternate reference locus', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28783, CurrentID=0, Status='live', Symbol='IGLV5-37', Aliases='IGLV537, V4-1', description='immunoglobulin lambda variable 5-37', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22427545, end_position_on_the_genomic_accession=22428035, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28784, CurrentID=0, Status='live', Symbol='IGLV4-69', Aliases='IGLV469, V5-6', description='immunoglobulin lambda variable 4-69', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22030994, end_position_on_the_genomic_accession=22031472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28785, CurrentID=0, Status='live', Symbol='IGLV4-60', Aliases='IGLV460, V5-4', description='immunoglobulin lambda variable 4-60', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22162198, end_position_on_the_genomic_accession=22162681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28786, CurrentID=0, Status='live', Symbol='IGLV4-3', Aliases='IGLV43, V5-1', description='immunoglobulin lambda variable 4-3', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22871539, end_position_on_the_genomic_accession=22872035, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28787, CurrentID=0, Status='live', Symbol='IGLV3-32', Aliases='IGLV332, V2-23P', description='immunoglobulin lambda variable 3-32 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22594567, end_position_on_the_genomic_accession=22595056, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28788, CurrentID=0, Status='live', Symbol='IGLV3-31', Aliases='IGLV331, V2-22P', description='immunoglobulin lambda variable 3-31 (pseudogene)', other_designations='immunoglobulin lambda variable 3-31 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22604693, end_position_on_the_genomic_accession=22605185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  791230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791230, CurrentID=0, Status='live', Symbol='TRV-AAC1-5', Aliases='HtV1, TRNAV7, TRV7', description='transfer RNA-Val (AAC) 1-5', other_designations='tRNA-Val-AAC-1-5|transfer RNA valine 7 (anticodon AAC)|transfer RNA valine 7 (anticodon RAC)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27753400, end_position_on_the_genomic_accession=27753472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28790, CurrentID=0, Status='live', Symbol='IGLV3-29', Aliases='IGLV329', description='immunoglobulin lambda variable 3-29 (pseudogene)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22660949, end_position_on_the_genomic_accession=22661577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28791, CurrentID=0, Status='live', Symbol='IGLV3-27', Aliases='IGLV327, V2-19', description='immunoglobulin lambda variable 3-27', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22668326, end_position_on_the_genomic_accession=22668808, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28792, CurrentID=0, Status='live', Symbol='IGLV3-26', Aliases='IGLV326, V2-18P', description='immunoglobulin lambda variable 3-26 (pseudogene)', other_designations='immunoglobulin lambda variable 3-26 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22673338, end_position_on_the_genomic_accession=22673637, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28793, CurrentID=0, Status='live', Symbol='IGLV3-25', Aliases='IGLV325, V2-17', description='immunoglobulin lambda variable 3-25', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22686769, end_position_on_the_genomic_accession=22687281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28794, CurrentID=0, Status='live', Symbol='IGLV3-24', Aliases='IGLV324, V2-16P', description='immunoglobulin lambda variable 3-24 (pseudogene)', other_designations='immunoglobulin lambda variable 3-24 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22694776, end_position_on_the_genomic_accession=22695070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28795, CurrentID=0, Status='live', Symbol='IGLV3-22', Aliases='IGLV322, V2-15', description='immunoglobulin lambda variable 3-22 (gene/pseudogene)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22704329, end_position_on_the_genomic_accession=22704818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28796, CurrentID=0, Status='live', Symbol='IGLV3-21', Aliases='IGLV321, V2-14', description='immunoglobulin lambda variable 3-21', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22712419, end_position_on_the_genomic_accession=22713199, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28797, CurrentID=0, Status='live', Symbol='IGLV3-19', Aliases='IGLV319, V2-13, VL3L', description='immunoglobulin lambda variable 3-19', other_designations='rearranged Vl3l gene segment|rearranged Vl3l segment', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22720663, end_position_on_the_genomic_accession=22721155, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28798, CurrentID=0, Status='live', Symbol='IGLV3-17', Aliases='IGLV317, V2-12P', description='immunoglobulin lambda variable 3-17 (pseudogene)', other_designations='immunoglobulin lambda variable 3-17 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22738911, end_position_on_the_genomic_accession=22739191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28799, CurrentID=0, Status='live', Symbol='IGLV3-16', Aliases='IGLV316, V2-11', description='immunoglobulin lambda variable 3-16', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22747421, end_position_on_the_genomic_accession=22747921, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28800, CurrentID=0, Status='live', Symbol='IGLV3-15', Aliases='IGLV315, V2-10P', description='immunoglobulin lambda variable 3-15 (pseudogene)', other_designations='immunoglobulin lambda variable 3-15 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22755553, end_position_on_the_genomic_accession=22755837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28801, CurrentID=0, Status='live', Symbol='IGLV3-13', Aliases='IGLV313, V2-9P', description='immunoglobulin lambda variable 3-13 (pseudogene)', other_designations='immunoglobulin lambda variable 3-13 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22762079, end_position_on_the_genomic_accession=22762574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28802, CurrentID=0, Status='live', Symbol='IGLV3-12', Aliases='IGLV312, V2-8', description='immunoglobulin lambda variable 3-12', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22771833, end_position_on_the_genomic_accession=22772582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28803, CurrentID=0, Status='live', Symbol='IGLV3-10', Aliases='IGLV310, V2-7', description='immunoglobulin lambda variable 3-10', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22811782, end_position_on_the_genomic_accession=22812281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28804, CurrentID=0, Status='live', Symbol='IGLV3-9', Aliases='IGLV39, V2-6', description='immunoglobulin lambda variable 3-9 (gene/pseudogene)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22819125, end_position_on_the_genomic_accession=22819752, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28805, CurrentID=0, Status='live', Symbol='IGLV3-7', Aliases='IGLV37, V2-5P', description='immunoglobulin lambda variable 3-7 (pseudogene)', other_designations='immunoglobulin lambda variable 3-7 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22838590, end_position_on_the_genomic_accession=22838888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28806, CurrentID=0, Status='live', Symbol='IGLV3-6', Aliases='IGLV36, V2-4P', description='immunoglobulin lambda variable 3-6 (pseudogene)', other_designations='immunoglobulin lambda variable 3-6 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22850373, end_position_on_the_genomic_accession=22850656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28807, CurrentID=0, Status='live', Symbol='IGLV3-4', Aliases='IGLV34, V2-3P', description='immunoglobulin lambda variable 3-4 (pseudogene)', other_designations='immunoglobulin lambda variable 3-4 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22857821, end_position_on_the_genomic_accession=22858302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28808, CurrentID=0, Status='live', Symbol='IGLV3-2', Aliases='IGLV32, V2-2P', description='immunoglobulin lambda variable 3-2 (pseudogene)', other_designations='immunoglobulin lambda variable 3-2 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22873008, end_position_on_the_genomic_accession=22873493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28809, CurrentID=0, Status='live', Symbol='IGLV3-1', Aliases='IGLV31, V2-1', description='immunoglobulin lambda variable 3-1', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22880912, end_position_on_the_genomic_accession=22881392, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28810, CurrentID=0, Status='live', Symbol='IGLV2-34', Aliases='IGLV234, V1-10P', description='immunoglobulin lambda variable 2-34 (pseudogene)', other_designations='immunoglobulin lambda variable 2-34 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22579842, end_position_on_the_genomic_accession=22580312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28811, CurrentID=0, Status='live', Symbol='IGLV2-33', Aliases='IGLV233, V1-9', description='immunoglobulin lambda variable 2-33 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22588213, end_position_on_the_genomic_accession=22588688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28812, CurrentID=0, Status='live', Symbol='IGLV2-28', Aliases='IGLV228, V1-8P', description='immunoglobulin lambda variable 2-28 (pseudogene)', other_designations='immunoglobulin lambda variable 2-28 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22664473, end_position_on_the_genomic_accession=22664931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28813, CurrentID=0, Status='live', Symbol='IGLV2-23', Aliases='IGLV223, V1-7', description='immunoglobulin lambda variable 2-23', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22697951, end_position_on_the_genomic_accession=22698421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28814, CurrentID=0, Status='live', Symbol='IGLV2-18', Aliases='IGLV218, V1-5', description='immunoglobulin lambda variable 2-18', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22734620, end_position_on_the_genomic_accession=22735089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28815, CurrentID=0, Status='live', Symbol='IGLV2-14', Aliases='IGLV214, V1-4', description='immunoglobulin lambda variable 2-14', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22758741, end_position_on_the_genomic_accession=22759211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28816, CurrentID=0, Status='live', Symbol='IGLV2-11', Aliases='IGLV211, V1-3', description='immunoglobulin lambda variable 2-11', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22792533, end_position_on_the_genomic_accession=22793003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28817, CurrentID=0, Status='live', Symbol='IGLV2-8', Aliases='IGLV28, V1-2', description='immunoglobulin lambda variable 2-8', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22822821, end_position_on_the_genomic_accession=22823289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28818, CurrentID=0, Status='live', Symbol='IGLV2-5', Aliases='IGLV25, V1-1P', description='immunoglobulin lambda variable 2-5 (pseudogene)', other_designations='immunoglobulin lambda variable 2-5 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22856590, end_position_on_the_genomic_accession=22857060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28819, CurrentID=0, Status='live', Symbol='IGLV1-62', Aliases='IGLV162, V1-23P', description='immunoglobulin lambda variable 1-62 (pseudogene)', other_designations='immunoglobulin lambda variable 1-62 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22086716, end_position_on_the_genomic_accession=22087024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28820, CurrentID=0, Status='live', Symbol='IGLV1-51', Aliases='IGLV151, V1-19', description='immunoglobulin lambda variable 1-51', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22322508, end_position_on_the_genomic_accession=22322969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28821, CurrentID=0, Status='live', Symbol='IGLV1-50', Aliases='IGLV150, V1-18', description='immunoglobulin lambda variable 1-50 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22327351, end_position_on_the_genomic_accession=22327816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28822, CurrentID=0, Status='live', Symbol='IGLV1-47', Aliases='IGLV147, V1-17', description='immunoglobulin lambda variable 1-47', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22357795, end_position_on_the_genomic_accession=22358262, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28823, CurrentID=0, Status='live', Symbol='IGLV1-44', Aliases='IGLV144, V1-16', description='immunoglobulin lambda variable 1-44', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22380880, end_position_on_the_genomic_accession=22381347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28824, CurrentID=0, Status='live', Symbol='IGLV1-41', Aliases='IGLV141, V1-14P', description='immunoglobulin lambda variable 1-41 (pseudogene)', other_designations='immunoglobulin lambda variable 1-41 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22404260, end_position_on_the_genomic_accession=22404721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28825, CurrentID=0, Status='live', Symbol='IGLV1-40', Aliases='IGLV140, V1-13', description='immunoglobulin lambda variable 1-40', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22409817, end_position_on_the_genomic_accession=22410282, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28826, CurrentID=0, Status='live', Symbol='IGLV1-36', Aliases='IGLV136, V1-11', description='immunoglobulin lambda variable 1-36', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22431998, end_position_on_the_genomic_accession=22432465, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553115, CurrentID=0, Status='live', Symbol='PEF1', Aliases='ABP32A, PEF1', description='penta-EF-hand domain containing 1', other_designations='PEF protein with a long N-terminal hydrophobic domain|penta-EF hand domain containing 1', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31629862, end_position_on_the_genomic_accession=31645237, orientation='minus', exon_count=9, OMIM=610033, no_hdr0=''),
  28828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28828, CurrentID=0, Status='live', Symbol='IGLJ6', Aliases='', description='immunoglobulin lambda joining 6', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22918164, end_position_on_the_genomic_accession=22918201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28829, CurrentID=0, Status='live', Symbol='IGLJ5', Aliases='', description='immunoglobulin lambda joining 5 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22914272, end_position_on_the_genomic_accession=22914309, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28830, CurrentID=0, Status='live', Symbol='IGLJ4', Aliases='', description='immunoglobulin lambda joining 4 (non-functional)', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22910570, end_position_on_the_genomic_accession=22910607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28831, CurrentID=0, Status='live', Symbol='IGLJ3', Aliases='J3', description='immunoglobulin lambda joining 3', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22904988, end_position_on_the_genomic_accession=22905025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28832, CurrentID=0, Status='live', Symbol='IGLJ2', Aliases='J2', description='immunoglobulin lambda joining 2', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22899618, end_position_on_the_genomic_accession=22899655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28833, CurrentID=0, Status='live', Symbol='IGLJ1', Aliases='J1', description='immunoglobulin lambda joining 1', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22893781, end_position_on_the_genomic_accession=22893818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28834, CurrentID=0, Status='live', Symbol='IGLC7', Aliases='C7', description='immunoglobulin lambda constant 7', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22922594, end_position_on_the_genomic_accession=22922913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729835, CurrentID=0, Status='live', Symbol='LOC729835', Aliases='', description='ubiquitin-fold modifier 1 pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35375236, end_position_on_the_genomic_accession=35377747, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553128, CurrentID=0, Status='live', Symbol='KIR2DL5B', Aliases='KIR2DL5, KIR2DL5.2, KIR2DLX', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 5B', other_designations='CD158 antigen-like family member F2|killer cell Ig-like receptor|killer cell immunoglobulin-like receptor 2DLX', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615727, no_hdr0=''),
  4807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4807, CurrentID=0, Status='live', Symbol='NHLH1', Aliases='HEN1, NSCL, NSCL1, bHLHa35', description='nescient helix-loop-helix 1', other_designations='HEN-1|NSCL-1|class A basic helix-loop-helix protein 35|nescient helix loop helix 1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160286369, end_position_on_the_genomic_accession=160372848, orientation='plus', exon_count=4, OMIM=162360, no_hdr0=''),
  28847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28847, CurrentID=0, Status='live', Symbol='IGKV2OR22-4', Aliases='IGKV2OR224, IGKVP6', description='immunoglobulin kappa variable 2/OR22-4 (pseudogene)', other_designations='IGKV2/OR22-4|immunoglobulin kappa variable 2/OR22-4 pseudogene|immunoglobulin kappa variable region pseudogene 6', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16914064, end_position_on_the_genomic_accession=16915177, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553137, CurrentID=0, Status='live', Symbol='LOC553137', Aliases='', description='uncharacterized LOC553137', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=107218001, end_position_on_the_genomic_accession=107222877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  28850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28850, CurrentID=0, Status='live', Symbol='IGKV1OR22-5', Aliases='IGKV1OR22-5A, IGKV1OR225, IGKV1OR225A, IGKVP7', description='immunoglobulin kappa variable 1/OR22-5 (pseudogene)', other_designations='IGKV1/OR22-5|IGKV1/OR22-5A|immunoglobulin kappa variable 1/OR22-5 pseudogene|immunoglobulin kappa variable 1/OR22-5a|immunoglobulin kappa variable region pseudogene 7', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553139, CurrentID=0, Status='live', Symbol='LOC553139', Aliases='', description='PAI-1 mRNA binding protein pseudogene', other_designations='SERPINE1 mRNA binding protein 1 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=80090177, end_position_on_the_genomic_accession=80092571, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553140, CurrentID=0, Status='live', Symbol='SLEP3', Aliases='', description='Serum leptin concentration QTL 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553141, CurrentID=0, Status='live', Symbol='BW88', Aliases='', description='Body weight QTL 88', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553142, CurrentID=0, Status='live', Symbol='AASTH29', Aliases='', description='Allergic/Atopic asthma related QTL 29', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28855, CurrentID=0, Status='live', Symbol='IGKV2OR2-10', Aliases='IGKV2OR210', description='immunoglobulin kappa variable 2/OR2-10 (pseudogene)', other_designations='IGKV2/OR2-10|immunoglobulin kappa variable 2/OR2-10 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97376255, end_position_on_the_genomic_accession=97376973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553144, CurrentID=0, Status='live', Symbol='AASTH56', Aliases='', description='Allergic/Atopic asthma related QTL 56', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553145, CurrentID=0, Status='live', Symbol='AASTH32', Aliases='', description='Allergic/Atopic asthma related QTL 32', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553146, CurrentID=0, Status='live', Symbol='AASTH41', Aliases='', description='Allergic/Atopic asthma related QTL 41', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553147, CurrentID=0, Status='live', Symbol='AASTH28', Aliases='', description='Allergic/Atopic asthma related QTL 28', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271450, CurrentID=0, Status='live', Symbol='RPL21P124', Aliases='RPL21_60_1556', description='ribosomal protein L21 pseudogene 124', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49400396, end_position_on_the_genomic_accession=49400800, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28861, CurrentID=0, Status='live', Symbol='IGKV2OR2-1', Aliases='IGKV2/OR2-1A, IGKV2OR21, IGKV2OR21A, IGKV2OR2-1', description='immunoglobulin kappa variable 2/OR2-1 (pseudogene)', other_designations='IGKV2/OR2-1A|immunoglobulin kappa variable 2/OR2-1 pseudogene|immunoglobulin kappa variable 2/OR2-1a', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28862, CurrentID=0, Status='live', Symbol='IGKV1OR2-108', Aliases='IGKV1OR2108, IGO1', description='immunoglobulin kappa variable 1/OR2-108 (non-functional)', other_designations='IGKV1/OR2-108|immunoglobulin orphon (transposed element) 1', map_location='2q12-q14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=147185, no_hdr0=''),
  28863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28863, CurrentID=0, Status='live', Symbol='IGKV1OR2-11', Aliases='IGKV1OR211', description='immunoglobulin kappa variable 1/OR2-11 (pseudogene)', other_designations='IGKV1/OR2-11|immunoglobulin kappa variable 1/OR2-11 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28864, CurrentID=0, Status='live', Symbol='IGKV1OR2-9', Aliases='IGKV1/OR2-9, IGKV1OR29', description='immunoglobulin kappa variable 1/OR2-9 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR2-9 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28865, CurrentID=0, Status='live', Symbol='IGKV1OR2-6', Aliases='IGKV1/OR2-6, IGKV1OR26', description='immunoglobulin kappa variable 1/OR2-6 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR2-6 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28866, CurrentID=0, Status='live', Symbol='IGKV1OR2-3', Aliases='IGKV1/OR2-3, IGKV1OR23', description='immunoglobulin kappa variable 1/OR2-3 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR2-3 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28867, CurrentID=0, Status='live', Symbol='IGKV1OR2-0', Aliases='IGKV1/OR2-0, IGKV1OR20', description='immunoglobulin kappa variable 1/OR2-0 (non-functional)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28869, CurrentID=0, Status='live', Symbol='IGKV6D-41', Aliases='A14', description='immunoglobulin kappa variable 6D-41 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90069688, end_position_on_the_genomic_accession=90070240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553158, CurrentID=0, Status='live', Symbol='PRR5-ARHGAP8', Aliases='', description='PRR5-ARHGAP8 readthrough', other_designations='PRR5-ARHGAP8 fusion', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44702198, end_position_on_the_genomic_accession=44862784, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  28871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28871, CurrentID=0, Status='live', Symbol='IGKV3D-34', Aliases='IGKV3D34, O7', description='immunoglobulin kappa variable 3D-34 (pseudogene)', other_designations='immunoglobulin kappa variable 1D-34|immunoglobulin kappa variable 3D-34 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89906759, end_position_on_the_genomic_accession=89907291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28872, CurrentID=0, Status='live', Symbol='IGKV3D-31', Aliases='IGKV3D31, O10', description='immunoglobulin kappa variable 3D-31 (pseudogene)', other_designations='immunoglobulin kappa variable 3D-31 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89929702, end_position_on_the_genomic_accession=89930239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28873, CurrentID=0, Status='live', Symbol='IGKV3D-25', Aliases='A6, IGKV3D25', description='immunoglobulin kappa variable 3D-25 (pseudogene)', other_designations='immunoglobulin kappa variable 3D-25 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89988500, end_position_on_the_genomic_accession=89990258, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28874, CurrentID=0, Status='live', Symbol='IGKV3D-20', Aliases='A11, A11a, IGKV3D20', description='immunoglobulin kappa variable 3D-20', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90038945, end_position_on_the_genomic_accession=90039479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28875, CurrentID=0, Status='live', Symbol='IGKV3D-15', Aliases='IGKV3D15, L16, L16a, L16b, L16c', description='immunoglobulin kappa variable 3D-15 (gene/pseudogene)', other_designations='immunoglobulin light chain variable region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90114887, end_position_on_the_genomic_accession=90115400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28876, CurrentID=0, Status='live', Symbol='IGKV3D-11', Aliases='IGKV3D11, L20', description='immunoglobulin kappa variable 3D-11', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90172899, end_position_on_the_genomic_accession=90173412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28877, CurrentID=0, Status='live', Symbol='IGKV3D-7', Aliases='IGKV3D7, L25', description='immunoglobulin kappa variable 3D-7', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90234837, end_position_on_the_genomic_accession=90235368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28878, CurrentID=0, Status='live', Symbol='IGKV2D-40', Aliases='IGKV2D40, O1', description='immunoglobulin kappa variable 2D-40', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89851758, end_position_on_the_genomic_accession=89852491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28879, CurrentID=0, Status='live', Symbol='IGKV2D-38', Aliases='IGKV2D38, O3', description='immunoglobulin kappa variable 2D-38 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-38 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89872463, end_position_on_the_genomic_accession=89872702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28880, CurrentID=0, Status='live', Symbol='IGKV2D-36', Aliases='IGKV2D36, O5', description='immunoglobulin kappa variable 2D-36 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-36 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89887021, end_position_on_the_genomic_accession=89887254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28881, CurrentID=0, Status='live', Symbol='IGKV2D-30', Aliases='A1, IGKV2D30', description='immunoglobulin kappa variable 2D-30', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89936894, end_position_on_the_genomic_accession=89937679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28882, CurrentID=0, Status='live', Symbol='IGKV2D-29', Aliases='A2a, A2c, IGKV2D29', description='immunoglobulin kappa variable 2D-29', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89947542, end_position_on_the_genomic_accession=89948271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28883, CurrentID=0, Status='live', Symbol='IGKV2D-28', Aliases='A3, IGKV2D28', description='immunoglobulin kappa variable 2D-28', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89960015, end_position_on_the_genomic_accession=89960747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28884, CurrentID=0, Status='live', Symbol='IGKV2D-26', Aliases='A5, IGKV2D26', description='immunoglobulin kappa variable 2D-26', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89985952, end_position_on_the_genomic_accession=89986704, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  135886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135886, CurrentID=0, Status='live', Symbol='WBSCR28', Aliases='', description='Williams-Beuren syndrome chromosome region 28', other_designations='Williams-Beuren syndrome critical region 28', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73861159, end_position_on_the_genomic_accession=73865894, orientation='plus', exon_count=4, OMIM=612547, no_hdr0=''),
  28886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28886, CurrentID=0, Status='live', Symbol='IGKV2D-23', Aliases='A8, IGKV2D23', description='immunoglobulin kappa variable 2D-23 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-23 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90008911, end_position_on_the_genomic_accession=90009916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28887, CurrentID=0, Status='live', Symbol='IGKV2D-19', Aliases='A12, IGKV2D19', description='immunoglobulin kappa variable 2D-19 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-19 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90046806, end_position_on_the_genomic_accession=90047072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28888, CurrentID=0, Status='live', Symbol='IGKV2D-18', Aliases='A13, IGKV2D18', description='immunoglobulin kappa variable 2D-18 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-18 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90052582, end_position_on_the_genomic_accession=90053358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28889, CurrentID=0, Status='live', Symbol='IGKV2D-14', Aliases='IGKV2D14, L17', description='immunoglobulin kappa variable 2D-14 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-14 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90121787, end_position_on_the_genomic_accession=90122552, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28890, CurrentID=0, Status='live', Symbol='IGKV2D-10', Aliases='IGKV2D10, L21', description='immunoglobulin kappa variable 2D-10 (pseudogene)', other_designations='immunoglobulin kappa variable 2D-10 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90179890, end_position_on_the_genomic_accession=90180655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  159963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159963, CurrentID=0, Status='live', Symbol='SLC5A12', Aliases='SMCT2', description='solute carrier family 5 (sodium/monocarboxylate cotransporter), member 12', other_designations='electroneutral sodium monocarboxylate cotransporter|low-affinity sodium-lactate cotransporter|sodium-iodide related cotransporter|solute carrier family 5 (sodium/glucose cotransporter), member 12|solute carrier family 5 member 12', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26667019, end_position_on_the_genomic_accession=26723356, orientation='minus', exon_count=19, OMIM=612455, no_hdr0=''),
  28892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28892, CurrentID=0, Status='live', Symbol='IGKV1D-42', Aliases='IGKV1D42, L22', description='immunoglobulin kappa variable 1D-42 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90190213, end_position_on_the_genomic_accession=90190679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28893, CurrentID=0, Status='live', Symbol='IGKV1D-39', Aliases='IGKV1D39, O2', description='immunoglobulin kappa variable 1D-39', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89862505, end_position_on_the_genomic_accession=89862979, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28894, CurrentID=0, Status='live', Symbol='IGKV1D-37', Aliases='IGKV1D37, O4', description='immunoglobulin kappa variable 1D-37 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89884740, end_position_on_the_genomic_accession=89885214, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28895, CurrentID=0, Status='live', Symbol='IGKV1D-35', Aliases='IGKV1D35, O6', description='immunoglobulin kappa variable 1D-35 (pseudogene)', other_designations='immunoglobulin kappa variable 1D-35 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89895314, end_position_on_the_genomic_accession=89895788, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28896, CurrentID=0, Status='live', Symbol='IGKV1D-33', Aliases='IGKV1D33, O8', description='immunoglobulin kappa variable 1D-33', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89914071, end_position_on_the_genomic_accession=89914545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28897, CurrentID=0, Status='live', Symbol='IGKV1D-32', Aliases='IGKV1D32, O9', description='immunoglobulin kappa variable 1D-32 (pseudogene)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89928422, end_position_on_the_genomic_accession=89928896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28898, CurrentID=0, Status='live', Symbol='IGKV1D-27', Aliases='A4, A4a, IGKV1D27', description='immunoglobulin kappa variable 1D-27 (pseudogene)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89968869, end_position_on_the_genomic_accession=89969344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28899, CurrentID=0, Status='live', Symbol='IGKV1D-22', Aliases='A9, IGKV1D22', description='immunoglobulin kappa variable 1D-22 (pseudogene)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90010741, end_position_on_the_genomic_accession=90011208, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28900, CurrentID=0, Status='live', Symbol='IGKV1D-17', Aliases='IGKV1D17, L14', description='immunoglobulin kappa variable 1D-17', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90082816, end_position_on_the_genomic_accession=90083291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28901, CurrentID=0, Status='live', Symbol='IGKV1D-16', Aliases='IGKV1D16, L15, L15a', description='immunoglobulin kappa variable 1D-16', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90100263, end_position_on_the_genomic_accession=90100738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553190, CurrentID=0, Status='live', Symbol='MYP7', Aliases='', description='myopia 7', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609256, no_hdr0=''),
  28903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28903, CurrentID=0, Status='live', Symbol='IGKV1D-12', Aliases='IGKV1D12, L19', description='immunoglobulin kappa variable 1D-12', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90159860, end_position_on_the_genomic_accession=90160335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553192, CurrentID=0, Status='live', Symbol='MYP8', Aliases='', description='myopia 8', other_designations='', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609257, no_hdr0=''),
  28905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28905, CurrentID=0, Status='live', Symbol='IGKV7-3', Aliases='B1, IGKV73', description='immunoglobulin kappa variable 7-3 (pseudogene)', other_designations='immunoglobulin kappa variable 7-3 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88915080, end_position_on_the_genomic_accession=88915678, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553194, CurrentID=0, Status='live', Symbol='MYP9', Aliases='', description='myopia 9', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609258, no_hdr0=''),
  553195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553195, CurrentID=0, Status='live', Symbol='MYP10', Aliases='', description='myopia 10', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609259, no_hdr0=''),
  28908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28908, CurrentID=0, Status='live', Symbol='IGKV4-1', Aliases='B3, IGKV41', description='immunoglobulin kappa variable 4-1', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88885572, end_position_on_the_genomic_accession=88886153, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28909, CurrentID=0, Status='live', Symbol='IGKV3-34', Aliases='IGKV334, O17', description='immunoglobulin kappa variable 3-34 (pseudogene)', other_designations='immunoglobulin kappa variable 3-34 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89275255, end_position_on_the_genomic_accession=89275787, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28910, CurrentID=0, Status='live', Symbol='IGKV3-31', Aliases='A16, A16a, IGKV331', description='immunoglobulin kappa variable 3-31 (pseudogene)', other_designations='immunoglobulin kappa variable 3-31 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89252198, end_position_on_the_genomic_accession=89252735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28911, CurrentID=0, Status='live', Symbol='IGKV3-25', Aliases='A22, IGKV325', description='immunoglobulin kappa variable 3-25 (pseudogene)', other_designations='immunoglobulin kappa variable 3-25 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89192491, end_position_on_the_genomic_accession=89194268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28912, CurrentID=0, Status='live', Symbol='IGKV3-20', Aliases='13K18, A27, IGKV320', description='immunoglobulin kappa variable 3-20', other_designations='Ig kappa chain V-III region HIC|anti-(ED-B) scFV|anti-FactorVIII scFv|anti-HCS scFv|anti-IFN-G scFv|anti-TeTox scFv|immunoglobulin light chain variable region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89142574, end_position_on_the_genomic_accession=89143108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28913, CurrentID=0, Status='live', Symbol='IGKV3-15', Aliases='IGKV315, L2', description='immunoglobulin kappa variable 3-15', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89085177, end_position_on_the_genomic_accession=89085690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28914, CurrentID=0, Status='live', Symbol='IGKV3-11', Aliases='IGKV311, L6', description='immunoglobulin kappa variable 3-11', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89027171, end_position_on_the_genomic_accession=89027684, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28915, CurrentID=0, Status='live', Symbol='IGKV3-7', Aliases='IGKV37, L10, L10a, Vh', description='immunoglobulin kappa variable 3-7 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88978468, end_position_on_the_genomic_accession=88978984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28916, CurrentID=0, Status='live', Symbol='IGKV2-40', Aliases='IGKV240, O11, O11a', description='immunoglobulin kappa variable 2-40', other_designations='Ig kappa chain V-II', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89330116, end_position_on_the_genomic_accession=89330430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  159989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=159989, CurrentID=0, Status='live', Symbol='CCDC67', Aliases='', description='coiled-coil domain containing 67', other_designations='coiled-coil domain-containing protein 67', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93329952, end_position_on_the_genomic_accession=93438470, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  28918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28918, CurrentID=0, Status='live', Symbol='IGKV2-36', Aliases='IGKV236, O15', description='immunoglobulin kappa variable 2-36 (pseudogene)', other_designations='immunoglobulin kappa variable 2-36 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89295224, end_position_on_the_genomic_accession=89295457, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28919, CurrentID=0, Status='live', Symbol='IGKV2-30', Aliases='A17, IGKV230', description='immunoglobulin kappa variable 2-30', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89244781, end_position_on_the_genomic_accession=89245566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28920, CurrentID=0, Status='live', Symbol='IGKV2-29', Aliases='A18, A18b, IGKV, IGKV229, IGVK-A18', description='immunoglobulin kappa variable 2-29 (gene/pseudogene)', other_designations='IGVK A18b', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89234185, end_position_on_the_genomic_accession=89234911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135892, CurrentID=0, Status='live', Symbol='TRIM50', Aliases='TRIM50A', description='tripartite motif containing 50', other_designations='tripartite motif-containing 50A|tripartite motif-containing protein 50', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73312472, end_position_on_the_genomic_accession=73328279, orientation='minus', exon_count=9, OMIM=612548, no_hdr0=''),
  28922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28922, CurrentID=0, Status='live', Symbol='IGKV2-26', Aliases='A21, IGKV226', description='immunoglobulin kappa variable 2-26 (pseudogene)', other_designations='immunoglobulin kappa variable 2-26 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89196076, end_position_on_the_genomic_accession=89196829, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28923, CurrentID=0, Status='live', Symbol='IGKV2-24', Aliases='A23, IGKV224', description='immunoglobulin kappa variable 2-24', other_designations='Ig kappa chain V-II region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89176328, end_position_on_the_genomic_accession=89177130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28924, CurrentID=0, Status='live', Symbol='IGKV2-23', Aliases='A24, IGKV223', description='immunoglobulin kappa variable 2-23 (pseudogene)', other_designations='immunoglobulin kappa variable 2-23 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89172328, end_position_on_the_genomic_accession=89173043, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28925, CurrentID=0, Status='live', Symbol='IGKV2-19', Aliases='A28, IGKV219', description='immunoglobulin kappa variable 2-19 (pseudogene)', other_designations='immunoglobulin kappa variable 2-19 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89134981, end_position_on_the_genomic_accession=89135247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28926, CurrentID=0, Status='live', Symbol='IGKV2-18', Aliases='A29, IGKV218', description='immunoglobulin kappa variable 2-18 (pseudogene)', other_designations='immunoglobulin kappa variable 2-18 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89128706, end_position_on_the_genomic_accession=89129482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28927, CurrentID=0, Status='live', Symbol='IGKV2-14', Aliases='IGKV214, L3', description='immunoglobulin kappa variable 2-14 (pseudogene)', other_designations='immunoglobulin kappa variable 2-14 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89078019, end_position_on_the_genomic_accession=89078784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28928, CurrentID=0, Status='live', Symbol='IGKV2-10', Aliases='IGKV210, L7', description='immunoglobulin kappa variable 2-10 (pseudogene)', other_designations='immunoglobulin kappa variable 2-10 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89019992, end_position_on_the_genomic_accession=89020685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28929, CurrentID=0, Status='live', Symbol='IGKV2-4', Aliases='IGKV24, L13', description='immunoglobulin kappa variable 2-4 (pseudogene)', other_designations='immunoglobulin kappa variable 2-4 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88931651, end_position_on_the_genomic_accession=88932378, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28930, CurrentID=0, Status='live', Symbol='IGKV1-39', Aliases='IGKV139, O12, O12a', description='immunoglobulin kappa variable 1-39 (gene/pseudogene)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89319625, end_position_on_the_genomic_accession=89320099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28931, CurrentID=0, Status='live', Symbol='IGKV1-37', Aliases='IGKV137, O14', description='immunoglobulin kappa variable 1-37 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89297264, end_position_on_the_genomic_accession=89297738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28932, CurrentID=0, Status='live', Symbol='IGKV1-35', Aliases='IGKV135, O16', description='immunoglobulin kappa variable 1-35 (pseudogene)', other_designations='immunoglobulin kappa variable 1-35 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89286696, end_position_on_the_genomic_accession=89287170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28933, CurrentID=0, Status='live', Symbol='IGKV1-33', Aliases='IGKV133, O18', description='immunoglobulin kappa variable 1-33', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89268001, end_position_on_the_genomic_accession=89268475, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28934, CurrentID=0, Status='live', Symbol='IGKV1-32', Aliases='A15, A15a, IGKV132', description='immunoglobulin kappa variable 1-32 (pseudogene)', other_designations='immunoglobulin kappa variable 1-32 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89253541, end_position_on_the_genomic_accession=89254015, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28935, CurrentID=0, Status='live', Symbol='IGKV1-27', Aliases='A20, IGKV127', description='immunoglobulin kappa variable 1-27', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89213423, end_position_on_the_genomic_accession=89213897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28936, CurrentID=0, Status='live', Symbol='IGKV1-22', Aliases='A25, IGKV122', description='immunoglobulin kappa variable 1-22 (pseudogene)', other_designations='immunoglobulin kappa variable 1-22 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89170745, end_position_on_the_genomic_accession=89171212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28937, CurrentID=0, Status='live', Symbol='IGKV1-17', Aliases='A30, IGKV117', description='immunoglobulin kappa variable 1-17', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89117342, end_position_on_the_genomic_accession=89117817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28938, CurrentID=0, Status='live', Symbol='IGKV1-16', Aliases='IGKV116, L1', description='immunoglobulin kappa variable 1-16', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89099859, end_position_on_the_genomic_accession=89100334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28939, CurrentID=0, Status='live', Symbol='IGKV1-13', Aliases='IGKV113, L18', description='immunoglobulin kappa variable 1-13 (gene/pseudogene)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89045990, end_position_on_the_genomic_accession=89046466, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28940, CurrentID=0, Status='live', Symbol='IGKV1-12', Aliases='IGKV112, L19', description='immunoglobulin kappa variable 1-12', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89040224, end_position_on_the_genomic_accession=89040698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28941, CurrentID=0, Status='live', Symbol='IGKV1-9', Aliases='IGKV19, L8', description='immunoglobulin kappa variable 1-9', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89009982, end_position_on_the_genomic_accession=89010457, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28942, CurrentID=0, Status='live', Symbol='IGKV1-8', Aliases='IGKV18, L9', description='immunoglobulin kappa variable 1-8', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88992409, end_position_on_the_genomic_accession=88992878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28943, CurrentID=0, Status='live', Symbol='IGKV1-6', Aliases='IGKV16, L11', description='immunoglobulin kappa variable 1-6', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88966262, end_position_on_the_genomic_accession=88966738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79525, CurrentID=0, Status='live', Symbol='OR51A5P', Aliases='', description='olfactory receptor family 51 subfamily A member 5 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 5 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4972931, end_position_on_the_genomic_accession=4973896, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28946, CurrentID=0, Status='live', Symbol='IGKJ5', Aliases='J5', description='immunoglobulin kappa joining 5', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88860568, end_position_on_the_genomic_accession=88860605, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28947, CurrentID=0, Status='live', Symbol='IGKJ4', Aliases='J4', description='immunoglobulin kappa joining 4', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88860886, end_position_on_the_genomic_accession=88860923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28948, CurrentID=0, Status='live', Symbol='IGKJ3', Aliases='J3', description='immunoglobulin kappa joining 3', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88861221, end_position_on_the_genomic_accession=88861258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28949, CurrentID=0, Status='live', Symbol='IGKJ2', Aliases='J2', description='immunoglobulin kappa joining 2', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88861525, end_position_on_the_genomic_accession=88861563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28950, CurrentID=0, Status='live', Symbol='IGKJ1', Aliases='J1', description='immunoglobulin kappa joining 1', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88861886, end_position_on_the_genomic_accession=88861923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28951, CurrentID=0, Status='live', Symbol='TRIB2', Aliases='C5FW, GS3955, TRB2', description='tribbles pseudokinase 2', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12716872, end_position_on_the_genomic_accession=12742734, orientation='plus', exon_count=3, OMIM=609462, no_hdr0=''),
  28952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28952, CurrentID=0, Status='live', Symbol='CCDC22', Aliases='CXorf37, JM1', description='coiled-coil domain containing 22', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49235467, end_position_on_the_genomic_accession=49250526, orientation='plus', exon_count=17, OMIM=300859, no_hdr0=''),
  28954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28954, CurrentID=0, Status='live', Symbol='REM1', Aliases='GD:REM, GES', description='RAS (RAD and GEM)-like GTP-binding 1', other_designations='GTPase-regulating endothelial cell sprouting', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31475288, end_position_on_the_genomic_accession=31484905, orientation='plus', exon_count=5, OMIM=610388, no_hdr0=''),
  28955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28955, CurrentID=0, Status='live', Symbol='DEXI', Aliases='MYLE', description='Dexi homolog (mouse)', other_designations='dexamethasone-induced transcript', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10928891, end_position_on_the_genomic_accession=10942400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28956, CurrentID=0, Status='live', Symbol='LAMTOR2', Aliases='ENDAP, HSPC003, MAPBPIP, MAPKSP1AP, ROBLD3, Ragulator2, p14', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 2', other_designations='MAPBP-interacting protein|MAPKSP1 adaptor protein|endosomal adaptor protein p14|late endosomal/lysosomal Mp1-interacting protein|late endosomal/lysosomal adaptor and MAPK and MTOR activator 2|mitogen-activated protein-binding protein-interacting protein (MAPBPIP)|roadblock domain containing 3|roadblock domain-containing protein 3', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156054726, end_position_on_the_genomic_accession=156058510, orientation='plus', exon_count=4, OMIM=610389, no_hdr0=''),
  28957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28957, CurrentID=0, Status='live', Symbol='MRPS28', Aliases='HSPC007, MRP-S28, MRP-S35, MRPS35', description='mitochondrial ribosomal protein S28', other_designations='28S ribosomal protein S35, mitochondrial|S28mt|S35mt|mitochondrial 28S ribosomal protein S35', map_location='8q21.1-q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79918860, end_position_on_the_genomic_accession=80030271, orientation='minus', exon_count=3, OMIM=611990, no_hdr0=''),
  28958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28958, CurrentID=0, Status='live', Symbol='COA3', Aliases='CCDC56, HSPC009, MITRAC12', description='cytochrome c oxidase assembly factor 3', other_designations='coiled-coil domain-containing protein 56|cytochrome C oxidase assembly factor 3 homolog, mitochondrial|cytochrome c oxidase assembly protein 3 homolog, mitochondrial|mitochondrial translation regulation assembly intermediate of cytochrome c oxidase protein of 12 kDa', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42797618, end_position_on_the_genomic_accession=42798725, orientation='minus', exon_count=2, OMIM=614775, no_hdr0=''),
  28959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28959, CurrentID=0, Status='live', Symbol='TMEM176B', Aliases='LR8, MS4B2', description='transmembrane protein 176B', other_designations='LR8-like protein', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150791287, end_position_on_the_genomic_accession=150801360, orientation='minus', exon_count=9, OMIM=610385, no_hdr0=''),
  28960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28960, CurrentID=0, Status='live', Symbol='DCPS', Aliases='ARS, DCS1, HINT-5, HINT5, HSL1, HSPC015', description='decapping enzyme, scavenger', other_designations='decapping scavenger enzyme|heat shock-like protein 1|hint-related 7meGMP-directed hydrolase|histidine triad nucleotide-binding protein 5|histidine triad protein member 5|homolog of C. elegans 7meGMP-directed hydrolase dcs-1|mRNA decapping enzyme|scavenger mRNA-decapping enzyme DcpS', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126303752, end_position_on_the_genomic_accession=126345754, orientation='plus', exon_count=8, OMIM=610534, no_hdr0=''),
  28962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28962, CurrentID=0, Status='live', Symbol='OSTM1', Aliases='GIPN, GL, HSPC019, OPTB5', description='osteopetrosis associated transmembrane protein 1', other_designations='CLCN7 accessory beta subunit|GAIP-interacting protein N terminus|chloride channel 7 beta subunit|grey-lethal osteopetrosis', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108041409, end_position_on_the_genomic_accession=108074737, orientation='minus', exon_count=8, OMIM=607649, no_hdr0=''),
  57703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57703, CurrentID=0, Status='live', Symbol='CWC22', Aliases='EIF4GL, NCM, fSAPb', description='CWC22 homolog, spliceosome-associated protein', other_designations='CWC22 spliceosome-associated protein homolog|functional spliceosome-associated protein b|nucampholin homolog', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179944876, end_position_on_the_genomic_accession=180007297, orientation='minus', exon_count=21, OMIM=615186, no_hdr0=''),
  28964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28964, CurrentID=0, Status='live', Symbol='GIT1', Aliases='', description='G protein-coupled receptor kinase interacting ArfGAP 1', other_designations='ARF GAP GIT1|CAT-1|CAT1|G protein-coupled receptor kinase interactor 1|G protein-coupled receptor kinase-interactor 1|GRK-interacting protein 1|cool-associated and tyrosine-phosphorylated protein 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29573469, end_position_on_the_genomic_accession=29589592, orientation='minus', exon_count=21, OMIM=608434, no_hdr0=''),
  28965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28965, CurrentID=0, Status='live', Symbol='SLC27A6', Aliases='ACSVL2, FACVL2, FATP6, VLCS-H1', description='solute carrier family 27 (fatty acid transporter), member 6', other_designations='acyl-CoA synthetase very long chain family, member 2|fatty-acid-Coenzyme A ligase, very long-chain 2|very long-chain acyl-CoA synthetase homolog 1', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128965475, end_position_on_the_genomic_accession=129033642, orientation='plus', exon_count=12, OMIM=604196, no_hdr0=''),
  28966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28966, CurrentID=0, Status='live', Symbol='SNX24', Aliases='PRO1284, SBBI31', description='sorting nexin 24', other_designations='sorting nexing 24', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122845465, end_position_on_the_genomic_accession=123029276, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  28968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28968, CurrentID=0, Status='live', Symbol='SLC6A16', Aliases='NT5, NTT5', description='solute carrier family 6 member 16', other_designations='solute carrier family 6 (neurotransmitter transporter), member 16|solute carrier family 6, member 16', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49289632, end_position_on_the_genomic_accession=49331701, orientation='minus', exon_count=15, OMIM=607972, no_hdr0=''),
  28969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28969, CurrentID=0, Status='live', Symbol='BZW2', Aliases='HSPC028, MST017, MSTP017', description='basic leucine zipper and W2 domains 2', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16646134, end_position_on_the_genomic_accession=16706523, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  28970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28970, CurrentID=0, Status='live', Symbol='C11orf54', Aliases='PTD012, PTOD012', description='chromosome 11 open reading frame 54', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93741594, end_position_on_the_genomic_accession=93763102, orientation='plus', exon_count=12, OMIM=615810, no_hdr0=''),
  28971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28971, CurrentID=0, Status='live', Symbol='AAMDC', Aliases='C11orf67, CK067, PTD015', description='adipogenesis associated, Mth938 domain containing', other_designations='UPF0366 protein C11orf67|adipogenesis associated Mth938 domain-containing protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77821107, end_position_on_the_genomic_accession=77872354, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  28972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28972, CurrentID=0, Status='live', Symbol='SPCS1', Aliases='HSPC033, SPC1, SPC12, YJR010C-A', description='signal peptidase complex subunit 1', other_designations='SPase 12 kDa subunit|microsomal signal peptidase 12 kDa subunit|signal peptidase 12kDa|signal peptidase complex subunit 1 homolog', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52705841, end_position_on_the_genomic_accession=52708182, orientation='plus', exon_count=4, OMIM=610358, no_hdr0=''),
  28973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28973, CurrentID=0, Status='live', Symbol='MRPS18B', Aliases='C6orf14, HSPC183, HumanS18a, MRP-S18-2, MRPS18-2, PTD017, S18amt', description='mitochondrial ribosomal protein S18B', other_designations='28S ribosomal protein S18-2, mitochondrial|MRP-S18-b|S18mt-b|mitochondrial ribosomal protein S18-2|mrps18-b', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30617320, end_position_on_the_genomic_accession=30626397, orientation='plus', exon_count=7, OMIM=611982, no_hdr0=''),
  28974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28974, CurrentID=0, Status='live', Symbol='C19orf53', Aliases='HSPC023, LYDG10', description='chromosome 19 open reading frame 53', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13774443, end_position_on_the_genomic_accession=13778773, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79526, CurrentID=0, Status='live', Symbol='OR4C10P', Aliases='', description='olfactory receptor family 4 subfamily C member 10 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 10 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48432217, end_position_on_the_genomic_accession=48433147, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28976, CurrentID=0, Status='live', Symbol='ACAD9', Aliases='NPD002', description='acyl-CoA dehydrogenase family member 9', other_designations='acyl-CoA dehydrogenase family, member 9|acyl-Coenzyme A dehydrogenase family, member 9|very-long-chain acyl-CoA dehydrogenase VLCAD', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128879490, end_position_on_the_genomic_accession=128913114, orientation='plus', exon_count=18, OMIM=611103, no_hdr0=''),
  28977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28977, CurrentID=0, Status='live', Symbol='MRPL42', Aliases='HSPC204, L31MT, L42MT, MRP-L31, MRP-L42, MRP-S32, MRPL31, MRPS32, PTD007, RPML31, S32MT', description='mitochondrial ribosomal protein L42', other_designations='28S ribosomal protein S32, mitochondrial|39S ribosomal protein L31, mitochondrial', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93467490, end_position_on_the_genomic_accession=93503772, orientation='plus', exon_count=7, OMIM=611847, no_hdr0=''),
  28978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28978, CurrentID=0, Status='live', Symbol='TMEM14A', Aliases='C6orf73, PTD011', description='transmembrane protein 14A', other_designations='', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52671086, end_position_on_the_genomic_accession=52686588, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  28981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28981, CurrentID=0, Status='live', Symbol='IFT81', Aliases='CDV-1, CDV-1R, CDV1, CDV1R, DV1', description='intraflagellar transport 81', other_designations='carnitine deficiency-associated gene expressed in ventricle 1|carnitine deficiency-associated protein expressed in ventricle 1|intraflagellar transport 81 homolog', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110124335, end_position_on_the_genomic_accession=110218795, orientation='plus', exon_count=21, OMIM=605489, no_hdr0=''),
  28982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28982, CurrentID=0, Status='live', Symbol='FLVCR1', Aliases='AXPC1, FLVCR, MFSD7B, PCA, PCARP', description='feline leukemia virus subgroup C cellular receptor 1', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212858255, end_position_on_the_genomic_accession=212899363, orientation='plus', exon_count=12, OMIM=609144, no_hdr0=''),
  28983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28983, CurrentID=0, Status='live', Symbol='TMPRSS11E', Aliases='DESC12, TMPRSS11E', description='transmembrane protease, serine 11E', other_designations='serine protease DESC1|transmembrane protease serine 11E2|transmembrane protease, serine 11E2', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68447449, end_position_on_the_genomic_accession=68497604, orientation='plus', exon_count=10, OMIM=610399, no_hdr0=''),
  28984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28984, CurrentID=0, Status='live', Symbol='RGCC', Aliases='C13orf15, RGC-32, RGC32, bA157L14.2', description='regulator of cell cycle', other_designations='response gene to complement 32 protein', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41457406, end_position_on_the_genomic_accession=41470877, orientation='plus', exon_count=7, OMIM=610077, no_hdr0=''),
  28985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28985, CurrentID=0, Status='live', Symbol='MCTS1', Aliases='MCT-1, MCT1', description='malignant T-cell amplified sequence 1', other_designations='malignant T cell amplified sequence 1|multiple copies T-cell malignancies', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120603889, end_position_on_the_genomic_accession=120621159, orientation='plus', exon_count=7, OMIM=300587, no_hdr0=''),
  28986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28986, CurrentID=0, Status='live', Symbol='MAGEH1', Aliases='APR-1, APR1, MAGEH', description='MAGE family member H1', other_designations='apoptosis-related protein 1|melanoma antigen family H, 1|melanoma antigen family H1|restin', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55452089, end_position_on_the_genomic_accession=55453568, orientation='plus', exon_count=1, OMIM=300548, no_hdr0=''),
  28987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28987, CurrentID=0, Status='live', Symbol='NOB1', Aliases='ART-4, MST158, MSTP158P, PSMD8BP1, NOB1', description='NIN1/RPN12 binding protein 1 homolog', other_designations='PSMD8 binding protein 1|adenocarcinoma antigen recognized by T lymphocytes 4|nin one binding protein|phosphorylation regulatory protein HP-10|protein ART-4', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69741854, end_position_on_the_genomic_accession=69754968, orientation='minus', exon_count=9, OMIM=613586, no_hdr0=''),
  28988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28988, CurrentID=0, Status='live', Symbol='DBNL', Aliases='ABP1, HIP-55, HIP55, SH3P7', description='drebrin-like', other_designations='HPK1-interacting protein of 55 kDa|SH3 domain-containing protein 7|cervical SH3P7|cervical mucin-associated protein|drebrin-F|src homology 3 domain-containing protein HIP-55', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44044640, end_position_on_the_genomic_accession=44061716, orientation='plus', exon_count=13, OMIM=610106, no_hdr0=''),
  28989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28989, CurrentID=0, Status='live', Symbol='NTMT1', Aliases='AD-003, C9orf32, HOMT1A, METTL11A, NRMT, NTM1A', description='N-terminal Xaa-Pro-Lys N-methyltransferase 1', other_designations='N-terminal RCC1 methyltransferase|X-Pro-Lys N-terminal protein methyltransferase 1A|alpha N-terminal protein methyltransferase 1A|methyltransferase-like protein 11A', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129608884, end_position_on_the_genomic_accession=129636131, orientation='plus', exon_count=7, OMIM=613560, no_hdr0=''),
  28990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28990, CurrentID=0, Status='live', Symbol='ASTE1', Aliases='HT001', description='asteroid homolog 1 (Drosophila)', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131013873, end_position_on_the_genomic_accession=131026854, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  28991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28991, CurrentID=0, Status='live', Symbol='COMMD5', Aliases='HCARG, HT002', description='COMM domain containing 5', other_designations='hypertension-related calcium-regulated gene protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144837978, end_position_on_the_genomic_accession=144853640, orientation='minus', exon_count=6, OMIM=608216, no_hdr0=''),
  28992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28992, CurrentID=0, Status='live', Symbol='MACROD1', Aliases='LRP16', description='MACRO domain containing 1', other_designations='MACRO domain-containing protein 1|[Protein ADP-ribosylglutamate] hydrolase', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63998556, end_position_on_the_genomic_accession=64166130, orientation='minus', exon_count=16, OMIM=610400, no_hdr0=''),
  160065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160065, CurrentID=0, Status='live', Symbol='PATE1', Aliases='PATE', description='prostate and testis expressed 1', other_designations='expressed in prostate and testis', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125746293, end_position_on_the_genomic_accession=125749848, orientation='plus', exon_count=5, OMIM=606861, no_hdr0=''),
  28994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28994, CurrentID=0, Status='live', Symbol='PRO0471', Aliases='', description='uncharacterized LOC28994', other_designations='', map_location='3p13-q26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478987, CurrentID=0, Status='live', Symbol='RNU6-62P', Aliases='RNU6-62', description='RNA, U6 small nuclear 62, pseudogene', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28996, CurrentID=0, Status='live', Symbol='HIPK2', Aliases='PRO0593', description='homeodomain interacting protein kinase 2', other_designations='hHIPk2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139561570, end_position_on_the_genomic_accession=139777894, orientation='minus', exon_count=19, OMIM=606868, no_hdr0=''),
  28997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28997, CurrentID=0, Status='live', Symbol='PRO0611', Aliases='', description='PRO0611 protein', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=31460912, end_position_on_the_genomic_accession=31461481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  28998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28998, CurrentID=0, Status='live', Symbol='MRPL13', Aliases='L13, L13A, L13mt, RPL13, RPML13', description='mitochondrial ribosomal protein L13', other_designations='MRP-L13', map_location='8q22.1-q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120395844, end_position_on_the_genomic_accession=120445407, orientation='minus', exon_count=7, OMIM=610200, no_hdr0=''),
  266977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266977, CurrentID=0, Status='live', Symbol='ADGRF1', Aliases='GPR110, KPG_012, PGR19, hGPCR36', description='adhesion G protein-coupled receptor F1', other_designations='G protein-coupled receptor 110|G protein-coupled receptor PGR19|G-protein coupled receptor KPG_012|probable G-protein coupled receptor 110|seven transmembrane helix receptor', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46997703, end_position_on_the_genomic_accession=47042363, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105373579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373579, CurrentID=0, Status='live', Symbol='LOC105373579', Aliases='', description='uncharacterized LOC105373579', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118459850, end_position_on_the_genomic_accession=118493394, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106478989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478989, CurrentID=0, Status='live', Symbol='RNU6-73P', Aliases='RNU6-73', description='RNA, U6 small nuclear 73, pseudogene', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271455, CurrentID=0, Status='live', Symbol='RPL22P21', Aliases='RPL22_11_1520', description='ribosomal protein L22 pseudogene 21', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16084040, end_position_on_the_genomic_accession=16084378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  266979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266979, CurrentID=0, Status='live', Symbol='CTRCT27', Aliases='CCNP', description='Cataract, congenital, nuclear progressive', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607304, no_hdr0=''),
  29013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29013, CurrentID=0, Status='live', Symbol='PRO1596', Aliases='', description='uncharacterized LOC29013', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29015, CurrentID=0, Status='live', Symbol='SLC43A3', Aliases='EEG1, FOAP-13, PRO1659, SEEEG-1', description='solute carrier family 43 member 3', other_designations='likely ortholog of mouse embryonic epithelial gene 1|solute carrier family 43, member 3', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57406954, end_position_on_the_genomic_accession=57427580, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  29018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29018, CurrentID=0, Status='live', Symbol='FOXN3-AS2', Aliases='PRO1768', description='FOXN3 antisense RNA 2', other_designations='CTD-2303B20.1|FOXN3 antisense RNA 2 (non-protein coding)', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89576216, end_position_on_the_genomic_accession=89577477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369215, CurrentID=0, Status='live', Symbol='LOC105369215', Aliases='', description='putative uncharacterized protein LOC388882', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21300985, end_position_on_the_genomic_accession=21308381, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106478992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478992, CurrentID=0, Status='live', Symbol='RNA5SP50', Aliases='RN5S50', description='RNA, 5S ribosomal pseudogene 50', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29028, CurrentID=0, Status='live', Symbol='ATAD2', Aliases='ANCCA, CT137, PRO2000', description='ATPase family, AAA domain containing 2', other_designations='AAA nuclear coregulator cancer-associated protein', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123319850, end_position_on_the_genomic_accession=123396465, orientation='minus', exon_count=28, OMIM=611941, no_hdr0=''),
  29034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29034, CurrentID=0, Status='live', Symbol='CPS1-IT1', Aliases='CPS1-IT, CPS1IT, CPS1IT1, PRO0132', description='CPS1 intronic transcript 1', other_designations='CPS1 intronic transcript (non-protein coding)|CPS1 intronic transcript 1 (non-protein coding)', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210617571, end_position_on_the_genomic_accession=210619876, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29035, CurrentID=0, Status='live', Symbol='C16orf72', Aliases='PRO0149', description='chromosome 16 open reading frame 72', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9091680, end_position_on_the_genomic_accession=9119698, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929974, CurrentID=0, Status='live', Symbol='LOC101929974', Aliases='', description='uncharacterized LOC101929974', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131453247, end_position_on_the_genomic_accession=131560378, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105369216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369216, CurrentID=0, Status='live', Symbol='LOC105369216', Aliases='', description='syncytin-2-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88036866, end_position_on_the_genomic_accession=88039819, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  29053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29053, CurrentID=0, Status='live', Symbol='PRO0628', Aliases='', description='uncharacterized LOC29053', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=39666473, end_position_on_the_genomic_accession=39667647, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29057, CurrentID=0, Status='live', Symbol='FAM156A', Aliases='PRO0659, TMEM29', description='family with sequence similarity 156 member A', other_designations='family with sequence similarity 156, member A|protein FAM156A|transmembrane protein 29|transmembrane protein 29/29B', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52947254, end_position_on_the_genomic_accession=52995472, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29058, CurrentID=0, Status='live', Symbol='TMEM230', Aliases='C20orf30, HSPC274, dJ1116H23.2.1', description='transmembrane protein 230', other_designations='UPF0414 transmembrane protein C20orf30', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5099838, end_position_on_the_genomic_accession=5113087, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106478998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478998, CurrentID=0, Status='live', Symbol='RNA5SP174', Aliases='RN5S174', description='RNA, 5S ribosomal pseudogene 174', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29062, CurrentID=0, Status='live', Symbol='WDR91', Aliases='HSPC049', description='WD repeat domain 91', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135183838, end_position_on_the_genomic_accession=135211564, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  29063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29063, CurrentID=0, Status='live', Symbol='ZCCHC4', Aliases='HSPC052, ZGRF4', description='zinc finger, CCHC domain containing 4', other_designations='DHHC domain-containing zinc finger protein|zinc finger, GRF-type containing 4', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25312774, end_position_on_the_genomic_accession=25373176, orientation='plus', exon_count=16, OMIM=611792, no_hdr0=''),
  29065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29065, CurrentID=0, Status='live', Symbol='ASAP1-IT1', Aliases='ASAP1-IT, ASAP1IT, ASAP1IT1, DDEF1IT1, HSPC054, NCRNA00050', description='ASAP1 intronic transcript 1', other_designations='ASAP1 intronic transcript (non-protein coding)|ASAP1 intronic transcript 1 (non-protein coding)|ArfGAP with SH3 domain, ankyrin repeat and PH domain 1 intronic transcript 1 (non-protein coding)|DDEF1 intronic transcript 1 (non-protein coding)', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130295355, end_position_on_the_genomic_accession=130296533, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  29066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29066, CurrentID=0, Status='live', Symbol='ZC3H7A', Aliases='HSPC055, ZC3H7, ZC3HDC7', description='zinc finger CCCH-type containing 7A', other_designations='zinc finger CCCH type domain containing 7|zinc finger CCCH-type domain containing 7|zinc-finger protein AY163807', map_location='16p13-p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11750586, end_position_on_the_genomic_accession=11797258, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  29068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29068, CurrentID=0, Status='live', Symbol='ZBTB44', Aliases='BTBD15, HSPC063, ZNF851', description='zinc finger and BTB domain containing 44', other_designations='BTB (POZ) domain containing 15|BTB/POZ domain-containing protein 15|zinc finger protein 851', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130226677, end_position_on_the_genomic_accession=130314769, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29070, CurrentID=0, Status='live', Symbol='CCDC113', Aliases='HSPC065', description='coiled-coil domain containing 113', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58196964, end_position_on_the_genomic_accession=58283836, orientation='plus', exon_count=15, OMIM=616070, no_hdr0=''),
  29071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29071, CurrentID=0, Status='live', Symbol='C1GALT1C1', Aliases='C1GALT2, C38H2-L1, COSMC, HSPC067, MST143, TNPS', description='C1GALT1 specific chaperone 1', other_designations='C1Gal-T2|C38H2-like protein 1|beta 1,3-galactosyltransferase 2|core 1 beta1,3-galactosyltransferase 2|core 1 beta3-Gal-T2|core 1 beta3-galactosyltransferase-specific molecular chaperone', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120625674, end_position_on_the_genomic_accession=120630150, orientation='minus', exon_count=3, OMIM=300611, no_hdr0=''),
  29072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29072, CurrentID=0, Status='live', Symbol='SETD2', Aliases='HBP231, HIF-1, HIP-1, HSPC069, HYPB, KMT3A, SET2, p231HBP', description='SET domain containing 2', other_designations='huntingtin interacting protein 1|huntingtin yeast partner B|huntingtin-interacting protein B|lysine N-methyltransferase 3A', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47016408, end_position_on_the_genomic_accession=47163977, orientation='minus', exon_count=26, OMIM=612778, no_hdr0=''),
  29074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29074, CurrentID=0, Status='live', Symbol='MRPL18', Aliases='HSPC071, L18mt, MRP-L18', description='mitochondrial ribosomal protein L18', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159790460, end_position_on_the_genomic_accession=159798429, orientation='plus', exon_count=4, OMIM=611831, no_hdr0=''),
  29075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29075, CurrentID=0, Status='live', Symbol='LINC00652', Aliases='HSPC072', description='long intergenic non-protein coding RNA 652', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18787971, end_position_on_the_genomic_accession=18794584, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  29078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29078, CurrentID=0, Status='live', Symbol='NDUFAF4', Aliases='C6orf66, HRPAP20, HSPC125, My013, bA22L21.1', description='NADH:ubiquinone oxidoreductase complex assembly factor 4', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 4|NADH dehydrogenase (ubiquinone) complex I, assembly factor 4|hormone-regulated proliferation-associated protein of 20 kDa|hormone-regulated proliferation-associated protein, 20 kDa', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96889311, end_position_on_the_genomic_accession=96897891, orientation='minus', exon_count=3, OMIM=611776, no_hdr0=''),
  29079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29079, CurrentID=0, Status='live', Symbol='MED4', Aliases='ARC36, DRIP36, HSPC126, TRAP36, VDRIP', description='mediator complex subunit 4', other_designations='TRAP/SMCC/PC2 subunit p36|activator-recruited cofactor 36 kDa component|mediator, 34-kD subunit, homolog|vitamin D receptor-interacting protein, 36-kD|vitamin D3 receptor-interacting protein complex 36 kDa component', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48075728, end_position_on_the_genomic_accession=48095141, orientation='minus', exon_count=7, OMIM=605718, no_hdr0=''),
  29080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29080, CurrentID=0, Status='live', Symbol='CCDC59', Aliases='BR22, HSPC128, TAP26', description='coiled-coil domain containing 59', other_designations='TTF-1-associated protein 26|TTF-1-associated protein BR2|coiled-coil domain-containing protein 59', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82352304, end_position_on_the_genomic_accession=82358805, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  29081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29081, CurrentID=0, Status='live', Symbol='METTL5', Aliases='HSPC133', description='methyltransferase like 5', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169811757, end_position_on_the_genomic_accession=169824931, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29082, CurrentID=0, Status='live', Symbol='CHMP4A', Aliases='C14orf123, CHMP4, CHMP4B, HSPC134, SHAX2, SNF7, SNF7-1, VPS32-1, VPS32A', description='charged multivesicular body protein 4A', other_designations='SNF7 homolog associated with Alix-2|Snf7 homologue associated with Alix 2|chromatin modifying protein 4A|vacuolar protein sorting-associated protein 32-1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24209581, end_position_on_the_genomic_accession=24213830, orientation='minus', exon_count=6, OMIM=610051, no_hdr0=''),
  29083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29083, CurrentID=0, Status='live', Symbol='GTPBP8', Aliases='HSPC135', description='GTP-binding protein 8 (putative)', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112990482, end_position_on_the_genomic_accession=113001374, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  29085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29085, CurrentID=0, Status='live', Symbol='PHPT1', Aliases='CGI-202, HEL-S-132P, HSPC141, PHP14', description='phosphohistidine phosphatase 1', other_designations='epididymis secretory sperm binding protein Li 132P|protein janus-A homolog|sex-regulated protein janus-a', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136848816, end_position_on_the_genomic_accession=136851043, orientation='plus', exon_count=4, OMIM=610167, no_hdr0=''),
  29086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29086, CurrentID=0, Status='live', Symbol='BABAM1', Aliases='C19orf62, HSPC142, MERIT40, NBA1', description='BRISC and BRCA1 A complex member 1', other_designations='BRCA1-A complex subunit MERIT40|mediator of RAP80 interactions and targeting subunit of 40 kDa|mediator of Rap80 interactions and targeting 40 kDa|new component of the BRCA1 A complex|new component of the BRCA1-A complex|new component of the BRCAA1 A complex', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17267376, end_position_on_the_genomic_accession=17279353, orientation='plus', exon_count=9, OMIM=612766, no_hdr0=''),
  29087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29087, CurrentID=0, Status='live', Symbol='THYN1', Aliases='HSPC144, MDS012, MY105, THY28, THY28KD', description='thymocyte nuclear protein 1', other_designations='thymocyte protein thy28', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134248278, end_position_on_the_genomic_accession=134253401, orientation='minus', exon_count=8, OMIM=613739, no_hdr0=''),
  29088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29088, CurrentID=0, Status='live', Symbol='MRPL15', Aliases='HSPC145, L15mt, MRP-L15, MRP-L7, RPML7', description='mitochondrial ribosomal protein L15', other_designations='', map_location='8q11.2-q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54135221, end_position_on_the_genomic_accession=54148514, orientation='plus', exon_count=5, OMIM=611828, no_hdr0=''),
  29089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29089, CurrentID=0, Status='live', Symbol='UBE2T', Aliases='FANCT, HSPC150, PIG50', description='ubiquitin conjugating enzyme E2T', other_designations='E2 ubiquitin-conjugating enzyme T|HSPC150 protein similar to ubiquitin-conjugating enzyme|cell proliferation-inducing gene 50 protein|ubiquitin carrier protein T|ubiquitin-conjugating enzyme E2T (putative)|ubiquitin-protein ligase T', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202331657, end_position_on_the_genomic_accession=202341966, orientation='minus', exon_count=7, OMIM=610538, no_hdr0=''),
  29090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29090, CurrentID=0, Status='live', Symbol='TIMM21', Aliases='C18orf55, HSPC154, TIM21', description='translocase of inner mitochondrial membrane 21 homolog (yeast)', other_designations='TIM21-like protein, mitochondrial', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74148511, end_position_on_the_genomic_accession=74158969, orientation='plus', exon_count=6, OMIM=615180, no_hdr0=''),
  29091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29091, CurrentID=0, Status='live', Symbol='STXBP6', Aliases='HSPC156, amisyn', description='syntaxin binding protein 6 (amisyn)', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24809654, end_position_on_the_genomic_accession=25050297, orientation='minus', exon_count=11, OMIM=607958, no_hdr0=''),
  29092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29092, CurrentID=0, Status='live', Symbol='LINC00339', Aliases='HSPC157, NCRNA00339', description='long intergenic non-protein coding RNA 339', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22025191, end_position_on_the_genomic_accession=22031224, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  29093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29093, CurrentID=0, Status='live', Symbol='MRPL22', Aliases='HSPC158, L22mt, MRP-L22, MRP-L25, RPML25', description='mitochondrial ribosomal protein L22', other_designations='39S ribosomal protein L25, mitochondrial|L25mt', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154941073, end_position_on_the_genomic_accession=154969411, orientation='plus', exon_count=7, OMIM=611835, no_hdr0=''),
  29094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29094, CurrentID=0, Status='live', Symbol='LGALSL', Aliases='GRP, HSPC159', description='lectin, galactoside-binding-like', other_designations='lectin galactoside-binding-like protein', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64454193, end_position_on_the_genomic_accession=64461383, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  29095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29095, CurrentID=0, Status='live', Symbol='ORMDL2', Aliases='HSPC160, MST095, MSTP095, adoplin-2', description='ORMDL sphingolipid biosynthesis regulator 2', other_designations='expressed in normal aorta', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55818022, end_position_on_the_genomic_accession=55821175, orientation='plus', exon_count=4, OMIM=610074, no_hdr0=''),
  29097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29097, CurrentID=0, Status='live', Symbol='CNIH4', Aliases='CNIH-4, HSPC163', description='cornichon family AMPA receptor auxiliary protein 4', other_designations='cornichon homolog 4', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224356811, end_position_on_the_genomic_accession=224379452, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  29098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29098, CurrentID=0, Status='live', Symbol='RANGRF', Aliases='HSPC165, HSPC236, MOG1, RANGNRF', description='RAN guanine nucleotide release factor', other_designations='MOG1 homolog|ran-binding protein MOG1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8288387, end_position_on_the_genomic_accession=8290092, orientation='plus', exon_count=5, OMIM=607954, no_hdr0=''),
  29099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29099, CurrentID=0, Status='live', Symbol='COMMD9', Aliases='HSPC166', description='COMM domain containing 9', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36272292, end_position_on_the_genomic_accession=36289449, orientation='minus', exon_count=6, OMIM=612299, no_hdr0=''),
  29100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29100, CurrentID=0, Status='live', Symbol='TMEM208', Aliases='HSPC171', description='transmembrane protein 208', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67227108, end_position_on_the_genomic_accession=67229279, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  29101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29101, CurrentID=0, Status='live', Symbol='SSU72', Aliases='HSPC182, PNAS-120', description='SSU72 homolog, RNA polymerase II CTD phosphatase', other_designations='CTD phosphatase SSU72', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1541673, end_position_on_the_genomic_accession=1574882, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  29102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29102, CurrentID=0, Status='live', Symbol='DROSHA', Aliases='ETOHI2, HSA242976, RANSE3L, RN3, RNASE3L, RNASEN', description='drosha, ribonuclease type III', other_designations='RNase III|drosha, double-stranded RNA-specific endoribonuclease|nuclear RNase III Drosha|p241|protein Drosha|putative protein p241 which interacts with transcription factor Sp1|putative ribonuclease III|ribonuclease III, nuclear|ribonuclease type III, nuclear', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31400494, end_position_on_the_genomic_accession=31532175, orientation='minus', exon_count=37, OMIM=608828, no_hdr0=''),
  29103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29103, CurrentID=0, Status='live', Symbol='DNAJC15', Aliases='DNAJD1, HSD18, MCJ', description='DnaJ heat shock protein family (Hsp40) member C15', other_designations='DNAJ domain-containing|DnaJ (Hsp40) homolog, subfamily C, member 15|DnaJ (Hsp40) homolog, subfamily D, member 1|cell growth-inhibiting gene 22 protein|methylation-controlled J protein', map_location='13q14.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43023226, end_position_on_the_genomic_accession=43109170, orientation='plus', exon_count=6, OMIM=615339, no_hdr0=''),
  29104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29104, CurrentID=0, Status='live', Symbol='N6AMT1', Aliases='C21orf127, HEMK2, MTQ2, N6AMT, PRED28, m.HsaHemK2P', description='N-6 adenine-specific DNA methyltransferase 1 (putative)', other_designations='N6-DNA-methyltransferase|n(6)-adenine-specific DNA methyltransferase 1', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28872191, end_position_on_the_genomic_accession=28885373, orientation='minus', exon_count=7, OMIM=614553, no_hdr0=''),
  29105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29105, CurrentID=0, Status='live', Symbol='CFAP20', Aliases='C16orf80, EVORF, GTL3, fSAP23', description='cilia and flagella associated protein 20', other_designations='UPF0468 protein C16orf80|basal body up-regulated protein 22|flagellar associated protein 20 homolog|functional spliceosome-associated protein 23|gene trap locus 3|transcription factor IIB', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58113592, end_position_on_the_genomic_accession=58129392, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  29106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29106, CurrentID=0, Status='live', Symbol='SCG3', Aliases='SGIII', description='secretogranin III', other_designations='', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51681353, end_position_on_the_genomic_accession=51721026, orientation='plus', exon_count=12, OMIM=611796, no_hdr0=''),
  29107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29107, CurrentID=0, Status='live', Symbol='NXT1', Aliases='MTR2, P15', description='nuclear transport factor 2-like export factor 1', other_designations='NTF2-like export factor 1|NTX2-like export factor1|NUTF-like export factor 1|protein p15', map_location='20p12-p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23350736, end_position_on_the_genomic_accession=23354771, orientation='plus', exon_count=3, OMIM=605811, no_hdr0=''),
  29108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29108, CurrentID=0, Status='live', Symbol='PYCARD', Aliases='ASC, CARD5, TMS, TMS-1, TMS1', description='PYD and CARD domain containing', other_designations='caspase recruitment domain-containing protein 5|target of methylation-induced silencing 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31201486, end_position_on_the_genomic_accession=31202776, orientation='minus', exon_count=3, OMIM=606838, no_hdr0=''),
  29109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29109, CurrentID=0, Status='live', Symbol='FHOD1', Aliases='FHOS', description='formin homology 2 domain containing 1', other_designations='formin homolog overexpressed in spleen 1', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67229389, end_position_on_the_genomic_accession=67247522, orientation='minus', exon_count=26, OMIM=606881, no_hdr0=''),
  29110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29110, CurrentID=0, Status='live', Symbol='TBK1', Aliases='NAK, T2K', description='TANK-binding kinase 1', other_designations='NF-kB-activating kinase|NF-kappa-B-activating kinase', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64452060, end_position_on_the_genomic_accession=64502119, orientation='plus', exon_count=24, OMIM=604834, no_hdr0=''),
  29113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29113, CurrentID=0, Status='live', Symbol='C6orf15', Aliases='STG', description='chromosome 6 open reading frame 15', other_designations='taste bud-specific protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31111223, end_position_on_the_genomic_accession=31112555, orientation='minus', exon_count=2, OMIM=611401, no_hdr0=''),
  29114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29114, CurrentID=0, Status='live', Symbol='TAGLN3', Aliases='NP22, NP24, NP25', description='transgelin 3', other_designations='neuronal protein 22|neuronal protein NP25', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111998739, end_position_on_the_genomic_accession=112013888, orientation='plus', exon_count=5, OMIM=607953, no_hdr0=''),
  29115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29115, CurrentID=0, Status='live', Symbol='SAP30BP', Aliases='HCNGP, HTRG, HTRP', description='SAP30 binding protein', other_designations='HSV-1 binding|transcriptional regulator protein HCNGP', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75667116, end_position_on_the_genomic_accession=75708062, orientation='plus', exon_count=15, OMIM=610218, no_hdr0=''),
  29116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29116, CurrentID=0, Status='live', Symbol='MYLIP', Aliases='IDOL, MIR', description='myosin regulatory light chain interacting protein', other_designations='E3 ubiquitin ligase-inducible degrader of the low density lipoprotein receptor|band 4.1 superfamily member BZF1|cellular modulator of immune recognition (c-MIR)|inducible degrader of the LDL-receptor|myosin regulatory light chain-interacting protein', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16129086, end_position_on_the_genomic_accession=16148248, orientation='plus', exon_count=7, OMIM=610082, no_hdr0=''),
  29117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29117, CurrentID=0, Status='live', Symbol='BRD7', Aliases='BP75, CELTIX1, NAG4', description='bromodomain containing 7', other_designations='75 kDa bromodomain protein|protein CELTIX-1', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50319018, end_position_on_the_genomic_accession=50368966, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  29118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29118, CurrentID=0, Status='live', Symbol='DDX25', Aliases='GRTH', description='DEAD (Asp-Glu-Ala-Asp) box helicase 25', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 25|DEAD box protein 25|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 25|gonadotropin-regulated testicular RNA helicase', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125904366, end_position_on_the_genomic_accession=125923110, orientation='plus', exon_count=13, OMIM=607663, no_hdr0=''),
  29119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29119, CurrentID=0, Status='live', Symbol='CTNNA3', Aliases='ARVD13, VR22', description='catenin alpha 3', other_designations='alpha-T-catenin|alpha-catenin-like protein|catenin (cadherin-associated protein), alpha 3', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65912518, end_position_on_the_genomic_accession=67763610, orientation='minus', exon_count=23, OMIM=607667, no_hdr0=''),
  29121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29121, CurrentID=0, Status='live', Symbol='CLEC2D', Aliases='CLAX, LLT1, OCIL', description='C-type lectin domain family 2 member D', other_designations='C-type lectin domain family 2, member D|C-type lectin related f|C-type lectin superfamily 2, member D|LLT-1|lectin-like NK cell receptor|lectin-like transcript 1|osteoclast inhibitory lectin', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9669708, end_position_on_the_genomic_accession=9699555, orientation='plus', exon_count=6, OMIM=605659, no_hdr0=''),
  29122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29122, CurrentID=0, Status='live', Symbol='PRSS50', Aliases='CT20, TSP50', description='protease, serine 50', other_designations='cancer/testis antigen 20|protease, serine, 50|serine protease 50|testes specific protease 50|testes-specific protease 50|testis-specific protease-like protein 50', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46712116, end_position_on_the_genomic_accession=46717883, orientation='minus', exon_count=6, OMIM=607950, no_hdr0=''),
  29123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29123, CurrentID=0, Status='live', Symbol='ANKRD11', Aliases='ANCO-1, ANCO1, LZ16, T13', description='ankyrin repeat domain 11', other_designations='ankyrin repeat-containing cofactor 1|nasopharyngeal carcinoma susceptibility protein|truncated ankyrin repeat domain 11 aberrant transcript 1|truncated ankyrin repeat domain 11 aberrant transcript 2', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89267619, end_position_on_the_genomic_accession=89490574, orientation='minus', exon_count=20, OMIM=611192, no_hdr0=''),
  29124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29124, CurrentID=0, Status='live', Symbol='LGALS13', Aliases='GAL13, PLAC8, PP13', description='lectin, galactoside-binding, soluble, 13', other_designations='beta-galactoside-binding lectin|gal-13|galectin 13|galectin-13|placental protein 13|placental tissue protein 13', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39602359, end_position_on_the_genomic_accession=39607474, orientation='plus', exon_count=4, OMIM=608717, no_hdr0=''),
  29125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29125, CurrentID=0, Status='live', Symbol='C11orf21', Aliases='', description='chromosome 11 open reading frame 21', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2296277, end_position_on_the_genomic_accession=2301913, orientation='minus', exon_count=5, OMIM=611033, no_hdr0=''),
  29126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29126, CurrentID=0, Status='live', Symbol='CD274', Aliases='B7-H, B7H1, PD-L1, PDCD1L1, PDCD1LG1, PDL1', description='CD274 molecule', other_designations='B7 homolog 1|CD274 antigen|PDCD1 ligand 1', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5450381, end_position_on_the_genomic_accession=5470567, orientation='plus', exon_count=8, OMIM=605402, no_hdr0=''),
  29127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29127, CurrentID=0, Status='live', Symbol='RACGAP1', Aliases='CYK4, HsCYK-4, ID-GAP, MgcRacGAP', description='Rac GTPase activating protein 1', other_designations='GTPase activating protein|male germ cell RacGap|protein CYK4 homolg|protein CYK4 homolog', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49989162, end_position_on_the_genomic_accession=50033136, orientation='minus', exon_count=22, OMIM=604980, no_hdr0=''),
  29128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29128, CurrentID=0, Status='live', Symbol='UHRF1', Aliases='ICBP90, Np95, RNF106, hNP95, hUHRF1, huNp95', description='ubiquitin-like with PHD and ring finger domains 1', other_designations='RING finger protein 106|inverted CCAAT box-binding protein of 90 kDa|nuclear phosphoprotein 95|nuclear protein 95|nuclear zinc finger protein Np95|transcription factor ICBP90|ubiquitin-like PHD and RING finger domain-containing protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4903080, end_position_on_the_genomic_accession=4962154, orientation='plus', exon_count=20, OMIM=607990, no_hdr0=''),
  135927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135927, CurrentID=0, Status='live', Symbol='C7orf34', Aliases='ctm-1', description='chromosome 7 open reading frame 34', other_designations='MSSP-binding protein CTM-1', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142939506, end_position_on_the_genomic_accession=142940870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26701, CurrentID=0, Status='live', Symbol='OR2N1P', Aliases='HS6M1-2, OLFR89, OR6-7', description='olfactory receptor family 2 subfamily N member 1 pseudogene', other_designations='olfactory receptor 89|olfactory receptor, family 2, subfamily N, member 1 pseudogene', map_location='6p22.2-p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29137770, end_position_on_the_genomic_accession=29138916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479011, CurrentID=0, Status='live', Symbol='RNA5SP377', Aliases='RN5S377', description='RNA, 5S ribosomal pseudogene 377', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369219, CurrentID=0, Status='live', Symbol='LOC105369219', Aliases='', description='uncharacterized LOC105369219', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1692913, end_position_on_the_genomic_accession=1727566, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  267002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267002, CurrentID=0, Status='live', Symbol='PGBD2', Aliases='', description='piggyBac transposable element derived 2', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248872850, end_position_on_the_genomic_accession=248919146, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100131795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131795, CurrentID=0, Status='live', Symbol='LOC100131795', Aliases='', description='uncharacterized LOC100131795', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=52174537, end_position_on_the_genomic_accession=52174917, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79532, CurrentID=0, Status='live', Symbol='OR51A3P', Aliases='', description='olfactory receptor family 51 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4937338, end_position_on_the_genomic_accession=4938294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  267004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267004, CurrentID=0, Status='live', Symbol='PGBD3', Aliases='', description='piggyBac transposable element derived 3', other_designations='', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49515105, end_position_on_the_genomic_accession=49524281, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106479015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479015, CurrentID=0, Status='live', Symbol='RNA5SP448', Aliases='RN5S448', description='RNA, 5S ribosomal pseudogene 448', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  267005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267005, CurrentID=0, Status='live', Symbol='PGBD3P1', Aliases='', description='piggyBac transposable element derived 3 pseudogene 1', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61230943, end_position_on_the_genomic_accession=61232952, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441768, CurrentID=0, Status='live', Symbol='LOC441768', Aliases='', description='RAB31, member RAS oncogene family pseudogene', other_designations='', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46571533, end_position_on_the_genomic_accession=46572133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369220, CurrentID=0, Status='live', Symbol='LOC105369220', Aliases='', description='uncharacterized LOC105369220', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28565898, end_position_on_the_genomic_accession=28569862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  267006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267006, CurrentID=0, Status='live', Symbol='PGBD3P2', Aliases='', description='piggyBac transposable element derived 3 pseudogene 2', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29881699, end_position_on_the_genomic_accession=29882953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135935, CurrentID=0, Status='live', Symbol='NOBOX', Aliases='OG-2, OG2, OG2X, POF5, TCAG_12042', description='NOBOX oogenesis homeobox', other_designations='newborn ovary homeobox-encoding', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144396914, end_position_on_the_genomic_accession=144410227, orientation='minus', exon_count=11, OMIM=610934, no_hdr0=''),
  105373585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373585, CurrentID=0, Status='live', Symbol='LOC105373585', Aliases='', description='uncharacterized LOC105373585', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120543559, end_position_on_the_genomic_accession=120553374, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  267008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267008, CurrentID=0, Status='live', Symbol='PGBD3P4', Aliases='', description='piggyBac transposable element derived 3 pseudogene 4', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43586320, end_position_on_the_genomic_accession=43588329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479019, CurrentID=0, Status='live', Symbol='RNA5SP522', Aliases='RN5S522', description='RNA, 5S ribosomal pseudogene 522', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271461, CurrentID=0, Status='live', Symbol='RPL23AP39', Aliases='RPL23A_15_341', description='ribosomal protein L23a pseudogene 39', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1730054, end_position_on_the_genomic_accession=1730474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419855, CurrentID=0, Status='live', Symbol='SALL4P4', Aliases='', description='spalt-like transcription factor 4 pseudogene 4', other_designations='sal-like 4 pseudogene 4', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111182111, end_position_on_the_genomic_accession=111182696, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479022, CurrentID=0, Status='live', Symbol='CHEK2P5', Aliases='', description='checkpoint kinase 2 pseudogene 5', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  267012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267012, CurrentID=0, Status='live', Symbol='DAOA', Aliases='LG72, SG72', description='D-amino acid oxidase activator', other_designations='G72 transcript|schizophrenia- and bipolar disorder-associated protein G72', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105465867, end_position_on_the_genomic_accession=105491034, orientation='plus', exon_count=7, OMIM=607408, no_hdr0=''),
  441775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441775, CurrentID=0, Status='live', Symbol='RPL18P13', Aliases='RPL18_5_1500', description='ribosomal protein L18 pseudogene 13', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76235017, end_position_on_the_genomic_accession=76235665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135941, CurrentID=0, Status='live', Symbol='OR2A14', Aliases='OR2A14P, OR2A6, OST182', description='olfactory receptor family 2 subfamily A member 14', other_designations='olfactory receptor 2A6|olfactory receptor OR7-12|olfactory receptor, family 2, subfamily A, member 14 pseudogene|olfactory receptor, family 2, subfamily A, member 6', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144129113, end_position_on_the_genomic_accession=144130045, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  790951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790951, CurrentID=0, Status='live', Symbol='TRS-AGA3-1', Aliases='TRNAS4, TRS4', description='transfer RNA-Ser (AGA) 3-1', other_designations='tRNA serine 4 (anticodon AGA)|tRNA-Ser-AGA-3-1|transfer RNA serine 4 (anticodon AGA)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27532208, end_position_on_the_genomic_accession=27532289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  160298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160298, CurrentID=0, Status='live', Symbol='C11orf42', Aliases='', description='chromosome 11 open reading frame 42', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6205043, end_position_on_the_genomic_accession=6231450, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  267015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267015, CurrentID=0, Status='live', Symbol='HLA-S', Aliases='HLA-17', description='major histocompatibility complex, class I, S (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31381569, end_position_on_the_genomic_accession=31382487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402509, CurrentID=0, Status='live', Symbol='SLC29A4P1', Aliases='', description='solute carrier family 29 (nucleoside transporters), member 4 pseudogene 1', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57014285, end_position_on_the_genomic_accession=57020271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  4872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4872, CurrentID=0, Status='live', Symbol='NPM1P3', Aliases='NG2-6, NG3-3, NPM1P15, NPMP3', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 3', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene|nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 15', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5365696, end_position_on_the_genomic_accession=5367008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131909, CurrentID=0, Status='live', Symbol='FAM172BP', Aliases='FAM172B', description='family with sequence similarity 172 member B, pseudogene', other_designations='family with sequence similarity 172, member B, pseudogene', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101518867, end_position_on_the_genomic_accession=101523887, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420534, CurrentID=0, Status='live', Symbol='LOC100420534', Aliases='', description='Rho GTPase activating protein 20 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43442902, end_position_on_the_genomic_accession=43451643, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  160313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160313, CurrentID=0, Status='live', Symbol='KRT19P2', Aliases='KRT19P5', description='keratin 19 pseudogene 2', other_designations='keratin 19 pseudogene 5', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94834454, end_position_on_the_genomic_accession=94835028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=4874, CurrentID=0, Status='live', Symbol='NPM1P5', Aliases='NG5-4, NPMP5', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 5', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 5', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92285730, end_position_on_the_genomic_accession=92287234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479029, CurrentID=0, Status='live', Symbol='WBP1LP1', Aliases='', description='WW domain binding protein 1-like pseudogene 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  790952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790952, CurrentID=0, Status='live', Symbol='ESRG', Aliases='HESRG', description='embryonic stem cell related (non-protein coding)', other_designations='embryonic stem cell related protein|human embryonic stem cell related', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54632124, end_position_on_the_genomic_accession=54639857, orientation='minus', exon_count=4, OMIM=611473, no_hdr0=''),
  135948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135948, CurrentID=0, Status='live', Symbol='OR2F2', Aliases='OR7-1', description='olfactory receptor family 2 subfamily F member 2', other_designations='olfactory receptor 7-1|olfactory receptor OR7-6|olfactory receptor, family 2, subfamily F, member 2', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143935233, end_position_on_the_genomic_accession=143936186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441783, CurrentID=0, Status='live', Symbol='ACTG1P24', Aliases='POTENP', description='actin gamma 1 pseudogene 24', other_designations='POTE ankyrin domain family, member I pseudogene|POTE ankyrin domain family, member N, pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17242020, end_position_on_the_genomic_accession=17242890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369223, CurrentID=0, Status='live', Symbol='LOC105369223', Aliases='', description='uncharacterized LOC105369223', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8244235, end_position_on_the_genomic_accession=8298345, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  160335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160335, CurrentID=0, Status='live', Symbol='TMTC2', Aliases='IBDBP1', description='transmembrane and tetratricopeptide repeat containing 2', other_designations='IBDBP1 transcript 2|IBDBP1, transcript 1', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82686952, end_position_on_the_genomic_accession=83134868, orientation='plus', exon_count=13, OMIM=615856, no_hdr0=''),
  441784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441784, CurrentID=0, Status='live', Symbol='LOC441784', Aliases='', description='TBC1 domain family member 25 pseudogene', other_designations='TBC1 domain family, member 25 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19755264, end_position_on_the_genomic_accession=19756206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373588, CurrentID=0, Status='live', Symbol='LOC105373588', Aliases='', description='uncharacterized LOC105373588', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121138062, end_position_on_the_genomic_accession=121140709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  790953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790953, CurrentID=0, Status='live', Symbol='TSL', Aliases='', description='testis-expressed, seven-twelve, leukemia', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27615974, end_position_on_the_genomic_accession=27627492, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369224, CurrentID=0, Status='live', Symbol='LOC105369224', Aliases='', description='zinc finger protein 806 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30549744, end_position_on_the_genomic_accession=30550808, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  160364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160364, CurrentID=0, Status='live', Symbol='CLEC12A', Aliases='CD371, CLL-1, CLL1, DCAL-2, MICL', description='C-type lectin domain family 12 member A', other_designations='C-type lectin domain family 12, member A|C-type lectin protein CLL-1|C-type lectin-like molecule-1|dendritic cell-associated lectin 2|myeloid inhibitory C-type lectin-like receptor', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9951316, end_position_on_the_genomic_accession=9999354, orientation='plus', exon_count=13, OMIM=612088, no_hdr0=''),
  160365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160365, CurrentID=0, Status='live', Symbol='CLECL1', Aliases='DCAL-1, DCAL1', description='C-type lectin like 1', other_designations='C-type lectin-like 1|DC-associated lectin-1|dendritic cell associated lectin 1|dendritic cell-associated lectin 1|dendritic cell-associated lectin-1|type II transmembrane protein DCAL1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9715860, end_position_on_the_genomic_accession=9733299, orientation='minus', exon_count=4, OMIM=607467, no_hdr0=''),
  79537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79537, CurrentID=0, Status='live', Symbol='OR7E90P', Aliases='OR7E123P, OST705', description='olfactory receptor family 7 subfamily E member 90 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 123 pseudogene|olfactory receptor, family 7, subfamily E, member 90 pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158874593, end_position_on_the_genomic_accession=158875766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369225, CurrentID=0, Status='live', Symbol='LOC105369225', Aliases='', description='uncharacterized LOC105369225', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45554189, end_position_on_the_genomic_accession=45558649, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100037264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037264, CurrentID=0, Status='live', Symbol='STQTL6', Aliases='', description='stature quantitative trait locus 6', other_designations='', map_location='Xq24-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300591, no_hdr0=''),
  100037265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037265, CurrentID=0, Status='live', Symbol='STQTL7', Aliases='', description='stature quantitative trait locus 7', other_designations='', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609822, no_hdr0=''),
  100037266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037266, CurrentID=0, Status='live', Symbol='STQTL8', Aliases='', description='stature quantitative trait locus 8', other_designations='', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610114, no_hdr0=''),
  106479043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479043, CurrentID=0, Status='live', Symbol='MTATP6P3', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 3', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79538, CurrentID=0, Status='live', Symbol='OR13I1P', Aliases='OR13I2P', description='olfactory receptor family 13 subfamily I member 1 pseudogene', other_designations='olfactory receptor, family 13, subfamily I, member 1 pseudogene|olfactory receptor, family 131, subfamily I, member 2 pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104629755, end_position_on_the_genomic_accession=104631019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  160410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160410, CurrentID=0, Status='live', Symbol='PSMC6P2', Aliases='', description='proteasome 26S subunit, ATPase, 6 pseudogene 2', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 6 pseudogene', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68473731, end_position_on_the_genomic_accession=68474921, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157807, CurrentID=0, Status='live', Symbol='CLVS1', Aliases='CRALBPL, RLBP1L1', description='clavesin 1', other_designations='retinaldehyde binding protein 1-like 1|retinaldehyde-binding protein 1-like 1', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60966438, end_position_on_the_genomic_accession=61501645, orientation='plus', exon_count=10, OMIM=611292, no_hdr0=''),
  100271466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271466, CurrentID=0, Status='live', Symbol='RPL23AP54', Aliases='RPL23A_22_870', description='ribosomal protein L23a pseudogene 54', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5476509, end_position_on_the_genomic_accession=5476969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378636, CurrentID=0, Status='live', Symbol='LOC105378636', Aliases='', description='uncharacterized LOC105378636', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33120352, end_position_on_the_genomic_accession=33140802, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100037280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037280, CurrentID=0, Status='live', Symbol='UBE2D3P1', Aliases='RP4-680N4, UBE2D3L, UBE2D3P, UBE2D5P', description='ubiquitin conjugating enzyme E2D 3 pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast) pseudogene 1|ubiquitin-conjugating enzyme E2D 3 pseudogene 1|ubiquitin-conjugating enzyme E2D 5 pseudogene', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5292088, end_position_on_the_genomic_accession=5292971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  160418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160418, CurrentID=0, Status='live', Symbol='TMTC3', Aliases='SMILE', description='transmembrane and tetratricopeptide repeat containing 3', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88142296, end_position_on_the_genomic_accession=88199887, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  160419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160419, CurrentID=0, Status='live', Symbol='C12orf50', Aliases='', description='chromosome 12 open reading frame 50', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87980039, end_position_on_the_genomic_accession=88035403, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105369226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369226, CurrentID=0, Status='live', Symbol='LOC105369226', Aliases='', description='uncharacterized LOC105369226', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160428, CurrentID=0, Status='live', Symbol='ALDH1L2', Aliases='mtFDH', description='aldehyde dehydrogenase 1 family member L2', other_designations='10-formyltetrahydrofolate dehydrogenase ALDH1L2|aldehyde dehydrogenase 1 family, member L2|aldehyde dehydrogenase family 1 member L2, mitochondrial|mitochondrial 10-FTHFDH|probable 10-formyltetrahydrofolate dehydrogenase ALDH1L2', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105019784, end_position_on_the_genomic_accession=105084583, orientation='minus', exon_count=27, OMIM=613584, no_hdr0=''),
  100271467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271467, CurrentID=0, Status='live', Symbol='RPL23AP55', Aliases='RPL23A_23_884', description='ribosomal protein L23a pseudogene 55', other_designations='', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23206918, end_position_on_the_genomic_accession=23207369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369227, CurrentID=0, Status='live', Symbol='LOC105369227', Aliases='', description='uncharacterized LOC105369227', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79540, CurrentID=0, Status='live', Symbol='OR7E89P', Aliases='', description='olfactory receptor family 7 subfamily E member 89 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 89 pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158853504, end_position_on_the_genomic_accession=158854672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441806, CurrentID=0, Status='live', Symbol='LOC441806', Aliases='', description='proteasome 26S subunit, non-ATPase 8 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 8 pseudogene', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=856527, end_position_on_the_genomic_accession=857421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369228, CurrentID=0, Status='live', Symbol='LOC105369228', Aliases='', description='uncharacterized LOC105369228', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69975859, end_position_on_the_genomic_accession=69978635, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724329, CurrentID=0, Status='live', Symbol='LOC102724329', Aliases='', description='uncharacterized LOC102724329', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161764120, end_position_on_the_genomic_accession=161766227, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  160492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160492, CurrentID=0, Status='live', Symbol='LMNTD1', Aliases='IFLTD1, PAS1C1', description='lamin tail domain containing 1', other_designations='intermediate filament tail domain containing 1|intermediate filament tail domain-containing protein 1', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25476082, end_position_on_the_genomic_accession=25648594, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  105373593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373593, CurrentID=0, Status='live', Symbol='LOC105373593', Aliases='', description='uncharacterized LOC105373593', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=122491374, end_position_on_the_genomic_accession=122505123, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317774, CurrentID=0, Status='live', Symbol='SEPT7P1', Aliases='CDC10P', description='septin 7 pseudogene 1', other_designations='CDC10 cell division cycle 10 homolog pseudogene', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35157696, end_position_on_the_genomic_accession=35158999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271469, CurrentID=0, Status='live', Symbol='RPL26P31', Aliases='RPL26_11_1166', description='ribosomal protein L26 pseudogene 31', other_designations='', map_location='11q21-q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93052378, end_position_on_the_genomic_accession=93052874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369229, CurrentID=0, Status='live', Symbol='LOC105369229', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160518, CurrentID=0, Status='live', Symbol='DENND5B', Aliases='', description='DENN/MADD domain containing 5B', other_designations='rab6IP1-like protein', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31382223, end_position_on_the_genomic_accession=31591018, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  390997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390997, CurrentID=0, Status='live', Symbol='LOC390997', Aliases='', description='SET binding factor 1 pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11877769, end_position_on_the_genomic_accession=11880758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271470, CurrentID=0, Status='live', Symbol='RPL26P34', Aliases='RPL26_13_1302', description='ribosomal protein L26 pseudogene 34', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24704338, end_position_on_the_genomic_accession=24704788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283888, CurrentID=0, Status='live', Symbol='IL21R-AS1', Aliases='', description='IL21R antisense RNA 1', other_designations='IL21R antisense RNA 1 (non-protein coding)', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27447669, end_position_on_the_genomic_accession=27453393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479066, CurrentID=0, Status='live', Symbol='RNU7-106P', Aliases='', description='RNA, U7 small nuclear 106 pseudogene', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369230, CurrentID=0, Status='live', Symbol='LOC105369230', Aliases='HLA-DRB1', description='HLA class II histocompatibility antigen, DRB1-7 beta chain', other_designations='MHC class II antigen', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479067, CurrentID=0, Status='live', Symbol='RNU7-107P', Aliases='', description='RNA, U7 small nuclear 107 pseudogene', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100037417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037417, CurrentID=0, Status='live', Symbol='DDTL', Aliases='KB-226F1.2', description='D-dopachrome tautomerase-like', other_designations='D-dopachrome tautomerase-like protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23966838, end_position_on_the_genomic_accession=23972559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479069, CurrentID=0, Status='live', Symbol='RNU7-115P', Aliases='', description='RNA, U7 small nuclear 115 pseudogene', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479071, CurrentID=0, Status='live', Symbol='RNU7-123P', Aliases='', description='RNA, U7 small nuclear 123 pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369231, CurrentID=0, Status='live', Symbol='LOC105369231', Aliases='', description='PRAME family member 22-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271472, CurrentID=0, Status='live', Symbol='RPL29P22', Aliases='RPL29_12_1148', description='ribosomal protein L29 pseudogene 22', other_designations='', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63115767, end_position_on_the_genomic_accession=63116383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479075, CurrentID=0, Status='live', Symbol='RNU7-133P', Aliases='', description='RNA, U7 small nuclear 133 pseudogene', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418614, CurrentID=0, Status='live', Symbol='LOC100418614', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100009601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009601, CurrentID=0, Status='live', Symbol='TRY-GTA7-1', Aliases='TRNAY1, Tyr-tRNA', description='transfer RNA-Tyr (GTA) 7-1', other_designations='transfer RNA tyrosine 1 (anticodon GUA)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20659958, end_position_on_the_genomic_accession=20660051, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369232, CurrentID=0, Status='live', Symbol='LOC105369232', Aliases='', description='peptidyl-prolyl cis-trans isomerase FKBP4 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26767, CurrentID=0, Status='live', Symbol='RNU105B', Aliases='105B, E1-2', description='RNA, U105B small nucleolar', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8831185, end_position_on_the_genomic_accession=8831393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  553820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553820, CurrentID=0, Status='live', Symbol='EEF1A1P31', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 31', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154908824, end_position_on_the_genomic_accession=154919212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100034704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034704, CurrentID=0, Status='live', Symbol='RHDNS1', Aliases='', description='Renal hypodysplasia, nonsyndromic, 1', other_designations='', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610805, no_hdr0=''),
  26769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26769, CurrentID=0, Status='live', Symbol='SNORD81', Aliases='RNU104, U81, Z23', description='small nucleolar RNA, C/D box 81', other_designations='RNA, U104 small nucleolar', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173864175, end_position_on_the_genomic_accession=173864217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373597, CurrentID=0, Status='live', Symbol='LOC105373597', Aliases='', description='uncharacterized LOC105373597', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=125710919, end_position_on_the_genomic_accession=125793559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  160622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160622, CurrentID=0, Status='live', Symbol='GRASP', Aliases='TAMALIN', description='GRP1 (general receptor for phosphoinositides 1)-associated scaffold protein', other_designations='GRP1-associated scaffold protein', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52006937, end_position_on_the_genomic_accession=52015889, orientation='plus', exon_count=10, OMIM=612027, no_hdr0=''),
  105375280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375280, CurrentID=0, Status='live', Symbol='LOC105375280', Aliases='', description='uncharacterized LOC105375280', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=52415923, end_position_on_the_genomic_accession=52426630, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100009602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009602, CurrentID=0, Status='live', Symbol='TRY-GTA5-2', Aliases='TRNAY2, Tyr-tRNA', description='transfer RNA-Tyr (GTA) 5-2', other_designations='transfer RNA tyrosine 2 (anticodon GUA)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20663192, end_position_on_the_genomic_accession=20663285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479081, CurrentID=0, Status='live', Symbol='RNU7-152P', Aliases='', description='RNA, U7 small nuclear 152 pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401730, CurrentID=0, Status='live', Symbol='IPMKP1', Aliases='', description='inositol polyphosphate multikinase pseudogene 1', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22832144, end_position_on_the_genomic_accession=22838500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373598, CurrentID=0, Status='live', Symbol='LOC105373598', Aliases='', description='uncharacterized LOC105373598', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=125706791, end_position_on_the_genomic_accession=125811356, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79546, CurrentID=0, Status='live', Symbol='OR6M2P', Aliases='', description='olfactory receptor family 6 subfamily M member 2 pseudogene', other_designations='olfactory receptor, family 6, subfamily M, member 2 pseudogene', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123840958, end_position_on_the_genomic_accession=123841903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271474, CurrentID=0, Status='live', Symbol='RPL29P29', Aliases='RPL29_18_1328', description='ribosomal protein L29 pseudogene 29', other_designations='', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88236562, end_position_on_the_genomic_accession=88236969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26776, CurrentID=0, Status='live', Symbol='SNORA71B', Aliases='RNU71B, U71b', description='small nucleolar RNA, H/ACA box 71B', other_designations='RNA, U71B small nucleolar', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38425195, end_position_on_the_genomic_accession=38425330, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  646934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646934, CurrentID=0, Status='live', Symbol='LOC646934', Aliases='', description='golgin A6 family, member D pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78728949, end_position_on_the_genomic_accession=78735123, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100009603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009603, CurrentID=0, Status='live', Symbol='TRA-AGC11-1', Aliases='TRNAA2', description='transfer RNA-Ala (AGC) 11-1', other_designations='tRNA-Ala-AGC-11-1|transfer RNA alanine 2 (anticodon AGC)|transfer RNA alanine 2 (anticodon UGC)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26571864, end_position_on_the_genomic_accession=26571936, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369234, CurrentID=0, Status='live', Symbol='LOC105369234', Aliases='', description='contactin-associated protein-like 3B', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  684959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=684959, CurrentID=0, Status='live', Symbol='SNORA25', Aliases='ACA25', description='small nucleolar RNA, H/ACA box 25', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93730513, end_position_on_the_genomic_accession=93730646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  645984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645984, CurrentID=0, Status='live', Symbol='LOC645984', Aliases='', description='uncharacterized LOC645984', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=31256940, end_position_on_the_genomic_accession=31259731, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271475, CurrentID=0, Status='live', Symbol='RPL31P51', Aliases='RPL31_17_1239', description='ribosomal protein L31 pseudogene 51', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53632735, end_position_on_the_genomic_accession=53633178, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100009604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009604, CurrentID=0, Status='live', Symbol='TRA-AGC8-1', Aliases='TRNAA3', description='transfer RNA-Ala (AGC) 8-1', other_designations='tRNA-Ala-AGC-8-1|transfer RNA alanine 3 (anticodon AGC)|transfer RNA alanine 3 (anticodon UGC)', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27051214, end_position_on_the_genomic_accession=27051286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479091, CurrentID=0, Status='live', Symbol='RNU7-175P', Aliases='', description='RNA, U7 small nuclear 175 pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26783, CurrentID=0, Status='live', Symbol='SNORA65', Aliases='RNU65, U65', description='small nucleolar RNA, H/ACA box 65', other_designations='RNA, U65 small nucleolar|U65 small nucleolar RNA|U65 snoRNA', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127448502, end_position_on_the_genomic_accession=127448637, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271476, CurrentID=0, Status='live', Symbol='RPL31P52', Aliases='RPL31_18_1289', description='ribosomal protein L31 pseudogene 52', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120450400, end_position_on_the_genomic_accession=120450802, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479095, CurrentID=0, Status='live', Symbol='RNU7-187P', Aliases='', description='RNA, U7 small nuclear 187 pseudogene', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100009605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009605, CurrentID=0, Status='live', Symbol='TRF-GAA2-1', Aliases='TRNAF1', description='transfer RNA-Phe (GAA) 2-1', other_designations='tRNA-Phe-GAA-2-1|transfer RNA phenylalanine 1 (anticodon GAA)', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59566380, end_position_on_the_genomic_accession=59566452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369236, CurrentID=0, Status='live', Symbol='LOC105369236', Aliases='', description='structure-specific endonuclease subunit SLX1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553939, CurrentID=0, Status='live', Symbol='LOC553939', Aliases='', description='trimethyllysine hydroxylase, epsilon pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155349581, end_position_on_the_genomic_accession=155352238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479097, CurrentID=0, Status='live', Symbol='RNU7-195P', Aliases='', description='RNA, U7 small nuclear 195 pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160728, CurrentID=0, Status='live', Symbol='SLC5A8', Aliases='AIT, SMCT, SMCT1', description='solute carrier family 5 (sodium/monocarboxylate cotransporter), member 8', other_designations='apical iodide transporter|electrogenic sodium monocarboxylate cotransporter|sodium iodide-related cotransporter|solute carrier family 5 (iodide transporter), member 8|solute carrier family 5 member 8', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101156216, end_position_on_the_genomic_accession=101210274, orientation='minus', exon_count=16, OMIM=608044, no_hdr0=''),
  157860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157860, CurrentID=0, Status='live', Symbol='LOC157860', Aliases='', description='uncharacterized LOC157860', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79549, CurrentID=0, Status='live', Symbol='OR6J1', Aliases='OR6J1P, OR6J2', description='olfactory receptor family 6 subfamily J member 1 (gene/pseudogene)', other_designations='Olfactory receptor 6J1|olfactory receptor, family 6, subfamily J, member 1 pseudogene|olfactory receptor, family 6, subfamily J, member 2|seven transmembrane helix receptor', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22633755, end_position_on_the_genomic_accession=22634744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  790966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=790966, CurrentID=0, Status='live', Symbol='TRK-CTT2-4', Aliases='TRK2, TRNAK2', description='transfer RNA-Lys (CTT) 2-4', other_designations='tRNA lysine 2 (anticodon CUU)|tRNA-Lys-CTT-2-4|transfer RNA lysine 2 (anticodon CUU)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181221979, end_position_on_the_genomic_accession=181222051, orientation='minus', exon_count=0, OMIM=612996, no_hdr0=''),
  100009606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009606, CurrentID=0, Status='live', Symbol='TRF-GAA1-2', Aliases='TRNAF2', description='transfer RNA-Phe (GAA) 1-2', other_designations='tRNA-Phe-GAA-1-2|transfer RNA phenylalanine 2 (anticodon GAA)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124927843, end_position_on_the_genomic_accession=124927915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369237, CurrentID=0, Status='live', Symbol='LOC105369237', Aliases='', description='sperm protein associated with the nucleus on the X chromosome B/F', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160760, CurrentID=0, Status='live', Symbol='PPTC7', Aliases='TA-PP2C, TAPP2C', description='PTC7 protein phosphatase homolog', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110533245, end_position_on_the_genomic_accession=110583259, orientation='minus', exon_count=7, OMIM=609668, no_hdr0=''),
  160762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160762, CurrentID=0, Status='live', Symbol='CCDC63', Aliases='ODA5', description='coiled-coil domain containing 63', other_designations='outer row dynein assembly 5 homolog', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110844486, end_position_on_the_genomic_accession=110907535, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  79550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79550, CurrentID=0, Status='live', Symbol='OR4C4P', Aliases='OR4C17, OR4C17P, OR4C47P', description='olfactory receptor family 4 subfamily C member 4 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 17 pseudogene|olfactory receptor, family 4, subfamily C, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48345348, end_position_on_the_genomic_accession=48352447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553989, CurrentID=0, Status='live', Symbol='LFS3', Aliases='', description='Li-Fraumeni syndrome 3', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609266, no_hdr0=''),
  553991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553991, CurrentID=0, Status='live', Symbol='LGMD1G', Aliases='', description='limb girdle muscular dystrophy 1G (autosomal dominant)', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609115, no_hdr0=''),
  553992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553992, CurrentID=0, Status='live', Symbol='MGR7', Aliases='', description='Migraine with aura, susceptibility to, 7', other_designations='', map_location='15q11.2-q12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609179, no_hdr0=''),
  160777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160777, CurrentID=0, Status='live', Symbol='CCDC60', Aliases='', description='coiled-coil domain containing 60', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119334712, end_position_on_the_genomic_accession=119541047, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  26797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26797, CurrentID=0, Status='live', Symbol='SNORD52', Aliases='RNU52, U52', description='small nucleolar RNA, C/D box 52', other_designations='RNA, U52 small nuclear', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31837076, end_position_on_the_genomic_accession=31837139, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271479, CurrentID=0, Status='live', Symbol='RPL31P46', Aliases='RPL31_23_1157', description='ribosomal protein L31 pseudogene 46', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74919575, end_position_on_the_genomic_accession=74920006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441862, CurrentID=0, Status='live', Symbol='LOC441862', Aliases='', description='DUT pseudogene', other_designations='deoxyuridine triphosphatase pseudogene', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52451211, end_position_on_the_genomic_accession=52452253, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479112, CurrentID=0, Status='live', Symbol='RN7SKP38', Aliases='', description='RNA, 7SK small nuclear pseudogene 38', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160824, CurrentID=0, Status='live', Symbol='SP3P', Aliases='', description='Sp3 transcription factor pseudogene', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89359753, end_position_on_the_genomic_accession=89363502, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  29760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29760, CurrentID=0, Status='live', Symbol='BLNK', Aliases='AGM4, BASH-S, LY57, SLP-65, SLP65, bca, BLNK', description='B-cell linker', other_designations='B cell adaptor containing SH2 domain|B cell linker protein|B-cell activation|B-cell adapter containing a SH2 domain protein|B-cell adapter containing a Src homology 2 domain protein|Src homology 2 domain-containing leukocyte protein of 65 kDa|Src homology [SH2] domain-containing leukocyte protein of 65 kD|cytoplasmic adapter protein', map_location='10q23.2-q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96191699, end_position_on_the_genomic_accession=96271576, orientation='minus', exon_count=18, OMIM=604515, no_hdr0=''),
  29761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29761, CurrentID=0, Status='live', Symbol='USP25', Aliases='USP21', description='ubiquitin specific peptidase 25', other_designations='USP on chromosome 21|deubiquitinating enzyme 25|ubiquitin thioesterase 25|ubiquitin thiolesterase 25|ubiquitin-specific processing protease 25', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15729947, end_position_on_the_genomic_accession=15880072, orientation='plus', exon_count=29, OMIM=604736, no_hdr0=''),
  105372999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372999, CurrentID=0, Status='live', Symbol='LOC105372999', Aliases='', description='uncharacterized LOC105372999', other_designations='CTA-342B11.2', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31970830, end_position_on_the_genomic_accession=32037566, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  29763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29763, CurrentID=0, Status='live', Symbol='PACSIN3', Aliases='SDPIII', description='protein kinase C and casein kinase substrate in neurons 3', other_designations='SH3 domain-containing protein 6511|syndapin III', map_location='11p12-p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47177522, end_position_on_the_genomic_accession=47186459, orientation='minus', exon_count=12, OMIM=606513, no_hdr0=''),
  29764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29764, CurrentID=0, Status='live', Symbol='DDU', Aliases='', description='dermo-distortive urticaria', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=125630, no_hdr0=''),
  29765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29765, CurrentID=0, Status='live', Symbol='TMOD4', Aliases='SK-TMOD', description='tropomodulin 4', other_designations='actin-capping protein|skeletal muscle tropomodulin|skeletal tropomodulin|tropomodulin 4 (muscle)', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151169986, end_position_on_the_genomic_accession=151176304, orientation='minus', exon_count=10, OMIM=605834, no_hdr0=''),
  29766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29766, CurrentID=0, Status='live', Symbol='TMOD3', Aliases='UTMOD', description='tropomodulin 3', other_designations='tropomodulin 3 (ubiquitous)|ubiquitous tropomodulin', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51829628, end_position_on_the_genomic_accession=51912138, orientation='plus', exon_count=10, OMIM=605112, no_hdr0=''),
  29767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29767, CurrentID=0, Status='live', Symbol='TMOD2', Aliases='N-TMOD, NTMOD', description='tropomodulin 2', other_designations='neuronal tropomodulin|tropomodulin 2 (neuronal)', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51751561, end_position_on_the_genomic_accession=51816363, orientation='plus', exon_count=10, OMIM=602928, no_hdr0=''),
  29768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29768, CurrentID=0, Status='live', Symbol='SDHDP6', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 6', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 6', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25293358, end_position_on_the_genomic_accession=25294669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  29769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29769, CurrentID=0, Status='live', Symbol='SDHDP5', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 5', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 5', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178018046, end_position_on_the_genomic_accession=178019256, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29770, CurrentID=0, Status='live', Symbol='SDHDP4', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 4', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 4', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45882827, end_position_on_the_genomic_accession=45884125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  29771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29771, CurrentID=0, Status='live', Symbol='SDHDP3', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 3', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 3', map_location='3q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169709272, end_position_on_the_genomic_accession=169710578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29774, CurrentID=0, Status='live', Symbol='POM121L9P', Aliases='DKFZP434P211, UNQ2565', description='POM121 transmembrane nucleoporin like 9, pseudogene', other_designations='POM121 membrane glycoprotein-like 1 pseudogene|POM121 membrane glycoprotein-like 9, pseudogene|POM121 transmembrane nucleoporin-like 9, pseudogene|POM121-like protein', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24251621, end_position_on_the_genomic_accession=24265524, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  29775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29775, CurrentID=0, Status='live', Symbol='CARD10', Aliases='BIMP1, CARMA3', description='caspase recruitment domain family member 10', other_designations='Bcl10 binding protein and activator of NFKB|CARD-containing MAGUK 3 protein|CARD-containing MAGUK protein 3|carma 3|caspase recruitment domain family, member 10', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37490362, end_position_on_the_genomic_accession=37519203, orientation='minus', exon_count=20, OMIM=607209, no_hdr0=''),
  105373872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373872, CurrentID=0, Status='live', Symbol='LOC105373872', Aliases='', description='uncharacterized LOC105373872', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217203345, end_position_on_the_genomic_accession=217220857, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  29777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29777, CurrentID=0, Status='live', Symbol='ABT1', Aliases='hABT1', description='activator of basal transcription 1', other_designations='TATA-binding protein-binding protein|basal transcriptional activator', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26596943, end_position_on_the_genomic_accession=26600050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  29778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29778, CurrentID=0, Status='live', Symbol='MAP2K1P1', Aliases='', description='mitogen-activated protein kinase kinase 1 pseudogene 1', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30027244, end_position_on_the_genomic_accession=30029839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29779, CurrentID=0, Status='live', Symbol='DFNA33', Aliases='', description='deafness, autosomal dominant 33', other_designations='', map_location='13q34-qter', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614211, no_hdr0=''),
  29780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29780, CurrentID=0, Status='live', Symbol='PARVB', Aliases='CGI-56', description='parvin beta', other_designations='affixin|parvin, beta', map_location='22q13.2-q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43999211, end_position_on_the_genomic_accession=44169233, orientation='plus', exon_count=17, OMIM=608121, no_hdr0=''),
  29781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29781, CurrentID=0, Status='live', Symbol='NCAPH2', Aliases='CAPH2', description='non-SMC condensin II complex subunit H2', other_designations='CAP-H2 subunit of the condensin II complex|CTA-384D8.36|chromosome-associated protein H2|kleisin beta|non-SMC condensin II complex, subunit H2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50508215, end_position_on_the_genomic_accession=50524780, orientation='plus', exon_count=22, OMIM=611230, no_hdr0=''),
  160857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160857, CurrentID=0, Status='live', Symbol='CCDC122', Aliases='', description='coiled-coil domain containing 122', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43823931, end_position_on_the_genomic_accession=43879690, orientation='minus', exon_count=11, OMIM=613408, no_hdr0=''),
  106479119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479119, CurrentID=0, Status='live', Symbol='RN7SKP55', Aliases='', description='RNA, 7SK small nuclear pseudogene 55', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29789, CurrentID=0, Status='live', Symbol='OLA1', Aliases='DOC45, GBP45, GTBP9, GTPBP9, PTD004', description='Obg-like ATPase 1', other_designations='DNA damage-regulated overexpressed in cancer 45 protein|GTP-binding protein 9 (putative)|GTP-binding protein PTD004|homologous yeast-44.2 protein', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174072447, end_position_on_the_genomic_accession=174248637, orientation='minus', exon_count=11, OMIM=611175, no_hdr0=''),
  100271481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271481, CurrentID=0, Status='live', Symbol='RPL31P47', Aliases='RPL31_24_1186', description='ribosomal protein L31 pseudogene 47', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123051113, end_position_on_the_genomic_accession=123051517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29796, CurrentID=0, Status='live', Symbol='UQCR10', Aliases='HSPC051, HSPC119, HSPC151, QCR9, UCCR7.2, UCRC', description='ubiquinol-cytochrome c reductase, complex III subunit X', other_designations='cytochrome C1, nonheme 7kDa protein|cytochrome c1 non-heme 7 kDa protein|ubiquinol-cytochrome c reductase complex (7.2 kD)|ubiquinol-cytochrome c reductase complex 7.2 kDa protein|ubiquinol-cytochrome c reductase, complex III subunit X, 7.2kDa', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29767369, end_position_on_the_genomic_accession=29770413, orientation='plus', exon_count=2, OMIM=610843, no_hdr0=''),
  29797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29797, CurrentID=0, Status='live', Symbol='POM121L8P', Aliases='DKFZp434K191', description='POM121 transmembrane nucleoporin like 8, pseudogene', other_designations='POM121 membrane glycoprotein-like 1 pseudogene|POM121 membrane glycoprotein-like 8 pseudogene|POM121 transmembrane nucleoporin-like 8 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21282425, end_position_on_the_genomic_accession=21297730, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  29798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29798, CurrentID=0, Status='live', Symbol='C2orf27A', Aliases='C2orf27, C2orf27B', description='chromosome 2 open reading frame 27A', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131722491, end_position_on_the_genomic_accession=131767404, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  29799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29799, CurrentID=0, Status='live', Symbol='YPEL1', Aliases='FKSG3', description='yippee like 1', other_designations='DiGeorge syndrome-related protein|yippee-like 1', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21697536, end_position_on_the_genomic_accession=21735834, orientation='minus', exon_count=5, OMIM=608082, no_hdr0=''),
  29800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29800, CurrentID=0, Status='live', Symbol='ZDHHC1', Aliases='C16orf1, HSU90653, ZNF377', description='zinc finger, DHHC-type containing 1', other_designations='DHHC domain-containing cysteine-rich protein 1|DHHC-1|DHHC-domain-containing cysteine-rich protein|zinc finger DHHC domain-containing protein 1|zinc finger protein 377|zinc finger, DHHC domain containing 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67394419, end_position_on_the_genomic_accession=67416436, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  29801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29801, CurrentID=0, Status='live', Symbol='ZDHHC8', Aliases='DHHC8, ZDHHCL1, ZNF378', description='zinc finger, DHHC-type containing 8', other_designations='membrane-associated DHHC8 zinc finger protein|zinc finger protein 378|zinc finger, DHHC domain like containing 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20131841, end_position_on_the_genomic_accession=20148007, orientation='plus', exon_count=11, OMIM=608784, no_hdr0=''),
  29802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29802, CurrentID=0, Status='live', Symbol='VPREB3', Aliases='8HS20, N27C7-2', description='pre-B lymphocyte 3', other_designations='pre-B lymphocyte gene 3', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23752743, end_position_on_the_genomic_accession=23754443, orientation='minus', exon_count=2, OMIM=605017, no_hdr0=''),
  29803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29803, CurrentID=0, Status='live', Symbol='REPIN1', Aliases='AP4, RIP60, ZNF464, Zfp464', description='replication initiator 1', other_designations='60 kDa origin-specific DNA-binding protein|60 kDa replication initiation region protein|ATT-binding protein|DHFR oribeta-binding protein RIP60|H_DJ0584D14.12|replication initiation region protein (60kD)|zinc finger protein 464 (RIP60)|zinc finger protein AP4', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150368229, end_position_on_the_genomic_accession=150374044, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106479124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479124, CurrentID=0, Status='live', Symbol='RN7SKP65', Aliases='', description='RNA, 7SK small nuclear pseudogene 65', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160897, CurrentID=0, Status='live', Symbol='GPR180', Aliases='ITR', description='G protein-coupled receptor 180', other_designations='intimal thickness-related receptor', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94601850, end_position_on_the_genomic_accession=94634645, orientation='plus', exon_count=9, OMIM=607787, no_hdr0=''),
  106479126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479126, CurrentID=0, Status='live', Symbol='RN7SKP71', Aliases='', description='RNA, 7SK small nuclear pseudogene 71', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369242, CurrentID=0, Status='live', Symbol='LOC105369242', Aliases='', description='E3 ubiquitin-protein ligase HERC2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373360, CurrentID=0, Status='live', Symbol='LOC105373360', Aliases='', description='uncharacterized LOC105373360', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1251793, end_position_on_the_genomic_accession=1262948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  29841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29841, CurrentID=0, Status='live', Symbol='GRHL1', Aliases='LBP32, MGR, NH32, TFCP2L2', description='grainyhead-like transcription factor 1', other_designations='LBP protein 32|grainyhead-like 1|mammalian grainyhead|transcription factor CP2-like 2|transcription factor LBP-32', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9951663, end_position_on_the_genomic_accession=10002284, orientation='plus', exon_count=17, OMIM=609786, no_hdr0=''),
  29842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29842, CurrentID=0, Status='live', Symbol='TFCP2L1', Aliases='CRTR1, LBP-9, LBP9', description='transcription factor CP2-like 1', other_designations='CP2-related transcriptional repressor 1|CRTR-1|transcription factor LBP-9', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121216587, end_position_on_the_genomic_accession=121285209, orientation='minus', exon_count=17, OMIM=609785, no_hdr0=''),
  29843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29843, CurrentID=0, Status='live', Symbol='SENP1', Aliases='SuPr-2', description='SUMO1/sentrin specific peptidase 1', other_designations='SUMO1/sentrin specific protease 1|sentrin/SUMO-specific protease SENP1', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48042898, end_position_on_the_genomic_accession=48106308, orientation='minus', exon_count=20, OMIM=612157, no_hdr0=''),
  29844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29844, CurrentID=0, Status='live', Symbol='TFPT', Aliases='FB1, INO80F, amida', description='TCF3 (E2A) fusion partner (in childhood Leukemia)', other_designations='INO80 complex subunit F|amida, partner of the E2A', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54107013, end_position_on_the_genomic_accession=54115675, orientation='minus', exon_count=7, OMIM=609519, no_hdr0=''),
  29850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29850, CurrentID=0, Status='live', Symbol='TRPM5', Aliases='LTRPC5, MTR1', description='transient receptor potential cation channel, subfamily M, member 5', other_designations='LTrpC-5|MLSN1 and TRP-related|MLSN1- and TRP-related gene 1 protein|long transient receptor potential channel 5', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2403952, end_position_on_the_genomic_accession=2435017, orientation='minus', exon_count=28, OMIM=604600, no_hdr0=''),
  29851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29851, CurrentID=0, Status='live', Symbol='ICOS', Aliases='AILIM, CD278, CVID1', description='inducible T-cell co-stimulator', other_designations='activation-inducible lymphocyte immunomediatory molecule|inducible costimulator', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203936731, end_position_on_the_genomic_accession=203961579, orientation='plus', exon_count=8, OMIM=604558, no_hdr0=''),
  29855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29855, CurrentID=0, Status='live', Symbol='UBN1', Aliases='VT, VT4', description='ubinuclein 1', other_designations='HIRA-binding protein|ubiquitously expressed nuclear protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4847446, end_position_on_the_genomic_accession=4882401, orientation='plus', exon_count=22, OMIM=609771, no_hdr0=''),
  105369243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369243, CurrentID=0, Status='live', Symbol='LOC105369243', Aliases='', description='sulfotransferase 1A3-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419877, CurrentID=0, Status='live', Symbol='LOC100419877', Aliases='', description='bromodomain containing 7 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109241507, end_position_on_the_genomic_accession=109243160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130372, CurrentID=0, Status='live', Symbol='LOC100130372', Aliases='', description='uncharacterized LOC100130372', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=46126687, end_position_on_the_genomic_accession=46127154, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441884, CurrentID=0, Status='live', Symbol='CALR4P', Aliases='', description='calreticulin 4, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51554728, end_position_on_the_genomic_accession=51569750, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616281, CurrentID=0, Status='live', Symbol='MIR4788', Aliases='', description='microRNA 4788', other_designations='hsa-mir-4788', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134437827, end_position_on_the_genomic_accession=134437906, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29880, CurrentID=0, Status='live', Symbol='ALG5', Aliases='bA421P11.2', description='ALG5, dolichyl-phosphate beta-glucosyltransferase', other_designations='Alg5, S. cerevisiae, homolog of|asparagine-linked glycosylation 5 homolog (S. cerevisiae, dolichyl-phosphate beta-glucosyltransferase)|asparagine-linked glycosylation 5 homolog (yeast, dolichyl-phosphate beta-glucosyltransferase)|asparagine-linked glycosylation 5, dolichyl-phosphate beta-glucosyltransferase homolog|asparagine-linked glycosylation protein 5 homolog|dolP-glucosyltransferase|dolichyl phosphate glucosyltransferase', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36949770, end_position_on_the_genomic_accession=36999367, orientation='minus', exon_count=11, OMIM=604565, no_hdr0=''),
  29881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29881, CurrentID=0, Status='live', Symbol='NPC1L1', Aliases='NPC11L1', description='NPC1-like 1', other_designations='NPC1 (Niemann-Pick disease, type C1, gene)-like 1', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44512535, end_position_on_the_genomic_accession=44541330, orientation='minus', exon_count=20, OMIM=608010, no_hdr0=''),
  29882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29882, CurrentID=0, Status='live', Symbol='ANAPC2', Aliases='APC2', description='anaphase promoting complex subunit 2', other_designations='cyclosome subunit 2', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137174784, end_position_on_the_genomic_accession=137188605, orientation='minus', exon_count=14, OMIM=606946, no_hdr0=''),
  29883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29883, CurrentID=0, Status='live', Symbol='CNOT7', Aliases='CAF1, Caf1a, hCAF-1', description='CCR4-NOT transcription complex subunit 7', other_designations='BTG1 binding factor 1|BTG1-binding factor 1|CAF-1|CCR4-NOT transcription complex, subunit 7|CCR4-associated factor 1|carbon catabolite repressor protein (CCR4)-associative factor 1', map_location='8p22-p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17226030, end_position_on_the_genomic_accession=17246878, orientation='minus', exon_count=10, OMIM=604913, no_hdr0=''),
  29886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29886, CurrentID=0, Status='live', Symbol='SNX8', Aliases='Mvp1', description='sorting nexin 8', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2251770, end_position_on_the_genomic_accession=2314475, orientation='minus', exon_count=13, OMIM=614905, no_hdr0=''),
  29887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29887, CurrentID=0, Status='live', Symbol='SNX10', Aliases='OPTB8', description='sorting nexin 10', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26291895, end_position_on_the_genomic_accession=26374383, orientation='plus', exon_count=11, OMIM=614780, no_hdr0=''),
  29888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29888, CurrentID=0, Status='live', Symbol='STRN4', Aliases='ZIN, zinedin', description='striatin 4', other_designations='striatin, calmodulin binding protein 4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46719509, end_position_on_the_genomic_accession=46746475, orientation='minus', exon_count=22, OMIM=614767, no_hdr0=''),
  29889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29889, CurrentID=0, Status='live', Symbol='GNL2', Aliases='HUMAUANTIG, NGP1, Ngp-1, Nog2, Nug2', description='guanine nucleotide binding protein-like 2 (nucleolar)', other_designations="autoantigen NGP-1|novel nucleolar guanosine 5'-triphosphate binding protein autoantigen|nucleolar G-protein gene 1|nucleolar GTPase|nucleostemin-2", map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37566812, end_position_on_the_genomic_accession=37595985, orientation='minus', exon_count=17, OMIM=609365, no_hdr0=''),
  29890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29890, CurrentID=0, Status='live', Symbol='RBM15B', Aliases='HUMAGCGB, OTT3', description='RNA binding motif protein 15B', other_designations='RNA-binding motif protein 15B|chromosome 3p21.1 gene sequence|huOTT3|one twenty two protein 3|one-twenty two protein 3', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51391268, end_position_on_the_genomic_accession=51397908, orientation='plus', exon_count=1, OMIM=612602, no_hdr0=''),
  100419878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419878, CurrentID=0, Status='live', Symbol='LOC100419878', Aliases='', description='methyltransferase like 5 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126624792, end_position_on_the_genomic_accession=126626108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29893, CurrentID=0, Status='live', Symbol='PSMC3IP', Aliases='GT198, HOP2, HUMGT198A, ODG3, TBPIP', description='PSMC3 interacting protein', other_designations='DBD-interacting|TBP-1-interacting protein|nuclear receptor coactivator GT198|proteasome 26S ATPase subunit 3-interacting protein|tat-binding protein 1-interacting protein', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42572310, end_position_on_the_genomic_accession=42577831, orientation='minus', exon_count=9, OMIM=608665, no_hdr0=''),
  29894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29894, CurrentID=0, Status='live', Symbol='CPSF1', Aliases='CPSF160, HSU37012, P/cl.18', description='cleavage and polyadenylation specific factor 1', other_designations='CPSF 160 kDa subunit|cleavage and polyadenylation specific factor 1, 160kDa|cleavage and polyadenylation specificity factor 160 kDa subunit|polyadenylation specificity factor', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144393231, end_position_on_the_genomic_accession=144409400, orientation='minus', exon_count=42, OMIM=606027, no_hdr0=''),
  29895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29895, CurrentID=0, Status='live', Symbol='MYLPF', Aliases='HUMMLC2B, MRLC2, MYL11', description='myosin light chain, phosphorylatable, fast skeletal muscle', other_designations='MLC2B|fast skeletal myosin light chain 2|myosin light chain 2|myosin, light chain 11, regulatory', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30374802, end_position_on_the_genomic_accession=30377989, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  29896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29896, CurrentID=0, Status='live', Symbol='TRA2A', Aliases='AWMS1, HSU53209', description='transformer 2 alpha homolog (Drosophila)', other_designations='TRA-2 alpha|TRA2-alpha|Tra2alpha|htra-2 alpha|htra-2-alpha|putative MAPK activating protein PM24|transformer-2 alpha|transformer-2 protein homolog A', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23504780, end_position_on_the_genomic_accession=23532041, orientation='minus', exon_count=10, OMIM=602718, no_hdr0=''),
  140660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140660, CurrentID=0, Status='live', Symbol='RPL21P4', Aliases='RPL21_58_1548', description='ribosomal protein L21 pseudogene 4', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43079261, end_position_on_the_genomic_accession=43079816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29899, CurrentID=0, Status='live', Symbol='GPSM2', Aliases='CMCS, DFNB82, LGN, PINS', description='G-protein signaling modulator 2', other_designations='G-protein signalling modulator 2 (AGS3-like, C. elegans)|mosaic protein LGN', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108875350, end_position_on_the_genomic_accession=108930422, orientation='plus', exon_count=18, OMIM=609245, no_hdr0=''),
  554188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554188, CurrentID=0, Status='live', Symbol='FCMTE2', Aliases='ADCME, BAFME2, FAME, FAME2', description='familial cortical myoclonic tremor with epilepsy 2', other_designations='benign adult familial myoclonic epilepsy 2|cortical myoclonus and epilepsy, autosomal dominant|familial adult myoclonic epilepsy 2', map_location='2p11.1-q12.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607876, no_hdr0=''),
  29901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29901, CurrentID=0, Status='live', Symbol='SAC3D1', Aliases='HSU79266, SHD1', description='SAC3 domain containing 1', other_designations='SAC3 homology domain-containing protein 1|Sac3 homology domain 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65040904, end_position_on_the_genomic_accession=65044828, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  29902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29902, CurrentID=0, Status='live', Symbol='FAM216A', Aliases='C12orf24, HSU79274', description='family with sequence similarity 216 member A', other_designations='family with sequence similarity 216, member A|protein predicted by clone 23733|uncharacterized protein C12orf24', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110468427, end_position_on_the_genomic_accession=110490387, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  29903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29903, CurrentID=0, Status='live', Symbol='CCDC106', Aliases='HSU79303, ZNF581', description='coiled-coil domain containing 106', other_designations='protein predicted by clone 23882', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55647568, end_position_on_the_genomic_accession=55653161, orientation='plus', exon_count=6, OMIM=613478, no_hdr0=''),
  29904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29904, CurrentID=0, Status='live', Symbol='EEF2K', Aliases='HSU93850, eEF-2K', description='eukaryotic elongation factor 2 kinase', other_designations='calcium/calmodulin-dependent eukaryotic elongation factor 2 kinase|calcium/calmodulin-dependent eukaryotic elongation factor-2 kinase|calmodulin-dependent protein kinase III|eEF-2 kinase|elongation factor-2 kinase|eukaroytic elongation factor 2 kinase|eukaryotic elongation factor-2 kinase', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22206271, end_position_on_the_genomic_accession=22288745, orientation='plus', exon_count=19, OMIM=606968, no_hdr0=''),
  105373609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373609, CurrentID=0, Status='live', Symbol='LOC105373609', Aliases='', description='uncharacterized LOC105373609', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127560045, end_position_on_the_genomic_accession=127571395, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  29906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29906, CurrentID=0, Status='live', Symbol='ST8SIA5', Aliases='SIAT8-E, SIAT8E, ST8SiaV', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 5', other_designations='sialyltransferase 8E (alpha-2, 8-polysialytransferase)|sialyltransferase St8Sia V', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46679118, end_position_on_the_genomic_accession=46757169, orientation='minus', exon_count=9, OMIM=607162, no_hdr0=''),
  29907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29907, CurrentID=0, Status='live', Symbol='SNX15', Aliases='HSAF001435', description='sorting nexin 15', other_designations='clone iota unknown protein', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65027408, end_position_on_the_genomic_accession=65040572, orientation='plus', exon_count=8, OMIM=605964, no_hdr0=''),
  29909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29909, CurrentID=0, Status='live', Symbol='GPR171', Aliases='H963', description='G protein-coupled receptor 171', other_designations='F730001G15Rik|G-protein coupled receptor H963|platelet activating receptor homolog', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151197754, end_position_on_the_genomic_accession=151203222, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  29911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29911, CurrentID=0, Status='live', Symbol='HOOK2', Aliases='HK2', description='hook microtubule-tethering protein 2', other_designations='h-hook2|hHK2|hook homolog 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12763003, end_position_on_the_genomic_accession=12778484, orientation='minus', exon_count=25, OMIM=607824, no_hdr0=''),
  102659288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102659288, CurrentID=0, Status='live', Symbol='KCCAT333', Aliases='', description='renal clear cell carcinoma-associated transcript 333', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17374917, end_position_on_the_genomic_accession=17467256, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  554201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554201, CurrentID=0, Status='live', Symbol='CCDC148-AS1', Aliases='', description='CCDC148 antisense RNA 1', other_designations='CCDC148 antisense RNA 1 (non-protein coding)|tmp_locus_26', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158166650, end_position_on_the_genomic_accession=158236169, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  554202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554202, CurrentID=0, Status='live', Symbol='MIR31HG', Aliases='', description='MIR31 host gene', other_designations='MIR31 host gene (non-protein coding)', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21454268, end_position_on_the_genomic_accession=21559698, orientation='minus', exon_count=4, OMIM=616356, no_hdr0=''),
  554203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554203, CurrentID=0, Status='live', Symbol='JPX', Aliases='DCBALD06, ENOX, LINC00183, NCRNA00183', description='JPX transcript, XIST activator (non-protein coding)', other_designations='expressed neighbor of XIST|just proximal to XIST', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73944324, end_position_on_the_genomic_accession=74070384, orientation='plus', exon_count=5, OMIM=300832, no_hdr0=''),
  29916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29916, CurrentID=0, Status='live', Symbol='SNX11', Aliases='', description='sorting nexin 11', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48107554, end_position_on_the_genomic_accession=48122745, orientation='plus', exon_count=8, OMIM=614906, no_hdr0=''),
  554206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554206, CurrentID=0, Status='live', Symbol='LOC554206', Aliases='', description='leucine carboxyl methyltransferase 1 pseudogene', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25031741, end_position_on_the_genomic_accession=25032777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29919, CurrentID=0, Status='live', Symbol='C18orf8', Aliases='HsT2591, MIC1, Mic-1', description='chromosome 18 open reading frame 8', other_designations='colon cancer-associated protein Mic1|macrophage inhibitory cytokine 1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23503470, end_position_on_the_genomic_accession=23531807, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  29920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29920, CurrentID=0, Status='live', Symbol='PYCR2', Aliases='HLD10, P5CR2', description='pyrroline-5-carboxylate reductase family member 2', other_designations='P5C reductase 2|pyrroline 5-carboxylate reductase isoform|pyrroline-5-carboxylate reductase family, member 2', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225919877, end_position_on_the_genomic_accession=225924340, orientation='minus', exon_count=7, OMIM=616406, no_hdr0=''),
  29922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29922, CurrentID=0, Status='live', Symbol='NME7', Aliases='CFAP67, MN23H7, NDK 7, NDK7, nm23-H7', description='NME/NM23 family member 7', other_designations='NDP kinase 7|cilia and flagella associated protein 67|non-metastatic cells 7, protein expressed in (nucleoside-diphosphate kinase)|nucleoside-diphosphate kinase 7', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169131998, end_position_on_the_genomic_accession=169367963, orientation='minus', exon_count=13, OMIM=613465, no_hdr0=''),
  29923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29923, CurrentID=0, Status='live', Symbol='HILPDA', Aliases='C7orf68, HIG-2, HIG2', description='hypoxia inducible lipid droplet-associated', other_designations='hypoxia inducible gene 2|hypoxia-inducible gene 2 protein|hypoxia-inducible protein 2', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128455830, end_position_on_the_genomic_accession=128458418, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  554212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554212, CurrentID=0, Status='live', Symbol='MIR448', Aliases='MIRN448, hsa-mir-448, miRNA448', description='microRNA 448', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114823454, end_position_on_the_genomic_accession=114823564, orientation='plus', exon_count=1, OMIM=300686, no_hdr0=''),
  29925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29925, CurrentID=0, Status='live', Symbol='GMPPB', Aliases='MDDGA14, MDDGB14, MDDGC14', description='GDP-mannose pyrophosphorylase B', other_designations='GTP-mannose-1-phosphate guanylyltransferase beta', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49721476, end_position_on_the_genomic_accession=49723974, orientation='minus', exon_count=9, OMIM=615320, no_hdr0=''),
  554214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554214, CurrentID=0, Status='live', Symbol='MIR450A1', Aliases='MIRN450, MIRN450-1, MIRN450A-1, MIRN450A1, hsa-mir-450, mir-450a-1', description='microRNA 450a-1', other_designations='hsa-mir-450-1|hsa-mir-450a-1|microRNA 450-1', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134540341, end_position_on_the_genomic_accession=134540431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  29927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29927, CurrentID=0, Status='live', Symbol='SEC61A1', Aliases='HSEC61, SEC61, SEC61A', description='Sec61 translocon alpha 1 subunit', other_designations='Sec61 alpha 1 subunit|Sec61 alpha-1|protein transport protein SEC61 alpha subunit|protein transport protein Sec61 subunit alpha|sec61 homolog', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128052369, end_position_on_the_genomic_accession=128071683, orientation='plus', exon_count=12, OMIM=609213, no_hdr0=''),
  29928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29928, CurrentID=0, Status='live', Symbol='TIMM22', Aliases='TEX4, TIM22', description='translocase of inner mitochondrial membrane 22 homolog (yeast)', other_designations='putative membrane protein|testis-expressed sequence 4', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=997117, end_position_on_the_genomic_accession=1002167, orientation='plus', exon_count=4, OMIM=607251, no_hdr0=''),
  29929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29929, CurrentID=0, Status='live', Symbol='ALG6', Aliases='CDG1C', description='ALG6, alpha-1,3-glucosyltransferase', other_designations='Man(9)GlcNAc(2)-PP-Dol alpha-1,3-glucosyltransferase|asparagine-linked glycosylation 6 homolog (S. cerevisiae, alpha-1,3-glucosyltransferase)|asparagine-linked glycosylation 6 homolog (yeast, alpha-1,3-glucosyltransferase)|asparagine-linked glycosylation 6, alpha-1,3-glucosyltransferase homolog|asparagine-linked glycosylation protein 6 homolog|dol-P-Glc:Man(9)GlcNAc(2)-PP-Dol alpha-1,3-glucosyltransferase|dolichyl-P-Glc:Man(9)GlcNAc(2)-PP-dolichol alpha- 1->3-glucosyltransferase|dolichyl-P-Glc:Man9GlcNAc2-PP-dolichyl glucosyltransferase|dolichyl-P-Glc:Man9GlcNAc2-PP-dolichylglucosyltransferase', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63367590, end_position_on_the_genomic_accession=63438562, orientation='plus', exon_count=15, OMIM=604566, no_hdr0=''),
  29930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29930, CurrentID=0, Status='live', Symbol='PCDHB1', Aliases='PCDH-BETA1', description='protocadherin beta 1', other_designations='PCDH-beta-1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141051376, end_position_on_the_genomic_accession=141053962, orientation='plus', exon_count=1, OMIM=606327, no_hdr0=''),
  29931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29931, CurrentID=0, Status='live', Symbol='LINC00312', Aliases='ERR-10, ERR10, LMCD1DN, LOH3CR2A, NAG-7, NAG7, NCRNA00312', description='long intergenic non-protein coding RNA 312', other_designations='LMCD1 downstream neighbor|estrogen receptor repressor 10|loss of heterozygosity, 3, chromosomal region 2, gene A|nasopharyngeal carcinoma candidate 7', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8571782, end_position_on_the_genomic_accession=8574668, orientation='plus', exon_count=1, OMIM=610485, no_hdr0=''),
  29933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29933, CurrentID=0, Status='live', Symbol='GPR132', Aliases='G2A', description='G protein-coupled receptor 132', other_designations='G2 accumulation protein', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105049389, end_position_on_the_genomic_accession=105065550, orientation='minus', exon_count=5, OMIM=606167, no_hdr0=''),
  29934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29934, CurrentID=0, Status='live', Symbol='SNX12', Aliases='', description='sorting nexin 12', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71056740, end_position_on_the_genomic_accession=71073426, orientation='minus', exon_count=6, OMIM=300883, no_hdr0=''),
  29935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29935, CurrentID=0, Status='live', Symbol='RPA4', Aliases='HSU24186', description='replication protein A4', other_designations='RF-A protein 4|RP-A p30|replication factor A protein 4|replication protein A complex 34 kd subunit homolog Rpa4|replication protein A4, 30kDa|replication protein A4, 34kDa', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96883908, end_position_on_the_genomic_accession=96885467, orientation='plus', exon_count=1, OMIM=300767, no_hdr0=''),
  554224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554224, CurrentID=0, Status='live', Symbol='PPIP5K1P1', Aliases='HISPPD2B, HISPPD2BP', description='diphosphoinositol pentakisphosphate kinase 1 pseudogene 1', other_designations='histidine acid phosphatase domain containing 2B (pseudogene)', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43664417, end_position_on_the_genomic_accession=43684727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  29937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29937, CurrentID=0, Status='live', Symbol='NENF', Aliases='CIR2, SCIRP10, SPUF', description='neudesin neurotrophic factor', other_designations='SCIRP10-related protein|Spinal cord injury related protein 10|cell growth-inhibiting protein 47|cell immortalization-related protein 2|neuron-derived neurotrophic factor|secreted protein of unknown function', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212432887, end_position_on_the_genomic_accession=212446379, orientation='plus', exon_count=4, OMIM=611874, no_hdr0=''),
  554226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554226, CurrentID=0, Status='live', Symbol='ANKRD30BL', Aliases='ANKRD30BP3, NCRNA00164', description='ankyrin repeat domain 30B-like', other_designations='ankyrin repeat domain 30B pseudogene 3', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132147591, end_position_on_the_genomic_accession=132257969, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29940, CurrentID=0, Status='live', Symbol='DSE', Aliases='DSEPPI, EDSMC2, SART-2, SART2, DSE', description='dermatan sulfate epimerase', other_designations='DS epimerase|chondroitin-glucuronate 5-epimerase|squamous cell carcinoma antigen recognized by T-cells 2', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116280120, end_position_on_the_genomic_accession=116443475, orientation='plus', exon_count=7, OMIM=605942, no_hdr0=''),
  29941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29941, CurrentID=0, Status='live', Symbol='PKN3', Aliases='UTDP4-1', description='protein kinase N3', other_designations='protein kinase PKN-beta|protein-kinase C-related kinase 3', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128702499, end_position_on_the_genomic_accession=128720920, orientation='plus', exon_count=23, OMIM=610714, no_hdr0=''),
  29942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29942, CurrentID=0, Status='live', Symbol='PURG', Aliases='PURG-A-B, PURG', description='purine-rich element binding protein G', other_designations='Pur-gamma', map_location='8p11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30995802, end_position_on_the_genomic_accession=31033715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  29943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29943, CurrentID=0, Status='live', Symbol='PADI1', Aliases='HPAD10, PAD1, PDI, PDI1', description='peptidyl arginine deiminase, type I', other_designations='hPAD-colony 10|peptidylarginine deiminase I|protein-arginine deiminase type I', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17205123, end_position_on_the_genomic_accession=17246006, orientation='plus', exon_count=19, OMIM=607934, no_hdr0=''),
  29944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29944, CurrentID=0, Status='live', Symbol='PNMA3', Aliases='MA3, MA5', description='paraneoplastic Ma antigen 3', other_designations='paraneoplastic cancer-testis-brain antigen', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153056409, end_position_on_the_genomic_accession=153060467, orientation='plus', exon_count=2, OMIM=300675, no_hdr0=''),
  29945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29945, CurrentID=0, Status='live', Symbol='ANAPC4', Aliases='APC4', description='anaphase promoting complex subunit 4', other_designations='cyclosome subunit 4', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25377226, end_position_on_the_genomic_accession=25418498, orientation='plus', exon_count=29, OMIM=606947, no_hdr0=''),
  29946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29946, CurrentID=0, Status='live', Symbol='SERTAD3', Aliases='RBT1', description='SERTA domain containing 3', other_designations='RPA-binding trans-activator|replication protein-binding trans-activator', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40440841, end_position_on_the_genomic_accession=40444375, orientation='minus', exon_count=4, OMIM=612125, no_hdr0=''),
  554235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554235, CurrentID=0, Status='live', Symbol='ASPDH', Aliases='', description='aspartate dehydrogenase domain containing', other_designations='aspartate dehydrogenase domain-containing protein', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50511600, end_position_on_the_genomic_accession=50514690, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29948, CurrentID=0, Status='live', Symbol='OSGIN1', Aliases='BDGI, OKL38', description='oxidative stress induced growth inhibitor 1', other_designations='bone marrow stromal cell-derived growth inhibitor|ovary, kidney and liver protein 38|pregnancy-induced growth inhibitor OKL38', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83953222, end_position_on_the_genomic_accession=83966332, orientation='plus', exon_count=6, OMIM=607975, no_hdr0=''),
  29949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29949, CurrentID=0, Status='live', Symbol='IL19', Aliases='IL-10C, MDA1, NG.1, ZMDA1', description='interleukin 19', other_designations='melanoma differentiation associated protein-like protein|melanoma differentiation-associated protein-like protein', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206798870, end_position_on_the_genomic_accession=206842981, orientation='plus', exon_count=8, OMIM=605687, no_hdr0=''),
  29950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29950, CurrentID=0, Status='live', Symbol='SERTAD1', Aliases='SEI1, TRIP-Br1', description='SERTA domain containing 1', other_designations='CDK4-binding protein p34SEI|CDK4-binding protein p34SEI1|SEI-1|transcriptional regulator interacting with the PHD-bromodomain 1', map_location='19q13.1-q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40422501, end_position_on_the_genomic_accession=40426025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  29951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29951, CurrentID=0, Status='live', Symbol='PDZRN4', Aliases='LNX4, SAMCAP3L', description='PDZ domain containing ring finger 4', other_designations='IMAGE5767589|SEMACAP3-like protein|ligand of Numb protein X 4', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41187850, end_position_on_the_genomic_accession=41574745, orientation='plus', exon_count=11, OMIM=609730, no_hdr0=''),
  29952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29952, CurrentID=0, Status='live', Symbol='DPP7', Aliases='DPP2, DPPII, QPP', description='dipeptidyl-peptidase 7', other_designations='DPP II|carboxytripeptidase|dipeptidyl aminopeptidase II|dipeptidyl arylamidase II|dipeptidyl peptidase II|dipeptidyl-peptidase II|quiescent cell proline dipeptidase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137110540, end_position_on_the_genomic_accession=137117980, orientation='minus', exon_count=14, OMIM=610537, no_hdr0=''),
  29953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29953, CurrentID=0, Status='live', Symbol='TRHDE', Aliases='PAP-II, PGPEP2, TRH-DE', description='thyrotropin-releasing hormone degrading enzyme', other_designations='TRH-degrading ectoenzyme|TRH-specific aminopeptidase|pyroglutamyl aminopeptidase II|pyroglutamyl-peptidase II|thyroliberinase|thyrotropin-releasing hormone degrading ectoenzyme', map_location='12q15-q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=72272580, end_position_on_the_genomic_accession=72665644, orientation='plus', exon_count=22, OMIM=606950, no_hdr0=''),
  29954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29954, CurrentID=0, Status='live', Symbol='POMT2', Aliases='LGMD2N, MDDGA2, MDDGB2, MDDGC2', description='protein-O-mannosyltransferase 2', other_designations='dolichyl-phosphate-mannose--protein mannosyltransferase 2', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77274956, end_position_on_the_genomic_accession=77320884, orientation='minus', exon_count=24, OMIM=607439, no_hdr0=''),
  441899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441899, CurrentID=0, Status='live', Symbol='VPS25P1', Aliases='', description='VPS25 pseudogene 1', other_designations='vacuolar protein sorting 25 homolog pseudogene 1', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117549415, end_position_on_the_genomic_accession=117549915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29956, CurrentID=0, Status='live', Symbol='CERS2', Aliases='L3, LASS2, SP260, TMSG1', description='ceramide synthase 2', other_designations='LAG1 homolog, ceramide synthase 2|LAG1 longevity assurance 2|longevity assurance (LAG1, S. cerevisiae) homolog 2|tumor metastasis-suppressor gene 1 protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150965173, end_position_on_the_genomic_accession=150975003, orientation='minus', exon_count=13, OMIM=606920, no_hdr0=''),
  29957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29957, CurrentID=0, Status='live', Symbol='SLC25A24', Aliases='APC1, SCAMC-1', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 24', other_designations='calcium-binding transporter|mitochondrial ATP-Mg/Pi carrier protein 1|mitochondrial Ca(2+)-dependent solute carrier protein 1|short calcium-binding mitochondrial carrier 1|small calcium-binding mitochondrial carrier protein 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108134722, end_position_on_the_genomic_accession=108200358, orientation='minus', exon_count=11, OMIM=608744, no_hdr0=''),
  29958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29958, CurrentID=0, Status='live', Symbol='DMGDH', Aliases='DMGDHD, ME2GLYDH', description='dimethylglycine dehydrogenase', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78997564, end_position_on_the_genomic_accession=79069674, orientation='minus', exon_count=19, OMIM=605849, no_hdr0=''),
  29959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29959, CurrentID=0, Status='live', Symbol='NRBP1', Aliases='BCON3, MADM, MUDPNP, NRBP', description='nuclear receptor binding protein 1', other_designations='multiple domain putative nuclear protein|myeloid leukemia factor 1 adaptor molecule', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27428606, end_position_on_the_genomic_accession=27442259, orientation='plus', exon_count=19, OMIM=606010, no_hdr0=''),
  29960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29960, CurrentID=0, Status='live', Symbol='FTSJ2', Aliases='FJH1, HEL97, MRM2, RRMJ2', description='FtsJ RNA methyltransferase homolog 2 (E. coli)', other_designations="16S rRNA (uridine(1369)-2'-O)-methyltransferase|16S rRNA [Um1369] 2'-O-methyltransferase|FtsJ homolog 2|MRM2 RNA methyltransferase homolog|cell division protein FtsJ|epididymis luminal protein 97|protein ftsJ homolog 2|putative ribosomal RNA methyltransferase 2|rRNA (uridine-2'-O-)-methyltransferase", map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2234291, end_position_on_the_genomic_accession=2242198, orientation='minus', exon_count=3, OMIM=606906, no_hdr0=''),
  106479148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479148, CurrentID=0, Status='live', Symbol='RN7SKP122', Aliases='', description='RNA, 7SK small nuclear pseudogene 122', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554250, CurrentID=0, Status='live', Symbol='GDF5OS', Aliases='', description='growth differentiation factor 5 opposite strand', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=34020829, end_position_on_the_genomic_accession=34023278, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  554251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554251, CurrentID=0, Status='live', Symbol='FBXO48', Aliases='', description='F-box protein 48', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68462366, end_position_on_the_genomic_accession=68467316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  29964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29964, CurrentID=0, Status='live', Symbol='PRICKLE4', Aliases='C6orf49, OBTP, OEBT, TOMM6', description='prickle homolog 4', other_designations='over-expressed breast tumor protein|overexpressed breast tumor protein|overexpressed breast tumour protein|translocase of outer mitochondrial membrane 6 homolog', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41780762, end_position_on_the_genomic_accession=41787372, orientation='plus', exon_count=8, OMIM=611389, no_hdr0=''),
  29965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29965, CurrentID=0, Status='live', Symbol='CDIP1', Aliases='C16orf5, CDIP, I1, LITAFL', description='cell death-inducing p53 target 1', other_designations='LITAF-like protein|cell death inducing protein|cell death involved p53-target|cell death-inducing protein|lipopolysaccharide-induced tumor necrosis factor-alpha-like protein|transmembrane protein I1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4510675, end_position_on_the_genomic_accession=4538815, orientation='minus', exon_count=8, OMIM=610503, no_hdr0=''),
  29966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29966, CurrentID=0, Status='live', Symbol='STRN3', Aliases='SG2NA', description='striatin 3', other_designations='cell cycle S/G2 nuclear autoantigen|cell cycle autoantigen SG2NA|nuclear autoantigen|s/G2 antigen|striatin, calmodulin binding protein 3', map_location='14q13-q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30893779, end_position_on_the_genomic_accession=31026528, orientation='minus', exon_count=19, OMIM=614766, no_hdr0=''),
  29967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29967, CurrentID=0, Status='live', Symbol='LRP12', Aliases='ST7', description='LDL receptor related protein 12', other_designations='low density lipoprotein receptor-related protein 12|suppressor of tumorigenicity 7 protein', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104489231, end_position_on_the_genomic_accession=104589024, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  29968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29968, CurrentID=0, Status='live', Symbol='PSAT1', Aliases='EPIP, NLS2, PSA, PSAT, PSATD', description='phosphoserine aminotransferase 1', other_designations='endometrial progesterone-induced protein|phosphohydroxythreonine aminotransferase', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78297075, end_position_on_the_genomic_accession=78330093, orientation='plus', exon_count=9, OMIM=610936, no_hdr0=''),
  29969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29969, CurrentID=0, Status='live', Symbol='MDFIC', Aliases='HIC', description='MyoD family inhibitor domain containing', other_designations='I-mfa domain-containing protein', map_location='7q31.1-q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=114922154, end_position_on_the_genomic_accession=115019916, orientation='plus', exon_count=6, OMIM=614511, no_hdr0=''),
  29970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29970, CurrentID=0, Status='live', Symbol='SCHIP1', Aliases='SCHIP-1', description='schwannomin interacting protein 1', other_designations='', map_location='3q25.32-q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159273247, end_position_on_the_genomic_accession=159897366, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100271487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271487, CurrentID=0, Status='live', Symbol='RPL31P61', Aliases='RPL31_34_1652', description='ribosomal protein L31 pseudogene 61', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36901719, end_position_on_the_genomic_accession=36902158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29974, CurrentID=0, Status='live', Symbol='A1CF', Aliases='ACF, ACF64, ACF65, APOBEC1CF, ASP', description='APOBEC1 complementation factor', other_designations='APOBEC-1 stimulating protein|apo-B RNA editing protein|apobec-1 complementation factor (ACF) (ASP)', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50799409, end_position_on_the_genomic_accession=50885675, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  102659353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102659353, CurrentID=0, Status='live', Symbol='THRIL', Aliases='BRI3BP-AS1, BRI3BPAS1, Linc1992, TCONS_00020260', description='TNF and HNRNPL related immunoregulatory long non-coding RNA', other_designations='BRI3BP antisense RNA 1|TNFalpha and hnRNPL related immunoregulatory LincRNA', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125025443, end_position_on_the_genomic_accession=125027423, orientation='minus', exon_count=1, OMIM=615622, no_hdr0=''),
  29978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29978, CurrentID=0, Status='live', Symbol='UBQLN2', Aliases='ALS15, CHAP1, DSK2, HRIHFB2157, N4BP4, PLIC2', description='ubiquilin 2', other_designations='Nedd4 binding protein 4|protein linking IAP with cytoskeleton 2|ubiquitin-like product Chap1/Dsk2', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56563593, end_position_on_the_genomic_accession=56567010, orientation='plus', exon_count=1, OMIM=300264, no_hdr0=''),
  29979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29979, CurrentID=0, Status='live', Symbol='UBQLN1', Aliases='DA41, DSK2, PLIC-1, UBQN, XDRP1', description='ubiquilin 1', other_designations='hPLIC-1|protein linking IAP with cytoskeleton 1', map_location='9q21.2-q21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83659963, end_position_on_the_genomic_accession=83708253, orientation='minus', exon_count=12, OMIM=605046, no_hdr0=''),
  29980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29980, CurrentID=0, Status='live', Symbol='DONSON', Aliases='B17, C21orf60', description='downstream neighbor of SON', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33577551, end_position_on_the_genomic_accession=33588708, orientation='minus', exon_count=10, OMIM=611428, no_hdr0=''),
  29982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29982, CurrentID=0, Status='live', Symbol='NRBF2', Aliases='COPR, COPR1, COPR2, NRBF-2', description='nuclear receptor binding factor 2', other_designations='comodulator of PPAR and RXR 1|comodulator of PPAR and RXR 2|nuclear receptor binding factor-2', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63133247, end_position_on_the_genomic_accession=63155031, orientation='plus', exon_count=5, OMIM=616477, no_hdr0=''),
  29984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29984, CurrentID=0, Status='live', Symbol='RHOD', Aliases='ARHD, RHOHP1, RHOM, Rho', description='ras homolog family member D', other_designations='Rho-related protein HP1|ras homolog D|ras homolog gene family, member A|ras homolog gene family, member D', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67056818, end_position_on_the_genomic_accession=67072017, orientation='plus', exon_count=5, OMIM=605781, no_hdr0=''),
  29985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29985, CurrentID=0, Status='live', Symbol='SLC39A3', Aliases='ZIP-3, ZIP3', description='solute carrier family 39 (zinc transporter), member 3', other_designations='solute carrier family 39 member 3|zrt- and Irt-like protein 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2732524, end_position_on_the_genomic_accession=2740076, orientation='minus', exon_count=3, OMIM=612168, no_hdr0=''),
  29986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29986, CurrentID=0, Status='live', Symbol='SLC39A2', Aliases='6A1, ETI-1, ZIP-2, ZIP2', description='solute carrier family 39 (zinc transporter), member 2', other_designations='Zrt- and Irt-like protein 2|solute carrier family 39 member 2|zinc uptake transporter 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20999255, end_position_on_the_genomic_accession=21001875, orientation='plus', exon_count=4, OMIM=612166, no_hdr0=''),
  29988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29988, CurrentID=0, Status='live', Symbol='SLC2A8', Aliases='GLUT8, GLUTX1', description='solute carrier family 2 (facilitated glucose transporter), member 8', other_designations='glucose transporter type 8|glucose transporter type X1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127397138, end_position_on_the_genomic_accession=127407898, orientation='plus', exon_count=11, OMIM=605245, no_hdr0=''),
  29989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29989, CurrentID=0, Status='live', Symbol='OBP2B', Aliases='LCN14, OBPIIb', description='odorant binding protein 2B', other_designations='odorant-binding protein IIb', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133205279, end_position_on_the_genomic_accession=133253397, orientation='minus', exon_count=12, OMIM=604606, no_hdr0=''),
  29990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29990, CurrentID=0, Status='live', Symbol='PILRB', Aliases='FDFACT1, FDFACT2', description='paired immunoglobin-like type 2 receptor beta', other_designations='activating receptor PILR-beta|activating receptor PILRbeta|cell surface receptor FDFACT|cell surface receptor FDFACT1|cell surface receptor FDFACT2|paired immunoglobin-like receptor beta|paired immunoglobulin-like receptor beta', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100358003, end_position_on_the_genomic_accession=100367831, orientation='plus', exon_count=4, OMIM=605342, no_hdr0=''),
  29991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29991, CurrentID=0, Status='live', Symbol='OBP2A', Aliases='LCN13, OBP, OBP2C, OBPIIa, hOBPIIa', description='odorant binding protein 2A', other_designations='odorant-binding protein IIa|putative odorant-binding protein 2c', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135546139, end_position_on_the_genomic_accession=135549969, orientation='plus', exon_count=7, OMIM=164320, no_hdr0=''),
  29992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29992, CurrentID=0, Status='live', Symbol='PILRA', Aliases='FDF03', description='paired immunoglobin-like type 2 receptor alpha', other_designations='cell surface receptor FDF03|inhibitory receptor PILR-alpha', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100373445, end_position_on_the_genomic_accession=100400099, orientation='plus', exon_count=7, OMIM=605341, no_hdr0=''),
  29993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29993, CurrentID=0, Status='live', Symbol='PACSIN1', Aliases='SDPI', description='protein kinase C and casein kinase substrate in neurons 1', other_designations='syndapin 1|syndapin I|syndapin-1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34466061, end_position_on_the_genomic_accession=34536262, orientation='plus', exon_count=12, OMIM=606512, no_hdr0=''),
  554282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554282, CurrentID=0, Status='live', Symbol='FAM72C', Aliases='', description='family with sequence similarity 72 member C', other_designations='RP5-998N21.9|family with sequence similarity 72, member C', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143955364, end_position_on_the_genomic_accession=143974205, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  29995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29995, CurrentID=0, Status='live', Symbol='LMCD1', Aliases='', description='LIM and cysteine-rich domains 1', other_designations='dyxin', map_location='3p26-p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8501807, end_position_on_the_genomic_accession=8568125, orientation='plus', exon_count=8, OMIM=604859, no_hdr0=''),
  29997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29997, CurrentID=0, Status='live', Symbol='GLTSCR2', Aliases='PICT-1, PICT1', description='glioma tumor suppressor candidate region gene 2', other_designations='p60|protein interacting with carboxyl terminus 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47745536, end_position_on_the_genomic_accession=47757066, orientation='plus', exon_count=13, OMIM=605691, no_hdr0=''),
  29998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29998, CurrentID=0, Status='live', Symbol='GLTSCR1', Aliases='', description='glioma tumor suppressor candidate region gene 1', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47608196, end_position_on_the_genomic_accession=47703277, orientation='plus', exon_count=17, OMIM=605690, no_hdr0=''),
  29999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29999, CurrentID=0, Status='live', Symbol='FSCN3', Aliases='', description='fascin actin-bundling protein 3', other_designations='fascin actin-bundling protein 3, testicular|fascin homolog 3, actin-bundling protein, testicular|testis fascin', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127593635, end_position_on_the_genomic_accession=127601797, orientation='plus', exon_count=7, OMIM=615800, no_hdr0=''),
  30000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30000, CurrentID=0, Status='live', Symbol='TNPO2', Aliases='IPO3, KPNB2B, TRN2', description='transportin 2', other_designations='importin 3|karyopherin beta 2b, transportin|karyopherin beta-2b|transportin 2 (importin 3, karyopherin beta 2b)', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12699194, end_position_on_the_genomic_accession=12723996, orientation='minus', exon_count=28, OMIM=603002, no_hdr0=''),
  30001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30001, CurrentID=0, Status='live', Symbol='ERO1A', Aliases='ERO1-L, ERO1-L-alpha, ERO1-alpha, ERO1L, ERO1LA, Ero1alpha', description='endoplasmic reticulum oxidoreductase alpha', other_designations='endoplasmic oxidoreductin-1-like protein|oxidoreductin-1-L-alpha', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52639915, end_position_on_the_genomic_accession=52695931, orientation='minus', exon_count=16, OMIM=615435, no_hdr0=''),
  100271488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271488, CurrentID=0, Status='live', Symbol='RPL32P32', Aliases='RPL32_10_1563', description='ribosomal protein L32 pseudogene 32', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60245304, end_position_on_the_genomic_accession=60245647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479155, CurrentID=0, Status='live', Symbol='RN7SKP136', Aliases='', description='RNA, 7SK small nuclear pseudogene 136', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  30008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30008, CurrentID=0, Status='live', Symbol='EFEMP2', Aliases='ARCL1B, FBLN4, MBP1, UPH1', description='EGF containing fibulin-like extracellular matrix protein 2', other_designations='FIBL-4|fibulin 4|fibulin-4|mutant p53 binding protein 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65866441, end_position_on_the_genomic_accession=65872934, orientation='minus', exon_count=12, OMIM=604633, no_hdr0=''),
  30009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30009, CurrentID=0, Status='live', Symbol='TBX21', Aliases='T-PET, T-bet, TBET, TBLYM', description='T-box 21', other_designations='T-box expressed in T cells|T-box protein 21|T-cell-specific T-box transcription factor T-bet|transcription factor TBLYM', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47733244, end_position_on_the_genomic_accession=47746122, orientation='plus', exon_count=8, OMIM=604895, no_hdr0=''),
  30010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30010, CurrentID=0, Status='live', Symbol='NXPH1', Aliases='NPH1, Nbla00697', description='neurexophilin 1', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8433955, end_position_on_the_genomic_accession=8752963, orientation='plus', exon_count=4, OMIM=604639, no_hdr0=''),
  30011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30011, CurrentID=0, Status='live', Symbol='SH3KBP1', Aliases='CD2BP3, CIN85, GIG10, HSB-1, HSB1, MIG18', description='SH3-domain kinase binding protein 1', other_designations='CD2-binding protein 3|Src family kinase-binding protein 1|c-Cbl-interacting protein|cbl-interacting protein of 85 kDa|human Src family kinase-binding protein 1|migration-inducing gene 18|src-related kinase binding protein-1', map_location='Xp22.1-p21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19533965, end_position_on_the_genomic_accession=19887626, orientation='minus', exon_count=25, OMIM=300374, no_hdr0=''),
  554300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554300, CurrentID=0, Status='live', Symbol='KIR2DP1', Aliases='KIR15, KIR2DL6, KIRY, KIRZ', description='killer cell immunoglobulin-like receptor, two domains, pseudogene 1', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54756012, end_position_on_the_genomic_accession=54760768, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  30014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30014, CurrentID=0, Status='live', Symbol='SPANXA1', Aliases='CT11.1, NAP-X, SPAN-X, SPAN-Xa, SPAN-Xb, SPANX, SPANX-A', description='sperm protein associated with the nucleus, X-linked, family member A1', other_designations='SPANX family member A|SPANX family, member A1|cancer/testis antigen 11.1|cancer/testis antigen family 11, member 1|nuclear-associated protein SPAN-Xa|sperm protein associated with the nucleus, X chromosome, family member A1', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141583674, end_position_on_the_genomic_accession=141584738, orientation='minus', exon_count=2, OMIM=300305, no_hdr0=''),
  100431168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431168, CurrentID=0, Status='live', Symbol='MYL6P2', Aliases='', description='myosin light chain 6 pseudogene 2', other_designations='myosin, light chain 6 pseudogene 2|myosin, light chain 6, alkali, smooth muscle and non-muscle pseudogene', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39488309, end_position_on_the_genomic_accession=39488955, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100431172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431172, CurrentID=0, Status='live', Symbol='KLRF2', Aliases='NKp65', description='killer cell lectin-like receptor subfamily F, member 2', other_designations='activating coreceptor NKp65|lectin-like receptor F2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9881489, end_position_on_the_genomic_accession=9895833, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100431173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431173, CurrentID=0, Status='live', Symbol='RNY4P16', Aliases='', description='RNA, Ro-associated Y4 pseudogene 16', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234837973, end_position_on_the_genomic_accession=234838070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100431174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431174, CurrentID=0, Status='live', Symbol='LOC100431174', Aliases='', description='clathrin, light chain A pseudogene', other_designations='clathrin, light chain (Lca) pseudogene', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=15688633, end_position_on_the_genomic_accession=15689638, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100431175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431175, CurrentID=0, Status='live', Symbol='LOC100431175', Aliases='', description='apolipoprotein O-like pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145997217, end_position_on_the_genomic_accession=145998374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  554313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554313, CurrentID=0, Status='live', Symbol='HIST2H4B', Aliases='H4/o', description='histone cluster 2, H4b', other_designations='histone 2, H4b', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149860764, end_position_on_the_genomic_accession=149861159, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  554314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554314, CurrentID=0, Status='live', Symbol='AUTS6', Aliases='', description='Autism, susceptibility to, 6', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609378, no_hdr0=''),
  554315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554315, CurrentID=0, Status='live', Symbol='BMND5', Aliases='', description='Bone mineral density variation QTL 5', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609354, no_hdr0=''),
  554316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554316, CurrentID=0, Status='live', Symbol='BMND4', Aliases='', description='Bone mineral density variation QTL 4', other_designations='', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300536, no_hdr0=''),
  100271489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271489, CurrentID=0, Status='live', Symbol='RPL32P24', Aliases='RPL32_15_1162', description='ribosomal protein L32 pseudogene 24', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83213005, end_position_on_the_genomic_accession=83213404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369249, CurrentID=0, Status='live', Symbol='LOC105369249', Aliases='', description='uncharacterized LOC105369249', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40214213, end_position_on_the_genomic_accession=40217916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441914, CurrentID=0, Status='live', Symbol='LOC441914', Aliases='', description='dynactin subunit 5 pseudogene', other_designations='dynactin 5 (p25) pseudogene', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168369054, end_position_on_the_genomic_accession=168369598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479163, CurrentID=0, Status='live', Symbol='RN7SKP152', Aliases='', description='RNA, 7SK small nuclear pseudogene 152', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26855, CurrentID=0, Status='live', Symbol='RNU2-2P', Aliases='RNU2-2, RNU2B, U2', description='RNA, U2 small nuclear 2, pseudogene', other_designations='RNA, U2B small nuclear', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62841624, end_position_on_the_genomic_accession=62841761, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  30061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30061, CurrentID=0, Status='live', Symbol='SLC40A1', Aliases='FPN1, HFE4, IREG1, MST079, MSTP079, MTP1, SLC11A3', description='solute carrier family 40 (iron-regulated transporter), member 1', other_designations='iron regulated gene 1|putative ferroportin 1 variant IIIA|putative ferroportin 1 variant IIIB|solute carrier family 11 (proton-coupled divalent metal ion transporters), member 3', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189560590, end_position_on_the_genomic_accession=189580811, orientation='minus', exon_count=9, OMIM=604653, no_hdr0=''),
  30062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30062, CurrentID=0, Status='live', Symbol='RAX', Aliases='MCOP3, RX', description='retina and anterior neural fold homeobox', other_designations='retina and anterior neural fold homeobox protein', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59267035, end_position_on_the_genomic_accession=59273393, orientation='minus', exon_count=3, OMIM=601881, no_hdr0=''),
  100270910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270910, CurrentID=0, Status='live', Symbol='RPS2P23', Aliases='RPS2_10_570', description='ribosomal protein S2 pseudogene 23', other_designations='', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65415333, end_position_on_the_genomic_accession=65416257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  554354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554354, CurrentID=0, Status='live', Symbol='LARP4P', Aliases='', description='La ribonucleoprotein domain family member 4 pseudogene', other_designations='La ribonucleoprotein domain family, member 4 pseudogene', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36617012, end_position_on_the_genomic_accession=36618423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479166, CurrentID=0, Status='live', Symbol='RN7SKP160', Aliases='', description='RNA, 7SK small nuclear pseudogene 160', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161142, CurrentID=0, Status='live', Symbol='FAM71D', Aliases='C14orf54', description='family with sequence similarity 71 member D', other_designations='family with sequence similarity 71, member D', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67189393, end_position_on_the_genomic_accession=67228550, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  161145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161145, CurrentID=0, Status='live', Symbol='TMEM229B', Aliases='C14orf83', description='transmembrane protein 229B', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67470266, end_position_on_the_genomic_accession=67533739, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106479167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479167, CurrentID=0, Status='live', Symbol='RN7SKP161', Aliases='', description='RNA, 7SK small nuclear pseudogene 161', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  157931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157931, CurrentID=0, Status='live', Symbol='LOC157931', Aliases='', description='uncharacterized LOC157931', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161159, CurrentID=0, Status='live', Symbol='ACTN1-AS1', Aliases='C14orf84', description='ACTN1 antisense RNA 1', other_designations='ACTN1 antisense RNA 1 (non-protein coding)', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68979682, end_position_on_the_genomic_accession=68987463, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271491, CurrentID=0, Status='live', Symbol='RPL32P26', Aliases='RPL32_17_1245', description='ribosomal protein L32 pseudogene 26', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62658823, end_position_on_the_genomic_accession=62659225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  594841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594841, CurrentID=0, Status='live', Symbol='WDR45P', Aliases='', description='WD repeat domain 45 pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153658466, end_position_on_the_genomic_accession=153660012, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161176, CurrentID=0, Status='live', Symbol='SYNE3', Aliases='C14orf49, NET53, Nesp3', description='spectrin repeat containing, nuclear envelope family member 3', other_designations='nuclear envelope spectrin repeat protein 3', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95416083, end_position_on_the_genomic_accession=95516668, orientation='minus', exon_count=23, OMIM=610861, no_hdr0=''),
  106479172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479172, CurrentID=0, Status='live', Symbol='RN7SKP174', Aliases='', description='RNA, 7SK small nuclear pseudogene 174', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  201626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201626, CurrentID=0, Status='live', Symbol='PDE12', Aliases="2'-PDE", description='phosphodiesterase 12', other_designations="2'-phosphodiesterase|2-PDE|mitochondrial deadenylase", map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57556252, end_position_on_the_genomic_accession=57654979, orientation='plus', exon_count=4, OMIM=616519, no_hdr0=''),
  441925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441925, CurrentID=0, Status='live', Symbol='BECN2', Aliases='BECN1L1, BECN1P1', description='beclin 2', other_designations='Beclin-1-like protein 1|beclin 1, autophagy related, pseudogene 1|beclin-1 autophagy-related pseudogene 1', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241957767, end_position_on_the_genomic_accession=241959062, orientation='plus', exon_count=1, OMIM=615687, no_hdr0=''),
  101929951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929951, CurrentID=0, Status='live', Symbol='LOC101929951', Aliases='', description='uncharacterized LOC101929951', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31693082, end_position_on_the_genomic_accession=31707630, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  441887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441887, CurrentID=0, Status='live', Symbol='LOC441887', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 7 pseudogene', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63359511, end_position_on_the_genomic_accession=63361181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271492, CurrentID=0, Status='live', Symbol='RPL32P30', Aliases='RPL32_19_1431', description='ribosomal protein L32 pseudogene 30', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50798670, end_position_on_the_genomic_accession=50798957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161198, CurrentID=0, Status='live', Symbol='CLEC14A', Aliases='C14orf27, CEG1, EGFR-5', description='C-type lectin domain family 14 member A', other_designations='C-type lectin domain family 14, member A|ClECT and EGF-like domain containing protein|epidermal growth factor receptor 5', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38254000, end_position_on_the_genomic_accession=38256370, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479179, CurrentID=0, Status='live', Symbol='RN7SKP188', Aliases='', description='RNA, 7SK small nuclear pseudogene 188', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271493, CurrentID=0, Status='live', Symbol='RPL32P33', Aliases='RPL32_20_1573', description='ribosomal protein L32 pseudogene 33', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72627231, end_position_on_the_genomic_accession=72627624, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441932, CurrentID=0, Status='live', Symbol='OR2W5', Aliases='OR2W5P, OST722', description='olfactory receptor family 2 subfamily W member 5 (gene/pseudogene)', other_designations='olfactory receptor 2W5|olfactory receptor, family 2, subfamily W, member 5 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247491068, end_position_on_the_genomic_accession=247492409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479181, CurrentID=0, Status='live', Symbol='RN7SKP193', Aliases='', description='RNA, 7SK small nuclear pseudogene 193', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161247, CurrentID=0, Status='live', Symbol='FITM1', Aliases='FIT1', description='fat storage-inducing transmembrane protein 1', other_designations='fat-inducing protein 1|fat-inducing transcript 1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24131466, end_position_on_the_genomic_accession=24132849, orientation='plus', exon_count=2, OMIM=612028, no_hdr0=''),
  161253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161253, CurrentID=0, Status='live', Symbol='REM2', Aliases='', description='RAS (RAD and GEM)-like GTP binding 2', other_designations='RAS-like GTP binding 2|Rad and Gem-related 2 (rat homolog)|rad and Gem-like GTP-binding protein 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22883165, end_position_on_the_genomic_accession=22887680, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  161271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161271, CurrentID=0, Status='live', Symbol='OR4H12P', Aliases='C14orf14, OR4H12, c14_5009', description='olfactory receptor family 4 subfamily H member 12 pseudogene', other_designations='olfactory receptor OR14-4|olfactory receptor, family 4, subfamily H, member 12 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19759910, end_position_on_the_genomic_accession=19760788, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441940, CurrentID=0, Status='live', Symbol='LIN28AP3', Aliases='', description='LIN28A pseudogene 3', other_designations='lin-28 homolog A pseudogene 3', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11608450, end_position_on_the_genomic_accession=11609050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642590, CurrentID=0, Status='live', Symbol='LOC642590', Aliases='', description='spermine synthase pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70607758, end_position_on_the_genomic_accession=70609425, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271495, CurrentID=0, Status='live', Symbol='RPL34P21', Aliases='RPL34_10_1189', description='ribosomal protein L34 pseudogene 21', other_designations='', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130082495, end_position_on_the_genomic_accession=130082781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270911, CurrentID=0, Status='live', Symbol='RPS2P30', Aliases='RPS2_10_800', description='ribosomal protein S2 pseudogene 30', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28615097, end_position_on_the_genomic_accession=28615864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  161291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161291, CurrentID=0, Status='live', Symbol='TMEM30B', Aliases='CDC50B', description='transmembrane protein 30B', other_designations='P4-ATPase flippase complex beta subunit TMEM30B', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61277370, end_position_on_the_genomic_accession=61281812, orientation='minus', exon_count=1, OMIM=611029, no_hdr0=''),
  100128355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128355, CurrentID=0, Status='live', Symbol='LOC100128355', Aliases='', description='ribosomal protein S27a pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3717109, end_position_on_the_genomic_accession=3717632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375621, CurrentID=0, Status='live', Symbol='LOC105375621', Aliases='', description='WAS/WASL-interacting protein family member 1-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157428233, end_position_on_the_genomic_accession=157441783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479199, CurrentID=0, Status='live', Symbol='RN7SKP235', Aliases='', description='RNA, 7SK small nuclear pseudogene 235', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161342, CurrentID=0, Status='live', Symbol='LINC00519', Aliases='C14orf30, c14_5317', description='long intergenic non-protein coding RNA 519', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161357, CurrentID=0, Status='live', Symbol='MDGA2', Aliases='MAMDC1, c14_5286', description='MAM domain containing glycosylphosphatidylinositol anchor 2', other_designations='MAM domain containing 1', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46839623, end_position_on_the_genomic_accession=47675334, orientation='minus', exon_count=23, OMIM=611128, no_hdr0=''),
  100913082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913082, CurrentID=0, Status='live', Symbol='LOC100913082', Aliases='', description='intraflagellar transport 20 homolog (Chlamydomonas) pseudogene', other_designations='', map_location='14q13-q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31018301, end_position_on_the_genomic_accession=31018791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479203, CurrentID=0, Status='live', Symbol='RN7SKP244', Aliases='', description='RNA, 7SK small nuclear pseudogene 244', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373622, CurrentID=0, Status='live', Symbol='LOC105373622', Aliases='', description='uncharacterized LOC105373622', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132268619, end_position_on_the_genomic_accession=132276904, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441956, CurrentID=0, Status='live', Symbol='LOC441956', Aliases='', description='uncharacterized LOC441956', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13679219, end_position_on_the_genomic_accession=13681132, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378598, CurrentID=0, Status='live', Symbol='LOC105378598', Aliases='', description='uncharacterized LOC105378598', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2763419, end_position_on_the_genomic_accession=2784733, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100271498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271498, CurrentID=0, Status='live', Symbol='RPL34P22', Aliases='RPL34_12_1141', description='ribosomal protein L34 pseudogene 22', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44629357, end_position_on_the_genomic_accession=44629674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479205, CurrentID=0, Status='live', Symbol='RN7SKP247', Aliases='', description='RNA, 7SK small nuclear pseudogene 247', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441958, CurrentID=0, Status='live', Symbol='GXYLT1P2', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 2', other_designations='glycosyltransferase 8 domain containing 3 pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13823671, end_position_on_the_genomic_accession=13825654, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100038246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100038246, CurrentID=0, Status='live', Symbol='TLX1NB', Aliases='APT-B7, TD1, TDI', description='TLX1 neighbor', other_designations='TLX1 divergent gene protein|putative TLX1 neighbor protein', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101089321, end_position_on_the_genomic_accession=101141266, orientation='minus', exon_count=5, OMIM=612734, no_hdr0=''),
  100038247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100038247, CurrentID=0, Status='live', Symbol='PTLS', Aliases='', description='Potocki-Lupski syndrome', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610883, no_hdr0=''),
  100038248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100038248, CurrentID=0, Status='live', Symbol='PSNP3', Aliases='', description='Supranuclear palsy, progressive, 3', other_designations='', map_location='11p12-p11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610898, no_hdr0=''),
  101928769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928769, CurrentID=0, Status='live', Symbol='LOC101928769', Aliases='', description='uncharacterized LOC101928769', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66507544, end_position_on_the_genomic_accession=66511604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  161394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161394, CurrentID=0, Status='live', Symbol='SAMD15', Aliases='C14orf174, FAM15A', description='sterile alpha motif domain containing 15', other_designations='SAM domain-containing protein 15|SAM domain-containing protein C14orf174|family with sequence similarity 15, member A', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77377419, end_position_on_the_genomic_accession=77391244, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271499, CurrentID=0, Status='live', Symbol='RPL34P23', Aliases='RPL34_13_1184', description='ribosomal protein L34 pseudogene 23', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123035869, end_position_on_the_genomic_accession=123036214, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161406, CurrentID=0, Status='live', Symbol='RPL7AP10', Aliases='RPL7AP3, RPL7AP68, RPL7A_31_1637, rpL7a', description='ribosomal protein L7a pseudogene 10', other_designations='ribosomal protein L7a pseudogene 3|ribosomal protein L7a pseudogene 68', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21149589, end_position_on_the_genomic_accession=21150458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376366, CurrentID=0, Status='live', Symbol='LOC105376366', Aliases='', description='uncharacterized LOC105376366', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3892821, end_position_on_the_genomic_accession=3905657, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419893, CurrentID=0, Status='live', Symbol='ZNF415P1', Aliases='', description='zinc finger protein 415 pseudogene 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9785127, end_position_on_the_genomic_accession=9787866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479213, CurrentID=0, Status='live', Symbol='RN7SKP264', Aliases='', description='RNA, 7SK small nuclear pseudogene 264', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161424, CurrentID=0, Status='live', Symbol='NOP9', Aliases='C14orf21', description='NOP9 nucleolar protein', other_designations='NOP9 nucleolar protein homolog|pumilio domain-containing protein C14orf21', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24299854, end_position_on_the_genomic_accession=24309126, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100271500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271500, CurrentID=0, Status='live', Symbol='RPL34P28', Aliases='RPL34_13_1367', description='ribosomal protein L34 pseudogene 28', other_designations='', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55282296, end_position_on_the_genomic_accession=55282643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  161436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161436, CurrentID=0, Status='live', Symbol='EML5', Aliases='EMAP-2', description='echinoderm microtubule associated protein like 5', other_designations='EMAP-5', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88612431, end_position_on_the_genomic_accession=88792752, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  101928546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928546, CurrentID=0, Status='live', Symbol='LOC101928546', Aliases='', description='uncharacterized LOC101928546', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31386504, end_position_on_the_genomic_accession=31451326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479217, CurrentID=0, Status='live', Symbol='RN7SKP272', Aliases='', description='RNA, 7SK small nuclear pseudogene 272', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271501, CurrentID=0, Status='live', Symbol='RPL34P25', Aliases='RPL34_14_1217', description='ribosomal protein L34 pseudogene 25', other_designations='', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19508904, end_position_on_the_genomic_accession=19509245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  157983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=157983, CurrentID=0, Status='live', Symbol='C9orf66', Aliases='', description='chromosome 9 open reading frame 66', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=213108, end_position_on_the_genomic_accession=215893, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271502, CurrentID=0, Status='live', Symbol='RPL34P32', Aliases='RPL34_14_1586', description='ribosomal protein L34 pseudogene 32', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21854645, end_position_on_the_genomic_accession=21854987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  161497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161497, CurrentID=0, Status='live', Symbol='STRC', Aliases='DFNB16', description='stereocilin', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43599243, end_position_on_the_genomic_accession=43711486, orientation='minus', exon_count=36, OMIM=606440, no_hdr0=''),
  161502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161502, CurrentID=0, Status='live', Symbol='CFAP161', Aliases='C15orf26', description='cilia and flagella associated protein 161', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81099408, end_position_on_the_genomic_accession=81149175, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  161514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161514, CurrentID=0, Status='live', Symbol='TBC1D21', Aliases='MgcRabGAP', description='TBC1 domain family member 21', other_designations='TBC1 domain family, member 21|male germ cell-specific expressed, containing a RabGAP domain', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73873575, end_position_on_the_genomic_accession=73909783, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100271503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271503, CurrentID=0, Status='live', Symbol='RPL34P29', Aliases='RPL34_15_1483', description='ribosomal protein L34 pseudogene 29', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49935026, end_position_on_the_genomic_accession=49935347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130376, CurrentID=0, Status='live', Symbol='MTRF1LP2', Aliases='', description='mitochondrial translational release factor 1 like pseudogene 2', other_designations='mitochondrial translational release factor 1-like pseudogene 2', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129728767, end_position_on_the_genomic_accession=129729086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140679, CurrentID=0, Status='live', Symbol='SLC32A1', Aliases='VGAT, VIAAT', description='solute carrier family 32 (GABA vesicular transporter), member 1', other_designations='GABA and glycine transporter|hVIAAT|solute carrier family 32 member 1|vesicular GABA transporter', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38724462, end_position_on_the_genomic_accession=38729372, orientation='plus', exon_count=2, OMIM=616440, no_hdr0=''),
  106146148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106146148, CurrentID=0, Status='live', Symbol='LINC01297', Aliases='', description='long intergenic non-protein coding RNA 1297', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288962, CurrentID=0, Status='live', Symbol='LOC100288962', Aliases='', description='S-phase kinase-associated protein 1 pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25622777, end_position_on_the_genomic_accession=25623595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271504, CurrentID=0, Status='live', Symbol='RPL34P35', Aliases='RPL34_16_1738', description='ribosomal protein L34 pseudogene 35', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19238443, end_position_on_the_genomic_accession=19238739, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441727, CurrentID=0, Status='live', Symbol='RPL36AP45', Aliases='RPL36A_20_1453', description='ribosomal protein L36a pseudogene 45', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75186651, end_position_on_the_genomic_accession=75186968, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479239, CurrentID=0, Status='live', Symbol='RN7SL39P', Aliases='', description='RNA, 7SL, cytoplasmic 39, pseudogene', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161582, CurrentID=0, Status='live', Symbol='DYX1C1', Aliases='CILD25, DNAAF4, DYX1, DYXC1, EKN1, RD', description='dyslexia susceptibility 1 candidate 1', other_designations='dynein, axonemal, assembly factor 4', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55417755, end_position_on_the_genomic_accession=55508234, orientation='minus', exon_count=11, OMIM=608706, no_hdr0=''),
  105369265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369265, CurrentID=0, Status='live', Symbol='LOC105369265', Aliases='', description='calponin-2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373630, CurrentID=0, Status='live', Symbol='LOC105373630', Aliases='', description='uncharacterized LOC105373630', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134503143, end_position_on_the_genomic_accession=134507488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  441996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441996, CurrentID=0, Status='live', Symbol='LOC441996', Aliases='', description='aconitase 2, mitochondrial pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34588754, end_position_on_the_genomic_accession=34591285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288964, CurrentID=0, Status='live', Symbol='FABP5P5', Aliases='FABP5L5', description='fatty acid binding protein 5 pseudogene 5', other_designations='fatty acid binding protein 5-like 5', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95972967, end_position_on_the_genomic_accession=95973454, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369266, CurrentID=0, Status='live', Symbol='LOC105369266', Aliases='', description='uncharacterized LOC105369266', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33329444, end_position_on_the_genomic_accession=33331189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  161635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161635, CurrentID=0, Status='live', Symbol='CSNK1A1P1', Aliases='CSNK1A1P', description='casein kinase 1, alpha 1 pseudogene 1', other_designations='casein kinase I alpha-like pseudogene', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36799100, end_position_on_the_genomic_accession=36818506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271507, CurrentID=0, Status='live', Symbol='RPL35AP22', Aliases='RPL35A_9_1011', description='ribosomal protein L35a pseudogene 22', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117685258, end_position_on_the_genomic_accession=117685585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873811, CurrentID=0, Status='live', Symbol='ZNF877P', Aliases='', description='zinc finger protein 877, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21266063, end_position_on_the_genomic_accession=21266424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442006, CurrentID=0, Status='live', Symbol='NPM1P48', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 48', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=4514119, end_position_on_the_genomic_accession=4515152, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271508, CurrentID=0, Status='live', Symbol='RPL36AP24', Aliases='RPL36A_10_733', description='ribosomal protein L36a pseudogene 24', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108318052, end_position_on_the_genomic_accession=108318454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873812, CurrentID=0, Status='live', Symbol='CFL1P8', Aliases='', description='cofilin 1 (non-muscle) pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99582708, end_position_on_the_genomic_accession=99583166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369268, CurrentID=0, Status='live', Symbol='LOC105369268', Aliases='', description='uncharacterized LOC105369268', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271509, CurrentID=0, Status='live', Symbol='RPL36AP28', Aliases='RPL36A_10_864', description='ribosomal protein L36a pseudogene 28', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152602154, end_position_on_the_genomic_accession=152602515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479261, CurrentID=0, Status='live', Symbol='RN7SL106P', Aliases='', description='RNA, 7SL, cytoplasmic 106, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369269, CurrentID=0, Status='live', Symbol='LOC105369269', Aliases='', description='poly(ADP-ribose) glycohydrolase-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161725, CurrentID=0, Status='live', Symbol='OTUD7A', Aliases='C15orf16, C16ORF15, CEZANNE2, OTUD7', description='OTU deubiquitinase 7A', other_designations='OTU domain containing 7A|OTU domain-containing 7A|cezanne 2|zinc finger protein Cezanne 2', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31481682, end_position_on_the_genomic_accession=31870685, orientation='minus', exon_count=15, OMIM=612024, no_hdr0=''),
  100271510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271510, CurrentID=0, Status='live', Symbol='RPL36AP29', Aliases='RPL36A_11_788', description='ribosomal protein L36a pseudogene 29', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16208948, end_position_on_the_genomic_accession=16209268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442017, CurrentID=0, Status='live', Symbol='LOC442017', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60682684, end_position_on_the_genomic_accession=60683294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646970, CurrentID=0, Status='live', Symbol='LOC646970', Aliases='', description='chromosome 5 open reading frame 15 pseudogene', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100331800, end_position_on_the_genomic_accession=100332616, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873814, CurrentID=0, Status='live', Symbol='RNU1-17P', Aliases='', description='RNA, U1 small nuclear 17, pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180729586, end_position_on_the_genomic_accession=180729737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161742, CurrentID=0, Status='live', Symbol='SPRED1', Aliases='NFLS, PPP1R147, hSpred1, spred-1', description='sprouty-related, EVH1 domain containing 1', other_designations='protein phosphatase 1, regulatory subunit 147|suppressor of Ras/MAPK activation', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38252087, end_position_on_the_genomic_accession=38357249, orientation='plus', exon_count=11, OMIM=609291, no_hdr0=''),
  26957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26957, CurrentID=0, Status='live', Symbol='ABCD1P4', Aliases='ALD22Q11', description='ATP binding cassette subfamily D member 1 pseudogene 4', other_designations='ATP-binding cassette, sub-family D (ALD), member 1 pseudogene|ATP-binding cassette, sub-family D (ALD), member 1, pseudogene 4', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16387689, end_position_on_the_genomic_accession=16391865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161753, CurrentID=0, Status='live', Symbol='ODF3L1', Aliases='', description='outer dense fiber of sperm tails 3-like 1', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75723944, end_position_on_the_genomic_accession=75727688, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873214, CurrentID=0, Status='live', Symbol='MTND1P26', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 26', other_designations='MT-ND1 pseudogene 26', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131385700, end_position_on_the_genomic_accession=131386423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373635, CurrentID=0, Status='live', Symbol='LOC105373635', Aliases='', description='uncharacterized LOC105373635', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138101884, end_position_on_the_genomic_accession=138105905, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271511, CurrentID=0, Status='live', Symbol='RPL36AP30', Aliases='RPL36A_11_867', description='ribosomal protein L36a pseudogene 30', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157263022, end_position_on_the_genomic_accession=157263401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873815, CurrentID=0, Status='live', Symbol='RNU1-18P', Aliases='', description='RNA, U1 small nuclear 18, pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122211648, end_position_on_the_genomic_accession=122211811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  161779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161779, CurrentID=0, Status='live', Symbol='PGBD4', Aliases='', description='piggyBac transposable element derived 4', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34102073, end_position_on_the_genomic_accession=34104390, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  140687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140687, CurrentID=0, Status='live', Symbol='GCNT7', Aliases='C20orf105, dJ1153D9.2, gcnt', description='glucosaminyl (N-acetyl) transferase family member 7', other_designations='beta 1,6-N-acetylglucosaminyltransferase', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56491492, end_position_on_the_genomic_accession=56525925, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  1023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1023, CurrentID=0, Status='live', Symbol='CDL2', Aliases='', description='Cornelia de Lange syndrome 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442028, CurrentID=0, Status='live', Symbol='LOC442028', Aliases='', description='uncharacterized LOC442028', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94868685, end_position_on_the_genomic_accession=94947342, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  140688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140688, CurrentID=0, Status='live', Symbol='NOL4L', Aliases='C20orf112, C20orf113, dJ1184F4.2, dJ1184F4.4', description='nucleolar protein 4-like', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32443059, end_position_on_the_genomic_accession=32585073, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  22846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22846, CurrentID=0, Status='live', Symbol='VASH1', Aliases='KIAA1036', description='vasohibin 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76761892, end_position_on_the_genomic_accession=76783022, orientation='plus', exon_count=11, OMIM=609011, no_hdr0=''),
  100169750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169750, CurrentID=0, Status='live', Symbol='PRINS', Aliases='NCRNA00074', description='psoriasis associated non-protein coding RNA induced by stress', other_designations='psoriasis associated RNA induced by stress (non-protein coding)', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=24536051, end_position_on_the_genomic_accession=24544975, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100169751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169751, CurrentID=0, Status='live', Symbol='RNA5S1', Aliases='RN5S1', description='RNA, 5S ribosomal 1', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228610266, end_position_on_the_genomic_accession=228610386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169752, CurrentID=0, Status='live', Symbol='TEX36-AS1', Aliases='', description='TEX36 antisense RNA 1', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125574371, end_position_on_the_genomic_accession=125578445, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100169753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169753, CurrentID=0, Status='live', Symbol='RNA5S2', Aliases='RN5S2', description='RNA, 5S ribosomal 2', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228612507, end_position_on_the_genomic_accession=228612627, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169754, CurrentID=0, Status='live', Symbol='RNA5S3', Aliases='RN5S3', description='RNA, 5S ribosomal 3', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228614748, end_position_on_the_genomic_accession=228614868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169755, CurrentID=0, Status='live', Symbol='RNA5S4', Aliases='RN5S4', description='RNA, 5S ribosomal 4', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228616989, end_position_on_the_genomic_accession=228617109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169756, CurrentID=0, Status='live', Symbol='RNA5S5', Aliases='RN5S5', description='RNA, 5S ribosomal 5', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228619230, end_position_on_the_genomic_accession=228619350, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169757, CurrentID=0, Status='live', Symbol='RNA5S6', Aliases='RN5S6', description='RNA, 5S ribosomal 6', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228621445, end_position_on_the_genomic_accession=228621565, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169758, CurrentID=0, Status='live', Symbol='RNA5S7', Aliases='RN5S7', description='RNA, 5S ribosomal 7', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228623665, end_position_on_the_genomic_accession=228623785, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  161823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161823, CurrentID=0, Status='live', Symbol='ADAL', Aliases='', description='adenosine deaminase-like', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43330356, end_position_on_the_genomic_accession=43354555, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100169760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169760, CurrentID=0, Status='live', Symbol='RNA5S9', Aliases='RN5S9', description='RNA, 5S ribosomal 9', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228628146, end_position_on_the_genomic_accession=228628266, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169761, CurrentID=0, Status='live', Symbol='RNA5S10', Aliases='RN5S10', description='RNA, 5S ribosomal 10', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228630388, end_position_on_the_genomic_accession=228630508, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169762, CurrentID=0, Status='live', Symbol='RNA5S11', Aliases='RN5S11', description='RNA, 5S ribosomal 11', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228632629, end_position_on_the_genomic_accession=228632749, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169763, CurrentID=0, Status='live', Symbol='RNA5S12', Aliases='RN5S12', description='RNA, 5S ribosomal 12', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228634869, end_position_on_the_genomic_accession=228634989, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169764, CurrentID=0, Status='live', Symbol='RNA5S13', Aliases='RN5S13', description='RNA, 5S ribosomal 13', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228637094, end_position_on_the_genomic_accession=228637214, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  161829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161829, CurrentID=0, Status='live', Symbol='EXD1', Aliases='EXDL1', description="exonuclease 3'-5' domain containing 1", other_designations="exonuclease 3'-5' domain-like-containing protein 1", map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41182728, end_position_on_the_genomic_accession=41231074, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100169766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169766, CurrentID=0, Status='live', Symbol='RNA5S15', Aliases='RN5S15', description='RNA, 5S ribosomal 15', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228641566, end_position_on_the_genomic_accession=228641686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169767, CurrentID=0, Status='live', Symbol='RNA5S16', Aliases='RN5S16', description='RNA, 5S ribosomal 16', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228643807, end_position_on_the_genomic_accession=228643927, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169768, CurrentID=0, Status='live', Symbol='RNA5S17', Aliases='RN5S17', description='RNA, 5S ribosomal 17', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228646038, end_position_on_the_genomic_accession=228646158, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  161835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161835, CurrentID=0, Status='live', Symbol='FSIP1', Aliases='HSD10', description='fibrous sheath interacting protein 1', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39599529, end_position_on_the_genomic_accession=39782838, orientation='minus', exon_count=17, OMIM=615795, no_hdr0=''),
  57763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57763, CurrentID=0, Status='live', Symbol='ANKRA2', Aliases='ANKRA', description='ankyrin repeat, family A (RFXANK-like), 2', other_designations='RFXANK-like protein 2', map_location='5q12-q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73552200, end_position_on_the_genomic_accession=73565705, orientation='minus', exon_count=9, OMIM=605787, no_hdr0=''),
  341457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341457, CurrentID=0, Status='live', Symbol='PPIAP8', Aliases='PPIP8', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 8', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 8', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98590515, end_position_on_the_genomic_accession=98591247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  26974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26974, CurrentID=0, Status='live', Symbol='ZNF285', Aliases='ZNF285A', description='zinc finger protein 285', other_designations='zinc finger protein 285A', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44385646, end_position_on_the_genomic_accession=44401611, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373638, CurrentID=0, Status='live', Symbol='LOC105373638', Aliases='', description='uncharacterized LOC105373638', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138596361, end_position_on_the_genomic_accession=138601765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479284, CurrentID=0, Status='live', Symbol='RN7SL173P', Aliases='', description='RNA, 7SL, cytoplasmic 173, pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479286, CurrentID=0, Status='live', Symbol='RN7SL178P', Aliases='', description='RNA, 7SL, cytoplasmic 178, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060015, CurrentID=0, Status='live', Symbol='LOC101060015', Aliases='', description='uncharacterized LOC101060015', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  131961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131961, CurrentID=0, Status='live', Symbol='IMPDH1P8', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 8', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15877950, end_position_on_the_genomic_accession=15880295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22848, CurrentID=0, Status='live', Symbol='AAK1', Aliases='', description='AP2 associated kinase 1', other_designations='adaptor-associated kinase 1', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69457995, end_position_on_the_genomic_accession=69643845, orientation='minus', exon_count=22, OMIM=616405, no_hdr0=''),
  442041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442041, CurrentID=0, Status='live', Symbol='LOC442041', Aliases='', description='zinc finger protein 532 pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110360609, end_position_on_the_genomic_accession=110363734, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928697, CurrentID=0, Status='live', Symbol='LOC101928697', Aliases='', description='uncharacterized LOC101928697', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19182033, end_position_on_the_genomic_accession=19187465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  161882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161882, CurrentID=0, Status='live', Symbol='ZFPM1', Aliases='FOG, FOG1, ZC2HC11A, ZNF408, ZNF89A', description='zinc finger protein, FOG family member 1', other_designations='FOG-1|friend of GATA 1|friend of GATA protein 1|friend of GATA-1|zinc finger protein 89A|zinc finger protein, multitype 1', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88451431, end_position_on_the_genomic_accession=88535166, orientation='plus', exon_count=15, OMIM=601950, no_hdr0=''),
  30811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30811, CurrentID=0, Status='live', Symbol='HUNK', Aliases='', description='hormonally up-regulated Neu-associated kinase', other_designations='B19|hormonally upregulated Neu-associated kinase|hormonally upregulated neu tumor-associated kinase|serine/threonine protein kinase MAK-V|serine/threonine-protein kinase MAK-V', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31873315, end_position_on_the_genomic_accession=32004064, orientation='plus', exon_count=11, OMIM=606532, no_hdr0=''),
  30812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30812, CurrentID=0, Status='live', Symbol='SOX8', Aliases='', description='SRY-box 8', other_designations='SRY (sex determining region Y)-box 8|SRY box 8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=981808, end_position_on_the_genomic_accession=986979, orientation='plus', exon_count=3, OMIM=605923, no_hdr0=''),
  30813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30813, CurrentID=0, Status='live', Symbol='VSX1', Aliases='CAASDS, KTCN, KTCN1, PPCD, PPCD1, PPD, RINX', description='visual system homeobox 1', other_designations='homeodomain protein RINX|retinal inner nuclear layer homeobox protein|transcription factor VSX1', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25070885, end_position_on_the_genomic_accession=25082379, orientation='minus', exon_count=7, OMIM=605020, no_hdr0=''),
  30814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30814, CurrentID=0, Status='live', Symbol='PLA2G2E', Aliases='GIIE sPLA2, sPLA2-IIE', description='phospholipase A2 group IIE', other_designations='phosphatidylcholine 2-acylhydrolase 2E|phospholipase A2, group IIE', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19920307, end_position_on_the_genomic_accession=19923617, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  30815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30815, CurrentID=0, Status='live', Symbol='ST6GALNAC6', Aliases='SIAT7-F, SIAT7F, ST6GALNACVI', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 6', other_designations='CMP-NeuAC:(beta)-N-acetylgalactosaminide (alpha)2,6-sialyltransferase member VI|ST6 -N-acetylgalactosaminide alpha-26-sialyltransferase 6|galNAc alpha-2,6-sialyltransferase VI|sialyltransferase 7F|sialytransferase 7 ((alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialytransferase) F', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127885321, end_position_on_the_genomic_accession=127906731, orientation='minus', exon_count=10, OMIM=610135, no_hdr0=''),
  30816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30816, CurrentID=0, Status='live', Symbol='ERVW-1', Aliases='ENV, ENVW, ERVWE1, HERV-7q, HERV-W-ENV, HERV7Q, HERVW, HERVWENV', description='endogenous retrovirus group W member 1', other_designations='HERV-7q envelope protein|HERV-W Env glycoprotein|HERV-W envelope protein|HERV-W_7q21.2 provirus ancestral Env polyprotein|HERV-tryptophan envelope protein|endogenous retroviral family W, env(C7), member 1|endogenous retrovirus group W, member 1|envelope glycoprotein|envelope polyprotein gPr73|enverin|human endogenous retrovirus W envC7-1 envelope protein|syncytin A', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92468380, end_position_on_the_genomic_accession=92477915, orientation='minus', exon_count=2, OMIM=604659, no_hdr0=''),
  30817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30817, CurrentID=0, Status='live', Symbol='ADGRE2', Aliases='CD312, EMR2', description='adhesion G protein-coupled receptor E2', other_designations='EGF-like module-containing mucin-like hormone receptor-like 2|egf-like module containing, mucin-like, hormone receptor-like 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14724170, end_position_on_the_genomic_accession=14778541, orientation='minus', exon_count=25, OMIM=606100, no_hdr0=''),
  30818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30818, CurrentID=0, Status='live', Symbol='KCNIP3', Aliases='CSEN, DREAM, KCHIP3', description='Kv channel interacting protein 3', other_designations='A-type potassium channel modulatory protein 3|DRE-antagonist modulator|Kv channel interacting protein 3, calsenilin|calsenilin, presenilin-binding protein, EF hand transcription factor|kv channel-interacting protein 3|potassium channel interacting protein 3', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95297324, end_position_on_the_genomic_accession=95386077, orientation='plus', exon_count=11, OMIM=604662, no_hdr0=''),
  30819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30819, CurrentID=0, Status='live', Symbol='KCNIP2', Aliases='KCHIP2', description='Kv channel interacting protein 2', other_designations='A-type potassium channel modulatory protein 2|cardiac voltage-gated potassium channel modulatory subunit|potassium channel-interacting protein 2', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101825974, end_position_on_the_genomic_accession=101843920, orientation='minus', exon_count=11, OMIM=604661, no_hdr0=''),
  30820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30820, CurrentID=0, Status='live', Symbol='KCNIP1', Aliases='KCHIP1, VABP', description='Kv channel interacting protein 1', other_designations='A-type potassium channel modulatory protein 1|potassium channel interacting protein 1|vesicle APC-binding protein', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170353487, end_position_on_the_genomic_accession=170736632, orientation='plus', exon_count=14, OMIM=604660, no_hdr0=''),
  100169834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169834, CurrentID=0, Status='live', Symbol='RNY4P19', Aliases='', description='RNA, Ro-associated Y4 pseudogene 19', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230058231, end_position_on_the_genomic_accession=230058324, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  30827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30827, CurrentID=0, Status='live', Symbol='CXXC1', Aliases='2410002I16Rik, 5830420C16Rik, CFP1, CGBP, HsT2645, PCCX1, PHF18, SPP1, ZCGPC1, hCGBP', description='CXXC finger protein 1', other_designations='CXXC finger 1 (PHD domain)|CpG binding protein|DNA-binding protein with PHD finger and CXXC domain|PHD finger and CXXC domain-containing protein 1|cpG-binding protein|zinc finger, CpG binding-type containing 1', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50282343, end_position_on_the_genomic_accession=50288322, orientation='minus', exon_count=16, OMIM=609150, no_hdr0=''),
  140691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140691, CurrentID=0, Status='live', Symbol='TRIM69', Aliases='HSD-34, HSD34, RNF36, Trif', description='tripartite motif containing 69', other_designations='RFP-like domain-containing protein trimless|ring finger protein 36', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44736362, end_position_on_the_genomic_accession=44767829, orientation='plus', exon_count=7, OMIM=616017, no_hdr0=''),
  30829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30829, CurrentID=0, Status='live', Symbol='IBD3', Aliases='', description='inflammatory bowel disease 3', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604519, no_hdr0=''),
  30832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30832, CurrentID=0, Status='live', Symbol='ZNF354C', Aliases='KID3', description='zinc finger protein 354C', other_designations='KRAB-zinc finger protein synten|kidney, ischemia, and developmentally-regulated protein 3', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179060606, end_position_on_the_genomic_accession=179086298, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  158056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158056, CurrentID=0, Status='live', Symbol='MAMDC4', Aliases='AEGP', description='MAM domain containing 4', other_designations='MAM domain-containing protein 4|apical early endosomal glycoprotein|endotubin', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136852367, end_position_on_the_genomic_accession=136860799, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  30834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30834, CurrentID=0, Status='live', Symbol='ZNRD1', Aliases='HTEX-6, HTEX6, Rpa12, TCTEX6, TEX6, ZR14, hZR14, tctex-6', description='zinc ribbon domain containing 1', other_designations='RNA polymerase I small specific subunit Rpa12|transcription-associated zinc ribbon protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30061240, end_position_on_the_genomic_accession=30064909, orientation='plus', exon_count=5, OMIM=607525, no_hdr0=''),
  30835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30835, CurrentID=0, Status='live', Symbol='CD209', Aliases='CDSIGN, CLEC4L, DC-SIGN, DC-SIGN1', description='CD209 molecule', other_designations='C-type lectin domain family 4 member L|C-type lectin domain family 4, member L|HIV gpl20-binding protein|dendritic cell-specific ICAM-3-grabbing non-integrin 1|dendritic cell-specific intercellular adhesion molecule-3-grabbing non-integrin|dendritic cell-specific intracellular adhesion molecules (ICAM)-3 grabbing non-integrin', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7739988, end_position_on_the_genomic_accession=7747605, orientation='minus', exon_count=7, OMIM=604672, no_hdr0=''),
  30836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30836, CurrentID=0, Status='live', Symbol='DNTTIP2', Aliases='ERBP, FCF2, HSU15552, LPTS-RP2, TdIF2', description='deoxynucleotidyltransferase, terminal, interacting protein 2', other_designations='LPTS-interacting protein 2|acidic 82 kDa protein mRNA|estrogen receptor binding protein|tdT-interacting factor 2', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93869458, end_position_on_the_genomic_accession=93879206, orientation='minus', exon_count=7, OMIM=611199, no_hdr0=''),
  30837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30837, CurrentID=0, Status='live', Symbol='SOCS7', Aliases='NAP4, NCKAP4', description='suppressor of cytokine signaling 7', other_designations='NAP-4|NCK-associated protein 4|Nck, Ash and phospholipase C binding protein|SOCS-7|nck, Ash and phospholipase C gamma-binding protein', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38352124, end_position_on_the_genomic_accession=38405593, orientation='plus', exon_count=10, OMIM=608788, no_hdr0=''),
  100169850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169850, CurrentID=0, Status='live', Symbol='RNY4P7', Aliases='', description='RNA, Ro-associated Y4 pseudogene 7', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127798906, end_position_on_the_genomic_accession=127798998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100169851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169851, CurrentID=0, Status='live', Symbol='PATE3', Aliases='HEL-127, PATE-DJ', description='prostate and testis expressed 3', other_designations='PATE-like protein DJ|acrosomal vesicle protein HEL-127|secreted TFP/Ly-6/uPAR protein PATE-DJ', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125788111, end_position_on_the_genomic_accession=125791600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30844, CurrentID=0, Status='live', Symbol='EHD4', Aliases='PAST4', description='EH domain containing 4', other_designations='EH-domain containing 4|PAST homolog 4|hepatocellular carcinoma-associated protein 10/11|hepatocellular carcinoma-associated protein HCA11|ortholog of rat pincher', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41899440, end_position_on_the_genomic_accession=41972557, orientation='minus', exon_count=6, OMIM=605892, no_hdr0=''),
  30845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30845, CurrentID=0, Status='live', Symbol='EHD3', Aliases='PAST3', description='EH domain containing 3', other_designations='EH-domain containing 3|PAST homolog 3', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31234014, end_position_on_the_genomic_accession=31269451, orientation='plus', exon_count=7, OMIM=605891, no_hdr0=''),
  30846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30846, CurrentID=0, Status='live', Symbol='EHD2', Aliases='PAST2', description='EH domain containing 2', other_designations='EH-domain containing 2|PAST homolog 2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47713344, end_position_on_the_genomic_accession=47743134, orientation='plus', exon_count=6, OMIM=605890, no_hdr0=''),
  30847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30847, CurrentID=0, Status='live', Symbol='CECR9', Aliases='', description='cat eye syndrome chromosome region, candidate 9 (non-protein coding)', other_designations='Putative cat eye syndrome critical region protein 9', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  30848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30848, CurrentID=0, Status='live', Symbol='CTAG2', Aliases='CAMEL, CT2, CT6.2, CT6.2a, CT6.2b, ESO2, LAGE-1, LAGE2B', description='cancer/testis antigen 2', other_designations='CTL-recognized antigen on melanoma|LAGE-1a protein|autoimmunogenic cancer/testis antigen NY-ESO-2|cancer/testis antigen 6.2|cancer/testis antigen family 6, member 2a|cancer/testis antigen family 6, member 2b|l antigen family member 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154651972, end_position_on_the_genomic_accession=154653579, orientation='minus', exon_count=3, OMIM=300396, no_hdr0=''),
  30849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30849, CurrentID=0, Status='live', Symbol='PIK3R4', Aliases='VPS15, p150', description='phosphoinositide-3-kinase regulatory subunit 4', other_designations='PI3-kinase p150 subunit|PI3-kinase regulatory subunit 4|phosphatidylinositol 3-kinase-associated p150|phosphoinositide 3-kinase adaptor protein|phosphoinositide-3-kinase, regulatory subunit 4|phosphoinositide-3-kinase, regulatory subunit 4, p150', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130678934, end_position_on_the_genomic_accession=130746852, orientation='minus', exon_count=20, OMIM=602610, no_hdr0=''),
  30850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30850, CurrentID=0, Status='live', Symbol='CDR2L', Aliases='HUMPPA', description='cerebellar degeneration-related protein 2-like', other_designations='paraneoplastic 62 kDa antigen|paraneoplastic antigen', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74987632, end_position_on_the_genomic_accession=75005800, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  30851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30851, CurrentID=0, Status='live', Symbol='TAX1BP3', Aliases='TIP-1', description='Tax1 (human T-cell leukemia virus type I) binding protein 3', other_designations='Tax interaction protein 1|glutaminase-interacting protein 3|tax-interacting protein 1', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3662893, end_position_on_the_genomic_accession=3668679, orientation='minus', exon_count=4, OMIM=616484, no_hdr0=''),
  140692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140692, CurrentID=0, Status='live', Symbol='COX7BP2', Aliases='dJ901O8.2', description='cytochrome c oxidase subunit VIIb pseudogene 2', other_designations='cytochrome c oxidase subunit VIIb (COX7B) pseudogene 2', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35850501, end_position_on_the_genomic_accession=35850958, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  161931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161931, CurrentID=0, Status='live', Symbol='ADAD2', Aliases='TENRL', description='adenosine deaminase domain containing 2', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84191117, end_position_on_the_genomic_accession=84197166, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100271517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271517, CurrentID=0, Status='live', Symbol='RPL39P25', Aliases='RPL39_11_1067', description='ribosomal protein L39 pseudogene 25', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75269200, end_position_on_the_genomic_accession=75269594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100169881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169881, CurrentID=0, Status='live', Symbol='MTPN-LUZP6', Aliases='', description='myotrophin, leucine zipper protein 6 transcription unit', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=135611503, end_position_on_the_genomic_accession=135662204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100169890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169890, CurrentID=0, Status='live', Symbol='PEG3-AS1', Aliases='APEG3, NCRNA00155, PEG3-AS, PEG3AS', description='PEG3 antisense RNA 1', other_designations='PEG3 antisense RNA (non-protein coding)|PEG3 antisense RNA 1 (non-protein coding)|antisense transcript gene of PEG3', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56812480, end_position_on_the_genomic_accession=56813793, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376853, CurrentID=0, Status='live', Symbol='LOC105376853', Aliases='', description='uncharacterized LOC105376853', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14860171, end_position_on_the_genomic_accession=14863223, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26995, CurrentID=0, Status='live', Symbol='TRUB2', Aliases='CLONE24922', description='TruB pseudouridine (psi) synthase family member 2', other_designations='TruB pseudouridine (psi) synthase homolog 2', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128309117, end_position_on_the_genomic_accession=128322418, orientation='minus', exon_count=8, OMIM=610727, no_hdr0=''),
  100271518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271518, CurrentID=0, Status='live', Symbol='RPL39P26', Aliases='RPL39_12_1121', description='ribosomal protein L39 pseudogene 26', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13685305, end_position_on_the_genomic_accession=13685704, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479305, CurrentID=0, Status='live', Symbol='RN7SL229P', Aliases='', description='RNA, 7SL, cytoplasmic 229, pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442058, CurrentID=0, Status='live', Symbol='LOC442058', Aliases='', description='replication protein A3, 14kDa pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174501952, end_position_on_the_genomic_accession=174502312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442060, CurrentID=0, Status='live', Symbol='SAP18P2', Aliases='', description='SAP18 pseudogene 2', other_designations='Sin3A-associated protein, 18kDa pseudogene', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181694424, end_position_on_the_genomic_accession=181695160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373643, CurrentID=0, Status='live', Symbol='LOC105373643', Aliases='', description='uncharacterized LOC105373643', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139362263, end_position_on_the_genomic_accession=139379146, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  136227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136227, CurrentID=0, Status='live', Symbol='COL26A1', Aliases='EMI6, EMID2, EMU2, SH2B', description='collagen, type XXVI, alpha 1', other_designations='EMI domain containing 2|emilin and multimerin domain-containing protein 2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101362820, end_position_on_the_genomic_accession=101559024, orientation='plus', exon_count=13, OMIM=608927, no_hdr0=''),
  100271519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271519, CurrentID=0, Status='live', Symbol='RPL39P30', Aliases='RPL39_12_1504', description='ribosomal protein L39 pseudogene 30', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87268131, end_position_on_the_genomic_accession=87268535, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442063, CurrentID=0, Status='live', Symbol='LOC442063', Aliases='', description='family with sequence similarity 64 member A pseudogene', other_designations='family with sequence similarity 64, member A pseudogene', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202335303, end_position_on_the_genomic_accession=202336805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83956, CurrentID=0, Status='live', Symbol='RACGAP1P', Aliases='FKSG42', description='Rac GTPase activating protein 1 pseudogene', other_designations='FKSG42', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45062618, end_position_on_the_genomic_accession=45065411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169951, CurrentID=0, Status='live', Symbol='SNAR-A3', Aliases='', description='small ILF3/NF90-associated RNA A3', other_designations='small nuclear ILF3/NF90-associated RNA A3|snaR-A53118848', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47923779, end_position_on_the_genomic_accession=47923899, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100169952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169952, CurrentID=0, Status='live', Symbol='SNAR-A5', Aliases='', description='small ILF3/NF90-associated RNA A5', other_designations='small nuclear ILF3/NF90-associated RNA A5|snaR-A55296080', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50100891, end_position_on_the_genomic_accession=50101011, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169953, CurrentID=0, Status='live', Symbol='SNAR-A7', Aliases='', description='small ILF3/NF90-associated RNA A7', other_designations='small nuclear ILF3/NF90-associated RNA A7|snaR-A55302203', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50107014, end_position_on_the_genomic_accession=50107134, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169954, CurrentID=0, Status='live', Symbol='SNAR-A11', Aliases='', description='small ILF3/NF90-associated RNA A11', other_designations='small nuclear ILF3/NF90-associated RNA A11|snaR-A55323591', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50128402, end_position_on_the_genomic_accession=50128522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169955, CurrentID=0, Status='live', Symbol='SNAR-A9', Aliases='', description='small ILF3/NF90-associated RNA A9', other_designations='small nuclear ILF3/NF90-associated RNA A9|snaR-A55312909', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50117720, end_position_on_the_genomic_accession=50117840, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169956, CurrentID=0, Status='live', Symbol='SNAR-A4', Aliases='', description='small ILF3/NF90-associated RNA A4', other_designations='small nuclear ILF3/NF90-associated RNA A4|snaR-A55287678', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50092489, end_position_on_the_genomic_accession=50092609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169957, CurrentID=0, Status='live', Symbol='SNAR-A6', Aliases='', description='small ILF3/NF90-associated RNA A6', other_designations='small nuclear ILF3/NF90-associated RNA A6|snaR-A55299144', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50103955, end_position_on_the_genomic_accession=50104075, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169958, CurrentID=0, Status='live', Symbol='SNAR-A8', Aliases='', description='small ILF3/NF90-associated RNA A8', other_designations='small nuclear ILF3/NF90-associated RNA A8|snaR-A55307557', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50112368, end_position_on_the_genomic_accession=50112488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100169959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169959, CurrentID=0, Status='live', Symbol='SNAR-A13', Aliases='', description='small ILF3/NF90-associated RNA A13', other_designations='small nuclear ILF3/NF90-associated RNA A13|snaR-A53139991', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47944922, end_position_on_the_genomic_accession=47945043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100288978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288978, CurrentID=0, Status='live', Symbol='LOC100288978', Aliases='', description='myeloma overexpressed 2 pseudogene', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52798707, end_position_on_the_genomic_accession=52799115, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271520, CurrentID=0, Status='live', Symbol='RPL39P28', Aliases='RPL39_14_1253', description='ribosomal protein L39 pseudogene 28', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68001394, end_position_on_the_genomic_accession=68001548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158078, CurrentID=0, Status='live', Symbol='EEF1A1P5', Aliases='EEF1AL3', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 5', other_designations='eukaryotic translation elongation factor 1 alpha-like 3', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133019435, end_position_on_the_genomic_accession=133021175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  30968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30968, CurrentID=0, Status='live', Symbol='STOML2', Aliases='HSPC108, SLP-2', description='stomatin like 2', other_designations='EPB72-like protein 2|paraprotein target 7|paratarg-7|stomatin (EPB72)-like 2|stomatin-like 2', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35099776, end_position_on_the_genomic_accession=35103195, orientation='minus', exon_count=10, OMIM=608292, no_hdr0=''),
  100169989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169989, CurrentID=0, Status='live', Symbol='DBIL5P2', Aliases='ELP1P', description='diazepam binding inhibitor-like 5 pseudogene 2', other_designations='endozepine-like peptide 1 pseudogene', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63117851, end_position_on_the_genomic_accession=63119542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271521, CurrentID=0, Status='live', Symbol='RPL39P33', Aliases='RPL39_14_1560', description='ribosomal protein L39 pseudogene 33', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56572995, end_position_on_the_genomic_accession=56573388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162073, CurrentID=0, Status='live', Symbol='ITPRIPL2', Aliases='', description='inositol 1,4,5-trisphosphate receptor interacting protein-like 2', other_designations='inositol 1,4,5-triphosphate receptor interacting protein-like 2|inositol 1,4,5-triphosphate receptor-interacting protein-like 2', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19113932, end_position_on_the_genomic_accession=19121630, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100419915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419915, CurrentID=0, Status='live', Symbol='LOC100419915', Aliases='', description='OTU deubiquitinase 1 pseudogene', other_designations='OTU domain containing 1 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22203049, end_position_on_the_genomic_accession=22204157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162083, CurrentID=0, Status='live', Symbol='C16orf82', Aliases='TNT', description='chromosome 16 open reading frame 82', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27066898, end_position_on_the_genomic_accession=27069166, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373646, CurrentID=0, Status='live', Symbol='LOC105373646', Aliases='', description='uncharacterized LOC105373646', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139991109, end_position_on_the_genomic_accession=140002659, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  136242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136242, CurrentID=0, Status='live', Symbol='PRSS37', Aliases='TRYX2', description='protease, serine 37', other_designations='Peptidase S1 domain-containing protein LOC136242|probable inactive trypsin-X2|protease, serine, 37', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141836278, end_position_on_the_genomic_accession=141842499, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  729435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729435, CurrentID=0, Status='live', Symbol='ATP6V1E1P3', Aliases='ATP6EL1, ATP6V1EL1', description='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E1 pseudogene 3', other_designations='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E-like 1 (pseudogene)', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64337949, end_position_on_the_genomic_accession=64338905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271522, CurrentID=0, Status='live', Symbol='RPL39P29', Aliases='RPL39_15_1333', description='ribosomal protein L39 pseudogene 29', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102262845, end_position_on_the_genomic_accession=102263253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442078, CurrentID=0, Status='live', Symbol='KRT8P18', Aliases='', description='keratin 8 pseudogene 18', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35215458, end_position_on_the_genomic_accession=35230634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402569, CurrentID=0, Status='live', Symbol='KPNA7', Aliases='', description='karyopherin alpha 7 (importin alpha 8)', other_designations='karyopherin 7|karyopherin subunit alpha-7', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99173574, end_position_on_the_genomic_accession=99219333, orientation='minus', exon_count=15, OMIM=614107, no_hdr0=''),
  101928381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928381, CurrentID=0, Status='live', Symbol='LOC101928381', Aliases='', description='uncharacterized LOC101928381', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65791121, end_position_on_the_genomic_accession=65802027, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  399994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399994, CurrentID=0, Status='live', Symbol='HADHAP2', Aliases='', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), alpha subunit pseudogene 2', other_designations='hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), alpha subunit pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8634303, end_position_on_the_genomic_accession=8638265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271523, CurrentID=0, Status='live', Symbol='RPL39P34', Aliases='RPL39_15_1666', description='ribosomal protein L39 pseudogene 34', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52564022, end_position_on_the_genomic_accession=52564422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479331, CurrentID=0, Status='live', Symbol='RN7SL299P', Aliases='', description='RNA, 7SL, cytoplasmic 299, pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419917, CurrentID=0, Status='live', Symbol='LOC100419917', Aliases='', description='lysine-rich coiled-coil 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88003008, end_position_on_the_genomic_accession=88003739, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  162137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162137, CurrentID=0, Status='live', Symbol='LOC162137', Aliases='', description='uncharacterized LOC162137', other_designations='CTD-2144E22.5', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35022007, end_position_on_the_genomic_accession=35024229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479334, CurrentID=0, Status='live', Symbol='RN7SL305P', Aliases='', description='RNA, 7SL, cytoplasmic 305, pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271524, CurrentID=0, Status='live', Symbol='RPL39P35', Aliases='RPL39_16_1667', description='ribosomal protein L39 pseudogene 35', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53386180, end_position_on_the_genomic_accession=53386580, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  284729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284729, CurrentID=0, Status='live', Symbol='ESPNP', Aliases='dJ1182A14.1', description='espin pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16691218, end_position_on_the_genomic_accession=16720157, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  158158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158158, CurrentID=0, Status='live', Symbol='RASEF', Aliases='RAB45', description='RAS and EF-hand domain containing', other_designations='ras-related protein Rab-45', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82979585, end_position_on_the_genomic_accession=83063177, orientation='minus', exon_count=18, OMIM=611344, no_hdr0=''),
  106479339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479339, CurrentID=0, Status='live', Symbol='RN7SL319P', Aliases='', description='RNA, 7SL, cytoplasmic 319, pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271525, CurrentID=0, Status='live', Symbol='RPL39P31', Aliases='RPL39_17_1495', description='ribosomal protein L39 pseudogene 31', other_designations='', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71991151, end_position_on_the_genomic_accession=71991305, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646985, CurrentID=0, Status='live', Symbol='RPS12P25', Aliases='RPS12_14_1457', description='ribosomal protein S12 pseudogene 25', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79751666, end_position_on_the_genomic_accession=79752049, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158104, CurrentID=0, Status='live', Symbol='RPS10P3', Aliases='RPS10_10_994', description='ribosomal protein S10 pseudogene 3', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88016262, end_position_on_the_genomic_accession=88016850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  136259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136259, CurrentID=0, Status='live', Symbol='KLF14', Aliases='BTEB5', description='Kruppel-like factor 14', other_designations='BTE-binding protein 5|basic transcription element-binding protein 5|transcription factor BTEB5', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130732554, end_position_on_the_genomic_accession=130734033, orientation='minus', exon_count=1, OMIM=609393, no_hdr0=''),
  646748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646748, CurrentID=0, Status='live', Symbol='FLJ42022', Aliases='', description='uncharacterized LOC646748', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=72720842, end_position_on_the_genomic_accession=72724916, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  201798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201798, CurrentID=0, Status='live', Symbol='TIGD4', Aliases='', description='tigger transposable element derived 4', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152769354, end_position_on_the_genomic_accession=152779789, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271526, CurrentID=0, Status='live', Symbol='RPL39P36', Aliases='RPL39_17_1668', description='ribosomal protein L39 pseudogene 36', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53413140, end_position_on_the_genomic_accession=53413539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  201799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201799, CurrentID=0, Status='live', Symbol='TMEM154', Aliases='', description='transmembrane protein 154', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152626114, end_position_on_the_genomic_accession=152680165, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  442098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442098, CurrentID=0, Status='live', Symbol='MTHFD2P7', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 7', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179464346, end_position_on_the_genomic_accession=179465331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372838, CurrentID=0, Status='live', Symbol='LOC105372838', Aliases='', description='uncharacterized LOC105372838', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45327344, end_position_on_the_genomic_accession=45330632, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419920, CurrentID=0, Status='live', Symbol='LOC100419920', Aliases='', description='ankyrin repeat domain 33B pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95525346, end_position_on_the_genomic_accession=95527220, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131973, CurrentID=0, Status='live', Symbol='RHBDF1P1', Aliases='', description='RHBDF1 pseudogene 1', other_designations='rhomboid 5 homolog 1 pseudogene 1', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14572654, end_position_on_the_genomic_accession=14574792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373651, CurrentID=0, Status='live', Symbol='LOC105373651', Aliases='', description='uncharacterized LOC105373651', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142299010, end_position_on_the_genomic_accession=142341746, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  162239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162239, CurrentID=0, Status='live', Symbol='ZFP1', Aliases='ZNF475', description='ZFP1 zinc finger protein', other_designations='zfp-1|zinc finger protein 475', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75132509, end_position_on_the_genomic_accession=75172234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106479352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479352, CurrentID=0, Status='live', Symbol='RN7SL352P', Aliases='', description='RNA, 7SL, cytoplasmic 352, pseudogene', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  648771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648771, CurrentID=0, Status='live', Symbol='LOC648771', Aliases='', description='60S ribosomal protein L12-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442105, CurrentID=0, Status='live', Symbol='OR7E85P', Aliases='OR7E118P, OR7E88P', description='olfactory receptor family 7 subfamily E member 85 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 118 pseudogene|olfactory receptor, family 7, subfamily E, member 85 pseudogene|olfactory receptor, family 7, subfamily E, member 88 pseudogene|seven transmembrane helix receptor', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9483603, end_position_on_the_genomic_accession=9484830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930127, CurrentID=0, Status='live', Symbol='LOC101930127', Aliases='', description='uncharacterized LOC101930127', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271528, CurrentID=0, Status='live', Symbol='RPL39P37', Aliases='RPL39_18_1669', description='ribosomal protein L39 pseudogene 37', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53495832, end_position_on_the_genomic_accession=53496229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100170216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170216, CurrentID=0, Status='live', Symbol='SNAR-A10', Aliases='', description='small ILF3/NF90-associated RNA A10', other_designations='snaR-A55318263', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50123074, end_position_on_the_genomic_accession=50123194, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170217, CurrentID=0, Status='live', Symbol='SNAR-B2', Aliases='', description='small ILF3/NF90-associated RNA B2', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50139118, end_position_on_the_genomic_accession=50139237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  162282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162282, CurrentID=0, Status='live', Symbol='ANKFN1', Aliases='', description='ankyrin repeat and fibronectin type III domain containing 1', other_designations='ankyrin-repeat and fibronectin type III domain containing 1', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55965604, end_position_on_the_genomic_accession=56511407, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  100170219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170219, CurrentID=0, Status='live', Symbol='SNAR-C4', Aliases='', description='small ILF3/NF90-associated RNA C4', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47939541, end_position_on_the_genomic_accession=47939660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170220, CurrentID=0, Status='live', Symbol='SNAR-E', Aliases='', description='small ILF3/NF90-associated RNA E', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46830585, end_position_on_the_genomic_accession=46830704, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170221, CurrentID=0, Status='live', Symbol='SNAR-H', Aliases='snaR-2', description='small ILF3/NF90-associated RNA H', other_designations='small ILF3/NF90-associated RNA 2', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77954907, end_position_on_the_genomic_accession=77955026, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170222, CurrentID=0, Status='live', Symbol='SNAR-I', Aliases='snaR-3', description='small ILF3/NF90-associated RNA I', other_designations='small ILF3/NF90-associated RNA 3', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190877930, end_position_on_the_genomic_accession=190878050, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170223, CurrentID=0, Status='live', Symbol='SNAR-C5', Aliases='', description='small ILF3/NF90-associated RNA C5', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47913051, end_position_on_the_genomic_accession=47913170, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170224, CurrentID=0, Status='live', Symbol='SNAR-B1', Aliases='', description='small ILF3/NF90-associated RNA B1', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50133745, end_position_on_the_genomic_accession=50133864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170225, CurrentID=0, Status='live', Symbol='SNAR-C1', Aliases='', description='small ILF3/NF90-associated RNA C1', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47955682, end_position_on_the_genomic_accession=47955801, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170226, CurrentID=0, Status='live', Symbol='SNAR-C3', Aliases='', description='small ILF3/NF90-associated RNA C3', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47950296, end_position_on_the_genomic_accession=47950414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170227, CurrentID=0, Status='live', Symbol='SNAR-D', Aliases='', description='small ILF3/NF90-associated RNA D', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50140202, end_position_on_the_genomic_accession=50140320, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170228, CurrentID=0, Status='live', Symbol='SNAR-G2', Aliases='', description='small ILF3/NF90-associated RNA G2', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49031669, end_position_on_the_genomic_accession=49031787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100170229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170229, CurrentID=0, Status='live', Symbol='SRRM5', Aliases='ZNF576', description='serine/arginine repetitive matrix 5', other_designations='zinc finger protein 576', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43612101, end_position_on_the_genomic_accession=43614498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479360, CurrentID=0, Status='live', Symbol='RN7SL377P', Aliases='', description='RNA, 7SL, cytoplasmic 377, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442113, CurrentID=0, Status='live', Symbol='LOC442113', Aliases='', description='protein tyrosine phosphatase, non-receptor type 11 pseudogene', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82580120, end_position_on_the_genomic_accession=82581695, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930047, CurrentID=0, Status='live', Symbol='LOC101930047', Aliases='', description='uncharacterized protein FLJ46347-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442114, CurrentID=0, Status='live', Symbol='KRT19P3', Aliases='', description='keratin 19 pseudogene 3', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109878947, end_position_on_the_genomic_accession=109879893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479363, CurrentID=0, Status='live', Symbol='RN7SL383P', Aliases='', description='RNA, 7SL, cytoplasmic 383, pseudogene', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  162333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162333, CurrentID=0, Status='live', Symbol='MARCH10', Aliases='MARCH-X, RNF190', description='membrane associated ring-CH-type finger 10', other_designations='membrane associated ring finger 10|membrane-associated RING finger protein 10|membrane-associated RING-CH protein X|membrane-associated ring finger (C3HC4) 10, E3 ubiquitin protein ligase|ring finger protein 190', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62701314, end_position_on_the_genomic_accession=62808381, orientation='minus', exon_count=17, OMIM=613337, no_hdr0=''),
  106479365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479365, CurrentID=0, Status='live', Symbol='RN7SL388P', Aliases='', description='RNA, 7SL, cytoplasmic 388, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419924, CurrentID=0, Status='live', Symbol='LOC100419924', Aliases='', description='uncharacterized LOC100419924', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40113780, end_position_on_the_genomic_accession=40114283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373655, CurrentID=0, Status='live', Symbol='LOC105373655', Aliases='', description='uncharacterized LOC105373655', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143217361, end_position_on_the_genomic_accession=143224173, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  136288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136288, CurrentID=0, Status='live', Symbol='C7orf57', Aliases='', description='chromosome 7 open reading frame 57', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48035511, end_position_on_the_genomic_accession=48061297, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101930049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930049, CurrentID=0, Status='live', Symbol='LOC101930049', Aliases='', description='uncharacterized LOC101930049', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158135, CurrentID=0, Status='live', Symbol='TTLL11', Aliases='C9orf20, bA244O19.1', description='tubulin tyrosine ligase like11', other_designations='tubulin tyrosine ligase-like family member 11|tubulin tyrosine ligase-like family, member 11|tubulin--tyrosine ligase-like protein 11', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121815674, end_position_on_the_genomic_accession=122093606, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  162387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162387, CurrentID=0, Status='live', Symbol='MFSD6L', Aliases='', description='major facilitator superfamily domain containing 6-like', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8797110, end_position_on_the_genomic_accession=8799349, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  162394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162394, CurrentID=0, Status='live', Symbol='SLFN5', Aliases='', description='schlafen family member 5', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35243067, end_position_on_the_genomic_accession=35267749, orientation='plus', exon_count=5, OMIM=614952, no_hdr0=''),
  100131868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131868, CurrentID=0, Status='live', Symbol='LOC100131868', Aliases='', description='serine/threonine-protein phosphatase 4 regulatory subunit 2-like', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197634245, end_position_on_the_genomic_accession=197635945, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162417, CurrentID=0, Status='live', Symbol='NAGS', Aliases='AGAS, ARGA', description='N-acetylglutamate synthase', other_designations='amino-acid acetyltransferase', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44004546, end_position_on_the_genomic_accession=44009068, orientation='plus', exon_count=8, OMIM=608300, no_hdr0=''),
  106479379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479379, CurrentID=0, Status='live', Symbol='RN7SL428P', Aliases='', description='RNA, 7SL, cytoplasmic 428, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442132, CurrentID=0, Status='live', Symbol='LOC442132', Aliases='', description='golgin A6 family-like 1 pseudogene', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7299374, end_position_on_the_genomic_accession=7306714, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  162427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162427, CurrentID=0, Status='live', Symbol='FAM134C', Aliases='', description='family with sequence similarity 134 member C', other_designations='family with sequence similarity 134, member C', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42579508, end_position_on_the_genomic_accession=42609427, orientation='minus', exon_count=10, OMIM=616498, no_hdr0=''),
  401742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401742, CurrentID=0, Status='live', Symbol='OR11K2P', Aliases='', description='olfactory receptor family 11 subfamily K member 2 pseudogene', other_designations='olfactory receptor, family 11, subfamily K, member 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19733345, end_position_on_the_genomic_accession=19734292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373658, CurrentID=0, Status='live', Symbol='LOC105373658', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144462383, end_position_on_the_genomic_accession=144470994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271534, CurrentID=0, Status='live', Symbol='RPL7AP52', Aliases='RPL7A_18_1082', description='ribosomal protein L7a pseudogene 52', other_designations='', map_location='10q24.1-q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94882564, end_position_on_the_genomic_accession=94883349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162461, CurrentID=0, Status='live', Symbol='TMEM92', Aliases='', description='transmembrane protein 92', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50271406, end_position_on_the_genomic_accession=50281485, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100419928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419928, CurrentID=0, Status='live', Symbol='LOC100419928', Aliases='', description='M-phase phosphoprotein 8 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7438761, end_position_on_the_genomic_accession=7439988, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  162466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162466, CurrentID=0, Status='live', Symbol='PHOSPHO1', Aliases='', description='phosphatase, orphan 1', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49223366, end_position_on_the_genomic_accession=49231692, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  140710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140710, CurrentID=0, Status='live', Symbol='SOGA1', Aliases='C20orf117, KIAA0889, SOGA', description='suppressor of glucose, autophagy associated 1', other_designations='SOGA family member 1|suppressor of glucose by autophagy|suppressor of glucose from autophagy|suppressor of glucose, autophagy-associated protein 1', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36777442, end_position_on_the_genomic_accession=36863684, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100271535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271535, CurrentID=0, Status='live', Symbol='RPL7AP55', Aliases='RPL7A_19_1118', description='ribosomal protein L7a pseudogene 55', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7783674, end_position_on_the_genomic_accession=7784466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646995, CurrentID=0, Status='live', Symbol='LOC646995', Aliases='', description='tripartite motif-containing protein 38-like', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165664092, end_position_on_the_genomic_accession=165665662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162494, CurrentID=0, Status='live', Symbol='RHBDL3', Aliases='RHBDL4, VRHO', description='rhomboid, veinlet-like 3 (Drosophila)', other_designations='rhomboid, veinlet-like 4|ventrhoid transmembrane protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32266131, end_position_on_the_genomic_accession=32324661, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  140711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140711, CurrentID=0, Status='live', Symbol='TLDC2', Aliases='C20orf118', description='TBC/LysM-associated domain containing 2', other_designations='TBC/LysM-associated domain-containing protein 2', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36876121, end_position_on_the_genomic_accession=36894235, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373660, CurrentID=0, Status='live', Symbol='LOC105373660', Aliases='', description='uncharacterized LOC105373660', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144440543, end_position_on_the_genomic_accession=144450689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100301521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301521, CurrentID=0, Status='live', Symbol='LOC100301521', Aliases='', description='chromosome 15 open reading frame 40 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41818946, end_position_on_the_genomic_accession=41819344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162514, CurrentID=0, Status='live', Symbol='TRPV3', Aliases='FNEPPK2, OLMS, VRL3', description='transient receptor potential cation channel, subfamily V, member 3', other_designations='VRL-3|vanilloid receptor-like 3|vanilloid receptor-related osmotically activated channel protein', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3510502, end_position_on_the_genomic_accession=3557995, orientation='minus', exon_count=19, OMIM=607066, no_hdr0=''),
  162515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162515, CurrentID=0, Status='live', Symbol='SLC16A11', Aliases='MCT11', description='solute carrier family 16 member 11', other_designations='MCT 11|monocarboxylic acid transporter 11|solute carrier family 16 (monocarboxylic acid transporters), member 11', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7041617, end_position_on_the_genomic_accession=7044400, orientation='minus', exon_count=4, OMIM=615765, no_hdr0=''),
  162517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162517, CurrentID=0, Status='live', Symbol='FBXO39', Aliases='CT144, Fbx39', description='F-box protein 39', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6776175, end_position_on_the_genomic_accession=6796474, orientation='plus', exon_count=6, OMIM=609106, no_hdr0=''),
  100301526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301526, CurrentID=0, Status='live', Symbol='STQTL17', Aliases='', description='Stature QTL 17', other_designations='', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612737, no_hdr0=''),
  100301527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301527, CurrentID=0, Status='live', Symbol='SYNSTH', Aliases='', description='Synesthesia', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612759, no_hdr0=''),
  162540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162540, CurrentID=0, Status='live', Symbol='SPPL2C', Aliases='IMP5', description='signal peptide peptidase like 2C', other_designations='IMP-5|SPP-like 2C|intramembrane protease 5', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45844890, end_position_on_the_genomic_accession=45847072, orientation='plus', exon_count=1, OMIM=608284, no_hdr0=''),
  100271537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271537, CurrentID=0, Status='live', Symbol='RPL7AP57', Aliases='RPL7A_23_1164', description='ribosomal protein L7a pseudogene 57', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92161081, end_position_on_the_genomic_accession=92161947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442153, CurrentID=0, Status='live', Symbol='PSMC1P11', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 11', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4702442, end_position_on_the_genomic_accession=4703975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479402, CurrentID=0, Status='live', Symbol='RN7SL492P', Aliases='', description='RNA, 7SL, cytoplasmic 492, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442155, CurrentID=0, Status='live', Symbol='LOC442155', Aliases='', description='transcription factor B2, mitochondrial pseudogene', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5065607, end_position_on_the_genomic_accession=5067041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100301572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301572, CurrentID=0, Status='live', Symbol='PRBNS', Aliases='', description='Pierre Robin syndrome', other_designations='', map_location='17q24.3-q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=261800, no_hdr0=''),
  442156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442156, CurrentID=0, Status='live', Symbol='PKMP5', Aliases='', description='pyruvate kinase, muscle pseudogene 5', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5972324, end_position_on_the_genomic_accession=5974206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442157, CurrentID=0, Status='live', Symbol='HNRNPLP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein L pseudogene 1', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7480864, end_position_on_the_genomic_accession=7482595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479406, CurrentID=0, Status='live', Symbol='RN7SL500P', Aliases='', description='RNA, 7SL, cytoplasmic 500, pseudogene', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22872, CurrentID=0, Status='live', Symbol='SEC31A', Aliases='ABP125, ABP130, HSPC275, HSPC334, SEC31L1', description='SEC31 homolog A, COPII coat complex component', other_designations='SEC31 homolog A, COPII coating complex component|SEC31-like protein 1|SEC31-related protein A|web1-like protein|yeast Sec31p homolog', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82818661, end_position_on_the_genomic_accession=82891280, orientation='minus', exon_count=30, OMIM=610257, no_hdr0=''),
  106479408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479408, CurrentID=0, Status='live', Symbol='RN7SL507P', Aliases='', description='RNA, 7SL, cytoplasmic 507, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373663, CurrentID=0, Status='live', Symbol='LOC105373663', Aliases='', description='uncharacterized LOC105373663', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144806056, end_position_on_the_genomic_accession=144810119, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479409, CurrentID=0, Status='live', Symbol='RN7SL509P', Aliases='', description='RNA, 7SL, cytoplasmic 509, pseudogene', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  162605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162605, CurrentID=0, Status='live', Symbol='KRT28', Aliases='KRT25D', description='keratin 28, type I', other_designations='CK-28|K25D|K28|cytokeratin-28|keratin 25D|keratin-25D|keratin-28|type I inner root sheath specific keratin 25 irs4|type I inner root sheath-specific keratin-K25irs4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40792196, end_position_on_the_genomic_accession=40799959, orientation='minus', exon_count=8, OMIM=616677, no_hdr0=''),
  105371265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371265, CurrentID=0, Status='live', Symbol='LOC105371265', Aliases='', description='uncharacterized LOC105371265', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52581252, end_position_on_the_genomic_accession=52588579, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100009668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009668, CurrentID=0, Status='live', Symbol='POU5F1P6', Aliases='', description='POU class 5 homeobox 1 pseudogene 6', other_designations='POU domain class 5, transcription factor 1 pseudogene 6', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128674646, end_position_on_the_genomic_accession=128676610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442164, CurrentID=0, Status='live', Symbol='DDX18P3', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 3', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18363096, end_position_on_the_genomic_accession=18365461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642635, CurrentID=0, Status='live', Symbol='LOC642635', Aliases='', description='tubulin, alpha 4a pseudogene', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153420677, end_position_on_the_genomic_accession=153421913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189219, CurrentID=0, Status='live', Symbol='TRY-GTA6-1', Aliases='TRNAY8', description='transfer RNA-Tyr (GTA) 6-1', other_designations='transfer RNA tyrosine 8 (anticodon GUA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26594874, end_position_on_the_genomic_accession=26594962, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162632, CurrentID=0, Status='live', Symbol='USP32P1', Aliases='', description='ubiquitin specific peptidase 32 pseudogene 1', other_designations='TL132 pseudogene|ubiquitin specific peptidase 6 (Tre-2 oncogene) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16786892, end_position_on_the_genomic_accession=16804505, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  136332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136332, CurrentID=0, Status='live', Symbol='LRGUK', Aliases='', description='leucine-rich repeats and guanylate kinase domain containing', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134127352, end_position_on_the_genomic_accession=134264181, orientation='plus', exon_count=23, OMIM=616478, no_hdr0=''),
  106479415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479415, CurrentID=0, Status='live', Symbol='RN7SL526P', Aliases='', description='RNA, 7SL, cytoplasmic 526, pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100009669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009669, CurrentID=0, Status='live', Symbol='POU5F1P7', Aliases='', description='POU class 5 homeobox 1 pseudogene 7', other_designations='POU domain class 5, transcription factor 1 pseudogene 7', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98119273, end_position_on_the_genomic_accession=98119975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162655, CurrentID=0, Status='live', Symbol='ZNF519', Aliases='HsT2362', description='zinc finger protein 519', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14056324, end_position_on_the_genomic_accession=14132490, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  442171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442171, CurrentID=0, Status='live', Symbol='RPL10P2', Aliases='RPL10_2_670, RPL10p, bA209A2.1', description='ribosomal protein L10 pseudogene 2', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27211221, end_position_on_the_genomic_accession=27211941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442172, CurrentID=0, Status='live', Symbol='LOC442172', Aliases='', description='cell division cycle associated 7 pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27268106, end_position_on_the_genomic_accession=27269229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162681, CurrentID=0, Status='live', Symbol='C18orf54', Aliases='LAS2', description='chromosome 18 open reading frame 54', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54357905, end_position_on_the_genomic_accession=54385218, orientation='plus', exon_count=14, OMIM=613258, no_hdr0=''),
  106479423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479423, CurrentID=0, Status='live', Symbol='RN7SL554P', Aliases='', description='RNA, 7SL, cytoplasmic 554, pseudogene', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271542, CurrentID=0, Status='live', Symbol='RPL7AP67', Aliases='RPL7A_29_1595', description='ribosomal protein L7a pseudogene 67', other_designations='', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35383345, end_position_on_the_genomic_accession=35384025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441897, CurrentID=0, Status='live', Symbol='PGCP1', Aliases='', description='progastricsin (pepsinogen C) pseudogene 1', other_designations='progastricsin pseudogene', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111384519, end_position_on_the_genomic_accession=111389851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  162699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162699, CurrentID=0, Status='live', Symbol='TCEB3C', Aliases='HsT829, TCEB3L2', description='transcription elongation factor B polypeptide 3C (elongin A3)', other_designations='eloA3|elongin-A3', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47028202, end_position_on_the_genomic_accession=47030078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  647002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647002, CurrentID=0, Status='live', Symbol='LOC647002', Aliases='', description='isopentenyl-diphosphate delta isomerase 1 pseudogene', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138307298, end_position_on_the_genomic_accession=138307973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479427, CurrentID=0, Status='live', Symbol='RN7SL563P', Aliases='', description='RNA, 7SL, cytoplasmic 563, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996249, CurrentID=0, Status='live', Symbol='LOC100996249', Aliases='', description='uncharacterized LOC100996249', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112618065, end_position_on_the_genomic_accession=112620649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441870, CurrentID=0, Status='live', Symbol='PRAMEF30P', Aliases='', description='PRAME family member 30, pseudogene', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12838470, end_position_on_the_genomic_accession=12841383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479429, CurrentID=0, Status='live', Symbol='RN7SL568P', Aliases='', description='RNA, 7SL, cytoplasmic 568, pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927126, CurrentID=0, Status='live', Symbol='LOC101927126', Aliases='', description='uncharacterized LOC101927126', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76077255, end_position_on_the_genomic_accession=76164388, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  100996251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996251, CurrentID=0, Status='live', Symbol='LOC100996251', Aliases='', description='uncharacterized LOC100996251', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113011687, end_position_on_the_genomic_accession=113073102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  442183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442183, CurrentID=0, Status='live', Symbol='OR2P1P', Aliases='hs6M1-26', description='olfactory receptor family 2 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily P, member 1 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29071832, end_position_on_the_genomic_accession=29072566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479432, CurrentID=0, Status='live', Symbol='RN7SL575P', Aliases='', description='RNA, 7SL, cytoplasmic 575, pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479433, CurrentID=0, Status='live', Symbol='RN7SL578P', Aliases='', description='RNA, 7SL, cytoplasmic 578, pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401744, CurrentID=0, Status='live', Symbol='SMARCE1P3', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 3', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21162512, end_position_on_the_genomic_accession=21163790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479434, CurrentID=0, Status='live', Symbol='RN7SL581P', Aliases='', description='RNA, 7SL, cytoplasmic 581, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479436, CurrentID=0, Status='live', Symbol='RN7SL585P', Aliases='', description='RNA, 7SL, cytoplasmic 585, pseudogene', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479437, CurrentID=0, Status='live', Symbol='RN7SL587P', Aliases='', description='RNA, 7SL, cytoplasmic 587, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479438, CurrentID=0, Status='live', Symbol='RN7SL589P', Aliases='', description='RNA, 7SL, cytoplasmic 589, pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132830, CurrentID=0, Status='live', Symbol='LOC100132830', Aliases='', description='zinc finger CCCH-type containing 14 pseudogene', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=94442798, end_position_on_the_genomic_accession=94443821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479439, CurrentID=0, Status='live', Symbol='RN7SL594P', Aliases='', description='RNA, 7SL, cytoplasmic 594, pseudogene', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287467, CurrentID=0, Status='live', Symbol='LOC100287467', Aliases='', description='uncharacterized LOC100287467', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116369367, end_position_on_the_genomic_accession=116372064, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442192, CurrentID=0, Status='live', Symbol='DDX6P1', Aliases='DDX6-Lp, DDX6P, bA150A6.3', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene 1', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 6 pseudogene 1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 6 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29329629, end_position_on_the_genomic_accession=29331061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996262, CurrentID=0, Status='live', Symbol='LOC100996262', Aliases='', description='uncharacterized LOC100996262', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147027979, end_position_on_the_genomic_accession=147028805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100529052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529052, CurrentID=0, Status='live', Symbol='LOC100529052', Aliases='', description='gamma-glutamylcyclotransferase pseudogene', other_designations='', map_location='20p11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18220260, end_position_on_the_genomic_accession=18221062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996263, CurrentID=0, Status='live', Symbol='LOC100996263', Aliases='', description='uncharacterized LOC100996263', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117596832, end_position_on_the_genomic_accession=117605770, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479442, CurrentID=0, Status='live', Symbol='RN7SL604P', Aliases='', description='RNA, 7SL, cytoplasmic 604, pseudogene', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479443, CurrentID=0, Status='live', Symbol='RN7SL606P', Aliases='', description='RNA, 7SL, cytoplasmic 606, pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271546, CurrentID=0, Status='live', Symbol='RPS10P25', Aliases='RPS10_10_1555', description='ribosomal protein S10 pseudogene 25', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49148316, end_position_on_the_genomic_accession=49148774, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100009675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009675, CurrentID=0, Status='live', Symbol='MRT4', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 4', other_designations='', map_location='1p21.3-p13.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611107, no_hdr0=''),
  105369306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369306, CurrentID=0, Status='live', Symbol='LOC105369306', Aliases='', description='uncharacterized LOC105369306', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36104434, end_position_on_the_genomic_accession=36109692, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  284751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284751, CurrentID=0, Status='live', Symbol='LINC01270', Aliases='', description='long intergenic non-protein coding RNA 1270', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50292720, end_position_on_the_genomic_accession=50314919, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100170765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170765, CurrentID=0, Status='live', Symbol='ERICH4', Aliases='C19orf69', description='glutamate rich 4', other_designations='glutamate-rich 4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41443158, end_position_on_the_genomic_accession=41444765, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271547, CurrentID=0, Status='live', Symbol='RPS10P20', Aliases='RPS10_11_1233', description='ribosomal protein S10 pseudogene 20', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48487913, end_position_on_the_genomic_accession=48488487, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479452, CurrentID=0, Status='live', Symbol='RN7SL629P', Aliases='', description='RNA, 7SL, cytoplasmic 629, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442205, CurrentID=0, Status='live', Symbol='KRT18P9', Aliases='bA513I15.2', description='keratin 18 pseudogene 9', other_designations='keratin 18 pseudogene 8', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34189776, end_position_on_the_genomic_accession=34191141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479454, CurrentID=0, Status='live', Symbol='RN7SL635P', Aliases='', description='RNA, 7SL, cytoplasmic 635, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271548, CurrentID=0, Status='live', Symbol='RPS10P26', Aliases='RPS10_11_1569', description='ribosomal protein S10 pseudogene 26', other_designations='', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62516275, end_position_on_the_genomic_accession=62516804, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27147, CurrentID=0, Status='live', Symbol='DENND2A', Aliases='FAM31D, KIAA1277', description='DENN/MADD domain containing 2A', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140518418, end_position_on_the_genomic_accession=140641732, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  442210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442210, CurrentID=0, Status='live', Symbol='TUBBP9', Aliases='', description='tubulin beta class I pseudogene 9', other_designations='tubulin, beta class I pseudogene 9', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39992652, end_position_on_the_genomic_accession=40000269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442211, CurrentID=0, Status='live', Symbol='ATP6V0CP3', Aliases='', description='ATPase, H+ transporting, lysosomal 16kDa, V0 subunit c pseudogene 3', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42727576, end_position_on_the_genomic_accession=42728194, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100170841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170841, CurrentID=0, Status='live', Symbol='C17orf96', Aliases='PRR28', description='chromosome 17 open reading frame 96', other_designations='Uncharacterized protein ENSP00000317905|proline rich 28', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38671703, end_position_on_the_genomic_accession=38674934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  425054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=425054, CurrentID=0, Status='live', Symbol='VCX3B', Aliases='VCX-C, VCXC', description='variable charge, X-linked 3B', other_designations='variably charged protein X-C', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8464830, end_position_on_the_genomic_accession=8466510, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479461, CurrentID=0, Status='live', Symbol='RN7SL654P', Aliases='', description='RNA, 7SL, cytoplasmic 654, pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  425057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=425057, CurrentID=0, Status='live', Symbol='TTTY9B', Aliases='NCRNA00132, TTTY9, TTTY9A, TTY9', description='testis-specific transcript, Y-linked 9B (non-protein coding)', other_designations='testis transcript Y 9|testis-specific transcript, Y-linked 9, centromeric', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18581206, end_position_on_the_genomic_accession=18590521, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132481, CurrentID=0, Status='live', Symbol='LOC100132481', Aliases='', description='uncharacterized LOC100132481', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=126426195, end_position_on_the_genomic_accession=126427493, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442215, CurrentID=0, Status='live', Symbol='EEF1A1P42', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 42', other_designations='eukaryotic translation elongation factor 1 alpha 2 pseudogene', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49358191, end_position_on_the_genomic_accession=49360404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479464, CurrentID=0, Status='live', Symbol='RN7SL664P', Aliases='', description='RNA, 7SL, cytoplasmic 664, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271550, CurrentID=0, Status='live', Symbol='RPS12P21', Aliases='RPS12_11_1175', description='ribosomal protein S12 pseudogene 21', other_designations='', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112218244, end_position_on_the_genomic_accession=112218737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  158228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158228, CurrentID=0, Status='live', Symbol='FAM201A', Aliases='C9orf122', description='family with sequence similarity 201 member A', other_designations='family with sequence similarity 201, member A', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38621088, end_position_on_the_genomic_accession=38623284, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479466, CurrentID=0, Status='live', Symbol='RN7SL668P', Aliases='', description='RNA, 7SL, cytoplasmic 668, pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479467, CurrentID=0, Status='live', Symbol='RN7SL671P', Aliases='', description='RNA, 7SL, cytoplasmic 671, pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929255, CurrentID=0, Status='live', Symbol='LOC101929255', Aliases='', description='uncharacterized LOC101929255', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22402789, end_position_on_the_genomic_accession=22409182, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  442221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442221, CurrentID=0, Status='live', Symbol='RCC2P7', Aliases='', description='regulator of chromosome condensation 2 pseudogene 7', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56431930, end_position_on_the_genomic_accession=56433429, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  162962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162962, CurrentID=0, Status='live', Symbol='ZNF836', Aliases='', description='zinc finger protein 836', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52154612, end_position_on_the_genomic_accession=52171667, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  162963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162963, CurrentID=0, Status='live', Symbol='ZNF610', Aliases='', description='zinc finger protein 610', other_designations='zink finger protein', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52331863, end_position_on_the_genomic_accession=52367123, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100271551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271551, CurrentID=0, Status='live', Symbol='RPS12P24', Aliases='RPS12_12_1308', description='ribosomal protein S12 pseudogene 24', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37220858, end_position_on_the_genomic_accession=37221252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  162966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162966, CurrentID=0, Status='live', Symbol='ZNF600', Aliases='KR-ZNF1', description='zinc finger protein 600', other_designations='zinc finger protein KR-ZNF1', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52765471, end_position_on_the_genomic_accession=52833628, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  162967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162967, CurrentID=0, Status='live', Symbol='ZNF320', Aliases='ZFPL', description='zinc finger protein 320', other_designations='zinc finger gene 320|zinc finger protein like', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52876172, end_position_on_the_genomic_accession=52897693, orientation='minus', exon_count=5, OMIM=606427, no_hdr0=''),
  162968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162968, CurrentID=0, Status='live', Symbol='ZNF497', Aliases='', description='zinc finger protein 497', other_designations='zinc finger-like protein', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58354357, end_position_on_the_genomic_accession=58362848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  162972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162972, CurrentID=0, Status='live', Symbol='ZNF550', Aliases='', description='zinc finger protein 550', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57541836, end_position_on_the_genomic_accession=57561647, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  162979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162979, CurrentID=0, Status='live', Symbol='ZNF296', Aliases='ZFP296, ZNF342', description='zinc finger protein 296', other_designations='zinc finger protein 342', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45071500, end_position_on_the_genomic_accession=45076430, orientation='minus', exon_count=3, OMIM=613226, no_hdr0=''),
  100301990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301990, CurrentID=0, Status='live', Symbol='LOC100301990', Aliases='', description='zinc finger protein 610 pseudogene', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52732379, end_position_on_the_genomic_accession=52734077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442225, CurrentID=0, Status='live', Symbol='LOC442225', Aliases='', description='keratin, type II pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63193058, end_position_on_the_genomic_accession=63193631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  162989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162989, CurrentID=0, Status='live', Symbol='DEDD2', Aliases='FLAME-3', description='death effector domain containing 2', other_designations='DED-containing protein FLAME-3|FADD-like anti-apoptotic molecule 3', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42198593, end_position_on_the_genomic_accession=42220152, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  162993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162993, CurrentID=0, Status='live', Symbol='ZNF846', Aliases='', description='zinc finger protein 846', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9752145, end_position_on_the_genomic_accession=9786177, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  106479475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479475, CurrentID=0, Status='live', Symbol='RN7SL690P', Aliases='', description='RNA, 7SL, cytoplasmic 690, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271552, CurrentID=0, Status='live', Symbol='RPS12P27', Aliases='RPS12_13_1490', description='ribosomal protein S12 pseudogene 27', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68209381, end_position_on_the_genomic_accession=68209824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  162998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=162998, CurrentID=0, Status='live', Symbol='OR7D2', Aliases='HTPCRH03, OR19-10, OR19-4', description='olfactory receptor family 7 subfamily D member 2', other_designations='olfactory receptor 19-4|olfactory receptor OR19-10|olfactory receptor, family 7, subfamily D, member 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9185594, end_position_on_the_genomic_accession=9188817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  647012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647012, CurrentID=0, Status='live', Symbol='YY1P2', Aliases='', description='YY1 transcription factor pseudogene 2', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138897324, end_position_on_the_genomic_accession=138899174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100887071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887071, CurrentID=0, Status='live', Symbol='LOC100887071', Aliases='', description='zinc finger protein 845 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  163007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163007, CurrentID=0, Status='live', Symbol='CYP2T3P', Aliases='', description='cytochrome P450 family 2 subfamily T member 3, pseudogene', other_designations='cytochrome P450, family 2, subfamily T, polypeptide 3 pseudogene|cytochrome P450, family 2, subfamily T, polypeptide 3, pseudogene|cytochrome P450, subfamily IIT, polypeptide 3 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41134719, end_position_on_the_genomic_accession=41137353, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441880, CurrentID=0, Status='live', Symbol='LOC441880', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31487591, end_position_on_the_genomic_accession=31488973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642648, CurrentID=0, Status='live', Symbol='LOC642648', Aliases='', description='uncharacterized LOC642648', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46064651, end_position_on_the_genomic_accession=46071037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479479, CurrentID=0, Status='live', Symbol='RN7SL703P', Aliases='', description='RNA, 7SL, cytoplasmic 703, pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479480, CurrentID=0, Status='live', Symbol='RN7SL706P', Aliases='', description='RNA, 7SL, cytoplasmic 706, pseudogene', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271553, CurrentID=0, Status='live', Symbol='RPS12P28', Aliases='RPS12_14_1532', description='ribosomal protein S12 pseudogene 28', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28655609, end_position_on_the_genomic_accession=28655981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442233, CurrentID=0, Status='live', Symbol='SMARCE1P2', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 2', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84429581, end_position_on_the_genomic_accession=84430801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163033, CurrentID=0, Status='live', Symbol='ZNF579', Aliases='', description='zinc finger protein 579', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55577523, end_position_on_the_genomic_accession=55580899, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369313, CurrentID=0, Status='live', Symbol='LOC105369313', Aliases='', description='uncharacterized LOC105369313', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58493615, end_position_on_the_genomic_accession=58526239, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  442234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442234, CurrentID=0, Status='live', Symbol='RCN1P1', Aliases='', description='reticulocalbin 1 pseudogene 1', other_designations='reticulocalbin 1, EF-hand calcium binding domain pseudogene 1', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87121696, end_position_on_the_genomic_accession=87122688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163049, CurrentID=0, Status='live', Symbol='ZNF791', Aliases='', description='zinc finger protein 791', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12610918, end_position_on_the_genomic_accession=12629862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  163050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163050, CurrentID=0, Status='live', Symbol='ZNF564', Aliases='', description='zinc finger protein 564', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12525370, end_position_on_the_genomic_accession=12551542, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  163051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163051, CurrentID=0, Status='live', Symbol='ZNF709', Aliases='', description='zinc finger protein 709', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12461184, end_position_on_the_genomic_accession=12484818, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479485, CurrentID=0, Status='live', Symbol='RN7SL718P', Aliases='', description='RNA, 7SL, cytoplasmic 718, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  163059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163059, CurrentID=0, Status='live', Symbol='ZNF433', Aliases='', description='zinc finger protein 433', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12014717, end_position_on_the_genomic_accession=12035710, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105369314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369314, CurrentID=0, Status='live', Symbol='LOC105369314', Aliases='', description='uncharacterized LOC105369314', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58770400, end_position_on_the_genomic_accession=58785705, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  442239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442239, CurrentID=0, Status='live', Symbol='PRDX2P4', Aliases='', description='peroxiredoxin 2 pseudogene 4', other_designations='', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100135992, end_position_on_the_genomic_accession=100136876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  163071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163071, CurrentID=0, Status='live', Symbol='ZNF114', Aliases='', description='zinc finger protein 114', other_designations='zinc finger protein 114 (Y18)', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48270099, end_position_on_the_genomic_accession=48287608, orientation='plus', exon_count=9, OMIM=603996, no_hdr0=''),
  132001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132001, CurrentID=0, Status='live', Symbol='TAMM41', Aliases='C3orf31, RAM41, TAM41', description='TAM41 mitochondrial translocator assembly and maintenance homolog', other_designations='CDP-DAG synthase|CDP-diacylglycerol synthase|MMP37-like protein, mitochondrial|TAM41, mitochondrial translocator assembly and maintenance protein, homolog|mitochondrial translocator assembly and maintenance protein 41 homolog', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11790442, end_position_on_the_genomic_accession=11846919, orientation='minus', exon_count=14, OMIM=614948, no_hdr0=''),
  100131890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131890, CurrentID=0, Status='live', Symbol='LOC100131890', Aliases='', description='neugrin, neurite outgrowth associated pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71550959, end_position_on_the_genomic_accession=71551670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373679, CurrentID=0, Status='live', Symbol='LOC105373679', Aliases='', description='uncharacterized LOC105373679', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149287752, end_position_on_the_genomic_accession=149290184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  163081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163081, CurrentID=0, Status='live', Symbol='ZNF567', Aliases='', description='zinc finger protein 567', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36687564, end_position_on_the_genomic_accession=36723358, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100302090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302090, CurrentID=0, Status='live', Symbol='DDX11L8', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 8', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 8', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12310, end_position_on_the_genomic_accession=14520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479490, CurrentID=0, Status='live', Symbol='RN7SL734P', Aliases='', description='RNA, 7SL, cytoplasmic 734, pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271555, CurrentID=0, Status='live', Symbol='RPS12P30', Aliases='RPS12_15_1561', description='ribosomal protein S12 pseudogene 30', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58899355, end_position_on_the_genomic_accession=58899692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  163087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163087, CurrentID=0, Status='live', Symbol='ZNF383', Aliases='HSD17', description='zinc finger protein 383', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37217926, end_position_on_the_genomic_accession=37243926, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  163088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163088, CurrentID=0, Status='live', Symbol='TAF9P3', Aliases='TAF2GL', description='TATA-box binding protein associated factor 9 pseudogene 3', other_designations='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene 3|TBP-associated factor 9 pseudogene', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=30037225, end_position_on_the_genomic_accession=30038390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442244, CurrentID=0, Status='live', Symbol='LOC442244', Aliases='', description='PDGFA associated protein 1 pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110923569, end_position_on_the_genomic_accession=110924108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100302111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302111, CurrentID=0, Status='live', Symbol='MIR1184-1', Aliases='MIR1184, MIRN1184, hsa-mir-1184, hsa-mir-1184-1', description='microRNA 1184-1', other_designations='microRNA 1184', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154887360, end_position_on_the_genomic_accession=154887458, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302112, CurrentID=0, Status='live', Symbol='MIR1284', Aliases='MIRN1284, hsa-mir-1284, mir-1284', description='microRNA 1284', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71541970, end_position_on_the_genomic_accession=71542089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302113, CurrentID=0, Status='live', Symbol='MIR1200', Aliases='MIRN1200, hsa-mir-1200', description='microRNA 1200', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36919357, end_position_on_the_genomic_accession=36919432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302114, CurrentID=0, Status='live', Symbol='MIR513C', Aliases='MIRN513C, hsa-mir-513c', description='microRNA 513c', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147189704, end_position_on_the_genomic_accession=147189787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302115, CurrentID=0, Status='live', Symbol='MIR1468', Aliases='MIRN1468, hsa-mir-1468, mir-1468', description='microRNA 1468', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63786002, end_position_on_the_genomic_accession=63786087, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302116, CurrentID=0, Status='live', Symbol='MIR1265', Aliases='MIRN1265, hsa-mir-1265', description='microRNA 1265', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14436576, end_position_on_the_genomic_accession=14436661, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302117, CurrentID=0, Status='live', Symbol='MIR320B1', Aliases='MIR320B-1, MIRN320B1, hsa-mir-320b-1, mir-320b-1', description='microRNA 320b-1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116671749, end_position_on_the_genomic_accession=116671827, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302118, CurrentID=0, Status='live', Symbol='MIR1286', Aliases='MIRN1286, hsa-mir-1286, mir-1286', description='microRNA 1286', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20249134, end_position_on_the_genomic_accession=20249211, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  158257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158257, CurrentID=0, Status='live', Symbol='MIRLET7DHG', Aliases='', description='MIRLET7D host gene', other_designations='MIRLET7D host gene (non-protein coding)', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94176569, end_position_on_the_genomic_accession=94204566, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100302121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302121, CurrentID=0, Status='live', Symbol='MIR1276', Aliases='MIRN1276, hsa-mir-1276, mir-1276', description='microRNA 1276', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85770496, end_position_on_the_genomic_accession=85770578, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302122, CurrentID=0, Status='live', Symbol='MIR1183', Aliases='MIRN1183, hsa-mir-1183', description='microRNA 1183', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21471058, end_position_on_the_genomic_accession=21471146, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302123, CurrentID=0, Status='live', Symbol='MIR1275', Aliases='MIRN1275, hsa-mir-1275, mir-1275', description='microRNA 1275', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33999972, end_position_on_the_genomic_accession=34000051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302124, CurrentID=0, Status='live', Symbol='MIR1288', Aliases='MIRN1288, hsa-mir-1288', description='microRNA 1288', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16282014, end_position_on_the_genomic_accession=16282088, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302125, CurrentID=0, Status='live', Symbol='MIR1289-1', Aliases='MIRN1289-1, hsa-mir-1289-1', description='microRNA 1289-1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35453954, end_position_on_the_genomic_accession=35454097, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302126, CurrentID=0, Status='live', Symbol='MIR1471', Aliases='MIRN1471, hsa-mir-1471', description='microRNA 1471', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231892242, end_position_on_the_genomic_accession=231892298, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302127, CurrentID=0, Status='live', Symbol='MIR1470', Aliases='MIRN1470, hsa-mir-1470', description='microRNA 1470', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15449548, end_position_on_the_genomic_accession=15449608, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302128, CurrentID=0, Status='live', Symbol='MIR1302-3', Aliases='MIRN1302-3, hsa-mir-1302-3', description='microRNA 1302-3', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113582959, end_position_on_the_genomic_accession=113583096, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302129, CurrentID=0, Status='live', Symbol='MIR1915', Aliases='MIRN1915, hsa-mir-1915', description='microRNA 1915', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21496562, end_position_on_the_genomic_accession=21496641, orientation='minus', exon_count=1, OMIM=615202, no_hdr0=''),
  100302130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302130, CurrentID=0, Status='live', Symbol='MIR1302-4', Aliases='MIRN1302-4, hsa-mir-1302-4', description='microRNA 1302-4', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207269275, end_position_on_the_genomic_accession=207269424, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302131, CurrentID=0, Status='live', Symbol='MIR302F', Aliases='MIRN302F, hsa-mir-302f', description='microRNA 302f', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=30298910, end_position_on_the_genomic_accession=30298960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302132, CurrentID=0, Status='live', Symbol='MIR1182', Aliases='MIRN1182, hsa-mir-1182', description='microRNA 1182', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231019828, end_position_on_the_genomic_accession=231019924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302133, CurrentID=0, Status='live', Symbol='MIR1287', Aliases='MIRN1287, hsa-mir-1287, mir-1287', description='microRNA 1287', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98395218, end_position_on_the_genomic_accession=98395307, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  163126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163126, CurrentID=0, Status='live', Symbol='EID2', Aliases='CRI2, EID-2', description='EP300 interacting inhibitor of differentiation 2', other_designations='CREBBP/EP300 inhibitor 2|CREBBP/EP300 inhibitory protein 2|E1A like inhibitor of differentiation 2|EID-1-like inhibitor of differentiation 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39538806, end_position_on_the_genomic_accession=39540198, orientation='minus', exon_count=1, OMIM=609773, no_hdr0=''),
  100302135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302135, CurrentID=0, Status='live', Symbol='MIR320C1', Aliases='MIR320C-1, MIRN320C1, hsa-mir-320c-1, mir-320c-1', description='microRNA 320c-1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21683510, end_position_on_the_genomic_accession=21683597, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302136, CurrentID=0, Status='live', Symbol='MIR1252', Aliases='MIRN1252, hsa-mir-1252', description='microRNA 1252', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79419257, end_position_on_the_genomic_accession=79419321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302137, CurrentID=0, Status='live', Symbol='MIR1914', Aliases='MIRN1914, hsa-mir-1914', description='microRNA 1914', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63941465, end_position_on_the_genomic_accession=63941544, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302138, CurrentID=0, Status='live', Symbol='MIR1292', Aliases='MIRN1292, hsa-mir-1292, mir-1292', description='microRNA 1292', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2652777, end_position_on_the_genomic_accession=2652842, orientation='plus', exon_count=1, OMIM=614155, no_hdr0=''),
  163131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163131, CurrentID=0, Status='live', Symbol='ZNF780B', Aliases='ZNF779', description='zinc finger protein 780B', other_designations='zinc finger protein 779', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40028260, end_position_on_the_genomic_accession=40059027, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100302140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302140, CurrentID=0, Status='live', Symbol='MIR1302-6', Aliases='MIRN1302-6, hsa-mir-1302-6', description='microRNA 1302-6', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18127220, end_position_on_the_genomic_accession=18127309, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302141, CurrentID=0, Status='live', Symbol='MIR1913', Aliases='MIRN1913, hsa-mir-1913', description='microRNA 1913', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166509354, end_position_on_the_genomic_accession=166509433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302142, CurrentID=0, Status='live', Symbol='MIR1246', Aliases='MIRN1246, hsa-mir-1246', description='microRNA 1246', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176600980, end_position_on_the_genomic_accession=176601052, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302143, CurrentID=0, Status='live', Symbol='MIR1248', Aliases='MIRN1248, hsa-mir-1248, mir-1248', description='microRNA 1248', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186786672, end_position_on_the_genomic_accession=186786777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302144, CurrentID=0, Status='live', Symbol='MIR1912', Aliases='MIRN1912, hsa-mir-1912, mir-1912', description='microRNA 1912', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114651544, end_position_on_the_genomic_accession=114651623, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302145, CurrentID=0, Status='live', Symbol='MIR1247', Aliases='MIRN1247, hsa-mir-1247, mir-1247', description='microRNA 1247', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101560287, end_position_on_the_genomic_accession=101560422, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302146, CurrentID=0, Status='live', Symbol='MIR1302-5', Aliases='MIRN1302-5, hsa-mir-1302-5', description='microRNA 1302-5', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50614636, end_position_on_the_genomic_accession=50614785, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302147, CurrentID=0, Status='live', Symbol='MIR1302-7', Aliases='MIRN1302-7, hsa-mir-1302-7', description='microRNA 1302-7', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141786242, end_position_on_the_genomic_accession=141786313, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302148, CurrentID=0, Status='live', Symbol='MIR1263', Aliases='MIRN1263, hsa-mir-1263', description='microRNA 1263', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164171471, end_position_on_the_genomic_accession=164171556, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302149, CurrentID=0, Status='live', Symbol='MIR1249', Aliases='MIRN1249, hsa-mir-1249', description='microRNA 1249', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45200954, end_position_on_the_genomic_accession=45201019, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302150, CurrentID=0, Status='live', Symbol='MIR1296', Aliases='MIRN1296, hsa-mir-1296, mir-1296', description='microRNA 1296', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63372957, end_position_on_the_genomic_accession=63373048, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302152, CurrentID=0, Status='live', Symbol='MIR548N', Aliases='MIRN548N, hsa-mir-548n', description='microRNA 548n', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34940760, end_position_on_the_genomic_accession=34940834, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302153, CurrentID=0, Status='live', Symbol='MIR1298', Aliases='MIRN1298, hsa-mir-1298', description='microRNA 1298', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114715233, end_position_on_the_genomic_accession=114715344, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271557, CurrentID=0, Status='live', Symbol='RPS15AP22', Aliases='RPS15A_11_848', description='ribosomal protein S15a pseudogene 22', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131702224, end_position_on_the_genomic_accession=131702681, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302155, CurrentID=0, Status='live', Symbol='MIR1256', Aliases='MIRN1256, hsa-mir-1256', description='microRNA 1256', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20988314, end_position_on_the_genomic_accession=20988432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302156, CurrentID=0, Status='live', Symbol='MIR1229', Aliases='MIRN1229, hsa-mir-1229', description='microRNA 1229', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179798278, end_position_on_the_genomic_accession=179798346, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302157, CurrentID=0, Status='live', Symbol='MIR1185-1', Aliases='MIRN1185-1, hsa-mir-1185-1, mir-1185-1', description='microRNA 1185-1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101042977, end_position_on_the_genomic_accession=101043062, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302158, CurrentID=0, Status='live', Symbol='MIR1231', Aliases='MIRN1231, hsa-mir-1231', description='microRNA 1231', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201808611, end_position_on_the_genomic_accession=201808702, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302159, CurrentID=0, Status='live', Symbol='MIR548F3', Aliases='MIR548F-3, MIRN548F3, hsa-mir-548f-3', description='microRNA 548f-3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110513829, end_position_on_the_genomic_accession=110513915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302160, CurrentID=0, Status='live', Symbol='MIR1233-1', Aliases='MIR1233, MIRN1233, hsa-mir-1233, hsa-mir-1233-1', description='microRNA 1233-1', other_designations='microRNA 1233', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34382069, end_position_on_the_genomic_accession=34382150, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302161, CurrentID=0, Status='live', Symbol='MIR1205', Aliases='MIRN1205, hsa-mir-1205', description='microRNA 1205', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127960633, end_position_on_the_genomic_accession=127960695, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  163154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163154, CurrentID=0, Status='live', Symbol='PRR22', Aliases='', description='proline rich 22', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5782960, end_position_on_the_genomic_accession=5784765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100302163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302163, CurrentID=0, Status='live', Symbol='MIR1278', Aliases='MIRN1278, hsa-mir-1278, mir-1278', description='microRNA 1278', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193136503, end_position_on_the_genomic_accession=193136583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302164, CurrentID=0, Status='live', Symbol='MIR2113', Aliases='hsa-mir-2113', description='microRNA 2113', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=98024531, end_position_on_the_genomic_accession=98024621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302165, CurrentID=0, Status='live', Symbol='MIR1273A', Aliases='MIR1273, MIRN1273, hsa-mir-1273, hsa-mir-1273a', description='microRNA 1273a', other_designations='microRNA 1273', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100023982, end_position_on_the_genomic_accession=100024084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302166, CurrentID=0, Status='live', Symbol='MIR1322', Aliases='MIRN1322, hsa-mir-1322', description='microRNA 1322', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10825373, end_position_on_the_genomic_accession=10825443, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302167, CurrentID=0, Status='live', Symbol='MIR1299', Aliases='MIRN1299, hsa-mir-1299', description='microRNA 1299', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40929010, end_position_on_the_genomic_accession=40929092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302168, CurrentID=0, Status='live', Symbol='MIR1257', Aliases='MIRN1257, hsa-mir-1257', description='microRNA 1257', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61953546, end_position_on_the_genomic_accession=61953662, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302169, CurrentID=0, Status='live', Symbol='MIR320D2', Aliases='MIR320D-2, MIRN320D2, hsa-mir-320d-2, mir-320d-2', description='microRNA 320d-2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140926172, end_position_on_the_genomic_accession=140926219, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302170, CurrentID=0, Status='live', Symbol='MIR1206', Aliases='MIRN1206, hsa-mir-1206', description='microRNA 1206', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128008898, end_position_on_the_genomic_accession=128008956, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302171, CurrentID=0, Status='live', Symbol='MIR1321', Aliases='MIRN1321, hsa-mir-1321', description='microRNA 1321', other_designations='', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85835780, end_position_on_the_genomic_accession=85835858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302172, CurrentID=0, Status='live', Symbol='MIR1258', Aliases='MIRN1258, hsa-mir-1258, mir-1258', description='microRNA 1258', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179860836, end_position_on_the_genomic_accession=179860908, orientation='minus', exon_count=1, OMIM=614488, no_hdr0=''),
  100302174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302174, CurrentID=0, Status='live', Symbol='MIR1307', Aliases='MIRN1307, hsa-mir-1307, mir-1307', description='microRNA 1307', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103394253, end_position_on_the_genomic_accession=103394401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302175, CurrentID=0, Status='live', Symbol='MIR1207', Aliases='MIRN1207, hsa-mir-1207', description='microRNA 1207', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128049152, end_position_on_the_genomic_accession=128049238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302177, CurrentID=0, Status='live', Symbol='MIR1269A', Aliases='MIR1269, MIRN1269, hsa-mir-1269, hsa-mir-1269a', description='microRNA 1269a', other_designations='microRNA 1269', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=66276824, end_position_on_the_genomic_accession=66276928, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302178, CurrentID=0, Status='live', Symbol='MIR1295A', Aliases='MIR1295, MIRN1295, hsa-mir-1295, hsa-mir-1295a, mir-1295a', description='microRNA 1295a', other_designations='microRNA 1295', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171101728, end_position_on_the_genomic_accession=171101806, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302179, CurrentID=0, Status='live', Symbol='MIR1270', Aliases='MIR1270-1-2, MIRN1270, hsa-mir-1270, hsa-mir-1270-1, mir-1270, MIR1270', description='microRNA 1270', other_designations='hsa-mir-1270-2|microRNA 1270-1|microRNA 1270-2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20399272, end_position_on_the_genomic_accession=20399354, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302181, CurrentID=0, Status='live', Symbol='MIR1294', Aliases='MIRN1294, hsa-mir-1294, mir-1294', description='microRNA 1294', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154347106, end_position_on_the_genomic_accession=154347247, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302182, CurrentID=0, Status='live', Symbol='MIR1279', Aliases='MIRN1279, hsa-mir-1279', description='microRNA 1279', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69273157, end_position_on_the_genomic_accession=69273218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  163175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163175, CurrentID=0, Status='live', Symbol='LGI4', Aliases='LGIL3', description='leucine-rich repeat LGI family member 4', other_designations='LGI1-like protein 3|leucine-rich glioma-inactivated gene 4|leucine-rich glioma-inactivated protein 4|leucine-rich repeat LGI family, member 4', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35124513, end_position_on_the_genomic_accession=35139451, orientation='minus', exon_count=11, OMIM=608303, no_hdr0=''),
  100302184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302184, CurrentID=0, Status='live', Symbol='MIR1272', Aliases='MIRN1272, hsa-mir-1272, mir-1272', description='microRNA 1272', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64762387, end_position_on_the_genomic_accession=64762515, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302185, CurrentID=0, Status='live', Symbol='MIR1204', Aliases='hsa-mir-1204', description='microRNA 1204', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127795962, end_position_on_the_genomic_accession=127796028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302186, CurrentID=0, Status='live', Symbol='MIR548I3', Aliases='MIR548I-3, MIRN548I3, hsa-mir-548i-3, mir-548i-3', description='microRNA 548i-3', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8088941, end_position_on_the_genomic_accession=8089089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302187, CurrentID=0, Status='live', Symbol='MIR1297', Aliases='MIRN1297, hsa-mir-1297, mir-1297', description='microRNA 1297', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54311972, end_position_on_the_genomic_accession=54312048, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302188, CurrentID=0, Status='live', Symbol='MIR1243', Aliases='MIRN1243, hsa-mir-1243', description='microRNA 1243', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113106863, end_position_on_the_genomic_accession=113106955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442258, CurrentID=0, Status='live', Symbol='EEF1DP5', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 5', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128580065, end_position_on_the_genomic_accession=128580952, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302190, CurrentID=0, Status='live', Symbol='MIR1976', Aliases='hsa-mir-1976', description='microRNA 1976', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26554542, end_position_on_the_genomic_accession=26554593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  163183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163183, CurrentID=0, Status='live', Symbol='SYNE4', Aliases='C19orf46, DFNB76, Nesp4', description='spectrin repeat containing, nuclear envelope family member 4', other_designations='deafness, autosomal recessive 76|nuclear envelope spectrin repeat protein 4', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36003000, end_position_on_the_genomic_accession=36008793, orientation='minus', exon_count=8, OMIM=615535, no_hdr0=''),
  100302192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302192, CurrentID=0, Status='live', Symbol='MIR548F1', Aliases='MIR548F-1, MIRN548F1, hsa-mir-548f-1', description='microRNA 548f-1', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=54607874, end_position_on_the_genomic_accession=54607957, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302193, CurrentID=0, Status='live', Symbol='MIR1255A', Aliases='MIRN1255A, hsa-mir-1255a', description='microRNA 1255a', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101330302, end_position_on_the_genomic_accession=101330414, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302195, CurrentID=0, Status='live', Symbol='MIR320C2', Aliases='MIR320C-2, MIRN320C2, hsa-mir-320c-2, mir-320c-2', description='microRNA 320c-2', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24321686, end_position_on_the_genomic_accession=24321735, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302196, CurrentID=0, Status='live', Symbol='MIR1234', Aliases='MIRN1234, hsa-mir-1234', description='microRNA 1234', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144400086, end_position_on_the_genomic_accession=144400165, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302197, CurrentID=0, Status='live', Symbol='MIR1306', Aliases='MIRN1306, hsa-mir-1306, mir-1306', description='microRNA 1306', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20086058, end_position_on_the_genomic_accession=20086142, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302201, CurrentID=0, Status='live', Symbol='MIR1228', Aliases='MIRN1228, hsa-mir-1228, mir-1228', description='microRNA 1228', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57194504, end_position_on_the_genomic_accession=57194576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302202, CurrentID=0, Status='live', Symbol='MIR1266', Aliases='MIRN1266, hsa-mir-1266, mir-1266', description='microRNA 1266', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52277117, end_position_on_the_genomic_accession=52277200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302203, CurrentID=0, Status='live', Symbol='MIR1271', Aliases='MIRN1271, hsa-mir-1271', description='microRNA 1271', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176367946, end_position_on_the_genomic_accession=176368031, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302204, CurrentID=0, Status='live', Symbol='MIR548I1', Aliases='MIR548I-1, MIRN548I1, hsa-mir-548i-1, mir-548i-1', description='microRNA 548i-1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125790404, end_position_on_the_genomic_accession=125790552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302205, CurrentID=0, Status='live', Symbol='MIR1283-2', Aliases='MIRN1283-2, hsa-mir-1283-2, mir-1283-2', description='microRNA 1283-2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53758232, end_position_on_the_genomic_accession=53758318, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373683, CurrentID=0, Status='live', Symbol='LOC105373683', Aliases='', description='uncharacterized LOC105373683', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150234658, end_position_on_the_genomic_accession=150257007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100302208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302208, CurrentID=0, Status='live', Symbol='MIR1253', Aliases='MIRN1253, hsa-mir-1253', description='microRNA 1253', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2748078, end_position_on_the_genomic_accession=2748182, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302209, CurrentID=0, Status='live', Symbol='MIR1185-2', Aliases='MIRN1185-2, hsa-mir-1185-2, mir-1185-2', description='microRNA 1185-2', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101044198, end_position_on_the_genomic_accession=101044283, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302210, CurrentID=0, Status='live', Symbol='MIR1909', Aliases='MIRN1909, hsa-mir-1909, mir-1909', description='microRNA 1909', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1816159, end_position_on_the_genomic_accession=1816238, orientation='minus', exon_count=1, OMIM=615201, no_hdr0=''),
  100302211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302211, CurrentID=0, Status='live', Symbol='MIR1203', Aliases='MIRN1203, hsa-mir-1203', description='microRNA 1203', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48156427, end_position_on_the_genomic_accession=48156511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302212, CurrentID=0, Status='live', Symbol='MIR1324', Aliases='MIRN1324, hsa-mir-1324', description='microRNA 1324', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75630763, end_position_on_the_genomic_accession=75630858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302213, CurrentID=0, Status='live', Symbol='MIR1181', Aliases='MIRN1181, hsa-mir-1181', description='microRNA 1181', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10403458, end_position_on_the_genomic_accession=10403538, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302214, CurrentID=0, Status='live', Symbol='MIR1277', Aliases='MIRN1277, hsa-mir-1277', description='microRNA 1277', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118386394, end_position_on_the_genomic_accession=118386471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302217, CurrentID=0, Status='live', Symbol='MIR1827', Aliases='MIRN1827, hsa-mir-1827', description='microRNA 1827', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100189884, end_position_on_the_genomic_accession=100189949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302218, CurrentID=0, Status='live', Symbol='MIR1285-1', Aliases='MIRN1285-1, hsa-mir-1285-1', description='microRNA 1285-1', other_designations='', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92204015, end_position_on_the_genomic_accession=92204098, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302219, CurrentID=0, Status='live', Symbol='MIR1245A', Aliases='MIR1245, MIRN1245, hsa-mir-1245, hsa-mir-1245a, mir-1245a', description='microRNA 1245a', other_designations='microRNA 1245', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188978092, end_position_on_the_genomic_accession=188978161, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302220, CurrentID=0, Status='live', Symbol='MIR1293', Aliases='MIRN1293, hsa-mir-1293, mir-1293', description='microRNA 1293', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50234142, end_position_on_the_genomic_accession=50234212, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302221, CurrentID=0, Status='live', Symbol='MIR1291', Aliases='MIRN1291, hsa-mir-1291, mir-1291', description='microRNA 1291', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48654444, end_position_on_the_genomic_accession=48654530, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302222, CurrentID=0, Status='live', Symbol='MIR1911', Aliases='MIRN1911, hsa-mir-1911, mir-1911', description='microRNA 1911', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114763184, end_position_on_the_genomic_accession=114763263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302223, CurrentID=0, Status='live', Symbol='MIR1302-8', Aliases='MIRN1302-8, hsa-mir-1302-8', description='microRNA 1302-8', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97363554, end_position_on_the_genomic_accession=97363681, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302224, CurrentID=0, Status='live', Symbol='MIR2110', Aliases='hsa-mir-2110, mir-2110', description='microRNA 2110', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114174105, end_position_on_the_genomic_accession=114174179, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302225, CurrentID=0, Status='live', Symbol='MIR2053', Aliases='hsa-mir-2053', description='microRNA 2053', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=112643493, end_position_on_the_genomic_accession=112643583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302226, CurrentID=0, Status='live', Symbol='MIR1238', Aliases='MIRN1238, hsa-mir-1238', description='microRNA 1238', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10552122, end_position_on_the_genomic_accession=10552204, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302227, CurrentID=0, Status='live', Symbol='MIR1302-1', Aliases='MIRN1302-1, hsa-mir-1302-1', description='microRNA 1302-1', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112695034, end_position_on_the_genomic_accession=112695176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302228, CurrentID=0, Status='live', Symbol='MIR1261', Aliases='MIRN1261, hsa-mir-1261', description='microRNA 1261', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90869121, end_position_on_the_genomic_accession=90869202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302229, CurrentID=0, Status='live', Symbol='MIR1250', Aliases='MIRN1250, hsa-mir-1250, mir-1250', description='microRNA 1250', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81133196, end_position_on_the_genomic_accession=81133308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  163223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163223, CurrentID=0, Status='live', Symbol='ZNF676', Aliases='', description='zinc finger protein 676', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22178936, end_position_on_the_genomic_accession=22248018, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100302232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302232, CurrentID=0, Status='live', Symbol='MIR1226', Aliases='MIRN1226, hsa-mir-1226, mir-1226', description='microRNA 1226', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47849555, end_position_on_the_genomic_accession=47849629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302233, CurrentID=0, Status='live', Symbol='MIR1268A', Aliases='MIR1268, MIRN1268, hsa-mir-1268, hsa-mir-1268a', description='microRNA 1268a', other_designations='microRNA 1268', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22225278, end_position_on_the_genomic_accession=22225329, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302234, CurrentID=0, Status='live', Symbol='MIR664A', Aliases='MIR664, MIRN664, hsa-mir-664, hsa-mir-664a, mir-664a', description='microRNA 664a', other_designations='microRNA 664', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220200538, end_position_on_the_genomic_accession=220200619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302235, CurrentID=0, Status='live', Symbol='MIR1179', Aliases='MIRN1179, hsa-mir-1179, mir-1179', description='microRNA 1179', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88608107, end_position_on_the_genomic_accession=88608197, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302236, CurrentID=0, Status='live', Symbol='MIR1260A', Aliases='MIR1260, MIRN1260, hsa-mir-1260, hsa-mir-1260a, mir-1260a', description='microRNA 1260a', other_designations='microRNA 1260', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77266218, end_position_on_the_genomic_accession=77266290, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302237, CurrentID=0, Status='live', Symbol='MIR1281', Aliases='MIRN1281, hsa-mir-1281', description='microRNA 1281', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41092513, end_position_on_the_genomic_accession=41092566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302238, CurrentID=0, Status='live', Symbol='MIR103B1', Aliases='MIR103-1AS, MIRN103-1AS, hsa-mir-103-1-as', description='microRNA 103b-1', other_designations='hsa-mir-103b-1|microRNA 103-1-as', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168560904, end_position_on_the_genomic_accession=168560965, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302239, CurrentID=0, Status='live', Symbol='MIR548F5', Aliases='MIR548F-5, MIRN548F5, hsa-mir-548f-5', description='microRNA 548f-5', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=32641474, end_position_on_the_genomic_accession=32641559, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302240, CurrentID=0, Status='live', Symbol='MIR1304', Aliases='MIRN1304, hsa-mir-1304', description='microRNA 1304', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93733674, end_position_on_the_genomic_accession=93733764, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302242, CurrentID=0, Status='live', Symbol='MIR1236', Aliases='MIRN1236, hsa-mir-1236', description='microRNA 1236', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31956839, end_position_on_the_genomic_accession=31956940, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302243, CurrentID=0, Status='live', Symbol='MIR1972-1', Aliases='MIR1972, hsa-mir-1972, hsa-mir-1972-1', description='microRNA 1972-1', other_designations='microRNA 1972', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15010321, end_position_on_the_genomic_accession=15010397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271560, CurrentID=0, Status='live', Symbol='RPS20P28', Aliases='RPS20_13_1202', description='ribosomal protein S20 pseudogene 28', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7994770, end_position_on_the_genomic_accession=7995078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302246, CurrentID=0, Status='live', Symbol='MIR1301', Aliases='MIRN1301, hsa-mir-1301, mir-1301', description='microRNA 1301', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25328640, end_position_on_the_genomic_accession=25328721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302250, CurrentID=0, Status='live', Symbol='MIR1197', Aliases='MIRN1197, hsa-mir-1197, mir-1197', description='microRNA 1197', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101025564, end_position_on_the_genomic_accession=101025651, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302251, CurrentID=0, Status='live', Symbol='MIR1264', Aliases='MIRN1264, hsa-mir-1264', description='microRNA 1264', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114652655, end_position_on_the_genomic_accession=114652723, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996338, CurrentID=0, Status='live', Symbol='LOC100996338', Aliases='', description='uncharacterized LOC100996338', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54245544, end_position_on_the_genomic_accession=54270879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100302254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302254, CurrentID=0, Status='live', Symbol='MIR1282', Aliases='MIRN1282, hsa-mir-1282', description='microRNA 1282', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43793659, end_position_on_the_genomic_accession=43793759, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302255, CurrentID=0, Status='live', Symbol='MIR1323', Aliases='MIRN1323, hsa-mir-1323, mir-1323', description='microRNA 1323', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53671968, end_position_on_the_genomic_accession=53672040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302256, CurrentID=0, Status='live', Symbol='MIR1180', Aliases='MIRN1180, hsa-mir-1180, mir-1180', description='microRNA 1180', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19344506, end_position_on_the_genomic_accession=19344574, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302257, CurrentID=0, Status='live', Symbol='MIR1539', Aliases='MIRN1539, hsa-mir-1539', description='microRNA 1539', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49487373, end_position_on_the_genomic_accession=49487422, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302258, CurrentID=0, Status='live', Symbol='MIR1469', Aliases='MIRN1469, hsa-mir-1469', description='microRNA 1469', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96333261, end_position_on_the_genomic_accession=96333307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302259, CurrentID=0, Status='live', Symbol='MIR1202', Aliases='MIRN1202, hsa-mir-1202', description='microRNA 1202', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155946797, end_position_on_the_genomic_accession=155946879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302260, CurrentID=0, Status='live', Symbol='MIR2052', Aliases='hsa-mir-2052', description='microRNA 2052', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74705693, end_position_on_the_genomic_accession=74705747, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302261, CurrentID=0, Status='live', Symbol='MIR1910', Aliases='MIRN1910, hsa-mir-1910, mir-1910', description='microRNA 1910', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85741621, end_position_on_the_genomic_accession=85741700, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302263, CurrentID=0, Status='live', Symbol='MIR1908', Aliases='MIRN1908, hsa-mir-1908, mir-1908', description='microRNA 1908', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61815161, end_position_on_the_genomic_accession=61815240, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302265, CurrentID=0, Status='live', Symbol='MIR1283-1', Aliases='MIRN1283-1, hsa-mir-1283-1, mir-1283-1', description='microRNA 1283-1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53688481, end_position_on_the_genomic_accession=53688567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  163259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163259, CurrentID=0, Status='live', Symbol='DENND2C', Aliases='dJ1156J9.1', description='DENN/MADD domain containing 2C', other_designations='RP5-1156J9.1', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114584575, end_position_on_the_genomic_accession=114670111, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100302268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302268, CurrentID=0, Status='live', Symbol='MIR1285-2', Aliases='MIRN1285-2, hsa-mir-1285-2', description='microRNA 1285-2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70252918, end_position_on_the_genomic_accession=70253005, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302270, CurrentID=0, Status='live', Symbol='MIR1305', Aliases='MIRN1305, hsa-mir-1305, mir-1305', description='microRNA 1305', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182169293, end_position_on_the_genomic_accession=182169378, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302273, CurrentID=0, Status='live', Symbol='MIR1254-1', Aliases='MIR1254, MIRN1254, hsa-mir-1254, hsa-mir-1254-1', description='microRNA 1254-1', other_designations='microRNA 1254', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68759318, end_position_on_the_genomic_accession=68759414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302274, CurrentID=0, Status='live', Symbol='MIR1178', Aliases='MIRN1178, hsa-mir-1178', description='microRNA 1178', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119713634, end_position_on_the_genomic_accession=119713724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302275, CurrentID=0, Status='live', Symbol='MIR548L', Aliases='MIRN548L, hsa-mir-548l', description='microRNA 548l', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94466495, end_position_on_the_genomic_accession=94466580, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302276, CurrentID=0, Status='live', Symbol='MIR1290', Aliases='MIRN1290, hsa-mir-1290', description='microRNA 1290', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18897071, end_position_on_the_genomic_accession=18897148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302277, CurrentID=0, Status='live', Symbol='MIR548I2', Aliases='MIR548I-2, MIRN548I2, hsa-mir-548i-2, mir-548i-2', description='microRNA 548i-2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9556168, end_position_on_the_genomic_accession=9556316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302278, CurrentID=0, Status='live', Symbol='MIR1302-2', Aliases='MIRN1302-2, hsa-mir-1302-2', description='microRNA 1302-2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30366, end_position_on_the_genomic_accession=30503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302279, CurrentID=0, Status='live', Symbol='MIR1262', Aliases='MIRN1262, hsa-mir-1262, mir-1262', description='microRNA 1262', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68183518, end_position_on_the_genomic_accession=68183610, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302280, CurrentID=0, Status='live', Symbol='MIR1237', Aliases='MIRN1237, hsa-mir-1237', description='microRNA 1237', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64368602, end_position_on_the_genomic_accession=64368703, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302281, CurrentID=0, Status='live', Symbol='MIR1208', Aliases='MIRN1208, hsa-mir-1208', description='microRNA 1208', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128150116, end_position_on_the_genomic_accession=128150188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302282, CurrentID=0, Status='live', Symbol='MIR103B2', Aliases='MIR103-2-AS, MIR103-2AS, MIRN103-2AS, hsa-mir-103-2-as', description='microRNA 103b-2', other_designations='hsa-mir-103b-2|microRNA 103-2-as', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3917502, end_position_on_the_genomic_accession=3917563, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302283, CurrentID=0, Status='live', Symbol='MIR1227', Aliases='MIRN1227, hsa-mir-1227, mir-1227', description='microRNA 1227', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2234062, end_position_on_the_genomic_accession=2234149, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302284, CurrentID=0, Status='live', Symbol='MIR1303', Aliases='MIRN1303, hsa-mir-1303', description='microRNA 1303', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154685776, end_position_on_the_genomic_accession=154685861, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302285, CurrentID=0, Status='live', Symbol='MIR1244-1', Aliases='MIR1244, MIRN1244, hsa-mir-1244, hsa-mir-1244-1', description='microRNA 1244-1', other_designations='microRNA 1244', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231713314, end_position_on_the_genomic_accession=231713398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302286, CurrentID=0, Status='live', Symbol='MIR1267', Aliases='MIRN1267, hsa-mir-1267', description='microRNA 1267', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107531171, end_position_on_the_genomic_accession=107531248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302287, CurrentID=0, Status='live', Symbol='MIR548H3', Aliases='MIR548H-3, MIRN548H3, hsa-mir-548h-3', description='microRNA 548h-3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13543529, end_position_on_the_genomic_accession=13543646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302288, CurrentID=0, Status='live', Symbol='MIR548P', Aliases='MIRN548P, hsa-mir-548p', description='microRNA 548p', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100816482, end_position_on_the_genomic_accession=100816565, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302289, CurrentID=0, Status='live', Symbol='MIR1251', Aliases='MIRN1251, hsa-mir-1251', description='microRNA 1251', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97491909, end_position_on_the_genomic_accession=97491978, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302290, CurrentID=0, Status='live', Symbol='MIR1973', Aliases='hsa-mir-1973, mir-1973', description='microRNA 1973', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116299725, end_position_on_the_genomic_accession=116299768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302291, CurrentID=0, Status='live', Symbol='LOC100302291', Aliases='', description='forty-two-three domain containing 1 pseudogene', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174360496, end_position_on_the_genomic_accession=174368102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373686, CurrentID=0, Status='live', Symbol='LOC105373686', Aliases='', description='uncharacterized LOC105373686', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151043836, end_position_on_the_genomic_accession=151050581, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271562, CurrentID=0, Status='live', Symbol='RPS20P30', Aliases='RPS20_14_1295', description='ribosomal protein S20 pseudogene 30', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130216807, end_position_on_the_genomic_accession=130217157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419956, CurrentID=0, Status='live', Symbol='LOC100419956', Aliases='', description='kinesin family member 15 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22961248, end_position_on_the_genomic_accession=22963207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479527, CurrentID=0, Status='live', Symbol='RN7SL838P', Aliases='', description='RNA, 7SL, cytoplasmic 838, pseudogene', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271563, CurrentID=0, Status='live', Symbol='RPS20P31', Aliases='RPS20_15_1287', description='ribosomal protein S20 pseudogene 31', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120247460, end_position_on_the_genomic_accession=120247791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  1074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1074, CurrentID=0, Status='live', Symbol='CFL1P5', Aliases='CFLL1, CFLP5', description='cofilin 1 (non-muscle) pseudogene 5', other_designations='cofilin pseudogene 5|cofilin-like 1', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69312859, end_position_on_the_genomic_accession=69313953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158295, CurrentID=0, Status='live', Symbol='MGC24103', Aliases='', description='uncharacterized MGC24103', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=16525990, end_position_on_the_genomic_accession=16527703, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441891, CurrentID=0, Status='live', Symbol='WDR82P2', Aliases='WDR82C', description='WD repeat domain 82 pseudogene 2', other_designations='WDR82 pseudogene 2', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91531954, end_position_on_the_genomic_accession=91535869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163351, CurrentID=0, Status='live', Symbol='GBP6', Aliases='', description='guanylate binding protein family member 6', other_designations='GBP-6|GTP-binding protein 6|guanine nucleotide-binding protein 6|guanylate binding protein family, member 6', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89363877, end_position_on_the_genomic_accession=89388160, orientation='plus', exon_count=11, OMIM=612467, no_hdr0=''),
  100271564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271564, CurrentID=0, Status='live', Symbol='RPS20P33', Aliases='RPS20_16_1416', description='ribosomal protein S20 pseudogene 33', other_designations='', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105374621, end_position_on_the_genomic_accession=105374971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369324, CurrentID=0, Status='live', Symbol='LOC105369324', Aliases='', description='uncharacterized LOC105369324', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61035351, end_position_on_the_genomic_accession=61041355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442290, CurrentID=0, Status='live', Symbol='NHP2P2', Aliases='TCAG_30283', description='NHP2 ribonucleoprotein pseudogene 2', other_designations='NHP2 ribonucleoprotein homolog pseudogene 2|nucleolar protein family A, member 2 pseudogene 2', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27048135, end_position_on_the_genomic_accession=27048879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442291, CurrentID=0, Status='live', Symbol='PSMC1P2', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 2', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27461685, end_position_on_the_genomic_accession=27463189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442292, CurrentID=0, Status='live', Symbol='LOC442292', Aliases='', description='coiled-coil domain containing 86 pseudogene', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27491682, end_position_on_the_genomic_accession=27492763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271565, CurrentID=0, Status='live', Symbol='RPS20P34', Aliases='RPS20_17_1430', description='ribosomal protein S20 pseudogene 34', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50466620, end_position_on_the_genomic_accession=50467138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441904, CurrentID=0, Status='live', Symbol='LOC441904', Aliases='', description='autoantigen La-like', other_designations='Sjogren syndrome antigen B (autoantigen La) pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147082034, end_position_on_the_genomic_accession=147083650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442293, CurrentID=0, Status='live', Symbol='LOC442293', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35765658, end_position_on_the_genomic_accession=35768652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302401, CurrentID=0, Status='live', Symbol='RASAL2-AS1', Aliases='', description='RASAL2 antisense RNA 1', other_designations='RASAL2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178091508, end_position_on_the_genomic_accession=178093993, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369325, CurrentID=0, Status='live', Symbol='LOC105369325', Aliases='', description='uncharacterized LOC105369325', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61029759, end_position_on_the_genomic_accession=61092441, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  163404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163404, CurrentID=0, Status='live', Symbol='PLPPR5', Aliases='LPPR5, PAP2, PAP2D, PRG5', description='phospholipid phosphatase related 5', other_designations='PRG-5|phosphatidic acid phosphatase 2d|phosphatidic acid phosphatase type 2|plasticity-related gene 5 protein|plasticity-related protein 5', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98890245, end_position_on_the_genomic_accession=99005809, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  441893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441893, CurrentID=0, Status='live', Symbol='GAPDHP29', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 29', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94301970, end_position_on_the_genomic_accession=94303228, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642661, CurrentID=0, Status='live', Symbol='LOC642661', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog (yeast) pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19267016, end_position_on_the_genomic_accession=19268560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271566, CurrentID=0, Status='live', Symbol='RPS20P35', Aliases='RPS20_18_1544', description='ribosomal protein S20 pseudogene 35', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38855194, end_position_on_the_genomic_accession=38855724, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369326, CurrentID=0, Status='live', Symbol='LOC105369326', Aliases='', description='uncharacterized LOC105369326', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61003721, end_position_on_the_genomic_accession=61008519, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616364, CurrentID=0, Status='live', Symbol='MIR4785', Aliases='mir-4785', description='microRNA 4785', other_designations='hsa-mir-4785', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160407810, end_position_on_the_genomic_accession=160407882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479551, CurrentID=0, Status='live', Symbol='RNU6ATAC18P', Aliases='', description='RNA, U6atac small nuclear 18, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158314, CurrentID=0, Status='live', Symbol='LINC00475', Aliases='C9orf44', description='long intergenic non-protein coding RNA 475', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92141467, end_position_on_the_genomic_accession=92159608, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369327, CurrentID=0, Status='live', Symbol='LOC105369327', Aliases='', description='uncharacterized LOC105369327', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61174663, end_position_on_the_genomic_accession=61199441, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  442304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442304, CurrentID=0, Status='live', Symbol='FTLP15', Aliases='', description='ferritin, light polypeptide pseudogene 15', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45997381, end_position_on_the_genomic_accession=45998274, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132014, CurrentID=0, Status='live', Symbol='IL17RE', Aliases='', description='interleukin 17 receptor E', other_designations='IL-17 receptor E', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9902062, end_position_on_the_genomic_accession=9916402, orientation='plus', exon_count=21, OMIM=614995, no_hdr0=''),
  642663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642663, CurrentID=0, Status='live', Symbol='CICP17', Aliases='', description='capicua transcriptional repressor pseudogene 17', other_designations='protein capicua homolog', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51382245, end_position_on_the_genomic_accession=51384089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56954, CurrentID=0, Status='live', Symbol='NIT2', Aliases='HEL-S-8a', description='nitrilase family member 2', other_designations='Nit protein 2|epididymis secretory sperm binding protein Li 8a|nitrilase family, member 2|nitrilase homolog 2', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100334718, end_position_on_the_genomic_accession=100355634, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100271568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271568, CurrentID=0, Status='live', Symbol='RPS26P42', Aliases='RPS26_14_1066', description='ribosomal protein S26 pseudogene 42', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75125616, end_position_on_the_genomic_accession=75125871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347422, CurrentID=0, Status='live', Symbol='LOC347422', Aliases='', description='tRNA methyltransferase 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145630119, end_position_on_the_genomic_accession=145632440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  163479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163479, CurrentID=0, Status='live', Symbol='FNDC7', Aliases='', description='fibronectin type III domain containing 7', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108712416, end_position_on_the_genomic_accession=108743262, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  442308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442308, CurrentID=0, Status='live', Symbol='TUBBP6', Aliases='', description='tubulin beta class I pseudogene 6', other_designations='tubulin, beta class I pseudogene 6|tubulin, beta pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55645466, end_position_on_the_genomic_accession=55647121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  163486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163486, CurrentID=0, Status='live', Symbol='DENND1B', Aliases='C1ORF18, C1orf218, FAM31B', description='DENN/MADD domain containing 1B', other_designations='connecdenn 2|family with sequence similarity 31, member B', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197504748, end_position_on_the_genomic_accession=197782175, orientation='minus', exon_count=30, OMIM=613292, no_hdr0=''),
  442309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442309, CurrentID=0, Status='live', Symbol='LOC442309', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56163113, end_position_on_the_genomic_accession=56163703, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106478979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478979, CurrentID=0, Status='live', Symbol='A2ML1-AS2', Aliases='', description='A2ML1 antisense RNA 2', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302509, CurrentID=0, Status='live', Symbol='LECD', Aliases='', description='Corneal dystrophy, Lisch epithelial', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300778, no_hdr0=''),
  100302510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302510, CurrentID=0, Status='live', Symbol='XECD', Aliases='', description='Corneal dystrophy, endothelial, X-linked', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300779, no_hdr0=''),
  100302511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302511, CurrentID=0, Status='live', Symbol='EOE1', Aliases='EE', description='Esophagitis, eosinophilic', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610247, no_hdr0=''),
  100302512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302512, CurrentID=0, Status='live', Symbol='RLS6', Aliases='', description='Restless legs syndrome, susceptibility to, 6', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611185, no_hdr0=''),
  100302513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302513, CurrentID=0, Status='live', Symbol='DFNB71', Aliases='', description='deafness, autosomal recessive 71', other_designations='', map_location='8p22-p21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612789, no_hdr0=''),
  100302514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302514, CurrentID=0, Status='live', Symbol='NRCLP5', Aliases='', description='Narcolepsy 5', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612851, no_hdr0=''),
  100302515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302515, CurrentID=0, Status='live', Symbol='OFC12', Aliases='', description='Orofacial cleft 12', other_designations='', map_location='8q24.31', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612858, no_hdr0=''),
  100302516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302516, CurrentID=0, Status='live', Symbol='CTEPH1', Aliases='', description='Pulmonary hypertension, chronic thromboembolic, without deep vein thrombosis, susceptibility to', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612862, no_hdr0=''),
  100302517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302517, CurrentID=0, Status='live', Symbol='MENOQ2', Aliases='', description='Menopause, natural, age at, QTL2', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612884, no_hdr0=''),
  100302518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302518, CurrentID=0, Status='live', Symbol='MENOQ3', Aliases='', description='Menopause, natural, age at, QTL3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612885, no_hdr0=''),
  100302519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302519, CurrentID=0, Status='live', Symbol='MENOQ4', Aliases='', description='Menopause, natural, age at, QTL4', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612886, no_hdr0=''),
  100302522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302522, CurrentID=0, Status='live', Symbol='PTCSC1', Aliases='NCRNA00197, PTCSC', description='papillary thyroid carcinoma susceptibility candidate 1 (non-protein coding)', other_designations='papillary thyroid carcinoma susceptibility candidate (non-protein coding)', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302526, CurrentID=0, Status='live', Symbol='FAM41AY2', Aliases='', description='family with sequence similarity 41 member A, Y-linked 2', other_designations='family with sequence similarity 41, member A, Y-linked 2', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18389270, end_position_on_the_genomic_accession=18405046, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27254, CurrentID=0, Status='live', Symbol='CSDC2', Aliases='PIPPIN, dJ347H13.2', description='cold shock domain containing C2, RNA binding', other_designations='RNA-binding protein pippin', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41561010, end_position_on_the_genomic_accession=41576666, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  442317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442317, CurrentID=0, Status='live', Symbol='SAPCD2P2', Aliases='', description='suppressor APC domain containing 2 pseudogene 2', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57424419, end_position_on_the_genomic_accession=57425879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271570, CurrentID=0, Status='live', Symbol='RPS26P45', Aliases='RPS26_17_1256', description='ribosomal protein S26 pseudogene 45', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69422677, end_position_on_the_genomic_accession=69422995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647030, CurrentID=0, Status='live', Symbol='EEF1B2P6', Aliases='', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 6', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131661900, end_position_on_the_genomic_accession=131662647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479567, CurrentID=0, Status='live', Symbol='RNU4-20P', Aliases='', description='RNA, U4 small nuclear 20, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302559, CurrentID=0, Status='live', Symbol='PURAQTL1', Aliases='', description='Polyunsaturated fatty acids plasma level QTL1', other_designations='', map_location='11q12-q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612795, no_hdr0=''),
  100302560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302560, CurrentID=0, Status='live', Symbol='MUHH2', Aliases='', description='Hypotrichosis, hereditary, Marie Unna type, 2', other_designations='', map_location='1p21.1-q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612841, no_hdr0=''),
  100302561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302561, CurrentID=0, Status='live', Symbol='RLS7', Aliases='', description='Restless legs syndrome 7', other_designations='', map_location='2p14-p13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612853, no_hdr0=''),
  100302562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302562, CurrentID=0, Status='live', Symbol='MENAQ2', Aliases='', description='Menarche, age at, QTL2', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612882, no_hdr0=''),
  100302563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302563, CurrentID=0, Status='live', Symbol='MENAQ3', Aliases='', description='Menarche, age at, QTL3', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612883, no_hdr0=''),
  100616368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616368, CurrentID=0, Status='live', Symbol='MIR4688', Aliases='mir-4688', description='microRNA 4688', other_designations='hsa-mir-4688', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46376402, end_position_on_the_genomic_accession=46376484, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369331, CurrentID=0, Status='live', Symbol='LOC105369331', Aliases='', description='uncharacterized LOC105369331', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61581406, end_position_on_the_genomic_accession=61587840, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100302591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302591, CurrentID=0, Status='live', Symbol='LOC100302591', Aliases='', description='TBC1 domain family member', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131907, CurrentID=0, Status='live', Symbol='LOC100131907', Aliases='', description='uncharacterized LOC100131907', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=377284, end_position_on_the_genomic_accession=378112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  163589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163589, CurrentID=0, Status='live', Symbol='TDRD5', Aliases='TUDOR3', description='tudor domain containing 5', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179591613, end_position_on_the_genomic_accession=179691272, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  163590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163590, CurrentID=0, Status='live', Symbol='TOR1AIP2', Aliases='IFRG15, LULL1, NET9', description='torsin A interacting protein 2', other_designations='15 kDa interferon-responsive protein|interferon responsive gene 15|lumenal domain like LAP1|lumenal domain-like LAP1', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179839967, end_position_on_the_genomic_accession=179877806, orientation='minus', exon_count=9, OMIM=614513, no_hdr0=''),
  102723771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723771, CurrentID=0, Status='live', Symbol='LOC102723771', Aliases='', description='uncharacterized LOC102723771', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144434, CurrentID=0, Status='live', Symbol='ETL4', Aliases='EPOLM, ETOLM', description='epilepsy, occipitotemporal lobe, and migraine with aura', other_designations='', map_location='9q21-q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611631, no_hdr0=''),
  100302640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302640, CurrentID=0, Status='live', Symbol='LINC00882', Aliases='', description='long intergenic non-protein coding RNA 882', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107109790, end_position_on_the_genomic_accession=107240638, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100144435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144435, CurrentID=0, Status='live', Symbol='VTRNA3-1P', Aliases='HVG4, HVGX, VAULTRC4, VTRNA3P, hvg-4', description='vault RNA 3-1, pseudogene', other_designations='vault RNA 3 (pseudogene)|vault RNA component 4 pseudogene', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53462210, end_position_on_the_genomic_accession=53462310, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442334, CurrentID=0, Status='live', Symbol='ARF1P1', Aliases='ARF1', description='ADP ribosylation factor 1 pseudogene 1', other_designations='ADP-ribosylation factor 1 pseudogene 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94833827, end_position_on_the_genomic_accession=94836304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100302650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302650, CurrentID=0, Status='live', Symbol='BRE-AS1', Aliases='', description='BRE antisense RNA 1', other_designations='BRE antisense RNA 1 (non-protein coding)', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27889456, end_position_on_the_genomic_accession=27891114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302652, CurrentID=0, Status='live', Symbol='GPR75-ASB3', Aliases='', description='GPR75-ASB3 readthrough', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53670293, end_position_on_the_genomic_accession=53860033, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  642669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642669, CurrentID=0, Status='live', Symbol='LOC642669', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19228925, end_position_on_the_genomic_accession=19229509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302657, CurrentID=0, Status='live', Symbol='DDX11L3', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 3', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 3|DEAD/H box polypeptide 11 like 3', map_location='3q29', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=102516761, end_position_on_the_genomic_accession=102519296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302666, CurrentID=0, Status='live', Symbol='CSPG4P13', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 13', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77878638, end_position_on_the_genomic_accession=77904669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479586, CurrentID=0, Status='live', Symbol='RNU4-62P', Aliases='', description='RNA, U4 small nuclear 62, pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302680, CurrentID=0, Status='live', Symbol='DELXQ28', Aliases='MICRODELXq28', description='Chromosome Xq28 microdeletion syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300475, no_hdr0=''),
  100302681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302681, CurrentID=0, Status='live', Symbol='PVNH3', Aliases='', description='Periventricular nodular heterotopia 3', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608098, no_hdr0=''),
  100302682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302682, CurrentID=0, Status='live', Symbol='PVNH5', Aliases='', description='Periventricular nodular heterotopia 5', other_designations='', map_location='5q14.3-q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612881, no_hdr0=''),
  100302683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302683, CurrentID=0, Status='live', Symbol='STQTL18', Aliases='', description='Stature QTL 18', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612892, no_hdr0=''),
  100302684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302684, CurrentID=0, Status='live', Symbol='STQTL19', Aliases='', description='Stature QTL 19', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612893, no_hdr0=''),
  100302685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302685, CurrentID=0, Status='live', Symbol='STQTL20', Aliases='', description='Stature QTL 20', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612894, no_hdr0=''),
  100302686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302686, CurrentID=0, Status='live', Symbol='MTBS3', Aliases='', description='Mycobacterium tuberculosis, susceptibility to, 3', other_designations='', map_location='19q13.31-q33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612929, no_hdr0=''),
  100302689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302689, CurrentID=0, Status='live', Symbol='GPN3P1', Aliases='', description='GPN-loop GTPase 3 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25647077, end_position_on_the_genomic_accession=25648104, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302690, CurrentID=0, Status='live', Symbol='DLG2-AS1', Aliases='DLG2-AS, DLG2AS, PSZA11q14, SZ-1', description='DLG2 antisense RNA 1', other_designations='DLG2 antisense RNA (non-protein coding)|DLG2 antisense RNA 1 (non-protein coding)', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302691, CurrentID=0, Status='live', Symbol='LINC00184', Aliases='HANC, NCRNA00184', description='long intergenic non-protein coding RNA 184', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234629311, end_position_on_the_genomic_accession=234634780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100302692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302692, CurrentID=0, Status='live', Symbol='FTX', Aliases='LINC00182, MIR374AHG, NCRNA00182', description='FTX transcript, XIST regulator (non-protein coding)', other_designations='five prime to XIST|mir-374a-545 cluster host gene (non-protein coding)', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74028136, end_position_on_the_genomic_accession=74293574, orientation='minus', exon_count=7, OMIM=300936, no_hdr0=''),
  100271575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271575, CurrentID=0, Status='live', Symbol='RPS26P49', Aliases='RPS26_27_1414', description='ribosomal protein S26 pseudogene 49', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104830836, end_position_on_the_genomic_accession=104831170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163688, CurrentID=0, Status='live', Symbol='CALML6', Aliases='CAGLP', description='calmodulin like 6', other_designations='EF-hand protein|calglandulin-like protein|calmodulin-like 6', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1913752, end_position_on_the_genomic_accession=1917294, orientation='plus', exon_count=6, OMIM=610171, no_hdr0=''),
  1086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1086, CurrentID=0, Status='live', Symbol='CEACAMP11', Aliases='CEACAM32P, CGM18', description='carcinoembryonic antigen-related cell adhesion molecule pseudogene 11', other_designations='carcinoembryonic antigen gene family member 18|carcinoembryonic antigen-related cell adhesion molecule 32, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43283436, end_position_on_the_genomic_accession=43284346, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644766, CurrentID=0, Status='live', Symbol='LOC644766', Aliases='', description='uncharacterized LOC644766', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=37360963, end_position_on_the_genomic_accession=37363560, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27283, CurrentID=0, Status='live', Symbol='TINAG', Aliases='TIN-AG', description='tubulointerstitial nephritis antigen', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54307792, end_position_on_the_genomic_accession=54390152, orientation='plus', exon_count=13, OMIM=606749, no_hdr0=''),
  101060373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060373, CurrentID=0, Status='live', Symbol='LOC101060373', Aliases='', description='uncharacterized LOC101060373', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  163702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163702, CurrentID=0, Status='live', Symbol='IFNLR1', Aliases='CRF2/12, IFNLR, IL-28R1, IL28RA, LICR2', description='interferon, lambda receptor 1', other_designations='CRF2-12|IFN-lambda receptor 1|IFN-lambda-R1|IL-28 receptor subunit alpha|IL-28R-alpha|IL-28RA|class II cytokine receptor CRF2/12|cytokine receptor class-II member 12|cytokine receptor family 2 member 12|interferon lambda, receptor 1|interleukin 28 alpha receptor|interleukin 28 receptor A|interleukin 28 receptor, alpha (interferon, lambda receptor)|interleukin or cytokine receptor 2|interleukin-28 receptor subunit alpha|likely interleukin or cytokine receptor 2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24154157, end_position_on_the_genomic_accession=24187275, orientation='minus', exon_count=8, OMIM=607404, no_hdr0=''),
  100302716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302716, CurrentID=0, Status='live', Symbol='HTGH', Aliases='DEL17q24', description='Hypertrichosis terminalis, generalized, with or without gingival hyperplasia', other_designations='', map_location='17q24.2-q24.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=135400, no_hdr0=''),
  100302718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302718, CurrentID=0, Status='live', Symbol='B6QTL1', Aliases='', description='Vitamin B6 plasma level QTL 1', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612957, no_hdr0=''),
  100302719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302719, CurrentID=0, Status='live', Symbol='BMIQ15', Aliases='', description='Body mass index QTL 15', other_designations='', map_location='17q23.2-q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612967, no_hdr0=''),
  100302720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302720, CurrentID=0, Status='live', Symbol='CTRCT34', Aliases='CATC3', description='Cataract, autosomal recessive congenital 3', other_designations='', map_location='1p34.3-p32.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612968, no_hdr0=''),
  100302721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302721, CurrentID=0, Status='live', Symbol='ARHI2', Aliases='', description='Age-related hearing impairment 2', other_designations='', map_location='3p26.1-p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612976, no_hdr0=''),
  27286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27286, CurrentID=0, Status='live', Symbol='SRPX2', Aliases='BPP, CBPS, PMGX, RESDX, SRPUL', description='sushi-repeat containing protein, X-linked 2', other_designations='sushi-repeat protein up-regulated in leukemia|sushi-repeat protein upregulated in leukemia|sushi-repeat-containing protein, X-linked 2', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100644166, end_position_on_the_genomic_accession=100671299, orientation='plus', exon_count=11, OMIM=300642, no_hdr0=''),
  163720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163720, CurrentID=0, Status='live', Symbol='CYP4Z2P', Aliases='', description='cytochrome P450 family 4 subfamily Z member 2, pseudogene', other_designations='cytochrome P450 4Z2 pseudogene|cytochrome P450, family 4, subfamily Z, polypeptide 2 pseudogene|cytochrome P450, family 4, subfamily Z, polypeptide 2, pseudogene', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46843095, end_position_on_the_genomic_accession=46900475, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105372840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372840, CurrentID=0, Status='live', Symbol='LOC105372840', Aliases='', description='uncharacterized LOC105372840', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45590787, end_position_on_the_genomic_accession=45597407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100302736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302736, CurrentID=0, Status='live', Symbol='TMED7-TICAM2', Aliases='', description='TMED7-TICAM2 readthrough', other_designations='TMED7-TICAM2 read-through transcript', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115578496, end_position_on_the_genomic_accession=115626179, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100302737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302737, CurrentID=0, Status='live', Symbol='ZNF861P', Aliases='', description='zinc finger protein 861, pseudogene', other_designations='zinc finger protein 709 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15821993, end_position_on_the_genomic_accession=15823787, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100302738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302738, CurrentID=0, Status='live', Symbol='RNU6-21P', Aliases='RNU6-21', description='RNA, U6 small nuclear 21, pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=62187540, end_position_on_the_genomic_accession=62187646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302739, CurrentID=0, Status='live', Symbol='PCNA-AS1', Aliases='PCNA-AS, PCNAAS', description='PCNA antisense RNA 1', other_designations='PCNA antisense RNA (non-protein coding)|PCNA antisense RNA 1 (non-protein coding)', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5119586, end_position_on_the_genomic_accession=5119969, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302740, CurrentID=0, Status='live', Symbol='FAS-AS1', Aliases='FAS-AS, FASAS, SAF', description='FAS antisense RNA 1', other_designations='FAS antisense RNA (non-protein coding)|FAS antisense RNA 1 (non-protein coding)', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88991422, end_position_on_the_genomic_accession=88992975, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302741, CurrentID=0, Status='live', Symbol='RNU6-15P', Aliases='RNU6-15', description='RNA, U6 small nuclear 15, pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21321961, end_position_on_the_genomic_accession=21322067, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302742, CurrentID=0, Status='live', Symbol='RNU6-42P', Aliases='RNU6-42', description='RNA, U6 small nuclear 42, pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196784980, end_position_on_the_genomic_accession=196785086, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100302743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302743, CurrentID=0, Status='live', Symbol='SNORA80B', Aliases='ACA67B', description='small nucleolar RNA, H/ACA box 80B', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10446714, end_position_on_the_genomic_accession=10446849, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100996420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996420, CurrentID=0, Status='live', Symbol='DNM1P50', Aliases='', description='dynamin 1 pseudogene 50', other_designations='DNM1 pseudogene 50', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302746, CurrentID=0, Status='live', Symbol='NCRUPAR', Aliases='NCRNA00193, ncR-uPAR', description='non-protein coding RNA, upstream of F2R/PAR1', other_designations='noncoding RNA upstream of the PAR-1', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76712403, end_position_on_the_genomic_accession=76712888, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  163742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163742, CurrentID=0, Status='live', Symbol='SLC25A3P1', Aliases='', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 3 pseudogene 1', other_designations='solute carrier family 24 (sodium/potassium/calcium exchanger), member 3 pseudogene 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53438370, end_position_on_the_genomic_accession=53440020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100302751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302751, CurrentID=0, Status='live', Symbol='NCLP1', Aliases='', description='nucleolin pseudogene 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136812788, end_position_on_the_genomic_accession=136815476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163747, CurrentID=0, Status='live', Symbol='LEXM', Aliases='C1orf177, LEM', description='lymphocyte expansion molecule', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54806063, end_position_on_the_genomic_accession=54842264, orientation='plus', exon_count=13, OMIM=616446, no_hdr0=''),
  100873881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873881, CurrentID=0, Status='live', Symbol='GRAMD4P2', Aliases='', description='GRAM domain containing 4 pseudogene 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15750362, end_position_on_the_genomic_accession=15750824, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  27295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27295, CurrentID=0, Status='live', Symbol='PDLIM3', Aliases='ALP', description='PDZ and LIM domain 3', other_designations='alpha-actinin-2-associated LIM protein|enigma homolog', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185500660, end_position_on_the_genomic_accession=185535607, orientation='minus', exon_count=9, OMIM=605889, no_hdr0=''),
  163778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163778, CurrentID=0, Status='live', Symbol='SPRR4', Aliases='', description='small proline-rich protein 4', other_designations='small proline rich protein 4', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152970653, end_position_on_the_genomic_accession=152972574, orientation='plus', exon_count=2, OMIM=616363, no_hdr0=''),
  163782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163782, CurrentID=0, Status='live', Symbol='KANK4', Aliases='ANKRD38, dJ1078M7.1', description='KN motif and ankyrin repeat domains 4', other_designations='ankyrin repeat domain 38|ankyrin repeat domain-containing protein 38|kidney ankyrin repeat-containing protein 4', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62236165, end_position_on_the_genomic_accession=62319412, orientation='minus', exon_count=12, OMIM=614612, no_hdr0=''),
  100996428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996428, CurrentID=0, Status='live', Symbol='LOC100996428', Aliases='', description='helicase SRCAP-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63758428, end_position_on_the_genomic_accession=63761708, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  163786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163786, CurrentID=0, Status='live', Symbol='SASS6', Aliases='MCPH14, SAS-6, SAS6', description='SAS-6 centriolar assembly protein', other_designations='hsSAS-6|spindle assembly 6 homolog', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100083544, end_position_on_the_genomic_accession=100133095, orientation='minus', exon_count=17, OMIM=609321, no_hdr0=''),
  106479609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479609, CurrentID=0, Status='live', Symbol='RNU6-98P', Aliases='', description='RNA, U6 small nuclear 98, pseudogene', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373703, CurrentID=0, Status='live', Symbol='LOC105373703', Aliases='', description='uncharacterized LOC105373703', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155866975, end_position_on_the_genomic_accession=155913995, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100289037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289037, CurrentID=0, Status='live', Symbol='LOC100289037', Aliases='', description='SNU13 homolog, small nuclear ribonucleoprotein (U4/U6.U5) pseudogene', other_designations='NHP2 non-histone chromosome protein 2-like 1 (S. cerevisiae) pseudogene', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97449365, end_position_on_the_genomic_accession=97449741, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  158373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158373, CurrentID=0, Status='live', Symbol='YBX1P10', Aliases='BA327L3.4', description='Y-box binding protein 1 pseudogene 10', other_designations='Y box binding protein 1 pseudogene 10|nuclease sensitive element binding protein 1 pseudogene', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35970944, end_position_on_the_genomic_accession=35972457, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928907, CurrentID=0, Status='live', Symbol='LOC101928907', Aliases='', description='uncharacterized LOC101928907', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62060503, end_position_on_the_genomic_accession=62061278, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  158376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158376, CurrentID=0, Status='live', Symbol='LINC00961', Aliases='', description='long intergenic non-protein coding RNA 961', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35909483, end_position_on_the_genomic_accession=35911620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479614, CurrentID=0, Status='live', Symbol='RNU6-107P', Aliases='', description='RNA, U6 small nuclear 107, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372491, CurrentID=0, Status='live', Symbol='LOC105372491', Aliases='', description='uncharacterized LOC105372491', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=660779, end_position_on_the_genomic_accession=692740, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378069, CurrentID=0, Status='live', Symbol='LOC105378069', Aliases='', description='uncharacterized LOC105378069', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155689642, end_position_on_the_genomic_accession=155697834, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369340, CurrentID=0, Status='live', Symbol='LOC105369340', Aliases='', description='uncharacterized LOC105369340', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64245377, end_position_on_the_genomic_accession=64248219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060443, CurrentID=0, Status='live', Symbol='LOC101060443', Aliases='', description='uncharacterized LOC101060443', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158381, CurrentID=0, Status='live', Symbol='ATP8B5P', Aliases='FetA', description='ATPase, class I, type 8B, member 5, pseudogene', other_designations='ATPase, Class I, type 8B family pseudogene|flippase expressed in testis splicing form A pseudogene|zinc finger, AN1-type domain 6 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35406755, end_position_on_the_genomic_accession=35483029, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  163859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163859, CurrentID=0, Status='live', Symbol='SDE2', Aliases='C1orf55, dJ671D7.1', description='SDE2 telomere maintenance homolog (S. pombe)', other_designations='UPF0667 protein C1orf55', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225982702, end_position_on_the_genomic_accession=225999365, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100419975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419975, CurrentID=0, Status='live', Symbol='LOC100419975', Aliases='', description='transmembrane protein 192 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70693839, end_position_on_the_genomic_accession=70694619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  136540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136540, CurrentID=0, Status='live', Symbol='PRSS3P3', Aliases='TRY3', description='protease, serine 3 pseudogene 3', other_designations='protease, serine, 3 pseudogene 3', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142287248, end_position_on_the_genomic_accession=142291610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163882, CurrentID=0, Status='live', Symbol='CNST', Aliases='C1orf71, PPP1R64', description='consortin, connexin sorting protein', other_designations='protein phosphatase 1, regulatory subunit 64', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246566337, end_position_on_the_genomic_accession=246668587, orientation='plus', exon_count=14, OMIM=613439, no_hdr0=''),
  106479623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479623, CurrentID=0, Status='live', Symbol='RNU6-129P', Aliases='', description='RNA, U6 small nuclear 129, pseudogene', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  136541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136541, CurrentID=0, Status='live', Symbol='PRSS58', Aliases='PRSS1, TRY1, TRYX3, UNQ2540', description='protease, serine 58', other_designations='protease, serine, 58|trypsin-X3', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142252138, end_position_on_the_genomic_accession=142258058, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100422110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422110, CurrentID=0, Status='live', Symbol='LOC100422110', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873886, CurrentID=0, Status='live', Symbol='SCAND3P1', Aliases='', description='SCAN domain containing 3 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35185294, end_position_on_the_genomic_accession=35186679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289569, CurrentID=0, Status='live', Symbol='LOC100289569', Aliases='', description='zinc finger protein 736 pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=106544099, end_position_on_the_genomic_accession=106544496, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442382, CurrentID=0, Status='live', Symbol='LOC442382', Aliases='', description='ATPase, Ca++ transporting, plasma membrane 1 pseudogene', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19238437, end_position_on_the_genomic_accession=19240873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271583, CurrentID=0, Status='live', Symbol='RPS27P26', Aliases='RPS27_12_1494', description='ribosomal protein S27 pseudogene 26', other_designations='ribosomal protein S27-like pseudogene', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70444674, end_position_on_the_genomic_accession=70444912, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  163933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163933, CurrentID=0, Status='live', Symbol='FAM43B', Aliases='', description='family with sequence similarity 43 member B', other_designations='family with sequence similarity 43, member B', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20552439, end_position_on_the_genomic_accession=20555020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479632, CurrentID=0, Status='live', Symbol='RNU6-150P', Aliases='', description='RNA, U6 small nuclear 150, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271584, CurrentID=0, Status='live', Symbol='RPS27P25', Aliases='RPS27_14_1288', description='ribosomal protein S27 pseudogene 25', other_designations='ribosomal protein S27-like pseudogene', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120371356, end_position_on_the_genomic_accession=120371695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479636, CurrentID=0, Status='live', Symbol='RNU6-157P', Aliases='', description='RNA, U6 small nuclear 157, pseudogene', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  158399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158399, CurrentID=0, Status='live', Symbol='ZNF483', Aliases='ZKSCAN16, ZSCAN48', description='zinc finger protein 483', other_designations='zinc finger protein HIT-10|zinc finger protein with KRAB and SCAN domains 16', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111524299, end_position_on_the_genomic_accession=111577844, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  442389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442389, CurrentID=0, Status='live', Symbol='LOC442389', Aliases='', description='interferon induced transmembrane protein pseudogene', other_designations='', map_location='8q12.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60906039, end_position_on_the_genomic_accession=60906600, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  158401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158401, CurrentID=0, Status='live', Symbol='C9orf84', Aliases='', description='chromosome 9 open reading frame 84', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111686175, end_position_on_the_genomic_accession=111795008, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  158402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158402, CurrentID=0, Status='live', Symbol='LOC158402', Aliases='', description='uncharacterized LOC158402', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144494, CurrentID=0, Status='live', Symbol='KDELC1P1', Aliases='', description='KDEL motif containing 1 pseudogene 1', other_designations='KDEL (Lys-Asp-Glu-Leu) containing 1 pseudogene 1|KDELC1 pseudogene', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128109205, end_position_on_the_genomic_accession=128111198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  27333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27333, CurrentID=0, Status='live', Symbol='GOLIM4', Aliases='GIMPC, GOLPH4, GPP130, P138', description='golgi integral membrane protein 4', other_designations='130 kDa golgi-localized phosphoprotein|cis Golgi-localized calcium-binding protein|golgi integral membrane protein, cis|golgi phosphoprotein 4|golgi phosphoprotein of 130 kDa|golgi-localized phosphoprotein of 130 kDa|type II Golgi membrane protein', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168008669, end_position_on_the_genomic_accession=168095924, orientation='minus', exon_count=16, OMIM=606805, no_hdr0=''),
  100131921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131921, CurrentID=0, Status='live', Symbol='SERF1AP1', Aliases='', description='small EDRK-rich factor 1A (telomeric) pseudogene 1', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137301004, end_position_on_the_genomic_accession=137301187, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442396, CurrentID=0, Status='live', Symbol='ARF1P3', Aliases='', description='ADP ribosylation factor 1 pseudogene 3', other_designations='ADP-ribosylation factor 1 pseudogene 3', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124151924, end_position_on_the_genomic_accession=124152436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646164, CurrentID=0, Status='live', Symbol='DDX39AP1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39A pseudogene 1', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39 pseudogene', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22702740, end_position_on_the_genomic_accession=22706205, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271586, CurrentID=0, Status='live', Symbol='RPS27P28', Aliases='RPS27_16_1593', description='ribosomal protein S27 pseudogene 28', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32779039, end_position_on_the_genomic_accession=32779286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369346, CurrentID=0, Status='live', Symbol='LOC105369346', Aliases='', description='uncharacterized LOC105369346', other_designations='CMB9-22P13.2', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65487241, end_position_on_the_genomic_accession=65489968, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  164045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164045, CurrentID=0, Status='live', Symbol='HFM1', Aliases='MER3, POF9, SEC63D1, Si-11, Si-11-6, helicase', description='HFM1, ATP-dependent DNA helicase homolog', other_designations='SEC63 domain-containing protein 1|helicase-like protein HFM1', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91260766, end_position_on_the_genomic_accession=91408007, orientation='minus', exon_count=43, OMIM=615684, no_hdr0=''),
  729973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729973, CurrentID=0, Status='live', Symbol='LOC729973', Aliases='', description='uncharacterized LOC729973', other_designations='coiled-coil domain containing 72 pseudogene', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27649400, end_position_on_the_genomic_accession=27649864, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369347, CurrentID=0, Status='live', Symbol='LOC105369347', Aliases='', description='uncharacterized LOC105369347', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65662992, end_position_on_the_genomic_accession=65671701, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  442405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442405, CurrentID=0, Status='live', Symbol='KRT18P66', Aliases='', description='keratin 18 pseudogene 66', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30773463, end_position_on_the_genomic_accession=30774807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642683, CurrentID=0, Status='live', Symbol='KIF18BP1', Aliases='', description='kinesin family member 18B pseudogene 1', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35638634, end_position_on_the_genomic_accession=35645158, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442406, CurrentID=0, Status='live', Symbol='KRT18P36', Aliases='', description='keratin 18 pseudogene 36', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30799658, end_position_on_the_genomic_accession=30800798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271588, CurrentID=0, Status='live', Symbol='RPS29P23', Aliases='RPS29_17_1627', description='ribosomal protein S29 pseudogene 23', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12373581, end_position_on_the_genomic_accession=12373731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  164091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164091, CurrentID=0, Status='live', Symbol='PAQR7', Aliases='MPRA, PGLP, mSR', description='progestin and adipoQ receptor family member VII', other_designations='2310021M12Rik|mPR alpha|progestin and adipoQ receptor family member 7', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25861484, end_position_on_the_genomic_accession=25876707, orientation='minus', exon_count=5, OMIM=607779, no_hdr0=''),
  102724449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724449, CurrentID=0, Status='live', Symbol='LOC102724449', Aliases='', description='uncharacterized LOC102724449', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31655387, end_position_on_the_genomic_accession=31659520, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100289047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289047, CurrentID=0, Status='live', Symbol='LOC100289047', Aliases='', description='dnaJ homolog subfamily B member 6-like', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26432644, end_position_on_the_genomic_accession=26434772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271589, CurrentID=0, Status='live', Symbol='RPS29P24', Aliases='RPS29_18_1653', description='ribosomal protein S29 pseudogene 24', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39588955, end_position_on_the_genomic_accession=39589125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996483, CurrentID=0, Status='live', Symbol='LOC100996483', Aliases='', description='caskin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53354522, end_position_on_the_genomic_accession=53358390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369349, CurrentID=0, Status='live', Symbol='LOC105369349', Aliases='', description='uncharacterized LOC105369349', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65922564, end_position_on_the_genomic_accession=65929479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  164118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164118, CurrentID=0, Status='live', Symbol='TTC24', Aliases='', description='tetratricopeptide repeat domain 24', other_designations='TPR repeat protein 24', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156579727, end_position_on_the_genomic_accession=156587797, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  442415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442415, CurrentID=0, Status='live', Symbol='ATP5A1P7', Aliases='ATP5A1P9', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 7', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 9', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62686910, end_position_on_the_genomic_accession=62688339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  164127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164127, CurrentID=0, Status='live', Symbol='CCDC185', Aliases='C1orf65', description='coiled-coil domain containing 185', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223393373, end_position_on_the_genomic_accession=223395470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479664, CurrentID=0, Status='live', Symbol='RNU6-219P', Aliases='', description='RNA, U6 small nuclear 219, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271590, CurrentID=0, Status='live', Symbol='RPS29P25', Aliases='RPS29_19_1654', description='ribosomal protein S29 pseudogene 25', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39590981, end_position_on_the_genomic_accession=39591151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422061, CurrentID=0, Status='live', Symbol='LOC100422061', Aliases='', description='olfactory receptor, family 11, subfamily H, member 4 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  164153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164153, CurrentID=0, Status='live', Symbol='UBL4B', Aliases='', description='ubiquitin like 4B', other_designations='ubiquitin-like 4B', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110112440, end_position_on_the_genomic_accession=110113947, orientation='plus', exon_count=1, OMIM=611127, no_hdr0=''),
  100996490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996490, CurrentID=0, Status='live', Symbol='LINC01206', Aliases='', description='long intergenic non-protein coding RNA 1206', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181952364, end_position_on_the_genomic_accession=182010678, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106479669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479669, CurrentID=0, Status='live', Symbol='RNU6-230P', Aliases='', description='RNA, U6 small nuclear 230, pseudogene', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  256676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256676, CurrentID=0, Status='live', Symbol='LOC256676', Aliases='', description='uncharacterized LOC256676', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873895, CurrentID=0, Status='live', Symbol='HMGB1P38', Aliases='', description='high mobility group box 1 pseudogene 38', other_designations='high-mobility group box 1 pseudogene 38', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=79957566, end_position_on_the_genomic_accession=79958142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479673, CurrentID=0, Status='live', Symbol='RNU6-240P', Aliases='', description='RNA, U6 small nuclear 240, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642687, CurrentID=0, Status='live', Symbol='RPS11P7', Aliases='RPS11_4_1771', description='ribosomal protein S11 pseudogene 7', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39865430, end_position_on_the_genomic_accession=39865941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479674, CurrentID=0, Status='live', Symbol='RNU6-241P', Aliases='', description='RNA, U6 small nuclear 241, pseudogene', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442427, CurrentID=0, Status='live', Symbol='LOC442427', Aliases='', description='putative UPF0607 protein ENSP00000381418', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82431889, end_position_on_the_genomic_accession=82433228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271592, CurrentID=0, Status='live', Symbol='RPS29P27', Aliases='RPS29_21_1656', description='ribosomal protein S29 pseudogene 27', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39697906, end_position_on_the_genomic_accession=39698073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442429, CurrentID=0, Status='live', Symbol='EEF1DP2', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 2', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92836766, end_position_on_the_genomic_accession=92837741, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103312105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103312105, CurrentID=0, Status='live', Symbol='LOC103312105', Aliases='', description='uncharacterized LOC103312105', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38618264, end_position_on_the_genomic_accession=38646356, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100996500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996500, CurrentID=0, Status='live', Symbol='PDSS1P2', Aliases='', description='prenyl (decaprenyl) diphosphate synthase, subunit 1 pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372107, CurrentID=0, Status='live', Symbol='LOC105372107', Aliases='', description='uncharacterized LOC105372107', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48673575, end_position_on_the_genomic_accession=48692858, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100533464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533464, CurrentID=0, Status='live', Symbol='PRH1-PRR4', Aliases='PRH1', description='PRH1-PRR4 readthrough', other_designations='', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10845849, end_position_on_the_genomic_accession=11171625, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100271593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271593, CurrentID=0, Status='live', Symbol='RPS3AP37', Aliases='RPS3A_14_1056', description='ribosomal protein S3a pseudogene 37', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68674304, end_position_on_the_genomic_accession=68675150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  164237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164237, CurrentID=0, Status='live', Symbol='WFDC13', Aliases='C20orf138, WAP13, dJ601O1.3', description='WAP four-disulfide core domain 13', other_designations='protease inhibitor WAP13|protein WFDC13', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45702016, end_position_on_the_genomic_accession=45708817, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106478984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478984, CurrentID=0, Status='live', Symbol='RNY3P1', Aliases='', description='RNA, Ro-associated Y3 pseudogene 1', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271594, CurrentID=0, Status='live', Symbol='RPS3AP38', Aliases='RPS3A_17_1053', description='ribosomal protein S3a pseudogene 38', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67960685, end_position_on_the_genomic_accession=67961494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369354, CurrentID=0, Status='live', Symbol='LOC105369354', Aliases='', description='uncharacterized LOC105369354', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66184340, end_position_on_the_genomic_accession=66190236, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  164284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164284, CurrentID=0, Status='live', Symbol='APCDD1L', Aliases='', description='adenomatosis polyposis coli down-regulated 1-like', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58452469, end_position_on_the_genomic_accession=58515393, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  442442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442442, CurrentID=0, Status='live', Symbol='RPL14P5', Aliases='', description='ribosomal protein L14 pseudogene 5', other_designations='60S ribosomal protein L14 pseudogene', map_location='Xp22.33;Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1009720, end_position_on_the_genomic_accession=1010100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271595, CurrentID=0, Status='live', Symbol='RPS3AP41', Aliases='RPS3A_17_1152', description='ribosomal protein S3a pseudogene 41', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71669496, end_position_on_the_genomic_accession=71670341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442443, CurrentID=0, Status='live', Symbol='LOC442443', Aliases='', description='chromodomain Y-like protein pseudogene', other_designations='chromodomain protein, Y-like pseudogene', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9402747, end_position_on_the_genomic_accession=9416082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479692, CurrentID=0, Status='live', Symbol='RNU6-283P', Aliases='', description='RNA, U6 small nuclear 283, pseudogene', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442445, CurrentID=0, Status='live', Symbol='LOC442445', Aliases='', description='ferritin, heavy polypeptide-like 17 pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37077869, end_position_on_the_genomic_accession=37078401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442446, CurrentID=0, Status='live', Symbol='UBTFL11', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 11 (pseudogene)', other_designations='upstream binding transcription factor, RNA polymerase I pseudogene', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38531408, end_position_on_the_genomic_accession=38534256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533467, CurrentID=0, Status='live', Symbol='BIVM-ERCC5', Aliases='ERCC5-202', description='BIVM-ERCC5 readthrough', other_designations='', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102807146, end_position_on_the_genomic_accession=102876001, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  164312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164312, CurrentID=0, Status='live', Symbol='LRRN4', Aliases='C20orf75, NLRR-4, NLRR4, dJ1056H1.1', description='leucine rich repeat neuronal 4', other_designations='neuronal leucine-rich repeat protein 4', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6039745, end_position_on_the_genomic_accession=6054080, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  442447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442447, CurrentID=0, Status='live', Symbol='CLIC4P3', Aliases='', description='chloride intracellular channel 4 pseudogene 3', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41076999, end_position_on_the_genomic_accession=41077754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873900, CurrentID=0, Status='live', Symbol='HMGB3P16', Aliases='', description='high mobility group box 3 pseudogene 16', other_designations='high-mobility group box 3 pseudogene 16', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112452703, end_position_on_the_genomic_accession=112452989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369356, CurrentID=0, Status='live', Symbol='LOC105369356', Aliases='', description='uncharacterized LOC105369356', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67267925, end_position_on_the_genomic_accession=67274059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642692, CurrentID=0, Status='live', Symbol='LOC642692', Aliases='', description='centrosomal protein 19kDa pseudogene', other_designations='', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235146834, end_position_on_the_genomic_accession=235147700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271597, CurrentID=0, Status='live', Symbol='RPS3AP42', Aliases='RPS3A_20_1165', description='ribosomal protein S3a pseudogene 42', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92498101, end_position_on_the_genomic_accession=92498958, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930115, CurrentID=0, Status='live', Symbol='LOC101930115', Aliases='', description='von Willebrand factor A domain-containing protein 3A', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479702, CurrentID=0, Status='live', Symbol='RNU6-303P', Aliases='', description='RNA, U6 small nuclear 303, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422529, CurrentID=0, Status='live', Symbol='LOC100422529', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18645991, end_position_on_the_genomic_accession=18646505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442455, CurrentID=0, Status='live', Symbol='KRT8P27', Aliases='', description='keratin 8 pseudogene 27', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64622280, end_position_on_the_genomic_accession=64624575, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616395, CurrentID=0, Status='live', Symbol='MIR4789', Aliases='', description='microRNA 4789', other_designations='hsa-mir-4789', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175369540, end_position_on_the_genomic_accession=175369621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  164380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164380, CurrentID=0, Status='live', Symbol='CST13P', Aliases='CSTT, CTES6', description='cystatin 13, pseudogene', other_designations='cystatin T|cystatin pseudogene', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23519146, end_position_on_the_genomic_accession=23542018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479707, CurrentID=0, Status='live', Symbol='RNU6-315P', Aliases='', description='RNA, U6 small nuclear 315, pseudogene', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442460, CurrentID=0, Status='live', Symbol='LOC442460', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100800103, end_position_on_the_genomic_accession=100800831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  164395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164395, CurrentID=0, Status='live', Symbol='TTLL9', Aliases='C20orf125', description='tubulin tyrosine ligase like 9', other_designations='tubulin tyrosine ligase-like family, member 9|tubulin--tyrosine ligase-like protein 9', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31870702, end_position_on_the_genomic_accession=31944959, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  106479709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479709, CurrentID=0, Status='live', Symbol='RNU6-321P', Aliases='', description='RNA, U6 small nuclear 321, pseudogene', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442465, CurrentID=0, Status='live', Symbol='SAGE4P', Aliases='', description='sarcoma antigen 4, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135598804, end_position_on_the_genomic_accession=135633643, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370746, CurrentID=0, Status='live', Symbol='LOC105370746', Aliases='', description='uncharacterized LOC105370746', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30169985, end_position_on_the_genomic_accession=30179938, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271600, CurrentID=0, Status='live', Symbol='RPS3AP51', Aliases='RPS3A_23_1744', description='ribosomal protein S3a pseudogene 51', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30136630, end_position_on_the_genomic_accession=30137398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100303453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303453, CurrentID=0, Status='live', Symbol='TSNAX-DISC1', Aliases='', description='TSNAX-DISC1 readthrough (NMD candidate)', other_designations='', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231528653, end_position_on_the_genomic_accession=232041272, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105369360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369360, CurrentID=0, Status='live', Symbol='LOC105369360', Aliases='', description='uncharacterized LOC105369360', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67884299, end_position_on_the_genomic_accession=67894096, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  644906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644906, CurrentID=0, Status='live', Symbol='LOC644906', Aliases='', description='malate dehydrogenase 1, NAD (soluble) pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16107976, end_position_on_the_genomic_accession=16108595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271601, CurrentID=0, Status='live', Symbol='RPS3AP50', Aliases='RPS3A_24_1646', description='ribosomal protein S3a pseudogene 50', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33341203, end_position_on_the_genomic_accession=33341984, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369361, CurrentID=0, Status='live', Symbol='LOC105369361', Aliases='', description='uncharacterized LOC105369361', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68023934, end_position_on_the_genomic_accession=68030592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100303491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303491, CurrentID=0, Status='live', Symbol='ZEB2-AS1', Aliases='ZEB2-AS, ZEB2AS, ZEB2NAT', description='ZEB2 antisense RNA 1', other_designations='ZEB2 antisense RNA 1 (non-protein coding)|ZEB2 natural antisense transcript', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144519614, end_position_on_the_genomic_accession=144520899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114796, CurrentID=0, Status='live', Symbol='PSMG3-AS1', Aliases='', description='PSMG3 antisense RNA 1 (head to head)', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1570073, end_position_on_the_genomic_accession=1589626, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100289060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289060, CurrentID=0, Status='live', Symbol='LOC100289060', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 10 pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58686387, end_position_on_the_genomic_accession=58687581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271602, CurrentID=0, Status='live', Symbol='RPL10AP13', Aliases='RPL10A_8_1594', description='ribosomal protein L10a pseudogene 13', other_designations='', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34573986, end_position_on_the_genomic_accession=34574499, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369362, CurrentID=0, Status='live', Symbol='LOC105369362', Aliases='', description='uncharacterized LOC105369362', other_designations='CTD-2655K5.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68121649, end_position_on_the_genomic_accession=68130507, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479727, CurrentID=0, Status='live', Symbol='RNU6-363P', Aliases='', description='RNA, U6 small nuclear 363, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996549, CurrentID=0, Status='live', Symbol='LOC100996549', Aliases='', description='uncharacterized LOC100996549', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9758590, end_position_on_the_genomic_accession=9778575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  136647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136647, CurrentID=0, Status='live', Symbol='MPLKIP', Aliases='ABHS, C7orf11, ORF20, TTD4', description='M-phase specific PLK1 interacting protein', other_designations='Russell-Silver syndrome region|TTD non-photosensitive 1 protein', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40132743, end_position_on_the_genomic_accession=40134652, orientation='minus', exon_count=2, OMIM=609188, no_hdr0=''),
  100289061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289061, CurrentID=0, Status='live', Symbol='LOC100289061', Aliases='', description='uncharacterized LOC100289061', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150548449, end_position_on_the_genomic_accession=150557938, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100271603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271603, CurrentID=0, Status='live', Symbol='RPL13AP24', Aliases='RPL13A_8_1193', description='ribosomal protein L13a pseudogene 24', other_designations='', map_location='12p13.3-p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3071546, end_position_on_the_genomic_accession=3072145, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114804, CurrentID=0, Status='live', Symbol='RNF157', Aliases='', description='ring finger protein 157', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76142453, end_position_on_the_genomic_accession=76240368, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  106479733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479733, CurrentID=0, Status='live', Symbol='RNU6-377P', Aliases='', description='RNA, U6 small nuclear 377, pseudogene', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131939, CurrentID=0, Status='live', Symbol='LOC100131939', Aliases='', description='PARK2 co-regulated like pseudogene', other_designations='PARK2 co-regulated-like pseudogene', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186580624, end_position_on_the_genomic_accession=186581194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479734, CurrentID=0, Status='live', Symbol='RNU6-379P', Aliases='', description='RNA, U6 small nuclear 379, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271604, CurrentID=0, Status='live', Symbol='RPL18AP12', Aliases='RPL18A_7_1543', description='ribosomal protein L18a pseudogene 12', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37162580, end_position_on_the_genomic_accession=37163196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369364, CurrentID=0, Status='live', Symbol='LOC105369364', Aliases='', description='uncharacterized LOC105369364', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68870422, end_position_on_the_genomic_accession=68874548, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479737, CurrentID=0, Status='live', Symbol='RNU6-388P', Aliases='', description='RNA, U6 small nuclear 388, pseudogene', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372151, CurrentID=0, Status='live', Symbol='LOC105372151', Aliases='', description='uncharacterized LOC105372151', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59927371, end_position_on_the_genomic_accession=59951306, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373729, CurrentID=0, Status='live', Symbol='LOC105373729', Aliases='', description='uncharacterized LOC105373729', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165624080, end_position_on_the_genomic_accession=165640659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  164592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164592, CurrentID=0, Status='live', Symbol='CCDC116', Aliases='', description='coiled-coil domain containing 116', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21632147, end_position_on_the_genomic_accession=21637327, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100873909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873909, CurrentID=0, Status='live', Symbol='DUTP2', Aliases='', description='deoxyuridine triphosphatase pseudogene 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123352307, end_position_on_the_genomic_accession=123352645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100144595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144595, CurrentID=0, Status='live', Symbol='LOC100144595', Aliases='', description='uncharacterized LOC100144595', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154435853, end_position_on_the_genomic_accession=154457438, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100144596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144596, CurrentID=0, Status='live', Symbol='LINC00029', Aliases='C20orf51, NCRNA00029', description='long intergenic non-protein coding RNA 29', other_designations='Putative uncharacterized protein C20orf51', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63034217, end_position_on_the_genomic_accession=63037028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479743, CurrentID=0, Status='live', Symbol='RNU6-401P', Aliases='', description='RNA, U6 small nuclear 401, pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284788, CurrentID=0, Status='live', Symbol='LOC284788', Aliases='', description='uncharacterized LOC284788', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22400333, end_position_on_the_genomic_accession=22420643, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=27434, CurrentID=0, Status='live', Symbol='POLM', Aliases='Pol Mu, Tdt-N', description='polymerase (DNA directed), mu', other_designations='DNA polymerase mu|Pol iota|polymerase (DNA-directed), mu|terminal transferase', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44072062, end_position_on_the_genomic_accession=44082540, orientation='minus', exon_count=14, OMIM=606344, no_hdr0=''),
  100144597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144597, CurrentID=0, Status='live', Symbol='LINC01056', Aliases='', description='long intergenic non-protein coding RNA 1056', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63038011, end_position_on_the_genomic_accession=63053863, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  442497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442497, CurrentID=0, Status='live', Symbol='LOC442497', Aliases='', description='uncharacterized LOC442497', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=379425, end_position_on_the_genomic_accession=382879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102030688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102030688, CurrentID=0, Status='live', Symbol='LOC102030688', Aliases='', description='testis expressed 28 pseudogene', other_designations='', map_location='Xp28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154233461, end_position_on_the_genomic_accession=154255096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  164633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164633, CurrentID=0, Status='live', Symbol='CABP7', Aliases='CALN2', description='calcium binding protein 7', other_designations='calneuron 2|calneuron II|calneuron-2', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29720090, end_position_on_the_genomic_accession=29731833, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  158511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158511, CurrentID=0, Status='live', Symbol='CSAG1', Aliases='CSAGE, CT24.1', description='chondrosarcoma associated gene 1', other_designations='cancer/testis antigen 24.1|cancer/testis antigen CSAGE|cancer/testis antigen family 24, member 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152727484, end_position_on_the_genomic_accession=152733736, orientation='minus', exon_count=5, OMIM=300944, no_hdr0=''),
  105371270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371270, CurrentID=0, Status='live', Symbol='LOC105371270', Aliases='', description='uncharacterized LOC105371270', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53560728, end_position_on_the_genomic_accession=53580951, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100144602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144602, CurrentID=0, Status='live', Symbol='EPHA5-AS1', Aliases='', description='EPHA5 antisense RNA 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65669961, end_position_on_the_genomic_accession=65693386, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100144603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144603, CurrentID=0, Status='live', Symbol='CHKB-AS1', Aliases='', description='CHKB antisense RNA 1 (head to head)', other_designations='CTA-384D8.20', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50583026, end_position_on_the_genomic_accession=50595281, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271607, CurrentID=0, Status='live', Symbol='RPS15AP30', Aliases='RPS15A_8_1098', description='ribosomal protein S15a pseudogene 30', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112987482, end_position_on_the_genomic_accession=112987864, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422111, CurrentID=0, Status='live', Symbol='LOC100422111', Aliases='', description='olfactory receptor, family 7, subfamily A, member 10 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144604, CurrentID=0, Status='live', Symbol='LINC00930', Aliases='', description='long intergenic non-protein coding RNA 930', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92567818, end_position_on_the_genomic_accession=92572263, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  164656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164656, CurrentID=0, Status='live', Symbol='TMPRSS6', Aliases='IRIDA', description='transmembrane protease, serine 6', other_designations='matriptase-2|membrane-bound mosaic serine proteinase matriptase-2|type II transmembrane serine protease 6', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37065436, end_position_on_the_genomic_accession=37110635, orientation='minus', exon_count=22, OMIM=609862, no_hdr0=''),
  164668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164668, CurrentID=0, Status='live', Symbol='APOBEC3H', Aliases='A3H, ARP-10, ARP10', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3H', other_designations='APOBEC-related protein 10', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39097223, end_position_on_the_genomic_accession=39104067, orientation='plus', exon_count=6, OMIM=610976, no_hdr0=''),
  100616405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616405, CurrentID=0, Status='live', Symbol='MIR4518', Aliases='mir-4518', description='microRNA 4518', other_designations='hsa-mir-4518', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30503919, end_position_on_the_genomic_accession=30504001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479755, CurrentID=0, Status='live', Symbol='RNU6-429P', Aliases='', description='RNA, U6 small nuclear 429, pseudogene', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479756, CurrentID=0, Status='live', Symbol='RNU6-430P', Aliases='', description='RNA, U6 small nuclear 430, pseudogene', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873912, CurrentID=0, Status='live', Symbol='DUTP6', Aliases='', description='deoxyuridine triphosphatase pseudogene 6', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166868734, end_position_on_the_genomic_accession=166869207, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  164684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164684, CurrentID=0, Status='live', Symbol='WBP2NL', Aliases='PAWP', description='WBP2 N-terminal like', other_designations='WW domain-binding protein 2-like', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41995346, end_position_on_the_genomic_accession=42059574, orientation='plus', exon_count=10, OMIM=610981, no_hdr0=''),
  105369368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369368, CurrentID=0, Status='live', Symbol='LOC105369368', Aliases='', description='uncharacterized LOC105369368', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69235246, end_position_on_the_genomic_accession=69251181, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996579, CurrentID=0, Status='live', Symbol='LOC100996579', Aliases='', description='uncharacterized LOC100996579', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161244739, end_position_on_the_genomic_accession=161249050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642677, CurrentID=0, Status='live', Symbol='LOC642677', Aliases='', description='stabilizer of axonemal microtubules 2 pseudogene', other_designations='family with sequence similarity 154, member B pseudogene', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84053117, end_position_on_the_genomic_accession=84054062, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100303715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303715, CurrentID=0, Status='live', Symbol='IDDM24', Aliases='', description='Diabetes mellitus, insulin-dependent, 24', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613006, no_hdr0=''),
  100303716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303716, CurrentID=0, Status='live', Symbol='PBC2', Aliases='', description='Biliary cirrhosis, primary, 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613007, no_hdr0=''),
  100303717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303717, CurrentID=0, Status='live', Symbol='PBC3', Aliases='', description='Biliary cirrhosis, primary, 3', other_designations='', map_location='1p31.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613008, no_hdr0=''),
  164714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164714, CurrentID=0, Status='live', Symbol='TTLL8', Aliases='', description='tubulin tyrosine ligase like 8', other_designations='tubulin tyrosine ligase-like family member 8|tubulin tyrosine ligase-like family, member 8|tubulin--tyrosine ligase-like protein 8', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50018575, end_position_on_the_genomic_accession=50059132, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105369369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369369, CurrentID=0, Status='live', Symbol='LOC105369369', Aliases='', description='uncharacterized LOC105369369', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69273082, end_position_on_the_genomic_accession=69293989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100303728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303728, CurrentID=0, Status='live', Symbol='SLC25A5-AS1', Aliases='', description='SLC25A5 antisense RNA 1', other_designations='SLC25A5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119466033, end_position_on_the_genomic_accession=119469120, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  114836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114836, CurrentID=0, Status='live', Symbol='SLAMF6', Aliases='CD352, KALI, KALIb, Ly108, NTB-A, NTBA, SF2000', description='SLAM family member 6', other_designations='NK-T-B-antigen|NTBA receptor|activating NK receptor|natural killer-, T- and B-cell antigen', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160485030, end_position_on_the_genomic_accession=160523262, orientation='minus', exon_count=8, OMIM=606446, no_hdr0=''),
  100303743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303743, CurrentID=0, Status='live', Symbol='LSM3P5', Aliases='', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 5', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated pseudogene 5|LSM3 homolog, U6 small nuclear RNA associated pseudogene 5|LSM3 pseudogene 5', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78355494, end_position_on_the_genomic_accession=78356060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100303745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303745, CurrentID=0, Status='live', Symbol='LSM3P2', Aliases='', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 2', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated pseudogene 2|LSM3 homolog, U6 small nuclear RNA associated pseudogene 2|LSM3 pseudogene 2', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94590738, end_position_on_the_genomic_accession=94591378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100303748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303748, CurrentID=0, Status='live', Symbol='LSM3P1', Aliases='', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 1', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated pseudogene 1|LSM3 homolog, U6 small nuclear RNA associated pseudogene 1|LSM3 pseudogene 1', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48211280, end_position_on_the_genomic_accession=48211786, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100303749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303749, CurrentID=0, Status='live', Symbol='LOC100303749', Aliases='', description='LSM3 homolog, U6 small nuclear RNA associated (S. cerevisiae) pseudogene', other_designations='', map_location='5q', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65944805, end_position_on_the_genomic_accession=65945573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  647070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647070, CurrentID=0, Status='live', Symbol='LOC647070', Aliases='', description='uncharacterized LOC647070', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182836005, end_position_on_the_genomic_accession=182840325, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100303755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100303755, CurrentID=0, Status='live', Symbol='PET117', Aliases='CSRP2BP', description='PET117 homolog', other_designations='alternative protein CSRP2BP', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18137855, end_position_on_the_genomic_accession=18143169, orientation='plus', exon_count=2, OMIM=614771, no_hdr0=''),
  106479770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479770, CurrentID=0, Status='live', Symbol='RNU6-462P', Aliases='', description='RNA, U6 small nuclear 462, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271611, CurrentID=0, Status='live', Symbol='RPL36AP53', Aliases='RPL36A_27_1803', description='ribosomal protein L36a pseudogene 53', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114589583, end_position_on_the_genomic_accession=114589985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479771, CurrentID=0, Status='live', Symbol='RNU6-464P', Aliases='', description='RNA, U6 small nuclear 464, pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369371, CurrentID=0, Status='live', Symbol='LOC105369371', Aliases='', description='uncharacterized LOC105369371', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70065376, end_position_on_the_genomic_accession=70071175, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479772, CurrentID=0, Status='live', Symbol='RNU6-467P', Aliases='', description='RNA, U6 small nuclear 467, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  164781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164781, CurrentID=0, Status='live', Symbol='DAW1', Aliases='ODA16, WDR69', description='dynein assembly factor with WDR repeat domains 1', other_designations='WD repeat domain 69|WD repeat-containing protein 69|outer row dynein assembly 16 homolog|outer row dynein assembly protein 16 homolog', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227871054, end_position_on_the_genomic_accession=227924347, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  442525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442525, CurrentID=0, Status='live', Symbol='SLC25A5P5', Aliases='ANTP5, TCAG_29001', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 5', other_designations='adenine nucleotide translocator pseudogene 5 (clone 19)', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32471429, end_position_on_the_genomic_accession=32473001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533483, CurrentID=0, Status='live', Symbol='DYX1C1-CCPG1', Aliases='', description='DYX1C1-CCPG1 readthrough (NMD candidate)', other_designations='DYX1C1-CCPG1 readthrough (non-protein coding)', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55355223, end_position_on_the_genomic_accession=55498584, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100271612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271612, CurrentID=0, Status='live', Symbol='RPL36AP54', Aliases='RPL36A_28_1814', description='ribosomal protein L36a pseudogene 54', other_designations='', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134522700, end_position_on_the_genomic_accession=134523000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479777, CurrentID=0, Status='live', Symbol='RNU6-476P', Aliases='', description='RNA, U6 small nuclear 476, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144631, CurrentID=0, Status='live', Symbol='PPP6CP', Aliases='', description='protein phosphatase 6, catalytic subunit pseudogene', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109638112, end_position_on_the_genomic_accession=109638606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100144632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144632, CurrentID=0, Status='live', Symbol='HPBP', Aliases='', description='phosphate binding apolipoprotein', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  164832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164832, CurrentID=0, Status='live', Symbol='LONRF2', Aliases='RNF192', description='LON peptidase N-terminal domain and ring finger 2', other_designations='RING finger protein 192|neuroblastoma apoptosis-related protease', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100273291, end_position_on_the_genomic_accession=100322733, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  164837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=164837, CurrentID=0, Status='live', Symbol='CYCSP7', Aliases='HCP7', description='cytochrome c, somatic pseudogene 7', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100369829, end_position_on_the_genomic_accession=100370551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479784, CurrentID=0, Status='live', Symbol='RNU6-491P', Aliases='', description='RNA, U6 small nuclear 491, pseudogene', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79686, CurrentID=0, Status='live', Symbol='LINC00341', Aliases='C14orf139, NCRNA00341', description='long intergenic non-protein coding RNA 341', other_designations='', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95407266, end_position_on_the_genomic_accession=95410090, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100996611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996611, CurrentID=0, Status='live', Symbol='LOC100996611', Aliases='', description='uncharacterized LOC100996611', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23321208, end_position_on_the_genomic_accession=23323757, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  377007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377007, CurrentID=0, Status='live', Symbol='KLHL30', Aliases='', description='kelch like family member 30', other_designations='kelch-like 30|kelch-like family member 30', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238138722, end_position_on_the_genomic_accession=238152906, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22949, CurrentID=0, Status='live', Symbol='PTGR1', Aliases='LTB4DH, PGR1, ZADH3', description='prostaglandin reductase 1', other_designations='15-oxoprostaglandin 13-reductase|NADP-dependent leukotriene B4 12-hydroxydehydrogenase|PRG-1|leukotriene B4 12-hydroxydehydrogenase|zinc binding alcohol dehydrogenase domain containing 3', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111549722, end_position_on_the_genomic_accession=111599855, orientation='minus', exon_count=14, OMIM=601274, no_hdr0=''),
  100271616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271616, CurrentID=0, Status='live', Symbol='RPL23AP58', Aliases='RPL23A_23_1090', description='ribosomal protein L23a pseudogene 58', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102572757, end_position_on_the_genomic_accession=102573219, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479796, CurrentID=0, Status='live', Symbol='RNU6-519P', Aliases='', description='RNA, U6 small nuclear 519, pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369376, CurrentID=0, Status='live', Symbol='LOC105369376', Aliases='', description='uncharacterized LOC105369376', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72525470, end_position_on_the_genomic_accession=72535026, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616414, CurrentID=0, Status='live', Symbol='MIR4761', Aliases='', description='microRNA 4761', other_designations='hsa-mir-4761', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19963753, end_position_on_the_genomic_accession=19963834, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271617, CurrentID=0, Status='live', Symbol='RPL23AP62', Aliases='RPL23A_24_1177', description='ribosomal protein L23a pseudogene 62', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112461483, end_position_on_the_genomic_accession=112461920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271618, CurrentID=0, Status='live', Symbol='RPL23AP59', Aliases='RPL23A_25_1093', description='ribosomal protein L23a pseudogene 59', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106550106, end_position_on_the_genomic_accession=106550484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271619, CurrentID=0, Status='live', Symbol='RPL23AP70', Aliases='RPL23A_27_1353', description='ribosomal protein L23a pseudogene 70', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35144021, end_position_on_the_genomic_accession=35144480, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372318, CurrentID=0, Status='live', Symbol='LOC105372318', Aliases='', description='uncharacterized LOC105372318', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20611223, end_position_on_the_genomic_accession=20617784, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930137, CurrentID=0, Status='live', Symbol='LOC101930137', Aliases='', description='uncharacterized LOC101930137', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241482075, end_position_on_the_genomic_accession=241483457, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283737, CurrentID=0, Status='live', Symbol='LOC283737', Aliases='', description='uncharacterized LOC283737', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271620, CurrentID=0, Status='live', Symbol='RPL23AP66', Aliases='RPL23A_30_1207', description='ribosomal protein L23a pseudogene 66', other_designations='', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12338776, end_position_on_the_genomic_accession=12339254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369380, CurrentID=0, Status='live', Symbol='LOC105369380', Aliases='', description='uncharacterized LOC105369380', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72793926, end_position_on_the_genomic_accession=72810568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  165055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165055, CurrentID=0, Status='live', Symbol='CCDC138', Aliases='', description='coiled-coil domain containing 138', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108786750, end_position_on_the_genomic_accession=108876769, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  646171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646171, CurrentID=0, Status='live', Symbol='EIF4A2P1', Aliases='', description='eukaryotic translation initiation factor 4A2 pseudogene 1', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12912803, end_position_on_the_genomic_accession=12914106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479820, CurrentID=0, Status='live', Symbol='RNU6-574P', Aliases='', description='RNA, U6 small nuclear 574, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271621, CurrentID=0, Status='live', Symbol='RPL23AP72', Aliases='RPL23A_30_1480', description='ribosomal protein L23a pseudogene 72', other_designations='', map_location='16q11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47150132, end_position_on_the_genomic_accession=47150566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  165082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165082, CurrentID=0, Status='live', Symbol='ADGRF3', Aliases='GPR113, PGR23, hGPCR37', description='adhesion G protein-coupled receptor F3', other_designations='G protein-coupled receptor 113|G protein-coupled receptor PGR23|G-protein coupled receptor 113|G-protein coupled receptor PGR23|probable G-protein coupled receptor 113|seven transmembrane helix receptor', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26308173, end_position_on_the_genomic_accession=26346817, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100996645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996645, CurrentID=0, Status='live', Symbol='LINCR-0003', Aliases='', description='uncharacterized LincR-0003', other_designations='CTD-2023N9.2|CTD-2023N9.3', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57570342, end_position_on_the_genomic_accession=57617430, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  165100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165100, CurrentID=0, Status='live', Symbol='C2orf57', Aliases='', description='chromosome 2 open reading frame 57', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231592864, end_position_on_the_genomic_accession=231594283, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479826, CurrentID=0, Status='live', Symbol='RNU6-588P', Aliases='', description='RNA, U6 small nuclear 588, pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245972, CurrentID=0, Status='live', Symbol='ATP6V0D2', Aliases='ATP6D2, VMA6', description='ATPase, H+ transporting, lysosomal 38kDa, V0 subunit d2', other_designations='V-ATPase subunit d 2|vacuolar proton pump subunit d 2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86098910, end_position_on_the_genomic_accession=86154225, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105373747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373747, CurrentID=0, Status='live', Symbol='LOC105373747', Aliases='', description='uncharacterized LOC105373747', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174210410, end_position_on_the_genomic_accession=174213844, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479830, CurrentID=0, Status='live', Symbol='RNU6-598P', Aliases='', description='RNA, U6 small nuclear 598, pseudogene', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  165140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165140, CurrentID=0, Status='live', Symbol='OXER1', Aliases='GPCR, GPR170, TG1019', description='oxoeicosanoid (OXE) receptor 1', other_designations='5-oxo-ETE G-protein coupled receptor|5-oxo-ETE acid G-protein-coupled receptor 1|G-protein coupled receptor 170|G-protein coupled receptor R527|G-protein coupled receptor TG1019', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42762499, end_position_on_the_genomic_accession=42764261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271624, CurrentID=0, Status='live', Symbol='RPL23AP78', Aliases='RPL23A_34_1620', description='ribosomal protein L23a pseudogene 78', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8728735, end_position_on_the_genomic_accession=8729204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479838, CurrentID=0, Status='live', Symbol='RNU6-616P', Aliases='', description='RNA, U6 small nuclear 616, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  165186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165186, CurrentID=0, Status='live', Symbol='FAM179A', Aliases='', description='family with sequence similarity 179 member A', other_designations='family with sequence similarity 179, member A', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28981298, end_position_on_the_genomic_accession=29052230, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100271625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271625, CurrentID=0, Status='live', Symbol='RPL23AP73', Aliases='RPL23A_36_1508', description='ribosomal protein L23a pseudogene 73', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6749718, end_position_on_the_genomic_accession=6749931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653155, CurrentID=0, Status='live', Symbol='LOC653155', Aliases='', description='pre-mRNA processing factor 4B pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116748140, end_position_on_the_genomic_accession=116748877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642721, CurrentID=0, Status='live', Symbol='CBX3P6', Aliases='', description='chromobox 3 pseudogene 6', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 6', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154940541, end_position_on_the_genomic_accession=154941067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  165215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165215, CurrentID=0, Status='live', Symbol='FAM171B', Aliases='KIAA1946', description='family with sequence similarity 171 member B', other_designations='family with sequence similarity 171, member B', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186694062, end_position_on_the_genomic_accession=186765959, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100271626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271626, CurrentID=0, Status='live', Symbol='RPL23AP79', Aliases='RPL23A_39_1676', description='ribosomal protein L23a pseudogene 79', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58598970, end_position_on_the_genomic_accession=58599435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103581030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103581030, CurrentID=0, Status='live', Symbol='TRN-ATT1-2', Aliases='', description='transfer RNA-Asn (ATT) 1-2', other_designations='tRNA-Asn-ATT-1-2', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103581031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103581031, CurrentID=0, Status='live', Symbol='CASC23', Aliases='LINC01464', description='cancer susceptibility candidate 23 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1464', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8011278, end_position_on_the_genomic_accession=8016509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369386, CurrentID=0, Status='live', Symbol='LOC105369386', Aliases='', description='uncharacterized LOC105369386', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74575982, end_position_on_the_genomic_accession=74578375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131962, CurrentID=0, Status='live', Symbol='LOC100131962', Aliases='', description='ubiquitin conjugating enzyme E2D 4 (putative) pseudogene', other_designations='ubiquitin-conjugating enzyme E2D 4 (putative) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143846099, end_position_on_the_genomic_accession=143846504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479851, CurrentID=0, Status='live', Symbol='RNU6-643P', Aliases='', description='RNA, U6 small nuclear 643, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  165257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165257, CurrentID=0, Status='live', Symbol='C1QL2', Aliases='C1QTNF10, CTRP10', description='complement component 1, q subcomponent-like 2', other_designations='C1q and tumor necrosis factor related protein 10|C1q and tumor necrosis factor-related protein 10|C1q-domain containing protein|C1q/TNF-related protein 10', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119156243, end_position_on_the_genomic_accession=119158895, orientation='minus', exon_count=2, OMIM=614330, no_hdr0=''),
  106479853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479853, CurrentID=0, Status='live', Symbol='RNU6-648P', Aliases='', description='RNA, U6 small nuclear 648, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103581096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103581096, CurrentID=0, Status='live', Symbol='OFC14', Aliases='', description='Orofacial cleft 14', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615892, no_hdr0=''),
  105373753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373753, CurrentID=0, Status='live', Symbol='LOC105373753', Aliases='', description='uncharacterized LOC105373753', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176164164, end_position_on_the_genomic_accession=176168203, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  165324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165324, CurrentID=0, Status='live', Symbol='UBXN2A', Aliases='UBXD4', description='UBX domain protein 2A', other_designations='UBX domain containing 4|UBX domain-containing protein 4', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23927271, end_position_on_the_genomic_accession=24001000, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  441957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441957, CurrentID=0, Status='live', Symbol='FEM1AP1', Aliases='FEM1DP', description='fem-1 homolog a (C. elegans) pseudogene 1', other_designations='fem-1 homolog a pseudogene|fem-1 homolog d pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13762338, end_position_on_the_genomic_accession=13764329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287523, CurrentID=0, Status='live', Symbol='LOC100287523', Aliases='', description='zinc finger protein 2-like', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95190634, end_position_on_the_genomic_accession=95192671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271630, CurrentID=0, Status='live', Symbol='RPL35AP26', Aliases='RPL35A_11_1187', description='ribosomal protein L35a pseudogene 26', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126278202, end_position_on_the_genomic_accession=126278618, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  202319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202319, CurrentID=0, Status='live', Symbol='LOC202319', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A)-like 4G pseudogene', other_designations='', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=50633097, end_position_on_the_genomic_accession=50633804, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105416157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105416157, CurrentID=0, Status='live', Symbol='NKILA', Aliases='', description='NF-kappaB interacting long non-coding RNA', other_designations='NF-KappaB Interacting LncRNA', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57710538, end_position_on_the_genomic_accession=57712775, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369390, CurrentID=0, Status='live', Symbol='LOC105369390', Aliases='', description='uncharacterized LOC105369390', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75569680, end_position_on_the_genomic_accession=75570992, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873935, CurrentID=0, Status='live', Symbol='MTOR-AS1', Aliases='', description='MTOR antisense RNA 1', other_designations='MTOR antisense RNA 1 (non-protein coding)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11143898, end_position_on_the_genomic_accession=11149538, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100289090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289090, CurrentID=0, Status='live', Symbol='LOC100289090', Aliases='', description='uncharacterized LOC100289090', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=42186434, end_position_on_the_genomic_accession=42188144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376521, CurrentID=0, Status='live', Symbol='LOC105376521', Aliases='', description='uncharacterized LOC105376521', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2435067, end_position_on_the_genomic_accession=2444570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102466986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466986, CurrentID=0, Status='live', Symbol='MIR6866', Aliases='hsa-mir-6866', description='microRNA 6866', other_designations='microRNA mir-6866', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40161933, end_position_on_the_genomic_accession=40162001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996699, CurrentID=0, Status='live', Symbol='LOC100996699', Aliases='', description='proline-rich protein 2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11626057, end_position_on_the_genomic_accession=11629489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996701, CurrentID=0, Status='live', Symbol='LOC100996701', Aliases='', description='uncharacterized LOC100996701', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131663774, end_position_on_the_genomic_accession=131665416, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271633, CurrentID=0, Status='live', Symbol='RPL35AP25', Aliases='RPL35AP23, RPL35A_10_1045, RPL35A_13_1043', description='ribosomal protein L35a pseudogene 25', other_designations='ribosomal protein L35a pseudogene 23', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45940454, end_position_on_the_genomic_accession=45940921, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369393, CurrentID=0, Status='live', Symbol='LOC105369393', Aliases='', description='uncharacterized LOC105369393', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75736938, end_position_on_the_genomic_accession=75741350, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130402, CurrentID=0, Status='live', Symbol='B3GALNT2P1', Aliases='', description='beta-1,3-N-acetylgalactosaminyltransferase 2 pseudogene 1', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129794024, end_position_on_the_genomic_accession=129801651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873938, CurrentID=0, Status='live', Symbol='GYG2-AS1', Aliases='', description='GYG2 antisense RNA 1', other_designations='GYG2 antisense RNA 1 (non-protein coding)', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369394, CurrentID=0, Status='live', Symbol='LOC105369394', Aliases='', description='uncharacterized LOC105369394', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76208980, end_position_on_the_genomic_accession=76211532, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642730, CurrentID=0, Status='live', Symbol='LOC642730', Aliases='', description='uncharacterized LOC642730', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=92039510, end_position_on_the_genomic_accession=92066655, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928573, CurrentID=0, Status='live', Symbol='LOC101928573', Aliases='', description='uncharacterized LOC101928573', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20265726, end_position_on_the_genomic_accession=20307258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271635, CurrentID=0, Status='live', Symbol='RPL35AP27', Aliases='RPL35A_14_1227', description='ribosomal protein L35a pseudogene 27', other_designations='', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32863031, end_position_on_the_genomic_accession=32863471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873939, CurrentID=0, Status='live', Symbol='LNX1-AS1', Aliases='', description='LNX1 antisense RNA 1', other_designations='LNX1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53500077, end_position_on_the_genomic_accession=53523274, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100101933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101933, CurrentID=0, Status='live', Symbol='GNG5P4', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 5 pseudogene 4', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73418428, end_position_on_the_genomic_accession=73418911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402682, CurrentID=0, Status='live', Symbol='UFSP1', Aliases='UFSP', description='UFM1-specific peptidase 1 (inactive)', other_designations='UFM1-specific peptidase 1 (non-functional)', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100888723, end_position_on_the_genomic_accession=100889718, orientation='minus', exon_count=1, OMIM=611481, no_hdr0=''),
  105376821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376821, CurrentID=0, Status='live', Symbol='LOC105376821', Aliases='', description='uncharacterized LOC105376821', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144748, CurrentID=0, Status='live', Symbol='KLLN', Aliases='CWS4, KILLIN', description='killin, p53-regulated DNA replication inhibitor', other_designations='', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87859161, end_position_on_the_genomic_accession=87863437, orientation='minus', exon_count=1, OMIM=612105, no_hdr0=''),
  100271636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271636, CurrentID=0, Status='live', Symbol='RPL35AP35', Aliases='RPL35A_14_1534', description='ribosomal protein L35a pseudogene 35', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29340438, end_position_on_the_genomic_accession=29340870, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873940, CurrentID=0, Status='live', Symbol='MYLK-AS2', Aliases='', description='MYLK antisense RNA 2', other_designations='MYLK antisense RNA 2 (non-protein coding)', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123689644, end_position_on_the_genomic_accession=123692407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369396, CurrentID=0, Status='live', Symbol='LOC105369396', Aliases='', description='uncharacterized LOC105369396', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76767363, end_position_on_the_genomic_accession=76768659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  165530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165530, CurrentID=0, Status='live', Symbol='CLEC4F', Aliases='CLECSF13, KCLR', description='C-type lectin domain family 4 member F', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 13|C-type lectin 13|C-type lectin domain family 4, member F|C-type lectin superfamily member 13', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70808643, end_position_on_the_genomic_accession=70823764, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  165545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165545, CurrentID=0, Status='live', Symbol='DQX1', Aliases='', description='DEAQ box RNA-dependent ATPase 1', other_designations='DEAQ RNA-dependent ATPase DQX1|DEAQ box polypeptide 1 (RNA-dependent ATPase)', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74518131, end_position_on_the_genomic_accession=74526281, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100271637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271637, CurrentID=0, Status='live', Symbol='RPL35AP28', Aliases='RPL35A_15_1235', description='ribosomal protein L35a pseudogene 28', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49863106, end_position_on_the_genomic_accession=49863541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420205, CurrentID=0, Status='live', Symbol='LOC100420205', Aliases='', description='ribonucleotide reductase M2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79984248, end_position_on_the_genomic_accession=79985721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873941, CurrentID=0, Status='live', Symbol='OPA1-AS1', Aliases='', description='OPA1 antisense RNA 1', other_designations='OPA1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193618599, end_position_on_the_genomic_accession=193627332, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369397, CurrentID=0, Status='live', Symbol='LOC105369397', Aliases='', description='uncharacterized LOC105369397', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76752687, end_position_on_the_genomic_accession=76758136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22971, CurrentID=0, Status='live', Symbol='RPL41P1', Aliases='RPL41L2, dJ1065O2.1', description='ribosomal protein L41 pseudogene 1', other_designations='ribosomal protein L41-like 2', map_location='20p11.23-p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21755230, end_position_on_the_genomic_accession=21755652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271638, CurrentID=0, Status='live', Symbol='RPL35AP29', Aliases='RPL35A_16_1237', description='ribosomal protein L35a pseudogene 29', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51106822, end_position_on_the_genomic_accession=51107144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873942, CurrentID=0, Status='live', Symbol='PHEX-AS1', Aliases='', description='PHEX antisense RNA 1', other_designations='PHEX antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22162732, end_position_on_the_genomic_accession=22172983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106479907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479907, CurrentID=0, Status='live', Symbol='RNU6-771P', Aliases='', description='RNA, U6 small nuclear 771, pseudogene', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130403, CurrentID=0, Status='live', Symbol='LOC100130403', Aliases='', description='replication protein A3, 14kDa pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53378503, end_position_on_the_genomic_accession=53378863, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479908, CurrentID=0, Status='live', Symbol='RNU6-772P', Aliases='', description='RNA, U6 small nuclear 772, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479909, CurrentID=0, Status='live', Symbol='RNU6-775P', Aliases='', description='RNA, U6 small nuclear 775, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271639, CurrentID=0, Status='live', Symbol='RPL35AP30', Aliases='RPL35A_17_1286', description='ribosomal protein L35a pseudogene 30', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119959424, end_position_on_the_genomic_accession=119959757, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873943, CurrentID=0, Status='live', Symbol='PLS1-AS1', Aliases='', description='PLS1 antisense RNA 1', other_designations='PLS1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288908, CurrentID=0, Status='live', Symbol='LOC100288908', Aliases='', description='PRO2898', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=5275316, end_position_on_the_genomic_accession=5276394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  165631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165631, CurrentID=0, Status='live', Symbol='PARP15', Aliases='ARTD7, BAL3, pART7', description='poly(ADP-ribose) polymerase family member 15', other_designations='ADP-ribosyltransferase diphtheria toxin-like 7|B-aggressive lymphoma 3|B-aggressive lymphoma protein 3|PARP-15|poly (ADP-ribose) polymerase family, member 15', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122577454, end_position_on_the_genomic_accession=122639047, orientation='plus', exon_count=14, OMIM=612066, no_hdr0=''),
  441907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441907, CurrentID=0, Status='live', Symbol='LOC441907', Aliases='', description='ribosomal protein L6 pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150053845, end_position_on_the_genomic_accession=150055040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442668, CurrentID=0, Status='live', Symbol='NECAP1P1', Aliases='', description='NECAP endocytosis associated 1 pseudogene 1', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37582855, end_position_on_the_genomic_accession=37586434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873944, CurrentID=0, Status='live', Symbol='PRKX-AS1', Aliases='', description='PRKX antisense RNA 1', other_designations='PRKX antisense RNA 1 (non-protein coding)', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3659487, end_position_on_the_genomic_accession=3668192, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100529147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529147, CurrentID=0, Status='live', Symbol='MMRFCGU', Aliases='', description='Microcephaly, mental retardation, and distinctive facies, with cardiac and genitourinary malformations', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613680, no_hdr0=''),
  100996741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996741, CurrentID=0, Status='live', Symbol='LOC100996741', Aliases='', description='uncharacterized LOC100996741', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145164099, end_position_on_the_genomic_accession=145207256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  153910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153910, CurrentID=0, Status='live', Symbol='LOC153910', Aliases='', description='uncharacterized LOC153910', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142526455, end_position_on_the_genomic_accession=142637889, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479921, CurrentID=0, Status='live', Symbol='RNU6-804P', Aliases='', description='RNA, U6 small nuclear 804, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873945, CurrentID=0, Status='live', Symbol='HLTF-AS1', Aliases='', description='HLTF antisense RNA 1', other_designations='HLTF antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149086332, end_position_on_the_genomic_accession=149102823, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369401, CurrentID=0, Status='live', Symbol='LOC105369401', Aliases='', description='uncharacterized LOC105369401', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78081401, end_position_on_the_genomic_accession=78087362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  165679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165679, CurrentID=0, Status='live', Symbol='SPTSSB', Aliases='ADMP, C3orf57, SSSPTB', description='serine palmitoyltransferase small subunit B', other_designations='androgen down regulated in mouse prostate|likely ortholog of androgen down regulated gene expressed in mouse prostate|serine palmitoyltransferase, small subunit B|small subunit of serine palmitoyltransferase B', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161344792, end_position_on_the_genomic_accession=161372938, orientation='minus', exon_count=4, OMIM=610412, no_hdr0=''),
  100131977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131977, CurrentID=0, Status='live', Symbol='LOC100131977', Aliases='', description='prickle homolog 1 (Drosophila) pseudogene', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183613563, end_position_on_the_genomic_accession=183615266, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442676, CurrentID=0, Status='live', Symbol='CICP8', Aliases='', description='capicua transcriptional repressor pseudogene 8', other_designations='capicua homolog pseudogene 8', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56362416, end_position_on_the_genomic_accession=56364243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271642, CurrentID=0, Status='live', Symbol='RPL36AP38', Aliases='RPL36A_16_1155', description='ribosomal protein L36a pseudogene 38', other_designations='', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74738480, end_position_on_the_genomic_accession=74738796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369402, CurrentID=0, Status='live', Symbol='LOC105369402', Aliases='', description='uncharacterized LOC105369402', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78296148, end_position_on_the_genomic_accession=78319731, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  165721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165721, CurrentID=0, Status='live', Symbol='DNAJB8', Aliases='CT156, DJ6', description='DnaJ heat shock protein family (Hsp40) member B8', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 8', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128462432, end_position_on_the_genomic_accession=128467248, orientation='minus', exon_count=3, OMIM=611337, no_hdr0=''),
  105369403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369403, CurrentID=0, Status='live', Symbol='LOC105369403', Aliases='', description='uncharacterized LOC105369403', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78473146, end_position_on_the_genomic_accession=78497594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  46 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=46, CurrentID=0, Status='live', Symbol='ACLS', Aliases='', description='acrocallosal syndrome', other_designations='', map_location='12p13.3-p11.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=200990, no_hdr0=''),
  100271644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271644, CurrentID=0, Status='live', Symbol='RPL36AP41', Aliases='RPL36A_17_1248', description='ribosomal protein L36a pseudogene 41', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63871739, end_position_on_the_genomic_accession=63872044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  5781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=5781, CurrentID=0, Status='live', Symbol='PTPN11', Aliases='BPTP3, CFC, JMML, METCDS, NS1, PTP-1D, PTP2C, SH-PTP2, SH-PTP3, SHP2', description='protein tyrosine phosphatase, non-receptor type 11', other_designations='PTP-2C|protein-tyrosine phosphatase 1D|protein-tyrosine phosphatase 2C', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112418732, end_position_on_the_genomic_accession=112509918, orientation='plus', exon_count=16, OMIM=176876, no_hdr0=''),
  105369404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369404, CurrentID=0, Status='live', Symbol='LOC105369404', Aliases='', description='uncharacterized LOC105369404', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=79437756, end_position_on_the_genomic_accession=79438948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373769, CurrentID=0, Status='live', Symbol='LOC105373769', Aliases='', description='uncharacterized LOC105373769', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180406866, end_position_on_the_genomic_accession=180432380, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271645, CurrentID=0, Status='live', Symbol='RPL36AP42', Aliases='RPL36A_18_1363', description='ribosomal protein L36a pseudogene 42', other_designations='', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49618742, end_position_on_the_genomic_accession=49619119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873949, CurrentID=0, Status='live', Symbol='IPO9-AS1', Aliases='', description='IPO9 antisense RNA 1', other_designations='IPO9 antisense RNA 1 (non-protein coding)', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201688256, end_position_on_the_genomic_accession=201829559, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  442695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442695, CurrentID=0, Status='live', Symbol='ZNF722P', Aliases='ZNF722', description='zinc finger protein 722, pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  140821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140821, CurrentID=0, Status='live', Symbol='RSS', Aliases='SRS', description='Russell Silver syndrome', other_designations='Russell-Silver dwarfism|Russell-Silver syndrome', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=180860, no_hdr0=''),
  100131981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131981, CurrentID=0, Status='live', Symbol='UBE2V1P9', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 9', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 9|ubiquitin-conjugating enzyme E2 variant 2 pseudogene', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=90289539, end_position_on_the_genomic_accession=90289849, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271646, CurrentID=0, Status='live', Symbol='RPL36AP43', Aliases='RPL36A_21_1460', description='ribosomal protein L36a pseudogene 43', other_designations='', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89687882, end_position_on_the_genomic_accession=89688280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873950, CurrentID=0, Status='live', Symbol='ZIC4-AS1', Aliases='', description='ZIC4 antisense RNA 1', other_designations='ZIC4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100909387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100909387, CurrentID=0, Status='live', Symbol='CMH21', Aliases='', description='Cardiomyopathy, familial hypertrophic, 21', other_designations='', map_location='7p12.1-q21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614676, no_hdr0=''),
  165829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165829, CurrentID=0, Status='live', Symbol='GPR156', Aliases='GABABL, PGR28', description='G protein-coupled receptor 156', other_designations='G protein-coupled receptor PGR28|G-protein coupled receptor PGR28|GABAB-related G-protein coupled receptor', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120165481, end_position_on_the_genomic_accession=120285124, orientation='minus', exon_count=15, OMIM=610464, no_hdr0=''),
  100873951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873951, CurrentID=0, Status='live', Symbol='MLIP-AS1', Aliases='', description='MLIP antisense RNA 1', other_designations='MLIP antisense RNA 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369407, CurrentID=0, Status='live', Symbol='LOC105369407', Aliases='', description='uncharacterized LOC105369407', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80088736, end_position_on_the_genomic_accession=80168657, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131983, CurrentID=0, Status='live', Symbol='HNRNPA1P25', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 25', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74472680, end_position_on_the_genomic_accession=74474013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373772, CurrentID=0, Status='live', Symbol='LOC105373772', Aliases='', description='uncharacterized LOC105373772', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180961079, end_position_on_the_genomic_accession=180980199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  442707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442707, CurrentID=0, Status='live', Symbol='MTHFD2P5', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 5', other_designations='methylene tetrahydrofolate dehydrogenase 2 pseudogene', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82588770, end_position_on_the_genomic_accession=82590763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271648, CurrentID=0, Status='live', Symbol='RPL36AP50', Aliases='RPL36A_26_1671', description='ribosomal protein L36a pseudogene 50', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54973817, end_position_on_the_genomic_accession=54974089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100101119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101119, CurrentID=0, Status='live', Symbol='SKP1P3', Aliases='', description='S-phase kinase-associated protein 1 pseudogene 3', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39485392, end_position_on_the_genomic_accession=39486103, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442709, CurrentID=0, Status='live', Symbol='EEF1A1P28', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 28', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88638952, end_position_on_the_genomic_accession=88640622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100101121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101121, CurrentID=0, Status='live', Symbol='TTTY23B', Aliases='TTTY23', description='testis-specific transcript, Y-linked 23B (non-protein coding)', other_designations='testis-specific transcript, Y-linked 23', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6242446, end_position_on_the_genomic_accession=6243610, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  165904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165904, CurrentID=0, Status='live', Symbol='XIRP1', Aliases='CMYA1, Xin', description='xin actin binding repeat containing 1', other_designations='cardiomyopathy associated 1|cardiomyopathy-associated protein 1|xin actin-binding repeat containing 1', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39183215, end_position_on_the_genomic_accession=39192620, orientation='minus', exon_count=3, OMIM=609777, no_hdr0=''),
  106479960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479960, CurrentID=0, Status='live', Symbol='RNU6-891P', Aliases='', description='RNA, U6 small nuclear 891, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479961, CurrentID=0, Status='live', Symbol='RNU6-893P', Aliases='', description='RNA, U6 small nuclear 893, pseudogene', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873953, CurrentID=0, Status='live', Symbol='WWC3-AS1', Aliases='', description='WWC3 antisense RNA 1', other_designations='WWC3 antisense RNA 1 (non-protein coding)', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  165918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=165918, CurrentID=0, Status='live', Symbol='RNF168', Aliases='hRNF168', description='ring finger protein 168, E3 ubiquitin protein ligase', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196468783, end_position_on_the_genomic_accession=196503768, orientation='minus', exon_count=7, OMIM=612688, no_hdr0=''),
  100101126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101126, CurrentID=0, Status='live', Symbol='LOC100101126', Aliases='', description='thioredoxin-related transmembrane protein 1 pseudogene', other_designations='', map_location='7p', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39567618, end_position_on_the_genomic_accession=39570231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480741, CurrentID=0, Status='live', Symbol='ZNF346-IT1', Aliases='', description='ZNF346 intronic transcript 1', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479965, CurrentID=0, Status='live', Symbol='RNU6-901P', Aliases='', description='RNA, U6 small nuclear 901, pseudogene', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873954, CurrentID=0, Status='live', Symbol='SNRK-AS1', Aliases='', description='SNRK antisense RNA 1', other_designations='SNRK antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43346919, end_position_on_the_genomic_accession=43351962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  442720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442720, CurrentID=0, Status='live', Symbol='EIF3IP1', Aliases='RG208K23', description='eukaryotic translation initiation factor 3 subunit I pseudogene 1', other_designations='eukaryotic translation initiation factor 3 subunit 2 pseudogene|eukaryotic translation initiation factor 3, subunit I pseudogene 1', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=109959227, end_position_on_the_genomic_accession=109960213, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479969, CurrentID=0, Status='live', Symbol='RNU6-911P', Aliases='', description='RNA, U6 small nuclear 911, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271651, CurrentID=0, Status='live', Symbol='RPS15AP37', Aliases='RPS15A_13_1746', description='ribosomal protein S15a pseudogene 37', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31112994, end_position_on_the_genomic_accession=31113428, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479971, CurrentID=0, Status='live', Symbol='RNU6-917P', Aliases='', description='RNA, U6 small nuclear 917, pseudogene', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873955, CurrentID=0, Status='live', Symbol='LNX1-AS2', Aliases='', description='LNX1 antisense RNA 2', other_designations='LNX1 antisense RNA 2 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53592956, end_position_on_the_genomic_accession=53605382, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479974, CurrentID=0, Status='live', Symbol='RNU6-922P', Aliases='', description='RNA, U6 small nuclear 922, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442727, CurrentID=0, Status='live', Symbol='LOC442727', Aliases='', description='prothymosin, alpha pseudogene', other_designations='prothymosin alpha pseudogene', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138403960, end_position_on_the_genomic_accession=138404651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  166012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166012, CurrentID=0, Status='live', Symbol='CHST13', Aliases='C4ST3', description='carbohydrate (chondroitin 4) sulfotransferase 13', other_designations='C4ST-3|chondroitin 4-O-sulfotransferase 3|chondroitin 4-sulfotransferase 3', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126524288, end_position_on_the_genomic_accession=126543291, orientation='plus', exon_count=3, OMIM=610124, no_hdr0=''),
  442730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442730, CurrentID=0, Status='live', Symbol='ST13P17', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 17', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139502454, end_position_on_the_genomic_accession=139503814, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373777, CurrentID=0, Status='live', Symbol='LOC105373777', Aliases='', description='uncharacterized LOC105373777', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184140727, end_position_on_the_genomic_accession=184204282, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873957, CurrentID=0, Status='live', Symbol='FABP5P14', Aliases='FABP5L14', description='fatty acid binding protein 5 pseudogene 14', other_designations='fatty acid binding protein 5-like 14 (pseudogene)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216805523, end_position_on_the_genomic_accession=216805992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  84090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84090, CurrentID=0, Status='live', Symbol='IGLVVI-25-1', Aliases='lambdavg3', description='immunoglobulin lambda variable (VI)-25-1 (pseudogene)', other_designations='IGLV(VI)-25-1|immunoglobulin lambda variable (VI)-25-1 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22678934, end_position_on_the_genomic_accession=22679397, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271276, CurrentID=0, Status='live', Symbol='RPL13AP11', Aliases='RPL13A_1_152', description='ribosomal protein L13a pseudogene 11', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204946607, end_position_on_the_genomic_accession=204947169, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379144, CurrentID=0, Status='live', Symbol='LOC105379144', Aliases='', description='uncharacterized LOC105379144', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119835266, end_position_on_the_genomic_accession=119977972, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  88455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=88455, CurrentID=0, Status='live', Symbol='ANKRD13A', Aliases='ANKRD13, NY-REN-25', description='ankyrin repeat domain 13A', other_designations='NY-REN-25 antigen', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109999169, end_position_on_the_genomic_accession=110039432, orientation='plus', exon_count=17, OMIM=615123, no_hdr0=''),
  100271654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271654, CurrentID=0, Status='live', Symbol='RPS15AP33', Aliases='RPS15A_21_1479', description='ribosomal protein S15a pseudogene 33', other_designations='', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28813942, end_position_on_the_genomic_accession=28814327, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060043, CurrentID=0, Status='live', Symbol='LOC101060043', Aliases='', description='sulfiredoxin 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76510406, end_position_on_the_genomic_accession=76511644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131990, CurrentID=0, Status='live', Symbol='LOC100131990', Aliases='', description='uncharacterized LOC100131990', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=168681898, end_position_on_the_genomic_accession=168686463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996813, CurrentID=0, Status='live', Symbol='LOC100996813', Aliases='', description='multidrug and toxin extrusion protein 2-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19577660, end_position_on_the_genomic_accession=19633825, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  442744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442744, CurrentID=0, Status='live', Symbol='LOC442744', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144434826, end_position_on_the_genomic_accession=144438713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271656, CurrentID=0, Status='live', Symbol='LOC100271656', Aliases='', description='N-acetylneuraminic acid phosphatase pseudogene', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108762499, end_position_on_the_genomic_accession=108767640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421482, CurrentID=0, Status='live', Symbol='LOC100421482', Aliases='', description='DnaJ (Hsp40) homolog, subfamily C, member 2 pseudogene', other_designations='DnaJ heat shock protein family (Hsp40) member C2 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112322473, end_position_on_the_genomic_accession=112323096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369416, CurrentID=0, Status='live', Symbol='LOC105369416', Aliases='', description='uncharacterized LOC105369416', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=84594827, end_position_on_the_genomic_accession=84608462, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728376, CurrentID=0, Status='live', Symbol='LOC728376', Aliases='', description='zinc finger protein 430 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56447958, end_position_on_the_genomic_accession=56467539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289607, CurrentID=0, Status='live', Symbol='HSPE1P22', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 22', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 22', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92027498, end_position_on_the_genomic_accession=92027821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873961, CurrentID=0, Status='live', Symbol='TRN-GTT21-1', Aliases='TRNAN38', description='transfer RNA-Asn (GTT) 21-1', other_designations='tRNA-Asn-GTT-21-1|transfer RNA asparagine 38 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369417, CurrentID=0, Status='live', Symbol='LOC105369417', Aliases='', description='uncharacterized LOC105369417', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=84189898, end_position_on_the_genomic_accession=84194286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401767, CurrentID=0, Status='live', Symbol='LOC401767', Aliases='', description='glycerol-3-phosphate acyltransferase 4 pseudogene', other_designations='1-acylglycerol-3-phosphate O-acyltransferase 6 (lysophosphatidic acid acyltransferase, zeta) pseudogene', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25424911, end_position_on_the_genomic_accession=25432236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131280, CurrentID=0, Status='live', Symbol='LOC100131280', Aliases='', description='integrin alpha FG-GAP repeat containing 2 pseudogene', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110982779, end_position_on_the_genomic_accession=110984455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371289, CurrentID=0, Status='live', Symbol='LOC105371289', Aliases='', description='uncharacterized LOC105371289', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57133495, end_position_on_the_genomic_accession=57135445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254312, CurrentID=0, Status='live', Symbol='LINC00710', Aliases='', description='long intergenic non-protein coding RNA 710', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10934940, end_position_on_the_genomic_accession=10952163, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369419, CurrentID=0, Status='live', Symbol='LOC105369419', Aliases='', description='uncharacterized LOC105369419', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83690746, end_position_on_the_genomic_accession=83693964, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480013, CurrentID=0, Status='live', Symbol='RNU6-1014P', Aliases='', description='RNA, U6 small nuclear 1014, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373035, CurrentID=0, Status='live', Symbol='LOC105373035', Aliases='', description='uncharacterized LOC105373035', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39500492, end_position_on_the_genomic_accession=39502150, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100270944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270944, CurrentID=0, Status='live', Symbol='RPL18P5', Aliases='RPL18_3_852', description='ribosomal protein L18 pseudogene 5', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137164031, end_position_on_the_genomic_accession=137164648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369420, CurrentID=0, Status='live', Symbol='LOC105369420', Aliases='', description='uncharacterized LOC105369420', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83306113, end_position_on_the_genomic_accession=83307142, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  158781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=158781, CurrentID=0, Status='live', Symbol='LOC158781', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41673852, end_position_on_the_genomic_accession=41676598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480019, CurrentID=0, Status='live', Symbol='RNU6-1029P', Aliases='', description='RNA, U6 small nuclear 1029, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369421, CurrentID=0, Status='live', Symbol='LOC105369421', Aliases='', description='uncharacterized LOC105369421', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86336615, end_position_on_the_genomic_accession=86353615, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480023, CurrentID=0, Status='live', Symbol='RNU6-1040P', Aliases='', description='RNA, U6 small nuclear 1040, pseudogene', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373786, CurrentID=0, Status='live', Symbol='LOC105373786', Aliases='', description='uncharacterized LOC105373786', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187461502, end_position_on_the_genomic_accession=187549450, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100144878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100144878, CurrentID=0, Status='live', Symbol='LRRC53', Aliases='', description='leucine rich repeat containing 53', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442777, CurrentID=0, Status='live', Symbol='TECRP1', Aliases='', description='trans-2,3-enoyl-CoA reductase pseudogene 1', other_designations='', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86949538, end_position_on_the_genomic_accession=86950105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873966, CurrentID=0, Status='live', Symbol='SPATA2P1', Aliases='', description='spermatogenesis associated 2 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=56214373, end_position_on_the_genomic_accession=56216942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616460, CurrentID=0, Status='live', Symbol='MIR3689E', Aliases='', description='microRNA 3689e', other_designations='hsa-mir-3689e', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134850570, end_position_on_the_genomic_accession=134850641, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873967, CurrentID=0, Status='live', Symbol='LSP1P2', Aliases='', description='lymphocyte-specific protein 1 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75395188, end_position_on_the_genomic_accession=75395322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  166336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166336, CurrentID=0, Status='live', Symbol='PRICKLE2', Aliases='EPM5', description='prickle homolog 2', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64093850, end_position_on_the_genomic_accession=64445476, orientation='minus', exon_count=14, OMIM=608501, no_hdr0=''),
  22997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=22997, CurrentID=0, Status='live', Symbol='IGSF9B', Aliases='', description='immunoglobulin superfamily member 9B', other_designations='immunoglobulin superfamily, member 9B', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133908302, end_position_on_the_genomic_accession=133956963, orientation='minus', exon_count=24, OMIM=613773, no_hdr0=''),
  166348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166348, CurrentID=0, Status='live', Symbol='KBTBD12', Aliases='KLHDC6', description='kelch repeat and BTB (POZ) domain containing 12', other_designations='kelch domain containing 6|kelch domain-containing protein 6', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127915232, end_position_on_the_genomic_accession=127987685, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  115110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115110, CurrentID=0, Status='live', Symbol='LOC115110', Aliases='', description='uncharacterized LOC115110', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2549920, end_position_on_the_genomic_accession=2557011, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  166378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166378, CurrentID=0, Status='live', Symbol='SPATA5', Aliases='AFG2, EHLMRS, SPAF', description='spermatogenesis associated 5', other_designations='ATPase family gene 2 homolog|ATPase family protein 2 homolog|spermatogenesis associated factor SPAF|spermatogenesis-associated factor protein', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122923070, end_position_on_the_genomic_accession=123319450, orientation='plus', exon_count=18, OMIM=613940, no_hdr0=''),
  166379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166379, CurrentID=0, Status='live', Symbol='BBS12', Aliases='C4orf24', description='Bardet-Biedl syndrome 12', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122700444, end_position_on_the_genomic_accession=122744943, orientation='plus', exon_count=4, OMIM=610683, no_hdr0=''),
  100270945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270945, CurrentID=0, Status='live', Symbol='RPL18P7', Aliases='RPL18_4_944', description='ribosomal protein L18 pseudogene 7', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=113376636, end_position_on_the_genomic_accession=113377290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132112, CurrentID=0, Status='live', Symbol='RTP1', Aliases='Z3CXXC1', description='receptor (chemosensory) transporter protein 1', other_designations='3CxxC-type zinc finger protein 1|receptor transporter protein 1|receptor transporting protein 1|zinc finger, 3CxxC-type 1', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187197486, end_position_on_the_genomic_accession=187201465, orientation='plus', exon_count=2, OMIM=609137, no_hdr0=''),
  100616463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616463, CurrentID=0, Status='live', Symbol='MIR4452', Aliases='', description='microRNA 4452', other_designations='hsa-mir-4452', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86542482, end_position_on_the_genomic_accession=86542552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374764, CurrentID=0, Status='live', Symbol='LOC105374764', Aliases='', description='uncharacterized LOC105374764', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62665819, end_position_on_the_genomic_accession=62673770, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369427, CurrentID=0, Status='live', Symbol='LOC105369427', Aliases='', description='uncharacterized LOC105369427', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=91158016, end_position_on_the_genomic_accession=91228398, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373792, CurrentID=0, Status='live', Symbol='LOC105373792', Aliases='', description='uncharacterized LOC105373792', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189296074, end_position_on_the_genomic_accession=189298666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  402715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402715, CurrentID=0, Status='live', Symbol='LOC402715', Aliases='', description='karyopherin alpha 2 (RAG cohort 1, importin alpha 1) pseudogene', other_designations='importin alpha 2 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145010053, end_position_on_the_genomic_accession=145011202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402716, CurrentID=0, Status='live', Symbol='RPL32P17', Aliases='RPL32_9_862', description='ribosomal protein L32 pseudogene 17', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148580401, end_position_on_the_genomic_accession=148588944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373794, CurrentID=0, Status='live', Symbol='LOC105373794', Aliases='', description='uncharacterized LOC105373794', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190129378, end_position_on_the_genomic_accession=190141478, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480065, CurrentID=0, Status='live', Symbol='RNU6-1134P', Aliases='', description='RNA, U6 small nuclear 1134, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270946, CurrentID=0, Status='live', Symbol='RPL19P6', Aliases='RPL19_2_411', description='ribosomal protein L19 pseudogene 6', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142043544, end_position_on_the_genomic_accession=142044225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480069, CurrentID=0, Status='live', Symbol='RNU6-1143P', Aliases='', description='RNA, U6 small nuclear 1143, pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  136991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136991, CurrentID=0, Status='live', Symbol='ASZ1', Aliases='ALP1, ANKL1, C7orf7, CT1.19, GASZ, Orf3', description='ankyrin repeat, SAM and basic leucine zipper domain containing 1', other_designations='ankyrin-like 1|ankyrin-like protein 1|germ cell-specific ankyrin, SAM and basic leucine zipper domain-containing protein', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117363222, end_position_on_the_genomic_accession=117427523, orientation='minus', exon_count=13, OMIM=605797, no_hdr0=''),
  100616469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616469, CurrentID=0, Status='live', Symbol='MIR4429', Aliases='mir-4429', description='microRNA 4429', other_designations='hsa-mir-4429', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11540605, end_position_on_the_genomic_accession=11540677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873976, CurrentID=0, Status='live', Symbol='CNTN4-AS2', Aliases='', description='CNTN4 antisense RNA 2', other_designations='CNTN4 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=2110409, end_position_on_the_genomic_accession=2144241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420066, CurrentID=0, Status='live', Symbol='LOC100420066', Aliases='', description='transmembrane protein 111 pseudogene', other_designations='ER membrane protein complex subunit 3 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69443227, end_position_on_the_genomic_accession=69444292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  166614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166614, CurrentID=0, Status='live', Symbol='DCLK2', Aliases='CL2, CLICK-II, CLICK2, CLIK2, DCAMKL2, DCDC3, DCDC3B, DCK2', description='doublecortin like kinase 2', other_designations='CaMK-like CREB regulatory kinase 2|doublecortin and CaM kinase-like 2|doublecortin domain-containing protein 3B|doublecortin-like and CAM kinase-like 2|doublecortin-like kinase 2', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150078274, end_position_on_the_genomic_accession=150257457, orientation='plus', exon_count=18, OMIM=613166, no_hdr0=''),
  106480079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480079, CurrentID=0, Status='live', Symbol='RNU6-1165P', Aliases='', description='RNA, U6 small nuclear 1165, pseudogene', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375288, CurrentID=0, Status='live', Symbol='LOC105375288', Aliases='', description='uncharacterized LOC105375288', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56119882, end_position_on_the_genomic_accession=56126445, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873977, CurrentID=0, Status='live', Symbol='RBMS3-AS1', Aliases='', description='RBMS3 antisense RNA 1', other_designations='RBMS3 antisense RNA 1 (non-protein coding)', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29926811, end_position_on_the_genomic_accession=29934156, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  166647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166647, CurrentID=0, Status='live', Symbol='ADGRA3', Aliases='GPR125, PGR21, TEM5L', description='adhesion G protein-coupled receptor A3', other_designations='G protein-coupled receptor 125|G-protein coupled receptor 125|probable G-protein coupled receptor 125', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22387374, end_position_on_the_genomic_accession=22516058, orientation='minus', exon_count=23, OMIM=612303, no_hdr0=''),
  401770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401770, CurrentID=0, Status='live', Symbol='LOC401770', Aliases='', description='DET1 and DDB1 associated 1 pseudogene', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44886839, end_position_on_the_genomic_accession=44887803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  166655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166655, CurrentID=0, Status='live', Symbol='TRIM60', Aliases='RNF129, RNF33', description='tripartite motif containing 60', other_designations='ring finger protein 129|ring finger protein 33|tripartite motif-containing 60', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165025762, end_position_on_the_genomic_accession=165041818, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873979, CurrentID=0, Status='live', Symbol='RBMS3-AS3', Aliases='', description='RBMS3 antisense RNA 3', other_designations='RBMS3 antisense RNA 3 (non-protein coding)', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29264194, end_position_on_the_genomic_accession=29280939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  137012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137012, CurrentID=0, Status='live', Symbol='PPM1AP1', Aliases='', description='protein phosphatase, Mg2+/Mn2+ dependent 1A pseudogene 1', other_designations='protein phosphatase 1A (formerly 2C), magnesium-dependent, alpha isoform pseudogene 1|protein phosphatase 1A pseudogene|protein phosphatase, Mg2+/Mn2+ dependent 1A, pseudogene 1|protein phosphatase, Mg2+/Mn2+ dependent, 1A pseudogene', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=15805925, end_position_on_the_genomic_accession=15807303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873980, CurrentID=0, Status='live', Symbol='RERG-AS1', Aliases='', description='RERG antisense RNA 1', other_designations='RERG antisense RNA 1 (non-protein coding)', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15151923, end_position_on_the_genomic_accession=15155283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646865, CurrentID=0, Status='live', Symbol='LOC646865', Aliases='', description='hCG1814936', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=157775096, end_position_on_the_genomic_accession=157783056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  166752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166752, CurrentID=0, Status='live', Symbol='FREM3', Aliases='', description='FRAS1 related extracellular matrix 3', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143577407, end_position_on_the_genomic_accession=143700675, orientation='minus', exon_count=8, OMIM=608946, no_hdr0=''),
  100873981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873981, CurrentID=0, Status='live', Symbol='ANO1-AS1', Aliases='', description='ANO1 antisense RNA 1', other_designations='ANO1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372387, CurrentID=0, Status='live', Symbol='LOC105372387', Aliases='', description='uncharacterized LOC105372387', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  166785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166785, CurrentID=0, Status='live', Symbol='MMAA', Aliases='cblA', description='methylmalonic aciduria (cobalamin deficiency) cblA type', other_designations='mutant adenosylcobalamin', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145619388, end_position_on_the_genomic_accession=145660035, orientation='plus', exon_count=10, OMIM=607481, no_hdr0=''),
  166793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166793, CurrentID=0, Status='live', Symbol='ZBTB49', Aliases='ZNF509', description='zinc finger and BTB domain containing 49', other_designations='zinc finger protein 509', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4290197, end_position_on_the_genomic_accession=4321786, orientation='plus', exon_count=8, OMIM=616238, no_hdr0=''),
  105373803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373803, CurrentID=0, Status='live', Symbol='LOC105373803', Aliases='', description='uncharacterized LOC105373803', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191320640, end_position_on_the_genomic_accession=191322584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480110, CurrentID=0, Status='live', Symbol='RNU6-1246P', Aliases='', description='RNA, U6 small nuclear 1246, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480111, CurrentID=0, Status='live', Symbol='RNU6-1248P', Aliases='', description='RNA, U6 small nuclear 1248, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  166815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166815, CurrentID=0, Status='live', Symbol='TIGD2', Aliases='HEL106', description='tigger transposable element derived 2', other_designations='epididymis luminal protein 106', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89112817, end_position_on_the_genomic_accession=89114901, orientation='plus', exon_count=1, OMIM=612973, no_hdr0=''),
  106480112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480112, CurrentID=0, Status='live', Symbol='RNU6-1251P', Aliases='', description='RNA, U6 small nuclear 1251, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480113, CurrentID=0, Status='live', Symbol='RNU6-1252P', Aliases='', description='RNA, U6 small nuclear 1252, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  166824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166824, CurrentID=0, Status='live', Symbol='RASSF6', Aliases='', description='Ras association (RalGDS/AF-6) domain family member 6', other_designations='Ras association (RalGDS/AF-6) domain family 6|putative RAS binding protein', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73571550, end_position_on_the_genomic_accession=73620631, orientation='minus', exon_count=16, OMIM=612620, no_hdr0=''),
  106480114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480114, CurrentID=0, Status='live', Symbol='RNU6-1256P', Aliases='', description='RNA, U6 small nuclear 1256, pseudogene', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373804, CurrentID=0, Status='live', Symbol='LOC105373804', Aliases='', description='uncharacterized LOC105373804', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191177889, end_position_on_the_genomic_accession=191207578, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480115, CurrentID=0, Status='live', Symbol='RNU6-1258P', Aliases='', description='RNA, U6 small nuclear 1258, pseudogene', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480116, CurrentID=0, Status='live', Symbol='RNU6-1261P', Aliases='', description='RNA, U6 small nuclear 1261, pseudogene', other_designations='', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270948, CurrentID=0, Status='live', Symbol='RPL19P10', Aliases='RPL19_3_638', description='ribosomal protein L19 pseudogene 10', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171344967, end_position_on_the_genomic_accession=171345662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  166863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166863, CurrentID=0, Status='live', Symbol='RBM46', Aliases='CT68', description='RNA binding motif protein 46', other_designations='RNA-binding motif protein 46|cancer/testis antigen 68', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154781252, end_position_on_the_genomic_accession=154828813, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105374785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374785, CurrentID=0, Status='live', Symbol='LOC105374785', Aliases='', description='uncharacterized LOC105374785', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66965145, end_position_on_the_genomic_accession=67015104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874121, CurrentID=0, Status='live', Symbol='FRMPD4-AS1', Aliases='', description='FRMPD4 antisense RNA 1', other_designations='FRMPD4 antisense RNA 1 (non-protein coding)', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100174857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100174857, CurrentID=0, Status='live', Symbol='GSTA9P', Aliases='GSTA4P, GSTAP4', description='glutathione S-transferase alpha 9, pseudogene', other_designations='glutathione S-transferase alpha 4 pseudogene', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52939907, end_position_on_the_genomic_accession=52957522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  166929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166929, CurrentID=0, Status='live', Symbol='SGMS2', Aliases='SMS2', description='sphingomyelin synthase 2', other_designations='SM synthase', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107824335, end_position_on_the_genomic_accession=107915047, orientation='plus', exon_count=12, OMIM=611574, no_hdr0=''),
  166968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166968, CurrentID=0, Status='live', Symbol='MIER3', Aliases='', description='mesoderm induction early response 1, family member 3', other_designations='mi-er3', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56919602, end_position_on_the_genomic_accession=56952130, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  106480138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480138, CurrentID=0, Status='live', Symbol='RNU6-1311P', Aliases='', description='RNA, U6 small nuclear 1311, pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  166979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166979, CurrentID=0, Status='live', Symbol='CDC20B', Aliases='G6VTS76519', description='cell division cycle 20B', other_designations='CDC20 cell division cycle 20 homolog B|CDC20-like protein|cell division cycle 20 homolog B', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55112971, end_position_on_the_genomic_accession=55173177, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105373809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373809, CurrentID=0, Status='live', Symbol='LOC105373809', Aliases='', description='uncharacterized LOC105373809', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191646544, end_position_on_the_genomic_accession=191675766, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480140, CurrentID=0, Status='live', Symbol='RNU6-1316P', Aliases='', description='RNA, U6 small nuclear 1316, pseudogene', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289143, CurrentID=0, Status='live', Symbol='LOC100289143', Aliases='', description='NOP10 ribonucleoprotein homolog (yeast) pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75964752, end_position_on_the_genomic_accession=75965249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480141, CurrentID=0, Status='live', Symbol='RNU6-1318P', Aliases='', description='RNA, U6 small nuclear 1318, pseudogene', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  166994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=166994, CurrentID=0, Status='live', Symbol='LOC166994', Aliases='', description='integral membrane glycoprotein-like', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26897223, end_position_on_the_genomic_accession=26900865, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873989, CurrentID=0, Status='live', Symbol='GRM5-AS1', Aliases='', description='GRM5 antisense RNA 1', other_designations='GRM5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88504576, end_position_on_the_genomic_accession=88524054, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480142, CurrentID=0, Status='live', Symbol='RNU6-1319P', Aliases='', description='RNA, U6 small nuclear 1319, pseudogene', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480143, CurrentID=0, Status='live', Symbol='RNU6-1322P', Aliases='', description='RNA, U6 small nuclear 1322, pseudogene', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23019, CurrentID=0, Status='live', Symbol='CNOT1', Aliases='AD-005, CDC39, NOT1, NOT1H', description='CCR4-NOT transcription complex subunit 1', other_designations='CCR4-NOT transcription complex, subunit 1|CCR4-associated factor 1|NOT1 (negative regulator of transcription 1, yeast) homolog|adrenal gland protein AD-005|negative regulator of transcription subunit 1 homolog', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58519946, end_position_on_the_genomic_accession=58629886, orientation='minus', exon_count=50, OMIM=604917, no_hdr0=''),
  106480144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480144, CurrentID=0, Status='live', Symbol='RNU6-1324P', Aliases='', description='RNA, U6 small nuclear 1324, pseudogene', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100174949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100174949, CurrentID=0, Status='live', Symbol='TRIM59-IFT80', Aliases='IFT80-L', description='TRIM59-IFT80 protein', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100174950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100174950, CurrentID=0, Status='live', Symbol='LOC100174950', Aliases='', description='speckle type POZ protein like pseudogene', other_designations='speckle-type POZ protein-like pseudogene', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68882301, end_position_on_the_genomic_accession=68884447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480145, CurrentID=0, Status='live', Symbol='RNU6-1325P', Aliases='', description='RNA, U6 small nuclear 1325, pseudogene', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271686, CurrentID=0, Status='live', Symbol='DELXP11.3', Aliases='CXDELp11.3', description='Chromosome Xp11.3 deletion syndrome', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300578, no_hdr0=''),
  106480146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480146, CurrentID=0, Status='live', Symbol='RNU6-1328P', Aliases='', description='RNA, U6 small nuclear 1328, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873990, CurrentID=0, Status='live', Symbol='HCFC1-AS1', Aliases='', description='HCFC1 antisense RNA 1', other_designations='HCFC1 antisense RNA 1 (non-protein coding)', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153968765, end_position_on_the_genomic_accession=153970091, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442899, CurrentID=0, Status='live', Symbol='MIR325', Aliases='MIRN325, hsa-mir-325', description='microRNA 325', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77005404, end_position_on_the_genomic_accession=77005501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480148, CurrentID=0, Status='live', Symbol='RNU6-1334P', Aliases='', description='RNA, U6 small nuclear 1334, pseudogene', other_designations='', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480149, CurrentID=0, Status='live', Symbol='RNU6-1335P', Aliases='', description='RNA, U6 small nuclear 1335, pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480150, CurrentID=0, Status='live', Symbol='RNU6-1338P', Aliases='', description='RNA, U6 small nuclear 1338, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100887745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887745, CurrentID=0, Status='live', Symbol='MTND1P13', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 13', other_designations='MT-ND1 pseudogene 13', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271687, CurrentID=0, Status='live', Symbol='LNCR3', Aliases='', description='Lung cancer susceptibility 3', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612571, no_hdr0=''),
  100188695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188695, CurrentID=0, Status='live', Symbol='DA4', Aliases='', description='arthrogryposis, distal, type 4', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609128, no_hdr0=''),
  106480152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480152, CurrentID=0, Status='live', Symbol='RNU1-6P', Aliases='RNU1-6', description='RNA, U1 small nuclear 6, pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480153, CurrentID=0, Status='live', Symbol='RPS4XP23', Aliases='', description='ribosomal protein S4X pseudogene 23', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100887748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887748, CurrentID=0, Status='live', Symbol='UNC93B8', Aliases='', description='unc-93 homolog B8 pseudogene (C. elegans)', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8959628, end_position_on_the_genomic_accession=8963801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480154, CurrentID=0, Status='live', Symbol='RNU1-29P', Aliases='', description='RNA, U1 small nuclear 29, pseudogene', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100887749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887749, CurrentID=0, Status='live', Symbol='MTND1P23', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 23', other_designations='MT-ND1 pseudogene 23', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=629062, end_position_on_the_genomic_accession=629440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480155, CurrentID=0, Status='live', Symbol='RNU1-32P', Aliases='', description='RNA, U1 small nuclear 32, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271688, CurrentID=0, Status='live', Symbol='ANIB9', Aliases='', description='Aneurysm, intracranial berry, 9', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612586, no_hdr0=''),
  106480156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480156, CurrentID=0, Status='live', Symbol='RNU1-33P', Aliases='', description='RNA, U1 small nuclear 33, pseudogene', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873992, CurrentID=0, Status='live', Symbol='ITGB5-AS1', Aliases='', description='ITGB5 antisense RNA 1', other_designations='ITGB5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480157, CurrentID=0, Status='live', Symbol='RNU1-35P', Aliases='', description='RNA, U1 small nuclear 35, pseudogene', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137075, CurrentID=0, Status='live', Symbol='CLDN23', Aliases='CLDNL, hCG1646163', description='claudin 23', other_designations='2310014B08Rik', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8702156, end_position_on_the_genomic_accession=8704107, orientation='plus', exon_count=1, OMIM=609203, no_hdr0=''),
  106480158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480158, CurrentID=0, Status='live', Symbol='RNU1-40P', Aliases='', description='RNA, U1 small nuclear 40, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480159, CurrentID=0, Status='live', Symbol='RNU1-43P', Aliases='', description='RNA, U1 small nuclear 43, pseudogene', other_designations='', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373813, CurrentID=0, Status='live', Symbol='LOC105373813', Aliases='', description='uncharacterized LOC105373813', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191922931, end_position_on_the_genomic_accession=192046161, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106480160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480160, CurrentID=0, Status='live', Symbol='RNU1-44P', Aliases='', description='RNA, U1 small nuclear 44, pseudogene', other_designations='', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271689, CurrentID=0, Status='live', Symbol='ANIB10', Aliases='', description='Aneurysm, intracranial berry, 10', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612587, no_hdr0=''),
  106480161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480161, CurrentID=0, Status='live', Symbol='RNU1-47P', Aliases='', description='RNA, U1 small nuclear 47, pseudogene', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480162, CurrentID=0, Status='live', Symbol='RNU1-49P', Aliases='', description='RNA, U1 small nuclear 49, pseudogene', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480163, CurrentID=0, Status='live', Symbol='RNU1-51P', Aliases='', description='RNA, U1 small nuclear 51, pseudogene', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167127, CurrentID=0, Status='live', Symbol='UGT3A2', Aliases='', description='UDP glycosyltransferase 3 family, polypeptide A2', other_designations='UDPGT 3A2', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36035017, end_position_on_the_genomic_accession=36066921, orientation='minus', exon_count=8, OMIM=616384, no_hdr0=''),
  106480164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480164, CurrentID=0, Status='live', Symbol='RNU1-52P', Aliases='', description='RNA, U1 small nuclear 52, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288910, CurrentID=0, Status='live', Symbol='LOC100288910', Aliases='', description='uncharacterized LOC100288910', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373814, CurrentID=0, Status='live', Symbol='LOC105373814', Aliases='', description='uncharacterized LOC105373814', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=192323298, end_position_on_the_genomic_accession=192372952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480165, CurrentID=0, Status='live', Symbol='RNU1-56P', Aliases='', description='RNA, U1 small nuclear 56, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271690, CurrentID=0, Status='live', Symbol='CRCS8', Aliases='', description='Colorectal cancer, susceptibility to, 8', other_designations='', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612589, no_hdr0=''),
  106480166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480166, CurrentID=0, Status='live', Symbol='RNU1-57P', Aliases='', description='RNA, U1 small nuclear 57, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  647150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647150, CurrentID=0, Status='live', Symbol='LOC647150', Aliases='', description='ATP binding cassette subfamily E member 1 pseudogene', other_designations='ATP-binding cassette, sub-family E (OABP), member 1 pseudogene', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192246709, end_position_on_the_genomic_accession=192247495, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873994, CurrentID=0, Status='live', Symbol='ITPKB-AS1', Aliases='', description='ITPKB antisense RNA 1', other_designations='ITPKB antisense RNA 1 (non-protein coding)', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480167, CurrentID=0, Status='live', Symbol='RNU1-59P', Aliases='', description='RNA, U1 small nuclear 59, pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167153, CurrentID=0, Status='live', Symbol='PAPD4', Aliases='GLD2, TUT2', description='PAP associated domain containing 4', other_designations='PAP-associated domain-containing protein 4|TUTase 2|hGLD-2|terminal uridylyltransferase 2', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79612416, end_position_on_the_genomic_accession=79686648, orientation='plus', exon_count=20, OMIM=614121, no_hdr0=''),
  100271691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271691, CurrentID=0, Status='live', Symbol='CRCS10', Aliases='', description='Colorectal cancer, susceptibility to, 10', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612591, no_hdr0=''),
  100873995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873995, CurrentID=0, Status='live', Symbol='KCND3-AS1', Aliases='', description='KCND3 antisense RNA 1', other_designations='KCND3 antisense RNA 1 (non-protein coding)', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111909657, end_position_on_the_genomic_accession=111910930, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271692, CurrentID=0, Status='live', Symbol='CRCS11', Aliases='', description='Colorectal cancer, susceptibility to, 11', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612592, no_hdr0=''),
  105369452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369452, CurrentID=0, Status='live', Symbol='LOC105369452', Aliases='', description='uncharacterized LOC105369452', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97668263, end_position_on_the_genomic_accession=97676750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  167227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167227, CurrentID=0, Status='live', Symbol='DCP2', Aliases='NUDT20', description='decapping mRNA 2', other_designations='DCP2 decapping enzyme homolog|hDpc|mRNA-decapping enzyme 2|nudix (nucleoside diphosphate linked moiety X)-type motif 20', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112976702, end_position_on_the_genomic_accession=113022195, orientation='plus', exon_count=11, OMIM=609844, no_hdr0=''),
  100271693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271693, CurrentID=0, Status='live', Symbol='LNCR4', Aliases='', description='Lung cancer susceptibility 4', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612593, no_hdr0=''),
  664611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664611, CurrentID=0, Status='live', Symbol='H3F3AP2', Aliases='H3F3CP', description='H3 histone, family 3A pseudogene 2', other_designations='H3 histone, family 3A, pseudogene 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23294546, end_position_on_the_genomic_accession=23295800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873997, CurrentID=0, Status='live', Symbol='KCNQ5-AS1', Aliases='', description='KCNQ5 antisense RNA 1', other_designations='KCNQ5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73134803, end_position_on_the_genomic_accession=73143514, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100293516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293516, CurrentID=0, Status='live', Symbol='ZNF587B', Aliases='', description='zinc finger protein 587B', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57830294, end_position_on_the_genomic_accession=57846238, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371296, CurrentID=0, Status='live', Symbol='LOC105371296', Aliases='', description='uncharacterized LOC105371296', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58847743, end_position_on_the_genomic_accession=58880381, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873998, CurrentID=0, Status='live', Symbol='MEIS1-AS2', Aliases='MEIS1-AS1', description='MEIS1 antisense RNA 2', other_designations='MEIS1 antisense RNA 1 (non-protein coding)|MEIS1 antisense RNA 2 (non-protein coding)', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66439087, end_position_on_the_genomic_accession=66440330, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  132141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132141, CurrentID=0, Status='live', Symbol='IQCF1', Aliases='', description='IQ motif containing F1', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51894876, end_position_on_the_genomic_accession=51903370, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480189, CurrentID=0, Status='live', Symbol='RNU1-117P', Aliases='', description='RNA, U1 small nuclear 117, pseudogene', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137107, CurrentID=0, Status='live', Symbol='RPL10AP3', Aliases='HsT23218, RPL10A_5_891', description='ribosomal protein L10a pseudogene 3', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34322960, end_position_on_the_genomic_accession=34323675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100568447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100568447, CurrentID=0, Status='live', Symbol='IGHV3OR16-14', Aliases='IGHV3OR1614', description='immunoglobulin heavy variable 3/OR16-14 (pseudogene)', other_designations='IGHV3/OR16-14|immunoglobulin heavy variable 3/OR16-14 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480192, CurrentID=0, Status='live', Symbol='RNU1-130P', Aliases='', description='RNA, U1 small nuclear 130, pseudogene', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420089, CurrentID=0, Status='live', Symbol='LOC100420089', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52612657, end_position_on_the_genomic_accession=52613526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642791, CurrentID=0, Status='live', Symbol='LOC642791', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92914603, end_position_on_the_genomic_accession=92916603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616493, CurrentID=0, Status='live', Symbol='MIR4441', Aliases='', description='microRNA 4441', other_designations='hsa-mir-4441', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239085827, end_position_on_the_genomic_accession=239085926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271696, CurrentID=0, Status='live', Symbol='MS4', Aliases='', description='Multiple sclerosis, susceptiblity to, 4', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612596, no_hdr0=''),
  664614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664614, CurrentID=0, Status='live', Symbol='MIR545', Aliases='MIRN545, hsa-mir-545, mir-545', description='microRNA 545', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74287104, end_position_on_the_genomic_accession=74287209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480199, CurrentID=0, Status='live', Symbol='RPL3P13', Aliases='', description='ribosomal protein L3 pseudogene 13', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271697, CurrentID=0, Status='live', Symbol='IDDM23', Aliases='', description='Diabetes mellitus, insulin-dependent, 23', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612622, no_hdr0=''),
  167359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167359, CurrentID=0, Status='live', Symbol='NIM1K', Aliases='NIM1', description='NIM1 serine/threonine protein kinase', other_designations='NIM1 serine/threonine-protein kinase', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43192054, end_position_on_the_genomic_accession=43280850, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480203, CurrentID=0, Status='live', Symbol='RNU2-13P', Aliases='', description='RNA, U2 small nuclear 13, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188748, CurrentID=0, Status='live', Symbol='STHAG5', Aliases='', description='tooth agenesis, selective, 5', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610926, no_hdr0=''),
  106480205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480205, CurrentID=0, Status='live', Symbol='RNU2-18P', Aliases='', description='RNA, U2 small nuclear 18, pseudogene', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480206, CurrentID=0, Status='live', Symbol='RNU2-21P', Aliases='', description='RNA, U2 small nuclear 21, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874002, CurrentID=0, Status='live', Symbol='OXCT1-AS1', Aliases='', description='OXCT1 antisense RNA 1', other_designations='OXCT1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41870030, end_position_on_the_genomic_accession=41872241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480207, CurrentID=0, Status='live', Symbol='RNU2-24P', Aliases='', description='RNA, U2 small nuclear 24, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369140, CurrentID=0, Status='live', Symbol='LOC105369140', Aliases='', description='putative neuroblastoma breakpoint family member 6-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143744880, end_position_on_the_genomic_accession=143785882, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106480208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480208, CurrentID=0, Status='live', Symbol='RNU2-25P', Aliases='', description='RNA, U2 small nuclear 25, pseudogene', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480209, CurrentID=0, Status='live', Symbol='RNU2-29P', Aliases='', description='RNA, U2 small nuclear 29, pseudogene', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188754, CurrentID=0, Status='live', Symbol='AD14', Aliases='', description='Alzheimer disease 14', other_designations='', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611154, no_hdr0=''),
  167410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167410, CurrentID=0, Status='live', Symbol='LIX1', Aliases='C5orf11, Lft', description='limb and CNS expressed 1', other_designations='Lix1 homolog|Lowfat homolog|limb expression 1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97091870, end_position_on_the_genomic_accession=97142816, orientation='minus', exon_count=6, OMIM=610466, no_hdr0=''),
  106480211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480211, CurrentID=0, Status='live', Symbol='RNU2-33P', Aliases='', description='RNA, U2 small nuclear 33, pseudogene', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480212, CurrentID=0, Status='live', Symbol='RNU2-35P', Aliases='', description='RNA, U2 small nuclear 35, pseudogene', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132487, CurrentID=0, Status='live', Symbol='PTCD2P2', Aliases='', description='pentatricopeptide repeat domain 2 pseudogene 2', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88876387, end_position_on_the_genomic_accession=88877542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  574037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574037, CurrentID=0, Status='live', Symbol='C1orf134', Aliases='', description='chromosome 1 open reading frame 134', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060052, CurrentID=0, Status='live', Symbol='LOC101060052', Aliases='', description='zinc transporter ZIP1-like', other_designations='solute carrier family 39 (zinc transporter), member 1 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30439843, end_position_on_the_genomic_accession=30440546, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480214, CurrentID=0, Status='live', Symbol='RNU2-39P', Aliases='', description='RNA, U2 small nuclear 39, pseudogene', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271700, CurrentID=0, Status='live', Symbol='NONOP2', Aliases='', description='non-POU domain containing, octamer-binding pseudogene 2', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60935930, end_position_on_the_genomic_accession=60938049, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480216, CurrentID=0, Status='live', Symbol='RNU2-43P', Aliases='', description='RNA, U2 small nuclear 43, pseudogene', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480217, CurrentID=0, Status='live', Symbol='RNU2-45P', Aliases='', description='RNA, U2 small nuclear 45, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480218, CurrentID=0, Status='live', Symbol='RNU2-46P', Aliases='', description='RNA, U2 small nuclear 46, pseudogene', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480219, CurrentID=0, Status='live', Symbol='RNU2-49P', Aliases='', description='RNA, U2 small nuclear 49, pseudogene', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167465, CurrentID=0, Status='live', Symbol='ZNF366', Aliases='DCSCRIPT', description='zinc finger protein 366', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72439899, end_position_on_the_genomic_accession=72507422, orientation='minus', exon_count=5, OMIM=610159, no_hdr0=''),
  100188765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188765, CurrentID=0, Status='live', Symbol='CFTDX', Aliases='', description='Myopathy, congenital, with fiber-type disproportion, X-linked', other_designations='', map_location='Xq13.1-q22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300580, no_hdr0=''),
  100188767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188767, CurrentID=0, Status='live', Symbol='AGSPX', Aliases='', description='Angio serpiginosum', other_designations='', map_location='Xp11.3-q12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300652, no_hdr0=''),
  100188768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188768, CurrentID=0, Status='live', Symbol='SCAX5', Aliases='', description='Spinocerebellar ataxia, X-linked 5', other_designations='', map_location='Xq25-q27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300703, no_hdr0=''),
  100188769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188769, CurrentID=0, Status='live', Symbol='HPCX2', Aliases='', description='Prostate cancer, hereditary, X-linked 2', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300704, no_hdr0=''),
  100616499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616499, CurrentID=0, Status='live', Symbol='MIR4435-1', Aliases='mir-4435-1', description='microRNA 4435-1', other_designations='hsa-mir-4435-1', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87629755, end_position_on_the_genomic_accession=87629834, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271702, CurrentID=0, Status='live', Symbol='LINC00940', Aliases='', description='long intergenic non-protein coding RNA 940', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1929202, end_position_on_the_genomic_accession=1936576, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480227, CurrentID=0, Status='live', Symbol='RNU2-70P', Aliases='', description='RNA, U2 small nuclear 70, pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369462, CurrentID=0, Status='live', Symbol='LOC105369462', Aliases='', description='uncharacterized LOC105369462', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102606158, end_position_on_the_genomic_accession=102626731, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480228, CurrentID=0, Status='live', Symbol='RNU2-71P', Aliases='', description='RNA, U2 small nuclear 71, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724563, CurrentID=0, Status='live', Symbol='LOC102724563', Aliases='', description='transcription factor E2F6 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480229, CurrentID=0, Status='live', Symbol='MED15P8', Aliases='', description='mediator complex subunit 15 pseudogene 8', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480230, CurrentID=0, Status='live', Symbol='ENPP7P12', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 12', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271703, CurrentID=0, Status='live', Symbol='HAUS1P3', Aliases='', description='HAUS augmin like complex subunit 1 pseudogene 3', other_designations='HAUS augmin-like complex, subunit 1 pseudogene 3', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72618511, end_position_on_the_genomic_accession=72619642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480231, CurrentID=0, Status='live', Symbol='RARRES2P5', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 5', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874007, CurrentID=0, Status='live', Symbol='VIPR1-AS1', Aliases='', description='VIPR1 antisense RNA 1', other_designations='VIPR1 antisense RNA 1 (non-protein coding)', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42530896, end_position_on_the_genomic_accession=42532606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480232, CurrentID=0, Status='live', Symbol='OSBPL9P5', Aliases='', description='oxysterol binding protein like 9 pseudogene 5', other_designations='oxysterol binding protein-like 9 pseudogene 5', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167555, CurrentID=0, Status='live', Symbol='FAM151B', Aliases='UNQ9217', description='family with sequence similarity 151 member B', other_designations='AASA9217|family with sequence similarity 151, member B', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80487981, end_position_on_the_genomic_accession=80542563, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  106480238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480238, CurrentID=0, Status='live', Symbol='KRT128P', Aliases='', description='keratin 128 pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480240, CurrentID=0, Status='live', Symbol='SLC25A36P1', Aliases='', description='SLC25A36 pseudogene 1', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372496, CurrentID=0, Status='live', Symbol='LOC105372496', Aliases='', description='uncharacterized LOC105372496', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1185661, end_position_on_the_genomic_accession=1191878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100188785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188785, CurrentID=0, Status='live', Symbol='ITC1', Aliases='', description='hypertrichosis universalis congenita, Ambras type', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874009, CurrentID=0, Status='live', Symbol='BIRC6-AS1', Aliases='pal3', description='BIRC6 antisense RNA 1', other_designations='BIRC6 antisense RNA 1 (non-protein coding)', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480243, CurrentID=0, Status='live', Symbol='PRELID3BP8', Aliases='', description='PRELI domain containing 3B pseudogene 8', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480244, CurrentID=0, Status='live', Symbol='AZU1P1', Aliases='', description='azurocidin 1 pseudogene 1', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373830, CurrentID=0, Status='live', Symbol='LOC105373830', Aliases='', description='uncharacterized LOC105373830', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=198562230, end_position_on_the_genomic_accession=198572584, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480245, CurrentID=0, Status='live', Symbol='RPL17P51', Aliases='', description='ribosomal protein L17 pseudogene 51', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480246, CurrentID=0, Status='live', Symbol='SRGNP1', Aliases='', description='serglycin pseudogene 1', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874010, CurrentID=0, Status='live', Symbol='NLGN1-AS1', Aliases='', description='NLGN1 antisense RNA 1', other_designations='NLGN1 antisense RNA 1 (non-protein coding)', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173910498, end_position_on_the_genomic_accession=173920796, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480247, CurrentID=0, Status='live', Symbol='NDE1P2', Aliases='', description='nudE neurodevelopment protein 1 pseudogene 2', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480248, CurrentID=0, Status='live', Symbol='GRAMD4P4', Aliases='', description='GRAM domain containing 4 pseudogene 4', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480249, CurrentID=0, Status='live', Symbol='RAP2CP1', Aliases='', description='RAP2C pseudogene 1', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480250, CurrentID=0, Status='live', Symbol='TCEB1P35', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 35', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480251, CurrentID=0, Status='live', Symbol='OOSP1P1', Aliases='', description='oocyte secreted protein 1 pseudogene 1', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874011, CurrentID=0, Status='live', Symbol='XIRP2-AS1', Aliases='', description='XIRP2 antisense RNA 1', other_designations='XIRP2 antisense RNA 1 (non-protein coding)', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=167122809, end_position_on_the_genomic_accession=167140955, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480252, CurrentID=0, Status='live', Symbol='OOSP1P2', Aliases='', description='oocyte secreted protein 1 pseudogene 2', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188797, CurrentID=0, Status='live', Symbol='BMND6', Aliases='', description='Bone mineral density QTL 6', other_designations='', map_location='21q22.13-qter', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609876, no_hdr0=''),
  100188798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188798, CurrentID=0, Status='live', Symbol='SLEB5', Aliases='', description='Systemic lupus erythematosus, susceptibility to, 5', other_designations='', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609903, no_hdr0=''),
  167681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167681, CurrentID=0, Status='live', Symbol='PRSS35', Aliases='C6orf158, dJ223E3.1', description='protease, serine 35', other_designations='protease, serine, 35', map_location='6q14.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83512475, end_position_on_the_genomic_accession=83525704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873539, CurrentID=0, Status='live', Symbol='RNA5SP286', Aliases='RN5S286', description='RNA, 5S ribosomal pseudogene 286', other_designations='RNA, 5S ribosomal 286', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73750972, end_position_on_the_genomic_accession=73751077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874012, CurrentID=0, Status='live', Symbol='NAV2-AS5', Aliases='', description='NAV2 antisense RNA 5', other_designations='NAV2 antisense RNA 5 (non-protein coding)', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19502672, end_position_on_the_genomic_accession=19507229, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480257, CurrentID=0, Status='live', Symbol='HSPE1P18', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 18', other_designations='heat shock 10kDa protein 1 pseudogene 18', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167691, CurrentID=0, Status='live', Symbol='LCA5', Aliases='C6orf152', description='Leber congenital amaurosis 5', other_designations='leber congenital amaurosis 5 protein', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79484991, end_position_on_the_genomic_accession=79537430, orientation='minus', exon_count=9, OMIM=611408, no_hdr0=''),
  106480258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480258, CurrentID=0, Status='live', Symbol='HSPE1P21', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 21', other_designations='heat shock 10kDa protein 1 pseudogene 21', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480259, CurrentID=0, Status='live', Symbol='HSPE1P24', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 24', other_designations='heat shock 10kDa protein 1 pseudogene 24', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480260, CurrentID=0, Status='live', Symbol='HSPE1P25', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 25', other_designations='heat shock 10kDa protein 1 pseudogene 25', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  93487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93487, CurrentID=0, Status='live', Symbol='MAPK1IP1L', Aliases='C14orf32, MISS, c14_5346', description='mitogen-activated protein kinase 1 interacting protein 1-like', other_designations='MAPK-interacting and spindle-stabilizing protein|mitogen activated protein kinase 1 interacting protein 1-like|mitogen-activated protein kinase 1-interacting protein 1-like', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55051644, end_position_on_the_genomic_accession=55070194, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480261, CurrentID=0, Status='live', Symbol='HSPE1P27', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 27', other_designations='heat shock 10kDa protein 1 pseudogene 27', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874013, CurrentID=0, Status='live', Symbol='NAV2-AS3', Aliases='', description='NAV2 antisense RNA 3', other_designations='NAV2 antisense RNA 3 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480262, CurrentID=0, Status='live', Symbol='SUMO2P12', Aliases='', description='SUMO2 pseudogene 12', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480263, CurrentID=0, Status='live', Symbol='SUMO2P13', Aliases='', description='SUMO2 pseudogene 13', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480264, CurrentID=0, Status='live', Symbol='SUMO2P14', Aliases='', description='SUMO2 pseudogene 14', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480265, CurrentID=0, Status='live', Symbol='COX6CP7', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 7', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480266, CurrentID=0, Status='live', Symbol='COX6CP11', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 11', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874014, CurrentID=0, Status='live', Symbol='NAV2-AS2', Aliases='', description='NAV2 antisense RNA 2', other_designations='NAV2 antisense RNA 2 (non-protein coding)', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20043775, end_position_on_the_genomic_accession=20049303, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480267, CurrentID=0, Status='live', Symbol='COX6CP13', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 13', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480268, CurrentID=0, Status='live', Symbol='COX6CP15', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 15', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480269, CurrentID=0, Status='live', Symbol='COX6CP17', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 17', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692053, CurrentID=0, Status='live', Symbol='SNORD9', Aliases='mgU6-53B', description='small nucleolar RNA, C/D box 9', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21392151, end_position_on_the_genomic_accession=21392253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692057, CurrentID=0, Status='live', Symbol='SNORD12', Aliases='HBII-99, MIR1259, MIRN1259', description='small nucleolar RNA, C/D box 12', other_designations='hsa-mir-1259|microRNA 1259', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49280683, end_position_on_the_genomic_accession=49280772, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692058, CurrentID=0, Status='live', Symbol='SNORD11', Aliases='HBII-95', description='small nucleolar RNA, C/D box 11', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202293051, end_position_on_the_genomic_accession=202293134, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480271, CurrentID=0, Status='live', Symbol='SNRPGP20', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 20', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692063, CurrentID=0, Status='live', Symbol='SNORA32', Aliases='ACA32', description='small nucleolar RNA, H/ACA box 32', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93730979, end_position_on_the_genomic_accession=93731099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100874015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874015, CurrentID=0, Status='live', Symbol='NAV2-AS1', Aliases='', description='NAV2 antisense RNA 1', other_designations='NAV2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480272, CurrentID=0, Status='live', Symbol='IMP3P2', Aliases='', description='IMP3 pseudogene 2', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100293534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293534, CurrentID=0, Status='live', Symbol='C4B_2', Aliases='', description='complement component 4B (Chido blood group), copy 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480273, CurrentID=0, Status='live', Symbol='YWHAEP3', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 3', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692072, CurrentID=0, Status='live', Symbol='SNORD5', Aliases='mgh28S-2410', description='small nucleolar RNA, C/D box 5', other_designations='mgh28S-2410 snoRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93733228, end_position_on_the_genomic_accession=93733300, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692073, CurrentID=0, Status='live', Symbol='SNORA16A', Aliases='ACA16', description='small nucleolar RNA, H/ACA box 16A', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28580920, end_position_on_the_genomic_accession=28581053, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692075, CurrentID=0, Status='live', Symbol='SNORD6', Aliases='mgh28S-2412', description='small nucleolar RNA, C/D box 6', other_designations='mgh28S-2412 snoRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93731503, end_position_on_the_genomic_accession=93731573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692076, CurrentID=0, Status='live', Symbol='SNORD7', Aliases='Z30, mgU6-47', description='small nucleolar RNA, C/D box 7', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35573657, end_position_on_the_genomic_accession=35573753, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480274, CurrentID=0, Status='live', Symbol='IGKV1OR9-2', Aliases='', description='immunoglobulin kappa variable 1/OR9-2 (pseudogene)', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480275, CurrentID=0, Status='live', Symbol='TXNP7', Aliases='', description='thioredoxin pseudogene 7', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692084, CurrentID=0, Status='live', Symbol='SNORD13', Aliases='RNU13, U13', description='small nucleolar RNA, C/D box 13', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33513475, end_position_on_the_genomic_accession=33513578, orientation='plus', exon_count=1, OMIM=616664, no_hdr0=''),
  692085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692085, CurrentID=0, Status='live', Symbol='SNORD45C', Aliases='U45C', description='small nucleolar RNA, C/D box 45C', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75787072, end_position_on_the_genomic_accession=75787149, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692086, CurrentID=0, Status='live', Symbol='SNORD17', Aliases='HBI-43', description='small nucleolar RNA, C/D box 17', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17962710, end_position_on_the_genomic_accession=17962946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692087, CurrentID=0, Status='live', Symbol='SNORD49B', Aliases='U49B', description='small nucleolar RNA, C/D box 49B', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16439509, end_position_on_the_genomic_accession=16439556, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692088, CurrentID=0, Status='live', Symbol='SNORD50B', Aliases="U50', U50B", description='small nucleolar RNA, C/D box 50B', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85677589, end_position_on_the_genomic_accession=85677659, orientation='minus', exon_count=1, OMIM=613264, no_hdr0=''),
  692089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692089, CurrentID=0, Status='live', Symbol='SNORD19', Aliases='HBII-108', description='small nucleolar RNA, C/D box 19', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52689240, end_position_on_the_genomic_accession=52689315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692090, CurrentID=0, Status='live', Symbol='SNORD59B', Aliases='U59B', description='small nucleolar RNA, C/D box 59B', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56643680, end_position_on_the_genomic_accession=56643754, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692091, CurrentID=0, Status='live', Symbol='SNORD23', Aliases='HBII-115', description='small nucleolar RNA, C/D box 23', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47755853, end_position_on_the_genomic_accession=47755962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692092, CurrentID=0, Status='live', Symbol='SNORD32B', Aliases='U32B', description='small nucleolar RNA, C/D box 32B', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29582252, end_position_on_the_genomic_accession=29582328, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692093, CurrentID=0, Status='live', Symbol='SNORD62B', Aliases='U62, U62B', description='small nucleolar RNA, C/D box 62B', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131490486, end_position_on_the_genomic_accession=131490571, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692094, CurrentID=0, Status='live', Symbol='MSMP', Aliases='PSMP', description='microseminoprotein, prostate associated', other_designations='PC3 prostate cancer cell-secreted microprotein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35752990, end_position_on_the_genomic_accession=35754277, orientation='minus', exon_count=3, OMIM=612191, no_hdr0=''),
  106480277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480277, CurrentID=0, Status='live', Symbol='EIF1P6', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 6', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692099, CurrentID=0, Status='live', Symbol='FAM86DP', Aliases='FAM86D', description='family with sequence similarity 86 member D, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75421552, end_position_on_the_genomic_accession=75435115, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106480278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480278, CurrentID=0, Status='live', Symbol='FAM195DP', Aliases='', description='family with sequence similarity 195 member D, pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137196, CurrentID=0, Status='live', Symbol='CCDC26', Aliases='RAM', description='CCDC26 long non-coding RNA', other_designations='coiled-coil domain containing 26|retinoic acid modulator', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129351694, end_position_on_the_genomic_accession=129680239, orientation='minus', exon_count=6, OMIM=613040, no_hdr0=''),
  23046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23046, CurrentID=0, Status='live', Symbol='KIF21B', Aliases='', description='kinesin family member 21B', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200969386, end_position_on_the_genomic_accession=201023700, orientation='minus', exon_count=36, OMIM=608322, no_hdr0=''),
  692107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692107, CurrentID=0, Status='live', Symbol='SNORD66', Aliases='HBII-142', description='small nucleolar RNA, C/D box 66', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184325696, end_position_on_the_genomic_accession=184325771, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692108, CurrentID=0, Status='live', Symbol='SNORD67', Aliases='HBII-166', description='small nucleolar RNA, C/D box 67', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46762389, end_position_on_the_genomic_accession=46762499, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692109, CurrentID=0, Status='live', Symbol='SNORD69', Aliases='HBII-210', description='small nucleolar RNA, C/D box 69', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52692736, end_position_on_the_genomic_accession=52692812, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692110, CurrentID=0, Status='live', Symbol='SNORD70', Aliases='HBII-234', description='small nucleolar RNA, C/D box 70', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202276431, end_position_on_the_genomic_accession=202276518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692111, CurrentID=0, Status='live', Symbol='SNORD71', Aliases='HBII-239, MIRN768, hsa-mir-768', description='small nucleolar RNA, C/D box 71', other_designations='microRNA 768', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71758402, end_position_on_the_genomic_accession=71758487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  167826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167826, CurrentID=0, Status='live', Symbol='OLIG3', Aliases='Bhlhb7, bHLHe20', description='oligodendrocyte transcription factor 3', other_designations='class B basic helix-loop-helix protein 7|class E basic helix-loop-helix protein 20|oligo3', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137492199, end_position_on_the_genomic_accession=137494394, orientation='minus', exon_count=1, OMIM=609323, no_hdr0=''),
  106480281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480281, CurrentID=0, Status='live', Symbol='BNIP3P19', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 19', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480282, CurrentID=0, Status='live', Symbol='BNIP3P31', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 31', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  167838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=167838, CurrentID=0, Status='live', Symbol='TXLNB', Aliases='C6orf198, LST001, MDP77, dJ522B19.2', description='taxilin beta', other_designations='muscle-derived protein 77', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139159160, end_position_on_the_genomic_accession=139324011, orientation='minus', exon_count=18, OMIM=611438, no_hdr0=''),
  106480283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480283, CurrentID=0, Status='live', Symbol='BNIP3P32', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 32', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132160, CurrentID=0, Status='live', Symbol='PPM1M', Aliases='PP2C-eta, PP2CE, PP2Ceta', description='protein phosphatase, Mg2+/Mn2+ dependent 1M', other_designations='protein phosphatase 1M (PP2C domain containing)|protein phosphatase 2C eta-2|protein phosphatase 2C isoform eta|protein phosphatase, Mg2+/Mn2+ dependent, 1M', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52245725, end_position_on_the_genomic_accession=52250599, orientation='plus', exon_count=11, OMIM=608979, no_hdr0=''),
  106480284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480284, CurrentID=0, Status='live', Symbol='BNIP3P35', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 35', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373838, CurrentID=0, Status='live', Symbol='LOC105373838', Aliases='', description='uncharacterized LOC105373838', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201511729, end_position_on_the_genomic_accession=201547908, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480285, CurrentID=0, Status='live', Symbol='BNIP3P37', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 37', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422639, CurrentID=0, Status='live', Symbol='LOC100422639', Aliases='', description='MX dynamin-like GTPase 1 pseudogene', other_designations='myxovirus (influenza virus) resistance 1, interferon-inducible protein p78 (mouse) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10294459, end_position_on_the_genomic_accession=10295546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692146, CurrentID=0, Status='live', Symbol='RPS26P4', Aliases='RPS26_32_1733', description='ribosomal protein S26 pseudogene 4', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39491476, end_position_on_the_genomic_accession=39491918, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692147, CurrentID=0, Status='live', Symbol='HMGB1P8', Aliases='HMG1L8, HMGB1L8', description='high mobility group box 1 pseudogene 8', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 8|high-mobility group box 1 pseudogene 8|high-mobility group box 1-like 8', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89135067, end_position_on_the_genomic_accession=89136675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692148, CurrentID=0, Status='live', Symbol='SCARNA10', Aliases='U85', description='small Cajal body-specific RNA 10', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6510222, end_position_on_the_genomic_accession=6510551, orientation='plus', exon_count=1, OMIM=615639, no_hdr0=''),
  692149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692149, CurrentID=0, Status='live', Symbol='SCARNA14', Aliases='U100', description='small Cajal body-specific RNA 14', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66347206, end_position_on_the_genomic_accession=66347342, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  268276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=268276, CurrentID=0, Status='live', Symbol='LOC268276', Aliases='', description='musashi RNA binding protein 1 pseudogene', other_designations='musashi 1 pseudogene|musashi RNA-binding protein 1 pseudogene|musashi homolog 1 pseudogene', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73691329, end_position_on_the_genomic_accession=73692529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692157, CurrentID=0, Status='live', Symbol='SNORA16B', Aliases='U98b', description='small nucleolar RNA, H/ACA box 16B', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212352818, end_position_on_the_genomic_accession=212352950, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692158, CurrentID=0, Status='live', Symbol='SNORA57', Aliases='U99', description='small nucleolar RNA, H/ACA box 57', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62665422, end_position_on_the_genomic_accession=62665570, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692159, CurrentID=0, Status='live', Symbol='GRASPOS', Aliases='', description='GRP1-associated scaffold protein opposite strand', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480288, CurrentID=0, Status='live', Symbol='LSM1P2', Aliases='', description='LSM1 homolog, mRNA degradation associated pseudogene 2', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442042, CurrentID=0, Status='live', Symbol='LOC442042', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide D pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112995517, end_position_on_the_genomic_accession=112995936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480290, CurrentID=0, Status='live', Symbol='SNRPCP5', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 5', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271715, CurrentID=0, Status='live', Symbol='ARHGEF33', Aliases='', description='Rho guanine nucleotide exchange factor 33', other_designations='DH and coiled-coil domain-containing protein ENSP00000381780|Rho guanine nucleotide exchange factor (GEF) 33', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38919363, end_position_on_the_genomic_accession=38975449, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  137209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137209, CurrentID=0, Status='live', Symbol='ZNF572', Aliases='', description='zinc finger protein 572', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124848552, end_position_on_the_genomic_accession=124979511, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  106480292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480292, CurrentID=0, Status='live', Symbol='SNRPCP20', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 20', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369475, CurrentID=0, Status='live', Symbol='LOC105369475', Aliases='', description='uncharacterized LOC105369475', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106856642, end_position_on_the_genomic_accession=106857718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480293, CurrentID=0, Status='live', Symbol='UBTFL10', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 10 (pseudogene)', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692195, CurrentID=0, Status='live', Symbol='SNORD75', Aliases='U75', description='small nucleolar RNA, C/D box 75', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173866879, end_position_on_the_genomic_accession=173866938, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23049, CurrentID=0, Status='live', Symbol='SMG1', Aliases='61E3.4, ATX, LIP', description='SMG1 phosphatidylinositol 3-kinase-related kinase', other_designations='PI-3-kinase-related kinase SMG-1|lambda-interacting protein|lambda/iota protein kinase C-interacting protein|smg-1 homolog, phosphatidylinositol 3-kinase-related kinase', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18804853, end_position_on_the_genomic_accession=18926409, orientation='minus', exon_count=63, OMIM=607032, no_hdr0=''),
  692197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692197, CurrentID=0, Status='live', Symbol='SNORD77', Aliases='U77', description='small nucleolar RNA, C/D box 77', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173866311, end_position_on_the_genomic_accession=173866371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692198, CurrentID=0, Status='live', Symbol='SNORD78', Aliases='U78', description='small nucleolar RNA, C/D box 78', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173865633, end_position_on_the_genomic_accession=173865686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692199, CurrentID=0, Status='live', Symbol='SNORD84', Aliases='U84', description='small nucleolar RNA, C/D box 84', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31541101, end_position_on_the_genomic_accession=31541178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692200, CurrentID=0, Status='live', Symbol='SNORD103C', Aliases='HBII-251, SNORD85', description='small nucleolar RNA, C/D box 103C', other_designations='small nucleolar RNA, C/D box 85', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30968163, end_position_on_the_genomic_accession=30968237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692201, CurrentID=0, Status='live', Symbol='SNORD86', Aliases='U86', description='small nucleolar RNA, C/D box 86', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2656097, end_position_on_the_genomic_accession=2656182, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692202, CurrentID=0, Status='live', Symbol='SNORD88A', Aliases='HBII-180A', description='small nucleolar RNA, C/D box 88A', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50799439, end_position_on_the_genomic_accession=50799535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692203, CurrentID=0, Status='live', Symbol='SNORD88B', Aliases='HBII-180B', description='small nucleolar RNA, C/D box 88B', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50799029, end_position_on_the_genomic_accession=50799125, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692204, CurrentID=0, Status='live', Symbol='SNORD88C', Aliases='HBII-180C', description='small nucleolar RNA, C/D box 88C', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50802325, end_position_on_the_genomic_accession=50802421, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692205, CurrentID=0, Status='live', Symbol='SNORD89', Aliases='HBII-289', description='small nucleolar RNA, C/D box 89', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101272936, end_position_on_the_genomic_accession=101273049, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692206, CurrentID=0, Status='live', Symbol='SNORD90', Aliases='HBII-295', description='small nucleolar RNA, C/D box 90', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122880213, end_position_on_the_genomic_accession=122880319, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692207, CurrentID=0, Status='live', Symbol='SNORD91A', Aliases='HBII-296a', description='small nucleolar RNA, C/D box 91A', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2330279, end_position_on_the_genomic_accession=2330370, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692208, CurrentID=0, Status='live', Symbol='SNORD91B', Aliases='HBII-296b', description='small nucleolar RNA, C/D box 91B', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2329125, end_position_on_the_genomic_accession=2329210, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692209, CurrentID=0, Status='live', Symbol='SNORD92', Aliases='HBII-316', description='small nucleolar RNA, C/D box 92', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28913662, end_position_on_the_genomic_accession=28913750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692210, CurrentID=0, Status='live', Symbol='SNORD93', Aliases='HBII-336', description='small nucleolar RNA, C/D box 93', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22856613, end_position_on_the_genomic_accession=22856686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692211, CurrentID=0, Status='live', Symbol='SNORD98', Aliases='HBII-419', description='small nucleolar RNA, C/D box 98', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68755172, end_position_on_the_genomic_accession=68755238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692212, CurrentID=0, Status='live', Symbol='SNORD99', Aliases='HBII-420', description='small nucleolar RNA, C/D box 99', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28578743, end_position_on_the_genomic_accession=28578822, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692213, CurrentID=0, Status='live', Symbol='SNORD110', Aliases='HBII-55', description='small nucleolar RNA, C/D box 110', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2654212, end_position_on_the_genomic_accession=2654286, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692214, CurrentID=0, Status='live', Symbol='SNORD111', Aliases='HBII-82', description='small nucleolar RNA, C/D box 111', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70538005, end_position_on_the_genomic_accession=70538098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692215, CurrentID=0, Status='live', Symbol='SNORD112', Aliases='14q(0)', description='small nucleolar RNA, C/D box 112', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100897920, end_position_on_the_genomic_accession=100897996, orientation='plus', exon_count=1, OMIM=613649, no_hdr0=''),
  100306937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306937, CurrentID=0, Status='live', Symbol='NBLST4', Aliases='', description='Neuroblastoma, susceptibility to, 4', other_designations='', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613015, no_hdr0=''),
  100306938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306938, CurrentID=0, Status='live', Symbol='NBLST5', Aliases='', description='Neuroblastoma, susceptibility to, 5', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613016, no_hdr0=''),
  692219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692219, CurrentID=0, Status='live', Symbol='DFNB65', Aliases='', description='deafness, autosomal recessive 65', other_designations='', map_location='20q13.2-q13.32', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610248, no_hdr0=''),
  692220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692220, CurrentID=0, Status='live', Symbol='DFNB62', Aliases='', description='deafness, autosomal recessive 62', other_designations='', map_location='12p13.2-p11.23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610143, no_hdr0=''),
  106480298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480298, CurrentID=0, Status='live', Symbol='NUTF2P3', Aliases='', description='nuclear transport factor 2 pseudogene 3', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692222, CurrentID=0, Status='live', Symbol='OPA5', Aliases='', description='optic atrophy 5 (autosomal dominant)', other_designations='', map_location='22q12.1-q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610708, no_hdr0=''),
  692223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692223, CurrentID=0, Status='live', Symbol='SNORD97', Aliases='U97', description='small nucleolar RNA, C/D box 97', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10801467, end_position_on_the_genomic_accession=10801608, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692224, CurrentID=0, Status='live', Symbol='FBXO22-AS1', Aliases='FBXO22OS, FISTC1OS, HsT18082, NCRNA00053', description='FBXO22 antisense RNA 1', other_designations='F-box only protein 22 opposite strand|F-box protein 22 opposite strand|FBXO22 antisense RNA 1 (non-protein coding)|FBXO22 opposite strand (non-protein coding)|FISTC1 opposite strand (non-protein coding)', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75932683, end_position_on_the_genomic_accession=75933130, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692225, CurrentID=0, Status='live', Symbol='SNORD94', Aliases='U94', description='small nucleolar RNA, C/D box 94', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86135870, end_position_on_the_genomic_accession=86136006, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692226, CurrentID=0, Status='live', Symbol='SNORD96B', Aliases='U96b', description='small nucleolar RNA, C/D box 96B', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110224989, end_position_on_the_genomic_accession=110225060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692227, CurrentID=0, Status='live', Symbol='SNORD104', Aliases='U104', description='small nucleolar RNA, C/D box 104', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64146078, end_position_on_the_genomic_accession=64146157, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  692229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692229, CurrentID=0, Status='live', Symbol='SNORD105', Aliases='U105', description='small nucleolar RNA, C/D box 105', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10107651, end_position_on_the_genomic_accession=10107735, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100306951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306951, CurrentID=0, Status='live', Symbol='PITPNA-AS1', Aliases='', description='PITPNA antisense RNA 1', other_designations='PITPNA antisense RNA 1 (non-protein coding)', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1516919, end_position_on_the_genomic_accession=1518096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  692232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692232, CurrentID=0, Status='live', Symbol='SNORD3@', Aliases='U3, U3A, U3B', description='small nucleolar RNA, C/D box 3 cluster', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692233, CurrentID=0, Status='live', Symbol='SNORD117', Aliases='U83', description='small nucleolar RNA, C/D box 117', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31536374, end_position_on_the_genomic_accession=31536449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692234, CurrentID=0, Status='live', Symbol='SNORD103A', Aliases='U103', description='small nucleolar RNA, C/D box 103A', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30935689, end_position_on_the_genomic_accession=30935776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692235, CurrentID=0, Status='live', Symbol='SNORD103B', Aliases='U103B', description='small nucleolar RNA, C/D box 103B', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30949118, end_position_on_the_genomic_accession=30949205, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  692236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692236, CurrentID=0, Status='live', Symbol='SNORD116@', Aliases='HBII-85, PET1, PWCR1', description='small nucleolar RNA, C/D box 116 cluster', other_designations='Prader-Willi syndrome chromosome region 1|non-coding RNA in the Prader-Willi critical region', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25051476, end_position_on_the_genomic_accession=25106603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480301, CurrentID=0, Status='live', Symbol='TOMM20P3', Aliases='', description='TOMM20 pseudogene 3', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480302, CurrentID=0, Status='live', Symbol='MZT1P1', Aliases='', description='mitotic spindle organizing protein 1 pseudogene 1', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692246, CurrentID=0, Status='live', Symbol='LOC692246', Aliases='', description='cytochrome c oxidase assembly factor 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='22q13.2-q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42853082, end_position_on_the_genomic_accession=42853940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692247, CurrentID=0, Status='live', Symbol='LOC692247', Aliases='', description='uncharacterized LOC692247', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=556958, end_position_on_the_genomic_accession=560113, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480303, CurrentID=0, Status='live', Symbol='PRR13P2', Aliases='', description='proline rich 13 pseudogene 2', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100306975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306975, CurrentID=0, Status='live', Symbol='NDUFAF4P1', Aliases='', description='NADH:ubiquinone oxidoreductase complex assembly factor 4 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 4 pseudogene 1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49156298, end_position_on_the_genomic_accession=49158625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480304, CurrentID=0, Status='live', Symbol='MREGP1', Aliases='', description='melanoregulin pseudogene 1', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100306978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306978, CurrentID=0, Status='live', Symbol='DEL19Q13.11', Aliases='C19DELq13.11', description='Chromosome 19q13.11 deletion syndrome', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613026, no_hdr0=''),
  100306979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306979, CurrentID=0, Status='live', Symbol='ATFB8', Aliases='', description='Atrial fibrillation, familial, 8', other_designations='', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613055, no_hdr0=''),
  106480305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480305, CurrentID=0, Status='live', Symbol='AP1S2P1', Aliases='', description='adaptor related protein complex 1 sigma 2 subunit pseudogene 1', other_designations='adaptor-related protein complex 1 sigma 2 subunit pseudogene 1|adaptor-related protein complex 1, sigma 2 subunit pseudogene 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271718, CurrentID=0, Status='live', Symbol='AD16', Aliases='', description='Alzheimer disease 16', other_designations='', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300756, no_hdr0=''),
  100874022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874022, CurrentID=0, Status='live', Symbol='ATG10-AS1', Aliases='', description='ATG10 antisense RNA 1', other_designations='ATG10 antisense RNA 1 (non-protein coding)', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130419, CurrentID=0, Status='live', Symbol='LOC100130419', Aliases='', description='uncharacterized LOC100130419', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=50331367, end_position_on_the_genomic_accession=50332283, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480308, CurrentID=0, Status='live', Symbol='RNY1P10', Aliases='', description='RNA, Ro-associated Y1 pseudogene 10', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100689074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689074, CurrentID=0, Status='live', Symbol='POU6F2-AS2', Aliases='', description='POU6F2 antisense RNA 2', other_designations='POU6F2 antisense RNA 2 (non-protein coding)', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38979895, end_position_on_the_genomic_accession=39013567, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480309, CurrentID=0, Status='live', Symbol='RNY1P12', Aliases='', description='RNA, Ro-associated Y1 pseudogene 12', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  168002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168002, CurrentID=0, Status='live', Symbol='DACT2', Aliases='C6orf116, DAPPER2, DPR2, bA503C24.7', description='dishevelled-binding antagonist of beta-catenin 2', other_designations='dapper antagonist of catenin 2|dapper homolog 2, antagonist of beta-catenin|dapper, antagonist of beta-catenin, homolog 2', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168292830, end_position_on_the_genomic_accession=168319754, orientation='minus', exon_count=9, OMIM=608966, no_hdr0=''),
  105373843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373843, CurrentID=0, Status='live', Symbol='LOC105373843', Aliases='', description='uncharacterized LOC105373843', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202075410, end_position_on_the_genomic_accession=202149600, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480310, CurrentID=0, Status='live', Symbol='RNY1P13', Aliases='', description='RNA, Ro-associated Y1 pseudogene 13', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271719, CurrentID=0, Status='live', Symbol='PSORS11', Aliases='', description='Psoriasis susceptibility 11', other_designations='', map_location='5q31.1-q33.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612599, no_hdr0=''),
  106480312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480312, CurrentID=0, Status='live', Symbol='RNY1P16', Aliases='', description='RNA, Ro-associated Y1 pseudogene 16', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480313, CurrentID=0, Status='live', Symbol='RNY4P36', Aliases='', description='RNA, Ro-associated Y4 pseudogene 36', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692312, CurrentID=0, Status='live', Symbol='PPAN-P2RY11', Aliases='', description='PPAN-P2RY11 readthrough', other_designations='PPAN-P2RY11 readthrough transcript|SSF1/P2Y11 protein', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10106223, end_position_on_the_genomic_accession=10115389, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  106480314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480314, CurrentID=0, Status='live', Symbol='RNY4P37', Aliases='', description='RNA, Ro-associated Y4 pseudogene 37', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480315, CurrentID=0, Status='live', Symbol='RNY3P12', Aliases='', description='RNA, Ro-associated Y3 pseudogene 12', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480316, CurrentID=0, Status='live', Symbol='RNY3P14', Aliases='', description='RNA, Ro-associated Y3 pseudogene 14', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480317, CurrentID=0, Status='live', Symbol='MLECP1', Aliases='', description='malectin pseudogene 1', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188862, CurrentID=0, Status='live', Symbol='DUP3Q29', Aliases='MICRODUP3Q29', description='chromosome 3q29 microduplication syndrome', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611936, no_hdr0=''),
  100188863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188863, CurrentID=0, Status='live', Symbol='ENDO1', Aliases='', description='Endometriosis, susceptibility to, 1', other_designations='', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=131200, no_hdr0=''),
  100132056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132056, CurrentID=0, Status='live', Symbol='LOC100132056', Aliases='', description='dihydrofolate reductase pseudogene', other_designations='', map_location='6q11.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=62460943, end_position_on_the_genomic_accession=62461503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188864, CurrentID=0, Status='live', Symbol='IH', Aliases='HHP', description='Hemihypertrophy', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=235000, no_hdr0=''),
  100101483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101483, CurrentID=0, Status='live', Symbol='LOC100101483', Aliases='', description='glycerophosphocholine phosphodiesterase 1 pseudogene', other_designations='glycerophosphocholine phosphodiesterase GDE1 homolog (S. cerevisiae) pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57146780, end_position_on_the_genomic_accession=57150547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480321, CurrentID=0, Status='live', Symbol='FRG2GP', Aliases='', description='FSHD region gene 2 family member G, pseudogene', other_designations='FSHD region gene 2 family, member G, pseudogene', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874025, CurrentID=0, Status='live', Symbol='GFOD1-AS1', Aliases='', description='GFOD1 antisense RNA 1', other_designations='GFOD1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480323, CurrentID=0, Status='live', Symbol='BOLA2P3', Aliases='', description='bolA family member 2 pseudogene 3', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188868, CurrentID=0, Status='live', Symbol='HPC15', Aliases='', description='Prostate cancer, hereditary, 15', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611959, no_hdr0=''),
  168090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168090, CurrentID=0, Status='live', Symbol='C6orf118', Aliases='bA85G2.1, dJ416F21.2', description='chromosome 6 open reading frame 118', other_designations='dJ416F21.2 (novel protein)', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165279664, end_position_on_the_genomic_accession=165309647, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  642817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642817, CurrentID=0, Status='live', Symbol='LOC642817', Aliases='', description='uncharacterized LOC642817', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=5554124, end_position_on_the_genomic_accession=5555471, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100188869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188869, CurrentID=0, Status='live', Symbol='DEL15Q13.3', Aliases='EIG7, MICRODEL15Q13.3, SCZD13', description='Chromosome 15q13.3 microdeletion syndrome', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612001, no_hdr0=''),
  100271722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271722, CurrentID=0, Status='live', Symbol='LINC00899', Aliases='', description='long intergenic non-protein coding RNA 899', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46039907, end_position_on_the_genomic_accession=46044868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480327, CurrentID=0, Status='live', Symbol='WBP1LP12', Aliases='', description='WW domain binding protein 1-like pseudogene 12', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100529229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529229, CurrentID=0, Status='live', Symbol='FAME3', Aliases='FCMTE3', description='Epilepsy, familial adult myoclonic, 3', other_designations='', map_location='5p15.31-p15.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613608, no_hdr0=''),
  100307119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307119, CurrentID=0, Status='live', Symbol='BCC2', Aliases='', description='Basal cell carcinoma, susceptibility to, 2', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613058, no_hdr0=''),
  100307120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307120, CurrentID=0, Status='live', Symbol='BCC3', Aliases='', description='Basal cell carcinoma, susceptibility to, 3', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613059, no_hdr0=''),
  100307121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307121, CurrentID=0, Status='live', Symbol='BCC4', Aliases='', description='Basal cell carcinoma, susceptibility to, 4', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613061, no_hdr0=''),
  100307122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307122, CurrentID=0, Status='live', Symbol='BCC5', Aliases='', description='Basal cell carcinoma, susceptibility to, 5', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613062, no_hdr0=''),
  100307123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307123, CurrentID=0, Status='live', Symbol='BCC6', Aliases='', description='Basal cell carcinoma, susceptibility to, 6', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613063, no_hdr0=''),
  100307126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307126, CurrentID=0, Status='live', Symbol='CHMP4BP1', Aliases='', description='charged multivesicular body protein 4B pseudogene 1', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55298602, end_position_on_the_genomic_accession=55299496, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480329, CurrentID=0, Status='live', Symbol='BSNDP1', Aliases='', description='barttin CLCNK-type chloride channel accessory beta subunit pseudogene 1', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373847, CurrentID=0, Status='live', Symbol='LOC105373847', Aliases='', description='uncharacterized LOC105373847', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=205798676, end_position_on_the_genomic_accession=205843342, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480330, CurrentID=0, Status='live', Symbol='BTG4P1', Aliases='', description='B-cell translocation gene 4 pseudogene 1', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480331, CurrentID=0, Status='live', Symbol='BUD31P1', Aliases='', description='BUD31 homolog pseudogene 1', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928389, CurrentID=0, Status='live', Symbol='LOC101928389', Aliases='', description='uncharacterized LOC101928389', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17528091, end_position_on_the_genomic_accession=17587160, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480332, CurrentID=0, Status='live', Symbol='BUD31P2', Aliases='', description='BUD31 homolog pseudogene 2', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420117, CurrentID=0, Status='live', Symbol='LOC100420117', Aliases='', description='methionyl aminopeptidase 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40728847, end_position_on_the_genomic_accession=40729897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480333, CurrentID=0, Status='live', Symbol='SULT1C2P2', Aliases='', description='sulfotransferase family, cytosolic, 1C, member 2 pseudogene 2', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642819, CurrentID=0, Status='live', Symbol='ZNF487', Aliases='KRBO1P, ZNF487', description='zinc finger protein 487', other_designations='KRAB domain only 1|zinc finger protein 487, pseudogene', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43437126, end_position_on_the_genomic_accession=43482559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927167, CurrentID=0, Status='live', Symbol='GATA2-AS1', Aliases='', description='GATA2 antisense RNA 1', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128489203, end_position_on_the_genomic_accession=128503207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480336, CurrentID=0, Status='live', Symbol='RPL23AP89', Aliases='', description='ribosomal protein L23a pseudogene 89', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480337, CurrentID=0, Status='live', Symbol='RPL23AP91', Aliases='', description='ribosomal protein L23a pseudogene 91', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369484, CurrentID=0, Status='live', Symbol='LOC105369484', Aliases='', description='uncharacterized LOC105369484', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109946550, end_position_on_the_genomic_accession=109972995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480338, CurrentID=0, Status='live', Symbol='RPL23AP95', Aliases='', description='ribosomal protein L23a pseudogene 95', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480339, CurrentID=0, Status='live', Symbol='TCF4-AS2', Aliases='', description='TCF4 antisense RNA 2', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480340, CurrentID=0, Status='live', Symbol='SEPT14P1', Aliases='', description='septin 14 pseudogene 1', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480341, CurrentID=0, Status='live', Symbol='SEPT14P8', Aliases='', description='septin 14 pseudogene 8', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480342, CurrentID=0, Status='live', Symbol='ZYXP1', Aliases='', description='zyxin pseudogene 1', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480343, CurrentID=0, Status='live', Symbol='SEPT14P12', Aliases='', description='septin 14 pseudogene 12', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289184, CurrentID=0, Status='live', Symbol='LOC100289184', Aliases='', description='ribosomal protein L22 pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136947751, end_position_on_the_genomic_accession=136948133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480347, CurrentID=0, Status='live', Symbol='GTF2IP22', Aliases='', description='general transcription factor IIi pseudogene 22', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369486, CurrentID=0, Status='live', Symbol='LOC105369486', Aliases='', description='uncharacterized LOC105369486', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110355114, end_position_on_the_genomic_accession=110406423, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421308, CurrentID=0, Status='live', Symbol='LOC100421308', Aliases='', description='phosphoribosyl pyrophosphate synthetase 2 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=18908901, end_position_on_the_genomic_accession=18909462, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480349, CurrentID=0, Status='live', Symbol='DUX4L46', Aliases='', description='double homeobox 4 like 46, pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289185, CurrentID=0, Status='live', Symbol='LOC100289185', Aliases='', description='telomeric repeat-binding factor 1 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6911032, end_position_on_the_genomic_accession=6912513, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369487, CurrentID=0, Status='live', Symbol='LOC105369487', Aliases='', description='uncharacterized LOC105369487', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373852, CurrentID=0, Status='live', Symbol='LOC105373852', Aliases='', description='uncharacterized LOC105373852', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207260432, end_position_on_the_genomic_accession=207271064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874032, CurrentID=0, Status='live', Symbol='PRRT3-AS1', Aliases='', description='PRRT3 antisense RNA 1', other_designations='PRRT3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9947404, end_position_on_the_genomic_accession=9954787, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369488, CurrentID=0, Status='live', Symbol='LOC105369488', Aliases='', description='uncharacterized LOC105369488', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110922007, end_position_on_the_genomic_accession=110923953, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645645, CurrentID=0, Status='live', Symbol='LOC645645', Aliases='', description='uncharacterized LOC645645', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=233818069, end_position_on_the_genomic_accession=233818744, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874033, CurrentID=0, Status='live', Symbol='SPIN4-AS1', Aliases='', description='SPIN4 antisense RNA 1', other_designations='SPIN4 antisense RNA 1 (non-protein coding)', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060058, CurrentID=0, Status='live', Symbol='LOC101060058', Aliases='', description='putative uncharacterized protein FLJ45355', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31122434, end_position_on_the_genomic_accession=31122998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  168330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168330, CurrentID=0, Status='live', Symbol='PRSS3P1', Aliases='TRY5', description='protease, serine 3 pseudogene 1', other_designations='protease, serine, 1 (trypsin 1)|protease, serine, 3 pseudogene 1', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142760415, end_position_on_the_genomic_accession=142763981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372497, CurrentID=0, Status='live', Symbol='LOC105372497', Aliases='', description='uncharacterized LOC105372497', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1405146, end_position_on_the_genomic_accession=1411146, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929465, CurrentID=0, Status='live', Symbol='LOC101929465', Aliases='', description='uncharacterized LOC101929465', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44292088, end_position_on_the_genomic_accession=44293360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  341674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341674, CurrentID=0, Status='live', Symbol='RPL5P31', Aliases='', description='ribosomal protein L5 pseudogene 31', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51161695, end_position_on_the_genomic_accession=51162352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  168374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168374, CurrentID=0, Status='live', Symbol='ZNF92', Aliases='HEL-203, HPF12, HTF12, TF12', description='zinc finger protein 92', other_designations='epididymis luminal protein 203|zinc finger protein HTF12', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65373799, end_position_on_the_genomic_accession=65401135, orientation='plus', exon_count=5, OMIM=603974, no_hdr0=''),
  100874035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874035, CurrentID=0, Status='live', Symbol='PLCH1-AS1', Aliases='', description='PLCH1 antisense RNA 1', other_designations='PLCH1 antisense RNA 1 (non-protein coding)', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155449185, end_position_on_the_genomic_accession=155457753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  137290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137290, CurrentID=0, Status='live', Symbol='MRPL9P1', Aliases='', description='mitochondrial ribosomal protein L9 pseudogene 1', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76602827, end_position_on_the_genomic_accession=76603873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480373, CurrentID=0, Status='live', Symbol='RN7SL258P', Aliases='', description='RNA, 7SL, cytoplasmic 258, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  168391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168391, CurrentID=0, Status='live', Symbol='GALNTL5', Aliases='GALNACT19, GALNT15, GalNAc-T5L', description='polypeptide N-acetylgalactosaminyltransferase-like 5', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 15|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 5|galNAc-T15|polypeptide GalNAc transferase 15|polypeptide GalNAc transferase-like 5|pp-GaNTase 15|protein-UDP acetylgalactosaminyltransferase 15|putative polypeptide N-acetylgalactosaminyltransferase-like protein 5', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151956379, end_position_on_the_genomic_accession=152019934, orientation='plus', exon_count=15, OMIM=615133, no_hdr0=''),
  168400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168400, CurrentID=0, Status='live', Symbol='DDX53', Aliases='CAGE, CT26', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 53', other_designations='DEAD-box protein CAGE|cancer-associated gene protein|cancer/testis antigen 26|probable ATP-dependent RNA helicase DDX53', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22999961, end_position_on_the_genomic_accession=23002089, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874036, CurrentID=0, Status='live', Symbol='PLCH1-AS2', Aliases='', description='PLCH1 antisense RNA 2', other_designations='PLCH1 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155485717, end_position_on_the_genomic_accession=155487121, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100529239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529239, CurrentID=0, Status='live', Symbol='RPS10-NUDT3', Aliases='', description='RPS10-NUDT3 readthrough', other_designations='ribosomal protein S10 homologue', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34287194, end_position_on_the_genomic_accession=34426125, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  168417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168417, CurrentID=0, Status='live', Symbol='ZNF679', Aliases='', description='zinc finger protein 679', other_designations='hypothetical protein MGC42415', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64228474, end_position_on_the_genomic_accession=64266931, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  168433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168433, CurrentID=0, Status='live', Symbol='RNF133', Aliases='', description='ring finger protein 133', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122697712, end_position_on_the_genomic_accession=122699154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  168448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168448, CurrentID=0, Status='live', Symbol='CDC14C', Aliases='CDC14B2, CDC14Bretro', description='cell division cycle 14C', other_designations='CDC14 cell division cycle 14 C pseudogene|CDC14 cell division cycle 14 homolog C', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48924561, end_position_on_the_genomic_accession=48927453, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  168451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168451, CurrentID=0, Status='live', Symbol='THAP5', Aliases='', description='THAP domain containing 5', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108562144, end_position_on_the_genomic_accession=108569768, orientation='minus', exon_count=4, OMIM=612534, no_hdr0=''),
  168455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168455, CurrentID=0, Status='live', Symbol='CCDC71L', Aliases='C7orf74', description='coiled-coil domain containing 71-like', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106656765, end_position_on_the_genomic_accession=106661188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  1245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1245, CurrentID=0, Status='live', Symbol='CMR1A', Aliases='', description='cardiomyopathy, restrictive 1A (autosomal dominant)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284939, CurrentID=0, Status='live', Symbol='LOC284939', Aliases='', description='uncharacterized LOC284939', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  168474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168474, CurrentID=0, Status='live', Symbol='SEPHS1P1', Aliases='SEPHS1P', description='selenophosphate synthetase 1 pseudogene 1', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64851320, end_position_on_the_genomic_accession=64853449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287544, CurrentID=0, Status='live', Symbol='LOC100287544', Aliases='', description='s-adenosylmethionine synthase isoform type-2-like', other_designations='methionine adenosyltransferase II, alpha pseudogene', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92858081, end_position_on_the_genomic_accession=92859272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480391, CurrentID=0, Status='live', Symbol='RNU6-373P', Aliases='', description='RNA, U6 small nuclear 373, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1246, CurrentID=0, Status='live', Symbol='CMR2A', Aliases='', description='cardiomyopathy, restrictive 2A (autosomal recessive)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874039, CurrentID=0, Status='live', Symbol='THOC7-AS1', Aliases='', description='THOC7 antisense RNA 1', other_designations='THOC7 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63860645, end_position_on_the_genomic_accession=63861819, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645321, CurrentID=0, Status='live', Symbol='LOC645321', Aliases='', description='uncharacterized LOC645321', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78976814, end_position_on_the_genomic_accession=78979526, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  168507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168507, CurrentID=0, Status='live', Symbol='PKD1L1', Aliases='PRO19563', description='polycystic kidney disease 1 like 1', other_designations='PC1-like 1 protein|polycystin-1L1', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47774652, end_position_on_the_genomic_accession=47964635, orientation='minus', exon_count=58, OMIM=609721, no_hdr0=''),
  101618237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101618237, CurrentID=0, Status='live', Symbol='LOC101618237', Aliases='', description='ubiquitin conjugating enzyme E2F (putative) pseudogene', other_designations='ubiquitin-conjugating enzyme E2F (putative) pseudogene', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3295496, end_position_on_the_genomic_accession=3296367, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1247, CurrentID=0, Status='live', Symbol='CMR3A', Aliases='', description='cardiomyopathy, restrictive 3A (X-linked)', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874040, CurrentID=0, Status='live', Symbol='PEX5L-AS1', Aliases='', description='PEX5L antisense RNA 1', other_designations='PEX5L antisense RNA 1 (non-protein coding)', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  168537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168537, CurrentID=0, Status='live', Symbol='GIMAP7', Aliases='IAN7, hIAN7', description='GTPase, IMAP family member 7', other_designations='IAN-7|immune associated nucleotide|immune-associated nucleotide-binding protein 7|immunity-associated nucleotide 7 protein', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150514857, end_position_on_the_genomic_accession=150521073, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  442064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442064, CurrentID=0, Status='live', Symbol='LOC442064', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class Y pseudogene', other_designations='phosphatidylinositol glycan anchor biosynthesis, class Y pseudogene', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202773152, end_position_on_the_genomic_accession=202774445, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  168544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168544, CurrentID=0, Status='live', Symbol='ZNF467', Aliases='EZI, Zfp467', description='zinc finger protein 467', other_designations='zinc finger protein EZI', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149764182, end_position_on_the_genomic_accession=149776323, orientation='minus', exon_count=9, OMIM=614040, no_hdr0=''),
  100188945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188945, CurrentID=0, Status='live', Symbol='CDCA4P4', Aliases='', description='cell division cycle associated 4 pseudogene 4', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207762589, end_position_on_the_genomic_accession=207764587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874041, CurrentID=0, Status='live', Symbol='MACC1-AS1', Aliases='', description='MACC1 antisense RNA 1', other_designations='MACC1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20141916, end_position_on_the_genomic_accession=20153531, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  284942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284942, CurrentID=0, Status='live', Symbol='RPL23AP82', Aliases='RPL23A_43_1761', description='ribosomal protein L23a pseudogene 82', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50757086, end_position_on_the_genomic_accession=50799637, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480403, CurrentID=0, Status='live', Symbol='RNU6-935P', Aliases='', description='RNA, U6 small nuclear 935, pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378433, CurrentID=0, Status='live', Symbol='LOC105378433', Aliases='', description='uncharacterized LOC105378433', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91553336, end_position_on_the_genomic_accession=91574916, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480405, CurrentID=0, Status='live', Symbol='RNU6-1028P', Aliases='', description='RNA, U6 small nuclear 1028, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420545, CurrentID=0, Status='live', Symbol='LOC100420545', Aliases='', description='abhydrolase domain containing 15 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57629285, end_position_on_the_genomic_accession=57630201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480408, CurrentID=0, Status='live', Symbol='RNU6-1188P', Aliases='', description='RNA, U6 small nuclear 1188, pseudogene', other_designations='', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480409, CurrentID=0, Status='live', Symbol='RNU6-1262P', Aliases='', description='RNA, U6 small nuclear 1262, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480410, CurrentID=0, Status='live', Symbol='RNU6-1309P', Aliases='', description='RNA, U6 small nuclear 1309, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6257, CurrentID=0, Status='live', Symbol='RXRB', Aliases='DAUDI6, H-2RIIBP, NR2B2, RCoR-1', description='retinoid X receptor beta', other_designations='MHC class I promoter binding protein|nuclear receptor subfamily 2 group B member 2|retinoid X receptor, beta', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33193585, end_position_on_the_genomic_accession=33200853, orientation='minus', exon_count=11, OMIM=180246, no_hdr0=''),
  168620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168620, CurrentID=0, Status='live', Symbol='BHLHA15', Aliases='BHLHB8, MIST1', description='basic helix-loop-helix family member a15', other_designations='MIST-1|basic helix-loop-helix domain containing, class B, 8|basic helix-loop-helix family, member a15|class B basic helix-loop-helix protein 8|class II bHLH protein MIST1|muscle, intestine and stomach expression 1', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98212254, end_position_on_the_genomic_accession=98212959, orientation='plus', exon_count=1, OMIM=608606, no_hdr0=''),
  399474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399474, CurrentID=0, Status='live', Symbol='TMEM200B', Aliases='TTMB', description='transmembrane protein 200B', other_designations='transmembrane protein TTMA|two transmembrane B|two transmembrane domain-containing family member B', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=29119425, end_position_on_the_genomic_accession=29129134, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480415, CurrentID=0, Status='live', Symbol='RNU2-51P', Aliases='', description='RNA, U2 small nuclear 51, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373865, CurrentID=0, Status='live', Symbol='LOC105373865', Aliases='', description='uncharacterized LOC105373865', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214491075, end_position_on_the_genomic_accession=214500502, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  168667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168667, CurrentID=0, Status='live', Symbol='BMPER', Aliases='CRIM3, CV-2, CV2', description='BMP binding endothelial regulator', other_designations='BMP-binding endothelial regulator precursor protein|bone morphogenetic protein-binding endothelial cell precursor-derived regulator|crossveinless 2|crossveinless-2|hCV2', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33904911, end_position_on_the_genomic_accession=34155872, orientation='plus', exon_count=16, OMIM=608699, no_hdr0=''),
  105369501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369501, CurrentID=0, Status='live', Symbol='LOC105369501', Aliases='', description='uncharacterized LOC105369501', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113488737, end_position_on_the_genomic_accession=113500738, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  791231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791231, CurrentID=0, Status='live', Symbol='TRV-AAC3-1', Aliases='HtV2, TRNAV8, TRV8', description='transfer RNA-Val (AAC) 3-1', other_designations='tRNA-Val-AAC-3-1|transfer RNA valine 8 (anticodon AAC)|transfer RNA valine 8 (anticodon RAC)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27650928, end_position_on_the_genomic_accession=27651000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102929163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102929163, CurrentID=0, Status='live', Symbol='LOC102929163', Aliases='', description='ADP-ribosylation factor-like 17A pseudogene', other_designations='ADP-ribosylation factor like GTPase 17A pseudogene', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64870885, end_position_on_the_genomic_accession=64882891, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23076, CurrentID=0, Status='live', Symbol='RRP1B', Aliases='KIAA0179, NNP1L, Nnp1, PPP1R136, RRP1', description='ribosomal RNA processing 1B', other_designations='RRP1-like protein B|protein phosphatase 1, regulatory subunit 136|ribosomal RNA processing 1 homolog B', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43659551, end_position_on_the_genomic_accession=43696079, orientation='plus', exon_count=16, OMIM=610654, no_hdr0=''),
  105373867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373867, CurrentID=0, Status='live', Symbol='LOC105373867', Aliases='', description='uncharacterized LOC105373867', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215356855, end_position_on_the_genomic_accession=215364872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  791232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=791232, CurrentID=0, Status='live', Symbol='TRV-CAC1-3', Aliases='HtV3, TRNAV9, TRV9', description='transfer RNA-Val (CAC) 1-3', other_designations='tRNA-Val-CAC-1-3|transfer RNA valine 9 (anticodon CAC)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181102253, end_position_on_the_genomic_accession=181102325, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874047, CurrentID=0, Status='live', Symbol='LINC00499', Aliases='', description='long intergenic non-protein coding RNA 499', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138309711, end_position_on_the_genomic_accession=138424344, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  168741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168741, CurrentID=0, Status='live', Symbol='PER4', Aliases='', description='period circadian clock 3 pseudogene', other_designations='period 3 pseudogene|period 4 pseudogene|period homolog 3 pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9634270, end_position_on_the_genomic_accession=9635817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  399495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399495, CurrentID=0, Status='live', Symbol='SLC2A3P4', Aliases='', description='solute carrier family 2 (facilitated glucose transporter), member 3 pseudogene 4', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86503416, end_position_on_the_genomic_accession=86506947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399496, CurrentID=0, Status='live', Symbol='SLC2AXP1', Aliases='', description='solute carrier family 2 (facilitated glucose transporter), pseudogene 1', other_designations='solute carrier family 2 (facilitated glucose transporter), member 1 pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95196495, end_position_on_the_genomic_accession=95197384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644249, CurrentID=0, Status='live', Symbol='LOC644249', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67903028, end_position_on_the_genomic_accession=67920548, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874048, CurrentID=0, Status='live', Symbol='DGUOK-AS1', Aliases='', description='DGUOK antisense RNA 1', other_designations='DGUOK antisense RNA 1 (non-protein coding)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73957967, end_position_on_the_genomic_accession=73981439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100529251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529251, CurrentID=0, Status='live', Symbol='CKLF-CMTM1', Aliases='', description='CKLF-CMTM1 readthrough', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66552563, end_position_on_the_genomic_accession=66579135, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372179, CurrentID=0, Status='live', Symbol='LOC105372179', Aliases='', description='uncharacterized LOC105372179', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=69469932, end_position_on_the_genomic_accession=69471736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480439, CurrentID=0, Status='live', Symbol='RNA5SP312', Aliases='RN5S312, RN5S313, RNA5SP313', description='RNA, 5S ribosomal pseudogene 312', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480441, CurrentID=0, Status='live', Symbol='GCSHP1', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene 1', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928595, CurrentID=0, Status='live', Symbol='LOC101928595', Aliases='', description='uncharacterized LOC101928595', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30096430, end_position_on_the_genomic_accession=30104116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480442, CurrentID=0, Status='live', Symbol='SUB1P3', Aliases='', description='SUB1 homolog, transcriptional regulator pseudogene 3', other_designations='SUB1 pseudogene 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284950, CurrentID=0, Status='live', Symbol='LOC284950', Aliases='', description='uncharacterized LOC284950', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85815130, end_position_on_the_genomic_accession=85826788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480443, CurrentID=0, Status='live', Symbol='PKMP4', Aliases='', description='pyruvate kinase, muscle pseudogene 4', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480444, CurrentID=0, Status='live', Symbol='FCF1P9', Aliases='', description='FCF1 pseudogene 9', other_designations='', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137362, CurrentID=0, Status='live', Symbol='GOT1L1', Aliases='', description='glutamic-oxaloacetic transaminase 1-like 1', other_designations='glutamate oxaloacetate transaminase 1-like protein 1|transaminase A-like protein 1', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37934281, end_position_on_the_genomic_accession=37940157, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100188989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188989, CurrentID=0, Status='live', Symbol='TRU-TCA3-1', Aliases='TRNAU3', description='transfer RNA-SeC (TCA) 3-1', other_designations='tRNA-SeC-TCA-3-1|transfer RNA selenocysteine 3 (anticodon UCA)', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480446, CurrentID=0, Status='live', Symbol='EIF4A2P5', Aliases='', description='eukaryotic translation initiation factor 4A2 pseudogene 5', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480447, CurrentID=0, Status='live', Symbol='RAB11AP2', Aliases='', description='RAB11A, member RAS oncogene family pseudogene 2', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  202025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202025, CurrentID=0, Status='live', Symbol='LOC202025', Aliases='', description='uncharacterized LOC202025', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  693120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693120, CurrentID=0, Status='live', Symbol='MIR33B', Aliases='MIRN33B, hsa-mir-33b, mir-33b', description='microRNA 33b', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17813836, end_position_on_the_genomic_accession=17813931, orientation='minus', exon_count=1, OMIM=613486, no_hdr0=''),
  693121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693121, CurrentID=0, Status='live', Symbol='MIR411', Aliases='MIRN411, hsa-mir-411, mir-411', description='microRNA 411', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101023325, end_position_on_the_genomic_accession=101023420, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693122, CurrentID=0, Status='live', Symbol='MIR421', Aliases='MIRN421, hsa-mir-421', description='microRNA 421', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74218377, end_position_on_the_genomic_accession=74218461, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693123, CurrentID=0, Status='live', Symbol='MIR449B', Aliases='MIRN449B, mir-449b', description='microRNA 449b', other_designations='hsa-mir-449b', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55170646, end_position_on_the_genomic_accession=55170742, orientation='minus', exon_count=1, OMIM=613132, no_hdr0=''),
  693124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693124, CurrentID=0, Status='live', Symbol='MIR532', Aliases='MIRN532, hsa-mir-532, mir-532', description='microRNA 532', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50003148, end_position_on_the_genomic_accession=50003238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693125, CurrentID=0, Status='live', Symbol='MIR548A1', Aliases='MIRN548A1', description='microRNA 548a-1', other_designations='hsa-mir-548a-1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18571784, end_position_on_the_genomic_accession=18571880, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693126, CurrentID=0, Status='live', Symbol='MIR548A2', Aliases='MIRN548A2', description='microRNA 548a-2', other_designations='hsa-mir-548a-2', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135239160, end_position_on_the_genomic_accession=135239256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693127, CurrentID=0, Status='live', Symbol='MIR548A3', Aliases='MIRN548A3', description='microRNA 548a-3', other_designations='hsa-mir-548a-3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104484369, end_position_on_the_genomic_accession=104484465, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693128, CurrentID=0, Status='live', Symbol='MIR548B', Aliases='MIRN548B', description='microRNA 548b', other_designations='hsa-mir-548b', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119069047, end_position_on_the_genomic_accession=119069143, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693129, CurrentID=0, Status='live', Symbol='MIR548C', Aliases='MIRN548C', description='microRNA 548c', other_designations='hsa-mir-548c', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64622509, end_position_on_the_genomic_accession=64622605, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693130, CurrentID=0, Status='live', Symbol='MIR548D1', Aliases='MIRN548D1', description='microRNA 548d-1', other_designations='hsa-mir-548d-1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123348034, end_position_on_the_genomic_accession=123348130, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  6295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6295, CurrentID=0, Status='live', Symbol='SAG', Aliases='RP47, S-AG', description='S-antigen; retina and pineal gland (arrestin)', other_designations='48 kDa protein|arrestin 1|retinal S-antigen (48 KDa protein)|rod photoreceptor arrestin', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233304300, end_position_on_the_genomic_accession=233347055, orientation='plus', exon_count=21, OMIM=181031, no_hdr0=''),
  693132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693132, CurrentID=0, Status='live', Symbol='MIR549A', Aliases='MIR549, MIRN549, hsa-mir-549, hsa-mir-549a, mir-549a', description='microRNA 549a', other_designations='microRNA 549', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80841978, end_position_on_the_genomic_accession=80842073, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693133, CurrentID=0, Status='live', Symbol='MIR550A1', Aliases='MIR550-1, MIRN550-1, hsa-mir-550a-1, mir-550a-1', description='microRNA 550a-1', other_designations='hsa-mir-550-1|microRNA 550-1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30289794, end_position_on_the_genomic_accession=30289890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693134, CurrentID=0, Status='live', Symbol='MIR550A2', Aliases='MIR550-2, MIRN550-2, hsa-mir-550a-2, mir-550a-2', description='microRNA 550a-2', other_designations='hsa-mir-550-2|microRNA 550-2', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32732981, end_position_on_the_genomic_accession=32733077, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693135, CurrentID=0, Status='live', Symbol='MIR551A', Aliases='MIRN551A, mir-551a', description='microRNA 551a', other_designations='hsa-mir-551a', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3560695, end_position_on_the_genomic_accession=3560790, orientation='minus', exon_count=1, OMIM=615148, no_hdr0=''),
  693136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693136, CurrentID=0, Status='live', Symbol='MIR551B', Aliases='MIRN551B, mir-551b', description='microRNA 551b', other_designations='hsa-mir-551b', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168551854, end_position_on_the_genomic_accession=168551949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693137, CurrentID=0, Status='live', Symbol='MIR552', Aliases='MIRN552, hsa-mir-552', description='microRNA 552', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34669599, end_position_on_the_genomic_accession=34669694, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  168850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168850, CurrentID=0, Status='live', Symbol='ZNF800', Aliases='', description='zinc finger protein 800', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127346721, end_position_on_the_genomic_accession=127392724, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  693139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693139, CurrentID=0, Status='live', Symbol='MIR554', Aliases='MIRN554, hsa-mir-554', description='microRNA 554', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151545796, end_position_on_the_genomic_accession=151545891, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693140, CurrentID=0, Status='live', Symbol='MIR555', Aliases='MIRN555, hsa-mir-555', description='microRNA 555', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155346350, end_position_on_the_genomic_accession=155346445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693141, CurrentID=0, Status='live', Symbol='MIR556', Aliases='MIRN556, hsa-mir-556, mir-556', description='microRNA 556', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162342546, end_position_on_the_genomic_accession=162342640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693142, CurrentID=0, Status='live', Symbol='MIR557', Aliases='MIRN557, hsa-mir-557', description='microRNA 557', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168375524, end_position_on_the_genomic_accession=168375621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693143, CurrentID=0, Status='live', Symbol='MIR558', Aliases='MIRN558, hsa-mir-558', description='microRNA 558', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32532153, end_position_on_the_genomic_accession=32532246, orientation='plus', exon_count=1, OMIM=616473, no_hdr0=''),
  693144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693144, CurrentID=0, Status='live', Symbol='MIR559', Aliases='MIRN559, hsa-mir-559', description='microRNA 559', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47377675, end_position_on_the_genomic_accession=47377770, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480452, CurrentID=0, Status='live', Symbol='RNU7-161P', Aliases='', description='RNA, U7 small nuclear 161 pseudogene', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  693146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693146, CurrentID=0, Status='live', Symbol='MIR561', Aliases='MIRN561, hsa-mir-561, mir-561', description='microRNA 561', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188297492, end_position_on_the_genomic_accession=188297588, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693147, CurrentID=0, Status='live', Symbol='MIR562', Aliases='MIRN562, hsa-mir-562', description='microRNA 562', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232172653, end_position_on_the_genomic_accession=232172747, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693148, CurrentID=0, Status='live', Symbol='MIR563', Aliases='MIRN563, hsa-mir-563', description='microRNA 563', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15873771, end_position_on_the_genomic_accession=15873849, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693149, CurrentID=0, Status='live', Symbol='MIR564', Aliases='MIRN564, hsa-mir-564', description='microRNA 564', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44861888, end_position_on_the_genomic_accession=44861981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693151, CurrentID=0, Status='live', Symbol='MIR566', Aliases='MIRN566, hsa-mir-566', description='microRNA 566', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50173326, end_position_on_the_genomic_accession=50173419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693152, CurrentID=0, Status='live', Symbol='MIR567', Aliases='MIRN567, hsa-mir-567', description='microRNA 567', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112112801, end_position_on_the_genomic_accession=112112898, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693153, CurrentID=0, Status='live', Symbol='MIR568', Aliases='MIRN568, hsa-mir-568', description='microRNA 568', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114316475, end_position_on_the_genomic_accession=114316569, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693154, CurrentID=0, Status='live', Symbol='MIR569', Aliases='MIRN569, hsa-mir-569', description='microRNA 569', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171106664, end_position_on_the_genomic_accession=171106759, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693155, CurrentID=0, Status='live', Symbol='MIR570', Aliases='MIRN570, hsa-mir-570', description='microRNA 570', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195699401, end_position_on_the_genomic_accession=195699497, orientation='plus', exon_count=1, OMIM=614538, no_hdr0=''),
  693156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693156, CurrentID=0, Status='live', Symbol='MIR571', Aliases='MIRN571, hsa-mir-571', description='microRNA 571', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=350157, end_position_on_the_genomic_accession=350252, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693157, CurrentID=0, Status='live', Symbol='MIR572', Aliases='MIRN572, hsa-mir-572', description='microRNA 572', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11368827, end_position_on_the_genomic_accession=11368921, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693158, CurrentID=0, Status='live', Symbol='MIR573', Aliases='MIRN573, hsa-mir-573, mir-573', description='microRNA 573', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24520192, end_position_on_the_genomic_accession=24520290, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693159, CurrentID=0, Status='live', Symbol='MIR574', Aliases='MIR574-3p, MIRN574, hsa-mir-574, mir-574', description='microRNA 574', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38868032, end_position_on_the_genomic_accession=38868127, orientation='plus', exon_count=1, OMIM=615469, no_hdr0=''),
  693160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693160, CurrentID=0, Status='live', Symbol='MIR575', Aliases='MIRN575, hsa-mir-575', description='microRNA 575', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82753337, end_position_on_the_genomic_accession=82753430, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693161, CurrentID=0, Status='live', Symbol='MIR576', Aliases='MIRN576, hsa-mir-576', description='microRNA 576', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109488698, end_position_on_the_genomic_accession=109488795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693162, CurrentID=0, Status='live', Symbol='MIR577', Aliases='MIRN577, hsa-mir-577, mir-577', description='microRNA 577', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114656759, end_position_on_the_genomic_accession=114656854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693163, CurrentID=0, Status='live', Symbol='MIR578', Aliases='MIRN578, hsa-mir-578', description='microRNA 578', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165386242, end_position_on_the_genomic_accession=165386337, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693164, CurrentID=0, Status='live', Symbol='MIR579', Aliases='MIRN579, hsa-mir-579', description='microRNA 579', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32394378, end_position_on_the_genomic_accession=32394475, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693165, CurrentID=0, Status='live', Symbol='MIR580', Aliases='MIRN580, hsa-mir-580', description='microRNA 580', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36147892, end_position_on_the_genomic_accession=36147988, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693166, CurrentID=0, Status='live', Symbol='MIR581', Aliases='MIRN581, hsa-mir-581', description='microRNA 581', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53951504, end_position_on_the_genomic_accession=53951599, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693167, CurrentID=0, Status='live', Symbol='MIR582', Aliases='MIRN582, hsa-mir-582', description='microRNA 582', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=59703606, end_position_on_the_genomic_accession=59703703, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693168, CurrentID=0, Status='live', Symbol='MIR583', Aliases='MIRN583, hsa-mir-583', description='microRNA 583', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96079138, end_position_on_the_genomic_accession=96079212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693169, CurrentID=0, Status='live', Symbol='MIR584', Aliases='MIRN584, hsa-mir-584', description='microRNA 584', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149062313, end_position_on_the_genomic_accession=149062409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693170, CurrentID=0, Status='live', Symbol='MIR585', Aliases='MIRN585, hsa-mir-585, mir-585', description='microRNA 585', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169263601, end_position_on_the_genomic_accession=169263694, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693171, CurrentID=0, Status='live', Symbol='MIR586', Aliases='MIRN586, hsa-mir-586', description='microRNA 586', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45197674, end_position_on_the_genomic_accession=45197770, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693172, CurrentID=0, Status='live', Symbol='MIR587', Aliases='MIRN587, hsa-mir-587', description='microRNA 587', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106784125, end_position_on_the_genomic_accession=106784220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693173, CurrentID=0, Status='live', Symbol='MIR588', Aliases='MIRN588, hsa-mir-588', description='microRNA 588', other_designations='', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126484631, end_position_on_the_genomic_accession=126484713, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693174, CurrentID=0, Status='live', Symbol='MIR589', Aliases='MIRN589, hsa-mir-589, mir-589', description='microRNA 589', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5495819, end_position_on_the_genomic_accession=5495917, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693175, CurrentID=0, Status='live', Symbol='MIR590', Aliases='MIRN590, hsa-mir-590, mir-590', description='microRNA 590', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74191198, end_position_on_the_genomic_accession=74191294, orientation='plus', exon_count=1, OMIM=615070, no_hdr0=''),
  693176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693176, CurrentID=0, Status='live', Symbol='MIR591', Aliases='MIRN591, hsa-mir-591', description='microRNA 591', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96219662, end_position_on_the_genomic_accession=96219756, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693177, CurrentID=0, Status='live', Symbol='MIR592', Aliases='MIRN592, hsa-mir-592, mir-592', description='microRNA 592', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127058088, end_position_on_the_genomic_accession=127058184, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693178, CurrentID=0, Status='live', Symbol='MIR593', Aliases='MIRN593, hsa-mir-593', description='microRNA 593', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128081861, end_position_on_the_genomic_accession=128081960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693180, CurrentID=0, Status='live', Symbol='MIR595', Aliases='MIRN595, hsa-mir-595', description='microRNA 595', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158532718, end_position_on_the_genomic_accession=158532813, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693181, CurrentID=0, Status='live', Symbol='MIR596', Aliases='MIRN596, hsa-mir-596', description='microRNA 596', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1817231, end_position_on_the_genomic_accession=1817307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693182, CurrentID=0, Status='live', Symbol='MIR597', Aliases='MIRN597, hsa-mir-597, mir-597', description='microRNA 597', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9741672, end_position_on_the_genomic_accession=9741768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693183, CurrentID=0, Status='live', Symbol='MIR598', Aliases='MIRN598, hsa-mir-598, mir-598', description='microRNA 598', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11035206, end_position_on_the_genomic_accession=11035302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693184, CurrentID=0, Status='live', Symbol='MIR599', Aliases='MIRN599, hsa-mir-599', description='microRNA 599', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99536636, end_position_on_the_genomic_accession=99536730, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  399520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399520, CurrentID=0, Status='live', Symbol='TIMM8AP1', Aliases='DDPP, TIMM8AP', description='translocase of inner mitochondrial membrane 8 homolog A (yeast) pseudogene 1', other_designations='deafness dystonia pseudogene', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162076515, end_position_on_the_genomic_accession=162077711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  693186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693186, CurrentID=0, Status='live', Symbol='MIR601', Aliases='MIRN601, hsa-mir-601', description='microRNA 601', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123402525, end_position_on_the_genomic_accession=123402603, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693187, CurrentID=0, Status='live', Symbol='MIR602', Aliases='MIRN602, hsa-mir-602', description='microRNA 602', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137838419, end_position_on_the_genomic_accession=137838516, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693188, CurrentID=0, Status='live', Symbol='MIR603', Aliases='MIRN603, hsa-mir-603', description='microRNA 603', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24275685, end_position_on_the_genomic_accession=24275781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693189, CurrentID=0, Status='live', Symbol='MIR604', Aliases='MIRN604, hsa-mir-604', description='microRNA 604', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29545004, end_position_on_the_genomic_accession=29545097, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693190, CurrentID=0, Status='live', Symbol='MIR605', Aliases='MIRN605, hsa-mir-605, mir-605', description='microRNA 605', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51299573, end_position_on_the_genomic_accession=51299655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693191, CurrentID=0, Status='live', Symbol='MIR606', Aliases='MIRN606, hsa-mir-606', description='microRNA 606', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75552458, end_position_on_the_genomic_accession=75552553, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693192, CurrentID=0, Status='live', Symbol='MIR607', Aliases='MIRN607, hsa-mir-607', description='microRNA 607', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96828669, end_position_on_the_genomic_accession=96828764, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693193, CurrentID=0, Status='live', Symbol='MIR608', Aliases='MIRN608, hsa-mir-608', description='microRNA 608', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100974985, end_position_on_the_genomic_accession=100975084, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693194, CurrentID=0, Status='live', Symbol='MIR609', Aliases='MIRN609, hsa-mir-609', description='microRNA 609', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104218789, end_position_on_the_genomic_accession=104218883, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693195, CurrentID=0, Status='live', Symbol='MIR610', Aliases='MIRN610, hsa-mir-610', description='microRNA 610', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28056815, end_position_on_the_genomic_accession=28056910, orientation='plus', exon_count=1, OMIM=612330, no_hdr0=''),
  693196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693196, CurrentID=0, Status='live', Symbol='MIR611', Aliases='MIRN611, hsa-mir-611', description='microRNA 611', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61792495, end_position_on_the_genomic_accession=61792561, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693197, CurrentID=0, Status='live', Symbol='MIR612', Aliases='MIRN612, hsa-mir-612', description='microRNA 612', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65444458, end_position_on_the_genomic_accession=65444557, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693198, CurrentID=0, Status='live', Symbol='MIR613', Aliases='MIRN613, hsa-mir-613', description='microRNA 613', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12764649, end_position_on_the_genomic_accession=12764743, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693199, CurrentID=0, Status='live', Symbol='MIR614', Aliases='MIRN614, hsa-mir-614', description='microRNA 614', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12915829, end_position_on_the_genomic_accession=12915918, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693200, CurrentID=0, Status='live', Symbol='MIR615', Aliases='MIRN615, hsa-mir-615, mir-615', description='microRNA 615', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54033950, end_position_on_the_genomic_accession=54034045, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693201, CurrentID=0, Status='live', Symbol='MIR616', Aliases='MIRN616, hsa-mir-616', description='microRNA 616', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57519163, end_position_on_the_genomic_accession=57519259, orientation='minus', exon_count=1, OMIM=614489, no_hdr0=''),
  693202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693202, CurrentID=0, Status='live', Symbol='MIR617', Aliases='MIRN617, hsa-mir-617', description='microRNA 617', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80832533, end_position_on_the_genomic_accession=80832629, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693203, CurrentID=0, Status='live', Symbol='MIR618', Aliases='MIRN618, hsa-mir-618, mir-618', description='microRNA 618', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80935736, end_position_on_the_genomic_accession=80935833, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693204, CurrentID=0, Status='live', Symbol='MIR619', Aliases='MIRN619, hsa-mir-619', description='microRNA 619', other_designations='miR-619-5p', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108836908, end_position_on_the_genomic_accession=108837006, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693205, CurrentID=0, Status='live', Symbol='MIR620', Aliases='MIRN620, hsa-mir-620', description='microRNA 620', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116148560, end_position_on_the_genomic_accession=116148654, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693206, CurrentID=0, Status='live', Symbol='MIR621', Aliases='MIRN621, hsa-mir-621', description='microRNA 621', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40810766, end_position_on_the_genomic_accession=40810861, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693207, CurrentID=0, Status='live', Symbol='MIR622', Aliases='MIRN622, hsa-mir-622', description='microRNA 622', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90231182, end_position_on_the_genomic_accession=90231277, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693208, CurrentID=0, Status='live', Symbol='MIR623', Aliases='MIRN623, hsa-mir-623', description='microRNA 623', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99356131, end_position_on_the_genomic_accession=99356228, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693209, CurrentID=0, Status='live', Symbol='MIR624', Aliases='MIRN624, hsa-mir-624, mir-624', description='microRNA 624', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31014646, end_position_on_the_genomic_accession=31014742, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693210, CurrentID=0, Status='live', Symbol='MIR625', Aliases='MIRN625, hsa-mir-625', description='microRNA 625', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65471102, end_position_on_the_genomic_accession=65471186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693211, CurrentID=0, Status='live', Symbol='MIR626', Aliases='MIRN626, hsa-mir-626', description='microRNA 626', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41691585, end_position_on_the_genomic_accession=41691678, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693212, CurrentID=0, Status='live', Symbol='MIR627', Aliases='MIRN627, hsa-mir-627, mir-627', description='microRNA 627', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42199570, end_position_on_the_genomic_accession=42199666, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693213, CurrentID=0, Status='live', Symbol='MIR628', Aliases='MIRN628, hsa-mir-628, mir-628', description='microRNA 628', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55372940, end_position_on_the_genomic_accession=55373034, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693214, CurrentID=0, Status='live', Symbol='MIR629', Aliases='MIRN629, hsa-mir-629, mir-629', description='microRNA 629', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70079372, end_position_on_the_genomic_accession=70079468, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693215, CurrentID=0, Status='live', Symbol='MIR630', Aliases='MIRN630, hsa-mir-630', description='microRNA 630', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72587217, end_position_on_the_genomic_accession=72587313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693216, CurrentID=0, Status='live', Symbol='MIR631', Aliases='MIRN631, hsa-mir-631', description='microRNA 631', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75353611, end_position_on_the_genomic_accession=75353685, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693217, CurrentID=0, Status='live', Symbol='MIR632', Aliases='MIRN632, hsa-mir-632', description='microRNA 632', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32350109, end_position_on_the_genomic_accession=32350202, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693218, CurrentID=0, Status='live', Symbol='MIR633', Aliases='MIRN633, hsa-mir-633', description='microRNA 633', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62944215, end_position_on_the_genomic_accession=62944312, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693219, CurrentID=0, Status='live', Symbol='MIR634', Aliases='MIRN634, hsa-mir-634', description='microRNA 634', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66787072, end_position_on_the_genomic_accession=66787168, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693220, CurrentID=0, Status='live', Symbol='MIR635', Aliases='MIRN635, hsa-mir-635', description='microRNA 635', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68424451, end_position_on_the_genomic_accession=68424548, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693221, CurrentID=0, Status='live', Symbol='MIR636', Aliases='MIRN636, hsa-mir-636', description='microRNA 636', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76736450, end_position_on_the_genomic_accession=76736548, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693222, CurrentID=0, Status='live', Symbol='MIR637', Aliases='MIRN637, hsa-mir-637', description='microRNA 637', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3961414, end_position_on_the_genomic_accession=3961512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693223, CurrentID=0, Status='live', Symbol='MIR638', Aliases='MIRN638, hsa-mir-638', description='microRNA 638', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10718404, end_position_on_the_genomic_accession=10718503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693224, CurrentID=0, Status='live', Symbol='MIR639', Aliases='MIRN639, hsa-mir-639', description='microRNA 639', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14529543, end_position_on_the_genomic_accession=14529640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693225, CurrentID=0, Status='live', Symbol='MIR640', Aliases='MIRN640, hsa-mir-640', description='microRNA 640', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19435063, end_position_on_the_genomic_accession=19435158, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693226, CurrentID=0, Status='live', Symbol='MIR641', Aliases='MIRN641, hsa-mir-641, mir-641', description='microRNA 641', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40282543, end_position_on_the_genomic_accession=40282641, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693227, CurrentID=0, Status='live', Symbol='MIR642A', Aliases='MIR642, MIRN642, hsa-mir-642, hsa-mir-642a, mir-642a', description='microRNA 642a', other_designations='microRNA 642|microRNA mir-642a', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45674928, end_position_on_the_genomic_accession=45675024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693228, CurrentID=0, Status='live', Symbol='MIR643', Aliases='MIRN643, hsa-mir-643, mir-643', description='microRNA 643', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52281797, end_position_on_the_genomic_accession=52281893, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693229, CurrentID=0, Status='live', Symbol='MIR644A', Aliases='MIR644, MIRN644, hsa-mir-644, hsa-mir-644a', description='microRNA 644a', other_designations='microRNA 644', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34466325, end_position_on_the_genomic_accession=34466418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693230, CurrentID=0, Status='live', Symbol='MIR645', Aliases='MIRN645, hsa-mir-645', description='microRNA 645', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50585786, end_position_on_the_genomic_accession=50585879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693231, CurrentID=0, Status='live', Symbol='MIR646', Aliases='MIRN646, hsa-mir-646', description='microRNA 646', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60308474, end_position_on_the_genomic_accession=60308567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  693232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693232, CurrentID=0, Status='live', Symbol='MIR647', Aliases='MIRN647, hsa-mir-647', description='microRNA 647', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63942631, end_position_on_the_genomic_accession=63942726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693233, CurrentID=0, Status='live', Symbol='MIR648', Aliases='MIRN648, hsa-mir-648', description='microRNA 648', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17980868, end_position_on_the_genomic_accession=17980961, orientation='minus', exon_count=1, OMIM=616205, no_hdr0=''),
  693234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693234, CurrentID=0, Status='live', Symbol='MIR649', Aliases='MIRN649, hsa-mir-649', description='microRNA 649', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21034176, end_position_on_the_genomic_accession=21034272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  693235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693235, CurrentID=0, Status='live', Symbol='MIR92B', Aliases='MIRN92B, hsa-mir-92b, mir-92b', description='microRNA 92b', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155195177, end_position_on_the_genomic_accession=155195272, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100189012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189012, CurrentID=0, Status='live', Symbol='TRR-TCT5-1', Aliases='TRNAR5', description='transfer RNA-Arg (TCT) 5-1', other_designations='transfer RNA arginine 5 (anticodon UCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27562184, end_position_on_the_genomic_accession=27562270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  677762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677762, CurrentID=0, Status='live', Symbol='LOC677762', Aliases='', description='Sin3A associated protein 30kDa pseudogene', other_designations='Sin3A-associated protein, 30kDa pseudogene|sin3 associated polypeptide pseudogene', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150611053, end_position_on_the_genomic_accession=150612090, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189013, CurrentID=0, Status='live', Symbol='TRK-CTT1-2', Aliases='TRNAK6', description='transfer RNA-Lys (CTT) 1-2', other_designations='tRNA-Lys-CTT-1-2|transfer RNA lysine 6 (anticodon CUU)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78860562, end_position_on_the_genomic_accession=78860634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480470, CurrentID=0, Status='live', Symbol='RN7SKP184', Aliases='', description='RNA, 7SK small nuclear pseudogene 184', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646197, CurrentID=0, Status='live', Symbol='LOC646197', Aliases='', description='heat shock protein 90kDa alpha family class B member 1 pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73215929, end_position_on_the_genomic_accession=73216627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480471, CurrentID=0, Status='live', Symbol='RN7SKP191', Aliases='', description='RNA, 7SK small nuclear pseudogene 191', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  168975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=168975, CurrentID=0, Status='live', Symbol='CNBD1', Aliases='', description='cyclic nucleotide binding domain containing 1', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86866448, end_position_on_the_genomic_accession=87382727, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106480472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480472, CurrentID=0, Status='live', Symbol='RN7SKP202', Aliases='', description='RNA, 7SK small nuclear pseudogene 202', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480473, CurrentID=0, Status='live', Symbol='RN7SKP214', Aliases='', description='RNA, 7SK small nuclear pseudogene 214', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6319, CurrentID=0, Status='live', Symbol='SCD', Aliases='FADS5, MSTP0081, SCDOS, hSCD1, SCD', description='stearoyl-CoA desaturase (delta-9-desaturase)', other_designations='delta(9)-desaturase|delta-9 desaturase|delta-9-desaturase|fatty acid desaturase|predicted protein of HQ0998|stearoyl-CoA desaturase opposite strand', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100347015, end_position_on_the_genomic_accession=100364831, orientation='plus', exon_count=6, OMIM=604031, no_hdr0=''),
  106480474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480474, CurrentID=0, Status='live', Symbol='RN7SKP221', Aliases='', description='RNA, 7SK small nuclear pseudogene 221', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137392, CurrentID=0, Status='live', Symbol='FAM92A1', Aliases='', description='family with sequence similarity 92 member A1', other_designations='family with sequence similarity 92, member A1', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93700252, end_position_on_the_genomic_accession=93729253, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100189019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189019, CurrentID=0, Status='live', Symbol='TRP-AGG2-1', Aliases='TRNAP4', description='transfer RNA-Pro (AGG) 2-1', other_designations='tRNA-Pro-AGG-2-1|transfer RNA proline 4 (anticodon AGG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167715488, end_position_on_the_genomic_accession=167715559, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480476, CurrentID=0, Status='live', Symbol='RN7SKP248', Aliases='', description='RNA, 7SK small nuclear pseudogene 248', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480477, CurrentID=0, Status='live', Symbol='RN7SKP261', Aliases='', description='RNA, 7SK small nuclear pseudogene 261', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189022, CurrentID=0, Status='live', Symbol='TRY-GTA2-1', Aliases='TRNAY4', description='transfer RNA-Tyr (GTA) 2-1', other_designations='transfer RNA tyrosine 4 (anticodon GUA)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27050782, end_position_on_the_genomic_accession=27050870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103191607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103191607, CurrentID=0, Status='live', Symbol='LOC103191607', Aliases='', description='uncharacterized LOC103191607', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52334295, end_position_on_the_genomic_accession=52341883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480480, CurrentID=0, Status='live', Symbol='RN7SKP290', Aliases='', description='RNA, 7SK small nuclear pseudogene 290', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169026, CurrentID=0, Status='live', Symbol='SLC30A8', Aliases='ZNT8, ZnT-8', description='solute carrier family 30 (zinc transporter), member 8', other_designations='zinc transporter ZnT-8', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116950180, end_position_on_the_genomic_accession=117176714, orientation='plus', exon_count=14, OMIM=611145, no_hdr0=''),
  106480481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480481, CurrentID=0, Status='live', Symbol='RN7SL14P', Aliases='', description='RNA, 7SL, cytoplasmic 14, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481554, CurrentID=0, Status='live', Symbol='RNU6-1187P', Aliases='', description='RNA, U6 small nuclear 1187, pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480482, CurrentID=0, Status='live', Symbol='RN7SL23P', Aliases='', description='RNA, 7SL, cytoplasmic 23, pseudogene', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874057, CurrentID=0, Status='live', Symbol='LINC00283', Aliases='NCRNA00283', description='long intergenic non-protein coding RNA 283', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284958, CurrentID=0, Status='live', Symbol='NT5DC4', Aliases='', description="5'-nucleotidase domain containing 4", other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112720810, end_position_on_the_genomic_accession=112742879, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  169044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169044, CurrentID=0, Status='live', Symbol='COL22A1', Aliases='', description='collagen, type XXII, alpha 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=138588235, end_position_on_the_genomic_accession=138914094, orientation='minus', exon_count=70, OMIM=610026, no_hdr0=''),
  132200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132200, CurrentID=0, Status='live', Symbol='C3orf49', Aliases='', description='chromosome 3 open reading frame 49', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63819365, end_position_on_the_genomic_accession=63848636, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106480485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480485, CurrentID=0, Status='live', Symbol='RN7SL66P', Aliases='', description='RNA, 7SL, cytoplasmic 66, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480486, CurrentID=0, Status='live', Symbol='RN7SL96P', Aliases='', description='RNA, 7SL, cytoplasmic 96, pseudogene', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480487, CurrentID=0, Status='live', Symbol='RN7SL105P', Aliases='', description='RNA, 7SL, cytoplasmic 105, pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874058, CurrentID=0, Status='live', Symbol='COX10-AS1', Aliases='COX10-AS, COX10AS', description='COX10 antisense RNA 1', other_designations='COX10 antisense RNA (non-protein coding)|COX10 antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14029292, end_position_on_the_genomic_accession=14069458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480488, CurrentID=0, Status='live', Symbol='RN7SL116P', Aliases='', description='RNA, 7SL, cytoplasmic 116, pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137406, CurrentID=0, Status='live', Symbol='MRPS16P1', Aliases='', description='mitochondrial ribosomal protein S16 pseudogene 1', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91917507, end_position_on_the_genomic_accession=91917922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480489, CurrentID=0, Status='live', Symbol='RN7SL138P', Aliases='', description='RNA, 7SL, cytoplasmic 138, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480490, CurrentID=0, Status='live', Symbol='RN7SL153P', Aliases='', description='RNA, 7SL, cytoplasmic 153, pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480491, CurrentID=0, Status='live', Symbol='RN7SL168P', Aliases='', description='RNA, 7SL, cytoplasmic 168, pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480492, CurrentID=0, Status='live', Symbol='RN7SL181P', Aliases='', description='RNA, 7SL, cytoplasmic 181, pseudogene', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480493, CurrentID=0, Status='live', Symbol='RN7SL193P', Aliases='', description='RNA, 7SL, cytoplasmic 193, pseudogene', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104502416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104502416, CurrentID=0, Status='live', Symbol='FOXP1-AS1', Aliases='', description='FOXP1 antisense RNA 1', other_designations='FOXP1 antisense RNA 1 (non-protein coding)', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71289758, end_position_on_the_genomic_accession=71305854, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  104502417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104502417, CurrentID=0, Status='live', Symbol='MGAT3-AS1', Aliases='TapSAKI', description='MGAT3 antisense RNA 1', other_designations='transcript predicting survival in AKI (acute kidney injury)', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39475807, end_position_on_the_genomic_accession=39476822, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189038, CurrentID=0, Status='live', Symbol='TRL-CAA3-1', Aliases='TRNAL8', description='transfer RNA-Leu (CAA) 3-1', other_designations='transfer RNA leucine 8 (anticodon CAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27602569, end_position_on_the_genomic_accession=27602675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480495, CurrentID=0, Status='live', Symbol='RN7SL216P', Aliases='', description='RNA, 7SL, cytoplasmic 216, pseudogene', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480496, CurrentID=0, Status='live', Symbol='RN7SL242P', Aliases='', description='RNA, 7SL, cytoplasmic 242, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480497, CurrentID=0, Status='live', Symbol='RN7SL260P', Aliases='', description='RNA, 7SL, cytoplasmic 260, pseudogene', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874060, CurrentID=0, Status='live', Symbol='SPATA20P1', Aliases='', description='spermatogenesis associated 20 pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38238227, end_position_on_the_genomic_accession=38238664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480498, CurrentID=0, Status='live', Symbol='RN7SL280P', Aliases='', description='RNA, 7SL, cytoplasmic 280, pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132203, CurrentID=0, Status='live', Symbol='SNTN', Aliases='S100A1L, S100AL, sentan', description='sentan, cilia apical structure protein', other_designations='S100A-like protein', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63652668, end_position_on_the_genomic_accession=63665217, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724617, CurrentID=0, Status='live', Symbol='LOC102724617', Aliases='', description='uncharacterized LOC102724617', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182644487, end_position_on_the_genomic_accession=182672846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480500, CurrentID=0, Status='live', Symbol='RN7SL307P', Aliases='', description='RNA, 7SL, cytoplasmic 307, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480501, CurrentID=0, Status='live', Symbol='RN7SL321P', Aliases='', description='RNA, 7SL, cytoplasmic 321, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480502, CurrentID=0, Status='live', Symbol='RN7SL334P', Aliases='', description='RNA, 7SL, cytoplasmic 334, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128192, CurrentID=0, Status='live', Symbol='LOC128192', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44988146, end_position_on_the_genomic_accession=44988782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399564, CurrentID=0, Status='live', Symbol='GLC1H', Aliases='', description='glaucoma 1, open angle, H (adult-onset)', other_designations='', map_location='2p16.3-p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611276, no_hdr0=''),
  106480503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480503, CurrentID=0, Status='live', Symbol='RN7SL344P', Aliases='', description='RNA, 7SL, cytoplasmic 344, pseudogene', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169166, CurrentID=0, Status='live', Symbol='SNX31', Aliases='', description='sorting nexin 31', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100572882, end_position_on_the_genomic_accession=100663410, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  399565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399565, CurrentID=0, Status='live', Symbol='GLC3C', Aliases='', description='glaucoma 3, primary congenital, C', other_designations='', map_location='14q24.3-q31.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613085, no_hdr0=''),
  132204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132204, CurrentID=0, Status='live', Symbol='SYNPR', Aliases='SPO', description='synaptoporin', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63200409, end_position_on_the_genomic_accession=63616924, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  106480505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480505, CurrentID=0, Status='live', Symbol='RN7SL379P', Aliases='', description='RNA, 7SL, cytoplasmic 379, pseudogene', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480506, CurrentID=0, Status='live', Symbol='RN7SL392P', Aliases='', description='RNA, 7SL, cytoplasmic 392, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480507, CurrentID=0, Status='live', Symbol='RN7SL401P', Aliases='', description='RNA, 7SL, cytoplasmic 401, pseudogene', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480508, CurrentID=0, Status='live', Symbol='RN7SL418P', Aliases='', description='RNA, 7SL, cytoplasmic 418, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480509, CurrentID=0, Status='live', Symbol='RN7SL434P', Aliases='', description='RNA, 7SL, cytoplasmic 434, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169200, CurrentID=0, Status='live', Symbol='TMEM64', Aliases='', description='transmembrane protein 64', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90621995, end_position_on_the_genomic_accession=90645905, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480510, CurrentID=0, Status='live', Symbol='RN7SL443P', Aliases='', description='RNA, 7SL, cytoplasmic 443, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480511, CurrentID=0, Status='live', Symbol='RN7SL456P', Aliases='', description='RNA, 7SL, cytoplasmic 456, pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480512, CurrentID=0, Status='live', Symbol='RN7SL469P', Aliases='', description='RNA, 7SL, cytoplasmic 469, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874063, CurrentID=0, Status='live', Symbol='CLYBL-AS2', Aliases='', description='CLYBL antisense RNA 2', other_designations='CLYBL antisense RNA 2 (non-protein coding)', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99690098, end_position_on_the_genomic_accession=99690971, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189057, CurrentID=0, Status='live', Symbol='TRY-GTA9-1', Aliases='TRNAY5', description='transfer RNA-Tyr (GTA) 9-1', other_designations='transfer RNA tyrosine 5 (anticodon GUA)', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480514, CurrentID=0, Status='live', Symbol='RN7SL491P', Aliases='', description='RNA, 7SL, cytoplasmic 491, pseudogene', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442087, CurrentID=0, Status='live', Symbol='LOC442087', Aliases='', description='prostate tumor overexpressed 1 pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120306562, end_position_on_the_genomic_accession=120308138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480515, CurrentID=0, Status='live', Symbol='RN7SL504P', Aliases='', description='RNA, 7SL, cytoplasmic 504, pseudogene', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480516, CurrentID=0, Status='live', Symbol='RN7SL525P', Aliases='', description='RNA, 7SL, cytoplasmic 525, pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419673, CurrentID=0, Status='live', Symbol='LOC100419673', Aliases='', description='PHD finger protein 10 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25349544, end_position_on_the_genomic_accession=25350593, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480517, CurrentID=0, Status='live', Symbol='RN7SL538P', Aliases='', description='RNA, 7SL, cytoplasmic 538, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874064, CurrentID=0, Status='live', Symbol='PCDH9-AS2', Aliases='', description='PCDH9 antisense RNA 2', other_designations='PCDH9 antisense RNA 2 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66825169, end_position_on_the_genomic_accession=66915031, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480518, CurrentID=0, Status='live', Symbol='RN7SL553P', Aliases='', description='RNA, 7SL, cytoplasmic 553, pseudogene', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480519, CurrentID=0, Status='live', Symbol='RN7SL584P', Aliases='', description='RNA, 7SL, cytoplasmic 584, pseudogene', other_designations='', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480520, CurrentID=0, Status='live', Symbol='RN7SL601P', Aliases='', description='RNA, 7SL, cytoplasmic 601, pseudogene', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533632, CurrentID=0, Status='live', Symbol='LOC100533632', Aliases='', description='zinc finger protein 736 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56649606, end_position_on_the_genomic_accession=56650698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  169270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169270, CurrentID=0, Status='live', Symbol='ZNF596', Aliases='', description='zinc finger protein 596', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=232137, end_position_on_the_genomic_accession=247340, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106480521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480521, CurrentID=0, Status='live', Symbol='RN7SL608P', Aliases='', description='RNA, 7SL, cytoplasmic 608, pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480522, CurrentID=0, Status='live', Symbol='RN7SL623P', Aliases='', description='RNA, 7SL, cytoplasmic 623, pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874065, CurrentID=0, Status='live', Symbol='PCDH9-AS1', Aliases='', description='PCDH9 antisense RNA 1', other_designations='PCDH9 antisense RNA 1 (non-protein coding)', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189067, CurrentID=0, Status='live', Symbol='TRL-TAA4-1', Aliases='TRNAL10', description='transfer RNA-Leu (TAA) 4-1', other_designations='tRNA-Leu-TAA-4-1|transfer RNA leucine 10 (anticodon UAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27230555, end_position_on_the_genomic_accession=27230637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480524, CurrentID=0, Status='live', Symbol='RN7SL658P', Aliases='', description='RNA, 7SL, cytoplasmic 658, pseudogene', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480525, CurrentID=0, Status='live', Symbol='RN7SL673P', Aliases='', description='RNA, 7SL, cytoplasmic 673, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533633, CurrentID=0, Status='live', Symbol='LOC100533633', Aliases='', description='zinc finger protein 257 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56685036, end_position_on_the_genomic_accession=56686196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289220, CurrentID=0, Status='live', Symbol='LTV1P1', Aliases='', description='LTV1 ribosome biogenesis factor pseudogene 1', other_designations='LTV1 homolog (S. cerevisiae) pseudogene 1', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120091049, end_position_on_the_genomic_accession=120092261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480526, CurrentID=0, Status='live', Symbol='RN7SL697P', Aliases='', description='RNA, 7SL, cytoplasmic 697, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420230, CurrentID=0, Status='live', Symbol='LOC100420230', Aliases='', description='MAGE family member C3 pseudogene', other_designations='melanoma antigen family C, 3 pseudogene|melanoma antigen family C3 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142205174, end_position_on_the_genomic_accession=142205749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480527, CurrentID=0, Status='live', Symbol='RN7SL714P', Aliases='', description='RNA, 7SL, cytoplasmic 714, pseudogene', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874066, CurrentID=0, Status='live', Symbol='LATS2-AS1', Aliases='', description='LATS2 antisense RNA 1', other_designations='LATS2 antisense RNA 1 (non-protein coding)', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284967, CurrentID=0, Status='live', Symbol='bA395L14.13', Aliases='', description='acrosin-like protease', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480529, CurrentID=0, Status='live', Symbol='RN7SL735P', Aliases='', description='RNA, 7SL, cytoplasmic 735, pseudogene', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480530, CurrentID=0, Status='live', Symbol='RN7SL757P', Aliases='', description='RNA, 7SL, cytoplasmic 757, pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533634, CurrentID=0, Status='live', Symbol='LOC100533634', Aliases='', description='zinc finger protein 737 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65355953, end_position_on_the_genomic_accession=65357289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369564, CurrentID=0, Status='live', Symbol='LOC105369564', Aliases='', description='uncharacterized LOC105369564', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128208757, end_position_on_the_genomic_accession=128241441, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480531, CurrentID=0, Status='live', Symbol='RN7SL764P', Aliases='', description='RNA, 7SL, cytoplasmic 764, pseudogene', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480533, CurrentID=0, Status='live', Symbol='RN7SL798P', Aliases='', description='RNA, 7SL, cytoplasmic 798, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480534, CurrentID=0, Status='live', Symbol='RN7SL813P', Aliases='', description='RNA, 7SL, cytoplasmic 813, pseudogene', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169355, CurrentID=0, Status='live', Symbol='IDO2', Aliases='INDOL1', description='indoleamine 2,3-dioxygenase 2', other_designations='IDO-2|indoleamine 2,3-dioxygenase-like 1 protein|indoleamine 2,3-dioxygenase-like protein 1|indoleamine-pyrrole 2,3 dioxygenase-like 1|indoleamine-pyrrole 2,3-dioxygenase-like protein 1', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39934955, end_position_on_the_genomic_accession=40016391, orientation='plus', exon_count=11, OMIM=612129, no_hdr0=''),
  100533635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533635, CurrentID=0, Status='live', Symbol='LOC100533635', Aliases='', description='zinc finger protein 384 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132086206, end_position_on_the_genomic_accession=132089319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480536, CurrentID=0, Status='live', Symbol='RN7SL837P', Aliases='', description='RNA, 7SL, cytoplasmic 837, pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480537, CurrentID=0, Status='live', Symbol='RN7SL851P', Aliases='', description='RNA, 7SL, cytoplasmic 851, pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369524, CurrentID=0, Status='live', Symbol='LOC105369524', Aliases='', description='uncharacterized LOC105369524', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119512230, end_position_on_the_genomic_accession=119517888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480538, CurrentID=0, Status='live', Symbol='RN7SL863P', Aliases='', description='RNA, 7SL, cytoplasmic 863, pseudogene', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23098, CurrentID=0, Status='live', Symbol='SARM1', Aliases='MyD88-5, SAMD2, SARM', description='sterile alpha and TIR motif containing 1', other_designations='SAM domain-containing protein 2|sterile alpha and Armadillo repeat protein|sterile alpha and HEAT/Armadillo motif protein, ortholog of Drosophila|sterile alpha motif domain-containing protein 2|tir-1 homolog', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28371662, end_position_on_the_genomic_accession=28401045, orientation='plus', exon_count=9, OMIM=607732, no_hdr0=''),
  102724625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724625, CurrentID=0, Status='live', Symbol='LOC102724625', Aliases='', description='uncharacterized LOC102724625', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32079345, end_position_on_the_genomic_accession=32095129, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480540, CurrentID=0, Status='live', Symbol='RNU6ATAC10P', Aliases='', description='RNA, U6atac small nuclear 10, pseudogene', other_designations='', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533636, CurrentID=0, Status='live', Symbol='LOC100533636', Aliases='', description='zinc finger protein 519 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101720482, end_position_on_the_genomic_accession=101721499, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189085, CurrentID=0, Status='live', Symbol='TRM-CAT7-1', Aliases='TRNAM5', description='transfer RNA-Met (CAT) 7-1', other_designations='tRNA-Met-CAT-7-1|transfer RNA methionine 5 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57842214, end_position_on_the_genomic_accession=57842286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100380878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380878, CurrentID=0, Status='live', Symbol='RNY5P8', Aliases='', description='RNA, Ro-associated Y5 pseudogene 8', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102819999, end_position_on_the_genomic_accession=102820066, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480542, CurrentID=0, Status='live', Symbol='RNU6ATAC26P', Aliases='', description='RNA, U6atac small nuclear 26, pseudogene', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874069, CurrentID=0, Status='live', Symbol='STK24-AS1', Aliases='', description='STK24 antisense RNA 1', other_designations='STK24 antisense RNA 1 (non-protein coding)', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98577198, end_position_on_the_genomic_accession=98578842, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420159, CurrentID=0, Status='live', Symbol='RRM2P3', Aliases='', description='ribonucleotide reductase M2 polypeptide pseudogene 3', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44307370, end_position_on_the_genomic_accession=44310536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480543, CurrentID=0, Status='live', Symbol='RNU6ATAC36P', Aliases='', description='RNA, U6atac small nuclear 36, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480544, CurrentID=0, Status='live', Symbol='RNU11-6P', Aliases='', description='RNA, U11 small nuclear 6, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480545, CurrentID=0, Status='live', Symbol='RNU4-18P', Aliases='', description='RNA, U4 small nuclear 18, pseudogene', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533637, CurrentID=0, Status='live', Symbol='LOC100533637', Aliases='', description='zinc finger protein 83 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15325231, end_position_on_the_genomic_accession=15326446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480546, CurrentID=0, Status='live', Symbol='RNU4-31P', Aliases='', description='RNA, U4 small nuclear 31, pseudogene', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189091, CurrentID=0, Status='live', Symbol='TRE-CTC1-2', Aliases='TRNAE13', description='transfer RNA-Glu (CTC) 1-2', other_designations='tRNA-Glu-CTC-1-2|transfer RNA glutamic acid 13 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161447228, end_position_on_the_genomic_accession=161447299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480548, CurrentID=0, Status='live', Symbol='RNU4-70P', Aliases='', description='RNA, U4 small nuclear 70, pseudogene', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169436, CurrentID=0, Status='live', Symbol='STKLD1', Aliases='C9orf96, SgK071, Sk521', description='serine/threonine kinase-like domain containing 1', other_designations='probable inactive protein kinase-like protein SgK071|serine/threonine kinase-like domain-containing protein 1|sugen kinase 071', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133376413, end_position_on_the_genomic_accession=133406096, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  106480549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480549, CurrentID=0, Status='live', Symbol='RNU4-76P', Aliases='', description='RNA, U4 small nuclear 76, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724627, CurrentID=0, Status='live', Symbol='LOC102724627', Aliases='', description='uncharacterized LOC102724627', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117734934, end_position_on_the_genomic_accession=117753377, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480550, CurrentID=0, Status='live', Symbol='RNU6-89P', Aliases='', description='RNA, U6 small nuclear 89, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533638, CurrentID=0, Status='live', Symbol='LOC100533638', Aliases='', description='zinc finger protein 85 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20695212, end_position_on_the_genomic_accession=20697232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480551, CurrentID=0, Status='live', Symbol='RNU6-97P', Aliases='', description='RNA, U6 small nuclear 97, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480552, CurrentID=0, Status='live', Symbol='RNU6-103P', Aliases='', description='RNA, U6 small nuclear 103, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874071, CurrentID=0, Status='live', Symbol='USP12-AS2', Aliases='', description='USP12 antisense RNA 2 (head to head)', other_designations='USP12 antisense RNA 2 (non-protein coding)', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27172259, end_position_on_the_genomic_accession=27183503, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480553, CurrentID=0, Status='live', Symbol='RNU6-116P', Aliases='', description='RNA, U6 small nuclear 116, pseudogene', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480554, CurrentID=0, Status='live', Symbol='RNU6-127P', Aliases='', description='RNA, U6 small nuclear 127, pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480555, CurrentID=0, Status='live', Symbol='RNU6-141P', Aliases='', description='RNA, U6 small nuclear 141, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480556, CurrentID=0, Status='live', Symbol='RNU6-147P', Aliases='', description='RNA, U6 small nuclear 147, pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421977, CurrentID=0, Status='live', Symbol='LOC100421977', Aliases='', description='olfactory receptor, family 6, subfamily M, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480557, CurrentID=0, Status='live', Symbol='RNU6-159P', Aliases='', description='RNA, U6 small nuclear 159, pseudogene', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130429, CurrentID=0, Status='live', Symbol='LOC100130429', Aliases='', description='uncharacterized LOC100130429', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=46732988, end_position_on_the_genomic_accession=46733365, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100420162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420162, CurrentID=0, Status='live', Symbol='LOC100420162', Aliases='', description='nuclear RNA export factor 5 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102549478, end_position_on_the_genomic_accession=102550104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189102, CurrentID=0, Status='live', Symbol='TRY-GTA5-5', Aliases='TRNAY6', description='transfer RNA-Tyr (GTA) 5-5', other_designations='transfer RNA tyrosine 6 (anticodon GUA)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66113988, end_position_on_the_genomic_accession=66114076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189103, CurrentID=0, Status='live', Symbol='TRR-ACG2-1', Aliases='TRNAR11', description='transfer RNA-Arg (ACG) 2-1', other_designations='tRNA-Arg-ACG-2-1|transfer RNA arginine 11 (anticodon ACG)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45688999, end_position_on_the_genomic_accession=45689071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  126826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126826, CurrentID=0, Status='live', Symbol='TGIF2P1', Aliases='', description='TGFB-induced factor homeobox 2 pseudogene 1', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244394975, end_position_on_the_genomic_accession=244407166, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642864, CurrentID=0, Status='live', Symbol='SPATA42', Aliases='AKNAD1-AS1, SRG7', description='spermatogenesis associated 42 (non-protein coding)', other_designations='AKNAD1 antisense RNA 1|spermatogenesis-related gene 7|spermatogenesis-related protein 7', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108857217, end_position_on_the_genomic_accession=108858524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480560, CurrentID=0, Status='live', Symbol='RNU6-193P', Aliases='', description='RNA, U6 small nuclear 193, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533640, CurrentID=0, Status='live', Symbol='LOC100533640', Aliases='', description='WEE1 homolog (S. pombe) pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22693319, end_position_on_the_genomic_accession=22694787, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480561, CurrentID=0, Status='live', Symbol='RNU6-221P', Aliases='', description='RNA, U6 small nuclear 221, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480562, CurrentID=0, Status='live', Symbol='RNU6-233P', Aliases='', description='RNA, U6 small nuclear 233, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874073, CurrentID=0, Status='live', Symbol='FREM2-AS1', Aliases='', description='FREM2 antisense RNA 1', other_designations='FREM2 antisense RNA 1 (non-protein coding)', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38821847, end_position_on_the_genomic_accession=38830835, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369529, CurrentID=0, Status='live', Symbol='LOC105369529', Aliases='', description='uncharacterized LOC105369529', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120128502, end_position_on_the_genomic_accession=120139424, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  169522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169522, CurrentID=0, Status='live', Symbol='KCNV2', Aliases='KV11.1, Kv8.2, RCD3B', description='potassium channel, voltage gated modifier subfamily V, member 2', other_designations='potassium channel, subfamily V, member 2|voltage-gated potassium channel subunit Kv8.2', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2717239, end_position_on_the_genomic_accession=2730037, orientation='plus', exon_count=4, OMIM=607604, no_hdr0=''),
  106480563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480563, CurrentID=0, Status='live', Symbol='RNU6-239P', Aliases='', description='RNA, U6 small nuclear 239, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  677781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677781, CurrentID=0, Status='live', Symbol='SCARNA16', Aliases='ACA47', description='small Cajal body-specific RNA 16', other_designations='ACA47 scaRNA', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77089307, end_position_on_the_genomic_accession=77089493, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480564, CurrentID=0, Status='live', Symbol='RNU6-252P', Aliases='', description='RNA, U6 small nuclear 252, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442097, CurrentID=0, Status='live', Symbol='ACTG1P23', Aliases='', description='actin gamma 1 pseudogene 23', other_designations='actin, gamma 1 pseudogene', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175977486, end_position_on_the_genomic_accession=175978575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480565, CurrentID=0, Status='live', Symbol='RNU6-267P', Aliases='', description='RNA, U6 small nuclear 267, pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533641, CurrentID=0, Status='live', Symbol='LOC100533641', Aliases='', description='zinc finger protein 100 pseudogene', other_designations='zinc finger protein 273 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57966959, end_position_on_the_genomic_accession=57968641, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480566, CurrentID=0, Status='live', Symbol='RNU6-273P', Aliases='', description='RNA, U6 small nuclear 273, pseudogene', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723169, CurrentID=0, Status='live', Symbol='IGHV1-69D', Aliases='', description='immunoglobulin heavy variable 1-69D', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270966, CurrentID=0, Status='live', Symbol='RPL26P25', Aliases='RPL26_6_894', description='ribosomal protein L26 pseudogene 25', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36987930, end_position_on_the_genomic_accession=36988498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189111, CurrentID=0, Status='live', Symbol='TRM-CAT3-2', Aliases='TRNAM8', description='transfer RNA-Met (CAT) 3-2', other_designations='tRNA-Met-CAT-3-2|transfer RNA methionine 8 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28953265, end_position_on_the_genomic_accession=28953337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874074, CurrentID=0, Status='live', Symbol='DLEU7-AS1', Aliases='', description='DLEU7 antisense RNA 1', other_designations='DLEU7 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50807855, end_position_on_the_genomic_accession=50849905, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480568, CurrentID=0, Status='live', Symbol='RNU6-308P', Aliases='', description='RNA, U6 small nuclear 308, pseudogene', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480569, CurrentID=0, Status='live', Symbol='RNU6-313P', Aliases='', description='RNA, U6 small nuclear 313, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  431704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431704, CurrentID=0, Status='live', Symbol='RGS21', Aliases='', description='regulator of G-protein signaling 21', other_designations='regulator of G-protein signalling 21', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192316992, end_position_on_the_genomic_accession=192367285, orientation='plus', exon_count=5, OMIM=612407, no_hdr0=''),
  431705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431705, CurrentID=0, Status='live', Symbol='ASTL', Aliases='SAS1B', description='astacin-like metallo-endopeptidase (M12 family)', other_designations='astacin-like metalloendopeptidase (M12 family)|oocyte astacin|ovastacin|sperm acrosomal SLLP1 binding', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96122817, end_position_on_the_genomic_accession=96143077, orientation='minus', exon_count=9, OMIM=608860, no_hdr0=''),
  431707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431707, CurrentID=0, Status='live', Symbol='LHX8', Aliases='LHX7', description='LIM homeobox 8', other_designations='LIM-homeodomain protein Lhx8', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75128434, end_position_on_the_genomic_accession=75161533, orientation='plus', exon_count=12, OMIM=604425, no_hdr0=''),
  431708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431708, CurrentID=0, Status='live', Symbol='MDD1', Aliases='', description='Major depressive disorder', other_designations='', map_location='12q22-q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608520, no_hdr0=''),
  431709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431709, CurrentID=0, Status='live', Symbol='MDD2', Aliases='', description='Major depressive disorder 2', other_designations='', map_location='15q25.3-q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608691, no_hdr0=''),
  431710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431710, CurrentID=0, Status='live', Symbol='ASPG1', Aliases='', description='Asperger syndrome, susceptibility to, 1', other_designations='', map_location='3q25-q27', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608638, no_hdr0=''),
  431711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431711, CurrentID=0, Status='live', Symbol='ASPG2', Aliases='', description='Asperger syndrome, susceptibility to, 2', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608631, no_hdr0=''),
  431712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=431712, CurrentID=0, Status='live', Symbol='CMT2G', Aliases='', description='Charcot-Marie-Tooth disease, axonal, type 2G', other_designations='', map_location='12q12-q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608591, no_hdr0=''),
  100189115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189115, CurrentID=0, Status='live', Symbol='TRR-TCT3-1', Aliases='TRNAR14', description='transfer RNA-Arg (TCT) 3-1', other_designations='transfer RNA arginine 14 (anticodon UCU)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59551294, end_position_on_the_genomic_accession=59551379, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480572, CurrentID=0, Status='live', Symbol='RNU6-344P', Aliases='', description='RNA, U6 small nuclear 344, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874075, CurrentID=0, Status='live', Symbol='ARAP1-AS1', Aliases='', description='ARAP1 antisense RNA 1', other_designations='ARAP1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369531, CurrentID=0, Status='live', Symbol='LOC105369531', Aliases='', description='uncharacterized LOC105369531', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120274115, end_position_on_the_genomic_accession=120291109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480573, CurrentID=0, Status='live', Symbol='RNU6-352P', Aliases='', description='RNA, U6 small nuclear 352, pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928621, CurrentID=0, Status='live', Symbol='LOC101928621', Aliases='', description='uncharacterized LOC101928621', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1323178, end_position_on_the_genomic_accession=1327620, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480574, CurrentID=0, Status='live', Symbol='RNU6-361P', Aliases='', description='RNA, U6 small nuclear 361, pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6420, CurrentID=0, Status='live', Symbol='SF', Aliases='', description='Stoltzfus blood group', other_designations='', map_location='4q28-q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=111800, no_hdr0=''),
  106480575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480575, CurrentID=0, Status='live', Symbol='RNU6-375P', Aliases='', description='RNA, U6 small nuclear 375, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533643, CurrentID=0, Status='live', Symbol='LOC100533643', Aliases='', description='zinc finger protein 91 pseudogene', other_designations='zinc finger protein 273 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63469376, end_position_on_the_genomic_accession=63471075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480576, CurrentID=0, Status='live', Symbol='RNU6-381P', Aliases='', description='RNA, U6 small nuclear 381, pseudogene', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376549, CurrentID=0, Status='live', Symbol='LOC105376549', Aliases='', description='activating transcription factor 7-interacting protein 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11137078, end_position_on_the_genomic_accession=11137933, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480577, CurrentID=0, Status='live', Symbol='RNU6-392P', Aliases='PBEF', description='RNA, U6 small nuclear 392, pseudogene', other_designations='H_NH0022N19_gi18089094.sp_cds.1', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169611, CurrentID=0, Status='live', Symbol='OLFML2A', Aliases='PRO34319', description='olfactomedin like 2A', other_designations='olfactomedin-like 2A|photomedin-1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124777158, end_position_on_the_genomic_accession=124814891, orientation='plus', exon_count=11, OMIM=615899, no_hdr0=''),
  106480578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480578, CurrentID=0, Status='live', Symbol='RNU6-403P', Aliases='', description='RNA, U6 small nuclear 403, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480579, CurrentID=0, Status='live', Symbol='RNU6-416P', Aliases='', description='RNA, U6 small nuclear 416, pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480580, CurrentID=0, Status='live', Symbol='RNU6-422P', Aliases='', description='RNA, U6 small nuclear 422, pseudogene', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533644, CurrentID=0, Status='live', Symbol='LOC100533644', Aliases='', description='zinc finger protein 628 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134171609, end_position_on_the_genomic_accession=134173684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480581, CurrentID=0, Status='live', Symbol='RNU6-434P', Aliases='', description='RNA, U6 small nuclear 434, pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574406, CurrentID=0, Status='live', Symbol='ADAMTSL4-AS1', Aliases='C1orf138', description='ADAMTSL4 antisense RNA 1', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150560895, end_position_on_the_genomic_accession=150574552, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874077, CurrentID=0, Status='live', Symbol='ATP1B3-AS1', Aliases='', description='ATP1B3 antisense RNA 1', other_designations='ATP1B3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130430, CurrentID=0, Status='live', Symbol='LOC100130430', Aliases='', description='uncharacterized LOC100130430', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=1298646, end_position_on_the_genomic_accession=1299415, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574407, CurrentID=0, Status='live', Symbol='C1orf145', Aliases='', description='chromosome 1 open reading frame 145', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228203506, end_position_on_the_genomic_accession=228213664, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100189128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189128, CurrentID=0, Status='live', Symbol='TRA-AGC10-1', Aliases='TRNAA18', description='transfer RNA-Ala (AGC) 10-1', other_designations='tRNA-Ala-AGC-10-1|transfer RNA alanine 18 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26687257, end_position_on_the_genomic_accession=26687329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189129, CurrentID=0, Status='live', Symbol='TRE-TTC11-1', Aliases='TRNAE15', description='transfer RNA-Glu (TTC) 11-1', other_designations='tRNA-Glu-TTC-11-1|transfer RNA glutamic acid 15 (anticodon UUC)', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189130, CurrentID=0, Status='live', Symbol='TRL-CAG2-1', Aliases='TRNAL13', description='transfer RNA-Leu (CAG) 2-1', other_designations='tRNA-Leu-CAG-2-1|transfer RNA leucine 13 (anticodon CAG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57299951, end_position_on_the_genomic_accession=57300033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480587, CurrentID=0, Status='live', Symbol='RNU6-527P', Aliases='', description='RNA, U6 small nuclear 527, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105694156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105694156, CurrentID=0, Status='live', Symbol='ERVW-28', Aliases='', description='endogenous retrovirus group W member 28', other_designations='endogenous retrovirus group W, member 28', map_location='17q12b', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480589, CurrentID=0, Status='live', Symbol='RNU6-553P', Aliases='', description='RNA, U6 small nuclear 553, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189134, CurrentID=0, Status='live', Symbol='TRD-GTC2-4', Aliases='TRNAD6', description='transfer RNA-Asp (GTC) 2-4', other_designations='tRNA-Asp-GTC-2-4|transfer RNA aspartic acid 6 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161463034, end_position_on_the_genomic_accession=161463105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480591, CurrentID=0, Status='live', Symbol='RNU6-573P', Aliases='', description='RNA, U6 small nuclear 573, pseudogene', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169693, CurrentID=0, Status='live', Symbol='TMEM252', Aliases='C9orf71', description='transmembrane protein 252', other_designations='transmembrane protein C9orf71', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68530662, end_position_on_the_genomic_accession=68540886, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480592, CurrentID=0, Status='live', Symbol='RNU6-586P', Aliases='', description='RNA, U6 small nuclear 586, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369535, CurrentID=0, Status='live', Symbol='LOC105369535', Aliases='', description='uncharacterized LOC105369535', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121478813, end_position_on_the_genomic_accession=121482907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480593, CurrentID=0, Status='live', Symbol='RNU6-590P', Aliases='', description='RNA, U6 small nuclear 590, pseudogene', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480594, CurrentID=0, Status='live', Symbol='RNU6-603P', Aliases='', description='RNA, U6 small nuclear 603, pseudogene', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399656, CurrentID=0, Status='live', Symbol='FAM32BP', Aliases='FAM32B', description='family with sequence similarity 32 member B, pseudogene', other_designations='family with sequence similarity 32 member B (pseudogene)', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17213467, end_position_on_the_genomic_accession=17215248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  169714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169714, CurrentID=0, Status='live', Symbol='QSOX2', Aliases='QSCN6L1, SOXN', description='quiescin sulfhydryl oxidase 2', other_designations='neuroblastoma-derived sulfhydryl oxidase|quiescin Q6 sulfhydryl oxidase 2|quiescin Q6-like 1|quiescin Q6-like protein 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136206333, end_position_on_the_genomic_accession=136245841, orientation='minus', exon_count=12, OMIM=612860, no_hdr0=''),
  106480595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480595, CurrentID=0, Status='live', Symbol='RNU6-615P', Aliases='', description='RNA, U6 small nuclear 615, pseudogene', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480596, CurrentID=0, Status='live', Symbol='RNU6-627P', Aliases='', description='RNA, U6 small nuclear 627, pseudogene', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480597, CurrentID=0, Status='live', Symbol='RNU6-636P', Aliases='', description='RNA, U6 small nuclear 636, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874080, CurrentID=0, Status='live', Symbol='FARP1-AS1', Aliases='', description='FARP1 antisense RNA 1', other_designations='FARP1 antisense RNA 1 (non-protein coding)', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369536, CurrentID=0, Status='live', Symbol='LOC105369536', Aliases='', description='uncharacterized LOC105369536', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121439082, end_position_on_the_genomic_accession=121453013, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480598, CurrentID=0, Status='live', Symbol='RNU6-647P', Aliases='', description='RNA, U6 small nuclear 647, pseudogene', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480599, CurrentID=0, Status='live', Symbol='RNU6-660P', Aliases='', description='RNA, U6 small nuclear 660, pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480600, CurrentID=0, Status='live', Symbol='RNU6-666P', Aliases='', description='RNA, U6 small nuclear 666, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533648, CurrentID=0, Status='live', Symbol='LOC100533648', Aliases='', description='uncharacterized LOC100533648', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56371936, end_position_on_the_genomic_accession=56373049, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480601, CurrentID=0, Status='live', Symbol='RNU6-680P', Aliases='', description='RNA, U6 small nuclear 680, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  346085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346085, CurrentID=0, Status='live', Symbol='GAPDHP72', Aliases='GAPDHL18, GAPDL18', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 72', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 18, pseudogene', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166064220, end_position_on_the_genomic_accession=166065502, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480602, CurrentID=0, Status='live', Symbol='RNU6-708P', Aliases='', description='RNA, U6 small nuclear 708, pseudogene', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399664, CurrentID=0, Status='live', Symbol='MEX3D', Aliases='MEX-3D, MEX3, OK/SW-cl.4, RKHD1, RNF193, TINO', description='mex-3 RNA binding family member D', other_designations='RING finger and KH domain-containing protein 1|RING finger protein 193|bcl-2 ARE RNA binding protein|ring finger (C3HC4 type) and KH domain containing 1|ring finger and KH domain containing 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1554669, end_position_on_the_genomic_accession=1568058, orientation='minus', exon_count=3, OMIM=611009, no_hdr0=''),
  100420171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420171, CurrentID=0, Status='live', Symbol='LOC100420171', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene', other_designations='fatty acyl CoA reductase 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19666309, end_position_on_the_genomic_accession=19683669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399665, CurrentID=0, Status='live', Symbol='FAM102A', Aliases='C9orf132, EEIG1, SYM-3A, bA203J24.7', description='family with sequence similarity 102 member A', other_designations='C230093N12Rik|early estrogen-induced gene 1 protein|family with sequence similarity 102, member A|sym-3 homolog A', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127940582, end_position_on_the_genomic_accession=127981027, orientation='minus', exon_count=14, OMIM=610891, no_hdr0=''),
  106480604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480604, CurrentID=0, Status='live', Symbol='RNU6-726P', Aliases='', description='RNA, U6 small nuclear 726, pseudogene', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399666, CurrentID=0, Status='live', Symbol='ARL2BPP1', Aliases='HsT2883', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 1', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 1|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 1|ADP-ribosylation factor-like 2 binding protein pseudogene 1', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76703863, end_position_on_the_genomic_accession=76704254, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480605, CurrentID=0, Status='live', Symbol='RNU6-743P', Aliases='', description='RNA, U6 small nuclear 743, pseudogene', other_designations='', map_location='20q13.11', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533649, CurrentID=0, Status='live', Symbol='LOC100533649', Aliases='', description='uncharacterized LOC100533649', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56374527, end_position_on_the_genomic_accession=56375653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480606, CurrentID=0, Status='live', Symbol='RNU6-749P', Aliases='', description='RNA, U6 small nuclear 749, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399668, CurrentID=0, Status='live', Symbol='SMIM10L2A', Aliases='LED, LINC00086, NCRNA00086', description='small integral membrane protein 10 like 2A', other_designations='Small integral membrane protein 10-like protein 2A|lncRNA activator of enhancer domains|long intergenic non-protein coding RNA 86', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135421943, end_position_on_the_genomic_accession=135426300, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189151, CurrentID=0, Status='live', Symbol='TRL-TAG2-1', Aliases='TRNAL15', description='transfer RNA-Leu (TAG) 2-1', other_designations='tRNA-Leu-TAG-2-1|transfer RNA leucine 15 (anticodon UAG)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20625370, end_position_on_the_genomic_accession=20625451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105682749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105682749, CurrentID=0, Status='live', Symbol='CCEPR', Aliases='CCHE1, lncRNA-CCHE1', description='cervical carcinoma expressed PCNA regulatory lncRNA', other_designations='cervical carcimona expressed PCNA regulatory lncRNA|cervical carcinoma high-expressed lncRNA 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399669, CurrentID=0, Status='live', Symbol='ZNF321P', Aliases='ZNF321', description='zinc finger protein 321, pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52927135, end_position_on_the_genomic_accession=52942594, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  169792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169792, CurrentID=0, Status='live', Symbol='GLIS3', Aliases='ZNF515', description='GLIS family zinc finger 3', other_designations='GLI-similar 3|OK/KNS-cl.4|zinc finger protein 515', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3824127, end_position_on_the_genomic_accession=4310506, orientation='minus', exon_count=19, OMIM=610192, no_hdr0=''),
  574432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574432, CurrentID=0, Status='live', Symbol='IBA57-AS1', Aliases='C1orf148', description='IBA57 antisense RNA 1 (head to head)', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228164086, end_position_on_the_genomic_accession=228165512, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28299, CurrentID=0, Status='live', Symbol='IGKV1-5', Aliases='IGKV, IGKV15, L12, L12a, V1', description='immunoglobulin kappa variable 1-5', other_designations='Ig kappa chain V-region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88947301, end_position_on_the_genomic_accession=88947776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189153, CurrentID=0, Status='live', Symbol='TRI-AAT5-3', Aliases='TRNAI9', description='transfer RNA-Ile (AAT) 5-3', other_designations='tRNA-Ile-AAT-5-3|transfer RNA isoleucine 9 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26554122, end_position_on_the_genomic_accession=26554195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399671, CurrentID=0, Status='live', Symbol='HEATR4', Aliases='', description='HEAT repeat containing 4', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73478484, end_position_on_the_genomic_accession=73558947, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  100189154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189154, CurrentID=0, Status='live', Symbol='TRQ-CTG8-2', Aliases='TRNAQ14', description='transfer RNA-Gln (CTG) 8-2', other_designations='tRNA-Gln-CTG-8-2|transfer RNA glutamine 14 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533650, CurrentID=0, Status='live', Symbol='LOC100533650', Aliases='', description='uncharacterized LOC100533650', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56817438, end_position_on_the_genomic_accession=56818588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189155, CurrentID=0, Status='live', Symbol='TRC-GCA5-1', Aliases='TRNAC12', description='transfer RNA-Cys (GCA) 5-1', other_designations='tRNA-Cys-GCA-5-1|transfer RNA cysteine 12 (anticodon GCA)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79744655, end_position_on_the_genomic_accession=79744727, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6457, CurrentID=0, Status='live', Symbol='SH3GL3', Aliases='CNSA3, EEN-B2, HsT19371, SH3D2C, SH3P13', description='SH3-domain GRB2-like 3', other_designations='SH3 domain protein 2C|SH3 domain-containing GRB2-like protein 3|endophilin-3', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83447123, end_position_on_the_genomic_accession=83618743, orientation='plus', exon_count=17, OMIM=603362, no_hdr0=''),
  100189156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189156, CurrentID=0, Status='live', Symbol='TRA-AGC9-1', Aliases='TRNAA20', description='transfer RNA-Ala (AGC) 9-1', other_designations='tRNA-Ala-AGC-9-1|transfer RNA alanine 20 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57813490, end_position_on_the_genomic_accession=57813562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480613, CurrentID=0, Status='live', Symbol='RNU6-830P', Aliases='', description='RNA, U6 small nuclear 830, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  149684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149684, CurrentID=0, Status='live', Symbol='LOC149684', Aliases='', description='uncharacterized LOC149684', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38260175, end_position_on_the_genomic_accession=38288947, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480614, CurrentID=0, Status='live', Symbol='RNU6-843P', Aliases='', description='RNA, U6 small nuclear 843, pseudogene', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169834, CurrentID=0, Status='live', Symbol='ZNF883', Aliases='', description='zinc finger protein 883', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112997120, end_position_on_the_genomic_accession=113012192, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480615, CurrentID=0, Status='live', Symbol='RNU6-857P', Aliases='', description='RNA, U6 small nuclear 857, pseudogene', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533651, CurrentID=0, Status='live', Symbol='LOC100533651', Aliases='', description='uncharacterized LOC100533651', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56819276, end_position_on_the_genomic_accession=56820426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  169841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169841, CurrentID=0, Status='live', Symbol='ZNF169', Aliases='', description='zinc finger protein 169', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94259266, end_position_on_the_genomic_accession=94303008, orientation='plus', exon_count=6, OMIM=603404, no_hdr0=''),
  100189161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189161, CurrentID=0, Status='live', Symbol='TRV-AAC4-1', Aliases='TRNAV17', description='transfer RNA-Val (AAC) 4-1', other_designations='tRNA-Val-AAC-4-1|transfer RNA valine 17 (anticodon AAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27681106, end_position_on_the_genomic_accession=27681178, orientation='minus', exon_count=0, OMIM=615308, no_hdr0=''),
  100420174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420174, CurrentID=0, Status='live', Symbol='LOC100420174', Aliases='', description='centrosomal protein 170kDa pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68115555, end_position_on_the_genomic_accession=68116115, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480618, CurrentID=0, Status='live', Symbol='RNU6-890P', Aliases='', description='RNA, U6 small nuclear 890, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189163, CurrentID=0, Status='live', Symbol='TRG-GCC2-3', Aliases='TRNAG16', description='transfer RNA-Gly (GCC) 2-3', other_designations='tRNA-Gly-GCC-2-3|transfer RNA glycine 16 (anticodon GCC)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70789507, end_position_on_the_genomic_accession=70789577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996291, CurrentID=0, Status='live', Symbol='LOC100996291', Aliases='', description='uncharacterized LOC100996291', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78261349, end_position_on_the_genomic_accession=78278492, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100189164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189164, CurrentID=0, Status='live', Symbol='TRL-TAA2-1', Aliases='TRNAL16', description='transfer RNA-Leu (TAA) 2-1', other_designations='tRNA-Leu-TAA-2-1|transfer RNA leucine 16 (anticodon UAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27721119, end_position_on_the_genomic_accession=27721201, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533652, CurrentID=0, Status='live', Symbol='LOC100533652', Aliases='', description='uncharacterized LOC100533652', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56803539, end_position_on_the_genomic_accession=56804648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646203, CurrentID=0, Status='live', Symbol='LOC646203', Aliases='', description='MRT4 homolog, ribosome maturation factor pseudogene', other_designations='mRNA turnover 4 homolog (S. cerevisiae) pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13139777, end_position_on_the_genomic_accession=13140489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189165, CurrentID=0, Status='live', Symbol='TRN-GTT16-4', Aliases='TRNAN16', description='transfer RNA-Asn (GTT) 16-4', other_designations='tRNA-Asn-GTT-16-4|transfer RNA asparagine 16 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421483, CurrentID=0, Status='live', Symbol='LOC100421483', Aliases='', description='transforming growth factor, beta receptor 1 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6553753, end_position_on_the_genomic_accession=6554487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480622, CurrentID=0, Status='live', Symbol='RNU6-930P', Aliases='', description='RNA, U6 small nuclear 930, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369541, CurrentID=0, Status='live', Symbol='LOC105369541', Aliases='', description='uncharacterized LOC105369541', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123313674, end_position_on_the_genomic_accession=123314972, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189167, CurrentID=0, Status='live', Symbol='TRR-ACG2-2', Aliases='TRNAR17', description='transfer RNA-Arg (ACG) 2-2', other_designations='tRNA-Arg-ACG-2-2|transfer RNA arginine 17 (anticodon ACG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27213844, end_position_on_the_genomic_accession=27213916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189168, CurrentID=0, Status='live', Symbol='TRR-CCT3-1', Aliases='TRNAR18', description='transfer RNA-Arg (CCT) 3-1', other_designations='tRNA-Arg-CCT-3-1|transfer RNA arginine 18 (anticodon CCU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3152900, end_position_on_the_genomic_accession=3152972, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132228, CurrentID=0, Status='live', Symbol='LSMEM2', Aliases='C3orf45', description='leucine-rich single-pass membrane protein 2', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50277915, end_position_on_the_genomic_accession=50288114, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724642, CurrentID=0, Status='live', Symbol='LOC102724642', Aliases='', description='anaphase-promoting complex subunit 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87125203, end_position_on_the_genomic_accession=87171388, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100189169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189169, CurrentID=0, Status='live', Symbol='TRL-AAG3-1', Aliases='TRNAL17', description='transfer RNA-Leu (AAG) 3-1', other_designations='tRNA-Leu-AAG-3-1|transfer RNA leucine 17 (anticodon AAG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28989002, end_position_on_the_genomic_accession=28989083, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533653, CurrentID=0, Status='live', Symbol='LOC100533653', Aliases='', description='fermitin family member 3 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21592678, end_position_on_the_genomic_accession=21594210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6471, CurrentID=0, Status='live', Symbol='SHMT1P1', Aliases='SHMT1P', description='serine hydroxymethyltransferase 1 (soluble) pseudogene 1', other_designations='', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43849421, end_position_on_the_genomic_accession=43851956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189170, CurrentID=0, Status='live', Symbol='TRF-GAA3-1', Aliases='TRNAF5', description='transfer RNA-Phe (GAA) 3-1', other_designations='tRNA-Phe-GAA-3-1|transfer RNA phenylalanine 5 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28807833, end_position_on_the_genomic_accession=28807905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189171, CurrentID=0, Status='live', Symbol='TRW-CCA4-1', Aliases='TRNAW5', description='transfer RNA-Trp (CCA) 4-1', other_designations='tRNA-Trp-CCA-4-1|transfer RNA tryptophan 5 (anticodon CCA)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98504252, end_position_on_the_genomic_accession=98504323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874086, CurrentID=0, Status='live', Symbol='PCDH9-AS3', Aliases='', description='PCDH9 antisense RNA 3', other_designations='PCDH9 antisense RNA 3 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66977389, end_position_on_the_genomic_accession=66985776, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  399839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399839, CurrentID=0, Status='live', Symbol='DUX4L11', Aliases='', description='double homeobox 4 like 11', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133746640, end_position_on_the_genomic_accession=133747918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189172, CurrentID=0, Status='live', Symbol='TRP-CGG2-1', Aliases='TRNAP10', description='transfer RNA-Pro (CGG) 2-1', other_designations='tRNA-Pro-CGG-2-1|transfer RNA proline 10 (anticodon CGG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27091742, end_position_on_the_genomic_accession=27091813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189173, CurrentID=0, Status='live', Symbol='TRR-CCG1-1', Aliases='TRNAR19', description='transfer RNA-Arg (CCG) 1-1', other_designations='tRNA-Arg-CCG-1-1|transfer RNA arginine 19 (anticodon CCG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3150674, end_position_on_the_genomic_accession=3150746, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189174, CurrentID=0, Status='live', Symbol='TRH-GTG1-4', Aliases='TRNAH7', description='transfer RNA-His (GTG) 1-4', other_designations='tRNA-His-GTG-1-4|transfer RNA histidine 7 (anticodon GUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148302734, end_position_on_the_genomic_accession=148302805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533654, CurrentID=0, Status='live', Symbol='LOC100533654', Aliases='', description='SDA1 domain containing 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1696691, end_position_on_the_genomic_accession=1698044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  574455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574455, CurrentID=0, Status='live', Symbol='MIR193B', Aliases='MIRN193B, mir-193b', description='microRNA 193b', other_designations='hsa-mir-193b', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14303967, end_position_on_the_genomic_accession=14304049, orientation='plus', exon_count=1, OMIM=614734, no_hdr0=''),
  399693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399693, CurrentID=0, Status='live', Symbol='CCDC187', Aliases='', description='coiled-coil domain containing 187', other_designations='uncharacterized MGC50722', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136266538, end_position_on_the_genomic_accession=136306901, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  100533711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533711, CurrentID=0, Status='live', Symbol='LOC100533711', Aliases='', description='cytoskeleton associated protein 2 like pseudogene', other_designations='cytoskeleton associated protein 2-like pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1756756, end_position_on_the_genomic_accession=1758940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189176, CurrentID=0, Status='live', Symbol='TRN-GTT2-5', Aliases='TRNAN17', description='transfer RNA-Asn (GTT) 2-5', other_designations='tRNA-Asn-GTT-2-5|transfer RNA asparagine 17 (anticodon GUU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38751781, end_position_on_the_genomic_accession=38751854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399694, CurrentID=0, Status='live', Symbol='SHC4', Aliases='RaLP, SHCD', description='SHC (Src homology 2 domain containing) family member 4', other_designations='SH2 domain protein C4|SHC (Src homology 2 domain containing) family, member 4|SHC-transforming protein D|hShcD|rai-like protein|src homology 2 domain-containing-transforming protein C4', map_location='15q21.1-q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48823737, end_position_on_the_genomic_accession=48963444, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100189177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189177, CurrentID=0, Status='live', Symbol='TRC-GCA19-1', Aliases='TRNAC13', description='transfer RNA-Cys (GCA) 19-1', other_designations='tRNA-Cys-GCA-19-1|transfer RNA cysteine 13 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149613065, end_position_on_the_genomic_accession=149613136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189178, CurrentID=0, Status='live', Symbol='TRV-AAC6-1', Aliases='TRNAV18', description='transfer RNA-Val (AAC) 6-1', other_designations='tRNA-Val-AAC-6-1|transfer RNA valine 18 (anticodon AAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28735429, end_position_on_the_genomic_accession=28735500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189179, CurrentID=0, Status='live', Symbol='TRC-GCA6-1', Aliases='TRNAC14', description='transfer RNA-Cys (GCA) 6-1', other_designations='tRNA-Cys-GCA-6-1|transfer RNA cysteine 14 (anticodon GCA)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132229100, end_position_on_the_genomic_accession=132229171, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533655, CurrentID=0, Status='live', Symbol='LOC100533655', Aliases='', description='aryl hydrocarbon receptor pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25218709, end_position_on_the_genomic_accession=25221168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189180, CurrentID=0, Status='live', Symbol='TRK-CTT9-1', Aliases='TRNAK18', description='transfer RNA-Lys (CTT) 9-1', other_designations='tRNA-Lys-CTT-9-1|transfer RNA lysine 18 (anticodon CUU)', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  169966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169966, CurrentID=0, Status='live', Symbol='FAM46D', Aliases='CT1.26, CT112', description='family with sequence similarity 46 member D', other_designations='cancer/testis antigen 112|family with sequence similarity 46, member D', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80335504, end_position_on_the_genomic_accession=80446885, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100189181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189181, CurrentID=0, Status='live', Symbol='TRR-CCT2-1', Aliases='TRNAR20', description='transfer RNA-Arg (CCT) 2-1', other_designations='tRNA-Arg-CCT-2-1|transfer RNA arginine 20 (anticodon CCU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75034431, end_position_on_the_genomic_accession=75034503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874088, CurrentID=0, Status='live', Symbol='PCYT1B-AS1', Aliases='', description='PCYT1B antisense RNA 1', other_designations='PCYT1B antisense RNA 1 (non-protein coding)', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24650073, end_position_on_the_genomic_accession=24658237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189182, CurrentID=0, Status='live', Symbol='TRK-TTT6-1', Aliases='TRNAK19', description='transfer RNA-Lys (TTT) 6-1', other_designations='tRNA-Lys-TTT-6-1|transfer RNA lysine 19 (anticodon UUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27334990, end_position_on_the_genomic_accession=27335062, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189183, CurrentID=0, Status='live', Symbol='TRN-GTT13-1', Aliases='TRNAN18', description='transfer RNA-Asn (GTT) 13-1', other_designations='tRNA-Asn-GTT-13-1|transfer RNA asparagine 18 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16875463, end_position_on_the_genomic_accession=16875536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  169981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=169981, CurrentID=0, Status='live', Symbol='SPIN3', Aliases='SPIN-3, bA445O16.1', description='spindlin family member 3', other_designations='bA445O16.1 (DXF34)|spindlin family, member 3|spindlin-like protein 3', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56976370, end_position_on_the_genomic_accession=56995601, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100189184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189184, CurrentID=0, Status='live', Symbol='TRS-AGA2-1', Aliases='TRNAS14', description='transfer RNA-Ser (AGA) 2-1', other_designations='tRNA-Ser-AGA-2-1|transfer RNA serine 14 (anticodon AGA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8226610, end_position_on_the_genomic_accession=8226691, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189185, CurrentID=0, Status='live', Symbol='TRQ-TTG3-2', Aliases='TRNAQ15', description='transfer RNA-Gln (TTG) 3-2', other_designations='tRNA-Gln-TTG-3-2|transfer RNA glutamine 15 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26311747, end_position_on_the_genomic_accession=26311818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189186, CurrentID=0, Status='live', Symbol='TRX-CAT3-1', Aliases='TRNAM10', description='transfer RNA-iMet (CAT) 3-1', other_designations='tRNA-iMet-CAT-3-1|transfer RNA initiator methionine (CAT) 3-1|transfer RNA methionine 10 (anticodon CAU)', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378274, CurrentID=0, Status='live', Symbol='LOC105378274', Aliases='', description='uncharacterized LOC105378274', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43649510, end_position_on_the_genomic_accession=43649790, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189187, CurrentID=0, Status='live', Symbol='TRT-AGT1-1', Aliases='TRNAT8', description='transfer RNA-Thr (AGT) 1-1', other_designations='tRNA-Thr-AGT-1-1|transfer RNA threonine 8 (anticodon AGU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8187160, end_position_on_the_genomic_accession=8187233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189188, CurrentID=0, Status='live', Symbol='TRG-TCC2-4', Aliases='TRNAG17', description='transfer RNA-Gly (TCC) 2-4', other_designations='tRNA-Gly-TCC-2-4|transfer RNA glycine 17 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161462376, end_position_on_the_genomic_accession=161462447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399706, CurrentID=0, Status='live', Symbol='LINC00200', Aliases='C10orf139, NCRNA00200', description='long intergenic non-protein coding RNA 200', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1159768, end_position_on_the_genomic_accession=1164672, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189189, CurrentID=0, Status='live', Symbol='TRQ-CTG1-4', Aliases='TRNAQ16', description='transfer RNA-Gln (CTG) 1-4', other_designations='tRNA-Gln-CTG-1-4|transfer RNA glutamine 16 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27519529, end_position_on_the_genomic_accession=27519600, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189190, CurrentID=0, Status='live', Symbol='TRW-CCA5-1', Aliases='TRNAW6', description='transfer RNA-Trp (CCA) 5-1', other_designations='tRNA-Trp-CCA-5-1|transfer RNA tryptophan 6 (anticodon CCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99469684, end_position_on_the_genomic_accession=99469755, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399708, CurrentID=0, Status='live', Symbol='LINC00701', Aliases='', description='long intergenic non-protein coding RNA 701', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2300319, end_position_on_the_genomic_accession=2315074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100189191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189191, CurrentID=0, Status='live', Symbol='TRV-TAC1-2', Aliases='TRNAV19', description='transfer RNA-Val (TAC) 1-2', other_designations='tRNA-Val-TAC-1-2|transfer RNA valine 19 (anticodon UAC)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18674909, end_position_on_the_genomic_accession=18674981, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189192, CurrentID=0, Status='live', Symbol='TRA-AGC21-1', Aliases='TRNAA21', description='transfer RNA-Ala (AGC) 21-1', other_designations='tRNA-Ala-AGC-21-1|transfer RNA alanine 21 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189193, CurrentID=0, Status='live', Symbol='TRN-GTT9-2', Aliases='TRNAN19', description='transfer RNA-Asn (GTT) 9-2', other_designations='tRNA-Asn-GTT-9-2|transfer RNA asparagine 19 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148048516, end_position_on_the_genomic_accession=148048589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189194, CurrentID=0, Status='live', Symbol='TRC-GCA23-1', Aliases='TRNAC15', description='transfer RNA-Cys (GCA) 23-1', other_designations='tRNA-Cys-GCA-23-1|transfer RNA cysteine 15 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149595214, end_position_on_the_genomic_accession=149595285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533658, CurrentID=0, Status='live', Symbol='HSPA8P17', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 17', other_designations='heat shock 70kDa protein 8 pseudogene 17', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30987928, end_position_on_the_genomic_accession=30990232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189195, CurrentID=0, Status='live', Symbol='TRN-GTT2-6', Aliases='TRNAN20', description='transfer RNA-Asn (GTT) 2-6', other_designations='tRNA-Asn-GTT-2-6|transfer RNA asparagine 20 (anticodon GUU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1383563, end_position_on_the_genomic_accession=1383636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189196, CurrentID=0, Status='live', Symbol='TRV-CAC9-1', Aliases='TRNAV20', description='transfer RNA-Val (CAC) 9-1', other_designations='tRNA-Val-CAC-9-1|transfer RNA valine 20 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27150243, end_position_on_the_genomic_accession=27150315, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170062, CurrentID=0, Status='live', Symbol='FAM47B', Aliases='', description='family with sequence similarity 47 member B', other_designations='family with sequence similarity 47, member B', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=34942796, end_position_on_the_genomic_accession=34944917, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100189197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189197, CurrentID=0, Status='live', Symbol='TRC-GCA2-4', Aliases='TRNAC16', description='transfer RNA-Cys (GCA) 2-4', other_designations='tRNA-Cys-GCA-2-4|transfer RNA cysteine 16 (anticodon GCA)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123508850, end_position_on_the_genomic_accession=123508921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170067, CurrentID=0, Status='live', Symbol='SUPT20HL2', Aliases='FAM48B1, FAM48B2', description='SPT20 homolog, SAGA complex component-like 2', other_designations='Putative protein FAM48B2|family with sequence similarity 48, member B1|family with sequence similarity 48, member B2|suppressor of Ty 20 homolog (S. cerevisiae)-like 2', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24310862, end_position_on_the_genomic_accession=24313315, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100189198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189198, CurrentID=0, Status='live', Symbol='TRK-CTT2-1', Aliases='TRNAK20', description='transfer RNA-Lys (CTT) 2-1', other_designations='tRNA-Lys-CTT-2-1|transfer RNA lysine 20 (anticodon CUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146039401, end_position_on_the_genomic_accession=146039473, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399716, CurrentID=0, Status='live', Symbol='LOC399716', Aliases='DKFZp667F0711', description='uncharacterized LOC399716', other_designations='uncharacterized protein DKFZp667F0711', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6347183, end_position_on_the_genomic_accession=6352762, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480655, CurrentID=0, Status='live', Symbol='RNU6-1279P', Aliases='', description='RNA, U6 small nuclear 1279, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533659, CurrentID=0, Status='live', Symbol='LOC100533659', Aliases='', description='endoplasmic reticulum oxidoreductase alpha pseudogene', other_designations='ERO1-like (S. cerevisiae) pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48171535, end_position_on_the_genomic_accession=48172055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189200, CurrentID=0, Status='live', Symbol='TRD-GTC3-1', Aliases='TRNAD9', description='transfer RNA-Asp (GTC) 3-1', other_designations='tRNA-Asp-GTC-3-1|transfer RNA aspartic acid 9 (anticodon GUC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27583457, end_position_on_the_genomic_accession=27583528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170082, CurrentID=0, Status='live', Symbol='TCEANC', Aliases='', description='transcription elongation factor A (SII) N-terminal and central domain containing', other_designations='TFIIS central domain-containing protein 1|TFS2-M domain-containing protein 1', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13652646, end_position_on_the_genomic_accession=13665409, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100420547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420547, CurrentID=0, Status='live', Symbol='LOC100420547', Aliases='', description='GTF2I repeat domain containing 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66511406, end_position_on_the_genomic_accession=66545672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189201, CurrentID=0, Status='live', Symbol='TRM-CAT1-1', Aliases='TRNAM11', description='transfer RNA-Met (CAT) 1-1', other_designations='tRNA-Met-CAT-1-1|transfer RNA methionine 11 (anticodon CAU)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123157230, end_position_on_the_genomic_accession=123157302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  219525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219525, CurrentID=0, Status='live', Symbol='OR5AK4P', Aliases='PJCG3', description='olfactory receptor family 5 subfamily AK member 4 pseudogene', other_designations='olfactory receptor, family 5, subfamily AK, member 2 pseudogene|olfactory receptor, family 5, subfamily AK, member 4 pseudogene|olfactory-like receptor pseudogene PJCG3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57037534, end_position_on_the_genomic_accession=57038460, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874092, CurrentID=0, Status='live', Symbol='UBE2E1-AS1', Aliases='', description='UBE2E1 antisense RNA 1', other_designations='UBE2E1 antisense RNA 1 (non-protein coding)', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23804024, end_position_on_the_genomic_accession=23806905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189202, CurrentID=0, Status='live', Symbol='TRL-AAG2-4', Aliases='TRNAL20', description='transfer RNA-Leu (AAG) 2-4', other_designations='tRNA-Leu-AAG-2-4|transfer RNA leucine 20 (anticodon AAG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28943622, end_position_on_the_genomic_accession=28943703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189203, CurrentID=0, Status='live', Symbol='TRW-CCA2-1', Aliases='TRNAW7', description='transfer RNA-Trp (CCA) 2-1', other_designations='tRNA-Trp-CCA-2-1|transfer RNA tryptophan 7 (anticodon CCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19508181, end_position_on_the_genomic_accession=19508252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189204, CurrentID=0, Status='live', Symbol='TRL-CAG3-1', Aliases='TRNAL21', description='transfer RNA-Leu (CAG) 3-1', other_designations='tRNA-Leu-CAG-3-1|transfer RNA leucine 21 (anticodon CAG)', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533660, CurrentID=0, Status='live', Symbol='EI24P5', Aliases='', description='etoposide induced 2.4 pseudogene 5', other_designations='etoposide induced 2.4 mRNA pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53194346, end_position_on_the_genomic_accession=53195375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189205, CurrentID=0, Status='live', Symbol='TRS-CGA3-1', Aliases='TRNAS15', description='transfer RNA-Ser (CGA) 3-1', other_designations='tRNA-Ser-CGA-3-1|transfer RNA serine 15 (anticodon CGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27672450, end_position_on_the_genomic_accession=27672531, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189206, CurrentID=0, Status='live', Symbol='TRG-TCC3-1', Aliases='TRNAG18', description='transfer RNA-Gly (TCC) 3-1', other_designations='tRNA-Gly-TCC-3-1|transfer RNA glycine 18 (anticodon UCC)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8221548, end_position_on_the_genomic_accession=8221619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874093, CurrentID=0, Status='live', Symbol='BRWD1-AS1', Aliases='', description='BRWD1 antisense RNA 1', other_designations='BRWD1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39315707, end_position_on_the_genomic_accession=39323218, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101930307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930307, CurrentID=0, Status='live', Symbol='LOC101930307', Aliases='', description='uncharacterized LOC101930307', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137101684, end_position_on_the_genomic_accession=137104460, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189207, CurrentID=0, Status='live', Symbol='TRD-GTC1-1', Aliases='TRNAD10', description='transfer RNA-Asp (GTC) 1-1', other_designations='tRNA-Asp-GTC-1-1|transfer RNA aspartic acid 10 (anticodon GUC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98503503, end_position_on_the_genomic_accession=98503574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100293612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293612, CurrentID=0, Status='live', Symbol='LOC100293612', Aliases='', description='uncharacterized LOC100293612', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13647105, end_position_on_the_genomic_accession=13654178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189208, CurrentID=0, Status='live', Symbol='TRG-GCC2-5', Aliases='TRNAG19', description='transfer RNA-Gly (GCC) 2-5', other_designations='tRNA-Gly-GCC-2-5|transfer RNA glycine 19 (anticodon GCC)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156401147, end_position_on_the_genomic_accession=156401217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399726, CurrentID=0, Status='live', Symbol='CASC10', Aliases='C10orf114, bA418C1.3', description='cancer susceptibility candidate 10', other_designations='cancer susceptibility candidate gene 10 protein', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21494492, end_position_on_the_genomic_accession=21497284, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189209, CurrentID=0, Status='live', Symbol='TRK-CTT5-1', Aliases='TRNAK21', description='transfer RNA-Lys (CTT) 5-1', other_designations='tRNA-Lys-CTT-5-1|transfer RNA lysine 21 (anticodon CUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3180554, end_position_on_the_genomic_accession=3180626, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533661, CurrentID=0, Status='live', Symbol='SGOL1P2', Aliases='', description='shugoshin-like 1 (S. pombe) pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=52896327, end_position_on_the_genomic_accession=52897808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189210, CurrentID=0, Status='live', Symbol='TRE-TTC16-1', Aliases='TRNAE16', description='transfer RNA-Glu (TTC) 16-1', other_designations='tRNA-Glu-TTC-16-1|transfer RNA glutamic acid 16 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189211, CurrentID=0, Status='live', Symbol='TRP-CGG1-1', Aliases='TRNAP11', description='transfer RNA-Pro (CGG) 1-1', other_designations='tRNA-Pro-CGG-1-1|transfer RNA proline 11 (anticodon CGG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167714725, end_position_on_the_genomic_accession=167714796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874094, CurrentID=0, Status='live', Symbol='ZNF197-AS1', Aliases='', description='ZNF197 antisense RNA 1', other_designations='ZNF197 antisense RNA 1 (non-protein coding)', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44617128, end_position_on_the_genomic_accession=44624797, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189212, CurrentID=0, Status='live', Symbol='TRA-AGC18-2', Aliases='TRNAA22', description='transfer RNA-Ala (AGC) 18-2', other_designations='tRNA-Ala-AGC-18-2|transfer RNA alanine 22 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189213, CurrentID=0, Status='live', Symbol='TRI-AAT3-1', Aliases='TRNAI11', description='transfer RNA-Ile (AAT) 3-1', other_designations='tRNA-Ile-AAT-3-1|transfer RNA isoleucine 11 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27275211, end_position_on_the_genomic_accession=27275284, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132126, CurrentID=0, Status='live', Symbol='LOC100132126', Aliases='', description='mitochondrial carrier 2 pseudogene', other_designations='mitochondrial carrier homolog 2 (C. elegans) pseudogene', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92530525, end_position_on_the_genomic_accession=92531433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189214, CurrentID=0, Status='live', Symbol='TRQ-CTG6-1', Aliases='TRNAQ17', description='transfer RNA-Gln (CTG) 6-1', other_designations='tRNA-Gln-CTG-6-1|transfer RNA glutamine 17 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27791356, end_position_on_the_genomic_accession=27791427, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533662, CurrentID=0, Status='live', Symbol='STK25P1', Aliases='', description='serine/threonine kinase 25 pseudogene 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12911217, end_position_on_the_genomic_accession=12912264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189215, CurrentID=0, Status='live', Symbol='NMTRQ-TTG3-1', Aliases='TRNAQ18', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 3-1', other_designations='nmt-tRNA-Gln-TTG-3-1|transfer RNA glutamine 18 (anticodon UUG)', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189216, CurrentID=0, Status='live', Symbol='TRM-CAT3-1', Aliases='TRNAM12', description='transfer RNA-Met (CAT) 3-1', other_designations='tRNA-Met-CAT-3-1|transfer RNA methionine 12 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28944575, end_position_on_the_genomic_accession=28944647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189217, CurrentID=0, Status='live', Symbol='TRQ-CTG8-1', Aliases='TRNAQ19', description='transfer RNA-Gln (CTG) 8-1', other_designations='tRNA-Gln-CTG-8-1|transfer RNA glutamine 19 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284996, CurrentID=0, Status='live', Symbol='RNF149', Aliases='DNAPTP2', description='ring finger protein 149', other_designations='DNA polymerase-transactivated protein 2', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101271219, end_position_on_the_genomic_accession=101308716, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100189218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189218, CurrentID=0, Status='live', Symbol='TRS-GCT6-1', Aliases='TRNAS16', description='transfer RNA-Ser (GCT) 6-1', other_designations='tRNA-Ser-GCT-6-1|transfer RNA serine 16 (anticodon GCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26305490, end_position_on_the_genomic_accession=26305573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724652, CurrentID=0, Status='live', Symbol='LOC102724652', Aliases='', description='alpha-crystallin A chain', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6560656, end_position_on_the_genomic_accession=6564489, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480675, CurrentID=0, Status='live', Symbol='HNRNPA1P76', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 76', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533663, CurrentID=0, Status='live', Symbol='LOC100533663', Aliases='', description='histidine-rich glycoprotein pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186687329, end_position_on_the_genomic_accession=186692506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189220, CurrentID=0, Status='live', Symbol='TRG-GCC5-1', Aliases='TRNAG20', description='transfer RNA-Gly (GCC) 5-1', other_designations='tRNA-Gly-GCC-5-1|transfer RNA glycine 20 (anticodon GCC)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70788694, end_position_on_the_genomic_accession=70788764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189221, CurrentID=0, Status='live', Symbol='TRP-CGG1-2', Aliases='TRNAP12', description='transfer RNA-Pro (CGG) 1-2', other_designations='tRNA-Pro-CGG-1-2|transfer RNA proline 12 (anticodon CGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3172048, end_position_on_the_genomic_accession=3172119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189222, CurrentID=0, Status='live', Symbol='TRT-AGT1-2', Aliases='TRNAT9', description='transfer RNA-Thr (AGT) 1-2', other_designations='tRNA-Thr-AGT-1-2|transfer RNA threonine 9 (anticodon AGU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8226235, end_position_on_the_genomic_accession=8226308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189223, CurrentID=0, Status='live', Symbol='TRP-AGG2-7', Aliases='TRNAP13', description='transfer RNA-Pro (AGG) 2-7', other_designations='tRNA-Pro-AGG-2-7|transfer RNA proline 13 (anticodon AGG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26555270, end_position_on_the_genomic_accession=26555341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642888, CurrentID=0, Status='live', Symbol='CNN2P2', Aliases='', description='calponin 2 pseudogene 2', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40345947, end_position_on_the_genomic_accession=40347085, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  432369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=432369, CurrentID=0, Status='live', Symbol='ATP5EP2', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F1 complex, epsilon subunit pseudogene 2', other_designations='novel protein similar to ATP synthase (H+ transporting, mitochondrial F1 complex, epsilon subunit)', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27945206, end_position_on_the_genomic_accession=27945573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100533664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533664, CurrentID=0, Status='live', Symbol='PTCHD3P2', Aliases='', description='patched domain containing 3 pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169767183, end_position_on_the_genomic_accession=169768749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189225, CurrentID=0, Status='live', Symbol='TRS-GCT4-3', Aliases='TRNAS17', description='transfer RNA-Ser (GCT) 4-3', other_designations='tRNA-Ser-GCT-4-3|transfer RNA serine 17 (anticodon GCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28597340, end_position_on_the_genomic_accession=28597421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189226, CurrentID=0, Status='live', Symbol='TRM-CAT6-1', Aliases='TRNAM13', description='transfer RNA-Met (CAT) 6-1', other_designations='tRNA-Met-CAT-6-1|transfer RNA methionine 13 (anticodon CAU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87384022, end_position_on_the_genomic_accession=87384094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647253, CurrentID=0, Status='live', Symbol='LOC647253', Aliases='', description='ring finger protein 13 pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180770719, end_position_on_the_genomic_accession=180774057, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399744, CurrentID=0, Status='live', Symbol='LINC00999', Aliases='', description='long intergenic non-protein coding RNA 999', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38428146, end_position_on_the_genomic_accession=38452153, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100189227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189227, CurrentID=0, Status='live', Symbol='TRV-CAC1-6', Aliases='TRNAV21', description='transfer RNA-Val (CAC) 1-6', other_designations='tRNA-Val-CAC-1-6|transfer RNA valine 21 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26538054, end_position_on_the_genomic_accession=26538126, orientation='plus', exon_count=0, OMIM=615306, no_hdr0=''),
  284998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284998, CurrentID=0, Status='live', Symbol='LINC01114', Aliases='PANCAT', description='long intergenic non-protein coding RNA 1114', other_designations='pancreatic cancer associated transcript', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104746637, end_position_on_the_genomic_accession=104757719, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100189228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189228, CurrentID=0, Status='live', Symbol='TRV-CAC5-1', Aliases='TRNAV22', description='transfer RNA-Val (CAC) 5-1', other_designations='tRNA-Val-CAC-5-1|transfer RNA valine 22 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149712552, end_position_on_the_genomic_accession=149712624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399746, CurrentID=0, Status='live', Symbol='ACTR3BP5', Aliases='FKSG74', description='ACTR3B pseudogene 5', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 5', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38696596, end_position_on_the_genomic_accession=38698241, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100189229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189229, CurrentID=0, Status='live', Symbol='TRS-GCT4-1', Aliases='TRNAS18', description='transfer RNA-Ser (GCT) 4-1', other_designations='tRNA-Ser-GCT-4-1|transfer RNA serine 18 (anticodon GCU)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40593825, end_position_on_the_genomic_accession=40593906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  432400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=432400, CurrentID=0, Status='live', Symbol='EIG3', Aliases='', description='Epilepsy, idiopathic generalized, susceptibility to 3', other_designations='', map_location='9q32-q33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608762, no_hdr0=''),
  100533665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533665, CurrentID=0, Status='live', Symbol='PTCHD3P3', Aliases='', description='patched domain containing 3 pseudogene 3', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109288571, end_position_on_the_genomic_accession=109290770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170261, CurrentID=0, Status='live', Symbol='ZCCHC12', Aliases='PNMA7A, SIZN, SIZN1', description='zinc finger, CCHC domain containing 12', other_designations='paraneoplastic Ma antigen family member 7A|smad-interacting zinc finger protein 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118823824, end_position_on_the_genomic_accession=118826968, orientation='plus', exon_count=4, OMIM=300701, no_hdr0=''),
  399748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399748, CurrentID=0, Status='live', Symbol='CAP1P2', Aliases='', description='CAP1 pseudogene 2', other_designations='CAP, adenylate cyclase-associated protein 1 pseudogene', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43604846, end_position_on_the_genomic_accession=43606204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189231, CurrentID=0, Status='live', Symbol='TRG-CCC2-2', Aliases='TRNAG22', description='transfer RNA-Gly (CCC) 2-2', other_designations='tRNA-Gly-CCC-2-2|transfer RNA glycine 22 (anticodon CCC)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70248991, end_position_on_the_genomic_accession=70249061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189232, CurrentID=0, Status='live', Symbol='TRA-TGC3-1', Aliases='TRNAA23', description='transfer RNA-Ala (TGC) 3-1', other_designations='tRNA-Ala-TGC-3-1|transfer RNA alanine 23 (anticodon UGC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124921755, end_position_on_the_genomic_accession=124921826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189233, CurrentID=0, Status='live', Symbol='TRN-GTT15-1', Aliases='TRNAN21', description='transfer RNA-Asn (GTT) 15-1', other_designations='tRNA-Asn-GTT-15-1|transfer RNA asparagine 21 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724655, CurrentID=0, Status='live', Symbol='AQP7P5', Aliases='', description='aquaporin 7 pseudogene 5', other_designations='putative aquaporin-7-like protein 3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40481501, end_position_on_the_genomic_accession=40498999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642890, CurrentID=0, Status='live', Symbol='LOC642890', Aliases='', description='N(alpha)-acetyltransferase 50, NatE catalytic subunit pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56205124, end_position_on_the_genomic_accession=56205624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189234, CurrentID=0, Status='live', Symbol='TRK-TTT3-5', Aliases='TRNAK22', description='transfer RNA-Lys (TTT) 3-5', other_designations='tRNA-Lys-TTT-3-5|transfer RNA lysine 22 (anticodon UUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28951029, end_position_on_the_genomic_accession=28951101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373919, CurrentID=0, Status='live', Symbol='LOC105373919', Aliases='', description='uncharacterized LOC105373919', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=228716688, end_position_on_the_genomic_accession=228737394, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533666, CurrentID=0, Status='live', Symbol='SGOL1P1', Aliases='', description='shugoshin-like 1 (S. pombe) pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69606855, end_position_on_the_genomic_accession=69608301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189235, CurrentID=0, Status='live', Symbol='TRD-GTC2-2', Aliases='TRNAD11', description='transfer RNA-Asp (GTC) 2-2', other_designations='tRNA-Asp-GTC-2-2|transfer RNA aspartic acid 11 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161448243, end_position_on_the_genomic_accession=161448314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  154064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154064, CurrentID=0, Status='live', Symbol='RAET1L', Aliases='ULBP6', description='retinoic acid early transcript 1L', other_designations='UL16 binding protein 6', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150018342, end_position_on_the_genomic_accession=150025596, orientation='minus', exon_count=5, OMIM=611047, no_hdr0=''),
  100189236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189236, CurrentID=0, Status='live', Symbol='TRE-CTC5-1', Aliases='TRNAE17', description='transfer RNA-Glu (CTC) 5-1', other_designations='tRNA-Glu-CTC-5-1|transfer RNA glutamic acid 17 (anticodon CUC)', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170302, CurrentID=0, Status='live', Symbol='ARX', Aliases='CT121, EIEE1, ISSX, MRX29, MRX32, MRX33, MRX36, MRX38, MRX43, MRX54, MRX76, MRX87, MRXS1, PRTS', description='aristaless related homeobox', other_designations='aristaless-related homeobox, X-linked|cancer/testis antigen 121', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=25003694, end_position_on_the_genomic_accession=25015948, orientation='minus', exon_count=5, OMIM=300382, no_hdr0=''),
  106480693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480693, CurrentID=0, Status='live', Symbol='RN7SL199P', Aliases='', description='RNA, 7SL, cytoplasmic 199, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189238, CurrentID=0, Status='live', Symbol='TRK-TTT5-1', Aliases='TRNAK23', description='transfer RNA-Lys (TTT) 5-1', other_designations='tRNA-Lys-TTT-5-1|transfer RNA lysine 23 (anticodon UUU)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59556429, end_position_on_the_genomic_accession=59556501, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480695, CurrentID=0, Status='live', Symbol='RN7SL738P', Aliases='', description='RNA, 7SL, cytoplasmic 738, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533667, CurrentID=0, Status='live', Symbol='LOC100533667', Aliases='', description='T, brachyury homolog (mouse) pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=28213369, end_position_on_the_genomic_accession=28214529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480696, CurrentID=0, Status='live', Symbol='RNU4ATAC17P', Aliases='', description='RNA, U4atac small nuclear 17, pseudogene', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480697, CurrentID=0, Status='live', Symbol='RNU6-253P', Aliases='', description='RNA, U6 small nuclear 253, pseudogene', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480698, CurrentID=0, Status='live', Symbol='RNU6-394P', Aliases='', description='RNA, U6 small nuclear 394, pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480699, CurrentID=0, Status='live', Symbol='RNU6-841P', Aliases='', description='RNA, U6 small nuclear 841, pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399761, CurrentID=0, Status='live', Symbol='BMS1P5', Aliases='BMS1LP1, BMS1LP5, BMS1P1, bA556L1.3', description='BMS1, ribosome biogenesis factor pseudogene 5', other_designations='BMS1 pseudogene 1|BMS1 pseudogene 5|BMS1 ribosome biogenesis factor pseudogene 5|BMS1L pseudogene 1|BMS1L pseudogene 5', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46786733, end_position_on_the_genomic_accession=46811989, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106480700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480700, CurrentID=0, Status='live', Symbol='RNU6-1096P', Aliases='', description='RNA, U6 small nuclear 1096, pseudogene', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533668, CurrentID=0, Status='live', Symbol='KLF2P4', Aliases='', description='Kruppel-like factor 2 pseudogene 4', other_designations='Kruppel-like factor 2 (lung) pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131298980, end_position_on_the_genomic_accession=131300444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480701, CurrentID=0, Status='live', Symbol='RNU6-1235P', Aliases='', description='RNA, U6 small nuclear 1235, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480702, CurrentID=0, Status='live', Symbol='RNU2-72P', Aliases='', description='RNA, U2 small nuclear 72, pseudogene', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480703, CurrentID=0, Status='live', Symbol='RNU6-40P', Aliases='RNU6-40', description='RNA, U6 small nuclear 40, pseudogene', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375286, CurrentID=0, Status='live', Symbol='LOC105375286', Aliases='', description='uncharacterized LOC105375286', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55553771, end_position_on_the_genomic_accession=55559862, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480704, CurrentID=0, Status='live', Symbol='RNU6-25P', Aliases='RNU6-25', description='RNA, U6 small nuclear 25, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170370, CurrentID=0, Status='live', Symbol='FAM170B', Aliases='C10orf73', description='family with sequence similarity 170 member B', other_designations='family with sequence similarity 170, member B', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49131154, end_position_on_the_genomic_accession=49134020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170371, CurrentID=0, Status='live', Symbol='C10orf128', Aliases='', description='chromosome 10 open reading frame 128', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49154724, end_position_on_the_genomic_accession=49188585, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106480705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480705, CurrentID=0, Status='live', Symbol='RNU6-17P', Aliases='RNU6-17', description='RNA, U6 small nuclear 17, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533669, CurrentID=0, Status='live', Symbol='LOC100533669', Aliases='', description='MAGE family member A10 pseudogene', other_designations='melanoma antigen family A, 10 pseudogene|melanoma antigen family A10 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142147840, end_position_on_the_genomic_accession=142148842, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480706, CurrentID=0, Status='live', Symbol='RNU6-20P', Aliases='RNU6-20', description='RNA, U6 small nuclear 20, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170384, CurrentID=0, Status='live', Symbol='FUT11', Aliases='', description='fucosyltransferase 11 (alpha (1,3) fucosyltransferase)', other_designations='fuc-TXI|fucT-XI|fucosyltransferase XI|galactoside 3-L-fucosyltransferase 11', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73772175, end_position_on_the_genomic_accession=73779075, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480707, CurrentID=0, Status='live', Symbol='RNU6-24P', Aliases='RNU6-24', description='RNA, U6 small nuclear 24, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170392, CurrentID=0, Status='live', Symbol='OIT3', Aliases='LZP', description='oncoprotein induced transcript 3', other_designations='liver-specific ZP domain-containing protein|liver-specific zona pellucida domain-containing protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72893556, end_position_on_the_genomic_accession=72933036, orientation='plus', exon_count=9, OMIM=609330, no_hdr0=''),
  170393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170393, CurrentID=0, Status='live', Symbol='C10orf91', Aliases='bA432J24.4', description='chromosome 10 open reading frame 91', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132445210, end_position_on_the_genomic_accession=132448321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170394, CurrentID=0, Status='live', Symbol='PWWP2B', Aliases='PWWP2, bA432J24.1, pp8607', description='PWWP domain containing 2B', other_designations='PWWP domain containing 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132397197, end_position_on_the_genomic_accession=132417863, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130435, CurrentID=0, Status='live', Symbol='RPS7P13', Aliases='RPS7_7_1794', description='ribosomal protein S7 pseudogene 13', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95679099, end_position_on_the_genomic_accession=95679731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  149703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149703, CurrentID=0, Status='live', Symbol='LOC149703', Aliases='', description='uncharacterized LOC149703', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189253, CurrentID=0, Status='live', Symbol='TRY-ATA1-1', Aliases='TRNAY9', description='transfer RNA-Tyr (ATA) 1-1', other_designations='transfer RNA tyrosine 9 (anticodon AUA)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218245826, end_position_on_the_genomic_accession=218245918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480710, CurrentID=0, Status='live', Symbol='RNU6-47P', Aliases='RNU6-47', description='RNA, U6 small nuclear 47, pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533670, CurrentID=0, Status='live', Symbol='LOC100533670', Aliases='', description='tripartite motif containing 60 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65034664, end_position_on_the_genomic_accession=65035990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480711, CurrentID=0, Status='live', Symbol='ZNF736P5Y', Aliases='', description='zinc finger protein 736 pseudogene 5, Y-linked', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480712, CurrentID=0, Status='live', Symbol='FABP5P13', Aliases='FABP5L13', description='fatty acid binding protein 5 pseudogene 13', other_designations='', map_location='Xp22.33 and Yp11.32', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399774, CurrentID=0, Status='live', Symbol='PRKRIRP3', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 3', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52411014, end_position_on_the_genomic_accession=52413539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369559, CurrentID=0, Status='live', Symbol='LOC105369559', Aliases='', description='uncharacterized LOC105369559', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126883039, end_position_on_the_genomic_accession=126940647, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100189257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189257, CurrentID=0, Status='live', Symbol='TRN-GTT11-2', Aliases='TRNAN22', description='transfer RNA-Asn (GTT) 11-2', other_designations='tRNA-Asn-GTT-11-2|transfer RNA asparagine 22 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149646451, end_position_on_the_genomic_accession=149646524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170425, CurrentID=0, Status='live', Symbol='LINC00858', Aliases='', description='long intergenic non-protein coding RNA 858', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84279980, end_position_on_the_genomic_accession=84294659, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480714, CurrentID=0, Status='live', Symbol='USP9YP17', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 17', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480715, CurrentID=0, Status='live', Symbol='USP9YP25', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 25', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533671, CurrentID=0, Status='live', Symbol='ADAM20P2', Aliases='', description='ADAM metallopeptidase domain 20 pseudogene 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175574602, end_position_on_the_genomic_accession=175576193, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480716, CurrentID=0, Status='live', Symbol='USP9YP27', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 27', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480717, CurrentID=0, Status='live', Symbol='USP9YP20', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 20', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874104, CurrentID=0, Status='live', Symbol='PLSCR5-AS1', Aliases='', description='PLSCR5 antisense RNA 1', other_designations='PLSCR5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84237, CurrentID=0, Status='live', Symbol='DKFZp547J222', Aliases='', description='uncharacterized LOC84237', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480718, CurrentID=0, Status='live', Symbol='USP9YP32', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 32', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100309464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100309464, CurrentID=0, Status='live', Symbol='OTX2-AS1', Aliases='OTX2OS1', description='OTX2 antisense RNA 1 (head to head)', other_designations='OTX2 antisense RNA 1 (non-protein coding)|Otx2 opposite strand transcript 1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56813183, end_position_on_the_genomic_accession=56930833, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480719, CurrentID=0, Status='live', Symbol='USP9YP35', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 35', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170463, CurrentID=0, Status='live', Symbol='SSBP4', Aliases='', description='single stranded DNA binding protein 4', other_designations='', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18418661, end_position_on_the_genomic_accession=18434562, orientation='plus', exon_count=21, OMIM=607391, no_hdr0=''),
  106480720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480720, CurrentID=0, Status='live', Symbol='GPM6BP1', Aliases='', description='glycoprotein M6B pseudogene 1', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533672, CurrentID=0, Status='live', Symbol='TRIM60P19', Aliases='', description='tripartite motif containing 60 pseudogene 19', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66269047, end_position_on_the_genomic_accession=66270412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189265, CurrentID=0, Status='live', Symbol='TRI-TAT2-1', Aliases='TRNAI13', description='transfer RNA-Ile (TAT) 2-1', other_designations='transfer RNA isoleucine 13 (anticodon UAU)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42810536, end_position_on_the_genomic_accession=42810628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372849, CurrentID=0, Status='live', Symbol='LOC105372849', Aliases='', description='uncharacterized LOC105372849', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17247489, end_position_on_the_genomic_accession=17261748, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170474, CurrentID=0, Status='live', Symbol='HFM', Aliases='', description='Hemifacial microsomia', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164210, no_hdr0=''),
  399783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399783, CurrentID=0, Status='live', Symbol='LOC399783', Aliases='', description='zinc finger protein 532 pseudogene', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77730772, end_position_on_the_genomic_accession=77734971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480722, CurrentID=0, Status='live', Symbol='HSFY6P', Aliases='', description='heat shock transcription factor, Y-linked 6, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170482, CurrentID=0, Status='live', Symbol='CLEC4C', Aliases='BDCA2, CD303, CLECSF11, CLECSF7, DLEC, HECL, PRO34150', description='C-type lectin domain family 4 member C', other_designations='BDCA-2|C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 11|C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 7|C-type lectin domain family 4, member C|C-type lectin superfamily member 7|blood dendritic cell antigen 2 protein|dendritic cell lectin b|dendritic lectin', map_location='12p13.2-p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7729415, end_position_on_the_genomic_accession=7749473, orientation='minus', exon_count=7, OMIM=606677, no_hdr0=''),
  106480723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480723, CurrentID=0, Status='live', Symbol='HSFY8P', Aliases='', description='heat shock transcription factor, Y-linked 8, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170485, CurrentID=0, Status='live', Symbol='ARF4P2', Aliases='dJ885L7.7', description='ADP ribosylation factor 4 pseudogene 2', other_designations='ADP-ribosylation factor 4 pseudogene 2', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62871979, end_position_on_the_genomic_accession=62873280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170487, CurrentID=0, Status='live', Symbol='ACTL10', Aliases='C20orf134, dJ63M2.2', description='actin like 10', other_designations='actin-like 10', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33666498, end_position_on_the_genomic_accession=33668525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480724, CurrentID=0, Status='live', Symbol='MLLT10P2', Aliases='', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 10 pseudogene 2', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189269, CurrentID=0, Status='live', Symbol='TRE-CTC1-7', Aliases='TRNAE19', description='transfer RNA-Glu (CTC) 1-7', other_designations='tRNA-Glu-CTC-1-7|transfer RNA glutamic acid 19 (anticodon CUC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125780247, end_position_on_the_genomic_accession=125780318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533673, CurrentID=0, Status='live', Symbol='OSBPL9P6', Aliases='', description='oxysterol binding protein like 9 pseudogene 6', other_designations='oxysterol binding protein-like 9 pseudogene 6', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30889250, end_position_on_the_genomic_accession=30890847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480726, CurrentID=0, Status='live', Symbol='MTND6P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 8', other_designations='MT-ND6 pseudogene 8', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170503, CurrentID=0, Status='live', Symbol='CDC42P1', Aliases='bK3216D2.3', description='cell division cycle 42 pseudogene 1', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34401295, end_position_on_the_genomic_accession=34402772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170504, CurrentID=0, Status='live', Symbol='COX6CP2', Aliases='dJ894K16.1', description='cytochrome c oxidase subunit VIc pseudogene 2', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50479716, end_position_on_the_genomic_accession=50480105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170506, CurrentID=0, Status='live', Symbol='DHX36', Aliases='DDX36, G4R1, MLEL1, RHAU', description='DEAH (Asp-Glu-Ala-His) box polypeptide 36', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 36|DEAH box protein 36|G4 resolvase-1|G4-resolvase 1|MLE-like protein 1|RNA helicase associated with AU-rich element ARE|probable ATP-dependent RNA helicase DHX36', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154275668, end_position_on_the_genomic_accession=154324497, orientation='minus', exon_count=25, OMIM=612767, no_hdr0=''),
  170507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170507, CurrentID=0, Status='live', Symbol='DFNA52', Aliases='DFNA42', description='deafness, autosomal dominant 52', other_designations='deafness, autosomal dominant 42', map_location='5q31.1-q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607683, no_hdr0=''),
  170508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170508, CurrentID=0, Status='live', Symbol='DFNB33', Aliases='', description='deafness, autosomal recessive 33', other_designations='', map_location='10p11.23-q21.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607239, no_hdr0=''),
  170510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170510, CurrentID=0, Status='live', Symbol='EIF4EBP2P1', Aliases='EIF4EBP2P, dJ862K6.3', description='eukaryotic translation initiation factor 4E binding protein 2 pseudogene 1', other_designations='', map_location='20q13.11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43422912, end_position_on_the_genomic_accession=43423594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170511, CurrentID=0, Status='live', Symbol='ENSAP1', Aliases='ENSAP, dJ822J19.1', description='endosulfine alpha pseudogene 1', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=15619394, end_position_on_the_genomic_accession=15619919, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369562, CurrentID=0, Status='live', Symbol='LOC105369562', Aliases='', description='uncharacterized LOC105369562', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=127109523, end_position_on_the_genomic_accession=127116762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100189272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189272, CurrentID=0, Status='live', Symbol='TRY-GTA1-1', Aliases='TRNAY10', description='transfer RNA-Tyr (GTA) 1-1', other_designations='transfer RNA tyrosine 10 (anticodon GUA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26568858, end_position_on_the_genomic_accession=26568948, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170514, CurrentID=0, Status='live', Symbol='FTLP1', Aliases='FTLP, bA465L10.3', description='ferritin, light polypeptide pseudogene 1', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45975117, end_position_on_the_genomic_accession=45976268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170516, CurrentID=0, Status='live', Symbol='GAPDHP2', Aliases='GAPDP2, dJ585I14.1', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 2', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13392614, end_position_on_the_genomic_accession=13393877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170517, CurrentID=0, Status='live', Symbol='GCNT1P1', Aliases='GCNT1P, bA189K21.2', description='glucosaminyl (N-acetyl) transferase 1, core 2 pseudogene 1', other_designations='glucosaminyl (N-acetyl) transferase 1, core 2 (beta-1,6-N-acetylglucosaminyltransferase) pseudogene', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18419977, end_position_on_the_genomic_accession=18421628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480729, CurrentID=0, Status='live', Symbol='RPL7L1P10', Aliases='', description='ribosomal protein L7 like 1 pseudogene 10', other_designations='ribosomal protein L7-like 1 pseudogene 10', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723332, CurrentID=0, Status='live', Symbol='LINC00412', Aliases='', description='long intergenic non-protein coding RNA 412', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27236282, end_position_on_the_genomic_accession=27237257, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170522, CurrentID=0, Status='live', Symbol='GLRXP1', Aliases='GLRXP, bA425M5.1', description='glutaredoxin (thioltransferase) pseudogene 1', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37602553, end_position_on_the_genomic_accession=37603249, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170523, CurrentID=0, Status='live', Symbol='GSTM3P1', Aliases='GSTM3P, dJ984P4.2', description='glutathione S-transferase mu 3 pseudogene 1', other_designations='glutathione S-transferase M3 pseudogene|glutathione S-transferase mu 3 (brain) pseudogene', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21501396, end_position_on_the_genomic_accession=21501969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170524, CurrentID=0, Status='live', Symbol='HNRNPA1P3', Aliases='HNRPA1P3, dJ447F3.1', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 3', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45745559, end_position_on_the_genomic_accession=45746220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480730, CurrentID=0, Status='live', Symbol='C4B-AS1', Aliases='', description='C4B antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170526, CurrentID=0, Status='live', Symbol='HNRNPA3P2', Aliases='HNRPA3P, HNRPA3P2, dJ977B1.2', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 2', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36620402, end_position_on_the_genomic_accession=36621446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533674, CurrentID=0, Status='live', Symbol='NPAP1P4', Aliases='', description='nuclear pore associated protein 1 pseudogene 4', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80357106, end_position_on_the_genomic_accession=80360218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170529, CurrentID=0, Status='live', Symbol='RPSAP1', Aliases='LAMR1P, LAMR1P1, RPSA_24_1712, dJ1193N1.1', description='ribosomal protein SA pseudogene 1', other_designations='laminin receptor 1 pseudogene 1', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51130704, end_position_on_the_genomic_accession=51131743, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170530, CurrentID=0, Status='live', Symbol='MRX78', Aliases='', description='mental retardation, X-linked 78', other_designations='', map_location='Xp11.4-q21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300551, no_hdr0=''),
  106480731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480731, CurrentID=0, Status='live', Symbol='C6orf47-AS1', Aliases='', description='C6orf47 antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170532, CurrentID=0, Status='live', Symbol='OTSC3', Aliases='', description='otosclerosis 3', other_designations='', map_location='6p22.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608244, no_hdr0=''),
  170533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170533, CurrentID=0, Status='live', Symbol='PA2G4P2', Aliases='PA2G4L5, bA102J14.2', description='proliferation-associated 2G4 pseudogene 2', other_designations='proliferation-associated 2G4, 38kDa pseudogene', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12380010, end_position_on_the_genomic_accession=12381523, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170534, CurrentID=0, Status='live', Symbol='PARK10', Aliases='AAOPD', description='Parkinson disease 10 (susceptibility)', other_designations='Parkinson disease, age at onset of', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606852, no_hdr0=''),
  170535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170535, CurrentID=0, Status='live', Symbol='PGAM3P', Aliases='dJ1128N12.1', description='phosphoglycerate mutase 3, pseudogene', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11621052, end_position_on_the_genomic_accession=11621478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170536, CurrentID=0, Status='live', Symbol='PPIAP21', Aliases='PPIAL, dJ269M15.1', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 21', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43230733, end_position_on_the_genomic_accession=43231461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480732, CurrentID=0, Status='live', Symbol='PSMC1P5', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 5', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874107, CurrentID=0, Status='live', Symbol='MTUS2-AS1', Aliases='', description='MTUS2 antisense RNA 1', other_designations='MTUS2 antisense RNA 1 (non-protein coding)|hCG2020170', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29476515, end_position_on_the_genomic_accession=29487750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  170540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170540, CurrentID=0, Status='live', Symbol='PPIAP17', Aliases='PPIP11, bA494B22.1', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 17', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 11', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16223750, end_position_on_the_genomic_accession=16224535, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170541, CurrentID=0, Status='live', Symbol='PSMD10P1', Aliases='PSMD10P, dJ914P20.4', description='proteasome 26S subunit, non-ATPase, 10 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 10 pseudogene', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50913715, end_position_on_the_genomic_accession=50914595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170542, CurrentID=0, Status='live', Symbol='PTMAP6', Aliases='dJ539E24.2', description='prothymosin, alpha pseudogene 6', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56701345, end_position_on_the_genomic_accession=56701514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480733, CurrentID=0, Status='live', Symbol='CYP4F44P', Aliases='', description='cytochrome P450 family 4 subfamily F member 44, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 44, pseudogene', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130436, CurrentID=0, Status='live', Symbol='LOC100130436', Aliases='', description='uncharacterized LOC100130436', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=2506666, end_position_on_the_genomic_accession=2507463, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  170547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170547, CurrentID=0, Status='live', Symbol='AIMP1P1', Aliases='SCYE1P, bA400P21.1', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 1 pseudogene 1', other_designations='small inducible cytokine subfamily E, member 1 (endothelial monocyte-activating) pseudogene', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14127331, end_position_on_the_genomic_accession=14128413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170548, CurrentID=0, Status='live', Symbol='SF3A3P1', Aliases='SF3A3P, dJ581P3.1', description='splicing factor 3a, subunit 3 pseudogene 1', other_designations='splicing factor 3a, subunit 3, 60kDa pseudogene', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3489299, end_position_on_the_genomic_accession=3491027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480734, CurrentID=0, Status='live', Symbol='CYP2C61P', Aliases='', description='cytochrome P450 family 2 subfamily C member 61, pseudogene', other_designations='cytochrome P450, family 2, subfamily C, polypeptide 61, pseudogene', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132250, CurrentID=0, Status='live', Symbol='DAP4C', Aliases='', description='death-associated protein 4C', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170552, CurrentID=0, Status='live', Symbol='SRMP1', Aliases='dJ1057D4.1', description='spermidine synthase pseudogene 1', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47859716, end_position_on_the_genomic_accession=47860949, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170554, CurrentID=0, Status='live', Symbol='DYNLT3P1', Aliases='TCTE1P, dJ531H16.2', description='dynein, light chain, Tctex-type 3 pseudogene 1', other_designations='dynein light chain Tctex-type 3 pseudogene|t-complex-associated-testis-expressed 1 pseudogene|t-complex-associated-testis-expressed 1-like pseudogene', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17479085, end_position_on_the_genomic_accession=17479431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480735, CurrentID=0, Status='live', Symbol='XIAP-AS1', Aliases='', description='XIAP antisense RNA 1', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170556, CurrentID=0, Status='live', Symbol='UBE2V1P1', Aliases='dJ687F11.3', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3360484, end_position_on_the_genomic_accession=3362346, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533675, CurrentID=0, Status='live', Symbol='NPAP1P6', Aliases='', description='nuclear pore associated protein 1 pseudogene 6', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79863020, end_position_on_the_genomic_accession=79866875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170559, CurrentID=0, Status='live', Symbol='XPOTP1', Aliases='XPO3P1, dJ785G19.1', description='exportin, tRNA pseudogene 1', other_designations='exportin, tRNA (nuclear export receptor for tRNAs) pseudogene 1', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34213486, end_position_on_the_genomic_accession=34216093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480736, CurrentID=0, Status='live', Symbol='MCCC1-AS1', Aliases='', description='MCCC1 antisense RNA 1', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480737, CurrentID=0, Status='live', Symbol='ITCH-AS1', Aliases='', description='ITCH antisense RNA 1', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170572, CurrentID=0, Status='live', Symbol='HTR3C', Aliases='', description='5-hydroxytryptamine (serotonin) receptor 3C, ionotropic', other_designations='5-HT3-C|5-HT3C|5-hydroxytryptamine (serotonin) receptor 3, family member C|5-hydroxytryptamine receptor 3 subunit C|serotonin receptor 3C', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184053047, end_position_on_the_genomic_accession=184060673, orientation='plus', exon_count=9, OMIM=610121, no_hdr0=''),
  106480738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480738, CurrentID=0, Status='live', Symbol='OSTN-AS1', Aliases='', description='OSTN antisense RNA 1', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170575, CurrentID=0, Status='live', Symbol='GIMAP1', Aliases='HIMAP1, IAN2, IMAP1, IMAP38', description='GTPase, IMAP family member 1', other_designations='immune-associated nucleotide-binding protein 2|immunity associated protein 1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150716557, end_position_on_the_genomic_accession=150724280, orientation='plus', exon_count=3, OMIM=608084, no_hdr0=''),
  106480739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480739, CurrentID=0, Status='live', Symbol='RHOA-IT1', Aliases='', description='RHOA intronic transcript 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23138, CurrentID=0, Status='live', Symbol='N4BP3', Aliases='LZTS4', description='NEDD4 binding protein 3', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178113506, end_position_on_the_genomic_accession=178127034, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480740, CurrentID=0, Status='live', Symbol='GK-IT1', Aliases='', description='GK intronic transcript 1', other_designations='', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533676, CurrentID=0, Status='live', Symbol='PHACTR2P1', Aliases='', description='phosphatase and actin regulator 2 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119815218, end_position_on_the_genomic_accession=119816648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170589, CurrentID=0, Status='live', Symbol='GPHA2', Aliases='A2, GPA2, ZSIG51', description='glycoprotein hormone alpha 2', other_designations='cysteine knot protein|glycoprotein alpha 2|putative secreted protein Zsig51|thyrostimulin subunit alpha', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64934471, end_position_on_the_genomic_accession=64938132, orientation='minus', exon_count=6, OMIM=609651, no_hdr0=''),
  170591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170591, CurrentID=0, Status='live', Symbol='S100Z', Aliases='Gm625, S100-zeta', description='S100 calcium binding protein Z', other_designations='S100 calcium binding protein, zeta|S100 calcium-binding protein Z', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76850001, end_position_on_the_genomic_accession=76952882, orientation='plus', exon_count=11, OMIM=610103, no_hdr0=''),
  170592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170592, CurrentID=0, Status='live', Symbol='UVM1', Aliases='', description='Melanoma, uveal, susceptibility to, 1', other_designations='', map_location='3q24-q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606660, no_hdr0=''),
  170593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170593, CurrentID=0, Status='live', Symbol='UVM2', Aliases='UBM2', description='Melanoma, uveal, susceptibility to, 2', other_designations='', map_location='3p25.2-p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606661, no_hdr0=''),
  170594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170594, CurrentID=0, Status='live', Symbol='WS2C', Aliases='', description='Waardenburg syndrome, type IIC', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606662, no_hdr0=''),
  170595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170595, CurrentID=0, Status='live', Symbol='IBD8', Aliases='', description='inflammatory bowel disease 8', other_designations='', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606668, no_hdr0=''),
  106480742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480742, CurrentID=0, Status='live', Symbol='RABGAP1L-IT1', Aliases='', description='RABGAP1L intronic transcript 1', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723345, CurrentID=0, Status='live', Symbol='LOC102723345', Aliases='', description='uncharacterized LOC102723345', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29647640, end_position_on_the_genomic_accession=29655478, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  399804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399804, CurrentID=0, Status='live', Symbol='NPM1P25', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 25', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96189273, end_position_on_the_genomic_accession=96190313, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369565, CurrentID=0, Status='live', Symbol='LOC105369565', Aliases='', description='uncharacterized LOC105369565', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128565279, end_position_on_the_genomic_accession=128586628, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480743, CurrentID=0, Status='live', Symbol='BRWD1-IT1', Aliases='', description='BRWD1 intronic transcript 1', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480744, CurrentID=0, Status='live', Symbol='MTND1P12', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 12', other_designations='MT-ND1 pseudogene 12', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399806, CurrentID=0, Status='live', Symbol='LBX1-AS1', Aliases='', description='LBX1 antisense RNA 1 (head to head)', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101229594, end_position_on_the_genomic_accession=101238859, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480745, CurrentID=0, Status='live', Symbol='TUBBP11', Aliases='', description='tubulin beta class I pseudogene 11', other_designations='tubulin, beta class I pseudogene 11', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533677, CurrentID=0, Status='live', Symbol='LOC100533677', Aliases='', description='basal body orientation factor 1 pseudogene', other_designations='coiled-coil domain containing 176 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29105614, end_position_on_the_genomic_accession=29106832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480746, CurrentID=0, Status='live', Symbol='TUBB4BP3', Aliases='', description='tubulin beta 4B class IVb pseudogene 3', other_designations='tubulin, beta 4B class IVb pseudogene 3', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170622, CurrentID=0, Status='live', Symbol='COMMD6', Aliases='Acrg', description='COMM domain containing 6', other_designations='Acrg embryonic lethality minimal region ortholog', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75525214, end_position_on_the_genomic_accession=75549439, orientation='minus', exon_count=7, OMIM=612377, no_hdr0=''),
  170626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170626, CurrentID=0, Status='live', Symbol='XAGE3', Aliases='CT12.3a, CT12.3b, GAGED4, PLAC6, XAGE-3, pp9012', description='X antigen family member 3', other_designations='CT12.3|G antigen, family D, 4|X antigen family, member 3|cancer/testis antigen 12.3|cancer/testis antigen family 12, member 3a|cancer/testis antigen family 12, member 3b|g antigen family D member 4|placenta-specific 6|placenta-specific gene 6 protein|protein XAGE-3', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52862528, end_position_on_the_genomic_accession=52868135, orientation='minus', exon_count=8, OMIM=300740, no_hdr0=''),
  170627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170627, CurrentID=0, Status='live', Symbol='XAGE5', Aliases='CT12.5, GAGED5, XAGE-5', description='X antigen family member 5', other_designations='G antigen, family D, 5|X antigen family, member 5|cancer/testis antigen 12.5|cancer/testis antigen family 12, member 5|g antigen family D member 5|protein XAGE-5', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52811294, end_position_on_the_genomic_accession=52818298, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102723350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723350, CurrentID=0, Status='live', Symbol='LOC102723350', Aliases='', description='uncharacterized LOC102723350', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69993790, end_position_on_the_genomic_accession=70019525, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480748, CurrentID=0, Status='live', Symbol='CYP2C60P', Aliases='', description='cytochrome P450 family 2 subfamily C member 60, pseudogene', other_designations='cytochrome P450, family 2, subfamily C, polypeptide 60, pseudogene', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480749, CurrentID=0, Status='live', Symbol='CYP3A54P', Aliases='', description='cytochrome P450 family 3 subfamily A member 54, pseudogene', other_designations='cytochrome P450, family 3, subfamily A, polypeptide 54, pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480750, CurrentID=0, Status='live', Symbol='RNU6-60P', Aliases='RNU6-60', description='RNA, U6 small nuclear 60, pseudogene', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533678, CurrentID=0, Status='live', Symbol='LOC100533678', Aliases='', description='serine/arginine repetitive matrix 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61858143, end_position_on_the_genomic_accession=61860506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480751, CurrentID=0, Status='live', Symbol='RNU6-65P', Aliases='RNU6-65', description='RNA, U6 small nuclear 65, pseudogene', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723354, CurrentID=0, Status='live', Symbol='LOC102723354', Aliases='', description='uncharacterized LOC102723354', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105094147, end_position_on_the_genomic_accession=105099496, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480752, CurrentID=0, Status='live', Symbol='RNU6-74P', Aliases='RNU6-74', description='RNA, U6 small nuclear 74, pseudogene', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399814, CurrentID=0, Status='live', Symbol='C10orf120', Aliases='bA318C4.1', description='chromosome 10 open reading frame 120', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122697709, end_position_on_the_genomic_accession=122699822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369567, CurrentID=0, Status='live', Symbol='LOC105369567', Aliases='', description='uncharacterized LOC105369567', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128597302, end_position_on_the_genomic_accession=128607392, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480753, CurrentID=0, Status='live', Symbol='RNU4ATAC3P', Aliases='', description='RNA, U4atac small nuclear 3, pseudogene', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  93980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93980, CurrentID=0, Status='live', Symbol='DYRK1AIP1', Aliases='', description='DYRK1A interacting protein 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399815, CurrentID=0, Status='live', Symbol='LOC399815', Aliases='', description='chromosome 10 open reading frame 88 pseudogene', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122879633, end_position_on_the_genomic_accession=122898714, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480754, CurrentID=0, Status='live', Symbol='RNA5SP57', Aliases='RN5S57, RN5S58, RNA5SP58', description='RNA, 5S ribosomal pseudogene 57', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  93981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93981, CurrentID=0, Status='live', Symbol='DYRK1AIP2', Aliases='', description='DYRK1A interacting protein 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480755, CurrentID=0, Status='live', Symbol='RNA5SP100', Aliases='RN5S100', description='RNA, 5S ribosomal pseudogene 100', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533679, CurrentID=0, Status='live', Symbol='LOC100533679', Aliases='', description='URI1, prefoldin-like chaperone pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22228710, end_position_on_the_genomic_accession=22230419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170679, CurrentID=0, Status='live', Symbol='PSORS1C1', Aliases='C6orf16, SEEK1', description='psoriasis susceptibility 1 candidate 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31114831, end_position_on_the_genomic_accession=31140092, orientation='plus', exon_count=6, OMIM=613525, no_hdr0=''),
  170680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170680, CurrentID=0, Status='live', Symbol='PSORS1C2', Aliases='C6orf17, SPR1', description='psoriasis susceptibility 1 candidate 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31137534, end_position_on_the_genomic_accession=31139350, orientation='minus', exon_count=2, OMIM=613526, no_hdr0=''),
  106480756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480756, CurrentID=0, Status='live', Symbol='RNA5SP134', Aliases='RN5S134', description='RNA, 5S ribosomal pseudogene 134', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170682, CurrentID=0, Status='live', Symbol='SLEH1', Aliases='', description='systemic lupus erythematosus with hemolytic anemia 1', other_designations='systematic lupus erythematosus with hemolytic anemia 1', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607279, no_hdr0=''),
  399818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399818, CurrentID=0, Status='live', Symbol='METTL10', Aliases='C10orf138', description='methyltransferase like 10', other_designations='methyltransferase-like protein 10', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124757831, end_position_on_the_genomic_accession=124791941, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100189301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189301, CurrentID=0, Status='live', Symbol='TRL-CAA2-1', Aliases='TRNAL28', description='transfer RNA-Leu (CAA) 2-1', other_designations='transfer RNA leucine 28 (anticodon CAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27605638, end_position_on_the_genomic_accession=27605745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  170688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170688, CurrentID=0, Status='live', Symbol='NUDT4P2', Aliases='', description='nudix hydrolase 4 pseudogene 2', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 4 pseudogene 2', map_location='1p12-p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146297773, end_position_on_the_genomic_accession=146301610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  170689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170689, CurrentID=0, Status='live', Symbol='ADAMTS15', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 15', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 15|ADAM-TS 15|ADAM-TS15|ADAMTS-15|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 15, preproprotein|metalloprotease disintegrin 15 with thrombospondin domains', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130448974, end_position_on_the_genomic_accession=130476645, orientation='plus', exon_count=9, OMIM=607509, no_hdr0=''),
  170690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170690, CurrentID=0, Status='live', Symbol='ADAMTS16', Aliases='ADAMTS16s', description='ADAM metallopeptidase with thrombospondin type 1 motif 16', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 16|ADAM-TS 16|ADAM-TS16|ADAMTS-16|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 16', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5140330, end_position_on_the_genomic_accession=5320299, orientation='plus', exon_count=23, OMIM=607510, no_hdr0=''),
  170691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170691, CurrentID=0, Status='live', Symbol='ADAMTS17', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 17', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 17|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 17', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99971438, end_position_on_the_genomic_accession=100342087, orientation='minus', exon_count=24, OMIM=607511, no_hdr0=''),
  170692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170692, CurrentID=0, Status='live', Symbol='ADAMTS18', Aliases='ADAMTS21, KNO2, MMCAT', description='ADAM metallopeptidase with thrombospondin type 1 motif 18', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 18|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 18|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 21|disintegrin and metalloprotease-like protein', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77282128, end_position_on_the_genomic_accession=77435114, orientation='minus', exon_count=23, OMIM=607512, no_hdr0=''),
  106480758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480758, CurrentID=0, Status='live', Symbol='RNA5SP175', Aliases='RN5S175', description='RNA, 5S ribosomal pseudogene 175', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28449, CurrentID=0, Status='live', Symbol='IGHV3-13', Aliases='IGHV313', description='immunoglobulin heavy variable 3-13', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106129540, end_position_on_the_genomic_accession=106129992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130437, CurrentID=0, Status='live', Symbol='LOC100130437', Aliases='', description='transmembrane protein 126A pseudogene', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65025955, end_position_on_the_genomic_accession=65026715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480759, CurrentID=0, Status='live', Symbol='RNA5SP221', Aliases='RN5S221', description='RNA, 5S ribosomal pseudogene 221', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23142, CurrentID=0, Status='live', Symbol='DCUN1D4', Aliases='', description='DCN1, defective in cullin neddylation 1, domain containing 4', other_designations='DCUN1 domain-containing protein 4|defective in cullin neddylation protein 1-like protein 4', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=51843000, end_position_on_the_genomic_accession=51916837, orientation='plus', exon_count=18, OMIM=612977, no_hdr0=''),
  399821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399821, CurrentID=0, Status='live', Symbol='FLJ37035', Aliases='', description='uncharacterized LOC399821', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125705290, end_position_on_the_genomic_accession=125719493, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480760, CurrentID=0, Status='live', Symbol='RNA5SP253', Aliases='RN5S253', description='RNA, 5S ribosomal pseudogene 253', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373933, CurrentID=0, Status='live', Symbol='LOC105373933', Aliases='', description='uncharacterized LOC105373933', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234285403, end_position_on_the_genomic_accession=234301820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480761, CurrentID=0, Status='live', Symbol='RNA5SP298', Aliases='RN5S298', description='RNA, 5S ribosomal pseudogene 298', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170712, CurrentID=0, Status='live', Symbol='COX7B2', Aliases='', description='cytochrome c oxidase subunit VIIb2', other_designations='cytochrome c oxidase polypeptide VIIb2', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46734827, end_position_on_the_genomic_accession=46909267, orientation='minus', exon_count=8, OMIM=609811, no_hdr0=''),
  399823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399823, CurrentID=0, Status='live', Symbol='FOXI2', Aliases='', description='forkhead box I2', other_designations='homolog of mouse Foxi2', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127737274, end_position_on_the_genomic_accession=127741186, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480762, CurrentID=0, Status='live', Symbol='RNA5SP310', Aliases='RN5S310, RN5S316, RNA5SP316', description='RNA, 5S ribosomal pseudogene 310', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480763, CurrentID=0, Status='live', Symbol='RNA5SP333', Aliases='RN5S333', description='RNA, 5S ribosomal pseudogene 333', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480764, CurrentID=0, Status='live', Symbol='RNA5SP351', Aliases='RN5S351', description='RNA, 5S ribosomal pseudogene 351', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137682, CurrentID=0, Status='live', Symbol='NDUFAF6', Aliases='C8orf38', description='NADH:ubiquinone oxidoreductase complex assembly factor 6', other_designations='UPF0551 protein C8orf38, mitochondrial|putative phytoene synthase', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94895451, end_position_on_the_genomic_accession=95058716, orientation='plus', exon_count=19, OMIM=612392, no_hdr0=''),
  100189309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189309, CurrentID=0, Status='live', Symbol='TRW-CCA6-1', Aliases='TRNAW8', description='transfer RNA-Trp (CCA) 6-1', other_designations='transfer RNA tryptophan 8 (anticodon CCA)', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399827, CurrentID=0, Status='live', Symbol='LINC01164', Aliases='', description='long intergenic non-protein coding RNA 1164', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131772398, end_position_on_the_genomic_accession=131790199, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480766, CurrentID=0, Status='live', Symbol='RNA5SP389', Aliases='RN5S389', description='RNA, 5S ribosomal pseudogene 389', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480767, CurrentID=0, Status='live', Symbol='RNA5SP429', Aliases='RN5S429', description='RNA, 5S ribosomal pseudogene 429', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723370, CurrentID=0, Status='live', Symbol='LOC102723370', Aliases='', description='uncharacterized LOC102723370', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=21753593, end_position_on_the_genomic_accession=22121531, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  399829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399829, CurrentID=0, Status='live', Symbol='LINC01168', Aliases='', description='long intergenic non-protein coding RNA 1168', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132965534, end_position_on_the_genomic_accession=132976497, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480768, CurrentID=0, Status='live', Symbol='RNA5SP442', Aliases='RN5S442', description='RNA, 5S ribosomal pseudogene 442', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480769, CurrentID=0, Status='live', Symbol='RNA5SP468', Aliases='RN5S468', description='RNA, 5S ribosomal pseudogene 468', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100101932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101932, CurrentID=0, Status='live', Symbol='GNG5P3', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 5 pseudogene 3', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111864193, end_position_on_the_genomic_accession=111864698, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480770, CurrentID=0, Status='live', Symbol='RNA5SP498', Aliases='RN5S498', description='RNA, 5S ribosomal pseudogene 498', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533682, CurrentID=0, Status='live', Symbol='MIPEPP1', Aliases='', description='mitochondrial intermediate peptidase pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112735000, end_position_on_the_genomic_accession=112738438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399832, CurrentID=0, Status='live', Symbol='DKFZp686M1136', Aliases='', description='uncharacterized LOC399832', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480771, CurrentID=0, Status='live', Symbol='RNA5SP520', Aliases='RN5S520', description='RNA, 5S ribosomal pseudogene 520', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399833, CurrentID=0, Status='live', Symbol='SPRNP1', Aliases='', description='shadow of prion protein homolog (zebrafish) pseudogene 1', other_designations='shadow of prion protein pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133566934, end_position_on_the_genomic_accession=133569958, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480772, CurrentID=0, Status='live', Symbol='RNA5SP523', Aliases='RN5S523', description='RNA, 5S ribosomal pseudogene 523', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100101935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101935, CurrentID=0, Status='live', Symbol='GNG5P5', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 5 pseudogene 5', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47307116, end_position_on_the_genomic_accession=47307540, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874115, CurrentID=0, Status='live', Symbol='PLCXD2-AS1', Aliases='', description='PLCXD2 antisense RNA 1', other_designations='PLCXD2 antisense RNA 1 (non-protein coding)', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111676266, end_position_on_the_genomic_accession=111677433, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480773, CurrentID=0, Status='live', Symbol='TBC1D3P7', Aliases='', description='TBC1 domain family member 3 pseudogene 7', other_designations='TBC1 domain family, member 3 pseudogene 7', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723376, CurrentID=0, Status='live', Symbol='LOC102723376', Aliases='', description='uncharacterized LOC102723376', other_designations='', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11721, end_position_on_the_genomic_accession=15576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480774, CurrentID=0, Status='live', Symbol='PRSS3P4', Aliases='', description='protease, serine 3 pseudogene 4', other_designations='protease, serine, 3 pseudogene 4', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480775, CurrentID=0, Status='live', Symbol='PTGES3P4', Aliases='', description='prostaglandin E synthase 3 (cytosolic) pseudogene 4', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723378, CurrentID=0, Status='live', Symbol='LINC01495', Aliases='', description='long intergenic non-protein coding RNA 1495', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22445672, end_position_on_the_genomic_accession=22492019, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480776, CurrentID=0, Status='live', Symbol='SLC25A18P1', Aliases='', description='solute carrier family 25 (glutamate carrier), member 18 pseudogene 1', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376557, CurrentID=0, Status='live', Symbol='LOC105376557', Aliases='', description='uncharacterized LOC105376557', other_designations='CTC-497E21.3', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12997047, end_position_on_the_genomic_accession=13009204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480777, CurrentID=0, Status='live', Symbol='FAM96AP1', Aliases='', description='family with sequence similarity 96 member A pseudogene 1', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137695, CurrentID=0, Status='live', Symbol='TMEM68', Aliases='', description='transmembrane protein 68', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55738744, end_position_on_the_genomic_accession=55773407, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106480778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480778, CurrentID=0, Status='live', Symbol='ARL5AP3', Aliases='', description='ADP ribosylation factor like GTPase 5A pseudogene 3', other_designations='ADP-ribosylation factor like GTPase 5A pseudogene 3|ADP-ribosylation factor-like 5A pseudogene 3', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170815, CurrentID=0, Status='live', Symbol='BMIQ1', Aliases='', description='body mass index QTL 1', other_designations='Body mass index QTL on chromosome 7|body mass index quantitative trait 1', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606641, no_hdr0=''),
  106480779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480779, CurrentID=0, Status='live', Symbol='WBP1LP4', Aliases='', description='WW domain binding protein 1-like pseudogene 4', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170825, CurrentID=0, Status='live', Symbol='GSX2', Aliases='GSH2', description='GS homeobox 2', other_designations='genetic-screened homeobox 2|homeobox protein GSH-2', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54100081, end_position_on_the_genomic_accession=54101955, orientation='plus', exon_count=2, OMIM=616253, no_hdr0=''),
  106480781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480781, CurrentID=0, Status='live', Symbol='TSEN15P2', Aliases='', description='tRNA splicing endonuclease subunit 15 pseudogene 2', other_designations='TSEN15 tRNA splicing endonuclease subunit pseudogene 2', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420548, CurrentID=0, Status='live', Symbol='LOC100420548', Aliases='', description='GTF2I repeat domain containing 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76175468, end_position_on_the_genomic_accession=76177182, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480782, CurrentID=0, Status='live', Symbol='POLHP1', Aliases='', description='polymerase (DNA directed), eta pseudogene 1', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723385, CurrentID=0, Status='live', Symbol='LOC102723385', Aliases='', description='uncharacterized LOC102723385', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19062754, end_position_on_the_genomic_accession=19067691, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  399844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399844, CurrentID=0, Status='live', Symbol='LINC01002', Aliases='', description='long intergenic non-protein coding RNA 1002', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=197016, end_position_on_the_genomic_accession=202209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480783, CurrentID=0, Status='live', Symbol='CIR1P3', Aliases='', description='corepressor interacting with RBPJ, 1 pseudogene 3', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480784, CurrentID=0, Status='live', Symbol='TRMT112P5', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 5', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170850, CurrentID=0, Status='live', Symbol='KCNG3', Aliases='KV10.1, KV6.3', description='potassium channel, voltage gated modifier subfamily G, member 3', other_designations='potassium voltage-gated channel, subfamily G, member 3|voltage-gated potassium channel 6.3|voltage-gated potassium channel Kv10.1|voltage-gated potassium channel subunit Kv10.1|voltage-gated potassium channel subunit Kv6.3|voltage-gated potassium channel subunit Kv6.4', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42442017, end_position_on_the_genomic_accession=42494097, orientation='minus', exon_count=2, OMIM=606767, no_hdr0=''),
  102723387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723387, CurrentID=0, Status='live', Symbol='LOC102723387', Aliases='', description='Ig heavy chain V-III region J606-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480785, CurrentID=0, Status='live', Symbol='NDUFS5P6', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 6', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 6', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480786, CurrentID=0, Status='live', Symbol='ATG4AP1', Aliases='', description='autophagy related 4A, cysteine peptidase pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480787, CurrentID=0, Status='live', Symbol='VWA8P1', Aliases='', description='von Willebrand factor A domain containing 8 pseudogene 1', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723390, CurrentID=0, Status='live', Symbol='FRG1EP', Aliases='', description='FSHD region gene 1 family member E, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29457393, end_position_on_the_genomic_accession=29496909, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100874118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874118, CurrentID=0, Status='live', Symbol='DDX26B-AS1', Aliases='', description='DDX26B antisense RNA 1', other_designations='DDX26B antisense RNA 1 (non-protein coding)', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135520083, end_position_on_the_genomic_accession=135520674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189332, CurrentID=0, Status='live', Symbol='TRK-TTT7-1', Aliases='TRNAK29', description='transfer RNA-Lys (TTT) 7-1', other_designations='tRNA-Lys-TTT-7-1|transfer RNA lysine 29 (anticodon UUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28747744, end_position_on_the_genomic_accession=28747816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480789, CurrentID=0, Status='live', Symbol='ALG1L11P', Aliases='', description='asparagine-linked glycosylation 1-like 11, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372193, CurrentID=0, Status='live', Symbol='LOC105372193', Aliases='', description='uncharacterized LOC105372193', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74495483, end_position_on_the_genomic_accession=74496163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  442142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442142, CurrentID=0, Status='live', Symbol='SOX30P1', Aliases='', description='SRY-box 30 pseudogene 1', other_designations='SRY (sex determining region Y)-box 30 pseudogene 1|SRY box 30 pseudogene 1', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157689933, end_position_on_the_genomic_accession=157691236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399851, CurrentID=0, Status='live', Symbol='LOC399851', Aliases='', description='uncharacterized LOC399851', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480790, CurrentID=0, Status='live', Symbol='ROBO2P1', Aliases='', description='roundabout guidance receptor 2 pseudogene 1', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723393, CurrentID=0, Status='live', Symbol='LOC102723393', Aliases='', description='uncharacterized LOC102723393', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480791, CurrentID=0, Status='live', Symbol='HMGN2P47', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 47', other_designations='', map_location='15q22.32', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480792, CurrentID=0, Status='live', Symbol='CHCHD4P3', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 4 pseudogene 3', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874119, CurrentID=0, Status='live', Symbol='TCEAL3-AS1', Aliases='', description='TCEAL3 antisense RNA 1', other_designations='TCEAL3 antisense RNA 1 (non-protein coding)', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189337, CurrentID=0, Status='live', Symbol='TRY-GTA8-1', Aliases='TRNAY12', description='transfer RNA-Tyr (GTA) 8-1', other_designations='transfer RNA tyrosine 12 (anticodon GUA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26575570, end_position_on_the_genomic_accession=26575659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480794, CurrentID=0, Status='live', Symbol='MRPL42P6', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 6', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480795, CurrentID=0, Status='live', Symbol='MTATP8P1', Aliases='', description='mitochondrially encoded ATP synthase 8 pseudogene 1', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399857, CurrentID=0, Status='live', Symbol='OR7E117P', Aliases='OST716', description='olfactory receptor family 7 subfamily E member 117 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 117 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3599643, end_position_on_the_genomic_accession=3600668, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480796, CurrentID=0, Status='live', Symbol='MTATP6P1', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 1', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480797, CurrentID=0, Status='live', Symbol='MTATP6P4', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 4', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874120, CurrentID=0, Status='live', Symbol='ZNF630-AS1', Aliases='', description='ZNF630 antisense RNA 1', other_designations='ZNF630 antisense RNA 1 (non-protein coding)', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48056310, end_position_on_the_genomic_accession=48066583, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480798, CurrentID=0, Status='live', Symbol='MTATP6P7', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 7', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102800310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800310, CurrentID=0, Status='live', Symbol='HAGLROS', Aliases='', description='HAGLR opposite strand (non-protein coding)', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176178166, end_position_on_the_genomic_accession=176179009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102800311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800311, CurrentID=0, Status='live', Symbol='TP53COR1', Aliases='TRP53COR1, linc-p21, lincRNA-p21', description='tumor protein p53 pathway corepressor 1 (non-protein coding)', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616343, no_hdr0=''),
  102800314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800314, CurrentID=0, Status='live', Symbol='MIR7515HG', Aliases='', description='MIR7515 host gene', other_designations='MIR7515 host gene (non-protein coding)|lncRNA LOC102800314', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6649107, end_position_on_the_genomic_accession=6650519, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102800315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800315, CurrentID=0, Status='live', Symbol='LINC01246', Aliases='', description='long intergenic non-protein coding RNA 1246', other_designations='lncRNA LOC102800315', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6626109, end_position_on_the_genomic_accession=6634706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  677828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677828, CurrentID=0, Status='live', Symbol='SNORA47', Aliases='HBI-115', description='small nucleolar RNA, H/ACA box 47', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77080434, end_position_on_the_genomic_accession=77080571, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102800317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800317, CurrentID=0, Status='live', Symbol='LOC400927-CSNK1E', Aliases='', description='LOC400927-CSNK1E readthrough', other_designations='CKI-epsilon', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38290691, end_position_on_the_genomic_accession=38398522, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  106480800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480800, CurrentID=0, Status='live', Symbol='UBL5P4', Aliases='', description='ubiquitin like 5 pseudogene 4', other_designations='ubiquitin-like 5 pseudogene 4', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723403, CurrentID=0, Status='live', Symbol='LOC102723403', Aliases='', description='uncharacterized LOC102723403', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30359366, end_position_on_the_genomic_accession=30363158, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170949, CurrentID=0, Status='live', Symbol='DEFT1P', Aliases='DEFQ1, DEFT1', description='defensin, theta 1 pseudogene', other_designations='circular minidefensin|defensin-related pseudogene|demidefensin|pro-retrocyclin|retrocyclin pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6987178, end_position_on_the_genomic_accession=6989721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480801, CurrentID=0, Status='live', Symbol='SDCBP2P1', Aliases='', description='syndecan binding protein (syntenin) 2 pseudogene 1', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170954, CurrentID=0, Status='live', Symbol='PPP1R18', Aliases='HKMT1098, KIAA1949', description='protein phosphatase 1 regulatory subunit 18', other_designations='protein phosphatase 1 F-actin cytoskeleton targeting subunit|protein phosphatase 1, regulatory subunit 18', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30676389, end_position_on_the_genomic_accession=30687895, orientation='minus', exon_count=4, OMIM=610990, no_hdr0=''),
  106480802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480802, CurrentID=0, Status='live', Symbol='UBE2V2P4', Aliases='', description='ubiquitin conjugating enzyme E2 variant 2 pseudogene 4', other_designations='ubiquitin-conjugating enzyme E2 variant 2 pseudogene 4', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170958, CurrentID=0, Status='live', Symbol='ZNF525', Aliases='', description='zinc finger protein 525', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53365715, end_position_on_the_genomic_accession=53386590, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  170959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170959, CurrentID=0, Status='live', Symbol='ZNF431', Aliases='', description='zinc finger protein 431', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21142004, end_position_on_the_genomic_accession=21186002, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  170960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170960, CurrentID=0, Status='live', Symbol='ZNF721', Aliases='', description='zinc finger protein 721', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=439984, end_position_on_the_genomic_accession=499171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  170961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170961, CurrentID=0, Status='live', Symbol='ANKRD24', Aliases='', description='ankyrin repeat domain 24', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4182625, end_position_on_the_genomic_accession=4224814, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  105378306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378306, CurrentID=0, Status='live', Symbol='LOC105378306', Aliases='', description='uncharacterized LOC105378306', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52790280, end_position_on_the_genomic_accession=52812468, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480803, CurrentID=0, Status='live', Symbol='HSPA8P13', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 13', other_designations='heat shock 70kDa protein 8 pseudogene 13', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723406, CurrentID=0, Status='live', Symbol='LOC102723406', Aliases='', description='uncharacterized LOC102723406', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33546743, end_position_on_the_genomic_accession=33611975, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100189348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189348, CurrentID=0, Status='live', Symbol='TRI-TAT3-1', Aliases='TRNAI21', description='transfer RNA-Ile (TAT) 3-1', other_designations='transfer RNA isoleucine 21 (anticodon UAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28537590, end_position_on_the_genomic_accession=28537683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  677829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677829, CurrentID=0, Status='live', Symbol='SNORA49', Aliases='ACA49', description='small nucleolar RNA, H/ACA box 49', other_designations='ACA49 snoRNA', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132031224, end_position_on_the_genomic_accession=132031360, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723407, CurrentID=0, Status='live', Symbol='LOC102723407', Aliases='', description='putative V-set and immunoglobulin domain-containing-like protein IGHV4OR15-8', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480805, CurrentID=0, Status='live', Symbol='NIFKP9', Aliases='MKI67IPP9', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 9', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723408, CurrentID=0, Status='live', Symbol='LOC102723408', Aliases='', description='tektin-4-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480806, CurrentID=0, Status='live', Symbol='NDUFA3P4', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A3 pseudogene 4', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3 pseudogene 4', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723409, CurrentID=0, Status='live', Symbol='LOC102723409', Aliases='', description='uncharacterized LOC102723409', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129502709, end_position_on_the_genomic_accession=129552937, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  106480807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480807, CurrentID=0, Status='live', Symbol='KRT19P6', Aliases='', description='keratin 19 pseudogene 6', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480808, CurrentID=0, Status='live', Symbol='MED28P6', Aliases='', description='mediator complex subunit 28 pseudogene 6', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480809, CurrentID=0, Status='live', Symbol='CACYBPP3', Aliases='', description='calcyclin binding protein pseudogene 3', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480810, CurrentID=0, Status='live', Symbol='RFPL4AP3', Aliases='', description='ret finger protein-like 4A pseudogene 3', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130746, CurrentID=0, Status='live', Symbol='LOC100130746', Aliases='', description='AKT interacting protein pseudogene', other_designations='', map_location='5p14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24170371, end_position_on_the_genomic_accession=24171245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480811, CurrentID=0, Status='live', Symbol='PIGPP4', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class P pseudogene 4', other_designations='phosphatidylinositol glycan anchor biosynthesis, class P pseudogene 4', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171013, CurrentID=0, Status='live', Symbol='SLI1', Aliases='', description='Specific language impairment QTL, 1', other_designations='', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606711, no_hdr0=''),
  171014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171014, CurrentID=0, Status='live', Symbol='SLI2', Aliases='', description='Specific language impairment QTL, 2', other_designations='', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606712, no_hdr0=''),
  171017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171017, CurrentID=0, Status='live', Symbol='ZNF384', Aliases='CAGH1, CAGH1A, CIZ, ERDA2, NMP4, NP, TNRC1', description='zinc finger protein 384', other_designations='CAG repeat protein 1|Cas-interacting zinc finger protein|expanded repeat domain, CAG/CTG 2|nuclear matrix transcription factor 4|trinucleotide repeat-containing gene 1 protein', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6666477, end_position_on_the_genomic_accession=6689572, orientation='minus', exon_count=14, OMIM=609951, no_hdr0=''),
  171019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171019, CurrentID=0, Status='live', Symbol='ADAMTS19', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 19', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 19|ADAM-TS 19|ADAM-TS19|ADAMTS-19|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 19', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129460399, end_position_on_the_genomic_accession=129738763, orientation='plus', exon_count=24, OMIM=607513, no_hdr0=''),
  100874123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874123, CurrentID=0, Status='live', Symbol='ZNRF3-AS1', Aliases='', description='ZNRF3 antisense RNA 1', other_designations='CTA-747E2.11|ZNRF3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29024999, end_position_on_the_genomic_accession=29031476, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  171022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171022, CurrentID=0, Status='live', Symbol='ABHD11-AS1', Aliases='LINC00035, NCRNA00035, WBSCR26', description='ABHD11 antisense RNA 1 (tail to tail)', other_designations='ABHD11 antisense RNA 1 (non-protein coding)|Williams Beuren syndrome chromosome region 26 (non-protein coding)|long intergenic non-protein coding RNA 35', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73735069, end_position_on_the_genomic_accession=73736000, orientation='plus', exon_count=4, OMIM=612545, no_hdr0=''),
  171023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171023, CurrentID=0, Status='live', Symbol='ASXL1', Aliases='BOPS, MDS', description='additional sex combs like 1, transcriptional regulator', other_designations='additional sex combs like 1|additional sex combs like transcriptional regulator 1', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32358344, end_position_on_the_genomic_accession=32439319, orientation='plus', exon_count=20, OMIM=612990, no_hdr0=''),
  171024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171024, CurrentID=0, Status='live', Symbol='SYNPO2', Aliases='', description='synaptopodin 2', other_designations='genethonin-2|myopodin', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118850688, end_position_on_the_genomic_accession=119061247, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100189358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189358, CurrentID=0, Status='live', Symbol='TRA-AGC14-1', Aliases='TRNAA36', description='transfer RNA-Ala (AGC) 14-1', other_designations='tRNA-Ala-AGC-14-1|transfer RNA alanine 36 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26673362, end_position_on_the_genomic_accession=26673434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  677831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=677831, CurrentID=0, Status='live', Symbol='SNORA51', Aliases='ACA51', description='small nucleolar RNA, H/ACA box 51', other_designations='ACA51 snoRNA', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2655067, end_position_on_the_genomic_accession=2655198, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723417, CurrentID=0, Status='live', Symbol='LOC102723417', Aliases='', description='uncharacterized LOC102723417', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123768972, end_position_on_the_genomic_accession=123777284, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  399876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399876, CurrentID=0, Status='live', Symbol='NAV2-AS4', Aliases='', description='NAV2 antisense RNA 4', other_designations='NAV2 antisense RNA 4 (non-protein coding)', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19510893, end_position_on_the_genomic_accession=19524078, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480815, CurrentID=0, Status='live', Symbol='RNU7-110P', Aliases='', description='RNA, U7 small nuclear 110 pseudogene', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480816, CurrentID=0, Status='live', Symbol='RNU7-116P', Aliases='', description='RNA, U7 small nuclear 116 pseudogene', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  154089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154089, CurrentID=0, Status='live', Symbol='LINC01312', Aliases='', description='long intergenic non-protein coding RNA 1312', other_designations='uncharacterized protein MGC34034', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133821147, end_position_on_the_genomic_accession=133853992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189361, CurrentID=0, Status='live', Symbol='TRP-AGG3-1', Aliases='TRNAP20', description='transfer RNA-Pro (AGG) 3-1', other_designations='transfer RNA proline 20 (anticodon AGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137735, CurrentID=0, Status='live', Symbol='ABRA', Aliases='STARS', description='actin binding Rho activating protein', other_designations='actin-binding Rho activating protein|striated muscle activator of Rho-dependent signaling', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106759483, end_position_on_the_genomic_accession=106770244, orientation='minus', exon_count=2, OMIM=609747, no_hdr0=''),
  106480818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480818, CurrentID=0, Status='live', Symbol='RNU7-125P', Aliases='', description='RNA, U7 small nuclear 125 pseudogene', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480819, CurrentID=0, Status='live', Symbol='RNU7-128P', Aliases='', description='RNA, U7 small nuclear 128 pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929483, CurrentID=0, Status='live', Symbol='LOC101929483', Aliases='', description='cyclin-dependent kinase 2-associated protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7324552, end_position_on_the_genomic_accession=7325050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399881, CurrentID=0, Status='live', Symbol='HNRNPKP3', Aliases='', description='heterogeneous nuclear ribonucleoprotein K pseudogene 3', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43261504, end_position_on_the_genomic_accession=43269369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480820, CurrentID=0, Status='live', Symbol='RNU7-134P', Aliases='', description='RNA, U7 small nuclear 134 pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102800446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800446, CurrentID=0, Status='live', Symbol='LINC01230', Aliases='TCONS_00015639, linc-DMRT2, lnc-DRMT2', description='long intergenic non-protein coding RNA 1230', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1048143, end_position_on_the_genomic_accession=1048641, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480821, CurrentID=0, Status='live', Symbol='RNU7-137P', Aliases='', description='RNA, U7 small nuclear 137 pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480822, CurrentID=0, Status='live', Symbol='RNU7-141P', Aliases='', description='RNA, U7 small nuclear 141 pseudogene', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399884, CurrentID=0, Status='live', Symbol='LOC399884', Aliases='', description='uncharacterized LOC399884', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480823, CurrentID=0, Status='live', Symbol='RNU7-148P', Aliases='', description='RNA, U7 small nuclear 148 pseudogene', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171089, CurrentID=0, Status='live', Symbol='DYX5', Aliases='', description='dyslexia susceptibility 5', other_designations='', map_location='3p12-q13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606896, no_hdr0=''),
  102723427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723427, CurrentID=0, Status='live', Symbol='LOC102723427', Aliases='', description='uncharacterized LOC102723427', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=68020253, end_position_on_the_genomic_accession=68032690, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  399886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399886, CurrentID=0, Status='live', Symbol='LOC399886', Aliases='', description='uncharacterized LOC399886', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45371397, end_position_on_the_genomic_accession=45388508, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480825, CurrentID=0, Status='live', Symbol='RNU7-155P', Aliases='', description='RNA, U7 small nuclear 155 pseudogene', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480826, CurrentID=0, Status='live', Symbol='RNU7-157P', Aliases='', description='RNA, U7 small nuclear 157 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399888, CurrentID=0, Status='live', Symbol='FAM180B', Aliases='', description='family with sequence similarity 180 member B', other_designations='family with sequence similarity 180, member B|hypothetical gene supported by BC065704', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47586678, end_position_on_the_genomic_accession=47589194, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480827, CurrentID=0, Status='live', Symbol='RNU7-169P', Aliases='', description='RNA, U7 small nuclear 169 pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480828, CurrentID=0, Status='live', Symbol='RNU7-172P', Aliases='', description='RNA, U7 small nuclear 172 pseudogene', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480829, CurrentID=0, Status='live', Symbol='RNU7-174P', Aliases='', description='RNA, U7 small nuclear 174 pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723432, CurrentID=0, Status='live', Symbol='LOC102723432', Aliases='', description='serine/threonine-protein kinase PAK 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372458, CurrentID=0, Status='live', Symbol='LOC105372458', Aliases='', description='uncharacterized LOC105372458', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54108402, end_position_on_the_genomic_accession=54110037, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480830, CurrentID=0, Status='live', Symbol='RNU7-180P', Aliases='', description='RNA, U7 small nuclear 180 pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373947, CurrentID=0, Status='live', Symbol='LOC105373947', Aliases='', description='uncharacterized LOC105373947', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236604266, end_position_on_the_genomic_accession=236609360, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480831, CurrentID=0, Status='live', Symbol='RNU7-182P', Aliases='', description='RNA, U7 small nuclear 182 pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  154092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154092, CurrentID=0, Status='live', Symbol='LINC01010', Aliases='', description='long intergenic non-protein coding RNA 1010', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134437716, end_position_on_the_genomic_accession=134504020, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480832, CurrentID=0, Status='live', Symbol='RNU7-188P', Aliases='', description='RNA, U7 small nuclear 188 pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480833, CurrentID=0, Status='live', Symbol='RNU7-192P', Aliases='', description='RNA, U7 small nuclear 192 pseudogene', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480834, CurrentID=0, Status='live', Symbol='RNU7-194P', Aliases='', description='RNA, U7 small nuclear 194 pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480835, CurrentID=0, Status='live', Symbol='RNU7-200P', Aliases='', description='RNA, U7 small nuclear 200 pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171157, CurrentID=0, Status='live', Symbol='SHFM5', Aliases='', description='split hand/foot malformation (ectrodactyly) type 5', other_designations='', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606708, no_hdr0=''),
  106480836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480836, CurrentID=0, Status='live', Symbol='RN7SKP12', Aliases='', description='RNA, 7SK small nuclear pseudogene 12', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399898, CurrentID=0, Status='live', Symbol='LOC399898', Aliases='', description='uncharacterized LOC399898', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=57808145, end_position_on_the_genomic_accession=57812436, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480837, CurrentID=0, Status='live', Symbol='RN7SKP16', Aliases='', description='RNA, 7SK small nuclear pseudogene 16', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171169, CurrentID=0, Status='live', Symbol='SPACA4', Aliases='SAMP14', description='sperm acrosome associated 4', other_designations='sperm acrosomal membrane protein 14|sperm acrosomal membrane-associated protein 14', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48606743, end_position_on_the_genomic_accession=48607714, orientation='plus', exon_count=1, OMIM=609932, no_hdr0=''),
  100874128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874128, CurrentID=0, Status='live', Symbol='LINC00333', Aliases='NCRNA00333', description='long intergenic non-protein coding RNA 333', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84140602, end_position_on_the_genomic_accession=84606768, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106480838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480838, CurrentID=0, Status='live', Symbol='RN7SKP18', Aliases='', description='RNA, 7SK small nuclear pseudogene 18', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171176, CurrentID=0, Status='live', Symbol='NICN2P', Aliases='', description='nicolin 2, pseudogene', other_designations='nicolin 1 pseudogene', map_location='Xp11.3-p11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47272141, end_position_on_the_genomic_accession=47274462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171177, CurrentID=0, Status='live', Symbol='RHOV', Aliases='ARHV, CHP, WRCH2', description='ras homolog family member V', other_designations='CDC42-like GTPase 2|GTP-binding protein-like 2|Rho GTPase-like protein ARHV|WRCH-2|WRCH1-related GTPase|Wnt-1 regulated Cdc42 homolog 2|Wnt-1 responsive Cdc42 homolog 2|ras homolog gene family, member V', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40872214, end_position_on_the_genomic_accession=40874289, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480839, CurrentID=0, Status='live', Symbol='RN7SKP21', Aliases='', description='RNA, 7SK small nuclear pseudogene 21', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480840, CurrentID=0, Status='live', Symbol='RN7SKP24', Aliases='', description='RNA, 7SK small nuclear pseudogene 24', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189385, CurrentID=0, Status='live', Symbol='TRE-CTC1-3', Aliases='TRNAE22', description='transfer RNA-Glu (CTC) 1-3', other_designations='tRNA-Glu-CTC-1-3|transfer RNA glutamic acid 22 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161454608, end_position_on_the_genomic_accession=161454679, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289283, CurrentID=0, Status='live', Symbol='LOC100289283', Aliases='', description='uncharacterized LOC100289283', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=29820000, end_position_on_the_genomic_accession=29823301, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480842, CurrentID=0, Status='live', Symbol='RN7SKP31', Aliases='', description='RNA, 7SK small nuclear pseudogene 31', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  1336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1336, CurrentID=0, Status='live', Symbol='COX5BP7', Aliases='COX5BL7', description='cytochrome c oxidase subunit Vb pseudogene 7', other_designations='cytochrome c oxidase subunit Vb-like 7', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39120204, end_position_on_the_genomic_accession=39120564, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480843, CurrentID=0, Status='live', Symbol='RN7SKP33', Aliases='', description='RNA, 7SK small nuclear pseudogene 33', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480844, CurrentID=0, Status='live', Symbol='RN7SKP37', Aliases='', description='RNA, 7SK small nuclear pseudogene 37', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723447, CurrentID=0, Status='live', Symbol='LOC102723447', Aliases='', description='uncharacterized LOC102723447', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116337122, end_position_on_the_genomic_accession=116450672, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480845, CurrentID=0, Status='live', Symbol='RN7SKP39', Aliases='', description='RNA, 7SK small nuclear pseudogene 39', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373950, CurrentID=0, Status='live', Symbol='LOC105373950', Aliases='', description='uncharacterized LOC105373950', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237062649, end_position_on_the_genomic_accession=237066802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171220, CurrentID=0, Status='live', Symbol='DSTNP2', Aliases='', description='destrin (actin depolymerizing factor) pseudogene 2', other_designations='destrin-2 pseudogene', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6884682, end_position_on_the_genomic_accession=6885786, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171221, CurrentID=0, Status='live', Symbol='DNAJB1P1', Aliases='DNAJB1P, HSP40, psiHSP40', description='DnaJ heat shock protein family (Hsp40) member B1 pseudogene 1', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 1 pseudogene 1|HSP40 pseudogene', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191879992, end_position_on_the_genomic_accession=191882291, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171222, CurrentID=0, Status='live', Symbol='UBE2L5P', Aliases='UBCH7N2, UBE2L5', description='ubiquitin conjugating enzyme E2L 5 (pseudogene)', other_designations='ubiquitin-conjugating enzyme E2L 5 (pseudogene)|ubiquitin-conjugating enzyme E2L 5, pseudogene', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30427985, end_position_on_the_genomic_accession=30428341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480847, CurrentID=0, Status='live', Symbol='RN7SKP46', Aliases='', description='RNA, 7SK small nuclear pseudogene 46', other_designations='', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399909, CurrentID=0, Status='live', Symbol='PCNXL3', Aliases='', description='pecanex-like 3 (Drosophila)', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65614322, end_position_on_the_genomic_accession=65637439, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  106480848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480848, CurrentID=0, Status='live', Symbol='RN7SKP49', Aliases='', description='RNA, 7SK small nuclear pseudogene 49', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480849, CurrentID=0, Status='live', Symbol='RN7SKP53', Aliases='', description='RNA, 7SK small nuclear pseudogene 53', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480850, CurrentID=0, Status='live', Symbol='RN7SKP58', Aliases='', description='RNA, 7SK small nuclear pseudogene 58', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928402, CurrentID=0, Status='live', Symbol='LOC101928402', Aliases='', description='uncharacterized LOC101928402', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123959233, end_position_on_the_genomic_accession=123961341, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480851, CurrentID=0, Status='live', Symbol='RN7SKP60', Aliases='', description='RNA, 7SK small nuclear pseudogene 60', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480852, CurrentID=0, Status='live', Symbol='RN7SKP64', Aliases='', description='RNA, 7SK small nuclear pseudogene 64', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874131, CurrentID=0, Status='live', Symbol='ZBTB20-AS4', Aliases='', description='ZBTB20 antisense RNA 4', other_designations='ZBTB20 antisense RNA 4 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115100423, end_position_on_the_genomic_accession=115103061, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480853, CurrentID=0, Status='live', Symbol='RN7SKP66', Aliases='', description='RNA, 7SK small nuclear pseudogene 66', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377433, CurrentID=0, Status='live', Symbol='LOC105377433', Aliases='', description='uncharacterized LOC105377433', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134453317, end_position_on_the_genomic_accession=134494241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480854, CurrentID=0, Status='live', Symbol='RN7SKP68', Aliases='', description='RNA, 7SK small nuclear pseudogene 68', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480855, CurrentID=0, Status='live', Symbol='RN7SKP73', Aliases='', description='RNA, 7SK small nuclear pseudogene 73', other_designations='', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723458, CurrentID=0, Status='live', Symbol='LOC102723458', Aliases='', description='uncharacterized LOC102723458', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88411483, end_position_on_the_genomic_accession=88467764, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106480856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480856, CurrentID=0, Status='live', Symbol='RN7SKP75', Aliases='', description='RNA, 7SK small nuclear pseudogene 75', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289286, CurrentID=0, Status='live', Symbol='HNRNPA1P58', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 58', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24001824, end_position_on_the_genomic_accession=24002426, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480857, CurrentID=0, Status='live', Symbol='RN7SKP79', Aliases='', description='RNA, 7SK small nuclear pseudogene 79', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480858, CurrentID=0, Status='live', Symbol='RN7SKP81', Aliases='', description='RNA, 7SK small nuclear pseudogene 81', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285033, CurrentID=0, Status='live', Symbol='STARD7-AS1', Aliases='', description='STARD7 antisense RNA 1', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96208416, end_position_on_the_genomic_accession=96242623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723462, CurrentID=0, Status='live', Symbol='LOC102723462', Aliases='', description='uncharacterized LOC102723462', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480860, CurrentID=0, Status='live', Symbol='RN7SKP88', Aliases='', description='RNA, 7SK small nuclear pseudogene 88', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189405, CurrentID=0, Status='live', Symbol='TRY-GTA4-1', Aliases='TRNAY13', description='transfer RNA-Tyr (GTA) 4-1', other_designations='transfer RNA tyrosine 13 (anticodon GUA)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20657464, end_position_on_the_genomic_accession=20657557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399923, CurrentID=0, Status='live', Symbol='FLJ42102', Aliases='', description='uncharacterized LOC399923', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71405746, end_position_on_the_genomic_accession=71423354, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480862, CurrentID=0, Status='live', Symbol='RN7SKP95', Aliases='', description='RNA, 7SK small nuclear pseudogene 95', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480863, CurrentID=0, Status='live', Symbol='RN7SKP98', Aliases='', description='RNA, 7SK small nuclear pseudogene 98', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480864, CurrentID=0, Status='live', Symbol='RN7SKP101', Aliases='', description='RNA, 7SK small nuclear pseudogene 101', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480865, CurrentID=0, Status='live', Symbol='RN7SKP104', Aliases='', description='RNA, 7SK small nuclear pseudogene 104', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480866, CurrentID=0, Status='live', Symbol='RN7SKP108', Aliases='', description='RNA, 7SK small nuclear pseudogene 108', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480867, CurrentID=0, Status='live', Symbol='RN7SKP110', Aliases='', description='RNA, 7SK small nuclear pseudogene 110', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480868, CurrentID=0, Status='live', Symbol='RN7SKP114', Aliases='', description='RNA, 7SK small nuclear pseudogene 114', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480869, CurrentID=0, Status='live', Symbol='RN7SKP116', Aliases='', description='RNA, 7SK small nuclear pseudogene 116', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480870, CurrentID=0, Status='live', Symbol='RN7SKP120', Aliases='', description='RNA, 7SK small nuclear pseudogene 120', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480871, CurrentID=0, Status='live', Symbol='RN7SKP124', Aliases='', description='RNA, 7SK small nuclear pseudogene 124', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480872, CurrentID=0, Status='live', Symbol='RN7SKP126', Aliases='', description='RNA, 7SK small nuclear pseudogene 126', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480873, CurrentID=0, Status='live', Symbol='RN7SKP130', Aliases='', description='RNA, 7SK small nuclear pseudogene 130', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171389, CurrentID=0, Status='live', Symbol='NLRP6', Aliases='AVR, CLR11.4, NALP6, NAVR, NAVR/AVR, PAN3, PYPAF5', description='NLR family, pyrin domain containing 6', other_designations='NACHT, LRR and PYD containing protein 6|NACHT, leucine rich repeat and PYD containing 6|PYRIN-containing APAF1-like protein 5|angiotensin II/vasopressin receptor|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 6', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=278365, end_position_on_the_genomic_accession=285388, orientation='plus', exon_count=8, OMIM=609650, no_hdr0=''),
  105375688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375688, CurrentID=0, Status='live', Symbol='LOC105375688', Aliases='', description='uncharacterized LOC105375688', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104074666, end_position_on_the_genomic_accession=104111648, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  552853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552853, CurrentID=0, Status='live', Symbol='NCRNA00250', Aliases='C8orf24', description='non-protein coding RNA 250', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134838069, end_position_on_the_genomic_accession=134842637, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171392, CurrentID=0, Status='live', Symbol='ZNF675', Aliases='TBZF, TIZ', description='zinc finger protein 675', other_designations='TRAF6 inhibitory zinc finger|TRAF6-binding zinc finger protein|TRAF6-inhibitory zinc finger protein', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23652781, end_position_on_the_genomic_accession=23687215, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480875, CurrentID=0, Status='live', Symbol='RN7SKP135', Aliases='', description='RNA, 7SK small nuclear pseudogene 135', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399937, CurrentID=0, Status='live', Symbol='TRIM51BP', Aliases='SPRYD5P, SPRYD5P1', description='tripartite motif-containing 51B, pseudogene', other_designations='SPRY domain containing 5 pseudogene 1|Tripartite motif-containing protein ENSP00000309378', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89852840, end_position_on_the_genomic_accession=89863065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480876, CurrentID=0, Status='live', Symbol='RN7SKP139', Aliases='', description='RNA, 7SK small nuclear pseudogene 139', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480877, CurrentID=0, Status='live', Symbol='RN7SKP142', Aliases='', description='RNA, 7SK small nuclear pseudogene 142', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723480, CurrentID=0, Status='live', Symbol='LOC102723480', Aliases='', description='uncharacterized LOC102723480', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93693196, end_position_on_the_genomic_accession=93699548, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874136, CurrentID=0, Status='live', Symbol='LINC00350', Aliases='', description='long intergenic non-protein coding RNA 350', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399939, CurrentID=0, Status='live', Symbol='TRIM49D1', Aliases='TRIM49DP, TRIM49D2, TRIM49D2P, TRIM49DP, TRIM49D1', description='tripartite motif containing 49D1', other_designations='tripartite motif containing 49D, pseudogene|tripartite motif containing 49D1, pseudogene|tripartite motif-containing protein 49-like protein|tripartite motif-containing protein 49D1|tripartite motif-containing protein 49D2 pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89911409, end_position_on_the_genomic_accession=89920388, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480878, CurrentID=0, Status='live', Symbol='RN7SKP145', Aliases='', description='RNA, 7SK small nuclear pseudogene 145', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171416, CurrentID=0, Status='live', Symbol='TXNP2', Aliases='', description='thioredoxin pseudogene 2', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68513844, end_position_on_the_genomic_accession=68514743, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171417, CurrentID=0, Status='live', Symbol='LOC171417', Aliases='', description='thioredoxin 1 pseudogene 7', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169765771, end_position_on_the_genomic_accession=169766620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171418, CurrentID=0, Status='live', Symbol='GLRXP2', Aliases='', description='glutaredoxin (thioltransferase) pseudogene 2', other_designations='glutaredoxin pseudogene 2', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90107611, end_position_on_the_genomic_accession=90109303, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  171419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171419, CurrentID=0, Status='live', Symbol='TXNP3', Aliases='', description='thioredoxin pseudogene 3', other_designations='', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112363199, end_position_on_the_genomic_accession=112363661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  171420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171420, CurrentID=0, Status='live', Symbol='HSPE1P28', Aliases='EPF5', description='heat shock protein family E (Hsp10) member 1 pseudogene 28', other_designations='EPF5 pseudogene|heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 28', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112280988, end_position_on_the_genomic_accession=112281408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171421, CurrentID=0, Status='live', Symbol='HSPE1P16', Aliases='EPF8', description='heat shock protein family E (Hsp10) member 1 pseudogene 16', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 16', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26062630, end_position_on_the_genomic_accession=26063160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171422, CurrentID=0, Status='live', Symbol='CRPP1', Aliases='', description='C-reactive protein pseudogene 1', other_designations='C-reactive protein, pentraxin-related pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159704950, end_position_on_the_genomic_accession=159705596, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  137797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137797, CurrentID=0, Status='live', Symbol='LYPD2', Aliases='LYPDC2, UNQ430', description='LY6/PLAUR domain containing 2', other_designations='RGTR430', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142750210, end_position_on_the_genomic_accession=142752534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  171424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171424, CurrentID=0, Status='live', Symbol='SPA17P1', Aliases='Sp17-2', description='sperm autoantigenic protein 17 pseudogene 1', other_designations='sperm protein Sp17-2 pseudogene', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75382217, end_position_on_the_genomic_accession=75382910, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  171425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171425, CurrentID=0, Status='live', Symbol='CLYBL', Aliases='CLB', description='citrate lyase beta like', other_designations='beta-methylmalate synthase|malate synthase', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99606664, end_position_on_the_genomic_accession=99909459, orientation='plus', exon_count=13, OMIM=609686, no_hdr0=''),
  171427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171427, CurrentID=0, Status='live', Symbol='CYP2F2P', Aliases='CYP2F1P', description='cytochrome P450 family 2 subfamily F member 2, pseudogene', other_designations='cytochrome P450, family 2, subfamily F, polypeptide 1 pseudogene|cytochrome P450, family 2, subfamily F, polypeptide 2, pseudogene|cytochrome P450, subfamily IIF, polypeptide 1 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40818477, end_position_on_the_genomic_accession=40826773, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399942, CurrentID=0, Status='live', Symbol='TUBAP2', Aliases='', description='tubulin alpha pseudogene 2', other_designations='tubulin, alpha 1b pseudogene|tubulin, alpha pseudogene 2', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90282560, end_position_on_the_genomic_accession=90284168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480881, CurrentID=0, Status='live', Symbol='RN7SKP154', Aliases='', description='RNA, 7SK small nuclear pseudogene 154', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  552860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552860, CurrentID=0, Status='live', Symbol='SAMD12-AS1', Aliases='C8orf26, NCRNA00252', description='SAMD12 antisense RNA 1', other_designations='SAMD12 antisense RNA 1 (non-protein coding)', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118621001, end_position_on_the_genomic_accession=118726067, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480882, CurrentID=0, Status='live', Symbol='RN7SKP156', Aliases='', description='RNA, 7SK small nuclear pseudogene 156', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480883, CurrentID=0, Status='live', Symbol='RN7SKP158', Aliases='', description='RNA, 7SK small nuclear pseudogene 158', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130442, CurrentID=0, Status='live', Symbol='LOC100130442', Aliases='', description='uncharacterized LOC100130442', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=17061997, end_position_on_the_genomic_accession=17063455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480884, CurrentID=0, Status='live', Symbol='RN7SKP162', Aliases='', description='RNA, 7SK small nuclear pseudogene 162', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442161, CurrentID=0, Status='live', Symbol='LOC442161', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10365511, end_position_on_the_genomic_accession=10366587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480885, CurrentID=0, Status='live', Symbol='RN7SKP164', Aliases='', description='RNA, 7SK small nuclear pseudogene 164', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533705, CurrentID=0, Status='live', Symbol='SLC9B1P5', Aliases='', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1 pseudogene 5', other_designations='Na+/H+ exchanger domain containing 1 pseudogene|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1 pseudogene 5', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32388131, end_position_on_the_genomic_accession=32436365, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399947, CurrentID=0, Status='live', Symbol='C11orf87', Aliases='LOH11CR1A, NEURIM1', description='chromosome 11 open reading frame 87', other_designations='neuronal integral membrane protein 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109421728, end_position_on_the_genomic_accession=109429167, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480886, CurrentID=0, Status='live', Symbol='RN7SKP169', Aliases='', description='RNA, 7SK small nuclear pseudogene 169', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399948, CurrentID=0, Status='live', Symbol='COLCA1', Aliases='C11orf92, CASC12, LOH11CR1F', description='colorectal cancer associated 1', other_designations='cancer susceptibility candidate 12', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111293389, end_position_on_the_genomic_accession=111305048, orientation='minus', exon_count=6, OMIM=615693, no_hdr0=''),
  106480887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480887, CurrentID=0, Status='live', Symbol='RN7SKP171', Aliases='', description='RNA, 7SK small nuclear pseudogene 171', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874138, CurrentID=0, Status='live', Symbol='LINC00352', Aliases='', description='long intergenic non-protein coding RNA 352', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399949, CurrentID=0, Status='live', Symbol='C11orf88', Aliases='', description='chromosome 11 open reading frame 88', other_designations='hypothetical gene supported by BC039505', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111514785, end_position_on_the_genomic_accession=111537031, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  171472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171472, CurrentID=0, Status='live', Symbol='SRSF10P2', Aliases='FUSIP1P2, SFRS13AP2, dJ836E8.1, pFUSIP1', description='serine/arginine-rich splicing factor 10 pseudogene 2', other_designations='FUS interacting protein (serine/arginine-rich) 1 pseudogene 2|FUSIP1 pseudogene|TLS-associated protein TASR-2 pseudogene|splicing factor, arginine/serine-rich 13A pseudogene 2', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7831501, end_position_on_the_genomic_accession=7832481, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480888, CurrentID=0, Status='live', Symbol='RN7SKP180', Aliases='', description='RNA, 7SK small nuclear pseudogene 180', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399950, CurrentID=0, Status='live', Symbol='ST13P10', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 10', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112267296, end_position_on_the_genomic_accession=112268371, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480889, CurrentID=0, Status='live', Symbol='RN7SKP183', Aliases='', description='RNA, 7SK small nuclear pseudogene 183', other_designations='', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171482, CurrentID=0, Status='live', Symbol='FAM9A', Aliases='TEX39A', description='family with sequence similarity 9 member A', other_designations='family with sequence similarity 9, member A|testis expressed 39A', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8790774, end_position_on_the_genomic_accession=8801383, orientation='minus', exon_count=11, OMIM=300477, no_hdr0=''),
  171483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171483, CurrentID=0, Status='live', Symbol='FAM9B', Aliases='TEX39B', description='family with sequence similarity 9 member B', other_designations='family with sequence similarity 9, member B|testis expressed 39B', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9024232, end_position_on_the_genomic_accession=9197249, orientation='minus', exon_count=17, OMIM=300478, no_hdr0=''),
  171484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171484, CurrentID=0, Status='live', Symbol='FAM9C', Aliases='TEX39C', description='family with sequence similarity 9 member C', other_designations='family with sequence similarity 9, member C|testis expressed 39C', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13035617, end_position_on_the_genomic_accession=13044798, orientation='minus', exon_count=8, OMIM=300479, no_hdr0=''),
  106480890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480890, CurrentID=0, Status='live', Symbol='RN7SKP187', Aliases='', description='RNA, 7SK small nuclear pseudogene 187', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533706, CurrentID=0, Status='live', Symbol='LOC100533706', Aliases='', description='NEDD8 activating enzyme E1 subunit 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26045984, end_position_on_the_genomic_accession=26047722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480891, CurrentID=0, Status='live', Symbol='RN7SKP190', Aliases='', description='RNA, 7SK small nuclear pseudogene 190', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271835, CurrentID=0, Status='live', Symbol='LIMS3-LOC440895', Aliases='', description='LIMS3-LOC440895 readthrough', other_designations='LIM and senescent cell antigen-like domains 3-two pore channel 3 pseudogene read-through', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109898432, end_position_on_the_genomic_accession=109968574, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  106480892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480892, CurrentID=0, Status='live', Symbol='RN7SKP192', Aliases='', description='RNA, 7SK small nuclear pseudogene 192', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270979, CurrentID=0, Status='live', Symbol='RPL30P7', Aliases='RPL30_3_541', description='ribosomal protein L30 pseudogene 7', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10488752, end_position_on_the_genomic_accession=10489190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189437, CurrentID=0, Status='live', Symbol='TRY-GTA5-4', Aliases='TRNAY15', description='transfer RNA-Tyr (GTA) 5-4', other_designations='transfer RNA tyrosine 15 (anticodon GUA)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66113367, end_position_on_the_genomic_accession=66113459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723496, CurrentID=0, Status='live', Symbol='LOC102723496', Aliases='', description='uncharacterized LOC102723496', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480894, CurrentID=0, Status='live', Symbol='RN7SKP197', Aliases='', description='RNA, 7SK small nuclear pseudogene 197', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171511, CurrentID=0, Status='live', Symbol='ADIPQTL2', Aliases='CAQ5', description='Circulating adiponectin QTL on chromosome 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606770, no_hdr0=''),
  171512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171512, CurrentID=0, Status='live', Symbol='ADIPQTL3', Aliases='CAQ14', description='Circulating adiponectin QTL on chromosome 14', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606771, no_hdr0=''),
  171513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171513, CurrentID=0, Status='live', Symbol='PAOD1', Aliases='', description='Peripheral arterial occlusive disease 1', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606787, no_hdr0=''),
  171514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171514, CurrentID=0, Status='live', Symbol='ANON1', Aliases='', description='Anorexia nervosa, susceptibility to, 1', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606788, no_hdr0=''),
  171515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171515, CurrentID=0, Status='live', Symbol='HBFQTL4', Aliases='FHQTL', description='Fetal hemoglobin QTL on chromosome 8', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606789, no_hdr0=''),
  102723498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723498, CurrentID=0, Status='live', Symbol='CDC27P4', Aliases='', description='cell division cycle 27 pseudogene 4', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30492954, end_position_on_the_genomic_accession=30494887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533707, CurrentID=0, Status='live', Symbol='LOC100533707', Aliases='', description='ZFP37 zinc finger protein pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64834434, end_position_on_the_genomic_accession=64835647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480896, CurrentID=0, Status='live', Symbol='RN7SKP204', Aliases='', description='RNA, 7SK small nuclear pseudogene 204', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171523, CurrentID=0, Status='live', Symbol='CYP2T1P', Aliases='CYP2T2P', description='cytochrome P450 family 2 subfamily T member 1, pseudogene', other_designations='cytochrome P450, family 2, subfamily F, polypeptide 1 pseudogene|cytochrome P450, family 2, subfamily T, polypeptide 1, pseudogene|cytochrome P450, family 2, subfamily T, polypeptide 2, pseudogene|cytochrome P450, subfamily IIT, polypeptide 2 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40808495, end_position_on_the_genomic_accession=40811392, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271836, CurrentID=0, Status='live', Symbol='SMG1P3', Aliases='', description='SMG1 pseudogene 3', other_designations='smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21446683, end_position_on_the_genomic_accession=21502281, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  137814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137814, CurrentID=0, Status='live', Symbol='NKX2-6', Aliases='CSX2, CTHM, NKX2F, NKX4-2', description='NK2 homeobox 6', other_designations='NK2 transcription factor related, locus 6|homeobox protein NK2 homolog F|homeobox protein NKX2.6|tinman paralog', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23702451, end_position_on_the_genomic_accession=23706598, orientation='minus', exon_count=2, OMIM=611770, no_hdr0=''),
  106480897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480897, CurrentID=0, Status='live', Symbol='RN7SKP207', Aliases='', description='RNA, 7SK small nuclear pseudogene 207', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480898, CurrentID=0, Status='live', Symbol='RN7SKP210', Aliases='', description='RNA, 7SK small nuclear pseudogene 210', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101752334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101752334, CurrentID=0, Status='live', Symbol='PVRIG2P', Aliases='', description='poliovirus receptor related immunoglobulin domain containing 2, pseudogene', other_designations='poliovirus receptor related immunoglobulin domain containing pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100352318, end_position_on_the_genomic_accession=100353936, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480899, CurrentID=0, Status='live', Symbol='RN7SKP213', Aliases='', description='RNA, 7SK small nuclear pseudogene 213', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132172, CurrentID=0, Status='live', Symbol='LOC100132172', Aliases='', description='histone deacetylase 2 pseudogene', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114583262, end_position_on_the_genomic_accession=114584549, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480900, CurrentID=0, Status='live', Symbol='RN7SKP218', Aliases='', description='RNA, 7SK small nuclear pseudogene 218', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171546, CurrentID=0, Status='live', Symbol='SPTSSA', Aliases='C14orf147, SSSPTA', description='serine palmitoyltransferase small subunit A', other_designations='serine palmitoyltransferase, small subunit A|small subunit of serine palmitoyltransferase A', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34432938, end_position_on_the_genomic_accession=34462262, orientation='minus', exon_count=2, OMIM=613540, no_hdr0=''),
  552879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552879, CurrentID=0, Status='live', Symbol='VENTXP6', Aliases='', description='VENT homeobox pseudogene 6', other_designations='VENT-like homeobox pseudogene 6', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73651010, end_position_on_the_genomic_accession=73651941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533708, CurrentID=0, Status='live', Symbol='LOC100533708', Aliases='', description='protein inhibitor of activated STAT 2 pseudogene', other_designations='protein inhibitor of activated STAT, 2 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29465719, end_position_on_the_genomic_accession=29467758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480901, CurrentID=0, Status='live', Symbol='RN7SKP220', Aliases='', description='RNA, 7SK small nuclear pseudogene 220', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480902, CurrentID=0, Status='live', Symbol='RN7SKP224', Aliases='', description='RNA, 7SK small nuclear pseudogene 224', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  171558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171558, CurrentID=0, Status='live', Symbol='PTCRA', Aliases='PT-ALPHA, PTA', description='pre T-cell antigen receptor alpha', other_designations='pT-alpha-TCR|pre-T-cell receptor alpha chain', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42915807, end_position_on_the_genomic_accession=42925838, orientation='plus', exon_count=6, OMIM=606817, no_hdr0=''),
  171559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171559, CurrentID=0, Status='live', Symbol='HMGA1P2', Aliases='HMGA1L2, HMGIYL2', description='high mobility group AT-hook 1 pseudogene 2', other_designations='high mobility group AT-hook 1-like 2|high mobility group protein isoforms I and Y-like protein 2 pseudogene', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73098741, end_position_on_the_genomic_accession=73100506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  171560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171560, CurrentID=0, Status='live', Symbol='TXNP6', Aliases='', description='thioredoxin pseudogene 6', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76958565, end_position_on_the_genomic_accession=76959364, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480903, CurrentID=0, Status='live', Symbol='RN7SKP226', Aliases='', description='RNA, 7SK small nuclear pseudogene 226', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399965, CurrentID=0, Status='live', Symbol='KRT18P59', Aliases='', description='keratin 18 pseudogene 59', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125113554, end_position_on_the_genomic_accession=125114919, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  171568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171568, CurrentID=0, Status='live', Symbol='POLR3H', Aliases='RPC22.9, RPC8', description='polymerase (RNA) III (DNA directed) polypeptide H (22.9kD)', other_designations='DNA-directed RNA polymerase III subunit 22.9 kDa polypeptide|DNA-directed RNA polymerase III subunit H|RNA nucleotidyltransferase (DNA-directed)|RNA polymerase III subunit C8', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41525799, end_position_on_the_genomic_accession=41544606, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106480904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480904, CurrentID=0, Status='live', Symbol='RN7SKP229', Aliases='', description='RNA, 7SK small nuclear pseudogene 229', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23171, CurrentID=0, Status='live', Symbol='GPD1L', Aliases='GPD1-L', description='glycerol-3-phosphate dehydrogenase 1-like', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32106511, end_position_on_the_genomic_accession=32168715, orientation='plus', exon_count=8, OMIM=611778, no_hdr0=''),
  106480905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480905, CurrentID=0, Status='live', Symbol='RN7SKP232', Aliases='', description='RNA, 7SK small nuclear pseudogene 232', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533709, CurrentID=0, Status='live', Symbol='LOC100533709', Aliases='', description='piggyBac transposable element derived 2 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=116818272, end_position_on_the_genomic_accession=116819697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399967, CurrentID=0, Status='live', Symbol='PATE2', Aliases='C11orf38, PATE-M, UNQ3112', description='prostate and testis expressed 2', other_designations='LVLF3112|PATE-like protein M|prostate and testis expression M', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125772311, end_position_on_the_genomic_accession=125778834, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480906, CurrentID=0, Status='live', Symbol='RN7SKP236', Aliases='', description='RNA, 7SK small nuclear pseudogene 236', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271838, CurrentID=0, Status='live', Symbol='IBD26', Aliases='', description='Inflammatory bowel disease-26', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612639, no_hdr0=''),
  399968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399968, CurrentID=0, Status='live', Symbol='PATE4', Aliases='PATE-B', description='prostate and testis expressed 4', other_designations='PATE-like protein B|prostate and testis expression B', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125833316, end_position_on_the_genomic_accession=125840072, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171586, CurrentID=0, Status='live', Symbol='ABHD3', Aliases='LABH3', description='abhydrolase domain containing 3', other_designations='abhydrolase domain-containing protein 3|alpha/beta hydrolase domain containing protein 3|lung alpha/beta hydrolase 3', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21650897, end_position_on_the_genomic_accession=21704805, orientation='minus', exon_count=11, OMIM=612197, no_hdr0=''),
  106480907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480907, CurrentID=0, Status='live', Symbol='RN7SKP239', Aliases='', description='RNA, 7SK small nuclear pseudogene 239', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723510, CurrentID=0, Status='live', Symbol='LCMT1-AS1', Aliases='', description='LCMT1 antisense RNA 1', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25100564, end_position_on_the_genomic_accession=25111555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480908, CurrentID=0, Status='live', Symbol='RN7SKP242', Aliases='', description='RNA, 7SK small nuclear pseudogene 242', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285043, CurrentID=0, Status='live', Symbol='LOC285043', Aliases='', description='uncharacterized LOC285043', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30346646, end_position_on_the_genomic_accession=30350158, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101752399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101752399, CurrentID=0, Status='live', Symbol='STAG3L5P-PVRIG2P-PILRB', Aliases='LOC101735302-LOC101752334-PILRB', description='STAG3L5P-PVRIG2P-PILRB readthrough', other_designations='LOC101735302-LOC101752334-PILRB readthrough', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100336065, end_position_on_the_genomic_accession=100367831, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  101752400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101752400, CurrentID=0, Status='live', Symbol='CAPN10-AS1', Aliases='locus959', description='CAPN10 antisense RNA 1 (head to head)', other_designations='locus 959', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240582750, end_position_on_the_genomic_accession=240586316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132174, CurrentID=0, Status='live', Symbol='LOC100132174', Aliases='', description='uncharacterized LOC100132174', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77722872, end_position_on_the_genomic_accession=77728559, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480910, CurrentID=0, Status='live', Symbol='RN7SKP249', Aliases='', description='RNA, 7SK small nuclear pseudogene 249', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28601, CurrentID=0, Status='live', Symbol='TRBV6-6', Aliases='TCRBV13S6A2T, TCRBV6S6, TRBV66', description='T cell receptor beta variable 6-6', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142469581, end_position_on_the_genomic_accession=142470013, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533710, CurrentID=0, Status='live', Symbol='LOC100533710', Aliases='', description='mannosidase, alpha, class 1B, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43063631, end_position_on_the_genomic_accession=43064856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399972, CurrentID=0, Status='live', Symbol='ST3GAL4-AS1', Aliases='DCPS-AS1', description='ST3GAL4 antisense RNA 1 (head to head)', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126341716, end_position_on_the_genomic_accession=126355587, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480911, CurrentID=0, Status='live', Symbol='RN7SKP256', Aliases='', description='RNA, 7SK small nuclear pseudogene 256', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94138, CurrentID=0, Status='live', Symbol='DFNA32', Aliases='', description='deafness, autosomal dominant 32', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480912, CurrentID=0, Status='live', Symbol='RN7SKP258', Aliases='', description='RNA, 7SK small nuclear pseudogene 258', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28603, CurrentID=0, Status='live', Symbol='TRBV6-4', Aliases='TCRBV13S5, TCRBV6S4, TRBV64', description='T cell receptor beta variable 6-4', other_designations='V_segment translation product', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142380827, end_position_on_the_genomic_accession=142381261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480913, CurrentID=0, Status='live', Symbol='RN7SKP263', Aliases='', description='RNA, 7SK small nuclear pseudogene 263', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723516, CurrentID=0, Status='live', Symbol='LOC102723516', Aliases='', description='uncharacterized LOC102723516', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38951673, end_position_on_the_genomic_accession=38973826, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  399975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399975, CurrentID=0, Status='live', Symbol='FLJ45950', Aliases='', description='FLJ45950 protein', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480914, CurrentID=0, Status='live', Symbol='RN7SKP265', Aliases='', description='RNA, 7SK small nuclear pseudogene 265', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480915, CurrentID=0, Status='live', Symbol='RN7SKP269', Aliases='', description='RNA, 7SK small nuclear pseudogene 269', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373964, CurrentID=0, Status='live', Symbol='LOC105373964', Aliases='', description='uncharacterized LOC105373964', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238692527, end_position_on_the_genomic_accession=238697520, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480916, CurrentID=0, Status='live', Symbol='RN7SKP271', Aliases='', description='RNA, 7SK small nuclear pseudogene 271', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480917, CurrentID=0, Status='live', Symbol='RN7SKP275', Aliases='', description='RNA, 7SK small nuclear pseudogene 275', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399979, CurrentID=0, Status='live', Symbol='SNX19', Aliases='CHET8', description='sorting nexin 19', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130875871, end_position_on_the_genomic_accession=130916487, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  106480918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480918, CurrentID=0, Status='live', Symbol='RN7SKP278', Aliases='', description='RNA, 7SK small nuclear pseudogene 278', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285045, CurrentID=0, Status='live', Symbol='LINC00486', Aliases='', description='long intergenic non-protein coding RNA 486', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32825443, end_position_on_the_genomic_accession=32937203, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  106480919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480919, CurrentID=0, Status='live', Symbol='RN7SKP282', Aliases='', description='RNA, 7SK small nuclear pseudogene 282', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  341784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341784, CurrentID=0, Status='live', Symbol='EIF4A1P7', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 7', other_designations='eukaryotic initiation factor 4AI pseudogene', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28598817, end_position_on_the_genomic_accession=28600204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480920, CurrentID=0, Status='live', Symbol='RN7SKP285', Aliases='', description='RNA, 7SK small nuclear pseudogene 285', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533712, CurrentID=0, Status='live', Symbol='LOC100533712', Aliases='', description='MDM4, p53 regulator pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17933615, end_position_on_the_genomic_accession=17934820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480921, CurrentID=0, Status='live', Symbol='RN7SKP287', Aliases='', description='RNA, 7SK small nuclear pseudogene 287', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  153684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153684, CurrentID=0, Status='live', Symbol='LOC153684', Aliases='', description='uncharacterized LOC153684', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43042134, end_position_on_the_genomic_accession=43045268, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  28612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28612, CurrentID=0, Status='live', Symbol='TRBV5-3', Aliases='TCRBV5S3, TCRBV5S5P, TRBV53', description='T cell receptor beta variable 5-3 (non-functional)', other_designations='T cell receptor beta variable 5-3 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142389202, end_position_on_the_genomic_accession=142389668, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480922, CurrentID=0, Status='live', Symbol='RN7SKP291', Aliases='', description='RNA, 7SK small nuclear pseudogene 291', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723525, CurrentID=0, Status='live', Symbol='LOC102723525', Aliases='', description='uncharacterized LOC102723525', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10737575, end_position_on_the_genomic_accession=10808919, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106480923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480923, CurrentID=0, Status='live', Symbol='RN7SKP293', Aliases='', description='RNA, 7SK small nuclear pseudogene 293', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94150, CurrentID=0, Status='live', Symbol='CDRT7', Aliases='LINC00025, NCRNA00025', description='CMT1A duplicated region transcript 7 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 25', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15030975, end_position_on_the_genomic_accession=15031957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480924, CurrentID=0, Status='live', Symbol='RN7SKP295', Aliases='', description='RNA, 7SK small nuclear pseudogene 295', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94151, CurrentID=0, Status='live', Symbol='CDRT8', Aliases='', description='CMT1A duplicated region transcript 8', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15104985, end_position_on_the_genomic_accession=15106187, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480925, CurrentID=0, Status='live', Symbol='RN7SKP299', Aliases='', description='RNA, 7SK small nuclear pseudogene 299', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533713, CurrentID=0, Status='live', Symbol='ELL2P3', Aliases='', description='elongation factor, RNA polymerase II, 2 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68289709, end_position_on_the_genomic_accession=68291030, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480926, CurrentID=0, Status='live', Symbol='RN7SL12P', Aliases='', description='RNA, 7SL, cytoplasmic 12, pseudogene', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723529, CurrentID=0, Status='live', Symbol='LOC102723529', Aliases='', description='uncharacterized LOC102723529', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203372931, end_position_on_the_genomic_accession=203398513, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  399988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399988, CurrentID=0, Status='live', Symbol='HSPA8P5', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 5', other_designations='heat shock 70kDa protein 8 pseudogene 5', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4097373, end_position_on_the_genomic_accession=4099720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480927, CurrentID=0, Status='live', Symbol='RN7SL18P', Aliases='', description='RNA, 7SL, cytoplasmic 18, pseudogene', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723530, CurrentID=0, Status='live', Symbol='LOC102723530', Aliases='', description='uncharacterized LOC102723530', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28448520, end_position_on_the_genomic_accession=28449989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874146, CurrentID=0, Status='live', Symbol='LINC00365', Aliases='', description='long intergenic non-protein coding RNA 365', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30103181, end_position_on_the_genomic_accession=30107930, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480928, CurrentID=0, Status='live', Symbol='RN7SL22P', Aliases='', description='RNA, 7SL, cytoplasmic 22, pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723531, CurrentID=0, Status='live', Symbol='LOC102723531', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399990, CurrentID=0, Status='live', Symbol='FLJ44874', Aliases='', description='FLJ44874 protein', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480929, CurrentID=0, Status='live', Symbol='RN7SL28P', Aliases='', description='RNA, 7SL, cytoplasmic 28, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723532, CurrentID=0, Status='live', Symbol='LOC102723532', Aliases='', description='olfactory receptor 4N4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480930, CurrentID=0, Status='live', Symbol='RN7SL32P', Aliases='', description='RNA, 7SL, cytoplasmic 32, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533714, CurrentID=0, Status='live', Symbol='DDX43P2', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 43 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=49258518, end_position_on_the_genomic_accession=49259780, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480931, CurrentID=0, Status='live', Symbol='RN7SL35P', Aliases='', description='RNA, 7SL, cytoplasmic 35, pseudogene', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723534, CurrentID=0, Status='live', Symbol='LOC102723534', Aliases='', description='E3 ubiquitin-protein ligase HERC2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23289204, end_position_on_the_genomic_accession=23303537, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480932, CurrentID=0, Status='live', Symbol='RN7SL38P', Aliases='', description='RNA, 7SL, cytoplasmic 38, pseudogene', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874147, CurrentID=0, Status='live', Symbol='LINC00366', Aliases='', description='long intergenic non-protein coding RNA 366', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38567724, end_position_on_the_genomic_accession=38579516, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480933, CurrentID=0, Status='live', Symbol='RN7SL41P', Aliases='', description='RNA, 7SL, cytoplasmic 41, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723536, CurrentID=0, Status='live', Symbol='LOC102723536', Aliases='', description='uncharacterized LOC102723536', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26318131, end_position_on_the_genomic_accession=26364057, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100310754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310754, CurrentID=0, Status='live', Symbol='DUPXP11.23P11.22', Aliases='CXDUPp11.23p11.22', description='Chromosome Xp11.23-p11.22 duplication syndrome', other_designations='', map_location='Xp11.23-p11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300801, no_hdr0=''),
  100310755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310755, CurrentID=0, Status='live', Symbol='ATOD7', Aliases='', description='Dermatitis, atopic, susceptibility to, 7', other_designations='', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613064, no_hdr0=''),
  100310756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310756, CurrentID=0, Status='live', Symbol='LOC100310756', Aliases='', description='uncharacterized LOC100310756', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480934, CurrentID=0, Status='live', Symbol='RN7SL44P', Aliases='', description='RNA, 7SL, cytoplasmic 44, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480935, CurrentID=0, Status='live', Symbol='RN7SL49P', Aliases='', description='RNA, 7SL, cytoplasmic 49, pseudogene', other_designations='', map_location='13q14.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94162, CurrentID=0, Status='live', Symbol='SNORD38A', Aliases='RNU38A, U38A', description='small nucleolar RNA, C/D box 38A', other_designations='RNA, U38A small nucleolar|U38A small nucleolar RNA|U38A snoRNA', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44777842, end_position_on_the_genomic_accession=44777912, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480936, CurrentID=0, Status='live', Symbol='RN7SL51P', Aliases='', description='RNA, 7SL, cytoplasmic 51, pseudogene', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036563, CurrentID=0, Status='live', Symbol='SNORD115-24', Aliases='HBII-52-24', description='small nucleolar RNA, C/D box 115-24', other_designations='HBII-52-24 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25213659, end_position_on_the_genomic_accession=25213729, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480937, CurrentID=0, Status='live', Symbol='RN7SL52P', Aliases='', description='RNA, 7SL, cytoplasmic 52, pseudogene', other_designations='', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036564, CurrentID=0, Status='live', Symbol='SNORD115-27', Aliases='HBII-52-27', description='small nucleolar RNA, C/D box 115-27', other_designations='HBII-52-27 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25220503, end_position_on_the_genomic_accession=25220578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103456506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456506, CurrentID=0, Status='live', Symbol='TRE-TTC4-1', Aliases='TRNAE44', description='transfer RNA-Glu (TTC) 4-1', other_designations='tRNA-Glu-TTC-4-1|transfer RNA glutamic acid 44 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103456507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456507, CurrentID=0, Status='live', Symbol='LINC01442', Aliases='', description='long intergenic non-protein coding RNA 1442', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103456508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456508, CurrentID=0, Status='live', Symbol='TRC-GCA7-1', Aliases='TRNAC33', description='transfer RNA-Cys (GCA) 7-1', other_designations='tRNA-Cys-GCA-7-1|transfer RNA cysteine 33 (anticodon GCA)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103456509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456509, CurrentID=0, Status='live', Symbol='TRN-GTT18-1', Aliases='', description='transfer RNA-Asn (GTT) 18-1', other_designations='tRNA-Asn-GTT-18-1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103456510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456510, CurrentID=0, Status='live', Symbol='TRV-CAC11-1', Aliases='TRNAV39', description='transfer RNA-Val (CAC) 11-1', other_designations='tRNA-Val-CAC-11-1|transfer RNA valine 39 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036565, CurrentID=0, Status='live', Symbol='SNORD115-28', Aliases='HBII-52-28', description='small nucleolar RNA, C/D box 115-28', other_designations='HBII-52-28 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25222354, end_position_on_the_genomic_accession=25222427, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100310785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310785, CurrentID=0, Status='live', Symbol='ALL1', Aliases='', description='Leukemia, acute lymphocytic, susceptibility to, 1', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613065, no_hdr0=''),
  100310786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310786, CurrentID=0, Status='live', Symbol='ALL2', Aliases='', description='Leukemia, acute lymphoblastic, susceptibility to, 2', other_designations='', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613067, no_hdr0=''),
  106480939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480939, CurrentID=0, Status='live', Symbol='RN7SL63P', Aliases='', description='RNA, 7SL, cytoplasmic 63, pseudogene', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036566, CurrentID=0, Status='live', Symbol='SNORD115-45', Aliases='HBII-52-45, SNORD115-48', description='small nucleolar RNA, C/D box 115-45', other_designations='HBII-52-45 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25264527, end_position_on_the_genomic_accession=25264579, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  727681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727681, CurrentID=0, Status='live', Symbol='LOC727681', Aliases='', description='NSL1, MIND kinetochore complex component, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74360284, end_position_on_the_genomic_accession=74361514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480940, CurrentID=0, Status='live', Symbol='RN7SL69P', Aliases='', description='RNA, 7SL, cytoplasmic 69, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036567, CurrentID=0, Status='live', Symbol='SNORD115-47', Aliases='HBII-52-46, HBII-52-47', description='small nucleolar RNA, C/D box 115-47', other_designations='HBII-52-46 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25268517, end_position_on_the_genomic_accession=25268549, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373969, CurrentID=0, Status='live', Symbol='LOC105373969', Aliases='', description='uncharacterized LOC105373969', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240857207, end_position_on_the_genomic_accession=240870637, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400002, CurrentID=0, Status='live', Symbol='LOC400002', Aliases='', description='uncharacterized FLJ46363', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9948137, end_position_on_the_genomic_accession=9953336, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480941, CurrentID=0, Status='live', Symbol='RN7SL72P', Aliases='', description='RNA, 7SL, cytoplasmic 72, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480942, CurrentID=0, Status='live', Symbol='RN7SL74P', Aliases='', description='RNA, 7SL, cytoplasmic 74, pseudogene', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480943, CurrentID=0, Status='live', Symbol='RN7SL79P', Aliases='', description='RNA, 7SL, cytoplasmic 79, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100310812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310812, CurrentID=0, Status='live', Symbol='SPDYE2B', Aliases='SPDYE2L', description='speedy/RINGO cell cycle regulator family member E2B', other_designations='WBSCR19-like protein 3|putative WBSCR19-like protein 3|putative speedy protein E2-like protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102650325, end_position_on_the_genomic_accession=102661400, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106480944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480944, CurrentID=0, Status='live', Symbol='RN7SL82P', Aliases='', description='RNA, 7SL, cytoplasmic 82, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480945, CurrentID=0, Status='live', Symbol='RN7SL88P', Aliases='', description='RNA, 7SL, cytoplasmic 88, pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480946, CurrentID=0, Status='live', Symbol='RN7SL91P', Aliases='', description='RNA, 7SL, cytoplasmic 91, pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379834, CurrentID=0, Status='live', Symbol='LOC105379834', Aliases='', description='uncharacterized LOC105379834', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480947, CurrentID=0, Status='live', Symbol='RN7SL93P', Aliases='', description='RNA, 7SL, cytoplasmic 93, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036574, CurrentID=0, Status='live', Symbol='HMGN1P3', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 3', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 3', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68886982, end_position_on_the_genomic_accession=68888183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310838, CurrentID=0, Status='live', Symbol='VDAC1P3', Aliases='', description='voltage-dependent anion channel 1 pseudogene 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91981804, end_position_on_the_genomic_accession=91983514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310839, CurrentID=0, Status='live', Symbol='VDAC1P7', Aliases='', description='voltage-dependent anion channel 1 pseudogene 7', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=77315896, end_position_on_the_genomic_accession=77317616, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310840, CurrentID=0, Status='live', Symbol='VDAC1P8', Aliases='', description='voltage-dependent anion channel 1 pseudogene 8', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143490508, end_position_on_the_genomic_accession=143492198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310841, CurrentID=0, Status='live', Symbol='VDAC1P11', Aliases='', description='voltage-dependent anion channel 1 pseudogene 11', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94286532, end_position_on_the_genomic_accession=94288555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310842, CurrentID=0, Status='live', Symbol='RBPJP1', Aliases='IGKJRBP1, RBPSUHP1', description='RBPJ pseudogene 1', other_designations='immunoglobulin kappa J region recombination signal binding protein pseudogene 1|recombining binding protein suppressor of hairless pseudogene 1', map_location='9p13-p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100036575, CurrentID=0, Status='live', Symbol='HMGN1P4', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 4', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 4', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182941306, end_position_on_the_genomic_accession=182942522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310846, CurrentID=0, Status='live', Symbol='ANKRD61', Aliases='', description='ankyrin repeat domain 61', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6031376, end_position_on_the_genomic_accession=6036552, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189493, CurrentID=0, Status='live', Symbol='TRL-AAG7-1', Aliases='TRNAL43P', description='transfer RNA-Leu (AAG) 7-1', other_designations='tRNA-Leu-AAG-7-1|transfer RNA leucine 43 (anticodon AAG) pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181164564, end_position_on_the_genomic_accession=181164645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310848, CurrentID=0, Status='live', Symbol='SNRPCP4', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 4', other_designations='', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2990650, end_position_on_the_genomic_accession=2991312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310849, CurrentID=0, Status='live', Symbol='SNRPCP19', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 19', other_designations='', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24906042, end_position_on_the_genomic_accession=24906602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310850, CurrentID=0, Status='live', Symbol='SNRPCP11', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 11', other_designations='', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125815508, end_position_on_the_genomic_accession=125816287, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310851, CurrentID=0, Status='live', Symbol='SNRPCP8', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 8', other_designations='', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130199711, end_position_on_the_genomic_accession=130200493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310852, CurrentID=0, Status='live', Symbol='SNRPCP7', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 7', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8340188, end_position_on_the_genomic_accession=8340967, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310853, CurrentID=0, Status='live', Symbol='SNRPCP9', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 9', other_designations='', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97885868, end_position_on_the_genomic_accession=97886659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310854, CurrentID=0, Status='live', Symbol='SNRPCP12', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 12', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67905397, end_position_on_the_genomic_accession=67906178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100310855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310855, CurrentID=0, Status='live', Symbol='SNRPCP14', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 14', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71689872, end_position_on_the_genomic_accession=71690653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310856, CurrentID=0, Status='live', Symbol='SNRPCP6', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 6', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7143059, end_position_on_the_genomic_accession=7143841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310857, CurrentID=0, Status='live', Symbol='SNRPCP15', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 15', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7170534, end_position_on_the_genomic_accession=7171318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189495, CurrentID=0, Status='live', Symbol='TRD-GTC10-1', Aliases='TRNAD20P', description='transfer RNA-Asp (GTC) 10-1', other_designations='tRNA-Asp-GTC-10-1|transfer RNA aspartic acid 20 (anticodon GUC) pseudogene', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161523145, end_position_on_the_genomic_accession=161523216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373200, CurrentID=0, Status='live', Symbol='LOC105373200', Aliases='', description='G antigen 2D-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189496, CurrentID=0, Status='live', Symbol='TRY-GTA12-1', Aliases='TRNAY16P', description='transfer RNA-Tyr (GTA) 12-1', other_designations='tRNA-Tyr-GTA-12-1|transfer RNA tyrosine 16 (anticodon GUA) pseudogene', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91711072, end_position_on_the_genomic_accession=91711144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100310869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310869, CurrentID=0, Status='live', Symbol='IGKV2OR2-7D', Aliases='', description='immunoglobulin kappa variable 2/OR2-7D (pseudogene)', other_designations='IGKV2/OR2-7D|immunoglobulin kappa variable 2/OR2-7D pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97335620, end_position_on_the_genomic_accession=97336659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189497, CurrentID=0, Status='live', Symbol='TRF-GAA10-1', Aliases='TRNAF16P', description='transfer RNA-Phe (GAA) 10-1', other_designations='tRNA-Phe-GAA-10-1|transfer RNA phenylalanine 16 (anticodon GAA) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121009444, end_position_on_the_genomic_accession=121009715, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480954, CurrentID=0, Status='live', Symbol='RN7SL122P', Aliases='', description='RNA, 7SL, cytoplasmic 122, pseudogene', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28645, CurrentID=0, Status='live', Symbol='TRAV37', Aliases='TCRAV37S1', description='T cell receptor alpha variable 37 (pseudogene)', other_designations='T cell receptor alpha variable 37 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22265749, end_position_on_the_genomic_accession=22267173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480955, CurrentID=0, Status='live', Symbol='RN7SL124P', Aliases='', description='RNA, 7SL, cytoplasmic 124, pseudogene', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533719, CurrentID=0, Status='live', Symbol='ANKRD20A14P', Aliases='', description='ankyrin repeat domain 20 family member A14, pseudogene', other_designations='ankyrin repeat domain 20 family, member A14, pseudogene|ankyrin repeat domain 20 family, member A2 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480956, CurrentID=0, Status='live', Symbol='RN7SL127P', Aliases='', description='RNA, 7SL, cytoplasmic 127, pseudogene', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189501, CurrentID=0, Status='live', Symbol='NMTRQ-TTG11-1', Aliases='TRNAQ43P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 11-1', other_designations='nmt-tRNA-Gln-TTG-11-1|transfer RNA glutamine 43 (anticodon UUG) pseudogene|transfer RNA-Gln (TTG) 11-1', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33015128, end_position_on_the_genomic_accession=33015199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480958, CurrentID=0, Status='live', Symbol='RN7SL135P', Aliases='', description='RNA, 7SL, cytoplasmic 135, pseudogene', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285053, CurrentID=0, Status='live', Symbol='RPL18AP6', Aliases='RPL18A_1_215', description='ribosomal protein L18a pseudogene 6', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47690667, end_position_on_the_genomic_accession=47691283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480959, CurrentID=0, Status='live', Symbol='RN7SL141P', Aliases='', description='RNA, 7SL, cytoplasmic 141, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727701, CurrentID=0, Status='live', Symbol='LINC00165', Aliases='C21orf135, NCRNA00165, NLC1-B, NLC1B', description='long intergenic non-protein coding RNA 165', other_designations='narcolepsy candidate region gene 1B|narcolepsy candidate-region 1 gene B', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480960, CurrentID=0, Status='live', Symbol='RN7SL144P', Aliases='', description='RNA, 7SL, cytoplasmic 144, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28651, CurrentID=0, Status='live', Symbol='TRAV31', Aliases='TCRAV31S1', description='T cell receptor alpha variable 31 (pseudogene)', other_designations='T cell receptor alpha variable 31 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22177273, end_position_on_the_genomic_accession=22177864, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533720, CurrentID=0, Status='live', Symbol='ANKRD20A15P', Aliases='', description='ankyrin repeat domain 20 family member A15, pseudogene', other_designations='ankyrin repeat domain 20 family, member A15, pseudogene|ankyrin repeat domain 20 family, member A2 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480961, CurrentID=0, Status='live', Symbol='RN7SL149P', Aliases='', description='RNA, 7SL, cytoplasmic 149, pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723564, CurrentID=0, Status='live', Symbol='LOC102723564', Aliases='', description='uncharacterized LOC102723564', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23303717, end_position_on_the_genomic_accession=23313129, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106480962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480962, CurrentID=0, Status='live', Symbol='RN7SL152P', Aliases='', description='RNA, 7SL, cytoplasmic 152, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874153, CurrentID=0, Status='live', Symbol='LINC00388', Aliases='', description='long intergenic non-protein coding RNA 388', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480963, CurrentID=0, Status='live', Symbol='RN7SL156P', Aliases='', description='RNA, 7SL, cytoplasmic 156, pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723566, CurrentID=0, Status='live', Symbol='LOC102723566', Aliases='', description='uncharacterized LOC102723566', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127816066, end_position_on_the_genomic_accession=127817166, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100189508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189508, CurrentID=0, Status='live', Symbol='TRC-ACA1-1', Aliases='TRNAC32P', description='transfer RNA-Cys (ACA) 1-1', other_designations='transfer RNA cysteine 32 (anticodon ACA) pseudogene', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152609036, end_position_on_the_genomic_accession=152609211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723567, CurrentID=0, Status='live', Symbol='LOC102723567', Aliases='', description='uncharacterized LOC102723567', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74056963, end_position_on_the_genomic_accession=74112891, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  400026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400026, CurrentID=0, Status='live', Symbol='LOC400026', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43569746, end_position_on_the_genomic_accession=43572171, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480965, CurrentID=0, Status='live', Symbol='RN7SL164P', Aliases='', description='RNA, 7SL, cytoplasmic 164, pseudogene', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533721, CurrentID=0, Status='live', Symbol='PNPT1P2', Aliases='', description='polyribonucleotide nucleotidyltransferase 1 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121843448, end_position_on_the_genomic_accession=121845212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  400027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400027, CurrentID=0, Status='live', Symbol='LINC00938', Aliases='', description='long intergenic non-protein coding RNA 938', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45725720, end_position_on_the_genomic_accession=45727921, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480966, CurrentID=0, Status='live', Symbol='RN7SL167P', Aliases='', description='RNA, 7SL, cytoplasmic 167, pseudogene', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28657, CurrentID=0, Status='live', Symbol='TRAV26-1', Aliases='TCRAV26S1, TCRAV4S2, TRAV261', description='T cell receptor alpha variable 26-1', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22123527, end_position_on_the_genomic_accession=22124285, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480967, CurrentID=0, Status='live', Symbol='RN7SL172P', Aliases='', description='RNA, 7SL, cytoplasmic 172, pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270982, CurrentID=0, Status='live', Symbol='RPL30P10', Aliases='RPL30_5_906', description='ribosomal protein L30 pseudogene 10', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57392546, end_position_on_the_genomic_accession=57392882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480968, CurrentID=0, Status='live', Symbol='RN7SL176P', Aliases='', description='RNA, 7SL, cytoplasmic 176, pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137886, CurrentID=0, Status='live', Symbol='UBXN2B', Aliases='p37', description='UBX domain protein 2B', other_designations='NSFL1 cofactor p37|p97 cofactor p37', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58411264, end_position_on_the_genomic_accession=58451501, orientation='plus', exon_count=9, OMIM=610686, no_hdr0=''),
  106480969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480969, CurrentID=0, Status='live', Symbol='RN7SL177P', Aliases='', description='RNA, 7SL, cytoplasmic 177, pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189514, CurrentID=0, Status='live', Symbol='TRK-TTT14-1', Aliases='TRNAK40P', description='transfer RNA-Lys (TTT) 14-1', other_designations='tRNA-Lys-TTT-14-1|transfer RNA lysine 40 (anticodon UUU) pseudogene', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73588825, end_position_on_the_genomic_accession=73588897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533722, CurrentID=0, Status='live', Symbol='LOC100533722', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95350164, end_position_on_the_genomic_accession=95351767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189515, CurrentID=0, Status='live', Symbol='TRP-TGG5-1', Aliases='TRNAP27P', description='transfer RNA-Pro (TGG) 5-1', other_designations='tRNA-Pro-TGG-5-1|transfer RNA proline 27 (anticodon UGG) pseudogene', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207004809, end_position_on_the_genomic_accession=207004885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480559, CurrentID=0, Status='live', Symbol='RNU6-181P', Aliases='', description='RNA, U6 small nuclear 181, pseudogene', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  127933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127933, CurrentID=0, Status='live', Symbol='UHMK1', Aliases='KIS, KIST, P-CIP2', description='U2AF homology motif (UHM) kinase 1', other_designations='KIS protein kinase|PAM COOH-terminal interactor protein 2|kinase interacting with leukemia-associated gene (stathmin)', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162497174, end_position_on_the_genomic_accession=162529629, orientation='plus', exon_count=9, OMIM=608849, no_hdr0=''),
  100189517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189517, CurrentID=0, Status='live', Symbol='TRN-GTT10-1', Aliases='TRNAN35, TRNAN35P', description='transfer RNA-Asn (GTT) 10-1', other_designations='tRNA-Asn-GTT-10-1|transfer RNA asparagine 35 (anticodon GUU) pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=143663558, end_position_on_the_genomic_accession=143663631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420245, CurrentID=0, Status='live', Symbol='LOC100420245', Aliases='', description='MAGE family member B3 pseudogene', other_designations='melanoma antigen family B, 3 pseudogene|melanoma antigen family B3 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26170258, end_position_on_the_genomic_accession=26170904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727715, CurrentID=0, Status='live', Symbol='SCAX3', Aliases='', description='spinocerebellar ataxia, X-linked 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301790, no_hdr0=''),
  106480974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480974, CurrentID=0, Status='live', Symbol='RN7SL192P', Aliases='', description='RNA, 7SL, cytoplasmic 192, pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400036, CurrentID=0, Status='live', Symbol='LOC400036', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 23 pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52730370, end_position_on_the_genomic_accession=52733062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189519, CurrentID=0, Status='live', Symbol='TRK-TTT17-1', Aliases='TRNAK41P', description='transfer RNA-Lys (TTT) 17-1', other_designations='tRNA-Lys-TTT-17-1|transfer RNA lysine 41 (anticodon UUU) pseudogene', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18262694, end_position_on_the_genomic_accession=18262763, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533723, CurrentID=0, Status='live', Symbol='LOC100533723', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3851224, end_position_on_the_genomic_accession=3852869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189520, CurrentID=0, Status='live', Symbol='NMTRL-TAA3-1', Aliases='TRNAL45P', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 3-1', other_designations='nmt-tRNA-Leu-TAA-3-1|transfer RNA leucine 45 (anticodon UAA) pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237943656, end_position_on_the_genomic_accession=237943730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28667, CurrentID=0, Status='live', Symbol='TRAV16', Aliases='TCRAV16S1, TCRAV9S1', description='T cell receptor alpha variable 16', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21990496, end_position_on_the_genomic_accession=21990938, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728811, CurrentID=0, Status='live', Symbol='LOC728811', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B15 pseudogene', other_designations='UDP glucuronosyltransferase 2 family, polypeptide B10 pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68704377, end_position_on_the_genomic_accession=68712500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189522, CurrentID=0, Status='live', Symbol='NMTRL-TAA2-1', Aliases='TRNAL46P', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 2-1', other_designations='nmt-tRNA-Leu-TAA-2-1|transfer RNA leucine 46 (anticodon UAA) pseudogene', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117023841, end_position_on_the_genomic_accession=117023915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189523, CurrentID=0, Status='live', Symbol='TRQ-CTG14-1', Aliases='TRNAQ55', description='transfer RNA-Gln (CTG) 14-1', other_designations='tRNA-Gln-CTG-14-1|transfer RNA glutamine 55 (anticodon CUG)', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=17180900, end_position_on_the_genomic_accession=17180971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132188, CurrentID=0, Status='live', Symbol='LOC100132188', Aliases='', description='LP7097', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=165993897, end_position_on_the_genomic_accession=165994448, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100189524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189524, CurrentID=0, Status='live', Symbol='NMTRQ-TTG10-1', Aliases='TRNAQ50P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 10-1', other_designations='nmt-tRNA-Gln-TTG-10-1|transfer RNA glutamine 50 (anticodon UUG) pseudogene', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36065610, end_position_on_the_genomic_accession=36065681, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480981, CurrentID=0, Status='live', Symbol='RN7SL218P', Aliases='', description='RNA, 7SL, cytoplasmic 218, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400043, CurrentID=0, Status='live', Symbol='LOC400043', Aliases='', description='uncharacterized LOC400043', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54126071, end_position_on_the_genomic_accession=54132843, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480982, CurrentID=0, Status='live', Symbol='RN7SL221P', Aliases='', description='RNA, 7SL, cytoplasmic 221, pseudogene', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100048903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048903, CurrentID=0, Status='live', Symbol='MCCD1P1', Aliases='', description='mitochondrial coiled-coil domain 1 pseudogene 1', other_designations='', map_location='6p22.1-p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29907783, end_position_on_the_genomic_accession=29908645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100048904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048904, CurrentID=0, Status='live', Symbol='DDX39BP1', Aliases='BAT1P1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39B pseudogene 1', other_designations='HLA-B associated transcript 1 pseudogene 1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29906547, end_position_on_the_genomic_accession=29907449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100048905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048905, CurrentID=0, Status='live', Symbol='GBD2', Aliases='', description='gallbladder disease 2', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609918, no_hdr0=''),
  100048906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048906, CurrentID=0, Status='live', Symbol='GBD3', Aliases='', description='gallbladder disease 3', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609919, no_hdr0=''),
  106480983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480983, CurrentID=0, Status='live', Symbol='RN7SL228P', Aliases='', description='RNA, 7SL, cytoplasmic 228, pseudogene', other_designations='', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100048908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048908, CurrentID=0, Status='live', Symbol='ZDHHC20P2', Aliases='', description='zinc finger, DHHC-type containing 20 pseudogene 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31380408, end_position_on_the_genomic_accession=31380839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100048909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048909, CurrentID=0, Status='live', Symbol='TMEM183AP1', Aliases='', description='transmembrane protein 183A pseudogene 1', other_designations='BPG126D10.10', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29577459, end_position_on_the_genomic_accession=29577748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130446, CurrentID=0, Status='live', Symbol='RPS27AP12', Aliases='RPS27A_5_866', description='ribosomal protein S27a pseudogene 12', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157192326, end_position_on_the_genomic_accession=157192862, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100048912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048912, CurrentID=0, Status='live', Symbol='CDKN2B-AS1', Aliases='ANRIL, CDKN2B-AS, CDKN2BAS, NCRNA00089, PCAT12, p15AS', description='CDKN2B antisense RNA 1', other_designations='CDKN2B antisense RNA 1 (non-protein coding)|antisense noncoding RNA in the INK4 locus|p15 antisense RNA|prostate cancer associated transcript 12', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21994791, end_position_on_the_genomic_accession=22121097, orientation='plus', exon_count=21, OMIM=613149, no_hdr0=''),
  106480984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480984, CurrentID=0, Status='live', Symbol='RN7SL231P', Aliases='', description='RNA, 7SL, cytoplasmic 231, pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137902, CurrentID=0, Status='live', Symbol='PXDNL', Aliases='PMR1, PRM1, VPO2', description='peroxidasin like', other_designations='cardiac peroxidase|cardiovascular peroxidase 2|peroxidasin homolog-like|peroxidasin-like|polysomal ribonuclease 1 homolog|vascular peroxidase 2', map_location='8q11.22-q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51319575, end_position_on_the_genomic_accession=51811029, orientation='minus', exon_count=25, OMIM=615904, no_hdr0=''),
  106480985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480985, CurrentID=0, Status='live', Symbol='RN7SL233P', Aliases='', description='RNA, 7SL, cytoplasmic 233, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533725, CurrentID=0, Status='live', Symbol='ERVFRD-3', Aliases='', description='endogenous retrovirus group FRD member 3', other_designations='HERV-FRD provirus ancestral Env polyprotein pseudogene|endogenous retrovirus group FRD, member 1 pseudogene|endogenous retrovirus group FRD, member 3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21929455, end_position_on_the_genomic_accession=21931073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100048923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048923, CurrentID=0, Status='live', Symbol='DDX39BP2', Aliases='BAT1P2, BPG309N1.15', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39B pseudogene 2', other_designations='HLA-B associated transcript 1 pseudogene 2', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29993209, end_position_on_the_genomic_accession=29994133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100048924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048924, CurrentID=0, Status='live', Symbol='MCCD1P2', Aliases='BPG309N1.16', description='mitochondrial coiled-coil domain 1 pseudogene 2', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29995142, end_position_on_the_genomic_accession=29995314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480986, CurrentID=0, Status='live', Symbol='RN7SL236P', Aliases='', description='RNA, 7SL, cytoplasmic 236, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480987, CurrentID=0, Status='live', Symbol='RN7SL238P', Aliases='', description='RNA, 7SL, cytoplasmic 238, pseudogene', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480988, CurrentID=0, Status='live', Symbol='RN7SL241P', Aliases='', description='RNA, 7SL, cytoplasmic 241, pseudogene', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480989, CurrentID=0, Status='live', Symbol='RN7SL246P', Aliases='', description='RNA, 7SL, cytoplasmic 246, pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480990, CurrentID=0, Status='live', Symbol='RN7SL251P', Aliases='', description='RNA, 7SL, cytoplasmic 251, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372502, CurrentID=0, Status='live', Symbol='LOC105372502', Aliases='', description='uncharacterized LOC105372502', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2194050, end_position_on_the_genomic_accession=2201397, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480991, CurrentID=0, Status='live', Symbol='RN7SL254P', Aliases='', description='RNA, 7SL, cytoplasmic 254, pseudogene', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723594, CurrentID=0, Status='live', Symbol='LOC102723594', Aliases='', description='uncharacterized LOC102723594', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30293035, end_position_on_the_genomic_accession=30300485, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480992, CurrentID=0, Status='live', Symbol='RN7SL259P', Aliases='', description='RNA, 7SL, cytoplasmic 259, pseudogene', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723595, CurrentID=0, Status='live', Symbol='LOC102723595', Aliases='', description='uncharacterized LOC102723595', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37923538, end_position_on_the_genomic_accession=37924366, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480993, CurrentID=0, Status='live', Symbol='RN7SL263P', Aliases='BCR, rbc', description='RNA, 7SL, cytoplasmic 263, pseudogene', other_designations='breakpoint cluster region protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400055, CurrentID=0, Status='live', Symbol='RPL26P32', Aliases='RPL26_12_1262', description='ribosomal protein L26 pseudogene 32', other_designations='hCG26523', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80102859, end_position_on_the_genomic_accession=80103357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480994, CurrentID=0, Status='live', Symbol='RN7SL266P', Aliases='', description='RNA, 7SL, cytoplasmic 266, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723597, CurrentID=0, Status='live', Symbol='LOC102723597', Aliases='', description='uncharacterized LOC102723597', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123275625, end_position_on_the_genomic_accession=123283983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480995, CurrentID=0, Status='live', Symbol='RN7SL270P', Aliases='', description='RNA, 7SL, cytoplasmic 270, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533727, CurrentID=0, Status='live', Symbol='LOC100533727', Aliases='', description='replication factor C subunit 2 pseudogene', other_designations='replication factor C (activator 1) 2, 40kDa pseudogene', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208135394, end_position_on_the_genomic_accession=208137697, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480996, CurrentID=0, Status='live', Symbol='RN7SL272P', Aliases='', description='RNA, 7SL, cytoplasmic 272, pseudogene', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189541, CurrentID=0, Status='live', Symbol='TRV-CAC1-4', Aliases='TRNAV4', description='transfer RNA-Val (CAC) 1-4', other_designations='tRNA-Val-CAC-1-4|transfer RNA valine 4 (anticodon CAC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181173650, end_position_on_the_genomic_accession=181173722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400061, CurrentID=0, Status='live', Symbol='LOC400061', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94520952, end_position_on_the_genomic_accession=94521980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533728, CurrentID=0, Status='live', Symbol='LOC100533728', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76250227, end_position_on_the_genomic_accession=76251796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723604, CurrentID=0, Status='live', Symbol='LOC102723604', Aliases='', description='uncharacterized LOC102723604', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51916145, end_position_on_the_genomic_accession=51918571, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373375, CurrentID=0, Status='live', Symbol='LOC105373375', Aliases='', description='uncharacterized LOC105373375', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723608, CurrentID=0, Status='live', Symbol='LOC102723608', Aliases='', description='uncharacterized LOC102723608', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37949739, end_position_on_the_genomic_accession=37962200, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727747, CurrentID=0, Status='live', Symbol='RPL22P12', Aliases='RPL22_6_262', description='ribosomal protein L22 pseudogene 12', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110633692, end_position_on_the_genomic_accession=110634497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28698, CurrentID=0, Status='live', Symbol='TRAJ57', Aliases='', description='T cell receptor alpha joining 57', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22478872, end_position_on_the_genomic_accession=22478934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420252, CurrentID=0, Status='live', Symbol='LOC100420252', Aliases='', description='transmembrane channel like 6 pseudogene', other_designations='transmembrane channel-like 6 pseudogene', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42886351, end_position_on_the_genomic_accession=42888720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723611, CurrentID=0, Status='live', Symbol='LOC102723611', Aliases='', description='uncharacterized LOC102723611', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123384395, end_position_on_the_genomic_accession=123387305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481009, CurrentID=0, Status='live', Symbol='RN7SL318P', Aliases='', description='RNA, 7SL, cytoplasmic 318, pseudogene', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373983, CurrentID=0, Status='live', Symbol='LOC105373983', Aliases='', description='uncharacterized LOC105373983', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97091741, end_position_on_the_genomic_accession=97100360, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533730, CurrentID=0, Status='live', Symbol='LOC100533730', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63280418, end_position_on_the_genomic_accession=63282010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100049076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049076, CurrentID=0, Status='live', Symbol='GUSBP9', Aliases='', description='glucuronidase, beta pseudogene 9', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71220226, end_position_on_the_genomic_accession=71259295, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400073, CurrentID=0, Status='live', Symbol='C12orf76', Aliases='', description='chromosome 12 open reading frame 76', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110028042, end_position_on_the_genomic_accession=110082224, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100419668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419668, CurrentID=0, Status='live', Symbol='LOC100419668', Aliases='', description='protein kinase C, eta pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67610986, end_position_on_the_genomic_accession=67613873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369619, CurrentID=0, Status='live', Symbol='LOC105369619', Aliases='', description='uncharacterized LOC105369619', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5404192, end_position_on_the_genomic_accession=5407370, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723617, CurrentID=0, Status='live', Symbol='LOC102723617', Aliases='', description='uncharacterized LOC102723617', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35597895, end_position_on_the_genomic_accession=35601501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100533731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533731, CurrentID=0, Status='live', Symbol='RLIMP2', Aliases='', description='ring finger protein, LIM domain interacting pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113125492, end_position_on_the_genomic_accession=113126955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874164, CurrentID=0, Status='live', Symbol='LINC00417', Aliases='', description='long intergenic non-protein coding RNA 417', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18738100, end_position_on_the_genomic_accession=18740099, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370931, CurrentID=0, Status='live', Symbol='LOC105370931', Aliases='', description='uncharacterized LOC105370931', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83096605, end_position_on_the_genomic_accession=83114828, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373985, CurrentID=0, Status='live', Symbol='LOC105373985', Aliases='', description='multiple epidermal growth factor-like domains protein 6', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110022300, end_position_on_the_genomic_accession=110029293, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  400084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400084, CurrentID=0, Status='live', Symbol='LINC00939', Aliases='', description='long intergenic non-protein coding RNA 939', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125958684, end_position_on_the_genomic_accession=125983374, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723627, CurrentID=0, Status='live', Symbol='LOC102723627', Aliases='', description='uncharacterized LOC102723627', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174523038, end_position_on_the_genomic_accession=174536676, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100049159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049159, CurrentID=0, Status='live', Symbol='SPG37', Aliases='', description='spastic paraplegia 37 (autosomal dominant)', other_designations='', map_location='8p21.2-q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611945, no_hdr0=''),
  100049160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049160, CurrentID=0, Status='live', Symbol='FEB7', Aliases='', description='febrile convulsions 7', other_designations='', map_location='21q22', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611515, no_hdr0=''),
  400087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400087, CurrentID=0, Status='live', Symbol='FLJ37505', Aliases='', description='uncharacterized LOC400087', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127881617, end_position_on_the_genomic_accession=127898639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100289320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289320, CurrentID=0, Status='live', Symbol='LOC100289320', Aliases='', description='NME/NM23 nucleoside diphosphate kinase 2 pseudogene', other_designations='NME1-NME2 readthrough pseudogene', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72275447, end_position_on_the_genomic_accession=72275828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723633, CurrentID=0, Status='live', Symbol='LOC102723633', Aliases='', description='uncharacterized LOC102723633', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137415101, end_position_on_the_genomic_accession=137463779, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373987, CurrentID=0, Status='live', Symbol='LOC105373987', Aliases='', description='uncharacterized LOC105373987', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111700614, end_position_on_the_genomic_accession=111722334, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100421996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421996, CurrentID=0, Status='live', Symbol='LOC100421996', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400094, CurrentID=0, Status='live', Symbol='SNX19P2', Aliases='', description='sorting nexin 19 pseudogene 2', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18890800, end_position_on_the_genomic_accession=18895740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533735, CurrentID=0, Status='live', Symbol='LOC100533735', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=475597, end_position_on_the_genomic_accession=477104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481036, CurrentID=0, Status='live', Symbol='RN7SL414P', Aliases='', description='RNA, 7SL, cytoplasmic 414, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400099, CurrentID=0, Status='live', Symbol='LOC400099', Aliases='', description='uncharacterized LOC400099', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=20437124, end_position_on_the_genomic_accession=20439190, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996375, CurrentID=0, Status='live', Symbol='LOC100996375', Aliases='', description='angiogenic factor with G patch and FHA domains 1-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533736, CurrentID=0, Status='live', Symbol='LOC100533736', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44994425, end_position_on_the_genomic_accession=44996022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723645, CurrentID=0, Status='live', Symbol='LOC102723645', Aliases='', description='uncharacterized LOC102723645', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21520775, end_position_on_the_genomic_accession=21528269, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874169, CurrentID=0, Status='live', Symbol='LINC00434', Aliases='', description='long intergenic non-protein coding RNA 434', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60214347, end_position_on_the_genomic_accession=60268104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369625, CurrentID=0, Status='live', Symbol='LOC105369625', Aliases='', description='uncharacterized LOC105369625', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481045, CurrentID=0, Status='live', Symbol='RN7SL449P', Aliases='', description='RNA, 7SL, cytoplasmic 449, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  137964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137964, CurrentID=0, Status='live', Symbol='GPAT4', Aliases='1-AGPAT 6, AGPAT6, LPAAT-zeta, LPAATZ, TSARG7', description='glycerol-3-phosphate acyltransferase 4', other_designations='1-AGP acyltransferase 6|1-acyl-sn-glycerol-3-phosphate acyltransferase zeta|1-acylglycerol-3-phosphate O-acyltransferase 6 (lysophosphatidic acid acyltransferase, zeta)|glycerol-3-phosphate acyltransferase 6|glycerol-3-phosphate acyltransferase-4|lysophosphatidic acid acyltransferase zeta|testis spermatogenesis apoptosis-related protein 7', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41578188, end_position_on_the_genomic_accession=41625001, orientation='plus', exon_count=17, OMIM=608143, no_hdr0=''),
  400110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400110, CurrentID=0, Status='live', Symbol='ANKRD20A19P', Aliases='', description='ankyrin repeat domain 20 family member A19, pseudogene', other_designations='FLJ46358 protein|ankyrin repeat domain 20 family, member A19, pseudogene', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23907284, end_position_on_the_genomic_accession=23949315, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481051, CurrentID=0, Status='live', Symbol='RN7SL471P', Aliases='', description='RNA, 7SL, cytoplasmic 471, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271867, CurrentID=0, Status='live', Symbol='MPVQTL1', Aliases='', description='Mean platelet volume QTL1', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612573, no_hdr0=''),
  137970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=137970, CurrentID=0, Status='live', Symbol='UNC5D', Aliases='PRO34692, Unc5h4', description='unc-5 netrin receptor D', other_designations='netrin receptor Unc5h4|protein unc-5 homolog 4|protein unc-5 homolog D|unc-5 homolog 4|unc-5 homolog D', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=35235457, end_position_on_the_genomic_accession=35796550, orientation='plus', exon_count=20, OMIM=616466, no_hdr0=''),
  100128703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128703, CurrentID=0, Status='live', Symbol='LOC100128703', Aliases='', description='uncharacterized LOC100128703', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369627, CurrentID=0, Status='live', Symbol='LOC105369627', Aliases='', description='uncharacterized LOC105369627', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6345019, end_position_on_the_genomic_accession=6347078, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369628, CurrentID=0, Status='live', Symbol='LOC105369628', Aliases='', description='uncharacterized LOC105369628', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6393366, end_position_on_the_genomic_accession=6395074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400120, CurrentID=0, Status='live', Symbol='SERTM1', Aliases='C13orf36', description='serine-rich and transmembrane domain containing 1', other_designations='serine-rich and transmembrane domain-containing 1', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36673912, end_position_on_the_genomic_accession=36697839, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400121, CurrentID=0, Status='live', Symbol='LINC00547', Aliases='', description='long intergenic non-protein coding RNA 547', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37534940, end_position_on_the_genomic_accession=37551536, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723663, CurrentID=0, Status='live', Symbol='LOC102723663', Aliases='', description='uncharacterized LOC102723663', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11193032, end_position_on_the_genomic_accession=11207321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271869, CurrentID=0, Status='live', Symbol='MPVQTL3', Aliases='', description='Mean platelet volume QTL3', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612575, no_hdr0=''),
  105369629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369629, CurrentID=0, Status='live', Symbol='LOC105369629', Aliases='', description='uncharacterized LOC105369629', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6606460, end_position_on_the_genomic_accession=6609304, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420263, CurrentID=0, Status='live', Symbol='CICP21', Aliases='', description='capicua transcriptional repressor pseudogene 21', other_designations='capicua homolog pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243049737, end_position_on_the_genomic_accession=243051569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285074, CurrentID=0, Status='live', Symbol='LOC285074', Aliases='', description='anaphase promoting complex subunit 1 pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87030675, end_position_on_the_genomic_accession=87076413, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  400126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400126, CurrentID=0, Status='live', Symbol='DGKZP1', Aliases='', description='diacylglycerol kinase, zeta pseudogene 1', other_designations='diacylglycerol kinase, zeta 104kDa pseudogene', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43968323, end_position_on_the_genomic_accession=43971710, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400129, CurrentID=0, Status='live', Symbol='SMARCE1P5', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 5', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44594858, end_position_on_the_genomic_accession=44596149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481068, CurrentID=0, Status='live', Symbol='RN7SL536P', Aliases='', description='RNA, 7SL, cytoplasmic 536, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727810, CurrentID=0, Status='live', Symbol='RPL21P17', Aliases='RPL21_9_342', description='ribosomal protein L21 pseudogene 17', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1905648, end_position_on_the_genomic_accession=1906169, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723673, CurrentID=0, Status='live', Symbol='LOC102723673', Aliases='', description='uncharacterized LOC102723673', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74209388, end_position_on_the_genomic_accession=74212973, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373995, CurrentID=0, Status='live', Symbol='LOC105373995', Aliases='', description='uncharacterized LOC105373995', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97717436, end_position_on_the_genomic_accession=97758619, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616668, CurrentID=0, Status='live', Symbol='TPTE2P5', Aliases='', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 5', other_designations='TPTE2 pseudogene', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40796985, end_position_on_the_genomic_accession=40921750, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271871, CurrentID=0, Status='live', Symbol='RBM22P2', Aliases='', description='RNA binding motif protein 22 pseudogene 2', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30539297, end_position_on_the_genomic_accession=30540584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481072, CurrentID=0, Status='live', Symbol='RN7SL552P', Aliases='', description='RNA, 7SL, cytoplasmic 552, pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723675, CurrentID=0, Status='live', Symbol='LOC102723675', Aliases='', description='uncharacterized LOC102723675', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24980352, end_position_on_the_genomic_accession=25026861, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874175, CurrentID=0, Status='live', Symbol='LINC00446', Aliases='', description='long intergenic non-protein coding RNA 446', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78012883, end_position_on_the_genomic_accession=78053595, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369631, CurrentID=0, Status='live', Symbol='LOC105369631', Aliases='', description='uncharacterized LOC105369631', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6613961, end_position_on_the_genomic_accession=6618323, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  28765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28765, CurrentID=0, Status='live', Symbol='IGLVI-63', Aliases='IGLV(I)-63, IGLVI63, V1-24P', description='immunoglobulin lambda variable (I)-63 (pseudogene)', other_designations='immunoglobulin lambda variable (I)-63 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22079770, end_position_on_the_genomic_accession=22080233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23205, CurrentID=0, Status='live', Symbol='ACSBG1', Aliases='BG, BG1, BGM, GR-LACS, LPD', description='acyl-CoA synthetase bubblegum family member 1', other_designations='bubblegum|lipidosin|very long-chain acyl-CoA synthetase', map_location='15q23-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78170845, end_position_on_the_genomic_accession=78234707, orientation='minus', exon_count=19, OMIM=614362, no_hdr0=''),
  102723678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723678, CurrentID=0, Status='live', Symbol='LOC102723678', Aliases='', description='uncharacterized LOC102723678', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61865632, end_position_on_the_genomic_accession=61869154, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481076, CurrentID=0, Status='live', Symbol='RN7SL565P', Aliases='', description='RNA, 7SL, cytoplasmic 565, pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  6922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6922, CurrentID=0, Status='live', Symbol='SKP1P1', Aliases='EMC19, OCP2, TCEB1L, TCEB1P, TCEB1P1', description='S-phase kinase-associated protein 1 pseudogene 1', other_designations='transcription elongation factor B (SIII), polypeptide 1 pseudogene 1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66422642, end_position_on_the_genomic_accession=66423986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481077, CurrentID=0, Status='live', Symbol='RN7SL569P', Aliases='', description='RNA, 7SL, cytoplasmic 569, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94304, CurrentID=0, Status='live', Symbol='PPP1R1AP1', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 1A pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 1A pseudogene 1', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75911200, end_position_on_the_genomic_accession=75911624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481078, CurrentID=0, Status='live', Symbol='RN7SL573P', Aliases='', description='RNA, 7SL, cytoplasmic 573, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723681, CurrentID=0, Status='live', Symbol='LOC102723681', Aliases='', description='protein capicua homolog pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727820, CurrentID=0, Status='live', Symbol='LOC727820', Aliases='', description='uncharacterized LOC727820', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=142529482, end_position_on_the_genomic_accession=142537643, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481079, CurrentID=0, Status='live', Symbol='RN7SL574P', Aliases='', description='RNA, 7SL, cytoplasmic 574, pseudogene', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400141, CurrentID=0, Status='live', Symbol='RPSAP53', Aliases='RPSA_19_1317', description='ribosomal protein SA pseudogene 53', other_designations='', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67266755, end_position_on_the_genomic_accession=67267780, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481080, CurrentID=0, Status='live', Symbol='RN7SL577P', Aliases='', description='RNA, 7SL, cytoplasmic 577, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28771, CurrentID=0, Status='live', Symbol='IGLV10-67', Aliases='IGLV1067, V1-25P', description='immunoglobulin lambda variable 10-67 (pseudogene)', other_designations='immunoglobulin lambda variable 10-67 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22043452, end_position_on_the_genomic_accession=22043914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481081, CurrentID=0, Status='live', Symbol='RN7SL580P', Aliases='', description='RNA, 7SL, cytoplasmic 580, pseudogene', other_designations='', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481082, CurrentID=0, Status='live', Symbol='RN7SL583P', Aliases='', description='RNA, 7SL, cytoplasmic 583, pseudogene', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533729, CurrentID=0, Status='live', Symbol='LOC100533729', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63087071, end_position_on_the_genomic_accession=63088652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481083, CurrentID=0, Status='live', Symbol='RN7SL586P', Aliases='', description='RNA, 7SL, cytoplasmic 586, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723686, CurrentID=0, Status='live', Symbol='LOC102723686', Aliases='', description='uncharacterized LOC102723686', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153225924, end_position_on_the_genomic_accession=153256244, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373998, CurrentID=0, Status='live', Symbol='LOC105373998', Aliases='', description='uncharacterized LOC105373998', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98184780, end_position_on_the_genomic_accession=98188858, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28777, CurrentID=0, Status='live', Symbol='IGLV7-35', Aliases='IGLV735, V3-1P', description='immunoglobulin lambda variable 7-35 (pseudogene)', other_designations='immunoglobulin lambda variable 7-35 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22450358, end_position_on_the_genomic_accession=22450652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271874, CurrentID=0, Status='live', Symbol='ZNRF2P2', Aliases='', description='zinc and ring finger 2 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29684772, end_position_on_the_genomic_accession=29685821, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100874178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874178, CurrentID=0, Status='live', Symbol='LINC00454', Aliases='', description='long intergenic non-protein coding RNA 454', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113878628, end_position_on_the_genomic_accession=113883404, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481088, CurrentID=0, Status='live', Symbol='RN7SL607P', Aliases='', description='RNA, 7SL, cytoplasmic 607, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369634, CurrentID=0, Status='live', Symbol='LOC105369634', Aliases='', description='uncharacterized LOC105369634', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6961632, end_position_on_the_genomic_accession=6963177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100049541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049541, CurrentID=0, Status='live', Symbol='STUT1', Aliases='', description='Stuttering, familial persistent 1', other_designations='', map_location='18p11.3-p11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=184450, no_hdr0=''),
  100049542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049542, CurrentID=0, Status='live', Symbol='SHFM3', Aliases='C10DUPq24, DUP10q24, SHSF3', description='Split-hand/foot malformation 3', other_designations='', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=246560, no_hdr0=''),
  100049543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049543, CurrentID=0, Status='live', Symbol='STUT2', Aliases='', description='Stuttering, familial persistent 2', other_designations='', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609261, no_hdr0=''),
  100049544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049544, CurrentID=0, Status='live', Symbol='APMR2', Aliases='', description='Alopecia with mental retardation syndrome 2', other_designations='', map_location='3q26.2-q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610422, no_hdr0=''),
  106481090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481090, CurrentID=0, Status='live', Symbol='RN7SL616P', Aliases='', description='RNA, 7SL, cytoplasmic 616, pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481092, CurrentID=0, Status='live', Symbol='RN7SL622P', Aliases='', description='RNA, 7SL, cytoplasmic 622, pseudogene', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369635, CurrentID=0, Status='live', Symbol='LOC105369635', Aliases='', description='uncharacterized LOC105369635', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6963204, end_position_on_the_genomic_accession=6964452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727835, CurrentID=0, Status='live', Symbol='RPL9P16', Aliases='RPL9_3_534', description='ribosomal protein L9 pseudogene 16', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168755798, end_position_on_the_genomic_accession=168756481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400156, CurrentID=0, Status='live', Symbol='RPS26P47', Aliases='RPS26, RPS26L, RPS26_24_1332', description='ribosomal protein S26 pseudogene 47', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100539816, end_position_on_the_genomic_accession=100540272, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100049579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049579, CurrentID=0, Status='live', Symbol='OCTN3', Aliases='', description='organic cation transporter 3', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481096, CurrentID=0, Status='live', Symbol='RN7SL634P', Aliases='', description='RNA, 7SL, cytoplasmic 634, pseudogene', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100049587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049587, CurrentID=0, Status='live', Symbol='SIGLEC14', Aliases='', description='sialic acid binding Ig-like lectin 14', other_designations='siglec-14', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51642553, end_position_on_the_genomic_accession=51646879, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100874180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874180, CurrentID=0, Status='live', Symbol='LINC00459', Aliases='', description='long intergenic non-protein coding RNA 459', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28789, CurrentID=0, Status='live', Symbol='IGLV3-30', Aliases='IGLV330, V2-21P', description='immunoglobulin lambda variable 3-30 (pseudogene)', other_designations='immunoglobulin lambda variable 3-30 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22618718, end_position_on_the_genomic_accession=22619211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481100, CurrentID=0, Status='live', Symbol='RN7SL649P', Aliases='', description='RNA, 7SL, cytoplasmic 649, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723703, CurrentID=0, Status='live', Symbol='LOC102723703', Aliases='', description='uncharacterized LOC102723703', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80531250, end_position_on_the_genomic_accession=80533123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100049613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049613, CurrentID=0, Status='live', Symbol='TRK-TTT3-3', Aliases='TRNAK3', description='transfer RNA-Lys (TTT) 3-3', other_designations='tRNA-Lys-TTT-3-3|transfer RNA lysine 3 (anticodon UUU)', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59560335, end_position_on_the_genomic_accession=59560407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100049614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049614, CurrentID=0, Status='live', Symbol='TRV-CAC1-5', Aliases='TRNAV6, TRV6', description='transfer RNA-Val (CAC) 1-5', other_designations='tRNA-Val-CAC-1-5|transfer RNA valine 6 (anticodon CAC)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181222395, end_position_on_the_genomic_accession=181222467, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100049615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049615, CurrentID=0, Status='live', Symbol='SKCG-1', Aliases='', description='sporadic kidney cancer gene 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=108829770, end_position_on_the_genomic_accession=108830144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106865367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106865367, CurrentID=0, Status='live', Symbol='MRD2', Aliases='', description='Mental retardation, autosomal dominant 2', other_designations='', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614113, no_hdr0=''),
  106481103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481103, CurrentID=0, Status='live', Symbol='RN7SL663P', Aliases='', description='RNA, 7SL, cytoplasmic 663, pseudogene', other_designations='', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106865372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106865372, CurrentID=0, Status='live', Symbol='RNU6-146P', Aliases='', description='RNA, U6 small nuclear 146, pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106865373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106865373, CurrentID=0, Status='live', Symbol='WRB-SH3BGR', Aliases='', description='WRB-SH3BGR readthrough', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106865374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106865374, CurrentID=0, Status='live', Symbol='RNU6-468P', Aliases='', description='RNA, U6 small nuclear 468, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400165, CurrentID=0, Status='live', Symbol='ATP11AUN', Aliases='C13orf35, SMABLO1', description='ATP11A upstream neighbor', other_designations='putative ATP11A upstream neighbor protein|small blood protein 1', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112647044, end_position_on_the_genomic_accession=112684497, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481104, CurrentID=0, Status='live', Symbol='RN7SL665P', Aliases='', description='RNA, 7SL, cytoplasmic 665, pseudogene', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389831, CurrentID=0, Status='live', Symbol='LOC389831', Aliases='', description='uncharacterized LOC389831', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421999, CurrentID=0, Status='live', Symbol='LOC100421999', Aliases='', description='olfactory receptor, family 5, subfamily J, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481108, CurrentID=0, Status='live', Symbol='RN7SL678P', Aliases='', description='RNA, 7SL, cytoplasmic 678, pseudogene', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723711, CurrentID=0, Status='live', Symbol='CDC27P7', Aliases='', description='cell division cycle 27 pseudogene 7', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11257794, end_position_on_the_genomic_accession=11259748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130451, CurrentID=0, Status='live', Symbol='LOC100130451', Aliases='', description='uncharacterized LOC100130451', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=213276553, end_position_on_the_genomic_accession=213284205, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723713, CurrentID=0, Status='live', Symbol='LOC102723713', Aliases='', description='TP53-target gene 3 protein', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188953, CurrentID=0, Status='live', Symbol='LINC00092', Aliases='NCRNA00092, bA346B7.1', description='long intergenic non-protein coding RNA 92', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96019732, end_position_on_the_genomic_accession=96021755, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928919, CurrentID=0, Status='live', Symbol='LOC101928919', Aliases='', description='uncharacterized LOC101928919', other_designations='CTD-2589M5.4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46256355, end_position_on_the_genomic_accession=46274643, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400174, CurrentID=0, Status='live', Symbol='LOC400174', Aliases='', description='single stranded DNA binding protein 4 pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19668829, end_position_on_the_genomic_accession=19670000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481113, CurrentID=0, Status='live', Symbol='RN7SL698P', Aliases='', description='RNA, 7SL, cytoplasmic 698, pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481114, CurrentID=0, Status='live', Symbol='RN7SL702P', Aliases='', description='RNA, 7SL, cytoplasmic 702, pseudogene', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106699568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106699568, CurrentID=0, Status='live', Symbol='ERVW-9', Aliases='', description='endogenous retrovirus group W member 9', other_designations='endogenous retrovirus group W, member 9', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481115, CurrentID=0, Status='live', Symbol='RN7SL704P', Aliases='', description='RNA, 7SL, cytoplasmic 704, pseudogene', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730021, CurrentID=0, Status='live', Symbol='LOC730021', Aliases='', description='eukaryotic translation initiation factor 3 subunit J pseudogene', other_designations='eukaryotic translation initiation factor 3, subunit J pseudogene', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150050645, end_position_on_the_genomic_accession=150052601, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  203569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203569, CurrentID=0, Status='live', Symbol='PAGE2', Aliases='CT16.4, GAGEC2, GAGEE2, PAGE-2', description='PAGE family member 2', other_designations='G antigen family C 2|G antigen, family C, 2|G antigen, family E, 2|P antigen family, member 2 (prostate associated)|g antigen family E member 2|prostate-associated gene 2 protein', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55089052, end_position_on_the_genomic_accession=55092836, orientation='plus', exon_count=5, OMIM=300738, no_hdr0=''),
  106481116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481116, CurrentID=0, Status='live', Symbol='RN7SL709P', Aliases='', description='RNA, 7SL, cytoplasmic 709, pseudogene', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100049707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049707, CurrentID=0, Status='live', Symbol='SPG38', Aliases='', description='spastic paraplegia 38 (autosomal dominant, Silver syndrome)', other_designations='', map_location='4p16-p15', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612335, no_hdr0=''),
  106699570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106699570, CurrentID=0, Status='live', Symbol='LOC106699570', Aliases='', description='uncharacterized LOC106699570', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100049716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049716, CurrentID=0, Status='live', Symbol='LOC100049716', Aliases='', description='uncharacterized LOC100049716', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=630891, end_position_on_the_genomic_accession=645878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100049717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100049717, CurrentID=0, Status='live', Symbol='JKAMPP1', Aliases='', description='JNK1/MAPK8-associated membrane protein pseudogene 1', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12287323, end_position_on_the_genomic_accession=12289233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100967223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100967223, CurrentID=0, Status='live', Symbol='LOC100967223', Aliases='', description='neuroepithelial cell transforming 1 pseudogene', other_designations='', map_location='7q11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=70837757, end_position_on_the_genomic_accession=70845353, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100967224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100967224, CurrentID=0, Status='live', Symbol='LOC100967224', Aliases='', description='neuroepithelial cell transforming 1 pseudogene', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187632152, end_position_on_the_genomic_accession=187635030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  6967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6967, CurrentID=0, Status='live', Symbol='TRGC2', Aliases='TCRGC2(2X), TRGC2(3X), TRGC2', description='T cell receptor gamma constant 2', other_designations='T-cell receptor, gamma, constant region C2', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38240024, end_position_on_the_genomic_accession=38249572, orientation='minus', exon_count=0, OMIM=615450, no_hdr0=''),
  6968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6968, CurrentID=0, Status='live', Symbol='TRGJ1', Aliases='J1, TCRGJ1', description='T cell receptor gamma joining 1', other_designations='T-cell receptor, gamma, joining segment J1', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38269491, end_position_on_the_genomic_accession=38269540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369641, CurrentID=0, Status='live', Symbol='LOC105369641', Aliases='', description='uncharacterized LOC105369641', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7623758, end_position_on_the_genomic_accession=7628695, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100293704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100293704, CurrentID=0, Status='live', Symbol='LOC100293704', Aliases='', description='serine/arginine repetitive matrix protein 3-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123363059, end_position_on_the_genomic_accession=123365789, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  553103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553103, CurrentID=0, Status='live', Symbol='LOC553103', Aliases='', description='uncharacterized LOC553103', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132311276, end_position_on_the_genomic_accession=132369916, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  6974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6974, CurrentID=0, Status='live', Symbol='TRGV2', Aliases='TCRGV2, VIS2', description='T cell receptor gamma variable 2', other_designations='T-cell receptor, gamma, variable region V2', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38362864, end_position_on_the_genomic_accession=38363328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481129, CurrentID=0, Status='live', Symbol='RN7SL760P', Aliases='', description='RNA, 7SL, cytoplasmic 760, pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23216, CurrentID=0, Status='live', Symbol='TBC1D1', Aliases='TBC, TBC1', description='TBC1 domain family member 1', other_designations='TBC1 (tre-2/USP6, BUB2, cdc16) domain family, member 1', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37891084, end_position_on_the_genomic_accession=38139175, orientation='plus', exon_count=31, OMIM=609850, no_hdr0=''),
  100533754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533754, CurrentID=0, Status='live', Symbol='MARK3P3', Aliases='', description='MAP/microtubule affinity-regulating kinase 3 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128847693, end_position_on_the_genomic_accession=128849411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  6978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=6978, CurrentID=0, Status='live', Symbol='TRGV5', Aliases='TCRGV5, V1S5', description='T cell receptor gamma variable 5', other_designations='T-cell gamma precursor protein|T-cell receptor, gamma, variable region V5', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38349355, end_position_on_the_genomic_accession=38349824, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553112, CurrentID=0, Status='live', Symbol='FERP1', Aliases='FERP, FERps', description='fer (fps/fes related) tyrosine kinase (phosphoprotein NCP94) pseudogene 1', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123603139, end_position_on_the_genomic_accession=123603931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533755, CurrentID=0, Status='live', Symbol='LOC100533755', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11905181, end_position_on_the_genomic_accession=11906306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28827, CurrentID=0, Status='live', Symbol='IGLJ7', Aliases='J7', description='immunoglobulin lambda joining 7', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22921398, end_position_on_the_genomic_accession=22921435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271884, CurrentID=0, Status='live', Symbol='SMARCE1P1', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 1', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107403370, end_position_on_the_genomic_accession=107404637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100270647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270647, CurrentID=0, Status='live', Symbol='LOC100270647', Aliases='', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15324260, end_position_on_the_genomic_accession=15325436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481138, CurrentID=0, Status='live', Symbol='RN7SL797P', Aliases='', description='RNA, 7SL, cytoplasmic 797, pseudogene', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553117, CurrentID=0, Status='live', Symbol='RPL39P5', Aliases='RPL39_7_406', description='ribosomal protein L39 pseudogene 5', other_designations='60 ribosomal protein L39 pseudogene 5', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134351739, end_position_on_the_genomic_accession=134352143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  553118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553118, CurrentID=0, Status='live', Symbol='STK33P1', Aliases='STK33P, STK33ps', description='serine/threonine kinase 33 pseudogene 1', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85492579, end_position_on_the_genomic_accession=85494094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  553121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553121, CurrentID=0, Status='live', Symbol='TLK2P2', Aliases='TLK2ps2', description='tousled like kinase 2 pseudogene 2', other_designations='tousled-like kinase 2 pseudogene 2', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37818606, end_position_on_the_genomic_accession=37820982, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132332, CurrentID=0, Status='live', Symbol='TMEM155', Aliases='', description='transmembrane protein 155', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121758930, end_position_on_the_genomic_accession=121766042, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  400207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400207, CurrentID=0, Status='live', Symbol='FLJ42220', Aliases='', description='FLJ42220 protein', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=36186635, end_position_on_the_genomic_accession=36187505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400208, CurrentID=0, Status='live', Symbol='LINC00517', Aliases='C14orf26', description='long intergenic non-protein coding RNA 517', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481147, CurrentID=0, Status='live', Symbol='RN7SL828P', Aliases='', description='RNA, 7SL, cytoplasmic 828, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138065, CurrentID=0, Status='live', Symbol='RNF183', Aliases='', description='ring finger protein 183', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113297093, end_position_on_the_genomic_accession=113303274, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  400212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400212, CurrentID=0, Status='live', Symbol='LOC400212', Aliases='', description='uncharacterized LOC400212', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50946794, end_position_on_the_genomic_accession=50962002, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128130, CurrentID=0, Status='live', Symbol='LOC100128130', Aliases='', description='uncharacterized LOC100128130', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=111994581, end_position_on_the_genomic_accession=111996624, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481153, CurrentID=0, Status='live', Symbol='RN7SL853P', Aliases='', description='RNA, 7SL, cytoplasmic 853, pseudogene', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401589, CurrentID=0, Status='live', Symbol='LOC401589', Aliases='', description='uncharacterized LOC401589', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52051171, end_position_on_the_genomic_accession=52054255, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723759, CurrentID=0, Status='live', Symbol='LOC102723759', Aliases='', description='uncharacterized LOC102723759', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128033395, end_position_on_the_genomic_accession=128052824, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100533732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533732, CurrentID=0, Status='live', Symbol='LOC100533732', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=188670702, end_position_on_the_genomic_accession=188671759, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369648, CurrentID=0, Status='live', Symbol='LOC105369648', Aliases='', description='uncharacterized LOC105369648', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130453, CurrentID=0, Status='live', Symbol='LOC100130453', Aliases='', description='uncharacterized LOC100130453', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=75615801, end_position_on_the_genomic_accession=75616229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481159, CurrentID=0, Status='live', Symbol='RNU4ATAC14P', Aliases='', description='RNA, U4atac small nuclear 14, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400221, CurrentID=0, Status='live', Symbol='FLJ22447', Aliases='', description='uncharacterized LOC400221', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61570540, end_position_on_the_genomic_accession=61654713, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724749, CurrentID=0, Status='live', Symbol='LOC102724749', Aliases='', description='uncharacterized LOC102724749', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23337575, end_position_on_the_genomic_accession=23405413, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105373506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373506, CurrentID=0, Status='live', Symbol='LOC105373506', Aliases='', description='uncharacterized LOC105373506', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100476033, end_position_on_the_genomic_accession=100494205, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7008, CurrentID=0, Status='live', Symbol='TEF', Aliases='', description='thyrotrophic embryonic factor', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41367333, end_position_on_the_genomic_accession=41399328, orientation='plus', exon_count=5, OMIM=188595, no_hdr0=''),
  28854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28854, CurrentID=0, Status='live', Symbol='IGKV3OR2-5', Aliases='IGKV3/OR2-5, IGKV3OR25', description='immunoglobulin kappa variable 3/OR2-5 (pseudogene)', other_designations='immunoglobulin kappa variable 3/OR2-5 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481164, CurrentID=0, Status='live', Symbol='RNU6ATAC17P', Aliases='', description='RNA, U6atac small nuclear 17, pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553143, CurrentID=0, Status='live', Symbol='BW65', Aliases='', description='Body weight QTL 65', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442217, CurrentID=0, Status='live', Symbol='RPL31P28', Aliases='RPL31_12_703', description='ribosomal protein L31 pseudogene 28', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53354661, end_position_on_the_genomic_accession=53355106, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28858, CurrentID=0, Status='live', Symbol='IGKV2OR2-4', Aliases='IGKV2/OR2-4, IGKV2OR24', description='immunoglobulin kappa variable 2/OR2-4 (pseudogene)', other_designations='immunoglobulin kappa variable 2/OR2-4 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28859, CurrentID=0, Status='live', Symbol='IGKV2OR2-2', Aliases='IGKV2/OR2-2, IGKV2OR22', description='immunoglobulin kappa variable 2/OR2-2 (pseudogene)', other_designations='immunoglobulin kappa variable 2/OR2-2 pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  553148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=553148, CurrentID=0, Status='live', Symbol='SERBP1P3', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene 3', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53058168, end_position_on_the_genomic_accession=53065117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  23224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23224, CurrentID=0, Status='live', Symbol='SYNE2', Aliases='EDMD5, NUA, NUANCE, Nesp2, Nesprin-2, SYNE-2, TROPH', description='spectrin repeat containing, nuclear envelope 2', other_designations='nesprin 2|nuclear envelope spectrin repeat protein 2|nucleus and actin connecting element protein|polytrophin|synaptic nuclear envelope protein 2|synaptic nuclei expressed gene 2', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63852586, end_position_on_the_genomic_accession=64226449, orientation='plus', exon_count=120, OMIM=608442, no_hdr0=''),
  106481171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481171, CurrentID=0, Status='live', Symbol='RNU6ATAC35P', Aliases='', description='RNA, U6atac small nuclear 35, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400236, CurrentID=0, Status='live', Symbol='FOXN3-AS1', Aliases='', description='FOXN3 antisense RNA 1', other_designations='FOXN3 antisense RNA 1 (non-protein coding)', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89417354, end_position_on_the_genomic_accession=89419793, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400238, CurrentID=0, Status='live', Symbol='LINC00642', Aliases='', description='long intergenic non-protein coding RNA 642', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90455230, end_position_on_the_genomic_accession=90458905, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  285097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285097, CurrentID=0, Status='live', Symbol='LOC285097', Aliases='', description='uncharacterized FLJ38379', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241996697, end_position_on_the_genomic_accession=242006013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481179, CurrentID=0, Status='live', Symbol='RNU4-24P', Aliases='', description='RNA, U4 small nuclear 24, pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28870, CurrentID=0, Status='live', Symbol='IGKV6D-21', Aliases='A10, IGKV6D21', description='immunoglobulin kappa variable 6D-21 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90021634, end_position_on_the_genomic_accession=90022185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400242, CurrentID=0, Status='live', Symbol='DICER1-AS1', Aliases='DICER1-AS', description='DICER1 antisense RNA 1', other_designations='DICER1 antisense RNA (non-protein coding)|DICER1 antisense RNA 1 (non-protein coding)', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95157688, end_position_on_the_genomic_accession=95179933, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481181, CurrentID=0, Status='live', Symbol='RNU4-29P', Aliases='', description='RNA, U4 small nuclear 29, pseudogene', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289351, CurrentID=0, Status='live', Symbol='LOC100289351', Aliases='', description='ribosomal protein S12 pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72470374, end_position_on_the_genomic_accession=72471073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271893, CurrentID=0, Status='live', Symbol='SAMM50P1', Aliases='', description='SAMM50 sorting and assembly machinery component pseudogene 1', other_designations='sorting and assembly machinery component 50 homolog pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17447002, end_position_on_the_genomic_accession=17448615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533733, CurrentID=0, Status='live', Symbol='LOC100533733', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=262820, end_position_on_the_genomic_accession=264256, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481183, CurrentID=0, Status='live', Symbol='RNU4-37P', Aliases='', description='RNA, U4 small nuclear 37, pseudogene', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369653, CurrentID=0, Status='live', Symbol='LOC105369653', Aliases='', description='uncharacterized LOC105369653', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9769179, end_position_on_the_genomic_accession=9822815, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374018, CurrentID=0, Status='live', Symbol='LOC105374018', Aliases='', description='uncharacterized LOC105374018', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102899864, end_position_on_the_genomic_accession=102926058, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369654, CurrentID=0, Status='live', Symbol='LOC105369654', Aliases='', description='uncharacterized LOC105369654', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9953002, end_position_on_the_genomic_accession=9956991, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927752, CurrentID=0, Status='live', Symbol='LOC101927752', Aliases='', description='uncharacterized LOC101927752', other_designations='CTD-2336O2.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1758208, end_position_on_the_genomic_accession=1763334, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928625, CurrentID=0, Status='live', Symbol='LOC101928625', Aliases='', description='uncharacterized LOC101928625', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27046193, end_position_on_the_genomic_accession=27068609, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723795, CurrentID=0, Status='live', Symbol='LOC102723795', Aliases='', description='uncharacterized LOC102723795', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156893362, end_position_on_the_genomic_accession=156922209, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874199, CurrentID=0, Status='live', Symbol='SHANK2-AS2', Aliases='', description='SHANK2 antisense RNA 2', other_designations='SHANK2 antisense RNA 2 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420289, CurrentID=0, Status='live', Symbol='LOC100420289', Aliases='', description='TBC1 domain family, member 3H pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155384887, end_position_on_the_genomic_accession=155385489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  28885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28885, CurrentID=0, Status='live', Symbol='IGKV2D-24', Aliases='A7, IGKV2D24', description='immunoglobulin kappa variable 2D-24 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90004827, end_position_on_the_genomic_accession=90005629, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379836, CurrentID=0, Status='live', Symbol='LOC105379836', Aliases='', description='uncharacterized LOC105379836', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400258, CurrentID=0, Status='live', Symbol='C14orf180', Aliases='C14orf77, NRAC', description='chromosome 14 open reading frame 180', other_designations='transmembrane protein C14orf180', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104579684, end_position_on_the_genomic_accession=104589847, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  394214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394214, CurrentID=0, Status='live', Symbol='COL11A2P1', Aliases='COL11A2P, cD0826Q1.1', description='collagen, type XI, alpha 2 pseudogene 1', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33103693, end_position_on_the_genomic_accession=33107144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28891, CurrentID=0, Status='live', Symbol='IGKV1D-43', Aliases='IGKV1D43, L23, L23a', description='immunoglobulin kappa variable 1D-43', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90210054, end_position_on_the_genomic_accession=90210529, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374021, CurrentID=0, Status='live', Symbol='LOC105374021', Aliases='', description='uncharacterized LOC105374021', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104522756, end_position_on_the_genomic_accession=104526870, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369885, CurrentID=0, Status='live', Symbol='LOC105369885', Aliases='', description='uncharacterized LOC105369885', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88597968, end_position_on_the_genomic_accession=88602195, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100533734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533734, CurrentID=0, Status='live', Symbol='LOC100533734', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=434345, end_position_on_the_genomic_accession=435790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874202, CurrentID=0, Status='live', Symbol='MIS18A-AS1', Aliases='', description='MIS18A antisense RNA 1', other_designations='MIS18A antisense RNA 1 (non-protein coding)', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130455, CurrentID=0, Status='live', Symbol='LOC100130455', Aliases='', description='chromosome 20 open reading frame 196 pseudogene', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88923388, end_position_on_the_genomic_accession=88923985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374023, CurrentID=0, Status='live', Symbol='LOC105374023', Aliases='', description='uncharacterized LOC105374023', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104995641, end_position_on_the_genomic_accession=105017986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  28902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28902, CurrentID=0, Status='live', Symbol='IGKV1D-13', Aliases='IGKV1D13, L18', description='immunoglobulin kappa variable 1D-13', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90154098, end_position_on_the_genomic_accession=90154574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132495, CurrentID=0, Status='live', Symbol='LOC100132495', Aliases='', description='uncharacterized LOC100132495', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=143963224, end_position_on_the_genomic_accession=143963631, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  28904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28904, CurrentID=0, Status='live', Symbol='IGKV1D-8', Aliases='IGKV1D8, L24, L24a', description='immunoglobulin kappa variable 1D-8', other_designations='immunoglobulin light chain variable region', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90220908, end_position_on_the_genomic_accession=90221382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  394217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394217, CurrentID=0, Status='live', Symbol='FGF7P2', Aliases='FGF7L, PRED3', description='fibroblast growth factor 7 pseudogene 2', other_designations='fibroblast growth factor 7 (keratinocyte growth factor) pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13349265, end_position_on_the_genomic_accession=13353460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372210, CurrentID=0, Status='live', Symbol='LOC105372210', Aliases='', description='uncharacterized LOC105372210', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76225185, end_position_on_the_genomic_accession=76232107, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28906, CurrentID=0, Status='live', Symbol='IGKV6-21', Aliases='A26, IGKV621', description='immunoglobulin kappa variable 6-21 (non-functional)', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89159751, end_position_on_the_genomic_accession=89160302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28907, CurrentID=0, Status='live', Symbol='IGKV5-2', Aliases='B2, IGKV52', description='immunoglobulin kappa variable 5-2', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88897295, end_position_on_the_genomic_accession=88897784, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874204, CurrentID=0, Status='live', Symbol='ALDH1L1-AS1', Aliases='', description='ALDH1L1 antisense RNA 1', other_designations='ALDH1L1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126103640, end_position_on_the_genomic_accession=126108069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369660, CurrentID=0, Status='live', Symbol='LOC105369660', Aliases='', description='uncharacterized LOC105369660', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10723107, end_position_on_the_genomic_accession=10727487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481219, CurrentID=0, Status='live', Symbol='RNU6-132P', Aliases='', description='RNA, U6 small nuclear 132, pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369661, CurrentID=0, Status='live', Symbol='LOC105369661', Aliases='', description='uncharacterized LOC105369661', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28917, CurrentID=0, Status='live', Symbol='IGKV2-38', Aliases='IGKV238, O13', description='immunoglobulin kappa variable 2-38 (pseudogene)', other_designations='immunoglobulin kappa variable 2-38 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89309903, end_position_on_the_genomic_accession=89310142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289360, CurrentID=0, Status='live', Symbol='LOC100289360', Aliases='', description='HIG1 hypoxia inducible domain family member 2A pseudogene', other_designations='HIG1 hypoxia inducible domain family, member 2A pseudogene', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231437109, end_position_on_the_genomic_accession=231437421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28921, CurrentID=0, Status='live', Symbol='IGKV2-28', Aliases='A19, IGKV228', description='immunoglobulin kappa variable 2-28', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89221698, end_position_on_the_genomic_accession=89222431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289361, CurrentID=0, Status='live', Symbol='LOC100289361', Aliases='', description='uncharacterized LOC100289361', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143000845, end_position_on_the_genomic_accession=143001467, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  138151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138151, CurrentID=0, Status='live', Symbol='NACC2', Aliases='BEND9, BTBD14, BTBD14A, BTBD31, NAC-2, RBB', description='NACC family member 2, BEN and BTB (POZ) domain containing', other_designations='BEN domain containing 9|BTB (POZ) domain containing 14A|BTB/POZ domain-containing protein 14A|repressor with BTB domain and BEN domain|transcription repressor with a BTB domain and a BEN domain', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136006537, end_position_on_the_genomic_accession=136095285, orientation='minus', exon_count=7, OMIM=615786, no_hdr0=''),
  105379791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379791, CurrentID=0, Status='live', Symbol='LOC105379791', Aliases='', description='uncharacterized LOC105379791', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132239, CurrentID=0, Status='live', Symbol='LOC100132239', Aliases='', description='calcium activated nucleotidase 1 pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25152650, end_position_on_the_genomic_accession=25155382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369664, CurrentID=0, Status='live', Symbol='LOC105369664', Aliases='', description='uncharacterized LOC105369664', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12879726, end_position_on_the_genomic_accession=12883081, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101060093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060093, CurrentID=0, Status='live', Symbol='LOC101060093', Aliases='', description='eukaryotic translation initiation factor 3 subunit J-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76992150, end_position_on_the_genomic_accession=76993176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929154, CurrentID=0, Status='live', Symbol='LOC101929154', Aliases='', description='uncharacterized LOC101929154', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77884656, end_position_on_the_genomic_accession=77959096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  138159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138159, CurrentID=0, Status='live', Symbol='LOC138159', Aliases='', description='progestagen-associated endometrial protein pseudogene', other_designations='beta-lactoglobulin-like pseudogene', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135588006, end_position_on_the_genomic_accession=135591893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481243, CurrentID=0, Status='live', Symbol='RNU6-212P', Aliases='', description='RNA, U6 small nuclear 212, pseudogene', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138162, CurrentID=0, Status='live', Symbol='C9orf116', Aliases='PIERCE1, RbEST47', description='chromosome 9 open reading frame 116', other_designations='p53-induced expression 1 in Rb&#8722;/&#8722; cells|p53-induced expression in RB-null cells protein 1|pierce 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135495180, end_position_on_the_genomic_accession=135501832, orientation='minus', exon_count=4, OMIM=614502, no_hdr0=''),
  105374030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374030, CurrentID=0, Status='live', Symbol='LOC105374030', Aliases='', description='uncharacterized LOC105374030', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108088662, end_position_on_the_genomic_accession=108090856, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874210, CurrentID=0, Status='live', Symbol='ZMYND10-AS1', Aliases='', description='ZMYND10 antisense RNA 1', other_designations='ZMYND10 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481249, CurrentID=0, Status='live', Symbol='RNU6-226P', Aliases='', description='RNA, U6 small nuclear 226, pseudogene', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374031, CurrentID=0, Status='live', Symbol='LOC105374031', Aliases='', description='uncharacterized LOC105374031', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108097992, end_position_on_the_genomic_accession=108122067, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100271907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271907, CurrentID=0, Status='live', Symbol='LOC100271907', Aliases='', description='myotubularin related protein 7 pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154494746, end_position_on_the_genomic_accession=154496702, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723855, CurrentID=0, Status='live', Symbol='LOC102723855', Aliases='', description='uncharacterized LOC102723855', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133494941, end_position_on_the_genomic_accession=133500627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442236, CurrentID=0, Status='live', Symbol='KRT18P50', Aliases='', description='keratin 18 pseudogene 50', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=95991098, end_position_on_the_genomic_accession=95992401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996419, CurrentID=0, Status='live', Symbol='LOC100996419', Aliases='', description='uncharacterized LOC100996419', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179858904, end_position_on_the_genomic_accession=179861283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400322, CurrentID=0, Status='live', Symbol='HERC2P2', Aliases='D15F37S3, MN7', description='hect domain and RLD 2 pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22494837, end_position_on_the_genomic_accession=22590831, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  106481261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481261, CurrentID=0, Status='live', Symbol='RNU6-262P', Aliases='', description='RNA, U6 small nuclear 262, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420303, CurrentID=0, Status='live', Symbol='FTH1P26', Aliases='FTHL26', description='ferritin, heavy polypeptide 1 pseudogene 26', other_designations='ferritin, heavy polypeptide-like 26', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133676888, end_position_on_the_genomic_accession=133677401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372212, CurrentID=0, Status='live', Symbol='LOC105372212', Aliases='', description='uncharacterized LOC105372212', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76683331, end_position_on_the_genomic_accession=76687030, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481265, CurrentID=0, Status='live', Symbol='RNU6-275P', Aliases='', description='RNA, U6 small nuclear 275, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481266, CurrentID=0, Status='live', Symbol='RNU6-277P', Aliases='', description='RNA, U6 small nuclear 277, pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481268, CurrentID=0, Status='live', Symbol='RNU6-284P', Aliases='', description='RNA, U6 small nuclear 284, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378399, CurrentID=0, Status='live', Symbol='LOC105378399', Aliases='', description='uncharacterized LOC105378399', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83971590, end_position_on_the_genomic_accession=84003484, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481269, CurrentID=0, Status='live', Symbol='RNU6-287P', Aliases='', description='RNA, U6 small nuclear 287, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7115, CurrentID=0, Status='live', Symbol='TMSB4XP1', Aliases='TMSL1', description='thymosin beta 4, X-linked pseudogene 1', other_designations='thymosin-like 1 (pseudogene)', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42499781, end_position_on_the_genomic_accession=42500417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312774, CurrentID=0, Status='live', Symbol='VN1R56P', Aliases='', description='vomeronasal 1 receptor 56 pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3349174, end_position_on_the_genomic_accession=3349617, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312775, CurrentID=0, Status='live', Symbol='VN1R20P', Aliases='', description='vomeronasal 1 receptor 20 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13926861, end_position_on_the_genomic_accession=13927760, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374035, CurrentID=0, Status='live', Symbol='LOC105374035', Aliases='', description='uncharacterized LOC105374035', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109648036, end_position_on_the_genomic_accession=109836157, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481271, CurrentID=0, Status='live', Symbol='RNU6-293P', Aliases='', description='RNA, U6 small nuclear 293, pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100312781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312781, CurrentID=0, Status='live', Symbol='VN1R24P', Aliases='', description='vomeronasal 1 receptor 24 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56500885, end_position_on_the_genomic_accession=56501143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312782, CurrentID=0, Status='live', Symbol='VN1R25P', Aliases='', description='vomeronasal 1 receptor 25 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56559113, end_position_on_the_genomic_accession=56560024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312783, CurrentID=0, Status='live', Symbol='VN1R26P', Aliases='', description='vomeronasal 1 receptor 26 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56681145, end_position_on_the_genomic_accession=56681657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312784, CurrentID=0, Status='live', Symbol='VN1R27P', Aliases='', description='vomeronasal 1 receptor 27 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57160159, end_position_on_the_genomic_accession=57160532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312785, CurrentID=0, Status='live', Symbol='VN1R28P', Aliases='', description='vomeronasal 1 receptor 28 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57422785, end_position_on_the_genomic_accession=57423703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312786, CurrentID=0, Status='live', Symbol='VN1R29P', Aliases='', description='vomeronasal 1 receptor 29 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57476622, end_position_on_the_genomic_accession=57477492, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312787, CurrentID=0, Status='live', Symbol='VN1R31P', Aliases='', description='vomeronasal 1 receptor 31 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63367325, end_position_on_the_genomic_accession=63368167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312788, CurrentID=0, Status='live', Symbol='VN1R32P', Aliases='', description='vomeronasal 1 receptor 32 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63377333, end_position_on_the_genomic_accession=63378249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312789, CurrentID=0, Status='live', Symbol='VN1R35P', Aliases='', description='vomeronasal 1 receptor 35 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63963431, end_position_on_the_genomic_accession=63964092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312790, CurrentID=0, Status='live', Symbol='VN1R36P', Aliases='', description='vomeronasal 1 receptor 36 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63965778, end_position_on_the_genomic_accession=63966148, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312791, CurrentID=0, Status='live', Symbol='VN1R39P', Aliases='', description='vomeronasal 1 receptor 39 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64273538, end_position_on_the_genomic_accession=64273989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312792, CurrentID=0, Status='live', Symbol='VN1R40P', Aliases='', description='vomeronasal 1 receptor 40 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64443760, end_position_on_the_genomic_accession=64444676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312793, CurrentID=0, Status='live', Symbol='VN1R41P', Aliases='', description='vomeronasal 1 receptor 41 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64475557, end_position_on_the_genomic_accession=64476162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312794, CurrentID=0, Status='live', Symbol='VN1R42P', Aliases='', description='vomeronasal 1 receptor 42 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64933273, end_position_on_the_genomic_accession=64933751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312795, CurrentID=0, Status='live', Symbol='VN1R43P', Aliases='', description='vomeronasal 1 receptor 43 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64972917, end_position_on_the_genomic_accession=64973224, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312796, CurrentID=0, Status='live', Symbol='VN1R45P', Aliases='', description='vomeronasal 1 receptor 45 pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43216484, end_position_on_the_genomic_accession=43216613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312797, CurrentID=0, Status='live', Symbol='VN1R47P', Aliases='', description='vomeronasal 1 receptor 47 pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33512526, end_position_on_the_genomic_accession=33512709, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312798, CurrentID=0, Status='live', Symbol='VN1R48P', Aliases='', description='vomeronasal 1 receptor 48 pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38504501, end_position_on_the_genomic_accession=38505437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312799, CurrentID=0, Status='live', Symbol='VN1R49P', Aliases='VN1R50P', description='vomeronasal 1 receptor 49 pseudogene', other_designations='vomeronasal 1 receptor 50 pseudogene', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64839439, end_position_on_the_genomic_accession=64840173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312801, CurrentID=0, Status='live', Symbol='VN1R51P', Aliases='', description='vomeronasal 1 receptor 51 pseudogene', other_designations='vomeronasal 1 receptor 1 pseudogene', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97093736, end_position_on_the_genomic_accession=97094638, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312802, CurrentID=0, Status='live', Symbol='VN1R52P', Aliases='', description='vomeronasal 1 receptor 52 pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97235268, end_position_on_the_genomic_accession=97235453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312803, CurrentID=0, Status='live', Symbol='VN1R110P', Aliases='', description='vomeronasal 1 receptor 110 pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48636228, end_position_on_the_genomic_accession=48637159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312804, CurrentID=0, Status='live', Symbol='VN1R15P', Aliases='', description='vomeronasal 1 receptor 15 pseudogene', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227659167, end_position_on_the_genomic_accession=227659730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312805, CurrentID=0, Status='live', Symbol='VN1R16P', Aliases='', description='vomeronasal 1 receptor 16 pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247180049, end_position_on_the_genomic_accession=247180963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312806, CurrentID=0, Status='live', Symbol='VN1R53P', Aliases='', description='vomeronasal 1 receptor 53 pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37348616, end_position_on_the_genomic_accession=37349468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312807, CurrentID=0, Status='live', Symbol='VN1R54P', Aliases='', description='vomeronasal 1 receptor 54 pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42660844, end_position_on_the_genomic_accession=42661730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312808, CurrentID=0, Status='live', Symbol='VN1R55P', Aliases='', description='vomeronasal 1 receptor 55 pseudogene', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88258557, end_position_on_the_genomic_accession=88259303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312809, CurrentID=0, Status='live', Symbol='VN1R57P', Aliases='', description='vomeronasal 1 receptor 57 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81868368, end_position_on_the_genomic_accession=81869098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312810, CurrentID=0, Status='live', Symbol='VN1R58P', Aliases='', description='vomeronasal 1 receptor 58 pseudogene', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18463430, end_position_on_the_genomic_accession=18463903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312811, CurrentID=0, Status='live', Symbol='VN1R59P', Aliases='', description='vomeronasal 1 receptor 59 pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60544394, end_position_on_the_genomic_accession=60544756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312812, CurrentID=0, Status='live', Symbol='VN1R60P', Aliases='', description='vomeronasal 1 receptor 60 pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20890117, end_position_on_the_genomic_accession=20890615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312813, CurrentID=0, Status='live', Symbol='VN1R61P', Aliases='', description='vomeronasal 1 receptor 61 pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20893107, end_position_on_the_genomic_accession=20893962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312814, CurrentID=0, Status='live', Symbol='VN1R62P', Aliases='', description='vomeronasal 1 receptor 62 pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21901097, end_position_on_the_genomic_accession=21901595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312815, CurrentID=0, Status='live', Symbol='VN1R63P', Aliases='', description='vomeronasal 1 receptor 63 pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21904087, end_position_on_the_genomic_accession=21904942, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312816, CurrentID=0, Status='live', Symbol='VN1R69P', Aliases='', description='vomeronasal 1 receptor 69 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35134465, end_position_on_the_genomic_accession=35135423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312817, CurrentID=0, Status='live', Symbol='VN1R71P', Aliases='', description='vomeronasal 1 receptor 71 pseudogene', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27057398, end_position_on_the_genomic_accession=27058345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312818, CurrentID=0, Status='live', Symbol='VN1R73P', Aliases='', description='vomeronasal 1 receptor 73 pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14075240, end_position_on_the_genomic_accession=14075740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312819, CurrentID=0, Status='live', Symbol='VN1R74P', Aliases='', description='vomeronasal 1 receptor 74 pseudogene', other_designations='vomeronasal 1 receptor 2 pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14680406, end_position_on_the_genomic_accession=14681322, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312820, CurrentID=0, Status='live', Symbol='VN1R75P', Aliases='', description='vomeronasal 1 receptor 75 pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15276415, end_position_on_the_genomic_accession=15276888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312821, CurrentID=0, Status='live', Symbol='VN1R76P', Aliases='', description='vomeronasal 1 receptor 76 pseudogene', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=52419157, end_position_on_the_genomic_accession=52419653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312822, CurrentID=0, Status='live', Symbol='VN1R77P', Aliases='', description='vomeronasal 1 receptor 77 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19979291, end_position_on_the_genomic_accession=19979610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312823, CurrentID=0, Status='live', Symbol='VN1R78P', Aliases='', description='vomeronasal 1 receptor 78 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20582930, end_position_on_the_genomic_accession=20583647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312824, CurrentID=0, Status='live', Symbol='VN1R79P', Aliases='', description='vomeronasal 1 receptor 79 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20758464, end_position_on_the_genomic_accession=20759138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312825, CurrentID=0, Status='live', Symbol='VN1R80P', Aliases='', description='vomeronasal 1 receptor 80 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21063924, end_position_on_the_genomic_accession=21064261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312826, CurrentID=0, Status='live', Symbol='VN1R81P', Aliases='', description='vomeronasal 1 receptor 81 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21123817, end_position_on_the_genomic_accession=21124532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312827, CurrentID=0, Status='live', Symbol='VN1R82P', Aliases='', description='vomeronasal 1 receptor 82 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21188540, end_position_on_the_genomic_accession=21189092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312828, CurrentID=0, Status='live', Symbol='VN1R83P', Aliases='', description='vomeronasal 1 receptor 83 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21289554, end_position_on_the_genomic_accession=21289998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312829, CurrentID=0, Status='live', Symbol='VN1R84P', Aliases='', description='vomeronasal 1 receptor 84 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21719627, end_position_on_the_genomic_accession=21720140, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312830, CurrentID=0, Status='live', Symbol='VN1R85P', Aliases='', description='vomeronasal 1 receptor 85 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22174766, end_position_on_the_genomic_accession=22175191, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312831, CurrentID=0, Status='live', Symbol='VN1R86P', Aliases='', description='vomeronasal 1 receptor 86 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22476506, end_position_on_the_genomic_accession=22476814, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312832, CurrentID=0, Status='live', Symbol='VN1R87P', Aliases='', description='vomeronasal 1 receptor 87 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22742924, end_position_on_the_genomic_accession=22743410, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312833, CurrentID=0, Status='live', Symbol='VN1R88P', Aliases='', description='vomeronasal 1 receptor 88 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22861205, end_position_on_the_genomic_accession=22861606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312834, CurrentID=0, Status='live', Symbol='VN1R89P', Aliases='', description='vomeronasal 1 receptor 89 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22970522, end_position_on_the_genomic_accession=22970916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312835, CurrentID=0, Status='live', Symbol='VN1R90P', Aliases='', description='vomeronasal 1 receptor 90 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23288614, end_position_on_the_genomic_accession=23289358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312836, CurrentID=0, Status='live', Symbol='VN1R91P', Aliases='', description='vomeronasal 1 receptor 91 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23290129, end_position_on_the_genomic_accession=23291017, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312837, CurrentID=0, Status='live', Symbol='VN1R92P', Aliases='', description='vomeronasal 1 receptor 92 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23444288, end_position_on_the_genomic_accession=23445189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312838, CurrentID=0, Status='live', Symbol='VN1R93P', Aliases='', description='vomeronasal 1 receptor 93 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23602567, end_position_on_the_genomic_accession=23603119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312839, CurrentID=0, Status='live', Symbol='VN1R94P', Aliases='', description='vomeronasal 1 receptor 94 pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23739602, end_position_on_the_genomic_accession=23739915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312840, CurrentID=0, Status='live', Symbol='VN1R97P', Aliases='', description='vomeronasal 1 receptor 97 pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48251755, end_position_on_the_genomic_accession=48252102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312841, CurrentID=0, Status='live', Symbol='VN1R98P', Aliases='', description='vomeronasal 1 receptor 98 pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51856700, end_position_on_the_genomic_accession=51857011, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312842, CurrentID=0, Status='live', Symbol='VN1R99P', Aliases='', description='vomeronasal 1 receptor 99 pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52241729, end_position_on_the_genomic_accession=52242471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  138199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138199, CurrentID=0, Status='live', Symbol='CARNMT1', Aliases='C9orf41', description='carnosine N-methyltransferase 1', other_designations='UPF0586 protein C9orf41', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74981020, end_position_on_the_genomic_accession=75028486, orientation='minus', exon_count=12, OMIM=616552, no_hdr0=''),
  100312844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312844, CurrentID=0, Status='live', Symbol='VN1R22P', Aliases='', description='vomeronasal 1 receptor 22 pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107135657, end_position_on_the_genomic_accession=107136055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312845, CurrentID=0, Status='live', Symbol='VN1R23P', Aliases='', description='vomeronasal 1 receptor 23 pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107136990, end_position_on_the_genomic_accession=107137289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312846, CurrentID=0, Status='live', Symbol='VN1R100P', Aliases='', description='vomeronasal 1 receptor 100 pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52263966, end_position_on_the_genomic_accession=52264398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312847, CurrentID=0, Status='live', Symbol='VN1R102P', Aliases='', description='vomeronasal 1 receptor 102 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53462968, end_position_on_the_genomic_accession=53463153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312848, CurrentID=0, Status='live', Symbol='VN1R106P', Aliases='FKSG41', description='vomeronasal 1 receptor 106 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55707577, end_position_on_the_genomic_accession=55707858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312849, CurrentID=0, Status='live', Symbol='VN1R107P', Aliases='', description='vomeronasal 1 receptor 107 pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57459911, end_position_on_the_genomic_accession=57460120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312850, CurrentID=0, Status='live', Symbol='VN1R108P', Aliases='', description='vomeronasal 1 receptor 108 pseudogene', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25734229, end_position_on_the_genomic_accession=25735114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100312851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312851, CurrentID=0, Status='live', Symbol='VN1R109P', Aliases='', description='vomeronasal 1 receptor 109 pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13091616, end_position_on_the_genomic_accession=13092089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312852, CurrentID=0, Status='live', Symbol='VN1R111P', Aliases='', description='vomeronasal 1 receptor 111 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18180390, end_position_on_the_genomic_accession=18180623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100312853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312853, CurrentID=0, Status='live', Symbol='VN1R72P', Aliases='', description='vomeronasal 1 receptor 72 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100312854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312854, CurrentID=0, Status='live', Symbol='VN1R112P', Aliases='', description='vomeronasal 1 receptor 112 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100312855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312855, CurrentID=0, Status='live', Symbol='VN1R44P', Aliases='', description='vomeronasal 1 receptor 44 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481285, CurrentID=0, Status='live', Symbol='RNU6-336P', Aliases='', description='RNA, U6 small nuclear 336, pseudogene', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481287, CurrentID=0, Status='live', Symbol='RNU6-341P', Aliases='', description='RNA, U6 small nuclear 341, pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723891, CurrentID=0, Status='live', Symbol='LOC102723891', Aliases='', description='ankyrin repeat domain-containing protein 20B-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369674, CurrentID=0, Status='live', Symbol='LOC105369674', Aliases='', description='uncharacterized LOC105369674', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15246459, end_position_on_the_genomic_accession=15265279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374039, CurrentID=0, Status='live', Symbol='LOC105374039', Aliases='', description='uncharacterized LOC105374039', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111306356, end_position_on_the_genomic_accession=111365814, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481291, CurrentID=0, Status='live', Symbol='RNU6-358P', Aliases='', description='RNA, U6 small nuclear 358, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481293, CurrentID=0, Status='live', Symbol='RNU6-364P', Aliases='', description='RNA, U6 small nuclear 364, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481294, CurrentID=0, Status='live', Symbol='RNU6-366P', Aliases='', description='RNA, U6 small nuclear 366, pseudogene', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723897, CurrentID=0, Status='live', Symbol='LOC102723897', Aliases='', description='WAS protein family homolog 2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184916, end_position_on_the_genomic_accession=200470, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106481295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481295, CurrentID=0, Status='live', Symbol='RNU6-370P', Aliases='', description='RNA, U6 small nuclear 370, pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374040, CurrentID=0, Status='live', Symbol='LOC105374040', Aliases='', description='uncharacterized LOC105374040', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111803811, end_position_on_the_genomic_accession=111861908, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481296, CurrentID=0, Status='live', Symbol='RNU6-374P', Aliases='', description='RNA, U6 small nuclear 374, pseudogene', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481297, CurrentID=0, Status='live', Symbol='RNU6-378P', Aliases='', description='RNA, U6 small nuclear 378, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400359, CurrentID=0, Status='live', Symbol='C15orf53', Aliases='', description='chromosome 15 open reading frame 53', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38696598, end_position_on_the_genomic_accession=38700038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481298, CurrentID=0, Status='live', Symbol='RNU6-380P', Aliases='', description='RNA, U6 small nuclear 380, pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930434, CurrentID=0, Status='live', Symbol='LOC101930434', Aliases='', description='golgin subfamily A member 8H-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30791502, end_position_on_the_genomic_accession=30801965, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481299, CurrentID=0, Status='live', Symbol='RNU6-384P', Aliases='', description='RNA, U6 small nuclear 384, pseudogene', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100312951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312951, CurrentID=0, Status='live', Symbol='NPCA2', Aliases='', description='Nasopharyngeal carcinoma, susceptibility to, 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=161550, no_hdr0=''),
  100312952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312952, CurrentID=0, Status='live', Symbol='PVOP1', Aliases='', description='Pelvic organ prolapse, susceptibility to, 1', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=176780, no_hdr0=''),
  106481300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481300, CurrentID=0, Status='live', Symbol='RNU6-386P', Aliases='', description='RNA, U6 small nuclear 386, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100312954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312954, CurrentID=0, Status='live', Symbol='GRD1', Aliases='', description='Graves disease, susceptiblity to, 1', other_designations='', map_location='14q31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=275000, no_hdr0=''),
  100312955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312955, CurrentID=0, Status='live', Symbol='PVOP2', Aliases='', description='Pelvic organ prolapse, susceptibility to, 2', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613088, no_hdr0=''),
  105374041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374041, CurrentID=0, Status='live', Symbol='LOC105374041', Aliases='', description='uncharacterized LOC105374041', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112375689, end_position_on_the_genomic_accession=112404394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481301, CurrentID=0, Status='live', Symbol='RNU6-389P', Aliases='', description='RNA, U6 small nuclear 389, pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481302, CurrentID=0, Status='live', Symbol='RNU6-393P', Aliases='', description='RNA, U6 small nuclear 393, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481303, CurrentID=0, Status='live', Symbol='RNU6-396P', Aliases='', description='RNA, U6 small nuclear 396, pseudogene', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420311, CurrentID=0, Status='live', Symbol='LOC100420311', Aliases='', description='TBC1 domain family, member 3 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30556345, end_position_on_the_genomic_accession=30558923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481304, CurrentID=0, Status='live', Symbol='RNU6-400P', Aliases='', description='RNA, U6 small nuclear 400, pseudogene', other_designations='', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104355287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355287, CurrentID=0, Status='live', Symbol='LINC01317', Aliases='', description='long intergenic non-protein coding RNA 1317', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33706886, end_position_on_the_genomic_accession=34297746, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  442245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442245, CurrentID=0, Status='live', Symbol='GSTM2P1', Aliases='', description='glutathione S-transferase mu 2 (muscle) pseudogene 1', other_designations='glutathione S-transferase mu 2 pseudogene', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111046419, end_position_on_the_genomic_accession=111047554, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100312985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312985, CurrentID=0, Status='live', Symbol='IGKV1OR15-118', Aliases='IGKV1/OR-118, IGKV1/OR15-118, IGKV1OR118, IGKVP2', description='immunoglobulin kappa variable 1/OR15-118 (pseudogene)', other_designations='immunoglobulin kappa variable 1/OR-118|immunoglobulin kappa variable 1/OR15-118 pseudogene|immunoglobulin kappa variable region pseudogene 2', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533789, CurrentID=0, Status='live', Symbol='LRRC37A8P', Aliases='', description='leucine-rich repeat containing 37 member A8, pseudogene', other_designations='leucine rich repeat containing 37, member A3 pseudogene|leucine rich repeat containing 37, member A8, pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35904220, end_position_on_the_genomic_accession=35911023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723909, CurrentID=0, Status='live', Symbol='LOC102723909', Aliases='', description='uncharacterized LOC102723909', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9102705, end_position_on_the_genomic_accession=9110351, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  28999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28999, CurrentID=0, Status='live', Symbol='KLF15', Aliases='KKLF', description='Kruppel-like factor 15', other_designations='kidney-enriched Kruppel-like factor|kidney-enriched krueppel-like factor', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126342635, end_position_on_the_genomic_accession=126357393, orientation='minus', exon_count=5, OMIM=606465, no_hdr0=''),
  106481310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481310, CurrentID=0, Status='live', Symbol='RNU6-418P', Aliases='', description='RNA, U6 small nuclear 418, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643014, CurrentID=0, Status='live', Symbol='LOC643014', Aliases='', description='coiled-coil domain containing 90B pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74085998, end_position_on_the_genomic_accession=74086672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533790, CurrentID=0, Status='live', Symbol='LRRC37A9P', Aliases='', description='leucine-rich repeat containing 37 member A9, pseudogene', other_designations='leucine rich repeat containing 37, member A3 pseudogene|leucine rich repeat containing 37, member A9, pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35911572, end_position_on_the_genomic_accession=35914842, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481312, CurrentID=0, Status='live', Symbol='RNU6-425P', Aliases='', description='RNA, U6 small nuclear 425, pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369679, CurrentID=0, Status='live', Symbol='LOC105369679', Aliases='', description='uncharacterized LOC105369679', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18714851, end_position_on_the_genomic_accession=18726817, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643015, CurrentID=0, Status='live', Symbol='LOC643015', Aliases='', description='nucleolar protein 11 pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150114366, end_position_on_the_genomic_accession=150116821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533791, CurrentID=0, Status='live', Symbol='LRRC37A10P', Aliases='', description='leucine-rich repeat containing 37 member A10, pseudogene', other_designations='leucine rich repeat containing 37, member A10, pseudogene|leucine rich repeat containing 37, member A3 pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43922775, end_position_on_the_genomic_accession=43926343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481316, CurrentID=0, Status='live', Symbol='RNU6-436P', Aliases='', description='RNA, U6 small nuclear 436, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138234, CurrentID=0, Status='live', Symbol='ATP5HP2', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit D pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d, pseudogene 2', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2900240, end_position_on_the_genomic_accession=2900846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481317, CurrentID=0, Status='live', Symbol='RNU6-439P', Aliases='', description='RNA, U6 small nuclear 439, pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481318, CurrentID=0, Status='live', Symbol='RNU6-443P', Aliases='', description='RNA, U6 small nuclear 443, pseudogene', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481319, CurrentID=0, Status='live', Symbol='RNU6-446P', Aliases='', description='RNA, U6 small nuclear 446, pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481320, CurrentID=0, Status='live', Symbol='RNU6-447P', Aliases='', description='RNA, U6 small nuclear 447, pseudogene', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105031671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105031671, CurrentID=0, Status='live', Symbol='VVS', Aliases='', description='Syncope, familial vasovagal', other_designations='', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609289, no_hdr0=''),
  100188995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188995, CurrentID=0, Status='live', Symbol='TRK-CTT6-1', Aliases='TRNAK4', description='transfer RNA-Lys (CTT) 6-1', other_designations='tRNA-Lys-CTT-6-1|transfer RNA lysine 4 (anticodon CUU)', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46089305, end_position_on_the_genomic_accession=46089377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271921, CurrentID=0, Status='live', Symbol='DEL15Q26QTER', Aliases='C15DELq26qter', description='Chromosome 15q26-qter deletion syndrome', other_designations='', map_location='15q26-qter', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612626, no_hdr0=''),
  138240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138240, CurrentID=0, Status='live', Symbol='C9orf57', Aliases='', description='chromosome 9 open reading frame 57', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72051381, end_position_on_the_genomic_accession=72060605, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100127892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127892, CurrentID=0, Status='live', Symbol='LOC100127892', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113984021, end_position_on_the_genomic_accession=113984655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481323, CurrentID=0, Status='live', Symbol='RNU6-456P', Aliases='', description='RNA, U6 small nuclear 456, pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138241, CurrentID=0, Status='live', Symbol='C9orf85', Aliases='', description='chromosome 9 open reading frame 85', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=71911321, end_position_on_the_genomic_accession=71973459, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  285126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285126, CurrentID=0, Status='live', Symbol='DNAJC5G', Aliases='CSP-gamma', description='DnaJ heat shock protein family (Hsp40) member C5 gamma', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5 gamma|cysteine string protein-gamma|gamma cysteine string protein|gamma-CSP|gamma-cysteine string protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27275421, end_position_on_the_genomic_accession=27281430, orientation='plus', exon_count=7, OMIM=613946, no_hdr0=''),
  106481325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481325, CurrentID=0, Status='live', Symbol='RNU6-463P', Aliases='', description='RNA, U6 small nuclear 463, pseudogene', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442249, CurrentID=0, Status='live', Symbol='KRT18P65', Aliases='', description='keratin 18 pseudogene 65', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112361383, end_position_on_the_genomic_accession=112362767, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481326, CurrentID=0, Status='live', Symbol='RNU6-465P', Aliases='', description='RNA, U6 small nuclear 465, pseudogene', other_designations='', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481327, CurrentID=0, Status='live', Symbol='RNU6-469P', Aliases='', description='RNA, U6 small nuclear 469, pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400389, CurrentID=0, Status='live', Symbol='RPL12P35', Aliases='RPL12_13_1452', description='ribosomal protein L12 pseudogene 35', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72379174, end_position_on_the_genomic_accession=72379798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481329, CurrentID=0, Status='live', Symbol='RNU6-474P', Aliases='', description='RNA, U6 small nuclear 474, pseudogene', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723933, CurrentID=0, Status='live', Symbol='LOC102723933', Aliases='', description='uncharacterized LOC102723933', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38153235, end_position_on_the_genomic_accession=38165697, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100533794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533794, CurrentID=0, Status='live', Symbol='MARK2P15', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 15', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83312369, end_position_on_the_genomic_accession=83314451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271923, CurrentID=0, Status='live', Symbol='FEB10', Aliases='', description='Febrile convulsions, familial, 10', other_designations='', map_location='3q26.2-q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612637, no_hdr0=''),
  106481333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481333, CurrentID=0, Status='live', Symbol='RNU6-485P', Aliases='', description='RNA, U6 small nuclear 485, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533795, CurrentID=0, Status='live', Symbol='MARK2P16', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 16', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89837545, end_position_on_the_genomic_accession=89839334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271924, CurrentID=0, Status='live', Symbol='DFNA59', Aliases='', description='deafness, autosomal dominant 59', other_designations='', map_location='11p14.2-q12.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612642, no_hdr0=''),
  7183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7183, CurrentID=0, Status='live', Symbol='TRC-GCA24-1', Aliases='TRC, TRNAC1', description='transfer RNA-Cys (GCA) 24-1', other_designations='tRNA cysteine|tRNA-Cys-GCA-24-1|transfer RNA cysteine 1 (anticodon GCA)', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372215, CurrentID=0, Status='live', Symbol='LOC105372215', Aliases='', description='uncharacterized LOC105372215', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77087368, end_position_on_the_genomic_accession=77094009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442252, CurrentID=0, Status='live', Symbol='KRT18P22', Aliases='', description='keratin 18 pseudogene 22', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116457260, end_position_on_the_genomic_accession=116458655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  400403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400403, CurrentID=0, Status='live', Symbol='DNM1P9', Aliases='DNM1DN9, DNM1DN9-1, DNM1DN9@', description='dynamin 1 pseudogene 9', other_designations='DNM1 duplicon 9 cluster|DNM1 pseudogene 9', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77906201, end_position_on_the_genomic_accession=77911621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928631, CurrentID=0, Status='live', Symbol='LOC101928631', Aliases='', description='uncharacterized LOC101928631', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2908628, end_position_on_the_genomic_accession=2926778, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723945, CurrentID=0, Status='live', Symbol='LOC102723945', Aliases='', description='sodium/hydrogen exchanger 9B1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7190, CurrentID=0, Status='live', Symbol='HSP90B2P', Aliases='GRP94P1, GRP94b, HSP, HSPCP2, TRA1P1, TRAP1', description='heat shock protein 90kDa beta family member 2, pseudogene', other_designations='heat shock protein 90kDa beta (Grp94), member 1 pseudogene|heat shock protein 90kDa beta (Grp94), member 2, pseudogene|heat shock protein 94b|tumor rejection antigen (gp96) 1 pseudogene 1', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99257525, end_position_on_the_genomic_accession=99260276, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481345, CurrentID=0, Status='live', Symbol='RNU6-520P', Aliases='', description='RNA, U6 small nuclear 520, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379488, CurrentID=0, Status='live', Symbol='LOC105379488', Aliases='', description='uncharacterized LOC105379488', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5506972, end_position_on_the_genomic_accession=5509729, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374050, CurrentID=0, Status='live', Symbol='LOC105374050', Aliases='', description='uncharacterized LOC105374050', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114530133, end_position_on_the_genomic_accession=114554889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723951, CurrentID=0, Status='live', Symbol='LOC102723951', Aliases='', description='uncharacterized LOC102723951', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24142345, end_position_on_the_genomic_accession=24153231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374051, CurrentID=0, Status='live', Symbol='LOC105374051', Aliases='', description='uncharacterized LOC105374051', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115437624, end_position_on_the_genomic_accession=115438600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189001, CurrentID=0, Status='live', Symbol='TRY-GTA5-3', Aliases='TRNAY3', description='transfer RNA-Tyr (GTA) 5-3', other_designations='transfer RNA tyrosine 3 (anticodon GUA)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20683273, end_position_on_the_genomic_accession=20683361, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271927, CurrentID=0, Status='live', Symbol='RASA4B', Aliases='', description='RAS p21 protein activator 4B', other_designations='putative Ras GTPase-activating protein 4B', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102483139, end_position_on_the_genomic_accession=102517777, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105369687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369687, CurrentID=0, Status='live', Symbol='LOC105369687', Aliases='', description='uncharacterized LOC105369687', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20212470, end_position_on_the_genomic_accession=20215215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874232, CurrentID=0, Status='live', Symbol='C1QTNF9-AS1', Aliases='', description='C1QTNF9 antisense RNA 1', other_designations='C1QTNF9 antisense RNA 1 (non-protein coding)', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106866913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106866913, CurrentID=0, Status='live', Symbol='RN7SL431P', Aliases='', description='RNA, 7SL, cytoplasmic 431, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106866914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106866914, CurrentID=0, Status='live', Symbol='RNU6-1240P', Aliases='', description='RNA, U6 small nuclear 1240, pseudogene', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106866915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106866915, CurrentID=0, Status='live', Symbol='RNU1-125P', Aliases='', description='RNA, U1 small nuclear 125, pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533800, CurrentID=0, Status='live', Symbol='BMPR1APS2', Aliases='', description='bone morphogenetic protein receptor type IA pseudogene 2', other_designations='bone morphogenetic protein receptor, type IA pseudogene 2', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121361853, end_position_on_the_genomic_accession=121362922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100127900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127900, CurrentID=0, Status='live', Symbol='LOC100127900', Aliases='', description='nucleoporin 50kDa pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=74849077, end_position_on_the_genomic_accession=74850403, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100124373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124373, CurrentID=0, Status='live', Symbol='LOC100124373', Aliases='', description='ankyrin repeat domain 39 pseudogene', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39353845, end_position_on_the_genomic_accession=39354498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928632, CurrentID=0, Status='live', Symbol='LOC101928632', Aliases='', description='uncharacterized LOC101928632', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130354731, end_position_on_the_genomic_accession=130357168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100420324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420324, CurrentID=0, Status='live', Symbol='LOC100420324', Aliases='', description='melanoma antigen family A10 pseudogene', other_designations='melanoma antigen family A, 10 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30209387, end_position_on_the_genomic_accession=30210238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129168, CurrentID=0, Status='live', Symbol='LOC100129168', Aliases='', description='chromosome 6 open reading frame 203 pseudogene', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31452452, end_position_on_the_genomic_accession=31453284, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106866980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106866980, CurrentID=0, Status='live', Symbol='RNU2-15P', Aliases='', description='RNA, U2 small nuclear 15, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106866981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106866981, CurrentID=0, Status='live', Symbol='RNU6-244P', Aliases='', description='RNA, U6 small nuclear 244, pseudogene', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400433, CurrentID=0, Status='live', Symbol='DNM1P40', Aliases='DNM1DN11-5, DNM1DN11.5, DNM1DN11@', description='dynamin 1 pseudogene 40', other_designations='DNM1 pseudogene 40|DNM1DN11.5 duplicon', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84503751, end_position_on_the_genomic_accession=84506432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874235, CurrentID=0, Status='live', Symbol='CACNA1C-AS2', Aliases='', description='CACNA1C antisense RNA 2', other_designations='CACNA1C antisense RNA 2 (non-protein coding)', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2668500, end_position_on_the_genomic_accession=2672220, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  139952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139952, CurrentID=0, Status='live', Symbol='LOC139952', Aliases='', description='LIM domain kinase 2 pseudogene', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11704123, end_position_on_the_genomic_accession=11705611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374056, CurrentID=0, Status='live', Symbol='LOC105374056', Aliases='', description='uncharacterized LOC105374056', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117513086, end_position_on_the_genomic_accession=117539770, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  160140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160140, CurrentID=0, Status='live', Symbol='C11orf65', Aliases='', description='chromosome 11 open reading frame 65', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108331426, end_position_on_the_genomic_accession=108467547, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100127903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127903, CurrentID=0, Status='live', Symbol='LOC100127903', Aliases='', description='speckle type POZ protein like pseudogene', other_designations='speckle-type POZ protein-like pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69125214, end_position_on_the_genomic_accession=69128072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377454, CurrentID=0, Status='live', Symbol='LOC105377454', Aliases='', description='uncharacterized LOC105377454', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140301437, end_position_on_the_genomic_accession=140303609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100189007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189007, CurrentID=0, Status='live', Symbol='TRE-CTC3-1', Aliases='TRNAE4', description='transfer RNA-Glu (CTC) 3-1', other_designations='tRNA-Glu-CTC-3-1|transfer RNA glutamic acid 4 (anticodon CUC)', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127904, CurrentID=0, Status='live', Symbol='LOC100127904', Aliases='', description='uncharacterized LOC100127904', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=45379758, end_position_on_the_genomic_accession=45380744, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  400446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400446, CurrentID=0, Status='live', Symbol='LOC400446', Aliases='', description='PERP, TP53 apoptosis effector pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90275046, end_position_on_the_genomic_accession=90277609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723989, CurrentID=0, Status='live', Symbol='LOC102723989', Aliases='', description='uncharacterized LOC102723989', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83219300, end_position_on_the_genomic_accession=83270833, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  138307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138307, CurrentID=0, Status='live', Symbol='LCN8', Aliases='EP17, LCN5', description='lipocalin 8', other_designations='lipocalin 5', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136754388, end_position_on_the_genomic_accession=136758279, orientation='minus', exon_count=6, OMIM=612902, no_hdr0=''),
  442262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442262, CurrentID=0, Status='live', Symbol='GAPDHP73', Aliases='GAPDHL19, GAPDL19', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 73', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 19, pseudogene', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135618967, end_position_on_the_genomic_accession=135620245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101362076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101362076, CurrentID=0, Status='live', Symbol='GVQW1', Aliases='TIGD1L2, bA205M20.5', description='GVQW motif containing 1', other_designations='tigger transposable element derived 1-like 2', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481959, CurrentID=0, Status='live', Symbol='RNU2-19P', Aliases='', description='RNA, U2 small nuclear 19, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7239, CurrentID=0, Status='live', Symbol='TRV-AAC1-4', Aliases='TRNAV1, TRV, TRV1', description='transfer RNA-Val (AAC) 1-4', other_designations='tRNA valine 1|tRNA-Val-AAC-1-4|transfer RNA valine 1 (anticodon AAC)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181218270, end_position_on_the_genomic_accession=181218342, orientation='minus', exon_count=0, OMIM=189921, no_hdr0=''),
  400456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400456, CurrentID=0, Status='live', Symbol='LINC01197', Aliases='', description='long intergenic non-protein coding RNA 1197', other_designations='CTD-2536I1.1', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95279288, end_position_on_the_genomic_accession=95327100, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  442263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442263, CurrentID=0, Status='live', Symbol='LOC442263', Aliases='', description='protein tyrosine phosphatase, non-receptor type 11 pseudogene', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137705823, end_position_on_the_genomic_accession=137707220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374060, CurrentID=0, Status='live', Symbol='LOC105374060', Aliases='', description='uncharacterized LOC105374060', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118507824, end_position_on_the_genomic_accession=118802454, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  102723999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723999, CurrentID=0, Status='live', Symbol='LOC102723999', Aliases='', description='rho GTPase-activating protein 23-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313532, CurrentID=0, Status='live', Symbol='MFSD1P1', Aliases='', description='major facilitator superfamily domain containing 1 pseudogene 1', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4301247, end_position_on_the_genomic_accession=4303382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481397, CurrentID=0, Status='live', Symbol='RNU6-686P', Aliases='', description='RNA, U6 small nuclear 686, pseudogene', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  394254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394254, CurrentID=0, Status='live', Symbol='GPR89P', Aliases='C11H1.2-Lp, SH120P', description='G protein-coupled receptor 89 pseudogene', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27736575, end_position_on_the_genomic_accession=27738514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421775, CurrentID=0, Status='live', Symbol='LOC100421775', Aliases='', description='leucine-rich repeat containing 8 family member B pseudogene', other_designations='leucine rich repeat containing 8 family, member B pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131825984, end_position_on_the_genomic_accession=131828413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481401, CurrentID=0, Status='live', Symbol='RNU6-698P', Aliases='', description='RNA, U6 small nuclear 698, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400464, CurrentID=0, Status='live', Symbol='LOC400464', Aliases='', description='uncharacterized LOC400464', other_designations='CTD-2054N24.2', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99807789, end_position_on_the_genomic_accession=99881880, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106481403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481403, CurrentID=0, Status='live', Symbol='RNU6-705P', Aliases='', description='RNA, U6 small nuclear 705, pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369697, CurrentID=0, Status='live', Symbol='LOC105369697', Aliases='', description='uncharacterized LOC105369697', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=23626960, end_position_on_the_genomic_accession=23640625, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  394255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394255, CurrentID=0, Status='live', Symbol='RPS10P1', Aliases='RPS10_7_669, RPS10p, dJ34B20.13', description='ribosomal protein S10 pseudogene 1', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26202123, end_position_on_the_genomic_accession=26202715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481404, CurrentID=0, Status='live', Symbol='RNU6-712P', Aliases='', description='RNA, U6 small nuclear 712, pseudogene', other_designations='', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481405, CurrentID=0, Status='live', Symbol='RNU6-715P', Aliases='', description='RNA, U6 small nuclear 715, pseudogene', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728147, CurrentID=0, Status='live', Symbol='LOC728147', Aliases='', description='uncharacterized LOC728147', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=2901735, end_position_on_the_genomic_accession=2904038, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481406, CurrentID=0, Status='live', Symbol='RNU6-718P', Aliases='', description='RNA, U6 small nuclear 718, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  650181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650181, CurrentID=0, Status='live', Symbol='LOC650181', Aliases='', description='AhpC/TSA antioxidant enzyme domain containing 1 pseudogene', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65356889, end_position_on_the_genomic_accession=65357520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481407, CurrentID=0, Status='live', Symbol='RNU6-722P', Aliases='', description='RNA, U6 small nuclear 722, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481408, CurrentID=0, Status='live', Symbol='RNU6-725P', Aliases='', description='RNA, U6 small nuclear 725, pseudogene', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130463, CurrentID=0, Status='live', Symbol='LOC100130463', Aliases='', description='uncharacterized LOC100130463', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=114075870, end_position_on_the_genomic_accession=114076649, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481409, CurrentID=0, Status='live', Symbol='RNU6-727P', Aliases='', description='RNA, U6 small nuclear 727, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7255, CurrentID=0, Status='live', Symbol='TSHRL2', Aliases='', description='thyroid stimulating hormone receptor-like 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  126860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126860, CurrentID=0, Status='live', Symbol='LOC126860', Aliases='', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 1 pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179586705, end_position_on_the_genomic_accession=179589174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481411, CurrentID=0, Status='live', Symbol='RNU6-731P', Aliases='', description='RNA, U6 small nuclear 731, pseudogene', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481412, CurrentID=0, Status='live', Symbol='RNU6-735P', Aliases='', description='RNA, U6 small nuclear 735, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373681, CurrentID=0, Status='live', Symbol='LOC105373681', Aliases='', description='uncharacterized LOC105373681', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149588278, end_position_on_the_genomic_accession=149596428, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724017, CurrentID=0, Status='live', Symbol='LOC102724017', Aliases='', description='uncharacterized LOC102724017', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481415, CurrentID=0, Status='live', Symbol='RNU6-745P', Aliases='', description='RNA, U6 small nuclear 745, pseudogene', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481416, CurrentID=0, Status='live', Symbol='RNU6-748P', Aliases='', description='RNA, U6 small nuclear 748, pseudogene', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481417, CurrentID=0, Status='live', Symbol='RNU6-752P', Aliases='', description='RNA, U6 small nuclear 752, pseudogene', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127912, CurrentID=0, Status='live', Symbol='LOC100127912', Aliases='', description='chromosome 2 open reading frame 27B pseudogene', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129133488, end_position_on_the_genomic_accession=129134243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874245, CurrentID=0, Status='live', Symbol='NAALADL2-AS1', Aliases='', description='NAALADL2 antisense RNA 1', other_designations='NAALADL2 antisense RNA 1 (non-protein coding)', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175773578, end_position_on_the_genomic_accession=175776333, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481424, CurrentID=0, Status='live', Symbol='RNU6-773P', Aliases='', description='RNA, U6 small nuclear 773, pseudogene', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481427, CurrentID=0, Status='live', Symbol='RNU6-780P', Aliases='', description='RNA, U6 small nuclear 780, pseudogene', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481428, CurrentID=0, Status='live', Symbol='RNU6-784P', Aliases='', description='RNA, U6 small nuclear 784, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420336, CurrentID=0, Status='live', Symbol='LOC100420336', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34276859, end_position_on_the_genomic_accession=34277364, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285147, CurrentID=0, Status='live', Symbol='LOC285147', Aliases='', description='uncharacterized LOC285147', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481430, CurrentID=0, Status='live', Symbol='RNU6-790P', Aliases='', description='RNA, U6 small nuclear 790, pseudogene', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374067, CurrentID=0, Status='live', Symbol='LOC105374067', Aliases='', description='uncharacterized LOC105374067', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120438120, end_position_on_the_genomic_accession=120442572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481432, CurrentID=0, Status='live', Symbol='RNU6-796P', Aliases='', description='RNA, U6 small nuclear 796, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  394261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394261, CurrentID=0, Status='live', Symbol='KLHL30-AS1', Aliases='C2orf19', description='KLHL30 antisense RNA 1', other_designations='KLHL30 antisense RNA 1 (non-protein coding)', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=238726269, end_position_on_the_genomic_accession=238729374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481434, CurrentID=0, Status='live', Symbol='RNU6-803P', Aliases='', description='RNA, U6 small nuclear 803, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481435, CurrentID=0, Status='live', Symbol='RNU6-806P', Aliases='', description='RNA, U6 small nuclear 806, pseudogene', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442271, CurrentID=0, Status='live', Symbol='TUBB4BP7', Aliases='', description='tubulin beta 4B class IVb pseudogene 7', other_designations='tubulin, beta 2C pseudogene|tubulin, beta 4B class IVb pseudogene 7', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152897916, end_position_on_the_genomic_accession=152899473, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100313769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313769, CurrentID=0, Status='live', Symbol='MIR320B2', Aliases='MIR320B-2, MIRN320B2, mir-320b-2', description='microRNA 320b-2', other_designations='hsa-mir-320b-2', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224257004, end_position_on_the_genomic_accession=224257141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313770, CurrentID=0, Status='live', Symbol='MIR548K', Aliases='MIRN548K', description='microRNA 548k', other_designations='hsa-mir-548k', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70283955, end_position_on_the_genomic_accession=70284070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313771, CurrentID=0, Status='live', Symbol='MIR548F2', Aliases='MIR548F-2, MIRN548F2, hsa-mir-548f-2', description='microRNA 548f-2', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212426263, end_position_on_the_genomic_accession=212426360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313772, CurrentID=0, Status='live', Symbol='MIR548M', Aliases='MIRN548M', description='microRNA 548m', other_designations='hsa-mir-548m', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95063141, end_position_on_the_genomic_accession=95063226, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313773, CurrentID=0, Status='live', Symbol='MIR548H2', Aliases='MIR548H-2, MIRN548H2', description='microRNA 548h-2', other_designations='hsa-mir-548h-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11306440, end_position_on_the_genomic_accession=11306527, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313774, CurrentID=0, Status='live', Symbol='MIR302E', Aliases='MIRN302E', description='microRNA 302e', other_designations='hsa-mir-302e', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7234766, end_position_on_the_genomic_accession=7234837, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313777, CurrentID=0, Status='live', Symbol='MIR670', Aliases='hsa-mir-670', description='microRNA 670', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43559656, end_position_on_the_genomic_accession=43559753, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313778, CurrentID=0, Status='live', Symbol='MIR759', Aliases='hsa-mir-759', description='microRNA 759', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52810050, end_position_on_the_genomic_accession=52810140, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313779, CurrentID=0, Status='live', Symbol='MIR2117', Aliases='', description='microRNA 2117', other_designations='hsa-mir-2117', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43444806, end_position_on_the_genomic_accession=43444885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313780, CurrentID=0, Status='live', Symbol='MIR2278', Aliases='mir-2278', description='microRNA 2278', other_designations='hsa-mir-2278', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94809962, end_position_on_the_genomic_accession=94810057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313781, CurrentID=0, Status='live', Symbol='MIR718', Aliases='hsa-mir-718', description='microRNA 718', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154019920, end_position_on_the_genomic_accession=154019989, orientation='minus', exon_count=1, OMIM=300929, no_hdr0=''),
  100313782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313782, CurrentID=0, Status='live', Symbol='PIRC7', Aliases='', description='piwi-interacting RNA cluster 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313783, CurrentID=0, Status='live', Symbol='PIRC10', Aliases='', description='piwi-interacting RNA cluster 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313784, CurrentID=0, Status='live', Symbol='PIRC14', Aliases='', description='piwi-interacting RNA cluster 14', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400500, CurrentID=0, Status='live', Symbol='BCAR4', Aliases='', description='breast cancer anti-estrogen resistance 4 (non-protein coding)', other_designations='breast cancer antiestrogen resistance 4 protein', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11819830, end_position_on_the_genomic_accession=11828832, orientation='minus', exon_count=5, OMIM=613746, no_hdr0=''),
  100313786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313786, CurrentID=0, Status='live', Symbol='PIRC23', Aliases='', description='piwi-interacting RNA cluster 23', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313787, CurrentID=0, Status='live', Symbol='PIRC31', Aliases='', description='piwi-interacting RNA cluster 31', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313788, CurrentID=0, Status='live', Symbol='PIRC38', Aliases='', description='piwi-interacting RNA cluster 38', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313789, CurrentID=0, Status='live', Symbol='PIRC42', Aliases='', description='piwi-interacting RNA cluster 42', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313790, CurrentID=0, Status='live', Symbol='PIRC48', Aliases='', description='piwi-interacting RNA cluster 48', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313791, CurrentID=0, Status='live', Symbol='PIRC52', Aliases='', description='piwi-interacting RNA cluster 52', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313792, CurrentID=0, Status='live', Symbol='PIRC55', Aliases='', description='piwi-interacting RNA cluster 55', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313793, CurrentID=0, Status='live', Symbol='PIRC63', Aliases='', description='piwi-interacting RNA cluster 63', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313794, CurrentID=0, Status='live', Symbol='PIRC65', Aliases='', description='piwi-interacting RNA cluster 65', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313795, CurrentID=0, Status='live', Symbol='PIRC73', Aliases='', description='piwi-interacting RNA cluster 73', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313796, CurrentID=0, Status='live', Symbol='PIRC76', Aliases='', description='piwi-interacting RNA cluster 76', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92573126, end_position_on_the_genomic_accession=92581675, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100313797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313797, CurrentID=0, Status='live', Symbol='PIRC79', Aliases='', description='piwi-interacting RNA cluster 79', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313798, CurrentID=0, Status='live', Symbol='PIRC90', Aliases='', description='piwi-interacting RNA cluster 90', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313799, CurrentID=0, Status='live', Symbol='PIRC94', Aliases='', description='piwi-interacting RNA cluster 94', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313800, CurrentID=0, Status='live', Symbol='PIRC96', Aliases='', description='piwi-interacting RNA cluster 96', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313801, CurrentID=0, Status='live', Symbol='PIRC104', Aliases='', description='piwi-interacting RNA cluster 104', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313802, CurrentID=0, Status='live', Symbol='PIRC107', Aliases='', description='piwi-interacting RNA cluster 107', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313803, CurrentID=0, Status='live', Symbol='PIRC112', Aliases='', description='piwi-interacting RNA cluster 112', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313806, CurrentID=0, Status='live', Symbol='MIR1255B1', Aliases='MIR1255B-1, MIRN1255B1', description='microRNA 1255b-1', other_designations='hsa-mir-1255b-1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36426366, end_position_on_the_genomic_accession=36426428, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313807, CurrentID=0, Status='live', Symbol='PIRC4', Aliases='', description='piwi-interacting RNA cluster 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313808, CurrentID=0, Status='live', Symbol='PIRC12', Aliases='', description='piwi-interacting RNA cluster 12', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313809, CurrentID=0, Status='live', Symbol='PIRC17', Aliases='', description='piwi-interacting RNA cluster 17', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313810, CurrentID=0, Status='live', Symbol='PIRC24', Aliases='', description='piwi-interacting RNA cluster 24', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313811, CurrentID=0, Status='live', Symbol='PIRC32', Aliases='', description='piwi-interacting RNA cluster 32', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313812, CurrentID=0, Status='live', Symbol='PIRC37', Aliases='', description='piwi-interacting RNA cluster 37', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313813, CurrentID=0, Status='live', Symbol='PIRC45', Aliases='', description='piwi-interacting RNA cluster 45', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313814, CurrentID=0, Status='live', Symbol='PIRC50', Aliases='', description='piwi-interacting RNA cluster 50', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313815, CurrentID=0, Status='live', Symbol='PIRC54', Aliases='', description='piwi-interacting RNA cluster 54', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  394263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=394263, CurrentID=0, Status='live', Symbol='MUC21', Aliases='C6orf205, KMQK697, MUC-21', description='mucin 21, cell surface associated', other_designations='epiglycanin', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30983708, end_position_on_the_genomic_accession=30988637, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100313817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313817, CurrentID=0, Status='live', Symbol='PIRC67', Aliases='', description='piwi-interacting RNA cluster 67', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313818, CurrentID=0, Status='live', Symbol='PIRC71', Aliases='', description='piwi-interacting RNA cluster 71', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313819, CurrentID=0, Status='live', Symbol='PIRC80', Aliases='', description='piwi-interacting RNA cluster 80', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313820, CurrentID=0, Status='live', Symbol='PIRC84', Aliases='', description='piwi-interacting RNA cluster 84', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400506, CurrentID=0, Status='live', Symbol='KNOP1', Aliases='101F10.1, C16orf88, FAM191A, TSG118', description='lysine-rich nucleolar protein 1', other_designations='family with sequence similarity 191, member A|protein C16orf88|testis-specific gene 118 protein', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19706347, end_position_on_the_genomic_accession=19718361, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100313822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313822, CurrentID=0, Status='live', Symbol='MIR513B', Aliases='MIRN513B', description='microRNA 513b', other_designations='hsa-mir-513b', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147199044, end_position_on_the_genomic_accession=147199127, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313824, CurrentID=0, Status='live', Symbol='MIR663B', Aliases='MIRN663B, mir-663b', description='microRNA 663b', other_designations='hsa-mir-663b', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132256966, end_position_on_the_genomic_accession=132257080, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313829, CurrentID=0, Status='live', Symbol='MIR548O', Aliases='MIRN548O', description='microRNA 548o', other_designations='hsa-mir-548o', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102405742, end_position_on_the_genomic_accession=102405855, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313830, CurrentID=0, Status='live', Symbol='MIR548H1', Aliases='MIR548H-1, MIRN548H1', description='microRNA 548h-1', other_designations='hsa-mir-548h-1', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64095024, end_position_on_the_genomic_accession=64095125, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  400508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400508, CurrentID=0, Status='live', Symbol='CRYM-AS1', Aliases='NCRNA00169', description='CRYM antisense RNA 1', other_designations='CRYM antisense RNA 1 (non-protein coding)', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21300849, end_position_on_the_genomic_accession=21318591, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100313835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313835, CurrentID=0, Status='live', Symbol='MIR1255B2', Aliases='MIR1255B-2, MIRN1255B2, hsa-mir-1255b-2', description='microRNA 1255b-2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167998660, end_position_on_the_genomic_accession=167998726, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313837, CurrentID=0, Status='live', Symbol='MIR762', Aliases='hsa-mir-762', description='microRNA 762', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30893903, end_position_on_the_genomic_accession=30893985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313838, CurrentID=0, Status='live', Symbol='MIR764', Aliases='hsa-mir-764', description='microRNA 764', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114639435, end_position_on_the_genomic_accession=114639519, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313839, CurrentID=0, Status='live', Symbol='MIR2114', Aliases='', description='microRNA 2114', other_designations='hsa-mir-2114', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150228004, end_position_on_the_genomic_accession=150228083, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313840, CurrentID=0, Status='live', Symbol='MIR2115', Aliases='', description='microRNA 2115', other_designations='hsa-mir-2115', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48316360, end_position_on_the_genomic_accession=48316459, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313841, CurrentID=0, Status='live', Symbol='MIR548Q', Aliases='', description='microRNA 548q', other_designations='hsa-mir-548q', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12725254, end_position_on_the_genomic_accession=12725353, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313842, CurrentID=0, Status='live', Symbol='MIR2276', Aliases='', description='microRNA 2276', other_designations='hsa-mir-2276', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24162416, end_position_on_the_genomic_accession=24162504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313843, CurrentID=0, Status='live', Symbol='MIR711', Aliases='hsa-mir-711', description='microRNA 711', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48578902, end_position_on_the_genomic_accession=48578977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313844, CurrentID=0, Status='live', Symbol='PIRC1', Aliases='', description='piwi-interacting RNA cluster 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313845, CurrentID=0, Status='live', Symbol='PIRC5', Aliases='', description='piwi-interacting RNA cluster 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313846, CurrentID=0, Status='live', Symbol='PIRC6', Aliases='', description='piwi-interacting RNA cluster 6', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313847, CurrentID=0, Status='live', Symbol='PIRC9', Aliases='', description='piwi-interacting RNA cluster 9', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313848, CurrentID=0, Status='live', Symbol='PIRC15', Aliases='', description='piwi-interacting RNA cluster 15', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313849, CurrentID=0, Status='live', Symbol='PIRC16', Aliases='', description='piwi-interacting RNA cluster 16', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313850, CurrentID=0, Status='live', Symbol='PIRC22', Aliases='', description='piwi-interacting RNA cluster 22', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400511, CurrentID=0, Status='live', Symbol='LINC01567', Aliases='', description='long intergenic non-protein coding RNA 1567', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24661422, end_position_on_the_genomic_accession=24671062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100313852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313852, CurrentID=0, Status='live', Symbol='PIRC26', Aliases='', description='piwi-interacting RNA cluster 26', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313853, CurrentID=0, Status='live', Symbol='PIRC30', Aliases='', description='piwi-interacting RNA cluster 30', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313854, CurrentID=0, Status='live', Symbol='PIRC33', Aliases='', description='piwi-interacting RNA cluster 33', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313855, CurrentID=0, Status='live', Symbol='PIRC35', Aliases='', description='piwi-interacting RNA cluster 35', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313856, CurrentID=0, Status='live', Symbol='PIRC39', Aliases='', description='piwi-interacting RNA cluster 39', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313857, CurrentID=0, Status='live', Symbol='PIRC43', Aliases='', description='piwi-interacting RNA cluster 43', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313858, CurrentID=0, Status='live', Symbol='PIRC46', Aliases='', description='piwi-interacting RNA cluster 46', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313859, CurrentID=0, Status='live', Symbol='PIRC47', Aliases='', description='piwi-interacting RNA cluster 47', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313860, CurrentID=0, Status='live', Symbol='PIRC51', Aliases='', description='piwi-interacting RNA cluster 51', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313861, CurrentID=0, Status='live', Symbol='PIRC53', Aliases='', description='piwi-interacting RNA cluster 53', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313862, CurrentID=0, Status='live', Symbol='PIRC56', Aliases='', description='piwi-interacting RNA cluster 56', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313863, CurrentID=0, Status='live', Symbol='PIRC60', Aliases='', description='piwi-interacting RNA cluster 60', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313864, CurrentID=0, Status='live', Symbol='PIRC64', Aliases='', description='piwi-interacting RNA cluster 64', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313865, CurrentID=0, Status='live', Symbol='PIRC66', Aliases='', description='piwi-interacting RNA cluster 66', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51037584, end_position_on_the_genomic_accession=51322397, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100313866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313866, CurrentID=0, Status='live', Symbol='PIRC70', Aliases='', description='piwi-interacting RNA cluster 70', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313867, CurrentID=0, Status='live', Symbol='PIRC72', Aliases='', description='piwi-interacting RNA cluster 72', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313868, CurrentID=0, Status='live', Symbol='PIRC77', Aliases='', description='piwi-interacting RNA cluster 77', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313869, CurrentID=0, Status='live', Symbol='PIRC78', Aliases='', description='piwi-interacting RNA cluster 78', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313870, CurrentID=0, Status='live', Symbol='PIRC81', Aliases='', description='piwi-interacting RNA cluster 81', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313871, CurrentID=0, Status='live', Symbol='PIRC85', Aliases='', description='piwi-interacting RNA cluster 85', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313872, CurrentID=0, Status='live', Symbol='PIRC87', Aliases='', description='piwi-interacting RNA cluster 87', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313873, CurrentID=0, Status='live', Symbol='PIRC89', Aliases='', description='piwi-interacting RNA cluster 89', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313874, CurrentID=0, Status='live', Symbol='PIRC98', Aliases='', description='piwi-interacting RNA cluster 98', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313875, CurrentID=0, Status='live', Symbol='PIRC99', Aliases='', description='piwi-interacting RNA cluster 99', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313876, CurrentID=0, Status='live', Symbol='PIRC102', Aliases='', description='piwi-interacting RNA cluster 102', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313877, CurrentID=0, Status='live', Symbol='PIRC106', Aliases='', description='piwi-interacting RNA cluster 106', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313878, CurrentID=0, Status='live', Symbol='PIRC108', Aliases='', description='piwi-interacting RNA cluster 108', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313879, CurrentID=0, Status='live', Symbol='PIRC110', Aliases='', description='piwi-interacting RNA cluster 110', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313880, CurrentID=0, Status='live', Symbol='PIRC113', Aliases='', description='piwi-interacting RNA cluster 113', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313881, CurrentID=0, Status='live', Symbol='PIRC114', Aliases='', description='piwi-interacting RNA cluster 114', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313884, CurrentID=0, Status='live', Symbol='MIR548H4', Aliases='MIR548H-4, MIRN548H4, hsa-mir-548h-4', description='microRNA 548h-4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27048853, end_position_on_the_genomic_accession=27048963, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313886, CurrentID=0, Status='live', Symbol='MIR2116', Aliases='mir-2116', description='microRNA 2116', other_designations='hsa-mir-2116', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59171183, end_position_on_the_genomic_accession=59171262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313887, CurrentID=0, Status='live', Symbol='MIR2277', Aliases='mir-2277', description='microRNA 2277', other_designations='hsa-mir-2277', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93620696, end_position_on_the_genomic_accession=93620788, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313888, CurrentID=0, Status='live', Symbol='PIRC101', Aliases='', description='piwi-interacting RNA cluster 101', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313890, CurrentID=0, Status='live', Symbol='PIRC34', Aliases='', description='piwi-interacting RNA cluster 34', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313891, CurrentID=0, Status='live', Symbol='PIRC83', Aliases='', description='piwi-interacting RNA cluster 83', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313892, CurrentID=0, Status='live', Symbol='MIR761', Aliases='hsa-mir-761', description='microRNA 761', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51836344, end_position_on_the_genomic_accession=51836402, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313893, CurrentID=0, Status='live', Symbol='PIRC28', Aliases='', description='piwi-interacting RNA cluster 28', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313894, CurrentID=0, Status='live', Symbol='PIRC88', Aliases='', description='piwi-interacting RNA cluster 88', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313895, CurrentID=0, Status='live', Symbol='MIR548F4', Aliases='MIR548F-4, MIRN548F4, hsa-mir-548f-4', description='microRNA 548f-4', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147378017, end_position_on_the_genomic_accession=147378121, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313896, CurrentID=0, Status='live', Symbol='MIR320D1', Aliases='MIR320D-1, MIRN320D1, mir-320d-1', description='microRNA 320d-1', other_designations='hsa-mir-320d-1', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40727828, end_position_on_the_genomic_accession=40727875, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313897, CurrentID=0, Status='live', Symbol='PIRC11', Aliases='', description='piwi-interacting RNA cluster 11', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313898, CurrentID=0, Status='live', Symbol='PIRC40', Aliases='', description='piwi-interacting RNA cluster 40', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313899, CurrentID=0, Status='live', Symbol='PIRC61', Aliases='', description='piwi-interacting RNA cluster 61', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313900, CurrentID=0, Status='live', Symbol='PIRC91', Aliases='', description='piwi-interacting RNA cluster 91', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313901, CurrentID=0, Status='live', Symbol='PIRC103', Aliases='', description='piwi-interacting RNA cluster 103', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378437, CurrentID=0, Status='live', Symbol='LOC105378437', Aliases='', description='uncharacterized LOC105378437', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93762880, end_position_on_the_genomic_accession=93788911, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100313905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313905, CurrentID=0, Status='live', Symbol='PIRC3', Aliases='', description='piwi-interacting RNA cluster 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313906, CurrentID=0, Status='live', Symbol='PIRC18', Aliases='', description='piwi-interacting RNA cluster 18', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313907, CurrentID=0, Status='live', Symbol='PIRC27', Aliases='', description='piwi-interacting RNA cluster 27', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313908, CurrentID=0, Status='live', Symbol='PIRC44', Aliases='', description='piwi-interacting RNA cluster 44', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313909, CurrentID=0, Status='live', Symbol='PIRC59', Aliases='', description='piwi-interacting RNA cluster 59', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313910, CurrentID=0, Status='live', Symbol='PIRC69', Aliases='', description='piwi-interacting RNA cluster 69', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313911, CurrentID=0, Status='live', Symbol='PIRC86', Aliases='', description='piwi-interacting RNA cluster 86', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313912, CurrentID=0, Status='live', Symbol='PIRC100', Aliases='', description='piwi-interacting RNA cluster 100', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313913, CurrentID=0, Status='live', Symbol='PIRC109', Aliases='', description='piwi-interacting RNA cluster 109', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313914, CurrentID=0, Status='live', Symbol='MIR548J', Aliases='MIRN548J, hsa-mir-548j', description='microRNA 548j', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26555212, end_position_on_the_genomic_accession=26555323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313915, CurrentID=0, Status='live', Symbol='PIRC8', Aliases='', description='piwi-interacting RNA cluster 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313916, CurrentID=0, Status='live', Symbol='PIRC21', Aliases='', description='piwi-interacting RNA cluster 21', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313917, CurrentID=0, Status='live', Symbol='PIRC41', Aliases='', description='piwi-interacting RNA cluster 41', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313918, CurrentID=0, Status='live', Symbol='PIRC58', Aliases='', description='piwi-interacting RNA cluster 58', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313919, CurrentID=0, Status='live', Symbol='PIRC75', Aliases='', description='piwi-interacting RNA cluster 75', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313920, CurrentID=0, Status='live', Symbol='PIRC92', Aliases='', description='piwi-interacting RNA cluster 92', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313921, CurrentID=0, Status='live', Symbol='MIR548E', Aliases='MIRN548E, hsa-mir-548e', description='microRNA 548e', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110988926, end_position_on_the_genomic_accession=110989013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100313923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313923, CurrentID=0, Status='live', Symbol='MIR449C', Aliases='mir-449c', description='microRNA 449c', other_designations='hsa-mir-449c', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55172262, end_position_on_the_genomic_accession=55172353, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313924, CurrentID=0, Status='live', Symbol='PIRC2', Aliases='', description='piwi-interacting RNA cluster 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313925, CurrentID=0, Status='live', Symbol='PIRC13', Aliases='', description='piwi-interacting RNA cluster 13', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313926, CurrentID=0, Status='live', Symbol='PIRC19', Aliases='', description='piwi-interacting RNA cluster 19', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  43847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=43847, CurrentID=0, Status='live', Symbol='KLK14', Aliases='KLK-L6', description='kallikrein related peptidase 14', other_designations='kallikrein-like protein 6|kallikrein-related peptidase 14', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51077879, end_position_on_the_genomic_accession=51086458, orientation='minus', exon_count=9, OMIM=606135, no_hdr0=''),
  100313928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313928, CurrentID=0, Status='live', Symbol='PIRC36', Aliases='', description='piwi-interacting RNA cluster 36', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  43849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=43849, CurrentID=0, Status='live', Symbol='KLK12', Aliases='KLK-L5, KLKL5', description='kallikrein related peptidase 12', other_designations='kallikrein 12|kallikrein-like protein 5|kallikrein-related peptidase 12', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51028960, end_position_on_the_genomic_accession=51035652, orientation='minus', exon_count=6, OMIM=605539, no_hdr0=''),
  100313930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313930, CurrentID=0, Status='live', Symbol='PIRC57', Aliases='', description='piwi-interacting RNA cluster 57', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313931, CurrentID=0, Status='live', Symbol='PIRC68', Aliases='', description='piwi-interacting RNA cluster 68', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313932, CurrentID=0, Status='live', Symbol='PIRC74', Aliases='', description='piwi-interacting RNA cluster 74', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313933, CurrentID=0, Status='live', Symbol='PIRC82', Aliases='', description='piwi-interacting RNA cluster 82', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313934, CurrentID=0, Status='live', Symbol='PIRC93', Aliases='', description='piwi-interacting RNA cluster 93', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313935, CurrentID=0, Status='live', Symbol='PIRC95', Aliases='', description='piwi-interacting RNA cluster 95', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313936, CurrentID=0, Status='live', Symbol='PIRC105', Aliases='', description='piwi-interacting RNA cluster 105', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313937, CurrentID=0, Status='live', Symbol='PIRC111', Aliases='', description='piwi-interacting RNA cluster 111', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313938, CurrentID=0, Status='live', Symbol='MIR548G', Aliases='MIRN548G, hsa-mir-548g', description='microRNA 548g', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147344629, end_position_on_the_genomic_accession=147344717, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100313942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313942, CurrentID=0, Status='live', Symbol='LOC100313942', Aliases='', description='transmembrane protein 92 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39199871, end_position_on_the_genomic_accession=39212739, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481465, CurrentID=0, Status='live', Symbol='RNU6-906P', Aliases='', description='RNA, U6 small nuclear 906, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106736472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106736472, CurrentID=0, Status='live', Symbol='LOC106736472', Aliases='', description='antisense transcript from human inducible nitric oxide synthase (iNOS) gene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106736475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106736475, CurrentID=0, Status='live', Symbol='PCF11-AS1', Aliases='', description='PCF11 anstisense RNA 1', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313948, CurrentID=0, Status='live', Symbol='BRV2', Aliases='', description='Vertigo, benign recurrent, 2', other_designations='', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613106, no_hdr0=''),
  106481466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481466, CurrentID=0, Status='live', Symbol='RNU6-909P', Aliases='', description='RNA, U6 small nuclear 909, pseudogene', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481467, CurrentID=0, Status='live', Symbol='RNU6-913P', Aliases='', description='RNA, U6 small nuclear 913, pseudogene', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481468, CurrentID=0, Status='live', Symbol='RNU6-916P', Aliases='', description='RNA, U6 small nuclear 916, pseudogene', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313962, CurrentID=0, Status='live', Symbol='RFKP1', Aliases='', description='riboflavin kinase pseudogene 1', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240822242, end_position_on_the_genomic_accession=240823670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481469, CurrentID=0, Status='live', Symbol='RNU6-918P', Aliases='', description='RNA, U6 small nuclear 918, pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313968, CurrentID=0, Status='live', Symbol='LOC100313968', Aliases='', description='riboflavin kinase pseudogene', other_designations='', map_location='22q', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40787707, end_position_on_the_genomic_accession=40789437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481470, CurrentID=0, Status='live', Symbol='RNU6-920P', Aliases='', description='RNA, U6 small nuclear 920, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374075, CurrentID=0, Status='live', Symbol='LOC105374075', Aliases='', description='uncharacterized LOC105374075', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123762938, end_position_on_the_genomic_accession=123765174, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400533, CurrentID=0, Status='live', Symbol='FLJ26245', Aliases='', description='uncharacterized LOC400533', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35746552, end_position_on_the_genomic_accession=35756624, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481472, CurrentID=0, Status='live', Symbol='RNU6-926P', Aliases='', description='RNA, U6 small nuclear 926, pseudogene', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481473, CurrentID=0, Status='live', Symbol='RNU6-929P', Aliases='', description='RNA, U6 small nuclear 929, pseudogene', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400535, CurrentID=0, Status='live', Symbol='FLJ44674', Aliases='', description='FLJ44674 protein', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=47940980, end_position_on_the_genomic_accession=47941966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481474, CurrentID=0, Status='live', Symbol='RNU6-932P', Aliases='', description='RNA, U6 small nuclear 932, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481475, CurrentID=0, Status='live', Symbol='RNU6-934P', Aliases='', description='RNA, U6 small nuclear 934, pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481476, CurrentID=0, Status='live', Symbol='RNU6-939P', Aliases='', description='RNA, U6 small nuclear 939, pseudogene', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  203930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203930, CurrentID=0, Status='live', Symbol='LOC203930', Aliases='', description='sorting nexin 7 pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107735961, end_position_on_the_genomic_accession=107738761, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481477, CurrentID=0, Status='live', Symbol='RNU6-942P', Aliases='', description='RNA, U6 small nuclear 942, pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481478, CurrentID=0, Status='live', Symbol='RNU6-946P', Aliases='', description='RNA, U6 small nuclear 946, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724081, CurrentID=0, Status='live', Symbol='LOC102724081', Aliases='', description='uncharacterized LOC102724081', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168422623, end_position_on_the_genomic_accession=168460221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106481479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481479, CurrentID=0, Status='live', Symbol='RNU6-948P', Aliases='', description='RNA, U6 small nuclear 948, pseudogene', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400541, CurrentID=0, Status='live', Symbol='LOC400541', Aliases='', description='uncharacterized LOC400541', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70156340, end_position_on_the_genomic_accession=70173448, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481480, CurrentID=0, Status='live', Symbol='RNU6-951P', Aliases='', description='RNA, U6 small nuclear 951, pseudogene', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481482, CurrentID=0, Status='live', Symbol='RNU6-958P', Aliases='', description='RNA, U6 small nuclear 958, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481483, CurrentID=0, Status='live', Symbol='RNU6-961P', Aliases='', description='RNA, U6 small nuclear 961, pseudogene', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420347, CurrentID=0, Status='live', Symbol='LOC100420347', Aliases='', description='polycomb group ring finger 3 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58100985, end_position_on_the_genomic_accession=58101572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481484, CurrentID=0, Status='live', Symbol='RNU6-965P', Aliases='', description='RNA, U6 small nuclear 965, pseudogene', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481485, CurrentID=0, Status='live', Symbol='RNU6-968P', Aliases='', description='RNA, U6 small nuclear 968, pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481486, CurrentID=0, Status='live', Symbol='RNU6-970P', Aliases='', description='RNA, U6 small nuclear 970, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189028, CurrentID=0, Status='live', Symbol='TRI-TAT2-2', Aliases='TRNAI2', description='transfer RNA-Ile (TAT) 2-2', other_designations='transfer RNA isoleucine 2 (anticodon UAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27020346, end_position_on_the_genomic_accession=27020439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481487, CurrentID=0, Status='live', Symbol='RNU6-974P', Aliases='', description='RNA, U6 small nuclear 974, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481488, CurrentID=0, Status='live', Symbol='RNU6-976P', Aliases='', description='RNA, U6 small nuclear 976, pseudogene', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874258, CurrentID=0, Status='live', Symbol='SCGB1D5P', Aliases='SCGB1D1P1', description='secretoglobin, family 1D, member 5, pseudogene', other_designations='secretoglobin, family 1D, member 1 pseudogene 1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165517276, end_position_on_the_genomic_accession=165517500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  7334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7334, CurrentID=0, Status='live', Symbol='UBE2N', Aliases='HEL-S-71, UBC13, UBCHBEN; UBC13, UbcH-ben, UbcH13', description='ubiquitin conjugating enzyme E2N', other_designations='E2 ubiquitin-conjugating enzyme N|bendless-like ubiquitin conjugating enzyme|bendless-like ubiquitin-conjugating enzyme|epididymis secretory protein Li 71|ubiquitin carrier protein N|ubiquitin-conjugating enzyme E2N (UBC13 homolog, yeast)|ubiquitin-conjugating enzyme E2N (homologous to yeast UBC13)|ubiquitin-protein ligase N|yeast UBC13 homolog', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93408312, end_position_on_the_genomic_accession=93443445, orientation='minus', exon_count=7, OMIM=603679, no_hdr0=''),
  106481489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481489, CurrentID=0, Status='live', Symbol='RNU6-980P', Aliases='', description='RNA, U6 small nuclear 980, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481490, CurrentID=0, Status='live', Symbol='RNU6-983P', Aliases='', description='RNA, U6 small nuclear 983, pseudogene', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481491, CurrentID=0, Status='live', Symbol='RNU6-986P', Aliases='', description='RNA, U6 small nuclear 986, pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400553, CurrentID=0, Status='live', Symbol='LOC400553', Aliases='', description='uncharacterized LOC400553', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88099409, end_position_on_the_genomic_accession=88100985, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481492, CurrentID=0, Status='live', Symbol='RNU6-989P', Aliases='', description='RNA, U6 small nuclear 989, pseudogene', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481493, CurrentID=0, Status='live', Symbol='RNU6-992P', Aliases='', description='RNA, U6 small nuclear 992, pseudogene', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874259, CurrentID=0, Status='live', Symbol='SLC25A30-AS1', Aliases='', description='SLC25A30 antisense RNA 1', other_designations='SLC25A30 antisense RNA 1 (non-protein coding)|XXyac-YMR12DG2.1', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45418162, end_position_on_the_genomic_accession=45420371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100533625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533625, CurrentID=0, Status='live', Symbol='LOC100533625', Aliases='', description='sorting nexin 18 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19750617, end_position_on_the_genomic_accession=19752539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400555, CurrentID=0, Status='live', Symbol='ZC3H18-AS1', Aliases='C16orf85, LINC00859', description='ZC3H18 antisense RNA 1 (head to head)', other_designations='long intergenic non-protein coding RNA 859', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88554002, end_position_on_the_genomic_accession=88567106, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481494, CurrentID=0, Status='live', Symbol='RNU6-995P', Aliases='', description='RNA, U6 small nuclear 995, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373967, CurrentID=0, Status='live', Symbol='LOC105373967', Aliases='', description='uncharacterized LOC105373967', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239728453, end_position_on_the_genomic_accession=239731002, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481495, CurrentID=0, Status='live', Symbol='RNU6-998P', Aliases='', description='RNA, U6 small nuclear 998, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481496, CurrentID=0, Status='live', Symbol='RNU6-1000P', Aliases='', description='RNA, U6 small nuclear 1000, pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374080, CurrentID=0, Status='live', Symbol='LOC105374080', Aliases='', description='uncharacterized LOC105374080', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125376862, end_position_on_the_genomic_accession=125386842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400558, CurrentID=0, Status='live', Symbol='LOC400558', Aliases='', description='uncharacterized LOC400558', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89166383, end_position_on_the_genomic_accession=89169147, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874260, CurrentID=0, Status='live', Symbol='UBE2CP5', Aliases='', description='ubiquitin conjugating enzyme E2C pseudogene 5', other_designations='ubiquitin-conjugating enzyme E2C pseudogene 5', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57867970, end_position_on_the_genomic_accession=57868463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481499, CurrentID=0, Status='live', Symbol='RNU6-1009P', Aliases='', description='RNA, U6 small nuclear 1009, pseudogene', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7350, CurrentID=0, Status='live', Symbol='UCP1', Aliases='SLC25A7, UCP', description='uncoupling protein 1 (mitochondrial, proton carrier)', other_designations='solute carrier family 25 member 7|thermogenin', map_location='4q28-q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140559394, end_position_on_the_genomic_accession=140568902, orientation='minus', exon_count=7, OMIM=113730, no_hdr0=''),
  106481506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481506, CurrentID=0, Status='live', Symbol='RNU6-1032P', Aliases='', description='RNA, U6 small nuclear 1032, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728248, CurrentID=0, Status='live', Symbol='ABCB10P3', Aliases='', description='ATP binding cassette subfamily B member 10 pseudogene 3', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene 3', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28402865, end_position_on_the_genomic_accession=28405031, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481507, CurrentID=0, Status='live', Symbol='RNU6-1035P', Aliases='', description='RNA, U6 small nuclear 1035, pseudogene', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400569, CurrentID=0, Status='live', Symbol='MED11', Aliases='HSPC296', description='mediator complex subunit 11', other_designations='mediator of RNA polymerase II transcription, subunit 11 homolog', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4731408, end_position_on_the_genomic_accession=4733610, orientation='plus', exon_count=3, OMIM=612383, no_hdr0=''),
  100874262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874262, CurrentID=0, Status='live', Symbol='UBE2D3P3', Aliases='', description='ubiquitin conjugating enzyme E2D 3 pseudogene 3', other_designations='ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast) pseudogene 3|ubiquitin-conjugating enzyme E2D 3 pseudogene 3', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150800473, end_position_on_the_genomic_accession=150801025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481510, CurrentID=0, Status='live', Symbol='RNU6-1043P', Aliases='', description='RNA, U6 small nuclear 1043, pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138428, CurrentID=0, Status='live', Symbol='PTRH1', Aliases='C9orf115, PTH1', description='peptidyl-tRNA hydrolase 1 homolog', other_designations='PTH', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127713948, end_position_on_the_genomic_accession=127716714, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  138429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138429, CurrentID=0, Status='live', Symbol='PIP5KL1', Aliases='PIPKH, bA203J24.5', description='phosphatidylinositol-4-phosphate 5-kinase-like 1', other_designations='PI(4)P 5-kinase-like protein 1|phosphatidylinositol phosphate kinase homolog|phosphatidylinositol-4-phosphate 5-kinase-like protein 1|ptdIns(4)P-5-kinase-like protein 1', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127921529, end_position_on_the_genomic_accession=127930797, orientation='minus', exon_count=10, OMIM=612865, no_hdr0=''),
  106481514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481514, CurrentID=0, Status='live', Symbol='RNU6-1056P', Aliases='', description='RNA, U6 small nuclear 1056, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400576, CurrentID=0, Status='live', Symbol='LOC400576', Aliases='', description='uncharacterized LOC400576', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=14670900, end_position_on_the_genomic_accession=14683514, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481516, CurrentID=0, Status='live', Symbol='RNU6-1067P', Aliases='', description='RNA, U6 small nuclear 1067, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724119, CurrentID=0, Status='live', Symbol='LOC102724119', Aliases='', description='uncharacterized LOC102724119', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13921, end_position_on_the_genomic_accession=21166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481517, CurrentID=0, Status='live', Symbol='RNU6-1071P', Aliases='', description='RNA, U6 small nuclear 1071, pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400579, CurrentID=0, Status='live', Symbol='FLJ35934', Aliases='', description='FLJ35934', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18411159, end_position_on_the_genomic_accession=18414388, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481518, CurrentID=0, Status='live', Symbol='RNU6-1073P', Aliases='', description='RNA, U6 small nuclear 1073, pseudogene', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420354, CurrentID=0, Status='live', Symbol='ECM1P1', Aliases='', description='extracellular matrix protein 1 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4768630, end_position_on_the_genomic_accession=4769895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481519, CurrentID=0, Status='live', Symbol='RNU6-1078P', Aliases='', description='RNA, U6 small nuclear 1078, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400581, CurrentID=0, Status='live', Symbol='GRAPL', Aliases='', description='GRB2-related adaptor protein-like', other_designations='GRB2-related adapter protein LOC400581', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19127469, end_position_on_the_genomic_accession=19158835, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481520, CurrentID=0, Status='live', Symbol='RNU6-1080P', Aliases='', description='RNA, U6 small nuclear 1080, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481521, CurrentID=0, Status='live', Symbol='RNU6-1083P', Aliases='', description='RNA, U6 small nuclear 1083, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374085, CurrentID=0, Status='live', Symbol='LOC105374085', Aliases='', description='uncharacterized LOC105374085', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126084227, end_position_on_the_genomic_accession=126086380, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724124, CurrentID=0, Status='live', Symbol='LOC102724124', Aliases='', description='uncharacterized LOC102724124', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16153388, end_position_on_the_genomic_accession=16167455, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481522, CurrentID=0, Status='live', Symbol='RNU6-1087P', Aliases='', description='RNA, U6 small nuclear 1087, pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481523, CurrentID=0, Status='live', Symbol='RNU6-1089P', Aliases='', description='RNA, U6 small nuclear 1089, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481524, CurrentID=0, Status='live', Symbol='RNU6-1093P', Aliases='', description='RNA, U6 small nuclear 1093, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  160287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160287, CurrentID=0, Status='live', Symbol='LDHAL6A', Aliases='LDH6A', description='lactate dehydrogenase A-like 6A', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18455827, end_position_on_the_genomic_accession=18479600, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106481525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481525, CurrentID=0, Status='live', Symbol='RNU6-1095P', Aliases='', description='RNA, U6 small nuclear 1095, pseudogene', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481526, CurrentID=0, Status='live', Symbol='RNU6-1099P', Aliases='', description='RNA, U6 small nuclear 1099, pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374086, CurrentID=0, Status='live', Symbol='LOC105374086', Aliases='', description='skin secretory protein xP2-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126213254, end_position_on_the_genomic_accession=126247330, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481527, CurrentID=0, Status='live', Symbol='RNU6-1102P', Aliases='', description='RNA, U6 small nuclear 1102, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400590, CurrentID=0, Status='live', Symbol='LOC400590', Aliases='', description='uncharacterized LOC400590', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481529, CurrentID=0, Status='live', Symbol='RNU6-1107P', Aliases='', description='RNA, U6 small nuclear 1107, pseudogene', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7375, CurrentID=0, Status='live', Symbol='USP4', Aliases='UNP, Unph', description='ubiquitin specific peptidase 4 (proto-oncogene)', other_designations='deubiquitinating enzyme 4|ubiquitin carboxyl-terminal esterase 4|ubiquitin specific protease 4 (proto-oncogene)|ubiquitin thioesterase 4|ubiquitin thiolesterase 4|ubiquitin-specific processing protease 4|ubiquitous nuclear protein homolog', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49277144, end_position_on_the_genomic_accession=49340103, orientation='minus', exon_count=23, OMIM=603486, no_hdr0=''),
  106481530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481530, CurrentID=0, Status='live', Symbol='RNU6-1111P', Aliases='', description='RNA, U6 small nuclear 1111, pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481531, CurrentID=0, Status='live', Symbol='RNU6-1114P', Aliases='', description='RNA, U6 small nuclear 1114, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374087, CurrentID=0, Status='live', Symbol='LOC105374087', Aliases='', description='uncharacterized LOC105374087', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126288121, end_position_on_the_genomic_accession=126301397, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481532, CurrentID=0, Status='live', Symbol='RNU6-1117P', Aliases='', description='RNA, U6 small nuclear 1117, pseudogene', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369604, CurrentID=0, Status='live', Symbol='LOC105369604', Aliases='', description='uncharacterized LOC105369604', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2032188, end_position_on_the_genomic_accession=2035065, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481533, CurrentID=0, Status='live', Symbol='RNU6-1119P', Aliases='', description='RNA, U6 small nuclear 1119, pseudogene', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481534, CurrentID=0, Status='live', Symbol='RNU6-1124P', Aliases='', description='RNA, U6 small nuclear 1124, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130468, CurrentID=0, Status='live', Symbol='LOC100130468', Aliases='', description='uncharacterized LOC100130468', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=10468965, end_position_on_the_genomic_accession=10469456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481535, CurrentID=0, Status='live', Symbol='RNU6-1126P', Aliases='', description='RNA, U6 small nuclear 1126, pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481536, CurrentID=0, Status='live', Symbol='RNU6-1129P', Aliases='', description='RNA, U6 small nuclear 1129, pseudogene', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374088, CurrentID=0, Status='live', Symbol='LOC105374088', Aliases='', description='uncharacterized LOC105374088', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126312420, end_position_on_the_genomic_accession=126323387, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481537, CurrentID=0, Status='live', Symbol='RNU6-1133P', Aliases='', description='RNA, U6 small nuclear 1133, pseudogene', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481538, CurrentID=0, Status='live', Symbol='RNU6-1135P', Aliases='', description='RNA, U6 small nuclear 1135, pseudogene', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420358, CurrentID=0, Status='live', Symbol='BUD13P1', Aliases='', description='BUD13 homolog pseudogene 1', other_designations='BUD13 homolog (S. cerevisiae) pseudogene 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95143986, end_position_on_the_genomic_accession=95145097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481539, CurrentID=0, Status='live', Symbol='RNU6-1139P', Aliases='', description='RNA, U6 small nuclear 1139, pseudogene', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481540, CurrentID=0, Status='live', Symbol='RNU6-1141P', Aliases='', description='RNA, U6 small nuclear 1141, pseudogene', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481541, CurrentID=0, Status='live', Symbol='RNU6-1145P', Aliases='', description='RNA, U6 small nuclear 1145, pseudogene', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481542, CurrentID=0, Status='live', Symbol='RNU6-1149P', Aliases='', description='RNA, U6 small nuclear 1149, pseudogene', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400604, CurrentID=0, Status='live', Symbol='TOB1-AS1', Aliases='', description='TOB1 antisense RNA 1', other_designations='TOB1 antisense RNA 1 (non-protein coding)', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50866679, end_position_on_the_genomic_accession=50868371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481543, CurrentID=0, Status='live', Symbol='RNU6-1151P', Aliases='', description='RNA, U6 small nuclear 1151, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481545, CurrentID=0, Status='live', Symbol='RNU6-1157P', Aliases='', description='RNA, U6 small nuclear 1157, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481546, CurrentID=0, Status='live', Symbol='RNU6-1160P', Aliases='', description='RNA, U6 small nuclear 1160, pseudogene', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481547, CurrentID=0, Status='live', Symbol='RNU6-1163P', Aliases='', description='RNA, U6 small nuclear 1163, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481548, CurrentID=0, Status='live', Symbol='RNU6-1168P', Aliases='', description='RNA, U6 small nuclear 1168, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481549, CurrentID=0, Status='live', Symbol='RNU6-1172P', Aliases='', description='RNA, U6 small nuclear 1172, pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481550, CurrentID=0, Status='live', Symbol='RNU6-1174P', Aliases='', description='RNA, U6 small nuclear 1174, pseudogene', other_designations='', map_location='3q22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400612, CurrentID=0, Status='live', Symbol='PRR29-AS1', Aliases='', description='PRR29 antisense RNA 1', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63996071, end_position_on_the_genomic_accession=63999899, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481551, CurrentID=0, Status='live', Symbol='RNU6-1176P', Aliases='', description='RNA, U6 small nuclear 1176, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374091, CurrentID=0, Status='live', Symbol='LOC105374091', Aliases='', description='uncharacterized LOC105374091', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126935167, end_position_on_the_genomic_accession=126950505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481552, CurrentID=0, Status='live', Symbol='RNU6-1180P', Aliases='', description='RNA, U6 small nuclear 1180, pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369727, CurrentID=0, Status='live', Symbol='LOC105369727', Aliases='', description='uncharacterized LOC105369727', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34036911, end_position_on_the_genomic_accession=34057906, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  285172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285172, CurrentID=0, Status='live', Symbol='FAM126B', Aliases='HYCC2', description='family with sequence similarity 126 member B', other_designations='family with sequence similarity 126, member B', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200973718, end_position_on_the_genomic_accession=201071671, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  106481555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481555, CurrentID=0, Status='live', Symbol='RNU6-1191P', Aliases='', description='RNA, U6 small nuclear 1191, pseudogene', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400617, CurrentID=0, Status='live', Symbol='KCNJ2-AS1', Aliases='', description='KCNJ2 antisense RNA 1 (head to head)', other_designations='KCNJ2 antisense RNA 1 (non-protein coding)', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70166961, end_position_on_the_genomic_accession=70169402, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481556, CurrentID=0, Status='live', Symbol='RNU6-1193P', Aliases='', description='RNA, U6 small nuclear 1193, pseudogene', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138474, CurrentID=0, Status='live', Symbol='TAF1L', Aliases='TAF(II)210, TAF2A2', description='TATA-box binding protein associated factor 1 like', other_designations='TAF1 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 210kDa-like|TBP-associated factor 1-like|TBP-associated factor 210 kDa|TBP-associated factor RNA polymerase 1-like|transcription initiation factor TFIID 210 kDa subunit', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32629454, end_position_on_the_genomic_accession=32635669, orientation='minus', exon_count=1, OMIM=607798, no_hdr0=''),
  106481557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481557, CurrentID=0, Status='live', Symbol='RNU6-1194P', Aliases='', description='RNA, U6 small nuclear 1194, pseudogene', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400619, CurrentID=0, Status='live', Symbol='LINC00511', Aliases='LCAL5, onco-lncRNA-12', description='long intergenic non-protein coding RNA 511', other_designations='lung cancer associated lncRNA 5', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72598041, end_position_on_the_genomic_accession=72640472, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481558, CurrentID=0, Status='live', Symbol='RNU6-1197P', Aliases='', description='RNA, U6 small nuclear 1197, pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400620, CurrentID=0, Status='live', Symbol='LOC400620', Aliases='', description='uncharacterized LOC400620', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73786822, end_position_on_the_genomic_accession=73800795, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106481559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481559, CurrentID=0, Status='live', Symbol='RNU6-1200P', Aliases='', description='RNA, U6 small nuclear 1200, pseudogene', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481560, CurrentID=0, Status='live', Symbol='RNU6-1204P', Aliases='', description='RNA, U6 small nuclear 1204, pseudogene', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481561, CurrentID=0, Status='live', Symbol='RNU6-1206P', Aliases='', description='RNA, U6 small nuclear 1206, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374093, CurrentID=0, Status='live', Symbol='LOC105374093', Aliases='', description='uncharacterized LOC105374093', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127276542, end_position_on_the_genomic_accession=127278237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189043, CurrentID=0, Status='live', Symbol='TRL-AAG5-1', Aliases='TRNAL9', description='transfer RNA-Leu (AAG) 5-1', other_designations='tRNA-Leu-AAG-5-1|transfer RNA leucine 9 (anticodon AAG)', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481562, CurrentID=0, Status='live', Symbol='RNU6-1210P', Aliases='', description='RNA, U6 small nuclear 1210, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400624, CurrentID=0, Status='live', Symbol='FLJ45079', Aliases='', description='FLJ45079 protein', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77879001, end_position_on_the_genomic_accession=77884087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481563, CurrentID=0, Status='live', Symbol='RNU6-1213P', Aliases='', description='RNA, U6 small nuclear 1213, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481564, CurrentID=0, Status='live', Symbol='RNU6-1217P', Aliases='', description='RNA, U6 small nuclear 1217, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372332, CurrentID=0, Status='live', Symbol='LOC105372332', Aliases='', description='uncharacterized LOC105372332', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22900594, end_position_on_the_genomic_accession=22920493, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106481565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481565, CurrentID=0, Status='live', Symbol='RNU6-1220P', Aliases='', description='RNA, U6 small nuclear 1220, pseudogene', other_designations='', map_location='8q21.12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400627, CurrentID=0, Status='live', Symbol='LOC400627', Aliases='', description='uncharacterized LOC400627', other_designations='CTD-2561B21.5', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80968621, end_position_on_the_genomic_accession=80971354, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481566, CurrentID=0, Status='live', Symbol='RNU6-1223P', Aliases='', description='RNA, U6 small nuclear 1223, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481567, CurrentID=0, Status='live', Symbol='RNU6-1227P', Aliases='', description='RNA, U6 small nuclear 1227, pseudogene', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481568, CurrentID=0, Status='live', Symbol='RNU6-1229P', Aliases='', description='RNA, U6 small nuclear 1229, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481569, CurrentID=0, Status='live', Symbol='RNU6-1233P', Aliases='', description='RNA, U6 small nuclear 1233, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481570, CurrentID=0, Status='live', Symbol='RNU6-1236P', Aliases='', description='RNA, U6 small nuclear 1236, pseudogene', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481571, CurrentID=0, Status='live', Symbol='RNU6-1239P', Aliases='', description='RNA, U6 small nuclear 1239, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533842, CurrentID=0, Status='live', Symbol='LOC100533842', Aliases='', description='ATP binding cassette subfamily G member 2 (Junior blood group) pseudogene', other_designations='ATP-binding cassette, sub-family G (WHITE), member 2 (Junior blood group) pseudogene|ATP-binding cassette, sub-family G (WHITE), member 2 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227802454, end_position_on_the_genomic_accession=227804674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481572, CurrentID=0, Status='live', Symbol='RNU6-1241P', Aliases='', description='RNA, U6 small nuclear 1241, pseudogene', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481573, CurrentID=0, Status='live', Symbol='RNU6-1243P', Aliases='', description='RNA, U6 small nuclear 1243, pseudogene', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728315, CurrentID=0, Status='live', Symbol='RPL7AP31', Aliases='RPL7A_7_480', description='ribosomal protein L7a pseudogene 31', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56356113, end_position_on_the_genomic_accession=56356989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481574, CurrentID=0, Status='live', Symbol='RNU6-1247P', Aliases='', description='RNA, U6 small nuclear 1247, pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481575, CurrentID=0, Status='live', Symbol='RNU6-1250P', Aliases='', description='RNA, U6 small nuclear 1250, pseudogene', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481576, CurrentID=0, Status='live', Symbol='RNU6-1254P', Aliases='', description='RNA, U6 small nuclear 1254, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131625, CurrentID=0, Status='live', Symbol='LOC100131625', Aliases='', description='ADP-ribosylation factor like GTPase 8B pseudogene', other_designations='ADP-ribosylation factor-like 8B pseudogene', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98234714, end_position_on_the_genomic_accession=98235561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724179, CurrentID=0, Status='live', Symbol='LOC102724179', Aliases='', description='putative BCR-like protein 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481578, CurrentID=0, Status='live', Symbol='RNU6-1260P', Aliases='', description='RNA, U6 small nuclear 1260, pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481579, CurrentID=0, Status='live', Symbol='RNU6-1265P', Aliases='', description='RNA, U6 small nuclear 1265, pseudogene', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481580, CurrentID=0, Status='live', Symbol='RNU6-1267P', Aliases='', description='RNA, U6 small nuclear 1267, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481581, CurrentID=0, Status='live', Symbol='RNU6-1270P', Aliases='', description='RNA, U6 small nuclear 1270, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533844, CurrentID=0, Status='live', Symbol='TARS2P1', Aliases='', description='threonyl-tRNA synthetase 2, mitochondrial pseudogene', other_designations='threonyl-tRNA synthetase 2, mitochondrial (putative) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135564894, end_position_on_the_genomic_accession=135567162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481582, CurrentID=0, Status='live', Symbol='RNU6-1274P', Aliases='', description='RNA, U6 small nuclear 1274, pseudogene', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400644, CurrentID=0, Status='live', Symbol='LINC01443', Aliases='', description='long intergenic non-protein coding RNA 1443', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14946267, end_position_on_the_genomic_accession=14973756, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481583, CurrentID=0, Status='live', Symbol='RNU6-1276P', Aliases='', description='RNA, U6 small nuclear 1276, pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400645, CurrentID=0, Status='live', Symbol='RBM22P1', Aliases='', description='RNA binding motif protein 22 pseudogene 1', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27418884, end_position_on_the_genomic_accession=27420436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285178, CurrentID=0, Status='live', Symbol='LOC285178', Aliases='', description='uncharacterized LOC285178', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481585, CurrentID=0, Status='live', Symbol='RNU6-1282P', Aliases='', description='RNA, U6 small nuclear 1282, pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481586, CurrentID=0, Status='live', Symbol='RNU6-1286P', Aliases='', description='RNA, U6 small nuclear 1286, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533845, CurrentID=0, Status='live', Symbol='LOC100533845', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69242628, end_position_on_the_genomic_accession=69250815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481587, CurrentID=0, Status='live', Symbol='RNU6-1289P', Aliases='', description='RNA, U6 small nuclear 1289, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481588, CurrentID=0, Status='live', Symbol='RNU6-1291P', Aliases='', description='RNA, U6 small nuclear 1291, pseudogene', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127945, CurrentID=0, Status='live', Symbol='SOCS5P3', Aliases='', description='suppressor of cytokine signaling 5 pseudogene 3', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44709311, end_position_on_the_genomic_accession=44710886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724191, CurrentID=0, Status='live', Symbol='LOC102724191', Aliases='', description='chondroitin sulfate proteoglycan 4-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84473892, end_position_on_the_genomic_accession=84488064, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  400650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400650, CurrentID=0, Status='live', Symbol='TPMTP1', Aliases='', description='thiopurine S-methyltransferase pseudogene 1', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47630015, end_position_on_the_genomic_accession=47631792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481589, CurrentID=0, Status='live', Symbol='RNU6-1295P', Aliases='', description='RNA, U6 small nuclear 1295, pseudogene', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481590, CurrentID=0, Status='live', Symbol='RNU6-1298P', Aliases='', description='RNA, U6 small nuclear 1298, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400652, CurrentID=0, Status='live', Symbol='RPS3AP49', Aliases='RPS3A_23_1608', description='ribosomal protein S3a pseudogene 49', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60149544, end_position_on_the_genomic_accession=60150407, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481591, CurrentID=0, Status='live', Symbol='RNU6-1301P', Aliases='', description='RNA, U6 small nuclear 1301, pseudogene', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533846, CurrentID=0, Status='live', Symbol='LOC100533846', Aliases='', description='solute carrier family 30 (zinc transporter), member 5 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46111077, end_position_on_the_genomic_accession=46113996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481592, CurrentID=0, Status='live', Symbol='RNU6-1305P', Aliases='', description='RNA, U6 small nuclear 1305, pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400654, CurrentID=0, Status='live', Symbol='LINC01538', Aliases='', description='long intergenic non-protein coding RNA 1538', other_designations='TCONS_00026184', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64213082, end_position_on_the_genomic_accession=64260055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481593, CurrentID=0, Status='live', Symbol='RNU6-1307P', Aliases='', description='RNA, U6 small nuclear 1307, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400655, CurrentID=0, Status='live', Symbol='LOC400655', Aliases='', description='uncharacterized LOC400655', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=73154058, end_position_on_the_genomic_accession=73264498, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  285180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285180, CurrentID=0, Status='live', Symbol='RUFY4', Aliases='ZFYVE31', description='RUN and FYVE domain containing 4', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218034934, end_position_on_the_genomic_accession=218090625, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  106481595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481595, CurrentID=0, Status='live', Symbol='RNU6-1314P', Aliases='', description='RNA, U6 small nuclear 1314, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400657, CurrentID=0, Status='live', Symbol='LINC00909', Aliases='', description='long intergenic non-protein coding RNA 909', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74591774, end_position_on_the_genomic_accession=74597835, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481596, CurrentID=0, Status='live', Symbol='RNU6-1317P', Aliases='', description='RNA, U6 small nuclear 1317, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372856, CurrentID=0, Status='live', Symbol='LOC105372856', Aliases='', description='uncharacterized LOC105372856', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400658, CurrentID=0, Status='live', Symbol='C18orf65', Aliases='', description='chromosome 18 open reading frame 65', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76495521, end_position_on_the_genomic_accession=76498088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481597, CurrentID=0, Status='live', Symbol='RNU6-1321P', Aliases='', description='RNA, U6 small nuclear 1321, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481598, CurrentID=0, Status='live', Symbol='RNU6-1323P', Aliases='', description='RNA, U6 small nuclear 1323, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400660, CurrentID=0, Status='live', Symbol='LINC00683', Aliases='', description='long intergenic non-protein coding RNA 683', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76619777, end_position_on_the_genomic_accession=76623559, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481599, CurrentID=0, Status='live', Symbol='RNU6-1327P', Aliases='', description='RNA, U6 small nuclear 1327, pseudogene', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481600, CurrentID=0, Status='live', Symbol='RNU6-1329P', Aliases='', description='RNA, U6 small nuclear 1329, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400662, CurrentID=0, Status='live', Symbol='LOC400662', Aliases='', description='uncharacterized LOC400662', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533848, CurrentID=0, Status='live', Symbol='LOC100533848', Aliases='', description='ATPase, H+/K+ transporting, nongastric, alpha polypeptide pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217693895, end_position_on_the_genomic_accession=217696752, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481602, CurrentID=0, Status='live', Symbol='RNU6-1337P', Aliases='', description='RNA, U6 small nuclear 1337, pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481605, CurrentID=0, Status='live', Symbol='TCP11X3P', Aliases='', description='t-complex 11 family, X-linked 3, pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7451, CurrentID=0, Status='live', Symbol='VWFP1', Aliases='F8VWFL, F8VWFP, VWFP', description='von Willebrand factor pseudogene 1', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16690035, end_position_on_the_genomic_accession=16704461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481606, CurrentID=0, Status='live', Symbol='RNU1-31P', Aliases='', description='RNA, U1 small nuclear 31, pseudogene', other_designations='', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481607, CurrentID=0, Status='live', Symbol='RNU1-34P', Aliases='', description='RNA, U1 small nuclear 34, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481608, CurrentID=0, Status='live', Symbol='RNU1-38P', Aliases='', description='RNA, U1 small nuclear 38, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51145, CurrentID=0, Status='live', Symbol='LOC51145', Aliases='', description='uncharacterized LOC51145', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125743754, end_position_on_the_genomic_accession=125745929, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481610, CurrentID=0, Status='live', Symbol='RNU1-45P', Aliases='', description='RNA, U1 small nuclear 45, pseudogene', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481611, CurrentID=0, Status='live', Symbol='RNU1-54P', Aliases='', description='RNA, U1 small nuclear 54, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7457, CurrentID=0, Status='live', Symbol='WBS2', Aliases='WBS, WS', description='Williams-Beuren syndrome type 2', other_designations='', map_location='4q33-q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481612, CurrentID=0, Status='live', Symbol='RNU1-58P', Aliases='', description='RNA, U1 small nuclear 58, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481613, CurrentID=0, Status='live', Symbol='RNU1-62P', Aliases='', description='RNA, U1 small nuclear 62, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369739, CurrentID=0, Status='live', Symbol='LOC105369739', Aliases='', description='uncharacterized LOC105369739', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42682793, end_position_on_the_genomic_accession=42717098, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106481614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481614, CurrentID=0, Status='live', Symbol='RNU1-65P', Aliases='', description='RNA, U1 small nuclear 65, pseudogene', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481615, CurrentID=0, Status='live', Symbol='RNU1-68P', Aliases='', description='RNA, U1 small nuclear 68, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372507, CurrentID=0, Status='live', Symbol='LOC105372507', Aliases='', description='uncharacterized LOC105372507', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2667713, end_position_on_the_genomic_accession=2692614, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481616, CurrentID=0, Status='live', Symbol='RNU1-73P', Aliases='', description='RNA, U1 small nuclear 73, pseudogene', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481618, CurrentID=0, Status='live', Symbol='RNU1-84P', Aliases='', description='RNA, U1 small nuclear 84, pseudogene', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400680, CurrentID=0, Status='live', Symbol='LINC00664', Aliases='', description='long intergenic non-protein coding RNA 664', other_designations='CTD-2561J22.4', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21483715, end_position_on_the_genomic_accession=21503238, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106481619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481619, CurrentID=0, Status='live', Symbol='RNU1-88P', Aliases='', description='RNA, U1 small nuclear 88, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373972, CurrentID=0, Status='live', Symbol='LOC105373972', Aliases='', description='uncharacterized LOC105373972', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241108801, end_position_on_the_genomic_accession=241111925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481620, CurrentID=0, Status='live', Symbol='RNU1-92P', Aliases='', description='RNA, U1 small nuclear 92, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23314, CurrentID=0, Status='live', Symbol='SATB2', Aliases='GLSS', description='SATB homeobox 2', other_designations='SATB family member 2|special AT-rich sequence-binding protein 2', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199269500, end_position_on_the_genomic_accession=199471266, orientation='minus', exon_count=14, OMIM=608148, no_hdr0=''),
  400682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400682, CurrentID=0, Status='live', Symbol='LOC400682', Aliases='', description='zinc finger protein 100-like', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21637371, end_position_on_the_genomic_accession=21657708, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481621, CurrentID=0, Status='live', Symbol='RNU1-95P', Aliases='', description='RNA, U1 small nuclear 95, pseudogene', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400684, CurrentID=0, Status='live', Symbol='LOC400684', Aliases='', description='uncharacterized LOC400684', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32390050, end_position_on_the_genomic_accession=32405539, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  400685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400685, CurrentID=0, Status='live', Symbol='LOC400685', Aliases='', description='uncharacterized LOC400685', other_designations='CTC-523E23.4', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34816154, end_position_on_the_genomic_accession=34832869, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106481624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481624, CurrentID=0, Status='live', Symbol='RNU1-106P', Aliases='', description='RNA, U1 small nuclear 106, pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421729, CurrentID=0, Status='live', Symbol='VPS26BP1', Aliases='', description='VPS26B pseudogene 1', other_designations='vacuolar protein sorting 26 homolog B pseudogene 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47960330, end_position_on_the_genomic_accession=47961081, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481625, CurrentID=0, Status='live', Symbol='RNU1-109P', Aliases='', description='RNA, U1 small nuclear 109, pseudogene', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481626, CurrentID=0, Status='live', Symbol='RNU1-114P', Aliases='', description='RNA, U1 small nuclear 114, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533853, CurrentID=0, Status='live', Symbol='LOC100533853', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45331807, end_position_on_the_genomic_accession=45332624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481627, CurrentID=0, Status='live', Symbol='RNU1-120P', Aliases='', description='RNA, U1 small nuclear 120, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481628, CurrentID=0, Status='live', Symbol='RNU1-123P', Aliases='', description='RNA, U1 small nuclear 123, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481630, CurrentID=0, Status='live', Symbol='RNU1-129P', Aliases='RNU1-137P', description='RNA, U1 small nuclear 129, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724233, CurrentID=0, Status='live', Symbol='LOC102724233', Aliases='', description='uncharacterized LOC102724233', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65030482, end_position_on_the_genomic_accession=65052504, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  400692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400692, CurrentID=0, Status='live', Symbol='LOC400692', Aliases='', description='uncharacterized LOC400692', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=38036372, end_position_on_the_genomic_accession=38040067, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481631, CurrentID=0, Status='live', Symbol='RNU1-133P', Aliases='', description='RNA, U1 small nuclear 133, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481632, CurrentID=0, Status='live', Symbol='RNU1-136P', Aliases='', description='RNA, U1 small nuclear 136, pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481634, CurrentID=0, Status='live', Symbol='RNU1-143P', Aliases='', description='RNA, U1 small nuclear 143, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400696, CurrentID=0, Status='live', Symbol='LGALS17A', Aliases='', description='Charcot-Leyden crystal protein pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39679374, end_position_on_the_genomic_accession=39686373, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106481635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481635, CurrentID=0, Status='live', Symbol='RNU1-148P', Aliases='', description='RNA, U1 small nuclear 148, pseudogene', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132430, CurrentID=0, Status='live', Symbol='PABPC4L', Aliases='', description='poly(A) binding protein, cytoplasmic 4-like', other_designations='PABP-4-like', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134196333, end_position_on_the_genomic_accession=134201748, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132319, CurrentID=0, Status='live', Symbol='LOC100132319', Aliases='', description='uncharacterized LOC100132319', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=188868602, end_position_on_the_genomic_accession=188868763, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481636, CurrentID=0, Status='live', Symbol='RNU2-9P', Aliases='', description='RNA, U2 small nuclear 9, pseudogene', other_designations='', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289442, CurrentID=0, Status='live', Symbol='LOC100289442', Aliases='', description='PHD finger protein 5A pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141307608, end_position_on_the_genomic_accession=141307933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481638, CurrentID=0, Status='live', Symbol='RNU2-14P', Aliases='', description='RNA, U2 small nuclear 14, pseudogene', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127955, CurrentID=0, Status='live', Symbol='LOC100127955', Aliases='', description='uncharacterized LOC100127955', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1837682, end_position_on_the_genomic_accession=1849931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481639, CurrentID=0, Status='live', Symbol='RNU2-16P', Aliases='', description='RNA, U2 small nuclear 16, pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481640, CurrentID=0, Status='live', Symbol='RNU2-20P', Aliases='', description='RNA, U2 small nuclear 20, pseudogene', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100037267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100037267, CurrentID=0, Status='live', Symbol='LOC100037267', Aliases='', description='developmental pluripotency associated 4 pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73989304, end_position_on_the_genomic_accession=73992184, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481641, CurrentID=0, Status='live', Symbol='RNU2-23P', Aliases='', description='RNA, U2 small nuclear 23, pseudogene', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481643, CurrentID=0, Status='live', Symbol='RNU2-30P', Aliases='', description='RNA, U2 small nuclear 30, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400706, CurrentID=0, Status='live', Symbol='LOC400706', Aliases='', description='uncharacterized LOC400706', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46057678, end_position_on_the_genomic_accession=46077629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481645, CurrentID=0, Status='live', Symbol='RNU2-36P', Aliases='', description='RNA, U2 small nuclear 36, pseudogene', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400707, CurrentID=0, Status='live', Symbol='LOC400707', Aliases='', description='uncharacterized LOC400707', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=51689605, end_position_on_the_genomic_accession=51691595, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481646, CurrentID=0, Status='live', Symbol='RNU2-38P', Aliases='', description='RNA, U2 small nuclear 38, pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481648, CurrentID=0, Status='live', Symbol='RNU2-44P', Aliases='', description='RNA, U2 small nuclear 44, pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400710, CurrentID=0, Status='live', Symbol='FLJ26850', Aliases='', description='FLJ26850 protein', other_designations='CTD-2126E3.1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50050680, end_position_on_the_genomic_accession=50066795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  414926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414926, CurrentID=0, Status='live', Symbol='LINC00593', Aliases='C15orf50', description='long intergenic non-protein coding RNA 593', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69835234, end_position_on_the_genomic_accession=69842967, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481649, CurrentID=0, Status='live', Symbol='RNU2-48P', Aliases='', description='RNA, U2 small nuclear 48, pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285191, CurrentID=0, Status='live', Symbol='LOC285191', Aliases='', description='uncharacterized LOC285191', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240685942, end_position_on_the_genomic_accession=240689527, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481650, CurrentID=0, Status='live', Symbol='RNU2-50P', Aliases='', description='RNA, U2 small nuclear 50, pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132322, CurrentID=0, Status='live', Symbol='SAR1AP4', Aliases='', description='SAR1 homolog A (S. cerevisiae) pseudogene 4', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75884546, end_position_on_the_genomic_accession=75885141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481651, CurrentID=0, Status='live', Symbol='RNU2-54P', Aliases='', description='RNA, U2 small nuclear 54, pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7497, CurrentID=0, Status='live', Symbol='XCE', Aliases='XCA, XITE', description='X chromosome controlling element', other_designations='', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300074, no_hdr0=''),
  106481652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481652, CurrentID=0, Status='live', Symbol='RNU2-58P', Aliases='', description='RNA, U2 small nuclear 58, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481655, CurrentID=0, Status='live', Symbol='RNU2-66P', Aliases='', description='RNA, U2 small nuclear 66, pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481656, CurrentID=0, Status='live', Symbol='RNU2-69P', Aliases='', description='RNA, U2 small nuclear 69, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374112, CurrentID=0, Status='live', Symbol='LOC105374112', Aliases='', description='uncharacterized LOC105374112', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131998043, end_position_on_the_genomic_accession=132013560, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481657, CurrentID=0, Status='live', Symbol='MED15P3', Aliases='', description='mediator complex subunit 15 pseudogene 3', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  346296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346296, CurrentID=0, Status='live', Symbol='LOC346296', Aliases='', description='insulin like growth factor 2 mRNA binding protein 3 pseudogene', other_designations='insulin-like growth factor 2 mRNA binding protein 3 pseudogene|pseudogene of IGF-II mRNA-binding protein 3', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56597548, end_position_on_the_genomic_accession=56601545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481658, CurrentID=0, Status='live', Symbol='ENPP7P3', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 3', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400720, CurrentID=0, Status='live', Symbol='ZNF772', Aliases='', description='zinc finger protein 772', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57469586, end_position_on_the_genomic_accession=57477570, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369748, CurrentID=0, Status='live', Symbol='LOC105369748', Aliases='', description='uncharacterized LOC105369748', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47706058, end_position_on_the_genomic_accession=47740134, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106481659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481659, CurrentID=0, Status='live', Symbol='RARRES2P3', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 3', other_designations='', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481660, CurrentID=0, Status='live', Symbol='RARRES2P7', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 7', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481662, CurrentID=0, Status='live', Symbol='HNRNPDLP3', Aliases='', description='heterogeneous nuclear ribonucleoprotein D-like pseudogene 3', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481663, CurrentID=0, Status='live', Symbol='HNRNPA1P49', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 49', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874293, CurrentID=0, Status='live', Symbol='GPC5-IT1', Aliases='', description='GPC5 intronic transcript 1', other_designations='GPC5 intronic transcript 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481664, CurrentID=0, Status='live', Symbol='HNRNPA1P65', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 65', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481665, CurrentID=0, Status='live', Symbol='MTHFD2P3', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724850, CurrentID=0, Status='live', Symbol='LOC102724850', Aliases='', description='uncharacterized LOC102724850', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481666, CurrentID=0, Status='live', Symbol='RPL37P25', Aliases='', description='ribosomal protein L37 pseudogene 25', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724269, CurrentID=0, Status='live', Symbol='LOC102724269', Aliases='', description='uncharacterized LOC102724269', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144860101, end_position_on_the_genomic_accession=144860807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  400728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400728, CurrentID=0, Status='live', Symbol='FAM87B', Aliases='', description='family with sequence similarity 87 member B', other_designations='family with sequence similarity 87, member B', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=817371, end_position_on_the_genomic_accession=819834, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481667, CurrentID=0, Status='live', Symbol='PRELID3BP2', Aliases='SLMO2P2', description='PRELI domain containing 3B pseudogene 2', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481668, CurrentID=0, Status='live', Symbol='MEMO1P5', Aliases='', description='mediator of cell motility 1 pseudogene 5', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481669, CurrentID=0, Status='live', Symbol='KLF17P2', Aliases='', description='Kruppel-like factor 17 pseudogene 2', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285195, CurrentID=0, Status='live', Symbol='SLC9A9', Aliases='AUTS16, NHE9', description='solute carrier family 9, subfamily A (NHE9, cation proton antiporter 9), member 9', other_designations='Na(+)/H(+) exchanger 9|putative protein product of Nbla00118|sodium/proton exchanger NHE9|solute carrier family 9 (sodium/hydrogen exchanger)', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143265221, end_position_on_the_genomic_accession=143848531, orientation='minus', exon_count=19, OMIM=608396, no_hdr0=''),
  106481670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481670, CurrentID=0, Status='live', Symbol='TCEB1P30', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 30', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442318, CurrentID=0, Status='live', Symbol='SAPCD2P4', Aliases='', description='suppressor APC domain containing 2 pseudogene 4', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63113440, end_position_on_the_genomic_accession=63114882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481671, CurrentID=0, Status='live', Symbol='TCEB1P34', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 34', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481672, CurrentID=0, Status='live', Symbol='KLF2P3', Aliases='', description='Kruppel-like factor 2 pseudogene 3', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724275, CurrentID=0, Status='live', Symbol='LOC102724275', Aliases='', description='uncharacterized LOC102724275', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70852088, end_position_on_the_genomic_accession=70872555, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106481673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481673, CurrentID=0, Status='live', Symbol='HSPE1P11', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 11', other_designations='heat shock 10kDa protein 1 pseudogene 11', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127962, CurrentID=0, Status='live', Symbol='LOC100127962', Aliases='', description='coiled-coil serine rich protein 2 pseudogene', other_designations='coiled-coil serine-rich protein 2 pseudogene|family with sequence similarity 190, member B pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102515164, end_position_on_the_genomic_accession=102516842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374777, CurrentID=0, Status='live', Symbol='LOC105374777', Aliases='', description='uncharacterized LOC105374777', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64761384, end_position_on_the_genomic_accession=64768186, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  400736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400736, CurrentID=0, Status='live', Symbol='PRAMEF13', Aliases='', description='PRAME family member 13', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13196188, end_position_on_the_genomic_accession=13201409, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374116, CurrentID=0, Status='live', Symbol='LOC105374116', Aliases='', description='uncharacterized LOC105374116', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133784139, end_position_on_the_genomic_accession=133805694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481677, CurrentID=0, Status='live', Symbol='HSPE1P26', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 26', other_designations='heat shock 10kDa protein 1 pseudogene 26', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481678, CurrentID=0, Status='live', Symbol='SUMO2P10', Aliases='', description='SUMO2 pseudogene 10', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442320, CurrentID=0, Status='live', Symbol='SAPCD2P1', Aliases='', description='suppressor APC domain containing 2 pseudogene 1', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64181612, end_position_on_the_genomic_accession=64183093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374117, CurrentID=0, Status='live', Symbol='LOC105374117', Aliases='', description='uncharacterized LOC105374117', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134064586, end_position_on_the_genomic_accession=134069053, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400743, CurrentID=0, Status='live', Symbol='LOC400743', Aliases='', description='uncharacterized LOC400743', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17194326, end_position_on_the_genomic_accession=17194898, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481682, CurrentID=0, Status='live', Symbol='COX6CP12', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 12', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422022, CurrentID=0, Status='live', Symbol='LOC100422022', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69306464, end_position_on_the_genomic_accession=69307173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  400745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400745, CurrentID=0, Status='live', Symbol='SH2D5', Aliases='', description='SH2 domain containing 5', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20719731, end_position_on_the_genomic_accession=20732640, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101101692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101692, CurrentID=0, Status='live', Symbol='HELLPAR', Aliases='LINC-HELLP, lncHELLP', description='HELLP associated long non-coding RNA', other_designations='long intergenic non-protein coding RNA associated with HELLP syndrome', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614985, no_hdr0=''),
  106481685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481685, CurrentID=0, Status='live', Symbol='SNRPGP17', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 17', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101101698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101698, CurrentID=0, Status='live', Symbol='POROK7', Aliases='', description='Porokeratosis 7, disseminated superficial actinic', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614714, no_hdr0=''),
  106481686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481686, CurrentID=0, Status='live', Symbol='SNRPGP19', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 19', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400748, CurrentID=0, Status='live', Symbol='LOC400748', Aliases='', description='uncharacterized LOC400748', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30858158, end_position_on_the_genomic_accession=30860254, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481688, CurrentID=0, Status='live', Symbol='YWHAEP2', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400750, CurrentID=0, Status='live', Symbol='LOC400750', Aliases='', description='heat shock protein family A (Hsp70) member 5 pseudogene', other_designations='heat shock 70kDa protein 5 (glucose-regulated protein, 78kDa) pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38708930, end_position_on_the_genomic_accession=38710889, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369754, CurrentID=0, Status='live', Symbol='LOC105369754', Aliases='', description='uncharacterized LOC105369754', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48357561, end_position_on_the_genomic_accession=48361380, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481689, CurrentID=0, Status='live', Symbol='IGKV1OR2-2', Aliases='', description='immunoglobulin kappa variable 1/OR2-2 (pseudogene)', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481690, CurrentID=0, Status='live', Symbol='EIF1P5', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 5', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400752, CurrentID=0, Status='live', Symbol='LINC01144', Aliases='', description='long intergenic non-protein coding RNA 1144', other_designations='TCONS_00000513', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45303910, end_position_on_the_genomic_accession=45305619, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481691, CurrentID=0, Status='live', Symbol='EIF1P7', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 7', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481692, CurrentID=0, Status='live', Symbol='BNIP3P15', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 15', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101232810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101232810, CurrentID=0, Status='live', Symbol='LOC101232810', Aliases='', description='stathmin 3 pseudogene', other_designations='stathmin-like 3 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25177662, end_position_on_the_genomic_accession=25179556, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481694, CurrentID=0, Status='live', Symbol='BNIP3P30', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 30', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481695, CurrentID=0, Status='live', Symbol='BNIP3P33', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 33', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400757, CurrentID=0, Status='live', Symbol='C1orf141', Aliases='', description='chromosome 1 open reading frame 141', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67092176, end_position_on_the_genomic_accession=67141645, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  106481696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481696, CurrentID=0, Status='live', Symbol='BNIP3P36', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 36', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481697, CurrentID=0, Status='live', Symbol='BNIP3P42', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 42', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101101768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101768, CurrentID=0, Status='live', Symbol='ECTD8', Aliases='', description='Ectodermal dysplasia 8, hair/tooth/nail type', other_designations='', map_location='18q22.1-q22.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602401, no_hdr0=''),
  101101769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101769, CurrentID=0, Status='live', Symbol='ECTD5', Aliases='', description='Ectodermal dysplasia 5, hair/nail type', other_designations='', map_location='10q24.32-q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614927, no_hdr0=''),
  101101770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101770, CurrentID=0, Status='live', Symbol='ECTD6', Aliases='', description='Ectodermal dysplasia 6, hair/nail type', other_designations='', map_location='17p12-q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614928, no_hdr0=''),
  101101771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101771, CurrentID=0, Status='live', Symbol='ECTD7', Aliases='', description='Ectodermal dysplasia 7, hair/nail type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614929, no_hdr0=''),
  101101772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101772, CurrentID=0, Status='live', Symbol='LINC00601', Aliases='', description='long intergenic non-protein coding RNA 601', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126413869, end_position_on_the_genomic_accession=126421879, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101101773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101773, CurrentID=0, Status='live', Symbol='LINC00609', Aliases='', description='long intergenic non-protein coding RNA 609', other_designations='', map_location='14q13.2-q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36070427, end_position_on_the_genomic_accession=36165288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101101775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101775, CurrentID=0, Status='live', Symbol='TMEM220-AS1', Aliases='', description='TMEM220 antisense RNA 1', other_designations='CTC-297N7.5', map_location='17p13.1-p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10729777, end_position_on_the_genomic_accession=10815164, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101101776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101776, CurrentID=0, Status='live', Symbol='BMS1P17', Aliases='LINC00516', description='BMS1, ribosome biogenesis factor pseudogene 17', other_designations='BMS1 pseudogene 17|BMS1 ribosome biogenesis factor pseudogene 17|ribosome biogenesis protein BMS1 homolog pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19306690, end_position_on_the_genomic_accession=19316914, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400761, CurrentID=0, Status='live', Symbol='FLJ27354', Aliases='', description='uncharacterized LOC400761', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89624849, end_position_on_the_genomic_accession=89632894, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481700, CurrentID=0, Status='live', Symbol='LARP7P3', Aliases='', description='La ribonucleoprotein domain family member 7 pseudogene 3', other_designations='La ribonucleoprotein domain family, member 7 pseudogene 3', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481701, CurrentID=0, Status='live', Symbol='SNRPCP10', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 10', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481703, CurrentID=0, Status='live', Symbol='NENFP3', Aliases='', description='neudesin neurotrophic factor pseudogene 3', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400765, CurrentID=0, Status='live', Symbol='MIR137HG', Aliases='', description='MIR137 host gene', other_designations='MIR137 host gene (non-protein coding)', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97988000, end_position_on_the_genomic_accession=98049693, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481705, CurrentID=0, Status='live', Symbol='FAM27AP1', Aliases='', description='family with sequence similarity 27 member A pseudogene 1', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400768, CurrentID=0, Status='live', Symbol='LOC400768', Aliases='', description='uncharacterized LOC400768', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108420689, end_position_on_the_genomic_accession=108433184, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400769, CurrentID=0, Status='live', Symbol='HNRNPA1P43', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 43', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115856630, end_position_on_the_genomic_accession=115857878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481708, CurrentID=0, Status='live', Symbol='VTA1P1', Aliases='', description='vesicle (multivesicular body) trafficking 1 pseudogene 1', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378487, CurrentID=0, Status='live', Symbol='LOC105378487', Aliases='', description='uncharacterized LOC105378487', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112799580, end_position_on_the_genomic_accession=112807617, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  285203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285203, CurrentID=0, Status='live', Symbol='EOGT', Aliases='AER61, AOS4, C3orf641, EOGT', description='EGF domain-specific O-linked N-acetylglucosamine (GlcNAc) transferase', other_designations='AER61 glycosyltransferase|EGF-O-GlcNAc transferase|extracellular O-linked N-acetylglucosamine transferase', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68975212, end_position_on_the_genomic_accession=69013961, orientation='minus', exon_count=21, OMIM=614789, no_hdr0=''),
  101101841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101101841, CurrentID=0, Status='live', Symbol='LOC101101841', Aliases='', description='mediator complex subunit 21 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103568337, end_position_on_the_genomic_accession=103568994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481710, CurrentID=0, Status='live', Symbol='RNY1P11', Aliases='', description='RNA, Ro-associated Y1 pseudogene 11', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481711, CurrentID=0, Status='live', Symbol='RNY1P15', Aliases='', description='RNA, Ro-associated Y1 pseudogene 15', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481712, CurrentID=0, Status='live', Symbol='RNY4P34', Aliases='', description='RNA, Ro-associated Y4 pseudogene 34', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481713, CurrentID=0, Status='live', Symbol='RNY3P11', Aliases='', description='RNA, Ro-associated Y3 pseudogene 11', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481714, CurrentID=0, Status='live', Symbol='RNY3P13', Aliases='', description='RNA, Ro-associated Y3 pseudogene 13', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285498, CurrentID=0, Status='live', Symbol='RNF212', Aliases='ZHP3', description='ring finger protein 212', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1056248, end_position_on_the_genomic_accession=1113794, orientation='minus', exon_count=18, OMIM=612041, no_hdr0=''),
  101060112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060112, CurrentID=0, Status='live', Symbol='LOC101060112', Aliases='', description='RAP1A, member of RAS oncogene family pseudogene', other_designations='ras-related protein Rap-1A-like', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28443357, end_position_on_the_genomic_accession=28445924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481715, CurrentID=0, Status='live', Symbol='IMMP1LP3', Aliases='', description='inner mitochondrial membrane peptidase subunit 1 pseudogene 3', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481716, CurrentID=0, Status='live', Symbol='SMARCE1P7', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 7', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374124, CurrentID=0, Status='live', Symbol='LOC105374124', Aliases='', description='uncharacterized LOC105374124', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135832431, end_position_on_the_genomic_accession=135840637, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481717, CurrentID=0, Status='live', Symbol='B4GALNT2P1', Aliases='', description='beta-1,4-N-acetyl-galactosaminyl transferase 2 pseudogene 1', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373103, CurrentID=0, Status='live', Symbol='LOC105373103', Aliases='', description='uncharacterized LOC105373103', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=1744535, end_position_on_the_genomic_accession=1776605, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283994, CurrentID=0, Status='live', Symbol='LINC00868', Aliases='C17orf52', description='long intergenic non-protein coding RNA 868', other_designations='', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127971, CurrentID=0, Status='live', Symbol='LOC100127971', Aliases='', description='uncharacterized LOC100127971', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=50487389, end_position_on_the_genomic_accession=50491626, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271012, CurrentID=0, Status='live', Symbol='RPL36P12', Aliases='RPL36_4_831', description='ribosomal protein L36 pseudogene 12', other_designations='', map_location='7q22-q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105244589, end_position_on_the_genomic_accession=105245015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481719, CurrentID=0, Status='live', Symbol='C8orf59P1', Aliases='', description='chromosome 8 open reading frame 59 pseudogene 1', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285205, CurrentID=0, Status='live', Symbol='LINC00636', Aliases='', description='long intergenic non-protein coding RNA 636', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107883205, end_position_on_the_genomic_accession=107928907, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481720, CurrentID=0, Status='live', Symbol='ENPP7P14', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 14', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724323, CurrentID=0, Status='live', Symbol='LOC102724323', Aliases='', description='uncharacterized LOC102724323', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45444568, end_position_on_the_genomic_accession=45453121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481721, CurrentID=0, Status='live', Symbol='BRD9P1', Aliases='', description='bromodomain containing 9 pseudogene 1', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138639, CurrentID=0, Status='live', Symbol='PTPDC1', Aliases='PTP9Q22', description='protein tyrosine phosphatase domain containing 1', other_designations='protein tyrosine phosphatase PTP9Q22', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94030794, end_position_on_the_genomic_accession=94109856, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  106481722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481722, CurrentID=0, Status='live', Symbol='GPAA1P2', Aliases='', description='glycosylphosphatidylinositol anchor attachment 1 pseudogene 2', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481723, CurrentID=0, Status='live', Symbol='ART2BP', Aliases='', description='ADP-ribosyltransferase 2B, pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874305, CurrentID=0, Status='live', Symbol='NAV2-IT1', Aliases='', description='NAV2 intronic transcript 1', other_designations='NAV2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481724, CurrentID=0, Status='live', Symbol='BRWD1P1', Aliases='', description='bromodomain and WD repeat domain containing 1 pseudogene 1', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374126, CurrentID=0, Status='live', Symbol='LOC105374126', Aliases='', description='uncharacterized LOC105374126', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137626912, end_position_on_the_genomic_accession=137713834, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481727, CurrentID=0, Status='live', Symbol='SETD6P1', Aliases='', description='SET domain containing 6 pseudogene 1', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481728, CurrentID=0, Status='live', Symbol='DUX4L28', Aliases='', description='double homeobox 4 like 28', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481729, CurrentID=0, Status='live', Symbol='C1DP5', Aliases='', description='C1D nuclear receptor corepressor pseudogene 5', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481730, CurrentID=0, Status='live', Symbol='C1GALT1P3', Aliases='', description='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase 1 pseudogene 3', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369317, CurrentID=0, Status='live', Symbol='LOC105369317', Aliases='', description='uncharacterized LOC105369317', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59560307, end_position_on_the_genomic_accession=59565495, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106481731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481731, CurrentID=0, Status='live', Symbol='RPL23AP92', Aliases='', description='ribosomal protein L23a pseudogene 92', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400793, CurrentID=0, Status='live', Symbol='C1orf226', Aliases='', description='chromosome 1 open reading frame 226', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162378906, end_position_on_the_genomic_accession=162386818, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481732, CurrentID=0, Status='live', Symbol='CAB39P1', Aliases='', description='calcium binding protein 39 pseudogene 1', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422024, CurrentID=0, Status='live', Symbol='LOC100422024', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B7 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69574341, end_position_on_the_genomic_accession=69585527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481734, CurrentID=0, Status='live', Symbol='GTF2IP9', Aliases='', description='general transcription factor IIi pseudogene 9', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130476, CurrentID=0, Status='live', Symbol='LOC100130476', Aliases='', description='uncharacterized LOC100130476', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137823669, end_position_on_the_genomic_accession=137868233, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  138652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138652, CurrentID=0, Status='live', Symbol='PRSS47', Aliases='', description='protease, serine 47', other_designations='protease, serine, 47', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=94939445, end_position_on_the_genomic_accession=94952858, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106481735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481735, CurrentID=0, Status='live', Symbol='GTF2IP14', Aliases='', description='general transcription factor IIi pseudogene 14', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400797, CurrentID=0, Status='live', Symbol='LINC00083', Aliases='NCRNA00083', description='long intergenic non-protein coding RNA 83', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178478153, end_position_on_the_genomic_accession=178499619, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106481736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481736, CurrentID=0, Status='live', Symbol='NF1P11', Aliases='', description='neurofibromin 1 pseudogene 11', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400798, CurrentID=0, Status='live', Symbol='C1orf220', Aliases='', description='chromosome 1 open reading frame 220', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178542796, end_position_on_the_genomic_accession=178548889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  378953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378953, CurrentID=0, Status='live', Symbol='RBMY2HP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member H pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7671468, end_position_on_the_genomic_accession=7677978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481738, CurrentID=0, Status='live', Symbol='DUX4L47', Aliases='', description='double homeobox 4 like 47, pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874308, CurrentID=0, Status='live', Symbol='RSF1-IT1', Aliases='', description='RSF1 intronic transcript 1', other_designations='RSF1 intronic transcript 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400800, CurrentID=0, Status='live', Symbol='LOC400800', Aliases='', description='uncharacterized LOC400800', other_designations='', map_location='1q31.3-q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199148598, end_position_on_the_genomic_accession=199393307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481739, CurrentID=0, Status='live', Symbol='FRG2MP', Aliases='', description='FSHD region gene 2 family member M, pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481740, CurrentID=0, Status='live', Symbol='DUX4L51', Aliases='', description='double homeobox 4 like 51, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481741, CurrentID=0, Status='live', Symbol='DUXAP11', Aliases='', description='double homeobox A pseudogene 11', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51278, CurrentID=0, Status='live', Symbol='IER5', Aliases='SBBI48', description='immediate early response 5', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181088502, end_position_on_the_genomic_accession=181090843, orientation='plus', exon_count=1, OMIM=607177, no_hdr0=''),
  100313785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313785, CurrentID=0, Status='live', Symbol='PIRC20', Aliases='', description='piwi-interacting RNA cluster 20', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481743, CurrentID=0, Status='live', Symbol='MTND6P12', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 12', other_designations='MT-ND6 pseudogene 12', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930523, CurrentID=0, Status='live', Symbol='LOC101930523', Aliases='', description='uncharacterized LOC101930523', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104890353, end_position_on_the_genomic_accession=104901853, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481744, CurrentID=0, Status='live', Symbol='AGBL5-IT1', Aliases='', description='AGBL5 intronic transcript 1', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481745, CurrentID=0, Status='live', Symbol='RNA5SP443', Aliases='RN5S443', description='RNA, 5S ribosomal pseudogene 443', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643101, CurrentID=0, Status='live', Symbol='RPL26P22', Aliases='RPL26_5_860', description='ribosomal protein L26 pseudogene 22', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143532694, end_position_on_the_genomic_accession=143533210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481746, CurrentID=0, Status='live', Symbol='UBE2V1P7', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 7', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 7', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374130, CurrentID=0, Status='live', Symbol='LOC105374130', Aliases='', description='uncharacterized LOC105374130', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139600522, end_position_on_the_genomic_accession=139606896, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481747, CurrentID=0, Status='live', Symbol='RNU7-159P', Aliases='', description='RNA, U7 small nuclear 159 pseudogene', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481748, CurrentID=0, Status='live', Symbol='RN7SKP72', Aliases='', description='RNA, 7SK small nuclear pseudogene 72', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481749, CurrentID=0, Status='live', Symbol='RN7SKP119', Aliases='', description='RNA, 7SK small nuclear pseudogene 119', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481750, CurrentID=0, Status='live', Symbol='RN7SKP212', Aliases='', description='RNA, 7SK small nuclear pseudogene 212', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724353, CurrentID=0, Status='live', Symbol='LOC102724353', Aliases='', description='uncharacterized LOC102724353', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130122077, end_position_on_the_genomic_accession=130125779, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  400812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400812, CurrentID=0, Status='live', Symbol='FLJ39095', Aliases='', description='uncharacterized LOC400812', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247104511, end_position_on_the_genomic_accession=247109561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481751, CurrentID=0, Status='live', Symbol='RN7SKP259', Aliases='', description='RNA, 7SK small nuclear pseudogene 259', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481753, CurrentID=0, Status='live', Symbol='RN7SL145P', Aliases='', description='RNA, 7SL, cytoplasmic 145, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127978, CurrentID=0, Status='live', Symbol='LOC100127978', Aliases='', description='starch binding domain 1 pseudogene', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47484689, end_position_on_the_genomic_accession=47485722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874311, CurrentID=0, Status='live', Symbol='ATG10-IT1', Aliases='', description='ATG10 intronic transcript 1', other_designations='ATG10 intronic transcript 1 (non-protein coding)', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930525, CurrentID=0, Status='live', Symbol='LOC101930525', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16215390, end_position_on_the_genomic_accession=16220617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481754, CurrentID=0, Status='live', Symbol='RN7SL223P', Aliases='', description='RNA, 7SL, cytoplasmic 223, pseudogene', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481755, CurrentID=0, Status='live', Symbol='RN7SL311P', Aliases='', description='RNA, 7SL, cytoplasmic 311, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481756, CurrentID=0, Status='live', Symbol='RN7SL393P', Aliases='', description='RNA, 7SL, cytoplasmic 393, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724359, CurrentID=0, Status='live', Symbol='LOC102724359', Aliases='', description='uncharacterized LOC102724359', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53393831, end_position_on_the_genomic_accession=53487960, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106481757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481757, CurrentID=0, Status='live', Symbol='RN7SL480P', Aliases='', description='RNA, 7SL, cytoplasmic 480, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369613, CurrentID=0, Status='live', Symbol='LOC105369613', Aliases='', description='uncharacterized LOC105369613', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4549908, end_position_on_the_genomic_accession=4556346, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481758, CurrentID=0, Status='live', Symbol='RN7SL505P', Aliases='', description='RNA, 7SL, cytoplasmic 505, pseudogene', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481759, CurrentID=0, Status='live', Symbol='RN7SL598P', Aliases='', description='RNA, 7SL, cytoplasmic 598, pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375476, CurrentID=0, Status='live', Symbol='LOC105375476', Aliases='', description='uncharacterized LOC105375476', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120166481, end_position_on_the_genomic_accession=120185032, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481760, CurrentID=0, Status='live', Symbol='RN7SL656P', Aliases='', description='RNA, 7SL, cytoplasmic 656, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132344, CurrentID=0, Status='live', Symbol='LOC100132344', Aliases='', description='uncharacterized LOC100132344', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=83186918, end_position_on_the_genomic_accession=83187364, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481761, CurrentID=0, Status='live', Symbol='RN7SL774P', Aliases='', description='RNA, 7SL, cytoplasmic 774, pseudogene', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481762, CurrentID=0, Status='live', Symbol='RN7SL827P', Aliases='', description='RNA, 7SL, cytoplasmic 827, pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724365, CurrentID=0, Status='live', Symbol='LOC102724365', Aliases='', description='otopetrin-1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89531543, end_position_on_the_genomic_accession=89539037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481763, CurrentID=0, Status='live', Symbol='RNU4-13P', Aliases='', description='RNA, U4 small nuclear 13, pseudogene', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481764, CurrentID=0, Status='live', Symbol='RNU4-59P', Aliases='', description='RNA, U4 small nuclear 59, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481765, CurrentID=0, Status='live', Symbol='RNU6-114P', Aliases='', description='RNA, U6 small nuclear 114, pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481766, CurrentID=0, Status='live', Symbol='RNU6-207P', Aliases='', description='RNA, U6 small nuclear 207, pseudogene', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481767, CurrentID=0, Status='live', Symbol='RNU6-278P', Aliases='', description='RNA, U6 small nuclear 278, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481768, CurrentID=0, Status='live', Symbol='RNU6-346P', Aliases='', description='RNA, U6 small nuclear 346, pseudogene', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724371, CurrentID=0, Status='live', Symbol='LOC102724371', Aliases='', description='uncharacterized LOC102724371', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24902309, end_position_on_the_genomic_accession=24904261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481769, CurrentID=0, Status='live', Symbol='RNU6-420P', Aliases='', description='RNA, U6 small nuclear 420, pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481770, CurrentID=0, Status='live', Symbol='RNU6-466P', Aliases='', description='RNA, U6 small nuclear 466, pseudogene', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481771, CurrentID=0, Status='live', Symbol='RNU6-559P', Aliases='', description='RNA, U6 small nuclear 559, pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481772, CurrentID=0, Status='live', Symbol='RNU6-607P', Aliases='', description='RNA, U6 small nuclear 607, pseudogene', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481773, CurrentID=0, Status='live', Symbol='RNU6-674P', Aliases='', description='RNA, U6 small nuclear 674, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481774, CurrentID=0, Status='live', Symbol='RNU6-721P', Aliases='', description='RNA, U6 small nuclear 721, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400836, CurrentID=0, Status='live', Symbol='SDAD1P2', Aliases='', description='SDA1 domain containing 1 pseudogene 2', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10385685, end_position_on_the_genomic_accession=10388625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481775, CurrentID=0, Status='live', Symbol='RNU6-768P', Aliases='', description='RNA, U6 small nuclear 768, pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481776, CurrentID=0, Status='live', Symbol='RNU6-864P', Aliases='', description='RNA, U6 small nuclear 864, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481777, CurrentID=0, Status='live', Symbol='RNU6-910P', Aliases='', description='RNA, U6 small nuclear 910, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481778, CurrentID=0, Status='live', Symbol='RNU6-1003P', Aliases='', description='RNA, U6 small nuclear 1003, pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481779, CurrentID=0, Status='live', Symbol='RNU6-1049P', Aliases='', description='RNA, U6 small nuclear 1049, pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400841, CurrentID=0, Status='live', Symbol='LINC01597', Aliases='', description='long intergenic non-protein coding RNA 1597', other_designations='', map_location='20q11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=29513582, end_position_on_the_genomic_accession=29521213, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481780, CurrentID=0, Status='live', Symbol='RNU6-1121P', Aliases='', description='RNA, U6 small nuclear 1121, pseudogene', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724383, CurrentID=0, Status='live', Symbol='LOC102724383', Aliases='', description='uncharacterized LOC102724383', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9146489, end_position_on_the_genomic_accession=9171264, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481781, CurrentID=0, Status='live', Symbol='RNU6-1214P', Aliases='', description='RNA, U6 small nuclear 1214, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106345152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106345152, CurrentID=0, Status='live', Symbol='LOC106345152', Aliases='', description='uncharacterized LOC106345152', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533884, CurrentID=0, Status='live', Symbol='LOC100533884', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481782, CurrentID=0, Status='live', Symbol='RNU6-1283P', Aliases='', description='RNA, U6 small nuclear 1283, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533757, CurrentID=0, Status='live', Symbol='LOC100533757', Aliases='', description='Sjogren syndrome antigen B (autoantigen La) pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45071774, end_position_on_the_genomic_accession=45073035, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481783, CurrentID=0, Status='live', Symbol='RNU1-36P', Aliases='', description='RNA, U1 small nuclear 36, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874317, CurrentID=0, Status='live', Symbol='FABP7P2', Aliases='', description='fatty acid binding protein 7, brain pseudogene 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95368542, end_position_on_the_genomic_accession=95368883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481784, CurrentID=0, Status='live', Symbol='NDUFA5P7', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 7', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 7', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481785, CurrentID=0, Status='live', Symbol='YWHAEP4', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 4', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481786, CurrentID=0, Status='live', Symbol='SNRPCP13', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 13', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533885, CurrentID=0, Status='live', Symbol='LOC100533885', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481787, CurrentID=0, Status='live', Symbol='FTH1P6', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene 6', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7633, CurrentID=0, Status='live', Symbol='ZNF79', Aliases='pT7', description='zinc finger protein 79', other_designations='ZNFpT7', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127424374, end_position_on_the_genomic_accession=127446964, orientation='plus', exon_count=6, OMIM=194552, no_hdr0=''),
  106481788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481788, CurrentID=0, Status='live', Symbol='MTND6P19', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 19', other_designations='MT-ND6 pseudogene 19', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481790, CurrentID=0, Status='live', Symbol='RC3H1-IT1', Aliases='', description='RC3H1 intronic transcript 1', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724875, CurrentID=0, Status='live', Symbol='LOC102724875', Aliases='', description='uncharacterized LOC102724875', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110494502, end_position_on_the_genomic_accession=110512013, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481791, CurrentID=0, Status='live', Symbol='CYP4A44P', Aliases='', description='cytochrome P450 family 4 subfamily A member 44, pseudogene', other_designations='cytochrome P450, family 4, subfamily A, polypeptide 44, pseudogene', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533886, CurrentID=0, Status='live', Symbol='LOC100533886', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728533, CurrentID=0, Status='live', Symbol='LOC728533', Aliases='', description='IST1 homolog', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37791772, end_position_on_the_genomic_accession=37792865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=305, CurrentID=0, Status='live', Symbol='ANXA2P3', Aliases='ANX2L3, ANX2P3, LIP2, LPC2C', description='annexin A2 pseudogene 3', other_designations='annexin II (lipocortin II) pseudogene 3', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64825528, end_position_on_the_genomic_accession=64826877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929099, CurrentID=0, Status='live', Symbol='LOC101929099', Aliases='', description='uncharacterized LOC101929099', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117025482, end_position_on_the_genomic_accession=117059490, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481793, CurrentID=0, Status='live', Symbol='RNA5SP254', Aliases='RN5S254', description='RNA, 5S ribosomal pseudogene 254', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100127986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127986, CurrentID=0, Status='live', Symbol='LOC100127986', Aliases='', description='cation channel, sperm associated 2 pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119011342, end_position_on_the_genomic_accession=119012138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481794, CurrentID=0, Status='live', Symbol='RNA5SP350', Aliases='RN5S350', description='RNA, 5S ribosomal pseudogene 350', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481796, CurrentID=0, Status='live', Symbol='LRRC37A15P', Aliases='', description='leucine-rich repeat containing 37 member A15, pseudogene', other_designations='leucine rich repeat containing 37, member A15, pseudogene', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481797, CurrentID=0, Status='live', Symbol='WBP1LP6', Aliases='', description='WW domain binding protein 1-like pseudogene 6', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138715, CurrentID=0, Status='live', Symbol='ARID3C', Aliases='', description='AT-rich interaction domain 3C', other_designations='ARID domain-containing protein 3C|AT rich interactive domain 3C (BRIGHT-like)|Brightlike', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34621286, end_position_on_the_genomic_accession=34628950, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106481798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481798, CurrentID=0, Status='live', Symbol='RPEP5', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 5', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138716, CurrentID=0, Status='live', Symbol='RPP25L', Aliases='C9orf23, bA296L22.5', description='ribonuclease P/MRP 25kDa subunit-like', other_designations='RNase P protein subunit-like p25|alba-like protein C9orf23|rpp25-like protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34610485, end_position_on_the_genomic_accession=34612113, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481799, CurrentID=0, Status='live', Symbol='RNU7-94P', Aliases='', description='RNA, U7 small nuclear 94 pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132352, CurrentID=0, Status='live', Symbol='FRG1HP', Aliases='', description='FSHD region gene 1 family member H, pseudogene', other_designations='FSHD region gene 1 pseudogene', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40992337, end_position_on_the_genomic_accession=41014139, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106481801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481801, CurrentID=0, Status='live', Symbol='RNU7-138P', Aliases='', description='RNA, U7 small nuclear 138 pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481802, CurrentID=0, Status='live', Symbol='RNU7-154P', Aliases='', description='RNA, U7 small nuclear 154 pseudogene', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481803, CurrentID=0, Status='live', Symbol='RNU7-176P', Aliases='', description='RNA, U7 small nuclear 176 pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481804, CurrentID=0, Status='live', Symbol='RNU7-193P', Aliases='', description='RNA, U7 small nuclear 193 pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481805, CurrentID=0, Status='live', Symbol='RN7SKP22', Aliases='', description='RNA, 7SK small nuclear pseudogene 22', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374142, CurrentID=0, Status='live', Symbol='LOC105374142', Aliases='', description='uncharacterized LOC105374142', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145556893, end_position_on_the_genomic_accession=145569213, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100533889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533889, CurrentID=0, Status='live', Symbol='LOC100533889', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138724, CurrentID=0, Status='live', Symbol='C9orf131', Aliases='', description='chromosome 9 open reading frame 131', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35041095, end_position_on_the_genomic_accession=35045991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481807, CurrentID=0, Status='live', Symbol='RN7SKP56', Aliases='', description='RNA, 7SK small nuclear pseudogene 56', other_designations='', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481811, CurrentID=0, Status='live', Symbol='RN7SKP121', Aliases='', description='RNA, 7SK small nuclear pseudogene 121', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533890, CurrentID=0, Status='live', Symbol='LOC100533890', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481812, CurrentID=0, Status='live', Symbol='RN7SKP133', Aliases='', description='RNA, 7SK small nuclear pseudogene 133', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378508, CurrentID=0, Status='live', Symbol='LOC105378508', Aliases='', description='uncharacterized LOC105378508', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118783980, end_position_on_the_genomic_accession=118866308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106481814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481814, CurrentID=0, Status='live', Symbol='RN7SKP166', Aliases='', description='RNA, 7SK small nuclear pseudogene 166', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285224, CurrentID=0, Status='live', Symbol='DNAJB8-AS1', Aliases='', description='DNAJB8 antisense RNA 1', other_designations='DNAJB8 antisense RNA 1 (non-protein coding)', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128463594, end_position_on_the_genomic_accession=128472317, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481815, CurrentID=0, Status='live', Symbol='RN7SKP179', Aliases='', description='RNA, 7SK small nuclear pseudogene 179', other_designations='', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481816, CurrentID=0, Status='live', Symbol='RN7SKP195', Aliases='', description='RNA, 7SK small nuclear pseudogene 195', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533891, CurrentID=0, Status='live', Symbol='LOC100533891', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400879, CurrentID=0, Status='live', Symbol='LOC400879', Aliases='', description='uncharacterized LOC400879', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=16157306, end_position_on_the_genomic_accession=16157938, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132849, CurrentID=0, Status='live', Symbol='LOC100132849', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F1 complex, O subunit pseudogene', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206236166, end_position_on_the_genomic_accession=206236883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100127991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127991, CurrentID=0, Status='live', Symbol='LOC100127991', Aliases='', description='uncharacterized LOC100127991', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=157044694, end_position_on_the_genomic_accession=157045245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271016, CurrentID=0, Status='live', Symbol='RPL37P11', Aliases='RPL37_3_203', description='ribosomal protein L37 pseudogene 11', other_designations='', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26678219, end_position_on_the_genomic_accession=26678510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724423, CurrentID=0, Status='live', Symbol='LOC102724423', Aliases='', description='uncharacterized LOC102724423', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89348086, end_position_on_the_genomic_accession=89349210, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724881, CurrentID=0, Status='live', Symbol='LOC102724881', Aliases='', description='uncharacterized LOC102724881', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19344397, end_position_on_the_genomic_accession=19403517, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374145, CurrentID=0, Status='live', Symbol='LOC105374145', Aliases='', description='uncharacterized LOC105374145', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146164657, end_position_on_the_genomic_accession=146190531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100289479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289479, CurrentID=0, Status='live', Symbol='LOC100289479', Aliases='', description='cytochrome c oxidase subunit VIIb pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173338222, end_position_on_the_genomic_accession=173338461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  346329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346329, CurrentID=0, Status='live', Symbol='LOC346329', Aliases='', description='guanine nucleotide binding protein (G protein), alpha 11 (Gq class) pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66504863, end_position_on_the_genomic_accession=66506431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481823, CurrentID=0, Status='live', Symbol='RN7SKP296', Aliases='', description='RNA, 7SK small nuclear pseudogene 296', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724426, CurrentID=0, Status='live', Symbol='LOC102724426', Aliases='', description='Sjoegren syndrome nuclear autoantigen 1 homolog', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49316657, end_position_on_the_genomic_accession=49318310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874325, CurrentID=0, Status='live', Symbol='ATP2B2-IT2', Aliases='', description='ATP2B2 intronic transcript 2', other_designations='ATP2B2 intronic transcript 2 (non-protein coding)', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10626015, end_position_on_the_genomic_accession=10626807, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481824, CurrentID=0, Status='live', Symbol='RN7SL34P', Aliases='', description='RNA, 7SL, cytoplasmic 34, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481825, CurrentID=0, Status='live', Symbol='RN7SL48P', Aliases='', description='RNA, 7SL, cytoplasmic 48, pseudogene', other_designations='', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481826, CurrentID=0, Status='live', Symbol='RN7SL128P', Aliases='', description='RNA, 7SL, cytoplasmic 128, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533893, CurrentID=0, Status='live', Symbol='LOC100533893', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481827, CurrentID=0, Status='live', Symbol='RN7SL146P', Aliases='', description='RNA, 7SL, cytoplasmic 146, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481828, CurrentID=0, Status='live', Symbol='RN7SL162P', Aliases='', description='RNA, 7SL, cytoplasmic 162, pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724431, CurrentID=0, Status='live', Symbol='LOC102724431', Aliases='', description='uncharacterized LOC102724431', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40189071, end_position_on_the_genomic_accession=40214200, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  106481829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481829, CurrentID=0, Status='live', Symbol='RN7SL187P', Aliases='', description='RNA, 7SL, cytoplasmic 187, pseudogene', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724432, CurrentID=0, Status='live', Symbol='LOC102724432', Aliases='', description='uncharacterized LOC102724432', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46983685, end_position_on_the_genomic_accession=46997099, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  400891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400891, CurrentID=0, Status='live', Symbol='LRRC74B', Aliases='', description='leucine rich repeat containing 74B', other_designations='leucine-rich repeat-containing protein LOC400891', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21045960, end_position_on_the_genomic_accession=21064168, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  106481830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481830, CurrentID=0, Status='live', Symbol='RN7SL201P', Aliases='', description='RNA, 7SL, cytoplasmic 201, pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481831, CurrentID=0, Status='live', Symbol='RN7SL222P', Aliases='', description='RNA, 7SL, cytoplasmic 222, pseudogene', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533894, CurrentID=0, Status='live', Symbol='LOC100533894', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481832, CurrentID=0, Status='live', Symbol='RN7SL252P', Aliases='', description='RNA, 7SL, cytoplasmic 252, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728575, CurrentID=0, Status='live', Symbol='LOC728575', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71233384, end_position_on_the_genomic_accession=71234073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724437, CurrentID=0, Status='live', Symbol='LOC102724437', Aliases='', description='profilin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149084455, end_position_on_the_genomic_accession=149085262, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481835, CurrentID=0, Status='live', Symbol='RN7SL293P', Aliases='', description='RNA, 7SL, cytoplasmic 293, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481836, CurrentID=0, Status='live', Symbol='RN7SL314P', Aliases='', description='RNA, 7SL, cytoplasmic 314, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724439, CurrentID=0, Status='live', Symbol='LOC102724439', Aliases='', description='uncharacterized LOC102724439', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106285468, end_position_on_the_genomic_accession=106310945, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533895, CurrentID=0, Status='live', Symbol='LOC100533895', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481837, CurrentID=0, Status='live', Symbol='RN7SL328P', Aliases='', description='RNA, 7SL, cytoplasmic 328, pseudogene', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481838, CurrentID=0, Status='live', Symbol='RN7SL354P', Aliases='', description='RNA, 7SL, cytoplasmic 354, pseudogene', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481839, CurrentID=0, Status='live', Symbol='RN7SL371P', Aliases='', description='RNA, 7SL, cytoplasmic 371, pseudogene', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481841, CurrentID=0, Status='live', Symbol='RN7SL408P', Aliases='', description='RNA, 7SL, cytoplasmic 408, pseudogene', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533896, CurrentID=0, Status='live', Symbol='LOC100533896', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481842, CurrentID=0, Status='live', Symbol='RN7SL426P', Aliases='', description='RNA, 7SL, cytoplasmic 426, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481843, CurrentID=0, Status='live', Symbol='RN7SL448P', Aliases='', description='RNA, 7SL, cytoplasmic 448, pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874329, CurrentID=0, Status='live', Symbol='KCNJ6-AS1', Aliases='KCNJ6-IT1', description='KCNJ6 antisense RNA 1', other_designations='KCNJ6 intronic transcript 1 (non-protein coding)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481844, CurrentID=0, Status='live', Symbol='RN7SL461P', Aliases='', description='RNA, 7SL, cytoplasmic 461, pseudogene', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728587, CurrentID=0, Status='live', Symbol='LOC728587', Aliases='', description='zinc finger protein 736 pseudogene', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46870809, end_position_on_the_genomic_accession=46872059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481846, CurrentID=0, Status='live', Symbol='RN7SL497P', Aliases='', description='RNA, 7SL, cytoplasmic 497, pseudogene', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372858, CurrentID=0, Status='live', Symbol='LOC105372858', Aliases='', description='uncharacterized LOC105372858', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18205810, end_position_on_the_genomic_accession=18211924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100533897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533897, CurrentID=0, Status='live', Symbol='LOC100533897', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481848, CurrentID=0, Status='live', Symbol='RN7SL545P', Aliases='', description='RNA, 7SL, cytoplasmic 545, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481849, CurrentID=0, Status='live', Symbol='RN7SL559P', Aliases='', description='RNA, 7SL, cytoplasmic 559, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481850, CurrentID=0, Status='live', Symbol='RN7SL566P', Aliases='', description='RNA, 7SL, cytoplasmic 566, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533898, CurrentID=0, Status='live', Symbol='LOC100533898', Aliases='', description='fibrosin-like 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481852, CurrentID=0, Status='live', Symbol='RN7SL617P', Aliases='', description='RNA, 7SL, cytoplasmic 617, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481853, CurrentID=0, Status='live', Symbol='RN7SL644P', Aliases='', description='RNA, 7SL, cytoplasmic 644, pseudogene', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378516, CurrentID=0, Status='live', Symbol='LOC105378516', Aliases='', description='uncharacterized LOC105378516', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120815003, end_position_on_the_genomic_accession=120825307, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106481854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481854, CurrentID=0, Status='live', Symbol='RN7SL666P', Aliases='', description='RNA, 7SL, cytoplasmic 666, pseudogene', other_designations='', map_location='20q13.11', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7700, CurrentID=0, Status='live', Symbol='ZNF141', Aliases='D4S90, PAPA6, pHZ-44', description='zinc finger protein 141', other_designations='zinc finger protein 141 (clone pHZ-44)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=337807, end_position_on_the_genomic_accession=373959, orientation='plus', exon_count=5, OMIM=194648, no_hdr0=''),
  102724458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724458, CurrentID=0, Status='live', Symbol='LOC102724458', Aliases='', description='uncharacterized LOC102724458', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163709170, end_position_on_the_genomic_accession=163731614, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100533899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533899, CurrentID=0, Status='live', Symbol='LOC100533899', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481857, CurrentID=0, Status='live', Symbol='RN7SL728P', Aliases='', description='RNA, 7SL, cytoplasmic 728, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481439, CurrentID=0, Status='live', Symbol='RNU6-818P', Aliases='', description='RNA, U6 small nuclear 818, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481861, CurrentID=0, Status='live', Symbol='RN7SL808P', Aliases='', description='RNA, 7SL, cytoplasmic 808, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533900, CurrentID=0, Status='live', Symbol='LOC100533900', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481862, CurrentID=0, Status='live', Symbol='RN7SL831P', Aliases='', description='RNA, 7SL, cytoplasmic 831, pseudogene', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116935, CurrentID=0, Status='live', Symbol='RPL3P2', Aliases='RPL3-Hom, RPL3_1_679', description='ribosomal protein L3 pseudogene 2', other_designations='ribosomal protein L3 homologous', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31280291, end_position_on_the_genomic_accession=31281571, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481864, CurrentID=0, Status='live', Symbol='RN7SL869P', Aliases='', description='RNA, 7SL, cytoplasmic 869, pseudogene', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724467, CurrentID=0, Status='live', Symbol='LOC102724467', Aliases='', description='uncharacterized LOC102724467', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87766331, end_position_on_the_genomic_accession=87779148, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400926, CurrentID=0, Status='live', Symbol='FLJ90680', Aliases='', description='FLJ90680 protein', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=37242075, end_position_on_the_genomic_accession=37244244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101060118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060118, CurrentID=0, Status='live', Symbol='LOC101060118', Aliases='', description='uncharacterized LOC101060118', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23123743, end_position_on_the_genomic_accession=23127048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372509, CurrentID=0, Status='live', Symbol='LOC105372509', Aliases='', description='uncharacterized LOC105372509', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3408017, end_position_on_the_genomic_accession=3420788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927779, CurrentID=0, Status='live', Symbol='LOC101927779', Aliases='', description='uncharacterized LOC101927779', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79358486, end_position_on_the_genomic_accession=79358780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400927, CurrentID=0, Status='live', Symbol='LOC400927', Aliases='', description='TPTE and PTEN homologous inositol lipid phosphatase pseudogene', other_designations='hypothetical protein LOC400927|transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38344665, end_position_on_the_genomic_accession=38398926, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132365, CurrentID=0, Status='live', Symbol='LOC100132365', Aliases='', description='uncharacterized LOC100132365', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=73833396, end_position_on_the_genomic_accession=73836135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100533901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533901, CurrentID=0, Status='live', Symbol='LOC100533901', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724471, CurrentID=0, Status='live', Symbol='LOC102724471', Aliases='', description='uncharacterized LOC102724471', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110667336, end_position_on_the_genomic_accession=110670836, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128001, CurrentID=0, Status='live', Symbol='LOC100128001', Aliases='', description='uncharacterized LOC100128001', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=464726, end_position_on_the_genomic_accession=465618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481870, CurrentID=0, Status='live', Symbol='RNU4-25P', Aliases='', description='RNA, U4 small nuclear 25, pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400932, CurrentID=0, Status='live', Symbol='LINC00898', Aliases='', description='long intergenic non-protein coding RNA 898', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47621043, end_position_on_the_genomic_accession=47631569, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374155, CurrentID=0, Status='live', Symbol='LOC105374155', Aliases='', description='uncharacterized LOC105374155', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150759105, end_position_on_the_genomic_accession=150768728, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400934, CurrentID=0, Status='live', Symbol='FLJ44385', Aliases='', description='uncharacterized FLJ44385', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=49942329, end_position_on_the_genomic_accession=49946817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  107001062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=107001062, CurrentID=0, Status='live', Symbol='LOC107001062', Aliases='', description='uncharacterized LOC107001062', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481873, CurrentID=0, Status='live', Symbol='RNU4-84P', Aliases='', description='RNA, U4 small nuclear 84, pseudogene', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128002, CurrentID=0, Status='live', Symbol='LOC100128002', Aliases='', description='uncharacterized LOC100128002', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131295522, end_position_on_the_genomic_accession=131298212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369791, CurrentID=0, Status='live', Symbol='LOC105369791', Aliases='', description='uncharacterized LOC105369791', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58920630, end_position_on_the_genomic_accession=59165442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106481875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481875, CurrentID=0, Status='live', Symbol='RNU6-108P', Aliases='', description='RNA, U6 small nuclear 108, pseudogene', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481876, CurrentID=0, Status='live', Symbol='RNU6-122P', Aliases='', description='RNA, U6 small nuclear 122, pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369792, CurrentID=0, Status='live', Symbol='LOC105369792', Aliases='', description='uncharacterized LOC105369792', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=59486918, end_position_on_the_genomic_accession=59524149, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724482, CurrentID=0, Status='live', Symbol='LOC102724482', Aliases='', description='uncharacterized LOC102724482', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74750190, end_position_on_the_genomic_accession=74778780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100533904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533904, CurrentID=0, Status='live', Symbol='LOC100533904', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138799, CurrentID=0, Status='live', Symbol='OR13C5', Aliases='OR9-11', description='olfactory receptor family 13 subfamily C member 5', other_designations='olfactory receptor OR9-11|olfactory receptor, family 13, subfamily C, member 5', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104598457, end_position_on_the_genomic_accession=104599413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724485, CurrentID=0, Status='live', Symbol='LOC102724485', Aliases='', description='putative cyclin-Y-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34351483, end_position_on_the_genomic_accession=34381482, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  102724486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724486, CurrentID=0, Status='live', Symbol='LOC102724486', Aliases='', description='uncharacterized LOC102724486', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  138802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138802, CurrentID=0, Status='live', Symbol='OR13C8', Aliases='OR37H, OR9-10', description='olfactory receptor family 13 subfamily C member 8', other_designations='olfactory receptor OR9-10 pseudogene|olfactory receptor, family 13, subfamily C, member 8', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104569168, end_position_on_the_genomic_accession=104570130, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481885, CurrentID=0, Status='live', Symbol='RNU6-261P', Aliases='', description='RNA, U6 small nuclear 261, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481886, CurrentID=0, Status='live', Symbol='RNU6-280P', Aliases='RNU6-1221P', description='RNA, U6 small nuclear 280, pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7732, CurrentID=0, Status='live', Symbol='RNF112', Aliases='BFP, ZNF179', description='ring finger protein 112', other_designations='brain finger protein|neurolastin|zinc finger protein 179', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19410790, end_position_on_the_genomic_accession=19417276, orientation='plus', exon_count=15, OMIM=601237, no_hdr0=''),
  102724490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724490, CurrentID=0, Status='live', Symbol='LOC102724490', Aliases='', description='uncharacterized LOC102724490', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15525591, end_position_on_the_genomic_accession=15531616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  138805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138805, CurrentID=0, Status='live', Symbol='OR13F1', Aliases='OR9-6', description='olfactory receptor family 13 subfamily F member 1', other_designations='olfactory receptor OR9-6|olfactory receptor, family 13, subfamily F, member 1', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104504263, end_position_on_the_genomic_accession=104505222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481890, CurrentID=0, Status='live', Symbol='RNU6-367P', Aliases='', description='RNA, U6 small nuclear 367, pseudogene', other_designations='', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927780, CurrentID=0, Status='live', Symbol='LOC101927780', Aliases='', description='uncharacterized LOC101927780', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61556272, end_position_on_the_genomic_accession=61570653, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106481891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481891, CurrentID=0, Status='live', Symbol='RNU6-387P', Aliases='', description='RNA, U6 small nuclear 387, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374159, CurrentID=0, Status='live', Symbol='LOC105374159', Aliases='', description='uncharacterized LOC105374159', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151894334, end_position_on_the_genomic_accession=151902924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481892, CurrentID=0, Status='live', Symbol='RNU6-399P', Aliases='', description='RNA, U6 small nuclear 399, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724495, CurrentID=0, Status='live', Symbol='LOC102724495', Aliases='', description='uncharacterized LOC102724495', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118259835, end_position_on_the_genomic_accession=118322890, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400954, CurrentID=0, Status='live', Symbol='EML6', Aliases='', description='echinoderm microtubule associated protein like 6', other_designations='EMAP-6', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54723670, end_position_on_the_genomic_accession=54972908, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  106481893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481893, CurrentID=0, Status='live', Symbol='RNU6-410P', Aliases='', description='RNA, U6 small nuclear 410, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103724386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103724386, CurrentID=0, Status='live', Symbol='LINC00377', Aliases='TCONS_00021861', description='long intergenic non-protein coding RNA 377', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81018391, end_position_on_the_genomic_accession=81044442, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  400955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400955, CurrentID=0, Status='live', Symbol='LINC01122', Aliases='', description='long intergenic non-protein coding RNA 1122', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=58520753, end_position_on_the_genomic_accession=59063766, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  103724388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103724388, CurrentID=0, Status='live', Symbol='LINC01068', Aliases='', description='long intergenic non-protein coding RNA 1068', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79566727, end_position_on_the_genomic_accession=79571446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103724389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103724389, CurrentID=0, Status='live', Symbol='LINC01170', Aliases='', description='long intergenic non-protein coding RNA 1170', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124059794, end_position_on_the_genomic_accession=124438520, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  103724390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103724390, CurrentID=0, Status='live', Symbol='LINC01287', Aliases='TCONS_l2_00027522', description='long intergenic non-protein coding RNA 1287', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153399919, end_position_on_the_genomic_accession=153412234, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724497, CurrentID=0, Status='live', Symbol='LOC102724497', Aliases='', description='uncharacterized LOC102724497', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74715209, end_position_on_the_genomic_accession=74754846, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23369, CurrentID=0, Status='live', Symbol='PUM2', Aliases='PUMH2, PUML2', description='pumilio RNA binding family member 2', other_designations='pumilio RNA-binding family member 2|pumilio-2', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20248692, end_position_on_the_genomic_accession=20352229, orientation='minus', exon_count=28, OMIM=607205, no_hdr0=''),
  400957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400957, CurrentID=0, Status='live', Symbol='LINC01185', Aliases='', description='long intergenic non-protein coding RNA 1185', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60847760, end_position_on_the_genomic_accession=60881314, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104384744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104384744, CurrentID=0, Status='live', Symbol='TET2-AS1', Aliases='', description='TET2 antisense RNA 1', other_designations='TET2 antisense RNA 1 (non-protein coding)', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105171354, end_position_on_the_genomic_accession=105352881, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106481897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481897, CurrentID=0, Status='live', Symbol='RNU6-461P', Aliases='', description='RNA, U6 small nuclear 461, pseudogene', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418539, CurrentID=0, Status='live', Symbol='LOC100418539', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481898, CurrentID=0, Status='live', Symbol='RNU6-486P', Aliases='', description='RNA, U6 small nuclear 486, pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128007, CurrentID=0, Status='live', Symbol='LOC100128007', Aliases='', description='Spi-C transcription factor (Spi-1/PU.1 related) pseudogene', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6927417, end_position_on_the_genomic_accession=6928228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481899, CurrentID=0, Status='live', Symbol='RNU6-502P', Aliases='', description='RNA, U6 small nuclear 502, pseudogene', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7745, CurrentID=0, Status='live', Symbol='ZKSCAN8', Aliases='LD5-1, ZNF192, ZSCAN40', description='zinc finger with KRAB and SCAN domains 8', other_designations='zinc finger protein 192', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28136743, end_position_on_the_genomic_accession=28159472, orientation='plus', exon_count=10, OMIM=602240, no_hdr0=''),
  105374161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374161, CurrentID=0, Status='live', Symbol='LOC105374161', Aliases='', description='uncharacterized LOC105374161', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152040960, end_position_on_the_genomic_accession=152152197, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  400963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400963, CurrentID=0, Status='live', Symbol='RPS2P17', Aliases='RPS2_6_247', description='ribosomal protein S2 pseudogene 17', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84915763, end_position_on_the_genomic_accession=84916711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481902, CurrentID=0, Status='live', Symbol='RNU6-547P', Aliases='', description='RNA, U6 small nuclear 547, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481903, CurrentID=0, Status='live', Symbol='RNU6-566P', Aliases='', description='RNA, U6 small nuclear 566, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724506, CurrentID=0, Status='live', Symbol='LOC102724506', Aliases='', description='uncharacterized LOC102724506', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88087376, end_position_on_the_genomic_accession=88089730, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481904, CurrentID=0, Status='live', Symbol='RNU6-578P', Aliases='', description='RNA, U6 small nuclear 578, pseudogene', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400966, CurrentID=0, Status='live', Symbol='RGPD1', Aliases='RGP1, RGPD2, RanBP2L2', description='RANBP2-like and GRIP domain containing 1', other_designations='RANBP2-like and GRIP domain-containing protein 1/2|ran-binding protein 2-like 2/6|ran-binding protein 2-like 6|ranBP2-like 2/6|ranBP2-like 6|ranBP2L6', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86913573, end_position_on_the_genomic_accession=87013976, orientation='plus', exon_count=31, OMIM=612704, no_hdr0=''),
  106481905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481905, CurrentID=0, Status='live', Symbol='RNU6-597P', Aliases='', description='RNA, U6 small nuclear 597, pseudogene', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481906, CurrentID=0, Status='live', Symbol='RNU6-609P', Aliases='', description='RNA, U6 small nuclear 609, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374162, CurrentID=0, Status='live', Symbol='LOC105374162', Aliases='', description='uncharacterized LOC105374162', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152317578, end_position_on_the_genomic_accession=152320701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400968, CurrentID=0, Status='live', Symbol='LOC400968', Aliases='', description='uncharacterized LOC400968', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=19620671, end_position_on_the_genomic_accession=19621307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481908, CurrentID=0, Status='live', Symbol='RNU6-641P', Aliases='', description='RNA, U6 small nuclear 641, pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724511, CurrentID=0, Status='live', Symbol='LOC102724511', Aliases='', description='uncharacterized LOC102724511', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170160642, end_position_on_the_genomic_accession=170163027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420432, CurrentID=0, Status='live', Symbol='LOC100420432', Aliases='', description='ubiquitin specific peptidase 26 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84756196, end_position_on_the_genomic_accession=84757675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724899, CurrentID=0, Status='live', Symbol='LOC102724899', Aliases='', description='uncharacterized LOC102724899', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=162705376, end_position_on_the_genomic_accession=162746263, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481911, CurrentID=0, Status='live', Symbol='RNU6-681P', Aliases='', description='RNA, U6 small nuclear 681, pseudogene', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619427, CurrentID=0, Status='live', Symbol='LOC619427', Aliases='', description='uncharacterized LOC619427', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481913, CurrentID=0, Status='live', Symbol='RNU6-719P', Aliases='', description='RNA, U6 small nuclear 719, pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481914, CurrentID=0, Status='live', Symbol='RNU6-737P', Aliases='', description='RNA, U6 small nuclear 737, pseudogene', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481915, CurrentID=0, Status='live', Symbol='RNU6-755P', Aliases='', description='RNA, U6 small nuclear 755, pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533911, CurrentID=0, Status='live', Symbol='LOC100533911', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481918, CurrentID=0, Status='live', Symbol='RNU6-836P', Aliases='', description='RNA, U6 small nuclear 836, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481919, CurrentID=0, Status='live', Symbol='RNU6-850P', Aliases='', description='RNA, U6 small nuclear 850, pseudogene', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481922, CurrentID=0, Status='live', Symbol='RNU6-896P', Aliases='', description='RNA, U6 small nuclear 896, pseudogene', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400984, CurrentID=0, Status='live', Symbol='OR7E102P', Aliases='OR7E102', description='olfactory receptor family 7 subfamily E member 102 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 102 pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95546579, end_position_on_the_genomic_accession=95547574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481924, CurrentID=0, Status='live', Symbol='RNU6-936P', Aliases='', description='RNA, U6 small nuclear 936, pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400986, CurrentID=0, Status='live', Symbol='ANKRD36C', Aliases='', description='ankyrin repeat domain 36C', other_designations='protein immuno-reactive with anti-PTH polyclonal antibodies', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95827619, end_position_on_the_genomic_accession=95992320, orientation='minus', exon_count=91, OMIM='', no_hdr0=''),
  728667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728667, CurrentID=0, Status='live', Symbol='LOC728667', Aliases='', description='dehydrogenase/reductase (SDR family) member 2 pseudogene', other_designations='dehydrogenase/reductase (SDR family) member 4 like 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23827193, end_position_on_the_genomic_accession=23827374, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374166, CurrentID=0, Status='live', Symbol='LOC105374166', Aliases='', description='uncharacterized LOC105374166', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153320382, end_position_on_the_genomic_accession=153386506, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106481927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481927, CurrentID=0, Status='live', Symbol='RNU6-984P', Aliases='', description='RNA, U6 small nuclear 984, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481929, CurrentID=0, Status='live', Symbol='RNU6-1010P', Aliases='', description='RNA, U6 small nuclear 1010, pseudogene', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420436, CurrentID=0, Status='live', Symbol='STIP1P2', Aliases='', description='stress induced phosphoprotein 1 pseudogene 2', other_designations='stress-induced phosphoprotein 1 pseudogene 2|stress-induced-phosphoprotein 1 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227335556, end_position_on_the_genomic_accession=227336993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481930, CurrentID=0, Status='live', Symbol='RNU6-1021P', Aliases='', description='RNA, U6 small nuclear 1021, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724533, CurrentID=0, Status='live', Symbol='LOC102724533', Aliases='', description='uncharacterized LOC102724533', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533914, CurrentID=0, Status='live', Symbol='LOC100533914', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619448, CurrentID=0, Status='live', Symbol='RPS12P2', Aliases='RPS12_9_1071, bA40F6.3', description='ribosomal protein S12 pseudogene 2', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80215851, end_position_on_the_genomic_accession=80216119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647503, CurrentID=0, Status='live', Symbol='LOC647503', Aliases='', description='SNU13 homolog, small nuclear ribonucleoprotein (U4/U6.U5) pseudogene', other_designations='NHP2 non-histone chromosome protein 2-like 1 (S. cerevisiae) pseudogene', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132899986, end_position_on_the_genomic_accession=132900368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101234261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101234261, CurrentID=0, Status='live', Symbol='LSINCT5', Aliases='', description='long stress-induced non-coding transcript 5', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615764, no_hdr0=''),
  101234262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101234262, CurrentID=0, Status='live', Symbol='COX5BP6', Aliases='COX5BL6', description='cytochrome c oxidase subunit Vb pseudogene 6', other_designations='cytochrome c oxidase subunit Vb-like 6', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100623673, end_position_on_the_genomic_accession=100624515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481935, CurrentID=0, Status='live', Symbol='RNU6-1098P', Aliases='', description='RNA, U6 small nuclear 1098, pseudogene', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  400997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400997, CurrentID=0, Status='live', Symbol='LOC400997', Aliases='', description='uncharacterized LOC400997', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111098341, end_position_on_the_genomic_accession=111115588, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481937, CurrentID=0, Status='live', Symbol='RNU6-1132P', Aliases='', description='RNA, U6 small nuclear 1132, pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439915, CurrentID=0, Status='live', Symbol='KRTAP5-5', Aliases='KRTAP5-11, KRTAP5.5', description='keratin associated protein 5-5', other_designations='keratin associated protein 5-11|keratin-associated protein 5-11|keratin-associated protein 5.11|keratin-associated protein 5.5|ultrahigh sulfur keratin-associated protein 5.5', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1629803, end_position_on_the_genomic_accession=1630930, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  439916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439916, CurrentID=0, Status='live', Symbol='LINC00615', Aliases='C12orf37', description='long intergenic non-protein coding RNA 615', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90918023, end_position_on_the_genomic_accession=90948669, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481938, CurrentID=0, Status='live', Symbol='RNU6-1144P', Aliases='', description='RNA, U6 small nuclear 1144, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439921, CurrentID=0, Status='live', Symbol='MXRA7', Aliases='PS1TP1, TMAP1', description='matrix-remodelling associated 7', other_designations='HBV PreS1-transactivated protein 1|transmembrane anchor protein 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76673650, end_position_on_the_genomic_accession=76711005, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106481939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481939, CurrentID=0, Status='live', Symbol='RNU6-1159P', Aliases='', description='RNA, U6 small nuclear 1159, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369804, CurrentID=0, Status='live', Symbol='LOC105369804', Aliases='', description='uncharacterized LOC105369804', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64599078, end_position_on_the_genomic_accession=64610368, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  439927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439927, CurrentID=0, Status='live', Symbol='LINC01555', Aliases='C1orf180', description='long intergenic non-protein coding RNA 1555', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84628230, end_position_on_the_genomic_accession=84635020, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  439931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439931, CurrentID=0, Status='live', Symbol='THAP7-AS1', Aliases='', description='THAP7 antisense RNA 1', other_designations='THAP7 antisense RNA 1 (non-protein coding)', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21001922, end_position_on_the_genomic_accession=21010374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  439933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439933, CurrentID=0, Status='live', Symbol='LOC439933', Aliases='', description='uncharacterized LOC439933', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36244116, end_position_on_the_genomic_accession=36274215, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  439934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439934, CurrentID=0, Status='live', Symbol='LINC00575', Aliases='C4orf11', description='long intergenic non-protein coding RNA 575', other_designations='', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82613113, end_position_on_the_genomic_accession=82621437, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481941, CurrentID=0, Status='live', Symbol='RNU6-1183P', Aliases='', description='RNA, U6 small nuclear 1183, pseudogene', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439936, CurrentID=0, Status='live', Symbol='C5orf17', Aliases='', description='chromosome 5 open reading frame 17', other_designations='', map_location='5p14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23951773, end_position_on_the_genomic_accession=24201313, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  439938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439938, CurrentID=0, Status='live', Symbol='LOC439938', Aliases='', description='uncharacterized LOC439938', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439941, CurrentID=0, Status='live', Symbol='FER1L6-AS1', Aliases='C8orf54', description='FER1L6 antisense RNA 1', other_designations='FER1L6 antisense RNA 1 (non-protein coding)', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123984138, end_position_on_the_genomic_accession=124040782, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724545, CurrentID=0, Status='live', Symbol='LOC102724545', Aliases='', description='uncharacterized LOC102724545', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439944, CurrentID=0, Status='live', Symbol='FLJ32742', Aliases='', description='uncharacterized locus FLJ32742', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439945, CurrentID=0, Status='live', Symbol='IL9RP2', Aliases='', description='interleukin 9 receptor pseudogene 2', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79976, end_position_on_the_genomic_accession=86243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481943, CurrentID=0, Status='live', Symbol='RNU6-1212P', Aliases='', description='RNA, U6 small nuclear 1212, pseudogene', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439949, CurrentID=0, Status='live', Symbol='PRKCQ-AS1', Aliases='', description='PRKCQ antisense RNA 1', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6580425, end_position_on_the_genomic_accession=6585361, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  439950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439950, CurrentID=0, Status='live', Symbol='CELF2-AS2', Aliases='', description='CELF2 antisense RNA 2', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11071541, end_position_on_the_genomic_accession=11105488, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  439951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439951, CurrentID=0, Status='live', Symbol='LOC439951', Aliases='', description='uncharacterized LOC439951', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11611307, end_position_on_the_genomic_accession=11612227, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481944, CurrentID=0, Status='live', Symbol='RNU6-1230P', Aliases='', description='RNA, U6 small nuclear 1230, pseudogene', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439954, CurrentID=0, Status='live', Symbol='RPL7P37', Aliases='RPL7_17_1035', description='ribosomal protein L7 pseudogene 37', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35697274, end_position_on_the_genomic_accession=35698105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  439957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439957, CurrentID=0, Status='live', Symbol='IGKV1ORY-1', Aliases='IGKV1/ORY-1', description='immunoglobulin kappa variable 1/ORY-1 (pseudogene)', other_designations='immunoglobulin kappa variable 1/ORY-1 pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=13629413, end_position_on_the_genomic_accession=13631274, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481945, CurrentID=0, Status='live', Symbol='RNU6-1244P', Aliases='', description='RNA, U6 small nuclear 1244, pseudogene', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401007, CurrentID=0, Status='live', Symbol='NF1P8', Aliases='NF1L2', description='neurofibromin 1 pseudogene 8', other_designations='neurofibromin 1-like 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131188448, end_position_on_the_genomic_accession=131198005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  439965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439965, CurrentID=0, Status='live', Symbol='FAM35DP', Aliases='FAM35B2', description='family with sequence similarity 35 member D, pseudogene', other_designations='family with sequence similarity 35, member B2 (pseudogene)|family with sequence similarity 35, member D, pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47689700, end_position_on_the_genomic_accession=47731218, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  138864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138864, CurrentID=0, Status='live', Symbol='ATP5HP3', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit D pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d, pseudogene 3', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14068878, end_position_on_the_genomic_accession=14069480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100316847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100316847, CurrentID=0, Status='live', Symbol='GLM8', Aliases='', description='Glioma susceptibility 8', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613033, no_hdr0=''),
  102724552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724552, CurrentID=0, Status='live', Symbol='LOC102724552', Aliases='', description='uncharacterized LOC102724552', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8424645, end_position_on_the_genomic_accession=8435016, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  439990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439990, CurrentID=0, Status='live', Symbol='LINC00857', Aliases='', description='long intergenic non-protein coding RNA 857', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80207710, end_position_on_the_genomic_accession=80219657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724553, CurrentID=0, Status='live', Symbol='LOC102724553', Aliases='', description='uncharacterized LOC102724553', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47960311, end_position_on_the_genomic_accession=48046612, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  439992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439992, CurrentID=0, Status='live', Symbol='RPS3AP5', Aliases='RPS3A_15_1074, bA486O22.3', description='ribosomal protein S3A pseudogene 5', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84560419, end_position_on_the_genomic_accession=84561306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132382, CurrentID=0, Status='live', Symbol='RPS14P4', Aliases='RPS14_2_264', description='ribosomal protein S14 pseudogene 4', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111295556, end_position_on_the_genomic_accession=111296081, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  439996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439996, CurrentID=0, Status='live', Symbol='IFIT1B', Aliases='IFIT1L, bA149I23.6', description='interferon induced protein with tetratricopeptide repeats 1B', other_designations='interferon-induced protein with tetratricopeptide repeats 1-like protein', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89378056, end_position_on_the_genomic_accession=89385205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374171, CurrentID=0, Status='live', Symbol='LOC105374171', Aliases='', description='uncharacterized LOC105374171', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154898740, end_position_on_the_genomic_accession=154970232, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100316867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100316867, CurrentID=0, Status='live', Symbol='ARF4P5', Aliases='', description='ADP ribosylation factor 4 pseudogene 5', other_designations='ADP-ribosylation factor 4 pseudogene 5', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83572718, end_position_on_the_genomic_accession=83573613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100316868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100316868, CurrentID=0, Status='live', Symbol='HOTTIP', Aliases='HOXA-AS6, HOXA13-AS1, NCRNA00213', description='HOXA distal transcript antisense RNA', other_designations='HOXA cluster antisense RNA 6 (non-protein coding)|HOXA distal transcript antisense RNA (non-protein coding)|HOXA13 antisense RNA 1 (non-protein coding)|HoxA transcript at the distal tip', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27200421, end_position_on_the_genomic_accession=27207259, orientation='plus', exon_count=3, OMIM=614060, no_hdr0=''),
  401014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401014, CurrentID=0, Status='live', Symbol='TEX41', Aliases='DKFZp686O1327, LINC00953', description='testis expressed 41 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 953', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144667967, end_position_on_the_genomic_accession=145076729, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481954, CurrentID=0, Status='live', Symbol='RNU1-55P', Aliases='', description='RNA, U1 small nuclear 55, pseudogene', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440014, CurrentID=0, Status='live', Symbol='DUX4L12', Aliases='', description='double homeobox 4 like 12', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133749949, end_position_on_the_genomic_accession=133751227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724557, CurrentID=0, Status='live', Symbol='LOC102724557', Aliases='', description='uncharacterized LOC102724557', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79801384, end_position_on_the_genomic_accession=79846392, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  440017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440017, CurrentID=0, Status='live', Symbol='DUX4L15', Aliases='', description='double homeobox 4 like 15', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133759857, end_position_on_the_genomic_accession=133761135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481955, CurrentID=0, Status='live', Symbol='RNU1-72P', Aliases='', description='RNA, U1 small nuclear 72, pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440021, CurrentID=0, Status='live', Symbol='KRTAP5-2', Aliases='KRTAP5-8, KRTAP5.2', description='keratin associated protein 5-2', other_designations='keratin-associated protein 5-8|keratin-associated protein 5.2|keratin-associated protein 5.8|ultrahigh sulfur keratin-associated protein 5.2', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1597177, end_position_on_the_genomic_accession=1598294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23381, CurrentID=0, Status='live', Symbol='SMG5', Aliases='EST1B, LPTS-RP1, LPTSRP1, SMG-5', description='SMG5 nonsense mediated mRNA decay factor', other_designations='EST1 telomerase component homolog B|EST1-like protein B|Est1p-like protein B|LPTS interacting protein|LPTS-interacting protein|RP11-54H19.7|ever shorter telomeres 1B|hSMG-5|smg-5 homolog, nonsense mediated mRNA decay factor', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156249224, end_position_on_the_genomic_accession=156282885, orientation='minus', exon_count=24, OMIM=610962, no_hdr0=''),
  440023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440023, CurrentID=0, Status='live', Symbol='KRTAP5-6', Aliases='KRTAP5.6', description='keratin associated protein 5-6', other_designations='keratin-associated protein 5.6|ultrahigh sulfur keratin-associated protein 5.6', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1697195, end_position_on_the_genomic_accession=1697755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440026, CurrentID=0, Status='live', Symbol='TMEM41B', Aliases='', description='transmembrane protein 41B', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9280654, end_position_on_the_genomic_accession=9314768, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  440027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440027, CurrentID=0, Status='live', Symbol='RPL23AP65', Aliases='RPL23A_27_1119', description='ribosomal protein L23a pseudogene 65', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9606134, end_position_on_the_genomic_accession=9606682, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440028, CurrentID=0, Status='live', Symbol='LOC440028', Aliases='', description='uncharacterized LOC440028', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9754770, end_position_on_the_genomic_accession=9759533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401018, CurrentID=0, Status='live', Symbol='CTAGE14P', Aliases='', description='CTAGE family member 14, pseudogene', other_designations='CTAGE family, member 14, pseudogene|CTAGE family, member 9 pseudogene', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=167713497, end_position_on_the_genomic_accession=167716065, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481957, CurrentID=0, Status='live', Symbol='RNU1-108P', Aliases='', description='RNA, U1 small nuclear 108, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440034, CurrentID=0, Status='live', Symbol='DKFZp686K1684', Aliases='', description='uncharacterized LOC440034', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31816566, end_position_on_the_genomic_accession=31887041, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401019, CurrentID=0, Status='live', Symbol='RPS15P4', Aliases='RPS15_2_291', description='ribosomal protein S15 pseudogene 4', other_designations='hCG1783417', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171517239, end_position_on_the_genomic_accession=171517737, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422033, CurrentID=0, Status='live', Symbol='LOC100422033', Aliases='', description='olfactory receptor, family 4, subfamily Q, member 3 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100316904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100316904, CurrentID=0, Status='live', Symbol='SAP25', Aliases='', description='Sin3A associated protein 25kDa', other_designations='25 kDa Sin3-associated polypeptide|Sin3 corepressor complex subunit SAP25|Sin3A-associated protein, 25kDa|sin3-associated polypeptide, 25kDa|sin3A-binding protein 25', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100572226, end_position_on_the_genomic_accession=100573864, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  440041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440041, CurrentID=0, Status='live', Symbol='TRIM51HP', Aliases='', description='tripartite motif-containing 51H, pseudogene', other_designations='SPRY domain containing 5 pseudogene|Tripartite motif-containing protein ENSP00000311270', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55293703, end_position_on_the_genomic_accession=55298232, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440043, CurrentID=0, Status='live', Symbol='PGAM1P8', Aliases='', description='phosphoglycerate mutase 1 pseudogene 8', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65174128, end_position_on_the_genomic_accession=65175491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440044, CurrentID=0, Status='live', Symbol='SLC22A20', Aliases='Oat6', description='solute carrier family 22 member 20', other_designations='organic anion transporter 6|solute carrier family 22, member 20', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65213840, end_position_on_the_genomic_accession=65242757, orientation='plus', exon_count=11, OMIM=611696, no_hdr0=''),
  100130485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130485, CurrentID=0, Status='live', Symbol='LOC100130485', Aliases='', description='FK506 binding protein 3, 25kDa pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236285474, end_position_on_the_genomic_accession=236287003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440047, CurrentID=0, Status='live', Symbol='LOC440047', Aliases='', description='uncharacterized LOC440047', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440049, CurrentID=0, Status='live', Symbol='FLJ42258', Aliases='', description='FLJ42258 protein', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69654681, end_position_on_the_genomic_accession=69659401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440050, CurrentID=0, Status='live', Symbol='KRTAP5-7', Aliases='KRTAP5-3, KRTAP5.7', description='keratin associated protein 5-7', other_designations='keratin-associated protein 5-3|keratin-associated protein 5.3|keratin-associated protein 5.7|ultrahigh sulfur keratin-associated protein 5.7', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71527267, end_position_on_the_genomic_accession=71528164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440051, CurrentID=0, Status='live', Symbol='KRTAP5-11', Aliases='KRTAP5-5, KRTAP5-6, KRTAP5.11', description='keratin associated protein 5-11', other_designations='keratin associated protein 5-5|keratin-associated protein 5.11|ultrahigh sulfur keratin-associated protein 5.11', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71581855, end_position_on_the_genomic_accession=71582875, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440055, CurrentID=0, Status='live', Symbol='RPS12P22', Aliases='RPS12_10_1156', description='ribosomal protein S12 pseudogene 22', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74745671, end_position_on_the_genomic_accession=74746194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440059, CurrentID=0, Status='live', Symbol='LOC440059', Aliases='', description='cell adhesion associated, oncogene regulated pseudogene', other_designations='Cdon homolog pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89897228, end_position_on_the_genomic_accession=89900449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440061, CurrentID=0, Status='live', Symbol='TRIM64EP', Aliases='', description='tripartite motif containing 64E, pseudogene', other_designations='tripartite motif containing 64B pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90057419, end_position_on_the_genomic_accession=90062997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440063, CurrentID=0, Status='live', Symbol='LOC440063', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100666426, end_position_on_the_genomic_accession=100667159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  7808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7808, CurrentID=0, Status='live', Symbol='AIR', Aliases='', description='Acute insulin response (2)', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601676, no_hdr0=''),
  440067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440067, CurrentID=0, Status='live', Symbol='CASP1P2', Aliases='', description='caspase 1 pseudogene 2', other_designations='caspase 1, apoptosis-related cysteine peptidase pseudogene 2|caspase-1-like', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105062602, end_position_on_the_genomic_accession=105071556, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  440068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440068, CurrentID=0, Status='live', Symbol='CARD17', Aliases='INCA', description='caspase recruitment domain family member 17', other_designations='Inhibitory CARD|caspase recruitment domain family, member 17|caspase-1 inhibitor INCA|inhibitory caspase recruitment domain (CARD) protein|inhibitory caspase recruitment domain protein', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105092469, end_position_on_the_genomic_accession=105101465, orientation='minus', exon_count=4, OMIM=609490, no_hdr0=''),
  102724566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724566, CurrentID=0, Status='live', Symbol='LOC102724566', Aliases='', description='uncharacterized LOC102724566', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89041027, end_position_on_the_genomic_accession=89082888, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  138881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138881, CurrentID=0, Status='live', Symbol='OR1L8', Aliases='OR9-24', description='olfactory receptor family 1 subfamily L member 8', other_designations='olfactory receptor OR9-24|olfactory receptor, family 1, subfamily L, member 8', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122546236, end_position_on_the_genomic_accession=122693914, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  440072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440072, CurrentID=0, Status='live', Symbol='LINC00167', Aliases='C11orf37, NCRNA00167', description='long intergenic non-protein coding RNA 167', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130002624, end_position_on_the_genomic_accession=130005486, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440073, CurrentID=0, Status='live', Symbol='IQSEC3', Aliases='', description='IQ motif and Sec7 domain 3', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66796, end_position_on_the_genomic_accession=178460, orientation='plus', exon_count=17, OMIM=612118, no_hdr0=''),
  440077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440077, CurrentID=0, Status='live', Symbol='ZNF705A', Aliases='', description='zinc finger protein 705A', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8157012, end_position_on_the_genomic_accession=8180046, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  440078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440078, CurrentID=0, Status='live', Symbol='FAM66C', Aliases='', description='family with sequence similarity 66 member C', other_designations='family with sequence similarity 66, member C', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8180209, end_position_on_the_genomic_accession=8201000, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106481965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481965, CurrentID=0, Status='live', Symbol='COX6CP16', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 16', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440081, CurrentID=0, Status='live', Symbol='DDX12P', Aliases='CHLR2, DDX12', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 12, pseudogene', other_designations='CHL1-like helicase homolog 2|CHL1-related protein 2|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 12 (CHL1-like helicase homolog, S. cerevisiae)|DEAD/H box protein 12|DEAD/H box-12|hCHLR2|probable ATP-dependent RNA helicase DDX12', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9417691, end_position_on_the_genomic_accession=9448172, orientation='minus', exon_count=27, OMIM=601151, no_hdr0=''),
  23383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23383, CurrentID=0, Status='live', Symbol='MAU2', Aliases='KIAA0892L, SCC4, mau-2, MAU2', description='MAU2, sister chromatid cohesion factor', other_designations='MAU2 sister chromatid cohesion factor|cohesin loading complex subunit SCC4 homolog|protein MAU-2|sister chromatid cohesion 4', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19320817, end_position_on_the_genomic_accession=19358754, orientation='plus', exon_count=20, OMIM=614560, no_hdr0=''),
  401027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401027, CurrentID=0, Status='live', Symbol='C2orf66', Aliases='IIDS6411, UNQ6411', description='chromosome 2 open reading frame 66', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196804415, end_position_on_the_genomic_accession=196810276, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440084, CurrentID=0, Status='live', Symbol='LOC440084', Aliases='', description='hCG1655019', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=11474411, end_position_on_the_genomic_accession=11480515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440086, CurrentID=0, Status='live', Symbol='RPS6P1', Aliases='RPS6_9_1210', description='ribosomal protein S6 pseudogene 1', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12850957, end_position_on_the_genomic_accession=12851751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440087, CurrentID=0, Status='live', Symbol='SMCO3', Aliases='C12orf69', description='single-pass membrane protein with coiled-coil domains 3', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14803723, end_position_on_the_genomic_accession=14814182, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481967, CurrentID=0, Status='live', Symbol='NOP56P3', Aliases='', description='NOP56 ribonucleoprotein pseudogene 3', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440093, CurrentID=0, Status='live', Symbol='H3F3C', Aliases='H3.5', description='H3 histone, family 3C', other_designations='histone H3.5|histone variant H3.5', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31791185, end_position_on_the_genomic_accession=31792241, orientation='minus', exon_count=1, OMIM=616134, no_hdr0=''),
  440097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440097, CurrentID=0, Status='live', Symbol='DBX2', Aliases='', description='developing brain homeobox 2', other_designations='developing brain homeobox protein 2', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45014756, end_position_on_the_genomic_accession=45051099, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440101, CurrentID=0, Status='live', Symbol='FLJ12825', Aliases='', description='uncharacterized LOC440101', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54058254, end_position_on_the_genomic_accession=54122234, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440104, CurrentID=0, Status='live', Symbol='TMEM198B', Aliases='', description='transmembrane protein 198B, pseudogene', other_designations='1110012D08Rik pseudogene|transmembrane protein 198, pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55829608, end_position_on_the_genomic_accession=55836246, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  440107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440107, CurrentID=0, Status='live', Symbol='PLEKHG7', Aliases='', description='pleckstrin homology and RhoGEF domain containing G7', other_designations='PH domain-containing family G member 7|pleckstrin homology domain containing, family G (with RhoGef domain) member 7', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92736489, end_position_on_the_genomic_accession=92772745, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  106481970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481970, CurrentID=0, Status='live', Symbol='DUX4L29', Aliases='', description='double homeobox 4 like 29', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481971, CurrentID=0, Status='live', Symbol='RPL23AP94', Aliases='', description='ribosomal protein L23a pseudogene 94', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440117, CurrentID=0, Status='live', Symbol='LOC440117', Aliases='', description='uncharacterized LOC440117', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126869494, end_position_on_the_genomic_accession=126874690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440119, CurrentID=0, Status='live', Symbol='FZD10-AS1', Aliases='', description='FZD10 antisense RNA 1 (head to head)', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130151593, end_position_on_the_genomic_accession=130161678, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440125, CurrentID=0, Status='live', Symbol='HNRNPA1P30', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 30', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20947946, end_position_on_the_genomic_accession=20949659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440129, CurrentID=0, Status='live', Symbol='FLJ26086', Aliases='', description='uncharacterized LOC440129', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440131, CurrentID=0, Status='live', Symbol='LINC00544', Aliases='', description='long intergenic non-protein coding RNA 544', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29936531, end_position_on_the_genomic_accession=29950488, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  440132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440132, CurrentID=0, Status='live', Symbol='LINC00545', Aliases='', description='long intergenic non-protein coding RNA 545', other_designations='hCG1815504', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30882562, end_position_on_the_genomic_accession=30883395, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481974, CurrentID=0, Status='live', Symbol='RN7SKP47', Aliases='', description='RNA, 7SK small nuclear pseudogene 47', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440137, CurrentID=0, Status='live', Symbol='FLJ31945', Aliases='', description='uncharacterized LOC440137', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50120837, end_position_on_the_genomic_accession=50128457, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  440138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440138, CurrentID=0, Status='live', Symbol='ALG11', Aliases='CDG1P, GT8', description='ALG11, alpha-1,2-mannosyltransferase', other_designations='GDP-Man:Man(3)GlcNAc(2)-PP-dolichol alpha-1,2-mannosyltransferase|asparagine-linked glycosylation 11, alpha-1,2-mannosyltransferase homolog|asparagine-linked glycosylation protein 11 homolog|glycolipid 2-alpha-mannosyltransferase', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52012387, end_position_on_the_genomic_accession=52029648, orientation='plus', exon_count=4, OMIM=613666, no_hdr0=''),
  440142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440142, CurrentID=0, Status='live', Symbol='TARDBPP2', Aliases='', description='TAR DNA binding protein pseudogene 2', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60269304, end_position_on_the_genomic_accession=60276137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440145, CurrentID=0, Status='live', Symbol='MZT1', Aliases='C13orf37, MOZART1', description='mitotic spindle organizing protein 1', other_designations='mitotic-spindle organizing protein associated with a ring of gamma-tubulin 1', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72708357, end_position_on_the_genomic_accession=72727800, orientation='minus', exon_count=4, OMIM=613448, no_hdr0=''),
  440149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440149, CurrentID=0, Status='live', Symbol='LOC440149', Aliases='', description='uncharacterized LOC440149', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481977, CurrentID=0, Status='live', Symbol='RNU6-139P', Aliases='', description='RNA, U6 small nuclear 139, pseudogene', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440153, CurrentID=0, Status='live', Symbol='OR11H12', Aliases='', description='olfactory receptor family 11 subfamily H member 12', other_designations='olfactory receptor, family 11, subfamily H, member 12', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18601117, end_position_on_the_genomic_accession=18602097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106481978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481978, CurrentID=0, Status='live', Symbol='RNU6-534P', Aliases='', description='RNA, U6 small nuclear 534, pseudogene', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440159, CurrentID=0, Status='live', Symbol='OR11H13P', Aliases='ORNP-NP1', description='olfactory receptor family 11 subfamily H member 13 pseudogene', other_designations='olfactory receptor ORNP-NP1 pseudogene|olfactory receptor, family 11, subfamily H, member 13 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19180291, end_position_on_the_genomic_accession=19181271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401040, CurrentID=0, Status='live', Symbol='LOC401040', Aliases='', description='uncharacterized LOC401040', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239578396, end_position_on_the_genomic_accession=239586088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440163, CurrentID=0, Status='live', Symbol='RNASE13', Aliases='HEL-S-86p, RAL1', description='ribonuclease, RNase A family, 13 (non-active)', other_designations='epididymis secretory sperm binding protein Li 86p|ribonuclease 13|ribonuclease A L1|ribonuclease-like protein 13', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21032820, end_position_on_the_genomic_accession=21034785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440173, CurrentID=0, Status='live', Symbol='LOC440173', Aliases='', description='uncharacterized LOC440173', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87008451, end_position_on_the_genomic_accession=87042126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106481981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481981, CurrentID=0, Status='live', Symbol='ZNF519P4', Aliases='ZNF141', description='zinc finger protein 519 pseudogene 4', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440176, CurrentID=0, Status='live', Symbol='RPL12P6', Aliases='RPL12_12_1352', description='ribosomal protein L12 pseudogene 6', other_designations='', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34740291, end_position_on_the_genomic_accession=34740904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724584, CurrentID=0, Status='live', Symbol='LOC102724584', Aliases='', description='uncharacterized LOC102724584', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34836589, end_position_on_the_genomic_accession=34848498, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440180, CurrentID=0, Status='live', Symbol='LOC440180', Aliases='', description='zinc finger, CCHC domain containing 7 pseudogene', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57040848, end_position_on_the_genomic_accession=57042701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440181, CurrentID=0, Status='live', Symbol='LOC440181', Aliases='', description='transmembrane protein 98 pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58827682, end_position_on_the_genomic_accession=58828355, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189127, CurrentID=0, Status='live', Symbol='TRN-GTT3-2', Aliases='TRNAN15', description='transfer RNA-Asn (GTT) 3-2', other_designations='tRNA-Asn-GTT-3-2|transfer RNA asparagine 15 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144567515, end_position_on_the_genomic_accession=144567588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289511, CurrentID=0, Status='live', Symbol='LOC100289511', Aliases='', description='uncharacterized LOC100289511', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69766283, end_position_on_the_genomic_accession=69767713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440184, CurrentID=0, Status='live', Symbol='LINC00238', Aliases='C14orf53, NCRNA00238', description='long intergenic non-protein coding RNA 238', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66486371, end_position_on_the_genomic_accession=66498553, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102724586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724586, CurrentID=0, Status='live', Symbol='LOC102724586', Aliases='', description='uncharacterized LOC102724586', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110335270, end_position_on_the_genomic_accession=110353143, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  440193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440193, CurrentID=0, Status='live', Symbol='CCDC88C', Aliases='DAPLE, HKRP2, KIAA1509, SCA40', description='coiled-coil domain containing 88C', other_designations='Dvl-associating protein with a high frequency of leucine residues|hook-related protein 2|spinocerebellar ataxia 40', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91271323, end_position_on_the_genomic_accession=91417844, orientation='minus', exon_count=34, OMIM=611204, no_hdr0=''),
  102724587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724587, CurrentID=0, Status='live', Symbol='LOC102724587', Aliases='', description='uncharacterized LOC102724587', other_designations='CTD-2647E9.3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49877104, end_position_on_the_genomic_accession=49883527, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372809, CurrentID=0, Status='live', Symbol='LOC105372809', Aliases='', description='uncharacterized LOC105372809', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41552804, end_position_on_the_genomic_accession=41593324, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  440200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440200, CurrentID=0, Status='live', Symbol='FLJ41170', Aliases='', description='uncharacterized LOC440200', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=100541398, end_position_on_the_genomic_accession=100543122, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724589, CurrentID=0, Status='live', Symbol='LOC102724589', Aliases='', description='uncharacterized LOC102724589', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114994657, end_position_on_the_genomic_accession=114996593, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481440, CurrentID=0, Status='live', Symbol='RNU6-821P', Aliases='', description='RNA, U6 small nuclear 821, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128136, CurrentID=0, Status='live', Symbol='LOC128136', Aliases='', description='3-hydroxy-3-methylglutaryl-Coenzyme A synthase 1 (soluble) pseudogene', other_designations='3-hydroxy-3-methylglutaryl-CoA synthase 1 (soluble) pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239972787, end_position_on_the_genomic_accession=239974244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724591, CurrentID=0, Status='live', Symbol='LOC102724591', Aliases='', description='uncharacterized LOC102724591', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17585356, end_position_on_the_genomic_accession=17600271, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100874358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874358, CurrentID=0, Status='live', Symbol='RNY1P9', Aliases='KLHL22-IT1, RNYP1', description='RNA, Ro-associated Y1 pseudogene 9', other_designations='KLHL22 intronic transcript 1 (non-protein coding)|RNA, Ro-associated Y, pseudogene 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20475194, end_position_on_the_genomic_accession=20475311, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440224, CurrentID=0, Status='live', Symbol='CXADRP3', Aliases='', description='coxsackie virus and adenovirus receptor pseudogene 3', other_designations='CTD-2594A10.3', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14477952, end_position_on_the_genomic_accession=14498706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440225, CurrentID=0, Status='live', Symbol='NF1P2', Aliases='NF1HHS', description='neurofibromin 1 pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21927657, end_position_on_the_genomic_accession=21940278, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724593, CurrentID=0, Status='live', Symbol='LOC102724593', Aliases='', description='uncharacterized LOC102724593', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47743699, end_position_on_the_genomic_accession=47759023, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440243, CurrentID=0, Status='live', Symbol='GOLGA6L22', Aliases='', description='golgin A6 family-like 22', other_designations='putative golgin subfamily A member 6-like', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22458903, end_position_on_the_genomic_accession=22468026, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  440248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440248, CurrentID=0, Status='live', Symbol='HERC2P9', Aliases='', description='hect domain and RLD 2 pseudogene 9', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28654442, end_position_on_the_genomic_accession=28685264, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  440253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440253, CurrentID=0, Status='live', Symbol='WHAMMP2', Aliases='WHAMML2, WHDC1L2', description='WAS protein homolog associated with actin, golgi membranes and microtubules pseudogene 2', other_designations='WAS protein homolog associated with actin, golgi membranes and microtubules-like 2 (pseudogene)|WAS protein homology region 2 domain containing 1-like 2 (pseudogene)', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28737583, end_position_on_the_genomic_accession=28758362, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100420449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420449, CurrentID=0, Status='live', Symbol='LOC100420449', Aliases='', description='tripartite motif containing 48 pseudogene', other_designations='tripartite motif-containing protein LOC642612-like pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48907103, end_position_on_the_genomic_accession=48908791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285260, CurrentID=0, Status='live', Symbol='RPL31P18', Aliases='RPL31_7_358', description='ribosomal protein L31 pseudogene 18', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32507931, end_position_on_the_genomic_accession=32508362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724598, CurrentID=0, Status='live', Symbol='LOC102724598', Aliases='', description='uncharacterized LOC102724598', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57490855, end_position_on_the_genomic_accession=57495841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100533927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533927, CurrentID=0, Status='live', Symbol='LOC100533927', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440270, CurrentID=0, Status='live', Symbol='GOLGA8B', Aliases='GOLGA5', description='golgin A8 family member B', other_designations='golgi autoantigen, golgin subfamily a, 8B|golgin A8 family, member B|golgin-67', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34525276, end_position_on_the_genomic_accession=34583651, orientation='minus', exon_count=26, OMIM=609619, no_hdr0=''),
  440272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440272, CurrentID=0, Status='live', Symbol='ANP32F', Aliases='', description='acidic (leucine-rich) nuclear phosphoprotein 32 family, member F', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440275, CurrentID=0, Status='live', Symbol='EIF2AK4', Aliases='GCN2, PVOD2', description='eukaryotic translation initiation factor 2 alpha kinase 4', other_designations='GCN2 eIF2alpha kinase|GCN2-like protein|general control nonderepressible 2', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39934124, end_position_on_the_genomic_accession=40035596, orientation='plus', exon_count=39, OMIM=609280, no_hdr0=''),
  51534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51534, CurrentID=0, Status='live', Symbol='VTA1', Aliases='C6orf55, DRG-1, DRG1, HSPC228, LIP5, My012, SBP1', description='vesicle (multivesicular body) trafficking 1', other_designations='LYST-interacting protein 5|SKD1-binding protein 1|Vps20-associated 1 homolog|dopamine-responsive gene 1 protein|homolog of mouse SKD1-binding protein 1', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142147162, end_position_on_the_genomic_accession=142220948, orientation='plus', exon_count=8, OMIM=610902, no_hdr0=''),
  440278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440278, CurrentID=0, Status='live', Symbol='CATSPER2P1', Aliases='', description='cation channel, sperm associated 2 pseudogene 1', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43735934, end_position_on_the_genomic_accession=43746298, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440279, CurrentID=0, Status='live', Symbol='UNC13C', Aliases='', description='unc-13 homolog C (C. elegans)', other_designations='Munc13-3|unc-13-like 3', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53978203, end_position_on_the_genomic_accession=54628608, orientation='plus', exon_count=33, OMIM=614568, no_hdr0=''),
  138916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138916, CurrentID=0, Status='live', Symbol='SYNGR2P2', Aliases='', description='synaptogyrin 2 pseudogene 2', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78149742, end_position_on_the_genomic_accession=78151210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378545, CurrentID=0, Status='live', Symbol='LOC105378545', Aliases='', description='uncharacterized LOC105378545', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125536833, end_position_on_the_genomic_accession=125539786, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440288, CurrentID=0, Status='live', Symbol='UBL7-AS1', Aliases='', description='UBL7 antisense RNA 1 (head to head)', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74461265, end_position_on_the_genomic_accession=74481292, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  440292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440292, CurrentID=0, Status='live', Symbol='LOC440292', Aliases='', description='COMM domain-containing protein 4-like', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77941393, end_position_on_the_genomic_accession=77944561, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440293, CurrentID=0, Status='live', Symbol='DNM1DN11-7', Aliases='DNM1DN11.7', description='DNM1DN11.7 duplicon', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440295, CurrentID=0, Status='live', Symbol='GOLGA6L9', Aliases='GOLGA6L20', description='golgin A6 family-like 9', other_designations='golgi autoantigen, golgin subfamily a, 6-like 9|golgin A6 family-like 20|putative golgin subfamily A member 6-like protein 20', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82417678, end_position_on_the_genomic_accession=82439387, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  440299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440299, CurrentID=0, Status='live', Symbol='DNM1P41', Aliases='DNM1DN11-6, DNM1DN11.6, DNM1DN11@', description='dynamin 1 pseudogene 41', other_designations='DNM1 pseudogene 41|DNM1DN11.6 duplicon', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84166547, end_position_on_the_genomic_accession=84170984, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440300, CurrentID=0, Status='live', Symbol='LOC440300', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84191848, end_position_on_the_genomic_accession=84209294, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100189131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189131, CurrentID=0, Status='live', Symbol='TRI-TAT2-3', Aliases='TRNAI6', description='transfer RNA-Ile (TAT) 2-3', other_designations='transfer RNA isoleucine 6 (anticodon UAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27631421, end_position_on_the_genomic_accession=27631514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440303, CurrentID=0, Status='live', Symbol='DNM1P44', Aliases='DNM1DN11-13, DNM1DN11.13, DNM1DN11@', description='dynamin 1 pseudogene 44', other_designations='DNM1 pseudogene 44|DNM1DN11.13 duplicon', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84426152, end_position_on_the_genomic_accession=84427645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440307, CurrentID=0, Status='live', Symbol='TTLL13P', Aliases='TTLL13', description='tubulin tyrosine ligase like 13, pseudogene', other_designations='tubulin polyglutamylase TTLL13|tubulin tyrosine ligase-like family member 13, pseudogene|tubulin tyrosine ligase-like family, member 13|tubulin--tyrosine ligase-like protein 13', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90249532, end_position_on_the_genomic_accession=90259088, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  440311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440311, CurrentID=0, Status='live', Symbol='LOC440311', Aliases='', description='glioma tumor suppressor candidate region gene 2 pseudogene', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94855363, end_position_on_the_genomic_accession=94857066, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440313, CurrentID=0, Status='live', Symbol='LOC440313', Aliases='', description='protein enabled homolog', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401067, CurrentID=0, Status='live', Symbol='IQCF3', Aliases='', description='IQ motif containing F3', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51826883, end_position_on_the_genomic_accession=51830858, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  401068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401068, CurrentID=0, Status='live', Symbol='LOC401068', Aliases='', description='uncharacterized LOC401068', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=53080108, end_position_on_the_genomic_accession=53082138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440330, CurrentID=0, Status='live', Symbol='LOC440330', Aliases='', description='uncharacterized LOC440330', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189132, CurrentID=0, Status='live', Symbol='TRI-GAT1-1', Aliases='TRNAI7', description='transfer RNA-Ile (GAT) 1-1', other_designations='tRNA-Ile-GAT-1-1|transfer RNA isoleucine 7 (anticodon GAU)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3838377, end_position_on_the_genomic_accession=3838450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440335, CurrentID=0, Status='live', Symbol='SMIM22', Aliases='', description='small integral membrane protein 22', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4788391, end_position_on_the_genomic_accession=4796491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  440345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440345, CurrentID=0, Status='live', Symbol='NPIPB4', Aliases='61E3.4', description='nuclear pore complex interacting protein family member B4', other_designations='nuclear pore complex interacting protein family, member B4|nuclear pore complex interacting protein-like', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21834568, end_position_on_the_genomic_accession=21880845, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  440346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440346, CurrentID=0, Status='live', Symbol='LOC440346', Aliases='', description='uncharacterized LOC440346', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22374323, end_position_on_the_genomic_accession=22376273, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440348, CurrentID=0, Status='live', Symbol='NPIPB15', Aliases='A-761H5.4, NPIPL2', description='nuclear pore complex interacting protein family member B15', other_designations='nuclear pore complex interacting protein family, member B15|nuclear pore complex interacting protein-like 2', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74375001, end_position_on_the_genomic_accession=74392115, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  440350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440350, CurrentID=0, Status='live', Symbol='NPIPB7', Aliases='A-575C2.4, A-761H5.5, NPIPL1', description='nuclear pore complex interacting protein family member B7', other_designations='nuclear pore complex interacting protein family, member B7|nuclear pore complex interacting protein-like 1|nuclear pore complex-interacting protein-like 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28456820, end_position_on_the_genomic_accession=28471296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440352, CurrentID=0, Status='live', Symbol='SNX29P2', Aliases='RUNDC2C', description='sorting nexin 29 pseudogene 2', other_designations='RUN domain containing 2C', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29302287, end_position_on_the_genomic_accession=29365059, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  440353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440353, CurrentID=0, Status='live', Symbol='LOC440353', Aliases='', description='nuclear pore complex interacting protein family member A1 pseudogene', other_designations='nuclear pore complex interacting protein family, member A1 pseudogene|nuclear pore complex interacting protein pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29485314, end_position_on_the_genomic_accession=29500205, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440354, CurrentID=0, Status='live', Symbol='SMG1P2', Aliases='', description='SMG1 pseudogene 2', other_designations='smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29527608, end_position_on_the_genomic_accession=29566219, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  440356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440356, CurrentID=0, Status='live', Symbol='CDIPT-AS1', Aliases='', description='CDIPT antisense RNA 1 (head to head)', other_designations='CTD-2574D22.5|Putative uncharacterized protein LOC440356', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29863683, end_position_on_the_genomic_accession=29868053, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  440359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440359, CurrentID=0, Status='live', Symbol='YBX3P1', Aliases='CSDAP1', description='Y-box binding protein 3 pseudogene 1', other_designations='Y box binding protein 3 pseudogene 1|cold shock domain protein A pseudogene 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31567767, end_position_on_the_genomic_accession=31569524, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  641372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641372, CurrentID=0, Status='live', Symbol='ACOT6', Aliases='C14orf42, c14_5530', description='acyl-CoA thioesterase 6', other_designations='putative acyl-CoA thioesterase 6', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73612296, end_position_on_the_genomic_accession=73619908, orientation='plus', exon_count=3, OMIM=614267, no_hdr0=''),
  100189133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189133, CurrentID=0, Status='live', Symbol='TRR-TCT1-1', Aliases='TRNAR15', description='transfer RNA-Arg (TCT) 1-1', other_designations='transfer RNA arginine 15 (anticodon UCU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93847573, end_position_on_the_genomic_accession=93847657, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440366, CurrentID=0, Status='live', Symbol='HERC2P8', Aliases='', description='hect domain and RLD 2 pseudogene 8', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33095032, end_position_on_the_genomic_accession=33128089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440375, CurrentID=0, Status='live', Symbol='RAB43P1', Aliases='RAB41P, RAB43P', description='RAB43 pseudogene 1', other_designations='RAB41, member RAS homolog family, pseudogene', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46622875, end_position_on_the_genomic_accession=46626994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440386, CurrentID=0, Status='live', Symbol='LOC440386', Aliases='', description='uncharacterized LOC440386', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74366843, end_position_on_the_genomic_accession=74369826, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440387, CurrentID=0, Status='live', Symbol='CTRB2', Aliases='', description='chymotrypsinogen B2', other_designations='chymotrypsin B2', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75204096, end_position_on_the_genomic_accession=75207174, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  440390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440390, CurrentID=0, Status='live', Symbol='LOC440390', Aliases='', description='uncharacterized LOC440390', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87057784, end_position_on_the_genomic_accession=87063991, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100289518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289518, CurrentID=0, Status='live', Symbol='LOC100289518', Aliases='', description='immunoglobin superfamily, member 21 pseudogene', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86283927, end_position_on_the_genomic_accession=86284671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440396, CurrentID=0, Status='live', Symbol='HNRNPA1P16', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 16', other_designations='heterogeneous nuclear ribonucleoprotein A1pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2306762, end_position_on_the_genomic_accession=2308048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440400, CurrentID=0, Status='live', Symbol='RNASEK', Aliases='', description='ribonuclease, RNase K', other_designations='RNase kappa|RNase kappa-02 isoform', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7012417, end_position_on_the_genomic_accession=7014533, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440406, CurrentID=0, Status='live', Symbol='UBE2SP2', Aliases='', description='ubiquitin conjugating enzyme E2S pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2S pseudogene 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18677261, end_position_on_the_genomic_accession=18678012, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401082, CurrentID=0, Status='live', Symbol='LINC01205', Aliases='', description='long intergenic non-protein coding RNA 1205', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109409990, end_position_on_the_genomic_accession=109495167, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  440414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440414, CurrentID=0, Status='live', Symbol='TRE17', Aliases='', description='TRE17 protein', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641381, CurrentID=0, Status='live', Symbol='LOC641381', Aliases='', description='ADP-ribosylation factor like GTPase 5A pseudogene', other_designations='ADP-ribosylation factor-like 5A pseudogene', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148331385, end_position_on_the_genomic_accession=148332037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440416, CurrentID=0, Status='live', Symbol='CCDC144NL-AS1', Aliases='', description='CCDC144NL antisense RNA 1', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20868433, end_position_on_the_genomic_accession=20905230, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724624, CurrentID=0, Status='live', Symbol='LOC102724624', Aliases='', description='uncharacterized protein ENSP00000382042-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19089448, end_position_on_the_genomic_accession=19097563, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440419, CurrentID=0, Status='live', Symbol='TBC1D3P5', Aliases='', description='TBC1 domain family member 3 pseudogene 5', other_designations='TBC1 domain family, member 29 pseudogene|TBC1 domain family, member 3 pseudogene 5|TBC1 domain family, member 3H pseudogene', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27418005, end_position_on_the_genomic_accession=27431623, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  440423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440423, CurrentID=0, Status='live', Symbol='SUZ12P1', Aliases='SUZ12P', description='SUZ12 polycomb repressive complex 2 subunit pseudogene 1', other_designations='suppressor of zeste 12 homolog pseudogene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30709608, end_position_on_the_genomic_accession=30770050, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  440434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440434, CurrentID=0, Status='live', Symbol='LOC440434', Aliases='', description='aminopeptidase puromycin sensitive pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38195740, end_position_on_the_genomic_accession=38257249, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  440435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440435, CurrentID=0, Status='live', Symbol='GPR179', Aliases='CSNB1E, GPR158L, GPR158L1', description='G protein-coupled receptor 179', other_designations='GPR158-like 1|probable G-protein coupled receptor 158-like 1', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38325610, end_position_on_the_genomic_accession=38343810, orientation='minus', exon_count=11, OMIM=614515, no_hdr0=''),
  285266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285266, CurrentID=0, Status='live', Symbol='ENTPD3-AS1', Aliases='', description='ENTPD3 antisense RNA 1', other_designations='ENTPD3 antisense RNA 1 (non-protein coding)', map_location='3p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40390951, end_position_on_the_genomic_accession=40453308, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440446, CurrentID=0, Status='live', Symbol='LOC440446', Aliases='', description='uncharacterized LOC440446', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51336715, end_position_on_the_genomic_accession=51342571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440447, CurrentID=0, Status='live', Symbol='RPS2P27', Aliases='RPS2_6_520', description='ribosomal protein S2 pseudogene 27', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134952136, end_position_on_the_genomic_accession=134953068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374186, CurrentID=0, Status='live', Symbol='LOC105374186', Aliases='', description='uncharacterized LOC105374186', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161173532, end_position_on_the_genomic_accession=161182263, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440452, CurrentID=0, Status='live', Symbol='TBC1D3P2', Aliases='', description='TBC1 domain family member 3 pseudogene 2', other_designations='TBC1 domain family, member 3 pseudogene 2', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62264705, end_position_on_the_genomic_accession=62275655, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  401089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401089, CurrentID=0, Status='live', Symbol='FOXL2NB', Aliases='C3orf72', description='FOXL2 neighbor', other_designations='FLJ43329 protein', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138947234, end_position_on_the_genomic_accession=138953990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440456, CurrentID=0, Status='live', Symbol='PLEKHM1P1', Aliases='PLEKHM1P', description='pleckstrin homology and RUN domain containing M1 pseudogene 1', other_designations='pleckstrin homology domain containing, family M (with RUN domain) member 1 pseudogene', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64784841, end_position_on_the_genomic_accession=64837184, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  440459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440459, CurrentID=0, Status='live', Symbol='SLC16A6P1', Aliases='', description='SLC16A6 pseudogene 1', other_designations='solute carrier family 16, member 6 (monocarboxylic acid transporter 7) pseudogene|solute carrier family 16, member 6 pseudogene', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64953683, end_position_on_the_genomic_accession=64957025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440461, CurrentID=0, Status='live', Symbol='LOC440461', Aliases='', description='Rho GTPase activating protein 27 pseudogene', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68198660, end_position_on_the_genomic_accession=68200302, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440465, CurrentID=0, Status='live', Symbol='BAIAP2-AS1', Aliases='', description='BAIAP2 antisense RNA 1 (head to head)', other_designations='BAIAP2 antisense RNA 1 (non-protein coding)', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81029133, end_position_on_the_genomic_accession=81034719, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724923, CurrentID=0, Status='live', Symbol='FRG1KP', Aliases='', description='FSHD region gene 1 family member K, pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65646906, end_position_on_the_genomic_accession=65669994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440479, CurrentID=0, Status='live', Symbol='FLJ34223', Aliases='', description='uncharacterized LOC440479', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440482, CurrentID=0, Status='live', Symbol='ANKRD20A5P', Aliases='ANKRD20A5', description='ankyrin repeat domain 20 family member A5, pseudogene', other_designations='Putative ankyrin repeat domain-containing protein 20A-like protein MGC26718|ankyrin repeat domain 20 family, member A4 pseudogene|ankyrin repeat domain 20 family, member A5, pseudogene|ankyrin repeat domain-containing protein 20A-like protein MGC26718|putative ankyrin repeat domain-containing protein 20A5', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14179097, end_position_on_the_genomic_accession=14227050, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  440487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440487, CurrentID=0, Status='live', Symbol='RPL21P128', Aliases='RPL21_62_1585', description='ribosomal protein L21 pseudogene 128', other_designations='', map_location='18q11.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21119150, end_position_on_the_genomic_accession=21119703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440489, CurrentID=0, Status='live', Symbol='LOC440489', Aliases='', description='ubiquitin fusion degradation 1 like (yeast) pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26830782, end_position_on_the_genomic_accession=26832521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130488, CurrentID=0, Status='live', Symbol='LOC100130488', Aliases='', description='glucoside xylosyltransferase 2 pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94734676, end_position_on_the_genomic_accession=94735790, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440498, CurrentID=0, Status='live', Symbol='HSBP1L1', Aliases='', description='heat shock factor binding protein 1-like 1', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79964582, end_position_on_the_genomic_accession=79970822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440503, CurrentID=0, Status='live', Symbol='PLIN5', Aliases='LSDA5, LSDP5, MLDP, OXPAT', description='perilipin 5', other_designations='lipid storage droplet protein 5', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4522531, end_position_on_the_genomic_accession=4535196, orientation='minus', exon_count=8, OMIM=613248, no_hdr0=''),
  440508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440508, CurrentID=0, Status='live', Symbol='CLEC4GP1', Aliases='', description='C-type lectin domain family 4 member G pseudogene 1', other_designations='C-type lectin domain family 4, member G pseudogene 1|C-type lectin superfamily 4, member G pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7787484, end_position_on_the_genomic_accession=7791012, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  440514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440514, CurrentID=0, Status='live', Symbol='LOC440514', Aliases='', description='citrate synthase, mitochondrial-like', other_designations='citrate synthase pseudogene', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18030512, end_position_on_the_genomic_accession=18033436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440515, CurrentID=0, Status='live', Symbol='ZNF506', Aliases='', description='zinc finger protein 506', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19792711, end_position_on_the_genomic_accession=19821751, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440518, CurrentID=0, Status='live', Symbol='GOLGA2P9', Aliases='', description='golgin A2 pseudogene 9', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22596257, end_position_on_the_genomic_accession=22603550, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  440519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440519, CurrentID=0, Status='live', Symbol='ZNF724P', Aliases='', description='zinc finger protein 724, pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23223805, end_position_on_the_genomic_accession=23250370, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373682, CurrentID=0, Status='live', Symbol='LOC105373682', Aliases='', description='uncharacterized LOC105373682', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150150633, end_position_on_the_genomic_accession=150166366, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440525, CurrentID=0, Status='live', Symbol='PRR13P5', Aliases='', description='proline rich 13 pseudogene 5', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39942656, end_position_on_the_genomic_accession=39943752, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724643, CurrentID=0, Status='live', Symbol='LOC102724643', Aliases='', description='uncharacterized LOC102724643', other_designations='', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47589051, end_position_on_the_genomic_accession=47619453, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927786, CurrentID=0, Status='live', Symbol='LOC101927786', Aliases='', description='uncharacterized LOC101927786', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130332669, end_position_on_the_genomic_accession=130336240, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440533, CurrentID=0, Status='live', Symbol='PSG8', Aliases='PSG1', description='pregnancy specific beta-1-glycoprotein 8', other_designations='C1 alternate|C2 alternate|C3 alternate|PS-beta-G-8|PSBG-8|pregnancy-specific beta-1 glycoprotein|pregnancy-specific glycoprotein 8', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42752687, end_position_on_the_genomic_accession=42765692, orientation='minus', exon_count=7, OMIM=176397, no_hdr0=''),
  100132400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132400, CurrentID=0, Status='live', Symbol='RPS15AP28', Aliases='RPS15A_16_1058', description='ribosomal protein S15a pseudogene 28', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69300393, end_position_on_the_genomic_accession=69300847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724644, CurrentID=0, Status='live', Symbol='LOC102724644', Aliases='', description='uncharacterized LOC102724644', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133250655, end_position_on_the_genomic_accession=133337349, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  401103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401103, CurrentID=0, Status='live', Symbol='FLJ46066', Aliases='', description='uncharacterized LOC401103', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182446970, end_position_on_the_genomic_accession=182486362, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  401105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401105, CurrentID=0, Status='live', Symbol='FLJ42393', Aliases='AVPC1948', description='uncharacterized LOC401105', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188178543, end_position_on_the_genomic_accession=188180808, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=26502, CurrentID=0, Status='live', Symbol='NARF', Aliases='IOP2', description='nuclear prelamin A recognition factor', other_designations='iron-only hydrogenase-like protein 2|prenyl-dependent prelamin A binding protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82458184, end_position_on_the_genomic_accession=82488267, orientation='plus', exon_count=16, OMIM=605349, no_hdr0=''),
  440556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440556, CurrentID=0, Status='live', Symbol='LINC00982', Aliases='', description='long intergenic non-protein coding RNA 982', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3059617, end_position_on_the_genomic_accession=3067725, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  440558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440558, CurrentID=0, Status='live', Symbol='LOC440558', Aliases='', description='uncharacterized LOC440558', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=53736201, end_position_on_the_genomic_accession=53739310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440560, CurrentID=0, Status='live', Symbol='PRAMEF11', Aliases='', description='PRAME family member 11', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12824380, end_position_on_the_genomic_accession=12831410, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440561, CurrentID=0, Status='live', Symbol='PRAMEF6', Aliases='', description='PRAME family member 6', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12938472, end_position_on_the_genomic_accession=12947580, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440563, CurrentID=0, Status='live', Symbol='HNRNPCL2', Aliases='HNRNPCP5', description='heterogeneous nuclear ribonucleoprotein C-like 2', other_designations='heterogeneous nuclear ribonucleoprotein C pseudogene 5|hnRNP C-like-2', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13115488, end_position_on_the_genomic_accession=13116854, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440567, CurrentID=0, Status='live', Symbol='UQCRHL', Aliases='', description='ubiquinol-cytochrome c reductase hinge protein like', other_designations='hCG25371|ubiquinol-cytochrome c reductase hinge protein-like', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15807162, end_position_on_the_genomic_accession=15807699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440568, CurrentID=0, Status='live', Symbol='LOC440568', Aliases='', description='family with sequence similarity 131 member C pseudogene', other_designations='family with sequence similarity 131, member C pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16034510, end_position_on_the_genomic_accession=16040031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374190, CurrentID=0, Status='live', Symbol='LOC105374190', Aliases='', description='uncharacterized LOC105374190', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164402462, end_position_on_the_genomic_accession=164419606, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440570, CurrentID=0, Status='live', Symbol='LOC440570', Aliases='', description='uncharacterized LOC440570', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440574, CurrentID=0, Status='live', Symbol='MINOS1', Aliases='C1orf151, MIO10, Mic10', description='mitochondrial inner membrane organizing system 1', other_designations='RP5-1056L3.2|UPF0327 protein C1orf151|mitochondrial inner membrane organizing system protein 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19596977, end_position_on_the_genomic_accession=19629821, orientation='plus', exon_count=7, OMIM=616574, no_hdr0=''),
  440575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440575, CurrentID=0, Status='live', Symbol='RPL21P29', Aliases='RPL21_1_19', description='ribosomal protein L21 pseudogene 29', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21908058, end_position_on_the_genomic_accession=21908614, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440577, CurrentID=0, Status='live', Symbol='NPM1P39', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 39', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27206835, end_position_on_the_genomic_accession=27207875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724651, CurrentID=0, Status='live', Symbol='LOC102724651', Aliases='', description='uncharacterized LOC102724651', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=53568447, end_position_on_the_genomic_accession=53597851, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  440584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440584, CurrentID=0, Status='live', Symbol='SLC2A1-AS1', Aliases='', description='SLC2A1 antisense RNA 1', other_designations='SLC2A1 antisense RNA 1 (non-protein coding)', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42959049, end_position_on_the_genomic_accession=42983358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440585, CurrentID=0, Status='live', Symbol='FAM183A', Aliases='', description='family with sequence similarity 183 member A', other_designations='family with sequence similarity 183, member A|hCG23177', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43142979, end_position_on_the_genomic_accession=43156396, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  440587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440587, CurrentID=0, Status='live', Symbol='RPL6P1', Aliases='RPL6_1_44', description='ribosomal protein L6 pseudogene 1', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45781259, end_position_on_the_genomic_accession=45781958, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440589, CurrentID=0, Status='live', Symbol='RPS2P8', Aliases='RPS2_1_48', description='ribosomal protein S2 pseudogene 8', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51250562, end_position_on_the_genomic_accession=51251494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440590, CurrentID=0, Status='live', Symbol='ZYG11A', Aliases='ZYG11', description='zyg-11 family member A, cell cycle regulator', other_designations='ZYG-11A early embryogenesis protein|zyg-11 homolog A', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52842511, end_position_on_the_genomic_accession=52894575, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  102724653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724653, CurrentID=0, Status='live', Symbol='LOC102724653', Aliases='', description='uncharacterized LOC102724653', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22087767, end_position_on_the_genomic_accession=22104619, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440594, CurrentID=0, Status='live', Symbol='FLJ31662', Aliases='', description='uncharacterized LOC440594', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95474737, end_position_on_the_genomic_accession=95479356, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440595, CurrentID=0, Status='live', Symbol='EEF1A1P11', Aliases='EEF1AL9', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 11', other_designations='eukaryotic translation elongation factor 1 alpha 1 (EEF1A1) pseudogene|eukaryotic translation elongation factor 1 alpha-like 9 pseudogene', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96446930, end_position_on_the_genomic_accession=96448623, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401113, CurrentID=0, Status='live', Symbol='FLJ41130', Aliases='', description='uncharacterized LOC401113', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440600, CurrentID=0, Status='live', Symbol='LOC440600', Aliases='', description='uncharacterized LOC440600', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110286375, end_position_on_the_genomic_accession=110339171, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440602, CurrentID=0, Status='live', Symbol='LOC440602', Aliases='', description='uncharacterized LOC440602', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110487680, end_position_on_the_genomic_accession=110490258, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440603, CurrentID=0, Status='live', Symbol='BCL2L15', Aliases='Bfk, C1orf178', description='BCL2-like 15', other_designations='bcl2-L-15|pro-apoptotic Bcl-2 protein', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113876814, end_position_on_the_genomic_accession=113887547, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  440606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440606, CurrentID=0, Status='live', Symbol='HSD3BP2', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta, pseudogene 2', other_designations='3-beta hydroxysteroid dehydrogenase pseudogene|hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 2 pseudogene', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119438528, end_position_on_the_genomic_accession=119446506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  138971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138971, CurrentID=0, Status='live', Symbol='DPP3P2', Aliases='', description='dipeptidyl-peptidase 3 pseudogene 2', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73474177, end_position_on_the_genomic_accession=73476080, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874371, CurrentID=0, Status='live', Symbol='ARHGEF7-IT1', Aliases='', description='ARHGEF7 intronic transcript 1', other_designations='ARHGEF7 intronic transcript 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724657, CurrentID=0, Status='live', Symbol='LOC102724657', Aliases='', description='uncharacterized LOC102724657', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178691534, end_position_on_the_genomic_accession=178708586, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102724658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724658, CurrentID=0, Status='live', Symbol='LOC102724658', Aliases='', description='C-terminal-binding protein 2 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724659, CurrentID=0, Status='live', Symbol='LOC102724659', Aliases='', description='uncharacterized LOC102724659', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11777077, end_position_on_the_genomic_accession=11779619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373558, CurrentID=0, Status='live', Symbol='LOC105373558', Aliases='', description='uncharacterized LOC105373558', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111279790, end_position_on_the_genomic_accession=111291564, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533940, CurrentID=0, Status='live', Symbol='LOC100533940', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724665, CurrentID=0, Status='live', Symbol='LOC102724665', Aliases='', description='putative protein PRCA2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92644473, end_position_on_the_genomic_accession=92644805, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  401124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401124, CurrentID=0, Status='live', Symbol='DTHD1', Aliases='', description='death domain containing 1', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36281611, end_position_on_the_genomic_accession=36345756, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  440672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440672, CurrentID=0, Status='live', Symbol='NUDT4P1', Aliases='', description='nudix hydrolase 4 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 4 pseudogene 1', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148748600, end_position_on_the_genomic_accession=148752401, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060126, CurrentID=0, Status='live', Symbol='LOC101060126', Aliases='', description='atrophin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11761379, end_position_on_the_genomic_accession=11762321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440683, CurrentID=0, Status='live', Symbol='LOC440683', Aliases='', description='seven transmembrane helix receptor', other_designations='dopamine receptor D5 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440686, CurrentID=0, Status='live', Symbol='HIST2H3PS2', Aliases='p06', description='histone cluster 2, H3, pseudogene 2', other_designations='histone 2, H3, pseudogene 2', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143905510, end_position_on_the_genomic_accession=143905966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440689, CurrentID=0, Status='live', Symbol='HIST2H2BF', Aliases='', description='histone cluster 2, H2bf', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149755104, end_position_on_the_genomic_accession=149813123, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440695, CurrentID=0, Status='live', Symbol='ETV3L', Aliases='', description='ets variant 3-like', other_designations='ets variant gene 3-like', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157092043, end_position_on_the_genomic_accession=157099808, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440699, CurrentID=0, Status='live', Symbol='LRRC52', Aliases='', description='leucine rich repeat containing 52', other_designations='BK channel auxiliary gamma subunit LRRC52|BK channel auxilliary gamma subunit LRRC52', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165544241, end_position_on_the_genomic_accession=165563948, orientation='plus', exon_count=2, OMIM=615218, no_hdr0=''),
  440700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440700, CurrentID=0, Status='live', Symbol='LOC440700', Aliases='', description='carbonic anhydrase XIV (CA14) pseudogene', other_designations='carbonic anhydrase XIV pseudogene', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165698750, end_position_on_the_genomic_accession=165709968, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874374, CurrentID=0, Status='live', Symbol='ARHGEF38-IT1', Aliases='', description='ARHGEF38 intronic transcript 1', other_designations='ARHGEF38 intronic transcript 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105561591, end_position_on_the_genomic_accession=105570238, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440704, CurrentID=0, Status='live', Symbol='LOC440704', Aliases='', description='uncharacterized LOC440704', other_designations='', map_location='1q31.1-q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=190624890, end_position_on_the_genomic_accession=190801658, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100420464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420464, CurrentID=0, Status='live', Symbol='LOC100420464', Aliases='', description='salvador family WW domain containing protein 1 pseudogene', other_designations='salvador homolog 1 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143911513, end_position_on_the_genomic_accession=143912250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401131, CurrentID=0, Status='live', Symbol='LOC401131', Aliases='', description='uncharacterized LOC401131', other_designations='', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=141959955, end_position_on_the_genomic_accession=142052593, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  554049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554049, CurrentID=0, Status='live', Symbol='C1DP1', Aliases='bA195O1.1', description='C1D nuclear receptor corepressor pseudogene 1', other_designations='C1D nuclear receptor co-repressor pseudogene 1|DNA-binding protein (C1D) pseudogene', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32510628, end_position_on_the_genomic_accession=32511819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440712, CurrentID=0, Status='live', Symbol='C1orf186', Aliases='RHEX', description='chromosome 1 open reading frame 186', other_designations='regulator of human erythroid cell expansion', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206052723, end_position_on_the_genomic_accession=206102701, orientation='plus', exon_count=7, OMIM=616088, no_hdr0=''),
  440714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440714, CurrentID=0, Status='live', Symbol='C1orf143', Aliases='', description='chromosome 1 open reading frame 143', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440716, CurrentID=0, Status='live', Symbol='LOC440716', Aliases='', description='uncharacterized LOC440716', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374195, CurrentID=0, Status='live', Symbol='LOC105374195', Aliases='', description='uncharacterized LOC105374195', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166971879, end_position_on_the_genomic_accession=166973797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401134, CurrentID=0, Status='live', Symbol='LOC401134', Aliases='', description='uncharacterized LOC401134', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64914281, end_position_on_the_genomic_accession=65004500, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  641433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641433, CurrentID=0, Status='live', Symbol='RP32', Aliases='', description='retinitis pigmentosa 32 (autosomal recessive)', other_designations='', map_location='1p21.2-p13.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609913, no_hdr0=''),
  440730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440730, CurrentID=0, Status='live', Symbol='TRIM67', Aliases='TNL', description='tripartite motif containing 67', other_designations='TRIM9-like protein TNL', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231162624, end_position_on_the_genomic_accession=231221568, orientation='plus', exon_count=10, OMIM=610584, no_hdr0=''),
  401135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401135, CurrentID=0, Status='live', Symbol='SYT14P1', Aliases='CHR415SYT, SYT14L, SYTDEP', description='synaptotagmin 14 pseudogene 1', other_designations='chr415 synaptotagmin|synaptotagmin XIV pseudogene 1|synaptotagmin XIV-derived|synaptotagmin XIV-like', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68060610, end_position_on_the_genomic_accession=68063297, orientation='minus', exon_count=2, OMIM=610892, no_hdr0=''),
  440732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440732, CurrentID=0, Status='live', Symbol='RPS7P3', Aliases='RPS7_2_175', description='ribosomal protein S7 pseudogene 3', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233288809, end_position_on_the_genomic_accession=233289491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724677, CurrentID=0, Status='live', Symbol='LOC102724677', Aliases='', description='carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6664290, end_position_on_the_genomic_accession=6666429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440737, CurrentID=0, Status='live', Symbol='RPL35P1', Aliases='RPL35_1_181', description='ribosomal protein L35 pseudogene 1', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236981309, end_position_on_the_genomic_accession=236981743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440738, CurrentID=0, Status='live', Symbol='MAP1LC3C', Aliases='ATG8J, LC3C', description='microtubule associated protein 1 light chain 3 gamma', other_designations='LC3-like protein 2|MAP1 light chain 3-like protein 2|MAP1 light chain 3-like protein 3|MAP1A/MAP1B LC3 C|MAP1A/MAP1B light chain 3 C|autophagy-related protein LC3 C|autophagy-related ubiquitin-like modifier LC3 C|microtubule-associated protein 1 light chain 3 gamma', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241995490, end_position_on_the_genomic_accession=242001399, orientation='minus', exon_count=5, OMIM=609605, no_hdr0=''),
  440742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440742, CurrentID=0, Status='live', Symbol='LOC440742', Aliases='', description='uncharacterized LOC440742', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244064614, end_position_on_the_genomic_accession=244065102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7921, CurrentID=0, Status='live', Symbol='D6S207E', Aliases='HLA-HA2', description='Minor histocompatibility antigen HA-2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401138, CurrentID=0, Status='live', Symbol='AMTN', Aliases='UNQ689', description='amelotin', other_designations='RSTI689', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70518572, end_position_on_the_genomic_accession=70532743, orientation='plus', exon_count=9, OMIM=610912, no_hdr0=''),
  440757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440757, CurrentID=0, Status='live', Symbol='LINC00851', Aliases='DZANK1-AS1', description='long intergenic non-protein coding RNA 851', other_designations='DZANK1 antisense RNA 1 (non-protein coding)|DZANK1 antisense RNA 1 (tail to tail)', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18379049, end_position_on_the_genomic_accession=18381484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724683, CurrentID=0, Status='live', Symbol='LOC102724683', Aliases='', description='uncharacterized LOC102724683', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58145550, end_position_on_the_genomic_accession=58166735, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23406, CurrentID=0, Status='live', Symbol='COTL1', Aliases='CLP', description='coactosin-like F-actin binding protein 1', other_designations='coactosin-like 1', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84565596, end_position_on_the_genomic_accession=84618096, orientation='minus', exon_count=4, OMIM=606748, no_hdr0=''),
  101928403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928403, CurrentID=0, Status='live', Symbol='LOC101928403', Aliases='', description='uncharacterized LOC101928403', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88627866, end_position_on_the_genomic_accession=88631825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100533944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533944, CurrentID=0, Status='live', Symbol='LOC100533944', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129617, CurrentID=0, Status='live', Symbol='LOC100129617', Aliases='', description='uncharacterized LOC100129617', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81665354, end_position_on_the_genomic_accession=81667276, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440786, CurrentID=0, Status='live', Symbol='LOC440786', Aliases='', description='uncharacterized LOC440786', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=17394562, end_position_on_the_genomic_accession=17395869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440792, CurrentID=0, Status='live', Symbol='LOC440792', Aliases='', description='proline dehydrogenase (oxidase) 1 pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20297310, end_position_on_the_genomic_accession=20313590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440795, CurrentID=0, Status='live', Symbol='CA15P2', Aliases='', description='carbonic anhydrase XV, pseudogene 2', other_designations='carbonic anhydrase family pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18635987, end_position_on_the_genomic_accession=18637497, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440799, CurrentID=0, Status='live', Symbol='P2RX6P', Aliases='', description='purinergic receptor P2X, ligand gated ion channel, 6 pseudogene', other_designations='purinergic receptor P2X, ligand-gated ion channel, 6 pseudogene|purinergic receptor P2X-like 1, orphan receptor pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21042392, end_position_on_the_genomic_accession=21044249, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372806, CurrentID=0, Status='live', Symbol='LOC105372806', Aliases='', description='uncharacterized LOC105372806', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41164261, end_position_on_the_genomic_accession=41167256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440804, CurrentID=0, Status='live', Symbol='RIMBP3B', Aliases='RIM-BP3.2, RIM-BP3.B, RIMBP3.2', description='RIMS binding protein 3B', other_designations='RIMS binding protein 3.2|RIMS-binding protein 3.2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21383751, end_position_on_the_genomic_accession=21389169, orientation='plus', exon_count=1, OMIM=612700, no_hdr0=''),
  401149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401149, CurrentID=0, Status='live', Symbol='LINC01061', Aliases='', description='long intergenic non-protein coding RNA 1061', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119405523, end_position_on_the_genomic_accession=119410660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  160851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=160851, CurrentID=0, Status='live', Symbol='DGKH', Aliases='DGKeta', description='diacylglycerol kinase, eta', other_designations='DAG kinase eta|diglyceride kinase eta', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42040036, end_position_on_the_genomic_accession=42256584, orientation='plus', exon_count=38, OMIM=604071, no_hdr0=''),
  100128045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128045, CurrentID=0, Status='live', Symbol='LOC100128045', Aliases='', description='homeobox protein cut-like 1', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75499257, end_position_on_the_genomic_accession=75580753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440822, CurrentID=0, Status='live', Symbol='PIWIL3', Aliases='HIWI3', description='piwi-like RNA-mediated gene silencing 3', other_designations='piwi-like 3', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24719034, end_position_on_the_genomic_accession=24774720, orientation='minus', exon_count=23, OMIM=610314, no_hdr0=''),
  440823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440823, CurrentID=0, Status='live', Symbol='MIAT', Aliases='C22orf35, GOMAFU, LINC00066, NCRNA00066, RNCR2, lncRNA-MIAT', description='myocardial infarction associated transcript (non-protein coding)', other_designations='MI related novel mRNA|long intergenic non-protein coding RNA 66', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26657482, end_position_on_the_genomic_accession=26676478, orientation='plus', exon_count=5, OMIM=611082, no_hdr0=''),
  440829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440829, CurrentID=0, Status='live', Symbol='SHISA8', Aliases='C22orf17, Orf26', description='shisa family member 8', other_designations='CTA-250D10.17|shisa homolog 8', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41909533, end_position_on_the_genomic_accession=41915093, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  401152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401152, CurrentID=0, Status='live', Symbol='C4orf3', Aliases='', description='chromosome 4 open reading frame 3', other_designations='HCV F-transactivated protein 1|hepatitis C virus F protein-transactivated protein 1', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119296419, end_position_on_the_genomic_accession=119304445, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440836, CurrentID=0, Status='live', Symbol='ODF3B', Aliases='ODF3L3', description='outer dense fiber of sperm tails 3B', other_designations="orf2 5' to PD-ECGF/TP|outer dense fiber protein 3-like protein 3", map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50530409, end_position_on_the_genomic_accession=50532579, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928661, CurrentID=0, Status='live', Symbol='LOC101928661', Aliases='', description='uncharacterized LOC101928661', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146594516, end_position_on_the_genomic_accession=146598931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874379, CurrentID=0, Status='live', Symbol='POM121L13P', Aliases='', description='POM121 transmembrane nucleoporin like 13, pseudogene', other_designations='POM121 membrane glycoprotein-like 13, pseudogene|POM121 transmembrane nucleoporin-like 13, pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28778558, end_position_on_the_genomic_accession=28778937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440854, CurrentID=0, Status='live', Symbol='CAPN14', Aliases='', description='calpain 14', other_designations='CANP 14|calcium-activated neutral proteinase 14', map_location='2p23.1-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31173056, end_position_on_the_genomic_accession=31233971, orientation='minus', exon_count=26, OMIM=610229, no_hdr0=''),
  29785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29785, CurrentID=0, Status='live', Symbol='CYP2S1', Aliases='CYPIIS1', description='cytochrome P450 family 2 subfamily S member 1', other_designations='cytochrome P450, family 2, subfamily S, polypeptide 1|cytochrome P540, subfamily IIS, polypeptide 1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41193207, end_position_on_the_genomic_accession=41207539, orientation='plus', exon_count=9, OMIM=611529, no_hdr0=''),
  440862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440862, CurrentID=0, Status='live', Symbol='LOC440862', Aliases='', description='uncharacterized LOC440862', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401157, CurrentID=0, Status='live', Symbol='FLJ44477', Aliases='', description='FLJ44477 protein', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=143766690, end_position_on_the_genomic_accession=143769122, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440864, CurrentID=0, Status='live', Symbol='LOC440864', Aliases='', description='uncharacterized LOC440864', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47695881, end_position_on_the_genomic_accession=47709696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440865, CurrentID=0, Status='live', Symbol='FLJ46838', Aliases='', description='FLJ46838 protein', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=48756528, end_position_on_the_genomic_accession=48759056, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440867, CurrentID=0, Status='live', Symbol='LOC440867', Aliases='', description='uncharacterized LOC440867', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65436702, end_position_on_the_genomic_accession=65459007, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374200, CurrentID=0, Status='live', Symbol='LOC105374200', Aliases='', description='uncharacterized LOC105374200', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168609441, end_position_on_the_genomic_accession=168627438, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440888, CurrentID=0, Status='live', Symbol='ACTR3BP2', Aliases='FKSG73', description='ACTR3B pseudogene 2', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 2', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91941133, end_position_on_the_genomic_accession=91942470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  440894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440894, CurrentID=0, Status='live', Symbol='LINC01123', Aliases='', description='long intergenic non-protein coding RNA 1123', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109986980, end_position_on_the_genomic_accession=109995281, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440895, CurrentID=0, Status='live', Symbol='LOC440895', Aliases='', description='two pore channel 3 pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109947690, end_position_on_the_genomic_accession=109968574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  440896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440896, CurrentID=0, Status='live', Symbol='LOC440896', Aliases='', description='uncharacterized LOC440896', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41100985, end_position_on_the_genomic_accession=41107812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374201, CurrentID=0, Status='live', Symbol='LOC105374201', Aliases='', description='uncharacterized LOC105374201', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169003014, end_position_on_the_genomic_accession=169006031, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440900, CurrentID=0, Status='live', Symbol='LINC01191', Aliases='VIN, lnc-ACTR3', description='long intergenic non-protein coding RNA 1191', other_designations='virus inducible lincRNa', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113979569, end_position_on_the_genomic_accession=114007310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440902, CurrentID=0, Status='live', Symbol='NPM1P32', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 32', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121708518, end_position_on_the_genomic_accession=121709112, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440905, CurrentID=0, Status='live', Symbol='FAR2P1', Aliases='HEL-182', description='fatty acyl-CoA reductase 2 pseudogene 1', other_designations='epididymis luminal protein 182|fatty acyl CoA reductase 2 pseudogene 1|fatty acyl-CoA reductase 2-like', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130025998, end_position_on_the_genomic_accession=130051131, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  440910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440910, CurrentID=0, Status='live', Symbol='LOC440910', Aliases='', description='uncharacterized LOC440910', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131279290, end_position_on_the_genomic_accession=131299819, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100128048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128048, CurrentID=0, Status='live', Symbol='LOC100128048', Aliases='', description='cysteine and histidine rich domain containing 1 pseudogene', other_designations='cysteine and histidine-rich domain (CHORD) containing 1 pseudogene', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42824728, end_position_on_the_genomic_accession=42827705, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440915, CurrentID=0, Status='live', Symbol='POTEKP', Aliases='ACT, ACTBL3, POTE2delta, POTEK', description='POTE ankyrin domain family member K, pseudogene', other_designations='POTE ankyrin domain family member K|POTE ankyrin domain family, member K, pseudogene|actin, beta-like 3|actin-like protein|beta-actin-like protein 3|kappa-actin', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131592093, end_position_on_the_genomic_accession=131609401, orientation='plus', exon_count=7, OMIM=611266, no_hdr0=''),
  440917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440917, CurrentID=0, Status='live', Symbol='YWHAEP5', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 5', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide pseudogene', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138287130, end_position_on_the_genomic_accession=138288880, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440918, CurrentID=0, Status='live', Symbol='FLJ46875', Aliases='', description='uncharacterized LOC440918', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=156901336, end_position_on_the_genomic_accession=156924941, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23411, CurrentID=0, Status='live', Symbol='SIRT1', Aliases='SIR2, SIR2L1, SIR2alpha', description='sirtuin 1', other_designations='SIR2-like protein 1|regulatory protein SIR2 homolog 1|sirtuin type 1', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67884662, end_position_on_the_genomic_accession=67918390, orientation='plus', exon_count=11, OMIM=604479, no_hdr0=''),
  440925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440925, CurrentID=0, Status='live', Symbol='LINC01124', Aliases='', description='long intergenic non-protein coding RNA 1124', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170712439, end_position_on_the_genomic_accession=170714567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  440926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440926, CurrentID=0, Status='live', Symbol='H3F3AP4', Aliases='p13', description='H3 histone, family 3A, pseudogene 4', other_designations='H3 histone, family 3A pseudogene', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174719800, end_position_on_the_genomic_accession=174720858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100533949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533949, CurrentID=0, Status='live', Symbol='LOC100533949', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440934, CurrentID=0, Status='live', Symbol='LOC440934', Aliases='', description='cancer/testis antigen 75', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222314143, end_position_on_the_genomic_accession=222321722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422039, CurrentID=0, Status='live', Symbol='LOC100422039', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440944, CurrentID=0, Status='live', Symbol='THUMPD3-AS1', Aliases='SETD5-AS1', description='THUMPD3 antisense RNA 1', other_designations='SETD5 antisense RNA 1', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9388853, end_position_on_the_genomic_accession=9397490, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422479, CurrentID=0, Status='live', Symbol='LOC100422479', Aliases='', description='integrin alpha X pseudogene', other_designations='integrin, alpha X (complement component 3 receptor 4 subunit) pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31360720, end_position_on_the_genomic_accession=31361704, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  440952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440952, CurrentID=0, Status='live', Symbol='EIF1B-AS1', Aliases='', description='EIF1B antisense RNA 1', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40173145, end_position_on_the_genomic_accession=40309698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401172, CurrentID=0, Status='live', Symbol='FLJ33360', Aliases='', description='FLJ33360 protein', other_designations='CTD-2324F15.2', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6310441, end_position_on_the_genomic_accession=6337292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440955, CurrentID=0, Status='live', Symbol='TMEM89', Aliases='', description='transmembrane protein 89', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48620842, end_position_on_the_genomic_accession=48621756, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  440956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440956, CurrentID=0, Status='live', Symbol='IQCF6', Aliases='', description='IQ motif containing F6', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51778561, end_position_on_the_genomic_accession=51779773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  440957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440957, CurrentID=0, Status='live', Symbol='SMIM4', Aliases='C3orf78', description='small integral membrane protein 4', other_designations='UPF0640 protein C3orf78', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52535986, end_position_on_the_genomic_accession=52540570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440970, CurrentID=0, Status='live', Symbol='LINC00971', Aliases='', description='long intergenic non-protein coding RNA 971', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=84638405, end_position_on_the_genomic_accession=84869575, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100128050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128050, CurrentID=0, Status='live', Symbol='LOC100128050', Aliases='', description='WD repeat domain 77 pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154523990, end_position_on_the_genomic_accession=154524806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440973, CurrentID=0, Status='live', Symbol='LOC440973', Aliases='', description='nuclear transcription factor Y subunit beta pseudogene', other_designations='nuclear transcription factor Y, beta pseudogene', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109915984, end_position_on_the_genomic_accession=109916940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401176, CurrentID=0, Status='live', Symbol='LOC401176', Aliases='', description='uncharacterized LOC401176', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16180492, end_position_on_the_genomic_accession=16187064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  440981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440981, CurrentID=0, Status='live', Symbol='LOC440981', Aliases='', description='phospholipid scramblase-like', other_designations='phospholipid scramblase 1-like', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=146109150, end_position_on_the_genomic_accession=146134385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  440982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440982, CurrentID=0, Status='live', Symbol='LOC440982', Aliases='', description='uncharacterized LOC440982', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147421327, end_position_on_the_genomic_accession=147509910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724719, CurrentID=0, Status='live', Symbol='LOC102724719', Aliases='', description='uncharacterized LOC102724719', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51062579, end_position_on_the_genomic_accession=51068552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100533951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533951, CurrentID=0, Status='live', Symbol='LOC100533951', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  440991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440991, CurrentID=0, Status='live', Symbol='RPS3P3', Aliases='RPS3_1_443', description='ribosomal protein S3 pseudogene 3', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184149674, end_position_on_the_genomic_accession=184150520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440992, CurrentID=0, Status='live', Symbol='RPS20P14', Aliases='RPS20_6_446', description='ribosomal protein S20 pseudogene 14', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186900073, end_position_on_the_genomic_accession=186900622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440993, CurrentID=0, Status='live', Symbol='LINC00969', Aliases='', description='long intergenic non-protein coding RNA 969', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195688498, end_position_on_the_genomic_accession=195711875, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  441009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441009, CurrentID=0, Status='live', Symbol='CPEB2-AS1', Aliases='', description='CPEB2 antisense RNA 1 (head to head)', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14909961, end_position_on_the_genomic_accession=15002045, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  441013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441013, CurrentID=0, Status='live', Symbol='RPS2P21', Aliases='RPS2_5_473', description='ribosomal protein S2 pseudogene 21', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43410004, end_position_on_the_genomic_accession=43410960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533952, CurrentID=0, Status='live', Symbol='RBAK-RBAKDN', Aliases='RBAK-LOC389458', description='RBAK-RBAKDN readthrough', other_designations='RBAK-LOC389458 readthrough', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5045821, end_position_on_the_genomic_accession=5073223, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  441024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441024, CurrentID=0, Status='live', Symbol='MTHFD2L', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2-like', other_designations='MTHFD2-like|NADP-dependent methylenetetrahydrofolate dehydrogenase 2-like protein', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74114183, end_position_on_the_genomic_accession=74303099, orientation='plus', exon_count=20, OMIM=614047, no_hdr0=''),
  441025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441025, CurrentID=0, Status='live', Symbol='LOC441025', Aliases='', description='uncharacterized LOC441025', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75354076, end_position_on_the_genomic_accession=75362566, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441026, CurrentID=0, Status='live', Symbol='HNRNPA3P13', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 13', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82128528, end_position_on_the_genomic_accession=82131631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441027, CurrentID=0, Status='live', Symbol='TMEM150C', Aliases='', description='transmembrane protein 150C', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82484451, end_position_on_the_genomic_accession=82561973, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  441032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441032, CurrentID=0, Status='live', Symbol='EEF1A1P9', Aliases='EEF1AL7', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 9', other_designations='eukaryotic translation elongation factor 1 alpha-like 7', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105484626, end_position_on_the_genomic_accession=105486376, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441034, CurrentID=0, Status='live', Symbol='RPL7AP30', Aliases='RPL7A_9_511', description='ribosomal protein L7a pseudogene 30', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112788024, end_position_on_the_genomic_accession=112788900, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441035, CurrentID=0, Status='live', Symbol='LOC441035', Aliases='', description='uncharacterized LOC441035', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=116093277, end_position_on_the_genomic_accession=116102640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  285286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285286, CurrentID=0, Status='live', Symbol='LINC00877', Aliases='', description='long intergenic non-protein coding RNA 877', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72035519, end_position_on_the_genomic_accession=72100455, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  441046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441046, CurrentID=0, Status='live', Symbol='GUSBP5', Aliases='GYPELOC441046', description='glucuronidase, beta pseudogene 5', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143559472, end_position_on_the_genomic_accession=143561460, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441050, CurrentID=0, Status='live', Symbol='PTGES3P3', Aliases='PTGES3P', description='prostaglandin E synthase 3 (cytosolic) pseudogene 3', other_designations='hCG2026038', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169791149, end_position_on_the_genomic_accession=169791836, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441051, CurrentID=0, Status='live', Symbol='HSP90AA6P', Aliases='HSP90Af', description='heat shock protein 90kDa alpha family class A member 6, pseudogene', other_designations='heat shock protein 90Af|heat shock protein 90kDa alpha (cytosolic), class A member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class A member 6, pseudogene', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170604715, end_position_on_the_genomic_accession=170605458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441052, CurrentID=0, Status='live', Symbol='LOC441052', Aliases='', description='uncharacterized LOC441052', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171812136, end_position_on_the_genomic_accession=171813601, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441054, CurrentID=0, Status='live', Symbol='C4orf47', Aliases='', description='chromosome 4 open reading frame 47', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185405235, end_position_on_the_genomic_accession=185449828, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  441056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441056, CurrentID=0, Status='live', Symbol='DUX4L4', Aliases='', description='double homeobox 4 like 4', other_designations='double homeobox protein 4-like protein 4', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190081118, end_position_on_the_genomic_accession=190082396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441058, CurrentID=0, Status='live', Symbol='MGC39584', Aliases='', description='uncharacterized LOC441058', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9781848, end_position_on_the_genomic_accession=9821061, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724731, CurrentID=0, Status='live', Symbol='LOC102724731', Aliases='', description='uncharacterized LOC102724731', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32944648, end_position_on_the_genomic_accession=32950361, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441061, CurrentID=0, Status='live', Symbol='MARCH11', Aliases='MARCH-XI', description='membrane associated ring-CH-type finger 11', other_designations='membrane associated ring finger 11|membrane-associated RING finger protein 11|membrane-associated RING-CH protein XI|membrane-associated ring finger (C3HC4) 11', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16067365, end_position_on_the_genomic_accession=16179788, orientation='minus', exon_count=4, OMIM=613338, no_hdr0=''),
  441066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441066, CurrentID=0, Status='live', Symbol='GUSBP8', Aliases='', description='glucuronidase, beta pseudogene 8', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=99532628, end_position_on_the_genomic_accession=99534367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401191, CurrentID=0, Status='live', Symbol='LOC401191', Aliases='', description='uncharacterized LOC401191', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66298389, end_position_on_the_genomic_accession=66537775, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  444882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=444882, CurrentID=0, Status='live', Symbol='IGFL4', Aliases='', description='IGF like family member 4', other_designations='IGF-like family member 4', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46039748, end_position_on_the_genomic_accession=46057132, orientation='minus', exon_count=5, OMIM=610547, no_hdr0=''),
  441072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441072, CurrentID=0, Status='live', Symbol='FLJ31104', Aliases='', description='uncharacterized LOC441072', other_designations='CTD-2031P19.3', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55995167, end_position_on_the_genomic_accession=56003649, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  441073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441073, CurrentID=0, Status='live', Symbol='RPL26P19', Aliases='RPL26_4_565', description='ribosomal protein L26 pseudogene 19', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56504592, end_position_on_the_genomic_accession=56505114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  51667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51667, CurrentID=0, Status='live', Symbol='NUB1', Aliases='BS4L, NYREN18, NUB1', description='negative regulator of ubiquitin-like proteins 1', other_designations='NEDD8 ultimate buster-1|renal carcinoma antigen NY-REN-18', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151341699, end_position_on_the_genomic_accession=151378461, orientation='plus', exon_count=16, OMIM=607981, no_hdr0=''),
  105374207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374207, CurrentID=0, Status='live', Symbol='LOC105374207', Aliases='', description='uncharacterized LOC105374207', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169170948, end_position_on_the_genomic_accession=169215685, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441081, CurrentID=0, Status='live', Symbol='LOC441081', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70486501, end_position_on_the_genomic_accession=70491787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441086, CurrentID=0, Status='live', Symbol='LOC441086', Aliases='', description='uncharacterized LOC441086', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74866171, end_position_on_the_genomic_accession=74867275, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102546175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546175, CurrentID=0, Status='live', Symbol='LINC01338', Aliases='', description='long intergenic non-protein coding RNA 1338', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82850864, end_position_on_the_genomic_accession=82859836, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  441089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441089, CurrentID=0, Status='live', Symbol='CRSP8P', Aliases='', description='mediator complex subunit 27 pseudogene', other_designations='CRSP8 pseudogene|cofactor required for Sp1 transcriptional activation subunit 8 34kDa pseudogene', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80350605, end_position_on_the_genomic_accession=80351966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441094, CurrentID=0, Status='live', Symbol='NR2F1-AS1', Aliases='', description='NR2F1 antisense RNA 1', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93409356, end_position_on_the_genomic_accession=93581320, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  441098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441098, CurrentID=0, Status='live', Symbol='LOC441098', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100388304, end_position_on_the_genomic_accession=100392932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441108, CurrentID=0, Status='live', Symbol='C5orf56', Aliases='', description='chromosome 5 open reading frame 56', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132410773, end_position_on_the_genomic_accession=132476044, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100533955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533955, CurrentID=0, Status='live', Symbol='SENP3-EIF4A1', Aliases='', description='SENP3-EIF4A1 readthrough (NMD candidate)', other_designations='', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7561992, end_position_on_the_genomic_accession=7579006, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  441114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441114, CurrentID=0, Status='live', Symbol='RPLP0P9', Aliases='', description='ribosomal protein, large, P0 pseudogene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166382305, end_position_on_the_genomic_accession=166382686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441116, CurrentID=0, Status='live', Symbol='FLJ16171', Aliases='', description='FLJ16171 protein', other_designations='CTC-281M20.1', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174919082, end_position_on_the_genomic_accession=174995731, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926917, CurrentID=0, Status='live', Symbol='LOC101926917', Aliases='', description='ankyrin repeat domain-containing protein 18A-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97148127, end_position_on_the_genomic_accession=97221318, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  441123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441123, CurrentID=0, Status='live', Symbol='LINC00266-3', Aliases='C6orf88, NCRNA00266-3, NCRNA00266C, dJ24O22.2', description='long intergenic non-protein coding RNA 266-3', other_designations='non-protein coding RNA 266C', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140264, end_position_on_the_genomic_accession=148159, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  441124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441124, CurrentID=0, Status='live', Symbol='GTF2IP20', Aliases='', description='general transcription factor IIi pseudogene 20', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223951266, end_position_on_the_genomic_accession=223983034, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  441131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441131, CurrentID=0, Status='live', Symbol='ARPC3P5', Aliases='', description='actin related protein 2/3 complex subunit 3 pseudogene 5', other_designations='actin related protein 2/3 complex, subunit 3 pseudogene 5|actin related protein 2/3 complex, subunit 3, 21kDa pseudogene', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15934743, end_position_on_the_genomic_accession=15935514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441132, CurrentID=0, Status='live', Symbol='IMPDH1P9', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 9', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18366016, end_position_on_the_genomic_accession=18368719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441133, CurrentID=0, Status='live', Symbol='KRT18P38', Aliases='', description='keratin 18 pseudogene 38', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19612739, end_position_on_the_genomic_accession=19614090, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724945, CurrentID=0, Status='live', Symbol='LOC102724945', Aliases='', description='uncharacterized LOC102724945', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=33958774, end_position_on_the_genomic_accession=34203705, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102546226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546226, CurrentID=0, Status='live', Symbol='LOC102546226', Aliases='', description='uncharacterized LOC102546226', other_designations='CTC-498M16.2', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88408982, end_position_on_the_genomic_accession=88439090, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102546227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546227, CurrentID=0, Status='live', Symbol='LINC01340', Aliases='', description='long intergenic non-protein coding RNA 1340', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97504696, end_position_on_the_genomic_accession=97671051, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102546228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546228, CurrentID=0, Status='live', Symbol='LOC102546228', Aliases='', description='uncharacterized LOC102546228', other_designations='CTC-339D2.1', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126179577, end_position_on_the_genomic_accession=126193850, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102546229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546229, CurrentID=0, Status='live', Symbol='LOC102546229', Aliases='', description='uncharacterized LOC102546229', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134429051, end_position_on_the_genomic_accession=134435047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  139060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139060, CurrentID=0, Status='live', Symbol='KRT18P68', Aliases='', description='keratin 18 pseudogene 68', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45983315, end_position_on_the_genomic_accession=45984369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441150, CurrentID=0, Status='live', Symbol='C6orf226', Aliases='', description='chromosome 6 open reading frame 226', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42890265, end_position_on_the_genomic_accession=42890816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441151, CurrentID=0, Status='live', Symbol='TMEM151B', Aliases='C6orf137, TMEM193, bA444E17.5', description='transmembrane protein 151B', other_designations='transmembrane protein 193', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44270743, end_position_on_the_genomic_accession=44279445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441154, CurrentID=0, Status='live', Symbol='RPL7AP34', Aliases='RPL7A_14_708', description='ribosomal protein L7a pseudogene 34', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63548650, end_position_on_the_genomic_accession=63549507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441155, CurrentID=0, Status='live', Symbol='LOC441155', Aliases='', description='zinc finger CCCH-type domain-containing-like', other_designations='zinc finger CCCH-type containing 11A-like', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=65301418, end_position_on_the_genomic_accession=65305613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7990, CurrentID=0, Status='live', Symbol='CRS1C', Aliases='', description='Cryptidin-related sequence-1C', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441160, CurrentID=0, Status='live', Symbol='KRT19P1', Aliases='', description='keratin 19 pseudogene 1', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71584660, end_position_on_the_genomic_accession=71586033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441161, CurrentID=0, Status='live', Symbol='OOEP', Aliases='C6orf156, FLOPED, HOEP19, KHDC2', description='oocyte expressed protein', other_designations='KH homology domain containing 2|KH homology domain-containing protein 2|oocyte and embryo protein 19|oocyte expressed protein homolog (dog)|oocyte- and embryo-specific protein 19', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73368555, end_position_on_the_genomic_accession=73369792, orientation='minus', exon_count=3, OMIM=611689, no_hdr0=''),
  401207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401207, CurrentID=0, Status='live', Symbol='C5orf63', Aliases='YDR286C', description='chromosome 5 open reading frame 63', other_designations='glutaredoxin-like protein YDR286C homolog', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127035666, end_position_on_the_genomic_accession=127073539, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  441166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441166, CurrentID=0, Status='live', Symbol='CNN2P9', Aliases='', description='calponin 2 pseudogene 9', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110858169, end_position_on_the_genomic_accession=110859344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441167, CurrentID=0, Status='live', Symbol='LOC441167', Aliases='', description='hCG1820801', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=114551305, end_position_on_the_genomic_accession=114560363, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441168, CurrentID=0, Status='live', Symbol='FAM26F', Aliases='C6orf187, dJ93H18.5', description='family with sequence similarity 26 member F', other_designations='family with sequence similarity 26, member F', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116461370, end_position_on_the_genomic_accession=116463782, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441172, CurrentID=0, Status='live', Symbol='FLJ46906', Aliases='', description='uncharacterized LOC441172', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138691668, end_position_on_the_genomic_accession=138697288, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  139065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139065, CurrentID=0, Status='live', Symbol='SLITRK4', Aliases='', description='SLIT and NTRK like family member 4', other_designations='SLIT and NTRK-like family, member 4|slit and trk like gene 4', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143622790, end_position_on_the_genomic_accession=143636107, orientation='minus', exon_count=6, OMIM=300562, no_hdr0=''),
  441177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441177, CurrentID=0, Status='live', Symbol='LINC00602', Aliases='', description='long intergenic non-protein coding RNA 602', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165987551, end_position_on_the_genomic_accession=165989615, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441178, CurrentID=0, Status='live', Symbol='LOC441178', Aliases='', description='uncharacterized LOC441178', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167679626, end_position_on_the_genomic_accession=167683787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  441179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441179, CurrentID=0, Status='live', Symbol='LOC441179', Aliases='', description='uncharacterized LOC441179', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167795029, end_position_on_the_genomic_accession=167796598, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  441191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441191, CurrentID=0, Status='live', Symbol='RNF216P1', Aliases='RNF216L, hCG2040556', description='ring finger protein 216 pseudogene 1', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4973985, end_position_on_the_genomic_accession=4998169, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  441194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441194, CurrentID=0, Status='live', Symbol='PMS2CL', Aliases='PMS2P13', description='PMS2 C-terminal like pseudogene', other_designations='PMS1 homolog 2, mismatch repair system component pseudogene 13|postmeiotic segregation increased 2 pseudogene 13', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6735305, end_position_on_the_genomic_accession=6751601, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  441204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441204, CurrentID=0, Status='live', Symbol='LOC441204', Aliases='', description='uncharacterized LOC441204', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26398614, end_position_on_the_genomic_accession=26462008, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102546294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546294, CurrentID=0, Status='live', Symbol='LOC102546294', Aliases='', description='uncharacterized LOC102546294', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148268307, end_position_on_the_genomic_accession=148383850, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  441208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441208, CurrentID=0, Status='live', Symbol='ZNRF2P1', Aliases='', description='zinc and ring finger 2 pseudogene 1', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32727950, end_position_on_the_genomic_accession=32729983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102546298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546298, CurrentID=0, Status='live', Symbol='LOC102546298', Aliases='', description='uncharacterized LOC102546298', other_designations='CTC-367J11.1', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150475516, end_position_on_the_genomic_accession=150485992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102546299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102546299, CurrentID=0, Status='live', Symbol='LOC102546299', Aliases='', description='uncharacterized LOC102546299', other_designations='CTC-340A15.2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=164470279, end_position_on_the_genomic_accession=164542984, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441212, CurrentID=0, Status='live', Symbol='RP9P', Aliases='', description='retinitis pigmentosa 9 pseudogene', other_designations='PNAS-13', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32916815, end_position_on_the_genomic_accession=32943170, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  441218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441218, CurrentID=0, Status='live', Symbol='KRT8P20', Aliases='', description='keratin 8 pseudogene 20', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38655144, end_position_on_the_genomic_accession=38656862, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441228, CurrentID=0, Status='live', Symbol='LOC441228', Aliases='', description='exportin, tRNA (nuclear export receptor for tRNAs) pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56287787, end_position_on_the_genomic_accession=56291981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189162, CurrentID=0, Status='live', Symbol='TRY-GTA3-1', Aliases='TRNAY7', description='transfer RNA-Tyr (GTA) 3-1', other_designations='transfer RNA tyrosine 7 (anticodon GUA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26577104, end_position_on_the_genomic_accession=26577192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441234, CurrentID=0, Status='live', Symbol='ZNF716', Aliases='', description='zinc finger protein 716', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57450177, end_position_on_the_genomic_accession=57473559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422041, CurrentID=0, Status='live', Symbol='LOC100422041', Aliases='', description='olfactory receptor, family 5, subfamily AC, member 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441238, CurrentID=0, Status='live', Symbol='NMD3P1', Aliases='NMD3P', description='NMD3 ribosome export adaptor pseudogene 1', other_designations='NMD3 homolog pseudogene 1|nonsense-mediated mRNA 3 homolog pseudogene 1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63908565, end_position_on_the_genomic_accession=63910561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441239, CurrentID=0, Status='live', Symbol='LOC441239', Aliases='', description='uncharacterized LOC441239', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64887751, end_position_on_the_genomic_accession=64890100, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  441241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441241, CurrentID=0, Status='live', Symbol='LOC441241', Aliases='', description='vitamin K epoxide reductase complex subunit 1 like 1 pseudogene', other_designations='vitamin K epoxide reductase complex, subunit 1-like 1 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65038265, end_position_on_the_genomic_accession=65038569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874392, CurrentID=0, Status='live', Symbol='ANKRD20A12P', Aliases='', description='ankyrin repeat domain 20 family member A12, pseudogene', other_designations='ankyrin repeat domain 20 family, member A pseudogene|ankyrin repeat domain 20 family, member A12, pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441246, CurrentID=0, Status='live', Symbol='RPL35P5', Aliases='RPL35_4_812', description='ribosomal protein L35 pseudogene 5', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66606703, end_position_on_the_genomic_accession=66607157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441250, CurrentID=0, Status='live', Symbol='TYW1B', Aliases='LINC00069, NCRNA00069, RSAFD2', description='tRNA-yW synthesizing protein 1 homolog B', other_designations='long intergenic non-protein coding RNA 69|radical S-adenosyl methionine and flavodoxin domain-containing protein 2|radical S-adenosyl methionine and flavodoxin domains 1|tRNA wybutosine-synthesizing protein 1 homolog B|tRNA-yW synthesizing protein 1 homolog B (non-protein coding)', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72574507, end_position_on_the_genomic_accession=72828234, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  441251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441251, CurrentID=0, Status='live', Symbol='SPDYE7P', Aliases='', description='speedy/RINGO cell cycle regulator family member E7, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E7, pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72862757, end_position_on_the_genomic_accession=72869091, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  441259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441259, CurrentID=0, Status='live', Symbol='LOC441259', Aliases='', description='PMS1 homolog 2, mismatch repair system component pseudogene', other_designations='postmeiotic segregation increased 2 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75295840, end_position_on_the_genomic_accession=75299895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374213, CurrentID=0, Status='live', Symbol='LOC105374213', Aliases='', description='uncharacterized LOC105374213', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170683121, end_position_on_the_genomic_accession=170754724, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  441261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441261, CurrentID=0, Status='live', Symbol='FDPSP7', Aliases='FDPSL2B', description='farnesyl diphosphate synthase pseudogene 7', other_designations='farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene|farnesyl diphosphate synthase-like 2B (farnesyl pyrophosphate synthetase-like 2B)|farnesyl pyrophosphate synthetase like', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76968136, end_position_on_the_genomic_accession=76969310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441263, CurrentID=0, Status='live', Symbol='DTX2P1-UPK3BP1-PMS2P11', Aliases='PMS2L11, PMS2P11, PMSR6', description='DTX2P1-UPK3BP1-PMS2P11 readthrough transcribed pseudogene', other_designations='DTX2P1-UPK3BP1-PMS2P11 readthrough (non-protein coding)|PMS2-DTX2-UPK3B pseudogene|deltex homolog 2 pseudogene|postmeiotic segregation increased 2 pseudogene 11|postmeiotic segregation increased 2-like 11 pseudogene|yeast mismatch repair gene PMS1 homologue', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76980822, end_position_on_the_genomic_accession=77023761, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  441268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441268, CurrentID=0, Status='live', Symbol='LOC441268', Aliases='', description='uncharacterized LOC441268', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8009, CurrentID=0, Status='live', Symbol='LALL', Aliases='', description='Lymphomatous acute lymphoblastic leukemia', other_designations='', map_location='9p22-p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=247640, no_hdr0=''),
  441272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441272, CurrentID=0, Status='live', Symbol='SPDYE3', Aliases='', description='speedy/RINGO cell cycle regulator family member E3', other_designations='Putative WBSCR19-like protein 7|putative Speedy protein E3|speedy homolog E3', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100306812, end_position_on_the_genomic_accession=100322196, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  441273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441273, CurrentID=0, Status='live', Symbol='SPDYE2', Aliases='SPDYB2-L1', description='speedy/RINGO cell cycle regulator family member E2', other_designations='Speedy B2-like 1|putative Speedy protein E2|speedy homolog E2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102550749, end_position_on_the_genomic_accession=102562633, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  441282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441282, CurrentID=0, Status='live', Symbol='AKR1B15', Aliases='AKR1B10L, AKR1B10L, AK1R1B7, AKR1R1B7', description='aldo-keto reductase family 1, member B15', other_designations='estradiol 17-beta-dehydrogenase AKR1B15|putative aldo-keto reductase family 1 member B15', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134546245, end_position_on_the_genomic_accession=134579861, orientation='plus', exon_count=12, OMIM=616336, no_hdr0=''),
  441294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441294, CurrentID=0, Status='live', Symbol='CTAGE15', Aliases='CTAGE15P, cTAGE-15', description='CTAGE family member 15', other_designations='CTAGE family, member 15|CTAGE family, member 15, pseudogene|protein cTAGE-15', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143571801, end_position_on_the_genomic_accession=143574387, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441295, CurrentID=0, Status='live', Symbol='OR2A9P', Aliases='FKSG35, HSDJ0798C17, OR2A19, OR2A22P, OR2A9', description='olfactory receptor family 2 subfamily A member 9 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 19|olfactory receptor, family 2, subfamily A, member 22 pseudogene|olfactory receptor, family 2, subfamily A, member 9 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144299521, end_position_on_the_genomic_accession=144300505, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724772, CurrentID=0, Status='live', Symbol='LOC102724772', Aliases='', description='uncharacterized LOC102724772', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71547632, end_position_on_the_genomic_accession=71553798, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  441307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441307, CurrentID=0, Status='live', Symbol='HRAT92', Aliases='', description='heart tissue-associated transcript 92', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=520391, end_position_on_the_genomic_accession=525232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441308, CurrentID=0, Status='live', Symbol='OR4F21', Aliases='OR4F21P', description='olfactory receptor family 4 subfamily F member 21', other_designations='olfactory receptor, family 4, subfamily F, member 21 pseudogene', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=166086, end_position_on_the_genomic_accession=167024, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724773, CurrentID=0, Status='live', Symbol='LOC102724773', Aliases='', description='N-acetyllactosaminide alpha-1,3-galactosyltransferase-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121444988, end_position_on_the_genomic_accession=121453628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441313, CurrentID=0, Status='live', Symbol='FAM90A4P', Aliases='FAM90A4', description='family with sequence similarity 90 member A4, pseudogene', other_designations='family with sequence similarity 90, member A4, pseudogene|family with sequence similarity 90, member A8 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7270488, end_position_on_the_genomic_accession=7275720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441314, CurrentID=0, Status='live', Symbol='FAM90A13P', Aliases='FAM90A13', description='family with sequence similarity 90 member A13, pseudogene', other_designations='family with sequence similarity 90, member A13, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7278110, end_position_on_the_genomic_accession=7283342, orientation='plus', exon_count=0, OMIM=613049, no_hdr0=''),
  441315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441315, CurrentID=0, Status='live', Symbol='FAM90A5P', Aliases='FAM90A5', description='family with sequence similarity 90 member A5, pseudogene', other_designations='family with sequence similarity 90, member A5, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7285732, end_position_on_the_genomic_accession=7290964, orientation='plus', exon_count=0, OMIM=613043, no_hdr0=''),
  441317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441317, CurrentID=0, Status='live', Symbol='FAM90A7P', Aliases='FAM90A7', description='family with sequence similarity 90 member A7, pseudogene', other_designations='family with sequence similarity 90, member A7, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7556138, end_position_on_the_genomic_accession=7559104, orientation='minus', exon_count=3, OMIM=613044, no_hdr0=''),
  441320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441320, CurrentID=0, Status='live', Symbol='LOC441320', Aliases='', description='chromosome 11 open reading frame2 pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7238286, end_position_on_the_genomic_accession=7238656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441323, CurrentID=0, Status='live', Symbol='FAM90A16P', Aliases='FAM90A16', description='family with sequence similarity 90 member A16, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7729077, end_position_on_the_genomic_accession=7734311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441324, CurrentID=0, Status='live', Symbol='FAM90A8P', Aliases='FAM90A8', description='family with sequence similarity 90 member A8, pseudogene', other_designations='family with sequence similarity 90, member A8, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7736725, end_position_on_the_genomic_accession=7741958, orientation='plus', exon_count=0, OMIM=613045, no_hdr0=''),
  441326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441326, CurrentID=0, Status='live', Symbol='FAM90A18P', Aliases='FAM90A18', description='family with sequence similarity 90 member A18, pseudogene', other_designations='family with sequence similarity 90, member A18, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7721429, end_position_on_the_genomic_accession=7726663, orientation='plus', exon_count=0, OMIM=613052, no_hdr0=''),
  441327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441327, CurrentID=0, Status='live', Symbol='FAM90A9P', Aliases='FAM90A9', description='family with sequence similarity 90 member A9, pseudogene', other_designations='family with sequence similarity 90, member A9, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7759668, end_position_on_the_genomic_accession=7764902, orientation='plus', exon_count=0, OMIM=613046, no_hdr0=''),
  441328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441328, CurrentID=0, Status='live', Symbol='FAM90A10P', Aliases='FAM90A10', description='family with sequence similarity 90 member A10, pseudogene', other_designations='family with sequence similarity 90, member A10, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7769584, end_position_on_the_genomic_accession=7771313, orientation='plus', exon_count=3, OMIM=613047, no_hdr0=''),
  441331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441331, CurrentID=0, Status='live', Symbol='FAM90A11P', Aliases='FAM90A11', description='family with sequence similarity 90 member A11, pseudogene', other_designations='family with sequence similarity 90, member A11, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8011782, end_position_on_the_genomic_accession=8017012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441332, CurrentID=0, Status='live', Symbol='FAM90A24P', Aliases='FAM90A24', description='family with sequence similarity 90 member A24, pseudogene', other_designations='family with sequence similarity 90, member A24, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8019426, end_position_on_the_genomic_accession=8024658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285296, CurrentID=0, Status='live', Symbol='UNC93B3', Aliases='UNC93B3P', description='unc-93 homolog B3 pseudogene (C. elegans)', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75619040, end_position_on_the_genomic_accession=75623850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401237, CurrentID=0, Status='live', Symbol='CASC15', Aliases='LINC00340', description='cancer susceptibility candidate 15 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 340', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21666444, end_position_on_the_genomic_accession=22194400, orientation='plus', exon_count=12, OMIM=616610, no_hdr0=''),
  441347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441347, CurrentID=0, Status='live', Symbol='LOC441347', Aliases='', description='hCG1795560', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=43100501, end_position_on_the_genomic_accession=43100665, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441355, CurrentID=0, Status='live', Symbol='MIR2052HG', Aliases='', description='MIR2052 host gene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74599866, end_position_on_the_genomic_accession=74758352, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  441359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441359, CurrentID=0, Status='live', Symbol='REXO1L3P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 3, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 3, pseudogene|REX1, RNA exonuclease 1 homolog-like 3 (pseudogene)', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85644388, end_position_on_the_genomic_accession=85656660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441361, CurrentID=0, Status='live', Symbol='REXO1L5P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 5, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 5, pseudogene|REX1, RNA exonuclease 1 homolog-like 5 (pseudogene)', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85804251, end_position_on_the_genomic_accession=85816478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441362, CurrentID=0, Status='live', Symbol='REXO1L6P', Aliases='REXO1L7P', description='REX1, RNA exonuclease 1 homolog-like 6, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 6, pseudogene|REX1, RNA exonuclease 1 homolog-like 7 (pseudogene)|REX1, RNA exonuclease 1 homolog-like 7, pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85826704, end_position_on_the_genomic_accession=85827839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441369, CurrentID=0, Status='live', Symbol='FLJ46284', Aliases='', description='uncharacterized LOC441369', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92712962, end_position_on_the_genomic_accession=92786060, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401242, CurrentID=0, Status='live', Symbol='LINC01623', Aliases='', description='long intergenic non-protein coding RNA 1623', other_designations='XXbac-BPG308K3.6', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28859625, end_position_on_the_genomic_accession=28863677, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  441374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441374, CurrentID=0, Status='live', Symbol='FLJ42969', Aliases='', description='uncharacterized LOC441374', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101052054, end_position_on_the_genomic_accession=101076251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441376, CurrentID=0, Status='live', Symbol='AARD', Aliases='C8orf85', description='alanine and arginine rich domain containing protein', other_designations='', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116938225, end_position_on_the_genomic_accession=116944000, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  441377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441377, CurrentID=0, Status='live', Symbol='RPS26P35', Aliases='RPS26_17_946', description='ribosomal protein S26 pseudogene 35', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118761774, end_position_on_the_genomic_accession=118762232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441381, CurrentID=0, Status='live', Symbol='LRRC24', Aliases='LRRC14OS', description='leucine rich repeat containing 24', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144522377, end_position_on_the_genomic_accession=144527032, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724785, CurrentID=0, Status='live', Symbol='LOC102724785', Aliases='', description='uncharacterized LOC102724785', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=156112591, end_position_on_the_genomic_accession=156143879, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  441386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441386, CurrentID=0, Status='live', Symbol='RPS26P3', Aliases='RPS26_11_962, bA338L20.1', description='ribosomal protein S26 pseudogene 3', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=9090874, end_position_on_the_genomic_accession=9091312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441389, CurrentID=0, Status='live', Symbol='LINC01239', Aliases='', description='long intergenic non-protein coding RNA 1239', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22646200, end_position_on_the_genomic_accession=22824213, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  441391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441391, CurrentID=0, Status='live', Symbol='RBMXP2', Aliases='', description='RNA binding motif protein, X-linked pseudogene 2', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30688948, end_position_on_the_genomic_accession=30690960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441394, CurrentID=0, Status='live', Symbol='SUGT1P1', Aliases='SGT1P, SUGT1P, bA255A11.1', description='SGT1 homolog, MIS12 kinetochore complex assembly cochaperone pseudogene 1', other_designations='SUGT1 pseudogene 1|bA255A11.10 (putative novel T cell receptor beta chain V region protein)|suppressor of G2 allele of SKP1 pseudogene 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33500756, end_position_on_the_genomic_accession=33511166, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  441400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441400, CurrentID=0, Status='live', Symbol='RAB1C', Aliases='', description='RAB1C, member RAS oncogene family pseudogene', other_designations='RAB1B, member RAS oncogene family pseudogene|small GTP-binding protein pseudogene', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37636688, end_position_on_the_genomic_accession=37637293, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  8031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8031, CurrentID=0, Status='live', Symbol='NCOA4', Aliases='ARA70, ELE1, PTC3, RFG', description='nuclear receptor coactivator 4', other_designations='70 kDa AR-activator|70 kDa androgen receptor coactivator|NCoA-4|RET-activating gene ELE1|androgen receptor-associated protein of 70 kDa|ret fused', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46005088, end_position_on_the_genomic_accession=46030714, orientation='minus', exon_count=14, OMIM=601984, no_hdr0=''),
  441410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441410, CurrentID=0, Status='live', Symbol='FKBP4P7', Aliases='FKBP4P3', description='FK506 binding protein 4 pseudogene 7', other_designations='FK506 binding protein 4, 59kDa pseudogene 3|FK506 binding protein 4, 59kDa pseudogene 7', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62709936, end_position_on_the_genomic_accession=62711538, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441416, CurrentID=0, Status='live', Symbol='ADGRF5P2', Aliases='GPR116P2', description='adhesion G protein-coupled receptor F5 pseudogene 2', other_designations='G protein-coupled receptor 116 pseudogene 2', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42795392, end_position_on_the_genomic_accession=42796797, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441420, CurrentID=0, Status='live', Symbol='MYO5BP1', Aliases='', description='myosin VB pseudogene 1', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41286820, end_position_on_the_genomic_accession=41294996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401250, CurrentID=0, Status='live', Symbol='MCCD1', Aliases='', description='mitochondrial coiled-coil domain 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31528962, end_position_on_the_genomic_accession=31530231, orientation='plus', exon_count=2, OMIM=609624, no_hdr0=''),
  441425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441425, CurrentID=0, Status='live', Symbol='ANKRD20A3', Aliases='', description='ankyrin repeat domain 20 family member A3', other_designations='ankyrin repeat domain 20 family, member A3|ankyrin repeat domain 20A-related', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66104344, end_position_on_the_genomic_accession=66153187, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  441426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441426, CurrentID=0, Status='live', Symbol='LINC00268', Aliases='NCRNA00268, NCRNA00268A', description='long intergenic non-protein coding RNA 268', other_designations='NCRNA00268-1', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441430, CurrentID=0, Status='live', Symbol='ANKRD20A2', Aliases='', description='ankyrin repeat domain 20 family member A2', other_designations='ankyrin repeat domain 20 family, member A2', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40223078, end_position_on_the_genomic_accession=40282461, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  441432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441432, CurrentID=0, Status='live', Symbol='AQP7P3', Aliases='AQPap-3', description='aquaporin 7 pseudogene 3', other_designations='aquaporin pseudogene AQPap-3', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65877574, end_position_on_the_genomic_accession=65912560, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441440, CurrentID=0, Status='live', Symbol='RPL7AP46', Aliases='RPL7A_20_981', description='ribosomal protein L7a pseudogene 46', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63581254, end_position_on_the_genomic_accession=63581855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441441, CurrentID=0, Status='live', Symbol='FLJ41455', Aliases='', description='uncharacterized LOC441441', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441442, CurrentID=0, Status='live', Symbol='MYO5BP3', Aliases='', description='myosin VB pseudogene 3', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63761544, end_position_on_the_genomic_accession=63762826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441452, CurrentID=0, Status='live', Symbol='SPATA31C1', Aliases='FAM75C1, SPATA31C2', description='SPATA31 subfamily C, member 1', other_designations='family with sequence similarity 75, member C1|spermatogenesis-associated protein 31C2', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87913936, end_position_on_the_genomic_accession=87923657, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  441454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441454, CurrentID=0, Status='live', Symbol='LOC441454', Aliases='', description='prothymosin, alpha pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96909075, end_position_on_the_genomic_accession=96910455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441455, CurrentID=0, Status='live', Symbol='LOC441455', Aliases='', description='makorin ring finger protein 1 pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96725821, end_position_on_the_genomic_accession=96727467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441457, CurrentID=0, Status='live', Symbol='NUTM2G', Aliases='FAM22G, NUTMG', description='NUT family member 2G', other_designations='family with sequence similarity 22, member G|family with sequence similarity 22, pseudogene|protein FAM22G', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96929004, end_position_on_the_genomic_accession=96942290, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  441459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441459, CurrentID=0, Status='live', Symbol='ANKRD18B', Aliases='bA255A11.3, bA255A11.5', description='ankyrin repeat domain 18B', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33524353, end_position_on_the_genomic_accession=33575969, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  441461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441461, CurrentID=0, Status='live', Symbol='STX17-AS1', Aliases='', description='STX17 antisense RNA 1', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99886317, end_position_on_the_genomic_accession=99906601, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  554175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554175, CurrentID=0, Status='live', Symbol='DNM1P33', Aliases='DNM1DN7-2, DNM1DN7@', description='dynamin 1 pseudogene 33', other_designations='DNM1 pseudogene 33|dynamin-related', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74063689, end_position_on_the_genomic_accession=74065758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441476, CurrentID=0, Status='live', Symbol='C9orf173', Aliases='', description='chromosome 9 open reading frame 173', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137250840, end_position_on_the_genomic_accession=137253483, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  441478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441478, CurrentID=0, Status='live', Symbol='NRARP', Aliases='', description='NOTCH-regulated ankyrin repeat protein', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137299631, end_position_on_the_genomic_accession=137302251, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441481, CurrentID=0, Status='live', Symbol='GPX1P1', Aliases='GPXL1, GPXL2, GPXP1', description='glutathione peroxidase pseudogene 1', other_designations='glutathione peroxidase 1 pseudogene', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13378515, end_position_on_the_genomic_accession=13379387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441484, CurrentID=0, Status='live', Symbol='HAUS1P2', Aliases='', description='HAUS augmin like complex subunit 1 pseudogene 2', other_designations='HAUS augmin-like complex, subunit 1 pseudogene 2', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18865370, end_position_on_the_genomic_accession=18866751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  572558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=572558, CurrentID=0, Status='live', Symbol='PGM5-AS1', Aliases='FAM233A', description='PGM5 antisense RNA 1', other_designations='PGM5 antisense RNA 1 (non-protein coding)|family with sequence similarity 233, member A', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68355164, end_position_on_the_genomic_accession=68357866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401261, CurrentID=0, Status='live', Symbol='LOC401261', Aliases='', description='zinc finger protein ZIC 5', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42782927, end_position_on_the_genomic_accession=42785561, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441488, CurrentID=0, Status='live', Symbol='LOC441488', Aliases='', description='transcription factor Dp-1-like pseudogene', other_designations='hCG1982709', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36961001, end_position_on_the_genomic_accession=36963558, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441490, CurrentID=0, Status='live', Symbol='FTH1P18', Aliases='FTHL18', description='ferritin, heavy polypeptide 1 pseudogene 18', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37042882, end_position_on_the_genomic_accession=37043794, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441495, CurrentID=0, Status='live', Symbol='CENPVP2', Aliases='', description='centromere protein V pseudogene 2', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51681212, end_position_on_the_genomic_accession=51682514, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441501, CurrentID=0, Status='live', Symbol='LINC00891', Aliases='', description='long intergenic non-protein coding RNA 891', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71697196, end_position_on_the_genomic_accession=71703406, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441502, CurrentID=0, Status='live', Symbol='RPS26P11', Aliases='RPS26L1, RPS26_22_1784, bA366E13.1', description='ribosomal protein S26 pseudogene 11', other_designations='ribosomal protein S26-like 1', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72044409, end_position_on_the_genomic_accession=72044974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  401264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401264, CurrentID=0, Status='live', Symbol='TRAM2-AS1', Aliases='', description='TRAM2 antisense RNA 1 (head to head)', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52577202, end_position_on_the_genomic_accession=52583993, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  441509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441509, CurrentID=0, Status='live', Symbol='GLRA4', Aliases='', description='glycine receptor alpha 4', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103707344, end_position_on_the_genomic_accession=103728624, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  441511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441511, CurrentID=0, Status='live', Symbol='LOC441511', Aliases='', description='mitofusin-1-like', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109568478, end_position_on_the_genomic_accession=109570289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441518, CurrentID=0, Status='live', Symbol='FAM127C', Aliases='CXX1c, MAR8B', description='family with sequence similarity 127 member C', other_designations='family with sequence similarity 127, member C|mammalian retrotransposon derived protein 8B', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135020504, end_position_on_the_genomic_accession=135022536, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441519, CurrentID=0, Status='live', Symbol='CT45A3', Aliases='CT45-3, CT45.3', description='cancer/testis antigen family 45, member A3', other_designations='cancer/testis antigen 45-3|cancer/testis antigen 45A3|cancer/testis antigen CT45-3', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135760067, end_position_on_the_genomic_accession=135777544, orientation='minus', exon_count=7, OMIM=300794, no_hdr0=''),
  441520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441520, CurrentID=0, Status='live', Symbol='CT45A4', Aliases='CT45-4, CT45.4', description='cancer/testis antigen family 45, member A4', other_designations='cancer/testis antigen 45-4|cancer/testis antigen 45-4/6|cancer/testis antigen 45A4|cancer/testis antigen 45A4/45A6|cancer/testis antigen CT45-4|cancer/testis antigen family 45 member A4/A6', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300795, no_hdr0=''),
  441521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441521, CurrentID=0, Status='live', Symbol='CT45A5', Aliases='CT45.5, CT455', description='cancer/testis antigen family 45, member A5', other_designations='cancer/testis antigen 45-5|cancer/testis antigen 45A5|cancer/testis antigen CT45-5', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135777130, end_position_on_the_genomic_accession=135794806, orientation='minus', exon_count=6, OMIM=300796, no_hdr0=''),
  441525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441525, CurrentID=0, Status='live', Symbol='SPANXN4', Aliases='CT11.9', description='SPANX family member N4', other_designations='SPANX family, member N4|cancer/testis antigen family 11, member 9|nuclear-associated protein SPAN-Xn4', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143025918, end_position_on_the_genomic_accession=143034280, orientation='plus', exon_count=2, OMIM=300667, no_hdr0=''),
  441528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441528, CurrentID=0, Status='live', Symbol='LOC441528', Aliases='', description='zinc finger protein 839-like', other_designations='MHC class I antigen', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441531, CurrentID=0, Status='live', Symbol='PGAM4', Aliases='PGAM-B, PGAM1, PGAM3, dJ1000K24.1', description='phosphoglycerate mutase family member 4', other_designations='phosphoglycerate mutase family 3|phosphoglycerate mutase processed protein|probable phosphoglycerate mutase 4', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77967961, end_position_on_the_genomic_accession=77969638, orientation='minus', exon_count=1, OMIM=300567, no_hdr0=''),
  441533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441533, CurrentID=0, Status='live', Symbol='RPL26P37', Aliases='RPL26_17_1822', description='ribosomal protein L26 pseudogene 37', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5793257, end_position_on_the_genomic_accession=5793835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441543, CurrentID=0, Status='live', Symbol='TTTY6B', Aliases='LINC00128, NCRNA00128, TTTY6, TTY6', description='testis-specific transcript, Y-linked 6B (non-protein coding)', other_designations='long intergenic non-protein coding RNA 128|testis-specific Y-linked 6, centromeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22144987, end_position_on_the_genomic_accession=22146834, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441548, CurrentID=0, Status='live', Symbol='BTBD7P1', Aliases='', description='BTB (POZ) domain containing 7 pseudogene 1', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13156681, end_position_on_the_genomic_accession=13158136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441549, CurrentID=0, Status='live', Symbol='CDNF', Aliases='ARMETL1', description='cerebral dopamine neurotrophic factor', other_designations='ARMET-like protein 1|arginine-rich, mutated in early stage tumors-like 1|conserved dopamine neurotrophic factor', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14819245, end_position_on_the_genomic_accession=14838098, orientation='minus', exon_count=4, OMIM=611233, no_hdr0=''),
  441550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441550, CurrentID=0, Status='live', Symbol='RPS4XP11', Aliases='RPS4P11, RPS4X_2_1029', description='ribosomal protein S4X pseudogene 11', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32102453, end_position_on_the_genomic_accession=32103341, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441554, CurrentID=0, Status='live', Symbol='DUSP8P3', Aliases='DUSP8P, DUSP8P1, DUSPP, HVH4', description='dual specificity phosphatase 8 pseudogene 3', other_designations='dual specificity phosphatase 8 pseudogene 1|dual specificity phosphatase pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47564128, end_position_on_the_genomic_accession=47567737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533970, CurrentID=0, Status='live', Symbol='P2RX5-TAX1BP3', Aliases='', description='P2RX5-TAX1BP3 readthrough (NMD candidate)', other_designations='', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3662893, end_position_on_the_genomic_accession=3696404, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  441572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441572, CurrentID=0, Status='live', Symbol='GAPDHP28', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 28', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91666581, end_position_on_the_genomic_accession=91667852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369859, CurrentID=0, Status='live', Symbol='LOC105369859', Aliases='', description='uncharacterized LOC105369859', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78335585, end_position_on_the_genomic_accession=78352674, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441581, CurrentID=0, Status='live', Symbol='FRG2B', Aliases='', description='FSHD region gene 2 family member B', other_designations='FSHD region gene 2 family, member B|FSHD region gene 2 protein family member B|HSA10-FRG2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133624796, end_position_on_the_genomic_accession=133627886, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  441584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441584, CurrentID=0, Status='live', Symbol='SSU72P4', Aliases='', description='SSU72 pseudogene 4', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4287499, end_position_on_the_genomic_accession=4288093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374224, CurrentID=0, Status='live', Symbol='LOC105374224', Aliases='', description='uncharacterized LOC105374224', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173259701, end_position_on_the_genomic_accession=173268650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  441592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441592, CurrentID=0, Status='live', Symbol='LOC441592', Aliases='', description='MAS related GPR family member X1 pseudogene', other_designations='MAS-related GPR, member X1 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18863546, end_position_on_the_genomic_accession=18864724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441601, CurrentID=0, Status='live', Symbol='LOC441601', Aliases='', description='septin 7 pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50279828, end_position_on_the_genomic_accession=50298462, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  401281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401281, CurrentID=0, Status='live', Symbol='LOC401281', Aliases='', description='uncharacterized LOC401281', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159588439, end_position_on_the_genomic_accession=159589980, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441608, CurrentID=0, Status='live', Symbol='OR5B3', Aliases='OR11-239, OR5B13, OST129', description='olfactory receptor family 5 subfamily B member 3', other_designations='olfactory receptor 5B13|olfactory receptor OR11-239|olfactory receptor, family 5, subfamily B, member 13|olfactory receptor, family 5, subfamily B, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58402465, end_position_on_the_genomic_accession=58403409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  441609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441609, CurrentID=0, Status='live', Symbol='YWHAZP9', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 9', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60077228, end_position_on_the_genomic_accession=60080571, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401282, CurrentID=0, Status='live', Symbol='DKFZp451B082', Aliases='', description='uncharacterized LOC401282', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163338342, end_position_on_the_genomic_accession=163347033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100189175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189175, CurrentID=0, Status='live', Symbol='TRL-TAA5-1', Aliases='TRNAL18', description='transfer RNA-Leu (TAA) 5-1', other_designations='tRNA-Leu-TAA-5-1|transfer RNA leucine 18 (anticodon UAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29914, CurrentID=0, Status='live', Symbol='UBIAD1', Aliases='SCCD, TERE1', description='UbiA prenyltransferase domain containing 1', other_designations='transitional epithelia response protein|transitional epithelial response protein 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11273198, end_position_on_the_genomic_accession=11296595, orientation='plus', exon_count=4, OMIM=611632, no_hdr0=''),
  441630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441630, CurrentID=0, Status='live', Symbol='BRI3P2', Aliases='', description='brain protein I3 pseudogene 2', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24967127, end_position_on_the_genomic_accession=24967504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  441631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441631, CurrentID=0, Status='live', Symbol='TSPAN11', Aliases='VSSW1971', description='tetraspanin 11', other_designations='tspan-11', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30926449, end_position_on_the_genomic_accession=31016490, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  441632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441632, CurrentID=0, Status='live', Symbol='RPL13AP22', Aliases='RPL13A_9_1223', description='ribosomal protein L13a pseudogene 22', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31251913, end_position_on_the_genomic_accession=31253807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  29915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29915, CurrentID=0, Status='live', Symbol='HCFC2', Aliases='HCF-2, HCF2', description='host cell factor C2', other_designations='C2 factor', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104064458, end_position_on_the_genomic_accession=104106525, orientation='plus', exon_count=17, OMIM=607926, no_hdr0=''),
  441639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441639, CurrentID=0, Status='live', Symbol='OR9K2', Aliases='OR12-2', description='olfactory receptor family 9 subfamily K member 2', other_designations='olfactory receptor OR12-2|olfactory receptor, family 9, subfamily K, member 2', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55129769, end_position_on_the_genomic_accession=55130776, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441641, CurrentID=0, Status='live', Symbol='RPL13AP23', Aliases='RPL13A_10_1240', description='ribosomal protein L13a pseudogene 23', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57674665, end_position_on_the_genomic_accession=57675250, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441642, CurrentID=0, Status='live', Symbol='RPL7AP9', Aliases='RPL7AP59, RPL7A_24_1258, rpL7a', description='ribosomal protein L7a pseudogene 9', other_designations='ribosomal protein L7a pseudogene 59', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76599652, end_position_on_the_genomic_accession=76600529, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401287, CurrentID=0, Status='live', Symbol='LOC401287', Aliases='', description='uncharacterized LOC401287', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169433369, end_position_on_the_genomic_accession=169453720, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  441644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441644, CurrentID=0, Status='live', Symbol='LOC441644', Aliases='', description='REST corepressor 2 pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94139936, end_position_on_the_genomic_accession=94141988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  441654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441654, CurrentID=0, Status='live', Symbol='SMPD4P2', Aliases='', description='sphingomyelin phosphodiesterase 4 pseudogene 2', other_designations='neutral sphingomyelinase 3 pseudogene|sphingomyelin phosphodiesterase 4, neutral membrane (neutral sphingomyelinase-3) pseudogene 2', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19151467, end_position_on_the_genomic_accession=19171401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401289, CurrentID=0, Status='live', Symbol='LINC01624', Aliases='', description='long intergenic non-protein coding RNA 1624', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170272474, end_position_on_the_genomic_accession=170279469, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  444980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=444980, CurrentID=0, Status='live', Symbol='HYT4', Aliases='', description='Hypertension, essential, susceptibility to, 4', other_designations='', map_location='12p12.2-p12.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608742, no_hdr0=''),
  554207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554207, CurrentID=0, Status='live', Symbol='LOC554207', Aliases='', description='uncharacterized LOC554207', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=21584645, end_position_on_the_genomic_accession=21584773, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106478909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478909, CurrentID=0, Status='live', Symbol='MTND4LP1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4L pseudogene 1', other_designations='MT-ND4L pseudogene 1', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478910, CurrentID=0, Status='live', Symbol='LAMTOR5P1', Aliases='HBXIPL, HBXIPP1', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 5 pseudogene 1', other_designations='hepatitis B virus x interacting protein-like', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478911, CurrentID=0, Status='live', Symbol='CST12P', Aliases='Cst', description='cystatin 12, pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478912, CurrentID=0, Status='live', Symbol='RNU6-41P', Aliases='RNU6-41', description='RNA, U6 small nuclear 41, pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478913, CurrentID=0, Status='live', Symbol='RNU6-5P', Aliases='RNU6-5', description='RNA, U6 small nuclear 5, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478914, CurrentID=0, Status='live', Symbol='RNU6-4P', Aliases='RNU6-4', description='RNA, U6 small nuclear 4, pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478915, CurrentID=0, Status='live', Symbol='RNU6-11P', Aliases='RNU6-11', description='RNA, U6 small nuclear 11, pseudogene', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478916, CurrentID=0, Status='live', Symbol='RNU6-18P', Aliases='RNU6-18', description='RNA, U6 small nuclear 18, pseudogene', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441669, CurrentID=0, Status='live', Symbol='OR4Q3', Aliases='C14orf13, HSA6, OR14-3, OR4Q4, c14_5008', description='olfactory receptor family 4 subfamily Q member 3', other_designations='olfactory receptor 4Q4|olfactory receptor OR14-3|olfactory receptor, family 4, subfamily Q, member 3|olfactory receptor, family 4, subfamily Q, member 4', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19747428, end_position_on_the_genomic_accession=19748369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  441670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441670, CurrentID=0, Status='live', Symbol='OR4M1', Aliases='OR14-7', description='olfactory receptor family 4 subfamily M member 1', other_designations='olfactory receptor OR14-7|olfactory receptor, family 4, subfamily M, member 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19780323, end_position_on_the_genomic_accession=19781264, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106478919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478919, CurrentID=0, Status='live', Symbol='RNU6-3P', Aliases='RNU6-3', description='RNA, U6 small nuclear 3, pseudogene', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478920, CurrentID=0, Status='live', Symbol='RNU6-37P', Aliases='RNU6-37', description='RNA, U6 small nuclear 37, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478921, CurrentID=0, Status='live', Symbol='RNU6-43P', Aliases='RNU6-43', description='RNA, U6 small nuclear 43, pseudogene', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478922, CurrentID=0, Status='live', Symbol='RNU6-44P', Aliases='RNU6-44', description='RNA, U6 small nuclear 44, pseudogene', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478923, CurrentID=0, Status='live', Symbol='RNU6-49P', Aliases='RNU6-49', description='RNA, U6 small nuclear 49, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478924, CurrentID=0, Status='live', Symbol='DHRSX-IT1', Aliases='DHRSX-IT, DHRSXIT1', description='DHRSX intronic transcript 1', other_designations='', map_location='Xp22.33 and Yp11', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554210, CurrentID=0, Status='live', Symbol='MIR429', Aliases='MIRN429, hsa-mir-429, mir-429', description='microRNA 429', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1169005, end_position_on_the_genomic_accession=1169087, orientation='plus', exon_count=1, OMIM=612094, no_hdr0=''),
  106478926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478926, CurrentID=0, Status='live', Symbol='DPH3P2', Aliases='', description='diphthamide biosynthesis 3 pseudogene 2', other_designations='', map_location='Xq28 and Yq12', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478927, CurrentID=0, Status='live', Symbol='USP9YP16', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 16', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478928, CurrentID=0, Status='live', Symbol='USP9YP24', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 24', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478929, CurrentID=0, Status='live', Symbol='USP9YP26', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 26', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478930, CurrentID=0, Status='live', Symbol='USP9YP31', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 31', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478931, CurrentID=0, Status='live', Symbol='USP9YP33', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 33', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478932, CurrentID=0, Status='live', Symbol='USP9YP36', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 36', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478933, CurrentID=0, Status='live', Symbol='PRYP5', Aliases='', description='PTPN13-like, Y-linked pseudogene 5', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478934, CurrentID=0, Status='live', Symbol='HSFY5P', Aliases='', description='heat shock transcription factor, Y-linked 5, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441687, CurrentID=0, Status='live', Symbol='TEX21P', Aliases='', description='testis expressed 21, pseudogene', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64345473, end_position_on_the_genomic_accession=64347611, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106478936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478936, CurrentID=0, Status='live', Symbol='FTH1P9', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene 9', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29924, CurrentID=0, Status='live', Symbol='EPN1', Aliases='', description='epsin 1', other_designations='EH domain-binding mitotic phosphoprotein|EPS-15-interacting protein 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55675195, end_position_on_the_genomic_accession=55695767, orientation='plus', exon_count=13, OMIM=607262, no_hdr0=''),
  106478938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478938, CurrentID=0, Status='live', Symbol='OOEP-AS1', Aliases='OOEP-AS', description='OOEP antisense RNA 1', other_designations='OOEP antisense RNA (non-protein coding)', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478939, CurrentID=0, Status='live', Symbol='PHBP5', Aliases='', description='prohibitin pseudogene 5', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478940, CurrentID=0, Status='live', Symbol='PHBP6', Aliases='', description='prohibitin pseudogene 6', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478941, CurrentID=0, Status='live', Symbol='MTND6P2', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 2', other_designations='MT-ND6 pseudogene 2', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478942, CurrentID=0, Status='live', Symbol='MTND6P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 3', other_designations='MT-ND6 pseudogene 3', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554213, CurrentID=0, Status='live', Symbol='MIR449A', Aliases='MIRN449, MIRN449A, hsa-mir-449, mir-449a', description='microRNA 449a', other_designations='hsa-mir-449a|microRNA 449', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55170532, end_position_on_the_genomic_accession=55170622, orientation='minus', exon_count=1, OMIM=613131, no_hdr0=''),
  106478944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478944, CurrentID=0, Status='live', Symbol='MTND6P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 6', other_designations='MT-ND6 pseudogene 6', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478945, CurrentID=0, Status='live', Symbol='MTND6P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 7', other_designations='MT-ND6 pseudogene 7', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478946, CurrentID=0, Status='live', Symbol='MTND6P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 9', other_designations='MT-ND6 pseudogene 9', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478947, CurrentID=0, Status='live', Symbol='MTND6P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 10', other_designations='MT-ND6 pseudogene 10', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478948, CurrentID=0, Status='live', Symbol='MTND6P14', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 14', other_designations='MT-ND6 pseudogene 14', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29926, CurrentID=0, Status='live', Symbol='GMPPA', Aliases='AAMR', description='GDP-mannose pyrophosphorylase A', other_designations='GMPP-alpha|GTP-mannose-1-phosphate guanylyltransferase alpha|mannose-1-phosphate guanylyltransferase (GDP)', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219498865, end_position_on_the_genomic_accession=219506996, orientation='plus', exon_count=15, OMIM=615495, no_hdr0=''),
  106478950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478950, CurrentID=0, Status='live', Symbol='MTND6P16', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 16', other_designations='MT-ND6 pseudogene 16', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478951, CurrentID=0, Status='live', Symbol='MTND6P18', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 18', other_designations='MT-ND6 pseudogene 18', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478952, CurrentID=0, Status='live', Symbol='MTND6P20', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 20', other_designations='MT-ND6 pseudogene 20', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478953, CurrentID=0, Status='live', Symbol='RBM22P7', Aliases='RBM22P9', description='RNA binding motif protein 22 pseudogene 7', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478954, CurrentID=0, Status='live', Symbol='EHMT2-AS1', Aliases='', description='EHMT2 antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478955, CurrentID=0, Status='live', Symbol='C4A-AS1', Aliases='', description='C4A antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478956, CurrentID=0, Status='live', Symbol='MDC1-AS1', Aliases='', description='MDC1 antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478957, CurrentID=0, Status='live', Symbol='DDX39B-AS1', Aliases='', description='DDX39B antisense RNA 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478958, CurrentID=0, Status='live', Symbol='CYP4F26P', Aliases='', description='cytochrome P450 family 4 subfamily F member 26, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 26, pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478959, CurrentID=0, Status='live', Symbol='CYP2C56P', Aliases='', description='cytochrome P450 family 2 subfamily C member 56, pseudogene', other_designations='cytochrome P450, family 2, subfamily C, polypeptide 56, pseudogene', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478960, CurrentID=0, Status='live', Symbol='CYP2AC1P', Aliases='', description='cytochrome P450 family 2 subfamily AC member 1, pseudogene', other_designations='cytochrome P450, family 2, subfamily AC, polypeptide 1, pseudogene', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478961, CurrentID=0, Status='live', Symbol='ATP6V0CP2', Aliases='', description='ATPase, H+ transporting, lysosomal 16kDa, V0 subunit c pseudogene 2', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478962, CurrentID=0, Status='live', Symbol='BACH1-AS1', Aliases='', description='BACH1 antisense RNA 1', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478963, CurrentID=0, Status='live', Symbol='BSN-AS1', Aliases='', description='BSN antisense RNA 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478964, CurrentID=0, Status='live', Symbol='CACNA1C-AS3', Aliases='', description='CACNA1C antisense RNA 3', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478965, CurrentID=0, Status='live', Symbol='IGBP1-AS2', Aliases='', description='IGBP1 antisense RNA 2', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478966, CurrentID=0, Status='live', Symbol='RRM1-AS1', Aliases='', description='RRM1 antisense RNA 1', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478967, CurrentID=0, Status='live', Symbol='SEC62-AS1', Aliases='', description='SEC62 antisense RNA 1', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478968, CurrentID=0, Status='live', Symbol='TRPC7-AS2', Aliases='', description='TRPC7 antisense RNA 2', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478969, CurrentID=0, Status='live', Symbol='DPP10-AS2', Aliases='', description='DPP10 antisense RNA 2', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478970, CurrentID=0, Status='live', Symbol='HTR3E-AS1', Aliases='', description='HTR3E antisense RNA 1', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478971, CurrentID=0, Status='live', Symbol='ALG13-AS1', Aliases='', description='ALG13 antisense RNA 1', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478972, CurrentID=0, Status='live', Symbol='BRD2-IT1', Aliases='', description='BRD2 intronic transcript 1', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478973, CurrentID=0, Status='live', Symbol='FMR1-IT1', Aliases='FMR1', description='FMR1 intronic transcript 1', other_designations='fragile X mental retardation syndrome protein', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478974, CurrentID=0, Status='live', Symbol='REV3L-IT1', Aliases='', description='REV3L intronic transcript 1', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478975, CurrentID=0, Status='live', Symbol='SMYD3-IT1', Aliases='', description='SMYD3 intronic transcript 1', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478976, CurrentID=0, Status='live', Symbol='ASH1L-IT1', Aliases='', description='ASH1L intronic transcript 1', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478977, CurrentID=0, Status='live', Symbol='WWTR1-IT1', Aliases='', description='WWTR1 intronic transcript 1', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478978, CurrentID=0, Status='live', Symbol='FAR1-IT1', Aliases='', description='FAR1 intronic transcript 1', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  161003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=161003, CurrentID=0, Status='live', Symbol='STOML3', Aliases='Epb7.2l, SRO', description='stomatin like 3', other_designations='SLP-3|erythrocyte band 7 integral membrane protein|protein 7.2B|stomatin (EPB72)-like 3', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38965925, end_position_on_the_genomic_accession=38990859, orientation='minus', exon_count=8, OMIM=608327, no_hdr0=''),
  106478980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478980, CurrentID=0, Status='live', Symbol='RNU1-23P', Aliases='', description='RNA, U1 small nuclear 23, pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441733, CurrentID=0, Status='live', Symbol='PRKXP1', Aliases='', description='protein kinase, X-linked, pseudogene 1', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100547752, end_position_on_the_genomic_accession=100559283, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106478982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478982, CurrentID=0, Status='live', Symbol='TUBB8P2', Aliases='', description='tubulin beta 8 class VIII pseudogene 2', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478983, CurrentID=0, Status='live', Symbol='CYP3A52P', Aliases='', description='cytochrome P450 family 3 subfamily A member 52, pseudogene', other_designations='cytochrome P450, family 3, subfamily A, polypeptide 52, pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441736, CurrentID=0, Status='live', Symbol='GOLGA8VP', Aliases='', description='golgin A8 family member V, pseudogene', other_designations='golgin A8 family, member A pseudogene|golgin A8 family, member V, pseudogene', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101775281, end_position_on_the_genomic_accession=101779514, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106478985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478985, CurrentID=0, Status='live', Symbol='RNY3P8', Aliases='', description='RNA, Ro-associated Y3 pseudogene 8', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478986, CurrentID=0, Status='live', Symbol='RNU6-54P', Aliases='RNU6-54', description='RNA, U6 small nuclear 54, pseudogene', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401303, CurrentID=0, Status='live', Symbol='ZNF815P', Aliases='ZNF815', description='zinc finger protein 815, pseudogene', other_designations='zinc finger protein pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5823160, end_position_on_the_genomic_accession=5854435, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106478988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478988, CurrentID=0, Status='live', Symbol='RNU6-70P', Aliases='RNU6-70', description='RNA, U6 small nuclear 70, pseudogene', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100190786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190786, CurrentID=0, Status='live', Symbol='CILD8', Aliases='', description='Ciliary dyskinesia, primary, 8', other_designations='', map_location='15q24-q25', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612274, no_hdr0=''),
  106478990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478990, CurrentID=0, Status='live', Symbol='RNU6-80P', Aliases='RNU6-80', description='RNA, U6 small nuclear 80, pseudogene', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478991, CurrentID=0, Status='live', Symbol='LINC00449', Aliases='', description='long intergenic non-protein coding RNA 449', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441744, CurrentID=0, Status='live', Symbol='RPS16P7', Aliases='RPS16_4_1467', description='ribosomal protein S16 pseudogene 7', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2332247, end_position_on_the_genomic_accession=2332733, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106478993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478993, CurrentID=0, Status='live', Symbol='RNA5SP99', Aliases='RN5S99', description='RNA, 5S ribosomal pseudogene 99', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478994, CurrentID=0, Status='live', Symbol='RNA5SP106', Aliases='RN5S106', description='RNA, 5S ribosomal pseudogene 106', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478995, CurrentID=0, Status='live', Symbol='RNA5SP130', Aliases='RN5S130', description='RNA, 5S ribosomal pseudogene 130', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478996, CurrentID=0, Status='live', Symbol='RNA5SP140', Aliases='RN5S140', description='RNA, 5S ribosomal pseudogene 140', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478997, CurrentID=0, Status='live', Symbol='RNA5SP172', Aliases='RN5S172', description='RNA, 5S ribosomal pseudogene 172', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271033, CurrentID=0, Status='live', Symbol='RPL4P6', Aliases='RPL4_4_1752', description='ribosomal protein L4 pseudogene 6', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40586625, end_position_on_the_genomic_accession=40588023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106478999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478999, CurrentID=0, Status='live', Symbol='RNA5SP206', Aliases='RN5S206', description='RNA, 5S ribosomal pseudogene 206', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479000, CurrentID=0, Status='live', Symbol='RNA5SP226', Aliases='RN5S226', description='RNA, 5S ribosomal pseudogene 226', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100190788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190788, CurrentID=0, Status='live', Symbol='IBD20', Aliases='', description='Inflammatory bowel disease-20', other_designations='', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612288, no_hdr0=''),
  106479002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479002, CurrentID=0, Status='live', Symbol='RNA5SP263', Aliases='RN5S263', description='RNA, 5S ribosomal pseudogene 263', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554223, CurrentID=0, Status='live', Symbol='LOC554223', Aliases='', description='histocompatibility antigen-related', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29791906, end_position_on_the_genomic_accession=29797807, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479004, CurrentID=0, Status='live', Symbol='RNA5SP302', Aliases='RN5S302', description='RNA, 5S ribosomal pseudogene 302', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479005, CurrentID=0, Status='live', Symbol='RNA5SP315', Aliases='RN5S315', description='RNA, 5S ribosomal pseudogene 315', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479006, CurrentID=0, Status='live', Symbol='RNA5SP317', Aliases='RN5S317', description='RNA, 5S ribosomal pseudogene 317', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479007, CurrentID=0, Status='live', Symbol='RNA5SP335', Aliases='RN5S335', description='RNA, 5S ribosomal pseudogene 335', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479008, CurrentID=0, Status='live', Symbol='RNA5SP342', Aliases='RN5S342', description='RNA, 5S ribosomal pseudogene 342', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479009, CurrentID=0, Status='live', Symbol='RNA5SP355', Aliases='RN5S355', description='RNA, 5S ribosomal pseudogene 355', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479010, CurrentID=0, Status='live', Symbol='RNA5SP376', Aliases='RN5S376', description='RNA, 5S ribosomal pseudogene 376', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  139163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139163, CurrentID=0, Status='live', Symbol='SALL1P1', Aliases='SALL1P', description='spalt-like transcription factor 1 pseudogene 1', other_designations='sal-like 1 pseudogene 1', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49663728, end_position_on_the_genomic_accession=49668690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479012, CurrentID=0, Status='live', Symbol='RNA5SP403', Aliases='RN5S403', description='RNA, 5S ribosomal pseudogene 403', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479013, CurrentID=0, Status='live', Symbol='RNA5SP435', Aliases='RN5S435', description='RNA, 5S ribosomal pseudogene 435', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479014, CurrentID=0, Status='live', Symbol='RNA5SP439', Aliases='RN5S439', description='RNA, 5S ribosomal pseudogene 439', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554225, CurrentID=0, Status='live', Symbol='STRCP1', Aliases='STRCP', description='stereocilin pseudogene 1', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43699379, end_position_on_the_genomic_accession=43718262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479016, CurrentID=0, Status='live', Symbol='RNA5SP458', Aliases='RN5S458', description='RNA, 5S ribosomal pseudogene 458', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479017, CurrentID=0, Status='live', Symbol='RNA5SP473', Aliases='RN5S473', description='RNA, 5S ribosomal pseudogene 473', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479018, CurrentID=0, Status='live', Symbol='RNA5SP521', Aliases='RN5S521', description='RNA, 5S ribosomal pseudogene 521', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441771, CurrentID=0, Status='live', Symbol='MT1CP', Aliases='', description='metallothionein 1C, pseudogene', other_designations='metallothionein 1C (pseudogene)', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56647689, end_position_on_the_genomic_accession=56649513, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479020, CurrentID=0, Status='live', Symbol='KIAA0196-AS1', Aliases='', description='KIAA0196 antisense RNA 1', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479021, CurrentID=0, Status='live', Symbol='CHEK2P4', Aliases='', description='checkpoint kinase 2 pseudogene 4', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441774, CurrentID=0, Status='live', Symbol='RPS4Y1P1', Aliases='RPS4Y_1_1499', description='ribosomal protein S4Y pseudogene 1', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75192400, end_position_on_the_genomic_accession=75193289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479023, CurrentID=0, Status='live', Symbol='HIST2H3DP1', Aliases='', description='histone cluster 2, H3d pseudogene 1', other_designations='', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479024, CurrentID=0, Status='live', Symbol='LRRC37A13P', Aliases='', description='leucine-rich repeat containing 37 member A13, pseudogene', other_designations='leucine rich repeat containing 37, member A13, pseudogene', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479025, CurrentID=0, Status='live', Symbol='LRRC37A14P', Aliases='', description='leucine-rich repeat containing 37 member A14, pseudogene', other_designations='leucine rich repeat containing 37, member A14, pseudogene', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479026, CurrentID=0, Status='live', Symbol='NECAP1P2', Aliases='', description='NECAP endocytosis associated 1 pseudogene 2', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479027, CurrentID=0, Status='live', Symbol='KRT18P56', Aliases='', description='keratin 18 pseudogene 56', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479028, CurrentID=0, Status='live', Symbol='AIG1P1', Aliases='', description='androgen-induced 1 pseudogene 1', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441781, CurrentID=0, Status='live', Symbol='RPL23AP76', Aliases='RPL23A_31_1517', description='ribosomal protein L23a pseudogene 76', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14947741, end_position_on_the_genomic_accession=14948198, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479030, CurrentID=0, Status='live', Symbol='WBP1LP7', Aliases='', description='WW domain binding protein 1-like pseudogene 7', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479031, CurrentID=0, Status='live', Symbol='WBP1LP3', Aliases='', description='WW domain binding protein 1-like pseudogene 3', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479032, CurrentID=0, Status='live', Symbol='PQLC1P1', Aliases='', description='PQ loop repeat containing 1 pseudogene 1', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479033, CurrentID=0, Status='live', Symbol='TRMT112P2', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 2', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479034, CurrentID=0, Status='live', Symbol='LARP1BP2', Aliases='PBEF', description='La ribonucleoprotein domain family member 1B pseudogene 2', other_designations='H_NH0022N19_gi18089094.sp_cds.1|La ribonucleoprotein domain family, member 1B pseudogene 2', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479035, CurrentID=0, Status='live', Symbol='GZMAP1', Aliases='', description='granzyme A pseudogene 1', other_designations='granzyme A (granzyme 1, cytotoxic T-lymphocyte-associated serine esterase 3) pseudogene 1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479036, CurrentID=0, Status='live', Symbol='EIF1AXP2', Aliases='', description='eukaryotic translation initiation factor 1A, X-linked pseudogene 2', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479037, CurrentID=0, Status='live', Symbol='ALG1L10P', Aliases='', description='asparagine-linked glycosylation 1-like 10, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479038, CurrentID=0, Status='live', Symbol='ALG1L13P', Aliases='', description='asparagine-linked glycosylation 1-like 13, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479039, CurrentID=0, Status='live', Symbol='CHCHD4P5', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 4 pseudogene 5', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479040, CurrentID=0, Status='live', Symbol='CRIP1P3', Aliases='', description='cysteine-rich protein 1 (intestinal) pseudogene 3', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479041, CurrentID=0, Status='live', Symbol='TUBAP4', Aliases='TUBA8P2', description='tubulin alpha pseudogene 4', other_designations='tubulin, alpha pseudogene 4', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479042, CurrentID=0, Status='live', Symbol='MTATP8P2', Aliases='', description='mitochondrially encoded ATP synthase 8 pseudogene 2', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441795, CurrentID=0, Status='live', Symbol='HMGB3P27', Aliases='', description='high mobility group box 3 pseudogene 27', other_designations='high-mobility group box 3 pseudogene 27|putative high mobility group protein B3-like protein', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42648154, end_position_on_the_genomic_accession=42648738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479044, CurrentID=0, Status='live', Symbol='MTATP6P9', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 9', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479045, CurrentID=0, Status='live', Symbol='FCF1P5', Aliases='', description='FCF1 pseudogene 5', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479046, CurrentID=0, Status='live', Symbol='FCF1P10', Aliases='', description='FCF1 pseudogene 10', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479047, CurrentID=0, Status='live', Symbol='EFCAB14P1', Aliases='', description='EF-hand calcium binding domain 14 pseudogene 1', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479048, CurrentID=0, Status='live', Symbol='UBE2V1P12', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 12', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 12', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479049, CurrentID=0, Status='live', Symbol='NIFKP7', Aliases='MKI67IPP7', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 7', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479050, CurrentID=0, Status='live', Symbol='KSR1P1', Aliases='', description='kinase suppressor of ras 1 pseudogene 1', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479051, CurrentID=0, Status='live', Symbol='NDUFA3P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A3 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3 pseudogene 2', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479052, CurrentID=0, Status='live', Symbol='NDUFA3P6', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A3 pseudogene 6', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3 pseudogene 6', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479053, CurrentID=0, Status='live', Symbol='MED28P5', Aliases='', description='mediator complex subunit 28 pseudogene 5', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479054, CurrentID=0, Status='live', Symbol='TMEM256P2', Aliases='', description='transmembrane protein 256 pseudogene 2', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479055, CurrentID=0, Status='live', Symbol='SOD1P1', Aliases='', description='superoxide dismutase 1, soluble pseudogene 1', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479056, CurrentID=0, Status='live', Symbol='PIGPP2', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class P pseudogene 2', other_designations='phosphatidylinositol glycan anchor biosynthesis, class P pseudogene 2', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479057, CurrentID=0, Status='live', Symbol='FABP12P1', Aliases='', description='fatty acid binding protein 12 pseudogene 1', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479058, CurrentID=0, Status='live', Symbol='COPS8P3', Aliases='', description='COP9 signalosome subunit 8 pseudogene 3', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479059, CurrentID=0, Status='live', Symbol='OLA1P2', Aliases='', description='Obg-like ATPase 1 pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479060, CurrentID=0, Status='live', Symbol='RNU7-90P', Aliases='', description='RNA, U7 small nuclear 90 pseudogene', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479061, CurrentID=0, Status='live', Symbol='RNU7-92P', Aliases='', description='RNA, U7 small nuclear 92 pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479062, CurrentID=0, Status='live', Symbol='RNU7-95P', Aliases='', description='RNA, U7 small nuclear 95 pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479063, CurrentID=0, Status='live', Symbol='RNU7-97P', Aliases='', description='RNA, U7 small nuclear 97 pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479064, CurrentID=0, Status='live', Symbol='RNU7-102P', Aliases='', description='RNA, U7 small nuclear 102 pseudogene', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479065, CurrentID=0, Status='live', Symbol='RNU7-105P', Aliases='', description='RNA, U7 small nuclear 105 pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441818, CurrentID=0, Status='live', Symbol='WBP11P1', Aliases='HsT3017', description='WW domain binding protein 11 pseudogene 1', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32511663, end_position_on_the_genomic_accession=32514634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100190799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190799, CurrentID=0, Status='live', Symbol='LDHAP2', Aliases='LDHAL2', description='lactate dehydrogenase A pseudogene 2', other_designations='lactate dehydrogenase A-like 2', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235737895, end_position_on_the_genomic_accession=235739609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479068, CurrentID=0, Status='live', Symbol='RNU7-113P', Aliases='', description='RNA, U7 small nuclear 113 pseudogene', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554234, CurrentID=0, Status='live', Symbol='RPL36AP40', Aliases='RPL36A_15_1127', description='ribosomal protein L36a pseudogene 40', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=25588365, end_position_on_the_genomic_accession=25588866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479070, CurrentID=0, Status='live', Symbol='RNU7-119P', Aliases='', description='RNA, U7 small nuclear 119 pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401317, CurrentID=0, Status='live', Symbol='LOC401317', Aliases='', description='uncharacterized LOC401317', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479072, CurrentID=0, Status='live', Symbol='RNU7-124P', Aliases='', description='RNA, U7 small nuclear 124 pseudogene', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479073, CurrentID=0, Status='live', Symbol='RNU7-126P', Aliases='', description='RNA, U7 small nuclear 126 pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479074, CurrentID=0, Status='live', Symbol='RNU7-130P', Aliases='', description='RNA, U7 small nuclear 130 pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  29947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29947, CurrentID=0, Status='live', Symbol='DNMT3L', Aliases='', description='DNA (cytosine-5-)-methyltransferase 3-like', other_designations='cytosine-5-methyltransferase 3-like protein|human cytosine-5-methyltransferase 3-like protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44246339, end_position_on_the_genomic_accession=44267138, orientation='minus', exon_count=13, OMIM=606588, no_hdr0=''),
  106479076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479076, CurrentID=0, Status='live', Symbol='RNU7-136P', Aliases='', description='RNA, U7 small nuclear 136 pseudogene', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479077, CurrentID=0, Status='live', Symbol='RNU7-140P', Aliases='', description='RNA, U7 small nuclear 140 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479078, CurrentID=0, Status='live', Symbol='RNU7-143P', Aliases='', description='RNA, U7 small nuclear 143 pseudogene', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479079, CurrentID=0, Status='live', Symbol='RNU7-144P', Aliases='', description='RNA, U7 small nuclear 144 pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479080, CurrentID=0, Status='live', Symbol='RNU7-149P', Aliases='', description='RNA, U7 small nuclear 149 pseudogene', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554236, CurrentID=0, Status='live', Symbol='DPY19L2P1', Aliases='', description='DPY19L2 pseudogene 1', other_designations='dpy-19-like 2 pseudogene 1', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35081287, end_position_on_the_genomic_accession=35186162, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  106479082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479082, CurrentID=0, Status='live', Symbol='RNU7-153P', Aliases='', description='RNA, U7 small nuclear 153 pseudogene', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479083, CurrentID=0, Status='live', Symbol='RNU7-156P', Aliases='', description='RNA, U7 small nuclear 156 pseudogene', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479084, CurrentID=0, Status='live', Symbol='RNU7-158P', Aliases='', description='RNA, U7 small nuclear 158 pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479085, CurrentID=0, Status='live', Symbol='RNU7-160P', Aliases='', description='RNA, U7 small nuclear 160 pseudogene', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479086, CurrentID=0, Status='live', Symbol='RNU7-164P', Aliases='', description='RNA, U7 small nuclear 164 pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479087, CurrentID=0, Status='live', Symbol='RNU7-165P', Aliases='', description='RNA, U7 small nuclear 165 pseudogene', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479088, CurrentID=0, Status='live', Symbol='RNU7-167P', Aliases='', description='RNA, U7 small nuclear 167 pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479089, CurrentID=0, Status='live', Symbol='RNU7-171P', Aliases='', description='RNA, U7 small nuclear 171 pseudogene', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479090, CurrentID=0, Status='live', Symbol='RNU7-173P', Aliases='', description='RNA, U7 small nuclear 173 pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441843, CurrentID=0, Status='live', Symbol='ZNF209P', Aliases='ZNF209', description='zinc finger protein 209, pseudogene', other_designations='zinc finger protein 675 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22470954, end_position_on_the_genomic_accession=22472536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479092, CurrentID=0, Status='live', Symbol='RNU7-179P', Aliases='', description='RNA, U7 small nuclear 179 pseudogene', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479093, CurrentID=0, Status='live', Symbol='RNU7-181P', Aliases='', description='RNA, U7 small nuclear 181 pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479094, CurrentID=0, Status='live', Symbol='RNU7-185P', Aliases='', description='RNA, U7 small nuclear 185 pseudogene', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401321, CurrentID=0, Status='live', Symbol='LINC00997', Aliases='', description='long intergenic non-protein coding RNA 997', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32758286, end_position_on_the_genomic_accession=32762924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479096, CurrentID=0, Status='live', Symbol='RNU7-190P', Aliases='', description='RNA, U7 small nuclear 190 pseudogene', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441849, CurrentID=0, Status='live', Symbol='MSRB1P1', Aliases='', description='methionine sulfoxide reductase B1 pseudogene 1', other_designations='selenoprotein X, 1 pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39254919, end_position_on_the_genomic_accession=39255269, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479098, CurrentID=0, Status='live', Symbol='RNU7-196P', Aliases='', description='RNA, U7 small nuclear 196 pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479099, CurrentID=0, Status='live', Symbol='RN7SKP11', Aliases='', description='RNA, 7SK small nuclear pseudogene 11', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479100, CurrentID=0, Status='live', Symbol='RN7SKP13', Aliases='', description='RNA, 7SK small nuclear pseudogene 13', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479101, CurrentID=0, Status='live', Symbol='RN7SKP14', Aliases='', description='RNA, 7SK small nuclear pseudogene 14', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479102, CurrentID=0, Status='live', Symbol='RN7SKP17', Aliases='', description='RNA, 7SK small nuclear pseudogene 17', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479103, CurrentID=0, Status='live', Symbol='RN7SKP19', Aliases='', description='RNA, 7SK small nuclear pseudogene 19', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479104, CurrentID=0, Status='live', Symbol='RN7SKP20', Aliases='', description='RNA, 7SK small nuclear pseudogene 20', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479105, CurrentID=0, Status='live', Symbol='RN7SKP23', Aliases='', description='RNA, 7SK small nuclear pseudogene 23', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479106, CurrentID=0, Status='live', Symbol='RN7SKP26', Aliases='', description='RNA, 7SK small nuclear pseudogene 26', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479107, CurrentID=0, Status='live', Symbol='RN7SKP28', Aliases='', description='RNA, 7SK small nuclear pseudogene 28', other_designations='', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479108, CurrentID=0, Status='live', Symbol='RN7SKP30', Aliases='', description='RNA, 7SK small nuclear pseudogene 30', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479109, CurrentID=0, Status='live', Symbol='RN7SKP32', Aliases='', description='RNA, 7SK small nuclear pseudogene 32', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479110, CurrentID=0, Status='live', Symbol='RN7SKP34', Aliases='', description='RNA, 7SK small nuclear pseudogene 34', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479111, CurrentID=0, Status='live', Symbol='RN7SKP35', Aliases='', description='RNA, 7SK small nuclear pseudogene 35', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441864, CurrentID=0, Status='live', Symbol='TARM1', Aliases='OLT-2', description='T cell-interacting, activating receptor on myeloid cells 1', other_designations='OSCAR-like transcript-2 protein', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54069487, end_position_on_the_genomic_accession=54081365, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106479113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479113, CurrentID=0, Status='live', Symbol='RN7SKP40', Aliases='', description='RNA, 7SK small nuclear pseudogene 40', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479114, CurrentID=0, Status='live', Symbol='RN7SKP44', Aliases='', description='RNA, 7SK small nuclear pseudogene 44', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479115, CurrentID=0, Status='live', Symbol='RN7SKP45', Aliases='', description='RNA, 7SK small nuclear pseudogene 45', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479116, CurrentID=0, Status='live', Symbol='RN7SKP48', Aliases='', description='RNA, 7SK small nuclear pseudogene 48', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441869, CurrentID=0, Status='live', Symbol='ANKRD65', Aliases='', description='ankyrin repeat domain 65', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1418420, end_position_on_the_genomic_accession=1421820, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479118, CurrentID=0, Status='live', Symbol='RN7SKP52', Aliases='', description='RNA, 7SK small nuclear pseudogene 52', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441871, CurrentID=0, Status='live', Symbol='PRAMEF7', Aliases='', description='PRAME family member 7', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12916647, end_position_on_the_genomic_accession=12920480, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479120, CurrentID=0, Status='live', Symbol='RN7SKP57', Aliases='', description='RNA, 7SK small nuclear pseudogene 57', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441873, CurrentID=0, Status='live', Symbol='PRAMEF25', Aliases='', description='PRAME family member 25', other_designations='Putative PRAME family member 25', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13073678, end_position_on_the_genomic_accession=13078038, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479122, CurrentID=0, Status='live', Symbol='RN7SKP61', Aliases='', description='RNA, 7SK small nuclear pseudogene 61', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479123, CurrentID=0, Status='live', Symbol='RN7SKP63', Aliases='', description='RNA, 7SK small nuclear pseudogene 63', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441876, CurrentID=0, Status='live', Symbol='RPS16P1', Aliases='RPS16_1_14', description='ribosomal protein S16 pseudogene 1', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15828212, end_position_on_the_genomic_accession=15828733, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479125, CurrentID=0, Status='live', Symbol='RN7SKP67', Aliases='', description='RNA, 7SK small nuclear pseudogene 67', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441878, CurrentID=0, Status='live', Symbol='RPS4XP4', Aliases='RPS4P4, RPS4X_1_17', description='ribosomal protein S4X pseudogene 4', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20525952, end_position_on_the_genomic_accession=20526810, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479127, CurrentID=0, Status='live', Symbol='RN7SKP74', Aliases='', description='RNA, 7SK small nuclear pseudogene 74', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479128, CurrentID=0, Status='live', Symbol='RN7SKP77', Aliases='', description='RNA, 7SK small nuclear pseudogene 77', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479129, CurrentID=0, Status='live', Symbol='RN7SKP78', Aliases='', description='RNA, 7SK small nuclear pseudogene 78', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479130, CurrentID=0, Status='live', Symbol='RN7SKP80', Aliases='', description='RNA, 7SK small nuclear pseudogene 80', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479131, CurrentID=0, Status='live', Symbol='RN7SKP83', Aliases='', description='RNA, 7SK small nuclear pseudogene 83', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132556, CurrentID=0, Status='live', Symbol='BTF3P13', Aliases='BTF3L3, HUMBTFD', description='basic transcription factor 3 pseudogene 13', other_designations='basic transcription factor 3, like 3|basic transcription factor 3, pseudogene|basic transcription factor 3-like 3, pseudogene|lambda h33', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98740625, end_position_on_the_genomic_accession=98741496, orientation='plus', exon_count=0, OMIM=603739, no_hdr0=''),
  106479133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479133, CurrentID=0, Status='live', Symbol='RN7SKP87', Aliases='', description='RNA, 7SK small nuclear pseudogene 87', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479134, CurrentID=0, Status='live', Symbol='RN7SKP89', Aliases='', description='RNA, 7SK small nuclear pseudogene 89', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479135, CurrentID=0, Status='live', Symbol='RN7SKP92', Aliases='', description='RNA, 7SK small nuclear pseudogene 92', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479136, CurrentID=0, Status='live', Symbol='RN7SKP93', Aliases='', description='RNA, 7SK small nuclear pseudogene 93', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479137, CurrentID=0, Status='live', Symbol='RN7SKP96', Aliases='', description='RNA, 7SK small nuclear pseudogene 96', other_designations='', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479138, CurrentID=0, Status='live', Symbol='RN7SKP99', Aliases='', description='RNA, 7SK small nuclear pseudogene 99', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479139, CurrentID=0, Status='live', Symbol='RN7SKP100', Aliases='', description='RNA, 7SK small nuclear pseudogene 100', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479140, CurrentID=0, Status='live', Symbol='RN7SKP103', Aliases='', description='RNA, 7SK small nuclear pseudogene 103', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479141, CurrentID=0, Status='live', Symbol='RN7SKP105', Aliases='', description='RNA, 7SK small nuclear pseudogene 105', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479142, CurrentID=0, Status='live', Symbol='RN7SKP106', Aliases='', description='RNA, 7SK small nuclear pseudogene 106', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479143, CurrentID=0, Status='live', Symbol='RN7SKP109', Aliases='', description='RNA, 7SK small nuclear pseudogene 109', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441896, CurrentID=0, Status='live', Symbol='RPL7P8', Aliases='RPL7_2_84', description='ribosomal protein L7 pseudogene 8', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109651363, end_position_on_the_genomic_accession=109651858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479145, CurrentID=0, Status='live', Symbol='RN7SKP112', Aliases='', description='RNA, 7SK small nuclear pseudogene 112', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441898, CurrentID=0, Status='live', Symbol='RPS15AP9', Aliases='RPS15A_2_91', description='ribosomal protein S15a pseudogene 9', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117138190, end_position_on_the_genomic_accession=117138648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479147, CurrentID=0, Status='live', Symbol='RN7SKP117', Aliases='', description='RNA, 7SK small nuclear pseudogene 117', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441900, CurrentID=0, Status='live', Symbol='HSD3BP3', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta, pseudogene 3', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119538509, end_position_on_the_genomic_accession=119546088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479149, CurrentID=0, Status='live', Symbol='RN7SKP123', Aliases='', description='RNA, 7SK small nuclear pseudogene 123', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479150, CurrentID=0, Status='live', Symbol='RN7SKP125', Aliases='', description='RNA, 7SK small nuclear pseudogene 125', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479151, CurrentID=0, Status='live', Symbol='RN7SKP128', Aliases='', description='RNA, 7SK small nuclear pseudogene 128', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479152, CurrentID=0, Status='live', Symbol='RN7SKP129', Aliases='', description='RNA, 7SK small nuclear pseudogene 129', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479153, CurrentID=0, Status='live', Symbol='RN7SKP131', Aliases='', description='RNA, 7SK small nuclear pseudogene 131', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479154, CurrentID=0, Status='live', Symbol='RN7SKP134', Aliases='', description='RNA, 7SK small nuclear pseudogene 134', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401331, CurrentID=0, Status='live', Symbol='RASA4CP', Aliases='RASA4P', description='RAS p21 protein activator 4C, pseudogene', other_designations='RAS p21 protein activator 4 pseudogene', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44028887, end_position_on_the_genomic_accession=44040623, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  441908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441908, CurrentID=0, Status='live', Symbol='NBPF18P', Aliases='', description='neuroblastoma breakpoint family member 18, pseudogene', other_designations='neuroblastoma breakpoint family, member 18 (pseudogene)|neuroblastoma breakpoint family, member 18, pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152018278, end_position_on_the_genomic_accession=152022509, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106479157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479157, CurrentID=0, Status='live', Symbol='RN7SKP141', Aliases='', description='RNA, 7SK small nuclear pseudogene 141', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479158, CurrentID=0, Status='live', Symbol='RN7SKP143', Aliases='', description='RNA, 7SK small nuclear pseudogene 143', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441911, CurrentID=0, Status='live', Symbol='OR10J3', Aliases='OR1-25P, OR10J3', description='olfactory receptor family 10 subfamily J member 3', other_designations='olfactory receptor OR1-25 pseudogene|olfactory receptor, family 10, subfamily J, member 3 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159313670, end_position_on_the_genomic_accession=159314659, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479160, CurrentID=0, Status='live', Symbol='RN7SKP146', Aliases='', description='RNA, 7SK small nuclear pseudogene 146', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479161, CurrentID=0, Status='live', Symbol='RN7SKP149', Aliases='', description='RNA, 7SK small nuclear pseudogene 149', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479162, CurrentID=0, Status='live', Symbol='RN7SKP151', Aliases='', description='RNA, 7SK small nuclear pseudogene 151', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441915, CurrentID=0, Status='live', Symbol='SLC25A38P1', Aliases='', description='solute carrier family 25, member 38 pseudogene 1', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172748252, end_position_on_the_genomic_accession=172750171, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479164, CurrentID=0, Status='live', Symbol='RN7SKP155', Aliases='', description='RNA, 7SK small nuclear pseudogene 155', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479165, CurrentID=0, Status='live', Symbol='RN7SKP157', Aliases='', description='RNA, 7SK small nuclear pseudogene 157', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187710, CurrentID=0, Status='live', Symbol='CRCS6', Aliases='', description='colorectal cancer, susceptibility to, 6', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612231, no_hdr0=''),
  100187711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187711, CurrentID=0, Status='live', Symbol='CRCS7', Aliases='', description='colorectal cancer, susceptibility to, 7', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612232, no_hdr0=''),
  106479168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479168, CurrentID=0, Status='live', Symbol='RN7SKP163', Aliases='', description='RNA, 7SK small nuclear pseudogene 163', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479169, CurrentID=0, Status='live', Symbol='RN7SKP167', Aliases='', description='RNA, 7SK small nuclear pseudogene 167', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479170, CurrentID=0, Status='live', Symbol='RN7SKP168', Aliases='', description='RNA, 7SK small nuclear pseudogene 168', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479171, CurrentID=0, Status='live', Symbol='RN7SKP170', Aliases='', description='RNA, 7SK small nuclear pseudogene 170', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187716, CurrentID=0, Status='live', Symbol='MIR1224', Aliases='MIRN1224, mir-1224', description='microRNA 1224', other_designations='hsa-mir-1224', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184241405, end_position_on_the_genomic_accession=184241489, orientation='plus', exon_count=1, OMIM=611620, no_hdr0=''),
  100187717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187717, CurrentID=0, Status='live', Symbol='NNO3', Aliases='', description='nanophthalmos 3', other_designations='', map_location='2q11-q14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611897, no_hdr0=''),
  106479174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479174, CurrentID=0, Status='live', Symbol='RN7SKP177', Aliases='', description='RNA, 7SK small nuclear pseudogene 177', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479175, CurrentID=0, Status='live', Symbol='RN7SKP178', Aliases='', description='RNA, 7SK small nuclear pseudogene 178', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479176, CurrentID=0, Status='live', Symbol='RN7SKP181', Aliases='', description='RNA, 7SK small nuclear pseudogene 181', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479177, CurrentID=0, Status='live', Symbol='RN7SKP182', Aliases='', description='RNA, 7SK small nuclear pseudogene 182', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479178, CurrentID=0, Status='live', Symbol='RN7SKP185', Aliases='', description='RNA, 7SK small nuclear pseudogene 185', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441931, CurrentID=0, Status='live', Symbol='VN1R17P', Aliases='GPCR', description='vomeronasal 1 receptor 17 pseudogene', other_designations='putative G-protein coupled receptor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247237224, end_position_on_the_genomic_accession=247237850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100187724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187724, CurrentID=0, Status='live', Symbol='DEL16P11.2', Aliases='AUTS14, AUTS14A, C16DELp11.2, C16DUPp11.2, DUP16p11.2', description='autism, susceptibility to, 14', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611913, no_hdr0=''),
  441933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441933, CurrentID=0, Status='live', Symbol='OR13G1', Aliases='OR1-37', description='olfactory receptor family 13 subfamily G member 1', other_designations='olfactory receptor OR1-37|olfactory receptor, family 13, subfamily G, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247672118, end_position_on_the_genomic_accession=247673041, orientation='minus', exon_count=1, OMIM=611677, no_hdr0=''),
  106479182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479182, CurrentID=0, Status='live', Symbol='RN7SKP196', Aliases='', description='RNA, 7SK small nuclear pseudogene 196', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479183, CurrentID=0, Status='live', Symbol='RN7SKP198', Aliases='', description='RNA, 7SK small nuclear pseudogene 198', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479184, CurrentID=0, Status='live', Symbol='RN7SKP199', Aliases='', description='RNA, 7SK small nuclear pseudogene 199', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479185, CurrentID=0, Status='live', Symbol='RN7SKP203', Aliases='', description='RNA, 7SK small nuclear pseudogene 203', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479186, CurrentID=0, Status='live', Symbol='RN7SKP205', Aliases='', description='RNA, 7SK small nuclear pseudogene 205', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479187, CurrentID=0, Status='live', Symbol='RN7SKP206', Aliases='', description='RNA, 7SK small nuclear pseudogene 206', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479188, CurrentID=0, Status='live', Symbol='RN7SKP209', Aliases='', description='RNA, 7SK small nuclear pseudogene 209', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479189, CurrentID=0, Status='live', Symbol='RN7SKP211', Aliases='', description='RNA, 7SK small nuclear pseudogene 211', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479190, CurrentID=0, Status='live', Symbol='RN7SKP216', Aliases='', description='RNA, 7SK small nuclear pseudogene 216', other_designations='', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479191, CurrentID=0, Status='live', Symbol='RN7SKP217', Aliases='', description='RNA, 7SK small nuclear pseudogene 217', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479192, CurrentID=0, Status='live', Symbol='RN7SKP219', Aliases='', description='RNA, 7SK small nuclear pseudogene 219', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479193, CurrentID=0, Status='live', Symbol='RN7SKP222', Aliases='', description='RNA, 7SK small nuclear pseudogene 222', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479194, CurrentID=0, Status='live', Symbol='RN7SKP223', Aliases='', description='RNA, 7SK small nuclear pseudogene 223', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479195, CurrentID=0, Status='live', Symbol='RN7SKP225', Aliases='', description='RNA, 7SK small nuclear pseudogene 225', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479196, CurrentID=0, Status='live', Symbol='RN7SKP228', Aliases='', description='RNA, 7SK small nuclear pseudogene 228', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479197, CurrentID=0, Status='live', Symbol='RN7SKP230', Aliases='', description='RNA, 7SK small nuclear pseudogene 230', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479198, CurrentID=0, Status='live', Symbol='RN7SKP231', Aliases='', description='RNA, 7SK small nuclear pseudogene 231', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  441951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=441951, CurrentID=0, Status='live', Symbol='ZFAS1', Aliases='C20orf199, HSUP1, HSUP2, NCRNA00275, ZNFX1-AS1', description='ZNFX1 antisense RNA 1', other_designations='Putative uncharacterized protein ZNFX1-AS1|ZNFX1 antisense RNA 1 (non-protein coding)|ZNFX1 antisense gene protein 1', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49278178, end_position_on_the_genomic_accession=49289260, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106479200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479200, CurrentID=0, Status='live', Symbol='RN7SKP237', Aliases='', description='RNA, 7SK small nuclear pseudogene 237', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479201, CurrentID=0, Status='live', Symbol='RN7SKP238', Aliases='', description='RNA, 7SK small nuclear pseudogene 238', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479202, CurrentID=0, Status='live', Symbol='RN7SKP240', Aliases='', description='RNA, 7SK small nuclear pseudogene 240', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8123, CurrentID=0, Status='live', Symbol='PWAR5', Aliases='D15S226E, PAR-5, PAR5', description='Prader Willi/Angelman region RNA 5', other_designations='Prader Willi/Angelman syndrome 5|Prader-Willi/Angelman syndrome-5', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24984860, end_position_on_the_genomic_accession=24988232, orientation='plus', exon_count=1, OMIM=600162, no_hdr0=''),
  106479204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479204, CurrentID=0, Status='live', Symbol='RN7SKP246', Aliases='', description='RNA, 7SK small nuclear pseudogene 246', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187749, CurrentID=0, Status='live', Symbol='ASDP', Aliases='', description='anal sphincter dysplasia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=105563, no_hdr0=''),
  106479206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479206, CurrentID=0, Status='live', Symbol='RN7SKP250', Aliases='', description='RNA, 7SK small nuclear pseudogene 250', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479207, CurrentID=0, Status='live', Symbol='RN7SKP252', Aliases='', description='RNA, 7SK small nuclear pseudogene 252', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479208, CurrentID=0, Status='live', Symbol='RN7SKP253', Aliases='', description='RNA, 7SK small nuclear pseudogene 253', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479209, CurrentID=0, Status='live', Symbol='RN7SKP255', Aliases='', description='RNA, 7SK small nuclear pseudogene 255', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479210, CurrentID=0, Status='live', Symbol='RN7SKP257', Aliases='', description='RNA, 7SK small nuclear pseudogene 257', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479211, CurrentID=0, Status='live', Symbol='RN7SKP260', Aliases='', description='RNA, 7SK small nuclear pseudogene 260', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479212, CurrentID=0, Status='live', Symbol='RN7SKP262', Aliases='', description='RNA, 7SK small nuclear pseudogene 262', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104644205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104644205, CurrentID=0, Status='live', Symbol='ENPP7P13', Aliases='DUNQU1', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 13', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33769377, end_position_on_the_genomic_accession=33784275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479214, CurrentID=0, Status='live', Symbol='RN7SKP266', Aliases='', description='RNA, 7SK small nuclear pseudogene 266', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479215, CurrentID=0, Status='live', Symbol='RN7SKP268', Aliases='', description='RNA, 7SK small nuclear pseudogene 268', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479216, CurrentID=0, Status='live', Symbol='RN7SKP270', Aliases='', description='RNA, 7SK small nuclear pseudogene 270', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187761, CurrentID=0, Status='live', Symbol='AN', Aliases='', description='blood group Ahonen', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=110350, no_hdr0=''),
  106479218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479218, CurrentID=0, Status='live', Symbol='RN7SKP276', Aliases='', description='RNA, 7SK small nuclear pseudogene 276', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479219, CurrentID=0, Status='live', Symbol='RN7SKP277', Aliases='', description='RNA, 7SK small nuclear pseudogene 277', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479220, CurrentID=0, Status='live', Symbol='RN7SKP280', Aliases='', description='RNA, 7SK small nuclear pseudogene 280', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479221, CurrentID=0, Status='live', Symbol='RN7SKP283', Aliases='', description='RNA, 7SK small nuclear pseudogene 283', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479222, CurrentID=0, Status='live', Symbol='RN7SKP284', Aliases='', description='RNA, 7SK small nuclear pseudogene 284', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479223, CurrentID=0, Status='live', Symbol='RN7SKP288', Aliases='', description='RNA, 7SK small nuclear pseudogene 288', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479224, CurrentID=0, Status='live', Symbol='RN7SKP289', Aliases='', description='RNA, 7SK small nuclear pseudogene 289', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479225, CurrentID=0, Status='live', Symbol='RN7SKP292', Aliases='', description='RNA, 7SK small nuclear pseudogene 292', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479226, CurrentID=0, Status='live', Symbol='RN7SKP294', Aliases='', description='RNA, 7SK small nuclear pseudogene 294', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479227, CurrentID=0, Status='live', Symbol='RN7SKP297', Aliases='', description='RNA, 7SK small nuclear pseudogene 297', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479228, CurrentID=0, Status='live', Symbol='RN7SKP298', Aliases='', description='RNA, 7SK small nuclear pseudogene 298', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479229, CurrentID=0, Status='live', Symbol='RN7SL11P', Aliases='', description='RNA, 7SL, cytoplasmic 11, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479230, CurrentID=0, Status='live', Symbol='RN7SL15P', Aliases='', description='RNA, 7SL, cytoplasmic 15, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479231, CurrentID=0, Status='live', Symbol='RN7SL16P', Aliases='', description='RNA, 7SL, cytoplasmic 16, pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479232, CurrentID=0, Status='live', Symbol='RN7SL20P', Aliases='', description='RNA, 7SL, cytoplasmic 20, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479233, CurrentID=0, Status='live', Symbol='RN7SL25P', Aliases='', description='RNA, 7SL, cytoplasmic 25, pseudogene', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479234, CurrentID=0, Status='live', Symbol='RN7SL26P', Aliases='', description='RNA, 7SL, cytoplasmic 26, pseudogene', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479235, CurrentID=0, Status='live', Symbol='RN7SL30P', Aliases='', description='RNA, 7SL, cytoplasmic 30, pseudogene', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479236, CurrentID=0, Status='live', Symbol='RN7SL33P', Aliases='', description='RNA, 7SL, cytoplasmic 33, pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479237, CurrentID=0, Status='live', Symbol='RN7SL36P', Aliases='', description='RNA, 7SL, cytoplasmic 36, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479238, CurrentID=0, Status='live', Symbol='RN7SL37P', Aliases='', description='RNA, 7SL, cytoplasmic 37, pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8129, CurrentID=0, Status='live', Symbol='HBHR', Aliases='ATR1', description='alpha-thalassemia/mental retardation syndrome, type 1', other_designations='', map_location='16pter-p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=141750, no_hdr0=''),
  106479240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479240, CurrentID=0, Status='live', Symbol='RN7SL42P', Aliases='', description='RNA, 7SL, cytoplasmic 42, pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479241, CurrentID=0, Status='live', Symbol='RN7SL43P', Aliases='', description='RNA, 7SL, cytoplasmic 43, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479242, CurrentID=0, Status='live', Symbol='RN7SL47P', Aliases='', description='RNA, 7SL, cytoplasmic 47, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479243, CurrentID=0, Status='live', Symbol='RN7SL50P', Aliases='', description='RNA, 7SL, cytoplasmic 50, pseudogene', other_designations='', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479244, CurrentID=0, Status='live', Symbol='RN7SL55P', Aliases='', description='RNA, 7SL, cytoplasmic 55, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479245, CurrentID=0, Status='live', Symbol='RN7SL58P', Aliases='', description='RNA, 7SL, cytoplasmic 58, pseudogene', other_designations='', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479246, CurrentID=0, Status='live', Symbol='RN7SL62P', Aliases='', description='RNA, 7SL, cytoplasmic 62, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479247, CurrentID=0, Status='live', Symbol='RN7SL65P', Aliases='', description='RNA, 7SL, cytoplasmic 65, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479248, CurrentID=0, Status='live', Symbol='RN7SL67P', Aliases='', description='RNA, 7SL, cytoplasmic 67, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479249, CurrentID=0, Status='live', Symbol='RN7SL70P', Aliases='', description='RNA, 7SL, cytoplasmic 70, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479250, CurrentID=0, Status='live', Symbol='RN7SL73P', Aliases='', description='RNA, 7SL, cytoplasmic 73, pseudogene', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479251, CurrentID=0, Status='live', Symbol='RN7SL76P', Aliases='', description='RNA, 7SL, cytoplasmic 76, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479252, CurrentID=0, Status='live', Symbol='RN7SL77P', Aliases='', description='RNA, 7SL, cytoplasmic 77, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479253, CurrentID=0, Status='live', Symbol='RN7SL81P', Aliases='', description='RNA, 7SL, cytoplasmic 81, pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479254, CurrentID=0, Status='live', Symbol='RN7SL83P', Aliases='', description='RNA, 7SL, cytoplasmic 83, pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479255, CurrentID=0, Status='live', Symbol='RN7SL87P', Aliases='', description='RNA, 7SL, cytoplasmic 87, pseudogene', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479256, CurrentID=0, Status='live', Symbol='RN7SL89P', Aliases='', description='RNA, 7SL, cytoplasmic 89, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479257, CurrentID=0, Status='live', Symbol='RN7SL92P', Aliases='', description='RNA, 7SL, cytoplasmic 92, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479258, CurrentID=0, Status='live', Symbol='RN7SL97P', Aliases='', description='RNA, 7SL, cytoplasmic 97, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479259, CurrentID=0, Status='live', Symbol='RN7SL98P', Aliases='', description='RNA, 7SL, cytoplasmic 98, pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479260, CurrentID=0, Status='live', Symbol='RN7SL101P', Aliases='', description='RNA, 7SL, cytoplasmic 101, pseudogene', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442013, CurrentID=0, Status='live', Symbol='LDHAP3', Aliases='LDHAL3', description='lactate dehydrogenase A pseudogene 3', other_designations='lactate dehydrogenase A-like 3', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41819648, end_position_on_the_genomic_accession=41821649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479262, CurrentID=0, Status='live', Symbol='RN7SL107P', Aliases='', description='RNA, 7SL, cytoplasmic 107, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479263, CurrentID=0, Status='live', Symbol='RN7SL111P', Aliases='', description='RNA, 7SL, cytoplasmic 111, pseudogene', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479264, CurrentID=0, Status='live', Symbol='RN7SL117P', Aliases='', description='RNA, 7SL, cytoplasmic 117, pseudogene', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479265, CurrentID=0, Status='live', Symbol='RN7SL121P', Aliases='', description='RNA, 7SL, cytoplasmic 121, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479266, CurrentID=0, Status='live', Symbol='RN7SL123P', Aliases='', description='RNA, 7SL, cytoplasmic 123, pseudogene', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479267, CurrentID=0, Status='live', Symbol='RN7SL125P', Aliases='', description='RNA, 7SL, cytoplasmic 125, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479268, CurrentID=0, Status='live', Symbol='RN7SL126P', Aliases='', description='RNA, 7SL, cytoplasmic 126, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479269, CurrentID=0, Status='live', Symbol='RN7SL130P', Aliases='', description='RNA, 7SL, cytoplasmic 130, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479270, CurrentID=0, Status='live', Symbol='RN7SL132P', Aliases='', description='RNA, 7SL, cytoplasmic 132, pseudogene', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479271, CurrentID=0, Status='live', Symbol='RN7SL134P', Aliases='', description='RNA, 7SL, cytoplasmic 134, pseudogene', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479272, CurrentID=0, Status='live', Symbol='RN7SL140P', Aliases='', description='RNA, 7SL, cytoplasmic 140, pseudogene', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479273, CurrentID=0, Status='live', Symbol='RN7SL143P', Aliases='', description='RNA, 7SL, cytoplasmic 143, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479274, CurrentID=0, Status='live', Symbol='RN7SL147P', Aliases='', description='RNA, 7SL, cytoplasmic 147, pseudogene', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479275, CurrentID=0, Status='live', Symbol='RN7SL148P', Aliases='', description='RNA, 7SL, cytoplasmic 148, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479276, CurrentID=0, Status='live', Symbol='RN7SL150P', Aliases='', description='RNA, 7SL, cytoplasmic 150, pseudogene', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479277, CurrentID=0, Status='live', Symbol='RN7SL154P', Aliases='', description='RNA, 7SL, cytoplasmic 154, pseudogene', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479278, CurrentID=0, Status='live', Symbol='RN7SL155P', Aliases='', description='RNA, 7SL, cytoplasmic 155, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479279, CurrentID=0, Status='live', Symbol='RN7SL159P', Aliases='', description='RNA, 7SL, cytoplasmic 159, pseudogene', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479280, CurrentID=0, Status='live', Symbol='RN7SL163P', Aliases='', description='RNA, 7SL, cytoplasmic 163, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479281, CurrentID=0, Status='live', Symbol='RN7SL165P', Aliases='', description='RNA, 7SL, cytoplasmic 165, pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479282, CurrentID=0, Status='live', Symbol='RN7SL166P', Aliases='', description='RNA, 7SL, cytoplasmic 166, pseudogene', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479283, CurrentID=0, Status='live', Symbol='RN7SL170P', Aliases='', description='RNA, 7SL, cytoplasmic 170, pseudogene', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187828, CurrentID=0, Status='live', Symbol='HBD', Aliases='', description='hypophosphatemic bone disease', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146350, no_hdr0=''),
  106479285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479285, CurrentID=0, Status='live', Symbol='RN7SL174P', Aliases='', description='RNA, 7SL, cytoplasmic 174, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442038, CurrentID=0, Status='live', Symbol='SULT1C3', Aliases='ST1C3', description='sulfotransferase family 1C member 3', other_designations='sulfotransferase family, cytosolic, 1C, member 3', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108239974, end_position_on_the_genomic_accession=108265608, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  106479287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479287, CurrentID=0, Status='live', Symbol='RN7SL182P', Aliases='', description='RNA, 7SL, cytoplasmic 182, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479288, CurrentID=0, Status='live', Symbol='RN7SL184P', Aliases='', description='RNA, 7SL, cytoplasmic 184, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479289, CurrentID=0, Status='live', Symbol='RN7SL185P', Aliases='', description='RNA, 7SL, cytoplasmic 185, pseudogene', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479290, CurrentID=0, Status='live', Symbol='RN7SL188P', Aliases='', description='RNA, 7SL, cytoplasmic 188, pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479291, CurrentID=0, Status='live', Symbol='RN7SL190P', Aliases='', description='RNA, 7SL, cytoplasmic 190, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479292, CurrentID=0, Status='live', Symbol='RN7SL191P', Aliases='', description='RNA, 7SL, cytoplasmic 191, pseudogene', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479293, CurrentID=0, Status='live', Symbol='RN7SL196P', Aliases='', description='RNA, 7SL, cytoplasmic 196, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479294, CurrentID=0, Status='live', Symbol='RN7SL198P', Aliases='', description='RNA, 7SL, cytoplasmic 198, pseudogene', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479295, CurrentID=0, Status='live', Symbol='RN7SL202P', Aliases='', description='RNA, 7SL, cytoplasmic 202, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479296, CurrentID=0, Status='live', Symbol='RN7SL203P', Aliases='', description='RNA, 7SL, cytoplasmic 203, pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479297, CurrentID=0, Status='live', Symbol='RN7SL205P', Aliases='', description='RNA, 7SL, cytoplasmic 205, pseudogene', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479298, CurrentID=0, Status='live', Symbol='RN7SL210P', Aliases='', description='RNA, 7SL, cytoplasmic 210, pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479299, CurrentID=0, Status='live', Symbol='RN7SL211P', Aliases='', description='RNA, 7SL, cytoplasmic 211, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479300, CurrentID=0, Status='live', Symbol='RN7SL214P', Aliases='', description='RNA, 7SL, cytoplasmic 214, pseudogene', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479301, CurrentID=0, Status='live', Symbol='RN7SL217P', Aliases='', description='RNA, 7SL, cytoplasmic 217, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479302, CurrentID=0, Status='live', Symbol='RN7SL219P', Aliases='', description='RNA, 7SL, cytoplasmic 219, pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479303, CurrentID=0, Status='live', Symbol='RN7SL220P', Aliases='', description='RNA, 7SL, cytoplasmic 220, pseudogene', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479304, CurrentID=0, Status='live', Symbol='RN7SL225P', Aliases='', description='RNA, 7SL, cytoplasmic 225, pseudogene', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  139212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139212, CurrentID=0, Status='live', Symbol='PIH1D3', Aliases='CXorf41, NYSAR97', description='PIH1 domain containing 3', other_designations='PIH1 domain-containing protein 3|sarcoma antigen NY-SAR-97', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107206559, end_position_on_the_genomic_accession=107244243, orientation='plus', exon_count=8, OMIM=300933, no_hdr0=''),
  106479306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479306, CurrentID=0, Status='live', Symbol='RN7SL230P', Aliases='', description='RNA, 7SL, cytoplasmic 230, pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479307, CurrentID=0, Status='live', Symbol='RN7SL234P', Aliases='', description='RNA, 7SL, cytoplasmic 234, pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479308, CurrentID=0, Status='live', Symbol='RN7SL237P', Aliases='', description='RNA, 7SL, cytoplasmic 237, pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479309, CurrentID=0, Status='live', Symbol='RN7SL239P', Aliases='', description='RNA, 7SL, cytoplasmic 239, pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479310, CurrentID=0, Status='live', Symbol='RN7SL240P', Aliases='', description='RNA, 7SL, cytoplasmic 240, pseudogene', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479311, CurrentID=0, Status='live', Symbol='RN7SL244P', Aliases='', description='RNA, 7SL, cytoplasmic 244, pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479312, CurrentID=0, Status='live', Symbol='RN7SL248P', Aliases='RN7SL527P', description='RNA, 7SL, cytoplasmic 248, pseudogene', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479313, CurrentID=0, Status='live', Symbol='RN7SL250P', Aliases='', description='RNA, 7SL, cytoplasmic 250, pseudogene', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479314, CurrentID=0, Status='live', Symbol='RN7SL253P', Aliases='', description='RNA, 7SL, cytoplasmic 253, pseudogene', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479315, CurrentID=0, Status='live', Symbol='RN7SL255P', Aliases='', description='RNA, 7SL, cytoplasmic 255, pseudogene', other_designations='', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479316, CurrentID=0, Status='live', Symbol='RN7SL261P', Aliases='', description='RNA, 7SL, cytoplasmic 261, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479317, CurrentID=0, Status='live', Symbol='RN7SL262P', Aliases='', description='RNA, 7SL, cytoplasmic 262, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479318, CurrentID=0, Status='live', Symbol='RN7SL265P', Aliases='', description='RNA, 7SL, cytoplasmic 265, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479319, CurrentID=0, Status='live', Symbol='RN7SL268P', Aliases='', description='RNA, 7SL, cytoplasmic 268, pseudogene', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479320, CurrentID=0, Status='live', Symbol='RN7SL269P', Aliases='', description='RNA, 7SL, cytoplasmic 269, pseudogene', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479321, CurrentID=0, Status='live', Symbol='RN7SL271P', Aliases='', description='RNA, 7SL, cytoplasmic 271, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479322, CurrentID=0, Status='live', Symbol='RN7SL274P', Aliases='', description='RNA, 7SL, cytoplasmic 274, pseudogene', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442075, CurrentID=0, Status='live', Symbol='EMC3-AS1', Aliases='', description='EMC3 antisense RNA 1', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9986893, end_position_on_the_genomic_accession=10007004, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479324, CurrentID=0, Status='live', Symbol='RN7SL278P', Aliases='', description='RNA, 7SL, cytoplasmic 278, pseudogene', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479325, CurrentID=0, Status='live', Symbol='RN7SL283P', Aliases='', description='RNA, 7SL, cytoplasmic 283, pseudogene', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479326, CurrentID=0, Status='live', Symbol='RN7SL285P', Aliases='', description='RNA, 7SL, cytoplasmic 285, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479327, CurrentID=0, Status='live', Symbol='RN7SL286P', Aliases='', description='RNA, 7SL, cytoplasmic 286, pseudogene', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479328, CurrentID=0, Status='live', Symbol='RN7SL291P', Aliases='', description='RNA, 7SL, cytoplasmic 291, pseudogene', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479329, CurrentID=0, Status='live', Symbol='RN7SL294P', Aliases='', description='RNA, 7SL, cytoplasmic 294, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479330, CurrentID=0, Status='live', Symbol='RN7SL297P', Aliases='', description='RNA, 7SL, cytoplasmic 297, pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442083, CurrentID=0, Status='live', Symbol='HSP90AB5P', Aliases='HSP90BE', description='heat shock protein 90kDa alpha family class B member 5, pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class B member 5, pseudogene|heat shock protein HSP 90-beta', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74101347, end_position_on_the_genomic_accession=74103828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479332, CurrentID=0, Status='live', Symbol='RN7SL302P', Aliases='', description='RNA, 7SL, cytoplasmic 302, pseudogene', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479333, CurrentID=0, Status='live', Symbol='RN7SL304P', Aliases='', description='RNA, 7SL, cytoplasmic 304, pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442086, CurrentID=0, Status='live', Symbol='RPL10P7', Aliases='RPL10_2_395', description='ribosomal protein L10 pseudogene 7', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119635466, end_position_on_the_genomic_accession=119636173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479335, CurrentID=0, Status='live', Symbol='RN7SL308P', Aliases='', description='RNA, 7SL, cytoplasmic 308, pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479336, CurrentID=0, Status='live', Symbol='RN7SL310P', Aliases='', description='RNA, 7SL, cytoplasmic 310, pseudogene', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479337, CurrentID=0, Status='live', Symbol='RN7SL316P', Aliases='', description='RNA, 7SL, cytoplasmic 316, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479338, CurrentID=0, Status='live', Symbol='RN7SL317P', Aliases='', description='RNA, 7SL, cytoplasmic 317, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554279, CurrentID=0, Status='live', Symbol='LINC00862', Aliases='C1orf98, SMIM16', description='long intergenic non-protein coding RNA 862', other_designations='small integral membrane protein 16', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200342544, end_position_on_the_genomic_accession=200373792, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106479340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479340, CurrentID=0, Status='live', Symbol='RN7SL322P', Aliases='', description='RNA, 7SL, cytoplasmic 322, pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479341, CurrentID=0, Status='live', Symbol='RN7SL323P', Aliases='', description='RNA, 7SL, cytoplasmic 323, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479342, CurrentID=0, Status='live', Symbol='RN7SL326P', Aliases='', description='RNA, 7SL, cytoplasmic 326, pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479343, CurrentID=0, Status='live', Symbol='RN7SL329P', Aliases='', description='RNA, 7SL, cytoplasmic 329, pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479344, CurrentID=0, Status='live', Symbol='RN7SL331P', Aliases='', description='RNA, 7SL, cytoplasmic 331, pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479345, CurrentID=0, Status='live', Symbol='RN7SL332P', Aliases='', description='RNA, 7SL, cytoplasmic 332, pseudogene', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479346, CurrentID=0, Status='live', Symbol='RN7SL336P', Aliases='', description='RNA, 7SL, cytoplasmic 336, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479347, CurrentID=0, Status='live', Symbol='RN7SL338P', Aliases='', description='RNA, 7SL, cytoplasmic 338, pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479348, CurrentID=0, Status='live', Symbol='RN7SL339P', Aliases='', description='RNA, 7SL, cytoplasmic 339, pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442101, CurrentID=0, Status='live', Symbol='OR7E162P', Aliases='OR7E77, OR7E77P, OST060', description='olfactory receptor family 7 subfamily E member 162 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 162 pseudogene|olfactory receptor, family 7, subfamily E, member 77 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3901506, end_position_on_the_genomic_accession=3902540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479350, CurrentID=0, Status='live', Symbol='RN7SL346P', Aliases='', description='RNA, 7SL, cytoplasmic 346, pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479351, CurrentID=0, Status='live', Symbol='RN7SL351P', Aliases='', description='RNA, 7SL, cytoplasmic 351, pseudogene', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442104, CurrentID=0, Status='live', Symbol='OR7E86P', Aliases='', description='olfactory receptor family 7 subfamily E member 86 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 86 pseudogene|seven transmembrane helix receptor', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9459122, end_position_on_the_genomic_accession=9460344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479353, CurrentID=0, Status='live', Symbol='RN7SL356P', Aliases='', description='RNA, 7SL, cytoplasmic 356, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479354, CurrentID=0, Status='live', Symbol='RN7SL359P', Aliases='', description='RNA, 7SL, cytoplasmic 359, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479355, CurrentID=0, Status='live', Symbol='RN7SL361P', Aliases='', description='RNA, 7SL, cytoplasmic 361, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442108, CurrentID=0, Status='live', Symbol='RPL22P13', Aliases='RPL22_7_478', description='ribosomal protein L22 pseudogene 13', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54221123, end_position_on_the_genomic_accession=54221472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  29994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=29994, CurrentID=0, Status='live', Symbol='BAZ2B', Aliases='WALp4', description='bromodomain adjacent to zinc finger domain 2B', other_designations='bromodomain adjacent to zinc finger domain, 2B', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159315312, end_position_on_the_genomic_accession=159712439, orientation='minus', exon_count=50, OMIM=605683, no_hdr0=''),
  106479358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479358, CurrentID=0, Status='live', Symbol='RN7SL372P', Aliases='', description='RNA, 7SL, cytoplasmic 372, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479359, CurrentID=0, Status='live', Symbol='RN7SL375P', Aliases='', description='RNA, 7SL, cytoplasmic 375, pseudogene', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442112, CurrentID=0, Status='live', Symbol='SERBP1P5', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene 5', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78180226, end_position_on_the_genomic_accession=78182145, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479361, CurrentID=0, Status='live', Symbol='RN7SL380P', Aliases='', description='RNA, 7SL, cytoplasmic 380, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479362, CurrentID=0, Status='live', Symbol='RN7SL381P', Aliases='', description='RNA, 7SL, cytoplasmic 381, pseudogene', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100187907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187907, CurrentID=0, Status='live', Symbol='TRAP', Aliases='', description='triiodothyronine receptor auxiliary protein', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=190445, no_hdr0=''),
  106479364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479364, CurrentID=0, Status='live', Symbol='RN7SL386P', Aliases='', description='RNA, 7SL, cytoplasmic 386, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442117, CurrentID=0, Status='live', Symbol='GALNTL6', Aliases='GALNACT20, GALNT17, GalNAc-T6L', description='polypeptide N-acetylgalactosaminyltransferase-like 6', other_designations='GaNTase 17|GalNAc transferase 17|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 17|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 20|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 6|galNAc-T17|polypeptide GalNAc transferase 17|polypeptide GalNAc transferase-like 6|pp-GaNTase 17|protein-UDP acetylgalactosaminyltransferase 17|putative polypeptide N-acetylgalactosaminyltransferase 17', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171813424, end_position_on_the_genomic_accession=173040408, orientation='plus', exon_count=21, OMIM=615138, no_hdr0=''),
  106479366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479366, CurrentID=0, Status='live', Symbol='RN7SL390P', Aliases='', description='RNA, 7SL, cytoplasmic 390, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479367, CurrentID=0, Status='live', Symbol='RN7SL394P', Aliases='', description='RNA, 7SL, cytoplasmic 394, pseudogene', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479368, CurrentID=0, Status='live', Symbol='RN7SL396P', Aliases='', description='RNA, 7SL, cytoplasmic 396, pseudogene', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479369, CurrentID=0, Status='live', Symbol='RN7SL397P', Aliases='', description='RNA, 7SL, cytoplasmic 397, pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479370, CurrentID=0, Status='live', Symbol='RN7SL399P', Aliases='', description='RNA, 7SL, cytoplasmic 399, pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479371, CurrentID=0, Status='live', Symbol='RN7SL402P', Aliases='', description='RNA, 7SL, cytoplasmic 402, pseudogene', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479372, CurrentID=0, Status='live', Symbol='RN7SL404P', Aliases='', description='RNA, 7SL, cytoplasmic 404, pseudogene', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479373, CurrentID=0, Status='live', Symbol='RN7SL405P', Aliases='', description='RNA, 7SL, cytoplasmic 405, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479374, CurrentID=0, Status='live', Symbol='RN7SL413P', Aliases='', description='RNA, 7SL, cytoplasmic 413, pseudogene', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479375, CurrentID=0, Status='live', Symbol='RN7SL415P', Aliases='', description='RNA, 7SL, cytoplasmic 415, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479376, CurrentID=0, Status='live', Symbol='RN7SL419P', Aliases='', description='RNA, 7SL, cytoplasmic 419, pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479377, CurrentID=0, Status='live', Symbol='RN7SL420P', Aliases='', description='RNA, 7SL, cytoplasmic 420, pseudogene', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479378, CurrentID=0, Status='live', Symbol='RN7SL423P', Aliases='', description='RNA, 7SL, cytoplasmic 423, pseudogene', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442131, CurrentID=0, Status='live', Symbol='ALG3P1', Aliases='', description='ALG3, alpha-1,3- mannosyltransferase pseudogene 1', other_designations='NOT56L pseudogene|asparagine-linked glycosylation 3, alpha-1,3- mannosyltransferase homolog pseudogene', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5375632, end_position_on_the_genomic_accession=5376844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479380, CurrentID=0, Status='live', Symbol='RN7SL429P', Aliases='', description='RNA, 7SL, cytoplasmic 429, pseudogene', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479381, CurrentID=0, Status='live', Symbol='RN7SL435P', Aliases='', description='RNA, 7SL, cytoplasmic 435, pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479382, CurrentID=0, Status='live', Symbol='RN7SL438P', Aliases='', description='RNA, 7SL, cytoplasmic 438, pseudogene', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479383, CurrentID=0, Status='live', Symbol='RN7SL441P', Aliases='', description='RNA, 7SL, cytoplasmic 441, pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479384, CurrentID=0, Status='live', Symbol='RN7SL444P', Aliases='', description='RNA, 7SL, cytoplasmic 444, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479385, CurrentID=0, Status='live', Symbol='RN7SL446P', Aliases='', description='RNA, 7SL, cytoplasmic 446, pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479386, CurrentID=0, Status='live', Symbol='RN7SL447P', Aliases='', description='RNA, 7SL, cytoplasmic 447, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479387, CurrentID=0, Status='live', Symbol='RN7SL452P', Aliases='', description='RNA, 7SL, cytoplasmic 452, pseudogene', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479388, CurrentID=0, Status='live', Symbol='RN7SL454P', Aliases='', description='RNA, 7SL, cytoplasmic 454, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479389, CurrentID=0, Status='live', Symbol='RN7SL455P', Aliases='', description='RNA, 7SL, cytoplasmic 455, pseudogene', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479390, CurrentID=0, Status='live', Symbol='RN7SL459P', Aliases='', description='RNA, 7SL, cytoplasmic 459, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479391, CurrentID=0, Status='live', Symbol='RN7SL462P', Aliases='', description='RNA, 7SL, cytoplasmic 462, pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479392, CurrentID=0, Status='live', Symbol='RN7SL465P', Aliases='', description='RNA, 7SL, cytoplasmic 465, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479393, CurrentID=0, Status='live', Symbol='RN7SL466P', Aliases='', description='RNA, 7SL, cytoplasmic 466, pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479394, CurrentID=0, Status='live', Symbol='RN7SL470P', Aliases='', description='RNA, 7SL, cytoplasmic 470, pseudogene', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479395, CurrentID=0, Status='live', Symbol='RN7SL472P', Aliases='', description='RNA, 7SL, cytoplasmic 472, pseudogene', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479396, CurrentID=0, Status='live', Symbol='RN7SL473P', Aliases='', description='RNA, 7SL, cytoplasmic 473, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479397, CurrentID=0, Status='live', Symbol='RN7SL477P', Aliases='', description='RNA, 7SL, cytoplasmic 477, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479398, CurrentID=0, Status='live', Symbol='RN7SL479P', Aliases='', description='RNA, 7SL, cytoplasmic 479, pseudogene', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479399, CurrentID=0, Status='live', Symbol='RN7SL483P', Aliases='', description='RNA, 7SL, cytoplasmic 483, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479400, CurrentID=0, Status='live', Symbol='RN7SL484P', Aliases='', description='RNA, 7SL, cytoplasmic 484, pseudogene', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479401, CurrentID=0, Status='live', Symbol='RN7SL488P', Aliases='', description='RNA, 7SL, cytoplasmic 488, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23452, CurrentID=0, Status='live', Symbol='ANGPTL2', Aliases='ARP2, HARP', description='angiopoietin like 2', other_designations='angiopoietin-like 2|angiopoietin-like protein 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127087349, end_position_on_the_genomic_accession=127122797, orientation='minus', exon_count=6, OMIM=605001, no_hdr0=''),
  106479403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479403, CurrentID=0, Status='live', Symbol='RN7SL493P', Aliases='', description='RNA, 7SL, cytoplasmic 493, pseudogene', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479404, CurrentID=0, Status='live', Symbol='RN7SL495P', Aliases='', description='RNA, 7SL, cytoplasmic 495, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479405, CurrentID=0, Status='live', Symbol='RN7SL498P', Aliases='', description='RNA, 7SL, cytoplasmic 498, pseudogene', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442158, CurrentID=0, Status='live', Symbol='RPS26P29', Aliases='RPS26_6_653', description='ribosomal protein S26 pseudogene 29', other_designations='40S ribosomal protein S26 pseudogene', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7506176, end_position_on_the_genomic_accession=7506628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479407, CurrentID=0, Status='live', Symbol='RN7SL502P', Aliases='', description='RNA, 7SL, cytoplasmic 502, pseudogene', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442160, CurrentID=0, Status='live', Symbol='RPL21P62', Aliases='RPL21_23_656', description='ribosomal protein L21 pseudogene 62', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10213850, end_position_on_the_genomic_accession=10214402, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533990, CurrentID=0, Status='live', Symbol='APOC4-APOC2', Aliases='', description='APOC4-APOC2 readthrough (NMD candidate)', other_designations='', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44942238, end_position_on_the_genomic_accession=44949565, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  442162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442162, CurrentID=0, Status='live', Symbol='RPS4XP7', Aliases='RPS4P7, RPS4X_3_659', description='ribosomal protein S4X pseudogene 7', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13521213, end_position_on_the_genomic_accession=13522116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479411, CurrentID=0, Status='live', Symbol='RN7SL516P', Aliases='', description='RNA, 7SL, cytoplasmic 516, pseudogene', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479412, CurrentID=0, Status='live', Symbol='RN7SL519P', Aliases='', description='RNA, 7SL, cytoplasmic 519, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479413, CurrentID=0, Status='live', Symbol='RN7SL521P', Aliases='', description='RNA, 7SL, cytoplasmic 521, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479414, CurrentID=0, Status='live', Symbol='RN7SL523P', Aliases='', description='RNA, 7SL, cytoplasmic 523, pseudogene', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442167, CurrentID=0, Status='live', Symbol='ASS1P1', Aliases='ASSP1', description='argininosuccinate synthetase 1 pseudogene 1', other_designations='argininosuccinate synthetase pseudogene 1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25023009, end_position_on_the_genomic_accession=25024591, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479416, CurrentID=0, Status='live', Symbol='RN7SL531P', Aliases='', description='RNA, 7SL, cytoplasmic 531, pseudogene', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479417, CurrentID=0, Status='live', Symbol='RN7SL535P', Aliases='', description='RNA, 7SL, cytoplasmic 535, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479418, CurrentID=0, Status='live', Symbol='RN7SL539P', Aliases='', description='RNA, 7SL, cytoplasmic 539, pseudogene', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401375, CurrentID=0, Status='live', Symbol='GTF2IRD2P1', Aliases='GTF2IRD2P', description='GTF2I repeat domain containing 2 pseudogene 1', other_designations='GTF2IRD2 pseudogene|general transcription factor II i repeat domain 2 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73242751, end_position_on_the_genomic_accession=73280119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479420, CurrentID=0, Status='live', Symbol='RN7SL546P', Aliases='', description='RNA, 7SL, cytoplasmic 546, pseudogene', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479421, CurrentID=0, Status='live', Symbol='RN7SL547P', Aliases='', description='RNA, 7SL, cytoplasmic 547, pseudogene', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479422, CurrentID=0, Status='live', Symbol='RN7SL551P', Aliases='', description='RNA, 7SL, cytoplasmic 551, pseudogene', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442175, CurrentID=0, Status='live', Symbol='RPLP2P1', Aliases='RPLP2_1_671, dJ408B20.3', description='ribosomal protein, large P2, pseudogene 1', other_designations='ribosomal protein LP2 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27965175, end_position_on_the_genomic_accession=27965456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479424, CurrentID=0, Status='live', Symbol='RN7SL555P', Aliases='', description='RNA, 7SL, cytoplasmic 555, pseudogene', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479425, CurrentID=0, Status='live', Symbol='RN7SL557P', Aliases='', description='RNA, 7SL, cytoplasmic 557, pseudogene', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479426, CurrentID=0, Status='live', Symbol='RN7SL560P', Aliases='', description='RNA, 7SL, cytoplasmic 560, pseudogene', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442179, CurrentID=0, Status='live', Symbol='OR1F12', Aliases='OR1F12PQ, OR6-12, hs6M1-35P, OR1F12', description='olfactory receptor family 1 subfamily F member 12', other_designations='olfactory receptor OR6-12|olfactory receptor, family 1, subfamily F, member 1 pseudogene|olfactory receptor, family 1, subfamily F, member 12 pseudogene|seven transmembrane helix receptor', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28073248, end_position_on_the_genomic_accession=28074224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479428, CurrentID=0, Status='live', Symbol='RN7SL564P', Aliases='', description='RNA, 7SL, cytoplasmic 564, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442181, CurrentID=0, Status='live', Symbol='RPSAP2', Aliases='LAMR1P2, RPSA_15_672, bA60E24.1', description='ribosomal protein SA pseudogene 2', other_designations='laminin receptor 1 pseudogene 2', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28731939, end_position_on_the_genomic_accession=28732974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479430, CurrentID=0, Status='live', Symbol='RN7SL570P', Aliases='', description='RNA, 7SL, cytoplasmic 570, pseudogene', other_designations='', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479431, CurrentID=0, Status='live', Symbol='RN7SL571P', Aliases='', description='RNA, 7SL, cytoplasmic 571, pseudogene', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442184, CurrentID=0, Status='live', Symbol='OR2B3', Aliases='6M1-1P, OR6-14, OR6-4, OR2B3', description='olfactory receptor family 2 subfamily B member 3', other_designations='hs6M1-1|olfactory receptor 6-4|olfactory receptor OR6-14', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29086208, end_position_on_the_genomic_accession=29087313, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442185, CurrentID=0, Status='live', Symbol='OR2J1', Aliases='6M1-4PP, OR6-15, OR6-5, dJ80I19.2, hs6M1-4, OR2J1', description='olfactory receptor family 2 subfamily J member 1 (gene/pseudogene)', other_designations='olfactory receptor OR6-15|olfactory receptor, family 2, subfamily J, member 1 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29100943, end_position_on_the_genomic_accession=29101878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442186, CurrentID=0, Status='live', Symbol='OR2J3', Aliases='6M1-3, C3HEXS, HS6M1-3, OR6-16, OR6-6, OR6.3.6, ORL671', description='olfactory receptor family 2 subfamily J member 3', other_designations='olfactory receptor 6-6|olfactory receptor OR6-16|olfactory receptor, family 2, subfamily J, member 3', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29111810, end_position_on_the_genomic_accession=29112884, orientation='plus', exon_count=1, OMIM=615016, no_hdr0=''),
  106479435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479435, CurrentID=0, Status='live', Symbol='RN7SL582P', Aliases='', description='RNA, 7SL, cytoplasmic 582, pseudogene', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442188, CurrentID=0, Status='live', Symbol='OR2J4P', Aliases='OR6-20, OR6-9, dJ80I19.5, hs6M1-5', description='olfactory receptor family 2 subfamily J member 4 pseudogene', other_designations='olfactory receptor OR6-20|olfactory receptor, family 2, subfamily J, member 4 pseudogene|seven transmembrane helix receptor', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29181510, end_position_on_the_genomic_accession=29182442, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442189, CurrentID=0, Status='live', Symbol='OR2H4P', Aliases='6M1-7P, OR2H4, OR6-21, OR6-3, dJ80I19.6, hs6M1-7', description='olfactory receptor family 2 subfamily H member 4 pseudogene', other_designations='olfactory receptor OR6-21 pseudogene|olfactory receptor, family 2, subfamily H, member 4 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29215236, end_position_on_the_genomic_accession=29216176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442190, CurrentID=0, Status='live', Symbol='OR2B4P', Aliases='hs6M1-22', description='olfactory receptor family 2 subfamily B member 4 pseudogene', other_designations='olfactory receptor, family 2, subfamily B, member 4 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29290596, end_position_on_the_genomic_accession=29291750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442191, CurrentID=0, Status='live', Symbol='OR14J1', Aliases='OR5U1, OR6-25, bA150A6.4, hs6M1-28', description='olfactory receptor family 14 subfamily J member 1', other_designations='olfactory receptor 5U1|olfactory receptor OR6-25|olfactory receptor, family 14, subfamily J, member 1|olfactory receptor, family 5, subfamily U member 1|olfactory receptor, family 5, subfamily U, member 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29306690, end_position_on_the_genomic_accession=29307655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479440, CurrentID=0, Status='live', Symbol='RN7SL597P', Aliases='', description='RNA, 7SL, cytoplasmic 597, pseudogene', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479441, CurrentID=0, Status='live', Symbol='RN7SL602P', Aliases='', description='RNA, 7SL, cytoplasmic 602, pseudogene', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442194, CurrentID=0, Status='live', Symbol='OR10C1', Aliases='OR10C1P, OR10C2, OR6-31, hs6M1-17', description='olfactory receptor family 10 subfamily C member 1 (gene/pseudogene)', other_designations='olfactory receptor 10C2|olfactory receptor OR6-31|olfactory receptor, family 10, subfamily C, member 1|olfactory receptor, family 10, subfamily C, member 2|seven transmembrane helix receptor', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29439939, end_position_on_the_genomic_accession=29440977, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442195, CurrentID=0, Status='live', Symbol='RPS17P1', Aliases='RPS17_3_674, RPS17p, dJ994E9.1', description='ribosomal protein S17 pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29489203, end_position_on_the_genomic_accession=29489693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479444, CurrentID=0, Status='live', Symbol='RN7SL610P', Aliases='', description='RNA, 7SL, cytoplasmic 610, pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442197, CurrentID=0, Status='live', Symbol='OR2I1P', Aliases='HS6M1-14, OR2I1, OR2I2, OR2I3P, OR2I4P, OR2I5P, OR2I6, OR2I7P, OR2I8P, OR2I9P, OR6-1, OR6-34', description='olfactory receptor family 2 subfamily I member 1 pseudogene', other_designations='olfactory receptor OR6-1 pseudogene|olfactory receptor OR6-34 pseudogene|olfactory receptor, family 2, subfamily I, member 1 pseudogene|olfactory receptor, family 2, subfamily I, member 2|olfactory receptor, family 2, subfamily I, member 3 pseudogene|olfactory receptor, family 2, subfamily I, member 4 pseudogene|olfactory receptor, family 2, subfamily I, member 5 pseudogene|olfactory receptor, family 2, subfamily I, member 6|olfactory receptor, family 2, subfamily I, member 7 pseudogene|olfactory receptor, family 2, subfamily I, member 8 pseudogene|olfactory receptor, family 2, subfamily I, member 9 pseudogene|seven transmembrane helix receptor', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29553105, end_position_on_the_genomic_accession=29554163, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479446, CurrentID=0, Status='live', Symbol='RN7SL615P', Aliases='', description='RNA, 7SL, cytoplasmic 615, pseudogene', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479447, CurrentID=0, Status='live', Symbol='RN7SL618P', Aliases='', description='RNA, 7SL, cytoplasmic 618, pseudogene', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479448, CurrentID=0, Status='live', Symbol='RN7SL620P', Aliases='', description='RNA, 7SL, cytoplasmic 620, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479449, CurrentID=0, Status='live', Symbol='RN7SL621P', Aliases='', description='RNA, 7SL, cytoplasmic 621, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479450, CurrentID=0, Status='live', Symbol='RN7SL625P', Aliases='', description='RNA, 7SL, cytoplasmic 625, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479451, CurrentID=0, Status='live', Symbol='RN7SL628P', Aliases='', description='RNA, 7SL, cytoplasmic 628, pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442204, CurrentID=0, Status='live', Symbol='MYL8P', Aliases='BING3', description='myosin light chain 8, pseudogene', other_designations='myosin, light chain 12B, regulatory pseudogene|myosin, light chain 8, pseudogene|myosin, light polypeptide 8, pseudogene', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33338892, end_position_on_the_genomic_accession=33339695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479453, CurrentID=0, Status='live', Symbol='RN7SL633P', Aliases='', description='RNA, 7SL, cytoplasmic 633, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442206, CurrentID=0, Status='live', Symbol='GPR166P', Aliases='GPCR, PGR9', description='G protein-coupled receptor 166 pseudogene', other_designations='growth hormone secretagogue receptor pseudogene', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36737110, end_position_on_the_genomic_accession=36737883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479455, CurrentID=0, Status='live', Symbol='RN7SL638P', Aliases='', description='RNA, 7SL, cytoplasmic 638, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479456, CurrentID=0, Status='live', Symbol='RN7SL640P', Aliases='', description='RNA, 7SL, cytoplasmic 640, pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479457, CurrentID=0, Status='live', Symbol='RN7SL645P', Aliases='', description='RNA, 7SL, cytoplasmic 645, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479458, CurrentID=0, Status='live', Symbol='RN7SL647P', Aliases='', description='RNA, 7SL, cytoplasmic 647, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479459, CurrentID=0, Status='live', Symbol='RN7SL648P', Aliases='', description='RNA, 7SL, cytoplasmic 648, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479460, CurrentID=0, Status='live', Symbol='RN7SL652P', Aliases='', description='RNA, 7SL, cytoplasmic 652, pseudogene', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442213, CurrentID=0, Status='live', Symbol='PTCHD4', Aliases='C6orf138, PTCH53, dJ402H5.2', description='patched domain containing 4', other_designations='patched domain-containing protein C6orf138', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47877421, end_position_on_the_genomic_accession=48111132, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  106479462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479462, CurrentID=0, Status='live', Symbol='RN7SL659P', Aliases='', description='RNA, 7SL, cytoplasmic 659, pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479463, CurrentID=0, Status='live', Symbol='RN7SL660P', Aliases='', description='RNA, 7SL, cytoplasmic 660, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442216, CurrentID=0, Status='live', Symbol='RPS17P5', Aliases='RPS17L4, RPS17_4_700, dJ753D5.2', description='ribosomal protein S17 pseudogene 5', other_designations='ribosomal protein S17-like 4', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50857209, end_position_on_the_genomic_accession=50857687, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  30012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30012, CurrentID=0, Status='live', Symbol='TLX3', Aliases='HOX11L2, RNX', description='T-cell leukemia homeobox 3', other_designations='T-cell leukemia, homeobox 3|homeo box 11-like 2|homeobox protein Hox-11L2', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171309284, end_position_on_the_genomic_accession=171312134, orientation='plus', exon_count=3, OMIM=604640, no_hdr0=''),
  442218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442218, CurrentID=0, Status='live', Symbol='RPL31P33', Aliases='RPL31_13_704', description='ribosomal protein L31 pseudogene 33', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53368670, end_position_on_the_genomic_accession=53369026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100188011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188011, CurrentID=0, Status='live', Symbol='DIP', Aliases='', description='interstitial pneumonitis, desquamative, familial', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=263000, no_hdr0=''),
  106479468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479468, CurrentID=0, Status='live', Symbol='RN7SL674P', Aliases='', description='RNA, 7SL, cytoplasmic 674, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479469, CurrentID=0, Status='live', Symbol='RN7SL676P', Aliases='', description='RNA, 7SL, cytoplasmic 676, pseudogene', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479470, CurrentID=0, Status='live', Symbol='RN7SL677P', Aliases='', description='RNA, 7SL, cytoplasmic 677, pseudogene', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479471, CurrentID=0, Status='live', Symbol='RN7SL681P', Aliases='', description='RNA, 7SL, cytoplasmic 681, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479472, CurrentID=0, Status='live', Symbol='RN7SL684P', Aliases='', description='RNA, 7SL, cytoplasmic 684, pseudogene', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479473, CurrentID=0, Status='live', Symbol='RN7SL685P', Aliases='', description='RNA, 7SL, cytoplasmic 685, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479474, CurrentID=0, Status='live', Symbol='RN7SL688P', Aliases='', description='RNA, 7SL, cytoplasmic 688, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442227, CurrentID=0, Status='live', Symbol='EEF1B2P5', Aliases='', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 5', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63480078, end_position_on_the_genomic_accession=63481926, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479476, CurrentID=0, Status='live', Symbol='RN7SL692P', Aliases='', description='RNA, 7SL, cytoplasmic 692, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442229, CurrentID=0, Status='live', Symbol='SLC25A51P1', Aliases='MCART3P, bA707M13.1', description='solute carrier family 25, member 51 pseudogene 1', other_designations='mitochondrial carrier triple repeat 3 pseudogene', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=65787879, end_position_on_the_genomic_accession=65789483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479478, CurrentID=0, Status='live', Symbol='RN7SL700P', Aliases='', description='RNA, 7SL, cytoplasmic 700, pseudogene', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442231, CurrentID=0, Status='live', Symbol='GAPDHP63', Aliases='GAPDHL8, GAPDL8', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 63', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 8, pseudogene', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79952806, end_position_on_the_genomic_accession=79954082, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442232, CurrentID=0, Status='live', Symbol='RPL17P25', Aliases='RPL17_6_718', description='ribosomal protein L17 pseudogene 25', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80373980, end_position_on_the_genomic_accession=80374604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479481, CurrentID=0, Status='live', Symbol='RN7SL708P', Aliases='', description='RNA, 7SL, cytoplasmic 708, pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479482, CurrentID=0, Status='live', Symbol='RN7SL711P', Aliases='', description='RNA, 7SL, cytoplasmic 711, pseudogene', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479483, CurrentID=0, Status='live', Symbol='RN7SL715P', Aliases='', description='RNA, 7SL, cytoplasmic 715, pseudogene', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479484, CurrentID=0, Status='live', Symbol='RN7SL716P', Aliases='', description='RNA, 7SL, cytoplasmic 716, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442237, CurrentID=0, Status='live', Symbol='RPS7P8', Aliases='RPS7_4_725', description='ribosomal protein S7 pseudogene 8', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96648455, end_position_on_the_genomic_accession=96649107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479486, CurrentID=0, Status='live', Symbol='RN7SL722P', Aliases='', description='RNA, 7SL, cytoplasmic 722, pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479487, CurrentID=0, Status='live', Symbol='RN7SL724P', Aliases='', description='RNA, 7SL, cytoplasmic 724, pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479488, CurrentID=0, Status='live', Symbol='RN7SL726P', Aliases='', description='RNA, 7SL, cytoplasmic 726, pseudogene', other_designations='', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479489, CurrentID=0, Status='live', Symbol='RN7SL733P', Aliases='', description='RNA, 7SL, cytoplasmic 733, pseudogene', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442242, CurrentID=0, Status='live', Symbol='RPL7P28', Aliases='RPL7_12_736', description='ribosomal protein L7 pseudogene 28', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109327175, end_position_on_the_genomic_accession=109327893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479491, CurrentID=0, Status='live', Symbol='RN7SL740P', Aliases='', description='RNA, 7SL, cytoplasmic 740, pseudogene', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479492, CurrentID=0, Status='live', Symbol='RN7SL743P', Aliases='', description='RNA, 7SL, cytoplasmic 743, pseudogene', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479493, CurrentID=0, Status='live', Symbol='RN7SL744P', Aliases='', description='RNA, 7SL, cytoplasmic 744, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479494, CurrentID=0, Status='live', Symbol='RN7SL749P', Aliases='', description='RNA, 7SL, cytoplasmic 749, pseudogene', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442247, CurrentID=0, Status='live', Symbol='RFPL4B', Aliases='RNF211', description='ret finger protein-like 4B', other_designations='RING finger protein 211', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112347330, end_position_on_the_genomic_accession=112351296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479496, CurrentID=0, Status='live', Symbol='RN7SL753P', Aliases='', description='RNA, 7SL, cytoplasmic 753, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479497, CurrentID=0, Status='live', Symbol='RN7SL754P', Aliases='', description='RNA, 7SL, cytoplasmic 754, pseudogene', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479498, CurrentID=0, Status='live', Symbol='RN7SL759P', Aliases='', description='RNA, 7SL, cytoplasmic 759, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479499, CurrentID=0, Status='live', Symbol='RN7SL761P', Aliases='', description='RNA, 7SL, cytoplasmic 761, pseudogene', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479500, CurrentID=0, Status='live', Symbol='RN7SL762P', Aliases='', description='RNA, 7SL, cytoplasmic 762, pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442253, CurrentID=0, Status='live', Symbol='NEPNP', Aliases='NEPN, NPN', description='nephrocan, pseudogene', other_designations='NEPN pseudogene|asporin pseudogene', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117602928, end_position_on_the_genomic_accession=117645087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479502, CurrentID=0, Status='live', Symbol='RN7SL769P', Aliases='', description='RNA, 7SL, cytoplasmic 769, pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442255, CurrentID=0, Status='live', Symbol='SLC25A5P7', Aliases='ANTP7', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 7', other_designations='adenine nucleotide translocator pseudogene 7 (clone 2)', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121653739, end_position_on_the_genomic_accession=121654948, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442256, CurrentID=0, Status='live', Symbol='PPP1R14BP5', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 14B pseudogene 5', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 14B pseudogene 5', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126257924, end_position_on_the_genomic_accession=126258330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442257, CurrentID=0, Status='live', Symbol='RPS4XP9', Aliases='RPS4P9, RPS4X_4_749', description='ribosomal protein S4X pseudogene 9', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126683013, end_position_on_the_genomic_accession=126683909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479506, CurrentID=0, Status='live', Symbol='RN7SL786P', Aliases='', description='RNA, 7SL, cytoplasmic 786, pseudogene', other_designations='', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479507, CurrentID=0, Status='live', Symbol='RN7SL787P', Aliases='', description='RNA, 7SL, cytoplasmic 787, pseudogene', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442260, CurrentID=0, Status='live', Symbol='RPL23AP46', Aliases='RPL23A_17_757', description='ribosomal protein L23a pseudogene 46', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132997570, end_position_on_the_genomic_accession=132998040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479509, CurrentID=0, Status='live', Symbol='RN7SL792P', Aliases='', description='RNA, 7SL, cytoplasmic 792, pseudogene', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479510, CurrentID=0, Status='live', Symbol='RN7SL794P', Aliases='', description='RNA, 7SL, cytoplasmic 794, pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479511, CurrentID=0, Status='live', Symbol='RN7SL795P', Aliases='', description='RNA, 7SL, cytoplasmic 795, pseudogene', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479512, CurrentID=0, Status='live', Symbol='RN7SL800P', Aliases='', description='RNA, 7SL, cytoplasmic 800, pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479513, CurrentID=0, Status='live', Symbol='RN7SL802P', Aliases='', description='RNA, 7SL, cytoplasmic 802, pseudogene', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442266, CurrentID=0, Status='live', Symbol='YAP1P1', Aliases='', description='Yes associated protein 1 pseudogene 1', other_designations='Yes-associated protein 1 pseudogene 1', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=147406889, end_position_on_the_genomic_accession=147408281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442267, CurrentID=0, Status='live', Symbol='CCT7P1', Aliases='CCT7-2P', description='chaperonin containing TCP1, subunit 7 (eta) pseudogene 1', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149879894, end_position_on_the_genomic_accession=149881722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479516, CurrentID=0, Status='live', Symbol='RN7SL809P', Aliases='', description='RNA, 7SL, cytoplasmic 809, pseudogene', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479517, CurrentID=0, Status='live', Symbol='RN7SL811P', Aliases='', description='RNA, 7SL, cytoplasmic 811, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442270, CurrentID=0, Status='live', Symbol='RPS12P11', Aliases='RPS12_6_769', description='ribosomal protein S12 pseudogene 11', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151225528, end_position_on_the_genomic_accession=151226029, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479519, CurrentID=0, Status='live', Symbol='RN7SL815P', Aliases='', description='RNA, 7SL, cytoplasmic 815, pseudogene', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479520, CurrentID=0, Status='live', Symbol='RN7SL818P', Aliases='', description='RNA, 7SL, cytoplasmic 818, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479521, CurrentID=0, Status='live', Symbol='RN7SL822P', Aliases='', description='RNA, 7SL, cytoplasmic 822, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479522, CurrentID=0, Status='live', Symbol='RN7SL823P', Aliases='', description='RNA, 7SL, cytoplasmic 823, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479523, CurrentID=0, Status='live', Symbol='RN7SL826P', Aliases='', description='RNA, 7SL, cytoplasmic 826, pseudogene', other_designations='', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479524, CurrentID=0, Status='live', Symbol='RN7SL832P', Aliases='', description='RNA, 7SL, cytoplasmic 832, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479525, CurrentID=0, Status='live', Symbol='RN7SL833P', Aliases='', description='RNA, 7SL, cytoplasmic 833, pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479526, CurrentID=0, Status='live', Symbol='RN7SL835P', Aliases='', description='RNA, 7SL, cytoplasmic 835, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8177, CurrentID=0, Status='live', Symbol='LPSA', Aliases='D19S381E', description='Oncogene liposarcoma (DNA segment, single copy, expressed, probes', other_designations='Oncogene liposarcoma (DNA segment, single copy, expressed, probes MC15, MC6)', map_location='19p13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164953, no_hdr0=''),
  106479528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479528, CurrentID=0, Status='live', Symbol='RN7SL840P', Aliases='', description='RNA, 7SL, cytoplasmic 840, pseudogene', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479529, CurrentID=0, Status='live', Symbol='RN7SL842P', Aliases='', description='RNA, 7SL, cytoplasmic 842, pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479530, CurrentID=0, Status='live', Symbol='RN7SL846P', Aliases='', description='RNA, 7SL, cytoplasmic 846, pseudogene', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479531, CurrentID=0, Status='live', Symbol='RN7SL849P', Aliases='', description='RNA, 7SL, cytoplasmic 849, pseudogene', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479532, CurrentID=0, Status='live', Symbol='RN7SL850P', Aliases='', description='RNA, 7SL, cytoplasmic 850, pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479533, CurrentID=0, Status='live', Symbol='RN7SL856P', Aliases='', description='RNA, 7SL, cytoplasmic 856, pseudogene', other_designations='tmp_locus_48', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479534, CurrentID=0, Status='live', Symbol='RN7SL861P', Aliases='', description='RNA, 7SL, cytoplasmic 861, pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479535, CurrentID=0, Status='live', Symbol='RN7SL862P', Aliases='', description='RNA, 7SL, cytoplasmic 862, pseudogene', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479536, CurrentID=0, Status='live', Symbol='RN7SL864P', Aliases='', description='RNA, 7SL, cytoplasmic 864, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479537, CurrentID=0, Status='live', Symbol='RN7SL868P', Aliases='', description='RNA, 7SL, cytoplasmic 868, pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479538, CurrentID=0, Status='live', Symbol='VTRNA2-2P', Aliases='', description='vault RNA 2-2, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479539, CurrentID=0, Status='live', Symbol='RNU4ATAC4P', Aliases='', description='RNA, U4atac small nuclear 4, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479540, CurrentID=0, Status='live', Symbol='RNU4ATAC6P', Aliases='', description='RNA, U4atac small nuclear 6, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479541, CurrentID=0, Status='live', Symbol='RNU4ATAC9P', Aliases='', description='RNA, U4atac small nuclear 9, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479542, CurrentID=0, Status='live', Symbol='RNU4ATAC10P', Aliases='', description='RNA, U4atac small nuclear 10, pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479543, CurrentID=0, Status='live', Symbol='RNU4ATAC13P', Aliases='', description='RNA, U4atac small nuclear 13, pseudogene', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479544, CurrentID=0, Status='live', Symbol='RNU4ATAC15P', Aliases='', description='RNA, U4atac small nuclear 15, pseudogene', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479545, CurrentID=0, Status='live', Symbol='RNU6ATAC6P', Aliases='', description='RNA, U6atac small nuclear 6, pseudogene', other_designations='', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479546, CurrentID=0, Status='live', Symbol='RNU6ATAC7P', Aliases='', description='RNA, U6atac small nuclear 7, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479547, CurrentID=0, Status='live', Symbol='RNU6ATAC9P', Aliases='', description='RNA, U6atac small nuclear 9, pseudogene', other_designations='', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479548, CurrentID=0, Status='live', Symbol='RNU6ATAC12P', Aliases='', description='RNA, U6atac small nuclear 12, pseudogene', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479549, CurrentID=0, Status='live', Symbol='RNU6ATAC13P', Aliases='', description='RNA, U6atac small nuclear 13, pseudogene', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479550, CurrentID=0, Status='live', Symbol='RNU6ATAC16P', Aliases='', description='RNA, U6atac small nuclear 16, pseudogene', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401397, CurrentID=0, Status='live', Symbol='LINC00998', Aliases='', description='long intergenic non-protein coding RNA 998', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=113116718, end_position_on_the_genomic_accession=113118582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479552, CurrentID=0, Status='live', Symbol='RNU6ATAC21P', Aliases='', description='RNA, U6atac small nuclear 21, pseudogene', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479553, CurrentID=0, Status='live', Symbol='RNU6ATAC22P', Aliases='', description='RNA, U6atac small nuclear 22, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479554, CurrentID=0, Status='live', Symbol='RNU6ATAC24P', Aliases='', description='RNA, U6atac small nuclear 24, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479555, CurrentID=0, Status='live', Symbol='RNU6ATAC28P', Aliases='', description='RNA, U6atac small nuclear 28, pseudogene', other_designations='', map_location='14q31.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479556, CurrentID=0, Status='live', Symbol='RNU6ATAC32P', Aliases='', description='RNA, U6atac small nuclear 32, pseudogene', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479557, CurrentID=0, Status='live', Symbol='RNU6ATAC34P', Aliases='', description='RNA, U6atac small nuclear 34, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479558, CurrentID=0, Status='live', Symbol='RNU6ATAC37P', Aliases='', description='RNA, U6atac small nuclear 37, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479559, CurrentID=0, Status='live', Symbol='RNU6ATAC38P', Aliases='', description='RNA, U6atac small nuclear 38, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479560, CurrentID=0, Status='live', Symbol='RNU6ATAC40P', Aliases='', description='RNA, U6atac small nuclear 40, pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479561, CurrentID=0, Status='live', Symbol='RNU11-2P', Aliases='', description='RNA, U11 small nuclear 2, pseudogene', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479562, CurrentID=0, Status='live', Symbol='RNU11-3P', Aliases='', description='RNA, U11 small nuclear 3, pseudogene', other_designations='', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479563, CurrentID=0, Status='live', Symbol='RNU11-5P', Aliases='', description='RNA, U11 small nuclear 5, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479564, CurrentID=0, Status='live', Symbol='RNU4-12P', Aliases='', description='RNA, U4 small nuclear 12, pseudogene', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479565, CurrentID=0, Status='live', Symbol='RNU4-15P', Aliases='', description='RNA, U4 small nuclear 15, pseudogene', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479566, CurrentID=0, Status='live', Symbol='RNU4-16P', Aliases='', description='RNA, U4 small nuclear 16, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442319, CurrentID=0, Status='live', Symbol='ZNF727', Aliases='ZNF727P', description='zinc finger protein 727', other_designations='zinc finger protein 727 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64044801, end_position_on_the_genomic_accession=64085184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106479568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479568, CurrentID=0, Status='live', Symbol='RNU4-23P', Aliases='', description='RNA, U4 small nuclear 23, pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479569, CurrentID=0, Status='live', Symbol='RNU4-26P', Aliases='', description='RNA, U4 small nuclear 26, pseudogene', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479570, CurrentID=0, Status='live', Symbol='RNU4-28P', Aliases='', description='RNA, U4 small nuclear 28, pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479571, CurrentID=0, Status='live', Symbol='RNU4-30P', Aliases='', description='RNA, U4 small nuclear 30, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479572, CurrentID=0, Status='live', Symbol='RNU4-32P', Aliases='', description='RNA, U4 small nuclear 32, pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442325, CurrentID=0, Status='live', Symbol='EEF1DP4', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 4', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64862951, end_position_on_the_genomic_accession=64864450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442326, CurrentID=0, Status='live', Symbol='RPL31P38', Aliases='RPL31_17_813', description='ribosomal protein L31 pseudogene 38', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66914570, end_position_on_the_genomic_accession=66914983, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479575, CurrentID=0, Status='live', Symbol='RNU4-39P', Aliases='', description='RNA, U4 small nuclear 39, pseudogene', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479576, CurrentID=0, Status='live', Symbol='RNU4-40P', Aliases='', description='RNA, U4 small nuclear 40, pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479577, CurrentID=0, Status='live', Symbol='RNU4-42P', Aliases='', description='RNA, U4 small nuclear 42, pseudogene', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479578, CurrentID=0, Status='live', Symbol='RNU4-44P', Aliases='', description='RNA, U4 small nuclear 44, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479579, CurrentID=0, Status='live', Symbol='RNU4-47P', Aliases='', description='RNA, U4 small nuclear 47, pseudogene', other_designations='', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479580, CurrentID=0, Status='live', Symbol='RNU4-48P', Aliases='', description='RNA, U4 small nuclear 48, pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479581, CurrentID=0, Status='live', Symbol='RNU4-51P', Aliases='', description='RNA, U4 small nuclear 51, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479582, CurrentID=0, Status='live', Symbol='RNU4-53P', Aliases='', description='RNA, U4 small nuclear 53, pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479583, CurrentID=0, Status='live', Symbol='RNU4-54P', Aliases='', description='RNA, U4 small nuclear 54, pseudogene', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479584, CurrentID=0, Status='live', Symbol='RNU4-57P', Aliases='', description='RNA, U4 small nuclear 57, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479585, CurrentID=0, Status='live', Symbol='RNU4-60P', Aliases='', description='RNA, U4 small nuclear 60, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442338, CurrentID=0, Status='live', Symbol='IRS3P', Aliases='IRS3L', description='insulin receptor substrate 3, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100570193, end_position_on_the_genomic_accession=100571401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479587, CurrentID=0, Status='live', Symbol='RNU4-63P', Aliases='', description='RNA, U4 small nuclear 63, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479588, CurrentID=0, Status='live', Symbol='RNU4-66P', Aliases='', description='RNA, U4 small nuclear 66, pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479589, CurrentID=0, Status='live', Symbol='RNU4-68P', Aliases='', description='RNA, U4 small nuclear 68, pseudogene', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479590, CurrentID=0, Status='live', Symbol='RNU4-69P', Aliases='', description='RNA, U4 small nuclear 69, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479591, CurrentID=0, Status='live', Symbol='RNU4-72P', Aliases='', description='RNA, U4 small nuclear 72, pseudogene', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479592, CurrentID=0, Status='live', Symbol='RNU4-74P', Aliases='', description='RNA, U4 small nuclear 74, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479593, CurrentID=0, Status='live', Symbol='RNU4-75P', Aliases='', description='RNA, U4 small nuclear 75, pseudogene', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479594, CurrentID=0, Status='live', Symbol='RNU4-78P', Aliases='', description='RNA, U4 small nuclear 78, pseudogene', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479595, CurrentID=0, Status='live', Symbol='RNU4-81P', Aliases='', description='RNA, U4 small nuclear 81, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479596, CurrentID=0, Status='live', Symbol='RNU4-83P', Aliases='', description='RNA, U4 small nuclear 83, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479597, CurrentID=0, Status='live', Symbol='RNU4-85P', Aliases='', description='RNA, U4 small nuclear 85, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479598, CurrentID=0, Status='live', Symbol='RNU4-87P', Aliases='', description='RNA, U4 small nuclear 87, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479599, CurrentID=0, Status='live', Symbol='RNU4-88P', Aliases='', description='RNA, U4 small nuclear 88, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479600, CurrentID=0, Status='live', Symbol='RNU4-90P', Aliases='', description='RNA, U4 small nuclear 90, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479601, CurrentID=0, Status='live', Symbol='RNU4-92P', Aliases='', description='RNA, U4 small nuclear 92, pseudogene', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479602, CurrentID=0, Status='live', Symbol='RPPH1-3P', Aliases='', description='ribonuclease P RNA component H1, 3 pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479603, CurrentID=0, Status='live', Symbol='RNU6-84P', Aliases='', description='RNA, U6 small nuclear 84, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479604, CurrentID=0, Status='live', Symbol='RNU6-86P', Aliases='', description='RNA, U6 small nuclear 86, pseudogene', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479605, CurrentID=0, Status='live', Symbol='RNU6-88P', Aliases='', description='RNA, U6 small nuclear 88, pseudogene', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479606, CurrentID=0, Status='live', Symbol='RNU6-91P', Aliases='', description='RNA, U6 small nuclear 91, pseudogene', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479607, CurrentID=0, Status='live', Symbol='RNU6-92P', Aliases='', description='RNA, U6 small nuclear 92, pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479608, CurrentID=0, Status='live', Symbol='RNU6-95P', Aliases='', description='RNA, U6 small nuclear 95, pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442361, CurrentID=0, Status='live', Symbol='OR2A2', Aliases='OR2A17PP, OR7-11, OST008, OR2A2', description='olfactory receptor family 2 subfamily A member 2', other_designations='olfactory receptor 2A17|olfactory receptor OR7-11|olfactory receptor, family 2, subfamily A, member 17 pseudogene|olfactory receptor, family 2, subfamily A, member 2 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144109583, end_position_on_the_genomic_accession=144110539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479610, CurrentID=0, Status='live', Symbol='RNU6-99P', Aliases='', description='RNA, U6 small nuclear 99, pseudogene', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479611, CurrentID=0, Status='live', Symbol='RNU6-101P', Aliases='', description='RNA, U6 small nuclear 101, pseudogene', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479612, CurrentID=0, Status='live', Symbol='RNU6-104P', Aliases='', description='RNA, U6 small nuclear 104, pseudogene', other_designations='', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479613, CurrentID=0, Status='live', Symbol='RNU6-106P', Aliases='', description='RNA, U6 small nuclear 106, pseudogene', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480570, CurrentID=0, Status='live', Symbol='RNU6-320P', Aliases='', description='RNA, U6 small nuclear 320, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479615, CurrentID=0, Status='live', Symbol='RNU6-110P', Aliases='', description='RNA, U6 small nuclear 110, pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479616, CurrentID=0, Status='live', Symbol='RNU6-113P', Aliases='', description='RNA, U6 small nuclear 113, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479617, CurrentID=0, Status='live', Symbol='RNU6-115P', Aliases='', description='RNA, U6 small nuclear 115, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479618, CurrentID=0, Status='live', Symbol='RNU6-118P', Aliases='', description='RNA, U6 small nuclear 118, pseudogene', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479619, CurrentID=0, Status='live', Symbol='RNU6-120P', Aliases='', description='RNA, U6 small nuclear 120, pseudogene', other_designations='', map_location='18q11.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479620, CurrentID=0, Status='live', Symbol='RNU6-123P', Aliases='', description='RNA, U6 small nuclear 123, pseudogene', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479621, CurrentID=0, Status='live', Symbol='RNU6-125P', Aliases='', description='RNA, U6 small nuclear 125, pseudogene', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479622, CurrentID=0, Status='live', Symbol='RNU6-126P', Aliases='', description='RNA, U6 small nuclear 126, pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8193, CurrentID=0, Status='live', Symbol='DPF1', Aliases='BAF45b, NEUD4, neuro-d4', description='D4, zinc and double PHD fingers family 1', other_designations='BRG1-associated factor 45B|neuro-d4 homolog', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38211006, end_position_on_the_genomic_accession=38229714, orientation='minus', exon_count=14, OMIM=601670, no_hdr0=''),
  106479624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479624, CurrentID=0, Status='live', Symbol='RNU6-131P', Aliases='', description='RNA, U6 small nuclear 131, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479625, CurrentID=0, Status='live', Symbol='RNU6-135P', Aliases='', description='RNA, U6 small nuclear 135, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479626, CurrentID=0, Status='live', Symbol='RNU6-136P', Aliases='', description='RNA, U6 small nuclear 136, pseudogene', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479627, CurrentID=0, Status='live', Symbol='RNU6-138P', Aliases='', description='RNA, U6 small nuclear 138, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479628, CurrentID=0, Status='live', Symbol='RNU6-142P', Aliases='', description='RNA, U6 small nuclear 142, pseudogene', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479629, CurrentID=0, Status='live', Symbol='RNU6-143P', Aliases='', description='RNA, U6 small nuclear 143, pseudogene', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479630, CurrentID=0, Status='live', Symbol='RNU6-145P', Aliases='', description='RNA, U6 small nuclear 145, pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479631, CurrentID=0, Status='live', Symbol='RNU6-148P', Aliases='', description='RNA, U6 small nuclear 148, pseudogene', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442384, CurrentID=0, Status='live', Symbol='VENTXP5', Aliases='', description='VENT homeobox pseudogene 5', other_designations='VENT homeobox homolog pseudogene|VENT-like homeobox pseudogene 5', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33721194, end_position_on_the_genomic_accession=33723089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479633, CurrentID=0, Status='live', Symbol='RNU6-151P', Aliases='', description='RNA, U6 small nuclear 151, pseudogene', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479634, CurrentID=0, Status='live', Symbol='RNU6-153P', Aliases='', description='RNA, U6 small nuclear 153, pseudogene', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479635, CurrentID=0, Status='live', Symbol='RNU6-156P', Aliases='', description='RNA, U6 small nuclear 156, pseudogene', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442388, CurrentID=0, Status='live', Symbol='SDR16C6P', Aliases='SDR16C6', description='short chain dehydrogenase/reductase family 16C, member 6, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56374718, end_position_on_the_genomic_accession=56390710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479637, CurrentID=0, Status='live', Symbol='RNU6-160P', Aliases='', description='RNA, U6 small nuclear 160, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479638, CurrentID=0, Status='live', Symbol='RNU6-163P', Aliases='', description='RNA, U6 small nuclear 163, pseudogene', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479639, CurrentID=0, Status='live', Symbol='RNU6-164P', Aliases='', description='RNA, U6 small nuclear 164, pseudogene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479640, CurrentID=0, Status='live', Symbol='RNU6-168P', Aliases='', description='RNA, U6 small nuclear 168, pseudogene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479641, CurrentID=0, Status='live', Symbol='RNU6-170P', Aliases='', description='RNA, U6 small nuclear 170, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479642, CurrentID=0, Status='live', Symbol='RNU6-172P', Aliases='', description='RNA, U6 small nuclear 172, pseudogene', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479643, CurrentID=0, Status='live', Symbol='RNU6-173P', Aliases='', description='RNA, U6 small nuclear 173, pseudogene', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479644, CurrentID=0, Status='live', Symbol='RNU6-176P', Aliases='', description='RNA, U6 small nuclear 176, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479645, CurrentID=0, Status='live', Symbol='RNU6-178P', Aliases='', description='RNA, U6 small nuclear 178, pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479646, CurrentID=0, Status='live', Symbol='RNU6-179P', Aliases='', description='RNA, U6 small nuclear 179, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479647, CurrentID=0, Status='live', Symbol='RNU6-182P', Aliases='', description='RNA, U6 small nuclear 182, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479648, CurrentID=0, Status='live', Symbol='RNU6-184P', Aliases='', description='RNA, U6 small nuclear 184, pseudogene', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479649, CurrentID=0, Status='live', Symbol='RNU6-188P', Aliases='', description='RNA, U6 small nuclear 188, pseudogene', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479650, CurrentID=0, Status='live', Symbol='RNU6-189P', Aliases='', description='RNA, U6 small nuclear 189, pseudogene', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479651, CurrentID=0, Status='live', Symbol='RNU6-191P', Aliases='', description='RNA, U6 small nuclear 191, pseudogene', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479652, CurrentID=0, Status='live', Symbol='RNU6-194P', Aliases='', description='RNA, U6 small nuclear 194, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479653, CurrentID=0, Status='live', Symbol='RNU6-195P', Aliases='', description='RNA, U6 small nuclear 195, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479654, CurrentID=0, Status='live', Symbol='RNU6-197P', Aliases='', description='RNA, U6 small nuclear 197, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479655, CurrentID=0, Status='live', Symbol='RNU6-200P', Aliases='', description='RNA, U6 small nuclear 200, pseudogene', other_designations='', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479656, CurrentID=0, Status='live', Symbol='RNU6-202P', Aliases='', description='RNA, U6 small nuclear 202, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479657, CurrentID=0, Status='live', Symbol='RNU6-203P', Aliases='', description='RNA, U6 small nuclear 203, pseudogene', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479658, CurrentID=0, Status='live', Symbol='RNU6-208P', Aliases='', description='RNA, U6 small nuclear 208, pseudogene', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479659, CurrentID=0, Status='live', Symbol='RNU6-210P', Aliases='', description='RNA, U6 small nuclear 210, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479660, CurrentID=0, Status='live', Symbol='RNU6-211P', Aliases='', description='RNA, U6 small nuclear 211, pseudogene', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479661, CurrentID=0, Status='live', Symbol='RNU6-214P', Aliases='', description='RNA, U6 small nuclear 214, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479662, CurrentID=0, Status='live', Symbol='RNU6-216P', Aliases='', description='RNA, U6 small nuclear 216, pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479663, CurrentID=0, Status='live', Symbol='RNU6-218P', Aliases='', description='RNA, U6 small nuclear 218, pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442416, CurrentID=0, Status='live', Symbol='ATP5A1P5', Aliases='ATP5A1P1, ATP5A1P6, ATP5AL1, ATP5AP1', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 5', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 1|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 6|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1, cardiac muscle-like 1 pseudogene|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, isoform 1, cardiac muscle-like 1', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42410727, end_position_on_the_genomic_accession=42412157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479665, CurrentID=0, Status='live', Symbol='RNU6-222P', Aliases='', description='RNA, U6 small nuclear 222, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479666, CurrentID=0, Status='live', Symbol='RNU6-224P', Aliases='', description='RNA, U6 small nuclear 224, pseudogene', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479667, CurrentID=0, Status='live', Symbol='RNU6-225P', Aliases='', description='RNA, U6 small nuclear 225, pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479668, CurrentID=0, Status='live', Symbol='RNU6-228P', Aliases='', description='RNA, U6 small nuclear 228, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442421, CurrentID=0, Status='live', Symbol='PTGER4P2-CDK2AP2P2', Aliases='PTGER4P2', description='PTGER4P2-CDK2AP2P2 readthrough transcribed pseudogene', other_designations='prostaglandin E receptor 4 (subtype EP4) pseudogene 2', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62838445, end_position_on_the_genomic_accession=62847206, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479670, CurrentID=0, Status='live', Symbol='RNU6-234P', Aliases='', description='RNA, U6 small nuclear 234, pseudogene', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479671, CurrentID=0, Status='live', Symbol='RNU6-235P', Aliases='', description='RNA, U6 small nuclear 235, pseudogene', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479672, CurrentID=0, Status='live', Symbol='RNU6-237P', Aliases='', description='RNA, U6 small nuclear 237, pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442425, CurrentID=0, Status='live', Symbol='FOXB2', Aliases='bA159H20.4', description='forkhead box B2', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77019655, end_position_on_the_genomic_accession=77020953, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442426, CurrentID=0, Status='live', Symbol='RPS2P34', Aliases='RPS2_16_990', description='ribosomal protein S2 pseudogene 34', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82130734, end_position_on_the_genomic_accession=82131666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479675, CurrentID=0, Status='live', Symbol='RNU6-243P', Aliases='', description='RNA, U6 small nuclear 243, pseudogene', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479676, CurrentID=0, Status='live', Symbol='RNU6-247P', Aliases='', description='RNA, U6 small nuclear 247, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479677, CurrentID=0, Status='live', Symbol='RNU6-249P', Aliases='', description='RNA, U6 small nuclear 249, pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479678, CurrentID=0, Status='live', Symbol='RNU6-250P', Aliases='', description='RNA, U6 small nuclear 250, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479679, CurrentID=0, Status='live', Symbol='RNU6-254P', Aliases='', description='RNA, U6 small nuclear 254, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479680, CurrentID=0, Status='live', Symbol='RNU6-256P', Aliases='', description='RNA, U6 small nuclear 256, pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479681, CurrentID=0, Status='live', Symbol='RNU6-257P', Aliases='', description='RNA, U6 small nuclear 257, pseudogene', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479682, CurrentID=0, Status='live', Symbol='RNU6-260P', Aliases='', description='RNA, U6 small nuclear 260, pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479683, CurrentID=0, Status='live', Symbol='RNU6-263P', Aliases='', description='RNA, U6 small nuclear 263, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479684, CurrentID=0, Status='live', Symbol='RNU6-266P', Aliases='', description='RNA, U6 small nuclear 266, pseudogene', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479685, CurrentID=0, Status='live', Symbol='RNU6-268P', Aliases='', description='RNA, U6 small nuclear 268, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479686, CurrentID=0, Status='live', Symbol='RNU6-270P', Aliases='', description='RNA, U6 small nuclear 270, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479687, CurrentID=0, Status='live', Symbol='RNU6-272P', Aliases='', description='RNA, U6 small nuclear 272, pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479688, CurrentID=0, Status='live', Symbol='RNU6-274P', Aliases='', description='RNA, U6 small nuclear 274, pseudogene', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479689, CurrentID=0, Status='live', Symbol='RNU6-276P', Aliases='', description='RNA, U6 small nuclear 276, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479690, CurrentID=0, Status='live', Symbol='RNU6-279P', Aliases='', description='RNA, U6 small nuclear 279, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479691, CurrentID=0, Status='live', Symbol='RNU6-282P', Aliases='', description='RNA, U6 small nuclear 282, pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442444, CurrentID=0, Status='live', Symbol='FAM47C', Aliases='', description='family with sequence similarity 47 member C', other_designations='family with sequence similarity 47, member C', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37008359, end_position_on_the_genomic_accession=37011666, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479693, CurrentID=0, Status='live', Symbol='RNU6-285P', Aliases='', description='RNA, U6 small nuclear 285, pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479694, CurrentID=0, Status='live', Symbol='RNU6-288P', Aliases='', description='RNA, U6 small nuclear 288, pseudogene', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479695, CurrentID=0, Status='live', Symbol='RNU6-289P', Aliases='', description='RNA, U6 small nuclear 289, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479696, CurrentID=0, Status='live', Symbol='RNU6-291P', Aliases='', description='RNA, U6 small nuclear 291, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479697, CurrentID=0, Status='live', Symbol='RNU6-294P', Aliases='', description='RNA, U6 small nuclear 294, pseudogene', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479698, CurrentID=0, Status='live', Symbol='RNU6-296P', Aliases='', description='RNA, U6 small nuclear 296, pseudogene', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479699, CurrentID=0, Status='live', Symbol='RNU6-297P', Aliases='', description='RNA, U6 small nuclear 297, pseudogene', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479700, CurrentID=0, Status='live', Symbol='RNU6-300P', Aliases='', description='RNA, U6 small nuclear 300, pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479701, CurrentID=0, Status='live', Symbol='RNU6-302P', Aliases='', description='RNA, U6 small nuclear 302, pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442454, CurrentID=0, Status='live', Symbol='UQCRBP1', Aliases='', description='ubiquinol-cytochrome c reductase binding protein pseudogene 1', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56736788, end_position_on_the_genomic_accession=56737584, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479703, CurrentID=0, Status='live', Symbol='RNU6-306P', Aliases='', description='RNA, U6 small nuclear 306, pseudogene', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479704, CurrentID=0, Status='live', Symbol='RNU6-309P', Aliases='', description='RNA, U6 small nuclear 309, pseudogene', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479705, CurrentID=0, Status='live', Symbol='RNU6-311P', Aliases='', description='RNA, U6 small nuclear 311, pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479706, CurrentID=0, Status='live', Symbol='RNU6-312P', Aliases='', description='RNA, U6 small nuclear 312, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442459, CurrentID=0, Status='live', Symbol='XRCC6P5', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 6 pseudogene 5', other_designations='DNA repair protein|X-ray repair complementing defective repair pseudogene', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=99461602, end_position_on_the_genomic_accession=99939843, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106479708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479708, CurrentID=0, Status='live', Symbol='RNU6-317P', Aliases='', description='RNA, U6 small nuclear 317, pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481443, CurrentID=0, Status='live', Symbol='RNU6-831P', Aliases='', description='RNA, U6 small nuclear 831, pseudogene', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479710, CurrentID=0, Status='live', Symbol='RNU6-324P', Aliases='', description='RNA, U6 small nuclear 324, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479711, CurrentID=0, Status='live', Symbol='RNU6-328P', Aliases='', description='RNA, U6 small nuclear 328, pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479712, CurrentID=0, Status='live', Symbol='RNU6-329P', Aliases='', description='RNA, U6 small nuclear 329, pseudogene', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479713, CurrentID=0, Status='live', Symbol='RNU6-331P', Aliases='', description='RNA, U6 small nuclear 331, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479714, CurrentID=0, Status='live', Symbol='RNU6-334P', Aliases='', description='RNA, U6 small nuclear 334, pseudogene', other_designations='', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479715, CurrentID=0, Status='live', Symbol='RNU6-335P', Aliases='', description='RNA, U6 small nuclear 335, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479716, CurrentID=0, Status='live', Symbol='RNU6-337P', Aliases='', description='RNA, U6 small nuclear 337, pseudogene', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479717, CurrentID=0, Status='live', Symbol='RNU6-340P', Aliases='', description='RNA, U6 small nuclear 340, pseudogene', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479718, CurrentID=0, Status='live', Symbol='RNU6-342P', Aliases='', description='RNA, U6 small nuclear 342, pseudogene', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479719, CurrentID=0, Status='live', Symbol='RNU6-343P', Aliases='', description='RNA, U6 small nuclear 343, pseudogene', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479720, CurrentID=0, Status='live', Symbol='RNU6-347P', Aliases='', description='RNA, U6 small nuclear 347, pseudogene', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479721, CurrentID=0, Status='live', Symbol='RNU6-349P', Aliases='', description='RNA, U6 small nuclear 349, pseudogene', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479722, CurrentID=0, Status='live', Symbol='RNU6-351P', Aliases='', description='RNA, U6 small nuclear 351, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479723, CurrentID=0, Status='live', Symbol='RNU6-353P', Aliases='', description='RNA, U6 small nuclear 353, pseudogene', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479724, CurrentID=0, Status='live', Symbol='RNU6-356P', Aliases='', description='RNA, U6 small nuclear 356, pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479725, CurrentID=0, Status='live', Symbol='RNU6-359P', Aliases='', description='RNA, U6 small nuclear 359, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479726, CurrentID=0, Status='live', Symbol='RNU6-360P', Aliases='', description='RNA, U6 small nuclear 360, pseudogene', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442479, CurrentID=0, Status='live', Symbol='HSFY3P', Aliases='HSFYP2', description='heat shock transcription factor, Y-linked 3, pseudogene', other_designations='heat shock transcription factor, Y-linked 1 pseudogene|heat shock transcription factor, Y-linked-like pseudogene 2', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2881565, end_position_on_the_genomic_accession=2883739, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479728, CurrentID=0, Status='live', Symbol='RNU6-365P', Aliases='', description='RNA, U6 small nuclear 365, pseudogene', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479729, CurrentID=0, Status='live', Symbol='RNU6-368P', Aliases='', description='RNA, U6 small nuclear 368, pseudogene', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479730, CurrentID=0, Status='live', Symbol='RNU6-369P', Aliases='', description='RNA, U6 small nuclear 369, pseudogene', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479731, CurrentID=0, Status='live', Symbol='RNU6-371P', Aliases='', description='RNA, U6 small nuclear 371, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479732, CurrentID=0, Status='live', Symbol='RNU6-376P', Aliases='', description='RNA, U6 small nuclear 376, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442485, CurrentID=0, Status='live', Symbol='OFD1P16Y', Aliases='OFD1P16, OFD1PY16, OFDYP16', description='OFD1 pseudogene 16, Y-linked', other_designations='OFD1 pseudogene, Y-linked 16', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21811387, end_position_on_the_genomic_accession=21818212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188278, CurrentID=0, Status='live', Symbol='FECD2', Aliases='FCD1', description='corneal dystrophy, Fuchs endothelial 2', other_designations='', map_location='13pter-q12.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610158, no_hdr0=''),
  106479735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479735, CurrentID=0, Status='live', Symbol='RNU6-382P', Aliases='', description='RNA, U6 small nuclear 382, pseudogene', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479736, CurrentID=0, Status='live', Symbol='RNU6-383P', Aliases='', description='RNA, U6 small nuclear 383, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401428, CurrentID=0, Status='live', Symbol='OR2A20P', Aliases='OR2A20', description='olfactory receptor family 2 subfamily A member 20 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 20 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144250674, end_position_on_the_genomic_accession=144251603, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106479738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479738, CurrentID=0, Status='live', Symbol='RNU6-390P', Aliases='', description='RNA, U6 small nuclear 390, pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479739, CurrentID=0, Status='live', Symbol='RNU6-391P', Aliases='', description='RNA, U6 small nuclear 391, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479740, CurrentID=0, Status='live', Symbol='RNU6-395P', Aliases='', description='RNA, U6 small nuclear 395, pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479741, CurrentID=0, Status='live', Symbol='RNU6-397P', Aliases='', description='RNA, U6 small nuclear 397, pseudogene', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479742, CurrentID=0, Status='live', Symbol='RNU6-398P', Aliases='', description='RNA, U6 small nuclear 398, pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8213, CurrentID=0, Status='live', Symbol='BAS', Aliases='', description='Beta-adrenergic stimulation, response to', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109670, no_hdr0=''),
  106479744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479744, CurrentID=0, Status='live', Symbol='RNU6-405P', Aliases='', description='RNA, U6 small nuclear 405, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479745, CurrentID=0, Status='live', Symbol='RNU6-407P', Aliases='', description='RNA, U6 small nuclear 407, pseudogene', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479746, CurrentID=0, Status='live', Symbol='RNU6-408P', Aliases='', description='RNA, U6 small nuclear 408, pseudogene', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479747, CurrentID=0, Status='live', Symbol='RNU6-411P', Aliases='', description='RNA, U6 small nuclear 411, pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479748, CurrentID=0, Status='live', Symbol='RNU6-413P', Aliases='', description='RNA, U6 small nuclear 413, pseudogene', other_designations='', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479749, CurrentID=0, Status='live', Symbol='RNU6-414P', Aliases='', description='RNA, U6 small nuclear 414, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479750, CurrentID=0, Status='live', Symbol='RNU6-417P', Aliases='', description='RNA, U6 small nuclear 417, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479751, CurrentID=0, Status='live', Symbol='RNU6-419P', Aliases='', description='RNA, U6 small nuclear 419, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479752, CurrentID=0, Status='live', Symbol='RNU6-423P', Aliases='', description='RNA, U6 small nuclear 423, pseudogene', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479753, CurrentID=0, Status='live', Symbol='RNU6-424P', Aliases='', description='RNA, U6 small nuclear 424, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479754, CurrentID=0, Status='live', Symbol='RNU6-426P', Aliases='', description='RNA, U6 small nuclear 426, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401431, CurrentID=0, Status='live', Symbol='ATP6V0E2-AS1', Aliases='', description='ATP6V0E2 antisense RNA 1', other_designations='ATP6V0E2 antisense RNA 1 (non-protein coding)', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149867694, end_position_on_the_genomic_accession=149873862, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  285335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285335, CurrentID=0, Status='live', Symbol='SLC9C1', Aliases='NHE, SLC9A10, sperm-NHE', description='solute carrier family 9, subfamily C (Na+-transporting carboxylic acid decarboxylase), member 1', other_designations='NHE-10|Na(+)/H(+) exchanger 10|sNHE|solute carrier family 10 member 10|solute carrier family 9|solute carrier family 9 member C1|solute carrier family 9, member 10|sperm-specific Na(+)/H(+) exchanger|sperm-specific sodium proton exchanger', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112140898, end_position_on_the_genomic_accession=112362986, orientation='minus', exon_count=37, OMIM=612738, no_hdr0=''),
  106479757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479757, CurrentID=0, Status='live', Symbol='RNU6-432P', Aliases='', description='RNA, U6 small nuclear 432, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479758, CurrentID=0, Status='live', Symbol='RNU6-435P', Aliases='', description='RNA, U6 small nuclear 435, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479759, CurrentID=0, Status='live', Symbol='RNU6-437P', Aliases='', description='RNA, U6 small nuclear 437, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479760, CurrentID=0, Status='live', Symbol='RNU6-438P', Aliases='', description='RNA, U6 small nuclear 438, pseudogene', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479761, CurrentID=0, Status='live', Symbol='RNU6-442P', Aliases='', description='RNA, U6 small nuclear 442, pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479762, CurrentID=0, Status='live', Symbol='RNU6-444P', Aliases='', description='RNA, U6 small nuclear 444, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479763, CurrentID=0, Status='live', Symbol='RNU6-445P', Aliases='', description='RNA, U6 small nuclear 445, pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479764, CurrentID=0, Status='live', Symbol='RNU6-448P', Aliases='', description='RNA, U6 small nuclear 448, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479765, CurrentID=0, Status='live', Symbol='RNU6-451P', Aliases='', description='RNA, U6 small nuclear 451, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479766, CurrentID=0, Status='live', Symbol='RNU6-453P', Aliases='', description='RNA, U6 small nuclear 453, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479767, CurrentID=0, Status='live', Symbol='RNU6-455P', Aliases='', description='RNA, U6 small nuclear 455, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479768, CurrentID=0, Status='live', Symbol='RNU6-457P', Aliases='', description='RNA, U6 small nuclear 457, pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479769, CurrentID=0, Status='live', Symbol='RNU6-460P', Aliases='', description='RNA, U6 small nuclear 460, pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188314, CurrentID=0, Status='live', Symbol='AUTS12', Aliases='', description='autism, susceptibility to, 12', other_designations='', map_location='21p13-q11', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610838, no_hdr0=''),
  442523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442523, CurrentID=0, Status='live', Symbol='DPY19L2P4', Aliases='', description='DPY19L2 pseudogene 4', other_designations='dpy-19-like 2 pseudogene 4', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90119400, end_position_on_the_genomic_accession=90125600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442524, CurrentID=0, Status='live', Symbol='DPY19L2P3', Aliases='', description='DPY19L2 pseudogene 3', other_designations='dpy-19-like 2 pseudogene 3', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29685154, end_position_on_the_genomic_accession=29688243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100188317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188317, CurrentID=0, Status='live', Symbol='AUTS13', Aliases='', description='autism, susceptibility to, 13', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610908, no_hdr0=''),
  106479774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479774, CurrentID=0, Status='live', Symbol='RNU6-471P', Aliases='', description='RNA, U6 small nuclear 471, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479775, CurrentID=0, Status='live', Symbol='RNU6-473P', Aliases='', description='RNA, U6 small nuclear 473, pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479776, CurrentID=0, Status='live', Symbol='RNU6-475P', Aliases='', description='RNA, U6 small nuclear 475, pseudogene', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188321, CurrentID=0, Status='live', Symbol='HYT8', Aliases='', description='hypertension, essential, susceptibility to, 8', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611014, no_hdr0=''),
  106479778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479778, CurrentID=0, Status='live', Symbol='RNU6-478P', Aliases='', description='RNA, U6 small nuclear 478, pseudogene', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479779, CurrentID=0, Status='live', Symbol='RNU6-481P', Aliases='', description='RNA, U6 small nuclear 481, pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479780, CurrentID=0, Status='live', Symbol='RNU6-483P', Aliases='', description='RNA, U6 small nuclear 483, pseudogene', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479781, CurrentID=0, Status='live', Symbol='RNU6-484P', Aliases='', description='RNA, U6 small nuclear 484, pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479782, CurrentID=0, Status='live', Symbol='RNU6-488P', Aliases='', description='RNA, U6 small nuclear 488, pseudogene', other_designations='', map_location='3q11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479783, CurrentID=0, Status='live', Symbol='RNU6-490P', Aliases='', description='RNA, U6 small nuclear 490, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188328, CurrentID=0, Status='live', Symbol='PNKD2', Aliases='', description='paroxysmal nonkinesigenic dyskinesia 2', other_designations='', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611147, no_hdr0=''),
  106479785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479785, CurrentID=0, Status='live', Symbol='RNU6-494P', Aliases='', description='RNA, U6 small nuclear 494, pseudogene', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479786, CurrentID=0, Status='live', Symbol='RNU6-497P', Aliases='', description='RNA, U6 small nuclear 497, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479787, CurrentID=0, Status='live', Symbol='RNU6-499P', Aliases='', description='RNA, U6 small nuclear 499, pseudogene', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479788, CurrentID=0, Status='live', Symbol='RNU6-500P', Aliases='', description='RNA, U6 small nuclear 500, pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479789, CurrentID=0, Status='live', Symbol='RNU6-503P', Aliases='', description='RNA, U6 small nuclear 503, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479790, CurrentID=0, Status='live', Symbol='RNU6-505P', Aliases='', description='RNA, U6 small nuclear 505, pseudogene', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479791, CurrentID=0, Status='live', Symbol='RNU6-506P', Aliases='', description='RNA, U6 small nuclear 506, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479792, CurrentID=0, Status='live', Symbol='RNU6-509P', Aliases='', description='RNA, U6 small nuclear 509, pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479793, CurrentID=0, Status='live', Symbol='RNU6-511P', Aliases='', description='RNA, U6 small nuclear 511, pseudogene', other_designations='', map_location='3q11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479794, CurrentID=0, Status='live', Symbol='RNU6-516P', Aliases='', description='RNA, U6 small nuclear 516, pseudogene', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479795, CurrentID=0, Status='live', Symbol='RNU6-517P', Aliases='', description='RNA, U6 small nuclear 517, pseudogene', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188340, CurrentID=0, Status='live', Symbol='AOS', Aliases='', description='Adams-Oliver syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=100300, no_hdr0=''),
  106479797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479797, CurrentID=0, Status='live', Symbol='RNU6-522P', Aliases='', description='RNA, U6 small nuclear 522, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479798, CurrentID=0, Status='live', Symbol='RNU6-523P', Aliases='', description='RNA, U6 small nuclear 523, pseudogene', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479799, CurrentID=0, Status='live', Symbol='RNU6-525P', Aliases='', description='RNA, U6 small nuclear 525, pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479800, CurrentID=0, Status='live', Symbol='RNU6-528P', Aliases='', description='RNA, U6 small nuclear 528, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479801, CurrentID=0, Status='live', Symbol='RNU6-530P', Aliases='', description='RNA, U6 small nuclear 530, pseudogene', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479802, CurrentID=0, Status='live', Symbol='RNU6-531P', Aliases='', description='RNA, U6 small nuclear 531, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479803, CurrentID=0, Status='live', Symbol='RNU6-535P', Aliases='', description='RNA, U6 small nuclear 535, pseudogene', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479804, CurrentID=0, Status='live', Symbol='RNU6-537P', Aliases='', description='RNA, U6 small nuclear 537, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479805, CurrentID=0, Status='live', Symbol='RNU6-538P', Aliases='', description='RNA, U6 small nuclear 538, pseudogene', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479806, CurrentID=0, Status='live', Symbol='RNU6-540P', Aliases='', description='RNA, U6 small nuclear 540, pseudogene', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479807, CurrentID=0, Status='live', Symbol='RNU6-543P', Aliases='', description='RNA, U6 small nuclear 543, pseudogene', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479808, CurrentID=0, Status='live', Symbol='RNU6-545P', Aliases='RNU6-292P', description='RNA, U6 small nuclear 545, pseudogene', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479809, CurrentID=0, Status='live', Symbol='RNU6-546P', Aliases='', description='RNA, U6 small nuclear 546, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479810, CurrentID=0, Status='live', Symbol='RNU6-549P', Aliases='', description='RNA, U6 small nuclear 549, pseudogene', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479811, CurrentID=0, Status='live', Symbol='RNU6-551P', Aliases='', description='RNA, U6 small nuclear 551, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479812, CurrentID=0, Status='live', Symbol='RNU6-554P', Aliases='', description='RNA, U6 small nuclear 554, pseudogene', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479813, CurrentID=0, Status='live', Symbol='RNU6-555P', Aliases='', description='RNA, U6 small nuclear 555, pseudogene', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479814, CurrentID=0, Status='live', Symbol='RNU6-557P', Aliases='', description='RNA, U6 small nuclear 557, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479815, CurrentID=0, Status='live', Symbol='RNU6-561P', Aliases='', description='RNA, U6 small nuclear 561, pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479816, CurrentID=0, Status='live', Symbol='RNU6-562P', Aliases='', description='RNA, U6 small nuclear 562, pseudogene', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479817, CurrentID=0, Status='live', Symbol='RNU6-564P', Aliases='', description='RNA, U6 small nuclear 564, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479818, CurrentID=0, Status='live', Symbol='RNU6-567P', Aliases='', description='RNA, U6 small nuclear 567, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479819, CurrentID=0, Status='live', Symbol='RNU6-571P', Aliases='', description='RNA, U6 small nuclear 571, pseudogene', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442572, CurrentID=0, Status='live', Symbol='SPDYE21P', Aliases='', description='speedy/RINGO cell cycle regulator family member E21, pseudogene', other_designations='Putative WBSCR19-like protein 8|speedy homolog E1 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67278293, end_position_on_the_genomic_accession=67288993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479821, CurrentID=0, Status='live', Symbol='RNU6-576P', Aliases='', description='RNA, U6 small nuclear 576, pseudogene', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479822, CurrentID=0, Status='live', Symbol='RNU6-577P', Aliases='', description='RNA, U6 small nuclear 577, pseudogene', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479823, CurrentID=0, Status='live', Symbol='RNU6-581P', Aliases='', description='RNA, U6 small nuclear 581, pseudogene', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479824, CurrentID=0, Status='live', Symbol='RNU6-583P', Aliases='', description='RNA, U6 small nuclear 583, pseudogene', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479825, CurrentID=0, Status='live', Symbol='RNU6-584P', Aliases='', description='RNA, U6 small nuclear 584, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442578, CurrentID=0, Status='live', Symbol='STAG3L3', Aliases='STAG3L1, STAG3L2P, STAG3L3', description='stromal antigen 3-like 3 (pseudogene)', other_designations='STAG3-like protein 3|stromal antigen 3 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72997295, end_position_on_the_genomic_accession=73005943, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106479827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479827, CurrentID=0, Status='live', Symbol='RNU6-591P', Aliases='', description='RNA, U6 small nuclear 591, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479828, CurrentID=0, Status='live', Symbol='RNU6-593P', Aliases='', description='RNA, U6 small nuclear 593, pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479829, CurrentID=0, Status='live', Symbol='RNU6-595P', Aliases='', description='RNA, U6 small nuclear 595, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442582, CurrentID=0, Status='live', Symbol='STAG3L2', Aliases='STAG3L1P, STAG3L3, STAG3L2', description='stromal antigen 3-like 2 (pseudogene)', other_designations='STAG3-like protein 2', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74881990, end_position_on_the_genomic_accession=74890612, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106479831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479831, CurrentID=0, Status='live', Symbol='RNU6-600P', Aliases='', description='RNA, U6 small nuclear 600, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479832, CurrentID=0, Status='live', Symbol='RNU6-601P', Aliases='', description='RNA, U6 small nuclear 601, pseudogene', other_designations='', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479833, CurrentID=0, Status='live', Symbol='RNU6-604P', Aliases='', description='RNA, U6 small nuclear 604, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479834, CurrentID=0, Status='live', Symbol='RNU6-606P', Aliases='', description='RNA, U6 small nuclear 606, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479835, CurrentID=0, Status='live', Symbol='RNU6-610P', Aliases='', description='RNA, U6 small nuclear 610, pseudogene', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479836, CurrentID=0, Status='live', Symbol='RNU6-611P', Aliases='', description='RNA, U6 small nuclear 611, pseudogene', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479837, CurrentID=0, Status='live', Symbol='RNU6-613P', Aliases='', description='RNA, U6 small nuclear 613, pseudogene', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442590, CurrentID=0, Status='live', Symbol='SPDYE5', Aliases='', description='speedy/RINGO cell cycle regulator family member E5', other_designations='speedy homolog E5', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75493182, end_position_on_the_genomic_accession=75504304, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106479839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479839, CurrentID=0, Status='live', Symbol='RNU6-617P', Aliases='', description='RNA, U6 small nuclear 617, pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479840, CurrentID=0, Status='live', Symbol='RNU6-619P', Aliases='', description='RNA, U6 small nuclear 619, pseudogene', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479841, CurrentID=0, Status='live', Symbol='RNU6-622P', Aliases='', description='RNA, U6 small nuclear 622, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479842, CurrentID=0, Status='live', Symbol='RNU6-624P', Aliases='', description='RNA, U6 small nuclear 624, pseudogene', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479843, CurrentID=0, Status='live', Symbol='RNU6-625P', Aliases='', description='RNA, U6 small nuclear 625, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479844, CurrentID=0, Status='live', Symbol='RNU6-629P', Aliases='', description='RNA, U6 small nuclear 629, pseudogene', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479845, CurrentID=0, Status='live', Symbol='RNU6-631P', Aliases='', description='RNA, U6 small nuclear 631, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479846, CurrentID=0, Status='live', Symbol='RNU6-632P', Aliases='', description='RNA, U6 small nuclear 632, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479847, CurrentID=0, Status='live', Symbol='RNU6-634P', Aliases='', description='RNA, U6 small nuclear 634, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479848, CurrentID=0, Status='live', Symbol='RNU6-637P', Aliases='', description='RNA, U6 small nuclear 637, pseudogene', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479849, CurrentID=0, Status='live', Symbol='RNU6-639P', Aliases='', description='RNA, U6 small nuclear 639, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479850, CurrentID=0, Status='live', Symbol='RNU6-640P', Aliases='', description='RNA, U6 small nuclear 640, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8231, CurrentID=0, Status='live', Symbol='CND', Aliases='', description='Corneal dermoids', other_designations='', map_location='Xq24-qter', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=304730, no_hdr0=''),
  106479852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479852, CurrentID=0, Status='live', Symbol='RNU6-645P', Aliases='', description='RNA, U6 small nuclear 645, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188397, CurrentID=0, Status='live', Symbol='HLP', Aliases='', description='hyperkeratosis lenticularis perstans', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=144150, no_hdr0=''),
  106479854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479854, CurrentID=0, Status='live', Symbol='RNU6-649P', Aliases='', description='RNA, U6 small nuclear 649, pseudogene', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479855, CurrentID=0, Status='live', Symbol='RNU6-651P', Aliases='', description='RNA, U6 small nuclear 651, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479856, CurrentID=0, Status='live', Symbol='RNU6-655P', Aliases='', description='RNA, U6 small nuclear 655, pseudogene', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479857, CurrentID=0, Status='live', Symbol='RNU6-656P', Aliases='', description='RNA, U6 small nuclear 656, pseudogene', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479858, CurrentID=0, Status='live', Symbol='RNU6-658P', Aliases='', description='RNA, U6 small nuclear 658, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479859, CurrentID=0, Status='live', Symbol='RNU6-661P', Aliases='', description='RNA, U6 small nuclear 661, pseudogene', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479860, CurrentID=0, Status='live', Symbol='RNU6-662P', Aliases='', description='RNA, U6 small nuclear 662, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479861, CurrentID=0, Status='live', Symbol='RNU6-664P', Aliases='FZD5', description='RNA, U6 small nuclear 664, pseudogene', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479862, CurrentID=0, Status='live', Symbol='RNU6-667P', Aliases='', description='RNA, U6 small nuclear 667, pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479863, CurrentID=0, Status='live', Symbol='RNU6-670P', Aliases='', description='RNA, U6 small nuclear 670, pseudogene', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479864, CurrentID=0, Status='live', Symbol='RNU6-675P', Aliases='', description='RNA, U6 small nuclear 675, pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479865, CurrentID=0, Status='live', Symbol='RNU6-678P', Aliases='', description='RNA, U6 small nuclear 678, pseudogene', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479866, CurrentID=0, Status='live', Symbol='RNU6-679P', Aliases='', description='RNA, U6 small nuclear 679, pseudogene', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479867, CurrentID=0, Status='live', Symbol='RNU6-682P', Aliases='', description='RNA, U6 small nuclear 682, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479868, CurrentID=0, Status='live', Symbol='RNU6-685P', Aliases='', description='RNA, U6 small nuclear 685, pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479869, CurrentID=0, Status='live', Symbol='RNU6-687P', Aliases='', description='RNA, U6 small nuclear 687, pseudogene', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479870, CurrentID=0, Status='live', Symbol='RNU6-689P', Aliases='', description='RNA, U6 small nuclear 689, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479871, CurrentID=0, Status='live', Symbol='RNU6-692P', Aliases='', description='RNA, U6 small nuclear 692, pseudogene', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479872, CurrentID=0, Status='live', Symbol='RNU6-694P', Aliases='', description='RNA, U6 small nuclear 694, pseudogene', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479873, CurrentID=0, Status='live', Symbol='RNU6-696P', Aliases='', description='RNA, U6 small nuclear 696, pseudogene', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479874, CurrentID=0, Status='live', Symbol='RNU6-697P', Aliases='', description='RNA, U6 small nuclear 697, pseudogene', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479875, CurrentID=0, Status='live', Symbol='RNU6-699P', Aliases='', description='RNA, U6 small nuclear 699, pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479876, CurrentID=0, Status='live', Symbol='RNU6-703P', Aliases='', description='RNA, U6 small nuclear 703, pseudogene', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479877, CurrentID=0, Status='live', Symbol='RNU6-704P', Aliases='', description='RNA, U6 small nuclear 704, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479878, CurrentID=0, Status='live', Symbol='RNU6-706P', Aliases='', description='RNA, U6 small nuclear 706, pseudogene', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479879, CurrentID=0, Status='live', Symbol='RNU6-709P', Aliases='', description='RNA, U6 small nuclear 709, pseudogene', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479880, CurrentID=0, Status='live', Symbol='RNU6-710P', Aliases='', description='RNA, U6 small nuclear 710, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479881, CurrentID=0, Status='live', Symbol='RNU6-711P', Aliases='', description='RNA, U6 small nuclear 711, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479882, CurrentID=0, Status='live', Symbol='RNU6-713P', Aliases='', description='RNA, U6 small nuclear 713, pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479883, CurrentID=0, Status='live', Symbol='RNU6-716P', Aliases='', description='RNA, U6 small nuclear 716, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479884, CurrentID=0, Status='live', Symbol='RNU6-717P', Aliases='', description='RNA, U6 small nuclear 717, pseudogene', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479885, CurrentID=0, Status='live', Symbol='RNU6-720P', Aliases='', description='RNA, U6 small nuclear 720, pseudogene', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479886, CurrentID=0, Status='live', Symbol='RNU6-723P', Aliases='', description='RNA, U6 small nuclear 723, pseudogene', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479887, CurrentID=0, Status='live', Symbol='RNU6-724P', Aliases='', description='RNA, U6 small nuclear 724, pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479888, CurrentID=0, Status='live', Symbol='RNU6-728P', Aliases='', description='RNA, U6 small nuclear 728, pseudogene', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479889, CurrentID=0, Status='live', Symbol='RNU6-730P', Aliases='', description='RNA, U6 small nuclear 730, pseudogene', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479890, CurrentID=0, Status='live', Symbol='RNU6-732P', Aliases='', description='RNA, U6 small nuclear 732, pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479891, CurrentID=0, Status='live', Symbol='RNU6-733P', Aliases='', description='RNA, U6 small nuclear 733, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479892, CurrentID=0, Status='live', Symbol='RNU6-738P', Aliases='', description='RNA, U6 small nuclear 738, pseudogene', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479893, CurrentID=0, Status='live', Symbol='RNU6-740P', Aliases='', description='RNA, U6 small nuclear 740, pseudogene', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479894, CurrentID=0, Status='live', Symbol='RNU6-741P', Aliases='', description='RNA, U6 small nuclear 741, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479895, CurrentID=0, Status='live', Symbol='RNU6-744P', Aliases='', description='RNA, U6 small nuclear 744, pseudogene', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479896, CurrentID=0, Status='live', Symbol='RNU6-746P', Aliases='', description='RNA, U6 small nuclear 746, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479897, CurrentID=0, Status='live', Symbol='RNU6-750P', Aliases='', description='RNA, U6 small nuclear 750, pseudogene', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479898, CurrentID=0, Status='live', Symbol='RNU6-751P', Aliases='', description='RNA, U6 small nuclear 751, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479899, CurrentID=0, Status='live', Symbol='RNU6-753P', Aliases='', description='RNA, U6 small nuclear 753, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479900, CurrentID=0, Status='live', Symbol='RNU6-756P', Aliases='', description='RNA, U6 small nuclear 756, pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479901, CurrentID=0, Status='live', Symbol='RNU6-757P', Aliases='', description='RNA, U6 small nuclear 757, pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479902, CurrentID=0, Status='live', Symbol='RNU6-759P', Aliases='', description='RNA, U6 small nuclear 759, pseudogene', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479903, CurrentID=0, Status='live', Symbol='RNU6-762P', Aliases='', description='RNA, U6 small nuclear 762, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479904, CurrentID=0, Status='live', Symbol='RNU6-764P', Aliases='', description='RNA, U6 small nuclear 764, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479905, CurrentID=0, Status='live', Symbol='RNU6-765P', Aliases='', description='RNA, U6 small nuclear 765, pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479906, CurrentID=0, Status='live', Symbol='RNU6-769P', Aliases='', description='RNA, U6 small nuclear 769, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442659, CurrentID=0, Status='live', Symbol='RPL23P7', Aliases='RPL23_5_797', description='ribosomal protein L23 pseudogene 7', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26276161, end_position_on_the_genomic_accession=26276666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442660, CurrentID=0, Status='live', Symbol='RPL7AP38', Aliases='RPL7A_11_798', description='ribosomal protein L7a pseudogene 38', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26922108, end_position_on_the_genomic_accession=26922966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442661, CurrentID=0, Status='live', Symbol='HMGB3P20', Aliases='', description='high mobility group box 3 pseudogene 20', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 20', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26982604, end_position_on_the_genomic_accession=26983459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479910, CurrentID=0, Status='live', Symbol='RNU6-778P', Aliases='', description='RNA, U6 small nuclear 778, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479911, CurrentID=0, Status='live', Symbol='RNU6-781P', Aliases='', description='RNA, U6 small nuclear 781, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479912, CurrentID=0, Status='live', Symbol='RNU6-783P', Aliases='', description='RNA, U6 small nuclear 783, pseudogene', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479913, CurrentID=0, Status='live', Symbol='RNU6-785P', Aliases='', description='RNA, U6 small nuclear 785, pseudogene', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479914, CurrentID=0, Status='live', Symbol='RNU6-787P', Aliases='', description='RNA, U6 small nuclear 787, pseudogene', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479915, CurrentID=0, Status='live', Symbol='RNU6-789P', Aliases='', description='RNA, U6 small nuclear 789, pseudogene', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479916, CurrentID=0, Status='live', Symbol='RNU6-791P', Aliases='', description='RNA, U6 small nuclear 791, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479917, CurrentID=0, Status='live', Symbol='RNU6-793P', Aliases='', description='RNA, U6 small nuclear 793, pseudogene', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479918, CurrentID=0, Status='live', Symbol='RNU6-797P', Aliases='', description='RNA, U6 small nuclear 797, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479919, CurrentID=0, Status='live', Symbol='RNU6-798P', Aliases='', description='RNA, U6 small nuclear 798, pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479920, CurrentID=0, Status='live', Symbol='RNU6-801P', Aliases='', description='RNA, U6 small nuclear 801, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442673, CurrentID=0, Status='live', Symbol='TUBG1P', Aliases='', description='tubulin gamma 1 pseudogene', other_designations='tubulin, gamma 1 pseudogene', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43918637, end_position_on_the_genomic_accession=43920252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479922, CurrentID=0, Status='live', Symbol='RNU6-805P', Aliases='', description='RNA, U6 small nuclear 805, pseudogene', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479923, CurrentID=0, Status='live', Symbol='RNU6-807P', Aliases='', description='RNA, U6 small nuclear 807, pseudogene', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479924, CurrentID=0, Status='live', Symbol='RNU6-809P', Aliases='', description='RNA, U6 small nuclear 809, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479925, CurrentID=0, Status='live', Symbol='RNU6-811P', Aliases='', description='RNA, U6 small nuclear 811, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479926, CurrentID=0, Status='live', Symbol='RNU6-812P', Aliases='', description='RNA, U6 small nuclear 812, pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479927, CurrentID=0, Status='live', Symbol='RNU6-817P', Aliases='', description='RNA, U6 small nuclear 817, pseudogene', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479928, CurrentID=0, Status='live', Symbol='RNU6-819P', Aliases='', description='RNA, U6 small nuclear 819, pseudogene', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479929, CurrentID=0, Status='live', Symbol='RNU6-820P', Aliases='', description='RNA, U6 small nuclear 820, pseudogene', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479930, CurrentID=0, Status='live', Symbol='RNU6-822P', Aliases='', description='RNA, U6 small nuclear 822, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479931, CurrentID=0, Status='live', Symbol='RNU6-826P', Aliases='', description='RNA, U6 small nuclear 826, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479932, CurrentID=0, Status='live', Symbol='RNU6-828P', Aliases='', description='RNA, U6 small nuclear 828, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479933, CurrentID=0, Status='live', Symbol='RNU6-829P', Aliases='', description='RNA, U6 small nuclear 829, pseudogene', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479934, CurrentID=0, Status='live', Symbol='RNU6-832P', Aliases='', description='RNA, U6 small nuclear 832, pseudogene', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479935, CurrentID=0, Status='live', Symbol='RNU6-834P', Aliases='', description='RNA, U6 small nuclear 834, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479936, CurrentID=0, Status='live', Symbol='RNU6-837P', Aliases='', description='RNA, U6 small nuclear 837, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479937, CurrentID=0, Status='live', Symbol='RNU6-838P', Aliases='', description='RNA, U6 small nuclear 838, pseudogene', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479938, CurrentID=0, Status='live', Symbol='RNU6-840P', Aliases='COPS7B, PTMAP7', description='RNA, U6 small nuclear 840, pseudogene', other_designations='tmp_locus_12|tmp_locus_14', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479939, CurrentID=0, Status='live', Symbol='RNU6-844P', Aliases='', description='RNA, U6 small nuclear 844, pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479940, CurrentID=0, Status='live', Symbol='RNU6-845P', Aliases='', description='RNA, U6 small nuclear 845, pseudogene', other_designations='', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479941, CurrentID=0, Status='live', Symbol='RNU6-848P', Aliases='', description='RNA, U6 small nuclear 848, pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479942, CurrentID=0, Status='live', Symbol='RNU6-851P', Aliases='', description='RNA, U6 small nuclear 851, pseudogene', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479943, CurrentID=0, Status='live', Symbol='RNU6-853P', Aliases='', description='RNA, U6 small nuclear 853, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479944, CurrentID=0, Status='live', Symbol='RNU6-855P', Aliases='', description='RNA, U6 small nuclear 855, pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479945, CurrentID=0, Status='live', Symbol='RNU6-858P', Aliases='', description='RNA, U6 small nuclear 858, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479946, CurrentID=0, Status='live', Symbol='RNU6-860P', Aliases='', description='RNA, U6 small nuclear 860, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479947, CurrentID=0, Status='live', Symbol='RNU6-861P', Aliases='', description='RNA, U6 small nuclear 861, pseudogene', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479948, CurrentID=0, Status='live', Symbol='RNU6-865P', Aliases='', description='RNA, U6 small nuclear 865, pseudogene', other_designations='', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479949, CurrentID=0, Status='live', Symbol='RNU6-867P', Aliases='', description='RNA, U6 small nuclear 867, pseudogene', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479950, CurrentID=0, Status='live', Symbol='RNU6-871P', Aliases='', description='RNA, U6 small nuclear 871, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479951, CurrentID=0, Status='live', Symbol='RNU6-873P', Aliases='', description='RNA, U6 small nuclear 873, pseudogene', other_designations='', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479952, CurrentID=0, Status='live', Symbol='RNU6-874P', Aliases='', description='RNA, U6 small nuclear 874, pseudogene', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479953, CurrentID=0, Status='live', Symbol='RNU6-875P', Aliases='', description='RNA, U6 small nuclear 875, pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479954, CurrentID=0, Status='live', Symbol='RNU6-878P', Aliases='', description='RNA, U6 small nuclear 878, pseudogene', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479955, CurrentID=0, Status='live', Symbol='RNU6-880P', Aliases='', description='RNA, U6 small nuclear 880, pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479956, CurrentID=0, Status='live', Symbol='RNU6-881P', Aliases='', description='RNA, U6 small nuclear 881, pseudogene', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479957, CurrentID=0, Status='live', Symbol='RNU6-884P', Aliases='', description='RNA, U6 small nuclear 884, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479958, CurrentID=0, Status='live', Symbol='RNU6-886P', Aliases='', description='RNA, U6 small nuclear 886, pseudogene', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479959, CurrentID=0, Status='live', Symbol='RNU6-888P', Aliases='', description='RNA, U6 small nuclear 888, pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442712, CurrentID=0, Status='live', Symbol='RPL7AP40', Aliases='RPL7A_18_822', description='ribosomal protein L7a pseudogene 40', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97200647, end_position_on_the_genomic_accession=97201470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1648, CurrentID=0, Status='live', Symbol='GADD45AP1', Aliases='DDIT1L, GADD45AL', description='growth arrest and DNA damage inducible alpha pseudogene 1', other_designations='growth arrest and DNA-damage-inducible, alpha pseudogene 1', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87820821, end_position_on_the_genomic_accession=87821569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479962, CurrentID=0, Status='live', Symbol='RNU6-895P', Aliases='', description='RNA, U6 small nuclear 895, pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479963, CurrentID=0, Status='live', Symbol='RNU6-897P', Aliases='', description='RNA, U6 small nuclear 897, pseudogene', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479964, CurrentID=0, Status='live', Symbol='RNU6-899P', Aliases='', description='RNA, U6 small nuclear 899, pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401466, CurrentID=0, Status='live', Symbol='C8orf59', Aliases='', description='chromosome 8 open reading frame 59', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85214053, end_position_on_the_genomic_accession=85220422, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106479966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479966, CurrentID=0, Status='live', Symbol='RNU6-904P', Aliases='', description='RNA, U6 small nuclear 904, pseudogene', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479967, CurrentID=0, Status='live', Symbol='RNU6-905P', Aliases='', description='RNA, U6 small nuclear 905, pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479968, CurrentID=0, Status='live', Symbol='RNU6-908P', Aliases='', description='RNA, U6 small nuclear 908, pseudogene', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442721, CurrentID=0, Status='live', Symbol='LMOD2', Aliases='C-LMOD, CLMOD', description='leiomodin 2', other_designations='cardiac leiomodin|leiomodin 2 (cardiac)', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123655807, end_position_on_the_genomic_accession=123664093, orientation='plus', exon_count=3, OMIM=608006, no_hdr0=''),
  106479970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479970, CurrentID=0, Status='live', Symbol='RNU6-912P', Aliases='', description='RNA, U6 small nuclear 912, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442723, CurrentID=0, Status='live', Symbol='RPL27P11', Aliases='RPL27_6_845', description='ribosomal protein L27 pseudogene 11', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131242555, end_position_on_the_genomic_accession=131243044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479972, CurrentID=0, Status='live', Symbol='RNU6-919P', Aliases='', description='RNA, U6 small nuclear 919, pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479973, CurrentID=0, Status='live', Symbol='RNU6-921P', Aliases='', description='RNA, U6 small nuclear 921, pseudogene', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442726, CurrentID=0, Status='live', Symbol='RPS17P12', Aliases='RPS17_5_854', description='ribosomal protein S17 pseudogene 12', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138298061, end_position_on_the_genomic_accession=138298507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479975, CurrentID=0, Status='live', Symbol='RNU6-925P', Aliases='', description='RNA, U6 small nuclear 925, pseudogene', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479976, CurrentID=0, Status='live', Symbol='RNU6-927P', Aliases='', description='RNA, U6 small nuclear 927, pseudogene', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479977, CurrentID=0, Status='live', Symbol='RNU6-928P', Aliases='', description='RNA, U6 small nuclear 928, pseudogene', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479978, CurrentID=0, Status='live', Symbol='RNU6-931P', Aliases='', description='RNA, U6 small nuclear 931, pseudogene', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479979, CurrentID=0, Status='live', Symbol='RNU6-933P', Aliases='', description='RNA, U6 small nuclear 933, pseudogene', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479980, CurrentID=0, Status='live', Symbol='RNU6-937P', Aliases='', description='RNA, U6 small nuclear 937, pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479981, CurrentID=0, Status='live', Symbol='RNU6-938P', Aliases='', description='RNA, U6 small nuclear 938, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479982, CurrentID=0, Status='live', Symbol='RNU6-940P', Aliases='', description='RNA, U6 small nuclear 940, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479983, CurrentID=0, Status='live', Symbol='RNU6-944P', Aliases='', description='RNA, U6 small nuclear 944, pseudogene', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479984, CurrentID=0, Status='live', Symbol='RNU6-945P', Aliases='', description='RNA, U6 small nuclear 945, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479985, CurrentID=0, Status='live', Symbol='RNU6-947P', Aliases='', description='RNA, U6 small nuclear 947, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479986, CurrentID=0, Status='live', Symbol='RNU6-950P', Aliases='', description='RNA, U6 small nuclear 950, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479987, CurrentID=0, Status='live', Symbol='RNU6-952P', Aliases='', description='RNA, U6 small nuclear 952, pseudogene', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479988, CurrentID=0, Status='live', Symbol='RNU6-953P', Aliases='', description='RNA, U6 small nuclear 953, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479989, CurrentID=0, Status='live', Symbol='RNU6-957P', Aliases='', description='RNA, U6 small nuclear 957, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479990, CurrentID=0, Status='live', Symbol='RNU6-959P', Aliases='', description='RNA, U6 small nuclear 959, pseudogene', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479991, CurrentID=0, Status='live', Symbol='RNU6-960P', Aliases='', description='RNA, U6 small nuclear 960, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479992, CurrentID=0, Status='live', Symbol='RNU6-964P', Aliases='', description='RNA, U6 small nuclear 964, pseudogene', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479993, CurrentID=0, Status='live', Symbol='RNU6-967P', Aliases='', description='RNA, U6 small nuclear 967, pseudogene', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479994, CurrentID=0, Status='live', Symbol='RNU6-969P', Aliases='', description='RNA, U6 small nuclear 969, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479995, CurrentID=0, Status='live', Symbol='RNU6-972P', Aliases='', description='RNA, U6 small nuclear 972, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479996, CurrentID=0, Status='live', Symbol='RNU6-973P', Aliases='', description='RNA, U6 small nuclear 973, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479997, CurrentID=0, Status='live', Symbol='RNU6-975P', Aliases='', description='RNA, U6 small nuclear 975, pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479998, CurrentID=0, Status='live', Symbol='RNU6-978P', Aliases='', description='RNA, U6 small nuclear 978, pseudogene', other_designations='', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479999, CurrentID=0, Status='live', Symbol='RNU6-979P', Aliases='', description='RNA, U6 small nuclear 979, pseudogene', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480000, CurrentID=0, Status='live', Symbol='RNU6-981P', Aliases='', description='RNA, U6 small nuclear 981, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480001, CurrentID=0, Status='live', Symbol='RNU6-985P', Aliases='', description='RNA, U6 small nuclear 985, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480002, CurrentID=0, Status='live', Symbol='RNU6-987P', Aliases='', description='RNA, U6 small nuclear 987, pseudogene', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480003, CurrentID=0, Status='live', Symbol='RNU6-988P', Aliases='', description='RNA, U6 small nuclear 988, pseudogene', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480004, CurrentID=0, Status='live', Symbol='RNU6-991P', Aliases='', description='RNA, U6 small nuclear 991, pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480005, CurrentID=0, Status='live', Symbol='RNU6-993P', Aliases='', description='RNA, U6 small nuclear 993, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480006, CurrentID=0, Status='live', Symbol='RNU6-994P', Aliases='', description='RNA, U6 small nuclear 994, pseudogene', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480007, CurrentID=0, Status='live', Symbol='RNU6-997P', Aliases='', description='RNA, U6 small nuclear 997, pseudogene', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480008, CurrentID=0, Status='live', Symbol='RNU6-999P', Aliases='', description='RNA, U6 small nuclear 999, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480009, CurrentID=0, Status='live', Symbol='RNU6-1004P', Aliases='', description='RNA, U6 small nuclear 1004, pseudogene', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480010, CurrentID=0, Status='live', Symbol='RNU6-1006P', Aliases='', description='RNA, U6 small nuclear 1006, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480011, CurrentID=0, Status='live', Symbol='RNU6-1011P', Aliases='', description='RNA, U6 small nuclear 1011, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480012, CurrentID=0, Status='live', Symbol='RNU6-1013P', Aliases='', description='RNA, U6 small nuclear 1013, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8258, CurrentID=0, Status='live', Symbol='GUST', Aliases='', description='Gustavson mental retardation syndrome (with microcephaly, optic', other_designations='Gustavson mental retardation syndrome (with microcephaly, optic atrophy, deafness)', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=309555, no_hdr0=''),
  106480014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480014, CurrentID=0, Status='live', Symbol='RNU6-1017P', Aliases='', description='RNA, U6 small nuclear 1017, pseudogene', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480015, CurrentID=0, Status='live', Symbol='RNU6-1019P', Aliases='', description='RNA, U6 small nuclear 1019, pseudogene', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480016, CurrentID=0, Status='live', Symbol='RNU6-1022P', Aliases='', description='RNA, U6 small nuclear 1022, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480017, CurrentID=0, Status='live', Symbol='RNU6-1023P', Aliases='', description='RNA, U6 small nuclear 1023, pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480018, CurrentID=0, Status='live', Symbol='RNU6-1025P', Aliases='', description='RNA, U6 small nuclear 1025, pseudogene', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401475, CurrentID=0, Status='live', Symbol='SRRM1P1', Aliases='SRRM1L, SRRM1P', description='serine/arginine repetitive matrix 1 pseudogene 1', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127086262, end_position_on_the_genomic_accession=127088530, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480020, CurrentID=0, Status='live', Symbol='RNU6-1031P', Aliases='', description='RNA, U6 small nuclear 1031, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480021, CurrentID=0, Status='live', Symbol='RNU6-1036P', Aliases='', description='RNA, U6 small nuclear 1036, pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480022, CurrentID=0, Status='live', Symbol='RNU6-1038P', Aliases='', description='RNA, U6 small nuclear 1038, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442775, CurrentID=0, Status='live', Symbol='RAC1P2', Aliases='Psi2Rac1', description='ras-related C3 botulinum toxin substrate 1 pseudogene 2', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=46723670, end_position_on_the_genomic_accession=46724607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480024, CurrentID=0, Status='live', Symbol='RNU6-1042P', Aliases='', description='RNA, U6 small nuclear 1042, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480025, CurrentID=0, Status='live', Symbol='RNU6-1045P', Aliases='', description='RNA, U6 small nuclear 1045, pseudogene', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480026, CurrentID=0, Status='live', Symbol='RNU6-1046P', Aliases='', description='RNA, U6 small nuclear 1046, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480027, CurrentID=0, Status='live', Symbol='RNU6-1048P', Aliases='', description='RNA, U6 small nuclear 1048, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480028, CurrentID=0, Status='live', Symbol='RNU6-1052P', Aliases='', description='RNA, U6 small nuclear 1052, pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480029, CurrentID=0, Status='live', Symbol='RNU6-1054P', Aliases='', description='RNA, U6 small nuclear 1054, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480030, CurrentID=0, Status='live', Symbol='RNU6-1057P', Aliases='', description='RNA, U6 small nuclear 1057, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480031, CurrentID=0, Status='live', Symbol='RNU6-1059P', Aliases='', description='RNA, U6 small nuclear 1059, pseudogene', other_designations='', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480032, CurrentID=0, Status='live', Symbol='RNU6-1060P', Aliases='', description='RNA, U6 small nuclear 1060, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480033, CurrentID=0, Status='live', Symbol='RNU6-1061P', Aliases='', description='RNA, U6 small nuclear 1061, pseudogene', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480034, CurrentID=0, Status='live', Symbol='RNU6-1065P', Aliases='', description='RNA, U6 small nuclear 1065, pseudogene', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480035, CurrentID=0, Status='live', Symbol='RNU6-1066P', Aliases='', description='RNA, U6 small nuclear 1066, pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480036, CurrentID=0, Status='live', Symbol='RNU6-1069P', Aliases='', description='RNA, U6 small nuclear 1069, pseudogene', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480037, CurrentID=0, Status='live', Symbol='RNU6-1072P', Aliases='', description='RNA, U6 small nuclear 1072, pseudogene', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480038, CurrentID=0, Status='live', Symbol='RNU6-1076P', Aliases='', description='RNA, U6 small nuclear 1076, pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480039, CurrentID=0, Status='live', Symbol='RNU6-1077P', Aliases='', description='RNA, U6 small nuclear 1077, pseudogene', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480040, CurrentID=0, Status='live', Symbol='RNU6-1079P', Aliases='', description='RNA, U6 small nuclear 1079, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480041, CurrentID=0, Status='live', Symbol='RNU6-1082P', Aliases='', description='RNA, U6 small nuclear 1082, pseudogene', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480042, CurrentID=0, Status='live', Symbol='RNU6-1084P', Aliases='', description='RNA, U6 small nuclear 1084, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480043, CurrentID=0, Status='live', Symbol='RNU6-1085P', Aliases='', description='RNA, U6 small nuclear 1085, pseudogene', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480044, CurrentID=0, Status='live', Symbol='RNU6-1088P', Aliases='', description='RNA, U6 small nuclear 1088, pseudogene', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480045, CurrentID=0, Status='live', Symbol='RNU6-1090P', Aliases='', description='RNA, U6 small nuclear 1090, pseudogene', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480046, CurrentID=0, Status='live', Symbol='RNU6-1092P', Aliases='', description='RNA, U6 small nuclear 1092, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480047, CurrentID=0, Status='live', Symbol='RNU6-1094P', Aliases='', description='RNA, U6 small nuclear 1094, pseudogene', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480048, CurrentID=0, Status='live', Symbol='RNU6-1097P', Aliases='', description='RNA, U6 small nuclear 1097, pseudogene', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480049, CurrentID=0, Status='live', Symbol='RNU6-1100P', Aliases='', description='RNA, U6 small nuclear 1100, pseudogene', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480050, CurrentID=0, Status='live', Symbol='RNU6-1101P', Aliases='', description='RNA, U6 small nuclear 1101, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480051, CurrentID=0, Status='live', Symbol='RNU6-1103P', Aliases='', description='RNA, U6 small nuclear 1103, pseudogene', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480052, CurrentID=0, Status='live', Symbol='RNU6-1105P', Aliases='', description='RNA, U6 small nuclear 1105, pseudogene', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480053, CurrentID=0, Status='live', Symbol='RNU6-1106P', Aliases='', description='RNA, U6 small nuclear 1106, pseudogene', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480054, CurrentID=0, Status='live', Symbol='RNU6-1109P', Aliases='', description='RNA, U6 small nuclear 1109, pseudogene', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480055, CurrentID=0, Status='live', Symbol='RNU6-1112P', Aliases='', description='RNA, U6 small nuclear 1112, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480056, CurrentID=0, Status='live', Symbol='RNU6-1113P', Aliases='', description='RNA, U6 small nuclear 1113, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480057, CurrentID=0, Status='live', Symbol='RNU6-1116P', Aliases='', description='RNA, U6 small nuclear 1116, pseudogene', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480058, CurrentID=0, Status='live', Symbol='RNU6-1118P', Aliases='', description='RNA, U6 small nuclear 1118, pseudogene', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480059, CurrentID=0, Status='live', Symbol='RNU6-1120P', Aliases='', description='RNA, U6 small nuclear 1120, pseudogene', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480060, CurrentID=0, Status='live', Symbol='RNU6-1123P', Aliases='', description='RNA, U6 small nuclear 1123, pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480061, CurrentID=0, Status='live', Symbol='RNU6-1125P', Aliases='', description='RNA, U6 small nuclear 1125, pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480062, CurrentID=0, Status='live', Symbol='RNU6-1128P', Aliases='', description='RNA, U6 small nuclear 1128, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480063, CurrentID=0, Status='live', Symbol='RNU6-1130P', Aliases='', description='RNA, U6 small nuclear 1130, pseudogene', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480064, CurrentID=0, Status='live', Symbol='RNU6-1131P', Aliases='', description='RNA, U6 small nuclear 1131, pseudogene', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132476, CurrentID=0, Status='live', Symbol='KRTAP4-7', Aliases='KAP4.7, KRTAP4.7', description='keratin associated protein 4-7', other_designations='Ultrahigh sulfur keratin-associated protein 4.7', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41084207, end_position_on_the_genomic_accession=41085144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480066, CurrentID=0, Status='live', Symbol='RNU6-1136P', Aliases='', description='RNA, U6 small nuclear 1136, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480067, CurrentID=0, Status='live', Symbol='RNU6-1137P', Aliases='', description='RNA, U6 small nuclear 1137, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480068, CurrentID=0, Status='live', Symbol='RNU6-1140P', Aliases='', description='RNA, U6 small nuclear 1140, pseudogene', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100534012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100534012, CurrentID=0, Status='live', Symbol='TNFAIP8L2-SCNM1', Aliases='', description='TNFAIP8L2-SCNM1 readthrough', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151156629, end_position_on_the_genomic_accession=151170297, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106480070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480070, CurrentID=0, Status='live', Symbol='RNU6-1146P', Aliases='', description='RNA, U6 small nuclear 1146, pseudogene', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480071, CurrentID=0, Status='live', Symbol='RNU6-1147P', Aliases='', description='RNA, U6 small nuclear 1147, pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480072, CurrentID=0, Status='live', Symbol='RNU6-1150P', Aliases='', description='RNA, U6 small nuclear 1150, pseudogene', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480073, CurrentID=0, Status='live', Symbol='RNU6-1152P', Aliases='', description='RNA, U6 small nuclear 1152, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480074, CurrentID=0, Status='live', Symbol='RNU6-1154P', Aliases='', description='RNA, U6 small nuclear 1154, pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480075, CurrentID=0, Status='live', Symbol='RNU6-1156P', Aliases='', description='RNA, U6 small nuclear 1156, pseudogene', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480076, CurrentID=0, Status='live', Symbol='RNU6-1158P', Aliases='', description='RNA, U6 small nuclear 1158, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480077, CurrentID=0, Status='live', Symbol='RNU6-1161P', Aliases='', description='RNA, U6 small nuclear 1161, pseudogene', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480078, CurrentID=0, Status='live', Symbol='RNU6-1162P', Aliases='', description='RNA, U6 small nuclear 1162, pseudogene', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8269, CurrentID=0, Status='live', Symbol='TMEM187', Aliases='CXorf12, DXS9878E, ITBA1', description='transmembrane protein 187', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153972540, end_position_on_the_genomic_accession=153983195, orientation='plus', exon_count=3, OMIM=300059, no_hdr0=''),
  106480080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480080, CurrentID=0, Status='live', Symbol='RNU6-1169P', Aliases='', description='RNA, U6 small nuclear 1169, pseudogene', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480081, CurrentID=0, Status='live', Symbol='RNU6-1170P', Aliases='', description='RNA, U6 small nuclear 1170, pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480082, CurrentID=0, Status='live', Symbol='RNU6-1175P', Aliases='', description='RNA, U6 small nuclear 1175, pseudogene', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480083, CurrentID=0, Status='live', Symbol='RNU6-1178P', Aliases='', description='RNA, U6 small nuclear 1178, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480084, CurrentID=0, Status='live', Symbol='RNU6-1179P', Aliases='', description='RNA, U6 small nuclear 1179, pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480085, CurrentID=0, Status='live', Symbol='RNU6-1181P', Aliases='', description='RNA, U6 small nuclear 1181, pseudogene', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480086, CurrentID=0, Status='live', Symbol='RNU6-1186P', Aliases='', description='RNA, U6 small nuclear 1186, pseudogene', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480087, CurrentID=0, Status='live', Symbol='RNU6-1190P', Aliases='', description='RNA, U6 small nuclear 1190, pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480088, CurrentID=0, Status='live', Symbol='RNU6-1192P', Aliases='', description='RNA, U6 small nuclear 1192, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480089, CurrentID=0, Status='live', Symbol='RNU6-1196P', Aliases='', description='RNA, U6 small nuclear 1196, pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480090, CurrentID=0, Status='live', Symbol='RNU6-1198P', Aliases='', description='RNA, U6 small nuclear 1198, pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480091, CurrentID=0, Status='live', Symbol='RNU6-1199P', Aliases='', description='RNA, U6 small nuclear 1199, pseudogene', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480092, CurrentID=0, Status='live', Symbol='RNU6-1203P', Aliases='', description='RNA, U6 small nuclear 1203, pseudogene', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480093, CurrentID=0, Status='live', Symbol='RNU6-1205P', Aliases='', description='RNA, U6 small nuclear 1205, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480094, CurrentID=0, Status='live', Symbol='RNU6-1208P', Aliases='', description='RNA, U6 small nuclear 1208, pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480095, CurrentID=0, Status='live', Symbol='RNU6-1209P', Aliases='', description='RNA, U6 small nuclear 1209, pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480096, CurrentID=0, Status='live', Symbol='RNU6-1211P', Aliases='', description='RNA, U6 small nuclear 1211, pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480097, CurrentID=0, Status='live', Symbol='RNU6-1215P', Aliases='', description='RNA, U6 small nuclear 1215, pseudogene', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480098, CurrentID=0, Status='live', Symbol='RNU6-1216P', Aliases='', description='RNA, U6 small nuclear 1216, pseudogene', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480099, CurrentID=0, Status='live', Symbol='RNU6-1218P', Aliases='', description='RNA, U6 small nuclear 1218, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480100, CurrentID=0, Status='live', Symbol='RNU6-1222P', Aliases='', description='RNA, U6 small nuclear 1222, pseudogene', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480101, CurrentID=0, Status='live', Symbol='RNU6-1224P', Aliases='', description='RNA, U6 small nuclear 1224, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480102, CurrentID=0, Status='live', Symbol='RNU6-1226P', Aliases='', description='RNA, U6 small nuclear 1226, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480103, CurrentID=0, Status='live', Symbol='RNU6-1228P', Aliases='', description='RNA, U6 small nuclear 1228, pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480104, CurrentID=0, Status='live', Symbol='RNU6-1231P', Aliases='', description='RNA, U6 small nuclear 1231, pseudogene', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480105, CurrentID=0, Status='live', Symbol='RNU6-1232P', Aliases='', description='RNA, U6 small nuclear 1232, pseudogene', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480106, CurrentID=0, Status='live', Symbol='RNU6-1234P', Aliases='', description='RNA, U6 small nuclear 1234, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480107, CurrentID=0, Status='live', Symbol='RNU6-1238P', Aliases='', description='RNA, U6 small nuclear 1238, pseudogene', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480108, CurrentID=0, Status='live', Symbol='RNU6-1242P', Aliases='', description='RNA, U6 small nuclear 1242, pseudogene', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480109, CurrentID=0, Status='live', Symbol='RNU6-1245P', Aliases='', description='RNA, U6 small nuclear 1245, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442862, CurrentID=0, Status='live', Symbol='PRY2', Aliases='PTPN13LY2', description='PTPN13-like, Y-linked 2', other_designations='PTPN13-like, Y-linked, centromeric|testis-specific PTP-BL-related Y protein|testis-specific PTP-BL-related protein on Y', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22071756, end_position_on_the_genomic_accession=22096007, orientation='minus', exon_count=5, OMIM=400041, no_hdr0=''),
  442863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442863, CurrentID=0, Status='live', Symbol='PRYP1', Aliases='PRY4P, PRYP7', description='PTPN13-like, Y-linked pseudogene 1', other_designations='PTPN13-like, Y-linked 4, pseudogene|PTPN13-like, Y-linked pseudogene 7', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17729799, end_position_on_the_genomic_accession=17744540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442864, CurrentID=0, Status='live', Symbol='PRYP2', Aliases='', description='PTPN13-like, Y-linked pseudogene 2', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18161364, end_position_on_the_genomic_accession=18177051, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442865, CurrentID=0, Status='live', Symbol='PRYP3', Aliases='', description='PTPN13-like, Y-linked pseudogene 3', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23681340, end_position_on_the_genomic_accession=23694681, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442866, CurrentID=0, Status='live', Symbol='PRYP4', Aliases='', description='PTPN13-like, Y-linked pseudogene 4', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25975395, end_position_on_the_genomic_accession=25988736, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  442867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442867, CurrentID=0, Status='live', Symbol='BPY2B', Aliases='VCY2B', description='basic charge, Y-linked, 2B', other_designations='basic charge, Y-linked 2|basic protein on Y chromosome 2|variable charge, Y-linked, 2|variably charged protein Y 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24618004, end_position_on_the_genomic_accession=24639207, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  442868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442868, CurrentID=0, Status='live', Symbol='BPY2C', Aliases='VCY2C', description='basic charge, Y-linked, 2C', other_designations='basic charge, Y-linked 2|basic protein on Y chromosome 2|variable charge, Y-linked, 2|variably charged protein Y 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25030901, end_position_on_the_genomic_accession=25052104, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106480117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480117, CurrentID=0, Status='live', Symbol='RNU6-1264P', Aliases='', description='RNA, U6 small nuclear 1264, pseudogene', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480118, CurrentID=0, Status='live', Symbol='RNU6-1266P', Aliases='', description='RNA, U6 small nuclear 1266, pseudogene', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480119, CurrentID=0, Status='live', Symbol='RNU6-1269P', Aliases='', description='RNA, U6 small nuclear 1269, pseudogene', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480120, CurrentID=0, Status='live', Symbol='RNU6-1271P', Aliases='', description='RNA, U6 small nuclear 1271, pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480121, CurrentID=0, Status='live', Symbol='RNU6-1272P', Aliases='', description='RNA, U6 small nuclear 1272, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480122, CurrentID=0, Status='live', Symbol='RNU6-1275P', Aliases='', description='RNA, U6 small nuclear 1275, pseudogene', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480123, CurrentID=0, Status='live', Symbol='RNU6-1277P', Aliases='', description='RNA, U6 small nuclear 1277, pseudogene', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480124, CurrentID=0, Status='live', Symbol='RNU6-1278P', Aliases='', description='RNA, U6 small nuclear 1278, pseudogene', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480125, CurrentID=0, Status='live', Symbol='RNU6-1281P', Aliases='', description='RNA, U6 small nuclear 1281, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480126, CurrentID=0, Status='live', Symbol='RNU6-1284P', Aliases='', description='RNA, U6 small nuclear 1284, pseudogene', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480127, CurrentID=0, Status='live', Symbol='RNU6-1287P', Aliases='', description='RNA, U6 small nuclear 1287, pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480128, CurrentID=0, Status='live', Symbol='RNU6-1288P', Aliases='', description='RNA, U6 small nuclear 1288, pseudogene', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480129, CurrentID=0, Status='live', Symbol='RNU6-1290P', Aliases='', description='RNA, U6 small nuclear 1290, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480130, CurrentID=0, Status='live', Symbol='RNU6-1293P', Aliases='', description='RNA, U6 small nuclear 1293, pseudogene', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480131, CurrentID=0, Status='live', Symbol='RNU6-1294P', Aliases='', description='RNA, U6 small nuclear 1294, pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480132, CurrentID=0, Status='live', Symbol='RNU6-1296P', Aliases='', description='RNA, U6 small nuclear 1296, pseudogene', other_designations='', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480133, CurrentID=0, Status='live', Symbol='RNU6-1299P', Aliases='', description='RNA, U6 small nuclear 1299, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480134, CurrentID=0, Status='live', Symbol='RNU6-1300P', Aliases='', description='RNA, U6 small nuclear 1300, pseudogene', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480135, CurrentID=0, Status='live', Symbol='RNU6-1304P', Aliases='', description='RNA, U6 small nuclear 1304, pseudogene', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480136, CurrentID=0, Status='live', Symbol='RNU6-1306P', Aliases='', description='RNA, U6 small nuclear 1306, pseudogene', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480137, CurrentID=0, Status='live', Symbol='RNU6-1308P', Aliases='', description='RNA, U6 small nuclear 1308, pseudogene', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  442890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442890, CurrentID=0, Status='live', Symbol='MIR133B', Aliases='MIRN133B, miRNA133B, mir-133b', description='microRNA 133b', other_designations='hsa-mir-133b', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52148923, end_position_on_the_genomic_accession=52149041, orientation='plus', exon_count=1, OMIM=610946, no_hdr0=''),
  442891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442891, CurrentID=0, Status='live', Symbol='MIR135B', Aliases='MIRN135B, mir-135b', description='microRNA 135b', other_designations='hsa-mir-135b', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205448302, end_position_on_the_genomic_accession=205448398, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442892, CurrentID=0, Status='live', Symbol='MIR148B', Aliases='MIRN148B, mir-148b', description='microRNA 148b', other_designations='hsa-mir-148b', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54337216, end_position_on_the_genomic_accession=54337314, orientation='plus', exon_count=1, OMIM=613787, no_hdr0=''),
  442893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442893, CurrentID=0, Status='live', Symbol='MIR151A', Aliases='MIR151, MIRN151, hsa-mir-151, hsa-mir-151a', description='microRNA 151a', other_designations='microRNA 151', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140732564, end_position_on_the_genomic_accession=140732653, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442894, CurrentID=0, Status='live', Symbol='MIR302B', Aliases='MIRN302B, mir-302b', description='microRNA 302b', other_designations='hsa-mir-302b', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112648485, end_position_on_the_genomic_accession=112648557, orientation='minus', exon_count=1, OMIM=614597, no_hdr0=''),
  442895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442895, CurrentID=0, Status='live', Symbol='MIR302C', Aliases='MIRN302C, mir-302c', description='microRNA 302c', other_designations='hsa-mir-302c', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112648363, end_position_on_the_genomic_accession=112648430, orientation='minus', exon_count=1, OMIM=614598, no_hdr0=''),
  442896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442896, CurrentID=0, Status='live', Symbol='MIR302D', Aliases='MIRN302D, mir-302d', description='microRNA 302d', other_designations='hsa-mir-302d', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112648004, end_position_on_the_genomic_accession=112648071, orientation='minus', exon_count=1, OMIM=614599, no_hdr0=''),
  442897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442897, CurrentID=0, Status='live', Symbol='MIR323A', Aliases='MIR323, MIRN323, hsa-mir-323, hsa-mir-323a, mir-323a', description='microRNA 323a', other_designations='microRNA 323', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101025732, end_position_on_the_genomic_accession=101025817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442898, CurrentID=0, Status='live', Symbol='MIR324', Aliases='MIRN324, hsa-mir-324, mir-324', description='microRNA 324', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7223297, end_position_on_the_genomic_accession=7223379, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  573971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=573971, CurrentID=0, Status='live', Symbol='DHFRP1', Aliases='', description='dihydrofolate reductase pseudogene 1', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26167862, end_position_on_the_genomic_accession=26171357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  442900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442900, CurrentID=0, Status='live', Symbol='MIR326', Aliases='MIRN326, hsa-mir-326, mir-326', description='microRNA 326', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75335092, end_position_on_the_genomic_accession=75335186, orientation='minus', exon_count=1, OMIM=613755, no_hdr0=''),
  442901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442901, CurrentID=0, Status='live', Symbol='MIR328', Aliases='MIRN328, hsa-mir-328, mir-328', description='microRNA 328', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67202321, end_position_on_the_genomic_accession=67202395, orientation='minus', exon_count=1, OMIM=613701, no_hdr0=''),
  442902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442902, CurrentID=0, Status='live', Symbol='MIR330', Aliases='MIRN330, hsa-mir-330, mir-330', description='microRNA 330', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45638994, end_position_on_the_genomic_accession=45639087, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442903, CurrentID=0, Status='live', Symbol='MIR331', Aliases='MIRN331, hsa-mir-331, mir-331', description='microRNA 331', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95308420, end_position_on_the_genomic_accession=95308513, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442904, CurrentID=0, Status='live', Symbol='MIR335', Aliases='MIRN335, hsa-mir-335, miRNA335, mir-335', description='microRNA 335', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130496111, end_position_on_the_genomic_accession=130496204, orientation='plus', exon_count=1, OMIM=611768, no_hdr0=''),
  442905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442905, CurrentID=0, Status='live', Symbol='MIR337', Aliases='MIRN337, hsa-mir-337, mir-337', description='microRNA 337', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100874493, end_position_on_the_genomic_accession=100874585, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442906, CurrentID=0, Status='live', Symbol='MIR338', Aliases='MIRN338, hsa-mir-338, mir-338', description='microRNA 338', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81125883, end_position_on_the_genomic_accession=81125949, orientation='minus', exon_count=1, OMIM=614059, no_hdr0=''),
  442907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442907, CurrentID=0, Status='live', Symbol='MIR339', Aliases='MIRN339, hsa-mir-339, mir-339', description='microRNA 339', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1022933, end_position_on_the_genomic_accession=1023026, orientation='minus', exon_count=1, OMIM=615977, no_hdr0=''),
  442908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442908, CurrentID=0, Status='live', Symbol='MIR340', Aliases='MIRN340, hsa-mir-340, mir-340', description='microRNA 340', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180015303, end_position_on_the_genomic_accession=180015397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442909, CurrentID=0, Status='live', Symbol='MIR342', Aliases='MIRN342, hsa-mir-342', description='microRNA 342', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100109655, end_position_on_the_genomic_accession=100109753, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442910, CurrentID=0, Status='live', Symbol='MIR345', Aliases='MIRN345, hsa-mir-345, mir-345', description='microRNA 345', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100307859, end_position_on_the_genomic_accession=100307956, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442911, CurrentID=0, Status='live', Symbol='MIR346', Aliases='MIRN346, hsa-mir-346, miR-346, miRNA346', description='microRNA 346', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86264694, end_position_on_the_genomic_accession=86264788, orientation='minus', exon_count=1, OMIM=611190, no_hdr0=''),
  442912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442912, CurrentID=0, Status='live', Symbol='MIR367', Aliases='MIRN367, hsa-mir-367', description='microRNA 367', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112647874, end_position_on_the_genomic_accession=112647941, orientation='minus', exon_count=1, OMIM=614600, no_hdr0=''),
  442913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442913, CurrentID=0, Status='live', Symbol='MIR376C', Aliases='MIR368, MIRN368, MIRN376C, hsa-mir-368, miRNA368, mir-376c', description='microRNA 376c', other_designations='hsa-mir-376c|microRNA 368', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101039690, end_position_on_the_genomic_accession=101039755, orientation='plus', exon_count=1, OMIM=610983, no_hdr0=''),
  442914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442914, CurrentID=0, Status='live', Symbol='MIR369', Aliases='MIR369-3, MIRN369, MIRN369-3, hsa-mir-369, mir-369', description='microRNA 369', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101065598, end_position_on_the_genomic_accession=101065667, orientation='plus', exon_count=1, OMIM=611794, no_hdr0=''),
  442915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442915, CurrentID=0, Status='live', Symbol='MIR370', Aliases='MIRN370, hsa-mir-370, mir-370', description='microRNA 370', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100911139, end_position_on_the_genomic_accession=100911213, orientation='plus', exon_count=1, OMIM=612553, no_hdr0=''),
  442916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442916, CurrentID=0, Status='live', Symbol='MIR371A', Aliases='MIR371, MIRN371, hsa-mir-371, hsa-mir-371a', description='microRNA 371a', other_designations='microRNA 371', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53787675, end_position_on_the_genomic_accession=53787741, orientation='plus', exon_count=1, OMIM=612043, no_hdr0=''),
  442917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442917, CurrentID=0, Status='live', Symbol='MIR372', Aliases='MIRN372, hsa-mir-372', description='microRNA 372', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53787890, end_position_on_the_genomic_accession=53787956, orientation='plus', exon_count=1, OMIM=612044, no_hdr0=''),
  442918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442918, CurrentID=0, Status='live', Symbol='MIR373', Aliases='MIRN373, hsa-mir-373, miRNA373, mir-373', description='microRNA 373', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53788705, end_position_on_the_genomic_accession=53788773, orientation='plus', exon_count=1, OMIM=611954, no_hdr0=''),
  442919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442919, CurrentID=0, Status='live', Symbol='MIR374A', Aliases='MIRN374, MIRN374A, hsa-mir-374, mir-374a', description='microRNA 374a', other_designations='hsa-mir-374a|microRNA 374', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74287286, end_position_on_the_genomic_accession=74287357, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442920, CurrentID=0, Status='live', Symbol='MIR196B', Aliases='MIRN196B, miR-196b, miRNA196B', description='microRNA 196b', other_designations='hsa-mir-196b', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27169480, end_position_on_the_genomic_accession=27169563, orientation='minus', exon_count=1, OMIM=609688, no_hdr0=''),
  106480169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480169, CurrentID=0, Status='live', Symbol='RNU1-67P', Aliases='', description='RNA, U1 small nuclear 67, pseudogene', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480170, CurrentID=0, Status='live', Symbol='RNU1-69P', Aliases='', description='RNA, U1 small nuclear 69, pseudogene', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480171, CurrentID=0, Status='live', Symbol='RNU1-74P', Aliases='', description='RNA, U1 small nuclear 74, pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480172, CurrentID=0, Status='live', Symbol='RNU1-75P', Aliases='', description='RNA, U1 small nuclear 75, pseudogene', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480173, CurrentID=0, Status='live', Symbol='RNU1-77P', Aliases='', description='RNA, U1 small nuclear 77, pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480174, CurrentID=0, Status='live', Symbol='RNU1-80P', Aliases='', description='RNA, U1 small nuclear 80, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480175, CurrentID=0, Status='live', Symbol='RNU1-82P', Aliases='', description='RNA, U1 small nuclear 82, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480176, CurrentID=0, Status='live', Symbol='RNU1-83P', Aliases='', description='RNA, U1 small nuclear 83, pseudogene', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480177, CurrentID=0, Status='live', Symbol='RNU1-85P', Aliases='', description='RNA, U1 small nuclear 85, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480178, CurrentID=0, Status='live', Symbol='RNU1-87P', Aliases='', description='RNA, U1 small nuclear 87, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480179, CurrentID=0, Status='live', Symbol='RNU1-91P', Aliases='', description='RNA, U1 small nuclear 91, pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480180, CurrentID=0, Status='live', Symbol='RNU1-93P', Aliases='HTR2B, PSMD1, TSARG1', description='RNA, U1 small nuclear 93, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480181, CurrentID=0, Status='live', Symbol='RNU1-96P', Aliases='', description='RNA, U1 small nuclear 96, pseudogene', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480182, CurrentID=0, Status='live', Symbol='RNU1-97P', Aliases='', description='RNA, U1 small nuclear 97, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480183, CurrentID=0, Status='live', Symbol='RNU1-100P', Aliases='', description='RNA, U1 small nuclear 100, pseudogene', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480184, CurrentID=0, Status='live', Symbol='RNU1-104P', Aliases='', description='RNA, U1 small nuclear 104, pseudogene', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480185, CurrentID=0, Status='live', Symbol='RNU1-105P', Aliases='', description='RNA, U1 small nuclear 105, pseudogene', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480186, CurrentID=0, Status='live', Symbol='RNU1-107P', Aliases='', description='RNA, U1 small nuclear 107, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480187, CurrentID=0, Status='live', Symbol='RNU1-112P', Aliases='', description='RNA, U1 small nuclear 112, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480188, CurrentID=0, Status='live', Symbol='RNU1-115P', Aliases='', description='RNA, U1 small nuclear 115, pseudogene', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574013, CurrentID=0, Status='live', Symbol='SMAN1', Aliases='', description='survival of motor and autonomic neurons 1', other_designations='', map_location='20q13.2-q13.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480190, CurrentID=0, Status='live', Symbol='RNU1-122P', Aliases='', description='RNA, U1 small nuclear 122, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480191, CurrentID=0, Status='live', Symbol='RNU1-128P', Aliases='', description='RNA, U1 small nuclear 128, pseudogene', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574016, CurrentID=0, Status='live', Symbol='CLLU1OS', Aliases='', description='chronic lymphocytic leukemia up-regulated 1 opposite strand', other_designations='chronic lymphocytic leukemia up-regulated 1 overlapping strand', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92420094, end_position_on_the_genomic_accession=92428545, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480193, CurrentID=0, Status='live', Symbol='RNU1-131P', Aliases='', description='RNA, U1 small nuclear 131, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480194, CurrentID=0, Status='live', Symbol='RNU1-134P', Aliases='', description='RNA, U1 small nuclear 134, pseudogene', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480195, CurrentID=0, Status='live', Symbol='RNU1-138P', Aliases='', description='RNA, U1 small nuclear 138, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480196, CurrentID=0, Status='live', Symbol='RNU1-139P', Aliases='', description='RNA, U1 small nuclear 139, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480197, CurrentID=0, Status='live', Symbol='RNU1-142P', Aliases='', description='RNA, U1 small nuclear 142, pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480198, CurrentID=0, Status='live', Symbol='RNU1-146P', Aliases='', description='RNA, U1 small nuclear 146, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401505, CurrentID=0, Status='live', Symbol='TOMM5', Aliases='C9orf105, Tom5, bA613M10.3', description='translocase of outer mitochondrial membrane 5 homolog (yeast)', other_designations='mitochondrial outer membrane protein TOM5', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37588413, end_position_on_the_genomic_accession=37592639, orientation='minus', exon_count=2, OMIM=616169, no_hdr0=''),
  106480200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480200, CurrentID=0, Status='live', Symbol='RNU2-8P', Aliases='', description='RNA, U2 small nuclear 8, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480201, CurrentID=0, Status='live', Symbol='RNU2-10P', Aliases='', description='RNA, U2 small nuclear 10, pseudogene', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480202, CurrentID=0, Status='live', Symbol='RNU2-12P', Aliases='', description='RNA, U2 small nuclear 12, pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271041, CurrentID=0, Status='live', Symbol='RPL7AP24', Aliases='RPL7A_6_424', description='ribosomal protein L7a pseudogene 24', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155760562, end_position_on_the_genomic_accession=155761223, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  574028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574028, CurrentID=0, Status='live', Symbol='CLLU1', Aliases='', description='chronic lymphocytic leukemia up-regulated 1', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92421531, end_position_on_the_genomic_accession=92431002, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  574029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574029, CurrentID=0, Status='live', Symbol='DUSP5P1', Aliases='DUSP5P', description='dual specificity phosphatase 5 pseudogene 1', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228644910, end_position_on_the_genomic_accession=228652412, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  574030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574030, CurrentID=0, Status='live', Symbol='MIR362', Aliases='MIRN362, hsa-mir-362, mir-362', description='microRNA 362', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50008964, end_position_on_the_genomic_accession=50009028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574031, CurrentID=0, Status='live', Symbol='MIR363', Aliases='MIR-363, MIRN363, hsa-mir-363', description='microRNA 363', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134169378, end_position_on_the_genomic_accession=134169452, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574032, CurrentID=0, Status='live', Symbol='MIR20B', Aliases='MIRN20B, hsa-mir-20b, mir-20b', description='microRNA 20b', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134169809, end_position_on_the_genomic_accession=134169877, orientation='minus', exon_count=1, OMIM=300950, no_hdr0=''),
  574033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574033, CurrentID=0, Status='live', Symbol='MIR18B', Aliases='MIRN18B, hsa-mir-18b, mir-18b', description='microRNA 18b', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134170041, end_position_on_the_genomic_accession=134170111, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574034, CurrentID=0, Status='live', Symbol='MIR433', Aliases='MIRN433, hsa-mir-433, miRNA433, mir-433', description='microRNA 433', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100881886, end_position_on_the_genomic_accession=100881978, orientation='plus', exon_count=1, OMIM=611711, no_hdr0=''),
  401507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401507, CurrentID=0, Status='live', Symbol='FAM74A1', Aliases='FAM74A5', description='family with sequence similarity 74 member A1', other_designations='family with sequence similarity 74, member A1|family with sequence similarity 74, member A5', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39371065, end_position_on_the_genomic_accession=39376916, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  574036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574036, CurrentID=0, Status='live', Symbol='SERTAD4-AS1', Aliases='C1orf133', description='SERTAD4 antisense RNA 1', other_designations='SERTAD4 antisense RNA 1 (non-protein coding)', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210231459, end_position_on_the_genomic_accession=210234121, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480213, CurrentID=0, Status='live', Symbol='RNU2-37P', Aliases='', description='RNA, U2 small nuclear 37, pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574038, CurrentID=0, Status='live', Symbol='MIR431', Aliases='MIRN431, hsa-mir-431, miRNA431, mir-431', description='microRNA 431', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100881007, end_position_on_the_genomic_accession=100881120, orientation='plus', exon_count=1, OMIM=611708, no_hdr0=''),
  106480215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480215, CurrentID=0, Status='live', Symbol='RNU2-41P', Aliases='', description='RNA, U2 small nuclear 41, pseudogene', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574040, CurrentID=0, Status='live', Symbol='SNORA6', Aliases='ACA6', description='small nucleolar RNA, H/ACA box 6', other_designations='ACA6 small nucleolar RNA|ACA6 snoRNA|small nucleolar RNA M2', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39408391, end_position_on_the_genomic_accession=39408539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  401508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401508, CurrentID=0, Status='live', Symbol='FAM74A4', Aliases='FAM74A2', description='family with sequence similarity 74 member A4', other_designations='family with sequence similarity 74, member A2|family with sequence similarity 74, member A4', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61205260, end_position_on_the_genomic_accession=61212373, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  574042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574042, CurrentID=0, Status='live', Symbol='SNORA10', Aliases='ACA10', description='small nucleolar RNA, H/ACA box 10', other_designations='ACA10 small nucleolar RNA within ribosomal protein 2 host|ACA10 snoRNA', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1962334, end_position_on_the_genomic_accession=1962466, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574043, CurrentID=0, Status='live', Symbol='CDAGS', Aliases='', description='Craniosynostosis, anal anomalies, and porokeratosis syndrome', other_designations='', map_location='22q12-q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603116, no_hdr0=''),
  574044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574044, CurrentID=0, Status='live', Symbol='EPRPDC', Aliases='', description="Epilepsy, rolandic, with paroxysmal exercise-induced dystonia and writer's cramp", other_designations='', map_location='16p12-p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608105, no_hdr0=''),
  574045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574045, CurrentID=0, Status='live', Symbol='HNP1', Aliases='', description='Hypertensive nephropathy', other_designations='', map_location='9q31-q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608026, no_hdr0=''),
  106480222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480222, CurrentID=0, Status='live', Symbol='RNU2-57P', Aliases='', description='RNA, U2 small nuclear 57, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574047, CurrentID=0, Status='live', Symbol='MRXSMP', Aliases='', description='Martin-Probst deafness-mental retardation syndrome', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300519, no_hdr0=''),
  574048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574048, CurrentID=0, Status='live', Symbol='OVCAS1', Aliases='', description='Ovarian cancer, epithelial, susceptibility to', other_designations='', map_location='3p25-p22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607893, no_hdr0=''),
  574049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574049, CurrentID=0, Status='live', Symbol='TUKLS', Aliases='', description='Tukel syndrome', other_designations='', map_location='21q22', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609428, no_hdr0=''),
  106480226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480226, CurrentID=0, Status='live', Symbol='RNU2-65P', Aliases='', description='RNA, U2 small nuclear 65, pseudogene', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188771, CurrentID=0, Status='live', Symbol='AGA2', Aliases='', description='Alopecia, androgenetic, 2', other_designations='', map_location='Xq11-q12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300710, no_hdr0=''),
  100188772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188772, CurrentID=0, Status='live', Symbol='IHPS4', Aliases='', description='Pyloric stenosis, infantile hypertrophic, 4', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300711, no_hdr0=''),
  100188773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188773, CurrentID=0, Status='live', Symbol='CFSS', Aliases='', description='craniofacioskeletal syndrome', other_designations='', map_location='Xq26-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300712, no_hdr0=''),
  100188774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188774, CurrentID=0, Status='live', Symbol='DFCTRPS', Aliases='', description='Deafness, cataract, retinitis pigmentosa, and sperm abnormalities', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300719, no_hdr0=''),
  100188775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188775, CurrentID=0, Status='live', Symbol='LOAS', Aliases='', description='Leber optic atrophy, susceptibility to', other_designations='', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=308905, no_hdr0=''),
  100188776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188776, CurrentID=0, Status='live', Symbol='HEY', Aliases='', description='hairy ears, Y-linked', other_designations='', map_location='Yq', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=425500, no_hdr0=''),
  106480233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480233, CurrentID=0, Status='live', Symbol='HNRNPH3P1', Aliases='', description='heterogeneous nuclear ribonucleoprotein H3 pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480234, CurrentID=0, Status='live', Symbol='HNRNPA3P15', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 15', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480235, CurrentID=0, Status='live', Symbol='HNRNPA1P62', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 62', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480236, CurrentID=0, Status='live', Symbol='HNRNPA1P72', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 72', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480237, CurrentID=0, Status='live', Symbol='MTHFD2P2', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 2', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188782, CurrentID=0, Status='live', Symbol='NIDDM4', Aliases='', description='Diabetes mellitus, noninsulin-dependent', other_designations='', map_location='5q34-q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608036, no_hdr0=''),
  106480239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480239, CurrentID=0, Status='live', Symbol='SLC2A13P1', Aliases='', description='SLC2A13 pseudogene 1', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188784, CurrentID=0, Status='live', Symbol='AFA1', Aliases='MPB', description='Alopecia, androgenetic', other_designations='', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109200, no_hdr0=''),
  106480241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480241, CurrentID=0, Status='live', Symbol='CLPTM1LP1', Aliases='', description='CLPTM1L pseudogene 1', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480242, CurrentID=0, Status='live', Symbol='PRELID3BP5', Aliases='', description='PRELI domain containing 3B pseudogene 5', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188787, CurrentID=0, Status='live', Symbol='WM1', Aliases='', description='Macroglobulinemia, Waldenstrom, susceptibility to, 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=153600, no_hdr0=''),
  100188788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188788, CurrentID=0, Status='live', Symbol='DEL3Q29', Aliases='MICRODEL3Q29', description='Chromosome 3q29 microdeletion syndrome', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609425, no_hdr0=''),
  100188789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188789, CurrentID=0, Status='live', Symbol='HPC6', Aliases='', description='Prostate cancer, susceptibility to', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609558, no_hdr0=''),
  100188790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188790, CurrentID=0, Status='live', Symbol='MGR8', Aliases='', description='Migraine, susceptibility to, 8', other_designations='', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609570, no_hdr0=''),
  100188791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188791, CurrentID=0, Status='live', Symbol='CLLS1', Aliases='', description='Leukemia, chronic lymphocytic, susceptibility to, 1', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609630, no_hdr0=''),
  100188792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188792, CurrentID=0, Status='live', Symbol='MAFD3', Aliases='BPEO', description='Major affective disorder 3, early onset', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609633, no_hdr0=''),
  100188793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188793, CurrentID=0, Status='live', Symbol='BSZQTL', Aliases='', description='Bone size QTL', other_designations='', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609656, no_hdr0=''),
  100188794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188794, CurrentID=0, Status='live', Symbol='BSZQTL2', Aliases='', description='Bone size QTL', other_designations='', map_location='5q', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609657, no_hdr0=''),
  100188795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188795, CurrentID=0, Status='live', Symbol='AD11', Aliases='', description='Alzheimer disease-11', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609790, no_hdr0=''),
  100188796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188796, CurrentID=0, Status='live', Symbol='GEFSP4', Aliases='', description='Epilepsy, generalized, with febrile seizures plus, type 4', other_designations='', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609800, no_hdr0=''),
  574077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574077, CurrentID=0, Status='live', Symbol='GLC1K', Aliases='JOAG3', description='glaucoma 1, open angle, K (juvenile-onset)', other_designations='', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608696, no_hdr0=''),
  574078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574078, CurrentID=0, Status='live', Symbol='GLC1J', Aliases='JOAG2', description='glaucoma 1, open angle, J (juvenile-onset)', other_designations='', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608695, no_hdr0=''),
  106480255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480255, CurrentID=0, Status='live', Symbol='HSPE1P9', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 9', other_designations='heat shock 10kDa protein 1 pseudogene 9', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188800, CurrentID=0, Status='live', Symbol='ASPG4', Aliases='', description='Asperger syndrome, susceptibility to, 4', other_designations='', map_location='3p24-p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609954, no_hdr0=''),
  100188801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188801, CurrentID=0, Status='live', Symbol='ASRT3', Aliases='', description='Asthma-related traits, susceptibility to, 3', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609958, no_hdr0=''),
  100188802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188802, CurrentID=0, Status='live', Symbol='ODS1', Aliases='', description='Opioid dependence, susceptibility to, 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610064, no_hdr0=''),
  100188803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188803, CurrentID=0, Status='live', Symbol='SLEB7', Aliases='', description='Systemic lupus erythematosus, susceptibility to, 7', other_designations='', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610065, no_hdr0=''),
  100188804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188804, CurrentID=0, Status='live', Symbol='SLEB8', Aliases='', description='Systemic lupus erythematosus, susceptibility to, 8', other_designations='', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610066, no_hdr0=''),
  100188805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188805, CurrentID=0, Status='live', Symbol='STBMS1', Aliases='', description='Strabismus, susceptibility to, 1', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=185100, no_hdr0=''),
  100188806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188806, CurrentID=0, Status='live', Symbol='HDLCQ4', Aliases='', description='High density lipoprotein cholesterol level QTL 4', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610239, no_hdr0=''),
  100188807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188807, CurrentID=0, Status='live', Symbol='HYT5', Aliases='', description='Hypertension, essential, susceptibility to, 5', other_designations='', map_location='20q11-q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610261, no_hdr0=''),
  100188808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188808, CurrentID=0, Status='live', Symbol='HYT6', Aliases='', description='Hypertension, essential, susceptibility to, 6', other_designations='', map_location='5p13-q12', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610262, no_hdr0=''),
  100188809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188809, CurrentID=0, Status='live', Symbol='HPC7', Aliases='', description='Prostate cancer, hereditary, 7', other_designations='', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610321, no_hdr0=''),
  100188810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188810, CurrentID=0, Status='live', Symbol='ANIB5', Aliases='', description='Aneurysm, intracranial berry, 5', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610402, no_hdr0=''),
  100188811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188811, CurrentID=0, Status='live', Symbol='WM2', Aliases='', description='Macroglobulinemia, Waldenstrom, susceptibility to, 2', other_designations='', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610430, no_hdr0=''),
  100188812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188812, CurrentID=0, Status='live', Symbol='RLS3', Aliases='', description='Restless legs syndrome 3', other_designations='', map_location='9p24-p22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610438, no_hdr0=''),
  100188813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188813, CurrentID=0, Status='live', Symbol='RLS4', Aliases='', description='Restless legs syndrome 4', other_designations='', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610439, no_hdr0=''),
  100188814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188814, CurrentID=0, Status='live', Symbol='DEL16P13.3', Aliases='RSTSS', description='Chromosome 16p13.3 deletion syndrome', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610543, no_hdr0=''),
  100188815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188815, CurrentID=0, Status='live', Symbol='BSZQTL3', Aliases='', description='Bone size quantitative trait locus 3', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610649, no_hdr0=''),
  100188816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188816, CurrentID=0, Status='live', Symbol='AUTS7', Aliases='', description='Autism, susceptibility to, 7', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610676, no_hdr0=''),
  100188817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188817, CurrentID=0, Status='live', Symbol='CLQTL2', Aliases='', description='Cholesterol level QTL 2', other_designations='', map_location='2p24-p22', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610760, no_hdr0=''),
  100188818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188818, CurrentID=0, Status='live', Symbol='HDLCQ5', Aliases='', description='High density lipoprotein cholesterol level QTL 5', other_designations='', map_location='3q24-q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610761, no_hdr0=''),
  100188819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188819, CurrentID=0, Status='live', Symbol='HDLCQ6', Aliases='', description='High density lipoprotein cholesterol level QTL 6', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610762, no_hdr0=''),
  100188820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188820, CurrentID=0, Status='live', Symbol='AUTS11', Aliases='', description='Autism, susceptibility to, 11', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610836, no_hdr0=''),
  100188821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188821, CurrentID=0, Status='live', Symbol='OS4', Aliases='GOA1', description='Osteoarthritis, generalized, without dysplasia, susceptibility to', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610839, no_hdr0=''),
  100188822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188822, CurrentID=0, Status='live', Symbol='MENAQ1', Aliases='', description='Menarche, age at, QTL', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610873, no_hdr0=''),
  100188823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188823, CurrentID=0, Status='live', Symbol='ASRT4', Aliases='', description='Asthma-related traits, susceptibility to, 4', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610906, no_hdr0=''),
  106480280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480280, CurrentID=0, Status='live', Symbol='BNIP3P9', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 9', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188825, CurrentID=0, Status='live', Symbol='HYT7', Aliases='', description='Hypertension, essential, susceptibility to, 7', other_designations='', map_location='3p14.1-q12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610948, no_hdr0=''),
  100188826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188826, CurrentID=0, Status='live', Symbol='HPC9', Aliases='', description='Prostate cancer, hereditary, 9', other_designations='', map_location='17q21-q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610997, no_hdr0=''),
  100188827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188827, CurrentID=0, Status='live', Symbol='SQTL1', Aliases='', description='Smoking as a quantitative trait locus 1', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611003, no_hdr0=''),
  100188828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188828, CurrentID=0, Status='live', Symbol='SQTL2', Aliases='', description='Smoking as a quantitative trait locus 2', other_designations='', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611004, no_hdr0=''),
  100188829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188829, CurrentID=0, Status='live', Symbol='MTBS2', Aliases='', description='Mycobacterium tuberculosis, susceptibility to, 2', other_designations='', map_location='8q12-q13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611046, no_hdr0=''),
  100188830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188830, CurrentID=0, Status='live', Symbol='AD12', Aliases='', description='Alzheimer disease 12', other_designations='', map_location='8p12-q22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611073, no_hdr0=''),
  106480287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480287, CurrentID=0, Status='live', Symbol='LARP7P1', Aliases='', description='La ribonucleoprotein domain family member 7 pseudogene 1', other_designations='La ribonucleoprotein domain family, member 7 pseudogene 1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188832, CurrentID=0, Status='live', Symbol='AUTS1', Aliases='', description='Autism, susceptibility to, 1', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=209850, no_hdr0=''),
  106480289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480289, CurrentID=0, Status='live', Symbol='C11orf98P1', Aliases='', description='chromosome 11 open reading frame 98 pseudogene 1', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188834, CurrentID=0, Status='live', Symbol='HPC10', Aliases='', description='Prostate cancer, hereditary, 10', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611100, no_hdr0=''),
  100188835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188835, CurrentID=0, Status='live', Symbol='CINN', Aliases='', description='Cinnamon odor, pleasantness of', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611109, no_hdr0=''),
  100188836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188836, CurrentID=0, Status='live', Symbol='CHDS8', Aliases='', description='Coronary heart disease, susceptibility to, 8', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611139, no_hdr0=''),
  100188837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188837, CurrentID=0, Status='live', Symbol='AD13', Aliases='', description='Alzheimer disease-13', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611152, no_hdr0=''),
  100188838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188838, CurrentID=0, Status='live', Symbol='AD15', Aliases='', description='Alzheimer disease-15', other_designations='', map_location='3q22-q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611155, no_hdr0=''),
  100188839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188839, CurrentID=0, Status='live', Symbol='RLS5', Aliases='', description='Restless legs syndrome, susceptibility to, 5', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611242, no_hdr0=''),
  100188840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188840, CurrentID=0, Status='live', Symbol='ASRT6', Aliases='', description='Asthma-related traits, susceptibility to, 6', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611403, no_hdr0=''),
  100188841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188841, CurrentID=0, Status='live', Symbol='CRCS2', Aliases='', description='Colorectal cancer, susceptibility to, 2', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611469, no_hdr0=''),
  100188842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188842, CurrentID=0, Status='live', Symbol='ATFB5', Aliases='', description='Atrial fibrillation, familial, 5', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611494, no_hdr0=''),
  100188843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188843, CurrentID=0, Status='live', Symbol='MAFD5', Aliases='', description='Major affective disorder 5', other_designations='', map_location='2q22-q24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611535, no_hdr0=''),
  100188844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188844, CurrentID=0, Status='live', Symbol='MAFD6', Aliases='BPAD', description='Major affective disorder 6', other_designations='', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611536, no_hdr0=''),
  100188845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188845, CurrentID=0, Status='live', Symbol='CODA', Aliases='', description='Cavitary optic disc anomalies', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611543, no_hdr0=''),
  100188846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188846, CurrentID=0, Status='live', Symbol='CELIAC6', Aliases='AIS5', description='Celiac disease, susceptibility to, 6', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611598, no_hdr0=''),
  100188847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188847, CurrentID=0, Status='live', Symbol='MIR1225', Aliases='MIRN1225', description='microRNA 1225', other_designations='hsa-mir-1225', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2090195, end_position_on_the_genomic_accession=2090284, orientation='minus', exon_count=1, OMIM=611621, no_hdr0=''),
  100188848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188848, CurrentID=0, Status='live', Symbol='FMTLE', Aliases='ETL3', description='Epilepsy, familial mesial temporal lobe', other_designations='', map_location='4q13.2-q21.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611630, no_hdr0=''),
  100188849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188849, CurrentID=0, Status='live', Symbol='FEB9', Aliases='', description='Febrile convulsions, familial, 9', other_designations='', map_location='3p24.2-p23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611634, no_hdr0=''),
  100188850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188850, CurrentID=0, Status='live', Symbol='HSCR9', Aliases='', description='Hirschsprung disease, susceptibility to, 9', other_designations='', map_location='4q31.3-q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611644, no_hdr0=''),
  100188851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188851, CurrentID=0, Status='live', Symbol='MGR12', Aliases='', description='Migraine, with or without aura, susceptibility to, 12', other_designations='', map_location='10q22-q23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611706, no_hdr0=''),
  100188852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188852, CurrentID=0, Status='live', Symbol='SHEP8', Aliases='', description='Skin/hair/eye pigmentation 8, freckling', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611724, no_hdr0=''),
  100188853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188853, CurrentID=0, Status='live', Symbol='BMND7', Aliases='', description='bone mineral density quantiative trait locus 7', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611738, no_hdr0=''),
  100188854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188854, CurrentID=0, Status='live', Symbol='BMND8', Aliases='', description='bone mineral density quantiative trait locus 8', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611739, no_hdr0=''),
  100188855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188855, CurrentID=0, Status='live', Symbol='MNDEC', Aliases='', description='Microtia with nasolacrimal duct imperforation and eye coloboma', other_designations='', map_location='4p16-p15', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611863, no_hdr0=''),
  100188856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188856, CurrentID=0, Status='live', Symbol='DEL22Q11.2', Aliases='C22DDELS, C22DELq11.2', description='Chromosome 22q11.2 deletion syndrome, distal', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611867, no_hdr0=''),
  100188857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188857, CurrentID=0, Status='live', Symbol='AAA3', Aliases='', description='Aneurysm, familial abdominal 3', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611891, no_hdr0=''),
  100188858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188858, CurrentID=0, Status='live', Symbol='ANIB6', Aliases='', description='Aneurysm, intracranial berry, 6', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611892, no_hdr0=''),
  100188859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188859, CurrentID=0, Status='live', Symbol='EA7', Aliases='', description='Episodic ataxia, type 7', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611907, no_hdr0=''),
  100188860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188860, CurrentID=0, Status='live', Symbol='CPROTQ', Aliases='', description='C-reactive protein QTL', other_designations='', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611920, no_hdr0=''),
  100188861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188861, CurrentID=0, Status='live', Symbol='EIG5', Aliases='', description='Epilepsy, idiopathic generalized, susceptibility to, 5', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611934, no_hdr0=''),
  49854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49854, CurrentID=0, Status='live', Symbol='ZBTB21', Aliases='ZNF295', description='zinc finger and BTB domain containing 21', other_designations='zinc finger protein 295', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41986831, end_position_on_the_genomic_accession=42010387, orientation='minus', exon_count=7, OMIM=616485, no_hdr0=''),
  49855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49855, CurrentID=0, Status='live', Symbol='SCAPER', Aliases='MSTP063, ZNF291, Zfp291', description='S-phase cyclin A-associated protein in the ER', other_designations='S phase cyclin A-associated protein in the ER|zinc finger protein 291', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76348185, end_position_on_the_genomic_accession=76905455, orientation='minus', exon_count=39, OMIM=611611, no_hdr0=''),
  49856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49856, CurrentID=0, Status='live', Symbol='WRAP73', Aliases='WDR8', description='WD repeat containing, antisense to TP73', other_designations='WD repeat domain 8', map_location='1p36.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3630767, end_position_on_the_genomic_accession=3650107, orientation='minus', exon_count=13, OMIM=606040, no_hdr0=''),
  100188865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188865, CurrentID=0, Status='live', Symbol='LMPH1B', Aliases='', description='Lymphedema, hereditary, IB', other_designations='', map_location='6q16.2-q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611944, no_hdr0=''),
  106480322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480322, CurrentID=0, Status='live', Symbol='BOLA2P2', Aliases='', description='bolA family member 2 pseudogene 2', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188867, CurrentID=0, Status='live', Symbol='HPC14', Aliases='', description='Prostate cancer, hereditary, 14', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611958, no_hdr0=''),
  49860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49860, CurrentID=0, Status='live', Symbol='CRNN', Aliases='C1orf10, DRC1, PDRC1, SEP53', description='cornulin', other_designations='53 kDa putative calcium-binding protein|53 kDa squamous epithelial-induced stress protein|58 kDa heat shock protein|squamous epithelial heat shock protein 53|tumor-related protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152409243, end_position_on_the_genomic_accession=152414274, orientation='minus', exon_count=3, OMIM=611312, no_hdr0=''),
  49861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=49861, CurrentID=0, Status='live', Symbol='CLDN20', Aliases='', description='claudin 20', other_designations='', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155264013, end_position_on_the_genomic_accession=155276548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100188870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188870, CurrentID=0, Status='live', Symbol='CELIAC8', Aliases='', description='Celiac disease, susceptibility to, 8', other_designations='', map_location='2q11-q12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612006, no_hdr0=''),
  100188871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188871, CurrentID=0, Status='live', Symbol='CELIAC9', Aliases='', description='Celiac disease, susceptibility to, 9', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612007, no_hdr0=''),
  100188872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188872, CurrentID=0, Status='live', Symbol='CELIAC10', Aliases='', description='Celiac disease, susceptibility to, 10', other_designations='', map_location='3q25-q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612008, no_hdr0=''),
  100188873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188873, CurrentID=0, Status='live', Symbol='CELIAC11', Aliases='', description='Celiac disease, susceptibility to, 11', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612009, no_hdr0=''),
  100188874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188874, CurrentID=0, Status='live', Symbol='CELIAC12', Aliases='', description='Celiac disease, susceptibility to, 12', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612010, no_hdr0=''),
  100188875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188875, CurrentID=0, Status='live', Symbol='CELIAC13', Aliases='', description='Celiac disease, susceptibility to, 13', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612011, no_hdr0=''),
  100188876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188876, CurrentID=0, Status='live', Symbol='IHPS3', Aliases='', description='Pyloric stenosis, infantile hypertrophic, 3', other_designations='', map_location='11q14-q22', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612017, no_hdr0=''),
  100188877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188877, CurrentID=0, Status='live', Symbol='CHDS9', Aliases='', description='Coronary heart disease, suscpetibility to, 9', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612030, no_hdr0=''),
  106480334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480334, CurrentID=0, Status='live', Symbol='BMS1P22', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 22', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480335, CurrentID=0, Status='live', Symbol='ACTP1', Aliases='', description='actin family pseudogene 1', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188880, CurrentID=0, Status='live', Symbol='HPRHP', Aliases='', description='Hypophosphatemic rickets and hyperparathyroidism', other_designations='', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612089, no_hdr0=''),
  100188881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188881, CurrentID=0, Status='live', Symbol='MFT2', Aliases='TEM', description='Trichoepithelioma, multiple familial, 2', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612099, no_hdr0=''),
  100188882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188882, CurrentID=0, Status='live', Symbol='BMND9', Aliases='', description='Bone mineral density QTL 9', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612110, no_hdr0=''),
  100188883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188883, CurrentID=0, Status='live', Symbol='BMND10', Aliases='', description='Bone mineral density QTL 10', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612113, no_hdr0=''),
  100188884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188884, CurrentID=0, Status='live', Symbol='BMND11', Aliases='', description='Bone mineral density QTL 11', other_designations='', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612114, no_hdr0=''),
  100188885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188885, CurrentID=0, Status='live', Symbol='ANIB7', Aliases='', description='Aneurysm, intracranial berry, 7', other_designations='', map_location='11q24-q25', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612161, no_hdr0=''),
  100188886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188886, CurrentID=0, Status='live', Symbol='ANIB8', Aliases='', description='Aneurysm, intracranial berry, 8', other_designations='', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612162, no_hdr0=''),
  100188887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188887, CurrentID=0, Status='live', Symbol='CRCS5', Aliases='', description='Colorectal cancer, susceptibility to, 5', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612230, no_hdr0=''),
  106480344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480344, CurrentID=0, Status='live', Symbol='SEPT14P17', Aliases='', description='septin 14 pseudogene 17', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480345, CurrentID=0, Status='live', Symbol='SEPT14P24', Aliases='', description='septin 14 pseudogene 24', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480346, CurrentID=0, Status='live', Symbol='GTF2IP11', Aliases='', description='general transcription factor IIi pseudogene 11', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188891, CurrentID=0, Status='live', Symbol='ZNF123P', Aliases='HZF-1, ZNF123', description='zinc finger protein 123, pseudogene', other_designations='zinc finger protein 123 (HZF-1)', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129295203, end_position_on_the_genomic_accession=129296615, orientation='minus', exon_count=0, OMIM=194630, no_hdr0=''),
  106480348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480348, CurrentID=0, Status='live', Symbol='NF1P10', Aliases='', description='neurofibromin 1 pseudogene 10', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188893, CurrentID=0, Status='live', Symbol='TOMM6', Aliases='OBTP, TOM6', description='translocase of outer mitochondrial membrane 6 homolog (yeast)', other_designations='over-expressed breast tumor protein|overexpressed breast tumor protein|translocase of outer membrane 6 kDa subunit homolog', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41787443, end_position_on_the_genomic_accession=41789896, orientation='plus', exon_count=3, OMIM=616168, no_hdr0=''),
  106480350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480350, CurrentID=0, Status='live', Symbol='DUX4L50', Aliases='', description='double homeobox 4 like 50, pseudogene', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480351, CurrentID=0, Status='live', Symbol='FRG2IP', Aliases='', description='FSHD region gene 2 family member I, pseudogene', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480352, CurrentID=0, Status='live', Symbol='FRG2LP', Aliases='', description='FSHD region gene 2 family member L, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480353, CurrentID=0, Status='live', Symbol='C1QL1P1', Aliases='', description='complement component 1, q subcomponent-like 1 pseudogene 1', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480354, CurrentID=0, Status='live', Symbol='DUX4L52', Aliases='', description='double homeobox 4 like 52, pseudogene', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480355, CurrentID=0, Status='live', Symbol='USP9YP21', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 21', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480356, CurrentID=0, Status='live', Symbol='SLC9A9-AS2', Aliases='', description='SLC9A9 antisense RNA 2', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480357, CurrentID=0, Status='live', Symbol='RNA5SP59', Aliases='RN5S59', description='RNA, 5S ribosomal pseudogene 59', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480358, CurrentID=0, Status='live', Symbol='RNA5SP311', Aliases='RN5S311, RN5S314, RNA5SP314', description='RNA, 5S ribosomal pseudogene 311', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480359, CurrentID=0, Status='live', Symbol='PTGES3P5', Aliases='', description='prostaglandin E synthase 3 (cytosolic) pseudogene 5', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480360, CurrentID=0, Status='live', Symbol='UBE2FP3', Aliases='', description='ubiquitin conjugating enzyme E2F (putative) pseudogene 3', other_designations='ubiquitin-conjugating enzyme E2F (putative) pseudogene 3', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480361, CurrentID=0, Status='live', Symbol='RNU7-96P', Aliases='', description='RNA, U7 small nuclear 96 pseudogene', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480362, CurrentID=0, Status='live', Symbol='RNU7-127P', Aliases='', description='RNA, U7 small nuclear 127 pseudogene', other_designations='', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480363, CurrentID=0, Status='live', Symbol='RNU7-186P', Aliases='', description='RNA, U7 small nuclear 186 pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480364, CurrentID=0, Status='live', Symbol='RN7SKP25', Aliases='', description='RNA, 7SK small nuclear pseudogene 25', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480365, CurrentID=0, Status='live', Symbol='RN7SKP94', Aliases='', description='RNA, 7SK small nuclear pseudogene 94', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480366, CurrentID=0, Status='live', Symbol='RN7SKP140', Aliases='', description='RNA, 7SK small nuclear pseudogene 140', other_designations='', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480367, CurrentID=0, Status='live', Symbol='RN7SKP165', Aliases='', description='RNA, 7SK small nuclear pseudogene 165', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480368, CurrentID=0, Status='live', Symbol='RN7SKP234', Aliases='', description='RNA, 7SK small nuclear pseudogene 234', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480369, CurrentID=0, Status='live', Symbol='RN7SKP281', Aliases='', description='RNA, 7SK small nuclear pseudogene 281', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480370, CurrentID=0, Status='live', Symbol='RN7SL19P', Aliases='', description='RNA, 7SL, cytoplasmic 19, pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480371, CurrentID=0, Status='live', Symbol='RN7SL110P', Aliases='', description='RNA, 7SL, cytoplasmic 110, pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480372, CurrentID=0, Status='live', Symbol='RN7SL169P', Aliases='', description='RNA, 7SL, cytoplasmic 169, pseudogene', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188917, CurrentID=0, Status='live', Symbol='ESP33', Aliases='', description='uncharacterized locus ESP33', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136249311, end_position_on_the_genomic_accession=136252057, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480374, CurrentID=0, Status='live', Symbol='RN7SL282P', Aliases='', description='RNA, 7SL, cytoplasmic 282, pseudogene', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480375, CurrentID=0, Status='live', Symbol='RN7SL335P', Aliases='', description='RNA, 7SL, cytoplasmic 335, pseudogene', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480376, CurrentID=0, Status='live', Symbol='RN7SL424P', Aliases='', description='RNA, 7SL, cytoplasmic 424, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480377, CurrentID=0, Status='live', Symbol='RN7SL451P', Aliases='', description='RNA, 7SL, cytoplasmic 451, pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480378, CurrentID=0, Status='live', Symbol='RN7SL543P', Aliases='', description='RNA, 7SL, cytoplasmic 543, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480379, CurrentID=0, Status='live', Symbol='RN7SL624P', Aliases='', description='RNA, 7SL, cytoplasmic 624, pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480380, CurrentID=0, Status='live', Symbol='RN7SL680P', Aliases='', description='RNA, 7SL, cytoplasmic 680, pseudogene', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480381, CurrentID=0, Status='live', Symbol='RN7SL713P', Aliases='', description='RNA, 7SL, cytoplasmic 713, pseudogene', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480382, CurrentID=0, Status='live', Symbol='RN7SL799P', Aliases='', description='RNA, 7SL, cytoplasmic 799, pseudogene', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480383, CurrentID=0, Status='live', Symbol='RN7SL854P', Aliases='', description='RNA, 7SL, cytoplasmic 854, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480384, CurrentID=0, Status='live', Symbol='RNU6ATAC25P', Aliases='', description='RNA, U6atac small nuclear 25, pseudogene', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480385, CurrentID=0, Status='live', Symbol='RNU4-34P', Aliases='', description='RNA, U4 small nuclear 34, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480386, CurrentID=0, Status='live', Symbol='RNU4-80P', Aliases='', description='RNA, U4 small nuclear 80, pseudogene', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480387, CurrentID=0, Status='live', Symbol='RNU6-161P', Aliases='', description='RNA, U6 small nuclear 161, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480388, CurrentID=0, Status='live', Symbol='RNU6-232P', Aliases='', description='RNA, U6 small nuclear 232, pseudogene', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480389, CurrentID=0, Status='live', Symbol='RNU6-299P', Aliases='', description='RNA, U6 small nuclear 299, pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480390, CurrentID=0, Status='live', Symbol='RNU6-325P', Aliases='', description='RNA, U6 small nuclear 325, pseudogene', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401537, CurrentID=0, Status='live', Symbol='RPSAP49', Aliases='RPSA_21_995', description='ribosomal protein SA pseudogene 49', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88324381, end_position_on_the_genomic_accession=88325410, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480392, CurrentID=0, Status='live', Symbol='RNU6-441P', Aliases='', description='RNA, U6 small nuclear 441, pseudogene', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480393, CurrentID=0, Status='live', Symbol='RNU6-487P', Aliases='', description='RNA, U6 small nuclear 487, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480394, CurrentID=0, Status='live', Symbol='RNU6-512P', Aliases='', description='RNA, U6 small nuclear 512, pseudogene', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480395, CurrentID=0, Status='live', Symbol='RNU6-580P', Aliases='', description='RNA, U6 small nuclear 580, pseudogene', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480396, CurrentID=0, Status='live', Symbol='RNU6-628P', Aliases='', description='RNA, U6 small nuclear 628, pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480397, CurrentID=0, Status='live', Symbol='RNU6-653P', Aliases='', description='RNA, U6 small nuclear 653, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480398, CurrentID=0, Status='live', Symbol='RNU6-700P', Aliases='', description='RNA, U6 small nuclear 700, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480399, CurrentID=0, Status='live', Symbol='RNU6-747P', Aliases='', description='RNA, U6 small nuclear 747, pseudogene', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480400, CurrentID=0, Status='live', Symbol='RNU6-794P', Aliases='', description='RNA, U6 small nuclear 794, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480401, CurrentID=0, Status='live', Symbol='RNU6-816P', Aliases='', description='RNA, U6 small nuclear 816, pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480402, CurrentID=0, Status='live', Symbol='RNU6-889P', Aliases='', description='RNA, U6 small nuclear 889, pseudogene', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188947, CurrentID=0, Status='live', Symbol='HECTD2-AS1', Aliases='', description='HECTD2 antisense RNA 1', other_designations='', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91306962, end_position_on_the_genomic_accession=91611460, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480404, CurrentID=0, Status='live', Symbol='RNU6-956P', Aliases='', description='RNA, U6 small nuclear 956, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188949, CurrentID=0, Status='live', Symbol='LINC00426', Aliases='', description='long intergenic non-protein coding RNA 426', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30340266, end_position_on_the_genomic_accession=30373914, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480406, CurrentID=0, Status='live', Symbol='RNU6-1075P', Aliases='', description='RNA, U6 small nuclear 1075, pseudogene', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480407, CurrentID=0, Status='live', Symbol='RNU6-1167P', Aliases='', description='RNA, U6 small nuclear 1167, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188952, CurrentID=0, Status='live', Symbol='IBD12', Aliases='', description='Inflammatory bowel disease 12', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612241, no_hdr0=''),
  401540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401540, CurrentID=0, Status='live', Symbol='OR7E109P', Aliases='OR7E113P, OST721', description='olfactory receptor family 7 subfamily E member 109 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 109 pseudogene|olfactory receptor, family 7, subfamily E, member 113 pseudogene', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90742502, end_position_on_the_genomic_accession=90743448, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188954, CurrentID=0, Status='live', Symbol='DNMBP-AS1', Aliases='NCRNA00093, bA287G8.2', description='DNMBP antisense RNA 1', other_designations='DNMBP antisense RNA 1 (non-protein coding)', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99927209, end_position_on_the_genomic_accession=99958998, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480411, CurrentID=0, Status='live', Symbol='RNU6-1330P', Aliases='', description='RNA, U6 small nuclear 1330, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480412, CurrentID=0, Status='live', Symbol='RNU1-89P', Aliases='', description='RNA, U1 small nuclear 89, pseudogene', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480413, CurrentID=0, Status='live', Symbol='RNU1-150P', Aliases='', description='RNA, U1 small nuclear 150, pseudogene', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480414, CurrentID=0, Status='live', Symbol='RNU2-26P', Aliases='', description='RNA, U2 small nuclear 26, pseudogene', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401541, CurrentID=0, Status='live', Symbol='CENPP', Aliases='CENP-P', description='centromere protein P', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92325457, end_position_on_the_genomic_accession=92615164, orientation='plus', exon_count=15, OMIM=611505, no_hdr0=''),
  106480416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480416, CurrentID=0, Status='live', Symbol='PRR20FP', Aliases='', description='proline rich 20F, pseudogene', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480417, CurrentID=0, Status='live', Symbol='SUMO2P15', Aliases='', description='SUMO2 pseudogene 15', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480418, CurrentID=0, Status='live', Symbol='BNIP3P16', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 16', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480419, CurrentID=0, Status='live', Symbol='B3GNTL1P2', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like 1 pseudogene 2', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480420, CurrentID=0, Status='live', Symbol='PHBP17', Aliases='', description='prohibitin pseudogene 17', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480421, CurrentID=0, Status='live', Symbol='DPY19L2P5', Aliases='', description='DPY19L2 pseudogene 5', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480422, CurrentID=0, Status='live', Symbol='KRTAP9-10P', Aliases='', description='keratin associated protein 9-10, pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480423, CurrentID=0, Status='live', Symbol='RNU6-13P', Aliases='RNU6-13', description='RNA, U6 small nuclear 13, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480424, CurrentID=0, Status='live', Symbol='USP9YP18', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 18', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480425, CurrentID=0, Status='live', Symbol='USP9YP29', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 29', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480426, CurrentID=0, Status='live', Symbol='GPM6BP2', Aliases='', description='glycoprotein M6B pseudogene 2', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480427, CurrentID=0, Status='live', Symbol='MTND6P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 5', other_designations='MT-ND6 pseudogene 5', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480428, CurrentID=0, Status='live', Symbol='MTND6P13', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 13', other_designations='MT-ND6 pseudogene 13', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480429, CurrentID=0, Status='live', Symbol='HLA-DQB1-AS1', Aliases='', description='HLA-DQB1 antisense RNA 1', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480430, CurrentID=0, Status='live', Symbol='CYP2C58P', Aliases='', description='cytochrome P450 family 2 subfamily C member 58, pseudogene', other_designations='cytochrome P450, family 2, subfamily C, polypeptide 58, pseudogene', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480431, CurrentID=0, Status='live', Symbol='ENO1-IT1', Aliases='', description='ENO1 intronic transcript 1', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480432, CurrentID=0, Status='live', Symbol='HM13-IT1', Aliases='', description='HM13 intronic transcript 1', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480433, CurrentID=0, Status='live', Symbol='MTND1P35', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 35', other_designations='MT-ND1 pseudogene 35', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480434, CurrentID=0, Status='live', Symbol='TUBB4BP4', Aliases='', description='tubulin beta 4B class IVb pseudogene 4', other_designations='tubulin, beta 4B class IVb pseudogene 4', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480435, CurrentID=0, Status='live', Symbol='RNU6-61P', Aliases='RNU6-61', description='RNA, U6 small nuclear 61, pseudogene', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480436, CurrentID=0, Status='live', Symbol='RNA5SP70', Aliases='RN5S70', description='RNA, 5S ribosomal pseudogene 70', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480437, CurrentID=0, Status='live', Symbol='RNA5SP149', Aliases='RN5S149', description='RNA, 5S ribosomal pseudogene 149', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480438, CurrentID=0, Status='live', Symbol='RNA5SP199', Aliases='RN5S199', description='RNA, 5S ribosomal pseudogene 199', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104472713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472713, CurrentID=0, Status='live', Symbol='PCAT29', Aliases='', description='prostate cancer associated transcript 29 (non-protein coding)', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69592200, end_position_on_the_genomic_accession=69695750, orientation='plus', exon_count=7, OMIM=616273, no_hdr0=''),
  106480440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480440, CurrentID=0, Status='live', Symbol='LINC00500', Aliases='', description='long intergenic non-protein coding RNA 500', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188985, CurrentID=0, Status='live', Symbol='TRG-GCC2-2', Aliases='TRNAG6', description='transfer RNA-Gly (GCC) 2-2', other_designations='tRNA-Gly-GCC-2-2|transfer RNA glycine 6 (anticodon GCC)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70779039, end_position_on_the_genomic_accession=70779109, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188986, CurrentID=0, Status='live', Symbol='TRH-GTG1-5', Aliases='TRNAH1', description='transfer RNA-His (GTG) 1-5', other_designations='tRNA-His-GTG-1-5|transfer RNA histidine 1 (anticodon GUG)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45198606, end_position_on_the_genomic_accession=45198677, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188987, CurrentID=0, Status='live', Symbol='TRH-GTG1-8', Aliases='TRNAH3', description='transfer RNA-His (GTG) 1-8', other_designations='tRNA-His-GTG-1-8|transfer RNA histidine 3 (anticodon GUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27158127, end_position_on_the_genomic_accession=27158198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100188988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188988, CurrentID=0, Status='live', Symbol='TRH-GTG1-6', Aliases='TRNAH2', description='transfer RNA-His (GTG) 1-6', other_designations='tRNA-His-GTG-1-6|transfer RNA histidine 2 (anticodon GUG)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45200413, end_position_on_the_genomic_accession=45200484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401546, CurrentID=0, Status='live', Symbol='C9orf152', Aliases='bA470J20.2', description='chromosome 9 open reading frame 152', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110199561, end_position_on_the_genomic_accession=110208133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100188990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188990, CurrentID=0, Status='live', Symbol='TRV-AAC1-2', Aliases='TRNAV5', description='transfer RNA-Val (AAC) 1-2', other_designations='tRNA-Val-AAC-1-2|transfer RNA valine 5 (anticodon AAC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181164154, end_position_on_the_genomic_accession=181164226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100188991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188991, CurrentID=0, Status='live', Symbol='TRH-GTG1-1', Aliases='TRNAH4', description='transfer RNA-His (GTG) 1-1', other_designations='tRNA-His-GTG-1-1|transfer RNA histidine 4 (anticodon GUG)', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146038044, end_position_on_the_genomic_accession=146038115, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100188992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188992, CurrentID=0, Status='live', Symbol='TRV-AAC2-1', Aliases='TRNAV10', description='transfer RNA-Val (AAC) 2-1', other_designations='tRNA-Val-AAC-2-1|transfer RNA valine 10 (anticodon AAC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181188416, end_position_on_the_genomic_accession=181188488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188993, CurrentID=0, Status='live', Symbol='TRV-TAC1-1', Aliases='TRNAV11', description='transfer RNA-Val (TAC) 1-1', other_designations='tRNA-Val-TAC-1-1|transfer RNA valine 11 (anticodon UAC)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59550629, end_position_on_the_genomic_accession=59550701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188994, CurrentID=0, Status='live', Symbol='TRD-GTC4-1', Aliases='TRNAD2', description='transfer RNA-Asp (GTC) 4-1', other_designations='tRNA-Asp-GTC-4-1|transfer RNA aspartic acid 2 (anticodon GUC)', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104472715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472715, CurrentID=0, Status='live', Symbol='SNHG14', Aliases='115HG, LNCAT, NCRNA00214, UBE3A-AS, UBE3A-AS1, UBE3AATS', description='small nucleolar RNA host gene 14', other_designations='UBE3A antisense RNA 1 (non-protein coding)|small nucleolar RNA host gene 14 (non-protein coding)', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616259, no_hdr0=''),
  100188996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188996, CurrentID=0, Status='live', Symbol='TRN-GTT14-1', Aliases='TRNAN4', description='transfer RNA-Asn (GTT) 14-1', other_designations='tRNA-Asn-GTT-14-1|transfer RNA asparagine 4 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100188997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188997, CurrentID=0, Status='live', Symbol='TRG-TCC4-1', Aliases='TRNAG7', description='transfer RNA-Gly (TCC) 4-1', other_designations='tRNA-Gly-TCC-4-1|transfer RNA glycine 7 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161440171, end_position_on_the_genomic_accession=161440242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100188998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188998, CurrentID=0, Status='live', Symbol='TRQ-CTG7-1', Aliases='TRNAQ3', description='transfer RNA-Gln (CTG) 7-1', other_designations='tRNA-Gln-CTG-7-1|transfer RNA glutamine 3 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148328812, end_position_on_the_genomic_accession=148328883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100188999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100188999, CurrentID=0, Status='live', Symbol='TRK-CTT15-1', Aliases='TRNAK5', description='transfer RNA-Lys (CTT) 15-1', other_designations='tRNA-Lys-CTT-15-1|transfer RNA lysine 5 (anticodon CUU)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189000, CurrentID=0, Status='live', Symbol='TRE-TTC12-1', Aliases='TRNAE3', description='transfer RNA-Glu (TTC) 12-1', other_designations='tRNA-Glu-TTC-12-1|transfer RNA glutamic acid 3 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401548, CurrentID=0, Status='live', Symbol='SNX30', Aliases='ATG24A', description='sorting nexin family member 30', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112749725, end_position_on_the_genomic_accession=112875851, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100189002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189002, CurrentID=0, Status='live', Symbol='TRS-GCT2-1', Aliases='TRNAS5', description='transfer RNA-Ser (GCT) 2-1', other_designations='tRNA-Ser-GCT-2-1|transfer RNA serine 5 (anticodon GCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27297996, end_position_on_the_genomic_accession=27298077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189003, CurrentID=0, Status='live', Symbol='TRL-CAA6-1', Aliases='TRNAL3', description='transfer RNA-Leu (CAA) 6-1', other_designations='tRNA-Leu-CAA-6-1|transfer RNA leucine 3 (anticodon CAA)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161611946, end_position_on_the_genomic_accession=161612029, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189004, CurrentID=0, Status='live', Symbol='TRL-AAG1-3', Aliases='TRNAL4', description='transfer RNA-Leu (AAG) 1-3', other_designations='tRNA-Leu-AAG-1-3|transfer RNA leucine 4 (anticodon AAG)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181174044, end_position_on_the_genomic_accession=181174125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189005, CurrentID=0, Status='live', Symbol='TRN-GTT7-1', Aliases='TRNAN5', description='transfer RNA-Asn (GTT) 7-1', other_designations='tRNA-Asn-GTT-7-1|transfer RNA asparagine 5 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148529257, end_position_on_the_genomic_accession=148529330, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189006, CurrentID=0, Status='live', Symbol='TRT-AGT1-3', Aliases='TRNAT4', description='transfer RNA-Thr (AGT) 1-3', other_designations='tRNA-Thr-AGT-1-3|transfer RNA threonine 4 (anticodon AGU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33177057, end_position_on_the_genomic_accession=33177130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8333, CurrentID=0, Status='live', Symbol='HIST1H2APS4', Aliases='H2A/k, H2AFK, H2AFKP, dJ34B20.2, pH2A/k', description='histone cluster 1, H2a, pseudogene 4', other_designations='H2A histone family, member K, pseudogene|histone 1, H2a, pseudogene 4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26272193, end_position_on_the_genomic_accession=26272540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189008, CurrentID=0, Status='live', Symbol='TRC-GCA11-1', Aliases='TRNAC2', description='transfer RNA-Cys (GCA) 11-1', other_designations='tRNA-Cys-GCA-11-1|transfer RNA cysteine 2 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149415138, end_position_on_the_genomic_accession=149415209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189009, CurrentID=0, Status='live', Symbol='TRL-CAG1-2', Aliases='TRNAL5', description='transfer RNA-Leu (CAG) 1-2', other_designations='tRNA-Leu-CAG-1-2|transfer RNA leucine 5 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161448951, end_position_on_the_genomic_accession=161449033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189010, CurrentID=0, Status='live', Symbol='TRQ-TTG10-1', Aliases='TRNAQ5', description='transfer RNA-Gln (TTG) 10-1', other_designations='tRNA-Gln-TTG-10-1|transfer RNA glutamine 5 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189011, CurrentID=0, Status='live', Symbol='TRG-TCC2-6', Aliases='TRNAG8', description='transfer RNA-Gly (TCC) 2-6', other_designations='tRNA-Gly-TCC-2-6|transfer RNA glycine 8 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161531113, end_position_on_the_genomic_accession=161531184, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480468, CurrentID=0, Status='live', Symbol='RN7SKP159', Aliases='', description='RNA, 7SK small nuclear pseudogene 159', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104472718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472718, CurrentID=0, Status='live', Symbol='LINC00508', Aliases='', description='long intergenic non-protein coding RNA 508', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127883989, end_position_on_the_genomic_accession=127983891, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100189014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189014, CurrentID=0, Status='live', Symbol='TRV-AAC5-1', Aliases='TRNAV12', description='transfer RNA-Val (AAC) 5-1', other_designations='tRNA-Val-AAC-5-1|transfer RNA valine 12 (anticodon AAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27235509, end_position_on_the_genomic_accession=27235581, orientation='plus', exon_count=0, OMIM=615307, no_hdr0=''),
  100189015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189015, CurrentID=0, Status='live', Symbol='TRX-CAT1-1', Aliases='TRNAM3', description='transfer RNA-iMet (CAT) 1-1', other_designations='tRNA-iMet-CAT-1-1|transfer RNA initiator methionine (CAT) 1-1|transfer RNA methionine 3 (anticodon CAU)', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153671250, end_position_on_the_genomic_accession=153671321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189016, CurrentID=0, Status='live', Symbol='TRQ-CTG15-1', Aliases='TRNAQ6', description='transfer RNA-Gln (CTG) 15-1', other_designations='tRNA-Gln-CTG-15-1|transfer RNA glutamine 6 (anticodon CUG)', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189017, CurrentID=0, Status='live', Symbol='TRE-CTC1-5', Aliases='TRNAE5', description='transfer RNA-Glu (CTC) 1-5', other_designations='tRNA-Glu-CTC-1-5|transfer RNA glutamic acid 5 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161469399, end_position_on_the_genomic_accession=161469470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189018, CurrentID=0, Status='live', Symbol='TRL-CAG1-6', Aliases='TRNAL6', description='transfer RNA-Leu (CAG) 1-6', other_designations='tRNA-Leu-CAG-1-6|transfer RNA leucine 6 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161530342, end_position_on_the_genomic_accession=161530424, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401551, CurrentID=0, Status='live', Symbol='WDR38', Aliases='', description='WD repeat domain 38', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124853347, end_position_on_the_genomic_accession=124857891, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100189020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189020, CurrentID=0, Status='live', Symbol='TRP-AGG1-1', Aliases='TRNAP5', description='transfer RNA-Pro (AGG) 1-1', other_designations='tRNA-Pro-AGG-1-1|transfer RNA proline 5 (anticodon AGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3191989, end_position_on_the_genomic_accession=3192060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189021, CurrentID=0, Status='live', Symbol='TRA-CGC3-1', Aliases='TRNAA4', description='transfer RNA-Ala (CGC) 3-1', other_designations='tRNA-Ala-CGC-3-1|transfer RNA alanine 4 (anticodon CGC)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156400769, end_position_on_the_genomic_accession=156400840, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480478, CurrentID=0, Status='live', Symbol='RN7SKP267', Aliases='', description='RNA, 7SK small nuclear pseudogene 267', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480479, CurrentID=0, Status='live', Symbol='RN7SKP279', Aliases='', description='RNA, 7SK small nuclear pseudogene 279', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189024, CurrentID=0, Status='live', Symbol='TRE-CTC1-6', Aliases='TRNAE6', description='transfer RNA-Glu (CTC) 1-6', other_designations='tRNA-Glu-CTC-1-6|transfer RNA glutamic acid 6 (anticodon CUC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28982199, end_position_on_the_genomic_accession=28982270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189025, CurrentID=0, Status='live', Symbol='TRR-TCG6-1', Aliases='TRNAR6', description='transfer RNA-Arg (TCG) 6-1', other_designations='tRNA-Arg-TCG-6-1|transfer RNA arginine 6 (anticodon UCG)', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110198523, end_position_on_the_genomic_accession=110198595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189026, CurrentID=0, Status='live', Symbol='TRQ-TTG6-1', Aliases='TRNAQ7', description='transfer RNA-Gln (TTG) 6-1', other_designations='tRNA-Gln-TTG-6-1|transfer RNA glutamine 7 (anticodon UUG)', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480483, CurrentID=0, Status='live', Symbol='RN7SL40P', Aliases='', description='RNA, 7SL, cytoplasmic 40, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480484, CurrentID=0, Status='live', Symbol='RN7SL57P', Aliases='', description='RNA, 7SL, cytoplasmic 57, pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189029, CurrentID=0, Status='live', Symbol='TRC-GCA22-1', Aliases='TRNAC3', description='transfer RNA-Cys (GCA) 22-1', other_designations='tRNA-Cys-GCA-22-1|transfer RNA cysteine 3 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149556711, end_position_on_the_genomic_accession=149556782, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189030, CurrentID=0, Status='live', Symbol='TRE-TTC2-1', Aliases='TRNAE7', description='transfer RNA-Glu (TTC) 2-1', other_designations='tRNA-Glu-TTC-2-1|transfer RNA glutamic acid 7 (anticodon UUC)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44917927, end_position_on_the_genomic_accession=44917998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189031, CurrentID=0, Status='live', Symbol='TRE-TTC5-1', Aliases='TRNAE8', description='transfer RNA-Glu (TTC) 5-1', other_designations='tRNA-Glu-TTC-5-1|transfer RNA glutamic acid 8 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189032, CurrentID=0, Status='live', Symbol='TRL-TAG3-1', Aliases='TRNAL7', description='transfer RNA-Leu (TAG) 3-1', other_designations='tRNA-Leu-TAG-3-1|transfer RNA leucine 7 (anticodon UAG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22195711, end_position_on_the_genomic_accession=22195792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189033, CurrentID=0, Status='live', Symbol='TRA-CGC4-1', Aliases='TRNAA5', description='transfer RNA-Ala (CGC) 4-1', other_designations='tRNA-Ala-CGC-4-1|transfer RNA alanine 5 (anticodon CGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28729315, end_position_on_the_genomic_accession=28729386, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189034, CurrentID=0, Status='live', Symbol='TRR-TCG5-1', Aliases='TRNAR7', description='transfer RNA-Arg (TCG) 5-1', other_designations='tRNA-Arg-TCG-5-1|transfer RNA arginine 7 (anticodon UCG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28543114, end_position_on_the_genomic_accession=28543186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189035, CurrentID=0, Status='live', Symbol='TRT-CGT1-1', Aliases='TRNAT5', description='transfer RNA-Thr (CGT) 1-1', other_designations='tRNA-Thr-CGT-1-1|transfer RNA threonine 5 (anticodon CGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28488993, end_position_on_the_genomic_accession=28489066, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189036, CurrentID=0, Status='live', Symbol='TRN-GTT2-4', Aliases='TRNAN7', description='transfer RNA-Asn (GTT) 2-4', other_designations='tRNA-Asn-GTT-2-4|transfer RNA asparagine 7 (anticodon GUU)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30673964, end_position_on_the_genomic_accession=30674037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189037, CurrentID=0, Status='live', Symbol='TRS-CGA1-1', Aliases='TRNAS6', description='transfer RNA-Ser (CGA) 1-1', other_designations='tRNA-Ser-CGA-1-1|transfer RNA serine 6 (anticodon CGA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8138881, end_position_on_the_genomic_accession=8138962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480494, CurrentID=0, Status='live', Symbol='RN7SL209P', Aliases='', description='RNA, 7SL, cytoplasmic 209, pseudogene', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189039, CurrentID=0, Status='live', Symbol='TRG-TCC2-5', Aliases='TRNAG9', description='transfer RNA-Gly (TCC) 2-5', other_designations='tRNA-Gly-TCC-2-5|transfer RNA glycine 9 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161469757, end_position_on_the_genomic_accession=161469828, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189040, CurrentID=0, Status='live', Symbol='TRX-CAT1-5', Aliases='TRNAM4', description='transfer RNA-iMet (CAT) 1-5', other_designations='tRNA-iMet-CAT-1-5|transfer RNA initiator methionine (CAT) 1-5|transfer RNA methionine 4 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26330301, end_position_on_the_genomic_accession=26330372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189041, CurrentID=0, Status='live', Symbol='TRS-CGA2-1', Aliases='TRNAS7', description='transfer RNA-Ser (CGA) 2-1', other_designations='tRNA-Ser-CGA-2-1|transfer RNA serine 7 (anticodon CGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27209849, end_position_on_the_genomic_accession=27209930, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189042, CurrentID=0, Status='live', Symbol='TRA-AGC6-1', Aliases='TRNAA6', description='transfer RNA-Ala (AGC) 6-1', other_designations='tRNA-Ala-AGC-6-1|transfer RNA alanine 6 (anticodon AGC)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28812072, end_position_on_the_genomic_accession=28812143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  139411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139411, CurrentID=0, Status='live', Symbol='PTCHD1', Aliases='', description='patched domain containing 1', other_designations='PTCHD1 isoform', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23334354, end_position_on_the_genomic_accession=23396801, orientation='plus', exon_count=4, OMIM=300828, no_hdr0=''),
  100189044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189044, CurrentID=0, Status='live', Symbol='TRE-CTC8-1', Aliases='TRNAE9', description='transfer RNA-Glu (CTC) 8-1', other_designations='tRNA-Glu-CTC-8-1|transfer RNA glutamic acid 9 (anticodon CUC)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189045, CurrentID=0, Status='live', Symbol='TRG-CCC5-1', Aliases='TRNAG10', description='transfer RNA-Gly (CCC) 5-1', other_designations='tRNA-Gly-CCC-5-1|transfer RNA glycine 10 (anticodon CCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16727285, end_position_on_the_genomic_accession=16727355, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189046, CurrentID=0, Status='live', Symbol='TRA-AGC4-1', Aliases='TRNAA7', description='transfer RNA-Ala (AGC) 4-1', other_designations='tRNA-Ala-AGC-4-1|transfer RNA alanine 7 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28658237, end_position_on_the_genomic_accession=28658308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189047, CurrentID=0, Status='live', Symbol='TRA-AGC5-1', Aliases='TRNAA8', description='transfer RNA-Ala (AGC) 5-1', other_designations='tRNA-Ala-AGC-5-1|transfer RNA alanine 8 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28710589, end_position_on_the_genomic_accession=28710660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189048, CurrentID=0, Status='live', Symbol='TRW-CCA3-3', Aliases='TRNAW2', description='transfer RNA-Trp (CCA) 3-3', other_designations='tRNA-Trp-CCA-3-3|transfer RNA tryptophan 2 (anticodon CCA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26331444, end_position_on_the_genomic_accession=26331515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189049, CurrentID=0, Status='live', Symbol='TRC-GCA18-1', Aliases='TRNAC4', description='transfer RNA-Cys (GCA) 18-1', other_designations='tRNA-Cys-GCA-18-1|transfer RNA cysteine 4 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149375759, end_position_on_the_genomic_accession=149375830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189050, CurrentID=0, Status='live', Symbol='TRI-GAT1-2', Aliases='TRNAI3', description='transfer RNA-Ile (GAT) 1-2', other_designations='tRNA-Ile-GAT-1-2|transfer RNA isoleucine 3 (anticodon GAU)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3876801, end_position_on_the_genomic_accession=3876874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189051, CurrentID=0, Status='live', Symbol='TRK-TTT12-1', Aliases='TRNAK7', description='transfer RNA-Lys (TTT) 12-1', other_designations='tRNA-Lys-TTT-12-1|transfer RNA lysine 7 (anticodon UUU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189052, CurrentID=0, Status='live', Symbol='TRC-GCA2-1', Aliases='TRNAC5', description='transfer RNA-Cys (GCA) 2-1', other_designations='tRNA-Cys-GCA-2-1|transfer RNA cysteine 5 (anticodon GCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38867645, end_position_on_the_genomic_accession=38867716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189053, CurrentID=0, Status='live', Symbol='TRK-TTT4-1', Aliases='TRNAK8', description='transfer RNA-Lys (TTT) 4-1', other_designations='tRNA-Lys-TTT-4-1|transfer RNA lysine 8 (anticodon UUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27591814, end_position_on_the_genomic_accession=27591886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189054, CurrentID=0, Status='live', Symbol='TRE-TTC13-1', Aliases='TRNAE10', description='transfer RNA-Glu (TTC) 13-1', other_designations='tRNA-Glu-TTC-13-1|transfer RNA glutamic acid 10 (anticodon UUC)', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189055, CurrentID=0, Status='live', Symbol='NMTRS-TGA1-1', Aliases='TRNAS8', description='nuclear-encoded mitochondrial transfer RNA-Ser (TGA) 1-1', other_designations='nmt-tRNA-Ser-TGA-1-1|transfer RNA serine 8 (anticodon UGA)', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189056, CurrentID=0, Status='live', Symbol='TRQ-CTG4-1', Aliases='TRNAQ8', description='transfer RNA-Gln (CTG) 4-1', other_designations='tRNA-Gln-CTG-4-1|transfer RNA glutamine 8 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148265108, end_position_on_the_genomic_accession=148265179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480513, CurrentID=0, Status='live', Symbol='RN7SL482P', Aliases='', description='RNA, 7SL, cytoplasmic 482, pseudogene', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189058, CurrentID=0, Status='live', Symbol='NMTRQ-TTG14-1', Aliases='TRNAQ9', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 14-1', other_designations='nmt-tRNA-Gln-TTG-14-1|transfer RNA glutamine 9 (anticodon UUG)|transfer RNA-Gln (TTG) 14-1', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189059, CurrentID=0, Status='live', Symbol='TRN-GTT8-1', Aliases='TRNAN8', description='transfer RNA-Asn (GTT) 8-1', other_designations='tRNA-Asn-GTT-8-1|transfer RNA asparagine 8 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149740248, end_position_on_the_genomic_accession=149740321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189060, CurrentID=0, Status='live', Symbol='TRA-AGC13-2', Aliases='TRNAA9', description='transfer RNA-Ala (AGC) 13-2', other_designations='tRNA-Ala-AGC-13-2|transfer RNA alanine 9 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57838350, end_position_on_the_genomic_accession=57838422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189061, CurrentID=0, Status='live', Symbol='TRD-GTC5-1', Aliases='TRNAD3', description='transfer RNA-Asp (GTC) 5-1', other_designations='tRNA-Asp-GTC-5-1|transfer RNA aspartic acid 3 (anticodon GUC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189062, CurrentID=0, Status='live', Symbol='TRK-CTT4-1', Aliases='TRNAK9', description='transfer RNA-Lys (CTT) 4-1', other_designations='tRNA-Lys-CTT-4-1|transfer RNA lysine 9 (anticodon CUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3191501, end_position_on_the_genomic_accession=3191573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189063, CurrentID=0, Status='live', Symbol='TRE-TTC2-2', Aliases='TRNAE11', description='transfer RNA-Glu (TTC) 2-2', other_designations='tRNA-Glu-TTC-2-2|transfer RNA glutamic acid 11 (anticodon UUC)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26082234, end_position_on_the_genomic_accession=26082305, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189064, CurrentID=0, Status='live', Symbol='TRG-TCC2-1', Aliases='TRNAG11', description='transfer RNA-Gly (TCC) 2-1', other_designations='tRNA-Gly-TCC-2-1|transfer RNA glycine 11 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146037061, end_position_on_the_genomic_accession=146037132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189065, CurrentID=0, Status='live', Symbol='TRH-GTG1-7', Aliases='TRNAH6', description='transfer RNA-His (GTG) 1-7', other_designations='tRNA-His-GTG-1-7|transfer RNA histidine 6 (anticodon GUG)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45201151, end_position_on_the_genomic_accession=45201222, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189066, CurrentID=0, Status='live', Symbol='TRA-TGC5-1', Aliases='TRNAA10', description='transfer RNA-Ala (TGC) 5-1', other_designations='tRNA-Ala-TGC-5-1|transfer RNA alanine 10 (anticodon UGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28817235, end_position_on_the_genomic_accession=28817306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189230, CurrentID=0, Status='live', Symbol='TRG-GCC1-3', Aliases='TRNAG21', description='transfer RNA-Gly (GCC) 1-3', other_designations='tRNA-Gly-GCC-1-3|transfer RNA glycine 21 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161458108, end_position_on_the_genomic_accession=161458178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189068, CurrentID=0, Status='live', Symbol='TRE-TTC8-1', Aliases='TRNAE12', description='transfer RNA-Glu (TTC) 8-1', other_designations='tRNA-Glu-TTC-8-1|transfer RNA glutamic acid 12 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189069, CurrentID=0, Status='live', Symbol='TRN-GTT16-1', Aliases='TRNAN9', description='transfer RNA-Asn (GTT) 16-1', other_designations='tRNA-Asn-GTT-16-1|transfer RNA asparagine 9 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189070, CurrentID=0, Status='live', Symbol='TRS-AGA4-1', Aliases='TRNAS9', description='transfer RNA-Ser (AGA) 4-1', other_designations='tRNA-Ser-AGA-4-1|transfer RNA serine 9 (anticodon AGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27553413, end_position_on_the_genomic_accession=27553494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189071, CurrentID=0, Status='live', Symbol='TRI-AAT2-1', Aliases='TRNAI4', description='transfer RNA-Ile (AAT) 2-1', other_designations='tRNA-Ile-AAT-2-1|transfer RNA isoleucine 4 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27688188, end_position_on_the_genomic_accession=27688261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189072, CurrentID=0, Status='live', Symbol='TRP-AGG2-2', Aliases='TRNAP6', description='transfer RNA-Pro (AGG) 2-2', other_designations='tRNA-Pro-AGG-2-2|transfer RNA proline 6 (anticodon AGG)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76235513, end_position_on_the_genomic_accession=76235584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189073, CurrentID=0, Status='live', Symbol='TRV-CAC12-1', Aliases='TRNAV14', description='transfer RNA-Val (CAC) 12-1', other_designations='tRNA-Val-CAC-12-1|transfer RNA valine 14 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189074, CurrentID=0, Status='live', Symbol='TRQ-TTG3-3', Aliases='TRNAQ10', description='transfer RNA-Gln (TTG) 3-3', other_designations='tRNA-Gln-TTG-3-3|transfer RNA glutamine 10 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27795861, end_position_on_the_genomic_accession=27795932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189075, CurrentID=0, Status='live', Symbol='TRC-GCA12-1', Aliases='TRNAC6', description='transfer RNA-Cys (GCA) 12-1', other_designations='tRNA-Cys-GCA-12-1|transfer RNA cysteine 6 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149646955, end_position_on_the_genomic_accession=149647026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480532, CurrentID=0, Status='live', Symbol='RN7SL785P', Aliases='', description='RNA, 7SL, cytoplasmic 785, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189077, CurrentID=0, Status='live', Symbol='TRS-AGA2-3', Aliases='TRNAS10', description='transfer RNA-Ser (AGA) 2-3', other_designations='tRNA-Ser-AGA-2-3|transfer RNA serine 10 (anticodon AGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27478812, end_position_on_the_genomic_accession=27478893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189078, CurrentID=0, Status='live', Symbol='TRR-CCT1-1', Aliases='TRNAR8', description='transfer RNA-Arg (CCT) 1-1', other_designations='tRNA-Arg-CCT-1-1|transfer RNA arginine 8 (anticodon CCU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75033906, end_position_on_the_genomic_accession=75033978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189079, CurrentID=0, Status='live', Symbol='TRN-GTT16-3', Aliases='TRNAN10', description='transfer RNA-Asn (GTT) 16-3', other_designations='tRNA-Asn-GTT-16-3|transfer RNA asparagine 10 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189080, CurrentID=0, Status='live', Symbol='TRSUP-TTA1-1', Aliases='TRNASUP1', description='transfer RNA suppressor (TTA) 1-1', other_designations='tRNA-Sup-TTA-1-1|transfer RNA suppressor 1 (anticodon UUA)', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60786232, end_position_on_the_genomic_accession=60786302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189081, CurrentID=0, Status='live', Symbol='TRG-GCC6-1', Aliases='TRNAG12', description='transfer RNA-Gly (GCC) 6-1', other_designations='tRNA-Gly-GCC-6-1|transfer RNA glycine 12 (anticodon GCC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189082, CurrentID=0, Status='live', Symbol='TRA-TGC6-1', Aliases='TRNAA11', description='transfer RNA-Ala (TGC) 6-1', other_designations='tRNA-Ala-TGC-6-1|transfer RNA alanine 11 (anticodon UGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28758364, end_position_on_the_genomic_accession=28758435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189083, CurrentID=0, Status='live', Symbol='TRN-GTT2-3', Aliases='TRNAN11', description='transfer RNA-Asn (GTT) 2-3', other_designations='tRNA-Asn-GTT-2-3|transfer RNA asparagine 11 (anticodon GUU)', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22229509, end_position_on_the_genomic_accession=22229582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189084, CurrentID=0, Status='live', Symbol='TRR-CCG1-3', Aliases='TRNAR9', description='transfer RNA-Arg (CCG) 1-3', other_designations='tRNA-Arg-CCG-1-3|transfer RNA arginine 9 (anticodon CCG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28881388, end_position_on_the_genomic_accession=28881460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401562, CurrentID=0, Status='live', Symbol='LCNL1', Aliases='', description='lipocalin-like 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136982993, end_position_on_the_genomic_accession=136985758, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189086, CurrentID=0, Status='live', Symbol='TRT-CGT2-1', Aliases='TRNAT6', description='transfer RNA-Thr (CGT) 2-1', other_designations='tRNA-Thr-CGT-2-1|transfer RNA threonine 6 (anticodon CGU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14285893, end_position_on_the_genomic_accession=14285964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189087, CurrentID=0, Status='live', Symbol='TRD-GTC7-1', Aliases='TRNAD4', description='transfer RNA-Asp (GTC) 7-1', other_designations='tRNA-Asp-GTC-7-1|transfer RNA aspartic acid 4 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189088, CurrentID=0, Status='live', Symbol='TRP-TGG3-1', Aliases='TRNAP7', description='transfer RNA-Pro (TGG) 3-1', other_designations='tRNA-Pro-TGG-3-1|transfer RNA proline 7 (anticodon UGG)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20684016, end_position_on_the_genomic_accession=20684087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189089, CurrentID=0, Status='live', Symbol='TRL-CAG2-2', Aliases='TRNAL11', description='transfer RNA-Leu (CAG) 2-2', other_designations='tRNA-Leu-CAG-2-2|transfer RNA leucine 11 (anticodon CAG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57300480, end_position_on_the_genomic_accession=57300562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189090, CurrentID=0, Status='live', Symbol='TRC-GCA4-1', Aliases='TRNAC7', description='transfer RNA-Cys (GCA) 4-1', other_designations='tRNA-Cys-GCA-4-1|transfer RNA cysteine 7 (anticodon GCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38869292, end_position_on_the_genomic_accession=38869363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401563, CurrentID=0, Status='live', Symbol='C9orf139', Aliases='', description='chromosome 9 open reading frame 139', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137027464, end_position_on_the_genomic_accession=137037767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100189092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189092, CurrentID=0, Status='live', Symbol='TRA-AGC7-1', Aliases='TRNAA12', description='transfer RNA-Ala (AGC) 7-1', other_designations='tRNA-Ala-AGC-7-1|transfer RNA alanine 12 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28719704, end_position_on_the_genomic_accession=28719775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189093, CurrentID=0, Status='live', Symbol='TRA-AGC2-1', Aliases='TRNAA13', description='transfer RNA-Ala (AGC) 2-1', other_designations='tRNA-Ala-AGC-2-1|transfer RNA alanine 13 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28838444, end_position_on_the_genomic_accession=28838515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189094, CurrentID=0, Status='live', Symbol='TRX-CAT1-4', Aliases='TRNAM6', description='transfer RNA-iMet (CAT) 1-4', other_designations='tRNA-iMet-CAT-1-4|transfer RNA initiator methionine (CAT) 1-4|transfer RNA methionine 6 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26313124, end_position_on_the_genomic_accession=26313195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189095, CurrentID=0, Status='live', Symbol='TRI-AAT8-1', Aliases='TRNAI5', description='transfer RNA-Ile (AAT) 8-1', other_designations='tRNA-Ile-AAT-8-1|transfer RNA isoleucine 5 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27668583, end_position_on_the_genomic_accession=27668656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189096, CurrentID=0, Status='live', Symbol='TRN-GTT17-1', Aliases='TRNAN12', description='transfer RNA-Asn (GTT) 17-1', other_designations='tRNA-Asn-GTT-17-1|transfer RNA asparagine 12 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189097, CurrentID=0, Status='live', Symbol='TRC-GCA3-1', Aliases='TRNAC8', description='transfer RNA-Cys (GCA) 3-1', other_designations='tRNA-Cys-GCA-3-1|transfer RNA cysteine 8 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149589073, end_position_on_the_genomic_accession=149589144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189098, CurrentID=0, Status='live', Symbol='TRK-TTT8-1', Aliases='TRNAK11', description='transfer RNA-Lys (TTT) 8-1', other_designations='tRNA-Lys-TTT-8-1|transfer RNA lysine 11 (anticodon UUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205474143, end_position_on_the_genomic_accession=205474215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189099, CurrentID=0, Status='live', Symbol='TRR-ACG2-3', Aliases='TRNAR10', description='transfer RNA-Arg (ACG) 2-3', other_designations='tRNA-Arg-ACG-2-3|transfer RNA arginine 10 (anticodon ACG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27215173, end_position_on_the_genomic_accession=27215245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189100, CurrentID=0, Status='live', Symbol='TRM-CAT4-1', Aliases='TRNAM7', description='transfer RNA-Met (CAT) 4-1', other_designations='tRNA-Met-CAT-4-1|transfer RNA methionine 7 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189101, CurrentID=0, Status='live', Symbol='TRV-CAC7-1', Aliases='TRNAV15', description='transfer RNA-Val (CAC) 7-1', other_designations='tRNA-Val-CAC-7-1|transfer RNA valine 15 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27728548, end_position_on_the_genomic_accession=27728620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480558, CurrentID=0, Status='live', Symbol='RNU6-175P', Aliases='', description='RNA, U6 small nuclear 175, pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401565, CurrentID=0, Status='live', Symbol='FAM166A', Aliases='HSD46', description='family with sequence similarity 166 member A', other_designations='family with sequence similarity 166, member A', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137243583, end_position_on_the_genomic_accession=137251193, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100189104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189104, CurrentID=0, Status='live', Symbol='TRK-TTT11-1', Aliases='TRNAK12', description='transfer RNA-Lys (TTT) 11-1', other_designations='tRNA-Lys-TTT-11-1|transfer RNA lysine 12 (anticodon UUU)', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189105, CurrentID=0, Status='live', Symbol='TRC-GCA2-2', Aliases='TRNAC9', description='transfer RNA-Cys (GCA) 2-2', other_designations='tRNA-Cys-GCA-2-2|transfer RNA cysteine 9 (anticodon GCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39153734, end_position_on_the_genomic_accession=39153805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189106, CurrentID=0, Status='live', Symbol='TRA-AGC12-3', Aliases='TRNAA14', description='transfer RNA-Ala (AGC) 12-3', other_designations='tRNA-Ala-AGC-12-3|transfer RNA alanine 14 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57856401, end_position_on_the_genomic_accession=57856473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189107, CurrentID=0, Status='live', Symbol='TRK-CTT1-1', Aliases='TRNAK13', description='transfer RNA-Lys (CTT) 1-1', other_designations='tRNA-Lys-CTT-1-1|transfer RNA lysine 13 (anticodon CUU)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58239895, end_position_on_the_genomic_accession=58239967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189108, CurrentID=0, Status='live', Symbol='TRN-GTT11-1', Aliases='TRNAN13', description='transfer RNA-Asn (GTT) 11-1', other_designations='tRNA-Asn-GTT-11-1|transfer RNA asparagine 13 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=144301611, end_position_on_the_genomic_accession=144301684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139422, CurrentID=0, Status='live', Symbol='MAGEB10', Aliases='', description='MAGE family member B10', other_designations='MAGE family testis and tumor-specific protein|MAGE-B10 antigen|melanoma antigen family B, 10|melanoma antigen family B10', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27807450, end_position_on_the_genomic_accession=27823532, orientation='plus', exon_count=4, OMIM=300761, no_hdr0=''),
  100189110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189110, CurrentID=0, Status='live', Symbol='TRK-CTT2-5', Aliases='TRNAK14', description='transfer RNA-Lys (CTT) 2-5', other_designations='tRNA-Lys-CTT-2-5|transfer RNA lysine 14 (anticodon CUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26556546, end_position_on_the_genomic_accession=26556618, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285362, CurrentID=0, Status='live', Symbol='SUMF1', Aliases='AAPA3037, FGE, UNQ3037', description='sulfatase modifying factor 1', other_designations='C-alpha-formylglycine-generating enzyme 1|FGly-generating enzyme', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4065934, end_position_on_the_genomic_accession=4467282, orientation='minus', exon_count=14, OMIM=607939, no_hdr0=''),
  100189112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189112, CurrentID=0, Status='live', Symbol='TRW-CCA3-1', Aliases='TRNAW3', description='transfer RNA-Trp (CCA) 3-1', other_designations='tRNA-Trp-CCA-3-1|transfer RNA tryptophan 3 (anticodon CCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8186358, end_position_on_the_genomic_accession=8186429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189113, CurrentID=0, Status='live', Symbol='TRE-CTC17-1', Aliases='TRNAE14', description='transfer RNA-Glu (CTC) 17-1', other_designations='tRNA-Glu-CTC-17-1|transfer RNA glutamic acid 14 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189114, CurrentID=0, Status='live', Symbol='TRA-TGC2-1', Aliases='TRNAA15', description='transfer RNA-Ala (TGC) 2-1', other_designations='tRNA-Ala-TGC-2-1|transfer RNA alanine 15 (anticodon UGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28643445, end_position_on_the_genomic_accession=28643516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480571, CurrentID=0, Status='live', Symbol='RNU6-333P', Aliases='', description='RNA, U6 small nuclear 333, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189116, CurrentID=0, Status='live', Symbol='TRA-AGC20-1', Aliases='TRNAA16', description='transfer RNA-Ala (AGC) 20-1', other_designations='tRNA-Ala-AGC-20-1|transfer RNA alanine 16 (anticodon AGC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189117, CurrentID=0, Status='live', Symbol='TRA-AGC15-1', Aliases='TRNAA17', description='transfer RNA-Ala (AGC) 15-1', other_designations='tRNA-Ala-AGC-15-1|transfer RNA alanine 17 (anticodon AGC)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88979098, end_position_on_the_genomic_accession=88979170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189118, CurrentID=0, Status='live', Symbol='TRL-AAG4-1', Aliases='TRNAL12', description='transfer RNA-Leu (AAG) 4-1', other_designations='tRNA-Leu-AAG-4-1|transfer RNA leucine 12 (anticodon AAG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28478623, end_position_on_the_genomic_accession=28478704, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189119, CurrentID=0, Status='live', Symbol='TRK-CTT3-1', Aliases='TRNAK15', description='transfer RNA-Lys (CTT) 3-1', other_designations='tRNA-Lys-CTT-3-1|transfer RNA lysine 15 (anticodon CUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3157405, end_position_on_the_genomic_accession=3157477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189120, CurrentID=0, Status='live', Symbol='TRD-GTC2-5', Aliases='TRNAD5', description='transfer RNA-Asp (GTC) 2-5', other_designations='tRNA-Asp-GTC-2-5|transfer RNA aspartic acid 5 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161470415, end_position_on_the_genomic_accession=161470486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189121, CurrentID=0, Status='live', Symbol='TRQ-CTG1-5', Aliases='TRNAQ11', description='transfer RNA-Gln (CTG) 1-5', other_designations='tRNA-Gln-CTG-1-5|transfer RNA glutamine 11 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28941601, end_position_on_the_genomic_accession=28941672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189122, CurrentID=0, Status='live', Symbol='TRK-TTT3-1', Aliases='TRNAK16', description='transfer RNA-Lys (TTT) 3-1', other_designations='tRNA-Lys-TTT-3-1|transfer RNA lysine 16 (anticodon UUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204506527, end_position_on_the_genomic_accession=204506599, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189123, CurrentID=0, Status='live', Symbol='TRW-CCA1-1', Aliases='TRNAW4', description='transfer RNA-Trp (CCA) 1-1', other_designations='tRNA-Trp-CCA-1-1|transfer RNA tryptophan 4 (anticodon CCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8220869, end_position_on_the_genomic_accession=8220940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189124, CurrentID=0, Status='live', Symbol='TRP-CGG1-3', Aliases='TRNAP8', description='transfer RNA-Pro (CGG) 1-3', other_designations='tRNA-Pro-CGG-1-3|transfer RNA proline 8 (anticodon CGG)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8222833, end_position_on_the_genomic_accession=8222904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189125, CurrentID=0, Status='live', Symbol='TRK-TTT1-1', Aliases='TRNAK17', description='transfer RNA-Lys (TTT) 1-1', other_designations='tRNA-Lys-TTT-1-1|transfer RNA lysine 17 (anticodon UUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73478317, end_position_on_the_genomic_accession=73478389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189126, CurrentID=0, Status='live', Symbol='TRN-GTT1-1', Aliases='TRNAN14', description='transfer RNA-Asn (GTT) 1-1', other_designations='tRNA-Asn-GTT-1-1|transfer RNA asparagine 14 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161540241, end_position_on_the_genomic_accession=161540314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139425, CurrentID=0, Status='live', Symbol='DCAF8L1', Aliases='WDR42B', description='DDB1 and CUL4 associated factor 8-like 1', other_designations='WD repeat domain 42B|WD repeat-containing protein 42B', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27977993, end_position_on_the_genomic_accession=27981449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574408, CurrentID=0, Status='live', Symbol='MIR329-1', Aliases='MIRN329-1, mir-329-1', description='microRNA 329-1', other_designations='hsa-mir-329-1', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101026785, end_position_on_the_genomic_accession=101026864, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574409, CurrentID=0, Status='live', Symbol='MIR329-2', Aliases='MIRN329-2, mir-329-2', description='microRNA 329-2', other_designations='hsa-mir-329-2', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101027100, end_position_on_the_genomic_accession=101027183, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574410, CurrentID=0, Status='live', Symbol='MIR323B', Aliases='MIR453, MIRN453, hsa-mir-453, mir-323b', description='microRNA 323b', other_designations='hsa-mir-323b|microRNA 453', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101056219, end_position_on_the_genomic_accession=101056300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574411, CurrentID=0, Status='live', Symbol='MIR451A', Aliases='MIR451, MIRN451, hsa-mir-451, hsa-mir-451a, mir-451a', description='microRNA 451a', other_designations='microRNA 451', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28861369, end_position_on_the_genomic_accession=28861440, orientation='minus', exon_count=1, OMIM=612071, no_hdr0=''),
  574412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574412, CurrentID=0, Status='live', Symbol='MIR452', Aliases='MIRN452, hsa-mir-452, mir-452', description='microRNA 452', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151959628, end_position_on_the_genomic_accession=151959712, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574413, CurrentID=0, Status='live', Symbol='MIR409', Aliases='MIRN409, hsa-mir-409, mir-409', description='microRNA 409', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101065300, end_position_on_the_genomic_accession=101065378, orientation='plus', exon_count=1, OMIM=614057, no_hdr0=''),
  574414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574414, CurrentID=0, Status='live', Symbol='PRR9', Aliases='', description='proline rich 9', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153205162, end_position_on_the_genomic_accession=153219317, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100189135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189135, CurrentID=0, Status='live', Symbol='TRM-CAT2-1', Aliases='TRNAM9', description='transfer RNA-Met (CAT) 2-1', other_designations='tRNA-Met-CAT-2-1|transfer RNA methionine 9 (anticodon CAU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71426493, end_position_on_the_genomic_accession=71426565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189136, CurrentID=0, Status='live', Symbol='TRP-AGG2-5', Aliases='TRNAP9', description='transfer RNA-Pro (AGG) 2-5', other_designations='tRNA-Pro-AGG-2-5|transfer RNA proline 9 (anticodon AGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3182635, end_position_on_the_genomic_accession=3182706, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189137, CurrentID=0, Status='live', Symbol='TRF-GAA1-5', Aliases='TRNAF3', description='transfer RNA-Phe (GAA) 1-5', other_designations='tRNA-Phe-GAA-1-5|transfer RNA phenylalanine 3 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28790722, end_position_on_the_genomic_accession=28790794, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189138, CurrentID=0, Status='live', Symbol='TRQ-TTG2-1', Aliases='TRNAQ12', description='transfer RNA-Gln (TTG) 2-1', other_designations='tRNA-Gln-TTG-2-1|transfer RNA glutamine 12 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28589379, end_position_on_the_genomic_accession=28589450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189139, CurrentID=0, Status='live', Symbol='TRL-CAG1-5', Aliases='TRNAL14', description='transfer RNA-Leu (CAG) 1-5', other_designations='tRNA-Leu-CAG-1-5|transfer RNA leucine 14 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161471123, end_position_on_the_genomic_accession=161471205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189140, CurrentID=0, Status='live', Symbol='TRQ-CTG3-2', Aliases='TRNAQ13', description='transfer RNA-Gln (CTG) 3-2', other_designations='tRNA-Gln-CTG-3-2|transfer RNA glutamine 13 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148032790, end_position_on_the_genomic_accession=148032861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189141, CurrentID=0, Status='live', Symbol='TRG-TCC2-3', Aliases='TRNAG13', description='transfer RNA-Gly (TCC) 2-3', other_designations='tRNA-Gly-TCC-2-3|transfer RNA glycine 13 (anticodon UCC)', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161454966, end_position_on_the_genomic_accession=161455037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189142, CurrentID=0, Status='live', Symbol='TRV-CAC8-1', Aliases='TRNAV16', description='transfer RNA-Val (CAC) 8-1', other_designations='tRNA-Val-CAC-8-1|transfer RNA valine 16 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143800041, end_position_on_the_genomic_accession=143800111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189143, CurrentID=0, Status='live', Symbol='TRT-AGT6-1', Aliases='TRNAT7', description='transfer RNA-Thr (AGT) 6-1', other_designations='tRNA-Thr-AGT-6-1|transfer RNA threonine 7 (anticodon AGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27162271, end_position_on_the_genomic_accession=27162344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189144, CurrentID=0, Status='live', Symbol='TRC-GCA1-1', Aliases='TRNAC10', description='transfer RNA-Cys (GCA) 1-1', other_designations='tRNA-Cys-GCA-1-1|transfer RNA cysteine 10 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149310190, end_position_on_the_genomic_accession=149310261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189145, CurrentID=0, Status='live', Symbol='TRF-GAA5-1', Aliases='TRNAF4', description='transfer RNA-Phe (GAA) 5-1', other_designations='tRNA-Phe-GAA-5-1|transfer RNA phenylalanine 4 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28764387, end_position_on_the_genomic_accession=28764459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189146, CurrentID=0, Status='live', Symbol='TRC-GCA16-1', Aliases='TRNAC11', description='transfer RNA-Cys (GCA) 16-1', other_designations='tRNA-Cys-GCA-16-1|transfer RNA cysteine 11 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149546540, end_position_on_the_genomic_accession=149546611, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480603, CurrentID=0, Status='live', Symbol='RNU6-714P', Aliases='', description='RNA, U6 small nuclear 714, pseudogene', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189148, CurrentID=0, Status='live', Symbol='TRI-AAT7-1', Aliases='TRNAI8', description='transfer RNA-Ile (AAT) 7-1', other_designations='tRNA-Ile-AAT-7-1|transfer RNA isoleucine 8 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189149, CurrentID=0, Status='live', Symbol='TRA-TGC1-1', Aliases='TRNAA19', description='transfer RNA-Ala (TGC) 1-1', other_designations='tRNA-Ala-TGC-1-1|transfer RNA alanine 19 (anticodon UGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28789770, end_position_on_the_genomic_accession=28789841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189150, CurrentID=0, Status='live', Symbol='TRD-GTC6-1', Aliases='TRNAD7', description='transfer RNA-Asp (GTC) 6-1', other_designations='tRNA-Asp-GTC-6-1|transfer RNA aspartic acid 7 (anticodon GUC)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574431, CurrentID=0, Status='live', Symbol='C1orf147', Aliases='', description='chromosome 1 open reading frame 147', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=206664449, end_position_on_the_genomic_accession=206671061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100189152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189152, CurrentID=0, Status='live', Symbol='TRS-CGA4-1', Aliases='TRNAS11', description='transfer RNA-Ser (CGA) 4-1', other_designations='tRNA-Ser-CGA-4-1|transfer RNA serine 11 (anticodon CGA)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56190364, end_position_on_the_genomic_accession=56190445, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  574433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574433, CurrentID=0, Status='live', Symbol='MIR412', Aliases='MIRN412, hsa-mir-412, mir-412', description='microRNA 412', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101065447, end_position_on_the_genomic_accession=101065537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574434, CurrentID=0, Status='live', Symbol='MIR410', Aliases='MIRN410, hsa-mir-410, mir-410', description='microRNA 410', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101065912, end_position_on_the_genomic_accession=101065991, orientation='plus', exon_count=1, OMIM=615036, no_hdr0=''),
  574435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574435, CurrentID=0, Status='live', Symbol='MIR376B', Aliases='MIRN376B, mir-376b', description='microRNA 376b', other_designations='hsa-mir-376b', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101040436, end_position_on_the_genomic_accession=101040535, orientation='plus', exon_count=1, OMIM=610961, no_hdr0=''),
  574436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574436, CurrentID=0, Status='live', Symbol='MIR485', Aliases='MIRN485, hsa-mir-485, mir-485', description='microRNA 485', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101055419, end_position_on_the_genomic_accession=101055491, orientation='plus', exon_count=1, OMIM=615385, no_hdr0=''),
  100189157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189157, CurrentID=0, Status='live', Symbol='TRS-GCT1-1', Aliases='TRNAS12', description='transfer RNA-Ser (GCT) 1-1', other_designations='tRNA-Ser-GCT-1-1|transfer RNA serine 12 (anticodon GCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27097306, end_position_on_the_genomic_accession=27097387, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189158, CurrentID=0, Status='live', Symbol='TRD-GTC2-10', Aliases='TRNAD8', description='transfer RNA-Asp (GTC) 2-10', other_designations='tRNA-Asp-GTC-2-10|transfer RNA aspartic acid 8 (anticodon GUC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27479674, end_position_on_the_genomic_accession=27479745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189159, CurrentID=0, Status='live', Symbol='TRS-GCT5-1', Aliases='TRNAS13', description='transfer RNA-Ser (GCT) 5-1', other_designations='tRNA-Ser-GCT-5-1|transfer RNA serine 13 (anticodon GCU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28213037, end_position_on_the_genomic_accession=28213118, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189160, CurrentID=0, Status='live', Symbol='TRG-GCC1-4', Aliases='TRNAG15', description='transfer RNA-Gly (GCC) 1-4', other_designations='tRNA-Gly-GCC-1-4|transfer RNA glycine 15 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161465468, end_position_on_the_genomic_accession=161465538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  574441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574441, CurrentID=0, Status='live', Symbol='MIR488', Aliases='MIRN488, hsa-mir-488, mir-488', description='microRNA 488', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177029363, end_position_on_the_genomic_accession=177029445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574442, CurrentID=0, Status='live', Symbol='MIR489', Aliases='MIRN489, hsa-mir-489, mir-489', description='microRNA 489', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93483936, end_position_on_the_genomic_accession=93484019, orientation='minus', exon_count=1, OMIM=614523, no_hdr0=''),
  574443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574443, CurrentID=0, Status='live', Symbol='MIR490', Aliases='MIRN490, hsa-mir-490, miR-490', description='microRNA 490', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136903167, end_position_on_the_genomic_accession=136903294, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574444, CurrentID=0, Status='live', Symbol='MIR491', Aliases='MIRN491, hsa-mir-491, mir-491', description='microRNA 491', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20716105, end_position_on_the_genomic_accession=20716188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574445, CurrentID=0, Status='live', Symbol='MIR511', Aliases='MIR511-1-2, MIRN511-1, MIRN511-2, hsa-mir-511, hsa-mir-511-2, mir-511, MIR511', description='microRNA 511', other_designations='hsa-mir-511-1|microRNA 511-1|microRNA 511-2', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17845107, end_position_on_the_genomic_accession=17845193, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100189166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189166, CurrentID=0, Status='live', Symbol='TRR-TCT4-1', Aliases='TRNAR16', description='transfer RNA-Arg (TCT) 4-1', other_designations='tRNA-Arg-TCT-4-1|transfer RNA arginine 16 (anticodon UCU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159141611, end_position_on_the_genomic_accession=159141684, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  574447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574447, CurrentID=0, Status='live', Symbol='MIR146B', Aliases='MIRN146B, miRNA146B, mir-146b', description='microRNA 146b', other_designations='hsa-mir-146b', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102436512, end_position_on_the_genomic_accession=102436584, orientation='plus', exon_count=1, OMIM=610567, no_hdr0=''),
  574448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574448, CurrentID=0, Status='live', Symbol='MIR202', Aliases='MIRN202, hsa-mir-202, mir-202', description='microRNA 202', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133247511, end_position_on_the_genomic_accession=133247620, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574449, CurrentID=0, Status='live', Symbol='MIR492', Aliases='MIRN492, hsa-mir-492', description='microRNA 492', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94834398, end_position_on_the_genomic_accession=94834513, orientation='plus', exon_count=1, OMIM=614384, no_hdr0=''),
  574450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574450, CurrentID=0, Status='live', Symbol='MIR493', Aliases='MIRN493, hsa-mir-493, mir-493', description='microRNA 493', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100869060, end_position_on_the_genomic_accession=100869148, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574451, CurrentID=0, Status='live', Symbol='MIR432', Aliases='MIRN432, hsa-mir-432, mir-432', description='microRNA 432', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100884483, end_position_on_the_genomic_accession=100884576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574452, CurrentID=0, Status='live', Symbol='MIR494', Aliases='MIRN494, hsa-mir-494, mir-494', description='microRNA 494', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101029634, end_position_on_the_genomic_accession=101029714, orientation='plus', exon_count=1, OMIM=616036, no_hdr0=''),
  574453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574453, CurrentID=0, Status='live', Symbol='MIR495', Aliases='MIRN495, hsa-mir-495, mir-495', description='microRNA 495', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101033755, end_position_on_the_genomic_accession=101033836, orientation='plus', exon_count=1, OMIM=615149, no_hdr0=''),
  574454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574454, CurrentID=0, Status='live', Symbol='MIR496', Aliases='MIRN496, hsa-mir-496, mir-496', description='microRNA 496', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101060573, end_position_on_the_genomic_accession=101060674, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8361, CurrentID=0, Status='live', Symbol='HIST1H4F', Aliases='H4, H4/c, H4FC', description='histone cluster 1, H4f', other_designations='H4 histone family, member C|histone 1, H4f', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26240426, end_position_on_the_genomic_accession=26240793, orientation='plus', exon_count=1, OMIM=602824, no_hdr0=''),
  574456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574456, CurrentID=0, Status='live', Symbol='MIR497', Aliases='MIRN497, hsa-mir-497, mir-497', description='microRNA 497', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7017911, end_position_on_the_genomic_accession=7018022, orientation='minus', exon_count=1, OMIM=615672, no_hdr0=''),
  574457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574457, CurrentID=0, Status='live', Symbol='MIR181D', Aliases='MIRN181D, mir-181d', description='microRNA 181d', other_designations='hsa-mir-181d', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13874875, end_position_on_the_genomic_accession=13875011, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574458, CurrentID=0, Status='live', Symbol='MIR512-1', Aliases='MIRN512-1, mir-512-1', description='microRNA 512-1', other_designations='hsa-mir-512-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53666679, end_position_on_the_genomic_accession=53666762, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574459, CurrentID=0, Status='live', Symbol='MIR512-2', Aliases='MIRN512-2, mir-512-2', description='microRNA 512-2', other_designations='hsa-mir-512-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53669157, end_position_on_the_genomic_accession=53669254, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574460, CurrentID=0, Status='live', Symbol='MIR498', Aliases='MIRN498, hsa-mir-498, mir-498', description='microRNA 498', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53674197, end_position_on_the_genomic_accession=53674320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574461, CurrentID=0, Status='live', Symbol='MIR520E', Aliases='MIRN520E', description='microRNA 520e', other_designations='hsa-mir-520e', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53675711, end_position_on_the_genomic_accession=53675797, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574462, CurrentID=0, Status='live', Symbol='MIR515-1', Aliases='MIRN515-1', description='microRNA 515-1', other_designations='hsa-mir-515-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53679003, end_position_on_the_genomic_accession=53679085, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574463, CurrentID=0, Status='live', Symbol='MIR519E', Aliases='MIRN519E', description='microRNA 519e', other_designations='hsa-mir-519e', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53679940, end_position_on_the_genomic_accession=53680023, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574464, CurrentID=0, Status='live', Symbol='MIR520F', Aliases='MIRN520F, mir-520f', description='microRNA 520f', other_designations='hsa-mir-520f', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53682159, end_position_on_the_genomic_accession=53682245, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574465, CurrentID=0, Status='live', Symbol='MIR515-2', Aliases='MIRN515-2', description='microRNA 515-2', other_designations='hsa-mir-515-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53685009, end_position_on_the_genomic_accession=53685091, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574466, CurrentID=0, Status='live', Symbol='MIR519C', Aliases='MIRN519C, mir-519c', description='microRNA 519c', other_designations='hsa-mir-519c', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53686469, end_position_on_the_genomic_accession=53686555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574467, CurrentID=0, Status='live', Symbol='MIR520A', Aliases='MIRN520A', description='microRNA 520a', other_designations='hsa-mir-520a', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53690881, end_position_on_the_genomic_accession=53690965, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574468, CurrentID=0, Status='live', Symbol='MIR526B', Aliases='MIRN526B', description='microRNA 526b', other_designations='hsa-mir-526b', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53694393, end_position_on_the_genomic_accession=53694475, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574469, CurrentID=0, Status='live', Symbol='MIR519B', Aliases='MIRN519B, mir-519b', description='microRNA 519b', other_designations='hsa-mir-519b', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53695213, end_position_on_the_genomic_accession=53695293, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574470, CurrentID=0, Status='live', Symbol='MIR525', Aliases='MIRN525, hsa-mir-525', description='microRNA 525', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53697533, end_position_on_the_genomic_accession=53697617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574471, CurrentID=0, Status='live', Symbol='MIR523', Aliases='MIRN523, hsa-mir-523, mir-523', description='microRNA 523', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53698385, end_position_on_the_genomic_accession=53698471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574472, CurrentID=0, Status='live', Symbol='MIR518F', Aliases='MIRN518F', description='microRNA 518f', other_designations='hsa-mir-518f', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53700015, end_position_on_the_genomic_accession=53700101, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574473, CurrentID=0, Status='live', Symbol='MIR520B', Aliases='MIRN520B', description='microRNA 520b', other_designations='hsa-mir-520b', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53701227, end_position_on_the_genomic_accession=53701287, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574474, CurrentID=0, Status='live', Symbol='MIR518B', Aliases='MIRN518B, mir-518b', description='microRNA 518b', other_designations='hsa-mir-518b', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53702737, end_position_on_the_genomic_accession=53702819, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574475, CurrentID=0, Status='live', Symbol='MIR526A1', Aliases='MIRN526A-1, MIRN526A1', description='microRNA 526a-1', other_designations='hsa-mir-526a-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53706252, end_position_on_the_genomic_accession=53706336, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574476, CurrentID=0, Status='live', Symbol='MIR520C', Aliases='MIRN520C', description='microRNA 520c', other_designations='hsa-mir-520c', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53707453, end_position_on_the_genomic_accession=53707539, orientation='plus', exon_count=1, OMIM=615908, no_hdr0=''),
  574477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574477, CurrentID=0, Status='live', Symbol='MIR518C', Aliases='MIRN518C', description='microRNA 518c', other_designations='hsa-mir-518c|miR-518', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53708735, end_position_on_the_genomic_accession=53708835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574478, CurrentID=0, Status='live', Symbol='MIR524', Aliases='MIRN524, hsa-mir-524, mir-524', description='microRNA 524', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53711002, end_position_on_the_genomic_accession=53711088, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574479, CurrentID=0, Status='live', Symbol='MIR517A', Aliases='MIRN517A, mir-517a', description='microRNA 517a', other_designations='hsa-mir-517a', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53712268, end_position_on_the_genomic_accession=53712354, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574480, CurrentID=0, Status='live', Symbol='MIR519D', Aliases='MIRN519D, mir-519d', description='microRNA 519d', other_designations='hsa-mir-519d', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53713347, end_position_on_the_genomic_accession=53713434, orientation='plus', exon_count=1, OMIM=614247, no_hdr0=''),
  574481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574481, CurrentID=0, Status='live', Symbol='MIR521-2', Aliases='MIRN521-2, mir-521-2', description='microRNA 521-2', other_designations='hsa-mir-521-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53716594, end_position_on_the_genomic_accession=53716680, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574482, CurrentID=0, Status='live', Symbol='MIR520D', Aliases='MIRN520D', description='microRNA 520d', other_designations='hsa-mir-520d', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53720096, end_position_on_the_genomic_accession=53720182, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574483, CurrentID=0, Status='live', Symbol='MIR517B', Aliases='MIRN517B, mir-517b', description='microRNA 517b', other_designations='hsa-mir-517b', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53721076, end_position_on_the_genomic_accession=53721142, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574484, CurrentID=0, Status='live', Symbol='MIR520G', Aliases='MIRN520G, mir-520g', description='microRNA 520g', other_designations='hsa-mir-520g', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53722166, end_position_on_the_genomic_accession=53722255, orientation='plus', exon_count=1, OMIM=616272, no_hdr0=''),
  574485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574485, CurrentID=0, Status='live', Symbol='MIR516B2', Aliases='MIRN516-3, MIRN516B-2, MIRN516B2, mir-516b-2', description='microRNA 516b-2', other_designations='hsa-mir-516-3|hsa-mir-516b-2|microRNA 516-3', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53725442, end_position_on_the_genomic_accession=53725526, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574486, CurrentID=0, Status='live', Symbol='MIR526A2', Aliases='MIRN526A-2, MIRN526A2', description='microRNA 526a-2', other_designations='hsa-mir-526a-2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53726922, end_position_on_the_genomic_accession=53726986, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574487, CurrentID=0, Status='live', Symbol='MIR518E', Aliases='MIRN518E, mir-518e', description='microRNA 518e', other_designations='hsa-mir-518e', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53729838, end_position_on_the_genomic_accession=53729925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574488, CurrentID=0, Status='live', Symbol='MIR518A1', Aliases='MIRN518A-1, MIRN518A1, mir-518a-1', description='microRNA 518a-1', other_designations='hsa-mir-518a-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53731006, end_position_on_the_genomic_accession=53731090, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574489, CurrentID=0, Status='live', Symbol='MIR518D', Aliases='MIRN518D', description='microRNA 518d', other_designations='hsa-mir-518d', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53734877, end_position_on_the_genomic_accession=53734963, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574490, CurrentID=0, Status='live', Symbol='MIR516B1', Aliases='MIRN516-4, MIRN516B-1, MIRN516B1, mir-516b-1', description='microRNA 516b-1', other_designations='hsa-mir-516-4|hsa-mir-516b-1|microRNA 516-4', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53736845, end_position_on_the_genomic_accession=53736934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574491, CurrentID=0, Status='live', Symbol='MIR518A2', Aliases='MIRN518A-2, MIRN518A2, mir-518a-2', description='microRNA 518a-2', other_designations='hsa-mir-518a-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53739333, end_position_on_the_genomic_accession=53739419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574492, CurrentID=0, Status='live', Symbol='MIR517C', Aliases='MIRN517C', description='microRNA 517c', other_designations='hsa-mir-517c', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53741313, end_position_on_the_genomic_accession=53741407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574493, CurrentID=0, Status='live', Symbol='MIR520H', Aliases='MIRN520H', description='microRNA 520h', other_designations='hsa-mir-520h', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53742512, end_position_on_the_genomic_accession=53742599, orientation='plus', exon_count=1, OMIM=614755, no_hdr0=''),
  574494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574494, CurrentID=0, Status='live', Symbol='MIR521-1', Aliases='MIRN521-1, mir-521-1', description='microRNA 521-1', other_designations='hsa-mir-521-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53748636, end_position_on_the_genomic_accession=53748722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574495, CurrentID=0, Status='live', Symbol='MIR522', Aliases='MIRN522, hsa-mir-522, mir-522', description='microRNA 522', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53751211, end_position_on_the_genomic_accession=53751297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574496, CurrentID=0, Status='live', Symbol='MIR519A1', Aliases='MIRN519A-1, MIRN519A1, mir-519a-1', description='microRNA 519a-1', other_designations='hsa-mir-519a-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53752397, end_position_on_the_genomic_accession=53752481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574497, CurrentID=0, Status='live', Symbol='MIR527', Aliases='MIRN527, hsa-mir-527', description='microRNA 527', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53754018, end_position_on_the_genomic_accession=53754102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574498, CurrentID=0, Status='live', Symbol='MIR516A1', Aliases='MIRN516-1, MIRN516A-1, MIRN516A1, mir-516a-1', description='microRNA 516a-1', other_designations='hsa-mir-516-1|hsa-mir-516a-1|microRNA 516-1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53756741, end_position_on_the_genomic_accession=53756830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574499, CurrentID=0, Status='live', Symbol='MIR516A2', Aliases='MIRN516-2, MIRN516A-2, MIRN516A2, mir-516a-2', description='microRNA 516a-2', other_designations='hsa-mir-516-2|hsa-mir-516a-2|microRNA 516-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53761133, end_position_on_the_genomic_accession=53761222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574500, CurrentID=0, Status='live', Symbol='MIR519A2', Aliases='MIRN519A-2, MIRN519A2, mir-519a-2', description='microRNA 519a-2', other_designations='hsa-mir-519a-2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53762344, end_position_on_the_genomic_accession=53762430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574501, CurrentID=0, Status='live', Symbol='MIR499A', Aliases='MIR499, MIRN499, hsa-mir-499a, mir-499a', description='microRNA 499a', other_designations='hsa-mir-499|microRNA 499', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34990376, end_position_on_the_genomic_accession=34990497, orientation='plus', exon_count=1, OMIM=613614, no_hdr0=''),
  574502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574502, CurrentID=0, Status='live', Symbol='MIR500A', Aliases='MIR500, MIRN500, hsa-mir-500, hsa-mir-500a, mir-500a', description='microRNA 500a', other_designations='microRNA 500', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50008431, end_position_on_the_genomic_accession=50008514, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574503, CurrentID=0, Status='live', Symbol='MIR501', Aliases='MIRN501, hsa-mir-501, mir-501', description='microRNA 501', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50009722, end_position_on_the_genomic_accession=50009805, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574504, CurrentID=0, Status='live', Symbol='MIR502', Aliases='MIRN502, hsa-mir-502, mir-502', description='microRNA 502', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50014598, end_position_on_the_genomic_accession=50014683, orientation='plus', exon_count=1, OMIM=300893, no_hdr0=''),
  574505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574505, CurrentID=0, Status='live', Symbol='MIR450A2', Aliases='MIRN450-2, MIRN450A-2, MIRN450A2, mir-450a-2', description='microRNA 450a-2', other_designations='hsa-mir-450-2|hsa-mir-450a-2|microRNA 450-2', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134540508, end_position_on_the_genomic_accession=134540607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574506, CurrentID=0, Status='live', Symbol='MIR503', Aliases='MIRN503, hsa-mir-503, mir-503', description='microRNA 503', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134546328, end_position_on_the_genomic_accession=134546398, orientation='minus', exon_count=1, OMIM=300865, no_hdr0=''),
  574507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574507, CurrentID=0, Status='live', Symbol='MIR504', Aliases='MIRN504, hsa-mir-504, mir-504', description='microRNA 504', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138667711, end_position_on_the_genomic_accession=138667793, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574508, CurrentID=0, Status='live', Symbol='MIR505', Aliases='MIRN505, hsa-mir-505, mir-505', description='microRNA 505', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139924148, end_position_on_the_genomic_accession=139924231, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574509, CurrentID=0, Status='live', Symbol='MIR513A1', Aliases='MIRN513-1, MIRN513A1', description='microRNA 513a-1', other_designations='hsa-mir-513-1|hsa-mir-513a-1|microRNA 513-1', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147213463, end_position_on_the_genomic_accession=147213591, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574510, CurrentID=0, Status='live', Symbol='MIR513A2', Aliases='MIRN513-2, MIRN513A-2, MIRN513A2, hsa-mir-513a-2', description='microRNA 513a-2', other_designations='hsa-mir-513-2|microRNA 513-2', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147225826, end_position_on_the_genomic_accession=147225952, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574511, CurrentID=0, Status='live', Symbol='MIR506', Aliases='MIRN506, hsa-mir-506, mir-506', description='microRNA 506', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147230720, end_position_on_the_genomic_accession=147230843, orientation='minus', exon_count=1, OMIM=300877, no_hdr0=''),
  574512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574512, CurrentID=0, Status='live', Symbol='MIR507', Aliases='MIRN507, hsa-mir-507, mir-507', description='microRNA 507', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147230984, end_position_on_the_genomic_accession=147231077, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574513, CurrentID=0, Status='live', Symbol='MIR508', Aliases='MIRN508, hsa-mir-508', description='microRNA 508', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147236913, end_position_on_the_genomic_accession=147237027, orientation='minus', exon_count=1, OMIM=300874, no_hdr0=''),
  574514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574514, CurrentID=0, Status='live', Symbol='MIR509-1', Aliases='MIRN509, MIRN509-1, hsa-mir-509, mir-509-1', description='microRNA 509-1', other_designations='hsa-mir-509-1|miR-509-5p|microRNA 509', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147260532, end_position_on_the_genomic_accession=147260625, orientation='minus', exon_count=1, OMIM=300875, no_hdr0=''),
  574515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574515, CurrentID=0, Status='live', Symbol='MIR510', Aliases='MIRN510, hsa-mir-510, mir-510', description='microRNA 510', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147272335, end_position_on_the_genomic_accession=147272408, orientation='minus', exon_count=1, OMIM=300866, no_hdr0=''),
  574516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574516, CurrentID=0, Status='live', Symbol='MIR514A1', Aliases='MIR514-1, MIRN514-1, hsa-mir-514a-1, mir-514a-1', description='microRNA 514a-1', other_designations='hsa-mir-514-1|microRNA 514-1', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147279247, end_position_on_the_genomic_accession=147279344, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574517, CurrentID=0, Status='live', Symbol='MIR514A2', Aliases='MIR514-2, MIRN514-2, hsa-mir-514a-2', description='microRNA 514a-2', other_designations='hsa-mir-514-2|microRNA 514-2', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147281943, end_position_on_the_genomic_accession=147282030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  574518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574518, CurrentID=0, Status='live', Symbol='MIR514A3', Aliases='MIR514-3, MIRN514-3, hsa-mir-514a-3', description='microRNA 514a-3', other_designations='hsa-mir-514-3|microRNA 514-3', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147284641, end_position_on_the_genomic_accession=147284728, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100189239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189239, CurrentID=0, Status='live', Symbol='TRD-GTC2-6', Aliases='TRNAD12', description='transfer RNA-Asp (GTC) 2-6', other_designations='tRNA-Asp-GTC-2-6|transfer RNA aspartic acid 12 (anticodon GUC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96036021, end_position_on_the_genomic_accession=96036092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189240, CurrentID=0, Status='live', Symbol='TRL-TAA3-1', Aliases='TRNAL23', description='transfer RNA-Leu (TAA) 3-1', other_designations='tRNA-Leu-TAA-3-1|transfer RNA leucine 23 (anticodon UAA)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59551755, end_position_on_the_genomic_accession=59551837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401588, CurrentID=0, Status='live', Symbol='ZNF674-AS1', Aliases='', description='ZNF674 antisense RNA 1 (head to head)', other_designations='ZNF674 antisense RNA 1 (non-protein coding)', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46545490, end_position_on_the_genomic_accession=46548475, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189242, CurrentID=0, Status='live', Symbol='TRH-GTG3-1', Aliases='TRNAH8', description='transfer RNA-His (GTG) 3-1', other_designations='tRNA-His-GTG-3-1|transfer RNA histidine 8 (anticodon GUG)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189243, CurrentID=0, Status='live', Symbol='TRR-ACG1-1', Aliases='TRNAR21', description='transfer RNA-Arg (ACG) 1-1', other_designations='tRNA-Arg-ACG-1-1|transfer RNA arginine 21 (anticodon ACG)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22929701, end_position_on_the_genomic_accession=22929773, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189244, CurrentID=0, Status='live', Symbol='TRK-CTT8-1', Aliases='TRNAK24', description='transfer RNA-Lys (CTT) 8-1', other_designations='tRNA-Lys-CTT-8-1|transfer RNA lysine 24 (anticodon CUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189245, CurrentID=0, Status='live', Symbol='TRT-CGT4-1', Aliases='TRNAT10', description='transfer RNA-Thr (CGT) 4-1', other_designations='tRNA-Thr-CGT-4-1|transfer RNA threonine 10 (anticodon CGU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31550074, end_position_on_the_genomic_accession=31550145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189246, CurrentID=0, Status='live', Symbol='TRA-AGC19-1', Aliases='TRNAA24', description='transfer RNA-Ala (AGC) 19-1', other_designations='tRNA-Ala-AGC-19-1|transfer RNA alanine 24 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189247, CurrentID=0, Status='live', Symbol='TRS-AGA2-5', Aliases='TRNAS19', description='transfer RNA-Ser (AGA) 2-5', other_designations='tRNA-Ser-AGA-2-5|transfer RNA serine 19 (anticodon AGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27503039, end_position_on_the_genomic_accession=27503120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189248, CurrentID=0, Status='live', Symbol='TRS-AGA2-6', Aliases='TRNAS20', description='transfer RNA-Ser (AGA) 2-6', other_designations='tRNA-Ser-AGA-2-6|transfer RNA serine 20 (anticodon AGA)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95269657, end_position_on_the_genomic_accession=95269738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189249, CurrentID=0, Status='live', Symbol='TRI-AAT1-1', Aliases='TRNAI12', description='transfer RNA-Ile (AAT) 1-1', other_designations='tRNA-Ile-AAT-1-1|transfer RNA isoleucine 12 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57800211, end_position_on_the_genomic_accession=57800284, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189250, CurrentID=0, Status='live', Symbol='TRC-GCA9-1', Aliases='TRNAC18', description='transfer RNA-Cys (GCA) 9-1', other_designations='tRNA-Cys-GCA-9-1|transfer RNA cysteine 18 (anticodon GCA)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132231798, end_position_on_the_genomic_accession=132231869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189251, CurrentID=0, Status='live', Symbol='TRF-GAA1-3', Aliases='TRNAF6', description='transfer RNA-Phe (GAA) 1-3', other_designations='tRNA-Phe-GAA-1-3|transfer RNA phenylalanine 6 (anticodon GAA)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94549650, end_position_on_the_genomic_accession=94549722, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189252, CurrentID=0, Status='live', Symbol='TRG-GCC1-2', Aliases='TRNAG23', description='transfer RNA-Gly (GCC) 1-2', other_designations='tRNA-Gly-GCC-1-2|transfer RNA glycine 23 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161450677, end_position_on_the_genomic_accession=161450747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480709, CurrentID=0, Status='live', Symbol='RNU6-38P', Aliases='RNU6-38', description='RNA, U6 small nuclear 38, pseudogene', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189254, CurrentID=0, Status='live', Symbol='TRF-GAA1-6', Aliases='TRNAF7', description='transfer RNA-Phe (GAA) 1-6', other_designations='tRNA-Phe-GAA-1-6|transfer RNA phenylalanine 7 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28981672, end_position_on_the_genomic_accession=28981744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189255, CurrentID=0, Status='live', Symbol='TRX-CAT1-2', Aliases='TRNAM14', description='transfer RNA-iMet (CAT) 1-2', other_designations='tRNA-iMet-CAT-1-2|transfer RNA initiator methionine (CAT) 1-2|transfer RNA methionine 14 (anticodon CAU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82494721, end_position_on_the_genomic_accession=82494792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189256, CurrentID=0, Status='live', Symbol='TRA-TGC8-1', Aliases='TRNAA25', description='transfer RNA-Ala (TGC) 8-1', other_designations='tRNA-Ala-TGC-8-1|transfer RNA alanine 25 (anticodon UGC)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  574537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574537, CurrentID=0, Status='live', Symbol='UGT2A2', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A2', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69588417, end_position_on_the_genomic_accession=69639642, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100189258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189258, CurrentID=0, Status='live', Symbol='TRT-CGT6-1', Aliases='TRNAT11', description='transfer RNA-Thr (CGT) 6-1', other_designations='tRNA-Thr-CGT-6-1|transfer RNA threonine 11 (anticodon CGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189259, CurrentID=0, Status='live', Symbol='TRE-CTC1-4', Aliases='TRNAE18', description='transfer RNA-Glu (CTC) 1-4', other_designations='tRNA-Glu-CTC-1-4|transfer RNA glutamic acid 18 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161462019, end_position_on_the_genomic_accession=161462090, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189260, CurrentID=0, Status='live', Symbol='TRG-TCC2-2', Aliases='TRNAG24', description='transfer RNA-Gly (TCC) 2-2', other_designations='tRNA-Gly-TCC-2-2|transfer RNA glycine 24 (anticodon UCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161447585, end_position_on_the_genomic_accession=161447656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189261, CurrentID=0, Status='live', Symbol='TRQ-CTG4-2', Aliases='TRNAQ20', description='transfer RNA-Gln (CTG) 4-2', other_designations='tRNA-Gln-CTG-4-2|transfer RNA glutamine 20 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143691474, end_position_on_the_genomic_accession=143691545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189262, CurrentID=0, Status='live', Symbol='TRF-GAA6-1', Aliases='TRNAF8', description='transfer RNA-Phe (GAA) 6-1', other_designations='tRNA-Phe-GAA-6-1|transfer RNA phenylalanine 8 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28763597, end_position_on_the_genomic_accession=28763670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189263, CurrentID=0, Status='live', Symbol='TRT-TGT5-1', Aliases='TRNAT12', description='transfer RNA-Thr (TGT) 5-1', other_designations='tRNA-Thr-TGT-5-1|transfer RNA threonine 12 (anticodon UGU)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20681690, end_position_on_the_genomic_accession=20681762, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189264, CurrentID=0, Status='live', Symbol='TRR-TCG1-1', Aliases='TRNAR22', description='transfer RNA-Arg (TCG) 1-1', other_designations='tRNA-Arg-TCG-1-1|transfer RNA arginine 22 (anticodon UCG)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89335073, end_position_on_the_genomic_accession=89335145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480721, CurrentID=0, Status='live', Symbol='PRYP6', Aliases='', description='PTPN13-like, Y-linked pseudogene 6', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189266, CurrentID=0, Status='live', Symbol='TRN-GTT20-1', Aliases='TRNAN23', description='transfer RNA-Asn (GTT) 20-1', other_designations='tRNA-Asn-GTT-20-1|transfer RNA asparagine 23 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189267, CurrentID=0, Status='live', Symbol='TRC-GCA15-1', Aliases='TRNAC19', description='transfer RNA-Cys (GCA) 15-1', other_designations='tRNA-Cys-GCA-15-1|transfer RNA cysteine 19 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149584725, end_position_on_the_genomic_accession=149584796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189268, CurrentID=0, Status='live', Symbol='TRQ-CTG2-1', Aliases='TRNAQ21', description='transfer RNA-Gln (CTG) 2-1', other_designations='tRNA-Gln-CTG-2-1|transfer RNA glutamine 21 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27547752, end_position_on_the_genomic_accession=27547823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480725, CurrentID=0, Status='live', Symbol='HMGN1P12', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 12', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189270, CurrentID=0, Status='live', Symbol='TRR-CCT5-1', Aliases='TRNAR23', description='transfer RNA-Arg (CCT) 5-1', other_designations='tRNA-Arg-CCT-5-1|transfer RNA arginine 23 (anticodon CCU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3193918, end_position_on_the_genomic_accession=3193990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189271, CurrentID=0, Status='live', Symbol='TRN-GTT16-2', Aliases='TRNAN24', description='transfer RNA-Asn (GTT) 16-2', other_designations='tRNA-Asn-GTT-16-2|transfer RNA asparagine 24 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480728, CurrentID=0, Status='live', Symbol='MTND6P17', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 17', other_designations='MT-ND6 pseudogene 17', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189273, CurrentID=0, Status='live', Symbol='TRL-AAG1-2', Aliases='TRNAL24', description='transfer RNA-Leu (AAG) 1-2', other_designations='tRNA-Leu-AAG-1-2|transfer RNA leucine 24 (anticodon AAG)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181101840, end_position_on_the_genomic_accession=181101921, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189274, CurrentID=0, Status='live', Symbol='TRG-GCC4-1', Aliases='TRNAG25', description='transfer RNA-Gly (GCC) 4-1', other_designations='tRNA-Gly-GCC-4-1|transfer RNA glycine 25 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161480566, end_position_on_the_genomic_accession=161480636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189275, CurrentID=0, Status='live', Symbol='TRI-AAT5-1', Aliases='TRNAI15', description='transfer RNA-Ile (AAT) 5-1', other_designations='tRNA-Ile-AAT-5-1|transfer RNA isoleucine 15 (anticodon AAU)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102317092, end_position_on_the_genomic_accession=102317165, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189276, CurrentID=0, Status='live', Symbol='TRP-AGG2-8', Aliases='TRNAP15', description='transfer RNA-Pro (AGG) 2-8', other_designations='tRNA-Pro-AGG-2-8|transfer RNA proline 15 (anticodon AGG)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128783450, end_position_on_the_genomic_accession=128783521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189277, CurrentID=0, Status='live', Symbol='TRG-CCC3-1', Aliases='TRNAG26', description='transfer RNA-Gly (CCC) 3-1', other_designations='tRNA-Gly-CCC-3-1|transfer RNA glycine 26 (anticodon CCC)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19860862, end_position_on_the_genomic_accession=19860932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189278, CurrentID=0, Status='live', Symbol='TRY-GTA10-1', Aliases='TRNAY11', description='transfer RNA-Tyr (GTA) 10-1', other_designations='tRNA-Tyr-GTA-10-1|transfer RNA tyrosine 11 (anticodon GUA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189279, CurrentID=0, Status='live', Symbol='TRS-TGA1-1', Aliases='TRNAS21', description='transfer RNA-Ser (TGA) 1-1', other_designations='tRNA-Ser-TGA-1-1|transfer RNA serine 21 (anticodon UGA)', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67764503, end_position_on_the_genomic_accession=67764584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189280, CurrentID=0, Status='live', Symbol='TRA-TGC3-2', Aliases='TRNAA26', description='transfer RNA-Ala (TGC) 3-2', other_designations='tRNA-Ala-TGC-3-2|transfer RNA alanine 26 (anticodon UGC)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181206868, end_position_on_the_genomic_accession=181206939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189281, CurrentID=0, Status='live', Symbol='TRP-TGG3-4', Aliases='TRNAP16', description='transfer RNA-Pro (TGG) 3-4', other_designations='tRNA-Pro-TGG-3-4|transfer RNA proline 16 (anticodon UGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3188094, end_position_on_the_genomic_accession=3188165, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189282, CurrentID=0, Status='live', Symbol='TRK-CTT2-2', Aliases='TRNAK26', description='transfer RNA-Lys (CTT) 2-2', other_designations='tRNA-Lys-CTT-2-2|transfer RNA lysine 26 (anticodon CUU)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3175691, end_position_on_the_genomic_accession=3175763, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189283, CurrentID=0, Status='live', Symbol='TRV-TAC2-1', Aliases='TRNAV23', description='transfer RNA-Val (TAC) 2-1', other_designations='tRNA-Val-TAC-2-1|transfer RNA valine 23 (anticodon UAC)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59550987, end_position_on_the_genomic_accession=59551059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189284, CurrentID=0, Status='live', Symbol='TRG-GCC1-5', Aliases='TRNAG27', description='transfer RNA-Gly (GCC) 1-5', other_designations='tRNA-Gly-GCC-1-5|transfer RNA glycine 27 (anticodon GCC)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17454789, end_position_on_the_genomic_accession=17454859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189285, CurrentID=0, Status='live', Symbol='TRS-GCT3-1', Aliases='TRNAS22', description='transfer RNA-Ser (GCT) 3-1', other_designations='tRNA-Ser-GCT-3-1|transfer RNA serine 22 (anticodon GCU)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66348120, end_position_on_the_genomic_accession=66348201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189286, CurrentID=0, Status='live', Symbol='TRP-TGG1-1', Aliases='TRNAP17', description='transfer RNA-Pro (TGG) 1-1', other_designations='tRNA-Pro-TGG-1-1|transfer RNA proline 17 (anticodon UGG)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20633006, end_position_on_the_genomic_accession=20633077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189287, CurrentID=0, Status='live', Symbol='TRQ-CTG1-3', Aliases='TRNAQ22', description='transfer RNA-Gln (CTG) 1-3', other_designations='tRNA-Gln-CTG-1-3|transfer RNA glutamine 22 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18836171, end_position_on_the_genomic_accession=18836242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189288, CurrentID=0, Status='live', Symbol='TRL-CAG1-7', Aliases='TRNAL26', description='transfer RNA-Leu (CAG) 1-7', other_designations='tRNA-Leu-CAG-1-7|transfer RNA leucine 26 (anticodon CAG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26521208, end_position_on_the_genomic_accession=26521290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189289, CurrentID=0, Status='live', Symbol='TRC-GCA17-1', Aliases='TRNAC20', description='transfer RNA-Cys (GCA) 17-1', other_designations='tRNA-Cys-GCA-17-1|transfer RNA cysteine 20 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149691181, end_position_on_the_genomic_accession=149691252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189290, CurrentID=0, Status='live', Symbol='TRD-GTC2-3', Aliases='TRNAD13', description='transfer RNA-Asp (GTC) 2-3', other_designations='tRNA-Asp-GTC-2-3|transfer RNA aspartic acid 13 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161455624, end_position_on_the_genomic_accession=161455695, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189291, CurrentID=0, Status='live', Symbol='TRL-AAG2-2', Aliases='TRNAL27', description='transfer RNA-Leu (AAG) 2-2', other_designations='tRNA-Leu-AAG-2-2|transfer RNA leucine 27 (anticodon AAG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22297140, end_position_on_the_genomic_accession=22297221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189292, CurrentID=0, Status='live', Symbol='TRD-GTC2-11', Aliases='TRNAD14', description='transfer RNA-Asp (GTC) 2-11', other_designations='tRNA-Asp-GTC-2-11|transfer RNA aspartic acid 14 (anticodon GUC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27503744, end_position_on_the_genomic_accession=27503815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189293, CurrentID=0, Status='live', Symbol='TRF-GAA4-1', Aliases='TRNAF9', description='transfer RNA-Phe (GAA) 4-1', other_designations='tRNA-Phe-GAA-4-1|transfer RNA phenylalanine 9 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28823316, end_position_on_the_genomic_accession=28823389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189294, CurrentID=0, Status='live', Symbol='TRK-TTT2-1', Aliases='TRNAK27', description='transfer RNA-Lys (TTT) 2-1', other_designations='tRNA-Lys-TTT-2-1|transfer RNA lysine 27 (anticodon UUU)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122559947, end_position_on_the_genomic_accession=122560019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189295, CurrentID=0, Status='live', Symbol='TRA-AGC2-2', Aliases='TRNAA27', description='transfer RNA-Ala (AGC) 2-2', other_designations='tRNA-Ala-AGC-2-2|transfer RNA alanine 27 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28863685, end_position_on_the_genomic_accession=28863756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189296, CurrentID=0, Status='live', Symbol='TRP-AGG2-6', Aliases='TRNAP18', description='transfer RNA-Pro (AGG) 2-6', other_designations='tRNA-Pro-AGG-2-6|transfer RNA proline 18 (anticodon AGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3189634, end_position_on_the_genomic_accession=3189705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189297, CurrentID=0, Status='live', Symbol='TRX-CAT1-7', Aliases='TRNAM15', description='transfer RNA-iMet (CAT) 1-7', other_designations='tRNA-iMet-CAT-1-7|transfer RNA initiator methionine (CAT) 1-7|transfer RNA methionine 15 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27592821, end_position_on_the_genomic_accession=27592892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189298, CurrentID=0, Status='live', Symbol='TRG-CCC1-2', Aliases='TRNAG28', description='transfer RNA-Gly (CCC) 1-2', other_designations='tRNA-Gly-CCC-1-2|transfer RNA glycine 28 (anticodon CCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16861921, end_position_on_the_genomic_accession=16861991, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189299, CurrentID=0, Status='live', Symbol='TRA-AGC3-1', Aliases='TRNAA28', description='transfer RNA-Ala (AGC) 3-1', other_designations='tRNA-Ala-AGC-3-1|transfer RNA alanine 28 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28607156, end_position_on_the_genomic_accession=28607227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189300, CurrentID=0, Status='live', Symbol='TRT-CGT5-1', Aliases='TRNAT13', description='transfer RNA-Thr (CGT) 5-1', other_designations='tRNA-Thr-CGT-5-1|transfer RNA threonine 13 (anticodon CGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27618356, end_position_on_the_genomic_accession=27618429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480757, CurrentID=0, Status='live', Symbol='RNA5SP161', Aliases='RN5S161', description='RNA, 5S ribosomal pseudogene 161', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189302, CurrentID=0, Status='live', Symbol='TRQ-CTG17-1', Aliases='TRNAQ23', description='transfer RNA-Gln (CTG) 17-1', other_designations='tRNA-Gln-CTG-17-1|transfer RNA glutamine 23 (anticodon CUG)', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189303, CurrentID=0, Status='live', Symbol='TRM-CAT5-1', Aliases='TRNAM16', description='transfer RNA-Met (CAT) 5-1', other_designations='tRNA-Met-CAT-5-1|transfer RNA methionine 16 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26800113, end_position_on_the_genomic_accession=26800185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189304, CurrentID=0, Status='live', Symbol='TRN-GTT9-1', Aliases='TRNAN25', description='transfer RNA-Asn (GTT) 9-1', other_designations='tRNA-Asn-GTT-9-1|transfer RNA asparagine 25 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=145979034, end_position_on_the_genomic_accession=145979107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189305, CurrentID=0, Status='live', Symbol='TRC-GCA10-1', Aliases='TRNAC21', description='transfer RNA-Cys (GCA) 10-1', other_designations='tRNA-Cys-GCA-10-1|transfer RNA cysteine 21 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149377510, end_position_on_the_genomic_accession=149377581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189306, CurrentID=0, Status='live', Symbol='TRF-GAA1-4', Aliases='TRNAF10', description='transfer RNA-Phe (GAA) 1-4', other_designations='tRNA-Phe-GAA-1-4|transfer RNA phenylalanine 10 (anticodon GAA)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1383362, end_position_on_the_genomic_accession=1383434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189307, CurrentID=0, Status='live', Symbol='TRH-GTG1-3', Aliases='TRNAH9', description='transfer RNA-His (GTG) 1-3', other_designations='tRNA-His-GTG-1-3|transfer RNA histidine 9 (anticodon GUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148281365, end_position_on_the_genomic_accession=148281436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189308, CurrentID=0, Status='live', Symbol='TRA-CGC1-1', Aliases='TRNAA29', description='transfer RNA-Ala (CGC) 1-1', other_designations='tRNA-Ala-CGC-1-1|transfer RNA alanine 29 (anticodon CGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26553503, end_position_on_the_genomic_accession=26553574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480765, CurrentID=0, Status='live', Symbol='RNA5SP370', Aliases='RN5S370', description='RNA, 5S ribosomal pseudogene 370', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189310, CurrentID=0, Status='live', Symbol='TRI-AAT5-5', Aliases='TRNAI17', description='transfer RNA-Ile (AAT) 5-5', other_designations='tRNA-Ile-AAT-5-5|transfer RNA isoleucine 17 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27237571, end_position_on_the_genomic_accession=27237644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189311, CurrentID=0, Status='live', Symbol='TRK-CTT11-1', Aliases='TRNAK28', description='transfer RNA-Lys (CTT) 11-1', other_designations='tRNA-Lys-CTT-11-1|transfer RNA lysine 28 (anticodon CUU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189312, CurrentID=0, Status='live', Symbol='TRA-AGC16-1', Aliases='TRNAA30', description='transfer RNA-Ala (AGC) 16-1', other_designations='tRNA-Ala-AGC-16-1|transfer RNA alanine 30 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57870345, end_position_on_the_genomic_accession=57870417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189313, CurrentID=0, Status='live', Symbol='TRL-TAA1-1', Aliases='TRNAL29', description='transfer RNA-Leu (TAA) 1-1', other_designations='tRNA-Leu-TAA-1-1|transfer RNA leucine 29 (anticodon UAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144216547, end_position_on_the_genomic_accession=144216629, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189314, CurrentID=0, Status='live', Symbol='TRG-GCC2-6', Aliases='TRNAG29', description='transfer RNA-Gly (GCC) 2-6', other_designations='tRNA-Gly-GCC-2-6|transfer RNA glycine 29 (anticodon GCC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27902908, end_position_on_the_genomic_accession=27902978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189315, CurrentID=0, Status='live', Symbol='TRX-CAT1-8', Aliases='TRNAM17', description='transfer RNA-iMet (CAT) 1-8', other_designations='tRNA-iMet-CAT-1-8|transfer RNA initiator methionine (CAT) 1-8|transfer RNA methionine 17 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27902493, end_position_on_the_genomic_accession=27902564, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189316, CurrentID=0, Status='live', Symbol='TRL-CAG1-3', Aliases='TRNAL30', description='transfer RNA-Leu (CAG) 1-3', other_designations='tRNA-Leu-CAG-1-3|transfer RNA leucine 30 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161456332, end_position_on_the_genomic_accession=161456414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189317, CurrentID=0, Status='live', Symbol='TRQ-CTG1-1', Aliases='TRNAQ24', description='transfer RNA-Gln (CTG) 1-1', other_designations='tRNA-Gln-CTG-1-1|transfer RNA glutamine 24 (anticodon CUG)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65869062, end_position_on_the_genomic_accession=65869133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189318, CurrentID=0, Status='live', Symbol='TRV-AAC1-1', Aliases='TRNAV24', description='transfer RNA-Val (AAC) 1-1', other_designations='tRNA-Val-AAC-1-1|transfer RNA valine 24 (anticodon AAC)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169772230, end_position_on_the_genomic_accession=169772302, orientation='plus', exon_count=0, OMIM=615310, no_hdr0=''),
  100189319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189319, CurrentID=0, Status='live', Symbol='TRT-AGT2-1', Aliases='TRNAT14', description='transfer RNA-Thr (AGT) 2-1', other_designations='tRNA-Thr-AGT-2-1|transfer RNA threonine 14 (anticodon AGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26532917, end_position_on_the_genomic_accession=26532990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189320, CurrentID=0, Status='live', Symbol='TRN-ATT1-1', Aliases='TRNAN26', description='transfer RNA-Asn (ATT) 1-1', other_designations='tRNA-Asn-ATT-1-1|transfer RNA asparagine 26 (anticodon AUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189321, CurrentID=0, Status='live', Symbol='TRQ-TTG3-1', Aliases='TRNAQ25', description='transfer RNA-Gln (TTG) 3-1', other_designations='tRNA-Gln-TTG-3-1|transfer RNA glutamine 25 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26311196, end_position_on_the_genomic_accession=26311267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189322, CurrentID=0, Status='live', Symbol='TRQ-CTG10-1', Aliases='TRNAQ26', description='transfer RNA-Gln (CTG) 10-1', other_designations='tRNA-Gln-CTG-10-1|transfer RNA glutamine 26 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189323, CurrentID=0, Status='live', Symbol='TRN-GTT3-1', Aliases='TRNAN27', description='transfer RNA-Asn (GTT) 3-1', other_designations='tRNA-Asn-GTT-3-1|transfer RNA asparagine 27 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144419267, end_position_on_the_genomic_accession=144419340, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480780, CurrentID=0, Status='live', Symbol='WBP1LP5', Aliases='', description='WW domain binding protein 1-like pseudogene 5', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189325, CurrentID=0, Status='live', Symbol='NMTRL-TAA1-1', Aliases='TRNAL31', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 1-1', other_designations='nmt-tRNA-Leu-TAA-1-1|transfer RNA leucine 31 (anticodon UAA)', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189326, CurrentID=0, Status='live', Symbol='TRV-CAC4-1', Aliases='TRNAV25', description='transfer RNA-Val (CAC) 4-1', other_designations='tRNA-Val-CAC-4-1|transfer RNA valine 25 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143803994, end_position_on_the_genomic_accession=143804066, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189327, CurrentID=0, Status='live', Symbol='TRA-AGC12-2', Aliases='TRNAA31', description='transfer RNA-Ala (AGC) 12-2', other_designations='tRNA-Ala-AGC-12-2|transfer RNA alanine 31 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189328, CurrentID=0, Status='live', Symbol='TRL-AAG2-3', Aliases='TRNAL32', description='transfer RNA-Leu (AAG) 2-3', other_designations='tRNA-Leu-AAG-2-3|transfer RNA leucine 32 (anticodon AAG)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181187701, end_position_on_the_genomic_accession=181187782, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189329, CurrentID=0, Status='live', Symbol='TRQ-CTG5-1', Aliases='TRNAQ27', description='transfer RNA-Gln (CTG) 5-1', other_designations='tRNA-Gln-CTG-5-1|transfer RNA glutamine 27 (anticodon CUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27295433, end_position_on_the_genomic_accession=27295504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189330, CurrentID=0, Status='live', Symbol='TRSUP-TTA2-1', Aliases='TRNASUP2', description='transfer RNA suppressor (TTA) 2-1', other_designations='tRNA-Sup-TTA-2-1|transfer RNA suppressor 2 (anticodon UUA)', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189331, CurrentID=0, Status='live', Symbol='TRS-GCT4-2', Aliases='TRNAS23', description='transfer RNA-Ser (GCT) 4-2', other_designations='tRNA-Ser-GCT-4-2|transfer RNA serine 23 (anticodon GCU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8186866, end_position_on_the_genomic_accession=8186947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480788, CurrentID=0, Status='live', Symbol='CST9LP2', Aliases='', description='cystatin 9-like pseudogene 2', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189333, CurrentID=0, Status='live', Symbol='TRE-TTC1-1', Aliases='TRNAE20', description='transfer RNA-Glu (TTC) 1-1', other_designations='tRNA-Glu-TTC-1-1|transfer RNA glutamic acid 20 (anticodon UUC)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41060738, end_position_on_the_genomic_accession=41060809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189334, CurrentID=0, Status='live', Symbol='TRH-GTG1-2', Aliases='TRNAH10', description='transfer RNA-His (GTG) 1-2', other_designations='tRNA-His-GTG-1-2|transfer RNA histidine 10 (anticodon GUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147073225, end_position_on_the_genomic_accession=147073296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189335, CurrentID=0, Status='live', Symbol='TRP-TGG3-2', Aliases='TRNAP19', description='transfer RNA-Pro (TGG) 3-2', other_designations='tRNA-Pro-TGG-3-2|transfer RNA proline 19 (anticodon UGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3158922, end_position_on_the_genomic_accession=3158993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189336, CurrentID=0, Status='live', Symbol='TRI-AAT5-4', Aliases='TRNAI20', description='transfer RNA-Ile (AAT) 5-4', other_designations='tRNA-Ile-AAT-5-4|transfer RNA isoleucine 20 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27177215, end_position_on_the_genomic_accession=27177288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480793, CurrentID=0, Status='live', Symbol='CRIP1P1', Aliases='', description='cysteine-rich protein 1 (intestinal) pseudogene 1', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189338, CurrentID=0, Status='live', Symbol='TRN-GTT16-5', Aliases='TRNAN28', description='transfer RNA-Asn (GTT) 16-5', other_designations='tRNA-Asn-GTT-16-5|transfer RNA asparagine 28 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189339, CurrentID=0, Status='live', Symbol='TRA-TGC4-1', Aliases='TRNAA32', description='transfer RNA-Ala (TGC) 4-1', other_designations='tRNA-Ala-TGC-4-1|transfer RNA alanine 32 (anticodon UGC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124939966, end_position_on_the_genomic_accession=124940037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189340, CurrentID=0, Status='live', Symbol='TRT-AGT2-2', Aliases='TRNAT15', description='transfer RNA-Thr (AGT) 2-2', other_designations='tRNA-Thr-AGT-2-2|transfer RNA threonine 15 (anticodon AGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27684695, end_position_on_the_genomic_accession=27684768, orientation='minus', exon_count=0, OMIM=615309, no_hdr0=''),
  100189341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189341, CurrentID=0, Status='live', Symbol='TRA-AGC18-1', Aliases='TRNAA33', description='transfer RNA-Ala (AGC) 18-1', other_designations='tRNA-Ala-AGC-18-1|transfer RNA alanine 33 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189342, CurrentID=0, Status='live', Symbol='TRV-TAC3-1', Aliases='TRNAV26', description='transfer RNA-Val (TAC) 3-1', other_designations='tRNA-Val-TAC-3-1|transfer RNA valine 26 (anticodon UAC)', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5853711, end_position_on_the_genomic_accession=5853783, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189343, CurrentID=0, Status='live', Symbol='TRD-GTC2-7', Aliases='TRNAD15', description='transfer RNA-Asp (GTC) 2-7', other_designations='tRNA-Asp-GTC-2-7|transfer RNA aspartic acid 15 (anticodon GUC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124927345, end_position_on_the_genomic_accession=124927416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189344, CurrentID=0, Status='live', Symbol='TRT-TGT1-1', Aliases='TRNAT16', description='transfer RNA-Thr (TGT) 1-1', other_designations='tRNA-Thr-TGT-1-1|transfer RNA threonine 16 (anticodon UGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28474552, end_position_on_the_genomic_accession=28474625, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189345, CurrentID=0, Status='live', Symbol='TRX-CAT1-6', Aliases='TRNAM18', description='transfer RNA-iMet (CAT) 1-6', other_designations='tRNA-iMet-CAT-1-6|transfer RNA initiator methionine (CAT) 1-6|transfer RNA methionine 18 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27332985, end_position_on_the_genomic_accession=27333056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189346, CurrentID=0, Status='live', Symbol='TRA-AGC1-1', Aliases='TRNAA34', description='transfer RNA-Ala (AGC) 1-1', other_designations='tRNA-Ala-AGC-1-1|transfer RNA alanine 34 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28795964, end_position_on_the_genomic_accession=28796035, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189347, CurrentID=0, Status='live', Symbol='TRD-GTC2-8', Aliases='TRNAD16', description='transfer RNA-Asp (GTC) 2-8', other_designations='tRNA-Asp-GTC-2-8|transfer RNA aspartic acid 16 (anticodon GUC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124939647, end_position_on_the_genomic_accession=124939718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480804, CurrentID=0, Status='live', Symbol='TPRKBP2', Aliases='', description='TP53RK binding protein pseudogene 2', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189349, CurrentID=0, Status='live', Symbol='TRL-AAG8-1', Aliases='TRNAL33', description='transfer RNA-Leu (AAG) 8-1', other_designations='tRNA-Leu-AAG-8-1|transfer RNA leucine 33 (anticodon AAG)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189350, CurrentID=0, Status='live', Symbol='TRH-GTG2-1', Aliases='TRNAH11', description='transfer RNA-His (GTG) 2-1', other_designations='tRNA-His-GTG-2-1|transfer RNA histidine 11 (anticodon GUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=149155828, end_position_on_the_genomic_accession=149155899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189351, CurrentID=0, Status='live', Symbol='TRQ-CTG18-1', Aliases='TRNAQ28', description='transfer RNA-Gln (CTG) 18-1', other_designations='tRNA-Gln-CTG-18-1|transfer RNA glutamine 28 (anticodon CUG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189352, CurrentID=0, Status='live', Symbol='TRS-AGA1-1', Aliases='TRNAS24', description='transfer RNA-Ser (AGA) 1-1', other_designations='tRNA-Ser-AGA-1-1|transfer RNA serine 24 (anticodon AGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27541775, end_position_on_the_genomic_accession=27541856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189353, CurrentID=0, Status='live', Symbol='TRA-AGC17-1', Aliases='TRNAA35', description='transfer RNA-Ala (AGC) 17-1', other_designations='tRNA-Ala-AGC-17-1|transfer RNA alanine 35 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189354, CurrentID=0, Status='live', Symbol='TRM-CAT4-2', Aliases='TRNAM19', description='transfer RNA-Met (CAT) 4-2', other_designations='tRNA-Met-CAT-4-2|transfer RNA methionine 19 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26743263, end_position_on_the_genomic_accession=26743335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189355, CurrentID=0, Status='live', Symbol='TRC-GCA20-1', Aliases='TRNAC22', description='transfer RNA-Cys (GCA) 20-1', other_designations='tRNA-Cys-GCA-20-1|transfer RNA cysteine 22 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149597955, end_position_on_the_genomic_accession=149598026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189356, CurrentID=0, Status='live', Symbol='TRN-GTT12-1', Aliases='TRNAN29', description='transfer RNA-Asn (GTT) 12-1', other_designations='tRNA-Asn-GTT-12-1|transfer RNA asparagine 29 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143831708, end_position_on_the_genomic_accession=143831781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189357, CurrentID=0, Status='live', Symbol='TRC-GCA13-1', Aliases='TRNAC23', description='transfer RNA-Cys (GCA) 13-1', other_designations='tRNA-Cys-GCA-13-1|transfer RNA cysteine 23 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149355675, end_position_on_the_genomic_accession=149355746, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  132612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132612, CurrentID=0, Status='live', Symbol='ADAD1', Aliases='Tenr', description='adenosine deaminase domain containing 1', other_designations='adenosine deaminase domain containing 1 (testis-specific)|testis nuclear RNA-binding protein', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122378966, end_position_on_the_genomic_accession=122429815, orientation='plus', exon_count=13, OMIM=614130, no_hdr0=''),
  100189359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189359, CurrentID=0, Status='live', Symbol='TRC-GCA8-1', Aliases='TRNAC24', description='transfer RNA-Cys (GCA) 8-1', other_designations='tRNA-Cys-GCA-8-1|transfer RNA cysteine 24 (anticodon GCA)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72962971, end_position_on_the_genomic_accession=72963042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189360, CurrentID=0, Status='live', Symbol='TRT-CGT3-1', Aliases='TRNAT17', description='transfer RNA-Thr (CGT) 3-1', other_designations='tRNA-Thr-CGT-3-1|transfer RNA threonine 17 (anticodon CGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28648207, end_position_on_the_genomic_accession=28648280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480817, CurrentID=0, Status='live', Symbol='RNU7-120P', Aliases='', description='RNA, U7 small nuclear 120 pseudogene', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189362, CurrentID=0, Status='live', Symbol='TRL-AAG1-1', Aliases='TRNAL34', description='transfer RNA-Leu (AAG) 1-1', other_designations='tRNA-Leu-AAG-1-1|transfer RNA leucine 34 (anticodon AAG)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181097474, end_position_on_the_genomic_accession=181097555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189363, CurrentID=0, Status='live', Symbol='TRV-CAC2-1', Aliases='TRNAV27', description='transfer RNA-Val (CAC) 2-1', other_designations='tRNA-Val-CAC-2-1|transfer RNA valine 27 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27280270, end_position_on_the_genomic_accession=27280342, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189364, CurrentID=0, Status='live', Symbol='TRL-CAG1-4', Aliases='TRNAL35', description='transfer RNA-Leu (CAG) 1-4', other_designations='tRNA-Leu-CAG-1-4|transfer RNA leucine 35 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161463742, end_position_on_the_genomic_accession=161463824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189365, CurrentID=0, Status='live', Symbol='TRR-CCG1-2', Aliases='TRNAR24', description='transfer RNA-Arg (CCG) 1-2', other_designations='tRNA-Arg-CCG-1-2|transfer RNA arginine 24 (anticodon CCG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28742952, end_position_on_the_genomic_accession=28743024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189366, CurrentID=0, Status='live', Symbol='TRG-GCC1-1', Aliases='TRNAG30', description='transfer RNA-Gly (GCC) 1-1', other_designations='tRNA-Gly-GCC-1-1|transfer RNA glycine 30 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161443304, end_position_on_the_genomic_accession=161443374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189367, CurrentID=0, Status='live', Symbol='TRK-CTT10-1', Aliases='TRNAK30', description='transfer RNA-Lys (CTT) 10-1', other_designations='tRNA-Lys-CTT-10-1|transfer RNA lysine 30 (anticodon CUU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189368, CurrentID=0, Status='live', Symbol='TRV-CAC6-1', Aliases='TRNAV28', description='transfer RNA-Val (CAC) 6-1', other_designations='tRNA-Val-CAC-6-1|transfer RNA valine 28 (anticodon CAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27206088, end_position_on_the_genomic_accession=27206160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189369, CurrentID=0, Status='live', Symbol='TRA-AGC12-1', Aliases='TRNAA37', description='transfer RNA-Ala (AGC) 12-1', other_designations='tRNA-Ala-AGC-12-1|transfer RNA alanine 37 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26819109, end_position_on_the_genomic_accession=26819181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189370, CurrentID=0, Status='live', Symbol='TRL-CAA5-1', Aliases='TRNAL36', description='transfer RNA-Leu (CAA) 5-1', other_designations='tRNA-Leu-CAA-5-1|transfer RNA leucine 36 (anticodon CAA)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9275243, end_position_on_the_genomic_accession=9275316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189371, CurrentID=0, Status='live', Symbol='TRR-TCG3-1', Aliases='TRNAR25', description='transfer RNA-Arg (TCG) 3-1', other_designations='tRNA-Arg-TCG-3-1|transfer RNA arginine 25 (anticodon UCG)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75035113, end_position_on_the_genomic_accession=75035185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189372, CurrentID=0, Status='live', Symbol='TRI-AAT6-1', Aliases='TRNAI22', description='transfer RNA-Ile (AAT) 6-1', other_designations='tRNA-Ile-AAT-6-1|transfer RNA isoleucine 22 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26756552, end_position_on_the_genomic_accession=26756625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189373, CurrentID=0, Status='live', Symbol='TRV-CAC1-1', Aliases='TRNAV29', description='transfer RNA-Val (CAC) 1-1', other_designations='tRNA-Val-CAC-1-1|transfer RNA valine 29 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161399700, end_position_on_the_genomic_accession=161399772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189374, CurrentID=0, Status='live', Symbol='TRT-AGT5-1', Aliases='TRNAT18', description='transfer RNA-Thr (AGT) 5-1', other_designations='tRNA-Thr-AGT-5-1|transfer RNA threonine 18 (anticodon AGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27726694, end_position_on_the_genomic_accession=27726767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189375, CurrentID=0, Status='live', Symbol='TRC-GCA21-1', Aliases='TRNAC25', description='transfer RNA-Cys (GCA) 21-1', other_designations='tRNA-Cys-GCA-21-1|transfer RNA cysteine 25 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149664824, end_position_on_the_genomic_accession=149664895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189376, CurrentID=0, Status='live', Symbol='TRC-GCA9-3', Aliases='TRNAC26', description='transfer RNA-Cys (GCA) 9-3', other_designations='tRNA-Cys-GCA-9-3|transfer RNA cysteine 26 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149635687, end_position_on_the_genomic_accession=149635758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189377, CurrentID=0, Status='live', Symbol='NMTRL-TAA4-1', Aliases='TRNAL37', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 4-1', other_designations='nmt-tRNA-Leu-TAA-4-1|transfer RNA leucine 37 (anticodon UAA)', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189378, CurrentID=0, Status='live', Symbol='TRT-TGT2-1', Aliases='TRNAT19', description='transfer RNA-Thr (TGT) 2-1', other_designations='tRNA-Thr-TGT-2-1|transfer RNA threonine 19 (anticodon UGU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222465005, end_position_on_the_genomic_accession=222465077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189379, CurrentID=0, Status='live', Symbol='TRC-GCA9-4', Aliases='TRNAC27', description='transfer RNA-Cys (GCA) 9-4', other_designations='tRNA-Cys-GCA-9-4|transfer RNA cysteine 27 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149707669, end_position_on_the_genomic_accession=149707740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189380, CurrentID=0, Status='live', Symbol='TRG-GCC2-1', Aliases='TRNAG31', description='transfer RNA-Gly (GCC) 2-1', other_designations='tRNA-Gly-GCC-2-1|transfer RNA glycine 31 (anticodon GCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161523847, end_position_on_the_genomic_accession=161523917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189381, CurrentID=0, Status='live', Symbol='TRN-GTT19-1', Aliases='TRNAN30', description='transfer RNA-Asn (GTT) 19-1', other_designations='tRNA-Asn-GTT-19-1|transfer RNA asparagine 30 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189382, CurrentID=0, Status='live', Symbol='TRS-AGA2-2', Aliases='TRNAS25', description='transfer RNA-Ser (AGA) 2-2', other_designations='tRNA-Ser-AGA-2-2|transfer RNA serine 25 (anticodon AGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26327589, end_position_on_the_genomic_accession=26327670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189383, CurrentID=0, Status='live', Symbol='TRD-GTC2-9', Aliases='TRNAD17', description='transfer RNA-Asp (GTC) 2-9', other_designations='tRNA-Asp-GTC-2-9|transfer RNA aspartic acid 17 (anticodon GUC)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8222238, end_position_on_the_genomic_accession=8222309, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189384, CurrentID=0, Status='live', Symbol='TRE-CTC1-1', Aliases='TRNAE21', description='transfer RNA-Glu (CTC) 1-1', other_designations='tRNA-Glu-CTC-1-1|transfer RNA glutamic acid 21 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146035692, end_position_on_the_genomic_accession=146035763, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  8396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8396, CurrentID=0, Status='live', Symbol='PIP4K2B', Aliases='PI5P4KB, PIP5K2B, PIP5KIIB, PIP5KIIbeta, PIP5P4KB', description='phosphatidylinositol-5-phosphate 4-kinase, type II, beta', other_designations='1-phosphatidylinositol 5-phosphate 4-kinase 2-beta|1-phosphatidylinositol-4-phosphate kinase|1-phosphatidylinositol-5-phosphate 4-kinase 2-beta|PI(5)P 4-kinase type II beta|PIP4KII-beta|PTDINS(4)P-5-kinase|diphosphoinositide kinase 2-beta|phosphatidylinositol 5-phosphate 4-kinase type II beta|phosphatidylinositol-4-phosphate 5-kinase, type II, beta|phosphatidylinositol-5-phosphate 4-kinase type-2 beta|ptdIns(5)P-4-kinase isoform 2-beta', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38765689, end_position_on_the_genomic_accession=38799965, orientation='minus', exon_count=15, OMIM=603261, no_hdr0=''),
  100189386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189386, CurrentID=0, Status='live', Symbol='TRQ-CTG3-1', Aliases='TRNAQ29', description='transfer RNA-Gln (CTG) 3-1', other_designations='tRNA-Gln-CTG-3-1|transfer RNA glutamine 29 (anticodon CUG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=145963304, end_position_on_the_genomic_accession=145963375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189387, CurrentID=0, Status='live', Symbol='TRI-AAT5-2', Aliases='TRNAI23', description='transfer RNA-Ile (AAT) 5-2', other_designations='tRNA-Ile-AAT-5-2|transfer RNA isoleucine 23 (anticodon AAU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8187593, end_position_on_the_genomic_accession=8187666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189388, CurrentID=0, Status='live', Symbol='TRP-TGG3-3', Aliases='TRNAP21', description='transfer RNA-Pro (TGG) 3-3', other_designations='tRNA-Pro-TGG-3-3|transfer RNA proline 21 (anticodon UGG)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3184133, end_position_on_the_genomic_accession=3184204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189389, CurrentID=0, Status='live', Symbol='TRQ-CTG12-1', Aliases='TRNAQ30', description='transfer RNA-Gln (CTG) 12-1', other_designations='tRNA-Gln-CTG-12-1|transfer RNA glutamine 30 (anticodon CUG)', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189390, CurrentID=0, Status='live', Symbol='TRK-TTT9-1', Aliases='TRNAK31', description='transfer RNA-Lys (TTT) 9-1', other_designations='tRNA-Lys-TTT-9-1|transfer RNA lysine 31 (anticodon UUU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27576067, end_position_on_the_genomic_accession=27576139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189391, CurrentID=0, Status='live', Symbol='TRM-CAT4-3', Aliases='TRNAM20', description='transfer RNA-Met (CAT) 4-3', other_designations='tRNA-Met-CAT-4-3|transfer RNA methionine 20 (anticodon CAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26735370, end_position_on_the_genomic_accession=26735442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189392, CurrentID=0, Status='live', Symbol='TRK-TTT16-1', Aliases='TRNAK32', description='transfer RNA-Lys (TTT) 16-1', other_designations='tRNA-Lys-TTT-16-1|transfer RNA lysine 32 (anticodon UUU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189393, CurrentID=0, Status='live', Symbol='TRI-AAT4-1', Aliases='TRNAI24', description='transfer RNA-Ile (AAT) 4-1', other_designations='tRNA-Ile-AAT-4-1|transfer RNA isoleucine 24 (anticodon AAU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8226991, end_position_on_the_genomic_accession=8227064, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189394, CurrentID=0, Status='live', Symbol='TRL-CAG1-1', Aliases='TRNAL38', description='transfer RNA-Leu (CAG) 1-1', other_designations='tRNA-Leu-CAG-1-1|transfer RNA leucine 38 (anticodon CAG)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161441533, end_position_on_the_genomic_accession=161441615, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189395, CurrentID=0, Status='live', Symbol='TRA-CGC5-1', Aliases='TRNAA38', description='transfer RNA-Ala (CGC) 5-1', other_designations='tRNA-Ala-CGC-5-1|transfer RNA alanine 38 (anticodon CGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28695932, end_position_on_the_genomic_accession=28696003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189396, CurrentID=0, Status='live', Symbol='TRR-ACG1-2', Aliases='TRNAR26', description='transfer RNA-Arg (ACG) 1-2', other_designations='tRNA-Arg-ACG-1-2|transfer RNA arginine 26 (anticodon ACG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26328140, end_position_on_the_genomic_accession=26328212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189397, CurrentID=0, Status='live', Symbol='NMTRQ-TTG5-1', Aliases='TRNAQ31', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 5-1', other_designations='nmt-tRNA-Gln-TTG-5-1|transfer RNA glutamine 31 (anticodon UUG)|transfer RNA-Gln (TTG) 5-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189398, CurrentID=0, Status='live', Symbol='TRA-AGC8-2', Aliases='TRNAA39', description='transfer RNA-Ala (AGC) 8-2', other_designations='tRNA-Ala-AGC-8-2|transfer RNA alanine 39 (anticodon AGC)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66114189, end_position_on_the_genomic_accession=66114261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189399, CurrentID=0, Status='live', Symbol='TRK-CTT2-3', Aliases='TRNAK33', description='transfer RNA-Lys (CTT) 2-3', other_designations='tRNA-Lys-CTT-2-3|transfer RNA lysine 33 (anticodon CUU)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181207755, end_position_on_the_genomic_accession=181207827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189400, CurrentID=0, Status='live', Symbol='TRA-CGC2-1', Aliases='TRNAA40', description='transfer RNA-Ala (CGC) 2-1', other_designations='tRNA-Ala-CGC-2-1|transfer RNA alanine 40 (anticodon CGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28673836, end_position_on_the_genomic_accession=28673907, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189401, CurrentID=0, Status='live', Symbol='TRI-AAT7-2', Aliases='TRNAI25', description='transfer RNA-Ile (AAT) 7-2', other_designations='tRNA-Ile-AAT-7-2|transfer RNA isoleucine 25 (anticodon AAU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57822973, end_position_on_the_genomic_accession=57823046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189402, CurrentID=0, Status='live', Symbol='TRN-GTT2-2', Aliases='TRNAN32', description='transfer RNA-Asn (GTT) 2-2', other_designations='tRNA-Asn-GTT-2-2|transfer RNA asparagine 32 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161428077, end_position_on_the_genomic_accession=161428150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480859, CurrentID=0, Status='live', Symbol='RN7SKP86', Aliases='', description='RNA, 7SK small nuclear pseudogene 86', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189404, CurrentID=0, Status='live', Symbol='TRV-AAC7-1', Aliases='TRNAV30', description='transfer RNA-Val (AAC) 7-1', other_designations='tRNA-Val-AAC-7-1|transfer RNA valine 30 (anticodon AAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480861, CurrentID=0, Status='live', Symbol='RN7SKP91', Aliases='', description='RNA, 7SK small nuclear pseudogene 91', other_designations='', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189406, CurrentID=0, Status='live', Symbol='TRR-CCT4-1', Aliases='TRNAR27', description='transfer RNA-Arg (CCT) 4-1', other_designations='tRNA-Arg-CCT-4-1|transfer RNA arginine 27 (anticodon CCU)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139340700, end_position_on_the_genomic_accession=139340772, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189407, CurrentID=0, Status='live', Symbol='TRE-TTC4-2', Aliases='TRNAE23', description='transfer RNA-Glu (TTC) 4-2', other_designations='tRNA-Glu-TTC-4-2|transfer RNA glutamic acid 23 (anticodon UUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161422093, end_position_on_the_genomic_accession=161422164, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189408, CurrentID=0, Status='live', Symbol='TRV-TAC4-1', Aliases='TRNAV31', description='transfer RNA-Val (TAC) 4-1', other_designations='tRNA-Val-TAC-4-1|transfer RNA valine 31 (anticodon UAC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27290626, end_position_on_the_genomic_accession=27290698, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189409, CurrentID=0, Status='live', Symbol='TRE-CTC2-1', Aliases='TRNAE24', description='transfer RNA-Glu (CTC) 2-1', other_designations='tRNA-Glu-CTC-2-1|transfer RNA glutamic acid 24 (anticodon CUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248874248, end_position_on_the_genomic_accession=248874319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189410, CurrentID=0, Status='live', Symbol='TRR-TCG2-1', Aliases='TRNAR28', description='transfer RNA-Arg (TCG) 2-1', other_designations='tRNA-Arg-TCG-2-1|transfer RNA arginine 28 (anticodon UCG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26322818, end_position_on_the_genomic_accession=26322890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189411, CurrentID=0, Status='live', Symbol='TRA-AGC13-1', Aliases='TRNAA41', description='transfer RNA-Ala (AGC) 13-1', other_designations='tRNA-Ala-AGC-13-1|transfer RNA alanine 41 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26796209, end_position_on_the_genomic_accession=26796281, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189412, CurrentID=0, Status='live', Symbol='TRD-GTC2-1', Aliases='TRNAD18', description='transfer RNA-Asp (GTC) 2-1', other_designations='tRNA-Asp-GTC-2-1|transfer RNA aspartic acid 18 (anticodon GUC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161440825, end_position_on_the_genomic_accession=161440896, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189413, CurrentID=0, Status='live', Symbol='TRC-GCA9-2', Aliases='TRNAC28', description='transfer RNA-Cys (GCA) 9-2', other_designations='tRNA-Cys-GCA-9-2|transfer RNA cysteine 28 (anticodon GCA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149331129, end_position_on_the_genomic_accession=149331200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189414, CurrentID=0, Status='live', Symbol='TRC-GCA2-3', Aliases='TRNAC29', description='transfer RNA-Cys (GCA) 2-3', other_designations='tRNA-Cys-GCA-2-3|transfer RNA cysteine 29 (anticodon GCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39154491, end_position_on_the_genomic_accession=39154562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189415, CurrentID=0, Status='live', Symbol='TRF-GAA1-1', Aliases='TRNAF11', description='transfer RNA-Phe (GAA) 1-1', other_designations='tRNA-Phe-GAA-1-1|transfer RNA phenylalanine 11 (anticodon GAA)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59557497, end_position_on_the_genomic_accession=59557569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189416, CurrentID=0, Status='live', Symbol='TRV-CAC3-1', Aliases='TRNAV32', description='transfer RNA-Val (CAC) 3-1', other_designations='tRNA-Val-CAC-3-1|transfer RNA valine 32 (anticodon CAC)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4724635, end_position_on_the_genomic_accession=4724707, orientation='minus', exon_count=0, OMIM=615304, no_hdr0=''),
  100189417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189417, CurrentID=0, Status='live', Symbol='TRE-CTC6-1', Aliases='TRNAE25', description='transfer RNA-Glu (CTC) 6-1', other_designations='tRNA-Glu-CTC-6-1|transfer RNA glutamic acid 25 (anticodon CUC)', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189418, CurrentID=0, Status='live', Symbol='TRE-TTC1-2', Aliases='TRNAE26', description='transfer RNA-Glu (TTC) 1-2', other_designations='tRNA-Glu-TTC-1-2|transfer RNA glutamic acid 26 (anticodon UUC)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130337128, end_position_on_the_genomic_accession=130337199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189419, CurrentID=0, Status='live', Symbol='TRT-TGT4-1', Aliases='TRNAT20', description='transfer RNA-Thr (TGT) 4-1', other_designations='tRNA-Thr-TGT-4-1|transfer RNA threonine 20 (anticodon UGU)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20631160, end_position_on_the_genomic_accession=20631232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189420, CurrentID=0, Status='live', Symbol='TRS-AGA6-1', Aliases='TRNAS26', description='transfer RNA-Ser (AGA) 6-1', other_designations='tRNA-Ser-AGA-6-1|transfer RNA serine 26 (anticodon AGA)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189421, CurrentID=0, Status='live', Symbol='TRF-GAA7-1', Aliases='TRNAF12', description='transfer RNA-Phe (GAA) 7-1', other_designations='tRNA-Phe-GAA-7-1|transfer RNA phenylalanine 12 (anticodon GAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189422, CurrentID=0, Status='live', Symbol='TRQ-TTG4-1', Aliases='TRNAQ32', description='transfer RNA-Gln (TTG) 4-1', other_designations='tRNA-Gln-TTG-4-1|transfer RNA glutamine 32 (anticodon UUG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145182723, end_position_on_the_genomic_accession=145182794, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189423, CurrentID=0, Status='live', Symbol='TRA-AGC9-2', Aliases='TRNAA42', description='transfer RNA-Ala (AGC) 9-2', other_designations='tRNA-Ala-AGC-9-2|transfer RNA alanine 42 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480880, CurrentID=0, Status='live', Symbol='RN7SKP150', Aliases='', description='RNA, 7SK small nuclear pseudogene 150', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189425, CurrentID=0, Status='live', Symbol='TRK-TTT3-2', Aliases='TRNAK34', description='transfer RNA-Lys (TTT) 3-2', other_designations='tRNA-Lys-TTT-3-2|transfer RNA lysine 34 (anticodon UUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204507030, end_position_on_the_genomic_accession=204507102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189426, CurrentID=0, Status='live', Symbol='TRSUP-CTA1-1', Aliases='TRNASUP3', description='transfer RNA suppressor (CTA) 1-1', other_designations='tRNA-Sup-CTA-1-1|transfer RNA suppressor 3 (anticodon CUA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189427, CurrentID=0, Status='live', Symbol='TRI-GAT1-3', Aliases='TRNAI26', description='transfer RNA-Ile (GAT) 1-3', other_designations='tRNA-Ile-GAT-1-3|transfer RNA isoleucine 26 (anticodon GAU)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3915230, end_position_on_the_genomic_accession=3915303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189428, CurrentID=0, Status='live', Symbol='TRR-ACG2-4', Aliases='TRNAR29', description='transfer RNA-Arg (ACG) 2-4', other_designations='tRNA-Arg-ACG-2-4|transfer RNA arginine 29 (anticodon ACG)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27670565, end_position_on_the_genomic_accession=27670637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189429, CurrentID=0, Status='live', Symbol='TRL-AAG6-1', Aliases='TRNAL39', description='transfer RNA-Leu (AAG) 6-1', other_designations='tRNA-Leu-AAG-6-1|transfer RNA leucine 39 (anticodon AAG)', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189430, CurrentID=0, Status='live', Symbol='TRS-TGA3-1', Aliases='TRNAS27', description='transfer RNA-Ser (TGA) 3-1', other_designations='tRNA-Ser-TGA-3-1|transfer RNA serine 27 (anticodon UGA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26312596, end_position_on_the_genomic_accession=26312677, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189431, CurrentID=0, Status='live', Symbol='TRD-GTC8-1', Aliases='TRNAD19', description='transfer RNA-Asp (GTC) 8-1', other_designations='tRNA-Asp-GTC-8-1|transfer RNA aspartic acid 19 (anticodon GUC)', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189432, CurrentID=0, Status='live', Symbol='TRS-AGA5-1', Aliases='TRNAS28', description='transfer RNA-Ser (AGA) 5-1', other_designations='tRNA-Ser-AGA-5-1|transfer RNA serine 28 (anticodon AGA)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149608376, end_position_on_the_genomic_accession=149608447, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189433, CurrentID=0, Status='live', Symbol='TRT-AGT3-1', Aliases='TRNAT21', description='transfer RNA-Thr (AGT) 3-1', other_designations='tRNA-Thr-AGT-3-1|transfer RNA threonine 21 (anticodon AGU)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28726018, end_position_on_the_genomic_accession=28726091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189434, CurrentID=0, Status='live', Symbol='TRA-AGC22-1', Aliases='TRNAA43', description='transfer RNA-Ala (AGC) 22-1', other_designations='tRNA-Ala-AGC-22-1|transfer RNA alanine 43 (anticodon AGC)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189435, CurrentID=0, Status='live', Symbol='TRC-GCA14-1', Aliases='TRNAC30', description='transfer RNA-Cys (GCA) 14-1', other_designations='tRNA-Cys-GCA-14-1|transfer RNA cysteine 30 (anticodon GCA)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38861684, end_position_on_the_genomic_accession=38861755, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189436, CurrentID=0, Status='live', Symbol='TRW-CCA3-2', Aliases='TRNAW9', description='transfer RNA-Trp (CCA) 3-2', other_designations='tRNA-Trp-CCA-3-2|transfer RNA tryptophan 9 (anticodon CCA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26319102, end_position_on_the_genomic_accession=26319173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480893, CurrentID=0, Status='live', Symbol='RN7SKP194', Aliases='', description='RNA, 7SK small nuclear pseudogene 194', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189438, CurrentID=0, Status='live', Symbol='TRT-AGT4-1', Aliases='TRNAT22', description='transfer RNA-Thr (AGT) 4-1', other_designations='tRNA-Thr-AGT-4-1|transfer RNA threonine 22 (anticodon AGU)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8139452, end_position_on_the_genomic_accession=8139525, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189439, CurrentID=0, Status='live', Symbol='TRK-CTT13-1', Aliases='TRNAK35P', description='transfer RNA-Lys (CTT) 13-1', other_designations='tRNA-Lys-CTT-13-1|transfer RNA lysine 35 (anticodon CUU) pseudogene', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165596913, end_position_on_the_genomic_accession=165596985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189440, CurrentID=0, Status='live', Symbol='TRE-CTC12-1', Aliases='TRNAE27P', description='transfer RNA-Glu (CTC) 12-1', other_designations='tRNA-Glu-CTC-12-1|transfer RNA glutamic acid 27 (anticodon CUC) pseudogene', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104160680, end_position_on_the_genomic_accession=104160751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189441, CurrentID=0, Status='live', Symbol='TRL-TAG4-1', Aliases='TRNAL40P', description='transfer RNA-Leu (TAG) 4-1', other_designations='tRNA-Leu-TAG-4-1|transfer RNA leucine 40 (anticodon UAG) pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20677019, end_position_on_the_genomic_accession=20677100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189442, CurrentID=0, Status='live', Symbol='TRSUP-CTA2-1', Aliases='TRNASUP4P', description='transfer RNA suppressor (CTA) 2-1', other_designations='tRNA-Sup-CTA-2-1|transfer RNA suppressor 4 (anticodon CUA) pseudogene', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13792389, end_position_on_the_genomic_accession=13792456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189443, CurrentID=0, Status='live', Symbol='TRR-CCT7-1', Aliases='TRNAR30P', description='transfer RNA-Arg (CCT) 7-1', other_designations='tRNA-Arg-CCT-7-1|transfer RNA arginine 30 (anticodon CCU) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143848514, end_position_on_the_genomic_accession=143848584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189444, CurrentID=0, Status='live', Symbol='TRQ-TTG7-1', Aliases='TRNAQ33P', description='transfer RNA-Gln (TTG) 7-1', other_designations='tRNA-Gln-TTG-7-1|transfer RNA glutamine 33 (anticodon UUG) pseudogene', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78022952, end_position_on_the_genomic_accession=78023024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189445, CurrentID=0, Status='live', Symbol='NMTRQ-TTG7-1', Aliases='TRNAQ34P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 7-1', other_designations='nmt-tRNA-Gln-TTG-7-1|transfer RNA glutamine 34 (anticodon UUG) pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130272338, end_position_on_the_genomic_accession=130272409, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189446, CurrentID=0, Status='live', Symbol='TRF-GAA8-1', Aliases='TRNAF13P', description='transfer RNA-Phe (GAA) 8-1', other_designations='tRNA-Phe-GAA-8-1|transfer RNA phenylalanine 13 (anticodon GAA) pseudogene', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=78958293, end_position_on_the_genomic_accession=78958365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189447, CurrentID=0, Status='live', Symbol='TRC-GCA25-1', Aliases='TRNAC31P', description='transfer RNA-Cys (GCA) 25-1', other_designations='tRNA-Cys-GCA-25-1|transfer RNA cysteine 31 (anticodon GCA) pseudogene', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17699900, end_position_on_the_genomic_accession=17699972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189448, CurrentID=0, Status='live', Symbol='NMTRQ-TTG8-1', Aliases='TRNAQ35P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 8-1', other_designations='nmt-tRNA-Gln-TTG-8-1|transfer RNA glutamine 35 (anticodon UUG) pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8372099, end_position_on_the_genomic_accession=8372173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189449, CurrentID=0, Status='live', Symbol='TRR-CCT6-2', Aliases='TRNAR31P, TRNAR33P', description='transfer RNA-Arg (CCT) 6-2', other_designations='tRNA-Arg-CCT-6-2|transfer RNA arginine 31 (anticodon CCU) pseudogene|transfer RNA arginine 33 (anticodon CCU) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148011059, end_position_on_the_genomic_accession=148011129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189450, CurrentID=0, Status='live', Symbol='TRE-CTC14-1', Aliases='TRNAE28P', description='transfer RNA-Glu (CTC) 14-1', other_designations='tRNA-Glu-CTC-14-1|transfer RNA glutamic acid 28 (anticodon CUC) pseudogene', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69160105, end_position_on_the_genomic_accession=69160179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189451, CurrentID=0, Status='live', Symbol='NMTRQ-TTG9-1', Aliases='TRNAQ36P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 9-1', other_designations='nmt-tRNA-Gln-TTG-9-1|transfer RNA glutamine 36 (anticodon UUG) pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131385560, end_position_on_the_genomic_accession=131385631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189452, CurrentID=0, Status='live', Symbol='TRQ-CTG11-1', Aliases='TRNAQ37P', description='transfer RNA-Gln (CTG) 11-1', other_designations='tRNA-Gln-CTG-11-1|transfer RNA glutamine 37 (anticodon CUG) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143584701, end_position_on_the_genomic_accession=143584772, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189453, CurrentID=0, Status='live', Symbol='TRSUP-TTA3-1', Aliases='TRNASUP5P', description='transfer RNA suppressor (TTA) 3-1', other_designations='tRNA-Sup-TTA-3-1|transfer RNA suppressor 5 (anticodon UUA) pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7324111, end_position_on_the_genomic_accession=7324182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189454, CurrentID=0, Status='live', Symbol='TRE-TTC10-1', Aliases='TRNAE29P', description='transfer RNA-Glu (TTC) 10-1', other_designations='tRNA-Glu-TTC-10-1|transfer RNA glutamic acid 29 (anticodon UUC) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145176839, end_position_on_the_genomic_accession=145177111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189455, CurrentID=0, Status='live', Symbol='TRP-GGG1-1', Aliases='TRNAP22P', description='transfer RNA-Pro (GGG) 1-1', other_designations='tRNA-Pro-GGG-1-1|transfer RNA proline 22 (anticodon GGG) pseudogene', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22563650, end_position_on_the_genomic_accession=22563722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189456, CurrentID=0, Status='live', Symbol='NMTRL-TAA6-1', Aliases='TRNAL41P', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 6-1', other_designations='nmt-tRNA-Leu-TAA-6-1|transfer RNA leucine 41 (anticodon UAA) pseudogene|transfer RNA-Leu (TAA) 6-1', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46828416, end_position_on_the_genomic_accession=46828490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189457, CurrentID=0, Status='live', Symbol='TRE-TTC7-1', Aliases='TRNAE30P', description='transfer RNA-Glu (TTC) 7-1', other_designations='tRNA-Glu-TTC-7-1|transfer RNA glutamic acid 30 (anticodon UUC) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143666988, end_position_on_the_genomic_accession=143667059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189458, CurrentID=0, Status='live', Symbol='TRG-CCC7-1', Aliases='TRNAG32P', description='transfer RNA-Gly (CCC) 7-1', other_designations='tRNA-Gly-CCC-7-1|transfer RNA glycine 32 (anticodon CCC) pseudogene', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11551701, end_position_on_the_genomic_accession=11551772, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189459, CurrentID=0, Status='live', Symbol='TRE-CTC13-1', Aliases='TRNAE31P', description='transfer RNA-Glu (CTC) 13-1', other_designations='tRNA-Glu-CTC-13-1|transfer RNA glutamic acid 31 (anticodon CUC) pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95551446, end_position_on_the_genomic_accession=95551517, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189460, CurrentID=0, Status='live', Symbol='TRG-TCC5-1', Aliases='TRNAG33P', description='transfer RNA-Gly (TCC) 5-1', other_designations='tRNA-Gly-TCC-5-1|transfer RNA glycine 33 (anticodon UCC) pseudogene', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57678622, end_position_on_the_genomic_accession=57678693, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189461, CurrentID=0, Status='live', Symbol='NMTRQ-CTG1-1', Aliases='TRNAQ38P', description='nuclear-encoded mitochondrial transfer RNA-Gln (CTG) 1-1', other_designations='nmt-tRNA-Gln-CTG-1-1|transfer RNA glutamine 38 (anticodon CUG) pseudogene', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19747679, end_position_on_the_genomic_accession=19747753, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189462, CurrentID=0, Status='live', Symbol='TRI-AAT10-1', Aliases='TRNAI28P', description='transfer RNA-Ile (AAT) 10-1', other_designations='tRNA-Ile-AAT-10-1|transfer RNA isoleucine 28 (anticodon AAU) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27284085, end_position_on_the_genomic_accession=27284158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189463, CurrentID=0, Status='live', Symbol='NMTRP-TGG1-1', Aliases='TRNAP23P', description='nuclear-encoded mitochondrial transfer RNA-Pro (TGG) 1-1', other_designations='nmt-tRNA-Pro-TGG-1-1|transfer RNA proline 23 (anticodon UGG) pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237942562, end_position_on_the_genomic_accession=237942633, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189464, CurrentID=0, Status='live', Symbol='TRK-CTT12-1', Aliases='TRNAK36P', description='transfer RNA-Lys (CTT) 12-1', other_designations='tRNA-Lys-CTT-12-1|transfer RNA lysine 36 (anticodon CUU) pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39504523, end_position_on_the_genomic_accession=39504595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189465, CurrentID=0, Status='live', Symbol='TRE-TTC6-1', Aliases='TRNAE32P', description='transfer RNA-Glu (TTC) 6-1', other_designations='tRNA-Glu-TTC-6-1|transfer RNA glutamic acid 32 (anticodon UUC) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148308922, end_position_on_the_genomic_accession=148308993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189466, CurrentID=0, Status='live', Symbol='NMTRL-TAA5-1', Aliases='TRNAL42P', description='nuclear-encoded mitochondrial transfer RNA-Leu (TAA) 5-1', other_designations='nmt-tRNA-Leu-TAA-5-1|transfer RNA leucine 42 (anticodon UAA) pseudogene', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57358372, end_position_on_the_genomic_accession=57358446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189467, CurrentID=0, Status='live', Symbol='TRK-CTT14-1', Aliases='TRNAK37P', description='transfer RNA-Lys (CTT) 14-1', other_designations='tRNA-Lys-CTT-14-1|transfer RNA lysine 37 (anticodon CUU) pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3196153, end_position_on_the_genomic_accession=3196225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189468, CurrentID=0, Status='live', Symbol='TRA-AGC23-1', Aliases='TRNAA44P', description='transfer RNA-Ala (AGC) 23-1', other_designations='tRNA-Ala-AGC-23-1|transfer RNA alanine 44 (anticodon AGC) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28778767, end_position_on_the_genomic_accession=28778838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189469, CurrentID=0, Status='live', Symbol='TRA-TGC10-1', Aliases='TRNAA45P', description='transfer RNA-Ala (TGC) 10-1', other_designations='tRNA-Ala-TGC-10-1|transfer RNA alanine 45 (anticodon UGC) pseudogene', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60996036, end_position_on_the_genomic_accession=60996105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189470, CurrentID=0, Status='live', Symbol='TRSUP-CTA3-1', Aliases='TRNASUP6P', description='transfer RNA suppressor (CTA) 3-1', other_designations='tRNA-Sup-CTA-3-1|transfer RNA suppressor 6 (anticodon CUA) pseudogene', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78234495, end_position_on_the_genomic_accession=78234567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189471, CurrentID=0, Status='live', Symbol='TRF-GAA12-1', Aliases='TRNAF14P', description='transfer RNA-Phe (GAA) 12-1', other_designations='tRNA-Phe-GAA-12-1|transfer RNA phenylalanine 14 (anticodon GAA) pseudogene', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123258485, end_position_on_the_genomic_accession=123258557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189472, CurrentID=0, Status='live', Symbol='TRF-GAA11-1', Aliases='TRNAF15P', description='transfer RNA-Phe (GAA) 11-1', other_designations='tRNA-Phe-GAA-11-1|transfer RNA phenylalanine 15 (anticodon GAA) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28727078, end_position_on_the_genomic_accession=28727150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189473, CurrentID=0, Status='live', Symbol='TRQ-TTG5-1', Aliases='TRNAQ39P', description='transfer RNA-Gln (TTG) 5-1', other_designations='tRNA-Gln-TTG-5-1|transfer RNA glutamine 39 (anticodon UUG) pseudogene', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45710302, end_position_on_the_genomic_accession=45710374, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189474, CurrentID=0, Status='live', Symbol='TRI-AAT11-1', Aliases='TRNAI29P', description='transfer RNA-Ile (AAT) 11-1', other_designations='tRNA-Ile-AAT-11-1|transfer RNA isoleucine 29 (anticodon AAU) pseudogene', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129231653, end_position_on_the_genomic_accession=129231727, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189475, CurrentID=0, Status='live', Symbol='NMTRQ-TTG1-1', Aliases='TRNAQ40P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 1-1', other_designations='nmt-tRNA-Gln-TTG-1-1|transfer RNA glutamine 40 (anticodon UUG) pseudogene', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55180233, end_position_on_the_genomic_accession=55180304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189476, CurrentID=0, Status='live', Symbol='TRP-AGG6-1', Aliases='TRNAP24P', description='transfer RNA-Pro (AGG) 6-1', other_designations='tRNA-Pro-AGG-6-1|transfer RNA proline 24 (anticodon AGG) pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3152608, end_position_on_the_genomic_accession=3152679, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189477, CurrentID=0, Status='live', Symbol='TRV-CAC14-1', Aliases='TRNAV33P', description='transfer RNA-Val (CAC) 14-1', other_designations='tRNA-Val-CAC-14-1|transfer RNA valine 33 (anticodon CAC) pseudogene', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45322361, end_position_on_the_genomic_accession=45322433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189478, CurrentID=0, Status='live', Symbol='NMTRQ-TTG12-1', Aliases='TRNAQ41P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 12-1', other_designations='nmt-tRNA-Gln-TTG-12-1|transfer RNA glutamine 41 (anticodon UUG) pseudogene|transfer RNA-Gln (TTG) 12-1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19603274, end_position_on_the_genomic_accession=19603345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189479, CurrentID=0, Status='live', Symbol='TRV-CAC11-2', Aliases='TRNAV34P', description='transfer RNA-Val (CAC) 11-2', other_designations='tRNA-Val-CAC-11-2|transfer RNA valine 34 (anticodon CAC) pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16860198, end_position_on_the_genomic_accession=16860270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189480, CurrentID=0, Status='live', Symbol='TRK-TTT10-1', Aliases='TRNAK38P', description='transfer RNA-Lys (TTT) 10-1', other_designations='tRNA-Lys-TTT-10-1|transfer RNA lysine 38 (anticodon UUU) pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41242237, end_position_on_the_genomic_accession=41242309, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189481, CurrentID=0, Status='live', Symbol='TRI-AAT12-1', Aliases='TRNAI30P', description='transfer RNA-Ile (AAT) 12-1', other_designations='tRNA-Ile-AAT-12-1|transfer RNA isoleucine 30 (anticodon AAU) pseudogene', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52601799, end_position_on_the_genomic_accession=52601876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189482, CurrentID=0, Status='live', Symbol='TRP-AGG5-1', Aliases='TRNAP25P', description='transfer RNA-Pro (AGG) 5-1', other_designations='tRNA-Pro-AGG-5-1|transfer RNA proline 25 (anticodon AGG) pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77668467, end_position_on_the_genomic_accession=77668537, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189483, CurrentID=0, Status='live', Symbol='TRP-TGG4-1', Aliases='TRNAP26P', description='transfer RNA-Pro (TGG) 4-1', other_designations='tRNA-Pro-TGG-4-1|transfer RNA proline 26 (anticodon UGG) pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3170960, end_position_on_the_genomic_accession=3171030, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189484, CurrentID=0, Status='live', Symbol='TRE-TTC15-1', Aliases='TRNAE33P', description='transfer RNA-Glu (TTC) 15-1', other_designations='tRNA-Glu-TTC-15-1|transfer RNA glutamic acid 33 (anticodon UUC) pseudogene', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162788982, end_position_on_the_genomic_accession=162789050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189485, CurrentID=0, Status='live', Symbol='TRE-TTC14-1', Aliases='TRNAE34P', description='transfer RNA-Glu (TTC) 14-1', other_designations='tRNA-Glu-TTC-14-1|transfer RNA glutamic acid 34 (anticodon UUC) pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147078630, end_position_on_the_genomic_accession=147078701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189486, CurrentID=0, Status='live', Symbol='TRG-CCC4-1', Aliases='TRNAG36, TRNAV35P', description='transfer RNA-Gly (CCC) 4-1', other_designations='tRNA-Gly-CCC-4-1|transfer RNA glycine 36 (anticodon CCC)|transfer RNA valine 35 (anticodon CAC) pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16547665, end_position_on_the_genomic_accession=16547737, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189487, CurrentID=0, Status='live', Symbol='TRS-AGA7-1', Aliases='TRNAS29P', description='transfer RNA-Ser (AGA) 7-1', other_designations='tRNA-Ser-AGA-7-1|transfer RNA serine 29 (anticodon AGA) pseudogene', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31500900, end_position_on_the_genomic_accession=31500971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189488, CurrentID=0, Status='live', Symbol='TRE-CTC11-1', Aliases='TRNAE35P', description='transfer RNA-Glu (CTC) 11-1', other_designations='tRNA-Glu-CTC-11-1|transfer RNA glutamic acid 35 (anticodon CUC) pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11935871, end_position_on_the_genomic_accession=11935942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189489, CurrentID=0, Status='live', Symbol='NMTRQ-TTG4-1', Aliases='TRNAQ42P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 4-1', other_designations='nmt-tRNA-Gln-TTG-4-1|transfer RNA glutamine 42 (anticodon UUG) pseudogene', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106902031, end_position_on_the_genomic_accession=106902102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189490, CurrentID=0, Status='live', Symbol='TRN-GTT23-1', Aliases='TRNAN34P', description='transfer RNA-Asn (GTT) 23-1', other_designations='tRNA-Asn-GTT-23-1|transfer RNA asparagine 34 (anticodon GUU) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148378816, end_position_on_the_genomic_accession=148379089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189491, CurrentID=0, Status='live', Symbol='TRE-CTC7-1', Aliases='TRNAE36P', description='transfer RNA-Glu (CTC) 7-1', other_designations='tRNA-Glu-CTC-7-1|transfer RNA glutamic acid 36 (anticodon CUC) pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158881659, end_position_on_the_genomic_accession=158881730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50484, CurrentID=0, Status='live', Symbol='RRM2B', Aliases='MTDPS8A, MTDPS8B, P53R2', description='ribonucleotide reductase M2 B (TP53 inducible)', other_designations='TP53-inducible ribonucleotide reductase M2 B|p53-inducible ribonucleotide reductase small subunit 2 homolog|p53-inducible ribonucleotide reductase small subunit 2 short form beta|p53-inducible ribonucleotide reductase small subunit 2-like protein', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102204501, end_position_on_the_genomic_accession=102239118, orientation='minus', exon_count=9, OMIM=604712, no_hdr0=''),
  50485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50485, CurrentID=0, Status='live', Symbol='SMARCAL1', Aliases='HARP, HHARP', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a-like 1', other_designations='ATP-driven annealing helicase|HepA-related protein|SMARCA-like protein 1|sucrose nonfermenting protein 2-like 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216412414, end_position_on_the_genomic_accession=216483053, orientation='plus', exon_count=19, OMIM=606622, no_hdr0=''),
  50486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50486, CurrentID=0, Status='live', Symbol='G0S2', Aliases='', description='G0/G1 switch 2', other_designations='G0/G1 switch regulatory protein 2|G0/G1switch 2', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209675325, end_position_on_the_genomic_accession=209676390, orientation='plus', exon_count=2, OMIM=614447, no_hdr0=''),
  50487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50487, CurrentID=0, Status='live', Symbol='PLA2G3', Aliases='GIII-SPLA2, SPLA2III, sPLA2-III', description='phospholipase A2 group III', other_designations='group III secreted phospholipase A2|phosphatidylcholine 2-acylhydrolase 3|phosphatidylcholine 2-acylhydrolase GIII|phospholipase A2, group III', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31134807, end_position_on_the_genomic_accession=31140757, orientation='minus', exon_count=9, OMIM=611651, no_hdr0=''),
  50488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50488, CurrentID=0, Status='live', Symbol='MINK1', Aliases='B55, MAP4K6, MINK, YSK2, ZC3, hMINK, hMINKbeta', description='misshapen-like kinase 1', other_designations='GCK family kinase MINK|MAPK/ERK kinase kinase kinase 6|MEK kinase kinase 6|MEKKK 6|misshapen/NIK-related kinase|mitogen-activated protein kinase kinase kinase kinase 6|serine/threonine protein kinase', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4833340, end_position_on_the_genomic_accession=4898061, orientation='plus', exon_count=35, OMIM=609426, no_hdr0=''),
  50489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50489, CurrentID=0, Status='live', Symbol='CD207', Aliases='CLEC4K', description='CD207 molecule, langerin', other_designations='C-type lectin domain family 4, member K|CD207 antigen, langerin|Langerhans cell specific c-type lectin', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70825208, end_position_on_the_genomic_accession=70836601, orientation='minus', exon_count=9, OMIM=604862, no_hdr0=''),
  100189498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189498, CurrentID=0, Status='live', Symbol='TRG-CCC6-1', Aliases='TRNAG34P', description='transfer RNA-Gly (CCC) 6-1', other_designations='tRNA-Gly-CCC-6-1|transfer RNA glycine 34 (anticodon CCC) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149708561, end_position_on_the_genomic_accession=149708835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189499, CurrentID=0, Status='live', Symbol='TRL-CAA7-1', Aliases='TRNAL44P', description='transfer RNA-Leu (CAA) 7-1', other_designations='tRNA-Leu-CAA-7-1|transfer RNA leucine 44 (anticodon CAA) pseudogene', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34302488, end_position_on_the_genomic_accession=34302555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189500, CurrentID=0, Status='live', Symbol='TRV-CAC13-1', Aliases='TRNAV36P', description='transfer RNA-Val (CAC) 13-1', other_designations='tRNA-Val-CAC-13-1|transfer RNA valine 36 (anticodon CAC) pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16725566, end_position_on_the_genomic_accession=16725638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285375, CurrentID=0, Status='live', Symbol='LINC00620', Aliases='', description='long intergenic non-protein coding RNA 620', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13650721, end_position_on_the_genomic_accession=13746635, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100189502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189502, CurrentID=0, Status='live', Symbol='TRR-CCT8-1', Aliases='TRNAR32P', description='transfer RNA-Arg (CCT) 8-1', other_designations='tRNA-Arg-CCT-8-1|transfer RNA arginine 32 (anticodon CCU) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149756721, end_position_on_the_genomic_accession=149756791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189503, CurrentID=0, Status='live', Symbol='TRQ-CTG16-1', Aliases='TRNAQ44P', description='transfer RNA-Gln (CTG) 16-1', other_designations='tRNA-Gln-CTG-16-1|transfer RNA glutamine 44 (anticodon CUG) pseudogene', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218626409, end_position_on_the_genomic_accession=218626484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189504, CurrentID=0, Status='live', Symbol='TRK-TTT15-1', Aliases='TRNAK39P', description='transfer RNA-Lys (TTT) 15-1', other_designations='tRNA-Lys-TTT-15-1|transfer RNA lysine 39 (anticodon UUU) pseudogene', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223321597, end_position_on_the_genomic_accession=223321669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189505, CurrentID=0, Status='live', Symbol='TRQ-TTG8-1', Aliases='TRNAQ45P', description='transfer RNA-Gln (TTG) 8-1', other_designations='tRNA-Gln-TTG-8-1|transfer RNA glutamine 45 (anticodon UUG) pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9039680, end_position_on_the_genomic_accession=9039752, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189506, CurrentID=0, Status='live', Symbol='TRE-CTC4-1', Aliases='TRNAE39P', description='transfer RNA-Glu (CTC) 4-1', other_designations='tRNA-Glu-CTC-4-1|transfer RNA glutamic acid 39 (anticodon CUC) pseudogene', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57884416, end_position_on_the_genomic_accession=57884487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189507, CurrentID=0, Status='live', Symbol='TRY-GTA11-1', Aliases='TRNAY17P', description='transfer RNA-Tyr (GTA) 11-1', other_designations='tRNA-Tyr-GTA-11-1|transfer RNA tyrosine 17 (anticodon GUA) pseudogene', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149356654, end_position_on_the_genomic_accession=149356729, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480964, CurrentID=0, Status='live', Symbol='RN7SL160P', Aliases='', description='RNA, 7SL, cytoplasmic 160, pseudogene', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189509, CurrentID=0, Status='live', Symbol='TRQ-TTG9-1', Aliases='TRNAQ46P', description='transfer RNA-Gln (TTG) 9-1', other_designations='tRNA-Gln-TTG-9-1|transfer RNA glutamine 46 (anticodon UUG) pseudogene', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97096913, end_position_on_the_genomic_accession=97096986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189510, CurrentID=0, Status='live', Symbol='NMTRQ-TTG6-1', Aliases='TRNAQ47P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 6-1', other_designations='nmt-tRNA-Gln-TTG-6-1|transfer RNA glutamine 47 (anticodon UUG) pseudogene', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49817407, end_position_on_the_genomic_accession=49817481, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189511, CurrentID=0, Status='live', Symbol='TRG-TCC6-1', Aliases='TRNAG35P', description='transfer RNA-Gly (TCC) 6-1', other_designations='tRNA-Gly-TCC-6-1|transfer RNA glycine 35 (anticodon UCC) pseudogene', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57678945, end_position_on_the_genomic_accession=57679016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189512, CurrentID=0, Status='live', Symbol='NMTRS-TGA3-1', Aliases='TRNAS30P', description='nuclear-encoded mitochondrial transfer RNA-Ser (TGA) 3-1', other_designations='nmt-tRNA-Ser-TGA-3-1|transfer RNA serine 30 (anticodon UGA) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22528853, end_position_on_the_genomic_accession=22528921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189513, CurrentID=0, Status='live', Symbol='NMTRV-TAC1-1', Aliases='TRNAV37P', description='nuclear-encoded mitochondrial transfer RNA-Val (TAC) 1-1', other_designations='nmt-tRNA-Val-TAC-1-1|transfer RNA valine 37 (anticodon UAC) pseudogene', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156547912, end_position_on_the_genomic_accession=156547986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50506, CurrentID=0, Status='live', Symbol='DUOX2', Aliases='LNOX2, NOXEF2, P138-TOX, TDH6, THOX2', description='dual oxidase 2', other_designations='NADH/NADPH thyroid oxidase p138-tox|NADPH oxidase/peroxidase DUOX2|NADPH thyroid oxidase 2|dual oxidase-like domains 2|flavoprotein NADPH oxidase|large NOX 2|long NOX 2|nicotinamide adenine dinucleotide phosphate oxidase|p138 thyroid oxidase|thyroid oxidase 2', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45092653, end_position_on_the_genomic_accession=45114161, orientation='minus', exon_count=34, OMIM=606759, no_hdr0=''),
  50507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50507, CurrentID=0, Status='live', Symbol='NOX4', Aliases='KOX, KOX-1, RENOX', description='NADPH oxidase 4', other_designations='kidney oxidase-1|kidney superoxide-producing NADPH oxidase|renal NAD(P)H-oxidase', map_location='11q14.2-q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89324353, end_position_on_the_genomic_accession=89589611, orientation='minus', exon_count=25, OMIM=605261, no_hdr0=''),
  50508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50508, CurrentID=0, Status='live', Symbol='NOX3', Aliases='GP91-3, MOX-2', description='NADPH oxidase 3', other_designations='NADPH oxidase catalytic subunit-like 3|gp91phox homolog 3|mitogenic oxidase 2', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155395368, end_position_on_the_genomic_accession=155455903, orientation='minus', exon_count=14, OMIM=607105, no_hdr0=''),
  50509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50509, CurrentID=0, Status='live', Symbol='COL5A3', Aliases='', description='collagen, type V, alpha 3', other_designations='pro-(alpha)3(V) collagen', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9959561, end_position_on_the_genomic_accession=10010538, orientation='minus', exon_count=67, OMIM=120216, no_hdr0=''),
  100189518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189518, CurrentID=0, Status='live', Symbol='TRE-TTC9-1', Aliases='TRNAE40P', description='transfer RNA-Glu (TTC) 9-1', other_designations='tRNA-Glu-TTC-9-1|transfer RNA glutamic acid 40 (anticodon UUC) pseudogene', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203364408, end_position_on_the_genomic_accession=203364478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50511, CurrentID=0, Status='live', Symbol='SYCP3', Aliases='COR1, SCP3, SPGF4', description='synaptonemal complex protein 3', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101728648, end_position_on_the_genomic_accession=101739513, orientation='minus', exon_count=8, OMIM=604759, no_hdr0=''),
  50512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50512, CurrentID=0, Status='live', Symbol='PODXL2', Aliases='EG, PODLX2', description='podocalyxin-like 2', other_designations='endoglycan', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127629159, end_position_on_the_genomic_accession=127672810, orientation='plus', exon_count=8, OMIM=616627, no_hdr0=''),
  106480977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480977, CurrentID=0, Status='live', Symbol='RN7SL204P', Aliases='', description='RNA, 7SL, cytoplasmic 204, pseudogene', other_designations='', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50514, CurrentID=0, Status='live', Symbol='DEC1', Aliases='CTS9', description='deleted in esophageal cancer 1', other_designations='candidate tumor suppressor CTS9', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115112617, end_position_on_the_genomic_accession=115402644, orientation='plus', exon_count=10, OMIM=604767, no_hdr0=''),
  50515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50515, CurrentID=0, Status='live', Symbol='CHST11', Aliases='C4ST, C4ST-1, C4ST1, HSA269537', description='carbohydrate (chondroitin 4) sulfotransferase 11', other_designations='C4S-1|IgH/CHST11 fusion|chondroitin 4-O-sulfotransferase 1', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104456914, end_position_on_the_genomic_accession=104762014, orientation='plus', exon_count=3, OMIM=610128, no_hdr0=''),
  50516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50516, CurrentID=0, Status='live', Symbol='PPP1R14BP1', Aliases='PLCB3NP', description='protein phosphatase 1 regulatory inhibitor subunit 14B pseudogene 1', other_designations='phospholipase C, beta 3, neighbor pseudogene|protein phosphatase 1, regulatory (inhibitor) subunit 14B pseudogene 1', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31276437, end_position_on_the_genomic_accession=31277415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189525, CurrentID=0, Status='live', Symbol='NMTRS-TGA2-1', Aliases='TRNAS31P', description='nuclear-encoded mitochondrial transfer RNA-Ser (TGA) 2-1', other_designations='nmt-tRNA-Ser-TGA-2-1|transfer RNA serine 31 (anticodon UGA) pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131382569, end_position_on_the_genomic_accession=131382638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189526, CurrentID=0, Status='live', Symbol='TRS-ACT1-1', Aliases='TRNAS32P', description='transfer RNA-Ser (ACT) 1-1', other_designations='tRNA-Ser-ACT-1-1|transfer RNA serine 32 (anticodon ACU) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27293892, end_position_on_the_genomic_accession=27293965, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189527, CurrentID=0, Status='live', Symbol='TRE-CTC15-1', Aliases='TRNAE41P', description='transfer RNA-Glu (CTC) 15-1', other_designations='tRNA-Glu-CTC-15-1|transfer RNA glutamic acid 41 (anticodon CUC) pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71046358, end_position_on_the_genomic_accession=71046430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189528, CurrentID=0, Status='live', Symbol='TRE-CTC16-1', Aliases='TRNAE42P', description='transfer RNA-Glu (CTC) 16-1', other_designations='tRNA-Glu-CTC-16-1|transfer RNA glutamic acid 42 (anticodon CUC) pseudogene', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113948740, end_position_on_the_genomic_accession=113948812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189529, CurrentID=0, Status='live', Symbol='TRA-TGC9-1', Aliases='TRNAA46P', description='transfer RNA-Ala (TGC) 9-1', other_designations='tRNA-Ala-TGC-9-1|transfer RNA alanine 46 (anticodon UGC) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28634082, end_position_on_the_genomic_accession=28634153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189530, CurrentID=0, Status='live', Symbol='TRP-AGG4-1', Aliases='TRNAP28P', description='transfer RNA-Pro (AGG) 4-1', other_designations='tRNA-Pro-AGG-4-1|transfer RNA proline 28 (anticodon AGG) pseudogene', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87112450, end_position_on_the_genomic_accession=87112521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189531, CurrentID=0, Status='live', Symbol='TRQ-CTG9-1', Aliases='TRNAQ51P', description='transfer RNA-Gln (CTG) 9-1', other_designations='tRNA-Gln-CTG-9-1|transfer RNA glutamine 51 (anticodon CUG) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148353561, end_position_on_the_genomic_accession=148353632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189532, CurrentID=0, Status='live', Symbol='TRE-CTC10-1', Aliases='TRNAE43P', description='transfer RNA-Glu (CTC) 10-1', other_designations='tRNA-Glu-CTC-10-1|transfer RNA glutamic acid 43 (anticodon CUC) pseudogene', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13989824, end_position_on_the_genomic_accession=13989895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189533, CurrentID=0, Status='live', Symbol='TRF-GAA9-1', Aliases='TRNAF17P', description='transfer RNA-Phe (GAA) 9-1', other_designations='tRNA-Phe-GAA-9-1|transfer RNA phenylalanine 17 (anticodon GAA) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143792624, end_position_on_the_genomic_accession=143792895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189534, CurrentID=0, Status='live', Symbol='TRN-GTT22-1', Aliases='TRNAN36P', description='transfer RNA-Asn (GTT) 22-1', other_designations='tRNA-Asn-GTT-22-1|transfer RNA asparagine 36 (anticodon GUU) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143717291, end_position_on_the_genomic_accession=143717364, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189535, CurrentID=0, Status='live', Symbol='NMTRQ-TTG13-1', Aliases='TRNAQ52P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 13-1', other_designations='nmt-tRNA-Gln-TTG-13-1|transfer RNA glutamine 52 (anticodon UUG) pseudogene|transfer RNA-Gln (TTG) 13-1', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57186273, end_position_on_the_genomic_accession=57186344, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189536, CurrentID=0, Status='live', Symbol='TRK-CTT16-1', Aliases='TRNAK42P', description='transfer RNA-Lys (CTT) 16-1', other_designations='tRNA-Lys-CTT-16-1|transfer RNA lysine 42 (anticodon CUU) pseudogene', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76382424, end_position_on_the_genomic_accession=76382497, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189537, CurrentID=0, Status='live', Symbol='TRR-CCT9-1', Aliases='TRNAR34P', description='transfer RNA-Arg (CCT) 9-1', other_designations='tRNA-Arg-CCT-9-1|transfer RNA arginine 34 (anticodon CCU) pseudogene', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63894295, end_position_on_the_genomic_accession=63894363, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189538, CurrentID=0, Status='live', Symbol='TRQ-CTG13-1', Aliases='TRNAQ53P', description='transfer RNA-Gln (CTG) 13-1', other_designations='tRNA-Gln-CTG-13-1|transfer RNA glutamine 53 (anticodon CUG) pseudogene', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151868513, end_position_on_the_genomic_accession=151868584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189539, CurrentID=0, Status='live', Symbol='TRK-TTT13-1', Aliases='TRNAK43P', description='transfer RNA-Lys (TTT) 13-1', other_designations='tRNA-Lys-TTT-13-1|transfer RNA lysine 43 (anticodon UUU) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28693210, end_position_on_the_genomic_accession=28693283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189540, CurrentID=0, Status='live', Symbol='NMTRQ-TTG15-1', Aliases='TRNAQ54P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 15-1', other_designations='nmt-tRNA-Gln-TTG-15-1|transfer RNA glutamine 54 (anticodon UUG) pseudogene|transfer RNA-Gln (TTG) 15-1', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155462736, end_position_on_the_genomic_accession=155462807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401638, CurrentID=0, Status='live', Symbol='RPSAP10', Aliases='LAMR1P10, RPSA_22_1026', description='ribosomal protein SA pseudogene 10', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28436126, end_position_on_the_genomic_accession=28437151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480998, CurrentID=0, Status='live', Symbol='RN7SL279P', Aliases='', description='RNA, 7SL, cytoplasmic 279, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480999, CurrentID=0, Status='live', Symbol='RN7SL284P', Aliases='', description='RNA, 7SL, cytoplasmic 284, pseudogene', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481000, CurrentID=0, Status='live', Symbol='RN7SL288P', Aliases='', description='RNA, 7SL, cytoplasmic 288, pseudogene', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481001, CurrentID=0, Status='live', Symbol='RN7SL290P', Aliases='', description='RNA, 7SL, cytoplasmic 290, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481002, CurrentID=0, Status='live', Symbol='RN7SL292P', Aliases='', description='RNA, 7SL, cytoplasmic 292, pseudogene', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481003, CurrentID=0, Status='live', Symbol='RN7SL296P', Aliases='', description='RNA, 7SL, cytoplasmic 296, pseudogene', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481004, CurrentID=0, Status='live', Symbol='RN7SL300P', Aliases='', description='RNA, 7SL, cytoplasmic 300, pseudogene', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481005, CurrentID=0, Status='live', Symbol='RN7SL303P', Aliases='', description='RNA, 7SL, cytoplasmic 303, pseudogene', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481006, CurrentID=0, Status='live', Symbol='RN7SL306P', Aliases='', description='RNA, 7SL, cytoplasmic 306, pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481007, CurrentID=0, Status='live', Symbol='RN7SL309P', Aliases='', description='RNA, 7SL, cytoplasmic 309, pseudogene', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481008, CurrentID=0, Status='live', Symbol='RN7SL313P', Aliases='', description='RNA, 7SL, cytoplasmic 313, pseudogene', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401640, CurrentID=0, Status='live', Symbol='RPL7AP53', Aliases='RPL7A_22_1031', description='ribosomal protein L7a pseudogene 53', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33057555, end_position_on_the_genomic_accession=33058432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481010, CurrentID=0, Status='live', Symbol='RN7SL320P', Aliases='', description='RNA, 7SL, cytoplasmic 320, pseudogene', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481011, CurrentID=0, Status='live', Symbol='RN7SL325P', Aliases='', description='RNA, 7SL, cytoplasmic 325, pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481012, CurrentID=0, Status='live', Symbol='RN7SL327P', Aliases='', description='RNA, 7SL, cytoplasmic 327, pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481013, CurrentID=0, Status='live', Symbol='RN7SL330P', Aliases='', description='RNA, 7SL, cytoplasmic 330, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481014, CurrentID=0, Status='live', Symbol='RN7SL333P', Aliases='', description='RNA, 7SL, cytoplasmic 333, pseudogene', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481015, CurrentID=0, Status='live', Symbol='RN7SL337P', Aliases='', description='RNA, 7SL, cytoplasmic 337, pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481016, CurrentID=0, Status='live', Symbol='RN7SL340P', Aliases='', description='RNA, 7SL, cytoplasmic 340, pseudogene', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481017, CurrentID=0, Status='live', Symbol='RN7SL341P', Aliases='', description='RNA, 7SL, cytoplasmic 341, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481018, CurrentID=0, Status='live', Symbol='RN7SL343P', Aliases='', description='RNA, 7SL, cytoplasmic 343, pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481019, CurrentID=0, Status='live', Symbol='RN7SL350P', Aliases='', description='RNA, 7SL, cytoplasmic 350, pseudogene', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481020, CurrentID=0, Status='live', Symbol='RN7SL353P', Aliases='', description='RNA, 7SL, cytoplasmic 353, pseudogene', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481021, CurrentID=0, Status='live', Symbol='RN7SL357P', Aliases='', description='RNA, 7SL, cytoplasmic 357, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481022, CurrentID=0, Status='live', Symbol='RN7SL362P', Aliases='', description='RNA, 7SL, cytoplasmic 362, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481023, CurrentID=0, Status='live', Symbol='RN7SL366P', Aliases='', description='RNA, 7SL, cytoplasmic 366, pseudogene', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481024, CurrentID=0, Status='live', Symbol='RN7SL370P', Aliases='', description='RNA, 7SL, cytoplasmic 370, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481025, CurrentID=0, Status='live', Symbol='RN7SL376P', Aliases='', description='RNA, 7SL, cytoplasmic 376, pseudogene', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481026, CurrentID=0, Status='live', Symbol='RN7SL378P', Aliases='', description='RNA, 7SL, cytoplasmic 378, pseudogene', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481027, CurrentID=0, Status='live', Symbol='RN7SL382P', Aliases='', description='RNA, 7SL, cytoplasmic 382, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481028, CurrentID=0, Status='live', Symbol='RN7SL384P', Aliases='', description='RNA, 7SL, cytoplasmic 384, pseudogene', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481029, CurrentID=0, Status='live', Symbol='RN7SL387P', Aliases='', description='RNA, 7SL, cytoplasmic 387, pseudogene', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481030, CurrentID=0, Status='live', Symbol='RN7SL391P', Aliases='', description='RNA, 7SL, cytoplasmic 391, pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481031, CurrentID=0, Status='live', Symbol='RN7SL395P', Aliases='', description='RNA, 7SL, cytoplasmic 395, pseudogene', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481032, CurrentID=0, Status='live', Symbol='RN7SL398P', Aliases='', description='RNA, 7SL, cytoplasmic 398, pseudogene', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481033, CurrentID=0, Status='live', Symbol='RN7SL400P', Aliases='', description='RNA, 7SL, cytoplasmic 400, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481034, CurrentID=0, Status='live', Symbol='RN7SL403P', Aliases='', description='RNA, 7SL, cytoplasmic 403, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481035, CurrentID=0, Status='live', Symbol='RN7SL411P', Aliases='', description='RNA, 7SL, cytoplasmic 411, pseudogene', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  154442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154442, CurrentID=0, Status='live', Symbol='BVES-AS1', Aliases='C6orf112, bA99L11.2', description='BVES antisense RNA 1', other_designations='BVES antisense RNA 1 (non-protein coding)|Putative uncharacterized protein C6orf112', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105137687, end_position_on_the_genomic_accession=105169945, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481037, CurrentID=0, Status='live', Symbol='RN7SL416P', Aliases='', description='RNA, 7SL, cytoplasmic 416, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481038, CurrentID=0, Status='live', Symbol='RN7SL422P', Aliases='', description='RNA, 7SL, cytoplasmic 422, pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481039, CurrentID=0, Status='live', Symbol='RN7SL425P', Aliases='', description='RNA, 7SL, cytoplasmic 425, pseudogene', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481040, CurrentID=0, Status='live', Symbol='RN7SL430P', Aliases='', description='RNA, 7SL, cytoplasmic 430, pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481041, CurrentID=0, Status='live', Symbol='RN7SL432P', Aliases='', description='RNA, 7SL, cytoplasmic 432, pseudogene', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481042, CurrentID=0, Status='live', Symbol='RN7SL440P', Aliases='', description='RNA, 7SL, cytoplasmic 440, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481043, CurrentID=0, Status='live', Symbol='RN7SL442P', Aliases='', description='RNA, 7SL, cytoplasmic 442, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481044, CurrentID=0, Status='live', Symbol='RN7SL445P', Aliases='', description='RNA, 7SL, cytoplasmic 445, pseudogene', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100189589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189589, CurrentID=0, Status='live', Symbol='DCTN1-AS1', Aliases='', description='DCTN1 antisense RNA 1', other_designations='DCTN1 antisense RNA 1 (non-protein coding)', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74385718, end_position_on_the_genomic_accession=74393881, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106481046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481046, CurrentID=0, Status='live', Symbol='RN7SL453P', Aliases='', description='RNA, 7SL, cytoplasmic 453, pseudogene', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481047, CurrentID=0, Status='live', Symbol='RN7SL457P', Aliases='', description='RNA, 7SL, cytoplasmic 457, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481048, CurrentID=0, Status='live', Symbol='RN7SL460P', Aliases='', description='RNA, 7SL, cytoplasmic 460, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481049, CurrentID=0, Status='live', Symbol='RN7SL464P', Aliases='', description='RNA, 7SL, cytoplasmic 464, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481050, CurrentID=0, Status='live', Symbol='RN7SL468P', Aliases='', description='RNA, 7SL, cytoplasmic 468, pseudogene', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401647, CurrentID=0, Status='live', Symbol='GOLGA7B', Aliases='C10orf132, C10orf133, bA451M19.3, bA459F3.4', description='golgin A7 family member B', other_designations='golgi autoantigen, golgin subfamily a, 7B|golgin A7 family, member B', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97850238, end_position_on_the_genomic_accession=97871578, orientation='plus', exon_count=5, OMIM=614189, no_hdr0=''),
  106481052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481052, CurrentID=0, Status='live', Symbol='RN7SL474P', Aliases='', description='RNA, 7SL, cytoplasmic 474, pseudogene', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481053, CurrentID=0, Status='live', Symbol='RN7SL478P', Aliases='', description='RNA, 7SL, cytoplasmic 478, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481054, CurrentID=0, Status='live', Symbol='RN7SL481P', Aliases='', description='RNA, 7SL, cytoplasmic 481, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481055, CurrentID=0, Status='live', Symbol='RN7SL487P', Aliases='', description='RNA, 7SL, cytoplasmic 487, pseudogene', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481056, CurrentID=0, Status='live', Symbol='RN7SL489P', Aliases='', description='RNA, 7SL, cytoplasmic 489, pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481057, CurrentID=0, Status='live', Symbol='RN7SL494P', Aliases='', description='RNA, 7SL, cytoplasmic 494, pseudogene', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481058, CurrentID=0, Status='live', Symbol='RN7SL496P', Aliases='', description='RNA, 7SL, cytoplasmic 496, pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481059, CurrentID=0, Status='live', Symbol='RN7SL499P', Aliases='', description='RNA, 7SL, cytoplasmic 499, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481060, CurrentID=0, Status='live', Symbol='RN7SL503P', Aliases='', description='RNA, 7SL, cytoplasmic 503, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481061, CurrentID=0, Status='live', Symbol='RN7SL508P', Aliases='', description='RNA, 7SL, cytoplasmic 508, pseudogene', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481062, CurrentID=0, Status='live', Symbol='RN7SL513P', Aliases='', description='RNA, 7SL, cytoplasmic 513, pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481063, CurrentID=0, Status='live', Symbol='RN7SL515P', Aliases='', description='RNA, 7SL, cytoplasmic 515, pseudogene', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481064, CurrentID=0, Status='live', Symbol='RN7SL517P', Aliases='', description='RNA, 7SL, cytoplasmic 517, pseudogene', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481065, CurrentID=0, Status='live', Symbol='RN7SL520P', Aliases='', description='RNA, 7SL, cytoplasmic 520, pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481066, CurrentID=0, Status='live', Symbol='RN7SL524P', Aliases='', description='RNA, 7SL, cytoplasmic 524, pseudogene', other_designations='', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481067, CurrentID=0, Status='live', Symbol='RN7SL530P', Aliases='', description='RNA, 7SL, cytoplasmic 530, pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50604, CurrentID=0, Status='live', Symbol='IL20', Aliases='IL-20, IL10D, ZCYTO10', description='interleukin 20', other_designations='cytokine Zcyto10|four alpha helix cytokine', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206865354, end_position_on_the_genomic_accession=206869223, orientation='plus', exon_count=6, OMIM=605619, no_hdr0=''),
  106481069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481069, CurrentID=0, Status='live', Symbol='RN7SL541P', Aliases='', description='RNA, 7SL, cytoplasmic 541, pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481070, CurrentID=0, Status='live', Symbol='RN7SL544P', Aliases='', description='RNA, 7SL, cytoplasmic 544, pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481071, CurrentID=0, Status='live', Symbol='RN7SL549P', Aliases='', description='RNA, 7SL, cytoplasmic 549, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50608, CurrentID=0, Status='live', Symbol='IBD4', Aliases='', description='inflammatory bowel disease 4', other_designations='', map_location='14q11-q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606675, no_hdr0=''),
  106481073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481073, CurrentID=0, Status='live', Symbol='RN7SL556P', Aliases='', description='RNA, 7SL, cytoplasmic 556, pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481074, CurrentID=0, Status='live', Symbol='RN7SL558P', Aliases='', description='RNA, 7SL, cytoplasmic 558, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481075, CurrentID=0, Status='live', Symbol='RN7SL561P', Aliases='', description='RNA, 7SL, cytoplasmic 561, pseudogene', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50612, CurrentID=0, Status='live', Symbol='CXCL1P', Aliases='GRO1P, GROP, MGSAP, SCYB1P', description='chemokine (C-X-C motif) ligand 1 pseudogene', other_designations='chemokine (C-X-C motif) ligand 1 (melanoma growth stimulating activity, alpha) pseudogene|small inducible cytokine B subfamily (Cys-X-Cys motif), member 1 (MGSA/GRO), pseudogene', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73943995, end_position_on_the_genomic_accession=73944493, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50613, CurrentID=0, Status='live', Symbol='UBQLN3', Aliases='TUP-1', description='ubiquilin 3', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5507300, end_position_on_the_genomic_accession=5509985, orientation='minus', exon_count=2, OMIM=605473, no_hdr0=''),
  50614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50614, CurrentID=0, Status='live', Symbol='GALNT9', Aliases='GALNAC-T9, GALNACT9', description='polypeptide N-acetylgalactosaminyltransferase 9', other_designations='GalNAc transferase 9|UDP-GalNAc: polypeptide N-acetylgalactosaminyltransferase 9|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 9|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 9 (GalNAc-T9)|polypeptide GalNAc transferase 9|pp-GaNTase 9|protein-UDP acetylgalactosaminyltransferase 9', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132238677, end_position_on_the_genomic_accession=132329319, orientation='minus', exon_count=15, OMIM=606251, no_hdr0=''),
  50615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50615, CurrentID=0, Status='live', Symbol='IL21R', Aliases='CD360, NILR', description='interleukin 21 receptor', other_designations='IL-21 receptor|novel interleukin receptor', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27402162, end_position_on_the_genomic_accession=27452042, orientation='plus', exon_count=11, OMIM=605383, no_hdr0=''),
  50616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50616, CurrentID=0, Status='live', Symbol='IL22', Aliases='IL-21, IL-22, IL-D110, IL-TIF, ILTIF, TIFIL-23, TIFa, zcyto18', description='interleukin 22', other_designations='IL-10-related T-cell-derived inducible factor|IL-10-related T-cell-derived-inducible factor|cytokine Zcyto18', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68248245, end_position_on_the_genomic_accession=68253501, orientation='minus', exon_count=5, OMIM=605330, no_hdr0=''),
  50617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50617, CurrentID=0, Status='live', Symbol='ATP6V0A4', Aliases='A4, ATP6N1B, ATP6N2, RDRTA2, RTA1C, RTADR, STV1, VPH1, VPP2', description='ATPase, H+ transporting, lysosomal V0 subunit a4', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) non-catalytic accessory protein 1B|ATPase, H+ transporting, lysosomal (vacuolar proton pump) non-catalytic accessory protein 2 (38kD)|H(+)-transporting two-sector ATPase, noncatalytic accessory protein 1B|V-ATPase 116 kDa|V-type proton ATPase 116 kDa subunit a|vacuolar proton pump 116 kDa accessory subunit|vacuolar proton pump, subunit 2|vacuolar proton translocating ATPase 116 kDa subunit a kidney isoform', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138706294, end_position_on_the_genomic_accession=138799632, orientation='minus', exon_count=24, OMIM=605239, no_hdr0=''),
  50618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50618, CurrentID=0, Status='live', Symbol='ITSN2', Aliases='PRO2015, SH3D1B, SH3P18, SWA, SWAP', description='intersectin 2', other_designations='SH3 domain-containing protein 1B|SH3P18-like WASP-associated protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24202866, end_position_on_the_genomic_accession=24360528, orientation='minus', exon_count=40, OMIM=604464, no_hdr0=''),
  50619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50619, CurrentID=0, Status='live', Symbol='DEF6', Aliases='IBP, SLAT, SWAP70L', description='DEF6, guanine nucleotide exchange factor', other_designations='DEF-6|DEF6 guanine nucleotide exchange factor|IRF4-binding protein|SWAP-70-like adaptor protein of T cells', map_location='6p21.33-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35297818, end_position_on_the_genomic_accession=35321771, orientation='plus', exon_count=11, OMIM=610094, no_hdr0=''),
  106481084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481084, CurrentID=0, Status='live', Symbol='RN7SL591P', Aliases='', description='RNA, 7SL, cytoplasmic 591, pseudogene', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481085, CurrentID=0, Status='live', Symbol='RN7SL596P', Aliases='', description='RNA, 7SL, cytoplasmic 596, pseudogene', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481086, CurrentID=0, Status='live', Symbol='RN7SL600P', Aliases='', description='RNA, 7SL, cytoplasmic 600, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481087, CurrentID=0, Status='live', Symbol='RN7SL605P', Aliases='', description='RNA, 7SL, cytoplasmic 605, pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50624, CurrentID=0, Status='live', Symbol='CUZD1', Aliases='ERG-1, UO-44', description='CUB and zona pellucida-like domains 1', other_designations='CUB and ZP domain-containing protein 1|estrogen regulated gene 1|transmembrane protein UO-44', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122832155, end_position_on_the_genomic_accession=122846175, orientation='minus', exon_count=9, OMIM=616644, no_hdr0=''),
  106481089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481089, CurrentID=0, Status='live', Symbol='RN7SL614P', Aliases='', description='RNA, 7SL, cytoplasmic 614, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50626, CurrentID=0, Status='live', Symbol='CYHR1', Aliases='CHRP', description='cysteine/histidine-rich 1', other_designations='cysteine and histidine rich 1|cysteine and histidine rich protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144449582, end_position_on_the_genomic_accession=144465691, orientation='minus', exon_count=8, OMIM=616635, no_hdr0=''),
  106481091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481091, CurrentID=0, Status='live', Symbol='RN7SL619P', Aliases='', description='RNA, 7SL, cytoplasmic 619, pseudogene', other_designations='', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50628, CurrentID=0, Status='live', Symbol='GEMIN4', Aliases='HC56, HCAP1, HHRF-1, p97', description='gem nuclear organelle associated protein 4', other_designations='HCC-associated protein 1|component of gems 4|gem (nuclear organelle) associated protein 4|gemin-4', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=744416, end_position_on_the_genomic_accession=752872, orientation='minus', exon_count=6, OMIM=606969, no_hdr0=''),
  106481093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481093, CurrentID=0, Status='live', Symbol='RN7SL627P', Aliases='', description='RNA, 7SL, cytoplasmic 627, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481094, CurrentID=0, Status='live', Symbol='RN7SL630P', Aliases='', description='RNA, 7SL, cytoplasmic 630, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50631, CurrentID=0, Status='live', Symbol='YBX1P1', Aliases='NSEP1P, PSDBPB1', description='Y-box binding protein 1 pseudogene 1', other_designations='Y box binding protein 1 pseudogene 1|nuclease sensitive element binding protein 1 pseudogene', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66012433, end_position_on_the_genomic_accession=66013933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50632, CurrentID=0, Status='live', Symbol='CALY', Aliases='DRD1IP, NSG3', description='calcyon neuron-specific vesicular protein', other_designations='D1 dopamine receptor-interacting protein|calcyon D1 dopamine receptor-interacting protein (CALCYON)|calcyon protein|dopamine receptor D1 interacting protein', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133324072, end_position_on_the_genomic_accession=133336971, orientation='minus', exon_count=10, OMIM=604647, no_hdr0=''),
  106481097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481097, CurrentID=0, Status='live', Symbol='RN7SL636P', Aliases='', description='RNA, 7SL, cytoplasmic 636, pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481098, CurrentID=0, Status='live', Symbol='RN7SL643P', Aliases='', description='RNA, 7SL, cytoplasmic 643, pseudogene', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481099, CurrentID=0, Status='live', Symbol='RN7SL646P', Aliases='', description='RNA, 7SL, cytoplasmic 646, pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50636, CurrentID=0, Status='live', Symbol='ANO7', Aliases='D-TMPP, DTMPP, IPCA-5, IPCA5, NGEP, PCANAP5, PCANAP5L, TMEM16G', description='anoctamin 7', other_designations='Dresden-transmembrane protein of the prostate|prostate cancer associated protein 5|transmembrane protein 16G', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241188509, end_position_on_the_genomic_accession=241239595, orientation='plus', exon_count=28, OMIM=605096, no_hdr0=''),
  106481101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481101, CurrentID=0, Status='live', Symbol='RN7SL653P', Aliases='', description='RNA, 7SL, cytoplasmic 653, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481102, CurrentID=0, Status='live', Symbol='RN7SL655P', Aliases='', description='RNA, 7SL, cytoplasmic 655, pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50639, CurrentID=0, Status='live', Symbol='MBL3P', Aliases='COLEC2, MBL', description='mannose-binding lectin family member 3, pseudogene', other_designations='collectin sub-family member 2', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79582298, end_position_on_the_genomic_accession=79584877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50640, CurrentID=0, Status='live', Symbol='PNPLA8', Aliases='IPLA2-2, IPLA2G, MMLA, PNPLA-gamma, iPLA2gamma', description='patatin like phospholipase domain containing 8', other_designations='intracellular membrane-associated calcium-independent phospholipase A2 gamma|membrane-associated calcium-independent phospholipase A2 gamma|patatin-like phospholipase domain containing 8', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108470422, end_position_on_the_genomic_accession=108528161, orientation='minus', exon_count=15, OMIM=612123, no_hdr0=''),
  106481105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481105, CurrentID=0, Status='live', Symbol='RN7SL670P', Aliases='', description='RNA, 7SL, cytoplasmic 670, pseudogene', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481106, CurrentID=0, Status='live', Symbol='RN7SL672P', Aliases='', description='RNA, 7SL, cytoplasmic 672, pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481107, CurrentID=0, Status='live', Symbol='RN7SL675P', Aliases='', description='RNA, 7SL, cytoplasmic 675, pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50644, CurrentID=0, Status='live', Symbol='MRXS11', Aliases='SMRXS', description='mental retardation, X-linked, syndromic 11', other_designations='Shashi X-linked mental retardation syndrome', map_location='Xq26-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300238, no_hdr0=''),
  106481109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481109, CurrentID=0, Status='live', Symbol='RN7SL683P', Aliases='', description='RNA, 7SL, cytoplasmic 683, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481110, CurrentID=0, Status='live', Symbol='RN7SL687P', Aliases='', description='RNA, 7SL, cytoplasmic 687, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481111, CurrentID=0, Status='live', Symbol='RN7SL689P', Aliases='', description='RNA, 7SL, cytoplasmic 689, pseudogene', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481112, CurrentID=0, Status='live', Symbol='RN7SL691P', Aliases='', description='RNA, 7SL, cytoplasmic 691, pseudogene', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50649, CurrentID=0, Status='live', Symbol='ARHGEF4', Aliases='ASEF, ASEF1, GEF4, STM6', description='Rho guanine nucleotide exchange factor 4', other_designations='APC-stimulated guanine nucleotide exchange factor 1|Rho guanine nucleotide exchange factor (GEF) 4', map_location='2q22', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130836916, end_position_on_the_genomic_accession=131047253, orientation='plus', exon_count=21, OMIM=605216, no_hdr0=''),
  50650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50650, CurrentID=0, Status='live', Symbol='ARHGEF3', Aliases='GEF3, STA3, XPLN', description='Rho guanine nucleotide exchange factor 3', other_designations='59.8 kDA protein|Rho guanine nucleotide exchange factor (GEF) 3|RhoGEF protein|exchange factor found in platelets and leukemic and neuronal tissues, XPLN', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56727418, end_position_on_the_genomic_accession=57079308, orientation='minus', exon_count=21, OMIM=612115, no_hdr0=''),
  50651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50651, CurrentID=0, Status='live', Symbol='SLC45A1', Aliases='DNB5', description='solute carrier family 45 member 1', other_designations='H+/sugar symporter|PAST-A|deleted in neuroblastoma 5 protein|solute carrier family 45, member 1', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8318085, end_position_on_the_genomic_accession=8344167, orientation='plus', exon_count=9, OMIM=605763, no_hdr0=''),
  50652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50652, CurrentID=0, Status='live', Symbol='PCA3', Aliases='DD3, NCRNA00019, PCAT3', description='prostate cancer associated 3 (non-protein coding)', other_designations='prostate cancer antigen 3 (non-protein coding)|prostate cancer associated transcript 3 (non-protein coding)|prostate-specific gene DD3', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76764438, end_position_on_the_genomic_accession=76787549, orientation='plus', exon_count=3, OMIM=604845, no_hdr0=''),
  106481117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481117, CurrentID=0, Status='live', Symbol='RN7SL712P', Aliases='', description='RNA, 7SL, cytoplasmic 712, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481118, CurrentID=0, Status='live', Symbol='RN7SL717P', Aliases='', description='RNA, 7SL, cytoplasmic 717, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481119, CurrentID=0, Status='live', Symbol='RN7SL719P', Aliases='', description='RNA, 7SL, cytoplasmic 719, pseudogene', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481120, CurrentID=0, Status='live', Symbol='RN7SL725P', Aliases='', description='RNA, 7SL, cytoplasmic 725, pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481121, CurrentID=0, Status='live', Symbol='RN7SL727P', Aliases='', description='RNA, 7SL, cytoplasmic 727, pseudogene', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481122, CurrentID=0, Status='live', Symbol='RN7SL732P', Aliases='', description='RNA, 7SL, cytoplasmic 732, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481123, CurrentID=0, Status='live', Symbol='RN7SL737P', Aliases='', description='RNA, 7SL, cytoplasmic 737, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481124, CurrentID=0, Status='live', Symbol='RN7SL741P', Aliases='', description='RNA, 7SL, cytoplasmic 741, pseudogene', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481125, CurrentID=0, Status='live', Symbol='RN7SL748P', Aliases='', description='RNA, 7SL, cytoplasmic 748, pseudogene', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481126, CurrentID=0, Status='live', Symbol='RN7SL750P', Aliases='', description='RNA, 7SL, cytoplasmic 750, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481127, CurrentID=0, Status='live', Symbol='RN7SL752P', Aliases='', description='RNA, 7SL, cytoplasmic 752, pseudogene', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481128, CurrentID=0, Status='live', Symbol='RN7SL756P', Aliases='', description='RNA, 7SL, cytoplasmic 756, pseudogene', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  139516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139516, CurrentID=0, Status='live', Symbol='AKAP17BP', Aliases='AKAP16B, AKAP16BP', description='A-kinase anchoring protein 17B, pseudogene', other_designations='A kinase (PRKA) anchor protein 16B, pseudogene|A kinase (PRKA) anchor protein 17B, pseudogene|splicing factor, arginine/serine-rich 17A pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119256918, end_position_on_the_genomic_accession=119257658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481130, CurrentID=0, Status='live', Symbol='RN7SL763P', Aliases='', description='RNA, 7SL, cytoplasmic 763, pseudogene', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481131, CurrentID=0, Status='live', Symbol='RN7SL767P', Aliases='', description='RNA, 7SL, cytoplasmic 767, pseudogene', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481132, CurrentID=0, Status='live', Symbol='RN7SL771P', Aliases='', description='RNA, 7SL, cytoplasmic 771, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481133, CurrentID=0, Status='live', Symbol='RN7SL782P', Aliases='', description='RNA, 7SL, cytoplasmic 782, pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481134, CurrentID=0, Status='live', Symbol='RN7SL784P', Aliases='', description='RNA, 7SL, cytoplasmic 784, pseudogene', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481135, CurrentID=0, Status='live', Symbol='RN7SL788P', Aliases='', description='RNA, 7SL, cytoplasmic 788, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481136, CurrentID=0, Status='live', Symbol='RN7SL790P', Aliases='', description='RNA, 7SL, cytoplasmic 790, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481137, CurrentID=0, Status='live', Symbol='RN7SL793P', Aliases='', description='RNA, 7SL, cytoplasmic 793, pseudogene', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50674, CurrentID=0, Status='live', Symbol='NEUROG3', Aliases='Atoh5, Math4B, NGN-3, bHLHa7, ngn3', description='neurogenin 3', other_designations='class A basic helix-loop-helix protein 7|protein atonal homolog 5', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69572035, end_position_on_the_genomic_accession=69573454, orientation='minus', exon_count=2, OMIM=604882, no_hdr0=''),
  106481139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481139, CurrentID=0, Status='live', Symbol='RN7SL801P', Aliases='', description='RNA, 7SL, cytoplasmic 801, pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481140, CurrentID=0, Status='live', Symbol='RN7SL804P', Aliases='', description='RNA, 7SL, cytoplasmic 804, pseudogene', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481141, CurrentID=0, Status='live', Symbol='RN7SL807P', Aliases='', description='RNA, 7SL, cytoplasmic 807, pseudogene', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481142, CurrentID=0, Status='live', Symbol='RN7SL810P', Aliases='', description='RNA, 7SL, cytoplasmic 810, pseudogene', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481143, CurrentID=0, Status='live', Symbol='RN7SL814P', Aliases='', description='RNA, 7SL, cytoplasmic 814, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481144, CurrentID=0, Status='live', Symbol='RN7SL817P', Aliases='', description='RNA, 7SL, cytoplasmic 817, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481145, CurrentID=0, Status='live', Symbol='RN7SL819P', Aliases='', description='RNA, 7SL, cytoplasmic 819, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481146, CurrentID=0, Status='live', Symbol='RN7SL824P', Aliases='', description='RNA, 7SL, cytoplasmic 824, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401663, CurrentID=0, Status='live', Symbol='OR51H1', Aliases='OR51H1P', description='olfactory receptor family 51 subfamily H member 1', other_designations='Olfactory receptor 51H1|olfactory receptor OR11-25|olfactory receptor, family 51, subfamily H, member 1 pseudogene|seven transmembrane helix receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4859619, end_position_on_the_genomic_accession=4860563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481148, CurrentID=0, Status='live', Symbol='RN7SL834P', Aliases='', description='RNA, 7SL, cytoplasmic 834, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481149, CurrentID=0, Status='live', Symbol='RN7SL836P', Aliases='', description='RNA, 7SL, cytoplasmic 836, pseudogene', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481150, CurrentID=0, Status='live', Symbol='RN7SL841P', Aliases='', description='RNA, 7SL, cytoplasmic 841, pseudogene', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481151, CurrentID=0, Status='live', Symbol='RN7SL843P', Aliases='', description='RNA, 7SL, cytoplasmic 843, pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481152, CurrentID=0, Status='live', Symbol='RN7SL847P', Aliases='', description='RNA, 7SL, cytoplasmic 847, pseudogene', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401664, CurrentID=0, Status='live', Symbol='OR51H2P', Aliases='', description='olfactory receptor family 51 subfamily H member 2 pseudogene', other_designations='olfactory receptor, family 51, subfamily H, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4876557, end_position_on_the_genomic_accession=4877503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481154, CurrentID=0, Status='live', Symbol='RN7SL865P', Aliases='', description='RNA, 7SL, cytoplasmic 865, pseudogene', other_designations='', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481155, CurrentID=0, Status='live', Symbol='RN7SL870P', Aliases='', description='RNA, 7SL, cytoplasmic 870, pseudogene', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481156, CurrentID=0, Status='live', Symbol='RNU4ATAC5P', Aliases='', description='RNA, U4atac small nuclear 5, pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481157, CurrentID=0, Status='live', Symbol='RNU4ATAC8P', Aliases='', description='RNA, U4atac small nuclear 8, pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481158, CurrentID=0, Status='live', Symbol='RNU4ATAC11P', Aliases='', description='RNA, U4atac small nuclear 11, pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8449, CurrentID=0, Status='live', Symbol='DHX16', Aliases='DBP2, DDX16, PRO2014, PRP8, PRPF2, Prp2', description='DEAH (Asp-Glu-Ala-His) box polypeptide 16', other_designations='ATP-dependent RNA helicase #3|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 16|DEAD/H box 16|DEAH-box protein 16|RNA helicase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30653119, end_position_on_the_genomic_accession=30673053, orientation='minus', exon_count=24, OMIM=603405, no_hdr0=''),
  106481160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481160, CurrentID=0, Status='live', Symbol='RNU4ATAC16P', Aliases='', description='RNA, U4atac small nuclear 16, pseudogene', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481161, CurrentID=0, Status='live', Symbol='RNU6ATAC8P', Aliases='', description='RNA, U6atac small nuclear 8, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481162, CurrentID=0, Status='live', Symbol='RNU6ATAC11P', Aliases='', description='RNA, U6atac small nuclear 11, pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481163, CurrentID=0, Status='live', Symbol='RNU6ATAC14P', Aliases='', description='RNA, U6atac small nuclear 14, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50700, CurrentID=0, Status='live', Symbol='RDH8', Aliases='PRRDH, SDR28C2', description='retinol dehydrogenase 8 (all-trans)', other_designations='photoreceptor outer segment all-trans retinol dehydrogenase|short chain dehydrogenase/reductase family 28C member 2|short chain dehydrogenase/reductase family 28C, member 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10013249, end_position_on_the_genomic_accession=10022278, orientation='plus', exon_count=6, OMIM=608575, no_hdr0=''),
  401666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401666, CurrentID=0, Status='live', Symbol='OR51A4', Aliases='', description='olfactory receptor family 51 subfamily A member 4', other_designations='olfactory receptor, family 51, subfamily A, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4946159, end_position_on_the_genomic_accession=4947100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481166, CurrentID=0, Status='live', Symbol='RNU6ATAC23P', Aliases='', description='RNA, U6atac small nuclear 23, pseudogene', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481167, CurrentID=0, Status='live', Symbol='RNU6ATAC27P', Aliases='', description='RNA, U6atac small nuclear 27, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481168, CurrentID=0, Status='live', Symbol='RNU6ATAC29P', Aliases='', description='RNA, U6atac small nuclear 29, pseudogene', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481169, CurrentID=0, Status='live', Symbol='RNU6ATAC31P', Aliases='', description='RNA, U6atac small nuclear 31, pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481170, CurrentID=0, Status='live', Symbol='RNU6ATAC33P', Aliases='', description='RNA, U6atac small nuclear 33, pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635523, CurrentID=0, Status='live', Symbol='SNORA90', Aliases='', description='small nucleolar RNA, H/ACA box 90', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481172, CurrentID=0, Status='live', Symbol='RNU6ATAC39P', Aliases='', description='RNA, U6atac small nuclear 39, pseudogene', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481173, CurrentID=0, Status='live', Symbol='RNU6ATAC42P', Aliases='', description='RNA, U6atac small nuclear 42, pseudogene', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481174, CurrentID=0, Status='live', Symbol='RNU11-4P', Aliases='', description='RNA, U11 small nuclear 4, pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481175, CurrentID=0, Status='live', Symbol='RNU4-11P', Aliases='', description='RNA, U4 small nuclear 11, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481176, CurrentID=0, Status='live', Symbol='RNU4-14P', Aliases='', description='RNA, U4 small nuclear 14, pseudogene', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481177, CurrentID=0, Status='live', Symbol='RNU4-17P', Aliases='', description='RNA, U4 small nuclear 17, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481178, CurrentID=0, Status='live', Symbol='RNU4-21P', Aliases='', description='RNA, U4 small nuclear 21, pseudogene', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50715, CurrentID=0, Status='live', Symbol='EJM2', Aliases='EIG7, JME', description='epilepsy, juvenile myoclonic 2', other_designations='myoclonic epilepsy, juvenile, 2', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604827, no_hdr0=''),
  106481180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481180, CurrentID=0, Status='live', Symbol='RNU4-27P', Aliases='', description='RNA, U4 small nuclear 27, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50717, CurrentID=0, Status='live', Symbol='DCAF8', Aliases='GAN2, H326, WDR42A', description='DDB1 and CUL4 associated factor 8', other_designations='WD repeat domain 42A|WD repeat-containing protein 42A', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160215715, end_position_on_the_genomic_accession=160262560, orientation='minus', exon_count=14, OMIM=615820, no_hdr0=''),
  106481182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481182, CurrentID=0, Status='live', Symbol='RNU4-35P', Aliases='', description='RNA, U4 small nuclear 35, pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729349, CurrentID=0, Status='live', Symbol='RPL17P16', Aliases='RPL17_4_369', description='ribosomal protein L17 pseudogene 16', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48120623, end_position_on_the_genomic_accession=48121287, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481184, CurrentID=0, Status='live', Symbol='RNU4-41P', Aliases='', description='RNA, U4 small nuclear 41, pseudogene', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481185, CurrentID=0, Status='live', Symbol='RNU4-43P', Aliases='', description='RNA, U4 small nuclear 43, pseudogene', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481186, CurrentID=0, Status='live', Symbol='RNU4-45P', Aliases='', description='RNA, U4 small nuclear 45, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481187, CurrentID=0, Status='live', Symbol='RNU4-49P', Aliases='', description='RNA, U4 small nuclear 49, pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481188, CurrentID=0, Status='live', Symbol='RNU4-52P', Aliases='', description='RNA, U4 small nuclear 52, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481189, CurrentID=0, Status='live', Symbol='RNU4-55P', Aliases='', description='RNA, U4 small nuclear 55, pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481190, CurrentID=0, Status='live', Symbol='RNU4-58P', Aliases='', description='RNA, U4 small nuclear 58, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481191, CurrentID=0, Status='live', Symbol='RNU4-61P', Aliases='', description='RNA, U4 small nuclear 61, pseudogene', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481192, CurrentID=0, Status='live', Symbol='RNU4-65P', Aliases='', description='RNA, U4 small nuclear 65, pseudogene', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481193, CurrentID=0, Status='live', Symbol='RNU4-67P', Aliases='', description='RNA, U4 small nuclear 67, pseudogene', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481194, CurrentID=0, Status='live', Symbol='RNU4-71P', Aliases='', description='RNA, U4 small nuclear 71, pseudogene', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481195, CurrentID=0, Status='live', Symbol='RNU4-73P', Aliases='', description='RNA, U4 small nuclear 73, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481196, CurrentID=0, Status='live', Symbol='RNU4-77P', Aliases='', description='RNA, U4 small nuclear 77, pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481197, CurrentID=0, Status='live', Symbol='RNU4-79P', Aliases='', description='RNA, U4 small nuclear 79, pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481198, CurrentID=0, Status='live', Symbol='RNU4-82P', Aliases='', description='RNA, U4 small nuclear 82, pseudogene', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481199, CurrentID=0, Status='live', Symbol='RNU4-86P', Aliases='', description='RNA, U4 small nuclear 86, pseudogene', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481200, CurrentID=0, Status='live', Symbol='RNU4-89P', Aliases='', description='RNA, U4 small nuclear 89, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481201, CurrentID=0, Status='live', Symbol='RNU4-91P', Aliases='', description='RNA, U4 small nuclear 91, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481202, CurrentID=0, Status='live', Symbol='RPPH1-2P', Aliases='', description='ribonuclease P RNA component H1, 2 pseudogene', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481203, CurrentID=0, Status='live', Symbol='RNU6-85P', Aliases='', description='RNA, U6 small nuclear 85, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481204, CurrentID=0, Status='live', Symbol='RNU6-87P', Aliases='', description='RNA, U6 small nuclear 87, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481205, CurrentID=0, Status='live', Symbol='RNU6-90P', Aliases='', description='RNA, U6 small nuclear 90, pseudogene', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481206, CurrentID=0, Status='live', Symbol='RNU6-94P', Aliases='', description='RNA, U6 small nuclear 94, pseudogene', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481207, CurrentID=0, Status='live', Symbol='RNU6-96P', Aliases='', description='RNA, U6 small nuclear 96, pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481208, CurrentID=0, Status='live', Symbol='RNU6-100P', Aliases='', description='RNA, U6 small nuclear 100, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481209, CurrentID=0, Status='live', Symbol='RNU6-102P', Aliases='', description='RNA, U6 small nuclear 102, pseudogene', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481210, CurrentID=0, Status='live', Symbol='RNU6-105P', Aliases='', description='RNA, U6 small nuclear 105, pseudogene', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481211, CurrentID=0, Status='live', Symbol='RNU6-109P', Aliases='', description='RNA, U6 small nuclear 109, pseudogene', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481212, CurrentID=0, Status='live', Symbol='RNU6-111P', Aliases='', description='RNA, U6 small nuclear 111, pseudogene', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481213, CurrentID=0, Status='live', Symbol='RNU6-117P', Aliases='', description='RNA, U6 small nuclear 117, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481214, CurrentID=0, Status='live', Symbol='RNU6-119P', Aliases='', description='RNA, U6 small nuclear 119, pseudogene', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481215, CurrentID=0, Status='live', Symbol='RNU6-121P', Aliases='', description='RNA, U6 small nuclear 121, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481216, CurrentID=0, Status='live', Symbol='RNU6-124P', Aliases='', description='RNA, U6 small nuclear 124, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481217, CurrentID=0, Status='live', Symbol='RNU6-128P', Aliases='', description='RNA, U6 small nuclear 128, pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481218, CurrentID=0, Status='live', Symbol='RNU6-130P', Aliases='', description='RNA, U6 small nuclear 130, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635531, CurrentID=0, Status='live', Symbol='SNORA99', Aliases='', description='small nucleolar RNA, H/ACA box 99', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481220, CurrentID=0, Status='live', Symbol='RNU6-137P', Aliases='', description='RNA, U6 small nuclear 137, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481221, CurrentID=0, Status='live', Symbol='RNU6-140P', Aliases='', description='RNA, U6 small nuclear 140, pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481222, CurrentID=0, Status='live', Symbol='RNU6-144P', Aliases='', description='RNA, U6 small nuclear 144, pseudogene', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481223, CurrentID=0, Status='live', Symbol='RNU6-152P', Aliases='', description='RNA, U6 small nuclear 152, pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481224, CurrentID=0, Status='live', Symbol='RNU6-155P', Aliases='', description='RNA, U6 small nuclear 155, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481225, CurrentID=0, Status='live', Symbol='RNU6-158P', Aliases='', description='RNA, U6 small nuclear 158, pseudogene', other_designations='', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481226, CurrentID=0, Status='live', Symbol='RNU6-162P', Aliases='', description='RNA, U6 small nuclear 162, pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481227, CurrentID=0, Status='live', Symbol='RNU6-165P', Aliases='', description='RNA, U6 small nuclear 165, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481228, CurrentID=0, Status='live', Symbol='RNU6-167P', Aliases='', description='RNA, U6 small nuclear 167, pseudogene', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481229, CurrentID=0, Status='live', Symbol='RNU6-169P', Aliases='', description='RNA, U6 small nuclear 169, pseudogene', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481230, CurrentID=0, Status='live', Symbol='RNU6-171P', Aliases='', description='RNA, U6 small nuclear 171, pseudogene', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481231, CurrentID=0, Status='live', Symbol='RNU6-174P', Aliases='', description='RNA, U6 small nuclear 174, pseudogene', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481232, CurrentID=0, Status='live', Symbol='RNU6-177P', Aliases='', description='RNA, U6 small nuclear 177, pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481233, CurrentID=0, Status='live', Symbol='RNU6-180P', Aliases='', description='RNA, U6 small nuclear 180, pseudogene', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481234, CurrentID=0, Status='live', Symbol='RNU6-183P', Aliases='', description='RNA, U6 small nuclear 183, pseudogene', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481235, CurrentID=0, Status='live', Symbol='RNU6-185P', Aliases='', description='RNA, U6 small nuclear 185, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481236, CurrentID=0, Status='live', Symbol='RNU6-190P', Aliases='', description='RNA, U6 small nuclear 190, pseudogene', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481237, CurrentID=0, Status='live', Symbol='RNU6-192P', Aliases='', description='RNA, U6 small nuclear 192, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481238, CurrentID=0, Status='live', Symbol='RNU6-196P', Aliases='', description='RNA, U6 small nuclear 196, pseudogene', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481239, CurrentID=0, Status='live', Symbol='RNU6-198P', Aliases='', description='RNA, U6 small nuclear 198, pseudogene', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481240, CurrentID=0, Status='live', Symbol='RNU6-201P', Aliases='', description='RNA, U6 small nuclear 201, pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481241, CurrentID=0, Status='live', Symbol='RNU6-204P', Aliases='', description='RNA, U6 small nuclear 204, pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481242, CurrentID=0, Status='live', Symbol='RNU6-209P', Aliases='', description='RNA, U6 small nuclear 209, pseudogene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635535, CurrentID=0, Status='live', Symbol='RN7SKP173', Aliases='', description='RNA, 7SK small nuclear pseudogene 173', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481244, CurrentID=0, Status='live', Symbol='RNU6-213P', Aliases='', description='RNA, U6 small nuclear 213, pseudogene', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481245, CurrentID=0, Status='live', Symbol='RNU6-215P', Aliases='', description='RNA, U6 small nuclear 215, pseudogene', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481246, CurrentID=0, Status='live', Symbol='RNU6-217P', Aliases='', description='RNA, U6 small nuclear 217, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481247, CurrentID=0, Status='live', Symbol='RNU6-220P', Aliases='', description='RNA, U6 small nuclear 220, pseudogene', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481248, CurrentID=0, Status='live', Symbol='RNU6-223P', Aliases='', description='RNA, U6 small nuclear 223, pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104472848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472848, CurrentID=0, Status='live', Symbol='IRAIN', Aliases='IGF1R-AS', description='IGF1R antisense imprinted non-protein coding RNA', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98645863, end_position_on_the_genomic_accession=98651221, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106481250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481250, CurrentID=0, Status='live', Symbol='RNU6-229P', Aliases='', description='RNA, U6 small nuclear 229, pseudogene', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481251, CurrentID=0, Status='live', Symbol='RNU6-231P', Aliases='', description='RNA, U6 small nuclear 231, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481252, CurrentID=0, Status='live', Symbol='RNU6-236P', Aliases='', description='RNA, U6 small nuclear 236, pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481253, CurrentID=0, Status='live', Symbol='RNU6-238P', Aliases='', description='RNA, U6 small nuclear 238, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481254, CurrentID=0, Status='live', Symbol='RNU6-242P', Aliases='', description='RNA, U6 small nuclear 242, pseudogene', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481255, CurrentID=0, Status='live', Symbol='RNU6-245P', Aliases='', description='RNA, U6 small nuclear 245, pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481256, CurrentID=0, Status='live', Symbol='RNU6-248P', Aliases='', description='RNA, U6 small nuclear 248, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481257, CurrentID=0, Status='live', Symbol='RNU6-251P', Aliases='', description='RNA, U6 small nuclear 251, pseudogene', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481258, CurrentID=0, Status='live', Symbol='RNU6-255P', Aliases='', description='RNA, U6 small nuclear 255, pseudogene', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481259, CurrentID=0, Status='live', Symbol='RNU6-258P', Aliases='', description='RNA, U6 small nuclear 258, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481260, CurrentID=0, Status='live', Symbol='RNU6-259P', Aliases='', description='RNA, U6 small nuclear 259, pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  8466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8466, CurrentID=0, Status='live', Symbol='ST11', Aliases='PETS1', description='suppression of tumorigenicity 11 (pancreas)', other_designations='Pancreatic endocrine tumor suppressor', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602011, no_hdr0=''),
  106481262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481262, CurrentID=0, Status='live', Symbol='RNU6-264P', Aliases='', description='RNA, U6 small nuclear 264, pseudogene', other_designations='', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481263, CurrentID=0, Status='live', Symbol='RNU6-269P', Aliases='', description='RNA, U6 small nuclear 269, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481264, CurrentID=0, Status='live', Symbol='RNU6-271P', Aliases='', description='RNA, U6 small nuclear 271, pseudogene', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50801, CurrentID=0, Status='live', Symbol='KCNK4', Aliases='K2p4.1, TRAAK, TRAAK1', description='potassium channel, two pore domain subfamily K, member 4', other_designations='K2P4.1 potassium channel|TWIK-related arachidonic acid-stimulated potassium channel protein|potassium channel, subfamily K, member 4|two pore K(+) channel KT4.1|two pore K+ channel KT4.1|two pore potassium channel KT4.1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64291321, end_position_on_the_genomic_accession=64300031, orientation='plus', exon_count=7, OMIM=605720, no_hdr0=''),
  50802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50802, CurrentID=0, Status='live', Symbol='IGK', Aliases='IGK@', description='immunoglobulin kappa locus', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89851758, end_position_on_the_genomic_accession=90235368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481267, CurrentID=0, Status='live', Symbol='RNU6-281P', Aliases='', description='RNA, U6 small nuclear 281, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50804, CurrentID=0, Status='live', Symbol='MYEF2', Aliases='HsT18564, MEF-2, MST156, MSTP156, myEF-2', description='myelin expression factor 2', other_designations='myelin gene expression factor 2', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48139428, end_position_on_the_genomic_accession=48178400, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  50805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50805, CurrentID=0, Status='live', Symbol='IRX4', Aliases='IRXA3', description='iroquois homeobox 4', other_designations='homeodomain protein IRXA3|iroquois homeobox protein 4', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1877427, end_position_on_the_genomic_accession=1887179, orientation='minus', exon_count=8, OMIM=606199, no_hdr0=''),
  106481270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481270, CurrentID=0, Status='live', Symbol='RNU6-290P', Aliases='', description='RNA, U6 small nuclear 290, pseudogene', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50807, CurrentID=0, Status='live', Symbol='ASAP1', Aliases='AMAP1, CENTB4, DDEF1, PAG2, PAP, ZG14P', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain 1', other_designations='130 kDa phosphatidylinositol 4,5-biphosphate-dependent ARF1 GTPase-activating protein|130 kDa phosphatidylinositol 4,5-bisphosphate-dependent ARF1 GTPase-activating protein|ADP-ribosylation factor-directed GTPase-activating protein 1|ARF GTPase-activating protein 1|DEF-1|PIP2-dependent ARF1 GAP|centaurin, beta 4|development and differentiation-enhancing factor 1', map_location='8q24.1-q24.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130052104, end_position_on_the_genomic_accession=130443660, orientation='minus', exon_count=36, OMIM=605953, no_hdr0=''),
  50808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50808, CurrentID=0, Status='live', Symbol='AK3', Aliases='AK3L1, AK6, AKL3L, AKL3L1, FIX', description='adenylate kinase 3', other_designations='GTP:AMP phosphotransferase, mitochondrial|adenylate kinase 3 alpha-like 1|adenylate kinase 6, adenylate kinase 3 like 1', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4709556, end_position_on_the_genomic_accession=4742043, orientation='minus', exon_count=8, OMIM=609290, no_hdr0=''),
  50809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50809, CurrentID=0, Status='live', Symbol='HP1BP3', Aliases='HP1-BP74, HP1BP74', description='heterochromatin protein 1, binding protein 3', other_designations='HP1-BP74', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20742677, end_position_on_the_genomic_accession=20787310, orientation='minus', exon_count=15, OMIM=616072, no_hdr0=''),
  50810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50810, CurrentID=0, Status='live', Symbol='HDGFRP3', Aliases='CGI-142, HDGF-2, HDGF2, HRP-3', description='hepatoma-derived growth factor, related protein 3', other_designations='hepatoma-derived growth factor 2', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83115376, end_position_on_the_genomic_accession=83208018, orientation='minus', exon_count=7, OMIM=616643, no_hdr0=''),
  106481275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481275, CurrentID=0, Status='live', Symbol='RNU6-304P', Aliases='', description='RNA, U6 small nuclear 304, pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481276, CurrentID=0, Status='live', Symbol='RNU6-307P', Aliases='', description='RNA, U6 small nuclear 307, pseudogene', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50813, CurrentID=0, Status='live', Symbol='COPS7A', Aliases='CSN7, CSN7A, SGN7a', description='COP9 signalosome subunit 7A', other_designations='COP9 complex subunit 7a|COP9 constitutive photomorphogenic homolog subunit 7A|JAB1-containing signalosome subunit 7a|dermal papilla-derived protein 10', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6723984, end_position_on_the_genomic_accession=6731875, orientation='plus', exon_count=9, OMIM=616009, no_hdr0=''),
  50814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50814, CurrentID=0, Status='live', Symbol='NSDHL', Aliases='H105E3, SDR31E1, XAP104', description='NAD(P) dependent steroid dehydrogenase-like', other_designations='protein H105e3|short chain dehydrogenase/reductase family 31E, member 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152830967, end_position_on_the_genomic_accession=152869363, orientation='plus', exon_count=10, OMIM=300275, no_hdr0=''),
  106481279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481279, CurrentID=0, Status='live', Symbol='RNU6-316P', Aliases='', description='RNA, U6 small nuclear 316, pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481280, CurrentID=0, Status='live', Symbol='RNU6-319P', Aliases='', description='RNA, U6 small nuclear 319, pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481281, CurrentID=0, Status='live', Symbol='RNU6-323P', Aliases='', description='RNA, U6 small nuclear 323, pseudogene', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481282, CurrentID=0, Status='live', Symbol='RNU6-326P', Aliases='', description='RNA, U6 small nuclear 326, pseudogene', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481283, CurrentID=0, Status='live', Symbol='RNU6-330P', Aliases='', description='RNA, U6 small nuclear 330, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481284, CurrentID=0, Status='live', Symbol='RNU6-332P', Aliases='', description='RNA, U6 small nuclear 332, pseudogene', other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  139542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139542, CurrentID=0, Status='live', Symbol='E2F6P4', Aliases='', description='E2F transcription factor 6 pseudogene 4', other_designations='E2F6 pseudogene 4', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136097499, end_position_on_the_genomic_accession=136098347, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481286, CurrentID=0, Status='live', Symbol='RNU6-338P', Aliases='', description='RNA, U6 small nuclear 338, pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50823, CurrentID=0, Status='live', Symbol='HMU', Aliases='MUHH', description='Hypotrichosis, Marie Unna type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146550, no_hdr0=''),
  106481288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481288, CurrentID=0, Status='live', Symbol='RNU6-345P', Aliases='', description='RNA, U6 small nuclear 345, pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481289, CurrentID=0, Status='live', Symbol='RNU6-348P', Aliases='', description='RNA, U6 small nuclear 348, pseudogene', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481290, CurrentID=0, Status='live', Symbol='RNU6-354P', Aliases='', description='RNA, U6 small nuclear 354, pseudogene', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401687, CurrentID=0, Status='live', Symbol='OR5J1P', Aliases='HTPCRH02, OR11-169, OR5J1', description='olfactory receptor family 5 subfamily J member 1 pseudogene', other_designations='olfactory receptor OR11-169 pseudogene|olfactory receptor, family 5, subfamily J, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56071111, end_position_on_the_genomic_accession=56072046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481292, CurrentID=0, Status='live', Symbol='RNU6-362P', Aliases='', description='RNA, U6 small nuclear 362, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50829, CurrentID=0, Status='live', Symbol='HDPA', Aliases='', description='Hodgkin disease, susceptibility, pseudoautosomal', other_designations='', map_location='Xpter-p22.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300221, no_hdr0=''),
  50830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50830, CurrentID=0, Status='live', Symbol='TGCT1', Aliases='', description='testicular germ cell tumor susceptibility 1', other_designations='Testicular germ cell tumor 1', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300228, no_hdr0=''),
  50831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50831, CurrentID=0, Status='live', Symbol='TAS2R3', Aliases='T2R3', description='taste 2 receptor member 3', other_designations='candidate taste receptor T2R3|taste receptor, type 2, member 3', map_location='7q31.3-q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141764097, end_position_on_the_genomic_accession=141765197, orientation='plus', exon_count=1, OMIM=604868, no_hdr0=''),
  50832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50832, CurrentID=0, Status='live', Symbol='TAS2R4', Aliases='T2R4', description='taste 2 receptor member 4', other_designations='candidate taste receptor T2R4|taste receptor, type 2, member 4', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141778489, end_position_on_the_genomic_accession=141779388, orientation='plus', exon_count=1, OMIM=604869, no_hdr0=''),
  50833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50833, CurrentID=0, Status='live', Symbol='TAS2R16', Aliases='T2R16', description='taste 2 receptor member 16', other_designations='candidate taste receptor T2R16|taste receptor, type 2, member 16', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122994705, end_position_on_the_genomic_accession=122995700, orientation='minus', exon_count=1, OMIM=604867, no_hdr0=''),
  50834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50834, CurrentID=0, Status='live', Symbol='TAS2R1', Aliases='T2R1, TRB7', description='taste 2 receptor member 1', other_designations='taste receptor family B member 7|taste receptor, family B, member 7|taste receptor, type 2, member 1', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9628997, end_position_on_the_genomic_accession=9630351, orientation='minus', exon_count=1, OMIM=604796, no_hdr0=''),
  50835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50835, CurrentID=0, Status='live', Symbol='TAS2R9', Aliases='T2R9, TRB6', description='taste 2 receptor member 9', other_designations='taste receptor family B member 6|taste receptor, family B, member 6|taste receptor, type 2, member 9', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10809094, end_position_on_the_genomic_accession=10810168, orientation='minus', exon_count=1, OMIM=604795, no_hdr0=''),
  50836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50836, CurrentID=0, Status='live', Symbol='TAS2R8', Aliases='T2R8, TRB5', description='taste 2 receptor member 8', other_designations='taste receptor family B member 5|taste receptor, family B, member 5|taste receptor, type 2, member 8', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10806051, end_position_on_the_genomic_accession=10806980, orientation='minus', exon_count=1, OMIM=604794, no_hdr0=''),
  50837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50837, CurrentID=0, Status='live', Symbol='TAS2R7', Aliases='T2R7, TRB4', description='taste 2 receptor member 7', other_designations='taste receptor family B member 4|taste receptor, family B, member 4|taste receptor, type 2, member 7', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10801532, end_position_on_the_genomic_accession=10802627, orientation='minus', exon_count=1, OMIM=604793, no_hdr0=''),
  50838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50838, CurrentID=0, Status='live', Symbol='TAS2R13', Aliases='T2R13, TRB3', description='taste 2 receptor member 13', other_designations='taste receptor family B member 3|taste receptor, family B, member 3|taste receptor, type 2, member 13', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10907926, end_position_on_the_genomic_accession=10909562, orientation='minus', exon_count=1, OMIM=604792, no_hdr0=''),
  50839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50839, CurrentID=0, Status='live', Symbol='TAS2R10', Aliases='T2R10, TRB2', description='taste 2 receptor member 10', other_designations='taste receptor family B member 2|taste receptor, family B, member 2|taste receptor, type 2, member 10', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10815851, end_position_on_the_genomic_accession=10828965, orientation='minus', exon_count=2, OMIM=604791, no_hdr0=''),
  50840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50840, CurrentID=0, Status='live', Symbol='TAS2R14', Aliases='T2R14, TRB1', description='taste 2 receptor member 14', other_designations='taste receptor family B member 1|taste receptor, family B, member 1|taste receptor, type 2, member 14', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10938254, end_position_on_the_genomic_accession=10939207, orientation='minus', exon_count=1, OMIM=604790, no_hdr0=''),
  106481305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481305, CurrentID=0, Status='live', Symbol='RNU6-402P', Aliases='', description='RNA, U6 small nuclear 402, pseudogene', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481306, CurrentID=0, Status='live', Symbol='RNU6-406P', Aliases='', description='RNA, U6 small nuclear 406, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481307, CurrentID=0, Status='live', Symbol='RNU6-409P', Aliases='', description='RNA, U6 small nuclear 409, pseudogene', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481308, CurrentID=0, Status='live', Symbol='RNU6-412P', Aliases='', description='RNA, U6 small nuclear 412, pseudogene', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481309, CurrentID=0, Status='live', Symbol='RNU6-415P', Aliases='', description='RNA, U6 small nuclear 415, pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50846, CurrentID=0, Status='live', Symbol='DHH', Aliases='GDXYM, HHG-3, SRXY7', description='desert hedgehog', other_designations='desert hedgehog homolog|mutant desert hedgehog', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49089421, end_position_on_the_genomic_accession=49094819, orientation='minus', exon_count=3, OMIM=605423, no_hdr0=''),
  106481311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481311, CurrentID=0, Status='live', Symbol='RNU6-421P', Aliases='', description='RNA, U6 small nuclear 421, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50848, CurrentID=0, Status='live', Symbol='F11R', Aliases='CD321, JAM, JAM1, JAMA, JCAM, KAT, PAM-1', description='F11 receptor', other_designations='junctional adhesion molecule 1|platelet F11 receptor|platelet adhesion molecule 1', map_location='1q21.2-q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160995211, end_position_on_the_genomic_accession=161021343, orientation='minus', exon_count=10, OMIM=605721, no_hdr0=''),
  106481313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481313, CurrentID=0, Status='live', Symbol='RNU6-427P', Aliases='', description='RNA, U6 small nuclear 427, pseudogene', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481314, CurrentID=0, Status='live', Symbol='RNU6-431P', Aliases='', description='RNA, U6 small nuclear 431, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481315, CurrentID=0, Status='live', Symbol='RNU6-433P', Aliases='', description='RNA, U6 small nuclear 433, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50852, CurrentID=0, Status='live', Symbol='TRAT1', Aliases='HSPC062, TCRIM, TRIM, pp29/30', description='T cell receptor associated transmembrane adaptor 1', other_designations='T cell receptor interacting molecule', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108822784, end_position_on_the_genomic_accession=108854867, orientation='plus', exon_count=6, OMIM=604962, no_hdr0=''),
  50853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50853, CurrentID=0, Status='live', Symbol='VILL', Aliases='', description='villin-like', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37987975, end_position_on_the_genomic_accession=38007185, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  50854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50854, CurrentID=0, Status='live', Symbol='C6orf48', Aliases='D6S57, G8', description='chromosome 6 open reading frame 48', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31834915, end_position_on_the_genomic_accession=31839766, orientation='plus', exon_count=5, OMIM=605447, no_hdr0=''),
  50855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50855, CurrentID=0, Status='live', Symbol='PARD6A', Aliases='PAR-6A, PAR6, PAR6C, PAR6alpha, TAX40, TIP-40', description='par-6 family cell polarity regulator alpha', other_designations='PAR-6 alpha|Tax-interacting protein 40|par-6 partitioning defective 6 homolog alpha|partitioning defective-6 homolog alpha|partitioning-defective protein 6|tax interaction protein 40', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67660948, end_position_on_the_genomic_accession=67662778, orientation='plus', exon_count=3, OMIM=607484, no_hdr0=''),
  50856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50856, CurrentID=0, Status='live', Symbol='CLEC4A', Aliases='CD367, CLECSF6, DCIR, DDB27, HDCGC13P, LLIR', description='C-type lectin domain family 4 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 6|C-type lectin DDB27|C-type lectin domain family 4, member A|C-type lectin superfamily member 6|dendritic cell immunoreceptor|lectin-like immunoreceptor', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8123632, end_position_on_the_genomic_accession=8138607, orientation='plus', exon_count=7, OMIM=605306, no_hdr0=''),
  401692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401692, CurrentID=0, Status='live', Symbol='OR5B1P', Aliases='OR3-144, OR5B14P, OR5B5, OR5B5P, OR5B6, OR5B6P, OR5B7, OR5B7P, OR5B8, OR5B8P, OR5B9, OR5B9P, OR6-55, OR6-57, OR8-122, OR8-123, OR912-92', description='olfactory receptor family 5 subfamily B member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily B, member 1 pseudogene|olfactory receptor, family 5, subfamily B, member 14 pseudogene|olfactory receptor, family 5, subfamily B, member 5 pseudogene|olfactory receptor, family 5, subfamily B, member 6 pseudogene|olfactory receptor, family 5, subfamily B, member 7 pseudogene|olfactory receptor, family 5, subfamily B, member 8 pseudogene|olfactory receptor, family 5, subfamily B, member 9 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58365921, end_position_on_the_genomic_accession=58366568, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50858, CurrentID=0, Status='live', Symbol='TTTY1', Aliases='LINC00033, NCRNA00033, TTY1', description='testis-specific transcript, Y-linked 1 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 33|testis-specific testis transcript Y 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9753156, end_position_on_the_genomic_accession=9774289, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50859, CurrentID=0, Status='live', Symbol='SPOCK3', Aliases='HSAJ1454, TES-3, TICN3', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan (testican) 3', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=166733384, end_position_on_the_genomic_accession=167234796, orientation='minus', exon_count=15, OMIM=607989, no_hdr0=''),
  106481324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481324, CurrentID=0, Status='live', Symbol='RNU6-458P', Aliases='', description='RNA, U6 small nuclear 458, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50861, CurrentID=0, Status='live', Symbol='STMN3', Aliases='SCLIP', description='stathmin 3', other_designations='SCG10-like protein|SCG10-like-protein|stathmin-like 3', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63639705, end_position_on_the_genomic_accession=63653610, orientation='minus', exon_count=6, OMIM=608362, no_hdr0=''),
  50862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50862, CurrentID=0, Status='live', Symbol='RNF141', Aliases='ZFP26, ZNF230', description='ring finger protein 141', other_designations='C3HC4-like zinc finger protein|zinc finger protein 230', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10511678, end_position_on_the_genomic_accession=10541227, orientation='minus', exon_count=6, OMIM=616641, no_hdr0=''),
  50863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50863, CurrentID=0, Status='live', Symbol='NTM', Aliases='HNT, IGLON2, NTRI', description='neurotrimin', other_designations='IgLON family member 2', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131370361, end_position_on_the_genomic_accession=132336822, orientation='plus', exon_count=15, OMIM=607938, no_hdr0=''),
  106481328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481328, CurrentID=0, Status='live', Symbol='RNU6-472P', Aliases='', description='RNA, U6 small nuclear 472, pseudogene', other_designations='', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50865, CurrentID=0, Status='live', Symbol='HEBP1', Aliases='HBP, HEBP', description='heme binding protein 1', other_designations='p22HBP', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12974864, end_position_on_the_genomic_accession=13000309, orientation='minus', exon_count=4, OMIM=605826, no_hdr0=''),
  106481330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481330, CurrentID=0, Status='live', Symbol='RNU6-477P', Aliases='', description='RNA, U6 small nuclear 477, pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481331, CurrentID=0, Status='live', Symbol='RNU6-479P', Aliases='', description='RNA, U6 small nuclear 479, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481332, CurrentID=0, Status='live', Symbol='RNU6-482P', Aliases='', description='RNA, U6 small nuclear 482, pseudogene', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729374, CurrentID=0, Status='live', Symbol='RPS10P6', Aliases='RPS10_1_98', description='ribosomal protein S10 pseudogene 6', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151557409, end_position_on_the_genomic_accession=151557987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481334, CurrentID=0, Status='live', Symbol='RNU6-489P', Aliases='', description='RNA, U6 small nuclear 489, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481335, CurrentID=0, Status='live', Symbol='RNU6-492P', Aliases='', description='RNA, U6 small nuclear 492, pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481336, CurrentID=0, Status='live', Symbol='RNU6-493P', Aliases='', description='RNA, U6 small nuclear 493, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481337, CurrentID=0, Status='live', Symbol='RNU6-495P', Aliases='', description='RNA, U6 small nuclear 495, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481338, CurrentID=0, Status='live', Symbol='RNU6-498P', Aliases='', description='RNA, U6 small nuclear 498, pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481339, CurrentID=0, Status='live', Symbol='RNU6-501P', Aliases='', description='RNA, U6 small nuclear 501, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481340, CurrentID=0, Status='live', Symbol='RNU6-504P', Aliases='', description='RNA, U6 small nuclear 504, pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481341, CurrentID=0, Status='live', Symbol='RNU6-507P', Aliases='', description='RNA, U6 small nuclear 507, pseudogene', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481342, CurrentID=0, Status='live', Symbol='RNU6-510P', Aliases='', description='RNA, U6 small nuclear 510, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481343, CurrentID=0, Status='live', Symbol='RNU6-513P', Aliases='', description='RNA, U6 small nuclear 513, pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481344, CurrentID=0, Status='live', Symbol='RNU6-518P', Aliases='', description='RNA, U6 small nuclear 518, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401696, CurrentID=0, Status='live', Symbol='OR4D8P', Aliases='', description='olfactory receptor family 4 subfamily D member 8 pseudogene', other_designations='olfactory receptor OR11-246 pseudogene|olfactory receptor, family 4, subfamily D, member 8 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59491653, end_position_on_the_genomic_accession=59492611, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481346, CurrentID=0, Status='live', Symbol='RNU6-524P', Aliases='', description='RNA, U6 small nuclear 524, pseudogene', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481347, CurrentID=0, Status='live', Symbol='RNU6-526P', Aliases='', description='RNA, U6 small nuclear 526, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481348, CurrentID=0, Status='live', Symbol='RNU6-529P', Aliases='', description='RNA, U6 small nuclear 529, pseudogene', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481349, CurrentID=0, Status='live', Symbol='RNU6-532P', Aliases='', description='RNA, U6 small nuclear 532, pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481350, CurrentID=0, Status='live', Symbol='RNU6-536P', Aliases='', description='RNA, U6 small nuclear 536, pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481351, CurrentID=0, Status='live', Symbol='RNU6-539P', Aliases='', description='RNA, U6 small nuclear 539, pseudogene', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481352, CurrentID=0, Status='live', Symbol='RNU6-541P', Aliases='', description='RNA, U6 small nuclear 541, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481353, CurrentID=0, Status='live', Symbol='RNU6-544P', Aliases='', description='RNA, U6 small nuclear 544, pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481354, CurrentID=0, Status='live', Symbol='RNU6-548P', Aliases='', description='RNA, U6 small nuclear 548, pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481355, CurrentID=0, Status='live', Symbol='RNU6-550P', Aliases='', description='RNA, U6 small nuclear 550, pseudogene', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481356, CurrentID=0, Status='live', Symbol='RNU6-552P', Aliases='', description='RNA, U6 small nuclear 552, pseudogene', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481357, CurrentID=0, Status='live', Symbol='RNU6-556P', Aliases='', description='RNA, U6 small nuclear 556, pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481358, CurrentID=0, Status='live', Symbol='RNU6-558P', Aliases='', description='RNA, U6 small nuclear 558, pseudogene', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481359, CurrentID=0, Status='live', Symbol='RNU6-563P', Aliases='', description='RNA, U6 small nuclear 563, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481360, CurrentID=0, Status='live', Symbol='RNU6-565P', Aliases='', description='RNA, U6 small nuclear 565, pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481361, CurrentID=0, Status='live', Symbol='RNU6-570P', Aliases='', description='RNA, U6 small nuclear 570, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481362, CurrentID=0, Status='live', Symbol='RNU6-572P', Aliases='', description='RNA, U6 small nuclear 572, pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481363, CurrentID=0, Status='live', Symbol='RNU6-575P', Aliases='', description='RNA, U6 small nuclear 575, pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481364, CurrentID=0, Status='live', Symbol='RNU6-579P', Aliases='', description='RNA, U6 small nuclear 579, pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481365, CurrentID=0, Status='live', Symbol='RNU6-582P', Aliases='', description='RNA, U6 small nuclear 582, pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481366, CurrentID=0, Status='live', Symbol='RNU6-587P', Aliases='', description='RNA, U6 small nuclear 587, pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481367, CurrentID=0, Status='live', Symbol='RNU6-589P', Aliases='', description='RNA, U6 small nuclear 589, pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481368, CurrentID=0, Status='live', Symbol='RNU6-592P', Aliases='', description='RNA, U6 small nuclear 592, pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481369, CurrentID=0, Status='live', Symbol='RNU6-596P', Aliases='', description='RNA, U6 small nuclear 596, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481370, CurrentID=0, Status='live', Symbol='RNU6-599P', Aliases='', description='RNA, U6 small nuclear 599, pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481371, CurrentID=0, Status='live', Symbol='RNU6-602P', Aliases='', description='RNA, U6 small nuclear 602, pseudogene', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481372, CurrentID=0, Status='live', Symbol='RNU6-605P', Aliases='', description='RNA, U6 small nuclear 605, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481373, CurrentID=0, Status='live', Symbol='RNU6-608P', Aliases='', description='RNA, U6 small nuclear 608, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481374, CurrentID=0, Status='live', Symbol='RNU6-612P', Aliases='', description='RNA, U6 small nuclear 612, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481375, CurrentID=0, Status='live', Symbol='RNU6-614P', Aliases='', description='RNA, U6 small nuclear 614, pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481376, CurrentID=0, Status='live', Symbol='RNU6-618P', Aliases='', description='RNA, U6 small nuclear 618, pseudogene', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481377, CurrentID=0, Status='live', Symbol='RNU6-620P', Aliases='', description='RNA, U6 small nuclear 620, pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481378, CurrentID=0, Status='live', Symbol='RNU6-623P', Aliases='', description='RNA, U6 small nuclear 623, pseudogene', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481379, CurrentID=0, Status='live', Symbol='RNU6-626P', Aliases='', description='RNA, U6 small nuclear 626, pseudogene', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481380, CurrentID=0, Status='live', Symbol='RNU6-630P', Aliases='', description='RNA, U6 small nuclear 630, pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481381, CurrentID=0, Status='live', Symbol='RNU6-633P', Aliases='', description='RNA, U6 small nuclear 633, pseudogene', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481382, CurrentID=0, Status='live', Symbol='RNU6-635P', Aliases='', description='RNA, U6 small nuclear 635, pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481383, CurrentID=0, Status='live', Symbol='RNU6-638P', Aliases='', description='RNA, U6 small nuclear 638, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481384, CurrentID=0, Status='live', Symbol='RNU6-642P', Aliases='', description='RNA, U6 small nuclear 642, pseudogene', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481385, CurrentID=0, Status='live', Symbol='RNU6-644P', Aliases='', description='RNA, U6 small nuclear 644, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481386, CurrentID=0, Status='live', Symbol='RNU6-646P', Aliases='', description='RNA, U6 small nuclear 646, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481387, CurrentID=0, Status='live', Symbol='RNU6-650P', Aliases='', description='RNA, U6 small nuclear 650, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481388, CurrentID=0, Status='live', Symbol='RNU6-652P', Aliases='', description='RNA, U6 small nuclear 652, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481389, CurrentID=0, Status='live', Symbol='RNU6-657P', Aliases='', description='RNA, U6 small nuclear 657, pseudogene', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481390, CurrentID=0, Status='live', Symbol='RNU6-659P', Aliases='', description='RNA, U6 small nuclear 659, pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481391, CurrentID=0, Status='live', Symbol='RNU6-663P', Aliases='', description='RNA, U6 small nuclear 663, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481392, CurrentID=0, Status='live', Symbol='RNU6-665P', Aliases='', description='RNA, U6 small nuclear 665, pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481393, CurrentID=0, Status='live', Symbol='RNU6-668P', Aliases='', description='RNA, U6 small nuclear 668, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481394, CurrentID=0, Status='live', Symbol='RNU6-673P', Aliases='', description='RNA, U6 small nuclear 673, pseudogene', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481395, CurrentID=0, Status='live', Symbol='RNU6-677P', Aliases='', description='RNA, U6 small nuclear 677, pseudogene', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481396, CurrentID=0, Status='live', Symbol='RNU6-684P', Aliases='', description='RNA, U6 small nuclear 684, pseudogene', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313821, CurrentID=0, Status='live', Symbol='PIRC97', Aliases='', description='piwi-interacting RNA cluster 97', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481398, CurrentID=0, Status='live', Symbol='RNU6-690P', Aliases='', description='RNA, U6 small nuclear 690, pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481399, CurrentID=0, Status='live', Symbol='RNU6-693P', Aliases='', description='RNA, U6 small nuclear 693, pseudogene', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481400, CurrentID=0, Status='live', Symbol='RNU6-695P', Aliases='', description='RNA, U6 small nuclear 695, pseudogene', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50937, CurrentID=0, Status='live', Symbol='CDON', Aliases='CDO1, HPE11, ORCAM, CDON', description='cell adhesion associated, oncogene regulated', other_designations='Cdon homolog|surface glycoprotein, Ig superfamily member', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125956818, end_position_on_the_genomic_accession=126063292, orientation='minus', exon_count=23, OMIM=608707, no_hdr0=''),
  106481402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481402, CurrentID=0, Status='live', Symbol='RNU6-701P', Aliases='', description='RNA, U6 small nuclear 701, pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50939, CurrentID=0, Status='live', Symbol='IMPG2', Aliases='IPM200, RP56, SPACRCAN, VMD5', description='interphotoreceptor matrix proteoglycan 2', other_designations='IPM 200|interphotoreceptor matrix proteoglycan IPM 200|interphotoreceptor matrix proteoglycan of 200 kDa|sialoprotein associated with cones and rods proteoglycan', map_location='3q12.2-q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101222546, end_position_on_the_genomic_accession=101320575, orientation='minus', exon_count=19, OMIM=607056, no_hdr0=''),
  50940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50940, CurrentID=0, Status='live', Symbol='PDE11A', Aliases='PPNAD2', description='phosphodiesterase 11A', other_designations='cAMP and cGMP cyclic nucleotide phosphodiesterase 11A', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177623249, end_position_on_the_genomic_accession=178108339, orientation='minus', exon_count=24, OMIM=604961, no_hdr0=''),
  50941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50941, CurrentID=0, Status='live', Symbol='IBD5', Aliases='', description='inflammatory bowel disease 5', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606348, no_hdr0=''),
  50942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50942, CurrentID=0, Status='live', Symbol='IBD6', Aliases='', description='inflammatory bowel disease 6', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606674, no_hdr0=''),
  50943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50943, CurrentID=0, Status='live', Symbol='FOXP3', Aliases='AIID, DIETER, IPEX, JM2, PIDX, XPID', description='forkhead box P3', other_designations='FOXP3delta7|immune dysregulation, polyendocrinopathy, enteropathy, X-linked|immunodeficiency, polyendocrinopathy, enteropathy, X-linked|scurfin', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49250436, end_position_on_the_genomic_accession=49269727, orientation='minus', exon_count=14, OMIM=300292, no_hdr0=''),
  50944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50944, CurrentID=0, Status='live', Symbol='SHANK1', Aliases='SPANK-1, SSTRIP, synamon', description='SH3 and multiple ankyrin repeat domains 1', other_designations='SSTR-interacting protein|somatostatin receptor-interacting protein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50659264, end_position_on_the_genomic_accession=50719783, orientation='minus', exon_count=25, OMIM=604999, no_hdr0=''),
  50945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50945, CurrentID=0, Status='live', Symbol='TBX22', Aliases='ABERS, CLPA, CPX, TBXX, dJ795G23.1', description='T-box 22', other_designations='T-box protein 22', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80014728, end_position_on_the_genomic_accession=80035828, orientation='plus', exon_count=11, OMIM=300307, no_hdr0=''),
  106481410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481410, CurrentID=0, Status='live', Symbol='RNU6-729P', Aliases='', description='RNA, U6 small nuclear 729, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50947, CurrentID=0, Status='live', Symbol='BHD', Aliases='BFHD', description='Beukes familial hip dysplasia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=142669, no_hdr0=''),
  50948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50948, CurrentID=0, Status='live', Symbol='OHDS', Aliases='', description='Orthostatic hypotensive disorder of Streeten', other_designations='', map_location='18q', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=143850, no_hdr0=''),
  106481413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481413, CurrentID=0, Status='live', Symbol='RNU6-739P', Aliases='', description='RNA, U6 small nuclear 739, pseudogene', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481414, CurrentID=0, Status='live', Symbol='RNU6-742P', Aliases='', description='RNA, U6 small nuclear 742, pseudogene', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50951, CurrentID=0, Status='live', Symbol='MMVP1', Aliases='MVP, PMV', description='myxomatous mitral valve prolapse 1', other_designations='Mitral valve prolapse, familial', map_location='16p12.1-p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=157700, no_hdr0=''),
  50952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50952, CurrentID=0, Status='live', Symbol='HMN7A', Aliases='DHMNVP', description='Motor neuronopathy, distal hereditary, type VIIA', other_designations='Motor neuronopathy, distal hereditary, with vocal cord paresis', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=158580, no_hdr0=''),
  401708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401708, CurrentID=0, Status='live', Symbol='GLULP3', Aliases='GLULL3', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 3', other_designations='glutamate-ammonia ligase (glutamine synthase)-like 3|glutamate-ammonia ligase (glutamine synthetase)-like 3', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122613326, end_position_on_the_genomic_accession=122614690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481418, CurrentID=0, Status='live', Symbol='RNU6-754P', Aliases='', description='RNA, U6 small nuclear 754, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481419, CurrentID=0, Status='live', Symbol='RNU6-758P', Aliases='', description='RNA, U6 small nuclear 758, pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481420, CurrentID=0, Status='live', Symbol='RNU6-760P', Aliases='', description='RNA, U6 small nuclear 760, pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481421, CurrentID=0, Status='live', Symbol='RNU6-763P', Aliases='', description='RNA, U6 small nuclear 763, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481422, CurrentID=0, Status='live', Symbol='RNU6-766P', Aliases='', description='RNA, U6 small nuclear 766, pseudogene', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481423, CurrentID=0, Status='live', Symbol='RNU6-770P', Aliases='', description='RNA, U6 small nuclear 770, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50960, CurrentID=0, Status='live', Symbol='MYMY1', Aliases='MYMY', description='moyamoya disease 1', other_designations='', map_location='3p26-p24.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=252350, no_hdr0=''),
  106481425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481425, CurrentID=0, Status='live', Symbol='RNU6-774P', Aliases='', description='RNA, U6 small nuclear 774, pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481426, CurrentID=0, Status='live', Symbol='RNU6-777P', Aliases='', description='RNA, U6 small nuclear 777, pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50963, CurrentID=0, Status='live', Symbol='BRKS1', Aliases='BRKS, TLH1', description='Telopeptide lysyl hydroxylase, bone-specific', other_designations='osteognesis imperfecta with congenital joint contractures', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=259450, no_hdr0=''),
  50964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50964, CurrentID=0, Status='live', Symbol='SOST', Aliases='CDD1, VBCH, SOST', description='sclerostin', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43753731, end_position_on_the_genomic_accession=43758788, orientation='minus', exon_count=2, OMIM=605740, no_hdr0=''),
  106481429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481429, CurrentID=0, Status='live', Symbol='RNU6-786P', Aliases='', description='RNA, U6 small nuclear 786, pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50966, CurrentID=0, Status='live', Symbol='ECA1', Aliases='', description='epilepsy, childhood absence 1', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600131, no_hdr0=''),
  106481431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481431, CurrentID=0, Status='live', Symbol='RNU6-792P', Aliases='', description='RNA, U6 small nuclear 792, pseudogene', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50968, CurrentID=0, Status='live', Symbol='FCMTE1', Aliases='BAFME, BAFME1, FAME, FAME1, MEBA', description='familial cortical myoclonic tremor with epilepsy 1', other_designations='benign adult familial myoclonic epilepsy 1|familial adult myoclonic epilepsy 1|myoclonic epilepsy, benign adult familial', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601068, no_hdr0=''),
  106481433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481433, CurrentID=0, Status='live', Symbol='RNU6-800P', Aliases='', description='RNA, U6 small nuclear 800, pseudogene', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50970, CurrentID=0, Status='live', Symbol='TSG11', Aliases='', description='Tumor suppressor gene on chromosome 11', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603040, no_hdr0=''),
  50971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50971, CurrentID=0, Status='live', Symbol='ENFL2', Aliases='', description='Epilepsy, nocturnal frontal lobe, type 2', other_designations='', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603204, no_hdr0=''),
  106481436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481436, CurrentID=0, Status='live', Symbol='RNU6-808P', Aliases='', description='RNA, U6 small nuclear 808, pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481437, CurrentID=0, Status='live', Symbol='RNU6-810P', Aliases='', description='RNA, U6 small nuclear 810, pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481438, CurrentID=0, Status='live', Symbol='RNU6-813P', Aliases='', description='RNA, U6 small nuclear 813, pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50975, CurrentID=0, Status='live', Symbol='TCO', Aliases='', description='Thyroid carcinoma, nonmedullary, with cell oxyphilia', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603386, no_hdr0=''),
  50976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50976, CurrentID=0, Status='live', Symbol='GRD2', Aliases='', description='Graves disease, susceptibility to, 2', other_designations='', map_location='20q13.11', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603388, no_hdr0=''),
  106481441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481441, CurrentID=0, Status='live', Symbol='RNU6-823P', Aliases='', description='RNA, U6 small nuclear 823, pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481442, CurrentID=0, Status='live', Symbol='RNU6-827P', Aliases='', description='RNA, U6 small nuclear 827, pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50979, CurrentID=0, Status='live', Symbol='MHW2', Aliases='', description='Mental health wellness 2', other_designations='', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603664, no_hdr0=''),
  106481444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481444, CurrentID=0, Status='live', Symbol='RNU6-833P', Aliases='', description='RNA, U6 small nuclear 833, pseudogene', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481445, CurrentID=0, Status='live', Symbol='RNU6-835P', Aliases='', description='RNA, U6 small nuclear 835, pseudogene', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50982, CurrentID=0, Status='live', Symbol='NIDDM3', Aliases='', description='Noninsulin-dependent diabetes mellitus 3', other_designations='', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603694, no_hdr0=''),
  106481447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481447, CurrentID=0, Status='live', Symbol='RNU6-842P', Aliases='', description='RNA, U6 small nuclear 842, pseudogene', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481448, CurrentID=0, Status='live', Symbol='RNU6-847P', Aliases='', description='RNA, U6 small nuclear 847, pseudogene', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481449, CurrentID=0, Status='live', Symbol='RNU6-849P', Aliases='', description='RNA, U6 small nuclear 849, pseudogene', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50986, CurrentID=0, Status='live', Symbol='HYT2', Aliases='', description='Hypertension, essential, susceptibility to, 2', other_designations='', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604329, no_hdr0=''),
  50987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50987, CurrentID=0, Status='live', Symbol='FPEVF', Aliases='', description='Epilepsy, partial, with variable foci', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604364, no_hdr0=''),
  50988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50988, CurrentID=0, Status='live', Symbol='WDM', Aliases='', description='Welander distal myopathy', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604454, no_hdr0=''),
  50989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50989, CurrentID=0, Status='live', Symbol='HMSNO', Aliases='HMSNP', description='Neuropathy, hereditary motor and sensory, Okinawa type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604484, no_hdr0=''),
  106481454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481454, CurrentID=0, Status='live', Symbol='RNU6-866P', Aliases='', description='RNA, U6 small nuclear 866, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481455, CurrentID=0, Status='live', Symbol='RNU6-876P', Aliases='', description='RNA, U6 small nuclear 876, pseudogene', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481456, CurrentID=0, Status='live', Symbol='RNU6-879P', Aliases='', description='RNA, U6 small nuclear 879, pseudogene', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481457, CurrentID=0, Status='live', Symbol='RNU6-882P', Aliases='', description='RNA, U6 small nuclear 882, pseudogene', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481458, CurrentID=0, Status='live', Symbol='RNU6-885P', Aliases='', description='RNA, U6 small nuclear 885, pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481459, CurrentID=0, Status='live', Symbol='RNU6-887P', Aliases='', description='RNA, U6 small nuclear 887, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481460, CurrentID=0, Status='live', Symbol='RNU6-892P', Aliases='', description='RNA, U6 small nuclear 892, pseudogene', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481461, CurrentID=0, Status='live', Symbol='RNU6-894P', Aliases='', description='RNA, U6 small nuclear 894, pseudogene', other_designations='', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481462, CurrentID=0, Status='live', Symbol='RNU6-898P', Aliases='', description='RNA, U6 small nuclear 898, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=50999, CurrentID=0, Status='live', Symbol='TMED5', Aliases='CGI-100, p28', description='transmembrane p24 trafficking protein 5', other_designations='p24 family protein gamma-2|p24gamma2|transmembrane emp24 protein transport domain containing 5', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93149742, end_position_on_the_genomic_accession=93180689, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  51000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51000, CurrentID=0, Status='live', Symbol='SLC35B3', Aliases='C6orf196, CGI-19, PAPST2', description="solute carrier family 35 (adenosine 3'-phospho 5'-phosphosulfate transporter), member B3", other_designations="3' phosphoadenosine 5' phosphosulfate transporter 2|PAPS transporter 2|solute carrier family 35, member B3", map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8413067, end_position_on_the_genomic_accession=8435567, orientation='minus', exon_count=13, OMIM=610845, no_hdr0=''),
  51001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51001, CurrentID=0, Status='live', Symbol='MTERF3', Aliases='CGI-12, MTERFD1', description='mitochondrial transcription termination factor 3', other_designations='MTERF domain containing 1|mTERF domain-containing protein 1, mitochondrial', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96239398, end_position_on_the_genomic_accession=96261613, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  51002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51002, CurrentID=0, Status='live', Symbol='TPRKB', Aliases='CGI-121', description='TP53RK binding protein', other_designations='PRPK (p53-related protein kinase)-binding protein|PRPK-binding protein|TP53RK-binding protein', map_location='2p24.3-p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73729830, end_position_on_the_genomic_accession=73740104, orientation='minus', exon_count=9, OMIM=608680, no_hdr0=''),
  51003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51003, CurrentID=0, Status='live', Symbol='MED31', Aliases='3110004H13Rik, CGI-125, Soh1', description='mediator complex subunit 31', other_designations='hSOH1|mediator complex subunit SOH1|mediator of RNA polymerase II transcription, subunit 31 homolog', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6643313, end_position_on_the_genomic_accession=6651634, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  51004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51004, CurrentID=0, Status='live', Symbol='COQ6', Aliases='CGI-10, CGI10, COQ10D6', description='coenzyme Q6 monooxygenase', other_designations='coenzyme Q10 monooxygenase 6|coenzyme Q6 homolog, monooxygenase', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73949934, end_position_on_the_genomic_accession=73963112, orientation='plus', exon_count=15, OMIM=614647, no_hdr0=''),
  51005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51005, CurrentID=0, Status='live', Symbol='AMDHD2', Aliases='CGI-14', description='amidohydrolase domain containing 2', other_designations='amidohydrolase domain-containing protein 2|glcNAc 6-P deacetylase', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2520362, end_position_on_the_genomic_accession=2530954, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  51006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51006, CurrentID=0, Status='live', Symbol='SLC35C2', Aliases='BA394O2.1, C20orf5, CGI-15, OVCOV1', description='solute carrier family 35 (GDP-fucose transporter), member C2', other_designations='ovarian cancer-overexpressed gene 1 protein', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46349528, end_position_on_the_genomic_accession=46364458, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106481471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481471, CurrentID=0, Status='live', Symbol='RNU6-923P', Aliases='', description='RNA, U6 small nuclear 923, pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51008, CurrentID=0, Status='live', Symbol='ASCC1', Aliases='ASC1p50, CGI-18, p50', description='activating signal cointegrator 1 complex subunit 1', other_designations='ASC-1 complex subunit P50|trip4 complex subunit p50', map_location='10pter-q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72096032, end_position_on_the_genomic_accession=72217134, orientation='minus', exon_count=15, OMIM=614215, no_hdr0=''),
  51009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51009, CurrentID=0, Status='live', Symbol='DERL2', Aliases='CGI-101, DERtrin-2, F-LAN-1, F-LANa, FLANa', description='derlin 2', other_designations='Der1-like domain family, member 2|carcinoma related|degradation in endoplasmic reticulum protein 2', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5471251, end_position_on_the_genomic_accession=5486512, orientation='minus', exon_count=10, OMIM=610304, no_hdr0=''),
  51010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51010, CurrentID=0, Status='live', Symbol='EXOSC3', Aliases='CGI-102, PCH1B, RRP40, Rrp40p, bA3J10.7, hRrp-40, p10', description='exosome component 3', other_designations='exosome complex exonuclease RRP40|ribosomal RNA-processing protein 40', map_location='9p11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37779714, end_position_on_the_genomic_accession=37785092, orientation='minus', exon_count=4, OMIM=606489, no_hdr0=''),
  51011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51011, CurrentID=0, Status='live', Symbol='FAHD2A', Aliases='CGI-105', description='fumarylacetoacetate hydrolase domain containing 2A', other_designations='fumarylacetoacetate hydrolase domain containing 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95402691, end_position_on_the_genomic_accession=95416609, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  51012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51012, CurrentID=0, Status='live', Symbol='PRELID3B', Aliases='C20orf45, SLMO2, dJ543J19.5', description='PRELI domain containing 3B', other_designations='protein slowmo homolog 2|slowmo homolog 2', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59033145, end_position_on_the_genomic_accession=59042846, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  51013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51013, CurrentID=0, Status='live', Symbol='EXOSC1', Aliases='CGI-108, CSL4, Csl4p, SKI4, Ski4p, hCsl4p, p13', description='exosome component 1', other_designations="3'-5' exoribonuclease CSL4 homolog|CSL4 exosomal core protein homolog|exosomal core protein CSL4|homolog of yeast exosomal core protein CSL4", map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97435909, end_position_on_the_genomic_accession=97446011, orientation='minus', exon_count=10, OMIM=606493, no_hdr0=''),
  51014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51014, CurrentID=0, Status='live', Symbol='TMED7', Aliases='CGI-109, p24gamma3, p27', description='transmembrane p24 trafficking protein 7', other_designations='p24 family protein gamma-3|transmembrane emp24 protein transport domain containing 7', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115613208, end_position_on_the_genomic_accession=115626179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  51015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51015, CurrentID=0, Status='live', Symbol='ISOC1', Aliases='CGI-111', description='isochorismatase domain containing 1', other_designations='', map_location='5q22.1-q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129094749, end_position_on_the_genomic_accession=129114028, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51016, CurrentID=0, Status='live', Symbol='EMC9', Aliases='C14orf122, CGI-112, FAM158A', description='ER membrane protein complex subunit 9', other_designations='UPF0172 protein FAM158A|family with sequence similarity 158, member A', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24138965, end_position_on_the_genomic_accession=24141588, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106481481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481481, CurrentID=0, Status='live', Symbol='RNU6-954P', Aliases='', description='RNA, U6 small nuclear 954, pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51018, CurrentID=0, Status='live', Symbol='RRP15', Aliases='CGI-115, KIAA0507', description='ribosomal RNA processing 15 homolog', other_designations='ribosomal RNA-processing protein 15', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218285285, end_position_on_the_genomic_accession=218337983, orientation='plus', exon_count=6, OMIM=611193, no_hdr0=''),
  51019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51019, CurrentID=0, Status='live', Symbol='CCDC53', Aliases='CGI-116', description='coiled-coil domain containing 53', other_designations='coiled-coil domain-containing protein 53', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102012840, end_position_on_the_genomic_accession=102062124, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  51020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51020, CurrentID=0, Status='live', Symbol='HDDC2', Aliases='C6orf74, CGI-130, NS5ATP2, dJ167O5.2', description='HD domain containing 2', other_designations='HCV NS5A-transactivated protein 2|hepatitis C virus NS5A-transactivated protein 2', map_location='6q13-q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125275350, end_position_on_the_genomic_accession=125302141, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  51021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51021, CurrentID=0, Status='live', Symbol='MRPS16', Aliases='CGI-132, COXPD2, MRP-S16, RPMS16', description='mitochondrial ribosomal protein S16', other_designations='S16mt', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73248843, end_position_on_the_genomic_accession=73252693, orientation='minus', exon_count=3, OMIM=609204, no_hdr0=''),
  51022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51022, CurrentID=0, Status='live', Symbol='GLRX2', Aliases='CGI-133, GRX2', description='glutaredoxin 2', other_designations='bA101E13.1 (GRX2 glutaredoxin (thioltransferase) 2)', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193096465, end_position_on_the_genomic_accession=193106549, orientation='minus', exon_count=6, OMIM=606820, no_hdr0=''),
  51023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51023, CurrentID=0, Status='live', Symbol='MRPS18C', Aliases='CGI-134, MRP-S18-1, MRP-S18-c, MRPS18-1, S18mt-c, mrps18-c', description='mitochondrial ribosomal protein S18C', other_designations='28S ribosomal protein S18-1, mitochondrial|mitochondrial ribosomal protein S18-1', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83455932, end_position_on_the_genomic_accession=83461723, orientation='plus', exon_count=6, OMIM=611983, no_hdr0=''),
  51024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51024, CurrentID=0, Status='live', Symbol='FIS1', Aliases='CGI-135, TTC11', description='fission, mitochondrial 1', other_designations='FIS1 homolog|H_NH0132A01.6|TPR repeat protein 11|fission 1 (mitochondrial outer membrane) homolog|hFis1|mitochondrial fission molecule|tetratricopeptide repeat domain 11|tetratricopeptide repeat protein 11', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101239612, end_position_on_the_genomic_accession=101245090, orientation='minus', exon_count=5, OMIM=609003, no_hdr0=''),
  51025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51025, CurrentID=0, Status='live', Symbol='PAM16', Aliases='CGI-136, MAGMAS, SMDMDM, TIM16, TIMM16', description='presequence translocase-associated motor 16 homolog (S. cerevisiae)', other_designations='magmas-like protein|mitochondria associated protein involved in granulocyte macrophage colony stimulating factor signal transduction|mitochondria-associated granulocyte macrophage CSF-signaling molecule', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4340251, end_position_on_the_genomic_accession=4351372, orientation='minus', exon_count=5, OMIM=614336, no_hdr0=''),
  51026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51026, CurrentID=0, Status='live', Symbol='GOLT1B', Aliases='CGI-141, GCT2, GOT1, GOT1B, YMR292W', description='golgi transport 1B', other_designations='germ cell tumor 2|hGOT1a|putative NF-kappa-B-activating protein 470', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21501765, end_position_on_the_genomic_accession=21518403, orientation='plus', exon_count=5, OMIM=615078, no_hdr0=''),
  51027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51027, CurrentID=0, Status='live', Symbol='BOLA1', Aliases='CGI-143', description='bolA family member 1', other_designations='bolA homolog 1|bolA-like 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149899602, end_position_on_the_genomic_accession=149900795, orientation='plus', exon_count=2, OMIM=613181, no_hdr0=''),
  51028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51028, CurrentID=0, Status='live', Symbol='VPS36', Aliases='C13orf9, CGI-145, EAP45', description='vacuolar protein sorting 36 homolog (S. cerevisiae)', other_designations='ELL-associated protein of 45 kDa|ESCRT-II complex subunit VPS36', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52412602, end_position_on_the_genomic_accession=52450678, orientation='minus', exon_count=15, OMIM=610903, no_hdr0=''),
  51029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51029, CurrentID=0, Status='live', Symbol='DESI2', Aliases='C1orf121, CGI-146, DESI, DESI1, DeSI-2, FAM152A, PNAS-4, PPPDE1', description='desumoylating isopeptidase 2', other_designations='PPPDE peptidase domain-containing protein 1|desumoylating isopeptidase 1|family with sequence similarity 152, member A', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244653048, end_position_on_the_genomic_accession=244709034, orientation='plus', exon_count=10, OMIM=614638, no_hdr0=''),
  51030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51030, CurrentID=0, Status='live', Symbol='TVP23B', Aliases='CGI-148, FAM18B, FAM18B1, NPD008, YDR084C', description='trans-golgi network vesicle protein 23 homolog B (S. cerevisiae)', other_designations='family with sequence similarity 18, member B|family with sequence similarity 18, member B1|protein FAM18B1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18781145, end_position_on_the_genomic_accession=18806714, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51031, CurrentID=0, Status='live', Symbol='GLOD4', Aliases='C17orf25, CGI-150, HC71', description='glyoxalase domain containing 4', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=759309, end_position_on_the_genomic_accession=782331, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  51032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51032, CurrentID=0, Status='live', Symbol='CELA2B', Aliases='ELA2B', description='chymotrypsin like elastase family member 2B', other_designations='RP11-265F14.2|chymotrypsin-like elastase family, member 2B|elastase 2B|elastase-2B|pancreatic elastase IIB (ELA2B)', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15476101, end_position_on_the_genomic_accession=15491400, orientation='plus', exon_count=8, OMIM=609444, no_hdr0=''),
  106481497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481497, CurrentID=0, Status='live', Symbol='RNU6-1005P', Aliases='', description='RNA, U6 small nuclear 1005, pseudogene', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481498, CurrentID=0, Status='live', Symbol='RNU6-1008P', Aliases='', description='RNA, U6 small nuclear 1008, pseudogene', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51035, CurrentID=0, Status='live', Symbol='UBXN1', Aliases='2B28, SAKS1, UBXD10', description='UBX domain protein 1', other_designations='SAPK substrate protein 1|UBA/UBX 33.3 kDa protein', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62676498, end_position_on_the_genomic_accession=62679117, orientation='minus', exon_count=9, OMIM=616378, no_hdr0=''),
  106481500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481500, CurrentID=0, Status='live', Symbol='RNU6-1012P', Aliases='', description='RNA, U6 small nuclear 1012, pseudogene', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481501, CurrentID=0, Status='live', Symbol='RNU6-1015P', Aliases='', description='RNA, U6 small nuclear 1015, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481502, CurrentID=0, Status='live', Symbol='RNU6-1018P', Aliases='', description='RNA, U6 small nuclear 1018, pseudogene', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481503, CurrentID=0, Status='live', Symbol='RNU6-1020P', Aliases='', description='RNA, U6 small nuclear 1020, pseudogene', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481504, CurrentID=0, Status='live', Symbol='RNU6-1024P', Aliases='', description='RNA, U6 small nuclear 1024, pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481505, CurrentID=0, Status='live', Symbol='RNU6-1027P', Aliases='', description='RNA, U6 small nuclear 1027, pseudogene', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51042, CurrentID=0, Status='live', Symbol='ZNF593', Aliases='ZT86', description='zinc finger protein 593', other_designations='zinc finger protein T86', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26169897, end_position_on_the_genomic_accession=26170873, orientation='plus', exon_count=3, OMIM=616698, no_hdr0=''),
  51043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51043, CurrentID=0, Status='live', Symbol='ZBTB7B', Aliases='CKROX, THPOK, ZBTB15, ZFP-67, ZFP67, ZNF857B, c-KROX, hcKROX', description='zinc finger and BTB domain containing 7B', other_designations='T-helper-inducing POZ/Krueppel-like factor|krueppel-related zinc finger protein cKrox|zinc finger and BTB domain containing 15|zinc finger protein 67 homolog|zinc finger protein 857B|zinc finger protein Th-POK', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155002627, end_position_on_the_genomic_accession=155018525, orientation='plus', exon_count=7, OMIM=607646, no_hdr0=''),
  106481508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481508, CurrentID=0, Status='live', Symbol='RNU6-1037P', Aliases='', description='RNA, U6 small nuclear 1037, pseudogene', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481509, CurrentID=0, Status='live', Symbol='RNU6-1041P', Aliases='', description='RNA, U6 small nuclear 1041, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51046, CurrentID=0, Status='live', Symbol='ST8SIA3', Aliases='SIAT8C, ST8SiaIII', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 3', other_designations='SIAT8-C|alpha-2,8-sialyltransferase 8C|alpha-2,8-sialyltransferase III|sialyltransferase 8C (alpha2,3Galbeta1,4GlcNAcalpha 2,8-sialyltransferase)|sialyltransferase St8Sia III|sialytransferase St8Sia III', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57352490, end_position_on_the_genomic_accession=57371800, orientation='plus', exon_count=5, OMIM=609478, no_hdr0=''),
  106481511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481511, CurrentID=0, Status='live', Symbol='RNU6-1047P', Aliases='', description='RNA, U6 small nuclear 1047, pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481512, CurrentID=0, Status='live', Symbol='RNU6-1050P', Aliases='', description='RNA, U6 small nuclear 1050, pseudogene', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481513, CurrentID=0, Status='live', Symbol='RNU6-1053P', Aliases='', description='RNA, U6 small nuclear 1053, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51050, CurrentID=0, Status='live', Symbol='PI15', Aliases='CRISP8, P24TI, P25TI', description='peptidase inhibitor 15', other_designations='25 kDa trypsin inhibitor|CRISP-8|PI-15|cysteine-rich secretory protein 8|protease inhibitor 15|sugarCrisp', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74824469, end_position_on_the_genomic_accession=74855044, orientation='plus', exon_count=7, OMIM=607076, no_hdr0=''),
  106481515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481515, CurrentID=0, Status='live', Symbol='RNU6-1064P', Aliases='', description='RNA, U6 small nuclear 1064, pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51052, CurrentID=0, Status='live', Symbol='PRLH', Aliases='PRH, PRRP', description='prolactin releasing hormone', other_designations='preproprolactin-releasing peptide|prolactin-releasing hormone', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237566574, end_position_on_the_genomic_accession=237567175, orientation='plus', exon_count=2, OMIM=602663, no_hdr0=''),
  51053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51053, CurrentID=0, Status='live', Symbol='GMNN', Aliases='Gem', description='geminin, DNA replication inhibitor', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24774931, end_position_on_the_genomic_accession=24786099, orientation='plus', exon_count=9, OMIM=602842, no_hdr0=''),
  51054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51054, CurrentID=0, Status='live', Symbol='PLEKHA8P1', Aliases='PLEKHA9', description='pleckstrin homology domain containing A8 pseudogene 1', other_designations='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 9|pleckstrin homology domain containing, family A member 8 pseudogene 1|putative glycolipid transfer protein', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45173034, end_position_on_the_genomic_accession=45216006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401725, CurrentID=0, Status='live', Symbol='RPL6P25', Aliases='RPL6_15_1263', description='ribosomal protein L6 pseudogene 25', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=83151311, end_position_on_the_genomic_accession=83152232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  51056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51056, CurrentID=0, Status='live', Symbol='LAP3', Aliases='HEL-S-106, LAP, LAPEP, PEPS', description='leucine aminopeptidase 3', other_designations='LAP-3|epididymis secretory protein Li 106|leucyl aminopeptidase|peptidase S|proline aminopeptidase|prolyl aminopeptidase', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17577304, end_position_on_the_genomic_accession=17607967, orientation='plus', exon_count=13, OMIM=170250, no_hdr0=''),
  51057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51057, CurrentID=0, Status='live', Symbol='WDPCP', Aliases='BBS15, C2orf86, CHDTHP, FRITZ, FRTZ', description='WD repeat containing planar cell polarity effector', other_designations='Bardet-Biedl syndrome 15 protein|WD repeat-containing protein C2orf86', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63120762, end_position_on_the_genomic_accession=63840825, orientation='minus', exon_count=31, OMIM=613580, no_hdr0=''),
  51058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51058, CurrentID=0, Status='live', Symbol='ZNF691', Aliases='Zfp691', description='zinc finger protein 691', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42846573, end_position_on_the_genomic_accession=42852477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51059, CurrentID=0, Status='live', Symbol='FAM135B', Aliases='C8ORFK32', description='family with sequence similarity 135 member B', other_designations='family with sequence similarity 135, member B', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=138130023, end_position_on_the_genomic_accession=138496822, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  51060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51060, CurrentID=0, Status='live', Symbol='TXNDC12', Aliases='AG1, AGR1, ERP16, ERP18, ERP19, PDIA16, TLP19, hAG-1, hTLP19', description='thioredoxin domain containing 12 (endoplasmic reticulum)', other_designations='ER protein 18|ER protein 19|anterior gradient homolog 1|endoplasmic reticulum protein ERp19|endoplasmic reticulum resident protein 18|endoplasmic reticulum resident protein 19|endoplasmic reticulum thioredoxin superfamily member, 18 kDa|protein disulfide isomerase family A, member 16|thioredoxin-like protein p19', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52020131, end_position_on_the_genomic_accession=52056171, orientation='minus', exon_count=8, OMIM=609448, no_hdr0=''),
  51061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51061, CurrentID=0, Status='live', Symbol='TXNDC11', Aliases='EFP1', description='thioredoxin domain containing 11', other_designations='EF-hand binding protein 1|EF-hand-binding protein 1', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11679080, end_position_on_the_genomic_accession=11742878, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  51062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51062, CurrentID=0, Status='live', Symbol='ATL1', Aliases='AD-FSP, FSP1, GBP3, HSN1D, SPG3, SPG3A, atlastin1', description='atlastin GTPase 1', other_designations='GBP-3|GTP-binding protein 3|brain-specific GTP-binding protein|guanine nucleotide-binding protein 3|guanylate-binding protein 3|hGBP3|spastic paraplegia 3 protein A', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50533082, end_position_on_the_genomic_accession=50633068, orientation='plus', exon_count=15, OMIM=606439, no_hdr0=''),
  51063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51063, CurrentID=0, Status='live', Symbol='CALHM2', Aliases='FAM26B', description='calcium homeostasis modulator 2', other_designations='family with sequence similarity 26, member B', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103446785, end_position_on_the_genomic_accession=103452405, orientation='minus', exon_count=4, OMIM=612235, no_hdr0=''),
  106481528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481528, CurrentID=0, Status='live', Symbol='RNU6-1104P', Aliases='', description='RNA, U6 small nuclear 1104, pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51065, CurrentID=0, Status='live', Symbol='RPS27L', Aliases='', description='ribosomal protein S27 like', other_designations='ribosomal protein S27-like', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63153340, end_position_on_the_genomic_accession=63157542, orientation='minus', exon_count=4, OMIM=612055, no_hdr0=''),
  51066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51066, CurrentID=0, Status='live', Symbol='SSUH2', Aliases='C3orf32, SSU-2, fls485', description='ssu-2 homolog (C. elegans)', other_designations='protein ssu-2 homolog', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8590122, end_position_on_the_genomic_accession=8745035, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  51067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51067, CurrentID=0, Status='live', Symbol='YARS2', Aliases='CGI-04, MLASA2, MT-TYRRS, TYRRS', description='tyrosyl-tRNA synthetase 2', other_designations='tyrosine tRNA ligase 2, mitochondrial|tyrosyl-tRNA synthetase 2, mitochondrial', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32725091, end_position_on_the_genomic_accession=32755961, orientation='minus', exon_count=9, OMIM=610957, no_hdr0=''),
  51068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51068, CurrentID=0, Status='live', Symbol='NMD3', Aliases='CGI-07', description='NMD3 ribosome export adaptor', other_designations='NMD3 homolog|hNMD3', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161221262, end_position_on_the_genomic_accession=161252294, orientation='plus', exon_count=16, OMIM=611021, no_hdr0=''),
  51069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51069, CurrentID=0, Status='live', Symbol='MRPL2', Aliases='CGI-22, MRP-L14, RPML14', description='mitochondrial ribosomal protein L2', other_designations='L2mt|MRP-L2', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43054029, end_position_on_the_genomic_accession=43060186, orientation='minus', exon_count=8, OMIM=611822, no_hdr0=''),
  51070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51070, CurrentID=0, Status='live', Symbol='NOSIP', Aliases='CGI-25', description='nitric oxide synthase interacting protein', other_designations='eNOS-interacting protein', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49555468, end_position_on_the_genomic_accession=49580601, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  51071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51071, CurrentID=0, Status='live', Symbol='DERA', Aliases='CGI-26, DEOC', description='deoxyribose-phosphate aldolase (putative)', other_designations='2-deoxy-D-ribose 5-phosphate aldolase|2-deoxyribose-5-phosphate aldolase homolog|deoxyriboaldolase|phosphodeoxyriboaldolase|putative deoxyribose-phosphate aldolase', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15911172, end_position_on_the_genomic_accession=16037381, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  51072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51072, CurrentID=0, Status='live', Symbol='MEMO1', Aliases='C2orf4, CGI-27, MEMO, NS5ATP7', description='mediator of cell motility 1', other_designations='C21orf19-like protein|HCV NS5A-transactivated protein 7|hepatitis C virus NS5A-transactivated protein 7|mediator of ErbB2-driven cell motility 1', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31867188, end_position_on_the_genomic_accession=32011052, orientation='minus', exon_count=13, OMIM=611786, no_hdr0=''),
  51073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51073, CurrentID=0, Status='live', Symbol='MRPL4', Aliases='CGI-28, L4mt', description='mitochondrial ribosomal protein L4', other_designations='MRP-L4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10251964, end_position_on_the_genomic_accession=10260060, orientation='plus', exon_count=10, OMIM=611823, no_hdr0=''),
  51074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51074, CurrentID=0, Status='live', Symbol='APIP', Aliases='APIP2, CGI-29, CGI29, MMRP19, hAPIP', description='APAF1 interacting protein', other_designations='APAF1-interacting protein|MTRu-1-P dehydratase|probable methylthioribulose-1-phosphate dehydratase', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34882295, end_position_on_the_genomic_accession=34916411, orientation='minus', exon_count=9, OMIM=612491, no_hdr0=''),
  51075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51075, CurrentID=0, Status='live', Symbol='TMX2', Aliases='CGI-31, PDIA12, PIG26, TXNDC14', description='thioredoxin-related transmembrane protein 2', other_designations='cell proliferation-inducing gene 26 protein|growth-inhibiting gene 11|protein disulfide isomerase family A, member 12|thioredoxin domain-containing protein 14', map_location='11cen-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57712523, end_position_on_the_genomic_accession=57740973, orientation='plus', exon_count=8, OMIM=616715, no_hdr0=''),
  51076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51076, CurrentID=0, Status='live', Symbol='CUTC', Aliases='CGI-32', description='cutC copper transporter', other_designations='cutC copper transporter homolog', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99732201, end_position_on_the_genomic_accession=99756137, orientation='plus', exon_count=9, OMIM=610101, no_hdr0=''),
  51077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51077, CurrentID=0, Status='live', Symbol='FCF1', Aliases='Bka, C14orf111, CGI-35, UTP24', description='FCF1 rRNA-processing protein', other_designations='FCF1 small subunit', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74713117, end_position_on_the_genomic_accession=74736691, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  51078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51078, CurrentID=0, Status='live', Symbol='THAP4', Aliases='CGI-36, PP238', description='THAP domain containing 4', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241584405, end_position_on_the_genomic_accession=241637525, orientation='minus', exon_count=8, OMIM=612533, no_hdr0=''),
  51079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51079, CurrentID=0, Status='live', Symbol='NDUFA13', Aliases='B16.6, CDA016, CGI-39, GRIM-19, GRIM19', description='NADH:ubiquinone oxidoreductase subunit A13', other_designations='CI-B16.6|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 13|NADH-ubiquinone oxidoreductase B16.6 subunit|cell death regulatory protein GRIM-19|cell death-regulatory protein GRIM19|complex I B16.6 subunit|complex I-B16.6|gene associated with retinoic and IFN-induced mortality 19 protein|gene associated with retinoic and interferon-induced mortality 19 protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19516210, end_position_on_the_genomic_accession=19528204, orientation='plus', exon_count=5, OMIM=609435, no_hdr0=''),
  106481544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481544, CurrentID=0, Status='live', Symbol='RNU6-1155P', Aliases='', description='RNA, U6 small nuclear 1155, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51081, CurrentID=0, Status='live', Symbol='MRPS7', Aliases='MRP-S, MRP-S7, RP-S7, RPMS7, S7mt, bMRP27a', description='mitochondrial ribosomal protein S7', other_designations='30S ribosomal protein S7 homolog|bMRP-27a', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75261668, end_position_on_the_genomic_accession=75266376, orientation='plus', exon_count=5, OMIM=611974, no_hdr0=''),
  51082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51082, CurrentID=0, Status='live', Symbol='POLR1D', Aliases='AC19, POLR1C, RPA16, RPA9, RPAC2, RPC16, RPO1-3, TCS2', description='polymerase (RNA) I polypeptide D', other_designations='DNA-directed RNA polymerase I subunit D|RNA polymerases I and III subunit AC2|polymerase (RNA) I polypeptide D, 16kDa', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27620743, end_position_on_the_genomic_accession=27667422, orientation='plus', exon_count=5, OMIM=613715, no_hdr0=''),
  51083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51083, CurrentID=0, Status='live', Symbol='GAL', Aliases='GAL-GMAPN, GLNN, GMAP, GAL', description='galanin/GMAP prepropeptide', other_designations='galanin prepropeptide|galanin-message-associated peptide|galanin-related peptide', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68684515, end_position_on_the_genomic_accession=68691175, orientation='plus', exon_count=6, OMIM=137035, no_hdr0=''),
  51084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51084, CurrentID=0, Status='live', Symbol='CRYL1', Aliases='GDH, HEL30, lambda-CRY', description='crystallin lambda 1', other_designations='L-gulonate 3-dehydrogenase|crystallin, lambda 1|crystallin, lamda 1|epididymis luminal protein 30|gul3DH', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20403667, end_position_on_the_genomic_accession=20525873, orientation='minus', exon_count=10, OMIM=609877, no_hdr0=''),
  51085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51085, CurrentID=0, Status='live', Symbol='MLXIPL', Aliases='CHREBP, MIO, MONDOB, WBSCR14, WS-bHLH, bHLHd14', description='MLX interacting protein-like', other_designations='MLX-interacting protein-like|Mlx interactor|WS basic-helix-loop-helix leucine zipper protein|Williams Beuren syndrome chromosome region 14|Williams-Beuren syndrome chromosome region 14 protein 1|Williams-Beuren syndrome chromosome region 14 protein 2|Williams-Beuren syndrome chromosome region 14 protein 3|carbohydrate response element binding protein|class D basic helix-loop-helix protein 14|williams-Beuren syndrome chromosomal region 14 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73593194, end_position_on_the_genomic_accession=73624540, orientation='minus', exon_count=18, OMIM=605678, no_hdr0=''),
  51086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51086, CurrentID=0, Status='live', Symbol='TNNI3K', Aliases='CARK, CCDD', description='TNNI3 interacting kinase', other_designations='cardiac ankyrin repeat kinase|cardiac troponin I-interacting kinase', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74235387, end_position_on_the_genomic_accession=74544432, orientation='plus', exon_count=25, OMIM=613932, no_hdr0=''),
  51087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51087, CurrentID=0, Status='live', Symbol='YBX2', Aliases='CONTRIN, CSDA3, DBPC, MSY2', description='Y-box binding protein 2', other_designations='DNA-binding protein C|MSY2 homolog|Y box binding protein 2|germ cell specific Y-box binding protein|germ cell-specific Y-box-binding protein', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7288252, end_position_on_the_genomic_accession=7294557, orientation='minus', exon_count=9, OMIM=611447, no_hdr0=''),
  51088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51088, CurrentID=0, Status='live', Symbol='KLHL5', Aliases='', description='kelch like family member 5', other_designations='kelch-like family member 5|lymphocyte activation-associated protein', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39044801, end_position_on_the_genomic_accession=39142823, orientation='plus', exon_count=16, OMIM=608064, no_hdr0=''),
  106481553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481553, CurrentID=0, Status='live', Symbol='RNU6-1184P', Aliases='', description='RNA, U6 small nuclear 1184, pseudogene', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51090, CurrentID=0, Status='live', Symbol='PLLP', Aliases='PMLP, TM4SF11', description='plasmolipin', other_designations='plasma membrane proteolipid (plasmolipin)|transmembrane 4 superfamily member 11 (plasmolipin)', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57256097, end_position_on_the_genomic_accession=57284672, orientation='minus', exon_count=4, OMIM=600340, no_hdr0=''),
  51091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51091, CurrentID=0, Status='live', Symbol='SEPSECS', Aliases='LP, PCH2D, SLA, SLA/LP', description='Sep (O-phosphoserine) tRNA:Sec (selenocysteine) tRNA synthase', other_designations='SLA-p35|SLA/LP autoantigen|UGA suppressor tRNA-associated protein|liver-pancreas antigen|soluble liver antigen|soluble liver antigen/liver pancreas antigen|tRNA(Ser/Sec)-associated antigenic protein', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25120005, end_position_on_the_genomic_accession=25160582, orientation='minus', exon_count=14, OMIM=613009, no_hdr0=''),
  51092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51092, CurrentID=0, Status='live', Symbol='SIDT2', Aliases='CGI-40', description='SID1 transmembrane family member 2', other_designations='SID1 transmembrane family, member 2', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117179223, end_position_on_the_genomic_accession=117197445, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  51093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51093, CurrentID=0, Status='live', Symbol='RRNAD1', Aliases='C1orf66, CGI-41', description='ribosomal RNA adenine dimethylase domain containing 1', other_designations='ribosomal RNA adenine dimethylase domain-containing protein 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156728471, end_position_on_the_genomic_accession=156736960, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51094, CurrentID=0, Status='live', Symbol='ADIPOR1', Aliases='ACDCR1, CGI-45, CGI45, PAQR1, TESBP1A', description='adiponectin receptor 1', other_designations='progestin and adipoQ receptor family member I', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202940825, end_position_on_the_genomic_accession=202958572, orientation='minus', exon_count=11, OMIM=607945, no_hdr0=''),
  51095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51095, CurrentID=0, Status='live', Symbol='TRNT1', Aliases='CCA1, CGI-47, MtCCA, SIFD', description='tRNA nucleotidyl transferase, CCA-adding, 1', other_designations='mitochondrial CCA-adding tRNA-nucleotidyltransferase|mt CCA-adding enzyme|mt tRNA CCA-diphosphorylase|mt tRNA CCA-pyrophosphorylase|mt tRNA adenylyltransferase|tRNA-nucleotidyltransferase 1, mitochondrial', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3126916, end_position_on_the_genomic_accession=3152503, orientation='plus', exon_count=12, OMIM=612907, no_hdr0=''),
  51096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51096, CurrentID=0, Status='live', Symbol='UTP18', Aliases='CGI-48, WDR50', description='UTP18 small subunit processome component', other_designations='UTP18 small subunit (SSU) processome component homolog|UTP18, small subunit (SSU) processome component, homolog|WD repeat domain 50|WD repeat-containing protein 50', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51260536, end_position_on_the_genomic_accession=51297933, orientation='plus', exon_count=16, OMIM=612816, no_hdr0=''),
  51097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51097, CurrentID=0, Status='live', Symbol='SCCPDH', Aliases='CGI-49, NET11', description='saccharopine dehydrogenase (putative)', other_designations='probable saccharopine dehydrogenase', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246724076, end_position_on_the_genomic_accession=246768137, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  51098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51098, CurrentID=0, Status='live', Symbol='IFT52', Aliases='C20orf9, CGI-53, NGD2, NGD5', description='intraflagellar transport 52', other_designations='protein NGD5 homolog', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43590613, end_position_on_the_genomic_accession=43647227, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  51099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51099, CurrentID=0, Status='live', Symbol='ABHD5', Aliases='CDS, CGI58, IECN2, NCIE2', description='abhydrolase domain containing 5', other_designations='abhydrolase domain-containing protein 5|lipid droplet-binding protein CGI-58|truncated abhydrolase domain-containing protein 5', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43690883, end_position_on_the_genomic_accession=43722725, orientation='plus', exon_count=8, OMIM=604780, no_hdr0=''),
  51100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51100, CurrentID=0, Status='live', Symbol='SH3GLB1', Aliases='Bif-1, CGI-61, PPP1R70, dJ612B15.2', description='SH3-domain GRB2-like endophilin B1', other_designations='Bax-interacting factor 1|SH3 domain-containing GRB2-like protein B1|SH3-containing protein SH3GLB1|protein phosphatase 1, regulatory subunit 70', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86704570, end_position_on_the_genomic_accession=86748184, orientation='plus', exon_count=11, OMIM=609287, no_hdr0=''),
  51101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51101, CurrentID=0, Status='live', Symbol='ZC2HC1A', Aliases='C8orf70, CGI-62, FAM164A', description='zinc finger, C2HC-type containing 1A', other_designations='family with sequence similarity 164, member A|protein FAM164A', map_location='8q21.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78666047, end_position_on_the_genomic_accession=78719765, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  51102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51102, CurrentID=0, Status='live', Symbol='MECR', Aliases='CGI-63, FASN2B, NRBF1', description='mitochondrial trans-2-enoyl-CoA reductase', other_designations='homolog of yeast 2-enoyl thioester reductase|mitochondrial 2-enoyl thioester reductase|nuclear receptor binding factor 1', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=29167696, end_position_on_the_genomic_accession=29230958, orientation='minus', exon_count=20, OMIM=608205, no_hdr0=''),
  51103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51103, CurrentID=0, Status='live', Symbol='NDUFAF1', Aliases='CGI-65, CGI65, CIA30', description='NADH:ubiquinone oxidoreductase complex assembly factor 1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 1|NADH dehydrogenase (ubiquinone) complex I, assembly factor 1|NADH-ubiquinone oxidoreductase 1 alpha subcomplex, assembly factor 1', map_location='15q11.2-q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41387349, end_position_on_the_genomic_accession=41402554, orientation='minus', exon_count=6, OMIM=606934, no_hdr0=''),
  51104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51104, CurrentID=0, Status='live', Symbol='ABHD17B', Aliases='C9orf77, CGI-67, FAM108B1', description='abhydrolase domain containing 17B', other_designations='abhydrolase domain-containing protein FAM108B1|family with sequence similarity 108, member B1', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=71862452, end_position_on_the_genomic_accession=71911233, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51105, CurrentID=0, Status='live', Symbol='PHF20L1', Aliases='CGI-72, TDRD20B, URLC1', description='PHD finger protein 20-like 1', other_designations='tudor domain containing 20B|tudor domain-containing protein PHF20L1|up-regulated in lung cancer 1', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132775358, end_position_on_the_genomic_accession=132848807, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  51106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51106, CurrentID=0, Status='live', Symbol='TFB1M', Aliases='CGI-75, CGI75, mtTFB, mtTFB1', description='transcription factor B1, mitochondrial', other_designations="S-adenosylmethionine-6-N', N'-adenosyl(rRNA) dimethyltransferase 1|h-mtTFB1|hTFB1M|homolog of yeast mitochondrial transcription factor B|mitochondrial 12S rRNA dimethylase 1|mitochondrial dimethyladenosine transferase 1", map_location='6q25.1-q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155247238, end_position_on_the_genomic_accession=155314497, orientation='minus', exon_count=12, OMIM=607033, no_hdr0=''),
  51107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51107, CurrentID=0, Status='live', Symbol='APH1A', Aliases='6530402N02Rik, APH-1, APH-1A, CGI-78', description='aph-1 homolog A, gamma secretase subunit', other_designations='APH1A gamma secretase subunit|anterior pharynx defective 1 homolog A|aph-1alpha|presenilin-stabilization factor', map_location='1p36.13-q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150265399, end_position_on_the_genomic_accession=150269209, orientation='minus', exon_count=7, OMIM=607629, no_hdr0=''),
  51108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51108, CurrentID=0, Status='live', Symbol='METTL9', Aliases='CGI-81, DREV, DREV1, PAP1', description='methyltransferase like 9', other_designations='CTB-31N19.3|DORA reverse strand protein 1|p53 activated protein 1', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21597221, end_position_on_the_genomic_accession=21657473, orientation='plus', exon_count=8, OMIM=609388, no_hdr0=''),
  51109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51109, CurrentID=0, Status='live', Symbol='RDH11', Aliases='ARSDR1, CGI82, HCBP12, MDT1, PSDR1, RALR1, RDJCSS, SCALD, SDR7C1', description='retinol dehydrogenase 11 (all-trans/9-cis/11-cis)', other_designations='HCV core-binding protein HCBP12|androgen-regulated short-chain dehydrogenase/reductase 1|prostate short-chain dehydrogenase reductase 1|retinal reductase 1|short chain dehydrogenase/reductase family 7C, member 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67676800, end_position_on_the_genomic_accession=67695793, orientation='minus', exon_count=7, OMIM=607849, no_hdr0=''),
  51110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51110, CurrentID=0, Status='live', Symbol='LACTB2', Aliases='CGI-83', description='lactamase beta 2', other_designations='lactamase, beta 2', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70637266, end_position_on_the_genomic_accession=70669212, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51111, CurrentID=0, Status='live', Symbol='KMT5B', Aliases='CGI-85, CGI85, SUV420H1', description='lysine (K)-specific methyltransferase 5B', other_designations='C630029K18Rik|lysine N-methyltransferase 5B|su(var)4-20 homolog 1|suppressor of variegation 4-20 homolog 1|suv4-20h1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68154863, end_position_on_the_genomic_accession=68213783, orientation='minus', exon_count=13, OMIM=610881, no_hdr0=''),
  51112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51112, CurrentID=0, Status='live', Symbol='TRAPPC12', Aliases='CGI-87, TTC-15, TTC15', description='trafficking protein particle complex 12', other_designations='TPR repeat protein 15|tetratricopeptide repeat domain 15|tetratricopeptide repeat protein 15', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3379672, end_position_on_the_genomic_accession=3479571, orientation='plus', exon_count=19, OMIM=614139, no_hdr0=''),
  106481577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481577, CurrentID=0, Status='live', Symbol='RNU6-1257P', Aliases='', description='RNA, U6 small nuclear 1257, pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51114, CurrentID=0, Status='live', Symbol='ZDHHC9', Aliases='CGI89, CXorf11, DHHC9, MMSA1, MRXSZ, ZDHHC10, ZNF379, ZNF380', description='zinc finger, DHHC-type containing 9', other_designations='Asp-His-His-Cys domain containing protein 9|antigen MMSA-1|zinc finger protein 379|zinc finger protein 380|zinc finger, DHHC domain containing 10', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129803288, end_position_on_the_genomic_accession=129843934, orientation='minus', exon_count=13, OMIM=300646, no_hdr0=''),
  51115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51115, CurrentID=0, Status='live', Symbol='RMDN1', Aliases='CGI-90, FAM82B, RMD-1, RMD1', description='regulator of microtubule dynamics 1', other_designations='family with sequence similarity 82, member B|microtubule-associated protein', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86467398, end_position_on_the_genomic_accession=86514338, orientation='minus', exon_count=14, OMIM=611871, no_hdr0=''),
  51116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51116, CurrentID=0, Status='live', Symbol='MRPS2', Aliases='CGI-91, MRP-S2, S2mt', description='mitochondrial ribosomal protein S2', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135499986, end_position_on_the_genomic_accession=135504673, orientation='plus', exon_count=7, OMIM=611971, no_hdr0=''),
  51117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51117, CurrentID=0, Status='live', Symbol='COQ4', Aliases='CGI-92, COQ10D7', description='coenzyme Q4', other_designations='coenzyme Q biosynthesis protein 4 homolog|coenzyme Q4 homolog', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128322512, end_position_on_the_genomic_accession=128334072, orientation='plus', exon_count=8, OMIM=612898, no_hdr0=''),
  51118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51118, CurrentID=0, Status='live', Symbol='UTP11L', Aliases='CGI-94, CGI94', description='UTP11-like, U3 small nucleolar ribonucleoprotein (yeast)', other_designations='U3 snoRNA-associated protein 11|UTP11-like protein', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38012712, end_position_on_the_genomic_accession=38024825, orientation='plus', exon_count=8, OMIM=609440, no_hdr0=''),
  51119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51119, CurrentID=0, Status='live', Symbol='SBDS', Aliases='CGI-97, SDS, SWDS', description='Shwachman-Bodian-Diamond syndrome', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66987703, end_position_on_the_genomic_accession=66995601, orientation='minus', exon_count=5, OMIM=607444, no_hdr0=''),
  106481584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481584, CurrentID=0, Status='live', Symbol='RNU6-1280P', Aliases='', description='RNA, U6 small nuclear 1280, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51121, CurrentID=0, Status='live', Symbol='RPL26L1', Aliases='RPL26P1', description='ribosomal protein L26 like 1', other_designations='ribosomal protein L26 homolog|ribosomal protein L26 pseudogene 1', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172954294, end_position_on_the_genomic_accession=172969771, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  51122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51122, CurrentID=0, Status='live', Symbol='COMMD2', Aliases='HSPC042', description='COMM domain containing 2', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149738470, end_position_on_the_genomic_accession=149752510, orientation='minus', exon_count=6, OMIM=616699, no_hdr0=''),
  51123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51123, CurrentID=0, Status='live', Symbol='ZNF706', Aliases='HSPC038, PNAS-106, PNAS-113', description='zinc finger protein 706', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101197038, end_position_on_the_genomic_accession=101206064, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51124, CurrentID=0, Status='live', Symbol='IER3IP1', Aliases='HSPC039, MEDS, PRO2309', description='immediate early response 3 interacting protein 1', other_designations='', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47155019, end_position_on_the_genomic_accession=47176374, orientation='minus', exon_count=3, OMIM=609382, no_hdr0=''),
  51125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51125, CurrentID=0, Status='live', Symbol='GOLGA7', Aliases='GCP16, GOLGA3AP1A, HSPC041, GOLGA7', description='golgin A7', other_designations='Golgi complex-associated protein of 16kDa|golgi autoantigen, golgin subfamily a, 7|golgi complex-associated protein of 16 kDa', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41490562, end_position_on_the_genomic_accession=41510980, orientation='plus', exon_count=6, OMIM=609453, no_hdr0=''),
  51126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51126, CurrentID=0, Status='live', Symbol='NAA20', Aliases='NAT3, NAT3P, NAT5, NAT5P, dJ1002M8.1', description='N(alpha)-acetyltransferase 20, NatB catalytic subunit', other_designations='N-acetyltransferase 3 homolog|N-acetyltransferase 5 (ARD1 homolog, S. cerevisiae)|N-acetyltransferase 5 (GCN5-related, putative)|N-acetyltransferase 5, ARD1 subunit (arrest-defective 1, S. cerevisiae, homolog)|N-alpha-acetyltransferase 20, NatB catalytic subunit|N-terminal acetyltransferase B complex catalytic subunit NAA20|N-terminal acetyltransferase B complex catalytic subunit NAT5|N-terminal acetyltransferase complex ARD1 subunit|methionine N-acetyltransferase|natB catalytic subunit|natB complex subunit NAT5', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20017290, end_position_on_the_genomic_accession=20033629, orientation='plus', exon_count=7, OMIM=610833, no_hdr0=''),
  51127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51127, CurrentID=0, Status='live', Symbol='TRIM17', Aliases='RBCC, RNF16, terf', description='tripartite motif containing 17', other_designations='RING finger protein terf|ring finger protein 16|testis RING finger protein|tripartite motif-containing 17|tripartite motif-containing protein 17', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228407935, end_position_on_the_genomic_accession=228416882, orientation='minus', exon_count=7, OMIM=606123, no_hdr0=''),
  51128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51128, CurrentID=0, Status='live', Symbol='SAR1B', Aliases='ANDD, CMRD, GTBPB, SARA2', description='secretion associated, Ras related GTPase 1B', other_designations='2310075M17Rik|GTP-binding protein B|GTP-binding protein Sara|SAR1 homolog B|SAR1a gene homolog 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134601149, end_position_on_the_genomic_accession=134632843, orientation='minus', exon_count=8, OMIM=607690, no_hdr0=''),
  51129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51129, CurrentID=0, Status='live', Symbol='ANGPTL4', Aliases='ARP4, FIAF, HARP, HFARP, NL2, PGAR, TGQTL, UNQ171, pp1158', description='angiopoietin like 4', other_designations='PPARG angiopoietin related protein|angiopoietin-like 4|fasting-induced adipose factor|hepatic angiopoietin-related protein|hepatic fibrinogen/angiopoietin-related protein|peroxisome proliferator-activated receptor (PPAR) gamma induced angiopoietin-related protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8364127, end_position_on_the_genomic_accession=8374375, orientation='plus', exon_count=7, OMIM=605910, no_hdr0=''),
  51130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51130, CurrentID=0, Status='live', Symbol='ASB3', Aliases='ASB-3', description='ankyrin repeat and SOCS box containing 3', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53669980, end_position_on_the_genomic_accession=53787009, orientation='minus', exon_count=10, OMIM=605760, no_hdr0=''),
  51131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51131, CurrentID=0, Status='live', Symbol='PHF11', Aliases='APY, BCAP, IGEL, IGER, IGHER, NY-REN-34, NYREN34', description='PHD finger protein 11', other_designations='BRCA1 C-terminus-associated protein|IgE responsiveness (atopic)|NY-REN-34 antigen|renal carcinoma antigen NY-REN-34', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49495610, end_position_on_the_genomic_accession=49528987, orientation='plus', exon_count=15, OMIM=607796, no_hdr0=''),
  51132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51132, CurrentID=0, Status='live', Symbol='RLIM', Aliases='NY-REN-43, RNF12', description='ring finger protein, LIM domain interacting', other_designations='E3 ubiquitin-protein ligase RNF12|LIM domain-interacting RING finger protein|R-LIM|RING finger LIM domain-binding protein|renal carcinoma antigen NY-REN-43|ring finger protein 12|ring zinc finger LIM domain binding protein|ring zinc finger protein NY-REN-43antigen', map_location='Xq13-q21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74582976, end_position_on_the_genomic_accession=74614626, orientation='minus', exon_count=5, OMIM=300379, no_hdr0=''),
  51133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51133, CurrentID=0, Status='live', Symbol='KCTD3', Aliases='NY-REN-45', description='potassium channel tetramerization domain containing 3', other_designations='NY-REN-45 antigen|potassium channel tetramerisation domain containing 3|renal carcinoma antigen NY-REN-45', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215567185, end_position_on_the_genomic_accession=215621807, orientation='plus', exon_count=18, OMIM=613272, no_hdr0=''),
  51134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51134, CurrentID=0, Status='live', Symbol='CEP83', Aliases='CCDC41, NPHP18, NY-REN-58', description='centrosomal protein 83kDa', other_designations='NY-REN-58 antigen|coiled-coil domain containing 41|coiled-coil domain-containing protein 41|renal carcinoma antigen NY-REN-58', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94265652, end_position_on_the_genomic_accession=94459988, orientation='minus', exon_count=28, OMIM=615847, no_hdr0=''),
  51135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51135, CurrentID=0, Status='live', Symbol='IRAK4', Aliases='IPD1, IRAK-4, NY-REN-64, REN64', description='interleukin 1 receptor associated kinase 4', other_designations='renal carcinoma antigen NY-REN-64', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43758944, end_position_on_the_genomic_accession=43789543, orientation='plus', exon_count=19, OMIM=606883, no_hdr0=''),
  51136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51136, CurrentID=0, Status='live', Symbol='RNFT1', Aliases='PTD016', description='ring finger protein, transmembrane 1', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59952362, end_position_on_the_genomic_accession=59964756, orientation='minus', exon_count=9, OMIM=615172, no_hdr0=''),
  106481601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481601, CurrentID=0, Status='live', Symbol='RNU6-1333P', Aliases='', description='RNA, U6 small nuclear 1333, pseudogene', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51138, CurrentID=0, Status='live', Symbol='COPS4', Aliases='CSN4', description='COP9 signalosome subunit 4', other_designations='COP9 constitutive photomorphogenic homolog subunit 4|COP9 constitutive photomorphogenic-like protein subunit 4|JAB1-containing signalosome subunit 4|SGN4|signalosome subunit 4', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83035086, end_position_on_the_genomic_accession=83075818, orientation='plus', exon_count=11, OMIM=616008, no_hdr0=''),
  106481603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481603, CurrentID=0, Status='live', Symbol='RNU6-1339P', Aliases='', description='RNA, U6 small nuclear 1339, pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481604, CurrentID=0, Status='live', Symbol='RNU1-8P', Aliases='RNU1-8', description='RNA, U1 small nuclear 8, pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51141, CurrentID=0, Status='live', Symbol='INSIG2', Aliases='', description='insulin induced gene 2', other_designations='INSIG-2|INSIG2 membrane protein|insulin induced protein 2', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118088474, end_position_on_the_genomic_accession=118110027, orientation='plus', exon_count=7, OMIM=608660, no_hdr0=''),
  51142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51142, CurrentID=0, Status='live', Symbol='CHCHD2', Aliases='C7orf17', description='coiled-coil-helix-coiled-coil-helix domain containing 2', other_designations='16.7kD protein|HCV NS2 trans-regulated protein|NS2TP|aging-associated gene 10 protein|coiled-coil-helix-coiled-coil-helix domain-containing protein 2, mitochondrial', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56101573, end_position_on_the_genomic_accession=56106494, orientation='minus', exon_count=4, OMIM=616244, no_hdr0=''),
  51143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51143, CurrentID=0, Status='live', Symbol='DYNC1LI1', Aliases='DNCLI1, LIC1', description='dynein, cytoplasmic 1, light intermediate chain 1', other_designations='DLC-A|dynein light chain A|dynein light chain-A|dynein light intermediate chain 1, cytosolic|dynein, cytoplasmic, light intermediate polypeptide 1', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32525971, end_position_on_the_genomic_accession=32570874, orientation='minus', exon_count=13, OMIM=615890, no_hdr0=''),
  51144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51144, CurrentID=0, Status='live', Symbol='HSD17B12', Aliases='KAR, SDR12C1', description='hydroxysteroid (17-beta) dehydrogenase 12', other_designations='17-beta-HSD 12|17-beta-hydroxysteroid dehydrogenase 12|17beta-HSD type 12|3-ketoacyl-CoA reductase|estradiol 17-beta-dehydrogenase 12|short chain dehydrogenase/reductase family 12C member 1|short chain dehydrogenase/reductase family 12C, member 1|steroid dehydrogenase homolog', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43680593, end_position_on_the_genomic_accession=43856619, orientation='plus', exon_count=11, OMIM=609574, no_hdr0=''),
  106481609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481609, CurrentID=0, Status='live', Symbol='RNU1-42P', Aliases='', description='RNA, U1 small nuclear 42, pseudogene', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51146, CurrentID=0, Status='live', Symbol='A4GNT', Aliases='alpha4GnT', description='alpha-1,4-N-acetylglucosaminyltransferase', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138123718, end_position_on_the_genomic_accession=138132387, orientation='minus', exon_count=3, OMIM=616709, no_hdr0=''),
  51147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51147, CurrentID=0, Status='live', Symbol='ING4', Aliases='my036, p29ING4', description='inhibitor of growth family member 4', other_designations='brain my036 protein|candidate tumor suppressor p33 ING1 homolog|inhibitor of growth family, member 4', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6650538, end_position_on_the_genomic_accession=6663149, orientation='minus', exon_count=9, OMIM=608524, no_hdr0=''),
  51148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51148, CurrentID=0, Status='live', Symbol='CERCAM', Aliases='CEECAM1, GLT25D3', description='cerebral endothelial cell adhesion molecule', other_designations='cerebral cell adhesion molecule|cerebral endothelial cell adhesion molecule 1|glycosyltransferase 25 domain containing 3|glycosyltransferase 25 family member 3', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128411575, end_position_on_the_genomic_accession=128437351, orientation='plus', exon_count=18, OMIM=616626, no_hdr0=''),
  51149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51149, CurrentID=0, Status='live', Symbol='C5orf45', Aliases='', description='chromosome 5 open reading frame 45', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179837266, end_position_on_the_genomic_accession=179858840, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51150, CurrentID=0, Status='live', Symbol='SDF4', Aliases='Cab45, SDF-4', description='stromal cell derived factor 4', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1216908, end_position_on_the_genomic_accession=1232067, orientation='minus', exon_count=8, OMIM=614282, no_hdr0=''),
  51151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51151, CurrentID=0, Status='live', Symbol='SLC45A2', Aliases='1A1, AIM1, MATP, OCA4, SHEP5', description='solute carrier family 45 member 2', other_designations='melanoma antigen AIM1|protein AIM-1|solute carrier family 45, member 2|underwhite', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33944616, end_position_on_the_genomic_accession=33985144, orientation='minus', exon_count=10, OMIM=606202, no_hdr0=''),
  51152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51152, CurrentID=0, Status='live', Symbol='LINC00328', Aliases='NCRNA00328, NCRNA00328A', description='long intergenic non-protein coding RNA 328', other_designations='NCRNA00328-1|melanoma antigen|non-protein coding RNA 328A', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481617, CurrentID=0, Status='live', Symbol='RNU1-76P', Aliases='', description='RNA, U1 small nuclear 76, pseudogene', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51154, CurrentID=0, Status='live', Symbol='MRTO4', Aliases='C1orf33, MRT4, dJ657E11.4', description='MRT4 homolog, ribosome maturation factor', other_designations='60S acidic ribosomal protein PO|MRT4, mRNA turnover 4, homolog|MRTO4 ribosome maturation factor|mRNA turnover 4 homolog|ribosome assembly factor MRTO4', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19251581, end_position_on_the_genomic_accession=19260128, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51155, CurrentID=0, Status='live', Symbol='HN1', Aliases='ARM2A, HN1', description='hematological and neurological expressed 1', other_designations='androgen-regulated protein 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75135243, end_position_on_the_genomic_accession=75154683, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51156, CurrentID=0, Status='live', Symbol='SERPINA10', Aliases='PZI, ZPI', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 10', other_designations='PZ-dependent protease inhibitor|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 10|serpin A10', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94283313, end_position_on_the_genomic_accession=94293271, orientation='minus', exon_count=6, OMIM=605271, no_hdr0=''),
  51157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51157, CurrentID=0, Status='live', Symbol='ZNF580', Aliases='', description='zinc finger protein 580', other_designations='LDL induced EC protein|LDL-induced EC protein', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55634903, end_position_on_the_genomic_accession=55643470, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106481622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481622, CurrentID=0, Status='live', Symbol='RNU1-98P', Aliases='', description='RNA, U1 small nuclear 98, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481623, CurrentID=0, Status='live', Symbol='RNU1-103P', Aliases='', description='RNA, U1 small nuclear 103, pseudogene', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51160, CurrentID=0, Status='live', Symbol='VPS28', Aliases='', description='vacuolar protein sorting 28 homolog (S. cerevisiae)', other_designations='ESCRT-I complex subunit VPS28|vacuolar protein sorting 28-like protein|yeast class E protein Vps28p homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144423601, end_position_on_the_genomic_accession=144428563, orientation='minus', exon_count=10, OMIM=611952, no_hdr0=''),
  51161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51161, CurrentID=0, Status='live', Symbol='C3orf18', Aliases='G20', description='chromosome 3 open reading frame 18', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50558025, end_position_on_the_genomic_accession=50574002, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  51162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51162, CurrentID=0, Status='live', Symbol='EGFL7', Aliases='NEU1, VE-STATIN, ZNEU1', description='EGF-like-domain, multiple 7', other_designations='EGF-like protein 7|NOTCH4-like protein|multiple EGF-like domains protein 7|multiple epidermal growth factor-like domains protein 7|vascular endothelial statin', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136658856, end_position_on_the_genomic_accession=136672678, orientation='plus', exon_count=13, OMIM=608582, no_hdr0=''),
  51163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51163, CurrentID=0, Status='live', Symbol='DBR1', Aliases='', description='debranching RNA lariats 1', other_designations='RNA lariat debranching enzyme|debranching enzyme homolog 1', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138160988, end_position_on_the_genomic_accession=138174949, orientation='minus', exon_count=8, OMIM=607024, no_hdr0=''),
  51164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51164, CurrentID=0, Status='live', Symbol='DCTN4', Aliases='DYN4, P62', description='dynactin subunit 4', other_designations='dynactin 4 (p62)|dynactin p62 subunit|dynactin subunit p62', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150708747, end_position_on_the_genomic_accession=150759095, orientation='minus', exon_count=19, OMIM=614758, no_hdr0=''),
  106481629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481629, CurrentID=0, Status='live', Symbol='RNU1-124P', Aliases='', description='RNA, U1 small nuclear 124, pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51166, CurrentID=0, Status='live', Symbol='AADAT', Aliases='KAT2, KATII', description='aminoadipate aminotransferase', other_designations='2-aminoadipate aminotransferase|2-aminoadipate transaminase|KAT/AadAT|L kynurenine/alpha aminoadipate aminotransferase|alpha-aminoadipate aminotransferase|kynurenine aminotransferase II|kynurenine--oxoglutarate aminotransferase II|kynurenine--oxoglutarate transaminase 2|kynurenine--oxoglutarate transaminase II', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170060222, end_position_on_the_genomic_accession=170091699, orientation='minus', exon_count=18, OMIM=611754, no_hdr0=''),
  51167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51167, CurrentID=0, Status='live', Symbol='CYB5R4', Aliases='NCB5OR, cb5/cb5R, dJ676J13.1', description='cytochrome b5 reductase 4', other_designations='N-terminal cytochrome b5 and cytochrome b5 oxidoreductase domain-containing protein|NADPH cytochrome B5 oxidoreductase|cytochrome b-type NAD(P)H oxidoreductase|flavohemoprotein b5+b5R|flavohemoprotein b5/b5R', map_location='6pter-q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83859651, end_position_on_the_genomic_accession=83960427, orientation='plus', exon_count=16, OMIM=608343, no_hdr0=''),
  51168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51168, CurrentID=0, Status='live', Symbol='MYO15A', Aliases='DFNB3, MYO15', description='myosin XVA', other_designations='myosin-XV|unconventional myosin-15', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18108706, end_position_on_the_genomic_accession=18180225, orientation='plus', exon_count=67, OMIM=602666, no_hdr0=''),
  106481633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481633, CurrentID=0, Status='live', Symbol='RNU1-141P', Aliases='', description='RNA, U1 small nuclear 141, pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51170, CurrentID=0, Status='live', Symbol='HSD17B11', Aliases='17-BETA-HSD11, 17-BETA-HSDXI, 17BHSD11, DHRS8, PAN1B, RETSDR2, SDR16C2', description='hydroxysteroid (17-beta) dehydrogenase 11', other_designations='17-beta-hydroxysteroid dehydrogenase type XI|CTCL tumor antigen HD-CL-03|CTCL-associated antigen HD-CL-03|cutaneous T-cell lymphoma-associated antigen HD-CL-03|dehydrogenase/reductase SDR family member 8|retinal short-chain dehydrogenase/reductase 2|short chain dehydrogenase/reductase family 16C member 2|short chain dehydrogenase/reductase family 16C, member 2', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87336522, end_position_on_the_genomic_accession=87391303, orientation='minus', exon_count=8, OMIM=612831, no_hdr0=''),
  51171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51171, CurrentID=0, Status='live', Symbol='HSD17B14', Aliases='DHRS10, SDR47C1, retSDR3', description='hydroxysteroid (17-beta) dehydrogenase 14', other_designations='17-beta-HSD 14|17-beta-hydroxysteroid dehydrogenase DHRS10|dehydrogenase/reductase (SDR family) member 10|dehydrogenase/reductase SDR family member 10|retinal short-chain dehydrogenase/reductase 3|retinal short-chain dehydrogenase/reductase retSDR3|short chain dehydrogenase/reductase family 47C member 1|short chain dehydrogenase/reductase family 47C, member 1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48813017, end_position_on_the_genomic_accession=48836703, orientation='minus', exon_count=9, OMIM=612832, no_hdr0=''),
  51172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51172, CurrentID=0, Status='live', Symbol='NAGPA', Aliases='APAA, UCE', description='N-acetylglucosamine-1-phosphodiester alpha-N-acetylglucosaminidase', other_designations='alpha-N-acetylglucosaminyl phosphodiesterase|lysosomal alpha-N-acetylglucosaminidase|mannose 6-phosphate-uncovering enzyme|phosphodiester alpha-GlcNAcase', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5024844, end_position_on_the_genomic_accession=5033941, orientation='minus', exon_count=11, OMIM=607985, no_hdr0=''),
  106481637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481637, CurrentID=0, Status='live', Symbol='RNU2-11P', Aliases='', description='RNA, U2 small nuclear 11, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51174, CurrentID=0, Status='live', Symbol='TUBD1', Aliases='TUBD', description='tubulin delta 1', other_designations='delta-tubulin|tubulin, delta 1', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59859479, end_position_on_the_genomic_accession=59892945, orientation='minus', exon_count=10, OMIM=607344, no_hdr0=''),
  51175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51175, CurrentID=0, Status='live', Symbol='TUBE1', Aliases='TUBE, dJ142L7.2', description='tubulin epsilon 1', other_designations='epsilon-tubulin|tubulin, epsilon 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112070657, end_position_on_the_genomic_accession=112087548, orientation='minus', exon_count=13, OMIM=607345, no_hdr0=''),
  51176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51176, CurrentID=0, Status='live', Symbol='LEF1', Aliases='LEF-1, TCF10, TCF1ALPHA, TCF7L3', description='lymphoid enhancer-binding factor 1', other_designations='T cell-specific transcription factor 1-alpha|TCF1-alpha', map_location='4q23-q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108047545, end_position_on_the_genomic_accession=108168956, orientation='minus', exon_count=13, OMIM=153245, no_hdr0=''),
  51177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51177, CurrentID=0, Status='live', Symbol='PLEKHO1', Aliases='CKIP-1, CKIP1, JBP, OC120', description='pleckstrin homology domain containing O1', other_designations='C-Jun-binding protein|CK2 interacting protein 1; HQ0024c protein|CK2-interacting protein 1|PH domain-containing family O member 1|casein kinase 2-interacting protein 1|osteoclast maturation-associated gene 120 protein|pleckstrin homology domain containing, family O member 1', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150149434, end_position_on_the_genomic_accession=150160065, orientation='plus', exon_count=7, OMIM=608335, no_hdr0=''),
  106481642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481642, CurrentID=0, Status='live', Symbol='RNU2-27P', Aliases='', description='RNA, U2 small nuclear 27, pseudogene', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51179, CurrentID=0, Status='live', Symbol='HAO2', Aliases='GIG16, HAOX2', description='hydroxyacid oxidase 2 (long chain)', other_designations='(S)-2-hydroxy-acid oxidase, peroxisomal|cell growth-inhibiting gene 16 protein|glycolate oxidase|long chain alpha-hydroxy acid oxidase|long-chain L-2-hydroxy acid oxidase', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119368776, end_position_on_the_genomic_accession=119394130, orientation='plus', exon_count=10, OMIM=605176, no_hdr0=''),
  106481644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481644, CurrentID=0, Status='live', Symbol='RNU2-32P', Aliases='', description='RNA, U2 small nuclear 32, pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51181, CurrentID=0, Status='live', Symbol='DCXR', Aliases='DCR, HCR2, HCRII, KIDCR, P34H, PNTSU, SDR20C1, XR', description='dicarbonyl/L-xylulose reductase', other_designations='carbonyl reductase 2|carbonyl reductase II|kidney dicarbonyl reductase|short chain dehydrogenase/reductase family 20C member 1|short chain dehydrogenase/reductase family 20C, member 1|sperm surface protein P34H', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82035881, end_position_on_the_genomic_accession=82037697, orientation='minus', exon_count=8, OMIM=608347, no_hdr0=''),
  51182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51182, CurrentID=0, Status='live', Symbol='HSPA14', Aliases='HSP70-4, HSP70L1', description='heat shock protein family A (Hsp70) member 14', other_designations='HSP70-like protein 1|heat shock 70kDa protein 14 isoform 1 variant 3|heat shock protein HSP60|heat shock protein hsp70-related protein', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14838160, end_position_on_the_genomic_accession=14871741, orientation='plus', exon_count=16, OMIM=610369, no_hdr0=''),
  106481647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481647, CurrentID=0, Status='live', Symbol='RNU2-42P', Aliases='', description='RNA, U2 small nuclear 42, pseudogene', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51184, CurrentID=0, Status='live', Symbol='GPN3', Aliases='ATPBD1C', description='GPN-loop GTPase 3', other_designations='ATP binding domain 1 family, member C|ATP-binding domain 1 family member C|protein x 0004', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110452486, end_position_on_the_genomic_accession=110468721, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  51185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51185, CurrentID=0, Status='live', Symbol='CRBN', Aliases='MRT2, MRT2A', description='cereblon', other_designations='protein x 0001', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3149633, end_position_on_the_genomic_accession=3179717, orientation='minus', exon_count=12, OMIM=609262, no_hdr0=''),
  51186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51186, CurrentID=0, Status='live', Symbol='WBP5', Aliases='TCEAL9, WEX6', description='WW domain binding protein 5', other_designations='WBP-5|WW domain binding protein 1|pp21 homolog', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103356452, end_position_on_the_genomic_accession=103358469, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  51187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51187, CurrentID=0, Status='live', Symbol='RSL24D1', Aliases='C15orf15, HRP-L30-iso, L30, RLP24, RPL24, RPL24L, TVAS3', description='ribosomal L24 domain containing 1', other_designations='60S ribosomal protein L30 isolog|homolog of yeast ribosomal like protein 24|my024 protein|ribosomal L24 domain-containing protein 1', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55181314, end_position_on_the_genomic_accession=55197033, orientation='minus', exon_count=6, OMIM=613262, no_hdr0=''),
  51188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51188, CurrentID=0, Status='live', Symbol='SS18L2', Aliases='KIAA-iso', description='synovial sarcoma translocation gene on chromosome 18-like 2', other_designations='SYT homolog 2', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42581840, end_position_on_the_genomic_accession=42594998, orientation='plus', exon_count=4, OMIM=606473, no_hdr0=''),
  106481653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481653, CurrentID=0, Status='live', Symbol='RNU2-60P', Aliases='', description='RNA, U2 small nuclear 60, pseudogene', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481654, CurrentID=0, Status='live', Symbol='RNU2-64P', Aliases='', description='RNA, U2 small nuclear 64, pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51191, CurrentID=0, Status='live', Symbol='HERC5', Aliases='CEB1, CEBP1', description='HECT and RLD domain containing E3 ubiquitin protein ligase 5', other_designations='HECT domain and RCC1-like domain-containing protein 5|cyclin-E-binding protein 1|hect domain and RLD 5|probable E3 ubiquitin-protein ligase HERC5', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88457117, end_position_on_the_genomic_accession=88506170, orientation='plus', exon_count=24, OMIM=608242, no_hdr0=''),
  51192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51192, CurrentID=0, Status='live', Symbol='CKLF', Aliases='C321, CKLF2, CKLF3, CKLF4, HSPC224, UCK-1, CKLF', description='chemokine-like factor', other_designations='chemokine-like factor 1|chemokine-like factor 2|chemokine-like factor 3|chemokine-like factor 4|transmembrane proteolipid', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66552563, end_position_on_the_genomic_accession=66566287, orientation='plus', exon_count=4, OMIM=616074, no_hdr0=''),
  51193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51193, CurrentID=0, Status='live', Symbol='ZNF639', Aliases='ANC-2H01, ANC_2H01, ZASC1', description='zinc finger protein 639', other_designations='zinc finger amplified in esophageal squamous cell carcinomas 1', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179322991, end_position_on_the_genomic_accession=179335535, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  51194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51194, CurrentID=0, Status='live', Symbol='IPO11', Aliases='RanBP11', description='importin 11', other_designations='Ran binding protein 11|imp11|ran-binding protein 11', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62412746, end_position_on_the_genomic_accession=62628589, orientation='plus', exon_count=31, OMIM=610889, no_hdr0=''),
  51195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51195, CurrentID=0, Status='live', Symbol='RAPGEFL1', Aliases='Link-GEFII', description='Rap guanine nucleotide exchange factor like 1', other_designations='Link guanine nucleotide exchange factor II|RAP guanine-nucleotide-exchange factor (GEF)-like 1|Rap guanine nucleotide exchange factor (GEF)-like 1|link GEFII', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40177010, end_position_on_the_genomic_accession=40195656, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  51196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51196, CurrentID=0, Status='live', Symbol='PLCE1', Aliases='NPHS3, PLCE, PPLC', description='phospholipase C epsilon 1', other_designations='PLC-epsilon-1|pancreas-enriched phospholipase C|phosphoinositide phospholipase C-epsilon-1|phosphoinositide-specific phospholipase C epsilon-1|phospholipase C, epsilon 1', map_location='10q23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93993989, end_position_on_the_genomic_accession=94328392, orientation='plus', exon_count=39, OMIM=608414, no_hdr0=''),
  106481661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481661, CurrentID=0, Status='live', Symbol='HNRNPDP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein D pseudogene 2', other_designations='heterogeneous nuclear ribonucleoprotein D (AU-rich element RNA binding protein 1, 37kDa) pseudogene 2', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51198, CurrentID=0, Status='live', Symbol='CDKN2A-AS1', Aliases='C9orf53, bA149I2.3', description='CDKN2A antisense RNA 1 (head to head)', other_designations='susceptibility protein NSG-x', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21967139, end_position_on_the_genomic_accession=21967754, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  51199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51199, CurrentID=0, Status='live', Symbol='NIN', Aliases='SCKL7', description='ninein', other_designations='glycogen synthase kinase 3 beta-interacting protein|hNinein|ninein (GSK3B interacting protein)|ninein centrosomal protein', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50719763, end_position_on_the_genomic_accession=50831165, orientation='minus', exon_count=35, OMIM=608684, no_hdr0=''),
  51200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51200, CurrentID=0, Status='live', Symbol='CPA4', Aliases='CPA3', description='carboxypeptidase A4', other_designations='carboxypeptidase A3', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130293134, end_position_on_the_genomic_accession=130324180, orientation='plus', exon_count=11, OMIM=607635, no_hdr0=''),
  51201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51201, CurrentID=0, Status='live', Symbol='ZDHHC2', Aliases='DHHC2, ZNF372', description='zinc finger, DHHC-type containing 2', other_designations='DHHC-2|ream|rec|reduced expression associated with metastasis protein|reduced expression in cancer protein|zinc finger DHHC domain-containing protein 2|zinc finger protein 372|zinc finger, DHHC domain containing 2', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17156020, end_position_on_the_genomic_accession=17222732, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  51202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51202, CurrentID=0, Status='live', Symbol='DDX47', Aliases='E4-DBP, HQ0256, MSTP162, RRP3', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 47', other_designations='DEAD box polypeptide 47|DEAD box protein 47|E4-DEAD box protein', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12813346, end_position_on_the_genomic_accession=12829981, orientation='plus', exon_count=12, OMIM=615428, no_hdr0=''),
  51203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51203, CurrentID=0, Status='live', Symbol='NUSAP1', Aliases='ANKT, BM037, LNP, NUSAP, PRO0310p1, Q0310, SAPL', description='nucleolar and spindle associated protein 1', other_designations='nucleolar protein ANKT', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41332694, end_position_on_the_genomic_accession=41381050, orientation='plus', exon_count=11, OMIM=612818, no_hdr0=''),
  51204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51204, CurrentID=0, Status='live', Symbol='TACO1', Aliases='CCDC44', description='translational activator of mitochondrially encoded cytochrome c oxidase I', other_designations='clone HQ0477 PRO0477p|coiled-coil domain-containing protein 44', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63600872, end_position_on_the_genomic_accession=63608365, orientation='plus', exon_count=5, OMIM=612958, no_hdr0=''),
  51205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51205, CurrentID=0, Status='live', Symbol='ACP6', Aliases='ACPL1, LPAP, PACPL1', description='acid phosphatase 6, lysophosphatidic', other_designations='acid phosphatase like 1|acid phosphatase-like protein 1|lysophosphatidic acid phosphatase 6', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147629286, end_position_on_the_genomic_accession=147670512, orientation='minus', exon_count=12, OMIM=611471, no_hdr0=''),
  51206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51206, CurrentID=0, Status='live', Symbol='GP6', Aliases='BDPLT11, GPIV, GPVI', description='glycoprotein VI (platelet)', other_designations='glycoprotein 6|platelet collagen receptor', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55013705, end_position_on_the_genomic_accession=55038264, orientation='minus', exon_count=8, OMIM=605546, no_hdr0=''),
  51207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51207, CurrentID=0, Status='live', Symbol='DUSP13', Aliases='BEDPA, DUSP13B, MDSP, SKRP4, TMDP, DUSP13', description='dual specificity phosphatase 13', other_designations='branching-enzyme interacting DSP|branching-enzyme interacting dual-specificity protein phosphatase|muscle-restricted DSP|testis- and skeletal-muscle-specific DSP', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75094432, end_position_on_the_genomic_accession=75109221, orientation='minus', exon_count=9, OMIM=613191, no_hdr0=''),
  51208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51208, CurrentID=0, Status='live', Symbol='CLDN18', Aliases='SFTA5, SFTPJ', description='claudin 18', other_designations='surfactant associated 5|surfactant associated protein J|surfactant, pulmonary associated protein J', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137998816, end_position_on_the_genomic_accession=138033652, orientation='plus', exon_count=6, OMIM=609210, no_hdr0=''),
  51209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51209, CurrentID=0, Status='live', Symbol='RAB9B', Aliases='RAB9L, Rab-9L', description='RAB9B, member RAS oncogene family', other_designations='RAB9-like protein|rab-9-like protein', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103822325, end_position_on_the_genomic_accession=103832282, orientation='minus', exon_count=3, OMIM=300285, no_hdr0=''),
  106481674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481674, CurrentID=0, Status='live', Symbol='HSPE1P14', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 14', other_designations='heat shock 10kDa protein 1 pseudogene 14', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481675, CurrentID=0, Status='live', Symbol='HSPE1P19', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 19', other_designations='heat shock 10kDa protein 1 pseudogene 19', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481676, CurrentID=0, Status='live', Symbol='HSPE1P23', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 23', other_designations='heat shock 10kDa protein 1 pseudogene 23', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51213, CurrentID=0, Status='live', Symbol='LUZP4', Aliases='CT-28, CT-8, CT28, HOM-TES-85', description='leucine zipper protein 4', other_designations='HOM-TES-85 tumor antigen|cancer/testis antigen 28|tumor antigen HOM-TES-85', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115289717, end_position_on_the_genomic_accession=115307556, orientation='plus', exon_count=4, OMIM=300616, no_hdr0=''),
  51214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51214, CurrentID=0, Status='live', Symbol='IGF2-AS', Aliases='IGF2-AS1, IGF2AS, PEG8', description='IGF2 antisense RNA', other_designations='IGF2 antisense RNA 1 (non-protein coding)|PEG8/IGF2AS, imprinting|insulin-like growth factor 2 antisense 1 (non-protein coding)|insulin-like growth factor II, antisense', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2140528, end_position_on_the_genomic_accession=2148666, orientation='plus', exon_count=2, OMIM=610146, no_hdr0=''),
  106481679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481679, CurrentID=0, Status='live', Symbol='SUMO2P18', Aliases='', description='SUMO2 pseudogene 18', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481680, CurrentID=0, Status='live', Symbol='COX6CP8', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 8', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481681, CurrentID=0, Status='live', Symbol='COX6CP10', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 10', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51218, CurrentID=0, Status='live', Symbol='GLRX5', Aliases='C14orf87, FLB4739, GRX5, PR01238, PRO1238, PRSA', description='glutaredoxin 5', other_designations='glutaredoxin 5 homolog|monothiol glutaredoxin-5', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95534986, end_position_on_the_genomic_accession=95544718, orientation='plus', exon_count=2, OMIM=609588, no_hdr0=''),
  106481683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481683, CurrentID=0, Status='live', Symbol='COX6CP14', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 14', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481684, CurrentID=0, Status='live', Symbol='COX6CP18', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 18', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  23528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23528, CurrentID=0, Status='live', Symbol='ZNF281', Aliases='ZBP-99, ZNP-99', description='zinc finger protein 281', other_designations='GC-box-binding zinc finger protein 1|ZNP-99 transcription factor|transcription factor ZBP-99|zinc finger DNA-binding protein 99', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200404947, end_position_on_the_genomic_accession=200410058, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  51222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51222, CurrentID=0, Status='live', Symbol='ZNF219', Aliases='ZFP219', description='zinc finger protein 219', other_designations='', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21090046, end_position_on_the_genomic_accession=21104704, orientation='minus', exon_count=10, OMIM=605036, no_hdr0=''),
  106481687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481687, CurrentID=0, Status='live', Symbol='PDIA3P2', Aliases='', description='protein disulfide isomerase family A member 3 pseudogene 2', other_designations='protein disulfide isomerase family A, member 3 pseudogene 2', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51224, CurrentID=0, Status='live', Symbol='TCEB3B', Aliases='ELOA2, HsT832, TCEB3L', description='transcription elongation factor B polypeptide 3B (elongin A2)', other_designations='elongin-A2|transcription elongation factor (SIII) elongin A2 elongin A2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47032572, end_position_on_the_genomic_accession=47035617, orientation='minus', exon_count=1, OMIM=609522, no_hdr0=''),
  51225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51225, CurrentID=0, Status='live', Symbol='ABI3', Aliases='NESH, SSH3BP3', description='ABI family member 3', other_designations='ABI family, member 3|new molecule including SH3', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49210227, end_position_on_the_genomic_accession=49223225, orientation='plus', exon_count=8, OMIM=606363, no_hdr0=''),
  51226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51226, CurrentID=0, Status='live', Symbol='COPZ2', Aliases='zeta2-COP', description='coatomer protein complex subunit zeta 2', other_designations='coatomer protein complex, subunit zeta 2|nonclathrin coat protein zeta-COP|zeta-2 COP|zeta-2-coat protein', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48026167, end_position_on_the_genomic_accession=48048091, orientation='minus', exon_count=11, OMIM=615526, no_hdr0=''),
  51227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51227, CurrentID=0, Status='live', Symbol='PIGP', Aliases='DCRC, DCRC-S, DSCR5, DSRC', description='phosphatidylinositol glycan anchor biosynthesis class P', other_designations='Down syndrome critical region gene 5|Down syndrome critical region protein 5|Down syndrome critical region protein C|PIG-P|phosphatidylinositol glycan anchor biosynthesis, class P|phosphatidylinositol glycan, class P|phosphatidylinositol-glycan biosynthesis class P protein|phosphatidylinositol-n-acetylglucosaminyltranferase subunit', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37065364, end_position_on_the_genomic_accession=37073158, orientation='minus', exon_count=7, OMIM=605938, no_hdr0=''),
  51228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51228, CurrentID=0, Status='live', Symbol='GLTP', Aliases='', description='glycolipid transfer protein', other_designations='truncated glycolipid transfer protein', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109850943, end_position_on_the_genomic_accession=109880488, orientation='minus', exon_count=6, OMIM=608949, no_hdr0=''),
  106481693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481693, CurrentID=0, Status='live', Symbol='BNIP3P18', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 18', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51230, CurrentID=0, Status='live', Symbol='PHF20', Aliases='C20orf104, GLEA2, HCA58, NZF, TDRD20A, TZP', description='PHD finger protein 20', other_designations='glioma-expressed antigen 2|hepatocellular carcinoma-associated antigen 58|novel zinc finger protein|transcription factor TZP|tudor domain containing 20A', map_location='20q11.22-q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35772001, end_position_on_the_genomic_accession=35950377, orientation='plus', exon_count=21, OMIM=610335, no_hdr0=''),
  51231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51231, CurrentID=0, Status='live', Symbol='VRK3', Aliases='', description='vaccinia related kinase 3', other_designations='serine/threonine-protein kinase VRK3|serine/threonine-protein pseudokinase VRK3|vaccinia-related kinase 3', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49976467, end_position_on_the_genomic_accession=50026017, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  51232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51232, CurrentID=0, Status='live', Symbol='CRIM1', Aliases='CRIM-1, S52', description='cysteine rich transmembrane BMP regulator 1 (chordin-like)', other_designations='cysteine-rich repeat-containing protein S52', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36355771, end_position_on_the_genomic_accession=36551135, orientation='plus', exon_count=20, OMIM=606189, no_hdr0=''),
  51233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51233, CurrentID=0, Status='live', Symbol='DRICH1', Aliases='C22orf43, KB-208E9.1', description='aspartate-rich 1', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23580872, end_position_on_the_genomic_accession=23637723, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  51234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51234, CurrentID=0, Status='live', Symbol='EMC4', Aliases='PIG17, TMEM85', description='ER membrane protein complex subunit 4', other_designations='cell proliferation-inducing gene 17 protein|transmembrane protein 85', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34224997, end_position_on_the_genomic_accession=34230165, orientation='plus', exon_count=5, OMIM=616245, no_hdr0=''),
  106481699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481699, CurrentID=0, Status='live', Symbol='LARP7P2', Aliases='', description='La ribonucleoprotein domain family member 7 pseudogene 2', other_designations='La ribonucleoprotein domain family, member 7 pseudogene 2', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51236, CurrentID=0, Status='live', Symbol='HGH1', Aliases='BRP16, BRP16L, C8orf30A, C8orf30B, FAM203A, FAM203B', description='HGH1 homolog', other_designations='brain protein 16|brain protein 16-like|family with sequence similarity 203, member A|family with sequence similarity 203, member B|protein FAM203A|protein FAM203B', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144137769, end_position_on_the_genomic_accession=144140843, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  51237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51237, CurrentID=0, Status='live', Symbol='MZB1', Aliases='MEDA-7, PACAP, pERp1', description='marginal zone B and B1 cell-specific protein', other_designations='HSPC190|caspase-2 binding protein|mesenteric estrogen-dependent adipose 7|mesenteric oestrogen-dependent adipose gene- 7|plasma cell-induced ER protein 1|plasma cell-induced resident ER protein|plasma cell-induced resident endoplasmic reticulum protein|proapoptotic caspase adapter protein|proapoptotic caspase adaptor protein', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139387568, end_position_on_the_genomic_accession=139389916, orientation='minus', exon_count=4, OMIM=609447, no_hdr0=''),
  106481702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481702, CurrentID=0, Status='live', Symbol='SNRPCP17', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 17', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51239, CurrentID=0, Status='live', Symbol='ANKRD39', Aliases='', description='ankyrin repeat domain 39', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96847985, end_position_on_the_genomic_accession=96858019, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481704, CurrentID=0, Status='live', Symbol='PAXBP1P1', Aliases='', description='PAX3 and PAX7 binding protein 1 pseudogene 1', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51241, CurrentID=0, Status='live', Symbol='COX16', Aliases='C14orf112, HSPC203', description='COX16 cytochrome c oxidase assembly homolog', other_designations='cytochrome c oxidase assembly factor', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70325081, end_position_on_the_genomic_accession=70359731, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481706, CurrentID=0, Status='live', Symbol='NUTF2P5', Aliases='', description='nuclear transport factor 2 pseudogene 5', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481707, CurrentID=0, Status='live', Symbol='CARSP2', Aliases='', description='cysteinyl-tRNA synthetase pseudogene 2', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51244, CurrentID=0, Status='live', Symbol='CCDC174', Aliases='C3orf19', description='coiled-coil domain containing 174', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14651746, end_position_on_the_genomic_accession=14672659, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106481709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481709, CurrentID=0, Status='live', Symbol='DDX43P3', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 43 pseudogene 3', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51246, CurrentID=0, Status='live', Symbol='SHISA5', Aliases='SCOTIN', description='shisa family member 5', other_designations='putative NF-kappa-B-activating protein 120|shisa homolog 5', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48467798, end_position_on_the_genomic_accession=48504826, orientation='minus', exon_count=10, OMIM=607290, no_hdr0=''),
  51247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51247, CurrentID=0, Status='live', Symbol='PAIP2', Aliases='PAIP-2A, PAIP2', description='poly(A) binding protein interacting protein 2', other_designations='PABC1-interacting protein 2|PABP-interacting protein 2|polyA-binding protein-interacting protein 2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139341832, end_position_on_the_genomic_accession=139369720, orientation='plus', exon_count=6, OMIM=605604, no_hdr0=''),
  51248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51248, CurrentID=0, Status='live', Symbol='PDZD11', Aliases='AIPP1, PDZK11, PISP', description='PDZ domain containing 11', other_designations='ATPase-interacting PDZ protein|PMCA-interacting single-PDZ protein|plasma membrane calcium ATPase-interacting single-PDZ protein', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70286361, end_position_on_the_genomic_accession=70289948, orientation='minus', exon_count=8, OMIM=300632, no_hdr0=''),
  51249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51249, CurrentID=0, Status='live', Symbol='TMEM69', Aliases='C1orf154', description='transmembrane protein 69', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45688175, end_position_on_the_genomic_accession=45694436, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  51250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51250, CurrentID=0, Status='live', Symbol='C6orf203', Aliases='HSPC230, PRED31', description='chromosome 6 open reading frame 203', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107028172, end_position_on_the_genomic_accession=107051586, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  51251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51251, CurrentID=0, Status='live', Symbol='NT5C3A', Aliases="NT5C3, P5'N-1, P5N-1, PN-I, POMP, PSN1, UMPH, UMPH1, cN-III, hUMP1, p36", description="5'-nucleotidase, cytosolic IIIA", other_designations="5'-nucleotidase, cytosolic III|7-methylguanosine nucleotidase|7-methylguanosine phosphate-specific 5'-nucleotidase|cytosolic 5'-nucleotidase 3|lupin|pyrimidine 5'-nucleotidase 1|uridine 5'-monophosphate hydrolase 1", map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33014113, end_position_on_the_genomic_accession=33062797, orientation='minus', exon_count=12, OMIM=606224, no_hdr0=''),
  51252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51252, CurrentID=0, Status='live', Symbol='FAM178B', Aliases='', description='family with sequence similarity 178 member B', other_designations='family with sequence similarity 178, member B', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96875882, end_position_on_the_genomic_accession=96986601, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  51253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51253, CurrentID=0, Status='live', Symbol='MRPL37', Aliases='L37mt, MRP-L2, MRP-L37, MRPL2, RPML2', description='mitochondrial ribosomal protein L37', other_designations='39S ribosomal protein L2, mitochondrial|L2mt|ribosomal protein, mitochondrial, L2', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54200167, end_position_on_the_genomic_accession=54226682, orientation='plus', exon_count=8, OMIM=611843, no_hdr0=''),
  106481718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481718, CurrentID=0, Status='live', Symbol='CFTRP3', Aliases='', description='cystic fibrosis transmembrane conductance regulator pseudogene 3', other_designations='', map_location='20q11.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51255, CurrentID=0, Status='live', Symbol='RNF181', Aliases='HSPC238', description='ring finger protein 181', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85595268, end_position_on_the_genomic_accession=85597708, orientation='plus', exon_count=5, OMIM=612490, no_hdr0=''),
  51256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51256, CurrentID=0, Status='live', Symbol='TBC1D7', Aliases='MGCPH, PIG51, TBC7', description='TBC1 domain family member 7', other_designations='TBC1 domain family, member 7|cell migration-inducing protein 23', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13304951, end_position_on_the_genomic_accession=13328583, orientation='minus', exon_count=9, OMIM=612655, no_hdr0=''),
  51257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51257, CurrentID=0, Status='live', Symbol='MARCH2', Aliases='HSPC240, MARCH-II, RNF172', description='membrane associated ring-CH-type finger 2', other_designations='RING finger protein 172|membrane associated ring finger 2|membrane-associated RING finger protein 2|membrane-associated RING-CH protein II|membrane-associated ring finger (C3HC4) 2, E3 ubiquitin protein ligase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8413272, end_position_on_the_genomic_accession=8439017, orientation='plus', exon_count=6, OMIM=613332, no_hdr0=''),
  51258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51258, CurrentID=0, Status='live', Symbol='MRPL51', Aliases='CDA09, HSPC241, MRP64, bMRP64', description='mitochondrial ribosomal protein L51', other_designations='L51mt|MRP-L51|bMRP-64|mitochondrial ribosomal protein 64|mitochondrial ribosomal protein bMRP64', map_location='12p13.3-p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6492150, end_position_on_the_genomic_accession=6493305, orientation='minus', exon_count=3, OMIM=611855, no_hdr0=''),
  51259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51259, CurrentID=0, Status='live', Symbol='TMEM216', Aliases='HSPC244', description='transmembrane protein 216', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61391687, end_position_on_the_genomic_accession=61398863, orientation='plus', exon_count=6, OMIM=613277, no_hdr0=''),
  51260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51260, CurrentID=0, Status='live', Symbol='PBDC1', Aliases='CXorf26', description='polysaccharide biosynthesis domain containing 1', other_designations='2610029G23Rik|UPF0368 protein Cxorf26|polysaccharide biosynthesis domain-containing protein 1', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76172929, end_position_on_the_genomic_accession=76178310, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106481725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481725, CurrentID=0, Status='live', Symbol='BSNDP3', Aliases='', description='barttin CLCNK-type chloride channel accessory beta subunit pseudogene 3', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481726, CurrentID=0, Status='live', Symbol='BSNDP4', Aliases='', description='barttin CLCNK-type chloride channel accessory beta subunit pseudogene 4', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51263, CurrentID=0, Status='live', Symbol='MRPL30', Aliases='L28MT, L30MT, MRP-L28, MRP-L30, MRPL28, MRPL28M, RPML28', description='mitochondrial ribosomal protein L30', other_designations='39S ribosomal protein L28, mitochondrial', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99181079, end_position_on_the_genomic_accession=99199557, orientation='plus', exon_count=6, OMIM=611838, no_hdr0=''),
  51264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51264, CurrentID=0, Status='live', Symbol='MRPL27', Aliases='L27mt', description='mitochondrial ribosomal protein L27', other_designations='MRP-L27', map_location='17q21.3-q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50367867, end_position_on_the_genomic_accession=50373201, orientation='minus', exon_count=4, OMIM=611837, no_hdr0=''),
  51265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51265, CurrentID=0, Status='live', Symbol='CDKL3', Aliases='NKIAMRE', description='cyclin-dependent kinase-like 3', other_designations='serine-threonine protein kinase NKIAMRE', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134236120, end_position_on_the_genomic_accession=134371469, orientation='minus', exon_count=23, OMIM=608459, no_hdr0=''),
  51266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51266, CurrentID=0, Status='live', Symbol='CLEC1B', Aliases='1810061I13Rik, CLEC2, CLEC2B, PRO1384, QDED721', description='C-type lectin domain family 1 member B', other_designations='C-type lectin domain family 1, member B|C-type lectin-like receptor 2|C-type lectin-like receptor-2|CLEC-2', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9985697, end_position_on_the_genomic_accession=10001865, orientation='minus', exon_count=9, OMIM=606783, no_hdr0=''),
  51267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51267, CurrentID=0, Status='live', Symbol='CLEC1A', Aliases='CLEC-1, CLEC1', description='C-type lectin domain family 1, member A', other_designations='C-type lectin-like receptor-1', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10069504, end_position_on_the_genomic_accession=10099065, orientation='minus', exon_count=8, OMIM=606782, no_hdr0=''),
  51268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51268, CurrentID=0, Status='live', Symbol='PIPOX', Aliases='LPIPOX', description='pipecolic acid oxidase', other_designations='L-pipecolate oxidase|L-pipecolic acid oxidase|PSO', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29042900, end_position_on_the_genomic_accession=29057218, orientation='plus', exon_count=8, OMIM=616713, no_hdr0=''),
  106481733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481733, CurrentID=0, Status='live', Symbol='SEPT14P13', Aliases='', description='septin 14 pseudogene 13', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51270, CurrentID=0, Status='live', Symbol='TFDP3', Aliases='CT30, DP4, HCA661', description='transcription factor Dp family member 3', other_designations='E2F-like protein|cancer/testis antigen 30|hepatocellular carcinoma-associated antigen 661|transcription factor Dp family, member 3', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133216669, end_position_on_the_genomic_accession=133218348, orientation='minus', exon_count=1, OMIM=300772, no_hdr0=''),
  51271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51271, CurrentID=0, Status='live', Symbol='UBAP1', Aliases='NAG20, UAP, UBAP, UBAP-1', description='ubiquitin associated protein 1', other_designations='nasopharyngeal carcinoma-associated gene 20 protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34179005, end_position_on_the_genomic_accession=34252523, orientation='plus', exon_count=12, OMIM=609787, no_hdr0=''),
  51272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51272, CurrentID=0, Status='live', Symbol='BET1L', Aliases='BET1L1, GOLIM3, GS15, HSPC197', description='Bet1 golgi vesicular membrane trafficking protein like', other_designations='Bet1 golgi vesicular membrane trafficking protein-like|GOS-15|blocked early in transport 1 homolog-like|golgi SNARE 15 kDa protein|golgi SNARE with a size of 15 kDa|golgi integral membrane protein 3|vesicle transport protein GOS15', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=202924, end_position_on_the_genomic_accession=207422, orientation='minus', exon_count=4, OMIM=615417, no_hdr0=''),
  106481737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481737, CurrentID=0, Status='live', Symbol='AGGF1P6', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 6', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 6', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51274, CurrentID=0, Status='live', Symbol='KLF3', Aliases='BKLF', description='Kruppel-like factor 3 (basic)', other_designations='CACCC-box-binding protein BKLF|TEF-2|basic Kruppel-like factor|basic krueppel-like factor|basic kruppel like factor|transcript ch138', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38664169, end_position_on_the_genomic_accession=38701508, orientation='plus', exon_count=8, OMIM=609392, no_hdr0=''),
  51275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51275, CurrentID=0, Status='live', Symbol='MAPKAPK5-AS1', Aliases='C12orf47', description='MAPKAPK5 antisense RNA 1', other_designations='MAPKAPK5 antisense RNA 1 (non-protein coding)|apoptosis-related protein PNAS-1', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111839769, end_position_on_the_genomic_accession=111842902, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  51276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51276, CurrentID=0, Status='live', Symbol='ZNF571', Aliases='HSPC059', description='zinc finger protein 571', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37564253, end_position_on_the_genomic_accession=37594792, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  51277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51277, CurrentID=0, Status='live', Symbol='DNAJC27', Aliases='RBJ, RabJS', description='DnaJ heat shock protein family (Hsp40) member C27', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 27|Ras-associated protein Rap1|rab and DnaJ domain-containing protein', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24943636, end_position_on_the_genomic_accession=24972094, orientation='minus', exon_count=9, OMIM=613527, no_hdr0=''),
  106481742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481742, CurrentID=0, Status='live', Symbol='FRGCA', Aliases='', description='FOXM1-regulated, gastric cancer associated', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51279, CurrentID=0, Status='live', Symbol='C1RL', Aliases='C1RL1P, C1r-LP, CLSPa, C1RL', description='complement component 1, r subcomponent-like', other_designations='C1r-like protein|C1r-like serine protease analog protein|complement C1r-like proteinase', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7093668, end_position_on_the_genomic_accession=7109278, orientation='minus', exon_count=9, OMIM=608974, no_hdr0=''),
  51280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51280, CurrentID=0, Status='live', Symbol='GOLM1', Aliases='C9orf155, GOLPH2, GP73, HEL46, PSEC0257, bA379P1.3', description='golgi membrane protein 1', other_designations='epididymis luminal protein 46|golgi membrane protein GP73|golgi phosphoprotein 2|golgi protein, 73-kD', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86026141, end_position_on_the_genomic_accession=86100201, orientation='minus', exon_count=11, OMIM=606804, no_hdr0=''),
  51281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51281, CurrentID=0, Status='live', Symbol='ANKMY1', Aliases='ZMYND13', description='ankyrin repeat and MYND domain containing 1', other_designations='testis-specific ankyrin-like protein 1|zinc finger MYND domain-containing protein 13', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240479422, end_position_on_the_genomic_accession=240561068, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  51282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51282, CurrentID=0, Status='live', Symbol='SCAND1', Aliases='RAZ1, SDP1', description='SCAN domain containing 1', other_designations='SCAN domain-containing 1|SCAN-related protein RAZ1', map_location='20q11.1-q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35953617, end_position_on_the_genomic_accession=35955359, orientation='minus', exon_count=3, OMIM=610416, no_hdr0=''),
  51283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51283, CurrentID=0, Status='live', Symbol='BFAR', Aliases='BAR, RNF47', description='bifunctional apoptosis regulator', other_designations='RING finger protein 47|bifunctional apoptosis inhibitor', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14632811, end_position_on_the_genomic_accession=14669236, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51284, CurrentID=0, Status='live', Symbol='TLR7', Aliases='TLR7-like', description='toll-like receptor 7', other_designations='toll-like receptor 7-like', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12867083, end_position_on_the_genomic_accession=12890361, orientation='plus', exon_count=3, OMIM=300365, no_hdr0=''),
  51285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51285, CurrentID=0, Status='live', Symbol='RASL12', Aliases='RIS', description='RAS-like family 12', other_designations='RAS-like, family 12|Ras family member Ris|ras-like protein Ris', map_location='15q11.2-q22.33', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65045582, end_position_on_the_genomic_accession=65076670, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  51286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51286, CurrentID=0, Status='live', Symbol='CEND1', Aliases='BM88', description='cell cycle exit and neuronal differentiation 1', other_designations='BM88 antigen', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=787110, end_position_on_the_genomic_accession=790126, orientation='minus', exon_count=2, OMIM=608213, no_hdr0=''),
  51287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51287, CurrentID=0, Status='live', Symbol='COA4', Aliases='CHCHD8, CMC3, E2IG2', description='cytochrome c oxidase assembly factor 4 homolog', other_designations='E2-induced gene 2 protein|coiled-coil-helix-coiled-coil-helix domain containing 8|coiled-coil-helix-coiled-coil-helix domain-containing protein 8', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73872667, end_position_on_the_genomic_accession=73876845, orientation='minus', exon_count=2, OMIM=608016, no_hdr0=''),
  106481752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481752, CurrentID=0, Status='live', Symbol='RN7SL78P', Aliases='', description='RNA, 7SL, cytoplasmic 78, pseudogene', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51289, CurrentID=0, Status='live', Symbol='RXFP3', Aliases='GPCR135, RLN3R1, RXFPR3, SALPR', description='relaxin/insulin-like family peptide receptor 3', other_designations='G-protein coupled receptor GPCR135|G-protein coupled receptor SALPR|RLN3 receptor 1|g protein-coupled receptor SALPR|relaxin 3 receptor 1|relaxin family peptide receptor 3|somatostatin and angiotensin-like peptide receptor|somatostatin- and angiotensin-like peptide receptor', map_location='5p15.1-p14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33936386, end_position_on_the_genomic_accession=33938918, orientation='plus', exon_count=1, OMIM=609445, no_hdr0=''),
  51290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51290, CurrentID=0, Status='live', Symbol='ERGIC2', Aliases='CDA14, Erv41, PTX1, cd002', description='ERGIC and golgi 2', other_designations='CD14 protein', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29340646, end_position_on_the_genomic_accession=29381210, orientation='minus', exon_count=14, OMIM=612236, no_hdr0=''),
  51291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51291, CurrentID=0, Status='live', Symbol='GMIP', Aliases='ARHGAP46', description='GEM interacting protein', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19629473, end_position_on_the_genomic_accession=19643667, orientation='minus', exon_count=21, OMIM=609694, no_hdr0=''),
  51292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51292, CurrentID=0, Status='live', Symbol='GMPR2', Aliases='GMPR 2', description='guanosine monophosphate reductase 2', other_designations="guanosine 5'-monophosphate oxidoreductase 2|guanosine monophosphate reductase isolog", map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24232422, end_position_on_the_genomic_accession=24239242, orientation='plus', exon_count=12, OMIM=610781, no_hdr0=''),
  51293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51293, CurrentID=0, Status='live', Symbol='CD320', Aliases='8D6, 8D6A, TCBLR', description='CD320 molecule', other_designations='8D6 antigen|FDC-SM-8D6|FDC-signaling molecule 8D6|transcobalamin receptor', map_location='19p13.3-p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8302127, end_position_on_the_genomic_accession=8308356, orientation='minus', exon_count=5, OMIM=606475, no_hdr0=''),
  51294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51294, CurrentID=0, Status='live', Symbol='PCDH12', Aliases='VE-cadherin-2, VECAD2', description='protocadherin 12', other_designations='VE-cad-2|VE-cadherin 2|vascular cadherin-2|vascular endothelial cadherin 2|vascular endothelial cadherin-2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141943585, end_position_on_the_genomic_accession=141959062, orientation='minus', exon_count=4, OMIM=605622, no_hdr0=''),
  51295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51295, CurrentID=0, Status='live', Symbol='ECSIT', Aliases='SITPEC', description='ECSIT signalling integrator', other_designations='ECSIT homolog|likely ortholog of mouse signaling intermediate in Toll pathway evolutionarily conserved', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11505916, end_position_on_the_genomic_accession=11529172, orientation='minus', exon_count=9, OMIM=608388, no_hdr0=''),
  51296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51296, CurrentID=0, Status='live', Symbol='SLC15A3', Aliases='OCTP, PHT2, PTR3', description='solute carrier family 15 (oligopeptide transporter), member 3', other_designations='osteoclast transporter|peptide transporter 3|peptide/histidine transporter 2|solute carrier family 15, member 3', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60937083, end_position_on_the_genomic_accession=60952526, orientation='minus', exon_count=9, OMIM=610408, no_hdr0=''),
  51297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51297, CurrentID=0, Status='live', Symbol='BPIFA1', Aliases='LUNX, NASG, PLUNC, SPLUNC1, SPURT, bA49G10.5', description='BPI fold containing family A member 1', other_designations='BPI fold containing family A, member 1|ligand-binding protein RYA3|lung-specific protein X|nasopharyngeal carcinoma-related protein|palate lung and nasal epithelium clone protein|palate, lung and nasal epithelium associated|protein Plunc|secretory protein in upper respiratory tracts|short PLUNC1|tracheal epithelium enriched protein|tracheal epithelium-enriched protein|von Ebner protein Hl', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33235996, end_position_on_the_genomic_accession=33243309, orientation='plus', exon_count=9, OMIM=607412, no_hdr0=''),
  51298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51298, CurrentID=0, Status='live', Symbol='THEG', Aliases='CT561, THEG', description='theg spermatid protein', other_designations='Theg homolog|cancer/testis antigen 56|testis-specific', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=362057, end_position_on_the_genomic_accession=376703, orientation='minus', exon_count=10, OMIM=609503, no_hdr0=''),
  51299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51299, CurrentID=0, Status='live', Symbol='NRN1', Aliases='NRN, dJ380B8.2', description='neuritin 1', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5997999, end_position_on_the_genomic_accession=6007605, orientation='minus', exon_count=7, OMIM=607409, no_hdr0=''),
  51300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51300, CurrentID=0, Status='live', Symbol='TIMMDC1', Aliases='C3orf1', description='translocase of inner mitochondrial membrane domain containing 1', other_designations='M5-14 protein|TIMM domain containing-protein 1|translocase of inner mitochondrial membrane domain-containing protein 1|transmembrane protein C3orf1', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119498521, end_position_on_the_genomic_accession=119524281, orientation='plus', exon_count=7, OMIM=615534, no_hdr0=''),
  51301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51301, CurrentID=0, Status='live', Symbol='GCNT4', Aliases='C2GNT3', description='glucosaminyl (N-acetyl) transferase 4, core 2', other_designations='core 2 beta-1,6-N-acetylglucosaminyltransferase 3|core 2-branching enzyme 3|core2-GlcNAc-transferase 3|glucosaminyl (N-acetyl) transferase 4, core 2 (beta-1,6-N-acetylglucosaminyltransferase)', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75027464, end_position_on_the_genomic_accession=75030899, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  51302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51302, CurrentID=0, Status='live', Symbol='CYP39A1', Aliases='', description='cytochrome P450 family 39 subfamily A member 1', other_designations='cytochrome P450, family 39, subfamily A, polypeptide 1|cytochrome P450, subfamily XXXIX (oxysterol 7 alpha-hydroxylase), polypeptide 1', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46549580, end_position_on_the_genomic_accession=46652847, orientation='minus', exon_count=14, OMIM=605994, no_hdr0=''),
  51303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51303, CurrentID=0, Status='live', Symbol='FKBP11', Aliases='FKBP19', description='FK506 binding protein 11', other_designations='19 kDa FK506-binding protein|19 kDa FKBP|FK506 binding protein 11, 19 kDa|FK506-binding protein 11|FKBP-11|FKBP-19|PPIase FKBP11|rotamase', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48921959, end_position_on_the_genomic_accession=48925547, orientation='minus', exon_count=8, OMIM=610571, no_hdr0=''),
  51304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51304, CurrentID=0, Status='live', Symbol='ZDHHC3', Aliases='DHHC-3, GODZ, ZNF373', description='zinc finger, DHHC-type containing 3', other_designations='DHHC1 protein|Golgi apparatus-specific protein with DHHC zinc finger domain|golgi-specific DHHC Zinc Finger Protein|zinc finger DHHC domain-containing protein 3|zinc finger protein 373|zinc finger, DHHC domain containing 3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44915261, end_position_on_the_genomic_accession=44976182, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  51305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51305, CurrentID=0, Status='live', Symbol='KCNK9', Aliases='K2p9.1, KT3.2, TASK-3, TASK3', description='potassium channel, two pore domain subfamily K, member 9', other_designations='TWIK-related acid-sensitive K(+) channel 3|acid-sensitive potassium channel protein TASK-3|two pore K(+) channel KT3.2|two pore potassium channel KT3.2', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=139600838, end_position_on_the_genomic_accession=139703068, orientation='minus', exon_count=5, OMIM=605874, no_hdr0=''),
  51306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51306, CurrentID=0, Status='live', Symbol='FAM13B', Aliases='ARHGAP49, C5orf51, KHCHP, N61, FAM13B', description='family with sequence similarity 13 member B', other_designations='GAP-like protein N61|family with sequence similarity 13, member B|family with sequence similarity 13, member B1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137937949, end_position_on_the_genomic_accession=138051961, orientation='minus', exon_count=29, OMIM=609371, no_hdr0=''),
  51307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51307, CurrentID=0, Status='live', Symbol='FAM53C', Aliases='C5orf6', description='family with sequence similarity 53 member C', other_designations='family with sequence similarity 53, member C|putative nuclear protein', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138337535, end_position_on_the_genomic_accession=138349729, orientation='plus', exon_count=6, OMIM=609372, no_hdr0=''),
  51308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51308, CurrentID=0, Status='live', Symbol='REEP2', Aliases='C5orf19, SGC32445, SPG72', description='receptor accessory protein 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138439001, end_position_on_the_genomic_accession=138446969, orientation='plus', exon_count=8, OMIM=609347, no_hdr0=''),
  51309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51309, CurrentID=0, Status='live', Symbol='ARMCX1', Aliases='ALEX1, GASP7', description='armadillo repeat containing, X-linked 1', other_designations='ARM protein lost in epithelial cancers on chromosome X 1|arm protein lost in epithelial cancers, X chromosome, 1', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101550531, end_position_on_the_genomic_accession=101554700, orientation='plus', exon_count=4, OMIM=300362, no_hdr0=''),
  51310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51310, CurrentID=0, Status='live', Symbol='SLC22A17', Aliases='24p3R, BOCT, BOIT, NGALR, NGALR2, NGALR3, hBOIT', description='solute carrier family 22 member 17', other_designations='24p3 receptor|NGAL receptor|brain-type organic cation transporter|lipocalin-2 receptor|neutrophil gelatinase-associated lipocalin receptor|potent brain type organic ion transporter|solute carrier family 22 (organic cation transporter), member 17|solute carrier family 22, member 17', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23346311, end_position_on_the_genomic_accession=23354976, orientation='minus', exon_count=10, OMIM=611461, no_hdr0=''),
  51311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51311, CurrentID=0, Status='live', Symbol='TLR8', Aliases='CD288', description='toll-like receptor 8', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12906620, end_position_on_the_genomic_accession=12923169, orientation='plus', exon_count=3, OMIM=300366, no_hdr0=''),
  51312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51312, CurrentID=0, Status='live', Symbol='SLC25A37', Aliases='HT015, MFRN, MFRN1, MSC, MSCP, PRO1278, PRO1584, PRO2217', description='solute carrier family 25 (mitochondrial iron transporter), member 37', other_designations='mitochondrial iron transporter 1|mitochondrial solute carrier protein|mitoferrin|predicted protein of HQ2217|solute carrier family 25, member 37', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23528850, end_position_on_the_genomic_accession=23572550, orientation='plus', exon_count=9, OMIM=610387, no_hdr0=''),
  51313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51313, CurrentID=0, Status='live', Symbol='FAM198B', Aliases='AD021, AD036, C4orf18', description='family with sequence similarity 198 member B', other_designations='expressed in nerve and epithelium during development|family with sequence similarity 198, member B', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158124580, end_position_on_the_genomic_accession=158173050, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51314, CurrentID=0, Status='live', Symbol='NME8', Aliases='CILD6, HEL-S-99, NM23-H8, SPTRX2, TXNDC3, sptrx-2', description='NME/NM23 family member 8', other_designations='epididymis secretory protein Li 99|sperm-specific thioredoxin 2|spermatid-specific thioredoxin-2|thioredoxin domain containing 3 (spermatozoa)', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37848597, end_position_on_the_genomic_accession=37900401, orientation='plus', exon_count=18, OMIM=607421, no_hdr0=''),
  51315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51315, CurrentID=0, Status='live', Symbol='KRCC1', Aliases='CHBP2', description='lysine-rich coiled-coil 1', other_designations='cryptogenic hepatitis-binding protein 2', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88027203, end_position_on_the_genomic_accession=88055801, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  51316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51316, CurrentID=0, Status='live', Symbol='PLAC8', Aliases='C15, DGIC, PNAS-144, onzin', description='placenta specific 8', other_designations='down-regulated in gastrointestinal cancer protein|placenta-specific 8', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83090048, end_position_on_the_genomic_accession=83114758, orientation='minus', exon_count=6, OMIM=607515, no_hdr0=''),
  51317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51317, CurrentID=0, Status='live', Symbol='PHF21A', Aliases='BHC80, BM-006', description='PHD finger protein 21A', other_designations='BHC80a|BRAF35-HDAC complex protein BHC80|BRAF35/HDAC2 complex (80 kDa)', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45929319, end_position_on_the_genomic_accession=46121434, orientation='minus', exon_count=25, OMIM=608325, no_hdr0=''),
  51318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51318, CurrentID=0, Status='live', Symbol='MRPL35', Aliases='L35mt, MRP-L35', description='mitochondrial ribosomal protein L35', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86199355, end_position_on_the_genomic_accession=86213794, orientation='plus', exon_count=5, OMIM=611841, no_hdr0=''),
  51319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51319, CurrentID=0, Status='live', Symbol='RSRC1', Aliases='BM-011, SFRS21, SRrp53', description='arginine/serine-rich coiled-coil 1', other_designations='arginine/serine-rich coiled-coil protein 1|splicing factor, arginine/serine-rich 21', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158110052, end_position_on_the_genomic_accession=158544835, orientation='plus', exon_count=11, OMIM=613352, no_hdr0=''),
  51320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51320, CurrentID=0, Status='live', Symbol='MEX3C', Aliases='BM-013, MEX-3C, RKHD2, RNF194', description='mex-3 RNA binding family member C', other_designations='MEX3C variant 1|MEX3C variant 2|RING finger and KH domain-containing protein 2|RING finger protein 194|RNA-binding protein MEX3C|ring finger and KH domain containing 2', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51174550, end_position_on_the_genomic_accession=51197681, orientation='minus', exon_count=2, OMIM=611005, no_hdr0=''),
  51321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51321, CurrentID=0, Status='live', Symbol='AMZ2', Aliases='', description='archaelysin family metallopeptidase 2', other_designations='archaemetzincins-2|archeobacterial metalloproteinase-like protein 2', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68205997, end_position_on_the_genomic_accession=68257164, orientation='plus', exon_count=11, OMIM=615169, no_hdr0=''),
  51322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51322, CurrentID=0, Status='live', Symbol='WAC', Aliases='BM-016, PRO1741, Wwp4', description='WW domain containing adaptor with coiled-coil', other_designations='', map_location='10p12.1-p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28532588, end_position_on_the_genomic_accession=28623112, orientation='plus', exon_count=16, OMIM=615049, no_hdr0=''),
  154467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154467, CurrentID=0, Status='live', Symbol='CCDC167', Aliases='C6orf129, HSPC265', description='coiled-coil domain containing 167', other_designations='transmembrane and coiled-coil domain-containing protein C6orf129', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37482920, end_position_on_the_genomic_accession=37499924, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  51324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51324, CurrentID=0, Status='live', Symbol='SPG21', Aliases='ACP33, BM-019, GL010, MAST', description='spastic paraplegia 21 (autosomal recessive, Mast syndrome)', other_designations='acid cluster protein 33', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64963021, end_position_on_the_genomic_accession=64989946, orientation='minus', exon_count=13, OMIM=608181, no_hdr0=''),
  106481789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481789, CurrentID=0, Status='live', Symbol='TRMT2B-AS1', Aliases='', description='TRMT2B antisense RNA 1', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51326, CurrentID=0, Status='live', Symbol='ARL17A', Aliases='ARF1P2, ARL17P1', description='ADP ribosylation factor like GTPase 17A', other_designations='ADP-ribosylation factor 1 pseudogene 2|ADP-ribosylation factor 7|ADP-ribosylation factor like GTPase 17A|ADP-ribosylation factor-like 17 pseudogene 1|ADP-ribosylation factor-like 17A', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46499818, end_position_on_the_genomic_accession=46579792, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  51327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51327, CurrentID=0, Status='live', Symbol='AHSP', Aliases='EDRF, ERAF', description='alpha hemoglobin stabilizing protein', other_designations='alpha hemoglobin stabilising protein|erythroid differentiation associated factor|erythroid differentiation-related factor', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31527852, end_position_on_the_genomic_accession=31528803, orientation='plus', exon_count=4, OMIM=605821, no_hdr0=''),
  106481792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481792, CurrentID=0, Status='live', Symbol='RNU6-77P', Aliases='RNU6-77', description='RNA, U6 small nuclear 77, pseudogene', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51329, CurrentID=0, Status='live', Symbol='ARL6IP4', Aliases='SFRS20, SR-25, SRp25, SRrp37', description='ADP ribosylation factor like GTPase 6 interacting protein 4', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 4|ADP-ribosylation factor like GTPase 6 interacting protein 4|ADP-ribosylation factor-like 6 interacting protein 4|ADP-ribosylation-like factor 6 interacting protein 4|ARL-6-interacting protein 4|HSP-975|HSVI binding protein|HSVI-binding protein|SR-15|SRp25 nuclear protein|aip-4|splicing factor SRrp37|splicing factor, arginine/serine-rich 20|splicing regulator SRrp38', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122980060, end_position_on_the_genomic_accession=122982913, orientation='plus', exon_count=6, OMIM=607668, no_hdr0=''),
  51330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51330, CurrentID=0, Status='live', Symbol='TNFRSF12A', Aliases='CD266, FN14, TWEAKR', description='tumor necrosis factor receptor superfamily member 12A', other_designations='FGF-inducible 14|fibroblast growth factor-inducible immediate-early response protein 14|tumor necrosis factor receptor superfamily, member 12A|tweak-receptor|type I transmembrane protein Fn14', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3020312, end_position_on_the_genomic_accession=3022383, orientation='plus', exon_count=4, OMIM=605914, no_hdr0=''),
  106481795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481795, CurrentID=0, Status='live', Symbol='HMBOX1-IT1', Aliases='', description='HMBOX1 intronic transcript 1', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51332, CurrentID=0, Status='live', Symbol='SPTBN5', Aliases='BSPECV, HUBSPECV, HUSPECV', description='spectrin, beta, non-erythrocytic 5', other_designations='beta V spectrin|beta-V spectrin|spectrin beta chain, brain 4|spectrin, non-erythroid beta chain 4', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41848146, end_position_on_the_genomic_accession=41894077, orientation='minus', exon_count=68, OMIM=605916, no_hdr0=''),
  51333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51333, CurrentID=0, Status='live', Symbol='ZNF771', Aliases='DSC43', description='zinc finger protein 771', other_designations='mesenchymal stem cell protein DSC43', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30407414, end_position_on_the_genomic_accession=30418595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  51334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51334, CurrentID=0, Status='live', Symbol='PRR16', Aliases='DSC54, LARGEN', description='proline rich 16', other_designations='mesenchymal stem cell protein DSC54|proline-rich protein 16', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=120464278, end_position_on_the_genomic_accession=120687332, orientation='plus', exon_count=10, OMIM=615931, no_hdr0=''),
  51335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51335, CurrentID=0, Status='live', Symbol='NGRN', Aliases='DSC92', description='neugrin, neurite outgrowth associated', other_designations='mesenchymal stem cell protein DSC92|neurite outgrowth associated protein|neurite outgrowth-associated protein|spinal cord-derived protein FI58G', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90265663, end_position_on_the_genomic_accession=90272211, orientation='plus', exon_count=2, OMIM=616718, no_hdr0=''),
  106481800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481800, CurrentID=0, Status='live', Symbol='RNU7-111P', Aliases='', description='RNA, U7 small nuclear 111 pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51337, CurrentID=0, Status='live', Symbol='THEM6', Aliases='C8orf55, DSCD75', description='thioesterase superfamily member 6', other_designations='UPF0670 protein C8orf55|UPF0670 protein THEM6|mesenchymal stem cell protein DSCD75', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142727203, end_position_on_the_genomic_accession=142736932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  51338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51338, CurrentID=0, Status='live', Symbol='MS4A4A', Aliases='4SPAN1, CD20-L1, CD20L1, HDCME31P, MS4A4, MS4A7', description='membrane-spanning 4-domains, subfamily A, member 4A', other_designations='CD20 antigen-like 1|Fc epsilon receptor beta subunit homolog|four-span transmembrane protein 1|membrane-spanning 4-domains, subfamily A, member 4', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60280541, end_position_on_the_genomic_accession=60308972, orientation='plus', exon_count=8, OMIM=606547, no_hdr0=''),
  51339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51339, CurrentID=0, Status='live', Symbol='DACT1', Aliases='DAPPER, DAPPER1, DPR1, FRODO, HDPR1, THYEX3', description='dishevelled-binding antagonist of beta-catenin 1', other_designations='dapper antagonist of catenin 1|dapper, antagonist of beta-catenin, homolog 1|hepatocellular carcinoma novel gene 3 protein|heptacellular carcinoma novel gene 3', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58634068, end_position_on_the_genomic_accession=58648321, orientation='plus', exon_count=6, OMIM=607861, no_hdr0=''),
  51340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51340, CurrentID=0, Status='live', Symbol='CRNKL1', Aliases='CLF, CRN, Clf1, HCRN, MSTP021, SYF3', description='crooked neck pre-mRNA splicing factor 1', other_designations='Crn, crooked neck-like 1|SYF3 pre-mRNA-splicing factor|crooked neck homolog|crooked neck pre-mRNA splicing factor-like 1|crooked neck-like 1', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20034361, end_position_on_the_genomic_accession=20056046, orientation='minus', exon_count=15, OMIM=610952, no_hdr0=''),
  51341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51341, CurrentID=0, Status='live', Symbol='ZBTB7A', Aliases='FBI-1, FBI1, LRF, TIP21, ZBTB7, ZNF857A, pokemon', description='zinc finger and BTB domain containing 7A', other_designations='HIV-1 1st-binding protein 1|HIV-1 inducer of short transcripts binding protein|POK erythroid myeloid ontogenic factor|POZ and Krueppel erythroid myeloid ontogenic factor|TTF-I-interacting peptide 21|factor binding IST protein 1|factor that binds to inducer of short transcripts protein 1|leukemia/lymphoma-related factor|lymphoma related factor|zinc finger and BTB domain containing 7A, HIV-1 inducer of short transcripts binding protein|zinc finger protein 857A', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4043303, end_position_on_the_genomic_accession=4066818, orientation='minus', exon_count=6, OMIM=605878, no_hdr0=''),
  106481806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481806, CurrentID=0, Status='live', Symbol='RN7SKP43', Aliases='', description='RNA, 7SK small nuclear pseudogene 43', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51343, CurrentID=0, Status='live', Symbol='FZR1', Aliases='CDC20C, CDH1, FZR, FZR2, HCDH, HCDH1', description='fizzy/cell division cycle 20 related 1', other_designations='CDC20 homolog 1|CDC20-like 1b|CDC20-like protein 1|cdh1/Hct1 homolog', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3506065, end_position_on_the_genomic_accession=3538539, orientation='plus', exon_count=16, OMIM=603619, no_hdr0=''),
  106481808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481808, CurrentID=0, Status='live', Symbol='RN7SKP76', Aliases='', description='RNA, 7SK small nuclear pseudogene 76', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481809, CurrentID=0, Status='live', Symbol='RN7SKP90', Aliases='', description='RNA, 7SK small nuclear pseudogene 90', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481810, CurrentID=0, Status='live', Symbol='RN7SKP102', Aliases='', description='RNA, 7SK small nuclear pseudogene 102', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51347, CurrentID=0, Status='live', Symbol='TAOK3', Aliases='DPK, JIK, MAP3K18', description='TAO kinase 3', other_designations='CTCL-associated antigen HD-CL-09|JNK/SAPK-inhibitory kinase|STE20-like kinase|cutaneous T-cell lymphoma-associated antigen HD-CL-09|dendritic cell-derived protein kinase|hKFC-A|jun kinase-inhibitory kinase|kinase from chicken homolog A|thousand and one amino acid protein 3', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118149801, end_position_on_the_genomic_accession=118372945, orientation='minus', exon_count=26, OMIM=616711, no_hdr0=''),
  51348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51348, CurrentID=0, Status='live', Symbol='KLRF1', Aliases='CLEC5C, NKp80', description='killer cell lectin-like receptor subfamily F, member 1', other_designations='C-type lectin domain family 5 member C|activating coreceptor NKp80|killer cell lectin-like receptor F1|lectin-like receptor F1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9827392, end_position_on_the_genomic_accession=9845005, orientation='plus', exon_count=6, OMIM=605029, no_hdr0=''),
  106481813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481813, CurrentID=0, Status='live', Symbol='RN7SKP153', Aliases='', description='RNA, 7SK small nuclear pseudogene 153', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51350, CurrentID=0, Status='live', Symbol='KRT76', Aliases='HUMCYT2A, KRT2B, KRT2P', description='keratin 76, type II', other_designations='CK-2P|K2P|K76|cytokeratin 2|cytokeratin-2P|keratin 2p|keratin-76|type-II keratin Kb9', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52768155, end_position_on_the_genomic_accession=52777345, orientation='minus', exon_count=9, OMIM=616671, no_hdr0=''),
  51351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51351, CurrentID=0, Status='live', Symbol='ZNF117', Aliases='H-plk, HPF9', description='zinc finger protein 117', other_designations='Krueppel-related zinc finger protein|provirus-linked krueppel|zinc finger protein HPF9', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64974452, end_position_on_the_genomic_accession=64991036, orientation='minus', exon_count=4, OMIM=194624, no_hdr0=''),
  51352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51352, CurrentID=0, Status='live', Symbol='WT1-AS', Aliases='WIT-1, WIT11, WT1AS, WT1-AS', description='WT1 antisense RNA', other_designations='WT1 antisense RNA (non-protein coding)|Wilms tumor associated protein|Wilms tumor upstream neighbor 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32435518, end_position_on_the_genomic_accession=32458769, orientation='plus', exon_count=8, OMIM=607899, no_hdr0=''),
  106481817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481817, CurrentID=0, Status='live', Symbol='RN7SKP208', Aliases='', description='RNA, 7SK small nuclear pseudogene 208', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481818, CurrentID=0, Status='live', Symbol='RN7SKP227', Aliases='', description='RNA, 7SK small nuclear pseudogene 227', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481819, CurrentID=0, Status='live', Symbol='RN7SKP243', Aliases='', description='RNA, 7SK small nuclear pseudogene 243', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481820, CurrentID=0, Status='live', Symbol='RN7SKP254', Aliases='', description='RNA, 7SK small nuclear pseudogene 254', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481821, CurrentID=0, Status='live', Symbol='RN7SKP273', Aliases='', description='RNA, 7SK small nuclear pseudogene 273', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481822, CurrentID=0, Status='live', Symbol='RN7SKP286', Aliases='', description='RNA, 7SK small nuclear pseudogene 286', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51359, CurrentID=0, Status='live', Symbol='ERVH-4', Aliases='HERV-HD1, HSU88895', description='endogenous retrovirus group H member 4', other_designations='clone c4.2|endogenous retrovirus H D1 leader region/integrase-derived ORF1, ORF2, and putative envelope protein|endogenous retrovirus group H, member 4', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51360, CurrentID=0, Status='live', Symbol='MBTPS2', Aliases='BRESEK, IFAP, KFSD, KFSDX, OLMSX, S2P', description='membrane bound transcription factor peptidase, site 2', other_designations='SREBPs intramembrane protease|endopeptidase S2P|keratosis follicularis spinulosa decalvans|membrane-bound transcription factor peptidase, site 2|membrane-bound transcription factor protease, site 2|sterol regulatory element-binding proteins intramembrane protease', map_location='Xp22.12-p22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21839538, end_position_on_the_genomic_accession=21885423, orientation='plus', exon_count=11, OMIM=300294, no_hdr0=''),
  51361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51361, CurrentID=0, Status='live', Symbol='HOOK1', Aliases='HK1', description='hook microtubule-tethering protein 1', other_designations='h-hook1|hHK1|hook homolog 1', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59814861, end_position_on_the_genomic_accession=59876378, orientation='plus', exon_count=23, OMIM=607820, no_hdr0=''),
  51362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51362, CurrentID=0, Status='live', Symbol='CDC40', Aliases='EHB3, PRP17, PRPF17', description='cell division cycle 40', other_designations='EH-binding protein 3|PRP17 homolog|cell division cycle 40 homolog|hPRP17|pre-mRNA splicing factor 17', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110180421, end_position_on_the_genomic_accession=110232232, orientation='plus', exon_count=16, OMIM=605585, no_hdr0=''),
  51363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51363, CurrentID=0, Status='live', Symbol='CHST15', Aliases='BRAG, GALNAC4S-6ST', description='carbohydrate (N-acetylgalactosamine 4-sulfate 6-O) sulfotransferase 15', other_designations='B cell RAG associated protein (GALNAC4S-6ST)|B-cell RAG-associated gene protein|hBRAG', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123954351, end_position_on_the_genomic_accession=124093684, orientation='minus', exon_count=17, OMIM=608277, no_hdr0=''),
  51364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51364, CurrentID=0, Status='live', Symbol='ZMYND10', Aliases='BLU, CILD22, FLU', description='zinc finger, MYND-type containing 10', other_designations='protein BLu', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50341106, end_position_on_the_genomic_accession=50346028, orientation='minus', exon_count=12, OMIM=607070, no_hdr0=''),
  51365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51365, CurrentID=0, Status='live', Symbol='PLA1A', Aliases='PS-PLA1, PSPLA1', description='phospholipase A1 member A', other_designations='phosphatidylserine-specific phospholipase A1alpha', map_location='3q13.13-q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119597848, end_position_on_the_genomic_accession=119629811, orientation='plus', exon_count=12, OMIM=607460, no_hdr0=''),
  51366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51366, CurrentID=0, Status='live', Symbol='UBR5', Aliases='DD5, EDD, EDD1, HYD', description='ubiquitin protein ligase E3 component n-recognin 5', other_designations='E3 identified by differential display|E3 ubiquitin protein ligase, HECT domain containing, 1|E3 ubiquitin-protein ligase, HECT domain-containing 1|hyperplastic discs protein homolog|progestin induced protein|progestin-induced protein|ubiquitin-protein ligase', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102252273, end_position_on_the_genomic_accession=102412689, orientation='minus', exon_count=60, OMIM=608413, no_hdr0=''),
  51367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51367, CurrentID=0, Status='live', Symbol='POP5', Aliases='HSPC004, RPP2, RPP20, hPop5', description='POP5 homolog, ribonuclease P/MRP subunit', other_designations='processing of precursor 5, ribonuclease P/MRP subunit', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120579045, end_position_on_the_genomic_accession=120581452, orientation='minus', exon_count=5, OMIM=609992, no_hdr0=''),
  51368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51368, CurrentID=0, Status='live', Symbol='TEX264', Aliases='ZSIG11', description='testis expressed 264', other_designations='testis expressed gene 264', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51671175, end_position_on_the_genomic_accession=51704323, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106481833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481833, CurrentID=0, Status='live', Symbol='RN7SL267P', Aliases='', description='RNA, 7SL, cytoplasmic 267, pseudogene', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481834, CurrentID=0, Status='live', Symbol='RN7SL273P', Aliases='', description='RNA, 7SL, cytoplasmic 273, pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51371, CurrentID=0, Status='live', Symbol='POMP', Aliases='C13orf12, HSPC014, PNAS-110, UMP1', description='proteasome maturation protein', other_designations='2510048O06Rik|hUMP1|proteassemblin|protein UMP1 homolog|voltage-gated K channel beta subunit 4.1|voltage-gated potassium channel beta subunit 4.1', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28659004, end_position_on_the_genomic_accession=28678957, orientation='plus', exon_count=6, OMIM=613386, no_hdr0=''),
  51372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51372, CurrentID=0, Status='live', Symbol='TMA7', Aliases='CCDC72, HSPC016', description='translation machinery associated 7 homolog', other_designations='coiled-coil domain containing 72|coiled-coil domain-containing protein 72|translational machinery associated 7 homolog', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48440276, end_position_on_the_genomic_accession=48444129, orientation='plus', exon_count=4, OMIM=615808, no_hdr0=''),
  51373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51373, CurrentID=0, Status='live', Symbol='MRPS17', Aliases='HSPC011, MRP-S17, RPMS17, S17mt', description='mitochondrial ribosomal protein S17', other_designations='', map_location='7p11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55951918, end_position_on_the_genomic_accession=55955341, orientation='plus', exon_count=3, OMIM=611980, no_hdr0=''),
  51374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51374, CurrentID=0, Status='live', Symbol='ATRAID', Aliases='APR--3, APR-3, APR3, C2orf28, HSPC013, PRO240, p18', description='all-trans retinoic acid induced differentiation factor', other_designations='apoptosis related protein 3|apoptosis related protein APR-3|apoptosis-related protein 3', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27212031, end_position_on_the_genomic_accession=27217178, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  51375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51375, CurrentID=0, Status='live', Symbol='SNX7', Aliases='', description='sorting nexin 7', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98661662, end_position_on_the_genomic_accession=98760500, orientation='plus', exon_count=11, OMIM=614904, no_hdr0=''),
  106481840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481840, CurrentID=0, Status='live', Symbol='RN7SL385P', Aliases='', description='RNA, 7SL, cytoplasmic 385, pseudogene', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51377, CurrentID=0, Status='live', Symbol='UCHL5', Aliases='CGI-70, INO80R, UCH-L5, UCH37', description='ubiquitin C-terminal hydrolase L5', other_designations='INO80 complex subunit R|ubiquitin C-terminal hydrolase UCH37|ubiquitin carboxyl-terminal esterase L5|ubiquitin carboxyl-terminal hydrolase L5|ubiquitin thioesterase L5', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193012366, end_position_on_the_genomic_accession=193060107, orientation='minus', exon_count=18, OMIM=610667, no_hdr0=''),
  51378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51378, CurrentID=0, Status='live', Symbol='ANGPT4', Aliases='AGP4, ANG-3, ANG4', description='angiopoietin 4', other_designations='ANG-4|angiopoietin-3|dJ824F16.2 (angiopoietin 4)', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=871243, end_position_on_the_genomic_accession=916375, orientation='minus', exon_count=14, OMIM=603705, no_hdr0=''),
  51379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51379, CurrentID=0, Status='live', Symbol='CRLF3', Aliases='CREME-9, CREME9, CRLM9, CYTOR4, FRWS, p48.2', description='cytokine receptor-like factor 3', other_designations='cytokine receptor-like molecule 9|cytokine receptor-related protein 4|type I cytokine receptor-like factor p48', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30782684, end_position_on_the_genomic_accession=30824760, orientation='minus', exon_count=10, OMIM=614853, no_hdr0=''),
  51380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51380, CurrentID=0, Status='live', Symbol='CSAD', Aliases='CSD, PCAP', description='cysteine sulfinic acid decarboxylase', other_designations='P-selectin cytoplasmic tail-associated protein|cysteine sulfinic acid decarboxylase-related protein|cysteine-sulfinate decarboxylase|sulfinoalanine decarboxylase', map_location='12q13.11-q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53157663, end_position_on_the_genomic_accession=53180909, orientation='minus', exon_count=21, OMIM=616569, no_hdr0=''),
  106481845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481845, CurrentID=0, Status='live', Symbol='RN7SL475P', Aliases='', description='RNA, 7SL, cytoplasmic 475, pseudogene', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51382, CurrentID=0, Status='live', Symbol='ATP6V1D', Aliases='ATP6M, VATD, VMA8', description='ATPase, H+ transporting, lysosomal 34kDa, V1 subunit D', other_designations='ATPase, H+ transporting lysosomal, member M|ATPase, H+ transporting, lysosomal (vacuolar proton pump)|H(+)-transporting two-sector ATPase, subunit M|V-ATPase 28 kDa accessory protein|V-ATPase D subunit|V-ATPase subunit D|vacuolar ATP synthase subunit D|vacuolar H-ATPase subunit D|vacuolar proton pump D subunit|vacuolar proton pump delta polypeptide|vacuolar proton pump subunit D|vacuolar proton-ATPase subunit D', map_location='14q23-q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67337864, end_position_on_the_genomic_accession=67360003, orientation='minus', exon_count=9, OMIM=609398, no_hdr0=''),
  106481847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481847, CurrentID=0, Status='live', Symbol='RN7SL518P', Aliases='', description='RNA, 7SL, cytoplasmic 518, pseudogene', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51384, CurrentID=0, Status='live', Symbol='WNT16', Aliases='', description='wingless-type MMTV integration site family member 16', other_designations='wingless-type MMTV integration site family member 16b|wingless-type MMTV integration site family, member 16', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121325367, end_position_on_the_genomic_accession=121341104, orientation='plus', exon_count=5, OMIM=606267, no_hdr0=''),
  51385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51385, CurrentID=0, Status='live', Symbol='ZNF589', Aliases='SZF1', description='zinc finger protein 589', other_designations='KRAB-zinc finger protein SZF1-1|stem cell zinc finger protein 1', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48241106, end_position_on_the_genomic_accession=48270990, orientation='plus', exon_count=4, OMIM=616702, no_hdr0=''),
  51386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51386, CurrentID=0, Status='live', Symbol='EIF3L', Aliases='EIF3EIP, EIF3S11, EIF3S6IP, HSPC021, HSPC025, MSTP005', description='eukaryotic translation initiation factor 3 subunit L', other_designations='eIEF associated protein HSPC021|eukaryotic translation initiation factor 3 subunit 6-interacting protein|eukaryotic translation initiation factor 3 subunit E-interacting protein|eukaryotic translation initiation factor 3, subunit L', map_location='22q', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37848448, end_position_on_the_genomic_accession=37888782, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  106481851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481851, CurrentID=0, Status='live', Symbol='RN7SL592P', Aliases='', description='RNA, 7SL, cytoplasmic 592, pseudogene', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51388, CurrentID=0, Status='live', Symbol='NIP7', Aliases='CGI-37, HSPC031, KD93', description='NIP7, nucleolar pre-rRNA processing protein', other_designations='nuclear import 7 homolog', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69339512, end_position_on_the_genomic_accession=69343111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51389, CurrentID=0, Status='live', Symbol='RWDD1', Aliases='CGI-24, PTD013', description='RWD domain containing 1', other_designations='DRG family-regulatory protein 2', map_location='6q13-q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116571367, end_position_on_the_genomic_accession=116593601, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  51390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51390, CurrentID=0, Status='live', Symbol='AIG1', Aliases='AIG-1, dJ95L4.1', description='androgen-induced 1', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143059363, end_position_on_the_genomic_accession=143340343, orientation='plus', exon_count=11, OMIM=608514, no_hdr0=''),
  106481855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481855, CurrentID=0, Status='live', Symbol='RN7SL679P', Aliases='', description='RNA, 7SL, cytoplasmic 679, pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481856, CurrentID=0, Status='live', Symbol='RN7SL705P', Aliases='', description='RNA, 7SL, cytoplasmic 705, pseudogene', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51393, CurrentID=0, Status='live', Symbol='TRPV2', Aliases='VRL, VRL-1, VRL1', description='transient receptor potential cation channel, subfamily V, member 2', other_designations='OTRPC2|osm-9-like TRP channel 2|vanilloid receptor-like protein 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16415542, end_position_on_the_genomic_accession=16437003, orientation='plus', exon_count=16, OMIM=606676, no_hdr0=''),
  106481858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481858, CurrentID=0, Status='live', Symbol='RN7SL745P', Aliases='', description='RNA, 7SL, cytoplasmic 745, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481859, CurrentID=0, Status='live', Symbol='RN7SL775P', Aliases='', description='RNA, 7SL, cytoplasmic 775, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481860, CurrentID=0, Status='live', Symbol='RN7SL791P', Aliases='', description='RNA, 7SL, cytoplasmic 791, pseudogene', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51397, CurrentID=0, Status='live', Symbol='COMMD10', Aliases='PTD002', description='COMM domain containing 10', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116083645, end_position_on_the_genomic_accession=116412683, orientation='plus', exon_count=19, OMIM=616704, no_hdr0=''),
  51398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51398, CurrentID=0, Status='live', Symbol='WDR83OS', Aliases='C19orf56, PTD008', description='WD repeat domain 83 opposite strand', other_designations='UPF0139 membrane protein C19orf56|WDR83 opposite strand', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12668067, end_position_on_the_genomic_accession=12669651, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  51399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51399, CurrentID=0, Status='live', Symbol='TRAPPC4', Aliases='CGI-104, HSPC172, PTD009, SBDN, SYNBINDIN, TRS23', description='trafficking protein particle complex 4', other_designations='TRS23 homolog|hematopoietic stem/progenitor cell protein 172', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119018531, end_position_on_the_genomic_accession=119023674, orientation='plus', exon_count=5, OMIM=610971, no_hdr0=''),
  51400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51400, CurrentID=0, Status='live', Symbol='PPME1', Aliases='PME-1', description='protein phosphatase methylesterase 1', other_designations='protein phosphatase methylesterase-1', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74170891, end_position_on_the_genomic_accession=74254703, orientation='plus', exon_count=14, OMIM=611117, no_hdr0=''),
  106481865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481865, CurrentID=0, Status='live', Symbol='RNU4ATAC12P', Aliases='', description='RNA, U4atac small nuclear 12, pseudogene', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481866, CurrentID=0, Status='live', Symbol='RNU4ATAC18P', Aliases='', description='RNA, U4atac small nuclear 18, pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481867, CurrentID=0, Status='live', Symbol='RNU6ATAC15P', Aliases='', description='RNA, U6atac small nuclear 15, pseudogene', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481868, CurrentID=0, Status='live', Symbol='RNU6ATAC30P', Aliases='', description='RNA, U6atac small nuclear 30, pseudogene', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481869, CurrentID=0, Status='live', Symbol='RNU6ATAC41P', Aliases='', description='RNA, U6atac small nuclear 41, pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51406, CurrentID=0, Status='live', Symbol='NOL7', Aliases='C6orf90, PQBP3, RARG-1, dJ223E5.2', description='nucleolar protein 7', other_designations='nucleolar protein 7, 27kDa|nucleolar protein of 27 kDa|polyglutamine binding protein 3|retinoic acid repressible protein', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13615327, end_position_on_the_genomic_accession=13632524, orientation='plus', exon_count=9, OMIM=611533, no_hdr0=''),
  106481871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481871, CurrentID=0, Status='live', Symbol='RNU4-50P', Aliases='', description='RNA, U4 small nuclear 50, pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481872, CurrentID=0, Status='live', Symbol='RNU4-64P', Aliases='', description='RNA, U4 small nuclear 64, pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51409, CurrentID=0, Status='live', Symbol='HEMK1', Aliases='HEMK, MTQ1', description='HemK methyltransferase family member 1', other_designations='HEMK homolog 7kb|m.HsaHemKP', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50569180, end_position_on_the_genomic_accession=50589050, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  106481874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481874, CurrentID=0, Status='live', Symbol='RMRPP5', Aliases='', description='RNA component of mitochondrial RNA processing endoribonuclease pseudogene 5', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51411, CurrentID=0, Status='live', Symbol='BIN2', Aliases='BRAP-1', description='bridging integrator 2', other_designations='breast cancer associated protein BRAP1|breast cancer-associated protein 1|bridging integrator-2', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51281038, end_position_on_the_genomic_accession=51324680, orientation='minus', exon_count=16, OMIM=605936, no_hdr0=''),
  51412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51412, CurrentID=0, Status='live', Symbol='ACTL6B', Aliases='ACTL6, BAF53B, arpNalpha', description='actin like 6B', other_designations='53 kDa BRG1-associated factor B|BRG1-associated factor 53B|actin-related protein Baf53b|hArpN alpha', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100643097, end_position_on_the_genomic_accession=100656461, orientation='minus', exon_count=14, OMIM=612458, no_hdr0=''),
  106481877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481877, CurrentID=0, Status='live', Symbol='RNU6-133P', Aliases='', description='RNA, U6 small nuclear 133, pseudogene', other_designations='', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481878, CurrentID=0, Status='live', Symbol='RNU6-154P', Aliases='', description='RNA, U6 small nuclear 154, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481879, CurrentID=0, Status='live', Symbol='RNU6-166P', Aliases='', description='RNA, U6 small nuclear 166, pseudogene', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481880, CurrentID=0, Status='live', Symbol='RNU6-187P', Aliases='', description='RNA, U6 small nuclear 187, pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481881, CurrentID=0, Status='live', Symbol='RNU6-199P', Aliases='', description='RNA, U6 small nuclear 199, pseudogene', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481882, CurrentID=0, Status='live', Symbol='RNU6-206P', Aliases='', description='RNA, U6 small nuclear 206, pseudogene', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481883, CurrentID=0, Status='live', Symbol='RNU6-227P', Aliases='', description='RNA, U6 small nuclear 227, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481884, CurrentID=0, Status='live', Symbol='RNU6-246P', Aliases='', description='RNA, U6 small nuclear 246, pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51421, CurrentID=0, Status='live', Symbol='AMOTL2', Aliases='LCCP', description='angiomotin like 2', other_designations='Leman coiled-coil protein', map_location='3q21-q22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134355345, end_position_on_the_genomic_accession=134375417, orientation='minus', exon_count=14, OMIM=614658, no_hdr0=''),
  51422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51422, CurrentID=0, Status='live', Symbol='PRKAG2', Aliases='AAKG, AAKG2, CMH6, H91620p, WPWS', description='protein kinase, AMP-activated, gamma 2 non-catalytic subunit', other_designations='AMPK subunit gamma-2', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151556114, end_position_on_the_genomic_accession=151877231, orientation='minus', exon_count=21, OMIM=602743, no_hdr0=''),
  106481887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481887, CurrentID=0, Status='live', Symbol='RNU6-327P', Aliases='', description='RNA, U6 small nuclear 327, pseudogene', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481888, CurrentID=0, Status='live', Symbol='RNU6-339P', Aliases='', description='RNA, U6 small nuclear 339, pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481889, CurrentID=0, Status='live', Symbol='RNU6-355P', Aliases='', description='RNA, U6 small nuclear 355, pseudogene', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51426, CurrentID=0, Status='live', Symbol='POLK', Aliases='DINB1, DINP, POLQ', description='polymerase (DNA directed) kappa', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75511043, end_position_on_the_genomic_accession=75599823, orientation='plus', exon_count=22, OMIM=605650, no_hdr0=''),
  51427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51427, CurrentID=0, Status='live', Symbol='ZNF107', Aliases='Y8, ZFD25, ZNF588, smap-7', description='zinc finger protein 107', other_designations='C2H2 type zinc-finger protein|zinc finger protein (ZFD25)|zinc finger protein 588', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64666083, end_position_on_the_genomic_accession=64711582, orientation='plus', exon_count=8, OMIM=603989, no_hdr0=''),
  51428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51428, CurrentID=0, Status='live', Symbol='DDX41', Aliases='ABS', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 41', other_designations='2900024F02Rik|DEAD box protein 41|DEAD box protein abstrakt homolog|DEAD-box protein abstrakt|putative RNA helicase', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177511577, end_position_on_the_genomic_accession=177516966, orientation='minus', exon_count=17, OMIM=608170, no_hdr0=''),
  51429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51429, CurrentID=0, Status='live', Symbol='SNX9', Aliases='SDP1, SH3PX1, SH3PXD3A, WISP', description='sorting nexin 9', other_designations='SH3 and PX domain-containing protein 1|SH3 and PX domain-containing protein 3A|Wiskott-Aldrich syndrome protein (WASP) interactor protein', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157716046, end_position_on_the_genomic_accession=157945077, orientation='plus', exon_count=22, OMIM=605952, no_hdr0=''),
  51430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51430, CurrentID=0, Status='live', Symbol='SUCO', Aliases='C1orf9, CH1, OPT, SLP1', description='SUN domain containing ossification factor', other_designations='SUN-like protein 1|membrane protein CH1|osteopotentia', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172532349, end_position_on_the_genomic_accession=172611833, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  106481895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481895, CurrentID=0, Status='live', Symbol='RNU6-440P', Aliases='', description='RNA, U6 small nuclear 440, pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481896, CurrentID=0, Status='live', Symbol='RNU6-449P', Aliases='', description='RNA, U6 small nuclear 449, pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51433, CurrentID=0, Status='live', Symbol='ANAPC5', Aliases='APC5', description='anaphase promoting complex subunit 5', other_designations='cyclosome subunit 5', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121308245, end_position_on_the_genomic_accession=121354209, orientation='minus', exon_count=18, OMIM=606948, no_hdr0=''),
  51434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51434, CurrentID=0, Status='live', Symbol='ANAPC7', Aliases='APC7', description='anaphase promoting complex subunit 7', other_designations='cyclosome subunit 7', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110372900, end_position_on_the_genomic_accession=110403730, orientation='minus', exon_count=13, OMIM=606949, no_hdr0=''),
  51435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51435, CurrentID=0, Status='live', Symbol='SCARA3', Aliases='APC7, CSR, CSR1, MSLR1, MSRL1', description='scavenger receptor class A member 3', other_designations='cellular stress response gene protein|cellular stress response protein|macrophage scavenger receptor-like 1|scavenger receptor class A, member 3', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27633894, end_position_on_the_genomic_accession=27712743, orientation='plus', exon_count=10, OMIM=602728, no_hdr0=''),
  106481900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481900, CurrentID=0, Status='live', Symbol='RNU6-521P', Aliases='', description='RNA, U6 small nuclear 521, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481901, CurrentID=0, Status='live', Symbol='RNU6-533P', Aliases='', description='RNA, U6 small nuclear 533, pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51438, CurrentID=0, Status='live', Symbol='MAGEC2', Aliases='CT10, HCA587, MAGEE1', description='MAGE family member C2', other_designations='MAGE-C2 antigen|MAGE-E1 antigen|cancer/testis antigen 10|hepatocellular cancer antigen 587|hepatocellular carcinoma-associated antigen 587|melanoma antigen family C, 2|melanoma antigen family C2|melanoma antigen, family E, 1, cancer/testis specific', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142202342, end_position_on_the_genomic_accession=142205290, orientation='minus', exon_count=3, OMIM=300468, no_hdr0=''),
  51439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51439, CurrentID=0, Status='live', Symbol='FAM8A1', Aliases='AHCP', description='family with sequence similarity 8 member A1', other_designations='Autosomal Highly Conserved Protein|family with sequence similarity 8, member A1', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17600287, end_position_on_the_genomic_accession=17611719, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51440, CurrentID=0, Status='live', Symbol='HPCAL4', Aliases='HLP4', description='hippocalcin like 4', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39678648, end_position_on_the_genomic_accession=39691710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  51441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51441, CurrentID=0, Status='live', Symbol='YTHDF2', Aliases='CAHL, HGRG8, NY-REN-2', description='YTH N(6)-methyladenosine RNA binding protein 2', other_designations='9430020E02Rik|CLL-associated antigen KW-14|YTH domain family, member 2|high-glucose-regulated protein 8|renal carcinoma antigen NY-REN-2', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28736621, end_position_on_the_genomic_accession=28769775, orientation='plus', exon_count=6, OMIM=610640, no_hdr0=''),
  51442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51442, CurrentID=0, Status='live', Symbol='VGLL1', Aliases='TDU, VGL1', description='vestigial like family member 1', other_designations='TONDU|WUGSC:H_GS188P18.1b|vestigial like 1|vestigial-like family member 1|vgl-1', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136532152, end_position_on_the_genomic_accession=136556807, orientation='plus', exon_count=5, OMIM=300583, no_hdr0=''),
  106481907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481907, CurrentID=0, Status='live', Symbol='RNU6-621P', Aliases='', description='RNA, U6 small nuclear 621, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51444, CurrentID=0, Status='live', Symbol='RNF138', Aliases='HSD-4, NARF, STRIN, hNARF', description='ring finger protein 138, E3 ubiquitin protein ligase', other_designations='NLK-associated RING finger protein|Nemo-like kinase-associated RING finger protein|nemo-like kinase associated ring finger protein', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32091855, end_position_on_the_genomic_accession=32131561, orientation='plus', exon_count=8, OMIM=616319, no_hdr0=''),
  106481909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481909, CurrentID=0, Status='live', Symbol='RNU6-654P', Aliases='', description='RNA, U6 small nuclear 654, pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481910, CurrentID=0, Status='live', Symbol='RNU6-672P', Aliases='', description='RNA, U6 small nuclear 672, pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51447, CurrentID=0, Status='live', Symbol='IP6K2', Aliases='IHPK2, PIUS', description='inositol hexakisphosphate kinase 2', other_designations='ATP:1D-myo-inositol-hexakisphosphate phosphotransferase|inositol hexaphosphate kinase 2|insp6 kinase 2|pi uptake stimulator', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48688003, end_position_on_the_genomic_accession=48717278, orientation='minus', exon_count=9, OMIM=606992, no_hdr0=''),
  106481912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481912, CurrentID=0, Status='live', Symbol='RNU6-702P', Aliases='', description='RNA, U6 small nuclear 702, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51449, CurrentID=0, Status='live', Symbol='PCYOX1', Aliases='PCL1', description='prenylcysteine oxidase 1', other_designations='prenylcysteine lyase', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70258099, end_position_on_the_genomic_accession=70281185, orientation='plus', exon_count=6, OMIM=610995, no_hdr0=''),
  51450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51450, CurrentID=0, Status='live', Symbol='PRRX2', Aliases='PMX2, PRX2', description='paired related homeobox 2', other_designations='PRX-2|paired-like homeodomain protein PRX2|paired-related homeobox protein 2', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129665641, end_position_on_the_genomic_accession=129722674, orientation='plus', exon_count=6, OMIM=604675, no_hdr0=''),
  51451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51451, CurrentID=0, Status='live', Symbol='LCMT1', Aliases='CGI-68, LCMT, PPMT1', description='leucine carboxyl methyltransferase 1', other_designations='[Phosphatase 2A protein]-leucine-carboxy methyltransferase 1|protein phosphatase methyltransferase 1|protein-leucine O-methyltransferase', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25111726, end_position_on_the_genomic_accession=25178231, orientation='plus', exon_count=13, OMIM=610286, no_hdr0=''),
  106481916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481916, CurrentID=0, Status='live', Symbol='RNU6-767P', Aliases='', description='RNA, U6 small nuclear 767, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481917, CurrentID=0, Status='live', Symbol='RNU6-782P', Aliases='', description='RNA, U6 small nuclear 782, pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51454, CurrentID=0, Status='live', Symbol='GULP1', Aliases='CED-6, CED6, GULP', description='GULP, engulfment adaptor PTB domain containing 1', other_designations='PTB domain adapter protein CED-6|PTB domain adaptor protein CED-6|cell death protein 6 homolog|engulfment adapter protein', map_location='2q32.3-q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188291669, end_position_on_the_genomic_accession=188595926, orientation='plus', exon_count=22, OMIM=608165, no_hdr0=''),
  51455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51455, CurrentID=0, Status='live', Symbol='REV1', Aliases='REV1L', description='REV1, DNA directed polymerase', other_designations='AIBP80|REV1 homolog|REV1, polymerase (DNA directed)|REV1- like|alpha integrin-binding protein 80|rev1-like terminal deoxycytidyl transferase', map_location='2q11.1-q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99400475, end_position_on_the_genomic_accession=99490030, orientation='minus', exon_count=26, OMIM=606134, no_hdr0=''),
  106481920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481920, CurrentID=0, Status='live', Symbol='RNU6-862P', Aliases='', description='RNA, U6 small nuclear 862, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481921, CurrentID=0, Status='live', Symbol='RNU6-883P', Aliases='', description='RNA, U6 small nuclear 883, pseudogene', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51458, CurrentID=0, Status='live', Symbol='RHCG', Aliases='C15orf6, PDRC2, RHGK, SLC42A3', description='Rh family, C glycoprotein', other_designations='Rh type C glycoprotein|Rhesus blood group, C glycoprotein|rh family type C glycoprotein|rh glycoprotein kidney|rhesus blood group family type C glycoprotein|tumor-related protein DRC2', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89471407, end_position_on_the_genomic_accession=89496613, orientation='minus', exon_count=11, OMIM=605381, no_hdr0=''),
  106481923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481923, CurrentID=0, Status='live', Symbol='RNU6-915P', Aliases='', description='RNA, U6 small nuclear 915, pseudogene', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51460, CurrentID=0, Status='live', Symbol='SFMBT1', Aliases='RU1, SFMBT, hSFMBT', description='Scm-like with four mbt domains 1', other_designations='Scm-related gene containing four mbt domains|Scm-related gene product containing four mbt domains|renal ubiquitous protein 1', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52899982, end_position_on_the_genomic_accession=53046655, orientation='minus', exon_count=24, OMIM=607319, no_hdr0=''),
  106481925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481925, CurrentID=0, Status='live', Symbol='RNU6-949P', Aliases='', description='RNA, U6 small nuclear 949, pseudogene', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481926, CurrentID=0, Status='live', Symbol='RNU6-971P', Aliases='', description='RNA, U6 small nuclear 971, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51463, CurrentID=0, Status='live', Symbol='GPR89B', Aliases='GPHR, GPR89, GPR89C, SH120, UNQ192', description='G protein-coupled receptor 89B', other_designations='G protein-coupled receptor 89C|Golgi pH regulator A|protein GPR89|putative Golgi pH regulator C|putative MAPK-activating protein PM01|putative NF-kappa-B-activating protein 90|putative protein GPR89C', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147928393, end_position_on_the_genomic_accession=147993521, orientation='plus', exon_count=18, OMIM=612806, no_hdr0=''),
  106481928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481928, CurrentID=0, Status='live', Symbol='RNU6-996P', Aliases='', description='RNA, U6 small nuclear 996, pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51465, CurrentID=0, Status='live', Symbol='UBE2J1', Aliases='CGI-76, HSPC153, HSPC205, HSU93243, NCUBE-1, NCUBE1, UBC6, UBC6E, Ubc6p', description='ubiquitin conjugating enzyme E2, J1', other_designations='E2 ubiquitin-conjugating enzyme J1|HSUBC6e|non-canonical ubiquitin-conjugating enzyme 1|ubiquitin-conjugating enzyme E2, J1 (UBC6 homolog, yeast)|ubiquitin-conjugating enzyme E2, J1, U|yeast ubiquitin-conjugating enzyme UBC6 homolog E', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89326625, end_position_on_the_genomic_accession=89352900, orientation='minus', exon_count=8, OMIM=616175, no_hdr0=''),
  51466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51466, CurrentID=0, Status='live', Symbol='EVL', Aliases='RNB6', description='Enah/Vasp-like', other_designations='ena/vasodilator-stimulated phosphoprotein-like', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100019387, end_position_on_the_genomic_accession=100144236, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  106481931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481931, CurrentID=0, Status='live', Symbol='RNU6-1039P', Aliases='', description='RNA, U6 small nuclear 1039, pseudogene', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481932, CurrentID=0, Status='live', Symbol='RNU6-1051P', Aliases='', description='RNA, U6 small nuclear 1051, pseudogene', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481933, CurrentID=0, Status='live', Symbol='RNU6-1068P', Aliases='', description='RNA, U6 small nuclear 1068, pseudogene', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481934, CurrentID=0, Status='live', Symbol='RNU6-1086P', Aliases='', description='RNA, U6 small nuclear 1086, pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51471, CurrentID=0, Status='live', Symbol='NAT8B', Aliases='CML2, Hcml2P, NAT8B', description='N-acetyltransferase 8B (GCN5-related, putative, gene/pseudogene)', other_designations='ATase1|N-acetyltransferase Camello 2|acetyltransferase 1|camello-like protein 2|probable N-acetyltransferase 8B', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73700509, end_position_on_the_genomic_accession=73701340, orientation='minus', exon_count=1, OMIM=608190, no_hdr0=''),
  106481936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481936, CurrentID=0, Status='live', Symbol='RNU6-1122P', Aliases='', description='RNA, U6 small nuclear 1122, pseudogene', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51473, CurrentID=0, Status='live', Symbol='DCDC2', Aliases='DCDC2A, DFNB66, NPHP19, RU2, RU2S', description='doublecortin domain containing 2', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24171755, end_position_on_the_genomic_accession=24383292, orientation='minus', exon_count=11, OMIM=605755, no_hdr0=''),
  51474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51474, CurrentID=0, Status='live', Symbol='LIMA1', Aliases='EPLIN, SREBP3', description='LIM domain and actin binding 1', other_designations='1110021C24Rik|epithelial protein lost in neoplasm beta|sterol regulatory element binding protein 3', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50175780, end_position_on_the_genomic_accession=50283570, orientation='minus', exon_count=12, OMIM=608364, no_hdr0=''),
  51475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51475, CurrentID=0, Status='live', Symbol='CABP2', Aliases='DFNB93', description='calcium binding protein 2', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67518947, end_position_on_the_genomic_accession=67523614, orientation='minus', exon_count=7, OMIM=607314, no_hdr0=''),
  106481940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481940, CurrentID=0, Status='live', Symbol='RNU6-1171P', Aliases='', description='RNA, U6 small nuclear 1171, pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51477, CurrentID=0, Status='live', Symbol='ISYNA1', Aliases='INO1, INOS, IPS, IPS 1, IPS-1', description='inositol-3-phosphate synthase 1', other_designations='MI-1-P synthase|MIP synthase|myo-inositol 1-phosphate synthase A1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18434388, end_position_on_the_genomic_accession=18438301, orientation='minus', exon_count=11, OMIM=611670, no_hdr0=''),
  51478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51478, CurrentID=0, Status='live', Symbol='HSD17B7', Aliases='PRAP, SDR37C1', description='hydroxysteroid (17-beta) dehydrogenase 7', other_designations='17 beta-hydroxysteroid dehydrogenase type VII|17-beta-HSD 7|17-beta-hydroxysteroid dehydrogenase 7|17beta hydroxysteroid dehydrogenase|estradiol 17-beta-dehydrogenase 7|short chain dehydrogenase/reductase family 37C member 1|short chain dehydrogenase/reductase family 37C, member 1', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162790702, end_position_on_the_genomic_accession=162812818, orientation='plus', exon_count=10, OMIM=606756, no_hdr0=''),
  51479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51479, CurrentID=0, Status='live', Symbol='ANKFY1', Aliases='ANKHZN, BTBD23, ZFYVE14', description='ankyrin repeat and FYVE domain containing 1', other_designations='Rabankyrin-5|ankyrin repeat and FYVE domain-containing protein 1|ankyrin repeat hooked to zinc finger motif|ankyrin repeats hooked to a zinc finger motif|rank-5', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4163371, end_position_on_the_genomic_accession=4264008, orientation='minus', exon_count=30, OMIM=607927, no_hdr0=''),
  51480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51480, CurrentID=0, Status='live', Symbol='VCX2', Aliases='VCX-2rR, VCXB, VCX2', description='variable charge, X-linked 2', other_designations='variable charge protein on X with two repeats|variably charged protein X-B|variably charged, X chromosome 2|variably charged, X chromosome B|variably charged, X chromosome, with 2 repeats', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8169944, end_position_on_the_genomic_accession=8171267, orientation='minus', exon_count=3, OMIM=300532, no_hdr0=''),
  51481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51481, CurrentID=0, Status='live', Symbol='VCX3A', Aliases='VCX-8r, VCX-A, VCX3, VCX8R, VCXA', description='variable charge, X-linked 3A', other_designations='variable charge protein on X with eight repeats|variably charged X-A|variably charged protein X-A', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=6533618, end_position_on_the_genomic_accession=6535118, orientation='minus', exon_count=3, OMIM=300533, no_hdr0=''),
  106481946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481946, CurrentID=0, Status='live', Symbol='RNU6-1255P', Aliases='', description='RNA, U6 small nuclear 1255, pseudogene', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481947, CurrentID=0, Status='live', Symbol='RNU6-1273P', Aliases='', description='RNA, U6 small nuclear 1273, pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481948, CurrentID=0, Status='live', Symbol='RNU6-1285P', Aliases='', description='RNA, U6 small nuclear 1285, pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481949, CurrentID=0, Status='live', Symbol='RNU6-1303P', Aliases='', description='RNA, U6 small nuclear 1303, pseudogene', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481950, CurrentID=0, Status='live', Symbol='RNU6-1315P', Aliases='', description='RNA, U6 small nuclear 1315, pseudogene', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481951, CurrentID=0, Status='live', Symbol='RNU6-1326P', Aliases='', description='RNA, U6 small nuclear 1326, pseudogene', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481952, CurrentID=0, Status='live', Symbol='RNU1-7P', Aliases='RNU1-7', description='RNA, U1 small nuclear 7, pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481953, CurrentID=0, Status='live', Symbol='RNU1-39P', Aliases='', description='RNA, U1 small nuclear 39, pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51490, CurrentID=0, Status='live', Symbol='C9orf114', Aliases='CENP-32, HSPC109', description='chromosome 9 open reading frame 114', other_designations='centromere protein 32', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128819651, end_position_on_the_genomic_accession=128829847, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  51491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51491, CurrentID=0, Status='live', Symbol='NOP16', Aliases='HSPC111, HSPC185', description='NOP16 nucleolar protein', other_designations='HBV pre-S2 trans-regulated protein 3|NOP16 nucleolar protein homolog|hypothetical protein HSPC111|nucleolar protein 16 homolog', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176383937, end_position_on_the_genomic_accession=176388762, orientation='minus', exon_count=6, OMIM=612861, no_hdr0=''),
  106481956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481956, CurrentID=0, Status='live', Symbol='RNU1-94P', Aliases='', description='RNA, U1 small nuclear 94, pseudogene', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51493, CurrentID=0, Status='live', Symbol='RTCB', Aliases='C22orf28, DJ149A16.6, FAAP, HSPC117', description="RNA 2',3'-cyclic phosphate and 5'-OH ligase", other_designations='ankyrin repeat domain 54|focal adhesion-associated protein', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32387575, end_position_on_the_genomic_accession=32412287, orientation='minus', exon_count=12, OMIM=613901, no_hdr0=''),
  106481958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481958, CurrentID=0, Status='live', Symbol='RNU1-132P', Aliases='', description='RNA, U1 small nuclear 132, pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51495, CurrentID=0, Status='live', Symbol='HACD3', Aliases='B-IND1, BIND1, HSPC121, PTPLAD1', description='3-hydroxyacyl-CoA dehydratase 3', other_designations='butyrate-induced protein 1|butyrate-induced transcript 1|protein tyrosine phosphatase-like A domain containing 1|protein tyrosine phosphatase-like protein PTPLAD1|protein-tyrosine phosphatase-like A domain-containing protein 1|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 3', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65530489, end_position_on_the_genomic_accession=65578355, orientation='plus', exon_count=12, OMIM=615940, no_hdr0=''),
  51496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51496, CurrentID=0, Status='live', Symbol='CTDSPL2', Aliases='HSPC058, HSPC129', description='CTD small phosphatase like 2', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase like 2|CTDSP-like 2', map_location='15q15.3-q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44427329, end_position_on_the_genomic_accession=44527257, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  51497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51497, CurrentID=0, Status='live', Symbol='NELFCD', Aliases='HSPC130, NELF-C, NELF-D, TH1, TH1L', description='negative elongation factor complex member C/D', other_designations='NELF-C/D|TH1 drosophila homolog|TH1-like protein|negative elongation factor proteins C and D|trihydrophobin 1', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58981208, end_position_on_the_genomic_accession=58995134, orientation='plus', exon_count=15, OMIM=605297, no_hdr0=''),
  106481962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481962, CurrentID=0, Status='live', Symbol='RNU2-68P', Aliases='', description='RNA, U2 small nuclear 68, pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51499, CurrentID=0, Status='live', Symbol='TRIAP1', Aliases='HSPC132, MDM35, P53CSV, WF-1', description='TP53 regulated inhibitor of apoptosis 1', other_designations='mitochondrial distribution and morphology 35 homolog|p53-inducible cell-survival factor|protein 15E1.1', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120443961, end_position_on_the_genomic_accession=120446412, orientation='minus', exon_count=2, OMIM=614943, no_hdr0=''),
  106481964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481964, CurrentID=0, Status='live', Symbol='TCEB1P32', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 32', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51501, CurrentID=0, Status='live', Symbol='C11orf73', Aliases='HSPC138, HSPC179, Hikeshi, L7RN6, OPI10', description='chromosome 11 open reading frame 73', other_designations='lethal, Chr 7, Rinchik 6', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86302211, end_position_on_the_genomic_accession=86345943, orientation='plus', exon_count=7, OMIM=614908, no_hdr0=''),
  106481966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481966, CurrentID=0, Status='live', Symbol='BNIP3P28', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 28', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51503, CurrentID=0, Status='live', Symbol='CWC15', Aliases='AD002, C11orf5, Cwf15, HSPC148, ORF5', description='CWC15 spliceosome-associated protein', other_designations='CWC15 spliceosome-associated protein homolog', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94962622, end_position_on_the_genomic_accession=94973612, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51504, CurrentID=0, Status='live', Symbol='TRMT112', Aliases='HSPC152, HSPC170, TRM112, TRMT11-2', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae)', other_designations='TRM112-like protein|tRNA methyltransferase 112 homolog', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64316492, end_position_on_the_genomic_accession=64318084, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106481969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481969, CurrentID=0, Status='live', Symbol='WBP1LP11', Aliases='', description='WW domain binding protein 1-like pseudogene 11', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51506, CurrentID=0, Status='live', Symbol='UFC1', Aliases='HSPC155', description='ubiquitin-fold modifier conjugating enzyme 1', other_designations='Ufm1-conjugating enzyme 1', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161153744, end_position_on_the_genomic_accession=161158856, orientation='plus', exon_count=6, OMIM=610554, no_hdr0=''),
  51507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51507, CurrentID=0, Status='live', Symbol='RTFDC1', Aliases='C20orf43, CDAO5, HSPC164, SHUJUN-3', description='replication termination factor 2 domain containing 1', other_designations='UPF0549 protein C20orf43|replication termination factor 2 domain-containing protein 1', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56468585, end_position_on_the_genomic_accession=56518886, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  106481972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481972, CurrentID=0, Status='live', Symbol='RARRES2P10', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 10', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481973, CurrentID=0, Status='live', Symbol='RNU6-12P', Aliases='RNU6-12', description='RNA, U6 small nuclear 12, pseudogene', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51510, CurrentID=0, Status='live', Symbol='CHMP5', Aliases='C9orf83, CGI-34, HSPC177, PNAS-2, SNF7DC2, Vps60', description='charged multivesicular body protein 5', other_designations='SNF7 domain containing 2|SNF7 domain-containing protein 2|apoptosis-related protein PNAS-2|chromatin modifying protein 5|chromatin-modifying protein 5|hVps60|vacuolar protein sorting-associated protein 60', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33264879, end_position_on_the_genomic_accession=33282069, orientation='plus', exon_count=8, OMIM=610900, no_hdr0=''),
  106481975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481975, CurrentID=0, Status='live', Symbol='RN7SL45P', Aliases='', description='RNA, 7SL, cytoplasmic 45, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51512, CurrentID=0, Status='live', Symbol='GTSE1', Aliases='B99', description='G2 and S-phase expressed 1', other_designations='G-2 and S-phase expressed 1|protein B99 homolog', map_location='22q13.2-q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46296014, end_position_on_the_genomic_accession=46333699, orientation='plus', exon_count=14, OMIM=607477, no_hdr0=''),
  51513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51513, CurrentID=0, Status='live', Symbol='ETV7', Aliases='TEL-2, TEL2, TELB', description='ets variant 7', other_designations='ETS translocation variant 7|ETS-related protein Tel2|Ets transcription factor TEL-2b|TEL2 oncogene|ets variant gene 7 (TEL2 oncogene)|tel-related Ets factor', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36354221, end_position_on_the_genomic_accession=36387800, orientation='minus', exon_count=10, OMIM=605255, no_hdr0=''),
  51514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51514, CurrentID=0, Status='live', Symbol='DTL', Aliases='CDT2, DCAF2, L2DTL, RAMP', description='denticleless E3 ubiquitin protein ligase homolog (Drosophila)', other_designations='DDB1 and CUL4 associated factor 2|DDB1- and CUL4-associated factor 2|RA regulated nuclear matrix associated protein|RA-regulated nuclear matrix-associated protein|lethal(2) denticleless protein homolog|retinoic acid-regulated nuclear matrix-associated protein', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212035553, end_position_on_the_genomic_accession=212105006, orientation='plus', exon_count=16, OMIM=610617, no_hdr0=''),
  106481979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481979, CurrentID=0, Status='live', Symbol='RNU6-982P', Aliases='', description='RNA, U6 small nuclear 982, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106481980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481980, CurrentID=0, Status='live', Symbol='RNU1-61P', Aliases='', description='RNA, U1 small nuclear 61, pseudogene', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51517, CurrentID=0, Status='live', Symbol='NCKIPSD', Aliases='AF3P21, DIP, DIP1, ORF1, SPIN90, VIP54, WASLBP, WISH', description='NCK interacting protein with SH3 domain', other_designations='54 kDa VacA-interacting protein|54 kDa vimentin-interacting protein|90 kDa SH3 protein interacting with Nck|SH3 adapter protein SPIN90|SH3 protein interacting with Nck, 90 kDa|WASP-interacting SH3-domain protein|dia interacting protein|dia-interacting protein 1|diaphanous protein interacting protein|wiskott-Aldrich syndrome protein-interacting protein', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48673839, end_position_on_the_genomic_accession=48685933, orientation='minus', exon_count=13, OMIM=606671, no_hdr0=''),
  51520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51520, CurrentID=0, Status='live', Symbol='LARS', Aliases='HSPC192, ILFS11, LEURS, LEUS, LFIS, LRS, PIG44, RNTLS, hr025Cl, LARS', description='leucyl-tRNA synthetase', other_designations='cytoplasmic leucyl-tRNA synthetase|cytosolic leucyl-tRNA synthetase|leucine tRNA ligase 1, cytoplasmic|leucine translase|proliferation-inducing gene 44', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146113026, end_position_on_the_genomic_accession=146182787, orientation='minus', exon_count=32, OMIM=151350, no_hdr0=''),
  51522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51522, CurrentID=0, Status='live', Symbol='TMEM14C', Aliases='C6orf53, HSPC194, MSTP073, NET26, bA421M1.6', description='transmembrane protein 14C', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10722915, end_position_on_the_genomic_accession=10731129, orientation='plus', exon_count=6, OMIM=615318, no_hdr0=''),
  51523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51523, CurrentID=0, Status='live', Symbol='CXXC5', Aliases='CF5, HSPC195, RINF, WID', description='CXXC finger protein 5', other_designations='CXXC finger 5 protein|WT1-induced Inhibitor of Dishevelled|putative MAPK-activating protein PM08|putative NF-kappa-B-activating protein 102|retinoid-inducible nuclear factor', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139648716, end_position_on_the_genomic_accession=139683882, orientation='plus', exon_count=6, OMIM=612752, no_hdr0=''),
  51524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51524, CurrentID=0, Status='live', Symbol='TMEM138', Aliases='HSPC196', description='transmembrane protein 138', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61362001, end_position_on_the_genomic_accession=61376502, orientation='plus', exon_count=8, OMIM=614459, no_hdr0=''),
  51526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51526, CurrentID=0, Status='live', Symbol='OSER1', Aliases='C20orf111, HSPC207, Osr1, Perit1, dJ1183I21.1', description='oxidative stress responsive serine-rich 1', other_designations='oxidative stress-responsive 1|oxidative stress-responsive protein 1|peroxide-inducible transcript 1 protein', map_location='20q13.11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44195939, end_position_on_the_genomic_accession=44210906, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51527, CurrentID=0, Status='live', Symbol='GSKIP', Aliases='C14orf129, HSPC210', description='GSK3B interacting protein', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96363452, end_position_on_the_genomic_accession=96387290, orientation='plus', exon_count=5, OMIM=616605, no_hdr0=''),
  51528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51528, CurrentID=0, Status='live', Symbol='JKAMP', Aliases='C14orf100, CDA06, HSPC213, HSPC327, JAMP', description='JNK1/MAPK8-associated membrane protein', other_designations='JNK-associated membrane protein|Jun N-terminal kinase 1-associated membrane protein|medulloblastoma antigen MU-MB-50.4', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59484443, end_position_on_the_genomic_accession=59505407, orientation='plus', exon_count=8, OMIM=611176, no_hdr0=''),
  51529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51529, CurrentID=0, Status='live', Symbol='ANAPC11', Aliases='APC11, Apc11p, HSPC214', description='anaphase promoting complex subunit 11', other_designations='APC11 anaphase promoting complex subunit 11 homolog|anaphase promoting complex subunit 11 (yeast APC11 homolog)|cyclosome subunit 11|hepatocellular carcinoma-associated RING finger protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81890790, end_position_on_the_genomic_accession=81900533, orientation='plus', exon_count=8, OMIM=614534, no_hdr0=''),
  51530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51530, CurrentID=0, Status='live', Symbol='ZC3HC1', Aliases='HSPC216, NIPA', description='zinc finger, C3HC-type containing 1', other_designations='hematopoietic stem/progenitor cell protein 216|nuclear interacting partner of anaplastic lymphoma kinase (ALK)', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130018286, end_position_on_the_genomic_accession=130051451, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  51531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51531, CurrentID=0, Status='live', Symbol='TRMO', Aliases='C9orf156, HSPC219, NAP1', description='tRNA methyltransferase O', other_designations='Nef (lentivirus myristoylated factor) associated protein 1|Nef associated protein 1|thioesterase NAP1', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97876324, end_position_on_the_genomic_accession=97922570, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  51533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51533, CurrentID=0, Status='live', Symbol='PHF7', Aliases='HSPC045, HSPC226, NYD-SP6', description='PHD finger protein 7', other_designations='testis development protein NYD-SP6', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52410558, end_position_on_the_genomic_accession=52423641, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  104909134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104909134, CurrentID=0, Status='live', Symbol='MGC57346-CRHR1', Aliases='', description='MGC57346-CRHR1 readthrough', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45620329, end_position_on_the_genomic_accession=45835828, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  51535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51535, CurrentID=0, Status='live', Symbol='PPHLN1', Aliases='CR, HSPC206, HSPC232', description='periphilin 1', other_designations='CDC7 expression repressor|gastric cancer antigen Ga50', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42326137, end_position_on_the_genomic_accession=42450965, orientation='plus', exon_count=17, OMIM=608150, no_hdr0=''),
  51537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51537, CurrentID=0, Status='live', Symbol='MTFP1', Aliases='HSPC242, MTP18', description='mitochondrial fission process 1', other_designations='mitochondrial 18 kDa protein|mitochondrial fission protein MTP18|mitochondrial protein 18 kDa', map_location='22q', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30425623, end_position_on_the_genomic_accession=30429054, orientation='plus', exon_count=4, OMIM=610235, no_hdr0=''),
  51538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51538, CurrentID=0, Status='live', Symbol='ZCCHC17', Aliases='HSPC251, PS1D, pNO40', description='zinc finger, CCHC domain containing 17', other_designations='nucleolar protein 40|pnn-interacting nucleolar protein|putative S1 RNA binding domain protein|putative S1 RNA-binding domain protein|zinc finger CCHC domain-containing protein 17', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31296982, end_position_on_the_genomic_accession=31365722, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  51540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51540, CurrentID=0, Status='live', Symbol='SCLY', Aliases='SCL, hSCL', description='selenocysteine lyase', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238060924, end_position_on_the_genomic_accession=238099413, orientation='plus', exon_count=12, OMIM=611056, no_hdr0=''),
  51542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51542, CurrentID=0, Status='live', Symbol='VPS54', Aliases='HCC8, PPP1R164, SLP-8pL, WR, hVps54L, VPS54', description='VPS54 GARP complex subunit', other_designations='hepatocellular carcinoma protein 8|protein phosphatase 1, regulatory subunit 164|tumor antigen HOM-HCC-8|tumor antigen SLP-8p|vacuolar protein sorting 54 homolog', map_location='2p13-p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63892150, end_position_on_the_genomic_accession=64019080, orientation='minus', exon_count=25, OMIM=614633, no_hdr0=''),
  51545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51545, CurrentID=0, Status='live', Symbol='ZNF581', Aliases='HSPC189', description='zinc finger protein 581', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55643620, end_position_on_the_genomic_accession=55645623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  51547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51547, CurrentID=0, Status='live', Symbol='SIRT7', Aliases='SIR2L7', description='sirtuin 7', other_designations='NAD-dependent deacetylase sirtuin-7|SIR2-like protein 7|regulatory protein SIR2 homolog 7|silent mating type information regulation 2, S.cerevisiae, homolog 7|sir2-related protein type 7|sirtuin type 7', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81911939, end_position_on_the_genomic_accession=81918182, orientation='minus', exon_count=12, OMIM=606212, no_hdr0=''),
  51548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51548, CurrentID=0, Status='live', Symbol='SIRT6', Aliases='SIR2L6', description='sirtuin 6', other_designations='NAD-dependent deacetylase sirtuin-6|SIR2-like protein 6|regulatory protein SIR2 homolog 6|sir2-related protein type 6|sirtuin type 6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4174109, end_position_on_the_genomic_accession=4182599, orientation='minus', exon_count=8, OMIM=606211, no_hdr0=''),
  51550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51550, CurrentID=0, Status='live', Symbol='CINP', Aliases='', description='cyclin-dependent kinase 2 interacting protein', other_designations='CDK2-interacting protein', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102348282, end_position_on_the_genomic_accession=102362916, orientation='minus', exon_count=6, OMIM=613362, no_hdr0=''),
  51552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51552, CurrentID=0, Status='live', Symbol='RAB14', Aliases='FBP, RAB-14', description='RAB14, member RAS oncogene family', other_designations='F protein-binding protein 1|bA165P4.3 (member RAS oncogene family)|small GTP binding protein RAB14', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121178137, end_position_on_the_genomic_accession=121202087, orientation='minus', exon_count=8, OMIM=612673, no_hdr0=''),
  51554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51554, CurrentID=0, Status='live', Symbol='ACKR4', Aliases='CC-CKR-11, CCBP2, CCR-11, CCR10, CCR11, CCRL1, CCX CKR, CCX-CKR, CKR-11, PPR1, VSHK1', description='atypical chemokine receptor 4', other_designations='C-C CKR-11|C-C chemokine receptor type 11|CC chemokine receptor-like 1|chemocentryx chemokine receptor|chemokine (C-C motif) receptor-like 1|chemokine, cc motif, receptor-like protein 1|orphan seven-transmembrane receptor, chemokine related', map_location='3q22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132597237, end_position_on_the_genomic_accession=132602644, orientation='plus', exon_count=2, OMIM=606065, no_hdr0=''),
  51555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51555, CurrentID=0, Status='live', Symbol='PEX5L', Aliases='PEX5R, PEX5RP, PXR2, PXR2B, TRIP8b', description='peroxisomal biogenesis factor 5-like', other_designations='HCN channel auxiliary subunit|PEX2-related protein|PEX5-like protein|Pex5p-related protein|peroxin-5-related protein|peroxisome biogenesis factor 5-like|tetratricopeptide repeat-containing Rab8b-interacting protein', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179794958, end_position_on_the_genomic_accession=180037053, orientation='minus', exon_count=22, OMIM=611058, no_hdr0=''),
  51557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51557, CurrentID=0, Status='live', Symbol='LGSN', Aliases='GLULD1, LGS', description='lengsin, lens protein with glutamine synthetase domain', other_designations='glutamate-ammonia ligase (glutamine synthase) domain containing 1|glutamate-ammonia ligase domain-containing protein 1', map_location='6pter-q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63275951, end_position_on_the_genomic_accession=63567921, orientation='minus', exon_count=14, OMIM=611470, no_hdr0=''),
  51559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51559, CurrentID=0, Status='live', Symbol='NT5DC3', Aliases='TU12B1-TY, TU12B1TY', description="5'-nucleotidase domain containing 3", other_designations='', map_location='12q22-q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103763312, end_position_on_the_genomic_accession=103841245, orientation='minus', exon_count=21, OMIM=611076, no_hdr0=''),
  51560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51560, CurrentID=0, Status='live', Symbol='RAB6B', Aliases='', description='RAB6B, member RAS oncogene family', other_designations='small GTP-binding protein|small GTPase RAB6B', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133824235, end_position_on_the_genomic_accession=133895847, orientation='minus', exon_count=9, OMIM=615852, no_hdr0=''),
  51561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51561, CurrentID=0, Status='live', Symbol='IL23A', Aliases='IL-23, IL-23A, IL23P19, P19, SGRF', description='interleukin 23 subunit alpha', other_designations='IL-23 subunit alpha|IL-23-A|IL-23p19|JKA3 induced upon T-cell activation|interleukin 23 p19 subunit|interleukin 23, alpha subunit p19|interleukin-23 subunit p19|interleukin-six, G-CSF related factor', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56334158, end_position_on_the_genomic_accession=56340410, orientation='plus', exon_count=6, OMIM=605580, no_hdr0=''),
  51562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51562, CurrentID=0, Status='live', Symbol='MBIP', Aliases='', description='MAP3K12 binding inhibitory protein 1', other_designations='MAPK upstream kinase-binding inhibitory protein|MUK-binding inhibitory protein', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36298558, end_position_on_the_genomic_accession=36320676, orientation='minus', exon_count=10, OMIM=609431, no_hdr0=''),
  51564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51564, CurrentID=0, Status='live', Symbol='HDAC7', Aliases='HD7, HD7AA, HDAC7', description='histone deacetylase 7', other_designations='histone deacetylase 7A', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47782722, end_position_on_the_genomic_accession=47820612, orientation='minus', exon_count=32, OMIM=606542, no_hdr0=''),
  51566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51566, CurrentID=0, Status='live', Symbol='ARMCX3', Aliases='ALEX3, GASP6, dJ545K15.2', description='armadillo repeat containing, X-linked 3', other_designations='1200004E24Rik|ARM protein lost in epithelial cancers on chromosome X 3|arm protein lost in epithelial cancers, X chromosome, 3', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101623127, end_position_on_the_genomic_accession=101627843, orientation='plus', exon_count=5, OMIM=300364, no_hdr0=''),
  51567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51567, CurrentID=0, Status='live', Symbol='TDP2', Aliases='AD022, EAP2, EAPII, TTRAP, dJ30M3.3, hTDP2', description='tyrosyl-DNA phosphodiesterase 2', other_designations="5'-Tyr-DNA phosphodiesterase|5'-tyrosyl-DNA phosphodiesterase|ETS1-associated protein 2|ETS1-associated protein II|TRAF and TNF receptor associated protein|TRAF and TNF receptor-associated protein|VPg unlinkase|tyr-DNA phosphodiesterase 2|tyrosyl-RNA phosphodiesterase", map_location='6p22.3-p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24649977, end_position_on_the_genomic_accession=24666989, orientation='minus', exon_count=8, OMIM=605764, no_hdr0=''),
  51569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51569, CurrentID=0, Status='live', Symbol='UFM1', Aliases='BM-002, C13orf20', description='ubiquitin-fold modifier 1', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38349771, end_position_on_the_genomic_accession=38363007, orientation='plus', exon_count=8, OMIM=610553, no_hdr0=''),
  51571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51571, CurrentID=0, Status='live', Symbol='FAM49B', Aliases='BM-009, L1', description='family with sequence similarity 49 member B', other_designations='family with sequence similarity 49, member B', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129839593, end_position_on_the_genomic_accession=130016697, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  51573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51573, CurrentID=0, Status='live', Symbol='GDE1', Aliases='363E6.2, MIR16', description='glycerophosphodiester phosphodiesterase 1', other_designations='RGS16-interacting membrane protein|membrane interacting protein of RGS16|membrane-interacting protein of RGS16', map_location='16p12-p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19501693, end_position_on_the_genomic_accession=19522128, orientation='minus', exon_count=6, OMIM=605943, no_hdr0=''),
  51574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51574, CurrentID=0, Status='live', Symbol='LARP7', Aliases='ALAZS, HDCMA18P, PIP7S', description='La ribonucleoprotein domain family member 7', other_designations='La ribonucleoprotein domain family, member 7|P-TEFb-interaction protein for 7SK stability', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112636964, end_position_on_the_genomic_accession=112657592, orientation='plus', exon_count=16, OMIM=612026, no_hdr0=''),
  51575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51575, CurrentID=0, Status='live', Symbol='ESF1', Aliases='ABTAP, C20orf6, HDCMC28P, bA526K24.1', description='ESF1 nucleolar pre-rRNA processing protein homolog', other_designations='ABT1-associated protein|ESF1, nucleolar pre-rRNA processing protein, homolog', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13714322, end_position_on_the_genomic_accession=13784933, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  51581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51581, CurrentID=0, Status='live', Symbol='ERVH-6', Aliases='', description='endogenous retrovirus group H member 6', other_designations='endogenous retrovirus group H, member 6', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51582, CurrentID=0, Status='live', Symbol='AZIN1', Aliases='AZI, AZIA1, OAZI, OAZIN, ODC1L', description='antizyme inhibitor 1', other_designations='ornithine decarboxylase antizyme inhibitor', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102826302, end_position_on_the_genomic_accession=102864200, orientation='minus', exon_count=13, OMIM=607909, no_hdr0=''),
  51585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51585, CurrentID=0, Status='live', Symbol='PCF11', Aliases='', description='PCF11 cleavage and polyadenylation factor subunit', other_designations='PCF11, cleavage and polyadenylation factor II subunit, homolog|PCF11, cleavage and polyadenylation factor subunit, homolog|PCF11p homolog|pre-mRNA cleavage complex II protein Pcf11', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83157095, end_position_on_the_genomic_accession=83185794, orientation='plus', exon_count=16, OMIM=608876, no_hdr0=''),
  51586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51586, CurrentID=0, Status='live', Symbol='MED15', Aliases='ARC105, CAG7A, CTG7A, PCQAP, TIG-1, TIG1, TNRC7', description='mediator complex subunit 15', other_designations='CTG repeat protein 7a|PC2 (positive cofactor 2, multiprotein complex) glutamine/Q-rich-associated protein|PC2 glutamine/Q-rich-associated protein|PC2-glutamine-rich-associated protein|TPA inducible gene-1|TPA inducible protein|TPA-inducible gene 1 protein|activator-recruited cofactor 105 kDa component|activator-recruited cofactor, 105-kD|positive cofactor 2 glutamine/Q-rich-associated protein|positive cofactor 2, glutamine/Q-rich-associated protein|trinucleotide repeat containing 7|trinucleotide repeat-containing gene 7 protein', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20507542, end_position_on_the_genomic_accession=20587632, orientation='plus', exon_count=25, OMIM=607372, no_hdr0=''),
  51588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51588, CurrentID=0, Status='live', Symbol='PIAS4', Aliases='PIAS-gamma, PIASY, Piasg, ZMIZ6', description='protein inhibitor of activated STAT 4', other_designations='protein inhibitor of activated STAT protein 4|protein inhibitor of activated STAT protein PIASy|protein inhibitor of activated STAT protein gamma|protein inhibitor of activated STAT, 4|zinc finger, MIZ-type containing 6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4007598, end_position_on_the_genomic_accession=4039386, orientation='plus', exon_count=12, OMIM=605989, no_hdr0=''),
  51592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51592, CurrentID=0, Status='live', Symbol='TRIM33', Aliases='ECTO, PTC7, RFG7, TF1G, TIF1G, TIF1GAMMA, TIFGAMMA', description='tripartite motif containing 33', other_designations='RET-fused gene 7 protein|TIF1-gamma|ectodermin homolog|protein Rfg7|transcriptional intermediary factor 1 gamma|tripartite motif-containing 33', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114392777, end_position_on_the_genomic_accession=114511160, orientation='minus', exon_count=21, OMIM=605769, no_hdr0=''),
  51593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51593, CurrentID=0, Status='live', Symbol='SRRT', Aliases='ARS2, ASR2, serrate', description='serrate, RNA effector molecule', other_designations='arsenate resistance protein 2|arsenate resistance protein ARS2|arsenite resistance protein|arsenite-resistance protein 2', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100874971, end_position_on_the_genomic_accession=100888664, orientation='plus', exon_count=23, OMIM=614469, no_hdr0=''),
  51594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51594, CurrentID=0, Status='live', Symbol='NBAS', Aliases='NAG, SOPH', description='neuroblastoma amplified sequence', other_designations='NAG/BC035112 fusion|NAG/FAM49A fusion|neuroblastoma-amplified gene protein', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15166908, end_position_on_the_genomic_accession=15561348, orientation='minus', exon_count=53, OMIM=608025, no_hdr0=''),
  51596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51596, CurrentID=0, Status='live', Symbol='CUTA', Aliases='ACHAP, C6orf82', description='cutA divalent cation tolerance homolog (E. coli)', other_designations='acetylcholinesterase-associated protein|brain acetylcholinesterase putative membrane anchor|divalent cation tolerant protein CUTA', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33416542, end_position_on_the_genomic_accession=33418288, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  51599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51599, CurrentID=0, Status='live', Symbol='LSR', Aliases='ILDR3, LISCH7', description='lipolysis stimulated lipoprotein receptor', other_designations='LISCH protein|immunoglobulin-like domain containing receptor 3|lipolysis-stimulated remnant|liver-specific bHLH-Zip transcription factor', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35248656, end_position_on_the_genomic_accession=35267964, orientation='plus', exon_count=10, OMIM=616582, no_hdr0=''),
  51601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51601, CurrentID=0, Status='live', Symbol='LIPT1', Aliases='LIPT1D', description='lipoyltransferase 1', other_designations='lipoate biosynthesis protein|lipoyl ligase', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99154955, end_position_on_the_genomic_accession=99163157, orientation='plus', exon_count=7, OMIM=610284, no_hdr0=''),
  51602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51602, CurrentID=0, Status='live', Symbol='NOP58', Aliases='HSPC120, NOP5, NOP5/NOP58', description='NOP58 ribonucleoprotein', other_designations='NOP58 ribonucleoprotein homolog|nucleolar protein 5|nucleolar protein NOP5/NOP58', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202265792, end_position_on_the_genomic_accession=202303661, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  51603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51603, CurrentID=0, Status='live', Symbol='METTL13', Aliases='5630401D24Rik, CGI-01, KIAA0859, feat', description='methyltransferase like 13', other_designations='antiapoptotic protein FEAT', map_location='1q24-q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171781621, end_position_on_the_genomic_accession=171797716, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51604, CurrentID=0, Status='live', Symbol='PIGT', Aliases='CGI-06, MCAHS3, NDAP, PNH2', description='phosphatidylinositol glycan anchor biosynthesis class T', other_designations='GPI transamidase subunit|neurotrophin-regulated neuronal development-associated protein|phosphatidylinositol glycan anchor biosynthesis, class T|phosphatidylinositol-glycan biosynthesis class T protein', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45416067, end_position_on_the_genomic_accession=45426245, orientation='plus', exon_count=12, OMIM=610272, no_hdr0=''),
  51605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51605, CurrentID=0, Status='live', Symbol='TRMT6', Aliases='CGI-09, GCD10, Gcd10p', description='tRNA methyltransferase 6', other_designations='tRNA methyltransferase 6 homolog|tRNA(m1A58)MTase subunit TRM6', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5937833, end_position_on_the_genomic_accession=5950558, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  51606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51606, CurrentID=0, Status='live', Symbol='ATP6V1H', Aliases='CGI-11, MSTP042, NBP1, SFD, SFDalpha, SFDbeta, VMA13', description='ATPase, H+ transporting, lysosomal 50/57kDa, V1 subunit H', other_designations='V-ATPase 50/57 kDa subunits|V-ATPase subunit H|nef-binding protein 1|protein VMA13 homolog|vacuolar ATP synthase subunit H|vacuolar ATPase subunit H|vacuolar proton pump subunit H|vacuolar proton pump subunit SFD', map_location='8q11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53715543, end_position_on_the_genomic_accession=53843311, orientation='minus', exon_count=16, OMIM=608861, no_hdr0=''),
  401817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401817, CurrentID=0, Status='live', Symbol='RPS10P22', Aliases='RPS10_12_1442', description='ribosomal protein S10 pseudogene 22', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60404419, end_position_on_the_genomic_accession=60405006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  51608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51608, CurrentID=0, Status='live', Symbol='GET4', Aliases='C7orf20, CEE, CGI-20, TRC35', description='golgi to ER traffic protein 4', other_designations='H_NH1244M04.5|conserved edge expressed protein|conserved edge protein|conserved edge-expressed protein|transmembrane domain recognition complex 35 kDa subunit|transmembrane domain recognition complex, 35kDa', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=876554, end_position_on_the_genomic_accession=896434, orientation='plus', exon_count=9, OMIM=612056, no_hdr0=''),
  51611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51611, CurrentID=0, Status='live', Symbol='DPH5', Aliases='AD-018, CGI-30, HSPC143, NPD015', description='diphthamide biosynthesis 5', other_designations='DPH5 homolog|diphthamide biosynthesis methyltransferase|diphthine synthase|protein x 0011', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100989623, end_position_on_the_genomic_accession=101025806, orientation='minus', exon_count=8, OMIM=611075, no_hdr0=''),
  100132543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132543, CurrentID=0, Status='live', Symbol='LOC100132543', Aliases='', description='family with sequence similarity 98, member B pseudogene', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31953853, end_position_on_the_genomic_accession=31955643, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  51614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51614, CurrentID=0, Status='live', Symbol='ERGIC3', Aliases='C20orf47, CGI-54, Erv46, NY-BR-84, PRO0989, SDBCAG84, dJ477O4.2', description='ERGIC and golgi 3', other_designations='endoplasmic reticulum-localized protein ERp43|serologically defined breast cancer antigen 84|serologically defined breast cancer antigen NY-BR-84', map_location='20pter-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35542029, end_position_on_the_genomic_accession=35557634, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  51616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51616, CurrentID=0, Status='live', Symbol='TAF9B', Aliases='DN-7, DN7, TAF9L, TAFII31L, TFIID-31', description='TATA-box binding protein associated factor 9b', other_designations='TAF9-like RNA polymerase II, TATA box binding protein (TBP)-associated factor, 31kDa|TAF9B RNA polymerase II, TATA box binding protein (TBP)-associated factor, 31kDa|TATA box binding protein associated factor 9b|TBP-associated factor 9L|neuronal cell death-related protein 7|transcription associated factor TAFII31L|transcription initiation factor IID, 31kD subunit|transcription-associated factor TAFII31L', map_location='Xq13.1-q21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78129748, end_position_on_the_genomic_accession=78139682, orientation='minus', exon_count=7, OMIM=300754, no_hdr0=''),
  51617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51617, CurrentID=0, Status='live', Symbol='HMP19', Aliases='', description='HMP19 protein', other_designations='NSG2|hypothalamus golgi apparatus expressed 19 kDa protein|p19 protein|protein p19', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174045604, end_position_on_the_genomic_accession=174109179, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51619, CurrentID=0, Status='live', Symbol='UBE2D4', Aliases='HBUCE1', description='ubiquitin conjugating enzyme E2D 4 (putative)', other_designations='E2 ubiquitin-conjugating enzyme D4|ubiquitin carrier protein D4|ubiquitin-conjugating enzyme E2D 4 (putative)|ubiquitin-conjugating enzyme HBUCE1|ubiquitin-protein ligase D4', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43926436, end_position_on_the_genomic_accession=43956136, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  51621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51621, CurrentID=0, Status='live', Symbol='KLF13', Aliases='BTEB3, FKLF2, NSLP1, RFLAT-1, RFLAT1', description='Kruppel-like factor 13', other_designations='BTE-binding protein 3|RANTES factor of late activated T lymphocytes-1|basic transcription element binding protein 3|novel Sp1-like zinc finger transcription factor 1|transcription factor BTEB3|transcription factor NSLP1', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31326855, end_position_on_the_genomic_accession=31435665, orientation='plus', exon_count=3, OMIM=605328, no_hdr0=''),
  51622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51622, CurrentID=0, Status='live', Symbol='CCZ1', Aliases='C7orf28AA, CGI-43, H_DJ1163J12.2, CCZ1', description='CCZ1 homolog, vacuolar protein trafficking and biogenesis associated', other_designations='CCZ1 vacuolar protein trafficking and biogenesis associated homolog|H_NH0577018.2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5898710, end_position_on_the_genomic_accession=5925976, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  51626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51626, CurrentID=0, Status='live', Symbol='DYNC2LI1', Aliases='CGI-60, D2LIC, LIC3', description='dynein, cytoplasmic 2, light intermediate chain 1', other_designations='dynein 2 light intermediate chain', map_location='2p25.1-p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43774039, end_position_on_the_genomic_accession=43827950, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  51629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51629, CurrentID=0, Status='live', Symbol='SLC25A39', Aliases='CGI-69, CGI69', description='solute carrier family 25 member 39', other_designations='solute carrier family 25, member 39', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44319625, end_position_on_the_genomic_accession=44324857, orientation='minus', exon_count=12, OMIM=610820, no_hdr0=''),
  51631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51631, CurrentID=0, Status='live', Symbol='LUC7L2', Aliases='CGI-59, CGI-74, LUC7B2', description='LUC7-like 2 pre-mRNA splicing factor', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139340359, end_position_on_the_genomic_accession=139423457, orientation='plus', exon_count=13, OMIM=613056, no_hdr0=''),
  51633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51633, CurrentID=0, Status='live', Symbol='OTUD6B', Aliases='CGI-77, DUBA-5, DUBA5', description='OTU domain containing 6B', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91070196, end_position_on_the_genomic_accession=91087095, orientation='plus', exon_count=8, OMIM=612021, no_hdr0=''),
  51634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51634, CurrentID=0, Status='live', Symbol='RBMX2', Aliases='CGI-79, Snu17', description='RNA binding motif protein, X-linked 2', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130401969, end_position_on_the_genomic_accession=130413343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  51635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51635, CurrentID=0, Status='live', Symbol='DHRS7', Aliases='CGI-86, SDR34C1, retDSR4, retSDR4', description='dehydrogenase/reductase (SDR family) member 7', other_designations='retinal short-chain dehydrogenase/reductase 4|short chain dehydrogenase/reductase family 34C member 1|short chain dehydrogenase/reductase family 34C, member 1', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60144778, end_position_on_the_genomic_accession=60165493, orientation='minus', exon_count=7, OMIM=612833, no_hdr0=''),
  51637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51637, CurrentID=0, Status='live', Symbol='C14orf166', Aliases='CGI-99, CGI99, CLE, CLE7, LCRP369, RLLM1', description='chromosome 14 open reading frame 166', other_designations='CLE7 homolog|RLL motif containing 1', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51989510, end_position_on_the_genomic_accession=52004702, orientation='plus', exon_count=8, OMIM=610858, no_hdr0=''),
  51639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51639, CurrentID=0, Status='live', Symbol='SF3B6', Aliases='CGI-110, HSPC175, Ht006, P14, SAP14, SAP14a, SF3B14, SF3B14a', description='splicing factor 3b subunit 6', other_designations='SF3b 14 kDa subunit|pre mRNA branch site protein p14|pre-mRNA branch site protein p14|spliceosome-associated protein, 14 kDa subunit|spliceosome-associated protein, 14-kDa|splicing factor 3B, 14 kDa subunit|splicing factor 3b, subunit 6, 14kDa', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24067584, end_position_on_the_genomic_accession=24076444, orientation='minus', exon_count=4, OMIM=607835, no_hdr0=''),
  51642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51642, CurrentID=0, Status='live', Symbol='MRPL48', Aliases='CGI-118, HSPC290, L48MT, MRP-L48', description='mitochondrial ribosomal protein L48', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73787872, end_position_on_the_genomic_accession=73864611, orientation='plus', exon_count=9, OMIM=611853, no_hdr0=''),
  51643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51643, CurrentID=0, Status='live', Symbol='TMBIM4', Aliases='CGI-119, GAAP, LFG4, S1R, ZPRO', description='transmembrane BAX inhibitor motif containing 4', other_designations='Golgi anti-apoptotic protein|Z-protein|transmembrane BAX inhibitor motif-containing protein 4', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66136936, end_position_on_the_genomic_accession=66170072, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  51645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51645, CurrentID=0, Status='live', Symbol='PPIL1', Aliases='CGI-124, CYPL1, PPIase, hCyPX', description='peptidylprolyl isomerase like 1', other_designations='cyclophilin-related gene 1|peptidyl-prolyl cis-trans isomerase|peptidylprolyl isomerase (cyclophilin)-like 1|rotamase PPIL1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36854829, end_position_on_the_genomic_accession=36875024, orientation='minus', exon_count=4, OMIM=601301, no_hdr0=''),
  51646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51646, CurrentID=0, Status='live', Symbol='YPEL5', Aliases='CGI-127', description='yippee like 5', other_designations='yippee-like 5', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30146884, end_position_on_the_genomic_accession=30160533, orientation='plus', exon_count=5, OMIM=609726, no_hdr0=''),
  51647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51647, CurrentID=0, Status='live', Symbol='FAM96B', Aliases='CGI-128, MIP18', description='family with sequence similarity 96 member B', other_designations='MSS19-interacting protein of 18 kDa|family with sequence similarity 96, member B', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66932055, end_position_on_the_genomic_accession=66934423, orientation='minus', exon_count=5, OMIM=614778, no_hdr0=''),
  51649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51649, CurrentID=0, Status='live', Symbol='MRPS23', Aliases='CGI-138, HSPC329, MRP-S23', description='mitochondrial ribosomal protein S23', other_designations='S23mt', map_location='17q22-q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57838926, end_position_on_the_genomic_accession=57850072, orientation='minus', exon_count=5, OMIM=611985, no_hdr0=''),
  51650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51650, CurrentID=0, Status='live', Symbol='MRPS33', Aliases='CGI-139, MRP-S33, PTD003, S33mt', description='mitochondrial ribosomal protein S33', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141006161, end_position_on_the_genomic_accession=141014981, orientation='minus', exon_count=4, OMIM=611993, no_hdr0=''),
  51651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51651, CurrentID=0, Status='live', Symbol='PTRH2', Aliases='BIT1, CFAP37, CGI-147, IMNEPD, PTH, PTH 2, PTH2', description='peptidyl-tRNA hydrolase 2', other_designations='bcl-2 inhibitor of transcription 1|cilia and flagella associated protein 37', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59697306, end_position_on_the_genomic_accession=59707495, orientation='minus', exon_count=3, OMIM=608625, no_hdr0=''),
  51652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51652, CurrentID=0, Status='live', Symbol='CHMP3', Aliases='CGI-149, NEDF, VPS24', description='charged multivesicular body protein 3', other_designations='25.1 protein|CHMP family, member 3|chromatin-modifying protein 3|neuroendocrine differentiation factor|vacuolar protein sorting 24 homolog|vacuolar protein sorting-associated protein 24', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86503430, end_position_on_the_genomic_accession=86563497, orientation='minus', exon_count=6, OMIM=610052, no_hdr0=''),
  51654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51654, CurrentID=0, Status='live', Symbol='CDK5RAP1', Aliases='C20orf34, C42, CGI-05, HSPC167', description='CDK5 regulatory subunit associated protein 1', other_designations='CDK5 activator-binding protein C42', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33358839, end_position_on_the_genomic_accession=33401902, orientation='minus', exon_count=15, OMIM=608200, no_hdr0=''),
  51655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51655, CurrentID=0, Status='live', Symbol='RASD1', Aliases='AGS1, DEXRAS1, MGC:26290', description='RAS, dexamethasone-induced 1', other_designations='activator of G-protein signaling 1|ras-related protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17494437, end_position_on_the_genomic_accession=17496395, orientation='minus', exon_count=2, OMIM=605550, no_hdr0=''),
  51657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51657, CurrentID=0, Status='live', Symbol='STYXL1', Aliases='DUSP24, MK-STYX, MKSTYX', description='serine/threonine/tyrosine interacting-like 1', other_designations='dual specificity phosphatase 24 (putative)|dual specificity phosphatase inhibitor MK-STYX|dual specificity protein phosphatase 24|map kinase phosphatase-like protein MK-STYX', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75995834, end_position_on_the_genomic_accession=76048003, orientation='minus', exon_count=10, OMIM=616695, no_hdr0=''),
  51659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51659, CurrentID=0, Status='live', Symbol='GINS2', Aliases='HSPC037, PSF2, Pfs2', description='GINS complex subunit 2 (Psf2 homolog)', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85677674, end_position_on_the_genomic_accession=85688982, orientation='minus', exon_count=5, OMIM=610609, no_hdr0=''),
  51660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51660, CurrentID=0, Status='live', Symbol='MPC1', Aliases='BRP44L, CGI-129, MPYCD, dJ68L15.3', description='mitochondrial pyruvate carrier 1', other_designations='HSPC040 protein|brain protein 44-like protein', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166364919, end_position_on_the_genomic_accession=166383013, orientation='minus', exon_count=7, OMIM=614738, no_hdr0=''),
  51661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51661, CurrentID=0, Status='live', Symbol='FKBP7', Aliases='FKBP23, PPIase', description='FK506 binding protein 7', other_designations='23 kDa FK506-binding protein|23 kDa FKBP|FK506-binding protein 23|FK506-binding protein 7|FKBP-23|FKBP-7|PPIase FKBP7|rotamase', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178463664, end_position_on_the_genomic_accession=178480393, orientation='minus', exon_count=7, OMIM=607062, no_hdr0=''),
  51663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51663, CurrentID=0, Status='live', Symbol='ZFR', Aliases='SPG711, ZFR', description='zinc finger RNA binding protein', other_designations='M-phase phosphoprotein homolog|hZFR', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32354350, end_position_on_the_genomic_accession=32444761, orientation='minus', exon_count=20, OMIM=615635, no_hdr0=''),
  51665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51665, CurrentID=0, Status='live', Symbol='ASB1', Aliases='ASB-1', description='ankyrin repeat and SOCS box containing 1', other_designations='ankyrin repeat and SOCS box-containing 1', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238426741, end_position_on_the_genomic_accession=238452250, orientation='plus', exon_count=6, OMIM=605758, no_hdr0=''),
  51666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51666, CurrentID=0, Status='live', Symbol='ASB4', Aliases='ASB-4', description='ankyrin repeat and SOCS box containing 4', other_designations='ankyrin repeat and SOCS box-containing 4', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95461542, end_position_on_the_genomic_accession=95540231, orientation='plus', exon_count=7, OMIM=605761, no_hdr0=''),
  401827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401827, CurrentID=0, Status='live', Symbol='MSLNL', Aliases='C16orf37, MPFL', description='mesothelin-like', other_designations='MPF-like|megakarycyte potentiating factor|pre-pro-megakaryocyte-potentiating-factor-like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=769428, end_position_on_the_genomic_accession=781996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  51668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51668, CurrentID=0, Status='live', Symbol='HSPB11', Aliases='C1orf41, HSPCO34, IFT25, PP25', description='heat shock protein family B (small) member 11', other_designations='heat shock protein beta-11|heat shock protein family B (small), member 11|intraflagellar transport 25 homolog|placental protein 25', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53921561, end_position_on_the_genomic_accession=53946303, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51669, CurrentID=0, Status='live', Symbol='SARAF', Aliases='FOAP-7, HSPC035, TMEM66, XTP3', description='store-operated calcium entry-associated regulatory factor', other_designations='HBV X-transactivated gene 3 protein|HBV XAg-transactivated protein 3|SARAF long isoform|SARAF short isoform|SOCE-associated regulatory factor|transmembrane protein 66', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30063012, end_position_on_the_genomic_accession=30083208, orientation='minus', exon_count=6, OMIM=614768, no_hdr0=''),
  51673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51673, CurrentID=0, Status='live', Symbol='TPPP3', Aliases='CGI-38, TPPP/p20, p20, p25gamma', description='tubulin polymerization-promoting protein family member 3', other_designations='brain specific protein', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67389807, end_position_on_the_genomic_accession=67393535, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  51676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51676, CurrentID=0, Status='live', Symbol='ASB2', Aliases='ASB-2', description='ankyrin repeat and SOCS box containing 2', other_designations='ankyrin repeat and SOCS box-containing protein 2a', map_location='14q31-q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93934153, end_position_on_the_genomic_accession=93976786, orientation='minus', exon_count=12, OMIM=605759, no_hdr0=''),
  51678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51678, CurrentID=0, Status='live', Symbol='MPP6', Aliases='PALS2, VAM-1, VAM1, p55T', description='membrane protein, palmitoylated 6', other_designations='MAGUK protein p55T|VELI-associated MAGUK 1|membrane protein, palmitoylated 6 (MAGUK p55 subfamily member 6)|protein associated with Lin7 2', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24573346, end_position_on_the_genomic_accession=24693703, orientation='plus', exon_count=15, OMIM=606959, no_hdr0=''),
  51684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51684, CurrentID=0, Status='live', Symbol='SUFU', Aliases='PRO1280H, SUFUXL, SUFU', description='suppressor of fused homolog (Drosophila)', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102502874, end_position_on_the_genomic_accession=102633457, orientation='plus', exon_count=16, OMIM=607035, no_hdr0=''),
  51686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51686, CurrentID=0, Status='live', Symbol='OAZ3', Aliases='AZ3, OAZ-t, TISP15', description='ornithine decarboxylase antizyme 3', other_designations='ODC-Az 3|antizyme 3', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151762969, end_position_on_the_genomic_accession=151771330, orientation='plus', exon_count=7, OMIM=605138, no_hdr0=''),
  51690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51690, CurrentID=0, Status='live', Symbol='LSM7', Aliases='YNL147W', description='LSM7 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM7 U6 small nuclear RNA and mRNA degradation associated|LSM7 homolog, U6 small nuclear RNA associated', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2321521, end_position_on_the_genomic_accession=2328615, orientation='minus', exon_count=5, OMIM=607287, no_hdr0=''),
  51691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51691, CurrentID=0, Status='live', Symbol='LSM8', Aliases='NAA38', description='LSM8 homolog, U6 small nuclear RNA associated', other_designations='LSM8 U6 small nuclear RNA associated|MAK31-like protein|N-alpha-acetyltransferase 38, NatC auxiliary subunit|U6 snRNA-associated Sm-like protein LSm8', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118184032, end_position_on_the_genomic_accession=118204039, orientation='plus', exon_count=4, OMIM=607288, no_hdr0=''),
  51692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51692, CurrentID=0, Status='live', Symbol='CPSF3', Aliases='CPSF-73, CPSF73', description='cleavage and polyadenylation specific factor 3', other_designations="cleavage and polyadenylation specific factor 3, 73kDa|mRNA 3'-end-processing endonuclease CPSF-73", map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9423537, end_position_on_the_genomic_accession=9473110, orientation='plus', exon_count=19, OMIM=606029, no_hdr0=''),
  51693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51693, CurrentID=0, Status='live', Symbol='TRAPPC2L', Aliases='HSPC176', description='trafficking protein particle complex 2-like', other_designations='hematopoietic stem/progenitor cells 176', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88857098, end_position_on_the_genomic_accession=88862678, orientation='plus', exon_count=5, OMIM=610970, no_hdr0=''),
  285415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285415, CurrentID=0, Status='live', Symbol='MRPL42P1', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119195266, end_position_on_the_genomic_accession=119195691, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  51696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51696, CurrentID=0, Status='live', Symbol='HECA', Aliases='HDC, HDCL, HHDC, dJ225E12.1', description='hdc homolog, cell cycle regulator', other_designations='headcase homolog', map_location='6q23-q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139135112, end_position_on_the_genomic_accession=139180809, orientation='plus', exon_count=4, OMIM=607977, no_hdr0=''),
  51699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51699, CurrentID=0, Status='live', Symbol='VPS29', Aliases='DC15, DC7, PEP11', description='VPS29 retromer complex component', other_designations='PEP11 homolog|hVPS29|retromer protein|vacuolar protein sorting 29 homolog|vacuolar sorting protein VPS29/PEP11|vesicle protein sorting 29|x 007 protein', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110491523, end_position_on_the_genomic_accession=110502142, orientation='minus', exon_count=8, OMIM=606932, no_hdr0=''),
  51700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51700, CurrentID=0, Status='live', Symbol='CYB5R2', Aliases='B5R.2', description='cytochrome b5 reductase 2', other_designations='cytochrome b5 reductase b5R.2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7665095, end_position_on_the_genomic_accession=7674341, orientation='minus', exon_count=12, OMIM=608342, no_hdr0=''),
  51701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51701, CurrentID=0, Status='live', Symbol='NLK', Aliases='', description='nemo-like kinase', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28042647, end_position_on_the_genomic_accession=28205140, orientation='plus', exon_count=13, OMIM=609476, no_hdr0=''),
  51702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51702, CurrentID=0, Status='live', Symbol='PADI3', Aliases='PAD3, PDI3', description='peptidyl arginine deiminase, type III', other_designations='peptidylarginine deiminase III|protein-arginine deiminase type III', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17249082, end_position_on_the_genomic_accession=17284233, orientation='plus', exon_count=18, OMIM=606755, no_hdr0=''),
  51703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51703, CurrentID=0, Status='live', Symbol='ACSL5', Aliases='ACS2, ACS5, FACL5', description='acyl-CoA synthetase long-chain family member 5', other_designations='FACL5 for fatty acid coenzyme A ligase 5|LACS 5|fatty acid coenzyme A ligase 5|fatty-acid-Coenzyme A ligase, long-chain 5|long-chain acyl-CoA synthetase 5|long-chain fatty acid coenzyme A ligase 5', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112374158, end_position_on_the_genomic_accession=112428380, orientation='plus', exon_count=23, OMIM=605677, no_hdr0=''),
  51704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51704, CurrentID=0, Status='live', Symbol='GPRC5B', Aliases='RAIG-2, RAIG2', description='G protein-coupled receptor, class C, group 5, member B', other_designations='G protein-coupled receptor, family C, group 1, member B|retinoic acid responsive gene protein|retinoic acid-induced gene 2 protein', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19858971, end_position_on_the_genomic_accession=19886045, orientation='minus', exon_count=7, OMIM=605948, no_hdr0=''),
  51705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51705, CurrentID=0, Status='live', Symbol='EMCN', Aliases='EMCN2, MUC14', description='endomucin', other_designations='MUC-14|endomucin-2|gastric cancer antigen Ga34|mucin-14', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100395341, end_position_on_the_genomic_accession=100518093, orientation='minus', exon_count=13, OMIM=608350, no_hdr0=''),
  51706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51706, CurrentID=0, Status='live', Symbol='CYB5R1', Aliases='B5R.1, B5R1, B5R2, NQO3A2, humb5R2', description='cytochrome b5 reductase 1', other_designations='NAD(P)H:quinone oxidoreductase type 3, polypeptide A2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202961873, end_position_on_the_genomic_accession=202967276, orientation='minus', exon_count=9, OMIM=608341, no_hdr0=''),
  51710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51710, CurrentID=0, Status='live', Symbol='ZNF44', Aliases='GIOT-2, KOX7, ZNF, ZNF504, ZNF55, ZNF58', description='zinc finger protein 44', other_designations='gonadotropin inducible transcription repressor-2|gonadotropin-inducible ovary transcription repressor 2|gonadotropin-inducible transcription repressor 2|zinc finger protein 55|zinc finger protein 58|zinc finger protein KOX7|zinc finger protein ZnFP12', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12247244, end_position_on_the_genomic_accession=12294899, orientation='minus', exon_count=10, OMIM=194542, no_hdr0=''),
  51714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51714, CurrentID=0, Status='live', Symbol='SELT', Aliases='', description='selenoprotein T', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150603279, end_position_on_the_genomic_accession=150630447, orientation='plus', exon_count=6, OMIM=607912, no_hdr0=''),
  51715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51715, CurrentID=0, Status='live', Symbol='RAB23', Aliases='HSPC137', description='RAB23, member RAS oncogene family', other_designations='RAB family small GTP binding protein RAB 23', map_location='6p11', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57186992, end_position_on_the_genomic_accession=57222314, orientation='minus', exon_count=8, OMIM=606144, no_hdr0=''),
  51716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51716, CurrentID=0, Status='live', Symbol='CES1P1', Aliases='CES1A2, CES1A3, CES4, CESR, PCE-3', description='carboxylesterase 1 pseudogene 1', other_designations='carboxylesterase 4, pseudogene|carboxylesterase-related protein', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55760599, end_position_on_the_genomic_accession=55774926, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  51719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51719, CurrentID=0, Status='live', Symbol='CAB39', Aliases='CGI-66, MO25', description='calcium binding protein 39', other_designations='MO25alpha', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230712842, end_position_on_the_genomic_accession=230821075, orientation='plus', exon_count=11, OMIM=612174, no_hdr0=''),
  51720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51720, CurrentID=0, Status='live', Symbol='UIMC1', Aliases='RAP80, X2HRIP110', description='ubiquitin interaction motif containing 1', other_designations='receptor-associated protein 80|retinoid X receptor-interacting protein 110', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176905005, end_position_on_the_genomic_accession=177022571, orientation='minus', exon_count=20, OMIM=609433, no_hdr0=''),
  51725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51725, CurrentID=0, Status='live', Symbol='FBXO40', Aliases='FBX40', description='F-box protein 40', other_designations='muscle disease-related protein', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121593323, end_position_on_the_genomic_accession=121630292, orientation='plus', exon_count=4, OMIM=609107, no_hdr0=''),
  51726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51726, CurrentID=0, Status='live', Symbol='DNAJB11', Aliases='ABBP-2, ABBP2, DJ9, Dj-9, EDJ, ERdj3, ERj3, ERj3p, PRO1080, UNQ537', description='DnaJ heat shock protein family (Hsp40) member B11', other_designations='APOBEC1-binding protein 2|DnaJ (Hsp40) homolog, subfamily B, member 11|DnaJ protein 9|ER-associated DNAJ protein 3|ER-associated Hsp40 co-chaperone|ER-resident protein ERdj3|PWP1-interacting protein 4|endoplasmic reticulum DNA J domain-containing protein 3', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186570676, end_position_on_the_genomic_accession=186585800, orientation='plus', exon_count=10, OMIM=611341, no_hdr0=''),
  51727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51727, CurrentID=0, Status='live', Symbol='CMPK1', Aliases='CK, CMK, CMPK, UMK, UMP-CMPK, UMPK', description='cytidine monophosphate (UMP-CMP) kinase 1, cytosolic', other_designations='CK|UMP/CMP kinase|cytidylate kinase|dCMP kinase|deoxycytidylate kinase|nucleoside-diphosphate kinase|uridine monophosphate kinase|uridine monophosphate/cytidine monophosphate kinase', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47333797, end_position_on_the_genomic_accession=47378839, orientation='plus', exon_count=7, OMIM=191710, no_hdr0=''),
  51728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51728, CurrentID=0, Status='live', Symbol='POLR3K', Aliases='C11, C11-RNP3, My010, RPC10, RPC11, RPC12.5', description='polymerase (RNA) III (DNA directed) polypeptide K, 12.3 kDa', other_designations='DNA-directed RNA polymerase III subunit K|DNA-directed RNA polymerases III 12.5 kDa polypeptide|RNA polymerase III 12.5 kDa subunit|RNA polymerase III subunit C10|RNA polymerase III subunit C11|RNA polymerase III subunit CII', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46979, end_position_on_the_genomic_accession=53632, orientation='minus', exon_count=3, OMIM=606007, no_hdr0=''),
  51729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51729, CurrentID=0, Status='live', Symbol='WBP11', Aliases='NPWBP, PPP1R165, SIPP1, WBP-11', description='WW domain binding protein 11', other_designations='Npw38-binding protein NpwBP|SH3 domain-binding protein SNP70|protein phosphatase 1, regulatory subunit 165|splicing factor that interacts with PQBP-1 and PP1|splicing factor, PQBP1 and PP1 interacting', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14786478, end_position_on_the_genomic_accession=14803467, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  51733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51733, CurrentID=0, Status='live', Symbol='UPB1', Aliases='BUP1', description='ureidopropionase, beta', other_designations='beta-alanine synthase|n-carbamoyl-beta-alanine amidohydrolase', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24494467, end_position_on_the_genomic_accession=24528390, orientation='plus', exon_count=13, OMIM=606673, no_hdr0=''),
  51734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51734, CurrentID=0, Status='live', Symbol='MSRB1', Aliases='HSPC270, SELR, SELX, SEPX1, SepR', description='methionine sulfoxide reductase B1', other_designations='selenoprotein R|selenoprotein X, 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1938233, end_position_on_the_genomic_accession=1943293, orientation='minus', exon_count=4, OMIM=606216, no_hdr0=''),
  51735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51735, CurrentID=0, Status='live', Symbol='RAPGEF6', Aliases='KIA001LB, PDZ-GEF2, PDZGEF2, RA-GEF-2, RAGEF2', description='Rap guanine nucleotide exchange factor 6', other_designations='PDZ domain containing guanine nucleotide exchange factor (GEF) 2|PDZ domain-containing guanine nucleotide exchange factor 2|PDZ domain-containing guanine nucleotide exchange factor I|Rap guanine nucleotide exchange factor (GEF) 6', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131423921, end_position_on_the_genomic_accession=131635236, orientation='minus', exon_count=31, OMIM=610499, no_hdr0=''),
  51738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51738, CurrentID=0, Status='live', Symbol='GHRL', Aliases='MTLRP', description='ghrelin/obestatin prepropeptide', other_designations='In2c-preproghrelin|ghrelin, growth hormone secretagogue receptor ligand|ghrelin/obestatin preprohormone|growth hormone-releasing peptide|motilin-related peptide|prepro-appetite regulatory hormone|preproghrelin', map_location='3p26-p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10285750, end_position_on_the_genomic_accession=10292947, orientation='minus', exon_count=7, OMIM=605353, no_hdr0=''),
  51741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51741, CurrentID=0, Status='live', Symbol='WWOX', Aliases='D16S432E, EIEE28, FOR, FRA16D, HHCMA56, PRO0128, SCAR12, SDR41C1, WOX1', description='WW domain containing oxidoreductase', other_designations='WW domain-containing protein WWOX|fragile site FRA16D oxidoreductase|short chain dehydrogenase/reductase family 41C member 1|short chain dehydrogenase/reductase family 41C, member 1', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78099413, end_position_on_the_genomic_accession=79212667, orientation='plus', exon_count=17, OMIM=605131, no_hdr0=''),
  51742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51742, CurrentID=0, Status='live', Symbol='ARID4B', Aliases='BCAA, BRCAA1, RBBP1L1, RBP1L1, SAP180', description='AT-rich interaction domain 4B', other_designations='180 kDa Sin3-associated polypeptide|ARID domain-containing protein 4B|AT rich interactive domain 4B (RBP1-like)|Rb-binding protein homolog|SIN3A-associated protein 180|breast cancer-associated antigen 1|breast cancer-associated antigen BRCAA1|breast carcinoma-associated antigen|histone deacetylase complex subunit SAP180|retinoblastoma-binding protein 1-like 1|sin3-associated polypeptide p180', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235166895, end_position_on_the_genomic_accession=235328298, orientation='minus', exon_count=28, OMIM=609696, no_hdr0=''),
  51744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51744, CurrentID=0, Status='live', Symbol='CD244', Aliases='2B4, NAIL, NKR2B4, Nmrk, SLAMF4', description='CD244 molecule, natural killer cell receptor 2B4', other_designations='NK cell activation inducing ligand NAIL|NK cell activation-inducing ligand|NK cell type I receptor protein 2B4|SLAM family member 4|h2B4|signaling lymphocytic activation molecule 4', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160830158, end_position_on_the_genomic_accession=160862902, orientation='minus', exon_count=10, OMIM=605554, no_hdr0=''),
  51747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51747, CurrentID=0, Status='live', Symbol='LUC7L3', Aliases='CRA, CREAP-1, CROP, LUC7A, OA48-18, hLuc7A', description='LUC7-like 3 pre-mRNA splicing factor', other_designations='CRE-associated protein 1|cAMP regulatory element-associated protein 1|cisplatin resistance associated overexpressed protein|cisplatin resistance-associated overexpressed protein|cisplatin resistance-associated-overexpressed protein|okadaic acid-inducible phosphoprotein OA48-18', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50719565, end_position_on_the_genomic_accession=50752711, orientation='plus', exon_count=15, OMIM=609434, no_hdr0=''),
  51750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51750, CurrentID=0, Status='live', Symbol='RTEL1', Aliases='C20orf41, DKCA4, DKCB5, NHL, PFBMFT3, RTEL', description='regulator of telomere elongation helicase 1', other_designations='regulator of telomere length', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63657810, end_position_on_the_genomic_accession=63696253, orientation='plus', exon_count=35, OMIM=608833, no_hdr0=''),
  51751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51751, CurrentID=0, Status='live', Symbol='HIGD1B', Aliases='CLST11240, CLST11240-15', description='HIG1 hypoxia inducible domain family member 1B', other_designations='HIG1 hypoxia inducible domain family, member 1B', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44844815, end_position_on_the_genomic_accession=44850480, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  51752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51752, CurrentID=0, Status='live', Symbol='ERAP1', Aliases='A-LAP, ALAP, APPILS, ARTS-1, ARTS1, ERAAP, ERAAP1, PILS-AP, PILSAP', description='endoplasmic reticulum aminopeptidase 1', other_designations='CTD-2260A17.2|adipocyte-derived leucine aminopeptidase|aminopeptidase PILS|aminopeptidase regulator of TNFR1 shedding|endoplasmic reticulum aminopeptidase 1 delta-Exon-11 isoform|endoplasmic reticulum aminopeptidase 1 delta-Exon-13 isoform|endoplasmic reticulum aminopeptidase 1 delta-Exon-14 isoform|endoplasmic reticulum aminopeptidase 1 delta-Exon-15 isoform|endoplasmic reticulum aminopeptidase associated with antigen processing|puromycin-insensitive leucyl-specific aminopeptidase|type 1 tumor necrosis factor receptor shedding aminopeptidase regulator', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96760810, end_position_on_the_genomic_accession=96935924, orientation='minus', exon_count=27, OMIM=606832, no_hdr0=''),
  51754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51754, CurrentID=0, Status='live', Symbol='TMEM8B', Aliases='C9orf127, NAG-5, NGX6', description='transmembrane protein 8B', other_designations='Protein NGX6|nasopharyngeal carcinoma expressed 6|nasopharyngeal carcinoma related protein|nasopharyngeal carcinoma-associated gene 6 protein|protein NAG-5', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35829225, end_position_on_the_genomic_accession=35854847, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  51755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51755, CurrentID=0, Status='live', Symbol='CDK12', Aliases='CRK7, CRKR, CRKRS', description='cyclin-dependent kinase 12', other_designations='CDC2-related protein kinase 7|Cdc2-related kinase, arginine/serine-rich|cell division cycle 2-related protein kinase 7|cell division protein kinase 12', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39461486, end_position_on_the_genomic_accession=39567559, orientation='plus', exon_count=25, OMIM=615514, no_hdr0=''),
  51759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51759, CurrentID=0, Status='live', Symbol='C9orf78', Aliases='HCA59, HSPC220, bA409K20.3', description='chromosome 9 open reading frame 78', other_designations='hepatocellular carcinoma-associated antigen 59', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129827285, end_position_on_the_genomic_accession=129835293, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  51760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51760, CurrentID=0, Status='live', Symbol='SYT17', Aliases='sytXVII', description='synaptotagmin 17', other_designations='B/K protein|synaptotagmin XVII', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19167976, end_position_on_the_genomic_accession=19268334, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  51761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51761, CurrentID=0, Status='live', Symbol='ATP8A2', Aliases='ATP, ATPIB, CAMRQ4, IB, ML-1', description='ATPase, aminophospholipid transporter, class I, type 8A, member 2', other_designations='ATPase, aminophospholipid transporter-like, class I, type 8A, member 2|P4-ATPase flippase complex alpha subunit ATP8A2|probable phospholipid-transporting ATPase IB', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25372011, end_position_on_the_genomic_accession=26025851, orientation='plus', exon_count=47, OMIM=605870, no_hdr0=''),
  51762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51762, CurrentID=0, Status='live', Symbol='RAB8B', Aliases='', description='RAB8B, member RAS oncogene family', other_designations='RAB-8b protein', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63189529, end_position_on_the_genomic_accession=63267776, orientation='plus', exon_count=11, OMIM=613532, no_hdr0=''),
  51763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51763, CurrentID=0, Status='live', Symbol='INPP5K', Aliases='PPS, SKIP', description='inositol polyphosphate-5-phosphatase K', other_designations='skeletal muscle and kidney-enriched inositol phosphatase', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1494575, end_position_on_the_genomic_accession=1516888, orientation='minus', exon_count=14, OMIM=607875, no_hdr0=''),
  51764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51764, CurrentID=0, Status='live', Symbol='GNG13', Aliases='G(gamma)13, h2-35', description='guanine nucleotide binding protein (G protein), gamma 13', other_designations='G gamma subunit, clone:h2-35|guanine nucleotide binding protein 13, gamma', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=798041, end_position_on_the_genomic_accession=800733, orientation='minus', exon_count=3, OMIM=607298, no_hdr0=''),
  51765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51765, CurrentID=0, Status='live', Symbol='STK26', Aliases='MASK, MST4', description='serine/threonine protein kinase 26', other_designations='Mst3 and SOK1-related kinase|STE20-like kinase 4|STE20-like kinase MST4|mammalian Ste20-like protein kinase 4|mammalian sterile 20-like 4|serine/threonine protein kinase MST4|serine/threonine-protein kinase MASK', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132023217, end_position_on_the_genomic_accession=132075943, orientation='plus', exon_count=13, OMIM=300547, no_hdr0=''),
  51768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51768, CurrentID=0, Status='live', Symbol='TM7SF3', Aliases='', description='transmembrane 7 superfamily member 3', other_designations='seven span transmembrane protein|seven transmembrane protein TM7SF3', map_location='12q11-q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26971570, end_position_on_the_genomic_accession=27014427, orientation='minus', exon_count=13, OMIM=605181, no_hdr0=''),
  51773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51773, CurrentID=0, Status='live', Symbol='RSF1', Aliases='HBXAP, RSF-1, XAP8, p325', description='remodeling and spacing factor 1', other_designations='HBV pX associated protein-8|HBV pX-associated protein 8|hepatitis B virus x associated protein|hepatitis B virus x-associated protein|p325 subunit of RSF chromatin-remodeling complex', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77660009, end_position_on_the_genomic_accession=77820869, orientation='minus', exon_count=16, OMIM=608522, no_hdr0=''),
  51776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51776, CurrentID=0, Status='live', Symbol='ZAK', Aliases='AZK, MLK7, MLT, MLTK, MRK, mlklak, pk', description='sterile alpha motif and leucine zipper containing kinase AZK', other_designations='HCCS-4|MLK-like mitogen-activated protein triple kinase|human cervical cancer suppressor gene 4 protein|mixed lineage kinase 7|mixed lineage kinase with a leucine zipper and a sterile alpha motif|mixed lineage kinase-related kinase MRK-beta', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173075837, end_position_on_the_genomic_accession=173268009, orientation='plus', exon_count=22, OMIM=609479, no_hdr0=''),
  100190785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190785, CurrentID=0, Status='live', Symbol='IS4', Aliases='', description='Scoliosis, idiopathic, susceptibility to, 4', other_designations='', map_location='9q31.2-q34.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612238, no_hdr0=''),
  51778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51778, CurrentID=0, Status='live', Symbol='MYOZ2', Aliases='C4orf5, CMH16, CS-1', description='myozenin 2', other_designations='FATZ-related protein 2|calcineurin-binding protein calsarcin-1|muscle-specific protein', map_location='4q26-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119135784, end_position_on_the_genomic_accession=119187789, orientation='plus', exon_count=7, OMIM=605602, no_hdr0=''),
  100190787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190787, CurrentID=0, Status='live', Symbol='GEFSP6', Aliases='', description='Generalized epilepsy with febrile seizures plus, type 6', other_designations='', map_location='8p23-p21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612279, no_hdr0=''),
  51780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51780, CurrentID=0, Status='live', Symbol='KDM3B', Aliases='5qNCA, C5orf7, JMJD1B, NET22', description='lysine (K)-specific demethylase 3B', other_designations='jmjC domain-containing histone demethylation protein 2B|jumonji domain containing 1B|jumonji domain-containing protein 1B|nuclear protein 5qNCA', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138352596, end_position_on_the_genomic_accession=138437027, orientation='plus', exon_count=26, OMIM=609373, no_hdr0=''),
  100190789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190789, CurrentID=0, Status='live', Symbol='TSHQTL1', Aliases='', description='Thyroid-stimulating hormone level QTL 1', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612306, no_hdr0=''),
  100190790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190790, CurrentID=0, Status='live', Symbol='ADHD5', Aliases='', description='Attention deficit-hyperactivity disorder, susceptibility to, 5', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612311, no_hdr0=''),
  100190791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190791, CurrentID=0, Status='live', Symbol='ADHD6', Aliases='', description='Attention deficit-hyperactivity disorder, susceptibility to, 6', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612312, no_hdr0=''),
  132660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132660, CurrentID=0, Status='live', Symbol='LIN54', Aliases='CXCDC1, JC8.6, MIP120', description='lin-54 DREAM MuvB core complex component', other_designations='CXC domain-containing protein 1|lin-54 homolog', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82924603, end_position_on_the_genomic_accession=83012941, orientation='minus', exon_count=14, OMIM=613367, no_hdr0=''),
  100190800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190800, CurrentID=0, Status='live', Symbol='LDHAP7', Aliases='LDHAL7', description='lactate dehydrogenase A pseudogene 7', other_designations='lactate dehydrogenase A-like 7', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84775989, end_position_on_the_genomic_accession=84778319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  135644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=135644, CurrentID=0, Status='live', Symbol='TRIM40', Aliases='RNF35', description='tripartite motif containing 40', other_designations='RING finger protein 35|probable E3 NEDD8-protein ligase|ring finger RNF35', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30134510, end_position_on_the_genomic_accession=30148773, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  51802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51802, CurrentID=0, Status='live', Symbol='ASIC5', Aliases='ACCN5, HINAC, INAC', description='acid sensing ion channel subunit family member 5', other_designations='acid sensing (proton gated) ion channel family member 5|acid sensing ion channel family member 5|acid-sensing (proton-gated) ion channel family member 5|amiloride-sensitive cation channel 5, intestinal|amiloride-sensitive sodium channel|human intestine Na(+) channel', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155829729, end_position_on_the_genomic_accession=155866273, orientation='minus', exon_count=10, OMIM=616693, no_hdr0=''),
  51804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51804, CurrentID=0, Status='live', Symbol='SIX4', Aliases='AREC3', description='SIX homeobox 4', other_designations='sine oculis homeobox homolog 4', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60709538, end_position_on_the_genomic_accession=60724321, orientation='minus', exon_count=4, OMIM=606342, no_hdr0=''),
  51805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51805, CurrentID=0, Status='live', Symbol='COQ3', Aliases='DHHBMT, DHHBMTASE, UG0215E05, bA9819.1', description='coenzyme Q3 methyltransferase', other_designations='2-polyprenyl-6-hydroxyphenol methylase|2-polyprenyl-6-hydroxyphenyl methylase|3,4-dihydroxy-5-hexaprenylbenzoate methyltransferase|3-demethylubiquinol 3-O-methyltransferase|3-demethylubiquinone-10 3-methyltransferase|DHHB methyltransferase|DHHB-MT|DHHB-MTase|coenzyme Q3 homolog, methyltransferase|dihydroxyhexaprenylbenzoate methyltransferase|hexaprenyldihydroxybenzoate methyltransferase, mitochondrial|methyltransferase COQ3|polyprenyldihydroxybenzoate methyltransferase', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99369401, end_position_on_the_genomic_accession=99394218, orientation='minus', exon_count=8, OMIM=605196, no_hdr0=''),
  51806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51806, CurrentID=0, Status='live', Symbol='CALML5', Aliases='CLSP', description='calmodulin like 5', other_designations='calmodulin-like 5|calmodulin-like skin protein', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5498695, end_position_on_the_genomic_accession=5499570, orientation='minus', exon_count=1, OMIM=605183, no_hdr0=''),
  51807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51807, CurrentID=0, Status='live', Symbol='TUBA8', Aliases='TUBAL2', description='tubulin alpha 8', other_designations='tubulin alpha chain-like 2|tubulin, alpha 8', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18110687, end_position_on_the_genomic_accession=18131731, orientation='plus', exon_count=5, OMIM=605742, no_hdr0=''),
  51808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51808, CurrentID=0, Status='live', Symbol='PHAX', Aliases='RNUXA', description='phosphorylated adaptor for RNA export', other_designations='RNA U small nuclear RNA export adapter protein|RNA U, small nuclear RNA export adapter (phosphorylation regulated)|RNA U, small nuclear RNA export adaptor (phosphorylation regulated)', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126600915, end_position_on_the_genomic_accession=126627252, orientation='plus', exon_count=5, OMIM=604924, no_hdr0=''),
  51809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51809, CurrentID=0, Status='live', Symbol='GALNT7', Aliases='GALNAC-T7, GalNAcT7', description='polypeptide N-acetylgalactosaminyltransferase 7', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 7|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 7 (GalNAc-T7)|polypeptide GalNAc transferase 7|pp-GaNTase 7|protein-UDP acetylgalactosaminyltransferase 7', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173168753, end_position_on_the_genomic_accession=173323967, orientation='plus', exon_count=14, OMIM=605005, no_hdr0=''),
  51816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=51816, CurrentID=0, Status='live', Symbol='CECR1', Aliases='ADA2, ADGF, IDGFL, PAN, SNEDS', description='cat eye syndrome chromosome region, candidate 1', other_designations='adenosine deaminase 2|cat eye syndrome critical region protein 1', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17178790, end_position_on_the_genomic_accession=17221854, orientation='minus', exon_count=13, OMIM=607575, no_hdr0=''),
  100499171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499171, CurrentID=0, Status='live', Symbol='LINC00276', Aliases='NCRNA00276', description='long intergenic non-protein coding RNA 276', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=14228874, end_position_on_the_genomic_accession=14400958, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  139716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139716, CurrentID=0, Status='live', Symbol='GAB3', Aliases='', description='GRB2 associated binding protein 3', other_designations='DOS/Gab family member 3|GRB2-associated binder 3|GRB2-associated binding protein 3|Gab3 scaffolding protein|growth factor receptor bound protein 2-associated protein 3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154675249, end_position_on_the_genomic_accession=154751583, orientation='minus', exon_count=11, OMIM=300482, no_hdr0=''),
  503647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503647, CurrentID=0, Status='live', Symbol='DPRXP6', Aliases='', description='divergent-paired related homeobox pseudogene 6', other_designations='divergent paired-related homeobox pseudogene 6', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40758431, end_position_on_the_genomic_accession=40759089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401863, CurrentID=0, Status='live', Symbol='RPL10AP12', Aliases='RPL10A_7_1502', description='ribosomal protein L10a pseudogene 12', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85815846, end_position_on_the_genomic_accession=85816567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401864, CurrentID=0, Status='live', Symbol='LOC401864', Aliases='', description='chloride intracellular channel 1 pseudogene', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86434844, end_position_on_the_genomic_accession=86435498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401865, CurrentID=0, Status='live', Symbol='RPL7AP63', Aliases='RPL7A_28_1503', description='ribosomal protein L7a pseudogene 63', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86525435, end_position_on_the_genomic_accession=86526314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100190922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190922, CurrentID=0, Status='live', Symbol='LOC100190922', Aliases='', description='eukaryotic translation initiation factor 4E binding protein 2 pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103627244, end_position_on_the_genomic_accession=103629231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190923, CurrentID=0, Status='live', Symbol='EIF4EBP2P2', Aliases='', description='eukaryotic translation initiation factor 4E binding protein 2 pseudogene 2', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42578632, end_position_on_the_genomic_accession=42581710, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190924, CurrentID=0, Status='live', Symbol='EIF4EBP2P3', Aliases='', description='eukaryotic translation initiation factor 4E binding protein 2 pseudogene 3', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=98178660, end_position_on_the_genomic_accession=98179840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190926, CurrentID=0, Status='live', Symbol='IBD19', Aliases='', description='Inflammatory bowel disease 19', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612278, no_hdr0=''),
  100190928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190928, CurrentID=0, Status='live', Symbol='SLEB12', Aliases='', description='Systemic lupus erythematosus, susceptibility to, 12', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612254, no_hdr0=''),
  100190929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190929, CurrentID=0, Status='live', Symbol='IBD15', Aliases='', description='Inflammatory bowel disease-15', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612255, no_hdr0=''),
  100190930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190930, CurrentID=0, Status='live', Symbol='IBD16', Aliases='', description='Inflammatory bowel disease-16', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612259, no_hdr0=''),
  100190931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190931, CurrentID=0, Status='live', Symbol='IBD18', Aliases='', description='Inflammatory bowel disease-18', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612262, no_hdr0=''),
  100190932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190932, CurrentID=0, Status='live', Symbol='CMM7', Aliases='', description='Melanoma, cutaneous malignant, susceptibility to, 7', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612263, no_hdr0=''),
  100190934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190934, CurrentID=0, Status='live', Symbol='LOC100190934', Aliases='', description='alcohol dehydrogenase 5 (class III), chi polypeptide pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52160006, end_position_on_the_genomic_accession=52160601, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190938, CurrentID=0, Status='live', Symbol='RAMP2-AS1', Aliases='', description='RAMP2 antisense RNA 1', other_designations='CTD-3193K9.7', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42753914, end_position_on_the_genomic_accession=42761257, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100190939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190939, CurrentID=0, Status='live', Symbol='TPT1-AS1', Aliases='', description='TPT1 antisense RNA 1', other_designations='TPT1 antisense RNA 1 (non-protein coding)', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45341345, end_position_on_the_genomic_accession=45391483, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100190940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190940, CurrentID=0, Status='live', Symbol='LOC100190940', Aliases='', description='uncharacterized LOC100190940', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130033812, end_position_on_the_genomic_accession=130042342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  285423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285423, CurrentID=0, Status='live', Symbol='LOC285423', Aliases='', description='uncharacterized LOC285423', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=149150322, end_position_on_the_genomic_accession=149815835, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  100190947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190947, CurrentID=0, Status='live', Symbol='SUPT4H1P1', Aliases='SUPT4H1P', description='SPT4 homolog, DSIF elongation factor subunit pseudogene', other_designations='SUPT4H1 pseudogene', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75650309, end_position_on_the_genomic_accession=75651648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190949, CurrentID=0, Status='live', Symbol='C5orf52', Aliases='', description='chromosome 5 open reading frame 52', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157671237, end_position_on_the_genomic_accession=157680176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102725079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725079, CurrentID=0, Status='live', Symbol='LOC102725079', Aliases='', description='uncharacterized LOC102725079', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212795320, end_position_on_the_genomic_accession=212817419, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100190952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190952, CurrentID=0, Status='live', Symbol='NDUFB1P1', Aliases='NDUFB1P, bcm3095', description='NADH:ubiquinone oxidoreductase subunit B1 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 1 pseudogene 1|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 1, 7kDa pseudogene', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48181471, end_position_on_the_genomic_accession=48181791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100190954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190954, CurrentID=0, Status='live', Symbol='EXOSC3P2', Aliases='', description='exosome component 3 pseudogene 2', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7812717, end_position_on_the_genomic_accession=7814428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190955, CurrentID=0, Status='live', Symbol='FAM32DP', Aliases='FAM32D', description='family with sequence similarity 32 member D, pseudogene', other_designations='family with sequence similarity 32 member D (pseudogene)', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67706805, end_position_on_the_genomic_accession=67708322, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100190956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190956, CurrentID=0, Status='live', Symbol='FAM32EP', Aliases='FAM32E', description='family with sequence similarity 32 member E, pseudogene', other_designations='family with sequence similarity 32 member E (pseudogene)', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126457537, end_position_on_the_genomic_accession=126458895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373369, CurrentID=0, Status='live', Symbol='LOC105373369', Aliases='', description='uncharacterized LOC105373369', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152111235, end_position_on_the_genomic_accession=152125008, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  401875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401875, CurrentID=0, Status='live', Symbol='LOC401875', Aliases='', description='differentially expressed in FDCP 8 homolog (mouse) pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20612536, end_position_on_the_genomic_accession=20613167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100190982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190982, CurrentID=0, Status='live', Symbol='POROK5', Aliases='DSAP3', description='Porokeratosis, disseminated superficial actinic, 3', other_designations='', map_location='1p31.3-p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612293, no_hdr0=''),
  100190983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190983, CurrentID=0, Status='live', Symbol='DEL2Q32Q33', Aliases='C2DELq32q33', description='Chromosome 2q32-q33 deletion syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612313, no_hdr0=''),
  100190984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190984, CurrentID=0, Status='live', Symbol='DEL1Q43Q44', Aliases='C1DELq42q44, C1DELq43q44, DEL1Q42Q44', description='Chromosome 1q42-q44 deletion syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612337, no_hdr0=''),
  100190985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190985, CurrentID=0, Status='live', Symbol='IS5', Aliases='', description='scoliosis, idiopathic, susceptibility to, 5', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612239, no_hdr0=''),
  100190986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190986, CurrentID=0, Status='live', Symbol='LOC100190986', Aliases='', description='uncharacterized LOC100190986', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21432003, end_position_on_the_genomic_accession=21434455, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  139735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139735, CurrentID=0, Status='live', Symbol='ZFP92', Aliases='ZNF897', description='ZFP92 zinc finger protein', other_designations='zfp-92|zinc finger protein homologous to Zfp92 in mouse', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153411470, end_position_on_the_genomic_accession=153426481, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100132857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132857, CurrentID=0, Status='live', Symbol='LOC100132857', Aliases='', description='heat shock protein 90kDa alpha family class B member 1 pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16474345, end_position_on_the_genomic_accession=16475352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401884, CurrentID=0, Status='live', Symbol='MGC57346', Aliases='', description='ADP-ribosylation factor pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45620329, end_position_on_the_genomic_accession=45637963, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  139741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139741, CurrentID=0, Status='live', Symbol='ACTRT1', Aliases='AIP1, ARIP1, ARPT1, HSD27', description='actin-related protein T1', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128050962, end_position_on_the_genomic_accession=128052403, orientation='minus', exon_count=1, OMIM=300487, no_hdr0=''),
  100191039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100191039, CurrentID=0, Status='live', Symbol='RPL36AP33', Aliases='RPL36A_15_975', description='ribosomal protein L36a pseudogene 33', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35507534, end_position_on_the_genomic_accession=35507931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100191040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100191040, CurrentID=0, Status='live', Symbol='C2CD4D', Aliases='FAM148D', description='C2 calcium-dependent domain containing 4D', other_designations='family with sequence similarity 148, member D', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151837818, end_position_on_the_genomic_accession=151845329, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100191063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100191063, CurrentID=0, Status='live', Symbol='SNAR-A14', Aliases='', description='small ILF3/NF90-associated RNA A14', other_designations='SNAR-A55293015|small nuclear ILF3/NF90-associated RNA A14', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50097826, end_position_on_the_genomic_accession=50097946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  139748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139748, CurrentID=0, Status='live', Symbol='KRT18P44', Aliases='', description='keratin 18 pseudogene 44', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128714456, end_position_on_the_genomic_accession=128715830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401898, CurrentID=0, Status='live', Symbol='ZNF833P', Aliases='ZNF833', description='zinc finger protein 833, pseudogene', other_designations='CTC-499B15.5|zinc finger protein pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11673998, end_position_on_the_genomic_accession=11686569, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  445328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445328, CurrentID=0, Status='live', Symbol='ARHGEF35', Aliases='ARHGEF5L', description='Rho guanine nucleotide exchange factor 35', other_designations='Rho guanine nucleotide exchange factor (GEF) 35|rho guanine nucleotide exchange factor 5-like protein', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144186083, end_position_on_the_genomic_accession=144195698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  445329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445329, CurrentID=0, Status='live', Symbol='SULT1A4', Aliases='HAST3, M-PST, ST1A3/ST1A4, ST1A4, TL-PST', description='sulfotransferase family 1A member 4', other_designations='SLX1B-SULT1A4|aryl sulfotransferase 1A3/1A4|catecholamine-sulfating phenol sulfotransferase|monoamine-sulfating phenol sulfotransferase|placental estrogen sulfotransferase|sulfokinase|sulfotransferase family, cytosolic, 1A, phenol-preferring, member 4|sulfotransferase, monoamine-preferring|thermolabile phenol sulfotransferase', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29459886, end_position_on_the_genomic_accession=29464980, orientation='plus', exon_count=8, OMIM=615819, no_hdr0=''),
  285429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285429, CurrentID=0, Status='live', Symbol='DCAF4L1', Aliases='WDR21B', description='DDB1 and CUL4 associated factor 4-like 1', other_designations='WD repeat domain 21B|WD repeat-containing protein 21B', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41981696, end_position_on_the_genomic_accession=41986467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  401903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401903, CurrentID=0, Status='live', Symbol='OR10B1P', Aliases='OR10B2', description='olfactory receptor family 10 subfamily B member 1 pseudogene', other_designations='Acetolactate synthase|Human homolog of bacterial acetolactate synthase|olfactory receptor, family 10, subfamily B, member 1 pseudogene|olfactory receptor, family 10, subfamily B, member 2|olfactory receptor, family 10, subfamily T, member 2 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15141132, end_position_on_the_genomic_accession=15142038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  445341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445341, CurrentID=0, Status='live', Symbol='TNAP', Aliases='', description='TRAFs and NIK-associated protein', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=38913518, end_position_on_the_genomic_accession=38913712, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=8688, CurrentID=0, Status='live', Symbol='KRT37', Aliases='HA7, K37, KRTHA7', description='keratin 37, type I', other_designations='hair keratin, type I Ha7|keratin, hair, acidic, 7|keratin-37', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41420557, end_position_on_the_genomic_accession=41424570, orientation='minus', exon_count=7, OMIM=604541, no_hdr0=''),
  445347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445347, CurrentID=0, Status='live', Symbol='TARP', Aliases='CD3G, TCRG, TCRGC1, TCRGC2, TCRGV', description='TCR gamma alternate reading frame protein', other_designations='T-cell receptor gamma-chain constant region|TCR- gamma V-IV|Vgamma9ChainTCRMOP', map_location='7p15-p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38257879, end_position_on_the_genomic_accession=38305646, orientation='minus', exon_count=8, OMIM=609642, no_hdr0=''),
  445372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445372, CurrentID=0, Status='live', Symbol='TRIM6-TRIM34', Aliases='IFP1, RNF21, TRIM34', description='TRIM6-TRIM34 readthrough', other_designations='interferon-responsive finger protein 1 long form', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5596635, end_position_on_the_genomic_accession=5644398, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  403273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403273, CurrentID=0, Status='live', Symbol='OR5H14', Aliases='', description='olfactory receptor family 5 subfamily H member 14', other_designations='olfactory receptor, family 5, subfamily H, member 14|seven transmembrane helix receptor', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98148718, end_position_on_the_genomic_accession=98150399, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  401913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401913, CurrentID=0, Status='live', Symbol='LOC401913', Aliases='', description='ABO blood group (transferase A, alpha 1-3-N-acetylgalactosaminyltransferase; transferase B, alpha 1-3-galactosyltransferase) pseudogene', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34866351, end_position_on_the_genomic_accession=34866911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401914, CurrentID=0, Status='live', Symbol='CTBP2P7', Aliases='', description='C-terminal binding protein 2 pseudogene 7', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36496916, end_position_on_the_genomic_accession=36498162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874532, CurrentID=0, Status='live', Symbol='IBGC2', Aliases='', description='Basal ganglia calcification, idiopathic, 2', other_designations='', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606656, no_hdr0=''),
  102288414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102288414, CurrentID=0, Status='live', Symbol='C11orf98', Aliases='C11orf48', description='chromosome 11 open reading frame 98', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62662817, end_position_on_the_genomic_accession=62665216, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060158, CurrentID=0, Status='live', Symbol='LOC101060158', Aliases='', description='diamine acetyltransferase 1-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127498328, end_position_on_the_genomic_accession=127498821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377140, CurrentID=0, Status='live', Symbol='LOC105377140', Aliases='', description='uncharacterized LOC105377140', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66620273, end_position_on_the_genomic_accession=66636841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23564, CurrentID=0, Status='live', Symbol='DDAH2', Aliases='DDAH, DDAHII, G6a, HEL-S-277, NG30', description='dimethylarginine dimethylaminohydrolase 2', other_designations='S-phase protein|dimethylargininase-2|dimethylarginine dimethylaminohydrolase II|epididymis secretory protein Li 277', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31727037, end_position_on_the_genomic_accession=31730265, orientation='minus', exon_count=6, OMIM=604744, no_hdr0=''),
  102725091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725091, CurrentID=0, Status='live', Symbol='LOC102725091', Aliases='', description='circumsporozoite protein-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873358, CurrentID=0, Status='live', Symbol='RNA5SP26', Aliases='RN5S26', description='RNA, 5S ribosomal pseudogene 26', other_designations='RNA, 5S ribosomal 26', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401934, CurrentID=0, Status='live', Symbol='RNF223', Aliases='', description='ring finger protein 223', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1071746, end_position_on_the_genomic_accession=1074307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102157400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102157400, CurrentID=0, Status='live', Symbol='LINC00384', Aliases='', description='long intergenic non-protein coding RNA 384', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30151886, end_position_on_the_genomic_accession=30154897, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102157401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102157401, CurrentID=0, Status='live', Symbol='PCBP2-OT1', Aliases='TUC338', description='PCBP2 overlapping transcript 1', other_designations='PCBP2 overlapping transcript 1 (non-protein coding)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53464468, end_position_on_the_genomic_accession=53465057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102157402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102157402, CurrentID=0, Status='live', Symbol='AK6', Aliases='AD-004, CGI-137, CINAP, CIP, hCINAP', description='adenylate kinase 6', other_designations='ATP-AMP transphosphorylase 6|adrenal gland protein AD-004|coilin interacting nuclear ATPase protein|coilin interacting protein', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69351726, end_position_on_the_genomic_accession=69370013, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  139793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139793, CurrentID=0, Status='live', Symbol='PAGE3', Aliases='CT16.6, GAGED1, PAGE-3', description='PAGE family member 3', other_designations='G antigen family D member 1|G antigen, family D, 1|P antigen family, member 3 (prostate associated)|prostate-associated gene 3 protein', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55258415, end_position_on_the_genomic_accession=55463452, orientation='minus', exon_count=7, OMIM=300739, no_hdr0=''),
  445571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445571, CurrentID=0, Status='live', Symbol='CBWD3', Aliases='bA561O23.1', description='COBW domain containing 3', other_designations='cobalamin synthase W domain-containing protein 3|cobalamin synthetase W domain-containing protein 3', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68241866, end_position_on_the_genomic_accession=68300035, orientation='plus', exon_count=18, OMIM=611080, no_hdr0=''),
  445573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445573, CurrentID=0, Status='live', Symbol='ATP5J2P3', Aliases='bA159H20.5', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2 pseudogene 3|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit f, isoform 2 pseudogene 3|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene 3', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77040390, end_position_on_the_genomic_accession=77040807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  445577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445577, CurrentID=0, Status='live', Symbol='C9orf129', Aliases='bA165J3.3', description='chromosome 9 open reading frame 129', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93318199, end_position_on_the_genomic_accession=93346414, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  445582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445582, CurrentID=0, Status='live', Symbol='POTEE', Aliases='A26C1, A26C1A, CT104.2, POTE-2, POTE2, POTE2gamma', description='POTE ankyrin domain family member E', other_designations='ANKRD26-like family C member 1A|ANKRD26-like family C, member 1A|POTE ankyrin domain family, member E|cancer/testis antigen family 104, member 2|prostate, ovary, testis-expressed protein on chromosome 2|protein expressed in prostate, ovary, testis, and placenta 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131211134, end_position_on_the_genomic_accession=131265448, orientation='plus', exon_count=16, OMIM=608914, no_hdr0=''),
  401944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401944, CurrentID=0, Status='live', Symbol='LDLRAD2', Aliases='', description='low density lipoprotein receptor class A domain containing 2', other_designations='low density lipoprotein receptor A domain containing 2', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21812265, end_position_on_the_genomic_accession=21825221, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  139804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139804, CurrentID=0, Status='live', Symbol='RBMXL3', Aliases='CXorf55', description='RNA binding motif protein, X-linked-like 3', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115189400, end_position_on_the_genomic_accession=115192868, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100147744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147744, CurrentID=0, Status='live', Symbol='RNU7-1', Aliases='RNU7, U7.1', description='RNA, U7 small nuclear 1', other_designations='RNA, small nuclear U7|RNA, small nuclear U7.1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6943816, end_position_on_the_genomic_accession=6943878, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100147745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147745, CurrentID=0, Status='live', Symbol='RNU7-2P', Aliases='U7.2', description='RNA, U7 small nuclear 2 pseudogene', other_designations='RNA, small nuclear U7.2 pseudogene', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146145156, end_position_on_the_genomic_accession=146145216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861445, CurrentID=0, Status='live', Symbol='DNAJC19P2', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 2', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 2', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41605290, end_position_on_the_genomic_accession=41605836, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147746, CurrentID=0, Status='live', Symbol='RNU7-3P', Aliases='U7.3', description='RNA, U7 small nuclear 3 pseudogene', other_designations='RNA, small nuclear U7.3 pseudogene', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149516957, end_position_on_the_genomic_accession=149517016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147747, CurrentID=0, Status='live', Symbol='RNU7-4P', Aliases='RNU7-132P, U7.4', description='RNA, U7 small nuclear 4 pseudogene', other_designations='RNA, U7 small nuclear 132 pseudogene|RNA, small nuclear U7.4 pseudogene', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68911921, end_position_on_the_genomic_accession=68911980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147748, CurrentID=0, Status='live', Symbol='RNU7-5P', Aliases='U7.5', description='RNA, U7 small nuclear 5 pseudogene', other_designations='RNA, small nuclear U7.5 pseudogen', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45446465, end_position_on_the_genomic_accession=45446524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147749, CurrentID=0, Status='live', Symbol='RNU7-6P', Aliases='U7.6', description='RNA, U7 small nuclear 6 pseudogene', other_designations='RNA, small nuclear U7.6 pseudogene', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51904969, end_position_on_the_genomic_accession=51905030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147750, CurrentID=0, Status='live', Symbol='RNU5B-2P', Aliases='', description='RNA, U5B small nuclear 2, pseudogene', other_designations='RNA, U5B small nuclear 2 pseudogene|RNA, U5B small nuclear pseudogene 2', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40498917, end_position_on_the_genomic_accession=40499003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873359, CurrentID=0, Status='live', Symbol='RNA5SP27', Aliases='RN5S27', description='RNA, 5S ribosomal pseudogene 27', other_designations='RNA, 5S ribosomal 27', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45551485, end_position_on_the_genomic_accession=45551603, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401959, CurrentID=0, Status='live', Symbol='VDAC2P3', Aliases='', description='voltage-dependent anion channel 2 pseudogene 3', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117640340, end_position_on_the_genomic_accession=117641727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499191, CurrentID=0, Status='live', Symbol='LOC100499191', Aliases='', description='male-specific lethal 3 homolog (Drosophila) pseudogene', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139171816, end_position_on_the_genomic_accession=139174446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147752, CurrentID=0, Status='live', Symbol='RNU5B-4P', Aliases='', description='RNA, U5B small nuclear 4, pseudogene', other_designations='RNA, U5B small nuclear 4 pseudogene|RNA, U5B small nuclear pseudogene 4', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139447907, end_position_on_the_genomic_accession=139448018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147753, CurrentID=0, Status='live', Symbol='RNU5B-5P', Aliases='', description='RNA, U5B small nuclear 5, pseudogene', other_designations='RNA, U5B small nuclear 5 pseudogene|RNA, U5B small nuclear pseudogene 5', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125425789, end_position_on_the_genomic_accession=125425876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84673, CurrentID=0, Status='live', Symbol='TTTY8', Aliases='LINC00130, NCRNA00130, TTY8', description='testis-specific transcript, Y-linked 8 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 130|testis transcript Y 8|transcript Y 8', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9691100, end_position_on_the_genomic_accession=9693699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100147754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147754, CurrentID=0, Status='live', Symbol='RNU5B-6P', Aliases='', description='RNA, U5B small nuclear 6, pseudogene', other_designations='RNA, U5B small nuclear 6 pseudogene|RNA, U5B small nuclear pseudogene 6', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109184830, end_position_on_the_genomic_accession=109185107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147755, CurrentID=0, Status='live', Symbol='RNU7-7P', Aliases='U7.7', description='RNA, U7 small nuclear 7 pseudogene', other_designations='RNA, small nuclear U7.7 pseudogene', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38608218, end_position_on_the_genomic_accession=38608275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147756, CurrentID=0, Status='live', Symbol='RNU7-8P', Aliases='RNU7-139P, U7.8', description='RNA, U7 small nuclear 8 pseudogene', other_designations='RNA, U7 small nuclear 139 pseudogene|RNA, small nuclear U7.8 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77420326, end_position_on_the_genomic_accession=77420384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147757, CurrentID=0, Status='live', Symbol='RNU7-9P', Aliases='U7.9', description='RNA, U7 small nuclear 9 pseudogene', other_designations='RNA, small nuclear U7.9 pseudogene', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229606239, end_position_on_the_genomic_accession=229606299, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401967, CurrentID=0, Status='live', Symbol='NBPF17P', Aliases='NBPF23, NBPF23P', description='neuroblastoma breakpoint family member 17, pseudogene', other_designations='neuroblastoma breakpoint family, member 17 (pseudogene)|neuroblastoma breakpoint family, member 17, pseudogene|neuroblastoma breakpoint family, member 23, pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143595216, end_position_on_the_genomic_accession=143635780, orientation='minus', exon_count=0, OMIM=612970, no_hdr0=''),
  105827617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105827617, CurrentID=0, Status='live', Symbol='CRAT40', Aliases='', description='cervical cancer-associated transcript 40', other_designations='CTD-2535L24.3', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369659, CurrentID=0, Status='live', Symbol='LOC105369659', Aliases='', description='uncharacterized LOC105369659', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10726192, end_position_on_the_genomic_accession=10741323, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100147762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147762, CurrentID=0, Status='live', Symbol='RNU7-14P', Aliases='U7.14', description='RNA, U7 small nuclear 14 pseudogene', other_designations='RNA, small nuclear U7.14 pseudogene', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53668698, end_position_on_the_genomic_accession=53668757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147763, CurrentID=0, Status='live', Symbol='RNU7-15P', Aliases='U7.15', description='RNA, U7 small nuclear 15 pseudogene', other_designations='RNA, small nuclear U7.15 pseudogene', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132841149, end_position_on_the_genomic_accession=132841407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147764, CurrentID=0, Status='live', Symbol='RNU7-16P', Aliases='U7.16', description='RNA, U7 small nuclear 16 pseudogene', other_designations='RNA, small nuclear U7.16 pseudogene', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129405635, end_position_on_the_genomic_accession=129405696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401973, CurrentID=0, Status='live', Symbol='OR10J6P', Aliases='OR10J6', description='olfactory receptor family 10 subfamily J member 6 pseudogene', other_designations='Olfactory receptor 10J6|olfactory receptor, family 10, subfamily J, member 6 pseudogene|seven transmembrane helix receptor', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159598299, end_position_on_the_genomic_accession=159599225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401974, CurrentID=0, Status='live', Symbol='RPS2P10', Aliases='RPS2_4_110', description='ribosomal protein S2 pseudogene 10', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165581558, end_position_on_the_genomic_accession=165582515, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106614088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106614088, CurrentID=0, Status='live', Symbol='MIR34AHG', Aliases='', description='MIR34A host gene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371405, CurrentID=0, Status='live', Symbol='LOC105371405', Aliases='', description='collagen alpha-1(XIII) chain', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149833133, end_position_on_the_genomic_accession=149840737, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100147767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147767, CurrentID=0, Status='live', Symbol='RNU7-19P', Aliases='U7.19', description='RNA, U7 small nuclear 19 pseudogene', other_designations='RNA, small nuclear U7.19 pseudogene', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73027811, end_position_on_the_genomic_accession=73027870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147768, CurrentID=0, Status='live', Symbol='RNU7-20P', Aliases='U7.20', description='RNA, U7 small nuclear 20 pseudogene', other_designations='RNA, small nuclear U7.20 pseudogene', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148821015, end_position_on_the_genomic_accession=148821069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420633, CurrentID=0, Status='live', Symbol='PGAM1P10', Aliases='', description='phosphoglycerate mutase 1 pseudogene 10', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73055097, end_position_on_the_genomic_accession=73055854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147769, CurrentID=0, Status='live', Symbol='RNU7-21P', Aliases='U7.21', description='RNA, U7 small nuclear 21 pseudogene', other_designations='RNA, small nuclear U7.21 pseudogene', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132963733, end_position_on_the_genomic_accession=132963793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147771, CurrentID=0, Status='live', Symbol='RNU7-23P', Aliases='U7.23', description='RNA, U7 small nuclear 23 pseudogene', other_designations='RNA, small nuclear U7.23 pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66919763, end_position_on_the_genomic_accession=66919822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723483, CurrentID=0, Status='live', Symbol='CSE1L-AS1', Aliases='', description='CSE1L antisense RNA 1', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49040463, end_position_on_the_genomic_accession=49046056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100147773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147773, CurrentID=0, Status='live', Symbol='LOC100147773', Aliases='', description='uncharacterized LOC100147773', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165768929, end_position_on_the_genomic_accession=165775449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401982, CurrentID=0, Status='live', Symbol='AKR1B1P1', Aliases='ALDRL1', description='aldehyde reductase family 1, member B1 pseudogene 1', other_designations='aldehyde reductase (aldose reductase)-like 1|aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224574401, end_position_on_the_genomic_accession=224575588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372692, CurrentID=0, Status='live', Symbol='LOC105372692', Aliases='', description='uncharacterized LOC105372692', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57720481, end_position_on_the_genomic_accession=57726370, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  445815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=445815, CurrentID=0, Status='live', Symbol='PALM2-AKAP2', Aliases='AKAP2', description='PALM2-AKAP2 readthrough', other_designations='PALM2-AKAP2 readthrough transcript', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109780297, end_position_on_the_genomic_accession=110172512, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  401983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401983, CurrentID=0, Status='live', Symbol='ZNF847P', Aliases='gm127', description='zinc finger protein 847, pseudogene', other_designations='zinc finger protein 717 pseudogene', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=227884733, end_position_on_the_genomic_accession=227885408, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873360, CurrentID=0, Status='live', Symbol='RNA5SP28', Aliases='RN5S28', description='RNA, 5S ribosomal pseudogene 28', other_designations='RNA, 5S ribosomal 28', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50812988, end_position_on_the_genomic_accession=50813106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100846978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846978, CurrentID=0, Status='live', Symbol='LINC00506', Aliases='', description='long intergenic non-protein coding RNA 506', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87089280, end_position_on_the_genomic_accession=87157069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100846991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846991, CurrentID=0, Status='live', Symbol='MIR5197', Aliases='', description='microRNA 5197', other_designations='hsa-mir-5197', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143679860, end_position_on_the_genomic_accession=143679971, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846992, CurrentID=0, Status='live', Symbol='MIR548AW', Aliases='', description='microRNA 548aw', other_designations='hsa-mir-548aw', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132945707, end_position_on_the_genomic_accession=132945771, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846993, CurrentID=0, Status='live', Symbol='MIR5009', Aliases='', description='microRNA 5009', other_designations='hsa-mir-5009', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89883931, end_position_on_the_genomic_accession=89884030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100846994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846994, CurrentID=0, Status='live', Symbol='MIR3670-2', Aliases='', description='microRNA 3670-2', other_designations='hsa-mir-3670-2', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16306370, end_position_on_the_genomic_accession=16306434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846995, CurrentID=0, Status='live', Symbol='MIR5000', Aliases='', description='microRNA 5000', other_designations='hsa-mir-5000', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75090812, end_position_on_the_genomic_accession=75090914, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846996, CurrentID=0, Status='live', Symbol='MIR5007', Aliases='', description='microRNA 5007', other_designations='hsa-mir-5007', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=55174454, end_position_on_the_genomic_accession=55174548, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846997, CurrentID=0, Status='live', Symbol='MIR5583-2', Aliases='', description='microRNA 5583-2', other_designations='hsa-mir-5583-2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39676719, end_position_on_the_genomic_accession=39676777, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100846998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846998, CurrentID=0, Status='live', Symbol='MIR5689', Aliases='', description='microRNA 5689', other_designations='hsa-mir-5689', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10439717, end_position_on_the_genomic_accession=10439794, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100846999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100846999, CurrentID=0, Status='live', Symbol='MIR4444-2', Aliases='', description='microRNA 4444-2', other_designations='hsa-mir-4444-2', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75214476, end_position_on_the_genomic_accession=75214549, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847000, CurrentID=0, Status='live', Symbol='MIR5579', Aliases='', description='microRNA 5579', other_designations='hsa-mir-5579', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=79422169, end_position_on_the_genomic_accession=79422226, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847001, CurrentID=0, Status='live', Symbol='MIR5680', Aliases='', description='microRNA 5680', other_designations='hsa-mir-5680', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102125432, end_position_on_the_genomic_accession=102125515, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847002, CurrentID=0, Status='live', Symbol='MIR5011', Aliases='', description='microRNA 5011', other_designations='hsa-mir-5011', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67081584, end_position_on_the_genomic_accession=67081686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847003, CurrentID=0, Status='live', Symbol='MIR5693', Aliases='', description='microRNA 5693', other_designations='hsa-mir-5693', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51348567, end_position_on_the_genomic_accession=51348639, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847004, CurrentID=0, Status='live', Symbol='MIR5188', Aliases='mir-5188', description='microRNA 5188', other_designations='hsa-mir-5188', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124915547, end_position_on_the_genomic_accession=124915659, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847005, CurrentID=0, Status='live', Symbol='MIR5701-2', Aliases='', description='microRNA 5701-2', other_designations='hsa-mir-5701-2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21513959, end_position_on_the_genomic_accession=21514040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847006, CurrentID=0, Status='live', Symbol='MIR5571', Aliases='', description='microRNA 5571', other_designations='hsa-mir-5571', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22886267, end_position_on_the_genomic_accession=22886379, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847007, CurrentID=0, Status='live', Symbol='MIR5696', Aliases='', description='microRNA 5696', other_designations='hsa-mir-5696', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101309450, end_position_on_the_genomic_accession=101309534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847008, CurrentID=0, Status='live', Symbol='MIR4524B', Aliases='', description='microRNA 4524b', other_designations='hsa-mir-4524b', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69099542, end_position_on_the_genomic_accession=69099656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847009, CurrentID=0, Status='live', Symbol='MIR1295B', Aliases='', description='microRNA 1295b', other_designations='hsa-mir-1295b', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171101739, end_position_on_the_genomic_accession=171101798, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847010, CurrentID=0, Status='live', Symbol='MIR5581', Aliases='', description='microRNA 5581', other_designations='hsa-mir-5581', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37500935, end_position_on_the_genomic_accession=37500994, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847012, CurrentID=0, Status='live', Symbol='MIR5004', Aliases='mir-5004', description='microRNA 5004', other_designations='hsa-mir-5004', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33438331, end_position_on_the_genomic_accession=33438437, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847013, CurrentID=0, Status='live', Symbol='MIR5692B', Aliases='', description='microRNA 5692b', other_designations='hsa-mir-5692b', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42950928, end_position_on_the_genomic_accession=42951014, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847014, CurrentID=0, Status='live', Symbol='MIR5100', Aliases='', description='microRNA 5100', other_designations='hsa-mir-5100', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42997563, end_position_on_the_genomic_accession=42997681, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847015, CurrentID=0, Status='live', Symbol='MIR5691', Aliases='', description='microRNA 5691', other_designations='hsa-mir-5691', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9090312, end_position_on_the_genomic_accession=9090379, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847016, CurrentID=0, Status='live', Symbol='MIR5695', Aliases='', description='microRNA 5695', other_designations='hsa-mir-5695', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12920320, end_position_on_the_genomic_accession=12920404, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847017, CurrentID=0, Status='live', Symbol='MIR5692C2', Aliases='', description='microRNA 5692c-2', other_designations='hsa-mir-5692c-2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97964405, end_position_on_the_genomic_accession=97964481, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847018, CurrentID=0, Status='live', Symbol='MIR5585', Aliases='', description='microRNA 5585', other_designations='hsa-mir-5585|miR-5585-3p', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32086949, end_position_on_the_genomic_accession=32087007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847019, CurrentID=0, Status='live', Symbol='MIR5687', Aliases='', description='microRNA 5687', other_designations='hsa-mir-5687', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55508850, end_position_on_the_genomic_accession=55508926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847020, CurrentID=0, Status='live', Symbol='MIR5582', Aliases='', description='microRNA 5582', other_designations='hsa-mir-5582', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46753125, end_position_on_the_genomic_accession=46753192, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847021, CurrentID=0, Status='live', Symbol='MIR5002', Aliases='', description='microRNA 5002', other_designations='hsa-mir-5002', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124132929, end_position_on_the_genomic_accession=124133025, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847022, CurrentID=0, Status='live', Symbol='MIR5093', Aliases='', description='microRNA 5093', other_designations='hsa-mir-5093', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85306226, end_position_on_the_genomic_accession=85306325, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847023, CurrentID=0, Status='live', Symbol='MIR5091', Aliases='mir-5091', description='microRNA 5091', other_designations='hsa-mir-5091', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13627865, end_position_on_the_genomic_accession=13627957, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847024, CurrentID=0, Status='live', Symbol='MIR5698', Aliases='', description='microRNA 5698', other_designations='hsa-mir-5698', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154104521, end_position_on_the_genomic_accession=154104592, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  401992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401992, CurrentID=0, Status='live', Symbol='OR2T2', Aliases='OR1-43P, OR2T2', description='olfactory receptor family 2 subfamily T member 2', other_designations='olfactory receptor OR1-43|olfactory receptor, family 2, subfamily T, member 2 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248449062, end_position_on_the_genomic_accession=248548832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100847026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847026, CurrentID=0, Status='live', Symbol='MIR5006', Aliases='', description='microRNA 5006', other_designations='hsa-mir-5006', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41568286, end_position_on_the_genomic_accession=41568395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847027, CurrentID=0, Status='live', Symbol='MIR5705', Aliases='', description='microRNA 5705', other_designations='hsa-mir-5705', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87300495, end_position_on_the_genomic_accession=87300583, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847028, CurrentID=0, Status='live', Symbol='MIR5587', Aliases='', description='microRNA 5587', other_designations='hsa-mir-5587', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=535316, end_position_on_the_genomic_accession=535368, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847029, CurrentID=0, Status='live', Symbol='MIR5003', Aliases='', description='microRNA 5003', other_designations='hsa-mir-5003', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172662165, end_position_on_the_genomic_accession=172662263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847030, CurrentID=0, Status='live', Symbol='MIR548AT', Aliases='', description='microRNA 548at', other_designations='hsa-mir-548at', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42494773, end_position_on_the_genomic_accession=42494830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  401993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401993, CurrentID=0, Status='live', Symbol='OR2T5', Aliases='OR1-62', description='olfactory receptor family 2 subfamily T member 5', other_designations='olfactory receptor OR1-62|olfactory receptor, family 2, subfamily T, member 5', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248488589, end_position_on_the_genomic_accession=248489536, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847032, CurrentID=0, Status='live', Symbol='MIR5707', Aliases='', description='microRNA 5707', other_designations='hsa-mir-5707', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158591616, end_position_on_the_genomic_accession=158591696, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847033, CurrentID=0, Status='live', Symbol='MIR4436B2', Aliases='', description='microRNA 4436b-2', other_designations='hsa-mir-4436b-2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110284853, end_position_on_the_genomic_accession=110284943, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847034, CurrentID=0, Status='live', Symbol='MIR5683', Aliases='', description='microRNA 5683', other_designations='hsa-mir-5683', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6169334, end_position_on_the_genomic_accession=6169409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847035, CurrentID=0, Status='live', Symbol='MIR548AR', Aliases='', description='microRNA 548ar', other_designations='hsa-mir-548ar', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114244505, end_position_on_the_genomic_accession=114244561, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847036, CurrentID=0, Status='live', Symbol='MIR5186', Aliases='', description='microRNA 5186', other_designations='hsa-mir-5186', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151565876, end_position_on_the_genomic_accession=151565995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  401994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401994, CurrentID=0, Status='live', Symbol='OR14I1', Aliases='OR5BU1, OR5BU1P', description='olfactory receptor family 14 subfamily I member 1', other_designations='olfactory receptor 5BU1|olfactory receptor, family 14, subfamily I, member 1|olfactory receptor, family 5, subfamily BU, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248483422, end_position_on_the_genomic_accession=248698899, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100847038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847038, CurrentID=0, Status='live', Symbol='MIR5692A2', Aliases='', description='microRNA 5692a-2', other_designations='hsa-mir-5692a-2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12719132, end_position_on_the_genomic_accession=12719190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847039, CurrentID=0, Status='live', Symbol='MIR5092', Aliases='', description='microRNA 5092', other_designations='hsa-mir-5092', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125151465, end_position_on_the_genomic_accession=125151552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847040, CurrentID=0, Status='live', Symbol='MIR5704', Aliases='', description='microRNA 5704', other_designations='hsa-mir-5704', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131985855, end_position_on_the_genomic_accession=131985931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847041, CurrentID=0, Status='live', Symbol='MIR3680-2', Aliases='', description='microRNA 3680-2', other_designations='hsa-mir-3680-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29599179, end_position_on_the_genomic_accession=29599265, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847042, CurrentID=0, Status='live', Symbol='MIR5572', Aliases='', description='microRNA 5572', other_designations='hsa-mir-5572', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80581103, end_position_on_the_genomic_accession=80581239, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847043, CurrentID=0, Status='live', Symbol='MIR5682', Aliases='', description='microRNA 5682', other_designations='hsa-mir-5682', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121049640, end_position_on_the_genomic_accession=121049715, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847044, CurrentID=0, Status='live', Symbol='MIR5087', Aliases='', description='microRNA 5087', other_designations='hsa-mir-5087', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148334479, end_position_on_the_genomic_accession=148334554, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847045, CurrentID=0, Status='live', Symbol='MIR548AU', Aliases='', description='microRNA 548au', other_designations='hsa-mir-548au', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93594841, end_position_on_the_genomic_accession=93594894, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847046, CurrentID=0, Status='live', Symbol='MIR5010', Aliases='mir-5010', description='microRNA 5010', other_designations='hsa-mir-5010', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42514188, end_position_on_the_genomic_accession=42514307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847047, CurrentID=0, Status='live', Symbol='MIR4666B', Aliases='', description='microRNA 4666b', other_designations='hsa-mir-4666b', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=29574278, end_position_on_the_genomic_accession=29574358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847048, CurrentID=0, Status='live', Symbol='MIR5690', Aliases='', description='microRNA 5690', other_designations='hsa-mir-5690', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35664717, end_position_on_the_genomic_accession=35664789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847049, CurrentID=0, Status='live', Symbol='MIR4999', Aliases='mir-4999', description='microRNA 4999', other_designations='hsa-mir-4999', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8389290, end_position_on_the_genomic_accession=8389380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847050, CurrentID=0, Status='live', Symbol='MIR5191', Aliases='', description='microRNA 5191', other_designations='hsa-mir-5191', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201719508, end_position_on_the_genomic_accession=201719627, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847051, CurrentID=0, Status='live', Symbol='MIR5194', Aliases='mir-5194', description='microRNA 5194', other_designations='hsa-mir-5194', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130008334, end_position_on_the_genomic_accession=130008453, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847052, CurrentID=0, Status='live', Symbol='MIR664B', Aliases='MIR644B, hsa-mir-664b, mir-664b', description='microRNA 664b', other_designations='hsa-mir-644b|microRNA 644b', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154768596, end_position_on_the_genomic_accession=154768656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847053, CurrentID=0, Status='live', Symbol='MIR5702', Aliases='', description='microRNA 5702', other_designations='hsa-mir-5702', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226658710, end_position_on_the_genomic_accession=226658793, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847054, CurrentID=0, Status='live', Symbol='MIR5588', Aliases='', description='microRNA 5588', other_designations='hsa-mir-5588', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185253210, end_position_on_the_genomic_accession=185253272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847055, CurrentID=0, Status='live', Symbol='MIR5697', Aliases='', description='microRNA 5697', other_designations='hsa-mir-5697', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9967381, end_position_on_the_genomic_accession=9967458, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847056, CurrentID=0, Status='live', Symbol='MIR5708', Aliases='', description='microRNA 5708', other_designations='hsa-mir-5708', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80241389, end_position_on_the_genomic_accession=80241473, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847057, CurrentID=0, Status='live', Symbol='MIR5189', Aliases='mir-5189', description='microRNA 5189', other_designations='hsa-mir-5189', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88468918, end_position_on_the_genomic_accession=88469031, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847058, CurrentID=0, Status='live', Symbol='MIR5681A', Aliases='', description='microRNA 5681a', other_designations='hsa-mir-5681a', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74548543, end_position_on_the_genomic_accession=74548617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847059, CurrentID=0, Status='live', Symbol='MIR5094', Aliases='', description='microRNA 5094', other_designations='hsa-mir-5094', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89850637, end_position_on_the_genomic_accession=89850721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847060, CurrentID=0, Status='live', Symbol='MIR5701-1', Aliases='', description='microRNA 5701-1', other_designations='hsa-mir-5701-1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20940252, end_position_on_the_genomic_accession=20940333, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847061, CurrentID=0, Status='live', Symbol='MIR4536-2', Aliases='', description='microRNA 4536-2', other_designations='hsa-mir-4536-2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55451495, end_position_on_the_genomic_accession=55451582, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847062, CurrentID=0, Status='live', Symbol='MIR5195', Aliases='', description='microRNA 5195', other_designations='hsa-mir-5195', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106850885, end_position_on_the_genomic_accession=106850999, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847063, CurrentID=0, Status='live', Symbol='MIR548AX', Aliases='', description='microRNA 548ax', other_designations='hsa-mir-548ax', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11318614, end_position_on_the_genomic_accession=11318686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847064, CurrentID=0, Status='live', Symbol='MIR5694', Aliases='', description='microRNA 5694', other_designations='hsa-mir-5694', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67441855, end_position_on_the_genomic_accession=67441930, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847065, CurrentID=0, Status='live', Symbol='MIR5591', Aliases='', description='microRNA 5591', other_designations='hsa-mir-5591', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39411910, end_position_on_the_genomic_accession=39411974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847066, CurrentID=0, Status='live', Symbol='MIR5692A1', Aliases='', description='microRNA 5692a-1', other_designations='hsa-mir-5692a-1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97963658, end_position_on_the_genomic_accession=97963726, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847067, CurrentID=0, Status='live', Symbol='MIR5089', Aliases='', description='microRNA 5089', other_designations='hsa-mir-5089', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46973017, end_position_on_the_genomic_accession=46973100, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847068, CurrentID=0, Status='live', Symbol='MIR548AO', Aliases='', description='microRNA 548ao', other_designations='hsa-mir-548ao', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41271048, end_position_on_the_genomic_accession=41271143, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847069, CurrentID=0, Status='live', Symbol='MIR5590', Aliases='', description='microRNA 5590', other_designations='hsa-mir-5590', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134857820, end_position_on_the_genomic_accession=134857873, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847070, CurrentID=0, Status='live', Symbol='MIR5196', Aliases='mir-5196', description='microRNA 5196', other_designations='hsa-mir-5196', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35345513, end_position_on_the_genomic_accession=35345627, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847071, CurrentID=0, Status='live', Symbol='MIR5684', Aliases='', description='microRNA 5684', other_designations='hsa-mir-5684', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12787128, end_position_on_the_genomic_accession=12787192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847072, CurrentID=0, Status='live', Symbol='MIR5008', Aliases='', description='microRNA 5008', other_designations='hsa-mir-5008', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227941590, end_position_on_the_genomic_accession=227941683, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105259600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105259600, CurrentID=0, Status='live', Symbol='CXDUPQ26.3', Aliases='DUPXq26.3, XLAG', description='Chromosome Xq26.3 duplication syndrome', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300942, no_hdr0=''),
  100847074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847074, CurrentID=0, Status='live', Symbol='MIR5088', Aliases='mir-5088', description='microRNA 5088', other_designations='hsa-mir-5088', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49682117, end_position_on_the_genomic_accession=49682195, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847075, CurrentID=0, Status='live', Symbol='MIR5685', Aliases='', description='microRNA 5685', other_designations='hsa-mir-5685', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53276993, end_position_on_the_genomic_accession=53277071, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847076, CurrentID=0, Status='live', Symbol='MIR5580', Aliases='', description='microRNA 5580', other_designations='hsa-mir-5580', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53948427, end_position_on_the_genomic_accession=53948484, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847077, CurrentID=0, Status='live', Symbol='MIR5688', Aliases='', description='microRNA 5688', other_designations='hsa-mir-5688', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=85385710, end_position_on_the_genomic_accession=85385792, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847078, CurrentID=0, Status='live', Symbol='MIR548AQ', Aliases='', description='microRNA 548aq', other_designations='hsa-mir-548aq', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185767847, end_position_on_the_genomic_accession=185767904, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847079, CurrentID=0, Status='live', Symbol='MIR5193', Aliases='mir-5193', description='microRNA 5193', other_designations='hsa-mir-5193', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49806137, end_position_on_the_genomic_accession=49806245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847080, CurrentID=0, Status='live', Symbol='MIR5190', Aliases='', description='microRNA 5190', other_designations='hsa-mir-5190', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13459947, end_position_on_the_genomic_accession=13460026, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847081, CurrentID=0, Status='live', Symbol='MIR5703', Aliases='mir-5703', description='microRNA 5703', other_designations='hsa-mir-5703', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227472132, end_position_on_the_genomic_accession=227472187, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847082, CurrentID=0, Status='live', Symbol='MIR5692C1', Aliases='', description='microRNA 5692c-1', other_designations='hsa-mir-5692c-1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135802985, end_position_on_the_genomic_accession=135803075, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847083, CurrentID=0, Status='live', Symbol='MIR548AV', Aliases='', description='microRNA 548av', other_designations='hsa-mir-548av', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72853321, end_position_on_the_genomic_accession=72853382, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847084, CurrentID=0, Status='live', Symbol='MIR548AP', Aliases='', description='microRNA 548ap', other_designations='hsa-mir-548ap', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85825635, end_position_on_the_genomic_accession=85825730, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847085, CurrentID=0, Status='live', Symbol='MIR5706', Aliases='', description='microRNA 5706', other_designations='hsa-mir-5706', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119154637, end_position_on_the_genomic_accession=119154716, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847086, CurrentID=0, Status='live', Symbol='MIR5699', Aliases='mir-5699', description='microRNA 5699', other_designations='hsa-mir-5699', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=641689, end_position_on_the_genomic_accession=641778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847087, CurrentID=0, Status='live', Symbol='MIR5192', Aliases='', description='microRNA 5192', other_designations='hsa-mir-5192', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62205826, end_position_on_the_genomic_accession=62205917, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847088, CurrentID=0, Status='live', Symbol='MIR5586', Aliases='', description='microRNA 5586', other_designations='hsa-mir-5586', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59646962, end_position_on_the_genomic_accession=59647020, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847089, CurrentID=0, Status='live', Symbol='MIR5584', Aliases='', description='microRNA 5584', other_designations='hsa-mir-5584', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44545493, end_position_on_the_genomic_accession=44545552, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847090, CurrentID=0, Status='live', Symbol='MIR5187', Aliases='mir-5187', description='microRNA 5187', other_designations='hsa-mir-5187', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161227186, end_position_on_the_genomic_accession=161227261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847091, CurrentID=0, Status='live', Symbol='MIR5681B', Aliases='', description='microRNA 5681b', other_designations='hsa-mir-5681b', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74548550, end_position_on_the_genomic_accession=74548609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100847092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847092, CurrentID=0, Status='live', Symbol='MIR548AS', Aliases='', description='microRNA 548as', other_designations='hsa-mir-548as', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92490163, end_position_on_the_genomic_accession=92490220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100847093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847093, CurrentID=0, Status='live', Symbol='MIR5589', Aliases='', description='microRNA 5589', other_designations='hsa-mir-5589', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10038354, end_position_on_the_genomic_accession=10038413, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271061, CurrentID=0, Status='live', Symbol='RPS13P5', Aliases='RPS13_2_201', description='ribosomal protein S13 pseudogene 5', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24968889, end_position_on_the_genomic_accession=24969237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100322878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100322878, CurrentID=0, Status='live', Symbol='SLEB14', Aliases='', description='Systemic lupus erythematosus, susceptiblity to, 14', other_designations='', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613145, no_hdr0=''),
  100322879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100322879, CurrentID=0, Status='live', Symbol='SPG45', Aliases='', description='spastic paraplegia 45 (autosomal recessive)', other_designations='', map_location='10q24.3-q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613162, no_hdr0=''),
  402016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402016, CurrentID=0, Status='live', Symbol='BPIFB9P', Aliases='C20orf115, dJ1187J4.2, dJ1187J4.3', description='BPI fold containing family B member 9, pseudogene', other_designations='BPI fold containing family B, member 9, pseudogene', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33347120, end_position_on_the_genomic_accession=33355044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147811, CurrentID=0, Status='live', Symbol='CBX1P2', Aliases='', description='chromobox 1 pseudogene 2', other_designations='chromobox homolog 1 (HP1 beta homolog Drosophila ) pseudogene|chromobox homolog 1 pseudogene 2', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17098380, end_position_on_the_genomic_accession=17100557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147812, CurrentID=0, Status='live', Symbol='RNU7-25P', Aliases='U7.25', description='RNA, U7 small nuclear 25 pseudogene', other_designations='RNA, small nuclear U7.25 pseudogene', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3093569, end_position_on_the_genomic_accession=3093631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147813, CurrentID=0, Status='live', Symbol='RNU7-26P', Aliases='U7.26', description='RNA, U7 small nuclear 26 pseudogene', other_designations='RNA, small nuclear U7.26 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27897506, end_position_on_the_genomic_accession=27897565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285453, CurrentID=0, Status='live', Symbol='LOC285453', Aliases='', description='TXNDC9 pseudogene', other_designations='thioredoxin domain containing 9 pseudogene', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56760912, end_position_on_the_genomic_accession=56761610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147814, CurrentID=0, Status='live', Symbol='RNU7-27P', Aliases='U7.27', description='RNA, U7 small nuclear 27 pseudogene', other_designations='RNA, small nuclear U7.27 pseudogene', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128344082, end_position_on_the_genomic_accession=128344139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147815, CurrentID=0, Status='live', Symbol='RNU7-28P', Aliases='U7.28', description='RNA, U7 small nuclear 28 pseudogene', other_designations='RNA, small nuclear U7.28 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9843817, end_position_on_the_genomic_accession=9843875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147816, CurrentID=0, Status='live', Symbol='RNU7-29P', Aliases='U7.29', description='RNA, U7 small nuclear 29 pseudogene', other_designations='RNA, small nuclear U7.29 pseudogene', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28144156, end_position_on_the_genomic_accession=28144214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147817, CurrentID=0, Status='live', Symbol='RNU7-30P', Aliases='U7.30', description='RNA, U7 small nuclear 30 pseudogene', other_designations='RNA, small nuclear U7.30 pseudogene', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91091076, end_position_on_the_genomic_accession=91091136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147819, CurrentID=0, Status='live', Symbol='RNU7-32P', Aliases='U7.32', description='RNA, U7 small nuclear 32 pseudogene', other_designations='RNA, small nuclear U7.32 pseudogene', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170160936, end_position_on_the_genomic_accession=170160994, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927823, CurrentID=0, Status='live', Symbol='LOC101927823', Aliases='', description='uncharacterized LOC101927823', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89817828, end_position_on_the_genomic_accession=89821753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100147820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147820, CurrentID=0, Status='live', Symbol='RNU7-33P', Aliases='U7.33', description='RNA, U7 small nuclear 33 pseudogene', other_designations='RNA, small nuclear U7.33 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3078484, end_position_on_the_genomic_accession=3078542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287647, CurrentID=0, Status='live', Symbol='PCMTD1P2', Aliases='', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 1 pseudogene 2', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34139179, end_position_on_the_genomic_accession=34139541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147821, CurrentID=0, Status='live', Symbol='RNU7-34P', Aliases='U7.34', description='RNA, U7 small nuclear 34 pseudogene', other_designations='RNA, small nuclear U7.34 pseudogene', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118758212, end_position_on_the_genomic_accession=118758273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189324, CurrentID=0, Status='live', Symbol='TRI-TAT1-1', Aliases='TRNAI18', description='transfer RNA-Ile (TAT) 1-1', other_designations='transfer RNA isoleucine 18 (anticodon UAU)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39412168, end_position_on_the_genomic_accession=39412260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  139886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139886, CurrentID=0, Status='live', Symbol='SPIN4', Aliases='', description='spindlin family member 4', other_designations='spindlin family, member 4', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63347228, end_position_on_the_genomic_accession=63351339, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100147823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147823, CurrentID=0, Status='live', Symbol='RNU7-36P', Aliases='RNU7-112P, U7.36', description='RNA, U7 small nuclear 36 pseudogene', other_designations='RNA, U7 small nuclear 112 pseudogene|RNA, small nuclear U7.36 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34018266, end_position_on_the_genomic_accession=34018320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403297, CurrentID=0, Status='live', Symbol='OR7E155P', Aliases='', description='olfactory receptor family 7 subfamily E member 155 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 155 pseudogene', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41439837, end_position_on_the_genomic_accession=41440862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147824, CurrentID=0, Status='live', Symbol='RNU7-37P', Aliases='U7.37', description='RNA, U7 small nuclear 37 pseudogene', other_designations='RNA, small nuclear U7.37 pseudogene', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53130482, end_position_on_the_genomic_accession=53130540, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147825, CurrentID=0, Status='live', Symbol='RNU7-38P', Aliases='U7.38', description='RNA, U7 small nuclear 38 pseudogene', other_designations='RNA, small nuclear U7.38 pseudogene', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71844737, end_position_on_the_genomic_accession=71844792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147826, CurrentID=0, Status='live', Symbol='RNU7-39P', Aliases='U7.39', description='RNA, U7 small nuclear 39 pseudogene', other_designations='RNA, small nuclear U7.39 pseudogene', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51010999, end_position_on_the_genomic_accession=51011057, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147827, CurrentID=0, Status='live', Symbol='RNU7-40P', Aliases='U7.40', description='RNA, U7 small nuclear 40 pseudogene', other_designations='RNA, small nuclear U7.40 pseudogene', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56352390, end_position_on_the_genomic_accession=56352449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147828, CurrentID=0, Status='live', Symbol='RNU7-41P', Aliases='U7.41', description='RNA, U7 small nuclear 41 pseudogene', other_designations='RNA, small nuclear U7.41 pseudogene', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34845301, end_position_on_the_genomic_accession=34845360, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642335, CurrentID=0, Status='live', Symbol='LOC642335', Aliases='', description='uncharacterized LOC642335', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=481199, end_position_on_the_genomic_accession=483632, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  285456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285456, CurrentID=0, Status='live', Symbol='RPL34-AS1', Aliases='', description='RPL34 antisense RNA 1 (head to head)', other_designations='RP11-462C24.1|RPL34 antisense RNA 1 (non-protein coding)', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108538190, end_position_on_the_genomic_accession=108620457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100147829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147829, CurrentID=0, Status='live', Symbol='RNU7-42P', Aliases='U7.42', description='RNA, U7 small nuclear 42 pseudogene', other_designations='RNA, small nuclear U7.42 pseudogene', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68926898, end_position_on_the_genomic_accession=68926957, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147830, CurrentID=0, Status='live', Symbol='RNU7-43P', Aliases='RNU7-118P, U7.43', description='RNA, U7 small nuclear 43 pseudogene', other_designations='RNA, U7 small nuclear 118 pseudogene|RNA, small nuclear U7.43 pseudogene', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8511126, end_position_on_the_genomic_accession=8511185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147831, CurrentID=0, Status='live', Symbol='RNU7-44P', Aliases='U7.44', description='RNA, U7 small nuclear 44 pseudogene', other_designations='RNA, small nuclear U7.44 pseudogene', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174566217, end_position_on_the_genomic_accession=174566278, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100147832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147832, CurrentID=0, Status='live', Symbol='RNU7-45P', Aliases='U7.45', description='RNA, U7 small nuclear 45 pseudogene', other_designations='RNA, small nuclear U7.45 pseudogene', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201141903, end_position_on_the_genomic_accession=201141963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147833, CurrentID=0, Status='live', Symbol='RNU7-46P', Aliases='U7.46', description='RNA, U7 small nuclear 46 pseudogene', other_designations='RNA, small nuclear U7.46 pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98840680, end_position_on_the_genomic_accession=98840735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369319, CurrentID=0, Status='live', Symbol='LOC105369319', Aliases='', description='uncharacterized LOC105369319', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60159241, end_position_on_the_genomic_accession=60164400, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100147834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147834, CurrentID=0, Status='live', Symbol='RNU7-47P', Aliases='U7.47', description='RNA, U7 small nuclear 47 pseudogene', other_designations='RNA, small nuclear U7.47 pseudogene', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142801210, end_position_on_the_genomic_accession=142801267, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147836, CurrentID=0, Status='live', Symbol='RNU7-49P', Aliases='U7.49', description='RNA, U7 small nuclear 49 pseudogene', other_designations='RNA, small nuclear U7.49 pseudogene', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14478894, end_position_on_the_genomic_accession=14478950, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100147837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147837, CurrentID=0, Status='live', Symbol='RNU7-50P', Aliases='U7.50', description='RNA, U7 small nuclear 50 pseudogene', other_designations='RNA, small nuclear U7.50 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3781396, end_position_on_the_genomic_accession=3781451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  446207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=446207, CurrentID=0, Status='live', Symbol='ACTBP10', Aliases='', description='actin, beta pseudogene 10', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18998860, end_position_on_the_genomic_accession=19000587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  446209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=446209, CurrentID=0, Status='live', Symbol='RPS26P9', Aliases='RPS26_25_1338', description='ribosomal protein S26 pseudogene 9', other_designations='', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20616644, end_position_on_the_genomic_accession=20616991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378743, CurrentID=0, Status='live', Symbol='LOC105378743', Aliases='', description='uncharacterized LOC105378743', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56194211, end_position_on_the_genomic_accession=56227331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372281, CurrentID=0, Status='live', Symbol='LOC105372281', Aliases='', description='uncharacterized LOC105372281', other_designations='CTD-2265O21.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12826785, end_position_on_the_genomic_accession=12833089, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  402055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402055, CurrentID=0, Status='live', Symbol='SRRD', Aliases='HC/HCC, SRR1L', description='SRR1 domain containing', other_designations='SRR1 domain-containing protein|hepatocellular carcinoma complicating hemochromatosis', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26483884, end_position_on_the_genomic_accession=26491938, orientation='plus', exon_count=7, OMIM=602254, no_hdr0=''),
  105378744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378744, CurrentID=0, Status='live', Symbol='LOC105378744', Aliases='', description='uncharacterized LOC105378744', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56414579, end_position_on_the_genomic_accession=56417085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  402057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402057, CurrentID=0, Status='live', Symbol='RPS17P16', Aliases='RPS17_8_1748', description='ribosomal protein S17 pseudogene 16', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32039465, end_position_on_the_genomic_accession=32039944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23589, CurrentID=0, Status='live', Symbol='CARHSP1', Aliases='CRHSP-24, CSDC1', description='calcium regulated heat stable protein 1', other_designations='calcium regulated heat stable protein 1, 24kDa|calcium-regulated heat-stable protein (24kD)|calcium-regulated heat-stable protein of 24 kDa', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8852942, end_position_on_the_genomic_accession=8869012, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105378745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378745, CurrentID=0, Status='live', Symbol='LOC105378745', Aliases='', description='uncharacterized LOC105378745', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56967536, end_position_on_the_genomic_accession=56987463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100127908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127908, CurrentID=0, Status='live', Symbol='LOC100127908', Aliases='', description='selenoprotein K pseudogene', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131819444, end_position_on_the_genomic_accession=131820123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402069, CurrentID=0, Status='live', Symbol='RPS16P2', Aliases='RPS16_1_197', description='ribosomal protein S16 pseudogene 2', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20155573, end_position_on_the_genomic_accession=20156113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391247, CurrentID=0, Status='live', Symbol='LOC391247', Aliases='', description='GINS complex subunit 2 (Psf2 homolog) pseudogene', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38614363, end_position_on_the_genomic_accession=38615560, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  1769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1769, CurrentID=0, Status='live', Symbol='DNAH8', Aliases='ATPase, hdhc9', description='dynein, axonemal, heavy chain 8', other_designations='axonemal beta dynein heavy chain 8|ciliary dynein heavy chain 8|dynein, axonemal, heavy polypeptide 8', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38715332, end_position_on_the_genomic_accession=39030798, orientation='plus', exon_count=95, OMIM=603337, no_hdr0=''),
  102725119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725119, CurrentID=0, Status='live', Symbol='LOC102725119', Aliases='', description='Ig heavy chain V region 5A-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32915074, end_position_on_the_genomic_accession=32916842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402077, CurrentID=0, Status='live', Symbol='OR7E46P', Aliases='MCEEP, OST379', description='olfactory receptor family 7 subfamily E member 46 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 46 pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71037722, end_position_on_the_genomic_accession=71038666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372282, CurrentID=0, Status='live', Symbol='LOC105372282', Aliases='', description='uncharacterized LOC105372282', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13104205, end_position_on_the_genomic_accession=13105328, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927825, CurrentID=0, Status='live', Symbol='LOC101927825', Aliases='', description='uncharacterized LOC101927825', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6317501, end_position_on_the_genomic_accession=6344772, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105378749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378749, CurrentID=0, Status='live', Symbol='LOC105378749', Aliases='', description='uncharacterized LOC105378749', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56998058, end_position_on_the_genomic_accession=57014700, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100192204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192204, CurrentID=0, Status='live', Symbol='PPIAP30', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 30', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15154722, end_position_on_the_genomic_accession=15155347, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102725121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725121, CurrentID=0, Status='live', Symbol='LOC102725121', Aliases='', description='putative ATP-dependent RNA helicase DDX12', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181045, end_position_on_the_genomic_accession=184254, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378750, CurrentID=0, Status='live', Symbol='LOC105378750', Aliases='', description='uncharacterized LOC105378750', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58739511, end_position_on_the_genomic_accession=58745474, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729769, CurrentID=0, Status='live', Symbol='UQCRHP2', Aliases='', description='ubiquinol-cytochrome c reductase hinge protein pseudogene 2', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25994678, end_position_on_the_genomic_accession=25995271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127909, CurrentID=0, Status='live', Symbol='LOC100127909', Aliases='', description='uncharacterized LOC100127909', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.9', start_position_on_the_genomic_accession=45659988, end_position_on_the_genomic_accession=45661502, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128233, CurrentID=0, Status='live', Symbol='LOC100128233', Aliases='', description='uncharacterized LOC100128233', other_designations='CTD-2509G16.2', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65212893, end_position_on_the_genomic_accession=65222347, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378751, CurrentID=0, Status='live', Symbol='LOC105378751', Aliases='', description='uncharacterized LOC105378751', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58761008, end_position_on_the_genomic_accession=58770704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  402093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402093, CurrentID=0, Status='live', Symbol='AHCYP3', Aliases='', description='adenosylhomocysteinase pseudogene 3', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104395003, end_position_on_the_genomic_accession=104396280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374387, CurrentID=0, Status='live', Symbol='LOC105374387', Aliases='', description='uncharacterized LOC105374387', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402096, CurrentID=0, Status='live', Symbol='LOC402096', Aliases='', description='anaphase-promoting complex subunit 1-like', other_designations='', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106369738, end_position_on_the_genomic_accession=106372422, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376318, CurrentID=0, Status='live', Symbol='LOC105376318', Aliases='', description='uncharacterized LOC105376318', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135678046, end_position_on_the_genomic_accession=135679562, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100192306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192306, CurrentID=0, Status='live', Symbol='JAWAD', Aliases='', description='Microcephaly with digital anomalies', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=251255, no_hdr0=''),
  100192307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192307, CurrentID=0, Status='live', Symbol='MUSQTL1', Aliases='', description='Musical aptitude QTL 1', other_designations='', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612343, no_hdr0=''),
  100192308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192308, CurrentID=0, Status='live', Symbol='DEL2Q31', Aliases='', description='Chromosome 2q31.2 deletion syndrome', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612345, no_hdr0=''),
  100192311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192311, CurrentID=0, Status='live', Symbol='MUSTQTL1', Aliases='', description='Muscle strength quantitative trait locus 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612083, no_hdr0=''),
  100192312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192312, CurrentID=0, Status='live', Symbol='IBD21', Aliases='', description='Inflammatory bowel disease-21', other_designations='', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612354, no_hdr0=''),
  402100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402100, CurrentID=0, Status='live', Symbol='RPL22P11', Aliases='RPL22P15, RPL22P9, RPL22_3_260, RPL22_5_261', description='ribosomal protein L22 pseudogene 11', other_designations='ribosomal protein L22 pseudogene 15|ribosomal protein L22 pseudogene 9', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109735509, end_position_on_the_genomic_accession=109737542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378753, CurrentID=0, Status='live', Symbol='LOC105378753', Aliases='', description='uncharacterized LOC105378753', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58882876, end_position_on_the_genomic_accession=58912038, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53335, CurrentID=0, Status='live', Symbol='BCL11A', Aliases='BCL11A-L-S, BCL11A-XL, BCL11a-M, CTIP1, EVI9, HBFQTL5, ZNF856, BCL11A', description='B-cell CLL/lymphoma 11A (zinc finger protein)', other_designations='B-cell CLL/lymphoma 11A (zinc finger protein) isoform 2|BCL-11A|BCL11A B-cell CLL/lymphoma 11A (zinc finger protein) isoform 1|C2H2-type zinc finger protein|COUP-TF-interacting protein 1|EVI-9|ecotropic viral integration site 9 homolog|ecotropic viral integration site 9 protein homolog|zinc finger protein 856', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60451167, end_position_on_the_genomic_accession=60553498, orientation='minus', exon_count=7, OMIM=606557, no_hdr0=''),
  53336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53336, CurrentID=0, Status='live', Symbol='CPXCR1', Aliases='CT77', description='CPX chromosome region, candidate 1', other_designations='cancer/testis antigen 77', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=88747225, end_position_on_the_genomic_accession=88754785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53338, CurrentID=0, Status='live', Symbol='CAPZA1P', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 1 pseudogene', other_designations='', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=88094920, end_position_on_the_genomic_accession=88097258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53339, CurrentID=0, Status='live', Symbol='BTBD1', Aliases='C15orf1, NS5ATP8', description='BTB (POZ) domain containing 1', other_designations='BTB domain containing 1|HCV NS5A-transactivated protein 8|hepatitis C virus NS5A-transactivated protein 8', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83016423, end_position_on_the_genomic_accession=83067434, orientation='minus', exon_count=9, OMIM=608530, no_hdr0=''),
  53340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53340, CurrentID=0, Status='live', Symbol='SPA17', Aliases='CT22, SP17, SP17-1', description='sperm autoantigenic protein 17', other_designations='cancer/testis antigen 22|sperm protein 17', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124673844, end_position_on_the_genomic_accession=124697545, orientation='plus', exon_count=6, OMIM=608621, no_hdr0=''),
  53341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53341, CurrentID=0, Status='live', Symbol='MHAC', Aliases='', description='microhydranencephaly', other_designations='', map_location='16p13.3-p12.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605013, no_hdr0=''),
  53342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53342, CurrentID=0, Status='live', Symbol='IL17D', Aliases='IL-17D, IL-27, IL27', description='interleukin 17D', other_designations='interleukin 27', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20702875, end_position_on_the_genomic_accession=20723100, orientation='plus', exon_count=6, OMIM=607587, no_hdr0=''),
  53343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53343, CurrentID=0, Status='live', Symbol='NUDT9', Aliases='NUDT10', description='nudix hydrolase 9', other_designations='ADP-ribose diphosphatase|ADP-ribose phosphohydrolase|ADP-ribose pyrosphosphatase NUDT9|ADPR-PPase|adenosine diphosphoribose pyrophosphatase|nucleoside diphosphate linked moiety X-type motif 9|nudix (nucleoside diphosphate linked moiety X)-type motif 9', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87422557, end_position_on_the_genomic_accession=87459454, orientation='plus', exon_count=10, OMIM=606022, no_hdr0=''),
  53344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53344, CurrentID=0, Status='live', Symbol='CHIC1', Aliases='BRX', description='cysteine rich hydrophobic domain 1', other_designations='brain X-linked protein|cystein-rich hydrophobic domain 1|cysteine-rich hydrophobic domain 1 protein', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73563148, end_position_on_the_genomic_accession=73687109, orientation='plus', exon_count=6, OMIM=300922, no_hdr0=''),
  53345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53345, CurrentID=0, Status='live', Symbol='TM6SF2', Aliases='', description='transmembrane 6 superfamily member 2', other_designations='', map_location='19p13.3-p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19264365, end_position_on_the_genomic_accession=19273265, orientation='minus', exon_count=10, OMIM=606563, no_hdr0=''),
  53346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53346, CurrentID=0, Status='live', Symbol='TM6SF1', Aliases='', description='transmembrane 6 superfamily member 1', other_designations='', map_location='15q24-q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83107475, end_position_on_the_genomic_accession=83138015, orientation='plus', exon_count=13, OMIM=606562, no_hdr0=''),
  53347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53347, CurrentID=0, Status='live', Symbol='UBASH3A', Aliases='CLIP4, STS-2, TULA, TULA-1', description='ubiquitin associated and SH3 domain containing A', other_designations='T-cell ubiquitin ligand 1|T-cell ubiquitin ligand protein|cbl-interacting protein 4|suppressor of T-cell receptor signaling 2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42403862, end_position_on_the_genomic_accession=42447688, orientation='plus', exon_count=18, OMIM=605736, no_hdr0=''),
  105372693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372693, CurrentID=0, Status='live', Symbol='LOC105372693', Aliases='', description='uncharacterized LOC105372693', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57985847, end_position_on_the_genomic_accession=57996859, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53349, CurrentID=0, Status='live', Symbol='ZFYVE1', Aliases='DFCP1, PPP1R172, SR3, TAFF1, ZNFN2A1', description='zinc finger, FYVE domain containing 1', other_designations='double FYVE-containing protein 1|phosphoinositide-binding protein SR3|protein phosphatase 1, regulatory subunit 172|tandem FYVE fingers-1 protein|zinc finger protein, subfamily 2A, member 1', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72969445, end_position_on_the_genomic_accession=73027212, orientation='minus', exon_count=14, OMIM=605471, no_hdr0=''),
  53353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53353, CurrentID=0, Status='live', Symbol='LRP1B', Aliases='LRP-DIT, LRPDIT', description='LDL receptor related protein 1B', other_designations='LRP-1B|LRP-deleted in tumors|low density lipoprotein receptor related protein-deleted in tumor|low density lipoprotein receptor-related protein 1B|low density lipoprotein-related protein 1B (deleted in tumors)|low-density lipoprotein receptor-related protein-deleted in tumor', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140231427, end_position_on_the_genomic_accession=142132474, orientation='minus', exon_count=91, OMIM=608766, no_hdr0=''),
  53354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53354, CurrentID=0, Status='live', Symbol='PANK1', Aliases='PANK', description='pantothenate kinase 1', other_designations='pantothenic acid kinase 1', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89579497, end_position_on_the_genomic_accession=89645572, orientation='minus', exon_count=8, OMIM=606160, no_hdr0=''),
  53358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53358, CurrentID=0, Status='live', Symbol='SHC3', Aliases='N-Shc, NSHC, RAI, SHCC', description='SHC (Src homology 2 domain containing) transforming protein 3', other_designations='SH2 domain protein C3|SHC-transforming protein C|neuronal Shc|protein Rai|src homology 2 domain-containing transforming protein C3|src homology 2 domain-containing-transforming protein C3', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89005771, end_position_on_the_genomic_accession=89182072, orientation='minus', exon_count=14, OMIM=605263, no_hdr0=''),
  105369203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369203, CurrentID=0, Status='live', Symbol='LOC105369203', Aliases='', description='uncharacterized LOC105369203', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94064379, end_position_on_the_genomic_accession=94099378, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53366, CurrentID=0, Status='live', Symbol='SM2', Aliases='', description='Hepatic fibrosis susceptibility due to Schistosoma mansoni infection', other_designations='', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604201, no_hdr0=''),
  53368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53368, CurrentID=0, Status='live', Symbol='MDC1B', Aliases='', description='Muscular dystrophy, congenital, 1B', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604801, no_hdr0=''),
  53369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53369, CurrentID=0, Status='live', Symbol='HDL3', Aliases='HLN2', description='Huntington-like neurodegenerative disorder 2', other_designations='', map_location='4p15.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604802, no_hdr0=''),
  100192378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192378, CurrentID=0, Status='live', Symbol='ZFHX4-AS1', Aliases='', description='ZFHX4 antisense RNA 1', other_designations='ZFHX4 antisense RNA 1 (non-protein coding)', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76610879, end_position_on_the_genomic_accession=76683275, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53371, CurrentID=0, Status='live', Symbol='NUP54', Aliases='', description='nucleoporin 54kDa', other_designations='54 kDa nucleoporin|nucleoporin 54kD', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76114659, end_position_on_the_genomic_accession=76148515, orientation='minus', exon_count=14, OMIM=607607, no_hdr0=''),
  53373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53373, CurrentID=0, Status='live', Symbol='TPCN1', Aliases='TPC1', description='two pore segment channel 1', other_designations='two-pore channel 1|two-pore channel 1, homolog|two-pore segment channel 1|voltage-dependent calcium channel protein TPC1', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113221080, end_position_on_the_genomic_accession=113298589, orientation='plus', exon_count=34, OMIM=609666, no_hdr0=''),
  105374766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374766, CurrentID=0, Status='live', Symbol='LOC105374766', Aliases='', description='uncharacterized LOC105374766', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63859209, end_position_on_the_genomic_accession=63868699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  402112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402112, CurrentID=0, Status='live', Symbol='HNRNPA1P39', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 39', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174309619, end_position_on_the_genomic_accession=174311023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100192386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192386, CurrentID=0, Status='live', Symbol='FLJ16779', Aliases='', description='uncharacterized LOC100192386', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63253978, end_position_on_the_genomic_accession=63261615, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100192388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192388, CurrentID=0, Status='live', Symbol='LOC100192388', Aliases='', description='anaphase promoting complex subunit 13 pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11838321, end_position_on_the_genomic_accession=11839978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100192389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192389, CurrentID=0, Status='live', Symbol='LOC100192389', Aliases='', description='anaphase promoting complex subunit 13 pseudogene', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117699429, end_position_on_the_genomic_accession=117700259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53383, CurrentID=0, Status='live', Symbol='CCDC90AP1', Aliases='C17orf23', description='coiled-coil domain containing 90A pseudogene 1', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2127428, end_position_on_the_genomic_accession=2128278, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53400, CurrentID=0, Status='live', Symbol='SMG6-IT1', Aliases='C17orf6', description='SMG6 intronic transcript 1', other_designations='SMG6 intronic transcript 1 (non-protein coding)', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402116, CurrentID=0, Status='live', Symbol='PKP4P1', Aliases='', description='plakophilin 4 pseudogene 1', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209178961, end_position_on_the_genomic_accession=209181263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  53405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53405, CurrentID=0, Status='live', Symbol='CLIC5', Aliases='DFNB102, DFNB103, MST130, MSTP130', description='chloride intracellular channel 5', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45898451, end_position_on_the_genomic_accession=46080348, orientation='minus', exon_count=14, OMIM=607293, no_hdr0=''),
  53407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53407, CurrentID=0, Status='live', Symbol='STX18', Aliases='Ufe1', description='syntaxin 18', other_designations='cell growth-inhibiting gene 9 protein', map_location='4p16.3-p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4418968, end_position_on_the_genomic_accession=4542232, orientation='minus', exon_count=15, OMIM=606046, no_hdr0=''),
  53409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53409, CurrentID=0, Status='live', Symbol='ATP5LP3', Aliases='ATP5KP3, ATPE5KP3', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 3|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit G pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 3|H+-transporting two-sector ATPase chain g', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24657662, end_position_on_the_genomic_accession=24658421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53410, CurrentID=0, Status='live', Symbol='ATP5LP2', Aliases='ATP5KP2', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 2|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 2', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122859603, end_position_on_the_genomic_accession=122860013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53411, CurrentID=0, Status='live', Symbol='ATP5LP1', Aliases='ATP5KP1, ATPE5KP1', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 1|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63834592, end_position_on_the_genomic_accession=63835044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100192420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192420, CurrentID=0, Status='live', Symbol='FLJ41941', Aliases='', description='uncharacterized LOC100192420', other_designations='XXbac-B476C20.17', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18029385, end_position_on_the_genomic_accession=18037968, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100653395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653395, CurrentID=0, Status='live', Symbol='LOC100653395', Aliases='', description='methionine adenosyltransferase II, alpha pseudogene', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112695400, end_position_on_the_genomic_accession=112703407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100192426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192426, CurrentID=0, Status='live', Symbol='LOC100192426', Aliases='', description='uncharacterized LOC100192426', other_designations='', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8360820, end_position_on_the_genomic_accession=8367034, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  402120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402120, CurrentID=0, Status='live', Symbol='RPL23P5', Aliases='RPL23_3_328', description='ribosomal protein L23 pseudogene 5', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221583512, end_position_on_the_genomic_accession=221592931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100169759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169759, CurrentID=0, Status='live', Symbol='RNA5S8', Aliases='RN5S8', description='RNA, 5S ribosomal 8', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228625907, end_position_on_the_genomic_accession=228626027, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  402123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402123, CurrentID=0, Status='live', Symbol='RPSAP32', Aliases='', description='ribosomal protein SA pseudogene 32', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=617789, end_position_on_the_genomic_accession=618389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100192455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192455, CurrentID=0, Status='live', Symbol='FASA', Aliases='', description='fertility associated sperm antigen', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642423, CurrentID=0, Status='live', Symbol='LOC642423', Aliases='', description='golgin A2 pseudogene', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85204676, end_position_on_the_genomic_accession=85205287, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  402125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402125, CurrentID=0, Status='live', Symbol='HSPA8P18', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 18', other_designations="heat shock 70kDa protein 6 (HSP70B') pseudogene|heat shock 70kDa protein 8 pseudogene 18", map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19940362, end_position_on_the_genomic_accession=19941573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100169765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169765, CurrentID=0, Status='live', Symbol='RNA5S14', Aliases='RN5S14', description='RNA, 5S ribosomal 14', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228639335, end_position_on_the_genomic_accession=228639455, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378759, CurrentID=0, Status='live', Symbol='LOC105378759', Aliases='', description='uncharacterized LOC105378759', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59913792, end_position_on_the_genomic_accession=59918682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  402133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402133, CurrentID=0, Status='live', Symbol='OR7E121P', Aliases='', description='olfactory receptor family 7 subfamily E member 121 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 121 pseudogene', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75598650, end_position_on_the_genomic_accession=75599672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402134, CurrentID=0, Status='live', Symbol='KRT8P25', Aliases='', description='keratin 8 pseudogene 25', other_designations='', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87323196, end_position_on_the_genomic_accession=87324906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375776, CurrentID=0, Status='live', Symbol='LOC105375776', Aliases='', description='uncharacterized LOC105375776', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=137763880, end_position_on_the_genomic_accession=137782112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420666, CurrentID=0, Status='live', Symbol='LOC100420666', Aliases='', description='lysine (K)-specific demethylase 3A pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189486476, end_position_on_the_genomic_accession=189488334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402143, CurrentID=0, Status='live', Symbol='HSPA8P9', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 9', other_designations='heat shock 70kDa protein 8 pseudogene 9', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137880210, end_position_on_the_genomic_accession=137881692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402146, CurrentID=0, Status='live', Symbol='B3GAT3P1', Aliases='B3GAT3P, PSIGLCAT-I', description='beta-1,3-glucuronyltransferase 3 pseudogene 1', other_designations='beta-1,3-glucuronyltransferase 3 (glucuronosyltransferase I) pseudogene', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160452368, end_position_on_the_genomic_accession=160453809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140004, CurrentID=0, Status='live', Symbol='TUBB1P2', Aliases='', description='tubulin beta 1 class VI pseudogene 2', other_designations='tubulin, beta 1 class VI pseudogene 2|tubulin, beta 1 pseudogene 2|tubulin, beta polypeptide 4, member Q pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17516596, end_position_on_the_genomic_accession=17519431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402149, CurrentID=0, Status='live', Symbol='RPL28P1', Aliases='RPL28_2_436, bcm2489', description='ribosomal protein L28 pseudogene 1', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170653807, end_position_on_the_genomic_accession=170654301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402150, CurrentID=0, Status='live', Symbol='RPL8P4', Aliases='RPL8_1_438', description='ribosomal protein L8 pseudogene 4', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=174377209, end_position_on_the_genomic_accession=174378040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378763, CurrentID=0, Status='live', Symbol='LOC105378763', Aliases='', description='uncharacterized LOC105378763', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60529214, end_position_on_the_genomic_accession=60622546, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  53615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53615, CurrentID=0, Status='live', Symbol='MBD3', Aliases='', description='methyl-CpG binding domain protein 3', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1576671, end_position_on_the_genomic_accession=1592761, orientation='minus', exon_count=7, OMIM=603573, no_hdr0=''),
  53616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53616, CurrentID=0, Status='live', Symbol='ADAM22', Aliases='ADAM 22, MDC2', description='ADAM metallopeptidase domain 22', other_designations='a disintegrin and metalloproteinase domain 22|metalloproteinase-disintegrin ADAM22-3|metalloproteinase-like, disintegrin-like, and cysteine-rich protein 2', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87934157, end_position_on_the_genomic_accession=88202889, orientation='plus', exon_count=35, OMIM=603709, no_hdr0=''),
  402152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402152, CurrentID=0, Status='live', Symbol='RPL7L1P8', Aliases='', description='ribosomal protein L7 like 1 pseudogene 8', other_designations='ribosomal protein L7-like 1 pseudogene 8', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182610669, end_position_on_the_genomic_accession=182611422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374399, CurrentID=0, Status='live', Symbol='LOC105374399', Aliases='', description='uncharacterized LOC105374399', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36355418, end_position_on_the_genomic_accession=36365931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53630, CurrentID=0, Status='live', Symbol='BCO1', Aliases='BCDO, BCDO1, BCMO, BCMO1, BCO', description='beta-carotene oxygenase 1', other_designations="beta-carotene 15, 15'-dioxygenase 1|beta-carotene 15,15'-monooxygenase 1|beta-carotene dioxygenase 1", map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81238642, end_position_on_the_genomic_accession=81291142, orientation='plus', exon_count=11, OMIM=605748, no_hdr0=''),
  53632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53632, CurrentID=0, Status='live', Symbol='PRKAG3', Aliases='AMPKG3', description='protein kinase, AMP-activated, gamma 3 non-catalytic subunit', other_designations="5'-AMP-activated protein kinase, gamma-3 subunit|AMPK gamma-3 chain|AMPK gamma3|AMPK subunit gamma-3", map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218822383, end_position_on_the_genomic_accession=218831789, orientation='minus', exon_count=14, OMIM=604976, no_hdr0=''),
  53635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53635, CurrentID=0, Status='live', Symbol='PTOV1', Aliases='ACID2, PTOV-1', description='prostate tumor overexpressed 1', other_designations='activator interaction domain-containing protein 2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49848844, end_position_on_the_genomic_accession=49860742, orientation='plus', exon_count=16, OMIM=610195, no_hdr0=''),
  53637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53637, CurrentID=0, Status='live', Symbol='S1PR5', Aliases='EDG8, Edg-8, S1P5, SPPR-1, SPPR-2', description='sphingosine-1-phosphate receptor 5', other_designations='S1P receptor 5|S1P receptor Edg-8|endothelial differentiation G-protein-coupled receptor 8|endothelial differentiation, sphingolipid G-protein-coupled receptor, 8|sphingosine 1-phosphate receptor EDG8|sphingosine 1-phosphate receptor Edg-8', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10512742, end_position_on_the_genomic_accession=10517992, orientation='minus', exon_count=3, OMIM=605146, no_hdr0=''),
  442603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442603, CurrentID=0, Status='live', Symbol='LOC442603', Aliases='', description='ADP ribosylation factor 6 pseudogene', other_designations='ADP-ribosylation factor 6 pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99635167, end_position_on_the_genomic_accession=99639023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643371, CurrentID=0, Status='live', Symbol='LOC643371', Aliases='', description='tubulin, beta 2B class IIb pseudogene', other_designations='tubulin, beta 8 class VIII pseudogene|tubulin, beta 8 pseudogene', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=94319037, end_position_on_the_genomic_accession=94320117, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402160, CurrentID=0, Status='live', Symbol='CFAP99', Aliases='', description='cilia and flagella associated protein 99', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2418893, end_position_on_the_genomic_accession=2462963, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  102682016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102682016, CurrentID=0, Status='live', Symbol='LINC01159', Aliases='linc-Brn1b', description='long intergenic non-protein coding RNA 1159', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104865497, end_position_on_the_genomic_accession=104872595, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128248, CurrentID=0, Status='live', Symbol='LOC100128248', Aliases='', description='mesoderm development candidate 2 pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45444796, end_position_on_the_genomic_accession=45445435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  132724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132724, CurrentID=0, Status='live', Symbol='TMPRSS11B', Aliases='', description='transmembrane protease, serine 11B', other_designations='airway trypsin-like protease 5', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68226653, end_position_on_the_genomic_accession=68245694, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  729852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729852, CurrentID=0, Status='live', Symbol='UMAD1', Aliases='RPA3-AS1, RPA3OS', description='UBAP1-MVB12-associated (UMA) domain containing 1', other_designations='RPA3 antisense RNA 1|RPA3 opposite strand', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7640636, end_position_on_the_genomic_accession=7879223, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102725139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725139, CurrentID=0, Status='live', Symbol='LOC102725139', Aliases='', description='uncharacterized LOC102725139', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402175, CurrentID=0, Status='live', Symbol='NDUFB4P12', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B4 pseudogene 12', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 4, 15kDa pseudogene', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43898915, end_position_on_the_genomic_accession=43899409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402176, CurrentID=0, Status='live', Symbol='RPL21P44', Aliases='RPL21_17_477', description='ribosomal protein L21 pseudogene 44', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53985486, end_position_on_the_genomic_accession=53987282, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132615, CurrentID=0, Status='live', Symbol='LOC100132615', Aliases='', description='chromosome 12 open reading frame 49 pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43728233, end_position_on_the_genomic_accession=43728922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729859, CurrentID=0, Status='live', Symbol='LOC729859', Aliases='', description='WW domain binding protein 1 pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87972501, end_position_on_the_genomic_accession=87973694, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928702, CurrentID=0, Status='live', Symbol='LOC101928702', Aliases='', description='uncharacterized LOC101928702', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74992238, end_position_on_the_genomic_accession=74999166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100847025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847025, CurrentID=0, Status='live', Symbol='MIR5583-1', Aliases='', description='microRNA 5583-1', other_designations='hsa-mir-5583-1', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39676721, end_position_on_the_genomic_accession=39676779, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23614, CurrentID=0, Status='live', Symbol='PPY2P', Aliases='PPY2', description='pancreatic polypeptide 2, pseudogene', other_designations='', map_location='17q11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28247444, end_position_on_the_genomic_accession=28248289, orientation='plus', exon_count=2, OMIM=606638, no_hdr0=''),
  402184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402184, CurrentID=0, Status='live', Symbol='RPS12P8', Aliases='RPS12_4_508', description='ribosomal protein S12 pseudogene 8', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112116490, end_position_on_the_genomic_accession=112116984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53820, CurrentID=0, Status='live', Symbol='RIPPLY3', Aliases='DSCR6', description='ripply transcriptional repressor 3', other_designations='Down syndrome critical region gene 6|down syndrome critical region protein 6|ripply3 homolog', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37006156, end_position_on_the_genomic_accession=37019659, orientation='plus', exon_count=5, OMIM=609892, no_hdr0=''),
  53822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53822, CurrentID=0, Status='live', Symbol='FXYD7', Aliases='', description='FXYD domain containing ion transport regulator 7', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35143250, end_position_on_the_genomic_accession=35154302, orientation='plus', exon_count=6, OMIM=606684, no_hdr0=''),
  53826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53826, CurrentID=0, Status='live', Symbol='FXYD6', Aliases='', description='FXYD domain containing ion transport regulator 6', other_designations='phosphohippolin', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117836976, end_position_on_the_genomic_accession=117877486, orientation='minus', exon_count=11, OMIM=606683, no_hdr0=''),
  53827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53827, CurrentID=0, Status='live', Symbol='FXYD5', Aliases='DYSAD, HSPC113, IWU1, KCT1, OIT2, PRO6241, RIC', description='FXYD domain containing ion transport regulator 5', other_designations='dysadherin|keratinocytes associated transmembrane protein 1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35154716, end_position_on_the_genomic_accession=35169885, orientation='plus', exon_count=10, OMIM=606669, no_hdr0=''),
  53828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53828, CurrentID=0, Status='live', Symbol='FXYD4', Aliases='CHIF', description='FXYD domain containing ion transport regulator 4', other_designations='channel-inducing factor', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43371644, end_position_on_the_genomic_accession=43376335, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53829, CurrentID=0, Status='live', Symbol='P2RY13', Aliases='FKSG77, GPCR1, GPR86, GPR94, P2Y13, SP174', description='purinergic receptor P2Y, G-protein coupled, 13', other_designations='G protein-coupled receptor 86|G-protein coupled receptor 86|G-protein coupled receptor 94|purinergic receptor P2Y G-protein coupled 13', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151326308, end_position_on_the_genomic_accession=151329549, orientation='minus', exon_count=2, OMIM=606380, no_hdr0=''),
  53831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53831, CurrentID=0, Status='live', Symbol='GPR84', Aliases='EX33, GPCR4', description='G protein-coupled receptor 84', other_designations='inflammation-related G protein-coupled receptor EX33|inflammation-related G-protein coupled receptor EX33', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54350784, end_position_on_the_genomic_accession=54365253, orientation='minus', exon_count=3, OMIM=606383, no_hdr0=''),
  53832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53832, CurrentID=0, Status='live', Symbol='IL20RA', Aliases='CRF2-8, IL-20R-alpha, IL-20R1, IL-20RA', description='interleukin 20 receptor subunit alpha', other_designations='class II cytokine receptor ZCYTOR7|cytokine receptor family 2 member 8|interleukin 20 receptor alpha subunit|interleukin 20 receptor, alpha|interleukin-20 receptor I', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136999971, end_position_on_the_genomic_accession=137045366, orientation='minus', exon_count=8, OMIM=605620, no_hdr0=''),
  53833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53833, CurrentID=0, Status='live', Symbol='IL20RB', Aliases='DIRS1, FNDC6, IL-20R2', description='interleukin 20 receptor subunit beta', other_designations='IL-20 receptor subunit beta|IL-20R-beta|IL-20RB|fibronectin type III domain containing 6|interleukin 20 receptor beta subunit|interleukin-20 receptor II', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136957865, end_position_on_the_genomic_accession=137011085, orientation='plus', exon_count=9, OMIM=605621, no_hdr0=''),
  53834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53834, CurrentID=0, Status='live', Symbol='FGFRL1', Aliases='FGFR5, FHFR', description='fibroblast growth factor receptor-like 1', other_designations='FGF homologous factor receptor|FGF receptor-like protein 1|FGFR-5|FGFR-like protein|fibroblast growth factor receptor 5', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1009979, end_position_on_the_genomic_accession=1026898, orientation='plus', exon_count=11, OMIM=605830, no_hdr0=''),
  100847031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847031, CurrentID=0, Status='live', Symbol='MIR5700', Aliases='', description='microRNA 5700', other_designations='hsa-mir-5700', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94561789, end_position_on_the_genomic_accession=94561859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53836, CurrentID=0, Status='live', Symbol='GPR87', Aliases='FKSG78, GPR95, KPG_002', description='G protein-coupled receptor 87', other_designations='G protein-coupled receptor 95|G-protein coupled receptor 95|orphan GPCR 87', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151294086, end_position_on_the_genomic_accession=151316848, orientation='minus', exon_count=3, OMIM=606379, no_hdr0=''),
  53838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53838, CurrentID=0, Status='live', Symbol='C11orf24', Aliases='DM4E3', description='chromosome 11 open reading frame 24', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68261335, end_position_on_the_genomic_accession=68272001, orientation='minus', exon_count=6, OMIM=610880, no_hdr0=''),
  53840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53840, CurrentID=0, Status='live', Symbol='TRIM34', Aliases='IFP1, RNF21', description='tripartite motif containing 34', other_designations='interferon-responsive finger protein 1|ring finger protein 21, interferon-responsive|tripartite motif-containing 34', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5619944, end_position_on_the_genomic_accession=5644398, orientation='plus', exon_count=9, OMIM=605684, no_hdr0=''),
  53841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53841, CurrentID=0, Status='live', Symbol='CDHR5', Aliases='MU-PCDH, MUCDHL, MUPCDH', description='cadherin related family member 5', other_designations='mu-protocadherin|mucin and cadherin-like protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=616565, end_position_on_the_genomic_accession=625067, orientation='minus', exon_count=15, OMIM=606839, no_hdr0=''),
  53842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53842, CurrentID=0, Status='live', Symbol='CLDN22', Aliases='CLDN21', description='claudin 22', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183318067, end_position_on_the_genomic_accession=183320774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53844, CurrentID=0, Status='live', Symbol='COPG2IT1', Aliases='CIT1, COPG2AS, DKFZP761N09121, NCRNA00170', description='COPG2 imprinted transcript 1 (non-protein coding)', other_designations='COPG2 antisense RNA (non-protein coding)|Copg2 intronic transcript 1|coatomer protein complex, subunit gamma 2, intronic transcript 1', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130543453, end_position_on_the_genomic_accession=130546900, orientation='plus', exon_count=1, OMIM=610581, no_hdr0=''),
  402191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402191, CurrentID=0, Status='live', Symbol='NOL8P1', Aliases='', description='nucleolar protein 8 pseudogene 1', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165547296, end_position_on_the_genomic_accession=165551232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  402192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402192, CurrentID=0, Status='live', Symbol='LOC402192', Aliases='', description='chloride channel, nucleotide-sensitive, 1A pseudogene', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169832564, end_position_on_the_genomic_accession=169833396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  402193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402193, CurrentID=0, Status='live', Symbol='RPL5P11', Aliases='RPL5_6_536', description='ribosomal protein L5 pseudogene 11', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173419187, end_position_on_the_genomic_accession=173420210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374407, CurrentID=0, Status='live', Symbol='LOC105374407', Aliases='', description='uncharacterized LOC105374407', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37929205, end_position_on_the_genomic_accession=37952375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100847037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847037, CurrentID=0, Status='live', Symbol='MIR5001', Aliases='mir-5001', description='microRNA 5001', other_designations='hsa-mir-5001', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232550474, end_position_on_the_genomic_accession=232550573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375487, CurrentID=0, Status='live', Symbol='LOC105375487', Aliases='', description='uncharacterized LOC105375487', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125845208, end_position_on_the_genomic_accession=125854216, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130532, CurrentID=0, Status='live', Symbol='LOC100130532', Aliases='', description='F-box protein 38 pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36506043, end_position_on_the_genomic_accession=36508866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100169835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169835, CurrentID=0, Status='live', Symbol='RNY5P7', Aliases='', description='RNA, Ro-associated Y5 pseudogene 7', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101166209, end_position_on_the_genomic_accession=101166283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100169836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169836, CurrentID=0, Status='live', Symbol='RNY5P9', Aliases='', description='RNA, Ro-associated Y5 pseudogene 9', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58728596, end_position_on_the_genomic_accession=58728676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374408, CurrentID=0, Status='live', Symbol='LOC105374408', Aliases='', description='uncharacterized LOC105374408', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38233564, end_position_on_the_genomic_accession=38243537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100169837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100169837, CurrentID=0, Status='live', Symbol='RNY5P10', Aliases='', description='RNA, Ro-associated Y5 pseudogene 10', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36049146, end_position_on_the_genomic_accession=36049219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  53904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53904, CurrentID=0, Status='live', Symbol='MYO3A', Aliases='DFNB30', description='myosin IIIA', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25934073, end_position_on_the_genomic_accession=26212536, orientation='plus', exon_count=46, OMIM=606808, no_hdr0=''),
  53905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53905, CurrentID=0, Status='live', Symbol='DUOX1', Aliases='LNOX1, NOXEF1, THOX1', description='dual oxidase 1', other_designations='NADPH thyroid oxidase 1|flavoprotein NADPH oxidase|large NOX 1|long NOX 1|nicotinamide adenine dinucleotide phosphate oxidase|thyroid oxidase 1', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45129992, end_position_on_the_genomic_accession=45165578, orientation='plus', exon_count=35, OMIM=606758, no_hdr0=''),
  53916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53916, CurrentID=0, Status='live', Symbol='RAB4B', Aliases='', description='RAB4B, member RAS oncogene family', other_designations='ras-related GTP-binding protein 4b|small GTP binding protein RAB4B', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40778219, end_position_on_the_genomic_accession=40796944, orientation='plus', exon_count=8, OMIM=612945, no_hdr0=''),
  53917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53917, CurrentID=0, Status='live', Symbol='RAB24', Aliases='', description='RAB24, member RAS oncogene family', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177301190, end_position_on_the_genomic_accession=177303744, orientation='minus', exon_count=8, OMIM=612415, no_hdr0=''),
  53918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53918, CurrentID=0, Status='live', Symbol='PELO', Aliases='CGI-17, PRO1770', description='pelota homolog (Drosophila)', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52787940, end_position_on_the_genomic_accession=52802617, orientation='plus', exon_count=3, OMIM=605757, no_hdr0=''),
  53919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53919, CurrentID=0, Status='live', Symbol='SLCO1C1', Aliases='OATP-F, OATP1, OATP14, OATP1C1, OATPF, OATPRP5, SLC21A14', description='solute carrier organic anion transporter family member 1C1', other_designations='OAT-RP-5|OATP-14|organic anion transporter F|organic anion transporter polypeptide-related protein 5|organic anion transporting polypeptide 14|organic anion-transporting polypeptide 14|solute carrier family 21 (organic anion transporter), member 14|solute carrier family 21 member 14|solute carrier organic anion transporter family, member 1C1|thyroxine transporter', map_location='12p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20695355, end_position_on_the_genomic_accession=20753386, orientation='plus', exon_count=20, OMIM=613389, no_hdr0=''),
  102725145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725145, CurrentID=0, Status='live', Symbol='LOC102725145', Aliases='', description='uncharacterized LOC102725145', other_designations='CTC-575I10.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73777601, end_position_on_the_genomic_accession=73787032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=30833, CurrentID=0, Status='live', Symbol='NT5C', Aliases='DNT, DNT1, HEL74, P5N2, PN-I, PN-II, UMPH2, cdN, dNT-1', description="5', 3'-nucleotidase, cytosolic", other_designations="5' nucleotidase, deoxy (pyrimidine), cytosolic type C|cytosolic 5',3'-pyrimidine nucleotidase|deoxy-5'-nucleotidase 1|epididymis luminal protein 74|uridine 5'-monophosphate phosphohydrolase 2|uridine 5-prime monophosphate hydrolase 2", map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75130225, end_position_on_the_genomic_accession=75131795, orientation='minus', exon_count=5, OMIM=191720, no_hdr0=''),
  53938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53938, CurrentID=0, Status='live', Symbol='PPIL3', Aliases='CYPJ', description='peptidylprolyl isomerase like 3', other_designations='PPIase|PPIase-like protein 3|cyclophilin J|cyclophilin-like protein 3|cyclophilin-like protein PPIL3|peptidylprolyl cis-trans isomerase-like protein 3|peptidylprolyl isomerase (cyclophilin)-like 3|peptidylprolyl isomerase -like 3|rotamase PPIL3', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200870907, end_position_on_the_genomic_accession=200889333, orientation='minus', exon_count=9, OMIM=615811, no_hdr0=''),
  53940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53940, CurrentID=0, Status='live', Symbol='FTHL17', Aliases='CT38', description='ferritin, heavy polypeptide-like 17', other_designations='cancer/testis antigen 38', map_location='Xp21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=31071241, end_position_on_the_genomic_accession=31072053, orientation='minus', exon_count=1, OMIM=300308, no_hdr0=''),
  53942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53942, CurrentID=0, Status='live', Symbol='CNTN5', Aliases='HNB-2s, NB-2', description='contactin 5', other_designations='neural adhesion molecule|neural recognition molecule NB-2', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=99020976, end_position_on_the_genomic_accession=100358885, orientation='plus', exon_count=26, OMIM=607219, no_hdr0=''),
  53944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53944, CurrentID=0, Status='live', Symbol='CSNK1G1', Aliases='CK1gamma1', description='casein kinase 1, gamma 1', other_designations='CKI-gamma 1', map_location='15q22.1-q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64165517, end_position_on_the_genomic_accession=64356243, orientation='minus', exon_count=12, OMIM=606274, no_hdr0=''),
  100420679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420679, CurrentID=0, Status='live', Symbol='LOC100420679', Aliases='', description='signal recognition particle 19kDa pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=242391, end_position_on_the_genomic_accession=243689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  53947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53947, CurrentID=0, Status='live', Symbol='A4GALT', Aliases='A14GALT1, Gb3S, P(k), P1, P1PK, PK, A4GALT', description='alpha 1,4-galactosyltransferase', other_designations='CD77 synthase|GB3 synthase|P blood group (P one antigen)|P(k) antigen synthase|P1/Pk synthase|UDP-galactose:beta-D-galactosyl-beta1-R 4-alpha-D-galactosyltransferase|alpha 14-galactosyltransferase|alpha-1,4-N-acetylglucosaminyltransferase|alpha4Gal-T1|globotriaosylceramide synthase', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42692117, end_position_on_the_genomic_accession=42721301, orientation='minus', exon_count=8, OMIM=607922, no_hdr0=''),
  105379677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379677, CurrentID=0, Status='live', Symbol='LOC105379677', Aliases='', description='uncharacterized LOC105379677', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132621, CurrentID=0, Status='live', Symbol='LOC100132621', Aliases='', description='mitochondrial fission regulator 2 pseudogene', other_designations='family with sequence similarity 54, member A pseudogene', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39750127, end_position_on_the_genomic_accession=39750829, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375980, CurrentID=0, Status='live', Symbol='LOC105375980', Aliases='', description='uncharacterized LOC105375980', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14987172, end_position_on_the_genomic_accession=14993254, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=53981, CurrentID=0, Status='live', Symbol='CPSF2', Aliases='CPSF100', description='cleavage and polyadenylation specific factor 2', other_designations='CPSF 100 kDa subunit|CPSF 100kDa subunit|cleavage and polyadenylation specific factor 2, 100kDa|cleavage and polyadenylation specificity factor 100 kDa subunit', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92121954, end_position_on_the_genomic_accession=92164199, orientation='plus', exon_count=16, OMIM=606028, no_hdr0=''),
  402213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402213, CurrentID=0, Status='live', Symbol='LOC402213', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42892119, end_position_on_the_genomic_accession=42892858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642515, CurrentID=0, Status='live', Symbol='LOC642515', Aliases='', description='proline-rich transmembrane protein 1-like', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131549603, end_position_on_the_genomic_accession=131559199, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54012, CurrentID=0, Status='live', Symbol='ZNF299P', Aliases='ZNF2991, ZNF299P', description='zinc finger protein 299, pseudogene', other_designations='zinc finger protein 299 pseudogene 1|zinc finger protein 670 pseudogene|zinc finger-like processed pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23090471, end_position_on_the_genomic_accession=23091744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54014, CurrentID=0, Status='live', Symbol='BRWD1', Aliases='C21orf107, N143, WDR9', description='bromodomain and WD repeat domain containing 1', other_designations='WD repeat protein WDR9-form2|WD repeat-containing protein 9|transcriptional unit N143', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39184176, end_position_on_the_genomic_accession=39313819, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  54015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54015, CurrentID=0, Status='live', Symbol='VDAC2P1', Aliases='VDAC2P', description='voltage-dependent anion channel 2 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16094189, end_position_on_the_genomic_accession=16095472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54016, CurrentID=0, Status='live', Symbol='TUBAP', Aliases='', description='tubulin alpha pseudogene', other_designations='alpha tubulin (TUBA2) processed pseudogene|tubulin, alpha pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23407492, end_position_on_the_genomic_accession=23409222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54019, CurrentID=0, Status='live', Symbol='SLC6A6P1', Aliases='SLC6A6P', description='solute carrier family 6 member 6 pseudogene 1', other_designations='solute carrier family 6, member 6 pseudogene 1|taurine transporter processed pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19244532, end_position_on_the_genomic_accession=19245133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54020, CurrentID=0, Status='live', Symbol='SLC37A1', Aliases='G3PP', description='solute carrier family 37 (glucose-6-phosphate transporter), member 1', other_designations='G-3-P permease|G-3-P transporter|glycerol-3-phosphate permease|solute carrier family 37 (glycerol-3-phosphate transporter), member 1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42499632, end_position_on_the_genomic_accession=42581440, orientation='plus', exon_count=24, OMIM=608094, no_hdr0=''),
  54021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54021, CurrentID=0, Status='live', Symbol='SRSF9P1', Aliases='SFRS9P1', description='serine/arginine-rich splicing factor 9 pseudogene 1', other_designations='splicing factor pseudogene|splicing factor, arginine/serine-rich 9 pseudogene 1', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36294831, end_position_on_the_genomic_accession=36295792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54022, CurrentID=0, Status='live', Symbol='RPS5P3', Aliases='RPS5L, RPS5_3_1726', description='ribosomal protein S5 pseudogene 3', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33481539, end_position_on_the_genomic_accession=33482262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54024, CurrentID=0, Status='live', Symbol='RPS26P1', Aliases='RPS26P', description='ribosomal protein S26 pseudogene 1', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36303886, end_position_on_the_genomic_accession=36304003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54025, CurrentID=0, Status='live', Symbol='RPS20P1', Aliases='RPS20P, RPS20_16_1728', description='ribosomal protein S20 pseudogene 1', other_designations='ribosomal protein S20, pseudogene', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35724708, end_position_on_the_genomic_accession=35725166, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54026, CurrentID=0, Status='live', Symbol='RPL34P3', Aliases='', description='ribosomal protein L34 pseudogene 3', other_designations='pseudogene with similarity to ribosomal protein L34', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35472076, end_position_on_the_genomic_accession=35472455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54027, CurrentID=0, Status='live', Symbol='RPL31P1', Aliases='RPL31P, RPL31_35_1734', description='ribosomal protein L31 pseudogene 1', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43551197, end_position_on_the_genomic_accession=43551645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54028, CurrentID=0, Status='live', Symbol='RPL23P2', Aliases='RPL23_4_1724', description='ribosomal protein L23 pseudogene 2', other_designations='60S ribosomal protein L23 pseudogene 2', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28997556, end_position_on_the_genomic_accession=28998070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54029, CurrentID=0, Status='live', Symbol='RPL23AP4', Aliases='RPL23A_42_1735', description='ribosomal protein L23a pseudogene 4', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46690750, end_position_on_the_genomic_accession=46691257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54031, CurrentID=0, Status='live', Symbol='RIMKLBP1', Aliases='FAM80B2, FAM80C, RIMKLB2, RIMKLP', description='ribosomal modification protein rimK-like family member B pseudogene 1', other_designations='family with sequence similarity 80, member B2|ribosomal modification protein rimK-like family member B2|ribosomal modification protein rimK-like pseudogene', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36050160, end_position_on_the_genomic_accession=36052187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54032, CurrentID=0, Status='live', Symbol='RBPMSLP', Aliases='', description='RNA binding protein with multiple splicing-like pseudogene', other_designations='RNA-binding protein hermes pseudogene|RNA-binding protein with multiple splicing-like pseudogene', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15744004, end_position_on_the_genomic_accession=15744328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54033, CurrentID=0, Status='live', Symbol='RBM11', Aliases='', description='RNA binding motif protein 11', other_designations='RNA-binding motif protein 11|putative RNA-binding protein 11', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14216106, end_position_on_the_genomic_accession=14228372, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54034, CurrentID=0, Status='live', Symbol='RAD23BLP', Aliases='', description='RAD23 homolog B-like (S. cerevisiae) pseudogene', other_designations='RAD23 homolog B-like pseudogene|UV excision repair protein pseudogene', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15694300, end_position_on_the_genomic_accession=15696411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54035, CurrentID=0, Status='live', Symbol='PSMD4P1', Aliases='PSMD15, PSMD4P', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 4 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 15|proteasome (prosome, macropain) 26S subunit, non-ATPase, 4, pseudogene|proteasome 26S subunit pseudogene', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36485870, end_position_on_the_genomic_accession=36487432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54036, CurrentID=0, Status='live', Symbol='PPP1R2P2', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 2', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 2|protein phosphatase inhibitor 2 pseudogene', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35886892, end_position_on_the_genomic_accession=35888917, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54038, CurrentID=0, Status='live', Symbol='POLR2CP', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide C, pseudogene', other_designations='polymerase (RNA) II (DNA directed) polypeptide C, 33kDa pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14757558, end_position_on_the_genomic_accession=14759221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54039, CurrentID=0, Status='live', Symbol='PCBP3', Aliases='ALPHA-CP3', description='poly(rC) binding protein 3', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45849961, end_position_on_the_genomic_accession=45942454, orientation='plus', exon_count=14, OMIM=608502, no_hdr0=''),
  54041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54041, CurrentID=0, Status='live', Symbol='VN1R8P', Aliases='ORLP2', description='vomeronasal 1 receptor 8 pseudogene', other_designations='olfactory receptor-like pseudogene 2|pheromone receptor pseudogene|vomeronasal 1 receptor 2 pseudogene', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13476241, end_position_on_the_genomic_accession=13477455, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54043, CurrentID=0, Status='live', Symbol='MYL6P1', Aliases='MYL6P', description='myosin light chain 6 pseudogene 1', other_designations='myosin, light chain 6 pseudogene 1|myosin, light chain 6, alkali, smooth muscle and non-muscle pseudogene|myosin, light chain 6, pseudogene|myosin, light polypeptide 6, pseudogene|myosine alkali light chain 6 pseudogene', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43855723, end_position_on_the_genomic_accession=43856387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54044, CurrentID=0, Status='live', Symbol='KRT18P2', Aliases='', description='keratin 18 pseudogene 2', other_designations='cytokeratin 18 processed pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20424895, end_position_on_the_genomic_accession=20426169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54045, CurrentID=0, Status='live', Symbol='IMMTP1', Aliases='IMMTP', description='inner membrane protein, mitochondrial (mitofilin) pseudogene 1', other_designations='motorprotein pseudogene', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44675528, end_position_on_the_genomic_accession=44678165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54047, CurrentID=0, Status='live', Symbol='HSPD1P7', Aliases='HSPD1-22P, HSPDP7', description='heat shock protein family D (Hsp60) member 1 pseudogene 7', other_designations='heat shock 60kD protein 1 (chaperonin) pseudogene 7|heat shock 60kDa protein 1 (chaperonin) pseudogene 7|human chaperonin pseudogene', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28886986, end_position_on_the_genomic_accession=28889334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54048, CurrentID=0, Status='live', Symbol='HMGN1P2', Aliases='HMG14P, HMGN1L2', description='high mobility group nucleosome binding domain 1 pseudogene 2', other_designations='high-mobility group (nonhistone chromosomal) protein 14 pseudogene|high-mobility group nucleosome binding domain 1 pseudogene 2|high-mobility group nucleosome binding domain 1-like 2|nonhistone chromosomal protein HMG-14 pseudogene', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31706608, end_position_on_the_genomic_accession=31707183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54049, CurrentID=0, Status='live', Symbol='H2AFZP1', Aliases='GT334, H2AFZP', description='H2A histone family member Z pseudogene 1', other_designations='H2A histone family, member Z pseudogene 1|H2A histone family, member Z, pseudogene|histone H2AZ pseudogene', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44045967, end_position_on_the_genomic_accession=44047146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54050, CurrentID=0, Status='live', Symbol='GAPDHP14', Aliases='GAPDP14', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 14', other_designations='glycerinaldehyde-3-phosphate dehydrogenase pseudogene', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29222254, end_position_on_the_genomic_accession=29223466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54051, CurrentID=0, Status='live', Symbol='FDPSP6', Aliases='FDPSP', description='farnesyl diphosphate synthase pseudogene 6', other_designations='farnesyl pyrophosphate synthetase processed pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20388181, end_position_on_the_genomic_accession=20388941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54052, CurrentID=0, Status='live', Symbol='EIF4A1P1', Aliases='EIF4A1P', description='eukaryotic translation initiation factor 4A1 pseudogene 1', other_designations='eukaryotic initiation factor 4AI pseudogene|eukaryotic translation initiation factor 4A, isoform 1, pseudogene', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27367012, end_position_on_the_genomic_accession=27367979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54053, CurrentID=0, Status='live', Symbol='EIF3FP1', Aliases='EIF3FP, EIF3S5P', description='eukaryotic translation initiation factor 3 subunit F pseudogene 1', other_designations='eucaryotic initiation factor-3, subunit 5 pseudogene|eukaryotic translation initiation factor 3, subunit 5 (epsilon, 47kD) pseudogene|eukaryotic translation initiation factor 3, subunit 5 epsilon, 47kDa pseudogene|eukaryotic translation initiation factor 3, subunit F pseudogene 1', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10330640, end_position_on_the_genomic_accession=10331695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54054, CurrentID=0, Status='live', Symbol='EEF1A1P1', Aliases='EEF1A1P', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 1', other_designations='eukaryotic translation elongation factor 1 alpha 2 pseudogene|human elongation factor EF-1-alpha processed pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23390259, end_position_on_the_genomic_accession=23391293, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54055, CurrentID=0, Status='live', Symbol='CYP4F29P', Aliases='4F-se4[6:7:8], C21orf15, CYP4F-se4[6:7:8], CYP4F3LP', description='cytochrome P450 family 4 subfamily F member 29, pseudogene', other_designations='cytochrome P450 pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 29, pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 3-like pseudogene|cytochrome P450, subfamily IVF, polypeptide 3-like pseudogene', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13843133, end_position_on_the_genomic_accession=13848364, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54058, CurrentID=0, Status='live', Symbol='C21orf58', Aliases='', description='chromosome 21 open reading frame 58', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46300430, end_position_on_the_genomic_accession=46323899, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54059, CurrentID=0, Status='live', Symbol='YBEY', Aliases='C21orf57', description='ybeY metallopeptidase (putative)', other_designations='putative metalloprotease C21orf57|rRNA maturation factor homolog', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46286310, end_position_on_the_genomic_accession=46313225, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54064, CurrentID=0, Status='live', Symbol='LINC00160', Aliases='C21orf52, NCRNA00160', description='long intergenic non-protein coding RNA 160', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34723807, end_position_on_the_genomic_accession=34737182, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54065, CurrentID=0, Status='live', Symbol='SMIM11A', Aliases='C21orf51, FAM165B, SMIM11', description='small integral membrane protein 11A', other_designations='family with sequence similarity 165, member B|protein FAM165B', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34375450, end_position_on_the_genomic_accession=34389154, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54067, CurrentID=0, Status='live', Symbol='C21orf62-AS1', Aliases='C21orf49', description='C21orf62 antisense RNA 1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32772100, end_position_on_the_genomic_accession=32797707, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54069, CurrentID=0, Status='live', Symbol='MIS18A', Aliases='B28, C21orf45, C21orf46, FASP1, MIS18alpha, hMis18alpha', description='MIS18 kinetochore protein A', other_designations='FAPP1-associated protein 1|MIS18 kinetochore protein homolog A', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32268219, end_position_on_the_genomic_accession=32279065, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54072, CurrentID=0, Status='live', Symbol='LINC00158', Aliases='C21orf42, NCRNA00158', description='long intergenic non-protein coding RNA 158', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25385820, end_position_on_the_genomic_accession=25431701, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54074, CurrentID=0, Status='live', Symbol='C21orf40', Aliases='', description='uncharacterized LOC54074', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54075, CurrentID=0, Status='live', Symbol='CHODL-AS1', Aliases='C21orf39, NCRNA00157', description='CHODL antisense RNA 1', other_designations='CHODL antisense RNA 1 (non-protein coding)', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17835672, end_position_on_the_genomic_accession=17885608, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  402229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402229, CurrentID=0, Status='live', Symbol='LOC402229', Aliases='', description='SURP and G-patch domain containing 1 pseudogene', other_designations='SURP and G patch domain containing 1 pseudogene|splicing factor 4 pseudogene', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130994243, end_position_on_the_genomic_accession=130996296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54082, CurrentID=0, Status='live', Symbol='TSPEAR-AS1', Aliases='C21orf31', description='TSPEAR antisense RNA 1', other_designations='TSPEAR antisense RNA 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44506044, end_position_on_the_genomic_accession=44516575, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54083, CurrentID=0, Status='live', Symbol='LRRC3DN', Aliases='C21orf30', description='LRRC3 downstream neighbor (non-protein coding)', other_designations='intronless long ORF, AL117578', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44458944, end_position_on_the_genomic_accession=44462196, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54084, CurrentID=0, Status='live', Symbol='TSPEAR', Aliases='C21orf29, DFNB98, TSP-EAR', description='thrombospondin-type laminin G domain and EAR repeats', other_designations='thrombospondin-type laminin G domain and EAR repeats-containing protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44497892, end_position_on_the_genomic_accession=44711580, orientation='minus', exon_count=13, OMIM=612920, no_hdr0=''),
  402230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402230, CurrentID=0, Status='live', Symbol='LOC402230', Aliases='', description='small nuclear ribonucleoprotein, U4/U6.U5 27kDa subunit pseudogene', other_designations='small nuclear ribonucleoprotein 27kDa (U4/U6.U5) pseudogene', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132986421, end_position_on_the_genomic_accession=132988138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100847073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100847073, CurrentID=0, Status='live', Symbol='MIR5090', Aliases='mir-5090', description='microRNA 5090', other_designations='hsa-mir-5090', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102465742, end_position_on_the_genomic_accession=102465826, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54088, CurrentID=0, Status='live', Symbol='LINC00113', Aliases='C21orf23, NCRNA00113', description='long intergenic non-protein coding RNA 113', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27722379, end_position_on_the_genomic_accession=27751233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54089, CurrentID=0, Status='live', Symbol='LINC00112', Aliases='C21orf22, NCRNA00112', description='long intergenic non-protein coding RNA 112', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41716436, end_position_on_the_genomic_accession=41717582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54090, CurrentID=0, Status='live', Symbol='LINC00111', Aliases='C21orf21, NCRNA00111', description='long intergenic non-protein coding RNA 111', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41679302, end_position_on_the_genomic_accession=41697336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54093, CurrentID=0, Status='live', Symbol='SETD4', Aliases='C21orf18, C21orf27', description='SET domain containing 4', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36034541, end_position_on_the_genomic_accession=36060818, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54097, CurrentID=0, Status='live', Symbol='FAM3B', Aliases='2-21, C21orf11, C21orf76, ORF9, PANDER, PRED44', description='family with sequence similarity 3 member B', other_designations='D21M16SJHU19e|cytokine-like protein 2-21|family with sequence similarity 3, member B|pancreatic derived factor|pancreatic-derived factor', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41304223, end_position_on_the_genomic_accession=41357727, orientation='plus', exon_count=11, OMIM=608617, no_hdr0=''),
  54098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54098, CurrentID=0, Status='live', Symbol='C1QBPP1', Aliases='C1QBPP', description='complement component 1, q subcomponent binding protein, pseudogene 1', other_designations='human splicing factor 2 hyaluronic acid-binding protein (SF2p32) pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19759120, end_position_on_the_genomic_accession=19760127, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54099, CurrentID=0, Status='live', Symbol='FBXW11P1', Aliases='BTRC2P, FBXW1BP, FBXW1BP1', description='F-box and WD repeat domain containing 11 pseudogene 1', other_designations='F-box and WD-40 domain protein 11 pseudogene 1|F-box and WD-40 domain protein 1B pseudogene 1|beta-transducin repeat containing 2 pseudogene', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31627057, end_position_on_the_genomic_accession=31628660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54101, CurrentID=0, Status='live', Symbol='RIPK4', Aliases='ANKK2, ANKRD3, DIK, NKRD3, PKK, PPS2, RIP4', description='receptor interacting serine/threonine kinase 4', other_designations='PKC-delta-interacting protein kinase|ankyrin repeat domain-containing protein 3|protein kinase C-associated kinase|receptor-interacting serine-threonine kinase 4|serine/threonine-protein kinase ANKRD3', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41739369, end_position_on_the_genomic_accession=41767089, orientation='minus', exon_count=8, OMIM=605706, no_hdr0=''),
  54102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54102, CurrentID=0, Status='live', Symbol='CLIC6', Aliases='CLIC1L', description='chloride intracellular channel 6', other_designations='chloride channel form A|parchorin', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34669292, end_position_on_the_genomic_accession=34718227, orientation='plus', exon_count=7, OMIM=615321, no_hdr0=''),
  54103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54103, CurrentID=0, Status='live', Symbol='GSAP', Aliases='PION', description='gamma-secretase activating protein', other_designations='pigeon homolog|protein pigeon homolog', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77310751, end_position_on_the_genomic_accession=77416408, orientation='minus', exon_count=34, OMIM=613552, no_hdr0=''),
  54106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54106, CurrentID=0, Status='live', Symbol='TLR9', Aliases='CD289', description='toll-like receptor 9', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52221080, end_position_on_the_genomic_accession=52226163, orientation='minus', exon_count=2, OMIM=605474, no_hdr0=''),
  54107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54107, CurrentID=0, Status='live', Symbol='POLE3', Aliases='CHARAC17, CHRAC17, YBL1, p17', description='polymerase (DNA directed), epsilon 3, accessory subunit', other_designations='CHRAC-17|DNA polymerase II subunit 3|DNA polymerase epsilon p17 subunit|DNA polymerase epsilon subunit p17|arsenic transactivated protein|arsenic-transactivated protein|asTP|chromatin accessibility complex 17 kDa protein|histone fold protein CHRAC17|huCHRAC17|polymerase (DNA directed), epsilon 3 (p17 subunit)', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113407235, end_position_on_the_genomic_accession=113410749, orientation='minus', exon_count=5, OMIM=607267, no_hdr0=''),
  54108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54108, CurrentID=0, Status='live', Symbol='CHRAC1', Aliases='CHARC1, CHARC15, CHRAC-1, CHRAC-155, YCL1, CHRAC1', description='chromatin accessibility complex 1', other_designations='DNA polymerase epsilon subunit p15|chromatin accessibility complex 15 kDa protein|histone-fold protein CHRAC15', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140511298, end_position_on_the_genomic_accession=140517153, orientation='plus', exon_count=4, OMIM=607268, no_hdr0=''),
  54109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54109, CurrentID=0, Status='live', Symbol='CORD8', Aliases='', description='cone rod dystrophy 8', other_designations='', map_location='1q12-q24', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605549, no_hdr0=''),
  54110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54110, CurrentID=0, Status='live', Symbol='RP29', Aliases='', description='retinitis pigmentosa 29 (autosomal recessive)', other_designations='', map_location='4q32-q34', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612165, no_hdr0=''),
  54111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54111, CurrentID=0, Status='live', Symbol='HNB1', Aliases='', description='hereditary neuroblastoma 1', other_designations='', map_location='16p13-p12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54112, CurrentID=0, Status='live', Symbol='GPR88', Aliases='STRG', description='G protein-coupled receptor 88', other_designations='G protein coupled receptor 88|G-protein coupled receptor 88|striatum-specific G-protein coupled receptor', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100538172, end_position_on_the_genomic_accession=100542027, orientation='plus', exon_count=2, OMIM=607468, no_hdr0=''),
  54113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54113, CurrentID=0, Status='live', Symbol='VUR', Aliases='', description='Vesicoureteral reflux', other_designations='', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=193000, no_hdr0=''),
  54114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54114, CurrentID=0, Status='live', Symbol='ODPF', Aliases='', description='Osseous dysplasia, digital, with facial pigmentary defects and multiple frenula', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300244, no_hdr0=''),
  54115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54115, CurrentID=0, Status='live', Symbol='PTOS2', Aliases='PTOSX', description='Ptosis, hereditary congenital 2', other_designations='', map_location='Xq24-q27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300245, no_hdr0=''),
  54119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54119, CurrentID=0, Status='live', Symbol='EIM', Aliases='', description='Infantile myoclonic epilepsy', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605021, no_hdr0=''),
  402236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402236, CurrentID=0, Status='live', Symbol='OR1X1P', Aliases='', description='olfactory receptor family 1 subfamily X member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily X, member 1 pseudogene', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176006145, end_position_on_the_genomic_accession=176007389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54143, CurrentID=0, Status='live', Symbol='LINC00308', Aliases='C21orf74, NCRNA00308, PRED16', description='long intergenic non-protein coding RNA 308', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=22098617, end_position_on_the_genomic_accession=22116528, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54145, CurrentID=0, Status='live', Symbol='H2BFS', Aliases='', description='H2B histone family member S (pseudogene)', other_designations='H2B histone family, member S (pseudogene)|histone protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43565239, end_position_on_the_genomic_accession=43565617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54147, CurrentID=0, Status='live', Symbol='BACH1-IT3', Aliases='C21orf8, NCRNA00097', description='BACH1 intronic transcript 3', other_designations='BACH1 intronic transcript 3 (non-protein coding)', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54148, CurrentID=0, Status='live', Symbol='MRPL39', Aliases='C21orf92, L39mt, MRP-L5, MRPL5, MSTP003, PRED22, PRED66, RPML5', description='mitochondrial ribosomal protein L39', other_designations='39S ribosomal protein L5, mitochondrial|L5mt|MRP-L39', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25585656, end_position_on_the_genomic_accession=25607846, orientation='minus', exon_count=12, OMIM=611845, no_hdr0=''),
  54149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54149, CurrentID=0, Status='live', Symbol='C21orf91', Aliases='C21orf14, C21orf38, CSSG1, EURL, YG81', description='chromosome 21 open reading frame 91', other_designations='cold sore susceptibility gene 1|early undifferentiated retina and lens', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17788967, end_position_on_the_genomic_accession=17819386, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373567, CurrentID=0, Status='live', Symbol='LOC105373567', Aliases='', description='uncharacterized LOC105373567', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113314501, end_position_on_the_genomic_accession=113325321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54165, CurrentID=0, Status='live', Symbol='DCUN1D1', Aliases='DCNL1, DCUN1L1, RP42, SCCRO, SCRO, Tes3', description='DCN1, defective in cullin neddylation 1, domain containing 1', other_designations='DCUN1 domain-containing protein 1|RP42 homolog|defective in cullin neddylation protein 1-like protein 1|squamous cell carcinoma related oncogene|squamous cell carcinoma-related oncogene', map_location='3q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182938158, end_position_on_the_genomic_accession=182985947, orientation='minus', exon_count=10, OMIM=605905, no_hdr0=''),
  140100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140100, CurrentID=0, Status='live', Symbol='RBMY2KP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member K pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8264176, end_position_on_the_genomic_accession=8279146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401021, CurrentID=0, Status='live', Symbol='LOC401021', Aliases='', description='uncharacterized LOC401021', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402246, CurrentID=0, Status='live', Symbol='OR10AH1P', Aliases='', description='olfactory receptor family 10 subfamily AH member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily AH, member 1 pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5117090, end_position_on_the_genomic_accession=5118088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378782, CurrentID=0, Status='live', Symbol='LOC105378782', Aliases='', description='uncharacterized LOC105378782', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68293382, end_position_on_the_genomic_accession=68303696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54187, CurrentID=0, Status='live', Symbol='NANS', Aliases='HEL-S-100, SAS', description='N-acetylneuraminate synthase', other_designations='N-acetylneuraminate-9-phosphate synthase|N-acetylneuraminic acid phosphate synthase|N-acetylneuraminic acid synthase|epididymis secretory protein Li 100|sialic acid phosphate synthase', map_location='9p24.1-p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98056677, end_position_on_the_genomic_accession=98083083, orientation='plus', exon_count=7, OMIM=605202, no_hdr0=''),
  402249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402249, CurrentID=0, Status='live', Symbol='OR7E59P', Aliases='OR7E137P, OR7E59, OST119', description='olfactory receptor family 7 subfamily E member 59 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 137 pseudogene|olfactory receptor, family 7, subfamily E, member 59 pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6879622, end_position_on_the_genomic_accession=6880580, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371416, CurrentID=0, Status='live', Symbol='LOC105371416', Aliases='', description='uncharacterized LOC105371416', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89682507, end_position_on_the_genomic_accession=89685298, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54205, CurrentID=0, Status='live', Symbol='CYCS', Aliases='CYC, HCS, THC4', description='cytochrome c, somatic', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25118651, end_position_on_the_genomic_accession=25125361, orientation='minus', exon_count=3, OMIM=123970, no_hdr0=''),
  54206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54206, CurrentID=0, Status='live', Symbol='ERRFI1', Aliases='GENE-33, MIG-6, MIG6, RALT', description='ERBB receptor feedback inhibitor 1', other_designations='mitogen-inducible gene 6 protein|receptor-associated late transducer', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8011719, end_position_on_the_genomic_accession=8026333, orientation='minus', exon_count=5, OMIM=608069, no_hdr0=''),
  54207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54207, CurrentID=0, Status='live', Symbol='KCNK10', Aliases='K2p10.1, PPP1R97, TREK-2, TREK2', description='potassium channel, two pore domain subfamily K, member 10', other_designations='2P domain potassium channel TREK2|TREK-2 K(+) channel subunit|TWIK-related K+ channel 2|outward rectifying potassium channel protein TREK-2|potassium channel TREK-2|potassium channel, subfamily K, member 10|protein phosphatase 1, regulatory subunit 97', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88180108, end_position_on_the_genomic_accession=88326912, orientation='minus', exon_count=10, OMIM=605873, no_hdr0=''),
  54209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54209, CurrentID=0, Status='live', Symbol='TREM2', Aliases='TREM-2, Trem2a, Trem2b, Trem2c', description='triggering receptor expressed on myeloid cells 2', other_designations='triggering receptor expressed on monocytes 2|triggering receptor expressed on myeloid cells 2a', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41158506, end_position_on_the_genomic_accession=41163198, orientation='minus', exon_count=5, OMIM=605086, no_hdr0=''),
  54210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54210, CurrentID=0, Status='live', Symbol='TREM1', Aliases='CD354, TREM-1', description='triggering receptor expressed on myeloid cells 1', other_designations='triggering receptor expressed on monocytes 1|triggering-receptor TREM1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41267385, end_position_on_the_genomic_accession=41286719, orientation='minus', exon_count=6, OMIM=605085, no_hdr0=''),
  54212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54212, CurrentID=0, Status='live', Symbol='SNTG1', Aliases='G1SYN, SYN4', description='syntrophin, gamma 1', other_designations='gamma1-syntrophin|syntrophin 4|syntrophin-4', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49909789, end_position_on_the_genomic_accession=50796442, orientation='plus', exon_count=24, OMIM=608714, no_hdr0=''),
  54221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54221, CurrentID=0, Status='live', Symbol='SNTG2', Aliases='G2SYN, SYN5', description='syntrophin, gamma 2', other_designations='syntrophin-5', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=950868, end_position_on_the_genomic_accession=1367645, orientation='plus', exon_count=21, OMIM=608715, no_hdr0=''),
  220032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220032, CurrentID=0, Status='live', Symbol='GDPD4', Aliases='GDE6', description='glycerophosphodiester phosphodiesterase domain containing 4', other_designations='GDPD domain containing protein|glycerophosphodiester phosphodiesterase 6|ugpQ', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77216558, end_position_on_the_genomic_accession=77301699, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  645433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645433, CurrentID=0, Status='live', Symbol='LOC645433', Aliases='', description='inositol-3-phosphate synthase 1 pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25504948, end_position_on_the_genomic_accession=25506769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285501, CurrentID=0, Status='live', Symbol='LINC01098', Aliases='', description='long intergenic non-protein coding RNA 1098', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177728757, end_position_on_the_genomic_accession=177990750, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130502, CurrentID=0, Status='live', Symbol='LOC100130502', Aliases='', description='uncharacterized LOC100130502', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44921080, end_position_on_the_genomic_accession=44923566, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374421, CurrentID=0, Status='live', Symbol='LOC105374421', Aliases='', description='uncharacterized LOC105374421', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41683005, end_position_on_the_genomic_accession=41697151, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101241875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241875, CurrentID=0, Status='live', Symbol='RPS3AP1', Aliases='RPS3AP', description='ribosomal protein S3A pseudogene 1', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20430400, end_position_on_the_genomic_accession=20430762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533695, CurrentID=0, Status='live', Symbol='RLIMP1', Aliases='', description='ring finger protein, LIM domain interacting pseudogene 1', other_designations='ring finger protein (C3H2C3 type) 6 pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40618626, end_position_on_the_genomic_accession=40621547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101241877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241877, CurrentID=0, Status='live', Symbol='ATP5J2LP', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2-like pseudogene', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2-like pseudogene|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit f, isoform 2-like pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2-like pseudogene|F1Fo-ATPase synthase f subunit pseudogene', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36388878, end_position_on_the_genomic_accession=36389112, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101241878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241878, CurrentID=0, Status='live', Symbol='USP17L23', Aliases='', description='ubiquitin specific peptidase 17-like family member 23', other_designations='Putative ubiquitin carboxyl-terminal hydrolase 17-like protein 23', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101241891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241891, CurrentID=0, Status='live', Symbol='LINC00850', Aliases='KUCG1', description='long intergenic non-protein coding RNA 850', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149825708, end_position_on_the_genomic_accession=149879799, orientation='plus', exon_count=3, OMIM=300892, no_hdr0=''),
  101241892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241892, CurrentID=0, Status='live', Symbol='NPTN-IT1', Aliases='lncRNA-LET', description='NPTN intronic transcript 1', other_designations='NPTN intronic transcript 1 (non-protein coding)|long non-coding RNA, low expression in tumor', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73566938, end_position_on_the_genomic_accession=73569543, orientation='minus', exon_count=1, OMIM=615176, no_hdr0=''),
  101241897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241897, CurrentID=0, Status='live', Symbol='PAPA5', Aliases='', description='Polydactyly, postaxial, type A5', other_designations='', map_location='13q13.3-q21', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=263450, no_hdr0=''),
  101241898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241898, CurrentID=0, Status='live', Symbol='KAZA2', Aliases='', description='Kala-azar, susceptibility to, 2', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611381, no_hdr0=''),
  101241899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241899, CurrentID=0, Status='live', Symbol='KAZA3', Aliases='', description='Kala-azar, susceptibility to, 3', other_designations='', map_location='6p27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611382, no_hdr0=''),
  101241900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241900, CurrentID=0, Status='live', Symbol='SMAJ', Aliases='', description='Spinal muscular atrophy, Jokela type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615048, no_hdr0=''),
  101241901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241901, CurrentID=0, Status='live', Symbol='LOC101241901', Aliases='', description='chromosome 4 open reading frame 46 pseudogene', other_designations='', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77962541, end_position_on_the_genomic_accession=77965212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101241902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101241902, CurrentID=0, Status='live', Symbol='LOC101241902', Aliases='', description='chromosome 4 open reading frame 46 pseudogene', other_designations='', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78722847, end_position_on_the_genomic_accession=78724374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  402269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402269, CurrentID=0, Status='live', Symbol='LOC402269', Aliases='', description='equilibrative nucleoside transporter 4-like', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63632667, end_position_on_the_genomic_accession=63640929, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374422, CurrentID=0, Status='live', Symbol='LOC105374422', Aliases='', description='uncharacterized LOC105374422', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41664007, end_position_on_the_genomic_accession=41665624, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54328, CurrentID=0, Status='live', Symbol='GPR173', Aliases='SREB3', description='G protein-coupled receptor 173', other_designations='super conserved receptor expressed in brain 3', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53049324, end_position_on_the_genomic_accession=53080615, orientation='plus', exon_count=4, OMIM=300253, no_hdr0=''),
  54329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54329, CurrentID=0, Status='live', Symbol='GPR85', Aliases='SREB, SREB2', description='G protein-coupled receptor 85', other_designations='seven transmembrane helix receptor|super conserved receptor expressed in brain 2', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=113080413, end_position_on_the_genomic_accession=113087778, orientation='minus', exon_count=4, OMIM=605188, no_hdr0=''),
  54331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54331, CurrentID=0, Status='live', Symbol='GNG2', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 2', other_designations='g gamma-I|guanine nucleotide binding protein gamma 2|guanine nucleotide-binding protein G(I)/G(O) gamma-2 subunit', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51847131, end_position_on_the_genomic_accession=51969800, orientation='plus', exon_count=7, OMIM=606981, no_hdr0=''),
  54332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54332, CurrentID=0, Status='live', Symbol='GDAP1', Aliases='CMT4, CMT4A, CMTRIA', description='ganglioside induced differentiation associated protein 1', other_designations='Charcot-Marie-Tooth neuropathy 4A|ganglioside differentiation associated protein 1', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74349251, end_position_on_the_genomic_accession=74367110, orientation='plus', exon_count=6, OMIM=606598, no_hdr0=''),
  402273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402273, CurrentID=0, Status='live', Symbol='VN1R37P', Aliases='', description='vomeronasal 1 receptor 37 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64149619, end_position_on_the_genomic_accession=64150529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54344, CurrentID=0, Status='live', Symbol='DPM3', Aliases='CDG1O', description='dolichyl-phosphate mannosyltransferase polypeptide 3', other_designations='DPM synthase complex subunit 3|DPM synthase subunit 3|MPD synthase subunit 3|dolichol-phosphate mannose synthase subunit 3|dolichyl-phosphate beta-D-mannosyltransferase subunit 3|mannose-P-dolichol synthase subunit 3|prostin 1|prostin-1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155139891, end_position_on_the_genomic_accession=155140520, orientation='minus', exon_count=2, OMIM=605951, no_hdr0=''),
  54345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54345, CurrentID=0, Status='live', Symbol='SOX18', Aliases='HLTRS, HLTS', description='SRY-box 18', other_designations='SRY (sex determining region Y)-box 18|SRY box 18', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64047726, end_position_on_the_genomic_accession=64049626, orientation='minus', exon_count=2, OMIM=601618, no_hdr0=''),
  54346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54346, CurrentID=0, Status='live', Symbol='UNC93A', Aliases='Unc-93A, dJ366N23.1, dJ366N23.2', description='unc-93 homolog A (C. elegans)', other_designations='dJ366N23.1 (putative C. elegans UNC-93 (protein 1, C46F11.1) LIKE)|dJ366N23.2 (putative C. elegans UNC-93 (protein 1, C46F11.1) C-terminal LIKE)|unc93 homolog A', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167271174, end_position_on_the_genomic_accession=167316029, orientation='plus', exon_count=12, OMIM=607995, no_hdr0=''),
  54360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54360, CurrentID=0, Status='live', Symbol='CYTL1', Aliases='C17, C4orf4', description='cytokine like 1', other_designations='cytokine-like 1|cytokine-like protein C17', map_location='4p16-p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5014586, end_position_on_the_genomic_accession=5019470, orientation='minus', exon_count=4, OMIM=607930, no_hdr0=''),
  54361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54361, CurrentID=0, Status='live', Symbol='WNT4', Aliases='SERKAL, WNT-4', description='wingless-type MMTV integration site family member 4', other_designations='wingless-type MMTV integration site family, member 4', map_location='1p36.23-p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22117305, end_position_on_the_genomic_accession=22143981, orientation='minus', exon_count=8, OMIM=603490, no_hdr0=''),
  54362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54362, CurrentID=0, Status='live', Symbol='DFNM1', Aliases='', description='deafness (recessive, non-syndromic) modifier 1', other_designations='deafness (autosomal recessive, nonsyndromic) modifier 1|deafness (recessive, nonsyndromic) modifier 1', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605429, no_hdr0=''),
  54363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54363, CurrentID=0, Status='live', Symbol='HAO1', Aliases='GOX, GOX1, HAOX1', description='hydroxyacid oxidase (glycolate oxidase) 1', other_designations='(S)-2-hydroxy-acid oxidase|glycolate oxidase', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7882984, end_position_on_the_genomic_accession=7940446, orientation='minus', exon_count=8, OMIM=605023, no_hdr0=''),
  101929227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929227, CurrentID=0, Status='live', Symbol='LOC101929227', Aliases='', description='uncharacterized LOC101929227', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120867933, end_position_on_the_genomic_accession=120894797, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  402279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402279, CurrentID=0, Status='live', Symbol='LOC402279', Aliases='', description='glutamate receptor, metabotropic 8 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65617082, end_position_on_the_genomic_accession=65619154, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100189424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189424, CurrentID=0, Status='live', Symbol='TRY-GTA5-1', Aliases='TRNAY14', description='transfer RNA-Tyr (GTA) 5-1', other_designations='transfer RNA tyrosine 14 (anticodon GUA)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20653099, end_position_on_the_genomic_accession=20653192, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54386, CurrentID=0, Status='live', Symbol='TERF2IP', Aliases='DRIP5, RAP1', description='telomeric repeat binding factor 2, interacting protein', other_designations='RAP1 homolog|TERF2-interacting telomeric protein 1|TRF2-interacting telomeric RAP1 protein|TRF2-interacting telomeric protein 1|dopamine receptor interacting protein 5|dopamine receptor-interacting protein 5|hRap1|repressor/activator protein 1 homolog', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75647704, end_position_on_the_genomic_accession=75657442, orientation='plus', exon_count=3, OMIM=605061, no_hdr0=''),
  105378789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378789, CurrentID=0, Status='live', Symbol='LOC105378789', Aliases='', description='uncharacterized LOC105378789', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69453676, end_position_on_the_genomic_accession=69557821, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  729962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729962, CurrentID=0, Status='live', Symbol='RPL21P71', Aliases='RPL21_12_416', description='ribosomal protein L21 pseudogene 71', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147276727, end_position_on_the_genomic_accession=147277287, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285505, CurrentID=0, Status='live', Symbol='LOC285505', Aliases='', description='uncharacterized LOC285505', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54407, CurrentID=0, Status='live', Symbol='SLC38A2', Aliases='ATA2, PRO1068, SAT2, SNAT2', description='solute carrier family 38 member 2', other_designations='amino acid transporter 2|amino acid transporter A2|protein 40-9-1|solute carrier family 38, member 2|system A amino acid transporter 2|system A transporter 1|system N amino acid transporter 2', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46358188, end_position_on_the_genomic_accession=46372862, orientation='minus', exon_count=16, OMIM=605180, no_hdr0=''),
  54413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54413, CurrentID=0, Status='live', Symbol='NLGN3', Aliases='HNL3', description='neuroligin 3', other_designations='gliotactin homolog', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71144823, end_position_on_the_genomic_accession=71175307, orientation='plus', exon_count=11, OMIM=300336, no_hdr0=''),
  54414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54414, CurrentID=0, Status='live', Symbol='SIAE', Aliases='AIS6, CSE-C, CSEC, LSE, YSG2', description='sialic acid acetylesterase', other_designations='H-Lse|cytosolic sialic acid 9-O-acetylesterase homolog|sialic acid-specific acetylesterase II', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124635789, end_position_on_the_genomic_accession=124676303, orientation='minus', exon_count=15, OMIM=610079, no_hdr0=''),
  100418592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418592, CurrentID=0, Status='live', Symbol='LOC100418592', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  402287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402287, CurrentID=0, Status='live', Symbol='RPS3AP21', Aliases='RPS3A_7_600', description='ribosomal protein S3a pseudogene 21', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111192185, end_position_on_the_genomic_accession=111193067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54429, CurrentID=0, Status='live', Symbol='TAS2R5', Aliases='T2R5', description='taste 2 receptor member 5', other_designations='taste receptor, type 2, member 5', map_location='7q31.3-q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141790217, end_position_on_the_genomic_accession=141791366, orientation='plus', exon_count=1, OMIM=605062, no_hdr0=''),
  54431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54431, CurrentID=0, Status='live', Symbol='DNAJC10', Aliases='ERdj5, JPDI, MTHr, PDIA19', description='DnaJ heat shock protein family (Hsp40) member C10', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 10|ER-resident protein ERdj5|J-domain-containing protein disulfide isomerase-like protein|endoplasmic reticulum DNA J domain-containing protein 5|macrothioredoxin|protein disulfide isomerase family A, member 19', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182716041, end_position_on_the_genomic_accession=182780025, orientation='plus', exon_count=26, OMIM=607987, no_hdr0=''),
  54432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54432, CurrentID=0, Status='live', Symbol='YIPF1', Aliases='DJ167A19.1, FinGER1', description='Yip1 domain family member 1', other_designations='YIP1 family member 1|Yip1 domain family, member 1', map_location='1p33-p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53849790, end_position_on_the_genomic_accession=53889820, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54433, CurrentID=0, Status='live', Symbol='GAR1', Aliases='NOLA1', description='GAR1 homolog, ribonucleoprotein', other_designations='GAR1 ribonucleoprotein homolog|nucleolar protein family A member 1|nucleolar protein family A, member 1 (H/ACA small nucleolar RNPs)|snoRNP protein GAR1', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109815510, end_position_on_the_genomic_accession=109824737, orientation='plus', exon_count=7, OMIM=606468, no_hdr0=''),
  54434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54434, CurrentID=0, Status='live', Symbol='SSH1', Aliases='SSH1L', description='slingshot protein phosphatase 1', other_designations='SSH-like protein 1|hSSH-1L|slingshot homolog 1', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108782690, end_position_on_the_genomic_accession=108857583, orientation='minus', exon_count=19, OMIM=606778, no_hdr0=''),
  54435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54435, CurrentID=0, Status='live', Symbol='HCG4', Aliases='HCG4P10, HCGIV-10, HCGIV.9', description='HLA complex group 4 (non-protein coding)', other_designations='HCGIV-10 pseudogene|HLA complex group 4 pseudogene 10', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29791031, end_position_on_the_genomic_accession=29793073, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54436, CurrentID=0, Status='live', Symbol='SH3TC1', Aliases='', description='SH3 domain and tetratricopeptide repeats 1', other_designations='SH3 domain and tetratricopeptide repeats-containing protein 1', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8182072, end_position_on_the_genomic_accession=8241103, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  54437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54437, CurrentID=0, Status='live', Symbol='SEMA5B', Aliases='SEMAG, SemG', description='sema domain, seven thrombospondin repeats (type 1 and type 1-like), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 5B', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122909193, end_position_on_the_genomic_accession=123028605, orientation='minus', exon_count=26, OMIM=609298, no_hdr0=''),
  54438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54438, CurrentID=0, Status='live', Symbol='GFOD1', Aliases='ADG-90, C6orf114', description='glucose-fructose oxidoreductase domain containing 1', other_designations='', map_location='6pter-p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13357830, end_position_on_the_genomic_accession=13487637, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54439, CurrentID=0, Status='live', Symbol='RBM27', Aliases='ARRS1, Psc1, ZC3H18, ZC3H20', description='RNA binding motif protein 27', other_designations='RNA-binding motif protein 27|acidic rich RS domain containing 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146203550, end_position_on_the_genomic_accession=146289223, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  54440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54440, CurrentID=0, Status='live', Symbol='SASH3', Aliases='753P9, CXorf9, HACS2, SH3D6C, SLY', description='SAM and SH3 domain containing 3', other_designations='SH3 protein expressed in lymphocytes homolog', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129779916, end_position_on_the_genomic_accession=129795201, orientation='plus', exon_count=8, OMIM=300441, no_hdr0=''),
  54441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54441, CurrentID=0, Status='live', Symbol='STAG3L1', Aliases='STAG3L1P, STAG3L2, STAG3L3', description='stromal antigen 3-like 1 (pseudogene)', other_designations='STAG3-like protein 1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75359200, end_position_on_the_genomic_accession=75367809, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54442, CurrentID=0, Status='live', Symbol='KCTD5', Aliases='', description='potassium channel tetramerization domain containing 5', other_designations='potassium channel tetramerisation domain containing 5', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2682494, end_position_on_the_genomic_accession=2709030, orientation='plus', exon_count=6, OMIM=611285, no_hdr0=''),
  54443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54443, CurrentID=0, Status='live', Symbol='ANLN', Aliases='FSGS8, Scraps, scra', description='anillin actin binding protein', other_designations='', map_location='7p15-p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36389803, end_position_on_the_genomic_accession=36453791, orientation='plus', exon_count=25, OMIM=616027, no_hdr0=''),
  100420338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420338, CurrentID=0, Status='live', Symbol='LOC100420338', Aliases='', description='charged multivesicular body protein 5 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203805621, end_position_on_the_genomic_accession=203806254, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54453, CurrentID=0, Status='live', Symbol='RIN2', Aliases='MACS, RASSF4', description='Ras and Rab interactor 2', other_designations='RAB5 interacting protein 2|RAS association (RalGDS/AF-6) domain containing protein JC265|RAS association domain family 4|RAS inhibitor JC265|RAS interaction/interference protein 2', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19757610, end_position_on_the_genomic_accession=20002459, orientation='plus', exon_count=22, OMIM=610222, no_hdr0=''),
  54454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54454, CurrentID=0, Status='live', Symbol='ATAD2B', Aliases='', description='ATPase family, AAA domain containing 2B', other_designations='', map_location='2p24.1-p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23685229, end_position_on_the_genomic_accession=23927114, orientation='minus', exon_count=41, OMIM=615347, no_hdr0=''),
  54455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54455, CurrentID=0, Status='live', Symbol='FBXO42', Aliases='Fbx42, JFK', description='F-box protein 42', other_designations='just one F-box and Kelch domain-containing protein', map_location='1p36.23-p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16246840, end_position_on_the_genomic_accession=16352491, orientation='minus', exon_count=11, OMIM=609109, no_hdr0=''),
  54456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54456, CurrentID=0, Status='live', Symbol='MOV10L1', Aliases='CHAMP, DJ402G11.8', description='Mov10 RISC complex RNA helicase like 1', other_designations='MOV10-like protein 1|Mov10-like 1|Mov10l1, Moloney leukemia virus 10-like 1, homolog|cardiac helicase activated by MEF2C protein|moloney leukemia virus 10-like protein 1', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50090006, end_position_on_the_genomic_accession=50161705, orientation='plus', exon_count=31, OMIM=605794, no_hdr0=''),
  54457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54457, CurrentID=0, Status='live', Symbol='TAF7L', Aliases='CT40, TAF2Q', description='TATA-box binding protein associated factor 7 like', other_designations='RNA polymerase II TBP-associated factor subunit Q|TAF7-like RNA polymerase II, TATA box binding protein (TBP)-associated factor, 50kDa|TATA box binding protein associated factor 7 like|TATA box-binding protein-associated factor 50 kDa|TBP-associated factor, RNA polymerase II, Q|cancer/testis antigen 40|transcription initiation factor TFIID 50 kDa subunit', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101268253, end_position_on_the_genomic_accession=101293071, orientation='minus', exon_count=14, OMIM=300314, no_hdr0=''),
  54458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54458, CurrentID=0, Status='live', Symbol='PRR13', Aliases='TXR1', description='proline rich 13', other_designations='taxane-resistance protein', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53441649, end_position_on_the_genomic_accession=53446643, orientation='plus', exon_count=4, OMIM=610459, no_hdr0=''),
  54460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54460, CurrentID=0, Status='live', Symbol='MRPS21', Aliases='MDS016, MRP-S21, RPMS21', description='mitochondrial ribosomal protein S21', other_designations='S21mt|mitochondrial 28S ribosomal protein S21', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150293842, end_position_on_the_genomic_accession=150308979, orientation='plus', exon_count=3, OMIM=611984, no_hdr0=''),
  54461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54461, CurrentID=0, Status='live', Symbol='FBXW5', Aliases='Fbw5', description='F-box and WD repeat domain containing 5', other_designations='F-box and WD-40 domain-containing protein 5|WD repeat-containing F-box protein FBW5', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136940433, end_position_on_the_genomic_accession=136944774, orientation='minus', exon_count=9, OMIM=609072, no_hdr0=''),
  54462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54462, CurrentID=0, Status='live', Symbol='CCSER2', Aliases='FAM190B, Gcap14, KIAA1128, bA486O22.1', description='coiled-coil serine rich protein 2', other_designations='coiled-coil serine-rich protein 2|family with sequence similarity 190, member B|protein FAM190B|protein GCAP14 homolog', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84328586, end_position_on_the_genomic_accession=84518521, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54463, CurrentID=0, Status='live', Symbol='FAM134B', Aliases='JK-1, JK1', description='family with sequence similarity 134 member B', other_designations='family with sequence similarity 134, member B|protein FAM134B', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16472816, end_position_on_the_genomic_accession=16617058, orientation='minus', exon_count=12, OMIM=613114, no_hdr0=''),
  54464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54464, CurrentID=0, Status='live', Symbol='XRN1', Aliases='SEP1', description="5'-3' exoribonuclease 1", other_designations='strand-exchange protein 1 homolog', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142306607, end_position_on_the_genomic_accession=142448062, orientation='minus', exon_count=44, OMIM=607994, no_hdr0=''),
  54465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54465, CurrentID=0, Status='live', Symbol='ETAA1', Aliases='ETAA16', description='Ewing tumor-associated antigen 1', other_designations="ETAA16|ewing's tumor-associated antigen 16", map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67397310, end_position_on_the_genomic_accession=67410413, orientation='plus', exon_count=8, OMIM=613196, no_hdr0=''),
  54466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54466, CurrentID=0, Status='live', Symbol='SPIN2A', Aliases='DXF34, SPIN2, dJ323P24.1', description='spindlin family member 2A', other_designations='centromeric spin2 copy|spindlin family, member 2A|spindlin homolog 2|spindlin-like protein 2|spindlin-like protein 2A', map_location='Xp11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57135650, end_position_on_the_genomic_accession=57137644, orientation='minus', exon_count=3, OMIM=300621, no_hdr0=''),
  54467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54467, CurrentID=0, Status='live', Symbol='ANKIB1', Aliases='', description='ankyrin repeat and IBR domain containing 1', other_designations='', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92246216, end_position_on_the_genomic_accession=92401770, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  54468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54468, CurrentID=0, Status='live', Symbol='MIOS', Aliases='', description='missing oocyte, meiosis regulator, homolog (Drosophila)', other_designations='H_DJ1159O04.1|WUGSC:H_DJ1159O04.1', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7566891, end_position_on_the_genomic_accession=7607514, orientation='plus', exon_count=14, OMIM=615359, no_hdr0=''),
  54469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54469, CurrentID=0, Status='live', Symbol='ZFAND6', Aliases='AWP1, ZA20D3, ZFAND5B', description='zinc finger, AN1-type domain 6', other_designations='protein associated with PRK1|zinc finger, A20 domain containing 3', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80059568, end_position_on_the_genomic_accession=80138393, orientation='plus', exon_count=14, OMIM=610183, no_hdr0=''),
  54470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54470, CurrentID=0, Status='live', Symbol='ARMCX6', Aliases='GASP10', description='armadillo repeat containing, X-linked 6', other_designations='', map_location='Xq21.33-q22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101615118, end_position_on_the_genomic_accession=101618001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54471, CurrentID=0, Status='live', Symbol='MIEF1', Aliases='HSU79252, MID51, SMCR7L, dJ1104E15.3', description='mitochondrial elongation factor 1', other_designations='SMCR7-like protein|Smith-Magenis syndrome chromosome region, candidate 7-like|mitochondrial dynamic protein MID51|mitochondrial dynamic protein of 51 kDa|mitochondrial dynamics protein of 51 kDa|smith-Magenis syndrome chromosomal region candidate gene 7 protein-like', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39500100, end_position_on_the_genomic_accession=39518134, orientation='plus', exon_count=11, OMIM=615497, no_hdr0=''),
  54472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54472, CurrentID=0, Status='live', Symbol='TOLLIP', Aliases='IL-1RAcPIP', description='toll interacting protein', other_designations='adapter protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1274368, end_position_on_the_genomic_accession=1309662, orientation='minus', exon_count=7, OMIM=606277, no_hdr0=''),
  54474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54474, CurrentID=0, Status='live', Symbol='KRT20', Aliases='CD20, CK-20, CK20, K20, KRT21', description='keratin 20, type I', other_designations='cytokeratin 20|cytokeratin-20|keratin-20|protein IT', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40875889, end_position_on_the_genomic_accession=40885243, orientation='minus', exon_count=8, OMIM=608218, no_hdr0=''),
  54475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54475, CurrentID=0, Status='live', Symbol='NLE1', Aliases='Nle', description='notchless homolog 1 (Drosophila)', other_designations='Notchless gene homolog', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35131322, end_position_on_the_genomic_accession=35142315, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54476, CurrentID=0, Status='live', Symbol='RNF216', Aliases='CAHH, TRIAD3, U7I1, UBCE7IP1, ZIN', description='ring finger protein 216', other_designations='triad domain-containing protein 3|ubiquitin-conjugating enzyme 7-interacting protein 1|zinc finger protein inhibiting NF-kappa-B', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5620041, end_position_on_the_genomic_accession=5781730, orientation='minus', exon_count=21, OMIM=609948, no_hdr0=''),
  54477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54477, CurrentID=0, Status='live', Symbol='PLEKHA5', Aliases='PEPP-2, PEPP2', description='pleckstrin homology domain containing A5', other_designations='PH domain-containing family A member 5|phosphoinositol 3-phosphate-binding protein-2|pleckstrin homology domain containing, family A member 5', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19129692, end_position_on_the_genomic_accession=19376400, orientation='plus', exon_count=35, OMIM=607770, no_hdr0=''),
  54478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54478, CurrentID=0, Status='live', Symbol='FAM64A', Aliases='CATS, RCS1', description='family with sequence similarity 64 member A', other_designations='CALM interacting protein expressed in thymus and spleen|CALM-interactor expressed in thymus and spleen|family with sequence similarity 64, member A|regulator of chromosome segregation 1|regulator of chromosome segregation protein 1', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6444205, end_position_on_the_genomic_accession=6451065, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54480, CurrentID=0, Status='live', Symbol='CHPF2', Aliases='CSGLCA-T, CSGlcAT, ChSy-3, chPF-2', description='chondroitin polymerizing factor 2', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase|chondroitin synthase-3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151232489, end_position_on_the_genomic_accession=151238827, orientation='plus', exon_count=5, OMIM=608037, no_hdr0=''),
  54482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54482, CurrentID=0, Status='live', Symbol='TRMT13', Aliases='CCDC76', description='tRNA methyltransferase 13 homolog (S. cerevisiae)', other_designations="coiled-coil domain containing 76|coiled-coil domain-containing protein 76|tRNA [Gm4] methyltransferase|tRNA guanosine-2'-O-methyltransferase TRM13 homolog", map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100133147, end_position_on_the_genomic_accession=100150498, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54487, CurrentID=0, Status='live', Symbol='DGCR8', Aliases='C22orf12, DGCRK6, Gy1, pasha', description='DGCR8 microprocessor complex subunit', other_designations='DiGeorge syndrome critical region 8|DiGeorge syndrome critical region gene 8', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20080232, end_position_on_the_genomic_accession=20111877, orientation='plus', exon_count=14, OMIM=609030, no_hdr0=''),
  54490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54490, CurrentID=0, Status='live', Symbol='UGT2B28', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B28', other_designations='UDP glycosyltransferase 2 family, polypeptide B28|UDPGT 2B28', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69280468, end_position_on_the_genomic_accession=69295050, orientation='plus', exon_count=6, OMIM=606497, no_hdr0=''),
  54491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54491, CurrentID=0, Status='live', Symbol='FAM105A', Aliases='NET20', description='family with sequence similarity 105 member A', other_designations='family with sequence similarity 105, member A|protein FAM105A', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14581782, end_position_on_the_genomic_accession=14616180, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54492, CurrentID=0, Status='live', Symbol='NEURL1B', Aliases='NEURL3, hNeur2, neur2', description='neuralized E3 ubiquitin protein ligase 1B', other_designations='E3 ubiquitin-protein ligase NEURL3|neuralized 2 alternative protein|neuralized homolog 1B|neuralized-2|neuralized-like protein 1B|neuralized-like protein 3', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172641247, end_position_on_the_genomic_accession=172691540, orientation='plus', exon_count=5, OMIM=615893, no_hdr0=''),
  54494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54494, CurrentID=0, Status='live', Symbol='C11orf71', Aliases='URLC7', description='chromosome 11 open reading frame 71', other_designations='up-regulated in lung cancer 7', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114391448, end_position_on_the_genomic_accession=114400550, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54495, CurrentID=0, Status='live', Symbol='TMX3', Aliases='PDIA13, TXNDC10', description='thioredoxin-related transmembrane protein 3', other_designations='protein disulfide isomerase family A, member 13|thioredoxin domain containing 10|thioredoxin domain-containing protein 10', map_location='18q22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68673688, end_position_on_the_genomic_accession=68715298, orientation='minus', exon_count=19, OMIM=616102, no_hdr0=''),
  54496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54496, CurrentID=0, Status='live', Symbol='PRMT7', Aliases='', description='protein arginine methyltransferase 7', other_designations='[Myelin basic protein]-arginine N-methyltransferase PRMT7|histone-arginine N-methyltransferase PRMT7|myelin basic protein-arginine N-methyltransferase', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68310974, end_position_on_the_genomic_accession=68358829, orientation='plus', exon_count=23, OMIM=610087, no_hdr0=''),
  54497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54497, CurrentID=0, Status='live', Symbol='HEATR5B', Aliases='', description='HEAT repeat containing 5B', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36980999, end_position_on_the_genomic_accession=37084375, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  54498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54498, CurrentID=0, Status='live', Symbol='SMOX', Aliases='C20orf16, PAO, PAO-1, PAO1, PAOH, PAOH1, SMO', description='spermine oxidase', other_designations='flavin containing amine oxidase|flavin-containing spermine oxidase|polyamine oxidase 1|putative cyclin G1 interacting protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4148779, end_position_on_the_genomic_accession=4187747, orientation='plus', exon_count=9, OMIM=615854, no_hdr0=''),
  54499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54499, CurrentID=0, Status='live', Symbol='TMCO1', Aliases='HP10122, PCIA3, PNAS-136, TMCC4', description='transmembrane and coiled-coil domains 1', other_designations='putative membrane protein|transmembrane and coiled-coil domains 4|xenogeneic cross-immune protein PCIA3', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165724291, end_position_on_the_genomic_accession=165768922, orientation='minus', exon_count=7, OMIM=614123, no_hdr0=''),
  54501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54501, CurrentID=0, Status='live', Symbol='CLQTL1', Aliases='CLF', description='cholesterol-lowering factor', other_designations='', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604595, no_hdr0=''),
  54502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54502, CurrentID=0, Status='live', Symbol='RBM47', Aliases='NET18', description='RNA binding motif protein 47', other_designations='RNA-binding motif protein 47', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40423255, end_position_on_the_genomic_accession=40630786, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  54503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54503, CurrentID=0, Status='live', Symbol='ZDHHC13', Aliases='HIP14L, HIP3RP', description='zinc finger, DHHC-type containing 13', other_designations='DHHC-13|HIP14-related protein|huntingtin interacting protein HIP3RP|huntingtin-interacting protein 14-related protein|huntingtin-interacting protein HIP3RP|probable palmitoyltransferase ZDHHC13|putative MAPK-activating protein PM03|putative NF-kappa-B-activating protein 209|zinc finger DHHC domain-containing protein 13|zinc finger, DHHC domain containing 13', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19117099, end_position_on_the_genomic_accession=19176420, orientation='plus', exon_count=20, OMIM=612815, no_hdr0=''),
  54504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54504, CurrentID=0, Status='live', Symbol='CPVL', Aliases='HVLP', description='carboxypeptidase, vitellogenic-like', other_designations='CP-Mac carboxypeptidase|VCP-like protein|carboxypeptidase WUG|vitellogenic carboxypeptidase-like protein', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28995631, end_position_on_the_genomic_accession=29146537, orientation='minus', exon_count=13, OMIM=609780, no_hdr0=''),
  54505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54505, CurrentID=0, Status='live', Symbol='DHX29', Aliases='DDX29', description='DEAH (Asp-Glu-Ala-His) box polypeptide 29', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 29|DEAH box protein 29|nucleic acid helicase DDXx', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55256245, end_position_on_the_genomic_accession=55307693, orientation='minus', exon_count=27, OMIM=612720, no_hdr0=''),
  54507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54507, CurrentID=0, Status='live', Symbol='ADAMTSL4', Aliases='ADAMTSL-4, ECTOL2, TSRC1', description='ADAMTS like 4', other_designations='ADAMTS-like 4|thrombospondin repeat-containing protein 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150549369, end_position_on_the_genomic_accession=150560937, orientation='plus', exon_count=22, OMIM=610113, no_hdr0=''),
  54508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54508, CurrentID=0, Status='live', Symbol='EPB41L4A-AS2', Aliases='', description='EPB41L4A antisense RNA 2 (head to head)', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112419583, end_position_on_the_genomic_accession=112420978, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54509, CurrentID=0, Status='live', Symbol='RHOF', Aliases='ARHF, RIF', description='ras homolog family member F (in filopodia)', other_designations='ras homolog gene family, member F (in filopodia)|rho family GTPase Rif|rho in filopodia', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121777754, end_position_on_the_genomic_accession=121793688, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54510, CurrentID=0, Status='live', Symbol='PCDH18', Aliases='PCDH68L', description='protocadherin 18', other_designations='protocadherin 68-like protein', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137518918, end_position_on_the_genomic_accession=137532498, orientation='minus', exon_count=4, OMIM=608287, no_hdr0=''),
  54511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54511, CurrentID=0, Status='live', Symbol='HMGCLL1', Aliases='bA418P12.1, er-cHL', description='3-hydroxymethyl-3-methylglutaryl-CoA lyase-like 1', other_designations='3-hydroxymethyl-3-methylglutaryl-CoA lyase-like protein 1|3-hydroxymethyl-3-methylglutaryl-Coenzyme A lyase-like 1|endoplasmic reticulum 3-hydroxymethyl-3-methylglutaryl-CoA lyase|probable 3-hydroxymethyl-3-methylglutaryl-CoA lyase 2', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55434373, end_position_on_the_genomic_accession=55679668, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54512, CurrentID=0, Status='live', Symbol='EXOSC4', Aliases='RRP41, RRP41A, Rrp41p, SKI6, Ski6p, hRrp41p, p12A', description='exosome component 4', other_designations='exosome complex exonuclease RRP41|exosome component Rrp41|ribosomal RNA-processing protein 41', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144064087, end_position_on_the_genomic_accession=144080648, orientation='plus', exon_count=4, OMIM=606491, no_hdr0=''),
  54514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54514, CurrentID=0, Status='live', Symbol='DDX4', Aliases='VASA', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 4', other_designations='DEAD box protein 4|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 4|vasa homolog', map_location='5p15.2-p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55737995, end_position_on_the_genomic_accession=55817147, orientation='plus', exon_count=23, OMIM=605281, no_hdr0=''),
  54516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54516, CurrentID=0, Status='live', Symbol='MTRF1L', Aliases='HMRF1L, MRF1L, mtRF1a', description='mitochondrial translational release factor 1 like', other_designations='mitochondrial release factor 1 like|mitochondrial translational release factor 1-like', map_location='6q25-q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152987366, end_position_on_the_genomic_accession=153002815, orientation='minus', exon_count=7, OMIM=613542, no_hdr0=''),
  54517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54517, CurrentID=0, Status='live', Symbol='PUS7', Aliases='', description='pseudouridylate synthase 7 (putative)', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105456511, end_position_on_the_genomic_accession=105522242, orientation='minus', exon_count=17, OMIM=616261, no_hdr0=''),
  54518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54518, CurrentID=0, Status='live', Symbol='APBB1IP', Aliases='INAG1, PREL1, RARP1, RIAM', description='amyloid beta (A4) precursor protein-binding, family B, member 1 interacting protein', other_designations='APBB1-interacting protein 1|PREL-1|RARP-1|Rap1-GTP-interacting adaptor molecule|Rap1-interacting adaptor molecule|proline rich EVH1 ligand 1|proline-rich EVH1 ligand 1|proline-rich protein 73|rap1-GTP-interacting adapter molecule|retinoic acid-responsive proline-rich protein 1', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26438203, end_position_on_the_genomic_accession=26567879, orientation='plus', exon_count=16, OMIM=609036, no_hdr0=''),
  54520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54520, CurrentID=0, Status='live', Symbol='CCDC93', Aliases='', description='coiled-coil domain containing 93', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117915478, end_position_on_the_genomic_accession=118014163, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  54521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54521, CurrentID=0, Status='live', Symbol='WDR44', Aliases='RAB11BP, RPH11', description='WD repeat domain 44', other_designations='rabphilin-11', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118346073, end_position_on_the_genomic_accession=118449961, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  54522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54522, CurrentID=0, Status='live', Symbol='ANKRD16', Aliases='', description='ankyrin repeat domain 16', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5861726, end_position_on_the_genomic_accession=5889897, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372291, CurrentID=0, Status='live', Symbol='LOC105372291', Aliases='', description='uncharacterized LOC105372291', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16014136, end_position_on_the_genomic_accession=16015056, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54529, CurrentID=0, Status='live', Symbol='ASNSD1', Aliases='NBLA00058, NS3TP1', description='asparagine synthetase domain containing 1', other_designations='HCV NS3-transactivated protein 1', map_location='2p24.3-q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189661399, end_position_on_the_genomic_accession=189670831, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54531, CurrentID=0, Status='live', Symbol='MIER2', Aliases='KIAA1193, Mi-er2', description='mesoderm induction early response 1, family member 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=305573, end_position_on_the_genomic_accession=344793, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  54532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54532, CurrentID=0, Status='live', Symbol='USP53', Aliases='', description='ubiquitin specific peptidase 53', other_designations='inactive ubiquitin-specific peptidase 53|ubiquitin specific protease 53|ubiquitin specific proteinase 53', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119212618, end_position_on_the_genomic_accession=119295518, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  54534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54534, CurrentID=0, Status='live', Symbol='MRPL50', Aliases='MRP-L50', description='mitochondrial ribosomal protein L50', other_designations='L50mt|mitochondrial 39S ribosomal protein L50', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101387633, end_position_on_the_genomic_accession=101398641, orientation='minus', exon_count=3, OMIM=611854, no_hdr0=''),
  54535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54535, CurrentID=0, Status='live', Symbol='CCHCR1', Aliases='C6orf18, HCR, SBP', description='coiled-coil alpha-helical rod protein 1', other_designations='HCR (a-helix coiled-coil rod homologue)|StAR-binding protein|alpha-helical coiled-coil rod protein|pg8|putative gene 8 protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31142439, end_position_on_the_genomic_accession=31158238, orientation='minus', exon_count=20, OMIM=605310, no_hdr0=''),
  54536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54536, CurrentID=0, Status='live', Symbol='EXOC6', Aliases='EXOC6A, SEC15, SEC15L, SEC15L1, SEC15L3, Sec15p', description='exocyst complex component 6', other_designations='SEC15-like 1|SEC15-like protein 1|SEC15-like protein 3|exocyst complex component Sec15A', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92834713, end_position_on_the_genomic_accession=93059494, orientation='plus', exon_count=24, OMIM=609672, no_hdr0=''),
  54537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54537, CurrentID=0, Status='live', Symbol='FAM35A', Aliases='FAM35A1, bA163M19.1', description='family with sequence similarity 35 member A', other_designations='family with sequence similarity 35, member A', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87095196, end_position_on_the_genomic_accession=87191469, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  54538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54538, CurrentID=0, Status='live', Symbol='ROBO4', Aliases='ECSM4, MRB', description='roundabout guidance receptor 4', other_designations='roundabout homolog 4, magic roundabout|roundabout, axon guidance receptor, homolog 4', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124884218, end_position_on_the_genomic_accession=124897935, orientation='minus', exon_count=18, OMIM=607528, no_hdr0=''),
  54539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54539, CurrentID=0, Status='live', Symbol='NDUFB11', Aliases='CI-ESSS, ESSS, LSDMCA3, NP17.3, Np15, P17.3', description='NADH:ubiquinone oxidoreductase subunit B11', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 11, 17.3kDa|NADH-ubiquinone oxidoreductase ESSS subunit|complex I NP17.3 subunit|complex I-ESSS|neuronal protein 17.3', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47142216, end_position_on_the_genomic_accession=47145210, orientation='minus', exon_count=3, OMIM=300403, no_hdr0=''),
  54540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54540, CurrentID=0, Status='live', Symbol='FAM193B', Aliases='IRIZIO', description='family with sequence similarity 193 member B', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177519789, end_position_on_the_genomic_accession=177554547, orientation='minus', exon_count=17, OMIM=615813, no_hdr0=''),
  54541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54541, CurrentID=0, Status='live', Symbol='DDIT4', Aliases='Dig2, REDD-1, REDD1', description='DNA damage inducible transcript 4', other_designations='DNA-damage-inducible transcript 4|HIF-1 responsive protein RTP801|protein regulated in development and DNA damage response 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72273919, end_position_on_the_genomic_accession=72276039, orientation='plus', exon_count=3, OMIM=607729, no_hdr0=''),
  54542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54542, CurrentID=0, Status='live', Symbol='RC3H2', Aliases='MNAB, RNF164', description='ring finger and CCCH-type domains 2', other_designations='RING finger protein 164|membrane associated DNA binding protein|membrane-associated nucleic acid binding protein|membrane-associated nucleic acid-binding protein|ring finger and CCCH-type zinc finger domain-containing protein 2|ring finger and CCCH-type zinc finger domains 2', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122849453, end_position_on_the_genomic_accession=122905283, orientation='minus', exon_count=21, OMIM=615231, no_hdr0=''),
  54543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54543, CurrentID=0, Status='live', Symbol='TOMM7', Aliases='TOM7', description='translocase of outer mitochondrial membrane 7 homolog (yeast)', other_designations='6.2 kd protein|translocase of outer membrane 7 kDa subunit homolog', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22812632, end_position_on_the_genomic_accession=22822852, orientation='minus', exon_count=3, OMIM=607980, no_hdr0=''),
  54544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54544, CurrentID=0, Status='live', Symbol='CRCT1', Aliases='C1orf42, NICE-1', description='cysteine rich C-terminal 1', other_designations='cysteine-rich C-terminal 1|protein NICE-1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152514502, end_position_on_the_genomic_accession=152516010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54545, CurrentID=0, Status='live', Symbol='MTMR12', Aliases='3-PAP, PIP3AP', description='myotubularin related protein 12', other_designations='3-phosphatase adapter protein|3-phosphatase adapter subunit|phosphatidylinositol 3 phosphate 3-phosphatase adapter subunit|phosphatidylinositol-3 phosphate 3-phosphatase adapter subunit|phosphatidylinositol-3 phosphate 3-phosphatase adaptor subunit|phosphatidylinositol-3-phosphate associated protein', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32227007, end_position_on_the_genomic_accession=32313008, orientation='minus', exon_count=18, OMIM=606501, no_hdr0=''),
  54546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54546, CurrentID=0, Status='live', Symbol='RNF186', Aliases='', description='ring finger protein 186', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19814029, end_position_on_the_genomic_accession=19815278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54549, CurrentID=0, Status='live', Symbol='SDK2', Aliases='', description='sidekick cell adhesion molecule 2', other_designations='sidekick homolog 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73334384, end_position_on_the_genomic_accession=73644484, orientation='minus', exon_count=46, OMIM=607217, no_hdr0=''),
  54550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54550, CurrentID=0, Status='live', Symbol='NECAB2', Aliases='EFCBP2', description='N-terminal EF-hand calcium binding protein 2', other_designations='EF hand calcium binding protein 2|EF-hand calcium binding protein 2|EF-hand calcium-binding protein 2|neuronal calcium binding 2|neuronal calcium-binding protein 2|stip-2|synaptotagmin-interacting protein 2', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83965185, end_position_on_the_genomic_accession=84002776, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54551, CurrentID=0, Status='live', Symbol='MAGEL2', Aliases='NDNL1, PWLS, SHFYNG, nM15', description='MAGE family member L2', other_designations='melanoma antigen family L2|necdin-like protein 1|protein nM15', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23643549, end_position_on_the_genomic_accession=23647846, orientation='minus', exon_count=1, OMIM=605283, no_hdr0=''),
  54552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54552, CurrentID=0, Status='live', Symbol='GNL3L', Aliases='', description='guanine nucleotide binding protein-like 3 (nucleolar)-like', other_designations='novel GTPase', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54530211, end_position_on_the_genomic_accession=54621565, orientation='plus', exon_count=17, OMIM=300873, no_hdr0=''),
  54553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54553, CurrentID=0, Status='live', Symbol='DKFZP434I0714', Aliases='', description='uncharacterized protein DKFZP434I0714', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152536264, end_position_on_the_genomic_accession=152539263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54554, CurrentID=0, Status='live', Symbol='WDR5B', Aliases='', description='WD repeat domain 5B', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122411853, end_position_on_the_genomic_accession=122416035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54555, CurrentID=0, Status='live', Symbol='DDX49', Aliases='R27090_2', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 49', other_designations='DEAD box protein 49', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18919675, end_position_on_the_genomic_accession=18928633, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54556, CurrentID=0, Status='live', Symbol='ING3', Aliases='Eaf4, ING2, MEAF4, p47ING3', description='inhibitor of growth family member 3', other_designations='inhibitor of growth family, member 3', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120950763, end_position_on_the_genomic_accession=120975657, orientation='plus', exon_count=13, OMIM=607493, no_hdr0=''),
  54557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54557, CurrentID=0, Status='live', Symbol='SGTB', Aliases='SGT2', description='small glutamine-rich tetratricopeptide repeat (TPR)-containing, beta', other_designations='beta-SGT|small glutamine rich protein with tetratricopeptide repeats 2|small glutamine-rich protein with tetratricopeptide repeats 2', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65665928, end_position_on_the_genomic_accession=65722821, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54558, CurrentID=0, Status='live', Symbol='SPATA6', Aliases='HASH, SRF-1, SRF1', description='spermatogenesis associated 6', other_designations='spermatogenesis-related factor-1', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48260553, end_position_on_the_genomic_accession=48472208, orientation='minus', exon_count=18, OMIM=613947, no_hdr0=''),
  54566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54566, CurrentID=0, Status='live', Symbol='EPB41L4B', Aliases='CG1, EHM2', description='erythrocyte membrane protein band 4.1 like 4B', other_designations='FERM-containing protein CG1', map_location='9q31-q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109171974, end_position_on_the_genomic_accession=109320964, orientation='minus', exon_count=29, OMIM=610340, no_hdr0=''),
  54567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54567, CurrentID=0, Status='live', Symbol='DLL4', Aliases='hdelta2', description='delta-like 4 (Drosophila)', other_designations='delta 4|delta ligand 4|delta-like 4 homolog|delta-like 4 protein|delta4|drosophila Delta homolog 4|notch ligand DLL4|notch ligand delta-2', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40929333, end_position_on_the_genomic_accession=40939060, orientation='plus', exon_count=11, OMIM=605185, no_hdr0=''),
  54568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54568, CurrentID=0, Status='live', Symbol='UGT2B29P', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B29 pseudogene', other_designations='UDP glucuronosyltransferase 2 family, polypeptide B15 pseudogene|UDP glycosyltransferase 2 family, polypeptide B28 pseudogene', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68516918, end_position_on_the_genomic_accession=68517653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54569, CurrentID=0, Status='live', Symbol='UGT2B27P', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B27 pseudogene', other_designations='UDP glycosyltransferase 2 family, polypeptide B27 pseudogene', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69019664, end_position_on_the_genomic_accession=69020671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54570, CurrentID=0, Status='live', Symbol='UGT2B26P', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B26 pseudogene', other_designations='UDP glycosyltransferase 2 family, polypeptide B26 pseudogene', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69027785, end_position_on_the_genomic_accession=69045018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54571, CurrentID=0, Status='live', Symbol='UGT2B25P', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B25 pseudogene', other_designations='UDP glycosyltransferase 2 family, polypeptide B25 pseudogene', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69389483, end_position_on_the_genomic_accession=69402194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54572, CurrentID=0, Status='live', Symbol='UGT2B24P', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B24 pseudogene', other_designations='UDP glucuronosyltransferase 2 family, polypeptide B11 pseudogene|UDP glycosyltransferase 2 family, polypeptide B24 pseudogene', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69408828, end_position_on_the_genomic_accession=69423275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54573, CurrentID=0, Status='live', Symbol='UGT1A12P', Aliases='UGT1L', description='UDP glucuronosyltransferase 1 family, polypeptide A12 pseudogene', other_designations='UDP glycosyltransferase 1 family, polypeptide A12 pseudogene', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233585439, end_position_on_the_genomic_accession=233586291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54574, CurrentID=0, Status='live', Symbol='UGT1A11P', Aliases='UGT1K', description='UDP glucuronosyltransferase 1 family, polypeptide A11 pseudogene', other_designations='UDP glycosyltransferase 1 family, polypeptide A11 pseudogene', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233603553, end_position_on_the_genomic_accession=233604394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54575, CurrentID=0, Status='live', Symbol='UGT1A10', Aliases='UDPGT, UGT-1J, UGT1-10, UGT1.10, UGT1J', description='UDP glucuronosyltransferase 1 family, polypeptide A10', other_designations='UDP glycosyltransferase 1 family, polypeptide A10|UDP-glucuronosyltransferase 1-J|UDP-glucuronosyltransferase 1A10|UDPGT 1-10', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233636477, end_position_on_the_genomic_accession=233773305, orientation='plus', exon_count=5, OMIM=606435, no_hdr0=''),
  54576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54576, CurrentID=0, Status='live', Symbol='UGT1A8', Aliases='UDPGT, UDPGT 1-8, UGT-1H, UGT1-08, UGT1.8S, UGT1H, UGT1A8', description='UDP glucuronosyltransferase 1 family, polypeptide A8', other_designations='UDP glycosyltransferase 1 family, polypeptide A8|UDP-glucuronosyltransferase 1 family polypeptide A8s|UDP-glucuronosyltransferase 1-H|UDP-glucuronosyltransferase 1A8', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233617645, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=606433, no_hdr0=''),
  54577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54577, CurrentID=0, Status='live', Symbol='UGT1A7', Aliases='UDPGT, UDPGT 1-7, UGT-1G, UGT1-07, UGT1.7, UGT1G', description='UDP glucuronosyltransferase 1 family, polypeptide A7', other_designations='UDP glycosyltransferase 1 family, polypeptide A7|UDP-glucuronosyltransferase 1 family polypeptide A7s|UDP-glucuronosyltransferase 1-G|UDP-glucuronosyltransferase 1A7', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233681938, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=606432, no_hdr0=''),
  54578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54578, CurrentID=0, Status='live', Symbol='UGT1A6', Aliases='GNT1, HLUGP, HLUGP1, UDPGT, UDPGT 1-6, UGT1S, UGT1F, UGT1A6', description='UDP glucuronosyltransferase 1 family, polypeptide A6', other_designations='UDP glycosyltransferase 1 family, polypeptide A6|UDP-glucuronosyltransferase 1 family polypeptide A6s|UDP-glucuronosyltransferase 1-F|UDP-glucuronosyltransferase 1A6|UGT-1F|UGT1*6|UGT1-06|UGT1.6|phenol-metabolizing UDP-glucuronosyltransferase', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233691675, end_position_on_the_genomic_accession=233773305, orientation='plus', exon_count=6, OMIM=606431, no_hdr0=''),
  54579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54579, CurrentID=0, Status='live', Symbol='UGT1A5', Aliases='UDPGT, UDPGT 1-5, UGT1E', description='UDP glucuronosyltransferase 1 family, polypeptide A5', other_designations='UDP glycosyltransferase 1 family, polypeptide A5|UDP-glucuronosyltransferase 1 family polypeptide A5s|UDP-glucuronosyltransferase 1-E|UDP-glucuronosyltransferase 1A5|UGT-1E|UGT1*5', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233712992, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=606430, no_hdr0=''),
  54580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54580, CurrentID=0, Status='live', Symbol='UGT1A2P', Aliases='UGT1BP', description='UDP glucuronosyltransferase 1 family, polypeptide A2 pseudogene', other_designations='UDP glycosyltransferase 1 family, polypeptide A2 pseudogene', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233747139, end_position_on_the_genomic_accession=233748540, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54581, CurrentID=0, Status='live', Symbol='SCAND2P', Aliases='SCAND2', description='SCAN domain containing 2 pseudogene', other_designations='SCAN domain-containing 2|egl nine homolog 1 pseudogene', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84631460, end_position_on_the_genomic_accession=84642463, orientation='plus', exon_count=5, OMIM=610417, no_hdr0=''),
  54583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54583, CurrentID=0, Status='live', Symbol='EGLN1', Aliases='C1orf12, ECYT3, HALAH, HIF-PH2, HIFPH2, HPH-2, HPH2, PHD2, SM20, ZMYND6', description='egl-9 family hypoxia-inducible factor 1', other_designations='HIF prolyl hydroxylase 2|HIF-prolyl hydroxylase 2|egl nine-like protein 1|hypoxia-inducible factor prolyl hydroxylase 2|prolyl hydroxylase domain-containing protein 2|zinc finger MYND domain-containing protein 6', map_location='1q42.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231363751, end_position_on_the_genomic_accession=231425044, orientation='minus', exon_count=5, OMIM=606425, no_hdr0=''),
  54584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54584, CurrentID=0, Status='live', Symbol='GNB1L', Aliases='DGCRK3, FKSG1, GY2, WDR14, WDVCF', description='guanine nucleotide binding protein (G protein), beta polypeptide 1-like', other_designations='G-protein beta subunit-like protein|WD repeat-containing protein 14|WD40 repeat-containing protein deleted in VCFS|g protein subunit beta-like protein 1|guanine nucleotide binding protein beta-subunit-like polypeptide', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19788411, end_position_on_the_genomic_accession=19854939, orientation='minus', exon_count=8, OMIM=610778, no_hdr0=''),
  54585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54585, CurrentID=0, Status='live', Symbol='LZTFL1', Aliases='BBS17', description='leucine zipper transcription factor like 1', other_designations='leucine zipper transcription factor-like 1', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45823316, end_position_on_the_genomic_accession=45915724, orientation='minus', exon_count=16, OMIM=606568, no_hdr0=''),
  54586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54586, CurrentID=0, Status='live', Symbol='EQTN', Aliases='AFAF, C9orf11, SPACA8', description='equatorin', other_designations='acrosome formation associated factor|acrosome formation-associated factor|equatorin, sperm acrosome associated|sperm acrosome associated 8', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27284654, end_position_on_the_genomic_accession=27297139, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54587, CurrentID=0, Status='live', Symbol='MXRA8', Aliases='', description='matrix-remodelling associated 8', other_designations='limitrin', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1352689, end_position_on_the_genomic_accession=1363541, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  729995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729995, CurrentID=0, Status='live', Symbol='LOC729995', Aliases='', description='hCG1817208', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=167871479, end_position_on_the_genomic_accession=167875742, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54596, CurrentID=0, Status='live', Symbol='L1TD1', Aliases='ECAT11', description='LINE-1 type transposase domain containing 1', other_designations='ES cell-associated protein 11', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62194802, end_position_on_the_genomic_accession=62212328, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369142, CurrentID=0, Status='live', Symbol='LOC105369142', Aliases='', description='uncharacterized LOC105369142', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113873517, end_position_on_the_genomic_accession=113889943, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54600, CurrentID=0, Status='live', Symbol='UGT1A9', Aliases='HLUGP4, LUGP4, UDPGT, UDPGT 1-9, UGT-1I, UGT1-09, UGT1-9, UGT1.9S, UGT1AI, UGT1I, UGT1A9', description='UDP glucuronosyltransferase 1 family, polypeptide A9', other_designations='UDP glycosyltransferase 1 family, polypeptide A9|UDP-glucuronosyltransferase 1 family polypeptide A9s|UDP-glucuronosyltransferase 1-I|UDP-glucuronosyltransferase 1A9', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233671898, end_position_on_the_genomic_accession=233773305, orientation='plus', exon_count=5, OMIM=606434, no_hdr0=''),
  54602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54602, CurrentID=0, Status='live', Symbol='NDFIP2', Aliases='N4WBP5A', description='Nedd4 family interacting protein 2', other_designations='MAPK-activating protein PM04 PM05 PM06 PM07|NEDD4 WW domain-binding protein 5A|NF-kappa-B-activating protein 413|putative MAPK-activating protein PM04/PM05/PM06/PM07|putative NF-kappa-B-activating protein 413', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79480722, end_position_on_the_genomic_accession=79556077, orientation='plus', exon_count=8, OMIM=610041, no_hdr0=''),
  54606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54606, CurrentID=0, Status='live', Symbol='DDX56', Aliases='DDX21, DDX26, NOH61', description='DEAD (Asp-Glu-Ala-Asp) box helicase 56', other_designations='61-kd nucleolar helicase|ATP-dependent 61 kDa nucleolar RNA helicase|DEAD (Asp-Glu-Ala-Asp) box polypeptide 56|DEAD box protein 21|DEAD box protein 56|DEAD-box RNA helicase|nucleolar helicase of 61 kDa|putative nucleolar RNA helicase', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44565417, end_position_on_the_genomic_accession=44574538, orientation='minus', exon_count=14, OMIM=608023, no_hdr0=''),
  402317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402317, CurrentID=0, Status='live', Symbol='OR2A42', Aliases='', description='olfactory receptor family 2 subfamily A member 42', other_designations='olfactory receptor OR7-16|olfactory receptor OR7-19|olfactory receptor, family 2, subfamily A, member 42', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144231911, end_position_on_the_genomic_accession=144232843, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373393, CurrentID=0, Status='live', Symbol='LOC105373393', Aliases='', description='uncharacterized LOC105373393', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3697341, end_position_on_the_genomic_accession=3703500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54617, CurrentID=0, Status='live', Symbol='INO80', Aliases='INO80A, INOC1, hINO80', description='INO80 complex subunit', other_designations='INO80 complex subunit A|INO80 homolog|homolog of yeast INO80|putative DNA helicase INO80 complex homolog 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40978880, end_position_on_the_genomic_accession=41116354, orientation='minus', exon_count=37, OMIM=610169, no_hdr0=''),
  54619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54619, CurrentID=0, Status='live', Symbol='CCNJ', Aliases='bA690P14.1', description='cyclin J', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96041774, end_position_on_the_genomic_accession=96060870, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54620, CurrentID=0, Status='live', Symbol='FBXL19', Aliases='CXXC11, Fbl19, JHDM1C', description='F-box and leucine-rich repeat protein 19', other_designations='jumonji C domain-containing histone demethylase 1C', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30923055, end_position_on_the_genomic_accession=30948783, orientation='plus', exon_count=12, OMIM=609085, no_hdr0=''),
  54621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54621, CurrentID=0, Status='live', Symbol='VSIG10', Aliases='', description='V-set and immunoglobulin domain containing 10', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118063593, end_position_on_the_genomic_accession=118104005, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54622, CurrentID=0, Status='live', Symbol='ARL15', Aliases='ARFRP2', description='ADP ribosylation factor like GTPase 15', other_designations='ADP-ribosylation factor like GTPase 15|ADP-ribosylation factor related protein 2|ADP-ribosylation factor-like 15|ADP-ribosylation factor-related protein 2|ARF-related protein 2', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53884748, end_position_on_the_genomic_accession=54310586, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54623, CurrentID=0, Status='live', Symbol='PAF1', Aliases='F23149_1, PD2', description='PAF1 homolog, Paf1/RNA polymerase II complex component', other_designations='Paf1, RNA polymerase II associated factor, homolog|pancreatic differentiation protein 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39385629, end_position_on_the_genomic_accession=39391195, orientation='minus', exon_count=15, OMIM=610506, no_hdr0=''),
  54625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54625, CurrentID=0, Status='live', Symbol='PARP14', Aliases='ARTD8, BAL2, PARP-14, pART8', description='poly(ADP-ribose) polymerase family member 14', other_designations='ADP-ribosyltransferase diphtheria toxin-like 8|B-aggressive lymphoma 2|b aggressive lymphoma protein 2|collaborator of STAT6|poly (ADP-ribose) polymerase family, member 14', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122680763, end_position_on_the_genomic_accession=122730840, orientation='plus', exon_count=19, OMIM=610028, no_hdr0=''),
  54626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54626, CurrentID=0, Status='live', Symbol='HES2', Aliases='bHLHb40', description='hes family bHLH transcription factor 2', other_designations='class B basic helix-loop-helix protein 40|hairy and enhancer of split 2', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6415232, end_position_on_the_genomic_accession=6424664, orientation='minus', exon_count=6, OMIM=609970, no_hdr0=''),
  54627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54627, CurrentID=0, Status='live', Symbol='MAP10', Aliases='KIAA1383, MTR120', description='microtubule associated protein 10', other_designations='microtubule regulator 120 KDa|microtubule regulator of 120 KDa', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232804892, end_position_on_the_genomic_accession=232810346, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54629, CurrentID=0, Status='live', Symbol='FAM63B', Aliases='', description='family with sequence similarity 63 member B', other_designations='family with sequence similarity 63, member B', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58771194, end_position_on_the_genomic_accession=58861900, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105369148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369148, CurrentID=0, Status='live', Symbol='LOC105369148', Aliases='', description='uncharacterized LOC105369148', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=145000545, end_position_on_the_genomic_accession=145035864, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369150, CurrentID=0, Status='live', Symbol='LOC105369150', Aliases='', description='uncharacterized LOC105369150', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50298563, end_position_on_the_genomic_accession=50360997, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105369151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369151, CurrentID=0, Status='live', Symbol='LOC105369151', Aliases='', description='uncharacterized LOC105369151', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22326022, end_position_on_the_genomic_accession=22333987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730005, CurrentID=0, Status='live', Symbol='SEC14L6', Aliases='', description='SEC14-like lipid binding 6', other_designations='SEC14-like 6', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30524606, end_position_on_the_genomic_accession=30546717, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  54657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54657, CurrentID=0, Status='live', Symbol='UGT1A4', Aliases='HUG-BR2, UDPGT, UDPGT 1-4, UGT-1D, UGT1-04, UGT1.4S, UGT1D, UGT1A4', description='UDP glucuronosyltransferase 1 family, polypeptide A4', other_designations='UDP glycosyltransferase 1 family, polypeptide A4|UDP-glucuronosyltransferase 1 family polypeptide A4s|UDP-glucuronosyltransferase 1-D|UDP-glucuronosyltransferase 1A4|bilirubin UDP-glucuronosyltransferase isozyme 2|bilirubin-specific UDPGT isozyme 2', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233718792, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=606429, no_hdr0=''),
  54658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54658, CurrentID=0, Status='live', Symbol='UGT1A1', Aliases='BILIQTL1, GNT1, HUG-BR1, UDPGT, UDPGT 1-1, UGT1, UGT1A', description='UDP glucuronosyltransferase 1 family, polypeptide A1', other_designations='UDP glycosyltransferase 1 family, polypeptide A1|UDP-glucuronosyltransferase 1-A|UDP-glucuronosyltransferase 1A1|UGT-1A|UGT1*1|UGT1-01|UGT1.1|bilirubin UDP-glucuronosyltranserase|bilirubin UDP-glucuronosyltransferase 1-1|bilirubin UDP-glucuronosyltransferase isozyme 1|bilirubin-specific UDPGT isozyme 1', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233760273, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=191740, no_hdr0=''),
  54659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54659, CurrentID=0, Status='live', Symbol='UGT1A3', Aliases='UDPGT, UDPGT 1-3, UGT-1C, UGT1-03, UGT1.3S, UGT1C, UGT1A3', description='UDP glucuronosyltransferase 1 family, polypeptide A3', other_designations='UDP glycosyltransferase 1 family, polypeptide A3|UDP-glucuronosyltransferase 1 family polypeptide A3s|UDP-glucuronosyltransferase 1-C|UDP-glucuronosyltransferase 1A3|UGT1*3', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233729127, end_position_on_the_genomic_accession=233773299, orientation='plus', exon_count=5, OMIM=606428, no_hdr0=''),
  54660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54660, CurrentID=0, Status='live', Symbol='PCDHB18P', Aliases='PCDH-psi2, PCDHB18', description='protocadherin beta 18 pseudogene', other_designations='protocadherin beta pseudogene', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141234333, end_position_on_the_genomic_accession=141237529, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54661, CurrentID=0, Status='live', Symbol='PCDHB17P', Aliases='ME4, PCDH-psi1, PCDHB17', description='protocadherin beta 17 pseudogene', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141155999, end_position_on_the_genomic_accession=141158409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54662, CurrentID=0, Status='live', Symbol='TBC1D13', Aliases='', description='TBC1 domain family member 13', other_designations='TBC1 domain family, member 13', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128787207, end_position_on_the_genomic_accession=128810432, orientation='plus', exon_count=12, OMIM=616218, no_hdr0=''),
  54663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54663, CurrentID=0, Status='live', Symbol='WDR74', Aliases='', description='WD repeat domain 74', other_designations='NOP seven-associated protein 1', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62832908, end_position_on_the_genomic_accession=62840572, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54664, CurrentID=0, Status='live', Symbol='TMEM106B', Aliases='', description='transmembrane protein 106B', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12211222, end_position_on_the_genomic_accession=12237264, orientation='plus', exon_count=9, OMIM=613413, no_hdr0=''),
  54665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54665, CurrentID=0, Status='live', Symbol='RSBN1', Aliases='ROSBIN', description='round spermatid basic protein 1', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113761832, end_position_on_the_genomic_accession=113812476, orientation='minus', exon_count=8, OMIM=615858, no_hdr0=''),
  54674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54674, CurrentID=0, Status='live', Symbol='LRRN3', Aliases='FIGLER5, NLRR-3, NLRR3', description='leucine rich repeat neuronal 3', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 5|leucine-rich repeat protein, neuronal 3|neuronal leucine-rich repeat protein 3', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111091006, end_position_on_the_genomic_accession=111125454, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54675, CurrentID=0, Status='live', Symbol='CRLS1', Aliases='C20orf155, CLS, CLS1, GCD10, dJ967N21.6', description='cardiolipin synthase 1', other_designations='', map_location='20p13-p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6006093, end_position_on_the_genomic_accession=6040072, orientation='plus', exon_count=12, OMIM=608188, no_hdr0=''),
  54676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54676, CurrentID=0, Status='live', Symbol='GTPBP2', Aliases='', description='GTP binding protein 2', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43620481, end_position_on_the_genomic_accession=43629246, orientation='minus', exon_count=13, OMIM=607434, no_hdr0=''),
  54677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54677, CurrentID=0, Status='live', Symbol='CROT', Aliases='COT', description='carnitine O-octanoyltransferase', other_designations='peroxisomal carnitine acyltransferase', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87345635, end_position_on_the_genomic_accession=87399796, orientation='plus', exon_count=20, OMIM=606090, no_hdr0=''),
  402329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402329, CurrentID=0, Status='live', Symbol='LOC402329', Aliases='', description='ubiquitin carboxyl-terminal hydrolase 17-like protein 2-like', other_designations='ubiquitin carboxyl-terminal hydrolase 17-like pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7341826, end_position_on_the_genomic_accession=7343020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54680, CurrentID=0, Status='live', Symbol='ZNHIT6', Aliases='BCD1, C1orf181, NY-BR-75', description='zinc finger, HIT-type containing 6', other_designations='box C/D snoRNA essential 1 homolog|serologically defined breast cancer antigen NY-BR-75|zinc finger HIT domain-containing protein 6|zinc finger, HIT type 6', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85649423, end_position_on_the_genomic_accession=85708463, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54681, CurrentID=0, Status='live', Symbol='P4HTM', Aliases='EGLN4, HIFPH4, P4H-TM, PH-4, PH4, PHD4', description='prolyl 4-hydroxylase, transmembrane (endoplasmic reticulum)', other_designations='HIF-PH4|HIF-prolyl hydroxylase 4|HPH-4|P4H with transmembrane domain|Prolyl hydroxlase domain-containing 4|hypoxia inducible factor prolyl 4 hydroxylase|hypoxia-inducible factor prolyl 4-hydroxylase|hypoxia-inducible factor prolyl hydroxylase 4|proline 4-hydroxylase', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48989100, end_position_on_the_genomic_accession=49007149, orientation='plus', exon_count=10, OMIM=614584, no_hdr0=''),
  54682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54682, CurrentID=0, Status='live', Symbol='MANSC1', Aliases='9130403P13Rik, LOH12CR3', description='MANSC domain containing 1', other_designations='loss of heterozygosity 12 chromosomal region 3 protein', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12329284, end_position_on_the_genomic_accession=12350235, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369157, CurrentID=0, Status='live', Symbol='LOC105369157', Aliases='', description='uncharacterized LOC105369157', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=159550, end_position_on_the_genomic_accession=203999, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100271074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271074, CurrentID=0, Status='live', Symbol='RPS17P14', Aliases='RPS17_6_941', description='ribosomal protein S17 pseudogene 14', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108379159, end_position_on_the_genomic_accession=108379545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369159, CurrentID=0, Status='live', Symbol='LOC105369159', Aliases='', description='uncharacterized LOC105369159', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16041292, end_position_on_the_genomic_accession=16640637, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54700, CurrentID=0, Status='live', Symbol='RRN3', Aliases='A-270G1.2, TIFIA', description='RRN3 homolog, RNA polymerase I transcription factor', other_designations='RRN3 RNA polymerase I transcription factor homolog|TIF-IA|transcription initiation factor IA|transcription initiation factor TIF-IA', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15060022, end_position_on_the_genomic_accession=15094335, orientation='minus', exon_count=19, OMIM=605121, no_hdr0=''),
  402333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402333, CurrentID=0, Status='live', Symbol='OR7E160P', Aliases='', description='olfactory receptor family 7 subfamily E member 160 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 160 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11997606, end_position_on_the_genomic_accession=12034640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  54704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54704, CurrentID=0, Status='live', Symbol='PDP1', Aliases='PDH, PDP, PDPC, PPM2C', description='pyruvate dehyrogenase phosphatase catalytic subunit 1', other_designations='PDP 1|PDPC 1|[Pyruvate dehydrogenase [acetyl-transferring]]-phosphatase 1, mitochondrial|protein phosphatase 2C, magnesium-dependent, catalytic subunit|pyruvate dehydrogenase (Lipoamide) phosphatase-phosphatase|pyruvate dehydrogenase phosphatase catalytic subunit 1', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93916855, end_position_on_the_genomic_accession=93926068, orientation='plus', exon_count=5, OMIM=605993, no_hdr0=''),
  54707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54707, CurrentID=0, Status='live', Symbol='GPN2', Aliases='ATPBD1B', description='GPN-loop GTPase 2', other_designations='ATP binding domain 1 family, member B|ATP-binding domain 1 family member B', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26879382, end_position_on_the_genomic_accession=26890378, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54708, CurrentID=0, Status='live', Symbol='MARCH5', Aliases='MARCH-V, MITOL, RNF153', description='membrane associated ring-CH-type finger 5', other_designations='membrane associated ring finger 5|membrane-associated RING finger protein 5|membrane-associated RING-CH protein V|membrane-associated ring finger (C3HC4) 5|mitochondrial ubiquitin ligase|ring finger protein 153', map_location='10q23.32-q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92291163, end_position_on_the_genomic_accession=92353964, orientation='plus', exon_count=7, OMIM=610637, no_hdr0=''),
  105369161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369161, CurrentID=0, Status='live', Symbol='LOC105369161', Aliases='', description='uncharacterized LOC105369161', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46948195, end_position_on_the_genomic_accession=46956082, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54714, CurrentID=0, Status='live', Symbol='CNGB3', Aliases='ACHM1', description='cyclic nucleotide gated channel beta 3', other_designations='CNG channel beta-3|cone photoreceptor cGMP-gated cation channel beta-subunit|cyclic nucleotide-gated cation channel modulatory subunit', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86573843, end_position_on_the_genomic_accession=86743675, orientation='minus', exon_count=19, OMIM=605080, no_hdr0=''),
  54715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54715, CurrentID=0, Status='live', Symbol='RBFOX1', Aliases='2BP1, A2BP1, FOX-1, FOX1, HRNBP1', description='RNA binding protein, fox-1 homolog (C. elegans) 1', other_designations='ataxin 2-binding protein 1|fox-1 homolog A|fox-1-like RNA-binding protein 1|hexaribonucleotide-binding protein 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5239782, end_position_on_the_genomic_accession=7713340, orientation='plus', exon_count=31, OMIM=605104, no_hdr0=''),
  54716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54716, CurrentID=0, Status='live', Symbol='SLC6A20', Aliases='SIT1, XT3, Xtrp3', description='solute carrier family 6 (proline IMINO transporter), member 20', other_designations='X transporter protein 3|neurotransmitter transporter RB21A|orphan transporter XT3|sodium/imino-acid transporter 1|solute carrier family 6 (neurotransmitter transporter), member 20|transporter rB21A homolog', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45755449, end_position_on_the_genomic_accession=45796561, orientation='minus', exon_count=13, OMIM=605616, no_hdr0=''),
  54718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54718, CurrentID=0, Status='live', Symbol='BTN2A3P', Aliases='BTN2.3, BTN2A3', description='butyrophilin subfamily 2 member A3, pseudogene', other_designations='butyrophilin, subfamily 2, member A pseudogene|butyrophilin, subfamily 2, member A3, pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26421391, end_position_on_the_genomic_accession=26430588, orientation='plus', exon_count=7, OMIM=613592, no_hdr0=''),
  54719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54719, CurrentID=0, Status='live', Symbol='RNANC', Aliases='NCRNA', description='Retinal nonattachment, nonsyndromic congenital', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=221900, no_hdr0=''),
  105369163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369163, CurrentID=0, Status='live', Symbol='LOC105369163', Aliases='', description='uncharacterized LOC105369163', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85140946, end_position_on_the_genomic_accession=85243794, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54726, CurrentID=0, Status='live', Symbol='OTUD4', Aliases='DUBA6, HIN1, HSHIN1', description='OTU deubiquitinase 4', other_designations='HIV-1 induced protein HIN-1|HIV-1-induced protein HIN-1|OTU domain containing 4', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145133650, end_position_on_the_genomic_accession=145179973, orientation='minus', exon_count=22, OMIM=611744, no_hdr0=''),
  54729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54729, CurrentID=0, Status='live', Symbol='NKX1-1', Aliases='HSPX153, NKX-1.1, SAX2', description='NK1 homeobox 1', other_designations='HPX-153 homeobox|NK1 transcription factor related, locus 1|homeobox protein 153|homeobox protein SAX-2', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1402932, end_position_on_the_genomic_accession=1406442, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54732, CurrentID=0, Status='live', Symbol='TMED9', Aliases='GMP25, HSGP25L2G, p25', description='transmembrane p24 trafficking protein 9', other_designations='glycoprotein 25L2|gp25L2 protein|p24 family protein alpha-2|p24alpha2|transmembrane emp24 protein transport domain containing 9', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177592212, end_position_on_the_genomic_accession=177596107, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  54733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54733, CurrentID=0, Status='live', Symbol='SLC35F2', Aliases='HSNOV1', description='solute carrier family 35 member F2', other_designations='lung squamous cell cancer related protein LSCC-3|solute carrier family 35, member F2', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107790991, end_position_on_the_genomic_accession=107859188, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54734, CurrentID=0, Status='live', Symbol='RAB39A', Aliases='RAB39', description='RAB39A, member RAS oncogene family', other_designations='RAB39, member RAS oncogene family|rab-39|rab-related GTP-binding protein', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107928475, end_position_on_the_genomic_accession=107963482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54737, CurrentID=0, Status='live', Symbol='MPHOSPH8', Aliases='HSMPP8, TWA3, mpp8', description='M-phase phosphoprotein 8', other_designations='M-phase phosphoprotein, mpp|M-phase phosphoprotein, mpp8|two hybrid-associated protein 3 with RanBPM', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19633647, end_position_on_the_genomic_accession=19673459, orientation='plus', exon_count=16, OMIM=611626, no_hdr0=''),
  54738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54738, CurrentID=0, Status='live', Symbol='FEV', Aliases='HSRNAFEV, PET-1', description='FEV (ETS oncogene family)', other_designations='PC12 ETS domain-containing transcription factor 1|fifth Ewing variant protein', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218981087, end_position_on_the_genomic_accession=218985657, orientation='minus', exon_count=3, OMIM=607150, no_hdr0=''),
  54739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54739, CurrentID=0, Status='live', Symbol='XAF1', Aliases='BIRC4BP, HSXIAPAF1, XIAPAF1', description='XIAP associated factor 1', other_designations='BIRC4 binding protein|BIRC4-binding protein', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6755572, end_position_on_the_genomic_accession=6775647, orientation='plus', exon_count=10, OMIM=606717, no_hdr0=''),
  54741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54741, CurrentID=0, Status='live', Symbol='LEPROT', Aliases='LEPR, OB-RGRP, OBRGRP, VPS55', description='leptin receptor overlapping transcript', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 6|OB-R gene-related protein|endospanin-1|leptin receptor gene related protein|leptin receptor overlapping transcript protein', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65420448, end_position_on_the_genomic_accession=65436007, orientation='plus', exon_count=6, OMIM=613461, no_hdr0=''),
  54742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54742, CurrentID=0, Status='live', Symbol='LY6K', Aliases='CT97, HSJ001348, URLC10, ly-6K', description='lymphocyte antigen 6 complex, locus K', other_designations='cancer/testis antigen 97|up-regulated in lung cancer 10', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142699928, end_position_on_the_genomic_accession=142704170, orientation='plus', exon_count=4, OMIM=615093, no_hdr0=''),
  105369167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369167, CurrentID=0, Status='live', Symbol='LOC105369167', Aliases='', description='uncharacterized LOC105369167', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65725676, end_position_on_the_genomic_accession=65788889, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54748, CurrentID=0, Status='live', Symbol='LINC00527', Aliases='C21orf104', description='long intergenic non-protein coding RNA 527', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54749, CurrentID=0, Status='live', Symbol='EPDR1', Aliases='EPDR, MERP-1, MERP1, UCC1', description='ependymin related 1', other_designations='ependymin related protein 1|mammalian ependymin related protein 1|upregulated in colorectal cancer gene 1 protein', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37920561, end_position_on_the_genomic_accession=37951941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54751, CurrentID=0, Status='live', Symbol='FBLIM1', Aliases='CAL, FBLP-1, FBLP1', description='filamin binding LIM protein 1', other_designations='CSX-associated LIM|MIG2-interacting protein|filamin-binding LIM protein-1|migfilin|mitogen-inducible 2 interacting protein|mitogen-inducible 2-interacting protein', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15756171, end_position_on_the_genomic_accession=15786589, orientation='plus', exon_count=13, OMIM=607747, no_hdr0=''),
  54752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54752, CurrentID=0, Status='live', Symbol='FNDC8', Aliases='', description='fibronectin type III domain containing 8', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35121612, end_position_on_the_genomic_accession=35130732, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54753, CurrentID=0, Status='live', Symbol='ZNF853', Aliases='', description='zinc finger protein 853', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6615719, end_position_on_the_genomic_accession=6624290, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54754, CurrentID=0, Status='live', Symbol='NUTM2F', Aliases='FAM22F, NUTMF', description='NUT family member 2F', other_designations='family with sequence similarity 22, member F|protein FAM22F', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94318196, end_position_on_the_genomic_accession=94554847, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54756, CurrentID=0, Status='live', Symbol='IL17RD', Aliases='HH18, IL-17RD, IL17RLM, SEF', description='interleukin 17 receptor D', other_designations='IL-17 receptor D|IL17Rhom|hSef|interleukin-17 receptor-like protein|sef homolog', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57089982, end_position_on_the_genomic_accession=57170317, orientation='minus', exon_count=17, OMIM=606807, no_hdr0=''),
  54757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54757, CurrentID=0, Status='live', Symbol='FAM20A', Aliases='AI1G, AIGFS, FP2747', description='family with sequence similarity 20 member A', other_designations='family with sequence similarity 20, member A|protein FAM20A', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68534561, end_position_on_the_genomic_accession=68600954, orientation='minus', exon_count=16, OMIM=611062, no_hdr0=''),
  54758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54758, CurrentID=0, Status='live', Symbol='KLHDC4', Aliases='', description='kelch domain containing 4', other_designations='', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87696569, end_position_on_the_genomic_accession=87765997, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54759, CurrentID=0, Status='live', Symbol='HSA251708', Aliases='', description='microtubule-binding protein-like', other_designations='putative microtubule-binding protein', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  54760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54760, CurrentID=0, Status='live', Symbol='PCSK4', Aliases='PC4, SPC5', description='proprotein convertase subtilisin/kexin type 4', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1481428, end_position_on_the_genomic_accession=1490888, orientation='minus', exon_count=18, OMIM=600487, no_hdr0=''),
  54762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54762, CurrentID=0, Status='live', Symbol='GRAMD1C', Aliases='', description='GRAM domain containing 1C', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113828182, end_position_on_the_genomic_accession=113947174, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  54763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54763, CurrentID=0, Status='live', Symbol='ROPN1', Aliases='CT91, ODF6, RHPNAP1A, ropporin, ROPN1', description='rhophilin associated tail protein 1', other_designations='cancer/testis antigen 91|outer dense fiber of sperm tails 6|rhophilin-associated protein 1A|ropporin, rhophilin associated protein 1', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123969006, end_position_on_the_genomic_accession=123992170, orientation='minus', exon_count=9, OMIM=611757, no_hdr0=''),
  54764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54764, CurrentID=0, Status='live', Symbol='ZRANB1', Aliases='TRABID', description='zinc finger, RAN-binding domain containing 1', other_designations='TRAF-binding domain-containing protein|TRAF-binding protein domain|hTrabid|zinc finger Ran-binding domain-containing protein 1', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124916899, end_position_on_the_genomic_accession=124988059, orientation='plus', exon_count=12, OMIM=611749, no_hdr0=''),
  54765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54765, CurrentID=0, Status='live', Symbol='TRIM44', Aliases='DIPB, HSA249128, MC7', description='tripartite motif containing 44', other_designations='tripartite motif-containing 44', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35662692, end_position_on_the_genomic_accession=35811053, orientation='plus', exon_count=5, OMIM=612298, no_hdr0=''),
  54766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54766, CurrentID=0, Status='live', Symbol='BTG4', Aliases='PC3B', description='BTG family member 4', other_designations='B-cell translocation gene 4|protein PC3b', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111384095, end_position_on_the_genomic_accession=111514733, orientation='minus', exon_count=24, OMIM=605673, no_hdr0=''),
  54768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54768, CurrentID=0, Status='live', Symbol='HYDIN', Aliases='CILD51, HYDIN2, PPP1R31, HYDIN', description='HYDIN, axonemal central pair apparatus protein', other_designations='protein phosphatase 1, regulatory subunit 31', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70802084, end_position_on_the_genomic_accession=71230722, orientation='minus', exon_count=91, OMIM=610812, no_hdr0=''),
  54769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54769, CurrentID=0, Status='live', Symbol='DIRAS2', Aliases='Di-Ras2', description='DIRAS family, GTP-binding RAS-like 2', other_designations='distinct subgroup of the Ras family member 2', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90609832, end_position_on_the_genomic_accession=90643105, orientation='minus', exon_count=2, OMIM=607863, no_hdr0=''),
  105369171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369171, CurrentID=0, Status='live', Symbol='LOC105369171', Aliases='', description='uncharacterized LOC105369171', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162367242, end_position_on_the_genomic_accession=162422268, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54776, CurrentID=0, Status='live', Symbol='PPP1R12C', Aliases='LENG3, MBS85, p84, p85', description='protein phosphatase 1 regulatory subunit 12C', other_designations='leukocyte receptor cluster (LRC) encoded novel gene 3|leukocyte receptor cluster (LRC) member 3|myosin-binding subunit 85|protein phosphatase 1 myosin-binding subunit of 85 kDa|protein phosphatase 1 myosin-binding subunit p85|protein phosphatase 1, regulatory (inhibitor) subunit 12C|protein phosphatase 1, regulatory subunit 12C', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55090913, end_position_on_the_genomic_accession=55117600, orientation='minus', exon_count=22, OMIM=613245, no_hdr0=''),
  54777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54777, CurrentID=0, Status='live', Symbol='CFAP46', Aliases='C10orf123, C10orf124, C10orf92, C10orf93, TTC40, bA288G11.4, bA288G11.5, bB137A17.2, bB137A17.3', description='cilia and flagella associated protein 46', other_designations='TPR repeat-containing protein C10orf93|protein CFAP46|tetratricopeptide repeat domain 40|tetratricopeptide repeat protein 40', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132808392, end_position_on_the_genomic_accession=132943163, orientation='minus', exon_count=68, OMIM='', no_hdr0=''),
  54778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54778, CurrentID=0, Status='live', Symbol='RNF111', Aliases='ARK', description='ring finger protein 111', other_designations='Arkadia', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58987666, end_position_on_the_genomic_accession=59097419, orientation='plus', exon_count=17, OMIM=605840, no_hdr0=''),
  54780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54780, CurrentID=0, Status='live', Symbol='NSMCE4A', Aliases='C10orf86, NS4EA, NSE4A', description='NSE4 homolog A, SMC5-SMC6 complex component', other_designations='NSE4 homolog, SMC5-SMC6 complex component A|non-SMC element 4 homolog A', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121957088, end_position_on_the_genomic_accession=121975228, orientation='minus', exon_count=14, OMIM=612987, no_hdr0=''),
  54784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54784, CurrentID=0, Status='live', Symbol='ALKBH4', Aliases='ABH4', description='alkB homolog 4, lysine demethylase', other_designations='alkB homolog 4, lysine demthylase|alkB, alkylation repair homolog 4|alkylated DNA repair protein alkB homolog 4|probable alpha-ketoglutarate-dependent dioxygenase ABH4', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102456220, end_position_on_the_genomic_accession=102464874, orientation='minus', exon_count=3, OMIM=613302, no_hdr0=''),
  54785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54785, CurrentID=0, Status='live', Symbol='BORCS6', Aliases='C17orf59, PRO2472', description='BLOC-1 related complex subunit 6', other_designations='lyspersin', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8188333, end_position_on_the_genomic_accession=8190246, orientation='minus', exon_count=1, OMIM=616599, no_hdr0=''),
  54788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54788, CurrentID=0, Status='live', Symbol='DNAJB12', Aliases='DJ10', description='DnaJ heat shock protein family (Hsp40) member B12', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 12', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72332830, end_position_on_the_genomic_accession=72355149, orientation='minus', exon_count=8, OMIM=608376, no_hdr0=''),
  54790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54790, CurrentID=0, Status='live', Symbol='TET2', Aliases='KIAA1546, MDS', description='tet methylcytosine dioxygenase 2', other_designations='probable methylcytosine dioxygenase TET2|tet oncogene family member 2', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105145875, end_position_on_the_genomic_accession=105279803, orientation='plus', exon_count=11, OMIM=612839, no_hdr0=''),
  54793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54793, CurrentID=0, Status='live', Symbol='KCTD9', Aliases='BTBD27', description='potassium channel tetramerization domain containing 9', other_designations='potassium channel tetramerisation domain containing 9', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25427842, end_position_on_the_genomic_accession=25458468, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54795, CurrentID=0, Status='live', Symbol='TRPM4', Aliases='PFHB1BB, TRPM4', description='transient receptor potential cation channel, subfamily M, member 4', other_designations='LTrpC4|calcium-activated non-selective cation channel 1|hTRPM4|long transient receptor potential channel 4|melastatin-4', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49157759, end_position_on_the_genomic_accession=49211841, orientation='plus', exon_count=25, OMIM=606936, no_hdr0=''),
  54796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54796, CurrentID=0, Status='live', Symbol='BNC2', Aliases='BSN2', description='basonuclin 2', other_designations='', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=16409503, end_position_on_the_genomic_accession=16870838, orientation='minus', exon_count=18, OMIM=608669, no_hdr0=''),
  54797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54797, CurrentID=0, Status='live', Symbol='MED18', Aliases='p28b', description='mediator complex subunit 18', other_designations='TRAP/mediator complex subunit p28b|mediator of RNA polymerase II transcription, subunit 18 homolog', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28329002, end_position_on_the_genomic_accession=28335967, orientation='plus', exon_count=4, OMIM=612384, no_hdr0=''),
  54798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54798, CurrentID=0, Status='live', Symbol='DCHS2', Aliases='CDH27, CDHJ, CDHR7, PCDH23, PCDHJ', description='dachsous cadherin-related 2', other_designations='cadherin-27|cadherin-like 27|cadherin-like protein CDHJ|cadherin-like protein VR8|cadherin-related family member 7|protein dachsous homolog 2|protocadherin 23|protocadherin PCDHJ', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154234375, end_position_on_the_genomic_accession=154491778, orientation='minus', exon_count=30, OMIM=612486, no_hdr0=''),
  54799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54799, CurrentID=0, Status='live', Symbol='MBTD1', Aliases='SA49P01', description='mbt domain containing 1', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51177419, end_position_on_the_genomic_accession=51260493, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  54800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54800, CurrentID=0, Status='live', Symbol='KLHL24', Aliases='DRE1, KRIP6', description='kelch like family member 24', other_designations='kainate receptor interacting protein for GluR6|kainate receptor-interacting protein for GluR6|kelch-like 24|kelch-like family member 24', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183635011, end_position_on_the_genomic_accession=183684519, orientation='plus', exon_count=10, OMIM=611295, no_hdr0=''),
  54801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54801, CurrentID=0, Status='live', Symbol='HAUS6', Aliases='Dgt6, FAM29A', description='HAUS augmin like complex subunit 6', other_designations='HAUS augmin-like complex, subunit 6|dim gamma-tubulin homolog|family with sequence similarity 29, member A', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19053137, end_position_on_the_genomic_accession=19102942, orientation='minus', exon_count=18, OMIM=613433, no_hdr0=''),
  54802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54802, CurrentID=0, Status='live', Symbol='TRIT1', Aliases='GRO1, IPPT, IPT, IPTase, MOD5, hGRO1', description='tRNA isopentenyltransferase 1', other_designations='IPP transferase|isopentenyl-diphosphate:tRNA isopentenyltransferase|tRNA isopentenylpyrophosphate transferase', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39841031, end_position_on_the_genomic_accession=39883511, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100131411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131411, CurrentID=0, Status='live', Symbol='LOC100131411', Aliases='', description='zinc finger protein 28 pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74929880, end_position_on_the_genomic_accession=74930569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  54805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54805, CurrentID=0, Status='live', Symbol='CNNM2', Aliases='ACDP2, HOMG6, HOMGSMR', description='cyclin and CBS domain divalent metal cation transport mediator 2', other_designations='ancient conserved domain-containing protein 2|cyclin M2', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102918293, end_position_on_the_genomic_accession=103078587, orientation='plus', exon_count=14, OMIM=607803, no_hdr0=''),
  54806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54806, CurrentID=0, Status='live', Symbol='AHI1', Aliases='AHI-1, JBTS3, ORF1, dJ71N10.1', description='Abelson helper integration site 1', other_designations='abelson helper integration site 1 protein homolog|contatins SH3 and WD40 domains', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135283972, end_position_on_the_genomic_accession=135497765, orientation='minus', exon_count=35, OMIM=608894, no_hdr0=''),
  54807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54807, CurrentID=0, Status='live', Symbol='ZNF586', Aliases='', description='zinc finger protein 586', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57769652, end_position_on_the_genomic_accession=57780616, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54808, CurrentID=0, Status='live', Symbol='DYM', Aliases='DMC, SMC', description='dymeclin', other_designations='dyggve-Melchior-Clausen syndrome protein', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49041474, end_position_on_the_genomic_accession=49460709, orientation='minus', exon_count=23, OMIM=607461, no_hdr0=''),
  54809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54809, CurrentID=0, Status='live', Symbol='SAMD9', Aliases='C7orf5, DRIF1, NFTC, OEF1, OEF2', description='sterile alpha motif domain containing 9', other_designations='SAM domain-containing protein 9|expressed in aggressive fibromatosis', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93099513, end_position_on_the_genomic_accession=93118023, orientation='minus', exon_count=3, OMIM=610456, no_hdr0=''),
  54810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54810, CurrentID=0, Status='live', Symbol='GIPC2', Aliases='SEMCAP-2, SEMCAP2', description='GIPC PDZ domain containing family member 2', other_designations='GIPC PDZ domain containing family, member 2|PDZ domain protein GIPC2|semaF cytoplasmic domain associated protein 2|semaphorin cytoplasmic domain associated protein 2', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78024370, end_position_on_the_genomic_accession=78138449, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  54811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54811, CurrentID=0, Status='live', Symbol='ZNF562', Aliases='', description='zinc finger protein 562', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9648654, end_position_on_the_genomic_accession=9675302, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54812, CurrentID=0, Status='live', Symbol='AFTPH', Aliases='Nbla10388', description='aftiphilin', other_designations='aftiphilin protein', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64524305, end_position_on_the_genomic_accession=64593005, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54813, CurrentID=0, Status='live', Symbol='KLHL28', Aliases='BTBD5', description='kelch like family member 28', other_designations='BTB (POZ) domain containing 5|BTB/POZ domain-containing protein 5|kelch-like 28|kelch-like family member 28', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44924319, end_position_on_the_genomic_accession=44978290, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54814, CurrentID=0, Status='live', Symbol='QPCTL', Aliases='gQC', description='glutaminyl-peptide cyclotransferase-like', other_designations='glutaminyl cyclase-like|golgi-resident glutaminyl-peptide cyclotransferase|isoQC', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45692381, end_position_on_the_genomic_accession=45703990, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54815, CurrentID=0, Status='live', Symbol='GATAD2A', Aliases='p66alpha', description='GATA zinc finger domain containing 2A', other_designations='GATA zinc finger domain-containing protein 2A', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19385833, end_position_on_the_genomic_accession=19508932, orientation='plus', exon_count=18, OMIM=614997, no_hdr0=''),
  54816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54816, CurrentID=0, Status='live', Symbol='ZNF280D', Aliases='SUHW4, ZNF634', description='zinc finger protein 280D', other_designations='suppressor of hairy wing homolog 4|zinc finger protein 634', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56630176, end_position_on_the_genomic_accession=56734086, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  54819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54819, CurrentID=0, Status='live', Symbol='ZCCHC10', Aliases='', description='zinc finger, CCHC domain containing 10', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132996985, end_position_on_the_genomic_accession=133026617, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54820, CurrentID=0, Status='live', Symbol='NDE1', Aliases='HOM-TES-87, LIS4, MHAC, NDE, NUDE, NUDE1', description='nudE neurodevelopment protein 1', other_designations='LIS1-interacting protein NUDE1, rat homolog|nudE nuclear distribution E homolog 1|nudE nuclear distribution gene E homolog 1', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15643267, end_position_on_the_genomic_accession=15726353, orientation='plus', exon_count=15, OMIM=609449, no_hdr0=''),
  54821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54821, CurrentID=0, Status='live', Symbol='ERCC6L', Aliases='PICH, RAD26L', description='excision repair cross-complementation group 6 like', other_designations='ATP-dependent helicase ERCC6-like|Plk1-interacting checkpoint helicase|SNF2/RAD54 family protein|excision repair cross-complementation group 6-like|excision repair cross-complementing rodent repair deficiency complementation group 6 - like|excision repair cross-complementing rodent repair deficiency, complementation group 6-like|excision repair protein ERCC6-like|tumor antigen BJ-HCC-15', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72204657, end_position_on_the_genomic_accession=72239008, orientation='minus', exon_count=2, OMIM=300687, no_hdr0=''),
  54822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54822, CurrentID=0, Status='live', Symbol='TRPM7', Aliases='ALSPDC, CHAK, CHAK1, LTRPC7, LTrpC-7, TRP-PLIK', description='transient receptor potential cation channel, subfamily M, member 7', other_designations='LTRPC ion channel family member 7|channel-kinase 1|long transient receptor potential channel 7|transient receptor potential-phospholipase C-interacting kinase', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50557158, end_position_on_the_genomic_accession=50686815, orientation='minus', exon_count=40, OMIM=605692, no_hdr0=''),
  54823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54823, CurrentID=0, Status='live', Symbol='SWT1', Aliases='C1orf26, HsSwt1', description='SWT1, RNA endoribonuclease homolog', other_designations='SWT1 RNA endoribonuclease homolog', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185157023, end_position_on_the_genomic_accession=185291781, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  54825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54825, CurrentID=0, Status='live', Symbol='CDHR2', Aliases='PCDH24, PCLKC', description='cadherin related family member 2', other_designations='protocadherin 24|protocadherin-24', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176542511, end_position_on_the_genomic_accession=176595768, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  54826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54826, CurrentID=0, Status='live', Symbol='GIN1', Aliases='GIN-1, TGIN1, ZH2C2', description='gypsy retrotransposon integrase 1', other_designations='Ty3/Gypsy integrase 1|zinc finger H2C2 domain-containing protein', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103086000, end_position_on_the_genomic_accession=103120169, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54827, CurrentID=0, Status='live', Symbol='NXPE4', Aliases='C11orf33, FAM55D', description='neurexophilin and PC-esterase domain family member 4', other_designations='family with sequence similarity 55, member D|neurexophilin and PC-esterase domain family, member 4|protein FAM55D', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114570591, end_position_on_the_genomic_accession=114595840, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54828, CurrentID=0, Status='live', Symbol='BCAS3', Aliases='GAOB1, MAAB', description='breast carcinoma amplified sequence 3', other_designations='BCAS4/BCAS3 fusion|breast carcinoma amplified sequence 4/3 fusion protein|metastasis associated antigen of breast cancer|protein Maab1', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60677811, end_position_on_the_genomic_accession=61392838, orientation='plus', exon_count=38, OMIM=607470, no_hdr0=''),
  54829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54829, CurrentID=0, Status='live', Symbol='ASPN', Aliases='OS3, PLAP-1, PLAP1, SLRR1C', description='asporin', other_designations='asporin (LRR class 1)|asporin proteoglycan|periodontal ligament associated protein 1|small leucine-rich protein 1C', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92456205, end_position_on_the_genomic_accession=92482562, orientation='minus', exon_count=8, OMIM=608135, no_hdr0=''),
  54830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54830, CurrentID=0, Status='live', Symbol='NUP62CL', Aliases='', description='nucleoporin 62kDa C-terminal like', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107119537, end_position_on_the_genomic_accession=107206440, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54831, CurrentID=0, Status='live', Symbol='BEST2', Aliases='VMD2L1', description='bestrophin 2', other_designations='VMD2-like gene 1|vitelliform macular dystrophy 2-like 1|vitelliform macular dystrophy 2-like protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12751676, end_position_on_the_genomic_accession=12758458, orientation='plus', exon_count=10, OMIM=607335, no_hdr0=''),
  54832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54832, CurrentID=0, Status='live', Symbol='VPS13C', Aliases='', description='vacuolar protein sorting 13 homolog C (S. cerevisiae)', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61852389, end_position_on_the_genomic_accession=62060465, orientation='minus', exon_count=88, OMIM=608879, no_hdr0=''),
  54834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54834, CurrentID=0, Status='live', Symbol='GDAP2', Aliases='MACROD3', description='ganglioside induced differentiation associated protein 2', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117863485, end_position_on_the_genomic_accession=117929679, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54836, CurrentID=0, Status='live', Symbol='BSPRY', Aliases='', description='B-box and SPRY domain containing', other_designations='B-box and SPRY-domain containing protein|zetin 1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113349532, end_position_on_the_genomic_accession=113371233, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369182, CurrentID=0, Status='live', Symbol='LOC105369182', Aliases='', description='uncharacterized LOC105369182', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28341324, end_position_on_the_genomic_accession=28366318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54838, CurrentID=0, Status='live', Symbol='WBP1L', Aliases='C10orf26, OPA1L, OPAL1', description='WW domain binding protein 1-like', other_designations='outcome predictor in acute leukemia 1', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102739283, end_position_on_the_genomic_accession=102816265, orientation='plus', exon_count=8, OMIM=611129, no_hdr0=''),
  54839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54839, CurrentID=0, Status='live', Symbol='LRRC49', Aliases='PGs4', description='leucine rich repeat containing 49', other_designations='tubulin polyglutamylase complex subunit 4', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70853239, end_position_on_the_genomic_accession=71050097, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  54840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54840, CurrentID=0, Status='live', Symbol='APTX', Aliases='AOA, AOA1, AXA1, EAOH, EOAHA, FHA-HIT', description='aprataxin', other_designations='forkhead-associated domain histidine triad-like protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32972606, end_position_on_the_genomic_accession=33025138, orientation='minus', exon_count=11, OMIM=606350, no_hdr0=''),
  54841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54841, CurrentID=0, Status='live', Symbol='BIVM', Aliases='', description='basic, immunoglobulin-like variable motif containing', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102799049, end_position_on_the_genomic_accession=102841538, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54842, CurrentID=0, Status='live', Symbol='MFSD6', Aliases='MMR2, hMMR2', description='major facilitator superfamily domain containing 6', other_designations='macrophage MHC class I receptor 2 homolog|macrophage MHC receptor 2', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190408355, end_position_on_the_genomic_accession=190502315, orientation='plus', exon_count=10, OMIM=613476, no_hdr0=''),
  54843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54843, CurrentID=0, Status='live', Symbol='SYTL2', Aliases='CHR11SYT, EXO4, PPP1R151, SGA72M, SLP2, SLP2A', description='synaptotagmin like 2', other_designations='breast cancer-associated antigen SGA-72M|chromosome 11 synaptotagmin|exophilin-4|protein phosphatase 1, regulatory subunit 151|synaptotagmin-like 2', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85694221, end_position_on_the_genomic_accession=85811159, orientation='minus', exon_count=23, OMIM=612880, no_hdr0=''),
  54845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54845, CurrentID=0, Status='live', Symbol='ESRP1', Aliases='RBM35A, RMB35A', description='epithelial splicing regulatory protein 1', other_designations='RNA binding motif protein 35A|RNA-binding motif protein 35A|RNA-binding protein 35A', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94641136, end_position_on_the_genomic_accession=94707466, orientation='plus', exon_count=16, OMIM=612959, no_hdr0=''),
  54847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54847, CurrentID=0, Status='live', Symbol='SIDT1', Aliases='B830021E24Rik, SID-1, SID1', description='SID1 transmembrane family member 1', other_designations='SID1 transmembrane family, member 1', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113532371, end_position_on_the_genomic_accession=113632097, orientation='plus', exon_count=27, OMIM=606816, no_hdr0=''),
  54848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54848, CurrentID=0, Status='live', Symbol='ARHGEF38', Aliases='', description='Rho guanine nucleotide exchange factor 38', other_designations='Rho guanine nucleotide exchange factor (GEF) 38', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105552522, end_position_on_the_genomic_accession=105682539, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  54849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54849, CurrentID=0, Status='live', Symbol='DEF8', Aliases='', description='differentially expressed in FDCP 8 homolog (mouse)', other_designations='DEF-8', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89948697, end_position_on_the_genomic_accession=89968060, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  54850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54850, CurrentID=0, Status='live', Symbol='FBXL12', Aliases='Fbl12', description='F-box and leucine-rich repeat protein 12', other_designations='F-box protein FBL12', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9810267, end_position_on_the_genomic_accession=9819425, orientation='minus', exon_count=6, OMIM=609079, no_hdr0=''),
  54851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54851, CurrentID=0, Status='live', Symbol='ANKRD49', Aliases='FGIF, GBIF', description='ankyrin repeat domain 49', other_designations='fetal globin-inducing factor', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94493987, end_position_on_the_genomic_accession=94499578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54852, CurrentID=0, Status='live', Symbol='PAQR5', Aliases='MPRG', description='progestin and adipoQ receptor family member V', other_designations='mPR gamma|progestin and adipoQ receptor family member 5', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69298902, end_position_on_the_genomic_accession=69407637, orientation='plus', exon_count=17, OMIM=607781, no_hdr0=''),
  54853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54853, CurrentID=0, Status='live', Symbol='WDR55', Aliases='', description='WD repeat domain 55', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140664799, end_position_on_the_genomic_accession=140672345, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54854, CurrentID=0, Status='live', Symbol='FAM83E', Aliases='', description='family with sequence similarity 83 member E', other_designations='family with sequence similarity 83, member E', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48600600, end_position_on_the_genomic_accession=48614841, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54855, CurrentID=0, Status='live', Symbol='FAM46C', Aliases='', description='family with sequence similarity 46 member C', other_designations='family with sequence similarity 46, member C', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117605982, end_position_on_the_genomic_accession=117628389, orientation='plus', exon_count=2, OMIM=613952, no_hdr0=''),
  54856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54856, CurrentID=0, Status='live', Symbol='GON4L', Aliases='GON-4, GON4, YARP', description='gon-4-like (C. elegans)', other_designations='2610100B20Rik|YY1AP-related protein1|gon-4 homolog', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155749658, end_position_on_the_genomic_accession=155859400, orientation='minus', exon_count=35, OMIM=610393, no_hdr0=''),
  54857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54857, CurrentID=0, Status='live', Symbol='GDPD2', Aliases='GDE3, OBDPF', description='glycerophosphodiester phosphodiesterase domain containing 2', other_designations='glycerophosphodiester phosphodiesterase 3|glycerophosphodiester phosphodiesterase domain-containing protein 2|osteoblast differentiation promoting factor', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70423031, end_position_on_the_genomic_accession=70433391, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54858, CurrentID=0, Status='live', Symbol='PGPEP1', Aliases='PAP-I, PGP, PGP-I, PGPI, Pcp', description='pyroglutamyl-peptidase I', other_designations='5-oxoprolyl-peptidase|pyroglutamyl aminopeptidase I|pyrrolidone-carboxylate peptidase', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18340582, end_position_on_the_genomic_accession=18369953, orientation='plus', exon_count=7, OMIM=610694, no_hdr0=''),
  54859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54859, CurrentID=0, Status='live', Symbol='ELP6', Aliases='C3orf75, TMEM103', description='elongator acetyltransferase complex subunit 6', other_designations='UPF0405 protein C3orf75|angiotonin-transactivated protein 1|elongator complex protein 6 homolog|transmembrane protein 103', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47495640, end_position_on_the_genomic_accession=47513722, orientation='minus', exon_count=13, OMIM=615020, no_hdr0=''),
  54860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54860, CurrentID=0, Status='live', Symbol='MS4A12', Aliases='Ms4a10', description='membrane-spanning 4-domains, subfamily A, member 12', other_designations='', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60492742, end_position_on_the_genomic_accession=60507430, orientation='plus', exon_count=8, OMIM=606550, no_hdr0=''),
  642658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642658, CurrentID=0, Status='live', Symbol='SCX', Aliases='SCXAB, bHLHa48, SCX', description='scleraxis bHLH transcription factor', other_designations='class A basic helix-loop-helix protein 41|class A basic helix-loop-helix protein 48|class II bHLH protein scleraxis|scleraxis basic helix-loop-helix transcription factor|scleraxis homolog A|scleraxis homolog B', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144265060, end_position_on_the_genomic_accession=144268483, orientation='plus', exon_count=3, OMIM=609067, no_hdr0=''),
  54862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54862, CurrentID=0, Status='live', Symbol='CC2D1A', Aliases='FREUD-1, Freud-1/Aki1, MRT3', description='coiled-coil and C2 domain containing 1A', other_designations='Akt kinase-interacting protein 1|FRE under dual repression-binding protein 1|five prime repressor element under dual repression-binding protein 1|five repressor element under dual repression-binding protein 1|mental retardation, nonsyndromic, autosomal recessive, 3|putative NF-kappa-B-activating protein 023N|putative NFkB activating protein', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13906143, end_position_on_the_genomic_accession=13930880, orientation='plus', exon_count=29, OMIM=610055, no_hdr0=''),
  54863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54863, CurrentID=0, Status='live', Symbol='TOR4A', Aliases='C9orf167', description='torsin family 4, member A', other_designations='torsin family 4 member A|torsin family protein C9orf167', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137277828, end_position_on_the_genomic_accession=137282641, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54865, CurrentID=0, Status='live', Symbol='GPATCH4', Aliases='GPATC4', description='G-patch domain containing 4', other_designations='G patch domain containing 4', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156594308, end_position_on_the_genomic_accession=156601496, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54866, CurrentID=0, Status='live', Symbol='PPP1R14D', Aliases='CPI17-like, GBPI-1, GBPI1', description='protein phosphatase 1 regulatory inhibitor subunit 14D', other_designations='PKC-dependent PP1 inhibitory protein|gastrointestinal and brain-specific PP1-inhibitory protein 1|gut and brain phosphatase inhibitor 1|protein phosphatase 1, regulatory (inhibitor) subunit 14D', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40815445, end_position_on_the_genomic_accession=40828709, orientation='minus', exon_count=5, OMIM=613256, no_hdr0=''),
  54867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54867, CurrentID=0, Status='live', Symbol='TMEM214', Aliases='', description='transmembrane protein 214', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27032906, end_position_on_the_genomic_accession=27041697, orientation='plus', exon_count=18, OMIM=615301, no_hdr0=''),
  54868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54868, CurrentID=0, Status='live', Symbol='TMEM104', Aliases='', description='transmembrane protein 104', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74776483, end_position_on_the_genomic_accession=74839783, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54869, CurrentID=0, Status='live', Symbol='EPS8L1', Aliases='DRC3, EPS8R1, PP10566', description='EPS8 like 1', other_designations='EPS8-like 1|EPS8-like protein 1|EPS8-related protein 1|epidermal growth factor receptor pathway substrate 8-related protein 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55075853, end_position_on_the_genomic_accession=55087923, orientation='plus', exon_count=23, OMIM=614987, no_hdr0=''),
  54870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54870, CurrentID=0, Status='live', Symbol='QRICH1', Aliases='', description='glutamine-rich 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49029707, end_position_on_the_genomic_accession=49094071, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54872, CurrentID=0, Status='live', Symbol='PIGG', Aliases='GPI7, LAS21, PRO4405, RLGS1930', description='phosphatidylinositol glycan anchor biosynthesis class G', other_designations='phosphatidylinositol glycan anchor biosynthesis, class G|phosphatidylinositol-glycan biosynthesis class G protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=499200, end_position_on_the_genomic_accession=539921, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  54873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54873, CurrentID=0, Status='live', Symbol='PALMD', Aliases='C1orf11, PALML', description='palmdelphin', other_designations='paralemmin-like protein|paralemnin-like protein', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99645875, end_position_on_the_genomic_accession=99694541, orientation='plus', exon_count=8, OMIM=610182, no_hdr0=''),
  54874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54874, CurrentID=0, Status='live', Symbol='FNBP1L', Aliases='C1orf39, TOCA1', description='formin binding protein 1-like', other_designations='toca-1|transducer of Cdc42-dependent actin assembly 1|transducer of Cdc42-dependent actin assembly protein 1', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93448076, end_position_on_the_genomic_accession=93570905, orientation='plus', exon_count=20, OMIM=608848, no_hdr0=''),
  54875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54875, CurrentID=0, Status='live', Symbol='CNTLN', Aliases='C9orf101, C9orf39, bA340N12.1', description='centlein', other_designations='centlein, centrosomal protein|centrosomal protein', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17134990, end_position_on_the_genomic_accession=17503923, orientation='plus', exon_count=33, OMIM=611870, no_hdr0=''),
  54876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54876, CurrentID=0, Status='live', Symbol='DCAF16', Aliases='C4orf30', description='DDB1 and CUL4 associated factor 16', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17800655, end_position_on_the_genomic_accession=17810769, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54877, CurrentID=0, Status='live', Symbol='ZCCHC2', Aliases='C18orf49', description='zinc finger, CCHC domain containing 2', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62523410, end_position_on_the_genomic_accession=62587709, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54878, CurrentID=0, Status='live', Symbol='DPP8', Aliases='DP8, DPRP1, MST097, MSTP097, MSTP135, MSTP141', description='dipeptidyl-peptidase 8', other_designations='DPP VIII|DPRP-1|dipeptidyl peptidase IV-related protein 1|dipeptidyl peptidase IV-related protein-1|dipeptidyl peptidase VIII|dipeptidylpeptidase 8|prolyl dipeptidase DPP8', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65442467, end_position_on_the_genomic_accession=65517702, orientation='minus', exon_count=23, OMIM=606819, no_hdr0=''),
  54879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54879, CurrentID=0, Status='live', Symbol='ST7L', Aliases='FAM4B, ST7R, STLR', description='suppression of tumorigenicity 7 like', other_designations='ST7-related protein|tumor suppressor ST7 like', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112517804, end_position_on_the_genomic_accession=112620838, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  54880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54880, CurrentID=0, Status='live', Symbol='BCOR', Aliases='ANOP2, MAA2, MCOPS2', description='BCL6 corepressor', other_designations='BCL-6 interacting corepressor', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40051246, end_position_on_the_genomic_accession=40177390, orientation='minus', exon_count=19, OMIM=300485, no_hdr0=''),
  54881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54881, CurrentID=0, Status='live', Symbol='TEX10', Aliases='Ipi1, bA208F1.2', description='testis expressed 10', other_designations='testis expressed gene 10|testis expressed sequence 10', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100302075, end_position_on_the_genomic_accession=100352977, orientation='minus', exon_count=16, OMIM=616717, no_hdr0=''),
  54882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54882, CurrentID=0, Status='live', Symbol='ANKHD1', Aliases='MASK, MASK1, PP2500, VBARP', description='ankyrin repeat and KH domain containing 1', other_designations='HIV-1 Vpr-binding ankyrin repeat protein|multiple ankyrin repeats, single KH-domain homolog', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140401814, end_position_on_the_genomic_accession=140539856, orientation='plus', exon_count=35, OMIM=610500, no_hdr0=''),
  54883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54883, CurrentID=0, Status='live', Symbol='CWC25', Aliases='CCDC49', description='CWC25 spliceosome-associated protein homolog', other_designations='coiled-coil domain-containing protein 49', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38800434, end_position_on_the_genomic_accession=38825350, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54884, CurrentID=0, Status='live', Symbol='RETSAT', Aliases='', description='retinol saturase (all-trans-retinol 13,14-reductase)', other_designations='PPAR-alpha-regulated and starvation-induced gene protein|all-trans-13,14-dihydroretinol saturase', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85341955, end_position_on_the_genomic_accession=85354698, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  54885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54885, CurrentID=0, Status='live', Symbol='TBC1D8B', Aliases='', description='TBC1 domain family member 8B', other_designations='RP11-321G1.1|TBC1 domain family, member 8B (with GRAM domain)', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106802686, end_position_on_the_genomic_accession=106876150, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  54886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54886, CurrentID=0, Status='live', Symbol='PLPPR1', Aliases='LPPR1, PRG-3', description='phospholipid phosphatase related 1', other_designations='plasticity related gene 3|plasticity-related gene 3 protein', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101028749, end_position_on_the_genomic_accession=101325135, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54887, CurrentID=0, Status='live', Symbol='UHRF1BP1', Aliases='C6orf107, ICBP90, dJ349A12.1', description='UHRF1 binding protein 1', other_designations='ICBP90 binding protein 1|ICBP90-binding protein 1|UHRF1 (ICBP90) binding protein 1|ubiquitin-like containing PHD and RING finger domains 1-binding protein 1', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34792017, end_position_on_the_genomic_accession=34877514, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  54888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54888, CurrentID=0, Status='live', Symbol='NSUN2', Aliases='MISU, MRT5, SAKI, TRM4', description='NOP2/Sun RNA methyltransferase family member 2', other_designations='5-methycytoisine methyltransferase|Myc-induced SUN-domain-containing protein|NOL1/NOP2/Sun domain family, member 2|NOP2/Sun RNA methyltransferase family, member 2|NOP2/Sun domain family, member 2|myc-induced SUN domain-containing protein|substrate of AIM1/Aurora kinase B|tRNA (cytosine-5-)-methyltransferase NSUN2|tRNA methyltransferase 4 homolog', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6599239, end_position_on_the_genomic_accession=6633360, orientation='minus', exon_count=19, OMIM=610916, no_hdr0=''),
  54890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54890, CurrentID=0, Status='live', Symbol='ALKBH5', Aliases='ABH5, OFOXD, OFOXD1', description='alkB homolog 5, RNA demethylase', other_designations='AlkB family member 5, RNA demethylase|alkB, alkylation repair homolog 5|alkylated DNA repair protein alkB homolog 5|alpha-ketoglutarate-dependent dioxygenase alkB homolog 5|oxoglutarate and iron-dependent oxygenase domain containing|probable alpha-ketoglutarate-dependent dioxygenase ABH5', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18183553, end_position_on_the_genomic_accession=18209954, orientation='plus', exon_count=4, OMIM=613303, no_hdr0=''),
  54891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54891, CurrentID=0, Status='live', Symbol='INO80D', Aliases='', description='INO80 complex subunit D', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=205993721, end_position_on_the_genomic_accession=206086182, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  54892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54892, CurrentID=0, Status='live', Symbol='NCAPG2', Aliases='CAP-G2, CAPG2, LUZP5, MTB, hCAP-G2', description='non-SMC condensin II complex subunit G2', other_designations='chromosome-associated protein G2|leucine zipper protein 5|more than blood homolog|non-SMC condensin II complex, subunit G2', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158631169, end_position_on_the_genomic_accession=158704831, orientation='minus', exon_count=35, OMIM=608532, no_hdr0=''),
  54893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54893, CurrentID=0, Status='live', Symbol='MTMR10', Aliases='', description='myotubularin related protein 10', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30919764, end_position_on_the_genomic_accession=30991604, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  54894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54894, CurrentID=0, Status='live', Symbol='RNF43', Aliases='RNF124, URCC', description='ring finger protein 43', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58353676, end_position_on_the_genomic_accession=58417620, orientation='minus', exon_count=10, OMIM=612482, no_hdr0=''),
  54896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54896, CurrentID=0, Status='live', Symbol='PQLC2', Aliases='', description='PQ loop repeat containing 2', other_designations='PQ-loop repeat-containing protein 2', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19312246, end_position_on_the_genomic_accession=19332431, orientation='plus', exon_count=9, OMIM=614760, no_hdr0=''),
  54897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54897, CurrentID=0, Status='live', Symbol='CASZ1', Aliases='CAS11, CST, SRG, ZNF693, dJ734G22.1', description='castor zinc finger 1', other_designations='castor homolog 1, zinc finger|castor-related protein|survival-related|zinc finger protein 693', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10636604, end_position_on_the_genomic_accession=10796676, orientation='minus', exon_count=23, OMIM=609895, no_hdr0=''),
  54898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54898, CurrentID=0, Status='live', Symbol='ELOVL2', Aliases='SSC2', description='ELOVL fatty acid elongase 2', other_designations='3-keto acyl-CoA synthase ELOVL2|ELOVL FA elongase 2|elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 2|very-long-chain 3-oxoacyl-CoA synthase 2', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10980759, end_position_on_the_genomic_accession=11044391, orientation='minus', exon_count=10, OMIM=611814, no_hdr0=''),
  54899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54899, CurrentID=0, Status='live', Symbol='PXK', Aliases='MONaKA', description='PX domain containing serine/threonine kinase', other_designations='PX ser/thr kinase v2|PX serine/threonine kinase|modulator of Na,K-ATPase long form', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58332890, end_position_on_the_genomic_accession=58426127, orientation='plus', exon_count=22, OMIM=611450, no_hdr0=''),
  54900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54900, CurrentID=0, Status='live', Symbol='LAX1', Aliases='LAX', description='lymphocyte transmembrane adaptor 1', other_designations='LAT-like membrane associated protein|linker for activation of x cells|membrane-associated adapter protein LAX', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203765146, end_position_on_the_genomic_accession=203776372, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54901, CurrentID=0, Status='live', Symbol='CDKAL1', Aliases='', description='CDK5 regulatory subunit associated protein 1-like 1', other_designations='tRNA-t(6)A37 methylthiotransferase', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20534457, end_position_on_the_genomic_accession=21232404, orientation='plus', exon_count=19, OMIM=611259, no_hdr0=''),
  54902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54902, CurrentID=0, Status='live', Symbol='TTC19', Aliases='2010204O13Rik, MC3DN2', description='tetratricopeptide repeat domain 19', other_designations='TPR repeat protein 19', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15999380, end_position_on_the_genomic_accession=16029409, orientation='plus', exon_count=11, OMIM=613814, no_hdr0=''),
  54903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54903, CurrentID=0, Status='live', Symbol='MKS1', Aliases='BBS13, MES, MKS, POC12', description='Meckel syndrome, type 1', other_designations='FABB proteome-like protein|Meckel syndrome type 1|POC12 centriolar protein homolog', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58205436, end_position_on_the_genomic_accession=58219605, orientation='minus', exon_count=20, OMIM=609883, no_hdr0=''),
  54904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54904, CurrentID=0, Status='live', Symbol='WHSC1L1', Aliases='NSD3, pp14328', description='Wolf-Hirschhorn syndrome candidate 1-like 1', other_designations='WHSC1-like 1 isoform 9 with methyltransferase activity to lysine|nuclear SET domain-containing protein 3|protein whistle', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38269699, end_position_on_the_genomic_accession=38382354, orientation='minus', exon_count=26, OMIM=607083, no_hdr0=''),
  54905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54905, CurrentID=0, Status='live', Symbol='CYP2W1', Aliases='', description='cytochrome P450 family 2 subfamily W member 1', other_designations='CYPIIW1|cytochrome P450, family 2, subfamily W, polypeptide 1', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=983156, end_position_on_the_genomic_accession=989640, orientation='plus', exon_count=10, OMIM=615967, no_hdr0=''),
  54906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54906, CurrentID=0, Status='live', Symbol='FAM208B', Aliases='C10orf18, bA318E3.2', description='family with sequence similarity 208 member B', other_designations='family with sequence similarity 208, member B', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5684772, end_position_on_the_genomic_accession=5765779, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  54908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54908, CurrentID=0, Status='live', Symbol='SPDL1', Aliases='CCDC99', description='spindle apparatus coiled-coil protein 1', other_designations='arsenite-related gene 1 protein|coiled-coil domain containing 99|coiled-coil domain-containing protein 99|hSpindly|rhabdomyosarcoma antigen MU-RMS-40.4A|rrhabdomyosarcoma antigen protein MU-RMS-40.4A|spindle apparatus coiled-coil domain-containing protein 1|spindly homolog', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169583634, end_position_on_the_genomic_accession=169604778, orientation='plus', exon_count=15, OMIM=616401, no_hdr0=''),
  105369194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369194, CurrentID=0, Status='live', Symbol='LOC105369194', Aliases='', description='uncharacterized LOC105369194', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123692834, end_position_on_the_genomic_accession=123717192, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  54910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54910, CurrentID=0, Status='live', Symbol='SEMA4C', Aliases='M-SEMA-F, SEMACL1, SEMAF, SEMAI', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4C', other_designations='M-Sema F', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96859736, end_position_on_the_genomic_accession=96870943, orientation='minus', exon_count=19, OMIM=604462, no_hdr0=''),
  54913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54913, CurrentID=0, Status='live', Symbol='RPP25', Aliases='', description='ribonuclease P/MRP 25kDa subunit', other_designations='RNase P protein subunit p25|ribonuclease P 25kDa subunit', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74955102, end_position_on_the_genomic_accession=74957434, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54914, CurrentID=0, Status='live', Symbol='FOCAD', Aliases='KIAA1797', description='focadhesin', other_designations='', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20658309, end_position_on_the_genomic_accession=20995955, orientation='plus', exon_count=49, OMIM=614606, no_hdr0=''),
  54915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54915, CurrentID=0, Status='live', Symbol='YTHDF1', Aliases='C20orf21', description='YTH N(6)-methyladenosine RNA binding protein 1', other_designations='DACA-1|YTH domain family 1|YTH domain family protein 1|YTH domain family, member 1|dermatomyositis associated with cancer putative autoantigen 1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63195429, end_position_on_the_genomic_accession=63216186, orientation='minus', exon_count=5, OMIM=616529, no_hdr0=''),
  54916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54916, CurrentID=0, Status='live', Symbol='TMEM260', Aliases='C14orf101', description='transmembrane protein 260', other_designations='UPF0679 protein C14orf101', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56579791, end_position_on_the_genomic_accession=56651884, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  54918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54918, CurrentID=0, Status='live', Symbol='CMTM6', Aliases='CKLFSF6, PRO2219', description='CKLF-like MARVEL transmembrane domain containing 6', other_designations='chemokine-like factor super family 6|chemokine-like factor superfamily 6|chemokine-like factor superfamily member 6', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32481312, end_position_on_the_genomic_accession=32502911, orientation='minus', exon_count=4, OMIM=607889, no_hdr0=''),
  54919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54919, CurrentID=0, Status='live', Symbol='DNAAF5', Aliases='CILD18, HEATR2', description='dynein, axonemal, assembly factor 5', other_designations='HEAT repeat containing 2|HEAT repeat-containing protein 2', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=726701, end_position_on_the_genomic_accession=786479, orientation='plus', exon_count=13, OMIM=614864, no_hdr0=''),
  54920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54920, CurrentID=0, Status='live', Symbol='DUS2', Aliases='DUS2L, SMM1, URLC8', description='dihydrouridine synthase 2', other_designations='SMM1 homolog|dihydrouridine synthase 2-like, SMM1 homolog|up-regulated in lung cancer protein 8', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68022944, end_position_on_the_genomic_accession=68079323, orientation='plus', exon_count=17, OMIM=609707, no_hdr0=''),
  54921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54921, CurrentID=0, Status='live', Symbol='CHTF8', Aliases='CTF8, DERPC', description='chromosome transmission fidelity factor 8', other_designations='CTF8, chromosome transmission fidelity factor 8 homolog|decreased expression in renal and prostate cancer protein', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69118009, end_position_on_the_genomic_accession=69132590, orientation='minus', exon_count=5, OMIM=613202, no_hdr0=''),
  54922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54922, CurrentID=0, Status='live', Symbol='RASIP1', Aliases='RAIN', description='Ras interacting protein 1', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48720585, end_position_on_the_genomic_accession=48740813, orientation='minus', exon_count=13, OMIM=609623, no_hdr0=''),
  54923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54923, CurrentID=0, Status='live', Symbol='LIME1', Aliases='LIME, dJ583P15.4', description='Lck interacting transmembrane adaptor 1', other_designations='lck-interacting membrane protein|lck-interacting molecule', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63736626, end_position_on_the_genomic_accession=63739107, orientation='plus', exon_count=6, OMIM=609809, no_hdr0=''),
  54925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54925, CurrentID=0, Status='live', Symbol='ZSCAN32', Aliases='HCCS-5, ZNF434', description='zinc finger and SCAN domain containing 32', other_designations='human cervical cancer suppressor gene 5 protein|zinc finger protein 434', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3382081, end_position_on_the_genomic_accession=3401065, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54926, CurrentID=0, Status='live', Symbol='UBE2R2', Aliases='CDC34B, E2-CDC34B, UBC3B', description='ubiquitin conjugating enzyme E2R 2', other_designations='E2 ubiquitin-conjugating enzyme R2|ubiquitin carrier protein R2|ubiquitin-conjugating enzyme E2-CDC34B|ubiquitin-conjugating enzyme E2R 2|ubiquitin-conjugating enzyme UBC3B|ubiquitin-protein ligase R2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33817164, end_position_on_the_genomic_accession=33920403, orientation='plus', exon_count=5, OMIM=612506, no_hdr0=''),
  54927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54927, CurrentID=0, Status='live', Symbol='CHCHD3', Aliases='MINOS3, Mic19, PPP1R22', description='coiled-coil-helix-coiled-coil-helix domain containing 3', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 3, mitochondrial|mitochondrial inner membrane organizing system 3|protein phosphatase 1, regulatory subunit 22', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132784862, end_position_on_the_genomic_accession=133082079, orientation='minus', exon_count=11, OMIM=613748, no_hdr0=''),
  54928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54928, CurrentID=0, Status='live', Symbol='IMPAD1', Aliases='GPAPP, IMP 3, IMP-3, IMPA3', description='inositol monophosphatase domain containing 1', other_designations='Golgi 3-prime phosphoadenosine 5-prime phosphate 3-prime phosphatase|IMPase 3|golgi-resident PAP phosphatase|golgi-resident nucleotide phosphatase|inositol monophosphatase domain-containing protein 1|inositol-1(or 4)-monophosphatase 3|myo-inositol monophosphatase A3', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56957929, end_position_on_the_genomic_accession=56993871, orientation='minus', exon_count=6, OMIM=614010, no_hdr0=''),
  54929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54929, CurrentID=0, Status='live', Symbol='TMEM161A', Aliases='AROS-29', description='transmembrane protein 161A', other_designations='adaptive response to oxidative stress protein 29', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19119616, end_position_on_the_genomic_accession=19138554, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54930, CurrentID=0, Status='live', Symbol='HAUS4', Aliases='C14orf94', description='HAUS augmin like complex subunit 4', other_designations='HAUS augmin-like complex, subunit 4|homologous to Augmin subunits', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22946228, end_position_on_the_genomic_accession=22957142, orientation='minus', exon_count=10, OMIM=613431, no_hdr0=''),
  54931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54931, CurrentID=0, Status='live', Symbol='TRMT10C', Aliases='HNYA, MRPP1, RG9MTD1', description='tRNA methyltransferase 10C, mitochondrial RNase P subunit', other_designations='HBV pre-S2 trans-regulated protein 2|RNA (guanine-9-) methyltransferase domain containing 1|mitochondrial RNase P subunit 1|renal carcinoma antigen NY-REN-49|tRNA methyltransferase 10 homolog C', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101561836, end_position_on_the_genomic_accession=101566446, orientation='plus', exon_count=2, OMIM=615423, no_hdr0=''),
  54932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54932, CurrentID=0, Status='live', Symbol='EXD3', Aliases='mut-7', description="exonuclease 3'-5' domain containing 3", other_designations="exonuclease 3'-5' domain-containing protein 3|probable exonuclease mut-7 homolog", map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137306894, end_position_on_the_genomic_accession=137423268, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  54933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54933, CurrentID=0, Status='live', Symbol='RHBDL2', Aliases='RRP2', description='rhomboid, veinlet-like 2 (Drosophila)', other_designations='rhomboid (veinlet, Drosophila)-like 2|rhomboid protease 2|rhomboid-like protein 2', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38885806, end_position_on_the_genomic_accession=38941830, orientation='minus', exon_count=9, OMIM=608962, no_hdr0=''),
  54934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54934, CurrentID=0, Status='live', Symbol='KANSL2', Aliases='C12orf41, NSL2', description='KAT8 regulatory NSL complex subunit 2', other_designations='NSL complex protein NSL2|non-specific lethal 2 homolog', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48653211, end_position_on_the_genomic_accession=48682252, orientation='minus', exon_count=10, OMIM=615488, no_hdr0=''),
  54935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54935, CurrentID=0, Status='live', Symbol='DUSP23', Aliases='DUSP25, LDP-3, MOSP, VHZ', description='dual specificity phosphatase 23', other_designations='VH1-like member Z|VH1-like phosphatase Z|low molecular mass dual specificity phosphatase 3|low-molecular-mass dual-specificity phosphatase 3', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159780940, end_position_on_the_genomic_accession=159782543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54936, CurrentID=0, Status='live', Symbol='ADPRHL2', Aliases='ARH3', description='ADP-ribosylhydrolase like 2', other_designations='ADP-ribosylhydrolase 3|[Protein ADP-ribosylarginine] hydrolase-like protein 2|protein ADP-ribosylarginine hydrolase-like protein 2', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36088852, end_position_on_the_genomic_accession=36093932, orientation='plus', exon_count=6, OMIM=610624, no_hdr0=''),
  54937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54937, CurrentID=0, Status='live', Symbol='SOHLH2', Aliases='SPATA28, TEB1, bHLHe81', description='spermatogenesis and oogenesis specific basic helix-loop-helix 2', other_designations='spermatogenesis associated 28', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36168208, end_position_on_the_genomic_accession=36214615, orientation='minus', exon_count=12, OMIM=616066, no_hdr0=''),
  54938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54938, CurrentID=0, Status='live', Symbol='SARS2', Aliases='SARS, SARSM, SERS, SYS, SerRS, SerRSmt, mtSerRS', description='seryl-tRNA synthetase 2, mitochondrial', other_designations='mitochondrial seryl-tRNA synthetase|serine tRNA ligase 2, mitochondrial|serine-tRNA ligase, mitochondrial|seryl-tRNA synthetase, mitochondrial|seryl-tRNA(Ser/Sec) synthetase', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38915264, end_position_on_the_genomic_accession=38930896, orientation='minus', exon_count=17, OMIM=612804, no_hdr0=''),
  54939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54939, CurrentID=0, Status='live', Symbol='COMMD4', Aliases='', description='COMM domain containing 4', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75335996, end_position_on_the_genomic_accession=75341927, orientation='plus', exon_count=10, OMIM=616701, no_hdr0=''),
  54940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54940, CurrentID=0, Status='live', Symbol='OCIAD1', Aliases='ASRIJ, OCIA, TPA018', description='OCIA domain containing 1', other_designations='ovarian carcinoma immunoreactive antigen', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48831043, end_position_on_the_genomic_accession=48861817, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54941, CurrentID=0, Status='live', Symbol='RNF125', Aliases='TNORS, TRAC-1, TRAC1', description='ring finger protein 125, E3 ubiquitin protein ligase', other_designations='T-cell RING activation protein 1|T-cell ring protein identified in activation screen', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32018223, end_position_on_the_genomic_accession=32088528, orientation='plus', exon_count=7, OMIM=610432, no_hdr0=''),
  54942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54942, CurrentID=0, Status='live', Symbol='FAM206A', Aliases='C9orf6, CG-8, Simiate', description='family with sequence similarity 206 member A', other_designations='family with sequence similarity 206, member A|protein FAM206A', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108934393, end_position_on_the_genomic_accession=108951041, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54943, CurrentID=0, Status='live', Symbol='DNAJC28', Aliases='C21orf55, C21orf78', description='DnaJ heat shock protein family (Hsp40) member C28', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 28', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33488055, end_position_on_the_genomic_accession=33491747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54944, CurrentID=0, Status='live', Symbol='LINC01521', Aliases='', description='long intergenic non-protein coding RNA 1521', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31346759, end_position_on_the_genomic_accession=31351155, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369200, CurrentID=0, Status='live', Symbol='LOC105369200', Aliases='', description='uncharacterized LOC105369200', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87930514, end_position_on_the_genomic_accession=87939304, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54946, CurrentID=0, Status='live', Symbol='SLC41A3', Aliases='SLC41A1-L2', description='solute carrier family 41 member 3', other_designations='SLC41A1-like 2|solute carrier family 41, member 3', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126006357, end_position_on_the_genomic_accession=126101550, orientation='minus', exon_count=17, OMIM=610803, no_hdr0=''),
  54947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54947, CurrentID=0, Status='live', Symbol='LPCAT2', Aliases='AYTL1, LysoPAFAT', description='lysophosphatidylcholine acyltransferase 2', other_designations='1-AGP acyltransferase 11|1-AGPAT 11|1-acylglycerol-3-phosphate O-acyltransferase 11|1-acylglycerophosphocholine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|LPAAT-alpha|LPC acyltransferase 2|LPCAT-2|acetyl-CoA:lyso-PAF acetyltransferase|acetyl-CoA:lyso-platelet-activating factor acetyltransferase|acyl-CoA:lysophosphatidylcholine acyltransferase 2|acyltransferase like 1|acyltransferase-like 1|lyso-PAF acetyltransferase|lyso-platelet-activating factor (PAF) acetyltransferase|lysoPC acyltransferase 2|lysophosphatidic acid acyltransferase alpha', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55509001, end_position_on_the_genomic_accession=55586670, orientation='plus', exon_count=17, OMIM=612040, no_hdr0=''),
  54948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54948, CurrentID=0, Status='live', Symbol='MRPL16', Aliases='L16mt, MRP-L16, PNAS-111', description='mitochondrial ribosomal protein L16', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59806135, end_position_on_the_genomic_accession=59810872, orientation='minus', exon_count=5, OMIM=611829, no_hdr0=''),
  54949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54949, CurrentID=0, Status='live', Symbol='SDHAF2', Aliases='C11orf79, PGL2, SDH5', description='succinate dehydrogenase complex assembly factor 2', other_designations='SDH assembly factor 2|hSDH5|succinate dehydrogenase subunit 5, mitochondrial', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61430125, end_position_on_the_genomic_accession=61446767, orientation='plus', exon_count=4, OMIM=613019, no_hdr0=''),
  54951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54951, CurrentID=0, Status='live', Symbol='COMMD8', Aliases='', description='COMM domain containing 8', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47450793, end_position_on_the_genomic_accession=47463665, orientation='minus', exon_count=5, OMIM=616656, no_hdr0=''),
  54952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54952, CurrentID=0, Status='live', Symbol='TRNAU1AP', Aliases='PRO1902, SECP43, TRSPAP1', description='tRNA selenocysteine 1 associated protein 1', other_designations='tRNA selenocysteine associated protein (SECP43)|tRNA selenocysteine associated protein 1|tRNA selenocysteine-associated protein 1', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28553017, end_position_on_the_genomic_accession=28578545, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54953, CurrentID=0, Status='live', Symbol='C1orf27', Aliases='ODR4, TTG1, odr-4', description='chromosome 1 open reading frame 27', other_designations='LAG1-interacting protein|odorant response abnormal 4|transactivated by recombinant transforming growth factor beta|transactivated by transforming growth factor beta protein 1', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186375758, end_position_on_the_genomic_accession=186421375, orientation='plus', exon_count=15, OMIM=609335, no_hdr0=''),
  54954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54954, CurrentID=0, Status='live', Symbol='FAM120C', Aliases='CXorf17, ORF34', description='family with sequence similarity 120C', other_designations='tumor antigen BJ-HCC-21', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54068324, end_position_on_the_genomic_accession=54183281, orientation='minus', exon_count=18, OMIM=300741, no_hdr0=''),
  54955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54955, CurrentID=0, Status='live', Symbol='C1orf109', Aliases='', description='chromosome 1 open reading frame 109', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37681570, end_position_on_the_genomic_accession=37692336, orientation='minus', exon_count=7, OMIM=614799, no_hdr0=''),
  54956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54956, CurrentID=0, Status='live', Symbol='PARP16', Aliases='ARTD15, C15orf30, pART15', description='poly(ADP-ribose) polymerase family member 16', other_designations='ADP-ribosyltransferase diphtheria toxin-like 15|PARP-16|poly (ADP-ribose) polymerase family, member 16|poly [ADP-ribose] polymerase 16', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65234464, end_position_on_the_genomic_accession=65286884, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54957, CurrentID=0, Status='live', Symbol='TXNL4B', Aliases='DLP, Dim2', description='thioredoxin like 4B', other_designations='Dim1-like protein|thioredoxin-like 4B', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72084857, end_position_on_the_genomic_accession=72094316, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54958, CurrentID=0, Status='live', Symbol='TMEM160', Aliases='', description='transmembrane protein 160', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47045909, end_position_on_the_genomic_accession=47048784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54959, CurrentID=0, Status='live', Symbol='ODAM', Aliases='APIN', description='odontogenic, ameloblast asssociated', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70195726, end_position_on_the_genomic_accession=70204576, orientation='plus', exon_count=12, OMIM=614843, no_hdr0=''),
  54960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54960, CurrentID=0, Status='live', Symbol='GEMIN8', Aliases='FAM51A1', description='gem nuclear organelle associated protein 8', other_designations='family with sequence similarity 51, member A1|gem (nuclear organelle) associated protein 8|gemin-8', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14002124, end_position_on_the_genomic_accession=14029916, orientation='minus', exon_count=8, OMIM=300962, no_hdr0=''),
  54961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54961, CurrentID=0, Status='live', Symbol='SSH3', Aliases='SSH3L', description='slingshot protein phosphatase 3', other_designations='SSH-3L|SSH-like protein 3|hSSH-3L|slingshot 3|slingshot homolog 3', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67303448, end_position_on_the_genomic_accession=67312607, orientation='plus', exon_count=14, OMIM=606780, no_hdr0=''),
  54962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54962, CurrentID=0, Status='live', Symbol='TIPIN', Aliases='', description='TIMELESS interacting protein', other_designations='CSM3 homolog', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66336191, end_position_on_the_genomic_accession=66386737, orientation='minus', exon_count=12, OMIM=610716, no_hdr0=''),
  54963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54963, CurrentID=0, Status='live', Symbol='UCKL1', Aliases='UCK1L, URKL1', description='uridine-cytidine kinase 1-like 1', other_designations='UCK1-LIKE', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63939829, end_position_on_the_genomic_accession=63959435, orientation='minus', exon_count=18, OMIM=610866, no_hdr0=''),
  54964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54964, CurrentID=0, Status='live', Symbol='C1orf56', Aliases='MENT', description='chromosome 1 open reading frame 56', other_designations='methylated in normal thymocytes protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151047783, end_position_on_the_genomic_accession=151051420, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54965, CurrentID=0, Status='live', Symbol='PIGX', Aliases='PIG-X', description='phosphatidylinositol glycan anchor biosynthesis class X', other_designations='GPI-mannosyltransferase subunit|phosphatidylinositol glycan anchor biosynthesis, class X', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196712374, end_position_on_the_genomic_accession=196736007, orientation='plus', exon_count=7, OMIM=610276, no_hdr0=''),
  54967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54967, CurrentID=0, Status='live', Symbol='CT55', Aliases='BJHCC20A, CXorf48', description='cancer/testis antigen 55', other_designations='BRCA2-interacting protein|tumor antigen BJ-HCC-20', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135156536, end_position_on_the_genomic_accession=135171827, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54968, CurrentID=0, Status='live', Symbol='TMEM70', Aliases='MC5DN2', description='transmembrane protein 70', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73976142, end_position_on_the_genomic_accession=73982783, orientation='plus', exon_count=4, OMIM=612418, no_hdr0=''),
  54969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54969, CurrentID=0, Status='live', Symbol='C4orf27', Aliases='', description='chromosome 4 open reading frame 27', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169729468, end_position_on_the_genomic_accession=169757942, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54970, CurrentID=0, Status='live', Symbol='TTC12', Aliases='TPARM', description='tetratricopeptide repeat domain 12', other_designations='TPR repeat protein 12|tetratricopeptide repeat protein 12 variant 1|tetratricopeptide repeat protein 12 variant 2', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113314529, end_position_on_the_genomic_accession=113385952, orientation='plus', exon_count=30, OMIM=610732, no_hdr0=''),
  54971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54971, CurrentID=0, Status='live', Symbol='BANP', Aliases='BEND1, SMAR1, SMARBP1', description='BTG3 associated nuclear protein', other_designations='BEN domain-containing protein 1|scaffold/matrix-associated region-1-binding protein', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87949172, end_position_on_the_genomic_accession=88077318, orientation='plus', exon_count=28, OMIM=611564, no_hdr0=''),
  54972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54972, CurrentID=0, Status='live', Symbol='TMEM132A', Aliases='GBP, HSPA5BP1', description='transmembrane protein 132A', other_designations='GRP78-binding protein|HSPA5-binding protein 1|glucose-regulated protein, 78kDa|heat shock 70kDa protein 5 (glucose-regulated protein, 78kDa) binding protein 1|heat shock 70kDa protein 5 binding protein 1', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60924441, end_position_on_the_genomic_accession=60937159, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54973, CurrentID=0, Status='live', Symbol='CPSF3L', Aliases='CPSF73L, INT11, INTS11, RC-68, RC68', description='cleavage and polyadenylation specific factor 3-like', other_designations='CPSF3-like protein|cleavage and polyadenylation-specific factor 3-like protein|protein related to CPSF subunits of 68 kDa|related to CPSF subunits 68 kDa', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1311585, end_position_on_the_genomic_accession=1324687, orientation='minus', exon_count=19, OMIM=611354, no_hdr0=''),
  54974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54974, CurrentID=0, Status='live', Symbol='THG1L', Aliases='ICF45, IHG-1', description='tRNA-histidine guanylyltransferase 1-like', other_designations='induced by high glucose-1|interphase cytoplasmic foci protein 45', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157731315, end_position_on_the_genomic_accession=157739818, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105369205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369205, CurrentID=0, Status='live', Symbol='LOC105369205', Aliases='', description='uncharacterized LOC105369205', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20921024, end_position_on_the_genomic_accession=20999752, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54976, CurrentID=0, Status='live', Symbol='C20orf27', Aliases='', description='chromosome 20 open reading frame 27', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3753499, end_position_on_the_genomic_accession=3768388, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54977, CurrentID=0, Status='live', Symbol='SLC25A38', Aliases='', description='solute carrier family 25 member 38', other_designations='appoptosin|solute carrier family 25, member 38', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39383324, end_position_on_the_genomic_accession=39397328, orientation='plus', exon_count=10, OMIM=610819, no_hdr0=''),
  54978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54978, CurrentID=0, Status='live', Symbol='SLC35F6', Aliases='ANT2BP, C2orf18, TANGO9', description='solute carrier family 35 member F6', other_designations='ANT2-binding protein|solute carrier family 35, member F6|transmembrane protein C2orf18|transport and golgi organization 9 homolog', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26764274, end_position_on_the_genomic_accession=26781231, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54979, CurrentID=0, Status='live', Symbol='HRASLS2', Aliases='PLA1/2-2', description='HRAS-like suppressor 2', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63552759, end_position_on_the_genomic_accession=63565089, orientation='minus', exon_count=6, OMIM=613866, no_hdr0=''),
  54980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54980, CurrentID=0, Status='live', Symbol='C2orf42', Aliases='', description='chromosome 2 open reading frame 42', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70149885, end_position_on_the_genomic_accession=70191019, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54981, CurrentID=0, Status='live', Symbol='NMRK1', Aliases='C9orf95, NRK1, bA235O14.2', description='nicotinamide riboside kinase 1', other_designations='NRK 1|RNK 1|nicotinic acid riboside kinase 1|nmR-K 1|ribosylnicotinamide kinase 1|ribosylnicotinic acid kinase 1', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75060577, end_position_on_the_genomic_accession=75088319, orientation='minus', exon_count=12, OMIM=608704, no_hdr0=''),
  54982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54982, CurrentID=0, Status='live', Symbol='CLN6', Aliases='CLN4A, HsT18960, nclf', description='ceroid-lipofuscinosis, neuronal 6, late infantile, variant', other_designations='ceroid-lipofuscinosis neuronal 6 late infantile', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68206992, end_position_on_the_genomic_accession=68229742, orientation='minus', exon_count=7, OMIM=606725, no_hdr0=''),
  54984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54984, CurrentID=0, Status='live', Symbol='PINX1', Aliases='LPTL, LPTS', description='PIN2/TERF1 interacting, telomerase inhibitor 1', other_designations='67-11-3 protein|PIN2 interacting protein 1|PIN2-interacting protein 1|TRF1-interacting protein 1|hepatocellular carcinoma-related putative tumor suppressor|liver-related putative tumor suppressor|pin2-interacting protein X1|protein 67-11-3', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10764961, end_position_on_the_genomic_accession=10839899, orientation='minus', exon_count=7, OMIM=606505, no_hdr0=''),
  54985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54985, CurrentID=0, Status='live', Symbol='HCFC1R1', Aliases='HPIP', description='host cell factor C1 regulator 1 (XPO1 dependent)', other_designations='HCF-1 beta-propeller-interacting protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3022620, end_position_on_the_genomic_accession=3024296, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54986, CurrentID=0, Status='live', Symbol='ULK4', Aliases='FAM7C1, REC01035', description='unc-51 like kinase 4', other_designations='unc-51-like kinase 4', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41239987, end_position_on_the_genomic_accession=41962887, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  54987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54987, CurrentID=0, Status='live', Symbol='C1orf123', Aliases='', description='chromosome 1 open reading frame 123', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53214099, end_position_on_the_genomic_accession=53220639, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54988, CurrentID=0, Status='live', Symbol='ACSM5', Aliases='', description='acyl-CoA synthetase medium-chain family member 5', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20409520, end_position_on_the_genomic_accession=20440987, orientation='plus', exon_count=14, OMIM=614361, no_hdr0=''),
  54989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54989, CurrentID=0, Status='live', Symbol='ZNF770', Aliases='PRO1914', description='zinc finger protein 770', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34978341, end_position_on_the_genomic_accession=34988296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54991, CurrentID=0, Status='live', Symbol='C1orf159', Aliases='', description='chromosome 1 open reading frame 159', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1081818, end_position_on_the_genomic_accession=1116356, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  54993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54993, CurrentID=0, Status='live', Symbol='ZSCAN2', Aliases='ZFP29, ZNF854', description='zinc finger and SCAN domain containing 2', other_designations='zfp-29|zinc finger protein 29 homolog|zinc finger protein 854', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84600986, end_position_on_the_genomic_accession=84623718, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54994, CurrentID=0, Status='live', Symbol='GID8', Aliases='C20orf11, TWA1', description='GID complex subunit 8', other_designations='GID complex subunit 8 homolog|protein C20orf11|two hybrid-associated protein 1 with RanBPM', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62938089, end_position_on_the_genomic_accession=62948475, orientation='plus', exon_count=5, OMIM=611625, no_hdr0=''),
  54995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54995, CurrentID=0, Status='live', Symbol='OXSM', Aliases='FASN2D, KASI, KS', description='3-oxoacyl-ACP synthase, mitochondrial', other_designations='3-ketoacyl-acyl carrier protein synthase|3-oxoacyl-acyl-carrier-protein synthase|acyl-malonyl acyl carrier protein-condensing enzyme|beta-ketoacyl synthase|beta-ketoacyl-ACP synthase|type II mitochondrial beta-ketoacyl synthase', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25790072, end_position_on_the_genomic_accession=25794534, orientation='plus', exon_count=5, OMIM=610324, no_hdr0=''),
  54996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54996, CurrentID=0, Status='live', Symbol='MARC2', Aliases='MOSC2', description='mitochondrial amidoxime reducing component 2', other_designations='MOCO sulphurase C-terminal domain containing 2|MOSC domain-containing protein 2, mitochondrial|moco sulfurase C-terminal domain-containing protein 2|molybdenum cofactor sulfurase C-terminal domain-containing protein 2', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220747502, end_position_on_the_genomic_accession=220784815, orientation='plus', exon_count=11, OMIM=614127, no_hdr0=''),
  54997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54997, CurrentID=0, Status='live', Symbol='TESC', Aliases='CHP3, TSC', description='tescalcin', other_designations='calcineurin-like EF hand protein 3', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117038923, end_position_on_the_genomic_accession=117099446, orientation='minus', exon_count=9, OMIM=611585, no_hdr0=''),
  54998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54998, CurrentID=0, Status='live', Symbol='AURKAIP1', Aliases='AIP, AKIP, MRP-S38', description='aurora kinase A interacting protein 1', other_designations='28S ribosomal protein S38, mitochondrial|AURKA-interacting protein|aurora-A kinase interacting protein', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1373730, end_position_on_the_genomic_accession=1375438, orientation='minus', exon_count=4, OMIM=609183, no_hdr0=''),
  105369209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369209, CurrentID=0, Status='live', Symbol='LOC105369209', Aliases='', description='putative uncharacterized protein FLJ44790', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62227876, end_position_on_the_genomic_accession=62235050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55000, CurrentID=0, Status='live', Symbol='TUG1', Aliases='LINC00080, NCRNA00080, TI-227H', description='taurine up-regulated 1 (non-protein coding)', other_designations='TUG1 noncoding RNA|long intergenic non-protein coding RNA 80|taurine upregulated gene 1', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30969211, end_position_on_the_genomic_accession=30979395, orientation='plus', exon_count=4, OMIM=614971, no_hdr0=''),
  55001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55001, CurrentID=0, Status='live', Symbol='TTC22', Aliases='', description='tetratricopeptide repeat domain 22', other_designations='TPR repeat protein 22', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54779712, end_position_on_the_genomic_accession=54801717, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55002, CurrentID=0, Status='live', Symbol='TMCO3', Aliases='C13orf11', description='transmembrane and coiled-coil domains 3', other_designations='B230339H12Rik|putative LAG1-interacting protein', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113490993, end_position_on_the_genomic_accession=113550229, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  55003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55003, CurrentID=0, Status='live', Symbol='PAK1IP1', Aliases='MAK11, PIP1, WDR84, bA421M1.5, hPIP1', description='PAK1 interacting protein 1', other_designations='PAK/PLC-interacting protein 1|PAK1-interacting protein 1|WD repeat-containing protein 84', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10690865, end_position_on_the_genomic_accession=10709782, orientation='plus', exon_count=13, OMIM=607811, no_hdr0=''),
  55004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55004, CurrentID=0, Status='live', Symbol='LAMTOR1', Aliases='C11orf59, PDRO, Ragulator1, p18, p27RF-Rho', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 1', other_designations='RhoA activator C11orf59|late endosomal/lysosomal adaptor and MAPK and MTOR activator 1|lipid raft adaptor protein p18|p27Kip1-releasing factor from RhoA|p27kip1 releasing factor from RhoA|protein associated with DRMs and endosomes|ragulator complex protein PDRO', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72097292, end_position_on_the_genomic_accession=72103387, orientation='minus', exon_count=5, OMIM=613510, no_hdr0=''),
  55005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55005, CurrentID=0, Status='live', Symbol='RMND1', Aliases='C6orf96, COXPD11, RMD1, bA351K16, bA351K16.3', description='required for meiotic nuclear division 1 homolog', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151404761, end_position_on_the_genomic_accession=151452181, orientation='minus', exon_count=13, OMIM=614917, no_hdr0=''),
  55006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55006, CurrentID=0, Status='live', Symbol='TRMT61B', Aliases='', description='tRNA methyltransferase 61B', other_designations='potential tRNA (adenine(58)-N(1))-methyltransferase catalytic subunit TRMT61B|potential tRNA (adenine-N(1)-)-methyltransferase catalytic subunit TRMT61B|tRNA methyltransferase 61 homolog B', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28849821, end_position_on_the_genomic_accession=28870309, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  55007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55007, CurrentID=0, Status='live', Symbol='FAM118A', Aliases='C22orf8', description='family with sequence similarity 118 member A', other_designations='bK268H5.C22.4|family with sequence similarity 118, member A', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45309200, end_position_on_the_genomic_accession=45341955, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55008, CurrentID=0, Status='live', Symbol='HERC6', Aliases='', description='HECT and RLD domain containing E3 ubiquitin protein ligase family member 6', other_designations='HECT domain and RCC1-like domain-containing protein 6|hect domain and RLD 6|potential ubiquitin ligase', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88378535, end_position_on_the_genomic_accession=88443097, orientation='plus', exon_count=23, OMIM=609249, no_hdr0=''),
  55009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55009, CurrentID=0, Status='live', Symbol='C19orf24', Aliases='', description='chromosome 19 open reading frame 24', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1275521, end_position_on_the_genomic_accession=1279244, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55010, CurrentID=0, Status='live', Symbol='PARPBP', Aliases='AROM, C12orf48, PARI', description='PARP1 binding protein', other_designations='PARP-1 binding protein|PARP1-binding protein', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102120171, end_position_on_the_genomic_accession=102197520, orientation='plus', exon_count=14, OMIM=613687, no_hdr0=''),
  55011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55011, CurrentID=0, Status='live', Symbol='PIH1D1', Aliases='NOP17', description='PIH1 domain containing 1', other_designations='nucleolar protein 17 homolog', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49446293, end_position_on_the_genomic_accession=49452028, orientation='minus', exon_count=9, OMIM=611480, no_hdr0=''),
  55012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55012, CurrentID=0, Status='live', Symbol='PPP2R3C', Aliases='C14orf10, G4-1, G5pr', description="protein phosphatase 2 regulatory subunit B'', gamma", other_designations="protein phosphatase 2, regulatory subunit B'', gamma|protein phosphatase subunit G5PR|rhabdomyosarcoma antigen MU-RMS-40.6A|rhabdomyosarcoma antigen Mu-RMS-40.6C", map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35085472, end_position_on_the_genomic_accession=35122501, orientation='minus', exon_count=15, OMIM=615902, no_hdr0=''),
  55013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55013, CurrentID=0, Status='live', Symbol='CCDC109B', Aliases='MCUb', description='coiled-coil domain containing 109B', other_designations='coiled-coil domain-containing protein 109B|mitochondrial calcium uniporter regulatory subunit MCUb', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109560199, end_position_on_the_genomic_accession=109688645, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  55014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55014, CurrentID=0, Status='live', Symbol='STX17', Aliases='', description='syntaxin 17', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99906633, end_position_on_the_genomic_accession=99979337, orientation='plus', exon_count=12, OMIM=604204, no_hdr0=''),
  55015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55015, CurrentID=0, Status='live', Symbol='PRPF39', Aliases='', description='pre-mRNA processing factor 39', other_designations='PRP39 homolog|PRP39 pre-mRNA processing factor 39 homolog', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45084099, end_position_on_the_genomic_accession=45115601, orientation='plus', exon_count=14, OMIM=614907, no_hdr0=''),
  55016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55016, CurrentID=0, Status='live', Symbol='MARCH1', Aliases='MARCH-I, RNF171', description='membrane associated ring-CH-type finger 1', other_designations='RING finger protein 171|membrane associated ring finger 1|membrane-associated RING finger protein 1|membrane-associated RING-CH protein I|membrane-associated ring finger (C3HC4) 1, E3 ubiquitin protein ligase', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163524298, end_position_on_the_genomic_accession=164383989, orientation='minus', exon_count=11, OMIM=613331, no_hdr0=''),
  55017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55017, CurrentID=0, Status='live', Symbol='C14orf119', Aliases='', description='chromosome 14 open reading frame 119', other_designations='My028 protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23095474, end_position_on_the_genomic_accession=23100456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55018, CurrentID=0, Status='live', Symbol='LINC00483', Aliases='C17orf73', description='long intergenic non-protein coding RNA 483', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50761034, end_position_on_the_genomic_accession=50767515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55020, CurrentID=0, Status='live', Symbol='TTC38', Aliases='LL22NC03-5H6.5', description='tetratricopeptide repeat domain 38', other_designations='TPR repeat protein 38', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46267964, end_position_on_the_genomic_accession=46294008, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  55022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55022, CurrentID=0, Status='live', Symbol='PID1', Aliases='HMFN2073, NYGGF4, P-CLI1, PCLI1', description='phosphotyrosine interaction domain containing 1', other_designations='phosphotyrosine interaction domain-containing protein 1', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229023973, end_position_on_the_genomic_accession=229271341, orientation='minus', exon_count=5, OMIM=612930, no_hdr0=''),
  55023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55023, CurrentID=0, Status='live', Symbol='PHIP', Aliases='BRWD2, DCAF14, WDR11, ndrp', description='pleckstrin homology domain interacting protein', other_designations='DDB1 and CUL4 associated factor 14|IRS-1 PH domain-binding protein|WD repeat-containing protein 11', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=78934419, end_position_on_the_genomic_accession=79078294, orientation='minus', exon_count=42, OMIM=612870, no_hdr0=''),
  55024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55024, CurrentID=0, Status='live', Symbol='BANK1', Aliases='BANK', description='B-cell scaffold protein with ankyrin repeats 1', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101790607, end_position_on_the_genomic_accession=102074812, orientation='plus', exon_count=18, OMIM=610292, no_hdr0=''),
  55025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55025, CurrentID=0, Status='live', Symbol='FLJ20712', Aliases='', description='uncharacterized FLJ20712', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33726517, end_position_on_the_genomic_accession=33728456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55026, CurrentID=0, Status='live', Symbol='TMEM255A', Aliases='FAM70A', description='transmembrane protein 255A', other_designations='family with sequence similarity 70, member A|protein FAM70A', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120251002, end_position_on_the_genomic_accession=120311556, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55027, CurrentID=0, Status='live', Symbol='HEATR3', Aliases='SYO1', description='HEAT repeat containing 3', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50065970, end_position_on_the_genomic_accession=50105481, orientation='plus', exon_count=16, OMIM=614951, no_hdr0=''),
  55028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55028, CurrentID=0, Status='live', Symbol='C17orf80', Aliases='HLC-8, MIG3, SPEP1', description='chromosome 17 open reading frame 80', other_designations='cell migration-inducing gene 3 protein|human lung cancer oncogene 8 protein|lung cancer-related protein 8|sperm-expressed protein 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73232233, end_position_on_the_genomic_accession=73248959, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105369214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369214, CurrentID=0, Status='live', Symbol='LOC105369214', Aliases='', description='uncharacterized LOC105369214', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10462873, end_position_on_the_genomic_accession=10464841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  55030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55030, CurrentID=0, Status='live', Symbol='FBXO34', Aliases='CGI-301, Fbx34', description='F-box protein 34', other_designations='protein CGI-301', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55271303, end_position_on_the_genomic_accession=55353611, orientation='plus', exon_count=5, OMIM=609104, no_hdr0=''),
  55031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55031, CurrentID=0, Status='live', Symbol='USP47', Aliases='TRFP', description='ubiquitin specific peptidase 47', other_designations='Trf (TATA binding protein-related factor)-proximal homolog|deubiquitinating enzyme 47|ubiquitin thioesterase 47|ubiquitin thiolesterase 47|ubiquitin-specific-processing protease 47', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11841423, end_position_on_the_genomic_accession=11959325, orientation='plus', exon_count=30, OMIM=614460, no_hdr0=''),
  55032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55032, CurrentID=0, Status='live', Symbol='SLC35A5', Aliases='', description='solute carrier family 35 member A5', other_designations='solute carrier family 35, member A5', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112561709, end_position_on_the_genomic_accession=112584439, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55033, CurrentID=0, Status='live', Symbol='FKBP14', Aliases='EDSKMH, FKBP22, IPBP12', description='FK506 binding protein 14', other_designations='22 kDa FK506-binding protein|22 kDa FKBP|FK506 binding protein 14, 22 kDa|FK506-binding protein 14|FKBP-22|PPIase FKBP14|rotamase', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30010583, end_position_on_the_genomic_accession=30026801, orientation='minus', exon_count=5, OMIM=614505, no_hdr0=''),
  55034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55034, CurrentID=0, Status='live', Symbol='MOCOS', Aliases='HMCS, MCS, MOS', description='molybdenum cofactor sulfurase', other_designations='MCS|MOS', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36187517, end_position_on_the_genomic_accession=36268722, orientation='plus', exon_count=15, OMIM=613274, no_hdr0=''),
  55035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55035, CurrentID=0, Status='live', Symbol='NOL8', Aliases='C9orf34, NOP132, bA62C3.3, bA62C3.4', description='nucleolar protein 8', other_designations='nucleolar protein Nop132', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92297358, end_position_on_the_genomic_accession=92325994, orientation='minus', exon_count=21, OMIM=611534, no_hdr0=''),
  55036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55036, CurrentID=0, Status='live', Symbol='CCDC40', Aliases='CILD15, FAP172', description='coiled-coil domain containing 40', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80036632, end_position_on_the_genomic_accession=80100613, orientation='plus', exon_count=25, OMIM=613799, no_hdr0=''),
  55037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55037, CurrentID=0, Status='live', Symbol='PTCD3', Aliases='MRP-S39', description='pentatricopeptide repeat domain 3', other_designations='28S ribosomal protein S39, mitochondrial|TRG-15|pentatricopeptide repeat-containing protein 3, mitochondrial|transformation-related gene 15 protein', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86106182, end_position_on_the_genomic_accession=86142157, orientation='plus', exon_count=24, OMIM=614918, no_hdr0=''),
  55038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55038, CurrentID=0, Status='live', Symbol='CDCA4', Aliases='HEPP, SEI-3/HEPP', description='cell division cycle associated 4', other_designations='hematopoietic progenitor protein', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105009573, end_position_on_the_genomic_accession=105021088, orientation='minus', exon_count=2, OMIM=612270, no_hdr0=''),
  55039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55039, CurrentID=0, Status='live', Symbol='TRMT12', Aliases='TRM12, TYW2', description='tRNA methyltransferase 12 homolog (S. cerevisiae)', other_designations='alpha-amino-alpha-carboxypropyl transferase TYW2|homolog of yeast tRNA methyltransferase|tRNA methyltranferase 12 homolog|tRNA(Phe) (4-demethylwyosine(37)-C(7)) aminocarboxypropyltransferase|tRNA-yW synthesizing protein 2|tRNA-yW-synthesizing protein 2', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124450807, end_position_on_the_genomic_accession=124453026, orientation='plus', exon_count=1, OMIM=611244, no_hdr0=''),
  55040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55040, CurrentID=0, Status='live', Symbol='EPN3', Aliases='', description='epsin 3', other_designations='EPS-15-interacting protein 3', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50532687, end_position_on_the_genomic_accession=50543750, orientation='plus', exon_count=10, OMIM=607264, no_hdr0=''),
  55041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55041, CurrentID=0, Status='live', Symbol='PLEKHB2', Aliases='EVT2', description='pleckstrin homology domain containing B2', other_designations='PH domain-containing family B member 2|evectin-2|pleckstrin homology domain containing, family B (evectins) member 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131104735, end_position_on_the_genomic_accession=131149852, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105369217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369217, CurrentID=0, Status='live', Symbol='LOC105369217', Aliases='', description='uncharacterized LOC105369217', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45700184, end_position_on_the_genomic_accession=45704983, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55048, CurrentID=0, Status='live', Symbol='VPS37C', Aliases='', description='vacuolar protein sorting 37 homolog C (S. cerevisiae)', other_designations='ESCRT-I complex subunit VPS37C|hVps37C|vacuolar protein sorting 37C', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61130256, end_position_on_the_genomic_accession=61161444, orientation='minus', exon_count=6, OMIM=610038, no_hdr0=''),
  55049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55049, CurrentID=0, Status='live', Symbol='C19orf60', Aliases='', description='chromosome 19 open reading frame 60', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18588685, end_position_on_the_genomic_accession=18592337, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  55051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55051, CurrentID=0, Status='live', Symbol='NRDE2', Aliases='C14orf102', description='NRDE-2, necessary for RNA interference, domain containing', other_designations='UPF0614 protein C14orf102', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90278054, end_position_on_the_genomic_accession=90332137, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  55052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55052, CurrentID=0, Status='live', Symbol='MRPL20', Aliases='L20mt, MRP-L20', description='mitochondrial ribosomal protein L20', other_designations='', map_location='1p36.3-p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1401896, end_position_on_the_genomic_accession=1407834, orientation='minus', exon_count=4, OMIM=611833, no_hdr0=''),
  105369218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369218, CurrentID=0, Status='live', Symbol='LOC105369218', Aliases='', description='uncharacterized LOC105369218', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55054, CurrentID=0, Status='live', Symbol='ATG16L1', Aliases='APG16L, ATG16A, ATG16L, IBD10, WDR30', description='autophagy related 16 like 1', other_designations='APG16L beta|ATG16 autophagy related 16-like 1|WD repeat domain 30|autophagy related 16-like 1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233251571, end_position_on_the_genomic_accession=233295674, orientation='plus', exon_count=19, OMIM=610767, no_hdr0=''),
  55055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55055, CurrentID=0, Status='live', Symbol='ZWILCH', Aliases='KNTC1AP, hZwilch', description='zwilch kinetochore protein', other_designations='Zwilch, kinetochore associated, homolog|homolog of Drosophila Zwilch', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66505083, end_position_on_the_genomic_accession=66549485, orientation='plus', exon_count=19, OMIM=609984, no_hdr0=''),
  55056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55056, CurrentID=0, Status='live', Symbol='FLJ10038', Aliases='', description='uncharacterized protein FLJ10038', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50348936, end_position_on_the_genomic_accession=50354879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55057, CurrentID=0, Status='live', Symbol='AIM1L', Aliases='CRYBG2', description='absent in melanoma 1-like', other_designations='beta-gamma crystallin domain containing 2|beta/gamma crystallin domain-containing protein 2', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26321859, end_position_on_the_genomic_accession=26354130, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  642691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642691, CurrentID=0, Status='live', Symbol='FLJ37786', Aliases='', description='uncharacterized LOC642691', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=148864462, end_position_on_the_genomic_accession=148866310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  55061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55061, CurrentID=0, Status='live', Symbol='SUSD4', Aliases='PRO222', description='sushi domain containing 4', other_designations='YHGM196', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223220819, end_position_on_the_genomic_accession=223364202, orientation='minus', exon_count=13, OMIM=615827, no_hdr0=''),
  55062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55062, CurrentID=0, Status='live', Symbol='WIPI1', Aliases='ATG18, ATG18A, WIPI49', description='WD repeat domain, phosphoinositide interacting 1', other_designations='WD40 repeat protein Interacting with phosphoInositides of 49kDa|WD40 repeat protein interacting with phosphoinositides of 49 kDa|WIPI 49 kDa|WIPI-1 alpha|atg18 protein homolog', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68421281, end_position_on_the_genomic_accession=68457512, orientation='minus', exon_count=13, OMIM=609224, no_hdr0=''),
  55063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55063, CurrentID=0, Status='live', Symbol='ZCWPW1', Aliases='ZCW1', description='zinc finger, CW type with PWWP domain 1', other_designations='zinc finger CW type with PWWP domain 1|zinc finger, CW-type with PWWP domain 1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100400872, end_position_on_the_genomic_accession=100429414, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55064, CurrentID=0, Status='live', Symbol='SPATA6L', Aliases='C9orf68, bA6J24.2', description='spermatogenesis associated 6-like', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4595837, end_position_on_the_genomic_accession=4666801, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55065, CurrentID=0, Status='live', Symbol='SLC52A1', Aliases='GPCR42, GPR172B, PAR2, RBFVD, RFT1, RFVT1, hRFT1', description='solute carrier family 52 (riboflavin transporter), member 1', other_designations='G protein-coupled receptor 172B|G-protein coupled receptor GPCR42|PERV-A receptor 2|porcine endogenous retrovirus A receptor 2', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5032602, end_position_on_the_genomic_accession=5035458, orientation='minus', exon_count=5, OMIM=607883, no_hdr0=''),
  55066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55066, CurrentID=0, Status='live', Symbol='PDPR', Aliases='PDP3', description='pyruvate dehydrogenase phosphatase regulatory subunit', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70113626, end_position_on_the_genomic_accession=70162534, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100132658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132658, CurrentID=0, Status='live', Symbol='LOC100132658', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100910614, end_position_on_the_genomic_accession=100911632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55068, CurrentID=0, Status='live', Symbol='ENOX1', Aliases='CNOX, PIG38, bA64J21.1, cCNOX', description='ecto-NOX disulfide-thiol exchanger 1', other_designations='candidate growth-related and time keeping constitutive hydroquinone (NADH) oxidase|candidate growth-related and time keeping constitutive hydroquinone [NADH] oxidase|cell proliferation-inducing gene 38 protein|constitutive Ecto-NOX', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43213130, end_position_on_the_genomic_accession=43786980, orientation='minus', exon_count=24, OMIM=610914, no_hdr0=''),
  55069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55069, CurrentID=0, Status='live', Symbol='TMEM248', Aliases='C7orf42', description='transmembrane protein 248', other_designations='UPF0458 protein C7orf42', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66921216, end_position_on_the_genomic_accession=66958551, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55070, CurrentID=0, Status='live', Symbol='DET1', Aliases='', description='de-etiolated homolog 1 (Arabidopsis)', other_designations='de-etiolated 1|de-etiolated-1 homolog', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88511482, end_position_on_the_genomic_accession=88546698, orientation='minus', exon_count=10, OMIM=608727, no_hdr0=''),
  55071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55071, CurrentID=0, Status='live', Symbol='C9orf40', Aliases='', description='chromosome 9 open reading frame 40', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74946583, end_position_on_the_genomic_accession=74952886, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55072, CurrentID=0, Status='live', Symbol='RNF31', Aliases='HOIP, ZIBRA', description='ring finger protein 31', other_designations='HOIL-1-interacting protein|zinc in-between-RING-finger ubiquitin-associated domain protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24147450, end_position_on_the_genomic_accession=24160661, orientation='plus', exon_count=21, OMIM=612487, no_hdr0=''),
  55073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55073, CurrentID=0, Status='live', Symbol='LRRC37A4P', Aliases='LRRC37, LRRC37A4', description='leucine-rich repeat containing 37 member A4, pseudogene', other_designations='leucine rich repeat containing 37, member A4 (pseudogene)|leucine rich repeat containing 37, member A4, pseudogene', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45505883, end_position_on_the_genomic_accession=45520523, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55074, CurrentID=0, Status='live', Symbol='OXR1', Aliases='Nbla00307, TLDC3', description='oxidation resistance 1', other_designations='TBC/LysM-associated domain containing 3|putative protein product of Nbla00307', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106270178, end_position_on_the_genomic_accession=106752694, orientation='plus', exon_count=25, OMIM=605609, no_hdr0=''),
  55075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55075, CurrentID=0, Status='live', Symbol='UACA', Aliases='NUCLING', description='uveal autoantigen with coiled-coil domains and ankyrin repeats', other_designations='nuclear membrane binding protein|uveal autoantigen with coiled coil domains and ankyrin repeats', map_location='15q22-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70654554, end_position_on_the_genomic_accession=70763511, orientation='minus', exon_count=22, OMIM=612516, no_hdr0=''),
  55076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55076, CurrentID=0, Status='live', Symbol='TMEM45A', Aliases='DERP7', description='transmembrane protein 45A', other_designations='DNA polymerase-transactivated protein 4|dermal papilla derived protein 7|dermal papilla-derived protein 7', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100492619, end_position_on_the_genomic_accession=100577453, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369222, CurrentID=0, Status='live', Symbol='LOC105369222', Aliases='', description='uncharacterized protein C6orf106 homolog pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106344566, end_position_on_the_genomic_accession=106345483, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55079, CurrentID=0, Status='live', Symbol='FEZF2', Aliases='FEZ, FEZL, FKSG36, TOF, ZFP312, ZNF312', description='FEZ family zinc finger 2', other_designations='forebrain embryonic zinc finger-like protein 2|zinc finger protein 312', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62369672, end_position_on_the_genomic_accession=62373515, orientation='minus', exon_count=5, OMIM=607414, no_hdr0=''),
  55080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55080, CurrentID=0, Status='live', Symbol='TAPBPL', Aliases='TAPBP-R, TAPBPR', description='TAP binding protein-like', other_designations='TAP binding protein related|TAP-binding protein-like|TAP-binding protein-related protein|TAPASIN-R|tapasin-like', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6451655, end_position_on_the_genomic_accession=6462322, orientation='plus', exon_count=10, OMIM=607081, no_hdr0=''),
  55081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55081, CurrentID=0, Status='live', Symbol='IFT57', Aliases='ESRRBL1, HIPPI, MHS4R2', description='intraflagellar transport 57', other_designations='HIP1 protein interactor|HIP1-interacting protein|dermal papilla-derived protein 8|estrogen-related receptor beta like 1|estrogen-related receptor beta-like protein 1|huntingtin interacting protein-1 interacting protein|intraflagellar transport 57 homolog', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108160812, end_position_on_the_genomic_accession=108222570, orientation='minus', exon_count=11, OMIM=606621, no_hdr0=''),
  55082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55082, CurrentID=0, Status='live', Symbol='ARGLU1', Aliases='', description='arginine and glutamate rich 1', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106543314, end_position_on_the_genomic_accession=106568166, orientation='minus', exon_count=5, OMIM=614046, no_hdr0=''),
  55083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55083, CurrentID=0, Status='live', Symbol='KIF26B', Aliases='', description='kinesin family member 26B', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245154985, end_position_on_the_genomic_accession=245709432, orientation='plus', exon_count=19, OMIM=614026, no_hdr0=''),
  55084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55084, CurrentID=0, Status='live', Symbol='SOBP', Aliases='JXC1, MRAMS', description='sine oculis binding protein homolog', other_designations='jackson circler protein 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107490098, end_position_on_the_genomic_accession=107661309, orientation='plus', exon_count=11, OMIM=613667, no_hdr0=''),
  55086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55086, CurrentID=0, Status='live', Symbol='CXorf57', Aliases='', description='chromosome X open reading frame 57', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106611930, end_position_on_the_genomic_accession=106679443, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  55088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55088, CurrentID=0, Status='live', Symbol='CCDC186', Aliases='C10orf118', description='coiled-coil domain containing 186', other_designations='CTCL tumor antigen HD-CL-01/L14-2|CTCL tumor antigen L14-2', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114120856, end_position_on_the_genomic_accession=114174605, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55089, CurrentID=0, Status='live', Symbol='SLC38A4', Aliases='ATA3, NAT3, PAAT, SNAT4', description='solute carrier family 38 member 4', other_designations='N amino acid transporter 3|Na(+)-coupled neutral amino acid transporter 4|amino acid transporter A3|amino acid transporter system A3|solute carrier family 38, member 4|system A amino acid transporter 3|system N amino acid transporter 3', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46764761, end_position_on_the_genomic_accession=46832422, orientation='minus', exon_count=18, OMIM=608065, no_hdr0=''),
  55090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55090, CurrentID=0, Status='live', Symbol='MED9', Aliases='MED25', description='mediator complex subunit 9', other_designations='mediator of RNA polymerase II transcription, subunit 9 homolog|mediator subunit 25', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17476986, end_position_on_the_genomic_accession=17493220, orientation='plus', exon_count=2, OMIM=609878, no_hdr0=''),
  55092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55092, CurrentID=0, Status='live', Symbol='TMEM51', Aliases='C1orf72', description='transmembrane protein 51', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15152532, end_position_on_the_genomic_accession=15220478, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  55093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55093, CurrentID=0, Status='live', Symbol='WDYHV1', Aliases='C8orf32', description='WDYHV motif containing 1', other_designations='N-terminal Gln amidase|WDYHV motif-containing protein 1|nt(Q)-amidase|protein NH2-terminal glutamine deamidase', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123416725, end_position_on_the_genomic_accession=123448380, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55094, CurrentID=0, Status='live', Symbol='GPATCH1', Aliases='ECGP, GPATC1', description='G-patch domain containing 1', other_designations='G patch domain containing 1|evolutionarily conserved G patch domain containing|evolutionarily conserved G-patch domain containing|evolutionarily conserved G-patch domain-containing protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33080880, end_position_on_the_genomic_accession=33130412, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55095, CurrentID=0, Status='live', Symbol='SAMD4B', Aliases='SMGB, Smaug2', description='sterile alpha motif domain containing 4B', other_designations='SAM domain-containing protein 4B|smaug 2|smaug homolog B|sterile alpha motif domain-containing protein 4B', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39342459, end_position_on_the_genomic_accession=39390760, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55096, CurrentID=0, Status='live', Symbol='EBLN2', Aliases='EBLN-2', description='endogenous Bornavirus-like nucleoprotein 2', other_designations='EBLN-1|endogenous Borna-like N element-1|endogenous Borna-like N element-2', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73061659, end_position_on_the_genomic_accession=73063337, orientation='plus', exon_count=1, OMIM=613250, no_hdr0=''),
  100132659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132659, CurrentID=0, Status='live', Symbol='LOC100132659', Aliases='', description='single-stranded DNA binding protein 2 pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82263999, end_position_on_the_genomic_accession=82264974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55100, CurrentID=0, Status='live', Symbol='WDR70', Aliases='', description='WD repeat domain 70', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37379289, end_position_on_the_genomic_accession=37752674, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  55101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55101, CurrentID=0, Status='live', Symbol='ATP5SL', Aliases='', description='ATP5S-like', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41431318, end_position_on_the_genomic_accession=41447827, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55102, CurrentID=0, Status='live', Symbol='ATG2B', Aliases='C14orf103', description='autophagy related 2B', other_designations='ATG2 autophagy related 2 homolog B', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96281258, end_position_on_the_genomic_accession=96363341, orientation='minus', exon_count=45, OMIM=616226, no_hdr0=''),
  55103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55103, CurrentID=0, Status='live', Symbol='RALGPS2', Aliases='dJ595C2.1', description='Ral GEF with PH domain and SH3 binding motif 2', other_designations='Ral-A exchange factor RalGPS2|ral GEF with PH domain and SH3-binding motif 2|ralA exchange factor RalGPS2', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178725147, end_position_on_the_genomic_accession=178921842, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55105, CurrentID=0, Status='live', Symbol='GPATCH2', Aliases='CT110, GPATC2, PPP1R30', description='G-patch domain containing 2', other_designations='G patch domain containing 2|cancer/testis antigen 110|protein phosphatase 1, regulatory subunit 30', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217426992, end_position_on_the_genomic_accession=217631102, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  55106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55106, CurrentID=0, Status='live', Symbol='SLFN12', Aliases='SLFN3', description='schlafen family member 12', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35409650, end_position_on_the_genomic_accession=35433229, orientation='minus', exon_count=7, OMIM=614955, no_hdr0=''),
  55107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55107, CurrentID=0, Status='live', Symbol='ANO1', Aliases='DOG1, ORAOV2, TAOS2, TMEM16A', description='anoctamin 1, calcium activated chloride channel', other_designations='Ca2+-activated Cl- channel|calcium activated chloride channel|discovered on gastrointestinal stromal tumors protein 1|oral cancer overexpressed 2|transmembrane protein 16A (eight membrane-spanning domains)|tumor-amplified and overexpressed sequence 2', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70078302, end_position_on_the_genomic_accession=70189545, orientation='plus', exon_count=30, OMIM=610108, no_hdr0=''),
  55108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55108, CurrentID=0, Status='live', Symbol='BSDC1', Aliases='', description='BSD domain containing 1', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32364633, end_position_on_the_genomic_accession=32394461, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55109, CurrentID=0, Status='live', Symbol='AGGF1', Aliases='GPATC7, GPATCH7, HSU84971, HUS84971, VG5Q', description='angiogenic factor with G-patch and FHA domains 1', other_designations='G patch domain-containing protein 7|angiogenic factor VG5Q|vasculogenesis gene on 5q protein', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77030385, end_position_on_the_genomic_accession=77065234, orientation='plus', exon_count=14, OMIM=608464, no_hdr0=''),
  55110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55110, CurrentID=0, Status='live', Symbol='MAGOHB', Aliases='MGN2, mago, magoh', description='mago homolog B, exon junction complex core component', other_designations='mago nashi protein|mago-nashi homolog 2|mago-nashi homolog B', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10603765, end_position_on_the_genomic_accession=10613609, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55111, CurrentID=0, Status='live', Symbol='PLEKHJ1', Aliases='GNRPX', description='pleckstrin homology domain containing J1', other_designations='PH domain-containing family J member 1|guanine nucleotide releasing protein x|pleckstrin homology domain containing, family J member 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2229952, end_position_on_the_genomic_accession=2237405, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55112, CurrentID=0, Status='live', Symbol='WDR60', Aliases='FAP163, SRPS6, SRTD8', description='WD repeat domain 60', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158856578, end_position_on_the_genomic_accession=158946192, orientation='plus', exon_count=27, OMIM=615462, no_hdr0=''),
  55113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55113, CurrentID=0, Status='live', Symbol='XKR8', Aliases='XRG8, hXkr8', description='X-linked Kx blood group related 8', other_designations='X Kell blood group precursor-related family, member 8|XK, Kell blood group complex subunit-related family, member 8', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27959462, end_position_on_the_genomic_accession=27968096, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  55114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55114, CurrentID=0, Status='live', Symbol='ARHGAP17', Aliases='MST066, MST110, MSTP038, MSTP066, MSTP110, NADRIN, PP367, PP4534, RICH-1, RICH1, WBP15', description='Rho GTPase activating protein 17', other_designations='RhoGAP interacting with CIP4 homologs 1|neuron-associated developmentally regulated protein|rho-type GTPase-activating protein 17', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24919384, end_position_on_the_genomic_accession=25015601, orientation='minus', exon_count=25, OMIM=608293, no_hdr0=''),
  55116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55116, CurrentID=0, Status='live', Symbol='TMEM39B', Aliases='', description='transmembrane protein 39B', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32072031, end_position_on_the_genomic_accession=32102866, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  55117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55117, CurrentID=0, Status='live', Symbol='SLC6A15', Aliases='NTT73, SBAT1, V7-3, hv7-3', description='solute carrier family 6 (neutral amino acid transporter), member 15', other_designations='homolog of rat orphan transporter v7-3|orphan sodium- and chloride-dependent neurotransmitter transporter NTT73|orphan transporter v7-3|sodium- and chloride-dependent neurotransmitter transporter NTT73|sodium-coupled branched-chain amino-acid transporter 1|sodium/chloride dependent neurotransmitter transporter Homo sapiens orphan neurotransmitter transporter NTT7|solute carrier family 6 (neurotransmitter transporter), member 15|solute carrier family 6 member 15|solute carrier family 6, member 15|transporter v7-3', map_location='12q21.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=84859488, end_position_on_the_genomic_accession=84913615, orientation='minus', exon_count=15, OMIM=607971, no_hdr0=''),
  55118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55118, CurrentID=0, Status='live', Symbol='CRTAC1', Aliases='ASPIC, ASPIC1, CEP-68', description='cartilage acidic protein 1', other_designations='acidic secreted protein in cartilage|chondrocyte expressed protein 68 kDa CEP-68', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97854711, end_position_on_the_genomic_accession=98030828, orientation='minus', exon_count=18, OMIM=606276, no_hdr0=''),
  55119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55119, CurrentID=0, Status='live', Symbol='PRPF38B', Aliases='NET1', description='pre-mRNA processing factor 38B', other_designations='PRP38 pre-mRNA processing factor 38 domain containing B|sarcoma antigen NY-SAR-27', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108692292, end_position_on_the_genomic_accession=108701803, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  55120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55120, CurrentID=0, Status='live', Symbol='FANCL', Aliases='FAAP43, PHF9, POG', description='Fanconi anemia complementation group L', other_designations='Fanconi anemia, complementation group L|PHD finger protein 9|fanconi anemia group L protein|fanconi anemia-associated polypeptide of 43 kDa', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=58159243, end_position_on_the_genomic_accession=58241681, orientation='minus', exon_count=15, OMIM=608111, no_hdr0=''),
  55122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55122, CurrentID=0, Status='live', Symbol='AKIRIN2', Aliases='C6orf166, FBI1, dJ486L4.2', description='akirin 2', other_designations='fourteen-three-three beta interactant 1', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87674860, end_position_on_the_genomic_accession=87702267, orientation='minus', exon_count=5, OMIM=615165, no_hdr0=''),
  55124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55124, CurrentID=0, Status='live', Symbol='PIWIL2', Aliases='CT80, HILI, PIWIL1L, mili', description='piwi-like RNA-mediated gene silencing 2', other_designations='80kDa PIWIL2 short isoform|Miwi like|cancer/testis antigen 80|piwi-like 2|piwil2-like protein', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22275297, end_position_on_the_genomic_accession=22357568, orientation='plus', exon_count=25, OMIM=610312, no_hdr0=''),
  55125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55125, CurrentID=0, Status='live', Symbol='CEP192', Aliases='PPP1R62', description='centrosomal protein 192kDa', other_designations='192 kDa centrosomal protein|protein phosphatase 1, regulatory subunit 62', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12991321, end_position_on_the_genomic_accession=13125053, orientation='plus', exon_count=47, OMIM=616426, no_hdr0=''),
  55127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55127, CurrentID=0, Status='live', Symbol='HEATR1', Aliases='BAP28, UTP10', description='HEAT repeat containing 1', other_designations='UTP10, small subunit (SSU) processome component, homolog|protein BAP28', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236536173, end_position_on_the_genomic_accession=236604541, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  55128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55128, CurrentID=0, Status='live', Symbol='TRIM68', Aliases='GC109, RNF137, SS-56, SS56', description='tripartite motif containing 68', other_designations='Ro/SSA1 related protein|SSA protein SS-56|ring finger protein 137|tripartite motif-containing protein 68', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4598672, end_position_on_the_genomic_accession=4608259, orientation='minus', exon_count=7, OMIM=613184, no_hdr0=''),
  55129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55129, CurrentID=0, Status='live', Symbol='ANO10', Aliases='SCAR10, TMEM16K', description='anoctamin 10', other_designations='transmembrane protein 16K', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43366326, end_position_on_the_genomic_accession=43691594, orientation='minus', exon_count=21, OMIM=613726, no_hdr0=''),
  55130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55130, CurrentID=0, Status='live', Symbol='ARMC4', Aliases='CILD23', description='armadillo repeat containing 4', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27812164, end_position_on_the_genomic_accession=27999067, orientation='minus', exon_count=30, OMIM=615408, no_hdr0=''),
  55131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55131, CurrentID=0, Status='live', Symbol='RBM28', Aliases='ANES', description='RNA binding motif protein 28', other_designations='2810480G15Rik', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128310383, end_position_on_the_genomic_accession=128343913, orientation='minus', exon_count=20, OMIM=612074, no_hdr0=''),
  55132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55132, CurrentID=0, Status='live', Symbol='LARP1B', Aliases='LARP2', description='La ribonucleoprotein domain family member 1B', other_designations='La ribonucleoprotein domain family, member 1B|La ribonucleoprotein domain family, member 2|la ribonucleoprotein domain family member 2|la-related protein 2', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128061266, end_position_on_the_genomic_accession=128222931, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  55133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55133, CurrentID=0, Status='live', Symbol='SRBD1', Aliases='', description='S1 RNA binding domain 1', other_designations='H_NH0576F01.1|WUGSC:H_NH0576F01.1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45388680, end_position_on_the_genomic_accession=45611295, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  55135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55135, CurrentID=0, Status='live', Symbol='WRAP53', Aliases='DKCB3, TCAB1, WDR79', description='WD repeat containing, antisense to TP53', other_designations='WD repeat-containing protein 79|WD repeat-containing protein WRAP53|WD40 protein Wrap53|WD40 repeat-containing protein antisense to TP53|WD40 repeat-containing protein encoding RNA antisense to p53', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7686071, end_position_on_the_genomic_accession=7703502, orientation='plus', exon_count=13, OMIM=612661, no_hdr0=''),
  55137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55137, CurrentID=0, Status='live', Symbol='FIGN', Aliases='', description='fidgetin', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=163604937, end_position_on_the_genomic_accession=163736197, orientation='minus', exon_count=7, OMIM=605295, no_hdr0=''),
  55138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55138, CurrentID=0, Status='live', Symbol='FAM90A1', Aliases='', description='family with sequence similarity 90 member A1', other_designations='family with sequence similarity 90, member A1', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8221260, end_position_on_the_genomic_accession=8227703, orientation='minus', exon_count=7, OMIM=613041, no_hdr0=''),
  55139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55139, CurrentID=0, Status='live', Symbol='ANKZF1', Aliases='ZNF744', description='ankyrin repeat and zinc finger domain containing 1', other_designations='zinc finger protein 744', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219229757, end_position_on_the_genomic_accession=219236669, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  55140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55140, CurrentID=0, Status='live', Symbol='ELP3', Aliases='KAT9', description='elongator acetyltransferase complex subunit 3', other_designations='elongation protein 3 homolog', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28090231, end_position_on_the_genomic_accession=28191153, orientation='plus', exon_count=16, OMIM=612722, no_hdr0=''),
  55142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55142, CurrentID=0, Status='live', Symbol='HAUS2', Aliases='C15orf25, CEP27, HsT17025', description='HAUS augmin like complex subunit 2', other_designations='HAUS augmin-like complex, subunit 2|centrosomal protein 27kDa|centrosomal protein of 27 kDa', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42548813, end_position_on_the_genomic_accession=42569994, orientation='plus', exon_count=6, OMIM=613429, no_hdr0=''),
  55143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55143, CurrentID=0, Status='live', Symbol='CDCA8', Aliases='BOR, BOREALIN, DasraB, MESRGP', description='cell division cycle associated 8', other_designations='Dasra B|cell division cycle-associated protein 8|dasra-B|hDasra-B|pluripotent embryonic stem cell-related gene 3 protein', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37692401, end_position_on_the_genomic_accession=37709719, orientation='plus', exon_count=11, OMIM=609977, no_hdr0=''),
  55144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55144, CurrentID=0, Status='live', Symbol='LRRC8D', Aliases='LRRC5', description='leucine-rich repeat containing 8 family member D', other_designations='leucine rich repeat containing 5|leucine rich repeat containing 8 family, member D|leucine-rich repeat-containing 5|leucine-rich repeat-containing protein 5|leucine-rich repeat-containing protein 8D', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89821014, end_position_on_the_genomic_accession=89936432, orientation='plus', exon_count=6, OMIM=612890, no_hdr0=''),
  55145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55145, CurrentID=0, Status='live', Symbol='THAP1', Aliases='DYT6', description='THAP domain containing, apoptosis associated protein 1', other_designations='4833431A01Rik|THAP domain protein 1|nuclear proapoptotic factor', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42836674, end_position_on_the_genomic_accession=42843331, orientation='minus', exon_count=3, OMIM=609520, no_hdr0=''),
  55146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55146, CurrentID=0, Status='live', Symbol='ZDHHC4', Aliases='ZNF374', description='zinc finger, DHHC-type containing 4', other_designations='DHHC-4|zinc finger DHHC domain-containing protein 4|zinc finger protein 374|zinc finger, DHHC domain containing 4', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6577412, end_position_on_the_genomic_accession=6588979, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55147, CurrentID=0, Status='live', Symbol='RBM23', Aliases='CAPERbeta, PP239, RNPC4', description='RNA binding motif protein 23', other_designations='RNA-binding motif protein 23|RNA-binding region (RNP1, RRM) containing 4|RNA-binding region-containing protein 4|coactivator of activating protein-1 and estrogen recep- tors beta|splicing factor SF2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22900645, end_position_on_the_genomic_accession=22919187, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55148, CurrentID=0, Status='live', Symbol='UBR7', Aliases='C14orf130', description='ubiquitin protein ligase E3 component n-recognin 7 (putative)', other_designations='N-recognin-7', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93207056, end_position_on_the_genomic_accession=93229215, orientation='plus', exon_count=11, OMIM=613816, no_hdr0=''),
  55149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55149, CurrentID=0, Status='live', Symbol='MTPAP', Aliases='PAPD1, SPAX4', description='mitochondrial poly(A) polymerase', other_designations='PAP-associated domain-containing protein 1|TUTase 1|polynucleotide adenylyltransferase|terminal uridylyltransferase 1', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30309801, end_position_on_the_genomic_accession=30349338, orientation='minus', exon_count=9, OMIM=613669, no_hdr0=''),
  55150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55150, CurrentID=0, Status='live', Symbol='C19orf73', Aliases='', description='chromosome 19 open reading frame 73', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49118397, end_position_on_the_genomic_accession=49119140, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55151, CurrentID=0, Status='live', Symbol='TMEM38B', Aliases='C9orf87, D4Ertd89e, OI14, TRIC-B, TRICB, bA219P18.1', description='transmembrane protein 38B', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105694525, end_position_on_the_genomic_accession=105776611, orientation='plus', exon_count=10, OMIM=611236, no_hdr0=''),
  55152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55152, CurrentID=0, Status='live', Symbol='DALRD3', Aliases='', description='DALR anticodon binding domain containing 3', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49015488, end_position_on_the_genomic_accession=49021386, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55153, CurrentID=0, Status='live', Symbol='SDAD1', Aliases='', description='SDA1 domain containing 1', other_designations='SDA1 domain-containing protein 1|nucleolar protein 130', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75949906, end_position_on_the_genomic_accession=75990962, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  55154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55154, CurrentID=0, Status='live', Symbol='MSTO1', Aliases='LST005, MST', description='misato 1, mitochondrial distribution and morphology regulator', other_designations='misato homolog 1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155610170, end_position_on_the_genomic_accession=155614967, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105369235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369235, CurrentID=0, Status='live', Symbol='LOC105369235', Aliases='', description='uncharacterized LOC105369235', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67642930, end_position_on_the_genomic_accession=67644461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55156, CurrentID=0, Status='live', Symbol='ARMC1', Aliases='Arcp', description='armadillo repeat containing 1', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65602456, end_position_on_the_genomic_accession=65634217, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55157, CurrentID=0, Status='live', Symbol='DARS2', Aliases='ASPRS, LBSL, MT-ASPRS', description='aspartyl-tRNA synthetase 2, mitochondrial', other_designations='aspartate tRNA ligase 2, mitochondrial|aspartyl-tRNA synthetase, mitochondrial', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173824645, end_position_on_the_genomic_accession=173858544, orientation='plus', exon_count=18, OMIM=610956, no_hdr0=''),
  55159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55159, CurrentID=0, Status='live', Symbol='RFWD3', Aliases='RNF201', description='ring finger and WD repeat domain 3', other_designations='RING finger and WD repeat domain-containing protein 3|RING finger protein 201', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74621394, end_position_on_the_genomic_accession=74666892, orientation='minus', exon_count=15, OMIM=614151, no_hdr0=''),
  55160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55160, CurrentID=0, Status='live', Symbol='ARHGEF10L', Aliases='GrinchGEF', description='Rho guanine nucleotide exchange factor 10 like', other_designations='Rho guanine nucleotide exchange factor (GEF) 10-like', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17520197, end_position_on_the_genomic_accession=17697875, orientation='plus', exon_count=34, OMIM=612494, no_hdr0=''),
  55161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55161, CurrentID=0, Status='live', Symbol='TMEM33', Aliases='1600019D15Rik, SHINC3', description='transmembrane protein 33', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41935120, end_position_on_the_genomic_accession=41960807, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55163, CurrentID=0, Status='live', Symbol='PNPO', Aliases='HEL-S-302, PDXPO', description="pyridoxamine 5'-phosphate oxidase", other_designations="epididymis secretory protein Li 302|pyridoxal 5'-phosphate synthase|pyridoxamine-phosphate oxidase|pyridoxine 5'-phosphate oxidase", map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47941523, end_position_on_the_genomic_accession=47949308, orientation='plus', exon_count=7, OMIM=603287, no_hdr0=''),
  55164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55164, CurrentID=0, Status='live', Symbol='SHQ1', Aliases='GRIM-1, Shq1p', description='SHQ1, H/ACA ribonucleoprotein assembly factor', other_designations='SHQ1 homolog|gene associated with retinoid and interferon-induced mortality 1', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72749277, end_position_on_the_genomic_accession=72848475, orientation='minus', exon_count=17, OMIM=613663, no_hdr0=''),
  55165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55165, CurrentID=0, Status='live', Symbol='CEP55', Aliases='C10orf3, CT111, URCC6', description='centrosomal protein 55kDa', other_designations='cancer/testis antigen 111|up-regulated in colon cancer 6', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93496612, end_position_on_the_genomic_accession=93529094, orientation='plus', exon_count=9, OMIM=610000, no_hdr0=''),
  55166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55166, CurrentID=0, Status='live', Symbol='CENPQ', Aliases='C6orf139, CENP-Q', description='centromere protein Q', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49463341, end_position_on_the_genomic_accession=49493107, orientation='plus', exon_count=9, OMIM=611506, no_hdr0=''),
  55167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55167, CurrentID=0, Status='live', Symbol='MSL2', Aliases='MSL-2L1, RNF184, MSL2', description='male-specific lethal 2 homolog (Drosophila)', other_designations='MSL2-like 1|male-specific lethal 2-like 1|male-specific lethal-2 homolog 1|ring finger protein 184', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136148918, end_position_on_the_genomic_accession=136197393, orientation='minus', exon_count=7, OMIM=614802, no_hdr0=''),
  55168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55168, CurrentID=0, Status='live', Symbol='MRPS18A', Aliases='HumanS18b, MRP-S18-3, MRPS18-3, S18bmt', description='mitochondrial ribosomal protein S18A', other_designations='28S ribosomal protein S18-3, mitochondrial|MRP-S18-a|S18mt-a|mitochondrial ribosomal protein S18-3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43671197, end_position_on_the_genomic_accession=43687812, orientation='minus', exon_count=6, OMIM=611981, no_hdr0=''),
  55170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55170, CurrentID=0, Status='live', Symbol='PRMT6', Aliases='HRMT1L6', description='protein arginine methyltransferase 6', other_designations='HMT1 hnRNP methyltransferase-like 6|heterogeneous nuclear ribonucleoprotein methyltransferase-like protein 6|histone-arginine N-methyltransferase PRMT6', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107056645, end_position_on_the_genomic_accession=107059294, orientation='plus', exon_count=1, OMIM=608274, no_hdr0=''),
  55171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55171, CurrentID=0, Status='live', Symbol='TBCCD1', Aliases='', description='TBCC domain containing 1', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186546067, end_position_on_the_genomic_accession=186570543, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  55172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55172, CurrentID=0, Status='live', Symbol='DNAAF2', Aliases='C14orf104, CILD10, KTU, PF13', description='dynein, axonemal, assembly factor 2', other_designations='dynein assembly factor 2, axonemal|kintoun', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49625174, end_position_on_the_genomic_accession=49635230, orientation='minus', exon_count=3, OMIM=612517, no_hdr0=''),
  55173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55173, CurrentID=0, Status='live', Symbol='MRPS10', Aliases='MRP-S10, PNAS-122', description='mitochondrial ribosomal protein S10', other_designations='S10mt|mitochondrial 28S ribosomal protein S10', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42206801, end_position_on_the_genomic_accession=42217907, orientation='minus', exon_count=8, OMIM=611976, no_hdr0=''),
  55174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55174, CurrentID=0, Status='live', Symbol='INTS10', Aliases='C8orf35, INT10', description='integrator complex subunit 10', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19817149, end_position_on_the_genomic_accession=19852078, orientation='plus', exon_count=19, OMIM=611353, no_hdr0=''),
  55175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55175, CurrentID=0, Status='live', Symbol='KLHL11', Aliases='', description='kelch like family member 11', other_designations='kelch-like family member 11', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41848518, end_position_on_the_genomic_accession=41865431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55176, CurrentID=0, Status='live', Symbol='SEC61A2', Aliases='', description='Sec61 translocon alpha 2 subunit', other_designations='Sec61 alpha 2 subunit|Sec61 alpha form 2|protein transport protein Sec61 subunit alpha|sec61 alpha-2', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12129641, end_position_on_the_genomic_accession=12169958, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  55177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55177, CurrentID=0, Status='live', Symbol='RMDN3', Aliases='FAM82A2, FAM82C, RMD-3, RMD3, ptpip51', description='regulator of microtubule dynamics 3', other_designations='TCPTP-interacting protein 51|cerebral protein 10|family with sequence similarity 82, member A2|family with sequence similarity 82, member C|microtubule-associated protein|protein tyrosine phosphatase-interacting protein 51', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40735884, end_position_on_the_genomic_accession=40755336, orientation='minus', exon_count=14, OMIM=611873, no_hdr0=''),
  55178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55178, CurrentID=0, Status='live', Symbol='RNMTL1', Aliases='MRM3, RMTL1', description='RNA methyltransferase like 1', other_designations="16S rRNA (guanosine(1370)-2'-O)-methyltransferase|16S rRNA [Gm1370] 2'-O-methyltransferase|RNA methyltransferase-like protein 1|putative RNA methyltransferase", map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=782273, end_position_on_the_genomic_accession=792509, orientation='plus', exon_count=4, OMIM=612600, no_hdr0=''),
  55179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55179, CurrentID=0, Status='live', Symbol='FAIM', Aliases='FAIM1', description='Fas apoptotic inhibitory molecule', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138608594, end_position_on_the_genomic_accession=138633376, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55180, CurrentID=0, Status='live', Symbol='LINS1', Aliases='LINS, MRT27, WINS1', description='lines homolog 1', other_designations='WINS1 protein with Drosophila Lines (Lin) homologous domain|wnt-signaling molecule Lines homolog 1', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100569223, end_position_on_the_genomic_accession=100602240, orientation='minus', exon_count=8, OMIM=610350, no_hdr0=''),
  55181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55181, CurrentID=0, Status='live', Symbol='SMG8', Aliases='C17orf71', description='SMG8 nonsense mediated mRNA decay factor', other_designations='amplified in breast cancer gene 2 protein|protein smg-8 homolog|smg-8 homolog, nonsense mediated mRNA decay factor', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59210010, end_position_on_the_genomic_accession=59215250, orientation='plus', exon_count=4, OMIM=613175, no_hdr0=''),
  55182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55182, CurrentID=0, Status='live', Symbol='RNF220', Aliases='C1orf164', description='ring finger protein 220', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44404783, end_position_on_the_genomic_accession=44651724, orientation='plus', exon_count=22, OMIM=616136, no_hdr0=''),
  55183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55183, CurrentID=0, Status='live', Symbol='RIF1', Aliases='', description='replication timing regulatory factor 1', other_designations='RAP1 interacting factor homolog|rap1-interacting factor 1 homolog', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151409883, end_position_on_the_genomic_accession=151534202, orientation='plus', exon_count=40, OMIM=608952, no_hdr0=''),
  55184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55184, CurrentID=0, Status='live', Symbol='DZANK1', Aliases='ANKRD64, C20orf12, C20orf84, bA189K21.1, bA189K21.8, dJ568F9.2', description='double zinc ribbon and ankyrin repeat domains 1', other_designations='ankyrin repeat domain 64|ankyrin repeat-containing protein C20orf12', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18383367, end_position_on_the_genomic_accession=18467185, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  105369240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369240, CurrentID=0, Status='live', Symbol='LOC105369240', Aliases='', description='beta-defensin 131-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55186, CurrentID=0, Status='live', Symbol='SLC25A36', Aliases='PNC2', description='solute carrier family 25 (pyrimidine nucleotide carrier), member 36', other_designations='solute carrier family 25 (pyrimidine nucleotide carrier ), member 36|solute carrier family 25, member 36', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140941820, end_position_on_the_genomic_accession=140979943, orientation='plus', exon_count=12, OMIM=616149, no_hdr0=''),
  55187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55187, CurrentID=0, Status='live', Symbol='VPS13D', Aliases='', description='vacuolar protein sorting 13 homolog D (S. cerevisiae)', other_designations='vacuolar protein sorting 13D', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12230039, end_position_on_the_genomic_accession=12512047, orientation='plus', exon_count=70, OMIM=608877, no_hdr0=''),
  55188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55188, CurrentID=0, Status='live', Symbol='RIC8B', Aliases='RIC8, hSyn', description='RIC8 guanine nucleotide exchange factor B', other_designations='brain synembrin|brain synembryn|resistance to inhibitors of cholinesterase 8 homolog B', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106774285, end_position_on_the_genomic_accession=106889316, orientation='plus', exon_count=16, OMIM=609147, no_hdr0=''),
  55190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55190, CurrentID=0, Status='live', Symbol='NUDT11', Aliases='APS1, ASP1, DIPP3b, DIPP3beta, hDIPP3beta', description='nudix hydrolase 11', other_designations="DIPP-3-beta|DIPP3-beta|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 3-beta|diadenosine hexaphosphate hydrolase (AMP-forming)|diphosphoinositol polyphosphate phosphohydrolase 3, beta|hAps1|nucleoside diphosphate-linked moiety X motif 11|nudix (nucleoside diphosphate linked moiety X)-type motif 11|nudix motif 11", map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51490011, end_position_on_the_genomic_accession=51496607, orientation='minus', exon_count=2, OMIM=300528, no_hdr0=''),
  55191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55191, CurrentID=0, Status='live', Symbol='NADSYN1', Aliases='', description='NAD synthetase 1', other_designations='NAD(+) synthase|NAD(+) synthetase|glutamine-dependent NAD synthetase', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71453171, end_position_on_the_genomic_accession=71501538, orientation='plus', exon_count=21, OMIM=608285, no_hdr0=''),
  55192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55192, CurrentID=0, Status='live', Symbol='DNAJC17', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C17', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 17', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40767867, end_position_on_the_genomic_accession=40807478, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55193, CurrentID=0, Status='live', Symbol='PBRM1', Aliases='BAF180, PB1', description='polybromo 1', other_designations='BRG1-associated factor 180|polybromo-1D', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52545352, end_position_on_the_genomic_accession=52685898, orientation='minus', exon_count=37, OMIM=606083, no_hdr0=''),
  55194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55194, CurrentID=0, Status='live', Symbol='EVA1B', Aliases='C1orf78, FAM176B', description='eva-1 homolog B (C. elegans)', other_designations='family with sequence similarity 176, member B|protein FAM176B', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36322030, end_position_on_the_genomic_accession=36324154, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55195, CurrentID=0, Status='live', Symbol='C14orf105', Aliases='', description='chromosome 14 open reading frame 105', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57469300, end_position_on_the_genomic_accession=57493891, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  55196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55196, CurrentID=0, Status='live', Symbol='KIAA1551', Aliases='C12orf35, UTA2-1', description='KIAA1551', other_designations='uncharacterized protein C12orf35', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31959419, end_position_on_the_genomic_accession=31993107, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  55197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55197, CurrentID=0, Status='live', Symbol='RPRD1A', Aliases='HsT3101, P15RS', description='regulation of nuclear pre-mRNA domain containing 1A', other_designations='CDKN2B-related protein|Cyclin-dependent kinase inhibitor 2B-related protein (p15INK4B-related protein)|cyclin-dependent kinase 2B-inhibitor-related protein|p15INK4B-related protein', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35989824, end_position_on_the_genomic_accession=36067594, orientation='minus', exon_count=12, OMIM=610347, no_hdr0=''),
  55198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55198, CurrentID=0, Status='live', Symbol='APPL2', Aliases='DIP13B', description='adaptor protein, phosphotyrosine interaction, PH domain and leucine zipper containing 2', other_designations='DIP13 beta|adapter protein containing PH domain, PTB domain and leucine zipper motif 2', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105173297, end_position_on_the_genomic_accession=105236230, orientation='minus', exon_count=24, OMIM=606231, no_hdr0=''),
  55199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55199, CurrentID=0, Status='live', Symbol='FAM86C1', Aliases='FAM86C', description='family with sequence similarity 86 member C1', other_designations='family with sequence similarity 86, member C|family with sequence similarity 86, member C1', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71787511, end_position_on_the_genomic_accession=71801234, orientation='plus', exon_count=6, OMIM=616124, no_hdr0=''),
  55200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55200, CurrentID=0, Status='live', Symbol='PLEKHG6', Aliases='MyoGEF', description='pleckstrin homology and RhoGEF domain containing G6', other_designations='PH domain-containing family G member 6|myosin II interacting GEF|myosin interacting guanine nucleotide exchange factor|myosin-interacting guanine nucleotide exchange factor|pleckstrin homology domain containing, family G (with RhoGef domain) member 6', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6310436, end_position_on_the_genomic_accession=6328506, orientation='plus', exon_count=20, OMIM=611743, no_hdr0=''),
  55201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55201, CurrentID=0, Status='live', Symbol='MAP1S', Aliases='BPY2IP1, C19orf5, MAP8, VCY2IP-1, VCY2IP1', description='microtubule associated protein 1S', other_designations='BPY2 interacting protein 1|BPY2-interacting protein 1|MAP-1S|VCY2 interacting protein 1|VCY2-interacting protein 1|microtubule-associated protein 8|variable charge Y chromosome 2-interacting protein 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17719494, end_position_on_the_genomic_accession=17734515, orientation='plus', exon_count=8, OMIM=607573, no_hdr0=''),
  55203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55203, CurrentID=0, Status='live', Symbol='LGI2', Aliases='LGIL2', description='leucine-rich repeat LGI family member 2', other_designations='LGI1-like protein 2|leucine-rich glioma-inactivated protein 2|leucine-rich repeat LGI family, member 2|leucine-rich, glioma inactivated 2', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24998849, end_position_on_the_genomic_accession=25030792, orientation='minus', exon_count=9, OMIM=608301, no_hdr0=''),
  55204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55204, CurrentID=0, Status='live', Symbol='GOLPH3L', Aliases='GPP34R', description='golgi phosphoprotein 3-like', other_designations='GPP34-related protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150646225, end_position_on_the_genomic_accession=150697956, orientation='minus', exon_count=7, OMIM=612208, no_hdr0=''),
  55205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55205, CurrentID=0, Status='live', Symbol='ZNF532', Aliases='', description='zinc finger protein 532', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58862600, end_position_on_the_genomic_accession=58986477, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  55206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55206, CurrentID=0, Status='live', Symbol='SBNO1', Aliases='MOP3, Sno', description='strawberry notch homolog 1 (Drosophila)', other_designations='monocyte protein 3', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123289109, end_position_on_the_genomic_accession=123365525, orientation='minus', exon_count=34, OMIM=614274, no_hdr0=''),
  55207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55207, CurrentID=0, Status='live', Symbol='ARL8B', Aliases='ARL10C, Gie1', description='ADP ribosylation factor like GTPase 8B', other_designations='ADP-ribosylation factor like GTPase 8B|ADP-ribosylation factor-like 10C|ADP-ribosylation factor-like 8B|ADP-ribosylation factor-like protein 10C|novel small G protein indispensable for equal chromosome segregation 1', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5122245, end_position_on_the_genomic_accession=5180916, orientation='plus', exon_count=7, OMIM=616596, no_hdr0=''),
  55208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55208, CurrentID=0, Status='live', Symbol='DCUN1D2', Aliases='C13orf17', description='DCN1, defective in cullin neddylation 1, domain containing 2', other_designations='DCUN1 domain-containing protein 2|defective in cullin neddylation protein 1-like protein 2', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113455819, end_position_on_the_genomic_accession=113490709, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55209, CurrentID=0, Status='live', Symbol='SETD5', Aliases='', description='SET domain containing 5', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9397700, end_position_on_the_genomic_accession=9478154, orientation='plus', exon_count=33, OMIM=615743, no_hdr0=''),
  55210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55210, CurrentID=0, Status='live', Symbol='ATAD3A', Aliases='', description='ATPase family, AAA domain containing 3A', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1512143, end_position_on_the_genomic_accession=1534687, orientation='plus', exon_count=17, OMIM=612316, no_hdr0=''),
  55211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55211, CurrentID=0, Status='live', Symbol='DPPA4', Aliases='2410091M23Rik', description='developmental pluripotency associated 4', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109326141, end_position_on_the_genomic_accession=109337572, orientation='minus', exon_count=8, OMIM=614125, no_hdr0=''),
  55212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55212, CurrentID=0, Status='live', Symbol='BBS7', Aliases='BBS2L1', description='Bardet-Biedl syndrome 7', other_designations='BBS2-like 1', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121824329, end_position_on_the_genomic_accession=121870497, orientation='minus', exon_count=19, OMIM=607590, no_hdr0=''),
  55213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55213, CurrentID=0, Status='live', Symbol='RCBTB1', Aliases='CLLD7, CLLL7, GLP', description='regulator of chromosome condensation (RCC1) and BTB (POZ) domain containing protein 1', other_designations='CLL deletion region gene 7 protein|GDP/GTP exchange factor (GEF)-like protein|chronic lymphocytic leukemia deletion region gene 7 protein|regulator of chromosome condensation and BTB domain-containing protein 1', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49531944, end_position_on_the_genomic_accession=49585606, orientation='minus', exon_count=16, OMIM=607867, no_hdr0=''),
  55214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55214, CurrentID=0, Status='live', Symbol='P3H2', Aliases='LEPREL1, MCVD, MLAT4', description='prolyl 3-hydroxylase 2', other_designations='leprecan-like 1|myxoid liposarcoma-associated protein 4|prolyl 3-hydroxylase 3', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189956728, end_position_on_the_genomic_accession=190122437, orientation='minus', exon_count=17, OMIM=610341, no_hdr0=''),
  55215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55215, CurrentID=0, Status='live', Symbol='FANCI', Aliases='KIAA1794', description='Fanconi anemia complementation group I', other_designations='Fanconi anemia, complementation group I', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89243948, end_position_on_the_genomic_accession=89317131, orientation='plus', exon_count=44, OMIM=611360, no_hdr0=''),
  55216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55216, CurrentID=0, Status='live', Symbol='C11orf57', Aliases='', description='chromosome 11 open reading frame 57', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112074244, end_position_on_the_genomic_accession=112085150, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55217, CurrentID=0, Status='live', Symbol='TMLHE', Aliases='AUTSX6, BBOX2, TMLD, TMLHD, XAP130, TMLHE', description='trimethyllysine hydroxylase, epsilon', other_designations='TML hydroxylase|TML-alpha-ketoglutarate dioxygenase|butyrobetaine (gamma), 2-oxoglutarate dioxygenase (gamma-butyrobetaine hydroxylase) 2|epsilon-trimethyllysine 2-oxoglutarate dioxygenase|epsilon-trimethyllysine hydroxylase', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155489011, end_position_on_the_genomic_accession=155612961, orientation='minus', exon_count=10, OMIM=300777, no_hdr0=''),
  55218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55218, CurrentID=0, Status='live', Symbol='EXD2', Aliases='C14orf114, EXDL2', description="exonuclease 3'-5' domain containing 2", other_designations="exonuclease 3'-5' domain-like 2|exonuclease 3'-5' domain-like-containing protein 2", map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69191477, end_position_on_the_genomic_accession=69244020, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  55219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55219, CurrentID=0, Status='live', Symbol='TMEM57', Aliases='MACOILIN', description='transmembrane protein 57', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25430858, end_position_on_the_genomic_accession=25500209, orientation='plus', exon_count=11, OMIM=610301, no_hdr0=''),
  55220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55220, CurrentID=0, Status='live', Symbol='KLHDC8A', Aliases='', description='kelch domain containing 8A', other_designations='S-delta-E1|substitute for delta-EGFR expression 1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205336065, end_position_on_the_genomic_accession=205357090, orientation='minus', exon_count=9, OMIM=614503, no_hdr0=''),
  105369246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369246, CurrentID=0, Status='live', Symbol='LOC105369246', Aliases='', description='uncharacterized LOC105369246', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88089867, end_position_on_the_genomic_accession=88090794, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55222, CurrentID=0, Status='live', Symbol='LRRC20', Aliases='', description='leucine rich repeat containing 20', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70298970, end_position_on_the_genomic_accession=70382650, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  55223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55223, CurrentID=0, Status='live', Symbol='TRIM62', Aliases='DEAR1', description='tripartite motif containing 62', other_designations='ductal epithelium-associated RING Chromosome 1|tripartite motif-containing 62|tripartite motif-containing protein 62', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33145399, end_position_on_the_genomic_accession=33182070, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55224, CurrentID=0, Status='live', Symbol='ETNK2', Aliases='EKI2, HMFT1716', description='ethanolamine kinase 2', other_designations='ethanolamine kinase-like protein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204131061, end_position_on_the_genomic_accession=204152182, orientation='minus', exon_count=9, OMIM=609859, no_hdr0=''),
  55225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55225, CurrentID=0, Status='live', Symbol='RAVER2', Aliases='', description='ribonucleoprotein, PTB-binding 2', other_designations='protein raver-2', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64744964, end_position_on_the_genomic_accession=64833232, orientation='plus', exon_count=13, OMIM=609953, no_hdr0=''),
  55226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55226, CurrentID=0, Status='live', Symbol='NAT10', Aliases='ALP, NET43', description='N-acetyltransferase 10 (GCN5-related)', other_designations='N-acetyltransferase-like protein', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34105564, end_position_on_the_genomic_accession=34151034, orientation='plus', exon_count=30, OMIM=609221, no_hdr0=''),
  55227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55227, CurrentID=0, Status='live', Symbol='LRRC1', Aliases='LANO, dJ523E19.1', description='leucine rich repeat containing 1', other_designations='LANO adapter protein|LAP (leucine-rich repeats and PDZ) and no PDZ protein|LAP and no PDZ protein|leucine-rich repeat-containing 1', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53794497, end_position_on_the_genomic_accession=53924610, orientation='plus', exon_count=16, OMIM=608195, no_hdr0=''),
  55228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55228, CurrentID=0, Status='live', Symbol='PNMAL1', Aliases='', description='paraneoplastic Ma antigen family-like 1', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46466491, end_position_on_the_genomic_accession=46471563, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  55229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55229, CurrentID=0, Status='live', Symbol='PANK4', Aliases='', description='pantothenate kinase 4', other_designations='hPanK4|pantothenic acid kinase 4', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2508531, end_position_on_the_genomic_accession=2526628, orientation='minus', exon_count=20, OMIM=606162, no_hdr0=''),
  55230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55230, CurrentID=0, Status='live', Symbol='USP40', Aliases='', description='ubiquitin specific peptidase 40', other_designations='deubiquitinating enzyme 40|ubiquitin specific protease 40|ubiquitin thioesterase 40|ubiquitin thiolesterase 40|ubiquitin-specific-processing protease 40', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233475519, end_position_on_the_genomic_accession=233566787, orientation='minus', exon_count=35, OMIM=610570, no_hdr0=''),
  55231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55231, CurrentID=0, Status='live', Symbol='CCDC87', Aliases='', description='coiled-coil domain containing 87', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66590169, end_position_on_the_genomic_accession=66593083, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55233, CurrentID=0, Status='live', Symbol='MOB1A', Aliases='C2orf6, MATS1, MOB1, MOBK1B, MOBKL1B, Mob4B', description='MOB kinase activator 1A', other_designations='MOB1 Mps One Binder homolog A|MOB1, Mps One Binder kinase activator-like 1B|mob1 alpha|mob1 homolog 1B|mps one binder kinase activator-like 1B', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74154409, end_position_on_the_genomic_accession=74178886, orientation='minus', exon_count=7, OMIM=609281, no_hdr0=''),
  55234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55234, CurrentID=0, Status='live', Symbol='SMU1', Aliases='BWD, SMU-1, fSAP57', description='smu-1 suppressor of mec-8 and unc-52 homolog (C. elegans)', other_designations='brain-enriched WD-repeat protein|functional spliceosome-associated protein 57|smu-1 suppressor of mec-8 and unc-52 protein homolog', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33041852, end_position_on_the_genomic_accession=33076716, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55236, CurrentID=0, Status='live', Symbol='UBA6', Aliases='E1-L2, MOP-4, UBE1L2', description='ubiquitin-like modifier activating enzyme 6', other_designations='UBA6, ubiquitin-activating enzyme E1|monocyte protein 4|ubiquitin-activating enzyme 6|ubiquitin-activating enzyme E1-like 2|ubiquitin-activating enzyme E1-like protein 2', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67615761, end_position_on_the_genomic_accession=67701171, orientation='minus', exon_count=33, OMIM=611361, no_hdr0=''),
  55237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55237, CurrentID=0, Status='live', Symbol='VRTN', Aliases='C14orf115, vertnin', description='vertebrae development associated', other_designations='vertebrae development homolog (pig)', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74303076, end_position_on_the_genomic_accession=74360008, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55238, CurrentID=0, Status='live', Symbol='SLC38A7', Aliases='SNAT7', description='solute carrier family 38 member 7', other_designations='amino acid transporter|solute carrier family 38, member 7', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58665109, end_position_on_the_genomic_accession=58684777, orientation='minus', exon_count=14, OMIM=614236, no_hdr0=''),
  55239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55239, CurrentID=0, Status='live', Symbol='OGFOD1', Aliases='TPA1', description='2-oxoglutarate and iron-dependent oxygenase domain containing 1', other_designations='2-oxoglutarate and iron-dependent oxygenase domain-containing protein 1|TPA1, termination and polyadenylation 1, homolog|termination and polyadenylation 1 homolog', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56451417, end_position_on_the_genomic_accession=56477495, orientation='plus', exon_count=13, OMIM=615857, no_hdr0=''),
  55240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55240, CurrentID=0, Status='live', Symbol='STEAP3', Aliases='AHMIO2, STMP3, TSAP6, dudlin-2, dudulin-2, pHyde', description='STEAP3 metalloreductase', other_designations='STEAP family member 3, metalloreductase|six-transmembrane epithelial antigen of prostate 3|tumor suppressor pHyde|tumor suppressor-activated pathway protein 6', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119222731, end_position_on_the_genomic_accession=119265652, orientation='plus', exon_count=9, OMIM=609671, no_hdr0=''),
  55243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55243, CurrentID=0, Status='live', Symbol='KIRREL', Aliases='NEPH1', description='kin of IRRE like (Drosophila)', other_designations='kin of irregular chiasm-like protein 1|nephrin related|nephrin-like protein 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157993273, end_position_on_the_genomic_accession=158100262, orientation='plus', exon_count=16, OMIM=607428, no_hdr0=''),
  55244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55244, CurrentID=0, Status='live', Symbol='SLC47A1', Aliases='MATE1', description='solute carrier family 47 (multidrug and toxin extrusion), member 1', other_designations='MATE-1|hMATE-1|multidrug and toxin extrusion 1|solute carrier family 47, member 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19533854, end_position_on_the_genomic_accession=19579033, orientation='plus', exon_count=17, OMIM=609832, no_hdr0=''),
  55245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55245, CurrentID=0, Status='live', Symbol='UQCC1', Aliases='BFZB, C20orf44, CBP3, UQCC', description='ubiquinol-cytochrome c reductase complex assembly factor 1', other_designations='bFGF-repressed Zic-binding protein|basic FGF-repressed Zic-binding protein|cytochrome B protein synthesis 3 homolog|ubiquinol-cytochrome c reductase complex chaperone CBP3 homolog', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35302566, end_position_on_the_genomic_accession=35412142, orientation='minus', exon_count=12, OMIM=611797, no_hdr0=''),
  55246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55246, CurrentID=0, Status='live', Symbol='CCDC25', Aliases='', description='coiled-coil domain containing 25', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27733316, end_position_on_the_genomic_accession=27772655, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55247, CurrentID=0, Status='live', Symbol='NEIL3', Aliases='FGP2, FPG2, NEI3, ZGRF3, hFPG2, hNEI3', description='nei-like DNA glycosylase 3', other_designations='DNA glycosylase FPG2|DNA glycosylase hFPG2|DNA glycosylase/AP lyase Neil3|endonuclease VIII-like 3|nei endonuclease VIII-like 3|nei-like protein 3|zinc finger, GRF-type containing 3', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177309837, end_position_on_the_genomic_accession=177362938, orientation='plus', exon_count=10, OMIM=608934, no_hdr0=''),
  55248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55248, CurrentID=0, Status='live', Symbol='TMEM206', Aliases='C1orf75', description='transmembrane protein 206', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212364474, end_position_on_the_genomic_accession=212415508, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  55249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55249, CurrentID=0, Status='live', Symbol='YY1AP1', Aliases='HCCA1, HCCA2, YAP, YY1AP', description='YY1 associated protein 1', other_designations='hepatocellular carcinoma susceptibility protein|hepatocellular carcinoma-associated protein 2', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155659442, end_position_on_the_genomic_accession=155689032, orientation='minus', exon_count=14, OMIM=607860, no_hdr0=''),
  55250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55250, CurrentID=0, Status='live', Symbol='ELP2', Aliases='SHINC-2, STATIP1, StIP', description='elongator acetyltransferase complex subunit 2', other_designations='STAT3-interacting protein 1|elongation protein 2 homolog|elongator protein 2|signal transducer and activator of transcription 3 interacting protein 1|signal transducer and activator of transcription interacting protein 1|stIP1', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36129874, end_position_on_the_genomic_accession=36174725, orientation='plus', exon_count=23, OMIM=616054, no_hdr0=''),
  55251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55251, CurrentID=0, Status='live', Symbol='PCMTD2', Aliases='C20orf36', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 2', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64255695, end_position_on_the_genomic_accession=64276226, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55252, CurrentID=0, Status='live', Symbol='ASXL2', Aliases='ASXH2', description='additional sex combs like 2, transcriptional regulator', other_designations='additional sex combs like 2|additional sex combs like transcriptional regulator 2|additional sex combs-like protein 2|polycomb group protein ASXH2', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25733753, end_position_on_the_genomic_accession=25878474, orientation='minus', exon_count=16, OMIM=612991, no_hdr0=''),
  55253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55253, CurrentID=0, Status='live', Symbol='TYW1', Aliases='RSAFD1A, YPL207W, TYW1', description='tRNA-yW synthesizing protein 1 homolog', other_designations='radical S-adenosyl methionine and flavodoxin domain-containing protein 1|radical S-adenosyl methionine and flavodoxin domains 1|tRNA wybutosine-synthesizing protein 1 homolog|tRNA-yW synthesizing protein 1 homolog A', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66996805, end_position_on_the_genomic_accession=67239520, orientation='plus', exon_count=19, OMIM=611243, no_hdr0=''),
  55254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55254, CurrentID=0, Status='live', Symbol='TMEM39A', Aliases='', description='transmembrane protein 39A', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119428960, end_position_on_the_genomic_accession=119463682, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55255, CurrentID=0, Status='live', Symbol='WDR41', Aliases='MSTP048', description='WD repeat domain 41', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77430933, end_position_on_the_genomic_accession=77620611, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  55256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55256, CurrentID=0, Status='live', Symbol='ADI1', Aliases='APL1, ARD, Fe-ARD, HMFT1638, MTCBP1, Ni-ARD, SIPL, mtnD', description='acireductone dioxygenase 1', other_designations='MT1-MMP cytoplasmic tail-binding protein-1|acireductone dioxygenase (Fe(2+)-requiring)|acireductone dioxygenase (Ni(2+)-requiring)|membrane-type 1 matrix metalloproteinase cytoplasmic tail binding protein-1|submergence induced protein 2|submergence-induced protein-like factor', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3497919, end_position_on_the_genomic_accession=3519579, orientation='minus', exon_count=5, OMIM=613400, no_hdr0=''),
  55257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55257, CurrentID=0, Status='live', Symbol='MRGBP', Aliases='C20orf20, Eaf7, MRG15BP, URCC4', description='MRG/MORF4L binding protein', other_designations='MRG(MORF4-related gene)-binding protein|MRG-binding protein|up-regulated in colon cancer 4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62796453, end_position_on_the_genomic_accession=62800593, orientation='plus', exon_count=5, OMIM=611157, no_hdr0=''),
  55258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55258, CurrentID=0, Status='live', Symbol='THNSL2', Aliases='SOFAT, THS2, TSH2', description='threonine synthase-like 2', other_designations='secreted osteoclastogenic factor of activated T cells', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88170295, end_position_on_the_genomic_accession=88186637, orientation='plus', exon_count=10, OMIM=611261, no_hdr0=''),
  55259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55259, CurrentID=0, Status='live', Symbol='CASC1', Aliases='LAS1, PPP1R54', description='cancer susceptibility candidate 1', other_designations='cancer susceptibility candidate gene 1 protein|cancer susceptibility candidate protein 1|lung adenoma susceptibility 1-like protein|protein phosphatase 1, regulatory subunit 54', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25107057, end_position_on_the_genomic_accession=25195162, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55260, CurrentID=0, Status='live', Symbol='TMEM143', Aliases='', description='transmembrane protein 143', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48332356, end_position_on_the_genomic_accession=48364237, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55262, CurrentID=0, Status='live', Symbol='C7orf43', Aliases='', description='chromosome 7 open reading frame 43', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100154420, end_position_on_the_genomic_accession=100158721, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105369253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369253, CurrentID=0, Status='live', Symbol='LOC105369253', Aliases='', description='rootletin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55266, CurrentID=0, Status='live', Symbol='TMEM19', Aliases='', description='transmembrane protein 19', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71686098, end_position_on_the_genomic_accession=71704060, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55267, CurrentID=0, Status='live', Symbol='PRR34', Aliases='C22orf26', description='proline rich 34', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46050459, end_position_on_the_genomic_accession=46054144, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55268, CurrentID=0, Status='live', Symbol='ECHDC2', Aliases='', description='enoyl-CoA hydratase domain containing 2', other_designations='enoyl CoA hydratase domain containing 2|enoyl Coenzyme A hydratase domain containing 2', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52895910, end_position_on_the_genomic_accession=52927212, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  55269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55269, CurrentID=0, Status='live', Symbol='PSPC1', Aliases='PSP1', description='paraspeckle component 1', other_designations='paraspeckle protein 1', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19674752, end_position_on_the_genomic_accession=19783019, orientation='minus', exon_count=17, OMIM=612408, no_hdr0=''),
  55270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55270, CurrentID=0, Status='live', Symbol='NUDT15', Aliases='MTH2', description='nudix hydrolase 15', other_designations='8-oxo-dGTPase NUDT15|mutT homolog 2|nudix (nucleoside diphosphate linked moiety X)-type motif 15|probable 7,8-dihydro-8-oxoguanine triphosphatase NUDT15', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48037567, end_position_on_the_genomic_accession=48047222, orientation='plus', exon_count=3, OMIM=615792, no_hdr0=''),
  55272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55272, CurrentID=0, Status='live', Symbol='IMP3', Aliases='BRMS2, C15orf12, MRPS4', description='IMP3, U3 small nucleolar ribonucleoprotein', other_designations='IMP3, U3 small nucleolar ribonucleoprotein, homolog|U3 snoRNP protein 3 homolog|U3 snoRNP protein IMP3', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75639085, end_position_on_the_genomic_accession=75640323, orientation='minus', exon_count=1, OMIM=612980, no_hdr0=''),
  55273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55273, CurrentID=0, Status='live', Symbol='TMEM100', Aliases='', description='transmembrane protein 100', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55719627, end_position_on_the_genomic_accession=55732121, orientation='minus', exon_count=4, OMIM=616334, no_hdr0=''),
  55274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55274, CurrentID=0, Status='live', Symbol='PHF10', Aliases='BAF45A, XAP135', description='PHD finger protein 10', other_designations='BRG1-associated factor 45a|BRG1-associated factor, 45-KD, A|PHD zinc finger protein XAP135', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169703905, end_position_on_the_genomic_accession=169724010, orientation='minus', exon_count=12, OMIM=613069, no_hdr0=''),
  55275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55275, CurrentID=0, Status='live', Symbol='VPS53', Aliases='HCCS1, PCH2E, hVps53L, pp13624', description='VPS53 GARP complex subunit', other_designations='vacuolar protein sorting 53 homolog', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=508668, end_position_on_the_genomic_accession=714856, orientation='minus', exon_count=24, OMIM=615850, no_hdr0=''),
  55276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55276, CurrentID=0, Status='live', Symbol='PGM2', Aliases='MSTP006', description='phosphoglucomutase 2', other_designations='PGM 2|glucose phosphomutase 2|phosphodeoxyribomutase', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37826660, end_position_on_the_genomic_accession=37862938, orientation='plus', exon_count=15, OMIM=172000, no_hdr0=''),
  55277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55277, CurrentID=0, Status='live', Symbol='FGGY', Aliases='', description='FGGY carbohydrate kinase domain containing', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59296405, end_position_on_the_genomic_accession=59762730, orientation='plus', exon_count=26, OMIM=611370, no_hdr0=''),
  55278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55278, CurrentID=0, Status='live', Symbol='QRSL1', Aliases='GatA', description='glutaminyl-tRNA synthase (glutamine-hydrolyzing)-like 1', other_designations='glu-AdT subunit A|glutaminyl-tRNA synthase-like protein 1|glutamyl-tRNA(Gln) amidotransferase subunit A homolog|glutamyl-tRNA(Gln) amidotransferase, subunit A', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106629566, end_position_on_the_genomic_accession=106668417, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  55279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55279, CurrentID=0, Status='live', Symbol='ZNF654', Aliases='', description='zinc finger protein 654', other_designations='melanoma-associated antigen', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88139112, end_position_on_the_genomic_accession=88144664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55280, CurrentID=0, Status='live', Symbol='CWF19L1', Aliases='C19L1, SCAR17', description='CWF19-like 1, cell cycle control (S. pombe)', other_designations='CWF19-like 1 cell cycle control', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100232296, end_position_on_the_genomic_accession=100267681, orientation='minus', exon_count=14, OMIM=616120, no_hdr0=''),
  55281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55281, CurrentID=0, Status='live', Symbol='TMEM140', Aliases='', description='transmembrane protein 140', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135148014, end_position_on_the_genomic_accession=135166215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55282, CurrentID=0, Status='live', Symbol='LRRC36', Aliases='RORBP70, XLHSRF2', description='leucine rich repeat containing 36', other_designations='ROR gamma binding protein 70|ROR gamma-binding protein 70|heat shock regulated 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67326844, end_position_on_the_genomic_accession=67385206, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55283, CurrentID=0, Status='live', Symbol='MCOLN3', Aliases='TRP-ML3, TRPML3', description='mucolipin 3', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85018082, end_position_on_the_genomic_accession=85048902, orientation='minus', exon_count=17, OMIM=607400, no_hdr0=''),
  55284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55284, CurrentID=0, Status='live', Symbol='UBE2W', Aliases='UBC-16, UBC16', description='ubiquitin conjugating enzyme E2W (putative)', other_designations='E2 ubiquitin-conjugating enzyme W|N-terminal E2 ubiquitin-conjugating enzyme|N-terminus-conjugating E2|ubiquitin carrier protein W|ubiquitin-conjugating enzyme 16|ubiquitin-conjugating enzyme 2W|ubiquitin-conjugating enzyme E2W (putative)|ubiquitin-protein ligase W', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73780097, end_position_on_the_genomic_accession=73878910, orientation='minus', exon_count=10, OMIM=614277, no_hdr0=''),
  55285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55285, CurrentID=0, Status='live', Symbol='RBM41', Aliases='', description='RNA binding motif protein 41', other_designations='RNA-binding motif protein 41', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107061885, end_position_on_the_genomic_accession=107118827, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55286, CurrentID=0, Status='live', Symbol='C4orf19', Aliases='', description='chromosome 4 open reading frame 19', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37453454, end_position_on_the_genomic_accession=37593510, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55287, CurrentID=0, Status='live', Symbol='TMEM40', Aliases='', description='transmembrane protein 40', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12733530, end_position_on_the_genomic_accession=12769457, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  55288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55288, CurrentID=0, Status='live', Symbol='RHOT1', Aliases='ARHT1, MIRO-1, MIRO1', description='ras homolog family member T1', other_designations='mitochondrial Rho (MIRO) GTPase 1|rac-GTP binding protein-like protein|ras homolog gene family, member T1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32142454, end_position_on_the_genomic_accession=32253374, orientation='plus', exon_count=28, OMIM=613888, no_hdr0=''),
  55289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55289, CurrentID=0, Status='live', Symbol='ACOXL', Aliases='', description='acyl-CoA oxidase-like', other_designations='acyl-CoA oxidase-like protein', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110732539, end_position_on_the_genomic_accession=111120610, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  55290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55290, CurrentID=0, Status='live', Symbol='BRF2', Aliases='BRFU, TFIIIB50', description='BRF2, RNA polymerase III transcription initiation factor 50 kDa subunit', other_designations='B-related factor 2|BRF-2|BRF2, subunit of RNA polymerase III transcription initiation factor, BRF1-like|RNA polymerase III transcription initiation factor BRFU|hBRFU|hTFIIIB50|transcription factor IIB- related factor, TFIIIB50', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37843880, end_position_on_the_genomic_accession=37849913, orientation='minus', exon_count=4, OMIM=607013, no_hdr0=''),
  55291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55291, CurrentID=0, Status='live', Symbol='PPP6R3', Aliases='C11orf23, PP6R3, SAP190, SAPL, SAPLa, SAPS3', description='protein phosphatase 6 regulatory subunit 3', other_designations='SAPS domain family, member 3|protein phosphatase 6, regulatory subunit 3|sporulation-induced transcript 4-associated protein SAPL', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68460718, end_position_on_the_genomic_accession=68615334, orientation='plus', exon_count=30, OMIM=610879, no_hdr0=''),
  55293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55293, CurrentID=0, Status='live', Symbol='UEVLD', Aliases='ATTP, UEV3', description='UEV and lactate/malate dehyrogenase domains', other_designations='EV and lactate/malate dehydrogenase domain-containing protein|UEV-3|signaling molecule ATTP', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18529609, end_position_on_the_genomic_accession=18588938, orientation='minus', exon_count=15, OMIM=610985, no_hdr0=''),
  55294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55294, CurrentID=0, Status='live', Symbol='FBXW7', Aliases='AGO, CDC4, FBW6, FBW7, FBX30, FBXO30, FBXW6, SEL-10, SEL10, hAgo, hCdc4', description='F-box and WD repeat domain containing 7', other_designations='F-box and WD repeat domain containing 7, E3 ubiquitin protein ligase|F-box and WD-40 domain protein 7 (archipelago homolog, Drosophila)|F-box protein FBW7|F-box protein FBX30|F-box protein SEL-10|archipelago|homolog of C elegans sel-10', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152321258, end_position_on_the_genomic_accession=152536095, orientation='minus', exon_count=16, OMIM=606278, no_hdr0=''),
  55295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55295, CurrentID=0, Status='live', Symbol='KLHL26', Aliases='', description='kelch like family member 26', other_designations='kelch-like 26|kelch-like family member 26', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18637028, end_position_on_the_genomic_accession=18670492, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55296, CurrentID=0, Status='live', Symbol='TBC1D19', Aliases='', description='TBC1 domain family member 19', other_designations='TBC1 domain family, member 19', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26583924, end_position_on_the_genomic_accession=26769771, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  55297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55297, CurrentID=0, Status='live', Symbol='CCDC91', Aliases='HSD8, p56', description='coiled-coil domain containing 91', other_designations='GGA binding partner|GGA-binding partner|p56 accessory protein', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28188259, end_position_on_the_genomic_accession=28550166, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55298, CurrentID=0, Status='live', Symbol='RNF121', Aliases='', description='ring finger protein 121', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71928701, end_position_on_the_genomic_accession=71997597, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  55299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55299, CurrentID=0, Status='live', Symbol='BRIX1', Aliases='BRIX, BXDC2', description='BRX1, biogenesis of ribosomes', other_designations='BRX1, biogenesis of ribosomes, homolog|brix domain containing 2|brix domain-containing protein 2', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34915715, end_position_on_the_genomic_accession=34925682, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55300, CurrentID=0, Status='live', Symbol='PI4K2B', Aliases='PI4KIIB, PIK42B', description='phosphatidylinositol 4-kinase type 2 beta', other_designations='PI4KII-BETA|phosphatidylinositol 4-kinase type II-beta|phosphatidylinositol 4-kinase type-II beta', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25234031, end_position_on_the_genomic_accession=25279209, orientation='plus', exon_count=11, OMIM=612101, no_hdr0=''),
  55301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55301, CurrentID=0, Status='live', Symbol='OLAH', Aliases='AURA1, SAST, THEDC1', description='oleoyl-ACP hydrolase', other_designations='augmented in rheumatoid arthritis 1|thioesterase II|thioesterase domain containing 1|thioesterase domain-containing protein 1', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15043896, end_position_on_the_genomic_accession=15073852, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55303, CurrentID=0, Status='live', Symbol='GIMAP4', Aliases='IAN-1, IAN1, IMAP4, MSTP062', description='GTPase, IMAP family member 4', other_designations='human immune associated nucleotide 1|immune-associated nucleotide-binding protein 1|immunity associated protein 4|immunity-associated nucleotide 1 protein|immunity-associated protein 4', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150567370, end_position_on_the_genomic_accession=150573955, orientation='plus', exon_count=3, OMIM=608087, no_hdr0=''),
  55304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55304, CurrentID=0, Status='live', Symbol='SPTLC3', Aliases='C20orf38, LCB2B, LCB3, SPT3, SPTLC2L, dJ718P11, dJ718P11.1, hLCB2b', description='serine palmitoyltransferase, long chain base subunit 3', other_designations='LCB 3|SPT 3|long chain base biosynthesis protein 2b|long chain base biosynthesis protein 3|serine palmitoyltransferase, long chain base subunit 2-like (aminotransferase 2)|serine-palmitoyl-CoA transferase 3', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13008979, end_position_on_the_genomic_accession=13170944, orientation='plus', exon_count=15, OMIM=611120, no_hdr0=''),
  55308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55308, CurrentID=0, Status='live', Symbol='DDX19A', Aliases='DDX19-DDX19L, DDX19L', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 19A', other_designations='DDX19-like protein|DEAD (Asp-Glu-Ala-As) box polypeptide 19A|DEAD box protein 19A|RNA helicase', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70346861, end_position_on_the_genomic_accession=70373383, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  140290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140290, CurrentID=0, Status='live', Symbol='TCP10L', Aliases='C21orf77, PRED77, TCP10A-2', description='t-complex 10-like', other_designations='T-complex 10A-2|T-complex protein 10A-2|TCP10-like|t-complex 10 (a murine tcp homolog)-like', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32574841, end_position_on_the_genomic_accession=32585535, orientation='minus', exon_count=5, OMIM=608365, no_hdr0=''),
  55311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55311, CurrentID=0, Status='live', Symbol='ZNF444', Aliases='EZF-2, EZF2, ZSCAN17', description='zinc finger protein 444', other_designations='endothelial zinc finger protein 2|zinc finger and SCAN domain-containing protein 17', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56140951, end_position_on_the_genomic_accession=56160893, orientation='plus', exon_count=9, OMIM=607874, no_hdr0=''),
  55312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55312, CurrentID=0, Status='live', Symbol='RFK', Aliases='RIFK', description='riboflavin kinase', other_designations="0610038L10Rik|ATP:riboflavin 5'-phosphotransferase|flavokinase", map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76385517, end_position_on_the_genomic_accession=76394528, orientation='minus', exon_count=4, OMIM=613010, no_hdr0=''),
  55313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55313, CurrentID=0, Status='live', Symbol='CPPED1', Aliases='CSTP1', description='calcineurin-like phosphoesterase domain containing 1', other_designations='calcineurin-like phosphoesterase domain-containing protein 1|complete S transactivated protein 1|complete S-transactivated protein 1', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12659799, end_position_on_the_genomic_accession=12803887, orientation='minus', exon_count=4, OMIM=615603, no_hdr0=''),
  55314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55314, CurrentID=0, Status='live', Symbol='TMEM144', Aliases='', description='transmembrane protein 144', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158201604, end_position_on_the_genomic_accession=158255287, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  55315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55315, CurrentID=0, Status='live', Symbol='SLC29A3', Aliases='ENT3, HCLAP, HJCD, PHID', description='solute carrier family 29 (equilibrative nucleoside transporter), member 3', other_designations='solute carrier family 29 (nucleoside transporters), member 3', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71319253, end_position_on_the_genomic_accession=71363396, orientation='plus', exon_count=6, OMIM=612373, no_hdr0=''),
  55316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55316, CurrentID=0, Status='live', Symbol='RSAD1', Aliases='', description='radical S-adenosyl methionine domain containing 1', other_designations='oxygen-independent coproporphyrinogen-III oxidase-like protein RSAD1', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50478800, end_position_on_the_genomic_accession=50502164, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  55317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55317, CurrentID=0, Status='live', Symbol='AP5S1', Aliases='C20orf29', description='adaptor related protein complex 5, sigma 1 subunit', other_designations='adapter-related protein complex 5 sigma subunit|adaptor-related protein complex 5 sigma subunit|adaptor-related protein complex 5, sigma 1 subunit|protein C20orf29|sigma5', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3820524, end_position_on_the_genomic_accession=3825307, orientation='plus', exon_count=3, OMIM=614824, no_hdr0=''),
  55319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55319, CurrentID=0, Status='live', Symbol='TMA16', Aliases='C4orf43', description='translation machinery associated 16 homolog', other_designations='UPF0534 protein C4orf43', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163494289, end_position_on_the_genomic_accession=163520539, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55320, CurrentID=0, Status='live', Symbol='MIS18BP1', Aliases='C14orf106, HSA242977, KNL2, M18BP1', description='MIS18 binding protein 1', other_designations='P243|hsKNL-2|kinetochore null 2 homolog|kinetochore-associated protein KNL-2 homolog|putative protein p243 which interacts with transcription factor Sp1', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45203190, end_position_on_the_genomic_accession=45253402, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  55321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55321, CurrentID=0, Status='live', Symbol='TMEM74B', Aliases='C20orf46', description='transmembrane protein 74B', other_designations='transmembrane protein C20orf46', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1180561, end_position_on_the_genomic_accession=1188128, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55322, CurrentID=0, Status='live', Symbol='C5orf22', Aliases='', description='chromosome 5 open reading frame 22', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31532266, end_position_on_the_genomic_accession=31555058, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55323, CurrentID=0, Status='live', Symbol='LARP6', Aliases='ACHN', description='La ribonucleoprotein domain family member 6', other_designations='La ribonucleoprotein domain family, member 6|acheron|death-associated LA motif protein', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70831524, end_position_on_the_genomic_accession=70854265, orientation='minus', exon_count=5, OMIM=611300, no_hdr0=''),
  55324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55324, CurrentID=0, Status='live', Symbol='ABCF3', Aliases='EST201864', description='ATP binding cassette subfamily F member 3', other_designations='ATP-binding cassette, sub-family F (GCN20), member 3', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184186075, end_position_on_the_genomic_accession=184194452, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55325, CurrentID=0, Status='live', Symbol='UFSP2', Aliases='BHD, C4orf20', description='UFM1-specific peptidase 2', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185399540, end_position_on_the_genomic_accession=185425985, orientation='minus', exon_count=13, OMIM=611482, no_hdr0=''),
  55326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55326, CurrentID=0, Status='live', Symbol='AGPAT5', Aliases='1AGPAT5, LPAATE', description='1-acylglycerol-3-phosphate O-acyltransferase 5', other_designations='1-AGP acyltransferase 5|1-AGPAT 5|1-acylglycerol-3-phosphate O-acyltransferase 5 (lysophosphatidic acid acyltransferase, epsilon)|lysophosphatidic acid acyltransferase epsilon', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6708357, end_position_on_the_genomic_accession=6761503, orientation='plus', exon_count=8, OMIM=614796, no_hdr0=''),
  55327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55327, CurrentID=0, Status='live', Symbol='LIN7C', Aliases='LIN-7-C, LIN-7C, MALS-3, MALS3, VELI3', description='lin-7 homolog C (C. elegans)', other_designations='LIN-7 protein 3|mammalian lin-seven protein 3|veli-3|vertebrate lin-7 homolog 3', map_location='11p14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27494418, end_position_on_the_genomic_accession=27506779, orientation='minus', exon_count=5, OMIM=612332, no_hdr0=''),
  55328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55328, CurrentID=0, Status='live', Symbol='RNLS', Aliases='C10orf59, RENALASE', description='renalase, FAD-dependent amine oxidase', other_designations='MAO-C|alpha-NAD(P)H oxidase/anomerase|monoamine oxidase-C', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88180643, end_position_on_the_genomic_accession=88583854, orientation='minus', exon_count=13, OMIM=609360, no_hdr0=''),
  55329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55329, CurrentID=0, Status='live', Symbol='MNS1', Aliases='SPATA40', description='meiosis specific nuclear structural 1', other_designations='meiosis-specific nuclear structural 1|spermatogenesis associated 40', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56428731, end_position_on_the_genomic_accession=56465137, orientation='minus', exon_count=10, OMIM=610766, no_hdr0=''),
  55330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55330, CurrentID=0, Status='live', Symbol='BLOC1S4', Aliases='BCAS4L, BLOS4, CNO', description='biogenesis of lysosomal organelles complex-1, subunit 4, cappuccino', other_designations='BLOC-1 subunit 4|cappuccino homolog|protein cappuccino homolog', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6716115, end_position_on_the_genomic_accession=6717660, orientation='plus', exon_count=1, OMIM=605695, no_hdr0=''),
  55331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55331, CurrentID=0, Status='live', Symbol='ACER3', Aliases='APHC, PHCA', description='alkaline ceramidase 3', other_designations='alkCDase 3|alkaline CDase 3|alkaline dihydroceramidase SB89|alkaline phytoceramidase|phytoceramidase, alkaline', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76860873, end_position_on_the_genomic_accession=77026797, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55332, CurrentID=0, Status='live', Symbol='DRAM1', Aliases='DRAM', description='DNA damage regulated autophagy modulator 1', other_designations='DNA-damage regulated autophagy modulator 1|damage-regulated autophagy modulator', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101877327, end_position_on_the_genomic_accession=101923623, orientation='plus', exon_count=8, OMIM=610776, no_hdr0=''),
  55333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55333, CurrentID=0, Status='live', Symbol='SYNJ2BP', Aliases='ARIP2, OMP25', description='synaptojanin 2 binding protein', other_designations='activin receptor interacting protein 5|mitochondrial outer membrane protein 25', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70366496, end_position_on_the_genomic_accession=70417090, orientation='minus', exon_count=4, OMIM=609411, no_hdr0=''),
  55334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55334, CurrentID=0, Status='live', Symbol='SLC39A9', Aliases='ZIP-9, ZIP9', description='solute carrier family 39 member 9', other_designations='solute carrier family 39 (zinc transporter), member 9|solute carrier family 39, member 9|zinc transporter SLC39A9|zrt- and Irt-like protein 9', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69398379, end_position_on_the_genomic_accession=69468403, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55335, CurrentID=0, Status='live', Symbol='NIPSNAP3B', Aliases='FP944, NIPSNAP3, SNAP1', description='nipsnap homolog 3B (C. elegans)', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104763741, end_position_on_the_genomic_accession=104777764, orientation='plus', exon_count=7, OMIM=608872, no_hdr0=''),
  55336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55336, CurrentID=0, Status='live', Symbol='FBXL8', Aliases='FBL8', description='F-box and leucine-rich repeat protein 8', other_designations='F-box protein FBL8', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67159988, end_position_on_the_genomic_accession=67164174, orientation='plus', exon_count=3, OMIM=609077, no_hdr0=''),
  55337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55337, CurrentID=0, Status='live', Symbol='C19orf66', Aliases='', description='chromosome 19 open reading frame 66', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10086119, end_position_on_the_genomic_accession=10093252, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  55338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55338, CurrentID=0, Status='live', Symbol='LOC55338', Aliases='', description='uncharacterized LOC55338', other_designations='CTC-242N15.1', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87215735, end_position_on_the_genomic_accession=87239201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  55339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55339, CurrentID=0, Status='live', Symbol='WDR33', Aliases='NET14, WDC146', description='WD repeat domain 33', other_designations='WD repeat-containing protein 33|WD repeat-containing protein WDC146', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127704234, end_position_on_the_genomic_accession=127811187, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  55340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55340, CurrentID=0, Status='live', Symbol='GIMAP5', Aliases='HIMAP3, IAN-5, IAN4, IAN4L1, IAN5, IMAP3, IROD', description='GTPase, IMAP family member 5', other_designations='immune associated nucleotide 4 like 1|immune-associated nucleotide-binding protein 5|immunity associated protein 3|immunity-associated nucleotide 5 protein|inhibitor of radiation- and OA-induced apoptosis, Irod/Ian5', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150737348, end_position_on_the_genomic_accession=150743649, orientation='plus', exon_count=3, OMIM=608086, no_hdr0=''),
  55341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55341, CurrentID=0, Status='live', Symbol='LSG1', Aliases='', description='large 60S subunit nuclear export GTPase 1', other_designations='hLsg1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194640788, end_position_on_the_genomic_accession=194672477, orientation='minus', exon_count=14, OMIM=610780, no_hdr0=''),
  55342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55342, CurrentID=0, Status='live', Symbol='STRBP', Aliases='HEL162, ILF3L, SPNR, p74', description='spermatid perinuclear RNA binding protein', other_designations='epididymis luminal protein 162', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123109494, end_position_on_the_genomic_accession=123268576, orientation='minus', exon_count=26, OMIM=611138, no_hdr0=''),
  55343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55343, CurrentID=0, Status='live', Symbol='SLC35C1', Aliases='CDG2C, FUCT1', description='solute carrier family 35 (GDP-fucose transporter), member C1', other_designations='solute carrier family 35, member C1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45804072, end_position_on_the_genomic_accession=45813016, orientation='plus', exon_count=5, OMIM=605881, no_hdr0=''),
  55344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55344, CurrentID=0, Status='live', Symbol='PLCXD1', Aliases='LL0XNC01-136G2.1', description='phosphatidylinositol-specific phospholipase C, X domain containing 1', other_designations='', map_location='Xp22.33; Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=276324, end_position_on_the_genomic_accession=303356, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  55345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55345, CurrentID=0, Status='live', Symbol='ZGRF1', Aliases='C4orf21', description='zinc finger, GRF-type containing 1', other_designations='GRF-type zinc finger domain-containing protein 1', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112539333, end_position_on_the_genomic_accession=112636995, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  55346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55346, CurrentID=0, Status='live', Symbol='TCP11L1', Aliases='dJ85M6.3', description='t-complex 11, testis-specific-like 1', other_designations='t-complex 11-like 1', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33039417, end_position_on_the_genomic_accession=33079454, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  55347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55347, CurrentID=0, Status='live', Symbol='ABHD10', Aliases='', description='abhydrolase domain containing 10', other_designations='abhydrolase domain-containing protein 10, mitochondrial|alpha/beta hydrolase domain-containing protein 10, mitochondrial', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111978876, end_position_on_the_genomic_accession=111993368, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55349, CurrentID=0, Status='live', Symbol='CHDH', Aliases='', description='choline dehydrogenase', other_designations='CDH|CHD', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53816297, end_position_on_the_genomic_accession=53846393, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55350, CurrentID=0, Status='live', Symbol='VNN3', Aliases='HSA238982', description='vanin 3', other_designations='pantetheinase pseudogene|pantetheinase-associated gene expressed in leukocytes (PAGEL)-alpha|vanin 1 pseudogene|vanin-3', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132722787, end_position_on_the_genomic_accession=132734765, orientation='minus', exon_count=7, OMIM=606592, no_hdr0=''),
  55351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55351, CurrentID=0, Status='live', Symbol='STK32B', Aliases='HSA250839, STK32, STKG6, YANK2', description='serine/threonine kinase 32B', other_designations='gene for serine/threonine protein kinase|yet another novel kinase 2', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5051211, end_position_on_the_genomic_accession=5500998, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  55352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55352, CurrentID=0, Status='live', Symbol='COPRS', Aliases='C17orf79, COPR5, HSA272196, TTP1', description='coordinator of PRMT5, differentiation stimulator', other_designations='cooperator of PRMT5', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31851865, end_position_on_the_genomic_accession=31859307, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  55353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55353, CurrentID=0, Status='live', Symbol='LAPTM4B', Aliases='LAPTM4beta, LC27', description='lysosomal protein transmembrane 4 beta', other_designations='lysosomal associated protein transmembrane 4 beta|lysosome-associated transmembrane protein 4-beta', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97775581, end_position_on_the_genomic_accession=97852602, orientation='plus', exon_count=7, OMIM=613296, no_hdr0=''),
  55354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55354, CurrentID=0, Status='live', Symbol='ORF1', Aliases='', description='uncharacterized protein, clone pT-Adv JuaX22', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55355, CurrentID=0, Status='live', Symbol='HJURP', Aliases='FAKTS, URLC9, hFLEG1', description='Holliday junction recognition protein', other_designations='14-3-3-associated AKT substrate|fetal liver expressing gene 1|fetal liver-expressing gene 1 protein|up-regulated in lung cancer 9', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233836701, end_position_on_the_genomic_accession=233854566, orientation='minus', exon_count=9, OMIM=612667, no_hdr0=''),
  55356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55356, CurrentID=0, Status='live', Symbol='SLC22A15', Aliases='FLIPT1, PRO34686', description='solute carrier family 22 member 15', other_designations='flipt 1|fly-like putative organic ion transporter 1|fly-like putative transporter 1|solute carrier family 22 (organic cation transporter), member 15|solute carrier family 22, member 15|trans-like protein', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115976498, end_position_on_the_genomic_accession=116070054, orientation='plus', exon_count=19, OMIM=608275, no_hdr0=''),
  55357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55357, CurrentID=0, Status='live', Symbol='TBC1D2', Aliases='PARIS-1, PARIS1A, TBC1D2', description='TBC1 domain family member 2', other_designations='TBC1 domain family, member 2|TBC1 domain family, member 2A|armus|prostate antigen recognized and identified by SEREX 1', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98198998, end_position_on_the_genomic_accession=98255721, orientation='minus', exon_count=16, OMIM=609871, no_hdr0=''),
  55359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55359, CurrentID=0, Status='live', Symbol='STYK1', Aliases='NOK, SuRTK106', description='serine/threonine/tyrosine kinase 1', other_designations='novel oncogene with kinase domain|protein PK-unique', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10618923, end_position_on_the_genomic_accession=10674318, orientation='minus', exon_count=14, OMIM=611433, no_hdr0=''),
  55361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55361, CurrentID=0, Status='live', Symbol='PI4K2A', Aliases='PI4KII, PIK42A', description='phosphatidylinositol 4-kinase type 2 alpha', other_designations='phosphatidylinositol 4-kinase type II (PI4KII)|phosphatidylinositol 4-kinase type II-alpha', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97640686, end_position_on_the_genomic_accession=97676434, orientation='plus', exon_count=9, OMIM=609763, no_hdr0=''),
  55362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55362, CurrentID=0, Status='live', Symbol='TMEM63B', Aliases='C6orf110', description='transmembrane protein 63B', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44126757, end_position_on_the_genomic_accession=44155519, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  55363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55363, CurrentID=0, Status='live', Symbol='HEMGN', Aliases='CT155, EDAG, EDAG-1, NDR', description='hemogen', other_designations='erythroid differentiation-associated gene protein|hemopoietic gene protein|negative differentiation regulator protein', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97926791, end_position_on_the_genomic_accession=97944915, orientation='minus', exon_count=6, OMIM=610715, no_hdr0=''),
  55364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55364, CurrentID=0, Status='live', Symbol='IMPACT', Aliases='RWDD5', description='impact RWD domain protein', other_designations='Impact homolog|RWD domain containing 5|imprinted and ancient gene protein homolog', map_location='18q11.2-q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24426645, end_position_on_the_genomic_accession=24453531, orientation='plus', exon_count=11, OMIM=615319, no_hdr0=''),
  55365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55365, CurrentID=0, Status='live', Symbol='TMEM176A', Aliases='GS188, HCA112, MS4B1', description='transmembrane protein 176A', other_designations='hepatocellular carcinoma-associated antigen 112|likley ortholog of mouse GS188', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150800525, end_position_on_the_genomic_accession=150805120, orientation='plus', exon_count=8, OMIM=610334, no_hdr0=''),
  55366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55366, CurrentID=0, Status='live', Symbol='LGR4', Aliases='BNMD17, GPR48', description='leucine-rich repeat containing G protein-coupled receptor 4', other_designations='G protein-coupled receptor 48', map_location='11p14-p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27365961, end_position_on_the_genomic_accession=27472787, orientation='minus', exon_count=18, OMIM=606666, no_hdr0=''),
  55367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55367, CurrentID=0, Status='live', Symbol='PIDD1', Aliases='LRDD, PIDD', description='p53-induced death domain protein 1', other_designations='leucine-rich repeat and death domain-containing protein|leucine-rich repeats and death domain containing|p53-induced protein with a death domain', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=799179, end_position_on_the_genomic_accession=809501, orientation='minus', exon_count=19, OMIM=605247, no_hdr0=''),
  55370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55370, CurrentID=0, Status='live', Symbol='PPP4R1L', Aliases='C20orf192, PPP4R1P1, PRO1085', description='protein phosphatase 4 regulatory subunit 1 like (pseudogene)', other_designations='protein phosphatase 4, regulatory subunit 1 pseudogene|protein phosphatase 4, regulatory subunit 1-like (pseudogene)', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58232777, end_position_on_the_genomic_accession=58309439, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  55374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55374, CurrentID=0, Status='live', Symbol='TMCO6', Aliases='PRO1580', description='transmembrane and coiled-coil domains 6', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140596540, end_position_on_the_genomic_accession=140647248, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  55379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55379, CurrentID=0, Status='live', Symbol='LRRC59', Aliases='PRO1855, p34', description='leucine rich repeat containing 59', other_designations='ribosome-binding protein p34', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50381233, end_position_on_the_genomic_accession=50397553, orientation='minus', exon_count=7, OMIM=614854, no_hdr0=''),
  55380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55380, CurrentID=0, Status='live', Symbol='EPS15P1', Aliases='EPS15L2, PRO1866', description='epidermal growth factor receptor pathway substrate 15 pseudogene 1', other_designations='epidermal growth factor receptor pathway substrate 15-like 2', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46781890, end_position_on_the_genomic_accession=46783520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55384, CurrentID=0, Status='live', Symbol='MEG3', Aliases='FP504, GTL2, LINC00023, NCRNA00023, PRO0518, PRO2160, onco-lncRNA-83, prebp1', description='maternally expressed 3 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 23', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100826108, end_position_on_the_genomic_accession=100861026, orientation='plus', exon_count=13, OMIM=605636, no_hdr0=''),
  55387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55387, CurrentID=0, Status='live', Symbol='PRO2214', Aliases='', description='uncharacterized protein PRO2214', other_designations='', map_location='5q12.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55388, CurrentID=0, Status='live', Symbol='MCM10', Aliases='CNA43, DNA43, PRO2249', description='minichromosome maintenance 10 replication initiation factor', other_designations='MCM10 minichromosome maintenance deficient 10|homolog of yeast MCM10|hsMCM10|minichromosome maintenance complex component 10', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13161554, end_position_on_the_genomic_accession=13211104, orientation='plus', exon_count=21, OMIM=609357, no_hdr0=''),
  105369274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369274, CurrentID=0, Status='live', Symbol='LOC105369274', Aliases='', description='olfactory receptor 4N4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21638503, end_position_on_the_genomic_accession=21652911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55390, CurrentID=0, Status='live', Symbol='PRO2268', Aliases='', description='uncharacterized protein PRO2268', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614354, no_hdr0=''),
  105369276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369276, CurrentID=0, Status='live', Symbol='LOC105369276', Aliases='', description='putative protein FAM90A23P', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369277, CurrentID=0, Status='live', Symbol='LOC105369277', Aliases='', description='transmembrane protein C16orf54', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55412, CurrentID=0, Status='live', Symbol='PRO2949', Aliases='', description='uncharacterized protein PRO2949', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55415, CurrentID=0, Status='live', Symbol='PRO2964', Aliases='', description='uncharacterized protein PRO2964', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100194425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100194425, CurrentID=0, Status='live', Symbol='BCAP31P1', Aliases='', description='B-cell receptor-associated protein 31 pseudogene 1', other_designations='TCONS_l2_00010135', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33987962, end_position_on_the_genomic_accession=33991551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100194426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100194426, CurrentID=0, Status='live', Symbol='BCAP31P2', Aliases='', description='B-cell receptor-associated protein 31 pseudogene 2', other_designations='TCONS_l2_00009689', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32869983, end_position_on_the_genomic_accession=32873572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55421, CurrentID=0, Status='live', Symbol='NCBP3', Aliases='C17orf85, ELG, HSA277841', description='nuclear cap binding subunit 3', other_designations='CTD-3195I5.1|CTD-3195I5.5', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3802160, end_position_on_the_genomic_accession=3846251, orientation='minus', exon_count=16, OMIM=616624, no_hdr0=''),
  55422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55422, CurrentID=0, Status='live', Symbol='ZNF331', Aliases='RITA, ZNF361, ZNF463', description='zinc finger protein 331', other_designations='C2H2-like zinc finger protein rearranged in thyroid adenomas|KRAB zinc finger protein|zinc finger protein 361|zinc finger protein 463', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53503880, end_position_on_the_genomic_accession=53580269, orientation='plus', exon_count=19, OMIM=606043, no_hdr0=''),
  55423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55423, CurrentID=0, Status='live', Symbol='SIRPG', Aliases='CD172g, SIRP-B2, SIRPB2, SIRPgamma, bA77C3.1', description='signal-regulatory protein gamma', other_designations='CD172 antigen-like family member B|CD172g antigen|SIRP beta 2|SIRP-beta-2|SIRP-gamma|signal-regulatory protein beta 2|signal-regulatory protein beta-2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1629152, end_position_on_the_genomic_accession=1686516, orientation='minus', exon_count=8, OMIM=605466, no_hdr0=''),
  55425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55425, CurrentID=0, Status='live', Symbol='GPALPP1', Aliases='AD029, KIAA1704, LSR7, bA245H20.2', description='GPALPP motifs containing 1', other_designations='lipopolysaccharide specific response-7 protein|lipopolysaccharide-specific response protein 7', map_location='13q14.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44989538, end_position_on_the_genomic_accession=45033611, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374267, CurrentID=0, Status='live', Symbol='LOC105374267', Aliases='', description='uncharacterized LOC105374267', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187958403, end_position_on_the_genomic_accession=187965520, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55432, CurrentID=0, Status='live', Symbol='YOD1', Aliases='DUBA8, OTUD2, PRO0907', description='YOD1 deubiquitinase', other_designations='DUBA-8|HIN-7|HIV-1-induced protease 7|OTU domain containing 2|OTU domain-containing protein 2|YOD1 OTU deubiquinating enzyme 1 homolog ( yeast)|hsHIN7', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207043849, end_position_on_the_genomic_accession=207053133, orientation='minus', exon_count=5, OMIM=612023, no_hdr0=''),
  55435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55435, CurrentID=0, Status='live', Symbol='AP1AR', Aliases='2C18, C4orf16, GBAR, PRO0971, gamma-BAR', description='adaptor related protein complex 1 associated regulatory protein', other_designations='adapter-related protein complex 1-associated regulatory protein|adaptor-related protein complex 1 associated regulatory protein|adaptor-related protein complex 1-associated regulatory protein|gadkin|gamma-1-adaptin brefeldin A resistance protein|gamma-A1-adaptin and kinesin interactor|gamma1-adaptin brefeldin A resistance protein, gamma-BAR', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112231737, end_position_on_the_genomic_accession=112272715, orientation='plus', exon_count=11, OMIM=610851, no_hdr0=''),
  55437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55437, CurrentID=0, Status='live', Symbol='STRADB', Aliases='ALS2CR2, CALS-21, ILPIP, ILPIPA, PAPK, PRO1038', description='STE20-related kinase adaptor beta', other_designations='ILP-interacting protein ILPIPA|STRAD beta|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 2|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 2 protein|pseudokinase ALS2CR2', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201451669, end_position_on_the_genomic_accession=201480851, orientation='plus', exon_count=12, OMIM=607333, no_hdr0=''),
  100313851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313851, CurrentID=0, Status='live', Symbol='PIRC25', Aliases='', description='piwi-interacting RNA cluster 25', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369283, CurrentID=0, Status='live', Symbol='LOC105369283', Aliases='', description='uncharacterized LOC105369283', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55448, CurrentID=0, Status='live', Symbol='PRO1483', Aliases='', description='uncharacterized LOC55448', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55449, CurrentID=0, Status='live', Symbol='DHRS4-AS1', Aliases='AS1DHRS4, C14orf167, PRO1488', description='DHRS4 antisense RNA 1', other_designations='DHRS4 antisense RNA 1 (non-protein coding)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23938731, end_position_on_the_genomic_accession=23955112, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55450, CurrentID=0, Status='live', Symbol='CAMK2N1', Aliases='PRO1489', description='calcium/calmodulin-dependent protein kinase II inhibitor 1', other_designations='CaMKIINalpha|caMKII inhibitory protein alpha|caMKIIN-alpha', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20482391, end_position_on_the_genomic_accession=20486235, orientation='minus', exon_count=2, OMIM=614986, no_hdr0=''),
  55451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55451, CurrentID=0, Status='live', Symbol='LINC00216', Aliases='C14orf107, NCRNA00216, PRO1617', description='long intergenic non-protein coding RNA 216', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55454, CurrentID=0, Status='live', Symbol='CSGALNACT2', Aliases='CHGN2, GALNACT-2, GALNACT2, PRO0082', description='chondroitin sulfate N-acetylgalactosaminyltransferase 2', other_designations='beta 4 GalNAcT-2|chondroitin beta1,4 N-acetylgalactosaminyltransferase 2', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43138445, end_position_on_the_genomic_accession=43185308, orientation='plus', exon_count=10, OMIM=616616, no_hdr0=''),
  105369285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369285, CurrentID=0, Status='live', Symbol='LOC105369285', Aliases='', description='zinc finger protein 813-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55460, CurrentID=0, Status='live', Symbol='PRO1082', Aliases='', description='uncharacterized protein PRO1082', other_designations='predicted protein of HQ1082', map_location='3p21.1-q13.13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55466, CurrentID=0, Status='live', Symbol='DNAJA4', Aliases='MST104, MSTP104, PRO1472', description='DnaJ heat shock protein family (Hsp40) member A4', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 4', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78264145, end_position_on_the_genomic_accession=78282196, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105369287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369287, CurrentID=0, Status='live', Symbol='LOC105369287', Aliases='', description='uncharacterized LOC105369287', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238152889, end_position_on_the_genomic_accession=238158749, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55471, CurrentID=0, Status='live', Symbol='NDUFAF7', Aliases='C2orf56, MidA, PRO1853', description='NADH:ubiquinone oxidoreductase complex assembly factor 7', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 7|mitochondrial dysfunction protein A homolog|protein midA homolog, mitochondrial', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37231631, end_position_on_the_genomic_accession=37249160, orientation='plus', exon_count=11, OMIM=615898, no_hdr0=''),
  55472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55472, CurrentID=0, Status='live', Symbol='RBM12B-AS1', Aliases='C8orf39, PRO1905', description='RBM12B antisense RNA 1', other_designations='RBM12B antisense RNA 1 (non-protein coding)', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93740111, end_position_on_the_genomic_accession=93740819, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369288, CurrentID=0, Status='live', Symbol='LOC105369288', Aliases='', description='uncharacterized LOC105369288', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7092220, end_position_on_the_genomic_accession=7114659, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55478, CurrentID=0, Status='live', Symbol='PRO2012', Aliases='', description='uncharacterized protein PRO2012', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369290, CurrentID=0, Status='live', Symbol='LOC105369290', Aliases='', description='uncharacterized LOC105369290', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14477329, end_position_on_the_genomic_accession=14539296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55486, CurrentID=0, Status='live', Symbol='PARL', Aliases='PRO2207, PSARL, PSARL1, PSENIP2, RHBDS1', description='presenilin associated, rhomboid-like', other_designations='mitochondrial intramembrane-cleaving protease PARL|rhomboid 7 homolog 1', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183826488, end_position_on_the_genomic_accession=183884905, orientation='minus', exon_count=11, OMIM=607858, no_hdr0=''),
  55500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55500, CurrentID=0, Status='live', Symbol='ETNK1', Aliases='EKI, EKI 1, EKI1, Nbla10396', description='ethanolamine kinase 1', other_designations='putative protein product of Nbla10396', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22625142, end_position_on_the_genomic_accession=22690674, orientation='plus', exon_count=14, OMIM=609858, no_hdr0=''),
  55501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55501, CurrentID=0, Status='live', Symbol='CHST12', Aliases='C4S-2, C4ST-2, C4ST2', description='carbohydrate (chondroitin 4) sulfotransferase 12', other_designations='chondroitin 4-O-sulfotransferase 2|chondroitin 4-sulfotransferase 2|sulfotransferase Hlo', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2403526, end_position_on_the_genomic_accession=2434607, orientation='plus', exon_count=4, OMIM=610129, no_hdr0=''),
  55502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55502, CurrentID=0, Status='live', Symbol='HES6', Aliases='C-HAIRY1, HES-6, bHLHb41, bHLHc23', description='hes family bHLH transcription factor 6', other_designations='class B basic helix-loop-helix protein 41|hairy and enhancer of split 6', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238238267, end_position_on_the_genomic_accession=238240124, orientation='minus', exon_count=4, OMIM=610331, no_hdr0=''),
  55503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55503, CurrentID=0, Status='live', Symbol='TRPV6', Aliases='ABP/ZF, CAT1, CATL, ECAC2, HSA277909, LP6728, ZFAB', description='transient receptor potential cation channel, subfamily V, member 6', other_designations='Alu-binding protein with zinc finger domain|calcium transport protein 1|calcium transporter-like protein|epithelial apical membrane calcium transporter/channel CaT1|epithelial calcium channel 2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142871203, end_position_on_the_genomic_accession=142885745, orientation='minus', exon_count=15, OMIM=606680, no_hdr0=''),
  55504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55504, CurrentID=0, Status='live', Symbol='TNFRSF19', Aliases='TAJ, TAJ-alpha, TRADE, TROY', description='tumor necrosis factor receptor superfamily member 19', other_designations='toxicity and JNK inducer|tumor necrosis factor receptor superfamily, member 19', map_location='13q12.11-q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23570370, end_position_on_the_genomic_accession=23676105, orientation='plus', exon_count=12, OMIM=606122, no_hdr0=''),
  317649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317649, CurrentID=0, Status='live', Symbol='EIF4E3', Aliases='eIF-4E3, eIF4E-3', description='eukaryotic translation initiation factor 4E family member 3', other_designations='', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71679289, end_position_on_the_genomic_accession=71754773, orientation='minus', exon_count=12, OMIM=609896, no_hdr0=''),
  55506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55506, CurrentID=0, Status='live', Symbol='H2AFY2', Aliases='macroH2A2', description='H2A histone family member Y2', other_designations='H2A histone family, member Y2|core histone macroH2A2.2|histone macroH2A2|mH2A2', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70052601, end_position_on_the_genomic_accession=70112284, orientation='plus', exon_count=9, OMIM=616141, no_hdr0=''),
  55507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55507, CurrentID=0, Status='live', Symbol='GPRC5D', Aliases='', description='G protein-coupled receptor, class C, group 5, member D', other_designations='G protein-coupled receptor, family C, group 5, member D|orphan G-protein coupled receptor', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12940554, end_position_on_the_genomic_accession=12979517, orientation='minus', exon_count=8, OMIM=607437, no_hdr0=''),
  55508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55508, CurrentID=0, Status='live', Symbol='SLC35E3', Aliases='BLOV1', description='solute carrier family 35 member E3', other_designations='bladder cancer overexpressed protein|bladder cancer-overexpressed gene 1 protein|solute carrier family 35, member E2|solute carrier family 35, member E3', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68746148, end_position_on_the_genomic_accession=68766073, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55509, CurrentID=0, Status='live', Symbol='BATF3', Aliases='JDP1, JUNDM1, SNFT', description='basic leucine zipper transcription factor, ATF-like 3', other_designations='21 kDa small nuclear factor isolated from T-cells|21-kD small nuclear factor isolated from T cells|B-ATF-3|Jun dimerization protein 1|Jun dimerization protein p21SNFT', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212686417, end_position_on_the_genomic_accession=212699985, orientation='minus', exon_count=4, OMIM=612470, no_hdr0=''),
  55510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55510, CurrentID=0, Status='live', Symbol='DDX43', Aliases='CT13, HAGE', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 43', other_designations='DEAD box protein 43|DEAD box protein HAGE|DEAD-box protein 43|cancer/testis antigen 13|helical antigen', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73394562, end_position_on_the_genomic_accession=73417566, orientation='plus', exon_count=18, OMIM=606286, no_hdr0=''),
  55511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55511, CurrentID=0, Status='live', Symbol='SAGE1', Aliases='CT14, SAGE', description='sarcoma antigen 1', other_designations='cancer/testis antigen 14|putative tumor antigen', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135893634, end_position_on_the_genomic_accession=135913061, orientation='plus', exon_count=20, OMIM=300359, no_hdr0=''),
  55512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55512, CurrentID=0, Status='live', Symbol='SMPD3', Aliases='NSMASE2', description='sphingomyelin phosphodiesterase 3', other_designations='nSMase-2|neutral sphingomyelinase 2|neutral sphingomyelinase II|sphingomyelin phosphodiesterase 3, neutral membrane (neutral sphingomyelinase II)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68358327, end_position_on_the_genomic_accession=68448506, orientation='minus', exon_count=14, OMIM=605777, no_hdr0=''),
  664612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664612, CurrentID=0, Status='live', Symbol='MIR539', Aliases='MIRN539, hsa-mir-539, mir-539', description='microRNA 539', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101047321, end_position_on_the_genomic_accession=101047398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  55515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55515, CurrentID=0, Status='live', Symbol='ASIC4', Aliases='ACCN4, BNAC4', description='acid sensing ion channel subunit family member 4', other_designations='acid sensing (proton gated) ion channel family member 4|acid sensing ion channel family member 4|acid-sensing (proton-gated) ion channel family member 4|amiloride-sensitive cation channel 4, pituitary|amiloride-sensitive cation channel family member 4, pituitary|brain sodium channel 4', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219514170, end_position_on_the_genomic_accession=219538772, orientation='plus', exon_count=12, OMIM=606715, no_hdr0=''),
  317662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317662, CurrentID=0, Status='live', Symbol='FAM149B1', Aliases='KIAA0974', description='family with sequence similarity 149 member B1', other_designations='family with sequence similarity 149, member B1', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73168095, end_position_on_the_genomic_accession=73242181, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  664613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664613, CurrentID=0, Status='live', Symbol='MIR544A', Aliases='MIR544, MIRN544, hsa-mir-544, hsa-mir-544a', description='microRNA 544a', other_designations='microRNA 544', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101048658, end_position_on_the_genomic_accession=101048748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317664, CurrentID=0, Status='live', Symbol='DFNA49', Aliases='', description='deafness, autosomal dominant 49', other_designations='', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608372, no_hdr0=''),
  55521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55521, CurrentID=0, Status='live', Symbol='TRIM36', Aliases='HAPRIN, RBCC728, RNF98', description='tripartite motif containing 36', other_designations='RING finger protein 98|tripartite motif protein 36|tripartite motif-containing 36|tripartite motif-containing protein 36|zinc-binding protein Rbcc728', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115124762, end_position_on_the_genomic_accession=115180546, orientation='minus', exon_count=14, OMIM=609317, no_hdr0=''),
  317668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317668, CurrentID=0, Status='live', Symbol='MCDR3', Aliases='', description='macular dystrophy, retinal 3', other_designations='', map_location='5p15.33-p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608850, no_hdr0=''),
  317669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317669, CurrentID=0, Status='live', Symbol='IBD9', Aliases='', description='inflammatory bowel disease 9', other_designations='', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608448, no_hdr0=''),
  55526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55526, CurrentID=0, Status='live', Symbol='DHTKD1', Aliases='AMOXAD, CMT2Q', description='dehydrogenase E1 and transketolase domain containing 1', other_designations='dehydrogenase E1 and transketolase domain-containing protein 1', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12068917, end_position_on_the_genomic_accession=12123228, orientation='plus', exon_count=17, OMIM=614984, no_hdr0=''),
  317671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317671, CurrentID=0, Status='live', Symbol='RFESD', Aliases='', description='Rieske (Fe-S) domain containing', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95646741, end_position_on_the_genomic_accession=95658082, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55529, CurrentID=0, Status='live', Symbol='TMEM55A', Aliases='', description='transmembrane protein 55A', other_designations='PtdIns-4,5-P(2) 4-phosphatase type II|ptdIns-4,5-P2 4-Ptase II|type 2 PtdIns-4,5-P2 4-Ptase|type II phosphatidylinositol 4,5-bisphosphate 4-phosphatase', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90994271, end_position_on_the_genomic_accession=91040975, orientation='minus', exon_count=8, OMIM=609864, no_hdr0=''),
  55530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55530, CurrentID=0, Status='live', Symbol='SVOP', Aliases='', description='SV2 related protein', other_designations='SV2 related protein homolog', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108908129, end_position_on_the_genomic_accession=109021240, orientation='minus', exon_count=16, OMIM=611699, no_hdr0=''),
  9255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9255, CurrentID=0, Status='live', Symbol='AIMP1', Aliases='EMAP2, EMAPII, HLD3, SCYE1, p43', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 1', other_designations='ARS-interacting multifunctional protein 1|endothelial monocyte-activating polypeptide 2|endothelial-monocyte activating polypeptide II|multisynthase complex auxiliary component p43|multisynthetase complex auxiliary component p43|small inducible cytokine subfamily E, member 1 (endothelial monocyte-activating)', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106315610, end_position_on_the_genomic_accession=106349225, orientation='plus', exon_count=9, OMIM=603605, no_hdr0=''),
  55532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55532, CurrentID=0, Status='live', Symbol='SLC30A10', Aliases='HMDPC, ZNT10, ZNT8, ZRC1, ZnT-10', description='solute carrier family 30 member 10', other_designations='solute carrier family 30, member 10|zinc transporter 8', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219913919, end_position_on_the_genomic_accession=219959315, orientation='minus', exon_count=5, OMIM=611146, no_hdr0=''),
  55534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55534, CurrentID=0, Status='live', Symbol='MAML3', Aliases='CAGH3, ERDA3, GDN, MAM-2, MAM2, TNRC3', description='mastermind like transcriptional coactivator 3', other_designations='CAG repeat containing (glia-derived nexin I alpha)|CAG repeat domain|expanded repeat domain, CAG/CTG 3|mam-3|mastermind-like 3|mastermind-like transcriptional coactivator 3|polyglutamine rich|trinucleotide repeat containing 3', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139716391, end_position_on_the_genomic_accession=140154079, orientation='minus', exon_count=5, OMIM=608991, no_hdr0=''),
  55536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55536, CurrentID=0, Status='live', Symbol='CDCA7L', Aliases='JPO2, R1, RAM2', description='cell division cycle associated 7-like', other_designations='transcription factor RAM2', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21900899, end_position_on_the_genomic_accession=21945924, orientation='minus', exon_count=11, OMIM=609685, no_hdr0=''),
  9256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9256, CurrentID=0, Status='live', Symbol='BZRAP1', Aliases='PBR-IP, PRAX-1, PRAX1, RIM-BP1, RIMBP1', description='benzodiazepine receptor (peripheral) associated protein 1', other_designations='RIMS-binding protein 1|benzodiazapine receptor (peripheral) associated protein 1|benzodiazapine receptor peripheral-associated protein 1|peripheral benzodiazepine receptor-interacting protein', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58301227, end_position_on_the_genomic_accession=58328795, orientation='minus', exon_count=34, OMIM=610764, no_hdr0=''),
  317682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317682, CurrentID=0, Status='live', Symbol='OTSC5', Aliases='', description='otosclerosis 5', other_designations='', map_location='3q22.1-q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608787, no_hdr0=''),
  55539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55539, CurrentID=0, Status='live', Symbol='KCNQ1DN', Aliases='BWRT, HSA404617', description='KCNQ1 downstream neighbor (non-protein coding)', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2870033, end_position_on_the_genomic_accession=2872105, orientation='plus', exon_count=2, OMIM=610980, no_hdr0=''),
  55540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55540, CurrentID=0, Status='live', Symbol='IL17RB', Aliases='CRL4, EVI27, IL17BR, IL17RH1', description='interleukin 17 receptor B', other_designations='IL-17 receptor B|IL-17 receptor homolog 1|IL-17B receptor|IL-17RB|IL-17Rh1|cytokine receptor CRL4|cytokine receptor-like 4|interleukin 17 receptor homolog 1|interleukin 17B receptor|interleukin-17B receptor', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53846550, end_position_on_the_genomic_accession=53865800, orientation='plus', exon_count=13, OMIM=605458, no_hdr0=''),
  317685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317685, CurrentID=0, Status='live', Symbol='NYS4', Aliases='', description='nystagmus 4, congenital autosomal dominant', other_designations='', map_location='13q31-q33', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=193003, no_hdr0=''),
  317686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317686, CurrentID=0, Status='live', Symbol='MAPK6PS1', Aliases='Erk3ps1', description='mitogen-activated protein kinase 6 pseudogene 1', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53536317, end_position_on_the_genomic_accession=53541566, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  317687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317687, CurrentID=0, Status='live', Symbol='BANF1P1', Aliases='BCRG1, BCRP1, D14S1460, D14S1460E', description='barrier to autointegration factor 1 pseudogene 1', other_designations='breakpoint cluster region protein, uterine leiomyoma, 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68936547, end_position_on_the_genomic_accession=68937244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55544, CurrentID=0, Status='live', Symbol='RBM38', Aliases='HSRNASEB, RNPC1, SEB4B, SEB4D, dJ800J21.2', description='RNA binding motif protein 38', other_designations='CLL-associated antigen KW-5|RNA-binding region (RNP1, RRM) containing 1|RNA-binding region-containing protein 1|ssDNA-binding protein SEB4', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57391398, end_position_on_the_genomic_accession=57409333, orientation='plus', exon_count=6, OMIM=612428, no_hdr0=''),
  55545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55545, CurrentID=0, Status='live', Symbol='MSX2P1', Aliases='HPX5, HSHPX5, MSX2P', description='msh homeobox 2 pseudogene 1', other_designations='msh homeobox homolog 2 pseudogene', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58156959, end_position_on_the_genomic_accession=58159120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317691, CurrentID=0, Status='live', Symbol='VN1R9P', Aliases='VN1R22-1P', description='vomeronasal 1 receptor 9 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17021355, end_position_on_the_genomic_accession=17022270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  664618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664618, CurrentID=0, Status='live', Symbol='HSP90AB4P', Aliases='HSP90Bd', description='heat shock protein 90kDa alpha family class B member 4, pseudogene', other_designations='heat shock protein 90Bd|heat shock protein 90kDa alpha (cytosolic), class B member 4, pseudogene', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58690473, end_position_on_the_genomic_accession=58693607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55552, CurrentID=0, Status='live', Symbol='ZNF823', Aliases='HSZFP36', description='zinc finger protein 823', other_designations='ZFP 36 for a zinc finger protein|zinc finger protein ZFP-36', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11721265, end_position_on_the_genomic_accession=11739009, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55553, CurrentID=0, Status='live', Symbol='SOX6', Aliases='HSSOX6, SOXD', description='SRY-box 6', other_designations='SRY (sex determining region Y)-box 6|SRY box 6|SRY-box containing gene 6', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15966449, end_position_on_the_genomic_accession=16476388, orientation='minus', exon_count=18, OMIM=607257, no_hdr0=''),
  55554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55554, CurrentID=0, Status='live', Symbol='KLK15', Aliases='ACO, HSRNASPH', description='kallikrein related peptidase 15', other_designations='ACO protease|kallikrein-like serine protease|kallikrein-related peptidase 15|prostinogen', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50825289, end_position_on_the_genomic_accession=50836035, orientation='minus', exon_count=8, OMIM=610601, no_hdr0=''),
  55556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55556, CurrentID=0, Status='live', Symbol='ENOSF1', Aliases='HSRTSBETA, RTS, TYMSAS', description='enolase superfamily member 1', other_designations='L-fuconate dehydratase|antisense RNA to thymidylate synthase', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=670324, end_position_on_the_genomic_accession=712662, orientation='minus', exon_count=19, OMIM=607427, no_hdr0=''),
  317701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317701, CurrentID=0, Status='live', Symbol='VN1R2', Aliases='V1RL2', description='vomeronasal 1 receptor 2', other_designations='G-protein coupled receptor GPCR25|V1R-like 2|V1r-like receptor 2|hGPCR25|pheromone receptor', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53258292, end_position_on_the_genomic_accession=53259602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  55558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55558, CurrentID=0, Status='live', Symbol='PLXNA3', Aliases='6.3, HSSEXGENE, PLXN3, PLXN4, XAP-6', description='plexin A3', other_designations='Sex chromosome X transmembrane protein of HGF receptor family 3|plexin-4|semaphorin receptor SEX', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154458281, end_position_on_the_genomic_accession=154473646, orientation='plus', exon_count=34, OMIM=300022, no_hdr0=''),
  317703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317703, CurrentID=0, Status='live', Symbol='VN1R4', Aliases='V1RL4', description='vomeronasal 1 receptor 4', other_designations='G-protein coupled receptor GPCR27|V1R-like 4|V1r-like receptor 4|hGPCR27|pheromone receptor', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53266760, end_position_on_the_genomic_accession=53267907, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55561, CurrentID=0, Status='live', Symbol='CDC42BPG', Aliases='DMPK2, HSMDPKIN, KAPPA-200, MRCKG, MRCKgamma', description='CDC42 binding protein kinase gamma (DMPK-like)', other_designations='CDC42-binding protein kinase gamma|DMPK-like gamma|MRCK gamma|myotonic dystrophy kinase-related CDC42-binding kinase gamma|myotonic dystrophy protein kinase like protein|myotonic dystrophy protein kinase-like alpha|myotonic dystrophy protein kinase-like gamma', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64823809, end_position_on_the_genomic_accession=64844673, orientation='minus', exon_count=40, OMIM=613991, no_hdr0=''),
  105369303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369303, CurrentID=0, Status='live', Symbol='LOC105369303', Aliases='', description='uncharacterized LOC105369303', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18437902, end_position_on_the_genomic_accession=18439277, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55565, CurrentID=0, Status='live', Symbol='ZNF821', Aliases='', description='zinc finger protein 821', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71859680, end_position_on_the_genomic_accession=71884867, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  55567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55567, CurrentID=0, Status='live', Symbol='DNAH3', Aliases='DLP3, DNAHC3B, HEL-36, Hsadhc3, dnahc3-b', description='dynein, axonemal, heavy chain 3', other_designations='axonemal beta dynein heavy chain 3|axonemal dynein, heavy chain|ciliary dynein heavy chain 3|dynein, axonemal, heavy polypeptide 3|epididymis luminal protein 36', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20933111, end_position_on_the_genomic_accession=21159441, orientation='minus', exon_count=68, OMIM=603334, no_hdr0=''),
  55568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55568, CurrentID=0, Status='live', Symbol='GALNT10', Aliases='GALNACT10, PPGALNACT10, PPGANTASE10', description='polypeptide N-acetylgalactosaminyltransferase 10', other_designations='GalNAc transferase 10|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 10|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 10 (GalNAc-T10)', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154190735, end_position_on_the_genomic_accession=154420984, orientation='plus', exon_count=12, OMIM=608043, no_hdr0=''),
  105369304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369304, CurrentID=0, Status='live', Symbol='LOC105369304', Aliases='', description='uncharacterized LOC105369304', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14383249, end_position_on_the_genomic_accession=14394031, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  317714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317714, CurrentID=0, Status='live', Symbol='DYT15', Aliases='', description='dystonia 15, myoclonic', other_designations='', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607488, no_hdr0=''),
  55571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55571, CurrentID=0, Status='live', Symbol='CNOT11', Aliases='C2orf29, C40', description='CCR4-NOT transcription complex subunit 11', other_designations='CCR4-NOT transcription complex, subunit 11|UPF0760 protein C2orf29', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101252883, end_position_on_the_genomic_accession=101270316, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  55572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55572, CurrentID=0, Status='live', Symbol='FOXRED1', Aliases='FP634, H17', description='FAD-dependent oxidoreductase domain containing 1', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126269040, end_position_on_the_genomic_accession=126278132, orientation='plus', exon_count=12, OMIM=613622, no_hdr0=''),
  55573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55573, CurrentID=0, Status='live', Symbol='CDV3', Aliases='H41', description='CDV3 homolog (mouse)', other_designations='carnitine deficiency-associated gene expressed in ventricle 3', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133573590, end_position_on_the_genomic_accession=133590274, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  317718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317718, CurrentID=0, Status='live', Symbol='DFNA37', Aliases='', description='deafness, autosomal dominant 37', other_designations='', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  317719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317719, CurrentID=0, Status='live', Symbol='KLHL10', Aliases='SPGF11', description='kelch like family member 10', other_designations='kelch-like 10|kelch-like family member 10', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41837490, end_position_on_the_genomic_accession=41848347, orientation='plus', exon_count=6, OMIM=608778, no_hdr0=''),
  55576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55576, CurrentID=0, Status='live', Symbol='STAB2', Aliases='FEEL2, FELE-2, FELL2, FEX2, HARE', description='stabilin 2', other_designations='CD44-like precursor FELL|FAS1 EGF-like and X-link domain containing adhesion molecule-2|fasciclin egf-like, laminin-type egf-like, and link domain-containing scavenger receptor-2|hepatic hyaluronan clearance receptor|hyaluronan receptor for endocytosis|hyaluronic acid receptor for endocytosis', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103587255, end_position_on_the_genomic_accession=103766729, orientation='plus', exon_count=73, OMIM=608561, no_hdr0=''),
  55577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55577, CurrentID=0, Status='live', Symbol='NAGK', Aliases='GNK, HSA242910', description='N-acetylglucosamine kinase', other_designations='glcNAc kinase', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71068241, end_position_on_the_genomic_accession=71078868, orientation='plus', exon_count=10, OMIM=606828, no_hdr0=''),
  55578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55578, CurrentID=0, Status='live', Symbol='SUPT20H', Aliases='C13, C13orf19, FAM48A, FP757, P38IP, SPT20', description='SPT20 homolog, SAGA complex component', other_designations='family with sequence similarity 48, member A|protein FAM48A|suppressor of Ty 20 homolog|transcription factor (p38 interacting protein)|tumor rejection antigen', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37009312, end_position_on_the_genomic_accession=37059713, orientation='minus', exon_count=29, OMIM=613417, no_hdr0=''),
  730159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730159, CurrentID=0, Status='live', Symbol='LOC730159', Aliases='', description='uncharacterized LOC730159', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173603008, end_position_on_the_genomic_accession=173637134, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  118490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118490, CurrentID=0, Status='live', Symbol='MSS51', Aliases='ZMYND17', description='MSS51 mitochondrial translational activator', other_designations='MSS51 mitochondrial translational activator homolog|zinc finger MYND domain-containing protein 17|zinc finger, MYND domain containing 17|zinc finger, MYND-type containing 17', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73423579, end_position_on_the_genomic_accession=73433561, orientation='minus', exon_count=7, OMIM=614773, no_hdr0=''),
  55582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55582, CurrentID=0, Status='live', Symbol='KIF27', Aliases='', description='kinesin family member 27', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83836698, end_position_on_the_genomic_accession=83921467, orientation='minus', exon_count=23, OMIM=611253, no_hdr0=''),
  317727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317727, CurrentID=0, Status='live', Symbol='LOC317727', Aliases='', description='ataxin 2 related protein pseudogene', other_designations='ataxin 2-like pseudogene', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16583248, end_position_on_the_genomic_accession=16586916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55584, CurrentID=0, Status='live', Symbol='CHRNA9', Aliases='HSA243342, NACHRA9', description='cholinergic receptor, nicotinic alpha 9', other_designations='NACHR alpha 9|acetylcholine receptor, neuronal nicotinic, alpha-9 subunit|cholinergic receptor, nicotinic, alpha 9 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 9|neuronal acetylcholine receptor protein, alpha-9 subunit|nicotinic acetylcholine receptor subunit alpha 9', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40335329, end_position_on_the_genomic_accession=40355217, orientation='plus', exon_count=5, OMIM=605116, no_hdr0=''),
  55585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55585, CurrentID=0, Status='live', Symbol='UBE2Q1', Aliases='GTAP, NICE-5, PRO3094, UBE2Q', description='ubiquitin conjugating enzyme E2Q family member 1', other_designations='E2 ubiquitin-conjugating enzyme Q1|galactosyl transferase-associated protein|protein NICE-5|ubiquitin carrier protein Q1|ubiquitin-conjugating enzyme E2Q (putative) 1|ubiquitin-conjugating enzyme E2Q family member 1|ubiquitin-protein ligase Q1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154548575, end_position_on_the_genomic_accession=154558644, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  55586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55586, CurrentID=0, Status='live', Symbol='MIOX', Aliases='ALDRL6', description='myo-inositol oxygenase', other_designations='MI oxygenase|aldehyde reductase (aldose reductase) like 6|aldehyde reductase-like 6|kidney-specific protein 32|renal-specific oxidoreductase', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50483988, end_position_on_the_genomic_accession=50490321, orientation='plus', exon_count=12, OMIM=606774, no_hdr0=''),
  105369307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369307, CurrentID=0, Status='live', Symbol='LOC105369307', Aliases='', description='uncharacterized LOC105369307', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15056029, end_position_on_the_genomic_accession=15063803, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55588, CurrentID=0, Status='live', Symbol='MED29', Aliases='IXL', description='mediator complex subunit 29', other_designations='intersex-like protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39391323, end_position_on_the_genomic_accession=39400563, orientation='plus', exon_count=4, OMIM=612914, no_hdr0=''),
  55589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55589, CurrentID=0, Status='live', Symbol='BMP2K', Aliases='BIKE, HRIHFB2017', description='BMP2 inducible kinase', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78776378, end_position_on_the_genomic_accession=78916365, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55591, CurrentID=0, Status='live', Symbol='VEZT', Aliases='VEZATIN', description='vezatin, adherens junctions transmembrane protein', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95217746, end_position_on_the_genomic_accession=95302790, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  55592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55592, CurrentID=0, Status='live', Symbol='GOLGA2P5', Aliases='GOLGA2B, GOLGA2L1', description='golgin A2 pseudogene 5', other_designations='golgi autoantigen, golgin subfamily a, 2-like 1|golgin A2 family, member B|golgin A2-like 1|golgin subfamily A member 2B', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100156391, end_position_on_the_genomic_accession=100173343, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  55593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55593, CurrentID=0, Status='live', Symbol='OTUD5', Aliases='DUBA', description='OTU deubiquitinase 5', other_designations='OTU domain containing 5|deubiquinating enzyme A|deubiquitinating enzyme A', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48922024, end_position_on_the_genomic_accession=48958624, orientation='minus', exon_count=13, OMIM=300713, no_hdr0=''),
  55596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55596, CurrentID=0, Status='live', Symbol='ZCCHC8', Aliases='', description='zinc finger, CCHC domain containing 8', other_designations='TRAMP-like complex RNA-binding factor ZCCHC8', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122471599, end_position_on_the_genomic_accession=122501073, orientation='minus', exon_count=17, OMIM=616381, no_hdr0=''),
  55599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55599, CurrentID=0, Status='live', Symbol='RNPC3', Aliases='RBM40, RNP, SNRNP65', description='RNA binding region (RNP1, RRM) containing 3', other_designations='RNA recognition protein|RNA-binding motif protein 40|RNA-binding region (RNP1, RRM) containing 3|RNA-binding region-containing protein 3|U11/U12 small nuclear ribonucleoprotein 65 kDa protein|U11/U12 snRNP 65 kDa protein|U11/U12 snRNP 65K|U11/U12-65K', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103525956, end_position_on_the_genomic_accession=103555239, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  55600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55600, CurrentID=0, Status='live', Symbol='ITLN1', Aliases='HL-1, HL1, INTL, ITLN, LFR, hIntL, omentin', description='intelectin 1 (galactofuranose binding)', other_designations='ITLN-1|endothelial lectin HL-1|galactofuranose-binding lectin|intestinal lactoferrin receptor', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160876540, end_position_on_the_genomic_accession=160885170, orientation='minus', exon_count=8, OMIM=609873, no_hdr0=''),
  55601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55601, CurrentID=0, Status='live', Symbol='DDX60', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 60', other_designations='DEAD box protein 60', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168216291, end_position_on_the_genomic_accession=168318807, orientation='minus', exon_count=40, OMIM=613974, no_hdr0=''),
  55602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55602, CurrentID=0, Status='live', Symbol='CDKN2AIP', Aliases='CARF', description='CDKN2A interacting protein', other_designations='collaborates/cooperates with ARF (alternate reading frame) protein|collaborator of ARF', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183444635, end_position_on_the_genomic_accession=183449064, orientation='plus', exon_count=3, OMIM=615914, no_hdr0=''),
  55603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55603, CurrentID=0, Status='live', Symbol='FAM46A', Aliases='C6orf37, XTP11', description='family with sequence similarity 46 member A', other_designations='HBV X-transactivated gene 11 protein|HBV XAg-transactivated protein 11|family with sequence similarity 46, member A|retinal expressed gene C6orf37', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81745730, end_position_on_the_genomic_accession=81752711, orientation='minus', exon_count=3, OMIM=611357, no_hdr0=''),
  55604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55604, CurrentID=0, Status='live', Symbol='LRRC16A', Aliases='CARMIL, CARMIL1, CARMIL1a, LRRC16, dJ501N12.1, dJ501N12.5', description='leucine rich repeat containing 16A', other_designations='CARMIL homolog|capping protein, Arp2/3, and Myosin-I Linker homolog 1|leucine rich repeat containing 16', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25279428, end_position_on_the_genomic_accession=25620530, orientation='plus', exon_count=41, OMIM=609593, no_hdr0=''),
  317749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317749, CurrentID=0, Status='live', Symbol='DHRS4L2', Aliases='SDR25C3', description='dehydrogenase/reductase (SDR family) member 4 like 2', other_designations='NADP(H)-dependent retinol dehydrogenase/reductase short isoform-like protein|dehydrogenase/reductase (SDR family) member 4 like 2A3|short chain dehydrogenase/reductase family 25C member 3|short chain dehydrogenase/reductase family 25C, member 3', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23969874, end_position_on_the_genomic_accession=24006408, orientation='plus', exon_count=10, OMIM=615196, no_hdr0=''),
  55607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55607, CurrentID=0, Status='live', Symbol='PPP1R9A', Aliases='NRB1, NRBI, Neurabin-I', description='protein phosphatase 1 regulatory subunit 9A', other_designations='neural tissue-specific F-actin-binding protein I|protein phosphatase 1, regulatory (inhibitor) subunit 9A|protein phosphatase 1, regulatory subunit 9A', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94907202, end_position_on_the_genomic_accession=95296415, orientation='plus', exon_count=26, OMIM=602468, no_hdr0=''),
  55608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55608, CurrentID=0, Status='live', Symbol='ANKRD10', Aliases='', description='ankyrin repeat domain 10', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110878540, end_position_on_the_genomic_accession=110915107, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55609, CurrentID=0, Status='live', Symbol='ZNF280C', Aliases='SUHW3, ZNF633', description='zinc finger protein 280C', other_designations='suppressor of hairy wing homolog 3|zinc finger protein 633', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130202699, end_position_on_the_genomic_accession=130269068, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  317754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317754, CurrentID=0, Status='live', Symbol='POTED', Aliases='A26B3, ANKRD21, CT104.1, POTE, POTE-21, POTE21', description='POTE ankyrin domain family member D', other_designations='ANKRD26-like family B member 3|ANKRD26-like family B, member 3|Expressed in prostate, ovary, testis, and placenta|POTE ankyrin domain family, member D|ankyrin repeat domain 21|ankyrin repeat domain-containing protein 21|cancer/testis antigen family 104, member 1|prostate, ovary, testis-expressed protein', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13610061, end_position_on_the_genomic_accession=13641585, orientation='plus', exon_count=12, OMIM=607549, no_hdr0=''),
  55611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55611, CurrentID=0, Status='live', Symbol='OTUB1', Aliases='HSPC263, OTB1, OTU1', description='OTU deubiquitinase, ubiquitin aldehyde binding 1', other_designations='OTU domain, ubiquitin aldehyde binding 1|OTU domain-containing ubiquitin aldehyde-binding protein 1|OTU-domain Ubal-binding 1|deubiquitinating enzyme OTUB1|otubain-1|ubiquitin-specific protease otubain 1|ubiquitin-specific-processing protease OTUB1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63985853, end_position_on_the_genomic_accession=63998420, orientation='plus', exon_count=7, OMIM=608337, no_hdr0=''),
  55612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55612, CurrentID=0, Status='live', Symbol='FERMT1', Aliases='C20orf42, DTGCU2, KIND1, UNC112A, URP1', description='fermitin family member 1', other_designations='UNC112 related protein 1|kindlerin|kindlin 1|kindlin syndrome protein|unc-112-related protein 1', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6074845, end_position_on_the_genomic_accession=6123544, orientation='minus', exon_count=15, OMIM=607900, no_hdr0=''),
  55613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55613, CurrentID=0, Status='live', Symbol='MTMR8', Aliases='', description='myotubularin related protein 8', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64268081, end_position_on_the_genomic_accession=64395453, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  55614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55614, CurrentID=0, Status='live', Symbol='KIF16B', Aliases='C20orf23, KISC20ORF, SNX23', description='kinesin family member 16B', other_designations='kinesin motor protein|sorting nexin 23', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16272097, end_position_on_the_genomic_accession=16573434, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  448831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448831, CurrentID=0, Status='live', Symbol='FRG2', Aliases='FRG2A', description='FSHD region gene 2', other_designations='FSHD region gene 2 protein', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190024367, end_position_on_the_genomic_accession=190028348, orientation='minus', exon_count=5, OMIM=609032, no_hdr0=''),
  55616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55616, CurrentID=0, Status='live', Symbol='ASAP3', Aliases='ACAP4, CENTB6, DDEFL1, UPLC1', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain 3', other_designations='ARF6 GTPase-activating protein|ArfGAP with SH3 domain, ankyrin repeat and PH domain 3 1|centaurin, beta 6|development and differentiation enhancing factor-like 1|development and differentiation-enhancing factor-like 1|protein up-regulated in liver cancer 1|up-regulated in liver cancer 1 (UPLC1)', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23428563, end_position_on_the_genomic_accession=23484567, orientation='minus', exon_count=27, OMIM=616594, no_hdr0=''),
  317761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317761, CurrentID=0, Status='live', Symbol='C14orf39', Aliases='Six6os1', description='chromosome 14 open reading frame 39', other_designations='six6 opposite strand transcript 1', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60353122, end_position_on_the_genomic_accession=60486069, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  317762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317762, CurrentID=0, Status='live', Symbol='CCDC85C', Aliases='', description='coiled-coil domain containing 85C', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99511266, end_position_on_the_genomic_accession=99604390, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  448835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448835, CurrentID=0, Status='live', Symbol='LCE6A', Aliases='C1orf44', description='late cornified envelope 6A', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152842854, end_position_on_the_genomic_accession=152843983, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55620, CurrentID=0, Status='live', Symbol='STAP2', Aliases='BKS', description='signal transducing adaptor family member 2', other_designations='BRK substrate|breast tumor kinase substrate|brk kinase substrate|signal-transducing adaptor protein-2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4324043, end_position_on_the_genomic_accession=4338877, orientation='minus', exon_count=13, OMIM=607881, no_hdr0=''),
  55621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55621, CurrentID=0, Status='live', Symbol='TRMT1', Aliases='TRM1', description='tRNA methyltransferase 1', other_designations='N(2),N(2)-dimethylguanosine tRNA methyltransferase|TRM1 tRNA methyltransferase 1 homolog|tRNA 2,2-dimethylguanosine-26 methyltransferase|tRNA methyltransferase 1 homolog|tRNA(guanine-26,N(2)-N(2)) methyltransferase|tRNA(m(2,2)G26)dimethyltransferase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13104900, end_position_on_the_genomic_accession=13117567, orientation='minus', exon_count=18, OMIM=611669, no_hdr0=''),
  55622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55622, CurrentID=0, Status='live', Symbol='TTC27', Aliases='', description='tetratricopeptide repeat domain 27', other_designations='TPR repeat protein 27', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32628020, end_position_on_the_genomic_accession=32821051, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  55623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55623, CurrentID=0, Status='live', Symbol='THUMPD1', Aliases='', description='THUMP domain containing 1', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20733664, end_position_on_the_genomic_accession=20741964, orientation='minus', exon_count=5, OMIM=616662, no_hdr0=''),
  55624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55624, CurrentID=0, Status='live', Symbol='POMGNT1', Aliases='GNTI.2, GnT I.2, LGMD2O, MEB, MGAT1.2, gnT-I.2', description='protein O-linked mannose N-acetylglucosaminyltransferase 1 (beta 1,2-)', other_designations='UDP-GlcNAc:alpha-D-mannoside beta-1,2-N-acetylglucosaminyltransferase I.2', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46188681, end_position_on_the_genomic_accession=46220305, orientation='minus', exon_count=25, OMIM=606822, no_hdr0=''),
  55625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55625, CurrentID=0, Status='live', Symbol='ZDHHC7', Aliases='DHHC7, SERZ-B, SERZ1, ZNF370', description='zinc finger, DHHC-type containing 7', other_designations='DHHC-7|Sertoli cell gene with zinc finger domain-&#946|zinc finger DHHC domain-containing protein 7|zinc finger protein 370|zinc finger, DHHC domain containing 7', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84974460, end_position_on_the_genomic_accession=85011732, orientation='minus', exon_count=10, OMIM=614604, no_hdr0=''),
  55626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55626, CurrentID=0, Status='live', Symbol='AMBRA1', Aliases='DCAF3, WDR94', description='autophagy/beclin-1 regulator 1', other_designations='DDB1 and CUL4 associated factor 3|WD repeat domain 94|activating molecule in beclin-1-regulated autophagy', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46396412, end_position_on_the_genomic_accession=46594069, orientation='minus', exon_count=23, OMIM=611359, no_hdr0=''),
  55627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55627, CurrentID=0, Status='live', Symbol='SMPD4', Aliases='NET13, NSMASE-3, NSMASE3', description='sphingomyelin phosphodiesterase 4', other_designations='SKNY protein|neutral sphingomyelinase 3|neutral sphingomyelinase III|sphingomyelin phosphodiesterase 4, neutral membrane (neutral sphingomyelinase-3)', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130151392, end_position_on_the_genomic_accession=130182750, orientation='minus', exon_count=20, OMIM=610457, no_hdr0=''),
  55628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55628, CurrentID=0, Status='live', Symbol='ZNF407', Aliases='', description='zinc finger protein 407', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74597870, end_position_on_the_genomic_accession=75065672, orientation='plus', exon_count=13, OMIM=615894, no_hdr0=''),
  317773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317773, CurrentID=0, Status='live', Symbol='MGR6', Aliases='FHM3, MGR', description='Migraine, several forms', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607516, no_hdr0=''),
  55630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55630, CurrentID=0, Status='live', Symbol='SLC39A4', Aliases='AEZ, AWMS2, ZIP4', description='solute carrier family 39 (zinc transporter), member 4', other_designations='zrt- and Irt-like protein 4', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144412414, end_position_on_the_genomic_accession=144418034, orientation='minus', exon_count=12, OMIM=607059, no_hdr0=''),
  55631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55631, CurrentID=0, Status='live', Symbol='LRRC40', Aliases='dJ677H15.1', description='leucine rich repeat containing 40', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70144802, end_position_on_the_genomic_accession=70205678, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  55632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55632, CurrentID=0, Status='live', Symbol='G2E3', Aliases='KIAA1333, PHF7B', description='G2/M-phase specific E3 ubiquitin protein ligase', other_designations='G2/M-phase specific E3 ubiquitin ligase|PHD finger protein 7B', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30559123, end_position_on_the_genomic_accession=30619840, orientation='plus', exon_count=16, OMIM=611299, no_hdr0=''),
  55633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55633, CurrentID=0, Status='live', Symbol='TBC1D22B', Aliases='C6orf197', description='TBC1 domain family member 22B', other_designations='TBC1 domain family, member 22B', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37257704, end_position_on_the_genomic_accession=37332970, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  55634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55634, CurrentID=0, Status='live', Symbol='KRBOX4', Aliases='ZNF673', description='KRAB box domain containing 4', other_designations='KRAB box domain-containing protein 4|putative zinc finger transcription factor|zinc finger family member 673|zinc finger protein 673', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46447189, end_position_on_the_genomic_accession=46474639, orientation='plus', exon_count=7, OMIM=300585, no_hdr0=''),
  55635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55635, CurrentID=0, Status='live', Symbol='DEPDC1', Aliases='DEP.8-V2, DEPDC1A, SDP35, DEPDC1', description='DEP domain containing 1', other_designations='cell cycle control protein SDP35', map_location='1p31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68474152, end_position_on_the_genomic_accession=68497221, orientation='minus', exon_count=12, OMIM=612002, no_hdr0=''),
  55636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55636, CurrentID=0, Status='live', Symbol='CHD7', Aliases='CRG, HH5, IS3, KAL5', description='chromodomain helicase DNA binding protein 7', other_designations='ATP-dependent helicase CHD7|CHARGE association|chromodomain helicase DNA binding protein 7 isoform CRA_e', map_location='8q12.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60678762, end_position_on_the_genomic_accession=60868028, orientation='plus', exon_count=44, OMIM=608892, no_hdr0=''),
  317781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317781, CurrentID=0, Status='live', Symbol='DDX51', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 51', other_designations='DEAD box protein 51', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132136594, end_position_on_the_genomic_accession=132144335, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  55638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55638, CurrentID=0, Status='live', Symbol='SYBU', Aliases='GOLSYN, OCSYN, SNPHL', description='syntabulin', other_designations='GOLSYN A protein|GOLSYN B protein|GOLSYN C protein|Golgi-localized protein|golgi-localized syntaphilin-related protein|implicated in syntaxin trafficking in neurons|microtubule-associated protein|syntabulin (syntaxin-interacting)|syntaxin-1-binding protein', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109574176, end_position_on_the_genomic_accession=109691791, orientation='minus', exon_count=15, OMIM=611568, no_hdr0=''),
  55640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55640, CurrentID=0, Status='live', Symbol='FLVCR2', Aliases='C14orf58, CCT, EPV, FLVCRL14q, MFSD7C, PVHH', description='feline leukemia virus subgroup C cellular receptor family member 2', other_designations='calcium-chelate transporter|feline leukemia virus subgroup C cellular receptor family, member 2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75578597, end_position_on_the_genomic_accession=75648169, orientation='plus', exon_count=11, OMIM=610865, no_hdr0=''),
  55643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55643, CurrentID=0, Status='live', Symbol='BTBD2', Aliases='', description='BTB (POZ) domain containing 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1985438, end_position_on_the_genomic_accession=2015703, orientation='minus', exon_count=11, OMIM=608531, no_hdr0=''),
  55644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55644, CurrentID=0, Status='live', Symbol='OSGEP', Aliases='GCPL1, KAE11, PRSMG1, OSGEP', description='O-sialoglycoprotein endopeptidase', other_designations='N6-L-threonylcarbamoyladenine synthase|hOSGEP|probable O-sialoglycoprotein endopeptidase|probable tRNA threonylcarbamoyladenosine biosynthesis protein OSGEP|t(6)A synthase|t(6)A37 threonylcarbamoyladenosine biosynthesis protein OSGEP|tRNA threonylcarbamoyladenosine biosynthesis protein OSGEP', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20447048, end_position_on_the_genomic_accession=20455108, orientation='minus', exon_count=11, OMIM=610107, no_hdr0=''),
  55646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55646, CurrentID=0, Status='live', Symbol='LYAR', Aliases='ZC2HC2, ZLYAR', description='Ly1 antibody reactive', other_designations='Ly1 antibody reactive homolog', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4267701, end_position_on_the_genomic_accession=4290169, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55647, CurrentID=0, Status='live', Symbol='RAB20', Aliases='', description='RAB20, member RAS oncogene family', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110523066, end_position_on_the_genomic_accession=110561737, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55650, CurrentID=0, Status='live', Symbol='PIGV', Aliases='GPI-MT-II, HPMRS1, PIG-V', description='phosphatidylinositol glycan anchor biosynthesis class V', other_designations='GPI mannosyltransferase II|Ybr004c homolog|dol-P-Man dependent GPI mannosyltransferase|phosphatidylinositol glycan anchor biosynthesis, class V', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26787963, end_position_on_the_genomic_accession=26798403, orientation='plus', exon_count=4, OMIM=610274, no_hdr0=''),
  55651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55651, CurrentID=0, Status='live', Symbol='NHP2', Aliases='DKCB2P, NOLA2, NHP2', description='NHP2 ribonucleoprotein', other_designations='NHP2 ribonucleoprotein homolog|NHP2-like protein|nucleolar protein family A, member 2 (H/ACA small nucleolar RNPs)|snoRNP protein NHP2', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178149463, end_position_on_the_genomic_accession=178153960, orientation='minus', exon_count=4, OMIM=606470, no_hdr0=''),
  55652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55652, CurrentID=0, Status='live', Symbol='SLC48A1', Aliases='HRG-1, HRG1, hHRG-1', description='solute carrier family 48 (heme transporter), member 1', other_designations='heme-responsive gene 1 protein homolog|solute carrier family 48 member 1', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47753842, end_position_on_the_genomic_accession=47782753, orientation='plus', exon_count=8, OMIM=612187, no_hdr0=''),
  55653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55653, CurrentID=0, Status='live', Symbol='BCAS4', Aliases='CNOL', description='breast carcinoma amplified sequence 4', other_designations='BCAS4/BCAS3 fusion|breast carcinoma amplified sequence 4/3 fusion protein', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50794894, end_position_on_the_genomic_accession=50877854, orientation='plus', exon_count=8, OMIM=607471, no_hdr0=''),
  55654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55654, CurrentID=0, Status='live', Symbol='TMEM127', Aliases='', description='transmembrane protein 127', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96250208, end_position_on_the_genomic_accession=96266013, orientation='minus', exon_count=4, OMIM=613403, no_hdr0=''),
  55655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55655, CurrentID=0, Status='live', Symbol='NLRP2', Aliases='CLR19.9, NALP2, NBS1, PAN1, PYPAF2', description='NLR family, pyrin domain containing 2', other_designations='NACHT, LRR and PYD containing protein 2|NACHT, leucine rich repeat and PYD containing 2|PYRIN domain and NACHT domain-containing protein 1|PYRIN-containing APAF1-like protein 2|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 2|nucleotide-binding site protein 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54965284, end_position_on_the_genomic_accession=55001142, orientation='plus', exon_count=14, OMIM=609364, no_hdr0=''),
  55656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55656, CurrentID=0, Status='live', Symbol='INTS8', Aliases='C8orf52, INT8', description='integrator complex subunit 8', other_designations='protein kaonashi-1', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94813311, end_position_on_the_genomic_accession=94880493, orientation='plus', exon_count=30, OMIM=611351, no_hdr0=''),
  55657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55657, CurrentID=0, Status='live', Symbol='ZNF692', Aliases='AREBP, Zfp692', description='zinc finger protein 692', other_designations='AICAR responsive element binding protein', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248850004, end_position_on_the_genomic_accession=248860798, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  55658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55658, CurrentID=0, Status='live', Symbol='RNF126', Aliases='', description='ring finger protein 126', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=647526, end_position_on_the_genomic_accession=663313, orientation='minus', exon_count=9, OMIM=615177, no_hdr0=''),
  55659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55659, CurrentID=0, Status='live', Symbol='ZNF416', Aliases='', description='zinc finger protein 416', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57571010, end_position_on_the_genomic_accession=57579180, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  55660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55660, CurrentID=0, Status='live', Symbol='PRPF40A', Aliases='FBP-11, FBP11, FLAF1, FNBP3, HIP-10, HIP10, HYPA, NY-REN-6, Prp40', description='pre-mRNA processing factor 40 homolog A', other_designations='Fas-ligand associated factor 1|Huntingtin-interacting protein A|NY-REN-6 antigen|PRP40 pre-mRNA processing factor 40 homolog A|fas ligand-associated factor 1|formin binding protein 3|formin-binding protein 11|huntingtin yeast partner A|huntingtin-interacting protein 10|renal carcinoma antigen NY-REN-6', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152651593, end_position_on_the_genomic_accession=152717916, orientation='minus', exon_count=29, OMIM=612941, no_hdr0=''),
  55661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55661, CurrentID=0, Status='live', Symbol='DDX27', Aliases='DRS1, Drs1p, HSPC259, PP3241, RHLP, dJ686N3.1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 27', other_designations='DEAD box protein 27|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 27|Deficiency of ribosomal subunits protein 1 homolog|RNA helicase-like protein', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49219295, end_position_on_the_genomic_accession=49244077, orientation='plus', exon_count=22, OMIM=616621, no_hdr0=''),
  55662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55662, CurrentID=0, Status='live', Symbol='HIF1AN', Aliases='FIH1', description='hypoxia inducible factor 1, alpha subunit inhibitor', other_designations='FIH-1|factor inhibiting HIF-1|factor inhibiting HIF1|hypoxia-inducible factor asparagine hydroxylase|peptide-aspartate beta-dioxygenase', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100535884, end_position_on_the_genomic_accession=100553924, orientation='plus', exon_count=8, OMIM=606615, no_hdr0=''),
  55663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55663, CurrentID=0, Status='live', Symbol='ZNF446', Aliases='ZKSCAN20, ZSCAN30, ZSCAN52', description='zinc finger protein 446', other_designations='zinc finger protein with KRAB and SCAN domains 20', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58474543, end_position_on_the_genomic_accession=58489533, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55664, CurrentID=0, Status='live', Symbol='CDC37L1', Aliases='CDC37B, HARC', description='cell division cycle 37-like 1', other_designations='CDC37 cell division cycle 37 homolog-like 1|Hsp90-associating relative of Cdc37|cell division cycle 37 homolog-like 1', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4679553, end_position_on_the_genomic_accession=4708399, orientation='plus', exon_count=7, OMIM=610346, no_hdr0=''),
  55665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55665, CurrentID=0, Status='live', Symbol='URGCP', Aliases='URG4', description='upregulator of cell proliferation', other_designations='HBV X protein up-regulated gene 4 protein|HBxAg up-regulated gene 4 protein|up-regulated gene 4', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43875894, end_position_on_the_genomic_accession=43926411, orientation='minus', exon_count=8, OMIM=610337, no_hdr0=''),
  55666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55666, CurrentID=0, Status='live', Symbol='NPLOC4', Aliases='NPL4', description='NPL4 homolog, ubiquitin recognition factor', other_designations='NPLOC4 ubiquitin recognition factor|nuclear protein localization 4 homolog', map_location='17qter', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81556885, end_position_on_the_genomic_accession=81637153, orientation='minus', exon_count=21, OMIM=606590, no_hdr0=''),
  55667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55667, CurrentID=0, Status='live', Symbol='DENND4C', Aliases='C9orf55, C9orf55B, bA513M16.3', description='DENN/MADD domain containing 4C', other_designations='DENN/MADD domain containing 4A pseudogene', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19230765, end_position_on_the_genomic_accession=19374268, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  55668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55668, CurrentID=0, Status='live', Symbol='GPATCH2L', Aliases='C14orf118', description='G-patch domain containing 2 like', other_designations='G patch domain containing 2-like', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76151885, end_position_on_the_genomic_accession=76214473, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  55669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55669, CurrentID=0, Status='live', Symbol='MFN1', Aliases='hfzo1, hfzo2', description='mitofusin 1', other_designations='fzo homolog|mitochondrial transmembrane GTPase FZO-2|mitochondrial transmembrane GTPase Fzo-1|putative transmembrane GTPase|transmembrane GTPase MFN1', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179347692, end_position_on_the_genomic_accession=179393226, orientation='plus', exon_count=19, OMIM=608506, no_hdr0=''),
  55670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55670, CurrentID=0, Status='live', Symbol='PEX26', Aliases='PBD7A, PBD7BM1T, Pex26pM1T, PEX26', description='peroxisomal biogenesis factor 26', other_designations='peroxin-26|peroxisome biogenesis disorder, complementation group 8|peroxisome biogenesis disorder, complementation group A|peroxisome biogenesis factor 26', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18077920, end_position_on_the_genomic_accession=18091031, orientation='plus', exon_count=6, OMIM=608666, no_hdr0=''),
  55671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55671, CurrentID=0, Status='live', Symbol='PPP4R3A', Aliases='FLFL1, KIAA2010, MSTP033, PP4R3, PP4R3A, SMEK1, smk-1, smk1', description='protein phosphatase 4 regulatory subunit 3A', other_designations='SMEK homolog 1, suppressor of mek1|protein phosphatase 4, regulatory subunit 3A', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91457481, end_position_on_the_genomic_accession=91510554, orientation='minus', exon_count=18, OMIM=610351, no_hdr0=''),
  55672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55672, CurrentID=0, Status='live', Symbol='NBPF1', Aliases='AB13, AB14, AB23, AD2, NBG, NBPF', description='neuroblastoma breakpoint family member 1', other_designations='neuroblastoma breakpoint family, member 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16562427, end_position_on_the_genomic_accession=16613605, orientation='minus', exon_count=28, OMIM=610501, no_hdr0=''),
  55676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55676, CurrentID=0, Status='live', Symbol='SLC30A6', Aliases='MST103, MSTP103, ZNT6', description='solute carrier family 30 (zinc transporter), member 6', other_designations='solute carrier family 30 member 6|znT-6', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32165841, end_position_on_the_genomic_accession=32224112, orientation='plus', exon_count=17, OMIM=611148, no_hdr0=''),
  55677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55677, CurrentID=0, Status='live', Symbol='IWS1', Aliases='', description='IWS1 homolog (S. cerevisiae)', other_designations='IWS1-like protein|interacts with Spt6', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127480807, end_position_on_the_genomic_accession=127526886, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  55679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55679, CurrentID=0, Status='live', Symbol='LIMS2', Aliases='PINCH-2, PINCH2', description='LIM and senescent cell antigen-like domains 2', other_designations='ILK-binding protein|particularly interesting new Cys-His protein 2', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127638421, end_position_on_the_genomic_accession=127681786, orientation='minus', exon_count=16, OMIM=607908, no_hdr0=''),
  55680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55680, CurrentID=0, Status='live', Symbol='RUFY2', Aliases='RABIP4R, ZFYVE13', description='RUN and FYVE domain containing 2', other_designations='RUN and FYVE domain-containing 2|Run- and FYVE-domain containing protein|antigen MU-RMS-40.17|rab4-interacting protein related', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68341107, end_position_on_the_genomic_accession=68407294, orientation='minus', exon_count=22, OMIM=610328, no_hdr0=''),
  55681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55681, CurrentID=0, Status='live', Symbol='SCYL2', Aliases='CVAK104', description='SCY1-like, kinase-like 2', other_designations='SCY1-like 2|coated vesicle-associated kinase of 104 kDa', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100267137, end_position_on_the_genomic_accession=100341722, orientation='plus', exon_count=22, OMIM=616365, no_hdr0=''),
  55683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55683, CurrentID=0, Status='live', Symbol='KANSL3', Aliases='KIAA1310, NSL3, Rcd1', description='KAT8 regulatory NSL complex subunit 3', other_designations='NSL complex protein NSL3|non-specific lethal 3 homolog|serum inhibited-related protein|testis development protein PRTD', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96593155, end_position_on_the_genomic_accession=96638379, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  55684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55684, CurrentID=0, Status='live', Symbol='RABL6', Aliases='C9orf86, PARF, RBEL1, pp8875', description='RAB, member RAS oncogene family-like 6', other_designations='Rab-like GTP-binding protein 1C|partner of ARF|putative GTP-binding protein Parf', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136807922, end_position_on_the_genomic_accession=136841187, orientation='plus', exon_count=16, OMIM=610615, no_hdr0=''),
  55686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55686, CurrentID=0, Status='live', Symbol='MREG', Aliases='DSU, WDT2', description='melanoregulin', other_designations='dilute suppressor protein homolog|whn-dependent transcript 2', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215942590, end_position_on_the_genomic_accession=216013623, orientation='minus', exon_count=7, OMIM=609207, no_hdr0=''),
  55687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55687, CurrentID=0, Status='live', Symbol='TRMU', Aliases='LCAL3, MTO2, MTU1, TRMT, TRMT1, TRNT1', description='tRNA 5-methylaminomethyl-2-thiouridylate methyltransferase', other_designations='MTO2 homolog|lung cancer associated lncRNA 3|mitochondrial 5-methylaminomethyl-2-thiouridylate-methyltransferase', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46335401, end_position_on_the_genomic_accession=46357340, orientation='plus', exon_count=12, OMIM=610230, no_hdr0=''),
  55689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55689, CurrentID=0, Status='live', Symbol='YEATS2', Aliases='', description='YEATS domain containing 2', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183697804, end_position_on_the_genomic_accession=183812625, orientation='plus', exon_count=35, OMIM=613373, no_hdr0=''),
  55690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55690, CurrentID=0, Status='live', Symbol='PACS1', Aliases='MRD17', description='phosphofurin acidic cluster sorting protein 1', other_designations='PACS-1|cytosolic sorting protein PACS-1', map_location='11q13.1-q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66070353, end_position_on_the_genomic_accession=66244747, orientation='plus', exon_count=27, OMIM=607492, no_hdr0=''),
  55691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55691, CurrentID=0, Status='live', Symbol='FRMD4A', Aliases='FRMD4, bA295P9.4', description='FERM domain containing 4A', other_designations='FERM domain containing 4', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13643706, end_position_on_the_genomic_accession=14330867, orientation='minus', exon_count=36, OMIM=616305, no_hdr0=''),
  55692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55692, CurrentID=0, Status='live', Symbol='LUC7L', Aliases='LUC7B1, Luc7, SR+89, hLuc7B1', description='LUC7-like', other_designations='putative SR protein LUC7B1|sarcoplasmic reticulum protein LUC7B1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=188971, end_position_on_the_genomic_accession=229450, orientation='minus', exon_count=13, OMIM=607782, no_hdr0=''),
  55693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55693, CurrentID=0, Status='live', Symbol='KDM4D', Aliases='JMJD2D', description='lysine (K)-specific demethylase 4D', other_designations='jmjC domain-containing histone demethylation protein 3D|jumonji domain containing 2D|jumonji domain-containing protein 2D', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94973681, end_position_on_the_genomic_accession=94999514, orientation='plus', exon_count=3, OMIM=609766, no_hdr0=''),
  55695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55695, CurrentID=0, Status='live', Symbol='NSUN5', Aliases='NOL1, NOL1RA, WBSCR20, WBSCR20A, p120, p120(NOL1), NSUN5', description='NOP2/Sun RNA methyltransferase family member 5', other_designations='NOL1-related protein|NOL1/NOP2/Sun domain family member 5|NOP2/Sun domain family, member 5|NOP2/Sun domain family, member 5A|Williams Beuren syndrome chromosome region 20A|Williams-Beuren syndrome chromosomal region 20A protein|Williams-Beuren syndrome critical region protein 20 copy A|putative methyltransferase NSUN5', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73302516, end_position_on_the_genomic_accession=73308867, orientation='minus', exon_count=10, OMIM=615732, no_hdr0=''),
  55696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55696, CurrentID=0, Status='live', Symbol='RBM22', Aliases='Cwc2, ZC3H16, fSAP47', description='RNA binding motif protein 22', other_designations='RNA-binding motif protein 22|functional spliceosome-associated protein 47|zinc finger CCCH domain-containing protein 16', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150690790, end_position_on_the_genomic_accession=150701107, orientation='minus', exon_count=11, OMIM=612430, no_hdr0=''),
  55697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55697, CurrentID=0, Status='live', Symbol='VAC14', Aliases='ArPIKfyve, TAX1BP2, TRX', description='Vac14 homolog (S. cerevisiae)', other_designations='Tax1 (human T-cell leukemia virus type I) binding protein 1|Tax1 (human T-cell leukemia virus type I) binding protein 2|tax1-binding protein 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70687439, end_position_on_the_genomic_accession=70801166, orientation='minus', exon_count=24, OMIM=604632, no_hdr0=''),
  55698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55698, CurrentID=0, Status='live', Symbol='RADIL', Aliases='RASIP2', description='Ras association and DIL domains', other_designations='Rap GTPase interactor', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4799109, end_position_on_the_genomic_accession=4883704, orientation='minus', exon_count=15, OMIM=611491, no_hdr0=''),
  55699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55699, CurrentID=0, Status='live', Symbol='IARS2', Aliases='CAGSSS, ILERS', description='isoleucyl-tRNA synthetase 2, mitochondrial', other_designations='isoleucine tRNA ligase 2, mitochondrial|isoleucine-tRNA synthetase 2, mitochondrial|isoleucyl-tRNA synthetase, mitochondrial|mitochondrial isoleucine tRNA synthetase', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220094113, end_position_on_the_genomic_accession=220148041, orientation='plus', exon_count=23, OMIM=612801, no_hdr0=''),
  55700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55700, CurrentID=0, Status='live', Symbol='MAP7D1', Aliases='PARCC1, RPRC1', description='MAP7 domain containing 1', other_designations='arginine/proline rich coiled-coil 1|arginine/proline-rich coiled-coil domain-containing protein 1|proline arginine rich coiled coil 1|proline/arginine-rich coiled-coil domain-containing protein 1', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36155965, end_position_on_the_genomic_accession=36180849, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  55701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55701, CurrentID=0, Status='live', Symbol='ARHGEF40', Aliases='SOLO', description='Rho guanine nucleotide exchange factor 40', other_designations='Rho guanine nucleotide exchange factor (GEF) 40|protein SOLO', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21070226, end_position_on_the_genomic_accession=21090248, orientation='plus', exon_count=25, OMIM=610018, no_hdr0=''),
  55702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55702, CurrentID=0, Status='live', Symbol='CCDC94', Aliases='', description='coiled-coil domain containing 94', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4247114, end_position_on_the_genomic_accession=4269088, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55703, CurrentID=0, Status='live', Symbol='POLR3B', Aliases='C128, HLD8, INMAP, RPC2', description='polymerase (RNA) III (DNA directed) polypeptide B', other_designations='DNA-directed RNA polymerase III 127.6 kDa polypeptide|DNA-directed RNA polymerase III subunit B|RNA polymerase III subunit C2|interphase nucleus and mitotic apparatus associated protein', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106357658, end_position_on_the_genomic_accession=106510198, orientation='plus', exon_count=29, OMIM=614366, no_hdr0=''),
  55704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55704, CurrentID=0, Status='live', Symbol='CCDC88A', Aliases='APE, GIRDIN, GIV, GRDN, HkRP1, KIAA1212', description='coiled-coil domain containing 88A', other_designations='AKT-phosphorylation enhancer|Hook-related protein 1|g alpha-interacting vesicle-associated protein|girders of actin filament|girders of actin filaments', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55287842, end_position_on_the_genomic_accession=55419921, orientation='minus', exon_count=36, OMIM=609736, no_hdr0=''),
  55705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55705, CurrentID=0, Status='live', Symbol='IPO9', Aliases='Imp9', description='importin 9', other_designations='ran-binding protein 9|ranBP9', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201829160, end_position_on_the_genomic_accession=201884294, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  55706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55706, CurrentID=0, Status='live', Symbol='NDC1', Aliases='NET3, TMEM48', description='NDC1 transmembrane nucleoporin', other_designations='nuclear division cycle 1 homolog|transmembrane protein 48', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53765460, end_position_on_the_genomic_accession=53838552, orientation='minus', exon_count=19, OMIM=610115, no_hdr0=''),
  55707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55707, CurrentID=0, Status='live', Symbol='NECAP2', Aliases='', description='NECAP endocytosis associated 2', other_designations='NECAP endocytosis-associated protein 2|NECAP-2|adaptin-ear-binding coat-associated protein 2', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16440672, end_position_on_the_genomic_accession=16460090, orientation='plus', exon_count=9, OMIM=611624, no_hdr0=''),
  55709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55709, CurrentID=0, Status='live', Symbol='KBTBD4', Aliases='BKLHD4, HSPC252', description='kelch repeat and BTB (POZ) domain containing 4', other_designations='BTB and kelch domain containing 4|BTB and kelch domain-containing protein 4', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47572197, end_position_on_the_genomic_accession=47579015, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  55711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55711, CurrentID=0, Status='live', Symbol='FAR2', Aliases='HEL-S-81, MLSTD1, SDR10E2', description='fatty acyl-CoA reductase 2', other_designations='epididymis secretory protein Li 81|fatty acyl CoA reductase 2|male sterility domain-containing protein 1|short chain dehydrogenase/reductase family 10E, member 2', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29149003, end_position_on_the_genomic_accession=29335616, orientation='plus', exon_count=14, OMIM=616156, no_hdr0=''),
  55713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55713, CurrentID=0, Status='live', Symbol='ZNF334', Aliases='', description='zinc finger protein 334', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46481908, end_position_on_the_genomic_accession=46513581, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  55714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55714, CurrentID=0, Status='live', Symbol='TENM3', Aliases='MCOPCB9, ODZ3, TNM3, Ten-m3, ten-3', description='teneurin transmembrane protein 3', other_designations='ODZ3-like protein|odz, odd Oz/ten-m homolog 3|protein Odd Oz/ten-m homolog 3|tenascin-M3', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182323984, end_position_on_the_genomic_accession=182803024, orientation='plus', exon_count=31, OMIM=610083, no_hdr0=''),
  55715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55715, CurrentID=0, Status='live', Symbol='DOK4', Aliases='IRS-5, IRS5', description='docking protein 4', other_designations='downstream of tyrosine kinase 4|insulin receptor substrate 5', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57471958, end_position_on_the_genomic_accession=57486473, orientation='minus', exon_count=9, OMIM=608333, no_hdr0=''),
  55716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55716, CurrentID=0, Status='live', Symbol='LMBR1L', Aliases='LIMR', description='limb development membrane protein 1-like', other_designations='limb region 1 homolog-like|limb region 1 protein homolog-like|limb region 1-like protein -like|lipocalin-1 interacting membrane receptor|lipocalin-1-interacting membrane receptor|lipocalin-interacting membrane receptor', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49097139, end_position_on_the_genomic_accession=49110900, orientation='minus', exon_count=20, OMIM=610007, no_hdr0=''),
  55717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55717, CurrentID=0, Status='live', Symbol='WDR11', Aliases='BRWD2, DR11, HH14, WDR15', description='WD repeat domain 11', other_designations='WD repeat domain 15|WD repeat-containing protein 15|bromodomain and WD repeat-containing protein 2', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120851175, end_position_on_the_genomic_accession=120909526, orientation='plus', exon_count=30, OMIM=606417, no_hdr0=''),
  55718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55718, CurrentID=0, Status='live', Symbol='POLR3E', Aliases='RPC5, SIN', description='polymerase (RNA) III (DNA directed) polypeptide E (80kD)', other_designations='DNA-directed RNA polymerase III 80 kDa polypeptide|RNA polymerase III 80 kDa subunit RPC5|RNA polymerase III subunit C5', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22297366, end_position_on_the_genomic_accession=22335103, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  55719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55719, CurrentID=0, Status='live', Symbol='SLF2', Aliases='C10orf6, FAM178A', description='SMC5-SMC6 complex localization factor 2', other_designations='family with sequence similarity 178, member A|protein FAM178A|smc5/6 localization factor 1', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100912569, end_position_on_the_genomic_accession=100965134, orientation='plus', exon_count=22, OMIM=610348, no_hdr0=''),
  55720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55720, CurrentID=0, Status='live', Symbol='TSR1', Aliases='', description='TSR1, 20S rRNA accumulation, homolog (S. cerevisiae)', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2322678, end_position_on_the_genomic_accession=2337384, orientation='minus', exon_count=15, OMIM=611214, no_hdr0=''),
  55721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55721, CurrentID=0, Status='live', Symbol='IQCC', Aliases='', description='IQ motif containing C', other_designations='', map_location='1p36.11-p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32205635, end_position_on_the_genomic_accession=32208687, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  55722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55722, CurrentID=0, Status='live', Symbol='CEP72', Aliases='', description='centrosomal protein 72kDa', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=612290, end_position_on_the_genomic_accession=663500, orientation='plus', exon_count=19, OMIM=616475, no_hdr0=''),
  55723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55723, CurrentID=0, Status='live', Symbol='ASF1B', Aliases='CIA-II', description='anti-silencing function 1B histone chaperone', other_designations='ASF1 anti-silencing function 1 homolog B|CCG1-interacting factor A-II|anti-silencing function protein 1 homolog B|hAsf1|hAsf1b|hCIA-II', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14119509, end_position_on_the_genomic_accession=14136628, orientation='minus', exon_count=4, OMIM=609190, no_hdr0=''),
  105369330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369330, CurrentID=0, Status='live', Symbol='LOC105369330', Aliases='', description='uncharacterized LOC105369330', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61511021, end_position_on_the_genomic_accession=61513544, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55726, CurrentID=0, Status='live', Symbol='ASUN', Aliases='C12orf11, GCT1, Mat89Bb, NET48, SPATA30', description='asunder spermatogenesis regulator', other_designations='asunder, spermatogenesis regulator homolog (Drosphila)|cell cycle regulator Mat89Bb homolog|germ cell tumor 1|sarcoma antigen NY-SAR-95|spermatogenesis associated 30', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26905057, end_position_on_the_genomic_accession=26938326, orientation='minus', exon_count=18, OMIM=615079, no_hdr0=''),
  55727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55727, CurrentID=0, Status='live', Symbol='BTBD7', Aliases='FUP1', description='BTB (POZ) domain containing 7', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93237550, end_position_on_the_genomic_accession=93333092, orientation='minus', exon_count=13, OMIM=610386, no_hdr0=''),
  55728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55728, CurrentID=0, Status='live', Symbol='N4BP2', Aliases='B3BP', description='NEDD4 binding protein 2', other_designations='BCL-3-binding protein', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40056212, end_position_on_the_genomic_accession=40185798, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55729, CurrentID=0, Status='live', Symbol='ATF7IP', Aliases='AM, ATF-IP, MCAF, MCAF1, p621', description='activating transcription factor 7 interacting protein', other_designations='ATF-interacting protein|ATF7-interacting protein|ATFa-associated modulator|MBD1-containing chromatin-associated factor 1', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14365632, end_position_on_the_genomic_accession=14502935, orientation='plus', exon_count=21, OMIM=613644, no_hdr0=''),
  55731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55731, CurrentID=0, Status='live', Symbol='FAM222B', Aliases='C17orf63', description='family with sequence similarity 222 member B', other_designations='family with sequence similarity 222, member B|uncharacterized protein C17orf63', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28755978, end_position_on_the_genomic_accession=28854985, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55732, CurrentID=0, Status='live', Symbol='C1orf112', Aliases='', description='chromosome 1 open reading frame 112', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169795005, end_position_on_the_genomic_accession=169853088, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  55733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55733, CurrentID=0, Status='live', Symbol='HHAT', Aliases='MART2, SKI1, Skn', description='hedgehog acyltransferase', other_designations='melanoma antigen recognized by T cells 2|melanoma antigen recognized by T-cells 2|skinny hedgehog protein 1', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210328252, end_position_on_the_genomic_accession=210676298, orientation='plus', exon_count=16, OMIM=605743, no_hdr0=''),
  55734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55734, CurrentID=0, Status='live', Symbol='ZFP64', Aliases='ZNF338', description='ZFP64 zinc finger protein', other_designations='zinc finger protein 338|zinc finger protein 64 homolog', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52084011, end_position_on_the_genomic_accession=52191985, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  55735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55735, CurrentID=0, Status='live', Symbol='DNAJC11', Aliases='dJ126A5.1', description='DnaJ heat shock protein family (Hsp40) member C11', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 11|novel DnaJ domain-containing protein', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6634168, end_position_on_the_genomic_accession=6701906, orientation='minus', exon_count=16, OMIM=614827, no_hdr0=''),
  55737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55737, CurrentID=0, Status='live', Symbol='VPS35', Aliases='MEM3, PARK17', description='VPS35 retromer complex component', other_designations='hVPS35|maternal-embryonic 3|vacuolar protein sorting 35 homolog', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46657995, end_position_on_the_genomic_accession=46689232, orientation='minus', exon_count=18, OMIM=601501, no_hdr0=''),
  55738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55738, CurrentID=0, Status='live', Symbol='ARFGAP1', Aliases='ARF1GAP, HRIHFB2281', description='ADP ribosylation factor GTPase activating protein 1', other_designations='ADP-ribosylation factor GTPase activating protein 1|ARF1-directed GTPase-activating protein', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63272785, end_position_on_the_genomic_accession=63289793, orientation='plus', exon_count=17, OMIM=608377, no_hdr0=''),
  55739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55739, CurrentID=0, Status='live', Symbol='CARKD', Aliases='LP3298', description='carbohydrate kinase domain containing', other_designations='ATP-dependent NAD(P)HX dehydratase|carbohydrate kinase domain-containing protein', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110615460, end_position_on_the_genomic_accession=110639996, orientation='plus', exon_count=10, OMIM=615910, no_hdr0=''),
  55740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55740, CurrentID=0, Status='live', Symbol='ENAH', Aliases='ENA, MENA, NDPP1', description='enabled homolog (Drosophila)', other_designations='mammalian enabled variant 11a|mammalian enabled variant pan', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225486832, end_position_on_the_genomic_accession=225661676, orientation='minus', exon_count=21, OMIM=609061, no_hdr0=''),
  55741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55741, CurrentID=0, Status='live', Symbol='EDEM2', Aliases='C20orf31, C20orf49, bA4204.1', description='ER degradation enhancer, mannosidase alpha-like 2', other_designations='ER degradation-enhancing alpha-mannosidase-like 2|ER degradation-enhancing-mannosidase-like protein 2', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35115357, end_position_on_the_genomic_accession=35147358, orientation='minus', exon_count=11, OMIM=610302, no_hdr0=''),
  55742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55742, CurrentID=0, Status='live', Symbol='PARVA', Aliases='CH-ILKBP, MXRA2', description='parvin alpha', other_designations='actopaxin|calponin-like integrin-linked kinase-binding protein|matrix-remodeling-associated protein 2|parvin, alpha', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12376441, end_position_on_the_genomic_accession=12535356, orientation='plus', exon_count=14, OMIM=608120, no_hdr0=''),
  55743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55743, CurrentID=0, Status='live', Symbol='CHFR', Aliases='RNF116, RNF196', description='checkpoint with forkhead and ring finger domains, E3 ubiquitin protein ligase', other_designations='RING finger protein 196', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132840352, end_position_on_the_genomic_accession=132887618, orientation='minus', exon_count=19, OMIM=605209, no_hdr0=''),
  55744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55744, CurrentID=0, Status='live', Symbol='COA1', Aliases='C7orf44, MITRAC15', description='cytochrome c oxidase assembly factor 1 homolog', other_designations='cytochrome c oxidase assembly protein 1 homolog|mitochondrial translation regulation assembly intermediate of cytochrome c oxidase protein of 15 kDa', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43608452, end_position_on_the_genomic_accession=43729541, orientation='minus', exon_count=13, OMIM=614769, no_hdr0=''),
  55745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55745, CurrentID=0, Status='live', Symbol='AP5M1', Aliases='C14orf108, MUDENG, Mu5, MuD', description='adaptor related protein complex 5, mu 1 subunit', other_designations='AP-5 complex subunit mu|MHD domain-containing death-inducing protein|MU-2/AP1M2 domain containing, death-inducing|Mu-2 related death-inducing|adapter-related protein complex 5 mu subunit|adapter-related protein complex 5 subunit mu-1|adaptor-related protein complex 5 subunit mu-1|adaptor-related protein complex 5, mu 1 subunit|mu-2-related death-inducing protein|putative HIV-1 infection-related protein', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57268888, end_position_on_the_genomic_accession=57298742, orientation='plus', exon_count=9, OMIM=614368, no_hdr0=''),
  55746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55746, CurrentID=0, Status='live', Symbol='NUP133', Aliases='hNUP133', description='nucleoporin 133kDa', other_designations='133 kDa nucleoporin|nucleoporin 133kD|nucleoporin Nup133', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229441297, end_position_on_the_genomic_accession=229508341, orientation='minus', exon_count=26, OMIM=607613, no_hdr0=''),
  448963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448963, CurrentID=0, Status='live', Symbol='DFNB51', Aliases='', description='deafness, autosomal recessive 51', other_designations='', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609941, no_hdr0=''),
  55748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55748, CurrentID=0, Status='live', Symbol='CNDP2', Aliases='CN2, CPGL, HEL-S-13, HsT2298, PEPA', description='CNDP dipeptidase 2 (metallopeptidase M20 family)', other_designations='cytosolic nonspecific dipeptidase|epididymis secretory protein Li 13|glutamate carboxypeptidase-like protein 1|peptidase A', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74496265, end_position_on_the_genomic_accession=74523454, orientation='plus', exon_count=15, OMIM=169800, no_hdr0=''),
  55749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55749, CurrentID=0, Status='live', Symbol='CCAR1', Aliases='', description='cell division cycle and apoptosis regulator 1', other_designations='cell cycle and apoptosis regulatory protein 1|death inducer with SAP domain', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68721144, end_position_on_the_genomic_accession=68792377, orientation='plus', exon_count=25, OMIM=612569, no_hdr0=''),
  55750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55750, CurrentID=0, Status='live', Symbol='AGK', Aliases='CATC5, CTRCT38, MTDPS10, MULK', description='acylglycerol kinase', other_designations='hAGK|hsMuLK|multi-substrate lipid kinase|multiple substrate lipid kinase', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141551278, end_position_on_the_genomic_accession=141654409, orientation='plus', exon_count=18, OMIM=610345, no_hdr0=''),
  55751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55751, CurrentID=0, Status='live', Symbol='TMEM184C', Aliases='TMEM34', description='transmembrane protein 184C', other_designations='transmembrane protein 34', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147617371, end_position_on_the_genomic_accession=147635523, orientation='plus', exon_count=10, OMIM=613937, no_hdr0=''),
  55752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55752, CurrentID=0, Status='live', Symbol='SEPT11', Aliases='', description='septin 11', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76949742, end_position_on_the_genomic_accession=77040154, orientation='plus', exon_count=13, OMIM=612887, no_hdr0=''),
  55753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55753, CurrentID=0, Status='live', Symbol='OGDHL', Aliases='', description='oxoglutarate dehydrogenase-like', other_designations='2-oxoglutarate dehydrogenase complex component E1-like|OGDC-E1-like|alpha-ketoglutarate dehydrogenase-like', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49734641, end_position_on_the_genomic_accession=49762379, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  55754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55754, CurrentID=0, Status='live', Symbol='TMEM30A', Aliases='C6orf67, CDC50A', description='transmembrane protein 30A', other_designations='P4-ATPase flippase complex beta subunit TMEM30A', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75252922, end_position_on_the_genomic_accession=75284916, orientation='minus', exon_count=7, OMIM=611028, no_hdr0=''),
  55755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55755, CurrentID=0, Status='live', Symbol='CDK5RAP2', Aliases='C48, Cep215, MCPH3', description='CDK5 regulatory subunit associated protein 2', other_designations='CDK5 activator-binding protein C48|centrosomal protein 215 kDa|centrosomin', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120388869, end_position_on_the_genomic_accession=120580170, orientation='minus', exon_count=39, OMIM=608201, no_hdr0=''),
  55756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55756, CurrentID=0, Status='live', Symbol='INTS9', Aliases='CPSF2L, INT9, RC74', description='integrator complex subunit 9', other_designations='protein related to CPSF subunits of 74 kDa', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28767658, end_position_on_the_genomic_accession=28890389, orientation='minus', exon_count=20, OMIM=611352, no_hdr0=''),
  55757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55757, CurrentID=0, Status='live', Symbol='UGGT2', Aliases='HUGT2, UGCGL2, UGT2', description='UDP-glucose glycoprotein glucosyltransferase 2', other_designations='UDP--Glc:glycoprotein glucosyltransferase 2|UDP-Glc:glycoprotein glucosyltransferase 2|UDP-glucose ceramide glucosyltransferase-like 1|UDP-glucose ceramide glucosyltransferase-like 2', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95801580, end_position_on_the_genomic_accession=96053482, orientation='minus', exon_count=44, OMIM=605898, no_hdr0=''),
  55758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55758, CurrentID=0, Status='live', Symbol='RCOR3', Aliases='', description='REST corepressor 3', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211259366, end_position_on_the_genomic_accession=211316385, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55759, CurrentID=0, Status='live', Symbol='WDR12', Aliases='YTM1', description='WD repeat domain 12', other_designations='WD repeat-containing protein 12', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202880600, end_position_on_the_genomic_accession=202912226, orientation='minus', exon_count=13, OMIM=616620, no_hdr0=''),
  55760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55760, CurrentID=0, Status='live', Symbol='DHX32', Aliases='DDX32, DHLP1', description='DEAH (Asp-Glu-Ala-His) box polypeptide 32', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 32|DEAD/H box 32|DEAD/H helicase-like protein 1|DEAD/H helicase-like protein-1|DEAH box protein 32|huDDX32', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125836340, end_position_on_the_genomic_accession=125881315, orientation='minus', exon_count=11, OMIM=607960, no_hdr0=''),
  55761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55761, CurrentID=0, Status='live', Symbol='TTC17', Aliases='', description='tetratricopeptide repeat domain 17', other_designations='TPR repeat protein 17', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43358885, end_position_on_the_genomic_accession=43494933, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  55762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55762, CurrentID=0, Status='live', Symbol='ZNF701', Aliases='', description='zinc finger protein 701', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52570273, end_position_on_the_genomic_accession=52587174, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55763, CurrentID=0, Status='live', Symbol='EXOC1', Aliases='BM-102, SEC3, SEC3L1, SEC3P', description='exocyst complex component 1', other_designations='SEC3-like 1|exocyst complex component Sec3', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55853636, end_position_on_the_genomic_accession=55905078, orientation='plus', exon_count=22, OMIM=607879, no_hdr0=''),
  55764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55764, CurrentID=0, Status='live', Symbol='IFT122', Aliases='CED, CED1, SPG, WDR10, WDR10p, WDR140', description='intraflagellar transport 122', other_designations='WD repeat domain 10|WD repeat-containing protein 10|WD repeat-containing protein 140|intraflagellar transport 122 homolog', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129440036, end_position_on_the_genomic_accession=129520507, orientation='plus', exon_count=34, OMIM=606045, no_hdr0=''),
  55765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55765, CurrentID=0, Status='live', Symbol='C1orf106', Aliases='', description='chromosome 1 open reading frame 106', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200891074, end_position_on_the_genomic_accession=200915736, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  55766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55766, CurrentID=0, Status='live', Symbol='H2AFJ', Aliases='H2AJ', description='H2A histone family member J', other_designations='H2A histone family, member J|H2a/j|buforin I', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14774336, end_position_on_the_genomic_accession=14778002, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369337, CurrentID=0, Status='live', Symbol='LOC105369337', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63836865, end_position_on_the_genomic_accession=63839417, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55768, CurrentID=0, Status='live', Symbol='NGLY1', Aliases='CDDG, CDG1V, PNG1, PNGase', description='N-glycanase 1', other_designations='hPNGase|peptide:N-glycanase', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25718944, end_position_on_the_genomic_accession=25790039, orientation='minus', exon_count=16, OMIM=610661, no_hdr0=''),
  55769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55769, CurrentID=0, Status='live', Symbol='ZNF83', Aliases='HPF1, ZNF816B', description='zinc finger protein 83', other_designations='zinc finger protein 816B|zinc finger protein HPF1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52612364, end_position_on_the_genomic_accession=52690581, orientation='minus', exon_count=13, OMIM=194558, no_hdr0=''),
  55770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55770, CurrentID=0, Status='live', Symbol='EXOC2', Aliases='SEC5, SEC5L1, Sec5p', description='exocyst complex component 2', other_designations='SEC5-like 1|exocyst complex component Sec5', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=485138, end_position_on_the_genomic_accession=693141, orientation='minus', exon_count=30, OMIM=615329, no_hdr0=''),
  55771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55771, CurrentID=0, Status='live', Symbol='PRR11', Aliases='', description='proline rich 11', other_designations='transcription repressor of MHCII', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59155499, end_position_on_the_genomic_accession=59206709, orientation='plus', exon_count=10, OMIM=615920, no_hdr0=''),
  55773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55773, CurrentID=0, Status='live', Symbol='TBC1D23', Aliases='NS4ATP1', description='TBC1 domain family member 23', other_designations='HCV non-structural protein 4A-transactivated protein 1|HCV nonstructural protein 4A-transactivated protein 1|TBC1 domain family, member 23', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100260817, end_position_on_the_genomic_accession=100325252, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  448990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448990, CurrentID=0, Status='live', Symbol='TAS2R6P', Aliases='PS3, T2R06, T2R6, TAS2R6', description='taste 2 receptor member 6 pseudogene', other_designations='bitter taste receptor PS3|taste receptor, type 2, member 6, pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141787814, end_position_on_the_genomic_accession=141788636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55775, CurrentID=0, Status='live', Symbol='TDP1', Aliases='', description='tyrosyl-DNA phosphodiesterase 1', other_designations='tyr-DNA phosphodiesterase 1', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89954935, end_position_on_the_genomic_accession=90044764, orientation='plus', exon_count=22, OMIM=607198, no_hdr0=''),
  55776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55776, CurrentID=0, Status='live', Symbol='SAYSD1', Aliases='C6orf64', description='SAYSVFN motif domain containing 1', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39104063, end_position_on_the_genomic_accession=39115186, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55777, CurrentID=0, Status='live', Symbol='MBD5', Aliases='MRD1', description='methyl-CpG binding domain protein 5', other_designations='methyl-CpG-binding protein MBD5', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148021011, end_position_on_the_genomic_accession=148513477, orientation='plus', exon_count=19, OMIM=611472, no_hdr0=''),
  55778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55778, CurrentID=0, Status='live', Symbol='ZNF839', Aliases='C14orf131', description='zinc finger protein 839', other_designations='renal carcinoma antigen NY-REN-50', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102317377, end_position_on_the_genomic_accession=102343174, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  55779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55779, CurrentID=0, Status='live', Symbol='CFAP44', Aliases='WDR52', description='cilia and flagella associated protein 44', other_designations='WD repeat domain 52|WD repeat-containing protein 52', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113286930, end_position_on_the_genomic_accession=113445190, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  55780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55780, CurrentID=0, Status='live', Symbol='ERMARD', Aliases='C6orf70, PVNH6, dJ266L20.3', description='ER membrane-associated RNA degradation', other_designations='transmembrane protein C6orf70', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169751622, end_position_on_the_genomic_accession=169781584, orientation='plus', exon_count=20, OMIM=615532, no_hdr0=''),
  55781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55781, CurrentID=0, Status='live', Symbol='RIOK2', Aliases='RIO2', description='RIO kinase 2', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97160867, end_position_on_the_genomic_accession=97183301, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55783, CurrentID=0, Status='live', Symbol='CMTR2', Aliases='AFT, FTSJD1, HMTr2, MTr2', description='cap methyltransferase 2', other_designations="FtsJ methyltransferase domain containing 1|adrift homolog|cap2 2'O-ribose methyltransferase 2|ftsJ methyltransferase domain-containing protein 1|protein adrift homolog", map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71282300, end_position_on_the_genomic_accession=71289606, orientation='minus', exon_count=3, OMIM=616190, no_hdr0=''),
  55784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55784, CurrentID=0, Status='live', Symbol='MCTP2', Aliases='', description='multiple C2 domains, transmembrane 2', other_designations='multiple C2-domains with two transmembrane regions 2 1', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94231538, end_position_on_the_genomic_accession=94483952, orientation='plus', exon_count=29, OMIM=616297, no_hdr0=''),
  55785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55785, CurrentID=0, Status='live', Symbol='FGD6', Aliases='ZFYVE24', description='FYVE, RhoGEF and PH domain containing 6', other_designations='zinc finger FYVE domain-containing protein 24', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95076749, end_position_on_the_genomic_accession=95217464, orientation='minus', exon_count=21, OMIM=613520, no_hdr0=''),
  55786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55786, CurrentID=0, Status='live', Symbol='ZNF415', Aliases='', description='zinc finger protein 415', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53107879, end_position_on_the_genomic_accession=53133077, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55787, CurrentID=0, Status='live', Symbol='TXLNG', Aliases='CXorf15, ELRG, FIAT, LSR5X, TXLNG', description='taxilin gamma', other_designations='environmental LPS-responding|environmental lipopolysaccharide-responding gene protein|factor inhibiting ATF4-mediated transcription|factor inhibiting activating transcription factor 4 (ATF4)-mediated transcription|lipopolysaccharide specific response-5 protein|lipopolysaccharide-specific response protein 5', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16786432, end_position_on_the_genomic_accession=16844519, orientation='plus', exon_count=12, OMIM=300677, no_hdr0=''),
  55788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55788, CurrentID=0, Status='live', Symbol='LMBRD1', Aliases='C6orf209, LMBD1, MAHCF, NESI', description='LMBR1 domain containing 1', other_designations='HDAg-L-interacting protein NESI|hepatitis delta antigen-L interacting protein|liver regeneration p-53 related protein|nuclear export signal-interacting protein', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=69675749, end_position_on_the_genomic_accession=69797157, orientation='minus', exon_count=18, OMIM=612625, no_hdr0=''),
  55789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55789, CurrentID=0, Status='live', Symbol='DEPDC1B', Aliases='BRCC3, XTP1', description='DEP domain containing 1B', other_designations='HBV X-transactivated gene 8 protein|HBV XAg-transactivated protein 8|HBxAg transactivated protein 1|breast cancer cell 3', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60596912, end_position_on_the_genomic_accession=60700166, orientation='minus', exon_count=14, OMIM=616073, no_hdr0=''),
  55790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55790, CurrentID=0, Status='live', Symbol='CSGALNACT1', Aliases='CSGalNAcT-1, ChGn, beta4GalNAcT', description='chondroitin sulfate N-acetylgalactosaminyltransferase 1', other_designations='beta4GalNAcT-1|chondroitin beta-1,4-N-acetylgalactosaminyltransferase 1|chondroitin beta1,4 N-acetylgalactosaminyltransferase', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19404161, end_position_on_the_genomic_accession=19757271, orientation='minus', exon_count=25, OMIM=616615, no_hdr0=''),
  55791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55791, CurrentID=0, Status='live', Symbol='LRIF1', Aliases='C1orf103, RIF1', description='ligand dependent nuclear receptor interacting factor 1', other_designations='receptor interacting factor 1|receptor-interacting factor 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110947190, end_position_on_the_genomic_accession=110964197, orientation='minus', exon_count=4, OMIM=615354, no_hdr0=''),
  55793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55793, CurrentID=0, Status='live', Symbol='FAM63A', Aliases='', description='family with sequence similarity 63 member A', other_designations='family with sequence similarity 63, member A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150996825, end_position_on_the_genomic_accession=151008378, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55794, CurrentID=0, Status='live', Symbol='DDX28', Aliases='MDDX28', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 28', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 28|mitochondrial DEAD box protein 28|mitochondrial DEAD-box polypeptide 28', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68021274, end_position_on_the_genomic_accession=68023867, orientation='minus', exon_count=1, OMIM=607618, no_hdr0=''),
  55795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55795, CurrentID=0, Status='live', Symbol='PCID2', Aliases='F10', description='PCI domain containing 2', other_designations='CSN12-like protein', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113177539, end_position_on_the_genomic_accession=113209466, orientation='minus', exon_count=18, OMIM=613713, no_hdr0=''),
  55796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55796, CurrentID=0, Status='live', Symbol='MBNL3', Aliases='CHCR, MBLX, MBLX39, MBXL', description='muscleblind-like splicing regulator 3', other_designations='Cys3His CCG1-required protein|muscleblind-like 3|muscleblind-like X-linked protein', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132369314, end_position_on_the_genomic_accession=132490030, orientation='minus', exon_count=17, OMIM=300413, no_hdr0=''),
  449013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449013, CurrentID=0, Status='live', Symbol='ANIB2', Aliases='AIB2', description='aneurysm, intracranial berry 2', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608542, no_hdr0=''),
  55798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55798, CurrentID=0, Status='live', Symbol='METTL2B', Aliases='METL, METTL2, METTL2A, PSENIP1', description='methyltransferase like 2B', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128476729, end_position_on_the_genomic_accession=128502924, orientation='plus', exon_count=9, OMIM=607846, no_hdr0=''),
  449015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449015, CurrentID=0, Status='live', Symbol='ASPG3', Aliases='', description='Asperger syndrome, susceptibility to, 3', other_designations='', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608781, no_hdr0=''),
  449016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449016, CurrentID=0, Status='live', Symbol='BMIQ5', Aliases='', description='body mass index QTL 5', other_designations='Body mass index quantitative trait locus on chromosome 16, in children', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608558, no_hdr0=''),
  55801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55801, CurrentID=0, Status='live', Symbol='IL26', Aliases='AK155, IL-26', description='interleukin 26', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68163987, end_position_on_the_genomic_accession=68225791, orientation='minus', exon_count=6, OMIM=605679, no_hdr0=''),
  449018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449018, CurrentID=0, Status='live', Symbol='EJM3', Aliases='', description='Epilepsy, juvenile myoclonic 3', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608816, no_hdr0=''),
  55803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55803, CurrentID=0, Status='live', Symbol='ADAP2', Aliases='CENTA2, HSA272195, cent-b', description='ArfGAP with dual PH domains 2', other_designations='centaurin beta|centaurin, alpha 2|centaurin-alpha 2 protein|centaurin-alpha-2|cnt-a2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30921685, end_position_on_the_genomic_accession=30959398, orientation='plus', exon_count=11, OMIM=608635, no_hdr0=''),
  55805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55805, CurrentID=0, Status='live', Symbol='LRP2BP', Aliases='', description='LRP2 binding protein', other_designations='megBP|megalin-binding protein', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185363878, end_position_on_the_genomic_accession=185396935, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55806, CurrentID=0, Status='live', Symbol='HR', Aliases='ALUNC, AU, HSA277165, HYPT4, MUHH, MUHH1', description='hair growth associated', other_designations='hairless homolog|protein hairless', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22114419, end_position_on_the_genomic_accession=22131053, orientation='minus', exon_count=19, OMIM=602302, no_hdr0=''),
  55808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55808, CurrentID=0, Status='live', Symbol='ST6GALNAC1', Aliases='HSY11339, SIAT7A, ST6GalNAcI, STYI', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 1', other_designations='GalNAc alpha-2, 6-sialyltransferase I, long form|SIAT7-A|ST6GalNAc I|galNAc alpha-2,6-sialyltransferase I|sialyltransferase 7 ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) A|sialyltransferase 7A', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76617774, end_position_on_the_genomic_accession=76643838, orientation='minus', exon_count=12, OMIM=610138, no_hdr0=''),
  55809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55809, CurrentID=0, Status='live', Symbol='TRERF1', Aliases='BCAR2, HSA277276, RAPA, TREP132, TReP-132, dJ139D8.5', description='transcriptional regulating factor 1', other_designations='breast cancer anti-estrogen resistance 2|rapa-1|rapa-2|transcriptional regulating protein 132|transcriptional-regulating protein 132|zinc finger protein rapa|zinc finger transcription factor TReP-132', map_location='6p21.1-p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42224933, end_position_on_the_genomic_accession=42452205, orientation='minus', exon_count=23, OMIM=610322, no_hdr0=''),
  55810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55810, CurrentID=0, Status='live', Symbol='FOXJ2', Aliases='FHX', description='forkhead box J2', other_designations='FOXJ2 forkhead factor|fork head homologous X', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8032708, end_position_on_the_genomic_accession=8055522, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  55811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55811, CurrentID=0, Status='live', Symbol='ADCY10', Aliases='HCA2, HEL-S-7a, SAC, SACI, Sacy, hsAC', description='adenylate cyclase 10 (soluble)', other_designations="3',5'-cyclic AMP synthetase|AH-related protein|ATP pyrophosphate-lyase|epididymis secretory sperm binding protein Li 7a|germ cell soluble adenylyl cyclase|testicular soluble adenylyl cyclase (SAC)", map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167809080, end_position_on_the_genomic_accession=167915526, orientation='minus', exon_count=36, OMIM=605205, no_hdr0=''),
  55812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55812, CurrentID=0, Status='live', Symbol='SPATA7', Aliases='HEL-S-296, HSD-3.1, HSD3, LCA3', description='spermatogenesis associated 7', other_designations='epididymis secretory protein Li 296|spermatogenesis-associated protein HSD3', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88385638, end_position_on_the_genomic_accession=88470350, orientation='plus', exon_count=15, OMIM=609868, no_hdr0=''),
  55813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55813, CurrentID=0, Status='live', Symbol='UTP6', Aliases='C17orf40, HCA66', description='UTP6 small subunit processome component', other_designations='UTP6, small subunit (SSU) processome component, homolog|hepatocellular carcinoma associated antigen 66|hepatocellular carcinoma-associated antigen 66|multiple hat domains protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31863171, end_position_on_the_genomic_accession=31901710, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  55814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55814, CurrentID=0, Status='live', Symbol='BDP1', Aliases="HSA238520, TAF3B1, TFC5, TFIIIB'', TFIIIB150, TFIIIB90, TFNR", description='B double prime 1, subunit of RNA polymerase III transcription initiation factor IIIB', other_designations="RNA polymerase III transcription initiation factor B''|TATA box binding protein (TBP)-associated factor, RNA polymerase III, GTF3B subunit 1|transcription factor IIIB 150|transcription factor-like nuclear regulator", map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71455383, end_position_on_the_genomic_accession=71581062, orientation='plus', exon_count=43, OMIM=607012, no_hdr0=''),
  55815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55815, CurrentID=0, Status='live', Symbol='TSNAXIP1', Aliases='TXI1', description='translin-associated factor X interacting protein 1', other_designations='trax-interacting protein 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67806656, end_position_on_the_genomic_accession=67828068, orientation='plus', exon_count=15, OMIM=607720, no_hdr0=''),
  55816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55816, CurrentID=0, Status='live', Symbol='DOK5', Aliases='C20orf180, IRS-6, IRS6', description='docking protein 5', other_designations='downstream of tyrosine kinase 5|insulin receptor substrate 6', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=54475472, end_position_on_the_genomic_accession=54651171, orientation='plus', exon_count=11, OMIM=608334, no_hdr0=''),
  55818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55818, CurrentID=0, Status='live', Symbol='KDM3A', Aliases='JHDM2A, JHMD2A, JMJD1, JMJD1A, TSGA', description='lysine (K)-specific demethylase 3A', other_designations='jmjC domain-containing histone demethylation protein 2A|jumonji C domain-containing histone demethylase 2A|jumonji domain containing 1|jumonji domain containing 1A|jumonji domain-containing protein 1A|testis-specific protein A', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86440647, end_position_on_the_genomic_accession=86492716, orientation='plus', exon_count=29, OMIM=611512, no_hdr0=''),
  55819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55819, CurrentID=0, Status='live', Symbol='RNF130', Aliases='G1RZFP, GOLIATH, GP', description='ring finger protein 130', other_designations='g1-related zinc finger protein|goliath homolog', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179920270, end_position_on_the_genomic_accession=180072130, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55821, CurrentID=0, Status='live', Symbol='ALLC', Aliases='ALC', description='allantoicase', other_designations='allantoate amidinohydrolase', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3658184, end_position_on_the_genomic_accession=3702671, orientation='plus', exon_count=15, OMIM=612396, no_hdr0=''),
  55823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55823, CurrentID=0, Status='live', Symbol='VPS11', Aliases='END1, PEP5, RNF108, hVPS11', description='VPS11, CORVET/HOPS core subunit', other_designations='RING finger protein 108|VPS11 CORVET/HOPS core subunit|vacuolar protein sorting 11 homolog', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119067752, end_position_on_the_genomic_accession=119081978, orientation='plus', exon_count=17, OMIM=608549, no_hdr0=''),
  55824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55824, CurrentID=0, Status='live', Symbol='PAG1', Aliases='CBP, PAG', description='phosphoprotein membrane anchor with glycosphingolipid microdomains 1', other_designations='Csk-binding protein|phosphoprotein associated with glycosphingolipid microdomains 1|transmembrane adapter protein PAG|transmembrane adaptor protein PAG|transmembrane phosphoprotein Cbp', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80967810, end_position_on_the_genomic_accession=81112068, orientation='minus', exon_count=14, OMIM=605767, no_hdr0=''),
  55825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55825, CurrentID=0, Status='live', Symbol='PECR', Aliases='DCRRP, HPDHASE, HSA250303, PVIARL, SDR29C1, TERP', description='peroxisomal trans-2-enoyl-CoA reductase', other_designations='2,4-dienoyl-CoA reductase-related protein|DCR-RP|pVI-ARL|putative short chain alcohol dehydrogenase|short chain dehydrogenase/reductase family 29C member 1|short chain dehydrogenase/reductase family 29C, member 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216038388, end_position_on_the_genomic_accession=216081816, orientation='minus', exon_count=9, OMIM=605843, no_hdr0=''),
  55827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55827, CurrentID=0, Status='live', Symbol='DCAF6', Aliases='1200006M05Rik, ARCAP, IQWD1, MSTP055, NRIP, PC326', description='DDB1 and CUL4 associated factor 6', other_designations='IQ motif and WD repeat-containing protein 1|IQ motif and WD repeats 1|androgen receptor complex-associated protein|nuclear receptor interaction protein', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167936559, end_position_on_the_genomic_accession=168075843, orientation='plus', exon_count=23, OMIM=610494, no_hdr0=''),
  55829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55829, CurrentID=0, Status='live', Symbol='VIMP', Aliases='AD-015, ADO15, SBBI8, SELS, SEPS1', description='VCP-interacting membrane selenoprotein', other_designations='VCP-interacting membrane protein|valosin-containing protein-interacting membrane protein', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101270909, end_position_on_the_genomic_accession=101277520, orientation='minus', exon_count=7, OMIM=607918, no_hdr0=''),
  55830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55830, CurrentID=0, Status='live', Symbol='GLT8D1', Aliases='AD-017, MSTP139', description='glycosyltransferase 8 domain containing 1', other_designations='glycosyltransferase AD-017', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52694484, end_position_on_the_genomic_accession=52706083, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  55831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55831, CurrentID=0, Status='live', Symbol='EMC3', Aliases='POB, TMEM111', description='ER membrane protein complex subunit 3', other_designations='30 kDa protein|partial optokinetic response b|transmembrane protein 111', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9962682, end_position_on_the_genomic_accession=10011116, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  55832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55832, CurrentID=0, Status='live', Symbol='CAND1', Aliases='TIP120, TIP120A', description='cullin-associated and neddylation-dissociated 1', other_designations='TBP interacting protein|TBP-interacting protein 120A|TBP-interacting protein of 120 kDa A|cullin-associated and neddylation-dissociated protein 1|p120 CAND1', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67269281, end_position_on_the_genomic_accession=67314692, orientation='plus', exon_count=15, OMIM=607727, no_hdr0=''),
  55833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55833, CurrentID=0, Status='live', Symbol='UBAP2', Aliases='UBAP-2', description='ubiquitin associated protein 2', other_designations='AD-012 protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33921693, end_position_on_the_genomic_accession=34048949, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  55835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55835, CurrentID=0, Status='live', Symbol='CENPJ', Aliases='BM032, CENP-J, CPAP, LAP, LIP1, MCPH6, SASS4, SCKL4, Sas-4', description='centromere protein J', other_designations='LAG-3-associated protein|LYST-interacting protein 1|LYST-interacting protein LIP1|LYST-interacting protein LIP7|centrosomal P4.1-associated protein', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24882274, end_position_on_the_genomic_accession=24922889, orientation='minus', exon_count=21, OMIM=609279, no_hdr0=''),
  55837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55837, CurrentID=0, Status='live', Symbol='EAPP', Aliases='BM036, C14orf11', description='E2F-associated phosphoprotein', other_designations='', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34515929, end_position_on_the_genomic_accession=34539737, orientation='minus', exon_count=6, OMIM=609486, no_hdr0=''),
  55839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55839, CurrentID=0, Status='live', Symbol='CENPN', Aliases='BM039, C16orf60, CENP-N, ICEN32', description='centromere protein N', other_designations='interphase centromere complex protein 32', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81006498, end_position_on_the_genomic_accession=81033107, orientation='plus', exon_count=14, OMIM=611509, no_hdr0=''),
  55840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55840, CurrentID=0, Status='live', Symbol='EAF2', Aliases='BM040, TRAITS, U19', description='ELL associated factor 2', other_designations='testosterone-regulated apoptosis inducer and tumor suppressor protein', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121835187, end_position_on_the_genomic_accession=121886526, orientation='plus', exon_count=8, OMIM=607659, no_hdr0=''),
  55841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55841, CurrentID=0, Status='live', Symbol='WWC3', Aliases='BM042', description='WWC family member 3', other_designations='', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=10015755, end_position_on_the_genomic_accession=10144478, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  55843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55843, CurrentID=0, Status='live', Symbol='ARHGAP15', Aliases='BM046', description='Rho GTPase activating protein 15', other_designations='rho-type GTPase-activating protein 15', map_location='2q22.2-q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143070968, end_position_on_the_genomic_accession=143768352, orientation='plus', exon_count=22, OMIM=610578, no_hdr0=''),
  55844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55844, CurrentID=0, Status='live', Symbol='PPP2R2D', Aliases='MDS026', description='protein phosphatase 2 regulatory subunit B, delta', other_designations='PP2A subunit B isoform B55-delta|PP2A subunit B isoform PR55-delta|PP2A subunit B isoform R2-delta|PP2A subunit B isoform delta|protein phosphatase 2, regulatory subunit B, delta isoform', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131900644, end_position_on_the_genomic_accession=131969980, orientation='plus', exon_count=15, OMIM=613992, no_hdr0=''),
  55845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55845, CurrentID=0, Status='live', Symbol='BRK1', Aliases='C3orf10, HSPC300, MDS027, hHBrk1', description='BRICK1, SCAR/WAVE actin-nucleating complex subunit', other_designations='BRICK1, SCAR/WAVE actin-nucleating complex subunit, homolog|haematopoietic stem/progenitor cell protein 300|probable protein BRICK1', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10115649, end_position_on_the_genomic_accession=10127190, orientation='plus', exon_count=3, OMIM=611183, no_hdr0=''),
  55846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55846, CurrentID=0, Status='live', Symbol='ITFG2', Aliases='MDS028', description='integrin alpha FG-GAP repeat containing 2', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2812621, end_position_on_the_genomic_accession=2836956, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  55847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55847, CurrentID=0, Status='live', Symbol='CISD1', Aliases='C10orf70, MDS029, ZCD1, mitoNEET', description='CDGSH iron sulfur domain 1', other_designations='CDGSH iron sulfur domain-containing protein 1|zinc finger CDGSH-type domain 1|zinc finger, CDGSH-type domain 1', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58269102, end_position_on_the_genomic_accession=58289259, orientation='plus', exon_count=3, OMIM=611932, no_hdr0=''),
  55848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55848, CurrentID=0, Status='live', Symbol='PLGRKT', Aliases='AD025, C9orf46, MDS030, PLG-RKT, Plg-R(KT)', description='plasminogen receptor, C-terminal lysine transmembrane protein', other_designations='5033414D02Rik|plasminogen receptor with a C-terminal lysine|transmembrane protein C9orf46', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5357966, end_position_on_the_genomic_accession=5438456, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55850, CurrentID=0, Status='live', Symbol='USE1', Aliases='D12, MDS032, P31, SLT1', description='unconventional SNARE in the ER 1 homolog (S. cerevisiae)', other_designations='Q-SNARE|SNARE-like tail-anchored protein 1 homolog|USE1-like protein|protein p31|putative MAPK activating protein PM26|putative MAPK-activating protein PM26', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17215346, end_position_on_the_genomic_accession=17219829, orientation='plus', exon_count=8, OMIM=610675, no_hdr0=''),
  55851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55851, CurrentID=0, Status='live', Symbol='PSENEN', Aliases='MDS033, MSTP064, PEN-2, PEN2', description='presenilin enhancer gamma secretase subunit', other_designations='hematopoietic stem/progenitor cells protein MDS033|presenilin enhancer 2 homolog', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35745577, end_position_on_the_genomic_accession=35747155, orientation='plus', exon_count=4, OMIM=607632, no_hdr0=''),
  55852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55852, CurrentID=0, Status='live', Symbol='TEX2', Aliases='HT008, TMEM96', description='testis expressed 2', other_designations='testis expressed sequence 2|transmembrane protein 96', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64147433, end_position_on_the_genomic_accession=64263323, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  55853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55853, CurrentID=0, Status='live', Symbol='IDI2-AS1', Aliases='C10orf110, HT009, IDI2-AS', description='IDI2 antisense RNA 1', other_designations='IDI2 antisense RNA (non-protein coding)|IDI2 antisense RNA 1 (non-protein coding)', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1022637, end_position_on_the_genomic_accession=1044201, orientation='plus', exon_count=5, OMIM=615391, no_hdr0=''),
  55854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55854, CurrentID=0, Status='live', Symbol='ZC3H15', Aliases='HT010, LEREPO4, MSTP012', description='zinc finger CCCH-type containing 15', other_designations='DRG family-regulatory protein 1|likely ortholog of mouse immediate early response erythropoietin 4', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186486158, end_position_on_the_genomic_accession=186509361, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  55855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55855, CurrentID=0, Status='live', Symbol='FAM45BP', Aliases='FAM45B, HT011', description='family with sequence similarity 45 member B, pseudogene', other_designations='family with sequence similarity 45, member B (pseudogene)', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130494941, end_position_on_the_genomic_accession=130497448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  55856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55856, CurrentID=0, Status='live', Symbol='ACOT13', Aliases='HT012, PNAS-27, THEM2', description='acyl-CoA thioesterase 13', other_designations='hypothalamus protein HT012|thioesterase superfamily member 2', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24667035, end_position_on_the_genomic_accession=24705069, orientation='plus', exon_count=4, OMIM=615652, no_hdr0=''),
  55857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55857, CurrentID=0, Status='live', Symbol='KIZ', Aliases='C20orf19, HT013, Kizuna, NCRNA00153, PLK1S1, RP69', description='kizuna centrosomal protein', other_designations='polo-like kinase 1 substrate 1', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21125983, end_position_on_the_genomic_accession=21246622, orientation='plus', exon_count=14, OMIM=615757, no_hdr0=''),
  55858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55858, CurrentID=0, Status='live', Symbol='TMEM165', Aliases='CDG2K, FT27, GDT1, TMPT27, TPARL', description='transmembrane protein 165', other_designations='TPA regulated locus|transmembrane protein PT27|transmembrane protein TPARL', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55395913, end_position_on_the_genomic_accession=55452974, orientation='plus', exon_count=8, OMIM=614726, no_hdr0=''),
  55859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55859, CurrentID=0, Status='live', Symbol='BEX1', Aliases='BEX2, HBEX2, HGR74-h', description='brain expressed X-linked 1', other_designations='brain expressed, X-linked 1|brain-expressed X-linked protein 1|ovarian granulosa cell 13.0 kDa protein hGR74', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103062653, end_position_on_the_genomic_accession=103064240, orientation='minus', exon_count=3, OMIM=300690, no_hdr0=''),
  55860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55860, CurrentID=0, Status='live', Symbol='ACTR10', Aliases='ACTR11, Arp10, Arp11, HARP11', description='actin-related protein 10 homolog (S. cerevisiae)', other_designations='actin-related protein 11', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58200115, end_position_on_the_genomic_accession=58235636, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  55861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55861, CurrentID=0, Status='live', Symbol='DBNDD2', Aliases='C20orf35, CK1BP, HSMNP1', description='dysbindin (dystrobrevin binding protein 1) domain containing 2', other_designations='SCF apoptosis response protein 1|casein kinase-1 binding protein', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45405993, end_position_on_the_genomic_accession=45410610, orientation='plus', exon_count=5, OMIM=611453, no_hdr0=''),
  55862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55862, CurrentID=0, Status='live', Symbol='ECHDC1', Aliases='HEL-S-76, MMCD, dJ351K20.2', description='ethylmalonyl-CoA decarboxylase 1', other_designations='enoyl CoA hydratase domain containing 1|enoyl Coenzyme A hydratase domain containing 1|enoyl-CoA hydratase domain-containing protein 1|epididymis secretory protein Li 76|methylmalonyl-CoA decarboxylase', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127288712, end_position_on_the_genomic_accession=127344069, orientation='minus', exon_count=10, OMIM=612136, no_hdr0=''),
  55863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55863, CurrentID=0, Status='live', Symbol='TMEM126B', Aliases='HT007', description='transmembrane protein 126B', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85628573, end_position_on_the_genomic_accession=85636539, orientation='plus', exon_count=7, OMIM=615533, no_hdr0=''),
  100128320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128320, CurrentID=0, Status='live', Symbol='LOC100128320', Aliases='', description='uncharacterized LOC100128320', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=98817858, end_position_on_the_genomic_accession=98818253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55867, CurrentID=0, Status='live', Symbol='SLC22A11', Aliases='OAT4, hOAT4', description='solute carrier family 22 (organic anion/urate transporter), member 11', other_designations='organic anion transporter 4|solute carrier family 22 (organic anion/cation transporter), member 11', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64555626, end_position_on_the_genomic_accession=64572875, orientation='plus', exon_count=10, OMIM=607097, no_hdr0=''),
  55869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55869, CurrentID=0, Status='live', Symbol='HDAC8', Aliases='CDA07, CDLS5, HD8, HDACL1, MRXS6, RPD3, WTS', description='histone deacetylase 8', other_designations='histone deacetylase-like 1', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72329516, end_position_on_the_genomic_accession=72573103, orientation='minus', exon_count=17, OMIM=300269, no_hdr0=''),
  55870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55870, CurrentID=0, Status='live', Symbol='ASH1L', Aliases='ASH11, KMT2H, ASH1L', description='ash1 (absent, small, or homeotic)-like (Drosophila)', other_designations='ASH1-like protein|absent small and homeotic disks protein 1 homolog|lysine N-methyltransferase 2H|probable histone-lysine N-methyltransferase ASH1L', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155335261, end_position_on_the_genomic_accession=155562850, orientation='minus', exon_count=30, OMIM=607999, no_hdr0=''),
  55871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55871, CurrentID=0, Status='live', Symbol='CBWD1', Aliases='COBP', description='COBW domain containing 1', other_designations='COBW-like protein|NPC-A-6 COBW domain-containing protein 1|cobalamin synthase W domain-containing protein 1|cobalamin synthetase W domain-containing protein 1', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121038, end_position_on_the_genomic_accession=179075, orientation='minus', exon_count=19, OMIM=611078, no_hdr0=''),
  55872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55872, CurrentID=0, Status='live', Symbol='PBK', Aliases='CT84, HEL164, Nori-3, SPK, TOPK', description='PDZ binding kinase', other_designations='MAPKK-like protein kinase|T-LAK cell-originated protein kinase|cancer/testis antigen 84|epididymis luminal protein 164|serine/threonine protein kinase|spermatogenesis-related protein kinase', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27809621, end_position_on_the_genomic_accession=27838095, orientation='minus', exon_count=8, OMIM=611210, no_hdr0=''),
  105369355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369355, CurrentID=0, Status='live', Symbol='LOC105369355', Aliases='', description='uncharacterized LOC105369355', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66954249, end_position_on_the_genomic_accession=66970128, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55876, CurrentID=0, Status='live', Symbol='GSDMB', Aliases='GSDML, PP4052, PRO2521', description='gasdermin B', other_designations='gasdermin-like protein', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39904595, end_position_on_the_genomic_accession=39919377, orientation='minus', exon_count=11, OMIM=611221, no_hdr0=''),
  55879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55879, CurrentID=0, Status='live', Symbol='GABRQ', Aliases='THETA', description='gamma-aminobutyric acid (GABA) A receptor, theta', other_designations='GABA(A) receptor subunit theta', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152637895, end_position_on_the_genomic_accession=152659510, orientation='plus', exon_count=10, OMIM=300349, no_hdr0=''),
  100126476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126476, CurrentID=0, Status='live', Symbol='TRNAL-CAA', Aliases='', description='transfer RNA leucine (anticodon CAA)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28896223, end_position_on_the_genomic_accession=28896328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55884, CurrentID=0, Status='live', Symbol='WSB2', Aliases='SBA2', description='WD repeat and SOCS box containing 2', other_designations='CS box-containing WD protein|WD repeat and SOCS box containing protein 2|WD repeat and SOCS box-containing 2|WSB-2', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118032687, end_position_on_the_genomic_accession=118062174, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55885, CurrentID=0, Status='live', Symbol='LMO3', Aliases='RBTN3, RBTNL2, RHOM3, Rhom-3', description='LIM domain only 3', other_designations='LIM domain only 3 (rhombotin-like 2)|neuronal-specific transcription factor DAT1|rhombotin-3', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16548372, end_position_on_the_genomic_accession=16610130, orientation='minus', exon_count=13, OMIM=180386, no_hdr0=''),
  105369357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369357, CurrentID=0, Status='live', Symbol='LOC105369357', Aliases='', description='uncharacterized LOC105369357', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67436357, end_position_on_the_genomic_accession=67438141, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  55888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55888, CurrentID=0, Status='live', Symbol='ZKSCAN7', Aliases='ZFP, ZNF167, ZNF448, ZNF64, ZSCAN39', description='zinc finger with KRAB and SCAN domains 7', other_designations='zinc finger protein 167|zinc finger protein 448|zinc finger protein 64', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44555175, end_position_on_the_genomic_accession=44583464, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  55889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55889, CurrentID=0, Status='live', Symbol='GOLGA6B', Aliases='GOLGA, GOLGA6D', description='golgin A6 family member B', other_designations='Golgin subfamily A member 6B|Putative golgin subfamily A member 6D|golgi autoantigen, golgin subfamily a, 6B|golgin A6 family, member B|golgin-like protein', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72608556, end_position_on_the_genomic_accession=72667397, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  55890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55890, CurrentID=0, Status='live', Symbol='GPRC5C', Aliases='RAIG-3, RAIG3', description='G protein-coupled receptor, class C, group 5, member C', other_designations='G protein-coupled receptor, family C, group 5, member C|orphan G-protein coupled receptor|retinoic acid responsive gene protein|retinoic acid-induced gene 3 protein', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74424503, end_position_on_the_genomic_accession=74451658, orientation='plus', exon_count=8, OMIM=605949, no_hdr0=''),
  55891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55891, CurrentID=0, Status='live', Symbol='LENEP', Aliases='LEP503', description='lens epithelial protein', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154993586, end_position_on_the_genomic_accession=154994315, orientation='plus', exon_count=2, OMIM=607377, no_hdr0=''),
  55892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55892, CurrentID=0, Status='live', Symbol='MYNN', Aliases='OSZF, SBBIZ1, ZBTB31, ZNF902', description='myoneurin', other_designations='zinc finger and BTB domain-containing protein 31|zinc finger protein with BTB/POZ domain', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169769649, end_position_on_the_genomic_accession=169789716, orientation='plus', exon_count=10, OMIM=606042, no_hdr0=''),
  55893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55893, CurrentID=0, Status='live', Symbol='ZNF395', Aliases='HDBP-2, HDBP2, HDRF-2, PBF, PRF-1, PRF1, Si-1-8-14', description='zinc finger protein 395', other_designations="HD gene regulatory region-binding protein 2|HD-regulating factor 2|Huntington's disease gene regulatory region-binding protein 2|huntington disease gene regulatory region-binding protein 2|papillomavirus regulatory factor 1|papillomavirus regulatory factor PRF-1|papillomavirus-binding factor", map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28345585, end_position_on_the_genomic_accession=28386460, orientation='minus', exon_count=10, OMIM=609494, no_hdr0=''),
  55894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55894, CurrentID=0, Status='live', Symbol='DEFB103B', Aliases='BD-3, DEFB-3, DEFB103, DEFB3, HBD-3, HBD3, HBP-3, HBP3', description='defensin beta 103B', other_designations='beta-defensin 3|defensin, beta 103|defensin, beta 103B|defensin-like protein', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7428888, end_position_on_the_genomic_accession=7430348, orientation='minus', exon_count=2, OMIM=606611, no_hdr0=''),
  55897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55897, CurrentID=0, Status='live', Symbol='MESP1', Aliases='bHLHc5', description='mesoderm posterior bHLH transcription factor 1', other_designations='class C basic helix-loop-helix protein 5|mesoderm posterior 1 homolog|mesoderm posterior basic helix-loop-helix transcription factor 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89749867, end_position_on_the_genomic_accession=89751309, orientation='minus', exon_count=2, OMIM=608689, no_hdr0=''),
  55898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55898, CurrentID=0, Status='live', Symbol='UNC45A', Aliases='GC-UNC45, GCUNC-45, GCUNC45, IRO039700, SMAP-1, SMAP1, UNC-45A', description='unc-45 myosin chaperone A', other_designations='general cell UNC45|smooth muscle cell associated protein-1|unc-45 homolog A', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90930008, end_position_on_the_genomic_accession=90954093, orientation='plus', exon_count=23, OMIM=611219, no_hdr0=''),
  105369359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369359, CurrentID=0, Status='live', Symbol='LOC105369359', Aliases='', description='40S ribosomal protein S24 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67935216, end_position_on_the_genomic_accession=67936204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55900, CurrentID=0, Status='live', Symbol='ZNF302', Aliases='HSD16, MST154, MSTP154, ZNF135L, ZNF140L, ZNF327', description='zinc finger protein 302', other_designations='zinc finger protein 327', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34677639, end_position_on_the_genomic_accession=34686397, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  55901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55901, CurrentID=0, Status='live', Symbol='THSD1', Aliases='TMTSP, UNQ3010', description='thrombospondin type 1 domain containing 1', other_designations='4833423O18Rik|thrombospondin, type I, domain 1|thrombospondin, type I, domain containing 1|transmembrane molecule with thrombospondin module', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52377167, end_position_on_the_genomic_accession=52406494, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  55902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55902, CurrentID=0, Status='live', Symbol='ACSS2', Aliases='ACAS2, ACECS, ACS, ACSA, dJ1161H23.1', description='acyl-CoA synthetase short-chain family member 2', other_designations='acetate thiokinase|acetate-CoA ligase|acetyl-Coenzyme A synthetase 2 (ADP forming)|acyl-activating enzyme|cytoplasmic acetyl-coenzyme A synthetase', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34874942, end_position_on_the_genomic_accession=34927966, orientation='plus', exon_count=22, OMIM=605832, no_hdr0=''),
  55904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55904, CurrentID=0, Status='live', Symbol='KMT2E', Aliases='HDCMC04P, MLL5, NKp44L', description='lysine (K)-specific methyltransferase 2E', other_designations='histone-lysine N-methyltransferase MLL5|myeloid/lymphoid or mixed-lineage leukemia 5 (trithorax homolog, Drosophila)|myeloid/lymphoid or mixed-lineage leukemia protein 5', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105014190, end_position_on_the_genomic_accession=105114085, orientation='plus', exon_count=29, OMIM=608444, no_hdr0=''),
  55905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55905, CurrentID=0, Status='live', Symbol='RNF114', Aliases='PSORS12, ZNF313', description='ring finger protein 114', other_designations='zinc finger protein 228|zinc finger protein 313', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49936377, end_position_on_the_genomic_accession=49953885, orientation='plus', exon_count=6, OMIM=612451, no_hdr0=''),
  55906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55906, CurrentID=0, Status='live', Symbol='ZC4H2', Aliases='HCA127, KIAA1166, WRWF, WWS', description='zinc finger, C4H2 domain containing', other_designations='hepatocellular carcinoma-associated antigen 127', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64915802, end_position_on_the_genomic_accession=65034744, orientation='minus', exon_count=6, OMIM=300897, no_hdr0=''),
  55907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55907, CurrentID=0, Status='live', Symbol='CMAS', Aliases='CSS', description='cytidine monophosphate N-acetylneuraminic acid synthetase', other_designations="CMP-N-acetylneuraminic acid synthase|CMP-N-acetylneuraminic acid synthetase|CMP-Neu5Ac synthetase|CMP-NeuNAc synthase|CMP-NeuNAc synthetase|CMP-sialic acid synthetase|cytidine 5'-monophosphate N-acetylneuraminic acid synthetase", map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22046174, end_position_on_the_genomic_accession=22065674, orientation='plus', exon_count=8, OMIM=603316, no_hdr0=''),
  55908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55908, CurrentID=0, Status='live', Symbol='ANGPTL8', Aliases='C19orf80, PRO1185, PVPA599, RIFL, TD26', description='angiopoietin like 8', other_designations='angiopoietin-like 8|betatrophin variant 1|betatrophin variant 2|hepatocellular carcinoma-associated gene TD26|hepatocellular carcinoma-associated protein TD26|lipasin|refeeding-induced fat and liver protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11239619, end_position_on_the_genomic_accession=11241943, orientation='plus', exon_count=4, OMIM=616223, no_hdr0=''),
  55909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55909, CurrentID=0, Status='live', Symbol='BIN3', Aliases='', description='bridging integrator 3', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22620418, end_position_on_the_genomic_accession=22669148, orientation='minus', exon_count=12, OMIM=606396, no_hdr0=''),
  55911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55911, CurrentID=0, Status='live', Symbol='APOBR', Aliases='APOB100R, APOB48R', description='apolipoprotein B receptor', other_designations='apoB-48R|apolipoprotein B-100 receptor|apolipoprotein B-48 receptor|apolipoprotein B100 receptor|apolipoprotein B48 receptor', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28494649, end_position_on_the_genomic_accession=28498970, orientation='plus', exon_count=4, OMIM=605220, no_hdr0=''),
  55914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55914, CurrentID=0, Status='live', Symbol='ERBB2IP', Aliases='ERBIN, HEL-S-78, LAP2', description='erbb2 interacting protein', other_designations='densin-180-like protein|epididymis secretory protein Li 78', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65926465, end_position_on_the_genomic_accession=66081023, orientation='plus', exon_count=29, OMIM=606944, no_hdr0=''),
  9319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9319, CurrentID=0, Status='live', Symbol='TRIP13', Aliases='16E1BP', description='thyroid hormone receptor interactor 13', other_designations='16E1-BP|HPV16 E1 protein binding protein|HPV16 E1 protein-binding protein|TR-interacting protein 13|TRIP-13|human papillomavirus type 16 E1 protein-binding protein|thyroid receptor interacting protein 13|thyroid receptor-interacting protein 13', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=892854, end_position_on_the_genomic_accession=919072, orientation='plus', exon_count=14, OMIM=604507, no_hdr0=''),
  55916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55916, CurrentID=0, Status='live', Symbol='NXT2', Aliases='P15-2', description='nuclear transport factor 2-like export factor 2', other_designations='protein p15-2', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109535781, end_position_on_the_genomic_accession=109544698, orientation='plus', exon_count=7, OMIM=300320, no_hdr0=''),
  55917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55917, CurrentID=0, Status='live', Symbol='CTTNBP2NL', Aliases='', description='CTTNBP2 N-terminal like', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112396178, end_position_on_the_genomic_accession=112461164, orientation='plus', exon_count=7, OMIM=615100, no_hdr0=''),
  55920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55920, CurrentID=0, Status='live', Symbol='RCC2', Aliases='TD-60', description='regulator of chromosome condensation 2', other_designations='RCC1-like protein TD-60|telophase disk protein of 60 kDa', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17406755, end_position_on_the_genomic_accession=17439754, orientation='minus', exon_count=13, OMIM=609587, no_hdr0=''),
  55922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55922, CurrentID=0, Status='live', Symbol='NKRF', Aliases='ITBA4, NRF', description='NFKB repressing factor', other_designations='transcription factor NRF', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119588337, end_position_on_the_genomic_accession=119606435, orientation='minus', exon_count=5, OMIM=300440, no_hdr0=''),
  105369363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369363, CurrentID=0, Status='live', Symbol='LOC105369363', Aliases='', description='uncharacterized LOC105369363', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68272087, end_position_on_the_genomic_accession=68284643, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  55924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55924, CurrentID=0, Status='live', Symbol='FAM212B', Aliases='C1orf183', description='family with sequence similarity 212 member B', other_designations='family with sequence similarity 212, member B', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111722063, end_position_on_the_genomic_accession=111755797, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55929, CurrentID=0, Status='live', Symbol='DMAP1', Aliases='DNMAP1, DNMTAP1, EAF2, MEAF2, SWC4', description='DNA methyltransferase 1 associated protein 1', other_designations='DNMT1 associated protein 1|DNMT1-associated protein 1', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44213432, end_position_on_the_genomic_accession=44220679, orientation='plus', exon_count=11, OMIM=605077, no_hdr0=''),
  55930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55930, CurrentID=0, Status='live', Symbol='MYO5C', Aliases='', description='myosin VC', other_designations='myosin 5C|myosin-Vc', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52192318, end_position_on_the_genomic_accession=52295798, orientation='minus', exon_count=41, OMIM=610022, no_hdr0=''),
  105375792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375792, CurrentID=0, Status='live', Symbol='LOC105375792', Aliases='', description='uncharacterized LOC105375792', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142185145, end_position_on_the_genomic_accession=142189736, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369365, CurrentID=0, Status='live', Symbol='LOC105369365', Aliases='', description='uncharacterized LOC105369365', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69033408, end_position_on_the_genomic_accession=69041393, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23685, CurrentID=0, Status='live', Symbol='PQBP4', Aliases='', description='polyglutamine binding protein 4', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55937, CurrentID=0, Status='live', Symbol='APOM', Aliases='G3a, HSPC336, NG20, apo-M', description='apolipoprotein M', other_designations='NG20-like protein|alternative name: G3a, NG20|protein G3a', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31652410, end_position_on_the_genomic_accession=31658210, orientation='plus', exon_count=7, OMIM=606907, no_hdr0=''),
  105369367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369367, CurrentID=0, Status='live', Symbol='LOC105369367', Aliases='', description='uncharacterized LOC105369367', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69228749, end_position_on_the_genomic_accession=69234945, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100126488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126488, CurrentID=0, Status='live', Symbol='TRNAV-CAC', Aliases='', description='transfer RNA valine (anticodon CAC)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149708660, end_position_on_the_genomic_accession=149708730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55954, CurrentID=0, Status='live', Symbol='ZMAT5', Aliases='SNRNP20, U11/U12-20K, ZC3H19', description='zinc finger, matrin-type 5', other_designations='U11/U12 small nuclear ribonucleoprotein 20 kDa protein|U11/U12 snRNP 20 kDa protein|U11/U12 snRNP 20K|zinc finger CCCH-type containing 19', map_location='22cen-q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29730956, end_position_on_the_genomic_accession=29767068, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55957, CurrentID=0, Status='live', Symbol='LIN37', Aliases='F25965, ZK418.4, lin-37', description='lin-37 DREAM MuvB core complex component', other_designations='antolefinin|lin-37 homolog|protein F25965', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35748361, end_position_on_the_genomic_accession=35754519, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  55958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55958, CurrentID=0, Status='live', Symbol='KLHL9', Aliases='', description='kelch like family member 9', other_designations='kelch-like 9|kelch-like family member 9', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21331019, end_position_on_the_genomic_accession=21335430, orientation='minus', exon_count=1, OMIM=611201, no_hdr0=''),
  55959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55959, CurrentID=0, Status='live', Symbol='SULF2', Aliases='HSULF-2', description='sulfatase 2', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47657406, end_position_on_the_genomic_accession=47786629, orientation='minus', exon_count=24, OMIM=610013, no_hdr0=''),
  101243544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101243544, CurrentID=0, Status='live', Symbol='ERCC6-PGBD3', Aliases='CPFP, CSB-PGBD3', description='ERCC6-PGBD3 readthrough', other_designations='CSB-PGBD3 fusion protein|CSB-piggyBac transposable element derived 3 fusion', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49515105, end_position_on_the_genomic_accession=49539538, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101243545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101243545, CurrentID=0, Status='live', Symbol='LOC101243545', Aliases='', description='uncharacterized LOC101243545', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161426427, end_position_on_the_genomic_accession=161448242, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  55964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55964, CurrentID=0, Status='live', Symbol='SEPT3', Aliases='SEP3, bK250D10.3', description='septin 3', other_designations='neuronal-specific septin 3', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41976927, end_position_on_the_genomic_accession=41998221, orientation='plus', exon_count=10, OMIM=608314, no_hdr0=''),
  55966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55966, CurrentID=0, Status='live', Symbol='AJAP1', Aliases='MOT8, SHREW-1, SHREW1', description='adherens junctions associated protein 1', other_designations='adherens junction associated protein 1|membrane protein shrew-1|transmembrane protein SHREW1', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4655045, end_position_on_the_genomic_accession=4792534, orientation='plus', exon_count=9, OMIM=610972, no_hdr0=''),
  55967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55967, CurrentID=0, Status='live', Symbol='NDUFA12', Aliases='B17.2, DAP13', description='NADH:ubiquinone oxidoreductase subunit A12', other_designations='13 kDa differentiation-associated protein|CI-B17.2|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 12|NADH-ubiquinone oxidoreductase subunit B17.2|complex I B17.2 subunit', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94971328, end_position_on_the_genomic_accession=95003713, orientation='minus', exon_count=4, OMIM=614530, no_hdr0=''),
  55968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55968, CurrentID=0, Status='live', Symbol='NSFL1C', Aliases='P47, UBX1, UBXD10, UBXN2C, dJ776F14.1', description='NSFL1 (p97) cofactor (p47)', other_designations='SHP1 homolog|UBX domain-containing protein 2C|p97 cofactor p47', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1442162, end_position_on_the_genomic_accession=1467720, orientation='minus', exon_count=14, OMIM=606610, no_hdr0=''),
  55969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55969, CurrentID=0, Status='live', Symbol='C20orf24', Aliases='PNAS-11, RIP5', description='chromosome 20 open reading frame 24', other_designations='rab5-interacting protein', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36605734, end_position_on_the_genomic_accession=36612557, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  55970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55970, CurrentID=0, Status='live', Symbol='GNG12', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 12', other_designations='G-protein gamma-12 subunit', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67701466, end_position_on_the_genomic_accession=67833779, orientation='minus', exon_count=6, OMIM=615405, no_hdr0=''),
  55971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55971, CurrentID=0, Status='live', Symbol='BAIAP2L1', Aliases='IRTKS', description='BAI1-associated protein 2-like 1', other_designations='BAI1-associated protein 2-like protein 1|insulin receptor tyrosine kinase substrate', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98291650, end_position_on_the_genomic_accession=98401115, orientation='minus', exon_count=14, OMIM=611877, no_hdr0=''),
  55972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55972, CurrentID=0, Status='live', Symbol='SLC25A40', Aliases='MCFP', description='solute carrier family 25 member 40', other_designations='mitochondrial carrier family protein|solute carrier family 25, member 40', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87833568, end_position_on_the_genomic_accession=87876377, orientation='minus', exon_count=16, OMIM=610821, no_hdr0=''),
  55973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55973, CurrentID=0, Status='live', Symbol='BCAP29', Aliases='BAP29', description='B-cell receptor-associated protein 29', other_designations='BCR-associated protein 29', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107579977, end_position_on_the_genomic_accession=107623317, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  55974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55974, CurrentID=0, Status='live', Symbol='SLC50A1', Aliases='HsSWEET1, RAG1AP1, SCP, SWEET1, slv', description='solute carrier family 50 (sugar efflux transporter), member 1', other_designations='RAG1-activating protein 1|RP11-540D14.5|RZPDo834D038D|probable sugar transporter RAG1AP1|recombination activating gene 1 activating protein 1|solute carrier family 50 (sugar transporter), member 1|solute carrier family 50 member 1|stromal cell protein', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155135344, end_position_on_the_genomic_accession=155138858, orientation='plus', exon_count=8, OMIM=613683, no_hdr0=''),
  55975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55975, CurrentID=0, Status='live', Symbol='KLHL7', Aliases='KLHL6, SBBI26', description='kelch like family member 7', other_designations='kelch-like 6|kelch-like 7|kelch-like family member 7|kelch/BTB', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23105734, end_position_on_the_genomic_accession=23175421, orientation='plus', exon_count=15, OMIM=611119, no_hdr0=''),
  105369372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369372, CurrentID=0, Status='live', Symbol='LOC105369372', Aliases='', description='uncharacterized LOC105369372', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70141569, end_position_on_the_genomic_accession=70142487, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422490, CurrentID=0, Status='live', Symbol='LOC100422490', Aliases='', description='UTP6, small subunit (SSU) processome component, homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51858994, end_position_on_the_genomic_accession=51861354, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369373, CurrentID=0, Status='live', Symbol='LOC105369373', Aliases='', description='uncharacterized LOC105369373', other_designations='CTA-797E19.2', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70324808, end_position_on_the_genomic_accession=70327459, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369375, CurrentID=0, Status='live', Symbol='LOC105369375', Aliases='', description='uncharacterized LOC105369375', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72298747, end_position_on_the_genomic_accession=72307190, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  55997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55997, CurrentID=0, Status='live', Symbol='CFC1', Aliases='CFC1B, CRYPTIC, DTGA2, HTX2', description='cripto, FRL-1, cryptic family 1', other_designations='cryptic family protein 1|heterotaxy 2 (autosomal dominant)', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130592165, end_position_on_the_genomic_accession=130599575, orientation='minus', exon_count=7, OMIM=605194, no_hdr0=''),
  55998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55998, CurrentID=0, Status='live', Symbol='NXF5', Aliases='', description='nuclear RNA export factor 5', other_designations='TAP-like protein 1|TAPL-1', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101832112, end_position_on_the_genomic_accession=101857577, orientation='minus', exon_count=21, OMIM=300319, no_hdr0=''),
  55999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55999, CurrentID=0, Status='live', Symbol='NXF4', Aliases='', description='nuclear RNA export factor 4 pseudogene', other_designations='', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102549965, end_position_on_the_genomic_accession=102571693, orientation='plus', exon_count=17, OMIM=300318, no_hdr0=''),
  56000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56000, CurrentID=0, Status='live', Symbol='NXF3', Aliases='', description='nuclear RNA export factor 3', other_designations='TAP-like protein 3|TAPL-3', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103075821, end_position_on_the_genomic_accession=103093094, orientation='minus', exon_count=20, OMIM=300316, no_hdr0=''),
  56001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56001, CurrentID=0, Status='live', Symbol='NXF2', Aliases='CT39, TAPL-2, TCP11X2', description='nuclear RNA export factor 2', other_designations='TAP-like protein 2|cancer/testis antigen 39|t-complex protein 11 homolog', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102247100, end_position_on_the_genomic_accession=102326720, orientation='plus', exon_count=23, OMIM=300315, no_hdr0=''),
  56005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56005, CurrentID=0, Status='live', Symbol='MYDGF', Aliases='C19orf10, EUROIMAGE1875335, IL25, IL27, IL27w, R33729_1, SF20', description='myeloid-derived growth factor', other_designations='UPF0556 protein C19orf10|interleukin 25|interleukin 27 working designation|interleukin-25|stromal cell-derived growth factor SF20', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4657545, end_position_on_the_genomic_accession=4670403, orientation='minus', exon_count=6, OMIM=606746, no_hdr0=''),
  56006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56006, CurrentID=0, Status='live', Symbol='SMG9', Aliases='C19orf61, F17127_1', description='SMG9 nonsense mediated mRNA decay factor', other_designations='protein smg-9 homolog|smg-9 homolog, nonsense mediated mRNA decay factor', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43727992, end_position_on_the_genomic_accession=43754990, orientation='minus', exon_count=15, OMIM=613176, no_hdr0=''),
  105369377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369377, CurrentID=0, Status='live', Symbol='LOC105369377', Aliases='', description='uncharacterized LOC105369377', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72606157, end_position_on_the_genomic_accession=72619225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  402779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402779, CurrentID=0, Status='live', Symbol='CSMD2-AS1', Aliases='', description='CSMD2 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33868956, end_position_on_the_genomic_accession=33885458, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369379, CurrentID=0, Status='live', Symbol='LOC105369379', Aliases='', description='uncharacterized LOC105369379', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72658076, end_position_on_the_genomic_accession=72670957, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372301, CurrentID=0, Status='live', Symbol='LOC105372301', Aliases='', description='uncharacterized LOC105372301', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19060831, end_position_on_the_genomic_accession=19063511, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642852, CurrentID=0, Status='live', Symbol='LOC642852', Aliases='', description='uncharacterized LOC642852', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45288052, end_position_on_the_genomic_accession=45297354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369381, CurrentID=0, Status='live', Symbol='LOC105369381', Aliases='', description='uncharacterized LOC105369381', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73156048, end_position_on_the_genomic_accession=73183578, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56033, CurrentID=0, Status='live', Symbol='BARX1', Aliases='', description='BARX homeobox 1', other_designations='BarH-like homeobox 1', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93951627, end_position_on_the_genomic_accession=93955326, orientation='minus', exon_count=4, OMIM=603260, no_hdr0=''),
  56034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56034, CurrentID=0, Status='live', Symbol='PDGFC', Aliases='FALLOTEIN, SCDGF', description='platelet derived growth factor C', other_designations='PDGF-C|VEGF-E|secretory growth factor-like protein|spinal cord-derived growth factor', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=156761611, end_position_on_the_genomic_accession=156971394, orientation='minus', exon_count=9, OMIM=608452, no_hdr0=''),
  390308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390308, CurrentID=0, Status='live', Symbol='LOC390308', Aliases='', description='CDGSH iron sulfur domain 1 pseudogene', other_designations='uncharacterized hematopoietic stem/progenitor cells protein MDS029 pseudogene 1', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42636632, end_position_on_the_genomic_accession=42637263, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369383, CurrentID=0, Status='live', Symbol='LOC105369383', Aliases='', description='uncharacterized LOC105369383', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74209208, end_position_on_the_genomic_accession=74210807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  664701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664701, CurrentID=0, Status='live', Symbol='ZNF826P', Aliases='ZNF826', description='zinc finger protein 826, pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20398658, end_position_on_the_genomic_accession=20424965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369384, CurrentID=0, Status='live', Symbol='LOC105369384', Aliases='', description='uncharacterized LOC105369384', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73994267, end_position_on_the_genomic_accession=74037418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56052, CurrentID=0, Status='live', Symbol='ALG1', Aliases='CDG1K, HMAT1, HMT-1, HMT1, MT-1, Mat-1, hMat-1', description='ALG1, chitobiosyldiphosphodolichol beta-mannosyltransferase', other_designations='GDP-Man:GlcNAc2-PP-dolichol mannosyltransferase|GDP-mannose-dolichol diphosphochitobiose mannosyltransferase|asparagine-linked glycosylation 1 homolog (yeast, beta-1,4-mannosyltransferase)|asparagine-linked glycosylation 1, beta-1,4-mannosyltransferase homolog|asparagine-linked glycosylation protein 1 homolog|beta-1,4 mannosyltransferase|beta-1,4-mannosyltransferase|mannosyltransferase-1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5071809, end_position_on_the_genomic_accession=5087379, orientation='plus', exon_count=14, OMIM=605907, no_hdr0=''),
  105369385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369385, CurrentID=0, Status='live', Symbol='LOC105369385', Aliases='', description='uncharacterized LOC105369385', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74328308, end_position_on_the_genomic_accession=74368479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56061, CurrentID=0, Status='live', Symbol='UBFD1', Aliases='UBPH', description='ubiquitin family domain containing 1', other_designations='ubiquitin-binding protein homolog', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23557541, end_position_on_the_genomic_accession=23574394, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56062, CurrentID=0, Status='live', Symbol='KLHL4', Aliases='DKELCHL, KHL4', description='kelch like family member 4', other_designations='kelch-like family member 4', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=87517541, end_position_on_the_genomic_accession=87670050, orientation='plus', exon_count=12, OMIM=300348, no_hdr0=''),
  56063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56063, CurrentID=0, Status='live', Symbol='TMEM234', Aliases='AASL548, C1orf91, PRO1105, RP4-622L5, dJ622L5.7', description='transmembrane protein 234', other_designations='', map_location='1p36.11-p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32204806, end_position_on_the_genomic_accession=32222353, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105369388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369388, CurrentID=0, Status='live', Symbol='LOC105369388', Aliases='', description='uncharacterized LOC105369388', other_designations='CTD-2562J17.4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75205717, end_position_on_the_genomic_accession=75225564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369389, CurrentID=0, Status='live', Symbol='LOC105369389', Aliases='', description='uncharacterized LOC105369389', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75509913, end_position_on_the_genomic_accession=75519957, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  642862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642862, CurrentID=0, Status='live', Symbol='LOC642862', Aliases='', description='Ras homolog enriched in brain pseudogene', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47705822, end_position_on_the_genomic_accession=47707168, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369391, CurrentID=0, Status='live', Symbol='LOC105369391', Aliases='', description='uncharacterized LOC105369391', other_designations='CTD-2530H12.4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75583196, end_position_on_the_genomic_accession=75605128, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  664709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664709, CurrentID=0, Status='live', Symbol='HNRNPA1P10', Aliases='HNRPA1L-2', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 10', other_designations='hnRNP A1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11665736, end_position_on_the_genomic_accession=11667089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56097, CurrentID=0, Status='live', Symbol='PCDHGC5', Aliases='PCDH-GAMMA-C5', description='protocadherin gamma subfamily C, 5', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141489241, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606306, no_hdr0=''),
  56098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56098, CurrentID=0, Status='live', Symbol='PCDHGC4', Aliases='PCDH-GAMMA-C4', description='protocadherin gamma subfamily C, 4', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141485174, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606305, no_hdr0=''),
  56099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56099, CurrentID=0, Status='live', Symbol='PCDHGB7', Aliases='ME6, PCDH-GAMMA-B7', description='protocadherin gamma subfamily B, 7', other_designations='cadherin ME6', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141417647, end_position_on_the_genomic_accession=141512981, orientation='plus', exon_count=4, OMIM=606304, no_hdr0=''),
  56100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56100, CurrentID=0, Status='live', Symbol='PCDHGB6', Aliases='PCDH-GAMMA-B6', description='protocadherin gamma subfamily B, 6', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141408203, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606303, no_hdr0=''),
  56101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56101, CurrentID=0, Status='live', Symbol='PCDHGB5', Aliases='PCDH-GAMMA-B5', description='protocadherin gamma subfamily B, 5', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141398128, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606302, no_hdr0=''),
  56102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56102, CurrentID=0, Status='live', Symbol='PCDHGB3', Aliases='PCDH-GAMMA-B3', description='protocadherin gamma subfamily B, 3', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141370264, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606301, no_hdr0=''),
  56103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56103, CurrentID=0, Status='live', Symbol='PCDHGB2', Aliases='PCDH-GAMMA-B2', description='protocadherin gamma subfamily B, 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141360136, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606300, no_hdr0=''),
  56104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56104, CurrentID=0, Status='live', Symbol='PCDHGB1', Aliases='PCDH-GAMMA-B1', description='protocadherin gamma subfamily B, 1', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141350261, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606299, no_hdr0=''),
  56105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56105, CurrentID=0, Status='live', Symbol='PCDHGA11', Aliases='PCDH-GAMMA-A11', description='protocadherin gamma subfamily A, 11', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141420970, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606298, no_hdr0=''),
  56106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56106, CurrentID=0, Status='live', Symbol='PCDHGA10', Aliases='PCDH-GAMMA-A10', description='protocadherin gamma subfamily A, 10', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141413176, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606297, no_hdr0=''),
  56107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56107, CurrentID=0, Status='live', Symbol='PCDHGA9', Aliases='PCDH-GAMMA-A9', description='protocadherin gamma subfamily A, 9', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141402953, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606296, no_hdr0=''),
  56108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56108, CurrentID=0, Status='live', Symbol='PCDHGA7', Aliases='PCDH-GAMMA-A7', description='protocadherin gamma subfamily A, 7', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141382739, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606294, no_hdr0=''),
  56109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56109, CurrentID=0, Status='live', Symbol='PCDHGA6', Aliases='PCDH-GAMMA-A6', description='protocadherin gamma subfamily A, 6', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141374084, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606293, no_hdr0=''),
  56110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56110, CurrentID=0, Status='live', Symbol='PCDHGA5', Aliases='CDH-GAMMA-A5, ME3, PCDH-GAMMA-A5', description='protocadherin gamma subfamily A, 5', other_designations='cadherin ME3', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141364331, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606292, no_hdr0=''),
  56111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56111, CurrentID=0, Status='live', Symbol='PCDHGA4', Aliases='PCDH-GAMMA-A4', description='protocadherin gamma subfamily A, 4', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141355025, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606291, no_hdr0=''),
  56112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56112, CurrentID=0, Status='live', Symbol='PCDHGA3', Aliases='PCDH-GAMMA-A3', description='protocadherin gamma subfamily A, 3', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141344034, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606290, no_hdr0=''),
  56113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56113, CurrentID=0, Status='live', Symbol='PCDHGA2', Aliases='PCDH-GAMMA-A2', description='protocadherin gamma subfamily A, 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141338760, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606289, no_hdr0=''),
  56114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56114, CurrentID=0, Status='live', Symbol='PCDHGA1', Aliases='PCDH-GAMMA-A1', description='protocadherin gamma subfamily A, 1', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141330685, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=4, OMIM=606288, no_hdr0=''),
  56115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56115, CurrentID=0, Status='live', Symbol='PCDHG@', Aliases='PCDH2, PCDH4, PCDHG', description='protocadherin gamma cluster', other_designations='protocadherin gamma cluster extended', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141330685, end_position_on_the_genomic_accession=141512979, orientation='plus', exon_count=0, OMIM=604968, no_hdr0=''),
  56116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56116, CurrentID=0, Status='live', Symbol='PCDHB@', Aliases='PCDH3', description='protocadherin beta cluster', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141051394, end_position_on_the_genomic_accession=141248234, orientation='plus', exon_count=0, OMIM=604967, no_hdr0=''),
  56117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56117, CurrentID=0, Status='live', Symbol='PCDHA@', Aliases='CNR1, CNRS1', description='protocadherin alpha cluster, complex locus', other_designations='cadherin-related neuronal receptor 1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140786291, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=0, OMIM=604966, no_hdr0=''),
  56118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56118, CurrentID=0, Status='live', Symbol='PCDHGCT', Aliases='PCDHGC', description='protocadherin gamma constant', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56119, CurrentID=0, Status='live', Symbol='PCDHACT', Aliases='PCDHAC', description='protocadherin alpha constant', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140978949, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56120, CurrentID=0, Status='live', Symbol='PCDHGB8P', Aliases='PCDH-PSI3', description='protocadherin gamma subfamily B, 8 pseudogene', other_designations='protocadherin gamma B8 pseudogene', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141426286, end_position_on_the_genomic_accession=141428258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56121, CurrentID=0, Status='live', Symbol='PCDHB15', Aliases='PCDH-BETA15', description='protocadherin beta 15', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141245349, end_position_on_the_genomic_accession=141248234, orientation='plus', exon_count=1, OMIM=606341, no_hdr0=''),
  56122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56122, CurrentID=0, Status='live', Symbol='PCDHB14', Aliases='PCDH-BETA14', description='protocadherin beta 14', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141222326, end_position_on_the_genomic_accession=141226288, orientation='plus', exon_count=1, OMIM=606340, no_hdr0=''),
  56123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56123, CurrentID=0, Status='live', Symbol='PCDHB13', Aliases='PCDH-BETA13', description='protocadherin beta 13', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141213919, end_position_on_the_genomic_accession=141217421, orientation='plus', exon_count=1, OMIM=606339, no_hdr0=''),
  56124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56124, CurrentID=0, Status='live', Symbol='PCDHB12', Aliases='PCDH-BETA12', description='protocadherin beta 12', other_designations='PCDH-beta-12', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141208719, end_position_on_the_genomic_accession=141212126, orientation='plus', exon_count=1, OMIM=606338, no_hdr0=''),
  56125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56125, CurrentID=0, Status='live', Symbol='PCDHB11', Aliases='ME2, PCDH-BETA11', description='protocadherin beta 11', other_designations='PCDH-beta-11|cadherin ME2', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141199775, end_position_on_the_genomic_accession=141203045, orientation='plus', exon_count=1, OMIM=606337, no_hdr0=''),
  56126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56126, CurrentID=0, Status='live', Symbol='PCDHB10', Aliases='PCDH-BETA10, PCHB10', description='protocadherin beta 10', other_designations='PCDH-beta-10', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141192379, end_position_on_the_genomic_accession=141195640, orientation='plus', exon_count=1, OMIM=606336, no_hdr0=''),
  56127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56127, CurrentID=0, Status='live', Symbol='PCDHB9', Aliases='PCDH-BETA9, PCDH3H', description='protocadherin beta 9', other_designations='PCDH-beta-9|protocadherin-3h', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141187127, end_position_on_the_genomic_accession=141191538, orientation='plus', exon_count=1, OMIM=606335, no_hdr0=''),
  56128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56128, CurrentID=0, Status='live', Symbol='PCDHB8', Aliases='PCDH-BETA8, PCDH3I', description='protocadherin beta 8', other_designations='protocadherin-3i', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141177790, end_position_on_the_genomic_accession=141180500, orientation='plus', exon_count=1, OMIM=606334, no_hdr0=''),
  56129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56129, CurrentID=0, Status='live', Symbol='PCDHB7', Aliases='PCDH-BETA7', description='protocadherin beta 7', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141172621, end_position_on_the_genomic_accession=141176376, orientation='plus', exon_count=1, OMIM=606333, no_hdr0=''),
  56130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56130, CurrentID=0, Status='live', Symbol='PCDHB6', Aliases='PCDH-BETA6', description='protocadherin beta 6', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141150022, end_position_on_the_genomic_accession=141153287, orientation='plus', exon_count=1, OMIM=606332, no_hdr0=''),
  56131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56131, CurrentID=0, Status='live', Symbol='PCDHB4', Aliases='PCDH-BETA4', description='protocadherin beta 4', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141121750, end_position_on_the_genomic_accession=141125623, orientation='plus', exon_count=1, OMIM=606330, no_hdr0=''),
  56132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56132, CurrentID=0, Status='live', Symbol='PCDHB3', Aliases='PCDH-BETA3', description='protocadherin beta 3', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141100242, end_position_on_the_genomic_accession=141103827, orientation='plus', exon_count=1, OMIM=606329, no_hdr0=''),
  56133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56133, CurrentID=0, Status='live', Symbol='PCDHB2', Aliases='PCDH-BETA2', description='protocadherin beta 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141094607, end_position_on_the_genomic_accession=141097380, orientation='plus', exon_count=1, OMIM=606328, no_hdr0=''),
  56134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56134, CurrentID=0, Status='live', Symbol='PCDHAC2', Aliases='PCDH-ALPHA-C2', description='protocadherin alpha subfamily C, 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140966162, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606321, no_hdr0=''),
  56135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56135, CurrentID=0, Status='live', Symbol='PCDHAC1', Aliases='PCDH-ALPHA-C1', description='protocadherin alpha subfamily C, 1', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140926717, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606320, no_hdr0=''),
  56136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56136, CurrentID=0, Status='live', Symbol='PCDHA13', Aliases='CNR5, CNRN5, CNRS5, CRNR5, PCDH-ALPHA13', description='protocadherin alpha 13', other_designations='KIAA0345-like 1|PCDH-alpha-13|ortholog of mouse CNR5', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140882269, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606319, no_hdr0=''),
  56137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56137, CurrentID=0, Status='live', Symbol='PCDHA12', Aliases='PCDH-ALPHA12', description='protocadherin alpha 12', other_designations='KIAA0345-like 2|PCDH-alpha-12', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140875302, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606318, no_hdr0=''),
  56138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56138, CurrentID=0, Status='live', Symbol='PCDHA11', Aliases='CNR7, CNRN7, CNRS7, CRNR7, PCDH-ALPHA11', description='protocadherin alpha 11', other_designations='KIAA0345-like 3', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140867513, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606317, no_hdr0=''),
  56139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56139, CurrentID=0, Status='live', Symbol='PCDHA10', Aliases='CNR8, CNRN8, CNRS8, CRNR8, PCDH-ALPHA10', description='protocadherin alpha 10', other_designations='KIAA0345-like 4|PCDH-alpha-10|ortholog to mouse CNR8', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140855883, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606316, no_hdr0=''),
  56140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56140, CurrentID=0, Status='live', Symbol='PCDHA8', Aliases='PCDH-ALPHA8', description='protocadherin alpha 8', other_designations='KIAA0345-like 6|PCDH-alpha-8', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140841322, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606314, no_hdr0=''),
  56141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56141, CurrentID=0, Status='live', Symbol='PCDHA7', Aliases='CNR4, CNRN4, CNRS4, CRNR4, PCDH-ALPHA7', description='protocadherin alpha 7', other_designations='KIAA0345-like 7|PCDH-alpha-7|ortholog to mouse CNR4', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140834384, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606313, no_hdr0=''),
  56142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56142, CurrentID=0, Status='live', Symbol='PCDHA6', Aliases='CNR2, CNRN2, CNRS2, CRNR2, PCDH-ALPHA6', description='protocadherin alpha 6', other_designations='KIAA0345-like 8|PCDH-alpha-6', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140827922, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606312, no_hdr0=''),
  56143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56143, CurrentID=0, Status='live', Symbol='PCDHA5', Aliases='CNR6, CNRN6, CNRS6, CRNR6, PCDH-ALPHA5', description='protocadherin alpha 5', other_designations='KIAA0345-like 9|PCDH-alpha-5|ortholog of mouse CNR6', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140821776, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606311, no_hdr0=''),
  56144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56144, CurrentID=0, Status='live', Symbol='PCDHA4', Aliases='CNR1, CNRN1, CRNR1, PCDH-ALPHA4', description='protocadherin alpha 4', other_designations='KIAA0345-like 10|PCDH-alpha-4|ortholog of mouse CNR1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140807074, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606310, no_hdr0=''),
  56145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56145, CurrentID=0, Status='live', Symbol='PCDHA3', Aliases='PCDH-ALPHA3', description='protocadherin alpha 3', other_designations='KIAA0345-like 11|PCDH-alpha-3', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140801198, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606309, no_hdr0=''),
  56146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56146, CurrentID=0, Status='live', Symbol='PCDHA2', Aliases='PCDH-ALPHA2', description='protocadherin alpha 2', other_designations='KIAA0345-like 12', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140794859, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=5, OMIM=606308, no_hdr0=''),
  56147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56147, CurrentID=0, Status='live', Symbol='PCDHA1', Aliases='PCDH-ALPHA1', description='protocadherin alpha 1', other_designations='KIAA0345-like 13|PCDH-alpha-1', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140786136, end_position_on_the_genomic_accession=141012344, orientation='plus', exon_count=4, OMIM=606307, no_hdr0=''),
  100126521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126521, CurrentID=0, Status='live', Symbol='TRNAW-CCA', Aliases='', description='transfer RNA tryptophan (anticodon CCA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56154, CurrentID=0, Status='live', Symbol='TEX15', Aliases='CT42', description='testis expressed 15', other_designations='cancer/testis antigen 42', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30831423, end_position_on_the_genomic_accession=30913015, orientation='minus', exon_count=17, OMIM=605795, no_hdr0=''),
  56155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56155, CurrentID=0, Status='live', Symbol='TEX14', Aliases='CT113', description='testis expressed 14', other_designations='cancer/testis antigen 113|protein kinase-like protein SgK307|sugen kinase 307|testis expressed sequence 14|testis-expressed protein 14|testis-expressed sequence 14 protein', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58556677, end_position_on_the_genomic_accession=58692079, orientation='minus', exon_count=34, OMIM=605792, no_hdr0=''),
  56156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56156, CurrentID=0, Status='live', Symbol='TEX13B', Aliases='TGC3B, TSGA5', description='testis expressed 13B', other_designations='testis expressed sequence 13B', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107980864, end_position_on_the_genomic_accession=107982370, orientation='minus', exon_count=3, OMIM=300313, no_hdr0=''),
  56157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56157, CurrentID=0, Status='live', Symbol='TEX13A', Aliases='', description='testis expressed 13A', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=105218929, end_position_on_the_genomic_accession=105220693, orientation='minus', exon_count=3, OMIM=300312, no_hdr0=''),
  56158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56158, CurrentID=0, Status='live', Symbol='TEX12', Aliases='', description='testis expressed 12', other_designations='testis expressed sequence 12|testis-expressed sequence 12 protein variant 1|testis-expressed sequence 12 protein variant 2|testis-expressed sequence 12 protein variant 3', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112167372, end_position_on_the_genomic_accession=112172556, orientation='plus', exon_count=5, OMIM=605791, no_hdr0=''),
  56159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56159, CurrentID=0, Status='live', Symbol='TEX11', Aliases='SPGFX2, TGC1, TSGA3', description='testis expressed 11', other_designations='testis expressed sequence 11', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70525561, end_position_on_the_genomic_accession=70908717, orientation='minus', exon_count=34, OMIM=300311, no_hdr0=''),
  56160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56160, CurrentID=0, Status='live', Symbol='NSMCE3', Aliases='HCA4, MAGEG1, MAGEL3, NDNL2, NSE3', description='NSE3 homolog, SMC5-SMC6 complex component', other_designations='MAGE-G1 antigen|hepatocellular carcinoma-associated protein 4|hepatocellular carcinoma-associated protein HCA4|melanoma antigen, family G, 1|necdin-like 2|necdin-like 2, SMC5-SMC6 complex component|necdin-like gene 2|necdin-like protein 2', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29268149, end_position_on_the_genomic_accession=29269816, orientation='minus', exon_count=1, OMIM=608243, no_hdr0=''),
  140432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140432, CurrentID=0, Status='live', Symbol='RNF113B', Aliases='RNF161, ZNF183L1, bA10G5.1', description='ring finger protein 113B', other_designations='zinc finger protein 183-like 1', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98175785, end_position_on_the_genomic_accession=98177267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56163, CurrentID=0, Status='live', Symbol='RNF17', Aliases='Mmip-2, SPATA23, TDRD4', description='ring finger protein 17', other_designations='spermatogenesis associated 23|tudor domain containing 4|tudor domain-containing protein 4', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24747672, end_position_on_the_genomic_accession=24889262, orientation='plus', exon_count=46, OMIM=605793, no_hdr0=''),
  56164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56164, CurrentID=0, Status='live', Symbol='STK31', Aliases='SGK396, TDRD8', description='serine/threonine kinase 31', other_designations='serine/threonine-protein kinase NYD-SPK|sugen kinase 396|tudor domain containing 8', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23710167, end_position_on_the_genomic_accession=23832513, orientation='plus', exon_count=25, OMIM=605790, no_hdr0=''),
  56165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56165, CurrentID=0, Status='live', Symbol='TDRD1', Aliases='CT41.1', description='tudor domain containing 1', other_designations='cancer/testis antigen 41.1|tudor domain containing protein 1', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114174442, end_position_on_the_genomic_accession=114232669, orientation='plus', exon_count=30, OMIM=605796, no_hdr0=''),
  664721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664721, CurrentID=0, Status='live', Symbol='HNRNPA1P27', Aliases='HNRPA1P5', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 27', other_designations='heterogeneous nuclear ribonucleoprotein A1 pseudogene 5', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100887929, end_position_on_the_genomic_accession=100888628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56168, CurrentID=0, Status='live', Symbol='MTND2P2', Aliases='bA522L3.2', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 2', other_designations='MT-ND2 pseudogene 2|MTND2 pseudogene 2|NADH dehydrogenase 2 pseudogene 2', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102786814, end_position_on_the_genomic_accession=102788388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56169, CurrentID=0, Status='live', Symbol='GSDMC', Aliases='MLZE', description='gasdermin C', other_designations='melanoma-derived leucine zipper, extra-nuclear factor|melanoma-derived leucine zipper-containing extranuclear factor', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129705903, end_position_on_the_genomic_accession=129786888, orientation='minus', exon_count=22, OMIM=608384, no_hdr0=''),
  56171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56171, CurrentID=0, Status='live', Symbol='DNAH7', Aliases='', description='dynein, axonemal, heavy chain 7', other_designations='axonemal beta dynein heavy chain 7|axonemal dynein heavy chain 7|ciliary dynein heavy chain 7|dynein heavy chain-like protein 2|dynein, axonemal, heavy polypeptide 7|hDHC2', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195737703, end_position_on_the_genomic_accession=196075186, orientation='minus', exon_count=70, OMIM=610061, no_hdr0=''),
  56172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56172, CurrentID=0, Status='live', Symbol='ANKH', Aliases='ANK, CCAL2, CMDJ, CPPDD, HANK, MANK', description='ANKH inorganic pyrophosphate transport regulator', other_designations='ankylosis, progressive homolog', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14704800, end_position_on_the_genomic_accession=14871778, orientation='minus', exon_count=13, OMIM=605145, no_hdr0=''),
  664722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664722, CurrentID=0, Status='live', Symbol='HNRNPA1P26', Aliases='HNRPA1P6', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 26', other_designations='heterogeneous nuclear ribonucleoprotein A1 pseudogene 6', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100854575, end_position_on_the_genomic_accession=100856441, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56180, CurrentID=0, Status='live', Symbol='MOSPD1', Aliases='DJ473B4', description='motile sperm domain containing 1', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134887626, end_position_on_the_genomic_accession=134915267, orientation='minus', exon_count=7, OMIM=300674, no_hdr0=''),
  56181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56181, CurrentID=0, Status='live', Symbol='MTFR1L', Aliases='FAM54B, HYST1888, MST116, MSTP116', description='mitochondrial fission regulator 1-like', other_designations='family with sequence similarity 54 member B|family with sequence similarity 54, member B|protein FAM54B', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25819906, end_position_on_the_genomic_accession=25832942, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  642879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642879, CurrentID=0, Status='live', Symbol='SMARCE1P4', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 4', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37095097, end_position_on_the_genomic_accession=37096407, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100170218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100170218, CurrentID=0, Status='live', Symbol='SNAR-C2', Aliases='', description='small ILF3/NF90-associated RNA C2', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47928818, end_position_on_the_genomic_accession=47928937, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369409, CurrentID=0, Status='live', Symbol='LOC105369409', Aliases='', description='uncharacterized LOC105369409', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80709047, end_position_on_the_genomic_accession=80737278, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56203, CurrentID=0, Status='live', Symbol='LMOD3', Aliases='NEM10', description='leiomodin 3', other_designations='leiomodin 3 (fetal)|leiomodin, fetal form', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69108672, end_position_on_the_genomic_accession=69122595, orientation='minus', exon_count=4, OMIM=616112, no_hdr0=''),
  56204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56204, CurrentID=0, Status='live', Symbol='FAM214A', Aliases='KIAA1370', description='family with sequence similarity 214 member A', other_designations='family with sequence similarity 214, member A', map_location='15q21.2-q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52581321, end_position_on_the_genomic_accession=52709817, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  664728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664728, CurrentID=0, Status='live', Symbol='CMD1Q', Aliases='', description='cardiomyopathy, dilated 1Q (autosomal dominant)', other_designations='', map_location='7q22.3-q31.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609915, no_hdr0=''),
  105369411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369411, CurrentID=0, Status='live', Symbol='LOC105369411', Aliases='', description='uncharacterized LOC105369411', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=81821274, end_position_on_the_genomic_accession=81824932, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369412, CurrentID=0, Status='live', Symbol='LOC105369412', Aliases='', description='uncharacterized LOC105369412', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82713069, end_position_on_the_genomic_accession=82718322, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369413, CurrentID=0, Status='live', Symbol='LOC105369413', Aliases='', description='uncharacterized LOC105369413', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82781524, end_position_on_the_genomic_accession=82849036, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369414, CurrentID=0, Status='live', Symbol='LOC105369414', Aliases='', description='uncharacterized LOC105369414', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83094781, end_position_on_the_genomic_accession=83117273, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369415, CurrentID=0, Status='live', Symbol='LOC105369415', Aliases='', description='uncharacterized LOC105369415', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82983640, end_position_on_the_genomic_accession=82998256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56241, CurrentID=0, Status='live', Symbol='SUSD2', Aliases='BK65A6.2', description='sushi domain containing 2', other_designations='Sushi domain (SCR repeat) containing', map_location='22q11-q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24181476, end_position_on_the_genomic_accession=24189106, orientation='plus', exon_count=15, OMIM=615825, no_hdr0=''),
  56242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56242, CurrentID=0, Status='live', Symbol='ZNF253', Aliases='BMZF-1, BMZF1, ZNF411', description='zinc finger protein 253', other_designations='DNA-binding protein|bone marrow zinc finger 1|zinc finger protein 411', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19865830, end_position_on_the_genomic_accession=19894449, orientation='plus', exon_count=4, OMIM=606954, no_hdr0=''),
  56243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56243, CurrentID=0, Status='live', Symbol='KIAA1217', Aliases='SKT', description='KIAA1217', other_designations='likely orthologue of Mus musculus enhancer trap locus 4', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23694746, end_position_on_the_genomic_accession=24547848, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  56244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56244, CurrentID=0, Status='live', Symbol='BTNL2', Aliases='BTL-II, BTN7, HSBLMHC1, SS2', description='butyrophilin-like 2', other_designations='butyrophilin-like 2 (MHC class II associated)', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32393339, end_position_on_the_genomic_accession=32408879, orientation='minus', exon_count=8, OMIM=606000, no_hdr0=''),
  56245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56245, CurrentID=0, Status='live', Symbol='C21orf62', Aliases='B37, C21orf120, PRED81', description='chromosome 21 open reading frame 62', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32790673, end_position_on_the_genomic_accession=32814181, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56246, CurrentID=0, Status='live', Symbol='MRAP', Aliases='B27, C21orf61, FALP, FGD2, GCCD2', description='melanocortin 2 receptor accessory protein', other_designations='fat cell-specific low molecular weight protein|fat tissue-specific low MW protein', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32291813, end_position_on_the_genomic_accession=32314784, orientation='plus', exon_count=6, OMIM=609196, no_hdr0=''),
  56247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56247, CurrentID=0, Status='live', Symbol='FMR3', Aliases='', description='fragile X mental retardation associated 3', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56252, CurrentID=0, Status='live', Symbol='YLPM1', Aliases='C14orf170, PPP1R169, ZAP113, ZAP3', description='YLP motif containing 1', other_designations='nuclear protein ZAP3|protein phosphatase 1, regulatory subunit 169', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74763322, end_position_on_the_genomic_accession=74837310, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  56253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56253, CurrentID=0, Status='live', Symbol='CRTAM', Aliases='CD355', description='cytotoxic and regulatory T-cell molecule', other_designations='class I MHC restricted T cell associated molecule|class-I MHC-restricted T cell associated molecule|class-I MHC-restricted T-cell-associated molecule|cytotoxic and regulatory T cell molecule', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122838431, end_position_on_the_genomic_accession=122872643, orientation='plus', exon_count=10, OMIM=612597, no_hdr0=''),
  56254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56254, CurrentID=0, Status='live', Symbol='RNF20', Aliases='BRE1, BRE1A, hBRE1', description='ring finger protein 20', other_designations='BRE1 E3 ubiquitin ligase homolog|BRE1-A|homolog of S. cerevisiae BRE1|ring finger protein 20, E3 ubiquitin protein ligase', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101533849, end_position_on_the_genomic_accession=101563344, orientation='plus', exon_count=21, OMIM=607699, no_hdr0=''),
  56255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56255, CurrentID=0, Status='live', Symbol='TMX4', Aliases='DJ971N18.2, PDIA14, TXNDC13', description='thioredoxin-related transmembrane protein 4', other_designations='protein disulfide isomerase family A, member 14|thioredoxin domain containing 13|thioredoxin domain-containing protein 13', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7977353, end_position_on_the_genomic_accession=8019829, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56256, CurrentID=0, Status='live', Symbol='SERTAD4', Aliases='DJ667H12.2', description='SERTA domain containing 4', other_designations='', map_location='1q32.1-q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210232850, end_position_on_the_genomic_accession=210246858, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56257, CurrentID=0, Status='live', Symbol='MEPCE', Aliases='BCDIN3', description='methylphosphate capping enzyme', other_designations='bicoid-interacting protein 3 homolog|bin3 homolog|bin3, bicoid-interacting 3, homolog', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100428698, end_position_on_the_genomic_accession=100434126, orientation='plus', exon_count=5, OMIM=611478, no_hdr0=''),
  56259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56259, CurrentID=0, Status='live', Symbol='CTNNBL1', Aliases='C20orf33, NAP, P14L, PP8304, dJ633O20.1', description='catenin beta like 1', other_designations='catenin, beta like 1|nuclear associated protein|testis development protein NYD-SP19', map_location='20q11.23-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37693955, end_position_on_the_genomic_accession=37872129, orientation='plus', exon_count=18, OMIM=611537, no_hdr0=''),
  56260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56260, CurrentID=0, Status='live', Symbol='C8orf44', Aliases='', description='chromosome 8 open reading frame 44', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66667552, end_position_on_the_genomic_accession=66681149, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56261, CurrentID=0, Status='live', Symbol='GPCPD1', Aliases='EDI3, GDE5, GDPD6, PREI4', description='glycerophosphocholine phosphodiesterase 1', other_designations='endometrial differential 3|glycerophosphocholine phosphodiesterase GDE1 homolog|glycerophosphodiester phosphodiesterase 5|preimplantation protein 4|putative glycerophosphocholine phosphodiesterase GPCPD1|putative glycerophosphodiester phosphodiesterase 5', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5544434, end_position_on_the_genomic_accession=5611026, orientation='minus', exon_count=23, OMIM=614124, no_hdr0=''),
  56262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56262, CurrentID=0, Status='live', Symbol='LRRC8A', Aliases='AGM5, LRRC8, SWELL1', description='leucine-rich repeat containing 8 family member A', other_designations='leucine rich repeat containing 8 family, member A|leucine-rich repeat-containing protein 8A|swelling protein 1', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128882112, end_position_on_the_genomic_accession=128918042, orientation='plus', exon_count=10, OMIM=608360, no_hdr0=''),
  56265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56265, CurrentID=0, Status='live', Symbol='CPXM1', Aliases='CPX1, CPXM', description='carboxypeptidase X (M14 family), member 1', other_designations='carboxypeptidase-like protein X1|metallocarboxypeptidase CPX-1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2794069, end_position_on_the_genomic_accession=2800646, orientation='minus', exon_count=14, OMIM=609555, no_hdr0=''),
  56267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56267, CurrentID=0, Status='live', Symbol='CCBL2', Aliases='KAT3, KATIII', description='cysteine conjugate-beta lyase 2', other_designations='RP11-82K18.3|cysteine-S-conjugate beta-lyase 2|kynurenine aminotransferase III|kynurenine--glyoxylate transaminase|kynurenine--oxoglutarate transaminase III|kynurenine-oxoglutarate transaminase 3', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88935773, end_position_on_the_genomic_accession=88992960, orientation='minus', exon_count=14, OMIM=610656, no_hdr0=''),
  449484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449484, CurrentID=0, Status='live', Symbol='DFNB44', Aliases='', description='deafness, autosomal recessive 44', other_designations='', map_location='7p14.1-q11.22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610154, no_hdr0=''),
  56269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56269, CurrentID=0, Status='live', Symbol='IRGC', Aliases='CINEMA, IFGGE1, Iigp5, R30953_1, IRGC', description='immunity-related GTPase family, cinema', other_designations='immunity-related GTPase cinema 1|immunity-related GTPase family, cinema 1|interferon inducible GTPase 5|interferon-gamma-inducible GTPase IFGGE protein', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43716058, end_position_on_the_genomic_accession=43720036, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56270, CurrentID=0, Status='live', Symbol='WDR45B', Aliases='WDR45L, WIPI-3, WIPI3', description='WD repeat domain 45B', other_designations='WD repeat protein 45-like|WD repeat-containing protein 45-like|WD repeat-containing protein 45B|WDR45-like protein|WIPI49-like protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82614562, end_position_on_the_genomic_accession=82648535, orientation='minus', exon_count=12, OMIM=609226, no_hdr0=''),
  56271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56271, CurrentID=0, Status='live', Symbol='BEX4', Aliases='BEXL1', description='brain expressed X-linked 4', other_designations='BEX family member 4|BEX1-like protein 1|brain expressed X-linked-like 1|brain expressed, X-linked 4|nerve growth factor receptor-associated protein 3', map_location='Xq22.1-q22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103215092, end_position_on_the_genomic_accession=103217200, orientation='plus', exon_count=3, OMIM=300692, no_hdr0=''),
  449488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449488, CurrentID=0, Status='live', Symbol='DFNB46', Aliases='', description='deafness, autosomal recessive 46', other_designations='', map_location='18p11.32-p11.31', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609647, no_hdr0=''),
  449489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449489, CurrentID=0, Status='live', Symbol='DFNB47', Aliases='', description='deafness, autosomal recessive 47', other_designations='', map_location='2p25.1-p24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609946, no_hdr0=''),
  9379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9379, CurrentID=0, Status='live', Symbol='NRXN2', Aliases='', description='neurexin 2', other_designations='neurexin II', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64606174, end_position_on_the_genomic_accession=64723188, orientation='minus', exon_count=30, OMIM=600566, no_hdr0=''),
  449492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449492, CurrentID=0, Status='live', Symbol='DEFA9P', Aliases='DEFAP2', description='defensin alpha 9, pseudogene', other_designations='defensin, alpha 9 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6959289, end_position_on_the_genomic_accession=6960161, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  449493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449493, CurrentID=0, Status='live', Symbol='DEFA10P', Aliases='', description='defensin alpha 10, pseudogene', other_designations='defensin, alpha 10 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6968141, end_position_on_the_genomic_accession=6969113, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246287, CurrentID=0, Status='live', Symbol='DFNA45', Aliases='', description='deafness, autosomal dominant 45', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56286, CurrentID=0, Status='live', Symbol='DAD1P1', Aliases='DAD1L, DADR', description='defender against cell death 1 pseudogene 1', other_designations='DAD1-related', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24917702, end_position_on_the_genomic_accession=24918384, orientation='minus', exon_count=0, OMIM=609860, no_hdr0=''),
  56287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56287, CurrentID=0, Status='live', Symbol='GKN1', Aliases='AMP18, BRICD1, CA11, FOV, foveolin', description='gastrokine 1', other_designations='18 kDa antrum mucosa protein|AMP-18|BRICHOS domain containing 1', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68974573, end_position_on_the_genomic_accession=68980980, orientation='plus', exon_count=6, OMIM=606402, no_hdr0=''),
  56288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56288, CurrentID=0, Status='live', Symbol='PARD3', Aliases='ASIP, Baz, PAR3, PAR3alpha, PARD-3A, PPP1R118, SE2-5L16, SE2-5LT1, SE2-5T2, PARD3', description='par-3 family cell polarity regulator', other_designations='CTCL tumor antigen se2-5|PAR3-alpha|atypical PKC isotype-specific interacting protein|atypical PKC isotype-specific-interacting protein|bazooka|par-3 family cell polarity regulator alpha|par-3 partitioning defective 3 homolog|protein phosphatase 1, regulatory subunit 118', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34109560, end_position_on_the_genomic_accession=34815325, orientation='minus', exon_count=30, OMIM=606745, no_hdr0=''),
  140454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140454, CurrentID=0, Status='live', Symbol='PSORS8', Aliases='', description='psoriasis susceptibility 8', other_designations='', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610707, no_hdr0=''),
  642897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642897, CurrentID=0, Status='live', Symbol='LOC642897', Aliases='', description='N-6 adenine-specific DNA methyltransferase 1 (putative) pseudogene', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93609432, end_position_on_the_genomic_accession=93610414, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56300, CurrentID=0, Status='live', Symbol='IL36G', Aliases='IL-1F9, IL-1H1, IL-1RP2, IL1E, IL1F9, IL1H1, IL1RP2', description='interleukin 36, gamma', other_designations='IL-1 epsilon|IL-1 related protein 2|IL-1(EPSILON)|IL-1-epsilon|IL-1-related protein 2|interleukin 1 family, member 9|interleukin 1-related protein 2|interleukin-1 epsilon|interleukin-1 family member 9|interleukin-1 homolog 1', map_location='2q12-q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112972329, end_position_on_the_genomic_accession=112985672, orientation='plus', exon_count=7, OMIM=605542, no_hdr0=''),
  56301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56301, CurrentID=0, Status='live', Symbol='SLC7A10', Aliases='ASC1, HASC-1, asc-1', description='solute carrier family 7 (neutral amino acid transporter light chain, asc system), member 10', other_designations='solute carrier family 7 member 10|solute carrier family 7, (cationic amino acid transporter, y+ system) member 10|solute carrier family 7, (neutral amino acid transporter, y+ system) member 10', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33208664, end_position_on_the_genomic_accession=33225850, orientation='minus', exon_count=9, OMIM=607959, no_hdr0=''),
  56302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56302, CurrentID=0, Status='live', Symbol='TRPV5', Aliases='CAT2, ECAC1, OTRPC3', description='transient receptor potential cation channel, subfamily V, member 5', other_designations='calcium transport protein 2|calcium transporter 2|epithelial calcium channel 1|osm-9-like TRP channel 3', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142908101, end_position_on_the_genomic_accession=142933723, orientation='minus', exon_count=15, OMIM=606679, no_hdr0=''),
  449519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449519, CurrentID=0, Status='live', Symbol='LOC449519', Aliases='', description='Mas-related gene LOC449519', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  449520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449520, CurrentID=0, Status='live', Symbol='GGNBP1', Aliases='', description='gametogenetin binding protein 1 (pseudogene)', other_designations='gametogenetin binding protein 1 pseudogene', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33583699, end_position_on_the_genomic_accession=33589026, orientation='plus', exon_count=6, OMIM=609495, no_hdr0=''),
  9384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9384, CurrentID=0, Status='live', Symbol='COX6CP1', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 1', other_designations='', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11903801, end_position_on_the_genomic_accession=11904193, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  449523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449523, CurrentID=0, Status='live', Symbol='JST', Aliases='', description='tumor suppressor gene JST', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56311, CurrentID=0, Status='live', Symbol='ANKRD7', Aliases='TSA806', description='ankyrin repeat domain 7', other_designations='testis-specific ankyrin motif containing protein|testis-specific protein TSA806', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118224658, end_position_on_the_genomic_accession=118242731, orientation='plus', exon_count=7, OMIM=610731, no_hdr0=''),
  100128335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128335, CurrentID=0, Status='live', Symbol='LOC100128335', Aliases='', description='WD repeat domain 12 pseudogene', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=84448630, end_position_on_the_genomic_accession=84449878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  9386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9386, CurrentID=0, Status='live', Symbol='COX7CP1', Aliases='', description='cytochrome c oxidase subunit VIIc pseudogene 1', other_designations='', map_location='13q14-q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49187359, end_position_on_the_genomic_accession=49187749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370124, CurrentID=0, Status='live', Symbol='LOC105370124', Aliases='', description='uncharacterized LOC105370124', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27046159, end_position_on_the_genomic_accession=27055767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  140460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140460, CurrentID=0, Status='live', Symbol='ASB7', Aliases='', description='ankyrin repeat and SOCS box containing 7', other_designations='ASB-7|ankyrin repeat and SOCS box-containing 7', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100602550, end_position_on_the_genomic_accession=100651701, orientation='plus', exon_count=6, OMIM=615052, no_hdr0=''),
  105374489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374489, CurrentID=0, Status='live', Symbol='LOC105374489', Aliases='', description='uncharacterized LOC105374489', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11822183, end_position_on_the_genomic_accession=11823880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9389, CurrentID=0, Status='live', Symbol='SLC22A14', Aliases='OCTL2, OCTL4, ORCTL4', description='solute carrier family 22 member 14', other_designations='ORCTL-4|organic cation transporter-like 4|organic cationic transporter-like 4|solute carrier family 22 (organic cation transporter), member 14|solute carrier family 22, member 14', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38281430, end_position_on_the_genomic_accession=38318522, orientation='plus', exon_count=12, OMIM=604048, no_hdr0=''),
  56339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56339, CurrentID=0, Status='live', Symbol='METTL3', Aliases='IME4, M6A, MT-A70, Spo8', description='methyltransferase like 3', other_designations="adoMet-binding subunit of the human mRNA (N6-adenosine)-methyltransferase|mRNA (2'-O-methyladenosine-N(6)-)-methyltransferase|mRNA m(6)A methyltransferase|methyltransferase-like protein 3", map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21498131, end_position_on_the_genomic_accession=21511375, orientation='minus', exon_count=11, OMIM=612472, no_hdr0=''),
  9390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9390, CurrentID=0, Status='live', Symbol='SLC22A13', Aliases='OAT10, OCTL1, OCTL3, ORCTL-3, ORCTL3', description='solute carrier family 22 (organic anion/urate transporter), member 13', other_designations='organic cation transporter-like 3|organic cationic transporter-like 3|organic-cation transporter like 3|solute carrier family 22 (organic anion transporter), member 13|solute carrier family 22, member 13', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38265807, end_position_on_the_genomic_accession=38278315, orientation='plus', exon_count=10, OMIM=604047, no_hdr0=''),
  56342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56342, CurrentID=0, Status='live', Symbol='PPAN', Aliases='BXDC3, SSF, SSF-1, SSF1, SSF2', description='peter pan homolog (Drosophila)', other_designations='brix domain-containing protein 3|homolog of S. cerevisiae SSF1|second-step splicing factor 1|suppressor of sterile four 1', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10106223, end_position_on_the_genomic_accession=10111299, orientation='plus', exon_count=12, OMIM=607793, no_hdr0=''),
  105369433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369433, CurrentID=0, Status='live', Symbol='LOC105369433', Aliases='', description='uncharacterized LOC105369433', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56344, CurrentID=0, Status='live', Symbol='CABP5', Aliases='CABP3', description='calcium binding protein 5', other_designations='calcium-binding protein 3', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48029383, end_position_on_the_genomic_accession=48044054, orientation='minus', exon_count=7, OMIM=607315, no_hdr0=''),
  402481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402481, CurrentID=0, Status='live', Symbol='PP13004', Aliases='', description='uncharacterized LOC402481', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=36118694, end_position_on_the_genomic_accession=36125338, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  140464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140464, CurrentID=0, Status='live', Symbol='PISRT1', Aliases='NCRNA00195', description='polled intersex syndrome regulated transcript 1 (non-protein coding RNA)', other_designations='polled intersex syndrome regulated transcript 1 homolog (goat)', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139232992, end_position_on_the_genomic_accession=139233522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369435, CurrentID=0, Status='live', Symbol='LOC105369435', Aliases='', description='uncharacterized LOC105369435', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94199727, end_position_on_the_genomic_accession=94282055, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369437, CurrentID=0, Status='live', Symbol='LOC105369437', Aliases='', description='uncharacterized LOC105369437', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94804059, end_position_on_the_genomic_accession=94807421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369439, CurrentID=0, Status='live', Symbol='LOC105369439', Aliases='', description='uncharacterized LOC105369439', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95466101, end_position_on_the_genomic_accession=95480923, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140469, CurrentID=0, Status='live', Symbol='MYO3B', Aliases='', description='myosin IIIB', other_designations='', map_location='2q31.1-q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170178145, end_position_on_the_genomic_accession=170655164, orientation='plus', exon_count=43, OMIM=610040, no_hdr0=''),
  449619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449619, CurrentID=0, Status='live', Symbol='ERVK-7', Aliases='ERVK7, HERV-K(III), HERV-K102, HERV-KIII', description='endogenous retrovirus group K member 7', other_designations='Envelope polyprotein|HERV-K_1q22 provirus Np9 protein|HERV-K_1q22 provirus ancestral Env polyprotein|HERV-K_1q22 provirus ancestral Gag polyprotein|HERV-K_1q22 provirus ancestral Pol protein|HERV-K_1q22 provirus ancestral Pro protein|endogenous retroviral sequence K, 7 (type 2)|endogenous retrovirus group K, member 7', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614013, no_hdr0=''),
  56413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56413, CurrentID=0, Status='live', Symbol='LTB4R2', Aliases='BLT2, BLTR2, JULF2, KPG_004, LTB4-R 2, LTB4-R2, NOP9', description='leukotriene B4 receptor 2', other_designations='LTB4 receptor JULF2|leukotriene B4 receptor BLT2|seven transmembrane receptor BLTR2', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24310151, end_position_on_the_genomic_accession=24312053, orientation='plus', exon_count=2, OMIM=605773, no_hdr0=''),
  449645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449645, CurrentID=0, Status='live', Symbol='SF3A3P2', Aliases='', description='splicing factor 3a, subunit 3 pseudogene 2', other_designations='SF3A3 pseudogene|splicing factor 3a, subunit 3, 60kDa pseudogene', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123596738, end_position_on_the_genomic_accession=123599474, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369449, CurrentID=0, Status='live', Symbol='LOC105369449', Aliases='', description='uncharacterized LOC105369449', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97222643, end_position_on_the_genomic_accession=97259565, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642924, CurrentID=0, Status='live', Symbol='LINC00535', Aliases='', description='long intergenic non-protein coding RNA 535', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93346467, end_position_on_the_genomic_accession=93700433, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105372304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372304, CurrentID=0, Status='live', Symbol='LOC105372304', Aliases='', description='uncharacterized LOC105372304', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19708400, end_position_on_the_genomic_accession=19710547, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56474, CurrentID=0, Status='live', Symbol='CTPS2', Aliases='', description='CTP synthase 2', other_designations="CTP synthase II|CTP synthetase type 2|UTP-ammonia ligase 2|cytidine 5'-triphosphate synthetase 2", map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16587999, end_position_on_the_genomic_accession=16712979, orientation='minus', exon_count=23, OMIM=300380, no_hdr0=''),
  56475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56475, CurrentID=0, Status='live', Symbol='RPRM', Aliases='REPRIMO', description='reprimo, TP53 dependent G2 arrest mediator candidate', other_designations='candidate mediator of the p53 dependent G2 arrest|candidate mediator of the p53-dependent G2 arrest|reprimo, TP53 dependant G2 arrest mediator candidate', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153477338, end_position_on_the_genomic_accession=153478808, orientation='minus', exon_count=1, OMIM=612171, no_hdr0=''),
  56476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56476, CurrentID=0, Status='live', Symbol='ABCB10P1', Aliases='ABCB10P, ABCB10P2, M-ABC2, MABC2', description='ATP binding cassette subfamily B member 10 pseudogene 1', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene 1|ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23179882, end_position_on_the_genomic_accession=23183659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56477, CurrentID=0, Status='live', Symbol='CCL28', Aliases='CCK1, MEC, SCYA28', description='chemokine (C-C motif) ligand 28', other_designations='CC chemokine CCL28|chemokine (C-C motif) ligand 28 splice variant chi|mucosae-associated epithelial chemokine|small inducible cytokine A28|small inducible cytokine subfamily A (Cys-Cys), member 28|small-inducible cytokine A28', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43356972, end_position_on_the_genomic_accession=43412391, orientation='minus', exon_count=8, OMIM=605240, no_hdr0=''),
  56478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56478, CurrentID=0, Status='live', Symbol='EIF4ENIF1', Aliases='4E-T, Clast4', description='eukaryotic translation initiation factor 4E nuclear import factor 1', other_designations='2610509L04Rik|eIF4E transporter|eIF4E-transporter', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31436963, end_position_on_the_genomic_accession=31489950, orientation='minus', exon_count=24, OMIM=607445, no_hdr0=''),
  56479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56479, CurrentID=0, Status='live', Symbol='KCNQ5', Aliases='Kv7.5', description='potassium channel, voltage gated KQT-like subfamily Q, member 5', other_designations='KQT-like 5|potassium channel protein|potassium channel subunit alpha KvLQT5|potassium voltage-gated channel, KQT-like subfamily, member 5|voltage-gated potassium channel subunit Kv7.5', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72621280, end_position_on_the_genomic_accession=73198851, orientation='plus', exon_count=16, OMIM=607357, no_hdr0=''),
  105374494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374494, CurrentID=0, Status='live', Symbol='LOC105374494', Aliases='', description='uncharacterized LOC105374494', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=12996485, end_position_on_the_genomic_accession=13241681, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369457, CurrentID=0, Status='live', Symbol='LOC105369457', Aliases='', description='uncharacterized LOC105369457', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100865745, end_position_on_the_genomic_accession=100866157, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929053, CurrentID=0, Status='live', Symbol='LOC101929053', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12247353, end_position_on_the_genomic_accession=12249430, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  23704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23704, CurrentID=0, Status='live', Symbol='KCNE4', Aliases='MIRP3', description='potassium channel, voltage gated subfamily E regulatory beta subunit 4', other_designations='MINK-related peptide 3|cardiac voltage-gated potassium channel accessory subunit 4|minimum potassium ion channel-related peptide 3|potassium channel subunit beta MiRP3|potassium voltage-gated channel, Isk-related family, member 4', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223051930, end_position_on_the_genomic_accession=223055637, orientation='plus', exon_count=2, OMIM=607775, no_hdr0=''),
  105369461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369461, CurrentID=0, Status='live', Symbol='LOC105369461', Aliases='', description='uncharacterized LOC105369461', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102466691, end_position_on_the_genomic_accession=102485765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  402635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402635, CurrentID=0, Status='live', Symbol='GRIFIN', Aliases='', description='galectin-related inter-fiber protein', other_designations='putative grifin', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2474636, end_position_on_the_genomic_accession=2476855, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100126582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126582, CurrentID=0, Status='live', Symbol='LOC100126582', Aliases='', description='uncharacterized LOC100126582', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62531308, end_position_on_the_genomic_accession=62533651, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  664780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=664780, CurrentID=0, Status='live', Symbol='MYP12', Aliases='', description='myopia 12 (high grade, autosomal dominant)', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609995, no_hdr0=''),
  105369463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369463, CurrentID=0, Status='live', Symbol='LOC105369463', Aliases='', description='uncharacterized LOC105369463', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103565865, end_position_on_the_genomic_accession=103644551, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100126584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126584, CurrentID=0, Status='live', Symbol='LOC100126584', Aliases='', description='uncharacterized LOC100126584', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125154598, end_position_on_the_genomic_accession=125197863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369466, CurrentID=0, Status='live', Symbol='LOC105369466', Aliases='', description='uncharacterized LOC105369466', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104671229, end_position_on_the_genomic_accession=104696722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374496, CurrentID=0, Status='live', Symbol='LOC105374496', Aliases='', description='uncharacterized LOC105374496', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14513612, end_position_on_the_genomic_accession=14518458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56547, CurrentID=0, Status='live', Symbol='MMP26', Aliases='', description='matrix metallopeptidase 26', other_designations='MMP-26|endometase|matrilysin 2|matrilysin-2|matrix metalloproteinase 26', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4988194, end_position_on_the_genomic_accession=4992429, orientation='plus', exon_count=7, OMIM=605470, no_hdr0=''),
  56548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56548, CurrentID=0, Status='live', Symbol='CHST7', Aliases='C6ST-2, GST-5', description='carbohydrate (N-acetylglucosamine 6-O) sulfotransferase 7', other_designations='N-acetylglucosamine 6-O-sulfotransferase 4|chondroitin 6-sulfotransferase-2|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 5|glcNAc6ST-4|gn6st-4', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46573687, end_position_on_the_genomic_accession=46598496, orientation='plus', exon_count=2, OMIM=300375, no_hdr0=''),
  255025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255025, CurrentID=0, Status='live', Symbol='LINC00879', Aliases='', description='long intergenic non-protein coding RNA 879', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94938263, end_position_on_the_genomic_accession=94991328, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  402641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402641, CurrentID=0, Status='live', Symbol='LOC402641', Aliases='', description='v-ral simian leukemia viral oncogene homolog A (ras related) pseudogene', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20588995, end_position_on_the_genomic_accession=20590089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285577, CurrentID=0, Status='live', Symbol='LINC01019', Aliases='', description='long intergenic non-protein coding RNA 1019', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3417152, end_position_on_the_genomic_accession=3536094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  402643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402643, CurrentID=0, Status='live', Symbol='TPM3P4', Aliases='', description='tropomyosin 3 pseudogene 4', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27023663, end_position_on_the_genomic_accession=27026490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369470, CurrentID=0, Status='live', Symbol='LOC105369470', Aliases='', description='uncharacterized LOC105369470', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105926784, end_position_on_the_genomic_accession=105974333, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369471, CurrentID=0, Status='live', Symbol='LOC105369471', Aliases='', description='uncharacterized LOC105369471', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105798267, end_position_on_the_genomic_accession=105802433, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369473, CurrentID=0, Status='live', Symbol='LOC105369473', Aliases='', description='uncharacterized LOC105369473', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106112456, end_position_on_the_genomic_accession=106132169, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378862, CurrentID=0, Status='live', Symbol='LOC105378862', Aliases='', description='uncharacterized LOC105378862', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94925901, end_position_on_the_genomic_accession=94931255, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100126594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126594, CurrentID=0, Status='live', Symbol='EJM4', Aliases='', description='Myoclonic epilepsy, juvenile, 4', other_designations='', map_location='5q12-q14', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611364, no_hdr0=''),
  642947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642947, CurrentID=0, Status='live', Symbol='LOC642947', Aliases='', description='uncharacterized LOC642947', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=78998528, end_position_on_the_genomic_accession=79000472, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369476, CurrentID=0, Status='live', Symbol='LOC105369476', Aliases='', description='uncharacterized LOC105369476', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106585990, end_position_on_the_genomic_accession=106778512, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56603, CurrentID=0, Status='live', Symbol='CYP26B1', Aliases='CYP26A2, P450RAI-2, P450RAI2, RHFCA', description='cytochrome P450 family 26 subfamily B member 1', other_designations='cytochrome P450 retinoic acid-inactivating 2|cytochrome P450 retinoid metabolizing protein|cytochrome P450, family 26, subfamily B, polypeptide 1|cytochrome P450, subfamily XXVIB, polypeptide 1|retinoic acid-metabolizing cytochrome', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72129238, end_position_on_the_genomic_accession=72147862, orientation='minus', exon_count=8, OMIM=605207, no_hdr0=''),
  56604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56604, CurrentID=0, Status='live', Symbol='TUBB7P', Aliases='TUBB4Q', description='tubulin beta 7 pseudogene', other_designations='putative tubulin beta-4q chain|tubulin, beta 7, pseudogene|tubulin, beta polypeptide 4, member Q, pseudogene', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189982335, end_position_on_the_genomic_accession=189984869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56605, CurrentID=0, Status='live', Symbol='ERO1B', Aliases='ERO1LB, Ero1beta', description='endoplasmic reticulum oxidoreductase beta', other_designations='ERO1-L-beta|ERO1-like beta|endoplasmic oxidoreductin-1-like protein B|endoplasmic reticulum oxidoreductin-1-like protein B|oxidoreductin-1-L-beta', map_location='1q42.2-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236215122, end_position_on_the_genomic_accession=236282039, orientation='minus', exon_count=16, OMIM=615437, no_hdr0=''),
  56606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56606, CurrentID=0, Status='live', Symbol='SLC2A9', Aliases='GLUT9, GLUTX, UAQTL2, URATv1', description='solute carrier family 2 (facilitated glucose transporter), member 9', other_designations='GLUT-9|glucose transporter type 9|human glucose transporter-like protein-9|urate voltage-driven efflux transporter 1', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9771125, end_position_on_the_genomic_accession=10040248, orientation='minus', exon_count=24, OMIM=606142, no_hdr0=''),
  105369477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369477, CurrentID=0, Status='live', Symbol='LOC105369477', Aliases='', description='uncharacterized LOC105369477', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107273289, end_position_on_the_genomic_accession=107314842, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56616, CurrentID=0, Status='live', Symbol='DIABLO', Aliases='DFNA64, SMAC', description='diablo, IAP-binding mitochondrial protein', other_designations='diablo-like protein|direct IAP-binding protein with low pI|second mitochondria-derived activator of caspase', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122207662, end_position_on_the_genomic_accession=122227534, orientation='minus', exon_count=8, OMIM=605219, no_hdr0=''),
  105369479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369479, CurrentID=0, Status='live', Symbol='LOC105369479', Aliases='', description='uncharacterized LOC105369479', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108478128, end_position_on_the_genomic_accession=108481246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56623, CurrentID=0, Status='live', Symbol='INPP5E', Aliases='CORS1, CPD4, JBTS1, MORMS, PPI5PIV', description='inositol polyphosphate-5-phosphatase E', other_designations='inositol polyphosphate-5-phosphatase, 72 kDa|phosphatidylinositol (4,5) bisphosphate 5-phosphatase|phosphatidylinositol 4,5-bisphosphate 5-phosphatase|phosphatidylinositol polyphosphate 5-phosphatase type IV|phosphatidylinositol-4,5-bisphosphate 5-phosphatase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136428615, end_position_on_the_genomic_accession=136439859, orientation='minus', exon_count=10, OMIM=613037, no_hdr0=''),
  56624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56624, CurrentID=0, Status='live', Symbol='ASAH2', Aliases='BCDase, HNAC1, LCDase, N-CDase, NCDase', description='N-acylsphingosine amidohydrolase (non-lysosomal ceramidase) 2', other_designations='N-acylsphingosine amidohydrolase 2|acylsphingosine deacylase 2|mitochondrial ceramidase|neutral/alkaline ceramidase|non-lysosomal ceramidase', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50182778, end_position_on_the_genomic_accession=50251577, orientation='minus', exon_count=22, OMIM=611202, no_hdr0=''),
  105369480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369480, CurrentID=0, Status='live', Symbol='LOC105369480', Aliases='', description='uncharacterized LOC105369480', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108699213, end_position_on_the_genomic_accession=108766450, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369482, CurrentID=0, Status='live', Symbol='LOC105369482', Aliases='', description='uncharacterized LOC105369482', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109780980, end_position_on_the_genomic_accession=109823844, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928721, CurrentID=0, Status='live', Symbol='LOC101928721', Aliases='', description='uncharacterized LOC101928721', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37073652, end_position_on_the_genomic_accession=37133925, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56647, CurrentID=0, Status='live', Symbol='BCCIP', Aliases='TOK-1, TOK1', description='BRCA2 and CDKN1A interacting protein', other_designations='BCCIPalpha|BCCIPbeta|BRCA2 and Cip1/p21 interacting protein|TOK-1alpha|TOK-1beta|cdk inhibitor p21 binding protein|p21- and CDK-associated protein 1|protein TOK-1', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125823535, end_position_on_the_genomic_accession=125853695, orientation='plus', exon_count=9, OMIM=611883, no_hdr0=''),
  56648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56648, CurrentID=0, Status='live', Symbol='EIF5A2', Aliases='EIF-5A2, eIF5AII', description='eukaryotic translation initiation factor 5A2', other_designations='eIF-5A-2|eukaryotic initiation factor 5A', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170888415, end_position_on_the_genomic_accession=170908637, orientation='minus', exon_count=5, OMIM=605782, no_hdr0=''),
  56649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56649, CurrentID=0, Status='live', Symbol='TMPRSS4', Aliases='CAPH2, MT-SP2, TMPRSS3', description='transmembrane protease, serine 4', other_designations='channel-activating protease 2|membrane-type serine protease 2|transmembrane serine protease 3|type II membrane serine protease', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118077012, end_position_on_the_genomic_accession=118125505, orientation='plus', exon_count=16, OMIM=606565, no_hdr0=''),
  56650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56650, CurrentID=0, Status='live', Symbol='CLDND1', Aliases='C3orf4, GENX-3745', description='claudin domain containing 1', other_designations='claudin domain containing 1 protein|membrane protein GENX-3745', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98515473, end_position_on_the_genomic_accession=98523066, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56651, CurrentID=0, Status='live', Symbol='LINC00470', Aliases='C18orf2', description='long intergenic non-protein coding RNA 470', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1268311, end_position_on_the_genomic_accession=1359629, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56652, CurrentID=0, Status='live', Symbol='C10orf2', Aliases='ATXN8, IOSCA, MTDPS7, PEO, PEO1, PEOA3, PRLTS5, SANDO, SCA8, TWINL', description='chromosome 10 open reading frame 2', other_designations='T7 gp4-like protein with intramitochondrial nucleoid localization|T7 helicase-related protein with intramitochondrial nucleoid localization|T7-like mitochondrial DNA helicase|ataxin 8|mitochondrial twinkle protein|progressive external ophthalmoplegia 1 protein', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100987378, end_position_on_the_genomic_accession=100994403, orientation='plus', exon_count=5, OMIM=606075, no_hdr0=''),
  56654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56654, CurrentID=0, Status='live', Symbol='NPDC1', Aliases='CAB, CAB-, CAB-1, CAB1, NPDC-1', description='neural proliferation, differentiation and control, 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137039457, end_position_on_the_genomic_accession=137046224, orientation='minus', exon_count=9, OMIM=605798, no_hdr0=''),
  56655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56655, CurrentID=0, Status='live', Symbol='POLE4', Aliases='YHHQ1, p12', description='polymerase (DNA-directed), epsilon 4, accessory subunit', other_designations='DNA polymerase II subunit 4|DNA polymerase epsilon p12 subunit|DNA polymerase epsilon subunit p12|polymerase (DNA-directed), epsilon 4 (p12 subunit)', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74958648, end_position_on_the_genomic_accession=74969732, orientation='plus', exon_count=4, OMIM=607269, no_hdr0=''),
  56656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56656, CurrentID=0, Status='live', Symbol='OR2S2', Aliases='OR37A, OST715', description='olfactory receptor family 2 subfamily S member 2 (gene/pseudogene)', other_designations='olfactory receptor OR9-3|olfactory receptor, family 2, subfamily S, member 2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35957108, end_position_on_the_genomic_accession=35958154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56658, CurrentID=0, Status='live', Symbol='TRIM39', Aliases='RNF23, TFPB, TRIM39', description='tripartite motif containing 39', other_designations='ring finger protein 23|testis-abundant finger protein|tripartite motif-containing 39|tripartite motif-containing protein 39', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30326844, end_position_on_the_genomic_accession=30343729, orientation='plus', exon_count=10, OMIM=605700, no_hdr0=''),
  56659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56659, CurrentID=0, Status='live', Symbol='KCNK13', Aliases='K2p13.1, THIK-1, THIK1', description='potassium channel, two pore domain subfamily K, member 13', other_designations='K2P13.1 potassium channel|potassium channel, subfamily K, member 13|tandem pore domain halothane-inhibited potassium channel 1|tandem pore domain potassium channel THIK-1', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90061765, end_position_on_the_genomic_accession=90185857, orientation='plus', exon_count=2, OMIM=607367, no_hdr0=''),
  56660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56660, CurrentID=0, Status='live', Symbol='KCNK12', Aliases='K2p12.1, THIK-2, THIK2', description='potassium channel, two pore domain subfamily K, member 12', other_designations='potassium channel, subfamily K, member 12|tandem pore domain halothane-inhibited potassium channel 2|tandem pore domain potassium channel THIK-2', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47516574, end_position_on_the_genomic_accession=47570331, orientation='minus', exon_count=3, OMIM=607366, no_hdr0=''),
  642958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642958, CurrentID=0, Status='live', Symbol='HNRNPA1P55', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 55', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73938274, end_position_on_the_genomic_accession=73939629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56662, CurrentID=0, Status='live', Symbol='VTRNA1-3', Aliases='HVG3, VAULTRC3, VR3, hvg-3', description='vault RNA 1-3', other_designations='vault RNA component 3', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140726159, end_position_on_the_genomic_accession=140726246, orientation='plus', exon_count=1, OMIM=612697, no_hdr0=''),
  56663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56663, CurrentID=0, Status='live', Symbol='VTRNA1-2', Aliases='HVG2, VAULTRC2, VR2, hvg-2', description='vault RNA 1-2', other_designations='vault RNA component 2', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140718926, end_position_on_the_genomic_accession=140719013, orientation='plus', exon_count=1, OMIM=612696, no_hdr0=''),
  56664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56664, CurrentID=0, Status='live', Symbol='VTRNA1-1', Aliases='HVG1, VAULTRC1, VR1, hvg-1, vRNA', description='vault RNA 1-1', other_designations='vault RNA component 1', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140711276, end_position_on_the_genomic_accession=140711373, orientation='plus', exon_count=1, OMIM=612695, no_hdr0=''),
  56666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56666, CurrentID=0, Status='live', Symbol='PANX2', Aliases='PX2, hPANX2', description='pannexin 2', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50170731, end_position_on_the_genomic_accession=50180295, orientation='plus', exon_count=4, OMIM=608421, no_hdr0=''),
  56667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56667, CurrentID=0, Status='live', Symbol='MUC13', Aliases='DRCC1, MUC-13', description='mucin 13, cell surface associated', other_designations='down-regulated in colon cancer 1|mucin 13, epithelial transmembrane', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124905442, end_position_on_the_genomic_accession=124934751, orientation='minus', exon_count=12, OMIM=612181, no_hdr0=''),
  56670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56670, CurrentID=0, Status='live', Symbol='SUCNR1', Aliases='GPR91', description='succinate receptor 1', other_designations='G protein-coupled receptor 91|G-protein coupled receptor 91|P2Y purinoceptor 1', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151873643, end_position_on_the_genomic_accession=151882088, orientation='plus', exon_count=3, OMIM=606381, no_hdr0=''),
  56672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56672, CurrentID=0, Status='live', Symbol='AKIP1', Aliases='BCA3, C11orf17', description='A-kinase interacting protein 1', other_designations='A kinase (PRKA) interacting protein 1|breast cancer associated gene 3|koyt binding protein 1|koyt binding protein 2|koyt binding protein 3|proline-rich protein BCA3', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8911118, end_position_on_the_genomic_accession=8920079, orientation='plus', exon_count=6, OMIM=609191, no_hdr0=''),
  56673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56673, CurrentID=0, Status='live', Symbol='C11orf16', Aliases='', description='chromosome 11 open reading frame 16', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8920076, end_position_on_the_genomic_accession=8933006, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56674, CurrentID=0, Status='live', Symbol='TMEM9B', Aliases='C11orf15', description='TMEM9 domain family member B', other_designations='TMEM9 domain family, member B', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8947201, end_position_on_the_genomic_accession=8965011, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56675, CurrentID=0, Status='live', Symbol='NRIP3', Aliases='C11orf14, NY-SAR-105', description='nuclear receptor interacting protein 3', other_designations='sarcoma antigen NY-SAR-105', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8980576, end_position_on_the_genomic_accession=9004049, orientation='minus', exon_count=7, OMIM=613125, no_hdr0=''),
  56676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56676, CurrentID=0, Status='live', Symbol='ASCL3', Aliases='HASH3, SGN1, bHLHa42', description='achaete-scute family bHLH transcription factor 3', other_designations='achaete-scute complex homolog 3|bHLH transcription factor Sgn-1 (Salivary Glands 1)|bHLH transcriptional regulator Sgn-1|class A basic helix-loop-helix protein 42', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8937572, end_position_on_the_genomic_accession=8943033, orientation='minus', exon_count=2, OMIM=609154, no_hdr0=''),
  56677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56677, CurrentID=0, Status='live', Symbol='FABP3P2', Aliases='FABP3-ps, FABP3P', description='fatty acid binding protein 3, pseudogene 2', other_designations='', map_location='13q13-q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42369114, end_position_on_the_genomic_accession=42369757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370136, CurrentID=0, Status='live', Symbol='LOC105370136', Aliases='', description='uncharacterized LOC105370136', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28659980, end_position_on_the_genomic_accession=28668466, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56681, CurrentID=0, Status='live', Symbol='SAR1A', Aliases='SAR1, SARA1, Sara, masra2', description='secretion associated, Ras related GTPase 1A', other_designations='COPII-associated small GTPase|SAR1 gene homolog A|SAR1 homolog A|SAR1a gene homolog 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70150204, end_position_on_the_genomic_accession=70170529, orientation='minus', exon_count=8, OMIM=607691, no_hdr0=''),
  56683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56683, CurrentID=0, Status='live', Symbol='C21orf59', Aliases='C21orf48, CILD26, FBB18', description='chromosome 21 open reading frame 59', other_designations='prostate cancer upregulated protein 1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32601674, end_position_on_the_genomic_accession=32612608, orientation='minus', exon_count=7, OMIM=615494, no_hdr0=''),
  56694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56694, CurrentID=0, Status='live', Symbol='BMIQ8', Aliases='OB10P', description='body mass index quantitative trait locus 8', other_designations='obesity, susceptibility to, on chromosome 10p', map_location='10p', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603188, no_hdr0=''),
  56704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56704, CurrentID=0, Status='live', Symbol='JPH1', Aliases='JP-1, JP1', description='junctophilin 1', other_designations='junctophilin type 1', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74234700, end_position_on_the_genomic_accession=74321500, orientation='minus', exon_count=6, OMIM=605266, no_hdr0=''),
  105369496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369496, CurrentID=0, Status='live', Symbol='LOC105369496', Aliases='', description='uncharacterized LOC105369496', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112698772, end_position_on_the_genomic_accession=112737683, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369497, CurrentID=0, Status='live', Symbol='LOC105369497', Aliases='', description='uncharacterized LOC105369497', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112781362, end_position_on_the_genomic_accession=112784905, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56729, CurrentID=0, Status='live', Symbol='RETN', Aliases='ADSF, FIZZ31, RSTN, XCP1, RETN', description='resistin', other_designations='C/EBP-epsilon regulated myeloid-specific secreted cysteine-rich protein precursor 1|adipose tissue-specific secretory factor|c/EBP-epsilon-regulated myeloid-specific secreted cysteine-rich protein|cysteine-rich secreted protein A12-alpha-like 2|cysteine-rich secreted protein FIZZ3|found in inflammatory zone 3|resistin delta2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7669086, end_position_on_the_genomic_accession=7670454, orientation='plus', exon_count=4, OMIM=605565, no_hdr0=''),
  56731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56731, CurrentID=0, Status='live', Symbol='SLC2A4RG', Aliases='GEF, HDBP-1, HDBP1, Si-1-2, Si-1-2-19', description='SLC2A4 regulator', other_designations="GLUT4 enhancer factor|Huntington's disease gene regulatory region-binding protein 1", map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63739858, end_position_on_the_genomic_accession=63744050, orientation='plus', exon_count=8, OMIM=609493, no_hdr0=''),
  105369499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369499, CurrentID=0, Status='live', Symbol='LOC105369499', Aliases='', description='uncharacterized LOC105369499', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113276702, end_position_on_the_genomic_accession=113314458, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56751, CurrentID=0, Status='live', Symbol='BARHL1', Aliases='', description='BarH-like homeobox 1', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132582606, end_position_on_the_genomic_accession=132590253, orientation='plus', exon_count=3, OMIM=605211, no_hdr0=''),
  105374503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374503, CurrentID=0, Status='live', Symbol='LOC105374503', Aliases='', description='uncharacterized LOC105374503', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16287658, end_position_on_the_genomic_accession=16355779, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369503, CurrentID=0, Status='live', Symbol='LOC105369503', Aliases='', description='uncharacterized LOC105369503', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114465766, end_position_on_the_genomic_accession=114497068, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369504, CurrentID=0, Status='live', Symbol='LOC105369504', Aliases='', description='uncharacterized LOC105369504', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114639577, end_position_on_the_genomic_accession=114651202, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56776, CurrentID=0, Status='live', Symbol='FMN2', Aliases='', description='formin 2', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240091885, end_position_on_the_genomic_accession=240475189, orientation='plus', exon_count=21, OMIM=606373, no_hdr0=''),
  402679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402679, CurrentID=0, Status='live', Symbol='MARK2P10', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 10', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96858164, end_position_on_the_genomic_accession=96859671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374504, CurrentID=0, Status='live', Symbol='LOC105374504', Aliases='', description='uncharacterized LOC105374504', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16744750, end_position_on_the_genomic_accession=16749331, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642980, CurrentID=0, Status='live', Symbol='LOC642980', Aliases='', description='uncharacterized LOC642980', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=148845967, end_position_on_the_genomic_accession=148847100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56797, CurrentID=0, Status='live', Symbol='HTGS', Aliases='', description='Hypertriglyceridemia, familial', other_designations='', map_location='15q11.2-q13.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145750, no_hdr0=''),
  56801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56801, CurrentID=0, Status='live', Symbol='HDLCQ14', Aliases='HDLD3, HYALP, HYLAP', description='Hypoalphalipoproteinemia, primary', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605201, no_hdr0=''),
  100287541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287541, CurrentID=0, Status='live', Symbol='LOC100287541', Aliases='', description='POM121 transmembrane nucleoporin pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18855105, end_position_on_the_genomic_accession=18861785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369511, CurrentID=0, Status='live', Symbol='LOC105369511', Aliases='', description='uncharacterized LOC105369511', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115518544, end_position_on_the_genomic_accession=115520220, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56829, CurrentID=0, Status='live', Symbol='ZC3HAV1', Aliases='ARTD13, FLB6421, PARP13, ZAP, ZC3H2, ZC3HDC2', description='zinc finger CCCH-type, antiviral 1', other_designations='ADP-ribosyltransferase diphtheria toxin-like 13|zinc finger CCCH domain-containing protein 2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139043520, end_position_on_the_genomic_accession=139109720, orientation='minus', exon_count=13, OMIM=607312, no_hdr0=''),
  56832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56832, CurrentID=0, Status='live', Symbol='IFNK', Aliases='IFNT1, INFE1', description='interferon, kappa', other_designations='IFN-kappa|interferon-like protein', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27524314, end_position_on_the_genomic_accession=27526498, orientation='plus', exon_count=2, OMIM=615326, no_hdr0=''),
  56833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56833, CurrentID=0, Status='live', Symbol='SLAMF8', Aliases='BLAME, CD353, SBBI42', description='SLAM family member 8', other_designations='B lymphocyte activator macrophage expressed|B-lymphocyte activator macrophage expressed|BCM-like membrane protein', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159826650, end_position_on_the_genomic_accession=159837497, orientation='plus', exon_count=6, OMIM=606620, no_hdr0=''),
  56834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56834, CurrentID=0, Status='live', Symbol='GPR137', Aliases='C11orf4A, TM7SF1L1, GPR137', description='G protein-coupled receptor 137', other_designations='transmembrane 7 superfamily member 1-like 1 protein', map_location='11cen-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64269260, end_position_on_the_genomic_accession=64289500, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105369515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369515, CurrentID=0, Status='live', Symbol='LOC105369515', Aliases='', description='uncharacterized LOC105369515', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117659055, end_position_on_the_genomic_accession=117668885, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140545, CurrentID=0, Status='live', Symbol='RNF32', Aliases='FKSG33, HSD15, LMBR2', description='ring finger protein 32', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156640286, end_position_on_the_genomic_accession=156677893, orientation='plus', exon_count=16, OMIM=610241, no_hdr0=''),
  105374506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374506, CurrentID=0, Status='live', Symbol='LOC105374506', Aliases='', description='uncharacterized LOC105374506', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41418928, end_position_on_the_genomic_accession=41760081, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376671, CurrentID=0, Status='live', Symbol='LOC105376671', Aliases='lnc-BBOX1-1', description='uncharacterized LOC105376671', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27472363, end_position_on_the_genomic_accession=27476635, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56848, CurrentID=0, Status='live', Symbol='SPHK2', Aliases='SK 2, SK-2, SPK 2, SPK-2', description='sphingosine kinase 2', other_designations='sphingosine kinase type 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48619291, end_position_on_the_genomic_accession=48630407, orientation='plus', exon_count=10, OMIM=607092, no_hdr0=''),
  56849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56849, CurrentID=0, Status='live', Symbol='TCEAL7', Aliases='MPMGp800C04260Q003, WEX5', description='transcription elongation factor A (SII)-like 7', other_designations='TCEA-like protein 7|transcription elongation factor S-II protein-like 7', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103330186, end_position_on_the_genomic_accession=103332323, orientation='plus', exon_count=3, OMIM=300771, no_hdr0=''),
  56850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56850, CurrentID=0, Status='live', Symbol='GRIPAP1', Aliases='GRASP-1', description='GRIP1 associated protein 1', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48973720, end_position_on_the_genomic_accession=49002264, orientation='minus', exon_count=27, OMIM=300408, no_hdr0=''),
  56851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56851, CurrentID=0, Status='live', Symbol='EMC7', Aliases='C11orf3, C15orf24, HT022, ORF1-FL1', description='ER membrane protein complex subunit 7', other_designations='UPF0480 protein C15orf24|chromosome 15 hypothetical ATG/GTP binding protein', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34084022, end_position_on_the_genomic_accession=34101852, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56852, CurrentID=0, Status='live', Symbol='RAD18', Aliases='RNF73', description='RAD18, E3 ubiquitin protein ligase', other_designations='RAD18 E3 ubiquitin protein ligase|RAD18 homolog|RAD18, S. cerevisiae, homolog|RING finger protein 73|hHR18|hRAD18|postreplication repair protein RAD18|postreplication repair protein hRAD18p', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8877196, end_position_on_the_genomic_accession=8963475, orientation='minus', exon_count=13, OMIM=605256, no_hdr0=''),
  56853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56853, CurrentID=0, Status='live', Symbol='CELF4', Aliases='BRUNOL4, CELF-4', description='CUGBP, Elav-like family member 4', other_designations='CUG-BP and ETR-3 like factor 4|CUG-BP- and ETR-3-like factor 4|LYST-interacting protein LIP9|RNA-binding protein BRUNOL4|bruno-like 4, RNA binding protein|bruno-like protein 4', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37190090, end_position_on_the_genomic_accession=37566037, orientation='minus', exon_count=17, OMIM=612679, no_hdr0=''),
  105378871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378871, CurrentID=0, Status='live', Symbol='LOC105378871', Aliases='', description='uncharacterized LOC105378871', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98709899, end_position_on_the_genomic_accession=98711484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  402694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402694, CurrentID=0, Status='live', Symbol='RPL15P11', Aliases='RPL15_5_851', description='ribosomal protein L15 pseudogene 11', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135660038, end_position_on_the_genomic_accession=135660652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  450086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450086, CurrentID=0, Status='live', Symbol='AD9', Aliases='', description='Alzheimer disease 9', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608907, no_hdr0=''),
  450087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450087, CurrentID=0, Status='live', Symbol='ADHD1', Aliases='', description='Attention deficit-hyperactivity disorder, susceptibility to, 1', other_designations='', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608903, no_hdr0=''),
  450088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450088, CurrentID=0, Status='live', Symbol='ADHD2', Aliases='', description='Attention deficit-hyperactivity disorder, susceptibility to, 2', other_designations='', map_location='17p11', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608904, no_hdr0=''),
  450089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450089, CurrentID=0, Status='live', Symbol='ADHD3', Aliases='LOAS', description='Attention deficit-hyperactivity disorder, susceptibility to, 2', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608905, no_hdr0=''),
  450090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450090, CurrentID=0, Status='live', Symbol='ADHD4', Aliases='', description='Attention deficit-hyperactivity disorder, susceptibility to, 4', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608906, no_hdr0=''),
  450092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450092, CurrentID=0, Status='live', Symbol='KTCN2', Aliases='', description='Keratoconus 2', other_designations='', map_location='16q22.3-q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608932, no_hdr0=''),
  450093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450093, CurrentID=0, Status='live', Symbol='LNCR1', Aliases='', description='Lung cancer 1', other_designations='', map_location='6q23-q25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608935, no_hdr0=''),
  450095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450095, CurrentID=0, Status='live', Symbol='PLF', Aliases='', description='Pulmonary function', other_designations='', map_location='6q21-q22', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608852, no_hdr0=''),
  450096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450096, CurrentID=0, Status='live', Symbol='PAPA4', Aliases='', description='Polydactyly, postaxial, type A4', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608562, no_hdr0=''),
  450097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450097, CurrentID=0, Status='live', Symbol='RLS2', Aliases='', description='Restless legs syndrome 2', other_designations='', map_location='14q13-q21', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608831, no_hdr0=''),
  56882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56882, CurrentID=0, Status='live', Symbol='CDC42SE1', Aliases='SCIP1, SPEC1', description='CDC42 small effector 1', other_designations='1300002M12Rik|CDC42-binding protein SCIP1|signaling molecule SPEC1 beta|small effector of CDC42 protein 1|small protein effector 1 of Cdc42', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151050971, end_position_on_the_genomic_accession=151059649, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369523, CurrentID=0, Status='live', Symbol='LOC105369523', Aliases='', description='uncharacterized LOC105369523', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119523189, end_position_on_the_genomic_accession=119526780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56884, CurrentID=0, Status='live', Symbol='FSTL5', Aliases='', description='follistatin-like 5', other_designations='follistatin-like protein 5', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=161383892, end_position_on_the_genomic_accession=162164034, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56886, CurrentID=0, Status='live', Symbol='UGGT1', Aliases='HUGT1, UGCGL1, UGT1', description='UDP-glucose glycoprotein glucosyltransferase 1', other_designations='UDP--Glc:glycoprotein glucosyltransferase|UDP-glucose ceramide glucosyltransferase-like 1', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128091180, end_position_on_the_genomic_accession=128195677, orientation='plus', exon_count=41, OMIM=605897, no_hdr0=''),
  56888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56888, CurrentID=0, Status='live', Symbol='KCMF1', Aliases='DEBT91, FIGC, PCMF, ZZZ1', description='potassium channel modulatory factor 1', other_designations='FGF-induced in gastric cancer|FGF-induced ubiquitin-protein ligase in gastric cancers|ZZ-type zinc finger-containing protein 1|differentially expressed in branching tubulogenesis 91|zinc finger, ZZ domain containing 1', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84971108, end_position_on_the_genomic_accession=85059472, orientation='plus', exon_count=8, OMIM=614719, no_hdr0=''),
  56889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56889, CurrentID=0, Status='live', Symbol='TM9SF3', Aliases='EP70-P-iso, SMBP', description='transmembrane 9 superfamily member 3', other_designations='SM-11044 binding protein|SM-11044-binding protein|dinucleotide oxidase disulfide thiol exchanger 3 superfamily member 3|endomembrane protein emp70 precursor isolog', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96518110, end_position_on_the_genomic_accession=96587052, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56890, CurrentID=0, Status='live', Symbol='MDM1', Aliases='', description='Mdm1 nuclear protein', other_designations='Mdm1 nuclear protein homolog|Mdm4, transformed 3T3 cell double minute 1, p53 binding protein|nuclear protein double minute 1', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68294566, end_position_on_the_genomic_accession=68332381, orientation='minus', exon_count=16, OMIM=613813, no_hdr0=''),
  56891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56891, CurrentID=0, Status='live', Symbol='LGALS14', Aliases='CLC2, PPL13', description='lectin, galactoside-binding, soluble, 14', other_designations='Charcot-Leyden crystal protein 2|gal-14|galectin-14|placental protein 13-like protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39704306, end_position_on_the_genomic_accession=39709448, orientation='plus', exon_count=5, OMIM=607260, no_hdr0=''),
  56892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56892, CurrentID=0, Status='live', Symbol='C8orf4', Aliases='TC-1, TC1', description='chromosome 8 open reading frame 4', other_designations='human thyroid cancer 1|thyroid cancer protein 1', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40153468, end_position_on_the_genomic_accession=40155308, orientation='plus', exon_count=1, OMIM=607702, no_hdr0=''),
  56893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56893, CurrentID=0, Status='live', Symbol='UBQLN4', Aliases='A1U, A1Up, C1orf6, CIP75, UBIN', description='ubiquilin 4', other_designations='ataxin-1 interacting ubiquitin-like protein|ataxin-1 ubiquitin-like interacting protein|ataxin-1 ubiquitin-like-interacting protein A1U|connexin43-interacting protein of 75 kDa', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156033135, end_position_on_the_genomic_accession=156053825, orientation='minus', exon_count=13, OMIM=605440, no_hdr0=''),
  56894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56894, CurrentID=0, Status='live', Symbol='AGPAT3', Aliases='LPAAT-GAMMA1, LPAAT3', description='1-acylglycerol-3-phosphate O-acyltransferase 3', other_designations='1-AGP acyltransferase 3|1-AGPAT 3|lysophosphatidic acid acyltransferase gamma|lysophosphatidic acid acyltransferase-gamma1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43865160, end_position_on_the_genomic_accession=43987594, orientation='plus', exon_count=18, OMIM=614794, no_hdr0=''),
  56895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56895, CurrentID=0, Status='live', Symbol='AGPAT4', Aliases='1-AGPAT4, LPAAT-delta, dJ473J16.2', description='1-acylglycerol-3-phosphate O-acyltransferase 4', other_designations='1-AGP acyltransferase 4|1-AGPAT 4|1-acylglycerol-3-phosphate O-acyltransferase 4 (lysophosphatidic acid acyltransferase, delta)|lysophosphatidic acid acyltransferase delta|lysophosphatidic acid acyltransferase-delta (LPAAT-delta)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=161130020, end_position_on_the_genomic_accession=161274075, orientation='minus', exon_count=10, OMIM=614795, no_hdr0=''),
  56896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56896, CurrentID=0, Status='live', Symbol='DPYSL5', Aliases='CRAM, CRMP-5, CRMP5, Ulip6', description='dihydropyrimidinase like 5', other_designations='CRMP3-associated molecule|DRP-5|ULIP-6|UNC33-like phosphoprotein 6|collapsin response mediator protein 5|collapsin response mediator protein-5|dihydropyrimidinase-like 5', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26848101, end_position_on_the_genomic_accession=26950351, orientation='plus', exon_count=15, OMIM=608383, no_hdr0=''),
  56897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56897, CurrentID=0, Status='live', Symbol='WRNIP1', Aliases='WHIP, bA420G6.2', description='Werner helicase interacting protein 1', other_designations='Werner helicase-interacting protein 1|putative helicase RUVBL', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2765341, end_position_on_the_genomic_accession=2785745, orientation='plus', exon_count=8, OMIM=608196, no_hdr0=''),
  56898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56898, CurrentID=0, Status='live', Symbol='BDH2', Aliases='DHRS6, EFA6R, PRO20933, SDR15C1, UCPA-OR, UNQ6308', description='3-hydroxybutyrate dehydrogenase, type 2', other_designations='R-beta-hydroxybutyrate dehydrogenase|dehydrogenase/reductase (SDR family) member 6|dehydrogenase/reductase SDR family member 6|oxidoreductase UCPA|short chain dehydrogenase/reductase family 15C member 1|short chain dehydrogenase/reductase family 15C, member 1', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103077625, end_position_on_the_genomic_accession=103099867, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56899, CurrentID=0, Status='live', Symbol='ANKS1B', Aliases='AIDA, AIDA-1, ANKS2, EB-1, EB1, cajalin-2', description='ankyrin repeat and sterile alpha motif domain containing 1B', other_designations='E2a-Pbx1-associated protein|amyloid-beta precursor protein intracellular domain associated protein 1|cajalin 2', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98729904, end_position_on_the_genomic_accession=99984727, orientation='minus', exon_count=34, OMIM=607815, no_hdr0=''),
  56900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56900, CurrentID=0, Status='live', Symbol='TMEM167B', Aliases='AD-020, C1orf119', description='transmembrane protein 167B', other_designations='protein x 013', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109090762, end_position_on_the_genomic_accession=109096934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56901, CurrentID=0, Status='live', Symbol='NDUFA4L2', Aliases='NUOMS', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4-like 2', other_designations='NADH-ubiquinone oxidoreductase MLRQ subunit homolog|NADH:ubiquinone oxidoreductase MLRQ subunit homolog', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57234903, end_position_on_the_genomic_accession=57240762, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56902, CurrentID=0, Status='live', Symbol='PNO1', Aliases='KHRBP1, RRP20', description='partner of NOB1 homolog', other_designations='KH-type RNA-binding protein 1|RNA binding protein', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68157873, end_position_on_the_genomic_accession=68175959, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56903, CurrentID=0, Status='live', Symbol='PAPOLB', Aliases='PAPT, TPAP', description='poly(A) polymerase beta', other_designations='PAP-beta|poly(A) polymerase beta (testis specific)|polynucleotide adenylyltransferase beta|testis-specific poly(A) polymerase', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4857738, end_position_on_the_genomic_accession=4861994, orientation='minus', exon_count=1, OMIM=607436, no_hdr0=''),
  56904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56904, CurrentID=0, Status='live', Symbol='SH3GLB2', Aliases='PP6569, PP9455, RRIG1', description='SH3-domain GRB2-like endophilin B2', other_designations='SH3 domain-containing GRB2-like protein B2|SH3-containing protein SH3GLB2|retinoid receptor-induced gene 1', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129007792, end_position_on_the_genomic_accession=129028353, orientation='minus', exon_count=13, OMIM=609288, no_hdr0=''),
  56905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56905, CurrentID=0, Status='live', Symbol='C15orf39', Aliases='', description='chromosome 15 open reading frame 39', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75198887, end_position_on_the_genomic_accession=75212169, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56906, CurrentID=0, Status='live', Symbol='THAP10', Aliases='', description='THAP domain containing 10', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70881342, end_position_on_the_genomic_accession=70892433, orientation='minus', exon_count=3, OMIM=612538, no_hdr0=''),
  56907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56907, CurrentID=0, Status='live', Symbol='SPIRE1', Aliases='Spir-1', description='spire-type actin nucleation factor 1', other_designations='spire actin nucleation factor 1|spire family actin nucleation factor 1|spire homolog 1', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12446512, end_position_on_the_genomic_accession=12658102, orientation='minus', exon_count=21, OMIM=609216, no_hdr0=''),
  56910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56910, CurrentID=0, Status='live', Symbol='STARD7', Aliases='GTT1', description='StAR related lipid transfer domain containing 7', other_designations='START domain containing 7|START domain-containing protein 7|StAR-related lipid transfer (START) domain containing 7|StAR-related lipid transfer domain containing 7|gestational trophoblastic tumor protein 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96184859, end_position_on_the_genomic_accession=96208846, orientation='minus', exon_count=8, OMIM=616712, no_hdr0=''),
  56911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56911, CurrentID=0, Status='live', Symbol='MAP3K7CL', Aliases='C21orf7, HC21ORF7, TAK1L, TAKL, TAKL-1, TAKL-2, TAKL-4', description='MAP3K7 C-terminal like', other_designations='TAK1-like protein 1|TAK1-like protein 2|TAK1-like protein 4|TGF-beta activated kinase', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29077471, end_position_on_the_genomic_accession=29175889, orientation='plus', exon_count=14, OMIM=611110, no_hdr0=''),
  56912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56912, CurrentID=0, Status='live', Symbol='IFT46', Aliases='C11orf2, C11orf60, CFAP32', description='intraflagellar transport 46', other_designations='UPF0360|cilia and flagella associated protein 32|intraflagellar transport 46 homolog|intraflagellar transport protein IFT46', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118544528, end_position_on_the_genomic_accession=118572965, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56913, CurrentID=0, Status='live', Symbol='C1GALT1', Aliases='C1GALT, T-synthase', description='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase 1', other_designations='B3Gal-T8|core 1 O-glycan T-synthase|core 1 UDP-galactose:N-acetylgalactosamine-alpha-R beta 1,3-galactosyltransferase 1|core 1 beta3-Gal-T1|core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase, 1', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7182547, end_position_on_the_genomic_accession=7248651, orientation='plus', exon_count=6, OMIM=610555, no_hdr0=''),
  56914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56914, CurrentID=0, Status='live', Symbol='OTOR', Aliases='FDP, MIAL1', description='otoraplin', other_designations='fibrocyte-derived protein|melanoma inhibitory activity-like protein', map_location='20p12.1-p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16748353, end_position_on_the_genomic_accession=16752164, orientation='plus', exon_count=5, OMIM=606067, no_hdr0=''),
  56915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56915, CurrentID=0, Status='live', Symbol='EXOSC5', Aliases='RRP41B, RRP46, Rrp46p, hRrp46p, p12B', description='exosome component 5', other_designations='chronic myelogenous leukemia tumor antigen 28|exosome complex exonuclease RRP46|exosome component Rrp46|ribosomal RNA-processing protein 46', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41386370, end_position_on_the_genomic_accession=41397351, orientation='minus', exon_count=6, OMIM=606492, no_hdr0=''),
  56916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56916, CurrentID=0, Status='live', Symbol='SMARCAD1', Aliases='ADERM, ETL1, HEL1', description='SWI/SNF-related, matrix-associated actin-dependent regulator of chromatin, subfamily a, containing DEAD/H box 1', other_designations='ATP-dependent helicase 1', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94207608, end_position_on_the_genomic_accession=94291292, orientation='plus', exon_count=30, OMIM=612761, no_hdr0=''),
  56917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56917, CurrentID=0, Status='live', Symbol='MEIS3', Aliases='MRG2', description='Meis homeobox 3', other_designations='Meis1, myeloid ecotropic viral integration site 1 homolog 3|meis1-related protein 2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47403118, end_position_on_the_genomic_accession=47422265, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56918, CurrentID=0, Status='live', Symbol='C2orf83', Aliases='', description='chromosome 2 open reading frame 83', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227610090, end_position_on_the_genomic_accession=227633320, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56919, CurrentID=0, Status='live', Symbol='DHX33', Aliases='DDX33', description='DEAH (Asp-Glu-Ala-His) box polypeptide 33', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 33|DEAH box protein 33', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5440912, end_position_on_the_genomic_accession=5469060, orientation='minus', exon_count=12, OMIM=614405, no_hdr0=''),
  56920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56920, CurrentID=0, Status='live', Symbol='SEMA3G', Aliases='sem2', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3G', other_designations='semaphorin sem2', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52433252, end_position_on_the_genomic_accession=52445027, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56922, CurrentID=0, Status='live', Symbol='MCCC1', Aliases='MCC-B, MCCA', description='methylcrotonoyl-CoA carboxylase 1', other_designations='3-methylcrotonyl-CoA carboxylase 1|3-methylcrotonyl-CoA carboxylase biotin-containing subunit|3-methylcrotonyl-CoA:carbon dioxide ligase subunit alpha|MCCase subunit alpha|methylcrotonoyl-CoA carboxylase 1 (alpha)|methylcrotonoyl-CoA carboxylase alpha|methylcrotonoyl-Coenzyme A carboxylase 1 (alpha)', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183015218, end_position_on_the_genomic_accession=183099587, orientation='minus', exon_count=21, OMIM=609010, no_hdr0=''),
  56923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56923, CurrentID=0, Status='live', Symbol='NMUR2', Aliases='FM-4, FM4, NMU-R2, NMU2R, TGR-1, TGR1', description='neuromedin U receptor 2', other_designations='G-protein coupled receptor FM-4|G-protein coupled receptor TGR-1|growth hormone secretagogue receptor family, member 4', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152391541, end_position_on_the_genomic_accession=152405281, orientation='minus', exon_count=4, OMIM=605108, no_hdr0=''),
  56924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56924, CurrentID=0, Status='live', Symbol='PAK6', Aliases='PAK5', description='p21 protein (Cdc42/Rac)-activated kinase 6', other_designations='p21(CDKN1A)-activated kinase 6', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40217428, end_position_on_the_genomic_accession=40277487, orientation='plus', exon_count=11, OMIM=608110, no_hdr0=''),
  56925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56925, CurrentID=0, Status='live', Symbol='LXN', Aliases='ECI, TCI', description='latexin', other_designations='MUM|endogenous carboxypeptidase inhibitor|tissue carboxypeptidase inhibitor', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158666414, end_position_on_the_genomic_accession=158672693, orientation='minus', exon_count=6, OMIM=609305, no_hdr0=''),
  56926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56926, CurrentID=0, Status='live', Symbol='NCLN', Aliases='NET59', description='nicalin', other_designations='nicalin homolog (zebrefish)|nicastrin-like protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3185877, end_position_on_the_genomic_accession=3209575, orientation='plus', exon_count=15, OMIM=609156, no_hdr0=''),
  56927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56927, CurrentID=0, Status='live', Symbol='GPR108', Aliases='LUSTR2', description='G protein-coupled receptor 108', other_designations='lung seven transmembrane receptor 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6729914, end_position_on_the_genomic_accession=6737622, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  56928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56928, CurrentID=0, Status='live', Symbol='SPPL2B', Aliases='IMP-4, IMP4, PSH4, PSL1', description='signal peptide peptidase like 2B', other_designations='SPP-like 2B|intramembrane protease 4|presenilin homologous protein 4|presenilin-like protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2328614, end_position_on_the_genomic_accession=2355102, orientation='plus', exon_count=15, OMIM=608239, no_hdr0=''),
  56929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56929, CurrentID=0, Status='live', Symbol='FEM1C', Aliases='EUROIMAGE686608, EUROIMAGE783647, FEM1A', description='fem-1 homolog c (C. elegans)', other_designations='FEM1-gamma', map_location='5q22', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115520908, end_position_on_the_genomic_accession=115544894, orientation='minus', exon_count=5, OMIM=608767, no_hdr0=''),
  56931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56931, CurrentID=0, Status='live', Symbol='DUS3L', Aliases='DUS3', description='dihydrouridine synthase 3-like', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5785140, end_position_on_the_genomic_accession=5791238, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105374509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374509, CurrentID=0, Status='live', Symbol='LOC105374509', Aliases='', description='uncharacterized LOC105374509', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17793446, end_position_on_the_genomic_accession=17798349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56934, CurrentID=0, Status='live', Symbol='CA10', Aliases='CA-RPX, CARPX, HUCEP-15', description='carbonic anhydrase X', other_designations='CA-RP X|CARP X|carbonic anhydrase-related protein X|cerebral protein 15|cerebral protein-15', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51630313, end_position_on_the_genomic_accession=52160017, orientation='minus', exon_count=13, OMIM=604642, no_hdr0=''),
  56935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56935, CurrentID=0, Status='live', Symbol='SMCO4', Aliases='C11orf75, FN5', description='single-pass membrane protein with coiled-coil domains 4', other_designations='UPF0443 protein C11orf75', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93478472, end_position_on_the_genomic_accession=93586855, orientation='minus', exon_count=8, OMIM=609477, no_hdr0=''),
  56936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56936, CurrentID=0, Status='live', Symbol='CCDC177', Aliases='C14orf162, PLPL', description='coiled-coil domain containing 177', other_designations='myelin proteolipid protein-like protein', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69569814, end_position_on_the_genomic_accession=69574883, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56937, CurrentID=0, Status='live', Symbol='PMEPA1', Aliases='STAG1, TMEPAI', description='prostate transmembrane protein, androgen induced 1', other_designations='prostate transmembrane protein androgen induced 1|solid tumor-associated 1 protein|transmembrane, prostate androgen induced RNA', map_location='20q13.31-q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57648392, end_position_on_the_genomic_accession=57711536, orientation='minus', exon_count=7, OMIM=606564, no_hdr0=''),
  56938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56938, CurrentID=0, Status='live', Symbol='ARNTL2', Aliases='BMAL2, CLIF, MOP9, PASD9, bHLHe6', description='aryl hydrocarbon receptor nuclear translocator like 2', other_designations='CYCLE-like factor|PAS domain-containing protein 9|aryl hydrocarbon receptor nuclear translocator-like 2|basic-helix-loop-helix-PAS protein MOP9|brain and muscle ARNT-like 2|class E basic helix-loop-helix protein 6|member of PAS protein 9|transcription factor BMAL2', map_location='12p12.2-p11.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27332854, end_position_on_the_genomic_accession=27425813, orientation='plus', exon_count=21, OMIM=614517, no_hdr0=''),
  56940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56940, CurrentID=0, Status='live', Symbol='DUSP22', Aliases='JKAP, JSP-1, JSP1, LMW-DSP2, LMWDSP2, MKP-x, MKPX, VHX', description='dual specificity phosphatase 22', other_designations='JNK-stimulating phosphatase 1|JNK-stimulatory phosphatase-1|MAP kinase phosphatase x|homolog of mouse dual specificity phosphatase LMW-DSP2|low molecular weight dual specificity phosphatase 2|mitogen-activated protein kinase phosphatase x', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=292057, end_position_on_the_genomic_accession=351355, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56941, CurrentID=0, Status='live', Symbol='HMCES', Aliases='C3orf37, DC12, SRAPD1', description='5-hydroxymethylcytosine (hmC) binding, ES cell-specific', other_designations='ES cell-specific 5hmC-binding protein|SRAP domain-containing protein 1|UPF0361 protein C3orf37|putative peptidase SRAPD1', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129278816, end_position_on_the_genomic_accession=129306186, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56942, CurrentID=0, Status='live', Symbol='CMC2', Aliases='2310061C15Rik, C16orf61, DC13', description='C-x(9)-C motif containing 2', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80975802, end_position_on_the_genomic_accession=81006897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56943, CurrentID=0, Status='live', Symbol='ENY2', Aliases='DC6, e(y)2', description='enhancer of yellow 2 homolog (Drosophila)', other_designations='enhancer of yellow 2 transcription factor homolog', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109334323, end_position_on_the_genomic_accession=109345960, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56944, CurrentID=0, Status='live', Symbol='OLFML3', Aliases='HNOEL-iso, OLF44', description='olfactomedin like 3', other_designations='olfactomedin-like 3', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113979391, end_position_on_the_genomic_accession=113982254, orientation='plus', exon_count=4, OMIM=610088, no_hdr0=''),
  319089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319089, CurrentID=0, Status='live', Symbol='TTC6', Aliases='C14orf25, NCRNA00291', description='tetratricopeptide repeat domain 6', other_designations='TPR repeat protein 6', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=37595199, end_position_on_the_genomic_accession=37842628, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  56946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56946, CurrentID=0, Status='live', Symbol='EMSY', Aliases='C11orf30, GL002', description='EMSY, BRCA2-interacting transcriptional repressor', other_designations='EMSY BRCA2-interacting transcriptional repressor', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76444928, end_position_on_the_genomic_accession=76552899, orientation='plus', exon_count=23, OMIM=608574, no_hdr0=''),
  56947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56947, CurrentID=0, Status='live', Symbol='MFF', Aliases='C2orf33, GL004', description='mitochondrial fission factor', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227325151, end_position_on_the_genomic_accession=227357836, orientation='plus', exon_count=13, OMIM=614785, no_hdr0=''),
  56948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56948, CurrentID=0, Status='live', Symbol='SDR39U1', Aliases='C14orf124, HCDI', description='short chain dehydrogenase/reductase family 39U, member 1', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24439766, end_position_on_the_genomic_accession=24442905, orientation='minus', exon_count=7, OMIM=616162, no_hdr0=''),
  56949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56949, CurrentID=0, Status='live', Symbol='XAB2', Aliases='HCNP, HCRN, NTC90, SYF1', description='XPA binding protein 2', other_designations='SYF1 homolog, RNA splicing factor|SYF1 pre-mRNA-splicing factor|XPA-binding protein 2|crn-related protein kim1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7619525, end_position_on_the_genomic_accession=7629553, orientation='minus', exon_count=19, OMIM=610850, no_hdr0=''),
  56950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56950, CurrentID=0, Status='live', Symbol='SMYD2', Aliases='HSKM-B, KMT3C, ZMYND14', description='SET and MYND domain containing 2', other_designations='SET and MYND domain-containing protein 2|histone methyltransferase SMYD2|lysine N-methyltransferase 3C|zinc finger, MYND domain containing 14', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214281222, end_position_on_the_genomic_accession=214337134, orientation='plus', exon_count=12, OMIM=610663, no_hdr0=''),
  56951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56951, CurrentID=0, Status='live', Symbol='C5orf15', Aliases='HTGN29, KCT2', description='chromosome 5 open reading frame 15', other_designations='keratinocytes associated transmembrane protein 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133955507, end_position_on_the_genomic_accession=133968715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56952, CurrentID=0, Status='live', Symbol='PRTFDC1', Aliases='HHGP', description='phosphoribosyl transferase domain containing 1', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24812957, end_position_on_the_genomic_accession=24953751, orientation='minus', exon_count=14, OMIM=610751, no_hdr0=''),
  56953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56953, CurrentID=0, Status='live', Symbol='NT5M', Aliases='dNT-2, dNT2, mdN', description="5',3'-nucleotidase, mitochondrial", other_designations="5' nucleotidase, mitochondrial|5(3)-deoxyribonucleotidase|deoxy-5'-nucleotidase 2|mitochondrial 5' nucleotidase", map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17303366, end_position_on_the_genomic_accession=17347663, orientation='plus', exon_count=10, OMIM=605292, no_hdr0=''),
  319098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319098, CurrentID=0, Status='live', Symbol='COX7A2P1', Aliases='COX7A3P, COX7AP1', description='cytochrome c oxidase subunit VIIa polypeptide 2 (liver) pseudogene 1', other_designations='cytochrome c oxidase subunit VIIa polypeptide 3 pseudogene|cytochrome c oxidase subunit VIIa pseudogene 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67652229, end_position_on_the_genomic_accession=67652629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56955, CurrentID=0, Status='live', Symbol='MEPE', Aliases='OF45', description='matrix extracellular phosphoglycoprotein', other_designations='matrix, extracellular phosphoglycoprotein with ASARM motif (bone)|osteoblast/osteocyte factor 45', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87821398, end_position_on_the_genomic_accession=87846817, orientation='plus', exon_count=7, OMIM=605912, no_hdr0=''),
  56956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56956, CurrentID=0, Status='live', Symbol='LHX9', Aliases='', description='LIM homeobox 9', other_designations='LIM homeobox protein 9', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197912505, end_position_on_the_genomic_accession=197935476, orientation='plus', exon_count=8, OMIM=606066, no_hdr0=''),
  319101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319101, CurrentID=0, Status='live', Symbol='KRT73', Aliases='IRT6IRS3, K6IRS3, KRT6IRS3', description='keratin 73, type II', other_designations='CK-73|K73|cytokeratin-73|keratin 6 irs3|keratin-73|type II inner root sheath-specific keratin-K6irs3|type-II keratin Kb36', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52602883, end_position_on_the_genomic_accession=52620116, orientation='minus', exon_count=10, OMIM=608247, no_hdr0=''),
  319102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319102, CurrentID=0, Status='live', Symbol='ARHGAP16P', Aliases='ARHGAP15P', description='Rho GTPase activating protein 16 pseudogene', other_designations='Rho GTPase activating protein 15 pseudogene', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43595455, end_position_on_the_genomic_accession=43596235, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319103, CurrentID=0, Status='live', Symbol='SNORD8', Aliases='RNU6C, mgU6-53', description='small nucleolar RNA, C/D box 8', other_designations='RNA, U6C small nucleolar', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21397293, end_position_on_the_genomic_accession=21397401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56961, CurrentID=0, Status='live', Symbol='SHD', Aliases='', description='Src homology 2 domain containing transforming protein D', other_designations='src homology 2 domain-containing transforming protein D', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4278601, end_position_on_the_genomic_accession=4290723, orientation='plus', exon_count=6, OMIM=610481, no_hdr0=''),
  56963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56963, CurrentID=0, Status='live', Symbol='RGMA', Aliases='RGM', description='repulsive guidance molecule family member a', other_designations='RGM domain family, member A', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93043406, end_position_on_the_genomic_accession=93089214, orientation='minus', exon_count=7, OMIM=607362, no_hdr0=''),
  56964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56964, CurrentID=0, Status='live', Symbol='WDR93', Aliases='', description='WD repeat domain 93', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89690085, end_position_on_the_genomic_accession=89743790, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  56965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56965, CurrentID=0, Status='live', Symbol='PARP6', Aliases='ARTD17, PARP-6-B1, PARP-6-C, pART17', description='poly(ADP-ribose) polymerase family member 6', other_designations='ADP-ribosyltransferase diphtheria toxin-like 17|PARP-6|poly (ADP-ribose) polymerase family, member 6', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72241181, end_position_on_the_genomic_accession=72272583, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  56967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56967, CurrentID=0, Status='live', Symbol='C14orf132', Aliases='C14orf88', description='chromosome 14 open reading frame 132', other_designations='putative uncharacterized protein C14orf132', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96039324, end_position_on_the_genomic_accession=96093971, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319112, CurrentID=0, Status='live', Symbol='EEF1A1P2', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 2', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81661433, end_position_on_the_genomic_accession=81663290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56969, CurrentID=0, Status='live', Symbol='RPL23AP32', Aliases='RPL23A_9_220', description='ribosomal protein L23a pseudogene 32', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54529222, end_position_on_the_genomic_accession=54529841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56970, CurrentID=0, Status='live', Symbol='ATXN7L3', Aliases='', description='ataxin 7-like 3', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44191805, end_position_on_the_genomic_accession=44199701, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56971, CurrentID=0, Status='live', Symbol='CEACAM19', Aliases='CEAL1', description='carcinoembryonic antigen-related cell adhesion molecule 19', other_designations='carcinoembryonic antigen-like 1', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44666592, end_position_on_the_genomic_accession=44684359, orientation='plus', exon_count=9, OMIM=606691, no_hdr0=''),
  100873256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873256, CurrentID=0, Status='live', Symbol='MTND4P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 10', other_designations='MT-ND4 pseudogene 10', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235540045, end_position_on_the_genomic_accession=235541416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319117, CurrentID=0, Status='live', Symbol='EIF3LP1', Aliases='EIF3EIPP1, EIF3S6IPP', description='eukaryotic translation initiation factor 3 subunit L pseudogene 1', other_designations='eukaryotic translation initiation factor 3, subunit 6 interacting protein pseudogene|eukaryotic translation initiation factor 3, subunit E interacting protein pseudogene 1|eukaryotic translation initiation factor 3, subunit L pseudogene 1', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81916200, end_position_on_the_genomic_accession=81918058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  319118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319118, CurrentID=0, Status='live', Symbol='EIF4BP1', Aliases='EIF4BP', description='eukaryotic translation initiation factor 4B pseudogene 1', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44077336, end_position_on_the_genomic_accession=44077882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56975, CurrentID=0, Status='live', Symbol='FAM20C', Aliases='DMP-4, DMP4, GEF-CK, RNS', description='family with sequence similarity 20 member C', other_designations='Golgi casein kinase|dentin matrix protein 4|family with sequence similarity 20, member C|golgi-enriched fraction casein kinase', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=192777, end_position_on_the_genomic_accession=260774, orientation='plus', exon_count=11, OMIM=611061, no_hdr0=''),
  319120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319120, CurrentID=0, Status='live', Symbol='YWHAZP1', Aliases='YWHAZP', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 1', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta peudogene|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44290659, end_position_on_the_genomic_accession=44291831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56977, CurrentID=0, Status='live', Symbol='STOX2', Aliases='', description='storkhead box 2', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183796331, end_position_on_the_genomic_accession=184023526, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56978, CurrentID=0, Status='live', Symbol='PRDM8', Aliases='PFM5', description='PR domain containing 8', other_designations='PR domain-containing protein 8|PR-domain containing protein 8', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80183741, end_position_on_the_genomic_accession=80204329, orientation='plus', exon_count=13, OMIM=616639, no_hdr0=''),
  56979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56979, CurrentID=0, Status='live', Symbol='PRDM9', Aliases='MEISETZ, MSBP3, PFM6, ZNF899', description='PR domain containing 9', other_designations='PR domain zinc finger protein 9|minisatellite binding protein 3 (115kD)', map_location='5p14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23507615, end_position_on_the_genomic_accession=23528597, orientation='plus', exon_count=11, OMIM=609760, no_hdr0=''),
  56980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56980, CurrentID=0, Status='live', Symbol='PRDM10', Aliases='PFM7', description='PR domain containing 10', other_designations='PR domain-containing protein 10|PR-domain family member 7|PRDM zinc finger transcription factor|tristanin', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129899706, end_position_on_the_genomic_accession=130003050, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  56981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56981, CurrentID=0, Status='live', Symbol='PRDM11', Aliases='PFM8', description='PR domain containing 11', other_designations='PR-domain containing protein 11', map_location='11p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45094013, end_position_on_the_genomic_accession=45235124, orientation='plus', exon_count=16, OMIM=616347, no_hdr0=''),
  56983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56983, CurrentID=0, Status='live', Symbol='POGLUT1', Aliases='C3orf9, CLP46, KDELCL1, KTELC1, MDS010, MDSRP, Rumi, hCLP46', description='protein O-glucosyltransferase 1', other_designations='9630046K23Rik|CAP10-like 46 kDa protein|KDELC family like 1|KTEL (Lys-Tyr-Glu-Leu) containing 1|KTEL motif-containing protein 1|O-glucosyltransferase Rumi homolog|hRumi|myelodysplastic syndromes relative protein|protein O-xylosyltransferase|x 010 protein', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119468938, end_position_on_the_genomic_accession=119494708, orientation='plus', exon_count=14, OMIM=615618, no_hdr0=''),
  56984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56984, CurrentID=0, Status='live', Symbol='PSMG2', Aliases='CLAST3, HCCA3, HsT1707, MDS003, PAC2, TNFSF5IP1', description='proteasome (prosome, macropain) assembly chaperone 2', other_designations='CD40 ligand-activated specific transcript 3|HDCMC29P|HSPC260|PAC-2|hepatocellular carcinoma susceptibility protein|hepatocellular carcinoma-susceptibility protein 3|likely ortholog of mouse CD40 ligand-activated specific transcript 3 (Clast3)|proteasome assembling chaperone 2|tumor necrosis factor superfamily member 5-induced protein 1|tumor necrosis factor superfamily, member 5-induced protein 1|x 003 protein', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12658739, end_position_on_the_genomic_accession=12725740, orientation='plus', exon_count=8, OMIM=609702, no_hdr0=''),
  56985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56985, CurrentID=0, Status='live', Symbol='ADPRM', Aliases='C17orf48, MDS006, NBLA03831', description='ADP-ribose/CDP-alcohol diphosphatase, manganese-dependent', other_designations='ADP-ribose/CDP-alcohol pyrophosphatase|ADPRibase-Mn|CDP-choline phosphohydrolase', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10697610, end_position_on_the_genomic_accession=10711558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56986, CurrentID=0, Status='live', Symbol='DTWD1', Aliases='MDS009', description='DTW domain containing 1', other_designations='x 009 protein', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49620980, end_position_on_the_genomic_accession=49654448, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56987, CurrentID=0, Status='live', Symbol='BBX', Aliases='ARTC1, HBP2, HSPC339, MDS001', description='bobby sox homolog (Drosophila)', other_designations='Ag recognized by Treg cells 1|HMG box-containing protein 2|x 001 protein', map_location='3q13.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107522936, end_position_on_the_genomic_accession=107811329, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  319132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319132, CurrentID=0, Status='live', Symbol='HMGN2P1', Aliases='HMGN2L1, HMGN2P', description='high mobility group nucleosomal binding domain 2 pseudogene 1', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 1|high-mobility group nucleosomal binding domain 2-like 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23263101, end_position_on_the_genomic_accession=23263791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319133, CurrentID=0, Status='live', Symbol='HNRNPCP1', Aliases='HNRPCP', description='heterogeneous nuclear ribonucleoprotein C pseudogene 1', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58520934, end_position_on_the_genomic_accession=58522746, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56990, CurrentID=0, Status='live', Symbol='CDC42SE2', Aliases='SPEC2', description='CDC42 small effector 2', other_designations='non-kinase Cdc42 effector protein SPEC2|small effector of CDC42 protein 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131264009, end_position_on_the_genomic_accession=131394690, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  319135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319135, CurrentID=0, Status='live', Symbol='RPL10AP1', Aliases='RPL10A_6_1413', description='ribosomal protein L10a pseudogene 1', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103412087, end_position_on_the_genomic_accession=103412792, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56992, CurrentID=0, Status='live', Symbol='KIF15', Aliases='HKLP2, KNSL7, NY-BR-62', description='kinesin family member 15', other_designations='kinesin-like 7|kinesin-like protein 2|kinesin-like protein 7|serologically defined breast cancer antigen NY-BR-62', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44761714, end_position_on_the_genomic_accession=44868656, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  56993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56993, CurrentID=0, Status='live', Symbol='TOMM22', Aliases='1C9-2, MST065, MSTP065, TOM22', description='translocase of outer mitochondrial membrane 22 homolog (yeast)', other_designations='mitochondrial import receptor Tom22|translocase of outer membrane 22 kDa subunit homolog', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38681949, end_position_on_the_genomic_accession=38684760, orientation='plus', exon_count=4, OMIM=607046, no_hdr0=''),
  56994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56994, CurrentID=0, Status='live', Symbol='CHPT1', Aliases='CPT, CPT1', description='choline phosphotransferase 1', other_designations='AAPT1-like protein|cholinephosphotransferase 1 alpha|diacylglycerol cholinephosphotransferase 1|hCPT1|phosphatidylcholine synthesizing enzyme', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101697639, end_position_on_the_genomic_accession=101729074, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56995, CurrentID=0, Status='live', Symbol='TULP4', Aliases='TUSP', description='tubby like protein 4', other_designations='tubby super-family protein|tubby superfamily protein|tubby-like protein 4', map_location='6q25-q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158232236, end_position_on_the_genomic_accession=158511828, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56996, CurrentID=0, Status='live', Symbol='SLC12A9', Aliases='CCC6, CIP1, WO3.3, hCCC6', description='solute carrier family 12, member 9', other_designations='CCC-interacting protein 1|cation-chloride cotransporter 6|cation-chloride cotransporter-interacting protein 1|potassium-chloride transporter 9|solute carrier family 12 (potassium/chloride transporters), member 9', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100834036, end_position_on_the_genomic_accession=100867012, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56997, CurrentID=0, Status='live', Symbol='ADCK3', Aliases='ARCA2, CABC1, COQ10D4, COQ8, SCAR9', description='aarF domain containing kinase 3', other_designations='aarF domain-containing protein kinase 3|chaperone activity of bc1 complex-like, mitochondrial|chaperone, ABC1 activity of bc1 complex homolog|coenzyme Q8 homolog', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226940237, end_position_on_the_genomic_accession=226987545, orientation='plus', exon_count=19, OMIM=606980, no_hdr0=''),
  56998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56998, CurrentID=0, Status='live', Symbol='CTNNBIP1', Aliases='ICAT', description='catenin beta interacting protein 1', other_designations='beta-catenin-interacting protein ICAT|catenin, beta interacting protein 1|inhibitor of beta-catenin and Tcf-4', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9848276, end_position_on_the_genomic_accession=9910258, orientation='minus', exon_count=6, OMIM=607758, no_hdr0=''),
  56999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56999, CurrentID=0, Status='live', Symbol='ADAMTS9', Aliases='', description='ADAM metallopeptidase with thrombospondin type 1 motif 9', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 9|ADAM-TS 9|ADAM-TS9|ADAMTS-9|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 9', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64515654, end_position_on_the_genomic_accession=64688000, orientation='minus', exon_count=41, OMIM=605421, no_hdr0=''),
  57000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57000, CurrentID=0, Status='live', Symbol='GSN-AS1', Aliases='C9orf31, MOST2', description='GSN antisense RNA 1', other_designations='GSN antisense RNA 1 (non-protein coding)', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121280768, end_position_on_the_genomic_accession=121285530, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57001, CurrentID=0, Status='live', Symbol='SDHAF3', Aliases='ACN9, DC11, Sdh7', description='succinate dehydrogenase complex assembly factor 3', other_designations='ACN9 homolog|SDH assembly factor 3|protein ACN9 homolog, mitochondrial', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97116593, end_position_on_the_genomic_accession=97181763, orientation='plus', exon_count=4, OMIM=615773, no_hdr0=''),
  57002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57002, CurrentID=0, Status='live', Symbol='YAE1D1', Aliases='C7orf36, GK003', description='Yae1 domain containing 1', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39566376, end_position_on_the_genomic_accession=39612089, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57003, CurrentID=0, Status='live', Symbol='CCDC47', Aliases='GK001, MSTP041', description='coiled-coil domain containing 47', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63745250, end_position_on_the_genomic_accession=63773728, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  57004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57004, CurrentID=0, Status='live', Symbol='UNGP1', Aliases='', description='uracil-DNA glycosylase pseudogene 1', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51278674, end_position_on_the_genomic_accession=51279705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  57005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57005, CurrentID=0, Status='live', Symbol='UNGP2', Aliases='', description='uracil-DNA glycosylase pseudogene 2', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26925434, end_position_on_the_genomic_accession=26926060, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57007, CurrentID=0, Status='live', Symbol='ACKR3', Aliases='CMKOR1, CXC-R7, CXCR-7, CXCR7, GPR159, RDC-1, RDC1', description='atypical chemokine receptor 3', other_designations='C-X-C chemokine receptor type 7|G protein-coupled receptor|G-protein coupled receptor 159|G-protein coupled receptor RDC1 homolog|chemokine (C-X-C motif) receptor 7|chemokine orphan receptor 1', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236537126, end_position_on_the_genomic_accession=236582358, orientation='plus', exon_count=6, OMIM=610376, no_hdr0=''),
  57008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57008, CurrentID=0, Status='live', Symbol='RLFP1', Aliases='RLFP', description='rearranged L-myc fusion pseudogene 1', other_designations='Zn-15 related zinc finger protein RLF pseudogene', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116095492, end_position_on_the_genomic_accession=116097100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57009, CurrentID=0, Status='live', Symbol='PTS-P1', Aliases='', description='6-pyruvoyltetrahydropterin synthase pseudogene', other_designations='', map_location='9p13-p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42196047, end_position_on_the_genomic_accession=42196732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57010, CurrentID=0, Status='live', Symbol='CABP4', Aliases='CRSD, CSNB2B', description='calcium binding protein 4', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67452415, end_position_on_the_genomic_accession=67461774, orientation='plus', exon_count=8, OMIM=608965, no_hdr0=''),
  57016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57016, CurrentID=0, Status='live', Symbol='AKR1B10', Aliases='AKR1B11, AKR1B12, ALDRLn, ARL-1, ARL1, HIS, HSI', description='aldo-keto reductase family 1, member B10 (aldose reductase)', other_designations='ARP|SI reductase|aldo-keto reductase family 1, member B11 (aldose reductase-like)|aldose reductase-like 1|aldose reductase-like peptide|aldose reductase-related protein|hARP|small intestine reductase', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134527592, end_position_on_the_genomic_accession=134541414, orientation='plus', exon_count=11, OMIM=604707, no_hdr0=''),
  57017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57017, CurrentID=0, Status='live', Symbol='COQ9', Aliases='C16orf49, COQ10D5', description='coenzyme Q9', other_designations='coenzyme Q9 homolog', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57447425, end_position_on_the_genomic_accession=57461275, orientation='plus', exon_count=9, OMIM=612837, no_hdr0=''),
  57018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57018, CurrentID=0, Status='live', Symbol='CCNL1', Aliases='ANIA6A, BM-001, PRO1073, ania-6a', description='cyclin L1', other_designations='cyclin L ania-6a|cyclin L gamma|cyclin-L', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157146364, end_position_on_the_genomic_accession=157160693, orientation='minus', exon_count=15, OMIM=613384, no_hdr0=''),
  57019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57019, CurrentID=0, Status='live', Symbol='CIAPIN1', Aliases='Anamorsin, DRE2, PRO0915', description='cytokine induced apoptosis inhibitor 1', other_designations='cytokine-induced apoptosis inhibitor 1|fe-S cluster assembly protein DRE2 homolog|predicted protein of HQ0915', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57428174, end_position_on_the_genomic_accession=57447457, orientation='minus', exon_count=9, OMIM=608943, no_hdr0=''),
  57020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57020, CurrentID=0, Status='live', Symbol='C16orf62', Aliases='', description='chromosome 16 open reading frame 62', other_designations='esophageal cancer associated protein|esophageal cancer-associated protein', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19555073, end_position_on_the_genomic_accession=19701163, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  450238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450238, CurrentID=0, Status='live', Symbol='PHF5DP', Aliases='', description='PHD finger protein 5D pseudogene', other_designations='PHF5A pseudogene 2', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113361857, end_position_on_the_genomic_accession=113362797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  450239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450239, CurrentID=0, Status='live', Symbol='PHF5EP', Aliases='', description='PHD finger protein 5E pseudogene', other_designations='PHF5A pseudogene 3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49775256, end_position_on_the_genomic_accession=49776184, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  57026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57026, CurrentID=0, Status='live', Symbol='PDXP', Aliases='CIN, PLP, dJ37E16.5', description='pyridoxal (pyridoxine, vitamin B6) phosphatase', other_designations='PLP phosphatase|chronophin', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37658730, end_position_on_the_genomic_accession=37666932, orientation='plus', exon_count=2, OMIM=609246, no_hdr0=''),
  57030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57030, CurrentID=0, Status='live', Symbol='SLC17A7', Aliases='BNPI, VGLUT1', description='solute carrier family 17 (vesicular glutamate transporter), member 7', other_designations='brain-specific Na(+)-dependent inorganic phosphate cotransporter|brain-specific Na-dependent inorganic phosphate cotransporter|solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 7|solute carrier family 17 member 7', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49429398, end_position_on_the_genomic_accession=49441551, orientation='minus', exon_count=12, OMIM=605208, no_hdr0=''),
  450250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450250, CurrentID=0, Status='live', Symbol='PHF5HP', Aliases='', description='PHD finger protein 5H pseudogene', other_designations='PHF5A pseudogene 6', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100845838, end_position_on_the_genomic_accession=100846795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57035, CurrentID=0, Status='live', Symbol='RSRP1', Aliases='C1orf63, DJ465N24.2.1, NPD014', description='arginine/serine-rich protein 1', other_designations='UPF0471 protein C1orf63', map_location='1p36.13-p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25242249, end_position_on_the_genomic_accession=25267915, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57037, CurrentID=0, Status='live', Symbol='ANKMY2', Aliases='ZMYND20', description='ankyrin repeat and MYND domain containing 2', other_designations='', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16599776, end_position_on_the_genomic_accession=16645817, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  57038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57038, CurrentID=0, Status='live', Symbol='RARS2', Aliases='ArgRS, DALRD2, PCH6, PRO1992, RARSL, dJ382I10.6', description='arginyl-tRNA synthetase 2, mitochondrial', other_designations='arginine--tRNA ligase|arginine-tRNA ligase|probable arginyl-tRNA synthetase, mitochondrial', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87499347, end_position_on_the_genomic_accession=87590017, orientation='minus', exon_count=26, OMIM=611524, no_hdr0=''),
  285593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285593, CurrentID=0, Status='live', Symbol='LOC285593', Aliases='', description='uncharacterized LOC285593', other_designations='CTB-33O18.3', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173579634, end_position_on_the_genomic_accession=173585068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57042, CurrentID=0, Status='live', Symbol='IBD7', Aliases='', description='inflammatory bowel disease 7', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605225, no_hdr0=''),
  57044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57044, CurrentID=0, Status='live', Symbol='IDDM18', Aliases='', description='insulin-dependent diabetes mellitus 18', other_designations='', map_location='5q31.1-q33.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605598, no_hdr0=''),
  57045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57045, CurrentID=0, Status='live', Symbol='TWSG1', Aliases='TSG', description='twisted gastrulation BMP signaling modulator 1', other_designations='twisted gastrulation homolog 1', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9334767, end_position_on_the_genomic_accession=9402420, orientation='plus', exon_count=6, OMIM=605049, no_hdr0=''),
  57046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57046, CurrentID=0, Status='live', Symbol='HRES2', Aliases='', description='HTLV-1 related endogenous sequence 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57047, CurrentID=0, Status='live', Symbol='PLSCR2', Aliases='', description='phospholipid scramblase 2', other_designations='PL scramblase 2|ca(2+)-dependent phospholipid scramblase 2', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146391363, end_position_on_the_genomic_accession=146496012, orientation='minus', exon_count=22, OMIM=607610, no_hdr0=''),
  57048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57048, CurrentID=0, Status='live', Symbol='PLSCR3', Aliases='', description='phospholipid scramblase 3', other_designations='PL scramblase 3|ca(2+)-dependent phospholipid scramblase 3', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7389727, end_position_on_the_genomic_accession=7394843, orientation='minus', exon_count=8, OMIM=607611, no_hdr0=''),
  57050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57050, CurrentID=0, Status='live', Symbol='UTP3', Aliases='CRL1, CRLZ1, SAS10', description='UTP3, small subunit (SSU) processome component, homolog (S. cerevisiae)', other_designations='UTP3 homolog|charged amino acid rich leucine zipper 1 homolog|charged amino acid-rich leucine zipper 1|disrupter of silencing 10|disrupter of silencing SAS10', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70688479, end_position_on_the_genomic_accession=70690551, orientation='plus', exon_count=1, OMIM=611614, no_hdr0=''),
  57051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57051, CurrentID=0, Status='live', Symbol='NAG18', Aliases='', description='NAG18 mRNA', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=63132859, end_position_on_the_genomic_accession=63143489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57053, CurrentID=0, Status='live', Symbol='CHRNA10', Aliases='', description='cholinergic receptor, nicotinic alpha 10', other_designations='NACHR alpha-10|acetylcholine receptor, nicotinic, alpha 10 (neuronal)|cholinergic receptor, nicotinic, alpha 10 (neuronal)|cholinergic receptor, nicotinic, alpha polypeptide 10|nicotinic acetylcholine receptor subunit alpha-10', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3665587, end_position_on_the_genomic_accession=3673208, orientation='minus', exon_count=5, OMIM=606372, no_hdr0=''),
  57054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57054, CurrentID=0, Status='live', Symbol='DAZ3', Aliases='pDP1679', description='deleted in azoospermia 3', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24763069, end_position_on_the_genomic_accession=24813505, orientation='minus', exon_count=21, OMIM=400027, no_hdr0=''),
  57055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57055, CurrentID=0, Status='live', Symbol='DAZ2', Aliases='pDP1678', description='deleted in azoospermia 2', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23219457, end_position_on_the_genomic_accession=23291356, orientation='plus', exon_count=24, OMIM=400026, no_hdr0=''),
  57057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57057, CurrentID=0, Status='live', Symbol='TBX20', Aliases='ASD4', description='T-box 20', other_designations='T-box protein 20', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35202430, end_position_on_the_genomic_accession=35254100, orientation='minus', exon_count=8, OMIM=606061, no_hdr0=''),
  57060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57060, CurrentID=0, Status='live', Symbol='PCBP4', Aliases='CBP, LIP4, MCG10', description='poly(rC) binding protein 4', other_designations='LYST-interacting protein|RNA binding protein MCG10|alpha-CP4', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51957454, end_position_on_the_genomic_accession=51967466, orientation='minus', exon_count=16, OMIM=608503, no_hdr0=''),
  57061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57061, CurrentID=0, Status='live', Symbol='HYMAI', Aliases='NCRNA00020', description='hydatidiform mole associated and imprinted (non-protein coding)', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144002886, end_position_on_the_genomic_accession=144008730, orientation='minus', exon_count=2, OMIM=606546, no_hdr0=''),
  57062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57062, CurrentID=0, Status='live', Symbol='DDX24', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 24', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 24|DEAD box protein 24|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 24|S. cerevisiae CHL1-like helicase', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94050922, end_position_on_the_genomic_accession=94081212, orientation='minus', exon_count=9, OMIM=606181, no_hdr0=''),
  105369553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369553, CurrentID=0, Status='live', Symbol='LOC105369553', Aliases='', description='uncharacterized LOC105369553', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125873449, end_position_on_the_genomic_accession=125875396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128360, CurrentID=0, Status='live', Symbol='LOC100128360', Aliases='', description='axin interactor, dorsalization associated pseudogene', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1927726, end_position_on_the_genomic_accession=1929087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369554, CurrentID=0, Status='live', Symbol='LOC105369554', Aliases='', description='uncharacterized LOC105369554', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125924221, end_position_on_the_genomic_accession=125939101, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372308, CurrentID=0, Status='live', Symbol='LOC105372308', Aliases='', description='uncharacterized LOC105372308', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20030717, end_position_on_the_genomic_accession=20033053, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57082, CurrentID=0, Status='live', Symbol='CASC5', Aliases='AF15Q14, CT29, D40, KNL1, PPP1R55, Spc7, hKNL-1, hSpc105', description='cancer susceptibility candidate 5', other_designations='ALL1-fused gene from chromosome 15q14 protein|blinkin, bub-linking kinetochore protein|cancer susceptibility candidate gene 5 protein|cancer/testis antigen 29|kinetochore null 1 homolog|kinetochore-null protein 1|protein phosphatase 1, regulatory subunit 55', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40594249, end_position_on_the_genomic_accession=40662683, orientation='plus', exon_count=27, OMIM=609173, no_hdr0=''),
  57084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57084, CurrentID=0, Status='live', Symbol='SLC17A6', Aliases='DNPI, VGLUT2', description='solute carrier family 17 (vesicular glutamate transporter), member 6', other_designations='differentiation-associated BNPI|differentiation-associated Na(+)-dependent inorganic phosphate cotransporter|differentiation-associated Na-dependent inorganic phosphate cotransporter|solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 6|solute carrier family 17 member 6', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22338121, end_position_on_the_genomic_accession=22379503, orientation='plus', exon_count=12, OMIM=607563, no_hdr0=''),
  57085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57085, CurrentID=0, Status='live', Symbol='AGTRAP', Aliases='ATRAP', description='angiotensin II receptor-associated protein', other_designations='AT1 receptor-associated protein|ATI receptor-associated protein|angiotensin II, type I receptor-associated protein', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11735474, end_position_on_the_genomic_accession=11750771, orientation='plus', exon_count=8, OMIM=608729, no_hdr0=''),
  57088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57088, CurrentID=0, Status='live', Symbol='PLSCR4', Aliases='TRA1', description='phospholipid scramblase 4', other_designations='Ca(2+)-dependent phospholipid scramblase 4|PL scramblase 4|cell growth inhibiting protein 43', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146192335, end_position_on_the_genomic_accession=146251179, orientation='minus', exon_count=12, OMIM=607612, no_hdr0=''),
  57089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57089, CurrentID=0, Status='live', Symbol='ENTPD7', Aliases='LALP1', description='ectonucleoside triphosphate diphosphohydrolase 7', other_designations='NTPDase 7|lysosomal apyrase-like protein 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99659506, end_position_on_the_genomic_accession=99711245, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101928724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928724, CurrentID=0, Status='live', Symbol='LOC101928724', Aliases='', description='eukaryotic translation initiation factor 3 subunit J-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62348945, end_position_on_the_genomic_accession=62365199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57091, CurrentID=0, Status='live', Symbol='CASS4', Aliases='C20orf32, CAS4, HEFL, HEPL', description='Cas scaffolding protein family member 4', other_designations='HEF-like protein|HEF1-EFS-p130Cas-like protein', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56411548, end_position_on_the_genomic_accession=56459340, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57092, CurrentID=0, Status='live', Symbol='PCNP', Aliases='', description='PEST proteolytic signal containing nuclear protein', other_designations='PEST-containing nuclear protein', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101573534, end_position_on_the_genomic_accession=101594437, orientation='plus', exon_count=5, OMIM=615210, no_hdr0=''),
  57093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57093, CurrentID=0, Status='live', Symbol='TRIM49', Aliases='RNF18A, TRIM49L2, TRIM49', description='tripartite motif containing 49', other_designations='ring finger protein 18|testis-specific RING Finger protein|testis-specific ring-finger protein|tripartite motif-containing 49', map_location='11p11.12-q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89797655, end_position_on_the_genomic_accession=89808575, orientation='minus', exon_count=8, OMIM=606124, no_hdr0=''),
  57094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57094, CurrentID=0, Status='live', Symbol='CPA6', Aliases='CPAH, ETL5, FEB11', description='carboxypeptidase A6', other_designations='carboxypeptidase B', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67422170, end_position_on_the_genomic_accession=67746571, orientation='minus', exon_count=12, OMIM=609562, no_hdr0=''),
  57095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57095, CurrentID=0, Status='live', Symbol='PITHD1', Aliases='C1orf128, HT014, TXNL1CL', description='PITH (C-terminal proteasome-interacting domain of thioredoxin-like) domain containing 1', other_designations='HT014|TXNL1 C-terminal like', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23778386, end_position_on_the_genomic_accession=23788232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57096, CurrentID=0, Status='live', Symbol='RPGRIP1', Aliases='CORD13, LCA6, RGI1, RGRIP, RPGRIPd, RPGRIP1', description='retinitis pigmentosa GTPase regulator interacting protein 1', other_designations='RPGR-interacting protein 1', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21287962, end_position_on_the_genomic_accession=21351301, orientation='plus', exon_count=25, OMIM=605446, no_hdr0=''),
  57097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57097, CurrentID=0, Status='live', Symbol='PARP11', Aliases='ARTD11, C12orf6, MIB006', description='poly(ADP-ribose) polymerase family member 11', other_designations='ADP-ribosyltransferase diphtheria toxin-like 11|poly (ADP-ribose) polymerase family, member 11', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3808861, end_position_on_the_genomic_accession=3873448, orientation='minus', exon_count=14, OMIM=616706, no_hdr0=''),
  57099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57099, CurrentID=0, Status='live', Symbol='AVEN', Aliases='PDCD12', description='apoptosis, caspase activation inhibitor', other_designations='programmed cell death 12', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33866227, end_position_on_the_genomic_accession=34074458, orientation='minus', exon_count=11, OMIM=605265, no_hdr0=''),
  57101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57101, CurrentID=0, Status='live', Symbol='ANO2', Aliases='C12orf3, TMEM16B', description='anoctamin 2, calcium activated chloride channel', other_designations='transmembrane protein 16B (eight membrane-spanning domains)', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5560241, end_position_on_the_genomic_accession=5948791, orientation='minus', exon_count=41, OMIM=610109, no_hdr0=''),
  57102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57102, CurrentID=0, Status='live', Symbol='C12orf4', Aliases='', description='chromosome 12 open reading frame 4', other_designations='', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4487730, end_position_on_the_genomic_accession=4538508, orientation='minus', exon_count=15, OMIM=616082, no_hdr0=''),
  57103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57103, CurrentID=0, Status='live', Symbol='TIGAR', Aliases='C12orf5, FR2BP', description='TP53 induced glycolysis regulatory phosphatase', other_designations='TP53-induced glycolysis and apoptosis regulator|TP53-induced glycolysis regulatory phosphatase|fructose-2,6-bisphosphate 2-phosphatase|probable fructose-2,6-bisphosphatase TIGAR|transactivated by NS3TP2 protein', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4321193, end_position_on_the_genomic_accession=4360028, orientation='plus', exon_count=6, OMIM=610775, no_hdr0=''),
  57104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57104, CurrentID=0, Status='live', Symbol='PNPLA2', Aliases='1110001C14Rik, ATGL, FP17548, PEDF-R, TTS-2.2, TTS2, iPLA2zeta', description='patatin like phospholipase domain containing 2', other_designations='IPLA2-zeta|TTS2.2|adipose triglyceride lipase|calcium-independent phospholipase A2|desnutrin|patatin-like phospholipase domain containing 2|patatin-like phospholipase domain containing protein 2|pigment epithelium-derived factor|transport-secretion protein 2.2|triglyceride hydrolase', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=818891, end_position_on_the_genomic_accession=825573, orientation='plus', exon_count=9, OMIM=609059, no_hdr0=''),
  57105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57105, CurrentID=0, Status='live', Symbol='CYSLTR2', Aliases='CYSLT2, CYSLT2R, GPCR21, HG57, HPN321, KPG_011, PSEC0146, hGPCR21', description='cysteinyl leukotriene receptor 2', other_designations='G-protein coupled receptor GPCR21|G-protein coupled receptor HG57|cysteinyl leukotriene CysLT2 receptor', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48706815, end_position_on_the_genomic_accession=48709362, orientation='plus', exon_count=1, OMIM=605666, no_hdr0=''),
  57106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57106, CurrentID=0, Status='live', Symbol='NAT14', Aliases='KLP1', description='N-acetyltransferase 14 (GCN5-related, putative)', other_designations='K562 cell-derived leucine-zipper-like protein 1|K562 cells-derived leucine zipper-like protein 1|K562 cells-derived leucine-zipper-like protein 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55485190, end_position_on_the_genomic_accession=55487568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57107, CurrentID=0, Status='live', Symbol='PDSS2', Aliases='C6orf210, COQ10D3, DLP1, bA59I9.3, hDLP1', description='prenyl (decaprenyl) diphosphate synthase, subunit 2', other_designations='all-trans-decaprenyl-diphosphate synthase subunit 2|decaprenyl pyrophosphate synthase subunit 2|decaprenyl pyrophosphate synthetase subunit 2|subunit 2 of decaprenyl diphosphate synthase', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107152557, end_position_on_the_genomic_accession=107459702, orientation='minus', exon_count=10, OMIM=610564, no_hdr0=''),
  57108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57108, CurrentID=0, Status='live', Symbol='SETD8P1', Aliases='SET07p, SET7', description='SET domain containing 8 pseudogene 1', other_designations='PR/SET domain containing protein 07 pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18297685, end_position_on_the_genomic_accession=18300377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  57109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57109, CurrentID=0, Status='live', Symbol='REXO4', Aliases='REX4, XPMC2, XPMC2H', description="REX4 homolog, 3'-5' exonuclease", other_designations='REX4, RNA exonuclease 4 homolog|XPMC2 prevents mitotic catastrophe 2 homolog|Xenopus prevents mitotic catastrophe 2 homolog|exonuclease XPMC2', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133406058, end_position_on_the_genomic_accession=133418172, orientation='minus', exon_count=8, OMIM=602930, no_hdr0=''),
  57110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57110, CurrentID=0, Status='live', Symbol='HRASLS', Aliases='A-C1, H-REV1071, HRSL1, HSD28, HRASLS', description='HRAS-like suppressor', other_designations='H-REV107 protein-related protein|HRAS-like suppressor 1|phospholipase A', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193241128, end_position_on_the_genomic_accession=193300939, orientation='plus', exon_count=10, OMIM=606487, no_hdr0=''),
  57111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57111, CurrentID=0, Status='live', Symbol='RAB25', Aliases='CATX-8, RAB11C', description='RAB25, member RAS oncogene family', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156061175, end_position_on_the_genomic_accession=156070504, orientation='plus', exon_count=5, OMIM=612942, no_hdr0=''),
  57113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57113, CurrentID=0, Status='live', Symbol='TRPC7', Aliases='TRP7', description='transient receptor potential cation channel, subfamily C, member 7', other_designations='TRP-7|hTRP7|likley ortholog of mouse transient receptor potential cation channel, subfamily C, member 7|putative capacitative calcium channel|transient receptor protein 7', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136212765, end_position_on_the_genomic_accession=136365476, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57115, CurrentID=0, Status='live', Symbol='PGLYRP4', Aliases='PGLYRPIbeta, PGRP-Ibeta, PGRPIB, SBBI67', description='peptidoglycan recognition protein 4', other_designations='PGRP-I-beta|peptidoglycan recognition protein I beta|peptidoglycan recognition protein I-beta|peptidoglycan recognition protein intermediate beta|peptidoglycan recognition protein-I-beta', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153327380, end_position_on_the_genomic_accession=153348841, orientation='minus', exon_count=11, OMIM=608198, no_hdr0=''),
  57116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57116, CurrentID=0, Status='live', Symbol='ZNF695', Aliases='SBZF3', description='zinc finger protein 695', other_designations='zinc finger protein SBZF3', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246945547, end_position_on_the_genomic_accession=247008093, orientation='minus', exon_count=6, OMIM=616348, no_hdr0=''),
  57117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57117, CurrentID=0, Status='live', Symbol='INTS12', Aliases='INT12, PHF22, SBBI22', description='integrator complex subunit 12', other_designations='PHD finger protein 22|hypothetical nuclear factor SBBI22', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105682627, end_position_on_the_genomic_accession=105708729, orientation='minus', exon_count=11, OMIM=611355, no_hdr0=''),
  57118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57118, CurrentID=0, Status='live', Symbol='CAMK1D', Aliases='CKLiK, CaM-K1, CaMKID', description='calcium/calmodulin dependent protein kinase ID', other_designations='CaM kinase ID|CamKI-like protein kinase|caM kinase I delta|caM-KI delta|caMKI delta|calcium/calmodulin-dependent protein kinase ID', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12349517, end_position_on_the_genomic_accession=12835545, orientation='plus', exon_count=17, OMIM=607957, no_hdr0=''),
  57119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57119, CurrentID=0, Status='live', Symbol='EPPIN', Aliases='CT71, CT72, SPINLW1, WAP7, WFDC7, dJ461P17.2', description='epididymal peptidase inhibitor', other_designations='WAP four-disulfide core domain protein 7|cancer/testis antigen 71|cancer/testis antigen 72|epididymal protease inhibitor|protease inhibitor WAP7|serine peptidase inhibitor-like, with Kunitz and WAP domains 1 (eppin)', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45540626, end_position_on_the_genomic_accession=45547426, orientation='minus', exon_count=4, OMIM=609031, no_hdr0=''),
  57120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57120, CurrentID=0, Status='live', Symbol='GOPC', Aliases='CAL, FIG1, PIST, dJ94G16.2, GOPC', description='golgi-associated PDZ and coiled-coil motif containing', other_designations='CFTR-associated ligand|Golgi associated PDZ and coiled-coil motif containing protein|PDZ protein interacting specifically with TC10|PDZ/coiled-coil domain binding partner for the rho-family GTPase TC10|dJ94G16.2 PIST|fused in glioblastoma|golgi-associated PDZ and coiled-coil motif containing protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117560269, end_position_on_the_genomic_accession=117602542, orientation='minus', exon_count=9, OMIM=606845, no_hdr0=''),
  57121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57121, CurrentID=0, Status='live', Symbol='LPAR5', Aliases='GPR92, GPR93, KPG_010, LPA5', description='lysophosphatidic acid receptor 5', other_designations='G protein-coupled receptor 92|G-protein coupled receptor 92|G-protein coupled receptor 93|LPA receptor 5|LPA-5', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6618835, end_position_on_the_genomic_accession=6636131, orientation='minus', exon_count=3, OMIM=606926, no_hdr0=''),
  57122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57122, CurrentID=0, Status='live', Symbol='NUP107', Aliases='NUP84', description='nucleoporin 107kDa', other_designations='107 kDa nucleoporin|nucleoporin Nup107', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68686951, end_position_on_the_genomic_accession=68742693, orientation='plus', exon_count=28, OMIM=607617, no_hdr0=''),
  57124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57124, CurrentID=0, Status='live', Symbol='CD248', Aliases='CD164L1, TEM1', description='CD248 molecule, endosialin', other_designations='2610111G01Rik|CD164 sialomucin-like 1|CD248 antigen, endosialin|tumor endothelial marker 1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66314487, end_position_on_the_genomic_accession=66317044, orientation='minus', exon_count=1, OMIM=606064, no_hdr0=''),
  57125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57125, CurrentID=0, Status='live', Symbol='PLXDC1', Aliases='TEM3, TEM7', description='plexin domain containing 1', other_designations='2410003I07Rik|tumor endothelial marker 3|tumor endothelial marker 7', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39063303, end_position_on_the_genomic_accession=39151649, orientation='minus', exon_count=14, OMIM=606826, no_hdr0=''),
  57126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57126, CurrentID=0, Status='live', Symbol='CD177', Aliases='HNA-2a, HNA2A, NB1, NB1 GP, PRV-1, PRV1', description='CD177 molecule', other_designations='NB1 glycoprotein|cell surface receptor|human neutrophil alloantigen 2a|polycythemia rubra vera protein 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43353659, end_position_on_the_genomic_accession=43363172, orientation='plus', exon_count=9, OMIM=162860, no_hdr0=''),
  57127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57127, CurrentID=0, Status='live', Symbol='RHBG', Aliases='SLC42A2', description='Rh family, B glycoprotein (gene/pseudogene)', other_designations='Rhesus blood group, B glycoprotein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156366043, end_position_on_the_genomic_accession=156385234, orientation='plus', exon_count=16, OMIM=607079, no_hdr0=''),
  57128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57128, CurrentID=0, Status='live', Symbol='LYRM4', Aliases='C6orf149, CGI-203, COXPD19, ISD11', description='LYR motif containing 4', other_designations='homolog of yeast Isd11|mitochondrial matrix Nfs1 interacting protein', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5031753, end_position_on_the_genomic_accession=5260950, orientation='minus', exon_count=12, OMIM=613311, no_hdr0=''),
  57129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57129, CurrentID=0, Status='live', Symbol='MRPL47', Aliases='CGI-204, L47mt, MRP-L47, NCM1', description='mitochondrial ribosomal protein L47', other_designations='nasopharyngeal carcinoma metastasis-related 1|nasopharyngeal carcinoma metastasis-related protein 1', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179588465, end_position_on_the_genomic_accession=179604646, orientation='minus', exon_count=7, OMIM=611852, no_hdr0=''),
  57130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57130, CurrentID=0, Status='live', Symbol='ATP13A1', Aliases='ATP13A, CGI-152', description='ATPase type 13A1', other_designations='cation transporting ATPase|probable cation-transporting ATPase 13A1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19645198, end_position_on_the_genomic_accession=19667823, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  57132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57132, CurrentID=0, Status='live', Symbol='CHMP1B', Aliases='C10orf2, C18-ORF2, C18orf2, CHMP1.5, Vps46-2, Vps46B, hVps46-2', description='charged multivesicular body protein 1B', other_designations='chromatin modifying protein 1B|chromatin-modifying protein 1b|vacuolar protein sorting 46-2|vacuolar protein sorting-associated protein 46-2', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11851390, end_position_on_the_genomic_accession=11854449, orientation='plus', exon_count=1, OMIM=606486, no_hdr0=''),
  57134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57134, CurrentID=0, Status='live', Symbol='MAN1C1', Aliases='HMIC, MAN1A3, MAN1C, pp6318', description='mannosidase, alpha, class 1C, member 1', other_designations='1,2-alpha-mannosidase IC|processing alpha-1,2-mannosidase IC', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25617468, end_position_on_the_genomic_accession=25786207, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57135, CurrentID=0, Status='live', Symbol='DAZ4', Aliases='pDP1680, pDP1681', description='deleted in azoospermia 4', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24833800, end_position_on_the_genomic_accession=24907040, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  57136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57136, CurrentID=0, Status='live', Symbol='APMAP', Aliases='BSCv, C20orf3', description='adipocyte plasma membrane associated protein', other_designations='protein BSCv', map_location='20p11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24962925, end_position_on_the_genomic_accession=24992974, orientation='minus', exon_count=9, OMIM=615884, no_hdr0=''),
  57139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57139, CurrentID=0, Status='live', Symbol='RGL3', Aliases='', description='ral guanine nucleotide dissociation stimulator-like 3', other_designations='RalGEF-like protein 3, mouse homolog|ralGDS-like 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11394056, end_position_on_the_genomic_accession=11419342, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  57140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57140, CurrentID=0, Status='live', Symbol='RNPEPL1', Aliases='', description='arginyl aminopeptidase (aminopeptidase B)-like 1', other_designations='RNPEP-like 1|argininyl aminopeptidase-like 1', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240568541, end_position_on_the_genomic_accession=240578732, orientation='plus', exon_count=11, OMIM=605287, no_hdr0=''),
  57142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57142, CurrentID=0, Status='live', Symbol='RTN4', Aliases='ASY, NI220/250, NOGO, NOGO-A, NOGOC, NSP, NSP-CL, Nbla00271, Nbla10545, Nogo-B, Nogo-C, RTN-X-A, RTN4-B1, RTN4-B2, RTN4-C, RTN4', description='reticulon 4', other_designations='Human NogoA|My043 protein|foocen|neurite growth inhibitor 220|neurite outgrowth inhibitor|neuroendocrine-specific protein C homolog|reticulon 5', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54972189, end_position_on_the_genomic_accession=55112616, orientation='minus', exon_count=15, OMIM=604475, no_hdr0=''),
  57143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57143, CurrentID=0, Status='live', Symbol='ADCK1', Aliases='', description='aarF domain containing kinase 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77800065, end_position_on_the_genomic_accession=77933954, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57144, CurrentID=0, Status='live', Symbol='PAK7', Aliases='PAK5', description='p21 protein (Cdc42/Rac)-activated kinase 7', other_designations='PAK-5|PAK-7|p21(CDKN1A)-activated kinase 7|p21-activated kinase 5|p21-activated kinase 7|p21CDKN1A-activated kinase 7|protein kinase PAK5|serine/threonine-protein kinase PAK7', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9537389, end_position_on_the_genomic_accession=9839039, orientation='minus', exon_count=14, OMIM=608038, no_hdr0=''),
  9524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9524, CurrentID=0, Status='live', Symbol='TECR', Aliases='GPSN2, MRT14, SC2, TER', description='trans-2,3-enoyl-CoA reductase', other_designations='glycoprotein, synaptic 2|synaptic glycoprotein SC2', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14527768, end_position_on_the_genomic_accession=14565980, orientation='plus', exon_count=15, OMIM=610057, no_hdr0=''),
  57146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57146, CurrentID=0, Status='live', Symbol='TMEM159', Aliases='', description='transmembrane protein 159', other_designations='', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21158377, end_position_on_the_genomic_accession=21180616, orientation='plus', exon_count=8, OMIM=611304, no_hdr0=''),
  57147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57147, CurrentID=0, Status='live', Symbol='SCYL3', Aliases='PACE-1, PACE1', description='SCY1-like, kinase-like 3', other_designations='SCY1-like 3|SCY1-like protein 3|ezrin-binding partner PACE-1 (PACE-1)|ezrin-binding protein PACE-1', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169853074, end_position_on_the_genomic_accession=169894267, orientation='minus', exon_count=16, OMIM=608192, no_hdr0=''),
  57148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57148, CurrentID=0, Status='live', Symbol='RALGAPB', Aliases='KIAA1219, RalGAPbeta', description='Ral GTPase activating protein, beta subunit (non-catalytic)', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38472806, end_position_on_the_genomic_accession=38578861, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  57149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57149, CurrentID=0, Status='live', Symbol='LYRM1', Aliases='A211C6.1', description='LYR motif containing 1', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20898805, end_position_on_the_genomic_accession=20925006, orientation='plus', exon_count=11, OMIM=614709, no_hdr0=''),
  57150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57150, CurrentID=0, Status='live', Symbol='SMIM8', Aliases='C6orf162, dJ102H19.2', description='small integral membrane protein 8', other_designations='UPF0708 protein C6orf162', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87322588, end_position_on_the_genomic_accession=87342328, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57151, CurrentID=0, Status='live', Symbol='LYZL6', Aliases='HEL-S-6a, LYC1, PRO1485, TKAL754, UNQ754', description='lysozyme like 6', other_designations='epididymis secretory sperm binding protein Li 6a|lysozyme homolog|lysozyme-like 6', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35934518, end_position_on_the_genomic_accession=35943710, orientation='minus', exon_count=5, OMIM=612751, no_hdr0=''),
  57152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57152, CurrentID=0, Status='live', Symbol='SLURP1', Aliases='ANUP, ARS, ArsB, LY6LS, MDM', description='secreted LY6/PLAUR domain containing 1', other_designations='ARS(component B)-81/S|SLURP-1|anti-neoplastic urinary protein|lymphocyte antigen 6-like secreted|secreted Ly6/uPAR related protein 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142740944, end_position_on_the_genomic_accession=142742411, orientation='minus', exon_count=3, OMIM=606119, no_hdr0=''),
  57153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57153, CurrentID=0, Status='live', Symbol='SLC44A2', Aliases='CTL2, PP1292', description='solute carrier family 44 (choline transporter), member 2', other_designations='solute carrier family 44, member 2', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10602445, end_position_on_the_genomic_accession=10644559, orientation='plus', exon_count=24, OMIM=606106, no_hdr0=''),
  57154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57154, CurrentID=0, Status='live', Symbol='SMURF1', Aliases='', description='SMAD specific E3 ubiquitin protein ligase 1', other_designations='E3 ubiquitin ligase SMURF1|SMAD-specific E3 ubiquitin-protein ligase 1|Smad ubiquitination regulatory factor 1|Smad-specific E3 ubiquitin ligase 1|hSMURF1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99027435, end_position_on_the_genomic_accession=99144120, orientation='minus', exon_count=21, OMIM=605568, no_hdr0=''),
  57156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57156, CurrentID=0, Status='live', Symbol='TMEM63C', Aliases='C14orf171, CSC1, hsCSC1', description='transmembrane protein 63C', other_designations='calcium permeable stress-gated cation channel 1 homolog', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77181759, end_position_on_the_genomic_accession=77259495, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  57157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57157, CurrentID=0, Status='live', Symbol='PHTF2', Aliases='', description='putative homeodomain transcription factor 2', other_designations='', map_location='7q11.23-q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77798792, end_position_on_the_genomic_accession=77957504, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  57158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57158, CurrentID=0, Status='live', Symbol='JPH2', Aliases='CMH17, JP-2, JP2', description='junctophilin 2', other_designations='junctophilin type 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44111697, end_position_on_the_genomic_accession=44187578, orientation='minus', exon_count=9, OMIM=605267, no_hdr0=''),
  57159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57159, CurrentID=0, Status='live', Symbol='TRIM54', Aliases='MURF, MURF-3, RNF30, muRF3', description='tripartite motif containing 54', other_designations='muscle-specific RING finger protein 3|muscle-specific RING-finger protein 3|ring finger protein 30|tripartite motif-containing 54', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27282429, end_position_on_the_genomic_accession=27307439, orientation='plus', exon_count=10, OMIM=606474, no_hdr0=''),
  57160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57160, CurrentID=0, Status='live', Symbol='TBX23P', Aliases='TBX23', description='T-box 23, pseudogene', other_designations='T-box 23 (pseudogene)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57161, CurrentID=0, Status='live', Symbol='PELI2', Aliases='', description='pellino E3 ubiquitin protein ligase family member 2', other_designations='pellino homolog 2|pellino-2|protein pellino homolog 2', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56118323, end_position_on_the_genomic_accession=56301581, orientation='plus', exon_count=10, OMIM=614798, no_hdr0=''),
  57162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57162, CurrentID=0, Status='live', Symbol='PELI1', Aliases='', description='pellino E3 ubiquitin protein ligase 1', other_designations='pellino homolog 1|pellino-1|pellino-related intracellular-signaling molecule|protein pellino homolog 1', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64092652, end_position_on_the_genomic_accession=64144471, orientation='minus', exon_count=11, OMIM=614797, no_hdr0=''),
  57165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57165, CurrentID=0, Status='live', Symbol='GJC2', Aliases='CX46.6, Cx47, GJA12, HLD2, LMPH1C, PMLDAR, SPG44', description='gap junction protein gamma 2', other_designations='connexin-46.6|connexin-47|gap junction alpha-12 protein|gap junction protein, gamma 2, 47kDa', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228149714, end_position_on_the_genomic_accession=228159826, orientation='plus', exon_count=2, OMIM=608803, no_hdr0=''),
  57167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57167, CurrentID=0, Status='live', Symbol='SALL4', Aliases='DRRS, HSAL4, ZNF797', description='spalt-like transcription factor 4', other_designations='zinc finger protein 797|zinc finger protein SALL4', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51784009, end_position_on_the_genomic_accession=51913245, orientation='minus', exon_count=10, OMIM=607343, no_hdr0=''),
  57168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57168, CurrentID=0, Status='live', Symbol='ASPHD2', Aliases='', description='aspartate beta-hydroxylase domain containing 2', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26429314, end_position_on_the_genomic_accession=26445012, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57169, CurrentID=0, Status='live', Symbol='ZNFX1', Aliases='', description='zinc finger, NFX1-type containing 1', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49245897, end_position_on_the_genomic_accession=49278244, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  57171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57171, CurrentID=0, Status='live', Symbol='DOLPP1', Aliases='LSFR2', description='dolichyldiphosphatase 1', other_designations='dolichyl pyrophosphate phosphatase 1|linked to Surfeit genes in Fugu rubripes 2', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129081104, end_position_on_the_genomic_accession=129090438, orientation='plus', exon_count=8, OMIM=614516, no_hdr0=''),
  57172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57172, CurrentID=0, Status='live', Symbol='CAMK1G', Aliases='CLICK3, CLICKIII, VWS1, dJ272L16.1', description='calcium/calmodulin-dependent protein kinase IG', other_designations='CLICK III|caM kinase I gamma|caM kinase IG|caM-KI gamma|caMK-like CREB kinase III|caMKI gamma|caMKIG', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209583700, end_position_on_the_genomic_accession=209613939, orientation='plus', exon_count=13, OMIM=614994, no_hdr0=''),
  57175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57175, CurrentID=0, Status='live', Symbol='CORO1B', Aliases='CORONIN-2', description='coronin, actin binding protein, 1B', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67438047, end_position_on_the_genomic_accession=67443821, orientation='minus', exon_count=12, OMIM=609849, no_hdr0=''),
  57176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57176, CurrentID=0, Status='live', Symbol='VARS2', Aliases='COXPD20, VALRSL, VARSL, VARS2', description='valyl-tRNA synthetase 2, mitochondrial', other_designations='valine tRNA ligase 2, mitochondrial (putative)|valyl-tRNA synthetase like', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30914208, end_position_on_the_genomic_accession=30926459, orientation='plus', exon_count=30, OMIM=612802, no_hdr0=''),
  57178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57178, CurrentID=0, Status='live', Symbol='ZMIZ1', Aliases='MIZ, RAI17, TRAFIP10, ZIMP10, hZIMP10', description='zinc finger, MIZ-type containing 1', other_designations='FLJ00092 protein|RP11-519K18.1|retinoic acid induced 17|zinc finger-containing, Miz1, PIAS-like protein on chromosome 10', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79068964, end_position_on_the_genomic_accession=79316528, orientation='plus', exon_count=28, OMIM=607159, no_hdr0=''),
  57179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57179, CurrentID=0, Status='live', Symbol='KIAA1191', Aliases='p60MONOX', description='KIAA1191', other_designations='brain-derived rescue factor p60MONOX|flavin monooxygenase motif-containing protein of 33 kDa|flavine monooxygenase motif-containing protein of 33 kDa', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176346061, end_position_on_the_genomic_accession=176361807, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57180, CurrentID=0, Status='live', Symbol='ACTR3B', Aliases='ARP11, ARP3BETA', description='ARP3 actin-related protein 3 homolog B (yeast)', other_designations='ARP3-beta|actin-like protein 3B|actin-related protein 3-beta|actin-related protein ARP4|actin-related protein Arp11', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152759749, end_position_on_the_genomic_accession=152887759, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  57181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57181, CurrentID=0, Status='live', Symbol='SLC39A10', Aliases='LZT-Hs2', description='solute carrier family 39 (zinc transporter), member 10', other_designations='ZIP-10|solute carrier family 39 (metal ion transporter), member 10|solute carrier family 39 member 10|zrt- and Irt-like protein 10', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195575977, end_position_on_the_genomic_accession=195737702, orientation='plus', exon_count=17, OMIM=608733, no_hdr0=''),
  57182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57182, CurrentID=0, Status='live', Symbol='ANKRD50', Aliases='', description='ankyrin repeat domain 50', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124664048, end_position_on_the_genomic_accession=124713220, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369573, CurrentID=0, Status='live', Symbol='LOC105369573', Aliases='', description='uncharacterized LOC105369573', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130220736, end_position_on_the_genomic_accession=130223631, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57184, CurrentID=0, Status='live', Symbol='FAM219B', Aliases='C15orf17', description='family with sequence similarity 219 member B', other_designations='family with sequence similarity 219, member B|uncharacterized protein C15orf17', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74899987, end_position_on_the_genomic_accession=74907307, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57185, CurrentID=0, Status='live', Symbol='NIPAL3', Aliases='DJ462O23.2, NPAL3', description='NIPA-like domain containing 3', other_designations='', map_location='1p36.12-p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24413674, end_position_on_the_genomic_accession=24472983, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  57186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57186, CurrentID=0, Status='live', Symbol='RALGAPA2', Aliases='AS250, C20orf74, bA287B20.1, dJ1049G11, dJ1049G11.4, p220', description='Ral GTPase activating protein, alpha subunit 2 (catalytic)', other_designations='250 kDa substrate of Akt|RapGAPalpha2|akt substrate AS250|ral GTPase-activating protein alpha subunit 2', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20389558, end_position_on_the_genomic_accession=20712638, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  57187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57187, CurrentID=0, Status='live', Symbol='THOC2', Aliases='CXorf3, MRX12, MRX35, THO2, dJ506G2.1, hTREX120', description='THO complex 2', other_designations='mental retardation, X-linked 12|mental retardation, X-linked 35', map_location='Xq25-q26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123600561, end_position_on_the_genomic_accession=123733054, orientation='minus', exon_count=41, OMIM=300395, no_hdr0=''),
  57188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57188, CurrentID=0, Status='live', Symbol='ADAMTSL3', Aliases='ADAMTSL-3', description='ADAMTS like 3', other_designations='ADAMTS-like 3|a disintegrin-like and metalloprotease domain with thrombospondin type I motifs-like 3|punctin-2', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83654086, end_position_on_the_genomic_accession=84039842, orientation='plus', exon_count=33, OMIM=609199, no_hdr0=''),
  57189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57189, CurrentID=0, Status='live', Symbol='KIAA1147', Aliases='LCHN, PRO2561', description='KIAA1147', other_designations='predicted protein of HQ2561', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141656728, end_position_on_the_genomic_accession=141702169, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  57190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57190, CurrentID=0, Status='live', Symbol='SEPN1', Aliases='CFTD, MDRS1, RSMD1, RSS, SELN', description='selenoprotein N, 1', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25800176, end_position_on_the_genomic_accession=25818222, orientation='plus', exon_count=13, OMIM=606210, no_hdr0=''),
  57191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57191, CurrentID=0, Status='live', Symbol='VN1R1', Aliases='V1RL1, VNR19I1, ZVNH1, ZVNR1', description='vomeronasal 1 receptor 1', other_designations='G-protein coupled receptor GPCR24|V1R-like 1|V1r-like receptor 1|V3r-related gene protein|hGPCR24|vomeronasal olfactory receptor chromosome 19 subtype I member 1|vomeronasal olfactory receptor, (chromosome 19) subtype I, member 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57455174, end_position_on_the_genomic_accession=57456739, orientation='minus', exon_count=1, OMIM=605234, no_hdr0=''),
  57192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57192, CurrentID=0, Status='live', Symbol='MCOLN1', Aliases='MG-2, ML4, MLIV, MST080, MSTP080, TRP-ML1, TRPM-L1, TRPML1', description='mucolipin 1', other_designations='mucolipidin|mucolipidosis type IV protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7522610, end_position_on_the_genomic_accession=7534009, orientation='plus', exon_count=14, OMIM=605248, no_hdr0=''),
  57194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57194, CurrentID=0, Status='live', Symbol='ATP10A', Aliases='ATP10C, ATPVA, ATPVC', description='ATPase, class V, type 10A', other_designations='ATPase type IV, phospholipid transporting (P-type)|ATPase, Class V, type 10C|P4-ATPase flippase complex alpha subunit ATP10A|aminophospholipid translocase VA|phospholipid-transporting ATPase VA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25841492, end_position_on_the_genomic_accession=25862117, orientation='minus', exon_count=29, OMIM=605855, no_hdr0=''),
  105369575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369575, CurrentID=0, Status='live', Symbol='LOC105369575', Aliases='', description='uncharacterized LOC105369575', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130723242, end_position_on_the_genomic_accession=130735530, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57198, CurrentID=0, Status='live', Symbol='ATP8B2', Aliases='ATPID', description='ATPase, aminophospholipid transporter, class I, type 8B, member 2', other_designations='36/8-9 fusion protein with epitope for anti-lectin antibody|ATPase, class I, type 8B, member 2|P4-ATPase flippase complex alpha subunit ATP8B2|probable phospholipid-transporting ATPase ID', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154325560, end_position_on_the_genomic_accession=154351304, orientation='plus', exon_count=30, OMIM=605867, no_hdr0=''),
  9534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9534, CurrentID=0, Status='live', Symbol='ZNF254', Aliases='BMZF-5, HD-ZNF1, ZNF539, ZNF91L', description='zinc finger protein 254', other_designations='CTD-2017D11.1|bone marrow zinc finger 5|hematopoietic cell-derived zinc finger protein 1|zinc finger protein 539', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=24033405, end_position_on_the_genomic_accession=24129968, orientation='plus', exon_count=13, OMIM=604768, no_hdr0=''),
  105369577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369577, CurrentID=0, Status='live', Symbol='LOC105369577', Aliases='', description='uncharacterized LOC105369577', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130994087, end_position_on_the_genomic_accession=130997506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  57209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57209, CurrentID=0, Status='live', Symbol='ZNF248', Aliases='bA162G10.3', description='zinc finger protein 248', other_designations='KRAB protein domain|Kruppel-type zinc finger protein|bA162G10.3 (zinc finger protein)', map_location='10p11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37744363, end_position_on_the_genomic_accession=37860851, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  57210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57210, CurrentID=0, Status='live', Symbol='SLC45A4', Aliases='', description='solute carrier family 45 member 4', other_designations='solute carrier family 45, member 4', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141207166, end_position_on_the_genomic_accession=141308310, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  57211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57211, CurrentID=0, Status='live', Symbol='ADGRG6', Aliases='APG1, DREG, GPR126, PS1TP2, VIGR', description='adhesion G protein-coupled receptor G6', other_designations='G protein-coupled receptor 126|HBV PreS1-transactivated protein 2|developmentally regulated G-protein-coupled receptor|vascular inducible G protein-coupled receptor|vascular-inducible G protein-coupled receptor', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142301919, end_position_on_the_genomic_accession=142446266, orientation='plus', exon_count=27, OMIM=612243, no_hdr0=''),
  57212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57212, CurrentID=0, Status='live', Symbol='TP73-AS1', Aliases='KIAA0495, PDAM', description='TP73 antisense RNA 1', other_designations='TP73 antisense RNA 1 (non-protein coding)|p53-dependent apoptosis modulator', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3735984, end_position_on_the_genomic_accession=3747373, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57213, CurrentID=0, Status='live', Symbol='SPRYD7', Aliases='C13orf1, CLLD6', description='SPRY domain containing 7', other_designations='CLL deletion region gene 6 protein|chronic lymphocytic leukemia deletion region gene 6 protein', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49912702, end_position_on_the_genomic_accession=49995962, orientation='minus', exon_count=11, OMIM=607866, no_hdr0=''),
  57214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57214, CurrentID=0, Status='live', Symbol='CEMIP', Aliases='CCSP1, KIAA1199, TMEM2L', description='cell migration inducing protein, hyaluronan binding', other_designations='colon cancer secreted protein 1', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80779343, end_position_on_the_genomic_accession=80951662, orientation='plus', exon_count=30, OMIM=608366, no_hdr0=''),
  57215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57215, CurrentID=0, Status='live', Symbol='THAP11', Aliases='CTG-B43a, CTG-B45d, HRIHFB2206, RONIN', description='THAP domain containing 11', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67842310, end_position_on_the_genomic_accession=67844195, orientation='plus', exon_count=1, OMIM=609119, no_hdr0=''),
  57216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57216, CurrentID=0, Status='live', Symbol='VANGL2', Aliases='LPP1, LTAP, STB1, STBM, STBM1', description='VANGL planar cell polarity protein 2', other_designations='loop-tail protein 1 homolog|loop-tail-associated protein|strabismus 1|van Gogh-like protein 2|vang-like 2 (van gogh, Drosophila)', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160400574, end_position_on_the_genomic_accession=160428678, orientation='plus', exon_count=11, OMIM=600533, no_hdr0=''),
  57217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57217, CurrentID=0, Status='live', Symbol='TTC7A', Aliases='GIDID, MINAT, TTC7', description='tetratricopeptide repeat domain 7A', other_designations='TPR repeat protein 7A', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46916129, end_position_on_the_genomic_accession=47076136, orientation='plus', exon_count=29, OMIM=609332, no_hdr0=''),
  57221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57221, CurrentID=0, Status='live', Symbol='ARFGEF3', Aliases='A7322, BIG3, C6orf92, KIAA1244, PPP1R33, dJ171N11.1', description='ARFGEF family member 3', other_designations='protein phosphatase 1, regulatory subunit 33', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138161916, end_position_on_the_genomic_accession=138344663, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  57222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57222, CurrentID=0, Status='live', Symbol='ERGIC1', Aliases='ERGIC-32, ERGIC32, NET24', description='endoplasmic reticulum-golgi intermediate compartment 1', other_designations='ER-Golgi intermediate compartment 32 kDa protein|endoplasmic reticulum-golgi intermediate compartment (ERGIC) 1|endoplasmic reticulum-golgi intermediate compartment 32 kDa protein', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172834220, end_position_on_the_genomic_accession=172952685, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57223, CurrentID=0, Status='live', Symbol='PPP4R3B', Aliases='FLFL2, PP4R3B, PSY2, SMEK2, smk1', description='protein phosphatase 4 regulatory subunit 3B', other_designations='SMEK homolog 2, suppressor of mek1|protein phosphatase 4, regulatory subunit 3B', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55547292, end_position_on_the_genomic_accession=55617725, orientation='minus', exon_count=19, OMIM=610352, no_hdr0=''),
  57224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57224, CurrentID=0, Status='live', Symbol='NHSL1', Aliases='C6orf63', description='NHS like 1', other_designations='NHS-like 1', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138422043, end_position_on_the_genomic_accession=138692616, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  57226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57226, CurrentID=0, Status='live', Symbol='LYRM2', Aliases='DJ122O8.2', description='LYR motif containing 2', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89632224, end_position_on_the_genomic_accession=89638755, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  57228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57228, CurrentID=0, Status='live', Symbol='SMAGP', Aliases='hSMAGP', description='small cell adhesion glycoprotein', other_designations='small cell transmembrane and glycosylated protein|small trans-membrane and glycosylated protein|small transmembrane and glycosylated protein', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51245349, end_position_on_the_genomic_accession=51270418, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57231, CurrentID=0, Status='live', Symbol='SNX14', Aliases='RGS-PX2, SCAR20', description='sorting nexin 14', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85505496, end_position_on_the_genomic_accession=85594156, orientation='minus', exon_count=31, OMIM=616105, no_hdr0=''),
  57232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57232, CurrentID=0, Status='live', Symbol='ZNF630', Aliases='dJ54B20.2', description='zinc finger protein 630', other_designations='dJ54B20.2 (novel KRAB box containing C2H2 type zinc finger protein)', map_location='Xp11.3-p11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48058178, end_position_on_the_genomic_accession=48071658, orientation='minus', exon_count=6, OMIM=300819, no_hdr0=''),
  57234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57234, CurrentID=0, Status='live', Symbol='LINC00869', Aliases='FAM91A2', description='long intergenic non-protein coding RNA 869', other_designations='family with sequence similarity 91, member A2', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149607012, end_position_on_the_genomic_accession=149679523, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57235, CurrentID=0, Status='live', Symbol='KIAA0485', Aliases='', description='uncharacterized LOC57235', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  154664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154664, CurrentID=0, Status='live', Symbol='ABCA13', Aliases='', description='ATP binding cassette subfamily A member 13', other_designations='ATP binding cassette transporter A13|ATP-binding cassette sub-family A member 13 variant 2|ATP-binding cassette sub-family A member 13 variant 3|ATP-binding cassette, sub-family A (ABC1), member 13', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48171448, end_position_on_the_genomic_accession=48647497, orientation='plus', exon_count=77, OMIM=607807, no_hdr0=''),
  140612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140612, CurrentID=0, Status='live', Symbol='ZFP28', Aliases='mkr5', description='ZFP28 zinc finger protein', other_designations='krueppel-like zinc finger factor X6|zfp-28|zinc finger factor X6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56537716, end_position_on_the_genomic_accession=56556810, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57242, CurrentID=0, Status='live', Symbol='KIAA0509', Aliases='', description='uncharacterized LOC57242', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369583, CurrentID=0, Status='live', Symbol='LOC105369583', Aliases='', description='uncharacterized LOC105369583', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134640184, end_position_on_the_genomic_accession=134646294, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370155, CurrentID=0, Status='live', Symbol='LOC105370155', Aliases='', description='uncharacterized LOC105370155', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33297008, end_position_on_the_genomic_accession=33322956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643058, CurrentID=0, Status='live', Symbol='LOC643058', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47135322, end_position_on_the_genomic_accession=47135948, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378885, CurrentID=0, Status='live', Symbol='LOC105378885', Aliases='', description='uncharacterized LOC105378885', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106047108, end_position_on_the_genomic_accession=106073162, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  285601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285601, CurrentID=0, Status='live', Symbol='GPR150', Aliases='PGR11', description='G protein-coupled receptor 150', other_designations='seven transmembrane helix receptor', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95618296, end_position_on_the_genomic_accession=95621672, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57282, CurrentID=0, Status='live', Symbol='SLC4A10', Aliases='NBCn2, NCBE', description='solute carrier family 4, sodium bicarbonate transporter, member 10', other_designations='solute carrier family 4, sodium bicarbonate cotransporter-like, member 10', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161624335, end_position_on_the_genomic_accession=161985276, orientation='plus', exon_count=30, OMIM=605556, no_hdr0=''),
  57289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57289, CurrentID=0, Status='live', Symbol='IGHV7-4-1', Aliases='IGHV7-41, IGHV741', description='immunoglobulin heavy variable 7-4-1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57290, CurrentID=0, Status='live', Symbol='IGHV3-30-3', Aliases='IGHV3-3, IGHV3303', description='immunoglobulin heavy variable 3-30-3', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57291, CurrentID=0, Status='live', Symbol='DANCR', Aliases='AGU2, ANCR, KIAA0114, SNHG13, lncRNA-ANCR', description='differentiation antagonizing non-protein coding RNA', other_designations='adipogenesis up-regulated transcript 2|anti-differentiation ncRNA|anti-differentiation noncoding RNA|small nucleolar RNA host gene 13 (non-protein coding)', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52712454, end_position_on_the_genomic_accession=52714138, orientation='plus', exon_count=3, OMIM=614625, no_hdr0=''),
  57292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57292, CurrentID=0, Status='live', Symbol='KIR2DL5A', Aliases='CD158F, KIR2DL5, KIR2DL5.1, KIR2DL5.3', description='killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 5A', other_designations='killer cell immunoglobulin-like receptor KIR2DL5A|killer cell immunoglobulin-like receptor, two domains, long cytoplasmic tail, 5|killer-cell Ig-like receptor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605305, no_hdr0=''),
  102031319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102031319, CurrentID=0, Status='live', Symbol='LOC102031319', Aliases='', description='uncharacterized LOC102031319', other_designations='', map_location='10p11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32347364, end_position_on_the_genomic_accession=32374488, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57306, CurrentID=0, Status='live', Symbol='CUP2Q35', Aliases='C2DUPq35, SD1, SDTY1', description='Syndactyly, type I', other_designations='', map_location='2q34-q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=185900, no_hdr0=''),
  140623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140623, CurrentID=0, Status='live', Symbol='FAUP1', Aliases='FAU1P, RPS30_2_1613', description='FBR-MuSV-associated ubiquitously expressed (fox derived) pseudogene 1', other_designations='FBR-MuSV-associated ubiquitously expressed gene pseudogene 1|ubiquitin-like protein fubi and ribosomal protein S30 precursor pseudogene 1|ubiquitin-like-S30 fusion protein pseudogene 1', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74389823, end_position_on_the_genomic_accession=74390413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57309, CurrentID=0, Status='live', Symbol='SPG14', Aliases='', description='spastic paraplegia 14 (autosomal recessive)', other_designations='', map_location='3q27-q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605229, no_hdr0=''),
  643067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643067, CurrentID=0, Status='live', Symbol='LOC643067', Aliases='', description='HORMA domain containing 1 pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72598242, end_position_on_the_genomic_accession=72599690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140625, CurrentID=0, Status='live', Symbol='ACTRT2', Aliases='ARPM2, ARPT2, Arp-T2, HARPM2', description='actin related protein T2', other_designations='actin-related protein M2|actin-related protein hArpM2', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3021482, end_position_on_the_genomic_accession=3022903, orientation='plus', exon_count=1, OMIM=608535, no_hdr0=''),
  57324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57324, CurrentID=0, Status='live', Symbol='TSE2', Aliases='', description='transsuppressor of expression 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601136, no_hdr0=''),
  57325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57325, CurrentID=0, Status='live', Symbol='CSRP2BP', Aliases='ATAC2, CRP2BP, KAT14, PRO1194, dJ717M23.1', description='CSRP2 binding protein', other_designations='ADA2A-containing complex subunit 2|ATAC component 2 homolog|CRP2 binding partner|CRP2 binding protein|CSRP2-binding protein', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18142224, end_position_on_the_genomic_accession=18188387, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57326, CurrentID=0, Status='live', Symbol='PBXIP1', Aliases='HPIP', description='pre-B-cell leukemia homeobox interacting protein 1', other_designations='hematopoietic PBX-interacting protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154944080, end_position_on_the_genomic_accession=154956091, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  57332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57332, CurrentID=0, Status='live', Symbol='CBX8', Aliases='PC3, RC1', description='chromobox 8', other_designations='Pc class 3 homolog|chromobox homolog 8 (Pc class homolog, Drosophila)|polycomb 3|rectachrome 1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79794377, end_position_on_the_genomic_accession=79797116, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57333, CurrentID=0, Status='live', Symbol='RCN3', Aliases='RLP49', description='reticulocalbin 3', other_designations='EF-hand calcium-binding protein RLP49|reticulocabin|reticulocalbin 3, EF-hand calcium binding domain', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49527579, end_position_on_the_genomic_accession=49543638, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57335, CurrentID=0, Status='live', Symbol='ZNF286A', Aliases='ZNF286', description='zinc finger protein 286A', other_designations='zinc finger protein ZNF286', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15699577, end_position_on_the_genomic_accession=15720787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57336, CurrentID=0, Status='live', Symbol='ZNF287', Aliases='ZKSCAN13, ZSCAN45', description='zinc finger protein 287', other_designations='zinc finger protein with KRAB and SCAN domains 13', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16549913, end_position_on_the_genomic_accession=16569206, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57337, CurrentID=0, Status='live', Symbol='SENP7', Aliases='', description='SUMO1/sentrin specific peptidase 7', other_designations='SUMO-1-specific protease 2|sentrin/SUMO-specific protease SENP7', map_location='3q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101324189, end_position_on_the_genomic_accession=101513883, orientation='minus', exon_count=27, OMIM=612846, no_hdr0=''),
  57338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57338, CurrentID=0, Status='live', Symbol='JPH3', Aliases='CAGL237, HDL2, JP-3, JP3, TNRC22', description='junctophilin 3', other_designations='junctophilin type 3|trinucleotide repeat containing 22', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87601835, end_position_on_the_genomic_accession=87698156, orientation='plus', exon_count=9, OMIM=605268, no_hdr0=''),
  57343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57343, CurrentID=0, Status='live', Symbol='ZNF304', Aliases='', description='zinc finger protein 304', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57351245, end_position_on_the_genomic_accession=57359898, orientation='plus', exon_count=6, OMIM=613840, no_hdr0=''),
  643072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643072, CurrentID=0, Status='live', Symbol='LOC643072', Aliases='', description='uncharacterized LOC643072', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159615294, end_position_on_the_genomic_accession=159618042, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57348, CurrentID=0, Status='live', Symbol='TTYH1', Aliases='', description='tweety family member 1', other_designations='hTTY1|tweety homolog 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54415426, end_position_on_the_genomic_accession=54436719, orientation='plus', exon_count=15, OMIM=605784, no_hdr0=''),
  105369601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369601, CurrentID=0, Status='live', Symbol='LOC105369601', Aliases='', description='uncharacterized LOC105369601', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1812847, end_position_on_the_genomic_accession=1817525, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57369, CurrentID=0, Status='live', Symbol='GJD2', Aliases='CX36, GJA9', description='gap junction protein delta 2', other_designations='connexin 36|connexin-36|gap junction alpha-9 protein|gap junction protein, delta 2, 36kDa', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34752441, end_position_on_the_genomic_accession=34754581, orientation='minus', exon_count=2, OMIM=607058, no_hdr0=''),
  402778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402778, CurrentID=0, Status='live', Symbol='IFITM10', Aliases='DSPA3', description='interferon induced transmembrane protein 10', other_designations='dispanin subfamily A member 3', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1732410, end_position_on_the_genomic_accession=1750594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57379, CurrentID=0, Status='live', Symbol='AICDA', Aliases='AID, ARP2, CDA2, HEL-S-284, HIGM2', description='activation-induced cytidine deaminase', other_designations="cytidine aminohydrolase|epididymis secretory protein Li 284|integrated into Burkitt's lymphoma cell line Ramos", map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8602166, end_position_on_the_genomic_accession=8612888, orientation='minus', exon_count=5, OMIM=605257, no_hdr0=''),
  57380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57380, CurrentID=0, Status='live', Symbol='MRS2', Aliases='HPTL, MRS2', description='MRS2, magnesium transporter', other_designations='MRS2 magnesium transporter|MRS2-like protein|MRS2-like, magnesium homeostasis factor', map_location='6p22.3-p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24402908, end_position_on_the_genomic_accession=24426194, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57381, CurrentID=0, Status='live', Symbol='RHOJ', Aliases='ARHJ, RASL7B, TC10B, TCL', description='ras homolog family member J', other_designations='RAS-like, family 7, member B|TC10-like Rho GTPase|ras homolog gene family, member J|ras-like protein family member 7B|tc10-like GTP-binding protein', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63204362, end_position_on_the_genomic_accession=63293512, orientation='plus', exon_count=7, OMIM=607653, no_hdr0=''),
  105369607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369607, CurrentID=0, Status='live', Symbol='LOC105369607', Aliases='', description='uncharacterized LOC105369607', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3419260, end_position_on_the_genomic_accession=3432395, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57393, CurrentID=0, Status='live', Symbol='TMEM27', Aliases='NX-17, NX17', description='transmembrane protein 27', other_designations='kidney-specific membrane protein', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15627316, end_position_on_the_genomic_accession=15665031, orientation='minus', exon_count=6, OMIM=300631, no_hdr0=''),
  57396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57396, CurrentID=0, Status='live', Symbol='CLK4', Aliases='', description='CDC like kinase 4', other_designations='CDC-like kinase 4|protein serine threonine kinase Clk4', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178602664, end_position_on_the_genomic_accession=178630639, orientation='minus', exon_count=17, OMIM=607969, no_hdr0=''),
  57399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57399, CurrentID=0, Status='live', Symbol='LOC57399', Aliases='', description='uncharacterized gastric protein ZA52P', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=60542152, end_position_on_the_genomic_accession=60542577, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  285605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285605, CurrentID=0, Status='live', Symbol='DTWD2', Aliases='', description='DTW domain containing 2', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118836074, end_position_on_the_genomic_accession=118988545, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57402, CurrentID=0, Status='live', Symbol='S100A14', Aliases='BCMP84, S100A15', description='S100 calcium binding protein A14', other_designations='S100 calcium-binding protein A14|S114|breast cancer membrane protein 84', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153614255, end_position_on_the_genomic_accession=153616332, orientation='minus', exon_count=5, OMIM=607986, no_hdr0=''),
  57403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57403, CurrentID=0, Status='live', Symbol='RAB22A', Aliases='', description='RAB22A, member RAS oncogene family', other_designations='GTP-binding protein RAB22A|rab-22', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58309715, end_position_on_the_genomic_accession=58367507, orientation='plus', exon_count=7, OMIM=612966, no_hdr0=''),
  57404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57404, CurrentID=0, Status='live', Symbol='CYP20A1', Aliases='CYP-M', description='cytochrome P450 family 20 subfamily A member 1', other_designations='cytochrome P450 monooxygenase|cytochrome P450, family 20, subfamily A, polypeptide 1', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203238441, end_position_on_the_genomic_accession=203305840, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  57405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57405, CurrentID=0, Status='live', Symbol='SPC25', Aliases='AD024, SPBC25, hSpc25', description='SPC25, NDC80 kinetochore complex component', other_designations='2600017H08Rik|SPC25, NDC80 kinetochore complex component, homolog|spindle pole body component 25 homolog', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168865144, end_position_on_the_genomic_accession=168890434, orientation='minus', exon_count=8, OMIM=609395, no_hdr0=''),
  57406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57406, CurrentID=0, Status='live', Symbol='ABHD6', Aliases='', description='abhydrolase domain containing 6', other_designations='2-arachidonoylglycerol hydrolase|abhydrolase domain-containing protein 6|lipase protein', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58237506, end_position_on_the_genomic_accession=58294738, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57407, CurrentID=0, Status='live', Symbol='NMRAL1', Aliases='HSCARG, SDR48A1', description='NmrA-like family domain containing 1', other_designations='short chain dehydrogenase/reductase family 48A, member 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4461689, end_position_on_the_genomic_accession=4474895, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57408, CurrentID=0, Status='live', Symbol='LRTM1', Aliases='HT017', description='leucine-rich repeats and transmembrane domains 1', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54918237, end_position_on_the_genomic_accession=54967088, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57409, CurrentID=0, Status='live', Symbol='MIF4GD', Aliases='AD023, MIFD, SLIP1', description='MIF4G domain containing', other_designations='SLBP (stem loop binding protein)-interacting protein 1|SLBP-interacting protein 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75266228, end_position_on_the_genomic_accession=75271292, orientation='minus', exon_count=11, OMIM=612072, no_hdr0=''),
  57410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57410, CurrentID=0, Status='live', Symbol='SCYL1', Aliases='GKLP, HT019, NKTL, NTKL, P105, TAPK, TEIF, TRAP', description='SCY1-like, kinase-like 1', other_designations='SCY1-like 1|SCY1-like protein 1|coated vesicle-associated kinase of 90 kDa|likely ortholog of mouse N-terminal kinase-like protein|telomerase regulation-associated protein|telomerase transcriptional element-interacting factor|telomerase transcriptional elements-interacting factor|teratoma-associated tyrosine kinase', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65525065, end_position_on_the_genomic_accession=65538711, orientation='plus', exon_count=18, OMIM=607982, no_hdr0=''),
  57412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57412, CurrentID=0, Status='live', Symbol='AS3MT', Aliases='CYT19', description='arsenite methyltransferase', other_designations='S-adenosyl-L-methionine:arsenic(III) methyltransferase|S-adenosylmethionine:arsenic (III) methyltransferase|arsenic (+3 oxidation state) methyltransferase|methylarsonite methyltransferase|methyltransferase cyt19', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102869453, end_position_on_the_genomic_accession=102901899, orientation='plus', exon_count=11, OMIM=611806, no_hdr0=''),
  57413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57413, CurrentID=0, Status='live', Symbol='TMIGD3', Aliases='AD026', description='transmembrane and immunoglobulin domain containing 3', other_designations='AD026 protein (AD026)|transmembrane domain-containing protein TMIGD3', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111483348, end_position_on_the_genomic_accession=111564003, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57414, CurrentID=0, Status='live', Symbol='RHBDD2', Aliases='NPD007, RHBDL7', description='rhomboid domain containing 2', other_designations='H_RG122E10.2a|H_RG122E10.2b|WUGSC:H_RG122E10.2a|WUGSC:H_RG122E10.2b|rhomboid, veinlet-like 7', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75878999, end_position_on_the_genomic_accession=75888926, orientation='plus', exon_count=5, OMIM=615203, no_hdr0=''),
  57415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57415, CurrentID=0, Status='live', Symbol='C3orf14', Aliases='HT021', description='chromosome 3 open reading frame 14', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62318973, end_position_on_the_genomic_accession=62336213, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57418, CurrentID=0, Status='live', Symbol='WDR18', Aliases='Ipi3, R32184_1', description='WD repeat domain 18', other_designations='Involved in Processing ITS2 3 homolog', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=984328, end_position_on_the_genomic_accession=994570, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57419, CurrentID=0, Status='live', Symbol='SLC24A3', Aliases='NCKX3', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 3', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 3|sodium calcium exchanger|solute carrier family 24 member 3', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19212646, end_position_on_the_genomic_accession=19722897, orientation='plus', exon_count=17, OMIM=609839, no_hdr0=''),
  643085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643085, CurrentID=0, Status='live', Symbol='LOC643085', Aliases='', description='uncharacterized LOC643085', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=96190718, end_position_on_the_genomic_accession=96192451, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378890, CurrentID=0, Status='live', Symbol='LOC105378890', Aliases='', description='uncharacterized LOC105378890', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108661626, end_position_on_the_genomic_accession=108667195, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369615, CurrentID=0, Status='live', Symbol='LOC105369615', Aliases='', description='uncharacterized LOC105369615', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5076992, end_position_on_the_genomic_accession=5090256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57446, CurrentID=0, Status='live', Symbol='NDRG3', Aliases='', description='NDRG family member 3', other_designations='N-myc downstream-regulated gene 3 protein', map_location='20q11.21-q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36651766, end_position_on_the_genomic_accession=36746138, orientation='minus', exon_count=19, OMIM=605273, no_hdr0=''),
  57447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57447, CurrentID=0, Status='live', Symbol='NDRG2', Aliases='SYLD', description='NDRG family member 2', other_designations='N-myc downstream regulator 2|N-myc downstream-regulated gene 2 protein|NDR1-related protein NDR2|cytoplasmic protein Ndr1|syld709613 protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21016763, end_position_on_the_genomic_accession=21070872, orientation='minus', exon_count=24, OMIM=605272, no_hdr0=''),
  57448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57448, CurrentID=0, Status='live', Symbol='BIRC6', Aliases='APOLLON, BRUCE', description='baculoviral IAP repeat containing 6', other_designations='BIR repeat-containing ubiquitin-conjugating enzyme|baculoviral IAP repeat-containing 6|ubiquitin-conjugating BIR domain enzyme apollon|ubiquitin-conjugating BIR-domain enzyme apollon', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32357028, end_position_on_the_genomic_accession=32618899, orientation='plus', exon_count=78, OMIM=605638, no_hdr0=''),
  57449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57449, CurrentID=0, Status='live', Symbol='PLEKHG5', Aliases='CMTRIC, DSMA4, GEF720, Syx, Tech', description='pleckstrin homology and RhoGEF domain containing G5', other_designations='NFkB activating protein|PH domain-containing family G member 5|guanine nucleotide exchange factor 720|novel PH domain-containing protein|pleckstrin homology domain containing family G member 5|pleckstrin homology domain containing, family G (with RhoGef domain) member 5|synectin-binding guanine exchange factor', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6466092, end_position_on_the_genomic_accession=6520061, orientation='minus', exon_count=28, OMIM=611101, no_hdr0=''),
  57451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57451, CurrentID=0, Status='live', Symbol='TENM2', Aliases='ODZ2, TEN-M2, TNM2, ten-2', description='teneurin transmembrane protein 2', other_designations='neurestin alpha|odd Oz/ten-m homolog 2|odz, odd Oz/ten-m homolog 2|protein Odd Oz/ten-m homolog 2|tenascin-M2', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166979078, end_position_on_the_genomic_accession=168264157, orientation='plus', exon_count=33, OMIM=610119, no_hdr0=''),
  57452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57452, CurrentID=0, Status='live', Symbol='GALNT16', Aliases='GALNACT16, GALNTL1, GalNAc-T16', description='polypeptide N-acetylgalactosaminyltransferase 16', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 1|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 16|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 1|galNAc-T-like protein 1|polypeptide GalNAc transferase 16|polypeptide GalNAc transferase-like protein 1|polypeptide N-acetylgalactosaminyltransferase-like protein 1|pp-GaNTase-like protein 1|protein-UDP acetylgalactosaminyltransferase-like protein 1|putative polypeptide N-acetylgalactosaminyltransferase-like protein 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69259546, end_position_on_the_genomic_accession=69379904, orientation='plus', exon_count=19, OMIM=615132, no_hdr0=''),
  57453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57453, CurrentID=0, Status='live', Symbol='DSCAML1', Aliases='DSCAM2', description='Down syndrome cell adhesion molecule like 1', other_designations='DSCAM-like 1|Downs syndrome cell adhesion molecule like 1|down syndrome cell adhesion molecule 2', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117427772, end_position_on_the_genomic_accession=117797261, orientation='minus', exon_count=38, OMIM=611782, no_hdr0=''),
  57455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57455, CurrentID=0, Status='live', Symbol='REXO1', Aliases='ELOABP1, EloA-BP1, REX1, TCEB3BP1', description='REX1, RNA exonuclease 1 homolog', other_designations='elongin A binding protein 1|elongin A-binding protein 1|elongin-A-binding protein 1|transcription elongation factor B polypeptide 3 binding protein 1|transcription elongation factor B polypeptide 3-binding protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1815246, end_position_on_the_genomic_accession=1848488, orientation='minus', exon_count=18, OMIM=609614, no_hdr0=''),
  57456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57456, CurrentID=0, Status='live', Symbol='KIAA1143', Aliases='', description='KIAA1143', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44748744, end_position_on_the_genomic_accession=44761697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57458, CurrentID=0, Status='live', Symbol='TMCC3', Aliases='', description='transmembrane and coiled-coil domain family 3', other_designations='transmembrane and coiled-coil domains 3', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94567124, end_position_on_the_genomic_accession=94650562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57459, CurrentID=0, Status='live', Symbol='GATAD2B', Aliases='MRD18, P66beta, p68', description='GATA zinc finger domain containing 2B', other_designations='GATA zinc finger domain-containing protein 2B|p66/p68|transcription repressor p66 beta component of the MeCP1 complex', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153804725, end_position_on_the_genomic_accession=153922975, orientation='minus', exon_count=14, OMIM=614998, no_hdr0=''),
  57460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57460, CurrentID=0, Status='live', Symbol='PPM1H', Aliases='ARHCL1, NERPP-2C, URCC2', description='protein phosphatase, Mg2+/Mn2+ dependent 1H', other_designations='neurite extension-related protein phosphatase related to PP2C|protein phosphatase 1H (PP2C domain containing)|protein phosphatase, Mg2+/Mn2+ dependent, 1H|ras homolog gene family, member C like 1', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62643982, end_position_on_the_genomic_accession=62934885, orientation='minus', exon_count=12, OMIM=616016, no_hdr0=''),
  57461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57461, CurrentID=0, Status='live', Symbol='ISY1', Aliases='FSAP33', description='ISY1 splicing factor homolog', other_designations='functional spliceosome-associated protein 33', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129127416, end_position_on_the_genomic_accession=129161230, orientation='minus', exon_count=12, OMIM=612764, no_hdr0=''),
  57462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57462, CurrentID=0, Status='live', Symbol='KIAA1161', Aliases='NET37', description='KIAA1161', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34366666, end_position_on_the_genomic_accession=34376896, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57463, CurrentID=0, Status='live', Symbol='AMIGO1', Aliases='ALI2, AMIGO, AMIGO-1', description='adhesion molecule with Ig-like domain 1', other_designations='alivin-2|amphoterin-induced gene and ORF', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109504178, end_position_on_the_genomic_accession=109509717, orientation='minus', exon_count=2, OMIM=615689, no_hdr0=''),
  57464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57464, CurrentID=0, Status='live', Symbol='STRIP2', Aliases='FAM40B, FAR11B', description='striatin interacting protein 2', other_designations='FAR11 factor arrest 11 homolog B|family with sequence similarity 40, member B|homolog of yeast FAR11 protein 2|protein FAM40B', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129434433, end_position_on_the_genomic_accession=129488399, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  57465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57465, CurrentID=0, Status='live', Symbol='TBC1D24', Aliases='DFNA65, DFNB86, DOORS, EIEE16, FIME, TLDC6', description='TBC1 domain family member 24', other_designations='TBC/LysM-associated domain containing 6|TBC1 domain family, member 24|skywalker homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2475146, end_position_on_the_genomic_accession=2505733, orientation='plus', exon_count=8, OMIM=613577, no_hdr0=''),
  57466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57466, CurrentID=0, Status='live', Symbol='SCAF4', Aliases='SFRS15, SRA4', description='SR-related CTD-associated factor 4', other_designations='CTD-binding SR-like protein rA4|SR-like CTD-associated factor 4|SR-related and CTD-associated factor 4|pre-mRNA splicing SR protein rA4|splicing factor serine alanine 15', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31671000, end_position_on_the_genomic_accession=31732118, orientation='minus', exon_count=20, OMIM=616023, no_hdr0=''),
  57467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57467, CurrentID=0, Status='live', Symbol='HHATL', Aliases='C3orf3, GUP1, MBOAT3, MSTP002, OACT3', description='hedgehog acyltransferase-like', other_designations='GUP1 glycerol uptake/transporter homolog|glycerol uptake/transporter homolog|hedgehog acyltransferase-like protein|membrane bound O-acyltransferase domain containing 3', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42692663, end_position_on_the_genomic_accession=42702848, orientation='minus', exon_count=14, OMIM=608116, no_hdr0=''),
  57468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57468, CurrentID=0, Status='live', Symbol='SLC12A5', Aliases='KCC2', description='solute carrier family 12 (potassium/chloride transporter), member 5', other_designations='K-Cl cotransporter 2|electroneutral potassium-chloride cotransporter 2|erythroid K-Cl cotransporter 2|hKCC2|neuronal K-Cl cotransporter', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46021690, end_position_on_the_genomic_accession=46060150, orientation='plus', exon_count=27, OMIM=606726, no_hdr0=''),
  57469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57469, CurrentID=0, Status='live', Symbol='PNMAL2', Aliases='', description='paraneoplastic Ma antigen family-like 2', other_designations='PNMA-like 2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46491191, end_position_on_the_genomic_accession=46495912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57470, CurrentID=0, Status='live', Symbol='LRRC47', Aliases='', description='leucine rich repeat containing 47', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3780220, end_position_on_the_genomic_accession=3796504, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57471, CurrentID=0, Status='live', Symbol='ERMN', Aliases='JN, KIAA1189', description='ermin', other_designations='ermin, ERM-like protein|juxtanodin', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157318613, end_position_on_the_genomic_accession=157327713, orientation='minus', exon_count=5, OMIM=610072, no_hdr0=''),
  57472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57472, CurrentID=0, Status='live', Symbol='CNOT6', Aliases='CCR4, Ccr4a', description='CCR4-NOT transcription complex subunit 6', other_designations='CCR4-NOT transcription complex, subunit 6|carbon catabolite repression 4 protein|carbon catabolite repressor protein 4 homolog|cytoplasmic deadenylase', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180494373, end_position_on_the_genomic_accession=180578405, orientation='plus', exon_count=14, OMIM=608951, no_hdr0=''),
  57473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57473, CurrentID=0, Status='live', Symbol='ZNF512B', Aliases='GM632', description='zinc finger protein 512B', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63956704, end_position_on_the_genomic_accession=63969870, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  57474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57474, CurrentID=0, Status='live', Symbol='ZNF490', Aliases='', description='zinc finger protein 490', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12576106, end_position_on_the_genomic_accession=12610809, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57475, CurrentID=0, Status='live', Symbol='PLEKHH1', Aliases='', description='pleckstrin homology, MyTH4 and FERM domain containing H1', other_designations='PH domain-containing family H member 1|pleckstrin homology domain containing, family H (with MyTH4 domain) member 1', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67533199, end_position_on_the_genomic_accession=67589538, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  57476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57476, CurrentID=0, Status='live', Symbol='GRAMD1B', Aliases='', description='GRAM domain containing 1B', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123358784, end_position_on_the_genomic_accession=123627771, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  57477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57477, CurrentID=0, Status='live', Symbol='SHROOM4', Aliases='SHAP', description='shroom family member 4', other_designations='second homolog of apical protein', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50576993, end_position_on_the_genomic_accession=50814139, orientation='minus', exon_count=13, OMIM=300579, no_hdr0=''),
  57478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57478, CurrentID=0, Status='live', Symbol='USP31', Aliases='', description='ubiquitin specific peptidase 31', other_designations='deubiquitinating enzyme 31|ubiquitin specific proteinase 31|ubiquitin thioesterase 31|ubiquitin thiolesterase 31|ubiquitin-specific-processing protease 31', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23061406, end_position_on_the_genomic_accession=23149270, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  57479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57479, CurrentID=0, Status='live', Symbol='PRR12', Aliases='KIAA1205', description='proline rich 12', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49591655, end_position_on_the_genomic_accession=49626439, orientation='plus', exon_count=14, OMIM=616633, no_hdr0=''),
  57480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57480, CurrentID=0, Status='live', Symbol='PLEKHG1', Aliases='ARHGEF41', description='pleckstrin homology and RhoGEF domain containing G1', other_designations='pleckstrin homology domain containing, family G (with RhoGef domain) member 1', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150599863, end_position_on_the_genomic_accession=150843665, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  57481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57481, CurrentID=0, Status='live', Symbol='KIAA1210', Aliases='', description='KIAA1210', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119078635, end_position_on_the_genomic_accession=119150579, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  57482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57482, CurrentID=0, Status='live', Symbol='KIAA1211', Aliases='', description='KIAA1211', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56049229, end_position_on_the_genomic_accession=56330724, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  57484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57484, CurrentID=0, Status='live', Symbol='RNF150', Aliases='', description='ring finger protein 150', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140859807, end_position_on_the_genomic_accession=141212869, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  57486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57486, CurrentID=0, Status='live', Symbol='NLN', Aliases='AGTBP, EP24.16, MEP, MOP', description='neurolysin (metallopeptidase M3 family)', other_designations='angiotensin binding protein|angiotensin-binding protein|endopeptidase 24.16|microsomal endopeptidase|mitochondrial oligopeptidase M|neurotensin endopeptidase', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65722196, end_position_on_the_genomic_accession=65829283, orientation='plus', exon_count=15, OMIM=611530, no_hdr0=''),
  57488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57488, CurrentID=0, Status='live', Symbol='ESYT2', Aliases='CHR2SYT, E-Syt2, FAM62B', description='extended synaptotagmin-like protein 2', other_designations='chr2 synaptotagmin|family with sequence similarity 62 (C2 domain containing), member B', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158730998, end_position_on_the_genomic_accession=158829628, orientation='minus', exon_count=22, OMIM=616691, no_hdr0=''),
  57489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57489, CurrentID=0, Status='live', Symbol='ODF2L', Aliases='dJ977L11.1', description='outer dense fiber of sperm tails 2-like', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86346824, end_position_on_the_genomic_accession=86396342, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  57491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57491, CurrentID=0, Status='live', Symbol='AHRR', Aliases='AHH, AHHR, bHLHe77', description='aryl-hydrocarbon receptor repressor', other_designations='ahR repressor|aryl hydrocarbon hydroxylase regulator|class E basic helix-loop-helix protein 77|dioxin receptor repressor', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=304176, end_position_on_the_genomic_accession=438291, orientation='plus', exon_count=12, OMIM=606517, no_hdr0=''),
  57492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57492, CurrentID=0, Status='live', Symbol='ARID1B', Aliases='6A3-5, BAF250B, BRIGHT, DAN15, ELD/OSA1, MRD12, OSA2, P250R', description='AT-rich interaction domain 1B', other_designations='ARID domain-containing protein 1B|AT rich interactive domain 1B (SWI1-like)|BRG1-associated factor 250b|BRG1-binding protein ELD/OSA1|ELD (eyelid)/OSA protein', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156777892, end_position_on_the_genomic_accession=157210779, orientation='plus', exon_count=26, OMIM=614556, no_hdr0=''),
  57493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57493, CurrentID=0, Status='live', Symbol='HEG1', Aliases='HEG, MST112, MSTP112', description='heart development protein with EGF-like domains 1', other_designations='HEG homolog 1|heart of glass', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124965710, end_position_on_the_genomic_accession=125055958, orientation='minus', exon_count=18, OMIM=614182, no_hdr0=''),
  57494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57494, CurrentID=0, Status='live', Symbol='RIMKLB', Aliases='FAM80B, NAAGS, NAAGS-I', description='ribosomal modification protein rimK-like family member B', other_designations='N-acetyl-aspartyl-glutamate synthetase B|N-acetyl-aspartylglutamate synthetase B|N-acetylaspartyl-glutamate synthetase B|NAAG synthetase B|beta-citryl-glutamate synthase B|family with sequence similarity 80, member B|ribosomal protein S6 modification-like protein B', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8681611, end_position_on_the_genomic_accession=8783098, orientation='plus', exon_count=13, OMIM=614054, no_hdr0=''),
  57495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57495, CurrentID=0, Status='live', Symbol='NWD2', Aliases='KIAA1239', description='NACHT and WD repeat domain containing 2', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37245068, end_position_on_the_genomic_accession=37449465, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57496, CurrentID=0, Status='live', Symbol='MKL2', Aliases='MRTF-B, MRTFB, NPD001', description='MKL/myocardin-like 2', other_designations='megakaryoblastic leukemia 2|myocardin-related transcription factor B', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14071321, end_position_on_the_genomic_accession=14266776, orientation='plus', exon_count=22, OMIM=609463, no_hdr0=''),
  57497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57497, CurrentID=0, Status='live', Symbol='LRFN2', Aliases='FIGLER2, KIAA1246, SALM1', description='leucine rich repeat and fibronectin type III domain containing 2', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 2|synaptic adhesion-like molecule 1', map_location='6p21.2-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40391099, end_position_on_the_genomic_accession=40587464, orientation='minus', exon_count=11, OMIM=612808, no_hdr0=''),
  57498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57498, CurrentID=0, Status='live', Symbol='KIDINS220', Aliases='ARMS', description='kinase D-interacting substrate 220kDa', other_designations='ankyrin repeat-rich membrane spanning|ankyrin repeat-rich membrane-spanning protein|kinase D-interacting substrate, 220kDa|truncated KIDINS220', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8728857, end_position_on_the_genomic_accession=8837625, orientation='minus', exon_count=30, OMIM=615759, no_hdr0=''),
  57501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57501, CurrentID=0, Status='live', Symbol='KIAA1257', Aliases='', description='KIAA1257', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128910186, end_position_on_the_genomic_accession=129026519, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  57502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57502, CurrentID=0, Status='live', Symbol='NLGN4X', Aliases='ASPGX2, AUTSX2, HLNX, HNL4X, NLGN4', description='neuroligin 4, X-linked', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=5890026, end_position_on_the_genomic_accession=6228882, orientation='minus', exon_count=12, OMIM=300427, no_hdr0=''),
  57504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57504, CurrentID=0, Status='live', Symbol='MTA3', Aliases='', description='metastasis associated 1 family member 3', other_designations='metastasis associated 1 family, member 3|metastasis associated family, member 3|metastasis associated gene family, member 3', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42494569, end_position_on_the_genomic_accession=42795030, orientation='plus', exon_count=23, OMIM=609050, no_hdr0=''),
  57505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57505, CurrentID=0, Status='live', Symbol='AARS2', Aliases='AARSL, COXPD8, LKENP, MT-ALARS, MTALARS', description='alanyl-tRNA synthetase 2, mitochondrial', other_designations='alanine tRNA ligase 2, mitochondrial (putative)|alanyl-tRNA synthetase 2, mitochondrial (putative)|alanyl-tRNA synthetase like|probable alanyl-tRNA synthetase, mitochondrial', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44297826, end_position_on_the_genomic_accession=44313358, orientation='minus', exon_count=23, OMIM=612035, no_hdr0=''),
  57506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57506, CurrentID=0, Status='live', Symbol='MAVS', Aliases='CARDIF, IPS-1, IPS1, VISA', description='mitochondrial antiviral signaling protein', other_designations='CARD adapter inducing interferon beta|CARD adaptor inducing IFN-beta|IFN-B promoter stimulator 1|interferon beta promoter stimulator protein 1|putative NF-kappa-B-activating protein 031N|virus-induced signaling adaptor|virus-induced-signaling adapter', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3846799, end_position_on_the_genomic_accession=3876123, orientation='plus', exon_count=7, OMIM=609676, no_hdr0=''),
  57507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57507, CurrentID=0, Status='live', Symbol='ZNF608', Aliases='NY-REN-36', description='zinc finger protein 608', other_designations='renal carcinoma antigen NY-REN-36', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124636913, end_position_on_the_genomic_accession=124748824, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  57508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57508, CurrentID=0, Status='live', Symbol='INTS2', Aliases='INT2, KIAA1287', description='integrator complex subunit 2', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61865367, end_position_on_the_genomic_accession=61928030, orientation='minus', exon_count=27, OMIM=611346, no_hdr0=''),
  57509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57509, CurrentID=0, Status='live', Symbol='MTUS1', Aliases='ATBP, ATIP, ICIS, MP44, MTSG1', description='microtubule associated tumor suppressor 1', other_designations='AT2 receptor-binding protein|AT2 receptor-interacting protein|AT2R binding protein|angiotensin-II type 2 receptor-interacting protein|erythroid differentiation-related|mitochondrial tumor suppressor gene 1|transcription factor MTSG1', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17643794, end_position_on_the_genomic_accession=17801220, orientation='minus', exon_count=20, OMIM=609589, no_hdr0=''),
  57510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57510, CurrentID=0, Status='live', Symbol='XPO5', Aliases='exp5', description='exportin 5', other_designations='ran-binding protein 21', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43522330, end_position_on_the_genomic_accession=43576075, orientation='minus', exon_count=32, OMIM=607845, no_hdr0=''),
  57511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57511, CurrentID=0, Status='live', Symbol='COG6', Aliases='CDG2L, COD2, SHNS', description='component of oligomeric golgi complex 6', other_designations='COG complex subunit 6|complexed with Dor1p 2|conserved oligomeric Golgi complex protein 6', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39655627, end_position_on_the_genomic_accession=39791665, orientation='plus', exon_count=23, OMIM=606977, no_hdr0=''),
  57512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57512, CurrentID=0, Status='live', Symbol='GPR158', Aliases='', description='G protein-coupled receptor 158', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25174737, end_position_on_the_genomic_accession=25602233, orientation='plus', exon_count=12, OMIM=614573, no_hdr0=''),
  57513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57513, CurrentID=0, Status='live', Symbol='CASKIN2', Aliases='ANKS5B', description='CASK interacting protein 2', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75500260, end_position_on_the_genomic_accession=75515583, orientation='minus', exon_count=21, OMIM=612185, no_hdr0=''),
  57514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57514, CurrentID=0, Status='live', Symbol='ARHGAP31', Aliases='AOS1, CDGAP', description='Rho GTPase activating protein 31', other_designations='Cdc42 GTPase-activating protein', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119294289, end_position_on_the_genomic_accession=119419476, orientation='plus', exon_count=13, OMIM=610911, no_hdr0=''),
  57515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57515, CurrentID=0, Status='live', Symbol='SERINC1', Aliases='TDE1L, TDE2, TMS-2, TMS2', description='serine incorporator 1', other_designations='tumor differentially expressed protein 2', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=122443348, end_position_on_the_genomic_accession=122471807, orientation='minus', exon_count=10, OMIM=614548, no_hdr0=''),
  105378893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378893, CurrentID=0, Status='live', Symbol='LOC105378893', Aliases='', description='uncharacterized LOC105378893', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109510613, end_position_on_the_genomic_accession=109522443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57519, CurrentID=0, Status='live', Symbol='STARD9', Aliases='KIF16A', description='StAR related lipid transfer domain containing 9', other_designations='START domain containing 9|START domain-containing protein 9|StAR-related lipid transfer (START) domain containing 9|StAR-related lipid transfer domain containing 9', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42575634, end_position_on_the_genomic_accession=42720998, orientation='plus', exon_count=35, OMIM=614642, no_hdr0=''),
  57520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57520, CurrentID=0, Status='live', Symbol='HECW2', Aliases='NEDL2', description='HECT, C2 and WW domain containing E3 ubiquitin protein ligase 2', other_designations='HECT, C2 and WW domain-containing protein 2|NEDD4-like E3 ubiquitin-protein ligase 2|NEDD4-related E3 ubiquitin ligase NEDL2', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196199000, end_position_on_the_genomic_accession=196593692, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  57521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57521, CurrentID=0, Status='live', Symbol='RPTOR', Aliases='KOG1, Mip1', description='regulatory associated protein of MTOR, complex 1', other_designations='p150 target of rapamycin (TOR)-scaffold protein containing WD-repeats|raptor', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80544825, end_position_on_the_genomic_accession=80966373, orientation='plus', exon_count=34, OMIM=607130, no_hdr0=''),
  57522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57522, CurrentID=0, Status='live', Symbol='SRGAP1', Aliases='ARHGAP13', description='SLIT-ROBO Rho GTPase activating protein 1', other_designations='rho GTPase-activating protein 13', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63844512, end_position_on_the_genomic_accession=64147863, orientation='plus', exon_count=26, OMIM=606523, no_hdr0=''),
  57523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57523, CurrentID=0, Status='live', Symbol='NYNRIN', Aliases='CGIN1, KIAA1305', description='NYN domain and retroviral integrase containing', other_designations='Cousin of GIN1|NYN domain and retroviral integrase catalytic domain-containing protein|protein cousin of GIN1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24398786, end_position_on_the_genomic_accession=24419288, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57524, CurrentID=0, Status='live', Symbol='CASKIN1', Aliases='ANKS5A', description='CASK interacting protein 1', other_designations='CASK-interacting protein 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2177183, end_position_on_the_genomic_accession=2196464, orientation='minus', exon_count=20, OMIM=612184, no_hdr0=''),
  105369630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369630, CurrentID=0, Status='live', Symbol='LOC105369630', Aliases='', description='uncharacterized LOC105369630', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6568402, end_position_on_the_genomic_accession=6570386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57526, CurrentID=0, Status='live', Symbol='PCDH19', Aliases='EFMR, EIEE9', description='protocadherin 19', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100291644, end_position_on_the_genomic_accession=100410273, orientation='minus', exon_count=6, OMIM=300460, no_hdr0=''),
  57528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57528, CurrentID=0, Status='live', Symbol='KCTD16', Aliases='', description='potassium channel tetramerization domain containing 16', other_designations='potassium channel tetramerisation domain containing 16|potassium channel tetramerization domain-containing protein 16', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=144170760, end_position_on_the_genomic_accession=144485686, orientation='plus', exon_count=8, OMIM=613423, no_hdr0=''),
  57529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57529, CurrentID=0, Status='live', Symbol='RGAG1', Aliases='MAR9, MART9', description='retrotransposon gag domain containing 1', other_designations='tumor antigen BJ-HCC-23|tumor antigen BJHCC23', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110358776, end_position_on_the_genomic_accession=110456334, orientation='plus', exon_count=5, OMIM=300965, no_hdr0=''),
  57530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57530, CurrentID=0, Status='live', Symbol='CGN', Aliases='', description='cingulin', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151510877, end_position_on_the_genomic_accession=151538692, orientation='plus', exon_count=22, OMIM=609473, no_hdr0=''),
  57531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57531, CurrentID=0, Status='live', Symbol='HACE1', Aliases='', description='HECT domain and ankyrin repeat containing E3 ubiquitin protein ligase 1', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104682659, end_position_on_the_genomic_accession=104859919, orientation='minus', exon_count=30, OMIM=610876, no_hdr0=''),
  57532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57532, CurrentID=0, Status='live', Symbol='NUFIP2', Aliases='182-FIP, 82-FIP, FIP-82, PIG1', description='nuclear fragile X mental retardation protein interacting protein 2', other_designations='82 kDa FMRP-interacting protein|82-kD FMRP Interacting Protein|FMRP-interacting protein 2|Fragile X Mental Retardation Protein|cell proliferation-inducing gene 1 protein|proliferation-inducing gene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29255836, end_position_on_the_genomic_accession=29294148, orientation='minus', exon_count=6, OMIM=609356, no_hdr0=''),
  57533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57533, CurrentID=0, Status='live', Symbol='TBC1D14', Aliases='', description='TBC1 domain family member 14', other_designations='TBC1 domain family, member 14', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6909444, end_position_on_the_genomic_accession=7033118, orientation='plus', exon_count=20, OMIM=614855, no_hdr0=''),
  57534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57534, CurrentID=0, Status='live', Symbol='MIB1', Aliases='DIP-1, DIP1, LVNC7, MIB, ZZANK2, ZZZ6', description='mindbomb E3 ubiquitin protein ligase 1', other_designations='DAPK-interacting protein 1|ubiquitin ligase mind bomb|zinc finger ZZ type with ankyrin repeat domain protein 2', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21740793, end_position_on_the_genomic_accession=21870957, orientation='plus', exon_count=25, OMIM=608677, no_hdr0=''),
  57535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57535, CurrentID=0, Status='live', Symbol='KIAA1324', Aliases='EIG121', description='KIAA1324', other_designations='estrogen-induced gene 121 protein|maba1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109113963, end_position_on_the_genomic_accession=109206781, orientation='plus', exon_count=24, OMIM=611298, no_hdr0=''),
  57536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57536, CurrentID=0, Status='live', Symbol='KIAA1328', Aliases='', description='KIAA1328', other_designations='hinderin', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36829109, end_position_on_the_genomic_accession=37247316, orientation='plus', exon_count=26, OMIM=616480, no_hdr0=''),
  57537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57537, CurrentID=0, Status='live', Symbol='SORCS2', Aliases='', description='sortilin-related VPS10 domain containing receptor 2', other_designations='VPS10 domain receptor protein', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7192607, end_position_on_the_genomic_accession=7742837, orientation='plus', exon_count=30, OMIM=606284, no_hdr0=''),
  57538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57538, CurrentID=0, Status='live', Symbol='ALPK3', Aliases='MAK, MIDORI', description='alpha kinase 3', other_designations='alpha-kinase 3|muscle alpha-kinase|muscle alpha-protein kinase|myocyte induction differentiation originator', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84816680, end_position_on_the_genomic_accession=84873482, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  57539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57539, CurrentID=0, Status='live', Symbol='WDR35', Aliases='CED2, IFT121, IFTA1, SRTD7', description='WD repeat domain 35', other_designations='intraflagellar transport protein 121 homolog|naofen', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19910260, end_position_on_the_genomic_accession=19990123, orientation='minus', exon_count=29, OMIM=613602, no_hdr0=''),
  57540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57540, CurrentID=0, Status='live', Symbol='PTCHD2', Aliases='DISP3', description='patched domain containing 2', other_designations='dispatched 3|protein dispatched homolog 3', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11479238, end_position_on_the_genomic_accession=11537583, orientation='plus', exon_count=25, OMIM=611251, no_hdr0=''),
  57541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57541, CurrentID=0, Status='live', Symbol='ZNF398', Aliases='P51, P71, ZER6', description='zinc finger protein 398', other_designations='zinc finger DNA binding protein ZER6|zinc finger DNA binding protein p52/p71|zinc finger DNA-binding protein p52/p71|zinc finger-estrogen receptor interaction, clone 6', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149126416, end_position_on_the_genomic_accession=149183042, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57542, CurrentID=0, Status='live', Symbol='KLHL42', Aliases='Ctb9, KLHDC5', description='kelch like family member 42', other_designations='cullin-3-binding protein 9|kelch domain containing 5|kelch domain-containing protein 5|kelch-like family member 42|kelch-like family, member 42', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27779958, end_position_on_the_genomic_accession=27803040, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57544, CurrentID=0, Status='live', Symbol='TXNDC16', Aliases='ERp90, KIAA1344', description='thioredoxin domain containing 16', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52430590, end_position_on_the_genomic_accession=52552583, orientation='minus', exon_count=23, OMIM=616179, no_hdr0=''),
  57545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57545, CurrentID=0, Status='live', Symbol='CC2D2A', Aliases='JBTS9, MKS6', description='coiled-coil and C2 domain containing 2A', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15469861, end_position_on_the_genomic_accession=15601971, orientation='plus', exon_count=40, OMIM=612013, no_hdr0=''),
  57546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57546, CurrentID=0, Status='live', Symbol='PDP2', Aliases='PPM2C2', description='pyruvate dehyrogenase phosphatase catalytic subunit 2', other_designations='PDPC 2|[pyruvate dehydrogenase (acetyl-transferring)]-phosphatase|protein phosphatase 2C, magnesium-dependent, catalytic subunit 2|pyruvate dehydrogenase phosphatase catalytic subunit 2|pyruvate dehydrogenase phosphatase isoenzyme 2|pyruvate dehydrogenase phosphatase, catalytic subunit 2', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66880457, end_position_on_the_genomic_accession=66891101, orientation='plus', exon_count=3, OMIM=615499, no_hdr0=''),
  57547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57547, CurrentID=0, Status='live', Symbol='ZNF624', Aliases='', description='zinc finger protein 624', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16599870, end_position_on_the_genomic_accession=16655626, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57549, CurrentID=0, Status='live', Symbol='IGSF9', Aliases='FP18798A, Nrt1, IGSF9', description='immunoglobulin superfamily member 9', other_designations='immunoglobulin superfamily member 9A|immunoglobulin superfamily, member 9', map_location='1q22-q23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159927039, end_position_on_the_genomic_accession=159945596, orientation='minus', exon_count=23, OMIM=609738, no_hdr0=''),
  57551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57551, CurrentID=0, Status='live', Symbol='TAOK1', Aliases='KFC-B, MAP3K16, MARKK, PSK-2, PSK2, TAO1, hKFC-B, hTAOK1', description='TAO kinase 1', other_designations='MARK Kinase|STE20-like kinase PSK2|kinase from chicken homolog B|microtubule affinity regulating kinase kinase|prostate-derived STE20-like kinase 2|prostate-derived sterile 20-like kinase 2|serine/threonine protein kinase TAO1 homolog|thousand and one amino acid protein 1|thousand and one amino acid protein kinase 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29390925, end_position_on_the_genomic_accession=29551904, orientation='plus', exon_count=21, OMIM=610266, no_hdr0=''),
  57552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57552, CurrentID=0, Status='live', Symbol='NCEH1', Aliases='AADACL1, NCEH', description='neutral cholesterol ester hydrolase 1', other_designations='arylacetamide deacetylase-like 1', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172630645, end_position_on_the_genomic_accession=172711218, orientation='minus', exon_count=5, OMIM=613234, no_hdr0=''),
  57553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57553, CurrentID=0, Status='live', Symbol='MICAL3', Aliases='MICAL-3', description='microtubule associated monooxygenase, calponin and LIM domain containing 3', other_designations='flavoprotein oxidoreductase MICAL3|molecule interacting with CasL protein 3|protein MICAL-3', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17787649, end_position_on_the_genomic_accession=18024559, orientation='minus', exon_count=47, OMIM=608882, no_hdr0=''),
  57554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57554, CurrentID=0, Status='live', Symbol='LRRC7', Aliases='DENSIN', description='leucine rich repeat containing 7', other_designations='densin-180', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69568773, end_position_on_the_genomic_accession=70148192, orientation='plus', exon_count=29, OMIM=614453, no_hdr0=''),
  57555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57555, CurrentID=0, Status='live', Symbol='NLGN2', Aliases='', description='neuroligin 2', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7408183, end_position_on_the_genomic_accession=7419864, orientation='plus', exon_count=7, OMIM=606479, no_hdr0=''),
  57556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57556, CurrentID=0, Status='live', Symbol='SEMA6A', Aliases='HT018, SEMA1, SEMAQ, VIA, SEMA6A', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6A', other_designations='SEMA6A-1|sema VIa|semaphorin 6A-1|semaphorin VIA|semaphorin-6A-1', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116443555, end_position_on_the_genomic_accession=116574926, orientation='minus', exon_count=21, OMIM=605885, no_hdr0=''),
  57558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57558, CurrentID=0, Status='live', Symbol='USP35', Aliases='', description='ubiquitin specific peptidase 35', other_designations='deubiquitinating enzyme 35|ubiquitin specific protease 35|ubiquitin thioesterase 35|ubiquitin thiolesterase 35|ubiquitin-specific-processing protease 35', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78188812, end_position_on_the_genomic_accession=78217644, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  57559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57559, CurrentID=0, Status='live', Symbol='STAMBPL1', Aliases='ALMalpha, AMSH-FP, AMSH-LP, bA399O19.2', description='STAM binding protein like 1', other_designations='STAM binding protein-like 1|associated molecule with the SH3 domain of STAM (AMSH) - Family Protein|associated molecule with the SH3 domain of STAM (AMSH) like protein', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88879368, end_position_on_the_genomic_accession=88923502, orientation='plus', exon_count=17, OMIM=612352, no_hdr0=''),
  57560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57560, CurrentID=0, Status='live', Symbol='IFT80', Aliases='ATD2, SRTD2, WDR56', description='intraflagellar transport 80', other_designations='WD repeat domain 56|WD repeat-containing protein 56|intraflagellar transport 80 homolog', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160256986, end_position_on_the_genomic_accession=160399532, orientation='minus', exon_count=21, OMIM=611177, no_hdr0=''),
  57561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57561, CurrentID=0, Status='live', Symbol='ARRDC3', Aliases='TLIMP', description='arrestin domain containing 3', other_designations='TBP-2-like inducible membrane protein|alpha-arrestin 3|thioredoxin-binding protein-2-like inducible membrane', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91368724, end_position_on_the_genomic_accession=91383359, orientation='minus', exon_count=8, OMIM=612464, no_hdr0=''),
  57562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57562, CurrentID=0, Status='live', Symbol='CEP126', Aliases='KIAA1377', description='centrosomal protein 126kDa', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101915015, end_position_on_the_genomic_accession=102001065, orientation='plus', exon_count=12, OMIM=614634, no_hdr0=''),
  57563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57563, CurrentID=0, Status='live', Symbol='KLHL8', Aliases='', description='kelch like family member 8', other_designations='kelch-like 8|kelch-like family member 8', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87160103, end_position_on_the_genomic_accession=87220608, orientation='minus', exon_count=11, OMIM=611967, no_hdr0=''),
  57565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57565, CurrentID=0, Status='live', Symbol='KLHL14', Aliases='', description='kelch like family member 14', other_designations='kelch-like 14|kelch-like family member 14|printor|protein interactor of Torsin-1A', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32672668, end_position_on_the_genomic_accession=32773011, orientation='minus', exon_count=11, OMIM=613772, no_hdr0=''),
  57567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57567, CurrentID=0, Status='live', Symbol='ZNF319', Aliases='ZFP319', description='zinc finger protein 319', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57994668, end_position_on_the_genomic_accession=58000684, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57568, CurrentID=0, Status='live', Symbol='SIPA1L2', Aliases='SPAL2', description='signal-induced proliferation-associated 1 like 2', other_designations='SIPA1-like protein 2|SPA-1-like 2', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232397965, end_position_on_the_genomic_accession=232630696, orientation='minus', exon_count=27, OMIM=611609, no_hdr0=''),
  57569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57569, CurrentID=0, Status='live', Symbol='ARHGAP20', Aliases='RARHOGAP', description='Rho GTPase activating protein 20', other_designations='RA and RhoGAP domain containing protein|rho GTPase activating protein 20 variant 2|rho-type GTPase-activating protein 20', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110576855, end_position_on_the_genomic_accession=110713189, orientation='minus', exon_count=20, OMIM=609568, no_hdr0=''),
  57570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57570, CurrentID=0, Status='live', Symbol='TRMT5', Aliases='KIAA1393, TRM5', description='tRNA methyltransferase 5', other_designations='M1G-methyltransferase|TRM5 tRNA methyltransferase 5 homolog|tRNA (guanine-N(1)-)-methyltransferase|tRNA [GM37] methyltransferase|tRNA-N1G37 methyltransferase', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60971441, end_position_on_the_genomic_accession=60981170, orientation='minus', exon_count=5, OMIM=611023, no_hdr0=''),
  57571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57571, CurrentID=0, Status='live', Symbol='CARNS1', Aliases='ATPGD1', description='carnosine synthase 1', other_designations='ATP-grasp domain containing 1|ATP-grasp domain-containing protein 1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67415678, end_position_on_the_genomic_accession=67425607, orientation='plus', exon_count=11, OMIM=613368, no_hdr0=''),
  57572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57572, CurrentID=0, Status='live', Symbol='DOCK6', Aliases='AOS2, ZIR1', description='dedicator of cytokinesis 6', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11199293, end_position_on_the_genomic_accession=11262501, orientation='minus', exon_count=53, OMIM=614194, no_hdr0=''),
  57573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57573, CurrentID=0, Status='live', Symbol='ZNF471', Aliases='ERP1, Z1971', description='zinc finger protein 471', other_designations='EZFIT-related protein 1', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56507843, end_position_on_the_genomic_accession=56530212, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57574, CurrentID=0, Status='live', Symbol='MARCH4', Aliases='MARCH-IV, RNF174', description='membrane associated ring-CH-type finger 4', other_designations='RING finger protein 174|membrane associated ring finger 4|membrane-associated RING finger protein 4|membrane-associated RING-CH protein IV|membrane-associated ring finger (C3HC4) 4, E3 ubiquitin protein ligase', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216257862, end_position_on_the_genomic_accession=216372027, orientation='minus', exon_count=5, OMIM=608208, no_hdr0=''),
  57575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57575, CurrentID=0, Status='live', Symbol='PCDH10', Aliases='OL-PCDH, PCDH19', description='protocadherin 10', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=133149290, end_position_on_the_genomic_accession=133208606, orientation='plus', exon_count=6, OMIM=608286, no_hdr0=''),
  57576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57576, CurrentID=0, Status='live', Symbol='KIF17', Aliases='KIF17B, KIF3X, KLP-2, OSM-3', description='kinesin family member 17', other_designations='KIF17 variant protein|KIF3-related motor protein', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20661643, end_position_on_the_genomic_accession=20718017, orientation='minus', exon_count=19, OMIM=605037, no_hdr0=''),
  57577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57577, CurrentID=0, Status='live', Symbol='CCDC191', Aliases='KIAA1407', description='coiled-coil domain containing 191', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113964137, end_position_on_the_genomic_accession=114056641, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  57578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57578, CurrentID=0, Status='live', Symbol='UNC79', Aliases='KIAA1409', description='unc-79 homolog (C. elegans)', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93333219, end_position_on_the_genomic_accession=93707878, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  57579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57579, CurrentID=0, Status='live', Symbol='FAM135A', Aliases='KIAA1411', description='family with sequence similarity 135 member A', other_designations='family with sequence similarity 135, member A', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70413404, end_position_on_the_genomic_accession=70561174, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  57580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57580, CurrentID=0, Status='live', Symbol='PREX1', Aliases='P-REX1', description='phosphatidylinositol-3,4,5-trisphosphate-dependent Rac exchange factor 1', other_designations='ptdIns(3,4,5)-dependent Rac exchanger 1', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48624249, end_position_on_the_genomic_accession=48827883, orientation='minus', exon_count=41, OMIM=606905, no_hdr0=''),
  100301522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100301522, CurrentID=0, Status='live', Symbol='CELIAC7', Aliases='', description='Celiac disease, susceptibility to, 7', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612005, no_hdr0=''),
  57582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57582, CurrentID=0, Status='live', Symbol='KCNT1', Aliases='EIEE14, ENFL5, KCa4.1, SLACK, Slo2.2, bA100C15.2', description='potassium channel, sodium activated subfamily T, member 1', other_designations='Sequence like a calcium-activated K+ channel|potassium channel, subfamily T, member 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135702185, end_position_on_the_genomic_accession=135793147, orientation='plus', exon_count=35, OMIM=608167, no_hdr0=''),
  57583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57583, CurrentID=0, Status='live', Symbol='TMEM181', Aliases='GPR178, KIAA1423', description='transmembrane protein 181', other_designations='G protein-coupled receptor 178', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158536432, end_position_on_the_genomic_accession=158635435, orientation='plus', exon_count=20, OMIM=613209, no_hdr0=''),
  57584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57584, CurrentID=0, Status='live', Symbol='ARHGAP21', Aliases='ARHGAP10', description='Rho GTPase activating protein 21', other_designations='Rho-GTPase activating protein 10|rho GTPase-activating protein 10|rho-type GTPase-activating protein 21', map_location='10p12.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24583609, end_position_on_the_genomic_accession=24723668, orientation='minus', exon_count=31, OMIM=609870, no_hdr0=''),
  57585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57585, CurrentID=0, Status='live', Symbol='CRAMP1', Aliases='CRAMP1L, HN1L, TCE4', description='cramped chromatin regulator homolog 1', other_designations='Crm, cramped-like|T-complex expressed gene 4|hematological and neurological expressed 1-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1614640, end_position_on_the_genomic_accession=1677908, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  57586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57586, CurrentID=0, Status='live', Symbol='SYT13', Aliases='', description='synaptotagmin 13', other_designations='synaptotagmin XIII|sytXIII', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45240302, end_position_on_the_genomic_accession=45286333, orientation='minus', exon_count=8, OMIM=607716, no_hdr0=''),
  57587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57587, CurrentID=0, Status='live', Symbol='CFAP97', Aliases='KIAA1430, hmw', description='cilia and flagella associated protein 97', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185151723, end_position_on_the_genomic_accession=185209495, orientation='minus', exon_count=10, OMIM=616047, no_hdr0=''),
  57589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57589, CurrentID=0, Status='live', Symbol='RIC1', Aliases='CIP150, KIAA1432, bA207C16.1', description='RIC1 homolog, RAB6A GEF complex partner 1', other_designations='RAB6A GEF complex partner 1|connexin 43-interacting protein 150 kDa|connexin-43-interacting protein of 150 kDa|protein RIC1 homolog', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5629030, end_position_on_the_genomic_accession=5778015, orientation='plus', exon_count=28, OMIM=610354, no_hdr0=''),
  57590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57590, CurrentID=0, Status='live', Symbol='WDFY1', Aliases='FENS-1, WDF1, ZFYVE17', description='WD repeat and FYVE domain containing 1', other_designations='WD40- and FYVE domain-containing protein 1|phosphoinositide-binding protein SR1|zinc finger FYVE domain-containing protein 17', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223875343, end_position_on_the_genomic_accession=223945387, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57591, CurrentID=0, Status='live', Symbol='MKL1', Aliases='BSAC, MAL, MRTF-A', description='megakaryoblastic leukemia (translocation) 1', other_designations='basic, SAP and coiled-coil domain|megakaryoblastic leukemia 1 protein|megakaryocytic acute leukemia protein|myocardin-related transcription factor A', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40410281, end_position_on_the_genomic_accession=40636719, orientation='minus', exon_count=19, OMIM=606078, no_hdr0=''),
  57592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57592, CurrentID=0, Status='live', Symbol='ZNF687', Aliases='', description='zinc finger protein 687', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151281555, end_position_on_the_genomic_accession=151291905, orientation='plus', exon_count=13, OMIM=610568, no_hdr0=''),
  57593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57593, CurrentID=0, Status='live', Symbol='EBF4', Aliases='COE4, O/E-4', description='early B-cell factor 4', other_designations='EBF-4|OE-4|RP5-860F19.3|olf-1/EBF-like 4', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2692878, end_position_on_the_genomic_accession=2760108, orientation='plus', exon_count=18, OMIM=609935, no_hdr0=''),
  57594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57594, CurrentID=0, Status='live', Symbol='HOMEZ', Aliases='KIAA1443', description='homeobox and leucine zipper encoding', other_designations='homeodomain leucine zipper-containing factor', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23273635, end_position_on_the_genomic_accession=23286100, orientation='minus', exon_count=2, OMIM=608119, no_hdr0=''),
  57595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57595, CurrentID=0, Status='live', Symbol='PDZD4', Aliases='LU1, PDZK4, PDZRN4L', description='PDZ domain containing 4', other_designations='PDZ domain-containing RING finger protein 4-like protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153802166, end_position_on_the_genomic_accession=153830567, orientation='minus', exon_count=10, OMIM=300634, no_hdr0=''),
  57596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57596, CurrentID=0, Status='live', Symbol='BEGAIN', Aliases='', description='brain enriched guanylate kinase associated', other_designations='brain-enriched guanylate kinase-associated homolog', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100537147, end_position_on_the_genomic_accession=100587404, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57597, CurrentID=0, Status='live', Symbol='BAHCC1', Aliases='BAHD2', description='BAH domain and coiled-coil containing 1', other_designations='BAH domain-containing protein 2|bromo adjacent homology domain-containing protein 2', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81395464, end_position_on_the_genomic_accession=81466332, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  57599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57599, CurrentID=0, Status='live', Symbol='WDR48', Aliases='P80, SPG60, UAF1', description='WD repeat domain 48', other_designations='USP1 associated factor 1|USP1-associated factor 1|WD repeat endosomal protein', map_location='3p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39051986, end_position_on_the_genomic_accession=39096671, orientation='plus', exon_count=21, OMIM=612167, no_hdr0=''),
  57600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57600, CurrentID=0, Status='live', Symbol='FNIP2', Aliases='FNIPL, MAPO1', description='folliculin interacting protein 2', other_designations='FNIP1-like protein|O6-methylguanine-induced apoptosis 1 protein', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158769030, end_position_on_the_genomic_accession=158908050, orientation='plus', exon_count=22, OMIM=612768, no_hdr0=''),
  57602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57602, CurrentID=0, Status='live', Symbol='USP36', Aliases='DUB1', description='ubiquitin specific peptidase 36', other_designations='deubiquitinating enzyme 1|deubiquitinating enzyme 36|ubiquitin specific protease 36|ubiquitin thioesterase 36|ubiquitin thiolesterase 36|ubiquitin-specific-processing protease 36', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78787385, end_position_on_the_genomic_accession=78840898, orientation='minus', exon_count=22, OMIM=612543, no_hdr0=''),
  105369643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369643, CurrentID=0, Status='live', Symbol='LOC105369643', Aliases='', description='uncharacterized LOC105369643', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8323584, end_position_on_the_genomic_accession=8329614, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57604, CurrentID=0, Status='live', Symbol='KIAA1456', Aliases='C8orf79, TRM9L', description='KIAA1456', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12945674, end_position_on_the_genomic_accession=13029777, orientation='plus', exon_count=14, OMIM=615666, no_hdr0=''),
  57605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57605, CurrentID=0, Status='live', Symbol='PITPNM2', Aliases='NIR-3, NIR3, RDGB2, RDGBA2', description='phosphatidylinositol transfer protein, membrane associated 2', other_designations='PYK2 N-terminal domain-interacting receptor 3|phosphatidylinositol transfer protein, membrane-associated 2|retinal degeneration B alpha 2', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122983480, end_position_on_the_genomic_accession=123150810, orientation='minus', exon_count=29, OMIM=608920, no_hdr0=''),
  57606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57606, CurrentID=0, Status='live', Symbol='SLAIN2', Aliases='KIAA1458', description='SLAIN motif family member 2', other_designations='SLAIN motif family, member 2', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48341480, end_position_on_the_genomic_accession=48426201, orientation='plus', exon_count=9, OMIM=610492, no_hdr0=''),
  57608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57608, CurrentID=0, Status='live', Symbol='KIAA1462', Aliases='JCAD', description='KIAA1462', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30012800, end_position_on_the_genomic_accession=30115486, orientation='minus', exon_count=7, OMIM=614398, no_hdr0=''),
  57609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57609, CurrentID=0, Status='live', Symbol='DIP2B', Aliases='', description='disco interacting protein 2 homolog B', other_designations='DIP2 disco-interacting protein 2 homolog B', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50504985, end_position_on_the_genomic_accession=50748667, orientation='plus', exon_count=40, OMIM=611379, no_hdr0=''),
  57610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57610, CurrentID=0, Status='live', Symbol='RANBP10', Aliases='', description='RAN binding protein 10', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67723070, end_position_on_the_genomic_accession=67806652, orientation='minus', exon_count=17, OMIM=614031, no_hdr0=''),
  57611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57611, CurrentID=0, Status='live', Symbol='ISLR2', Aliases='LINX', description='immunoglobulin superfamily containing leucine-rich repeat 2', other_designations='leucine-rich repeat domain and immunoglobulin domain-containing axon extension protein', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74128203, end_position_on_the_genomic_accession=74141841, orientation='plus', exon_count=8, OMIM=614179, no_hdr0=''),
  57612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57612, CurrentID=0, Status='live', Symbol='ERV3-2', Aliases='', description='endogenous retrovirus group 3 member 2', other_designations='endogenous retrovirus group 3, member 2', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57613, CurrentID=0, Status='live', Symbol='FAM234B', Aliases='KIAA1467', description='family with sequence similarity 234 member B', other_designations='family with sequence similarity 234, member B', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13044381, end_position_on_the_genomic_accession=13083449, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  57614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57614, CurrentID=0, Status='live', Symbol='KIAA1468', Aliases='HsT3308, HsT885', description='KIAA1468', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62187244, end_position_on_the_genomic_accession=62307122, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  57615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57615, CurrentID=0, Status='live', Symbol='ZNF492', Aliases='ZNF115', description='zinc finger protein 492', other_designations='zinc finger protein 115 (Y20)', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22634324, end_position_on_the_genomic_accession=22667670, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57616, CurrentID=0, Status='live', Symbol='TSHZ3', Aliases='TSH3, ZNF537', description='teashirt zinc finger homeobox 3', other_designations='teashirt 3|teashirt family zinc finger 3|zinc finger protein 537', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31274945, end_position_on_the_genomic_accession=31349284, orientation='minus', exon_count=2, OMIM=614119, no_hdr0=''),
  57617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57617, CurrentID=0, Status='live', Symbol='VPS18', Aliases='PEP3', description='VPS18, CORVET/HOPS core subunit', other_designations='VPS18 CORVET/HOPS core subunit|hVPS18|vacuolar protein sorting 18 homolog|vacuolar protein sorting protein 18', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40894430, end_position_on_the_genomic_accession=40903975, orientation='plus', exon_count=7, OMIM=608551, no_hdr0=''),
  57619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57619, CurrentID=0, Status='live', Symbol='SHROOM3', Aliases='APXL3, MSTP013, SHRM, ShrmL', description='shroom family member 3', other_designations='F-actin-binding protein|shroom-related protein', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76435100, end_position_on_the_genomic_accession=76783253, orientation='plus', exon_count=18, OMIM=604570, no_hdr0=''),
  57620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57620, CurrentID=0, Status='live', Symbol='STIM2', Aliases='', description='stromal interaction molecule 2', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26860691, end_position_on_the_genomic_accession=27025381, orientation='plus', exon_count=14, OMIM=610841, no_hdr0=''),
  57621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57621, CurrentID=0, Status='live', Symbol='ZBTB2', Aliases='ZNF437', description='zinc finger and BTB domain containing 2', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151364115, end_position_on_the_genomic_accession=151391700, orientation='minus', exon_count=6, OMIM=616595, no_hdr0=''),
  57622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57622, CurrentID=0, Status='live', Symbol='LRFN1', Aliases='SALM2', description='leucine rich repeat and fibronectin type III domain containing 1', other_designations='CTC-246B18.8|synaptic adhesion-like molecule 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39306817, end_position_on_the_genomic_accession=39320854, orientation='minus', exon_count=5, OMIM=612807, no_hdr0=''),
  57623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57623, CurrentID=0, Status='live', Symbol='ZFAT', Aliases='AITD31, ZNF406, ZFAT', description='zinc finger and AT-hook domain containing', other_designations='zinc finger and AT hook domain containing|zinc finger gene in autoimmune thyroid disease|zinc finger protein 406|zinc-finger gene in AITD susceptibility region', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134477788, end_position_on_the_genomic_accession=134713049, orientation='minus', exon_count=23, OMIM=610931, no_hdr0=''),
  57624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57624, CurrentID=0, Status='live', Symbol='NYAP2', Aliases='KIAA1486', description='neuronal tyrosine-phosphorylated phosphoinositide-3-kinase adaptor 2', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=225399710, end_position_on_the_genomic_accession=225703653, orientation='plus', exon_count=9, OMIM=615478, no_hdr0=''),
  57626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57626, CurrentID=0, Status='live', Symbol='KLHL1', Aliases='MRP2', description='kelch like family member 1', other_designations='Mayven-related protein 2|kelch-like family member 1', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69700593, end_position_on_the_genomic_accession=70108493, orientation='minus', exon_count=11, OMIM=605332, no_hdr0=''),
  57628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57628, CurrentID=0, Status='live', Symbol='DPP10', Aliases='DPL2, DPPY, DPRP-3, DPRP3', description='dipeptidyl-peptidase 10 (inactive)', other_designations='DPP X|dipeptidyl peptidase 10|dipeptidyl peptidase IV-related protein 3|dipeptidyl peptidase X|dipeptidyl peptidase-like protein 2|dipeptidyl-peptidase 10 (non-functional)', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=114442322, end_position_on_the_genomic_accession=115845752, orientation='plus', exon_count=32, OMIM=608209, no_hdr0=''),
  57630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57630, CurrentID=0, Status='live', Symbol='SH3RF1', Aliases='POSH, RNF142, SH3MD2', description='SH3 domain containing ring finger 1', other_designations='SH3 domain-containing RING finger protein 1|SH3 multiple domains 2|SH3 multiple domains protein 2|plenty of SH3 domains|plenty of SH3s|putative E3 ubiquitin-protein ligase SH3RF1|ring finger protein 142', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169094256, end_position_on_the_genomic_accession=169271098, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57631, CurrentID=0, Status='live', Symbol='LRCH2', Aliases='dA204F4.4', description='leucine-rich repeats and calponin homology (CH) domain containing 2', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115110616, end_position_on_the_genomic_accession=115234075, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  57633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57633, CurrentID=0, Status='live', Symbol='LRRN1', Aliases='FIGLER3, NLRR-1', description='leucine rich repeat neuronal 1', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 3|neuronal leucine-rich repeat protein 1', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3799402, end_position_on_the_genomic_accession=3849834, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57634, CurrentID=0, Status='live', Symbol='EP400', Aliases='CAGH32, P400, TNRC12', description='E1A binding protein p400', other_designations='CAG repeat protein 32|domino homolog|hDomino|p400 SWI2/SNF2-related protein|p400 kDa SWI2/SNF2-related protein|trinucleotide repeat containing 12|trinucleotide repeat-containing gene 12 protein', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131949920, end_position_on_the_genomic_accession=132080466, orientation='plus', exon_count=53, OMIM=606265, no_hdr0=''),
  100128379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128379, CurrentID=0, Status='live', Symbol='LOC100128379', Aliases='', description='TRK-fused gene pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38587280, end_position_on_the_genomic_accession=38588726, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  57636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57636, CurrentID=0, Status='live', Symbol='ARHGAP23', Aliases='', description='Rho GTPase activating protein 23', other_designations='rho-type GTPase-activating protein 23', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38419286, end_position_on_the_genomic_accession=38512392, orientation='plus', exon_count=30, OMIM=610590, no_hdr0=''),
  57639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57639, CurrentID=0, Status='live', Symbol='CCDC146', Aliases='', description='coiled-coil domain containing 146', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77122617, end_position_on_the_genomic_accession=77295204, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  285613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285613, CurrentID=0, Status='live', Symbol='RELL2', Aliases='C5orf16', description='RELT-like 2', other_designations='receptor expressed in lymphoid tissues like 2', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141636946, end_position_on_the_genomic_accession=141641064, orientation='plus', exon_count=6, OMIM=611213, no_hdr0=''),
  57642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57642, CurrentID=0, Status='live', Symbol='COL20A1', Aliases='bA261N11.4', description='collagen, type XX, alpha 1', other_designations='collagen-like protein', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63293169, end_position_on_the_genomic_accession=63334847, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  57643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57643, CurrentID=0, Status='live', Symbol='ZSWIM5', Aliases='', description='zinc finger, SWIM-type containing 5', other_designations='zinc finger, SWIM domain containing 5', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45016399, end_position_on_the_genomic_accession=45206578, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  57644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57644, CurrentID=0, Status='live', Symbol='MYH7B', Aliases='MHC14, MYH14', description='myosin, heavy chain 7B, cardiac muscle, beta', other_designations='U937-associated antigen|antigen MLAA-21|myosin heavy chain 7B, cardiac muscle beta isoform|myosin, heavy polypeptide 7B, cardiac muscle, beta|slow A MYH14', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34955835, end_position_on_the_genomic_accession=35002437, orientation='plus', exon_count=49, OMIM=609928, no_hdr0=''),
  57645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57645, CurrentID=0, Status='live', Symbol='POGK', Aliases='BASS2, KRBOX2, LST003', description='pogo transposable element with KRAB domain', other_designations='KRAB box domain containing 2', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166839449, end_position_on_the_genomic_accession=166854472, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57646, CurrentID=0, Status='live', Symbol='USP28', Aliases='', description='ubiquitin specific peptidase 28', other_designations='deubiquitinating enzyme 28|ubiquitin carboxyl-terminal hydrolase 28 variant 1|ubiquitin specific protease 28|ubiquitin thioesterase 28|ubiquitin thiolesterase 28|ubiquitin-specific-processing protease 28', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113797875, end_position_on_the_genomic_accession=113875570, orientation='minus', exon_count=29, OMIM=610748, no_hdr0=''),
  57647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57647, CurrentID=0, Status='live', Symbol='DHX37', Aliases='DDX37', description='DEAH (Asp-Glu-Ala-His) box polypeptide 37', other_designations='DEAD/DEAH box helicase DDX37|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 37|DEAH box protein 37', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124946824, end_position_on_the_genomic_accession=124989131, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  57648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57648, CurrentID=0, Status='live', Symbol='KIAA1522', Aliases='', description='KIAA1522', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32741911, end_position_on_the_genomic_accession=32774970, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57649, CurrentID=0, Status='live', Symbol='PHF12', Aliases='PF1', description='PHD finger protein 12', other_designations='PHD factor 1|PHD zinc finger transcription factor', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28905247, end_position_on_the_genomic_accession=28951605, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  57650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57650, CurrentID=0, Status='live', Symbol='KIAA1524', Aliases='CIP2A, p90', description='KIAA1524', other_designations='cancerous inhibitor of PP2A|cancerous inhibitor of protein phosphatase 2A|p90 autoantigen', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108545752, end_position_on_the_genomic_accession=108589644, orientation='minus', exon_count=22, OMIM=610643, no_hdr0=''),
  57653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57653, CurrentID=0, Status='live', Symbol='LOC100499484-C9ORF174', Aliases='', description='LOC100499484-C9orf174 readthrough', other_designations="Behcet's Disease Associated Gene 1", map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97238426, end_position_on_the_genomic_accession=97377295, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  57654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57654, CurrentID=0, Status='live', Symbol='UVSSA', Aliases='KIAA1530, UVSS3', description='UV stimulated scaffold protein A', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1345658, end_position_on_the_genomic_accession=1388049, orientation='plus', exon_count=31, OMIM=614632, no_hdr0=''),
  57655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57655, CurrentID=0, Status='live', Symbol='GRAMD1A', Aliases='KIAA1533', description='GRAM domain containing 1A', other_designations='', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34994784, end_position_on_the_genomic_accession=35026471, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  57657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57657, CurrentID=0, Status='live', Symbol='HCN3', Aliases='', description='hyperpolarization activated cyclic nucleotide gated potassium channel 3', other_designations='hyperpolarization activated cyclic nucleotide-gated potassium channel 3', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155277427, end_position_on_the_genomic_accession=155289848, orientation='plus', exon_count=10, OMIM=609973, no_hdr0=''),
  57658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57658, CurrentID=0, Status='live', Symbol='CALCOCO1', Aliases='Cocoa, PP13275, calphoglin', description='calcium binding and coiled-coil domain 1', other_designations='coiled-coil coactivator protein|coiled-coil leucine zipper coactivator 1|inorganic pyrophosphatase activator|sarcoma antigen NY-SAR-3', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53711118, end_position_on_the_genomic_accession=53727523, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  57659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57659, CurrentID=0, Status='live', Symbol='ZBTB4', Aliases='KAISO-L1, ZNF903', description='zinc finger and BTB domain containing 4', other_designations='KAISO-like zinc finger protein 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7459366, end_position_on_the_genomic_accession=7484249, orientation='minus', exon_count=8, OMIM=612308, no_hdr0=''),
  57661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57661, CurrentID=0, Status='live', Symbol='PHRF1', Aliases='PPP1R125, RNF221', description='PHD and ring finger domains 1', other_designations='CTD-binding SR-like protein rA9|protein phosphatase 1, regulatory subunit 125', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=576220, end_position_on_the_genomic_accession=612222, orientation='plus', exon_count=18, OMIM=611780, no_hdr0=''),
  57662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57662, CurrentID=0, Status='live', Symbol='CAMSAP3', Aliases='KIAA1543, NEZHA, PPP1R80', description='calmodulin regulated spectrin associated protein family member 3', other_designations='calmodulin regulated spectrin-associated protein family, member 3|protein phosphatase 1, regulatory subunit 80', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7595902, end_position_on_the_genomic_accession=7618310, orientation='plus', exon_count=19, OMIM=612685, no_hdr0=''),
  57663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57663, CurrentID=0, Status='live', Symbol='USP29', Aliases='HOM-TES-84/86', description='ubiquitin specific peptidase 29', other_designations='deubiquitinating enzyme 29|ubiquitin thioesterase 29|ubiquitin thiolesterase 29|ubiquitin-specific processing protease|ubiquitin-specific-processing protease 29', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57120141, end_position_on_the_genomic_accession=57131926, orientation='plus', exon_count=4, OMIM=609546, no_hdr0=''),
  57664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57664, CurrentID=0, Status='live', Symbol='PLEKHA4', Aliases='PEPP1', description='pleckstrin homology domain containing A4', other_designations='PEPP-1|PH domain-containing family A member 4|phosphoinositol 3-phosphate binding protein-1|phosphoinositol 3-phosphate-binding PH domain protein 1|phosphoinositol 3-phosphate-binding protein 1|pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 4', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48837092, end_position_on_the_genomic_accession=48868627, orientation='minus', exon_count=21, OMIM=607769, no_hdr0=''),
  57665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57665, CurrentID=0, Status='live', Symbol='RDH14', Aliases='PAN2, SDR7C4', description='retinol dehydrogenase 14 (all-trans/9-cis/11-cis)', other_designations='alcohol dehydrogenase PAN2|pancreas protein 2|retinol dehydrogenase 14 (all-trans and 9-cis)|short chain dehydrogenase/reductase family 7C member 4|short chain dehydrogenase/reductase family 7C, member 4', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18554723, end_position_on_the_genomic_accession=18560693, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57666, CurrentID=0, Status='live', Symbol='FBRSL1', Aliases='', description='fibrosin-like 1', other_designations='AUTS2-like protein|HBV X-transactivated gene 9 protein|HBV XAg-transactivated protein 9', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132490428, end_position_on_the_genomic_accession=132585188, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  140683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140683, CurrentID=0, Status='live', Symbol='BPIFA2', Aliases='C20orf70, PSP, SPLUNC2, bA49G10.1', description='BPI fold containing family A member 2', other_designations='BPI fold containing family A, member 2|parotid secretory protein|short palate, lung and nasal epithelium carcinoma associated 2|short palate, lung and nasal epithelium carcinoma-associated protein 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33161836, end_position_on_the_genomic_accession=33181417, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57669, CurrentID=0, Status='live', Symbol='EPB41L5', Aliases='BE37, YMO1, YRT', description='erythrocyte membrane protein band 4.1 like 5', other_designations='yurt homolog', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120013028, end_position_on_the_genomic_accession=120179121, orientation='plus', exon_count=26, OMIM=611730, no_hdr0=''),
  57670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57670, CurrentID=0, Status='live', Symbol='KIAA1549', Aliases='', description='KIAA1549', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138831381, end_position_on_the_genomic_accession=138981318, orientation='minus', exon_count=20, OMIM=613344, no_hdr0=''),
  57673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57673, CurrentID=0, Status='live', Symbol='BEND3', Aliases='KIAA1553', description='BEN domain containing 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107065180, end_position_on_the_genomic_accession=107115182, orientation='minus', exon_count=8, OMIM=616374, no_hdr0=''),
  57674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57674, CurrentID=0, Status='live', Symbol='RNF213', Aliases='ALO17, C17orf27, KIAA1618, MYMY2, MYSTR, NET57', description='ring finger protein 213', other_designations='ALK lymphoma oligomerization partner on chromosome 17|mysterin', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80260840, end_position_on_the_genomic_accession=80398786, orientation='plus', exon_count=72, OMIM=613768, no_hdr0=''),
  105369655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369655, CurrentID=0, Status='live', Symbol='LOC105369655', Aliases='', description='uncharacterized LOC105369655', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10115386, end_position_on_the_genomic_accession=10133805, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57677, CurrentID=0, Status='live', Symbol='ZFP14', Aliases='ZNF531', description='ZFP14 zinc finger protein', other_designations='zinc finger protein 531', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36334453, end_position_on_the_genomic_accession=36379203, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57678, CurrentID=0, Status='live', Symbol='GPAM', Aliases='GPAT, GPAT1', description='glycerol-3-phosphate acyltransferase, mitochondrial', other_designations='GPAT-1|glycerol 3-phosphate acyltransferase, mitochondrial', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112149864, end_position_on_the_genomic_accession=112215395, orientation='minus', exon_count=23, OMIM=602395, no_hdr0=''),
  57679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57679, CurrentID=0, Status='live', Symbol='ALS2', Aliases='ALS2CR6, ALSJ, IAHSP, PLSJ', description='ALS2, alsin Rho guanine nucleotide exchange factor', other_designations='amyotrophic lateral sclerosis 2 chromosomal region candidate gene 6 protein|amyotrophic lateral sclerosis 2 protein', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201700263, end_position_on_the_genomic_accession=201781172, orientation='minus', exon_count=38, OMIM=606352, no_hdr0=''),
  57680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57680, CurrentID=0, Status='live', Symbol='CHD8', Aliases='AUTS18, HELSNF1', description='chromodomain helicase DNA binding protein 8', other_designations='ATP-dependent helicase CHD8|axis duplication inhibitor|duplin|helicase with SNF2 domain 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21385194, end_position_on_the_genomic_accession=21437298, orientation='minus', exon_count=39, OMIM=610528, no_hdr0=''),
  57683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57683, CurrentID=0, Status='live', Symbol='ZDBF2', Aliases='', description='zinc finger, DBF-type containing 2', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206274641, end_position_on_the_genomic_accession=206314438, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57684, CurrentID=0, Status='live', Symbol='ZBTB26', Aliases='ZNF481, bioref', description='zinc finger and BTB domain containing 26', other_designations='zinc finger protein 481|zinc finger protein 483|zinc finger protein Bioref', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122918029, end_position_on_the_genomic_accession=122931551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57685, CurrentID=0, Status='live', Symbol='CACHD1', Aliases='', description='cache domain containing 1', other_designations='cache domain-containing protein 1|von Willebrand factor type A and cache domain containing 1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64470793, end_position_on_the_genomic_accession=64693058, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  57687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57687, CurrentID=0, Status='live', Symbol='VAT1L', Aliases='', description='vesicle amine transport 1-like', other_designations='vesicle amine transport protein 1 homolog (T. californica)-like|vesicle amine transport protein 1 homolog-like (T. californica)', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77788564, end_position_on_the_genomic_accession=77980107, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57688, CurrentID=0, Status='live', Symbol='ZSWIM6', Aliases='AFND', description='zinc finger, SWIM-type containing 6', other_designations='zinc finger, SWIM domain containing 6', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61332273, end_position_on_the_genomic_accession=61546172, orientation='plus', exon_count=14, OMIM=615951, no_hdr0=''),
  57689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57689, CurrentID=0, Status='live', Symbol='LRRC4C', Aliases='NGL-1, NGL1', description='leucine rich repeat containing 4C', other_designations='netrin-G1 ligand', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=40109938, end_position_on_the_genomic_accession=41460212, orientation='minus', exon_count=10, OMIM=608817, no_hdr0=''),
  57690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57690, CurrentID=0, Status='live', Symbol='TNRC6C', Aliases='', description='trinucleotide repeat containing 6C', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78004170, end_position_on_the_genomic_accession=78108835, orientation='plus', exon_count=23, OMIM=610741, no_hdr0=''),
  57691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57691, CurrentID=0, Status='live', Symbol='KIAA1586', Aliases='', description='KIAA1586', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57046532, end_position_on_the_genomic_accession=57059326, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57692, CurrentID=0, Status='live', Symbol='MAGEE1', Aliases='DAMAGE, HCA1', description='MAGE family member E1', other_designations='MAGE-E1 antigen|alpha-dystrobrevin-associated MAGE Protein|dystrobrevin-associated MAGE protein|hepatocellular carcinoma-associated HCA1|hepatocellular carcinoma-associated protein 1|melanoma antigen family E, 1|melanoma antigen family E1', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76427653, end_position_on_the_genomic_accession=76431353, orientation='plus', exon_count=1, OMIM=300759, no_hdr0=''),
  57693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57693, CurrentID=0, Status='live', Symbol='ZNF317', Aliases='', description='zinc finger protein 317', other_designations='KRAB-containing zinc finger protein 317', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9140380, end_position_on_the_genomic_accession=9163415, orientation='plus', exon_count=8, OMIM=613864, no_hdr0=''),
  57695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57695, CurrentID=0, Status='live', Symbol='USP37', Aliases='', description='ubiquitin specific peptidase 37', other_designations='deubiquitinating enzyme 37|tmp_locus_50|ubiquitin specific protease 37|ubiquitin thioesterase 37|ubiquitin thiolesterase 37|ubiquitin-specific-processing protease 37', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218450251, end_position_on_the_genomic_accession=218568366, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  57696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57696, CurrentID=0, Status='live', Symbol='DDX55', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 55', other_designations='DEAD box protein 55', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123602092, end_position_on_the_genomic_accession=123620938, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  57697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57697, CurrentID=0, Status='live', Symbol='FANCM', Aliases='FAAP250, KIAA1596', description='Fanconi anemia complementation group M', other_designations='ATP-dependent RNA helicase FANCM|Fanconi anemia, complementation group M|fanconi anemia-associated polypeptide of 250 kDa|protein Hef ortholog', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45135933, end_position_on_the_genomic_accession=45200890, orientation='plus', exon_count=24, OMIM=609644, no_hdr0=''),
  57698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57698, CurrentID=0, Status='live', Symbol='SHTN1', Aliases='KIAA1598', description='shootin 1', other_designations='shootin1', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116883377, end_position_on_the_genomic_accession=117126586, orientation='minus', exon_count=21, OMIM=611171, no_hdr0=''),
  57699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57699, CurrentID=0, Status='live', Symbol='CPNE5', Aliases='COPN5, CPN5', description='copine V', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36740772, end_position_on_the_genomic_accession=36839444, orientation='minus', exon_count=29, OMIM=604209, no_hdr0=''),
  57700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57700, CurrentID=0, Status='live', Symbol='FAM160B1', Aliases='KIAA1600, bA106M7.3', description='family with sequence similarity 160 member B1', other_designations='family with sequence similarity 160, member B1', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114821744, end_position_on_the_genomic_accession=114899827, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  57701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57701, CurrentID=0, Status='live', Symbol='NCKAP5L', Aliases='FP1193, KIAA1602', description='NCK-associated protein 5-like', other_designations='NCKAP5-like', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49791139, end_position_on_the_genomic_accession=49828750, orientation='minus', exon_count=14, OMIM=615104, no_hdr0=''),
  9617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9617, CurrentID=0, Status='live', Symbol='MTRF1', Aliases='MRF1, MTTRF1, RF1', description='mitochondrial translational release factor 1', other_designations='MRF-1|mitochontrial peptide chain release factor 1|mtRF-1', map_location='13q14.1-q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41216369, end_position_on_the_genomic_accession=41263797, orientation='minus', exon_count=17, OMIM=604601, no_hdr0=''),
  57704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57704, CurrentID=0, Status='live', Symbol='GBA2', Aliases='AD035, NLGase, SPG46', description='glucosidase, beta (bile acid) 2', other_designations='beta-glucocerebrosidase 2|glucosylceramidase 2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35736866, end_position_on_the_genomic_accession=35749228, orientation='minus', exon_count=19, OMIM=609471, no_hdr0=''),
  57705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57705, CurrentID=0, Status='live', Symbol='WDFY4', Aliases='C10orf64', description='WDFY family member 4', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48684830, end_position_on_the_genomic_accession=48982956, orientation='plus', exon_count=69, OMIM=613316, no_hdr0=''),
  57706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57706, CurrentID=0, Status='live', Symbol='DENND1A', Aliases='FAM31A, KIAA1608', description='DENN/MADD domain containing 1A', other_designations='connecdenn 1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123379654, end_position_on_the_genomic_accession=123930138, orientation='minus', exon_count=27, OMIM=613633, no_hdr0=''),
  57707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57707, CurrentID=0, Status='live', Symbol='TLDC1', Aliases='KIAA1609', description='TBC/LysM-associated domain containing 1', other_designations='TBC/LysM-associated domain-containing protein 1|TLD domain containing 1', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84476360, end_position_on_the_genomic_accession=84554045, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57708, CurrentID=0, Status='live', Symbol='MIER1', Aliases='ER1, MI-ER1', description='mesoderm induction early response 1, transcriptional regulator', other_designations='mesoderm induction early response 1 homolog', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66924895, end_position_on_the_genomic_accession=66988619, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  140690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140690, CurrentID=0, Status='live', Symbol='CTCFL', Aliases='BORIS, CT27, CTCF-T, HMGB1L1, dJ579F20.2', description='CCCTC-binding factor (zinc finger protein)-like', other_designations='BORIS-like protein|CTCF paralog|CTCF-like protein|HMG-1L1|brother of the regulator of imprinted sites|cancer/testis antigen 27|putative high mobility group protein 1-like 1|putative high mobility group protein B1-like 1|zinc finger protein CTCF-T', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57495965, end_position_on_the_genomic_accession=57525652, orientation='minus', exon_count=16, OMIM=607022, no_hdr0=''),
  57710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57710, CurrentID=0, Status='live', Symbol='KIAA1614', Aliases='', description='KIAA1614', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180913177, end_position_on_the_genomic_accession=180946103, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57711, CurrentID=0, Status='live', Symbol='ZNF529', Aliases='', description='zinc finger protein 529', other_designations='', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36534518, end_position_on_the_genomic_accession=36605411, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  57713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57713, CurrentID=0, Status='live', Symbol='SFMBT2', Aliases='', description='Scm-like with four mbt domains 2', other_designations='Scm-related gene containing four mbt domains 2|scm-like with 4 MBT domains protein 2', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7158624, end_position_on_the_genomic_accession=7411486, orientation='minus', exon_count=24, OMIM=615392, no_hdr0=''),
  57715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57715, CurrentID=0, Status='live', Symbol='SEMA4G', Aliases='', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4G', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100969482, end_position_on_the_genomic_accession=100985616, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  57716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57716, CurrentID=0, Status='live', Symbol='PRX', Aliases='CMT4F', description='periaxin', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40393764, end_position_on_the_genomic_accession=40414718, orientation='minus', exon_count=8, OMIM=605725, no_hdr0=''),
  57717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57717, CurrentID=0, Status='live', Symbol='PCDHB16', Aliases='ME1, PCDH-BETA16, PCDH3X, PCDHB8a', description='protocadherin beta 16', other_designations='cadherin ME1|protocadherin beta 8a|protocadherin-3x', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141181690, end_position_on_the_genomic_accession=141186221, orientation='plus', exon_count=1, OMIM=606345, no_hdr0=''),
  57718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57718, CurrentID=0, Status='live', Symbol='PPP4R4', Aliases='CFAP14, KIAA1622, PP4R4', description='protein phosphatase 4, regulatory subunit 4', other_designations='HEAT-like repeat-containing protein|cilia and flagella associated protein 14', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94174312, end_position_on_the_genomic_accession=94279755, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  57719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57719, CurrentID=0, Status='live', Symbol='ANO8', Aliases='KIAA1623, TMEM16H', description='anoctamin 8', other_designations='transmembrane protein 16H', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17323223, end_position_on_the_genomic_accession=17334834, orientation='minus', exon_count=18, OMIM=610216, no_hdr0=''),
  57720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57720, CurrentID=0, Status='live', Symbol='GPR107', Aliases='GCDRP, LUSTR1, bA138E2.2', description='G protein-coupled receptor 107', other_designations='lung seven transmembrane receptor 1', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130053706, end_position_on_the_genomic_accession=130140169, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  57721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57721, CurrentID=0, Status='live', Symbol='METTL14', Aliases='', description='methyltransferase like 14', other_designations='methyltransferase-like protein 14', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118685364, end_position_on_the_genomic_accession=118712270, orientation='plus', exon_count=12, OMIM=616504, no_hdr0=''),
  57722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57722, CurrentID=0, Status='live', Symbol='IGDCC4', Aliases='DDM36, NOPE', description='immunoglobulin superfamily, DCC subclass, member 4', other_designations='likely ortholog of mouse neighbor of Punc E11|neighbor of Punc E11', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65381480, end_position_on_the_genomic_accession=65423072, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  57724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57724, CurrentID=0, Status='live', Symbol='EPG5', Aliases='HEEW1, KIAA1632, VICIS', description='ectopic P-granules autophagy protein 5 homolog (C. elegans)', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45800580, end_position_on_the_genomic_accession=45967339, orientation='minus', exon_count=47, OMIM=615068, no_hdr0=''),
  57727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57727, CurrentID=0, Status='live', Symbol='NCOA5', Aliases='CIA, bA465L10.6', description='nuclear receptor coactivator 5', other_designations='NCoA-5|coactivator independent of AF-2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46060985, end_position_on_the_genomic_accession=46089952, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57728, CurrentID=0, Status='live', Symbol='WDR19', Aliases='ATD5, CED4, DYF-2, IFT144, NPHP13, ORF26, Oseg6, PWDMP, SRTD5', description='WD repeat domain 19', other_designations='WD repeat membrane protein PWDMP|intraflagellar transport 144 homolog', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39182189, end_position_on_the_genomic_accession=39285810, orientation='plus', exon_count=37, OMIM=608151, no_hdr0=''),
  100126784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126784, CurrentID=0, Status='live', Symbol='LOC100126784', Aliases='', description='uncharacterized LOC100126784', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19710934, end_position_on_the_genomic_accession=19714672, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57730, CurrentID=0, Status='live', Symbol='ANKRD36B', Aliases='KIAA1641', description='ankyrin repeat domain 36B', other_designations='CLL-associated antigen KW-1|melanoma-associated antigen', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97492663, end_position_on_the_genomic_accession=97590245, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  57731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57731, CurrentID=0, Status='live', Symbol='SPTBN4', Aliases='QV, SPNB4, SPTBN3', description='spectrin, beta, non-erythrocytic 4', other_designations='beta-IV spectrin|spectrin beta chain, brain 3|spectrin, non-erythroid beta chain 3', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40466107, end_position_on_the_genomic_accession=40576459, orientation='plus', exon_count=38, OMIM=606214, no_hdr0=''),
  57732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57732, CurrentID=0, Status='live', Symbol='ZFYVE28', Aliases='LST2, LYST2', description='zinc finger, FYVE domain containing 28', other_designations='hLst2|zinc finger FYVE domain-containing protein 28', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2267524, end_position_on_the_genomic_accession=2418690, orientation='minus', exon_count=18, OMIM=614176, no_hdr0=''),
  57733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57733, CurrentID=0, Status='live', Symbol='GBA3', Aliases='CBG, CBGL1, GLUC, KLRP', description='glucosidase, beta, acid 3 (gene/pseudogene)', other_designations='cytosolic beta-glucosidase-like protein 1|glucosidase, beta, acid 3 (cytosolic)|klotho-related protein', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22692914, end_position_on_the_genomic_accession=22819575, orientation='plus', exon_count=6, OMIM=606619, no_hdr0=''),
  140695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140695, CurrentID=0, Status='live', Symbol='RPL7P2', Aliases='bA550O8.1', description='ribosomal protein L7 pseudogene 2', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2075556, end_position_on_the_genomic_accession=2075823, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374536, CurrentID=0, Status='live', Symbol='LOC105374536', Aliases='', description='uncharacterized LOC105374536', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25425403, end_position_on_the_genomic_accession=25467849, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369667, CurrentID=0, Status='live', Symbol='LOC105369667', Aliases='', description='uncharacterized LOC105369667', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13833335, end_position_on_the_genomic_accession=13838128, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140697, CurrentID=0, Status='live', Symbol='RPL36P1', Aliases='RPL36_9_1713, bA80K6.3', description='ribosomal protein L36 pseudogene 1', other_designations='', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52890157, end_position_on_the_genomic_accession=52890384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  57758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57758, CurrentID=0, Status='live', Symbol='SCUBE2', Aliases='CEGB1, CEGF1, CEGP1', description='signal peptide, CUB domain, EGF-like 2', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9019500, end_position_on_the_genomic_accession=9091603, orientation='minus', exon_count=25, OMIM=611747, no_hdr0=''),
  57760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57760, CurrentID=0, Status='live', Symbol='SPG16', Aliases='SPG', description='spastic paraplegia 16 (complicated, X-linked recessive)', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300266, no_hdr0=''),
  57761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57761, CurrentID=0, Status='live', Symbol='TRIB3', Aliases='C20orf97, NIPK, SINK, SKIP3, TRB3', description='tribbles pseudokinase 3', other_designations='TRB-3|neuronal cell death inducible putative kinase|neuronal cell death-inducible putative kinase|p65-interacting inhibitor of NF-kappa-B|p65-interacting inhibitor of NF-kappaB', map_location='20p13-p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=380629, end_position_on_the_genomic_accession=397559, orientation='plus', exon_count=6, OMIM=607898, no_hdr0=''),
  57762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57762, CurrentID=0, Status='live', Symbol='NS2', Aliases='', description='Noonan syndrome 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605275, no_hdr0=''),
  140699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140699, CurrentID=0, Status='live', Symbol='MROH8', Aliases='C20orf131, C20orf132, dJ621N11.3, dJ621N11.4', description='maestro heat-like repeat family member 8', other_designations='maestro heat-like repeat-containing protein family member 8', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37101226, end_position_on_the_genomic_accession=37179588, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  140700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140700, CurrentID=0, Status='live', Symbol='SAMD10', Aliases='C20orf136, dJ591C20, dJ591C20.7', description='sterile alpha motif domain containing 10', other_designations='D930033N23Rik|SAM domain-containing protein 10', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63974113, end_position_on_the_genomic_accession=63980986, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369671, CurrentID=0, Status='live', Symbol='LOC105369671', Aliases='', description='uncharacterized LOC105369671', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15162911, end_position_on_the_genomic_accession=15172272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140701, CurrentID=0, Status='live', Symbol='ABHD16B', Aliases='C20orf135, dJ591C20.1', description='abhydrolase domain containing 16B', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63861213, end_position_on_the_genomic_accession=63862988, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369672, CurrentID=0, Status='live', Symbol='LOC105369672', Aliases='', description='uncharacterized LOC105369672', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15521942, end_position_on_the_genomic_accession=15555769, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369673, CurrentID=0, Status='live', Symbol='LOC105369673', Aliases='', description='uncharacterized LOC105369673', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15239348, end_position_on_the_genomic_accession=15348497, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57786, CurrentID=0, Status='live', Symbol='RBAK', Aliases='ZNF769', description='RB-associated KRAB zinc finger', other_designations='RB-associated KRAB repressor|zinc finger protein 769', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5045821, end_position_on_the_genomic_accession=5069488, orientation='plus', exon_count=5, OMIM=608191, no_hdr0=''),
  57787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57787, CurrentID=0, Status='live', Symbol='MARK4', Aliases='MARK4LS, MARKL1, MARKL1L, PAR-1D, MARK4', description='MAP/microtubule affinity-regulating kinase 4', other_designations='MAP/microtubule affinity-regulating kinase like 1|MARK4 serine/threonine protein kinase', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45251258, end_position_on_the_genomic_accession=45305283, orientation='plus', exon_count=18, OMIM=606495, no_hdr0=''),
  57788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57788, CurrentID=0, Status='live', Symbol='RFMN', Aliases='', description='Roifman syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300258, no_hdr0=''),
  57789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57789, CurrentID=0, Status='live', Symbol='MTBSX', Aliases='MTBS', description='Mycobacterium tuberculosis, susceptibility to infection by', other_designations='', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300259, no_hdr0=''),
  57790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57790, CurrentID=0, Status='live', Symbol='MRXSL', Aliases='', description='Lubs X-linked mental retardation syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300260, no_hdr0=''),
  57791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57791, CurrentID=0, Status='live', Symbol='MRXSA', Aliases='', description='Armfield X-linked mental retardation syndrome', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300261, no_hdr0=''),
  57792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57792, CurrentID=0, Status='live', Symbol='MRXSAB', Aliases='', description='Abidi X-linked mental retardation syndrome', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300262, no_hdr0=''),
  57794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57794, CurrentID=0, Status='live', Symbol='SUGP1', Aliases='F23858, RBP, SF4', description='SURP and G-patch domain containing 1', other_designations='RNA-binding protein RBP|SURP and G patch domain containing 1|splicing factor 4', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19276511, end_position_on_the_genomic_accession=19320512, orientation='minus', exon_count=14, OMIM=607992, no_hdr0=''),
  57795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57795, CurrentID=0, Status='live', Symbol='BRINP2', Aliases='DBCCR1L2, FAM5B', description='bone morphogenetic protein/retinoic acid inducible neural-specific 2', other_designations='DBCCR1-like protein 2|DBCCR1-like2|bone morphogenic protein/retinoic acid inducible neural-specific 2|family with sequence similarity 5, member B|protein FAM5B', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177170922, end_position_on_the_genomic_accession=177282422, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  57796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57796, CurrentID=0, Status='live', Symbol='DKFZP761C1711', Aliases='', description='uncharacterized protein DKFZp761C1711', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57798, CurrentID=0, Status='live', Symbol='GATAD1', Aliases='CMD2B, ODAG, RG083M05.2', description='GATA zinc finger domain containing 1', other_designations='ocular development-associated gene protein', map_location='7q21-q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92447448, end_position_on_the_genomic_accession=92491328, orientation='plus', exon_count=12, OMIM=614518, no_hdr0=''),
  57799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57799, CurrentID=0, Status='live', Symbol='RAB40C', Aliases='RARL, RASL8C', description='RAB40C, member RAS oncogene family', other_designations='RAR (RAS like GTPASE) like|RAR like|RAS-like GTPase|RAS-like, family 8, member C|SOCS box-containing protein RAR3|rar-like protein|ras-like protein family member 8C', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=589357, end_position_on_the_genomic_accession=629273, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57801, CurrentID=0, Status='live', Symbol='HES4', Aliases='bHLHb42', description='hes family bHLH transcription factor 4', other_designations='bHLH factor Hes4|class B basic helix-loop-helix protein 42|hHES4|hairy and enhancer of split 4', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=998962, end_position_on_the_genomic_accession=1001285, orientation='minus', exon_count=4, OMIM=608060, no_hdr0=''),
  57804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57804, CurrentID=0, Status='live', Symbol='POLD4', Aliases='POLDS, p12', description='polymerase (DNA-directed), delta 4, accessory subunit', other_designations='DNA polymerase delta smallest subunit p12', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67350765, end_position_on_the_genomic_accession=67353596, orientation='minus', exon_count=4, OMIM=611525, no_hdr0=''),
  57805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57805, CurrentID=0, Status='live', Symbol='CCAR2', Aliases='DBC-1, DBC1, KIAA1967, NET35, p30 DBC, p30DBC', description='cell cycle and apoptosis regulator 2', other_designations='cell division cycle and apoptosis regulator protein 2|deleted in breast cancer 1|p30 DBC protein', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22604632, end_position_on_the_genomic_accession=22620471, orientation='plus', exon_count=22, OMIM=607359, no_hdr0=''),
  140707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140707, CurrentID=0, Status='live', Symbol='BRI3BP', Aliases='BNAS1, HCCR-1, HCCR-2, HCCRBP-1, KG19', description='BRI3 binding protein', other_designations='I3-binding protein|cervical cancer 1 proto-oncogene-binding protein KG19|cervical cancer oncogene binding protein', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124993648, end_position_on_the_genomic_accession=125051845, orientation='plus', exon_count=4, OMIM=615627, no_hdr0=''),
  105369678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369678, CurrentID=0, Status='live', Symbol='LOC105369678', Aliases='', description='uncharacterized LOC105369678', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=17581388, end_position_on_the_genomic_accession=17589080, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57817, CurrentID=0, Status='live', Symbol='HAMP', Aliases='HEPC, HFE2B, LEAP1, PLTR', description='hepcidin antimicrobial peptide', other_designations='liver-expressed antimicrobial peptide 1|putative liver tumor regressor', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35282507, end_position_on_the_genomic_accession=35285143, orientation='plus', exon_count=3, OMIM=606464, no_hdr0=''),
  57818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57818, CurrentID=0, Status='live', Symbol='G6PC2', Aliases='IGRP', description='glucose-6-phosphatase, catalytic, 2', other_designations='G-6-Pase 2|G6Pase 2|islet-specific G6CP-related protein|islet-specific glucose-6-phosphatase catalytic subunit-related protein|islet-specific glucose-6-phosphatase-related protein', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168901240, end_position_on_the_genomic_accession=168910000, orientation='plus', exon_count=5, OMIM=608058, no_hdr0=''),
  57819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57819, CurrentID=0, Status='live', Symbol='LSM2', Aliases='C6orf28, G7B, YBL026W, snRNP', description='LSM2 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM2 U6 small nuclear RNA and mRNA degradation associated|LSM2 homolog, U6 small nuclear RNA associated|protein G7b|small nuclear ribonuclear protein D homolog|snRNP core Sm-like protein Sm-x5', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31797392, end_position_on_the_genomic_accession=31806984, orientation='minus', exon_count=5, OMIM=607282, no_hdr0=''),
  57820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57820, CurrentID=0, Status='live', Symbol='CCNB1IP1', Aliases='C14orf18, HEI10', description='cyclin B1 interacting protein 1, E3 ubiquitin protein ligase', other_designations='human enhancer of invasion 10', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20311368, end_position_on_the_genomic_accession=20333312, orientation='minus', exon_count=8, OMIM=608249, no_hdr0=''),
  57821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57821, CurrentID=0, Status='live', Symbol='CCDC181', Aliases='C1orf114', description='coiled-coil domain containing 181', other_designations='', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169394870, end_position_on_the_genomic_accession=169460669, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57822, CurrentID=0, Status='live', Symbol='GRHL3', Aliases='SOM, TFCP2L4, VWS2', description='grainyhead-like transcription factor 3', other_designations='grainyhead-like 3|sister of mammalian grainyhead|sister-of-mammalian grainyhead|transcription factor CP2-like 4|transcription factor hSOM1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24319322, end_position_on_the_genomic_accession=24364482, orientation='plus', exon_count=19, OMIM=608317, no_hdr0=''),
  57823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57823, CurrentID=0, Status='live', Symbol='SLAMF7', Aliases='19A, CD319, CRACC, CS1', description='SLAM family member 7', other_designations='19A24 protein|CD2 subset 1|CD2-like receptor activating cytotoxic cells|CD2-like receptor-activating cytotoxic cells|membrane protein FOAP-12|novel LY9 (lymphocyte antigen 9) like protein|protein 19A', map_location='1q23.1-q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160739057, end_position_on_the_genomic_accession=160754821, orientation='plus', exon_count=7, OMIM=606625, no_hdr0=''),
  57824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57824, CurrentID=0, Status='live', Symbol='HMHB1', Aliases='HB-1, HB-1Y, HLA-HB1', description='histocompatibility (minor) HB-1', other_designations='minor histocompatibility antigen HB-1', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143812161, end_position_on_the_genomic_accession=143820719, orientation='plus', exon_count=2, OMIM=609961, no_hdr0=''),
  105369680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369680, CurrentID=0, Status='live', Symbol='LOC105369680', Aliases='', description='uncharacterized LOC105369680', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18413138, end_position_on_the_genomic_accession=18501176, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57826, CurrentID=0, Status='live', Symbol='RAP2C', Aliases='', description='RAP2C, member of RAS oncogene family', other_designations='small GTPase RAP2C', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132203024, end_position_on_the_genomic_accession=132219480, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57827, CurrentID=0, Status='live', Symbol='C6orf47', Aliases='D6S53E, G4, NG34', description='chromosome 6 open reading frame 47', other_designations='protein G4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31658298, end_position_on_the_genomic_accession=31660772, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57828, CurrentID=0, Status='live', Symbol='CATSPERG', Aliases='C19orf15', description='catsper channel auxiliary subunit gamma', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38335783, end_position_on_the_genomic_accession=38370949, orientation='plus', exon_count=34, OMIM=613452, no_hdr0=''),
  57829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57829, CurrentID=0, Status='live', Symbol='ZP4', Aliases='ZBP, ZP1, ZPB, Zp-4', description='zona pellucida glycoprotein 4', other_designations='zona pellucida B protein|zona pellucida protein B', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237877864, end_position_on_the_genomic_accession=237890922, orientation='minus', exon_count=13, OMIM=613514, no_hdr0=''),
  57830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57830, CurrentID=0, Status='live', Symbol='KRTAP5-8', Aliases='KRTAP5-2, KRTAP5.8, UHSKerB', description='keratin associated protein 5-8', other_designations='UHS KERB-like protein|UHS KerB|UHS keratin B|keratin, ultra high-sulfur matrix protein B|keratin, ultrahigh sulfur, B|keratin-associated protein 5.8|ultrahigh sulfur keratin-associated protein 5.8', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71538025, end_position_on_the_genomic_accession=71539207, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  57834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57834, CurrentID=0, Status='live', Symbol='CYP4F11', Aliases='CYPIVF11', description='cytochrome P450 family 4 subfamily F member 11', other_designations='3-hydroxy fatty acids omega-hydroxylase CYP4F11|cytochrome P450 4F11|cytochrome P450, family 4, subfamily F, polypeptide 11|cytochrome P450, subfamily IVF, polypeptide 11', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15912370, end_position_on_the_genomic_accession=15934866, orientation='minus', exon_count=13, OMIM=611517, no_hdr0=''),
  57835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57835, CurrentID=0, Status='live', Symbol='SLC4A5', Aliases='NBC4', description='solute carrier family 4 (sodium bicarbonate cotransporter), member 5', other_designations='NBCe2', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74216242, end_position_on_the_genomic_accession=74343407, orientation='minus', exon_count=32, OMIM=606757, no_hdr0=''),
  105369683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369683, CurrentID=0, Status='live', Symbol='LOC105369683', Aliases='', description='uncharacterized LOC105369683', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20009882, end_position_on_the_genomic_accession=20012616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140713, CurrentID=0, Status='live', Symbol='RPS3AP3', Aliases='bA332A4.2', description='ribosomal protein S3A pseudogene 3', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37021743, end_position_on_the_genomic_accession=37022009, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378904, CurrentID=0, Status='live', Symbol='LOC105378904', Aliases='', description='uncharacterized LOC105378904', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111361574, end_position_on_the_genomic_accession=111371692, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369685, CurrentID=0, Status='live', Symbol='LOC105369685', Aliases='', description='uncharacterized LOC105369685', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20104587, end_position_on_the_genomic_accession=20109904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  140715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140715, CurrentID=0, Status='live', Symbol='RPL21P2', Aliases='RPL21_65_1679, bA352D3.1', description='ribosomal protein L21 pseudogene 2', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4059601, end_position_on_the_genomic_accession=4060152, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  57862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57862, CurrentID=0, Status='live', Symbol='ZNF410', Aliases='APA-1, APA1', description='zinc finger protein 410', other_designations='another partner for ARF 1|clones 23667 and 23775 zinc finger protein|zinc finger protein APA-1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73886615, end_position_on_the_genomic_accession=73932288, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57863, CurrentID=0, Status='live', Symbol='CADM3', Aliases='BIgR, IGSF4B, NECL1, Necl-1, TSLL1, synCAM3', description='cell adhesion molecule 3', other_designations='TSLC1-like 1|TSLC1-like protein 1|brain immunoglobulin receptor|dendritic cell nectin-like protein 1 short isoform|immunoglobulin superfamily member 4B|immunoglobulin superfamily, member 4B|nectin-like 1|nectin-like protein 1|synaptic cell adhesion molecule 3', map_location='1q21.2-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159171587, end_position_on_the_genomic_accession=159203142, orientation='plus', exon_count=10, OMIM=609743, no_hdr0=''),
  57864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57864, CurrentID=0, Status='live', Symbol='SLC46A2', Aliases='Ly110, TSCOT', description='solute carrier family 46 member 2', other_designations='solute carrier family 46, member 2|thymic stromal co-transporter', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112878920, end_position_on_the_genomic_accession=112890942, orientation='minus', exon_count=4, OMIM=608956, no_hdr0=''),
  140716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140716, CurrentID=0, Status='live', Symbol='RPL35AP', Aliases='RPL35A_15_1710, bA456N23.1', description='ribosomal protein L35a pseudogene', other_designations='', map_location='20q12-q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47478214, end_position_on_the_genomic_accession=47479010, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646330, CurrentID=0, Status='live', Symbol='BTBD6P1', Aliases='', description='BTB (POZ) domain containing 6 pseudogene 1', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23901049, end_position_on_the_genomic_accession=23903008, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140717, CurrentID=0, Status='live', Symbol='RPL37AP1', Aliases='RPL37A_5_1704, dJ1013A22.4', description='ribosomal protein L37a pseudogene 1', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44466503, end_position_on_the_genomic_accession=44466873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369688, CurrentID=0, Status='live', Symbol='LOC105369688', Aliases='', description='uncharacterized LOC105369688', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20320671, end_position_on_the_genomic_accession=20328654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57876, CurrentID=0, Status='live', Symbol='MUC3B', Aliases='MUC3, MUC3A', description='mucin 3B, cell surface associated', other_designations='MUC-3B|intestinal mucin MUC3B|intestinal mucin-3B', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605633, no_hdr0=''),
  140718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140718, CurrentID=0, Status='live', Symbol='RPL31P3', Aliases='RPL31_24_1693, dJ1018D12.4', description='ribosomal protein L31 pseudogene 3', other_designations='', map_location='20p11-q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31311204, end_position_on_the_genomic_accession=31312055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378905, CurrentID=0, Status='live', Symbol='LOC105378905', Aliases='', description='uncharacterized LOC105378905', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111788775, end_position_on_the_genomic_accession=111795743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369689, CurrentID=0, Status='live', Symbol='LOC105369689', Aliases='', description='uncharacterized LOC105369689', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21662313, end_position_on_the_genomic_accession=21828991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  140719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140719, CurrentID=0, Status='live', Symbol='RPL39P', Aliases='RPL39_20_1715, dJ1167H4.5', description='ribosomal protein L39 pseudogene', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56459235, end_position_on_the_genomic_accession=56459632, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369690, CurrentID=0, Status='live', Symbol='LOC105369690', Aliases='', description='uncharacterized LOC105369690', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22063889, end_position_on_the_genomic_accession=22083113, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369691, CurrentID=0, Status='live', Symbol='LOC105369691', Aliases='', description='uncharacterized LOC105369691', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22576571, end_position_on_the_genomic_accession=22625099, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100196909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196909, CurrentID=0, Status='live', Symbol='ALPQTL1', Aliases='', description='Alkaline phosphatase, plasma level of, QTL1', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=171720, no_hdr0=''),
  100196910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196910, CurrentID=0, Status='live', Symbol='POROK2', Aliases='PPPD1', description='Porokeratosis, palmar, plantar, and disseminated 1', other_designations='', map_location='12q24.1-q24.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=175850, no_hdr0=''),
  100196911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196911, CurrentID=0, Status='live', Symbol='POROK6', Aliases='DSAP4', description='Porokeratosis, disseminated superficial actinic, 4', other_designations='', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612353, no_hdr0=''),
  100196912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196912, CurrentID=0, Status='live', Symbol='MAFD8', Aliases='', description='Major affective disorder-8, susceptibility to', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612357, no_hdr0=''),
  100196913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196913, CurrentID=0, Status='live', Symbol='SCZD14', Aliases='', description='schizophrenia, susceptibility to, 14', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612361, no_hdr0=''),
  100196914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196914, CurrentID=0, Status='live', Symbol='ALPQTL2', Aliases='', description='Alkaline phosphatase, plasma level of, QTL 2', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612367, no_hdr0=''),
  100196915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196915, CurrentID=0, Status='live', Symbol='ALPQTL3', Aliases='', description='Alkaline phosphatase, plasma level of, QTL3', other_designations='', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612368, no_hdr0=''),
  100196916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196916, CurrentID=0, Status='live', Symbol='ALPQTL4', Aliases='', description='Alkaline phosphatase, plasma level of, QTL4', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612369, no_hdr0=''),
  100196917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196917, CurrentID=0, Status='live', Symbol='MAFD9', Aliases='', description='Major affective disorder-9, susceptibility to', other_designations='', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612372, no_hdr0=''),
  100196918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196918, CurrentID=0, Status='live', Symbol='SLEB13', Aliases='', description='Systemic lupus erythematosus, susceptibility to, 13', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612378, no_hdr0=''),
  100196919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196919, CurrentID=0, Status='live', Symbol='SS3', Aliases='', description='Sarcoidosis, susceptibility to, 3', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612388, no_hdr0=''),
  643167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643167, CurrentID=0, Status='live', Symbol='LOC643167', Aliases='', description='RNA binding motif protein 39 pseudogene', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44029135, end_position_on_the_genomic_accession=44031118, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369696, CurrentID=0, Status='live', Symbol='LOC105369696', Aliases='', description='uncharacterized LOC105369696', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=23738650, end_position_on_the_genomic_accession=23757019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100189492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189492, CurrentID=0, Status='live', Symbol='TRE-CTC18-1', Aliases='TRNAE37P', description='transfer RNA-Glu (CTC) 18-1', other_designations='tRNA-Glu-CTC-18-1|transfer RNA glutamic acid 37 (anticodon CUC) pseudogene', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51563260, end_position_on_the_genomic_accession=51563331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100196944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100196944, CurrentID=0, Status='live', Symbol='LOC100196944', Aliases='', description='CD2 (cytoplasmic tail) binding protein 2 pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127857537, end_position_on_the_genomic_accession=127858471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369699, CurrentID=0, Status='live', Symbol='LOC105369699', Aliases='', description='uncharacterized LOC105369699', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24818670, end_position_on_the_genomic_accession=24836555, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643172, CurrentID=0, Status='live', Symbol='LOC643172', Aliases='', description='neurofilament, heavy polypeptide pseudogene', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41996935, end_position_on_the_genomic_accession=41999447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341689, CurrentID=0, Status='live', Symbol='CTAGE16P', Aliases='', description='CTAGE family member 16, pseudogene', other_designations='CTAGE family, member 16, pseudogene|CTAGE family, member 9 pseudogene', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58528615, end_position_on_the_genomic_accession=58531145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140730, CurrentID=0, Status='live', Symbol='RIMS4', Aliases='C20orf190, RIM 4, RIM-4, RIM4, RIM4-gamma, RIM4gamma', description='regulating synaptic membrane exocytosis 4', other_designations='RIM4 gamma|rab3-interacting molecule 4', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44751804, end_position_on_the_genomic_accession=44810338, orientation='minus', exon_count=6, OMIM=611601, no_hdr0=''),
  140731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140731, CurrentID=0, Status='live', Symbol='ANKRD60', Aliases='C20orf86, bA196N14.3', description='ankyrin repeat domain 60', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58216126, end_position_on_the_genomic_accession=58228653, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100126822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126822, CurrentID=0, Status='live', Symbol='LOC100126822', Aliases='', description='ubiquitin-like modifier activating enzyme 5 pseudogene', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227263326, end_position_on_the_genomic_accession=227265190, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131432, CurrentID=0, Status='live', Symbol='LOC100131432', Aliases='', description='uncharacterized LOC100131432', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=45029527, end_position_on_the_genomic_accession=45029901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372314, CurrentID=0, Status='live', Symbol='LOC105372314', Aliases='', description='uncharacterized LOC105372314', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20276134, end_position_on_the_genomic_accession=20284676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369705, CurrentID=0, Status='live', Symbol='LOC105369705', Aliases='', description='uncharacterized LOC105369705', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26252874, end_position_on_the_genomic_accession=26304350, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  140735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140735, CurrentID=0, Status='live', Symbol='DYNLL2', Aliases='DNCL1B, Dlc2, RSPH22', description='dynein, light chain, LC8-type 2', other_designations='8 kDa dynein light chain b|DLC8b|radial spoke 22 homolog', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58083419, end_position_on_the_genomic_accession=58090257, orientation='plus', exon_count=3, OMIM=608942, no_hdr0=''),
  100189494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189494, CurrentID=0, Status='live', Symbol='TRE-CTC9-1', Aliases='TRNAE38P', description='transfer RNA-Glu (CTC) 9-1', other_designations='tRNA-Glu-CTC-9-1|transfer RNA glutamic acid 38 (anticodon CUC) pseudogene', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150362763, end_position_on_the_genomic_accession=150362835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9665, CurrentID=0, Status='live', Symbol='KIAA0430', Aliases='LKAP, MARF1, PPP1R34', description='KIAA0430', other_designations='limkain b1|limkain-b1|meiosis arrest female 1|protein phosphatase 1, regulatory subunit 34', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15594369, end_position_on_the_genomic_accession=15643166, orientation='minus', exon_count=29, OMIM=614593, no_hdr0=''),
  105373187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373187, CurrentID=0, Status='live', Symbol='LOC105373187', Aliases='', description='uncharacterized LOC105373187', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=42456599, end_position_on_the_genomic_accession=42710060, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369708, CurrentID=0, Status='live', Symbol='LOC105369708', Aliases='', description='uncharacterized LOC105369708', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27280648, end_position_on_the_genomic_accession=27291550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  140738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140738, CurrentID=0, Status='live', Symbol='TMEM37', Aliases='PR, PR1', description='transmembrane protein 37', other_designations='neuronal voltage-gated calcium channel gamma-like subunit|voltage-dependent calcium channel gamma subunit-like protein', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119429912, end_position_on_the_genomic_accession=119438520, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369709, CurrentID=0, Status='live', Symbol='LOC105369709', Aliases='', description='uncharacterized LOC105369709', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27811102, end_position_on_the_genomic_accession=27824097, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  140739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140739, CurrentID=0, Status='live', Symbol='UBE2F', Aliases='NCE2', description='ubiquitin conjugating enzyme E2F (putative)', other_designations='NEDD8 carrier protein UBE2F|NEDD8 protein ligase UBE2F|NEDD8-conjugating enzyme 2|ubiquitin-conjugating enzyme E2F (putative)', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237966945, end_position_on_the_genomic_accession=238042782, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105369710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369710, CurrentID=0, Status='live', Symbol='LOC105369710', Aliases='', description='uncharacterized LOC105369710', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28036664, end_position_on_the_genomic_accession=28129572, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100852393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852393, CurrentID=0, Status='live', Symbol='MRT29', Aliases='', description='Mental retardation, autosomal recessive 29', other_designations='', map_location='4q27-q28.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614333, no_hdr0=''),
  100852394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852394, CurrentID=0, Status='live', Symbol='MRT33', Aliases='', description='Mental retardation, autosomal recessive 33', other_designations='', map_location='17p13.2-p13.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614341, no_hdr0=''),
  100852395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852395, CurrentID=0, Status='live', Symbol='MRT19', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 19', other_designations='', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614343, no_hdr0=''),
  100852396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852396, CurrentID=0, Status='live', Symbol='MRT28', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 28', other_designations='', map_location='6q26-q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614347, no_hdr0=''),
  100852397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852397, CurrentID=0, Status='live', Symbol='MRT27', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 27', other_designations='', map_location='15q23-q26', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614340, no_hdr0=''),
  100852398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852398, CurrentID=0, Status='live', Symbol='MRT23', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 23', other_designations='', map_location='11p13-q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614344, no_hdr0=''),
  100852399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852399, CurrentID=0, Status='live', Symbol='MRT31', Aliases='', description='Mental retardation, autosomal recessive 31', other_designations='', map_location='4q12-q13.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614329, no_hdr0=''),
  100852400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852400, CurrentID=0, Status='live', Symbol='MRT32', Aliases='', description='Mental retardation, autosomal recessive 32', other_designations='', map_location='14q32.13-q32.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614339, no_hdr0=''),
  100852401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852401, CurrentID=0, Status='live', Symbol='MRT30', Aliases='', description='Mental retardation, autosomal recessive 30', other_designations='', map_location='6q12-q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614342, no_hdr0=''),
  100852402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852402, CurrentID=0, Status='live', Symbol='MRT24', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 24', other_designations='', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614345, no_hdr0=''),
  100852403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852403, CurrentID=0, Status='live', Symbol='MRT25', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 25', other_designations='', map_location='12q13.11-q15', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614346, no_hdr0=''),
  100852404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852404, CurrentID=0, Status='live', Symbol='DEL17Q11.2', Aliases='C17DELq11.2', description='chromosome 17q11.2 deletion syndrome', other_designations='NF1 microdeletion syndrome|neurofibromatosis 1 microdeletion syndrome', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613675, no_hdr0=''),
  100852405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852405, CurrentID=0, Status='live', Symbol='RNF185-AS1', Aliases='', description='RNF185 antisense RNA 1', other_designations='RNF185 antisense RNA 1 (non-protein coding)', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100852406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852406, CurrentID=0, Status='live', Symbol='POTEH-AS1', Aliases='LA16c-3G11.5', description='POTEH antisense RNA 1', other_designations='POTEH antisense RNA 1 (non-protein coding)', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15700386, end_position_on_the_genomic_accession=15703354, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100852407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852407, CurrentID=0, Status='live', Symbol='CARS-AS1', Aliases='', description='CARS antisense RNA 1', other_designations='CARS antisense RNA 1 (non-protein coding)', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100852408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852408, CurrentID=0, Status='live', Symbol='EGFLAM-AS4', Aliases='', description='EGFLAM antisense RNA 4', other_designations='CTD-2022B6.1|EGFLAM antisense RNA 4 (non-protein coding)', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38282269, end_position_on_the_genomic_accession=38290986, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100852409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852409, CurrentID=0, Status='live', Symbol='NEGR1-IT1', Aliases='', description='NEGR1 intronic transcript 1', other_designations='NEGR1 intronic transcript 1 (non-protein coding)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71794232, end_position_on_the_genomic_accession=71837012, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100852410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100852410, CurrentID=0, Status='live', Symbol='ZRANB2-AS2', Aliases='', description='ZRANB2 antisense RNA 2 (head to head)', other_designations='ZRANB2 antisense RNA 2 (non-protein coding)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71081324, end_position_on_the_genomic_accession=71237723, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369719, CurrentID=0, Status='live', Symbol='LOC105369719', Aliases='', description='uncharacterized LOC105369719', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30581698, end_position_on_the_genomic_accession=30600830, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  285627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285627, CurrentID=0, Status='live', Symbol='LOC285627', Aliases='', description='uncharacterized LOC285627', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159448556, end_position_on_the_genomic_accession=159466276, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  140749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140749, CurrentID=0, Status='live', Symbol='RPL5P2', Aliases='RPL5_14_1705, dJ461P17.5', description='ribosomal protein L5 pseudogene 2', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45490818, end_position_on_the_genomic_accession=45491671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140750, CurrentID=0, Status='live', Symbol='RPL36P4', Aliases='RPL36_7_1697, dJ477O4.3', description='ribosomal protein L36 pseudogene 4', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35544377, end_position_on_the_genomic_accession=35544806, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369721, CurrentID=0, Status='live', Symbol='LOC105369721', Aliases='', description='uncharacterized LOC105369721', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31558032, end_position_on_the_genomic_accession=31568862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140751, CurrentID=0, Status='live', Symbol='RPL36P2', Aliases='RPL36_8_1711, dJ530I15.5', description='ribosomal protein L36 pseudogene 2', other_designations='', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50620721, end_position_on_the_genomic_accession=50621121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100192379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100192379, CurrentID=0, Status='live', Symbol='PP12613', Aliases='', description='uncharacterized LOC100192379', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121764585, end_position_on_the_genomic_accession=121766808, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  140753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140753, CurrentID=0, Status='live', Symbol='RPL31P2', Aliases='RPL31_25_1695, dJ553F4.5', description='ribosomal protein L31 pseudogene 2', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33752128, end_position_on_the_genomic_accession=33752567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285628, CurrentID=0, Status='live', Symbol='LOC285628', Aliases='', description='MIR146A host gene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  140754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140754, CurrentID=0, Status='live', Symbol='RPS3P1', Aliases='dJ631M13.3', description='ribosomal protein S3 pseudogene 1', other_designations='', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14003509, end_position_on_the_genomic_accession=14003700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130560, CurrentID=0, Status='live', Symbol='LOC100130560', Aliases='', description='uncharacterized LOC100130560', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=88711033, end_position_on_the_genomic_accession=88711830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369725, CurrentID=0, Status='live', Symbol='LOC105369725', Aliases='', description='uncharacterized LOC105369725', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31958493, end_position_on_the_genomic_accession=31959219, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  140755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140755, CurrentID=0, Status='live', Symbol='RPS2P1', Aliases='RPS2_26_1696, dJ64K7.1', description='ribosomal protein S2 pseudogene 1', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34122346, end_position_on_the_genomic_accession=34123342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369726, CurrentID=0, Status='live', Symbol='LOC105369726', Aliases='', description='uncharacterized LOC105369726', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=33685981, end_position_on_the_genomic_accession=33691838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140756, CurrentID=0, Status='live', Symbol='RPL7AP14', Aliases='RPL7AL4, RPL7A_32_1700, dJ823N20.2', description='ribosomal protein L7a pseudogene 14', other_designations='ribosomal protein L7a-like 4', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37444674, end_position_on_the_genomic_accession=37445556, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643199, CurrentID=0, Status='live', Symbol='TAB3P1', Aliases='', description='TGF-beta activated kinase 1/MAP3K7 binding protein 3 pseudogene 1', other_designations='TAK1-binding protein 3 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13151542, end_position_on_the_genomic_accession=13162217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140757, CurrentID=0, Status='live', Symbol='RSL24D1P6', Aliases='RPL24P1, dJ836N17.3', description='ribosomal L24 domain containing 1 pseudogene 6', other_designations='ribosomal protein L24 pseudogene 1', map_location='20q11', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32170298, end_position_on_the_genomic_accession=32170778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140758, CurrentID=0, Status='live', Symbol='RPS10P2', Aliases='RPS10_15_1685, dJ858M22.1', description='ribosomal protein S10 pseudogene 2', other_designations='', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14757533, end_position_on_the_genomic_accession=14758118, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643201, CurrentID=0, Status='live', Symbol='LOC643201', Aliases='', description='centrosomal protein 192kDa pseudogene', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176143085, end_position_on_the_genomic_accession=176199295, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  140759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140759, CurrentID=0, Status='live', Symbol='RPL19P1', Aliases='RPL19_11_1678, dJ860F19.2', description='ribosomal protein L19 pseudogene 1', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2780615, end_position_on_the_genomic_accession=2781042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140760, CurrentID=0, Status='live', Symbol='RPL24P2', Aliases='RPL24_5_1691, dJ872K7.2', description='ribosomal protein L24 pseudogene 2', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21114681, end_position_on_the_genomic_accession=21115230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374549, CurrentID=0, Status='live', Symbol='LOC105374549', Aliases='', description='uncharacterized LOC105374549', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27736622, end_position_on_the_genomic_accession=27750258, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140761, CurrentID=0, Status='live', Symbol='RPS3P2', Aliases='RPS3_4_1701, dJ927M24.3', description='ribosomal protein S3 pseudogene 2', other_designations='', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38537750, end_position_on_the_genomic_accession=38538459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369733, CurrentID=0, Status='live', Symbol='LOC105369733', Aliases='', description='uncharacterized LOC105369733', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39823695, end_position_on_the_genomic_accession=39829359, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379508, CurrentID=0, Status='live', Symbol='LOC105379508', Aliases='', description='uncharacterized LOC105379508', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8405254, end_position_on_the_genomic_accession=8406461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58155, CurrentID=0, Status='live', Symbol='PTBP2', Aliases='PTBLP, brPTB, nPTB', description='polypyrimidine tract binding protein 2', other_designations='PTB-like protein|neural polypyrimidine tract binding protein|neurally-enriched homolog of PTB', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96721605, end_position_on_the_genomic_accession=96815049, orientation='plus', exon_count=16, OMIM=608449, no_hdr0=''),
  58156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58156, CurrentID=0, Status='live', Symbol='OPA4', Aliases='', description='optic atrophy 4 (autosomal dominant)', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605293, no_hdr0=''),
  58157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58157, CurrentID=0, Status='live', Symbol='NGB', Aliases='', description='neuroglobin', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77265491, end_position_on_the_genomic_accession=77271312, orientation='minus', exon_count=4, OMIM=605304, no_hdr0=''),
  58158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58158, CurrentID=0, Status='live', Symbol='NEUROD4', Aliases='ATH-3, ATH3, Atoh3, MATH-3, MATH3, bHLHa4', description='neuronal differentiation 4', other_designations='class A basic helix-loop-helix protein 4|neurogenic differentiation 4|protein atonal homolog 3', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55019945, end_position_on_the_genomic_accession=55030017, orientation='plus', exon_count=2, OMIM=611635, no_hdr0=''),
  58160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58160, CurrentID=0, Status='live', Symbol='NFE4', Aliases='NF-E4', description='nuclear factor, erythroid 4', other_designations='fetal globin activator NF-E4|gamma-globin gene activator', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102973437, end_position_on_the_genomic_accession=102978667, orientation='plus', exon_count=3, OMIM=612133, no_hdr0=''),
  58163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58163, CurrentID=0, Status='live', Symbol='RBPJP4', Aliases='K7, RBPSUHP4', description='RBPJ pseudogene 4', other_designations='rec|recombining binding protein suppressor of hairless pseudogene 4', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  58164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58164, CurrentID=0, Status='live', Symbol='RBPJP3', Aliases='K2, RBPSUHP3', description='RBPJ pseudogene 3', other_designations='recombining binding protein suppressor of hairless pseudogene 3', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  9694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9694, CurrentID=0, Status='live', Symbol='EMC2', Aliases='KIAA0103, TTC35', description='ER membrane protein complex subunit 2', other_designations='TPR repeat protein 35|tetratricopeptide repeat domain 35|tetratricopeptide repeat protein 35', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108443624, end_position_on_the_genomic_accession=108486907, orientation='plus', exon_count=11, OMIM=607722, no_hdr0=''),
  140767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140767, CurrentID=0, Status='live', Symbol='NRSN1', Aliases='VMP, p24', description='neurensin 1', other_designations='neuro-p24|vesicular membrain protein p24|vesicular membrane protein of 24 kDa|vesicular membrane protein p24', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24126186, end_position_on_the_genomic_accession=24147529, orientation='plus', exon_count=4, OMIM=616630, no_hdr0=''),
  105369738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369738, CurrentID=0, Status='live', Symbol='LOC105369738', Aliases='', description='uncharacterized LOC105369738', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42646583, end_position_on_the_genomic_accession=42652583, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372874, CurrentID=0, Status='live', Symbol='LOC105372874', Aliases='', description='uncharacterized LOC105372874', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206207623, end_position_on_the_genomic_accession=206213266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378915, CurrentID=0, Status='live', Symbol='LOC105378915', Aliases='', description='uncharacterized LOC105378915', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114899374, end_position_on_the_genomic_accession=114902463, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369740, CurrentID=0, Status='live', Symbol='LOC105369740', Aliases='', description='uncharacterized LOC105369740', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43154443, end_position_on_the_genomic_accession=43163820, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  58189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58189, CurrentID=0, Status='live', Symbol='WFDC1', Aliases='PS20', description='WAP four-disulfide core domain 1', other_designations='WAP four-disulfide core domain 1 homolog|prostate stromal protein ps20|ps20 growth inhibitor', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84294706, end_position_on_the_genomic_accession=84329851, orientation='plus', exon_count=7, OMIM=605322, no_hdr0=''),
  58190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58190, CurrentID=0, Status='live', Symbol='CTDSP1', Aliases='NIF3, NLI-IF, NLIIF, SCP1', description='CTD small phosphatase 1', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 1|NLI-interacting factor 3|nuclear LIM interactor-interacting factor 3|small C-terminal domain phosphatase 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218398338, end_position_on_the_genomic_accession=218405941, orientation='plus', exon_count=9, OMIM=605323, no_hdr0=''),
  58191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58191, CurrentID=0, Status='live', Symbol='CXCL16', Aliases='CXCLG16, SR-PSOX, SRPSOX', description='chemokine (C-X-C motif) ligand 16', other_designations='CXC chemokine ligand 16|scavenger receptor for phosphatidylserine and oxidized low density lipoprotein|small-inducible cytokine B16|transmembrane chemokine CXCL16', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4733529, end_position_on_the_genomic_accession=4739928, orientation='minus', exon_count=6, OMIM=605398, no_hdr0=''),
  140771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140771, CurrentID=0, Status='live', Symbol='SMCR5', Aliases='NCRNA00034', description='Smith-Magenis syndrome chromosome region, candidate 5 (non-protein coding)', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17776686, end_position_on_the_genomic_accession=17779529, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  140772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140772, CurrentID=0, Status='live', Symbol='SMCR6', Aliases='', description='Smith-Magenis syndrome chromosome region, candidate 6 (non-protein coding)', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370187, CurrentID=0, Status='live', Symbol='LOC105370187', Aliases='', description='uncharacterized LOC105370187', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44840895, end_position_on_the_genomic_accession=44896666, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369745, CurrentID=0, Status='live', Symbol='LOC105369745', Aliases='', description='uncharacterized LOC105369745', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45825919, end_position_on_the_genomic_accession=45849646, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  140775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140775, CurrentID=0, Status='live', Symbol='SMCR8', Aliases='', description='Smith-Magenis syndrome chromosome region, candidate 8', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18315280, end_position_on_the_genomic_accession=18328056, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643219, CurrentID=0, Status='live', Symbol='LOC643219', Aliases='', description='glycerol-3-phosphate acyltransferase 2, mitochondrial pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95792148, end_position_on_the_genomic_accession=95800191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140777, CurrentID=0, Status='live', Symbol='PDB5', Aliases='', description='Paget disease of bone 5', other_designations='', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643220, CurrentID=0, Status='live', Symbol='RPL3P6', Aliases='RPL3_2_568', description='ribosomal protein L3 pseudogene 6', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61390616, end_position_on_the_genomic_accession=61391897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140778, CurrentID=0, Status='live', Symbol='PDB6', Aliases='', description='Paget disease of bone 6', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378917, CurrentID=0, Status='live', Symbol='LOC105378917', Aliases='', description='uncharacterized LOC105378917', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115222474, end_position_on_the_genomic_accession=115282555, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369751, CurrentID=0, Status='live', Symbol='LOC105369751', Aliases='', description='uncharacterized LOC105369751', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47996599, end_position_on_the_genomic_accession=47998004, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9709, CurrentID=0, Status='live', Symbol='HERPUD1', Aliases='HERP, Mif1, SUP', description='homocysteine-inducible, endoplasmic reticulum stress-inducible, ubiquitin-like domain member 1', other_designations='MMS-inducible|homocysteine-inducible endoplasmic reticulum stress-inducible ubiquitin-like domain member 1 protein|methyl methanesulfonate (MMF)-inducible fragment protein 1', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56931765, end_position_on_the_genomic_accession=56943881, orientation='plus', exon_count=8, OMIM=608070, no_hdr0=''),
  105369752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369752, CurrentID=0, Status='live', Symbol='LOC105369752', Aliases='', description='uncharacterized LOC105369752', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47980657, end_position_on_the_genomic_accession=47982875, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378918, CurrentID=0, Status='live', Symbol='LOC105378918', Aliases='', description='uncharacterized LOC105378918', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116121320, end_position_on_the_genomic_accession=116124054, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369755, CurrentID=0, Status='live', Symbol='LOC105369755', Aliases='', description='uncharacterized LOC105369755', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48851439, end_position_on_the_genomic_accession=48852549, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369756, CurrentID=0, Status='live', Symbol='LOC105369756', Aliases='', description='uncharacterized LOC105369756', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48987781, end_position_on_the_genomic_accession=48988743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369757, CurrentID=0, Status='live', Symbol='LOC105369757', Aliases='', description='uncharacterized LOC105369757', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48982499, end_position_on_the_genomic_accession=48987194, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369760, CurrentID=0, Status='live', Symbol='LOC105369760', Aliases='', description='uncharacterized LOC105369760', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49131626, end_position_on_the_genomic_accession=49147865, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369761, CurrentID=0, Status='live', Symbol='LOC105369761', Aliases='', description='uncharacterized LOC105369761', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49550340, end_position_on_the_genomic_accession=49553199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369763, CurrentID=0, Status='live', Symbol='LOC105369763', Aliases='', description='uncharacterized LOC105369763', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49942747, end_position_on_the_genomic_accession=49946888, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369764, CurrentID=0, Status='live', Symbol='LOC105369764', Aliases='', description='uncharacterized LOC105369764', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49965170, end_position_on_the_genomic_accession=49970532, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  285636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285636, CurrentID=0, Status='live', Symbol='C5orf51', Aliases='', description='chromosome 5 open reading frame 51', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41904289, end_position_on_the_genomic_accession=41921636, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369769, CurrentID=0, Status='live', Symbol='LOC105369769', Aliases='', description='uncharacterized LOC105369769', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  9729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9729, CurrentID=0, Status='live', Symbol='KIAA0408', Aliases='', description='KIAA0408', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127438406, end_position_on_the_genomic_accession=127459390, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369772, CurrentID=0, Status='live', Symbol='LOC105369772', Aliases='', description='uncharacterized LOC105369772', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52629765, end_position_on_the_genomic_accession=52631549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  140802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140802, CurrentID=0, Status='live', Symbol='LGV1', Aliases='', description='longevity 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=152430, no_hdr0=''),
  9731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9731, CurrentID=0, Status='live', Symbol='CEP104', Aliases='GlyBP, KIAA0562', description='centrosomal protein 104kDa', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3812081, end_position_on_the_genomic_accession=3857387, orientation='minus', exon_count=24, OMIM=616690, no_hdr0=''),
  105369774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369774, CurrentID=0, Status='live', Symbol='LOC105369774', Aliases='', description='uncharacterized LOC105369774', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53726426, end_position_on_the_genomic_accession=53731221, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  140805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140805, CurrentID=0, Status='live', Symbol='HT', Aliases='', description='Hashimoto thyroiditis', other_designations='', map_location='8q23-q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=140300, no_hdr0=''),
  140807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140807, CurrentID=0, Status='live', Symbol='KRT72', Aliases='K6IRS2, K6irs, KRT6, KRT6IRS2', description='keratin 72, type II', other_designations='CK-72|K72|cytokeratin-72|keratin 6, inner root sheath, 2|keratin protein K6irs|keratin-72|type II inner root sheath-specific keratin-K6irs2|type-II keratin Kb35', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52585589, end_position_on_the_genomic_accession=52602944, orientation='minus', exon_count=14, OMIM=608246, no_hdr0=''),
  105378923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378923, CurrentID=0, Status='live', Symbol='LOC105378923', Aliases='', description='uncharacterized LOC105378923', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116487140, end_position_on_the_genomic_accession=116502637, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  9737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9737, CurrentID=0, Status='live', Symbol='GPRASP1', Aliases='GASP, GASP-1, GASP1', description='G protein-coupled receptor associated sorting protein 1', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102651366, end_position_on_the_genomic_accession=102659083, orientation='plus', exon_count=6, OMIM=300417, no_hdr0=''),
  105369781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369781, CurrentID=0, Status='live', Symbol='LOC105369781', Aliases='', description='uncharacterized LOC105369781', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56010092, end_position_on_the_genomic_accession=56014750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928733, CurrentID=0, Status='live', Symbol='LOC101928733', Aliases='', description='uncharacterized LOC101928733', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149398204, end_position_on_the_genomic_accession=149401456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369783, CurrentID=0, Status='live', Symbol='LOC105369783', Aliases='', description='uncharacterized LOC105369783', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57826409, end_position_on_the_genomic_accession=57843761, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374063, CurrentID=0, Status='live', Symbol='LOC105374063', Aliases='', description='uncharacterized LOC105374063', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119665615, end_position_on_the_genomic_accession=119668565, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369787, CurrentID=0, Status='live', Symbol='LOC105369787', Aliases='', description='uncharacterized LOC105369787', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58335839, end_position_on_the_genomic_accession=58336628, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  58472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58472, CurrentID=0, Status='live', Symbol='SQRDL', Aliases='CGI-44, PRO1975, SQOR', description='sulfide quinone reductase-like (yeast)', other_designations='sulfide dehydrogenase like', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45631148, end_position_on_the_genomic_accession=45691294, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  58473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58473, CurrentID=0, Status='live', Symbol='PLEKHB1', Aliases='KPL1, PHR1, PHRET1', description='pleckstrin homology domain containing B1', other_designations='PH domain containing protein in retina 1|PH domain containing, retinal 1|PH domain-containing family B member 1|PH domain-containing protein in retina 1|evectin-1|pleckstrin homology domain containing, family B (evectins) member 1|pleckstrin homology domain retinal protein 1', map_location='11q13.5-q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73646178, end_position_on_the_genomic_accession=73662819, orientation='plus', exon_count=11, OMIM=607651, no_hdr0=''),
  58475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58475, CurrentID=0, Status='live', Symbol='MS4A7', Aliases='4SPAN2, CD20L4, CFFM4, MS4A8', description='membrane-spanning 4-domains, subfamily A, member 7', other_designations='CD20 antigen-like 4|CD20/Fc-epsilon-RI-beta family member 4|four-span transmembrane protein 2|high affinity immunoglobulin epsilon receptor beta subunit', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60378485, end_position_on_the_genomic_accession=60395954, orientation='plus', exon_count=7, OMIM=606502, no_hdr0=''),
  58476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58476, CurrentID=0, Status='live', Symbol='TP53INP2', Aliases='C20orf110, DOR, PINH, dJ1181N3.1', description='tumor protein p53 inducible nuclear protein 2', other_designations='PIG-U|diabetes and obesity regulated|diabetes and obesity-regulated|p53-inducible protein U', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34704311, end_position_on_the_genomic_accession=34713439, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  58477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58477, CurrentID=0, Status='live', Symbol='SRPRB', Aliases='APMCF1', description='SRP receptor beta subunit', other_designations='SR-beta|signal recognition particle receptor, B subunit|signal recognition particle receptor, beta subunit', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133784033, end_position_on_the_genomic_accession=133821492, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58478, CurrentID=0, Status='live', Symbol='ENOPH1', Aliases='E1, MASA, MST145, mtnC', description='enolase-phosphatase 1', other_designations='2,3-diketo-5-methylthio-1-phosphopentane phosphatase|acireductone synthase', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82430480, end_position_on_the_genomic_accession=82461175, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  58480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58480, CurrentID=0, Status='live', Symbol='RHOU', Aliases='ARHU, CDC42L1, G28K, WRCH1, hG28K', description='ras homolog family member U', other_designations='2310026M05Rik|CDC42-like GTPase 1|GTP-binding protein SB128|GTP-binding protein like 1|Ryu GTPase|ras homolog gene family, member U|ras-like gene family member U|rho GTPase-like protein ARHU|wnt-1 responsive Cdc42 homolog 1', map_location='1q42.11-q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228644647, end_position_on_the_genomic_accession=228746669, orientation='plus', exon_count=4, OMIM=606366, no_hdr0=''),
  58483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58483, CurrentID=0, Status='live', Symbol='LINC00474', Aliases='C9orf27, EST-YD1', description='long intergenic non-protein coding RNA 474', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115888265, end_position_on_the_genomic_accession=115925098, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  58484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58484, CurrentID=0, Status='live', Symbol='NLRC4', Aliases='AIFEC, CARD12, CLAN, CLAN1, CLANA, CLANB, CLANC, CLAND, CLR2.1, FCAS4, IPAF', description='NLR family, CARD domain containing 4', other_designations='CARD, LRR, and NACHT-containing protein|ICE-protease activating factor|NOD-like receptor C4|caspase recruitment domain family, member 12|caspase recruitment domain-containing protein 12|nucleotide-binding oligomerization domain, leucine rich repeat and CARD domain containing 4', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32224449, end_position_on_the_genomic_accession=32265743, orientation='minus', exon_count=10, OMIM=606831, no_hdr0=''),
  58485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58485, CurrentID=0, Status='live', Symbol='TRAPPC1', Aliases='BET5, MUM2', description='trafficking protein particle complex 1', other_designations='BET5 homolog|MUM-2|melanoma ubiquitous mutated 2|multiple myeloma protein 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7930345, end_position_on_the_genomic_accession=7931999, orientation='minus', exon_count=5, OMIM=610969, no_hdr0=''),
  58486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58486, CurrentID=0, Status='live', Symbol='ZBED5', Aliases='Buster1', description='zinc finger, BED-type containing 5', other_designations='transposon-derived Buster1 transposase-like protein', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10852704, end_position_on_the_genomic_accession=10858073, orientation='minus', exon_count=3, OMIM=615251, no_hdr0=''),
  58487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58487, CurrentID=0, Status='live', Symbol='CREBZF', Aliases='SMILE, ZF', description='CREB/ATF bZIP transcription factor', other_designations='HCF-binding transcription factor Zhangfei|SHP-interacting leucine zipper protein|Zhangfei|host cell factor-binding transcription factor Zhangfei', map_location='11q14', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85657564, end_position_on_the_genomic_accession=85665138, orientation='minus', exon_count=3, OMIM=606444, no_hdr0=''),
  58488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58488, CurrentID=0, Status='live', Symbol='PCTP', Aliases='PC-TP, STARD2', description='phosphatidylcholine transfer protein', other_designations='START domain-containing protein 2|StAR-related lipid transfer (START) domain containing 2|stAR-related lipid transfer protein 2', map_location='17q21-q24', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55750979, end_position_on_the_genomic_accession=55843209, orientation='plus', exon_count=9, OMIM=606055, no_hdr0=''),
  58489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58489, CurrentID=0, Status='live', Symbol='ABHD17C', Aliases='FAM108C1', description='abhydrolase domain containing 17C', other_designations='abhydrolase domain-containing protein FAM108C1|family with sequence similarity 108, member C1', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80695311, end_position_on_the_genomic_accession=80755621, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58490, CurrentID=0, Status='live', Symbol='RPRD1B', Aliases='C20orf77, CREPT, NET60, dJ1057B20.2', description='regulation of nuclear pre-mRNA domain containing 1B', other_designations='cell cycle-related and expression-elevated protein in tumor|cell-cycle related and expression-elevated protein in tumor', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38033487, end_position_on_the_genomic_accession=38092364, orientation='plus', exon_count=11, OMIM=614694, no_hdr0=''),
  58491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58491, CurrentID=0, Status='live', Symbol='ZNF71', Aliases='EZFIT', description='zinc finger protein 71', other_designations='Kruppel-related zinc finger protein', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56595296, end_position_on_the_genomic_accession=56624481, orientation='plus', exon_count=4, OMIM=194545, no_hdr0=''),
  58492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58492, CurrentID=0, Status='live', Symbol='ZNF77', Aliases='pT1', description='zinc finger protein 77', other_designations='ZNFpT1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2933218, end_position_on_the_genomic_accession=2945279, orientation='minus', exon_count=5, OMIM=194551, no_hdr0=''),
  58493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58493, CurrentID=0, Status='live', Symbol='INIP', Aliases='C9orf80, HSPC043, MISE, SOSSC, SSBIP1, hSSBIP1', description='INTS3 and NABP interacting protein', other_designations='INTS3- and NABP-interacting protein|SOSS-C|SSB-interacting protein 1|hSSB-interacting protein 1|minute INTS3/hSSB-associated element|sensor of single-strand DNA complex subunit C|sensor of ssDNA subunit C|single-stranded DNA-binding protein-interacting protein 1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112685564, end_position_on_the_genomic_accession=112718107, orientation='minus', exon_count=7, OMIM=613273, no_hdr0=''),
  58494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58494, CurrentID=0, Status='live', Symbol='JAM2', Aliases='C21orf43, CD322, JAM-B, JAMB, PRO245, VE-JAM, VEJAM', description='junctional adhesion molecule 2', other_designations='JAM-2|JAM-IT/VE-JAM|vascular endothelial junction-associated molecule', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25639282, end_position_on_the_genomic_accession=25717562, orientation='plus', exon_count=12, OMIM=606870, no_hdr0=''),
  58495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58495, CurrentID=0, Status='live', Symbol='OVOL2', Aliases='EUROIMAGE566589, ZNF339', description='ovo-like zinc finger 2', other_designations='zinc finger protein 339', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18024152, end_position_on_the_genomic_accession=18059188, orientation='minus', exon_count=5, OMIM=616441, no_hdr0=''),
  58496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58496, CurrentID=0, Status='live', Symbol='LY6G5B', Aliases='C6orf19, G5b', description='lymphocyte antigen 6 complex, locus G5B', other_designations='lymphocyte antigen-6 G5B splicing isoform 288|lymphocyte antigen-6 G5B splicing isoform 452|lymphocyte antigen-6 G5B splicing isoform 690|lymphocyte antigen-6 G5B splicing isoform 837', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31670951, end_position_on_the_genomic_accession=31672450, orientation='plus', exon_count=3, OMIM=610433, no_hdr0=''),
  58497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58497, CurrentID=0, Status='live', Symbol='PRUNE', Aliases='DRES-17, DRES17, HTCD37', description='prune exopolyphosphatase', other_designations='Drosophila-related expressed sequence 17', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151008391, end_position_on_the_genomic_accession=151035713, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58498, CurrentID=0, Status='live', Symbol='MYL7', Aliases='MYL2A, MYLC2A', description='myosin light chain 7', other_designations='MLC-2a|MLC2a|myosin light chain 2a|myosin regulatory light chain 7|myosin, light chain 7, regulatory|myosin, light polypeptide 7, regulatory', map_location='7p21-p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44138864, end_position_on_the_genomic_accession=44141388, orientation='minus', exon_count=7, OMIM=613993, no_hdr0=''),
  58499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58499, CurrentID=0, Status='live', Symbol='ZNF462', Aliases='Zfp462', description='zinc finger protein 462', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106860165, end_position_on_the_genomic_accession=107012455, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  58500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58500, CurrentID=0, Status='live', Symbol='ZNF250', Aliases='ZFP647, ZNF647', description='zinc finger protein 250', other_designations='zinc finger protein (clone 647)|zinc finger protein 647', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144876951, end_position_on_the_genomic_accession=144902305, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58503, CurrentID=0, Status='live', Symbol='PROL1', Aliases='BPLP, PRL1, opiorphin', description='proline rich, lacrimal 1', other_designations='basic proline-rich lacrimal protein|basic proline-rich protein|proline-rich 1', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70397882, end_position_on_the_genomic_accession=70410197, orientation='plus', exon_count=4, OMIM=608936, no_hdr0=''),
  58504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58504, CurrentID=0, Status='live', Symbol='ARHGAP22', Aliases='RhoGAP2, RhoGap22', description='Rho GTPase activating protein 22', other_designations='rho-type GTPase-activating protein 22', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48430318, end_position_on_the_genomic_accession=48656265, orientation='minus', exon_count=20, OMIM=610585, no_hdr0=''),
  58505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58505, CurrentID=0, Status='live', Symbol='OSTC', Aliases='DC2', description='oligosaccharyltransferase complex subunit (non-catalytic)', other_designations='hydrophobic protein HSF-28', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108650585, end_position_on_the_genomic_accession=108667822, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  58506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58506, CurrentID=0, Status='live', Symbol='SCAF1', Aliases='SRA1', description='SR-related CTD-associated factor 1', other_designations='SCAF|SR-related and CTD-associated factor 1|SR-related-CTD-associated factor|serine arginine-rich pre-mRNA splicing factor SR-A1', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49640503, end_position_on_the_genomic_accession=49658649, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  9751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9751, CurrentID=0, Status='live', Symbol='SNPH', Aliases='', description='syntaphilin', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1266297, end_position_on_the_genomic_accession=1309327, orientation='plus', exon_count=8, OMIM=604942, no_hdr0=''),
  58508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58508, CurrentID=0, Status='live', Symbol='KMT2C', Aliases='HALR, MLL3', description='lysine (K)-specific methyltransferase 2C', other_designations='ALR-like protein|histone-lysine N-methyltransferase MLL3|histone-lysine N-methyltransferase, H3 lysine-4 specific|homologous to ALR protein|myeloid/lymphoid or mixed-lineage leukemia protein 3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152134925, end_position_on_the_genomic_accession=152436638, orientation='minus', exon_count=64, OMIM=606833, no_hdr0=''),
  58509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58509, CurrentID=0, Status='live', Symbol='CACTIN', Aliases='C19orf29, NY-REN-24, fSAPc', description='cactin, spliceosome C complex subunit', other_designations='NY REN 24 antigen|functional spliceosome-associated protein c|renal carcinoma antigen NY-REN-24', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3610628, end_position_on_the_genomic_accession=3626815, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  58510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58510, CurrentID=0, Status='live', Symbol='PRODH2', Aliases='HSPOX1, HYPDH', description='proline dehydrogenase (oxidase) 2', other_designations='kidney and liver proline oxidase 1|probable proline oxidase 2', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35799990, end_position_on_the_genomic_accession=35813299, orientation='minus', exon_count=11, OMIM=616377, no_hdr0=''),
  58511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58511, CurrentID=0, Status='live', Symbol='DNASE2B', Aliases='DLAD', description='deoxyribonuclease II beta', other_designations='DNase II beta|DNase II-like acid DNase|DNase2-like acid DNase|endonuclease DLAD|lysosomal DNase II', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84398532, end_position_on_the_genomic_accession=84415895, orientation='plus', exon_count=6, OMIM=608057, no_hdr0=''),
  58512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58512, CurrentID=0, Status='live', Symbol='DLGAP3', Aliases='DAP3, SAPAP3', description='discs, large (Drosophila) homolog-associated protein 3', other_designations='DAP-3|PSD-95/SAP90-binding protein 3|SAP90/PSD 95 associated protein 3|SAP90/PSD-95-associated protein 3', map_location='1p35.3-p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34862120, end_position_on_the_genomic_accession=34929585, orientation='minus', exon_count=14, OMIM=611413, no_hdr0=''),
  58513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58513, CurrentID=0, Status='live', Symbol='EPS15L1', Aliases='EPS15R', description='epidermal growth factor receptor pathway substrate 15-like 1', other_designations='epidermal growth factor receptor substrate EPS15R|eps15-related protein', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16355244, end_position_on_the_genomic_accession=16472012, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  58515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58515, CurrentID=0, Status='live', Symbol='SELK', Aliases='HSPC030, HSPC297, SelK', description='selenoprotein K', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53885199, end_position_on_the_genomic_accession=53891962, orientation='minus', exon_count=5, OMIM=607916, no_hdr0=''),
  58516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58516, CurrentID=0, Status='live', Symbol='FAM60A', Aliases='C12orf14, L4, TERA', description='family with sequence similarity 60 member A', other_designations='family with sequence similarity 60, member A|tera protein homolog', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31280422, end_position_on_the_genomic_accession=31326225, orientation='minus', exon_count=10, OMIM=615027, no_hdr0=''),
  58517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58517, CurrentID=0, Status='live', Symbol='RBM25', Aliases='NET52, RED120, RNPC7, S164, Snu71, fSAP94', description='RNA binding motif protein 25', other_designations='Arg/Glu/Asp-rich protein, 120 kDa|RNA-binding motif protein 25|RNA-binding region (RNP1, RRM) containing 7|RNA-binding region-containing protein 7|U1 small nuclear ribonucleoprotein 1SNRP homolog|arg/Glu/Asp-rich protein of 120 kDa|functional spliceosome-associated protein 94', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73058436, end_position_on_the_genomic_accession=73121368, orientation='plus', exon_count=21, OMIM=612427, no_hdr0=''),
  9753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9753, CurrentID=0, Status='live', Symbol='ZSCAN12', Aliases='ZFP96, ZNF29K1, ZNF305, ZNF96, dJ29K1.2', description='zinc finger and SCAN domain containing 12', other_designations='zinc finger protein 305|zinc finger protein 96', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28378821, end_position_on_the_genomic_accession=28399767, orientation='minus', exon_count=7, OMIM=603978, no_hdr0=''),
  105369796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369796, CurrentID=0, Status='live', Symbol='LOC105369796', Aliases='', description='uncharacterized LOC105369796', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  58524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58524, CurrentID=0, Status='live', Symbol='DMRT3', Aliases='DMRTA3', description='doublesex and mab-3 related transcription factor 3', other_designations='DMRT-like family A3|testis-specific protein', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=976968, end_position_on_the_genomic_accession=991736, orientation='plus', exon_count=3, OMIM=614754, no_hdr0=''),
  58525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58525, CurrentID=0, Status='live', Symbol='WIZ', Aliases='ZNF803', description='widely interspaced zinc finger motifs', other_designations='WIZ zinc finger|widely-interspaced zinc finger motifs|widely-interspaced zinc finger-containing protein|zinc finger protein 803', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15421507, end_position_on_the_genomic_accession=15449956, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  58526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58526, CurrentID=0, Status='live', Symbol='MID1IP1', Aliases='G12-like, MIG12, S14R, STRAIT11499, THRSPL', description='MID1 interacting protein 1', other_designations='MID1 interacting G12-like protein|MID1 interacting protein 1 (gastrulation specific G12-like)|gastrulation specific G12 homolog|spot 14-R|spot 14-related protein', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38801432, end_position_on_the_genomic_accession=38806532, orientation='plus', exon_count=3, OMIM=300961, no_hdr0=''),
  58527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58527, CurrentID=0, Status='live', Symbol='ABRACL', Aliases='C6orf115, Costars, HSPC280, PRO2013', description='ABRA C-terminal like', other_designations='ABRA C-terminal-like protein', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139028682, end_position_on_the_genomic_accession=139043302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58528, CurrentID=0, Status='live', Symbol='RRAGD', Aliases='RAGD, bA11D8.2.1', description='Ras-related GTP binding D', other_designations='Rag D protein', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89364616, end_position_on_the_genomic_accession=89412276, orientation='minus', exon_count=8, OMIM=608268, no_hdr0=''),
  58529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58529, CurrentID=0, Status='live', Symbol='MYOZ1', Aliases='CS-2, FATZ, MYOZ', description='myozenin 1', other_designations='calsarcin-2|filamin-, actinin- and telethonin-binding protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73631612, end_position_on_the_genomic_accession=73641757, orientation='minus', exon_count=6, OMIM=605603, no_hdr0=''),
  58530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58530, CurrentID=0, Status='live', Symbol='LY6G6D', Aliases='C6orf23, G6D, LY6-D, MEGT1, NG25', description='lymphocyte antigen 6 complex, locus G6D', other_designations='lymphocyte antigen-6 G6D|megakaryocyte-enhanced gene transcript 1 protein|protein Ly6-D', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31715356, end_position_on_the_genomic_accession=31717804, orientation='plus', exon_count=3, OMIM=606038, no_hdr0=''),
  58531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58531, CurrentID=0, Status='live', Symbol='PRM3', Aliases='', description='protamine 3', other_designations='gene 4|sperm protamine P3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11273199, end_position_on_the_genomic_accession=11273595, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  58533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58533, CurrentID=0, Status='live', Symbol='SNX6', Aliases='MSTP010, TFAF2', description='sorting nexin 6', other_designations='TRAF4-associated factor 2|tumor necrosis factor receptor-associated factor 4(TRAF4)-associated factor 2', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34561409, end_position_on_the_genomic_accession=34630160, orientation='minus', exon_count=14, OMIM=606098, no_hdr0=''),
  58538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58538, CurrentID=0, Status='live', Symbol='MPP4', Aliases='ALS2CR5, DLG6', description='membrane protein, palmitoylated 4', other_designations='amyotrophic lateral sclerosis 2 chromosomal region candidate gene 5 protein|discs large homolog 6|discs, large homolog 6|membrane protein, palmitoylated 4 (MAGUK p55 subfamily member 4)', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201644874, end_position_on_the_genomic_accession=201698834, orientation='minus', exon_count=22, OMIM=606575, no_hdr0=''),
  285643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285643, CurrentID=0, Status='live', Symbol='KIF4B', Aliases='', description='kinesin family member 4B', other_designations='chromokinesin-B', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=155013700, end_position_on_the_genomic_accession=155018125, orientation='plus', exon_count=1, OMIM=609184, no_hdr0=''),
  140831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140831, CurrentID=0, Status='live', Symbol='ZSWIM3', Aliases='C20orf164, PPP1R174', description='zinc finger, SWIM-type containing 3', other_designations='protein phosphatase 1, regulatory subunit 174|zinc finger, SWIM domain containing 3', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45857581, end_position_on_the_genomic_accession=45879131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140832, CurrentID=0, Status='live', Symbol='WFDC10A', Aliases='C20orf146, WAP10, dJ688G8.3', description='WAP four-disulfide core domain 10A', other_designations='putative protease inhibitor WAP10A', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45629746, end_position_on_the_genomic_accession=45631192, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369803, CurrentID=0, Status='live', Symbol='LOC105369803', Aliases='', description='uncharacterized LOC105369803', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64515381, end_position_on_the_genomic_accession=64537416, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378928, CurrentID=0, Status='live', Symbol='LOC105378928', Aliases='', description='uncharacterized LOC105378928', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117759856, end_position_on_the_genomic_accession=117796481, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140834, CurrentID=0, Status='live', Symbol='LINC01620', Aliases='C20orf62, dJ1013A22.3', description='long intergenic non-protein coding RNA 1620', other_designations='uncharacterized protein C20orf62', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44461763, end_position_on_the_genomic_accession=44465344, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132524, CurrentID=0, Status='live', Symbol='LOC100132524', Aliases='', description='UBX domain protein 2A pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34373029, end_position_on_the_genomic_accession=34373827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  9764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9764, CurrentID=0, Status='live', Symbol='KIAA0513', Aliases='', description='KIAA0513', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85027751, end_position_on_the_genomic_accession=85094230, orientation='plus', exon_count=15, OMIM=611675, no_hdr0=''),
  140838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140838, CurrentID=0, Status='live', Symbol='NANP', Aliases='C20orf147, HDHD4, dJ694B14.3', description='N-acetylneuraminic acid phosphatase', other_designations='Neu5Ac-9-Pase|haloacid dehalogenase-like hydrolase domain containing 4|haloacid dehalogenase-like hydrolase domain-containing protein 4', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25612935, end_position_on_the_genomic_accession=25624012, orientation='minus', exon_count=2, OMIM=610763, no_hdr0=''),
  105369809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369809, CurrentID=0, Status='live', Symbol='LOC105369809', Aliases='', description='uncharacterized LOC105369809', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65473891, end_position_on_the_genomic_accession=65489889, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369812, CurrentID=0, Status='live', Symbol='LOC105369812', Aliases='', description='uncharacterized LOC105369812', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67354654, end_position_on_the_genomic_accession=67455618, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  647208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647208, CurrentID=0, Status='live', Symbol='LOC647208', Aliases='', description="protein phosphatase 2, regulatory subunit B'', beta pseudogene", other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33495946, end_position_on_the_genomic_accession=33496092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140846, CurrentID=0, Status='live', Symbol='FAM83C-AS1', Aliases='C20orf120, NCRNA00154, dJ614O4.3', description='FAM83C antisense RNA 1', other_designations='FAM83C antisense RNA 1 (non-protein coding)', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35285251, end_position_on_the_genomic_accession=35285772, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100189516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189516, CurrentID=0, Status='live', Symbol='NMTRQ-TTG2-1', Aliases='TRNAQ48P', description='nuclear-encoded mitochondrial transfer RNA-Gln (TTG) 2-1', other_designations='nmt-tRNA-Gln-TTG-2-1|transfer RNA glutamine 48 (anticodon UUG) pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141803557, end_position_on_the_genomic_accession=141803628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140848, CurrentID=0, Status='live', Symbol='MACROD2-IT1', Aliases='C20orf68, NCRNA00227, bA467D7.4', description='MACROD2 intronic transcript 1', other_designations='MACROD2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14554384, end_position_on_the_genomic_accession=14628908, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  285647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285647, CurrentID=0, Status='live', Symbol='SUDS3P1', Aliases='', description='SDS3 homolog, SIN3A corepressor complex component pseudogene 1', other_designations='SUDS3 pseudogene 1|suppressor of defective silencing 3 homolog pseudogene 1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177971178, end_position_on_the_genomic_accession=177973821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140849, CurrentID=0, Status='live', Symbol='LINC00266-1', Aliases='C20orf69, NCRNA00266, NCRNA00266-1, bA476I15.3', description='long intergenic non-protein coding RNA 266-1', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64290385, end_position_on_the_genomic_accession=64303354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9778, CurrentID=0, Status='live', Symbol='KIAA0232', Aliases='', description='KIAA0232', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6782732, end_position_on_the_genomic_accession=6884172, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100132778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132778, CurrentID=0, Status='live', Symbol='NACAP6', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 6', other_designations='nascent polypeptide-associated complex alpha subunit pseudogene|nascent-polypeptide-associated complex alpha polypeptide pseudogene 6', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16192098, end_position_on_the_genomic_accession=16192707, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140851, CurrentID=0, Status='live', Symbol='MT1P3', Aliases='C20orf127, MTL4, dJ614O4.6', description='metallothionein 1 pseudogene 3', other_designations='metallothionein-like 4', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35218009, end_position_on_the_genomic_accession=35218185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140856, CurrentID=0, Status='live', Symbol='SCP2D1', Aliases='C20orf79, HSD22, dJ1068E13.2', description='SCP2 sterol-binding domain containing 1', other_designations='sterol carrier protein 2-like protein', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18813726, end_position_on_the_genomic_accession=18814392, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  140857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140857, CurrentID=0, Status='live', Symbol='RPL7P3', Aliases='dJ353C17.2', description='ribosomal protein L7 pseudogene 3', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62572671, end_position_on_the_genomic_accession=62573385, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369828, CurrentID=0, Status='live', Symbol='LOC105369828', Aliases='', description='uncharacterized LOC105369828', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70521520, end_position_on_the_genomic_accession=70571019, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378933, CurrentID=0, Status='live', Symbol='LOC105378933', Aliases='', description='uncharacterized LOC105378933', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119000344, end_position_on_the_genomic_accession=119001461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369832, CurrentID=0, Status='live', Symbol='LOC105369832', Aliases='', description='uncharacterized LOC105369832', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71276086, end_position_on_the_genomic_accession=71315893, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  140862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140862, CurrentID=0, Status='live', Symbol='ISM1', Aliases='C20orf82, ISM, Isthmin, bA149I18.1, dJ1077I2.1', description='isthmin 1, angiogenesis inhibitor', other_designations='isthmin 1 homolog', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13221771, end_position_on_the_genomic_accession=13300650, orientation='plus', exon_count=6, OMIM=615793, no_hdr0=''),
  140863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140863, CurrentID=0, Status='live', Symbol='SPINT5P', Aliases='C20orf168, dJ447F3.6', description='serine peptidase inhibitor, Kunitz type 5, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45749920, end_position_on_the_genomic_accession=45750074, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420839, CurrentID=0, Status='live', Symbol='LOC100420839', Aliases='', description='chromosome 18 open reading frame 21 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157372554, end_position_on_the_genomic_accession=157373170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643307, CurrentID=0, Status='live', Symbol='LOC643307', Aliases='', description='actin, beta pseudogene', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61831947, end_position_on_the_genomic_accession=61833024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140865, CurrentID=0, Status='live', Symbol='LINC00686', Aliases='C20orf90', description='long intergenic non-protein coding RNA 686', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369836, CurrentID=0, Status='live', Symbol='LOC105369836', Aliases='', description='uncharacterized LOC105369836', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71736407, end_position_on_the_genomic_accession=71738181, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369837, CurrentID=0, Status='live', Symbol='LOC105369837', Aliases='', description='uncharacterized LOC105369837', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=72148017, end_position_on_the_genomic_accession=72181492, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369838, CurrentID=0, Status='live', Symbol='LOC105369838', Aliases='', description='uncharacterized LOC105369838', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=72919850, end_position_on_the_genomic_accession=73042903, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  285829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285829, CurrentID=0, Status='live', Symbol='SUMO2P1', Aliases='SMT3Bp, SMT3H2P, SUMO2P, dJ271M21', description='SUMO2 pseudogene 1', other_designations='SMT3 (suppressor of mif two 3, yeast) homolog 2 pseudogene|SMT3 suppressor of mif two 3 homolog 2 pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29635454, end_position_on_the_genomic_accession=29636504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369839, CurrentID=0, Status='live', Symbol='LOC105369839', Aliases='', description='uncharacterized LOC105369839', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=73290799, end_position_on_the_genomic_accession=73342971, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  140870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140870, CurrentID=0, Status='live', Symbol='WFDC6', Aliases='C20orf171, HEL-S-295, WAP6, dJ461P17.11', description='WAP four-disulfide core domain 6', other_designations='epididymis secretory protein Li 295|protease inhibitor WAP6', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45534197, end_position_on_the_genomic_accession=45539495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374571, CurrentID=0, Status='live', Symbol='LOC105374571', Aliases='', description='uncharacterized LOC105374571', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43774157, end_position_on_the_genomic_accession=43774992, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369842, CurrentID=0, Status='live', Symbol='LOC105369842', Aliases='', description='uncharacterized LOC105369842', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74847451, end_position_on_the_genomic_accession=74898846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  9801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9801, CurrentID=0, Status='live', Symbol='MRPL19', Aliases='L19mt, MRP-L15, MRP-L19, MRPL15, RLX1, RPML15', description='mitochondrial ribosomal protein L19', other_designations='39S ribosomal protein L15, mitochondrial|L15mt', map_location='2p11.1-q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75646783, end_position_on_the_genomic_accession=75662208, orientation='plus', exon_count=6, OMIM=611832, no_hdr0=''),
  9804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9804, CurrentID=0, Status='live', Symbol='TOMM20', Aliases='MAS20, MOM19, TOM20', description='translocase of outer mitochondrial membrane 20 homolog (yeast)', other_designations='mitochondrial 20 kDa outer membrane protein|outer mitochondrial membrane receptor Tom20|translocase of outer mitochondrial membrane 20 homolog type II', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235109341, end_position_on_the_genomic_accession=235128941, orientation='minus', exon_count=5, OMIM=601848, no_hdr0=''),
  105369848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369848, CurrentID=0, Status='live', Symbol='LOC105369848', Aliases='', description='uncharacterized LOC105369848', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76259852, end_position_on_the_genomic_accession=76293837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370208, CurrentID=0, Status='live', Symbol='LOC105370208', Aliases='', description='uncharacterized LOC105370208', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52552097, end_position_on_the_genomic_accession=52600372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  9808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9808, CurrentID=0, Status='live', Symbol='KIAA0087', Aliases='', description='KIAA0087', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26533121, end_position_on_the_genomic_accession=26538825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369853, CurrentID=0, Status='live', Symbol='LOC105369853', Aliases='', description='uncharacterized LOC105369853', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77122435, end_position_on_the_genomic_accession=77152104, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  140883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140883, CurrentID=0, Status='live', Symbol='ZNF280B', Aliases="5'OY11.1, D87009.C22.3, SUHW2, ZNF279, ZNF632", description='zinc finger protein 280B', other_designations='suppressor of hairy wing homolog 2|zinc finger protein 279|zinc finger protein 632', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22484421, end_position_on_the_genomic_accession=22509187, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378938, CurrentID=0, Status='live', Symbol='LOC105378938', Aliases='', description='uncharacterized LOC105378938', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119829294, end_position_on_the_genomic_accession=119846812, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369855, CurrentID=0, Status='live', Symbol='LOC105369855', Aliases='', description='uncharacterized LOC105369855', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77309929, end_position_on_the_genomic_accession=77317199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372320, CurrentID=0, Status='live', Symbol='LOC105372320', Aliases='', description='uncharacterized LOC105372320', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20864384, end_position_on_the_genomic_accession=20972138, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  9813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9813, CurrentID=0, Status='live', Symbol='EFCAB14', Aliases='KIAA0494', description='EF-hand calcium binding domain 14', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46675159, end_position_on_the_genomic_accession=46719064, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101901828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101901828, CurrentID=0, Status='live', Symbol='LOC101901828', Aliases='', description='dynein, light chain, roadblock-type 1 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60495416, end_position_on_the_genomic_accession=60496488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  140886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140886, CurrentID=0, Status='live', Symbol='PABPC5', Aliases='PABP5', description='poly(A) binding protein, cytoplasmic 5', other_designations='', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91434598, end_position_on_the_genomic_accession=91438584, orientation='plus', exon_count=2, OMIM=300407, no_hdr0=''),
  105369857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369857, CurrentID=0, Status='live', Symbol='LOC105369857', Aliases='', description='uncharacterized LOC105369857', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77379760, end_position_on_the_genomic_accession=77402954, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373193, CurrentID=0, Status='live', Symbol='LOC105373193', Aliases='', description='uncharacterized LOC105373193', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47125809, end_position_on_the_genomic_accession=47129454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378939, CurrentID=0, Status='live', Symbol='LOC105378939', Aliases='', description='uncharacterized LOC105378939', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120093631, end_position_on_the_genomic_accession=120100787, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100328982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100328982, CurrentID=0, Status='live', Symbol='LOC100328982', Aliases='', description='eukaryotic translation initiation factor 2 subunit gamma pseudogene', other_designations='eukaryotic translation initiation factor 2, subunit 3 gamma, 52kDa pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40559340, end_position_on_the_genomic_accession=40561060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369860, CurrentID=0, Status='live', Symbol='LOC105369860', Aliases='', description='uncharacterized LOC105369860', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78326680, end_position_on_the_genomic_accession=78359746, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929291, CurrentID=0, Status='live', Symbol='LOC101929291', Aliases='', description='uncharacterized LOC101929291', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128909331, end_position_on_the_genomic_accession=128915390, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  9818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9818, CurrentID=0, Status='live', Symbol='NUP58', Aliases='NUPL1, PRO2463', description='nucleoporin 58kDa', other_designations='nucleoporin like 1|nucleoporin-like protein 1', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25301084, end_position_on_the_genomic_accession=25349795, orientation='plus', exon_count=19, OMIM=607615, no_hdr0=''),
  100141515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100141515, CurrentID=0, Status='live', Symbol='C17orf99', Aliases='UNQ464', description='chromosome 17 open reading frame 99', other_designations='GLPG464', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78146040, end_position_on_the_genomic_accession=78166904, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100873402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873402, CurrentID=0, Status='live', Symbol='RNA5SP133', Aliases='RN5S133', description='RNA, 5S ribosomal pseudogene 133', other_designations='RNA, 5S ribosomal 133', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56185104, end_position_on_the_genomic_accession=56185222, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  140892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140892, CurrentID=0, Status='live', Symbol='TCPT', Aliases='', description='thrombocytopenia, Paris-Trousseau type', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=188025, no_hdr0=''),
  105372875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372875, CurrentID=0, Status='live', Symbol='LOC105372875', Aliases='', description='uncharacterized LOC105372875', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206680457, end_position_on_the_genomic_accession=206682988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140893, CurrentID=0, Status='live', Symbol='RBBP8NL', Aliases='C20orf151', description='RBBP8 N-terminal like', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62410237, end_position_on_the_genomic_accession=62427680, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105369864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369864, CurrentID=0, Status='live', Symbol='LOC105369864', Aliases='', description='uncharacterized LOC105369864', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79715242, end_position_on_the_genomic_accession=79778543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140895, CurrentID=0, Status='live', Symbol='HSPE1P1', Aliases='HSPEP1, bA101E14.1', description='heat shock protein family E (Hsp10) member 1 pseudogene 1', other_designations='heat shock 10kD protein 1 (chaperonin 10) pseudogene 1|heat shock 10kDa protein 1 (chaperonin 10) pseudogene 1|heat shock 10kDa protein 1 pseudogene 1', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39932922, end_position_on_the_genomic_accession=39933239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369866, CurrentID=0, Status='live', Symbol='LOC105369866', Aliases='', description='uncharacterized LOC105369866', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80030445, end_position_on_the_genomic_accession=80083692, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369867, CurrentID=0, Status='live', Symbol='LOC105369867', Aliases='', description='uncharacterized LOC105369867', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80550626, end_position_on_the_genomic_accession=80580408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378941, CurrentID=0, Status='live', Symbol='LOC105378941', Aliases='', description='uncharacterized LOC105378941', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120323845, end_position_on_the_genomic_accession=120328644, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643342, CurrentID=0, Status='live', Symbol='LOC643342', Aliases='', description='ATM interactor pseudogene', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95292338, end_position_on_the_genomic_accession=95294807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369871, CurrentID=0, Status='live', Symbol='LOC105369871', Aliases='', description='uncharacterized LOC105369871', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81417081, end_position_on_the_genomic_accession=81471193, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  9829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9829, CurrentID=0, Status='live', Symbol='DNAJC6', Aliases='DJC6, PARK19', description='DnaJ heat shock protein family (Hsp40) member C6', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 6|DnaJ (Hsp40) homolog, subfamily C, member 6|auxilin', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65248148, end_position_on_the_genomic_accession=65415869, orientation='plus', exon_count=26, OMIM=608375, no_hdr0=''),
  140902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140902, CurrentID=0, Status='live', Symbol='R3HDML', Aliases='dJ881L22.3', description='R3H domain containing-like', other_designations='cysteine-rich secretory protein R3HDML', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44337158, end_position_on_the_genomic_accession=44350792, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  58985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58985, CurrentID=0, Status='live', Symbol='IL22RA1', Aliases='CRF2-9, IL22R, IL22R1', description='interleukin 22 receptor subunit alpha 1', other_designations='IL-22 receptor subunit alpha-1|IL-22R-alpha-1|IL-22RA1|cytokine receptor class-II member 9|cytokine receptor family 2 member 9|interleukin 22 receptor, alpha 1|zcytoR11', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24119771, end_position_on_the_genomic_accession=24143285, orientation='minus', exon_count=8, OMIM=605457, no_hdr0=''),
  58986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=58986, CurrentID=0, Status='live', Symbol='TMEM8A', Aliases='M83, TMEM6, TMEM8', description='transmembrane protein 8A', other_designations='five-span transmembrane protein M83|transmembrane protein 6|transmembrane protein 8 (five membrane-spanning domains)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=370776, end_position_on_the_genomic_accession=381950, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105369876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369876, CurrentID=0, Status='live', Symbol='LOC105369876', Aliases='', description='uncharacterized LOC105369876', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85318060, end_position_on_the_genomic_accession=85355057, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  140906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140906, CurrentID=0, Status='live', Symbol='SPG18', Aliases='', description='spastic paraplegia 18 (autosomal dominant)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611225, no_hdr0=''),
  140909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140909, CurrentID=0, Status='live', Symbol='MTCO2P1', Aliases='MTCO2L, dJ342K12.1', description='MT-CO2 pseudogene 1', other_designations='MTCO2 like|MTCO2 pseudogene 1|cytochrome c oxidase subunit II (MTCO2) pseudogene', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60414396, end_position_on_the_genomic_accession=60414872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100329108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329108, CurrentID=0, Status='live', Symbol='GCSHP5', Aliases='', description='glycine cleavage system protein H pseudogene 5', other_designations='glycine cleavage system protein H (aminomethyl carrier) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168055359, end_position_on_the_genomic_accession=168056493, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100329109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329109, CurrentID=0, Status='live', Symbol='GCSHP3', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206115573, end_position_on_the_genomic_accession=206116572, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369881, CurrentID=0, Status='live', Symbol='LOC105369881', Aliases='', description='uncharacterized LOC105369881', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87612217, end_position_on_the_genomic_accession=87670542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369883, CurrentID=0, Status='live', Symbol='LOC105369883', Aliases='', description='uncharacterized LOC105369883', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88054420, end_position_on_the_genomic_accession=88082376, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140913, CurrentID=0, Status='live', Symbol='PPIAP10', Aliases='CRP, PPIAL2, PPIP10, dJ823G15.1', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 10', other_designations='cyclophillin related pseudogene|peptidylprolyl isomerase (cyclophilin) pseudogene 10|peptidylprolyl isomerase A (cyclophilin A) (PPIA) pseudogene|peptidylprolyl isomerase A (cyclophilin A)-like 2', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53421192, end_position_on_the_genomic_accession=53421678, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100329129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329129, CurrentID=0, Status='live', Symbol='LOC100329129', Aliases='', description='proteolipid protein 2 (colonic epithelium-enriched) pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180533778, end_position_on_the_genomic_accession=180534658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100329135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329135, CurrentID=0, Status='live', Symbol='TRPC5OS', Aliases='TRPC5-AS1', description='TRPC5 opposite strand', other_designations='TRPC5 antisense RNA 1 (non-protein coding)|TRPC5 opposite strand protein|TRPC5-antisense RNA 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111876051, end_position_on_the_genomic_accession=111903990, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369886, CurrentID=0, Status='live', Symbol='LOC105369886', Aliases='', description='uncharacterized LOC105369886', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88749453, end_position_on_the_genomic_accession=88770156, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873403, CurrentID=0, Status='live', Symbol='RNA5SP135', Aliases='RN5S135', description='RNA, 5S ribosomal pseudogene 135', other_designations='RNA, 5S ribosomal 135', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68882967, end_position_on_the_genomic_accession=68883091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  59067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59067, CurrentID=0, Status='live', Symbol='IL21', Aliases='CVID11, IL-21, Za11', description='interleukin 21', other_designations='interleukin-21 isoform', map_location='4q26-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122612628, end_position_on_the_genomic_accession=122621057, orientation='minus', exon_count=5, OMIM=605384, no_hdr0=''),
  59081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59081, CurrentID=0, Status='live', Symbol='MNG2', Aliases='', description='multinodular goiter 2', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300273, no_hdr0=''),
  59082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59082, CurrentID=0, Status='live', Symbol='CARD18', Aliases='ICEBERG, UNQ5804, pseudo-ICE', description='caspase recruitment domain family member 18', other_designations='ICEBERG caspase-1 inhibitor|caspase recruitment domain family, member 18|caspase-1 inhibitor Iceberg', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105137721, end_position_on_the_genomic_accession=105139734, orientation='minus', exon_count=3, OMIM=605354, no_hdr0=''),
  59084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59084, CurrentID=0, Status='live', Symbol='ENPP5', Aliases='NPP-5', description='ectonucleotide pyrophosphatase/phosphodiesterase 5 (putative)', other_designations='E-NPP 5', map_location='6p21.1-p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46159182, end_position_on_the_genomic_accession=46171010, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100329167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329167, CurrentID=0, Status='live', Symbol='AAA1', Aliases='AAA', description='aortic aneurysm, familial abdominal 1', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=100070, no_hdr0=''),
  100329168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329168, CurrentID=0, Status='live', Symbol='LGV2', Aliases='LVG2', description='longevity 2', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606460, no_hdr0=''),
  100329169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329169, CurrentID=0, Status='live', Symbol='CRCS9', Aliases='', description='colorectal cancer, susceptibility to, 9', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612590, no_hdr0=''),
  100329170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100329170, CurrentID=0, Status='live', Symbol='SCZD13', Aliases='', description='schizophrenia 13', other_designations='', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613025, no_hdr0=''),
  140921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140921, CurrentID=0, Status='live', Symbol='PPP2R5CP', Aliases='', description="protein phosphatase 2 regulatory subunit B', gamma pseudogene", other_designations="PP2A gamma1 pseudogene|protein phosphatase 2, regulatory subunit B (B56), gamma isoform pseudogene|protein phosphatase 2, regulatory subunit B', gamma isoform pseudogene|protein phosphatase 2, regulatory subunit B', gamma pseudogene", map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52373652, end_position_on_the_genomic_accession=52375049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369893, CurrentID=0, Status='live', Symbol='LOC105369893', Aliases='', description='uncharacterized LOC105369893', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90293318, end_position_on_the_genomic_accession=90300378, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373399, CurrentID=0, Status='live', Symbol='LOC105373399', Aliases='', description='uncharacterized LOC105373399', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5401308, end_position_on_the_genomic_accession=5409084, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369895, CurrentID=0, Status='live', Symbol='LOC105369895', Aliases='', description='uncharacterized LOC105369895', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90761803, end_position_on_the_genomic_accession=90809151, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369899, CurrentID=0, Status='live', Symbol='LOC105369899', Aliases='', description='uncharacterized LOC105369899', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91485902, end_position_on_the_genomic_accession=91633824, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420852, CurrentID=0, Status='live', Symbol='LOC100420852', Aliases='', description='nitric oxide synthase 2, inducible pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36373459, end_position_on_the_genomic_accession=36425589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643373, CurrentID=0, Status='live', Symbol='LOC643373', Aliases='', description='glucuronidase, beta pseudogene', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34175822, end_position_on_the_genomic_accession=34182868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128430, CurrentID=0, Status='live', Symbol='LOC100128430', Aliases='', description='uncharacterized LOC100128430', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=114545293, end_position_on_the_genomic_accession=114546167, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369904, CurrentID=0, Status='live', Symbol='LOC105369904', Aliases='', description='uncharacterized LOC105369904', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92505323, end_position_on_the_genomic_accession=92512399, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643381, CurrentID=0, Status='live', Symbol='LOC643381', Aliases='', description='zinc finger protein 614 pseudogene', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=98565077, end_position_on_the_genomic_accession=98566829, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369910, CurrentID=0, Status='live', Symbol='LOC105369910', Aliases='', description='uncharacterized LOC105369910', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93388389, end_position_on_the_genomic_accession=93398761, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374068, CurrentID=0, Status='live', Symbol='LOC105374068', Aliases='', description='uncharacterized LOC105374068', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120423142, end_position_on_the_genomic_accession=120431143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873404, CurrentID=0, Status='live', Symbol='RNA5SP136', Aliases='RN5S136', description='RNA, 5S ribosomal pseudogene 136', other_designations='RNA, 5S ribosomal 136', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72691705, end_position_on_the_genomic_accession=72691811, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369912, CurrentID=0, Status='live', Symbol='LOC105369912', Aliases='', description='uncharacterized LOC105369912', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93969696, end_position_on_the_genomic_accession=94012119, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376367, CurrentID=0, Status='live', Symbol='LOC105376367', Aliases='', description='uncharacterized LOC105376367', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3942329, end_position_on_the_genomic_accession=3949354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369915, CurrentID=0, Status='live', Symbol='LOC105369915', Aliases='', description='uncharacterized LOC105369915', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94897055, end_position_on_the_genomic_accession=94928258, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  9875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9875, CurrentID=0, Status='live', Symbol='URB1', Aliases='C21orf108, NPA1', description='URB1 ribosome biogenesis 1 homolog (S. cerevisiae)', other_designations='nucleolar preribosomal-associated protein 1|nucleolar protein 254 kDa', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32311019, end_position_on_the_genomic_accession=32393003, orientation='minus', exon_count=39, OMIM=608865, no_hdr0=''),
  105378951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378951, CurrentID=0, Status='live', Symbol='LOC105378951', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39784636, end_position_on_the_genomic_accession=39808042, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59269, CurrentID=0, Status='live', Symbol='HIVEP3', Aliases='KBP-1, KBP1, KRC, SHN3, Schnurri-3, ZAS3, ZNF40C', description='human immunodeficiency virus type I enhancer binding protein 3', other_designations='ZAS family, member 3|kappa-B and V(D)J recombination signal sequences-binding protein|kappa-binding protein 1|zinc finger protein ZAS3', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41506365, end_position_on_the_genomic_accession=42035931, orientation='minus', exon_count=15, OMIM=606649, no_hdr0=''),
  59271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59271, CurrentID=0, Status='live', Symbol='EVA1C', Aliases='B18, B19, C21orf63, C21orf64, FAM176C, PRED34, SUE21', description='eva-1 homolog C (C. elegans)', other_designations='family with sequence similarity 176, member C|protein FAM176C', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32411692, end_position_on_the_genomic_accession=32515400, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59272, CurrentID=0, Status='live', Symbol='ACE2', Aliases='ACEH', description='angiotensin I converting enzyme 2', other_designations='ACE-related carboxypeptidase|angiotensin I converting enzyme (peptidyl-dipeptidase A) 2|angiotensin-converting enzyme homolog|metalloprotease MPROT15|peptidyl-dipeptidase A', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15494402, end_position_on_the_genomic_accession=15602069, orientation='minus', exon_count=22, OMIM=300335, no_hdr0=''),
  59274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59274, CurrentID=0, Status='live', Symbol='MESDC1', Aliases='', description='mesoderm development candidate 1', other_designations='', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81000954, end_position_on_the_genomic_accession=81004004, orientation='plus', exon_count=1, OMIM=615466, no_hdr0=''),
  59277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59277, CurrentID=0, Status='live', Symbol='NTN4', Aliases='PRO3091', description='netrin 4', other_designations='beta-netrin|hepar-derived netrin-like protein', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95657807, end_position_on_the_genomic_accession=95790787, orientation='minus', exon_count=11, OMIM=610401, no_hdr0=''),
  59283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59283, CurrentID=0, Status='live', Symbol='CACNG8', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 8', other_designations='TARP gamma-8|neuronal voltage-gated calcium channel gamma-8 subunit|transmembrane AMPAR regulatory protein gamma-8', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53963036, end_position_on_the_genomic_accession=53990215, orientation='plus', exon_count=4, OMIM=606900, no_hdr0=''),
  59284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59284, CurrentID=0, Status='live', Symbol='CACNG7', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 7', other_designations='TARP gamma-7|neuronal voltage-gated calcium channel gamma-7 subunit|transmembrane AMPAR regulatory protein gamma-7', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53909425, end_position_on_the_genomic_accession=53943715, orientation='plus', exon_count=6, OMIM=606899, no_hdr0=''),
  59285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59285, CurrentID=0, Status='live', Symbol='CACNG6', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 6', other_designations='neuronal voltage-gated calcium channel gamma-6 subunit', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53991149, end_position_on_the_genomic_accession=54012666, orientation='plus', exon_count=7, OMIM=606898, no_hdr0=''),
  59286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59286, CurrentID=0, Status='live', Symbol='UBL5', Aliases='HUB1', description='ubiquitin like 5', other_designations='beacon|ubiquitin-like 5', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9827880, end_position_on_the_genomic_accession=9830121, orientation='plus', exon_count=5, OMIM=606849, no_hdr0=''),
  59307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59307, CurrentID=0, Status='live', Symbol='SIGIRR', Aliases='TIR8', description='single immunoglobulin and toll-interleukin 1 receptor (TIR) domain', other_designations='single Ig IL-1R-related molecule|single immunoglobulin domain IL1R1 related|single immunoglobulin domain-containing IL1R-related protein|toll/interleukin-1 receptor 8', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=405716, end_position_on_the_genomic_accession=417454, orientation='minus', exon_count=11, OMIM=605478, no_hdr0=''),
  102557615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102557615, CurrentID=0, Status='live', Symbol='CTB-7E3.1', Aliases='', description='uncharacterized LOC102557615', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166905222, end_position_on_the_genomic_accession=166926370, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369929, CurrentID=0, Status='live', Symbol='LOC105369929', Aliases='', description='uncharacterized LOC105369929', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97270099, end_position_on_the_genomic_accession=97275112, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369930, CurrentID=0, Status='live', Symbol='LOC105369930', Aliases='', description='uncharacterized LOC105369930', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97270021, end_position_on_the_genomic_accession=97271530, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59330, CurrentID=0, Status='live', Symbol='GER', Aliases='', description='Gastroesophageal reflux', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=109350, no_hdr0=''),
  59331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59331, CurrentID=0, Status='live', Symbol='RFH1', Aliases='AORF', description='Renal failure, progressive, with hypertension', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=161900, no_hdr0=''),
  59332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59332, CurrentID=0, Status='live', Symbol='VSPA', Aliases='', description='Visuospatial/perceptual abilities', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=313000, no_hdr0=''),
  59333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59333, CurrentID=0, Status='live', Symbol='NMSR', Aliases='HMSNR', description='Neuropathy, hereditary motor and sensory, Russe type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605285, no_hdr0=''),
  59335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59335, CurrentID=0, Status='live', Symbol='PRDM12', Aliases='PFM9', description='PR domain containing 12', other_designations='PR domain-containing protein 12|PR-domain containing protein 12|PR-domain zinc finger protein 12', map_location='9q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130664594, end_position_on_the_genomic_accession=130682997, orientation='plus', exon_count=5, OMIM=616458, no_hdr0=''),
  59336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59336, CurrentID=0, Status='live', Symbol='PRDM13', Aliases='MU-MB-20.220, PFM10', description='PR domain containing 13', other_designations='PR domain-containing protein 13|PR-domain containing protein 13|PR-domain zinc finger protein 13', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99606774, end_position_on_the_genomic_accession=99615578, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59338, CurrentID=0, Status='live', Symbol='PLEKHA1', Aliases='TAPP1', description='pleckstrin homology domain containing A1', other_designations='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 1|tandem PH domain-containing protein 1', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122374578, end_position_on_the_genomic_accession=122432355, orientation='plus', exon_count=19, OMIM=607772, no_hdr0=''),
  59339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59339, CurrentID=0, Status='live', Symbol='PLEKHA2', Aliases='TAPP2', description='pleckstrin homology domain containing A2', other_designations='PH domain-containing family A member 2|TAPP-2|pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 2|pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 2|tandem PH Domain containing protein-2|tandem PH domain-containing protein 2', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38901235, end_position_on_the_genomic_accession=38973912, orientation='plus', exon_count=14, OMIM=607773, no_hdr0=''),
  59340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59340, CurrentID=0, Status='live', Symbol='HRH4', Aliases='AXOR35, BG26, GPCR105, GPRv53, H4, H4R, HH4R', description='histamine receptor H4', other_designations='G-protein coupled receptor 105|SP9144|pfi-013', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24460568, end_position_on_the_genomic_accession=24487427, orientation='plus', exon_count=5, OMIM=606792, no_hdr0=''),
  59341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59341, CurrentID=0, Status='live', Symbol='TRPV4', Aliases='BCYM3, CMT2C, HMSN2C, OTRPC4, SMAL, SPSMA, SSQTL1, TRP12, VRL2, VROAC', description='transient receptor potential cation channel, subfamily V, member 4', other_designations='OSM9-like transient receptor potential channel 4|osm-9-like TRP channel 4|osmosensitive transient receptor potential channel 4|transient receptor potential protein 12|vanilloid receptor-like channel 2|vanilloid receptor-related osmotically activated channel', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109783087, end_position_on_the_genomic_accession=109833407, orientation='minus', exon_count=18, OMIM=605427, no_hdr0=''),
  59342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59342, CurrentID=0, Status='live', Symbol='SCPEP1', Aliases='HSCP1, RISC', description='serine carboxypeptidase 1', other_designations='retinoid inducible serine carboxypeptidase|serine carboxypeptidase 1 precursor protein', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56978107, end_position_on_the_genomic_accession=57006768, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  59343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59343, CurrentID=0, Status='live', Symbol='SENP2', Aliases='AXAM2, SMT3IP2', description='SUMO1/sentrin/SMT3 specific peptidase 2', other_designations='SMT3-specific isopeptidase 2|SUMO1/sentrin/SMT3 specific protease 2|sentrin/SUMO-specific protease SENP2', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185586228, end_position_on_the_genomic_accession=185631101, orientation='plus', exon_count=17, OMIM=608261, no_hdr0=''),
  59344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59344, CurrentID=0, Status='live', Symbol='ALOXE3', Aliases='ARCI3, E-LOX, eLOX-3, eLOX3', description='arachidonate lipoxygenase 3', other_designations='e-LOX-3|epidermal LOX-3|epidermal lipoxygenase|epidermis-type lipoxygenase 3|hydroperoxy icosatetraenoate dehydratase', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8095900, end_position_on_the_genomic_accession=8118916, orientation='minus', exon_count=17, OMIM=607206, no_hdr0=''),
  59345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59345, CurrentID=0, Status='live', Symbol='GNB4', Aliases='CMTD1F', description='guanine nucleotide binding protein (G protein), beta polypeptide 4', other_designations='G protein beta-4 subunit|transducin beta chain 4', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179396088, end_position_on_the_genomic_accession=179451583, orientation='minus', exon_count=13, OMIM=610863, no_hdr0=''),
  59347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59347, CurrentID=0, Status='live', Symbol='TPT1P8', Aliases='FKSG2', description='tumor protein, translationally-controlled 1 pseudogene 8', other_designations='apoptosis inhibitor pseudogene', map_location='8p11.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36888391, end_position_on_the_genomic_accession=36889522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  59348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59348, CurrentID=0, Status='live', Symbol='ZNF350', Aliases='ZBRK1, ZFQR', description='zinc finger protein 350', other_designations='KRAB zinc finger protein ZFQR|zinc finger and BRCA1-interacting protein with a KRAB domain 1|zinc finger protein ZBRK1|zinc-finger protein ZBRK1', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51964340, end_position_on_the_genomic_accession=51986856, orientation='minus', exon_count=5, OMIM=605422, no_hdr0=''),
  59349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59349, CurrentID=0, Status='live', Symbol='KLHL12', Aliases='C3IP1, DKIR', description='kelch like family member 12', other_designations='CUL3-interacting protein 1|DKIR homolog|kelch-like family member 12|kelch-like protein C3IP1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202891096, end_position_on_the_genomic_accession=202928644, orientation='minus', exon_count=14, OMIM=614522, no_hdr0=''),
  59350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59350, CurrentID=0, Status='live', Symbol='RXFP1', Aliases='LGR7, RXFPR1', description='relaxin/insulin-like family peptide receptor 1', other_designations='leucine-rich repeat-containing G protein-coupled receptor 7', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158521714, end_position_on_the_genomic_accession=158653372, orientation='plus', exon_count=22, OMIM=606654, no_hdr0=''),
  59351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59351, CurrentID=0, Status='live', Symbol='PBOV1', Aliases='UC28, UROC28, dJ171N11.2', description='prostate and breast cancer overexpressed 1', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138215986, end_position_on_the_genomic_accession=138218490, orientation='minus', exon_count=1, OMIM=605669, no_hdr0=''),
  59352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59352, CurrentID=0, Status='live', Symbol='LGR6', Aliases='GPCR, VTS20631', description='leucine-rich repeat containing G protein-coupled receptor 6', other_designations='gonadotropin receptor|leucine-rich repeat-containing G protein-coupled receptor 6', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202193990, end_position_on_the_genomic_accession=202319785, orientation='plus', exon_count=25, OMIM=606653, no_hdr0=''),
  59353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=59353, CurrentID=0, Status='live', Symbol='TMEM35', Aliases='', description='transmembrane protein 35', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101078847, end_position_on_the_genomic_accession=101096367, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369935, CurrentID=0, Status='live', Symbol='LOC105369935', Aliases='', description='uncharacterized LOC105369935', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101526354, end_position_on_the_genomic_accession=101534280, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374069, CurrentID=0, Status='live', Symbol='LOC105374069', Aliases='', description='uncharacterized LOC105374069', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120863325, end_position_on_the_genomic_accession=120908079, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100418625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418625, CurrentID=0, Status='live', Symbol='LOC100418625', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369940, CurrentID=0, Status='live', Symbol='LOC105369940', Aliases='', description='uncharacterized LOC105369940', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101743990, end_position_on_the_genomic_accession=101745279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369941, CurrentID=0, Status='live', Symbol='LOC105369941', Aliases='', description='uncharacterized LOC105369941', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101923364, end_position_on_the_genomic_accession=101924719, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374591, CurrentID=0, Status='live', Symbol='LOC105374591', Aliases='', description='uncharacterized LOC105374591', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47924438, end_position_on_the_genomic_accession=48004929, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369942, CurrentID=0, Status='live', Symbol='LOC105369942', Aliases='', description='uncharacterized LOC105369942', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102281270, end_position_on_the_genomic_accession=102463647, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105369946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369946, CurrentID=0, Status='live', Symbol='LOC105369946', Aliases='', description='uncharacterized LOC105369946', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103668738, end_position_on_the_genomic_accession=103671526, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369947, CurrentID=0, Status='live', Symbol='LOC105369947', Aliases='', description='uncharacterized LOC105369947', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104140258, end_position_on_the_genomic_accession=104148544, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422117, CurrentID=0, Status='live', Symbol='LOC100422117', Aliases='', description='olfactory receptor, family 51, subfamily L, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369949, CurrentID=0, Status='live', Symbol='LOC105369949', Aliases='', description='uncharacterized LOC105369949', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104377588, end_position_on_the_genomic_accession=104398688, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369953, CurrentID=0, Status='live', Symbol='LOC105369953', Aliases='', description='uncharacterized LOC105369953', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105200061, end_position_on_the_genomic_accession=105211913, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480660, CurrentID=0, Status='live', Symbol='RNU6-1332P', Aliases='', description='RNA, U6 small nuclear 1332, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369955, CurrentID=0, Status='live', Symbol='LOC105369955', Aliases='', description='uncharacterized LOC105369955', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105424100, end_position_on_the_genomic_accession=105442764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060206, CurrentID=0, Status='live', Symbol='LOC101060206', Aliases='', description='transcription termination factor 1, mitochondrial-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60614150, end_position_on_the_genomic_accession=60618101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271106, CurrentID=0, Status='live', Symbol='RPS26P13', Aliases='RPS26_4_158', description='ribosomal protein S26 pseudogene 13', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208697286, end_position_on_the_genomic_accession=208697720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369960, CurrentID=0, Status='live', Symbol='LOC105369960', Aliases='', description='uncharacterized LOC105369960', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105783978, end_position_on_the_genomic_accession=105787678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128442, CurrentID=0, Status='live', Symbol='LOC100128442', Aliases='', description='serine/arginine-rich splicing factor 6 pseudogene', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45913118, end_position_on_the_genomic_accession=45913874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285676, CurrentID=0, Status='live', Symbol='ZNF454', Aliases='', description='zinc finger protein 454', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178941193, end_position_on_the_genomic_accession=178966218, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369966, CurrentID=0, Status='live', Symbol='LOC105369966', Aliases='', description='uncharacterized LOC105369966', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108473207, end_position_on_the_genomic_accession=108475906, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369969, CurrentID=0, Status='live', Symbol='LOC105369969', Aliases='', description='uncharacterized LOC105369969', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108733409, end_position_on_the_genomic_accession=108754415, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369973, CurrentID=0, Status='live', Symbol='LOC105369973', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63246984, end_position_on_the_genomic_accession=63272273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369974, CurrentID=0, Status='live', Symbol='LOC105369974', Aliases='', description='uncharacterized LOC105369974', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109118687, end_position_on_the_genomic_accession=109131869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369975, CurrentID=0, Status='live', Symbol='LOC105369975', Aliases='', description='uncharacterized LOC105369975', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109840327, end_position_on_the_genomic_accession=109843373, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  403150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403150, CurrentID=0, Status='live', Symbol='FLJ31356', Aliases='', description='uncharacterized protein FLJ31356', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28384409, end_position_on_the_genomic_accession=28394672, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  154743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154743, CurrentID=0, Status='live', Symbol='C7orf60', Aliases='', description='chromosome 7 open reading frame 60', other_designations='UPF0532 protein C7orf60', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112819147, end_position_on_the_genomic_accession=112939877, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369982, CurrentID=0, Status='live', Symbol='LOC105369982', Aliases='', description='uncharacterized LOC105369982', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111026327, end_position_on_the_genomic_accession=111032964, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369983, CurrentID=0, Status='live', Symbol='LOC105369983', Aliases='', description='uncharacterized LOC105369983', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111285227, end_position_on_the_genomic_accession=111320162, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369985, CurrentID=0, Status='live', Symbol='LOC105369985', Aliases='', description='uncharacterized LOC105369985', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111436872, end_position_on_the_genomic_accession=111439030, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643459, CurrentID=0, Status='live', Symbol='DUSP5P2', Aliases='', description='dual specificity phosphatase 5 pseudogene 2', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9601694, end_position_on_the_genomic_accession=9602725, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372876, CurrentID=0, Status='live', Symbol='LOC105372876', Aliases='', description='uncharacterized LOC105372876', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206634196, end_position_on_the_genomic_accession=206635780, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369989, CurrentID=0, Status='live', Symbol='LOC105369989', Aliases='', description='uncharacterized LOC105369989', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113353457, end_position_on_the_genomic_accession=113360663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369996, CurrentID=0, Status='live', Symbol='LOC105369996', Aliases='', description='uncharacterized LOC105369996', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114238222, end_position_on_the_genomic_accession=114242173, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369998, CurrentID=0, Status='live', Symbol='LOC105369998', Aliases='', description='uncharacterized LOC105369998', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114414357, end_position_on_the_genomic_accession=114448646, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369999, CurrentID=0, Status='live', Symbol='LOC105369999', Aliases='', description='uncharacterized LOC105369999', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114733760, end_position_on_the_genomic_accession=114735788, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370001, CurrentID=0, Status='live', Symbol='LOC105370001', Aliases='', description='uncharacterized LOC105370001', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115040651, end_position_on_the_genomic_accession=115077476, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927009, CurrentID=0, Status='live', Symbol='LOC101927009', Aliases='', description='uncharacterized LOC101927009', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9401727, end_position_on_the_genomic_accession=9408093, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370007, CurrentID=0, Status='live', Symbol='LOC105370007', Aliases='', description='uncharacterized LOC105370007', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116484428, end_position_on_the_genomic_accession=116488505, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100499435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499435, CurrentID=0, Status='live', Symbol='TEX16P', Aliases='', description='testis expressed 16, pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84717747, end_position_on_the_genomic_accession=84762308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373199, CurrentID=0, Status='live', Symbol='LOC105373199', Aliases='', description='uncharacterized LOC105373199', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232718052, end_position_on_the_genomic_accession=232762405, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369210, CurrentID=0, Status='live', Symbol='LOC105369210', Aliases='', description='uncharacterized LOC105369210', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16929740, end_position_on_the_genomic_accession=16936518, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378969, CurrentID=0, Status='live', Symbol='LOC105378969', Aliases='', description='uncharacterized LOC105378969', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54691905, end_position_on_the_genomic_accession=54737635, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370014, CurrentID=0, Status='live', Symbol='LOC105370014', Aliases='', description='uncharacterized LOC105370014', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118758802, end_position_on_the_genomic_accession=118761689, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132817, CurrentID=0, Status='live', Symbol='LOC100132817', Aliases='', description='C-x(9)-C motif containing 2 pseudogene', other_designations='COX assembly mitochondrial protein 2 homolog pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22756513, end_position_on_the_genomic_accession=22756729, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643494, CurrentID=0, Status='live', Symbol='LOC643494', Aliases='', description='RNA binding motif protein 4B pseudogene', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98904200, end_position_on_the_genomic_accession=98905736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370024, CurrentID=0, Status='live', Symbol='LOC105370024', Aliases='', description='uncharacterized LOC105370024', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119174065, end_position_on_the_genomic_accession=119176484, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643499, CurrentID=0, Status='live', Symbol='RPL11P5', Aliases='RPL11_3_1291', description='ribosomal protein L11 pseudogene 5', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120593964, end_position_on_the_genomic_accession=120594566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378973, CurrentID=0, Status='live', Symbol='LOC105378973', Aliases='', description='uncharacterized LOC105378973', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55279606, end_position_on_the_genomic_accession=55295302, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370031, CurrentID=0, Status='live', Symbol='LOC105370031', Aliases='', description='uncharacterized LOC105370031', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121100691, end_position_on_the_genomic_accession=121108922, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  154754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154754, CurrentID=0, Status='live', Symbol='PRSS3P2', Aliases='T6, TRY6', description='protease, serine 3 pseudogene 2', other_designations='protease, serine, 1 (trypsin 1)|protease, serine, 3 pseudogene 2|trypsinogen C', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285690, CurrentID=0, Status='live', Symbol='LOC285690', Aliases='', description='uncharacterized LOC285690', other_designations='', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370035, CurrentID=0, Status='live', Symbol='LOC105370035', Aliases='', description='uncharacterized LOC105370035', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121857750, end_position_on_the_genomic_accession=121870211, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374073, CurrentID=0, Status='live', Symbol='LOC105374073', Aliases='', description='uncharacterized LOC105374073', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123335164, end_position_on_the_genomic_accession=123337826, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200609, CurrentID=0, Status='live', Symbol='LOC200609', Aliases='', description='uncharacterized LOC200609', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128457, CurrentID=0, Status='live', Symbol='LOC100128457', Aliases='', description='uncharacterized LOC100128457', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=141028994, end_position_on_the_genomic_accession=141037153, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103344718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344718, CurrentID=0, Status='live', Symbol='HOTS', Aliases='', description='H19 opposite tumor suppressor', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643513, CurrentID=0, Status='live', Symbol='LOC643513', Aliases='', description='karyopherin alpha 2 (RAG cohort 1, importin alpha 1) pseudogene', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80079411, end_position_on_the_genomic_accession=80081368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285692, CurrentID=0, Status='live', Symbol='LOC285692', Aliases='', description='uncharacterized LOC285692', other_designations='CTD-2143L24.1', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9641315, end_position_on_the_genomic_accession=9903824, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  403219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403219, CurrentID=0, Status='live', Symbol='OR7E157P', Aliases='', description='olfactory receptor family 7 subfamily E member 157 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 157 pseudogene', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7591988, end_position_on_the_genomic_accession=7592956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131023, CurrentID=0, Status='live', Symbol='LOC100131023', Aliases='', description='uncharacterized LOC100131023', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=40089625, end_position_on_the_genomic_accession=40090699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643520, CurrentID=0, Status='live', Symbol='LOC643520', Aliases='', description='uncharacterized LOC643520', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=97839739, end_position_on_the_genomic_accession=97897179, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370049, CurrentID=0, Status='live', Symbol='LOC105370049', Aliases='', description='uncharacterized LOC105370049', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124620829, end_position_on_the_genomic_accession=124684204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  403223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403223, CurrentID=0, Status='live', Symbol='OR10AE3P', Aliases='', description='olfactory receptor family 10 subfamily AE member 3 pseudogene', other_designations='olfactory receptor, family 10, subfamily AE, member 3 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55646194, end_position_on_the_genomic_accession=55646662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403224, CurrentID=0, Status='live', Symbol='OR10AK1P', Aliases='', description='olfactory receptor family 10 subfamily AK member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily AK, member 1 pseudogene', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55957215, end_position_on_the_genomic_accession=55957629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132824, CurrentID=0, Status='live', Symbol='LOC100132824', Aliases='', description='tubulin alpha-1C chain-like', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34848983, end_position_on_the_genomic_accession=34850484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403225, CurrentID=0, Status='live', Symbol='OR10V7P', Aliases='', description='olfactory receptor family 10 subfamily V member 7 pseudogene', other_designations='olfactory receptor, family 10, subfamily V, member 7 pseudogene', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=42105878, end_position_on_the_genomic_accession=42106177, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403226, CurrentID=0, Status='live', Symbol='OR11J5P', Aliases='', description='olfactory receptor family 11 subfamily J member 5 pseudogene', other_designations='olfactory receptor, family 11, subfamily J, member 5 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20969250, end_position_on_the_genomic_accession=20970209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370053, CurrentID=0, Status='live', Symbol='LOC105370053', Aliases='', description='uncharacterized LOC105370053', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125151514, end_position_on_the_genomic_accession=125160922, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  403227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403227, CurrentID=0, Status='live', Symbol='OR11Q1P', Aliases='', description='olfactory receptor family 11 subfamily Q member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily Q, member 1 pseudogene', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131204096, end_position_on_the_genomic_accession=131205044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378978, CurrentID=0, Status='live', Symbol='LOC105378978', Aliases='', description='uncharacterized LOC105378978', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56536583, end_position_on_the_genomic_accession=56555886, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370055, CurrentID=0, Status='live', Symbol='LOC105370055', Aliases='', description='uncharacterized LOC105370055', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125704486, end_position_on_the_genomic_accession=125707819, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372328, CurrentID=0, Status='live', Symbol='LOC105372328', Aliases='', description='uncharacterized LOC105372328', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403229, CurrentID=0, Status='live', Symbol='OR13Z2P', Aliases='', description='olfactory receptor family 13 subfamily Z member 2 pseudogene', other_designations='olfactory receptor, family 13, subfamily Z, member 2 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147445479, end_position_on_the_genomic_accession=147446234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403230, CurrentID=0, Status='live', Symbol='OR13Z3P', Aliases='', description='olfactory receptor family 13 subfamily Z member 3 pseudogene', other_designations='olfactory receptor, family 13, subfamily Z, member 3 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147482238, end_position_on_the_genomic_accession=147482628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370057, CurrentID=0, Status='live', Symbol='LOC105370057', Aliases='', description='uncharacterized LOC105370057', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125985712, end_position_on_the_genomic_accession=126043480, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  403231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403231, CurrentID=0, Status='live', Symbol='OR1AC1P', Aliases='', description='olfactory receptor family 1 subfamily AC member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily AC, member 1 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3377467, end_position_on_the_genomic_accession=3377713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403232, CurrentID=0, Status='live', Symbol='OR1M4P', Aliases='', description='olfactory receptor family 1 subfamily M member 4 pseudogene', other_designations='olfactory receptor, family 1, subfamily M, member 4 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9082936, end_position_on_the_genomic_accession=9083398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370059, CurrentID=0, Status='live', Symbol='LOC105370059', Aliases='', description='uncharacterized LOC105370059', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126425958, end_position_on_the_genomic_accession=126434206, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271110, CurrentID=0, Status='live', Symbol='RPS26P19', Aliases='RPS26_6_273', description='ribosomal protein S26 pseudogene 19', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127846731, end_position_on_the_genomic_accession=127847080, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403234, CurrentID=0, Status='live', Symbol='OR2A41P', Aliases='', description='olfactory receptor family 2 subfamily A member 41 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 41 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144077393, end_position_on_the_genomic_accession=144077854, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403235, CurrentID=0, Status='live', Symbol='OR2AO1P', Aliases='', description='olfactory receptor family 2 subfamily AO member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AO, member 1 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144176741, end_position_on_the_genomic_accession=144177517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403236, CurrentID=0, Status='live', Symbol='OR2AS2P', Aliases='', description='olfactory receptor family 2 subfamily AS member 2 pseudogene', other_designations='olfactory receptor, family 2, subfamily AS, member 2 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248498311, end_position_on_the_genomic_accession=248498660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370063, CurrentID=0, Status='live', Symbol='LOC105370063', Aliases='', description='uncharacterized LOC105370063', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126975914, end_position_on_the_genomic_accession=127028040, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105370064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370064, CurrentID=0, Status='live', Symbol='LOC105370064', Aliases='', description='uncharacterized LOC105370064', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403238, CurrentID=0, Status='live', Symbol='OR2BH1P', Aliases='', description='olfactory receptor family 2 subfamily BH member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily BH, member 1 pseudogene', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28986635, end_position_on_the_genomic_accession=28987498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285696, CurrentID=0, Status='live', Symbol='LOC285696', Aliases='', description='uncharacterized LOC285696', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17130028, end_position_on_the_genomic_accession=17217422, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  403239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403239, CurrentID=0, Status='live', Symbol='OR2T27', Aliases='', description='olfactory receptor family 2 subfamily T member 27', other_designations='olfactory receptor OR1-67|olfactory receptor, family 2, subfamily T, member 27', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248649931, end_position_on_the_genomic_accession=248650884, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  403241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403241, CurrentID=0, Status='live', Symbol='OR2T32P', Aliases='', description='olfactory receptor family 2 subfamily T member 32 pseudogene', other_designations='olfactory receptor, family 2, subfamily T, member 32 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248083633, end_position_on_the_genomic_accession=248084596, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376693, CurrentID=0, Status='live', Symbol='LOC105376693', Aliases='', description='uncharacterized LOC105376693', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7773794, end_position_on_the_genomic_accession=7783107, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  285697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285697, CurrentID=0, Status='live', Symbol='LOC285697', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17486328, end_position_on_the_genomic_accession=17486885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643542, CurrentID=0, Status='live', Symbol='LOC643542', Aliases='', description='uncharacterized LOC643542', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67516546, end_position_on_the_genomic_accession=67899619, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  403245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403245, CurrentID=0, Status='live', Symbol='OR2X1P', Aliases='', description='olfactory receptor family 2 subfamily X member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily X, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247939072, end_position_on_the_genomic_accession=247939816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403247, CurrentID=0, Status='live', Symbol='OR4A41P', Aliases='', description='olfactory receptor family 4 subfamily A member 41 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 41 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48589740, end_position_on_the_genomic_accession=48590691, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128464, CurrentID=0, Status='live', Symbol='LOC100128464', Aliases='', description='PET117 homolog (S. cerevisiae) pseudogene', other_designations='cytochrome c oxidase assembly factor-like pseudogene', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88513507, end_position_on_the_genomic_accession=88513749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103344926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344926, CurrentID=0, Status='live', Symbol='LINCR-0002', Aliases='', description='uncharacterized LincR-0002', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191425524, end_position_on_the_genomic_accession=191590356, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  103344927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344927, CurrentID=0, Status='live', Symbol='FRIASS', Aliases='C14DELq22, DEL14q22', description='Frias syndrome', other_designations='', map_location='14q22.1-q22.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609640, no_hdr0=''),
  103344928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344928, CurrentID=0, Status='live', Symbol='KIAA1614-AS1', Aliases='', description='KIAA1614 antisense RNA 1', other_designations='RP11-46A10.4', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103344929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344929, CurrentID=0, Status='live', Symbol='BGLT3', Aliases='BGL3, LINC01083, lncRNA-BGL3', description='beta globin locus transcript 3 (non-protein coding)', other_designations='CTD-2643I7.1|long intergenic non-protein coding RNA 1083', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5244554, end_position_on_the_genomic_accession=5245546, orientation='minus', exon_count=1, OMIM=616308, no_hdr0=''),
  103344930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344930, CurrentID=0, Status='live', Symbol='LINC01391', Aliases='', description='long intergenic non-protein coding RNA 1391', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138935189, end_position_on_the_genomic_accession=138944020, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103344931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344931, CurrentID=0, Status='live', Symbol='LOC103344931', Aliases='', description='uncharacterized LOC103344931', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112823496, end_position_on_the_genomic_accession=112827732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103344932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103344932, CurrentID=0, Status='live', Symbol='PGM5P4-AS1', Aliases='FAM233C', description='PGM5P4 antisense RNA 1', other_designations='family with sequence similarity 233, member C', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113528647, end_position_on_the_genomic_accession=113542690, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  403250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403250, CurrentID=0, Status='live', Symbol='OR4A44P', Aliases='', description='olfactory receptor family 4 subfamily A member 44 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 44 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48627481, end_position_on_the_genomic_accession=48628461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370077, CurrentID=0, Status='live', Symbol='LOC105370077', Aliases='', description='uncharacterized LOC105370077', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130067243, end_position_on_the_genomic_accession=130071281, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370078, CurrentID=0, Status='live', Symbol='LOC105370078', Aliases='', description='uncharacterized LOC105370078', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130071331, end_position_on_the_genomic_accession=130072685, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  403252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403252, CurrentID=0, Status='live', Symbol='OR4A46P', Aliases='', description='olfactory receptor family 4 subfamily A member 46 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 46 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48496319, end_position_on_the_genomic_accession=48497259, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643551, CurrentID=0, Status='live', Symbol='LOC643551', Aliases='', description='uncharacterized LOC643551', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=100125315, end_position_on_the_genomic_accession=100126079, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378983, CurrentID=0, Status='live', Symbol='LOC105378983', Aliases='', description='uncharacterized LOC105378983', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58188296, end_position_on_the_genomic_accession=58218900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  403254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403254, CurrentID=0, Status='live', Symbol='OR4A48P', Aliases='', description='olfactory receptor family 4 subfamily A member 48 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 48 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48491722, end_position_on_the_genomic_accession=48492495, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374279, CurrentID=0, Status='live', Symbol='LOC105374279', Aliases='', description='uncharacterized LOC105374279', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192294174, end_position_on_the_genomic_accession=192299758, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  403255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403255, CurrentID=0, Status='live', Symbol='OR4A49P', Aliases='', description='olfactory receptor family 4 subfamily A member 49 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 49 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49915012, end_position_on_the_genomic_accession=49915643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370082, CurrentID=0, Status='live', Symbol='LOC105370082', Aliases='', description='uncharacterized LOC105370082', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131237947, end_position_on_the_genomic_accession=131268678, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  403256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403256, CurrentID=0, Status='live', Symbol='OR4A50P', Aliases='', description='olfactory receptor family 4 subfamily A member 50 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 50 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55475944, end_position_on_the_genomic_accession=55476429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403259, CurrentID=0, Status='live', Symbol='OR4C48P', Aliases='', description='olfactory receptor family 4 subfamily C member 48 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 48 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50018789, end_position_on_the_genomic_accession=50019713, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370086, CurrentID=0, Status='live', Symbol='LOC105370086', Aliases='', description='uncharacterized LOC105370086', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131659474, end_position_on_the_genomic_accession=131661886, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132831, CurrentID=0, Status='live', Symbol='LOC100132831', Aliases='', description='A20-binding inhibitor of NF-kappaB activation 2 pseudogene', other_designations='TNFAIP3 interacting protein 2 pseudogene', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40831217, end_position_on_the_genomic_accession=40833196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  403260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403260, CurrentID=0, Status='live', Symbol='OR4C49P', Aliases='', description='olfactory receptor family 4 subfamily C member 49 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 49 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50053412, end_position_on_the_genomic_accession=50054308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418631, CurrentID=0, Status='live', Symbol='LOC100418631', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378985, CurrentID=0, Status='live', Symbol='LOC105378985', Aliases='', description='uncharacterized LOC105378985', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58503120, end_position_on_the_genomic_accession=58507784, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  403263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403263, CurrentID=0, Status='live', Symbol='OR4G11P', Aliases='', description='olfactory receptor family 4 subfamily G member 11 pseudogene', other_designations='olfactory receptor, family 4, subfamily G, member 11 pseudogene', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63016, end_position_on_the_genomic_accession=63885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643562, CurrentID=0, Status='live', Symbol='LOC643562', Aliases='', description='Rho GTPase activating protein 21 pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80063496, end_position_on_the_genomic_accession=80070583, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403264, CurrentID=0, Status='live', Symbol='OR4X7P', Aliases='', description='olfactory receptor family 4 subfamily X member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily X, member 7 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55411527, end_position_on_the_genomic_accession=55411718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370091, CurrentID=0, Status='live', Symbol='LOC105370091', Aliases='', description='uncharacterized LOC105370091', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132199274, end_position_on_the_genomic_accession=132205472, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  403265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403265, CurrentID=0, Status='live', Symbol='OR51AB1P', Aliases='', description='olfactory receptor family 51 subfamily AB member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily AB, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5291467, end_position_on_the_genomic_accession=5292351, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  403266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403266, CurrentID=0, Status='live', Symbol='OR51B8P', Aliases='', description='olfactory receptor family 51 subfamily B member 8 pseudogene', other_designations='olfactory receptor, family 51, subfamily B, member 8 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5330588, end_position_on_the_genomic_accession=5331345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  60312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60312, CurrentID=0, Status='live', Symbol='AFAP1', Aliases='AFAP, AFAP-11010, AFAP1', description='actin filament associated protein 1', other_designations='110 kDa actin filament-associated protein|actin filament-associated protein, 110 kDa', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7758713, end_position_on_the_genomic_accession=7939926, orientation='minus', exon_count=20, OMIM=608252, no_hdr0=''),
  60313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60313, CurrentID=0, Status='live', Symbol='GPBP1L1', Aliases='SP192', description='GC-rich promoter binding protein 1-like 1', other_designations='GC-rich promoter-binding protein 1-like 1', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45627304, end_position_on_the_genomic_accession=45688110, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  60314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60314, CurrentID=0, Status='live', Symbol='C12orf10', Aliases='Gamm1, MST024, MSTP024, MYG, MYG1', description='chromosome 12 open reading frame 10', other_designations='melanocyte proliferating gene 1|melanocyte related', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53299686, end_position_on_the_genomic_accession=53307181, orientation='plus', exon_count=7, OMIM=611366, no_hdr0=''),
  403271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403271, CurrentID=0, Status='live', Symbol='OR5AC4P', Aliases='', description='olfactory receptor family 5 subfamily AC member 4 pseudogene', other_designations='olfactory receptor, family 5, subfamily AC, member 4 pseudogene', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98104754, end_position_on_the_genomic_accession=98105669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403272, CurrentID=0, Status='live', Symbol='OR5AO1P', Aliases='', description='olfactory receptor family 5 subfamily AO member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AO, member 1 pseudogene', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57045080, end_position_on_the_genomic_accession=57045410, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  60343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60343, CurrentID=0, Status='live', Symbol='FAM3A', Aliases='2.19, DLD, DXS560S, XAP-7', description='family with sequence similarity 3 member A', other_designations='cytokine-like protein 2-19|family with sequence similarity 3, member A', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154506159, end_position_on_the_genomic_accession=154516247, orientation='minus', exon_count=11, OMIM=300492, no_hdr0=''),
  403274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403274, CurrentID=0, Status='live', Symbol='OR5H15', Aliases='', description='olfactory receptor family 5 subfamily H member 15', other_designations='olfactory receptor, family 5, subfamily H, member 15', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98168700, end_position_on_the_genomic_accession=98169641, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132834, CurrentID=0, Status='live', Symbol='LOC100132834', Aliases='', description='coiled-coil domain containing 112 pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71251564, end_position_on_the_genomic_accession=71253013, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403275, CurrentID=0, Status='live', Symbol='OR9S24P', Aliases='OR5J6P', description='olfactory receptor family 9 subfamily S member 24 pseudogene', other_designations='olfactory receptor, family 5, subfamily J, member 6 pseudogene|olfactory receptor, family 9, subfamily S, member 24 pseudogene', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240079433, end_position_on_the_genomic_accession=240080360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374623, CurrentID=0, Status='live', Symbol='LOC105374623', Aliases='', description='uncharacterized LOC105374623', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3344010, end_position_on_the_genomic_accession=3346863, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  60358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60358, CurrentID=0, Status='live', Symbol='ERVK-5', Aliases='ERVK5, HERV-K(II), HERV-KII', description='endogenous retrovirus group K member 5', other_designations='endogenous retroviral sequence K (G2), 5|endogenous retrovirus group K, member 5', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614012, no_hdr0=''),
  60359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60359, CurrentID=0, Status='live', Symbol='ERVK-4', Aliases='ERVK4, HERV-K(I), HERV-KI', description='endogenous retrovirus group K member 4', other_designations='endogenous retroviral sequence K (G1), 4|endogenous retrovirus group K, member 4', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614011, no_hdr0=''),
  403276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403276, CurrentID=0, Status='live', Symbol='OR5J7P', Aliases='', description='olfactory receptor family 5 subfamily J member 7 pseudogene', other_designations='olfactory receptor, family 5, subfamily J, member 7 pseudogene', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56165501, end_position_on_the_genomic_accession=56165922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  403277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403277, CurrentID=0, Status='live', Symbol='OR5K3', Aliases='', description='olfactory receptor family 5 subfamily K member 3', other_designations='olfactory receptor, family 5, subfamily K, member 3', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98390666, end_position_on_the_genomic_accession=98391631, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370104, CurrentID=0, Status='live', Symbol='LOC105370104', Aliases='', description='uncharacterized LOC105370104', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21067532, end_position_on_the_genomic_accession=21075623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  60370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60370, CurrentID=0, Status='live', Symbol='AVPI1', Aliases='PP5395, VIP32, VIT32', description='arginine vasopressin induced 1', other_designations='AVP-induced protein 1|arginine vasopressin-induced 1|vasopressin-induced transcript', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97677424, end_position_on_the_genomic_accession=97687258, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  403278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403278, CurrentID=0, Status='live', Symbol='OR5K4', Aliases='', description='olfactory receptor family 5 subfamily K member 4', other_designations='olfactory receptor, family 5, subfamily K, member 4', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98353854, end_position_on_the_genomic_accession=98354819, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  132946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132946, CurrentID=0, Status='live', Symbol='ARL9', Aliases='', description='ADP ribosylation factor like GTPase 9', other_designations='ADP-ribosylation factor like GTPase 9|ADP-ribosylation factor-like 9', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56505209, end_position_on_the_genomic_accession=56523983, orientation='plus', exon_count=7, OMIM=612405, no_hdr0=''),
  60385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60385, CurrentID=0, Status='live', Symbol='TSKS', Aliases='PPP1R161, STK22S11, TSSKS, TSKS', description='testis-specific serine kinase substrate', other_designations='STK22 substrate 1|protein phosphatase 1, regulatory subunit 161|testis specific serine/threonine kinase substrate|testis-specific kinase substrate', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49739753, end_position_on_the_genomic_accession=49763328, orientation='minus', exon_count=11, OMIM=608253, no_hdr0=''),
  60386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60386, CurrentID=0, Status='live', Symbol='SLC25A19', Aliases='DNC, MCPHA, MUP1, THMD3, THMD4, TPC', description='solute carrier family 25 (mitochondrial thiamine pyrophosphate carrier), member 19', other_designations='microcephaly, Amish|mitochondrial uncoupling protein 1|solute carrier family 25 (mitochondrial deoxynucleotide carrier), member 19', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75272980, end_position_on_the_genomic_accession=75289957, orientation='minus', exon_count=10, OMIM=606521, no_hdr0=''),
  403282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403282, CurrentID=0, Status='live', Symbol='OR6C65', Aliases='', description='olfactory receptor family 6 subfamily C member 65', other_designations='olfactory receptor, family 6, subfamily C, member 65', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55400529, end_position_on_the_genomic_accession=55401467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370109, CurrentID=0, Status='live', Symbol='LOC105370109', Aliases='', description='uncharacterized LOC105370109', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22903263, end_position_on_the_genomic_accession=22909059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  60401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60401, CurrentID=0, Status='live', Symbol='EDA2R', Aliases='EDA-A2R, EDAA2R, TNFRSF27, XEDAR', description='ectodysplasin A2 receptor', other_designations='EDA-A2 receptor|X-linked ectodysplasin-A2 receptor|tumor necrosis factor receptor superfamily member XEDAR', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66594384, end_position_on_the_genomic_accession=66639298, orientation='minus', exon_count=9, OMIM=300276, no_hdr0=''),
  403283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403283, CurrentID=0, Status='live', Symbol='OR6C66P', Aliases='', description='olfactory receptor family 6 subfamily C member 66 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 66 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55388445, end_position_on_the_genomic_accession=55389377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643582, CurrentID=0, Status='live', Symbol='AKR1B10P2', Aliases='', description='aldo-keto reductase family 1, member B10 pseudogene 2', other_designations='aldo-keto reductase family 1, member B10 (aldose reductase) pseudogene', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68455950, end_position_on_the_genomic_accession=68456870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374076, CurrentID=0, Status='live', Symbol='LOC105374076', Aliases='', description='uncharacterized LOC105374076', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124545196, end_position_on_the_genomic_accession=124584583, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  403284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403284, CurrentID=0, Status='live', Symbol='OR6C68', Aliases='', description='olfactory receptor family 6 subfamily C member 68', other_designations='olfactory receptor, family 6, subfamily C, member 68', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55492378, end_position_on_the_genomic_accession=55493316, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  60412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60412, CurrentID=0, Status='live', Symbol='EXOC4', Aliases='SEC8, SEC8L1, Sec8p', description='exocyst complex component 4', other_designations='SEC8-like 1|exocyst complex component Sec8', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133253067, end_position_on_the_genomic_accession=134065761, orientation='plus', exon_count=20, OMIM=608185, no_hdr0=''),
  403285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403285, CurrentID=0, Status='live', Symbol='OR6C69P', Aliases='', description='olfactory receptor family 6 subfamily C member 69 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 69 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55262539, end_position_on_the_genomic_accession=55263434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418632, CurrentID=0, Status='live', Symbol='LOC100418632', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370114, CurrentID=0, Status='live', Symbol='LOC105370114', Aliases='', description='uncharacterized LOC105370114', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23669192, end_position_on_the_genomic_accession=23714932, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  403288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403288, CurrentID=0, Status='live', Symbol='OR6C72P', Aliases='', description='olfactory receptor family 6 subfamily C member 72 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 72 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55283750, end_position_on_the_genomic_accession=55284683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285706, CurrentID=0, Status='live', Symbol='LOC285706', Aliases='', description='cytokine receptor-like factor 3 pseudogene', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=99947726, end_position_on_the_genomic_accession=99949500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  60436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60436, CurrentID=0, Status='live', Symbol='TGIF2', Aliases='', description='TGFB-induced factor homeobox 2', other_designations="5'-TG-3' interacting factor 2|TGF(beta)-induced transcription factor 2|TGF-beta-induced transcription factor 2|TGFB-induced factor 2 (TALE family homeobox)|transcription growth factor-beta-induced factor 2", map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36573473, end_position_on_the_genomic_accession=36593952, orientation='plus', exon_count=5, OMIM=607294, no_hdr0=''),
  60437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60437, CurrentID=0, Status='live', Symbol='CDH26', Aliases='VR20', description='cadherin 26', other_designations='cadherin-like 26|cadherin-like protein VR20', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59955831, end_position_on_the_genomic_accession=60034338, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  60439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60439, CurrentID=0, Status='live', Symbol='TTTY2', Aliases='NCRNA00109, TTY2, lINC00109', description='testis-specific transcript, Y-linked 2 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 109|testis-specific testis transcript Y 2', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9736286, end_position_on_the_genomic_accession=9758476, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  60467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60467, CurrentID=0, Status='live', Symbol='BPESC1', Aliases='NCRNA00187', description='blepharophimosis, epicanthus inversus and ptosis, candidate 1 (non-protein coding)', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139104185, end_position_on_the_genomic_accession=139125167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  60468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60468, CurrentID=0, Status='live', Symbol='BACH2', Aliases='BTBD25', description='BTB and CNC homology 1, basic leucine zipper transcription factor 2', other_designations='BTB and CNC homolog 2', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89926528, end_position_on_the_genomic_accession=90296908, orientation='minus', exon_count=13, OMIM=605394, no_hdr0=''),
  403294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403294, CurrentID=0, Status='live', Symbol='OR6R2P', Aliases='', description='olfactory receptor family 6 subfamily R member 2 pseudogene', other_designations='olfactory receptor, family 6, subfamily R, member 2 pseudogene', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21797086, end_position_on_the_genomic_accession=21797896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421251, CurrentID=0, Status='live', Symbol='LOC100421251', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403295, CurrentID=0, Status='live', Symbol='OR7D11P', Aliases='', description='olfactory receptor family 7 subfamily D member 11 pseudogene', other_designations='olfactory receptor, family 7, subfamily D, member 11 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9216188, end_position_on_the_genomic_accession=9216389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374627, CurrentID=0, Status='live', Symbol='LOC105374627', Aliases='', description='uncharacterized LOC105374627', other_designations='CTD-2008N3.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=4135682, end_position_on_the_genomic_accession=4154136, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  60481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60481, CurrentID=0, Status='live', Symbol='ELOVL5', Aliases='HELO1, SCA38, dJ483K16.1', description='ELOVL fatty acid elongase 5', other_designations='3-keto acyl-CoA synthase ELOVL5|ELOVL FA elongase 5|ELOVL family member 5, elongation of long chain fatty acids (FEN1/Elo2, SUR4/Elo3-like, yeast)|fatty acid elongase 1|homolog of yeast long chain polyunsaturated fatty acid elongation enzyme 2|spinocerebellar ataxia 38|very-long-chain 3-oxoacyl-CoA synthase 5', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53267398, end_position_on_the_genomic_accession=53349179, orientation='minus', exon_count=10, OMIM=611805, no_hdr0=''),
  60482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60482, CurrentID=0, Status='live', Symbol='SLC5A7', Aliases='CHT, CHT1, HMN7A', description='solute carrier family 5 (sodium/choline cotransporter), member 7', other_designations='high affinity choline transporter; hemicholinium-3-sensitive choline transporter', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107986487, end_position_on_the_genomic_accession=108014207, orientation='plus', exon_count=10, OMIM=608761, no_hdr0=''),
  105370123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370123, CurrentID=0, Status='live', Symbol='LOC105370123', Aliases='', description='uncharacterized LOC105370123', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26876606, end_position_on_the_genomic_accession=26880941, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  60484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60484, CurrentID=0, Status='live', Symbol='HAPLN2', Aliases='BRAL1', description='hyaluronan and proteoglycan link protein 2', other_designations='brain link protein 1|brain link protein-1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156618801, end_position_on_the_genomic_accession=156625725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  60485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60485, CurrentID=0, Status='live', Symbol='SAV1', Aliases='SAV, WW45, WWP4', description='salvador family WW domain containing protein 1', other_designations='1700040G09Rik|45 kDa WW domain protein|WW domain-containing adaptor 45|hWW45|salvador homolog 1', map_location='14q13-q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50633580, end_position_on_the_genomic_accession=50668357, orientation='minus', exon_count=7, OMIM=607203, no_hdr0=''),
  100422124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422124, CurrentID=0, Status='live', Symbol='LOC100422124', Aliases='', description='olfactory receptor, family 51, subfamily B, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60487, CurrentID=0, Status='live', Symbol='TRMT11', Aliases='C6orf75, MDS024, TRM11-1, TRMT11', description='tRNA methyltransferase 11 homolog', other_designations="tRNA guanosine-2'-O-methyltransferase TRM11 homolog", map_location='6q11.1-q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125986430, end_position_on_the_genomic_accession=126039276, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  60488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60488, CurrentID=0, Status='live', Symbol='MRPS35', Aliases='HDCMD11P, MDS023, MRP-S28, MRPS28', description='mitochondrial ribosomal protein S35', other_designations='28S ribosomal protein S28, mitochondrial|MRP-S35|S28mt|S35mt|mitochondrial ribosomal protein S28', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27710773, end_position_on_the_genomic_accession=27756304, orientation='plus', exon_count=8, OMIM=611995, no_hdr0=''),
  60489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60489, CurrentID=0, Status='live', Symbol='APOBEC3G', Aliases='A3G, ARCD, ARP-9, ARP9, CEM-15, CEM15, MDS019, bK150C2.7, dJ494G10.1', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3G', other_designations='APOBEC-related cytidine deaminase|APOBEC-related protein 9|DNA dC->dU editing enzyme|apolipoprotein B editing enzyme catalytic polypeptide-like 3G|apolipoprotein B mRNA editing enzyme cytidine deaminase|apolipoprotein B mRNA-editing enzyme catalytic polypeptide 3G|deoxycytidine deaminase|phorbolin-like protein MDS019', map_location='22q13.1-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39077005, end_position_on_the_genomic_accession=39087743, orientation='plus', exon_count=8, OMIM=607113, no_hdr0=''),
  60490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60490, CurrentID=0, Status='live', Symbol='PPCDC', Aliases='MDS018, PPC-DC, coaC', description='phosphopantothenoylcysteine decarboxylase', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75023555, end_position_on_the_genomic_accession=75050726, orientation='plus', exon_count=7, OMIM=609854, no_hdr0=''),
  60491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60491, CurrentID=0, Status='live', Symbol='NIF3L1', Aliases='ALS2CR1, CALS-7, MDS015', description='NGG1 interacting factor 3 like 1', other_designations='GTP cyclohydrolase I|NIF3 (Ngg1 interacting factor 3, S.pombe homolog)-like 1|NIF3 NGG1 interacting factor 3-like 1|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 1|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 1 protein|putative GTP cyclohydrolase 1 type 2 NIF3L1', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200889327, end_position_on_the_genomic_accession=200903932, orientation='plus', exon_count=8, OMIM=605778, no_hdr0=''),
  60492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60492, CurrentID=0, Status='live', Symbol='CCDC90B', Aliases='MDS011, MDS025', description='coiled-coil domain containing 90B', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83259093, end_position_on_the_genomic_accession=83286407, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  60493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60493, CurrentID=0, Status='live', Symbol='FASTKD5', Aliases='dJ1187M17.5', description='FAST kinase domains 5', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3146519, end_position_on_the_genomic_accession=3159886, orientation='minus', exon_count=2, OMIM=614272, no_hdr0=''),
  60494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60494, CurrentID=0, Status='live', Symbol='CCDC81', Aliases='', description='coiled-coil domain containing 81', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86374736, end_position_on_the_genomic_accession=86423109, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  60495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60495, CurrentID=0, Status='live', Symbol='HPSE2', Aliases='HPA2, HPR2, UFS, UFS1', description='heparanase 2 (inactive)', other_designations='heparanase 3|heparanase-like protein', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98457077, end_position_on_the_genomic_accession=99235875, orientation='minus', exon_count=16, OMIM=613469, no_hdr0=''),
  60496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60496, CurrentID=0, Status='live', Symbol='AASDHPPT', Aliases='AASD-PPT, CGI-80, LYS2, LYS5', description='aminoadipate-semialdehyde dehydrogenase-phosphopantetheinyl transferase', other_designations="4'-phosphopantetheinyl transferase|LYS5 ortholog|alpha-aminoadipic semialdehyde dehydrogenase-phosphopantetheinyl transferase", map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106077565, end_position_on_the_genomic_accession=106098692, orientation='plus', exon_count=6, OMIM=607756, no_hdr0=''),
  60498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60498, CurrentID=0, Status='live', Symbol='IGAN1', Aliases='IGAN', description='IgA nephropathy', other_designations='', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=161950, no_hdr0=''),
  403299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403299, CurrentID=0, Status='live', Symbol='OR7E159P', Aliases='', description='olfactory receptor family 7 subfamily E member 159 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 159 pseudogene', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51771788, end_position_on_the_genomic_accession=51772096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  60500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60500, CurrentID=0, Status='live', Symbol='BRCA3', Aliases='BRCAX', description='breast cancer 3', other_designations='Breast cancer, type 3', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605365, no_hdr0=''),
  60502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60502, CurrentID=0, Status='live', Symbol='CPAT1', Aliases='ACP', description='cerebral palsy, ataxic 1', other_designations='', map_location='9p12-q12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605388, no_hdr0=''),
  403300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403300, CurrentID=0, Status='live', Symbol='OR7G15P', Aliases='', description='olfactory receptor family 7 subfamily G member 15 pseudogene', other_designations='olfactory receptor, family 7, subfamily G, member 15 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9122083, end_position_on_the_genomic_accession=9122486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  60506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60506, CurrentID=0, Status='live', Symbol='NYX', Aliases='CLRP, CSNB1, CSNB1A, CSNB4, NBM1', description='nyctalopin', other_designations='leucine-rich repeat protein', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41445637, end_position_on_the_genomic_accession=41476421, orientation='plus', exon_count=4, OMIM=300278, no_hdr0=''),
  60509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60509, CurrentID=0, Status='live', Symbol='AGBL5', Aliases='CCP5', description='ATP/GTP binding protein-like 5', other_designations='ATP/GTP-binding protein-like 5|cytosolic carboxypeptidase 5', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27051423, end_position_on_the_genomic_accession=27070622, orientation='plus', exon_count=18, OMIM=615900, no_hdr0=''),
  403301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403301, CurrentID=0, Status='live', Symbol='OR7H2P', Aliases='', description='olfactory receptor family 7 subfamily H member 2 pseudogene', other_designations='olfactory receptor, family 7, subfamily H, member 2 pseudogene', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=101815883, end_position_on_the_genomic_accession=101816780, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  191585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=191585, CurrentID=0, Status='live', Symbol='PLAC4', Aliases='C21orf115, PRED78', description='placenta specific 4', other_designations='placenta-specific 4', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41175231, end_position_on_the_genomic_accession=41185239, orientation='minus', exon_count=1, OMIM=613770, no_hdr0=''),
  403302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403302, CurrentID=0, Status='live', Symbol='OR8A3P', Aliases='', description='olfactory receptor family 8 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 8, subfamily A, member 3 pseudogene', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124531607, end_position_on_the_genomic_accession=124531985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378993, CurrentID=0, Status='live', Symbol='LOC105378993', Aliases='', description='uncharacterized LOC105378993', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61305916, end_position_on_the_genomic_accession=61309317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  403303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403303, CurrentID=0, Status='live', Symbol='OR8I4P', Aliases='', description='olfactory receptor family 8 subfamily I member 4 pseudogene', other_designations='olfactory receptor, family 8, subfamily I, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56097243, end_position_on_the_genomic_accession=56097383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  60526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60526, CurrentID=0, Status='live', Symbol='LDAH', Aliases='C2orf43, hLDAH', description='lipid droplet associated hydrolase', other_designations='UPF0554 protein C2orf43|lipid droplet-associated serine hydrolase', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20684014, end_position_on_the_genomic_accession=20823137, orientation='minus', exon_count=7, OMIM=613570, no_hdr0=''),
  60528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60528, CurrentID=0, Status='live', Symbol='ELAC2', Aliases='COXPD17, ELC2, HPC2', description='elaC ribonuclease Z 2', other_designations='RNase Z 2|elaC homolog 2|elaC homolog protein 2|elaC-like protein 2|heredity prostate cancer protein 2|putative prostate cancer susceptibility protein HPC2/ELAC2|ribonuclease Z 2|tRNA 3 endonuclease 2|tRNase Z (long form)|tRNase Z 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12991612, end_position_on_the_genomic_accession=13018064, orientation='minus', exon_count=24, OMIM=605367, no_hdr0=''),
  60529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60529, CurrentID=0, Status='live', Symbol='ALX4', Aliases='CRS5, FND2', description='ALX homeobox 4', other_designations='aristaless-like homeobox 4|homeodomain transcription factor ALX4', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44260438, end_position_on_the_genomic_accession=44319220, orientation='minus', exon_count=7, OMIM=605420, no_hdr0=''),
  403305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403305, CurrentID=0, Status='live', Symbol='OR8X1P', Aliases='', description='olfactory receptor family 8 subfamily X member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily X, member 1 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124404233, end_position_on_the_genomic_accession=124405101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372528, CurrentID=0, Status='live', Symbol='LOC105372528', Aliases='', description='uncharacterized LOC105372528', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11171603, end_position_on_the_genomic_accession=11178453, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643605, CurrentID=0, Status='live', Symbol='LOC643605', Aliases='', description='CTD small phosphatase like 2 pseudogene', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase like 2 pseudogene', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100149785, end_position_on_the_genomic_accession=100151377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378994, CurrentID=0, Status='live', Symbol='LOC105378994', Aliases='', description='uncharacterized LOC105378994', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61321490, end_position_on_the_genomic_accession=61332216, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370265, CurrentID=0, Status='live', Symbol='LOC105370265', Aliases='', description='uncharacterized LOC105370265', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76712001, end_position_on_the_genomic_accession=76806086, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105370135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370135, CurrentID=0, Status='live', Symbol='LOC105370135', Aliases='', description='uncharacterized LOC105370135', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28574988, end_position_on_the_genomic_accession=28584344, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  60558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60558, CurrentID=0, Status='live', Symbol='GUF1', Aliases='EF-4', description='GUF1 homolog, GTPase', other_designations='EF-4|GTP-binding protein GUF1 homolog|GTPase GUF1|GTPase of unknown function 1|GUF1 GTPase homolog|elongation factor 4 homolog|ribosomal back-translocase', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44678395, end_position_on_the_genomic_accession=44700928, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  60559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60559, CurrentID=0, Status='live', Symbol='SPCS3', Aliases='PRO3567, SPC22/23, SPC3, YLR066W', description='signal peptidase complex subunit 3', other_designations='SPase 22 kDa subunit|SPase 22/23 kDa subunit|microsomal signal peptidase 22/23 kDa subunit|microsomal signal peptidase 23 kDa subunit|signal peptidase complex subunit 3 homolog', map_location='4q34.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176319939, end_position_on_the_genomic_accession=176332245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  60560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60560, CurrentID=0, Status='live', Symbol='NAA35', Aliases='EGAP, MAK10, MAK10P, bA379P1.1', description='N(alpha)-acetyltransferase 35, NatC auxiliary subunit', other_designations='MAK10 homolog, amino-acid N-acetyltransferase subunit|corneal wound healing-related protein|embryonic growth-associated protein homolog|protein MAK10 homolog', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85941142, end_position_on_the_genomic_accession=86025462, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  60561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60561, CurrentID=0, Status='live', Symbol='RINT1', Aliases='RINT-1', description='RAD50 interactor 1', other_designations='hsRINT-1', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105532085, end_position_on_the_genomic_accession=105567678, orientation='plus', exon_count=16, OMIM=610089, no_hdr0=''),
  105370137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370137, CurrentID=0, Status='live', Symbol='LOC105370137', Aliases='', description='uncharacterized LOC105370137', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28659195, end_position_on_the_genomic_accession=28661269, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418633, CurrentID=0, Status='live', Symbol='LOC100418633', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403312, CurrentID=0, Status='live', Symbol='LOC403312', Aliases='', description='putative uncharacterized protein MGC39545', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125495214, end_position_on_the_genomic_accession=125499526, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  403313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403313, CurrentID=0, Status='live', Symbol='PLPP6', Aliases='PDP1, PPAPDC2, PSDP, bA6J24.6', description='phospholipid phosphatase 6', other_designations='PPAP2 domain-containing protein 2|phosphatidic acid phosphatase type 2 domain containing 2|phosphatidic acid phosphatase type 2 domain-containing protein 2|polyisoprenoid diphosphate phosphatase type 1', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4662294, end_position_on_the_genomic_accession=4665274, orientation='plus', exon_count=1, OMIM=611666, no_hdr0=''),
  403314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403314, CurrentID=0, Status='live', Symbol='APOBEC4', Aliases='C1orf169', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 4 (putative)', other_designations='apolipoprotein B mRNA-editing enzyme catalytic polypeptide-like 4', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183646275, end_position_on_the_genomic_accession=183653313, orientation='minus', exon_count=2, OMIM=609908, no_hdr0=''),
  60592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60592, CurrentID=0, Status='live', Symbol='SCOC', Aliases='HRIHFB2072O, SCOC', description='short coiled-coil protein', other_designations='short coiled coil protein', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140257286, end_position_on_the_genomic_accession=140382556, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  403315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403315, CurrentID=0, Status='live', Symbol='FAM92A1P2', Aliases='FAM92A3', description='family with sequence similarity 92 member A1 pseudogene 2', other_designations='family with sequence similarity 92, member A1 pseudogene 2', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183037665, end_position_on_the_genomic_accession=183040119, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374631, CurrentID=0, Status='live', Symbol='LOC105374631', Aliases='', description='uncharacterized LOC105374631', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5069192, end_position_on_the_genomic_accession=5078410, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  60598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60598, CurrentID=0, Status='live', Symbol='KCNK15', Aliases='K2p15.1, KCNK11, KCNK14, KT3.3, TASK-5, TASK5, dJ781B1.1', description='potassium channel, two pore domain subfamily K, member 15', other_designations='TWIK-related acid-sensitive K(+) channel 5|acid-sensitive potassium channel protein TASK-5|potassium channel, subfamily K, member 14|two pore K(+) channel KT3.3|two pore potassium channel KT3.3', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44745847, end_position_on_the_genomic_accession=44752313, orientation='plus', exon_count=2, OMIM=607368, no_hdr0=''),
  105370143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370143, CurrentID=0, Status='live', Symbol='LOC105370143', Aliases='', description='uncharacterized LOC105370143', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30422604, end_position_on_the_genomic_accession=30427567, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370144, CurrentID=0, Status='live', Symbol='LOC105370144', Aliases='', description='uncharacterized LOC105370144', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30613657, end_position_on_the_genomic_accession=30617704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84944, CurrentID=0, Status='live', Symbol='MAEL', Aliases='CT128, SPATA35', description='maelstrom spermatogenic transposon silencer', other_designations='cancer/testis antigen 128|maelstrom homolog|spermatogenesis associated 35', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166975582, end_position_on_the_genomic_accession=167022214, orientation='plus', exon_count=14, OMIM=611368, no_hdr0=''),
  105370145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370145, CurrentID=0, Status='live', Symbol='LOC105370145', Aliases='', description='uncharacterized LOC105370145', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30973073, end_position_on_the_genomic_accession=30977640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  60625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60625, CurrentID=0, Status='live', Symbol='DHX35', Aliases='C20orf15, DDX35, KAIA0875', description='DEAH (Asp-Glu-Ala-His) box polypeptide 35', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 35|DEAH box protein 35|DEAH-box protein 35', map_location='20q11.22-q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38962338, end_position_on_the_genomic_accession=39039723, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  60626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60626, CurrentID=0, Status='live', Symbol='RIC8A', Aliases='RIC8', description='RIC8 guanine nucleotide exchange factor A', other_designations='resistance to inhibitors of cholinesterase 8 homolog A', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=207511, end_position_on_the_genomic_accession=215175, orientation='plus', exon_count=10, OMIM=609146, no_hdr0=''),
  105180377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105180377, CurrentID=0, Status='live', Symbol='CMTS', Aliases='', description='Chronic mountain sickness, susceptibility to', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616182, no_hdr0=''),
  100422125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422125, CurrentID=0, Status='live', Symbol='LOC100422125', Aliases='', description='olfactory receptor, family 51, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403323, CurrentID=0, Status='live', Symbol='LOC403323', Aliases='', description='uncharacterized LOC403323', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62857856, end_position_on_the_genomic_accession=62898087, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105180390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105180390, CurrentID=0, Status='live', Symbol='SPDYE13P', Aliases='', description='speedy/RINGO cell cycle regulator family member E13, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75282192, end_position_on_the_genomic_accession=75289489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105180391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105180391, CurrentID=0, Status='live', Symbol='SPDYE15P', Aliases='', description='speedy/RINGO cell cycle regulator family member E15, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75338087, end_position_on_the_genomic_accession=75345384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370151, CurrentID=0, Status='live', Symbol='LOC105370151', Aliases='', description='uncharacterized LOC105370151', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32131824, end_position_on_the_genomic_accession=32136896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374633, CurrentID=0, Status='live', Symbol='LOC105374633', Aliases='', description='uncharacterized LOC105374633', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5584188, end_position_on_the_genomic_accession=5595078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370154, CurrentID=0, Status='live', Symbol='LOC105370154', Aliases='', description='uncharacterized LOC105370154', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33333890, end_position_on_the_genomic_accession=33335266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378998, CurrentID=0, Status='live', Symbol='LOC105378998', Aliases='', description='uncharacterized LOC105378998', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61902626, end_position_on_the_genomic_accession=61916275, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  60672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60672, CurrentID=0, Status='live', Symbol='MIIP', Aliases='IIP45', description='migration and invasion inhibitory protein', other_designations='IGFBP2-binding protein|invasion-inhibitory protein 45', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12019242, end_position_on_the_genomic_accession=12032049, orientation='plus', exon_count=13, OMIM=608772, no_hdr0=''),
  60673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60673, CurrentID=0, Status='live', Symbol='ATG101', Aliases='C12orf44', description='autophagy related 101', other_designations='Atg13-interacting protein', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52069974, end_position_on_the_genomic_accession=52077495, orientation='plus', exon_count=4, OMIM=615089, no_hdr0=''),
  60674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60674, CurrentID=0, Status='live', Symbol='GAS5', Aliases='NCRNA00030, SNHG2', description='growth arrest specific 5 (non-protein coding)', other_designations='growth arrest specific 5|growth arrest specific transcript 5|growth arrest-specific 5 (non-protein coding)|small nucleolar RNA host gene (non-protein coding) 2', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173863901, end_position_on_the_genomic_accession=173867987, orientation='minus', exon_count=12, OMIM=608280, no_hdr0=''),
  60675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60675, CurrentID=0, Status='live', Symbol='PROK2', Aliases='BV8, HH4, KAL4, MIT1, PK2', description='prokineticin 2', other_designations='protein Bv8 homolog', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71771655, end_position_on_the_genomic_accession=71785206, orientation='minus', exon_count=4, OMIM=607002, no_hdr0=''),
  60676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60676, CurrentID=0, Status='live', Symbol='PAPPA2', Aliases='PAPP-A2, PAPP-E, PAPPE, PLAC3', description='pappalysin 2', other_designations='placenta-specific 3|pregnancy-associated plasma preproprotein-A2|pregnancy-associated plasma protein E1', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176463171, end_position_on_the_genomic_accession=176845601, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  60677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60677, CurrentID=0, Status='live', Symbol='CELF6', Aliases='BRUNOL6', description='CUGBP, Elav-like family member 6', other_designations='Bruno -like 6, RNA binding protein|CELF-6|CUG-BP and ETR-3 like factor 6|CUG-BP- and ETR-3-like factor 6|RNA-binding protein BRUNOL-6|bruno-like 6, RNA binding protein|bruno-like protein 6', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72284727, end_position_on_the_genomic_accession=72320184, orientation='minus', exon_count=14, OMIM=612681, no_hdr0=''),
  60678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60678, CurrentID=0, Status='live', Symbol='EEFSEC', Aliases='EFSEC, SELB', description='eukaryotic elongation factor, selenocysteine-tRNA-specific', other_designations='elongation factor for selenoprotein translation|elongation factor sec', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128153457, end_position_on_the_genomic_accession=128426193, orientation='plus', exon_count=10, OMIM=607695, no_hdr0=''),
  60680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60680, CurrentID=0, Status='live', Symbol='CELF5', Aliases='BRUNOL-5, BRUNOL5, CELF-5', description='CUGBP, Elav-like family member 5', other_designations='CUG-BP and ETR-3 like factor 5|RNA-binding protein BRUNOL-5|bruno-like 5 RNA binding protein', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3224676, end_position_on_the_genomic_accession=3297076, orientation='plus', exon_count=16, OMIM=612680, no_hdr0=''),
  60681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60681, CurrentID=0, Status='live', Symbol='FKBP10', Aliases='BRKS1, FKBP65, OI11, OI6, PPIASE, hFKBP65', description='FK506 binding protein 10', other_designations='65 kDa FK506-binding protein|65 kDa FKBP|FK506 binding protein 10, 65 kDa|FK506-binding protein 10|FKBP-10|FKBP-65|PPIase FKBP10|immunophilin FKBP65|rotamase', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41812262, end_position_on_the_genomic_accession=41823217, orientation='plus', exon_count=11, OMIM=607063, no_hdr0=''),
  60682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60682, CurrentID=0, Status='live', Symbol='SMAP1', Aliases='SMAP-1', description='small ArfGAP 1', other_designations='stromal membrane-associated GTPase-activating protein 1', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70667771, end_position_on_the_genomic_accession=70862015, orientation='plus', exon_count=16, OMIM=611372, no_hdr0=''),
  60684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60684, CurrentID=0, Status='live', Symbol='TRAPPC11', Aliases='C4orf41, FOIGR, GRY, LGMD2S', description='trafficking protein particle complex 11', other_designations='foie gras homolog|gryzun homolog', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183659267, end_position_on_the_genomic_accession=183713594, orientation='plus', exon_count=31, OMIM=614138, no_hdr0=''),
  60685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60685, CurrentID=0, Status='live', Symbol='ZFAND3', Aliases='TEX27', description='zinc finger, AN1-type domain 3', other_designations='testis expressed sequence 27', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37819496, end_position_on_the_genomic_accession=38154624, orientation='plus', exon_count=7, OMIM=607455, no_hdr0=''),
  60686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=60686, CurrentID=0, Status='live', Symbol='C14orf93', Aliases='', description='chromosome 14 open reading frame 93', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22986901, end_position_on_the_genomic_accession=23010201, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105373205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373205, CurrentID=0, Status='live', Symbol='LOC105373205', Aliases='', description='uncharacterized LOC105373205', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51629461, end_position_on_the_genomic_accession=51644084, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643630, CurrentID=0, Status='live', Symbol='LOC643630', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62266324, end_position_on_the_genomic_accession=62266897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643634, CurrentID=0, Status='live', Symbol='LOC643634', Aliases='', description='tropomyosin 1 (alpha) pseudogene', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27633103, end_position_on_the_genomic_accession=27633711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128482, CurrentID=0, Status='live', Symbol='LOC100128482', Aliases='', description='uncharacterized LOC100128482', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370165, CurrentID=0, Status='live', Symbol='LOC105370165', Aliases='', description='uncharacterized LOC105370165', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37011559, end_position_on_the_genomic_accession=37019746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  403339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403339, CurrentID=0, Status='live', Symbol='DEFB133', Aliases='', description='defensin beta 133', other_designations='defensin, beta 133', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49946101, end_position_on_the_genomic_accession=49949444, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  403340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403340, CurrentID=0, Status='live', Symbol='MGC70870', Aliases='', description='C-terminal binding protein 2 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  403341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=403341, CurrentID=0, Status='live', Symbol='ZBTB34', Aliases='ZNF918', description='zinc finger and BTB domain containing 34', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126860665, end_position_on_the_genomic_accession=126885878, orientation='plus', exon_count=4, OMIM=611692, no_hdr0=''),
  105370169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370169, CurrentID=0, Status='live', Symbol='LOC105370169', Aliases='', description='uncharacterized LOC105370169', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39081231, end_position_on_the_genomic_accession=39091690, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  119358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119358, CurrentID=0, Status='live', Symbol='SNRPD2P1', Aliases='SNRPD2P, bA478K7.1', description='small nuclear ribonucleoprotein D2 pseudogene 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89978662, end_position_on_the_genomic_accession=89979168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100302058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302058, CurrentID=0, Status='live', Symbol='LKMCD', Aliases='', description='Leukoencephalopathy with metaphyseal chondrodysplasia', other_designations='', map_location='Xq25-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300660, no_hdr0=''),
  105370176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370176, CurrentID=0, Status='live', Symbol='LOC105370176', Aliases='', description='uncharacterized LOC105370176', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41974442, end_position_on_the_genomic_accession=42013965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100302059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302059, CurrentID=0, Status='live', Symbol='IBD27', Aliases='', description='Inflammatory bowel disease-27', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612796, no_hdr0=''),
  105370178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370178, CurrentID=0, Status='live', Symbol='LOC105370178', Aliases='', description='uncharacterized LOC105370178', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42927884, end_position_on_the_genomic_accession=42963222, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370181, CurrentID=0, Status='live', Symbol='LOC105370181', Aliases='', description='uncharacterized LOC105370181', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43811790, end_position_on_the_genomic_accession=43814816, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370185, CurrentID=0, Status='live', Symbol='LOC105370185', Aliases='', description='uncharacterized LOC105370185', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44684509, end_position_on_the_genomic_accession=44724311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374079, CurrentID=0, Status='live', Symbol='LOC105374079', Aliases='', description='uncharacterized LOC105374079', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125152359, end_position_on_the_genomic_accession=125155198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643659, CurrentID=0, Status='live', Symbol='LOC643659', Aliases='', description='hCG2019139', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=20004002, end_position_on_the_genomic_accession=20005068, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370189, CurrentID=0, Status='live', Symbol='LOC105370189', Aliases='', description='uncharacterized LOC105370189', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45079348, end_position_on_the_genomic_accession=45085072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370192, CurrentID=0, Status='live', Symbol='LOC105370192', Aliases='', description='uncharacterized LOC105370192', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46187661, end_position_on_the_genomic_accession=46194105, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370195, CurrentID=0, Status='live', Symbol='LOC105370195', Aliases='', description='uncharacterized LOC105370195', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47266126, end_position_on_the_genomic_accession=47270340, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370196, CurrentID=0, Status='live', Symbol='LOC105370196', Aliases='', description='uncharacterized LOC105370196', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47696392, end_position_on_the_genomic_accession=47705385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370197, CurrentID=0, Status='live', Symbol='LOC105370197', Aliases='', description='uncharacterized LOC105370197', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47945442, end_position_on_the_genomic_accession=47948275, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369715, CurrentID=0, Status='live', Symbol='LOC105369715', Aliases='', description='uncharacterized LOC105369715', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29868738, end_position_on_the_genomic_accession=30051496, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370201, CurrentID=0, Status='live', Symbol='LOC105370201', Aliases='', description='uncharacterized LOC105370201', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48567598, end_position_on_the_genomic_accession=48569871, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370203, CurrentID=0, Status='live', Symbol='LOC105370203', Aliases='', description='uncharacterized LOC105370203', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49203295, end_position_on_the_genomic_accession=49247853, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370205, CurrentID=0, Status='live', Symbol='LOC105370205', Aliases='', description='uncharacterized LOC105370205', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49635468, end_position_on_the_genomic_accession=49640311, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370206, CurrentID=0, Status='live', Symbol='LOC105370206', Aliases='', description='uncharacterized LOC105370206', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51279052, end_position_on_the_genomic_accession=51324734, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370207, CurrentID=0, Status='live', Symbol='LOC105370207', Aliases='', description='uncharacterized LOC105370207', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52344619, end_position_on_the_genomic_accession=52346942, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643680, CurrentID=0, Status='live', Symbol='MS4A4E', Aliases='', description='membrane-spanning 4-domains, subfamily A, member 4E', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60201253, end_position_on_the_genomic_accession=60243114, orientation='minus', exon_count=11, OMIM=608401, no_hdr0=''),
  105379009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379009, CurrentID=0, Status='live', Symbol='LOC105379009', Aliases='', description='uncharacterized LOC105379009', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68373030, end_position_on_the_genomic_accession=68374710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370211, CurrentID=0, Status='live', Symbol='LOC105370211', Aliases='', description='uncharacterized LOC105370211', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54097103, end_position_on_the_genomic_accession=54105803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  192111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192111, CurrentID=0, Status='live', Symbol='PGAM5', Aliases='BXLBV68', description='PGAM family member 5, serine/threonine protein phosphatase, mitochondrial', other_designations='Bcl-XL-binding protein v68|phosphoglycerate mutase family member 5', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132710807, end_position_on_the_genomic_accession=132722737, orientation='plus', exon_count=7, OMIM=614939, no_hdr0=''),
  192115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192115, CurrentID=0, Status='live', Symbol='MGR1', Aliases='MA, MGAU', description='Migraine with aura, susceptibility to', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=157300, no_hdr0=''),
  105370217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370217, CurrentID=0, Status='live', Symbol='LOC105370217', Aliases='', description='uncharacterized LOC105370217', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57457461, end_position_on_the_genomic_accession=57520231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  154791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154791, CurrentID=0, Status='live', Symbol='C7orf55', Aliases='FMC1, HSPC268', description='chromosome 7 open reading frame 55', other_designations='formation of mitochondrial complexes 1 homolog', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139340450, end_position_on_the_genomic_accession=139346319, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192127, CurrentID=0, Status='live', Symbol='SLC20A1P1', Aliases='', description='solute carrier family 20 (phosphate transporter), member 1 pseudogene 1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106682630, end_position_on_the_genomic_accession=106684349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192128, CurrentID=0, Status='live', Symbol='SLC20A1P2', Aliases='', description='solute carrier family 20 (phosphate transporter), member 1 pseudogene 2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106178885, end_position_on_the_genomic_accession=106180492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192130, CurrentID=0, Status='live', Symbol='GOLGA4P3', Aliases='', description='golgin A4 pseudogene 3', other_designations='golgi autoantigen, golgin subfamily a, 4 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106653466, end_position_on_the_genomic_accession=106654382, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192131, CurrentID=0, Status='live', Symbol='GOLGA4P1', Aliases='', description='golgin A4 pseudogene 1', other_designations='GOLGA4 pseudogene 1|golgi autoantigen, golgin subfamily a, 4 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106357284, end_position_on_the_genomic_accession=106357797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192132, CurrentID=0, Status='live', Symbol='GOLGA4P2', Aliases='', description='golgin A4 pseudogene 2', other_designations='GOLGA4 pseudogene 2|golgi autoantigen, golgin subfamily a, 4 pseudogene', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106332236, end_position_on_the_genomic_accession=106333081, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192133, CurrentID=0, Status='live', Symbol='RPS8P1', Aliases='RPS8_5_1417', description='ribosomal protein S8 pseudogene 1', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105978467, end_position_on_the_genomic_accession=105979169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192134, CurrentID=0, Status='live', Symbol='B3GNT6', Aliases='B3Gn-T6, BGnT-6, beta-1,3-Gn-T6, beta3Gn-T6', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 6', other_designations='beta-1,3-N-acetylglucosaminyltransferase 6|beta-1,3-N-acetylglucosaminyltransferase protein|core 3 synthase', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77034341, end_position_on_the_genomic_accession=77041973, orientation='plus', exon_count=2, OMIM=615315, no_hdr0=''),
  192142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192142, CurrentID=0, Status='live', Symbol='RLS1', Aliases='RLS', description='Restless legs syndrome, susceptibility to', other_designations='', map_location='12q12-q21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=102300, no_hdr0=''),
  192143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192143, CurrentID=0, Status='live', Symbol='STQTL2', Aliases='', description='stature quantitative trait locus 2', other_designations='', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606256, no_hdr0=''),
  192144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192144, CurrentID=0, Status='live', Symbol='HDLCQ1', Aliases='HDLC1', description='high density lipoprotein cholesterol level QTL 1', other_designations='', map_location='9p', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606613, no_hdr0=''),
  192146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192146, CurrentID=0, Status='live', Symbol='DDX50P1', Aliases='RH-II/GuBp1', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 50 pseudogene 1', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 50 pseudogene|RH-II/GuB pseudogene 1', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32201498, end_position_on_the_genomic_accession=32204218, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  192149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192149, CurrentID=0, Status='live', Symbol='BMIQ2', Aliases='', description='body mass index QTL 2', other_designations='Body mass index QTL on chromosome 13|body mass index quantitative trait 2', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606643, no_hdr0=''),
  105370225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370225, CurrentID=0, Status='live', Symbol='LOC105370225', Aliases='', description='uncharacterized LOC105370225', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59660656, end_position_on_the_genomic_accession=59665267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370227, CurrentID=0, Status='live', Symbol='LOC105370227', Aliases='', description='uncharacterized LOC105370227', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60164076, end_position_on_the_genomic_accession=60172615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372335, CurrentID=0, Status='live', Symbol='LOC105372335', Aliases='', description='uncharacterized LOC105372335', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23402529, end_position_on_the_genomic_accession=23438216, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  192217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192217, CurrentID=0, Status='live', Symbol='OXCT2P1', Aliases='OXCT2P', description='3-oxoacid CoA-transferase 2 pseudogene 1', other_designations='3-oxoacid CoA transferase 2 pseudogene 1|scot-t pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39514863, end_position_on_the_genomic_accession=39516667, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370285, CurrentID=0, Status='live', Symbol='LOC105370285', Aliases='', description='uncharacterized LOC105370285', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=82350203, end_position_on_the_genomic_accession=82396708, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370235, CurrentID=0, Status='live', Symbol='LOC105370235', Aliases='', description='uncharacterized LOC105370235', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63728291, end_position_on_the_genomic_accession=63735437, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100302119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302119, CurrentID=0, Status='live', Symbol='MIR1538', Aliases='MIRN1538, hsa-mir-1538', description='microRNA 1538', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69565808, end_position_on_the_genomic_accession=69565868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643709, CurrentID=0, Status='live', Symbol='LOC643709', Aliases='', description='macrophage expressed 1 pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59287342, end_position_on_the_genomic_accession=59289485, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370239, CurrentID=0, Status='live', Symbol='LOC105370239', Aliases='', description='uncharacterized LOC105370239', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=64691034, end_position_on_the_genomic_accession=64702076, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  163115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163115, CurrentID=0, Status='live', Symbol='ZNF781', Aliases='', description='zinc finger protein 781', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37667749, end_position_on_the_genomic_accession=37692315, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370241, CurrentID=0, Status='live', Symbol='LOC105370241', Aliases='', description='uncharacterized LOC105370241', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65835888, end_position_on_the_genomic_accession=65865951, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370242, CurrentID=0, Status='live', Symbol='LINC01052', Aliases='', description='long intergenic non-protein coding RNA 1052', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65866047, end_position_on_the_genomic_accession=65878058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10200, CurrentID=0, Status='live', Symbol='MPHOSPH6', Aliases='MPP, MPP-6, MPP6', description='M-phase phosphoprotein 6', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82148161, end_position_on_the_genomic_accession=82170224, orientation='minus', exon_count=6, OMIM=605500, no_hdr0=''),
  100128498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128498, CurrentID=0, Status='live', Symbol='LOC100128498', Aliases='', description='uncharacterized LOC100128498', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=67182786, end_position_on_the_genomic_accession=67183831, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  192286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192286, CurrentID=0, Status='live', Symbol='HIGD2A', Aliases='RCF1b', description='HIG1 hypoxia inducible domain family member 2A', other_designations='HIG1 domain family, member 2A|HIG1 hypoxia inducible domain family, member 2A|RCF1 homolog B', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176388783, end_position_on_the_genomic_accession=176389750, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  359729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359729, CurrentID=0, Status='live', Symbol='MRPL57P2', Aliases='MRP63P2', description='mitochondrial ribosomal protein L57 pseudogene 2', other_designations='mitochondrial ribosomal protein 63 pseudogene 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227572240, end_position_on_the_genomic_accession=227572447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370247, CurrentID=0, Status='live', Symbol='LOC105370247', Aliases='', description='uncharacterized LOC105370247', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66560385, end_position_on_the_genomic_accession=66660491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105379796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379796, CurrentID=0, Status='live', Symbol='LOC105379796', Aliases='', description='uncharacterized LOC105379796', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  359731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359731, CurrentID=0, Status='live', Symbol='MRPL57P7', Aliases='MRP63P7', description='mitochondrial ribosomal protein L57 pseudogene 7', other_designations='mitochondrial ribosomal protein 63 pseudogene 7', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98906255, end_position_on_the_genomic_accession=98906527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370249, CurrentID=0, Status='live', Symbol='LOC105370249', Aliases='', description='uncharacterized LOC105370249', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67596363, end_position_on_the_genomic_accession=67649363, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105379017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379017, CurrentID=0, Status='live', Symbol='LOC105379017', Aliases='', description='uncharacterized LOC105379017', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70926382, end_position_on_the_genomic_accession=70930090, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370251, CurrentID=0, Status='live', Symbol='LOC105370251', Aliases='', description='uncharacterized LOC105370251', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67783791, end_position_on_the_genomic_accession=67858048, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100302134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302134, CurrentID=0, Status='live', Symbol='MIR1289-2', Aliases='MIRN1289-2, hsa-mir-1289-2', description='microRNA 1289-2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133427596, end_position_on_the_genomic_accession=133427706, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372862, CurrentID=0, Status='live', Symbol='LOC105372862', Aliases='', description='uncharacterized LOC105372862', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20206038, end_position_on_the_genomic_accession=20212157, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  359735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359735, CurrentID=0, Status='live', Symbol='MRPL14P1', Aliases='', description='mitochondrial ribosomal protein L14 pseudogene 1', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1155977, end_position_on_the_genomic_accession=1156173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370253, CurrentID=0, Status='live', Symbol='LOC105370253', Aliases='', description='uncharacterized LOC105370253', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68878384, end_position_on_the_genomic_accession=68894417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  359736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359736, CurrentID=0, Status='live', Symbol='MRPL15P1', Aliases='MRPL15P2', description='mitochondrial ribosomal protein L15 pseudogene 1', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89752344, end_position_on_the_genomic_accession=89753533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359737, CurrentID=0, Status='live', Symbol='MRPL20P1', Aliases='', description='mitochondrial ribosomal protein L20 pseudogene 1', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36994558, end_position_on_the_genomic_accession=36995248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370255, CurrentID=0, Status='live', Symbol='LOC105370255', Aliases='', description='uncharacterized LOC105370255', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=70765778, end_position_on_the_genomic_accession=70827935, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  359738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359738, CurrentID=0, Status='live', Symbol='MRPL22P1', Aliases='', description='mitochondrial ribosomal protein L22 pseudogene 1', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56320647, end_position_on_the_genomic_accession=56321334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372336, CurrentID=0, Status='live', Symbol='LOC105372336', Aliases='', description='uncharacterized LOC105372336', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23520827, end_position_on_the_genomic_accession=23522928, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100302139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302139, CurrentID=0, Status='live', Symbol='MIR1537', Aliases='MIRN1537, hsa-mir-1537, mir-1537', description='microRNA 1537', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235853000, end_position_on_the_genomic_accession=235853060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132865, CurrentID=0, Status='live', Symbol='LOC100132865', Aliases='', description='PRAME family member 17-like', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  359740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359740, CurrentID=0, Status='live', Symbol='MRPL30P1', Aliases='', description='mitochondrial ribosomal protein L30 pseudogene 1', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57029496, end_position_on_the_genomic_accession=57029866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359741, CurrentID=0, Status='live', Symbol='MRPL32P1', Aliases='', description='mitochondrial ribosomal protein L32 pseudogene 1', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48583093, end_position_on_the_genomic_accession=48583356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370259, CurrentID=0, Status='live', Symbol='LOC105370259', Aliases='', description='uncharacterized LOC105370259', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74254221, end_position_on_the_genomic_accession=74258200, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  359742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359742, CurrentID=0, Status='live', Symbol='MRPL35P1', Aliases='', description='mitochondrial ribosomal protein L35 pseudogene 1', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13949095, end_position_on_the_genomic_accession=13949466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370290, CurrentID=0, Status='live', Symbol='LOC105370290', Aliases='', description='uncharacterized LOC105370290', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84872099, end_position_on_the_genomic_accession=84902431, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  285735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285735, CurrentID=0, Status='live', Symbol='LINC00326', Aliases='NCRNA00326', description='long intergenic non-protein coding RNA 326', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133088080, end_position_on_the_genomic_accession=133106578, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  359743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359743, CurrentID=0, Status='live', Symbol='MRPL35P2', Aliases='', description='mitochondrial ribosomal protein L35 pseudogene 2', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63634457, end_position_on_the_genomic_accession=63634827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370261, CurrentID=0, Status='live', Symbol='LOC105370261', Aliases='', description='uncharacterized LOC105370261', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76267535, end_position_on_the_genomic_accession=76281295, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102464832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464832, CurrentID=0, Status='live', Symbol='MIR6083', Aliases='hsa-mir-6083', description='microRNA 6083', other_designations='microRNA mir-6083', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124374332, end_position_on_the_genomic_accession=124374437, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  359745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359745, CurrentID=0, Status='live', Symbol='MRPL35P4', Aliases='', description='mitochondrial ribosomal protein L35 pseudogene 4', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12845062, end_position_on_the_genomic_accession=12845430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370263, CurrentID=0, Status='live', Symbol='LOC105370263', Aliases='', description='uncharacterized LOC105370263', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76459888, end_position_on_the_genomic_accession=76527128, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  359746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359746, CurrentID=0, Status='live', Symbol='MRPL42P3', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 3', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144136437, end_position_on_the_genomic_accession=144136775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359747, CurrentID=0, Status='live', Symbol='MRPL45P1', Aliases='', description='mitochondrial ribosomal protein L45 pseudogene 1', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88364495, end_position_on_the_genomic_accession=88365636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  285736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285736, CurrentID=0, Status='live', Symbol='FLJ35700', Aliases='', description='uncharacterized protein FLJ35700', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  359748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359748, CurrentID=0, Status='live', Symbol='MRPL49P1', Aliases='', description='mitochondrial ribosomal protein L49 pseudogene 1', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64674854, end_position_on_the_genomic_accession=64675390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359749, CurrentID=0, Status='live', Symbol='MRPL50P2', Aliases='', description='mitochondrial ribosomal protein L50 pseudogene 2', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203315390, end_position_on_the_genomic_accession=203315871, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643739, CurrentID=0, Status='live', Symbol='GAPDHP31', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 31', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7566298, end_position_on_the_genomic_accession=7567579, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131822, CurrentID=0, Status='live', Symbol='LOC100131822', Aliases='', description='uncharacterized LOC100131822', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=32957662, end_position_on_the_genomic_accession=32958540, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  285737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285737, CurrentID=0, Status='live', Symbol='CCRL1P1', Aliases='CCRL1P, dJ509I19.4', description='chemokine (C-C motif) receptor-like 1 pseudogene', other_designations='pseudogene of chemokine (C-C motif) receptor-like 1 (CCRL1)', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138822707, end_position_on_the_genomic_accession=138824563, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370272, CurrentID=0, Status='live', Symbol='LOC105370272', Aliases='', description='uncharacterized LOC105370272', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78760826, end_position_on_the_genomic_accession=78762918, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370277, CurrentID=0, Status='live', Symbol='LOC105370277', Aliases='', description='uncharacterized LOC105370277', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  359761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359761, CurrentID=0, Status='live', Symbol='MRPS18AP1', Aliases='', description='mitochondrial ribosomal protein S18A pseudogene 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48256040, end_position_on_the_genomic_accession=48256961, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370279, CurrentID=0, Status='live', Symbol='LOC105370279', Aliases='', description='uncharacterized LOC105370279', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370280, CurrentID=0, Status='live', Symbol='LOC105370280', Aliases='', description='uncharacterized LOC105370280', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81008053, end_position_on_the_genomic_accession=81009261, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370283, CurrentID=0, Status='live', Symbol='LOC105370283', Aliases='', description='uncharacterized LOC105370283', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81539918, end_position_on_the_genomic_accession=81592715, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370284, CurrentID=0, Status='live', Symbol='LOC105370284', Aliases='', description='uncharacterized LOC105370284', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81802360, end_position_on_the_genomic_accession=81848750, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271119, CurrentID=0, Status='live', Symbol='RPS27P16', Aliases='RPS27_9_801', description='ribosomal protein S27 pseudogene 16', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30167828, end_position_on_the_genomic_accession=30168004, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285740, CurrentID=0, Status='live', Symbol='PHACTR2-AS1', Aliases='', description='PHACTR2 antisense RNA 1', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143554115, end_position_on_the_genomic_accession=143569339, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  285741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285741, CurrentID=0, Status='live', Symbol='TPT1P4', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 4', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144200240, end_position_on_the_genomic_accession=144201057, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419512, CurrentID=0, Status='live', Symbol='LOC100419512', Aliases='', description='zinc finger protein 19 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39929194, end_position_on_the_genomic_accession=39929974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370291, CurrentID=0, Status='live', Symbol='LOC105370291', Aliases='', description='uncharacterized LOC105370291', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85111801, end_position_on_the_genomic_accession=85205470, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100462647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462647, CurrentID=0, Status='live', Symbol='NKAIN1P1', Aliases='', description='Na+/K+ transporting ATPase interacting 1 pseudogene 1', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143487662, end_position_on_the_genomic_accession=143489198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462648, CurrentID=0, Status='live', Symbol='LOC100462648', Aliases='', description='Na+/K+ transporting ATPase interacting 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91723013, end_position_on_the_genomic_accession=91724351, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462652, CurrentID=0, Status='live', Symbol='LOC100462652', Aliases='', description='solute carrier family 6 (amino acid transporter), member 14 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116688829, end_position_on_the_genomic_accession=116692685, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370294, CurrentID=0, Status='live', Symbol='LOC105370294', Aliases='', description='uncharacterized LOC105370294', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379026, CurrentID=0, Status='live', Symbol='LOC105379026', Aliases='', description='uncharacterized LOC105379026', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71337349, end_position_on_the_genomic_accession=71346915, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100462676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462676, CurrentID=0, Status='live', Symbol='RCHTS', Aliases='', description='Roifman-Chitayat syndrome', other_designations='', map_location='15q11-q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613328, no_hdr0=''),
  105370298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370298, CurrentID=0, Status='live', Symbol='LOC105370298', Aliases='', description='proline-rich protein 20A', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100302183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302183, CurrentID=0, Status='live', Symbol='MIR1825', Aliases='MIRN1825, hsa-mir-1825', description='microRNA 1825', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32237795, end_position_on_the_genomic_accession=32237847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370301, CurrentID=0, Status='live', Symbol='LOC105370301', Aliases='', description='uncharacterized LOC105370301', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=86984386, end_position_on_the_genomic_accession=87051179, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927025, CurrentID=0, Status='live', Symbol='LOC101927025', Aliases='', description='uncharacterized LOC101927025', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93089403, end_position_on_the_genomic_accession=93170550, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927008, CurrentID=0, Status='live', Symbol='LOC101927008', Aliases='', description='uncharacterized LOC101927008', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2566027, end_position_on_the_genomic_accession=2574621, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100462721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462721, CurrentID=0, Status='live', Symbol='INTLQ1', Aliases='', description='Intelligence QTL1', other_designations='', map_location='4p16-q34', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603783, no_hdr0=''),
  100462722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462722, CurrentID=0, Status='live', Symbol='INTLQ3', Aliases='', description='Intelligence QTL3', other_designations='', map_location='6p25.3-p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610294, no_hdr0=''),
  100462723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462723, CurrentID=0, Status='live', Symbol='INTLQ2', Aliases='', description='Intelligence QTL3', other_designations='', map_location='2q24.1-q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610295, no_hdr0=''),
  192644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192644, CurrentID=0, Status='live', Symbol='DFNA21', Aliases='', description='deafness, autosomal dominant 21', other_designations='', map_location='6p24.1-p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607017, no_hdr0=''),
  100462725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462725, CurrentID=0, Status='live', Symbol='NAFLD2', Aliases='', description='Fatty liver disease, nonalcoholic, susceptibility to, 2', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613387, no_hdr0=''),
  100462732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462732, CurrentID=0, Status='live', Symbol='ATP6V1G1P6', Aliases='', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 6', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78222255, end_position_on_the_genomic_accession=78222623, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370307, CurrentID=0, Status='live', Symbol='LOC105370307', Aliases='', description='uncharacterized LOC105370307', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89035781, end_position_on_the_genomic_accession=89087431, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192666, CurrentID=0, Status='live', Symbol='KRT24', Aliases='K24, KA24', description='keratin 24, type I', other_designations='CK-24|cytokeratin-24|keratin-24|type I keratin-24', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40697991, end_position_on_the_genomic_accession=40703750, orientation='minus', exon_count=8, OMIM=607742, no_hdr0=''),
  100302191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302191, CurrentID=0, Status='live', Symbol='MIR548I4', Aliases='MIR548I-4, MIRN548I4, hsa-mir-548i-4', description='microRNA 548i-4', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84225752, end_position_on_the_genomic_accession=84225828, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  192668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192668, CurrentID=0, Status='live', Symbol='CYS1', Aliases='', description='cystin 1', other_designations='cilia-associated protein', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10056799, end_position_on_the_genomic_accession=10080411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  192669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192669, CurrentID=0, Status='live', Symbol='AGO3', Aliases='EIF2C3', description='argonaute 3, RISC catalytic component', other_designations='argonaute 3|argonaute RISC catalytic component 3|eukaryotic translation initiation factor 2C, 3|hAgo3', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35923403, end_position_on_the_genomic_accession=36056747, orientation='plus', exon_count=24, OMIM=607355, no_hdr0=''),
  192670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192670, CurrentID=0, Status='live', Symbol='AGO4', Aliases='EIF2C4', description='argonaute 4, RISC catalytic component', other_designations='argonaute 4|argonaute RISC catalytic component 4|argonaute4|eIF-2C 4|eIF2C 4|eukaryotic translation initiation factor 2C, 4|hAgo4', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35807630, end_position_on_the_genomic_accession=35857890, orientation='plus', exon_count=21, OMIM=607356, no_hdr0=''),
  192676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192676, CurrentID=0, Status='live', Symbol='DFNA43', Aliases='', description='deafness, autosomal dominant 43', other_designations='deafness, autosomal dominant 40', map_location='2p12-p11.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608394, no_hdr0=''),
  192677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192677, CurrentID=0, Status='live', Symbol='SLEN1', Aliases='', description='systemic lupus erythematosus with nephritis 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607965, no_hdr0=''),
  192679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192679, CurrentID=0, Status='live', Symbol='SLEN2', Aliases='', description='systemic lupus erythematosus with nephritis 2', other_designations='', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607966, no_hdr0=''),
  192683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192683, CurrentID=0, Status='live', Symbol='SCAMP5', Aliases='', description='secretory carrier membrane protein 5', other_designations='hSCAMP5', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74995535, end_position_on_the_genomic_accession=75021495, orientation='plus', exon_count=11, OMIM=613766, no_hdr0=''),
  100462767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462767, CurrentID=0, Status='live', Symbol='ACTBP12', Aliases='ACTGP5', description='actin, beta pseudogene 12', other_designations='actin, gamma pseudogene 5', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92229016, end_position_on_the_genomic_accession=92229456, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643784, CurrentID=0, Status='live', Symbol='LOC643784', Aliases='', description='NLR family, apoptosis inhibitory protein pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69609426, end_position_on_the_genomic_accession=69623646, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462772, CurrentID=0, Status='live', Symbol='FTH1P22', Aliases='FTHL22', description='ferritin, heavy polypeptide 1 pseudogene 22', other_designations='ferritin, heavy polypeptide-like 22', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116775078, end_position_on_the_genomic_accession=116775567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462776, CurrentID=0, Status='live', Symbol='RRAGAP1', Aliases='RRAGAP', description='Ras-related GTP binding A pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44888558, end_position_on_the_genomic_accession=44889053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462780, CurrentID=0, Status='live', Symbol='FTLP9', Aliases='', description='ferritin, light polypeptide pseudogene 9', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77216419, end_position_on_the_genomic_accession=77216866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462782, CurrentID=0, Status='live', Symbol='MEMO1P3', Aliases='', description='mediator of cell motility 1 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167888413, end_position_on_the_genomic_accession=167888878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462785, CurrentID=0, Status='live', Symbol='MEMO1P4', Aliases='', description='mediator of cell motility 1 pseudogene 4', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123765322, end_position_on_the_genomic_accession=123765692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462786, CurrentID=0, Status='live', Symbol='FTLP18', Aliases='', description='ferritin, light polypeptide pseudogene 18', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36630335, end_position_on_the_genomic_accession=36630812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462787, CurrentID=0, Status='live', Symbol='CICP10', Aliases='', description='capicua transcriptional repressor pseudogene 10', other_designations='capicua homolog pseudogene 10', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242119877, end_position_on_the_genomic_accession=242120142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462788, CurrentID=0, Status='live', Symbol='FTLP6', Aliases='', description='ferritin, light polypeptide pseudogene 6', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66771253, end_position_on_the_genomic_accession=66771670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462790, CurrentID=0, Status='live', Symbol='FTLP8', Aliases='', description='ferritin, light polypeptide pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97986949, end_position_on_the_genomic_accession=97987225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462792, CurrentID=0, Status='live', Symbol='FTLP12', Aliases='', description='ferritin, light polypeptide pseudogene 12', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16391434, end_position_on_the_genomic_accession=16391883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462793, CurrentID=0, Status='live', Symbol='FTLP13', Aliases='', description='ferritin, light polypeptide pseudogene 13', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22221220, end_position_on_the_genomic_accession=22221645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462794, CurrentID=0, Status='live', Symbol='FTLP14', Aliases='', description='ferritin, light polypeptide pseudogene 14', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68822449, end_position_on_the_genomic_accession=68823232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462795, CurrentID=0, Status='live', Symbol='H2AFZP4', Aliases='', description='H2A histone family member Z pseudogene 4', other_designations='H2A histone family, member Z pseudogene 4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70278915, end_position_on_the_genomic_accession=70279297, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462797, CurrentID=0, Status='live', Symbol='FTH1P24', Aliases='FTHL24', description='ferritin, heavy polypeptide 1 pseudogene 24', other_designations='ferritin, heavy polypeptide-like 24', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139546266, end_position_on_the_genomic_accession=139546724, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462799, CurrentID=0, Status='live', Symbol='FTH1P25', Aliases='FTHL25', description='ferritin, heavy polypeptide 1 pseudogene 25', other_designations='ferritin, heavy polypeptide-like 25', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185071565, end_position_on_the_genomic_accession=185072035, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462800, CurrentID=0, Status='live', Symbol='H2AFZP6', Aliases='', description='H2A histone family member Z pseudogene 6', other_designations='H2A histone family, member Z pseudogene 6', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31521202, end_position_on_the_genomic_accession=31521587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370318, CurrentID=0, Status='live', Symbol='LOC105370318', Aliases='', description='uncharacterized LOC105370318', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93801687, end_position_on_the_genomic_accession=93809266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100462809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462809, CurrentID=0, Status='live', Symbol='RSL24D1P8', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 8', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12096424, end_position_on_the_genomic_accession=12096909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462810, CurrentID=0, Status='live', Symbol='RHOT1P3', Aliases='', description='ras homolog family member T1 pseudogene 3', other_designations='ras homolog gene family, member T1 pseudogene 3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18837767, end_position_on_the_genomic_accession=18838097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462811, CurrentID=0, Status='live', Symbol='RHOT1P2', Aliases='', description='ras homolog family member T1 pseudogene 2', other_designations='ras homolog gene family, member T1 pseudogene 2', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13936993, end_position_on_the_genomic_accession=13937325, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462812, CurrentID=0, Status='live', Symbol='FGFR3P1', Aliases='D84394.2, FGFR3P, FGFR6', description='fibroblast growth factor receptor 3 pseudogene 1', other_designations='fibroblast growth factor receptor 6', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31377717, end_position_on_the_genomic_accession=31378028, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462813, CurrentID=0, Status='live', Symbol='RHOT1P1', Aliases='RHOT1P, RHOT3', description='ras homolog family member T1 pseudogene 1', other_designations='ras homolog gene family, member T1 pseudogene 1|ras homolog gene family, member T3', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14222008, end_position_on_the_genomic_accession=14222340, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462815, CurrentID=0, Status='live', Symbol='FGFR3P2', Aliases='', description='fibroblast growth factor receptor 3 pseudogene 2', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95756954, end_position_on_the_genomic_accession=95757444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462816, CurrentID=0, Status='live', Symbol='FGFR3P3', Aliases='', description='fibroblast growth factor receptor 3 pseudogene 3', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6220330, end_position_on_the_genomic_accession=6220745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370320, CurrentID=0, Status='live', Symbol='LOC105370320', Aliases='', description='uncharacterized LOC105370320', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95274007, end_position_on_the_genomic_accession=95280235, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100462820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462820, CurrentID=0, Status='live', Symbol='USP9YP6', Aliases='CDY24P', description='ubiquitin specific peptidase 9, Y-linked pseudogene 6', other_designations='USP9Y pseudogene 6|chromodomain protein, Y-linked 24 pseudogene|ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17914911, end_position_on_the_genomic_accession=17921630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462821, CurrentID=0, Status='live', Symbol='USP9YP7', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 7', other_designations='USP9Y pseudogene 7', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17910591, end_position_on_the_genomic_accession=17910787, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462824, CurrentID=0, Status='live', Symbol='EIF4A1P3', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131782573, end_position_on_the_genomic_accession=131782791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462825, CurrentID=0, Status='live', Symbol='ATP7BP1', Aliases='', description='ATPase, Cu++ transporting, beta polypeptide pseudogene 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22418817, end_position_on_the_genomic_accession=22419296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462826, CurrentID=0, Status='live', Symbol='EIF4A1P13', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 13', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88448788, end_position_on_the_genomic_accession=88449026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462829, CurrentID=0, Status='live', Symbol='GCNT1P4', Aliases='', description='glucosaminyl (N-acetyl) transferase 1, core 2 pseudogene 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63857318, end_position_on_the_genomic_accession=63857770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462830, CurrentID=0, Status='live', Symbol='EIF4A1P12', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 12', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28593773, end_position_on_the_genomic_accession=28593916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462832, CurrentID=0, Status='live', Symbol='PNPLA4P1', Aliases='', description='patatin like phospholipase domain containing 4 pseudogene 1', other_designations='patatin-like phospholipase domain containing 4 pseudogene 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14081075, end_position_on_the_genomic_accession=14081461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462833, CurrentID=0, Status='live', Symbol='ISCUP1', Aliases='', description='iron-sulfur cluster assembly enzyme pseudogene 1', other_designations='iron-sulfur cluster scaffold homolog pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170211010, end_position_on_the_genomic_accession=170211397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462836, CurrentID=0, Status='live', Symbol='EIF3EP3', Aliases='', description='eukaryotic translation initiation factor 3 subunit E pseudogene 3', other_designations='eukaryotic translation initiation factor 3, subunit E pseudogene 3', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164573740, end_position_on_the_genomic_accession=164574015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462838, CurrentID=0, Status='live', Symbol='ATP6V1G1P3', Aliases='', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136779938, end_position_on_the_genomic_accession=136780215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462840, CurrentID=0, Status='live', Symbol='ATP6V1G1P5', Aliases='', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151319425, end_position_on_the_genomic_accession=151319778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462843, CurrentID=0, Status='live', Symbol='ATP6V1G1P4', Aliases='', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35158306, end_position_on_the_genomic_accession=35158637, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462844, CurrentID=0, Status='live', Symbol='ATP6V1E1P2', Aliases='ATP6EP2, ATP6V1EP2', description='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E1 pseudogene 2', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) 31kD pseudogene 2|ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E pseudogene 2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26952201, end_position_on_the_genomic_accession=26952504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370325, CurrentID=0, Status='live', Symbol='LOC105370325', Aliases='', description='uncharacterized LOC105370325', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97539579, end_position_on_the_genomic_accession=97550113, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100462850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462850, CurrentID=0, Status='live', Symbol='COX6B1P4', Aliases='COX6BP4', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 4', other_designations='cytochrome c oxidase subunit VIb pseudogene 4', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26835983, end_position_on_the_genomic_accession=26836248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462851, CurrentID=0, Status='live', Symbol='COX6B1P6', Aliases='', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 6', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89823542, end_position_on_the_genomic_accession=89823830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462857, CurrentID=0, Status='live', Symbol='LSM6P2', Aliases='', description='LSM6 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 2', other_designations='LSM6 U6 small nuclear RNA and mRNA degradation associated pseudogene 2|LSM6 homolog, U6 small nuclear RNA associated pseudogene 2|LSM6 pseudogene 2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49770960, end_position_on_the_genomic_accession=49771136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462862, CurrentID=0, Status='live', Symbol='ATP5LP8', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 8', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 8|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 8', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28144903, end_position_on_the_genomic_accession=28145205, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462865, CurrentID=0, Status='live', Symbol='ATP5LP6', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 6', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 6|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 6', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76202526, end_position_on_the_genomic_accession=76202796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462866, CurrentID=0, Status='live', Symbol='TCEB1P21', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 21', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70955700, end_position_on_the_genomic_accession=70956318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462867, CurrentID=0, Status='live', Symbol='TCEB1P6', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 6', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17837201, end_position_on_the_genomic_accession=17837528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462868, CurrentID=0, Status='live', Symbol='TCEB1P7', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 7', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18532326, end_position_on_the_genomic_accession=18532656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462869, CurrentID=0, Status='live', Symbol='TCEB1P8', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 8', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22517242, end_position_on_the_genomic_accession=22517573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462870, CurrentID=0, Status='live', Symbol='TCEB1P9', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 9', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23808821, end_position_on_the_genomic_accession=23809146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462871, CurrentID=0, Status='live', Symbol='TCEB1P10', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 10', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24006911, end_position_on_the_genomic_accession=24007237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462872, CurrentID=0, Status='live', Symbol='TCEB1P11', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 11', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25991241, end_position_on_the_genomic_accession=25991572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462873, CurrentID=0, Status='live', Symbol='TCEB1P4', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 4', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9230926, end_position_on_the_genomic_accession=9231247, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462874, CurrentID=0, Status='live', Symbol='TCEB1P12', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 12', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18068483, end_position_on_the_genomic_accession=18068810, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462875, CurrentID=0, Status='live', Symbol='TCEB1P13', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 13', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18440797, end_position_on_the_genomic_accession=18441125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462876, CurrentID=0, Status='live', Symbol='TCEB1P14', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 14', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18787665, end_position_on_the_genomic_accession=18788326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462877, CurrentID=0, Status='live', Symbol='TCEB1P15', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 15', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22068820, end_position_on_the_genomic_accession=22069151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462878, CurrentID=0, Status='live', Symbol='TCEB1P16', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 16', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23678504, end_position_on_the_genomic_accession=23678835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462879, CurrentID=0, Status='live', Symbol='TCEB1P17', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 17', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25860943, end_position_on_the_genomic_accession=25861268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462880, CurrentID=0, Status='live', Symbol='TCEB1P19', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 19', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88829113, end_position_on_the_genomic_accession=88829416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462881, CurrentID=0, Status='live', Symbol='TCEB1P23', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 23', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78488182, end_position_on_the_genomic_accession=78488512, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643803, CurrentID=0, Status='live', Symbol='KRTAP24-1', Aliases='KAP24.1', description='keratin associated protein 24-1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30281309, end_position_on_the_genomic_accession=30282958, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100462886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462886, CurrentID=0, Status='live', Symbol='TAF13P2', Aliases='', description='TATA-box binding protein associated factor 13 pseudogene 2', other_designations='TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 18kDa pseudogene 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74351086, end_position_on_the_genomic_accession=74351419, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  10289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10289, CurrentID=0, Status='live', Symbol='EIF1B', Aliases='GC20', description='eukaryotic translation initiation factor 1B', other_designations='protein translation factor SUI1 homolog GC20|translation factor sui1 homolog', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40309682, end_position_on_the_genomic_accession=40312424, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379034, CurrentID=0, Status='live', Symbol='LOC105379034', Aliases='', description='uncharacterized LOC105379034', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73967619, end_position_on_the_genomic_accession=73985925, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370337, CurrentID=0, Status='live', Symbol='LOC105370337', Aliases='', description='uncharacterized LOC105370337', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102760392, end_position_on_the_genomic_accession=102768293, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379035, CurrentID=0, Status='live', Symbol='LOC105379035', Aliases='', description='uncharacterized LOC105379035', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74039887, end_position_on_the_genomic_accession=74054359, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370340, CurrentID=0, Status='live', Symbol='LOC105370340', Aliases='', description='uncharacterized LOC105370340', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=104125899, end_position_on_the_genomic_accession=104134288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100462953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462953, CurrentID=0, Status='live', Symbol='MINOS1P1', Aliases='CG012, N4BP2L2-IT, N4BP2L2-IT1, N4BP2L2IT1', description='mitochondrial inner membrane organizing system 1 pseudogene 1', other_designations='N4BP2L2 intronic transcript (non-protein coding)|N4BP2L2 intronic transcript 1 (non-protein coding)|N4BPL2 intronic transcript 1 (non-protein coding)', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32496206, end_position_on_the_genomic_accession=32498492, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100462954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462954, CurrentID=0, Status='live', Symbol='MINOS1P3', Aliases='', description='mitochondrial inner membrane organizing system 1 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27214562, end_position_on_the_genomic_accession=27215089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370343, CurrentID=0, Status='live', Symbol='LOC105370343', Aliases='', description='uncharacterized LOC105370343', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=104946966, end_position_on_the_genomic_accession=104969010, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100462957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462957, CurrentID=0, Status='live', Symbol='MINOS1P4', Aliases='', description='mitochondrial inner membrane organizing system 1 pseudogene 4', other_designations='', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78379190, end_position_on_the_genomic_accession=78379658, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373228, CurrentID=0, Status='live', Symbol='LOC105373228', Aliases='', description='uncharacterized LOC105373228', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240685683, end_position_on_the_genomic_accession=240710019, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100462977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462977, CurrentID=0, Status='live', Symbol='MTRNR2L1', Aliases='HN1', description='MT-RNR2-like 1', other_designations='MTRNR2-like 1|humanin-like protein 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22523111, end_position_on_the_genomic_accession=22524665, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100462981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462981, CurrentID=0, Status='live', Symbol='MTRNR2L2', Aliases='HN2', description='MT-RNR2-like 2', other_designations='MTRNR2-like 2|humanin-like protein 2', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80650000, end_position_on_the_genomic_accession=80651035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100462982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462982, CurrentID=0, Status='live', Symbol='CCT5-2P', Aliases='', description='chaperonin containing TCP1, subunit 5 (epsilon) pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78910722, end_position_on_the_genomic_accession=78910831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462983, CurrentID=0, Status='live', Symbol='MTRNR2L3', Aliases='HN3', description='MT-RNR2-like 3', other_designations='MTRNR2-like 3|humanin-like protein 3', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57358440, end_position_on_the_genomic_accession=57359822, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100462984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462984, CurrentID=0, Status='live', Symbol='HSPD1P22', Aliases='HSPD1-21P', description='heat shock protein family D (Hsp60) member 1 pseudogene 22', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 22', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61662846, end_position_on_the_genomic_accession=61663292, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462985, CurrentID=0, Status='live', Symbol='HSPD1P16', Aliases='HSPD1-13P', description='heat shock protein family D (Hsp60) member 1 pseudogene 16', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 16', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152705798, end_position_on_the_genomic_accession=152706115, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462986, CurrentID=0, Status='live', Symbol='HSPD1P17', Aliases='HSPD1-19P', description='heat shock protein family D (Hsp60) member 1 pseudogene 17', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 17', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49440816, end_position_on_the_genomic_accession=49440956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462987, CurrentID=0, Status='live', Symbol='HSPD1P13', Aliases='HSPD1-16P', description='heat shock protein family D (Hsp60) member 1 pseudogene 13', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 13', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105706866, end_position_on_the_genomic_accession=105707819, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100462988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462988, CurrentID=0, Status='live', Symbol='LOC100462988', Aliases='', description='heat shock protein family D (Hsp60) member 1 pseudogene', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8016711, end_position_on_the_genomic_accession=8017860, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462989, CurrentID=0, Status='live', Symbol='HSPD1P20', Aliases='HSPD1-20P', description='heat shock protein family D (Hsp60) member 1 pseudogene 20', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 20', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80005987, end_position_on_the_genomic_accession=80006575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100462990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462990, CurrentID=0, Status='live', Symbol='HSPD1P21', Aliases='HSPD1-3P', description='heat shock protein family D (Hsp60) member 1 pseudogene 21', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 21', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45539975, end_position_on_the_genomic_accession=45540286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  613206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613206, CurrentID=0, Status='live', Symbol='LOC613206', Aliases='', description='myeloproliferative disease associated tumor antigen 5', other_designations='', map_location='9q33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124258585, end_position_on_the_genomic_accession=124261310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  163227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163227, CurrentID=0, Status='live', Symbol='ZNF100', Aliases='', description='zinc finger protein 100', other_designations='zinc finger protein 100 (Y1)', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21724041, end_position_on_the_genomic_accession=21767628, orientation='minus', exon_count=8, OMIM=603982, no_hdr0=''),
  101927010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927010, CurrentID=0, Status='live', Symbol='LOC101927010', Aliases='', description='uncharacterized LOC101927010', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121988417, end_position_on_the_genomic_accession=121990450, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379038, CurrentID=0, Status='live', Symbol='LOC105379038', Aliases='', description='uncharacterized LOC105379038', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74682891, end_position_on_the_genomic_accession=74685097, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370355, CurrentID=0, Status='live', Symbol='LOC105370355', Aliases='', description='uncharacterized LOC105370355', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108328055, end_position_on_the_genomic_accession=108365281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106099130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106099130, CurrentID=0, Status='live', Symbol='GSTT1-AS1', Aliases='', description='GSTT1 antisense RNA 1', other_designations='LOC106099130 uncharacterized lncRNA', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390503, CurrentID=0, Status='live', Symbol='ADIPOR1P2', Aliases='', description='adiponectin receptor 1 pseudogene 2', other_designations='', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94633672, end_position_on_the_genomic_accession=94634525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370359, CurrentID=0, Status='live', Symbol='LOC105370359', Aliases='', description='uncharacterized LOC105370359', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109371446, end_position_on_the_genomic_accession=109731829, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128521, CurrentID=0, Status='live', Symbol='TPT1P14', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 14', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14977752, end_position_on_the_genomic_accession=14978540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420944, CurrentID=0, Status='live', Symbol='LOC100420944', Aliases='', description='nuclear transcription factor Y subunit beta pseudogene', other_designations='nuclear transcription factor Y, beta pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39997869, end_position_on_the_genomic_accession=39998492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370361, CurrentID=0, Status='live', Symbol='LOC105370361', Aliases='', description='uncharacterized LOC105370361', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110121708, end_position_on_the_genomic_accession=110125234, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370362, CurrentID=0, Status='live', Symbol='LOC105370362', Aliases='', description='uncharacterized LOC105370362', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110813203, end_position_on_the_genomic_accession=110820209, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  119548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119548, CurrentID=0, Status='live', Symbol='PNLIPRP3', Aliases='', description='pancreatic lipase-related protein 3', other_designations='PL-RP3', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116427879, end_position_on_the_genomic_accession=116477961, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105370311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370311, CurrentID=0, Status='live', Symbol='LOC105370311', Aliases='', description='uncharacterized LOC105370311', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90880315, end_position_on_the_genomic_accession=90896811, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370365, CurrentID=0, Status='live', Symbol='LOC105370365', Aliases='', description='uncharacterized LOC105370365', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111317105, end_position_on_the_genomic_accession=111320539, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  338004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338004, CurrentID=0, Status='live', Symbol='LINC00226', Aliases='C14orf97, NCRNA00226', description='long intergenic non-protein coding RNA 226', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106287674, end_position_on_the_genomic_accession=106288372, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374676, CurrentID=0, Status='live', Symbol='LOC105374676', Aliases='', description='uncharacterized LOC105374676', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20611817, end_position_on_the_genomic_accession=20937711, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  119559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119559, CurrentID=0, Status='live', Symbol='SFXN4', Aliases='BCRM1, COXPD18', description='sideroflexin 4', other_designations='breast cancer resistance marker 1', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119140913, end_position_on_the_genomic_accession=119165709, orientation='minus', exon_count=16, OMIM=615564, no_hdr0=''),
  105370379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370379, CurrentID=0, Status='live', Symbol='LOC105370379', Aliases='', description='uncharacterized LOC105370379', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  163255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163255, CurrentID=0, Status='live', Symbol='ZNF540', Aliases='Nbla10512', description='zinc finger protein 540', other_designations='CTD-3064H18.6|putative protein product of Nbla10512', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37551371, end_position_on_the_genomic_accession=37614179, orientation='plus', exon_count=6, OMIM=613903, no_hdr0=''),
  105370381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370381, CurrentID=0, Status='live', Symbol='LOC105370381', Aliases='', description='uncharacterized LOC105370381', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370382, CurrentID=0, Status='live', Symbol='LOC105370382', Aliases='', description='uncharacterized LOC105370382', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114111299, end_position_on_the_genomic_accession=114116622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369213, CurrentID=0, Status='live', Symbol='LOC105369213', Aliases='', description='uncharacterized LOC105369213', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81739031, end_position_on_the_genomic_accession=81777318, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100302267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100302267, CurrentID=0, Status='live', Symbol='MIR2054', Aliases='hsa-mir-2054', description='microRNA 2054', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=125507259, end_position_on_the_genomic_accession=125507307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645485, CurrentID=0, Status='live', Symbol='LOC645485', Aliases='', description='uncharacterized LOC645485', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30755029, end_position_on_the_genomic_accession=30780739, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370386, CurrentID=0, Status='live', Symbol='LOC105370386', Aliases='', description='uncharacterized LOC105370386', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18590021, end_position_on_the_genomic_accession=18616881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643859, CurrentID=0, Status='live', Symbol='FEM1AP3', Aliases='', description='fem-1 homolog a (C. elegans) pseudogene 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112365707, end_position_on_the_genomic_accession=112367444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873423, CurrentID=0, Status='live', Symbol='RNA5SP157', Aliases='RN5S157', description='RNA, 5S ribosomal pseudogene 157', other_designations='RNA, 5S ribosomal 157', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=19181814, end_position_on_the_genomic_accession=19181856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370390, CurrentID=0, Status='live', Symbol='LOC105370390', Aliases='', description='uncharacterized LOC105370390', other_designations='CTD-2311B13.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18920139, end_position_on_the_genomic_accession=18946272, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  285761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285761, CurrentID=0, Status='live', Symbol='DCBLD1', Aliases='dJ94G16.1', description='discoidin, CUB and LCCL domain containing 1', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117453819, end_position_on_the_genomic_accession=117569858, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105370391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370391, CurrentID=0, Status='live', Symbol='LOC105370391', Aliases='', description='uncharacterized LOC105370391', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370393, CurrentID=0, Status='live', Symbol='LOC105370393', Aliases='', description='uncharacterized LOC105370393', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19902461, end_position_on_the_genomic_accession=19948734, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370394, CurrentID=0, Status='live', Symbol='LOC105370394', Aliases='', description='uncharacterized LOC105370394', other_designations='CTD-2292M16.7', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20305887, end_position_on_the_genomic_accession=20306811, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377579, CurrentID=0, Status='live', Symbol='LOC105377579', Aliases='', description='uncharacterized LOC105377579', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183559182, end_position_on_the_genomic_accession=183564692, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100463285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463285, CurrentID=0, Status='live', Symbol='MTRNR2L4', Aliases='HN4', description='MT-RNR2-like 4', other_designations='MTRNR2-like 4|humanin-like protein 4', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3371053, end_position_on_the_genomic_accession=3372283, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100463286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463286, CurrentID=0, Status='live', Symbol='CCT6P5', Aliases='CCT6-2P', description='chaperonin containing TCP1, subunit 6 (zeta) pseudogene 5', other_designations='chaperonin containing TCP1, subunit 6A (zeta 1) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109637242, end_position_on_the_genomic_accession=109638102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100463288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463288, CurrentID=0, Status='live', Symbol='HSPD1P19', Aliases='HSPD1-18P', description='heat shock protein family D (Hsp60) member 1 pseudogene 19', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 18|heat shock 60kDa protein 1 (chaperonin) pseudogene 19', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58097222, end_position_on_the_genomic_accession=58100274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100463289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463289, CurrentID=0, Status='live', Symbol='MTRNR2L5', Aliases='HN5', description='MT-RNR2-like 5', other_designations='MTRNR2-like 5|humanin-like protein 5', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=55598990, end_position_on_the_genomic_accession=55600728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128529, CurrentID=0, Status='live', Symbol='LOC100128529', Aliases='', description='uncharacterized LOC100128529', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=51401874, end_position_on_the_genomic_accession=51402314, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130587, CurrentID=0, Status='live', Symbol='LOC100130587', Aliases='', description='uncharacterized LOC100130587', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63359988, end_position_on_the_genomic_accession=63371177, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370400, CurrentID=0, Status='live', Symbol='LOC105370400', Aliases='', description='uncharacterized LOC105370400', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22556311, end_position_on_the_genomic_accession=22557062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370401, CurrentID=0, Status='live', Symbol='LOC105370401', Aliases='', description='uncharacterized LOC105370401', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22461268, end_position_on_the_genomic_accession=22484675, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  643875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643875, CurrentID=0, Status='live', Symbol='CNN3P1', Aliases='', description='calponin 3, acidic pseudogene 1', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6534697, end_position_on_the_genomic_accession=6535628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927012, CurrentID=0, Status='live', Symbol='LOC101927012', Aliases='', description='uncharacterized LOC101927012', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147666598, end_position_on_the_genomic_accession=147694809, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101060224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060224, CurrentID=0, Status='live', Symbol='LOC101060224', Aliases='', description='uncharacterized LOC101060224', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50081831, end_position_on_the_genomic_accession=50088477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373215, CurrentID=0, Status='live', Symbol='LOC105373215', Aliases='', description='uncharacterized LOC105373215', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235892284, end_position_on_the_genomic_accession=235957995, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105370408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370408, CurrentID=0, Status='live', Symbol='LOC105370408', Aliases='', description='uncharacterized LOC105370408', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23692767, end_position_on_the_genomic_accession=23698934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374276, CurrentID=0, Status='live', Symbol='LOC105374276', Aliases='', description='uncharacterized LOC105374276', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191781109, end_position_on_the_genomic_accession=191808573, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370412, CurrentID=0, Status='live', Symbol='LOC105370412', Aliases='', description='uncharacterized LOC105370412', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24066784, end_position_on_the_genomic_accession=24068184, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643888, CurrentID=0, Status='live', Symbol='LOC643888', Aliases='', description='coiled-coil domain containing 121 pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112380781, end_position_on_the_genomic_accession=112383556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  731275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731275, CurrentID=0, Status='live', Symbol='LINC01347', Aliases='', description='long intergenic non-protein coding RNA 1347', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243056314, end_position_on_the_genomic_accession=243101744, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100422136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422136, CurrentID=0, Status='live', Symbol='LOC100422136', Aliases='', description='olfactory receptor, family 4, subfamily A, member 16 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370424, CurrentID=0, Status='live', Symbol='LOC105370424', Aliases='', description='uncharacterized LOC105370424', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28783491, end_position_on_the_genomic_accession=28785798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  285768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285768, CurrentID=0, Status='live', Symbol='LINC01622', Aliases='', description='long intergenic non-protein coding RNA 1622', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=961003, end_position_on_the_genomic_accession=1101332, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370426, CurrentID=0, Status='live', Symbol='LOC105370426', Aliases='', description='uncharacterized LOC105370426', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927013, CurrentID=0, Status='live', Symbol='LOC101927013', Aliases='', description='uncharacterized LOC101927013', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93746023, end_position_on_the_genomic_accession=93757715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100463482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463482, CurrentID=0, Status='live', Symbol='MTRNR2L6', Aliases='HN6', description='MT-RNR2-like 6', other_designations='MTRNR2-like 6|humanin-like protein 6', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142666299, end_position_on_the_genomic_accession=142667693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100463486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463486, CurrentID=0, Status='live', Symbol='MTRNR2L8', Aliases='HN8', description='MT-RNR2-like 8', other_designations='MTRNR2-like 8|humanin-like protein 8', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10507887, end_position_on_the_genomic_accession=10509176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100463487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463487, CurrentID=0, Status='live', Symbol='MTRNR2L9', Aliases='HN9', description='MT-RNR2-like 9', other_designations='MT-RNR2-like 9 (pseudogene)|MTRNR2-like 9|humanin-like protein 9', map_location='6q11.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=61574103, end_position_on_the_genomic_accession=61574629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100463488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463488, CurrentID=0, Status='live', Symbol='MTRNR2L10', Aliases='HN10', description='MT-RNR2-like 10', other_designations='MTRNR2-like 10|humanin-like protein 10', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55181391, end_position_on_the_genomic_accession=55182511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100463489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463489, CurrentID=0, Status='live', Symbol='MTRNR2L11', Aliases='HN11', description='MT-RNR2-like 11 (pseudogene)', other_designations='MTRNR2-like 11', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237943724, end_position_on_the_genomic_accession=237945213, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370433, CurrentID=0, Status='live', Symbol='LOC105370433', Aliases='', description='uncharacterized LOC105370433', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30472438, end_position_on_the_genomic_accession=30485239, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100463498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463498, CurrentID=0, Status='live', Symbol='MTRNR2L12', Aliases='HN12', description='MT-RNR2-like 12', other_designations='MT-RNR2-like 12 (pseudogene)|MTRNR2-like 12|humanin-like 12', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96617186, end_position_on_the_genomic_accession=96618223, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373216, CurrentID=0, Status='live', Symbol='LOC105373216', Aliases='', description='uncharacterized LOC105373216', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55908180, end_position_on_the_genomic_accession=56205098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100463500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463500, CurrentID=0, Status='live', Symbol='MTRNR2L13', Aliases='HN13', description='MT-RNR2-like 13', other_designations='MT-RNR2-like 13 (pseudogene)|MTRNR2-like 13|humanin-like 13', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116298860, end_position_on_the_genomic_accession=116300322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370434, CurrentID=0, Status='live', Symbol='LOC105370434', Aliases='', description='uncharacterized LOC105370434', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30559279, end_position_on_the_genomic_accession=30577650, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370437, CurrentID=0, Status='live', Symbol='LOC105370437', Aliases='', description='uncharacterized LOC105370437', other_designations='CTD-2213F21.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31553355, end_position_on_the_genomic_accession=31561331, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100463528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100463528, CurrentID=0, Status='live', Symbol='LSP1P1', Aliases='', description='lymphocyte-specific protein 1 pseudogene 1', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25017402, end_position_on_the_genomic_accession=25017538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370439, CurrentID=0, Status='live', Symbol='LOC105370439', Aliases='', description='uncharacterized LOC105370439', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31943455, end_position_on_the_genomic_accession=31960678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370443, CurrentID=0, Status='live', Symbol='LOC105370443', Aliases='', description='uncharacterized LOC105370443', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=33221916, end_position_on_the_genomic_accession=33227649, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643916, CurrentID=0, Status='live', Symbol='LOC643916', Aliases='', description='RAB1A, member RAS oncogene family pseudogene', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86170444, end_position_on_the_genomic_accession=86171036, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370445, CurrentID=0, Status='live', Symbol='LOC105370445', Aliases='', description='uncharacterized LOC105370445', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34039984, end_position_on_the_genomic_accession=34060069, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100856791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856791, CurrentID=0, Status='live', Symbol='LOC100856791', Aliases='', description='TBP-like 1 pseudogene', other_designations='', map_location='3q26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172522228, end_position_on_the_genomic_accession=172522901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374692, CurrentID=0, Status='live', Symbol='LOC105374692', Aliases='', description='uncharacterized LOC105374692', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24877924, end_position_on_the_genomic_accession=24880474, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100856809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856809, CurrentID=0, Status='live', Symbol='OR3A5P', Aliases='OLFRA06, OR17-25, OR25', description='olfactory receptor family 3 subfamily A member 5 pseudogene', other_designations='RP11-64J4.6-001|olfactory receptor, family 3, subfamily A pseudogene|olfactory receptor, family 3, subfamily A, member 5 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3311200, end_position_on_the_genomic_accession=3311892, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100856810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856810, CurrentID=0, Status='live', Symbol='TCEB2P3', Aliases='', description='transcription elongation factor B (SIII), polypeptide 2 (18kDa, elongin B) pseudogene 3', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48780602, end_position_on_the_genomic_accession=48780897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100856811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856811, CurrentID=0, Status='live', Symbol='NRG1-IT1', Aliases='', description='NRG1 intronic transcript 1', other_designations='NRG1 intronic transcript 1 (non-protein coding)', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=32025736, end_position_on_the_genomic_accession=32139477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100294391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100294391, CurrentID=0, Status='live', Symbol='LOC100294391', Aliases='', description='uncharacterized LOC100294391', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643923, CurrentID=0, Status='live', Symbol='LOC643923', Aliases='', description='uncharacterized LOC643923', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107591745, end_position_on_the_genomic_accession=107593223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84214, CurrentID=0, Status='live', Symbol='LOC84214', Aliases='', description='uncharacterized LOC84214', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65837860, end_position_on_the_genomic_accession=65842398, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106478917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478917, CurrentID=0, Status='live', Symbol='RNU6-22P', Aliases='RNU6-22', description='RNA, U6 small nuclear 22, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372356, CurrentID=0, Status='live', Symbol='LOC105372356', Aliases='', description='uncharacterized LOC105372356', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31125853, end_position_on_the_genomic_accession=31139925, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374694, CurrentID=0, Status='live', Symbol='LOC105374694', Aliases='', description='uncharacterized LOC105374694', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=25088082, end_position_on_the_genomic_accession=25190522, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106478918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478918, CurrentID=0, Status='live', Symbol='RNU6-27P', Aliases='RNU6-27', description='RNA, U6 small nuclear 27, pseudogene', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370460, CurrentID=0, Status='live', Symbol='LOC105370460', Aliases='', description='uncharacterized LOC105370460', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39474847, end_position_on_the_genomic_accession=39480342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100856877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856877, CurrentID=0, Status='live', Symbol='SETP15', Aliases='', description='SET pseudogene 15', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15917548, end_position_on_the_genomic_accession=15917846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100856878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856878, CurrentID=0, Status='live', Symbol='SETP21', Aliases='', description='SET pseudogene 21', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60459852, end_position_on_the_genomic_accession=60460739, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100856881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100856881, CurrentID=0, Status='live', Symbol='MTND1P32', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 32', other_designations='MT-ND1 pseudogene 32', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102785678, end_position_on_the_genomic_accession=102786604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643936, CurrentID=0, Status='live', Symbol='LOC643936', Aliases='', description='uncharacterized LOC643936', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=7052562, end_position_on_the_genomic_accession=7053218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370331, CurrentID=0, Status='live', Symbol='LOC105370331', Aliases='', description='uncharacterized LOC105370331', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99281276, end_position_on_the_genomic_accession=99283698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  193629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=193629, CurrentID=0, Status='live', Symbol='LINC00189', Aliases='C21orf109, NCRNA00189', description='long intergenic non-protein coding RNA 189', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29193494, end_position_on_the_genomic_accession=29288205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370469, CurrentID=0, Status='live', Symbol='LOC105370469', Aliases='', description='uncharacterized LOC105370469', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=41588081, end_position_on_the_genomic_accession=41607111, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370475, CurrentID=0, Status='live', Symbol='LOC105370475', Aliases='', description='uncharacterized LOC105370475', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45369151, end_position_on_the_genomic_accession=45389286, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  150478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150478, CurrentID=0, Status='live', Symbol='bA395L14.12', Aliases='', description="small nuclear ribonucleoprotein polypeptide A' pseudogene", other_designations="novel protein similar to small nuclear ribonucleoprotein polypeptide A' SNRPA1|u2 small nuclear ribonucleoprotein polypeptide A' pseudogene", map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113657873, end_position_on_the_genomic_accession=113658892, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130745, CurrentID=0, Status='live', Symbol='LOC100130745', Aliases='', description='uncharacterized LOC100130745', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=1473164, end_position_on_the_genomic_accession=1473541, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370477, CurrentID=0, Status='live', Symbol='LOC105370477', Aliases='', description='uncharacterized LOC105370477', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45622643, end_position_on_the_genomic_accession=45626357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132910, CurrentID=0, Status='live', Symbol='LOC100132910', Aliases='', description='PRO1477', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=35168764, end_position_on_the_genomic_accession=35169665, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370482, CurrentID=0, Status='live', Symbol='LOC105370482', Aliases='', description='uncharacterized LOC105370482', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48234199, end_position_on_the_genomic_accession=48253153, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370483, CurrentID=0, Status='live', Symbol='LOC105370483', Aliases='', description='uncharacterized LOC105370483', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48400411, end_position_on_the_genomic_accession=48429248, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128546, CurrentID=0, Status='live', Symbol='LOC100128546', Aliases='', description='glutaredoxin 5 pseudogene', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120732920, end_position_on_the_genomic_accession=120733859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370488, CurrentID=0, Status='live', Symbol='LOC105370488', Aliases='', description='uncharacterized LOC105370488', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50668447, end_position_on_the_genomic_accession=50670138, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643962, CurrentID=0, Status='live', Symbol='LOC643962', Aliases='', description='uncharacterized LOC643962', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86728142, end_position_on_the_genomic_accession=86937522, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370493, CurrentID=0, Status='live', Symbol='LOC105370493', Aliases='', description='uncharacterized LOC105370493', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51089237, end_position_on_the_genomic_accession=51091288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370494, CurrentID=0, Status='live', Symbol='LOC105370494', Aliases='', description='uncharacterized LOC105370494', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51130106, end_position_on_the_genomic_accession=51198011, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106478925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478925, CurrentID=0, Status='live', Symbol='KRT18P53', Aliases='', description='keratin 18 pseudogene 53', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128548, CurrentID=0, Status='live', Symbol='LOC100128548', Aliases='', description='ribosomal protein L17 pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94891201, end_position_on_the_genomic_accession=94892143, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370495, CurrentID=0, Status='live', Symbol='LOC105370495', Aliases='', description='uncharacterized LOC105370495', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51333399, end_position_on_the_genomic_accession=51348521, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370498, CurrentID=0, Status='live', Symbol='LOC105370498', Aliases='', description='uncharacterized LOC105370498', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52111174, end_position_on_the_genomic_accession=52126619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422139, CurrentID=0, Status='live', Symbol='LOC100422139', Aliases='', description='olfactory receptor, family 5, subfamily B, member 12 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370499, CurrentID=0, Status='live', Symbol='LOC105370499', Aliases='', description='uncharacterized LOC105370499', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52415825, end_position_on_the_genomic_accession=52430046, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370500, CurrentID=0, Status='live', Symbol='LOC105370500', Aliases='', description='uncharacterized LOC105370500', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52791803, end_position_on_the_genomic_accession=52930234, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370502, CurrentID=0, Status='live', Symbol='LOC105370502', Aliases='', description='uncharacterized LOC105370502', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53168872, end_position_on_the_genomic_accession=53187311, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370503, CurrentID=0, Status='live', Symbol='LOC105370503', Aliases='', description='uncharacterized LOC105370503', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53685007, end_position_on_the_genomic_accession=53850837, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927016, CurrentID=0, Status='live', Symbol='LOC101927016', Aliases='', description='keratin-associated protein 21-1-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63832066, end_position_on_the_genomic_accession=63833720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370504, CurrentID=0, Status='live', Symbol='LOC105370504', Aliases='', description='uncharacterized LOC105370504', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53369449, end_position_on_the_genomic_accession=53883349, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105370505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370505, CurrentID=0, Status='live', Symbol='LOC105370505', Aliases='', description='uncharacterized LOC105370505', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54114815, end_position_on_the_genomic_accession=54116076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370508, CurrentID=0, Status='live', Symbol='LOC105370508', Aliases='', description='uncharacterized LOC105370508', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55191726, end_position_on_the_genomic_accession=55196037, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101060064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060064, CurrentID=0, Status='live', Symbol='LOC101060064', Aliases='', description='Fanconi anemia-associated protein of 24 kDa pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8678677, end_position_on_the_genomic_accession=8679917, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370509, CurrentID=0, Status='live', Symbol='LOC105370509', Aliases='', description='uncharacterized LOC105370509', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55262231, end_position_on_the_genomic_accession=55272091, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128551, CurrentID=0, Status='live', Symbol='LOC100128551', Aliases='', description='uncharacterized LOC100128551', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=157990592, end_position_on_the_genomic_accession=157991232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370510, CurrentID=0, Status='live', Symbol='LOC105370510', Aliases='', description='uncharacterized LOC105370510', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55709812, end_position_on_the_genomic_accession=55715100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  285785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285785, CurrentID=0, Status='live', Symbol='RPL39P3', Aliases='RPL39_9_712', description='ribosomal protein L39 pseudogene 3', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73372930, end_position_on_the_genomic_accession=73373313, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370511, CurrentID=0, Status='live', Symbol='LOC105370511', Aliases='', description='uncharacterized LOC105370511', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55899411, end_position_on_the_genomic_accession=55900801, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370512, CurrentID=0, Status='live', Symbol='LOC105370512', Aliases='', description='uncharacterized LOC105370512', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56005324, end_position_on_the_genomic_accession=56046068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10821, CurrentID=0, Status='live', Symbol='OR7E12P', Aliases='HG1058, OR11-3, OR7E58P, OR7E79P, OR7E81P, OST021', description='olfactory receptor family 7 subfamily E member 12 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 12 pseudogene|olfactory receptor, family 7, subfamily E, member 58 pseudogene|olfactory receptor, family 7, subfamily E, member 79 pseudogene|olfactory receptor, family 7, subfamily E, member 81 pseudogene', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3390707, end_position_on_the_genomic_accession=3391915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370515, CurrentID=0, Status='live', Symbol='LOC105370515', Aliases='', description='uncharacterized LOC105370515', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56633249, end_position_on_the_genomic_accession=56639338, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370516, CurrentID=0, Status='live', Symbol='LOC105370516', Aliases='', description='uncharacterized LOC105370516', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57320608, end_position_on_the_genomic_accession=57344016, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370520, CurrentID=0, Status='live', Symbol='LOC105370520', Aliases='', description='uncharacterized LOC105370520', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57550477, end_position_on_the_genomic_accession=57569888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370524, CurrentID=0, Status='live', Symbol='LOC105370524', Aliases='', description='uncharacterized LOC105370524', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59969150, end_position_on_the_genomic_accession=60091780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  643997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643997, CurrentID=0, Status='live', Symbol='LOC643997', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173485659, end_position_on_the_genomic_accession=173486405, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370526, CurrentID=0, Status='live', Symbol='LOC105370526', Aliases='', description='uncharacterized LOC105370526', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61708704, end_position_on_the_genomic_accession=61751097, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130602, CurrentID=0, Status='live', Symbol='LOC100130602', Aliases='', description='regulation of nuclear pre-mRNA domain containing 1A pseudogene', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50046458, end_position_on_the_genomic_accession=50047862, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370528, CurrentID=0, Status='live', Symbol='LOC105370528', Aliases='', description='uncharacterized LOC105370528', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62487476, end_position_on_the_genomic_accession=62492624, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644006, CurrentID=0, Status='live', Symbol='LOC644006', Aliases='', description='ring finger protein 4 pseudogene', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231021373, end_position_on_the_genomic_accession=231024205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379074, CurrentID=0, Status='live', Symbol='LOC105379074', Aliases='', description='uncharacterized LOC105379074', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89362922, end_position_on_the_genomic_accession=89394360, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370538, CurrentID=0, Status='live', Symbol='LOC105370538', Aliases='', description='uncharacterized LOC105370538', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67072420, end_position_on_the_genomic_accession=67189354, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379075, CurrentID=0, Status='live', Symbol='LOC105379075', Aliases='', description='uncharacterized LOC105379075', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89349253, end_position_on_the_genomic_accession=89360723, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370541, CurrentID=0, Status='live', Symbol='LOC105370541', Aliases='', description='uncharacterized LOC105370541', other_designations='5SN4 snoRNA', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67259863, end_position_on_the_genomic_accession=67337302, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370544, CurrentID=0, Status='live', Symbol='LOC105370544', Aliases='', description='uncharacterized LOC105370544', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68652592, end_position_on_the_genomic_accession=68662833, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106478935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478935, CurrentID=0, Status='live', Symbol='HSFY7P', Aliases='', description='heat shock transcription factor, Y-linked 7, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370545, CurrentID=0, Status='live', Symbol='LOC105370545', Aliases='', description='uncharacterized LOC105370545', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68559610, end_position_on_the_genomic_accession=68565235, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  63027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63027, CurrentID=0, Status='live', Symbol='SLC22A23', Aliases='C6orf85', description='solute carrier family 22 member 23', other_designations='ion transporter protein|solute carrier family 22, member 23', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3268973, end_position_on_the_genomic_accession=3456587, orientation='minus', exon_count=16, OMIM=611697, no_hdr0=''),
  105370548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370548, CurrentID=0, Status='live', Symbol='LOC105370548', Aliases='', description='atherin', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68801294, end_position_on_the_genomic_accession=68828141, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  63035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63035, CurrentID=0, Status='live', Symbol='BCORL1', Aliases='BCoR-L1, CXorf10', description='BCL6 corepressor-like 1', other_designations='BCoR-like protein 1', map_location='Xq25-q26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129980283, end_position_on_the_genomic_accession=130058083, orientation='plus', exon_count=20, OMIM=300688, no_hdr0=''),
  63036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63036, CurrentID=0, Status='live', Symbol='CELA2A', Aliases='ELA2A, PE-1', description='chymotrypsin like elastase family member 2A', other_designations='chymotrypsin-like elastase family, member 2A|elastase 2A|elastase-2A|pancreatic elastase 2|pancreatic elastase IIA', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15456728, end_position_on_the_genomic_accession=15472091, orientation='plus', exon_count=8, OMIM=609443, no_hdr0=''),
  105370549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370549, CurrentID=0, Status='live', Symbol='LOC105370549', Aliases='', description='uncharacterized LOC105370549', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69297674, end_position_on_the_genomic_accession=69300959, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370550, CurrentID=0, Status='live', Symbol='LOC105370550', Aliases='', description='uncharacterized LOC105370550', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69183018, end_position_on_the_genomic_accession=69260467, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927018, CurrentID=0, Status='live', Symbol='LOC101927018', Aliases='', description='uncharacterized LOC101927018', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28926275, end_position_on_the_genomic_accession=28944749, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106478937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478937, CurrentID=0, Status='live', Symbol='TPM3P3', Aliases='', description='tropomyosin 3 pseudogene 3', other_designations='', map_location='8q21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370555, CurrentID=0, Status='live', Symbol='LOC105370555', Aliases='', description='uncharacterized LOC105370555', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70417195, end_position_on_the_genomic_accession=70460964, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370556, CurrentID=0, Status='live', Symbol='LOC105370556', Aliases='', description='uncharacterized LOC105370556', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70525556, end_position_on_the_genomic_accession=70538883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370557, CurrentID=0, Status='live', Symbol='LOC105370557', Aliases='', description='uncharacterized LOC105370557', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70932064, end_position_on_the_genomic_accession=70965423, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370558, CurrentID=0, Status='live', Symbol='LOC105370558', Aliases='', description='uncharacterized LOC105370558', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71867397, end_position_on_the_genomic_accession=71910831, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370560, CurrentID=0, Status='live', Symbol='LOC105370560', Aliases='', description='uncharacterized LOC105370560', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73027700, end_position_on_the_genomic_accession=73060061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370563, CurrentID=0, Status='live', Symbol='LOC105370563', Aliases='', description='uncharacterized LOC105370563', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74096495, end_position_on_the_genomic_accession=74140657, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644037, CurrentID=0, Status='live', Symbol='HNRNPA1P12', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 12', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80358880, end_position_on_the_genomic_accession=80360099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370566, CurrentID=0, Status='live', Symbol='LOC105370566', Aliases='', description='uncharacterized LOC105370566', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74513799, end_position_on_the_genomic_accession=74519450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370569, CurrentID=0, Status='live', Symbol='LOC105370569', Aliases='', description='uncharacterized LOC105370569', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74839303, end_position_on_the_genomic_accession=74842149, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128563, CurrentID=0, Status='live', Symbol='LOC100128563', Aliases='', description='uncharacterized LOC100128563', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=240269215, end_position_on_the_genomic_accession=240270842, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370570, CurrentID=0, Status='live', Symbol='LOC105370570', Aliases='', description='uncharacterized LOC105370570', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644050, CurrentID=0, Status='live', Symbol='LINC01529', Aliases='', description='long intergenic non-protein coding RNA 1529', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35788876, end_position_on_the_genomic_accession=35797909, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927019, CurrentID=0, Status='live', Symbol='LOC101927019', Aliases='', description='uncharacterized LOC101927019', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39315932, end_position_on_the_genomic_accession=39329491, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644051, CurrentID=0, Status='live', Symbol='LOC644051', Aliases='', description='isocitrate dehydrogenase pseudogene', other_designations='isocitrate dehydrogenase 1 (NADP+), soluble pseudogene', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7516094, end_position_on_the_genomic_accession=7517878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370580, CurrentID=0, Status='live', Symbol='LOC105370580', Aliases='', description='uncharacterized LOC105370580', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77081149, end_position_on_the_genomic_accession=77086696, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370581, CurrentID=0, Status='live', Symbol='LOC105370581', Aliases='', description='uncharacterized LOC105370581', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77966162, end_position_on_the_genomic_accession=77968049, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370582, CurrentID=0, Status='live', Symbol='LOC105370582', Aliases='', description='uncharacterized LOC105370582', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79874221, end_position_on_the_genomic_accession=79917104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927551, CurrentID=0, Status='live', Symbol='LOC101927551', Aliases='', description='uncharacterized LOC101927551', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44997734, end_position_on_the_genomic_accession=45003619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373222, CurrentID=0, Status='live', Symbol='LOC105373222', Aliases='', description='uncharacterized LOC105373222', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238836446, end_position_on_the_genomic_accession=238848059, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106478943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478943, CurrentID=0, Status='live', Symbol='MTND6P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 4', other_designations='MT-ND6 pseudogene 4', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370586, CurrentID=0, Status='live', Symbol='LOC105370586', Aliases='', description='uncharacterized LOC105370586', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79199982, end_position_on_the_genomic_accession=79201636, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370587, CurrentID=0, Status='live', Symbol='LOC105370587', Aliases='', description='uncharacterized LOC105370587', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79154930, end_position_on_the_genomic_accession=79197864, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105374720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374720, CurrentID=0, Status='live', Symbol='LOC105374720', Aliases='', description='uncharacterized LOC105374720', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34329894, end_position_on_the_genomic_accession=34465090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370588, CurrentID=0, Status='live', Symbol='LOC105370588', Aliases='', description='uncharacterized LOC105370588', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79077160, end_position_on_the_genomic_accession=79091590, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370590, CurrentID=0, Status='live', Symbol='LOC105370590', Aliases='', description='uncharacterized LOC105370590', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79972368, end_position_on_the_genomic_accession=79987108, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  119775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119775, CurrentID=0, Status='live', Symbol='OR52P2P', Aliases='', description='olfactory receptor family 52 subfamily P member 2 pseudogene', other_designations='olfactory receptor, family 52, subfamily P, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4431332, end_position_on_the_genomic_accession=4432228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128568, CurrentID=0, Status='live', Symbol='LOC100128568', Aliases='', description='uncharacterized LOC100128568', other_designations='CTC-232P5.1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5978403, end_position_on_the_genomic_accession=6020363, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370596, CurrentID=0, Status='live', Symbol='LOC105370596', Aliases='', description='uncharacterized LOC105370596', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81013558, end_position_on_the_genomic_accession=81025299, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375325, CurrentID=0, Status='live', Symbol='LOC105375325', Aliases='', description='uncharacterized LOC105375325', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64600306, end_position_on_the_genomic_accession=64604054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370598, CurrentID=0, Status='live', Symbol='LOC105370598', Aliases='', description='uncharacterized LOC105370598', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81811520, end_position_on_the_genomic_accession=81820061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370604, CurrentID=0, Status='live', Symbol='LOC105370604', Aliases='', description='uncharacterized LOC105370604', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85148645, end_position_on_the_genomic_accession=85194702, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79518, CurrentID=0, Status='live', Symbol='OR5M4P', Aliases='', description='olfactory receptor family 5 subfamily M member 4 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56448669, end_position_on_the_genomic_accession=56449593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370607, CurrentID=0, Status='live', Symbol='LOC105370607', Aliases='', description='uncharacterized LOC105370607', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=86331570, end_position_on_the_genomic_accession=86357640, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370609, CurrentID=0, Status='live', Symbol='LOC105370609', Aliases='', description='uncharacterized LOC105370609', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=87164530, end_position_on_the_genomic_accession=87180479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370610, CurrentID=0, Status='live', Symbol='LOC105370610', Aliases='', description='uncharacterized LOC105370610', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=87345525, end_position_on_the_genomic_accession=87613463, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105370611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370611, CurrentID=0, Status='live', Symbol='LOC105370611', Aliases='', description='uncharacterized LOC105370611', other_designations='CTD-2007A10.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=87633875, end_position_on_the_genomic_accession=87655294, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370613, CurrentID=0, Status='live', Symbol='LOC105370613', Aliases='', description='uncharacterized LOC105370613', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88156117, end_position_on_the_genomic_accession=88161181, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  731467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731467, CurrentID=0, Status='live', Symbol='LOC731467', Aliases='', description='uncharacterized LOC731467', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106478949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478949, CurrentID=0, Status='live', Symbol='MTND6P15', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 15', other_designations='MT-ND6 pseudogene 15', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644090, CurrentID=0, Status='live', Symbol='LOC644090', Aliases='', description='uncharacterized LOC644090', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151806141, end_position_on_the_genomic_accession=151810820, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370621, CurrentID=0, Status='live', Symbol='LOC105370621', Aliases='', description='uncharacterized LOC105370621', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90812031, end_position_on_the_genomic_accession=90815908, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370622, CurrentID=0, Status='live', Symbol='LOC105370622', Aliases='', description='uncharacterized LOC105370622', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90697358, end_position_on_the_genomic_accession=90699374, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370624, CurrentID=0, Status='live', Symbol='LOC105370624', Aliases='', description='uncharacterized LOC105370624', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91378640, end_position_on_the_genomic_accession=91400786, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370625, CurrentID=0, Status='live', Symbol='LOC105370625', Aliases='', description='uncharacterized LOC105370625', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91418209, end_position_on_the_genomic_accession=91437633, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370628, CurrentID=0, Status='live', Symbol='LOC105370628', Aliases='', description='uncharacterized LOC105370628', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92514493, end_position_on_the_genomic_accession=92515507, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370629, CurrentID=0, Status='live', Symbol='LOC105370629', Aliases='', description='uncharacterized LOC105370629', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92890509, end_position_on_the_genomic_accession=92893792, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370364, CurrentID=0, Status='live', Symbol='LOC105370364', Aliases='', description='uncharacterized LOC105370364', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110945456, end_position_on_the_genomic_accession=110971654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370631, CurrentID=0, Status='live', Symbol='LOC105370631', Aliases='', description='uncharacterized LOC105370631', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93880426, end_position_on_the_genomic_accession=93885404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374097, CurrentID=0, Status='live', Symbol='LOC105374097', Aliases='', description='uncharacterized LOC105374097', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128202684, end_position_on_the_genomic_accession=128220190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  284988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284988, CurrentID=0, Status='live', Symbol='RHOQP3', Aliases='', description='ras homolog family member Q pseudogene 3', other_designations='ras homolog gene family, member Q pseudogene 3', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130210879, end_position_on_the_genomic_accession=130213809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370637, CurrentID=0, Status='live', Symbol='LOC105370637', Aliases='', description='uncharacterized LOC105370637', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95326169, end_position_on_the_genomic_accession=95328014, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644110, CurrentID=0, Status='live', Symbol='LOC644110', Aliases='', description='hyperpolarization activated cyclic nucleotide gated potassium channel 2 pseudogene', other_designations='hyperpolarization activated cyclic nucleotide-gated potassium channel 2 pseudogene|hyperpolarization activated cyclic nucleotide-gated potassium channel pseudogene', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32312445, end_position_on_the_genomic_accession=32315620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374970, CurrentID=0, Status='live', Symbol='LOC105374970', Aliases='', description='uncharacterized LOC105374970', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21825749, end_position_on_the_genomic_accession=21828652, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132942, CurrentID=0, Status='live', Symbol='LOC100132942', Aliases='', description='uncharacterized LOC100132942', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=14912036, end_position_on_the_genomic_accession=14912876, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370646, CurrentID=0, Status='live', Symbol='LOC105370646', Aliases='', description='uncharacterized LOC105370646', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370647, CurrentID=0, Status='live', Symbol='LOC105370647', Aliases='', description='uncharacterized LOC105370647', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97027717, end_position_on_the_genomic_accession=97049357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130597, CurrentID=0, Status='live', Symbol='LOC100130597', Aliases='', description='uncharacterized LOC100130597', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.7', start_position_on_the_genomic_accession=45779264, end_position_on_the_genomic_accession=45780055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  401957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401957, CurrentID=0, Status='live', Symbol='LOC401957', Aliases='', description='matrin 3 pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105891562, end_position_on_the_genomic_accession=105893519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370655, CurrentID=0, Status='live', Symbol='LOC105370655', Aliases='', description='uncharacterized LOC105370655', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98079600, end_position_on_the_genomic_accession=98167999, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370656, CurrentID=0, Status='live', Symbol='LOC105370656', Aliases='', description='uncharacterized LOC105370656', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98214005, end_position_on_the_genomic_accession=98242774, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372352, CurrentID=0, Status='live', Symbol='LOC105372352', Aliases='CTB-32O4.2, CTB-32O4.3, CTB-32O4.4', description='uncharacterized LOC105372352', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29213235, end_position_on_the_genomic_accession=29223137, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370657, CurrentID=0, Status='live', Symbol='LOC105370657', Aliases='', description='uncharacterized LOC105370657', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98546279, end_position_on_the_genomic_accession=98550388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370659, CurrentID=0, Status='live', Symbol='LOC105370659', Aliases='', description='uncharacterized LOC105370659', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99123701, end_position_on_the_genomic_accession=99137560, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928768, CurrentID=0, Status='live', Symbol='LOC101928768', Aliases='', description='uncharacterized LOC101928768', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=784848, end_position_on_the_genomic_accession=787141, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374735, CurrentID=0, Status='live', Symbol='LOC105374735', Aliases='', description='uncharacterized LOC105374735', other_designations='CTD-2196P11.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38576091, end_position_on_the_genomic_accession=38579594, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370663, CurrentID=0, Status='live', Symbol='LOC105370663', Aliases='', description='uncharacterized LOC105370663', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100453755, end_position_on_the_genomic_accession=100455640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370666, CurrentID=0, Status='live', Symbol='LOC105370666', Aliases='', description='uncharacterized LOC105370666', other_designations='CTD-2062F14.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100539127, end_position_on_the_genomic_accession=100540488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370667, CurrentID=0, Status='live', Symbol='LOC105370667', Aliases='', description='uncharacterized LOC105370667', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100663247, end_position_on_the_genomic_accession=100664612, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370673, CurrentID=0, Status='live', Symbol='LOC105370673', Aliases='', description='uncharacterized LOC105370673', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101519517, end_position_on_the_genomic_accession=101545328, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370674, CurrentID=0, Status='live', Symbol='LOC105370674', Aliases='', description='uncharacterized LOC105370674', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101545862, end_position_on_the_genomic_accession=101547964, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  360157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360157, CurrentID=0, Status='live', Symbol='CYCSP4', Aliases='HCP4', description='cytochrome c, somatic pseudogene 4', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202369526, end_position_on_the_genomic_accession=202369802, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102465246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465246, CurrentID=0, Status='live', Symbol='MIR6499', Aliases='hsa-mir-6499, mir-6499', description='microRNA 6499', other_designations='microRNA mir-6499', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151522087, end_position_on_the_genomic_accession=151522148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  360159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360159, CurrentID=0, Status='live', Symbol='CYCSP6', Aliases='HCP6', description='cytochrome c, somatic pseudogene 6', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78412784, end_position_on_the_genomic_accession=78413094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102465248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465248, CurrentID=0, Status='live', Symbol='MIR6501', Aliases='hsa-mir-6501, mir-6501', description='microRNA 6501', other_designations='microRNA mir-6501', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33550662, end_position_on_the_genomic_accession=33550728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131471, CurrentID=0, Status='live', Symbol='LOC100131471', Aliases='', description='presenilins-associated rhomboid-like protein, mitochondrial-like', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143342239, end_position_on_the_genomic_accession=143343577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102465249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465249, CurrentID=0, Status='live', Symbol='MIR6502', Aliases='hsa-mir-6502', description='microRNA 6502', other_designations='microRNA mir-6502', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66251082, end_position_on_the_genomic_accession=66251157, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927023, CurrentID=0, Status='live', Symbol='LOC101927023', Aliases='', description='uncharacterized LOC101927023', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112228283, end_position_on_the_genomic_accession=112257309, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  63826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63826, CurrentID=0, Status='live', Symbol='SRR', Aliases='ILV1, ISO1', description='serine racemase', other_designations='D-serine ammonia-lyase|D-serine dehydratase|L-serine ammonia-lyase|L-serine dehydratase', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2303371, end_position_on_the_genomic_accession=2325264, orientation='plus', exon_count=11, OMIM=606477, no_hdr0=''),
  63827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63827, CurrentID=0, Status='live', Symbol='BCAN', Aliases='BEHAB, CSPG7', description='brevican', other_designations='brain-enriched hyaluronan-binding protein|chondroitin sulfate proteoglycan 7|chondroitin sulfate proteoglycan BEHAB', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156641948, end_position_on_the_genomic_accession=156659532, orientation='plus', exon_count=14, OMIM=600347, no_hdr0=''),
  105370681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370681, CurrentID=0, Status='live', Symbol='LOC105370681', Aliases='', description='uncharacterized LOC105370681', other_designations='CTD-2555C10.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102545229, end_position_on_the_genomic_accession=102589559, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10639, CurrentID=0, Status='live', Symbol='HHCM', Aliases='hhc(M)', description='Mahlavu hepatocellular carcinoma', other_designations='Hepatocellular carcinoma protein HHCM', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=144241549, end_position_on_the_genomic_accession=144242955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370682, CurrentID=0, Status='live', Symbol='LOC105370682', Aliases='', description='uncharacterized LOC105370682', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102770047, end_position_on_the_genomic_accession=102775040, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  360165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360165, CurrentID=0, Status='live', Symbol='NDUFA4P1', Aliases='NDUFA4', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa, pseudogene 1', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107505128, end_position_on_the_genomic_accession=107505628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360166, CurrentID=0, Status='live', Symbol='CYCSP10', Aliases='HCP10', description='cytochrome c, somatic pseudogene 10', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10000640, end_position_on_the_genomic_accession=10000947, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373226, CurrentID=0, Status='live', Symbol='LOC105373226', Aliases='', description='uncharacterized LOC105373226', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239386518, end_position_on_the_genomic_accession=239387227, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271135, CurrentID=0, Status='live', Symbol='RPS6P18', Aliases='RPS6_8_1191', description='ribosomal protein S6 pseudogene 18', other_designations='', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2659895, end_position_on_the_genomic_accession=2660546, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360167, CurrentID=0, Status='live', Symbol='CYCSP11', Aliases='HCP11', description='cytochrome c, somatic pseudogene 11', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10057662, end_position_on_the_genomic_accession=10058356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370685, CurrentID=0, Status='live', Symbol='LOC105370685', Aliases='', description='uncharacterized LOC105370685', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103164194, end_position_on_the_genomic_accession=103175421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370375, CurrentID=0, Status='live', Symbol='LOC105370375', Aliases='', description='uncharacterized LOC105370375', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112923550, end_position_on_the_genomic_accession=112929527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  360168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360168, CurrentID=0, Status='live', Symbol='CYCSP12', Aliases='CYCSP13, HCP12, HCP13', description='cytochrome c, somatic pseudogene 12', other_designations='cytochrome c, somatic pseudogene 13', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11878817, end_position_on_the_genomic_accession=11879120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644158, CurrentID=0, Status='live', Symbol='LOC644158', Aliases='', description='leucine rich repeat containing 2 pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174300000, end_position_on_the_genomic_accession=174300956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102465257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465257, CurrentID=0, Status='live', Symbol='MIR6514', Aliases='hsa-mir-6514, mir-6514', description='microRNA 6514', other_designations='microRNA mir-6514', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62792702, end_position_on_the_genomic_accession=62792771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370687, CurrentID=0, Status='live', Symbol='LOC105370687', Aliases='', description='uncharacterized LOC105370687', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63869, CurrentID=0, Status='live', Symbol='PSORS6', Aliases='', description='psoriasis susceptibility 6', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605364, no_hdr0=''),
  63870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63870, CurrentID=0, Status='live', Symbol='PSORS5', Aliases='', description='psoriasis susceptibility 5', other_designations='', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604316, no_hdr0=''),
  105370688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370688, CurrentID=0, Status='live', Symbol='LOC105370688', Aliases='', description='uncharacterized LOC105370688', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103627502, end_position_on_the_genomic_accession=103629141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  63874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63874, CurrentID=0, Status='live', Symbol='ABHD4', Aliases='ABH4', description='abhydrolase domain containing 4', other_designations='alpha/beta-hydrolase 4|lyso-N-acylphosphatidylethanolamine lipase', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22598238, end_position_on_the_genomic_accession=22612923, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  63875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63875, CurrentID=0, Status='live', Symbol='MRPL17', Aliases='L17mt, LIP2, MRP-L17, MRP-L26, RPL17L, RPML26', description='mitochondrial ribosomal protein L17', other_designations='LYST-interacting protein 2|LYST-interacting protein LIP2', map_location='11p15.5-p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6680385, end_position_on_the_genomic_accession=6683401, orientation='minus', exon_count=3, OMIM=611830, no_hdr0=''),
  63876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63876, CurrentID=0, Status='live', Symbol='PKNOX2', Aliases='PREP2', description='PBX/knotted 1 homeobox 2', other_designations='PBX/knotted homeobox 2|homeobox protein PREP-2', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125164663, end_position_on_the_genomic_accession=125433389, orientation='plus', exon_count=15, OMIM=613066, no_hdr0=''),
  63877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63877, CurrentID=0, Status='live', Symbol='FAM204A', Aliases='C10orf84, bA319I23.1', description='family with sequence similarity 204 member A', other_designations='family with sequence similarity 204, member A', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118305889, end_position_on_the_genomic_accession=118342327, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  63891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63891, CurrentID=0, Status='live', Symbol='RNF123', Aliases='FP1477, KPC1', description='ring finger protein 123', other_designations='kip1 ubiquitination-promoting complex protein 1', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49689143, end_position_on_the_genomic_accession=49721529, orientation='plus', exon_count=40, OMIM=614472, no_hdr0=''),
  63892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63892, CurrentID=0, Status='live', Symbol='THADA', Aliases='GITA', description='thyroid adenoma associated', other_designations='death receptor-interacting protein|gene inducing thyroid adenomas protein', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43230836, end_position_on_the_genomic_accession=43596056, orientation='minus', exon_count=41, OMIM=611800, no_hdr0=''),
  63893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63893, CurrentID=0, Status='live', Symbol='UBE2O', Aliases='E2-230K', description='ubiquitin conjugating enzyme E2O', other_designations='E2/E3 hybrid ubiquitin-protein ligase UBE2O|ubiquitin carrier protein O|ubiquitin-conjugating enzyme E2 of 230 kDa|ubiquitin-conjugating enzyme E2-230K|ubiquitin-conjugating enzyme E2O|ubiquitin-protein ligase O', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76389532, end_position_on_the_genomic_accession=76453206, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  63894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63894, CurrentID=0, Status='live', Symbol='VIPAS39', Aliases='C14orf133, SPE-39, SPE39, VIPAR, VPS16B, hSPE-39', description='VPS33B interacting protein, apical-basolateral polarity regulator, spe-39 homolog', other_designations='VPS33B-interacting protein involved in polarity and apical protein restriction', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77426675, end_position_on_the_genomic_accession=77457640, orientation='minus', exon_count=21, OMIM=613401, no_hdr0=''),
  63895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63895, CurrentID=0, Status='live', Symbol='PIEZO2', Aliases='C18orf30, C18orf58, DA3, DA5, FAM38B, FAM38B2, HsT748, HsT771, MWKS', description='piezo-type mechanosensitive ion channel component 2', other_designations='family with sequence similarity 38, member A pseudogene|family with sequence similarity 38, member B2|piezo-type mechanosensitive ion channel component 1 pseudogene|protein PIEZO2', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10670134, end_position_on_the_genomic_accession=11148762, orientation='minus', exon_count=56, OMIM=613629, no_hdr0=''),
  63897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63897, CurrentID=0, Status='live', Symbol='HEATR6', Aliases='ABC1', description='HEAT repeat containing 6', other_designations='amplified in breast cancer 1|amplified in breast cancer protein 1', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60043191, end_position_on_the_genomic_accession=60078931, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  63898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63898, CurrentID=0, Status='live', Symbol='SH2D4A', Aliases='PPP1R38, SH2A', description='SH2 domain containing 4A', other_designations='protein SH(2)A|protein phosphatase 1 regulatory subunit 38|protein phosphatase 1, regulatory subunit 38', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19313443, end_position_on_the_genomic_accession=19396218, orientation='plus', exon_count=11, OMIM=614968, no_hdr0=''),
  63899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63899, CurrentID=0, Status='live', Symbol='NSUN3', Aliases='MST077, MSTP077', description='NOP2/Sun RNA methyltransferase family member 3', other_designations='NOL1/NOP2/Sun domain family 3|NOL1/NOP2/Sun domain family member 3|NOL1/NOP2/Sun domain family, member 3|NOP2/Sun domain family, member 3', map_location='3q11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94063011, end_position_on_the_genomic_accession=94127740, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  63901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63901, CurrentID=0, Status='live', Symbol='FAM111A', Aliases='GCLEB, KCS2', description='family with sequence similarity 111 member A', other_designations='family with sequence similarity 111, member A', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59142845, end_position_on_the_genomic_accession=59155039, orientation='plus', exon_count=6, OMIM=615292, no_hdr0=''),
  105370693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370693, CurrentID=0, Status='live', Symbol='LOC105370693', Aliases='', description='uncharacterized LOC105370693', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104223584, end_position_on_the_genomic_accession=104288671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  63904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63904, CurrentID=0, Status='live', Symbol='DUSP21', Aliases='LMWDSP21', description='dual specificity phosphatase 21', other_designations='BJ-HCC-26 tumor antigen|LMW-DSP21|low molecular weight dual specificity phosphatase 21', map_location='Xp11.4-p11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44844003, end_position_on_the_genomic_accession=44844888, orientation='plus', exon_count=1, OMIM=300678, no_hdr0=''),
  63905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63905, CurrentID=0, Status='live', Symbol='MANBAL', Aliases='', description='mannosidase, beta A, lysosomal-like', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37289648, end_position_on_the_genomic_accession=37317260, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  63906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63906, CurrentID=0, Status='live', Symbol='GPATCH3', Aliases='GPATC3', description='G-patch domain containing 3', other_designations='G patch domain containing 3', map_location='1p35.3-p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26890488, end_position_on_the_genomic_accession=26901065, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  63908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63908, CurrentID=0, Status='live', Symbol='NAPB', Aliases='SNAP-BETA, SNAPB', description='NSF attachment protein beta', other_designations='N-ethylmaleimide-sensitive factor attachment protein, beta|beta soluble NSF attachment protein', map_location='20p12.3-p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23374519, end_position_on_the_genomic_accession=23421519, orientation='minus', exon_count=13, OMIM=611270, no_hdr0=''),
  105370694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370694, CurrentID=0, Status='live', Symbol='LOC105370694', Aliases='', description='uncharacterized LOC105370694', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104353099, end_position_on_the_genomic_accession=104365561, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  63910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63910, CurrentID=0, Status='live', Symbol='SLC17A9', Aliases='C20orf59, POROK8, VNUT', description='solute carrier family 17 (vesicular nucleotide transporter), member 9', other_designations='vesicular nucleotide transporter SLC17A9', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62952647, end_position_on_the_genomic_accession=62969585, orientation='plus', exon_count=14, OMIM=612107, no_hdr0=''),
  63914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63914, CurrentID=0, Status='live', Symbol='LINC01590', Aliases='C6orf164, dJ102H19.4', description='long intergenic non-protein coding RNA 1590', other_designations='', map_location='6q15-q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87397124, end_position_on_the_genomic_accession=87399741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  63915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63915, CurrentID=0, Status='live', Symbol='BLOC1S5', Aliases='BLOS5, MU, MUTED', description='biogenesis of lysosomal organelles complex-1, subunit 5, muted', other_designations='BLOC-1 subunit 5|muted homolog|protein Muted homolog', map_location='6p25.1-p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8013567, end_position_on_the_genomic_accession=8064414, orientation='minus', exon_count=6, OMIM=607289, no_hdr0=''),
  63916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63916, CurrentID=0, Status='live', Symbol='ELMO2', Aliases='CED-12, CED12, Ced-12A, ELMO-2', description='engulfment and cell motility 2', other_designations='PH domain protein CED12A|ced-12 homolog 2|protein ced-12 homolog A', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46366049, end_position_on_the_genomic_accession=46407051, orientation='minus', exon_count=22, OMIM=606421, no_hdr0=''),
  63917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63917, CurrentID=0, Status='live', Symbol='GALNT11', Aliases='GALNAC-T11, GALNACT11', description='polypeptide N-acetylgalactosaminyltransferase 11', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 11|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 11 (GalNAc-T11)|polypeptide GalNAc transferase 11|pp-GaNTase 11|protein-UDP acetylgalactosaminyltransferase 11', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152025674, end_position_on_the_genomic_accession=152122347, orientation='plus', exon_count=15, OMIM=615130, no_hdr0=''),
  63920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63920, CurrentID=0, Status='live', Symbol='ZBED8', Aliases='Buster3, C5orf54', description='zinc finger, BED-type containing 8', other_designations='transposon-derived Buster3 transposase-like protein|zinc finger BED domain-containing protein 8', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160393148, end_position_on_the_genomic_accession=160400097, orientation='minus', exon_count=3, OMIM=615253, no_hdr0=''),
  105370696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370696, CurrentID=0, Status='live', Symbol='LOC105370696', Aliases='', description='uncharacterized LOC105370696', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104735054, end_position_on_the_genomic_accession=104737974, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  63922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63922, CurrentID=0, Status='live', Symbol='CHTF18', Aliases='C16orf41, C321D2.2, C321D2.3, C321D2.4, CHL12, Ctf18, RUVBL', description='chromosome transmission fidelity factor 18', other_designations='C321D2.3 (novel protein)|C321D2.4 (novel protein)|CTF18, chromosome transmission fidelity factor 18 homolog|hCTF18|homolog of yeast CHL12|some homology with holliday junction DNA helicase RUVB like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=788622, end_position_on_the_genomic_accession=798074, orientation='plus', exon_count=22, OMIM=613201, no_hdr0=''),
  63923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63923, CurrentID=0, Status='live', Symbol='TNN', Aliases='TN-W', description='tenascin N', other_designations='TN-N', map_location='1q23-q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175067858, end_position_on_the_genomic_accession=175148066, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  63924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63924, CurrentID=0, Status='live', Symbol='CIDEC', Aliases='CIDE-3, CIDE3, FPLD5, FSP27', description='cell death-inducing DFFA-like effector c', other_designations='fat specific protein 27', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9866710, end_position_on_the_genomic_accession=9880262, orientation='minus', exon_count=9, OMIM=612120, no_hdr0=''),
  63925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63925, CurrentID=0, Status='live', Symbol='ZNF335', Aliases='MCPH10, NIF-1, NIF1, NIF2', description='zinc finger protein 335', other_designations='NRC-interacting factor 1|zinc-finger/leucine-zipper co-transducer NIF1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45948653, end_position_on_the_genomic_accession=45972194, orientation='minus', exon_count=28, OMIM=610827, no_hdr0=''),
  63926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63926, CurrentID=0, Status='live', Symbol='ANKEF1', Aliases='ANKRD5', description='ankyrin repeat and EF-hand domain containing 1', other_designations='ankyrin repeat domain-containing protein 5', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10034999, end_position_on_the_genomic_accession=10056764, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105370697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370697, CurrentID=0, Status='live', Symbol='LOC105370697', Aliases='', description='uncharacterized LOC105370697', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105531643, end_position_on_the_genomic_accession=105534318, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  63928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63928, CurrentID=0, Status='live', Symbol='CHP2', Aliases='', description='calcineurin-like EF-hand protein 2', other_designations='calcineurin-like EF hand protein 2|hepatocellular carcinoma antigen gene 520|hepatocellular carcinoma-associated antigen 520', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23754627, end_position_on_the_genomic_accession=23758951, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  63929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63929, CurrentID=0, Status='live', Symbol='XPNPEP3', Aliases='APP3, ICP55, NPHPL1', description='X-prolyl aminopeptidase 3, mitochondrial', other_designations='Intermediate Cleaving Peptidase 55|X-Pro aminopeptidase 3|X-prolyl aminopeptidase (aminopeptidase P) 3, putative', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40857081, end_position_on_the_genomic_accession=40932819, orientation='plus', exon_count=12, OMIM=613553, no_hdr0=''),
  63930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63930, CurrentID=0, Status='live', Symbol='LOC63930', Aliases='', description='uncharacterized LOC63930', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63009383, end_position_on_the_genomic_accession=63085071, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  63931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63931, CurrentID=0, Status='live', Symbol='MRPS14', Aliases='DJ262D12.2, HSMRPS14, MRP-S14, S14mt', description='mitochondrial ribosomal protein S14', other_designations='mitochondrial 28S ribosomal protein S14', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175012958, end_position_on_the_genomic_accession=175023455, orientation='minus', exon_count=3, OMIM=611978, no_hdr0=''),
  63932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63932, CurrentID=0, Status='live', Symbol='CXorf56', Aliases='', description='chromosome X open reading frame 56', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119538149, end_position_on_the_genomic_accession=119565434, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  63933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63933, CurrentID=0, Status='live', Symbol='MCUR1', Aliases='C6orf79, CCDC90A', description='mitochondrial calcium uniporter regulator 1', other_designations='coiled-coil domain containing 90A|coiled-coil domain-containing protein 90A, mitochondrial', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13786549, end_position_on_the_genomic_accession=13814563, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  63934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63934, CurrentID=0, Status='live', Symbol='ZNF667', Aliases='MIPU1', description='zinc finger protein 667', other_designations='myocardial ischemic preconditioning upregulated 1 ortholog', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56438512, end_position_on_the_genomic_accession=56478065, orientation='minus', exon_count=9, OMIM=611024, no_hdr0=''),
  63935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63935, CurrentID=0, Status='live', Symbol='PCIF1', Aliases='C20orf67, PPP1R121', description='PDX1 C-terminal inhibiting factor 1', other_designations='PDX-1 C terminus-interacting factor 1|protein phosphatase 1, regulatory subunit 121', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45934674, end_position_on_the_genomic_accession=45948023, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  63939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63939, CurrentID=0, Status='live', Symbol='FAM217B', Aliases='C20orf177, dJ551D2.5', description='family with sequence similarity 217 member B', other_designations='family with sequence similarity 217, member B|uncharacterized protein C20orf177', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59933764, end_position_on_the_genomic_accession=59948680, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  63940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63940, CurrentID=0, Status='live', Symbol='GPSM3', Aliases='AGS4, C6orf9, G18, G18.1a, G18.1b, G18.2, NG1', description='G-protein signaling modulator 3', other_designations='G-protein signaling modulator 3 (AGS3-like, C. elegans)|G-protein signalling modulator 3 (AGS3-like, C. elegans)|activator of G-protein signaling 4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32190766, end_position_on_the_genomic_accession=32195523, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  63941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63941, CurrentID=0, Status='live', Symbol='NECAB3', Aliases='APBA2BP, EFCBP3, NIP1, STIP3, SYTIP2, XB51, dJ63M2.4, dJ63M2.5', description='N-terminal EF-hand calcium binding protein 3', other_designations='EF-hand calcium binding protein 3|Nek2-interacting protein 1|X11L-binding protein 51|amyloid beta (A4) precursor protein-binding, family A, member 2 binding protein|amyloid beta A4 protein-binding family A member 2-binding protein|neuronal calcium-binding protein 3|neuronal calcium-binding protein NECAB3|synaptotagmin interacting protein 2|synaptotagmin interacting protein STIP3', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33657087, end_position_on_the_genomic_accession=33674458, orientation='minus', exon_count=16, OMIM=612478, no_hdr0=''),
  63943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63943, CurrentID=0, Status='live', Symbol='FKBPL', Aliases='DIR1, FKBP4, NG7, WISP39', description='FK506 binding protein like', other_designations='FK506-binding protein like|WAF-1/CIP1 stabilizing protein 39|peptidyl-prolyl cis-trans isomerase', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32128707, end_position_on_the_genomic_accession=32130308, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  63944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63944, CurrentID=0, Status='live', Symbol='SCZD10', Aliases='', description='schizophrenia disorder 10 (periodic catatonia)', other_designations='', map_location='15q15', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605419, no_hdr0=''),
  63945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63945, CurrentID=0, Status='live', Symbol='DFNA40', Aliases='', description='deafness, autosomal dominant 40', other_designations='deafness, autosomal dominant 40, sensorineural', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63946, CurrentID=0, Status='live', Symbol='DMRTC2', Aliases='', description='DMRT-like family C2', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41844743, end_position_on_the_genomic_accession=41852330, orientation='plus', exon_count=13, OMIM=614806, no_hdr0=''),
  63947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63947, CurrentID=0, Status='live', Symbol='DMRTC1', Aliases='', description='DMRT-like family C1', other_designations='doublesex-mab-3', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72872025, end_position_on_the_genomic_accession=72943653, orientation='minus', exon_count=9, OMIM=300878, no_hdr0=''),
  63948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63948, CurrentID=0, Status='live', Symbol='DMRTB1', Aliases='', description='DMRT like family B with proline rich C-terminal 1', other_designations='DMRT-like family B with proline-rich C-terminal 1|DMRT-like family B with proline-rich C-terminal, 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53459399, end_position_on_the_genomic_accession=53467488, orientation='plus', exon_count=4, OMIM=614805, no_hdr0=''),
  63950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63950, CurrentID=0, Status='live', Symbol='DMRTA2', Aliases='', description='DMRT-like family A2', other_designations='doublesex- and mab-3-related transcription factor 5', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50417550, end_position_on_the_genomic_accession=50423447, orientation='minus', exon_count=3, OMIM=614804, no_hdr0=''),
  63951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63951, CurrentID=0, Status='live', Symbol='DMRTA1', Aliases='DMO, DMRT4', description='DMRT-like family A1', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22446841, end_position_on_the_genomic_accession=22452473, orientation='plus', exon_count=2, OMIM=614803, no_hdr0=''),
  100499192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499192, CurrentID=0, Status='live', Symbol='LOC100499192', Aliases='', description='male-specific lethal 3 homolog (Drosophila) pseudogene', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99339630, end_position_on_the_genomic_accession=99341879, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326105, CurrentID=0, Status='live', Symbol='UBE2CP1', Aliases='', description='ubiquitin conjugating enzyme E2C pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2C pseudogene 1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30682862, end_position_on_the_genomic_accession=30683650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326106, CurrentID=0, Status='live', Symbol='TUBBP3', Aliases='', description='tubulin beta pseudogene 3', other_designations='tubulin, beta polypeptide pseudogene 3|tubulin, beta pseudogene 3', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43274968, end_position_on_the_genomic_accession=43275849, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326107, CurrentID=0, Status='live', Symbol='STK16P1', Aliases='STK16P', description='serine/threonine kinase 16 pseudogene 1', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49676668, end_position_on_the_genomic_accession=49678511, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499194, CurrentID=0, Status='live', Symbol='LOC100499194', Aliases='', description='uncharacterized LOC100499194', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113977669, end_position_on_the_genomic_accession=113979984, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  63967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63967, CurrentID=0, Status='live', Symbol='CLSPN', Aliases='', description='claspin', other_designations='claspin homolog|hClaspin', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35720691, end_position_on_the_genomic_accession=35769997, orientation='minus', exon_count=26, OMIM=605434, no_hdr0=''),
  101927024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927024, CurrentID=0, Status='live', Symbol='LOC101927024', Aliases='', description='uncharacterized LOC101927024', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38694290, end_position_on_the_genomic_accession=38720073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  63970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63970, CurrentID=0, Status='live', Symbol='TP53AIP1', Aliases='P53AIP1', description='tumor protein p53 regulated apoptosis inducing protein 1', other_designations='', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128934732, end_position_on_the_genomic_accession=128943399, orientation='minus', exon_count=5, OMIM=605426, no_hdr0=''),
  63971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63971, CurrentID=0, Status='live', Symbol='KIF13A', Aliases='RBKIN, bA500C11.2', description='kinesin family member 13A', other_designations='homolog of mouse KIF13A mannose-6-phosphate receptor transporter|kinesin-like protein RBKIN', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17759183, end_position_on_the_genomic_accession=17987629, orientation='minus', exon_count=47, OMIM=605433, no_hdr0=''),
  63973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63973, CurrentID=0, Status='live', Symbol='NEUROG2', Aliases='Atoh4, Math4A, NGN2, bHLHa8, ngn-2', description='neurogenin 2', other_designations='class A basic helix-loop-helix protein 8|protein atonal homolog 4', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112513516, end_position_on_the_genomic_accession=112516172, orientation='minus', exon_count=2, OMIM=606624, no_hdr0=''),
  63974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63974, CurrentID=0, Status='live', Symbol='NEUROD6', Aliases='Atoh2, MATH2, Math-2, NEX1M, Nex1, bHLHa2', description='neuronal differentiation 6', other_designations='brain my051 protein|class A basic helix-loop-helix protein 2|protein atonal homolog 2', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31337461, end_position_on_the_genomic_accession=31340924, orientation='minus', exon_count=2, OMIM=611513, no_hdr0=''),
  63976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63976, CurrentID=0, Status='live', Symbol='PRDM16', Aliases='CMD1LL, LVNC8, MEL1, PFM13', description='PR domain containing 16', other_designations='MDS1/EVI1-like gene 1|transcription factor MEL1', map_location='1p36.23-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3068227, end_position_on_the_genomic_accession=3438621, orientation='plus', exon_count=18, OMIM=605557, no_hdr0=''),
  63977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63977, CurrentID=0, Status='live', Symbol='PRDM15', Aliases='C21orf83, PFM15, ZNF298', description='PR domain containing 15', other_designations='zinc finger protein 298', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41798225, end_position_on_the_genomic_accession=41879482, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  63978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63978, CurrentID=0, Status='live', Symbol='PRDM14', Aliases='PFM11', description='PR domain containing 14', other_designations='PR domain-containing 14|PR domain-containing protein 14', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70051126, end_position_on_the_genomic_accession=70071327, orientation='minus', exon_count=9, OMIM=611781, no_hdr0=''),
  63979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63979, CurrentID=0, Status='live', Symbol='FIGNL1', Aliases='', description='fidgetin-like 1', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50444129, end_position_on_the_genomic_accession=50450390, orientation='minus', exon_count=5, OMIM=615383, no_hdr0=''),
  63982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=63982, CurrentID=0, Status='live', Symbol='ANO3', Aliases='C11orf25, DYT23, DYT24, GENX-3947, TMEM16C', description='anoctamin 3', other_designations='transmembrane protein 16C (eight membrane-spanning domains)', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26188821, end_position_on_the_genomic_accession=26668409, orientation='plus', exon_count=30, OMIM=610110, no_hdr0=''),
  105373227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373227, CurrentID=0, Status='live', Symbol='LOC105373227', Aliases='', description='uncharacterized LOC105373227', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240553784, end_position_on_the_genomic_accession=240572211, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64002, CurrentID=0, Status='live', Symbol='PCGEM1', Aliases='LINC00071, NCRNA00071, PCAT9', description='PCGEM1, prostate-specific transcript (non-protein coding)', other_designations='long intergenic non-protein coding RNA 71|prostate cancer associated transcript 9|prostate-specific non-coding|prostate-specific transcript 1 (non-protein coding)', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=192749845, end_position_on_the_genomic_accession=192776899, orientation='plus', exon_count=3, OMIM=605443, no_hdr0=''),
  64005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64005, CurrentID=0, Status='live', Symbol='MYO1G', Aliases='HA2, HLA-HA2, MHAG', description='myosin IG', other_designations='minor histocompatibility antigen HA-2|myosin-Ig', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44962661, end_position_on_the_genomic_accession=44979105, orientation='minus', exon_count=22, OMIM=600642, no_hdr0=''),
  64006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64006, CurrentID=0, Status='live', Symbol='ERVK-6', Aliases='ERVK6, HERV-K(C7), HERV-K108, K-Rev, c-orf, cORF', description='endogenous retrovirus group K, member 6', other_designations='HERV-K (HML-2.HOM)|HERV-K(C7) envelope protein|HERV-K108 envelope protein|HERV-K_7p22.1 provirus Rec protein|HERV-K_7p22.1 provirus ancestral Env polyprotein|HERV-K_7p22.1 provirus ancestral Pol protein|HERV-K_7p22.1 provirus ancestral Pro protein|Rev-like protein|Rev/Rex homolog|endogenous retroviral sequence K, 6|envelope glycoprotein|envelope protein', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=4591306, end_position_on_the_genomic_accession=4596943, orientation='minus', exon_count=1, OMIM=605626, no_hdr0=''),
  105374745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374745, CurrentID=0, Status='live', Symbol='LOC105374745', Aliases='', description='uncharacterized LOC105374745', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42957579, end_position_on_the_genomic_accession=42965053, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370714, CurrentID=0, Status='live', Symbol='LOC105370714', Aliases='', description='uncharacterized LOC105370714', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21267005, end_position_on_the_genomic_accession=21273998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370717, CurrentID=0, Status='live', Symbol='LOC105370717', Aliases='', description='uncharacterized LOC105370717', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21151416, end_position_on_the_genomic_accession=21154404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374746, CurrentID=0, Status='live', Symbol='LOC105374746', Aliases='', description='uncharacterized LOC105374746', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42996584, end_position_on_the_genomic_accession=43000906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100127981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127981, CurrentID=0, Status='live', Symbol='LOC100127981', Aliases='', description='fibrillarin pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183418885, end_position_on_the_genomic_accession=183419805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370719, CurrentID=0, Status='live', Symbol='LOC105370719', Aliases='', description='uncharacterized LOC105370719', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20282015, end_position_on_the_genomic_accession=20284930, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64061, CurrentID=0, Status='live', Symbol='TSPYL2', Aliases='CDA1, CINAP, CTCL, DENTT, HRIHFB2216, NP79, SE204, TSPX', description='TSPY-like 2', other_designations='CASK-interacting nucleosome assembly protein|CTCL tumor antigen se20-4|CTCL-associated antigen se20-4|TSPY-like protein 2|cell division autoantigen 1|cutaneous T-cell lymphoma-associated antigen se20-4|cutaneous T-cell lymphoma-associated tumor antigen se20-4|differentially expressed nucleolar TGF-beta1 target|differentially-expressed nucleolar TGF-beta1 target protein|nuclear protein of 79 kDa|testis-specific protein Y encoded-like 2', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53082360, end_position_on_the_genomic_accession=53088546, orientation='plus', exon_count=7, OMIM=300564, no_hdr0=''),
  64062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64062, CurrentID=0, Status='live', Symbol='RBM26', Aliases='ARRS2, C13orf10, PPP1R132, PRO1777, SE70-2, ZC3H17', description='RNA binding motif protein 26', other_designations='CTCL tumor antigen se70-2|acidic rich RS domain containing 2|cutaneous T-cell lymphoma tumor antigen se70-2|protein phosphatase 1, regulatory 132', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79311827, end_position_on_the_genomic_accession=79406258, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  64063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64063, CurrentID=0, Status='live', Symbol='PRSS22', Aliases='BSSP-4, SP001LA, hBSSP-4', description='protease, serine 22', other_designations='prosemin|protease, serine S1 family member 22|protease, serine, 22|serine protease 22|serine protease 26|tryptase epsilon', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2852727, end_position_on_the_genomic_accession=2858630, orientation='minus', exon_count=8, OMIM=609343, no_hdr0=''),
  64064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64064, CurrentID=0, Status='live', Symbol='OXCT2', Aliases='FKSG25, SCOTT', description='3-oxoacid CoA-transferase 2', other_designations='3-oxoacid CoA transferase 2|3-oxoacid CoA-transferase 2A|3-oxoacid-CoA transferase 2A|succinyl-CoA:3-ketoacid-coenzyme A transferase 2, mitochondrial|testis-specific succinyl CoA:3-oxoacid CoA-transferase|testis-specific succinyl-CoA:3-oxoacid CoA-transferase', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39769525, end_position_on_the_genomic_accession=39771348, orientation='minus', exon_count=1, OMIM=610289, no_hdr0=''),
  64065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64065, CurrentID=0, Status='live', Symbol='PERP', Aliases='KCP1, KRTCAP1, PIGPC1, THW, dJ496H19.1', description='PERP, TP53 apoptosis effector', other_designations='1110017A08Rik|KCP-1|keratinocyte associated protein 1|keratinocyte-associated protein 1|keratinocytes associated protein 1|p53 apoptosis effector related to PMP22|p53-induced protein PIGPC1|transmembrane protein THW', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138088505, end_position_on_the_genomic_accession=138107523, orientation='minus', exon_count=3, OMIM=609301, no_hdr0=''),
  64066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64066, CurrentID=0, Status='live', Symbol='MMP27', Aliases='MMP-27', description='matrix metallopeptidase 27', other_designations='matrix metalloprotease 27|matrix metalloproteinase 27', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102691510, end_position_on_the_genomic_accession=102705755, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  64067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64067, CurrentID=0, Status='live', Symbol='NPAS3', Aliases='MOP6, PASD6, bHLHe12', description='neuronal PAS domain protein 3', other_designations='PAS domain-containing protein 6|basic-helix-loop-helix-PAS protein MOP6|class E basic helix-loop-helix protein 12|member of PAS protein 6|neuronal PAS3', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32934909, end_position_on_the_genomic_accession=33804176, orientation='plus', exon_count=16, OMIM=609430, no_hdr0=''),
  64069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64069, CurrentID=0, Status='live', Symbol='ATOD1', Aliases='ATOD', description='Dermatitis, atopic', other_designations='', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603165, no_hdr0=''),
  105370721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370721, CurrentID=0, Status='live', Symbol='LOC105370721', Aliases='', description='uncharacterized LOC105370721', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22290064, end_position_on_the_genomic_accession=22294417, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64072, CurrentID=0, Status='live', Symbol='CDH23', Aliases='CDHR23, USH1D', description='cadherin-related 23', other_designations='cadherin-like 23|cadherin-related family member 23|otocadherin', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71396934, end_position_on_the_genomic_accession=71815947, orientation='plus', exon_count=80, OMIM=605516, no_hdr0=''),
  64073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64073, CurrentID=0, Status='live', Symbol='C19orf33', Aliases='H2RSP, IMUP, IMUP-1, IMUP-2', description='chromosome 19 open reading frame 33', other_designations='HAI-2 related small protein|HAI-2-related small protein|hepatocyte growth factor activator inhibitor type 2-related small protein|immortalization-upregulated protein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38304164, end_position_on_the_genomic_accession=38305006, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64077, CurrentID=0, Status='live', Symbol='LHPP', Aliases='HDHD2B', description='phospholysine phosphohistidine inorganic pyrophosphate phosphatase', other_designations='hLHPP', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124461772, end_position_on_the_genomic_accession=124614141, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  64078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64078, CurrentID=0, Status='live', Symbol='SLC28A3', Aliases='CNT3', description='solute carrier family 28 (concentrative nucleoside transporter), member 3', other_designations='concentrative Na(+)-nucleoside cotransporter 3|concentrative Na+-nucleoside cotransporter|solute carrier family 28 (sodium-coupled nucleoside transporter), member 3', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84275850, end_position_on_the_genomic_accession=84368632, orientation='minus', exon_count=22, OMIM=608269, no_hdr0=''),
  64080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64080, CurrentID=0, Status='live', Symbol='RBKS', Aliases='RBSK', description='ribokinase', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27781364, end_position_on_the_genomic_accession=27890396, orientation='minus', exon_count=9, OMIM=611132, no_hdr0=''),
  64081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64081, CurrentID=0, Status='live', Symbol='PBLD', Aliases='HEL-S-306, MAWBP, MAWDBP', description='phenazine biosynthesis-like protein domain containing', other_designations='MAWD-binding protein|epididymis secretory protein Li 306', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68282660, end_position_on_the_genomic_accession=68332928, orientation='minus', exon_count=10, OMIM=612189, no_hdr0=''),
  64083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64083, CurrentID=0, Status='live', Symbol='GOLPH3', Aliases='GOPP1, GPP34, MIDAS, Vps74', description='golgi phosphoprotein 3 (coat-protein)', other_designations='coat protein GPP34|coat-protein|golgi peripheral membrane protein 1, 34 kDa|golgi protein|golgi-associated protein|mitochondrial DNA absence factor', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32124718, end_position_on_the_genomic_accession=32174319, orientation='minus', exon_count=4, OMIM=612207, no_hdr0=''),
  64084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64084, CurrentID=0, Status='live', Symbol='CLSTN2', Aliases='ALC-GAMMA, CDHR13, CS2, CSTN2, alcagamma', description='calsyntenin 2', other_designations='alcadein gamma|alcadein-gamma|cadherin-related family member 13', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139935185, end_position_on_the_genomic_accession=140568077, orientation='plus', exon_count=17, OMIM=611323, no_hdr0=''),
  64087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64087, CurrentID=0, Status='live', Symbol='MCCC2', Aliases='MCCB', description='methylcrotonoyl-CoA carboxylase 2', other_designations='3-methylcrotonyl-CoA carboxylase 2|3-methylcrotonyl-CoA carboxylase non-biotin-containing subunit|3-methylcrotonyl-CoA:carbon dioxide ligase subunit beta|MCCase subunit beta|biotin carboxylase|methylcrotonoyl-CoA carboxylase 2 (beta)|methylcrotonoyl-Coenzyme A carboxylase 2 (beta)|non-biotin containing subunit of 3-methylcrotonyl-CoA carboxylase', map_location='5q12-q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71587288, end_position_on_the_genomic_accession=71658706, orientation='plus', exon_count=19, OMIM=609014, no_hdr0=''),
  64089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64089, CurrentID=0, Status='live', Symbol='SNX16', Aliases='', description='sorting nexin 16', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81799583, end_position_on_the_genomic_accession=81842325, orientation='minus', exon_count=11, OMIM=614903, no_hdr0=''),
  64090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64090, CurrentID=0, Status='live', Symbol='GAL3ST2', Aliases='GAL3ST-2, GP3ST', description='galactose-3-O-sulfotransferase 2', other_designations="beta-galactose-3-O-sulfotransferase 2|gal 3-O-sulphotransferase|gal-beta-1, 3-GalNAc 3'-sulfotransferase 2|galbeta1-3GalNAc 3'-sulfotransferase 2|glycoprotein beta-Gal 3'-sulfotransferase 2", map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241776825, end_position_on_the_genomic_accession=241804287, orientation='plus', exon_count=4, OMIM=608237, no_hdr0=''),
  64091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64091, CurrentID=0, Status='live', Symbol='POPDC2', Aliases='POP2', description='popeye domain containing 2', other_designations='popeye protein 2', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119642055, end_position_on_the_genomic_accession=119660590, orientation='minus', exon_count=6, OMIM=605823, no_hdr0=''),
  64092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64092, CurrentID=0, Status='live', Symbol='SAMSN1', Aliases='HACS1, NASH1, SASH2, SH3D6B, SLy2', description='SAM domain, SH3 domain and nuclear localization signals 1', other_designations='SAM and SH3 domain containing 2|SAM domain, SH3 domain and nuclear localisation signals, 1|SAM domain, SH3 domain and nuclear localization signals protein 1|SH3-SAM adaptor protein|Src homology domain 3 (SH3)-containing adapter protein SH3 lymphocyte protein 2|hematopoietic adapter-containing SH3 and sterile &#945;-motif (SAM) domains 1|hematopoietic adapter-containing SH3 and sterile alpha-motif (SAM) domains 1|hematopoietic adaptor containing SH3 and SAM domains 1|nuclear localization signals, SAM and SH3 domain containing 1', map_location='21q11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14485228, end_position_on_the_genomic_accession=14583402, orientation='minus', exon_count=14, OMIM=607978, no_hdr0=''),
  64093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64093, CurrentID=0, Status='live', Symbol='SMOC1', Aliases='OAS', description='SPARC related modular calcium binding 1', other_designations='secreted modular calcium-binding protein 1', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69879388, end_position_on_the_genomic_accession=70032366, orientation='plus', exon_count=12, OMIM=608488, no_hdr0=''),
  64094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64094, CurrentID=0, Status='live', Symbol='SMOC2', Aliases='DTDP1, MST117, MSTP117, MSTP140, SMAP2, bA270C4A.1, bA37D8.1, dJ421D16.1', description='SPARC related modular calcium binding 2', other_designations='SMAP-2|SMOC-2|secreted modular calcium-binding protein 2|smooth muscle associated protein 2|smooth muscle-associated protein 2|thyroglobulin type-1 repeat containing protein', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168441151, end_position_on_the_genomic_accession=168667994, orientation='plus', exon_count=14, OMIM=607223, no_hdr0=''),
  64096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64096, CurrentID=0, Status='live', Symbol='GFRA4', Aliases='', description='GDNF family receptor alpha 4', other_designations='GDNF receptor alpha-4|GDNFR-alpha-4|GFR receptor alpha 4|GFR-alpha-4|persephin receptor', map_location='20p13-p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3659292, end_position_on_the_genomic_accession=3663399, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  64097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64097, CurrentID=0, Status='live', Symbol='EPB41L4A', Aliases='EPB41L4, NBL4', description='erythrocyte membrane protein band 4.1 like 4A', other_designations='erythrocyte protein band 4.1-like 4', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112142319, end_position_on_the_genomic_accession=112419490, orientation='minus', exon_count=26, OMIM=612141, no_hdr0=''),
  64098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64098, CurrentID=0, Status='live', Symbol='PARVG', Aliases='', description='parvin gamma', other_designations='parvin, gamma', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44170227, end_position_on_the_genomic_accession=44208469, orientation='plus', exon_count=19, OMIM=608122, no_hdr0=''),
  64100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64100, CurrentID=0, Status='live', Symbol='ELSPBP1', Aliases='E12, EDDM12, EL149, HE12', description='epididymal sperm binding protein 1', other_designations='epididymal protein 12|epididymal secretory protein 12|epididymis luminal secretory protein 149', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47994650, end_position_on_the_genomic_accession=48025156, orientation='plus', exon_count=7, OMIM=607443, no_hdr0=''),
  64101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64101, CurrentID=0, Status='live', Symbol='LRRC4', Aliases='NAG14, NGL-2', description='leucine rich repeat containing 4', other_designations='brain tumor associated protein LRRC4|brain tumor-associated protein BAG|leucine-rich repeat-containing 4|nasopharyngeal carcinoma-associated gene 14 protein|netrin-G2 ligand', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128027071, end_position_on_the_genomic_accession=128032107, orientation='minus', exon_count=3, OMIM=610486, no_hdr0=''),
  64102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64102, CurrentID=0, Status='live', Symbol='TNMD', Aliases='BRICD4, CHM1L, TEM', description='tenomodulin', other_designations='BRICHOS domain containing 4|chondromodulin-1-like protein|chondromodulin-I-like protein|chondromodulin-IB|hChM1L|hTeM|myodulin|tendin', map_location='Xq21.33-q23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100584793, end_position_on_the_genomic_accession=100600983, orientation='plus', exon_count=9, OMIM=300459, no_hdr0=''),
  64105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64105, CurrentID=0, Status='live', Symbol='CENPK', Aliases='AF5alpha, CENP-K, FKSG14, P33, Solt', description='centromere protein K', other_designations='SoxLZ/Sox6-binding protein Solt|interphase centromere complex protein 37|leucine zipper protein FKSG14|protein AF-5alpha', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65508067, end_position_on_the_genomic_accession=65563189, orientation='minus', exon_count=16, OMIM=611502, no_hdr0=''),
  64106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64106, CurrentID=0, Status='live', Symbol='NPFFR1', Aliases='GPR147, NPFF1, NPFF1R1, OT7T022', description='neuropeptide FF receptor 1', other_designations='G protein-coupled receptor 147|G-protein coupled receptor 147|RFamide-related peptide receptor OT7T022|neuropeptide FF 1', map_location='10q21-q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70254957, end_position_on_the_genomic_accession=70283694, orientation='minus', exon_count=4, OMIM=607448, no_hdr0=''),
  64108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64108, CurrentID=0, Status='live', Symbol='RTP4', Aliases='IFRG28, Z3CXXC4', description='receptor (chemosensory) transporter protein 4', other_designations='28 kDa interferon-responsive protein|28kD interferon responsive protein|3CxxC-type zinc finger protein 4|receptor transporter protein 4|receptor transporting protein 4|zinc finger, 3CxxC-type 4', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187368380, end_position_on_the_genomic_accession=187371584, orientation='plus', exon_count=2, OMIM=609350, no_hdr0=''),
  64109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64109, CurrentID=0, Status='live', Symbol='CRLF2', Aliases='CRL2Y, TSLPR, CRLF2', description='cytokine receptor-like factor 2', other_designations='IL-XR|TSLP receptor|cytokine receptor CRL2 precusor|thymic stromal lymphopoietin protein receptor|thymic stromal-derived lymphopoietin receptor', map_location='Xp22.3; Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1190449, end_position_on_the_genomic_accession=1212815, orientation='minus', exon_count=9, OMIM='300357,400023', no_hdr0=''),
  64110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64110, CurrentID=0, Status='live', Symbol='MAGEF1', Aliases='', description='MAGE family member F1', other_designations='MAGE-F1 antigen|melanoma antigen family F, 1|melanoma antigen family F1', map_location='3q13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184710367, end_position_on_the_genomic_accession=184712048, orientation='minus', exon_count=1, OMIM=609267, no_hdr0=''),
  64111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64111, CurrentID=0, Status='live', Symbol='NPVF', Aliases='C7orf9, RFRP', description='neuropeptide VF precursor', other_designations='FMRFamide-related peptide|FMRFamide-related peptides|RFamide-related peptide|neuropeptide NPVF', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25224570, end_position_on_the_genomic_accession=25228486, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64112, CurrentID=0, Status='live', Symbol='MOAP1', Aliases='MAP-1, PNMA4', description='modulator of apoptosis 1', other_designations='MAP1|paraneoplastic Ma antigen family member 4|paraneoplastic antigen Ma4|paraneoplastic antigen like 4', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93182196, end_position_on_the_genomic_accession=93184904, orientation='minus', exon_count=3, OMIM=609485, no_hdr0=''),
  64114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64114, CurrentID=0, Status='live', Symbol='TMBIM1', Aliases='LFG3, MST100, MSTP100, PP1201, RECS1', description='transmembrane BAX inhibitor motif containing 1', other_designations='transmembrane BAX inhibitor motif-containing protein 1', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218274194, end_position_on_the_genomic_accession=218292557, orientation='minus', exon_count=16, OMIM=610364, no_hdr0=''),
  64115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64115, CurrentID=0, Status='live', Symbol='C10orf54', Aliases='B7-H5, B7H5, DD1alpha, GI24, PP2135, SISP1, VISTA', description='chromosome 10 open reading frame 54', other_designations='Death Domain1alpha|V-domain Ig suppressor of T cell activation|sisp-1|stress-induced secreted protein-1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71747556, end_position_on_the_genomic_accession=71773580, orientation='minus', exon_count=7, OMIM=615608, no_hdr0=''),
  64116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64116, CurrentID=0, Status='live', Symbol='SLC39A8', Aliases='BIGM103, LZT-Hs6, PP3105, ZIP8', description='solute carrier family 39 (zinc transporter), member 8', other_designations='BCG induced integral membrane protein BIGM103|BCG-induced integral membrane protein in monocyte clone 103 protein|LIV-1 subfamily of ZIP zinc transporter 6|ZIP-8|Zrt- and Irt-like protein 8|solute carrier family 39 (metal ion transporter), member 8|solute carrier family 39 member 8', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102251041, end_position_on_the_genomic_accession=102345498, orientation='minus', exon_count=15, OMIM=608732, no_hdr0=''),
  64118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64118, CurrentID=0, Status='live', Symbol='DUS1L', Aliases='DUS1, PP3111', description='dihydrouridine synthase 1-like', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82057872, end_position_on_the_genomic_accession=82065821, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  64121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64121, CurrentID=0, Status='live', Symbol='RRAGC', Aliases='GTR2, RAGC, TIB929', description='Ras-related GTP binding C', other_designations='GTPase-interacting protein 2|Rag C protein', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38838197, end_position_on_the_genomic_accession=38859823, orientation='minus', exon_count=7, OMIM=608267, no_hdr0=''),
  64122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64122, CurrentID=0, Status='live', Symbol='FN3K', Aliases='', description='fructosamine 3 kinase', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82735576, end_position_on_the_genomic_accession=82751197, orientation='plus', exon_count=6, OMIM=608425, no_hdr0=''),
  64123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64123, CurrentID=0, Status='live', Symbol='ADGRL4', Aliases='ELTD1, ETL, KPG_003', description='adhesion G protein-coupled receptor L4', other_designations='EGF, latrophilin and seven transmembrane domain containing 1|EGF, latrophilin and seven transmembrane domain-containing protein 1|EGF-TM7-latrophilin-related protein', map_location='1p33-p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78889764, end_position_on_the_genomic_accession=79006810, orientation='minus', exon_count=15, OMIM=616419, no_hdr0=''),
  326268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326268, CurrentID=0, Status='live', Symbol='ABI1P1', Aliases='ABI1P, SSH3BP1P', description='abl-interactor 1 pseudogene 1', other_designations='spectrin SH3 domain binding protein 1 pseudogene', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55546559, end_position_on_the_genomic_accession=55550363, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326269, CurrentID=0, Status='live', Symbol='SRMP2', Aliases='', description='spermidine synthase pseudogene 2', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60918982, end_position_on_the_genomic_accession=60920163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64127, CurrentID=0, Status='live', Symbol='NOD2', Aliases='ACUG, BLAU, CARD15, CD, CLR16.3, IBD1, NLRC2B, PSORAS1, NOD2', description='nucleotide binding oligomerization domain containing 2', other_designations='NLR family, CARD domain containing 2|NOD-like receptor C2|caspase recruitment domain family, member 15|caspase recruitment domain protein 15|caspase recruitment domain-containing protein 15|inflammatory bowel disease protein 1|nucleotide-binding oligomerization domain 2|nucleotide-binding oligomerization domain containing 2|nucleotide-binding oligomerization domain, leucine rich repeat and CARD domain containing 2', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50693581, end_position_on_the_genomic_accession=50733081, orientation='plus', exon_count=16, OMIM=605956, no_hdr0=''),
  326272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326272, CurrentID=0, Status='live', Symbol='SNRPGP1', Aliases='SNRPGP', description='small nuclear ribonucleoprotein polypeptide G pseudogene 1', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50571774, end_position_on_the_genomic_accession=50572173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64129, CurrentID=0, Status='live', Symbol='TINAGL1', Aliases='ARG1, LCN7, LIECG3, TINAGRP', description='tubulointerstitial nephritis antigen-like 1', other_designations='OLRG-2|P3ECSL|TIN Ag-related protein|TIN-Ag-RP|TINAG-like 1|androgen-regulated gene 1|glucocorticoid-inducible protein 5|lipocalin 7|oxidized-LDL responsive gene 2|tubulointerstitial nephritis antigen-related protein', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31576384, end_position_on_the_genomic_accession=31587686, orientation='plus', exon_count=13, OMIM=616064, no_hdr0=''),
  64130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64130, CurrentID=0, Status='live', Symbol='LIN7B', Aliases='LIN-7B, MALS-2, MALS2, VELI2', description='lin-7 homolog B (C. elegans)', other_designations='hLin7B|hVeli2|mammalian lin-seven protein 2|veli-2|vertebrate lin-7 homolog 2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49114361, end_position_on_the_genomic_accession=49118460, orientation='plus', exon_count=6, OMIM=612331, no_hdr0=''),
  64131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64131, CurrentID=0, Status='live', Symbol='XYLT1', Aliases='DBQD2, PXYLT1, XT-I, XT1, XTI, XYLTI', description='xylosyltransferase I', other_designations='O-xylosyltransferase 1|beta-D-xylosyltransferase 1|peptide O-xylosyltransferase 1|xylT-I|xylosyltransferase iota', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17102324, end_position_on_the_genomic_accession=17470881, orientation='minus', exon_count=13, OMIM=608124, no_hdr0=''),
  64132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64132, CurrentID=0, Status='live', Symbol='XYLT2', Aliases='PXYLT2, SOS, XT-II, XT2, xylT-II', description='xylosyltransferase II', other_designations='UDP-D-xylose:proteoglycan core protein beta-D-xylosyltransferase|peptide O-xylosyltransferase 1|protein xylosyltransferase 2', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50346032, end_position_on_the_genomic_accession=50361185, orientation='plus', exon_count=13, OMIM=608125, no_hdr0=''),
  326277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326277, CurrentID=0, Status='live', Symbol='SETP2', Aliases='', description='SET pseudogene 2', other_designations='SET nuclear oncogene pseudogene', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51436150, end_position_on_the_genomic_accession=51439313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326278, CurrentID=0, Status='live', Symbol='RPL13AP2', Aliases='RPL13A_12_1359', description='ribosomal protein L13a pseudogene 2', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=47200549, end_position_on_the_genomic_accession=47201121, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64135, CurrentID=0, Status='live', Symbol='IFIH1', Aliases='AGS7, Hlcd, IDDM19, MDA-5, MDA5, RLR-2, SGMRT1', description='interferon induced, with helicase C domain 1', other_designations='CADM-140 autoantigen|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide|RIG-I-like receptor 2|RNA helicase-DEAD box protein 116|clinically amyopathic dermatomyositis autoantigen 140 kDa|helicard|helicase with 2 CARD domains|interferon induced with helicase C domain 1|melanoma differentiation associated protein-5|melanoma differentiation-associated protein 5|murabutide down-regulated protein', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162267079, end_position_on_the_genomic_accession=162318764, orientation='minus', exon_count=17, OMIM=606951, no_hdr0=''),
  64137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64137, CurrentID=0, Status='live', Symbol='ABCG4', Aliases='WHITE2', description='ATP binding cassette subfamily G member 4', other_designations='ATP-binding cassette, sub-family G (WHITE), member 4|ATP-binding cassette, subfamily G, member 4|putative ABC transporter', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119149040, end_position_on_the_genomic_accession=119162666, orientation='plus', exon_count=17, OMIM=607784, no_hdr0=''),
  326286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326286, CurrentID=0, Status='live', Symbol='RPL17P3', Aliases='', description='ribosomal protein L17 pseudogene 3', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81001134, end_position_on_the_genomic_accession=81001817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326288, CurrentID=0, Status='live', Symbol='RPL17P4', Aliases='RPL17_23_1411', description='ribosomal protein L17 pseudogene 4', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103282898, end_position_on_the_genomic_accession=103283498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64145, CurrentID=0, Status='live', Symbol='RBSN', Aliases='Rabenosyn-5, ZFYVE20', description='rabenosyn, RAB effector', other_designations='110 kDa protein|FYVE finger-containing Rab5 effector protein rabenosyn-5|RAB effector RBSN|zinc finger, FYVE domain containing 20', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15070069, end_position_on_the_genomic_accession=15099157, orientation='minus', exon_count=16, OMIM=609511, no_hdr0=''),
  64146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64146, CurrentID=0, Status='live', Symbol='PDF', Aliases='', description='peptide deformylase (mitochondrial)', other_designations='peptide deformylase-like protein|polypeptide deformylase', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69328621, end_position_on_the_genomic_accession=69330595, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64147, CurrentID=0, Status='live', Symbol='KIF9', Aliases='', description='kinesin family member 9', other_designations='kinesin protein 9', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47228026, end_position_on_the_genomic_accession=47282847, orientation='minus', exon_count=24, OMIM=607910, no_hdr0=''),
  326292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326292, CurrentID=0, Status='live', Symbol='RPL22P2', Aliases='RPL22_9_1393', description='ribosomal protein L22 pseudogene 2', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76886320, end_position_on_the_genomic_accession=76886775, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64149, CurrentID=0, Status='live', Symbol='C17orf75', Aliases='NJMU-R1', description='chromosome 17 open reading frame 75', other_designations='spermatogenesis-related protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32328441, end_position_on_the_genomic_accession=32350047, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  64150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64150, CurrentID=0, Status='live', Symbol='DIO3OS', Aliases='C14orf134, DIO3-AS1, DIO3-OS, NCRNA00041', description='DIO3 opposite strand/antisense RNA (head to head)', other_designations='DIO3 opposite strand (non-protein coding)|DIO3 opposite strand/antisense RNA (non-protein coding)|deiodinase, iodothyronine, type III opposite strand|uterine-derived 14 kDa protein', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101552223, end_position_on_the_genomic_accession=101555676, orientation='minus', exon_count=1, OMIM=608523, no_hdr0=''),
  64151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64151, CurrentID=0, Status='live', Symbol='NCAPG', Aliases='CAPG, CHCG, NY-MEL-3, YCG1', description='non-SMC condensin I complex subunit G', other_designations='XCAP-G homolog|chromosome condensation protein G|chromosome-associated protein G|condensin subunit CAP-G|melanoma antigen NY-MEL-3|non-SMC condensin I complex, subunit G', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17810813, end_position_on_the_genomic_accession=17844865, orientation='plus', exon_count=21, OMIM=606280, no_hdr0=''),
  326296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326296, CurrentID=0, Status='live', Symbol='HSBP1P1', Aliases='', description='heat shock factor binding protein 1 pseudogene 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58463889, end_position_on_the_genomic_accession=58464409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326297, CurrentID=0, Status='live', Symbol='RPL26P2', Aliases='RPL26_14_1341', description='ribosomal protein L26 pseudogene 2', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22694052, end_position_on_the_genomic_accession=22694581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326298, CurrentID=0, Status='live', Symbol='RPL26P3', Aliases='RPL26_17_1344', description='ribosomal protein L26 pseudogene 3', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28489905, end_position_on_the_genomic_accession=28490392, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326299, CurrentID=0, Status='live', Symbol='RPL27P1', Aliases='RPL27_7_1347', description='ribosomal protein L27 pseudogene 1', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30821391, end_position_on_the_genomic_accession=30821885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326300, CurrentID=0, Status='live', Symbol='HSPE1P2', Aliases='HSPEP2', description='heat shock protein family E (Hsp10) member 1 pseudogene 2', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene 2|heat shock 10kDa protein 1 pseudogene 2', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63601471, end_position_on_the_genomic_accession=63601934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326301, CurrentID=0, Status='live', Symbol='RPL36AP2', Aliases='RPL36A_18_1380', description='ribosomal protein L36a pseudogene 2', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65268813, end_position_on_the_genomic_accession=65269579, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326302, CurrentID=0, Status='live', Symbol='NEK2P1', Aliases='NEK2P', description='NEK2 pseudogene 1', other_designations='NIMA (never in mitosis gene a)-related kinase 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18888449, end_position_on_the_genomic_accession=18890566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326303, CurrentID=0, Status='live', Symbol='RPL36P3', Aliases='RPL36AP3, RPL36_6_1387', description='ribosomal protein L36 pseudogene 3', other_designations='ribosomal protein L36a pseudogene 3', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72929569, end_position_on_the_genomic_accession=72929964, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326304, CurrentID=0, Status='live', Symbol='RPL36AP4', Aliases='RPL36A_19_1402', description='ribosomal protein L36a pseudogene 4', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93243960, end_position_on_the_genomic_accession=93244423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326305, CurrentID=0, Status='live', Symbol='MOCS3P1', Aliases='MOCS3P', description='molybdenum cofactor synthesis 3 pseudogene 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61864216, end_position_on_the_genomic_accession=61866049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326306, CurrentID=0, Status='live', Symbol='RPL39P2', Aliases='RPL39_16_1342', description='ribosomal protein L39 pseudogene 2', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23229768, end_position_on_the_genomic_accession=23230132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499227, CurrentID=0, Status='live', Symbol='USP2-AS1', Aliases='THY1-AS1', description='USP2 antisense RNA 1 (head to head)', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119381778, end_position_on_the_genomic_accession=119499233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326308, CurrentID=0, Status='live', Symbol='RPL41P4', Aliases='', description='ribosomal protein L41 pseudogene 4', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74003353, end_position_on_the_genomic_accession=74003781, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326309, CurrentID=0, Status='live', Symbol='RPL7AP5', Aliases='RPL7A_26_1382', description='ribosomal protein L7a pseudogene 5', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68338669, end_position_on_the_genomic_accession=68339549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326310, CurrentID=0, Status='live', Symbol='RPL7AP6', Aliases='RPL7A_27_1386', description='ribosomal protein L7a pseudogene 6', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69885282, end_position_on_the_genomic_accession=69886162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64167, CurrentID=0, Status='live', Symbol='ERAP2', Aliases='L-RAP, LRAP', description='endoplasmic reticulum aminopeptidase 2', other_designations='leukocyte-derived arginine aminopeptidase', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96875940, end_position_on_the_genomic_accession=96919703, orientation='plus', exon_count=20, OMIM=609497, no_hdr0=''),
  64168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64168, CurrentID=0, Status='live', Symbol='NECAB1', Aliases='EFCBP1, STIP-1', description='N-terminal EF-hand calcium binding protein 1', other_designations='EF hand calcium binding protein 1|EF-hand calcium binding protein 1|EF-hand calcium-binding protein 1|neuronal calcium binding protein|neuronal calcium-binding protein 1|synaptotagmin interacting protein 1', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90791693, end_position_on_the_genomic_accession=90959402, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64170, CurrentID=0, Status='live', Symbol='CARD9', Aliases='CANDF2, hCARD9', description='caspase recruitment domain family member 9', other_designations='caspase recruitment domain family, member 9', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136363956, end_position_on_the_genomic_accession=136373681, orientation='minus', exon_count=13, OMIM=607212, no_hdr0=''),
  326315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326315, CurrentID=0, Status='live', Symbol='PTTG4P', Aliases='', description='pituitary tumor-transforming 4 pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71085397, end_position_on_the_genomic_accession=71085876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64172, CurrentID=0, Status='live', Symbol='OSGEPL1', Aliases='Qri7', description='O-sialoglycoprotein endopeptidase-like 1', other_designations='N6-L-threonylcarbamoyladenine synthase|O-sialoglycoprotein endopeptidase like 1|O-sialoglycoprotein endopeptidase-like protein 1|probable O-sialoglycoprotein endopeptidase 2|probable tRNA threonylcarbamoyladenosine biosynthesis protein OSGEPL1|putative sialoglycoprotease type 2|t(6)A synthase|t(6)A37 threonylcarbamoyladenosine biosynthesis protein OSGEPL1|tRNA threonylcarbamoyladenosine biosynthesis protein OSGEPL1', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189746660, end_position_on_the_genomic_accession=189763198, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  326317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326317, CurrentID=0, Status='live', Symbol='RPS15AP3', Aliases='RPS15A_19_1360', description='ribosomal protein S15a pseudogene 3', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=47444071, end_position_on_the_genomic_accession=47444524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64174, CurrentID=0, Status='live', Symbol='DPEP2', Aliases='MBD2', description='dipeptidase 2', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67987390, end_position_on_the_genomic_accession=68000611, orientation='minus', exon_count=12, OMIM=609925, no_hdr0=''),
  64175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64175, CurrentID=0, Status='live', Symbol='P3H1', Aliases='GROS1, LEPRE1, OI8', description='prolyl 3-hydroxylase 1', other_designations='growth suppressor 1|leprecan|leucine proline-enriched proteoglycan (leprecan) 1|leucine- and proline-enriched proteoglycan 1|procollagen-proline 3-dioxygenase', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42746335, end_position_on_the_genomic_accession=42767084, orientation='minus', exon_count=16, OMIM=610339, no_hdr0=''),
  326320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326320, CurrentID=0, Status='live', Symbol='RPS18P2', Aliases='RPS18_7_1400', description='ribosomal protein S18 pseudogene 2', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90766495, end_position_on_the_genomic_accession=90767027, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326321, CurrentID=0, Status='live', Symbol='RPS24P2', Aliases='RPS24_8_1390', description='ribosomal protein S24 pseudogene 2', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75595729, end_position_on_the_genomic_accession=75596243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326322, CurrentID=0, Status='live', Symbol='RPS24P3', Aliases='RPS24_9_1397', description='ribosomal protein S24 pseudogene 3', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81332837, end_position_on_the_genomic_accession=81333393, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326323, CurrentID=0, Status='live', Symbol='RPS27AP4', Aliases='RPS27A_7_1343', description='ribosomal protein S27a pseudogene 4', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27258686, end_position_on_the_genomic_accession=27259192, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64180, CurrentID=0, Status='live', Symbol='DPEP3', Aliases='MBD3', description='dipeptidase 3', other_designations='membrane-bound dipeptidase 3|metallopeptidase (family M19)', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67975663, end_position_on_the_genomic_accession=67980549, orientation='minus', exon_count=10, OMIM=609926, no_hdr0=''),
  326325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326325, CurrentID=0, Status='live', Symbol='RPS2P2', Aliases='RPS2_17_1389', description='ribosomal protein S2 pseudogene 2', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74377675, end_position_on_the_genomic_accession=74378623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326326, CurrentID=0, Status='live', Symbol='RPS2P3', Aliases='RPS2_19_1406', description='ribosomal protein S2 pseudogene 3', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99974290, end_position_on_the_genomic_accession=99975033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64184, CurrentID=0, Status='live', Symbol='EDDM3B', Aliases='EP3B, FAM12B, HE3-BETA, HE3B, RAM2', description='epididymal protein 3B', other_designations='epididymal secretory protein E3 beta|family with sequence similarity 12, member B (epididymal)|human epididymis-specific 3 beta|human epididymis-specific protein 3-beta|ribonuclease A M2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20768427, end_position_on_the_genomic_accession=20770948, orientation='plus', exon_count=2, OMIM=611582, no_hdr0=''),
  105370741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370741, CurrentID=0, Status='live', Symbol='LOC105370741', Aliases='', description='uncharacterized LOC105370741', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27418796, end_position_on_the_genomic_accession=27421743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326337, CurrentID=0, Status='live', Symbol='SSXP4', Aliases='SSX10, psiSSX4', description='SSX family pseudogene 4', other_designations='SSX4 pseudogene|synovial sarcoma, X breakpoint 10|synovial sarcoma, X breakpoint pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52597942, end_position_on_the_genomic_accession=52599441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326339, CurrentID=0, Status='live', Symbol='SSXP5', Aliases='psiSSX5', description='SSX family pseudogene 5', other_designations='SSX5 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52672522, end_position_on_the_genomic_accession=52678581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326340, CurrentID=0, Status='live', Symbol='ZAR1', Aliases='Z3CXXC6', description='zygote arrest 1', other_designations='oocyte-specific maternal effect factor|zinc finger, 3CxxC-type 6', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48490252, end_position_on_the_genomic_accession=48498122, orientation='plus', exon_count=5, OMIM=607520, no_hdr0=''),
  326341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326341, CurrentID=0, Status='live', Symbol='SSXP3', Aliases='psiSSX3', description='SSX family pseudogene 3', other_designations='SSX3 pseudogene|synovial sarcoma, X breakpoint 5 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48156380, end_position_on_the_genomic_accession=48166349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326342, CurrentID=0, Status='live', Symbol='ADGRE4P', Aliases='EMR4, EMR4P, FIRE, GPR127, PGR16', description='adhesion G protein-coupled receptor E4, pseudogene', other_designations='EGF-TM7 receptor EMR4|G protein-coupled receptor 127|egf-like module containing, mucin-like, hormone receptor-like 4 pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6952500, end_position_on_the_genomic_accession=6990846, orientation='minus', exon_count=16, OMIM=612305, no_hdr0=''),
  103611079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103611079, CurrentID=0, Status='live', Symbol='ASOBS', Aliases='', description='Asthma and obesity susceptibility, chromosome 16p11 inversion related', other_designations='', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615835, no_hdr0=''),
  103611081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103611081, CurrentID=0, Status='live', Symbol='LOC103611081', Aliases='', description='uncharacterized LOC103611081', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130866249, end_position_on_the_genomic_accession=130870247, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370743, CurrentID=0, Status='live', Symbol='LOC105370743', Aliases='', description='uncharacterized LOC105370743', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29685001, end_position_on_the_genomic_accession=29686956, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64207, CurrentID=0, Status='live', Symbol='IRF2BPL', Aliases='C14orf4, EAP1', description='interferon regulatory factor 2 binding protein-like', other_designations='enhanced at puberty 1|enhanced at puberty protein 1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77024543, end_position_on_the_genomic_accession=77028699, orientation='minus', exon_count=1, OMIM=611720, no_hdr0=''),
  64208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64208, CurrentID=0, Status='live', Symbol='POPDC3', Aliases='POP3, bA355M14.1', description='popeye domain containing 3', other_designations='popeye protein 3', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105157900, end_position_on_the_genomic_accession=105180009, orientation='minus', exon_count=8, OMIM=605824, no_hdr0=''),
  103611090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103611090, CurrentID=0, Status='live', Symbol='CDR1-AS', Aliases='CDR1NAT, CDR1as, ciRS-7', description='CDR1 antisense RNA', other_designations='CDR1 natural antisense transcript|circular RNA sponge for miR-7', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64211, CurrentID=0, Status='live', Symbol='LHX5', Aliases='', description='LIM homeobox 5', other_designations='LIM homeobox protein 5', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113462889, end_position_on_the_genomic_accession=113472072, orientation='minus', exon_count=5, OMIM=605992, no_hdr0=''),
  64215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64215, CurrentID=0, Status='live', Symbol='DNAJC1', Aliases='DNAJL1, ERdj1, HTJ1, MTJ1', description='DnaJ heat shock protein family (Hsp40) member C1', other_designations='DnaJ-like protein|dnaJ protein homolog MTJ1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21756548, end_position_on_the_genomic_accession=22003768, orientation='minus', exon_count=15, OMIM=611207, no_hdr0=''),
  64216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64216, CurrentID=0, Status='live', Symbol='TFB2M', Aliases='Hkp1, mtTFB2', description='transcription factor B2, mitochondrial', other_designations="HCV NS5A-transactivated protein 5|S-adenosylmethionine-6-N', N'-adenosyl(rRNA) dimethyltransferase 2|h-mtTFB|h-mtTFB2|hTFB2M|hepatitis C virus NS5A-transactivated protein 5|mitochondrial 12S rRNA dimethylase 2|mitochondrial transcription factor B2", map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246540561, end_position_on_the_genomic_accession=246566319, orientation='minus', exon_count=8, OMIM=607055, no_hdr0=''),
  64218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64218, CurrentID=0, Status='live', Symbol='SEMA4A', Aliases='CORD10, RP35, SEMAB, SEMB', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4A', other_designations='sema B|semaphorin-B', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156147359, end_position_on_the_genomic_accession=156177751, orientation='plus', exon_count=19, OMIM=607292, no_hdr0=''),
  64219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64219, CurrentID=0, Status='live', Symbol='PJA1', Aliases='PRAJA1, RNF70', description='praja ring finger 1, E3 ubiquitin protein ligase', other_designations='RING finger protein 70|praja 1', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69160738, end_position_on_the_genomic_accession=69165522, orientation='minus', exon_count=2, OMIM=300420, no_hdr0=''),
  64220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64220, CurrentID=0, Status='live', Symbol='STRA6', Aliases='MCOPCB8, MCOPS9, PP14296', description='stimulated by retinoic acid 6', other_designations='retinol binding protein 4 receptor|stimulated by retinoic acid 6 homolog|stimulated by retinoic acid gene 6 homolog', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74179466, end_position_on_the_genomic_accession=74212259, orientation='minus', exon_count=26, OMIM=610745, no_hdr0=''),
  64221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64221, CurrentID=0, Status='live', Symbol='ROBO3', Aliases='HGPPS, HGPS, RBIG1, RIG1', description='roundabout guidance receptor 3', other_designations='retinoblastoma inhibiting gene 1|roundabout, axon guidance receptor, homolog 3|roundabout-like protein 3', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124865006, end_position_on_the_genomic_accession=124881474, orientation='plus', exon_count=31, OMIM=608630, no_hdr0=''),
  64222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64222, CurrentID=0, Status='live', Symbol='TOR3A', Aliases='ADIR, ADIR2', description='torsin family 3, member A', other_designations='ATP-dependant interferon response protein 1|ATP-dependant interferon responsive|ATP-dependent interferon-responsive protein', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179081977, end_position_on_the_genomic_accession=179095996, orientation='plus', exon_count=6, OMIM=607555, no_hdr0=''),
  64223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64223, CurrentID=0, Status='live', Symbol='MLST8', Aliases='GBL, GbetaL, LST8, POP3, WAT1', description='MTOR associated protein, LST8 homolog', other_designations='TORC subunit LST8|gable|mammalian lethal with SEC13 protein 8|protein GbetaL', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2205177, end_position_on_the_genomic_accession=2209417, orientation='plus', exon_count=10, OMIM=612190, no_hdr0=''),
  64224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64224, CurrentID=0, Status='live', Symbol='HERPUD2', Aliases='', description='HERPUD family member 2', other_designations='homocysteine-inducible, endoplasmic reticulum stress-inducible, ubiquitin-like domain member 2', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35632659, end_position_on_the_genomic_accession=35695187, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64225, CurrentID=0, Status='live', Symbol='ATL2', Aliases='ARL3IP2, ARL6IP2, atlastin2', description='atlastin GTPase 2', other_designations='ADP-ribosylation factor-like 6 interacting protein 2|ADP-ribosylation factor-like protein 6-interacting protein 2|ADP-ribosylation-like factor 6 interacting protein 2|ARL-6-interacting protein 2|aip-2', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38294094, end_position_on_the_genomic_accession=38378139, orientation='minus', exon_count=20, OMIM=609368, no_hdr0=''),
  64231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64231, CurrentID=0, Status='live', Symbol='MS4A6A', Aliases='4SPAN3, 4SPAN3.2, CD20L3, CDA01, MS4A6, MST090, MSTP090', description='membrane-spanning 4-domains, subfamily A, member 6A', other_designations='CD20 antigen-like 3|CD20-like precusor|HAIRB-iso|MS4A6A-polymorph|four-span transmembrane protein 3|four-span transmembrane protein 3.1|four-span transmembrane protein 3.2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60171607, end_position_on_the_genomic_accession=60184666, orientation='minus', exon_count=12, OMIM=606548, no_hdr0=''),
  64232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64232, CurrentID=0, Status='live', Symbol='MS4A5', Aliases='CD20-L2, CD20L2, TETM4', description='membrane-spanning 4-domains, subfamily A, member 5', other_designations='CD20 antigen-like 2|testis-expressed transmembrane protein 4|testis-expressed transmembrane-4 protein', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60429589, end_position_on_the_genomic_accession=60447792, orientation='plus', exon_count=5, OMIM=606499, no_hdr0=''),
  105370748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370748, CurrentID=0, Status='live', Symbol='LOC105370748', Aliases='', description='golgin subfamily A member 2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30799309, end_position_on_the_genomic_accession=30800017, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  64236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64236, CurrentID=0, Status='live', Symbol='PDLIM2', Aliases='MYSTIQUE, SLIM', description='PDZ and LIM domain 2 (mystique)', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22578741, end_position_on_the_genomic_accession=22598025, orientation='plus', exon_count=11, OMIM=609722, no_hdr0=''),
  64239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64239, CurrentID=0, Status='live', Symbol='RSCIS', Aliases='', description='Radiation sensitivity/chromosome instability syndrome, autosomal dominant', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605463, no_hdr0=''),
  64240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64240, CurrentID=0, Status='live', Symbol='ABCG5', Aliases='STSL', description='ATP binding cassette subfamily G member 5', other_designations='ATP-binding cassette, subfamily G, member 5|sterolin 1|sterolin-1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43806155, end_position_on_the_genomic_accession=43838865, orientation='minus', exon_count=15, OMIM=605459, no_hdr0=''),
  64241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64241, CurrentID=0, Status='live', Symbol='ABCG8', Aliases='GBD4, STSL', description='ATP binding cassette subfamily G member 8', other_designations='ATP-binding cassette, subfamily G, member 8|sterolin 2|sterolin-2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43828954, end_position_on_the_genomic_accession=43880189, orientation='plus', exon_count=15, OMIM=605460, no_hdr0=''),
  100421022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421022, CurrentID=0, Status='live', Symbol='LOC100421022', Aliases='', description='tigger transposable element derived 6 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87334612, end_position_on_the_genomic_accession=87335589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370751, CurrentID=0, Status='live', Symbol='LOC105370751', Aliases='', description='uncharacterized LOC105370751', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30365902, end_position_on_the_genomic_accession=30372969, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731605, CurrentID=0, Status='live', Symbol='BCLAF1P2', Aliases='', description='BCL2-associated transcription factor 1 pseudogene 2', other_designations='bcl-2-associated transcription factor 1 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34265826, end_position_on_the_genomic_accession=34268832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103611155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103611155, CurrentID=0, Status='live', Symbol='PAPPA-AS2', Aliases='AGU1', description='PAPPA antisense RNA 2', other_designations='adipogenesis up-regulated transcript 1', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103611157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103611157, CurrentID=0, Status='live', Symbol='TGFB2-OT1', Aliases='', description='TGFB2 overlapping transcript 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218442626, end_position_on_the_genomic_accession=218443995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  64282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64282, CurrentID=0, Status='live', Symbol='PAPD5', Aliases='TRF4-2', description='PAP associated domain containing 5', other_designations='PAP-associated domain-containing protein 5|TUTase 3|terminal uridylyltransferase 3|topoisomerase-related function protein 4-2', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50152918, end_position_on_the_genomic_accession=50235310, orientation='plus', exon_count=13, OMIM=605540, no_hdr0=''),
  64283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64283, CurrentID=0, Status='live', Symbol='ARHGEF28', Aliases='RGNEF, RIP2, p190RHOGEF', description='Rho guanine nucleotide exchange factor 28', other_designations='190 kDa guanine nucleotide exchange factor|Rho guanine nucleotide exchange factor (GEF) 28|Rho interacting protein 2|p190-RhoGEF|rho-guanine nucleotide exchange factor', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73626158, end_position_on_the_genomic_accession=73941993, orientation='plus', exon_count=44, OMIM=612790, no_hdr0=''),
  64284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64284, CurrentID=0, Status='live', Symbol='RAB17', Aliases='', description='RAB17, member RAS oncogene family', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237574322, end_position_on_the_genomic_accession=237591126, orientation='minus', exon_count=6, OMIM=602206, no_hdr0=''),
  64285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64285, CurrentID=0, Status='live', Symbol='RHBDF1', Aliases='C16orf8, Dist1, EGFR-RS, gene-89, gene-90, hDist1', description='rhomboid 5 homolog 1 (Drosophila)', other_designations='epidermal growth factor receptor, related sequence|epidermal growth factor receptor-related protein|iRhom1|p100hRho|rhomboid family 1|rhomboid family member 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58059, end_position_on_the_genomic_accession=76355, orientation='minus', exon_count=19, OMIM=614403, no_hdr0=''),
  64288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64288, CurrentID=0, Status='live', Symbol='ZSCAN31', Aliases='ZNF20-Lp, ZNF310P, ZNF323, dJ874C20.2', description='zinc finger and SCAN domain containing 31', other_designations='zinc finger protein 310 pseudogene|zinc finger protein 323', map_location='6p22.3-p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28312837, end_position_on_the_genomic_accession=28354506, orientation='minus', exon_count=14, OMIM=610794, no_hdr0=''),
  105373229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373229, CurrentID=0, Status='live', Symbol='LOC105373229', Aliases='', description='uncharacterized LOC105373229', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240823572, end_position_on_the_genomic_accession=240852725, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  285835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285835, CurrentID=0, Status='live', Symbol='LOC285835', Aliases='', description='uncharacterized LOC285835', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370761, CurrentID=0, Status='live', Symbol='LOC105370761', Aliases='', description='uncharacterized LOC105370761', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34498930, end_position_on_the_genomic_accession=34513743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64318, CurrentID=0, Status='live', Symbol='NOC3L', Aliases='AD24, C10orf117, FAD24', description='NOC3-like DNA replication regulator', other_designations='NOC3 protein homolog|NOC3L DNA replication regulator|factor for adipocyte differentiation 24|nucleolar complex associated 3 homolog|nucleolar complex-associated protein 3-like protein', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94314907, end_position_on_the_genomic_accession=94362976, orientation='minus', exon_count=22, OMIM=610769, no_hdr0=''),
  64319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64319, CurrentID=0, Status='live', Symbol='FBRS', Aliases='FBS, FBS1', description='fibrosin', other_designations='fibrogenic lymphokine|fibrosin 1|probable fibrosin-1 long transcript protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30658431, end_position_on_the_genomic_accession=30670810, orientation='plus', exon_count=18, OMIM=608601, no_hdr0=''),
  64320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64320, CurrentID=0, Status='live', Symbol='RNF25', Aliases='AO7', description='ring finger protein 25', other_designations='ring finger protein AO7', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218663864, end_position_on_the_genomic_accession=218672058, orientation='minus', exon_count=10, OMIM=616014, no_hdr0=''),
  64321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64321, CurrentID=0, Status='live', Symbol='SOX17', Aliases='VUR3', description='SRY-box 17', other_designations='SRY (sex determining region Y)-box 17|SRY box 17|SRY-related HMG-box transcription factor SOX17', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54457935, end_position_on_the_genomic_accession=54460896, orientation='plus', exon_count=2, OMIM=610928, no_hdr0=''),
  105370763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370763, CurrentID=0, Status='live', Symbol='LOC105370763', Aliases='', description='uncharacterized LOC105370763', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35003515, end_position_on_the_genomic_accession=35015337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64324, CurrentID=0, Status='live', Symbol='NSD1', Aliases='ARA267, KMT3B, SOTOS, SOTOS1, STO', description='nuclear receptor binding SET domain protein 1', other_designations='H3-K36-HMTase|H4-K20-HMTase|NR-binding SET domain-containing protein|androgen receptor coactivator 267 kDa protein|androgen receptor-associated coregulator 267|androgen receptor-associated protein of 267 kDa|lysine N-methyltransferase 3B|nuclear receptor-binding SET domain-containing protein 1', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177133025, end_position_on_the_genomic_accession=177300213, orientation='plus', exon_count=29, OMIM=606681, no_hdr0=''),
  64326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64326, CurrentID=0, Status='live', Symbol='RFWD2', Aliases='COP1, RNF200', description='ring finger and WD repeat domain 2, E3 ubiquitin protein ligase', other_designations='RING finger protein 200|constitutive photomorphogenesis protein 1 homolog|putative ubiquitin ligase COP1', map_location='1q25.1-q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175944826, end_position_on_the_genomic_accession=176207279, orientation='minus', exon_count=21, OMIM=608067, no_hdr0=''),
  64327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64327, CurrentID=0, Status='live', Symbol='LMBR1', Aliases='ACHP, C7orf2, DIF14, LSS, PPD2, THYP, TPT, ZRS', description='limb development membrane protein 1', other_designations='differentiation-related gene 14 protein|limb region 1 homolog', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156668946, end_position_on_the_genomic_accession=156893208, orientation='minus', exon_count=22, OMIM=605522, no_hdr0=''),
  64328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64328, CurrentID=0, Status='live', Symbol='XPO4', Aliases='exp4', description='exportin 4', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20777329, end_position_on_the_genomic_accession=20902774, orientation='minus', exon_count=26, OMIM=611449, no_hdr0=''),
  105370764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370764, CurrentID=0, Status='live', Symbol='LOC105370764', Aliases='', description='uncharacterized LOC105370764', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34989690, end_position_on_the_genomic_accession=35011182, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64332, CurrentID=0, Status='live', Symbol='NFKBIZ', Aliases='IKBZ, INAP, MAIL', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, zeta', other_designations='I-kappa-B-zeta|IL-1 inducible nuclear ankyrin-repeat protein|Ikappa B-zeta variant 3|IkappaB-zeta|ikB-zeta|ikappaBzeta|molecule possessing ankyrin repeats induced by lipopolysaccharide', map_location='3p12-q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101827990, end_position_on_the_genomic_accession=101861025, orientation='plus', exon_count=14, OMIM=608004, no_hdr0=''),
  64333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64333, CurrentID=0, Status='live', Symbol='ARHGAP9', Aliases='10C, RGL1', description='Rho GTPase activating protein 9', other_designations='rho-type GTPase-activating protein 9', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57472255, end_position_on_the_genomic_accession=57483959, orientation='minus', exon_count=22, OMIM=610576, no_hdr0=''),
  105370765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370765, CurrentID=0, Status='live', Symbol='LOC105370765', Aliases='', description='uncharacterized LOC105370765', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35079811, end_position_on_the_genomic_accession=35100420, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64342, CurrentID=0, Status='live', Symbol='HS1BP3', Aliases='ETM2, HS1-BP3', description='HCLS1 binding protein 3', other_designations='HS1-binding protein 3|HSP1BP-3', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20617804, end_position_on_the_genomic_accession=20651104, orientation='minus', exon_count=7, OMIM=609359, no_hdr0=''),
  64343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64343, CurrentID=0, Status='live', Symbol='AZI2', Aliases='AZ2, NAP1, TILP', description='5-azacytidine induced 2', other_designations='5-azacytidine induced gene 2|NAK-associated protein 1|NF-kappa-B-activating kinase-associated protein 1', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28322353, end_position_on_the_genomic_accession=28349424, orientation='minus', exon_count=11, OMIM=609916, no_hdr0=''),
  64344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64344, CurrentID=0, Status='live', Symbol='HIF3A', Aliases='HIF-3A, IPAS, MOP7, PASD7, bHLHe17', description='hypoxia inducible factor 3, alpha subunit', other_designations='HIF3-alpha-1|PAS domain-containing protein 7|basic-helix-loop-helix-PAS protein MOP7|class E basic helix-loop-helix protein 17|inhibitory PAS domain protein|member of PAS protein 7', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46297046, end_position_on_the_genomic_accession=46343433, orientation='plus', exon_count=18, OMIM=609976, no_hdr0=''),
  105370768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370768, CurrentID=0, Status='live', Symbol='LOC105370768', Aliases='', description='uncharacterized LOC105370768', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36446706, end_position_on_the_genomic_accession=36485034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64359, CurrentID=0, Status='live', Symbol='NXN', Aliases='NRX, TRG-4', description='nucleoredoxin', other_designations='nucleoredoxin 1', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=799313, end_position_on_the_genomic_accession=979767, orientation='minus', exon_count=11, OMIM=612895, no_hdr0=''),
  105370770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370770, CurrentID=0, Status='live', Symbol='LOC105370770', Aliases='', description='uncharacterized LOC105370770', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36485084, end_position_on_the_genomic_accession=36507209, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64374, CurrentID=0, Status='live', Symbol='SIL1', Aliases='BAP, MSS, ULG5', description='SIL1 nucleotide exchange factor', other_designations='BiP-associated protein|SIL1 homolog, endoplasmic reticulum chaperone|SIL1-like protein endoplasmic reticulum chaperone', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138946720, end_position_on_the_genomic_accession=139198376, orientation='minus', exon_count=12, OMIM=608005, no_hdr0=''),
  64375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64375, CurrentID=0, Status='live', Symbol='IKZF4', Aliases='EOS, ZNFN1A4', description='IKAROS family zinc finger 4', other_designations='IKAROS family zinc finger 4 (Eos)|ikaros family zinc finger protein 4|zinc finger protein, subfamily 1A, 4 (Eos)|zinc finger transcription factor Eos', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56007647, end_position_on_the_genomic_accession=56038435, orientation='plus', exon_count=13, OMIM=606239, no_hdr0=''),
  64376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64376, CurrentID=0, Status='live', Symbol='IKZF5', Aliases='PEGASUS, ZNFN1A5', description='IKAROS family zinc finger 5', other_designations='IKAROS family zinc finger 5 (Pegasus)|zinc finger protein, subfamily 1A, 5 (Pegasus)|zinc finger transcription factor Pegasus', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122990806, end_position_on_the_genomic_accession=123008850, orientation='minus', exon_count=7, OMIM=606238, no_hdr0=''),
  64377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64377, CurrentID=0, Status='live', Symbol='CHST8', Aliases='GALNAC4ST1, GalNAc4ST, PSS3', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 8', other_designations='GALNAC-4-ST1|GalNAc-4-O-sulfotransferase 1|N-acetylgalactosamine-4-O-sulfotransferase 1|galNAc4ST-1', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33621955, end_position_on_the_genomic_accession=33773509, orientation='plus', exon_count=10, OMIM=610190, no_hdr0=''),
  105370773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370773, CurrentID=0, Status='live', Symbol='LOC105370773', Aliases='', description='uncharacterized LOC105370773', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=37672769, end_position_on_the_genomic_accession=37679682, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64386, CurrentID=0, Status='live', Symbol='MMP25', Aliases='MMP-25, MMP20, MMP20A, MMPL1, MT-MMP 6, MT-MMP6, MT6-MMP, MT6MMP, MTMMP6', description='matrix metallopeptidase 25', other_designations='leukolysin|matrix metallopeptidase-like 1|matrix metalloproteinase 20|matrix metalloproteinase 25|matrix metalloproteinase-like 1|membrane-type 6 matrix metalloproteinase|membrane-type matrix metalloproteinase 6|membrane-type-6 matrix metalloproteinase', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3045963, end_position_on_the_genomic_accession=3060729, orientation='plus', exon_count=15, OMIM=608482, no_hdr0=''),
  64388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64388, CurrentID=0, Status='live', Symbol='GREM2', Aliases='CKTSF1B2, DAND3, PRDC', description='gremlin 2, DAN family BMP antagonist', other_designations='DAN domain family member 3|cysteine knot superfamily 1, BMP antagonist 2|gremlin 2, cysteine knot superfamily, homolog|protein related to DAN and cerberus', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240489573, end_position_on_the_genomic_accession=240612272, orientation='minus', exon_count=4, OMIM=608832, no_hdr0=''),
  105370774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370774, CurrentID=0, Status='live', Symbol='LOC105370774', Aliases='', description='uncharacterized LOC105370774', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38502436, end_position_on_the_genomic_accession=38515819, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106478981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106478981, CurrentID=0, Status='live', Symbol='TUBB4BP1', Aliases='', description='tubulin beta 4B class IVb pseudogene 1', other_designations='tubulin, beta 4B class IVb pseudogene 1', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64393, CurrentID=0, Status='live', Symbol='ZMAT3', Aliases='PAG608, WIG-1, WIG1', description='zinc finger, matrin-type 3', other_designations='WIG-1/PAG608 protein|p53 target zinc finger protein|p53-activated gene 608 protein|zinc finger protein WIG-1|zinc finger protein WIG1|zinc finger, matrin type 3', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179017223, end_position_on_the_genomic_accession=179072279, orientation='minus', exon_count=7, OMIM=606452, no_hdr0=''),
  64395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64395, CurrentID=0, Status='live', Symbol='GMCL1', Aliases='BTBD13, GCL, GCL1, SPATA29', description='germ cell-less, spermatogenesis associated 1', other_designations='germ cell-less homolog 1|spermatogenesis associated 29', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69829642, end_position_on_the_genomic_accession=69881386, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  64396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64396, CurrentID=0, Status='live', Symbol='GMCL1P1', Aliases='GCL, GMCL1L, GMCL2', description='germ cell-less, spermatogenesis associated 1 pseudogene 1', other_designations='germ cell-less homolog 1 pseudogene 1|germ cell-less related', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178184510, end_position_on_the_genomic_accession=178187432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  64397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64397, CurrentID=0, Status='live', Symbol='ZNF106', Aliases='SH3BP3, ZFP106, ZNF474', description='zinc finger protein 106', other_designations='SH3-domain binding protein 3|zinc finger protein 474', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42412437, end_position_on_the_genomic_accession=42491197, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  64398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64398, CurrentID=0, Status='live', Symbol='MPP5', Aliases='PALS1', description='membrane protein, palmitoylated 5', other_designations='membrane protein, palmitoylated 5 (MAGUK p55 subfamily member 5)|protein associated with lin seven 1|stardust', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67241295, end_position_on_the_genomic_accession=67336061, orientation='plus', exon_count=17, OMIM=606958, no_hdr0=''),
  64399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64399, CurrentID=0, Status='live', Symbol='HHIP', Aliases='HIP', description='hedgehog interacting protein', other_designations='', map_location='4q28-q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144645996, end_position_on_the_genomic_accession=144741400, orientation='plus', exon_count=13, OMIM=606178, no_hdr0=''),
  64400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64400, CurrentID=0, Status='live', Symbol='AKTIP', Aliases='FT1, FTS', description='AKT interacting protein', other_designations='fused toes homolog', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53491280, end_position_on_the_genomic_accession=53504411, orientation='minus', exon_count=11, OMIM=608483, no_hdr0=''),
  105370776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370776, CurrentID=0, Status='live', Symbol='LOC105370776', Aliases='', description='uncharacterized LOC105370776', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38729743, end_position_on_the_genomic_accession=38732505, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64403, CurrentID=0, Status='live', Symbol='CDH24', Aliases='CDH11L', description='cadherin 24, type 2', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23047061, end_position_on_the_genomic_accession=23057619, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  64405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64405, CurrentID=0, Status='live', Symbol='CDH22', Aliases='C20orf25, dJ998H6.1', description='cadherin 22, type 2', other_designations='PB-cadherin|cadherin-like 22|ortholog of rat PB-cadherin|pituitary and brain cadherin', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46173733, end_position_on_the_genomic_accession=46308829, orientation='minus', exon_count=15, OMIM=609920, no_hdr0=''),
  64407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64407, CurrentID=0, Status='live', Symbol='RGS18', Aliases='RGS13', description='regulator of G-protein signaling 18', other_designations='regulator of G-protein signalling 13|regulator of G-protein signalling 18', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192158462, end_position_on_the_genomic_accession=192187172, orientation='plus', exon_count=6, OMIM=607192, no_hdr0=''),
  64409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64409, CurrentID=0, Status='live', Symbol='WBSCR17', Aliases='GALNACT17, GALNT16, GALNT20, GALNTL3, GalNAc-T5L', description='Williams-Beuren syndrome chromosome region 17', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like 3|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 3|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 16|galNAc-T-like protein 3|polypeptide GalNAc transferase 3|polypeptide GalNAc transferase-like protein 3|polypeptide N-acetylgalactosaminyltransferase WBSCR17|polypeptide N-acetylgalactosaminyltransferase-like 3|pp-GaNTase-like protein 3|pp-GalNAc-transferase 20|protein-UDP acetylgalactosaminyltransferase-like protein 3|williams-Beuren syndrome chromosomal region 17 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71132414, end_position_on_the_genomic_accession=71713601, orientation='plus', exon_count=14, OMIM=615137, no_hdr0=''),
  64410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64410, CurrentID=0, Status='live', Symbol='KLHL25', Aliases='ENC-2, ENC2', description='kelch like family member 25', other_designations='BTB/POZ KELCH domain protein|ectoderm-neural cortex protein 2|ectodermal-neural cortex 2|kelch-like 25|kelch-like family member 25', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85759326, end_position_on_the_genomic_accession=85794958, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64411, CurrentID=0, Status='live', Symbol='ARAP3', Aliases='CENTD3, DRAG1', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 3', other_designations='ARF-GAP, RHO-GAP, ankyrin repeat and plekstrin homology domains-containing protein 3|Arf and Rho GAP adapter protein 3|PtdIns(3,4,5)P3-binding protein|centaurin, delta 3|centaurin-delta-3|phosphoinositide binding protein', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141653401, end_position_on_the_genomic_accession=141682233, orientation='minus', exon_count=35, OMIM=606647, no_hdr0=''),
  64412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64412, CurrentID=0, Status='live', Symbol='GZF1', Aliases='ZBTB23, ZNF336', description='GDNF-inducible zinc finger protein 1', other_designations='zinc finger and BTB domain-containing protein 23|zinc finger protein 336', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23361639, end_position_on_the_genomic_accession=23373046, orientation='plus', exon_count=9, OMIM=613842, no_hdr0=''),
  64417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64417, CurrentID=0, Status='live', Symbol='C5orf28', Aliases='', description='chromosome 5 open reading frame 28', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43444252, end_position_on_the_genomic_accession=43485178, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64418, CurrentID=0, Status='live', Symbol='TMEM168', Aliases='', description='transmembrane protein 168', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112762382, end_position_on_the_genomic_accession=112790713, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  64419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64419, CurrentID=0, Status='live', Symbol='MTMR14', Aliases='C3orf29', description='myotubularin related protein 14', other_designations='HCV NS5A-transactivated protein 4 splice variant A-binding protein 1|NS5ATP4ABP1|egg-derived tyrosine phosphatase homolog|jumpy', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9649433, end_position_on_the_genomic_accession=9702394, orientation='plus', exon_count=20, OMIM=611089, no_hdr0=''),
  64420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64420, CurrentID=0, Status='live', Symbol='SUSD1', Aliases='', description='sushi domain containing 1', other_designations='', map_location='9q31.3-q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112040781, end_position_on_the_genomic_accession=112175520, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  64421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64421, CurrentID=0, Status='live', Symbol='DCLRE1C', Aliases='A-SCID, DCLREC1C, RS-SCID, SCIDA, SNM1C', description='DNA cross-link repair 1C', other_designations='DNA cross-link repair 1C (PSO2 homolog, S. cerevisiae)|SNM1 homolog C|SNM1-like protein|severe combined immunodeficiency, type a (Athabascan)', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14904611, end_position_on_the_genomic_accession=14954432, orientation='minus', exon_count=19, OMIM=605988, no_hdr0=''),
  64422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64422, CurrentID=0, Status='live', Symbol='ATG3', Aliases='APG3, APG3-LIKE, APG3L, PC3-96', description='autophagy related 3', other_designations='2610016C12Rik|ATG3 autophagy related 3 homolog|autophagy-related protein 3|hApg3', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112532507, end_position_on_the_genomic_accession=112561963, orientation='minus', exon_count=11, OMIM=609606, no_hdr0=''),
  64423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64423, CurrentID=0, Status='live', Symbol='INF2', Aliases='C14orf151, C14orf173, CMTDIE, FSGS5, pp9484', description='inverted formin, FH2 and WH2 domain containing', other_designations='HBEAG-binding protein 2 binding protein C|HBEBP2-binding protein C', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104689606, end_position_on_the_genomic_accession=104719610, orientation='plus', exon_count=25, OMIM=610982, no_hdr0=''),
  64425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64425, CurrentID=0, Status='live', Symbol='POLR1E', Aliases='PAF53, PRAF1', description='polymerase (RNA) I polypeptide E', other_designations='DNA-directed RNA polymerase I subunit E|RNA polymerase I subunit A49|RNA polymerase I-associated factor 53|polymerase (RNA) I associated factor 1|polymerase (RNA) I polypeptide E, 53kDa', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37485935, end_position_on_the_genomic_accession=37503697, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64426, CurrentID=0, Status='live', Symbol='SUDS3', Aliases='SAP45, SDS3', description='SDS3 homolog, SIN3A corepressor complex component', other_designations='45 kDa Sin3-associated polypeptide|sin3A-associated protein, 45kDa|suppressor of defective silencing 3 protein homolog', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118373189, end_position_on_the_genomic_accession=118418035, orientation='plus', exon_count=15, OMIM=608250, no_hdr0=''),
  64427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64427, CurrentID=0, Status='live', Symbol='TTC31', Aliases='', description='tetratricopeptide repeat domain 31', other_designations='TPR repeat protein 31', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74483032, end_position_on_the_genomic_accession=74494564, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64428, CurrentID=0, Status='live', Symbol='NARFL', Aliases='HPRN, IOP1, LET1L, PRN', description='nuclear prelamin A recognition factor-like', other_designations='LET1 like/JFP15|iron-only hydrogenase-like protein 1|protein related to Narf', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=729753, end_position_on_the_genomic_accession=741038, orientation='minus', exon_count=12, OMIM=611118, no_hdr0=''),
  64429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64429, CurrentID=0, Status='live', Symbol='ZDHHC6', Aliases='DHHC-6, ZNF376', description='zinc finger, DHHC-type containing 6', other_designations='transmembrane protein H4|zinc finger DHHC domain-containing protein 6|zinc finger protein 376', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112425439, end_position_on_the_genomic_accession=112447234, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  64430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64430, CurrentID=0, Status='live', Symbol='PCNXL4', Aliases='C14orf135, FBP2', description='pecanex-like 4 (Drosophila)', other_designations='HCV F protein-binding protein 2|HCV F-interacting protein|hepatitis C virus F protein-binding protein 2|pecanex-like protein C14orf135', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60091827, end_position_on_the_genomic_accession=60134814, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  64431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64431, CurrentID=0, Status='live', Symbol='ACTR6', Aliases='ARP6, CDA12, HSPC281, MSTP136, hARP6, hARPX', description='ARP6 actin-related protein 6 homolog (yeast)', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100200087, end_position_on_the_genomic_accession=100224424, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  64432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64432, CurrentID=0, Status='live', Symbol='MRPS25', Aliases='MRP-S25, RPMS25', description='mitochondrial ribosomal protein S25', other_designations='S25mt|mitochondrial 28S ribosomal protein S25', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15042251, end_position_on_the_genomic_accession=15065311, orientation='minus', exon_count=6, OMIM=611987, no_hdr0=''),
  64433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64433, CurrentID=0, Status='live', Symbol='LINC00244', Aliases='C7orf4, NCRNA00244', description='long intergenic non-protein coding RNA 244', other_designations='Putative uncharacterized protein C7orf4', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156540491, end_position_on_the_genomic_accession=156541101, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  64434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64434, CurrentID=0, Status='live', Symbol='NOM1', Aliases='C7orf3, PPP1R113, SGD1', description='nucleolar protein with MIF4G domain 1', other_designations='SGD1 homolog|protein phosphatase 1, regulatory subunit 113', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156949723, end_position_on_the_genomic_accession=156973182, orientation='plus', exon_count=12, OMIM=611269, no_hdr0=''),
  105370782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370782, CurrentID=0, Status='live', Symbol='LOC105370782', Aliases='', description='uncharacterized LOC105370782', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38885146, end_position_on_the_genomic_accession=38894736, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105274375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105274375, CurrentID=0, Status='live', Symbol='TEMPS', Aliases='', description='Temple syndrome', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616222, no_hdr0=''),
  326585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326585, CurrentID=0, Status='live', Symbol='SSXP8', Aliases='psiSSX8', description='SSX family pseudogene 8', other_designations='psiSSX8 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48276592, end_position_on_the_genomic_accession=48279351, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326586, CurrentID=0, Status='live', Symbol='SSXP9', Aliases='psiSSX9', description='SSX family pseudogene 9', other_designations='SSX9 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48322193, end_position_on_the_genomic_accession=48328372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326587, CurrentID=0, Status='live', Symbol='SSXP6', Aliases='psiSSX6', description='SSX family pseudogene 6', other_designations='SSX6 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48433201, end_position_on_the_genomic_accession=48433274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64446, CurrentID=0, Status='live', Symbol='DNAI2', Aliases='CILD9, DIC2', description='dynein, axonemal, intermediate chain 2', other_designations='dynein, axonemal, intermediate polypeptide 2', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74274232, end_position_on_the_genomic_accession=74314884, orientation='plus', exon_count=17, OMIM=605483, no_hdr0=''),
  326591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326591, CurrentID=0, Status='live', Symbol='MORF4L1P1', Aliases='MORF4LP1, MRG1', description='mortality factor 4 like 1 pseudogene 1', other_designations='mortality factor 4 like pseudogene 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220253834, end_position_on_the_genomic_accession=220254541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326593, CurrentID=0, Status='live', Symbol='MORF4L1P3', Aliases='MORF4LP3, MRG11', description='mortality factor 4 like 1 pseudogene 3', other_designations='mortality factor 4 like pseudogene 3', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14673582, end_position_on_the_genomic_accession=14674289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326594, CurrentID=0, Status='live', Symbol='MORF4L1P2', Aliases='MORF4LP2, MORF4LP4, MRG5', description='mortality factor 4 like 1 pseudogene 2', other_designations='mortality factor 4 like pseudogene 2|mortality factor 4 like pseudogene 4', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12198897, end_position_on_the_genomic_accession=12200866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326596, CurrentID=0, Status='live', Symbol='MAD2L1P1', Aliases='MAD2L1P', description='MAD2 mitotic arrest deficient-like 1 (yeast) pseudogene 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60779865, end_position_on_the_genomic_accession=60781267, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127905, CurrentID=0, Status='live', Symbol='SMIM11P1', Aliases='C6orf161, FAM165A, bA33E24.4', description='small integral membrane protein 11 pseudogene 1', other_designations='family with sequence similarity 165, member A|family with sequence similarity 165, member B pseudogene', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85735209, end_position_on_the_genomic_accession=85736068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326603, CurrentID=0, Status='live', Symbol='MTCO1P2', Aliases='', description='MT-CO1 pseudogene 2', other_designations='MTCO1 pseudogene 2|cytochrome c oxidase pseudogene 2', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32484093, end_position_on_the_genomic_accession=32485122, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326605, CurrentID=0, Status='live', Symbol='NCOA4P1', Aliases='NCOA4P', description='nuclear receptor coactivator 4 pseudogene 1', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65878033, end_position_on_the_genomic_accession=65881067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326606, CurrentID=0, Status='live', Symbol='NDUFB8P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B8 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, 19kDa pseudogene|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, pseudogene 1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73631456, end_position_on_the_genomic_accession=73632121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326607, CurrentID=0, Status='live', Symbol='NMNAT1P1', Aliases='NMNATP', description='NMNAT1 pseudogene 1', other_designations='nicotinamide nucleotide adenylyltransferase 1 pseudogene|nicotinamide nucleotide adenylyltransferase pseudogene', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81032475, end_position_on_the_genomic_accession=81033541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326608, CurrentID=0, Status='live', Symbol='NPM1P20', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 20', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101842761, end_position_on_the_genomic_accession=101844130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326609, CurrentID=0, Status='live', Symbol='NT5CP1', Aliases='', description="5',3'-nucleotidase, cytosolic pseudogene 1", other_designations="5', 3'-nucleotidase, cytosolic pseudogene", map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73570997, end_position_on_the_genomic_accession=73571652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326611, CurrentID=0, Status='live', Symbol='NT5CP2', Aliases='', description="5',3'-nucleotidase, cytosolic pseudogene 2", other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73538975, end_position_on_the_genomic_accession=73539832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326613, CurrentID=0, Status='live', Symbol='NUTF2P2', Aliases='', description='nuclear transport factor 2 pseudogene 2', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35819758, end_position_on_the_genomic_accession=35820609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326614, CurrentID=0, Status='live', Symbol='PEBP1P1', Aliases='PBPP1', description='phosphatidylethanolamine binding protein 1 pseudogene 1', other_designations='prostatic binding protein pseudogene 1', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96422928, end_position_on_the_genomic_accession=96424297, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326615, CurrentID=0, Status='live', Symbol='MED15P1', Aliases='MED15P, PCQAPP', description='mediator complex subunit 15 pseudogene 1', other_designations='PCQAP pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19485315, end_position_on_the_genomic_accession=19489687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326616, CurrentID=0, Status='live', Symbol='PDLIM1P1', Aliases='PDLIM1P', description='PDZ and LIM domain 1 pseudogene 1', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50050289, end_position_on_the_genomic_accession=50051686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326617, CurrentID=0, Status='live', Symbol='PSMA3P', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 3 pseudogene', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60179324, end_position_on_the_genomic_accession=60180133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64478, CurrentID=0, Status='live', Symbol='CSMD1', Aliases='PPP1R24', description='CUB and Sushi multiple domains 1', other_designations='protein phosphatase 1, regulatory subunit 24', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2935353, end_position_on_the_genomic_accession=4994806, orientation='minus', exon_count=72, OMIM=608397, no_hdr0=''),
  105370789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370789, CurrentID=0, Status='live', Symbol='LOC105370789', Aliases='', description='uncharacterized LOC105370789', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40962503, end_position_on_the_genomic_accession=40974961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326624, CurrentID=0, Status='live', Symbol='RAB37', Aliases='', description='RAB37, member RAS oncogene family', other_designations='RAB37, member of RAS oncogene family', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74671117, end_position_on_the_genomic_accession=74747335, orientation='plus', exon_count=13, OMIM=609956, no_hdr0=''),
  326625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326625, CurrentID=0, Status='live', Symbol='MMAB', Aliases='ATR, CFAP23, cblB, cob', description='methylmalonic aciduria (cobalamin deficiency) cblB type', other_designations='ATP:cob(I)alamin adenosyltransferase|ATP:corrinoid adenosyltransferase|aquocob(I)alamin vitamin B12s adenosyltransferase|cilia and flagella associated protein 23|methylmalonic aciduria type B protein', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109553715, end_position_on_the_genomic_accession=109573553, orientation='minus', exon_count=13, OMIM=607568, no_hdr0=''),
  326626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326626, CurrentID=0, Status='live', Symbol='PTMAP7', Aliases='', description='prothymosin, alpha pseudogene 7', other_designations='prothymosin a14', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92026422, end_position_on_the_genomic_accession=92027567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326627, CurrentID=0, Status='live', Symbol='RANBP20P', Aliases='', description='RAN binding protein 20 pseudogene', other_designations='exportin 6 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20720805, end_position_on_the_genomic_accession=20723141, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326628, CurrentID=0, Status='live', Symbol='RPA2P1', Aliases='RPA2P', description='replication protein A2 pseudogene 1', other_designations='replication protein A2, 32kDa pseudogene', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46996379, end_position_on_the_genomic_accession=46997805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127910, CurrentID=0, Status='live', Symbol='LOC100127910', Aliases='', description='uncharacterized LOC100127910', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=170636347, end_position_on_the_genomic_accession=170638593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  415056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=415056, CurrentID=0, Status='live', Symbol='CNTFR-AS1', Aliases='', description='CNTFR antisense RNA 1', other_designations='', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34568012, end_position_on_the_genomic_accession=34583072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370791, CurrentID=0, Status='live', Symbol='LOC105370791', Aliases='', description='uncharacterized LOC105370791', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41609414, end_position_on_the_genomic_accession=41621165, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64493, CurrentID=0, Status='live', Symbol='LINC00235', Aliases='C16orf10, LA16c-356B8.1, NCRNA00235', description='long intergenic non-protein coding RNA 235', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=526847, end_position_on_the_genomic_accession=527407, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132972, CurrentID=0, Status='live', Symbol='LOC100132972', Aliases='', description='synaptophysin-like 1 pseudogene', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27351861, end_position_on_the_genomic_accession=27352576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64499, CurrentID=0, Status='live', Symbol='TPSB2', Aliases='TPS2, tryptaseB, tryptaseC', description='tryptase beta 2 (gene/pseudogene)', other_designations='mast cell tryptase beta II|mast cell tryptase beta III', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1228336, end_position_on_the_genomic_accession=1230184, orientation='minus', exon_count=6, OMIM=191081, no_hdr0=''),
  644265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644265, CurrentID=0, Status='live', Symbol='LOC644265', Aliases='', description='capping protein (actin filament) muscle Z-line, beta pseudogene', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=103623554, end_position_on_the_genomic_accession=103624172, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128160, CurrentID=0, Status='live', Symbol='LOC100128160', Aliases='', description='uncharacterized LOC100128160', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=70246077, end_position_on_the_genomic_accession=70264477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64506, CurrentID=0, Status='live', Symbol='CPEB1', Aliases='CPE-BP1, CPEB, CPEB-1, h-CPEB, hCPEB-1', description='cytoplasmic polyadenylation element binding protein 1', other_designations='CPE-binding protein 1', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82543201, end_position_on_the_genomic_accession=82648016, orientation='minus', exon_count=14, OMIM=607342, no_hdr0=''),
  10877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10877, CurrentID=0, Status='live', Symbol='CFHR4', Aliases='CFHL4, FHR-4, FHR4', description='complement factor H-related 4', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196888014, end_position_on_the_genomic_accession=196918812, orientation='plus', exon_count=14, OMIM=605337, no_hdr0=''),
  64518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64518, CurrentID=0, Status='live', Symbol='TEKT3', Aliases='', description='tektin 3', other_designations='testicular microtubules-related protein', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15303811, end_position_on_the_genomic_accession=15341889, orientation='minus', exon_count=10, OMIM=612683, no_hdr0=''),
  100127917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127917, CurrentID=0, Status='live', Symbol='LOC100127917', Aliases='', description='heat shock protein 90kDa alpha family class B member 1 pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85731293, end_position_on_the_genomic_accession=85732031, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127919, CurrentID=0, Status='live', Symbol='LOC100127919', Aliases='', description='developmental pluripotency associated 3 pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93035148, end_position_on_the_genomic_accession=93036231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370398, CurrentID=0, Status='live', Symbol='LOC105370398', Aliases='', description='uncharacterized LOC105370398', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21046478, end_position_on_the_genomic_accession=21049742, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370802, CurrentID=0, Status='live', Symbol='LOC105370802', Aliases='', description='uncharacterized LOC105370802', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45926829, end_position_on_the_genomic_accession=45931069, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731656, CurrentID=0, Status='live', Symbol='LINC01348', Aliases='', description='long intergenic non-protein coding RNA 1348', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235065474, end_position_on_the_genomic_accession=235069818, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370804, CurrentID=0, Status='live', Symbol='LOC105370804', Aliases='', description='uncharacterized LOC105370804', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47599570, end_position_on_the_genomic_accession=47607012, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100128610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128610, CurrentID=0, Status='live', Symbol='LOC100128610', Aliases='', description='putative UPF0607 protein ENSP00000383144 pseudogene', other_designations='UPF0607 protein ENSP00000383144-like pseudogene', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63229089, end_position_on_the_genomic_accession=63230111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370805, CurrentID=0, Status='live', Symbol='LOC105370805', Aliases='', description='uncharacterized LOC105370805', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47884365, end_position_on_the_genomic_accession=48005833, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  64577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64577, CurrentID=0, Status='live', Symbol='ALDH8A1', Aliases='ALDH12, DJ352A20.2', description='aldehyde dehydrogenase 8 family member A1', other_designations='aldehyde dehydrogenase 12|aldehyde dehydrogenase 8 family, member A1|aldehyde dehydrogenase family protein', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134917390, end_position_on_the_genomic_accession=134950122, orientation='minus', exon_count=7, OMIM=606467, no_hdr0=''),
  64579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64579, CurrentID=0, Status='live', Symbol='NDST4', Aliases='N-HSST, N-HSST 4, NDST-4, NHSST4', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 4', other_designations='N-heparan sulfate sulfotransferase 4|glucosaminyl N-deacetylase/N-sulfotransferase 4', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114827775, end_position_on_the_genomic_accession=115113876, orientation='minus', exon_count=14, OMIM=615039, no_hdr0=''),
  64581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64581, CurrentID=0, Status='live', Symbol='CLEC7A', Aliases='BGR, CANDF4, CD369, CLECSF12, DECTIN1', description='C-type lectin domain family 7 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 12|C-type lectin domain family 7, member A|C-type lectin superfamily member 12|DC-associated C-type lectin 1|beta-glucan receptor|dectin-1|dendritic cell-associated C-type lectin 1|dendritic cell-associated C-type lectin-1|lectin-like receptor 1', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10116777, end_position_on_the_genomic_accession=10130273, orientation='minus', exon_count=9, OMIM=606264, no_hdr0=''),
  64582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64582, CurrentID=0, Status='live', Symbol='GPR135', Aliases='HUMNPIIY20', description='G protein-coupled receptor 135', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59452773, end_position_on_the_genomic_accession=59465396, orientation='minus', exon_count=7, OMIM=607970, no_hdr0=''),
  64585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64585, CurrentID=0, Status='live', Symbol='BOS2', Aliases='', description='Branchiootic syndrome 2', other_designations='', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=120502, no_hdr0=''),
  64591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64591, CurrentID=0, Status='live', Symbol='TSPY2', Aliases='TSPYQ1', description='testis specific protein, Y-linked 2', other_designations='testis-specific Y-encoded protein Q1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6246223, end_position_on_the_genomic_accession=6249018, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64593, CurrentID=0, Status='live', Symbol='RBMY3AP', Aliases='RBMY3P, RBMY4P, SPATA14', description='RNA binding motif protein, Y-linked, family 3, member A pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 3 pseudogene|RNA binding motif protein, Y chromosome, family 4 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9610721, end_position_on_the_genomic_accession=9615153, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64595, CurrentID=0, Status='live', Symbol='TTTY15', Aliases='NCRNA00138', description='testis-specific transcript, Y-linked 15 (non-protein coding)', other_designations='', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12662367, end_position_on_the_genomic_accession=12692224, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64596, CurrentID=0, Status='live', Symbol='PSMA6P1', Aliases='PSMA6-LIKE, PSMA6P, PSMA6Y', description='proteasome (prosome, macropain) subunit, alpha type, 6 pseudogene 1', other_designations='PSMA6 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13286561, end_position_on_the_genomic_accession=13287546, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64598, CurrentID=0, Status='live', Symbol='MOSPD3', Aliases='CDS3, NET30', description='motile sperm domain containing 3', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100612102, end_position_on_the_genomic_accession=100615384, orientation='plus', exon_count=8, OMIM=609125, no_hdr0=''),
  64599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64599, CurrentID=0, Status='live', Symbol='GIGYF1', Aliases='GYF1, PERQ1', description='GRB10 interacting GYF protein 1', other_designations='GRB10-interacting GYF protein 1|GYF domain containing 1|PERQ amino acid rich, with GYF domain 1|postmeiotic segregation increased 2-like 12', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100679507, end_position_on_the_genomic_accession=100694268, orientation='minus', exon_count=29, OMIM=612064, no_hdr0=''),
  64600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64600, CurrentID=0, Status='live', Symbol='PLA2G2F', Aliases='', description='phospholipase A2 group IIF', other_designations='GIIF sPLA2|phosphatidylcholine 2-acylhydrolase 2F|phospholipase A2, group IIF|sPLA2-IIF', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20139326, end_position_on_the_genomic_accession=20150386, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64601, CurrentID=0, Status='live', Symbol='VPS16', Aliases='hVPS16', description='VPS16, CORVET/HOPS core subunit', other_designations='VPS16 CORVET/HOPS core subunit|vacuolar protein sorting 16 homolog|vacuolar protein sorting protein 16', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2840697, end_position_on_the_genomic_accession=2866732, orientation='plus', exon_count=24, OMIM=608550, no_hdr0=''),
  105370813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370813, CurrentID=0, Status='live', Symbol='LOC105370813', Aliases='', description='uncharacterized LOC105370813', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50807548, end_position_on_the_genomic_accession=50810059, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127934, CurrentID=0, Status='live', Symbol='LOC100127934', Aliases='', description='meiotic nuclear divisions 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92931579, end_position_on_the_genomic_accession=92932088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64641, CurrentID=0, Status='live', Symbol='EBF2', Aliases='COE2, EBF-2, O/E-3, OE-3', description='early B-cell factor 2', other_designations='Collier, Olf and EBF 2|OLF-1/EBF-LIKE 3|metencephalon-mesencephalnon-olfactory transcription factor 1', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25841730, end_position_on_the_genomic_accession=26045124, orientation='minus', exon_count=16, OMIM=609934, no_hdr0=''),
  64644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64644, CurrentID=0, Status='live', Symbol='GINGF2', Aliases='GGF2, HGF2', description='gingival fibromatosis, hereditary, 2', other_designations='', map_location='5q13-q22', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605544, no_hdr0=''),
  64645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64645, CurrentID=0, Status='live', Symbol='MFSD14A', Aliases='HIAT1', description='major facilitator superfamily domain containing 14A', other_designations='hippocampus abundant gene transcript 1|putative tetracycline transporter-like protein|tetracycline transporter-like protein', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100038078, end_position_on_the_genomic_accession=100083377, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64648, CurrentID=0, Status='live', Symbol='SPANXD', Aliases='CT11.4, SPANX-C, SPANX-D, SPANX-E, SPANXE, dJ171K16.1', description='SPANX family member D', other_designations='SPANX family, member D|SPANX family, member E|cancer/testis antigen 11.4|cancer/testis antigen family 11, member 4|nuclear-associated protein SPAN-Xd|nuclear-associated protein SPAN-Xe|sperm protein associated with the nucleus, X chromosome, family member D|sperm protein associated with the nucleus, X chromosome, family member E', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141697401, end_position_on_the_genomic_accession=141698739, orientation='minus', exon_count=2, OMIM='300670,300671', no_hdr0=''),
  64651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64651, CurrentID=0, Status='live', Symbol='CSRNP1', Aliases='AXUD1, CSRNP-1, FAM130B, TAIP-3, URAX1', description='cysteine-serine-rich nuclear protein 1', other_designations='AXIN1 up-regulated 1|TGF-beta induced apoptosis protein 3|TGF-beta-induced apoptosis protein 3|axin-1 up-regulated gene 1 protein', map_location='3p22', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39141851, end_position_on_the_genomic_accession=39154597, orientation='minus', exon_count=7, OMIM=606458, no_hdr0=''),
  100127938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127938, CurrentID=0, Status='live', Symbol='LOC100127938', Aliases='', description='KIAA1191 pseudogene', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101594077, end_position_on_the_genomic_accession=101594931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64663, CurrentID=0, Status='live', Symbol='SPANXC', Aliases='C, CT11.3, CTp11, SPANX-C', description='SPANX family member C', other_designations='SPAN-Xc protein|SPANX family, member C|cancer-testis-associated protein CTp11|cancer/testis antigen 11.3|cancer/testis antigen family 11, member 3|cancer/testis-associated protein CTp11|cancer/testis-associated protein of 11 kD|nuclear-associated protein SPAN-Xc|sperm protein associated with the nucleus, X chromosome, family member C', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141241463, end_position_on_the_genomic_accession=141242550, orientation='minus', exon_count=2, OMIM=300330, no_hdr0=''),
  105370820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370820, CurrentID=0, Status='live', Symbol='LOC105370820', Aliases='', description='uncharacterized LOC105370820', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52225501, end_position_on_the_genomic_accession=52237475, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616285, CurrentID=0, Status='live', Symbol='MIR4645', Aliases='mir-4645', description='microRNA 4645', other_designations='hsa-mir-4645', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2854031, end_position_on_the_genomic_accession=2854107, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  64682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64682, CurrentID=0, Status='live', Symbol='ANAPC1', Aliases='APC1, MCPR, TSG24', description='anaphase promoting complex subunit 1', other_designations='anaphase-promoting complex 1 (meiotic checkpoint regulator)|cyclosome subunit 1|mitotic checkpoint regulator|testis-specific gene 24 protein', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111767637, end_position_on_the_genomic_accession=111884186, orientation='minus', exon_count=49, OMIM=608473, no_hdr0=''),
  105370823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370823, CurrentID=0, Status='live', Symbol='LOC105370823', Aliases='', description='uncharacterized LOC105370823', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53082851, end_position_on_the_genomic_accession=53103770, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  64689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64689, CurrentID=0, Status='live', Symbol='GORASP1', Aliases='GOLPH5, GRASP65, P65', description='golgi reassembly stacking protein 1', other_designations='Golgi peripheral membrane protein p65|Golgi phosphoprotein 5|Golgi reassembly and stacking protein 1|golgi reassembly stacking protein 1, 65kDa|golgi reassembly-stacking protein of 65 kDa', map_location='3p22-p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39096599, end_position_on_the_genomic_accession=39108363, orientation='minus', exon_count=12, OMIM=606867, no_hdr0=''),
  64693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64693, CurrentID=0, Status='live', Symbol='CTAGE1', Aliases='CT21.1, CT21.2, CTAGE, CTAGE-1, CTAGE-2', description='cutaneous T-cell lymphoma-associated antigen 1', other_designations='cancer/testis antigen 21.2|cancer/testis antigen family 21, member 1|cancer/testis antigen family 21, member 2|cutaneous T-cell lymphoma-associated antigen 2|protein cTAGE-2', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22413601, end_position_on_the_genomic_accession=22417915, orientation='minus', exon_count=1, OMIM=608856, no_hdr0=''),
  64695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64695, CurrentID=0, Status='live', Symbol='SLEB3', Aliases='', description='systemic lupus erythematosus susceptibility 3', other_designations='', map_location='4p16-p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605480, no_hdr0=''),
  105370403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370403, CurrentID=0, Status='live', Symbol='LOC105370403', Aliases='', description='uncharacterized LOC105370403', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22167738, end_position_on_the_genomic_accession=22175228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64699, CurrentID=0, Status='live', Symbol='TMPRSS3', Aliases='DFNB10, DFNB8, ECHOS1, TADG12', description='transmembrane protease, serine 3', other_designations='serine protease TADG-12|tumor-associated differentially-expressed gene 12 protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42371887, end_position_on_the_genomic_accession=42396846, orientation='minus', exon_count=13, OMIM=605511, no_hdr0=''),
  105370826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370826, CurrentID=0, Status='live', Symbol='LOC105370826', Aliases='', description='uncharacterized LOC105370826', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53445721, end_position_on_the_genomic_accession=53514089, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100127947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127947, CurrentID=0, Status='live', Symbol='LOC100127947', Aliases='', description='uncharacterized LOC100127947', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=36789103, end_position_on_the_genomic_accession=36794822, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  64708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64708, CurrentID=0, Status='live', Symbol='COPS7B', Aliases='CSN7B, SGN7b', description='COP9 signalosome subunit 7B', other_designations='COP9 constitutive photomorphogenic homolog subunit 7B|JAB1-containing signalosome subunit 7b|signalosome subunit 7b', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231781671, end_position_on_the_genomic_accession=231809254, orientation='plus', exon_count=12, OMIM=616010, no_hdr0=''),
  64710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64710, CurrentID=0, Status='live', Symbol='NUCKS1', Aliases='JC7, NUCKS', description='nuclear casein kinase and cyclin-dependent kinase substrate 1', other_designations='P1|nuclear ubiquitous casein kinase and cyclin-dependent kinase substrate|potential LAG1 interactor', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205712819, end_position_on_the_genomic_accession=205750244, orientation='minus', exon_count=7, OMIM=611912, no_hdr0=''),
  64711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64711, CurrentID=0, Status='live', Symbol='HS3ST6', Aliases='3-OST-6, HS3ST5', description='heparan sulfate-glucosamine 3-sulfotransferase 6', other_designations='heparan sulfate (glucosamine) 3-O-sulfotransferase 5|heparan sulfate (glucosamine) 3-O-sulfotransferase 6|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 6|heparan sulphate D-glucosaminyl 3-O-sulfotransferase-3B like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1911463, end_position_on_the_genomic_accession=1921422, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64714, CurrentID=0, Status='live', Symbol='PDIA2', Aliases='PDA2, PDI, PDIP, PDIR', description='protein disulfide isomerase family A member 2', other_designations='Rho GDP dissociation inhibitor gamma|pancreas-specific protein disulfide isomerase|pancreatic protein disulfide isomerase|protein disulfide isomerase A2|protein disulfide isomerase family A, member 2|protein disulfide isomerase, pancreatic|protein disulfide isomerase-associated 2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=283118, end_position_on_the_genomic_accession=287209, orientation='plus', exon_count=11, OMIM=608012, no_hdr0=''),
  104266956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266956, CurrentID=0, Status='live', Symbol='LINC01044', Aliases='', description='long intergenic non-protein coding RNA 1044', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112322567, end_position_on_the_genomic_accession=112331225, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64717, CurrentID=0, Status='live', Symbol='TRG-CCC2-1', Aliases='TRG4, TRNAG4', description='transfer RNA-Gly (CCC) 2-1', other_designations='tRNA glycine 4|tRNA-Gly-CCC-2-1|transfer RNA glycine 4 (anticodon CCC)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=636736, end_position_on_the_genomic_accession=636806, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64718, CurrentID=0, Status='live', Symbol='UNKL', Aliases='C16orf28, ZC3H5L, ZC3HDC5L', description='unkempt family zinc finger-like', other_designations='RING finger protein unkempt-like|unkempt homolog-like|zinc finger CCCH domain-containing protein 5-like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1363205, end_position_on_the_genomic_accession=1414749, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  104266959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266959, CurrentID=0, Status='live', Symbol='LINC01195', Aliases='', description='long intergenic non-protein coding RNA 1195', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9446013, end_position_on_the_genomic_accession=9455505, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  104266960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266960, CurrentID=0, Status='live', Symbol='LINC01201', Aliases='TCONS_00017091', description='long intergenic non-protein coding RNA 1201', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131016469, end_position_on_the_genomic_accession=131058146, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104266961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266961, CurrentID=0, Status='live', Symbol='LINC01221', Aliases='TCONS_00000366', description='long intergenic non-protein coding RNA 1221', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199016133, end_position_on_the_genomic_accession=199076735, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104266962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266962, CurrentID=0, Status='live', Symbol='LINC01228', Aliases='lincRNA-DYNLRB2-2', description='long intergenic non-protein coding RNA 1228', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104266963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266963, CurrentID=0, Status='live', Symbol='LINC01264', Aliases='', description='long intergenic non-protein coding RNA 1264', other_designations='RP11-124O11.2', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42979017, end_position_on_the_genomic_accession=42981507, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64725, CurrentID=0, Status='live', Symbol='RPS3AP2', Aliases='RPS3A_21_1466', description='ribosomal protein S3A pseudogene 2', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1477983, end_position_on_the_genomic_accession=1478590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64726, CurrentID=0, Status='live', Symbol='RPS20P2', Aliases='', description='ribosomal protein S20 pseudogene 2', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1328626, end_position_on_the_genomic_accession=1329809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285849, CurrentID=0, Status='live', Symbol='COX6A1P2', Aliases='', description='cytochrome c oxidase subunit VIa polypeptide 1 pseudogene 2', other_designations='ytochrome c oxidase subunit VIa polypeptide 1 pseudogene 2', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37044831, end_position_on_the_genomic_accession=37045382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64730, CurrentID=0, Status='live', Symbol='TJP1P', Aliases='', description='tight junction protein 1 pseudogene', other_designations='tight junction protein 1 (zona occludens 1) pseudogene|tight junction protein 1 (zonula occludens 1) pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1371113, end_position_on_the_genomic_accession=1371883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370831, CurrentID=0, Status='live', Symbol='LOC105370831', Aliases='', description='uncharacterized LOC105370831', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56479587, end_position_on_the_genomic_accession=56483837, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  678656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=678656, CurrentID=0, Status='live', Symbol='ZRF1PS', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C2 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 2 pseudogene|zuotin related factor 1 pseudogene', map_location='6p21.31-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38762856, end_position_on_the_genomic_accession=38766883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64735, CurrentID=0, Status='live', Symbol='LINC00254', Aliases='C16orf32, NCRNA00254', description='long intergenic non-protein coding RNA 254', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1878285, end_position_on_the_genomic_accession=1884231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370832, CurrentID=0, Status='live', Symbol='LOC105370832', Aliases='', description='uncharacterized LOC105370832', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56492180, end_position_on_the_genomic_accession=56605536, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105374769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374769, CurrentID=0, Status='live', Symbol='LOC105374769', Aliases='', description='uncharacterized LOC105374769', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64330484, end_position_on_the_genomic_accession=64332803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  195814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=195814, CurrentID=0, Status='live', Symbol='SDR16C5', Aliases='EPHD-2, RDH#2, RDH-E2, RDHE2, retSDR2', description='short chain dehydrogenase/reductase family 16C, member 5', other_designations='epidermal retinal dehydrogenase 2|retinal short chain dehydrogenase reductase|retinal short-chain dehydrogenase reductase 2|short-chain dehydrogenase/reductase family 16C member 5', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56300011, end_position_on_the_genomic_accession=56320776, orientation='minus', exon_count=8, OMIM=608989, no_hdr0=''),
  64743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64743, CurrentID=0, Status='live', Symbol='WDR13', Aliases='MG21', description='WD repeat domain 13', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48590043, end_position_on_the_genomic_accession=48605194, orientation='plus', exon_count=11, OMIM=300512, no_hdr0=''),
  64744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64744, CurrentID=0, Status='live', Symbol='SMAP2', Aliases='SMAP1L', description='small ArfGAP2', other_designations='stromal membrane-associated GTPase-activating protein 2', map_location='1p35.3-p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40344844, end_position_on_the_genomic_accession=40423326, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  64745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64745, CurrentID=0, Status='live', Symbol='METTL17', Aliases='METT11D1', description='methyltransferase like 17', other_designations='false p73 target gene protein|methyltransferase 11 domain containing 1|methyltransferase 11 domain-containing protein 1|protein RSM22 homolog, mitochondrial', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20989806, end_position_on_the_genomic_accession=20997035, orientation='plus', exon_count=13, OMIM=616091, no_hdr0=''),
  64746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64746, CurrentID=0, Status='live', Symbol='ACBD3', Aliases='GCP60, GOCAP1, GOLPH1, PAP7', description='acyl-CoA binding domain containing 3', other_designations='PBR- and PKA-associated protein 7|PKA (RIalpha)-associated protein|acyl-Coenzyme A binding domain containing 3|golgi complex associated protein 1, 60kDa|golgi phosphoprotein 1|peripheral benzodiazepine receptor-associated protein PAP7', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226144679, end_position_on_the_genomic_accession=226186722, orientation='minus', exon_count=8, OMIM=606809, no_hdr0=''),
  64747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64747, CurrentID=0, Status='live', Symbol='MFSD1', Aliases='SMAP4', description='major facilitator superfamily domain containing 1', other_designations='smooth muscle cell-associated protein 4', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158801450, end_position_on_the_genomic_accession=158829719, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  64748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64748, CurrentID=0, Status='live', Symbol='PLPPR2', Aliases='LPPR2, PRG4', description='phospholipid phosphatase related 2', other_designations='PRG-4|plasticity related gene 4|plasticity-related gene 4 protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11355386, end_position_on_the_genomic_accession=11365698, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100127954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127954, CurrentID=0, Status='live', Symbol='LOC100127954', Aliases='', description='pituitary tumor-transforming 1 pseudogene', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63032447, end_position_on_the_genomic_accession=63033108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64750, CurrentID=0, Status='live', Symbol='SMURF2', Aliases='', description='SMAD specific E3 ubiquitin protein ligase 2', other_designations='E3 ubiquitin ligase SMURF2|SMAD ubiquitination regulatory factor 2|SMAD-specific E3 ubiquitin-protein ligase 2|hSMURF2', map_location='17q22-q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64542282, end_position_on_the_genomic_accession=64662268, orientation='minus', exon_count=20, OMIM=605532, no_hdr0=''),
  64753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64753, CurrentID=0, Status='live', Symbol='CCDC136', Aliases='NAG6', description='coiled-coil domain containing 136', other_designations='nasopharyngeal carcinoma-associated gene 6 protein', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128791410, end_position_on_the_genomic_accession=128822133, orientation='plus', exon_count=22, OMIM=611902, no_hdr0=''),
  64754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64754, CurrentID=0, Status='live', Symbol='SMYD3', Aliases='KMT3E, ZMYND1, ZNFN3A1, bA74P14.1', description='SET and MYND domain containing 3', other_designations='SET and MYND domain-containing protein 3|bA74P14.1 (novel protein)|zinc finger MYND domain-containing protein 1|zinc finger protein, subfamily 3A (MYND domain containing), 1|zinc finger, MYND domain containing 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245749340, end_position_on_the_genomic_accession=246507342, orientation='minus', exon_count=22, OMIM=608783, no_hdr0=''),
  64755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64755, CurrentID=0, Status='live', Symbol='C16orf58', Aliases='RUS', description='chromosome 16 open reading frame 58', other_designations='UPF0420 protein C16orf58', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31489469, end_position_on_the_genomic_accession=31508419, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  195828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=195828, CurrentID=0, Status='live', Symbol='ZNF367', Aliases='AFF29, CDC14B, ZFF29', description='zinc finger protein 367', other_designations='C2H2 zinc finger protein ZFF29|C2H2 zinc finger protein ZFF29a|CDC14 cell division cycle 14 homolog B', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96385941, end_position_on_the_genomic_accession=96418720, orientation='minus', exon_count=9, OMIM=610160, no_hdr0=''),
  64757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64757, CurrentID=0, Status='live', Symbol='MARC1', Aliases='MOSC1', description='mitochondrial amidoxime reducing component 1', other_designations='MOCO sulphurase C-terminal domain containing 1|MOSC domain-containing protein 1, mitochondrial|moco sulfurase C-terminal domain-containing protein 1|molybdenum cofactor sulfurase C-terminal domain-containing protein 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220786697, end_position_on_the_genomic_accession=220814399, orientation='plus', exon_count=11, OMIM=614126, no_hdr0=''),
  64759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64759, CurrentID=0, Status='live', Symbol='TNS3', Aliases='TEM6, TENS1', description='tensin 3', other_designations='tensin-like SH2 domain containing 1|tensin-like SH2 domain-containing 1|tensin-like SH2 domain-containing protein 1|thyroid specific PTB domain protein|tumor endothelial marker 6', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47275144, end_position_on_the_genomic_accession=47582553, orientation='minus', exon_count=41, OMIM=606825, no_hdr0=''),
  64760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64760, CurrentID=0, Status='live', Symbol='FAM160B2', Aliases='RAI16', description='family with sequence similarity 160 member B2', other_designations='family with sequence similarity 160, member B2|retinoic acid induced 16|retinoic acid-induced protein 16', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22089184, end_position_on_the_genomic_accession=22104380, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  64761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64761, CurrentID=0, Status='live', Symbol='PARP12', Aliases='ARTD12, MST109, MSTP109, ZC3H1, ZC3HDC1', description='poly(ADP-ribose) polymerase family member 12', other_designations='ADP-ribosyltransferase diphtheria toxin-like 12|poly (ADP-ribose) polymerase family, member 12|zinc finger CCCH type domain containing 1', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140023744, end_position_on_the_genomic_accession=140063828, orientation='minus', exon_count=14, OMIM=612481, no_hdr0=''),
  64762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64762, CurrentID=0, Status='live', Symbol='GAREM1', Aliases='C18orf11, FAM59A, GAREM, Gm944', description='GRB2 associated regulator of MAPK1 subtype 1', other_designations='GRB2 associated regulator of MAPK1 1|GRB2 associated, regulator of MAPK1|GRB2-associated and regulator of MAPK1|Grb2-associated and regulator of Erk/MAPK|family with sequence similarity 59, member A|protein FAM59A', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32263521, end_position_on_the_genomic_accession=32470496, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  64763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64763, CurrentID=0, Status='live', Symbol='ZNF574', Aliases='FP972', description='zinc finger protein 574', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42074905, end_position_on_the_genomic_accession=42081567, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64764, CurrentID=0, Status='live', Symbol='CREB3L2', Aliases='BBF2H7', description='cAMP responsive element binding protein 3-like 2', other_designations='B-ZIB transcription factor|BBF2 human homolog on chromosome 7|FUS/BBF2H7 protein|TCAG_1951439|basic transcription factor 2|cAMP-responsive element-binding protein 3-like protein 2', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137874979, end_position_on_the_genomic_accession=138002101, orientation='minus', exon_count=14, OMIM=608834, no_hdr0=''),
  64766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64766, CurrentID=0, Status='live', Symbol='S100PBP', Aliases='S100PBPR', description='S100P binding protein', other_designations='S100P binding protein 1', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32816767, end_position_on_the_genomic_accession=32858879, orientation='plus', exon_count=13, OMIM=611889, no_hdr0=''),
  105370837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370837, CurrentID=0, Status='live', Symbol='LOC105370837', Aliases='', description='uncharacterized LOC105370837', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58919883, end_position_on_the_genomic_accession=58921815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64768, CurrentID=0, Status='live', Symbol='IPPK', Aliases='C9orf12, INSP5K2, IP5K, IPK1, bA476B13.1', description='inositol-pentakisphosphate 2-kinase', other_designations='IPK1 homolog|bA476B13.1 (novel protein)|inositol 1,3,4,5,6-pentakisphosphate 2-kinase|ins(1,3,4,5,6)P5 2-kinase|insP5 2-kinase', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92613184, end_position_on_the_genomic_accession=92670265, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  64769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64769, CurrentID=0, Status='live', Symbol='MEAF6', Aliases='C1orf149, CENP-28, EAF6, NY-SAR-91', description='MYST/Esa1-associated factor 6', other_designations='Esa1p-associated factor 6 homolog|centromere protein 28|protein EAF6 homolog|sarcoma antigen NY-SAR-91', map_location='1p35.3-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37489960, end_position_on_the_genomic_accession=37514819, orientation='minus', exon_count=11, OMIM=611001, no_hdr0=''),
  64770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64770, CurrentID=0, Status='live', Symbol='CCDC14', Aliases='', description='coiled-coil domain containing 14', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123887353, end_position_on_the_genomic_accession=123961409, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  64771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64771, CurrentID=0, Status='live', Symbol='C6orf106', Aliases='FP852, dJ391O22.4', description='chromosome 6 open reading frame 106', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34587280, end_position_on_the_genomic_accession=34696850, orientation='minus', exon_count=7, OMIM=612217, no_hdr0=''),
  64772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64772, CurrentID=0, Status='live', Symbol='ENGASE', Aliases='', description='endo-beta-N-acetylglucosaminidase', other_designations='Di-N-acetylchitobiosyl beta-N-acetylglucosaminidase|Mannosyl-glycoprotein endo-beta-N-acetylglucosaminidase', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79071523, end_position_on_the_genomic_accession=79088603, orientation='plus', exon_count=17, OMIM=611898, no_hdr0=''),
  64773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64773, CurrentID=0, Status='live', Symbol='PCED1A', Aliases='C20orf81, FAM113A, bA12M19.1', description='PC-esterase domain containing 1A', other_designations='family with sequence similarity 113, member A|sarcoma antigen NY-SAR-23', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2835314, end_position_on_the_genomic_accession=2841243, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64776, CurrentID=0, Status='live', Symbol='C11orf1', Aliases='', description='chromosome 11 open reading frame 1', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111879198, end_position_on_the_genomic_accession=111884074, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  64777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64777, CurrentID=0, Status='live', Symbol='RMND5B', Aliases='GID2, GID2B', description='required for meiotic nuclear division 5 homolog B', other_designations='GID complex subunit 2 homolog B', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178130532, end_position_on_the_genomic_accession=178148570, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  64778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64778, CurrentID=0, Status='live', Symbol='FNDC3B', Aliases='FAD104, PRO4979, YVTM2421', description='fibronectin type III domain containing 3B', other_designations='HCV NS5A-binding protein 37|factor for adipocyte differentiation 104', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172039628, end_position_on_the_genomic_accession=172400703, orientation='plus', exon_count=28, OMIM=611909, no_hdr0=''),
  64779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64779, CurrentID=0, Status='live', Symbol='MTHFSD', Aliases='', description='methenyltetrahydrofolate synthetase domain containing', other_designations='methenyltetrahydrofolate synthetase domain-containing protein', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86530176, end_position_on_the_genomic_accession=86555235, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  64780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64780, CurrentID=0, Status='live', Symbol='MICAL1', Aliases='MICAL, MICAL-1, NICAL', description='microtubule associated monooxygenase, calponin and LIM domain containing 1', other_designations='NEDD9-interacting protein with calponin homology and LIM domains|molecule interacting with CasL protein 1', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109444062, end_position_on_the_genomic_accession=109465968, orientation='minus', exon_count=26, OMIM=607129, no_hdr0=''),
  64781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64781, CurrentID=0, Status='live', Symbol='CERK', Aliases='LK4, dA59H18.2, dA59H18.3, hCERK', description='ceramide kinase', other_designations='acylsphingosine kinase|lipid kinase 4|lipid kinase LK4', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46684410, end_position_on_the_genomic_accession=46738255, orientation='minus', exon_count=14, OMIM=610307, no_hdr0=''),
  64782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64782, CurrentID=0, Status='live', Symbol='AEN', Aliases='ISG20L1, pp12744', description='apoptosis enhancing nuclease', other_designations='interferon stimulated exonuclease gene 20kDa-like 1|interferon-stimulated 20 kDa exonuclease-like 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88604681, end_position_on_the_genomic_accession=88632281, orientation='plus', exon_count=8, OMIM=610177, no_hdr0=''),
  64783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64783, CurrentID=0, Status='live', Symbol='RBM15', Aliases='OTT, OTT1, SPEN', description='RNA binding motif protein 15', other_designations='RNA-binding motif protein 15|one twenty two protein|one twenty-two|one-twenty two protein 1', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110338426, end_position_on_the_genomic_accession=110373218, orientation='plus', exon_count=5, OMIM=606077, no_hdr0=''),
  64784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64784, CurrentID=0, Status='live', Symbol='CRTC3', Aliases='TORC-3, TORC3', description='CREB regulated transcription coactivator 3', other_designations='transducer of regulated CREB protein 3|transducer of regulated cAMP response element-binding protein (CREB) 3|transducer of regulated cAMP response element-binding protein 3', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90529886, end_position_on_the_genomic_accession=90645345, orientation='plus', exon_count=15, OMIM=608986, no_hdr0=''),
  64785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64785, CurrentID=0, Status='live', Symbol='GINS3', Aliases='PSF3', description='GINS complex subunit 3 (Psf3 homolog)', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58392394, end_position_on_the_genomic_accession=58406144, orientation='plus', exon_count=4, OMIM=610610, no_hdr0=''),
  64786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64786, CurrentID=0, Status='live', Symbol='TBC1D15', Aliases='RAB7-GAP', description='TBC1 domain family member 15', other_designations='GAP for RAB7|GTPase-activating protein RAB7|TBC1 domain family, member 15|Tre-2-budding uninhibited by benzimidazole-cell division cycle 16 domain, domain family member 15|Tre2/Bub2/Cdc16 domain family member 15', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71839707, end_position_on_the_genomic_accession=71926849, orientation='plus', exon_count=26, OMIM=612662, no_hdr0=''),
  64787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64787, CurrentID=0, Status='live', Symbol='EPS8L2', Aliases='EPS8R2', description='EPS8-like 2', other_designations='EPS8-like protein 2|EPS8-related protein 2|epidermal growth factor receptor pathway substrate 8-related protein 2', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=706120, end_position_on_the_genomic_accession=727727, orientation='plus', exon_count=21, OMIM=614988, no_hdr0=''),
  64788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64788, CurrentID=0, Status='live', Symbol='LMF1', Aliases='C16orf26, HMFN1876, JFP11, TMEM112, TMEM112A', description='lipase maturation factor 1', other_designations='transmembrane protein 112', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=853634, end_position_on_the_genomic_accession=981318, orientation='minus', exon_count=24, OMIM=611761, no_hdr0=''),
  64789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64789, CurrentID=0, Status='live', Symbol='EXO5', Aliases='C1orf176, DEM1, Exo V, hExo5', description='exonuclease 5', other_designations='defects in morphology 1 homolog|defects in morphology protein 1 homolog|probable exonuclease V', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40508719, end_position_on_the_genomic_accession=40517725, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370841, CurrentID=0, Status='live', Symbol='LOC105370841', Aliases='', description='uncharacterized LOC105370841', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61219547, end_position_on_the_genomic_accession=61222257, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64792, CurrentID=0, Status='live', Symbol='IFT22', Aliases='RABL5', description='intraflagellar transport 22', other_designations='RAB, member RAS oncogene family-like 5|RAB, member of RAS oncogene family-like 5|intraflagellar transport 22 homolog|rab-like protein 5', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101313367, end_position_on_the_genomic_accession=101321823, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  64793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64793, CurrentID=0, Status='live', Symbol='CEP85', Aliases='CCDC21', description='centrosomal protein 85kDa', other_designations='coiled-coil domain-containing protein 21', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26234153, end_position_on_the_genomic_accession=26279038, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  64794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64794, CurrentID=0, Status='live', Symbol='DDX31', Aliases='PPP1R25', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 31', other_designations='DEAD box protein 31|DEAD/DEXH helicase DDX31|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 31|G2 helicase|helicain|protein phosphatase 1, regulatory subunit 25', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132592997, end_position_on_the_genomic_accession=132670401, orientation='minus', exon_count=24, OMIM=616533, no_hdr0=''),
  64795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64795, CurrentID=0, Status='live', Symbol='RMND5A', Aliases='CTLH, GID2, GID2A, RMD5, p44CTLH', description='required for meiotic nuclear division 5 homolog A', other_designations='C-terminal to LisH motif, 44 kDa|GID complex subunit 2 homolog A', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86720291, end_position_on_the_genomic_accession=86778041, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105370842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370842, CurrentID=0, Status='live', Symbol='LOC105370842', Aliases='', description='uncharacterized LOC105370842', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60883416, end_position_on_the_genomic_accession=60895643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64798, CurrentID=0, Status='live', Symbol='DEPTOR', Aliases='DEP.6, DEPDC6', description='DEP domain containing MTOR-interacting protein', other_designations='DEP domain containing 6|DEP domain containing mTOR interacting protein|DEP domain-containing protein 6', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119873655, end_position_on_the_genomic_accession=120050918, orientation='plus', exon_count=9, OMIM=612974, no_hdr0=''),
  64799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64799, CurrentID=0, Status='live', Symbol='IQCH', Aliases='NYDSP5', description='IQ motif containing H', other_designations='testis development protein NYD-SP5', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67254774, end_position_on_the_genomic_accession=67506612, orientation='plus', exon_count=27, OMIM=612523, no_hdr0=''),
  64800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64800, CurrentID=0, Status='live', Symbol='EFCAB6', Aliases='DJBP, HSCBCIP1, dJ185D5.1', description='EF-hand calcium binding domain 6', other_designations='CAP-binding protein complex-interacting protein 1|DJ-1-binding protein', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43528561, end_position_on_the_genomic_accession=43812337, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  64801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64801, CurrentID=0, Status='live', Symbol='ARV1', Aliases='', description='ARV1 homolog, fatty acid homeostasis modulator', other_designations='ARV1 fatty acid homeostatsis modulator|ARV1 homolog, fatty acid homeostatsis modulator|hARV1', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230979075, end_position_on_the_genomic_accession=231000733, orientation='plus', exon_count=7, OMIM=611647, no_hdr0=''),
  64802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64802, CurrentID=0, Status='live', Symbol='NMNAT1', Aliases='LCA9, NMNAT, PNAT1', description='nicotinamide nucleotide adenylyltransferase 1', other_designations='NMN adenylyltransferase 1|NMN/NaMN adenylyltransferase 1|NaMN adenylyltransferase 1|nicotinamide mononucleotide adenylyltransferase 1|nicotinamide-nucleotide adenylyltransferase 1|nicotinate-nucleotide adenylyltransferase 1|pyridine nucleotide adenylyltransferase 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9942923, end_position_on_the_genomic_accession=9985501, orientation='plus', exon_count=7, OMIM=608700, no_hdr0=''),
  644315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644315, CurrentID=0, Status='live', Symbol='RPS7P11', Aliases='RPS7_6_1552', description='ribosomal protein S7 pseudogene 11', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46721548, end_position_on_the_genomic_accession=46722240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64805, CurrentID=0, Status='live', Symbol='P2RY12', Aliases='ADPG-R, BDPLT8, HORK3, P2T(AC), P2Y(12)R, P2Y(AC), P2Y(ADP), P2Y(cyc), P2Y12, SP1999', description='purinergic receptor P2Y, G-protein coupled, 12', other_designations='ADP-glucose receptor|G-protein coupled receptor SP1999|Gi-coupled ADP receptor HORK3|P2Y12 platelet ADP receptor|purinergic receptor P2RY12|putative G-protein coupled receptor', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151336843, end_position_on_the_genomic_accession=151384812, orientation='minus', exon_count=3, OMIM=600515, no_hdr0=''),
  64806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64806, CurrentID=0, Status='live', Symbol='IL25', Aliases='IL17E', description='interleukin 25', other_designations='interleukin-17E', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23372809, end_position_on_the_genomic_accession=23376403, orientation='plus', exon_count=2, OMIM=605658, no_hdr0=''),
  105373232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373232, CurrentID=0, Status='live', Symbol='LOC105373232', Aliases='', description='uncharacterized LOC105373232', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241723366, end_position_on_the_genomic_accession=241742111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370844, CurrentID=0, Status='live', Symbol='LOC105370844', Aliases='', description='uncharacterized LOC105370844', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60763075, end_position_on_the_genomic_accession=60765317, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100127965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127965, CurrentID=0, Status='live', Symbol='LOC100127965', Aliases='', description='ribosomal protein L7-like 1 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30537204, end_position_on_the_genomic_accession=30538232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64816, CurrentID=0, Status='live', Symbol='CYP3A43', Aliases='', description='cytochrome P450 family 3 subfamily A member 43', other_designations='cytochrome P450, family 3, subfamily A, polypeptide 43|cytochrome P450, subfamily IIIA, polypeptide 43', map_location='7q21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99828013, end_position_on_the_genomic_accession=99867092, orientation='plus', exon_count=16, OMIM=606534, no_hdr0=''),
  285852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285852, CurrentID=0, Status='live', Symbol='TREML4', Aliases='TLT-4, TLT4', description='triggering receptor expressed on myeloid cells like 4', other_designations='TREM like transcript 4|triggering receptor expressed on myeloid cells-like 4|triggering receptor expressed on myeloid cells-like protein 4', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41228324, end_position_on_the_genomic_accession=41238382, orientation='plus', exon_count=6, OMIM=614664, no_hdr0=''),
  100418661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418661, CurrentID=0, Status='live', Symbol='LOC100418661', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370847, CurrentID=0, Status='live', Symbol='LOC105370847', Aliases='', description='uncharacterized LOC105370847', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61404197, end_position_on_the_genomic_accession=61411242, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64834, CurrentID=0, Status='live', Symbol='ELOVL1', Aliases='CGI-88, Ssc1', description='ELOVL fatty acid elongase 1', other_designations='3-keto acyl-CoA synthase ELOVL1|ELOVL FA elongase 1|elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 1|very-long-chain 3-oxoacyl-CoA synthase 1', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43363397, end_position_on_the_genomic_accession=43368074, orientation='minus', exon_count=9, OMIM=611813, no_hdr0=''),
  64837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64837, CurrentID=0, Status='live', Symbol='KLC2', Aliases='', description='kinesin light chain 2', other_designations='KLC 2', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66257294, end_position_on_the_genomic_accession=66267861, orientation='plus', exon_count=17, OMIM=611729, no_hdr0=''),
  64838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64838, CurrentID=0, Status='live', Symbol='FNDC4', Aliases='FRCP1', description='fibronectin type III domain containing 4', other_designations='fibronectin type III repeat-containing protein 1', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27491883, end_position_on_the_genomic_accession=27495296, orientation='minus', exon_count=7, OMIM=611905, no_hdr0=''),
  64839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64839, CurrentID=0, Status='live', Symbol='FBXL17', Aliases='FBXO13, Fbl17, Fbx13', description='F-box and leucine-rich repeat protein 17', other_designations='F-box only protein 13', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=107859033, end_position_on_the_genomic_accession=108382098, orientation='minus', exon_count=18, OMIM=609083, no_hdr0=''),
  64840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64840, CurrentID=0, Status='live', Symbol='PORCN', Aliases='DHOF, FODH, MG61, PORC, PPN', description='porcupine homolog (Drosophila)', other_designations='probable protein-cysteine N-palmitoyltransferase porcupine', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48508959, end_position_on_the_genomic_accession=48520814, orientation='plus', exon_count=16, OMIM=300651, no_hdr0=''),
  64841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64841, CurrentID=0, Status='live', Symbol='GNPNAT1', Aliases='GNPNAT, Gpnat1', description='glucosamine-phosphate N-acetyltransferase 1', other_designations='phosphoglucosamine acetylase|phosphoglucosamine transacetylase', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52775193, end_position_on_the_genomic_accession=52791668, orientation='minus', exon_count=8, OMIM=616510, no_hdr0=''),
  64843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64843, CurrentID=0, Status='live', Symbol='ISL2', Aliases='', description='ISL LIM homeobox 2', other_designations='ISL2 transcription factor, LIM/homeodomain, (islet-2)|islet-2', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76335942, end_position_on_the_genomic_accession=76342476, orientation='plus', exon_count=8, OMIM=609481, no_hdr0=''),
  64844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64844, CurrentID=0, Status='live', Symbol='MARCH7', Aliases='AXO, AXOT, MARCH-VII, RNF177', description='membrane associated ring-CH-type finger 7', other_designations='RING finger protein 177|axotrophin|membrane associated ring finger 7|membrane-associated RING finger protein 7|membrane-associated RING-CH protein VII|membrane-associated ring finger (C3HC4) 7, E3 ubiquitin protein ligase', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159712457, end_position_on_the_genomic_accession=159768583, orientation='plus', exon_count=13, OMIM=613334, no_hdr0=''),
  64847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64847, CurrentID=0, Status='live', Symbol='SPATA20', Aliases='HEL-S-98, SSP411, Tisp78', description='spermatogenesis associated 20', other_designations='epididymis secretory protein Li 98|sperm protein SSP411|sperm-specific protein 411|transcript increased in spermiogenesis 78', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50547089, end_position_on_the_genomic_accession=50555852, orientation='plus', exon_count=17, OMIM=613939, no_hdr0=''),
  64848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64848, CurrentID=0, Status='live', Symbol='YTHDC2', Aliases='CAHL', description='YTH domain containing 2', other_designations='CsA-associated helicase-like protein|YTH domain-containing protein 2', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=113513683, end_position_on_the_genomic_accession=113595287, orientation='plus', exon_count=33, OMIM=616530, no_hdr0=''),
  64849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64849, CurrentID=0, Status='live', Symbol='SLC13A3', Aliases='NADC3, SDCT2', description='solute carrier family 13 (sodium-dependent dicarboxylate transporter), member 3', other_designations='Na(+)/dicarboxylate cotransporter 3|hNaDC3|naDC-3|sodium-dependent high affinity dicarboxylate transporter 3|sodium-dependent high-affinity dicarboxylate transporter 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46557823, end_position_on_the_genomic_accession=46684485, orientation='minus', exon_count=15, OMIM=606411, no_hdr0=''),
  64850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64850, CurrentID=0, Status='live', Symbol='ETNPPL', Aliases='AGXT2L1', description='ethanolamine-phosphate phospho-lyase', other_designations='alanine--glyoxylate aminotransferase 2-like 1|alanine-glyoxylate aminotransferase 2-like 1', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108742040, end_position_on_the_genomic_accession=108763079, orientation='minus', exon_count=14, OMIM=614682, no_hdr0=''),
  64851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64851, CurrentID=0, Status='live', Symbol='AD6', Aliases='', description='Alzheimer disease 6', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605526, no_hdr0=''),
  64852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64852, CurrentID=0, Status='live', Symbol='TUT1', Aliases='PAPD2, RBM21, STARPAP, TUTase, URLC6', description='terminal uridylyl transferase 1, U6 snRNA-specific', other_designations='PAP-associated domain-containing 2|RNA binding motif protein 21|RNA uridylyltransferase|RNA-binding motif protein 21|RNA-binding protein 21|TUTase 6|U6 TUTase|U6 snRNA-specific terminal uridylyltransferase 1|U6-TUTase|nuclear speckle targeted phosphatidylinositol 4-phosphate 5-kinase type I-alpha regulated-poly(A) polymerase|nuclear speckle-targeted PIPK1A-regulated-poly(A) polymerase|poly(A) polymerase associated domain containing 2|star-PAP|terminal uridylyl transferase 1 U6 snRNA-specific|up-regulated in lung cancer 6', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62575045, end_position_on_the_genomic_accession=62591637, orientation='minus', exon_count=9, OMIM=610641, no_hdr0=''),
  64853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64853, CurrentID=0, Status='live', Symbol='AIDA', Aliases='C1orf80', description='axin interactor, dorsalization associated', other_designations='axin interaction partner and dorsalization antagonist', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222668013, end_position_on_the_genomic_accession=222712901, orientation='minus', exon_count=10, OMIM=612375, no_hdr0=''),
  64854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64854, CurrentID=0, Status='live', Symbol='USP46', Aliases='', description='ubiquitin specific peptidase 46', other_designations='deubiquitinating enzyme 46|ubiquitin thioesterase 46|ubiquitin thiolesterase 46|ubiquitin-specific-processing protease 46', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52590960, end_position_on_the_genomic_accession=52659335, orientation='minus', exon_count=11, OMIM=612849, no_hdr0=''),
  64855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64855, CurrentID=0, Status='live', Symbol='FAM129B', Aliases='C9orf88, MEG-3, MINERVA, OC58, bA356B19.6', description='family with sequence similarity 129 member B', other_designations='family with sequence similarity 129, member B|melanoma invasion by ERK', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127505338, end_position_on_the_genomic_accession=127579007, orientation='minus', exon_count=17, OMIM=614045, no_hdr0=''),
  64856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64856, CurrentID=0, Status='live', Symbol='VWA1', Aliases='WARP', description='von Willebrand factor A domain containing 1', other_designations='von Willebrand factor A domain-related protein', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1435523, end_position_on_the_genomic_accession=1442882, orientation='plus', exon_count=3, OMIM=611901, no_hdr0=''),
  64857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64857, CurrentID=0, Status='live', Symbol='PLEKHG2', Aliases='ARHGEF42, CLG', description='pleckstrin homology and RhoGEF domain containing G2', other_designations='CTB-60E11.4|PH domain-containing family G member 2|common-site lymphoma/leukemia guanine nucleotide exchange factor|pleckstrin homology domain containing, family G (with RhoGef domain) member 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39412582, end_position_on_the_genomic_accession=39428415, orientation='plus', exon_count=20, OMIM=611893, no_hdr0=''),
  64858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64858, CurrentID=0, Status='live', Symbol='DCLRE1B', Aliases='APOLLO, SNM1B, SNMIB', description='DNA cross-link repair 1B', other_designations='DNA cross-link repair 1B (PSO2 homolog, S. cerevisiae)|PSO2 homolog|SNM1 homolog B', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113905293, end_position_on_the_genomic_accession=113914086, orientation='plus', exon_count=4, OMIM=609683, no_hdr0=''),
  64859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64859, CurrentID=0, Status='live', Symbol='NABP1', Aliases='OBFC2A, SOSS-B2, SSB2', description='nucleic acid binding protein 1', other_designations='nucleic acid-binding protein 1|oligonucleotide/oligosaccharide-binding fold containing 2A|oligonucleotide/oligosaccharide-binding fold-containing protein 2A|sensor of single-strand DNA complex subunit B2|sensor of ssDNA subunit B2|single-stranded DNA-binding protein 2', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191678072, end_position_on_the_genomic_accession=191688522, orientation='plus', exon_count=8, OMIM=612103, no_hdr0=''),
  64860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64860, CurrentID=0, Status='live', Symbol='ARMCX5', Aliases='GASP5', description='armadillo repeat containing, X-linked 5', other_designations='', map_location='Xq22.1-q22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102599168, end_position_on_the_genomic_accession=102604159, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  404026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404026, CurrentID=0, Status='live', Symbol='MTCO3P1', Aliases='U92032.1', description='MT-CO3 pseudogene 1', other_designations='MTCO3 pseudogene 1|cytochrome c oxidase III pseudogene 1', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32706144, end_position_on_the_genomic_accession=32706803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64863, CurrentID=0, Status='live', Symbol='METTL4', Aliases='HsT661', description='methyltransferase like 4', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2537525, end_position_on_the_genomic_accession=2571490, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  64864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64864, CurrentID=0, Status='live', Symbol='RFX7', Aliases='RFXDC2', description='regulatory factor X7', other_designations='regulatory factor X domain containing 2|regulatory factor X domain-containing protein 2|regulatory factor X, 7', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56090533, end_position_on_the_genomic_accession=56243536, orientation='minus', exon_count=13, OMIM=612660, no_hdr0=''),
  64866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64866, CurrentID=0, Status='live', Symbol='CDCP1', Aliases='CD318, SIMA135, TRASK', description='CUB domain containing protein 1', other_designations='membrane glycoprotein gp140|subtractive immunization M plus HEp3-associated 135 kDa protein|transmembrane and associated with src kinases', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45082274, end_position_on_the_genomic_accession=45146422, orientation='minus', exon_count=9, OMIM=611735, no_hdr0=''),
  101927030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927030, CurrentID=0, Status='live', Symbol='LOC101927030', Aliases='', description='uncharacterized LOC101927030', other_designations='CTD-2035E11.4', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43024401, end_position_on_the_genomic_accession=43034635, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370854, CurrentID=0, Status='live', Symbol='LOC105370854', Aliases='', description='uncharacterized LOC105370854', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62833637, end_position_on_the_genomic_accession=62838155, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64881, CurrentID=0, Status='live', Symbol='PCDH20', Aliases='PCDH13', description='protocadherin 20', other_designations='protocadherin 13|protocadherin-13', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61409685, end_position_on_the_genomic_accession=61415522, orientation='minus', exon_count=2, OMIM=614449, no_hdr0=''),
  105372360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372360, CurrentID=0, Status='live', Symbol='LOC105372360', Aliases='', description='uncharacterized LOC105372360', other_designations='CTD-2050I18.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31520148, end_position_on_the_genomic_accession=31554795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  163732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=163732, CurrentID=0, Status='live', Symbol='CITED4', Aliases='', description='Cbp/p300-interacting transactivator, with Glu/Asp rich carboxy-terminal domain, 4', other_designations='Cbp/p300-interacting transactivator, with Glu/Asp-rich carboxy-terminal domain, 4|MRG-2|MSG1-related protein 2|transcriptional co-activator 4', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40861056, end_position_on_the_genomic_accession=40862346, orientation='minus', exon_count=1, OMIM=606815, no_hdr0=''),
  105370858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370858, CurrentID=0, Status='live', Symbol='LOC105370858', Aliases='', description='uncharacterized LOC105370858', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63320324, end_position_on_the_genomic_accession=63323792, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  64895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64895, CurrentID=0, Status='live', Symbol='PAPOLG', Aliases='', description='poly(A) polymerase gamma', other_designations="PAP-gamma|SRP RNA 3' adenylating enzyme/pap2|SRP RNA 3'-adenylating enzyme|neo-PAP|neo-poly(A) polymerase|nuclear poly(A) polymerase gamma|polynucleotide adenylyltransferase gamma", map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60756230, end_position_on_the_genomic_accession=60802086, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  64897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64897, CurrentID=0, Status='live', Symbol='C12orf43', Aliases='', description='chromosome 12 open reading frame 43', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121002422, end_position_on_the_genomic_accession=121016502, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370859, CurrentID=0, Status='live', Symbol='LOC105370859', Aliases='', description='uncharacterized LOC105370859', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63317988, end_position_on_the_genomic_accession=63345458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64900, CurrentID=0, Status='live', Symbol='LPIN3', Aliases='LIPN3L, SMP2', description='lipin 3', other_designations='lipin-3', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41340780, end_position_on_the_genomic_accession=41360585, orientation='plus', exon_count=23, OMIM=605520, no_hdr0=''),
  64901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64901, CurrentID=0, Status='live', Symbol='RANBP17', Aliases='', description='RAN binding protein 17', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170861882, end_position_on_the_genomic_accession=171300450, orientation='plus', exon_count=38, OMIM=606141, no_hdr0=''),
  64902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64902, CurrentID=0, Status='live', Symbol='AGXT2', Aliases='AGT2, DAIBAT', description='alanine--glyoxylate aminotransferase 2', other_designations='(R)-3-amino-2-methylpropionate--pyruvate transaminase|beta-ALAAT II|beta-alanine-pyruvate aminotransferase', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34998101, end_position_on_the_genomic_accession=35048188, orientation='minus', exon_count=14, OMIM=612471, no_hdr0=''),
  195977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=195977, CurrentID=0, Status='live', Symbol='ANTXRL', Aliases='', description='anthrax toxin receptor-like', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46286997, end_position_on_the_genomic_accession=46330419, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  105370861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370861, CurrentID=0, Status='live', Symbol='LOC105370861', Aliases='', description='uncharacterized LOC105370861', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64444747, end_position_on_the_genomic_accession=64460587, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645969, CurrentID=0, Status='live', Symbol='LOC645969', Aliases='', description='selenoprotein T pseudogene', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6278596, end_position_on_the_genomic_accession=6281993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127982, CurrentID=0, Status='live', Symbol='LOC100127982', Aliases='', description='transmembrane protein 69 pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97852822, end_position_on_the_genomic_accession=97853925, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  64919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64919, CurrentID=0, Status='live', Symbol='BCL11B', Aliases='ATL1, ATL1-alpha, ATL1-beta, ATL1-delta, ATL1-gamma, CTIP-2, CTIP2, RIT1, ZNF856B, hRIT1-alpha', description='B-cell CLL/lymphoma 11B (zinc finger protein)', other_designations='B-cell CLL/lymphoma 11B/T-cell receptor delta constant region fusion protein|B-cell lymphoma/leukaemia 11B|BCL-11B|BCL11B/TRDC fusion|COUP-TF-interacting protein 2|hRit1|radiation-induced tumor suppressor gene 1 protein|zinc finger protein hRit1 alpha', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99169287, end_position_on_the_genomic_accession=99271713, orientation='minus', exon_count=5, OMIM=606558, no_hdr0=''),
  64921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64921, CurrentID=0, Status='live', Symbol='CASD1', Aliases='C7orf12, NBLA04196', description='CAS1 domain containing 1', other_designations='O-acetyltransferase|WUGSC:H_GS542D18.2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94509858, end_position_on_the_genomic_accession=94631872, orientation='plus', exon_count=22, OMIM=611686, no_hdr0=''),
  64922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64922, CurrentID=0, Status='live', Symbol='LRRC19', Aliases='', description='leucine rich repeat containing 19', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26993136, end_position_on_the_genomic_accession=27005693, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370863, CurrentID=0, Status='live', Symbol='LOC105370863', Aliases='', description='uncharacterized LOC105370863', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65188995, end_position_on_the_genomic_accession=65196677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  64924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64924, CurrentID=0, Status='live', Symbol='SLC30A5', Aliases='ZNT5, ZNTL1, ZTL1, ZnT-5', description='solute carrier family 30 (zinc transporter), member 5', other_designations='zinc transporter ZTL1|znT-like transporter 1', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69093949, end_position_on_the_genomic_accession=69131072, orientation='plus', exon_count=16, OMIM=607819, no_hdr0=''),
  64925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64925, CurrentID=0, Status='live', Symbol='CCDC71', Aliases='', description='coiled-coil domain containing 71', other_designations='2600016J21Rik', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49162535, end_position_on_the_genomic_accession=49166352, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64926, CurrentID=0, Status='live', Symbol='RASAL3', Aliases='', description='RAS protein activator like 3', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15451624, end_position_on_the_genomic_accession=15464571, orientation='minus', exon_count=18, OMIM=616561, no_hdr0=''),
  64927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64927, CurrentID=0, Status='live', Symbol='TTC23', Aliases='HCC-8', description='tetratricopeptide repeat domain 23', other_designations='TPR repeat protein 23|cervical cancer proto-oncogene 8 protein', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99136323, end_position_on_the_genomic_accession=99251226, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  64928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64928, CurrentID=0, Status='live', Symbol='MRPL14', Aliases='L14mt, L32mt, MRP-L14, MRP-L32, MRPL32, RMPL32, RPML32', description='mitochondrial ribosomal protein L14', other_designations='39S ribosomal protein L32, mitochondrial', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44113635, end_position_on_the_genomic_accession=44127454, orientation='minus', exon_count=6, OMIM=611827, no_hdr0=''),
  100418662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418662, CurrentID=0, Status='live', Symbol='LOC100418662', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128622, CurrentID=0, Status='live', Symbol='LOC100128622', Aliases='', description='uncharacterized LOC100128622', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64939, CurrentID=0, Status='live', Symbol='SMNP', Aliases='', description='survival of motor neuron 1, telomeric pseudogene', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20331408, end_position_on_the_genomic_accession=20332377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64940, CurrentID=0, Status='live', Symbol='STAG3L4', Aliases='STAG3L4P', description='stromal antigen 3-like 4 (pseudogene)', other_designations='STAG3-like protein 4|stromal antigen 3 pseudogene', map_location='7p11.2-q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67302638, end_position_on_the_genomic_accession=67321526, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370866, CurrentID=0, Status='live', Symbol='LOC105370866', Aliases='', description='uncharacterized LOC105370866', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65792378, end_position_on_the_genomic_accession=65860476, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  64943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64943, CurrentID=0, Status='live', Symbol='NT5DC2', Aliases='', description="5'-nucleotidase domain containing 2", other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52524369, end_position_on_the_genomic_accession=52535077, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  64946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64946, CurrentID=0, Status='live', Symbol='CENPH', Aliases='', description='centromere protein H', other_designations='CENP-H|interphase centromere complex protein 35|kinetochore protein CENP-H', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69189548, end_position_on_the_genomic_accession=69210357, orientation='plus', exon_count=9, OMIM=605607, no_hdr0=''),
  644339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644339, CurrentID=0, Status='live', Symbol='LOC644339', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64075758, end_position_on_the_genomic_accession=64134813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64949, CurrentID=0, Status='live', Symbol='MRPS26', Aliases='C20orf193, GI008, MRP-S13, MRP-S26, MRPS13, NY-BR-87, RPMS13, dJ534B8.3', description='mitochondrial ribosomal protein S26', other_designations='28S ribosomal protein S13, mitochondrial|S13mt|S26mt|serologically defined breast cancer antigen NY-BR-87', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3046029, end_position_on_the_genomic_accession=3048250, orientation='plus', exon_count=4, OMIM=611988, no_hdr0=''),
  100335030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100335030, CurrentID=0, Status='live', Symbol='LOC100335030', Aliases='', description='FGFR1 oncogene partner 2 pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49389174, end_position_on_the_genomic_accession=49392335, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  64951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64951, CurrentID=0, Status='live', Symbol='MRPS24', Aliases='HSPC335, MRP-S24, S24mt, bMRP-47, bMRP47', description='mitochondrial ribosomal protein S24', other_designations='mitochondrial 28S ribosomal protein S24', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43866558, end_position_on_the_genomic_accession=43869546, orientation='minus', exon_count=4, OMIM=611986, no_hdr0=''),
  105370869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370869, CurrentID=0, Status='live', Symbol='LOC105370869', Aliases='', description='uncharacterized LOC105370869', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66841300, end_position_on_the_genomic_accession=66842469, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64960, CurrentID=0, Status='live', Symbol='MRPS15', Aliases='DC37, MPR-S15, RPMS15, S15mt', description='mitochondrial ribosomal protein S15', other_designations='MRP-S15', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36455761, end_position_on_the_genomic_accession=36464439, orientation='minus', exon_count=8, OMIM=611979, no_hdr0=''),
  64963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64963, CurrentID=0, Status='live', Symbol='MRPS11', Aliases='HCC-2', description='mitochondrial ribosomal protein S11', other_designations='MRP-S11|S11mt|cervical cancer proto-oncogene 2 protein', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88467453, end_position_on_the_genomic_accession=88478630, orientation='plus', exon_count=7, OMIM=611977, no_hdr0=''),
  64965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64965, CurrentID=0, Status='live', Symbol='MRPS9', Aliases='MRP-S9, RPMS9, S9mt', description='mitochondrial ribosomal protein S9', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105038025, end_position_on_the_genomic_accession=105099960, orientation='plus', exon_count=13, OMIM=611975, no_hdr0=''),
  64968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64968, CurrentID=0, Status='live', Symbol='MRPS6', Aliases='C21orf101, MRP-S6, RPMS6, S6mt', description='mitochondrial ribosomal protein S6', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34073523, end_position_on_the_genomic_accession=34143034, orientation='plus', exon_count=3, OMIM=611973, no_hdr0=''),
  64969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64969, CurrentID=0, Status='live', Symbol='MRPS5', Aliases='MRP-S5, S5mt', description='mitochondrial ribosomal protein S5', other_designations='mitochondrial 28S ribosomal protein S5', map_location='2p11.2-q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95087207, end_position_on_the_genomic_accession=95122009, orientation='minus', exon_count=12, OMIM=611972, no_hdr0=''),
  105370871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370871, CurrentID=0, Status='live', Symbol='LOC105370871', Aliases='', description='uncharacterized LOC105370871', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68182814, end_position_on_the_genomic_accession=68183623, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  196047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196047, CurrentID=0, Status='live', Symbol='EMX2OS', Aliases='EMX2-AS1, NCRNA00045', description='EMX2 opposite strand/antisense RNA', other_designations='EMX2 opposite strand (non-protein coding)|EMX2 opposite strand/antisense RNA (non-protein coding)|empty spiracles homeobox 2 opposite strand|empty spiracles homolog 2 opposite strand', map_location='10q26.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117484293, end_position_on_the_genomic_accession=117545068, orientation='minus', exon_count=4, OMIM=607637, no_hdr0=''),
  64976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64976, CurrentID=0, Status='live', Symbol='MRPL40', Aliases='L40mt, MRP-L22, MRP-L40, MRPL22, NLVCF, URIM', description='mitochondrial ribosomal protein L40', other_designations='nuclear localization signal-containing protein deleted in velocardiofacial syndrome|up-regulated in metastasis', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19432513, end_position_on_the_genomic_accession=19436077, orientation='plus', exon_count=4, OMIM=605089, no_hdr0=''),
  64978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64978, CurrentID=0, Status='live', Symbol='MRPL38', Aliases='HSPC262, L38MT, MRP-L3, MRP-L38, RPML3', description='mitochondrial ribosomal protein L38', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75898643, end_position_on_the_genomic_accession=75905100, orientation='minus', exon_count=9, OMIM=611844, no_hdr0=''),
  196051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196051, CurrentID=0, Status='live', Symbol='PLPP4', Aliases='DPPL2, PPAPDC1, PPAPDC1A', description='phospholipid phosphatase 4', other_designations='diacylglycerol pyrophosphate like 2|phosphatidic acid phosphatase type 2 domain containing 1A|phosphatidic acid phosphatase type 2 domain-containing protein 1A', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120456954, end_position_on_the_genomic_accession=120589855, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  64981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64981, CurrentID=0, Status='live', Symbol='MRPL34', Aliases='L34mt', description='mitochondrial ribosomal protein L34', other_designations='MRP-L34', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17292778, end_position_on_the_genomic_accession=17306847, orientation='plus', exon_count=4, OMIM=611840, no_hdr0=''),
  64983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64983, CurrentID=0, Status='live', Symbol='MRPL32', Aliases='HSPC283, L32mt, MRP-L32, bMRP-59b', description='mitochondrial ribosomal protein L32', other_designations='', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42932205, end_position_on_the_genomic_accession=42937857, orientation='plus', exon_count=4, OMIM=611839, no_hdr0=''),
  100127994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127994, CurrentID=0, Status='live', Symbol='LOC100127994', Aliases='', description='uncharacterized LOC100127994', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=75357666, end_position_on_the_genomic_accession=75358064, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479001, CurrentID=0, Status='live', Symbol='RNA5SP243', Aliases='RN5S243', description='RNA, 5S ribosomal pseudogene 243', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370875, CurrentID=0, Status='live', Symbol='LOC105370875', Aliases='', description='uncharacterized LOC105370875', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69278315, end_position_on_the_genomic_accession=69284443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  196074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196074, CurrentID=0, Status='live', Symbol='METTL15', Aliases='METT5D1', description='methyltransferase like 15', other_designations='methyltransferase 5 domain containing 1|methyltransferase 5 domain-containing protein 1|probable S-adenosyl-L-methionine-dependent methyltransferase METT5D1|putative S-adenosyl-L-methionine-dependent methyltransferase METT5D1', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28108251, end_position_on_the_genomic_accession=28527041, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  65003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65003, CurrentID=0, Status='live', Symbol='MRPL11', Aliases='CGI-113, L11MT, MRP-L11', description='mitochondrial ribosomal protein L11', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66435075, end_position_on_the_genomic_accession=66438848, orientation='minus', exon_count=6, OMIM=611826, no_hdr0=''),
  65005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65005, CurrentID=0, Status='live', Symbol='MRPL9', Aliases='L9mt', description='mitochondrial ribosomal protein L9', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151759643, end_position_on_the_genomic_accession=151763916, orientation='minus', exon_count=7, OMIM=611824, no_hdr0=''),
  65008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65008, CurrentID=0, Status='live', Symbol='MRPL1', Aliases='BM022, L1MT, MRP-L1', description='mitochondrial ribosomal protein L1', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77861955, end_position_on_the_genomic_accession=77952790, orientation='plus', exon_count=12, OMIM=611821, no_hdr0=''),
  65009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65009, CurrentID=0, Status='live', Symbol='NDRG4', Aliases='BDM1, SMAP-8, SMAP8', description='NDRG family member 4', other_designations='N-myc downstream-regulated gene 4 protein|brain development-related molecule 1|smooth muscle-associated protein 8|vascular smooth muscle cell-associated protein 8', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58463645, end_position_on_the_genomic_accession=58513619, orientation='plus', exon_count=23, OMIM=614463, no_hdr0=''),
  65010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65010, CurrentID=0, Status='live', Symbol='SLC26A6', Aliases='', description='solute carrier family 26 (anion exchanger), member 6', other_designations='anion exchange transporter|anion transporter 1|pendrin L1|solute carrier family 26, member 6|sulfate anion transporter', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48625723, end_position_on_the_genomic_accession=48635493, orientation='minus', exon_count=21, OMIM=610068, no_hdr0=''),
  65012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65012, CurrentID=0, Status='live', Symbol='SLC26A10', Aliases='', description='solute carrier family 26 member 10', other_designations='solute carrier family 26, member 10', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57619910, end_position_on_the_genomic_accession=57626151, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  65014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65014, CurrentID=0, Status='live', Symbol='CMD1K', Aliases='', description='cardiomyopathy, dilated 1K (autosomal dominant)', other_designations='', map_location='6q12-q16', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605582, no_hdr0=''),
  65018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65018, CurrentID=0, Status='live', Symbol='PINK1', Aliases='BRPK, PARK6', description='PTEN induced putative kinase 1', other_designations='PTEN-induced putative kinase protein 1|protein kinase BRPK', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20633455, end_position_on_the_genomic_accession=20651511, orientation='plus', exon_count=8, OMIM=608309, no_hdr0=''),
  105370879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370879, CurrentID=0, Status='live', Symbol='LOC105370879', Aliases='', description='uncharacterized LOC105370879', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644353, CurrentID=0, Status='live', Symbol='ZCCHC18', Aliases='PNMA7B, SIZN2', description='zinc finger, CCHC domain containing 18', other_designations='paraneoplastic Ma antigen family member 7B|putative zinc finger CCHC domain-containing protein 18|zinc finger, CCHC domain containing 12 pseudogene 1', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104112416, end_position_on_the_genomic_accession=104115847, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370882, CurrentID=0, Status='live', Symbol='LOC105370882', Aliases='', description='uncharacterized LOC105370882', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71428693, end_position_on_the_genomic_accession=71432186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  196120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196120, CurrentID=0, Status='live', Symbol='SSU72P5', Aliases='', description='SSU72 pseudogene 5', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 5', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4233278, end_position_on_the_genomic_accession=4233872, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370884, CurrentID=0, Status='live', Symbol='LOC105370884', Aliases='', description='uncharacterized LOC105370884', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71253871, end_position_on_the_genomic_accession=71262976, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106479003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479003, CurrentID=0, Status='live', Symbol='RNA5SP297', Aliases='RN5S297', description='RNA, 5S ribosomal pseudogene 297', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  65055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65055, CurrentID=0, Status='live', Symbol='REEP1', Aliases='C2orf23, HMN5B, SPG31', description='receptor accessory protein 1', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86213993, end_position_on_the_genomic_accession=86338418, orientation='minus', exon_count=12, OMIM=609139, no_hdr0=''),
  65056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65056, CurrentID=0, Status='live', Symbol='GPBP1', Aliases='GPBP, SSH6, VASCULIN', description='GC-rich promoter binding protein 1', other_designations='vascular wall-linked protein', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57173849, end_position_on_the_genomic_accession=57264679, orientation='plus', exon_count=15, OMIM=608412, no_hdr0=''),
  65057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65057, CurrentID=0, Status='live', Symbol='ACD', Aliases='PIP1, PTOP, TINT1, TPP1', description='adrenocortical dysplasia homolog (mouse)', other_designations='POT1 and TIN2-interacting protein|TIN2 interacting protein 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67657512, end_position_on_the_genomic_accession=67660815, orientation='minus', exon_count=12, OMIM=609377, no_hdr0=''),
  65059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65059, CurrentID=0, Status='live', Symbol='RAPH1', Aliases='ALS2CR18, ALS2CR9, LPD, PREL-2, PREL2, RMO1, RalGDS/AF-6', description='Ras association (RalGDS/AF-6) and pleckstrin homology domains 1', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 18|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 9|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 18 protein|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 9 protein|lamellipodin|proline-rich EVH1 ligand 2', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203433682, end_position_on_the_genomic_accession=203535335, orientation='minus', exon_count=20, OMIM=609035, no_hdr0=''),
  65061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65061, CurrentID=0, Status='live', Symbol='CDK15', Aliases='ALS2CR7, PFTAIRE2, PFTK2', description='cyclin-dependent kinase 15', other_designations='PFTAIRE protein kinase 2|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 7|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 7 protein|cell division protein kinase 15|serine/threonine protein kinase|serine/threonine-protein kinase ALS2CR7|serine/threonine-protein kinase PFTAIRE-2', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201790459, end_position_on_the_genomic_accession=201895550, orientation='plus', exon_count=26, OMIM=616147, no_hdr0=''),
  65062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65062, CurrentID=0, Status='live', Symbol='TMEM237', Aliases='ALS2CR4, JBTS14', description='transmembrane protein 237', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 4|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 4 protein', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201620184, end_position_on_the_genomic_accession=201643529, orientation='minus', exon_count=14, OMIM=614423, no_hdr0=''),
  65065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65065, CurrentID=0, Status='live', Symbol='NBEAL1', Aliases='A530083I02Rik, ALS2CR16, ALS2CR17', description='neurobeachin like 1', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 16|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 17|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 16 protein|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 17 protein|beach|neurobeachin-like 1', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203014617, end_position_on_the_genomic_accession=203226353, orientation='plus', exon_count=62, OMIM=609816, no_hdr0=''),
  105370887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370887, CurrentID=0, Status='live', Symbol='LOC105370887', Aliases='', description='uncharacterized LOC105370887', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72274680, end_position_on_the_genomic_accession=72284233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  65072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65072, CurrentID=0, Status='live', Symbol='CFLAR-AS1', Aliases='ALS2CR10', description='CFLAR antisense RNA 1', other_designations='CFLAR antisense RNA 1 (non-protein coding)|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 10', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201140289, end_position_on_the_genomic_accession=201157792, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  65076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65076, CurrentID=0, Status='live', Symbol='AOMS1', Aliases='SYNX', description='Abdominal obesity-metabolic syndrome QTL1', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605552, no_hdr0=''),
  65077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65077, CurrentID=0, Status='live', Symbol='AOMS2', Aliases='', description='abdominal obesity-metabolic syndrome QTL2', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605572, no_hdr0=''),
  65078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65078, CurrentID=0, Status='live', Symbol='RTN4R', Aliases='NGR, NOGOR', description='reticulon 4 receptor', other_designations='Nogo-66 receptor|UNQ330/PRO526|nogo receptor', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20241415, end_position_on_the_genomic_accession=20268293, orientation='minus', exon_count=2, OMIM=605566, no_hdr0=''),
  65080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65080, CurrentID=0, Status='live', Symbol='MRPL44', Aliases='COXPD16, L44MT, MRP-L44', description='mitochondrial ribosomal protein L44', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223950890, end_position_on_the_genomic_accession=223967714, orientation='plus', exon_count=5, OMIM=611849, no_hdr0=''),
  65082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65082, CurrentID=0, Status='live', Symbol='VPS33A', Aliases='', description='VPS33A, CORVET/HOPS core subunit', other_designations='VPS33A CORVET/HOPS core subunit|hVPS33A|vacuolar protein sorting 33 homolog A|vacuolar protein sorting 33A', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122231547, end_position_on_the_genomic_accession=122266521, orientation='minus', exon_count=13, OMIM=610034, no_hdr0=''),
  65083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65083, CurrentID=0, Status='live', Symbol='NOL6', Aliases='NRAP, UTP22, bA311H10.1', description='nucleolar protein 6', other_designations='nucleolar RNA-associated protein|nucleolar protein 6 (RNA-associated)|nucleolar protein family 6 (RNA-associated)', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33461353, end_position_on_the_genomic_accession=33473943, orientation='minus', exon_count=26, OMIM=611532, no_hdr0=''),
  65084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65084, CurrentID=0, Status='live', Symbol='TMEM135', Aliases='PMP52', description='transmembrane protein 135', other_designations='peroxisomal membrane protein 52', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87037844, end_position_on_the_genomic_accession=87328834, orientation='plus', exon_count=15, OMIM=616360, no_hdr0=''),
  105370890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370890, CurrentID=0, Status='live', Symbol='LOC105370890', Aliases='', description='uncharacterized LOC105370890', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73335242, end_position_on_the_genomic_accession=73343376, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370891, CurrentID=0, Status='live', Symbol='LOC105370891', Aliases='', description='uncharacterized LOC105370891', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73636588, end_position_on_the_genomic_accession=73655020, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  65094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65094, CurrentID=0, Status='live', Symbol='JMJD4', Aliases='', description='jumonji domain containing 4', other_designations='2-oxoglutarate- and Fe(II)-dependent oxygenase|C4 lysyl hydroxylase|jumonji domain-containing protein 4', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227731189, end_position_on_the_genomic_accession=227735499, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  65095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65095, CurrentID=0, Status='live', Symbol='KRI1', Aliases='', description='KRI1 homolog', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10553085, end_position_on_the_genomic_accession=10566031, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105370892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370892, CurrentID=0, Status='live', Symbol='LOC105370892', Aliases='', description='uncharacterized LOC105370892', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73806733, end_position_on_the_genomic_accession=73810299, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  65108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65108, CurrentID=0, Status='live', Symbol='MARCKSL1', Aliases='F52, MACMARCKS, MLP, MLP1, MRP', description='MARCKS-like 1', other_designations='MARCKS-like protein 1|mac-MARCKS|macrophage myristoylated alanine-rich C kinase substrate', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32333829, end_position_on_the_genomic_accession=32336239, orientation='minus', exon_count=2, OMIM=602940, no_hdr0=''),
  65109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65109, CurrentID=0, Status='live', Symbol='UPF3B', Aliases='HUPF3B, MRX62, MRXS14, RENT3B, UPF3X', description='UPF3 regulator of nonsense transcripts homolog B (yeast)', other_designations='hUpf3p-X|mental retardation, X-linked 62|nonsense mRNA reducing factor 3B|up-frameshift suppressor 3 homolog B|up-frameshift suppressor 3 homolog on chromosome X', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119805311, end_position_on_the_genomic_accession=119853028, orientation='minus', exon_count=16, OMIM=300298, no_hdr0=''),
  65110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65110, CurrentID=0, Status='live', Symbol='UPF3A', Aliases='HUPF3A, RENT3A, UPF3', description='UPF3 regulator of nonsense transcripts homolog A (yeast)', other_designations='hUpf3|nonsense mRNA reducing factor 3A|up-frameshift suppressor 3 homolog A', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114281584, end_position_on_the_genomic_accession=114305817, orientation='plus', exon_count=13, OMIM=605530, no_hdr0=''),
  105370895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370895, CurrentID=0, Status='live', Symbol='LOC105370895', Aliases='', description='uncharacterized LOC105370895', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  196188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196188, CurrentID=0, Status='live', Symbol='RPS2P38', Aliases='RPS2_18_1126', description='ribosomal protein S2 pseudogene 38', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23520295, end_position_on_the_genomic_accession=23521239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  65117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65117, CurrentID=0, Status='live', Symbol='RSRC2', Aliases='', description='arginine/serine-rich coiled-coil 2', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122504643, end_position_on_the_genomic_accession=122527013, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  65121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65121, CurrentID=0, Status='live', Symbol='PRAMEF1', Aliases='', description='PRAME family member 1', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12791397, end_position_on_the_genomic_accession=12796628, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65122, CurrentID=0, Status='live', Symbol='PRAMEF2', Aliases='DJ845O24.3', description='PRAME family member 2', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12857086, end_position_on_the_genomic_accession=12861909, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65123, CurrentID=0, Status='live', Symbol='INTS3', Aliases='C1orf193, C1orf60, INT3, SOSS-A, SOSSA', description='integrator complex subunit 3', other_designations='SOSS complex subunit A|sensor of single-strand DNA complex subunit A|sensor of ssDNA A|sensor of ssDNA subunit A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153728050, end_position_on_the_genomic_accession=153774808, orientation='plus', exon_count=34, OMIM=611347, no_hdr0=''),
  65124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65124, CurrentID=0, Status='live', Symbol='SOWAHC', Aliases='ANKRD57, C2orf26', description='sosondowah ankyrin repeat domain family member C', other_designations='ankyrin repeat domain 57|ankyrin repeat domain-containing protein 57|protein sosondowah homolog C', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109614334, end_position_on_the_genomic_accession=109618987, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  65125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65125, CurrentID=0, Status='live', Symbol='WNK1', Aliases='HSAN2, HSN2, KDP, PPP1R167, PRKWNK1, PSK, p65', description='WNK lysine deficient protein kinase 1', other_designations='WNK lysine deficient protein kinase 1 isoform|erythrocyte 65 kDa protein|prostate-derived sterile 20-like kinase|protein kinase with no lysine 1|protein phosphatase 1, regulatory subunit 167|serine/threonine-protein kinase WNK1 1|serine/threonine-protein kinase WNK1 2', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=752593, end_position_on_the_genomic_accession=911452, orientation='plus', exon_count=31, OMIM=605232, no_hdr0=''),
  105370897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370897, CurrentID=0, Status='live', Symbol='LOC105370897', Aliases='', description='uncharacterized LOC105370897', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74828132, end_position_on_the_genomic_accession=74829701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374782, CurrentID=0, Status='live', Symbol='LOC105374782', Aliases='', description='uncharacterized LOC105374782', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66227793, end_position_on_the_genomic_accession=66236237, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128019, CurrentID=0, Status='live', Symbol='LOC100128019', Aliases='', description='uncharacterized LOC100128019', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=55473098, end_position_on_the_genomic_accession=55473935, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370900, CurrentID=0, Status='live', Symbol='LOC105370900', Aliases='', description='uncharacterized LOC105370900', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465472, CurrentID=0, Status='live', Symbol='MIR6787', Aliases='hsa-mir-6787', description='microRNA 6787', other_designations='microRNA mir-6787', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82236668, end_position_on_the_genomic_accession=82236728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  731755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731755, CurrentID=0, Status='live', Symbol='LOC731755', Aliases='', description='eukaryotic translation initiation factor 2 alpha kinase 1 pseudogene', other_designations='eukaryotic translation initiation factor 2-alpha kinase 1 pseudogene', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142450130, end_position_on_the_genomic_accession=142452172, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128022, CurrentID=0, Status='live', Symbol='LOC100128022', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59929230, end_position_on_the_genomic_accession=59930330, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132994, CurrentID=0, Status='live', Symbol='CXorf49B', Aliases='', description='chromosome X open reading frame 49B', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71763290, end_position_on_the_genomic_accession=71767201, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102465473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465473, CurrentID=0, Status='live', Symbol='MIR6790', Aliases='hsa-mir-6790', description='microRNA 6790', other_designations='microRNA mir-6790', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6392921, end_position_on_the_genomic_accession=6392983, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370903, CurrentID=0, Status='live', Symbol='LOC105370903', Aliases='', description='uncharacterized LOC105370903', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76107202, end_position_on_the_genomic_accession=76127865, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927032, CurrentID=0, Status='live', Symbol='LOC101927032', Aliases='', description='uncharacterized LOC101927032', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93762166, end_position_on_the_genomic_accession=93771976, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370905, CurrentID=0, Status='live', Symbol='LOC105370905', Aliases='', description='uncharacterized LOC105370905', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77041611, end_position_on_the_genomic_accession=77046378, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370906, CurrentID=0, Status='live', Symbol='LOC105370906', Aliases='', description='uncharacterized LOC105370906', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77568858, end_position_on_the_genomic_accession=77573031, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465477, CurrentID=0, Status='live', Symbol='MIR6796', Aliases='hsa-mir-6796', description='microRNA 6796', other_designations='microRNA mir-6796', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40369846, end_position_on_the_genomic_accession=40369907, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128027, CurrentID=0, Status='live', Symbol='LOC100128027', Aliases='', description='cysteine rich, DPF motif domain containing 1 pseudogene', other_designations='cysteine-rich, DPF motif domain containing 1 pseudogene', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55224697, end_position_on_the_genomic_accession=55224841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102465478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465478, CurrentID=0, Status='live', Symbol='MIR6797', Aliases='hsa-mir-6797', description='microRNA 6797', other_designations='microRNA mir-6797', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41869627, end_position_on_the_genomic_accession=41869698, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  196264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196264, CurrentID=0, Status='live', Symbol='MPZL3', Aliases='', description='myelin protein zero-like 3', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118226690, end_position_on_the_genomic_accession=118252368, orientation='minus', exon_count=7, OMIM=611707, no_hdr0=''),
  100128028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128028, CurrentID=0, Status='live', Symbol='LOC100128028', Aliases='', description='uncharacterized LOC100128028', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=44642139, end_position_on_the_genomic_accession=44650712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100499399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499399, CurrentID=0, Status='live', Symbol='LOC100499399', Aliases='', description='adenylate kinase 3 pseudogene', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52735296, end_position_on_the_genomic_accession=52736812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370909, CurrentID=0, Status='live', Symbol='LOC105370909', Aliases='', description='uncharacterized LOC105370909', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370910, CurrentID=0, Status='live', Symbol='LOC105370910', Aliases='', description='uncharacterized LOC105370910', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78274514, end_position_on_the_genomic_accession=78278720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  613204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613204, CurrentID=0, Status='live', Symbol='IFITM8P', Aliases='', description='interferon induced transmembrane protein 8 pseudogene', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63408975, end_position_on_the_genomic_accession=63409628, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  65211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65211, CurrentID=0, Status='live', Symbol='FRTS1', Aliases='FRTS, RFS', description='Fanconi renotubular syndrome', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=134600, no_hdr0=''),
  65217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65217, CurrentID=0, Status='live', Symbol='PCDH15', Aliases='CDHR15, DFNB23, USH1F', description='protocadherin-related 15', other_designations='cadherin-related family member 15', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=53802771, end_position_on_the_genomic_accession=54801291, orientation='minus', exon_count=40, OMIM=605514, no_hdr0=''),
  65220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65220, CurrentID=0, Status='live', Symbol='NADK', Aliases='dJ283E3.1', description='NAD kinase', other_designations='poly(P)/ATP NAD kinase', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1751232, end_position_on_the_genomic_accession=1780514, orientation='minus', exon_count=20, OMIM=611616, no_hdr0=''),
  196294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196294, CurrentID=0, Status='live', Symbol='IMMP1L', Aliases='IMMP1, IMP1, IMP1-LIKE', description='inner mitochondrial membrane peptidase subunit 1', other_designations='IMP1 inner mitochondrial membrane peptidase-like', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31432399, end_position_on_the_genomic_accession=31509632, orientation='minus', exon_count=10, OMIM=612323, no_hdr0=''),
  105370913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370913, CurrentID=0, Status='live', Symbol='LOC105370913', Aliases='', description='uncharacterized LOC105370913', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78692826, end_position_on_the_genomic_accession=78728055, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100499405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499405, CurrentID=0, Status='live', Symbol='LINC00987', Aliases='', description='long intergenic non-protein coding RNA 987', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9240003, end_position_on_the_genomic_accession=9243049, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128035, CurrentID=0, Status='live', Symbol='LOC100128035', Aliases='', description='VGSA5840', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=111576840, end_position_on_the_genomic_accession=111578440, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=10873, CurrentID=0, Status='live', Symbol='ME3', Aliases='NADP-ME', description='malic enzyme 3, NADP(+)-dependent, mitochondrial', other_designations='malate dehydrogenase|malic enzyme, NADP+-dependent, mitochondrial|mitochondrial NADP(+)-dependent malic enzyme 3|pyruvic-malic carboxylase', map_location='11cen-q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86441108, end_position_on_the_genomic_accession=86672636, orientation='minus', exon_count=22, OMIM=604626, no_hdr0=''),
  100128036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128036, CurrentID=0, Status='live', Symbol='KLF7P1', Aliases='KLF7P', description='kruppel-like factor 7 pseudogene 1', other_designations='Kruppel-like factor 7 (ubiquitous) pseudogene', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170952859, end_position_on_the_genomic_accession=170994882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  65243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65243, CurrentID=0, Status='live', Symbol='ZFP69B', Aliases='ZKSCAN23B, ZNF643, ZSCAN54B', description='ZFP69 zinc finger protein B', other_designations='novel zinc finger protein|zinc finger protein 643', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40450086, end_position_on_the_genomic_accession=40463718, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  65244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65244, CurrentID=0, Status='live', Symbol='SPATS2', Aliases='Nbla00526, P59SCR, SCR59, SPATA10', description='spermatogenesis associated, serine rich 2', other_designations='putative protein product of Nbla00526|serine-rich spermatocytes and round spermatid 59 kDa protein|serine-rich spermatocytes and round spermatid protein|spermatogenesis associated, serine-rich 2', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49366905, end_position_on_the_genomic_accession=49527426, orientation='plus', exon_count=17, OMIM=611667, no_hdr0=''),
  65245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65245, CurrentID=0, Status='live', Symbol='PSORS7', Aliases='', description='psoriasis susceptibility 7', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605606, no_hdr0=''),
  65249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65249, CurrentID=0, Status='live', Symbol='ZSWIM4', Aliases='', description='zinc finger, SWIM-type containing 4', other_designations='zinc finger, SWIM domain containing 4', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13795434, end_position_on_the_genomic_accession=13832254, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  65250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65250, CurrentID=0, Status='live', Symbol='C5orf42', Aliases='Hug, JBTS17, OFD6', description='chromosome 5 open reading frame 42', other_designations='Transmembrane protein ENSP00000382582', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37065171, end_position_on_the_genomic_accession=37249428, orientation='minus', exon_count=58, OMIM=614571, no_hdr0=''),
  65251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65251, CurrentID=0, Status='live', Symbol='ZNF649', Aliases='', description='zinc finger protein 649', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51888538, end_position_on_the_genomic_accession=51905052, orientation='minus', exon_count=6, OMIM=611903, no_hdr0=''),
  105374786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374786, CurrentID=0, Status='live', Symbol='LOC105374786', Aliases='', description='uncharacterized LOC105374786', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67553910, end_position_on_the_genomic_accession=67564516, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370918, CurrentID=0, Status='live', Symbol='LOC105370918', Aliases='', description='uncharacterized LOC105370918', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79463780, end_position_on_the_genomic_accession=79475148, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102465489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465489, CurrentID=0, Status='live', Symbol='MIR6815', Aliases='hsa-mir-6815', description='microRNA 6815', other_designations='microRNA mir-6815', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45478266, end_position_on_the_genomic_accession=45478326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  65258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65258, CurrentID=0, Status='live', Symbol='MPPE1', Aliases='PGAP5', description='metallophosphoesterase 1', other_designations='metallo phosphoesterase|post-GPI attachment to proteins factor 5', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11883471, end_position_on_the_genomic_accession=11908797, orientation='minus', exon_count=17, OMIM=611900, no_hdr0=''),
  105370919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370919, CurrentID=0, Status='live', Symbol='LOC105370919', Aliases='', description='uncharacterized LOC105370919', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79494428, end_position_on_the_genomic_accession=79520597, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65260, CurrentID=0, Status='live', Symbol='COA7', Aliases='C1orf163, RESA1, SELRC1', description='cytochrome c oxidase assembly factor 7 (putative)', other_designations='Sel1 repeat containing 1|beta-lactamase hcp-like protein|hcp beta-lactamase-like protein C1orf163|respiratory chain assembly 1|respiratory chain assembly factor 1|sel1 repeat-containing protein 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52686342, end_position_on_the_genomic_accession=52698366, orientation='minus', exon_count=3, OMIM=615623, no_hdr0=''),
  196335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196335, CurrentID=0, Status='live', Symbol='OR56B4', Aliases='OR11-67', description='olfactory receptor family 56 subfamily B member 4', other_designations='olfactory receptor OR11-67|olfactory receptor, family 56, subfamily B, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6107684, end_position_on_the_genomic_accession=6108835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  65264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65264, CurrentID=0, Status='live', Symbol='UBE2Z', Aliases='HOYS7, USE1', description='ubiquitin conjugating enzyme E2Z', other_designations='E2 ubiquitin-conjugating enzyme Z|UBA6-specific enzyme E2|uba6-specific E2 conjugating enzyme 1|ubiquitin carrier protein Z|ubiquitin-conjugating enzyme E2Z|ubiquitin-protein ligase Z', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48908369, end_position_on_the_genomic_accession=48929060, orientation='plus', exon_count=7, OMIM=611362, no_hdr0=''),
  65265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65265, CurrentID=0, Status='live', Symbol='C8orf33', Aliases='', description='chromosome 8 open reading frame 33', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=145052438, end_position_on_the_genomic_accession=145065864, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  65266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65266, CurrentID=0, Status='live', Symbol='WNK4', Aliases='PHA2B, PRKWNK4', description='WNK lysine deficient protein kinase 4', other_designations='protein kinase lysine-deficient 4|protein kinase with no lysine 4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42779673, end_position_on_the_genomic_accession=42797066, orientation='plus', exon_count=19, OMIM=601844, no_hdr0=''),
  65267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65267, CurrentID=0, Status='live', Symbol='WNK3', Aliases='PRKWNK3', description='WNK lysine deficient protein kinase 3', other_designations='protein kinase with no lysine 3', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54192823, end_position_on_the_genomic_accession=54358695, orientation='minus', exon_count=27, OMIM=300358, no_hdr0=''),
  65268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65268, CurrentID=0, Status='live', Symbol='WNK2', Aliases='NY-CO-43, P/OKcl.13, PRKWNK2, SDCCAG43', description='WNK lysine deficient protein kinase 2', other_designations='antigen NY-CO-43|mitogen-activated protein kinase kinase kinase|protein kinase lysine-deficient 2|protein kinase with no lysine 2|protein kinase, lysine deficient 2|serologically defined colon cancer antigen 43', map_location='9q22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93184155, end_position_on_the_genomic_accession=93320572, orientation='plus', exon_count=37, OMIM=606249, no_hdr0=''),
  105370921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370921, CurrentID=0, Status='live', Symbol='LOC105370921', Aliases='', description='uncharacterized LOC105370921', other_designations='CTD-2034I4.2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81667011, end_position_on_the_genomic_accession=81689650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465492, CurrentID=0, Status='live', Symbol='MIR6820', Aliases='hsa-mir-6820', description='microRNA 6820', other_designations='microRNA mir-6820', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37967563, end_position_on_the_genomic_accession=37967624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370922, CurrentID=0, Status='live', Symbol='LOC105370922', Aliases='', description='uncharacterized LOC105370922', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81978659, end_position_on_the_genomic_accession=82013691, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374787, CurrentID=0, Status='live', Symbol='LOC105374787', Aliases='', description='uncharacterized LOC105374787', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67549148, end_position_on_the_genomic_accession=67551853, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370923, CurrentID=0, Status='live', Symbol='LOC105370923', Aliases='', description='uncharacterized LOC105370923', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82026119, end_position_on_the_genomic_accession=82030058, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370924, CurrentID=0, Status='live', Symbol='LOC105370924', Aliases='', description='uncharacterized LOC105370924', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82096582, end_position_on_the_genomic_accession=82099672, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370925, CurrentID=0, Status='live', Symbol='LOC105370925', Aliases='', description='uncharacterized LOC105370925', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82347640, end_position_on_the_genomic_accession=82349718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379152, CurrentID=0, Status='live', Symbol='LOC105379152', Aliases='', description='uncharacterized LOC105379152', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123086231, end_position_on_the_genomic_accession=123090149, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100499416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499416, CurrentID=0, Status='live', Symbol='TOMM22P1', Aliases='', description='TOMM22 pseudogene 1', other_designations='translocase of outer mitochondrial membrane 22 homolog pseudogene 1', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21130870, end_position_on_the_genomic_accession=21131181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370926, CurrentID=0, Status='live', Symbol='LOC105370926', Aliases='', description='uncharacterized LOC105370926', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82487783, end_position_on_the_genomic_accession=82493706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  196374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196374, CurrentID=0, Status='live', Symbol='KRT78', Aliases='CK-78, K5B, K78, Kb40', description='keratin 78, type II', other_designations='cytokeratin-78|keratin-5b|type-II keratin Kb40', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52837804, end_position_on_the_genomic_accession=52848994, orientation='minus', exon_count=10, OMIM=611159, no_hdr0=''),
  100499417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499417, CurrentID=0, Status='live', Symbol='TOMM22P2', Aliases='', description='TOMM22 pseudogene 2', other_designations='translocase of outer mitochondrial membrane 22 homolog pseudogene 2', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2827982, end_position_on_the_genomic_accession=2828218, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499418, CurrentID=0, Status='live', Symbol='MTCYBP1', Aliases='', description='MT-CYB pseudogene 1', other_designations='MTCYB pseudogene 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8364032, end_position_on_the_genomic_accession=8365150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  196383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196383, CurrentID=0, Status='live', Symbol='RILPL2', Aliases='RLP2', description='Rab interacting lysosomal protein-like 2', other_designations='p40phox-binding protein|rab-interacting lysosomal protein-like 2', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123410076, end_position_on_the_genomic_accession=123436733, orientation='minus', exon_count=5, OMIM=614093, no_hdr0=''),
  196385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196385, CurrentID=0, Status='live', Symbol='DNAH10', Aliases='', description='dynein, axonemal, heavy chain 10', other_designations='axonemal beta dynein heavy chain 10|ciliary dynein heavy chain 10|dynein, axonemal, heavy polypeptide 10', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123761133, end_position_on_the_genomic_accession=123935720, orientation='plus', exon_count=79, OMIM=605884, no_hdr0=''),
  196394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196394, CurrentID=0, Status='live', Symbol='AMN1', Aliases='', description='antagonist of mitotic exit network 1 homolog', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31671137, end_position_on_the_genomic_accession=31730392, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105370930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370930, CurrentID=0, Status='live', Symbol='LOC105370930', Aliases='', description='uncharacterized LOC105370930', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83179123, end_position_on_the_genomic_accession=83186582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  196403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196403, CurrentID=0, Status='live', Symbol='DTX3', Aliases='RNF154, deltex3', description='deltex 3, E3 ubiquitin ligase', other_designations='RING finger protein 154|deltex homolog 3|protein deltex-3', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57604327, end_position_on_the_genomic_accession=57609804, orientation='plus', exon_count=8, OMIM=613142, no_hdr0=''),
  100128052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128052, CurrentID=0, Status='live', Symbol='LOC100128052', Aliases='', description="protein phosphatase 2 regulatory subunit B', gamma pseudogene", other_designations="protein phosphatase 2, regulatory subunit B', delta pseudogene|protein phosphatase 2, regulatory subunit B', gamma pseudogene", map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138022472, end_position_on_the_genomic_accession=138024446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  196410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196410, CurrentID=0, Status='live', Symbol='METTL7B', Aliases='ALDI', description='methyltransferase like 7B', other_designations='associated with lipid droplets 1', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55681546, end_position_on_the_genomic_accession=55684611, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  196415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196415, CurrentID=0, Status='live', Symbol='C12orf77', Aliases='', description='chromosome 12 open reading frame 77', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24993424, end_position_on_the_genomic_accession=24997439, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370934, CurrentID=0, Status='live', Symbol='LOC105370934', Aliases='', description='uncharacterized LOC105370934', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83365011, end_position_on_the_genomic_accession=83368757, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370935, CurrentID=0, Status='live', Symbol='LOC105370935', Aliases='', description='uncharacterized LOC105370935', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83944103, end_position_on_the_genomic_accession=83945909, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379154, CurrentID=0, Status='live', Symbol='LOC105379154', Aliases='', description='uncharacterized LOC105379154', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122787495, end_position_on_the_genomic_accession=122844943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128056, CurrentID=0, Status='live', Symbol='LOC100128056', Aliases='', description='CTD small phosphatase 2 pseudogene', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 2 pseudogene', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=109838843, end_position_on_the_genomic_accession=109840614, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370937, CurrentID=0, Status='live', Symbol='LOC105370937', Aliases='', description='uncharacterized LOC105370937', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84180883, end_position_on_the_genomic_accession=84189179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  196441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196441, CurrentID=0, Status='live', Symbol='ZFC3H1', Aliases='CCDC131, PSRC2', description='zinc finger, C3H1-type containing', other_designations='coiled-coil domain containing 131|coiled-coil domain-containing protein 131|proline/serine-rich coiled-coil 2|proline/serine-rich coiled-coil protein 2', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71609599, end_position_on_the_genomic_accession=71663969, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  105369392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369392, CurrentID=0, Status='live', Symbol='LOC105369392', Aliases='', description='uncharacterized LOC105369392', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75701619, end_position_on_the_genomic_accession=75718752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374790, CurrentID=0, Status='live', Symbol='LOC105374790', Aliases='', description='uncharacterized LOC105374790', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68750613, end_position_on_the_genomic_accession=68763013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  196446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196446, CurrentID=0, Status='live', Symbol='MYRFL', Aliases='C12orf15, C12orf28, bcm1377', description='myelin regulatory factor-like', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69824569, end_position_on_the_genomic_accession=69958724, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  100128059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128059, CurrentID=0, Status='live', Symbol='LOC100128059', Aliases='', description='uncharacterized LOC100128059', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=169735702, end_position_on_the_genomic_accession=169739720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370940, CurrentID=0, Status='live', Symbol='LOC105370940', Aliases='', description='uncharacterized LOC105370940', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32028535, end_position_on_the_genomic_accession=32030596, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  196463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196463, CurrentID=0, Status='live', Symbol='PLBD2', Aliases='P76', description='phospholipase B domain containing 2', other_designations='76 kDa protein|LAMA-like protein 2|PLB homolog 2|lamina ancestor homolog 2|mannose-6-phosphate protein associated protein p76|phospholipase B domain-containing protein 2|phospholipase B-like 2 32 kDa form|phospholipase B-like 2 45 kDa form', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113358566, end_position_on_the_genomic_accession=113389715, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100128062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128062, CurrentID=0, Status='live', Symbol='SLC2A3P1', Aliases='GLUT3P1, GLUT6, SLC2A3P', description='solute carrier family 2 (facilitated glucose transporter), member 3 pseudogene 1', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168552494, end_position_on_the_genomic_accession=168553653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  196472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196472, CurrentID=0, Status='live', Symbol='FAM71C', Aliases='', description='family with sequence similarity 71 member C', other_designations='family with sequence similarity 71, member C', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=99647750, end_position_on_the_genomic_accession=99650114, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  196475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196475, CurrentID=0, Status='live', Symbol='RMST', Aliases='LINC00054, NCRMS, NCRNA00054', description='rhabdomyosarcoma 2 associated transcript (non-protein coding)', other_designations='long intergenic non-protein coding RNA 54|non-coding RNA in rhabdomyosarcoma (RMS)|rhabdomyosarcoma 2 associated transcript (non-coding RNA)', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97465021, end_position_on_the_genomic_accession=97533766, orientation='plus', exon_count=9, OMIM=607045, no_hdr0=''),
  196477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196477, CurrentID=0, Status='live', Symbol='CCER1', Aliases='C12orf12', description='coiled-coil glutamate rich protein 1', other_designations='coiled-coil glutamate-rich protein 1', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90952215, end_position_on_the_genomic_accession=90955176, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100073347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100073347, CurrentID=0, Status='live', Symbol='MIMT1', Aliases='LINC00067, MIM1, NCRNA00067', description='MER1 repeat containing imprinted transcript 1 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 67', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56840902, end_position_on_the_genomic_accession=56848554, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  613211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613211, CurrentID=0, Status='live', Symbol='DEFB134', Aliases='', description='defensin beta 134', other_designations='beta-defensin 135|defensin, beta 134', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11993177, end_position_on_the_genomic_accession=12000752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100499437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499437, CurrentID=0, Status='live', Symbol='DYZ1L19', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  196500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196500, CurrentID=0, Status='live', Symbol='PIANP', Aliases='C12orf53, LEDA1, PANP, leda-1', description='PILR alpha associated neural protein', other_designations='PILR-associating neural protein|liver endothelial differentiation-associated protein-1|paired immunoglobin-like type 2 receptor-associating neural protein', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6690934, end_position_on_the_genomic_accession=6700843, orientation='minus', exon_count=9, OMIM=616065, no_hdr0=''),
  102465518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465518, CurrentID=0, Status='live', Symbol='MIR6860', Aliases='hsa-mir-6860', description='microRNA 6860', other_designations='microRNA mir-6860', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67045643, end_position_on_the_genomic_accession=67045708, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  196513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196513, CurrentID=0, Status='live', Symbol='DCP1B', Aliases='DCP1, hDcp1b', description='decapping mRNA 1B', other_designations='DCP1 decapping enzyme homolog B|decapping enzyme hDcp1b', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1946048, end_position_on_the_genomic_accession=2004511, orientation='minus', exon_count=12, OMIM=609843, no_hdr0=''),
  644422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644422, CurrentID=0, Status='live', Symbol='LOC644422', Aliases='', description='splicing factor, arginine/serine-rich 6 pseudogene', other_designations='serine/arginine-rich splicing factor 6 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16616851, end_position_on_the_genomic_accession=16617881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370951, CurrentID=0, Status='live', Symbol='LOC105370951', Aliases='', description='uncharacterized LOC105370951', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85382849, end_position_on_the_genomic_accession=85384484, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102465522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465522, CurrentID=0, Status='live', Symbol='MIR6865', Aliases='hsa-mir-6865', description='microRNA 6865', other_designations='microRNA mir-6865', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4970086, end_position_on_the_genomic_accession=4970150, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  196527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196527, CurrentID=0, Status='live', Symbol='ANO6', Aliases='BDPLT7, SCTS, TMEM16F', description='anoctamin 6', other_designations='SCAN channel|small-conductance calcium-activated nonselective cation channel|transmembrane protein 16F', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45215987, end_position_on_the_genomic_accession=45440404, orientation='plus', exon_count=24, OMIM=608663, no_hdr0=''),
  196528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196528, CurrentID=0, Status='live', Symbol='ARID2', Aliases='BAF200, p200', description='AT-rich interaction domain 2', other_designations='ARID domain-containing protein 2|AT rich interactive domain 2 (ARID, RFX-like)|BRG1-associated factor 200|zinc finger protein with activation potential|zipzap|zipzap/p200', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45729689, end_position_on_the_genomic_accession=45908037, orientation='plus', exon_count=24, OMIM=609539, no_hdr0=''),
  105370952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370952, CurrentID=0, Status='live', Symbol='LOC105370952', Aliases='', description='uncharacterized LOC105370952', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85837291, end_position_on_the_genomic_accession=85841114, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465523, CurrentID=0, Status='live', Symbol='MIR6867', Aliases='hsa-mir-6867', description='microRNA 6867', other_designations='microRNA mir-6867', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40193597, end_position_on_the_genomic_accession=40193663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370953, CurrentID=0, Status='live', Symbol='LOC105370953', Aliases='', description='uncharacterized LOC105370953', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  196541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196541, CurrentID=0, Status='live', Symbol='METTL21C', Aliases='C13orf39', description='methyltransferase like 21C', other_designations='methyltransferase-like protein 21C', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102685747, end_position_on_the_genomic_accession=102694521, orientation='minus', exon_count=4, OMIM=615259, no_hdr0=''),
  100128075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128075, CurrentID=0, Status='live', Symbol='LOC100128075', Aliases='', description='uncharacterized LOC100128075', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=89734782, end_position_on_the_genomic_accession=89735210, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  196549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196549, CurrentID=0, Status='live', Symbol='EEF1DP3', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 3', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31846783, end_position_on_the_genomic_accession=31959584, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370956, CurrentID=0, Status='live', Symbol='LOC105370956', Aliases='', description='uncharacterized LOC105370956', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=87352179, end_position_on_the_genomic_accession=87422830, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102465527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465527, CurrentID=0, Status='live', Symbol='MIR6876', Aliases='hsa-mir-6876', description='microRNA 6876', other_designations='microRNA mir-6876', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25345402, end_position_on_the_genomic_accession=25345474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370957, CurrentID=0, Status='live', Symbol='LOC105370957', Aliases='', description='uncharacterized LOC105370957', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88277152, end_position_on_the_genomic_accession=88279983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128078, CurrentID=0, Status='live', Symbol='LOC100128078', Aliases='', description='NIPA-like domain containing 1 pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25161090, end_position_on_the_genomic_accession=25162061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128079, CurrentID=0, Status='live', Symbol='LOC100128079', Aliases='', description='uncharacterized LOC100128079', other_designations='CTD-2049O4.1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27650134, end_position_on_the_genomic_accession=27719029, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370961, CurrentID=0, Status='live', Symbol='LOC105370961', Aliases='', description='uncharacterized LOC105370961', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  327657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=327657, CurrentID=0, Status='live', Symbol='SERPINA9', Aliases='GCET1, SERPINA11, SERPINA11b', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 9', other_designations='centerin|germinal center B-cell-expressed transcript 1 protein|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 9|serine proteinase inhibitor A11', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94462721, end_position_on_the_genomic_accession=94476333, orientation='minus', exon_count=7, OMIM=615677, no_hdr0=''),
  327658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=327658, CurrentID=0, Status='live', Symbol='PUDPP1', Aliases='FAM16AY, HDHD1BP, HDHD1P1', description="pseudouridine 5'-phosphatase pseudogene 1", other_designations='HDHD family member 1B, pseudogene|family with sequence similarity 16, member A, Y-linked|haloacid dehalogenase-like hydrolase domain containing 1 pseudogene 1|haloacid dehalogenase-like hydrolase domain containing 1A pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=15373504, end_position_on_the_genomic_accession=15375458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  327659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=327659, CurrentID=0, Status='live', Symbol='PUDPP2', Aliases='FAM16B, HDHD1CP, HDHD1P2', description="pseudouridine 5'-phosphatase pseudogene 2", other_designations='HDHD family member 1C, pseudogene|family with sequence similarity 16, member B|haloacid dehalogenase-like hydrolase domain containing 1 pseudogene 2|haloacid dehalogenase-like hydrolase domain containing 1A pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152123160, end_position_on_the_genomic_accession=152125170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370962, CurrentID=0, Status='live', Symbol='LOC105370962', Aliases='', description='uncharacterized LOC105370962', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89334780, end_position_on_the_genomic_accession=89336161, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644435, CurrentID=0, Status='live', Symbol='AGGF1P2', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 2', other_designations='angiogenic factor VG5Q pseudogene|angiogenic factor with G patch and FHA domains 1 pseudogene 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133642087, end_position_on_the_genomic_accession=133642895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644436, CurrentID=0, Status='live', Symbol='LOC644436', Aliases='', description='tubulin, alpha 4a pseudogene', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63046726, end_position_on_the_genomic_accession=63048195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128086, CurrentID=0, Status='live', Symbol='LOC100128086', Aliases='', description='APAF1 interacting protein pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107116752, end_position_on_the_genomic_accession=107117907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289473, CurrentID=0, Status='live', Symbol='LOC100289473', Aliases='', description='cytoskeleton associated protein 2-like pseudogene', other_designations='immunoglobulin C1-set domain containing pseudogene', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1773365, end_position_on_the_genomic_accession=1779746, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100499457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499457, CurrentID=0, Status='live', Symbol='MTND2P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 3', other_designations='MT-ND2 pseudogene 3|MTND2 pseudogene 3', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8372241, end_position_on_the_genomic_accession=8372710, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133007, CurrentID=0, Status='live', Symbol='LOC100133007', Aliases='', description='DENN/MADD domain containing 1C pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63395552, end_position_on_the_genomic_accession=63396564, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100859919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859919, CurrentID=0, Status='live', Symbol='LOC100859919', Aliases='', description='chromosome 9 open reading frame 86 pseudogene 2', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131833380, end_position_on_the_genomic_accession=131836358, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100859920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859920, CurrentID=0, Status='live', Symbol='HIGD1AP3', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 3', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174632929, end_position_on_the_genomic_accession=174633206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100859921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859921, CurrentID=0, Status='live', Symbol='LINC00536', Aliases='', description='long intergenic non-protein coding RNA 536', other_designations='', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=115950511, end_position_on_the_genomic_accession=116325059, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100859922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859922, CurrentID=0, Status='live', Symbol='ERVK11-1', Aliases='', description='endogenous retrovirus group K11 member 1', other_designations='Envelope polyprotein|HERV-K_1p13.3 provirus ancestral Env polyprotein|endogenous retrovirus group K11, member 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100859924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859924, CurrentID=0, Status='live', Symbol='SUMO2P2', Aliases='', description='SUMO2 pseudogene 2', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 2', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23632106, end_position_on_the_genomic_accession=23632355, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100859925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859925, CurrentID=0, Status='live', Symbol='COX5BP8', Aliases='', description='cytochrome c oxidase subunit Vb pseudogene 8', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179255733, end_position_on_the_genomic_accession=179256081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100859926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859926, CurrentID=0, Status='live', Symbol='NANOGNBP1', Aliases='', description='NANOGNB pseudogene 1', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100657878, end_position_on_the_genomic_accession=100660944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100859927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859927, CurrentID=0, Status='live', Symbol='AAA4', Aliases='', description='Aortic aneurysm, familial abdominal, 4', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614375, no_hdr0=''),
  100859928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859928, CurrentID=0, Status='live', Symbol='BMIQ16', Aliases='C16DELp11.2, DEL16p.11.2', description='Chromosome 16p11.2 deletion syndrome, 220kb', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613444, no_hdr0=''),
  100859930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100859930, CurrentID=0, Status='live', Symbol='HEIH', Aliases='HCCAT2, LINC-HEIH, LINC00848, lncRNA-HEIH', description='hepatocellular carcinoma up-regulated EZH2-associated long non-coding RNA', other_designations='hepatocellular carcinoma associated transcript 2 (non-protein coding)|highly expressed in hepatocellular carcinoma long non-coding RNA|long intergenic non-protein coding RNA 848|long intergenic non-protein coding RNA, highly expressed in hepatocellular carcinoma|long non-coding RNA, high expression in hepatocellular carcinoma', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180829954, end_position_on_the_genomic_accession=180831618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370970, CurrentID=0, Status='live', Symbol='LOC105370970', Aliases='', description='uncharacterized LOC105370970', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91010554, end_position_on_the_genomic_accession=91013670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370432, CurrentID=0, Status='live', Symbol='LOC105370432', Aliases='', description='uncharacterized LOC105370432', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30503438, end_position_on_the_genomic_accession=30513381, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128091, CurrentID=0, Status='live', Symbol='LOC100128091', Aliases='', description='coiled-coil domain containing 25 pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39718028, end_position_on_the_genomic_accession=39718593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128092, CurrentID=0, Status='live', Symbol='PRDX3P2', Aliases='', description='peroxiredoxin 3 pseudogene 2', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28526358, end_position_on_the_genomic_accession=28527238, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370973, CurrentID=0, Status='live', Symbol='LOC105370973', Aliases='', description='uncharacterized LOC105370973', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91643213, end_position_on_the_genomic_accession=91649301, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370974, CurrentID=0, Status='live', Symbol='LOC105370974', Aliases='', description='uncharacterized LOC105370974', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370976, CurrentID=0, Status='live', Symbol='LOC105370976', Aliases='', description='uncharacterized LOC105370976', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92623130, end_position_on_the_genomic_accession=92628100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370977, CurrentID=0, Status='live', Symbol='LOC105370977', Aliases='', description='uncharacterized LOC105370977', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92612226, end_position_on_the_genomic_accession=92618839, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370978, CurrentID=0, Status='live', Symbol='LOC105370978', Aliases='', description='uncharacterized LOC105370978', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92591621, end_position_on_the_genomic_accession=92611049, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370979, CurrentID=0, Status='live', Symbol='LOC105370979', Aliases='', description='uncharacterized LOC105370979', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93041407, end_position_on_the_genomic_accession=93043253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132526, CurrentID=0, Status='live', Symbol='FGD5P1', Aliases='', description='FYVE, RhoGEF and PH domain containing 5 pseudogene 1', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13933056, end_position_on_the_genomic_accession=13936944, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370982, CurrentID=0, Status='live', Symbol='LOC105370982', Aliases='', description='uncharacterized LOC105370982', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93553620, end_position_on_the_genomic_accession=93593361, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370983, CurrentID=0, Status='live', Symbol='LOC105370983', Aliases='', description='uncharacterized LOC105370983', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93228784, end_position_on_the_genomic_accession=93541186, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105373238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373238, CurrentID=0, Status='live', Symbol='LOC105373238', Aliases='', description='uncharacterized LOC105373238', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242700942, end_position_on_the_genomic_accession=242773796, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644456, CurrentID=0, Status='live', Symbol='LOC644456', Aliases='', description='IK cytokine, down-regulator of HLA II pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=58460170, end_position_on_the_genomic_accession=58462032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370985, CurrentID=0, Status='live', Symbol='LOC105370985', Aliases='', description='uncharacterized LOC105370985', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94583558, end_position_on_the_genomic_accession=94589215, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79521, CurrentID=0, Status='live', Symbol='OR52S1P', Aliases='', description='olfactory receptor family 52 subfamily S member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily S, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5076112, end_position_on_the_genomic_accession=5077066, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370987, CurrentID=0, Status='live', Symbol='LOC105370987', Aliases='', description='uncharacterized LOC105370987', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94724933, end_position_on_the_genomic_accession=94726513, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  196740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196740, CurrentID=0, Status='live', Symbol='VSTM4', Aliases='C10orf72', description='V-set and transmembrane domain containing 4', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49014245, end_position_on_the_genomic_accession=49129144, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  196743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196743, CurrentID=0, Status='live', Symbol='PAOX', Aliases='PAO', description='polyamine oxidase (exo-N4-amino)', other_designations='peroxisomal N1-acetyl-spermine/spermidine oxidase', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133379237, end_position_on_the_genomic_accession=133391696, orientation='plus', exon_count=8, OMIM=615853, no_hdr0=''),
  100128108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128108, CurrentID=0, Status='live', Symbol='LOC100128108', Aliases='', description='putative ubiquitin-conjugating enzyme E2Q2-like protein', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101724664, end_position_on_the_genomic_accession=101747827, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105370989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370989, CurrentID=0, Status='live', Symbol='LOC105370989', Aliases='', description='uncharacterized LOC105370989', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95001369, end_position_on_the_genomic_accession=95018127, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418667, CurrentID=0, Status='live', Symbol='LOC100418667', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644462, CurrentID=0, Status='live', Symbol='LOC644462', Aliases='', description='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 2 pseudogene', other_designations='STE20-related kinase adaptor beta pseudogene', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48435702, end_position_on_the_genomic_accession=48437897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370992, CurrentID=0, Status='live', Symbol='LOC105370992', Aliases='', description='uncharacterized LOC105370992', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133012, CurrentID=0, Status='live', Symbol='CHEK2P1', Aliases='CHEK2L1', description='checkpoint kinase 2 pseudogene 1', other_designations='CHK2 checkpoint homolog pseudogene|CHK2 checkpoint homolog-like 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11412547, end_position_on_the_genomic_accession=11418872, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374801, CurrentID=0, Status='live', Symbol='LOC105374801', Aliases='', description='uncharacterized LOC105374801', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73113111, end_position_on_the_genomic_accession=73113833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370993, CurrentID=0, Status='live', Symbol='LOC105370993', Aliases='', description='uncharacterized LOC105370993', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95463592, end_position_on_the_genomic_accession=95469889, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370995, CurrentID=0, Status='live', Symbol='LOC105370995', Aliases='', description='uncharacterized LOC105370995', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95564367, end_position_on_the_genomic_accession=95592086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  196792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196792, CurrentID=0, Status='live', Symbol='FAM24B', Aliases='', description='family with sequence similarity 24 member B', other_designations='family with sequence similarity 24, member B', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122849094, end_position_on_the_genomic_accession=122879641, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370997, CurrentID=0, Status='live', Symbol='LOC105370997', Aliases='', description='uncharacterized LOC105370997', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96170825, end_position_on_the_genomic_accession=96173957, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103164619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103164619, CurrentID=0, Status='live', Symbol='PCAT2', Aliases='CARLo-4, PCA2, TCONS_00015167', description='prostate cancer associated transcript 2 (non-protein coding)', other_designations='cancer-associated region long noncoding RNA 4', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127072694, end_position_on_the_genomic_accession=127082221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370998, CurrentID=0, Status='live', Symbol='LOC105370998', Aliases='', description='uncharacterized LOC105370998', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96044557, end_position_on_the_genomic_accession=96053241, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370999, CurrentID=0, Status='live', Symbol='LOC105370999', Aliases='', description='uncharacterized LOC105370999', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96422921, end_position_on_the_genomic_accession=96438474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128121, CurrentID=0, Status='live', Symbol='LOC100128121', Aliases='', description='60S ribosomal protein L9 pseudogene', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=144511338, end_position_on_the_genomic_accession=144522988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105447645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105447645, CurrentID=0, Status='live', Symbol='LOC105447645', Aliases='', description='uncharacterized LOC105447645', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105447646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105447646, CurrentID=0, Status='live', Symbol='PNKY', Aliases='', description='long intergenic non-protein coding RNA PNKY', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616328, no_hdr0=''),
  105447648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105447648, CurrentID=0, Status='live', Symbol='LOC105447648', Aliases='', description='uncharacterized LOC105447648', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371004, CurrentID=0, Status='live', Symbol='LOC105371004', Aliases='', description='uncharacterized LOC105371004', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97071052, end_position_on_the_genomic_accession=97102856, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105371006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371006, CurrentID=0, Status='live', Symbol='LOC105371006', Aliases='', description='uncharacterized LOC105371006', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97272209, end_position_on_the_genomic_accession=97319366, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371007, CurrentID=0, Status='live', Symbol='LOC105371007', Aliases='', description='uncharacterized LOC105371007', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97539146, end_position_on_the_genomic_accession=97576622, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373239, CurrentID=0, Status='live', Symbol='LOC105373239', Aliases='', description='uncharacterized LOC105373239', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64842760, end_position_on_the_genomic_accession=64851049, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  196872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196872, CurrentID=0, Status='live', Symbol='LINC00638', Aliases='', description='long intergenic non-protein coding RNA 638', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104821201, end_position_on_the_genomic_accession=104823718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371010, CurrentID=0, Status='live', Symbol='LOC105371010', Aliases='', description='uncharacterized LOC105371010', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98111503, end_position_on_the_genomic_accession=98131668, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  196883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196883, CurrentID=0, Status='live', Symbol='ADCY4', Aliases='AC4', description='adenylate cyclase 4', other_designations='ATP pyrophosphate-lyase 4|adenylate cyclase type IV|adenylyl cyclase 4', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24318349, end_position_on_the_genomic_accession=24335071, orientation='minus', exon_count=26, OMIM=600292, no_hdr0=''),
  285003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285003, CurrentID=0, Status='live', Symbol='MRPL36P1', Aliases='', description='mitochondrial ribosomal protein L36 pseudogene 1', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70102489, end_position_on_the_genomic_accession=70103086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128135, CurrentID=0, Status='live', Symbol='GVINP2', Aliases='', description='GTPase, very large interferon inducible pseudogene 2', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6748125, end_position_on_the_genomic_accession=6750369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  196913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196913, CurrentID=0, Status='live', Symbol='LINC01599', Aliases='C14orf183', description='long intergenic non-protein coding RNA 1599', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50007313, end_position_on_the_genomic_accession=50039884, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105371017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371017, CurrentID=0, Status='live', Symbol='LOC105371017', Aliases='', description='uncharacterized LOC105371017', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99397396, end_position_on_the_genomic_accession=99417208, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  644491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644491, CurrentID=0, Status='live', Symbol='ICE2P1', Aliases='NARG2P1', description='interactor of little elongation complex ELL subunit 2 pseudogene 1', other_designations='NMDA receptor regulated 2 pseudogene 1|interactor of little elongator complex ELL subunit 2 pseudogene 1', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188387286, end_position_on_the_genomic_accession=188389663, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371020, CurrentID=0, Status='live', Symbol='LOC105371020', Aliases='', description='uncharacterized LOC105371020', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99737425, end_position_on_the_genomic_accession=99774638, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  196951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196951, CurrentID=0, Status='live', Symbol='FAM227B', Aliases='C15orf33', description='family with sequence similarity 227 member B', other_designations='family with sequence similarity 227, member B', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49328393, end_position_on_the_genomic_accession=49620931, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  105371023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371023, CurrentID=0, Status='live', Symbol='LOC105371023', Aliases='', description='uncharacterized LOC105371023', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99952765, end_position_on_the_genomic_accession=99953693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371024, CurrentID=0, Status='live', Symbol='LOC105371024', Aliases='', description='uncharacterized LOC105371024', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100716052, end_position_on_the_genomic_accession=100832535, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422160, CurrentID=0, Status='live', Symbol='LOC100422160', Aliases='', description='olfactory receptor, family 56, subfamily B, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371025, CurrentID=0, Status='live', Symbol='LOC105371025', Aliases='', description='uncharacterized LOC105371025', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101162704, end_position_on_the_genomic_accession=101164792, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  196968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196968, CurrentID=0, Status='live', Symbol='DNM1P46', Aliases='C15orf51, DNM1DN14.2, DNM1DN14@', description='dynamin 1 pseudogene 46', other_designations='DNM1 pseudogene 46', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99790156, end_position_on_the_genomic_accession=99806927, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  404199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404199, CurrentID=0, Status='live', Symbol='PPIHP2', Aliases='', description='peptidylprolyl isomerase H (cyclophilin H) pseudogene 2', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=113522438, end_position_on_the_genomic_accession=113523185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371026, CurrentID=0, Status='live', Symbol='LOC105371026', Aliases='', description='uncharacterized LOC105371026', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101050701, end_position_on_the_genomic_accession=101151837, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105371027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371027, CurrentID=0, Status='live', Symbol='PCSK6-AS1', Aliases='', description='PCSK6 antisense RNA 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  404201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404201, CurrentID=0, Status='live', Symbol='WDFY3-AS2', Aliases='C4orf12, FBI4, NCRNA00247', description='WDFY3 antisense RNA 2', other_designations='WDFY3 antisense RNA 2 (non-protein coding)', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84966818, end_position_on_the_genomic_accession=85007015, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371029, CurrentID=0, Status='live', Symbol='LOC105371029', Aliases='', description='uncharacterized LOC105371029', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101621473, end_position_on_the_genomic_accession=101626482, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  196993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196993, CurrentID=0, Status='live', Symbol='CT62', Aliases='', description='cancer/testis antigen 62', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71110244, end_position_on_the_genomic_accession=71115500, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  404203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404203, CurrentID=0, Status='live', Symbol='SPINK6', Aliases='BUSI2, UNQ844', description='serine peptidase inhibitor, Kazal type 6', other_designations='acrosin inhibitor|kallikrein inhibitor|protease inhibitor H', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148202794, end_position_on_the_genomic_accession=148215137, orientation='plus', exon_count=5, OMIM=615868, no_hdr0=''),
  196996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196996, CurrentID=0, Status='live', Symbol='GRAMD2', Aliases='', description='GRAM domain containing 2', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72159806, end_position_on_the_genomic_accession=72197795, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105371030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371030, CurrentID=0, Status='live', Symbol='LOC105371030', Aliases='', description='uncharacterized LOC105371030', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101772694, end_position_on_the_genomic_accession=101799800, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379173, CurrentID=0, Status='live', Symbol='LOC105379173', Aliases='', description='uncharacterized LOC105379173', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131393809, end_position_on_the_genomic_accession=131417924, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  404204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404204, CurrentID=0, Status='live', Symbol='UGT1A13P', Aliases='', description='UDP glucuronosyltransferase 1 family, polypeptide A13 pseudogene', other_designations='UDP glycosyltransferase 1 family, polypeptide A13 pseudogene', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233636523, end_position_on_the_genomic_accession=233649027, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  197003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197003, CurrentID=0, Status='live', Symbol='MGC15885', Aliases='', description='uncharacterized protein MGC15885', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62637172, end_position_on_the_genomic_accession=62645181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372367, CurrentID=0, Status='live', Symbol='LOC105372367', Aliases='', description='uncharacterized LOC105372367', other_designations='CTD-2538C1.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32687089, end_position_on_the_genomic_accession=32691759, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644504, CurrentID=0, Status='live', Symbol='LOC644504', Aliases='', description='novel protein similar to chondroitin sulfate GalNAcT-2 (GALNACT-2)', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=105530405, end_position_on_the_genomic_accession=105572817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371033, CurrentID=0, Status='live', Symbol='LOC105371033', Aliases='', description='uncharacterized LOC105371033', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  65944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65944, CurrentID=0, Status='live', Symbol='OR2B8P', Aliases='OR2B8, OR6-10, dJ313I6.4, hs6M1-29P', description='olfactory receptor family 2 subfamily B member 8 pseudogene', other_designations='Olfactory receptor 2B8|olfactory receptor, family 2, subfamily B, member 8 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28053128, end_position_on_the_genomic_accession=28054265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  197021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197021, CurrentID=0, Status='live', Symbol='LCTL', Aliases='KLG, KLPH', description='lactase like', other_designations='KL lactase phlorizin hydrolase|klotho gamma|klotho/lactase-phlorizin hydrolase-related protein|lactase-like', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66547468, end_position_on_the_genomic_accession=66565979, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105379174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379174, CurrentID=0, Status='live', Symbol='LOC105379174', Aliases='', description='uncharacterized LOC105379174', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132048496, end_position_on_the_genomic_accession=132062721, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371037, CurrentID=0, Status='live', Symbol='LOC105371037', Aliases='', description='uncharacterized LOC105371037', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=417948, end_position_on_the_genomic_accession=425520, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  65975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65975, CurrentID=0, Status='live', Symbol='STK33', Aliases='', description='serine/threonine kinase 33', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8391866, end_position_on_the_genomic_accession=8594289, orientation='minus', exon_count=20, OMIM=607670, no_hdr0=''),
  65977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65977, CurrentID=0, Status='live', Symbol='PLEKHA3', Aliases='FAPP1', description='pleckstrin homology domain containing A3', other_designations='FAPP-1|PH domain-containing family A member 3|four-phosphate-adaptor protein 1|phosphatidylinositol-four-phosphate adapter protein 1|phosphoinositol 4-phosphate adapter protein 1|pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 3|pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 3', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178480410, end_position_on_the_genomic_accession=178505586, orientation='plus', exon_count=9, OMIM=607774, no_hdr0=''),
  65978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65978, CurrentID=0, Status='live', Symbol='IGSF6-DREV1', Aliases='', description='region containing immunoglobulin superfamily, member  6 and DREV1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  65979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65979, CurrentID=0, Status='live', Symbol='PHACTR4', Aliases='PPP1R124', description='phosphatase and actin regulator 4', other_designations='protein phosphatase 1, regulatory subunit 124', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28369582, end_position_on_the_genomic_accession=28500369, orientation='plus', exon_count=19, OMIM=608726, no_hdr0=''),
  65980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65980, CurrentID=0, Status='live', Symbol='BRD9', Aliases='LAVS3040, PRO9856', description='bromodomain containing 9', other_designations='rhabdomyosarcoma antigen MU-RMS-40.8|sarcoma antigen NY-SAR-29', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=863735, end_position_on_the_genomic_accession=892824, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  65981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65981, CurrentID=0, Status='live', Symbol='CAPRIN2', Aliases='C1QDC1, EEG-1, EEG1, RNG140', description='caprin family member 2', other_designations='C1q domain-containing protein 1|RNA granule protein 140|cytoplasmic activation/proliferation-associated protein 2|gastric cancer multidrug resistance-associated protein', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30709552, end_position_on_the_genomic_accession=30823235, orientation='minus', exon_count=23, OMIM=610375, no_hdr0=''),
  65982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65982, CurrentID=0, Status='live', Symbol='ZSCAN18', Aliases='ZNF447', description='zinc finger and SCAN domain containing 18', other_designations='zinc finger protein 447', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58083834, end_position_on_the_genomic_accession=58118433, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  65983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65983, CurrentID=0, Status='live', Symbol='GRAMD3', Aliases='NS3TP2', description='GRAM domain containing 3', other_designations='HCV NS3-transactivated protein 2|hepatitis C virus nonstructural protein 3-transactivating protein 2', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126360096, end_position_on_the_genomic_accession=126494364, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  65985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65985, CurrentID=0, Status='live', Symbol='AACS', Aliases='ACSF1, SUR-5', description='acetoacetyl-CoA synthetase', other_designations='acetoacetate-CoA ligase|acyl-CoA synthetase family member 1|homolog of C. elegans supressor of ras 5 (sur-5)|protein sur-5 homolog', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125065379, end_position_on_the_genomic_accession=125143327, orientation='plus', exon_count=19, OMIM=614364, no_hdr0=''),
  65986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65986, CurrentID=0, Status='live', Symbol='ZBTB10', Aliases='RINZF', description='zinc finger and BTB domain containing 10', other_designations='zinc finger protein RIN ZF|zinc finger protein RINZF', map_location='8q13-q21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80485619, end_position_on_the_genomic_accession=80526265, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  65987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65987, CurrentID=0, Status='live', Symbol='KCTD14', Aliases='', description='potassium channel tetramerization domain containing 14', other_designations='potassium channel tetramerisation domain containing 14', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78015715, end_position_on_the_genomic_accession=78046191, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  65988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65988, CurrentID=0, Status='live', Symbol='ZNF747', Aliases='', description='zinc finger protein 747', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30530367, end_position_on_the_genomic_accession=30534970, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  65989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65989, CurrentID=0, Status='live', Symbol='DLK2', Aliases='DLK-2, EGFL9', description='delta-like 2 homolog (Drosophila)', other_designations='EGF-like protein 9|EGF-like-domain, multiple 9|epidermal growth factor-like protein 9', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43450352, end_position_on_the_genomic_accession=43456632, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  65990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65990, CurrentID=0, Status='live', Symbol='FAM173A', Aliases='C16orf24', description='family with sequence similarity 173 member A', other_designations='family with sequence similarity 173, member A', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=720581, end_position_on_the_genomic_accession=722590, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65991, CurrentID=0, Status='live', Symbol='FUNDC2', Aliases='DC44, HCBP6, HCC3, PD03104', description='FUN14 domain containing 2', other_designations='HCC-3|cervical cancer oncogene 3|cervical cancer proto-oncogene 3 protein|hepatitis C virus core-binding protein 6', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155026789, end_position_on_the_genomic_accession=155056916, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65992, CurrentID=0, Status='live', Symbol='DDRGK1', Aliases='C20orf116, UFBP1, dJ1187M17.3', description='DDRGK domain containing 1', other_designations='Dashurin|UFM1-binding and PCI domain-containing protein 1|UFM1-binding protein 1 containing a PCI domain', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3190366, end_position_on_the_genomic_accession=3204649, orientation='minus', exon_count=9, OMIM=616177, no_hdr0=''),
  65993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65993, CurrentID=0, Status='live', Symbol='MRPS34', Aliases='MRP-S12, MRP-S34, MRPS12', description='mitochondrial ribosomal protein S34', other_designations='S34mt|mitochondrial 28S ribosomal protein S34', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1771895, end_position_on_the_genomic_accession=1773139, orientation='minus', exon_count=3, OMIM=611994, no_hdr0=''),
  65996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65996, CurrentID=0, Status='live', Symbol='CENPBD1P1', Aliases='', description='CENPB DNA-binding domains containing 1 pseudogene 1', other_designations='CENPBD1 pseudogene 1', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58575399, end_position_on_the_genomic_accession=58584395, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  65997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65997, CurrentID=0, Status='live', Symbol='RASL11B', Aliases='', description='RAS-like family 11 member B', other_designations='RAS-like, family 11, member B', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52862328, end_position_on_the_genomic_accession=52866835, orientation='plus', exon_count=4, OMIM=612404, no_hdr0=''),
  65998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65998, CurrentID=0, Status='live', Symbol='C11orf95', Aliases='', description='chromosome 11 open reading frame 95', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63759892, end_position_on_the_genomic_accession=63768641, orientation='minus', exon_count=5, OMIM=615699, no_hdr0=''),
  65999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65999, CurrentID=0, Status='live', Symbol='LRRC61', Aliases='HSPC295', description='leucine rich repeat containing 61', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150322639, end_position_on_the_genomic_accession=150338156, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66000, CurrentID=0, Status='live', Symbol='TMEM108', Aliases='CT124', description='transmembrane protein 108', other_designations='cancer/testis antigen 124', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133038288, end_position_on_the_genomic_accession=133397775, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66001, CurrentID=0, Status='live', Symbol='MLVI4', Aliases='', description='Moloney leukemia virus integration site 4, mouse, homolog of', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  66002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66002, CurrentID=0, Status='live', Symbol='CYP4F12', Aliases='CYPIVF12, F22329_1', description='cytochrome P450 family 4 subfamily F member 12', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 12|cytochrome P450, subfamily IVF, polypeptide 12', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15673018, end_position_on_the_genomic_accession=15698819, orientation='plus', exon_count=16, OMIM=611485, no_hdr0=''),
  105371043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371043, CurrentID=0, Status='live', Symbol='LOC105371043', Aliases='', description='uncharacterized LOC105371043', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1200823, end_position_on_the_genomic_accession=1201711, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66004, CurrentID=0, Status='live', Symbol='LYNX1', Aliases='SLURP2', description='Ly6/neurotoxin 1', other_designations='Ly-6 neurotoxin-like protein 1|SLURP-2|secreted Ly-6/uPAR domain-containing protein 2|secreted Ly-6/uPAR-related protein 2|secreted Ly6/uPAR related protein 2', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142764334, end_position_on_the_genomic_accession=142777848, orientation='minus', exon_count=9, OMIM=606110, no_hdr0=''),
  66005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66005, CurrentID=0, Status='live', Symbol='CHID1', Aliases='GL008, SI-CLP, SICLP', description='chitinase domain containing 1', other_designations='stabilin-1 interacting chitinase-like protein|stabilin-1-interacting chitinase-like protein', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=867859, end_position_on_the_genomic_accession=915058, orientation='minus', exon_count=19, OMIM=615692, no_hdr0=''),
  404217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404217, CurrentID=0, Status='live', Symbol='CTXN1', Aliases='CTXN', description='cortexin 1', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7924496, end_position_on_the_genomic_accession=7926166, orientation='minus', exon_count=2, OMIM=600135, no_hdr0=''),
  66008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66008, CurrentID=0, Status='live', Symbol='TRAK2', Aliases='ALS2CR3, CALS-C, GRIF-1, GRIF1, MILT2, OIP98', description='trafficking protein, kinesin binding 2', other_designations='O-linked N-acetylglucosamine transferase interacting protein 98|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 3|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 3 protein|gamma-aminobutyric acid(A) receptor-interacting factor|milton homolog 2', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201377207, end_position_on_the_genomic_accession=201451596, orientation='minus', exon_count=18, OMIM=607334, no_hdr0=''),
  105371044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371044, CurrentID=0, Status='live', Symbol='LOC105371044', Aliases='', description='uncharacterized LOC105371044', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1425551, end_position_on_the_genomic_accession=1428843, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371045, CurrentID=0, Status='live', Symbol='LOC105371045', Aliases='gs104', description='uncharacterized LOC105371045', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1430309, end_position_on_the_genomic_accession=1433421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379176, CurrentID=0, Status='live', Symbol='LOC105379176', Aliases='', description='uncharacterized LOC105379176', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132679217, end_position_on_the_genomic_accession=132682629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  404220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404220, CurrentID=0, Status='live', Symbol='C6orf201', Aliases='dJ1013A10.5', description='chromosome 6 open reading frame 201', other_designations='protein MGC87625', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4079206, end_position_on_the_genomic_accession=4130766, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105371047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371047, CurrentID=0, Status='live', Symbol='LOC105371047', Aliases='', description='uncharacterized LOC105371047', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1530388, end_position_on_the_genomic_accession=1533569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371048, CurrentID=0, Status='live', Symbol='LOC105371048', Aliases='', description='uncharacterized LOC105371048', other_designations='XX-DJ76P10__A.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1888947, end_position_on_the_genomic_accession=1890434, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66035, CurrentID=0, Status='live', Symbol='SLC2A11', Aliases='GLUT10, GLUT11', description='solute carrier family 2 (facilitated glucose transporter), member 11', other_designations='facilitative glucose transporter GLUT11|glucose transporter protein 10|glucose transporter protein 11|glucose transporter type 10|glucose transporter type 11|glucose transporter-like protein XI', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23856703, end_position_on_the_genomic_accession=23886112, orientation='plus', exon_count=14, OMIM=610367, no_hdr0=''),
  66036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66036, CurrentID=0, Status='live', Symbol='MTMR9', Aliases='C8orf9, LIP-STYX, MTMR8', description='myotubularin related protein 9', other_designations='myotubularin related protein 8', map_location='8p23-p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11284491, end_position_on_the_genomic_accession=11328146, orientation='plus', exon_count=12, OMIM=606260, no_hdr0=''),
  66037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=66037, CurrentID=0, Status='live', Symbol='BOLL', Aliases='BOULE', description='boule homolog, RNA binding protein', other_designations='bol, boule-like|boule-like RNA binding protein|boule-like RNA-binding protein', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197726879, end_position_on_the_genomic_accession=197786574, orientation='minus', exon_count=15, OMIM=606165, no_hdr0=''),
  100533892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533892, CurrentID=0, Status='live', Symbol='LOC100533892', Aliases='', description='ubiquilin 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371050, CurrentID=0, Status='live', Symbol='LOC105371050', Aliases='', description='uncharacterized LOC105371050', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2451612, end_position_on_the_genomic_accession=2460022, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128171, CurrentID=0, Status='live', Symbol='LOC100128171', Aliases='', description='mitochondrial fission regulator 1 pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66359581, end_position_on_the_genomic_accession=66361865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371052, CurrentID=0, Status='live', Symbol='LOC105371052', Aliases='', description='uncharacterized LOC105371052', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2464150, end_position_on_the_genomic_accession=2471062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  197131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197131, CurrentID=0, Status='live', Symbol='UBR1', Aliases='JBS', description='ubiquitin protein ligase E3 component n-recognin 1', other_designations='E3a ligase|N-recognin-1|ubiquitin ligase E3 alpha-I|ubiquitin-protein ligase E3-alpha|ubiquitin-protein ligase E3-alpha-1|ubiquitin-protein ligase E3-alpha-I', map_location='15q13', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42942897, end_position_on_the_genomic_accession=43106088, orientation='minus', exon_count=47, OMIM=605981, no_hdr0=''),
  197135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197135, CurrentID=0, Status='live', Symbol='PATL2', Aliases='Pat1a, hPat1a', description='protein associated with topoisomerase II homolog 2 (yeast)', other_designations='PAT1-like protein 2|protein PAT1 homolog a', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44665732, end_position_on_the_genomic_accession=44711392, orientation='minus', exon_count=22, OMIM=614661, no_hdr0=''),
  100128174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128174, CurrentID=0, Status='live', Symbol='LOC100128174', Aliases='', description='defensin beta 131 pseudogene', other_designations='defensin, beta 131 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12043503, end_position_on_the_genomic_accession=12049625, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  648149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648149, CurrentID=0, Status='live', Symbol='LOC648149', Aliases='', description='uncharacterized LOC648149', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=210262332, end_position_on_the_genomic_accession=210265320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128175, CurrentID=0, Status='live', Symbol='LOC100128175', Aliases='', description='uncharacterized LOC100128175', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=116322448, end_position_on_the_genomic_accession=116322937, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371056, CurrentID=0, Status='live', Symbol='LOC105371056', Aliases='', description='uncharacterized LOC105371056', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3006567, end_position_on_the_genomic_accession=3007459, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128177, CurrentID=0, Status='live', Symbol='NDUFS5P5', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 5', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 5', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118400396, end_position_on_the_genomic_accession=118400868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373241, CurrentID=0, Status='live', Symbol='LOC105373241', Aliases='', description='uncharacterized LOC105373241', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=67745789, end_position_on_the_genomic_accession=67761958, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371059, CurrentID=0, Status='live', Symbol='LOC105371059', Aliases='', description='uncharacterized LOC105371059', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3364676, end_position_on_the_genomic_accession=3399669, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371061, CurrentID=0, Status='live', Symbol='LOC105371061', Aliases='', description='uncharacterized LOC105371061', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3651928, end_position_on_the_genomic_accession=3657767, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929657, CurrentID=0, Status='live', Symbol='LOC101929657', Aliases='', description='uncharacterized LOC101929657', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52128895, end_position_on_the_genomic_accession=52132723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  197187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197187, CurrentID=0, Status='live', Symbol='SNAI3-AS1', Aliases='', description='SNAI3 antisense RNA 1', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88663373, end_position_on_the_genomic_accession=88687186, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371062, CurrentID=0, Status='live', Symbol='LOC105371062', Aliases='', description='uncharacterized LOC105371062', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4180117, end_position_on_the_genomic_accession=4181987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371063, CurrentID=0, Status='live', Symbol='LOC105371063', Aliases='', description='putative uncharacterized protein encoded by LINC00596', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4254514, end_position_on_the_genomic_accession=4255336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  197196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197196, CurrentID=0, Status='live', Symbol='LINC00311', Aliases='NCRNA00311, TMEM148', description='long intergenic non-protein coding RNA 311', other_designations='transmembrane protein 148', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85282958, end_position_on_the_genomic_accession=85288082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128184, CurrentID=0, Status='live', Symbol='LOC100128184', Aliases='', description='uncharacterized LOC100128184', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=60382268, end_position_on_the_genomic_accession=60383215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128185, CurrentID=0, Status='live', Symbol='LOC100128185', Aliases='', description='PNAS-19', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=687886, end_position_on_the_genomic_accession=688026, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371066, CurrentID=0, Status='live', Symbol='LOC105371066', Aliases='', description='uncharacterized LOC105371066', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5213787, end_position_on_the_genomic_accession=5220692, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371068, CurrentID=0, Status='live', Symbol='LOC105371068', Aliases='', description='uncharacterized LOC105371068', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=7042801, end_position_on_the_genomic_accession=7055584, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371069, CurrentID=0, Status='live', Symbol='LOC105371069', Aliases='', description='uncharacterized LOC105371069', other_designations='CTD-2535I10.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=7855125, end_position_on_the_genomic_accession=7894384, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371070, CurrentID=0, Status='live', Symbol='LOC105371070', Aliases='', description='uncharacterized LOC105371070', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=7982101, end_position_on_the_genomic_accession=8112633, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128191, CurrentID=0, Status='live', Symbol='TMPO-AS1', Aliases='', description='TMPO antisense RNA 1', other_designations='TMPO antisense RNA 1 (non-protein coding)', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98512973, end_position_on_the_genomic_accession=98516226, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11029, CurrentID=0, Status='live', Symbol='SERPINB8P1', Aliases='PI8L1, SERPINB11, SERPINBP1', description='serpin peptidase inhibitor, clade B8 (ovalbumin) pseudogene 1', other_designations='protease inhibitor 8 (ovalbumin type)-like 1|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin) pseudogene 1|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 11|serpin peptidase inhibitor, clade B (ovalbumin) pseudogene 1', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3048838, end_position_on_the_genomic_accession=3049234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421289, CurrentID=0, Status='live', Symbol='LOC100421289', Aliases='', description='prolylcarboxypeptidase (angiotensinase C) pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101520830, end_position_on_the_genomic_accession=101521994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929237, CurrentID=0, Status='live', Symbol='LOC101929237', Aliases='', description='uncharacterized LOC101929237', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22877972, end_position_on_the_genomic_accession=22888022, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  197257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197257, CurrentID=0, Status='live', Symbol='LDHD', Aliases='DLD', description='lactate dehydrogenase D', other_designations='D-lactate dehydrogenase', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75111860, end_position_on_the_genomic_accession=75116772, orientation='minus', exon_count=11, OMIM=607490, no_hdr0=''),
  197258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197258, CurrentID=0, Status='live', Symbol='FUK', Aliases='1110046B12Rik', description='fucokinase', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70454584, end_position_on_the_genomic_accession=70480274, orientation='plus', exon_count=27, OMIM=608675, no_hdr0=''),
  197259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197259, CurrentID=0, Status='live', Symbol='MLKL', Aliases='hMLKL', description='mixed lineage kinase domain-like', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74671855, end_position_on_the_genomic_accession=74701148, orientation='minus', exon_count=11, OMIM=615153, no_hdr0=''),
  100128686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128686, CurrentID=0, Status='live', Symbol='LOC100128686', Aliases='', description='programmed cell death 10 pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50859530, end_position_on_the_genomic_accession=50860059, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379182, CurrentID=0, Status='live', Symbol='LOC105379182', Aliases='', description='uncharacterized LOC105379182', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133927734, end_position_on_the_genomic_accession=133954074, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371076, CurrentID=0, Status='live', Symbol='LOC105371076', Aliases='', description='uncharacterized LOC105371076', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10031637, end_position_on_the_genomic_accession=10037297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128198, CurrentID=0, Status='live', Symbol='LOC100128198', Aliases='', description='uncharacterized LOC100128198', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.7', start_position_on_the_genomic_accession=32577676, end_position_on_the_genomic_accession=32578343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105317032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105317032, CurrentID=0, Status='live', Symbol='OI16', Aliases='C16DELp11.2, DEL16p11.2', description='Osteogenesis imperfecta, type XVI', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616229, no_hdr0=''),
  390099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390099, CurrentID=0, Status='live', Symbol='LOC390099', Aliases='', description='MAS-related GPR, member X3 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18910234, end_position_on_the_genomic_accession=18911422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379183, CurrentID=0, Status='live', Symbol='LOC105379183', Aliases='', description='uncharacterized LOC105379183', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133968689, end_position_on_the_genomic_accession=133970223, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371082, CurrentID=0, Status='live', Symbol='LOC105371082', Aliases='', description='uncharacterized LOC105371082', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11249046, end_position_on_the_genomic_accession=11396486, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371083, CurrentID=0, Status='live', Symbol='LOC105371083', Aliases='', description='uncharacterized LOC105371083', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11341809, end_position_on_the_genomic_accession=11345299, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  197320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197320, CurrentID=0, Status='live', Symbol='ZNF778', Aliases='', description='zinc finger protein 778', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89202038, end_position_on_the_genomic_accession=89232637, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  197322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197322, CurrentID=0, Status='live', Symbol='ACSF3', Aliases='', description='acyl-CoA synthetase family member 3', other_designations='malonyl-CoA synthetase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89093809, end_position_on_the_genomic_accession=89155846, orientation='plus', exon_count=17, OMIM=614245, no_hdr0=''),
  101515984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101515984, CurrentID=0, Status='live', Symbol='LINC01080', Aliases='TCONS_00021856', description='long intergenic non-protein coding RNA 1080', other_designations='Long intergenic noncoding RNA TCONS_00021856', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80011077, end_position_on_the_genomic_accession=80026663, orientation='plus', exon_count=2, OMIM=615230, no_hdr0=''),
  197331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197331, CurrentID=0, Status='live', Symbol='TUBB8P7', Aliases='', description='tubulin beta 8 class VIII pseudogene 7', other_designations='HSA16q24 beta-tubulin 4Q pseudogene|tubulin, beta 8 class VIII pseudogene 7', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90094023, end_position_on_the_genomic_accession=90096358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  197335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197335, CurrentID=0, Status='live', Symbol='WDR90', Aliases='C16orf15, C16orf16, C16orf17, C16orf18, C16orf19', description='WD repeat domain 90', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=649363, end_position_on_the_genomic_accession=667833, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  197342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197342, CurrentID=0, Status='live', Symbol='EME2', Aliases='SLX2B, gs125', description='essential meiotic structure-specific endonuclease subunit 2', other_designations='SLX2 structure-specific endonuclease subunit homolog B|essential meiotic endonuclease 1 homolog 2|homolog of yeast EME1 endonuclease 2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1773228, end_position_on_the_genomic_accession=1776238, orientation='plus', exon_count=8, OMIM=610886, no_hdr0=''),
  197350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197350, CurrentID=0, Status='live', Symbol='CASP16P', Aliases='CASP16', description='caspase 16, pseudogene', other_designations='caspase 16, apoptosis-related cysteine peptidase pseudogene|putative caspase-14-like pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3144015, end_position_on_the_genomic_accession=3149960, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100128210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128210, CurrentID=0, Status='live', Symbol='LOC100128210', Aliases='', description='ubiquitin C-terminal hydrolase L1 pseudogene', other_designations='ubiquitin carboxyl-terminal esterase L1 (ubiquitin thiolesterase) pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56495761, end_position_on_the_genomic_accession=56496413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  197358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197358, CurrentID=0, Status='live', Symbol='NLRC3', Aliases='CLR16.2, NOD3', description='NLR family, CARD domain containing 3', other_designations='CARD15-like protein|NOD-like receptor C3|caterpiller protein 16.2|nucleotide-binding oligomerization domain, leucine rich repeat and CARD domain containing 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3539033, end_position_on_the_genomic_accession=3578373, orientation='minus', exon_count=25, OMIM=615648, no_hdr0=''),
  644563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644563, CurrentID=0, Status='live', Symbol='LOC644563', Aliases='', description='general transcription factor IIIC subunit 6 pseudogene', other_designations='general transcription factor IIIC, polypeptide 6, alpha 35kDa pseudogene', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107131574, end_position_on_the_genomic_accession=107132367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  197370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197370, CurrentID=0, Status='live', Symbol='NSMCE1', Aliases='NSE1', description='NSE1 homolog, SMC5-SMC6 complex component', other_designations='non-SMC element 1 homolog', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27224994, end_position_on_the_genomic_accession=27268809, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100133032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133032, CurrentID=0, Status='live', Symbol='LOC100133032', Aliases='', description='uncharacterized LOC100133032', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=47620433, end_position_on_the_genomic_accession=47620876, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371093, CurrentID=0, Status='live', Symbol='LOC105371093', Aliases='', description='uncharacterized LOC105371093', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=13882848, end_position_on_the_genomic_accession=13926606, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  390650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390650, CurrentID=0, Status='live', Symbol='OR4F14P', Aliases='OR4F14', description='olfactory receptor family 4 subfamily F member 14 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 14 pseudogene', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101828741, end_position_on_the_genomic_accession=101829676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371094, CurrentID=0, Status='live', Symbol='LOC105371094', Aliases='', description='uncharacterized LOC105371094', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14486796, end_position_on_the_genomic_accession=14489038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  197387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197387, CurrentID=0, Status='live', Symbol='LOC197387', Aliases='', description='uncharacterized LOC197387', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370457, CurrentID=0, Status='live', Symbol='LOC105370457', Aliases='', description='uncharacterized LOC105370457', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38878702, end_position_on_the_genomic_accession=38919201, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421091, CurrentID=0, Status='live', Symbol='LOC100421091', Aliases='', description='LSM14A, SCD6 homolog A (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75611054, end_position_on_the_genomic_accession=75612001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371097, CurrentID=0, Status='live', Symbol='LOC105371097', Aliases='', description='uncharacterized LOC105371097', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15282376, end_position_on_the_genomic_accession=15300457, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374822, CurrentID=0, Status='live', Symbol='LOC105374822', Aliases='', description='uncharacterized LOC105374822', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78900475, end_position_on_the_genomic_accession=78931140, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  197407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=197407, CurrentID=0, Status='live', Symbol='ZNF48', Aliases='ZNF553', description='zinc finger protein 48', other_designations='zinc finger protein 553', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30378312, end_position_on_the_genomic_accession=30400108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371099, CurrentID=0, Status='live', Symbol='LOC105371099', Aliases='', description='uncharacterized LOC105371099', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16294416, end_position_on_the_genomic_accession=16312316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371100, CurrentID=0, Status='live', Symbol='LOC105371100', Aliases='', description='uncharacterized LOC105371100', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16208605, end_position_on_the_genomic_accession=16218808, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379187, CurrentID=0, Status='live', Symbol='LOC105379187', Aliases='', description='uncharacterized LOC105379187', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134874906, end_position_on_the_genomic_accession=134875546, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  374387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374387, CurrentID=0, Status='live', Symbol='DKFZp779M0652', Aliases='', description='uncharacterized DKFZp779M0652', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45771432, end_position_on_the_genomic_accession=45772358, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371101, CurrentID=0, Status='live', Symbol='LOC105371101', Aliases='', description='uncharacterized LOC105371101', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16143753, end_position_on_the_genomic_accession=16164106, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371104, CurrentID=0, Status='live', Symbol='LOC105371104', Aliases='', description='uncharacterized LOC105371104', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421093, CurrentID=0, Status='live', Symbol='LOC100421093', Aliases='', description='karyopherin alpha 3 (importin alpha 4) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85286339, end_position_on_the_genomic_accession=85290753, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644578, CurrentID=0, Status='live', Symbol='LARP1BP1', Aliases='', description='La ribonucleoprotein domain family member 1B pseudogene 1', other_designations='La ribonucleoprotein domain family, member 1B pseudogene 1', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=63350017, end_position_on_the_genomic_accession=63351474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927913, CurrentID=0, Status='live', Symbol='LOC101927913', Aliases='', description='uncharacterized LOC101927913', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4774526, end_position_on_the_genomic_accession=4775548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371107, CurrentID=0, Status='live', Symbol='LOC105371107', Aliases='', description='uncharacterized LOC105371107', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17826026, end_position_on_the_genomic_accession=17831835, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11065, CurrentID=0, Status='live', Symbol='UBE2C', Aliases='UBCH10, dJ447F3.2', description='ubiquitin conjugating enzyme E2C', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme C|E2 ubiquitin-conjugating enzyme C|cyclin-selective ubiquitin carrier protein|mitotic-specific ubiquitin-conjugating enzyme|ubiquitin-conjugating enzyme E2C|ubiquitin-protein ligase C', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45812576, end_position_on_the_genomic_accession=45816957, orientation='plus', exon_count=8, OMIM=605574, no_hdr0=''),
  105371111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371111, CurrentID=0, Status='live', Symbol='LOC105371111', Aliases='', description='uncharacterized LOC105371111', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18820569, end_position_on_the_genomic_accession=18827194, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644584, CurrentID=0, Status='live', Symbol='LOC644584', Aliases='', description='EWS RNA binding protein 1 pseudogene', other_designations='EWS RNA-binding protein 1 pseudogene|Ewing sarcoma breakpoint region 1 pseudogene', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35880956, end_position_on_the_genomic_accession=35881719, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374825, CurrentID=0, Status='live', Symbol='LOC105374825', Aliases='', description='uncharacterized LOC105374825', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=80028006, end_position_on_the_genomic_accession=80035348, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  103875461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103875461, CurrentID=0, Status='live', Symbol='SPDT', Aliases='', description='Spondyloepiphyseal dysplasia tarda, autosomal dominant', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=184100, no_hdr0=''),
  644588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644588, CurrentID=0, Status='live', Symbol='DNAJA1P3', Aliases='', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 3', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 3', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107351487, end_position_on_the_genomic_accession=107352974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419545, CurrentID=0, Status='live', Symbol='LOC100419545', Aliases='', description='zinc finger protein 35 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42466691, end_position_on_the_genomic_accession=42468522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644589, CurrentID=0, Status='live', Symbol='DNAJC19P9', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 9', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 9', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45290036, end_position_on_the_genomic_accession=45290383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128238, CurrentID=0, Status='live', Symbol='LOC100128238', Aliases='', description='KIAA0895-like pseudogene', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10724618, end_position_on_the_genomic_accession=10725462, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128239, CurrentID=0, Status='live', Symbol='LOC100128239', Aliases='', description='uncharacterized LOC100128239', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134032272, end_position_on_the_genomic_accession=134041341, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371120, CurrentID=0, Status='live', Symbol='LOC105371120', Aliases='', description='uncharacterized LOC105371120', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20580371, end_position_on_the_genomic_accession=20583653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128242, CurrentID=0, Status='live', Symbol='LOC100128242', Aliases='', description='uncharacterized LOC100128242', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123430236, end_position_on_the_genomic_accession=123435715, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371124, CurrentID=0, Status='live', Symbol='LOC105371124', Aliases='', description='uncharacterized LOC105371124', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21252996, end_position_on_the_genomic_accession=21277422, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377075, CurrentID=0, Status='live', Symbol='LOC105377075', Aliases='', description='uncharacterized LOC105377075', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47599334, end_position_on_the_genomic_accession=47615687, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133039, CurrentID=0, Status='live', Symbol='LOC100133039', Aliases='', description='uncharacterized LOC100133039', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=11603076, end_position_on_the_genomic_accession=11605949, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371128, CurrentID=0, Status='live', Symbol='LOC105371128', Aliases='', description='serine/threonine-protein kinase SMG1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371129, CurrentID=0, Status='live', Symbol='LOC105371129', Aliases='', description='uncharacterized LOC105371129', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22086201, end_position_on_the_genomic_accession=22092293, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371130, CurrentID=0, Status='live', Symbol='MFSD13B', Aliases='', description='major facilitator superfamily domain containing 13B', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22382084, end_position_on_the_genomic_accession=22407457, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927914, CurrentID=0, Status='live', Symbol='LOC101927914', Aliases='', description='uncharacterized LOC101927914', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157466231, end_position_on_the_genomic_accession=157499716, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128252, CurrentID=0, Status='live', Symbol='ZNF667-AS1', Aliases='', description='ZNF667 antisense RNA 1 (head to head)', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56477874, end_position_on_the_genomic_accession=56495437, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371133, CurrentID=0, Status='live', Symbol='LOC105371133', Aliases='', description='uncharacterized LOC105371133', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21399787, end_position_on_the_genomic_accession=21424972, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128254, CurrentID=0, Status='live', Symbol='FAM90A28P', Aliases='', description='family with sequence similarity 90 member A28, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53299048, end_position_on_the_genomic_accession=53302126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371137, CurrentID=0, Status='live', Symbol='LOC105371137', Aliases='', description='uncharacterized LOC105371137', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371138, CurrentID=0, Status='live', Symbol='LOC105371138', Aliases='', description='uncharacterized LOC105371138', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371140, CurrentID=0, Status='live', Symbol='LOC105371140', Aliases='', description='uncharacterized LOC105371140', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23560806, end_position_on_the_genomic_accession=23562004, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371142, CurrentID=0, Status='live', Symbol='LOC105371142', Aliases='', description='uncharacterized LOC105371142', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24236053, end_position_on_the_genomic_accession=24252860, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  643802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643802, CurrentID=0, Status='live', Symbol='LOC643802', Aliases='', description='u3 small nucleolar ribonucleoprotein protein MPP10-like', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53369362, end_position_on_the_genomic_accession=53371129, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644616, CurrentID=0, Status='live', Symbol='ILF2P2', Aliases='', description='interleukin enhancer binding factor 2 pseudogene 2', other_designations='interleukin enhancer binding factor 2, 45kDa pseudogene', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36053826, end_position_on_the_genomic_accession=36055391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371145, CurrentID=0, Status='live', Symbol='LOC105371145', Aliases='', description='uncharacterized LOC105371145', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25069454, end_position_on_the_genomic_accession=25083036, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371146, CurrentID=0, Status='live', Symbol='LOC105371146', Aliases='', description='uncharacterized LOC105371146', other_designations='CTA-249B10.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25418283, end_position_on_the_genomic_accession=25433666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371147, CurrentID=0, Status='live', Symbol='LOC105371147', Aliases='', description='uncharacterized LOC105371147', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25490962, end_position_on_the_genomic_accession=25503824, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128268, CurrentID=0, Status='live', Symbol='LOC100128268', Aliases='', description='actin-related protein T1 pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128126972, end_position_on_the_genomic_accession=128128410, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371151, CurrentID=0, Status='live', Symbol='LOC105371151', Aliases='', description='uncharacterized LOC105371151', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26299975, end_position_on_the_genomic_accession=26302012, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371153, CurrentID=0, Status='live', Symbol='LOC105371153', Aliases='', description='uncharacterized LOC105371153', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27041193, end_position_on_the_genomic_accession=27048445, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100131490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131490, CurrentID=0, Status='live', Symbol='LOC100131490', Aliases='', description='uncharacterized LOC100131490', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371154, CurrentID=0, Status='live', Symbol='LOC105371154', Aliases='', description='uncharacterized LOC105371154', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27061984, end_position_on_the_genomic_accession=27083988, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128275, CurrentID=0, Status='live', Symbol='SYF2P2', Aliases='', description='SYF2 pre-mRNA-splicing factor pseudogene 2', other_designations='SYF2 homolog, RNA splicing factor pseudogene 2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35010735, end_position_on_the_genomic_accession=35011269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128276, CurrentID=0, Status='live', Symbol='LOC100128276', Aliases='', description='uncharacterized LOC100128276', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128849172, end_position_on_the_genomic_accession=128853404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060254, CurrentID=0, Status='live', Symbol='LOC101060254', Aliases='', description='myomegalin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145507430, end_position_on_the_genomic_accession=145539922, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105371157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371157, CurrentID=0, Status='live', Symbol='LOC105371157', Aliases='', description='uncharacterized LOC105371157', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27746108, end_position_on_the_genomic_accession=27747890, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371159, CurrentID=0, Status='live', Symbol='LOC105371159', Aliases='', description='uncharacterized LOC105371159', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29291886, end_position_on_the_genomic_accession=29294351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271842, CurrentID=0, Status='live', Symbol='LOC100271842', Aliases='', description='phosphatidylinositol 4-kinase type 2 beta pseudogene', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227278450, end_position_on_the_genomic_accession=227279517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371160, CurrentID=0, Status='live', Symbol='LOC105371160', Aliases='', description='uncharacterized LOC105371160', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29174987, end_position_on_the_genomic_accession=29183990, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128281, CurrentID=0, Status='live', Symbol='LOC100128281', Aliases='', description='uncharacterized LOC100128281', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=7716106, end_position_on_the_genomic_accession=7717998, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  103482523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482523, CurrentID=0, Status='live', Symbol='TRUND-NNN2-1', Aliases='', description='transfer RNA-undetermined (NNN) 2-1', other_designations='tRNA-Und-NNN-2-1', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482524, CurrentID=0, Status='live', Symbol='TRUND-NNN5-1', Aliases='', description='transfer RNA-undetermined (NNN) 5-1', other_designations='tRNA-Und-NNN-5-1', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482525, CurrentID=0, Status='live', Symbol='TRUND-NNN6-1', Aliases='', description='transfer RNA-undetermined (NNN) 6-1', other_designations='tRNA-Und-NNN-6-1', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482526, CurrentID=0, Status='live', Symbol='TRUND-NNN9-1', Aliases='', description='transfer RNA-undetermined (NNN) 9-1', other_designations='tRNA-Und-NNN-9-1', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482527, CurrentID=0, Status='live', Symbol='TRUND-NNN3-1', Aliases='', description='transfer RNA-undetermined (NNN) 3-1', other_designations='tRNA-Und-NNN-3-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482528, CurrentID=0, Status='live', Symbol='TRUND-NNN8-1', Aliases='', description='transfer RNA-undetermined (NNN) 8-1', other_designations='tRNA-Und-NNN-8-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482529, CurrentID=0, Status='live', Symbol='TRUND-NNN1-1', Aliases='', description='transfer RNA-undetermined (NNN) 1-1', other_designations='tRNA-Und-NNN-1-1', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482530, CurrentID=0, Status='live', Symbol='TRUND-NNN4-1', Aliases='', description='transfer RNA-undetermined (NNN) 4-1', other_designations='tRNA-Und-NNN-4-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482531, CurrentID=0, Status='live', Symbol='TRUND-NNN7-1', Aliases='', description='transfer RNA-undetermined (NNN) 7-1', other_designations='tRNA-Und-NNN-7-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103482532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103482532, CurrentID=0, Status='live', Symbol='TRUND-NNN10-1', Aliases='', description='transfer RNA-undetermined (NNN) 10-1', other_designations='tRNA-Und-NNN-10-1', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128284, CurrentID=0, Status='live', Symbol='LOC100128284', Aliases='', description='nipsnap homolog 1 (C. elegans) pseudogene', other_designations='glioblastoma amplified sequence pseudogene', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5730318, end_position_on_the_genomic_accession=5731808, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128285, CurrentID=0, Status='live', Symbol='DNM1P35', Aliases='DNM1DN8-2, DNM1DN8.2, DNM1DN8@, FKSG88', description='dynamin 1 pseudogene 35', other_designations='DNM1 pseudogene 35|DNM1DN8.2 duplicon', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75727670, end_position_on_the_genomic_accession=75740077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374836, CurrentID=0, Status='live', Symbol='LOC105374836', Aliases='', description='uncharacterized LOC105374836', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84901643, end_position_on_the_genomic_accession=84907108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371172, CurrentID=0, Status='live', Symbol='LOC105371172', Aliases='', description='uncharacterized LOC105371172', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143323047, end_position_on_the_genomic_accession=143327030, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128293, CurrentID=0, Status='live', Symbol='LOC100128293', Aliases='', description='histone-lysine N-methyltransferase SUV39H2-like', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67933501, end_position_on_the_genomic_accession=67934676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422166, CurrentID=0, Status='live', Symbol='LOC100422166', Aliases='', description='olfactory receptor, family 4, subfamily A, member 16 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371177, CurrentID=0, Status='live', Symbol='LOC105371177', Aliases='', description='uncharacterized LOC105371177', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32693599, end_position_on_the_genomic_accession=32694298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371178, CurrentID=0, Status='live', Symbol='LOC105371178', Aliases='', description='uncharacterized LOC105371178', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32638675, end_position_on_the_genomic_accession=32666543, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100128299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128299, CurrentID=0, Status='live', Symbol='SOCS5P2', Aliases='', description='suppressor of cytokine signaling 5 pseudogene 2', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135447446, end_position_on_the_genomic_accession=135449797, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371180, CurrentID=0, Status='live', Symbol='LOC105371180', Aliases='', description='uncharacterized LOC105371180', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32356877, end_position_on_the_genomic_accession=32359017, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379203, CurrentID=0, Status='live', Symbol='LOC105379203', Aliases='', description='uncharacterized LOC105379203', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20120874, end_position_on_the_genomic_accession=20128834, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421108, CurrentID=0, Status='live', Symbol='LOC100421108', Aliases='', description='inositol 1,3,4-triphosphate 5/6 kinase pseudogene', other_designations='inositol-tetrakisphosphate 1-kinase pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47156641, end_position_on_the_genomic_accession=47158199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371182, CurrentID=0, Status='live', Symbol='LOC105371182', Aliases='', description='uncharacterized LOC105371182', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32126414, end_position_on_the_genomic_accession=32131304, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133050, CurrentID=0, Status='live', Symbol='LOC100133050', Aliases='', description='glucuronidase, beta pseudogene', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100379505, end_position_on_the_genomic_accession=100388254, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371184, CurrentID=0, Status='live', Symbol='LOC105371184', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=678507, end_position_on_the_genomic_accession=680527, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371185, CurrentID=0, Status='live', Symbol='LOC105371185', Aliases='', description='uncharacterized LOC105371185', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128307, CurrentID=0, Status='live', Symbol='PIGCP2', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class C pseudogene 2', other_designations='phosphatidylinositol glycan anchor biosynthesis, class C pseudogene 2', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107808737, end_position_on_the_genomic_accession=107809592, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371188, CurrentID=0, Status='live', Symbol='LOC105371188', Aliases='', description='uncharacterized LOC105371188', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644661, CurrentID=0, Status='live', Symbol='LOC644661', Aliases='', description='importin subunit alpha-1 pseudogene', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107169426, end_position_on_the_genomic_accession=107171097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371190, CurrentID=0, Status='live', Symbol='LOC105371190', Aliases='', description='Ig heavy chain V-I region V35-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32059084, end_position_on_the_genomic_accession=32066382, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371191, CurrentID=0, Status='live', Symbol='LOC105371191', Aliases='', description='uncharacterized LOC105371191', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32329727, end_position_on_the_genomic_accession=32349365, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128312, CurrentID=0, Status='live', Symbol='SLC25A6P6', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene 6', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70663502, end_position_on_the_genomic_accession=70664279, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102902672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102902672, CurrentID=0, Status='live', Symbol='LINC00843', Aliases='', description='long intergenic non-protein coding RNA 843', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644667, CurrentID=0, Status='live', Symbol='LOC644667', Aliases='', description='golgin A7 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66069500, end_position_on_the_genomic_accession=66071339, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377611, CurrentID=0, Status='live', Symbol='LOC105377611', Aliases='', description='uncharacterized LOC105377611', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189054894, end_position_on_the_genomic_accession=189058836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644669, CurrentID=0, Status='live', Symbol='LOC644669', Aliases='', description='ankyrin repeat domain 30B pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15313556, end_position_on_the_genomic_accession=15325919, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371198, CurrentID=0, Status='live', Symbol='LOC105371198', Aliases='', description='uncharacterized LOC105371198', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35327467, end_position_on_the_genomic_accession=35333893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  151457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151457, CurrentID=0, Status='live', Symbol='LOC151457', Aliases='', description='developmentally regulated GTP binding protein 1 pseudogene', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20239664, end_position_on_the_genomic_accession=20241108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371200, CurrentID=0, Status='live', Symbol='LOC105371200', Aliases='', description='uncharacterized LOC105371200', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35207860, end_position_on_the_genomic_accession=35244710, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379207, CurrentID=0, Status='live', Symbol='LOC105379207', Aliases='', description='uncharacterized LOC105379207', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21832295, end_position_on_the_genomic_accession=21842313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374843, CurrentID=0, Status='live', Symbol='LOC105374843', Aliases='', description='uncharacterized LOC105374843', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85982594, end_position_on_the_genomic_accession=85998113, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371204, CurrentID=0, Status='live', Symbol='LOC105371204', Aliases='', description='uncharacterized LOC105371204', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35506331, end_position_on_the_genomic_accession=35524631, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100128325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128325, CurrentID=0, Status='live', Symbol='LOC100128325', Aliases='', description='uncharacterized LOC100128325', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130141517, end_position_on_the_genomic_accession=130143423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372374, CurrentID=0, Status='live', Symbol='LOC105372374', Aliases='', description='uncharacterized LOC105372374', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34555938, end_position_on_the_genomic_accession=34558329, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372362, CurrentID=0, Status='live', Symbol='LOC105372362', Aliases='', description='uncharacterized LOC105372362', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31791075, end_position_on_the_genomic_accession=31797986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371209, CurrentID=0, Status='live', Symbol='LOC105371209', Aliases='', description='uncharacterized LOC105371209', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143907388, end_position_on_the_genomic_accession=143934791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371212, CurrentID=0, Status='live', Symbol='LOC105371212', Aliases='', description='uncharacterized LOC105371212', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144501113, end_position_on_the_genomic_accession=144503718, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=54861, CurrentID=0, Status='live', Symbol='SNRK', Aliases='HSNFRK', description='SNF related kinase', other_designations='SNF-1 related kinase|SNF1-related kinase', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43286512, end_position_on_the_genomic_accession=43351143, orientation='plus', exon_count=9, OMIM=612760, no_hdr0=''),
  100128334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128334, CurrentID=0, Status='live', Symbol='LOC100128334', Aliases='', description='gap junction protein gamma 3 pseudogene', other_designations='gap junction protein, gamma 3, 30.2kDa pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99992397, end_position_on_the_genomic_accession=99993062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861402, CurrentID=0, Status='live', Symbol='CERS6-AS1', Aliases='', description='CERS6 antisense RNA 1', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168771951, end_position_on_the_genomic_accession=168786429, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371215, CurrentID=0, Status='live', Symbol='LOC105371215', Aliases='', description='uncharacterized LOC105371215', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144652206, end_position_on_the_genomic_accession=144654427, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371216, CurrentID=0, Status='live', Symbol='LOC105371216', Aliases='', description='uncharacterized LOC105371216', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144795133, end_position_on_the_genomic_accession=144813218, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100861412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861412, CurrentID=0, Status='live', Symbol='FSBP', Aliases='', description='fibrinogen silencer binding protein', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94427712, end_position_on_the_genomic_accession=94436952, orientation='minus', exon_count=2, OMIM=616306, no_hdr0=''),
  105371217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371217, CurrentID=0, Status='live', Symbol='LOC105371217', Aliases='', description='uncharacterized LOC105371217', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144814153, end_position_on_the_genomic_accession=144818136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371218, CurrentID=0, Status='live', Symbol='LOC105371218', Aliases='', description='uncharacterized LOC105371218', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145085613, end_position_on_the_genomic_accession=145091250, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128339, CurrentID=0, Status='live', Symbol='BCAS2P3', Aliases='', description='breast carcinoma amplified sequence 2 pseudogene 3', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79166227, end_position_on_the_genomic_accession=79166889, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861429, CurrentID=0, Status='live', Symbol='YRDCP3', Aliases='', description='yrdC N(6)-threonylcarbamoyltransferase domain containing pseudogene 3', other_designations='yrdC domain containing pseudogene 3', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40864018, end_position_on_the_genomic_accession=40864470, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861430, CurrentID=0, Status='live', Symbol='PDCD6IPP1', Aliases='', description='PDCD6IP pseudogene 1', other_designations='programmed cell death 6 interacting protein pseudogene', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22727104, end_position_on_the_genomic_accession=22744699, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861431, CurrentID=0, Status='live', Symbol='YBX1P7', Aliases='', description='Y-box binding protein 1 pseudogene 7', other_designations='Y box binding protein 1 pseudogene 7', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132488555, end_position_on_the_genomic_accession=132489395, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128340, CurrentID=0, Status='live', Symbol='LOC100128340', Aliases='', description='uncharacterized LOC100128340', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177883544, end_position_on_the_genomic_accession=177957757, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100861434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861434, CurrentID=0, Status='live', Symbol='USP9YP13', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 13', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23919905, end_position_on_the_genomic_accession=23920123, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861435, CurrentID=0, Status='live', Symbol='SRIP3', Aliases='', description='sorcin pseudogene 3', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6718950, end_position_on_the_genomic_accession=6719180, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100861436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861436, CurrentID=0, Status='live', Symbol='PRPS1P1', Aliases='', description='phosphoribosyl pyrophosphate synthetase 1 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164213539, end_position_on_the_genomic_accession=164214414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861437, CurrentID=0, Status='live', Symbol='NARR', Aliases='', description='nine-amino acid residue-repeats', other_designations='nine amino-acid residue-repeats', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28715993, end_position_on_the_genomic_accession=28717890, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100861439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861439, CurrentID=0, Status='live', Symbol='CTRCT37', Aliases='CCA5', description='Cataract, congenital cerulean type, 5', other_designations='', map_location='12q24.2-q24.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614422, no_hdr0=''),
  100861440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861440, CurrentID=0, Status='live', Symbol='DFNB96', Aliases='', description='deafness, autosomal recessive 96', other_designations='', map_location='1p36.31-p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614414, no_hdr0=''),
  100861443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861443, CurrentID=0, Status='live', Symbol='DNAJC19P3', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 3', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 3', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41638480, end_position_on_the_genomic_accession=41638991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644694, CurrentID=0, Status='live', Symbol='CDRT15P2', Aliases='CDRT15L1', description='CMT1A duplicated region transcript 15 pseudogene 2', other_designations='CMT1A duplicated region transcript 15-like 1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15764703, end_position_on_the_genomic_accession=15765689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100861449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861449, CurrentID=0, Status='live', Symbol='DNAJC19P6', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 6', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 6', map_location='6q', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89797446, end_position_on_the_genomic_accession=89797898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861453, CurrentID=0, Status='live', Symbol='DNAJC19P5', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 5', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 5', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177229097, end_position_on_the_genomic_accession=177229520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100861455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861455, CurrentID=0, Status='live', Symbol='DNAJC19P8', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 8', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 8', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245279753, end_position_on_the_genomic_accession=245283207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100861458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861458, CurrentID=0, Status='live', Symbol='DNAJC19P7', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 7', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 7', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52251971, end_position_on_the_genomic_accession=52252883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128345, CurrentID=0, Status='live', Symbol='LOC100128345', Aliases='', description='nuclear transcription factor Y subunit gamma pseudogene', other_designations='nuclear transcription factor Y, gamma pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35761499, end_position_on_the_genomic_accession=35762484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861467, CurrentID=0, Status='live', Symbol='ERVK-13', Aliases='c3_D', description='endogenous retrovirus group K member 13', other_designations='endogenous retrovirus group K, member 13', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861468, CurrentID=0, Status='live', Symbol='LINC00492', Aliases='', description='long intergenic non-protein coding RNA 492', other_designations='CTD-2340E1.1', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102581368, end_position_on_the_genomic_accession=102617589, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100861469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861469, CurrentID=0, Status='live', Symbol='PTP4A1P3', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 3', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91107391, end_position_on_the_genomic_accession=91107788, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861470, CurrentID=0, Status='live', Symbol='YBX1P9', Aliases='', description='Y-box binding protein 1 pseudogene 9', other_designations='Y box binding protein 1 pseudogene 9', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226318014, end_position_on_the_genomic_accession=226318703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371227, CurrentID=0, Status='live', Symbol='LOC105371227', Aliases='', description='uncharacterized LOC105371227', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148155577, end_position_on_the_genomic_accession=148158546, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100861476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861476, CurrentID=0, Status='live', Symbol='YBX1P8', Aliases='', description='Y-box binding protein 1 pseudogene 8', other_designations='Y box binding protein 1 pseudogene 8', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46684788, end_position_on_the_genomic_accession=46685498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133059, CurrentID=0, Status='live', Symbol='LOC100133059', Aliases='', description='centromere protein N pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111743314, end_position_on_the_genomic_accession=111744662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128348, CurrentID=0, Status='live', Symbol='LOC100128348', Aliases='', description='uncharacterized LOC100128348', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=33254707, end_position_on_the_genomic_accession=33255460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861485, CurrentID=0, Status='live', Symbol='DUTP3', Aliases='', description='deoxyuridine triphosphatase pseudogene 3', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147146272, end_position_on_the_genomic_accession=147146760, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646034, CurrentID=0, Status='live', Symbol='LOC646034', Aliases='', description='uncharacterized LOC646034', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31602570, end_position_on_the_genomic_accession=31607714, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371231, CurrentID=0, Status='live', Symbol='LOC105371231', Aliases='', description='uncharacterized LOC105371231', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147053869, end_position_on_the_genomic_accession=147073065, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100861504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861504, CurrentID=0, Status='live', Symbol='LINC00538', Aliases='PROX1UT, Yiya', description='long intergenic non-protein coding RNA 538', other_designations='PROX1 upstream transcript', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213924749, end_position_on_the_genomic_accession=213926654, orientation='plus', exon_count=1, OMIM=614635, no_hdr0=''),
  100861506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861506, CurrentID=0, Status='live', Symbol='LINC00903', Aliases='LSAMP-AS2', description='long intergenic non-protein coding RNA 903', other_designations='LSAMP antisense RNA 2 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861507, CurrentID=0, Status='live', Symbol='FTCD-AS1', Aliases='', description='FTCD antisense RNA 1', other_designations='FTCD antisense RNA 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861510, CurrentID=0, Status='live', Symbol='LRRC3-AS1', Aliases='', description='LRRC3 antisense RNA 1 (head to head)', other_designations='LRRC3 antisense RNA 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44450986, end_position_on_the_genomic_accession=44455284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100861512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861512, CurrentID=0, Status='live', Symbol='CDKN1A-AS1', Aliases='p21NAT', description='CDKN1A antisense RNA 1', other_designations='CDKN1A antisense RNA 1 (non-protein coding)', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861513, CurrentID=0, Status='live', Symbol='ZMYM4-AS1', Aliases='', description='ZMYM4 antisense RNA 1', other_designations='ZMYM4 antisense RNA 1 (non-protein coding)', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861514, CurrentID=0, Status='live', Symbol='ELMO1-AS1', Aliases='', description='ELMO1 antisense RNA 1', other_designations='ELMO1 antisense RNA 1 (non-protein coding)', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36997796, end_position_on_the_genomic_accession=37013630, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100861515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861515, CurrentID=0, Status='live', Symbol='TRAPPC12-AS1', Aliases='', description='TRAPPC12 antisense RNA 1', other_designations='TRAPPC12 antisense RNA 1 (non-protein coding)', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861516, CurrentID=0, Status='live', Symbol='RAI1-AS1', Aliases='', description='RAI1 antisense RNA 1', other_designations='RAI1 antisense RNA 1 (non-protein coding)', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17759154, end_position_on_the_genomic_accession=17766302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100861517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861517, CurrentID=0, Status='live', Symbol='OCIAD1-AS1', Aliases='', description='OCIAD1 antisense RNA 1', other_designations='OCIAD1 antisense RNA 1 (non-protein coding)', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861518, CurrentID=0, Status='live', Symbol='P4HA2-AS1', Aliases='', description='P4HA2 antisense RNA 1', other_designations='P4HA2 antisense RNA 1 (non-protein coding)', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132184876, end_position_on_the_genomic_accession=132192805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100861519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861519, CurrentID=0, Status='live', Symbol='GDNF-AS1', Aliases='GDNFOS', description='GDNF antisense RNA 1 (head to head)', other_designations='GDNF opposite strand', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37840438, end_position_on_the_genomic_accession=37875799, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100861520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861520, CurrentID=0, Status='live', Symbol='POU6F2-AS1', Aliases='', description='POU6F2 antisense RNA 1', other_designations='POU6F2 antisense RNA 1 (non-protein coding)', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39404598, end_position_on_the_genomic_accession=39406346, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100861522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861522, CurrentID=0, Status='live', Symbol='LINC00489', Aliases='', description='long intergenic non-protein coding RNA 489', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37619298, end_position_on_the_genomic_accession=37623119, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100861523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861523, CurrentID=0, Status='live', Symbol='RPS6KA2-AS1', Aliases='', description='RPS6KA2 antisense RNA 1', other_designations='RPS6KA2 antisense RNA 1 (non-protein coding)', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166903698, end_position_on_the_genomic_accession=166905069, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100861529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861529, CurrentID=0, Status='live', Symbol='HHPP', Aliases='', description='Hyperhidrosis palmaris et plantaris', other_designations='', map_location='14q11.2-q13', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=144110, no_hdr0=''),
  100861530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861530, CurrentID=0, Status='live', Symbol='UFL1-AS1', Aliases='', description='UFL1 antisense RNA 1', other_designations='UFL1 antisense RNA 1 (non-protein coding)', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861532, CurrentID=0, Status='live', Symbol='RNA45S5', Aliases='RN45S', description='RNA, 45S pre-ribosomal 5', other_designations='45S pre-RNA|45S pre-ribosomal RNA', map_location='22p12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861540, CurrentID=0, Status='live', Symbol='CYP3A7-CYP3A51P', Aliases='CYP3A7-3AP1, CYP3A7-CYP3AP1, CYP3A7.1L', description='CYP3A7-CYP3A51P readthrough', other_designations='CYP3A7-CYP3AP1 readthrough', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99684679, end_position_on_the_genomic_accession=99735200, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100861541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861541, CurrentID=0, Status='live', Symbol='DOCK9-AS2', Aliases='', description='DOCK9 antisense RNA 2 (head to head)', other_designations='DOCK9 antisense RNA 2 (non-protein coding)', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99086723, end_position_on_the_genomic_accession=99088625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100861542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861542, CurrentID=0, Status='live', Symbol='LINC00554', Aliases='', description='long intergenic non-protein coding RNA 554', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99994899, end_position_on_the_genomic_accession=99997909, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861543, CurrentID=0, Status='live', Symbol='LINC00555', Aliases='', description='long intergenic non-protein coding RNA 555', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861544, CurrentID=0, Status='live', Symbol='LINC00557', Aliases='', description='long intergenic non-protein coding RNA 557', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94960041, end_position_on_the_genomic_accession=94961319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100861545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861545, CurrentID=0, Status='live', Symbol='LINC00561', Aliases='', description='long intergenic non-protein coding RNA 561', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861546, CurrentID=0, Status='live', Symbol='LINC00564', Aliases='', description='long intergenic non-protein coding RNA 564', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81225865, end_position_on_the_genomic_accession=81226983, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861547, CurrentID=0, Status='live', Symbol='LINC00566', Aliases='', description='long intergenic non-protein coding RNA 566', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861548, CurrentID=0, Status='live', Symbol='PINK1-AS', Aliases='NAPINK11, PINK1AS, PINK1-AS', description='PINK1 antisense RNA', other_designations='FLJ00387 protein|PINK1 antisense RNA 1 (non-protein coding)', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20642657, end_position_on_the_genomic_accession=20652193, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100861549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861549, CurrentID=0, Status='live', Symbol='LINC00562', Aliases='', description='long intergenic non-protein coding RNA 562', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47930153, end_position_on_the_genomic_accession=47932622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861550, CurrentID=0, Status='live', Symbol='PDX1-AS1', Aliases='', description='PDX1 antisense RNA 1', other_designations='PDX1 antisense RNA 1 (non-protein coding)', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27829759, end_position_on_the_genomic_accession=27921404, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100861551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861551, CurrentID=0, Status='live', Symbol='LINC00556', Aliases='', description='long intergenic non-protein coding RNA 556', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100861552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861552, CurrentID=0, Status='live', Symbol='LINC00558', Aliases='', description='long intergenic non-protein coding RNA 558', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53815419, end_position_on_the_genomic_accession=53876119, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100861553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861553, CurrentID=0, Status='live', Symbol='LINC00560', Aliases='', description='long intergenic non-protein coding RNA 560', other_designations='', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88626048, end_position_on_the_genomic_accession=88630397, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100861554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861554, CurrentID=0, Status='live', Symbol='LINC00563', Aliases='', description='long intergenic non-protein coding RNA 563', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46296445, end_position_on_the_genomic_accession=46297844, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861555, CurrentID=0, Status='live', Symbol='LINC00565', Aliases='', description='long intergenic non-protein coding RNA 565', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113926514, end_position_on_the_genomic_accession=113928991, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861556, CurrentID=0, Status='live', Symbol='YBX1P3', Aliases='', description='Y-box binding protein 1 pseudogene 3', other_designations='Y box binding protein 1 pseudogene 3', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114930541, end_position_on_the_genomic_accession=114931449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128361, CurrentID=0, Status='live', Symbol='LOC100128361', Aliases='', description='uncharacterized LOC100128361', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92618050, end_position_on_the_genomic_accession=92620533, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100861563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861563, CurrentID=0, Status='live', Symbol='SCAANT1', Aliases='ATXN7-AS1', description='SCA7/ATXN7 antisense RNA 1', other_designations='ATXN7 antisense RNA 1 (non-protein coding)|SCA7/ATXN7 antisense RNA 1 (non-protein coding)|spinocerebellar ataxia-7 antisense noncoding transcript 1', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614481, no_hdr0=''),
  105371242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371242, CurrentID=0, Status='live', Symbol='LOC105371242', Aliases='', description='peptidyl-prolyl cis-trans isomerase A-like 4G', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146343119, end_position_on_the_genomic_accession=146344870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100861569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861569, CurrentID=0, Status='live', Symbol='YBX1P5', Aliases='', description='Y-box binding protein 1 pseudogene 5', other_designations='Y box binding protein 1 pseudogene 5', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72417560, end_position_on_the_genomic_accession=72418334, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100861573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861573, CurrentID=0, Status='live', Symbol='LINC00572', Aliases='', description='long intergenic non-protein coding RNA 572', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29918647, end_position_on_the_genomic_accession=29926651, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100861574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100861574, CurrentID=0, Status='live', Symbol='SNRPGP6', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 6', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116409260, end_position_on_the_genomic_accession=116409610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128364, CurrentID=0, Status='live', Symbol='LOC100128364', Aliases='', description='uncharacterized LOC100128364', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45077672, end_position_on_the_genomic_accession=45080011, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371245, CurrentID=0, Status='live', Symbol='LOC105371245', Aliases='', description='uncharacterized LOC105371245', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49454229, end_position_on_the_genomic_accession=49460767, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100124334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124334, CurrentID=0, Status='live', Symbol='LOC100124334', Aliases='', description='Y-box binding protein 2 pseudogene', other_designations='Y box binding protein 2 pseudogene', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93065168, end_position_on_the_genomic_accession=93066683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371248, CurrentID=0, Status='live', Symbol='LOC105371248', Aliases='', description='uncharacterized LOC105371248', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50390916, end_position_on_the_genomic_accession=50395692, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644721, CurrentID=0, Status='live', Symbol='LOC644721', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99990728, end_position_on_the_genomic_accession=99991481, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128699, CurrentID=0, Status='live', Symbol='LOC100128699', Aliases='', description='uncharacterized LOC100128699', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=127355505, end_position_on_the_genomic_accession=127356427, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101910199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101910199, CurrentID=0, Status='live', Symbol='PFFE1', Aliases='', description='Plasmodium falciparum fever episodes QTL1', other_designations='', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611384, no_hdr0=''),
  105371252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371252, CurrentID=0, Status='live', Symbol='LOC105371252', Aliases='', description='uncharacterized LOC105371252', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50805440, end_position_on_the_genomic_accession=50807629, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371253, CurrentID=0, Status='live', Symbol='LOC105371253', Aliases='', description='uncharacterized LOC105371253', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51126627, end_position_on_the_genomic_accession=51135774, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371254, CurrentID=0, Status='live', Symbol='LOC105371254', Aliases='', description='uncharacterized LOC105371254', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146229323, end_position_on_the_genomic_accession=146236689, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371255, CurrentID=0, Status='live', Symbol='LOC105371255', Aliases='', description='uncharacterized LOC105371255', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51151942, end_position_on_the_genomic_accession=51189290, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372376, CurrentID=0, Status='live', Symbol='LOC105372376', Aliases='', description='uncharacterized LOC105372376', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34710926, end_position_on_the_genomic_accession=34725895, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371257, CurrentID=0, Status='live', Symbol='LOC105371257', Aliases='', description='uncharacterized LOC105371257', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51719080, end_position_on_the_genomic_accession=51767172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371258, CurrentID=0, Status='live', Symbol='LOC105371258', Aliases='', description='uncharacterized LOC105371258', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51918216, end_position_on_the_genomic_accession=51921151, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371259, CurrentID=0, Status='live', Symbol='LOC105371259', Aliases='', description='uncharacterized LOC105371259', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51985399, end_position_on_the_genomic_accession=51994589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371264, CurrentID=0, Status='live', Symbol='LOC105371264', Aliases='', description='uncharacterized LOC105371264', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52607054, end_position_on_the_genomic_accession=52613908, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  120449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120449, CurrentID=0, Status='live', Symbol='CTBP2P6', Aliases='', description='C-terminal binding protein 2 pseudogene 6', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43521976, end_position_on_the_genomic_accession=43523264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371266, CurrentID=0, Status='live', Symbol='LOC105371266', Aliases='', description='uncharacterized LOC105371266', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52763335, end_position_on_the_genomic_accession=52764102, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371267, CurrentID=0, Status='live', Symbol='LOC105371267', Aliases='', description='uncharacterized LOC105371267', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52975953, end_position_on_the_genomic_accession=53052917, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  198437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=198437, CurrentID=0, Status='live', Symbol='LKAAEAR1', Aliases='C20orf201', description='LKAAEAR motif containing 1', other_designations='LKAAEAR motif-containing protein 1|bA299N6.3', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64083376, end_position_on_the_genomic_accession=64085144, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371271, CurrentID=0, Status='live', Symbol='LOC105371271', Aliases='', description='uncharacterized LOC105371271', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54056198, end_position_on_the_genomic_accession=54057437, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377614, CurrentID=0, Status='live', Symbol='LOC105377614', Aliases='', description='uncharacterized LOC105377614', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189364340, end_position_on_the_genomic_accession=189391692, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128393, CurrentID=0, Status='live', Symbol='RPL35AP36', Aliases='RPL35A_20_1756', description='ribosomal protein L35a pseudogene 36', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43956966, end_position_on_the_genomic_accession=43957392, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926886, CurrentID=0, Status='live', Symbol='LOC101926886', Aliases='', description='uncharacterized LOC101926886', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115147622, end_position_on_the_genomic_accession=115334252, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  644746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644746, CurrentID=0, Status='live', Symbol='LOC644746', Aliases='', description='ret finger protein-like 3 pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107887730, end_position_on_the_genomic_accession=107889767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371277, CurrentID=0, Status='live', Symbol='LOC105371277', Aliases='', description='uncharacterized LOC105371277', other_designations='CTD-3032H12.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371278, CurrentID=0, Status='live', Symbol='LOC105371278', Aliases='', description='uncharacterized LOC105371278', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55042998, end_position_on_the_genomic_accession=55049302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926887, CurrentID=0, Status='live', Symbol='LOC101926887', Aliases='', description='uncharacterized LOC101926887', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89332107, end_position_on_the_genomic_accession=89342706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371280, CurrentID=0, Status='live', Symbol='LOC105371280', Aliases='', description='uncharacterized LOC105371280', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55714266, end_position_on_the_genomic_accession=55721836, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100128401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128401, CurrentID=0, Status='live', Symbol='LOC100128401', Aliases='', description='uncharacterized LOC100128401', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=25172067, end_position_on_the_genomic_accession=25172821, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128402, CurrentID=0, Status='live', Symbol='LOC100128402', Aliases='', description='uncharacterized LOC100128402', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=8381402, end_position_on_the_genomic_accession=8381982, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928793, CurrentID=0, Status='live', Symbol='LOC101928793', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1081751, end_position_on_the_genomic_accession=1083738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644758, CurrentID=0, Status='live', Symbol='PABPC1P13', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 13', other_designations='polyadenylate-binding protein 1 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32438631, end_position_on_the_genomic_accession=32440578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371287, CurrentID=0, Status='live', Symbol='LOC105371287', Aliases='', description='uncharacterized LOC105371287', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56708829, end_position_on_the_genomic_accession=56724575, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371288, CurrentID=0, Status='live', Symbol='LOC105371288', Aliases='', description='uncharacterized LOC105371288', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145432150, end_position_on_the_genomic_accession=145443352, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374860, CurrentID=0, Status='live', Symbol='LOC105374860', Aliases='', description='uncharacterized LOC105374860', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38075387, end_position_on_the_genomic_accession=38095569, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644761, CurrentID=0, Status='live', Symbol='PIGHP1', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class H pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class H pseudogene 1', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59676757, end_position_on_the_genomic_accession=59677318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644762, CurrentID=0, Status='live', Symbol='LOC644762', Aliases='', description='mitochondrial fission factor pseudogene', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149931358, end_position_on_the_genomic_accession=149932977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421130, CurrentID=0, Status='live', Symbol='LOC100421130', Aliases='', description='LSM14A, SCD6 homolog A (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54228990, end_position_on_the_genomic_accession=54230343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371291, CurrentID=0, Status='live', Symbol='LOC105371291', Aliases='', description='uncharacterized LOC105371291', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57567663, end_position_on_the_genomic_accession=57607352, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371292, CurrentID=0, Status='live', Symbol='LOC105371292', Aliases='', description='uncharacterized LOC105371292', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57661858, end_position_on_the_genomic_accession=57663443, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128413, CurrentID=0, Status='live', Symbol='LOC100128413', Aliases='', description='X-linked inhibitor of apoptosis, E3 ubiquitin protein ligase pseudogene', other_designations='X-linked inhibitor of apoptosis pseudogene', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112853326, end_position_on_the_genomic_accession=112854314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374861, CurrentID=0, Status='live', Symbol='LOC105374861', Aliases='', description='putative uncharacterized protein C8orf44', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39148212, end_position_on_the_genomic_accession=39150198, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100337589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337589, CurrentID=0, Status='live', Symbol='FMO7P', Aliases='', description='flavin containing monooxygenase 7 pseudogene', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166474745, end_position_on_the_genomic_accession=166481565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100337590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337590, CurrentID=0, Status='live', Symbol='FMO11P', Aliases='', description='flavin containing monooxygenase 11, pseudogene', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166767570, end_position_on_the_genomic_accession=166792767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100337591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337591, CurrentID=0, Status='live', Symbol='SNORA70F', Aliases='U70F', description='small nucleolar RNA, H/ACA box 70F', other_designations='small nucleolar RNA, H/ACA box 70F (retrotransposed)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164687643, end_position_on_the_genomic_accession=164687777, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371295, CurrentID=0, Status='live', Symbol='LOC105371295', Aliases='', description='uncharacterized LOC105371295', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58461538, end_position_on_the_genomic_accession=58462856, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644768, CurrentID=0, Status='live', Symbol='SLC9B1P4', Aliases='', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1 pseudogene 4', other_designations='Na+/H+ exchanger domain containing 1 pseudogene|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1 pseudogene 4', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16428755, end_position_on_the_genomic_accession=16476875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133073, CurrentID=0, Status='live', Symbol='LOC100133073', Aliases='', description='family with sequence similarity 175, member A pseudogene', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18088919, end_position_on_the_genomic_accession=18089902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100337616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337616, CurrentID=0, Status='live', Symbol='CSN1S2BP', Aliases='CSN1S2B', description='casein alpha s2-like B, pseudogene', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70133604, end_position_on_the_genomic_accession=70146704, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100337622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337622, CurrentID=0, Status='live', Symbol='DDX11L4', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 4', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 4|DEAD/H box polypeptide 11 like 4', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100533902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533902, CurrentID=0, Status='live', Symbol='LOC100533902', Aliases='', description='ribosomal protein L3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371300, CurrentID=0, Status='live', Symbol='LOC105371300', Aliases='', description='uncharacterized LOC105371300', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59916262, end_position_on_the_genomic_accession=59942774, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371301, CurrentID=0, Status='live', Symbol='LOC105371301', Aliases='', description='uncharacterized LOC105371301', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61229804, end_position_on_the_genomic_accession=61230416, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371303, CurrentID=0, Status='live', Symbol='LOC105371303', Aliases='', description='uncharacterized LOC105371303', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145423357, end_position_on_the_genomic_accession=145425966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374863, CurrentID=0, Status='live', Symbol='LOC105374863', Aliases='', description='uncharacterized LOC105374863', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90126275, end_position_on_the_genomic_accession=90133710, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100337646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100337646, CurrentID=0, Status='live', Symbol='CT45B1P', Aliases='', description='cancer/testis antigen family 45, member B1 pseudogene', other_designations='cancer/testis antigen family 45, member A5 pseudogene', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109839949, end_position_on_the_genomic_accession=109840697, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644776, CurrentID=0, Status='live', Symbol='LOC644776', Aliases='', description='RAD52 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179400341, end_position_on_the_genomic_accession=179400893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371305, CurrentID=0, Status='live', Symbol='LOC105371305', Aliases='', description='uncharacterized LOC105371305', other_designations='CTC-420A11.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61918321, end_position_on_the_genomic_accession=61940988, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128427, CurrentID=0, Status='live', Symbol='LOC100128427', Aliases='', description='upstream binding transcription factor, RNA polymerase I pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49133807, end_position_on_the_genomic_accession=49134819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128428, CurrentID=0, Status='live', Symbol='LOC100128428', Aliases='', description='UBX domain protein 2A pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9348123, end_position_on_the_genomic_accession=9348710, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374864, CurrentID=0, Status='live', Symbol='LOC105374864', Aliases='', description='uncharacterized LOC105374864', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128429, CurrentID=0, Status='live', Symbol='LOC100128429', Aliases='', description='uncharacterized LOC100128429', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=19432039, end_position_on_the_genomic_accession=19432715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271160, CurrentID=0, Status='live', Symbol='RPL21P31', Aliases='RPL21_12_296', description='ribosomal protein L21 pseudogene 31', other_designations='', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175033785, end_position_on_the_genomic_accession=175034234, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371311, CurrentID=0, Status='live', Symbol='LOC105371311', Aliases='', description='uncharacterized LOC105371311', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=64345030, end_position_on_the_genomic_accession=64346589, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129484, CurrentID=0, Status='live', Symbol='LOC100129484', Aliases='', description='uncharacterized LOC100129484', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5426281, end_position_on_the_genomic_accession=5431023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128437, CurrentID=0, Status='live', Symbol='LOC100128437', Aliases='', description='uncharacterized LOC100128437', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=60337448, end_position_on_the_genomic_accession=60340906, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419553, CurrentID=0, Status='live', Symbol='LOC100419553', Aliases='', description='zinc finger protein 35 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150978419, end_position_on_the_genomic_accession=150979175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133077, CurrentID=0, Status='live', Symbol='LOC100133077', Aliases='', description='uncharacterized LOC100133077', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137867925, end_position_on_the_genomic_accession=137892570, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371318, CurrentID=0, Status='live', Symbol='LOC105371318', Aliases='', description='uncharacterized LOC105371318', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66363896, end_position_on_the_genomic_accession=66375804, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128713, CurrentID=0, Status='live', Symbol='LOC100128713', Aliases='', description='zinc finger protein 816A pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53952722, end_position_on_the_genomic_accession=53954813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128441, CurrentID=0, Status='live', Symbol='LOC100128441', Aliases='', description='general transcription factor IIA 2 pseudogene', other_designations='general transcription factor IIA, 2, 12kDa pseudogene', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30331964, end_position_on_the_genomic_accession=30332277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371322, CurrentID=0, Status='live', Symbol='LOC105371322', Aliases='', description='uncharacterized LOC105371322', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69176076, end_position_on_the_genomic_accession=69185253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371326, CurrentID=0, Status='live', Symbol='LOC105371326', Aliases='', description='uncharacterized LOC105371326', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69965014, end_position_on_the_genomic_accession=69973864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128448, CurrentID=0, Status='live', Symbol='UQCRHP4', Aliases='', description='ubiquinol-cytochrome c reductase hinge protein pseudogene 4', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70140217, end_position_on_the_genomic_accession=70140488, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926897, CurrentID=0, Status='live', Symbol='LOC101926897', Aliases='', description='uncharacterized LOC101926897', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58165827, end_position_on_the_genomic_accession=58209483, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371329, CurrentID=0, Status='live', Symbol='LOC105371329', Aliases='', description='uncharacterized LOC105371329', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70417366, end_position_on_the_genomic_accession=70426903, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927049, CurrentID=0, Status='live', Symbol='PLCE1-AS2', Aliases='', description='PLCE1 antisense RNA 2', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94081950, end_position_on_the_genomic_accession=94108794, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371330, CurrentID=0, Status='live', Symbol='LOC105371330', Aliases='', description='uncharacterized LOC105371330', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145408580, end_position_on_the_genomic_accession=145410723, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371331, CurrentID=0, Status='live', Symbol='LOC105371331', Aliases='', description='uncharacterized LOC105371331', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71230089, end_position_on_the_genomic_accession=71264909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371332, CurrentID=0, Status='live', Symbol='LOC105371332', Aliases='', description='uncharacterized LOC105371332', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71366589, end_position_on_the_genomic_accession=71392136, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128454, CurrentID=0, Status='live', Symbol='LOC100128454', Aliases='', description='39S ribosomal protein L32, mitochondrial pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53760506, end_position_on_the_genomic_accession=53808735, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100652743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652743, CurrentID=0, Status='live', Symbol='LOC100652743', Aliases='', description='ig kappa chain V-III region VH-like pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87338725, end_position_on_the_genomic_accession=87339075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371336, CurrentID=0, Status='live', Symbol='LOC105371336', Aliases='', description='uncharacterized LOC105371336', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71629621, end_position_on_the_genomic_accession=71631470, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100288619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288619, CurrentID=0, Status='live', Symbol='LOC100288619', Aliases='', description='uncharacterized LOC100288619', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=1017086, end_position_on_the_genomic_accession=1034281, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371337, CurrentID=0, Status='live', Symbol='LOC105371337', Aliases='', description='parathymosin pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71777148, end_position_on_the_genomic_accession=71778098, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926899, CurrentID=0, Status='live', Symbol='LOC101926899', Aliases='', description='uncharacterized LOC101926899', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90294927, end_position_on_the_genomic_accession=90364844, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371339, CurrentID=0, Status='live', Symbol='LOC105371339', Aliases='', description='uncharacterized LOC105371339', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72787278, end_position_on_the_genomic_accession=72822786, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371340, CurrentID=0, Status='live', Symbol='LOC105371340', Aliases='', description='uncharacterized LOC105371340', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72266838, end_position_on_the_genomic_accession=72287212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371341, CurrentID=0, Status='live', Symbol='LOC105371341', Aliases='', description='uncharacterized LOC105371341', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72203689, end_position_on_the_genomic_accession=72207719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128462, CurrentID=0, Status='live', Symbol='RPF2P2', Aliases='', description='ribosome production factor 2 homolog pseudogene 2', other_designations='ribosome production factor 2 homolog (S. cerevisiae) pseudogene 2', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=115628809, end_position_on_the_genomic_accession=115629704, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103352539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103352539, CurrentID=0, Status='live', Symbol='LINC01410', Aliases='', description='long intergenic non-protein coding RNA 1410', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62801465, end_position_on_the_genomic_accession=62813486, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  103352540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103352540, CurrentID=0, Status='live', Symbol='IGKV1OR9-1', Aliases='IGKV1/OR9-1', description='immunoglobulin kappa variable 1/OR9-1 (pseudogene)', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65250499, end_position_on_the_genomic_accession=65251012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103352541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103352541, CurrentID=0, Status='live', Symbol='LOC103352541', Aliases='', description='uncharacterized LOC103352541', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139978350, end_position_on_the_genomic_accession=139991261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371344, CurrentID=0, Status='live', Symbol='LOC105371344', Aliases='', description='uncharacterized LOC105371344', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75255953, end_position_on_the_genomic_accession=75263167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371345, CurrentID=0, Status='live', Symbol='LOC105371345', Aliases='', description='uncharacterized LOC105371345', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75280281, end_position_on_the_genomic_accession=75293602, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421141, CurrentID=0, Status='live', Symbol='LOC100421141', Aliases='', description='developmental pluripotency associated 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77311425, end_position_on_the_genomic_accession=77312117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189109, CurrentID=0, Status='live', Symbol='TRR-TCT3-2', Aliases='TRNAR12', description='transfer RNA-Arg (TCT) 3-2', other_designations='transfer RNA arginine 12 (anticodon UCU)', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128340076, end_position_on_the_genomic_accession=128340166, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128467, CurrentID=0, Status='live', Symbol='LOC100128467', Aliases='', description='nuclear distribution protein nudE-like 1-like', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68474611, end_position_on_the_genomic_accession=68476179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371348, CurrentID=0, Status='live', Symbol='LOC105371348', Aliases='', description='uncharacterized LOC105371348', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75905099, end_position_on_the_genomic_accession=75915651, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926901, CurrentID=0, Status='live', Symbol='LOC101926901', Aliases='', description='uncharacterized LOC101926901', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222510532, end_position_on_the_genomic_accession=222519974, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371349, CurrentID=0, Status='live', Symbol='LOC105371349', Aliases='', description='uncharacterized LOC105371349', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76064640, end_position_on_the_genomic_accession=76121808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100652758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652758, CurrentID=0, Status='live', Symbol='LOC100652758', Aliases='', description='uncharacterized LOC100652758', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151769793, end_position_on_the_genomic_accession=151771508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371351, CurrentID=0, Status='live', Symbol='LOC105371351', Aliases='', description='uncharacterized LOC105371351', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77934505, end_position_on_the_genomic_accession=77949736, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371355, CurrentID=0, Status='live', Symbol='LOC105371355', Aliases='', description='uncharacterized LOC105371355', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78495771, end_position_on_the_genomic_accession=78506613, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929894, CurrentID=0, Status='live', Symbol='LOC101929894', Aliases='', description='uncharacterized LOC101929894', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128477, CurrentID=0, Status='live', Symbol='LOC100128477', Aliases='', description='uncharacterized LOC100128477', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=111736707, end_position_on_the_genomic_accession=111737749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101926903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926903, CurrentID=0, Status='live', Symbol='LSAMP-AS1', Aliases='', description='uncharacterized LOC101926903', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116360024, end_position_on_the_genomic_accession=116370085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100652768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652768, CurrentID=0, Status='live', Symbol='LOC100652768', Aliases='', description='uncharacterized LOC100652768', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117195613, end_position_on_the_genomic_accession=117201914, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100128721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128721, CurrentID=0, Status='live', Symbol='LOC100128721', Aliases='', description='stromal cell-derived factor 2 pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75114261, end_position_on_the_genomic_accession=75114869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100652770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652770, CurrentID=0, Status='live', Symbol='DSG2-AS1', Aliases='', description='DSG2 antisense RNA 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31542146, end_position_on_the_genomic_accession=31556911, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371363, CurrentID=0, Status='live', Symbol='LOC105371363', Aliases='', description='uncharacterized LOC105371363', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82044333, end_position_on_the_genomic_accession=82064420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371364, CurrentID=0, Status='live', Symbol='LOC105371364', Aliases='', description='uncharacterized LOC105371364', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149140045, end_position_on_the_genomic_accession=149142712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103352670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103352670, CurrentID=0, Status='live', Symbol='LINC01419', Aliases='LVCAT7, TCONS_00014497', description='long intergenic non-protein coding RNA 1419', other_designations='liver cancer-associated transcript 7', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=83403758, end_position_on_the_genomic_accession=83408900, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371365, CurrentID=0, Status='live', Symbol='LOC105371365', Aliases='', description='uncharacterized LOC105371365', other_designations='CTD-2588J6.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82170309, end_position_on_the_genomic_accession=82179276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371366, CurrentID=0, Status='live', Symbol='LOC105371366', Aliases='', description='uncharacterized LOC105371366', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83720401, end_position_on_the_genomic_accession=83729740, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371367, CurrentID=0, Status='live', Symbol='LOC105371367', Aliases='', description='uncharacterized LOC105371367', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83701656, end_position_on_the_genomic_accession=83717920, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  404542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404542, CurrentID=0, Status='live', Symbol='DFNB50', Aliases='', description='deafness, autosomal recessive 50', other_designations='', map_location='12q23-qter', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371369, CurrentID=0, Status='live', Symbol='LOC105371369', Aliases='', description='uncharacterized LOC105371369', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82707318, end_position_on_the_genomic_accession=82716463, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=11328, CurrentID=0, Status='live', Symbol='FKBP9', Aliases='FKBP60, FKBP63, PPIase', description='FK506 binding protein 9', other_designations='63 kDa FK506-binding protein|63 kDa FKBP|FK506 binding protein 9, 63 kDa|FK506-binding protein 9|FKBP-63|FKBP-9|PPIase FKBP9|rotamase', map_location='7p11.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32957404, end_position_on_the_genomic_accession=33006931, orientation='plus', exon_count=13, OMIM=616257, no_hdr0=''),
  105371372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371372, CurrentID=0, Status='live', Symbol='LOC105371372', Aliases='', description='uncharacterized LOC105371372', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83929795, end_position_on_the_genomic_accession=83931113, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128493, CurrentID=0, Status='live', Symbol='UBE2V2P1', Aliases='', description='ubiquitin conjugating enzyme E2 variant 2 pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2 variant 2 pseudogene 1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19051455, end_position_on_the_genomic_accession=19052172, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128494, CurrentID=0, Status='live', Symbol='LOC100128494', Aliases='', description='uncharacterized LOC100128494', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72014291, end_position_on_the_genomic_accession=72020910, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  732229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732229, CurrentID=0, Status='live', Symbol='LOC732229', Aliases='', description='AN1-type zinc finger protein 5 pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42485054, end_position_on_the_genomic_accession=42485765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371376, CurrentID=0, Status='live', Symbol='LOC105371376', Aliases='', description='uncharacterized LOC105371376', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84642649, end_position_on_the_genomic_accession=84650843, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  404550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404550, CurrentID=0, Status='live', Symbol='C16orf74', Aliases='', description='chromosome 16 open reading frame 74', other_designations='hypothetical protein 1|hypothetical protein 2|hypothetical protein 3', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85707510, end_position_on_the_genomic_accession=85751083, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371377, CurrentID=0, Status='live', Symbol='LOC105371377', Aliases='', description='uncharacterized LOC105371377', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84729015, end_position_on_the_genomic_accession=84730815, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  404551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404551, CurrentID=0, Status='live', Symbol='RWDD1P1', Aliases='', description='RWD domain containing 1 pseudogene 1', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78382573, end_position_on_the_genomic_accession=78383200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371378, CurrentID=0, Status='live', Symbol='LOC105371378', Aliases='', description='uncharacterized LOC105371378', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84780930, end_position_on_the_genomic_accession=84795698, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  404552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404552, CurrentID=0, Status='live', Symbol='SCGB1D4', Aliases='', description='secretoglobin, family 1D, member 4', other_designations='IFN-gamma inducible SCGB (IIS)|IFN-gamma-inducible secretoglobin|IIS', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62296282, end_position_on_the_genomic_accession=62299064, orientation='minus', exon_count=3, OMIM=615062, no_hdr0=''),
  644852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644852, CurrentID=0, Status='live', Symbol='LOC644852', Aliases='', description='uncharacterized LOC644852', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=247681165, end_position_on_the_genomic_accession=247683938, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128503, CurrentID=0, Status='live', Symbol='DPPA2P2', Aliases='', description='developmental pluripotency associated 2 pseudogene 2', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26519117, end_position_on_the_genomic_accession=26520264, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652792, CurrentID=0, Status='live', Symbol='LOC100652792', Aliases='', description='ferritin, light polypeptide pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92762757, end_position_on_the_genomic_accession=92763793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271163, CurrentID=0, Status='live', Symbol='RPL21P43', Aliases='RPL21_15_434', description='ribosomal protein L21 pseudogene 43', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168858622, end_position_on_the_genomic_accession=168859172, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379244, CurrentID=0, Status='live', Symbol='LOC105379244', Aliases='', description='uncharacterized LOC105379244', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11942304, end_position_on_the_genomic_accession=11944137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100652794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652794, CurrentID=0, Status='live', Symbol='LOC100652794', Aliases='', description='UMP-CMP kinase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35765992, end_position_on_the_genomic_accession=35774263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133091, CurrentID=0, Status='live', Symbol='LOC100133091', Aliases='', description='uncharacterized LOC100133091', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76549341, end_position_on_the_genomic_accession=76627982, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371388, CurrentID=0, Status='live', Symbol='LOC105371388', Aliases='', description='uncharacterized LOC105371388', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85926020, end_position_on_the_genomic_accession=85949182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371389, CurrentID=0, Status='live', Symbol='LOC105371389', Aliases='', description='uncharacterized LOC105371389', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86047819, end_position_on_the_genomic_accession=86049340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371390, CurrentID=0, Status='live', Symbol='LOC105371390', Aliases='', description='uncharacterized LOC105371390', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86645221, end_position_on_the_genomic_accession=86647356, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371391, CurrentID=0, Status='live', Symbol='LOC105371391', Aliases='', description='uncharacterized LOC105371391', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86682344, end_position_on_the_genomic_accession=86696445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  360158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360158, CurrentID=0, Status='live', Symbol='CYCSP5', Aliases='HCP5', description='cytochrome c, somatic pseudogene 5', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244598390, end_position_on_the_genomic_accession=244598687, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371392, CurrentID=0, Status='live', Symbol='LOC105371392', Aliases='', description='uncharacterized LOC105371392', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86761598, end_position_on_the_genomic_accession=86763965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371393, CurrentID=0, Status='live', Symbol='LOC105371393', Aliases='', description='uncharacterized LOC105371393', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86846427, end_position_on_the_genomic_accession=87057690, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371394, CurrentID=0, Status='live', Symbol='LOC105371394', Aliases='', description='uncharacterized LOC105371394', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87117560, end_position_on_the_genomic_accession=87120017, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371396, CurrentID=0, Status='live', Symbol='LOC105371396', Aliases='', description='uncharacterized LOC105371396', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422157, CurrentID=0, Status='live', Symbol='LOC100422157', Aliases='', description='olfactory receptor, family 52, subfamily J, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128517, CurrentID=0, Status='live', Symbol='LOC100128517', Aliases='', description='uncharacterized LOC100128517', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=14323069, end_position_on_the_genomic_accession=14323715, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644871, CurrentID=0, Status='live', Symbol='TRAF6P1', Aliases='', description='TNF receptor associated factor 6 pseudogene 1', other_designations='TNF receptor-associated factor 6 pseudogene 1|TNF receptor-associated factor 6, E3 ubiquitin protein ligase pseudogene', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59136625, end_position_on_the_genomic_accession=59138182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  199221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199221, CurrentID=0, Status='live', Symbol='DZIP1L', Aliases='DZIP2', description='DAZ interacting zinc finger protein 1-like', other_designations='DAZ interacting protein 1-like|DAZ-interacting protein 1-like protein', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138061985, end_position_on_the_genomic_accession=138115818, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  199223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199223, CurrentID=0, Status='live', Symbol='TTC21A', Aliases='STI2', description='tetratricopeptide repeat domain 21A', other_designations='TPR domain containing STI2|TPR repeat protein 21A|stress-inducible protein 2', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39107661, end_position_on_the_genomic_accession=39138994, orientation='plus', exon_count=31, OMIM=611430, no_hdr0=''),
  100130627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130627, CurrentID=0, Status='live', Symbol='LOC100130627', Aliases='', description='uncharacterized LOC100130627', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=39137775, end_position_on_the_genomic_accession=39138479, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  360160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360160, CurrentID=0, Status='live', Symbol='CYCSP8', Aliases='HCP8', description='cytochrome c, somatic pseudogene 8', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130776646, end_position_on_the_genomic_accession=130776957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128523, CurrentID=0, Status='live', Symbol='LOC100128523', Aliases='', description='mitochondrial carrier 2 pseudogene', other_designations='mitochondrial carrier homolog 2 (C. elegans) pseudogene', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53010281, end_position_on_the_genomic_accession=53011296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644877, CurrentID=0, Status='live', Symbol='CACYBPP2', Aliases='', description='calcyclin binding protein pseudogene 2', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183607438, end_position_on_the_genomic_accession=183608119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927383, CurrentID=0, Status='live', Symbol='LOC101927383', Aliases='', description='uncharacterized LOC101927383', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104703591, end_position_on_the_genomic_accession=104705913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643855, CurrentID=0, Status='live', Symbol='LOC643855', Aliases='', description='G1 to S phase transition pseudogene', other_designations='G1 to S phase transition 1 pseudogene', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106310104, end_position_on_the_genomic_accession=106312221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371409, CurrentID=0, Status='live', Symbol='LOC105371409', Aliases='', description='uncharacterized LOC105371409', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88925447, end_position_on_the_genomic_accession=88930657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371411, CurrentID=0, Status='live', Symbol='LOC105371411', Aliases='', description='uncharacterized LOC105371411', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89060338, end_position_on_the_genomic_accession=89062351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105601840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105601840, CurrentID=0, Status='live', Symbol='WDR7-OT1', Aliases='WDR7-UA1, WDR7-UT1', description='WDR7 overlapping transcript 1', other_designations="WDR7 3'UTR-associated RNA 1|WDR7 overlapping transcript 1 (non-protein coding)", map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131482, CurrentID=0, Status='live', Symbol='LOC100131482', Aliases='', description='thioredoxin like 4A pseudogene', other_designations='thioredoxin-like 4A pseudogene', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179814094, end_position_on_the_genomic_accession=179814838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371415, CurrentID=0, Status='live', Symbol='LOC105371415', Aliases='', description='uncharacterized LOC105371415', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149665715, end_position_on_the_genomic_accession=149669149, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379250, CurrentID=0, Status='live', Symbol='LOC105379250', Aliases='', description='uncharacterized LOC105379250', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65720105, end_position_on_the_genomic_accession=65723902, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128536, CurrentID=0, Status='live', Symbol='SETP8', Aliases='', description='SET pseudogene 8', other_designations='SET nuclear oncogene pseudogene', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116963049, end_position_on_the_genomic_accession=116963676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360163, CurrentID=0, Status='live', Symbol='CYCTP', Aliases='HCP9', description='cytochrome c, testis, pseudogene', other_designations='cytochrome c, somatic pseudogene', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178092500, end_position_on_the_genomic_accession=178104470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371418, CurrentID=0, Status='live', Symbol='LOC105371418', Aliases='', description='uncharacterized LOC105371418', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89871065, end_position_on_the_genomic_accession=89873296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128540, CurrentID=0, Status='live', Symbol='LOC100128540', Aliases='', description='TAR DNA-binding protein 43-like', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63136397, end_position_on_the_genomic_accession=63138084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371421, CurrentID=0, Status='live', Symbol='LOC105371421', Aliases='', description='uncharacterized LOC105371421', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89986149, end_position_on_the_genomic_accession=89987852, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339059, CurrentID=0, Status='live', Symbol='LOC339059', Aliases='', description='uncharacterized LOC339059', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88742767, end_position_on_the_genomic_accession=88745748, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371422, CurrentID=0, Status='live', Symbol='LOC105371422', Aliases='', description='uncharacterized LOC105371422', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371423, CurrentID=0, Status='live', Symbol='LOC105371423', Aliases='', description='uncharacterized LOC105371423', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371424, CurrentID=0, Status='live', Symbol='LOC105371424', Aliases='', description='uncharacterized LOC105371424', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149853544, end_position_on_the_genomic_accession=149860683, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100862662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862662, CurrentID=0, Status='live', Symbol='ALDH1L1-AS2', Aliases='', description='ALDH1L1 antisense RNA 2', other_designations='ALDH1L1 antisense RNA 2 (non-protein coding)', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126180065, end_position_on_the_genomic_accession=126210169, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100862663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862663, CurrentID=0, Status='live', Symbol='RNY4P21', Aliases='', description='RNA, Ro-associated Y4 pseudogene 21', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74091413, end_position_on_the_genomic_accession=74091485, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862664, CurrentID=0, Status='live', Symbol='RNY4P24', Aliases='', description='RNA, Ro-associated Y4 pseudogene 24', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52400004, end_position_on_the_genomic_accession=52400096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862665, CurrentID=0, Status='live', Symbol='RNY4P25', Aliases='', description='RNA, Ro-associated Y4 pseudogene 25', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151439002, end_position_on_the_genomic_accession=151439093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862666, CurrentID=0, Status='live', Symbol='RNY1P2', Aliases='', description='RNA, Ro-associated Y1 pseudogene 2', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102016953, end_position_on_the_genomic_accession=102017069, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862667, CurrentID=0, Status='live', Symbol='RNY1P5', Aliases='', description='RNA, Ro-associated Y1 pseudogene 5', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74289100, end_position_on_the_genomic_accession=74289214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862668, CurrentID=0, Status='live', Symbol='RNY3P2', Aliases='', description='RNA, Ro-associated Y3 pseudogene 2', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49095392, end_position_on_the_genomic_accession=49095489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862669, CurrentID=0, Status='live', Symbol='RNY3P3', Aliases='', description='RNA, Ro-associated Y3 pseudogene 3', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78355227, end_position_on_the_genomic_accession=78355328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862670, CurrentID=0, Status='live', Symbol='RNY3P7', Aliases='', description='RNA, Ro-associated Y3 pseudogene 7', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77623372, end_position_on_the_genomic_accession=77623472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862671, CurrentID=0, Status='live', Symbol='TMEM265', Aliases='IFITMD8', description='transmembrane protein 265', other_designations='', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30740191, end_position_on_the_genomic_accession=30745196, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100862672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862672, CurrentID=0, Status='live', Symbol='RNY3P9', Aliases='', description='RNA, Ro-associated Y3 pseudogene 9', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40226827, end_position_on_the_genomic_accession=40226932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862673, CurrentID=0, Status='live', Symbol='RNY4P30', Aliases='', description='RNA, Ro-associated Y4 pseudogene 30', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49889634, end_position_on_the_genomic_accession=49889728, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862674, CurrentID=0, Status='live', Symbol='RNY4P31', Aliases='', description='RNA, Ro-associated Y4 pseudogene 31', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60816701, end_position_on_the_genomic_accession=60816788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862675, CurrentID=0, Status='live', Symbol='RNU4-9P', Aliases='', description='RNA, U4 small nuclear 9, pseudogene', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21038529, end_position_on_the_genomic_accession=21038667, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862676, CurrentID=0, Status='live', Symbol='RNU4-10P', Aliases='', description='RNA, U4 small nuclear 10, pseudogene', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73191528, end_position_on_the_genomic_accession=73191667, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862677, CurrentID=0, Status='live', Symbol='RBMX2P4', Aliases='', description='RNA binding motif protein, X-linked 2 pseudogene 4', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12922344, end_position_on_the_genomic_accession=12922920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862678, CurrentID=0, Status='live', Symbol='IL20RB-AS1', Aliases='', description='IL20RB antisense RNA 1', other_designations='IL20RB antisense RNA 1 (non-protein coding)', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862679, CurrentID=0, Status='live', Symbol='NAALADL2-AS3', Aliases='', description='NAALADL2 antisense RNA 3', other_designations='NAALADL2 antisense RNA 3 (non-protein coding)', map_location='3q26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175079312, end_position_on_the_genomic_accession=175115242, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100862680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862680, CurrentID=0, Status='live', Symbol='LINC00507', Aliases='', description='long intergenic non-protein coding RNA 507', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127915410, end_position_on_the_genomic_accession=127951552, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100862681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862681, CurrentID=0, Status='live', Symbol='RP63', Aliases='', description='retinitis pigmentosa 63 (autosomal dominant)', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614494, no_hdr0=''),
  100862682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862682, CurrentID=0, Status='live', Symbol='LOC100862682', Aliases='', description='nucleus accumbens associated 1, BEN and BTB (POZ) domain containing pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33288232, end_position_on_the_genomic_accession=33290003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862683, CurrentID=0, Status='live', Symbol='ERVK-25', Aliases='c11_A', description='endogenous retrovirus group K member 25', other_designations='Envelope polyprotein|HERV-K_11q22.1 provirus Rec protein|HERV-K_11q22.1 provirus ancestral Env polyprotein|HERV-K_11q22.1 provirus ancestral Pol protein|HERV-K_11q22.1 provirus ancestral Pro protein|endogenous retrovirus group K, member 25', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862684, CurrentID=0, Status='live', Symbol='ERVK-24', Aliases='HERV-K101, c22_A', description='endogenous retrovirus group K, member 24', other_designations='HERV-K_22q11.21 provirus Np9 protein|HERV-K_22q11.21 provirus ancestral Env polyprotein|HERV-K_22q11.21 provirus ancestral Gag polyprotein|HERV-K_22q11.21 provirus ancestral Pro protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862685, CurrentID=0, Status='live', Symbol='ERVK-19', Aliases='HERV-K(C19), P1.8, envK3, polymerase', description='endogenous retrovirus group K member 19', other_designations='Envelope polyprotein|HERV-K_19q11 provirus ancestral Env polyprotein|HERV-K_19q11 provirus ancestral Gag polyprotein|HERV-K_19q11 provirus ancestral Pol protein|HERV-K_19q12 provirus ancestral Pro protein|endogenous retrovirus group K, member 19', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862686, CurrentID=0, Status='live', Symbol='ERVK-17', Aliases='', description='endogenous retrovirus group K member 17', other_designations='endogenous retrovirus group K, member 17', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862687, CurrentID=0, Status='live', Symbol='ERVK-14', Aliases='c7_B', description='endogenous retrovirus group K member 14', other_designations='endogenous retrovirus group K, member 14', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862688, CurrentID=0, Status='live', Symbol='ERVK-11', Aliases='N8.4, c3_E, polymerase', description='endogenous retrovirus group K member 11', other_designations='HERV-K_3q27.3 provirus ancestral Pol protein|endogenous retrovirus group K, member 11|env protein', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862689, CurrentID=0, Status='live', Symbol='ERVK-3', Aliases='c3_B', description='endogenous retrovirus group K member 3', other_designations='endogenous retrovirus group K, member 3', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862690, CurrentID=0, Status='live', Symbol='ERVK-1', Aliases='c1_A', description='endogenous retrovirus group K member 1', other_designations='endogenous retrovirus group K, member 1', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862691, CurrentID=0, Status='live', Symbol='ERVE-2', Aliases='', description='endogenous retrovirus group E member 2', other_designations='endogenous retrovirus group E, member 2', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862692, CurrentID=0, Status='live', Symbol='DIP2A-IT1', Aliases='', description='DIP2A intronic transcript 1', other_designations='DIP2A intronic transcript 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46462471, end_position_on_the_genomic_accession=46469306, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100862693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862693, CurrentID=0, Status='live', Symbol='ERVPABLB-1', Aliases='envR(b)', description='endogenous retrovirus group PABLB member 1', other_designations='Envelope polyprotein|HERV-R(b)_3p24.3 provirus ancestral Env polyprotein|endogenous retrovirus group PABLB, member 1', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862694, CurrentID=0, Status='live', Symbol='ERVS71-1', Aliases='envT', description='endogenous retrovirus group S71 member 1', other_designations='Envelope polyprotein|HERV-T Env protein|HERV-T_19 provirus ancestral Env polyprotein|endogenous retrovirus group S71, member 1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862695, CurrentID=0, Status='live', Symbol='ERVW-5', Aliases='CL2', description='endogenous retrovirus group W member 5', other_designations='endogenous retrovirus group W, member 5', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862696, CurrentID=0, Status='live', Symbol='ERVW-3', Aliases='CL4', description='endogenous retrovirus group W member 3', other_designations='endogenous retrovirus group W, member 3', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862697, CurrentID=0, Status='live', Symbol='ERVK3-2', Aliases='HML6-c14, PP3227', description='endogenous retrovirus group K3 member 2', other_designations='PP3227 protein|endogenous retrovirus group K3, member 2', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862698, CurrentID=0, Status='live', Symbol='ERVH-7', Aliases='', description='endogenous retrovirus group H member 7', other_designations='TCONS_00022381|endogenous retrovirus group H, member 7', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862699, CurrentID=0, Status='live', Symbol='ERVH-5', Aliases='', description='endogenous retrovirus group H member 5', other_designations='clone c14.6|endogenous retrovirus group H, member 5', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862700, CurrentID=0, Status='live', Symbol='ERVE-3', Aliases='', description='endogenous retrovirus group E member 3', other_designations='endogenous retrovirus group E, member 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372383, CurrentID=0, Status='live', Symbol='LOC105372383', Aliases='', description='uncharacterized LOC105372383', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35956822, end_position_on_the_genomic_accession=35959285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100862704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862704, CurrentID=0, Status='live', Symbol='LINC00441', Aliases='ncRNA-RB1', description='long intergenic non-protein coding RNA 441', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48296513, end_position_on_the_genomic_accession=48303661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100862705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862705, CurrentID=0, Status='live', Symbol='LOC100862705', Aliases='', description='KN motif and ankyrin repeat domains 1 pseudogene', other_designations='', map_location='20p', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5927129, end_position_on_the_genomic_accession=5929328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862706, CurrentID=0, Status='live', Symbol='CATMANS', Aliases='', description='Catel-Manzke syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302380, no_hdr0=''),
  100289117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289117, CurrentID=0, Status='live', Symbol='LOC100289117', Aliases='', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224294930, end_position_on_the_genomic_accession=224295088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371433, CurrentID=0, Status='live', Symbol='LOC105371433', Aliases='', description='uncharacterized LOC105371433', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150160728, end_position_on_the_genomic_accession=150181331, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105371434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371434, CurrentID=0, Status='live', Symbol='LOC105371434', Aliases='', description='uncharacterized LOC105371434', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150293302, end_position_on_the_genomic_accession=150293722, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271165, CurrentID=0, Status='live', Symbol='RPL21P48', Aliases='RPL21_17_503', description='ribosomal protein L21 pseudogene 48', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98409250, end_position_on_the_genomic_accession=98409796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862725, CurrentID=0, Status='live', Symbol='TIMM9P3', Aliases='', description='TIMM9 pseudogene 3', other_designations='translocase of inner mitochondrial membrane 9 homolog pseudogene 3', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45552604, end_position_on_the_genomic_accession=45553363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862726, CurrentID=0, Status='live', Symbol='TIMM9P1', Aliases='', description='TIMM9 pseudogene 1', other_designations='translocase of inner mitochondrial membrane 9 homolog pseudogene 1', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74344270, end_position_on_the_genomic_accession=74344929, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862727, CurrentID=0, Status='live', Symbol='TIMM9P2', Aliases='', description='TIMM9 pseudogene 2', other_designations='translocase of inner mitochondrial membrane 9 homolog pseudogene 2', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39216654, end_position_on_the_genomic_accession=39216788, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862728, CurrentID=0, Status='live', Symbol='UPK1A-AS1', Aliases='', description='UPK1A antisense RNA 1', other_designations='UPK1A antisense RNA 1 (non-protein coding)', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35667948, end_position_on_the_genomic_accession=35673291, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  136263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=136263, CurrentID=0, Status='live', Symbol='SSMEM1', Aliases='C7orf45', description='serine-rich single-pass membrane protein 1', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130205403, end_position_on_the_genomic_accession=130216844, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  388022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388022, CurrentID=0, Status='live', Symbol='LOC388022', Aliases='', description='uncharacterized LOC388022', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104719713, end_position_on_the_genomic_accession=104722537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644909, CurrentID=0, Status='live', Symbol='LOC644909', Aliases='', description='ligand of numb-protein X 1, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17422677, end_position_on_the_genomic_accession=17423729, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371438, CurrentID=0, Status='live', Symbol='LOC105371438', Aliases='', description='uncharacterized LOC105371438', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150979827, end_position_on_the_genomic_accession=150981661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371439, CurrentID=0, Status='live', Symbol='LOC105371439', Aliases='', description='uncharacterized LOC105371439', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151273221, end_position_on_the_genomic_accession=151278047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371440, CurrentID=0, Status='live', Symbol='LOC105371440', Aliases='', description='uncharacterized LOC105371440', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151735538, end_position_on_the_genomic_accession=151759648, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371442, CurrentID=0, Status='live', Symbol='LOC105371442', Aliases='', description='uncharacterized LOC105371442', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152018878, end_position_on_the_genomic_accession=152027518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419558, CurrentID=0, Status='live', Symbol='KARSP1', Aliases='', description='lysyl-tRNA synthetase pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41925572, end_position_on_the_genomic_accession=41927525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133102, CurrentID=0, Status='live', Symbol='LOC100133102', Aliases='', description='ras-related C3 botulinum toxin substrate 2 (rho family, small GTP binding protein Rac2) pseudogene', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85868944, end_position_on_the_genomic_accession=85869499, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644915, CurrentID=0, Status='live', Symbol='METTL15P2', Aliases='', description='methyltransferase like 15 pseudogene 2', other_designations='methyltransferase 5 domain containing 1 pseudogene', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=59053321, end_position_on_the_genomic_accession=59054184, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371445, CurrentID=0, Status='live', Symbol='LOC105371445', Aliases='', description='uncharacterized LOC105371445', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152869330, end_position_on_the_genomic_accession=152872495, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  723778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723778, CurrentID=0, Status='live', Symbol='MIR650', Aliases='MIRN650, hsa-mir-650', description='microRNA 650', other_designations='', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22822776, end_position_on_the_genomic_accession=22822871, orientation='plus', exon_count=1, OMIM=615379, no_hdr0=''),
  723779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723779, CurrentID=0, Status='live', Symbol='MIR651', Aliases='MIRN651, hsa-mir-651, mir-651', description='microRNA 651', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8126965, end_position_on_the_genomic_accession=8127061, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371447, CurrentID=0, Status='live', Symbol='LOC105371447', Aliases='', description='uncharacterized LOC105371447', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153164351, end_position_on_the_genomic_accession=153203313, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  723788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723788, CurrentID=0, Status='live', Symbol='MIG7', Aliases='', description='mig-7', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  723790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723790, CurrentID=0, Status='live', Symbol='HIST2H2AA4', Aliases='H2A/R', description='histone cluster 2, H2aa4', other_designations='histone 2, H2aa4|histone H2A/r', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149851061, end_position_on_the_genomic_accession=149851594, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371448, CurrentID=0, Status='live', Symbol='LOC105371448', Aliases='', description='immunoglobulin superfamily DCC subclass member 3-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153800481, end_position_on_the_genomic_accession=153804151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100652857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652857, CurrentID=0, Status='live', Symbol='TMCO5B', Aliases='', description='transmembrane and coiled-coil domains 5B, pseudogene', other_designations='transmembrane and coiled-coil domain-containing protein 5B', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33236476, end_position_on_the_genomic_accession=33247555, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  723805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723805, CurrentID=0, Status='live', Symbol='LOC723805', Aliases='', description='interleukin-like', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39439264, end_position_on_the_genomic_accession=39442019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  723809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723809, CurrentID=0, Status='live', Symbol='LHFPL3-AS2', Aliases='', description='LHFPL3 antisense RNA 2', other_designations='LHFPL3 antisense RNA 2 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104894628, end_position_on_the_genomic_accession=104926645, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128571, CurrentID=0, Status='live', Symbol='LOC100128571', Aliases='', description='uncharacterized LOC100128571', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=47194515, end_position_on_the_genomic_accession=47194883, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371452, CurrentID=0, Status='live', Symbol='LOC105371452', Aliases='', description='uncharacterized LOC105371452', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155609609, end_position_on_the_genomic_accession=155612837, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128573, CurrentID=0, Status='live', Symbol='LOC100128573', Aliases='', description='uncharacterized LOC100128573', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7472837, end_position_on_the_genomic_accession=7473361, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371454, CurrentID=0, Status='live', Symbol='LOC105371454', Aliases='', description='uncharacterized LOC105371454', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156706329, end_position_on_the_genomic_accession=156713174, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479117, CurrentID=0, Status='live', Symbol='RN7SKP51', Aliases='', description='RNA, 7SK small nuclear pseudogene 51', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100862847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862847, CurrentID=0, Status='live', Symbol='USP17L', Aliases='', description='ubiquitin carboxyl-terminal hydrolase 17-like protein', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371456, CurrentID=0, Status='live', Symbol='LOC105371456', Aliases='', description='uncharacterized LOC105371456', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157225413, end_position_on_the_genomic_accession=157279272, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100862850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862850, CurrentID=0, Status='live', Symbol='EEF1A1P39', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 39', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30301823, end_position_on_the_genomic_accession=30302072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100862852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862852, CurrentID=0, Status='live', Symbol='SLAMF6P1', Aliases='', description='SLAM family member 6 pseudogene 1', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162445549, end_position_on_the_genomic_accession=162445886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862853, CurrentID=0, Status='live', Symbol='MTND4P13', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 13', other_designations='MT-ND4 pseudogene 13', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153668879, end_position_on_the_genomic_accession=153669826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862854, CurrentID=0, Status='live', Symbol='MTND5P16', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 16', other_designations='MT-ND5 pseudogene 16', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106898889, end_position_on_the_genomic_accession=106900585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100862855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100862855, CurrentID=0, Status='live', Symbol='COX6CP5', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 5', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7977496, end_position_on_the_genomic_accession=7977705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371459, CurrentID=0, Status='live', Symbol='LOC105371459', Aliases='', description='uncharacterized LOC105371459', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157979311, end_position_on_the_genomic_accession=157986183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271166, CurrentID=0, Status='live', Symbol='RPL21P49', Aliases='RPL21_18_504', description='ribosomal protein L21 pseudogene 49', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102662581, end_position_on_the_genomic_accession=102663116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371462, CurrentID=0, Status='live', Symbol='LOC105371462', Aliases='', description='uncharacterized LOC105371462', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159466104, end_position_on_the_genomic_accession=159490491, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100133106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133106, CurrentID=0, Status='live', Symbol='LOC100133106', Aliases='', description='VCEW9374', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=169372500, end_position_on_the_genomic_accession=169373461, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371463, CurrentID=0, Status='live', Symbol='LOC105371463', Aliases='', description='uncharacterized LOC105371463', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159509919, end_position_on_the_genomic_accession=159534693, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371464, CurrentID=0, Status='live', Symbol='LOC105371464', Aliases='', description='uncharacterized LOC105371464', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159800481, end_position_on_the_genomic_accession=159806589, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  204800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204800, CurrentID=0, Status='live', Symbol='LOC204800', Aliases='', description='actin related protein 2/3 complex subunit 1A pseudogene', other_designations='actin related protein 2/3 complex, subunit 1A, 41kDa pseudogene', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53003943, end_position_on_the_genomic_accession=53005346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371465, CurrentID=0, Status='live', Symbol='LOC105371465', Aliases='', description='uncharacterized LOC105371465', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160057212, end_position_on_the_genomic_accession=160079682, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371468, CurrentID=0, Status='live', Symbol='LOC105371468', Aliases='', description='uncharacterized LOC105371468', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160536577, end_position_on_the_genomic_accession=160571133, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371469, CurrentID=0, Status='live', Symbol='LOC105371469', Aliases='', description='endogenous retrovirus group K member 8 Gag polyprotein-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160691859, end_position_on_the_genomic_accession=160693868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371470, CurrentID=0, Status='live', Symbol='LOC105371470', Aliases='', description='uncharacterized LOC105371470', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160773528, end_position_on_the_genomic_accession=160789878, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371471, CurrentID=0, Status='live', Symbol='LOC105371471', Aliases='', description='uncharacterized LOC105371471', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161083826, end_position_on_the_genomic_accession=161087575, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128592, CurrentID=0, Status='live', Symbol='LOC100128592', Aliases='', description='family with sequence similarity 136, member A pseudogene', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=74282266, end_position_on_the_genomic_accession=74282674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371473, CurrentID=0, Status='live', Symbol='LOC105371473', Aliases='', description='uncharacterized LOC105371473', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161489310, end_position_on_the_genomic_accession=161495515, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926926, CurrentID=0, Status='live', Symbol='RDH10-AS1', Aliases='', description='RDH10 antisense RNA 1', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73310906, end_position_on_the_genomic_accession=73356461, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371474, CurrentID=0, Status='live', Symbol='LOC105371474', Aliases='', description='uncharacterized LOC105371474', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162288049, end_position_on_the_genomic_accession=162292930, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  120126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120126, CurrentID=0, Status='live', Symbol='UBTFL2', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 2 (pseudogene)', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89753012, end_position_on_the_genomic_accession=89753980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479121, CurrentID=0, Status='live', Symbol='RN7SKP59', Aliases='', description='RNA, 7SK small nuclear pseudogene 59', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  723961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723961, CurrentID=0, Status='live', Symbol='INS-IGF2', Aliases='INSIGF', description='INS-IGF2 readthrough', other_designations='INS-IGF2 readthrough transcript protein|insulin- insulin-like growth factor 2 read-through product', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2129112, end_position_on_the_genomic_accession=2161209, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  199675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199675, CurrentID=0, Status='live', Symbol='MCEMP1', Aliases='C19orf59', description='mast cell-expressed membrane protein 1', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7677057, end_position_on_the_genomic_accession=7679833, orientation='plus', exon_count=7, OMIM=609565, no_hdr0=''),
  360175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360175, CurrentID=0, Status='live', Symbol='CYCSP22', Aliases='HCP22', description='cytochrome c, somatic pseudogene 22', other_designations='', map_location='8q11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50762486, end_position_on_the_genomic_accession=50762798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  723972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=723972, CurrentID=0, Status='live', Symbol='ANP32AP1', Aliases='', description='acidic nuclear phosphoprotein 32 family member A pseudogene 1', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member A pseudogene 1|hepatopoietin PCn127', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35237326, end_position_on_the_genomic_accession=35238063, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  199692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199692, CurrentID=0, Status='live', Symbol='ZNF627', Aliases='', description='zinc finger protein 627', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11575272, end_position_on_the_genomic_accession=11619159, orientation='plus', exon_count=6, OMIM=612248, no_hdr0=''),
  199699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199699, CurrentID=0, Status='live', Symbol='DAND5', Aliases='CER2, CERL2, CKTSF1B3, COCO, CRL2, DANTE, GREM3, SP1', description='DAN domain family member 5, BMP antagonist', other_designations='DAN domain family, member 5|cerberus 2|cerberus-like 2|cerberus-like protein 2|cerl-2|cysteine knot superfamily 1, BMP antagonist 3|gremlin-3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12969618, end_position_on_the_genomic_accession=12974753, orientation='plus', exon_count=2, OMIM=609068, no_hdr0=''),
  100128601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128601, CurrentID=0, Status='live', Symbol='LOC100128601', Aliases='', description='zinc finger protein 770 pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103087196, end_position_on_the_genomic_accession=103089276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  199704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199704, CurrentID=0, Status='live', Symbol='ZNF585A', Aliases='', description='zinc finger protein 585A', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37147438, end_position_on_the_genomic_accession=37172741, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927209, CurrentID=0, Status='live', Symbol='LOC101927209', Aliases='', description='uncharacterized LOC101927209', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49161314, end_position_on_the_genomic_accession=49207720, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  199713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199713, CurrentID=0, Status='live', Symbol='NLRP7', Aliases='CLR19.4, HYDM, NALP7, NOD12, PAN7, PYPAF3', description='NLR family, pyrin domain containing 7', other_designations='NACHT, LRR and PYD containing protein 7|NACHT, leucine rich repeat and PYD containing 7|PYRIN-containing Apaf1-like protein 3|nucleotide-binding oligomerization domain protein 12|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 7', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54923509, end_position_on_the_genomic_accession=54969965, orientation='minus', exon_count=15, OMIM=609661, no_hdr0=''),
  105371483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371483, CurrentID=0, Status='live', Symbol='LOC105371483', Aliases='', description='uncharacterized LOC105371483', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1492819, end_position_on_the_genomic_accession=1495093, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  199720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199720, CurrentID=0, Status='live', Symbol='GGN', Aliases='', description='gametogenetin', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38384349, end_position_on_the_genomic_accession=38388583, orientation='minus', exon_count=5, OMIM=609966, no_hdr0=''),
  105371485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371485, CurrentID=0, Status='live', Symbol='LOC105371485', Aliases='', description='uncharacterized LOC105371485', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2017734, end_position_on_the_genomic_accession=2020706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  199731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199731, CurrentID=0, Status='live', Symbol='CADM4', Aliases='IGSF4C, NECL4, Necl-4, TSLL2, synCAM4', description='cell adhesion molecule 4', other_designations='TSLC1-like 2|TSLC1-like protein 2|immunoglobulin superfamily member 4C|immunoglobulin superfamily, member 4C|nectin-like 4|nectin-like protein 4', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43622367, end_position_on_the_genomic_accession=43639839, orientation='minus', exon_count=9, OMIM=609744, no_hdr0=''),
  100652894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652894, CurrentID=0, Status='live', Symbol='LOC100652894', Aliases='', description='uncharacterized LOC100652894', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=24530296, end_position_on_the_genomic_accession=24535705, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  724022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724022, CurrentID=0, Status='live', Symbol='MIR652', Aliases='MIRN652, hsa-mir-652', description='microRNA 652', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110055329, end_position_on_the_genomic_accession=110055426, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  724023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724023, CurrentID=0, Status='live', Symbol='MIR653', Aliases='MIRN653, hsa-mir-653, mir-653', description='microRNA 653', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93482760, end_position_on_the_genomic_accession=93482855, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  724024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724024, CurrentID=0, Status='live', Symbol='MIR654', Aliases='MIRN654, hsa-mir-654, mir-654', description='microRNA 654', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101040219, end_position_on_the_genomic_accession=101040299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  724025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724025, CurrentID=0, Status='live', Symbol='MIR655', Aliases='MIRN655, hsa-mir-655, mir-655', description='microRNA 655', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101049550, end_position_on_the_genomic_accession=101049646, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  724026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724026, CurrentID=0, Status='live', Symbol='MIR656', Aliases='MIRN656, hsa-mir-656, mir-656', description='microRNA 656', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101066724, end_position_on_the_genomic_accession=101066801, orientation='plus', exon_count=1, OMIM=616376, no_hdr0=''),
  724027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724027, CurrentID=0, Status='live', Symbol='MIR657', Aliases='MIRN657, hsa-mir-657', description='microRNA 657', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81125276, end_position_on_the_genomic_accession=81125373, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  724028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724028, CurrentID=0, Status='live', Symbol='MIR658', Aliases='MIRN658, hsa-mir-658', description='microRNA 658', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37844272, end_position_on_the_genomic_accession=37844371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  724029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724029, CurrentID=0, Status='live', Symbol='MIR659', Aliases='MIRN659, hsa-mir-659, mir-659', description='microRNA 659', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37847678, end_position_on_the_genomic_accession=37847774, orientation='minus', exon_count=1, OMIM=613556, no_hdr0=''),
  724030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724030, CurrentID=0, Status='live', Symbol='MIR660', Aliases='MIRN660, hsa-mir-660, mir-660', description='microRNA 660', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50013241, end_position_on_the_genomic_accession=50013337, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  724031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724031, CurrentID=0, Status='live', Symbol='MIR661', Aliases='MIRN661, hsa-mir-661', description='microRNA 661', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143945191, end_position_on_the_genomic_accession=143945279, orientation='minus', exon_count=1, OMIM=613716, no_hdr0=''),
  724032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724032, CurrentID=0, Status='live', Symbol='MIR662', Aliases='MIRN662, hsa-mir-662', description='microRNA 662', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=770183, end_position_on_the_genomic_accession=770277, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  724033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724033, CurrentID=0, Status='live', Symbol='MIR663A', Aliases='MIR663, MIRN663, hsa-mir-663, hsa-mir-663a, mir-663a', description='microRNA 663a', other_designations='microRNA 663', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=26208186, end_position_on_the_genomic_accession=26208278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  199746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199746, CurrentID=0, Status='live', Symbol='U2AF1L4', Aliases='U2AF1-RS3, U2AF1L3, U2AF1L3V1, U2AF1RS3, U2af26', description='U2 small nuclear RNA auxiliary factor 1-like 4', other_designations='U2 auxiliary factor 26|U2 small nuclear RNA auxiliary factor 1-like 3|U2 small nuclear RNA auxiliary factor 1-like protein 3|U2 small nuclear RNA auxiliary factor 1-like protein 4|U2(RNU2) small nuclear RNA auxiliary factor 1-like 3|U2(RNU2) small nuclear RNA auxiliary factor 1-like protein 3|U2AF1-like 4|U2AF1-like protein 3', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35742527, end_position_on_the_genomic_accession=35745442, orientation='minus', exon_count=7, OMIM=601080, no_hdr0=''),
  724038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724038, CurrentID=0, Status='live', Symbol='BDP1P', Aliases='', description='B double prime 1 pseudogene', other_designations='B double prime 1, subunit of RNA polymerase III transcription initiation factor IIIB pseudogene', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77432057, end_position_on_the_genomic_accession=77437249, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371489, CurrentID=0, Status='live', Symbol='LOC105371489', Aliases='', description='uncharacterized LOC105371489', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2563251, end_position_on_the_genomic_accession=2569136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371490, CurrentID=0, Status='live', Symbol='LOC105371490', Aliases='', description='uncharacterized LOC105371490', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2688240, end_position_on_the_genomic_accession=2689308, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371491, CurrentID=0, Status='live', Symbol='LOC105371491', Aliases='', description='uncharacterized LOC105371491', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2720801, end_position_on_the_genomic_accession=2723258, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  724058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724058, CurrentID=0, Status='live', Symbol='ANIB4', Aliases='', description='aneurysm, intracranial berry 4', other_designations='', map_location='5p15.2-p14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610213, no_hdr0=''),
  724059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724059, CurrentID=0, Status='live', Symbol='TMEM97P2', Aliases='', description='transmembrane protein 97 pseudogene 2', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20932572, end_position_on_the_genomic_accession=20934863, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724060, CurrentID=0, Status='live', Symbol='LOC724060', Aliases='', description='TRA2A pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52161043, end_position_on_the_genomic_accession=52163486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652901, CurrentID=0, Status='live', Symbol='LOC100652901', Aliases='', description='uncharacterized LOC100652901', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25343582, end_position_on_the_genomic_accession=25345261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  199777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199777, CurrentID=0, Status='live', Symbol='ZNF626', Aliases='', description='zinc finger protein 626', other_designations='CTC-513N18.7', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20619939, end_position_on_the_genomic_accession=20661596, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  724066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724066, CurrentID=0, Status='live', Symbol='ATXN8', Aliases='ATXN3', description='ataxin 8', other_designations='protein 1C2', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613289, no_hdr0=''),
  724067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724067, CurrentID=0, Status='live', Symbol='DEFA7P', Aliases='DEFA7', description='defensin alpha 7, pseudogene', other_designations='defensin, alpha 7 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7038481, end_position_on_the_genomic_accession=7039439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724068, CurrentID=0, Status='live', Symbol='DEFA11P', Aliases='DEFAP3', description='defensin alpha 11, pseudogene', other_designations='defensin, alpha 11 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7028601, end_position_on_the_genomic_accession=7029489, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  199786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199786, CurrentID=0, Status='live', Symbol='FAM129C', Aliases='BCNP1', description='family with sequence similarity 129 member C', other_designations='B-cell novel protein 1|family with sequence similarity 129, member C', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17523301, end_position_on_the_genomic_accession=17555545, orientation='plus', exon_count=21, OMIM=609967, no_hdr0=''),
  105379266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379266, CurrentID=0, Status='live', Symbol='LOC105379266', Aliases='', description='uncharacterized LOC105379266', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9288977, end_position_on_the_genomic_accession=9291228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  724084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724084, CurrentID=0, Status='live', Symbol='LOC724084', Aliases='', description='zinc finger protein 706 pseudogene', other_designations='', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224175147, end_position_on_the_genomic_accession=224175910, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724085, CurrentID=0, Status='live', Symbol='LOC724085', Aliases='', description='immediate early response 3 interacting protein 1 pseudogene', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111999233, end_position_on_the_genomic_accession=112001068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724087, CurrentID=0, Status='live', Symbol='LINC01545', Aliases='CXorf31', description='long intergenic non-protein coding RNA 1545', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46887417, end_position_on_the_genomic_accession=46899704, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  199800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199800, CurrentID=0, Status='live', Symbol='ADM5', Aliases='AM5, C19orf76', description='adrenomedullin 5 (putative)', other_designations='adrenomedullin 5 homolog (pig)', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49688685, end_position_on_the_genomic_accession=49690990, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371498, CurrentID=0, Status='live', Symbol='LOC105371498', Aliases='', description='uncharacterized LOC105371498', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4616764, end_position_on_the_genomic_accession=4629907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  724094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724094, CurrentID=0, Status='live', Symbol='FAM27D1', Aliases='', description='family with sequence similarity 27 member D1', other_designations='family with sequence similarity 27, member D1', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=46390277, end_position_on_the_genomic_accession=46424817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104533120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104533120, CurrentID=0, Status='live', Symbol='TRIM31-AS1', Aliases='', description='TRIM31 antisense RNA 1', other_designations='TRIM31 antisense RNA 1 (non-protein coding)|XXbac-BPG250I8.13', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30105240, end_position_on_the_genomic_accession=30114724, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  724101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724101, CurrentID=0, Status='live', Symbol='SS18L2P2', Aliases='', description='synovial sarcoma translocation gene on chromosome 18-like 2 pseudogene 2', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51420653, end_position_on_the_genomic_accession=51421241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724102, CurrentID=0, Status='live', Symbol='SNHG4', Aliases='NCRNA00059, U19H', description='small nucleolar RNA host gene 4', other_designations='U19 host|small nucleolar RNA host gene (non-protein coding) 4|small nucleolar RNA host gene 4 (non-protein coding)', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139273752, end_position_on_the_genomic_accession=139283184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  724104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724104, CurrentID=0, Status='live', Symbol='LOC724104', Aliases='', description='SH3 domain binding glutamate-rich protein like 3 pseudogene', other_designations='SH3 domain binding glutamic acid-rich protein like 3 pseudogene', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52657317, end_position_on_the_genomic_accession=52657897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371500, CurrentID=0, Status='live', Symbol='LOC105371500', Aliases='', description='uncharacterized LOC105371500', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4713950, end_position_on_the_genomic_accession=4714980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422179, CurrentID=0, Status='live', Symbol='KCTD9P5', Aliases='', description='potassium channel tetramerization domain containing 9 pseudogene 5', other_designations='potassium channel tetramerisation domain containing 9 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37505691, end_position_on_the_genomic_accession=37506858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  724107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724107, CurrentID=0, Status='live', Symbol='SPG32', Aliases='SPG29', description='spastic paraplegia 32 (autosomal recessive)', other_designations='', map_location='14q12-q21', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611252, no_hdr0=''),
  724110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724110, CurrentID=0, Status='live', Symbol='SPG34', Aliases='', description='spastic paraplegia 34 (autosomal dominant)', other_designations='', map_location='Xq24-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300750, no_hdr0=''),
  100652909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652909, CurrentID=0, Status='live', Symbol='LINC00904', Aliases='', description='long intergenic non-protein coding RNA 904', other_designations='CTC-523E23.13', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34892029, end_position_on_the_genomic_accession=34898943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  199827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199827, CurrentID=0, Status='live', Symbol='RNASEH1P3', Aliases='', description='ribonuclease H1 pseudogene 3', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210303676, end_position_on_the_genomic_accession=210304818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371502, CurrentID=0, Status='live', Symbol='LOC105371502', Aliases='', description='uncharacterized LOC105371502', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5025614, end_position_on_the_genomic_accession=5031613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  199834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199834, CurrentID=0, Status='live', Symbol='LCE4A', Aliases='LEP8, SPRL4A', description='late cornified envelope 4A', other_designations='late envelope protein 8|small proline rich-like (epidermal differentiation complex) 4A|small proline-rich-like epidermal differentiation complex protein 4A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152709047, end_position_on_the_genomic_accession=152709434, orientation='plus', exon_count=1, OMIM=612618, no_hdr0=''),
  105371503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371503, CurrentID=0, Status='live', Symbol='LOC105371503', Aliases='', description='uncharacterized LOC105371503', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162534828, end_position_on_the_genomic_accession=162549066, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371505, CurrentID=0, Status='live', Symbol='LOC105371505', Aliases='', description='uncharacterized LOC105371505', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5360961, end_position_on_the_genomic_accession=5362905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371506, CurrentID=0, Status='live', Symbol='LOC105371506', Aliases='', description='uncharacterized LOC105371506', other_designations='CTC-524C5.5', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5469057, end_position_on_the_genomic_accession=5470392, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  199857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199857, CurrentID=0, Status='live', Symbol='ALG14', Aliases='CMS15', description='ALG14, UDP-N-acetylglucosaminyltransferase subunit', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94981565, end_position_on_the_genomic_accession=95072964, orientation='minus', exon_count=6, OMIM=612866, no_hdr0=''),
  105371507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371507, CurrentID=0, Status='live', Symbol='LOC105371507', Aliases='', description='uncharacterized LOC105371507', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5539697, end_position_on_the_genomic_accession=5554775, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371508, CurrentID=0, Status='live', Symbol='LOC105371508', Aliases='', description='uncharacterized LOC105371508', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6149980, end_position_on_the_genomic_accession=6190547, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101926933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926933, CurrentID=0, Status='live', Symbol='LOC101926933', Aliases='', description='uncharacterized LOC101926933', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22929609, end_position_on_the_genomic_accession=22954692, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  199870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199870, CurrentID=0, Status='live', Symbol='FAM76A', Aliases='', description='family with sequence similarity 76 member A', other_designations='family with sequence similarity 76, member A', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27725979, end_position_on_the_genomic_accession=27762915, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105371510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371510, CurrentID=0, Status='live', Symbol='LOC105371510', Aliases='', description='uncharacterized LOC105371510', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  199882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199882, CurrentID=0, Status='live', Symbol='LOC199882', Aliases='', description='phosphodiesterase 4D interacting protein-like', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143811357, end_position_on_the_genomic_accession=143840630, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101929675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929675, CurrentID=0, Status='live', Symbol='LOC101929675', Aliases='', description='uncharacterized LOC101929675', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64410274, end_position_on_the_genomic_accession=64462935, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  139249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139249, CurrentID=0, Status='live', Symbol='FAM47DP', Aliases='', description='family with sequence similarity 47 member D, pseudogene', other_designations='family with sequence similarity 47, member D, pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37540929, end_position_on_the_genomic_accession=37545259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926934, CurrentID=0, Status='live', Symbol='LOC101926934', Aliases='', description='uncharacterized LOC101926934', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46670444, end_position_on_the_genomic_accession=46687833, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  199899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199899, CurrentID=0, Status='live', Symbol='LINC00466', Aliases='', description='long intergenic non-protein coding RNA 466', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63159083, end_position_on_the_genomic_accession=63317230, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  199900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199900, CurrentID=0, Status='live', Symbol='MRPS21P1', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 1', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65092202, end_position_on_the_genomic_accession=65092685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371516, CurrentID=0, Status='live', Symbol='LOC105371516', Aliases='', description='uncharacterized LOC105371516', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7832891, end_position_on_the_genomic_accession=7834111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371517, CurrentID=0, Status='live', Symbol='LOC105371517', Aliases='', description='uncharacterized LOC105371517', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7951369, end_position_on_the_genomic_accession=7961886, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  199920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199920, CurrentID=0, Status='live', Symbol='C1orf168', Aliases='', description='chromosome 1 open reading frame 168', other_designations='RP4-758N20.2', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56718804, end_position_on_the_genomic_accession=56819696, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  105371518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371518, CurrentID=0, Status='live', Symbol='LOC105371518', Aliases='', description='uncharacterized LOC105371518', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8056900, end_position_on_the_genomic_accession=8063456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371519, CurrentID=0, Status='live', Symbol='LOC105371519', Aliases='', description='uncharacterized LOC105371519', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8191288, end_position_on_the_genomic_accession=8199496, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371520, CurrentID=0, Status='live', Symbol='LOC105371520', Aliases='', description='uncharacterized LOC105371520', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8174623, end_position_on_the_genomic_accession=8183616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371521, CurrentID=0, Status='live', Symbol='LOC105371521', Aliases='', description='uncharacterized LOC105371521', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163769385, end_position_on_the_genomic_accession=163842492, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371522, CurrentID=0, Status='live', Symbol='LOC105371522', Aliases='', description='uncharacterized LOC105371522', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8119746, end_position_on_the_genomic_accession=8144268, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  199953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199953, CurrentID=0, Status='live', Symbol='TMEM201', Aliases='NET5', description='transmembrane protein 201', other_designations='RP13-15M17.2|spindle-associated membrane protein 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9588853, end_position_on_the_genomic_accession=9614877, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105371523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371523, CurrentID=0, Status='live', Symbol='LOC105371523', Aliases='', description='uncharacterized LOC105371523', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8365569, end_position_on_the_genomic_accession=8382136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128644, CurrentID=0, Status='live', Symbol='LOC100128644', Aliases='', description='LMNE6487', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=12948885, end_position_on_the_genomic_accession=12949492, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  199964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199964, CurrentID=0, Status='live', Symbol='TMEM61', Aliases='', description='transmembrane protein 61', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54980595, end_position_on_the_genomic_accession=54992293, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644997, CurrentID=0, Status='live', Symbol='PIK3CD-AS1', Aliases='C1orf200', description='PIK3CD antisense RNA 1', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9652610, end_position_on_the_genomic_accession=9654586, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644998, CurrentID=0, Status='live', Symbol='KRT8P19', Aliases='', description='keratin 8 pseudogene 19', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61879244, end_position_on_the_genomic_accession=61880972, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  199974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199974, CurrentID=0, Status='live', Symbol='CYP4Z1', Aliases='CYP4A20', description='cytochrome P450 family 4 subfamily Z member 1', other_designations='CYPIVZ1|cytochrome P450, family 4, subfamily Z, polypeptide 1', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47067448, end_position_on_the_genomic_accession=47118320, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100130632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130632, CurrentID=0, Status='live', Symbol='LOC100130632', Aliases='', description='zinc finger protein 530-like', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34635013, end_position_on_the_genomic_accession=34636132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371528, CurrentID=0, Status='live', Symbol='LOC105371528', Aliases='', description='uncharacterized LOC105371528', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8718594, end_position_on_the_genomic_accession=8721377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  199990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=199990, CurrentID=0, Status='live', Symbol='FAAP20', Aliases='C1orf86, FP7162', description='Fanconi anemia core complex associated protein 20', other_designations='FANCA-associated protein of 20 kDa|Fanconi anemia-associated protein of 20 kDa', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2184460, end_position_on_the_genomic_accession=2212720, orientation='minus', exon_count=20, OMIM=615183, no_hdr0=''),
  105371529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371529, CurrentID=0, Status='live', Symbol='LOC105371529', Aliases='', description='uncharacterized LOC105371529', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8744956, end_position_on_the_genomic_accession=8749113, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479132, CurrentID=0, Status='live', Symbol='RN7SKP85', Aliases='', description='RNA, 7SK small nuclear pseudogene 85', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128651, CurrentID=0, Status='live', Symbol='LOC100128651', Aliases='', description='uncharacterized LOC100128651', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=5953402, end_position_on_the_genomic_accession=5954868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  200008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200008, CurrentID=0, Status='live', Symbol='CDCP2', Aliases='', description='CUB domain containing protein 2', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54138995, end_position_on_the_genomic_accession=54153006, orientation='minus', exon_count=4, OMIM=612320, no_hdr0=''),
  200010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200010, CurrentID=0, Status='live', Symbol='SLC5A9', Aliases='SGLT4', description='solute carrier family 5 (sodium/sugar cotransporter), member 9', other_designations='Na(+)/glucose cotransporter 4|hSGLT4|solute carrier family 5 (sodium/glucose cotransporter), member 9|solute carrier family 5 member 9', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48222270, end_position_on_the_genomic_accession=48248644, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  283278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283278, CurrentID=0, Status='live', Symbol='LOC283278', Aliases='', description='uncharacterized LOC283278', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=16800509, end_position_on_the_genomic_accession=16809037, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  200014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200014, CurrentID=0, Status='live', Symbol='CC2D1B', Aliases='', description='coiled-coil and C2 domain containing 1B', other_designations='FRE under dual repression-binding protein 2|five prime repressor element under dual repression-binding protein 2|freud-2', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52350593, end_position_on_the_genomic_accession=52366205, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  100128653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128653, CurrentID=0, Status='live', Symbol='LOC100128653', Aliases='', description='uncharacterized LOC100128653', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1532454, end_position_on_the_genomic_accession=1534880, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  200025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200025, CurrentID=0, Status='live', Symbol='PFN1P3', Aliases='COAS-3, COAS3', description='profilin 1 pseudogene 3', other_designations='Putative chromosome 1-amplified sequence 3', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145407635, end_position_on_the_genomic_accession=145408413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128655, CurrentID=0, Status='live', Symbol='LOC100128655', Aliases='', description='ribosomal protein S2 pseudogene', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39039524, end_position_on_the_genomic_accession=39039985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200030, CurrentID=0, Status='live', Symbol='NBPF11', Aliases='NBPF24', description='neuroblastoma breakpoint family member 11', other_designations='neuroblastoma breakpoint family member 24|neuroblastoma breakpoint family, member 11|neuroblastoma breakpoint family, member 24', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148102046, end_position_on_the_genomic_accession=148152190, orientation='minus', exon_count=25, OMIM=614001, no_hdr0=''),
  100421179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421179, CurrentID=0, Status='live', Symbol='LOC100421179', Aliases='', description='component of oligomeric golgi complex 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55137918, end_position_on_the_genomic_accession=55138559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200035, CurrentID=0, Status='live', Symbol='NUDT17', Aliases='', description='nudix hydrolase 17', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 17|nudix motif 17', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145844985, end_position_on_the_genomic_accession=145848753, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371537, CurrentID=0, Status='live', Symbol='LOC105371537', Aliases='', description='uncharacterized LOC105371537', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11836963, end_position_on_the_genomic_accession=11837976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371538, CurrentID=0, Status='live', Symbol='LOC105371538', Aliases='', description='uncharacterized LOC105371538', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11560805, end_position_on_the_genomic_accession=11598460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390739, CurrentID=0, Status='live', Symbol='LOC390739', Aliases='', description='MYC binding protein pseudogene', other_designations='c-myc binding protein pseudogene', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72278241, end_position_on_the_genomic_accession=72280776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371539, CurrentID=0, Status='live', Symbol='LOC105371539', Aliases='', description='uncharacterized LOC105371539', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12002357, end_position_on_the_genomic_accession=12020722, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  200058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200058, CurrentID=0, Status='live', Symbol='FLJ23867', Aliases='', description='uncharacterized protein FLJ23867', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180198009, end_position_on_the_genomic_accession=180200725, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  140123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=140123, CurrentID=0, Status='live', Symbol='RBMY2CP', Aliases='RBMY2', description='RNA binding motif protein, Y-linked, family 2, member C pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 2 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23536902, end_position_on_the_genomic_accession=23549031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371542, CurrentID=0, Status='live', Symbol='LOC105371542', Aliases='', description='uncharacterized LOC105371542', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165267631, end_position_on_the_genomic_accession=165287477, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371543, CurrentID=0, Status='live', Symbol='LOC105371543', Aliases='', description='uncharacterized LOC105371543', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13353148, end_position_on_the_genomic_accession=13387067, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374911, CurrentID=0, Status='live', Symbol='LOC105374911', Aliases='', description='uncharacterized LOC105374911', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8283203, end_position_on_the_genomic_accession=8307516, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379786, CurrentID=0, Status='live', Symbol='LOC105379786', Aliases='', description='putative exonuclease GOR-like protein', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  200081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200081, CurrentID=0, Status='live', Symbol='TXLNA', Aliases='IL14, TXLN', description='taxilin alpha', other_designations='interleukin 14', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32179695, end_position_on_the_genomic_accession=32198285, orientation='plus', exon_count=11, OMIM=608676, no_hdr0=''),
  105371545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371545, CurrentID=0, Status='live', Symbol='LOC105371545', Aliases='', description='uncharacterized LOC105371545', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14800206, end_position_on_the_genomic_accession=14802273, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  404719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404719, CurrentID=0, Status='live', Symbol='HSCR8', Aliases='HSCRM2', description='Hirschsprung disease modifier 2', other_designations='', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608462, no_hdr0=''),
  105371546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371546, CurrentID=0, Status='live', Symbol='LOC105371546', Aliases='', description='uncharacterized LOC105371546', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165247305, end_position_on_the_genomic_accession=165258654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100652955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652955, CurrentID=0, Status='live', Symbol='LOC100652955', Aliases='', description='COP9 signalosome complex subunit 3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187441651, end_position_on_the_genomic_accession=187444298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128668, CurrentID=0, Status='live', Symbol='LOC100128668', Aliases='', description='uncharacterized LOC100128668', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=8176923, end_position_on_the_genomic_accession=8177312, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128670, CurrentID=0, Status='live', Symbol='LOC100128670', Aliases='', description='uncharacterized LOC100128670', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=91276321, end_position_on_the_genomic_accession=91276698, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421182, CurrentID=0, Status='live', Symbol='PGBD4P7', Aliases='', description='piggyBac transposable element derived 4 pseudogene 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94672921, end_position_on_the_genomic_accession=94674169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371552, CurrentID=0, Status='live', Symbol='LOC105371552', Aliases='', description='uncharacterized LOC105371552', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16822427, end_position_on_the_genomic_accession=16829837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  200132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200132, CurrentID=0, Status='live', Symbol='TCTEX1D1', Aliases='', description='Tctex1 domain containing 1', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66752457, end_position_on_the_genomic_accession=66779047, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100128673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128673, CurrentID=0, Status='live', Symbol='LOC100128673', Aliases='', description='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74727889, end_position_on_the_genomic_accession=74728959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926942, CurrentID=0, Status='live', Symbol='LOC101926942', Aliases='', description='uncharacterized LOC101926942', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90402521, end_position_on_the_genomic_accession=90540805, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100131506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131506, CurrentID=0, Status='live', Symbol='LOC100131506', Aliases='', description='uncharacterized LOC100131506', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=10114051, end_position_on_the_genomic_accession=10116458, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128674, CurrentID=0, Status='live', Symbol='LOC100128674', Aliases='', description='SURP and G-patch domain containing 1 pseudogene', other_designations='SURP and G patch domain containing 1 pseudogene|splicing factor 4 pseudogene', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74170436, end_position_on_the_genomic_accession=74171885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927058, CurrentID=0, Status='live', Symbol='LOC101927058', Aliases='', description='uncharacterized LOC101927058', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42615503, end_position_on_the_genomic_accession=42646498, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  645027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645027, CurrentID=0, Status='live', Symbol='EVPLL', Aliases='', description='envoplakin-like', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18377692, end_position_on_the_genomic_accession=18389652, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  200149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200149, CurrentID=0, Status='live', Symbol='TUBB8P6', Aliases='', description='tubulin beta 8 class VIII pseudogene 6', other_designations='HSA1q43-44 beta-tubulin 4Q (TUBB4Q) pseudogene|tubulin, beta 8 class VIII pseudogene 6|tubulin, beta 8 pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242057085, end_position_on_the_genomic_accession=242060446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200150, CurrentID=0, Status='live', Symbol='PLD5', Aliases='PLDC', description='phospholipase D family member 5', other_designations='inactive PLD 5|inactive choline phosphatase 5|inactive phosphatidylcholine-hydrolyzing phospholipase D5|phospholipase D family, member 5', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242088387, end_position_on_the_genomic_accession=242530544, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  100128676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128676, CurrentID=0, Status='live', Symbol='LOC100128676', Aliases='', description='TBCC domain containing 1 pseudogene', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56378831, end_position_on_the_genomic_accession=56382850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200159, CurrentID=0, Status='live', Symbol='C1orf100', Aliases='', description='chromosome 1 open reading frame 100', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244352069, end_position_on_the_genomic_accession=244389874, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  200162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200162, CurrentID=0, Status='live', Symbol='SPAG17', Aliases='CT143, PF6', description='sperm associated antigen 17', other_designations='projection protein PF6 homolog', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117953665, end_position_on_the_genomic_accession=118185225, orientation='minus', exon_count=56, OMIM=616554, no_hdr0=''),
  100128678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128678, CurrentID=0, Status='live', Symbol='BTBD10P1', Aliases='', description='BTB (POZ) domain containing 10 pseudogene 1', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52748759, end_position_on_the_genomic_accession=52750324, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926943, CurrentID=0, Status='live', Symbol='LOC101926943', Aliases='', description='uncharacterized LOC101926943', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74688937, end_position_on_the_genomic_accession=74728918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100652967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652967, CurrentID=0, Status='live', Symbol='LOC100652967', Aliases='', description='chromosome 14 open reading frame 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52925250, end_position_on_the_genomic_accession=52925432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200172, CurrentID=0, Status='live', Symbol='SLFNL1', Aliases='', description='schlafen like 1', other_designations='schlafen-like 1', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41015597, end_position_on_the_genomic_accession=41023444, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  266694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=266694, CurrentID=0, Status='live', Symbol='LOC266694', Aliases='', description='embryonic ectoderm development pseudogene', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140397057, end_position_on_the_genomic_accession=140398543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371561, CurrentID=0, Status='live', Symbol='LOC105371561', Aliases='', description='uncharacterized LOC105371561', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165204047, end_position_on_the_genomic_accession=165213091, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100288628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288628, CurrentID=0, Status='live', Symbol='LOC100288628', Aliases='', description='SUMO1/sentrin specific peptidase 6 pseudogene', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=58071691, end_position_on_the_genomic_accession=58072722, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200185, CurrentID=0, Status='live', Symbol='KRTCAP2', Aliases='KCP2', description='keratinocyte associated protein 2', other_designations='KCP-2|keratinocytes associated protein 2', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155169408, end_position_on_the_genomic_accession=155173328, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  200186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200186, CurrentID=0, Status='live', Symbol='CRTC2', Aliases='TORC-2, TORC2', description='CREB regulated transcription coactivator 2', other_designations='transducer of regulated cAMP response element-binding protein (CREB) 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153947672, end_position_on_the_genomic_accession=153958656, orientation='minus', exon_count=15, OMIM=608972, no_hdr0=''),
  105371562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371562, CurrentID=0, Status='live', Symbol='LOC105371562', Aliases='', description='uncharacterized LOC105371562', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16982763, end_position_on_the_genomic_accession=17006087, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  200197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200197, CurrentID=0, Status='live', Symbol='TMEM51-AS1', Aliases='C1orf126', description='TMEM51 antisense RNA 1', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15111815, end_position_on_the_genomic_accession=15152464, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101926944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926944, CurrentID=0, Status='live', Symbol='LOC101926944', Aliases='', description='uncharacterized LOC101926944', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59772795, end_position_on_the_genomic_accession=59788832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100652972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652972, CurrentID=0, Status='live', Symbol='LOC100652972', Aliases='', description='uncharacterized LOC100652972', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=40685804, end_position_on_the_genomic_accession=40686841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  200205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200205, CurrentID=0, Status='live', Symbol='IBA57', Aliases='C1orf69, MMDS3, SPG74', description='IBA57 homolog, iron-sulfur cluster assembly', other_designations='IBA57, iron-sulfur cluster assembly homolog|iron-sulfur cluster assembly factor for biotin synthase- and aconitase-like mitochondrial proteins, with a mass of 57kDa|putative transferase C1orf69, mitochondrial', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228165728, end_position_on_the_genomic_accession=228197905, orientation='plus', exon_count=5, OMIM=615316, no_hdr0=''),
  100128685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128685, CurrentID=0, Status='live', Symbol='SLC16A14P1', Aliases='', description='solute carrier family 16 member 14 pseudogene 1', other_designations='solute carrier family 16, member 14 (monocarboxylic acid transporter 14) pseudogene 1|solute carrier family 16, member 14 pseudogene 1', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106215765, end_position_on_the_genomic_accession=106217305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371566, CurrentID=0, Status='live', Symbol='LOC105371566', Aliases='', description='uncharacterized LOC105371566', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18102005, end_position_on_the_genomic_accession=18117789, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100421185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421185, CurrentID=0, Status='live', Symbol='PGBD4P8', Aliases='', description='piggyBac transposable element derived 4 pseudogene 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60097401, end_position_on_the_genomic_accession=60097977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371568, CurrentID=0, Status='live', Symbol='LOC105371568', Aliases='', description='uncharacterized LOC105371568', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18146972, end_position_on_the_genomic_accession=18153195, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  200232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200232, CurrentID=0, Status='live', Symbol='FAM209A', Aliases='C20orf106, dJ1153D9.3', description='family with sequence similarity 209 member A', other_designations='family with sequence similarity 209, member A|uncharacterized protein C20orf106', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56524729, end_position_on_the_genomic_accession=56526152, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131200, CurrentID=0, Status='live', Symbol='LOC100131200', Aliases='', description='mannose-6-phosphate receptor (cation dependent) pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110636866, end_position_on_the_genomic_accession=110637674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371570, CurrentID=0, Status='live', Symbol='LOC105371570', Aliases='', description='uncharacterized LOC105371570', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18437978, end_position_on_the_genomic_accession=18440057, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128691, CurrentID=0, Status='live', Symbol='LOC100128691', Aliases='', description='hCG1810796-like', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=115783249, end_position_on_the_genomic_accession=115785557, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724581, CurrentID=0, Status='live', Symbol='LOC102724581', Aliases='', description='uncharacterized LOC102724581', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65385613, end_position_on_the_genomic_accession=65396023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926946, CurrentID=0, Status='live', Symbol='LOC101926946', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90300808, end_position_on_the_genomic_accession=90301522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200261, CurrentID=0, Status='live', Symbol='LINC00656', Aliases='', description='long intergenic non-protein coding RNA 656', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23125068, end_position_on_the_genomic_accession=23132636, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100533913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533913, CurrentID=0, Status='live', Symbol='LOC100533913', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371575, CurrentID=0, Status='live', Symbol='LOC105371575', Aliases='', description='uncharacterized LOC105371575', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165598380, end_position_on_the_genomic_accession=165606862, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422182, CurrentID=0, Status='live', Symbol='LOC100422182', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371576, CurrentID=0, Status='live', Symbol='LOC105371576', Aliases='', description='uncharacterized LOC105371576', other_designations='CTB-187M2.3', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19470544, end_position_on_the_genomic_accession=19493214, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128697, CurrentID=0, Status='live', Symbol='LOC100128697', Aliases='', description='uncharacterized LOC100128697', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=134929408, end_position_on_the_genomic_accession=134929845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371579, CurrentID=0, Status='live', Symbol='LOC105371579', Aliases='', description='uncharacterized LOC105371579', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20084462, end_position_on_the_genomic_accession=20086872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927059, CurrentID=0, Status='live', Symbol='LOC101927059', Aliases='', description='uncharacterized LOC101927059', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=114576041, end_position_on_the_genomic_accession=114579970, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371580, CurrentID=0, Status='live', Symbol='LOC105371580', Aliases='', description='uncharacterized LOC105371580', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20574079, end_position_on_the_genomic_accession=20579627, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  200298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200298, CurrentID=0, Status='live', Symbol='LINC00528', Aliases='C22orf37', description='long intergenic non-protein coding RNA 528', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17777290, end_position_on_the_genomic_accession=17779481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371581, CurrentID=0, Status='live', Symbol='LOC105371581', Aliases='', description='uncharacterized LOC105371581', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165707131, end_position_on_the_genomic_accession=165709500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371582, CurrentID=0, Status='live', Symbol='LOC105371582', Aliases='', description='uncharacterized LOC105371582', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20614277, end_position_on_the_genomic_accession=20633230, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  200312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200312, CurrentID=0, Status='live', Symbol='RNF215', Aliases='', description='ring finger protein 215', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30378812, end_position_on_the_genomic_accession=30387313, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  200315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200315, CurrentID=0, Status='live', Symbol='APOBEC3A', Aliases='A3A, ARP3, PHRBN, bK150C2.1', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3A', other_designations='phorbolin 1|phorbolin-1|probable DNA dC->dU-editing enzyme APOBEC-3A', map_location='22q13.1-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38957522, end_position_on_the_genomic_accession=38963183, orientation='plus', exon_count=5, OMIM=607109, no_hdr0=''),
  200316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200316, CurrentID=0, Status='live', Symbol='APOBEC3F', Aliases='A3F, ARP8, BK150C2.4.MRNA, KA6', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 3F', other_designations='apolipoprotein B editing enzyme catalytic polypeptide-like 3F|apolipoprotein B mRNA editing enzyme cytidine deaminase|apolipoprotein B mRNA-editing enzyme catalytic polypeptide-like 3F|induced upon T-cell activation', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39040604, end_position_on_the_genomic_accession=39055972, orientation='plus', exon_count=8, OMIM=608993, no_hdr0=''),
  105371585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371585, CurrentID=0, Status='live', Symbol='LOC105371585', Aliases='', description='uncharacterized LOC105371585', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421189, CurrentID=0, Status='live', Symbol='GAPDHP49', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 49', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222503919, end_position_on_the_genomic_accession=222504836, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371587, CurrentID=0, Status='live', Symbol='LOC105371587', Aliases='', description='uncharacterized LOC105371587', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21078584, end_position_on_the_genomic_accession=21085271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103091863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103091863, CurrentID=0, Status='live', Symbol='LOC103091863', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3276053, end_position_on_the_genomic_accession=3276394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103091864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103091864, CurrentID=0, Status='live', Symbol='SNHG22', Aliases='SCARNA17, SCARNA17HG', description='small nucleolar RNA host gene 22', other_designations='small nucleolar RNA host gene 22 (non-protein coding)', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49814023, end_position_on_the_genomic_accession=49851059, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103091865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103091865, CurrentID=0, Status='live', Symbol='BRWD1-IT2', Aliases='', description='BRWD1 intronic transcript 2 (non-protein coding)', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39313907, end_position_on_the_genomic_accession=39314915, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103091866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103091866, CurrentID=0, Status='live', Symbol='LOC103091866', Aliases='', description='uncharacterized LOC103091866', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145168405, end_position_on_the_genomic_accession=145216071, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  200350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200350, CurrentID=0, Status='live', Symbol='FOXD4L1', Aliases='FOXD5, bA395L14.1', description='forkhead box D4-like 1', other_designations='FOXD4-like 1|forkhead box D4 like 1', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113499084, end_position_on_the_genomic_accession=113501150, orientation='plus', exon_count=1, OMIM=611084, no_hdr0=''),
  105371589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371589, CurrentID=0, Status='live', Symbol='LOC105371589', Aliases='', description='uncharacterized LOC105371589', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21496983, end_position_on_the_genomic_accession=21521455, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371590, CurrentID=0, Status='live', Symbol='LOC105371590', Aliases='', description='uncharacterized LOC105371590', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166475844, end_position_on_the_genomic_accession=166490223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479144, CurrentID=0, Status='live', Symbol='RN7SKP111', Aliases='', description='RNA, 7SK small nuclear pseudogene 111', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128711, CurrentID=0, Status='live', Symbol='NDUFB10P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B10 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 10, 22kDa pseudogene', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57325394, end_position_on_the_genomic_accession=57326513, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128712, CurrentID=0, Status='live', Symbol='PRELID3BP3', Aliases='SLMO2P3', description='PRELI domain containing 3B pseudogene 3', other_designations='SLMO2 pseudogene 3|slowmo homolog 2 pseudogene 3', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62841008, end_position_on_the_genomic_accession=62841834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200373, CurrentID=0, Status='live', Symbol='CFAP221', Aliases='FAP221, PCDP1', description='cilia and flagella associated protein 221', other_designations='flagellar associated protein 221 homolog|primary ciliary dyskinesia 1 homolog|primary ciliary dyskinesia protein 1', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119544423, end_position_on_the_genomic_accession=119660321, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  105371593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371593, CurrentID=0, Status='live', Symbol='LOC105371593', Aliases='', description='uncharacterized LOC105371593', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166894906, end_position_on_the_genomic_accession=166908107, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371594, CurrentID=0, Status='live', Symbol='LOC105371594', Aliases='', description='uncharacterized LOC105371594', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166947755, end_position_on_the_genomic_accession=166973184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377627, CurrentID=0, Status='live', Symbol='LOC105377627', Aliases='', description='uncharacterized LOC105377627', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51332725, end_position_on_the_genomic_accession=51342650, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  200403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200403, CurrentID=0, Status='live', Symbol='VWA3B', Aliases='', description='von Willebrand factor A domain containing 3B', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98087132, end_position_on_the_genomic_accession=98330681, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  200407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200407, CurrentID=0, Status='live', Symbol='CREG2', Aliases='', description='cellular repressor of E1A-stimulated genes 2', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101348354, end_position_on_the_genomic_accession=101387988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371599, CurrentID=0, Status='live', Symbol='LOC105371599', Aliases='', description='uncharacterized LOC105371599', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167079856, end_position_on_the_genomic_accession=167088513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479146, CurrentID=0, Status='live', Symbol='RN7SKP115', Aliases='', description='RNA, 7SK small nuclear pseudogene 115', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  200420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200420, CurrentID=0, Status='live', Symbol='ALMS1P1', Aliases='ALMS1L, ALMS1P', description='ALMS1, centrosome and basal body associated protein pseudogene 1', other_designations='Alstrom syndrome 1 pseudogene', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73644919, end_position_on_the_genomic_accession=73685572, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371601, CurrentID=0, Status='live', Symbol='LOC105371601', Aliases='', description='uncharacterized LOC105371601', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167166958, end_position_on_the_genomic_accession=167172159, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200424, CurrentID=0, Status='live', Symbol='TET3', Aliases='hCG_40738', description='tet methylcytosine dioxygenase 3', other_designations='probable methylcytosine dioxygenase TET3|putative methylcytosine dioxygenase|tet oncogene family member 3', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73983335, end_position_on_the_genomic_accession=74108176, orientation='plus', exon_count=15, OMIM=613555, no_hdr0=''),
  100128723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128723, CurrentID=0, Status='live', Symbol='CBX1P3', Aliases='', description='chromobox 1 pseudogene 3', other_designations='chromobox homolog 1 (HP1 beta homolog Drosophila ) pseudogene|chromobox homolog 1 pseudogene 3', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203954398, end_position_on_the_genomic_accession=203958029, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371605, CurrentID=0, Status='live', Symbol='LOC105371605', Aliases='', description='uncharacterized LOC105371605', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169371461, end_position_on_the_genomic_accession=169373946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371606, CurrentID=0, Status='live', Symbol='LOC105371606', Aliases='', description='uncharacterized LOC105371606', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168803948, end_position_on_the_genomic_accession=168870199, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128727, CurrentID=0, Status='live', Symbol='LOC100128727', Aliases='', description='uncharacterized LOC100128727', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=136722150, end_position_on_the_genomic_accession=136723387, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371609, CurrentID=0, Status='live', Symbol='LOC105371609', Aliases='', description='uncharacterized LOC105371609', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170173928, end_position_on_the_genomic_accession=170241844, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105371611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371611, CurrentID=0, Status='live', Symbol='LOC105371611', Aliases='', description='uncharacterized LOC105371611', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171120402, end_position_on_the_genomic_accession=171251882, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645084, CurrentID=0, Status='live', Symbol='DBF4P1', Aliases='', description='DBF4 zinc finger pseudogene 1', other_designations='DBF4 homolog pseudogene 1|hCG2024596', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64168953, end_position_on_the_genomic_accession=64169690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371613, CurrentID=0, Status='live', Symbol='LOC105371613', Aliases='', description='uncharacterized LOC105371613', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171655658, end_position_on_the_genomic_accession=171663109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371614, CurrentID=0, Status='live', Symbol='LOC105371614', Aliases='', description='uncharacterized LOC105371614', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171773106, end_position_on_the_genomic_accession=171779610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200504, CurrentID=0, Status='live', Symbol='GKN2', Aliases='BRICD1B, GDDR, PRO813, TFIZ1, VLTI465', description='gastrokine 2', other_designations='BRICHOS domain containing 1B|blottin|down-regulated in gastric cancer GDDR|trefoil factor interactions(z) 1', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68945232, end_position_on_the_genomic_accession=68952970, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246725, CurrentID=0, Status='live', Symbol='LOC246725', Aliases='', description='DNA directed RNA polymerase II polypeptide J-related gene', other_designations='', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371617, CurrentID=0, Status='live', Symbol='LOC105371617', Aliases='', description='uncharacterized LOC105371617', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172679677, end_position_on_the_genomic_accession=172789878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419565, CurrentID=0, Status='live', Symbol='LOC100419565', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47630765, end_position_on_the_genomic_accession=47631438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200523, CurrentID=0, Status='live', Symbol='TEX37', Aliases='C2orf51, TSC21', description='testis expressed 37', other_designations='Testis-Specific Conserved gene 21kDa|protein TSC21|testis-specific conserved protein of 21 kDa', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88524317, end_position_on_the_genomic_accession=88529585, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100133137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133137, CurrentID=0, Status='live', Symbol='LOC100133137', Aliases='', description='geminin, DNA replication inhibitor pseudogene 2', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18170368, end_position_on_the_genomic_accession=18170982, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128738, CurrentID=0, Status='live', Symbol='LOC100128738', Aliases='', description='leucine rich repeat (in FLII) interacting protein 2 pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73237650, end_position_on_the_genomic_accession=73238709, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128739, CurrentID=0, Status='live', Symbol='LOC100128739', Aliases='', description='ubiquitin domain containing 2 pseudogene', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143627172, end_position_on_the_genomic_accession=143627692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200539, CurrentID=0, Status='live', Symbol='ANKRD23', Aliases='DARP, MARP3', description='ankyrin repeat domain 23', other_designations='diabetes related ankyrin repeat protein|diabetes-related ankyrin repeat protein|muscle ankyrin repeat protein 3', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96837914, end_position_on_the_genomic_accession=96844021, orientation='minus', exon_count=9, OMIM=610736, no_hdr0=''),
  105371622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371622, CurrentID=0, Status='live', Symbol='LOC105371622', Aliases='', description='uncharacterized LOC105371622', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174992545, end_position_on_the_genomic_accession=174999352, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  200558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200558, CurrentID=0, Status='live', Symbol='APLF', Aliases='APFL, C2orf13, PALF, Xip1, ZCCHH1', description='aprataxin and PNKP like factor', other_designations='PNK and APTX-like FHA domain-containing protein|PNK and APTX-like FHA protein|XRCC1-interacting protein 1|aprataxin- and PNK-like factor|apurinic-apyrimidinic endonuclease APLF|zinc finger, CX5CX6HX5H motif containing 1', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68467559, end_position_on_the_genomic_accession=68580162, orientation='plus', exon_count=10, OMIM=611035, no_hdr0=''),
  105371625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371625, CurrentID=0, Status='live', Symbol='LOC105371625', Aliases='', description='uncharacterized LOC105371625', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177243736, end_position_on_the_genomic_accession=177255930, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421197, CurrentID=0, Status='live', Symbol='LOC100421197', Aliases='', description='zinc finger and BTB domain containing 10 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101358932, end_position_on_the_genomic_accession=101360372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200575, CurrentID=0, Status='live', Symbol='CRYGEP', Aliases='CCL, CRYG51, D2S1472, G2, CRYGEP', description='crystallin gamma E, pseudogene', other_designations='crystallin, gamma D pseudogene|crystallin, gamma E pseudogene 1|crystallin, gamma E, pseudogene', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208108164, end_position_on_the_genomic_accession=208112717, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200576, CurrentID=0, Status='live', Symbol='PIKFYVE', Aliases='CFD, FAB1, HEL37, PIP5K, PIP5K3, ZFYVE29', description='phosphoinositide kinase, FYVE finger containing', other_designations='1-phosphatidylinositol-3-phosphate 5-kinase|PIPkin-III|epididymis luminal protein 37|phosphatidylinositol 3-phosphate 5-kinase type III|phosphatidylinositol-3-phosphate/phosphatidylinositol 5-kinase, type III|type III PIP kinase|zinc finger, FYVE domain containing 29', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208265753, end_position_on_the_genomic_accession=208358751, orientation='plus', exon_count=44, OMIM=609414, no_hdr0=''),
  105371628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371628, CurrentID=0, Status='live', Symbol='LOC105371628', Aliases='', description='uncharacterized LOC105371628', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178058472, end_position_on_the_genomic_accession=178076755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926957, CurrentID=0, Status='live', Symbol='ISX-AS1', Aliases='', description='ISX antisense RNA 1', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34922609, end_position_on_the_genomic_accession=35072889, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105371629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371629, CurrentID=0, Status='live', Symbol='LOC105371629', Aliases='', description='uncharacterized LOC105371629', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178323658, end_position_on_the_genomic_accession=178341380, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371630, CurrentID=0, Status='live', Symbol='LOC105371630', Aliases='', description='uncharacterized LOC105371630', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178125463, end_position_on_the_genomic_accession=178135525, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371631, CurrentID=0, Status='live', Symbol='LOC105371631', Aliases='', description='uncharacterized LOC105371631', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178526180, end_position_on_the_genomic_accession=178532732, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371632, CurrentID=0, Status='live', Symbol='LOC105371632', Aliases='', description='uncharacterized LOC105371632', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178554390, end_position_on_the_genomic_accession=178580176, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374929, CurrentID=0, Status='live', Symbol='LOC105374929', Aliases='', description='uncharacterized LOC105374929', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11417612, end_position_on_the_genomic_accession=11429540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371635, CurrentID=0, Status='live', Symbol='LOC105371635', Aliases='', description='uncharacterized LOC105371635', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179690334, end_position_on_the_genomic_accession=179730060, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645108, CurrentID=0, Status='live', Symbol='NACAP5', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 5', other_designations='nascent polypeptide-associated complex alpha subunit pseudogene|nascent-polypeptide-associated complex alpha polypeptide pseudogene 5', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17552004, end_position_on_the_genomic_accession=17552803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200634, CurrentID=0, Status='live', Symbol='KRTCAP3', Aliases='KCP3, MRV222, PRO9898', description='keratinocyte associated protein 3', other_designations='KCP-3|keratinocytes associated protein 3', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27442351, end_position_on_the_genomic_accession=27446481, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100128759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128759, CurrentID=0, Status='live', Symbol='LOC100128759', Aliases='', description='heme binding protein 2 pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156033249, end_position_on_the_genomic_accession=156034070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646408, CurrentID=0, Status='live', Symbol='LOC646408', Aliases='', description='chromosome 9 open reading frame 16 pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28835614, end_position_on_the_genomic_accession=28836214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371640, CurrentID=0, Status='live', Symbol='LOC105371640', Aliases='', description='uncharacterized LOC105371640', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182002837, end_position_on_the_genomic_accession=182004330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371641, CurrentID=0, Status='live', Symbol='LOC105371641', Aliases='', description='uncharacterized LOC105371641', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182064295, end_position_on_the_genomic_accession=182244444, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100128762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128762, CurrentID=0, Status='live', Symbol='LOC100128762', Aliases='', description='telomeric repeat binding factor 2, interacting protein pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180967189, end_position_on_the_genomic_accession=180970807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371644, CurrentID=0, Status='live', Symbol='LOC105371644', Aliases='', description='uncharacterized LOC105371644', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182712676, end_position_on_the_genomic_accession=182714544, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371645, CurrentID=0, Status='live', Symbol='LOC105371645', Aliases='', description='uncharacterized LOC105371645', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182792371, end_position_on_the_genomic_accession=182794349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371646, CurrentID=0, Status='live', Symbol='LOC105371646', Aliases='', description='uncharacterized LOC105371646', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183433890, end_position_on_the_genomic_accession=183462082, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421201, CurrentID=0, Status='live', Symbol='LOC100421201', Aliases='', description='piggyBac transposable element derived 3 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126510906, end_position_on_the_genomic_accession=126511711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371648, CurrentID=0, Status='live', Symbol='LOC105371648', Aliases='', description='uncharacterized LOC105371648', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184663181, end_position_on_the_genomic_accession=184672035, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371649, CurrentID=0, Status='live', Symbol='LOC105371649', Aliases='', description='uncharacterized LOC105371649', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184754973, end_position_on_the_genomic_accession=184776029, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128770, CurrentID=0, Status='live', Symbol='LOC100128770', Aliases='', description='uncharacterized LOC100128770', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3032481, end_position_on_the_genomic_accession=3039133, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106479156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479156, CurrentID=0, Status='live', Symbol='RN7SKP137', Aliases='', description='RNA, 7SK small nuclear pseudogene 137', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371651, CurrentID=0, Status='live', Symbol='LOC105371651', Aliases='', description='uncharacterized LOC105371651', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185267686, end_position_on_the_genomic_accession=185283809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421202, CurrentID=0, Status='live', Symbol='LOC100421202', Aliases='', description='C-type lectin domain family 1, member B pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10040847, end_position_on_the_genomic_accession=10045626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200726, CurrentID=0, Status='live', Symbol='LOC200726', Aliases='', description='hCG1657980', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206642418, end_position_on_the_genomic_accession=206649449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200728, CurrentID=0, Status='live', Symbol='TMEM17', Aliases='', description='transmembrane protein 17', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62447595, end_position_on_the_genomic_accession=62506469, orientation='minus', exon_count=8, OMIM=614950, no_hdr0=''),
  100128772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128772, CurrentID=0, Status='live', Symbol='LOC100128772', Aliases='', description='proline-rich protein 3 pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13460291, end_position_on_the_genomic_accession=13460855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200734, CurrentID=0, Status='live', Symbol='SPRED2', Aliases='Spred-2', description='sprouty-related, EVH1 domain containing 2', other_designations='sprouty protein with EVH-1 domain 2, related sequence', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65307435, end_position_on_the_genomic_accession=65432637, orientation='minus', exon_count=9, OMIM=609292, no_hdr0=''),
  100653061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653061, CurrentID=0, Status='live', Symbol='LOC100653061', Aliases='', description='beta-glucuronidase-like protein SMA4', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374933, CurrentID=0, Status='live', Symbol='LOC105374933', Aliases='', description='uncharacterized LOC105374933', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13223049, end_position_on_the_genomic_accession=13230030, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371654, CurrentID=0, Status='live', Symbol='LOC105371654', Aliases='', description='uncharacterized LOC105371654', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186182362, end_position_on_the_genomic_accession=186222715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128775, CurrentID=0, Status='live', Symbol='LOC100128775', Aliases='', description='neural precursor cell expressed, developmentally down-regulated 8 pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84244251, end_position_on_the_genomic_accession=84244847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128776, CurrentID=0, Status='live', Symbol='LOC100128776', Aliases='', description='NSE4 homolog A, SMC5-SMC6 complex component pseudogene', other_designations='NSE4 homolog, SMC5-SMC6 complex component A pseudogene|non-SMC element 4 homolog A (S. cerevisiae) pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14053906, end_position_on_the_genomic_accession=14054945, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128777, CurrentID=0, Status='live', Symbol='KLF8P1', Aliases='', description='Kruppel-like factor 8 pseudogene 1', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48717925, end_position_on_the_genomic_accession=48732399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372392, CurrentID=0, Status='live', Symbol='LOC105372392', Aliases='', description='uncharacterized LOC105372392', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37328188, end_position_on_the_genomic_accession=37329717, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133145, CurrentID=0, Status='live', Symbol='LOC100133145', Aliases='', description='uncharacterized LOC100133145', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=74900889, end_position_on_the_genomic_accession=74901646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  200765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200765, CurrentID=0, Status='live', Symbol='TIGD1', Aliases='EEYORE', description='tigger transposable element derived 1', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232547970, end_position_on_the_genomic_accession=232550590, orientation='minus', exon_count=1, OMIM=612972, no_hdr0=''),
  200772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200772, CurrentID=0, Status='live', Symbol='LOC200772', Aliases='', description='uncharacterized LOC200772', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240954617, end_position_on_the_genomic_accession=240967451, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371661, CurrentID=0, Status='live', Symbol='LOC105371661', Aliases='', description='uncharacterized LOC105371661', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192106283, end_position_on_the_genomic_accession=192139550, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371662, CurrentID=0, Status='live', Symbol='LOC105371662', Aliases='', description='uncharacterized LOC105371662', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192235841, end_position_on_the_genomic_accession=192239786, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645135, CurrentID=0, Status='live', Symbol='LOC645135', Aliases='', description='zinc finger protein 532 pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110007657, end_position_on_the_genomic_accession=110010782, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371664, CurrentID=0, Status='live', Symbol='LOC105371664', Aliases='', description='uncharacterized LOC105371664', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192491751, end_position_on_the_genomic_accession=192629012, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371665, CurrentID=0, Status='live', Symbol='LOC105371665', Aliases='', description='uncharacterized LOC105371665', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192763829, end_position_on_the_genomic_accession=192792523, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106479159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479159, CurrentID=0, Status='live', Symbol='RN7SKP144', Aliases='', description='RNA, 7SK small nuclear pseudogene 144', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  200810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200810, CurrentID=0, Status='live', Symbol='ALG1L', Aliases='ALG1L1', description='ALG1, chitobiosyldiphosphodolichol beta-mannosyltransferase-like', other_designations='asparagine-linked glycosylation 1-like 1|beta-1,4-mannosyltransferase-like', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125928840, end_position_on_the_genomic_accession=125952098, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105371667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371667, CurrentID=0, Status='live', Symbol='LOC105371667', Aliases='', description='uncharacterized LOC105371667', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192836182, end_position_on_the_genomic_accession=192853029, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371668, CurrentID=0, Status='live', Symbol='LOC105371668', Aliases='', description='uncharacterized LOC105371668', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192831863, end_position_on_the_genomic_accession=192864092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200830, CurrentID=0, Status='live', Symbol='LOC200830', Aliases='', description='uncharacterized LOC200830', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128790, CurrentID=0, Status='live', Symbol='LYPLA2P3', Aliases='', description='lysophospholipase II pseudogene 3', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76970320, end_position_on_the_genomic_accession=76971225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371671, CurrentID=0, Status='live', Symbol='LOC105371671', Aliases='', description='uncharacterized LOC105371671', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=195574400, end_position_on_the_genomic_accession=195721899, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  200844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200844, CurrentID=0, Status='live', Symbol='C3orf67', Aliases='', description='chromosome 3 open reading frame 67', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58734472, end_position_on_the_genomic_accession=59050078, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  200845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200845, CurrentID=0, Status='live', Symbol='KCTD6', Aliases='KCASH3', description='potassium channel tetramerization domain containing 6', other_designations='potassium channel tetramerisation domain containing 6', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58492096, end_position_on_the_genomic_accession=58502360, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128792, CurrentID=0, Status='live', Symbol='LOC100128792', Aliases='', description='uncharacterized LOC100128792', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=118328774, end_position_on_the_genomic_accession=118329363, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371673, CurrentID=0, Status='live', Symbol='LOC105371673', Aliases='', description='uncharacterized LOC105371673', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196044962, end_position_on_the_genomic_accession=196086021, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374937, CurrentID=0, Status='live', Symbol='LOC105374937', Aliases='', description='uncharacterized LOC105374937', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13572072, end_position_on_the_genomic_accession=13574137, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  200860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200860, CurrentID=0, Status='live', Symbol='DKFZp547K1610', Aliases='', description='uncharacterized protein DKFZp547K1610', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371675, CurrentID=0, Status='live', Symbol='LOC105371675', Aliases='', description='uncharacterized LOC105371675', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196912092, end_position_on_the_genomic_accession=196933716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379302, CurrentID=0, Status='live', Symbol='LOC105379302', Aliases='', description='uncharacterized LOC105379302', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19003942, end_position_on_the_genomic_accession=19015343, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421207, CurrentID=0, Status='live', Symbol='LOC100421207', Aliases='', description='Fanconi anemia complementation group L pseudogene', other_designations='Fanconi anemia, complementation group L pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75100462, end_position_on_the_genomic_accession=75101132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631239, CurrentID=0, Status='live', Symbol='KRT18P67', Aliases='', description='keratin 18 pseudogene 67', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122837552, end_position_on_the_genomic_accession=122838891, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200879, CurrentID=0, Status='live', Symbol='LIPH', Aliases='AH, ARWH2, HYPT7, LAH2, LPDLR, PLA1B, mPA-PLA1', description='lipase, member H', other_designations='LPD lipase-related protein|lipase H|mPA-PLA1 alpha|membrane-associated phosphatidic acid-selective phospholipase A1-alpha|membrane-bound phosphatidic acid-selective phospholipase A1|phospholipase A1 member B', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185506866, end_position_on_the_genomic_accession=185552655, orientation='minus', exon_count=12, OMIM=607365, no_hdr0=''),
  105371679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371679, CurrentID=0, Status='live', Symbol='LOC105371679', Aliases='', description='uncharacterized LOC105371679', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198976178, end_position_on_the_genomic_accession=198983925, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100631242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631242, CurrentID=0, Status='live', Symbol='LOC100631242', Aliases='', description='zinc finger protein 852 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51856898, end_position_on_the_genomic_accession=51858089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200894, CurrentID=0, Status='live', Symbol='ARL13B', Aliases='ARL2L1, JBTS8', description='ADP ribosylation factor like GTPase 13B', other_designations='ADP-ribosylation factor like GTPase 13B|ADP-ribosylation factor-like 13B|ADP-ribosylation factor-like 2-like 1|ARL2-like protein 1', map_location='3q11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=93980134, end_position_on_the_genomic_accession=94055678, orientation='plus', exon_count=14, OMIM=608922, no_hdr0=''),
  200895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200895, CurrentID=0, Status='live', Symbol='DHFRL1', Aliases='DHFRP4', description='dihydrofolate reductase like 1', other_designations='dihydrofolate reductase pseudogene 4|dihydrofolate reductase-like 1|dihydrofolate reductase-like protein 1', map_location='3q11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94057922, end_position_on_the_genomic_accession=94063389, orientation='minus', exon_count=3, OMIM=616588, no_hdr0=''),
  100128800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128800, CurrentID=0, Status='live', Symbol='LOC100128800', Aliases='', description='cell adhesion associated, oncogene regulated pseudogene', other_designations='Cdon homolog pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89944100, end_position_on_the_genomic_accession=89947320, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631243, CurrentID=0, Status='live', Symbol='LOC100631243', Aliases='', description='zinc finger protein 211 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=103527626, end_position_on_the_genomic_accession=103529272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128801, CurrentID=0, Status='live', Symbol='DRAXINP1', Aliases='', description='dorsal inhibitory axon guidance protein pseudogene 1', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8783514, end_position_on_the_genomic_accession=8784439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  200909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200909, CurrentID=0, Status='live', Symbol='HTR3D', Aliases='5HT3D', description='5-hydroxytryptamine (serotonin) receptor 3D, ionotropic', other_designations='5-hydroxytryptamine (serotonin) receptor 3 family member D|5-hydroxytryptamine receptor 3 subunit D|serotonin 5-HT-3D receptor|serotonin receptor 3D', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184031544, end_position_on_the_genomic_accession=184039369, orientation='plus', exon_count=10, OMIM=610122, no_hdr0=''),
  100128803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128803, CurrentID=0, Status='live', Symbol='LOC100128803', Aliases='', description='Spi-C transcription factor (Spi-1/PU.1 related) pseudogene', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58175, end_position_on_the_genomic_accession=59069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  200916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200916, CurrentID=0, Status='live', Symbol='RPL22L1', Aliases='', description='ribosomal protein L22 like 1', other_designations='ribosomal protein L22-like 1', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170864875, end_position_on_the_genomic_accession=170870256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101926968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926968, CurrentID=0, Status='live', Symbol='LOC101926968', Aliases='', description='uncharacterized LOC101926968', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118641768, end_position_on_the_genomic_accession=118644157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371684, CurrentID=0, Status='live', Symbol='LOC105371684', Aliases='', description='uncharacterized LOC105371684', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202087727, end_position_on_the_genomic_accession=202119949, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371685, CurrentID=0, Status='live', Symbol='LOC105371685', Aliases='', description='uncharacterized LOC105371685', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202602017, end_position_on_the_genomic_accession=202604010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  200931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200931, CurrentID=0, Status='live', Symbol='SLC51A', Aliases='OSTA, OSTalpha', description='solute carrier family 51 alpha subunit', other_designations='OST-alpha|organic solute transporter alpha|organic solute transporter, alpha subunit|solute carrier family 51 subunit alpha|solute carrier family 51, alpha subunit', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196216512, end_position_on_the_genomic_accession=196233430, orientation='plus', exon_count=9, OMIM=612084, no_hdr0=''),
  200933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200933, CurrentID=0, Status='live', Symbol='FBXO45', Aliases='Fbx45', description='F-box protein 45', other_designations='F-box only protein 45|hFbxo45', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196568854, end_position_on_the_genomic_accession=196589059, orientation='plus', exon_count=3, OMIM=609112, no_hdr0=''),
  105371687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371687, CurrentID=0, Status='live', Symbol='LOC105371687', Aliases='', description='uncharacterized LOC105371687', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203067327, end_position_on_the_genomic_accession=203067688, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  200942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200942, CurrentID=0, Status='live', Symbol='KLHDC8B', Aliases='CHL', description='kelch domain containing 8B', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49171120, end_position_on_the_genomic_accession=49176486, orientation='plus', exon_count=6, OMIM=613169, no_hdr0=''),
  120872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120872, CurrentID=0, Status='live', Symbol='RPL7P38', Aliases='RPL7_24_1261', description='ribosomal protein L7 pseudogene 38', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80028877, end_position_on_the_genomic_accession=80029700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631251, CurrentID=0, Status='live', Symbol='GRAMD4P7', Aliases='', description='GRAM domain containing 4 pseudogene 7', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14485804, end_position_on_the_genomic_accession=14487494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371689, CurrentID=0, Status='live', Symbol='LOC105371689', Aliases='', description='uncharacterized LOC105371689', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203996805, end_position_on_the_genomic_accession=204008534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  200958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200958, CurrentID=0, Status='live', Symbol='MUC20', Aliases='MUC-20', description='mucin 20, cell surface associated', other_designations='transmembrane mucin MUC20S', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195720882, end_position_on_the_genomic_accession=195733553, orientation='plus', exon_count=4, OMIM=610360, no_hdr0=''),
  200959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=200959, CurrentID=0, Status='live', Symbol='GABRR3', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, rho 3 (gene/pseudogene)', other_designations='GABA(A) receptor subunit rho-3|GABA(C) receptor', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97986683, end_position_on_the_genomic_accession=98035304, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105371691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371691, CurrentID=0, Status='live', Symbol='LOC105371691', Aliases='', description='uncharacterized LOC105371691', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204659290, end_position_on_the_genomic_accession=204678290, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375885, CurrentID=0, Status='live', Symbol='LOC105375885', Aliases='', description='uncharacterized LOC105375885', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67028578, end_position_on_the_genomic_accession=67037616, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371694, CurrentID=0, Status='live', Symbol='LOC105371694', Aliases='', description='uncharacterized LOC105371694', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204726229, end_position_on_the_genomic_accession=204733988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  574080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574080, CurrentID=0, Status='live', Symbol='LOC574080', Aliases='', description='CDGSH iron sulfur domain 1 pseudogene', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=165782343, end_position_on_the_genomic_accession=165783271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371696, CurrentID=0, Status='live', Symbol='LOC105371696', Aliases='', description='uncharacterized LOC105371696', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205282417, end_position_on_the_genomic_accession=205298122, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371697, CurrentID=0, Status='live', Symbol='LOC105371697', Aliases='', description='uncharacterized LOC105371697', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205441393, end_position_on_the_genomic_accession=205452029, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100631260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631260, CurrentID=0, Status='live', Symbol='MARK2P13', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 13', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36109024, end_position_on_the_genomic_accession=36110872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371698, CurrentID=0, Status='live', Symbol='LOC105371698', Aliases='', description='uncharacterized LOC105371698', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205376982, end_position_on_the_genomic_accession=205378597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926971, CurrentID=0, Status='live', Symbol='LINC01284', Aliases='', description='long intergenic non-protein coding RNA 1284', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51095836, end_position_on_the_genomic_accession=51117830, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100533918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533918, CurrentID=0, Status='live', Symbol='LOC100533918', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371701, CurrentID=0, Status='live', Symbol='LOC105371701', Aliases='', description='uncharacterized LOC105371701', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205680415, end_position_on_the_genomic_accession=205690644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128822, CurrentID=0, Status='live', Symbol='LINC01003', Aliases='', description='long intergenic non-protein coding RNA 1003', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152464124, end_position_on_the_genomic_accession=152465545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423060, CurrentID=0, Status='live', Symbol='ASNSP6', Aliases='', description='asparagine synthetase (glutamine-hydrolyzing) pseudogene 6', other_designations='', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11745669, end_position_on_the_genomic_accession=11747521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371703, CurrentID=0, Status='live', Symbol='LOC105371703', Aliases='', description='uncharacterized LOC105371703', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27072388, end_position_on_the_genomic_accession=27089240, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128824, CurrentID=0, Status='live', Symbol='LOC100128824', Aliases='', description='uncharacterized LOC100128824', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=171185909, end_position_on_the_genomic_accession=171186476, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645177, CurrentID=0, Status='live', Symbol='LOC645177', Aliases='', description='lymphoid-restricted membrane protein-like', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24997562, end_position_on_the_genomic_accession=25066516, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105371707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371707, CurrentID=0, Status='live', Symbol='LOC105371707', Aliases='', description='uncharacterized LOC105371707', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27649394, end_position_on_the_genomic_accession=27650394, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371708, CurrentID=0, Status='live', Symbol='LOC105371708', Aliases='', description='uncharacterized LOC105371708', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27930393, end_position_on_the_genomic_accession=27991694, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480975, CurrentID=0, Status='live', Symbol='RN7SL197P', Aliases='', description='RNA, 7SL, cytoplasmic 197, pseudogene', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101926973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926973, CurrentID=0, Status='live', Symbol='LOC101926973', Aliases='', description='uncharacterized LOC101926973', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107851278, end_position_on_the_genomic_accession=107939909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645181, CurrentID=0, Status='live', Symbol='LOC645181', Aliases='', description='PDGFA associated protein 1 pseudogene', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85549410, end_position_on_the_genomic_accession=85549953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371710, CurrentID=0, Status='live', Symbol='LOC105371710', Aliases='', description='uncharacterized LOC105371710', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28204999, end_position_on_the_genomic_accession=28210272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371711, CurrentID=0, Status='live', Symbol='LOC105371711', Aliases='', description='uncharacterized LOC105371711', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28132682, end_position_on_the_genomic_accession=28144415, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128833, CurrentID=0, Status='live', Symbol='LOC100128833', Aliases='', description='vacuolar protein sorting 37 homolog A pseudogene', other_designations='vacuolar protein sorting 37 homolog A (S. cerevisiae) pseudogene', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154493278, end_position_on_the_genomic_accession=154494751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926974, CurrentID=0, Status='live', Symbol='LOC101926974', Aliases='', description='uncharacterized LOC101926974', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46246437, end_position_on_the_genomic_accession=46291290, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371714, CurrentID=0, Status='live', Symbol='LOC105371714', Aliases='', description='uncharacterized LOC105371714', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28770821, end_position_on_the_genomic_accession=28783525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371716, CurrentID=0, Status='live', Symbol='LOC105371716', Aliases='', description='uncharacterized LOC105371716', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29012787, end_position_on_the_genomic_accession=29016512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371719, CurrentID=0, Status='live', Symbol='LOC105371719', Aliases='', description='uncharacterized LOC105371719', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30090366, end_position_on_the_genomic_accession=30118073, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  201134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201134, CurrentID=0, Status='live', Symbol='CEP112', Aliases='CCDC46, MACOCO', description='centrosomal protein 112kDa', other_designations='coiled-coil domain-containing protein 46', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65635540, end_position_on_the_genomic_accession=66192131, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  100128840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128840, CurrentID=0, Status='live', Symbol='LOC100128840', Aliases='', description='uncharacterized LOC100128840', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=131472857, end_position_on_the_genomic_accession=131474573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  201140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201140, CurrentID=0, Status='live', Symbol='DHRS7C', Aliases='SDR32C2', description='dehydrogenase/reductase (SDR family) member 7C', other_designations='short chain dehydrogenase/reductase family 32C, member 2|short-chain dehydrogenase/reductase family 32C member 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9771434, end_position_on_the_genomic_accession=9791577, orientation='minus', exon_count=7, OMIM=616161, no_hdr0=''),
  100128843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128843, CurrentID=0, Status='live', Symbol='LOC100128843', Aliases='', description='uncharacterized LOC100128843', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=133413289, end_position_on_the_genomic_accession=133414204, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369922, CurrentID=0, Status='live', Symbol='LOC105369922', Aliases='', description='uncharacterized LOC105369922', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95858237, end_position_on_the_genomic_accession=95858830, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  201158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201158, CurrentID=0, Status='live', Symbol='TVP23C', Aliases='FAM18B2', description='trans-golgi network vesicle protein 23 homolog C (S. cerevisiae)', other_designations='family with sequence similarity 18, member B2|protein FAM18B2', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15502264, end_position_on_the_genomic_accession=15563631, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  201161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201161, CurrentID=0, Status='live', Symbol='CENPV', Aliases='3110013H01Rik, CENP-V, PRR6, p30', description='centromere protein V', other_designations='nuclear protein p30|proline rich 6|proline-rich protein 6', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16342534, end_position_on_the_genomic_accession=16353498, orientation='minus', exon_count=5, OMIM=608139, no_hdr0=''),
  201163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201163, CurrentID=0, Status='live', Symbol='FLCN', Aliases='BHD, FLCL', description='folliculin', other_designations='BHD skin lesion fibrofolliculoma protein|birt-Hogg-Dube syndrome protein', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17206924, end_position_on_the_genomic_accession=17237188, orientation='minus', exon_count=17, OMIM=607273, no_hdr0=''),
  201164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201164, CurrentID=0, Status='live', Symbol='PLD6', Aliases='ZUC', description='phospholipase D family member 6', other_designations='PLD 6|choline phosphatase 6|mitoPLD|mitochondrial phospholipase|phosphatidylcholine-hydrolyzing phospholipase D6|phospholipase D family, member 6|phospholipase D6|protein zucchini homolog', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17200995, end_position_on_the_genomic_accession=17206332, orientation='minus', exon_count=2, OMIM=614960, no_hdr0=''),
  100128845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128845, CurrentID=0, Status='live', Symbol='CDC42EP3P1', Aliases='', description='CDC42 effector protein (Rho GTPase binding) 3 pseudogene 1', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23384530, end_position_on_the_genomic_accession=23385290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128407, CurrentID=0, Status='live', Symbol='LOC100128407', Aliases='', description='N-acetylneuraminic acid phosphatase pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119009814, end_position_on_the_genomic_accession=119012756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371726, CurrentID=0, Status='live', Symbol='LOC105371726', Aliases='', description='uncharacterized LOC105371726', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31599955, end_position_on_the_genomic_accession=31606408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  201176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201176, CurrentID=0, Status='live', Symbol='ARHGAP27', Aliases='CAMGAP1, PP905, SH3D20, SH3P20', description='Rho GTPase activating protein 27', other_designations='CIN85-associated multi-domain-containing Rho GTPase-activating protein 1|SH3 domain containing 20|SH3 domain-containing protein 20|rho-type GTPase-activating protein 27', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45393902, end_position_on_the_genomic_accession=45432916, orientation='minus', exon_count=23, OMIM=610591, no_hdr0=''),
  201181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201181, CurrentID=0, Status='live', Symbol='ZNF385C', Aliases='', description='zinc finger protein 385C', other_designations='CTD-2132N18.2|CTD-2132N18.4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42024876, end_position_on_the_genomic_accession=42072630, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  101926977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926977, CurrentID=0, Status='live', Symbol='LOC101926977', Aliases='', description='uncharacterized LOC101926977', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94553622, end_position_on_the_genomic_accession=94594761, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  201191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201191, CurrentID=0, Status='live', Symbol='SAMD14', Aliases='', description='sterile alpha motif domain containing 14', other_designations='SAM domain-containing protein 14', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50111309, end_position_on_the_genomic_accession=50129882, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105371730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371730, CurrentID=0, Status='live', Symbol='LOC105371730', Aliases='', description='uncharacterized LOC105371730', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32141226, end_position_on_the_genomic_accession=32143117, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128795, CurrentID=0, Status='live', Symbol='LOC100128795', Aliases='', description='coiled-coil domain-containing protein 144A-like', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371731, CurrentID=0, Status='live', Symbol='LOC105371731', Aliases='', description='uncharacterized LOC105371731', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32258045, end_position_on_the_genomic_accession=32265451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371732, CurrentID=0, Status='live', Symbol='LOC105371732', Aliases='', description='uncharacterized LOC105371732', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32280387, end_position_on_the_genomic_accession=32281034, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372395, CurrentID=0, Status='live', Symbol='LOC105372395', Aliases='', description='uncharacterized LOC105372395', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37855777, end_position_on_the_genomic_accession=37896673, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371733, CurrentID=0, Status='live', Symbol='LOC105371733', Aliases='', description='uncharacterized LOC105371733', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32519631, end_position_on_the_genomic_accession=32531468, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926978, CurrentID=0, Status='live', Symbol='LINC01111', Aliases='', description='long intergenic non-protein coding RNA 1111', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76406654, end_position_on_the_genomic_accession=76524356, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371734, CurrentID=0, Status='live', Symbol='LOC105371734', Aliases='', description='uncharacterized LOC105371734', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32509954, end_position_on_the_genomic_accession=32523674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371735, CurrentID=0, Status='live', Symbol='LOC105371735', Aliases='', description='uncharacterized LOC105371735', other_designations='CTC-304I17.5', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34156634, end_position_on_the_genomic_accession=34164238, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  201229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201229, CurrentID=0, Status='live', Symbol='LYRM9', Aliases='C17orf108, HSD24', description='LYR motif containing 9', other_designations='UPF0631 protein C17orf108', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27878314, end_position_on_the_genomic_accession=27893383, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106479173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479173, CurrentID=0, Status='live', Symbol='RN7SKP175', Aliases='', description='RNA, 7SK small nuclear pseudogene 175', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  201232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201232, CurrentID=0, Status='live', Symbol='SLC16A13', Aliases='MCT13', description='solute carrier family 16 member 13', other_designations='MCT 13|monocarboxylic acid transporter 13|solute carrier family 16 (monocarboxylic acid transporters), member 13|solute carrier family 16, member 13 (monocarboxylic acid transporter 13)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7036075, end_position_on_the_genomic_accession=7040121, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100631299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631299, CurrentID=0, Status='live', Symbol='LOC100631299', Aliases='', description='perilipin 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371737, CurrentID=0, Status='live', Symbol='LOC105371737', Aliases='', description='uncharacterized LOC105371737', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=33111917, end_position_on_the_genomic_accession=33155533, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100631300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631300, CurrentID=0, Status='live', Symbol='LOC100631300', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  201243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201243, CurrentID=0, Status='live', Symbol='C17orf74', Aliases='', description='chromosome 17 open reading frame 74', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7424881, end_position_on_the_genomic_accession=7427568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371738, CurrentID=0, Status='live', Symbol='LOC105371738', Aliases='', description='uncharacterized LOC105371738', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34473797, end_position_on_the_genomic_accession=34477274, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100631301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631301, CurrentID=0, Status='live', Symbol='LOC100631301', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371739, CurrentID=0, Status='live', Symbol='LOC105371739', Aliases='', description='uncharacterized LOC105371739', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34479269, end_position_on_the_genomic_accession=34483012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  201254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201254, CurrentID=0, Status='live', Symbol='STRA13', Aliases='CENP-X, CENPX, D9, FAAP10, MHF2', description='stimulated by retinoic acid 13', other_designations='FANCM-interacting histone fold protein 2|Fanconi anemia-associated polypeptide of 10 kDa|retinoic acid-inducible gene D9 protein homolog|stimulated by retinoic acid 13 homolog|stimulated by retinoic acid gene 13 protein homolog', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82018703, end_position_on_the_genomic_accession=82022930, orientation='minus', exon_count=5, OMIM=615128, no_hdr0=''),
  201255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201255, CurrentID=0, Status='live', Symbol='LRRC45', Aliases='', description='leucine rich repeat containing 45', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82023250, end_position_on_the_genomic_accession=82031151, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105371740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371740, CurrentID=0, Status='live', Symbol='LOC105371740', Aliases='', description='uncharacterized LOC105371740', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34725509, end_position_on_the_genomic_accession=34727062, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  201259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201259, CurrentID=0, Status='live', Symbol='RPL32P31', Aliases='RPL32_21_1577', description='ribosomal protein L32 pseudogene 31', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80542203, end_position_on_the_genomic_accession=80545134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  201266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201266, CurrentID=0, Status='live', Symbol='SLC39A11', Aliases='C17orf26, ZIP11', description='solute carrier family 39 member 11', other_designations='ZIP-11|Zrt- and Irt-like protein 11|solute carrier family 39 (metal ion transporter), member 11|solute carrier family 39, member 11', map_location='17q24.3-q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72645946, end_position_on_the_genomic_accession=73092714, orientation='minus', exon_count=19, OMIM=616508, no_hdr0=''),
  105371743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371743, CurrentID=0, Status='live', Symbol='LOC105371743', Aliases='', description='uncharacterized LOC105371743', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35568089, end_position_on_the_genomic_accession=35574313, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  201283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201283, CurrentID=0, Status='live', Symbol='AMZ2P1', Aliases='', description='archaelysin family metallopeptidase 2 pseudogene 1', other_designations='putative archaemetzincin-2-like protein', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64966550, end_position_on_the_genomic_accession=64975585, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100128865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128865, CurrentID=0, Status='live', Symbol='LOC100128865', Aliases='', description='methyltransferase like 5 pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55219344, end_position_on_the_genomic_accession=55219973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  201288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201288, CurrentID=0, Status='live', Symbol='NOS2P2', Aliases='NOS2B', description='nitric oxide synthase 2 pseudogene 2', other_designations='nitric oxide synthase 2, inducible pseudogene|nitric oxide synthase 2A (inducible, hepatocytes) pseudogene|nitric oxide synthase 2B (pseudogene)', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18497090, end_position_on_the_genomic_accession=18503285, orientation='minus', exon_count=0, OMIM=600719, no_hdr0=''),
  201292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201292, CurrentID=0, Status='live', Symbol='TRIM65', Aliases='4732463G12Rik', description='tripartite motif containing 65', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75880379, end_position_on_the_genomic_accession=75897003, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  201294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201294, CurrentID=0, Status='live', Symbol='UNC13D', Aliases='FHL3, HLH3, HPLH3, Munc13-4', description='unc-13 homolog D (C. elegans)', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75827225, end_position_on_the_genomic_accession=75844717, orientation='minus', exon_count=33, OMIM=608897, no_hdr0=''),
  201299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201299, CurrentID=0, Status='live', Symbol='RDM1', Aliases='RAD52B', description='RAD52 motif containing 1', other_designations='RAD52 homolog B|RAD52 motif 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35918080, end_position_on_the_genomic_accession=35930811, orientation='minus', exon_count=8, OMIM=612896, no_hdr0=''),
  201305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201305, CurrentID=0, Status='live', Symbol='SPNS3', Aliases='', description='spinster homolog 3 (Drosophila)', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4433924, end_position_on_the_genomic_accession=4488204, orientation='plus', exon_count=14, OMIM=611701, no_hdr0=''),
  106660623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660623, CurrentID=0, Status='live', Symbol='CNTNAP3P3', Aliases='', description='contactin associated protein-like 3 pseudogene 3', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631777, CurrentID=0, Status='live', Symbol='RNA45S4', Aliases='', description='RNA, 45S pre-ribosomal 4', other_designations='', map_location='21p12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631778, CurrentID=0, Status='live', Symbol='RNA45S3', Aliases='', description='RNA, 45S pre-ribosomal 3', other_designations='', map_location='15p12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631779, CurrentID=0, Status='live', Symbol='RNA45S2', Aliases='', description='RNA, 45S pre-ribosomal 2', other_designations='', map_location='14p12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631780, CurrentID=0, Status='live', Symbol='RNA45S1', Aliases='', description='RNA, 45S pre-ribosomal 1', other_designations='', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631781, CurrentID=0, Status='live', Symbol='RNA18S4', Aliases='', description='RNA, 18S ribosomal 4', other_designations='', map_location='21p12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106631782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106631782, CurrentID=0, Status='live', Symbol='RNA18S3', Aliases='', description='RNA, 18S ribosomal 3', other_designations='', map_location='15p12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371751, CurrentID=0, Status='live', Symbol='LOC105371751', Aliases='', description='uncharacterized LOC105371751', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36672230, end_position_on_the_genomic_accession=36676488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371752, CurrentID=0, Status='live', Symbol='LOC105371752', Aliases='', description='proline-rich protein 36-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371755, CurrentID=0, Status='live', Symbol='LOC105371755', Aliases='', description='uncharacterized LOC105371755', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631319, CurrentID=0, Status='live', Symbol='LOC100631319', Aliases='', description='zinc finger protein 273 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379175, CurrentID=0, Status='live', Symbol='LOC105379175', Aliases='', description='uncharacterized LOC105379175', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132468863, end_position_on_the_genomic_accession=132472685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133165, CurrentID=0, Status='live', Symbol='LOC100133165', Aliases='', description='Dexi homolog (mouse) pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22714544, end_position_on_the_genomic_accession=22716050, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631321, CurrentID=0, Status='live', Symbol='LOC100631321', Aliases='', description='zinc finger protein 100 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  399512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399512, CurrentID=0, Status='live', Symbol='SLC25A35', Aliases='', description='solute carrier family 25 member 35', other_designations='solute carrier family 25, member 35', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8287764, end_position_on_the_genomic_accession=8295400, orientation='minus', exon_count=7, OMIM=610818, no_hdr0=''),
  105371760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371760, CurrentID=0, Status='live', Symbol='LOC105371760', Aliases='', description='uncharacterized LOC105371760', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38266771, end_position_on_the_genomic_accession=38295603, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100631323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631323, CurrentID=0, Status='live', Symbol='LOC100631323', Aliases='', description='fermitin family member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645233, CurrentID=0, Status='live', Symbol='TDGP1', Aliases='', description='thymine-DNA glycosylase pseudogene 1', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25803193, end_position_on_the_genomic_accession=25806259, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128882, CurrentID=0, Status='live', Symbol='LOC100128882', Aliases='', description='uncharacterized LOC100128882', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88512836, end_position_on_the_genomic_accession=88531995, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101926984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926984, CurrentID=0, Status='live', Symbol='LOC101926984', Aliases='', description='putative uncharacterized protein FLJ43826', other_designations='CTD-2008P7.1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28263634, end_position_on_the_genomic_accession=28275902, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422063, CurrentID=0, Status='live', Symbol='LOC100422063', Aliases='', description='olfactory receptor, family 13, subfamily H, member 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128885, CurrentID=0, Status='live', Symbol='LOC100128885', Aliases='', description='uncharacterized LOC100128885', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64574695, end_position_on_the_genomic_accession=64618589, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379320, CurrentID=0, Status='live', Symbol='LOC105379320', Aliases='', description='uncharacterized LOC105379320', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22564628, end_position_on_the_genomic_accession=22567122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370591, CurrentID=0, Status='live', Symbol='LOC105370591', Aliases='', description='uncharacterized LOC105370591', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80147500, end_position_on_the_genomic_accession=80153683, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371768, CurrentID=0, Status='live', Symbol='LOC105371768', Aliases='', description='uncharacterized LOC105371768', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39175767, end_position_on_the_genomic_accession=39184330, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371769, CurrentID=0, Status='live', Symbol='LOC105371769', Aliases='', description='uncharacterized LOC105371769', other_designations='CTB-131K11.1', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371770, CurrentID=0, Status='live', Symbol='LOC105371770', Aliases='', description='uncharacterized LOC105371770', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39567266, end_position_on_the_genomic_accession=39573134, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106479180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479180, CurrentID=0, Status='live', Symbol='RN7SKP189', Aliases='', description='RNA, 7SK small nuclear pseudogene 189', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371771, CurrentID=0, Status='live', Symbol='LOC105371771', Aliases='', description='uncharacterized LOC105371771', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39594724, end_position_on_the_genomic_accession=39598250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371773, CurrentID=0, Status='live', Symbol='LOC105371773', Aliases='', description='uncharacterized LOC105371773', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40516942, end_position_on_the_genomic_accession=40522565, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  201456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201456, CurrentID=0, Status='live', Symbol='FBXO15', Aliases='FBX15', description='F-box protein 15', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74073343, end_position_on_the_genomic_accession=74147865, orientation='minus', exon_count=13, OMIM=609093, no_hdr0=''),
  105371774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371774, CurrentID=0, Status='live', Symbol='LOC105371774', Aliases='', description='uncharacterized LOC105371774', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40548316, end_position_on_the_genomic_accession=40554664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371775, CurrentID=0, Status='live', Symbol='LOC105371775', Aliases='', description='uncharacterized LOC105371775', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40687265, end_position_on_the_genomic_accession=40697840, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379322, CurrentID=0, Status='live', Symbol='LOC105379322', Aliases='', description='uncharacterized LOC105379322', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22693695, end_position_on_the_genomic_accession=22694275, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  201475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201475, CurrentID=0, Status='live', Symbol='RAB12', Aliases='', description='RAB12, member RAS oncogene family', other_designations='putative Ras-related protein Rab-12', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8609445, end_position_on_the_genomic_accession=8639383, orientation='plus', exon_count=7, OMIM=616448, no_hdr0=''),
  100130642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130642, CurrentID=0, Status='live', Symbol='LOC100130642', Aliases='', description='uncharacterized LOC100130642', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=46727075, end_position_on_the_genomic_accession=46728012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  201477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201477, CurrentID=0, Status='live', Symbol='DLGAP1-AS3', Aliases='', description='DLGAP1 antisense RNA 3', other_designations='DLGAP1 antisense RNA 3 (non-protein coding)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3878180, end_position_on_the_genomic_accession=3897069, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371777, CurrentID=0, Status='live', Symbol='LOC105371777', Aliases='', description='uncharacterized LOC105371777', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40849361, end_position_on_the_genomic_accession=40921604, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128898, CurrentID=0, Status='live', Symbol='LOC100128898', Aliases='', description='latexin pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158078099, end_position_on_the_genomic_accession=158078764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399516, CurrentID=0, Status='live', Symbol='OTSC7', Aliases='', description='otosclerosis 7', other_designations='', map_location='6q13-q16.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611572, no_hdr0=''),
  645251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645251, CurrentID=0, Status='live', Symbol='CBX1P1', Aliases='', description='chromobox 1 pseudogene 1', other_designations='chromobox homolog 1 (HP1 beta homolog Drosophila ) pseudogene|chromobox homolog 1 pseudogene 1', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63297674, end_position_on_the_genomic_accession=63299885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  201501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201501, CurrentID=0, Status='live', Symbol='ZBTB7C', Aliases='APM-1, APM1, ZBTB36, ZNF857C', description='zinc finger and BTB domain containing 7C', other_designations='B230208J24Rik|affected by papillomavirus DNA integration in ME180 cells protein 1|zinc finger and BTB domain containing 36|zinc finger and BTB domain-containing 7C|zinc finger and BTB domain-containing protein 36|zinc finger protein 857C', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48026864, end_position_on_the_genomic_accession=48409422, orientation='minus', exon_count=12, OMIM=616591, no_hdr0=''),
  105371782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371782, CurrentID=0, Status='live', Symbol='LOC105371782', Aliases='', description='uncharacterized LOC105371782', other_designations='CTD-3193K9.4', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  201514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201514, CurrentID=0, Status='live', Symbol='ZNF584', Aliases='', description='zinc finger protein 584', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58407873, end_position_on_the_genomic_accession=58418327, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371783, CurrentID=0, Status='live', Symbol='LOC105371783', Aliases='', description='uncharacterized LOC105371783', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43316254, end_position_on_the_genomic_accession=43317113, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  201516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201516, CurrentID=0, Status='live', Symbol='ZSCAN4', Aliases='ZNF494', description='zinc finger and SCAN domain containing 4', other_designations='zinc finger protein 494', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57651504, end_position_on_the_genomic_accession=57679152, orientation='plus', exon_count=9, OMIM=613419, no_hdr0=''),
  105371784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371784, CurrentID=0, Status='live', Symbol='LOC105371784', Aliases='', description='uncharacterized LOC105371784', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43377835, end_position_on_the_genomic_accession=43381168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128905, CurrentID=0, Status='live', Symbol='LOC100128905', Aliases='', description='uncharacterized LOC100128905', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174024721, end_position_on_the_genomic_accession=174027158, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128906, CurrentID=0, Status='live', Symbol='LOC100128906', Aliases='', description='WDR45-like pseudogene', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104328323, end_position_on_the_genomic_accession=104329396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371788, CurrentID=0, Status='live', Symbol='LOC105371788', Aliases='', description='leucine-rich repeat-containing protein 37B-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43926633, end_position_on_the_genomic_accession=43927412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645261, CurrentID=0, Status='live', Symbol='LOC645261', Aliases='', description='PP565', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87050926, end_position_on_the_genomic_accession=87056013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645262, CurrentID=0, Status='live', Symbol='LOC645262', Aliases='', description='phosphodiesterase 4D interacting protein-like', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=149305909, end_position_on_the_genomic_accession=149314610, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  201562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201562, CurrentID=0, Status='live', Symbol='HACD2', Aliases='PTPLB', description='3-hydroxyacyl-CoA dehydratase 2', other_designations='protein tyrosine phosphatase-like (proline instead of catalytic arginine), member b|protein-tyrosine phosphatase-like member B|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 2', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123494516, end_position_on_the_genomic_accession=123585077, orientation='minus', exon_count=7, OMIM=615939, no_hdr0=''),
  100128911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128911, CurrentID=0, Status='live', Symbol='PCNPP2', Aliases='', description='PEST containing nuclear protein pseudogene 2', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88970809, end_position_on_the_genomic_accession=88971348, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645264, CurrentID=0, Status='live', Symbol='OSTCP4', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 4', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112685055, end_position_on_the_genomic_accession=112685500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419572, CurrentID=0, Status='live', Symbol='LOC100419572', Aliases='', description='uracil DNA glycosylase pseudogene', other_designations='uracil-DNA glycosylase pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87166943, end_position_on_the_genomic_accession=87167731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371794, CurrentID=0, Status='live', Symbol='LOC105371794', Aliases='', description='uncharacterized LOC105371794', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128808, CurrentID=0, Status='live', Symbol='LOC100128808', Aliases='', description='chromosome 11 open reading frame 51 pseudogene', other_designations='anaphase promoting complex subunit 15 pseudogene', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8856466, end_position_on_the_genomic_accession=8857123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  201595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201595, CurrentID=0, Status='live', Symbol='STT3B', Aliases='CDG1X, SIMP, STT3-B', description='STT3B, subunit of the oligosaccharyltransferase complex (catalytic)', other_designations='STT3, subunit of the oligosaccharyltransferase complex, homolog B|dolichyl-diphosphooligosaccharide protein glycotransferase|homolog of yeast STT3|oligosaccharyl transferase subunit STT3B|source of immunodominant MHC-associated peptides homolog', map_location='3p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31532501, end_position_on_the_genomic_accession=31637622, orientation='plus', exon_count=17, OMIM=608605, no_hdr0=''),
  105371797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371797, CurrentID=0, Status='live', Symbol='LOC105371797', Aliases='', description='uncharacterized LOC105371797', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369779, CurrentID=0, Status='live', Symbol='LOC105369779', Aliases='', description='uncharacterized LOC105369779', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55729282, end_position_on_the_genomic_accession=55730862, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  594832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594832, CurrentID=0, Status='live', Symbol='MYP11', Aliases='', description='myopia 11 (high grade, autosomal dominant)', other_designations='', map_location='4q22-q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609994, no_hdr0=''),
  201617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201617, CurrentID=0, Status='live', Symbol='LINC00870', Aliases='', description='long intergenic non-protein coding RNA 870', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72151257, end_position_on_the_genomic_accession=72174339, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  594837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594837, CurrentID=0, Status='live', Symbol='SNORD101', Aliases='U101', description='small nucleolar RNA, C/D box 101', other_designations='C/D box snoRNA U101', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132815307, end_position_on_the_genomic_accession=132815379, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  594838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594838, CurrentID=0, Status='live', Symbol='SNORD100', Aliases='HBII-429', description='small nucleolar RNA, C/D box 100', other_designations='C/D box snoRNA HBII-429', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132816802, end_position_on_the_genomic_accession=132816877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  594839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594839, CurrentID=0, Status='live', Symbol='SNORA33', Aliases='ACA33', description='small nucleolar RNA, H/ACA box 33', other_designations='ACA33 small nucleolar RNA|ACA33 snoRNA', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132817219, end_position_on_the_genomic_accession=132817351, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  594840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594840, CurrentID=0, Status='live', Symbol='ZFAT-AS1', Aliases='NCRNA00070, SAS-ZFAT, ZFAT-AS, ZFATAS', description='ZFAT antisense RNA 1', other_designations='ZFAT antisense RNA (non-protein coding)|ZFAT antisense RNA 1 (non-protein coding)|zinc finger protein 406 antisense transcript', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134598071, end_position_on_the_genomic_accession=134600689, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  201625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201625, CurrentID=0, Status='live', Symbol='DNAH12', Aliases='DHC3, DLP12L, DNAH7L, DNAHC12, DNAHC3, DNHD2, HDHC3, HL-19, HL19, DNAH12', description='dynein, axonemal, heavy chain 12', other_designations='axonemal beta dynein heavy chain 12|axonemal dynein heavy chain isotype3|ciliary dynein heavy chain 12|dynein heavy chain domain-containing protein 2|dynein, axonemal, heavy polypeptide 12|dynein, heavy chain-5', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57293699, end_position_on_the_genomic_accession=57544344, orientation='minus', exon_count=79, OMIM=603340, no_hdr0=''),
  594842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594842, CurrentID=0, Status='live', Symbol='HAS2-AS1', Aliases='HAS2-AS, HAS2AS, HASNT, NCRNA00077', description='HAS2 antisense RNA 1', other_designations='HAS2 antisense RNA (non-protein coding)|HAS2 antisense RNA 1 (non-protein coding)|hyaluronan synthase 2 antisense', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121639346, end_position_on_the_genomic_accession=121645325, orientation='plus', exon_count=4, OMIM=614353, no_hdr0=''),
  201627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201627, CurrentID=0, Status='live', Symbol='DENND6A', Aliases='AFI1A, FAM116A', description='DENN/MADD domain containing 6A', other_designations='DENN domain-containing protein 6A|family with sequence similarity 116, member A|protein FAM116A', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57625454, end_position_on_the_genomic_accession=57693089, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100128922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128922, CurrentID=0, Status='live', Symbol='LOC100128922', Aliases='', description='connexin', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=48727949, end_position_on_the_genomic_accession=48728197, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  201633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201633, CurrentID=0, Status='live', Symbol='TIGIT', Aliases='VSIG9, VSTM3, WUCAM', description='T-cell immunoreceptor with Ig and ITIM domains', other_designations='T cell immunoreceptor with Ig and ITIM domains|V-set and immunoglobulin domain containing 9|V-set and immunoglobulin domain-containing protein 9|V-set and transmembrane domain containing 3|V-set and transmembrane domain-containing protein 3|Washington University cell adhesion molecule', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114293986, end_position_on_the_genomic_accession=114310288, orientation='plus', exon_count=4, OMIM=612859, no_hdr0=''),
  105371803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371803, CurrentID=0, Status='live', Symbol='LOC105371803', Aliases='', description='uncharacterized LOC105371803', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  594855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594855, CurrentID=0, Status='live', Symbol='CPLX3', Aliases='CPX-III, CPXIII, Nbla11589', description='complexin 3', other_designations='CPX III|complexin III', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74826610, end_position_on_the_genomic_accession=74831795, orientation='plus', exon_count=3, OMIM=609585, no_hdr0=''),
  594857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=594857, CurrentID=0, Status='live', Symbol='NPS', Aliases='', description='neuropeptide S', other_designations='prepro-neuropeptide S', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127549349, end_position_on_the_genomic_accession=127552671, orientation='plus', exon_count=3, OMIM=609513, no_hdr0=''),
  100129280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129280, CurrentID=0, Status='live', Symbol='LOC100129280', Aliases='', description='uncharacterized LOC100129280', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=97646471, end_position_on_the_genomic_accession=97647823, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  201651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201651, CurrentID=0, Status='live', Symbol='AADACP1', Aliases='', description='arylacetamide deacetylase pseudogene 1', other_designations='arylacetamide deacetylase (esterase) pseudogene', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151770456, end_position_on_the_genomic_accession=151784894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128926, CurrentID=0, Status='live', Symbol='LOC100128926', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1-like', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2568390, end_position_on_the_genomic_accession=2569122, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128927, CurrentID=0, Status='live', Symbol='ZBTB42', Aliases='LCCS6, ZNF925', description='zinc finger and BTB domain containing 42', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104800596, end_position_on_the_genomic_accession=104804712, orientation='plus', exon_count=2, OMIM=613915, no_hdr0=''),
  105372398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372398, CurrentID=0, Status='live', Symbol='LOC105372398', Aliases='', description='uncharacterized LOC105372398', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39014032, end_position_on_the_genomic_accession=39014508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371808, CurrentID=0, Status='live', Symbol='LOC105371808', Aliases='', description='uncharacterized LOC105371808', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48520541, end_position_on_the_genomic_accession=48527507, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100631371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631371, CurrentID=0, Status='live', Symbol='LOC100631371', Aliases='', description='cytochrome P450 family 2 subfamily D member 6 pseudogene', other_designations='cytochrome P450, family 2, subfamily D, polypeptide 6 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101926993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926993, CurrentID=0, Status='live', Symbol='LOC101926993', Aliases='', description='protein SPHAR-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43284599, end_position_on_the_genomic_accession=43291735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128929, CurrentID=0, Status='live', Symbol='LOC100128929', Aliases='', description='archaelysin family metallopeptidase 2 pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27218134, end_position_on_the_genomic_accession=27218974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371810, CurrentID=0, Status='live', Symbol='LOC105371810', Aliases='', description='uncharacterized LOC105371810', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48623142, end_position_on_the_genomic_accession=48625737, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  347148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347148, CurrentID=0, Status='live', Symbol='QRFP', Aliases='26RFa, P518', description='pyroglutamylated RFamide peptide', other_designations='P518 precursor protein|RF(Arg-Phe)amide family 26 amino acid peptide (P518)|prepro-QRFP', map_location='9q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130893428, end_position_on_the_genomic_accession=130893838, orientation='minus', exon_count=1, OMIM=609795, no_hdr0=''),
  105374965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374965, CurrentID=0, Status='live', Symbol='LOC105374965', Aliases='', description='uncharacterized LOC105374965', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20069576, end_position_on_the_genomic_accession=20074664, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371814, CurrentID=0, Status='live', Symbol='LOC105371814', Aliases='', description='uncharacterized LOC105371814', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48874860, end_position_on_the_genomic_accession=48892677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371816, CurrentID=0, Status='live', Symbol='LOC105371816', Aliases='', description='uncharacterized LOC105371816', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49224873, end_position_on_the_genomic_accession=49226640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  201725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201725, CurrentID=0, Status='live', Symbol='C4orf46', Aliases='RCDG1', description='chromosome 4 open reading frame 46', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158666675, end_position_on_the_genomic_accession=158672255, orientation='minus', exon_count=2, OMIM=616210, no_hdr0=''),
  105374966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374966, CurrentID=0, Status='live', Symbol='LOC105374966', Aliases='', description='uncharacterized LOC105374966', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21136037, end_position_on_the_genomic_accession=21140935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128939, CurrentID=0, Status='live', Symbol='LOC100128939', Aliases='', description='pyruvate dehydrogenase kinase, isozyme 3 pseudogene', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91709115, end_position_on_the_genomic_accession=91710644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371820, CurrentID=0, Status='live', Symbol='LOC105371820', Aliases='', description='uncharacterized LOC105371820', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50213813, end_position_on_the_genomic_accession=50216750, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371821, CurrentID=0, Status='live', Symbol='LOC105371821', Aliases='', description='uncharacterized LOC105371821', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50209365, end_position_on_the_genomic_accession=50214138, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371822, CurrentID=0, Status='live', Symbol='LOC105371822', Aliases='', description='uncharacterized LOC105371822', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50256613, end_position_on_the_genomic_accession=50263339, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377636, CurrentID=0, Status='live', Symbol='LOC105377636', Aliases='', description='uncharacterized LOC105377636', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88897957, end_position_on_the_genomic_accession=88905372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128944, CurrentID=0, Status='live', Symbol='YWHAQP3', Aliases='', description='YWHAQ pseudogene 3', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56871798, end_position_on_the_genomic_accession=56872404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653233, CurrentID=0, Status='live', Symbol='LOC100653233', Aliases='', description='uncharacterized LOC100653233', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57404767, end_position_on_the_genomic_accession=57412147, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371826, CurrentID=0, Status='live', Symbol='LOC105371826', Aliases='', description='uncharacterized LOC105371826', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50908915, end_position_on_the_genomic_accession=50923919, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105371827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371827, CurrentID=0, Status='live', Symbol='LOC105371827', Aliases='', description='uncharacterized LOC105371827', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51041801, end_position_on_the_genomic_accession=51085953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  201780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201780, CurrentID=0, Status='live', Symbol='SLC10A4', Aliases='P4', description='solute carrier family 10 member 4', other_designations='Na(+)/bile acid cotransporter 4|bile acid transporter SLC10A4|solute carrier family 10 (sodium/bile acid cotransporter family), member 4|solute carrier family 10, member 4', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48483343, end_position_on_the_genomic_accession=48489524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128948, CurrentID=0, Status='live', Symbol='LOC100128948', Aliases='', description='zinc finger protein 607 pseudogene', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37802988, end_position_on_the_genomic_accession=37804526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106632258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632258, CurrentID=0, Status='live', Symbol='RNA18S2', Aliases='', description='RNA, 18S ribosomal 2', other_designations='', map_location='14p12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632259, CurrentID=0, Status='live', Symbol='RNA18S1', Aliases='', description='RNA, 18S ribosomal 1', other_designations='', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632260, CurrentID=0, Status='live', Symbol='RNA5-8S4', Aliases='', description='RNA, 5.8S ribosomal 4', other_designations='', map_location='21p12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632261, CurrentID=0, Status='live', Symbol='RNA5-8S3', Aliases='', description='RNA, 5.8S ribosomal 3', other_designations='', map_location='15p12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632262, CurrentID=0, Status='live', Symbol='RNA5-8S2', Aliases='', description='RNA, 5.8S ribosomal 2', other_designations='', map_location='14p12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632263, CurrentID=0, Status='live', Symbol='RNA5-8S1', Aliases='', description='RNA, 5.8S ribosomal 1', other_designations='', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632264, CurrentID=0, Status='live', Symbol='RNA28S4', Aliases='', description='RNA, 28S ribosomal 4', other_designations='', map_location='21p12', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632265, CurrentID=0, Status='live', Symbol='RNA28S3', Aliases='', description='RNA, 28S ribosomal 3', other_designations='', map_location='15p12', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632266, CurrentID=0, Status='live', Symbol='RNA28S2', Aliases='', description='RNA, 28S ribosomal 2', other_designations='', map_location='14p12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632267, CurrentID=0, Status='live', Symbol='RNA28S1', Aliases='', description='RNA, 28S ribosomal 1', other_designations='', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632269, CurrentID=0, Status='live', Symbol='CCDC58P4', Aliases='', description='coiled-coil domain containing 58 pseudogene 4', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632270, CurrentID=0, Status='live', Symbol='DUX4L35', Aliases='', description='double homeobox 4 like 35, pseudogene', other_designations='', map_location='20q11.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106632271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106632271, CurrentID=0, Status='live', Symbol='SNORD128', Aliases='ZL43', description='small nucleolar RNA, C/D box 128', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128958, CurrentID=0, Status='live', Symbol='LOC100128958', Aliases='', description='ferredoxin 1-like pseudogene', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240654114, end_position_on_the_genomic_accession=240654624, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216340, CurrentID=0, Status='live', Symbol='RPS24P16', Aliases='RPS24_6_1447', description='ribosomal protein S24 pseudogene 16', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67232407, end_position_on_the_genomic_accession=67232914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371839, CurrentID=0, Status='live', Symbol='LOC105371839', Aliases='', description='uncharacterized LOC105371839', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57771781, end_position_on_the_genomic_accession=57835995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  201853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201853, CurrentID=0, Status='live', Symbol='LINC00504', Aliases='', description='long intergenic non-protein coding RNA 504', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14470465, end_position_on_the_genomic_accession=14888169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371840, CurrentID=0, Status='live', Symbol='LOC105371840', Aliases='', description='uncharacterized LOC105371840', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58045800, end_position_on_the_genomic_accession=58059944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128961, CurrentID=0, Status='live', Symbol='GAPDHP62', Aliases='GAPDHL7, GAPDL7', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 62', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene|glyceraldehyde-3-phosphate dehydrogenase-like 7, pseudogene', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100550540, end_position_on_the_genomic_accession=100551790, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371842, CurrentID=0, Status='live', Symbol='LOC105371842', Aliases='', description='uncharacterized LOC105371842', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58632920, end_position_on_the_genomic_accession=58651649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371843, CurrentID=0, Status='live', Symbol='LOC105371843', Aliases='', description='uncharacterized LOC105371843', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58724111, end_position_on_the_genomic_accession=58756105, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996246, CurrentID=0, Status='live', Symbol='LOC100996246', Aliases='', description='uncharacterized LOC100996246', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132186735, end_position_on_the_genomic_accession=132189727, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996247, CurrentID=0, Status='live', Symbol='LOC100996247', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18874919, end_position_on_the_genomic_accession=18875911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927000, CurrentID=0, Status='live', Symbol='LOC101927000', Aliases='', description='uncharacterized LOC101927000', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=561481, end_position_on_the_genomic_accession=566230, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  595097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595097, CurrentID=0, Status='live', Symbol='SNORD16', Aliases='U16', description='small nucleolar RNA, C/D box 16', other_designations='U16 small nucleolar RNA|U16 snoRNA', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66502811, end_position_on_the_genomic_accession=66502910, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  595098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595098, CurrentID=0, Status='live', Symbol='SNORD18A', Aliases='U18A', description='small nucleolar RNA, C/D box 18A', other_designations='U18A small nucleolar RNA|U18A snoRNA', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66503245, end_position_on_the_genomic_accession=66503314, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  595099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595099, CurrentID=0, Status='live', Symbol='SNORD18B', Aliases='U18B', description='small nucleolar RNA, C/D box 18B', other_designations='U18B small nucleolar RNA|U18B snoRNA', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66502022, end_position_on_the_genomic_accession=66502091, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  595100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595100, CurrentID=0, Status='live', Symbol='SNORD18C', Aliases='U18C', description='small nucleolar RNA, C/D box 18C', other_designations='U18C small nucleolar RNA|U18C snoRNA', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66501252, end_position_on_the_genomic_accession=66501318, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  595101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595101, CurrentID=0, Status='live', Symbol='SMG1P5', Aliases='', description='SMG1 pseudogene 5', other_designations='smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30267593, end_position_on_the_genomic_accession=30335374, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  100996255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996255, CurrentID=0, Status='live', Symbol='LOC100996255', Aliases='', description='uncharacterized LOC100996255', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32536759, end_position_on_the_genomic_accession=32580609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996259, CurrentID=0, Status='live', Symbol='FAM106B', Aliases='', description='family with sequence similarity 106 member B', other_designations='family with sequence similarity 106, member B', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20417292, end_position_on_the_genomic_accession=20419050, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128966, CurrentID=0, Status='live', Symbol='LOC100128966', Aliases='', description='uncharacterized LOC100128966', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138332055, end_position_on_the_genomic_accession=138336085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379336, CurrentID=0, Status='live', Symbol='LOC105379336', Aliases='', description='uncharacterized LOC105379336', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26136157, end_position_on_the_genomic_accession=26206775, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  201895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201895, CurrentID=0, Status='live', Symbol='SMIM14', Aliases='C4orf34', description='small integral membrane protein 14', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39547297, end_position_on_the_genomic_accession=39638861, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100996266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996266, CurrentID=0, Status='live', Symbol='LINC01510', Aliases='', description='long intergenic non-protein coding RNA 1510', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116563594, end_position_on_the_genomic_accession=116614820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645319, CurrentID=0, Status='live', Symbol='LOC645319', Aliases='', description='MAS-related GPR, member X3 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18196660, end_position_on_the_genomic_accession=18197345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996273, CurrentID=0, Status='live', Symbol='LOC100996273', Aliases='', description='uncharacterized LOC100996273', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77439912, end_position_on_the_genomic_accession=77446373, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996274, CurrentID=0, Status='live', Symbol='LOC100996274', Aliases='', description='uncharacterized LOC100996274', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79144752, end_position_on_the_genomic_accession=79149693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996275, CurrentID=0, Status='live', Symbol='LOC100996275', Aliases='', description='uncharacterized LOC100996275', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129484143, end_position_on_the_genomic_accession=129496804, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996279, CurrentID=0, Status='live', Symbol='LINC00269', Aliases='CXorf62, NCRNA00269', description='long intergenic non-protein coding RNA 269', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69179557, end_position_on_the_genomic_accession=69209924, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100996280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996280, CurrentID=0, Status='live', Symbol='LINC00911', Aliases='', description='long intergenic non-protein coding RNA 911', other_designations='CTD-2128A3.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85393879, end_position_on_the_genomic_accession=85420074, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996284, CurrentID=0, Status='live', Symbol='LOC100996284', Aliases='', description='Fc receptor-like protein 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157862846, end_position_on_the_genomic_accession=157931144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996286, CurrentID=0, Status='live', Symbol='LOC100996286', Aliases='', description='uncharacterized LOC100996286', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152100753, end_position_on_the_genomic_accession=152104720, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  595135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=595135, CurrentID=0, Status='live', Symbol='PGM5P2', Aliases='', description='phosphoglucomutase 5 pseudogene 2', other_designations='', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41007011, end_position_on_the_genomic_accession=41074625, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100996288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996288, CurrentID=0, Status='live', Symbol='LOC100996288', Aliases='', description='uncharacterized LOC100996288', other_designations='CTD-2623N2.11', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9834079, end_position_on_the_genomic_accession=9835147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371851, CurrentID=0, Status='live', Symbol='LOC105371851', Aliases='', description='uncharacterized LOC105371851', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61740195, end_position_on_the_genomic_accession=61744291, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996294, CurrentID=0, Status='live', Symbol='LOC100996294', Aliases='', description='uncharacterized LOC100996294', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20530042, end_position_on_the_genomic_accession=20533271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996295, CurrentID=0, Status='live', Symbol='DNAH17-AS1', Aliases='', description='DNAH17 antisense RNA 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78484910, end_position_on_the_genomic_accession=78503056, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  645324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645324, CurrentID=0, Status='live', Symbol='LOC645324', Aliases='', description='GrpE-like 1, mitochondrial (E. coli) pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76112910, end_position_on_the_genomic_accession=76113435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  201931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201931, CurrentID=0, Status='live', Symbol='TMEM192', Aliases='', description='transmembrane protein 192', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165076078, end_position_on_the_genomic_accession=165112913, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100996301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996301, CurrentID=0, Status='live', Symbol='FOXD3-AS1', Aliases='pasFOXD3', description='FOXD3 antisense RNA 1 (head to head)', other_designations='FOXD3 promoter-associated antisense long non-coding RNA', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63320884, end_position_on_the_genomic_accession=63324441, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100996306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996306, CurrentID=0, Status='live', Symbol='TUBB4BP6', Aliases='', description='tubulin beta 4B class IVb pseudogene 6', other_designations='tubulin, beta 4B class IVb pseudogene 6', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=121526157, end_position_on_the_genomic_accession=121526623, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996307, CurrentID=0, Status='live', Symbol='LIPE-AS1', Aliases='', description='LIPE antisense RNA 1', other_designations='CTB-50E14.6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42397148, end_position_on_the_genomic_accession=42652355, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128974, CurrentID=0, Status='live', Symbol='RBM17P4', Aliases='', description='RNA binding motif protein 17 pseudogene 4', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35298047, end_position_on_the_genomic_accession=35299610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996315, CurrentID=0, Status='live', Symbol='CHIAP3', Aliases='', description='chitinase, acidic pseudogene 3', other_designations='acidic mammalian chitinase-like', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=111895850, end_position_on_the_genomic_accession=111913893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996316, CurrentID=0, Status='live', Symbol='LOC100996316', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996318, CurrentID=0, Status='live', Symbol='LOC100996318', Aliases='', description='uncharacterized LOC100996318', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121097142, end_position_on_the_genomic_accession=121098265, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317648, CurrentID=0, Status='live', Symbol='NOP14-AS1', Aliases='C4orf10, RES4-24', description='NOP14 antisense RNA 1', other_designations='NOP14 antisense RNA 1 (non-protein coding)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2935546, end_position_on_the_genomic_accession=2951078, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421243, CurrentID=0, Status='live', Symbol='LOC100421243', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996324, CurrentID=0, Status='live', Symbol='LOC100996324', Aliases='', description='uncharacterized LOC100996324', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12739486, end_position_on_the_genomic_accession=12749422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996325, CurrentID=0, Status='live', Symbol='LOC100996325', Aliases='', description='uncharacterized LOC100996325', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=602169, end_position_on_the_genomic_accession=612210, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  55505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55505, CurrentID=0, Status='live', Symbol='NOP10', Aliases='DKCB1, NOLA3P, NOP10', description='NOP10 ribonucleoprotein', other_designations='NOP10 ribonucleoprotein homolog|homolog of yeast Nop10p|nucleolar protein 10|nucleolar protein family A member 3|nucleolar protein family A, member 3 (H/ACA small nucleolar RNPs)|snoRNP protein NOP10', map_location='15q14-q15', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34341716, end_position_on_the_genomic_accession=34343161, orientation='minus', exon_count=2, OMIM=606471, no_hdr0=''),
  105372400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372400, CurrentID=0, Status='live', Symbol='LOC105372400', Aliases='', description='uncharacterized LOC105372400', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40286431, end_position_on_the_genomic_accession=40293192, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996331, CurrentID=0, Status='live', Symbol='POTEB', Aliases='A26B1, CT104.5, POTE-15, POTE15', description='POTE ankyrin domain family member B', other_designations='ANKRD26-like family B, member 1|POTE ankyrin domain family member B-like|POTE ankyrin domain family, member B|cancer/testis antigen family 104, member 5|prostate, ovary, testis-expressed protein on chromosome 15|protein expressed in prostate, ovary, testis, and placenta 15', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21846329, end_position_on_the_genomic_accession=21877396, orientation='minus', exon_count=12, OMIM=608912, no_hdr0=''),
  201965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201965, CurrentID=0, Status='live', Symbol='RWDD4', Aliases='FAM28AA, RWDD4', description='RWD domain containing 4', other_designations='RWD domain containing 4A|RWD domain-containing protein 4A|family with sequence similarity 28, member A', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183639636, end_position_on_the_genomic_accession=183659178, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100996335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996335, CurrentID=0, Status='live', Symbol='LOC100996335', Aliases='', description='putative uncharacterized protein LOC388882', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21308504, end_position_on_the_genomic_accession=21325019, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100996337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996337, CurrentID=0, Status='live', Symbol='LOC100996337', Aliases='', description='transmembrane protein 191B-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927003, CurrentID=0, Status='live', Symbol='PKIA-AS1', Aliases='', description='PKIA antisense RNA 1', other_designations='', map_location='8q21.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78533892, end_position_on_the_genomic_accession=78558503, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371859, CurrentID=0, Status='live', Symbol='LOC105371859', Aliases='', description='uncharacterized LOC105371859', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  201973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201973, CurrentID=0, Status='live', Symbol='PRIMPOL', Aliases='CCDC111, MYP22', description='primase and DNA directed polymerase', other_designations='coiled-coil domain containing 111|coiled-coil domain-containing protein 111|hPrimpol1|primase and polymerase (DNA-directed)', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184649613, end_position_on_the_genomic_accession=184694959, orientation='plus', exon_count=16, OMIM=615421, no_hdr0=''),
  100996342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996342, CurrentID=0, Status='live', Symbol='LOC100996342', Aliases='', description='uncharacterized LOC100996342', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17121595, end_position_on_the_genomic_accession=17132104, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996345, CurrentID=0, Status='live', Symbol='LOC100996345', Aliases='', description='uncharacterized LOC100996345', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54366004, end_position_on_the_genomic_accession=54370660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996348, CurrentID=0, Status='live', Symbol='CA3-AS1', Aliases='', description='CA3 antisense RNA 1', other_designations='', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85441806, end_position_on_the_genomic_accession=85464915, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100996349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996349, CurrentID=0, Status='live', Symbol='LOC100996349', Aliases='', description='testis expressed 264 pseudogene', other_designations='testis expressed 264-like', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22608016, end_position_on_the_genomic_accession=22608501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996350, CurrentID=0, Status='live', Symbol='DUX4L26', Aliases='', description='double homeobox 4 like 26', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75668931, end_position_on_the_genomic_accession=75670327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996351, CurrentID=0, Status='live', Symbol='LOC100996351', Aliases='', description='uncharacterized LOC100996351', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3141576, end_position_on_the_genomic_accession=3155175, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  201989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=201989, CurrentID=0, Status='live', Symbol='GK6P', Aliases='GK7P, GKP1', description='glycerol kinase 6 pseudogene', other_designations='glycerol kinase 7 pseudogene|glycerol kinase pseudogene 1', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119392863, end_position_on_the_genomic_accession=119395070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645544, CurrentID=0, Status='live', Symbol='LOC645544', Aliases='', description='cuticle collagen 7-like', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71630771, end_position_on_the_genomic_accession=71748771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996361, CurrentID=0, Status='live', Symbol='LOC100996361', Aliases='', description='ribosomal protein S6 kinase, 70kDa, polypeptide 1 pseudogene', other_designations='ribosomal protein S6 kinase beta-1-like', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62231469, end_position_on_the_genomic_accession=62247172, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996362, CurrentID=0, Status='live', Symbol='LOC100996362', Aliases='', description='C-terminal-binding protein 2 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371863, CurrentID=0, Status='live', Symbol='LOC105371863', Aliases='', description='uncharacterized LOC105371863', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64951467, end_position_on_the_genomic_accession=64953225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996364, CurrentID=0, Status='live', Symbol='LOC100996364', Aliases='', description='40S ribosomal protein SA-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10580457, end_position_on_the_genomic_accession=10580988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996365, CurrentID=0, Status='live', Symbol='LOC100996365', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63756809, end_position_on_the_genomic_accession=63757407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371864, CurrentID=0, Status='live', Symbol='LOC105371864', Aliases='', description='uncharacterized LOC105371864', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65078158, end_position_on_the_genomic_accession=65100712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375019, CurrentID=0, Status='live', Symbol='LOC105375019', Aliases='', description='uncharacterized LOC105375019', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31682895, end_position_on_the_genomic_accession=31687124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996379, CurrentID=0, Status='live', Symbol='LOC100996379', Aliases='', description='E3 ubiquitin-protein ligase HERC2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20515748, end_position_on_the_genomic_accession=20520710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874355, CurrentID=0, Status='live', Symbol='ARHGEF9-IT1', Aliases='', description='ARHGEF9 intronic transcript 1', other_designations='ARHGEF9 intronic transcript 1 (non-protein coding)', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63670196, end_position_on_the_genomic_accession=63671502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996384, CurrentID=0, Status='live', Symbol='LOC100996384', Aliases='', description='aconitate hydratase, mitochondrial pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75920666, end_position_on_the_genomic_accession=75944298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996385, CurrentID=0, Status='live', Symbol='LOC100996385', Aliases='', description='uncharacterized LOC100996385', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176049678, end_position_on_the_genomic_accession=176062055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  202018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202018, CurrentID=0, Status='live', Symbol='TAPT1', Aliases='CMVFR', description='transmembrane anterior posterior transformation 1', other_designations='cytomegalovirus partial fusion receptor', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16160505, end_position_on_the_genomic_accession=16227158, orientation='minus', exon_count=20, OMIM=612758, no_hdr0=''),
  100996387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996387, CurrentID=0, Status='live', Symbol='CICP24', Aliases='', description='capicua transcriptional repressor pseudogene 24', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=63228569, end_position_on_the_genomic_accession=63228975, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  202020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202020, CurrentID=0, Status='live', Symbol='TAPT1-AS1', Aliases='', description='TAPT1 antisense RNA 1 (head to head)', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16226663, end_position_on_the_genomic_accession=16258187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371868, CurrentID=0, Status='live', Symbol='LOC105371868', Aliases='', description='uncharacterized LOC105371868', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67560394, end_position_on_the_genomic_accession=67575444, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371869, CurrentID=0, Status='live', Symbol='LOC105371869', Aliases='', description='uncharacterized LOC105371869', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67431310, end_position_on_the_genomic_accession=67444617, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100996401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996401, CurrentID=0, Status='live', Symbol='LOC100996401', Aliases='', description='transmembrane protein 191B', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18527067, end_position_on_the_genomic_accession=18527706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996404, CurrentID=0, Status='live', Symbol='LOC100996404', Aliases='', description='uncharacterized LOC100996404', other_designations='CTA-407F11.8', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25561120, end_position_on_the_genomic_accession=25564893, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  121054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121054, CurrentID=0, Status='live', Symbol='KRT18P20', Aliases='', description='keratin 18 pseudogene 20', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104976684, end_position_on_the_genomic_accession=104978077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371871, CurrentID=0, Status='live', Symbol='LOC105371871', Aliases='', description='uncharacterized LOC105371871', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68147505, end_position_on_the_genomic_accession=68161137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996413, CurrentID=0, Status='live', Symbol='LOC100996413', Aliases='', description='uncharacterized LOC100996413', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996414, CurrentID=0, Status='live', Symbol='LOC100996414', Aliases='', description='putative uncharacterized protein FLJ92257', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63754798, end_position_on_the_genomic_accession=63755530, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996415, CurrentID=0, Status='live', Symbol='LOC100996415', Aliases='', description='putative uncharacterized protein LOC388882', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18178026, end_position_on_the_genomic_accession=18202437, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  55520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55520, CurrentID=0, Status='live', Symbol='ELAC1', Aliases='D29', description='elaC ribonuclease Z 1', other_designations="RNase Z 1|RNaseZ(S)|deleted in Ma29|elaC homolog 1|elaC homolog protein 1|ribonuclease Z 1|tRNA 3 endonuclease 1|tRNA 3' processing endoribonuclease|tRNA Z (short form)|tRNase Z 1|tRNase ZS", map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50968017, end_position_on_the_genomic_accession=50988120, orientation='plus', exon_count=4, OMIM=608079, no_hdr0=''),
  202051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202051, CurrentID=0, Status='live', Symbol='SPATA24', Aliases='CCDC161, T6441', description='spermatogenesis associated 24', other_designations='coiled-coil domain containing 161|testis protein T6441 homolog', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139395940, end_position_on_the_genomic_accession=139404155, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  202052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202052, CurrentID=0, Status='live', Symbol='DNAJC18', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C18', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 18', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139410203, end_position_on_the_genomic_accession=139439525, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100128993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128993, CurrentID=0, Status='live', Symbol='LOC100128993', Aliases='', description='uncharacterized LOC100128993', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19183674, end_position_on_the_genomic_accession=19245522, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100996425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996425, CurrentID=0, Status='live', Symbol='PRCAT47', Aliases='', description='prostate cancer-associated 47', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80828735, end_position_on_the_genomic_accession=80840222, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927006, CurrentID=0, Status='live', Symbol='RPSAP76', Aliases='', description='ribosomal protein SA pseudogene 76', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125194561, end_position_on_the_genomic_accession=125195893, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996432, CurrentID=0, Status='live', Symbol='LOC100996432', Aliases='', description='uncharacterized LOC100996432', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18872379, end_position_on_the_genomic_accession=18893080, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996434, CurrentID=0, Status='live', Symbol='LOC100996434', Aliases='', description='uncharacterized protein FLJ39653 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66473792, end_position_on_the_genomic_accession=66474713, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996435, CurrentID=0, Status='live', Symbol='CA15P1', Aliases='', description='carbonic anhydrase XV, pseudogene 1', other_designations='carbonic anhydrase 15-like', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=19018766, end_position_on_the_genomic_accession=19022784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996437, CurrentID=0, Status='live', Symbol='LOC100996437', Aliases='', description='uncharacterized LOC100996437', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66654519, end_position_on_the_genomic_accession=66669602, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996438, CurrentID=0, Status='live', Symbol='MED15P6', Aliases='', description='mediator complex subunit 15 pseudogene 6', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18914213, end_position_on_the_genomic_accession=18919822, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371876, CurrentID=0, Status='live', Symbol='LOC105371876', Aliases='', description='uncharacterized LOC105371876', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68798729, end_position_on_the_genomic_accession=68800150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996442, CurrentID=0, Status='live', Symbol='LOC100996442', Aliases='', description='uncharacterized LOC100996442', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=141934, end_position_on_the_genomic_accession=174394, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100996447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996447, CurrentID=0, Status='live', Symbol='LOC100996447', Aliases='', description='uncharacterized LOC100996447', other_designations='', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158545220, end_position_on_the_genomic_accession=158571066, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  104797536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104797536, CurrentID=0, Status='live', Symbol='LINC01402', Aliases='GS1-421I3.4, TCONS_00017087', description='long intergenic non-protein coding RNA 1402', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120117642, end_position_on_the_genomic_accession=120119700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104797537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104797537, CurrentID=0, Status='live', Symbol='GACAT3', Aliases='LINC01458, lncRNA-AC130710', description='gastric cancer associated transcript 3 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1458', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16050427, end_position_on_the_genomic_accession=16085689, orientation='plus', exon_count=4, OMIM=616132, no_hdr0=''),
  104797538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104797538, CurrentID=0, Status='live', Symbol='KCNMB2-AS1', Aliases='', description='KCNMB2 antisense RNA 1', other_designations='RP11-385J1.2', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178525467, end_position_on_the_genomic_accession=178860405, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371878, CurrentID=0, Status='live', Symbol='LOC105371878', Aliases='', description='uncharacterized LOC105371878', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69331844, end_position_on_the_genomic_accession=69343487, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996455, CurrentID=0, Status='live', Symbol='LOC100996455', Aliases='', description='uncharacterized LOC100996455', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64449057, end_position_on_the_genomic_accession=64451657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927007, CurrentID=0, Status='live', Symbol='LINC01365', Aliases='', description='long intergenic non-protein coding RNA 1365', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119799106, end_position_on_the_genomic_accession=119804515, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  55527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55527, CurrentID=0, Status='live', Symbol='FEM1A', Aliases='EPRAP', description='fem-1 homolog a (C. elegans)', other_designations='FEM1-alpha|prostaglandin E receptor 4-associated protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4791716, end_position_on_the_genomic_accession=4795559, orientation='plus', exon_count=1, OMIM=613538, no_hdr0=''),
  100996461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996461, CurrentID=0, Status='live', Symbol='LOC100996461', Aliases='', description='uncharacterized LOC100996461', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=73927556, end_position_on_the_genomic_accession=73934687, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996465, CurrentID=0, Status='live', Symbol='LCA10', Aliases='LCAP', description='lung carcinoma-associated 10', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153880672, end_position_on_the_genomic_accession=153888990, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100996466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996466, CurrentID=0, Status='live', Symbol='MED15P7', Aliases='', description='mediator complex subunit 15 pseudogene 7', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15631942, end_position_on_the_genomic_accession=15639329, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129001, CurrentID=0, Status='live', Symbol='LOC100129001', Aliases='', description='general transcription factor IIIC subunit 6 pseudogene', other_designations='general transcription factor IIIC, polypeptide 6, alpha 35kDa pseudogene', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95584238, end_position_on_the_genomic_accession=95585037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996476, CurrentID=0, Status='live', Symbol='LOC100996476', Aliases='', description='uncharacterized protein FLJ39653 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76149797, end_position_on_the_genomic_accession=76150613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371882, CurrentID=0, Status='live', Symbol='LOC105371882', Aliases='', description='uncharacterized LOC105371882', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70150905, end_position_on_the_genomic_accession=70158192, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996478, CurrentID=0, Status='live', Symbol='LOC100996478', Aliases='', description='uncharacterized LOC100996478', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85064110, end_position_on_the_genomic_accession=85067535, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996481, CurrentID=0, Status='live', Symbol='PRIM2B', Aliases='', description='primase, DNA, polypeptide 2 (58kDa) pseudogene', other_designations='DNA primase large subunit-like', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=60281438, end_position_on_the_genomic_accession=60545629, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55531, CurrentID=0, Status='live', Symbol='ELMOD1', Aliases='', description='ELMO/CED-12 domain containing 1', other_designations='ELMO domain containing 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107591091, end_position_on_the_genomic_accession=107666779, orientation='plus', exon_count=13, OMIM=615456, no_hdr0=''),
  100996485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996485, CurrentID=0, Status='live', Symbol='C5orf66', Aliases='', description='chromosome 5 open reading frame 66', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135033280, end_position_on_the_genomic_accession=135344680, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100996488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996488, CurrentID=0, Status='live', Symbol='FKSG48', Aliases='', description='FKSG48', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=33439154, end_position_on_the_genomic_accession=33440162, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371884, CurrentID=0, Status='live', Symbol='LOC105371884', Aliases='', description='uncharacterized LOC105371884', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70811471, end_position_on_the_genomic_accession=70835275, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  202122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202122, CurrentID=0, Status='live', Symbol='RPL27P10', Aliases='RPL27_5_551', description='ribosomal protein L27 pseudogene 10', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32221897, end_position_on_the_genomic_accession=32222388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996492, CurrentID=0, Status='live', Symbol='LINC01314', Aliases='', description='long intergenic non-protein coding RNA 1314', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80195479, end_position_on_the_genomic_accession=80252261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645357, CurrentID=0, Status='live', Symbol='LOC645357', Aliases='', description='kinesin family member pseudogene', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2381982, end_position_on_the_genomic_accession=2392275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271183, CurrentID=0, Status='live', Symbol='RPL21P77', Aliases='RPL21_29_883', description='ribosomal protein L21 pseudogene 77', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22389835, end_position_on_the_genomic_accession=22390286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996501, CurrentID=0, Status='live', Symbol='HNRNPA1P34', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 34', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9027399, end_position_on_the_genomic_accession=9028116, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  202134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202134, CurrentID=0, Status='live', Symbol='FAM153B', Aliases='', description='family with sequence similarity 153 member B', other_designations='family with sequence similarity 153, member B', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176063709, end_position_on_the_genomic_accession=176114798, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100996503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996503, CurrentID=0, Status='live', Symbol='LOC100996503', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74220927, end_position_on_the_genomic_accession=74221476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996506, CurrentID=0, Status='live', Symbol='LOC100996506', Aliases='', description='uncharacterized LOC100996506', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85314216, end_position_on_the_genomic_accession=85316844, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129007, CurrentID=0, Status='live', Symbol='FMO8P', Aliases='', description='flavin containing monooxygenase 8 pseudogene', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166566178, end_position_on_the_genomic_accession=166580711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996511, CurrentID=0, Status='live', Symbol='LINC01355', Aliases='', description='long intergenic non-protein coding RNA 1355', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23281309, end_position_on_the_genomic_accession=23285518, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129008, CurrentID=0, Status='live', Symbol='CUL4AP1', Aliases='', description='cullin 4A pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116426657, end_position_on_the_genomic_accession=116430446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996517, CurrentID=0, Status='live', Symbol='LOC100996517', Aliases='', description='vesicle-trafficking protein SEC22b-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148767753, end_position_on_the_genomic_accession=148791814, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  202151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202151, CurrentID=0, Status='live', Symbol='RANBP3L', Aliases='', description='RAN binding protein 3-like', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36247874, end_position_on_the_genomic_accession=36301926, orientation='minus', exon_count=16, OMIM=616391, no_hdr0=''),
  100996520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996520, CurrentID=0, Status='live', Symbol='RPSAP68', Aliases='', description='ribosomal protein SA pseudogene 68', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8603459, end_position_on_the_genomic_accession=8603990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996521, CurrentID=0, Status='live', Symbol='LOC100996521', Aliases='', description='uncharacterized LOC100996521', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150887253, end_position_on_the_genomic_accession=150912512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  131408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131408, CurrentID=0, Status='live', Symbol='FAM131A', Aliases='C3orf40, FLAT715, PRO1378', description='family with sequence similarity 131 member A', other_designations='family with sequence similarity 131, member A', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184335929, end_position_on_the_genomic_accession=184346275, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105371891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371891, CurrentID=0, Status='live', Symbol='LOC105371891', Aliases='', description='uncharacterized LOC105371891', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74312100, end_position_on_the_genomic_accession=74317239, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317684, CurrentID=0, Status='live', Symbol='MAPK6PS3', Aliases='', description='mitogen-activated protein kinase 6 pseudogene 3', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42067717, end_position_on_the_genomic_accession=42071878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996538, CurrentID=0, Status='live', Symbol='CCDC74BP1', Aliases='', description='coiled-coil domain containing 74B pseudogene 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20590393, end_position_on_the_genomic_accession=20592352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996539, CurrentID=0, Status='live', Symbol='ITPR1-AS1', Aliases='', description='ITPR1 antisense RNA 1 (head to head)', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4490317, end_position_on_the_genomic_accession=4493163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996541, CurrentID=0, Status='live', Symbol='FAM157C', Aliases='', description='family with sequence similarity 157 member C', other_designations='family with sequence similarity 157, member C', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90102264, end_position_on_the_genomic_accession=90177606, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371893, CurrentID=0, Status='live', Symbol='LOC105371893', Aliases='', description='uncharacterized LOC105371893', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74747017, end_position_on_the_genomic_accession=74748912, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  202181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202181, CurrentID=0, Status='live', Symbol='LOC202181', Aliases='', description='SUMO-interacting motifs containing 1 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177618500, end_position_on_the_genomic_accession=177672277, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371895, CurrentID=0, Status='live', Symbol='LOC105371895', Aliases='', description='uncharacterized LOC105371895', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75114728, end_position_on_the_genomic_accession=75119443, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996556, CurrentID=0, Status='live', Symbol='CA15P3', Aliases='', description='carbonic anhydrase XV, pseudogene 3', other_designations='carbonic anhydrase 15-like', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=21708005, end_position_on_the_genomic_accession=21714771, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371896, CurrentID=0, Status='live', Symbol='LOC105371896', Aliases='', description='uncharacterized LOC105371896', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76243188, end_position_on_the_genomic_accession=76263848, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100996564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996564, CurrentID=0, Status='live', Symbol='SSXP1', Aliases='psiSSX1', description='SSX family pseudogene 1', other_designations='SSX1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52604982, end_position_on_the_genomic_accession=52611758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996567, CurrentID=0, Status='live', Symbol='ANTXRLP1', Aliases='', description='anthrax toxin receptor-like pseudogene 1', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46219025, end_position_on_the_genomic_accession=46269608, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100996569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996569, CurrentID=0, Status='live', Symbol='NAMA', Aliases='', description='non-protein coding RNA, associated with MAP kinase pathway and growth arrest', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99355340, end_position_on_the_genomic_accession=99375257, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100996570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996570, CurrentID=0, Status='live', Symbol='NFIA-AS2', Aliases='', description='NFIA antisense RNA 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60940244, end_position_on_the_genomic_accession=60970776, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100996571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996571, CurrentID=0, Status='live', Symbol='CYYR1-AS1', Aliases='', description='cysteine/tyrosine-rich 1 antisense RNA 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26405943, end_position_on_the_genomic_accession=26569263, orientation='plus', exon_count=5, OMIM=616021, no_hdr0=''),
  100996573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996573, CurrentID=0, Status='live', Symbol='LOC100996573', Aliases='', description='uncharacterized LOC100996573', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132550729, end_position_on_the_genomic_accession=132554890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996574, CurrentID=0, Status='live', Symbol='LOC100996574', Aliases='', description='uncharacterized LOC100996574', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133058029, end_position_on_the_genomic_accession=133061567, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  55547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55547, CurrentID=0, Status='live', Symbol='HAB1', Aliases='', description='B1 for mucin', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996582, CurrentID=0, Status='live', Symbol='FAM74A7', Aliases='', description='family with sequence similarity 74 member A7', other_designations='family with sequence similarity 74, member A7|family with sequence similarity 74, member A7, pseudogene', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42198862, end_position_on_the_genomic_accession=42206016, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996583, CurrentID=0, Status='live', Symbol='LOC100996583', Aliases='', description='uncharacterized LOC100996583', other_designations='', map_location='1p356.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2566535, end_position_on_the_genomic_accession=2584534, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996585, CurrentID=0, Status='live', Symbol='LOC100996585', Aliases='', description='uncharacterized LOC100996585', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67380664, end_position_on_the_genomic_accession=67381548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996586, CurrentID=0, Status='live', Symbol='LOC100996586', Aliases='', description='Ig lambda chain V region 4A-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47202431, end_position_on_the_genomic_accession=47202897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996590, CurrentID=0, Status='live', Symbol='LINC01505', Aliases='', description='long intergenic non-protein coding RNA 1505', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106616058, end_position_on_the_genomic_accession=106679802, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371901, CurrentID=0, Status='live', Symbol='LOC105371901', Aliases='', description='uncharacterized LOC105371901', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77123966, end_position_on_the_genomic_accession=77141439, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  202227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202227, CurrentID=0, Status='live', Symbol='LOC202227', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66572020, end_position_on_the_genomic_accession=66572749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996597, CurrentID=0, Status='live', Symbol='LINC00690', Aliases='', description='long intergenic non-protein coding RNA 690', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16536318, end_position_on_the_genomic_accession=16541373, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996598, CurrentID=0, Status='live', Symbol='LOC100996598', Aliases='', description='uncharacterized LOC100996598', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41369082, end_position_on_the_genomic_accession=41370396, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996600, CurrentID=0, Status='live', Symbol='IGKV2OR2-8', Aliases='IGKV2/OR2-8, IGKV2OR28', description='immunoglobulin kappa variable 2/OR2-8 (pseudogene)', other_designations='Ig kappa chain V-II region RPMI 6410-like|immunoglobulin kappa variable 2/OR2-8 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97331411, end_position_on_the_genomic_accession=97332505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371903, CurrentID=0, Status='live', Symbol='LOC105371903', Aliases='', description='uncharacterized LOC105371903', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77460092, end_position_on_the_genomic_accession=77461042, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928609, CurrentID=0, Status='live', Symbol='LOC101928609', Aliases='', description='uncharacterized LOC101928609', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105198637, end_position_on_the_genomic_accession=105245168, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374983, CurrentID=0, Status='live', Symbol='LOC105374983', Aliases='', description='uncharacterized LOC105374983', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25682024, end_position_on_the_genomic_accession=25703140, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100996609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996609, CurrentID=0, Status='live', Symbol='LINC01436', Aliases='', description='long intergenic non-protein coding RNA 1436', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36005320, end_position_on_the_genomic_accession=36008295, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  202243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202243, CurrentID=0, Status='live', Symbol='CCDC125', Aliases='KENAE', description='coiled-coil domain containing 125', other_designations='Kenae', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69273101, end_position_on_the_genomic_accession=69332852, orientation='minus', exon_count=17, OMIM=613781, no_hdr0=''),
  100131520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131520, CurrentID=0, Status='live', Symbol='LOC100131520', Aliases='', description='ribosomal protein L6 pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=155111152, end_position_on_the_genomic_accession=155112341, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996624, CurrentID=0, Status='live', Symbol='LOC100996624', Aliases='', description='uncharacterized LOC100996624', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27830888, end_position_on_the_genomic_accession=27834136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371907, CurrentID=0, Status='live', Symbol='LOC105371907', Aliases='', description='uncharacterized LOC105371907', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77257737, end_position_on_the_genomic_accession=77282699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996629, CurrentID=0, Status='live', Symbol='LINC00849', Aliases='', description='long intergenic non-protein coding RNA 849', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=70236918, end_position_on_the_genomic_accession=70240681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996630, CurrentID=0, Status='live', Symbol='LOC100996630', Aliases='', description='uncharacterized LOC100996630', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95310928, end_position_on_the_genomic_accession=95318282, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996631, CurrentID=0, Status='live', Symbol='TCP11X1', Aliases='', description='t-complex 11 family, X-linked 1', other_designations='T-complex protein 11 homolog', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102215256, end_position_on_the_genomic_accession=102248522, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105371908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371908, CurrentID=0, Status='live', Symbol='LOC105371908', Aliases='', description='uncharacterized LOC105371908', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77797656, end_position_on_the_genomic_accession=77840697, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100996634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996634, CurrentID=0, Status='live', Symbol='LOC100996634', Aliases='', description='transmembrane protein FLJ37396', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109217301, end_position_on_the_genomic_accession=109285010, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100996635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996635, CurrentID=0, Status='live', Symbol='LOC100996635', Aliases='', description='uncharacterized LOC100996635', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95510116, end_position_on_the_genomic_accession=95515464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996637, CurrentID=0, Status='live', Symbol='LOC100996637', Aliases='', description='P-selectin glycoprotein ligand 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=494739, end_position_on_the_genomic_accession=496631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371909, CurrentID=0, Status='live', Symbol='LOC105371909', Aliases='', description='uncharacterized LOC105371909', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77926230, end_position_on_the_genomic_accession=77935348, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996643, CurrentID=0, Status='live', Symbol='LOC100996643', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63334547, end_position_on_the_genomic_accession=63378132, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  317702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317702, CurrentID=0, Status='live', Symbol='VN1R3', Aliases='FKSG46, V1RL3, V1RL3p', description='vomeronasal 1 receptor 3 (gene/pseudogene)', other_designations='V1R-like 3|V1RL3 pseudogene|vomeronasal 1 receptor 3 pseudogene|vomeronasal 1 receptor 4 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31807590, end_position_on_the_genomic_accession=31808846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996648, CurrentID=0, Status='live', Symbol='TCP11X2', Aliases='', description='t-complex 11 family, X-linked 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102460319, end_position_on_the_genomic_accession=102471812, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100996649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996649, CurrentID=0, Status='live', Symbol='ALG1L14P', Aliases='', description='asparagine-linked glycosylation 1-like 14, pseudogene', other_designations='putative glycosyltransferase ALG1-like', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9166116, end_position_on_the_genomic_accession=9170298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55559, CurrentID=0, Status='live', Symbol='HAUS7', Aliases='UCHL5IP, UIP1', description='HAUS augmin like complex subunit 7', other_designations='26S proteasome-associated UCH37-interacting protein 1|HAUS augmin-like complex, subunit 7|UCHL5 interacting protein|X-linked protein STS1769', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153447665, end_position_on_the_genomic_accession=153495525, orientation='minus', exon_count=14, OMIM=300540, no_hdr0=''),
  100996654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996654, CurrentID=0, Status='live', Symbol='LOC100996654', Aliases='', description='uncharacterized LOC100996654', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54759313, end_position_on_the_genomic_accession=54804969, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371912, CurrentID=0, Status='live', Symbol='LOC105371912', Aliases='', description='uncharacterized LOC105371912', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78342113, end_position_on_the_genomic_accession=78346377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996660, CurrentID=0, Status='live', Symbol='LOC100996660', Aliases='', description='uncharacterized LOC100996660', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60126535, end_position_on_the_genomic_accession=60135748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100996662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996662, CurrentID=0, Status='live', Symbol='LOC100996662', Aliases='', description='uncharacterized LOC100996662', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144700394, end_position_on_the_genomic_accession=144709597, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317705, CurrentID=0, Status='live', Symbol='VN1R5', Aliases='V1RL5', description='vomeronasal 1 receptor 5 (gene/pseudogene)', other_designations='G-protein coupled receptor GPCR26|V1R-like 5|V1r-like receptor 5|hGPCR26|pheromone receptor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247256072, end_position_on_the_genomic_accession=247257145, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996664, CurrentID=0, Status='live', Symbol='LOC100996664', Aliases='', description='uncharacterized LOC100996664', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68605297, end_position_on_the_genomic_accession=68686949, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100996665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996665, CurrentID=0, Status='live', Symbol='LINC01533', Aliases='CTC-360P9.3', description='long intergenic non-protein coding RNA 1533', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32025862, end_position_on_the_genomic_accession=32048899, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  202299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202299, CurrentID=0, Status='live', Symbol='LINC01554', Aliases='C5orf27, FIS', description='long intergenic non-protein coding RNA 1554', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95852232, end_position_on_the_genomic_accession=95860133, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996669, CurrentID=0, Status='live', Symbol='LINC01544', Aliases='', description='long intergenic non-protein coding RNA 1544', other_designations='TCONS_00026194', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61748176, end_position_on_the_genomic_accession=61754695, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996670, CurrentID=0, Status='live', Symbol='LOC100996670', Aliases='', description='zinc finger, FYVE domain containing 9 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28762802, end_position_on_the_genomic_accession=28764142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996671, CurrentID=0, Status='live', Symbol='LOC100996671', Aliases='', description='uncharacterized LOC100996671', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126652946, end_position_on_the_genomic_accession=126690318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371915, CurrentID=0, Status='live', Symbol='LOC105371915', Aliases='', description='uncharacterized LOC105371915', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78985913, end_position_on_the_genomic_accession=78993240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  202309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202309, CurrentID=0, Status='live', Symbol='GAPT', Aliases='C5orf29', description='GRB2-binding adaptor protein, transmembrane', other_designations='GRB2-binding transmembrane adaptor|growth factor receptor-bound protein 2-binding adapter protein, transmembrane', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58491435, end_position_on_the_genomic_accession=58496358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996679, CurrentID=0, Status='live', Symbol='LOC100996679', Aliases='', description='uncharacterized LOC100996679', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128086982, end_position_on_the_genomic_accession=128118141, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100996681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996681, CurrentID=0, Status='live', Symbol='LOC100996681', Aliases='', description='uncharacterized LOC100996681', other_designations='CTD-2329C7.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33553113, end_position_on_the_genomic_accession=33557506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996684, CurrentID=0, Status='live', Symbol='AKR1C5P', Aliases='', description='aldo-keto reductase family 1, member C5, pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=5161924, end_position_on_the_genomic_accession=5177418, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996686, CurrentID=0, Status='live', Symbol='METTL7AP1', Aliases='', description='methyltransferase like 7A pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52229005, end_position_on_the_genomic_accession=52233281, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129037, CurrentID=0, Status='live', Symbol='LOC100129037', Aliases='', description='WEE1 homolog (S. pombe) pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105960, end_position_on_the_genomic_accession=107757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996693, CurrentID=0, Status='live', Symbol='LOC100996693', Aliases='', description='CAVP-target protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219494300, end_position_on_the_genomic_accession=219498287, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100996694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996694, CurrentID=0, Status='live', Symbol='LOC100996694', Aliases='', description='uncharacterized LOC100996694', other_designations='', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119939540, end_position_on_the_genomic_accession=119964293, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996696, CurrentID=0, Status='live', Symbol='LOC100996696', Aliases='', description='polyadenylate-binding protein 4 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=59812079, end_position_on_the_genomic_accession=59812618, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927015, CurrentID=0, Status='live', Symbol='LINC01506', Aliases='', description='long intergenic non-protein coding RNA 1506', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68543541, end_position_on_the_genomic_accession=68546602, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  202333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202333, CurrentID=0, Status='live', Symbol='CMYA5', Aliases='C5orf10, SPRYD2, TRIM76', description='cardiomyopathy associated 5', other_designations='2310076E16Rik|SPRY domain-containing protein 2|dystrobrevin-binding protein 2|genethonin-3|myospryn|tripartite motif-containing 76|tripartite motif-containing protein 76', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79689836, end_position_on_the_genomic_accession=79800226, orientation='plus', exon_count=13, OMIM=612193, no_hdr0=''),
  100996702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996702, CurrentID=0, Status='live', Symbol='LINC01356', Aliases='', description='long intergenic non-protein coding RNA 1356', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112820169, end_position_on_the_genomic_accession=112850643, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100996703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996703, CurrentID=0, Status='live', Symbol='LOC100996703', Aliases='', description='uncharacterized protein C2orf27 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371920, CurrentID=0, Status='live', Symbol='LOC105371920', Aliases='', description='uncharacterized LOC105371920', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80200673, end_position_on_the_genomic_accession=80204262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996709, CurrentID=0, Status='live', Symbol='LOC100996709', Aliases='', description='ADP-ribosylation factor-like protein 17', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996712, CurrentID=0, Status='live', Symbol='SRGAP2D', Aliases='', description='SLIT-ROBO Rho GTPase activating protein 2D (pseudogene)', other_designations='SLIT-ROBO Rho GTPase activating protein 2D pseudogene|SLIT-ROBO Rho GTPase-activating protein 2-like|SLIT-ROBO Rho GTPase-activating protein 2D', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143972639, end_position_on_the_genomic_accession=144069704, orientation='plus', exon_count=8, OMIM=614705, no_hdr0=''),
  100996713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996713, CurrentID=0, Status='live', Symbol='LOC100996713', Aliases='', description='uncharacterized LOC100996713', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143883904, end_position_on_the_genomic_accession=143885954, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421256, CurrentID=0, Status='live', Symbol='LOC100421256', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140039758, end_position_on_the_genomic_accession=140040273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996716, CurrentID=0, Status='live', Symbol='LOC100996716', Aliases='', description='vesicle-trafficking protein SEC22b-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149338438, end_position_on_the_genomic_accession=149362499, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100996717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996717, CurrentID=0, Status='live', Symbol='LOC100996717', Aliases='', description='uncharacterized LOC100996717', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149390593, end_position_on_the_genomic_accession=149414468, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996720, CurrentID=0, Status='live', Symbol='LOC100996720', Aliases='', description='uncharacterized LOC100996720', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996721, CurrentID=0, Status='live', Symbol='LOC100996721', Aliases='', description='protein FAM231D', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121012216, end_position_on_the_genomic_accession=121013923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996722, CurrentID=0, Status='live', Symbol='LOC100996722', Aliases='', description='ubiquitin-conjugating enzyme E2 D4 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121058944, end_position_on_the_genomic_accession=121059353, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996723, CurrentID=0, Status='live', Symbol='LOC100996723', Aliases='', description='uncharacterized LOC100996723', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120197070, end_position_on_the_genomic_accession=120320122, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996724, CurrentID=0, Status='live', Symbol='LOC100996724', Aliases='', description='phosphodiesterase 4D interacting protein-like', other_designations='myomegalin-like', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120488844, end_position_on_the_genomic_accession=120701360, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  100996727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996727, CurrentID=0, Status='live', Symbol='LOC100996727', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60731993, end_position_on_the_genomic_accession=60732636, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317716, CurrentID=0, Status='live', Symbol='BPIFA4P', Aliases='BASE', description='BPI fold containing family A member 4, pseudogene', other_designations='BPI fold containing family A, member 4, pseudogene|PLUNC family pseudogene|breast cancer and salivary gland expression|gene expressed only in the salivary gland and breast cancers', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33193605, end_position_on_the_genomic_accession=33210462, orientation='plus', exon_count=9, OMIM=607627, no_hdr0=''),
  100996731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996731, CurrentID=0, Status='live', Symbol='LOC100996731', Aliases='', description='otopetrin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144179807, end_position_on_the_genomic_accession=144213226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996732, CurrentID=0, Status='live', Symbol='LOC100996732', Aliases='', description='uncharacterized LOC100996732', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144208089, end_position_on_the_genomic_accession=144250332, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105371925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371925, CurrentID=0, Status='live', Symbol='LOC105371925', Aliases='', description='uncharacterized LOC105371925', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81229483, end_position_on_the_genomic_accession=81234029, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100996736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996736, CurrentID=0, Status='live', Symbol='LOC100996736', Aliases='', description='otopetrin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143354221, end_position_on_the_genomic_accession=143387619, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996737, CurrentID=0, Status='live', Symbol='LOC100996737', Aliases='', description='otopetrin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144753165, end_position_on_the_genomic_accession=144786581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100996738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996738, CurrentID=0, Status='live', Symbol='HERC2P11', Aliases='', description='hect domain and RLD 2 pseudogene 11', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28574940, end_position_on_the_genomic_accession=28589267, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100996740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996740, CurrentID=0, Status='live', Symbol='LOC100996740', Aliases='', description='uncharacterized LOC100996740', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145214738, end_position_on_the_genomic_accession=145215769, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371926, CurrentID=0, Status='live', Symbol='LOC105371926', Aliases='', description='uncharacterized LOC105371926', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81452858, end_position_on_the_genomic_accession=81455260, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  202374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202374, CurrentID=0, Status='live', Symbol='STK32A', Aliases='YANK1', description='serine/threonine kinase 32A', other_designations='A930015B13Rik|yet another novel kinase 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147234963, end_position_on_the_genomic_accession=147401533, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100996745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996745, CurrentID=0, Status='live', Symbol='LOC100996745', Aliases='', description='dead end protein homolog 1 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996746, CurrentID=0, Status='live', Symbol='SPDYE11', Aliases='', description='speedy/RINGO cell cycle regulator family member E11', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73049244, end_position_on_the_genomic_accession=73057172, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100996747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996747, CurrentID=0, Status='live', Symbol='LOC100996747', Aliases='', description='40S ribosomal protein S26', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100996750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996750, CurrentID=0, Status='live', Symbol='LOC100996750', Aliases='', description='keratin-associated protein 4-7', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129048, CurrentID=0, Status='live', Symbol='LOC100129048', Aliases='', description='uncharacterized LOC100129048', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=7370049, end_position_on_the_genomic_accession=7370661, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100996756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996756, CurrentID=0, Status='live', Symbol='LOC100996756', Aliases='', description='uncharacterized LOC100996756', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146387141, end_position_on_the_genomic_accession=146388091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100996758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996758, CurrentID=0, Status='live', Symbol='CH17-360D5.1', Aliases='', description='neuropeptide Y receptor type 4-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47914012, end_position_on_the_genomic_accession=47926069, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101927017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927017, CurrentID=0, Status='live', Symbol='LINC01180', Aliases='', description='long intergenic non-protein coding RNA 1180', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44646364, end_position_on_the_genomic_accession=44650349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996763, CurrentID=0, Status='live', Symbol='LOC100996763', Aliases='', description='notch homolog 2 N-terminal-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148595857, end_position_on_the_genomic_accession=148712520, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100129050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129050, CurrentID=0, Status='live', Symbol='CICP20', Aliases='', description='capicua transcriptional repressor pseudogene 20', other_designations='capicua homolog pseudogene', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45815326, end_position_on_the_genomic_accession=45819058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927073, CurrentID=0, Status='live', Symbol='LOC101927073', Aliases='', description='uncharacterized LOC101927073', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179655871, end_position_on_the_genomic_accession=179679990, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100996770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996770, CurrentID=0, Status='live', Symbol='LOC100996770', Aliases='', description='uncharacterized LOC100996770', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65687656, end_position_on_the_genomic_accession=65692177, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421258, CurrentID=0, Status='live', Symbol='LOC100421258', Aliases='', description='NOP58 ribonucleoprotein homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=60024588, end_position_on_the_genomic_accession=60025141, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129052, CurrentID=0, Status='live', Symbol='LOC100129052', Aliases='', description='N(alpha)-acetyltransferase 40, NatD catalytic subunit, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4189068, end_position_on_the_genomic_accession=4190008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129053, CurrentID=0, Status='live', Symbol='LOC100129053', Aliases='', description='lysine (K)-specific demethylase 4E pseudogene', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95049080, end_position_on_the_genomic_accession=95051457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  317726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317726, CurrentID=0, Status='live', Symbol='HMGN2P2', Aliases='HMGN2L2', description='high mobility group nucleosomal binding domain 2 pseudogene 2', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 2|high-mobility group nucleosomal binding domain 2-like 2', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80822429, end_position_on_the_genomic_accession=80823606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100996792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996792, CurrentID=0, Status='live', Symbol='LOC100996792', Aliases='', description='dual specificity mitogen-activated protein kinase kinase 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21773015, end_position_on_the_genomic_accession=21783473, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129055, CurrentID=0, Status='live', Symbol='LOC100129055', Aliases='', description='cyclin Y like 1 pseudogene', other_designations='cyclin Y-like 1 pseudogene', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38175671, end_position_on_the_genomic_accession=38214345, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  103689845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103689845, CurrentID=0, Status='live', Symbol='LINC01050', Aliases='', description='long intergenic non-protein coding RNA 1050', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42810366, end_position_on_the_genomic_accession=42812562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317729, CurrentID=0, Status='live', Symbol='BLZF2P', Aliases='', description='basic leucine zipper nuclear factor 2 pseudogene', other_designations='basic leucine zipper nuclear factor 1 pseudogene', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68866658, end_position_on_the_genomic_accession=68870110, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317730, CurrentID=0, Status='live', Symbol='C14orf99', Aliases='', description='uncharacterized LOC317730', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106809028, end_position_on_the_genomic_accession=106810171, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371939, CurrentID=0, Status='live', Symbol='LOC105371939', Aliases='', description='uncharacterized LOC105371939', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81975968, end_position_on_the_genomic_accession=81977429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129060, CurrentID=0, Status='live', Symbol='SEMA3F-AS1', Aliases='RBM5-AS1', description='SEMA3F antisense RNA 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=50159685, end_position_on_the_genomic_accession=50193518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  202459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202459, CurrentID=0, Status='live', Symbol='OSTCP1', Aliases='DC2L, OSTCL', description='oligosaccharyltransferase complex subunit pseudogene 1', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158841117, end_position_on_the_genomic_accession=158857632, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129061, CurrentID=0, Status='live', Symbol='RPL7P25', Aliases='RPL7_12_684', description='ribosomal protein L7 pseudogene 25', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34616522, end_position_on_the_genomic_accession=34617324, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371942, CurrentID=0, Status='live', Symbol='LOC105371942', Aliases='', description='uncharacterized LOC105371942', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82582754, end_position_on_the_genomic_accession=82588712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100996842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996842, CurrentID=0, Status='live', Symbol='LOC100996842', Aliases='', description='uncharacterized LOC100996842', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7582033, end_position_on_the_genomic_accession=7584784, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645415, CurrentID=0, Status='live', Symbol='LOC645415', Aliases='', description='MAS-related GPR, member X1 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18956045, end_position_on_the_genomic_accession=18957232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927020, CurrentID=0, Status='live', Symbol='LOC101927020', Aliases='', description='uncharacterized LOC101927020', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49811540, end_position_on_the_genomic_accession=49820503, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100996851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996851, CurrentID=0, Status='live', Symbol='LOC100996851', Aliases='', description='E3 ubiquitin-protein ligase XIAP pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109661814, end_position_on_the_genomic_accession=109681201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371945, CurrentID=0, Status='live', Symbol='LOC105371945', Aliases='', description='uncharacterized LOC105371945', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82848862, end_position_on_the_genomic_accession=82849945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996860, CurrentID=0, Status='live', Symbol='LOC100996860', Aliases='', description='TAR DNA-binding protein 43 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=102453146, end_position_on_the_genomic_accession=102494365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129066, CurrentID=0, Status='live', Symbol='UNQ6494', Aliases='', description='uncharacterized LOC100129066', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89639783, end_position_on_the_genomic_accession=89719759, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  121131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121131, CurrentID=0, Status='live', Symbol='PSMB3P', Aliases='', description='proteasome (prosome, macropain) subunit, beta type, 3 pseudogene', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55654534, end_position_on_the_genomic_accession=55655234, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  202500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202500, CurrentID=0, Status='live', Symbol='TCTE1', Aliases='D6S46, FAP155', description='t-complex-associated-testis-expressed 1', other_designations='tcte-1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44280160, end_position_on_the_genomic_accession=44297721, orientation='minus', exon_count=8, OMIM=186975, no_hdr0=''),
  100996876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996876, CurrentID=0, Status='live', Symbol='RORA-AS2', Aliases='', description='RORA antisense RNA 2', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60681569, end_position_on_the_genomic_accession=60687249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927021, CurrentID=0, Status='live', Symbol='LOC101927021', Aliases='', description='uncharacterized LOC101927021', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1160374, end_position_on_the_genomic_accession=1165958, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371950, CurrentID=0, Status='live', Symbol='LOC105371950', Aliases='', description='uncharacterized LOC105371950', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=43088, end_position_on_the_genomic_accession=47016, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100996886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996886, CurrentID=0, Status='live', Symbol='LOC100996886', Aliases='', description='complement factor H-related protein 3-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196850249, end_position_on_the_genomic_accession=196884811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371951, CurrentID=0, Status='live', Symbol='LOC105371951', Aliases='', description='uncharacterized LOC105371951', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=151413, end_position_on_the_genomic_accession=158169, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130649, CurrentID=0, Status='live', Symbol='LOC100130649', Aliases='', description='G protein-coupled receptor 160 pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85685936, end_position_on_the_genomic_accession=85687192, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129072, CurrentID=0, Status='live', Symbol='LOC100129072', Aliases='', description='uncharacterized LOC100129072', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=99057338, end_position_on_the_genomic_accession=99057760, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100996902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996902, CurrentID=0, Status='live', Symbol='LOC100996902', Aliases='', description='uncharacterized LOC100996902', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38431186, end_position_on_the_genomic_accession=38433573, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371953, CurrentID=0, Status='live', Symbol='LOC105371953', Aliases='', description='uncharacterized LOC105371953', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1099260, end_position_on_the_genomic_accession=1183954, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371954, CurrentID=0, Status='live', Symbol='LINC00470', Aliases='', description='long intergenic non-protein coding RNA 470', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1365917, end_position_on_the_genomic_accession=1392185, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100996911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996911, CurrentID=0, Status='live', Symbol='LOC100996911', Aliases='', description='putative ubiquitin-conjugating enzyme E2Q2-like protein', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24196515, end_position_on_the_genomic_accession=24206155, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645427, CurrentID=0, Status='live', Symbol='LOC645427', Aliases='', description='uncharacterized LOC645427', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=21256376, end_position_on_the_genomic_accession=21256753, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  202559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202559, CurrentID=0, Status='live', Symbol='KHDRBS2', Aliases='SLM-1, SLM1, bA535F17.1', description='KH domain containing, RNA binding, signal transduction associated 2', other_designations='Sam68-like mammalian protein 1|hSLM-1', map_location='6q11.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=61630484, end_position_on_the_genomic_accession=62286195, orientation='minus', exon_count=21, OMIM=610487, no_hdr0=''),
  100996928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996928, CurrentID=0, Status='live', Symbol='C7orf55-LUC7L2', Aliases='', description='C7orf55-LUC7L2 readthrough', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139341132, end_position_on_the_genomic_accession=139423457, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100996930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996930, CurrentID=0, Status='live', Symbol='LINC00621', Aliases='', description='long intergenic non-protein coding RNA 621', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22859875, end_position_on_the_genomic_accession=22896401, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371958, CurrentID=0, Status='live', Symbol='LOC105371958', Aliases='', description='uncharacterized LOC105371958', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1906016, end_position_on_the_genomic_accession=1923073, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100996934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996934, CurrentID=0, Status='live', Symbol='UAQTL5', Aliases='', description='Uric acid concentration, serum, QTL5', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614746, no_hdr0=''),
  100996935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996935, CurrentID=0, Status='live', Symbol='UAQTL6', Aliases='', description='Uric acid concentration, serum, QTL6', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614747, no_hdr0=''),
  100996939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996939, CurrentID=0, Status='live', Symbol='PYURF', Aliases='PREY', description='PIGY upstream reading frame', other_designations='PIGY upstream reading frame protein', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88520978, end_position_on_the_genomic_accession=88523801, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100996941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996941, CurrentID=0, Status='live', Symbol='IL21-AS1', Aliases='', description='IL21 antisense RNA 1', other_designations='IL21 antisense RNA 1 (non-protein coding)', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122618983, end_position_on_the_genomic_accession=122689156, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100996944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996944, CurrentID=0, Status='live', Symbol='NDUFS5P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 1', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136475871, end_position_on_the_genomic_accession=136476150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100653368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653368, CurrentID=0, Status='live', Symbol='SCA32', Aliases='', description='spinocerebellar ataxia 32', other_designations='', map_location='7q32-q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613909, no_hdr0=''),
  100996949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100996949, CurrentID=0, Status='live', Symbol='PDA1', Aliases='', description='Patent ductus arteriosus, susceptibility to', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607411, no_hdr0=''),
  317753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317753, CurrentID=0, Status='live', Symbol='PSMD12P', Aliases='', description='proteasome 26S subunit, non-ATPase, 12 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 12 pseudogene|proteasome 26S non-ATPase subunit 12 pseudogene', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72876669, end_position_on_the_genomic_accession=72878186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55610, CurrentID=0, Status='live', Symbol='VPS50', Aliases='CCDC132, VPS54L', description='VPS50 EARP/GARPII complex subunit', other_designations='EARP/GARPII complex subunit VPS50|coiled-coil domain containing 132|coiled-coil domain-containing protein 132', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93232340, end_position_on_the_genomic_accession=93361123, orientation='plus', exon_count=30, OMIM=616465, no_hdr0=''),
  100129083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129083, CurrentID=0, Status='live', Symbol='LOC100129083', Aliases='', description='uncharacterized LOC100129083', other_designations='CTD-2616J11.2', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51415906, end_position_on_the_genomic_accession=51417425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419892, CurrentID=0, Status='live', Symbol='LOC100419892', Aliases='', description='zinc finger protein 592 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3580160, end_position_on_the_genomic_accession=3580756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645405, CurrentID=0, Status='live', Symbol='LOC645405', Aliases='', description='oxidation resistance 1 pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48725345, end_position_on_the_genomic_accession=48725828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645436, CurrentID=0, Status='live', Symbol='HNRNPA1P63', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 63', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54536740, end_position_on_the_genomic_accession=54537951, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371965, CurrentID=0, Status='live', Symbol='LOC105371965', Aliases='', description='uncharacterized LOC105371965', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3461844, end_position_on_the_genomic_accession=3481678, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129086, CurrentID=0, Status='live', Symbol='HIGD1AP1', Aliases='HIG1, HIGD1D, HIGD1DP', description='HIG1 hypoxia inducible domain family member 1A pseudogene 1', other_designations='HIG1 domain family, member 1D|HIG1 hypoxia inducible domain family, member 1A pseudogene 1|HIG1 hypoxia inducible domain family, member 1D, pseudogene', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53141960, end_position_on_the_genomic_accession=53143304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55615, CurrentID=0, Status='live', Symbol='PRR5', Aliases='FLJ20185k, PP610, PROTOR-1, PROTOR1', description='proline rich 5', other_designations='Rho GTPase activating protein 8|proline rich 5 (renal)|protein observed with Rictor-1', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44668547, end_position_on_the_genomic_accession=44737681, orientation='plus', exon_count=11, OMIM=609406, no_hdr0=''),
  100419579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419579, CurrentID=0, Status='live', Symbol='LOC100419579', Aliases='', description='HECT and RLD domain containing E3 ubiquitin protein ligase 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22584966, end_position_on_the_genomic_accession=22587799, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317760, CurrentID=0, Status='live', Symbol='ADAM20P1', Aliases='C14orf55', description='ADAM metallopeptidase domain 20 pseudogene 1', other_designations='CTD-2540L5.4|metalloproteinase-disintegrin 1-2 pseudogene', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70468876, end_position_on_the_genomic_accession=70483756, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100270801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270801, CurrentID=0, Status='live', Symbol='STQTL15', Aliases='', description='Stature QTL 15', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612578, no_hdr0=''),
  105374996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374996, CurrentID=0, Status='live', Symbol='LOC105374996', Aliases='', description='uncharacterized LOC105374996', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28638794, end_position_on_the_genomic_accession=28653234, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  55617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55617, CurrentID=0, Status='live', Symbol='TASP1', Aliases='C20orf13, dJ585I14.2', description='taspase 1', other_designations='taspase, threonine aspartase, 1|taspase-1', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13389389, end_position_on_the_genomic_accession=13638936, orientation='minus', exon_count=20, OMIM=608270, no_hdr0=''),
  448834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448834, CurrentID=0, Status='live', Symbol='KPRP', Aliases='C1orf45', description='keratinocyte proline-rich protein', other_designations='hKPRP|keratinocyte expressed, proline-rich protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152758030, end_position_on_the_genomic_accession=152762053, orientation='plus', exon_count=3, OMIM=613260, no_hdr0=''),
  55619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55619, CurrentID=0, Status='live', Symbol='DOCK10', Aliases='DRIP2, Nbla10300, ZIZ3', description='dedicator of cytokinesis 10', other_designations='dopamine receptor interacting protein 2|zizimin3', map_location='2q36.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224765090, end_position_on_the_genomic_accession=225042613, orientation='minus', exon_count=60, OMIM=611518, no_hdr0=''),
  105371972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371972, CurrentID=0, Status='live', Symbol='LOC105371972', Aliases='', description='uncharacterized LOC105371972', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6558573, end_position_on_the_genomic_accession=6569397, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371973, CurrentID=0, Status='live', Symbol='LOC105371973', Aliases='', description='uncharacterized LOC105371973', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7247705, end_position_on_the_genomic_accession=7268603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  202658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202658, CurrentID=0, Status='live', Symbol='TRIM39-RPP21', Aliases='TRIM39R', description='TRIM39-RPP21 readthrough', other_designations='Rpp21 domain-containing TRIM protein|TRIM39-RPP21 read-through transcript', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30329311, end_position_on_the_genomic_accession=30346858, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100129094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129094, CurrentID=0, Status='live', Symbol='BTNL10', Aliases='BTN4, BUTR1', description='butyrophilin-like 10', other_designations='butyrophilin related 1', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228508027, end_position_on_the_genomic_accession=228514017, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317767, CurrentID=0, Status='live', Symbol='SCOCP1', Aliases='', description='short coiled-coil protein pseudogene 1', other_designations='short coiled coil protein pseudogene 1', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60159011, end_position_on_the_genomic_accession=60161425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533929, CurrentID=0, Status='live', Symbol='LOC100533929', Aliases='', description='eukaryotic translation termination factor 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  317768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317768, CurrentID=0, Status='live', Symbol='HOMER2P1', Aliases='', description='homer scaffolding protein 2 pseudogene 1', other_designations='HOMER2 pseudogene 1|homer homolog 2 pseudogene 1', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106458942, end_position_on_the_genomic_accession=106461306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421267, CurrentID=0, Status='live', Symbol='LOC100421267', Aliases='', description='phosphodiesterase 4D interacting protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16548652, end_position_on_the_genomic_accession=16551230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317769, CurrentID=0, Status='live', Symbol='HOMER2P2', Aliases='', description='homer scaffolding protein 2 pseudogene 2', other_designations='HOMER2 pseudogene 2|homer homolog 2 pseudogene 2', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106285995, end_position_on_the_genomic_accession=106287892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129098, CurrentID=0, Status='live', Symbol='LOC100129098', Aliases='', description='serine/arginine repetitive matrix protein 2-like', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54465594, end_position_on_the_genomic_accession=54470426, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927027, CurrentID=0, Status='live', Symbol='LOC101927027', Aliases='', description='uncharacterized LOC101927027', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178413659, end_position_on_the_genomic_accession=178440243, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317772, CurrentID=0, Status='live', Symbol='HIST2H2AB', Aliases='H2AB', description='histone cluster 2, H2ab', other_designations='histone 2, H2ab', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149887469, end_position_on_the_genomic_accession=149887916, orientation='minus', exon_count=1, OMIM=615014, no_hdr0=''),
  55629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55629, CurrentID=0, Status='live', Symbol='PNRC2', Aliases='', description='proline rich nuclear receptor coactivator 2', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23959811, end_position_on_the_genomic_accession=23963462, orientation='plus', exon_count=3, OMIM=611882, no_hdr0=''),
  100129102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129102, CurrentID=0, Status='live', Symbol='RPL22P4', Aliases='RPL22_2_154', description='ribosomal protein L22 pseudogene 4', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206160887, end_position_on_the_genomic_accession=206161276, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317775, CurrentID=0, Status='live', Symbol='CHORDC2P', Aliases='', description='cysteine and histidine rich domain containing 2, pseudogene', other_designations='cysteine and histidine-rich domain (CHORD) containing 2 pseudogene|cysteine and histidine-rich domain (CHORD)-containing 2 pseudogene|cysteine and histidine-rich domain-containing 2 pseudogene', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89734678, end_position_on_the_genomic_accession=89738442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129104, CurrentID=0, Status='live', Symbol='LOC100129104', Aliases='', description='uncharacterized LOC100129104', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=134584414, end_position_on_the_genomic_accession=134586099, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371985, CurrentID=0, Status='live', Symbol='LOC105371985', Aliases='', description='uncharacterized LOC105371985', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10308373, end_position_on_the_genomic_accession=10323653, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79529, CurrentID=0, Status='live', Symbol='OR52J2P', Aliases='OR52J4P', description='olfactory receptor family 52 subfamily J member 2 pseudogene', other_designations='olfactory receptor, family 52, subfamily J, member 2 pseudogene|olfactory receptor, family 52, subfamily J, member 4 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5036916, end_position_on_the_genomic_accession=5038074, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271187, CurrentID=0, Status='live', Symbol='RPL21P84', Aliases='RPL21_32_987', description='ribosomal protein L21 pseudogene 84', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78180180, end_position_on_the_genomic_accession=78180693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  317778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317778, CurrentID=0, Status='live', Symbol='SYF2P1', Aliases='CBPINP, SYF2P', description='SYF2 pre-mRNA-splicing factor pseudogene 1', other_designations='CCNDBP1 interactor pseudogene|SYF2 homolog, RNA splicing factor pseudogene 1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30541750, end_position_on_the_genomic_accession=30542188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104798195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104798195, CurrentID=0, Status='live', Symbol='GS1-594A7.3', Aliases='', description='uncharacterized LOC104798195', other_designations='', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15602881, end_position_on_the_genomic_accession=15621484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317780, CurrentID=0, Status='live', Symbol='CKS1BP1', Aliases='CKS1BP', description='CDC28 protein kinase regulatory subunit 1B pseudogene 1', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96223037, end_position_on_the_genomic_accession=96223830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927029, CurrentID=0, Status='live', Symbol='LOC101927029', Aliases='', description='uncharacterized LOC101927029', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90048134, end_position_on_the_genomic_accession=90061730, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317782, CurrentID=0, Status='live', Symbol='CELIAC2', Aliases='CD, GSE', description='celiac disease 2', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609754, no_hdr0=''),
  100129111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129111, CurrentID=0, Status='live', Symbol='LOC100129111', Aliases='', description='uncharacterized LOC100129111', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=8016351, end_position_on_the_genomic_accession=8016743, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317784, CurrentID=0, Status='live', Symbol='COILP1', Aliases='COILP', description='coilin pseudogene 1', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39384520, end_position_on_the_genomic_accession=39387128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  202781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202781, CurrentID=0, Status='live', Symbol='PAXIP1-AS1', Aliases='', description='PAXIP1 antisense RNA 1 (head to head)', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155003433, end_position_on_the_genomic_accession=155005703, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371995, CurrentID=0, Status='live', Symbol='LOC105371995', Aliases='', description='uncharacterized LOC105371995', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11390033, end_position_on_the_genomic_accession=11391767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  202789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202789, CurrentID=0, Status='live', Symbol='RPL17P27', Aliases='RPL17_17_857', description='ribosomal protein L17 pseudogene 27', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138688939, end_position_on_the_genomic_accession=138689560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2388, CurrentID=0, Status='live', Symbol='FRA4A', Aliases='', description='fragile site, aphidicolin type, common, fra(4)(p16.1)', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371996, CurrentID=0, Status='live', Symbol='LOC105371996', Aliases='', description='uncharacterized LOC105371996', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12254964, end_position_on_the_genomic_accession=12257456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129118, CurrentID=0, Status='live', Symbol='LOC100129118', Aliases='', description='transmembrane protein 167A pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14792265, end_position_on_the_genomic_accession=14792954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533930, CurrentID=0, Status='live', Symbol='LOC100533930', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128849, CurrentID=0, Status='live', Symbol='BNIP3P34', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 34', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22773853, end_position_on_the_genomic_accession=22774421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372002, CurrentID=0, Status='live', Symbol='LOC105372002', Aliases='', description='uncharacterized LOC105372002', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372003, CurrentID=0, Status='live', Symbol='LOC105372003', Aliases='', description='uncharacterized LOC105372003', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14403006, end_position_on_the_genomic_accession=14431347, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372004, CurrentID=0, Status='live', Symbol='LOC105372004', Aliases='', description='uncharacterized LOC105372004', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14555209, end_position_on_the_genomic_accession=14642507, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100129125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129125, CurrentID=0, Status='live', Symbol='LOC100129125', Aliases='', description='uncharacterized LOC100129125', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=19562964, end_position_on_the_genomic_accession=19563826, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372006, CurrentID=0, Status='live', Symbol='LOC105372006', Aliases='', description='uncharacterized LOC105372006', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14711262, end_position_on_the_genomic_accession=14728178, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  333926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=333926, CurrentID=0, Status='live', Symbol='PPM1J', Aliases='PP2CZ, PP2Czeta, PPP2CZ', description='protein phosphatase, Mg2+/Mn2+ dependent 1J', other_designations='PP2C-zeta|protein phosphatase 1J (PP2C domain containing)|protein phosphatase 2C isoform zeta|protein phosphatase 2C zeta|protein phosphatase 2a, catalytic subunit, zeta isoform|protein phosphatase, Mg2+/Mn2+ dependent, 1J', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112709994, end_position_on_the_genomic_accession=112715328, orientation='minus', exon_count=10, OMIM=609957, no_hdr0=''),
  333929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=333929, CurrentID=0, Status='live', Symbol='SNAI3', Aliases='SMUC, SNAIL3, ZNF293, Zfp293', description='snail family zinc finger 3', other_designations='protein snail homolog 3|snail homolog 3|zinc finger protein 293', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88677682, end_position_on_the_genomic_accession=88686474, orientation='minus', exon_count=3, OMIM=612741, no_hdr0=''),
  333932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=333932, CurrentID=0, Status='live', Symbol='HIST2H3A', Aliases='H3/n, H3/o', description='histone cluster 2, H3a', other_designations='histone 2, H3a|histone H3/o', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149852619, end_position_on_the_genomic_accession=149853125, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  202861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202861, CurrentID=0, Status='live', Symbol='OR2A3P', Aliases='OR2A18P', description='olfactory receptor family 2 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 18 pseudogene|olfactory receptor, family 2, subfamily A, member 3 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144157026, end_position_on_the_genomic_accession=144158359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  202865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202865, CurrentID=0, Status='live', Symbol='C7orf33', Aliases='', description='chromosome 7 open reading frame 33', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148590565, end_position_on_the_genomic_accession=148615860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372009, CurrentID=0, Status='live', Symbol='LOC105372009', Aliases='', description='uncharacterized LOC105372009', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14902200, end_position_on_the_genomic_accession=14941129, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100129130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129130, CurrentID=0, Status='live', Symbol='LOC100129130', Aliases='', description='THO complex 4 pseudogene', other_designations='Aly/REF export factor pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75835581, end_position_on_the_genomic_accession=75836141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128851, CurrentID=0, Status='live', Symbol='LOC100128851', Aliases='', description='uncharacterized LOC100128851', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=19163539, end_position_on_the_genomic_accession=19164242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129133, CurrentID=0, Status='live', Symbol='LOC100129133', Aliases='', description='transmembrane protein 184C pseudogene', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=88061566, end_position_on_the_genomic_accession=88062841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927034, CurrentID=0, Status='live', Symbol='ROR1-AS1', Aliases='', description='ROR1 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64094442, end_position_on_the_genomic_accession=64113791, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  390834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390834, CurrentID=0, Status='live', Symbol='FEM1AP2', Aliases='', description='fem-1 homolog a (C. elegans) pseudogene 2', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14391916, end_position_on_the_genomic_accession=14393906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372015, CurrentID=0, Status='live', Symbol='LOC105372015', Aliases='', description='uncharacterized LOC105372015', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6406091, end_position_on_the_genomic_accession=6414772, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  202915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=202915, CurrentID=0, Status='live', Symbol='TMEM184A', Aliases='', description='transmembrane protein 184A', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1542235, end_position_on_the_genomic_accession=1560821, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100129137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129137, CurrentID=0, Status='live', Symbol='LOC100129137', Aliases='', description='CD53 molecule pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88527562, end_position_on_the_genomic_accession=88528167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372018, CurrentID=0, Status='live', Symbol='LOC105372018', Aliases='', description='uncharacterized LOC105372018', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22242199, end_position_on_the_genomic_accession=22267655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129140, CurrentID=0, Status='live', Symbol='LOC100129140', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B12 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 12 pseudogene', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65500716, end_position_on_the_genomic_accession=65502220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372021, CurrentID=0, Status='live', Symbol='LOC105372021', Aliases='', description='uncharacterized LOC105372021', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22609557, end_position_on_the_genomic_accession=22641001, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372023, CurrentID=0, Status='live', Symbol='LOC105372023', Aliases='', description='uncharacterized LOC105372023', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23959975, end_position_on_the_genomic_accession=23967824, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375007, CurrentID=0, Status='live', Symbol='LOC105375007', Aliases='', description='uncharacterized LOC105375007', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29490320, end_position_on_the_genomic_accession=29494236, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129144, CurrentID=0, Status='live', Symbol='LOC100129144', Aliases='', description='zinc finger protein 681 pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=67373362, end_position_on_the_genomic_accession=67375536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129148, CurrentID=0, Status='live', Symbol='LOC100129148', Aliases='', description='uncharacterized LOC100129148', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139417462, end_position_on_the_genomic_accession=139427526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129149, CurrentID=0, Status='live', Symbol='LOC100129149', Aliases='', description='uncharacterized LOC100129149', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=37898114, end_position_on_the_genomic_accession=37899889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372030, CurrentID=0, Status='live', Symbol='LOC105372030', Aliases='', description='uncharacterized LOC105372030', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24932779, end_position_on_the_genomic_accession=24987979, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372031, CurrentID=0, Status='live', Symbol='LOC105372031', Aliases='', description='uncharacterized LOC105372031', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=25233721, end_position_on_the_genomic_accession=25258818, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372034, CurrentID=0, Status='live', Symbol='LOC105372034', Aliases='', description='uncharacterized LOC105372034', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26405858, end_position_on_the_genomic_accession=26409379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928823, CurrentID=0, Status='live', Symbol='LOC101928823', Aliases='', description='uncharacterized LOC101928823', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112482232, end_position_on_the_genomic_accession=112541916, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373280, CurrentID=0, Status='live', Symbol='LOC105373280', Aliases='', description='RNA-binding protein 33-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220875193, end_position_on_the_genomic_accession=220879330, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372036, CurrentID=0, Status='live', Symbol='LOC105372036', Aliases='', description='uncharacterized LOC105372036', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  203054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203054, CurrentID=0, Status='live', Symbol='ADCK5', Aliases='', description='aarF domain containing kinase 5', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144374015, end_position_on_the_genomic_accession=144393242, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105372040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372040, CurrentID=0, Status='live', Symbol='LOC105372040', Aliases='', description='uncharacterized LOC105372040', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27344300, end_position_on_the_genomic_accession=27346111, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  203062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203062, CurrentID=0, Status='live', Symbol='TSNARE1', Aliases='', description='t-SNARE domain containing 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142212080, end_position_on_the_genomic_accession=142403291, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  645513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645513, CurrentID=0, Status='live', Symbol='LOC645513', Aliases='', description='uncharacterized LOC645513', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119454783, end_position_on_the_genomic_accession=119499592, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  203068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203068, CurrentID=0, Status='live', Symbol='TUBB', Aliases='CDCBM6, M40, OK/SW-cl.561, TUBB5, TUBB', description='tubulin beta class I', other_designations='beta Ib tubulin|tubulin beta-1 chain|tubulin beta-5 chain|tubulin, beta class I|tubulin, beta polypeptide', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30720201, end_position_on_the_genomic_accession=30725426, orientation='plus', exon_count=5, OMIM=191130, no_hdr0=''),
  203069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203069, CurrentID=0, Status='live', Symbol='R3HCC1', Aliases='', description='R3H domain and coiled-coil containing 1', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23288092, end_position_on_the_genomic_accession=23296279, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  203074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203074, CurrentID=0, Status='live', Symbol='PRSS55', Aliases='CT153, T-SP1, TSP1, UNQ9391', description='protease, serine 55', other_designations='probable serine protease UNQ9391/PRO34284|protease, serine, 55|testis serine protease 1|tryptophan/serine protease', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10525477, end_position_on_the_genomic_accession=10554166, orientation='plus', exon_count=11, OMIM=615144, no_hdr0=''),
  105372043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372043, CurrentID=0, Status='live', Symbol='LOC105372043', Aliases='', description='uncharacterized LOC105372043', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=28719197, end_position_on_the_genomic_accession=28724767, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  203076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203076, CurrentID=0, Status='live', Symbol='C8orf74', Aliases='', description='chromosome 8 open reading frame 74', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10672592, end_position_on_the_genomic_accession=10700613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927040, CurrentID=0, Status='live', Symbol='LOC101927040', Aliases='', description='uncharacterized LOC101927040', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79769372, end_position_on_the_genomic_accession=79871737, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372046, CurrentID=0, Status='live', Symbol='LOC105372046', Aliases='', description='uncharacterized LOC105372046', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=30317403, end_position_on_the_genomic_accession=30323796, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421281, CurrentID=0, Status='live', Symbol='LOC100421281', Aliases='', description='high density lipoprotein binding protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118712993, end_position_on_the_genomic_accession=118713676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372047, CurrentID=0, Status='live', Symbol='LOC105372047', Aliases='', description='uncharacterized LOC105372047', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=30325739, end_position_on_the_genomic_accession=30333984, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  203100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203100, CurrentID=0, Status='live', Symbol='HTRA4', Aliases='', description='HtrA serine peptidase 4', other_designations='high-temperature requirement factor A4|probable serine protease HTRA4', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38974149, end_position_on_the_genomic_accession=38988662, orientation='plus', exon_count=11, OMIM=610700, no_hdr0=''),
  203102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203102, CurrentID=0, Status='live', Symbol='ADAM32', Aliases='', description='ADAM metallopeptidase domain 32', other_designations='a disintegrin and metalloprotease domain 32|a disintegrin and metalloproteinase domain 32|metalloproteinase 12-like protein', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39107531, end_position_on_the_genomic_accession=39284917, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  339685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339685, CurrentID=0, Status='live', Symbol='LOC339685', Aliases='', description='uncharacterized LOC339685', other_designations='CITF22-49D8.1', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47324623, end_position_on_the_genomic_accession=47366807, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372048, CurrentID=0, Status='live', Symbol='LOC105372048', Aliases='', description='uncharacterized LOC105372048', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31257314, end_position_on_the_genomic_accession=31280655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  203111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203111, CurrentID=0, Status='live', Symbol='ERICH5', Aliases='C8orf47', description='glutamate rich 5', other_designations='glutamate-rich 5', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98064522, end_position_on_the_genomic_accession=98093610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129171, CurrentID=0, Status='live', Symbol='LOC100129171', Aliases='', description='uncharacterized LOC100129171', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=224562814, end_position_on_the_genomic_accession=224563281, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372052, CurrentID=0, Status='live', Symbol='LOC105372052', Aliases='', description='uncharacterized LOC105372052', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31725360, end_position_on_the_genomic_accession=31778950, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372054, CurrentID=0, Status='live', Symbol='LOC105372054', Aliases='', description='uncharacterized LOC105372054', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32175745, end_position_on_the_genomic_accession=32185169, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372055, CurrentID=0, Status='live', Symbol='LOC105372055', Aliases='', description='uncharacterized LOC105372055', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32772101, end_position_on_the_genomic_accession=32778697, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645528, CurrentID=0, Status='live', Symbol='LINC00264', Aliases='C10orf50, NCRNA00264, bA128B16.2', description='long intergenic non-protein coding RNA 264', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26589865, end_position_on_the_genomic_accession=26594324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372057, CurrentID=0, Status='live', Symbol='LOC105372057', Aliases='', description='uncharacterized LOC105372057', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32886752, end_position_on_the_genomic_accession=32903419, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372059, CurrentID=0, Status='live', Symbol='LOC105372059', Aliases='', description='uncharacterized LOC105372059', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32938099, end_position_on_the_genomic_accession=32958140, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927043, CurrentID=0, Status='live', Symbol='LOC101927043', Aliases='', description='uncharacterized LOC101927043', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47192405, end_position_on_the_genomic_accession=47345074, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372060, CurrentID=0, Status='live', Symbol='LOC105372060', Aliases='', description='uncharacterized LOC105372060', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=33911245, end_position_on_the_genomic_accession=33922191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  203190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203190, CurrentID=0, Status='live', Symbol='LGI3', Aliases='LGIL4', description='leucine-rich repeat LGI family member 3', other_designations='LGI1-like protein 4|leucine-rich glioma-inactivated protein 3|leucine-rich repeat LGI family, member 3', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22146830, end_position_on_the_genomic_accession=22156831, orientation='minus', exon_count=8, OMIM=608302, no_hdr0=''),
  203197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203197, CurrentID=0, Status='live', Symbol='C9orf91', Aliases='', description='chromosome 9 open reading frame 91', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114603942, end_position_on_the_genomic_accession=114646423, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100129184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129184, CurrentID=0, Status='live', Symbol='LOC100129184', Aliases='', description='uncharacterized LOC100129184', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29129557, end_position_on_the_genomic_accession=29130248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927044, CurrentID=0, Status='live', Symbol='LOC101927044', Aliases='', description='uncharacterized LOC101927044', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3196904, end_position_on_the_genomic_accession=3247341, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372066, CurrentID=0, Status='live', Symbol='LOC105372066', Aliases='', description='uncharacterized LOC105372066', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35951812, end_position_on_the_genomic_accession=35972527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  203228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203228, CurrentID=0, Status='live', Symbol='C9orf72', Aliases='ALSFTD, FTDALS, FTDALS1', description='chromosome 9 open reading frame 72', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27546546, end_position_on_the_genomic_accession=27573866, orientation='minus', exon_count=12, OMIM=614260, no_hdr0=''),
  101927045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927045, CurrentID=0, Status='live', Symbol='LOC101927045', Aliases='', description='uncharacterized LOC101927045', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24442953, end_position_on_the_genomic_accession=24502197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  203238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203238, CurrentID=0, Status='live', Symbol='CCDC171', Aliases='C9orf93, bA536D16.1, bA778P13.1', description='coiled-coil domain containing 171', other_designations='myosin tail domain containing protein', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15552874, end_position_on_the_genomic_accession=15971899, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  203245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203245, CurrentID=0, Status='live', Symbol='NAIF1', Aliases='C9orf90, bA379C10.2', description='nuclear apoptosis inducing factor 1', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128061233, end_position_on_the_genomic_accession=128067321, orientation='minus', exon_count=3, OMIM=610673, no_hdr0=''),
  100129192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129192, CurrentID=0, Status='live', Symbol='MICC', Aliases='PERB11.3', description='MHC class I polypeptide-related sequence C (pseudogene)', other_designations='MHC class I polypeptide-related sequence C, pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30414713, end_position_on_the_genomic_accession=30419766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645545, CurrentID=0, Status='live', Symbol='IGBP1P2', Aliases='', description='immunoglobulin (CD79A) binding protein 1 pseudogene 2', other_designations='hCG1642624', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58659652, end_position_on_the_genomic_accession=58661608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  203259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203259, CurrentID=0, Status='live', Symbol='FAM219A', Aliases='C9orf25', description='family with sequence similarity 219 member A', other_designations='family with sequence similarity 219, member A|uncharacterized protein C9orf25', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34398184, end_position_on_the_genomic_accession=34458610, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  203260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203260, CurrentID=0, Status='live', Symbol='CCDC107', Aliases='PSEC0222', description='coiled-coil domain containing 107', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35658282, end_position_on_the_genomic_accession=35661503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927046, CurrentID=0, Status='live', Symbol='LOC101927046', Aliases='', description='uncharacterized LOC101927046', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149382061, end_position_on_the_genomic_accession=149384359, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372075, CurrentID=0, Status='live', Symbol='LOC105372075', Aliases='', description='uncharacterized LOC105372075', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=38591683, end_position_on_the_genomic_accession=38594635, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100533933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533933, CurrentID=0, Status='live', Symbol='LOC100533933', Aliases='', description='leucine rich repeat containing 37A pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645548, CurrentID=0, Status='live', Symbol='HSPD1P6', Aliases='HSPD1-6P, HSPDP6', description='heat shock protein family D (Hsp60) member 1 pseudogene 6', other_designations='heat shock 60kD protein 1 (chaperonin) pseudogene 6|heat shock 60kDa protein 1 (chaperonin) pseudogene 6', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36766643, end_position_on_the_genomic_accession=36768887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  203274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203274, CurrentID=0, Status='live', Symbol='LINC00537', Aliases='', description='long intergenic non-protein coding RNA 537', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=68408723, end_position_on_the_genomic_accession=68414763, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421287, CurrentID=0, Status='live', Symbol='LOC100421287', Aliases='', description='cAMP responsive element binding protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95459606, end_position_on_the_genomic_accession=95460162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129198, CurrentID=0, Status='live', Symbol='LOC100129198', Aliases='', description='PRO2866', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=3185514, end_position_on_the_genomic_accession=3185654, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  203286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203286, CurrentID=0, Status='live', Symbol='ANKS6', Aliases='ANKRD14, NPHP16, PKDR1, SAMD6', description='ankyrin repeat and sterile alpha motif domain containing 6', other_designations='SAM domain-containing protein 6|ankyrin repeat domain 14|samCystin', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98732009, end_position_on_the_genomic_accession=98796531, orientation='minus', exon_count=19, OMIM=615370, no_hdr0=''),
  105372079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372079, CurrentID=0, Status='live', Symbol='LOC105372079', Aliases='', description='uncharacterized LOC105372079', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39840335, end_position_on_the_genomic_accession=39882483, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100129201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129201, CurrentID=0, Status='live', Symbol='PRR13P3', Aliases='', description='proline rich 13 pseudogene 3', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117335600, end_position_on_the_genomic_accession=117336683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421288, CurrentID=0, Status='live', Symbol='LOC100421288', Aliases='', description='cAMP responsive element binding protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97008455, end_position_on_the_genomic_accession=97009011, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129202, CurrentID=0, Status='live', Symbol='LOC100129202', Aliases='', description='arginine/serine-rich coiled-coil 1 pseudogene', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88144677, end_position_on_the_genomic_accession=88145989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129203, CurrentID=0, Status='live', Symbol='LOC100129203', Aliases='', description='uncharacterized LOC100129203', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95230186, end_position_on_the_genomic_accession=95234404, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129204, CurrentID=0, Status='live', Symbol='LOC100129204', Aliases='', description='uncharacterized LOC100129204', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=852871, end_position_on_the_genomic_accession=853557, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129205, CurrentID=0, Status='live', Symbol='LOC100129205', Aliases='', description='ribosomal protein S24 pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65077412, end_position_on_the_genomic_accession=65077622, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  203328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203328, CurrentID=0, Status='live', Symbol='SUSD3', Aliases='', description='sushi domain containing 3', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93058688, end_position_on_the_genomic_accession=93085138, orientation='plus', exon_count=9, OMIM=616429, no_hdr0=''),
  100271191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271191, CurrentID=0, Status='live', Symbol='RPL26P20', Aliases='RPL26_10_715', description='ribosomal protein L26 pseudogene 20', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75499705, end_position_on_the_genomic_accession=75500142, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  121270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121270, CurrentID=0, Status='live', Symbol='OR11M1P', Aliases='', description='olfactory receptor family 11 subfamily M member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily M, member 1 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48627841, end_position_on_the_genomic_accession=48628813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106633800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633800, CurrentID=0, Status='live', Symbol='SNORD133', Aliases='ZL142, ggn68', description='small nucleolar RNA, C/D box 133', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633801, CurrentID=0, Status='live', Symbol='SCARNA28', Aliases='ZL1', description='small Cajal body-specific RNA 28', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633802, CurrentID=0, Status='live', Symbol='RN7SKP118', Aliases='', description='RNA, 7SK small nuclear pseudogene 118', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633803, CurrentID=0, Status='live', Symbol='SNORD134', Aliases='ZL12', description='small nucleolar RNA, C/D box 134', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633804, CurrentID=0, Status='live', Symbol='SNORD135', Aliases='ZL2', description='small nucleolar RNA, C/D box 135', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633805, CurrentID=0, Status='live', Symbol='SNORD136', Aliases='ZL26', description='small nucleolar RNA, C/D box 136', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633806, CurrentID=0, Status='live', Symbol='SNORD142', Aliases='SNORA85, SNORD174, ZL68', description='small nucleolar RNA, C/D box 142', other_designations='', map_location='10p15.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633807, CurrentID=0, Status='live', Symbol='RN7SKP251', Aliases='', description='RNA, 7SK small nuclear pseudogene 251', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633808, CurrentID=0, Status='live', Symbol='SNORD143', Aliases='SNORA96, SNORD171, ZL5', description='small nucleolar RNA, C/D box 143', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633809, CurrentID=0, Status='live', Symbol='SNORD144', Aliases='SNORA97, ZL6', description='small nucleolar RNA, C/D box 144', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633810, CurrentID=0, Status='live', Symbol='SCARNA26A', Aliases='', description='small Cajal body-specific RNA 26A', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633813, CurrentID=0, Status='live', Symbol='SIRT1-AS', Aliases='', description='SIRT1 antisense RNA', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633814, CurrentID=0, Status='live', Symbol='SNORA86', Aliases='', description='small nucleolar RNA, H/ACA box 86', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633815, CurrentID=0, Status='live', Symbol='RNU6-205P', Aliases='', description='RNA, U6 small nuclear 205, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106633816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106633816, CurrentID=0, Status='live', Symbol='SCARNA26B', Aliases='', description='small Cajal body-specific RNA 26B', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372090, CurrentID=0, Status='live', Symbol='LOC105372090', Aliases='', description='uncharacterized LOC105372090', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44793926, end_position_on_the_genomic_accession=44800836, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372091, CurrentID=0, Status='live', Symbol='LOC105372091', Aliases='', description='uncharacterized LOC105372091', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45069276, end_position_on_the_genomic_accession=45156463, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372093, CurrentID=0, Status='live', Symbol='LOC105372093', Aliases='', description='uncharacterized LOC105372093', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45731213, end_position_on_the_genomic_accession=45782818, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  645566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645566, CurrentID=0, Status='live', Symbol='FLJ45482', Aliases='', description='uncharacterized LOC645566', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927050, CurrentID=0, Status='live', Symbol='LOC101927050', Aliases='', description='uncharacterized LOC101927050', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89625443, end_position_on_the_genomic_accession=89637953, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129215, CurrentID=0, Status='live', Symbol='LOC100129215', Aliases='', description='uncharacterized LOC100129215', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87693537, end_position_on_the_genomic_accession=87696147, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  727676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727676, CurrentID=0, Status='live', Symbol='SNORD118', Aliases='U8', description='small nucleolar RNA, C/D box 118', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8173453, end_position_on_the_genomic_accession=8173588, orientation='minus', exon_count=1, OMIM=616663, no_hdr0=''),
  727677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727677, CurrentID=0, Status='live', Symbol='CASC8', Aliases='CARLo-1, LINC00860', description='cancer susceptibility candidate 8 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 860', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127289676, end_position_on_the_genomic_accession=127482139, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372096, CurrentID=0, Status='live', Symbol='LOC105372096', Aliases='', description='uncharacterized LOC105372096', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46660210, end_position_on_the_genomic_accession=46663217, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  727682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727682, CurrentID=0, Status='live', Symbol='TAB3-AS1', Aliases='CXorf29', description='TAB3 antisense RNA 1', other_designations='TAB3 antisense RNA 1 (non-protein coding)', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727684, CurrentID=0, Status='live', Symbol='C1orf195', Aliases='', description='chromosome 1 open reading frame 195', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15164196, end_position_on_the_genomic_accession=15171624, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727686, CurrentID=0, Status='live', Symbol='SNURFL', Aliases='CXorf19', description='SNRPN upstream reading frame-like, pseudogene', other_designations='', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139361940, end_position_on_the_genomic_accession=139362482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  448962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448962, CurrentID=0, Status='live', Symbol='DFNA54', Aliases='', description='deafness, autosomal dominant 54', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615649, no_hdr0=''),
  727699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727699, CurrentID=0, Status='live', Symbol='LINC00163', Aliases='C21orf134, NCRNA00163, NLC1-A, NLC1A', description='long intergenic non-protein coding RNA 163', other_designations='narcolepsy candidate region gene 1A|narcolepsy candidate-region 1 gene A', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44989864, end_position_on_the_genomic_accession=44994086, orientation='minus', exon_count=2, OMIM=610259, no_hdr0=''),
  101927051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927051, CurrentID=0, Status='live', Symbol='LOC101927051', Aliases='', description='uncharacterized LOC101927051', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37641832, end_position_on_the_genomic_accession=37658377, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  203413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203413, CurrentID=0, Status='live', Symbol='CT83', Aliases='CXorf61, KK-LC-1, KKLC1', description='cancer/testis antigen 83', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116461686, end_position_on_the_genomic_accession=116463033, orientation='minus', exon_count=2, OMIM=300625, no_hdr0=''),
  203414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203414, CurrentID=0, Status='live', Symbol='LINC01560', Aliases='CXorf24', description='long intergenic non-protein coding RNA 1560', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47482716, end_position_on_the_genomic_accession=47485227, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  727703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727703, CurrentID=0, Status='live', Symbol='IGFL1P1', Aliases='IGFL-5P', description='IGF like family member 1 pseudogene 1', other_designations='IGF-like family member 1 pseudogene 1', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46197021, end_position_on_the_genomic_accession=46197280, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727704, CurrentID=0, Status='live', Symbol='IGFL1P2', Aliases='IGFL-6P', description='IGF like family member 1 pseudogene 2', other_designations='IGF-like family member 1 pseudogene 2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46260435, end_position_on_the_genomic_accession=46260861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372100, CurrentID=0, Status='live', Symbol='LOC105372100', Aliases='', description='uncharacterized LOC105372100', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47668151, end_position_on_the_genomic_accession=47694841, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  727708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727708, CurrentID=0, Status='live', Symbol='SNORD116-19', Aliases='HBII-85-19', description='small nucleolar RNA, C/D box 116-19', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25086526, end_position_on_the_genomic_accession=25086619, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  727709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727709, CurrentID=0, Status='live', Symbol='LOC727709', Aliases='', description='DNA damage regulated autophagy modulator 1 pseudogene', other_designations='DNA-damage regulated autophagy modulator 1 pseudogene', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122739374, end_position_on_the_genomic_accession=122740720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727710, CurrentID=0, Status='live', Symbol='LOC727710', Aliases='', description='uncharacterized LOC727710', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85286304, end_position_on_the_genomic_accession=85303464, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372101, CurrentID=0, Status='live', Symbol='LOC105372101', Aliases='', description='uncharacterized LOC105372101', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47955471, end_position_on_the_genomic_accession=47958018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  727713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727713, CurrentID=0, Status='live', Symbol='LOC727713', Aliases='', description='methionyl aminopeptidase 2 pseudogene', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145825536, end_position_on_the_genomic_accession=145827842, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727714, CurrentID=0, Status='live', Symbol='SCAX2', Aliases='', description='spinocerebellar ataxia, X-linked 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=302600, no_hdr0=''),
  203427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203427, CurrentID=0, Status='live', Symbol='SLC25A43', Aliases='', description='solute carrier family 25 member 43', other_designations='solute carrier family 25, member 43', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119399295, end_position_on_the_genomic_accession=119454478, orientation='plus', exon_count=7, OMIM=300641, no_hdr0=''),
  727716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727716, CurrentID=0, Status='live', Symbol='SCAX4', Aliases='', description='spinocerebellar ataxia, X-linked 4', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=301840, no_hdr0=''),
  203429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203429, CurrentID=0, Status='live', Symbol='FAM226A', Aliases='CXorf50, LINC00246A, NCRNA00246, NCRNA00246A', description='family with sequence similarity 226 member A (non-protein coding)', other_designations='family with sequence similarity 226, member A (non-protein coding)|long intergenic non-protein coding RNA 246A', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72941728, end_position_on_the_genomic_accession=72943750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  203430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203430, CurrentID=0, Status='live', Symbol='ZCCHC5', Aliases='Mar3, Mart3, ZHC5', description='zinc finger, CCHC domain containing 5', other_designations='mammalian retrotransposon-derived 3', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78656069, end_position_on_the_genomic_accession=78659502, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  727721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727721, CurrentID=0, Status='live', Symbol='LOC727721', Aliases='', description='uncharacterized LOC727721', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=9241898, end_position_on_the_genomic_accession=9257040, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372103, CurrentID=0, Status='live', Symbol='LOC105372103', Aliases='', description='uncharacterized LOC105372103', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48342652, end_position_on_the_genomic_accession=48354203, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129224, CurrentID=0, Status='live', Symbol='LOC100129224', Aliases='', description='ankyrin repeat domain-containing protein SOWAHC-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64066534, end_position_on_the_genomic_accession=64067497, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  203447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203447, CurrentID=0, Status='live', Symbol='NRK', Aliases='NESK', description='Nik related kinase', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=105821856, end_position_on_the_genomic_accession=105958612, orientation='plus', exon_count=30, OMIM=300791, no_hdr0=''),
  727736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727736, CurrentID=0, Status='live', Symbol='SH3GL1P3', Aliases='CNSA-P3, SH3GLP3', description='SH3-domain GRB2-like 1 pseudogene 3', other_designations='SH3-domain GRB2-like pseudogene 3', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68133216, end_position_on_the_genomic_accession=68135668, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421293, CurrentID=0, Status='live', Symbol='LOC100421293', Aliases='', description='serine/arginine-rich splicing factor 10 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105710033, end_position_on_the_genomic_accession=105710979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129227, CurrentID=0, Status='live', Symbol='RPS6P22', Aliases='RPS6_10_1242', description='ribosomal protein S6 pseudogene 22', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=59020144, end_position_on_the_genomic_accession=59020925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727751, CurrentID=0, Status='live', Symbol='LOC727751', Aliases='', description='golgin A2 pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727758, CurrentID=0, Status='live', Symbol='ROCK1P1', Aliases='', description='Rho-associated, coiled-coil containing protein kinase 1 pseudogene 1', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=109065, end_position_on_the_genomic_accession=122219, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372109, CurrentID=0, Status='live', Symbol='LOC105372109', Aliases='', description='uncharacterized LOC105372109', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49017679, end_position_on_the_genomic_accession=49019771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927053, CurrentID=0, Status='live', Symbol='LOC101927053', Aliases='', description='uncharacterized LOC101927053', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96919106, end_position_on_the_genomic_accession=96925837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727764, CurrentID=0, Status='live', Symbol='MAFIP', Aliases='MIP, TEKT4P4, pp5644', description='MAFF interacting protein (pseudogene)', other_designations='maFF-interacting protein|tektin 4 pseudogene 4|tektin-4 like protein PP5644', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  203477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203477, CurrentID=0, Status='live', Symbol='HMGA1P1', Aliases='HMGA1L1, HMGIYL1', description='high mobility group AT-hook 1 pseudogene 1', other_designations='HMGIY-related retropseudogene|high mobility group AT-hook 1-like 1', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26747924, end_position_on_the_genomic_accession=26749711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373283, CurrentID=0, Status='live', Symbol='LOC105373283', Aliases='', description='uncharacterized LOC105373283', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80842046, end_position_on_the_genomic_accession=80847124, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727768, CurrentID=0, Status='live', Symbol='TEKT4P1', Aliases='', description='tektin 4 pseudogene 1', other_designations='', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11153795, end_position_on_the_genomic_accession=11159193, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372111, CurrentID=0, Status='live', Symbol='LOC105372111', Aliases='', description='uncharacterized LOC105372111', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49205491, end_position_on_the_genomic_accession=49209136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727775, CurrentID=0, Status='live', Symbol='LOC727775', Aliases='', description='uncharacterized LOC727775', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  121296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121296, CurrentID=0, Status='live', Symbol='LOC121296', Aliases='', description='transmembrane protein 132B pseudogene', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126864421, end_position_on_the_genomic_accession=126865463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129233, CurrentID=0, Status='live', Symbol='LOC100129233', Aliases='', description='uncharacterized LOC100129233', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=106757101, end_position_on_the_genomic_accession=106758966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  727789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727789, CurrentID=0, Status='live', Symbol='RWDD1P3', Aliases='RWDD1L1', description='RWD domain containing 1 pseudogene 3', other_designations='RWD domain containing 1-like 1', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76757098, end_position_on_the_genomic_accession=76758093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727792, CurrentID=0, Status='live', Symbol='RPL31P31', Aliases='RPL31_10_468', description='ribosomal protein L31 pseudogene 31', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33968147, end_position_on_the_genomic_accession=33968567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727795, CurrentID=0, Status='live', Symbol='HMGN2P5', Aliases='HMGN2L5', description='high mobility group nucleosomal binding domain 2 pseudogene 5', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 5|high-mobility group nucleosomal binding domain 2-like 5', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29730612, end_position_on_the_genomic_accession=29731806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  203510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203510, CurrentID=0, Status='live', Symbol='HMGB3P30', Aliases='', description='high mobility group box 3 pseudogene 30', other_designations='high-mobility group box 3 pseudogene 30', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112689892, end_position_on_the_genomic_accession=112690644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727799, CurrentID=0, Status='live', Symbol='LOC727799', Aliases='', description='uncharacterized LOC727799', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=787879, end_position_on_the_genomic_accession=789286, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727800, CurrentID=0, Status='live', Symbol='RNF208', Aliases='', description='ring finger protein 208', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137220247, end_position_on_the_genomic_accession=137221323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  727803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727803, CurrentID=0, Status='live', Symbol='LOC727803', Aliases='', description='RAN binding protein 1 pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76984238, end_position_on_the_genomic_accession=76984669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727805, CurrentID=0, Status='live', Symbol='POM121L14P', Aliases='', description='POM121 transmembrane nucleoporin like 14, pseudogene', other_designations='POM121 transmembrane nucleoporin-like 14, pseudogene|putative POM121-like protein 1-like', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=58259807, end_position_on_the_genomic_accession=58271194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372117, CurrentID=0, Status='live', Symbol='LOC105372117', Aliases='', description='uncharacterized LOC105372117', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51025096, end_position_on_the_genomic_accession=51030169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  727808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727808, CurrentID=0, Status='live', Symbol='LOC727808', Aliases='', description='uncharacterized LOC727808', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=30114824, end_position_on_the_genomic_accession=30116302, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  203522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203522, CurrentID=0, Status='live', Symbol='DDX26B', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 26B', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135520630, end_position_on_the_genomic_accession=135582535, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  203523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203523, CurrentID=0, Status='live', Symbol='ZNF449', Aliases='ZSCAN19', description='zinc finger protein 449', other_designations='zinc finger and SCAN domain-containing protein 19', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135344757, end_position_on_the_genomic_accession=135363413, orientation='plus', exon_count=7, OMIM=300627, no_hdr0=''),
  645591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645591, CurrentID=0, Status='live', Symbol='LHFPL3-AS1', Aliases='', description='LHFPL3 antisense RNA 1', other_designations='LHFPL3 antisense RNA 1 (non-protein coding)', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104796506, end_position_on_the_genomic_accession=104804092, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927055, CurrentID=0, Status='live', Symbol='LOC101927055', Aliases='', description='uncharacterized LOC101927055', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178776926, end_position_on_the_genomic_accession=178779963, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727821, CurrentID=0, Status='live', Symbol='RPL21P45', Aliases='RPL21_15_469', description='ribosomal protein L21 pseudogene 45', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37821321, end_position_on_the_genomic_accession=37821878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372120, CurrentID=0, Status='live', Symbol='LOC105372120', Aliases='', description='uncharacterized LOC105372120', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51645003, end_position_on_the_genomic_accession=51656433, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727826, CurrentID=0, Status='live', Symbol='RPS11P5', Aliases='RPS11_3_1298', description='ribosomal protein S11 pseudogene 5', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132825681, end_position_on_the_genomic_accession=132826237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727828, CurrentID=0, Status='live', Symbol='TRIM64DP', Aliases='', description='tripartite motif containing 64D, pseudogene', other_designations='tripartite motif containing 64B pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89776981, end_position_on_the_genomic_accession=89782560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727830, CurrentID=0, Status='live', Symbol='SPATA31A3', Aliases='FAM75A3', description='SPATA31 subfamily A, member 3', other_designations='family with sequence similarity 75, member A3|protein FAM75A3', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66986301, end_position_on_the_genomic_accession=66992550, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  727831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727831, CurrentID=0, Status='live', Symbol='RPL13AP8', Aliases='RPL13A_2_157', description='ribosomal protein L13a pseudogene 8', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206907595, end_position_on_the_genomic_accession=206908735, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727832, CurrentID=0, Status='live', Symbol='GOLGA6L6', Aliases='', description='golgin A6 family-like 6', other_designations='golgi autoantigen, golgin subfamily a, 6-like 6', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20531856, end_position_on_the_genomic_accession=20541800, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  727833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727833, CurrentID=0, Status='live', Symbol='IMPDH1P11', Aliases='IMPDH1, IMPDHL1', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 11', other_designations='IMP (inosine monophosphate) dehydrogenase-like 1', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10111309, end_position_on_the_genomic_accession=10113590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  203547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203547, CurrentID=0, Status='live', Symbol='VMA21', Aliases='MEAX, XMEA', description='VMA21 vacuolar H+-ATPase homolog (S. cerevisiae)', other_designations='myopathy with excessive autophagy protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151396745, end_position_on_the_genomic_accession=151409364, orientation='plus', exon_count=4, OMIM=300913, no_hdr0=''),
  727837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727837, CurrentID=0, Status='live', Symbol='SSX2B', Aliases='CT5.2, CT5.2b, HOM-MEL-40, SSX', description='synovial sarcoma, X breakpoint 2B', other_designations='cancer/testis antigen 5.2|cancer/testis antigen family 5, member 2b|tumor antigen HOM-MEL-40', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52751204, end_position_on_the_genomic_accession=52761536, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  727838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727838, CurrentID=0, Status='live', Symbol='LOC727838', Aliases='', description='cancer/testis antigen family 47, member A11 pseudogene', other_designations='cancer/testis antigen family 47, member A pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119072969, end_position_on_the_genomic_accession=119076383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727842, CurrentID=0, Status='live', Symbol='RBBP4P4', Aliases='', description='retinoblastoma binding protein 4 pseudogene 4', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=58119744, end_position_on_the_genomic_accession=58120924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372123, CurrentID=0, Status='live', Symbol='LOC105372123', Aliases='', description='uncharacterized LOC105372123', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375027, CurrentID=0, Status='live', Symbol='LOC105375027', Aliases='', description='uncharacterized LOC105375027', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33856202, end_position_on_the_genomic_accession=33877132, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  203562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203562, CurrentID=0, Status='live', Symbol='TMEM31', Aliases='', description='transmembrane protein 31', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103710909, end_position_on_the_genomic_accession=103714032, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727851, CurrentID=0, Status='live', Symbol='RGPD8', Aliases='RANBP2L1, RGP8, RanBP2alpha', description='RANBP2-like and GRIP domain containing 8', other_designations='RAN binding protein 2-like 1|ran-binding protein 2-like 3|ranBP2-like 3|ranBP2L3', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112368369, end_position_on_the_genomic_accession=112433645, orientation='minus', exon_count=25, OMIM=602752, no_hdr0=''),
  100129245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129245, CurrentID=0, Status='live', Symbol='LOC100129245', Aliases='', description='GAR1 ribonucleoprotein homolog (yeast) pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184558978, end_position_on_the_genomic_accession=184559796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727856, CurrentID=0, Status='live', Symbol='DDX11L16', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 16', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 16|DEAD/H box polypeptide 11 like 16', map_location='Xq28; Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57212178, end_position_on_the_genomic_accession=57214703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  727857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727857, CurrentID=0, Status='live', Symbol='BHLHA9', Aliases='BHLHF42', description='basic helix-loop-helix family member a9', other_designations='basic helix-loop-helix family, member a9|class F basic helix-loop-helix factor 42|class II basic helix-loop-helix protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1270564, end_position_on_the_genomic_accession=1271271, orientation='plus', exon_count=1, OMIM=615416, no_hdr0=''),
  727859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727859, CurrentID=0, Status='live', Symbol='C4BPAP1', Aliases='C4BPAL1', description='complement component 4 binding protein, alpha pseudogene 1', other_designations='complement component 4 binding protein, alpha-like 1|complement component 4-binding protein, alpha-like 1', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207165495, end_position_on_the_genomic_accession=207184061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372126, CurrentID=0, Status='live', Symbol='LOC105372126', Aliases='', description='uncharacterized LOC105372126', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55589899, end_position_on_the_genomic_accession=55606698, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  727865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727865, CurrentID=0, Status='live', Symbol='RPL9P32', Aliases='RPL9_18_1643', description='ribosomal protein L9 pseudogene 32', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29956306, end_position_on_the_genomic_accession=29957013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727866, CurrentID=0, Status='live', Symbol='FAM156B', Aliases='TMEM29B', description='family with sequence similarity 156 member B', other_designations='family with sequence similarity 156, member B|protein FAM156A|transmembrane protein 29|transmembrane protein 29/29B|transmembrane protein 29B', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52891306, end_position_on_the_genomic_accession=52908560, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  448991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=448991, CurrentID=0, Status='live', Symbol='TAS2R67P', Aliases='PS5', description='taste 2 receptor member 67 pseudogene', other_designations='taste receptor, type 2, member 43 pseudogene|taste receptor, type 2, member 67 pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11179541, end_position_on_the_genomic_accession=11180462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372128, CurrentID=0, Status='live', Symbol='LOC105372128', Aliases='', description='uncharacterized LOC105372128', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55331236, end_position_on_the_genomic_accession=55340409, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  727874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727874, CurrentID=0, Status='live', Symbol='LOC727874', Aliases='', description='WW domain binding protein 11 pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80559979, end_position_on_the_genomic_accession=80562452, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727879, CurrentID=0, Status='live', Symbol='C1DP4', Aliases='', description='C1D nuclear receptor corepressor pseudogene 4', other_designations='C1D nuclear receptor co-repressor pseudogene 4', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80040264, end_position_on_the_genomic_accession=80041122, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727880, CurrentID=0, Status='live', Symbol='LOC727880', Aliases='', description='uncharacterized LOC727880', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=30636502, end_position_on_the_genomic_accession=30640439, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  727884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727884, CurrentID=0, Status='live', Symbol='LOC727884', Aliases='', description='insulin like growth factor 2 mRNA binding protein 2 pseudogene', other_designations='insulin-like growth factor 2 mRNA binding protein 2 pseudogene', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42286911, end_position_on_the_genomic_accession=42288557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372130, CurrentID=0, Status='live', Symbol='LOC105372130', Aliases='', description='uncharacterized LOC105372130', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55665544, end_position_on_the_genomic_accession=55839862, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105372131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372131, CurrentID=0, Status='live', Symbol='LOC105372131', Aliases='', description='uncharacterized LOC105372131', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55863352, end_position_on_the_genomic_accession=55870232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727896, CurrentID=0, Status='live', Symbol='LOC727896', Aliases='', description='cysteine and histidine rich domain containing 1 pseudogene', other_designations='cysteine and histidine-rich domain (CHORD) containing 1 pseudogene', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2943212, end_position_on_the_genomic_accession=2946623, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  727897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727897, CurrentID=0, Status='live', Symbol='MUC5B', Aliases='MG1, MUC-5B, MUC5, MUC9', description='mucin 5B, oligomeric mucus/gel-forming', other_designations='cervical mucin MUC5B|high molecular weight salivary mucin MG1|mucin 5, subtype B, tracheobronchial|sublingual gland mucin', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1223065, end_position_on_the_genomic_accession=1262176, orientation='plus', exon_count=49, OMIM=600770, no_hdr0=''),
  203611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203611, CurrentID=0, Status='live', Symbol='CDY2B', Aliases='CDY', description='chromodomain protein, Y-linked, 2B', other_designations='Y chromosome chromodomain protein 2B|chromodomain protein, Y chromosome, 2 related|testis-specific chromodomain protein Y protein 2 related', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17878260, end_position_on_the_genomic_accession=17880219, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130964, CurrentID=0, Status='live', Symbol='LOC100130964', Aliases='', description='ADAM metallopeptidase domain 3A-like', other_designations='CTD-2024D23.1', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39563112, end_position_on_the_genomic_accession=39580134, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  727905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727905, CurrentID=0, Status='live', Symbol='SPATA31A5', Aliases='FAM75A5', description='SPATA31 subfamily A, member 5', other_designations='family with sequence similarity 75, member A5|protein FAM75A5', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=60914372, end_position_on_the_genomic_accession=60920653, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  727909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727909, CurrentID=0, Status='live', Symbol='GOLGA8Q', Aliases='', description='golgin A8 family member Q', other_designations='golgin A8 family, member Q', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30556266, end_position_on_the_genomic_accession=30562498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727910, CurrentID=0, Status='live', Symbol='TLCD2', Aliases='', description='TLC domain containing 2', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1702790, end_position_on_the_genomic_accession=1710368, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727914, CurrentID=0, Status='live', Symbol='BMS1P16', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 16', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 16|BMS1 ribosome biogenesis factor pseudogene 16', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21160305, end_position_on_the_genomic_accession=21167357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727915, CurrentID=0, Status='live', Symbol='AGBL1-AS1', Aliases='', description='AGBL1 antisense RNA 1', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86296833, end_position_on_the_genomic_accession=86316968, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727916, CurrentID=0, Status='live', Symbol='LOC727916', Aliases='', description='uncharacterized LOC727916', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=64280910, end_position_on_the_genomic_accession=64283473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373284, CurrentID=0, Status='live', Symbol='LOC105373284', Aliases='', description='uncharacterized LOC105373284', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81420992, end_position_on_the_genomic_accession=81425448, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  727919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727919, CurrentID=0, Status='live', Symbol='LOC727919', Aliases='', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 2 pseudogene', other_designations='hCG1795065', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20529396, end_position_on_the_genomic_accession=20530096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372136, CurrentID=0, Status='live', Symbol='LOC105372136', Aliases='', description='uncharacterized LOC105372136', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57050096, end_position_on_the_genomic_accession=57051426, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727924, CurrentID=0, Status='live', Symbol='LOC727924', Aliases='', description='uncharacterized LOC727924', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21552795, end_position_on_the_genomic_accession=21640521, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  727925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727925, CurrentID=0, Status='live', Symbol='LOC727925', Aliases='', description='hCG2031213', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68720717, end_position_on_the_genomic_accession=68761467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727930, CurrentID=0, Status='live', Symbol='LOC727930', Aliases='', description='chondroitin sulfate proteoglycan 4-like', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54810139, end_position_on_the_genomic_accession=54872034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727936, CurrentID=0, Status='live', Symbol='GXYLT2', Aliases='GLT8D4', description='glucoside xylosyltransferase 2', other_designations='glycosyltransferase 8 domain containing 4|glycosyltransferase 8 domain-containing protein 4', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72848522, end_position_on_the_genomic_accession=72975374, orientation='plus', exon_count=9, OMIM=613322, no_hdr0=''),
  105372139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372139, CurrentID=0, Status='live', Symbol='LOC105372139', Aliases='', description='uncharacterized LOC105372139', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57444376, end_position_on_the_genomic_accession=57469778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  727940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727940, CurrentID=0, Status='live', Symbol='RHOXF2B', Aliases='', description='Rhox homeobox family member 2B', other_designations='Rhox homeobox family, member 2B', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120072264, end_position_on_the_genomic_accession=120077742, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727941, CurrentID=0, Status='live', Symbol='SLC25A24P1', Aliases='', description='SLC25A24 pseudogene 1', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108273139, end_position_on_the_genomic_accession=108337853, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727944, CurrentID=0, Status='live', Symbol='LOC727944', Aliases='', description='uncharacterized LOC727944', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=490771, end_position_on_the_genomic_accession=492887, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372140, CurrentID=0, Status='live', Symbol='LOC105372140', Aliases='', description='uncharacterized LOC105372140', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57896335, end_position_on_the_genomic_accession=57901924, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  727947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727947, CurrentID=0, Status='live', Symbol='LOC727947', Aliases='', description='ubiquinol-cytochrome c reductase binding protein pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160258239, end_position_on_the_genomic_accession=160258700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727951, CurrentID=0, Status='live', Symbol='RPS3AP16', Aliases='RPS3A_3_456', description='ribosomal protein S3a pseudogene 16', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4076667, end_position_on_the_genomic_accession=4077515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421300, CurrentID=0, Status='live', Symbol='LOC100421300', Aliases='', description='pregnancy specific beta-1-glycoprotein 9 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43178876, end_position_on_the_genomic_accession=43179157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727956, CurrentID=0, Status='live', Symbol='SDHAP2', Aliases='SDHAL2, SDHALP2', description='succinate dehydrogenase complex subunit A, flavoprotein pseudogene 2', other_designations='succinate dehydrogenase complex, subunit A, flavoprotein pseudogene 2|succinate dehydrogenase complex, subunit A, flavoprotein-like 2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195658039, end_position_on_the_genomic_accession=195688871, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  727957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727957, CurrentID=0, Status='live', Symbol='MROH1', Aliases='HEATR7A', description='maestro heat-like repeat family member 1', other_designations='HEAT repeat containing 7A|HEAT repeat-containing protein 7A', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144148016, end_position_on_the_genomic_accession=144261940, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  727961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727961, CurrentID=0, Status='live', Symbol='LOC727961', Aliases='', description='hCG1776047', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  727964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727964, CurrentID=0, Status='live', Symbol='C2orf69P1', Aliases='', description='chromosome 2 open reading frame 69 pseudogene 1', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35552907, end_position_on_the_genomic_accession=35555525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727969, CurrentID=0, Status='live', Symbol='TSPY16P', Aliases='', description='testis specific protein, Y-linked 16, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9640249, end_position_on_the_genomic_accession=9640761, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727970, CurrentID=0, Status='live', Symbol='RPS3AP43', Aliases='RPS3A_21_1203', description='ribosomal protein S3a pseudogene 43', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8359010, end_position_on_the_genomic_accession=8359858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390860, CurrentID=0, Status='live', Symbol='LOC390860', Aliases='', description='ribosomal protein, large, P0 pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59823950, end_position_on_the_genomic_accession=59825335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129265, CurrentID=0, Status='live', Symbol='LOC100129265', Aliases='', description='zinc finger protein 66-like', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20295093, end_position_on_the_genomic_accession=20303960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727978, CurrentID=0, Status='live', Symbol='LOC727978', Aliases='', description='mucin-5AC-like', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195935613, end_position_on_the_genomic_accession=195942008, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727980, CurrentID=0, Status='live', Symbol='LOC727980', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog (yeast) pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19131445, end_position_on_the_genomic_accession=19132989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  727982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727982, CurrentID=0, Status='live', Symbol='LINC01249', Aliases='', description='long intergenic non-protein coding RNA 1249', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=4628218, end_position_on_the_genomic_accession=4656222, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  727983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727983, CurrentID=0, Status='live', Symbol='LOC727983', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=18833962, end_position_on_the_genomic_accession=18839322, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  727984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727984, CurrentID=0, Status='live', Symbol='RPL17P22', Aliases='RPL17_13_564', description='ribosomal protein L17 pseudogene 22', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56136944, end_position_on_the_genomic_accession=56137578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727987, CurrentID=0, Status='live', Symbol='FAM207CP', Aliases='', description='family with sequence similarity 207 member C, pseudogene', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13792633, end_position_on_the_genomic_accession=13793313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727991, CurrentID=0, Status='live', Symbol='SLC35E1P1', Aliases='', description='solute carrier family 35 member E1 pseudogene 1', other_designations='solute carrier family 35, member E1 pseudogene 1', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20607079, end_position_on_the_genomic_accession=20610582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  727993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727993, CurrentID=0, Status='live', Symbol='LOC727993', Aliases='', description='uncharacterized LOC727993', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1947246, end_position_on_the_genomic_accession=2007517, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  727997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=727997, CurrentID=0, Status='live', Symbol='RPS12P9', Aliases='RPS12_3_484', description='ribosomal protein S12 pseudogene 9', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=61775427, end_position_on_the_genomic_accession=61776266, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372149, CurrentID=0, Status='live', Symbol='LOC105372149', Aliases='', description='uncharacterized LOC105372149', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59563762, end_position_on_the_genomic_accession=59572450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728002, CurrentID=0, Status='live', Symbol='RPL15P17', Aliases='RPL15_10_1204', description='ribosomal protein L15 pseudogene 17', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8460584, end_position_on_the_genomic_accession=8461232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728005, CurrentID=0, Status='live', Symbol='AGAP13P', Aliases='CTGLF9P, bA548K23.1', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 13, pseudogene', other_designations='centaurin, gamma-like family, member 9 pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46816582, end_position_on_the_genomic_accession=46835825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100533936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533936, CurrentID=0, Status='live', Symbol='LOC100533936', Aliases='', description='major histocompatibility complex, class I, B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728010, CurrentID=0, Status='live', Symbol='RPL35P3', Aliases='RPL35_2_728', description='ribosomal protein L35 pseudogene 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105302420, end_position_on_the_genomic_accession=105302859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  55799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55799, CurrentID=0, Status='live', Symbol='CACNA2D3', Aliases='HSA272268', description='calcium channel, voltage-dependent, alpha 2/delta subunit 3', other_designations='calcium channel alpha2-delta3 subunit|calcium channel, voltage-dependent, alpha 2/delta 3 subunit|voltage-gated calcium channel subunit alpha-2/delta-3', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54122593, end_position_on_the_genomic_accession=55074557, orientation='plus', exon_count=43, OMIM=606399, no_hdr0=''),
  728012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728012, CurrentID=0, Status='live', Symbol='MCHR2-AS1', Aliases='', description='MCHR2 antisense RNA 1', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99994038, end_position_on_the_genomic_accession=100076419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728013, CurrentID=0, Status='live', Symbol='CBWD7', Aliases='', description='COBW domain containing 7', other_designations='putative cobalamin synthase W domain-containing protein 7', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=42671863, end_position_on_the_genomic_accession=42714961, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  728015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728015, CurrentID=0, Status='live', Symbol='RPS7P15', Aliases='', description='ribosomal protein S7 pseudogene 15', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4507045, end_position_on_the_genomic_accession=4507833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55800, CurrentID=0, Status='live', Symbol='SCN3B', Aliases='ATFB16, BRGDA7, HSA243396, SCNB3', description='sodium channel, voltage gated, type III beta subunit', other_designations='sodium channel, voltage-gated, type III, beta subunit|voltage-gated sodium channel beta-3 subunit', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123629187, end_position_on_the_genomic_accession=123654607, orientation='minus', exon_count=8, OMIM=608214, no_hdr0=''),
  728019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728019, CurrentID=0, Status='live', Symbol='HNRNPA1P67', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 67', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=72807187, end_position_on_the_genomic_accession=72808384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728022, CurrentID=0, Status='live', Symbol='DUX4L14', Aliases='', description='double homeobox 4 like 14', other_designations='double homeobox, 4 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133756558, end_position_on_the_genomic_accession=133757836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  449017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449017, CurrentID=0, Status='live', Symbol='BMIQ6', Aliases='', description='body mass index QTL 6', other_designations='Body mass index quantitative trait locus on chromosome 20, in children', map_location='20pter-p11.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608559, no_hdr0=''),
  728024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728024, CurrentID=0, Status='live', Symbol='LOC728024', Aliases='', description='chromosome X open reading frame 56 pseudogene', other_designations='hCG1640171', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37746554, end_position_on_the_genomic_accession=37748046, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100133244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133244, CurrentID=0, Status='live', Symbol='LOC100133244', Aliases='', description='leucine rich repeat containing 37, member A3 pseudogene', other_designations='leucine rich repeat containing 37A pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35915313, end_position_on_the_genomic_accession=35916324, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728026, CurrentID=0, Status='live', Symbol='LOC728026', Aliases='', description='prothymosin alpha-like', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94347117, end_position_on_the_genomic_accession=94347586, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55802, CurrentID=0, Status='live', Symbol='DCP1A', Aliases='HSA275986, Nbla00360, SMAD4IP1, SMIF', description='decapping mRNA 1A', other_designations='DCP1 decapping enzyme homolog A|DCP1 decapping enzyme-like protein A|Smad4-interacting transcriptional co-activator|decapping enzyme hDcp1a|putative protein product of Nbla00360|transcription factor SMIF', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53283423, end_position_on_the_genomic_accession=53347627, orientation='minus', exon_count=11, OMIM=607010, no_hdr0=''),
  728034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728034, CurrentID=0, Status='live', Symbol='LOC728034', Aliases='', description='BMS1 ribosome biogenesis factor pseudogene', other_designations='BMS1 homolog, ribosome assembly protein pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94196055, end_position_on_the_genomic_accession=94203288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129275, CurrentID=0, Status='live', Symbol='LOC100129275', Aliases='', description='uncharacterized LOC100129275', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=121036413, end_position_on_the_genomic_accession=121037016, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728036, CurrentID=0, Status='live', Symbol='CT47A10', Aliases='CT47, CT47.10', description='cancer/testis antigen family 47, member A10', other_designations='cancer/testis CT47 family, member 10|cancer/testis antigen 47', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120938701, end_position_on_the_genomic_accession=120942059, orientation='minus', exon_count=4, OMIM=300789, no_hdr0=''),
  728039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728039, CurrentID=0, Status='live', Symbol='SSR4P1', Aliases='C21orf122, PRED57, PRED90', description='signal sequence receptor, delta pseudogene 1', other_designations='hCG401283', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45070955, end_position_on_the_genomic_accession=45073211, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728040, CurrentID=0, Status='live', Symbol='LOC728040', Aliases='', description='hCG1813624', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73508803, end_position_on_the_genomic_accession=73528533, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372156, CurrentID=0, Status='live', Symbol='LOC105372156', Aliases='', description='uncharacterized LOC105372156', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60637967, end_position_on_the_genomic_accession=60646250, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  728042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728042, CurrentID=0, Status='live', Symbol='CT47A9', Aliases='CT47.9', description='cancer/testis antigen family 47, member A9', other_designations='cancer/testis CT47 family, member 9', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120943561, end_position_on_the_genomic_accession=120946919, orientation='minus', exon_count=4, OMIM=300788, no_hdr0=''),
  728045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728045, CurrentID=0, Status='live', Symbol='PPBPP1', Aliases='PPBPL1, TGB2', description='pro-platelet basic protein pseudogene 1', other_designations='connective tissue-activating peptide I|neutrophil-activating peptide-2-like 1|pro-platelet basic protein-like 1|thromboglobulin, beta-2', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73848736, end_position_on_the_genomic_accession=73849253, orientation='plus', exon_count=0, OMIM=188035, no_hdr0=''),
  728047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728047, CurrentID=0, Status='live', Symbol='GOLGA8O', Aliases='', description='golgin A8 family member O', other_designations='Golgin subfamily A member 6-like protein 11|Golgin subfamily A member 8O|golgin A8 family, member O|golgin subfamily A member 8-like protein 2-like', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32441914, end_position_on_the_genomic_accession=32456324, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  728048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728048, CurrentID=0, Status='live', Symbol='LOC728048', Aliases='', description='interferon induced transmembrane protein pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=66094070, end_position_on_the_genomic_accession=66094686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728049, CurrentID=0, Status='live', Symbol='CT47A8', Aliases='CT47.8', description='cancer/testis antigen family 47, member A8', other_designations='cancer/testis CT47 family, member 8', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120948422, end_position_on_the_genomic_accession=120951780, orientation='minus', exon_count=4, OMIM=300787, no_hdr0=''),
  101060294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060294, CurrentID=0, Status='live', Symbol='LOC101060294', Aliases='', description='PRAME family member 13-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645630, CurrentID=0, Status='live', Symbol='RPS13P8', Aliases='RPS13_5_1432', description='ribosomal protein S13 pseudogene 8', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52095272, end_position_on_the_genomic_accession=52095789, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728056, CurrentID=0, Status='live', Symbol='IQSEC3P3', Aliases='', description='IQ motif and Sec7 domain 3 pseudogene 3', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64327409, end_position_on_the_genomic_accession=64327972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372159, CurrentID=0, Status='live', Symbol='LOC105372159', Aliases='', description='uncharacterized LOC105372159', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61894556, end_position_on_the_genomic_accession=61905636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728061, CurrentID=0, Status='live', Symbol='LOC728061', Aliases='', description='hCG2003663', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=96863820, end_position_on_the_genomic_accession=96866034, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728062, CurrentID=0, Status='live', Symbol='CT47A6', Aliases='CT47.6', description='cancer/testis antigen family 47, member A6', other_designations='cancer/testis CT47 family, member 6', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120958165, end_position_on_the_genomic_accession=120961483, orientation='minus', exon_count=3, OMIM=300785, no_hdr0=''),
  728064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728064, CurrentID=0, Status='live', Symbol='CUBNP1', Aliases='', description='cubilin (intrinsic factor-cobalamin receptor) pseudogene 1', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42694371, end_position_on_the_genomic_accession=42714576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372160, CurrentID=0, Status='live', Symbol='LOC105372160', Aliases='', description='uncharacterized LOC105372160', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62756307, end_position_on_the_genomic_accession=62774027, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  728066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728066, CurrentID=0, Status='live', Symbol='FAM133DP', Aliases='', description='family with sequence similarity 133 member D, pseudogene', other_designations='family with sequence similarity 133, member B pseudogene|family with sequence similarity 133, member D, pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157378487, end_position_on_the_genomic_accession=157380509, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271194, CurrentID=0, Status='live', Symbol='RPL31P24', Aliases='RPL31_11_499', description='ribosomal protein L31 pseudogene 24', other_designations='', map_location='4q22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88163539, end_position_on_the_genomic_accession=88164010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728072, CurrentID=0, Status='live', Symbol='CT47A5', Aliases='CT47.5', description='cancer/testis antigen family 47, member A5', other_designations='CT47|cancer/testis CT47 family, member 5|cancer/testis antigen 47', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120963026, end_position_on_the_genomic_accession=120966384, orientation='minus', exon_count=4, OMIM=300784, no_hdr0=''),
  728073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728073, CurrentID=0, Status='live', Symbol='LOC728073', Aliases='', description='uncharacterized LOC728073', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=69026722, end_position_on_the_genomic_accession=69029141, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728075, CurrentID=0, Status='live', Symbol='CT47A4', Aliases='CT47.4', description='cancer/testis antigen family 47, member A4', other_designations='cancer/testis CT47 family, member 4', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120967886, end_position_on_the_genomic_accession=120971244, orientation='minus', exon_count=4, OMIM=300783, no_hdr0=''),
  728081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728081, CurrentID=0, Status='live', Symbol='LINC00290', Aliases='NCRNA00290', description='long intergenic non-protein coding RNA 290', other_designations='hCG2025798', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181064089, end_position_on_the_genomic_accession=181159149, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728082, CurrentID=0, Status='live', Symbol='CT47A3', Aliases='CT47.3', description='cancer/testis antigen family 47, member A3', other_designations='cancer/testis CT47 family, member 3', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120972746, end_position_on_the_genomic_accession=120976104, orientation='minus', exon_count=4, OMIM=300782, no_hdr0=''),
  105372163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372163, CurrentID=0, Status='live', Symbol='LOC105372163', Aliases='', description='uncharacterized LOC105372163', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63397260, end_position_on_the_genomic_accession=63403695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728084, CurrentID=0, Status='live', Symbol='LOC728084', Aliases='', description='uncharacterized LOC728084', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89011126, end_position_on_the_genomic_accession=89019692, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728088, CurrentID=0, Status='live', Symbol='RPL15P18', Aliases='RPL15_5_1331', description='ribosomal protein L15 pseudogene 18', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100150440, end_position_on_the_genomic_accession=100151143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372164, CurrentID=0, Status='live', Symbol='LOC105372164', Aliases='', description='uncharacterized LOC105372164', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63724289, end_position_on_the_genomic_accession=63737861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728090, CurrentID=0, Status='live', Symbol='CT47A2', Aliases='CT47.2', description='cancer/testis antigen family 47, member A2', other_designations='cancer/testis CT47 family, member 2', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120977606, end_position_on_the_genomic_accession=120980964, orientation='minus', exon_count=4, OMIM=300781, no_hdr0=''),
  728095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728095, CurrentID=0, Status='live', Symbol='LOC728095', Aliases='', description='uncharacterized LOC728095', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169013227, end_position_on_the_genomic_accession=169037998, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  728096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728096, CurrentID=0, Status='live', Symbol='CT47A1', Aliases='CT47.1', description='cancer/testis antigen family 47, member A1', other_designations='cancer/testis CT47 family, member 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120982466, end_position_on_the_genomic_accession=120985824, orientation='minus', exon_count=4, OMIM=300780, no_hdr0=''),
  728097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728097, CurrentID=0, Status='live', Symbol='FAM8A2P', Aliases='', description='family with sequence similarity 8 member A2, pseudogene', other_designations='family with sequence similarity 8, member A1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56327844, end_position_on_the_genomic_accession=56333227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728098, CurrentID=0, Status='live', Symbol='LOC728098', Aliases='', description='mitogen-activated protein kinase 1 interacting protein 1-like pseudogene', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100889603, end_position_on_the_genomic_accession=100890337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728099, CurrentID=0, Status='live', Symbol='LOC728099', Aliases='', description='uncharacterized LOC728099', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21348765, end_position_on_the_genomic_accession=21361469, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728100, CurrentID=0, Status='live', Symbol='LOC728100', Aliases='', description='uncharacterized LOC728100', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=199421470, end_position_on_the_genomic_accession=199428605, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372166, CurrentID=0, Status='live', Symbol='LOC105372166', Aliases='', description='uncharacterized LOC105372166', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64538100, end_position_on_the_genomic_accession=64539134, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728106, CurrentID=0, Status='live', Symbol='RPL23AP50', Aliases='RPL23A_20_731', description='ribosomal protein L23a pseudogene 50', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107931800, end_position_on_the_genomic_accession=107932254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728111, CurrentID=0, Status='live', Symbol='HMGN1P31', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 31', other_designations='hCG1659830|high-mobility group nucleosome binding domain 1 pseudogene 31', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60794607, end_position_on_the_genomic_accession=60794903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728113, CurrentID=0, Status='live', Symbol='ANXA8L1', Aliases='ANXA8, ANXA8L2, VAC-beta, bA145E20.2', description='annexin A8-like 1', other_designations='annexin A8-like protein 2|annexin A8L2|annexin VIII|annexin-8|vascular anticoagulant-beta', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46375590, end_position_on_the_genomic_accession=46391784, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  728114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728114, CurrentID=0, Status='live', Symbol='LOC728114', Aliases='', description='uncharacterized LOC728114', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48019771, end_position_on_the_genomic_accession=48025382, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728115, CurrentID=0, Status='live', Symbol='FAM60CP', Aliases='', description='family with sequence similarity 60 member C, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60016775, end_position_on_the_genomic_accession=60017569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728116, CurrentID=0, Status='live', Symbol='ZBTB8B', Aliases='ZNF916B', description='zinc finger and BTB domain containing 8B', other_designations='putative zinc finger and BTB domain-containing protein 8B', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32465057, end_position_on_the_genomic_accession=32487859, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728118, CurrentID=0, Status='live', Symbol='NUTM2A', Aliases='FAM22A', description='NUT family member 2A', other_designations='family with sequence similarity 22, member A|protein FAM22A', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87223041, end_position_on_the_genomic_accession=87235907, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372169, CurrentID=0, Status='live', Symbol='LOC105372169', Aliases='', description='uncharacterized LOC105372169', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=65464892, end_position_on_the_genomic_accession=65506660, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728121, CurrentID=0, Status='live', Symbol='CSPG4P12', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 12', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85191412, end_position_on_the_genomic_accession=85201673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728126, CurrentID=0, Status='live', Symbol='RPSAP30', Aliases='RPSA_7_452', description='ribosomal protein SA pseudogene 30', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196990592, end_position_on_the_genomic_accession=197000423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728128, CurrentID=0, Status='live', Symbol='RPL7AP26', Aliases='RPL7A_8_496', description='ribosomal protein L7a pseudogene 26', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82490763, end_position_on_the_genomic_accession=82491655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728130, CurrentID=0, Status='live', Symbol='NUTM2D', Aliases='FAM22D', description='NUT family member 2D', other_designations='NUT family member 2A pseudogene|family with sequence similarity 22, member D', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87357720, end_position_on_the_genomic_accession=87370695, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  728131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728131, CurrentID=0, Status='live', Symbol='RPS6P6', Aliases='', description='ribosomal protein S6 pseudogene', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82494806, end_position_on_the_genomic_accession=82495934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728132, CurrentID=0, Status='live', Symbol='TSPY9P', Aliases='TSPY9', description='testis specific protein, Y-linked 9, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9487267, end_position_on_the_genomic_accession=9488988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728134, CurrentID=0, Status='live', Symbol='RPL23AP18', Aliases='RPL23A_3_161', description='ribosomal protein L23a pseudogene 18', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212309051, end_position_on_the_genomic_accession=212309518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421306, CurrentID=0, Status='live', Symbol='LOC100421306', Aliases='', description='RecQ helicase-like pseudogene', other_designations='RecQ protein-like (DNA helicase Q1-like) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106824046, end_position_on_the_genomic_accession=106824954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728137, CurrentID=0, Status='live', Symbol='TSPY3', Aliases='CT78', description='testis specific protein, Y-linked 3', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9398421, end_position_on_the_genomic_accession=9401222, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728138, CurrentID=0, Status='live', Symbol='LOC728138', Aliases='', description='KIAA2013 pseudogene', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14988009, end_position_on_the_genomic_accession=14990326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728139, CurrentID=0, Status='live', Symbol='RPL7AP70', Aliases='RPL7A_32_1739', description='ribosomal protein L7a pseudogene 70', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19792275, end_position_on_the_genomic_accession=19793153, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728143, CurrentID=0, Status='live', Symbol='CHCHD4P4', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 4 pseudogene 4', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11234256, end_position_on_the_genomic_accession=11234680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728145, CurrentID=0, Status='live', Symbol='LOC728145', Aliases='', description='uncharacterized LOC728145', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=170732985, end_position_on_the_genomic_accession=170735809, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  203859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=203859, CurrentID=0, Status='live', Symbol='ANO5', Aliases='GDD1, LGMD2L, TMEM16E', description='anoctamin 5', other_designations='gnathodiaphyseal dysplasia 1 protein|integral membrane protein GDD1|transmembrane protein 16E', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22192513, end_position_on_the_genomic_accession=22283367, orientation='plus', exon_count=23, OMIM=608662, no_hdr0=''),
  105372174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372174, CurrentID=0, Status='live', Symbol='LOC105372174', Aliases='', description='uncharacterized LOC105372174', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67587245, end_position_on_the_genomic_accession=67623863, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129295, CurrentID=0, Status='live', Symbol='LOC100129295', Aliases='', description='ribosomal protein L22 pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185171337, end_position_on_the_genomic_accession=185171710, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728157, CurrentID=0, Status='live', Symbol='ANKRD20A6P', Aliases='', description='ankyrin repeat domain 20 family member A6, pseudogene', other_designations='ankyrin repeat domain 20 family, member A6, pseudogene', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11173618, end_position_on_the_genomic_accession=11174439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728158, CurrentID=0, Status='live', Symbol='LOC728158', Aliases='', description='hCG2044975', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126423112, end_position_on_the_genomic_accession=126433986, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100533937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533937, CurrentID=0, Status='live', Symbol='LOC100533937', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372176, CurrentID=0, Status='live', Symbol='LOC105372176', Aliases='', description='uncharacterized LOC105372176', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68753296, end_position_on_the_genomic_accession=68784216, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728162, CurrentID=0, Status='live', Symbol='ST13P11', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 11', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94913047, end_position_on_the_genomic_accession=94914137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421307, CurrentID=0, Status='live', Symbol='DDX43P1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 43 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115974505, end_position_on_the_genomic_accession=115975854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728167, CurrentID=0, Status='live', Symbol='KRT8P31', Aliases='', description='keratin 8 pseudogene 31', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60743317, end_position_on_the_genomic_accession=60745044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728169, CurrentID=0, Status='live', Symbol='DUX4L16', Aliases='DUXY1', description='double homeobox 4 like 16', other_designations='double homeobox, 4-like pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11306918, end_position_on_the_genomic_accession=11308191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728170, CurrentID=0, Status='live', Symbol='HNRNPA1P50', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 50', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92883692, end_position_on_the_genomic_accession=92884900, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874150, CurrentID=0, Status='live', Symbol='LINC00379', Aliases='', description='long intergenic non-protein coding RNA 379', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91127613, end_position_on_the_genomic_accession=91211698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372178, CurrentID=0, Status='live', Symbol='LOC105372178', Aliases='', description='uncharacterized LOC105372178', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=69715333, end_position_on_the_genomic_accession=69733135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728175, CurrentID=0, Status='live', Symbol='LOC728175', Aliases='', description='uncharacterized LOC728175', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184341031, end_position_on_the_genomic_accession=184353977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728178, CurrentID=0, Status='live', Symbol='LOC728178', Aliases='', description='uncharacterized LOC728178', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=14661917, end_position_on_the_genomic_accession=14664713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728179, CurrentID=0, Status='live', Symbol='RPL7AP36', Aliases='RPL7A_9_655', description='ribosomal protein L7a pseudogene 36', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10118434, end_position_on_the_genomic_accession=10119311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645652, CurrentID=0, Status='live', Symbol='MT1XP1', Aliases='MTL1', description='metallothionein 1X pseudogene 1', other_designations='metallothionein 1M pseudogene|metallothionein-like 1 pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16241216, end_position_on_the_genomic_accession=16241398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728190, CurrentID=0, Status='live', Symbol='NUTM2A-AS1', Aliases='FAM22A-AS1', description='NUTM2A antisense RNA 1', other_designations='FAM22A antisense RNA 1', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87238667, end_position_on_the_genomic_accession=87342558, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  728192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728192, CurrentID=0, Status='live', Symbol='LINC00460', Aliases='', description='long intergenic non-protein coding RNA 460', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106376563, end_position_on_the_genomic_accession=106377792, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728194, CurrentID=0, Status='live', Symbol='RSPH10B2', Aliases='', description='radial spoke head 10 homolog B2 (Chlamydomonas)', other_designations='radial spoke head 10 homolog B', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6754108, end_position_on_the_genomic_accession=6798775, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  728195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728195, CurrentID=0, Status='live', Symbol='LOC728195', Aliases='', description='fibroblast growth factor 7 pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42664677, end_position_on_the_genomic_accession=42668886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728196, CurrentID=0, Status='live', Symbol='LOC728196', Aliases='', description='uncharacterized LOC728196', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=111381327, end_position_on_the_genomic_accession=111384617, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372182, CurrentID=0, Status='live', Symbol='LOC105372182', Aliases='', description='uncharacterized LOC105372182', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728202, CurrentID=0, Status='live', Symbol='RPL36AP7', Aliases='RPL36A_24_1574', description='ribosomal protein L36a pseudogene 7', other_designations='hCG1789827', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75338800, end_position_on_the_genomic_accession=75339197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372183, CurrentID=0, Status='live', Symbol='LOC105372183', Aliases='', description='uncharacterized LOC105372183', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728207, CurrentID=0, Status='live', Symbol='RPL23AP75', Aliases='RPL23A_32_1546', description='ribosomal protein L23a pseudogene 75', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40439430, end_position_on_the_genomic_accession=40439935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728208, CurrentID=0, Status='live', Symbol='LOC728208', Aliases='', description='uncharacterized LOC728208', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=241920932, end_position_on_the_genomic_accession=241925514, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372184, CurrentID=0, Status='live', Symbol='LOC105372184', Aliases='', description='uncharacterized LOC105372184', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70756207, end_position_on_the_genomic_accession=70770863, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929135, CurrentID=0, Status='live', Symbol='LOC101929135', Aliases='', description='RWD domain-containing protein 2B pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171111586, end_position_on_the_genomic_accession=171114581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728215, CurrentID=0, Status='live', Symbol='FAM155A', Aliases='', description='family with sequence similarity 155 member A', other_designations='family with sequence similarity 155, member A', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107168531, end_position_on_the_genomic_accession=107867255, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373286, CurrentID=0, Status='live', Symbol='LOC105373286', Aliases='', description='uncharacterized LOC105373286', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=82119281, end_position_on_the_genomic_accession=82125542, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728218, CurrentID=0, Status='live', Symbol='LINC00864', Aliases='', description='long intergenic non-protein coding RNA 864', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87396574, end_position_on_the_genomic_accession=87407700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372186, CurrentID=0, Status='live', Symbol='LOC105372186', Aliases='', description='uncharacterized LOC105372186', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71642691, end_position_on_the_genomic_accession=71646621, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728224, CurrentID=0, Status='live', Symbol='KRTAP4-8', Aliases='KAP4.8, KRTAP4.8', description='keratin associated protein 4-8', other_designations='keratin associated protein 4.8|keratin-associated protein 4.8|ultrahigh sulfur keratin-associated protein 4.8', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41096981, end_position_on_the_genomic_accession=41098123, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421309, CurrentID=0, Status='live', Symbol='LOC100421309', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 25 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61347120, end_position_on_the_genomic_accession=61347713, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728226, CurrentID=0, Status='live', Symbol='GGTLC3', Aliases='GGT', description='gamma-glutamyltransferase light chain 3', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18516338, end_position_on_the_genomic_accession=18518161, orientation='minus', exon_count=0, OMIM=612340, no_hdr0=''),
  105372187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372187, CurrentID=0, Status='live', Symbol='LOC105372187', Aliases='', description='uncharacterized LOC105372187', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71215626, end_position_on_the_genomic_accession=71217974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728228, CurrentID=0, Status='live', Symbol='LINC01433', Aliases='', description='long intergenic non-protein coding RNA 1433', other_designations='hCG2045830', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4193090, end_position_on_the_genomic_accession=4195953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728229, CurrentID=0, Status='live', Symbol='TMEM191B', Aliases='', description='transmembrane protein 191B', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18527802, end_position_on_the_genomic_accession=18530794, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  728231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728231, CurrentID=0, Status='live', Symbol='UBTFL5', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 5 (pseudogene)', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95450305, end_position_on_the_genomic_accession=95451486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728233, CurrentID=0, Status='live', Symbol='PI4KAP1', Aliases='', description='phosphatidylinositol 4-kinase, catalytic, alpha pseudogene 1', other_designations='phosphatidylinositol 4-kinase, catalytic, alpha polypeptide pseudogene 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18533864, end_position_on_the_genomic_accession=18548828, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  728234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728234, CurrentID=0, Status='live', Symbol='HMGN2P16', Aliases='HMGN2L', description='high mobility group nucleosomal binding domain 2 pseudogene 16', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 16', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15588355, end_position_on_the_genomic_accession=15588615, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728239, CurrentID=0, Status='live', Symbol='MAGED4', Aliases='MAGE-E1, MAGE1, MAGEE1', description='melanoma antigen family D4', other_designations='MAGE-D4 antigen|MAGE-E1 antigen|melanoma antigen family D, 4', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52184823, end_position_on_the_genomic_accession=52192268, orientation='plus', exon_count=14, OMIM=300702, no_hdr0=''),
  101927069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927069, CurrentID=0, Status='live', Symbol='LOC101927069', Aliases='', description='uncharacterized LOC101927069', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68953088, end_position_on_the_genomic_accession=68975841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728244, CurrentID=0, Status='live', Symbol='RPS12P29', Aliases='RPS12_15_1507', description='ribosomal protein S12 pseudogene 29', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4704799, end_position_on_the_genomic_accession=4705299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129310, CurrentID=0, Status='live', Symbol='RPL31P14', Aliases='RPL31_4_324', description='ribosomal protein L31 pseudogene 14', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216785735, end_position_on_the_genomic_accession=216786179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  28614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=28614, CurrentID=0, Status='live', Symbol='TRBV5-1', Aliases='TCRBV5S1, TCRBV5S1A1T, TRBV51', description='T cell receptor beta variable 5-1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142321075, end_position_on_the_genomic_accession=142321544, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129311, CurrentID=0, Status='live', Symbol='LOC100129311', Aliases='', description='uncharacterized LOC100129311', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.7', start_position_on_the_genomic_accession=42877289, end_position_on_the_genomic_accession=42878109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728254, CurrentID=0, Status='live', Symbol='LOC728254', Aliases='', description='uncharacterized LOC728254', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=114500897, end_position_on_the_genomic_accession=114504749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728255, CurrentID=0, Status='live', Symbol='KRTAP1-4', Aliases='KAP1.4, KRTAP1.4', description='keratin associated protein 1-4', other_designations='high sulfur keratin-associated protein 1.4|keratin-associated protein 1.4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41029677, end_position_on_the_genomic_accession=41030125, orientation='minus', exon_count=1, OMIM=608821, no_hdr0=''),
  105372192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372192, CurrentID=0, Status='live', Symbol='LOC105372192', Aliases='', description='uncharacterized LOC105372192', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74454602, end_position_on_the_genomic_accession=74456948, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728262, CurrentID=0, Status='live', Symbol='FAM157A', Aliases='', description='family with sequence similarity 157 member A', other_designations='family with sequence similarity 157, member A', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198151974, end_position_on_the_genomic_accession=198211810, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  728263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728263, CurrentID=0, Status='live', Symbol='ALG1L7P', Aliases='', description='asparagine-linked glycosylation 1-like 7, pseudogene', other_designations='asparagine-linked glycosylation 1 pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3935446, end_position_on_the_genomic_accession=3945048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728264, CurrentID=0, Status='live', Symbol='CARMN', Aliases='CARMEN, MIR143HG', description='cardiac mesoderm enhancer-associated non-coding RNA', other_designations='MIR143 host gene (non-protein coding)|cardiac mesoderm enhancer-associated noncoding RNA', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149406845, end_position_on_the_genomic_accession=149432836, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728269, CurrentID=0, Status='live', Symbol='MAGEA9B', Aliases='', description='MAGE family member A9B', other_designations='CT1.9|MAGE-9 antigen|cancer/testis antigen 1.9|melanoma antigen family A, 9B|melanoma antigen family A9B', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149581651, end_position_on_the_genomic_accession=149587468, orientation='minus', exon_count=5, OMIM=300764, no_hdr0=''),
  101927070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927070, CurrentID=0, Status='live', Symbol='LOC101927070', Aliases='', description='uncharacterized LOC101927070', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98761063, end_position_on_the_genomic_accession=98771898, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372195, CurrentID=0, Status='live', Symbol='LOC105372195', Aliases='', description='uncharacterized LOC105372195', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75066354, end_position_on_the_genomic_accession=75074205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728276, CurrentID=0, Status='live', Symbol='CLEC19A', Aliases='', description='C-type lectin domain family 19 member A', other_designations='C-type lectin domain family 19, member A', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19285755, end_position_on_the_genomic_accession=19324537, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728279, CurrentID=0, Status='live', Symbol='KRTAP2-2', Aliases='KAP2.2', description='keratin associated protein 2-2', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41054498, end_position_on_the_genomic_accession=41055211, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372196, CurrentID=0, Status='live', Symbol='LOC105372196', Aliases='', description='uncharacterized LOC105372196', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74895671, end_position_on_the_genomic_accession=74936725, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728283, CurrentID=0, Status='live', Symbol='LOC728283', Aliases='', description='uncharacterized LOC728283', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5315582, end_position_on_the_genomic_accession=5317717, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728287, CurrentID=0, Status='live', Symbol='LOC728287', Aliases='', description='uncharacterized LOC728287', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=149359596, end_position_on_the_genomic_accession=149361141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728290, CurrentID=0, Status='live', Symbol='LOC728290', Aliases='', description='uncharacterized LOC728290', other_designations='', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80993834, end_position_on_the_genomic_accession=81095425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728294, CurrentID=0, Status='live', Symbol='D2HGDH', Aliases='D2HGD', description='D-2-hydroxyglutarate dehydrogenase', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241734579, end_position_on_the_genomic_accession=241768816, orientation='plus', exon_count=20, OMIM=609186, no_hdr0=''),
  204010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204010, CurrentID=0, Status='live', Symbol='RPSAP52', Aliases='RPSA_17_1251', description='ribosomal protein SA pseudogene 52', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65758020, end_position_on_the_genomic_accession=65826974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728299, CurrentID=0, Status='live', Symbol='KRTAP19-8', Aliases='', description='keratin associated protein 19-8', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31038159, end_position_on_the_genomic_accession=31038476, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728300, CurrentID=0, Status='live', Symbol='KRT18P5', Aliases='', description='keratin 18 pseudogene 5', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20482748, end_position_on_the_genomic_accession=20484155, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372200, CurrentID=0, Status='live', Symbol='LOC105372200', Aliases='', description='uncharacterized LOC105372200', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75406947, end_position_on_the_genomic_accession=75425040, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728307, CurrentID=0, Status='live', Symbol='LOC728307', Aliases='', description='melanoma associated antigen (mutated) 1-like 1 pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153122086, end_position_on_the_genomic_accession=153128423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100533938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533938, CurrentID=0, Status='live', Symbol='LOC100533938', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728310, CurrentID=0, Status='live', Symbol='GOLGA6L7P', Aliases='GOLGA6L7', description='golgin A6 family-like 7, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 6-like 7 (pseudogene)|golgin A6 family-like 7 (pseudogene)|golgin A6 family-like 9 pseudogene', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28844961, end_position_on_the_genomic_accession=28848675, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  728311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728311, CurrentID=0, Status='live', Symbol='RPL26P24', Aliases='RPL26_12_861', description='ribosomal protein L26 pseudogene 24', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143796253, end_position_on_the_genomic_accession=143796769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100736250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100736250, CurrentID=0, Status='live', Symbol='HYPT9', Aliases='', description='Hypotrichosis 9', other_designations='', map_location='10q11.23-q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614237, no_hdr0=''),
  100736251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100736251, CurrentID=0, Status='live', Symbol='HYP10', Aliases='', description='Hypotrichosis 10', other_designations='', map_location='7p22.3-p21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614238, no_hdr0=''),
  728317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728317, CurrentID=0, Status='live', Symbol='LOC728317', Aliases='', description='melanoma associated antigen (mutated) 1-like 1 pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153271618, end_position_on_the_genomic_accession=153277755, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728318, CurrentID=0, Status='live', Symbol='KRTAP9-1', Aliases='KAP9.1, KRTAP9L3', description='keratin associated protein 9-1', other_designations='Putative keratin-associated protein 9-2-like 3|keratin associated protein 9-like 3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41189887, end_position_on_the_genomic_accession=41190639, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728323, CurrentID=0, Status='live', Symbol='LOC728323', Aliases='', description='uncharacterized LOC728323', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242088633, end_position_on_the_genomic_accession=242160503, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  728324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728324, CurrentID=0, Status='live', Symbol='CHCHD2P2', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 2', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69333866, end_position_on_the_genomic_accession=69334627, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728325, CurrentID=0, Status='live', Symbol='HMGN1P21', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 21', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 21', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3988354, end_position_on_the_genomic_accession=3989603, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728327, CurrentID=0, Status='live', Symbol='LOC728327', Aliases='', description='hCG1795091', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128912830, end_position_on_the_genomic_accession=128916286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372204, CurrentID=0, Status='live', Symbol='LOC105372204', Aliases='', description='uncharacterized LOC105372204', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76053684, end_position_on_the_genomic_accession=76056412, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927072, CurrentID=0, Status='live', Symbol='LOC101927072', Aliases='', description='uncharacterized LOC101927072', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68025812, end_position_on_the_genomic_accession=68043675, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728333, CurrentID=0, Status='live', Symbol='LOC728333', Aliases='', description='nuclear receptor coactivator 4 pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88507003, end_position_on_the_genomic_accession=88510487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372205, CurrentID=0, Status='live', Symbol='LOC105372205', Aliases='', description='uncharacterized LOC105372205', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728339, CurrentID=0, Status='live', Symbol='LOC728339', Aliases='', description='uncharacterized LOC728339', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189863452, end_position_on_the_genomic_accession=189940713, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728340, CurrentID=0, Status='live', Symbol='GTF2H2C', Aliases='GTF2H2C_2', description='GTF2H2 family member C', other_designations='general transcription factor IIH polypeptide 2-like protein|general transcription factor IIH, polypeptide 2C', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69560143, end_position_on_the_genomic_accession=69592903, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  728342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728342, CurrentID=0, Status='live', Symbol='LINC00992', Aliases='', description='long intergenic non-protein coding RNA 992', other_designations='CTC-504A5.1', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117415512, end_position_on_the_genomic_accession=117579744, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728343, CurrentID=0, Status='live', Symbol='NXF2B', Aliases='TCP11X1, bA353J17.1', description='nuclear RNA export factor 2B', other_designations='TAP-like protein 2|cancer/testis antigen 39', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102360395, end_position_on_the_genomic_accession=102440008, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  728344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728344, CurrentID=0, Status='live', Symbol='GLRX3P2', Aliases='', description='glutaredoxin 3 pseudogene 2', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3978041, end_position_on_the_genomic_accession=3979369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372207, CurrentID=0, Status='live', Symbol='LOC105372207', Aliases='', description='uncharacterized LOC105372207', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76209203, end_position_on_the_genomic_accession=76217864, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  728350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728350, CurrentID=0, Status='live', Symbol='EIF2S2P4', Aliases='', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 4', other_designations='eukaryotic translation initiation factor 2 beta pseudogene|eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 4', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170751709, end_position_on_the_genomic_accession=170753078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728353, CurrentID=0, Status='live', Symbol='LOC728353', Aliases='', description='uncharacterized LOC728353', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728358, CurrentID=0, Status='live', Symbol='DEFA1B', Aliases='HNP-1, HP-1, HP1', description='defensin alpha 1B', other_designations='defensin, alpha 1B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6996766, end_position_on_the_genomic_accession=6999202, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372209, CurrentID=0, Status='live', Symbol='LOC105372209', Aliases='', description='uncharacterized LOC105372209', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76233241, end_position_on_the_genomic_accession=76260114, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  728361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728361, CurrentID=0, Status='live', Symbol='OVOL3', Aliases='HOVO3', description='ovo-like zinc finger 3', other_designations='OVO homolog-like 3|ovo-like 3', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36108829, end_position_on_the_genomic_accession=36113720, orientation='plus', exon_count=5, OMIM=616442, no_hdr0=''),
  728365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728365, CurrentID=0, Status='live', Symbol='RPL6P16', Aliases='RPL6_12_538', description='ribosomal protein L6 pseudogene 16', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184254535, end_position_on_the_genomic_accession=184255321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728366, CurrentID=0, Status='live', Symbol='BCLAF1P1', Aliases='', description='BCL2-associated transcription factor 1 pseudogene 1', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110946273, end_position_on_the_genomic_accession=110951024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728369, CurrentID=0, Status='live', Symbol='USP17L24', Aliases='', description='ubiquitin specific peptidase 17-like family member 24', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9325165, end_position_on_the_genomic_accession=9326757, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129331, CurrentID=0, Status='live', Symbol='RAB28P2', Aliases='', description='RAB28, member RAS oncogene family pseudogene 2', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62605958, end_position_on_the_genomic_accession=62607138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728373, CurrentID=0, Status='live', Symbol='USP17L25', Aliases='', description='ubiquitin specific peptidase 17-like family member 25', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9329911, end_position_on_the_genomic_accession=9331503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728374, CurrentID=0, Status='live', Symbol='RPL9P21', Aliases='RPL9_8_1037', description='ribosomal protein L9 pseudogene 21', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44414415, end_position_on_the_genomic_accession=44415202, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728375, CurrentID=0, Status='live', Symbol='LOC728375', Aliases='', description='uncharacterized LOC728375', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=168128040, end_position_on_the_genomic_accession=168128630, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100421314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421314, CurrentID=0, Status='live', Symbol='LOC100421314', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728377, CurrentID=0, Status='live', Symbol='ARHGEF34P', Aliases='', description='Rho guanine nucleotide exchange factor 34, pseudogene', other_designations='Rho guanine nucleotide exchange factor (GEF) 34, pseudogene|Rho guanine nucleotide exchange factor (GEF) 5 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144258996, end_position_on_the_genomic_accession=144286003, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  728378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728378, CurrentID=0, Status='live', Symbol='POTEF', Aliases='A26C1B, ACTB, POTE2alpha, POTEACTIN', description='POTE ankyrin domain family member F', other_designations='ANKRD26-like family C member 1B|ANKRD26-like family C, member 1B|POTE ankyrin domain family, member F|POTE-2 alpha-actin|beta actin|chimeric POTE-actin protein|prostate, ovary, testis expressed protein on chromosome 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130073479, end_position_on_the_genomic_accession=130129222, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  728379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728379, CurrentID=0, Status='live', Symbol='USP17L26', Aliases='', description='ubiquitin specific peptidase 17-like family member 26', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9334658, end_position_on_the_genomic_accession=9336250, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728380, CurrentID=0, Status='live', Symbol='RPL7P26', Aliases='RPL7_9_661', description='ribosomal protein L7 pseudogene 26', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17530828, end_position_on_the_genomic_accession=17531457, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728381, CurrentID=0, Status='live', Symbol='NRBF2P4', Aliases='', description='nuclear receptor binding factor 2 pseudogene 4', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107982969, end_position_on_the_genomic_accession=107984766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728386, CurrentID=0, Status='live', Symbol='USP17L5', Aliases='', description='ubiquitin specific peptidase 17-like family member 5', other_designations='deubiquitinating enzyme 17-like protein 5|ubiquitin specific peptidase 17-like 5|ubiquitin thioesterase 17-like protein 5|ubiquitin thiolesterase 17-like protein 5|ubiquitin-specific-processing protease 17-like protein 5', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9339403, end_position_on_the_genomic_accession=9340995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728388, CurrentID=0, Status='live', Symbol='DNM1P30', Aliases='DNM1DN3-4, DNM1DN3.4, DNM1DN3@, DNM1DN4-3, DNM1DN4@', description='dynamin 1 pseudogene 30', other_designations='DNM1 pseudogene 30|DNM1DN3.4 duplicon', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30150897, end_position_on_the_genomic_accession=30154110, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372214, CurrentID=0, Status='live', Symbol='LOC105372214', Aliases='', description='uncharacterized LOC105372214', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76842732, end_position_on_the_genomic_accession=76858167, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728392, CurrentID=0, Status='live', Symbol='LOC728392', Aliases='', description='uncharacterized LOC728392', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5499427, end_position_on_the_genomic_accession=5500999, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728393, CurrentID=0, Status='live', Symbol='USP17L27', Aliases='', description='ubiquitin specific peptidase 17-like family member 27', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9344148, end_position_on_the_genomic_accession=9345740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728395, CurrentID=0, Status='live', Symbol='TSPY4', Aliases='TSPY10, TSPY8', description='testis specific protein, Y-linked 4', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9337483, end_position_on_the_genomic_accession=9340283, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728400, CurrentID=0, Status='live', Symbol='USP17L28', Aliases='', description='ubiquitin specific peptidase 17-like family member 28', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9348893, end_position_on_the_genomic_accession=9350485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372216, CurrentID=0, Status='live', Symbol='LOC105372216', Aliases='', description='uncharacterized LOC105372216', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77062783, end_position_on_the_genomic_accession=77065933, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728402, CurrentID=0, Status='live', Symbol='TPI1P3', Aliases='', description='triosephosphate isomerase 1 pseudogene 3', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116038731, end_position_on_the_genomic_accession=116039944, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728403, CurrentID=0, Status='live', Symbol='TSPY8', Aliases='', description='testis specific protein, Y-linked 8', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9357828, end_position_on_the_genomic_accession=9360598, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728405, CurrentID=0, Status='live', Symbol='USP17L29', Aliases='', description='ubiquitin specific peptidase 17-like family member 29', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9353638, end_position_on_the_genomic_accession=9355230, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728407, CurrentID=0, Status='live', Symbol='PARGP1', Aliases='', description='poly(ADP-ribose) glycohydrolase pseudogene 1', other_designations='PRO1102|poly (ADP-ribose) glycohydrolase pseudogene 1', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45854835, end_position_on_the_genomic_accession=45972428, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  728409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728409, CurrentID=0, Status='live', Symbol='LINC01548', Aliases='C21orf54', description='long intergenic non-protein coding RNA 1548', other_designations='hCG2045804', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33165470, end_position_on_the_genomic_accession=33170236, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728410, CurrentID=0, Status='live', Symbol='DUX4L2', Aliases='DUX4', description='double homeobox 4 like 2', other_designations='double homeobox protein 4-like protein 2', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190091005, end_position_on_the_genomic_accession=190092289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728411, CurrentID=0, Status='live', Symbol='GUSBP1', Aliases='', description='glucuronidase, beta pseudogene 1', other_designations='Putative beta-glucuronidase-like protein MGC87375', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21459480, end_position_on_the_genomic_accession=21589372, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  728416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728416, CurrentID=0, Status='live', Symbol='LOC728416', Aliases='', description='translation machinery associated 7 homolog (S. cerevisiae) pseudogene', other_designations='translational machinery associated 7 homolog pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56736210, end_position_on_the_genomic_accession=56736611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728417, CurrentID=0, Status='live', Symbol='LOC728417', Aliases='', description='uncharacterized LOC728417', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222471774, end_position_on_the_genomic_accession=222477700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728418, CurrentID=0, Status='live', Symbol='POM121L7', Aliases='', description='POM121 transmembrane nucleoporin like 7', other_designations='POM121 membrane glycoprotein-like 1 pseudogene|POM121 membrane glycoprotein-like 7|POM121 transmembrane nucleoporin-like 7', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21112788, end_position_on_the_genomic_accession=21128092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728419, CurrentID=0, Status='live', Symbol='USP17L30', Aliases='', description='ubiquitin specific peptidase 17-like family member 30', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9363129, end_position_on_the_genomic_accession=9364721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728424, CurrentID=0, Status='live', Symbol='LOC728424', Aliases='', description='hyperpolarization activated cyclic nucleotide gated potassium channel 2 pseudogene', other_designations='hyperpolarization activated cyclic nucleotide-gated potassium channel 2 pseudogene', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30223068, end_position_on_the_genomic_accession=30226244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372220, CurrentID=0, Status='live', Symbol='LOC105372220', Aliases='', description='uncharacterized LOC105372220', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78556094, end_position_on_the_genomic_accession=78558962, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728428, CurrentID=0, Status='live', Symbol='RPS15P1', Aliases='RPS15_1_87', description='ribosomal protein S15 pseudogene 1', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113198807, end_position_on_the_genomic_accession=113199291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728430, CurrentID=0, Status='live', Symbol='FAM90A20P', Aliases='FAM90A20', description='family with sequence similarity 90 member A20, pseudogene', other_designations='family with sequence similarity 90, member A20, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7293354, end_position_on_the_genomic_accession=7298584, orientation='plus', exon_count=0, OMIM=613054, no_hdr0=''),
  728431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728431, CurrentID=0, Status='live', Symbol='LINC01137', Aliases='', description='long intergenic non-protein coding RNA 1137', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37454879, end_position_on_the_genomic_accession=37474443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728432, CurrentID=0, Status='live', Symbol='ICE2P2', Aliases='NARG2P2', description='interactor of little elongation complex ELL subunit 2 pseudogene 2', other_designations='NMDA receptor regulated 2 pseudogene 2|interactor of little elongator complex ELL subunit 2 pseudogene 2', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88948198, end_position_on_the_genomic_accession=88950993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728433, CurrentID=0, Status='live', Symbol='LOC728433', Aliases='', description='fibroblast growth factor 7 pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62472124, end_position_on_the_genomic_accession=62476333, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728434, CurrentID=0, Status='live', Symbol='C20orf187', Aliases='dJ727I10.1', description='chromosome 20 open reading frame 187', other_designations='hCG2045828|putative uncharacterized protein C20orf187', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10862085, end_position_on_the_genomic_accession=11029529, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105372222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372222, CurrentID=0, Status='live', Symbol='LOC105372222', Aliases='', description='uncharacterized LOC105372222', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78669937, end_position_on_the_genomic_accession=78670997, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728441, CurrentID=0, Status='live', Symbol='GGT2', Aliases='GGT, GGT 2', description='gamma-glutamyltransferase 2', other_designations='gamma-glutamyltranspeptidase 2|glutathione hydrolase 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21207972, end_position_on_the_genomic_accession=21239982, orientation='minus', exon_count=21, OMIM=137181, no_hdr0=''),
  728445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728445, CurrentID=0, Status='live', Symbol='LOC728445', Aliases='', description='uncharacterized LOC728445', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=42128447, end_position_on_the_genomic_accession=42130267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728447, CurrentID=0, Status='live', Symbol='TEX28P1', Aliases='pTEX', description='testis expressed 28 pseudogene 1', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154196321, end_position_on_the_genomic_accession=154217295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728448, CurrentID=0, Status='live', Symbol='PPIEL', Aliases='PPIEP1', description='peptidylprolyl isomerase E-like pseudogene', other_designations='peptidylprolyl isomerase E (cyclophilin E) pseudogene 1|peptidylprolyl isomerase E pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39522280, end_position_on_the_genomic_accession=39559698, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100129344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129344, CurrentID=0, Status='live', Symbol='LOC100129344', Aliases='', description='developmental pluripotency associated 4 pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10199829, end_position_on_the_genomic_accession=10200595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728450, CurrentID=0, Status='live', Symbol='LINC00687', Aliases='C20orf61, dJ1012F16.1', description='long intergenic non-protein coding RNA 687', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11809987, end_position_on_the_genomic_accession=11870715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728452, CurrentID=0, Status='live', Symbol='LOC728452', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71215363, end_position_on_the_genomic_accession=71216052, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728453, CurrentID=0, Status='live', Symbol='RPS28P6', Aliases='RPS28_2_814', description='ribosomal protein S28 pseudogene 6', other_designations='', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71942233, end_position_on_the_genomic_accession=71942616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372225, CurrentID=0, Status='live', Symbol='LOC105372225', Aliases='', description='uncharacterized LOC105372225', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728458, CurrentID=0, Status='live', Symbol='OPN1MW2', Aliases='GOP', description='opsin 1 (cone pigments), medium-wave-sensitive 2', other_designations='green cone photoreceptor pigment|green-sensitive opsin', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154219734, end_position_on_the_genomic_accession=154233286, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728463, CurrentID=0, Status='live', Symbol='TGFB2-AS1', Aliases='', description='TGFB2 antisense RNA 1 (head to head)', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218344196, end_position_on_the_genomic_accession=218345678, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728464, CurrentID=0, Status='live', Symbol='METTL24', Aliases='C6orf186, dJ71D21.2', description='methyltransferase like 24', other_designations='UPF0624 protein C6orf186', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110245946, end_position_on_the_genomic_accession=110358272, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728467, CurrentID=0, Status='live', Symbol='SNX5P2', Aliases='', description='sorting nexin 5 pseudogene 2', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101066607, end_position_on_the_genomic_accession=101070000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728468, CurrentID=0, Status='live', Symbol='BCRP6', Aliases='', description='breakpoint cluster region pseudogene 6', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21288229, end_position_on_the_genomic_accession=21294589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728470, CurrentID=0, Status='live', Symbol='LOC728470', Aliases='', description='cancer/testis antigen 55 pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135179499, end_position_on_the_genomic_accession=135195333, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100736408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100736408, CurrentID=0, Status='live', Symbol='LOC100736408', Aliases='', description='NAD kinase 2, mitochondrial pseudogene', other_designations='NAD kinase domain containing 1 pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22103509, end_position_on_the_genomic_accession=22104795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129348, CurrentID=0, Status='live', Symbol='CDCA4P2', Aliases='', description='cell division cycle associated 4 pseudogene 2', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86551316, end_position_on_the_genomic_accession=86553563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100736411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100736411, CurrentID=0, Status='live', Symbol='LOC100736411', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila) pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22339726, end_position_on_the_genomic_accession=22341167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100736412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100736412, CurrentID=0, Status='live', Symbol='LOC100736412', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila) pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22446836, end_position_on_the_genomic_accession=22449007, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728477, CurrentID=0, Status='live', Symbol='GOLGA2P8', Aliases='', description='golgin A2 pseudogene 8', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90293500, end_position_on_the_genomic_accession=90300975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728481, CurrentID=0, Status='live', Symbol='RPL23AP21', Aliases='RPL23A_1_1', description='ribosomal protein L23a pseudogene 21', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=258019, end_position_on_the_genomic_accession=259027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728484, CurrentID=0, Status='live', Symbol='RPL23AP61', Aliases='RPL23A_22_1044', description='ribosomal protein L23a pseudogene 61', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46063250, end_position_on_the_genomic_accession=46063720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728485, CurrentID=0, Status='live', Symbol='LOC728485', Aliases='', description='uncharacterized LOC728485', other_designations='CTD-2162K18.4', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36773118, end_position_on_the_genomic_accession=36777078, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728486, CurrentID=0, Status='live', Symbol='RPL7AP64', Aliases='RPL7A_29_1509', description='ribosomal protein L7a pseudogene 64', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7140945, end_position_on_the_genomic_accession=7141592, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728488, CurrentID=0, Status='live', Symbol='LOC728488', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69631756, end_position_on_the_genomic_accession=69636953, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728489, CurrentID=0, Status='live', Symbol='DNLZ', Aliases='C9orf151, HEP, HEP1, TIMM15, ZIM17, bA413M3.2', description='DNL-type zinc finger', other_designations='HSP70 escort protein|RP11-413M3.2|hsp70-escort protein 1|mtHsp70-escort protein|translocase of inner mitochondrial membrane 15 homolog', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136361900, end_position_on_the_genomic_accession=136363811, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372231, CurrentID=0, Status='live', Symbol='LOC105372231', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728492, CurrentID=0, Status='live', Symbol='SERF1B', Aliases='FAM2B, H4F5C, h4F5', description='small EDRK-rich factor 1B (centromeric)', other_designations='SMA modifier 1|protein 4F5', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70025251, end_position_on_the_genomic_accession=70043108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728493, CurrentID=0, Status='live', Symbol='DPPA3P1', Aliases='', description='developmental pluripotency associated 3 pseudogene 1', other_designations='developmental pluripotency-associated 3 pseudogene', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103961328, end_position_on_the_genomic_accession=103962389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728495, CurrentID=0, Status='live', Symbol='FAM74A3', Aliases='', description='family with sequence similarity 74 member A3', other_designations='family with sequence similarity 74, member A3', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66970162, end_position_on_the_genomic_accession=66977317, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129352, CurrentID=0, Status='live', Symbol='LOC100129352', Aliases='', description='protein PAT1 homolog 1-like', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128444588, end_position_on_the_genomic_accession=128446631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728498, CurrentID=0, Status='live', Symbol='GOLGA8H', Aliases='GOLGA6L11', description='golgin A8 family member H', other_designations='Golgin subfamily A member 8-like protein 1|golgi autoantigen, golgin subfamily a, 6-like 11|golgin A8 family, member H', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30545870, end_position_on_the_genomic_accession=30617827, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  728499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728499, CurrentID=0, Status='live', Symbol='LOC728499', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69653308, end_position_on_the_genomic_accession=69653997, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728501, CurrentID=0, Status='live', Symbol='RPL21P75', Aliases='RPL21_29_789', description='ribosomal protein L21 pseudogene 75', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20002725, end_position_on_the_genomic_accession=20003292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728505, CurrentID=0, Status='live', Symbol='WDR82P1', Aliases='SW2, SWD2, WDR82, WDR82B, hCG26824', description='WD repeat domain 82 pseudogene 1', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94935599, end_position_on_the_genomic_accession=94938284, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728506, CurrentID=0, Status='live', Symbol='LOC728506', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69898215, end_position_on_the_genomic_accession=69903409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  204219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204219, CurrentID=0, Status='live', Symbol='CERS3', Aliases='ARCI9, LASS3', description='ceramide synthase 3', other_designations='LAG1 homolog, ceramide synthase 3|LAG1 longevity assurance homolog 3|dihydroceramide synthase 3', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100400395, end_position_on_the_genomic_accession=100544786, orientation='minus', exon_count=16, OMIM=615276, no_hdr0=''),
  728509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728509, CurrentID=0, Status='live', Symbol='RPS19P7', Aliases='RPS19_3_1038', description='ribosomal protein S19 pseudogene 7', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45073089, end_position_on_the_genomic_accession=45075906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375049, CurrentID=0, Status='live', Symbol='LOC105375049', Aliases='', description='uncharacterized LOC105375049', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40164686, end_position_on_the_genomic_accession=40171667, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927078, CurrentID=0, Status='live', Symbol='LOC101927078', Aliases='', description='uncharacterized LOC101927078', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=114447418, end_position_on_the_genomic_accession=114773413, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  728513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728513, CurrentID=0, Status='live', Symbol='PNMA6B', Aliases='PNMA6D', description='paraneoplastic Ma antigen family member 6B (pseudogene)', other_designations='paraneoplastic Ma antigen family member 6D (pseudogene)|paraneoplastic antigen like 6B|paraneoplastic antigen like 6D', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153075613, end_position_on_the_genomic_accession=153077617, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644686, CurrentID=0, Status='live', Symbol='LOC644686', Aliases='', description='uncharacterized LOC644686', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372235, CurrentID=0, Status='live', Symbol='LOC105372235', Aliases='', description='uncharacterized LOC105372235', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1007927, end_position_on_the_genomic_accession=1008665, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728517, CurrentID=0, Status='live', Symbol='RPL23AP24', Aliases='RPL23A_1_2', description='ribosomal protein L23a pseudogene 24', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=348069, end_position_on_the_genomic_accession=348323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728519, CurrentID=0, Status='live', Symbol='LOC728519', Aliases='BIRC, NAIP1B', description='NLR family, apoptosis inhibitory protein pseudogene', other_designations='neuronal apoptosis inhibitory protein (centromeric)', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70093038, end_position_on_the_genomic_accession=70128437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728523, CurrentID=0, Status='live', Symbol='RPL9P17', Aliases='RPL9_5_548', description='ribosomal protein L9 pseudogene 17', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31822460, end_position_on_the_genomic_accession=31823164, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728524, CurrentID=0, Status='live', Symbol='SPDYE8P', Aliases='', description='speedy/RINGO cell cycle regulator family member E8, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E pseudogene|speedy homolog E8, pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73019731, end_position_on_the_genomic_accession=73029762, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  728526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728526, CurrentID=0, Status='live', Symbol='LOC728526', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70233049, end_position_on_the_genomic_accession=70233738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372237, CurrentID=0, Status='live', Symbol='LOC105372237', Aliases='', description='uncharacterized LOC105372237', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1747996, end_position_on_the_genomic_accession=1748745, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728532, CurrentID=0, Status='live', Symbol='DYNC1I2P1', Aliases='', description='dynein, cytoplasmic 1, intermediate chain 2 pseudogene 1', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50264964, end_position_on_the_genomic_accession=50267012, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372238, CurrentID=0, Status='live', Symbol='LOC105372238', Aliases='', description='uncharacterized LOC105372238', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1940114, end_position_on_the_genomic_accession=1940782, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728535, CurrentID=0, Status='live', Symbol='LOC728535', Aliases='', description='NLR family, apoptosis inhibitory protein pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70458612, end_position_on_the_genomic_accession=70478581, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728537, CurrentID=0, Status='live', Symbol='LINC01125', Aliases='', description='long intergenic non-protein coding RNA 1125', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97669743, end_position_on_the_genomic_accession=97703066, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  728538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728538, CurrentID=0, Status='live', Symbol='RPL23AP77', Aliases='RPL23A_33_1589', description='ribosomal protein L23a pseudogene 77', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23709798, end_position_on_the_genomic_accession=23710298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728540, CurrentID=0, Status='live', Symbol='LOC728540', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70529425, end_position_on_the_genomic_accession=70530114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927079, CurrentID=0, Status='live', Symbol='LOC101927079', Aliases='', description='uncharacterized LOC101927079', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21990080, end_position_on_the_genomic_accession=22095613, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  728549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728549, CurrentID=0, Status='live', Symbol='LOC728549', Aliases='', description='calcium binding protein P22 pseudogene', other_designations='calcineurin-like EF-hand protein 1 pseudogene', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73019603, end_position_on_the_genomic_accession=73021387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645713, CurrentID=0, Status='live', Symbol='RPS10P23', Aliases='RPS10_13_1498', description='ribosomal protein S10 pseudogene 23', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74552214, end_position_on_the_genomic_accession=74552792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728553, CurrentID=0, Status='live', Symbol='RPS8P8', Aliases='RPS8_1_552', description='ribosomal protein S8 pseudogene 8', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33162156, end_position_on_the_genomic_accession=33162859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728554, CurrentID=0, Status='live', Symbol='LOC728554', Aliases='', description='THO complex 3 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177875261, end_position_on_the_genomic_accession=177884268, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728556, CurrentID=0, Status='live', Symbol='MEMO1P1', Aliases='C21orf19, C2orf4P, MEMO1P, rcC2orf4', description='mediator of cell motility 1 pseudogene 1', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36130072, end_position_on_the_genomic_accession=36131916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728558, CurrentID=0, Status='live', Symbol='ENTPD1-AS1', Aliases='', description='ENTPD1 antisense RNA 1', other_designations='weakly similar to zinc finger protein 195', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96021200, end_position_on_the_genomic_accession=96022588, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  728562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728562, CurrentID=0, Status='live', Symbol='RPSAP45', Aliases='RPSA_11_739', description='ribosomal protein SA pseudogene 45', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112355752, end_position_on_the_genomic_accession=112356793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372243, CurrentID=0, Status='live', Symbol='LOC105372243', Aliases='', description='uncharacterized LOC105372243', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3121311, end_position_on_the_genomic_accession=3125016, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728564, CurrentID=0, Status='live', Symbol='GMPSP1', Aliases='', description='guanine monophosphate synthase pseudogene 1', other_designations='guanine monophosphate synthetase pseudogene 1', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8174421, end_position_on_the_genomic_accession=8175020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728567, CurrentID=0, Status='live', Symbol='RPL23AP48', Aliases='RPL23A_16_748', description='ribosomal protein L23a pseudogene 48', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121679972, end_position_on_the_genomic_accession=121680459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728568, CurrentID=0, Status='live', Symbol='C12orf73', Aliases='', description='chromosome 12 open reading frame 73', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103950202, end_position_on_the_genomic_accession=103957215, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  728572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728572, CurrentID=0, Status='live', Symbol='RPL32P11', Aliases='RPL32_2_350', description='ribosomal protein L32 pseudogene 11', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25749458, end_position_on_the_genomic_accession=25749896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728573, CurrentID=0, Status='live', Symbol='LOC728573', Aliases='', description='uncharacterized LOC728573', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=11665792, end_position_on_the_genomic_accession=11668000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372245, CurrentID=0, Status='live', Symbol='LOC105372245', Aliases='', description='uncharacterized LOC105372245', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3496115, end_position_on_the_genomic_accession=3500521, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  728576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728576, CurrentID=0, Status='live', Symbol='RPL15P7', Aliases='RPL15_2_475', description='ribosomal protein L15 pseudogene 7', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47706322, end_position_on_the_genomic_accession=47707287, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728577, CurrentID=0, Status='live', Symbol='CNTNAP3B', Aliases='', description='contactin associated protein-like 3B', other_designations='cell recognition molecule Caspr3b', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41890921, end_position_on_the_genomic_accession=42129504, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  728586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728586, CurrentID=0, Status='live', Symbol='CCDC192', Aliases='LINC01183', description='coiled-coil domain containing 192', other_designations='long intergenic non-protein coding RNA 1183', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127702216, end_position_on_the_genomic_accession=128000995, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  100129367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129367, CurrentID=0, Status='live', Symbol='LOC100129367', Aliases='', description='nucleolar and coiled-body phosphoprotein 1 pseudogene', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=137696724, end_position_on_the_genomic_accession=137698464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728588, CurrentID=0, Status='live', Symbol='MS4A18', Aliases='', description='membrane-spanning 4-domains, subfamily A, member 18', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60724009, end_position_on_the_genomic_accession=60744213, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  728589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728589, CurrentID=0, Status='live', Symbol='NAP1L4P1', Aliases='', description='nucleosome assembly protein 1-like 4 pseudogene 1', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116531756, end_position_on_the_genomic_accession=116534111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728590, CurrentID=0, Status='live', Symbol='RPS27AP11', Aliases='RPS27A_4_740', description='ribosomal protein S27a pseudogene 11', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113581463, end_position_on_the_genomic_accession=113581999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728591, CurrentID=0, Status='live', Symbol='CCDC169', Aliases='C13orf38', description='coiled-coil domain containing 169', other_designations='RP11-251J8.1', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36227042, end_position_on_the_genomic_accession=36297855, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105372248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372248, CurrentID=0, Status='live', Symbol='LOC105372248', Aliases='', description='uncharacterized LOC105372248', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3761754, end_position_on_the_genomic_accession=3762640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728597, CurrentID=0, Status='live', Symbol='DCDC2C', Aliases='', description='doublecortin domain containing 2C', other_designations='doublecortin domain containing 2C pseudogene', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3703592, end_position_on_the_genomic_accession=3847411, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  728599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728599, CurrentID=0, Status='live', Symbol='CIAPIN1P', Aliases='HsT2384', description='cytokine induced apoptosis inhibitor 1 pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26599319, end_position_on_the_genomic_accession=26601327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728602, CurrentID=0, Status='live', Symbol='LOC728602', Aliases='', description='ornithine decarboxylase antizyme 1 pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40132709, end_position_on_the_genomic_accession=40133576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372250, CurrentID=0, Status='live', Symbol='LOC105372250', Aliases='', description='uncharacterized LOC105372250', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4722942, end_position_on_the_genomic_accession=4725174, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  728606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728606, CurrentID=0, Status='live', Symbol='PCAT18', Aliases='LINC01092', description='prostate cancer associated transcript 18 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1092', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26687621, end_position_on_the_genomic_accession=26703638, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728609, CurrentID=0, Status='live', Symbol='SDHAP3', Aliases='SDHACL, SDHAL', description='succinate dehydrogenase complex subunit A, flavoprotein pseudogene 3', other_designations='SDHA C-terminal like|succinate dehydrogenase complex, subunit A, flavoprotein pseudogene 3', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1571957, end_position_on_the_genomic_accession=1594531, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  728611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728611, CurrentID=0, Status='live', Symbol='BMS1P10', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 10', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 10|BMS1 ribosome biogenesis factor pseudogene 10', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63372803, end_position_on_the_genomic_accession=63380012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728613, CurrentID=0, Status='live', Symbol='LOC728613', Aliases='', description='programmed cell death 6 pseudogene', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1597557, end_position_on_the_genomic_accession=1634005, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728614, CurrentID=0, Status='live', Symbol='DNAJA1P4', Aliases='', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 4', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 4', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114349486, end_position_on_the_genomic_accession=114350657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728615, CurrentID=0, Status='live', Symbol='CICP13', Aliases='', description='capicua transcriptional repressor pseudogene 13', other_designations='capicua homolog pseudogene 13', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222468098, end_position_on_the_genomic_accession=222469897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421322, CurrentID=0, Status='live', Symbol='LOC100421322', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728619, CurrentID=0, Status='live', Symbol='ASB9P1', Aliases='', description='ankyrin repeat and SOCS box containing 9 pseudogene 1', other_designations='ankyrin repeat and SOCS box-containing 9 pseudogene 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92795484, end_position_on_the_genomic_accession=92797108, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728620, CurrentID=0, Status='live', Symbol='RPL7P22', Aliases='RPL7_6_574', description='ribosomal protein L7 pseudogene 22', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72725417, end_position_on_the_genomic_accession=72726154, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728621, CurrentID=0, Status='live', Symbol='CCDC30', Aliases='PFD6L, PFDN6L', description='coiled-coil domain containing 30', other_designations='prefoldin subunit 6-like protein', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42463318, end_position_on_the_genomic_accession=42654664, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  728622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728622, CurrentID=0, Status='live', Symbol='SKP1P2', Aliases='SKP1B, p19B', description='S-phase kinase-associated protein 1 pseudogene 2', other_designations='S-phase kinase-associated protein 1B (p19B)', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16988747, end_position_on_the_genomic_accession=16990628, orientation='plus', exon_count=1, OMIM=601435, no_hdr0=''),
  100129373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129373, CurrentID=0, Status='live', Symbol='LOC100129373', Aliases='', description='uncharacterized LOC100129373', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=149660273, end_position_on_the_genomic_accession=149660722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728629, CurrentID=0, Status='live', Symbol='TPI1P4', Aliases='', description='triosephosphate isomerase 1 pseudogene 4', other_designations='', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49016277, end_position_on_the_genomic_accession=49017190, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927082, CurrentID=0, Status='live', Symbol='LOC101927082', Aliases='', description='uncharacterized LOC101927082', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51621870, end_position_on_the_genomic_accession=51622577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728632, CurrentID=0, Status='live', Symbol='HMGN2P3', Aliases='HMGN2L3', description='high mobility group nucleosomal binding domain 2 pseudogene 3', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 3|high-mobility group nucleosomal binding domain 2-like 3', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26031717, end_position_on_the_genomic_accession=26032931, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728633, CurrentID=0, Status='live', Symbol='GTF2F2P2', Aliases='', description='general transcription factor IIF subunit 2 pseudogene 2', other_designations='general transcription factor IIF, polypeptide 2 pseudogene 2|general transcription factor IIF, polypeptide 2, 30kDa pseudogene', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40593633, end_position_on_the_genomic_accession=40594323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728635, CurrentID=0, Status='live', Symbol='DHRS4L1', Aliases='SDR25C4', description='dehydrogenase/reductase (SDR family) member 4 like 1', other_designations='Putative dehydrogenase/reductase SDR family member 4-like 2|short chain dehydrogenase/reductase family 25C member 4|short chain dehydrogenase/reductase family 25C, member 4', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24007009, end_position_on_the_genomic_accession=24051377, orientation='plus', exon_count=9, OMIM=615195, no_hdr0=''),
  728637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728637, CurrentID=0, Status='live', Symbol='MEIKIN', Aliases='', description='meiotic kinetochore factor', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131806990, end_position_on_the_genomic_accession=131945698, orientation='minus', exon_count=13, OMIM=616232, no_hdr0=''),
  728638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728638, CurrentID=0, Status='live', Symbol='KRT8P3', Aliases='', description='keratin 8 pseudogene 3', other_designations='hCG1988300', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61578144, end_position_on_the_genomic_accession=61579915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728640, CurrentID=0, Status='live', Symbol='FAM133CP', Aliases='', description='family with sequence similarity 133 member C, pseudogene', other_designations='family with sequence similarity 133, member C, pseudogene', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58715015, end_position_on_the_genomic_accession=58717534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728641, CurrentID=0, Status='live', Symbol='FABP5P7', Aliases='FABP5L7', description='fatty acid binding protein 5 pseudogene 7', other_designations='fatty acid binding protein 5-like 7', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59781103, end_position_on_the_genomic_accession=59781751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728642, CurrentID=0, Status='live', Symbol='CDK11A', Aliases='CDC2L2, CDC2L3, CDK11-p110, CDK11-p46, CDK11-p58, PITSLRE, p58GTA', description='cyclin-dependent kinase 11A', other_designations='PITSLRE B|PITSLRE protein kinase beta|PITSLRE serine/threonine-protein kinase CDC2L2|cell division cycle 2-like 2 (PITSLRE proteins)|cell division cycle 2-like protein kinase 2|cell division protein kinase 11A|galactosyltransferase-associated protein kinase p58/GTA', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1702730, end_position_on_the_genomic_accession=1724352, orientation='minus', exon_count=20, OMIM=116951, no_hdr0=''),
  728643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728643, CurrentID=0, Status='live', Symbol='HNRNPA1P33', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 33', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46415915, end_position_on_the_genomic_accession=46416456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129377, CurrentID=0, Status='live', Symbol='RPL23P9', Aliases='RPL23_3_937', description='ribosomal protein L23 pseudogene 9', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104981137, end_position_on_the_genomic_accession=104984885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728655, CurrentID=0, Status='live', Symbol='HULC', Aliases='HCCAT1, LINC00078, NCRNA00078', description='hepatocellular carcinoma up-regulated long non-coding RNA', other_designations='hepatocellular carcinoma associated transcript 1 (non-protein coding)|highly up-regulated in liver cancer (non-protein coding)|highly up-regulated in liver cancer long non-coding RNA|long intergenic non-protein coding RNA 78', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8652209, end_position_on_the_genomic_accession=8653846, orientation='plus', exon_count=2, OMIM=612210, no_hdr0=''),
  728656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728656, CurrentID=0, Status='live', Symbol='DMRTC1B', Aliases='', description='DMRT-like family C1B', other_designations='DMRT-like family C1', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72777170, end_position_on_the_genomic_accession=72848802, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  728657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728657, CurrentID=0, Status='live', Symbol='DUSP8P4', Aliases='DUSP8P2', description='dual specificity phosphatase 8 pseudogene 4', other_designations='dual specificity phosphatase 8 pseudogene 2', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46772700, end_position_on_the_genomic_accession=46776308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728658, CurrentID=0, Status='live', Symbol='RPL13AP5', Aliases='RPL13A_8_1086, bA196N24.2', description='ribosomal protein L13a pseudogene 5', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96750266, end_position_on_the_genomic_accession=96750923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372259, CurrentID=0, Status='live', Symbol='LOC105372259', Aliases='', description='uncharacterized LOC105372259', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7075752, end_position_on_the_genomic_accession=7092337, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728660, CurrentID=0, Status='live', Symbol='LOC728660', Aliases='', description='uncharacterized LOC728660', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140017243, end_position_on_the_genomic_accession=140017694, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728661, CurrentID=0, Status='live', Symbol='SLC35E2B', Aliases='SLC35E2', description='solute carrier family 35 member E2B', other_designations='solute carrier family 35, member E2B', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1660025, end_position_on_the_genomic_accession=1692804, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  728664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728664, CurrentID=0, Status='live', Symbol='API5P2', Aliases='', description='apoptosis inhibitor 5 pseudogene 2', other_designations='API5 pseudogene 2', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177994817, end_position_on_the_genomic_accession=177998877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129380, CurrentID=0, Status='live', Symbol='LOC100129380', Aliases='', description='uncharacterized LOC100129380', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=124189946, end_position_on_the_genomic_accession=124192209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728666, CurrentID=0, Status='live', Symbol='PRELID1P1', Aliases='', description='PRELI domain containing 1 pseudogene 1', other_designations='', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126643460, end_position_on_the_genomic_accession=126644398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373289, CurrentID=0, Status='live', Symbol='LOC105373289', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227786481, end_position_on_the_genomic_accession=227790651, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129381, CurrentID=0, Status='live', Symbol='LOC100129381', Aliases='', description='ribosomal protein S7 pseudogene', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1735819, end_position_on_the_genomic_accession=1737291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728673, CurrentID=0, Status='live', Symbol='LOC728673', Aliases='', description='uncharacterized LOC728673', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62897376, end_position_on_the_genomic_accession=62899777, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728675, CurrentID=0, Status='live', Symbol='SMCO4P1', Aliases='', description='single-pass membrane protein with coiled-coil domains 4 pseudogene 1', other_designations='UPF0443 protein C11orf75-like', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44645266, end_position_on_the_genomic_accession=44646418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372262, CurrentID=0, Status='live', Symbol='LOC105372262', Aliases='', description='uncharacterized LOC105372262', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7703080, end_position_on_the_genomic_accession=7707221, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728684, CurrentID=0, Status='live', Symbol='LOC728684', Aliases='', description='uncharacterized LOC728684', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=47124566, end_position_on_the_genomic_accession=47128180, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728688, CurrentID=0, Status='live', Symbol='LOC728688', Aliases='', description='ubiquitin-like, containing PHD and RING finger domains, 1 pseudogene', other_designations='ubiquitin-like with PHD and ring finger domains 1 pseudogene', map_location='12p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20551569, end_position_on_the_genomic_accession=20554360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728689, CurrentID=0, Status='live', Symbol='EIF3CL', Aliases='', description='eukaryotic translation initiation factor 3 subunit C-like', other_designations='eIF3 p110|eukaryotic translation initiation factor 3 subunit 8|eukaryotic translation initiation factor 3, subunit 8, 110kDa-like|eukaryotic translation initiation factor 3, subunit C-like', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28379579, end_position_on_the_genomic_accession=28403841, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  728693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728693, CurrentID=0, Status='live', Symbol='RPL21P105', Aliases='RPL21_43_1282', description='ribosomal protein L21 pseudogene 105', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116755353, end_position_on_the_genomic_accession=116755917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728695, CurrentID=0, Status='live', Symbol='SPANXB1', Aliases='B1, CT11.2, SPANX-B, SPANXB, SPANXB2, SPANXF1, SPANXF2', description='SPANX family member B1', other_designations='SPANX family member B/F|SPANX family member F1|SPANX family, member B1|SPANX family, member B2|SPANX family, member F1|SPANX family, member F2|cancer/testis antigen 11.2|cancer/testis antigen family 11, member 2|nuclear-associated protein SPAN-Xb|nuclear-associated protein SPAN-Xf|sperm protein associated with the nucleus on the X chromosome B/F|sperm protein associated with the nucleus, X chromosome, family member B1', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141002591, end_position_on_the_genomic_accession=141003706, orientation='plus', exon_count=2, OMIM=300669, no_hdr0=''),
  728698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728698, CurrentID=0, Status='live', Symbol='EIF4A1P4', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 4', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53153258, end_position_on_the_genomic_accession=53154630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728700, CurrentID=0, Status='live', Symbol='RPS3AP19', Aliases='RPS3A_8_458', description='ribosomal protein S3a pseudogene 19', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9563147, end_position_on_the_genomic_accession=9563994, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728705, CurrentID=0, Status='live', Symbol='LOC728705', Aliases='', description='uncharacterized LOC728705', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=105871937, end_position_on_the_genomic_accession=105882738, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728706, CurrentID=0, Status='live', Symbol='SUGT1P2', Aliases='', description='SGT1 homolog, MIS12 kinetochore complex assembly cochaperone pseudogene 2', other_designations='SGT1, suppressor of G2 allele of SKP1 pseudogene|SUGT1 pseudogene 2|suppressor of G2 allele of SKP1 pseudogene 2', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32752830, end_position_on_the_genomic_accession=32753920, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  55915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55915, CurrentID=0, Status='live', Symbol='LANCL2', Aliases='GPR69B, TASP', description='LanC lantibiotic synthetase component C-like 2 (bacterial)', other_designations='G protein-coupled receptor 69B|LanC (bacterial lantibiotic synthetase component C)-like 2|testis-specific adriamycin sensitivity protein', map_location='7q31.1-q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55365448, end_position_on_the_genomic_accession=55433742, orientation='plus', exon_count=10, OMIM=612919, no_hdr0=''),
  728710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728710, CurrentID=0, Status='live', Symbol='FAM177A1P1', Aliases='FAM177A2', description='family with sequence similarity 177 member A1 pseudogene 1', other_designations='family with sequence similarity 177, member A1 pseudogene 1|family with sequence similarity 177, member A2', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98955999, end_position_on_the_genomic_accession=98964956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728712, CurrentID=0, Status='live', Symbol='SPANXA2', Aliases='CT11.1, SPANX, SPANX-A, SPANXA', description='SPANX family member A2', other_designations='SPANX family member A|SPANX family, member A2|cancer/testis antigen 11.1|nuclear-associated protein SPAN-Xa|sperm protein associated with the nucleus, X chromosome, family member A2', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141589708, end_position_on_the_genomic_accession=141590772, orientation='plus', exon_count=2, OMIM=300493, no_hdr0=''),
  105372268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372268, CurrentID=0, Status='live', Symbol='LOC105372268', Aliases='', description='uncharacterized LOC105372268', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9507811, end_position_on_the_genomic_accession=9514673, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728715, CurrentID=0, Status='live', Symbol='LOC728715', Aliases='', description='ovostatin homolog 2', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9397062, end_position_on_the_genomic_accession=9403060, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  728716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728716, CurrentID=0, Status='live', Symbol='LINC01346', Aliases='', description='long intergenic non-protein coding RNA 1346', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3940612, end_position_on_the_genomic_accession=3952583, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728723, CurrentID=0, Status='live', Symbol='ZBED3-AS1', Aliases='', description='ZBED3 antisense RNA 1', other_designations='CTC-564N23.3|ZBED3 antisense RNA 1 (non-protein coding)', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77086798, end_position_on_the_genomic_accession=77148351, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728724, CurrentID=0, Status='live', Symbol='LINC00977', Aliases='', description='long intergenic non-protein coding RNA 977', other_designations='hCG1814486', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129216467, end_position_on_the_genomic_accession=129241240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728729, CurrentID=0, Status='live', Symbol='FABP5P8', Aliases='FABP5L8', description='fatty acid binding protein 5 pseudogene 8', other_designations='fatty acid binding protein 5-like 8', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83044307, end_position_on_the_genomic_accession=83044977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728730, CurrentID=0, Status='live', Symbol='LOC728730', Aliases='', description='uncharacterized LOC728730', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39437416, end_position_on_the_genomic_accession=39601344, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129391, CurrentID=0, Status='live', Symbol='LOC100129391', Aliases='', description='zinc finger (CCCH type), RNA binding motif and serine/arginine rich 2 pseudogene', other_designations='zinc finger (CCCH type), RNA-binding motif and serine/arginine rich 2 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7740655, end_position_on_the_genomic_accession=7742160, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728732, CurrentID=0, Status='live', Symbol='HNRNPA1P53', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 53', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5570915, end_position_on_the_genomic_accession=5572253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728734, CurrentID=0, Status='live', Symbol='NPIPB8', Aliases='', description='nuclear pore complex interacting protein family member B8', other_designations='Nuclear pore complex-interacting protein family member B8|nuclear pore complex interacting protein family, member B8', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28624517, end_position_on_the_genomic_accession=28658725, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  728737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728737, CurrentID=0, Status='live', Symbol='TATDN1P1', Aliases='', description='TatD DNase domain containing 1 pseudogene 1', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63222124, end_position_on_the_genomic_accession=63223134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728739, CurrentID=0, Status='live', Symbol='LOC728739', Aliases='', description='programmed cell death 2 pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107903150, end_position_on_the_genomic_accession=107903771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728741, CurrentID=0, Status='live', Symbol='NPIPB6', Aliases='NPIPB', description='nuclear pore complex interacting protein family member B6', other_designations='nuclear pore complex interacting protein family, member B6|nuclear pore complex-interacting protein B type', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28342512, end_position_on_the_genomic_accession=28364929, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  645745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645745, CurrentID=0, Status='live', Symbol='MT1HL1', Aliases='MT1P2', description='metallothionein 1H-like 1', other_designations='MT-1H-like protein', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237004103, end_position_on_the_genomic_accession=237004418, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728746, CurrentID=0, Status='live', Symbol='FAM90A17P', Aliases='FAM90A17', description='family with sequence similarity 90 member A17, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7744372, end_position_on_the_genomic_accession=7749606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728747, CurrentID=0, Status='live', Symbol='ANKRD20A4', Aliases='', description='ankyrin repeat domain 20 family member A4', other_designations='ankyrin repeat domain 20 family, member A4', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64369563, end_position_on_the_genomic_accession=64412691, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  728748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728748, CurrentID=0, Status='live', Symbol='PRKRIRP1', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 1', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74396572, end_position_on_the_genomic_accession=74399637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372274, CurrentID=0, Status='live', Symbol='HSPC102', Aliases='', description='uncharacterized LOC105372274', other_designations='CTC-510F12.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11300777, end_position_on_the_genomic_accession=11321262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927086, CurrentID=0, Status='live', Symbol='LOC101927086', Aliases='', description='uncharacterized LOC101927086', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70418592, end_position_on_the_genomic_accession=70608272, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102725219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725219, CurrentID=0, Status='live', Symbol='LOC102725219', Aliases='', description='PRAME family member 17-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728753, CurrentID=0, Status='live', Symbol='FAM90A19P', Aliases='FAM90A19', description='family with sequence similarity 90 member A19, pseudogene', other_designations='family with sequence similarity 90, member A19, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7752020, end_position_on_the_genomic_accession=7757254, orientation='plus', exon_count=0, OMIM=613053, no_hdr0=''),
  728755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728755, CurrentID=0, Status='live', Symbol='LOC728755', Aliases='', description='uncharacterized LOC728755', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27317383, end_position_on_the_genomic_accession=27673221, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728758, CurrentID=0, Status='live', Symbol='PIN4P1', Aliases='', description='protein (peptidylprolyl cis/trans isomerase) NIMA-interacting, 4 pseudogene 1', other_designations='hCG1789710|protein (peptidylprolyl cis/trans isomerase) NIMA-interacting, 4 (parvulin) pseudogene', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43875806, end_position_on_the_genomic_accession=43878171, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100533941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533941, CurrentID=0, Status='live', Symbol='LOC100533941', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372276, CurrentID=0, Status='live', Symbol='LOC105372276', Aliases='', description='uncharacterized LOC105372276', other_designations='CTD-3105H18.7', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12450440, end_position_on_the_genomic_accession=12460839, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  204474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204474, CurrentID=0, Status='live', Symbol='PDILT', Aliases='PDIA7', description='protein disulfide isomerase-like, testis expressed', other_designations='protein disulfide isomerase family A, member 7|protein disulfide isomerase-like protein of the testis', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20359170, end_position_on_the_genomic_accession=20404742, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  728763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728763, CurrentID=0, Status='live', Symbol='CROCC2', Aliases='', description='ciliary rootlet coiled-coil, rootletin family member 2', other_designations='Putative ciliary rootlet coiled-coil protein-like 3 protein|rootletin-like', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240906359, end_position_on_the_genomic_accession=240993307, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  105372277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372277, CurrentID=0, Status='live', Symbol='LOC105372277', Aliases='', description='zinc finger protein 799-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12379217, end_position_on_the_genomic_accession=12386256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  728769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728769, CurrentID=0, Status='live', Symbol='SCAMP1-AS1', Aliases='', description='SCAMP1 antisense RNA 1', other_designations='CTD-2037K23.2', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78358764, end_position_on_the_genomic_accession=78360541, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728770, CurrentID=0, Status='live', Symbol='DNM1P5', Aliases='DNM1DN5, DNM1DN5@', description='dynamin 1 pseudogene 5', other_designations='DNM1 duplicon 5|DNM1 pseudogene 5|DNM1DN5 duplicon', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34520607, end_position_on_the_genomic_accession=34523423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728773, CurrentID=0, Status='live', Symbol='PABPC1P2', Aliases='PABP2, PABP4, PABPCP2, PABPCP4', description='poly(A) binding protein, cytoplasmic 1 pseudogene 2', other_designations='poly(A) binding protein, cytoplasmic, pseudogene 2|poly(A) binding protein, cytoplasmic, pseudogene 4', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146587057, end_position_on_the_genomic_accession=146590990, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728774, CurrentID=0, Status='live', Symbol='RPS15AP25', Aliases='RPS15A_14_912', description='ribosomal protein S15a pseudogene 25', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68525483, end_position_on_the_genomic_accession=68525939, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728776, CurrentID=0, Status='live', Symbol='HMGN1P37', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 37', other_designations='hCG17955|high-mobility group nucleosome binding domain 1 pseudogene 37', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154816299, end_position_on_the_genomic_accession=154817521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129399, CurrentID=0, Status='live', Symbol='LOC100129399', Aliases='', description='uncharacterized LOC100129399', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=112095889, end_position_on_the_genomic_accession=112096251, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728780, CurrentID=0, Status='live', Symbol='ANKDD1B', Aliases='', description='ankyrin repeat and death domain containing 1B', other_designations='ankyrin repeat and death domain-containing protein ENSP00000345065', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75611218, end_position_on_the_genomic_accession=75671846, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  728782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728782, CurrentID=0, Status='live', Symbol='RPL21P134', Aliases='RPL21_61_1788', description='ribosomal protein L21 pseudogene 134', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75384306, end_position_on_the_genomic_accession=75384863, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371544, CurrentID=0, Status='live', Symbol='LOC105371544', Aliases='', description='uncharacterized LOC105371544', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13619735, end_position_on_the_genomic_accession=13627647, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645752, CurrentID=0, Status='live', Symbol='LOC645752', Aliases='', description='golgin A6 family, member A pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 6 pseudogene', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77914217, end_position_on_the_genomic_accession=77926846, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  728788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728788, CurrentID=0, Status='live', Symbol='ANKRD20A20P', Aliases='CCDC29', description='ankyrin repeat domain 20 family member A20, pseudogene', other_designations='ankyrin repeat domain 20 family, member A20, pseudogene|ankyrin repeat domain pseudogene|coiled-coil domain containing 29', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64410782, end_position_on_the_genomic_accession=64436159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728791, CurrentID=0, Status='live', Symbol='RPS10P4', Aliases='RPS10_2_393, bcm1851', description='ribosomal protein S10 pseudogene 4', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113850188, end_position_on_the_genomic_accession=113850770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645754, CurrentID=0, Status='live', Symbol='LGALS9DP', Aliases='', description='lectin, galactoside-binding, soluble, 9D, pseudogene', other_designations='lectin, galactoside-binding, soluble, 9 pseudogene', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27737777, end_position_on_the_genomic_accession=27756703, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728798, CurrentID=0, Status='live', Symbol='FRMPD2B', Aliases='FRMPD2L1, FRMPD2L2, FRMPD2P1, FRMPD2P2, FRMPDP2, PDZD5A, PDZD5B, PDZK5A, PDZK5B, bA556L1.2, yX59F3.2', description='FERM and PDZ domain containing 2B, pseudogene', other_designations='FERM and PDZ domain containing 2 like 1|FERM and PDZ domain containing 2 like 2|FERM and PDZ domain containing 2 pseudogene 1|FERM and PDZ domain containing 2 pseudogene 2|FRMPD2 related 1|FRMPD2 related 2|PDZ domain containing 5A|PDZ domain containing 5B', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46876674, end_position_on_the_genomic_accession=46895326, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372283, CurrentID=0, Status='live', Symbol='LOC105372283', Aliases='', description='uncharacterized LOC105372283', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13409196, end_position_on_the_genomic_accession=13438775, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  728805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728805, CurrentID=0, Status='live', Symbol='LOC728805', Aliases='', description='uncharacterized LOC728805', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53168643, end_position_on_the_genomic_accession=53180017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728806, CurrentID=0, Status='live', Symbol='NSFP1', Aliases='NSFP', description='N-ethylmaleimide-sensitive factor pseudogene 1', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46372813, end_position_on_the_genomic_accession=46423097, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  728807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728807, CurrentID=0, Status='live', Symbol='LOC728807', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B15 pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68626844, end_position_on_the_genomic_accession=68628908, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129404, CurrentID=0, Status='live', Symbol='LOC100129404', Aliases='', description='matrix metallopeptidase 12 (macrophage elastase) pseudogene', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26254916, end_position_on_the_genomic_accession=26256631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927088, CurrentID=0, Status='live', Symbol='LOC101927088', Aliases='', description='uncharacterized LOC101927088', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374291, CurrentID=0, Status='live', Symbol='LOC105374291', Aliases='', description='uncharacterized LOC105374291', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194600735, end_position_on_the_genomic_accession=194603448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288803, CurrentID=0, Status='live', Symbol='LOC100288803', Aliases='', description='F-box and WD repeat domain containing 2 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178997965, end_position_on_the_genomic_accession=178998813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728819, CurrentID=0, Status='live', Symbol='C1GALT1C1L', Aliases='', description='C1GALT1-specific chaperone 1 like', other_designations='C1GALT1-specific chaperone 1-like|hCG1645220', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43675153, end_position_on_the_genomic_accession=43676322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728820, CurrentID=0, Status='live', Symbol='RPL29P15', Aliases='RPL29_4_582', description='ribosomal protein L29 pseudogene 15', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79510382, end_position_on_the_genomic_accession=79511017, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372286, CurrentID=0, Status='live', Symbol='LOC105372286', Aliases='', description='uncharacterized LOC105372286', other_designations='CTB-55O6.10', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14114286, end_position_on_the_genomic_accession=14119432, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728823, CurrentID=0, Status='live', Symbol='RPS26P53', Aliases='RPS26_19_1511', description='ribosomal protein S26 pseudogene 53', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8561203, end_position_on_the_genomic_accession=8561659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728825, CurrentID=0, Status='live', Symbol='LOC728825', Aliases='', description='small ubiquitin-like modifier 2', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115718340, end_position_on_the_genomic_accession=115719398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372287, CurrentID=0, Status='live', Symbol='LOC105372287', Aliases='', description='uncharacterized LOC105372287', other_designations='CTD-2189E23.1|CTD-2189E23.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14267816, end_position_on_the_genomic_accession=14292913, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728833, CurrentID=0, Status='live', Symbol='FAM72D', Aliases='GCUD2', description='family with sequence similarity 72 member D', other_designations='family with sequence similarity 72 member|family with sequence similarity 72, member D|gastric cancer up-regulated protein 2|gastric cancer up-regulated-2', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145093798, end_position_on_the_genomic_accession=145112692, orientation='plus', exon_count=7, OMIM=614712, no_hdr0=''),
  728841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728841, CurrentID=0, Status='live', Symbol='NBPF8', Aliases='NBPF8P', description='neuroblastoma breakpoint family member 8', other_designations='neuroblastoma breakpoint family, member 8|neuroblastoma breakpoint family, member 8, pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120416595, end_position_on_the_genomic_accession=120467844, orientation='plus', exon_count=28, OMIM=613998, no_hdr0=''),
  728843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728843, CurrentID=0, Status='live', Symbol='RPL7P20', Aliases='RPL7_8_636', description='ribosomal protein L7 pseudogene 20', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166028484, end_position_on_the_genomic_accession=166029331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728851, CurrentID=0, Status='live', Symbol='HMGN1P36', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 36', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 36', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97827152, end_position_on_the_genomic_accession=97828350, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728853, CurrentID=0, Status='live', Symbol='LOC728853', Aliases='', description='WD repeat domain 87 pseudogene', other_designations='', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37829067, end_position_on_the_genomic_accession=37831916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728855, CurrentID=0, Status='live', Symbol='LINC00623', Aliases='CH17-118O6.2', description='long intergenic non-protein coding RNA 623', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120912238, end_position_on_the_genomic_accession=120955811, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  728858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728858, CurrentID=0, Status='live', Symbol='C12orf71', Aliases='', description='chromosome 12 open reading frame 71', other_designations='hCG1794298', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27081057, end_position_on_the_genomic_accession=27084144, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728868, CurrentID=0, Status='live', Symbol='LOC728868', Aliases='', description='uncharacterized LOC728868', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=10318828, end_position_on_the_genomic_accession=10321453, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728873, CurrentID=0, Status='live', Symbol='PTMAP8', Aliases='', description='prothymosin, alpha pseudogene 8', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117026698, end_position_on_the_genomic_accession=117027039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728877, CurrentID=0, Status='live', Symbol='LOC728877', Aliases='', description='COBW domain containing 3 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63974762, end_position_on_the_genomic_accession=64004181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728882, CurrentID=0, Status='live', Symbol='FAM182B', Aliases='', description='family with sequence similarity 182 member B', other_designations='family with sequence similarity 182, member B', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25763466, end_position_on_the_genomic_accession=25801291, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  728887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728887, CurrentID=0, Status='live', Symbol='CCNB1IP1P1', Aliases='', description='cyclin B1 interacting protein 1 pseudogene 1', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44958088, end_position_on_the_genomic_accession=44959576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728888, CurrentID=0, Status='live', Symbol='NPIPB11', Aliases='NPIP', description='nuclear pore complex interacting protein family member B11', other_designations='NPIP-like locus|Nuclear pore complex-interacting protein family member B11|hepatitis B virus pre-pre-S promoter DNA-binding protein|nuclear pore complex interacting protein family, member B11|pps22-1 protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29381183, end_position_on_the_genomic_accession=29405678, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105372298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372298, CurrentID=0, Status='live', Symbol='LOC105372298', Aliases='', description='uncharacterized LOC105372298', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17378790, end_position_on_the_genomic_accession=17379772, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372299, CurrentID=0, Status='live', Symbol='LOC105372299', Aliases='', description='uncharacterized LOC105372299', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17733338, end_position_on_the_genomic_accession=17750695, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  728903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728903, CurrentID=0, Status='live', Symbol='LOC728903', Aliases='', description='uncharacterized LOC728903', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=46683325, end_position_on_the_genomic_accession=46690200, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100533942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533942, CurrentID=0, Status='live', Symbol='LOC100533942', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728911, CurrentID=0, Status='live', Symbol='CT45A2', Aliases='CT45-2, CT45.2', description='cancer/testis antigen family 45, member A2', other_designations='cancer/testis antigen 45-2|cancer/testis antigen 45A2|cancer/testis antigen CT45', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135811979, end_position_on_the_genomic_accession=135829348, orientation='minus', exon_count=7, OMIM=300793, no_hdr0=''),
  728913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728913, CurrentID=0, Status='live', Symbol='RCN1P2', Aliases='', description='reticulocalbin 1 pseudogene 2', other_designations='reticulocalbin 1 like|reticulocalbin 1, EF-hand calcium binding domain pseudogene 2', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45389294, end_position_on_the_genomic_accession=45391426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728917, CurrentID=0, Status='live', Symbol='ABHD17AP1', Aliases='ABHD17AP2, C1orf47, FAM108A11P, FAM108A2, FAM108A3, FAM108A3P', description='abhydrolase domain containing 17A pseudogene 1', other_designations='abhydrolase domain containing 17A pseudogene 2|abhydrolase domain-containing protein FAM108A2|family with sequence similarity 108, member A11, pseudogene|family with sequence similarity 108, member A2|family with sequence similarity 108, member A3, pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148146066, end_position_on_the_genomic_accession=148149788, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724917, CurrentID=0, Status='live', Symbol='LOC102724917', Aliases='', description='uncharacterized LOC102724917', other_designations='CTD-2384A14.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28830248, end_position_on_the_genomic_accession=28968386, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  728920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728920, CurrentID=0, Status='live', Symbol='PDE4DIPP1', Aliases='PDE4DIP1', description='phosphodiesterase 4D interacting protein pseudogene 1', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148080595, end_position_on_the_genomic_accession=148093980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728921, CurrentID=0, Status='live', Symbol='HAUS1P1', Aliases='', description='HAUS augmin like complex subunit 1 pseudogene 1', other_designations='HAUS augmin-like complex, subunit 1 pseudogene 1', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140581493, end_position_on_the_genomic_accession=140582541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372303, CurrentID=0, Status='live', Symbol='LOC105372303', Aliases='', description='uncharacterized LOC105372303', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379401, CurrentID=0, Status='live', Symbol='LOC105379401', Aliases='', description='uncharacterized LOC105379401', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  728927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728927, CurrentID=0, Status='live', Symbol='ZNF736', Aliases='', description='zinc finger protein 736', other_designations='KRAB-containing zinc-finger repressor protein', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64313873, end_position_on_the_genomic_accession=64356634, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  728929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728929, CurrentID=0, Status='live', Symbol='TCEB3CL', Aliases='HsT828, eloA3-like-1', description='transcription elongation factor B polypeptide 3C-like', other_designations='elongin-A3-like-1|transcription elongation factor B polypeptide 3C-like-1', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46968695, end_position_on_the_genomic_accession=46969912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728931, CurrentID=0, Status='live', Symbol='AMD1P3', Aliases='', description='adenosylmethionine decarboxylase 1 pseudogene 3', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43586363, end_position_on_the_genomic_accession=43588822, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728933, CurrentID=0, Status='live', Symbol='RPL23AP60', Aliases='RPL23A_26_1111', description='ribosomal protein L23a pseudogene 60', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133778241, end_position_on_the_genomic_accession=133779250, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728937, CurrentID=0, Status='live', Symbol='RPS26P25', Aliases='RPS26_10_512', description='ribosomal protein S26 pseudogene 25', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113213969, end_position_on_the_genomic_accession=113214420, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728938, CurrentID=0, Status='live', Symbol='SUB1P1', Aliases='hCG17750', description='SUB1 homolog, transcriptional regulator pseudogene 1', other_designations='SUB1 homolog pseudogene 1|SUB1 pseudogene 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11789952, end_position_on_the_genomic_accession=11791085, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  2450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=2450, CurrentID=0, Status='live', Symbol='FRA12C', Aliases='', description='fragile site, BrdU type, rare, fra(12)(q24.2)', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100500797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500797, CurrentID=0, Status='live', Symbol='USP9YP19', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 19', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25748860, end_position_on_the_genomic_accession=25749988, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728945, CurrentID=0, Status='live', Symbol='PPIAL4F', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A)-like 4F', other_designations='PPIase A-like 4F|cyclophilin-LC (COAS2) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144592868, end_position_on_the_genomic_accession=144593652, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  728947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728947, CurrentID=0, Status='live', Symbol='ZNF680P1', Aliases='', description='ZNF680 pseudogene 1', other_designations='zinc finger protein 680 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64469174, end_position_on_the_genomic_accession=64470179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100500798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500798, CurrentID=0, Status='live', Symbol='USP9YP23', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 23', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17782231, end_position_on_the_genomic_accession=17784816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728953, CurrentID=0, Status='live', Symbol='RPS19P3', Aliases='RPS19_4_1350', description='ribosomal protein S19 pseudogene 3', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34568652, end_position_on_the_genomic_accession=34569169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728955, CurrentID=0, Status='live', Symbol='FAM25D', Aliases='bA592B15.5', description='family with sequence similarity 25 member D', other_designations='family with sequence similarity 25, member D', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=50884513, end_position_on_the_genomic_accession=50886406, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728957, CurrentID=0, Status='live', Symbol='ZNF705D', Aliases='ZNF705C', description='zinc finger protein 705D', other_designations='putative zinc finger protein 705C|zinc finger protein 705C', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12089338, end_position_on_the_genomic_accession=12115516, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  728962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728962, CurrentID=0, Status='live', Symbol='RPL7P56', Aliases='RPL7_33_1817', description='ribosomal protein L7 pseudogene 56', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136791037, end_position_on_the_genomic_accession=136791848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728963, CurrentID=0, Status='live', Symbol='RPS15AP10', Aliases='RPS15A_4_42', description='ribosomal protein S15a pseudogene 10', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45645772, end_position_on_the_genomic_accession=45646685, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  728970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728970, CurrentID=0, Status='live', Symbol='PPP1R2P4', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 4', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 4', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46436727, end_position_on_the_genomic_accession=46438185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728971, CurrentID=0, Status='live', Symbol='RPS27AP7', Aliases='RPS27A_2_217', description='ribosomal protein S27a pseudogene 7', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47882967, end_position_on_the_genomic_accession=47883914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  728975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728975, CurrentID=0, Status='live', Symbol='LOC728975', Aliases='', description='uncharacterized LOC728975', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65179871, end_position_on_the_genomic_accession=65181337, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370700, CurrentID=0, Status='live', Symbol='LOC105370700', Aliases='', description='uncharacterized LOC105370700', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106202390, end_position_on_the_genomic_accession=106207786, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372312, CurrentID=0, Status='live', Symbol='LOC105372312', Aliases='', description='uncharacterized LOC105372312', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20126125, end_position_on_the_genomic_accession=20127032, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  728978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728978, CurrentID=0, Status='live', Symbol='UNC5B-AS1', Aliases='', description='UNC5B antisense RNA 1', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71217224, end_position_on_the_genomic_accession=71218228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728979, CurrentID=0, Status='live', Symbol='RPL10AP9', Aliases='RPL10A_3_593', description='ribosomal protein L10a pseudogene 9', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86884210, end_position_on_the_genomic_accession=86884915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129434, CurrentID=0, Status='live', Symbol='LOC100129434', Aliases='', description='uncharacterized LOC100129434', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56173534, end_position_on_the_genomic_accession=56185770, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728990, CurrentID=0, Status='live', Symbol='SLC9A3P3', Aliases='', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 pseudogene 3', other_designations='solute carrier family 9 (sodium/hydrogen exchanger), member 3 pseudogene 3', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50025392, end_position_on_the_genomic_accession=50027167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728992, CurrentID=0, Status='live', Symbol='RPL7AP11', Aliases='RPL7A_5_397', description='ribosomal protein L7a pseudogene 11', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121494087, end_position_on_the_genomic_accession=121494963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928693, CurrentID=0, Status='live', Symbol='LOC101928693', Aliases='', description='glycoprotein Xg-like', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12436699, end_position_on_the_genomic_accession=12507408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728996, CurrentID=0, Status='live', Symbol='METTL8P1', Aliases='', description='methyltransferase like 8 pseudogene 1', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15254740, end_position_on_the_genomic_accession=15255823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  728999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728999, CurrentID=0, Status='live', Symbol='LOC728999', Aliases='', description='hCG1783052', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=15627052, end_position_on_the_genomic_accession=15631766, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729004, CurrentID=0, Status='live', Symbol='RPL9P2', Aliases='RPL9_10_1518', description='ribosomal protein L9 pseudogene 2', other_designations='60S ribosomal protein L9 pseudogene 2', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15443613, end_position_on_the_genomic_accession=15444290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729009, CurrentID=0, Status='live', Symbol='FTH1P20', Aliases='FTHL20', description='ferritin, heavy polypeptide 1 pseudogene 20', other_designations='ferritin, heavy polypeptide-like 20', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180872705, end_position_on_the_genomic_accession=180873565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729012, CurrentID=0, Status='live', Symbol='SKA2P1', Aliases='Em:AC006313.1, FAM33B, SKA2L', description='spindle and kinetochore associated complex subunit 2 pseudogene 1', other_designations='family with sequence similarity 33, member B|spindle and kinetochore associated complex subunit 2-like', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122761850, end_position_on_the_genomic_accession=122762656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729013, CurrentID=0, Status='live', Symbol='ZBED5-AS1', Aliases='', description='ZBED5 antisense RNA 1', other_designations='CTD-2003C8.1', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10858217, end_position_on_the_genomic_accession=10879276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102466889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466889, CurrentID=0, Status='live', Symbol='MIR8071-2', Aliases='hsa-mir-8071-2', description='microRNA 8071-2', other_designations='microRNA mir-8071-2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105640168, end_position_on_the_genomic_accession=105640232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729020, CurrentID=0, Status='live', Symbol='RPEL1', Aliases='', description='ribulose-5-phosphate-3-epimerase-like 1', other_designations='hCG2024410|rcRPE protein|ribulose-5-phosphate-3-epimerase-like protein 1', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103245887, end_position_on_the_genomic_accession=103248016, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729023, CurrentID=0, Status='live', Symbol='NUTM2HP', Aliases='FAM22HP, NUTMHP', description='NUT family member 2H, pseudogene', other_designations='family with sequence similarity 22, member D pseudogene|family with sequence similarity 22, member H, pseudogene', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50676455, end_position_on_the_genomic_accession=50685882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729025, CurrentID=0, Status='live', Symbol='SLC15A5', Aliases='', description='solute carrier family 15, member 5', other_designations='Peptide/histidine transporter ENSP00000340402', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16188485, end_position_on_the_genomic_accession=16277685, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  729032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729032, CurrentID=0, Status='live', Symbol='RPL36P20', Aliases='', description='ribosomal protein L36 pseudogene 20', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41955445, end_position_on_the_genomic_accession=41955760, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729033, CurrentID=0, Status='live', Symbol='LOC729033', Aliases='', description='uncharacterized LOC729033', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227972508, end_position_on_the_genomic_accession=227974357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372322, CurrentID=0, Status='live', Symbol='LOC105372322', Aliases='', description='uncharacterized LOC105372322', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21432485, end_position_on_the_genomic_accession=21443021, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729040, CurrentID=0, Status='live', Symbol='LOC729040', Aliases='', description='uncharacterized LOC729040', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91279993, end_position_on_the_genomic_accession=91283183, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729041, CurrentID=0, Status='live', Symbol='FAAHP1', Aliases='', description='fatty acid amide hydrolase pseudogene 1', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46433827, end_position_on_the_genomic_accession=46445702, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  729042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729042, CurrentID=0, Status='live', Symbol='RPL29P3', Aliases='RPL29_19_1356', description='ribosomal protein L29 pseudogene 3', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36577394, end_position_on_the_genomic_accession=36578035, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101243555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101243555, CurrentID=0, Status='live', Symbol='PTENP1-AS', Aliases='PTENP1as, PTENpg1-asRNA', description='PTENP1 antisense RNA', other_designations='PTENpg1 antisense RNA', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33677268, end_position_on_the_genomic_accession=33688011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729046, CurrentID=0, Status='live', Symbol='RPL17P50', Aliases='', description='ribosomal protein L17 pseudogene 50', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73005796, end_position_on_the_genomic_accession=73006627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729047, CurrentID=0, Status='live', Symbol='LOC729047', Aliases='', description='uncharacterized LOC729047', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50737567, end_position_on_the_genomic_accession=50740317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729050, CurrentID=0, Status='live', Symbol='KRT18P29', Aliases='', description='keratin 18 pseudogene 29', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181961149, end_position_on_the_genomic_accession=181962480, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927096, CurrentID=0, Status='live', Symbol='CTB-113P19.1', Aliases='', description='uncharacterized LOC101927096', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151676945, end_position_on_the_genomic_accession=151687910, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729056, CurrentID=0, Status='live', Symbol='LOC729056', Aliases='', description='KIAA1462 pseudogene', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39525583, end_position_on_the_genomic_accession=39529852, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729057, CurrentID=0, Status='live', Symbol='LOC729057', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog (yeast) pseudogene', other_designations='mitochondrial import receptor subunit TOM40 homolog', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15853681, end_position_on_the_genomic_accession=15855342, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729059, CurrentID=0, Status='live', Symbol='C1orf234', Aliases='', description='chromosome 1 open reading frame 234', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23010834, end_position_on_the_genomic_accession=23015850, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372326, CurrentID=0, Status='live', Symbol='LOC105372326', Aliases='', description='uncharacterized LOC105372326', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21902528, end_position_on_the_genomic_accession=21917829, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  204777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204777, CurrentID=0, Status='live', Symbol='IGLJCOR18', Aliases='IGLJ-C/OR18, IGLJ-COR18', description='immunoglobulin lambda joining-constant/OR18 (pseudogene)', other_designations='immunoglobulin lambda joining-constant/OR18 pseudogene', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3394886, end_position_on_the_genomic_accession=3395317, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372327, CurrentID=0, Status='live', Symbol='LOC105372327', Aliases='', description='uncharacterized LOC105372327', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22018223, end_position_on_the_genomic_accession=22046295, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100130664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130664, CurrentID=0, Status='live', Symbol='LOC100130664', Aliases='', description='cytidine monophosphate (UMP-CMP) kinase 1, cytosolic pseudogene', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89458791, end_position_on_the_genomic_accession=89459437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729073, CurrentID=0, Status='live', Symbol='FAM92A1P1', Aliases='FAM92A2', description='family with sequence similarity 92 member A1 pseudogene 1', other_designations='family with sequence similarity 92, member A1 pseudogene 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41163124, end_position_on_the_genomic_accession=41164697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729075, CurrentID=0, Status='live', Symbol='RPL36AP23', Aliases='RPL36A_6_507', description='ribosomal protein L36a pseudogene 23', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111551837, end_position_on_the_genomic_accession=111552228, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729078, CurrentID=0, Status='live', Symbol='LOC729078', Aliases='', description='uncharacterized LOC729078', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=139298185, end_position_on_the_genomic_accession=139298956, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729079, CurrentID=0, Status='live', Symbol='RPSAP21', Aliases='RPSA_2_180', description='ribosomal protein SA pseudogene 21', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236819556, end_position_on_the_genomic_accession=236820590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729080, CurrentID=0, Status='live', Symbol='LOC729080', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141895626, end_position_on_the_genomic_accession=141896695, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927097, CurrentID=0, Status='live', Symbol='LOC101927097', Aliases='', description='uncharacterized LOC101927097', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198222627, end_position_on_the_genomic_accession=198233940, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729082, CurrentID=0, Status='live', Symbol='OIP5-AS1', Aliases='cyrano', description='OIP5 antisense RNA 1', other_designations='OIP5 antisense RNA 1 (non-protein coding)', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41284003, end_position_on_the_genomic_accession=41299597, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729083, CurrentID=0, Status='live', Symbol='LOC729083', Aliases='', description='uncharacterized LOC729083', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42770612, end_position_on_the_genomic_accession=42773635, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729085, CurrentID=0, Status='live', Symbol='FAM198A', Aliases='C3orf41', description='family with sequence similarity 198 member A', other_designations='family with sequence similarity 198, member A', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42979267, end_position_on_the_genomic_accession=43057715, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729086, CurrentID=0, Status='live', Symbol='LOC729086', Aliases='', description='vesicular, overexpressed in cancer, prosurvival protein 1 pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145232840, end_position_on_the_genomic_accession=145235776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729088, CurrentID=0, Status='live', Symbol='LOC729088', Aliases='', description='uncharacterized LOC729088', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  204801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204801, CurrentID=0, Status='live', Symbol='NLRP11', Aliases='CLR19.6, NALP11, NOD17, PAN10, PYPAF6, PYPAF7', description='NLR family, pyrin domain containing 11', other_designations='NACHT, leucine rich repeat and PYD containing 11|PAAD- and NACHT-containing protein 10B|PAAD-and NACHT domain-containing protein 10|PYRIN-containing APAF1-like protein 6|nucleotide-binding oligomerization domain protein 17|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 11', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55785397, end_position_on_the_genomic_accession=55836762, orientation='minus', exon_count=12, OMIM=609664, no_hdr0=''),
  729092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729092, CurrentID=0, Status='live', Symbol='AGAP5', Aliases='CTGLF2', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 5', other_designations='AGAP-5|centaurin, gamma-like family, member 2|centaurin-gamma-like family member 2', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73674275, end_position_on_the_genomic_accession=73697796, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  729096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729096, CurrentID=0, Status='live', Symbol='BMS1P4', Aliases='BMS1LP4', description='BMS1, ribosome biogenesis factor pseudogene 4', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 4|BMS1 ribosome biogenesis factor pseudogene 4|BMS1L pseudogene 4', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73699151, end_position_on_the_genomic_accession=73730514, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  729097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729097, CurrentID=0, Status='live', Symbol='RPS4XP14', Aliases='RPS4P14, RPS4X_5_1190', description='ribosomal protein S4X pseudogene 14', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1757166, end_position_on_the_genomic_accession=1758032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729101, CurrentID=0, Status='live', Symbol='IMMP1LP1', Aliases='', description='inner mitochondrial membrane peptidase subunit 1 pseudogene 1', other_designations='IMP1 inner mitochondrial membrane peptidase-like (S. cerevisiae) pseudogene 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63632194, end_position_on_the_genomic_accession=63632822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729102, CurrentID=0, Status='live', Symbol='HNRNPA1P22', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 22', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42889273, end_position_on_the_genomic_accession=42890503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372333, CurrentID=0, Status='live', Symbol='LOC105372333', Aliases='', description='uncharacterized LOC105372333', other_designations='CTD-2291D10.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23074945, end_position_on_the_genomic_accession=23102535, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372334, CurrentID=0, Status='live', Symbol='LOC105372334', Aliases='', description='uncharacterized LOC105372334', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23479678, end_position_on_the_genomic_accession=23489468, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927098, CurrentID=0, Status='live', Symbol='LOC101927098', Aliases='', description='uncharacterized LOC101927098', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40694036, end_position_on_the_genomic_accession=40717316, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129455, CurrentID=0, Status='live', Symbol='LOC100129455', Aliases='', description='uncharacterized LOC100129455', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176124296, end_position_on_the_genomic_accession=176129631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466906, CurrentID=0, Status='live', Symbol='MIR6124', Aliases='hsa-mir-6124', description='microRNA 6124', other_designations='microRNA mir-6124', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12163683, end_position_on_the_genomic_accession=12163767, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729119, CurrentID=0, Status='live', Symbol='HMGB2P1', Aliases='', description='high mobility group box 2 pseudogene 1', other_designations='high-mobility group box 2 pseudogene 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1203209, end_position_on_the_genomic_accession=1203820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729121, CurrentID=0, Status='live', Symbol='RGPD4-AS1', Aliases='', description='RGPD4 antisense RNA 1 (head to head)', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107823064, end_position_on_the_genomic_accession=107826769, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729123, CurrentID=0, Status='live', Symbol='RPL7P21', Aliases='RPL7_10_623', description='ribosomal protein L7 pseudogene 21', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143319595, end_position_on_the_genomic_accession=143333790, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379434, CurrentID=0, Status='live', Symbol='LOC105379434', Aliases='', description='uncharacterized LOC105379434', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=60898314, end_position_on_the_genomic_accession=60899726, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729126, CurrentID=0, Status='live', Symbol='LOC729126', Aliases='', description='GrpE-like 1, mitochondrial (E. coli) pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66434633, end_position_on_the_genomic_accession=66435157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372337, CurrentID=0, Status='live', Symbol='LOC105372337', Aliases='', description='uncharacterized LOC105372337', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23525631, end_position_on_the_genomic_accession=23532679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729131, CurrentID=0, Status='live', Symbol='KRT8P7', Aliases='', description='keratin 8 pseudogene 7', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119602619, end_position_on_the_genomic_accession=119604380, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729138, CurrentID=0, Status='live', Symbol='RPLP1P2', Aliases='RPLP1_2_185', description='ribosomal protein, large, P1 pseudogene 2', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240998409, end_position_on_the_genomic_accession=240998867, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  204851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204851, CurrentID=0, Status='live', Symbol='HIPK1', Aliases='Myak, Nbak2', description='homeodomain interacting protein kinase 1', other_designations='homeodomain interacting protein kinase 1-like protein|nuclear body associated kinase 2b|nuclear body-associated kinase 2', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113929197, end_position_on_the_genomic_accession=113977869, orientation='plus', exon_count=24, OMIM=608003, no_hdr0=''),
  729141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729141, CurrentID=0, Status='live', Symbol='LOC729141', Aliases='', description='chromosome 18 open reading frame 25 pseudogene', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188227189, end_position_on_the_genomic_accession=188228379, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372340, CurrentID=0, Status='live', Symbol='LOC105372340', Aliases='', description='uncharacterized LOC105372340', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23937669, end_position_on_the_genomic_accession=23947817, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729146, CurrentID=0, Status='live', Symbol='RPL9P28', Aliases='RPL9_16_1554', description='ribosomal protein L9 pseudogene 28', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48691017, end_position_on_the_genomic_accession=48691718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102466911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466911, CurrentID=0, Status='live', Symbol='MIR6785', Aliases='hsa-mir-6785', description='microRNA 6785', other_designations='microRNA mir-6785', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75498548, end_position_on_the_genomic_accession=75498628, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729148, CurrentID=0, Status='live', Symbol='NUS1P1', Aliases='', description='NUS1 dehydrodolichyl diphosphate synthase subunit pseudogene 1', other_designations='nuclear undecaprenyl pyrophosphate synthase 1 homolog (S. cerevisiae) pseudogene 1|nuclear undecaprenyl pyrophosphate synthase 1 pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47509795, end_position_on_the_genomic_accession=47513660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729150, CurrentID=0, Status='live', Symbol='RPS3AP14', Aliases='RPS3A_7_401', description='ribosomal protein S3a pseudogene 14', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125795082, end_position_on_the_genomic_accession=125795929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372341, CurrentID=0, Status='live', Symbol='LOC105372341', Aliases='', description='uncharacterized LOC105372341', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102466912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466912, CurrentID=0, Status='live', Symbol='MIR6871', Aliases='hsa-mir-6871', description='microRNA 6871', other_designations='microRNA mir-6871', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41169023, end_position_on_the_genomic_accession=41169078, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729156, CurrentID=0, Status='live', Symbol='GTF2IRD1P1', Aliases='', description='GTF2I repeat domain containing 1 pseudogene 1', other_designations='GTF2I repeat domain containing 1 pseusogene 1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66809993, end_position_on_the_genomic_accession=66844826, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  729159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729159, CurrentID=0, Status='live', Symbol='LOC729159', Aliases='', description='UPF0607 protein ENSP00000381418-like', other_designations='Putative UPF0607 protein ENSP00000381418', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=60358455, end_position_on_the_genomic_accession=60359763, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729162, CurrentID=0, Status='live', Symbol='LOC729162', Aliases='', description='uncharacterized LOC729162', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3890928, end_position_on_the_genomic_accession=3926111, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729163, CurrentID=0, Status='live', Symbol='FABP5P2', Aliases='FABP5L2', description='fatty acid binding protein 5 pseudogene 2', other_designations='Fatty acid-binding protein, epidermal (E-FABP) (Psoriasis-associated fatty acid-binding protein homolog) (PA-FABP)|fatty acid binding protein 5-like 2', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51966681, end_position_on_the_genomic_accession=51967340, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729164, CurrentID=0, Status='live', Symbol='LOC729164', Aliases='', description='hCG1732469', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109343760, end_position_on_the_genomic_accession=109345032, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372344, CurrentID=0, Status='live', Symbol='LOC105372344', Aliases='', description='uncharacterized LOC105372344', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729171, CurrentID=0, Status='live', Symbol='ANKRD20A8P', Aliases='ANKRD20B', description='ankyrin repeat domain 20 family member A8, pseudogene', other_designations='ankyrin repeat domain 20 family, member A8, pseudogene|ankyrin repeat domain 20B|ankyrin repeat domain protein 18A pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94760928, end_position_on_the_genomic_accession=94857075, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  729173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729173, CurrentID=0, Status='live', Symbol='LOC729173', Aliases='', description='uncharacterized LOC729173', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120168544, end_position_on_the_genomic_accession=120171679, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100508631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508631, CurrentID=0, Status='live', Symbol='LOC100508631', Aliases='', description='uncharacterized LOC100508631', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37001802, end_position_on_the_genomic_accession=37020706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729176, CurrentID=0, Status='live', Symbol='KATNBL1P6', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 6', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146801669, end_position_on_the_genomic_accession=146803824, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729177, CurrentID=0, Status='live', Symbol='NBAT1', Aliases='CASC14, NBAT-1', description='neuroblastoma associated transcript 1', other_designations='cancer susceptibility candidate 14 (non-protein coding)|neuroblastoma associated transcript-1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22134602, end_position_on_the_genomic_accession=22147193, orientation='minus', exon_count=3, OMIM=616206, no_hdr0=''),
  729178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729178, CurrentID=0, Status='live', Symbol='STXBP5-AS1', Aliases='', description='STXBP5 antisense RNA 1', other_designations='STXBP5 antisense RNA 1 (non-protein coding)', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146841388, end_position_on_the_genomic_accession=147204614, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  729182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729182, CurrentID=0, Status='live', Symbol='FAM91A3P', Aliases='', description='family with sequence similarity 91 member A3, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143766479, end_position_on_the_genomic_accession=143771822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729187, CurrentID=0, Status='live', Symbol='LOC729187', Aliases='', description='uncharacterized LOC729187', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=78884873, end_position_on_the_genomic_accession=78886481, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729188, CurrentID=0, Status='live', Symbol='RPS26P58', Aliases='RPS26_33_1769', description='ribosomal protein S26 pseudogene 58', other_designations='40S ribosomal protein S26 pseudogene', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24429490, end_position_on_the_genomic_accession=24429949, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102466918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466918, CurrentID=0, Status='live', Symbol='MIR8059', Aliases='hsa-mir-8059', description='microRNA 8059', other_designations='microRNA mir-8059', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50768650, end_position_on_the_genomic_accession=50768730, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372348, CurrentID=0, Status='live', Symbol='LOC105372348', Aliases='', description='uncharacterized LOC105372348', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27849465, end_position_on_the_genomic_accession=27861940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729196, CurrentID=0, Status='live', Symbol='UNC93B5', Aliases='', description='unc-93 homolog B5 pseudogene (C. elegans)', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67711255, end_position_on_the_genomic_accession=67716006, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729198, CurrentID=0, Status='live', Symbol='ANAPC10P1', Aliases='ANAPC10P', description='anaphase promoting complex subunit 10 pseudogene 1', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52253567, end_position_on_the_genomic_accession=52254395, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729200, CurrentID=0, Status='live', Symbol='LOC729200', Aliases='', description='small nuclear ribonucleoprotein polypeptide B pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149111788, end_position_on_the_genomic_accession=149112744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729201, CurrentID=0, Status='live', Symbol='SPACA5B', Aliases='', description='sperm acrosome associated 5B', other_designations='SLLP-X|lysozyme-like protein 5|sperm-specific lysozyme-like protein X', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48130657, end_position_on_the_genomic_accession=48132613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729203, CurrentID=0, Status='live', Symbol='RPL31P30', Aliases='RPL31_2_226', description='ribosomal protein L31 pseudogene 30', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61856695, end_position_on_the_genomic_accession=61857062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729208, CurrentID=0, Status='live', Symbol='RPL21P14', Aliases='HsT16795, RPL21_53_1436', description='ribosomal protein L21 pseudogene 14', other_designations='', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58869619, end_position_on_the_genomic_accession=58870167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729209, CurrentID=0, Status='live', Symbol='RPSAP26', Aliases='RPSA_3_227', description='ribosomal protein SA pseudogene 26', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62146297, end_position_on_the_genomic_accession=62147326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729211, CurrentID=0, Status='live', Symbol='KRT18P15', Aliases='KRT18L5', description='keratin 18 pseudogene 15', other_designations='keratin 18-like 5', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32258908, end_position_on_the_genomic_accession=32260289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729212, CurrentID=0, Status='live', Symbol='RPL13AP26', Aliases='RPL13A_13_1745', description='ribosomal protein L13a pseudogene 26', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30653877, end_position_on_the_genomic_accession=30655315, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729215, CurrentID=0, Status='live', Symbol='RPL32P29', Aliases='RPL32_8_1362', description='ribosomal protein L32 pseudogene 29', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49577355, end_position_on_the_genomic_accession=49577842, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129472, CurrentID=0, Status='live', Symbol='PTP4A1P6', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 6', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86431407, end_position_on_the_genomic_accession=86434877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729218, CurrentID=0, Status='live', Symbol='LOC729218', Aliases='', description='uncharacterized LOC729218', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118591744, end_position_on_the_genomic_accession=118635025, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  729220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729220, CurrentID=0, Status='live', Symbol='FLJ45513', Aliases='', description='uncharacterized LOC729220', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49844715, end_position_on_the_genomic_accession=49848837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729222, CurrentID=0, Status='live', Symbol='LOC729222', Aliases='', description='liprin-beta-1-like', other_designations='', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27100129, end_position_on_the_genomic_accession=27172070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129473, CurrentID=0, Status='live', Symbol='LOC100129473', Aliases='', description='uncharacterized LOC100129473', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61766243, end_position_on_the_genomic_accession=61767601, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729224, CurrentID=0, Status='live', Symbol='LOC729224', Aliases='', description='uncharacterized LOC729224', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202113618, end_position_on_the_genomic_accession=202117070, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729230, CurrentID=0, Status='live', Symbol='CCR2', Aliases='CC-CKR-2, CCR-2A, CCR2B, CD192, CKR2, CKR2A, CKR2B, CMKBR2, MCP-1-R, CCR2', description='chemokine (C-C motif) receptor 2', other_designations='MCP-1 receptor|monocyte chemoattractant protein 1 receptor|monocyte chemotactic protein 1 receptor', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46353744, end_position_on_the_genomic_accession=46360940, orientation='plus', exon_count=4, OMIM=601267, no_hdr0=''),
  729233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729233, CurrentID=0, Status='live', Symbol='PRR20B', Aliases='', description='proline rich 20B', other_designations='FLJ40296 protein family member', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57147488, end_position_on_the_genomic_accession=57150509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729234, CurrentID=0, Status='live', Symbol='FAHD2CP', Aliases='', description='fumarylacetoacetate hydrolase domain containing 2C, pseudogene', other_designations='fumarylacetoacetate hydrolase domain containing 2 pseudogene|fumarylacetoacetate hydrolase domain containing 2A pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96010551, end_position_on_the_genomic_accession=96023136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729236, CurrentID=0, Status='live', Symbol='RPS13P2', Aliases='RPS13_1_51', description='ribosomal protein S13 pseudogene 2', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52772153, end_position_on_the_genomic_accession=52772674, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729238, CurrentID=0, Status='live', Symbol='SFTPA2', Aliases='COLEC5, PSAP, PSP-A, PSPA, SFTP1B, SP-A, SPA2, SPAII, SFTPA2', description='surfactant protein A2', other_designations='35 kDa pulmonary surfactant-associated protein|alveolar proteinosis protein|collectin 5|surfactant, pulmonary-associated protein A2A', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79555852, end_position_on_the_genomic_accession=79560459, orientation='minus', exon_count=6, OMIM=178642, no_hdr0=''),
  729240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729240, CurrentID=0, Status='live', Symbol='PRR20C', Aliases='', description='proline rich 20C', other_designations='FLJ40296 protein family member', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57154061, end_position_on_the_genomic_accession=57157082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729241, CurrentID=0, Status='live', Symbol='RPL21P117', Aliases='RPL21_53_1439', description='ribosomal protein L21 pseudogene 117', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59552572, end_position_on_the_genomic_accession=59553133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729246, CurrentID=0, Status='live', Symbol='PRR20D', Aliases='', description='proline rich 20D', other_designations='FLJ40296 protein family member', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57160632, end_position_on_the_genomic_accession=57163653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729248, CurrentID=0, Status='live', Symbol='ZBTB45P2', Aliases='', description='zinc finger and BTB domain containing 45 pseudogene 2', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110383111, end_position_on_the_genomic_accession=110384843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729249, CurrentID=0, Status='live', Symbol='KLHL2P1', Aliases='MAVP1', description='kelch like family member 2 pseudogene 1', other_designations='kelch-like 2, Mayven pseudogene 1|kelch-like family member 2 pseudogene 1', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119334329, end_position_on_the_genomic_accession=119378233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729250, CurrentID=0, Status='live', Symbol='PRR20E', Aliases='', description='proline rich 20E', other_designations='FLJ40296 protein family member', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57167197, end_position_on_the_genomic_accession=57170218, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729252, CurrentID=0, Status='live', Symbol='KRT16P1', Aliases='KRT14P, KRT16P', description='keratin 16 pseudogene 1', other_designations='keratin 14 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18439908, end_position_on_the_genomic_accession=18442895, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100129478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129478, CurrentID=0, Status='live', Symbol='LOC100129478', Aliases='', description='chromosome 4 open reading frame 46 pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=231084, end_position_on_the_genomic_accession=232010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729254, CurrentID=0, Status='live', Symbol='LOC729254', Aliases='', description='hCG2045843', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191792840, end_position_on_the_genomic_accession=191820249, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  729257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729257, CurrentID=0, Status='live', Symbol='RPS24P4', Aliases='RPS24_2_174', description='ribosomal protein S24 pseudogene 4', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230611943, end_position_on_the_genomic_accession=230612448, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729258, CurrentID=0, Status='live', Symbol='RPS11P6', Aliases='RPS11_2_1249', description='ribosomal protein S11 pseudogene 6', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64396010, end_position_on_the_genomic_accession=64397000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729259, CurrentID=0, Status='live', Symbol='RPL17P43', Aliases='RPL17_24_1526', description='ribosomal protein L17 pseudogene 43', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19496904, end_position_on_the_genomic_accession=19497564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729262, CurrentID=0, Status='live', Symbol='NUTM2B', Aliases='FAM22B, bA119F19.1', description='NUT family member 2B', other_designations='family with sequence similarity 22, member B|protein FAM22B', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79699700, end_position_on_the_genomic_accession=79715486, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  729264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729264, CurrentID=0, Status='live', Symbol='TP53TG3D', Aliases='TP53TG3, TP53TG3B, TP53TG3C, TP53TG3E, TP53TG3F', description='TP53 target 3D', other_designations='TP53-inducible gene 3 protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32253324, end_position_on_the_genomic_accession=32255928, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645832, CurrentID=0, Status='live', Symbol='SEBOX', Aliases='OG-9, OG9, OG9X', description='SEBOX homeobox', other_designations='homeobox OG-9|skin-, embryo-, brain- and oocyte-specific homeobox', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28364268, end_position_on_the_genomic_accession=28365151, orientation='minus', exon_count=3, OMIM=610975, no_hdr0=''),
  729268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729268, CurrentID=0, Status='live', Symbol='LOC729268', Aliases='', description='hCG2013212', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=49543781, end_position_on_the_genomic_accession=49575694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729274, CurrentID=0, Status='live', Symbol='RPS2P53', Aliases='RPS2_24_1626', description='ribosomal protein S2 pseudogene 53', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12284314, end_position_on_the_genomic_accession=12285229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271202, CurrentID=0, Status='live', Symbol='RPL37P17', Aliases='RPL37_7_1015', description='ribosomal protein L37 pseudogene 17', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130724786, end_position_on_the_genomic_accession=130725050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370710, CurrentID=0, Status='live', Symbol='LOC105370710', Aliases='', description='uncharacterized LOC105370710', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729277, CurrentID=0, Status='live', Symbol='RPL23AP47', Aliases='RPL23A_21_778', description='ribosomal protein L23a pseudogene 47', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170739837, end_position_on_the_genomic_accession=170745879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729279, CurrentID=0, Status='live', Symbol='RPL28P5', Aliases='RPL28_2_1625', description='ribosomal protein L28 pseudogene 5', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12260126, end_position_on_the_genomic_accession=12260582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129483, CurrentID=0, Status='live', Symbol='LOC100129483', Aliases='', description='solute carrier family 35 (GDP-fucose transporter), member C2 pseudogene', other_designations='solute carrier family 35, member C2 pseudogene', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16897654, end_position_on_the_genomic_accession=16898521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729288, CurrentID=0, Status='live', Symbol='ZNF286B', Aliases='ZNF286C, ZNF286L, ZNF590', description='zinc finger protein 286B', other_designations='zinc finger 286C pseudogene|zinc finger protein 590', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18658429, end_position_on_the_genomic_accession=18682259, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372364, CurrentID=0, Status='live', Symbol='LOC105372364', Aliases='', description='uncharacterized LOC105372364', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31980165, end_position_on_the_genomic_accession=31985978, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729291, CurrentID=0, Status='live', Symbol='LOC729291', Aliases='', description='uncharacterized LOC729291', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28188334, end_position_on_the_genomic_accession=28240184, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  414246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414246, CurrentID=0, Status='live', Symbol='CDH23-AS1', Aliases='C10orf106, NCRNA00223, bA327E2.3', description='CDH23 antisense RNA 1', other_designations='CDH23 antisense RNA 1 (non-protein coding)', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729293, CurrentID=0, Status='live', Symbol='RPS18P10', Aliases='RPS18_2_763', description='ribosomal protein S18 pseudogene 10', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140922417, end_position_on_the_genomic_accession=140922954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372365, CurrentID=0, Status='live', Symbol='LOC105372365', Aliases='', description='uncharacterized LOC105372365', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31837380, end_position_on_the_genomic_accession=31850720, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729296, CurrentID=0, Status='live', Symbol='LOC729296', Aliases='', description='uncharacterized LOC729296', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60087686, end_position_on_the_genomic_accession=60101387, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729298, CurrentID=0, Status='live', Symbol='PCNPP3', Aliases='', description='PEST containing nuclear protein pseudogene 3', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65644310, end_position_on_the_genomic_accession=65646476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729299, CurrentID=0, Status='live', Symbol='PMS2P8', Aliases='PMS2L16', description='PMS1 homolog 2, mismatch repair system component pseudogene 8', other_designations='postmeiotic segregation increased 2 pseudogene 8|postmeiotic segregation increased 2-like 16', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73037373, end_position_on_the_genomic_accession=73041445, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729301, CurrentID=0, Status='live', Symbol='RPL17P9', Aliases='RPL17_1_25', description='ribosomal protein L17 pseudogene 9', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26454599, end_position_on_the_genomic_accession=26455260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729305, CurrentID=0, Status='live', Symbol='LOC729305', Aliases='', description='uncharacterized LOC729305', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=134605490, end_position_on_the_genomic_accession=134633703, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129487, CurrentID=0, Status='live', Symbol='RPL24P9', Aliases='', description='ribosomal protein L24 pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3953016, end_position_on_the_genomic_accession=3953606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729308, CurrentID=0, Status='live', Symbol='RPL31P45', Aliases='RPL31_22_1025', description='ribosomal protein L31 pseudogene 45', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22266990, end_position_on_the_genomic_accession=22267205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369578, CurrentID=0, Status='live', Symbol='LOC105369578', Aliases='', description='uncharacterized LOC105369578', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131175563, end_position_on_the_genomic_accession=131187098, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  729310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729310, CurrentID=0, Status='live', Symbol='KYNUP2', Aliases='', description='kynureninase pseudogene 2', other_designations='kynureninase (L-kynurenine hydrolase) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19082861, end_position_on_the_genomic_accession=19084831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372368, CurrentID=0, Status='live', Symbol='LOC105372368', Aliases='', description='uncharacterized LOC105372368', other_designations='CTD-2538C1.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32718298, end_position_on_the_genomic_accession=32719801, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729316, CurrentID=0, Status='live', Symbol='LOC729316', Aliases='', description='POM121 membrane glycoprotein pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=46737447, end_position_on_the_genomic_accession=46747179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729317, CurrentID=0, Status='live', Symbol='LOC729317', Aliases='', description='voltage-dependent anion channel 2 pseudogene', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65205076, end_position_on_the_genomic_accession=65206278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372369, CurrentID=0, Status='live', Symbol='LOC105372369', Aliases='', description='uncharacterized LOC105372369', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33304498, end_position_on_the_genomic_accession=33323203, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729321, CurrentID=0, Status='live', Symbol='KYNUP1', Aliases='', description='kynureninase pseudogene 1', other_designations='kynureninase (L-kynurenine hydrolase) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19167258, end_position_on_the_genomic_accession=19169228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729323, CurrentID=0, Status='live', Symbol='KYNUP3', Aliases='', description='kynureninase pseudogene 3', other_designations='kynureninase (L-kynurenine hydrolase) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19209033, end_position_on_the_genomic_accession=19210997, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729324, CurrentID=0, Status='live', Symbol='LOC729324', Aliases='', description='hCG1986447', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=65452469, end_position_on_the_genomic_accession=65455079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372370, CurrentID=0, Status='live', Symbol='LOC105372370', Aliases='', description='uncharacterized LOC105372370', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33521942, end_position_on_the_genomic_accession=33525662, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729327, CurrentID=0, Status='live', Symbol='MED28P7', Aliases='', description='mediator complex subunit 28 pseudogene 7', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25669660, end_position_on_the_genomic_accession=25670404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729330, CurrentID=0, Status='live', Symbol='OC90', Aliases='PLA2L', description='otoconin 90', other_designations='phospholipase A2 homolog', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132024220, end_position_on_the_genomic_accession=132059380, orientation='minus', exon_count=14, OMIM=601658, no_hdr0=''),
  105372371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372371, CurrentID=0, Status='live', Symbol='LOC105372371', Aliases='', description='uncharacterized LOC105372371', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34000326, end_position_on_the_genomic_accession=34005416, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729332, CurrentID=0, Status='live', Symbol='RPL27P8', Aliases='RPL27_3_316', description='ribosomal protein L27 pseudogene 8', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206040389, end_position_on_the_genomic_accession=206040877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729334, CurrentID=0, Status='live', Symbol='LOC729334', Aliases='', description='cytospin B pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20647300, end_position_on_the_genomic_accession=20655932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129492, CurrentID=0, Status='live', Symbol='LOC100129492', Aliases='', description='small nuclear ribonucleoprotein Sm D1 pseudogene', other_designations='small nuclear ribonucleoprotein D1 polypeptide 16kDa pseudogene', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44122054, end_position_on_the_genomic_accession=44122595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729338, CurrentID=0, Status='live', Symbol='CETN4P', Aliases='', description='centrin 4, pseudogene', other_designations='CETN4 pseudogene|centrin EF-hand protein 4, pseudogene', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122730189, end_position_on_the_genomic_accession=122732458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729340, CurrentID=0, Status='live', Symbol='RPL17P36', Aliases='RPL17_11_1102', description='ribosomal protein L17 pseudogene 36', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118872014, end_position_on_the_genomic_accession=118872637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729342, CurrentID=0, Status='live', Symbol='NPM1P46', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 46', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197379385, end_position_on_the_genomic_accession=197380989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106635520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635520, CurrentID=0, Status='live', Symbol='SNORA87', Aliases='', description='small nucleolar RNA, H/ACA box 87', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635521, CurrentID=0, Status='live', Symbol='SNORA88', Aliases='', description='small nucleolar RNA, H/ACA box 88', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635522, CurrentID=0, Status='live', Symbol='SNORA89', Aliases='', description='small nucleolar RNA, H/ACA box 89', other_designations='', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729347, CurrentID=0, Status='live', Symbol='AKR1B1P3', Aliases='ALDRL3', description='aldo-keto reductase family 1, member B1 pseudogene 3', other_designations='CoroMarker|aldehyde reductase (aldose reductase)-like 3|aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16483362, end_position_on_the_genomic_accession=16484709, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106635524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635524, CurrentID=0, Status='live', Symbol='SNORA91', Aliases='', description='small nucleolar RNA, H/ACA box 91', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635525, CurrentID=0, Status='live', Symbol='SNORA92', Aliases='', description='small nucleolar RNA, H/ACA box 92', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635526, CurrentID=0, Status='live', Symbol='SNORA93', Aliases='', description='small nucleolar RNA, H/ACA box 93', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635527, CurrentID=0, Status='live', Symbol='LINC01323', Aliases='', description='long intergenic non-protein coding RNA 1323', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635528, CurrentID=0, Status='live', Symbol='SNORA94', Aliases='', description='small nucleolar RNA, H/ACA box 94', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635529, CurrentID=0, Status='live', Symbol='SNORA95', Aliases='', description='small nucleolar RNA, H/ACA box 95', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635530, CurrentID=0, Status='live', Symbol='SNORA98', Aliases='', description='small nucleolar RNA, H/ACA box 98', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729355, CurrentID=0, Status='live', Symbol='TP53TG3B', Aliases='TP53TG3E, TP53TG3F', description='TP53 target 3B', other_designations='TP53-inducible gene 3 protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33360877, end_position_on_the_genomic_accession=33363478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106635532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635532, CurrentID=0, Status='live', Symbol='SNORA100', Aliases='', description='small nucleolar RNA, H/ACA box 100', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635533, CurrentID=0, Status='live', Symbol='SNORA103', Aliases='', description='small nucleolar RNA, H/ACA box 103', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635534, CurrentID=0, Status='live', Symbol='SNORA104', Aliases='', description='small nucleolar RNA, H/ACA box 104', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729359, CurrentID=0, Status='live', Symbol='PLIN4', Aliases='KIAA1881, S3-12', description='perilipin 4', other_designations='adipocyte protein S3-12', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4502180, end_position_on_the_genomic_accession=4518541, orientation='minus', exon_count=9, OMIM=613247, no_hdr0=''),
  106635536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635536, CurrentID=0, Status='live', Symbol='RN7SKP176', Aliases='', description='RNA, 7SK small nuclear pseudogene 176', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635537, CurrentID=0, Status='live', Symbol='RN7SL860P', Aliases='', description='RNA, 7SL, cytoplasmic 860, pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729362, CurrentID=0, Status='live', Symbol='RPL36AP37', Aliases='RPL36A_14_1122', description='ribosomal protein L36a pseudogene 37', other_designations='cell growth-inhibiting protein 15', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16974629, end_position_on_the_genomic_accession=16975044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106635539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635539, CurrentID=0, Status='live', Symbol='SNORA105A', Aliases='', description='small nucleolar RNA, H/ACA box 105A', other_designations='', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635540, CurrentID=0, Status='live', Symbol='SNORA105B', Aliases='', description='small nucleolar RNA, H/ACA box 105B', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635541, CurrentID=0, Status='live', Symbol='SNORA105C', Aliases='', description='small nucleolar RNA, H/ACA box 105C', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635542, CurrentID=0, Status='live', Symbol='SNORA107', Aliases='', description='small nucleolar RNA, H/ACA box 107', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635543, CurrentID=0, Status='live', Symbol='SNORA108', Aliases='', description='small nucleolar RNA, H/ACA box 108', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635544, CurrentID=0, Status='live', Symbol='SNORA109', Aliases='', description='small nucleolar RNA, H/ACA box 109', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635545, CurrentID=0, Status='live', Symbol='SNORA110', Aliases='', description='small nucleolar RNA, H/ACA box 110', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635546, CurrentID=0, Status='live', Symbol='SNORA111', Aliases='', description='small nucleolar RNA, H/ACA box 111', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635547, CurrentID=0, Status='live', Symbol='SNORD129', Aliases='', description='small nucleolar RNA, C/D box 129', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635548, CurrentID=0, Status='live', Symbol='SNORD130', Aliases='', description='small nucleolar RNA, C/D box 130', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635549, CurrentID=0, Status='live', Symbol='SNORD131', Aliases='', description='small nucleolar RNA, C/D box 131', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635550, CurrentID=0, Status='live', Symbol='SNORD132', Aliases='', description='small nucleolar RNA, C/D box 132', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729375, CurrentID=0, Status='live', Symbol='FAM86HP', Aliases='', description='family with sequence similarity 86 member H, pseudogene', other_designations='family with sequence similarity 86, member H, pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130097782, end_position_on_the_genomic_accession=130111433, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  729384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729384, CurrentID=0, Status='live', Symbol='TRIM49D2', Aliases='TRIM48L1, TRIM49D1P, TRIM49L, TRIM49L1, TRIM49D2', description='tripartite motif containing 49D2', other_designations='tripartite motif containing 49D2, pseudogene|tripartite motif-containing 49-like|tripartite motif-containing protein 49-like protein 1|tripartite motif-containing protein 49D1|tripartite motif-containing protein 49D2', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89924064, end_position_on_the_genomic_accession=89933063, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  645852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645852, CurrentID=0, Status='live', Symbol='LOC645852', Aliases='', description='splicing factor 3a subunit 2 pseudogene', other_designations='splicing factor 3a, subunit 2, 66kDa pseudogene', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126672044, end_position_on_the_genomic_accession=126673315, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729389, CurrentID=0, Status='live', Symbol='RPS6P25', Aliases='RPS6_13_1629', description='ribosomal protein S6 pseudogene 25', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12894094, end_position_on_the_genomic_accession=12894911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729391, CurrentID=0, Status='live', Symbol='RFWD2P1', Aliases='RFWD2P', description='ring finger and WD repeat domain 2, E3 ubiquitin protein ligase pseudogene 1', other_designations='ring finger and WD repeat domain 2 pseudogene 1', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8411985, end_position_on_the_genomic_accession=8414631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729392, CurrentID=0, Status='live', Symbol='POM121L6P', Aliases='bA239L20.2', description='POM121 transmembrane nucleoporin like 6, pseudogene', other_designations='POM121 membrane glycoprotein-like 6 pseudogene|POM121 transmembrane nucleoporin-like 6 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26870667, end_position_on_the_genomic_accession=26898161, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729396, CurrentID=0, Status='live', Symbol='GAGE12J', Aliases='GAGE11', description='G antigen 12J', other_designations='G antigen 11', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49322030, end_position_on_the_genomic_accession=49329387, orientation='plus', exon_count=5, OMIM=300733, no_hdr0=''),
  102466953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466953, CurrentID=0, Status='live', Symbol='MIR6082', Aliases='hsa-mir-6082', description='microRNA 6082', other_designations='microRNA mir-6082', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171186184, end_position_on_the_genomic_accession=171186292, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729402, CurrentID=0, Status='live', Symbol='RPL21P16', Aliases='RPL21_38_1104', description='ribosomal protein L21 pseudogene 16', other_designations='', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120354645, end_position_on_the_genomic_accession=120355226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729403, CurrentID=0, Status='live', Symbol='GAPDHP61', Aliases='GAPDHL6, GAPDL6', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 61', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 6, pseudogene', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64528468, end_position_on_the_genomic_accession=64529740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729406, CurrentID=0, Status='live', Symbol='PPP1R14BP2', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 14B pseudogene 2', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 14B pseudogene 2', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207681390, end_position_on_the_genomic_accession=207682230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729408, CurrentID=0, Status='live', Symbol='GAGE2D', Aliases='CT4.8, GAGE-2D, GAGE-8, GAGE8', description='G antigen 2D', other_designations='cancer/testis antigen 4.8', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49560844, end_position_on_the_genomic_accession=49568208, orientation='plus', exon_count=5, OMIM=300735, no_hdr0=''),
  729409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729409, CurrentID=0, Status='live', Symbol='C1GALT1P1', Aliases='', description='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase 1 pseudogene 1', other_designations='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase, 1 pseudogene', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69284217, end_position_on_the_genomic_accession=69285684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729412, CurrentID=0, Status='live', Symbol='PSMA6P2', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 6 pseudogene 2', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12825810, end_position_on_the_genomic_accession=12826996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372385, CurrentID=0, Status='live', Symbol='LOC105372385', Aliases='', description='uncharacterized LOC105372385', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36127955, end_position_on_the_genomic_accession=36140023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729416, CurrentID=0, Status='live', Symbol='RPLP1P6', Aliases='RPLP1_5_629', description='ribosomal protein, large, P1 pseudogene 6', other_designations='hCG1641617', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151765860, end_position_on_the_genomic_accession=151766368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729420, CurrentID=0, Status='live', Symbol='LMO7DN', Aliases='C13orf45', description='LMO7 downstream neighbor', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75870571, end_position_on_the_genomic_accession=75883812, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729422, CurrentID=0, Status='live', Symbol='GAGE12C', Aliases='GAGE-12B', description='G antigen 12C', other_designations='GAGE-12C|g antigen 12C/D/E', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49532177, end_position_on_the_genomic_accession=49539541, orientation='plus', exon_count=5, OMIM=300727, no_hdr0=''),
  729423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729423, CurrentID=0, Status='live', Symbol='HNRNPA1P6', Aliases='HNRNPA1L, HNRNPA1L1, HNRPA1L', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 6', other_designations='heterogeneous nuclear ribonucleoprotein A1-like 1', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58047849, end_position_on_the_genomic_accession=58049373, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271203, CurrentID=0, Status='live', Symbol='RPL39P23', Aliases='RPL39_10_809', description='ribosomal protein L39 pseudogene 23', other_designations='', map_location='7p12-p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50843391, end_position_on_the_genomic_accession=50843782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645859, CurrentID=0, Status='live', Symbol='CCDC58P1', Aliases='', description='coiled-coil domain containing 58 pseudogene 1', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10623616, end_position_on_the_genomic_accession=10624309, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729428, CurrentID=0, Status='live', Symbol='GAGE12B', Aliases='GAGE-12B', description='G antigen 12B', other_designations='g antigen 12C/D/E', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49341183, end_position_on_the_genomic_accession=49529921, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729431, CurrentID=0, Status='live', Symbol='GAGE12E', Aliases='GAGE-12B', description='G antigen 12E', other_designations='GAGE-12E|g antigen 12C/D/E', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49551289, end_position_on_the_genomic_accession=49558649, orientation='plus', exon_count=5, OMIM=300729, no_hdr0=''),
  205147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205147, CurrentID=0, Status='live', Symbol='AMER3', Aliases='FAM123C', description='APC membrane recruitment protein 3', other_designations='family with sequence similarity 123C|protein FAM123C', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130755504, end_position_on_the_genomic_accession=130768134, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729438, CurrentID=0, Status='live', Symbol='GATSL2', Aliases='GATSL1', description='GATS protein-like 2', other_designations='GATS protein-like 1|GATS-like protein 1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74964818, end_position_on_the_genomic_accession=75031532, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106635615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635615, CurrentID=0, Status='live', Symbol='SNORD137', Aliases='', description='small nucleolar RNA, C/D box 137', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729440, CurrentID=0, Status='live', Symbol='CCDC61', Aliases='', description='coiled-coil domain containing 61', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45995451, end_position_on_the_genomic_accession=46018616, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  729442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729442, CurrentID=0, Status='live', Symbol='GAGE12H', Aliases='', description='G antigen 12H', other_designations='GAGE-12H', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49579949, end_position_on_the_genomic_accession=49587304, orientation='plus', exon_count=5, OMIM=300732, no_hdr0=''),
  105372390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372390, CurrentID=0, Status='live', Symbol='LOC105372390', Aliases='', description='uncharacterized LOC105372390', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37076087, end_position_on_the_genomic_accession=37078071, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729447, CurrentID=0, Status='live', Symbol='GAGE2A', Aliases='CT4.2, GAGE-2, GAGE-2A, GAGE2', description='G antigen 2A', other_designations='G antigen 2|cancer/testis antigen 4.2|cancer/testis antigen family 4, member 2', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49589515, end_position_on_the_genomic_accession=49596827, orientation='plus', exon_count=5, OMIM=300720, no_hdr0=''),
  729449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729449, CurrentID=0, Status='live', Symbol='RPL7P50', Aliases='RPL7_28_1618', description='ribosomal protein L7 pseudogene 50', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6593306, end_position_on_the_genomic_accession=6594120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729450, CurrentID=0, Status='live', Symbol='SUSD2P2', Aliases='', description='sushi domain containing 2 pseudogene 2', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18623339, end_position_on_the_genomic_accession=18625227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729451, CurrentID=0, Status='live', Symbol='LOC729451', Aliases='', description='uncharacterized LOC729451', other_designations='', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54671068, end_position_on_the_genomic_accession=54687918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729453, CurrentID=0, Status='live', Symbol='PMS2P6', Aliases='PMS2L14', description='PMS1 homolog 2, mismatch repair system component pseudogene 6', other_designations='postmeiotic segregation increased 2 pseudogene 6|postmeiotic segregation increased 2-like 14', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73093644, end_position_on_the_genomic_accession=73097591, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729454, CurrentID=0, Status='live', Symbol='DSTNP4', Aliases='', description='destrin (actin depolymerizing factor) pseudogene 4', other_designations='destrin-like', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39212661, end_position_on_the_genomic_accession=39214678, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129512, CurrentID=0, Status='live', Symbol='POU5F1P2', Aliases='', description='POU class 5 homeobox 1 pseudogene 2', other_designations='POU domain class 5, transcription factor 1 pseudogene 2', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102621236, end_position_on_the_genomic_accession=102622331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729458, CurrentID=0, Status='live', Symbol='LOC729458', Aliases='', description='methyl-CpG-binding domain protein 3-like 2', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729461, CurrentID=0, Status='live', Symbol='LOC729461', Aliases='', description='uncharacterized LOC729461', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18369332, end_position_on_the_genomic_accession=18391105, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  729467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729467, CurrentID=0, Status='live', Symbol='HSD52', Aliases='', description='uncharacterized LOC729467', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59131936, end_position_on_the_genomic_accession=59146807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729468, CurrentID=0, Status='live', Symbol='PGM5P4', Aliases='', description='phosphoglucomutase 5 pseudogene 4', other_designations='putative PGM5-like protein 1', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113541191, end_position_on_the_genomic_accession=113570962, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375081, CurrentID=0, Status='live', Symbol='LOC105375081', Aliases='', description='uncharacterized LOC105375081', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47129172, end_position_on_the_genomic_accession=47138637, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  729475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729475, CurrentID=0, Status='live', Symbol='RAD51AP2', Aliases='', description='RAD51 associated protein 2', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17510584, end_position_on_the_genomic_accession=17520580, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  729480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729480, CurrentID=0, Status='live', Symbol='RPL23AP5', Aliases='RPL23AL, RPL23A_35_1465, RPL2B, c367G8.3, rjd10', description='ribosomal protein L23a pseudogene 5', other_designations='60S ribosomal protein L23A like', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=386746, end_position_on_the_genomic_accession=387290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372396, CurrentID=0, Status='live', Symbol='LOC105372396', Aliases='', description='uncharacterized LOC105372396', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38024951, end_position_on_the_genomic_accession=38028428, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102466967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466967, CurrentID=0, Status='live', Symbol='MIR6130', Aliases='hsa-mir-6130', description='microRNA 6130', other_designations='microRNA mir-6130', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23079284, end_position_on_the_genomic_accession=23079392, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729484, CurrentID=0, Status='live', Symbol='RPL21P18', Aliases='RPL21_46_1252', description='ribosomal protein L21 pseudogene 18', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66037195, end_position_on_the_genomic_accession=66037744, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729486, CurrentID=0, Status='live', Symbol='IL9RP3', Aliases='', description='interleukin 9 receptor pseudogene 3', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28928, end_position_on_the_genomic_accession=36491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372397, CurrentID=0, Status='live', Symbol='LOC105372397', Aliases='', description='uncharacterized LOC105372397', other_designations='CTB-186G2.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38597319, end_position_on_the_genomic_accession=38601694, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729490, CurrentID=0, Status='live', Symbol='NCOR1P2', Aliases='', description='nuclear receptor corepressor 1 pseudogene 2', other_designations='nuclear receptor co-repressor 1 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22061960, end_position_on_the_genomic_accession=22070605, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729493, CurrentID=0, Status='live', Symbol='GAPDHP38', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 38', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46558560, end_position_on_the_genomic_accession=46559844, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729494, CurrentID=0, Status='live', Symbol='ARPC3P3', Aliases='', description='actin related protein 2/3 complex subunit 3 pseudogene 3', other_designations='actin related protein 2/3 complex, subunit 3 pseudogene 3|actin related protein 2/3 complex, subunit 3, 21kDa pseudogene', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94188204, end_position_on_the_genomic_accession=94189036, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729495, CurrentID=0, Status='live', Symbol='ABHD17AP4', Aliases='FAM108A5, FAM108A5P', description='abhydrolase domain containing 17A pseudogene P4', other_designations='family with sequence similarity 108, member A5, pseudogene|putative abhydrolase domain-containing protein FAM108A5', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20667597, end_position_on_the_genomic_accession=20671339, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729496, CurrentID=0, Status='live', Symbol='LOC729496', Aliases='', description='coiled-coil domain containing 59 pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149491256, end_position_on_the_genomic_accession=149492312, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729497, CurrentID=0, Status='live', Symbol='LOC729497', Aliases='', description='nuclear receptor coactivator 4 pseudogene', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145622597, end_position_on_the_genomic_accession=145626132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729500, CurrentID=0, Status='live', Symbol='RPL12P14', Aliases='RPL12_1_13', description='ribosomal protein L12 pseudogene 14', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15792727, end_position_on_the_genomic_accession=15793346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729501, CurrentID=0, Status='live', Symbol='ZNF962P', Aliases='', description='zinc finger protein 962, pseudogene', other_designations='Putative zinc finger protein ENSP00000344568', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18466319, end_position_on_the_genomic_accession=18487412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927111, CurrentID=0, Status='live', Symbol='SUCLG2-AS1', Aliases='', description='SUCLG2 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67654697, end_position_on_the_genomic_accession=67947713, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  729505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729505, CurrentID=0, Status='live', Symbol='HMGN2P24', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 24', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 24', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40698249, end_position_on_the_genomic_accession=40699435, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106635682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635682, CurrentID=0, Status='live', Symbol='SNORD140', Aliases='', description='small nucleolar RNA, C/D box 140', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635683, CurrentID=0, Status='live', Symbol='SNORD141A', Aliases='', description='small nucleolar RNA, C/D box 141A', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106635684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106635684, CurrentID=0, Status='live', Symbol='SNORD141B', Aliases='', description='small nucleolar RNA, C/D box 141B', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729515, CurrentID=0, Status='live', Symbol='TMEM242', Aliases='BM033, C6orf35', description='transmembrane protein 242', other_designations='UPF0463 transmembrane protein C6orf35', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157289022, end_position_on_the_genomic_accession=157324259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729522, CurrentID=0, Status='live', Symbol='AACSP1', Aliases='AACSL', description='acetoacetyl-CoA synthetase pseudogene 1', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178764861, end_position_on_the_genomic_accession=178776276, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  729524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729524, CurrentID=0, Status='live', Symbol='FEM1AP4', Aliases='', description='fem-1 homolog a (C. elegans) pseudogene 4', other_designations='fem-1 homolog a pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18665114, end_position_on_the_genomic_accession=18667934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103908604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103908604, CurrentID=0, Status='live', Symbol='LOC103908604', Aliases='', description='uncharacterized LOC103908604', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729528, CurrentID=0, Status='live', Symbol='PRAMEF14', Aliases='', description='PRAME family member 14', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13341892, end_position_on_the_genomic_accession=13347134, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  729530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729530, CurrentID=0, Status='live', Symbol='CNOT6LP1', Aliases='', description='CCR4-NOT transcription complex subunit 6-like pseudogene 1', other_designations='CCR4-NOT transcription complex, subunit 6-like pseudogene 1', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56003305, end_position_on_the_genomic_accession=56007179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729532, CurrentID=0, Status='live', Symbol='LOC729532', Aliases='', description='zinc finger protein 207 pseudogene', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203634512, end_position_on_the_genomic_accession=203636139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729533, CurrentID=0, Status='live', Symbol='FAM72A', Aliases='LMPIP, Ugene', description='family with sequence similarity 72 member A', other_designations='LMP1-induced protein|family with sequence similarity 72, member A|latent membrane protein 1-induced protein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206186182, end_position_on_the_genomic_accession=206204919, orientation='minus', exon_count=8, OMIM=614710, no_hdr0=''),
  729535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729535, CurrentID=0, Status='live', Symbol='FAM207BP', Aliases='', description='family with sequence similarity 207 member B, pseudogene', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18697003, end_position_on_the_genomic_accession=18697686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729536, CurrentID=0, Status='live', Symbol='RPL34P31', Aliases='RPL34_16_1528', description='ribosomal protein L34 pseudogene 31', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27274107, end_position_on_the_genomic_accession=27274522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927092, CurrentID=0, Status='live', Symbol='LOC101927092', Aliases='', description='protein GVQW1-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47135491, end_position_on_the_genomic_accession=47137144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  205251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205251, CurrentID=0, Status='live', Symbol='LINC00116', Aliases='NCRNA00116', description='long intergenic non-protein coding RNA 116', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110211529, end_position_on_the_genomic_accession=110222940, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  729540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729540, CurrentID=0, Status='live', Symbol='RGPD6', Aliases='RGP6, RGPD7, RanBP2L1, RanBP2L2', description='RANBP2-like and GRIP domain containing 6', other_designations='RANBP2-like and GRIP domain-containing protein 7|ran-binding protein 2-like 1/2|ranBP2-like 1/2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110513802, end_position_on_the_genomic_accession=110610888, orientation='minus', exon_count=31, OMIM=612709, no_hdr0=''),
  105372406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372406, CurrentID=0, Status='live', Symbol='LOC105372406', Aliases='', description='uncharacterized LOC105372406', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42590073, end_position_on_the_genomic_accession=42590848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729545, CurrentID=0, Status='live', Symbol='LOC729545', Aliases='', description='WBSCR19-like protein 5-like', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=76517185, end_position_on_the_genomic_accession=76524125, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372407, CurrentID=0, Status='live', Symbol='LOC105372407', Aliases='', description='uncharacterized LOC105372407', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42522769, end_position_on_the_genomic_accession=42525966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101060304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060304, CurrentID=0, Status='live', Symbol='LOC101060304', Aliases='', description='protein phosphatase 1 regulatory subunit 26-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372408, CurrentID=0, Status='live', Symbol='LOC105372408', Aliases='', description='uncharacterized LOC105372408', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42510906, end_position_on_the_genomic_accession=42533791, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729557, CurrentID=0, Status='live', Symbol='LOC729557', Aliases='', description='density-regulated protein pseudogene', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34245761, end_position_on_the_genomic_accession=34246537, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729558, CurrentID=0, Status='live', Symbol='LOC729558', Aliases='', description='uncharacterized LOC729558', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=151500241, end_position_on_the_genomic_accession=151502697, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  205272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205272, CurrentID=0, Status='live', Symbol='CNN2P8', Aliases='', description='calponin 2 pseudogene 8', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94737296, end_position_on_the_genomic_accession=94738437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729561, CurrentID=0, Status='live', Symbol='NAP1L1P1', Aliases='', description='nucleosome assembly protein 1-like 1 pseudogene 1', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126067540, end_position_on_the_genomic_accession=126068600, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729562, CurrentID=0, Status='live', Symbol='LOC729562', Aliases='', description='uncharacterized LOC729562', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=23013918, end_position_on_the_genomic_accession=23023728, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  729564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729564, CurrentID=0, Status='live', Symbol='HMGN2P41', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 41', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 41', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35802090, end_position_on_the_genomic_accession=35803283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729565, CurrentID=0, Status='live', Symbol='NDUFB8P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B8 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, 19kDa pseudogene|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, pseudogene 2', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59884990, end_position_on_the_genomic_accession=59885649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729566, CurrentID=0, Status='live', Symbol='ZBTB8OSP1', Aliases='', description='zinc finger and BTB domain containing 8 opposite strand pseudogene 1', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150720975, end_position_on_the_genomic_accession=150721616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729568, CurrentID=0, Status='live', Symbol='SUSD2P1', Aliases='', description='sushi domain containing 2 pseudogene 1', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21370061, end_position_on_the_genomic_accession=21371949, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102466982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466982, CurrentID=0, Status='live', Symbol='MIR6729', Aliases='hsa-mir-6729', description='microRNA 6729', other_designations='microRNA mir-6729', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12029158, end_position_on_the_genomic_accession=12029222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729574, CurrentID=0, Status='live', Symbol='FAM231A', Aliases='FAM231C', description='family with sequence similarity 231 member A', other_designations='UPF0627 protein ENSP00000341061/ENSP00000339743|family with sequence similarity 231, member A|protein FAM231A/C', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16673003, end_position_on_the_genomic_accession=16673512, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100271204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271204, CurrentID=0, Status='live', Symbol='RPL5P25', Aliases='RPL5_10_1021', description='ribosomal protein L5 pseudogene 25', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13057890, end_position_on_the_genomic_accession=13058725, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372412, CurrentID=0, Status='live', Symbol='LOC105372412', Aliases='', description='phospholipase A2 inhibitor and Ly6/PLAUR domain-containing protein-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43687591, end_position_on_the_genomic_accession=43700078, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102466983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466983, CurrentID=0, Status='live', Symbol='MIR6766', Aliases='hsa-mir-6766', description='microRNA 6766', other_designations='microRNA mir-6766', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89326739, end_position_on_the_genomic_accession=89326810, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729582, CurrentID=0, Status='live', Symbol='DIRC3', Aliases='', description='disrupted in renal carcinoma 3', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217284023, end_position_on_the_genomic_accession=217756593, orientation='minus', exon_count=13, OMIM=608262, no_hdr0=''),
  729583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729583, CurrentID=0, Status='live', Symbol='LINC01556', Aliases='C6orf100, dJ25J6.5', description='long intergenic non-protein coding RNA 1556', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28943784, end_position_on_the_genomic_accession=28944538, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466984, CurrentID=0, Status='live', Symbol='MIR6801', Aliases='hsa-mir-6801', description='microRNA 6801', other_designations='microRNA mir-6801', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52222020, end_position_on_the_genomic_accession=52222098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729587, CurrentID=0, Status='live', Symbol='FAM231C', Aliases='', description='family with sequence similarity 231 member C', other_designations='family with sequence similarity 231, member C', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372414, CurrentID=0, Status='live', Symbol='LOC105372414', Aliases='', description='uncharacterized LOC105372414', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44022976, end_position_on_the_genomic_accession=44025263, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102466985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466985, CurrentID=0, Status='live', Symbol='MIR6837', Aliases='hsa-mir-6837', description='microRNA 6837', other_designations='microRNA mir-6837', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44051766, end_position_on_the_genomic_accession=44051829, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729595, CurrentID=0, Status='live', Symbol='HMGB3P22', Aliases='', description='high mobility group box 3 pseudogene 22', other_designations='high-mobility group box 3 pseudogene 22', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179694178, end_position_on_the_genomic_accession=179694786, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729597, CurrentID=0, Status='live', Symbol='SPDYE6', Aliases='', description='speedy/RINGO cell cycle regulator family member E6', other_designations='speedy homolog E6', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102345752, end_position_on_the_genomic_accession=102357047, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  729602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729602, CurrentID=0, Status='live', Symbol='NPIPB1P', Aliases='', description='nuclear pore complex interacting protein family member B1, pseudogene', other_designations='nuclear pore complex interacting protein family, member B1, pseudogene|nuclear pore complex interacting protein pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11619578, end_position_on_the_genomic_accession=11634053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729603, CurrentID=0, Status='live', Symbol='LOC729603', Aliases='', description='calcineurin-like EF-hand protein 1 pseudogene', other_designations='calcium binding protein P22 pseudogene', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160093082, end_position_on_the_genomic_accession=160096212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372417, CurrentID=0, Status='live', Symbol='LOC105372417', Aliases='', description='uncharacterized LOC105372417', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44742615, end_position_on_the_genomic_accession=44748454, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729608, CurrentID=0, Status='live', Symbol='RPF2P1', Aliases='BXDC1P, C20orf53, RPF2P, bA353C18.4', description='ribosome production factor 2 homolog pseudogene 1', other_designations='brix domain containing 1 pseudogene|ribosome production factor 2 homolog (S. cerevisiae) pseudogene 1|ribosome production factor 2 homolog pseudogene|ribosome production factor 2 pseudogene', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35752795, end_position_on_the_genomic_accession=35753737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729609, CurrentID=0, Status='live', Symbol='LOC729609', Aliases='', description='uncharacterized LOC729609', other_designations='', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19986817, end_position_on_the_genomic_accession=19989779, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729610, CurrentID=0, Status='live', Symbol='FAHD2P1', Aliases='FAHD2P', description='fumarylacetoacetate hydrolase domain containing 2 pseudogene 1', other_designations='fumarylacetoacetate hydrolase domain containing 2A pseudogene', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70671789, end_position_on_the_genomic_accession=70673032, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729611, CurrentID=0, Status='live', Symbol='RPL29P30', Aliases='RPL29_20_1448', description='ribosomal protein L29 pseudogene 30', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70796520, end_position_on_the_genomic_accession=70797146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729614, CurrentID=0, Status='live', Symbol='FLJ37453', Aliases='', description='uncharacterized LOC729614', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15834215, end_position_on_the_genomic_accession=15848147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  205327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205327, CurrentID=0, Status='live', Symbol='C2orf69', Aliases='', description='chromosome 2 open reading frame 69', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199911256, end_position_on_the_genomic_accession=199928273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729617, CurrentID=0, Status='live', Symbol='RPL23AP74', Aliases='RPL23A_37_1566', description='ribosomal protein L23a pseudogene 74', other_designations='hCG2001000', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61067208, end_position_on_the_genomic_accession=61070947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729621, CurrentID=0, Status='live', Symbol='RPL36AP46', Aliases='RPL36A_22_1565', description='ribosomal protein L36a pseudogene 46', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60727165, end_position_on_the_genomic_accession=60727482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927115, CurrentID=0, Status='live', Symbol='CTB-12O2.1', Aliases='', description='uncharacterized LOC101927115', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151958898, end_position_on_the_genomic_accession=152270449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372420, CurrentID=0, Status='live', Symbol='LOC105372420', Aliases='', description='uncharacterized LOC105372420', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45153503, end_position_on_the_genomic_accession=45157364, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729627, CurrentID=0, Status='live', Symbol='PRR23A', Aliases='', description='proline rich 23A', other_designations='UPF0572 protein ENSP00000372650|proline-rich protein 23C', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139003962, end_position_on_the_genomic_accession=139006268, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729630, CurrentID=0, Status='live', Symbol='DYNLL1P7', Aliases='', description='dynein, light chain, LC8-type 1 pseudogene 7', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=85381105, end_position_on_the_genomic_accession=85381471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729633, CurrentID=0, Status='live', Symbol='MRS2P2', Aliases='', description='MRS2 pseudogene 2', other_designations='MRS2 magnesium homeostasis factor homolog pseudogene 2', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71848292, end_position_on_the_genomic_accession=71850983, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729635, CurrentID=0, Status='live', Symbol='HMGB3P19', Aliases='', description='high mobility group box 3 pseudogene 19', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 19', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153938261, end_position_on_the_genomic_accession=153939125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372422, CurrentID=0, Status='live', Symbol='LOC105372422', Aliases='', description='uncharacterized LOC105372422', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46198212, end_position_on_the_genomic_accession=46204862, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372423, CurrentID=0, Status='live', Symbol='LOC105372423', Aliases='', description='uncharacterized LOC105372423', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46195233, end_position_on_the_genomic_accession=46197647, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729646, CurrentID=0, Status='live', Symbol='RPL31P4', Aliases='RPL31_28_1375', description='ribosomal protein L31 pseudogene 4', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58794630, end_position_on_the_genomic_accession=58795071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729648, CurrentID=0, Status='live', Symbol='ZNF812P', Aliases='ZNF812', description='zinc finger protein 812, pseudogene', other_designations='zinc finger protein 812 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9688855, end_position_on_the_genomic_accession=9700817, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  729650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729650, CurrentID=0, Status='live', Symbol='METTL15P3', Aliases='', description='methyltransferase like 15 pseudogene 3', other_designations='methyltransferase 5 domain containing 1 pseudogene', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22435598, end_position_on_the_genomic_accession=22436815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102466995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466995, CurrentID=0, Status='live', Symbol='MIR7156', Aliases='hsa-mir-7156', description='microRNA 7156', other_designations='microRNA mir-7156', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77060143, end_position_on_the_genomic_accession=77060202, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729652, CurrentID=0, Status='live', Symbol='LOC729652', Aliases='', description='spidroin-2', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2464936, end_position_on_the_genomic_accession=2468207, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729654, CurrentID=0, Status='live', Symbol='LOC729654', Aliases='', description='inositol polyphosphate multikinase pseudogene', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15907767, end_position_on_the_genomic_accession=15912139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729658, CurrentID=0, Status='live', Symbol='PACRG-AS3', Aliases='', description='PACRG antisense RNA 3', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163181645, end_position_on_the_genomic_accession=163192002, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729659, CurrentID=0, Status='live', Symbol='S100A11P1', Aliases='S100A11P, S100A14', description='S100 calcium binding protein A11 pseudogene 1', other_designations='S100 calcium-binding protein A14 (calgizzarin)', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103261951, end_position_on_the_genomic_accession=103262481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729662, CurrentID=0, Status='live', Symbol='RPL22P3', Aliases='RPL22_1_15', description='ribosomal protein L22 pseudogene 3', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16369093, end_position_on_the_genomic_accession=16369551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729665, CurrentID=0, Status='live', Symbol='CCDC175', Aliases='C14orf38, c14_5395', description='coiled-coil domain containing 175', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59505067, end_position_on_the_genomic_accession=59582114, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  729666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729666, CurrentID=0, Status='live', Symbol='LDHAP5', Aliases='LDHAL5', description='lactate dehydrogenase A pseudogene 5', other_designations='lactate dehydrogenase A-like 5', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118932096, end_position_on_the_genomic_accession=118933283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729667, CurrentID=0, Status='live', Symbol='TRMT112P3', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 3', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62950664, end_position_on_the_genomic_accession=62951188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729668, CurrentID=0, Status='live', Symbol='GOLGA2P6', Aliases='', description='golgin A2 pseudogene 6', other_designations='golgi autoantigen, golgin subfamily a, 6 pseudogene|golgin A6 family, member A pseudogene', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30364327, end_position_on_the_genomic_accession=30371907, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  729673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729673, CurrentID=0, Status='live', Symbol='PPIAP18', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 18', other_designations='', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44873322, end_position_on_the_genomic_accession=44874063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729674, CurrentID=0, Status='live', Symbol='LOC729674', Aliases='', description='proteasome 26S subunit, ATPase 2 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 2 pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132175334, end_position_on_the_genomic_accession=132176811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729677, CurrentID=0, Status='live', Symbol='RPL7P16', Aliases='RPL7_5_405', description='ribosomal protein L7 pseudogene 16', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132243460, end_position_on_the_genomic_accession=132244286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729678, CurrentID=0, Status='live', Symbol='LINC00847', Aliases='', description='long intergenic non-protein coding RNA 847', other_designations='CTC-205M6.2', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180830957, end_position_on_the_genomic_accession=180835726, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729679, CurrentID=0, Status='live', Symbol='RPS2P51', Aliases='RPS2_25_1632', description='ribosomal protein S2 pseudogene 51', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16412661, end_position_on_the_genomic_accession=16413626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729680, CurrentID=0, Status='live', Symbol='LOC729680', Aliases='', description='uncharacterized LOC729680', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=21276266, end_position_on_the_genomic_accession=21277998, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729681, CurrentID=0, Status='live', Symbol='LOC729681', Aliases='', description='uncharacterized LOC729681', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=166511208, end_position_on_the_genomic_accession=166511878, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927117, CurrentID=0, Status='live', Symbol='PLCG1-AS1', Aliases='TOP1-AS1', description='PLCG1 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41097994, end_position_on_the_genomic_accession=41138000, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  729683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729683, CurrentID=0, Status='live', Symbol='LOC729683', Aliases='', description='uncharacterized LOC729683', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63700338, end_position_on_the_genomic_accession=63702685, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129550, CurrentID=0, Status='live', Symbol='LOC100129550', Aliases='', description='uncharacterized LOC100129550', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122886513, end_position_on_the_genomic_accession=122892416, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729686, CurrentID=0, Status='live', Symbol='NPM1P42', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 42', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72899197, end_position_on_the_genomic_accession=72900360, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729687, CurrentID=0, Status='live', Symbol='HMGN2P25', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 25', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 25', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141863869, end_position_on_the_genomic_accession=141865377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729689, CurrentID=0, Status='live', Symbol='FAM90A2P', Aliases='', description='family with sequence similarity 90 member A2, pseudogene', other_designations='family with sequence similarity 90, member A2, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12173454, end_position_on_the_genomic_accession=12177267, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372431, CurrentID=0, Status='live', Symbol='LOC105372431', Aliases='', description='uncharacterized LOC105372431', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48673261, end_position_on_the_genomic_accession=48677067, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729696, CurrentID=0, Status='live', Symbol='LOC729696', Aliases='', description='uncharacterized LOC729696', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=74819711, end_position_on_the_genomic_accession=74822229, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372432, CurrentID=0, Status='live', Symbol='LOC105372432', Aliases='', description='uncharacterized LOC105372432', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48807081, end_position_on_the_genomic_accession=48810798, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372433, CurrentID=0, Status='live', Symbol='LOC105372433', Aliases='', description='uncharacterized LOC105372433', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49412952, end_position_on_the_genomic_accession=49416158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  342908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342908, CurrentID=0, Status='live', Symbol='ZNF404', Aliases='', description='zinc finger protein 404', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43872363, end_position_on_the_genomic_accession=43883964, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729707, CurrentID=0, Status='live', Symbol='LOC729707', Aliases='', description='transmembrane and tetratricopeptide repeat containing 4 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181113930, end_position_on_the_genomic_accession=181115654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729708, CurrentID=0, Status='live', Symbol='TPI1P1', Aliases='RCTPI1', description='triosephosphate isomerase 1 pseudogene 1', other_designations='TPI1 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76699755, end_position_on_the_genomic_accession=76701025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129554, CurrentID=0, Status='live', Symbol='LOC100129554', Aliases='', description='thyroid hormone receptor interactor 4 pseudogene', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139659902, end_position_on_the_genomic_accession=139661700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927118, CurrentID=0, Status='live', Symbol='LOC101927118', Aliases='', description='uncharacterized LOC101927118', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81523068, end_position_on_the_genomic_accession=81579444, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105372435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372435, CurrentID=0, Status='live', Symbol='LOC105372435', Aliases='', description='uncharacterized LOC105372435', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49808832, end_position_on_the_genomic_accession=49809816, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  205428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205428, CurrentID=0, Status='live', Symbol='C3orf58', Aliases='DIA1, GoPro49, HASF', description='chromosome 3 open reading frame 58', other_designations='Golgi Protein of 49 kDa|Golgi protein GoPro49|UPF0672 protein C3orf58|deleted in autism 1|hypoxia and AKT-induced stem cell factor|hypoxia and Akt induced stem cell factor', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143971798, end_position_on_the_genomic_accession=143992368, orientation='plus', exon_count=4, OMIM=612200, no_hdr0=''),
  105372436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372436, CurrentID=0, Status='live', Symbol='LOC105372436', Aliases='', description='uncharacterized LOC105372436', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50137496, end_position_on_the_genomic_accession=50148459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729722, CurrentID=0, Status='live', Symbol='LOC729722', Aliases='', description='ankyrin repeat domain-containing protein ENSP00000383090-like', other_designations='putative ankyrin repeat domain-containing protein ENSP00000383090', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=20077202, end_position_on_the_genomic_accession=20087671, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729723, CurrentID=0, Status='live', Symbol='DNAJC27-AS1', Aliases='', description='DNAJC27 antisense RNA 1', other_designations='DNAJC27 antisense RNA 1 (non-protein coding)', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24972112, end_position_on_the_genomic_accession=25039694, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729727, CurrentID=0, Status='live', Symbol='RPL12P1', Aliases='CICK0721Q.1, RPL12-L, RPL12_5_682, dJ570F3.5', description='ribosomal protein L12 pseudogene 1', other_designations='cICK0721Q.2 (60S Ribosomal Protein L12 LIKE protein)', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33400015, end_position_on_the_genomic_accession=33400644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729732, CurrentID=0, Status='live', Symbol='LOC729732', Aliases='', description='uncharacterized LOC729732', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12537079, end_position_on_the_genomic_accession=12665611, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  729737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729737, CurrentID=0, Status='live', Symbol='LOC729737', Aliases='', description='uncharacterized LOC729737', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=134773, end_position_on_the_genomic_accession=140566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729739, CurrentID=0, Status='live', Symbol='LOC729739', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74361552, end_position_on_the_genomic_accession=74362041, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375090, CurrentID=0, Status='live', Symbol='LOC105375090', Aliases='', description='uncharacterized LOC105375090', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52780884, end_position_on_the_genomic_accession=52787690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729747, CurrentID=0, Status='live', Symbol='ZNF878', Aliases='', description='zinc finger protein 878', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12043648, end_position_on_the_genomic_accession=12052967, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129561, CurrentID=0, Status='live', Symbol='MYL6P3', Aliases='', description='myosin light chain 6 pseudogene 3', other_designations='myosin, light chain 6 pseudogene 3', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65169271, end_position_on_the_genomic_accession=65169898, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729756, CurrentID=0, Status='live', Symbol='PRSS44', Aliases='TESSP-4, TESSP4', description='protease, serine 44', other_designations='protease, serine, 44|testis serine protease 4', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46811532, end_position_on_the_genomic_accession=46812558, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370726, CurrentID=0, Status='live', Symbol='LOC105370726', Aliases='', description='uncharacterized LOC105370726', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23361466, end_position_on_the_genomic_accession=23365069, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729759, CurrentID=0, Status='live', Symbol='OR4F29', Aliases='OR7-21', description='olfactory receptor family 4 subfamily F member 29', other_designations='olfactory receptor OR1-1|olfactory receptor OR7-21|olfactory receptor, family 4, subfamily F, member 29|seven transmembrane helix receptor', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=450740, end_position_on_the_genomic_accession=451678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729760, CurrentID=0, Status='live', Symbol='RPL12P18', Aliases='RPL12_3_246', description='ribosomal protein L12 pseudogene 18', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84874527, end_position_on_the_genomic_accession=84875128, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372443, CurrentID=0, Status='live', Symbol='LOC105372443', Aliases='', description='uncharacterized LOC105372443', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51209831, end_position_on_the_genomic_accession=51212934, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  729764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729764, CurrentID=0, Status='live', Symbol='RPL23AP45', Aliases='RPL23A_19_648', description='ribosomal protein L23a pseudogene 45', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181471961, end_position_on_the_genomic_accession=181474520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729767, CurrentID=0, Status='live', Symbol='CEACAM18', Aliases='', description='carcinoembryonic antigen-related cell adhesion molecule 18', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51478547, end_position_on_the_genomic_accession=51491261, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100129564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129564, CurrentID=0, Status='live', Symbol='PTP4A1P4', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 4', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86086940, end_position_on_the_genomic_accession=86089288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729770, CurrentID=0, Status='live', Symbol='LOC729770', Aliases='', description='wiskott-Aldrich syndrome protein homolog', other_designations='CTD-2308L22.1', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221572506, end_position_on_the_genomic_accession=221575974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927120, CurrentID=0, Status='live', Symbol='LOC101927120', Aliases='', description='uncharacterized LOC101927120', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56848478, end_position_on_the_genomic_accession=56878078, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  729774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729774, CurrentID=0, Status='live', Symbol='LOC729774', Aliases='', description='family with sequence similarity 207, member A pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14361262, end_position_on_the_genomic_accession=14361945, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372445, CurrentID=0, Status='live', Symbol='LOC105372445', Aliases='', description='uncharacterized LOC105372445', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51370195, end_position_on_the_genomic_accession=51373876, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376788, CurrentID=0, Status='live', Symbol='LOC105376788', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90167497, end_position_on_the_genomic_accession=90168771, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729779, CurrentID=0, Status='live', Symbol='PSAT1P3', Aliases='', description='phosphoserine aminotransferase 1 pseudogene 3', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=79054948, end_position_on_the_genomic_accession=79056055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729780, CurrentID=0, Status='live', Symbol='RPS10P8', Aliases='RPS10_1_104', description='ribosomal protein S10 pseudogene 8', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158195582, end_position_on_the_genomic_accession=158196167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129566, CurrentID=0, Status='live', Symbol='SNRPCP2', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 2', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76376626, end_position_on_the_genomic_accession=76377441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729786, CurrentID=0, Status='live', Symbol='GOLGA8CP', Aliases='GOLGA8C', description='golgin A8 family member C, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 8C|golgin A8 family, member C, pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20562349, end_position_on_the_genomic_accession=20575703, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  729789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729789, CurrentID=0, Status='live', Symbol='RPS15P5', Aliases='RPS15_3_372', description='ribosomal protein S15 pseudogene 5', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54626153, end_position_on_the_genomic_accession=54626680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729790, CurrentID=0, Status='live', Symbol='ARL6IP1P3', Aliases='', description='ADP ribosylation factor like GTPase 6 interacting protein 1 pseudogene 3', other_designations='ADP-ribosylation factor GTPase 6 interacting protein 1 pseudogene 3|ADP-ribosylation factor like GTPase 6 interacting protein 1 pseudogene 3|ADP-ribosylation factor-like 6 interacting protein 1 pseudogene 3', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80653425, end_position_on_the_genomic_accession=80653984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729798, CurrentID=0, Status='live', Symbol='RPL23AP63', Aliases='RPL23A_28_1140', description='ribosomal protein L23a pseudogene 63', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43718656, end_position_on_the_genomic_accession=43719383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729799, CurrentID=0, Status='live', Symbol='SEC14L1P1', Aliases='', description='SEC14-like 1 pseudogene 1', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43897303, end_position_on_the_genomic_accession=43899888, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  729800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729800, CurrentID=0, Status='live', Symbol='LOC729800', Aliases='', description='ZP domain-containing protein LOC729800-like', other_designations='Putative ZP domain-containing protein LOC729800', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122820363, end_position_on_the_genomic_accession=122826448, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129570, CurrentID=0, Status='live', Symbol='PEBP1P3', Aliases='', description='phosphatidylethanolamine binding protein 1 pseudogene 3', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198678338, end_position_on_the_genomic_accession=198680139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729806, CurrentID=0, Status='live', Symbol='ZNF731P', Aliases='ZNF731', description='zinc finger protein 731, pseudogene', other_designations='zinc finger protein 682 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247190342, end_position_on_the_genomic_accession=247191017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729809, CurrentID=0, Status='live', Symbol='DNM1P34', Aliases='DNM1DN8-1, DNM1DN8-5, DNM1DN8@', description='dynamin 1 pseudogene 34', other_designations='DNM1 pseudogene 34|Putative GED domain-containing protein LOC554175', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75298143, end_position_on_the_genomic_accession=75302920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729810, CurrentID=0, Status='live', Symbol='LOC729810', Aliases='', description='uncharacterized LOC729810', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=247493000, end_position_on_the_genomic_accession=247497878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729815, CurrentID=0, Status='live', Symbol='LOC729815', Aliases='', description='keratin-associated protein 5-1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79503400, end_position_on_the_genomic_accession=79512577, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  729816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729816, CurrentID=0, Status='live', Symbol='DHFRP2', Aliases='', description='dihydrofolate reductase pseudogene 2', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31363467, end_position_on_the_genomic_accession=31366965, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729817, CurrentID=0, Status='live', Symbol='HNRNPA3P14', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 14', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81426525, end_position_on_the_genomic_accession=81427411, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129573, CurrentID=0, Status='live', Symbol='LOC100129573', Aliases='', description='tRNA methyltransferase 2 homolog B (S. cerevisiae) pseudogene', other_designations='TRM2 tRNA methyltransferase 2 homolog B (S. cerevisiae) pseudogene', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184566480, end_position_on_the_genomic_accession=184567858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729828, CurrentID=0, Status='live', Symbol='ST13P20', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 20', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81721700, end_position_on_the_genomic_accession=81723058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129574, CurrentID=0, Status='live', Symbol='LOC100129574', Aliases='', description='chromosome 12 open reading frame 29 pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57120758, end_position_on_the_genomic_accession=57122622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729830, CurrentID=0, Status='live', Symbol='FAM160A1', Aliases='', description='family with sequence similarity 160 member A1', other_designations='family with sequence similarity 160, member A1', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151409156, end_position_on_the_genomic_accession=151670503, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  645927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645927, CurrentID=0, Status='live', Symbol='MCFD2P1', Aliases='C6orf44, MCFD2L, NUCB2-L, dJ153G14.2', description='multiple coagulation factor deficiency 2 pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27407723, end_position_on_the_genomic_accession=27409691, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729838, CurrentID=0, Status='live', Symbol='GGTLC4P', Aliases='GGT', description='gamma-glutamyltransferase light chain 4 pseudogene', other_designations='gamma-glutamyltransferase pseudogene', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24247948, end_position_on_the_genomic_accession=24249500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729839, CurrentID=0, Status='live', Symbol='LOC729839', Aliases='', description='DTW domain containing 2 pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35406945, end_position_on_the_genomic_accession=35407836, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729842, CurrentID=0, Status='live', Symbol='RPS2P50', Aliases='RPS2_23_1541', description='ribosomal protein S2 pseudogene 50', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36574408, end_position_on_the_genomic_accession=36575345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129577, CurrentID=0, Status='live', Symbol='LOC100129577', Aliases='', description='mitochondrial carrier 1 pseudogene', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138649472, end_position_on_the_genomic_accession=138650975, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  205564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205564, CurrentID=0, Status='live', Symbol='SENP5', Aliases='', description='SUMO1/sentrin specific peptidase 5', other_designations='SUMO1/sentrin specific protease 5|sentrin/SUMO-specific protease SENP5', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196867856, end_position_on_the_genomic_accession=196934714, orientation='plus', exon_count=12, OMIM=612845, no_hdr0=''),
  645930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645930, CurrentID=0, Status='live', Symbol='HMGN2P31', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 31', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 31', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5311423, end_position_on_the_genomic_accession=5312732, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729856, CurrentID=0, Status='live', Symbol='LOC729856', Aliases='', description='elongation factor 1-alpha-like', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23670777, end_position_on_the_genomic_accession=23671244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729857, CurrentID=0, Status='live', Symbol='RGPD2', Aliases='NUP358, RANBP2L2, RGP2', description='RANBP2-like and GRIP domain containing 2', other_designations='Ran binding protein 2-like 2|nucleoporin|ran-binding protein 2-like 2|ranBP2-like 2', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87757198, end_position_on_the_genomic_accession=87985814, orientation='minus', exon_count=29, OMIM=612705, no_hdr0=''),
  645931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645931, CurrentID=0, Status='live', Symbol='FEN1P1', Aliases='FENL1', description='flap structure-specific endonuclease 1 pseudogene 1', other_designations='flap endonuclease-like 1', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91328238, end_position_on_the_genomic_accession=91330256, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729860, CurrentID=0, Status='live', Symbol='LOC729860', Aliases='', description='uncharacterized LOC729860', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=134144782, end_position_on_the_genomic_accession=134146451, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729862, CurrentID=0, Status='live', Symbol='LSP1P3', Aliases='', description='lymphocyte-specific protein 1 pseudogene 3', other_designations='striatin, calmodulin binding protein pseudogene', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=28926870, end_position_on_the_genomic_accession=28927313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372460, CurrentID=0, Status='live', Symbol='LOC105372460', Aliases='', description='uncharacterized LOC105372460', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729866, CurrentID=0, Status='live', Symbol='LOC729866', Aliases='', description='hCG1994895', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=157893622, end_position_on_the_genomic_accession=157895688, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729867, CurrentID=0, Status='live', Symbol='LOC729867', Aliases='', description='uncharacterized LOC729867', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160202199, end_position_on_the_genomic_accession=160208869, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  729870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729870, CurrentID=0, Status='live', Symbol='LOC729870', Aliases='', description='serine/arginine repetitive matrix protein 1', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152934516, end_position_on_the_genomic_accession=152938450, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129581, CurrentID=0, Status='live', Symbol='LOC100129581', Aliases='', description='uncharacterized LOC100129581', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=7907028, end_position_on_the_genomic_accession=7907486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  729873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729873, CurrentID=0, Status='live', Symbol='TBC1D3', Aliases='PRC17A, TBC1D3F, TBC1D3', description='TBC1 domain family member 3', other_designations='Rab GTPase-activating protein PRC17|TBC1 domain family, member 3|TBC1 domain family, member 3F|prostate cancer gene 17 protein|protein TRE17-alpha', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38181476, end_position_on_the_genomic_accession=38192555, orientation='minus', exon_count=14, OMIM=607741, no_hdr0=''),
  729877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729877, CurrentID=0, Status='live', Symbol='TBC1D3H', Aliases='', description='TBC1 domain family member 3H', other_designations='TBC1 domain family member 3-like|TBC1 domain family, member 3H', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36377475, end_position_on_the_genomic_accession=36388441, orientation='minus', exon_count=15, OMIM=610811, no_hdr0=''),
  729879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729879, CurrentID=0, Status='live', Symbol='LOC729879', Aliases='', description='uncharacterized LOC729879', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=231275415, end_position_on_the_genomic_accession=231277244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372463, CurrentID=0, Status='live', Symbol='LOC105372463', Aliases='', description='uncharacterized LOC105372463', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55553867, end_position_on_the_genomic_accession=55555350, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929713, CurrentID=0, Status='live', Symbol='LOC101929713', Aliases='', description='uncharacterized LOC101929713', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219218524, end_position_on_the_genomic_accession=219445121, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  729885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729885, CurrentID=0, Status='live', Symbol='VN1R101P', Aliases='', description='vomeronasal 1 receptor 101 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53422798, end_position_on_the_genomic_accession=53423689, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729887, CurrentID=0, Status='live', Symbol='LOC729887', Aliases='', description='uncharacterized LOC729887', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84116181, end_position_on_the_genomic_accession=84117571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372464, CurrentID=0, Status='live', Symbol='LOC105372464', Aliases='', description='uncharacterized LOC105372464', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375095, CurrentID=0, Status='live', Symbol='LOC105375095', Aliases='', description='uncharacterized LOC105375095', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53474102, end_position_on_the_genomic_accession=53506911, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729894, CurrentID=0, Status='live', Symbol='ELMO2P1', Aliases='', description='engulfment and cell motility 2 pseudogene 1', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23083094, end_position_on_the_genomic_accession=23095571, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645937, CurrentID=0, Status='live', Symbol='LOC645937', Aliases='', description='zinc finger protein 598 pseudogene', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87944274, end_position_on_the_genomic_accession=87945023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729898, CurrentID=0, Status='live', Symbol='ZBTB8OSP2', Aliases='', description='zinc finger and BTB domain containing 8 opposite strand pseudogene 2', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231433772, end_position_on_the_genomic_accession=231434430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729900, CurrentID=0, Status='live', Symbol='LOC729900', Aliases='', description='golgin subfamily A member 6-like protein 2', other_designations='golgin subfamily A member 6-like protein 1-like', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22769365, end_position_on_the_genomic_accession=22776520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129586, CurrentID=0, Status='live', Symbol='LOC100129586', Aliases='', description='ubiquitin conjugating enzyme E2D 4 (putative) pseudogene', other_designations='ubiquitin-conjugating enzyme E2D 4 (putative) pseudogene', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149754302, end_position_on_the_genomic_accession=149754708, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729903, CurrentID=0, Status='live', Symbol='RPS16P10', Aliases='RPS16_5_1636', description='ribosomal protein S16 pseudogene 10', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20124980, end_position_on_the_genomic_accession=20125550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645939, CurrentID=0, Status='live', Symbol='PRKRIRP8', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 8', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240769221, end_position_on_the_genomic_accession=240771543, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729909, CurrentID=0, Status='live', Symbol='RPL36AP26', Aliases='RPL36A_8_787', description='ribosomal protein L36a pseudogene 26', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15701378, end_position_on_the_genomic_accession=15701781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729911, CurrentID=0, Status='live', Symbol='ANKRD34C-AS1', Aliases='', description='ANKRD34C antisense RNA 1', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79191707, end_position_on_the_genomic_accession=79283945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372468, CurrentID=0, Status='live', Symbol='LOC105372468', Aliases='', description='uncharacterized LOC105372468', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56314999, end_position_on_the_genomic_accession=56316029, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  729915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729915, CurrentID=0, Status='live', Symbol='LOC729915', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=34189680, end_position_on_the_genomic_accession=34193766, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129589, CurrentID=0, Status='live', Symbol='LOC100129589', Aliases='', description='TM2 domain containing 3 pseudogene', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85781891, end_position_on_the_genomic_accession=85782512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729920, CurrentID=0, Status='live', Symbol='ISPD', Aliases='MDDGA7, MDDGC7, Nip, hCG_1745121', description='isoprenoid synthase domain containing', other_designations='2-C-methyl-D-erythritol 4-phosphate cytidylyltransferase-like protein|4-diphosphocytidyl-2C-methyl-D-erythritol synthase homolog|notch1-induced protein', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15916851, end_position_on_the_genomic_accession=16530558, orientation='minus', exon_count=24, OMIM=614631, no_hdr0=''),
  729927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729927, CurrentID=0, Status='live', Symbol='RPS20P27', Aliases='RPS20_12_1158', description='ribosomal protein S20 pseudogene 27', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77813276, end_position_on_the_genomic_accession=77813796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729930, CurrentID=0, Status='live', Symbol='LOC729930', Aliases='', description='chromosome 11 open reading frame 74 pseudogene', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89289258, end_position_on_the_genomic_accession=89290362, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372471, CurrentID=0, Status='live', Symbol='LOC105372471', Aliases='', description='uncharacterized LOC105372471', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56485523, end_position_on_the_genomic_accession=56507755, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  729941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729941, CurrentID=0, Status='live', Symbol='LOC729941', Aliases='', description='PDZ domain containing 11 pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71330339, end_position_on_the_genomic_accession=71330758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  205655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205655, CurrentID=0, Status='live', Symbol='NBPF21P', Aliases='', description='neuroblastoma breakpoint family member 21, pseudogene', other_designations='neuroblastoma breakpoint family, member 21 (pseudogene)|neuroblastoma breakpoint family, member 21, pseudogene', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36616009, end_position_on_the_genomic_accession=36637797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729944, CurrentID=0, Status='live', Symbol='ZNF75CP', Aliases='ZNF75C', description='zinc finger protein 75C, pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78384033, end_position_on_the_genomic_accession=78387026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729945, CurrentID=0, Status='live', Symbol='LOC729945', Aliases='', description='Rho GTPase activating protein 21 pseudogene', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10839598, end_position_on_the_genomic_accession=10841553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729948, CurrentID=0, Status='live', Symbol='CDC26P1', Aliases='CDC26P', description='cell division cycle 26 pseudogene 1', other_designations='TCAG_1970866|cell division cycle 26 homolog pseudogene 1', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129409853, end_position_on_the_genomic_accession=129410556, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372474, CurrentID=0, Status='live', Symbol='LOC105372474', Aliases='', description='uncharacterized LOC105372474', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56863396, end_position_on_the_genomic_accession=56894601, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729950, CurrentID=0, Status='live', Symbol='LOC729950', Aliases='', description='uncharacterized LOC729950', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24628182, end_position_on_the_genomic_accession=24662198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729952, CurrentID=0, Status='live', Symbol='HMGB3P6', Aliases='', description='high mobility group box 3 pseudogene 6', other_designations='high-mobility group box 2 pseudogene|high-mobility group box 3 pseudogene 6', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164356767, end_position_on_the_genomic_accession=164357367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729955, CurrentID=0, Status='live', Symbol='RPL18AP5', Aliases='RPL18A_1_27', description='ribosomal protein L18a pseudogene 5', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27176711, end_position_on_the_genomic_accession=27177330, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729956, CurrentID=0, Status='live', Symbol='SHISA7', Aliases='', description='shisa family member 7', other_designations='UPF0626 protein A|shisa homolog 7', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55428738, end_position_on_the_genomic_accession=55442863, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729960, CurrentID=0, Status='live', Symbol='LOC729960', Aliases='', description='NADPH oxidase 4 pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49305715, end_position_on_the_genomic_accession=49380100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129596, CurrentID=0, Status='live', Symbol='LOC100129596', Aliases='', description='uncharacterized LOC100129596', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=146030740, end_position_on_the_genomic_accession=146046151, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103752554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752554, CurrentID=0, Status='live', Symbol='MYCNUT', Aliases='MYCNUN, lncUSMycN', description='MYCN upstream transcript (non-protein coding)', other_designations='MYCN upstream neighbor', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15920399, end_position_on_the_genomic_accession=15936018, orientation='plus', exon_count=3, OMIM=615968, no_hdr0=''),
  103752555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752555, CurrentID=0, Status='live', Symbol='SPATA17-AS1', Aliases='', description='SPATA17 antisense RNA 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217781198, end_position_on_the_genomic_accession=217785120, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94139, CurrentID=0, Status='live', Symbol='DFNA34', Aliases='', description='deafness, autosomal dominant 34', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  729966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729966, CurrentID=0, Status='live', Symbol='LOC729966', Aliases='', description='uncharacterized LOC729966', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18249950, end_position_on_the_genomic_accession=18255419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729967, CurrentID=0, Status='live', Symbol='MORN2', Aliases='BLOCK27, MOPT', description='MORN repeat containing 2', other_designations='MORN motif protein in testis|protein containing single MORN motif in testis', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38875962, end_position_on_the_genomic_accession=38882709, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  729968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729968, CurrentID=0, Status='live', Symbol='LOC729968', Aliases='', description='spore coat protein SP96', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227616991, end_position_on_the_genomic_accession=227618655, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729970, CurrentID=0, Status='live', Symbol='LOC729970', Aliases='', description='hCG2028352-like', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94928028, end_position_on_the_genomic_accession=94963270, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372478, CurrentID=0, Status='live', Symbol='LOC105372478', Aliases='', description='taste receptor type 1 member 2-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58019302, end_position_on_the_genomic_accession=58025640, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729974, CurrentID=0, Status='live', Symbol='RFPL4AL1', Aliases='', description='ret finger protein-like 4A-like 1', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55769118, end_position_on_the_genomic_accession=55773423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  729975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729975, CurrentID=0, Status='live', Symbol='LINC01530', Aliases='', description='long intergenic non-protein coding RNA 1530', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51591783, end_position_on_the_genomic_accession=51594380, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  729978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729978, CurrentID=0, Status='live', Symbol='LOC729978', Aliases='', description='nuclear pore complex interacting protein family, member A1 pseudogene', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11927245, end_position_on_the_genomic_accession=11942045, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372479, CurrentID=0, Status='live', Symbol='LOC105372479', Aliases='', description='uncharacterized LOC105372479', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58216034, end_position_on_the_genomic_accession=58217927, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129316, CurrentID=0, Status='live', Symbol='LOC100129316', Aliases='', description='uncharacterized LOC100129316', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91063294, end_position_on_the_genomic_accession=91075132, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729987, CurrentID=0, Status='live', Symbol='LOC729987', Aliases='', description='uncharacterized LOC729987', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98210711, end_position_on_the_genomic_accession=98272658, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103752581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752581, CurrentID=0, Status='live', Symbol='LINC01055', Aliases='', description='long intergenic non-protein coding RNA 1055', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45681199, end_position_on_the_genomic_accession=45701184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729991, CurrentID=0, Status='live', Symbol='BORCS8', Aliases='MEF2BNB', description='BLOC-1 related complex subunit 8', other_designations='MEF2B neighbor gene protein', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19176904, end_position_on_the_genomic_accession=19192591, orientation='minus', exon_count=6, OMIM=616601, no_hdr0=''),
  729992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729992, CurrentID=0, Status='live', Symbol='ST13P1', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 1', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12023560, end_position_on_the_genomic_accession=12024632, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729993, CurrentID=0, Status='live', Symbol='SHISA9', Aliases='CKAMP44', description='shisa family member 9', other_designations='cystine-knot AMPAR modulating protein|shisa homolog 9', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12901620, end_position_on_the_genomic_accession=13562919, orientation='plus', exon_count=10, OMIM=613346, no_hdr0=''),
  103752586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752586, CurrentID=0, Status='live', Symbol='BIRC6-AS2', Aliases='linc-birc6, megamind', description='BIRC6 antisense RNA 2', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32557520, end_position_on_the_genomic_accession=32574818, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103752587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752587, CurrentID=0, Status='live', Symbol='FOXC2-AS1', Aliases='ODRUL', description='FOXC2 antisense RNA 1', other_designations='osteosarcoma doxorubicin-resistance related up-regulated lncRNA|osteosarcoma doxorubicin-resistance related upregulated lncRNA', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86565145, end_position_on_the_genomic_accession=86567761, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103752588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752588, CurrentID=0, Status='live', Symbol='PACERR', Aliases='PACER, PTGS2-AS1', description='PTGS2 antisense NFKB1 complex-mediated expression regulator RNA', other_designations='PTGS2 antisense RNA 1 (head to head)|p50-associated COX-2 extragenic RNA', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186680654, end_position_on_the_genomic_accession=186681446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103752589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752589, CurrentID=0, Status='live', Symbol='TMEM92-AS1', Aliases='TCONS_00025237, lncRNA-508851', description='TMEM92 antisense RNA 1', other_designations='RP11-893F2.9', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50281579, end_position_on_the_genomic_accession=50287855, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  729998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729998, CurrentID=0, Status='live', Symbol='LOC729998', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133034515, end_position_on_the_genomic_accession=133035940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  729999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729999, CurrentID=0, Status='live', Symbol='LOC729999', Aliases='', description='glycerol-3-phosphate dehydrogenase 2 (mitochondrial) pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56318760, end_position_on_the_genomic_accession=56322465, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372483, CurrentID=0, Status='live', Symbol='LOC105372483', Aliases='', description='uncharacterized LOC105372483', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58361589, end_position_on_the_genomic_accession=58367615, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  205717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205717, CurrentID=0, Status='live', Symbol='USF3', Aliases='KIAA2018', description='upstream transcription factor family member 3', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113648385, end_position_on_the_genomic_accession=113696646, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372484, CurrentID=0, Status='live', Symbol='LOC105372484', Aliases='', description='uncharacterized LOC105372484', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58377586, end_position_on_the_genomic_accession=58379025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730010, CurrentID=0, Status='live', Symbol='BRI3P1', Aliases='', description='brain protein I3 pseudogene 1', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100213204, end_position_on_the_genomic_accession=100213950, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730013, CurrentID=0, Status='live', Symbol='ABCC6P2', Aliases='', description='ATP binding cassette subfamily C member 6 pseudogene 2', other_designations='ATP-binding cassette, sub-family C, member 6 pseudogene 2', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14822432, end_position_on_the_genomic_accession=14824702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372485, CurrentID=0, Status='live', Symbol='LOC105372485', Aliases='', description='uncharacterized LOC105372485', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58499697, end_position_on_the_genomic_accession=58509961, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730016, CurrentID=0, Status='live', Symbol='QRSL1P1', Aliases='', description='glutaminyl-tRNA synthase (glutamine-hydrolyzing)-like 1 pseudogene 1', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168449676, end_position_on_the_genomic_accession=168451316, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373298, CurrentID=0, Status='live', Symbol='LOC105373298', Aliases='', description='uncharacterized LOC105373298', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100605362, end_position_on_the_genomic_accession=100622156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  730020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730020, CurrentID=0, Status='live', Symbol='LOC730020', Aliases='', description='uncharacterized LOC730020', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=23242018, end_position_on_the_genomic_accession=23247664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129606, CurrentID=0, Status='live', Symbol='HNRNPA1P73', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 73', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106479323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479323, CurrentID=0, Status='live', Symbol='RN7SL277P', Aliases='', description='RNA, 7SL, cytoplasmic 277, pseudogene', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730023, CurrentID=0, Status='live', Symbol='KRT8P13', Aliases='', description='keratin 8 pseudogene 13', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170495365, end_position_on_the_genomic_accession=170497096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129607, CurrentID=0, Status='live', Symbol='LOC100129607', Aliases='', description='ubiquitin C-terminal hydrolase L1 pseudogene', other_designations='ubiquitin carboxyl-terminal esterase L1 (ubiquitin thiolesterase) pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56509574, end_position_on_the_genomic_accession=56510196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730029, CurrentID=0, Status='live', Symbol='RPSAP19', Aliases='RPSA_2_81', description='ribosomal protein SA pseudogene 19', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101786266, end_position_on_the_genomic_accession=101787307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645960, CurrentID=0, Status='live', Symbol='LOC645960', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8961747, end_position_on_the_genomic_accession=8962124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372489, CurrentID=0, Status='live', Symbol='LOC105372489', Aliases='', description='uncharacterized LOC105372489', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40443184, end_position_on_the_genomic_accession=40445263, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  730041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730041, CurrentID=0, Status='live', Symbol='RHOQP2', Aliases='', description='ras homolog family member Q pseudogene 2', other_designations='ras homolog gene family, member Q pseudogene 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131460682, end_position_on_the_genomic_accession=131463626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730043, CurrentID=0, Status='live', Symbol='LOC730043', Aliases='', description='uncharacterized LOC730043', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44996412, end_position_on_the_genomic_accession=44996904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372490, CurrentID=0, Status='live', Symbol='LOC105372490', Aliases='', description='sialic acid-binding Ig-like lectin 14', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51559870, end_position_on_the_genomic_accession=51564820, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  730051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730051, CurrentID=0, Status='live', Symbol='ZNF814', Aliases='', description='zinc finger protein 814', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57869379, end_position_on_the_genomic_accession=57889074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730052, CurrentID=0, Status='live', Symbol='UBE2V1P2', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 2', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175717993, end_position_on_the_genomic_accession=175718673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730058, CurrentID=0, Status='live', Symbol='DNM1P49', Aliases='DNM1DN8@', description='dynamin 1 pseudogene 49', other_designations='DNM1 pseudogene 49', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75791517, end_position_on_the_genomic_accession=75792683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730061, CurrentID=0, Status='live', Symbol='RPL34P33', Aliases='RPL34_18_1640', description='ribosomal protein L34 pseudogene 33', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22699352, end_position_on_the_genomic_accession=22699691, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645965, CurrentID=0, Status='live', Symbol='LOC645965', Aliases='', description='adipose differentiation-related protein pseudogene', other_designations='perilipin 2 pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153851622, end_position_on_the_genomic_accession=153853451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730066, CurrentID=0, Status='live', Symbol='VN1R103P', Aliases='', description='vomeronasal 1 receptor 103 pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53503392, end_position_on_the_genomic_accession=53504283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730069, CurrentID=0, Status='live', Symbol='LOC730069', Aliases='', description='nuclear receptor binding factor 2 pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232343029, end_position_on_the_genomic_accession=232344711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730070, CurrentID=0, Status='live', Symbol='RPL26P11', Aliases='RPL26_5_125', description='ribosomal protein L26 pseudogene 11', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173345860, end_position_on_the_genomic_accession=173352010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372495, CurrentID=0, Status='live', Symbol='LOC105372495', Aliases='', description='uncharacterized LOC105372495', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1175046, end_position_on_the_genomic_accession=1179952, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730076, CurrentID=0, Status='live', Symbol='LOC730076', Aliases='', description='zinc finger domain containing pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132085422, end_position_on_the_genomic_accession=132105081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730081, CurrentID=0, Status='live', Symbol='LOC730081', Aliases='', description='uncharacterized LOC730081', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=100433942, end_position_on_the_genomic_accession=100435838, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730086, CurrentID=0, Status='live', Symbol='ALG1L15P', Aliases='', description='asparagine-linked glycosylation 1-like 15, pseudogene', other_designations='asparagine-linked glycosylation 1, beta-1,4-mannosyltransferase homolog pseudogene|asparagine-linked glycosylation 1-like 2 pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156044129, end_position_on_the_genomic_accession=156044950, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730087, CurrentID=0, Status='live', Symbol='ZNF726', Aliases='', description='zinc finger protein 726', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23914882, end_position_on_the_genomic_accession=23933967, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  730091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730091, CurrentID=0, Status='live', Symbol='LINC00886', Aliases='', description='long intergenic non-protein coding RNA 886', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156747343, end_position_on_the_genomic_accession=156817062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730092, CurrentID=0, Status='live', Symbol='RRN3P1', Aliases='', description='RRN3 homolog, RNA polymerase I transcription factor pseudogene 1', other_designations='RNA polymerase I transcription factor homolog (S. cerevisiae) pseudogene 1|RRN3 RNA polymerase I transcription factor homolog pseudogene', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21796630, end_position_on_the_genomic_accession=21819174, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105372498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372498, CurrentID=0, Status='live', Symbol='LOC105372498', Aliases='', description='uncharacterized LOC105372498', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1515294, end_position_on_the_genomic_accession=1521871, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730094, CurrentID=0, Status='live', Symbol='C16orf52', Aliases='', description='chromosome 16 open reading frame 52', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22008135, end_position_on_the_genomic_accession=22084654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730098, CurrentID=0, Status='live', Symbol='LOC730098', Aliases='', description='uncharacterized LOC730098', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34665028, end_position_on_the_genomic_accession=34666050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  730100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730100, CurrentID=0, Status='live', Symbol='LOC730100', Aliases='', description='uncharacterized LOC730100', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51032601, end_position_on_the_genomic_accession=52407917, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  730101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730101, CurrentID=0, Status='live', Symbol='LOC730101', Aliases='', description='uncharacterized LOC730101', other_designations='', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52664401, end_position_on_the_genomic_accession=52669153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730102, CurrentID=0, Status='live', Symbol='LOC730102', Aliases='', description='quinone oxidoreductase-like protein 2 pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178006136, end_position_on_the_genomic_accession=178038007, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105372500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372500, CurrentID=0, Status='live', Symbol='LOC105372500', Aliases='', description='uncharacterized LOC105372500', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730109, CurrentID=0, Status='live', Symbol='LINC01100', Aliases='', description='long intergenic non-protein coding RNA 1100', other_designations='CTD-2049J23.3', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160020649, end_position_on_the_genomic_accession=160031423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730110, CurrentID=0, Status='live', Symbol='LOC730110', Aliases='', description='zinc finger protein 492 pseudogene', other_designations='Zinc finger protein ENSP00000351689', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35145432, end_position_on_the_genomic_accession=35149298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372501, CurrentID=0, Status='live', Symbol='LOC105372501', Aliases='', description='uncharacterized LOC105372501', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730112, CurrentID=0, Status='live', Symbol='FAM166B', Aliases='', description='family with sequence similarity 166 member B', other_designations='family with sequence similarity 166, member B', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35561830, end_position_on_the_genomic_accession=35563899, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129622, CurrentID=0, Status='live', Symbol='RSU1P1', Aliases='Y214H10.6', description='Ras suppressor protein 1 pseudogene 1', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42734707, end_position_on_the_genomic_accession=42738795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102467073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467073, CurrentID=0, Status='live', Symbol='HRAT5', Aliases='', description='heart tissue-associated transcript 5', other_designations='CTD-2083E4.6', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=269556, end_position_on_the_genomic_accession=271524, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372503, CurrentID=0, Status='live', Symbol='LOC105372503', Aliases='', description='uncharacterized LOC105372503', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2296078, end_position_on_the_genomic_accession=2301491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730129, CurrentID=0, Status='live', Symbol='LOC730129', Aliases='', description='neugrin, neurite outgrowth associated pseudogene', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164001173, end_position_on_the_genomic_accession=164002554, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730130, CurrentID=0, Status='live', Symbol='TMEM229A', Aliases='', description='transmembrane protein 229A', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124030916, end_position_on_the_genomic_accession=124033469, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927132, CurrentID=0, Status='live', Symbol='LOC101927132', Aliases='', description='uncharacterized LOC101927132', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47858581, end_position_on_the_genomic_accession=47908431, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  730138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730138, CurrentID=0, Status='live', Symbol='PFN1P1', Aliases='', description='profilin 1 pseudogene 1', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171670277, end_position_on_the_genomic_accession=171671088, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129626, CurrentID=0, Status='live', Symbol='LOC100129626', Aliases='', description='translocase of outer mitochondrial membrane 5 homolog (yeast) pseudogene', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3693595, end_position_on_the_genomic_accession=3694108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128950, CurrentID=0, Status='live', Symbol='LOC100128950', Aliases='', description='uncharacterized LOC100128950', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=198177584, end_position_on_the_genomic_accession=198177982, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  730144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730144, CurrentID=0, Status='live', Symbol='EIF1P3', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 3', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182335858, end_position_on_the_genomic_accession=182336514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645979, CurrentID=0, Status='live', Symbol='RPS26P31', Aliases='RPS26_8_840', description='ribosomal protein S26 pseudogene 31', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122681293, end_position_on_the_genomic_accession=122681726, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  205860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=205860, CurrentID=0, Status='live', Symbol='TRIML2', Aliases='SPRYD6', description='tripartite motif family like 2', other_designations='SPRY domain-containing protein 6|tripartite motif family-like 2|tripartite motif family-like protein 2', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188091255, end_position_on_the_genomic_accession=188109603, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  730152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730152, CurrentID=0, Status='live', Symbol='GSTA7P', Aliases='GSTAP2, GSTAP6', description='glutathione S-transferase alpha 7, pseudogene', other_designations='glutathione S-transferase alpha pseudogene 2|glutathione-S-transferase A pseudogene 2', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52739463, end_position_on_the_genomic_accession=52745159, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  730153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730153, CurrentID=0, Status='live', Symbol='NPIP', Aliases='', description='nuclear pore complex interacting protein family, member A1 pseudogene', other_designations='nuclear pore complex interacting protein pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29037834, end_position_on_the_genomic_accession=29052307, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129629, CurrentID=0, Status='live', Symbol='EIF2S2P6', Aliases='', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 6', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 6', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82828271, end_position_on_the_genomic_accession=82837242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372511, CurrentID=0, Status='live', Symbol='LOC105372511', Aliases='', description='uncharacterized LOC105372511', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4777723, end_position_on_the_genomic_accession=4779349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730174, CurrentID=0, Status='live', Symbol='NAP1L4P3', Aliases='', description='nucleosome assembly protein 1-like 4 pseudogene 3', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=47763559, end_position_on_the_genomic_accession=47764572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271208, CurrentID=0, Status='live', Symbol='RPL6P28', Aliases='RPL6_16_1757', description='ribosomal protein L6 pseudogene 28', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44901571, end_position_on_the_genomic_accession=44902450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730177, CurrentID=0, Status='live', Symbol='RPS26P16', Aliases='RPS26_3_133', description='ribosomal protein S26 pseudogene 16', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179035231, end_position_on_the_genomic_accession=179035678, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730179, CurrentID=0, Status='live', Symbol='LOC730179', Aliases='', description='uncharacterized LOC730179', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730180, CurrentID=0, Status='live', Symbol='RBBP4P3', Aliases='', description='retinoblastoma binding protein 4 pseudogene 3', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=60873363, end_position_on_the_genomic_accession=60874643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374172, CurrentID=0, Status='live', Symbol='LOC105374172', Aliases='', description='uncharacterized LOC105374172', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155315383, end_position_on_the_genomic_accession=155318933, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730183, CurrentID=0, Status='live', Symbol='LOC730183', Aliases='', description='uncharacterized LOC730183', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30697704, end_position_on_the_genomic_accession=30698489, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730187, CurrentID=0, Status='live', Symbol='RPS10P9', Aliases='RPS10_3_235', description='ribosomal protein S10 pseudogene 9', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64665571, end_position_on_the_genomic_accession=64666123, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372514, CurrentID=0, Status='live', Symbol='LOC105372514', Aliases='', description='uncharacterized LOC105372514', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5703632, end_position_on_the_genomic_accession=5708551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730190, CurrentID=0, Status='live', Symbol='ZNF101P2', Aliases='', description='zinc finger protein 101 pseudogene 2', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192993423, end_position_on_the_genomic_accession=192994806, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730196, CurrentID=0, Status='live', Symbol='VN1R64P', Aliases='', description='vomeronasal 1 receptor 64 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31546805, end_position_on_the_genomic_accession=31547737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730198, CurrentID=0, Status='live', Symbol='MEIS1-AS3', Aliases='', description='MEIS1 antisense RNA 3', other_designations='MEIS1 antisense RNA 3 (non-protein coding)', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66423342, end_position_on_the_genomic_accession=66433470, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372516, CurrentID=0, Status='live', Symbol='LOC105372516', Aliases='', description='uncharacterized LOC105372516', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6730791, end_position_on_the_genomic_accession=6736277, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  730202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730202, CurrentID=0, Status='live', Symbol='LOC730202', Aliases='', description='uncharacterized LOC730202', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96501648, end_position_on_the_genomic_accession=96502326, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  730211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730211, CurrentID=0, Status='live', Symbol='HSP90AA5P', Aliases='HSP90Ae', description='heat shock protein 90kDa alpha family class A member 5, pseudogene', other_designations='heat shock protein 90Ae|heat shock protein 90kDa alpha (cytosolic), class A member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class A member 5, pseudogene|heat shock protein HSP 90-alpha (HSP 86)', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184115842, end_position_on_the_genomic_accession=184118003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730213, CurrentID=0, Status='live', Symbol='RPL5P5', Aliases='RPL5_1_139', description='ribosomal protein L5 pseudogene 5', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185226747, end_position_on_the_genomic_accession=185227772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  56167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56167, CurrentID=0, Status='live', Symbol='MTCO1P1', Aliases='', description='MT-CO1 pseudogene 1', other_designations='MTCO1 pseudogene 1|cytochrome c oxidase pseudogene 1', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8845707, end_position_on_the_genomic_accession=8846548, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730222, CurrentID=0, Status='live', Symbol='LOC730222', Aliases='', description='tubulin, alpha 3d pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202853012, end_position_on_the_genomic_accession=202859099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372520, CurrentID=0, Status='live', Symbol='LOC105372520', Aliases='', description='uncharacterized LOC105372520', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8018552, end_position_on_the_genomic_accession=8028623, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  730227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730227, CurrentID=0, Status='live', Symbol='LINC01136', Aliases='', description='long intergenic non-protein coding RNA 1136', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203298758, end_position_on_the_genomic_accession=203305325, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129641, CurrentID=0, Status='live', Symbol='BTG1P1', Aliases='', description='B-cell translocation gene 1 pseudogene 1', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9134922, end_position_on_the_genomic_accession=9135597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730232, CurrentID=0, Status='live', Symbol='EEF1A1P32', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 32', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197688557, end_position_on_the_genomic_accession=197690648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  648442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648442, CurrentID=0, Status='live', Symbol='LOC648442', Aliases='', description='eukaryotic translation initiation factor 3 subunit I pseudogene', other_designations='eukaryotic translation initiation factor 3, subunit I pseudogene', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118950388, end_position_on_the_genomic_accession=118951267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730234, CurrentID=0, Status='live', Symbol='LOC730234', Aliases='', description='uncharacterized LOC730234', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45806944, end_position_on_the_genomic_accession=45815389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730236, CurrentID=0, Status='live', Symbol='LOC730236', Aliases='', description='uncharacterized LOC730236', other_designations='', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=67804715, end_position_on_the_genomic_accession=67805815, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370742, CurrentID=0, Status='live', Symbol='LOC105370742', Aliases='', description='uncharacterized LOC105370742', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27039078, end_position_on_the_genomic_accession=27072198, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  730239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730239, CurrentID=0, Status='live', Symbol='ZDHHC20P4', Aliases='', description='zinc finger, DHHC-type containing 20 pseudogene 4', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68984943, end_position_on_the_genomic_accession=68986326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372523, CurrentID=0, Status='live', Symbol='LOC105372523', Aliases='', description='uncharacterized LOC105372523', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9563015, end_position_on_the_genomic_accession=9616108, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  730246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730246, CurrentID=0, Status='live', Symbol='HNRNPA1P59', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 59', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202911751, end_position_on_the_genomic_accession=202912928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730249, CurrentID=0, Status='live', Symbol='IRG1', Aliases='CAD', description='immunoresponsive 1 homolog (mouse)', other_designations='aconitate decarboxylase|cis-aconitic acid decarboxylase|immune-responsive gene 1 protein homolog', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76948559, end_position_on_the_genomic_accession=76958659, orientation='plus', exon_count=7, OMIM=615275, no_hdr0=''),
  730255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730255, CurrentID=0, Status='live', Symbol='RPL17P8', Aliases='RPL17_4_153', description='ribosomal protein L17 pseudogene 8', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205202157, end_position_on_the_genomic_accession=205202771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730257, CurrentID=0, Status='live', Symbol='LOC730257', Aliases='', description='uncharacterized LOC730257', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149176141, end_position_on_the_genomic_accession=149177139, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129646, CurrentID=0, Status='live', Symbol='LOC100129646', Aliases='', description='intraflagellar transport 57 homolog (Chlamydomonas) pseudogene', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28164958, end_position_on_the_genomic_accession=28165662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730262, CurrentID=0, Status='live', Symbol='PPIAL4E', Aliases='COAS2', description='peptidylprolyl isomerase A (cyclophilin A)-like 4E', other_designations='PPIase A-like 4E', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144372875, end_position_on_the_genomic_accession=144373659, orientation='minus', exon_count=1, OMIM=608608, no_hdr0=''),
  105372527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372527, CurrentID=0, Status='live', Symbol='LOC105372527', Aliases='', description='uncharacterized LOC105372527', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11157485, end_position_on_the_genomic_accession=11158782, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730268, CurrentID=0, Status='live', Symbol='LOC730268', Aliases='', description='anaphase-promoting complex subunit 1-like', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87700445, end_position_on_the_genomic_accession=87747727, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100129648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129648, CurrentID=0, Status='live', Symbol='LOC100129648', Aliases='', description='uncharacterized LOC100129648', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=40098598, end_position_on_the_genomic_accession=40099173, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372529, CurrentID=0, Status='live', Symbol='LOC105372529', Aliases='', description='uncharacterized LOC105372529', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11309808, end_position_on_the_genomic_accession=11427287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372530, CurrentID=0, Status='live', Symbol='LOC105372530', Aliases='', description='uncharacterized LOC105372530', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11713041, end_position_on_the_genomic_accession=11728808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  730291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730291, CurrentID=0, Status='live', Symbol='ZNF735', Aliases='ZNF735P', description='zinc finger protein 735', other_designations='zinc finger protein 735 pseudogene|zinc finger protein 735, pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64207203, end_position_on_the_genomic_accession=64220290, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372427, CurrentID=0, Status='live', Symbol='LOC105372427', Aliases='', description='uncharacterized LOC105372427', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47280119, end_position_on_the_genomic_accession=47292076, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372534, CurrentID=0, Status='live', Symbol='LOC105372534', Aliases='', description='uncharacterized LOC105372534', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12374136, end_position_on_the_genomic_accession=12393694, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271209, CurrentID=0, Status='live', Symbol='RPL7AP28', Aliases='RPL7A_10_519', description='ribosomal protein L7a pseudogene 28', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131857340, end_position_on_the_genomic_accession=131857907, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372537, CurrentID=0, Status='live', Symbol='LOC105372537', Aliases='', description='uncharacterized LOC105372537', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13321162, end_position_on_the_genomic_accession=13325145, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730338, CurrentID=0, Status='live', Symbol='LOC730338', Aliases='', description='uncharacterized LOC730338', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46687879, end_position_on_the_genomic_accession=46697122, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372539, CurrentID=0, Status='live', Symbol='LOC105372539', Aliases='', description='uncharacterized LOC105372539', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13239080, end_position_on_the_genomic_accession=13245243, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375110, CurrentID=0, Status='live', Symbol='LOC105375110', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35684467, end_position_on_the_genomic_accession=35689106, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646012, CurrentID=0, Status='live', Symbol='LOC646012', Aliases='', description='aurora kinase A interacting protein 1 pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17306364, end_position_on_the_genomic_accession=17307033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129661, CurrentID=0, Status='live', Symbol='LOC100129661', Aliases='', description='HCLS1 associated protein X-1 pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=148197828, end_position_on_the_genomic_accession=148198979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372542, CurrentID=0, Status='live', Symbol='LOC105372542', Aliases='', description='uncharacterized LOC105372542', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16610526, end_position_on_the_genomic_accession=16640181, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372543, CurrentID=0, Status='live', Symbol='LOC105372543', Aliases='', description='uncharacterized LOC105372543', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16965438, end_position_on_the_genomic_accession=16968001, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646016, CurrentID=0, Status='live', Symbol='YWHAZP8', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 8', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73274785, end_position_on_the_genomic_accession=73275516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927140, CurrentID=0, Status='live', Symbol='LOC101927140', Aliases='', description='probable ribosome biogenesis protein RLP24 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101561777, end_position_on_the_genomic_accession=101562488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  730376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730376, CurrentID=0, Status='live', Symbol='LOC730376', Aliases='', description='uncharacterized LOC730376', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=55808486, end_position_on_the_genomic_accession=55809263, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  730378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730378, CurrentID=0, Status='live', Symbol='LOC730378', Aliases='', description='uncharacterized LOC730378', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=51459730, end_position_on_the_genomic_accession=51460937, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372547, CurrentID=0, Status='live', Symbol='LOC105372547', Aliases='', description='uncharacterized LOC105372547', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17697616, end_position_on_the_genomic_accession=17699583, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730394, CurrentID=0, Status='live', Symbol='GTF2H2C_2', Aliases='GTF2H2D', description='GTF2H2 family member C, copy 2', other_designations='general transcription factor IIH polypeptide 2-like protein|general transcription factor IIH, polypeptide 2D', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372549, CurrentID=0, Status='live', Symbol='LOC105372549', Aliases='', description='uncharacterized LOC105372549', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17912579, end_position_on_the_genomic_accession=17925217, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927141, CurrentID=0, Status='live', Symbol='LOC101927141', Aliases='', description='uncharacterized LOC101927141', other_designations='CTD-2272D18.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=82912104, end_position_on_the_genomic_accession=82913023, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372550, CurrentID=0, Status='live', Symbol='LOC105372550', Aliases='', description='uncharacterized LOC105372550', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18312988, end_position_on_the_genomic_accession=18359309, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378538, CurrentID=0, Status='live', Symbol='LOC105378538', Aliases='', description='uncharacterized LOC105378538', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124623611, end_position_on_the_genomic_accession=124629802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372551, CurrentID=0, Status='live', Symbol='LOC105372551', Aliases='', description='uncharacterized LOC105372551', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18270624, end_position_on_the_genomic_accession=18271104, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372552, CurrentID=0, Status='live', Symbol='LOC105372552', Aliases='', description='uncharacterized LOC105372552', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18579501, end_position_on_the_genomic_accession=18587830, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130673, CurrentID=0, Status='live', Symbol='LOC100130673', Aliases='', description='phosphoribosyl pyrophosphate synthetase 2 pseudogene', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32456897, end_position_on_the_genomic_accession=32458494, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129675, CurrentID=0, Status='live', Symbol='LOC100129675', Aliases='', description='uncharacterized LOC100129675', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=242163674, end_position_on_the_genomic_accession=242164455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372557, CurrentID=0, Status='live', Symbol='LOC105372557', Aliases='', description='uncharacterized LOC105372557', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21145662, end_position_on_the_genomic_accession=21162880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646030, CurrentID=0, Status='live', Symbol='LOC646030', Aliases='', description='leucine-rich repeat-containing protein 37A2-like', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31038574, end_position_on_the_genomic_accession=31051037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373301, CurrentID=0, Status='live', Symbol='LOC105373301', Aliases='', description='uncharacterized LOC105373301', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103609677, end_position_on_the_genomic_accession=103615966, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271210, CurrentID=0, Status='live', Symbol='RPL7AP35', Aliases='RPL7A_10_727', description='ribosomal protein L7a pseudogene 35', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105298141, end_position_on_the_genomic_accession=105298946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129682, CurrentID=0, Status='live', Symbol='LOC100129682', Aliases='', description='ribosomal protein L17 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8346480, end_position_on_the_genomic_accession=8347331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372722, CurrentID=0, Status='live', Symbol='LOC105372722', Aliases='', description='uncharacterized LOC105372722', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63419679, end_position_on_the_genomic_accession=63424047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372565, CurrentID=0, Status='live', Symbol='LOC105372565', Aliases='', description='uncharacterized LOC105372565', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22596076, end_position_on_the_genomic_accession=22618421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129687, CurrentID=0, Status='live', Symbol='LOC100129687', Aliases='', description='C-x(9)-C motif containing 2 pseudogene', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28830966, end_position_on_the_genomic_accession=28831182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372568, CurrentID=0, Status='live', Symbol='LOC105372568', Aliases='', description='uncharacterized LOC105372568', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22805757, end_position_on_the_genomic_accession=22824384, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372570, CurrentID=0, Status='live', Symbol='LOC105372570', Aliases='', description='uncharacterized LOC105372570', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23346627, end_position_on_the_genomic_accession=23348876, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372572, CurrentID=0, Status='live', Symbol='LOC105372572', Aliases='', description='uncharacterized LOC105372572', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23309456, end_position_on_the_genomic_accession=23351410, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421386, CurrentID=0, Status='live', Symbol='LOC100421386', Aliases='', description='glutamyl aminopeptidase (aminopeptidase A) pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63209222, end_position_on_the_genomic_accession=63211816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129693, CurrentID=0, Status='live', Symbol='LOC100129693', Aliases='', description='mitochondrial ribosomal protein L19 pseudogene', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87410729, end_position_on_the_genomic_accession=87411385, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372574, CurrentID=0, Status='live', Symbol='LOC105372574', Aliases='', description='uncharacterized LOC105372574', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23722547, end_position_on_the_genomic_accession=23729631, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372576, CurrentID=0, Status='live', Symbol='LOC105372576', Aliases='', description='uncharacterized LOC105372576', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24092278, end_position_on_the_genomic_accession=24112234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928639, CurrentID=0, Status='live', Symbol='LOC101928639', Aliases='', description='uncharacterized LOC101928639', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219851639, end_position_on_the_genomic_accession=219872916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128964, CurrentID=0, Status='live', Symbol='LOC100128964', Aliases='', description='serine/arginine-rich splicing factor 3 pseudogene', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107920607, end_position_on_the_genomic_accession=107921291, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372577, CurrentID=0, Status='live', Symbol='LOC105372577', Aliases='', description='uncharacterized LOC105372577', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24309217, end_position_on_the_genomic_accession=24318114, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100421387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421387, CurrentID=0, Status='live', Symbol='LOC100421387', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79124739, end_position_on_the_genomic_accession=79125631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130674, CurrentID=0, Status='live', Symbol='LOC100130674', Aliases='', description='cathepsin B pseudogene', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=30006947, end_position_on_the_genomic_accession=30008319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646050, CurrentID=0, Status='live', Symbol='LOC646050', Aliases='', description='axin interactor, dorsalization associated pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11308025, end_position_on_the_genomic_accession=11308938, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927147, CurrentID=0, Status='live', Symbol='LOC101927147', Aliases='', description='uncharacterized LOC101927147', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239879393, end_position_on_the_genomic_accession=239913187, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372580, CurrentID=0, Status='live', Symbol='LOC105372580', Aliases='', description='uncharacterized LOC105372580', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25238999, end_position_on_the_genomic_accession=25247970, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129703, CurrentID=0, Status='live', Symbol='NAG20', Aliases='', description='NAG20', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=137352554, end_position_on_the_genomic_accession=137354168, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372584, CurrentID=0, Status='live', Symbol='LOC105372584', Aliases='', description='uncharacterized LOC105372584', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372585, CurrentID=0, Status='live', Symbol='LOC105372585', Aliases='', description='formin-like protein 5', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24468226, end_position_on_the_genomic_accession=24475800, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129706, CurrentID=0, Status='live', Symbol='HMGB1P13', Aliases='HMGB1L13', description='high mobility group box 1 pseudogene 13', other_designations='high-mobility group box 1 pseudogene 13|high-mobility group box 1-like 13|high-mobility group box 2 pseudogene', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132868230, end_position_on_the_genomic_accession=132868859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271211, CurrentID=0, Status='live', Symbol='RPL7AP47', Aliases='RPL7A_13_964', description='ribosomal protein L7a pseudogene 47', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14204218, end_position_on_the_genomic_accession=14205094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  206338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=206338, CurrentID=0, Status='live', Symbol='LVRN', Aliases='APQ, AQPEP, TAQPEP', description='laeverin', other_designations='AP-Q|CHL2 antigen', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115954903, end_position_on_the_genomic_accession=116027606, orientation='plus', exon_count=26, OMIM=610046, no_hdr0=''),
  105372587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372587, CurrentID=0, Status='live', Symbol='LOC105372587', Aliases='', description='uncharacterized LOC105372587', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30350382, end_position_on_the_genomic_accession=30362209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372588, CurrentID=0, Status='live', Symbol='LOC105372588', Aliases='', description='uncharacterized LOC105372588', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31604691, end_position_on_the_genomic_accession=31605190, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  89797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89797, CurrentID=0, Status='live', Symbol='NAV2', Aliases='HELAD1, POMFIL2, RAINB1, STEERIN2, UNC53H2', description='neuron navigator 2', other_designations='helicase, APC down-regulated 1|pore membrane and/or filament-interacting-like protein 2|retinoic acid inducible gene in neuroblastoma 1|steerin-2|unc-53 homolog 2', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19345247, end_position_on_the_genomic_accession=20121601, orientation='plus', exon_count=49, OMIM=607026, no_hdr0=''),
  206358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=206358, CurrentID=0, Status='live', Symbol='SLC36A1', Aliases='Dct1, LYAAT1, PAT1, TRAMD3', description='solute carrier family 36 (proton/amino acid symporter), member 1', other_designations='lysosomal amino acid transporter 1|proton/amino acid transporter 1', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151411344, end_position_on_the_genomic_accession=151561961, orientation='plus', exon_count=27, OMIM=606561, no_hdr0=''),
  730647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730647, CurrentID=0, Status='live', Symbol='HINT1P1', Aliases='HINTP1', description='histidine triad nucleotide binding protein 1 pseudogene 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143312530, end_position_on_the_genomic_accession=143313123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372591, CurrentID=0, Status='live', Symbol='LOC105372591', Aliases='', description='uncharacterized LOC105372591', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32561040, end_position_on_the_genomic_accession=32574387, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372593, CurrentID=0, Status='live', Symbol='LOC105372593', Aliases='', description='uncharacterized LOC105372593', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33242536, end_position_on_the_genomic_accession=33257496, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730668, CurrentID=0, Status='live', Symbol='LOC730668', Aliases='', description='dynein heavy chain -like pseudogene', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46006616, end_position_on_the_genomic_accession=46010777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372594, CurrentID=0, Status='live', Symbol='LOC105372594', Aliases='', description='uncharacterized LOC105372594', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33298756, end_position_on_the_genomic_accession=33299853, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927150, CurrentID=0, Status='live', Symbol='L3MBTL4-AS1', Aliases='', description='L3MBTL4 antisense RNA 1', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6256744, end_position_on_the_genomic_accession=6260934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372596, CurrentID=0, Status='live', Symbol='LOC105372596', Aliases='', description='uncharacterized LOC105372596', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34014737, end_position_on_the_genomic_accession=34017502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129717, CurrentID=0, Status='live', Symbol='LOC100129717', Aliases='', description='uncharacterized LOC100129717', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646070, CurrentID=0, Status='live', Symbol='CICP18', Aliases='', description='capicua transcriptional repressor pseudogene 18', other_designations='capicua homolog pseudogene', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142230, end_position_on_the_genomic_accession=145994, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646071, CurrentID=0, Status='live', Symbol='LOC646071', Aliases='', description='disintegrin and metalloproteinase domain-containing protein 21-like', other_designations='', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20046294, end_position_on_the_genomic_accession=20048297, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  206412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=206412, CurrentID=0, Status='live', Symbol='C6orf163', Aliases='', description='chromosome 6 open reading frame 163', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87343080, end_position_on_the_genomic_accession=87365553, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129721, CurrentID=0, Status='live', Symbol='LOC100129721', Aliases='', description='uncharacterized LOC100129721', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=61138544, end_position_on_the_genomic_accession=61139334, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  206426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=206426, CurrentID=0, Status='live', Symbol='PIP5K1P1', Aliases='', description='phosphatidylinositol-4-phosphate 5-kinase, type I, pseudogene 1', other_designations='PIP5K1A pseudogene', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7986102, end_position_on_the_genomic_accession=7990344, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129722, CurrentID=0, Status='live', Symbol='C9orf173-AS1', Aliases='', description='C9orf173 antisense RNA 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137250219, end_position_on_the_genomic_accession=137253497, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372603, CurrentID=0, Status='live', Symbol='LOC105372603', Aliases='', description='uncharacterized LOC105372603', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36174995, end_position_on_the_genomic_accession=36197852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129724, CurrentID=0, Status='live', Symbol='LOC100129724', Aliases='', description='uncharacterized LOC100129724', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26979572, end_position_on_the_genomic_accession=27014739, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100129725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129725, CurrentID=0, Status='live', Symbol='LOC100129725', Aliases='', description='PERP, TP53 apoptosis effector pseudogene', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191951111, end_position_on_the_genomic_accession=191951747, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372606, CurrentID=0, Status='live', Symbol='LOC105372606', Aliases='', description='uncharacterized LOC105372606', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37337440, end_position_on_the_genomic_accession=37344570, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  730747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730747, CurrentID=0, Status='live', Symbol='RPL14P3', Aliases='RPL14_2_522', description='ribosomal protein L14 pseudogene 3', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140366202, end_position_on_the_genomic_accession=140366892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372608, CurrentID=0, Status='live', Symbol='LOC105372608', Aliases='', description='uncharacterized LOC105372608', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37571118, end_position_on_the_genomic_accession=37583924, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  730755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730755, CurrentID=0, Status='live', Symbol='KRTAP2-3', Aliases='KAP2.3, KRTAP2.3', description='keratin associated protein 2-3', other_designations='high sulfur keratin-associated protein 2.4|keratin-associated protein 2.3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41059240, end_position_on_the_genomic_accession=41060092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372609, CurrentID=0, Status='live', Symbol='LOC105372609', Aliases='', description='uncharacterized LOC105372609', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37686042, end_position_on_the_genomic_accession=37691276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372610, CurrentID=0, Status='live', Symbol='LOC105372610', Aliases='', description='uncharacterized LOC105372610', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37751146, end_position_on_the_genomic_accession=37757532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129731, CurrentID=0, Status='live', Symbol='RAB11AP1', Aliases='', description='RAB11A, member RAS oncogene family pseudogene 1', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93881289, end_position_on_the_genomic_accession=93882017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271212, CurrentID=0, Status='live', Symbol='RPL7AP48', Aliases='RPL7A_14_968', description='ribosomal protein L7a pseudogene 48', other_designations='', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20532526, end_position_on_the_genomic_accession=20533398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372612, CurrentID=0, Status='live', Symbol='LOC105372612', Aliases='', description='uncharacterized LOC105372612', other_designations='CTD-2308N23.2', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38288600, end_position_on_the_genomic_accession=38317559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372613, CurrentID=0, Status='live', Symbol='LOC105372613', Aliases='', description='uncharacterized LOC105372613', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39785314, end_position_on_the_genomic_accession=39788283, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372614, CurrentID=0, Status='live', Symbol='LOC105372614', Aliases='', description='uncharacterized LOC105372614', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39757263, end_position_on_the_genomic_accession=39813637, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  730792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730792, CurrentID=0, Status='live', Symbol='SFR1P2', Aliases='', description='SFR1 pseudogene 2', other_designations='MEI5 meiotic recombination protein homolog pseudogene|SWI5-dependent recombination repair 1 pseudogene', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85849412, end_position_on_the_genomic_accession=85850139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927154, CurrentID=0, Status='live', Symbol='LOC101927154', Aliases='', description='uncharacterized LOC101927154', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372615, CurrentID=0, Status='live', Symbol='LOC105372615', Aliases='', description='uncharacterized LOC105372615', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40022595, end_position_on_the_genomic_accession=40044813, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372616, CurrentID=0, Status='live', Symbol='LOC105372616', Aliases='', description='uncharacterized LOC105372616', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40116426, end_position_on_the_genomic_accession=40121077, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479349, CurrentID=0, Status='live', Symbol='RN7SL342P', Aliases='', description='RNA, 7SL, cytoplasmic 342, pseudogene', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  449481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449481, CurrentID=0, Status='live', Symbol='DFNB43', Aliases='', description='deafness, autosomal recessive 43', other_designations='', map_location='15q24.1-q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730811, CurrentID=0, Status='live', Symbol='MYT1L-AS1', Aliases='', description='MYT1L antisense RNA 1', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=2319232, end_position_on_the_genomic_accession=2327108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129738, CurrentID=0, Status='live', Symbol='SETP9', Aliases='', description='SET pseudogene 9', other_designations='SET nuclear oncogene pseudogene', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160670151, end_position_on_the_genomic_accession=160670981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  449483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449483, CurrentID=0, Status='live', Symbol='DFNB45', Aliases='', description='deafness, autosomal recessive 45', other_designations='', map_location='1q43-q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612433, no_hdr0=''),
  105372621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372621, CurrentID=0, Status='live', Symbol='LOC105372621', Aliases='', description='uncharacterized LOC105372621', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40995614, end_position_on_the_genomic_accession=41005881, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372622, CurrentID=0, Status='live', Symbol='LOC105372622', Aliases='', description='uncharacterized LOC105372622', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42014562, end_position_on_the_genomic_accession=42024246, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372623, CurrentID=0, Status='live', Symbol='LOC105372623', Aliases='', description='uncharacterized LOC105372623', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42949325, end_position_on_the_genomic_accession=42955794, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372624, CurrentID=0, Status='live', Symbol='LOC105372624', Aliases='', description='uncharacterized LOC105372624', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42630824, end_position_on_the_genomic_accession=42648771, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375127, CurrentID=0, Status='live', Symbol='LOC105375127', Aliases='', description='uncharacterized LOC105375127', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2113564, end_position_on_the_genomic_accession=2117740, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730861, CurrentID=0, Status='live', Symbol='RPS3AP29', Aliases='RPS3A_10_823', description='ribosomal protein S3a pseudogene 29', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97898296, end_position_on_the_genomic_accession=97899144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  449491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449491, CurrentID=0, Status='live', Symbol='DEFA8P', Aliases='DEFAP1', description='defensin alpha 8, pseudogene', other_designations='defensin, alpha 8 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6950726, end_position_on_the_genomic_accession=6951599, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130676, CurrentID=0, Status='live', Symbol='RPSAP38', Aliases='RPSA_9_557', description='ribosomal protein SA pseudogene 38', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41951240, end_position_on_the_genomic_accession=41952086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372628, CurrentID=0, Status='live', Symbol='LOC105372628', Aliases='', description='uncharacterized LOC105372628', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44211499, end_position_on_the_genomic_accession=44220256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  402036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402036, CurrentID=0, Status='live', Symbol='LOC402036', Aliases='', description='hCG1646661-like', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=19019041, end_position_on_the_genomic_accession=19020562, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927157, CurrentID=0, Status='live', Symbol='LOC101927157', Aliases='', description='uncharacterized LOC101927157', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47914227, end_position_on_the_genomic_accession=47990737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  730888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730888, CurrentID=0, Status='live', Symbol='DEPDC1P1', Aliases='', description='DEP domain containing 1 pseudogene 1', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31764101, end_position_on_the_genomic_accession=31768339, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371558, CurrentID=0, Status='live', Symbol='LOC105371558', Aliases='', description='cytospin-B-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15817863, end_position_on_the_genomic_accession=15930886, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  646103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646103, CurrentID=0, Status='live', Symbol='LOC646103', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17518267, end_position_on_the_genomic_accession=17519063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372632, CurrentID=0, Status='live', Symbol='LOC105372632', Aliases='', description='uncharacterized LOC105372632', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46022984, end_position_on_the_genomic_accession=46023581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  730909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730909, CurrentID=0, Status='live', Symbol='LOC730909', Aliases='', description='uncharacterized LOC730909', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=28909730, end_position_on_the_genomic_accession=28917477, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100129755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129755, CurrentID=0, Status='live', Symbol='RSL24D1P1', Aliases='RPL24p, RPL30BP, RSL24D1P, dJ97D16.2', description='ribosomal L24 domain containing 1 pseudogene 1', other_designations='ribosomal protein L30B pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27780683, end_position_on_the_genomic_accession=27781111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421294, CurrentID=0, Status='live', Symbol='LOC100421294', Aliases='', description='phosphoribosyl pyrophosphate synthetase 2 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102922572, end_position_on_the_genomic_accession=102923238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928847, CurrentID=0, Status='live', Symbol='LOC101928847', Aliases='', description='uncharacterized LOC101928847', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112959279, end_position_on_the_genomic_accession=112963460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372636, CurrentID=0, Status='live', Symbol='LOC105372636', Aliases='', description='uncharacterized LOC105372636', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47840984, end_position_on_the_genomic_accession=47843284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271213, CurrentID=0, Status='live', Symbol='RPL7AP41', Aliases='RPL7A_16_795', description='ribosomal protein L7a pseudogene 41', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25212164, end_position_on_the_genomic_accession=25212985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129840, CurrentID=0, Status='live', Symbol='LOC100129840', Aliases='', description='uncharacterized LOC100129840', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=66652549, end_position_on_the_genomic_accession=66653166, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  103753527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103753527, CurrentID=0, Status='live', Symbol='SCA37', Aliases='', description='spinocerebellar ataxia 37', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615945, no_hdr0=''),
  103753528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103753528, CurrentID=0, Status='live', Symbol='DUPXP11.22', Aliases='MRX17, MRX31', description='Xp11.22 microduplication syndrome', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300705, no_hdr0=''),
  101927159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927159, CurrentID=0, Status='live', Symbol='LOC101927159', Aliases='', description='uncharacterized LOC101927159', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42685483, end_position_on_the_genomic_accession=42688562, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646112, CurrentID=0, Status='live', Symbol='LOC646112', Aliases='', description='origin recognition complex subunit 3 pseudogene', other_designations='origin recognition complex, subunit 3-like (yeast) pseudogene', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=79987471, end_position_on_the_genomic_accession=79989817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129761, CurrentID=0, Status='live', Symbol='AMD1P4', Aliases='', description='adenosylmethionine decarboxylase 1 pseudogene 4', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11861626, end_position_on_the_genomic_accession=11862698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930593, CurrentID=0, Status='live', Symbol='LOC101930593', Aliases='', description='uncharacterized LOC101930593', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55160024, end_position_on_the_genomic_accession=55166322, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372642, CurrentID=0, Status='live', Symbol='LOC105372642', Aliases='', description='uncharacterized LOC105372642', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48216271, end_position_on_the_genomic_accession=48231631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  730961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730961, CurrentID=0, Status='live', Symbol='LOC730961', Aliases='', description='uncharacterized LOC730961', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129763, CurrentID=0, Status='live', Symbol='LOC100129763', Aliases='', description='uncharacterized LOC100129763', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=8294512, end_position_on_the_genomic_accession=8295185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  730971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730971, CurrentID=0, Status='live', Symbol='FLJ36777', Aliases='', description='uncharacterized LOC730971', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7097424, end_position_on_the_genomic_accession=7103376, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927160, CurrentID=0, Status='live', Symbol='LOC101927160', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76071322, end_position_on_the_genomic_accession=76072392, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  730974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730974, CurrentID=0, Status='live', Symbol='PRELID3BP7', Aliases='', description='PRELI domain containing 3B pseudogene 7', other_designations='slowmo homolog 2 (Drosophila) pseudogene', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42972624, end_position_on_the_genomic_accession=42973179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  730978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730978, CurrentID=0, Status='live', Symbol='CICP5', Aliases='', description='capicua transcriptional repressor pseudogene 5', other_designations='capicua homolog pseudogene 5', map_location='1q42.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223947558, end_position_on_the_genomic_accession=223949393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372646, CurrentID=0, Status='live', Symbol='LOC105372646', Aliases='', description='uncharacterized LOC105372646', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48526380, end_position_on_the_genomic_accession=48563403, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372647, CurrentID=0, Status='live', Symbol='LOC105372647', Aliases='', description='uncharacterized LOC105372647', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  730993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=730993, CurrentID=0, Status='live', Symbol='RPS3AP39', Aliases='RPS3A_18_1114', description='ribosomal protein S3a pseudogene 39', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3549695, end_position_on_the_genomic_accession=3550543, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372649, CurrentID=0, Status='live', Symbol='LOC105372649', Aliases='', description='uncharacterized LOC105372649', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49331413, end_position_on_the_genomic_accession=49340733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130620, CurrentID=0, Status='live', Symbol='LOC100130620', Aliases='', description='tropomyosin 1 (alpha) pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138806374, end_position_on_the_genomic_accession=138807130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422225, CurrentID=0, Status='live', Symbol='LOC100422225', Aliases='', description='ATP binding cassette subfamily G member 2 (Junior blood group) pseudogene', other_designations='ATP-binding cassette, sub-family G (WHITE), member 2 (Junior blood group) pseudogene|ATP-binding cassette, sub-family G (WHITE), member 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77359416, end_position_on_the_genomic_accession=77359999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479356, CurrentID=0, Status='live', Symbol='RN7SL364P', Aliases='', description='RNA, 7SL, cytoplasmic 364, pseudogene', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372652, CurrentID=0, Status='live', Symbol='LOC105372652', Aliases='', description='uncharacterized LOC105372652', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49764899, end_position_on_the_genomic_accession=49775396, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421402, CurrentID=0, Status='live', Symbol='LOC100421402', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113449706, end_position_on_the_genomic_accession=113450685, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724931, CurrentID=0, Status='live', Symbol='LOC102724931', Aliases='', description='C-terminal-binding protein 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7128654, end_position_on_the_genomic_accession=7130923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  449518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=449518, CurrentID=0, Status='live', Symbol='P2RY10P2', Aliases='', description='purinergic receptor P2Y, G-protein coupled, 10 pseudogene 2', other_designations='P2Y10-like pseudogene', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79084928, end_position_on_the_genomic_accession=79085987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927162, CurrentID=0, Status='live', Symbol='LOC101927162', Aliases='', description='histone H3.3 type 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115770586, end_position_on_the_genomic_accession=115771036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372655, CurrentID=0, Status='live', Symbol='LOC105372655', Aliases='', description='uncharacterized LOC105372655', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50053031, end_position_on_the_genomic_accession=50057196, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129776, CurrentID=0, Status='live', Symbol='LOC100129776', Aliases='', description='eukaryotic translation initiation factor 2 subunit gamma pseudogene', other_designations='eukaryotic translation initiation factor 2, subunit 3 gamma, 52kDa pseudogene', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8189826, end_position_on_the_genomic_accession=8192773, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479357, CurrentID=0, Status='live', Symbol='RN7SL368P', Aliases='', description='RNA, 7SL, cytoplasmic 368, pseudogene', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372657, CurrentID=0, Status='live', Symbol='LOC105372657', Aliases='', description='uncharacterized LOC105372657', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50350698, end_position_on_the_genomic_accession=50368831, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370769, CurrentID=0, Status='live', Symbol='LOC105370769', Aliases='', description='uncharacterized LOC105370769', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36437593, end_position_on_the_genomic_accession=36447487, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129779, CurrentID=0, Status='live', Symbol='LOC100129779', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7 pseudogene', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69737354, end_position_on_the_genomic_accession=69738016, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927163, CurrentID=0, Status='live', Symbol='LOC101927163', Aliases='', description='uncharacterized LOC101927163', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127933053, end_position_on_the_genomic_accession=127938367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372661, CurrentID=0, Status='live', Symbol='LOC105372661', Aliases='', description='uncharacterized LOC105372661', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51117390, end_position_on_the_genomic_accession=51142931, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  731075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731075, CurrentID=0, Status='live', Symbol='LOC731075', Aliases='', description='uncharacterized LOC731075', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152121954, end_position_on_the_genomic_accession=152133298, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372662, CurrentID=0, Status='live', Symbol='LOC105372662', Aliases='', description='uncharacterized LOC105372662', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51117876, end_position_on_the_genomic_accession=51122709, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  337867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337867, CurrentID=0, Status='live', Symbol='UBAC2', Aliases='PHGDHL1', description='UBA domain containing 2', other_designations='RP11-178C10.1|UBA domain-containing protein 2|phosphoglycerate dehydrogenase like 1|phosphoglycerate dehydrogenase-like protein 1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99200425, end_position_on_the_genomic_accession=99386499, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  337872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337872, CurrentID=0, Status='live', Symbol='HIST3H2BA', Aliases='', description='histone cluster 3, H2ba, pseudogene', other_designations='histone 3, H2ba', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228464170, end_position_on_the_genomic_accession=228464575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337873, CurrentID=0, Status='live', Symbol='HIST2H2BC', Aliases='H2B/t, HIST2H2BD', description='histone cluster 2, H2bc (pseudogene)', other_designations='histone 2, H2bc|histone H2B/s', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149850192, end_position_on_the_genomic_accession=149850773, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337874, CurrentID=0, Status='live', Symbol='HIST2H2BD', Aliases='H2B/o, H2B/s, H2BFO', description='histone cluster 2, H2bd (pseudogene)', other_designations='H2B histone family, member O|histone 2, H2bd', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149843039, end_position_on_the_genomic_accession=149843510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  337875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337875, CurrentID=0, Status='live', Symbol='HIST2H2BA', Aliases='', description='histone cluster 2, H2ba (pseudogene)', other_designations='histone 2, H2ba', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121108441, end_position_on_the_genomic_accession=121117251, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  337876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337876, CurrentID=0, Status='live', Symbol='CHSY3', Aliases='CHSY2, CSS3', description='chondroitin sulfate synthase 3', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase 3|N-acetylgalactosaminyltransferase 3|carbohydrate synthase 2|chSy-2|chondroitin glucuronyltransferase 3|chondroitin synthase 2|chondroitin synthase-2|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase II', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129903408, end_position_on_the_genomic_accession=130186678, orientation='plus', exon_count=8, OMIM=609963, no_hdr0=''),
  101927164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927164, CurrentID=0, Status='live', Symbol='LOC101927164', Aliases='', description='uncharacterized LOC101927164', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224219613, end_position_on_the_genomic_accession=224228043, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  337878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337878, CurrentID=0, Status='live', Symbol='KRTAP7-1', Aliases='KAP7.1', description='keratin associated protein 7-1 (gene/pseudogene)', other_designations='high tyrosine-glycine keratin-associated protein 7.1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30829039, end_position_on_the_genomic_accession=30829732, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337879, CurrentID=0, Status='live', Symbol='KRTAP8-1', Aliases='KAP8.1', description='keratin associated protein 8-1', other_designations='high glycine-tyrosine keratin-associated protein 8.1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30812697, end_position_on_the_genomic_accession=30813252, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337880, CurrentID=0, Status='live', Symbol='KRTAP11-1', Aliases='HACL-1, HACL1, KAP11.1', description='keratin associated protein 11-1', other_designations='high sulfur keratin-associated protein 11.1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30880644, end_position_on_the_genomic_accession=30881555, orientation='minus', exon_count=1, OMIM=600064, no_hdr0=''),
  337882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337882, CurrentID=0, Status='live', Symbol='KRTAP19-1', Aliases='KAP19.1', description='keratin associated protein 19-1', other_designations='high tyrosine-glycine keratin-associated protein 19.1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30480045, end_position_on_the_genomic_accession=30480317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337892, CurrentID=0, Status='live', Symbol='MGR3', Aliases='', description='Migraine, familial, with or without aura, susceptibility to', other_designations='', map_location='6p21.1-p12.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607498, no_hdr0=''),
  337893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337893, CurrentID=0, Status='live', Symbol='BULN', Aliases='', description='Bulimia nervosa, susceptibility to', other_designations='', map_location='10p', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607499, no_hdr0=''),
  337895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337895, CurrentID=0, Status='live', Symbol='BMIQ3', Aliases='', description='body mass index QTL 3', other_designations='body mass index quantitative trait 3|body mass index quantitative trait locus 3', map_location='6q23-q25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607446, no_hdr0=''),
  646140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646140, CurrentID=0, Status='live', Symbol='LOC646140', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3323177, end_position_on_the_genomic_accession=3323930, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927165, CurrentID=0, Status='live', Symbol='LOC101927165', Aliases='', description='uncharacterized LOC101927165', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56074654, end_position_on_the_genomic_accession=56090382, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128983, CurrentID=0, Status='live', Symbol='LOC100128983', Aliases='', description='mbt domain containing 1 pseudogene', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129897146, end_position_on_the_genomic_accession=129897901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372672, CurrentID=0, Status='live', Symbol='LOC105372672', Aliases='', description='uncharacterized LOC105372672', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53608354, end_position_on_the_genomic_accession=53634615, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372673, CurrentID=0, Status='live', Symbol='LOC105372673', Aliases='', description='uncharacterized LOC105372673', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53744989, end_position_on_the_genomic_accession=53751251, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372674, CurrentID=0, Status='live', Symbol='LOC105372674', Aliases='', description='uncharacterized LOC105372674', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53845514, end_position_on_the_genomic_accession=53852486, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927166, CurrentID=0, Status='live', Symbol='LOC101927166', Aliases='', description='uncharacterized LOC101927166', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48444833, end_position_on_the_genomic_accession=48466040, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372675, CurrentID=0, Status='live', Symbol='LOC105372675', Aliases='', description='uncharacterized LOC105372675', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=54178398, end_position_on_the_genomic_accession=54189802, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731157, CurrentID=0, Status='live', Symbol='LOC731157', Aliases='', description='uncharacterized LOC731157', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90410000, end_position_on_the_genomic_accession=90410669, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372676, CurrentID=0, Status='live', Symbol='LOC105372676', Aliases='', description='uncharacterized LOC105372676', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55577627, end_position_on_the_genomic_accession=55745007, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  337959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337959, CurrentID=0, Status='live', Symbol='KRTAP13-2', Aliases='KAP13-2', description='keratin associated protein 13-2', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30371391, end_position_on_the_genomic_accession=30372239, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337960, CurrentID=0, Status='live', Symbol='KRTAP13-3', Aliases='KAP13.3', description='keratin associated protein 13-3', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30425393, end_position_on_the_genomic_accession=30425912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337963, CurrentID=0, Status='live', Symbol='KRTAP23-1', Aliases='KAP23.1', description='keratin associated protein 23-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30348399, end_position_on_the_genomic_accession=30348606, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337964, CurrentID=0, Status='live', Symbol='KRTAP13-5P', Aliases='KAP13A, KRTAP13P1', description='keratin associated protein 13-5, pseudogene', other_designations='keratin associated protein 13 pseudogene 1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30436654, end_position_on_the_genomic_accession=30436953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337965, CurrentID=0, Status='live', Symbol='KRTAP13-6P', Aliases='KAP13B, KRTAP13P2', description='keratin associated protein 13-6, pseudogene', other_designations='keratin associated protein 13 pseudogene 2', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30356775, end_position_on_the_genomic_accession=30356875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337966, CurrentID=0, Status='live', Symbol='KRTAP6-1', Aliases='C21orf103, KAP6.1', description='keratin associated protein 6-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30613686, end_position_on_the_genomic_accession=30613904, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337967, CurrentID=0, Status='live', Symbol='KRTAP6-2', Aliases='KAP6.2', description='keratin associated protein 6-2', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30598686, end_position_on_the_genomic_accession=30598874, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337968, CurrentID=0, Status='live', Symbol='KRTAP6-3', Aliases='KAP6.3', description='keratin associated protein 6-3', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30592440, end_position_on_the_genomic_accession=30593055, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  337969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337969, CurrentID=0, Status='live', Symbol='KRTAP19-2', Aliases='KAP19.2', description='keratin associated protein 19-2', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30487190, end_position_on_the_genomic_accession=30487348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337970, CurrentID=0, Status='live', Symbol='KRTAP19-3', Aliases='GTRHP, KAP19.3', description='keratin associated protein 19-3', other_designations='GTHRP|glycine/tyrosine-rich protein', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30491464, end_position_on_the_genomic_accession=30491957, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337971, CurrentID=0, Status='live', Symbol='KRTAP19-4', Aliases='KAP19.4', description='keratin associated protein 19-4', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30496856, end_position_on_the_genomic_accession=30497110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337972, CurrentID=0, Status='live', Symbol='KRTAP19-5', Aliases='KAP19.5', description='keratin associated protein 19-5', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30501872, end_position_on_the_genomic_accession=30502090, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337973, CurrentID=0, Status='live', Symbol='KRTAP19-6', Aliases='KAP19.6', description='keratin associated protein 19-6', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30541535, end_position_on_the_genomic_accession=30541864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337974, CurrentID=0, Status='live', Symbol='KRTAP19-7', Aliases='KAP19.7', description='keratin associated protein 19-7', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30561098, end_position_on_the_genomic_accession=30561289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337975, CurrentID=0, Status='live', Symbol='KRTAP20-1', Aliases='KAP20.1', description='keratin associated protein 20-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30616455, end_position_on_the_genomic_accession=30616625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  337976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337976, CurrentID=0, Status='live', Symbol='KRTAP20-2', Aliases='KAP20.2', description='keratin associated protein 20-2', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30635264, end_position_on_the_genomic_accession=30635461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  337977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337977, CurrentID=0, Status='live', Symbol='KRTAP21-1', Aliases='KAP21.1', description='keratin associated protein 21-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30755139, end_position_on_the_genomic_accession=30755378, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337978, CurrentID=0, Status='live', Symbol='KRTAP21-2', Aliases='KAP21.2', description='keratin associated protein 21-2', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30746951, end_position_on_the_genomic_accession=30747202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  337979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337979, CurrentID=0, Status='live', Symbol='KRTAP22-1', Aliases='KAP22.1', description='keratin associated protein 22-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30601121, end_position_on_the_genomic_accession=30601267, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  337980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337980, CurrentID=0, Status='live', Symbol='KRTAP8-2P', Aliases='KAP8A, KRTAP8P1', description='keratin associated protein 8-2, pseudogene', other_designations='keratin associated protein 8 pseudogene 1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30802243, end_position_on_the_genomic_accession=30802365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337981, CurrentID=0, Status='live', Symbol='KRTAP8-3P', Aliases='KAP8B, KRTAP8P2', description='keratin associated protein 8-3, pseudogene', other_designations='keratin associated protein 8 pseudogene 2', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30806990, end_position_on_the_genomic_accession=30807100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337982, CurrentID=0, Status='live', Symbol='KRTAP19-9P', Aliases='KAP19A, KRTAP19P1', description='keratin associated protein 19-9, pseudogene', other_designations='keratin associated protein 19 pseudogene 1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30510287, end_position_on_the_genomic_accession=30510489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337983, CurrentID=0, Status='live', Symbol='KRTAP19-10P', Aliases='KAP19B, KRTAP19P2', description='keratin associated protein 19-10, pseudogene', other_designations='keratin associated protein 19 pseudogene 2', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30514580, end_position_on_the_genomic_accession=30514831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337984, CurrentID=0, Status='live', Symbol='KRTAP19-11P', Aliases='KAP19C, KRTAP19P3', description='keratin associated protein 19-11, pseudogene', other_designations='keratin associated protein 19 pseudogene 3', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30537178, end_position_on_the_genomic_accession=30537408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337985, CurrentID=0, Status='live', Symbol='KRTAP20-3', Aliases='KAP19D, KAP20.3, KRTAP19P4', description='keratin associated protein 20-3', other_designations='keratin associated protein 19 pseudogene 4', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30642864, end_position_on_the_genomic_accession=30643136, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  337986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337986, CurrentID=0, Status='live', Symbol='KRTAP21-4P', Aliases='KAP21A, KRTAP21P1', description='keratin associated protein 21-4, pseudogene', other_designations='keratin associated protein 21 pseudogene 1', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30741781, end_position_on_the_genomic_accession=30742614, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  337991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=337991, CurrentID=0, Status='live', Symbol='DFNB38', Aliases='', description='deafness, autosomal recessive 38', other_designations='', map_location='6q25.3-q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608219, no_hdr0=''),
  101927168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927168, CurrentID=0, Status='live', Symbol='LOC101927168', Aliases='', description='uncharacterized LOC101927168', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6728927, end_position_on_the_genomic_accession=6729862, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373306, CurrentID=0, Status='live', Symbol='LOC105373306', Aliases='', description='uncharacterized LOC105373306', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106465249, end_position_on_the_genomic_accession=106468013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338002, CurrentID=0, Status='live', Symbol='LINC00524', Aliases='C14orf71', description='long intergenic non-protein coding RNA 524', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101405987, end_position_on_the_genomic_accession=101407922, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  731220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731220, CurrentID=0, Status='live', Symbol='RFX8', Aliases='', description='RFX family member 8, lacking RFX DNA binding domain', other_designations='RFX gene family member 8, lacking RFX DNA binding domain|regulatory factor X, 8', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101397344, end_position_on_the_genomic_accession=101475096, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  338005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338005, CurrentID=0, Status='live', Symbol='LINC00221', Aliases='C14orf98, NCRNA00221', description='long intergenic non-protein coding RNA 221', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106482439, end_position_on_the_genomic_accession=106495515, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731223, CurrentID=0, Status='live', Symbol='LINC01220', Aliases='', description='long intergenic non-protein coding RNA 1220', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75294404, end_position_on_the_genomic_accession=75296410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271215, CurrentID=0, Status='live', Symbol='RPL7AP42', Aliases='RPL7A_19_836', description='ribosomal protein L7a pseudogene 42', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111971164, end_position_on_the_genomic_accession=111972025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372687, CurrentID=0, Status='live', Symbol='LOC105372687', Aliases='', description='uncharacterized LOC105372687', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57274075, end_position_on_the_genomic_accession=57350728, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  338017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338017, CurrentID=0, Status='live', Symbol='KRT8P1', Aliases='', description='keratin 8 pseudogene 1', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38726964, end_position_on_the_genomic_accession=38728761, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060256, CurrentID=0, Status='live', Symbol='LOC101060256', Aliases='', description='tripartite motif-containing protein 51-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372689, CurrentID=0, Status='live', Symbol='LOC105372689', Aliases='', description='uncharacterized LOC105372689', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57546265, end_position_on_the_genomic_accession=57548897, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338027, CurrentID=0, Status='live', Symbol='BMND2', Aliases='', description='Bone mineral density variation 2', other_designations='', map_location='1q21-q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605833, no_hdr0=''),
  252946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252946, CurrentID=0, Status='live', Symbol='FAM197Y2P', Aliases='CYorf16, FAM197Y2', description='family with sequence similarity 197 Y-linked member 2, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9519666, end_position_on_the_genomic_accession=9525268, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  338030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338030, CurrentID=0, Status='live', Symbol='GLM4', Aliases='GLM1', description='Glioma, familial, 1', other_designations='', map_location='15q23-q26.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607248, no_hdr0=''),
  252948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252948, CurrentID=0, Status='live', Symbol='TTTY16', Aliases='NCRNA00139', description='testis-specific transcript, Y-linked 16 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7699357, end_position_on_the_genomic_accession=7701247, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56341, CurrentID=0, Status='live', Symbol='PRMT8', Aliases='HRMT1L3, HRMT1L4', description='protein arginine methyltransferase 8', other_designations='HMT1 hnRNP methyltransferase-like 3|arginine methyltransferase 8|heterogeneous nuclear ribonucleoprotein methyltransferase-like protein 4|protein arginine N-methyltransferase 4', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3381349, end_position_on_the_genomic_accession=3593973, orientation='plus', exon_count=12, OMIM=610086, no_hdr0=''),
  105372694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372694, CurrentID=0, Status='live', Symbol='LOC105372694', Aliases='', description='uncharacterized LOC105372694', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58164070, end_position_on_the_genomic_accession=58179679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129815, CurrentID=0, Status='live', Symbol='PGAM1P13', Aliases='', description='phosphoglycerate mutase 1 pseudogene 13', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107647280, end_position_on_the_genomic_accession=107648033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646168, CurrentID=0, Status='live', Symbol='LINC01330', Aliases='', description='long intergenic non-protein coding RNA 1330', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167895948, end_position_on_the_genomic_accession=167924011, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  338069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338069, CurrentID=0, Status='live', Symbol='ST7-OT4', Aliases='NCRNA00042, ST7OT4', description='ST7 overlapping transcript 4', other_designations='ST7 overlapping transcript 4 (non-coding RNA)|ST7 overlapping transcript 4 (non-protein coding)', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116953899, end_position_on_the_genomic_accession=116959813, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379506, CurrentID=0, Status='live', Symbol='LOC105379506', Aliases='', description='uncharacterized LOC105379506', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8222255, end_position_on_the_genomic_accession=8223462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372698, CurrentID=0, Status='live', Symbol='LOC105372698', Aliases='', description='uncharacterized LOC105372698', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60257758, end_position_on_the_genomic_accession=60261182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129819, CurrentID=0, Status='live', Symbol='RPL22P7', Aliases='RPL22_4_275', description='ribosomal protein L22 pseudogene 7', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129939848, end_position_on_the_genomic_accession=129940217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100151665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100151665, CurrentID=0, Status='live', Symbol='RNU7-67P', Aliases='U7.67', description='RNA, U7 small nuclear 67 pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927171, CurrentID=0, Status='live', Symbol='LINC01564', Aliases='TCONS_00011314, XLOC_005327', description='long intergenic non-protein coding RNA 1564', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53628380, end_position_on_the_genomic_accession=53631394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372700, CurrentID=0, Status='live', Symbol='LOC105372700', Aliases='', description='uncharacterized LOC105372700', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60804745, end_position_on_the_genomic_accession=60808967, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338090, CurrentID=0, Status='live', Symbol='SPG24', Aliases='', description='spastic paraplegia 24 (autosomal recessive)', other_designations='Spastic paraplegia locus', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607584, no_hdr0=''),
  338091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338091, CurrentID=0, Status='live', Symbol='PSMD10P3', Aliases='', description='proteasome 26S subunit, non-ATPase, 10 pseudogene 3', other_designations='proteasome 26S non-ATPase subunit 10 pseudogene', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73036385, end_position_on_the_genomic_accession=73036845, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  731308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731308, CurrentID=0, Status='live', Symbol='DYNLL1P2', Aliases='DNCL1P2', description='dynein, light chain, LC8-type 1 pseudogene 2', other_designations='dynein, cytoplasmic, light polypeptide 1 pseudogene 2', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81412859, end_position_on_the_genomic_accession=81413439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338094, CurrentID=0, Status='live', Symbol='FAM151A', Aliases='C1orf179', description='family with sequence similarity 151 member A', other_designations='family with sequence similarity 151, member A|hyporthetical protein MGC27169', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54609177, end_position_on_the_genomic_accession=54623527, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  338095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338095, CurrentID=0, Status='live', Symbol='PSME2P3', Aliases='', description='proteasome activator subunit 2 pseudogene 3', other_designations='proteasome (prosome, macropain) activator subunit 2 (PA28 beta) pseudogene', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159007088, end_position_on_the_genomic_accession=159007877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338096, CurrentID=0, Status='live', Symbol='PSME2P4', Aliases='', description='proteasome activator subunit 2 pseudogene 4', other_designations='proteasome (prosome, macropain) activator subunit 2 (PA28 beta) pseudogene', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37995451, end_position_on_the_genomic_accession=37996265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338097, CurrentID=0, Status='live', Symbol='PSME2P5', Aliases='', description='proteasome activator subunit 2 pseudogene 5', other_designations='proteasome (prosome, macropain) activator subunit 2 (PA28 beta) pseudogene', map_location='8p21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26547669, end_position_on_the_genomic_accession=26548449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338098, CurrentID=0, Status='live', Symbol='PSME2P6', Aliases='', description='proteasome activator subunit 2 pseudogene 6', other_designations='', map_location='10p12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22225697, end_position_on_the_genomic_accession=22226071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338099, CurrentID=0, Status='live', Symbol='PSME2P2', Aliases='', description='proteasome activator subunit 2 pseudogene 2', other_designations='proteasome (prosome, macropain) activator subunit 2 (PA28 beta) pseudogene', map_location='13q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48771096, end_position_on_the_genomic_accession=48771870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372702, CurrentID=0, Status='live', Symbol='LOC105372702', Aliases='', description='uncharacterized LOC105372702', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  207032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207032, CurrentID=0, Status='live', Symbol='RPL36AP52', Aliases='dJ507I15.1', description='ribosomal protein L36a pseudogene', other_designations='60S ribosomal protein L44 (L41, L36) like', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141138441, end_position_on_the_genomic_accession=141139224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060152, CurrentID=0, Status='live', Symbol='LOC101060152', Aliases='', description='NCK-associated protein 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=97906647, end_position_on_the_genomic_accession=97908084, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372703, CurrentID=0, Status='live', Symbol='LOC105372703', Aliases='', description='uncharacterized LOC105372703', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61433090, end_position_on_the_genomic_accession=61437453, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372704, CurrentID=0, Status='live', Symbol='LOC105372704', Aliases='', description='uncharacterized LOC105372704', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62000728, end_position_on_the_genomic_accession=62002317, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372705, CurrentID=0, Status='live', Symbol='LOC105372705', Aliases='', description='uncharacterized LOC105372705', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62066200, end_position_on_the_genomic_accession=62068533, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646179, CurrentID=0, Status='live', Symbol='LOC646179', Aliases='', description='hCG2045614', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  207063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207063, CurrentID=0, Status='live', Symbol='DHRSX', Aliases='CXorf11, DHRS5X, DHRS5YY, DHRSY, SDR46C1, SDR7C6, DHRSX', description='dehydrogenase/reductase (SDR family) X-linked', other_designations='dehydrogenase/reductase (SDR family) X chromosome|dehydrogenase/reductase (SDR family) Y-linked|short chain dehydrogenase/reductase family 46C member 1|short chain dehydrogenase/reductase family 46C, member 1|short chain dehydrogenase/reductase family 7C member 6|short chain dehydrogenase/reductase family 7C, member 6', map_location='Xp22.33; Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2219506, end_position_on_the_genomic_accession=2500974, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375144, CurrentID=0, Status='live', Symbol='LOC105375144', Aliases='', description='uncharacterized LOC105375144', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8563494, end_position_on_the_genomic_accession=8632624, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100129830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129830, CurrentID=0, Status='live', Symbol='LOC100129830', Aliases='', description='5-azacytidine induced 2 pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44919957, end_position_on_the_genomic_accession=44921037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372711, CurrentID=0, Status='live', Symbol='LOC105372711', Aliases='', description='uncharacterized LOC105372711', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62481285, end_position_on_the_genomic_accession=62490856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271216, CurrentID=0, Status='live', Symbol='RPL7P35', Aliases='RPL7_15_1095', description='ribosomal protein L7 pseudogene 35', other_designations='', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110550046, end_position_on_the_genomic_accession=110550873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  252968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252968, CurrentID=0, Status='live', Symbol='TNPO1P2', Aliases='', description='transportin 1 pseudogene 2', other_designations='karyopherin (importin) beta 2 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18460332, end_position_on_the_genomic_accession=18463332, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372713, CurrentID=0, Status='live', Symbol='LOC105372713', Aliases='', description='uncharacterized LOC105372713', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62589665, end_position_on_the_genomic_accession=62598358, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375145, CurrentID=0, Status='live', Symbol='LOC105375145', Aliases='', description='uncharacterized LOC105375145', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9639546, end_position_on_the_genomic_accession=9667092, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927174, CurrentID=0, Status='live', Symbol='LOC101927174', Aliases='', description='uncharacterized LOC101927174', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112166, end_position_on_the_genomic_accession=163349, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  207107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207107, CurrentID=0, Status='live', Symbol='SFTA1P', Aliases='SFTPF', description='surfactant associated 1, pseudogene', other_designations='surfactant associated 1 (pseudogene)|surfactant associated protein F', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10784439, end_position_on_the_genomic_accession=10794914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372716, CurrentID=0, Status='live', Symbol='LOC105372716', Aliases='', description='uncharacterized LOC105372716', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62783218, end_position_on_the_genomic_accession=62794298, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  207115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207115, CurrentID=0, Status='live', Symbol='LATD', Aliases='', description='Laterality defects, autosomal dominant', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601086, no_hdr0=''),
  105372717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372717, CurrentID=0, Status='live', Symbol='LOC105372717', Aliases='', description='uncharacterized LOC105372717', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62861426, end_position_on_the_genomic_accession=62878190, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370781, CurrentID=0, Status='live', Symbol='LOC105370781', Aliases='', description='uncharacterized LOC105370781', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39019180, end_position_on_the_genomic_accession=39027446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442486, CurrentID=0, Status='live', Symbol='ZNF736P1Y', Aliases='ZNF736P1, ZNF736PY1', description='zinc finger protein 736 pseudogene 1, Y-linked', other_designations='zinc finger protein 736 pseudogene, Y-linked 1', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24985203, end_position_on_the_genomic_accession=24990367, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  731424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731424, CurrentID=0, Status='live', Symbol='MIR3945HG', Aliases='', description='MIR3945 host gene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184843296, end_position_on_the_genomic_accession=184847046, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372720, CurrentID=0, Status='live', Symbol='LOC105372720', Aliases='', description='uncharacterized LOC105372720', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63447024, end_position_on_the_genomic_accession=63450897, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372721, CurrentID=0, Status='live', Symbol='LOC105372721', Aliases='', description='uncharacterized LOC105372721', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63435346, end_position_on_the_genomic_accession=63436091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420482, CurrentID=0, Status='live', Symbol='LOC100420482', Aliases='', description='G protein-coupled receptor 160 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31759826, end_position_on_the_genomic_accession=31760944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  207147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207147, CurrentID=0, Status='live', Symbol='TRY2P', Aliases='', description='trypsinogen-like pseudogene', other_designations='trypsin X3 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142268284, end_position_on_the_genomic_accession=142272249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  207148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=207148, CurrentID=0, Status='live', Symbol='TRY7', Aliases='', description='trypsinogen D', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  121906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121906, CurrentID=0, Status='live', Symbol='PSMA6P4', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 6 pseudogene 4', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97475703, end_position_on_the_genomic_accession=97476625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372723, CurrentID=0, Status='live', Symbol='LOC105372723', Aliases='', description='uncharacterized LOC105372723', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372724, CurrentID=0, Status='live', Symbol='LOC105372724', Aliases='', description='uncharacterized LOC105372724', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63409539, end_position_on_the_genomic_accession=63413375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375147, CurrentID=0, Status='live', Symbol='LOC105375147', Aliases='', description='uncharacterized LOC105375147', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9708053, end_position_on_the_genomic_accession=10085470, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927176, CurrentID=0, Status='live', Symbol='LINC01069', Aliases='', description='long intergenic non-protein coding RNA 1069', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77979766, end_position_on_the_genomic_accession=77994773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372725, CurrentID=0, Status='live', Symbol='LOC105372725', Aliases='', description='uncharacterized LOC105372725', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63502058, end_position_on_the_genomic_accession=63505726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129846, CurrentID=0, Status='live', Symbol='LOC100129846', Aliases='', description='uncharacterized LOC100129846', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=30545233, end_position_on_the_genomic_accession=30547826, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372727, CurrentID=0, Status='live', Symbol='LOC105372727', Aliases='', description='uncharacterized LOC105372727', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64139445, end_position_on_the_genomic_accession=64143682, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370783, CurrentID=0, Status='live', Symbol='LOC105370783', Aliases='', description='uncharacterized LOC105370783', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39473243, end_position_on_the_genomic_accession=39493031, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372729, CurrentID=0, Status='live', Symbol='LOC105372729', Aliases='', description='uncharacterized LOC105372729', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102443352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102443352, CurrentID=0, Status='live', Symbol='LOC102443352', Aliases='', description='ubiquitin-fold modifier 1 pseudogene', other_designations='', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145768026, end_position_on_the_genomic_accession=145769946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372730, CurrentID=0, Status='live', Symbol='LOC105372730', Aliases='', description='uncharacterized LOC105372730', other_designations='bP-2189O9.3', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8759113, end_position_on_the_genomic_accession=8805942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131553, CurrentID=0, Status='live', Symbol='LOC100131553', Aliases='', description='disintegrin and metalloproteinase domain-containing protein 21-like', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175016936, end_position_on_the_genomic_accession=175019136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129851, CurrentID=0, Status='live', Symbol='LOC100129851', Aliases='', description='claudin 7 pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40760092, end_position_on_the_genomic_accession=40761319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372732, CurrentID=0, Status='live', Symbol='LOC105372732', Aliases='', description='uncharacterized LOC105372732', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9757291, end_position_on_the_genomic_accession=9758365, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370784, CurrentID=0, Status='live', Symbol='LOC105370784', Aliases='', description='uncharacterized LOC105370784', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39512406, end_position_on_the_genomic_accession=39519937, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372733, CurrentID=0, Status='live', Symbol='LOC105372733', Aliases='', description='uncharacterized LOC105372733', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9996067, end_position_on_the_genomic_accession=10119507, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129854, CurrentID=0, Status='live', Symbol='LOC100129854', Aliases='', description='arginase 2 pseudogene', other_designations='arginase, type II pseudogene', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100177178, end_position_on_the_genomic_accession=100178192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646208, CurrentID=0, Status='live', Symbol='NUS1P2', Aliases='', description='NUS1 dehydrodolichyl diphosphate synthase subunit pseudogene 2', other_designations='nuclear undecaprenyl pyrophosphate synthase 1 homolog (S. cerevisiae) pseudogene 2|nuclear undecaprenyl pyrophosphate synthase 1 pseudogene', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22915461, end_position_on_the_genomic_accession=22917346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370785, CurrentID=0, Status='live', Symbol='LOC105370785', Aliases='', description='uncharacterized LOC105370785', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39628409, end_position_on_the_genomic_accession=39650632, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645579, CurrentID=0, Status='live', Symbol='EIF4BP6', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 6', other_designations='', map_location='7q22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104667733, end_position_on_the_genomic_accession=104671585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338321, CurrentID=0, Status='live', Symbol='NLRP9', Aliases='CLR19.1, NALP9, NOD6, PAN12', description='NLR family, pyrin domain containing 9', other_designations='NACHT, LRR and PYD containing protein 9|NACHT, leucine rich repeat and PYD containing 9|PYRIN and NACHT-containing protein 12|nucleotide-binding oligomerization domain protein 6|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 9', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55708432, end_position_on_the_genomic_accession=55745528, orientation='minus', exon_count=12, OMIM=609663, no_hdr0=''),
  338322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338322, CurrentID=0, Status='live', Symbol='NLRP10', Aliases='CLR11.1, NALP10, NOD8, PAN5, PYNOD', description='NLR family, pyrin domain containing 10', other_designations='NACHT, leucine rich repeat and PYD containing 10|nucleotide-binding oligomerization domain protein 8|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 10', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7959023, end_position_on_the_genomic_accession=7965424, orientation='minus', exon_count=3, OMIM=609662, no_hdr0=''),
  338323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338323, CurrentID=0, Status='live', Symbol='NLRP14', Aliases='CLR11.2, GC-LRR, NALP14, NOD5, PAN8', description='NLR family, pyrin domain containing 14', other_designations='NACHT, LRR and PYD containing protein 14|NACHT, leucine rich repeat and PYD containing 14|nucleotide-binding oligomerization domain protein 5|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 14', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7020446, end_position_on_the_genomic_accession=7089025, orientation='plus', exon_count=13, OMIM=609665, no_hdr0=''),
  338324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338324, CurrentID=0, Status='live', Symbol='S100A7A', Aliases='NICE-2, S100A15, S100A7L1, S100A7f', description='S100 calcium binding protein A7A', other_designations='S100 calcium binding protein A15|S100 calcium binding protein A7-like 1|S100 calcium-binding protein A15|S100 calcium-binding protein A7-like 1|S100 calcium-binding protein A7A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153416524, end_position_on_the_genomic_accession=153423225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338328, CurrentID=0, Status='live', Symbol='GPIHBP1', Aliases='GPI-HBP1, HYPL1D', description='glycosylphosphatidylinositol anchored high density lipoprotein binding protein 1', other_designations='GPI anchored high density lipoprotein binding protein 1|GPI-anchored HDL-binding protein 1|endothelial cell LPL transporter', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143213193, end_position_on_the_genomic_accession=143217170, orientation='plus', exon_count=4, OMIM=612757, no_hdr0=''),
  105372740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372740, CurrentID=0, Status='live', Symbol='LOC105372740', Aliases='', description='uncharacterized LOC105372740', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  338331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338331, CurrentID=0, Status='live', Symbol='AZON', Aliases='', description='Azoospremia, nonobstructive', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606766, no_hdr0=''),
  338332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338332, CurrentID=0, Status='live', Symbol='CELIAC5', Aliases='GSES', description='Celiac disease, susceptibility to, 5', other_designations='Gluten-sensitive enteropathy, susceptibility to', map_location='15q11-q13', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607202, no_hdr0=''),
  338333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338333, CurrentID=0, Status='live', Symbol='PAPA3', Aliases='', description='Polydactyly, postaxial, type A3', other_designations='', map_location='19p13.2-p13.1', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607324, no_hdr0=''),
  338334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338334, CurrentID=0, Status='live', Symbol='CHDS1', Aliases='', description='Coronary heart disease, susceptibility to, 1', other_designations='', map_location='16pter-p13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607339, no_hdr0=''),
  646213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646213, CurrentID=0, Status='live', Symbol='SS18L2P1', Aliases='', description='synovial sarcoma translocation gene on chromosome 18-like 2 pseudogene 1', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34875161, end_position_on_the_genomic_accession=34876225, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338339, CurrentID=0, Status='live', Symbol='CLEC4D', Aliases='CD368, CLEC-6, CLEC6, CLECSF8, MCL, MPCL', description='C-type lectin domain family 4 member D', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 8|C-type lectin domain family 4, member D|C-type lectin receptor|C-type lectin superfamily member 8|C-type lectin-like receptor 6|macrophage C-type lectin', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8513482, end_position_on_the_genomic_accession=8522366, orientation='plus', exon_count=6, OMIM=609964, no_hdr0=''),
  338340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338340, CurrentID=0, Status='live', Symbol='MMS', Aliases='', description='Malignant mesothelioma, susceptibility to', other_designations='', map_location='9p', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=156240, no_hdr0=''),
  105379515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379515, CurrentID=0, Status='live', Symbol='LOC105379515', Aliases='', description='uncharacterized LOC105379515', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11969667, end_position_on_the_genomic_accession=11971963, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338342, CurrentID=0, Status='live', Symbol='MGR4', Aliases='MGOA', description='Migraine, susceptibility to, 4', other_designations='', map_location='14q21.2-q22.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607501, no_hdr0=''),
  105370786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370786, CurrentID=0, Status='live', Symbol='LOC105370786', Aliases='', description='uncharacterized LOC105370786', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39821491, end_position_on_the_genomic_accession=39826901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129863, CurrentID=0, Status='live', Symbol='SUCLA2P2', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene 2', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76106001, end_position_on_the_genomic_accession=76107199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646216, CurrentID=0, Status='live', Symbol='LOC646216', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26422023, end_position_on_the_genomic_accession=26423218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927180, CurrentID=0, Status='live', Symbol='LOC101927180', Aliases='', description='V-type proton ATPase subunit F pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48084898, end_position_on_the_genomic_accession=48085740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646218, CurrentID=0, Status='live', Symbol='PRDX2P2', Aliases='', description='peroxiredoxin 2 pseudogene 2', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34965069, end_position_on_the_genomic_accession=34966256, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379516, CurrentID=0, Status='live', Symbol='LOC105379516', Aliases='', description='uncharacterized LOC105379516', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=12131701, end_position_on_the_genomic_accession=12167053, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370787, CurrentID=0, Status='live', Symbol='LOC105370787', Aliases='', description='uncharacterized LOC105370787', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40075943, end_position_on_the_genomic_accession=40082302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338376, CurrentID=0, Status='live', Symbol='IFNE', Aliases='IFN-E1, IFNT1, INFE1, PRO655, IFNE', description='interferon, epsilon', other_designations='IFN-epsilon|interferon epsilon 1|interferon epsilon-1|interferon tau-1|interferon-epsilon', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21480839, end_position_on_the_genomic_accession=21482313, orientation='minus', exon_count=1, OMIM=615223, no_hdr0=''),
  338382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338382, CurrentID=0, Status='live', Symbol='RAB7B', Aliases='RAB7', description='RAB7B, member RAS oncogene family', other_designations='Ras-related protein Rab-7', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205976740, end_position_on_the_genomic_accession=206003495, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100129869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129869, CurrentID=0, Status='live', Symbol='LOC100129869', Aliases='', description='uncharacterized LOC100129869', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57958681, end_position_on_the_genomic_accession=57959615, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  338386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338386, CurrentID=0, Status='live', Symbol='HSR', Aliases='', description='Hand skill, relative', other_designations='', map_location='2p12-q22', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=139900, no_hdr0=''),
  338388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338388, CurrentID=0, Status='live', Symbol='MRPS16P3', Aliases='', description='mitochondrial ribosomal protein S16 pseudogene 3', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35703765, end_position_on_the_genomic_accession=35704174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338389, CurrentID=0, Status='live', Symbol='MRPS24P1', Aliases='', description='mitochondrial ribosomal protein S24 pseudogene 1', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=303710, end_position_on_the_genomic_accession=304376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338390, CurrentID=0, Status='live', Symbol='MRPS25P1', Aliases='', description='mitochondrial ribosomal protein S25 pseudogene 1', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80321540, end_position_on_the_genomic_accession=80322059, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338391, CurrentID=0, Status='live', Symbol='HIST2H2BB', Aliases='', description='histone cluster 2, H2bb (pseudogene)', other_designations='histone 2, H2bb', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143904223, end_position_on_the_genomic_accession=143904603, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372751, CurrentID=0, Status='live', Symbol='LOC105372751', Aliases='', description='uncharacterized LOC105372751', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=24530664, end_position_on_the_genomic_accession=24547942, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338396, CurrentID=0, Status='live', Symbol='TAS2R2P', Aliases='PS9, T2R02, T2R2, TAS2R2', description='taste 2 receptor member 2 pseudogene', other_designations='taste receptor Ps9 pseudogene|taste receptor, type 2, member 2 pseudogene|taste receptor, type 2, member 2, pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12491095, end_position_on_the_genomic_accession=12492004, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338398, CurrentID=0, Status='live', Symbol='TAS2R60', Aliases='T2R56, T2R60', description='taste 2 receptor member 60', other_designations='taste receptor type 2 member 56|taste receptor, type 2, member 60', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143443453, end_position_on_the_genomic_accession=143444409, orientation='plus', exon_count=1, OMIM=613968, no_hdr0=''),
  338399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338399, CurrentID=0, Status='live', Symbol='TAS2R62P', Aliases='PS1, T2R62, TAS2R62', description='taste 2 receptor member 62 pseudogene', other_designations='taste receptor, type 2, member 62, pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143437035, end_position_on_the_genomic_accession=143437973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129872, CurrentID=0, Status='live', Symbol='LOC100129872', Aliases='', description='uncharacterized LOC100129872', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=45740052, end_position_on_the_genomic_accession=45746558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338411, CurrentID=0, Status='live', Symbol='TELAB1', Aliases='', description='telangiectasia, benign 1', other_designations='', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=187260, no_hdr0=''),
  338412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338412, CurrentID=0, Status='live', Symbol='TAS2R64P', Aliases='PS2, T2R64, T2R64P', description='taste 2 receptor member 64 pseudogene', other_designations='taste receptor, type 2, member 19 pseudogene|taste receptor, type 2, member 64 pseudogene|taste receptor, type 2, member 64, pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11077345, end_position_on_the_genomic_accession=11078242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338413, CurrentID=0, Status='live', Symbol='TAS2R63P', Aliases='PS6, T2R63', description='taste 2 receptor member 63 pseudogene', other_designations='taste receptor, type 2, member 63 pseudogene|taste receptor, type 2, member 63, pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11048357, end_position_on_the_genomic_accession=11049256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338414, CurrentID=0, Status='live', Symbol='TAS2R18P', Aliases='PS4, T2R18, T2R65, TAS2R18, TAS2R65, TAS2R65P', description='taste 2 receptor member 18 pseudogene', other_designations='taste receptor, type 2, member 18 pseudogene|taste receptor, type 2, member 18, pseudogene|taste receptor, type 2, member 65 pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11158776, end_position_on_the_genomic_accession=11159694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  731631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731631, CurrentID=0, Status='live', Symbol='LOC731631', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372755, CurrentID=0, Status='live', Symbol='LOC105372755', Aliases='', description='uncharacterized LOC105372755', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25744060, end_position_on_the_genomic_accession=25754784, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372756, CurrentID=0, Status='live', Symbol='LOC105372756', Aliases='', description='uncharacterized LOC105372756', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25862371, end_position_on_the_genomic_accession=25865989, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  731642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731642, CurrentID=0, Status='live', Symbol='RPL18AP9', Aliases='', description='ribosomal protein L18a pseudogene', other_designations='', map_location='3p22.1-p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43484919, end_position_on_the_genomic_accession=43485916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338427, CurrentID=0, Status='live', Symbol='SNORD108', Aliases='HBII-437', description='small nucleolar RNA, C/D box 108', other_designations='HBII-437 C/D box snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24986925, end_position_on_the_genomic_accession=24986993, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  338428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338428, CurrentID=0, Status='live', Symbol='SNORD109A', Aliases='HBII-438A', description='small nucleolar RNA, C/D box 109A', other_designations='HBII-438A C/D box snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25041974, end_position_on_the_genomic_accession=25042040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  338429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338429, CurrentID=0, Status='live', Symbol='SNORD109B', Aliases='HBII-438B', description='small nucleolar RNA, C/D box 109B', other_designations='HBII-438B C/D box snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25278343, end_position_on_the_genomic_accession=25278409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372757, CurrentID=0, Status='live', Symbol='LOC105372757', Aliases='', description='uncharacterized LOC105372757', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25800176, end_position_on_the_genomic_accession=25806990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338433, CurrentID=0, Status='live', Symbol='SNORD115-1', Aliases='HBII-52-1, RNHBII52', description='small nucleolar RNA, C/D box 115-1', other_designations='HBII-52-1 snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25170723, end_position_on_the_genomic_accession=25170804, orientation='plus', exon_count=1, OMIM=609837, no_hdr0=''),
  338434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338434, CurrentID=0, Status='live', Symbol='CANDN1', Aliases='CANDF3, FCNC', description='candidiasis, nail 1 (with ICAM1 deficiency)', other_designations='', map_location='11p12-q12.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607644, no_hdr0=''),
  338435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338435, CurrentID=0, Status='live', Symbol='SCA25', Aliases='', description='spinocerebellar ataxia 25', other_designations='', map_location='2p21-p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608703, no_hdr0=''),
  338436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338436, CurrentID=0, Status='live', Symbol='BLACE', Aliases='', description='B-cell acute lymphoblastic leukemia expressed', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155356988, end_position_on_the_genomic_accession=155367933, orientation='minus', exon_count=6, OMIM=608450, no_hdr0=''),
  100129878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129878, CurrentID=0, Status='live', Symbol='LOC100129878', Aliases='', description='uncharacterized LOC100129878', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49024628, end_position_on_the_genomic_accession=49048469, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338440, CurrentID=0, Status='live', Symbol='ANO9', Aliases='PIG5, TMEM16J, TP53I5', description='anoctamin 9', other_designations='p53-induced gene 5 protein|transmembrane protein 16J|tumor protein p53 inducible protein 5|tumor protein p53-inducible protein 5', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=417930, end_position_on_the_genomic_accession=442011, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  338442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338442, CurrentID=0, Status='live', Symbol='HCAR2', Aliases='GPR109A, HCA2, HM74a, HM74b, NIACR1, PUMAG, Puma-g', description='hydroxycarboxylic acid receptor 2', other_designations='G protein-coupled receptor 109A|G protein-coupled receptor HM74a|G-protein coupled receptor 109A|G-protein coupled receptor HM74A|hydroxy-carboxylic acid receptor 2|niacin receptor 1|nicotinic acid receptor', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122701293, end_position_on_the_genomic_accession=122703357, orientation='minus', exon_count=1, OMIM=609163, no_hdr0=''),
  105372759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372759, CurrentID=0, Status='live', Symbol='LOC105372759', Aliases='', description='uncharacterized LOC105372759', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27204884, end_position_on_the_genomic_accession=27234255, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369295, CurrentID=0, Status='live', Symbol='LOC105369295', Aliases='', description='uncharacterized LOC105369295', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36061896, end_position_on_the_genomic_accession=36068569, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338448, CurrentID=0, Status='live', Symbol='GTF2IP2', Aliases='', description='general transcription factor IIi pseudogene 2', other_designations='general transcription factor IIi, pseudogene 2', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13443354, end_position_on_the_genomic_accession=13462471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372760, CurrentID=0, Status='live', Symbol='LOC105372760', Aliases='', description='uncharacterized LOC105372760', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26854880, end_position_on_the_genomic_accession=26953019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646233, CurrentID=0, Status='live', Symbol='RPS27P10', Aliases='RPS27_2_322', description='ribosomal protein S27 pseudogene 10', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=211298878, end_position_on_the_genomic_accession=211299221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129882, CurrentID=0, Status='live', Symbol='RPL31P49', Aliases='RPL31_26_1276', description='ribosomal protein L31 pseudogene 49', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110460960, end_position_on_the_genomic_accession=110461401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338470, CurrentID=0, Status='live', Symbol='MESTP1', Aliases='', description='mesoderm specific transcript pseudogene 1', other_designations='mesoderm specific transcript homolog pseudogene 1', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128927756, end_position_on_the_genomic_accession=128930216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  731688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731688, CurrentID=0, Status='live', Symbol='RPL21P120', Aliases='RPL21_57_1527', description='ribosomal protein L21 pseudogene 120', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21504518, end_position_on_the_genomic_accession=21505094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372764, CurrentID=0, Status='live', Symbol='LOC105372764', Aliases='', description='uncharacterized LOC105372764', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28391708, end_position_on_the_genomic_accession=28393335, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375155, CurrentID=0, Status='live', Symbol='LOC105375155', Aliases='', description='uncharacterized LOC105375155', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12433491, end_position_on_the_genomic_accession=12443799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372767, CurrentID=0, Status='live', Symbol='LOC105372767', Aliases='', description='uncharacterized LOC105372767', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29178119, end_position_on_the_genomic_accession=29183572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379520, CurrentID=0, Status='live', Symbol='LOC105379520', Aliases='', description='uncharacterized LOC105379520', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421425, CurrentID=0, Status='live', Symbol='LOC100421425', Aliases='', description='calmodulin 2 (phosphorylase kinase, delta) pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372768, CurrentID=0, Status='live', Symbol='LOC105372768', Aliases='', description='uncharacterized LOC105372768', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29561624, end_position_on_the_genomic_accession=29577698, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731716, CurrentID=0, Status='live', Symbol='PRKRAP1', Aliases='', description='protein kinase, interferon-inducible double stranded RNA dependent activator pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372769, CurrentID=0, Status='live', Symbol='LOC105372769', Aliases='', description='uncharacterized LOC105372769', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29555309, end_position_on_the_genomic_accession=29559910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128063, CurrentID=0, Status='live', Symbol='LOC100128063', Aliases='', description='uncharacterized LOC100128063', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372771, CurrentID=0, Status='live', Symbol='LOC105372771', Aliases='', description='uncharacterized LOC105372771', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29413561, end_position_on_the_genomic_accession=29417792, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  731741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731741, CurrentID=0, Status='live', Symbol='RPL32P12', Aliases='RPL32_4_459', description='ribosomal protein L32 pseudogene 12', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13647694, end_position_on_the_genomic_accession=13648400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372773, CurrentID=0, Status='live', Symbol='LOC105372773', Aliases='', description='uncharacterized LOC105372773', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30707278, end_position_on_the_genomic_accession=30725702, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372774, CurrentID=0, Status='live', Symbol='LOC105372774', Aliases='', description='uncharacterized LOC105372774', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31030142, end_position_on_the_genomic_accession=31036268, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927186, CurrentID=0, Status='live', Symbol='ADGRL3-AS1', Aliases='LPHN3-AS1', description='adhesion G protein-coupled receptor L3 antisense RNA 1', other_designations='LPHN3 antisense RNA 1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62071755, end_position_on_the_genomic_accession=62161765, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  399715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399715, CurrentID=0, Status='live', Symbol='LOC399715', Aliases='', description='uncharacterized LOC399715', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6326544, end_position_on_the_genomic_accession=6335982, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372777, CurrentID=0, Status='live', Symbol='LOC105372777', Aliases='', description='uncharacterized LOC105372777', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31255432, end_position_on_the_genomic_accession=31262458, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421427, CurrentID=0, Status='live', Symbol='ANKRD20A16P', Aliases='', description='ankyrin repeat domain 20 family member A16, pseudogene', other_designations='ankyrin repeat domain 20 family, member A16, pseudogene|ankyrin repeat domain 20 family, member A2 pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8877128, end_position_on_the_genomic_accession=8880976, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338557, CurrentID=0, Status='live', Symbol='FFAR4', Aliases='BMIQ10, GPR120, GPR129, GT01, O3FAR1, PGR4', description='free fatty acid receptor 4', other_designations='G-protein coupled receptor 120|G-protein coupled receptor 129|G-protein coupled receptor GT01|G-protein coupled receptor PGR4|omega-3 fatty acid receptor 1', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93566204, end_position_on_the_genomic_accession=93590072, orientation='plus', exon_count=4, OMIM=609044, no_hdr0=''),
  731779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731779, CurrentID=0, Status='live', Symbol='LINC01300', Aliases='', description='long intergenic non-protein coding RNA 1300', other_designations='CTD-3064M3.1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141340549, end_position_on_the_genomic_accession=141344621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338567, CurrentID=0, Status='live', Symbol='KCNK18', Aliases='K2p18.1, MGR13, TRESK, TRESK-2, TRESK2, TRIK', description='potassium channel, two pore domain subfamily K, member 18', other_designations='TWIK-related individual K+ channel|TWIK-related individual potassium channel|TWIK-related spinal cord K+ channel|TWIK-related spinal cord potassium channel|potassium channel, subfamily K, member 18', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117197489, end_position_on_the_genomic_accession=117210299, orientation='plus', exon_count=3, OMIM=613655, no_hdr0=''),
  101927187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927187, CurrentID=0, Status='live', Symbol='LOC101927187', Aliases='', description='uncharacterized LOC101927187', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241687765, end_position_on_the_genomic_accession=241694219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129900, CurrentID=0, Status='live', Symbol='LOC100129900', Aliases='', description='P008-1', other_designations='putative P008-2', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=29436144, end_position_on_the_genomic_accession=29436314, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  731789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731789, CurrentID=0, Status='live', Symbol='LINC00202-2', Aliases='', description='long intergenic non-protein coding RNA 202-2', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26643108, end_position_on_the_genomic_accession=26653454, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  338579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338579, CurrentID=0, Status='live', Symbol='ANKRD30BP3', Aliases='', description='ankyrin repeat domain 30B pseudogene 3', other_designations='ankyrin repeat domain-containing protein 30B pseudogene', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45154662, end_position_on_the_genomic_accession=45186041, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372782, CurrentID=0, Status='live', Symbol='LOC105372782', Aliases='', description='uncharacterized LOC105372782', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32586431, end_position_on_the_genomic_accession=32592239, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379523, CurrentID=0, Status='live', Symbol='LOC105379523', Aliases='', description='uncharacterized LOC105379523', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646255, CurrentID=0, Status='live', Symbol='RPL19P16', Aliases='RPL19_5_1105', description='ribosomal protein L19 pseudogene 16', other_designations='', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121133016, end_position_on_the_genomic_accession=121133754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338588, CurrentID=0, Status='live', Symbol='LINC00705', Aliases='', description='long intergenic non-protein coding RNA 705', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4656156, end_position_on_the_genomic_accession=4662419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338591, CurrentID=0, Status='live', Symbol='CHCHD3P1', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 3 pseudogene 1', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8619933, end_position_on_the_genomic_accession=8620610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338596, CurrentID=0, Status='live', Symbol='ST8SIA6', Aliases='SIA8F, SIAT8F, ST8SIA-VI', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 6', other_designations='SIAT8-F|ST8SiaVI|alpha-2,8-sialyltransferase 8F variant 1|alpha-2,8-sialyltransferase 8F variant 2|alpha-2,8-sialyltransferase 8F variant 3|sialyltransferase 8F (alpha-2, 8-sialyltransferase)|sialyltransferase St8Sia VI|sialytransferase St8Sia VI', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17315414, end_position_on_the_genomic_accession=17455510, orientation='minus', exon_count=10, OMIM=610139, no_hdr0=''),
  101927188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927188, CurrentID=0, Status='live', Symbol='LOC101927188', Aliases='', description='uncharacterized LOC101927188', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6954677, end_position_on_the_genomic_accession=6957419, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338599, CurrentID=0, Status='live', Symbol='DUPD1', Aliases='DUSP27, FMDSP', description='dual specificity phosphatase and pro isomerase domain containing 1', other_designations='atypical dual-specific protein phosphatase|dual specificity phosphatase 27', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75037472, end_position_on_the_genomic_accession=75073632, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  338611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338611, CurrentID=0, Status='live', Symbol='RPS26P40', Aliases='RPS26_20_1061', description='ribosomal protein S26 pseudogene 40', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70794116, end_position_on_the_genomic_accession=70794549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421429, CurrentID=0, Status='live', Symbol='RTEL1P1', Aliases='', description='regulator of telomere elongation helicase 1 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112355537, end_position_on_the_genomic_accession=112359913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646260, CurrentID=0, Status='live', Symbol='SAR1AP1', Aliases='SAR1P1, SARAP, SARAP1, dJ88J8.3, hsSARA-p', description='SAR1 homolog A (S. cerevisiae) pseudogene 1', other_designations='SAR1 gene homolog pseudogene 1|SAR1 homolog pseudogene 1|SAR1a gene homolog pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29074416, end_position_on_the_genomic_accession=29077241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929726, CurrentID=0, Status='live', Symbol='LOC101929726', Aliases='', description='uncharacterized LOC101929726', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44192365, end_position_on_the_genomic_accession=44218236, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  338620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338620, CurrentID=0, Status='live', Symbol='LOC338620', Aliases='', description='uncharacterized LOC338620', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927189, CurrentID=0, Status='live', Symbol='LOC101927189', Aliases='', description='uncharacterized LOC101927189', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53929982, end_position_on_the_genomic_accession=53997667, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372790, CurrentID=0, Status='live', Symbol='LOC105372790', Aliases='', description='uncharacterized LOC105372790', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33982966, end_position_on_the_genomic_accession=34051939, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  338637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338637, CurrentID=0, Status='live', Symbol='OR52M2P', Aliases='OR52M4', description='olfactory receptor family 52 subfamily M member 2 pseudogene', other_designations='olfactory receptor, family 52, subfamily M, member 2 pseudogene|olfactory receptor, family 52, subfamily M, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4514945, end_position_on_the_genomic_accession=4515924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372792, CurrentID=0, Status='live', Symbol='LOC105372792', Aliases='', description='uncharacterized LOC105372792', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34484016, end_position_on_the_genomic_accession=34486898, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  338645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338645, CurrentID=0, Status='live', Symbol='LUZP2', Aliases='KFSP2566, PRO6246', description='leucine zipper protein 2', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=24496488, end_position_on_the_genomic_accession=25082640, orientation='plus', exon_count=15, OMIM=608178, no_hdr0=''),
  338651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338651, CurrentID=0, Status='live', Symbol='KRTAP5-AS1', Aliases='', description='KRTAP5-1/KRTAP5-2 antisense RNA 1', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1572741, end_position_on_the_genomic_accession=1599184, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338653, CurrentID=0, Status='live', Symbol='KCNQ1-AS1', Aliases='', description='KCNQ1 antisense RNA 1', other_designations='KCNQ1 antisense RNA 1 (non-protein coding)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2840135, end_position_on_the_genomic_accession=2861569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338657, CurrentID=0, Status='live', Symbol='CCDC84', Aliases='DLNB14', description='coiled-coil domain containing 84', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118994353, end_position_on_the_genomic_accession=119015793, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101927190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927190, CurrentID=0, Status='live', Symbol='LOC101927190', Aliases='', description='uncharacterized LOC101927190', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116367416, end_position_on_the_genomic_accession=116374570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338661, CurrentID=0, Status='live', Symbol='TMEM225', Aliases='PMP22CD, PPP1R154', description='transmembrane protein 225', other_designations='PMP22 claudin domain-containing protein|protein phosphatase 1, regulatory subunit 154', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123882920, end_position_on_the_genomic_accession=123885650, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  338662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338662, CurrentID=0, Status='live', Symbol='OR8D4', Aliases='OR11-275', description='olfactory receptor family 8 subfamily D member 4', other_designations='olfactory receptor OR11-275|olfactory receptor, family 8, subfamily D, member 4', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123906432, end_position_on_the_genomic_accession=123907376, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646268, CurrentID=0, Status='live', Symbol='LOC646268', Aliases='', description='hCG1654703', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158132044, end_position_on_the_genomic_accession=158140640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338667, CurrentID=0, Status='live', Symbol='LOC338667', Aliases='', description='V-set and immunoglobulin domain-containing protein 10-like', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125947424, end_position_on_the_genomic_accession=125957040, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  121981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121981, CurrentID=0, Status='live', Symbol='PPIAP26', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 26', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52820461, end_position_on_the_genomic_accession=52821528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338674, CurrentID=0, Status='live', Symbol='OR5F1', Aliases='OR11-10', description='olfactory receptor family 5 subfamily F member 1', other_designations='olfactory receptor 11-10|olfactory receptor OR11-167|olfactory receptor, family 5, subfamily F, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55993681, end_position_on_the_genomic_accession=55994625, orientation='minus', exon_count=1, OMIM=608492, no_hdr0=''),
  338675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338675, CurrentID=0, Status='live', Symbol='OR5AP2', Aliases='OR9J1', description='olfactory receptor family 5 subfamily AP member 2', other_designations='olfactory receptor, family 5, subfamily AP, member 2|olfactory receptor, family 9, subfamily J, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56641489, end_position_on_the_genomic_accession=56642439, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372798, CurrentID=0, Status='live', Symbol='LOC105372798', Aliases='', description='uncharacterized LOC105372798', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37563973, end_position_on_the_genomic_accession=37570518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372799, CurrentID=0, Status='live', Symbol='LOC105372799', Aliases='', description='uncharacterized LOC105372799', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37715917, end_position_on_the_genomic_accession=37739035, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927191, CurrentID=0, Status='live', Symbol='LOC101927191', Aliases='', description='uncharacterized LOC101927191', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72557604, end_position_on_the_genomic_accession=72577348, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372800, CurrentID=0, Status='live', Symbol='LOC105372800', Aliases='', description='uncharacterized LOC105372800', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38248949, end_position_on_the_genomic_accession=38256605, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338692, CurrentID=0, Status='live', Symbol='ANKRD13D', Aliases='', description='ankyrin repeat domain 13 family member D', other_designations='ankyrin repeat domain 13 family, member D', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67289291, end_position_on_the_genomic_accession=67302484, orientation='plus', exon_count=17, OMIM=615126, no_hdr0=''),
  338694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338694, CurrentID=0, Status='live', Symbol='LOC338694', Aliases='', description='uncharacterized LOC338694', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69147228, end_position_on_the_genomic_accession=69171562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129921, CurrentID=0, Status='live', Symbol='RPL15P4', Aliases='Em:AC006046.2, RPL15_4_680', description='ribosomal protein L15 pseudogene 4', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31528076, end_position_on_the_genomic_accession=31528721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338699, CurrentID=0, Status='live', Symbol='ANKRD42', Aliases='PPP1R79, SARP', description='ankyrin repeat domain 42', other_designations='protein phosphatase 1, regulatory subunit 79|several ankyrin repeat protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83193739, end_position_on_the_genomic_accession=83256099, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  646274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646274, CurrentID=0, Status='live', Symbol='LOC646274', Aliases='', description='stathmin 2 pseudogene', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154706479, end_position_on_the_genomic_accession=154710045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338707, CurrentID=0, Status='live', Symbol='B4GALNT4', Aliases='', description='beta-1,4-N-acetyl-galactosaminyl transferase 4', other_designations='NGalNAc-T1|beta-1,4-N-acetylgalactosaminyltransferase IV|beta1,4-N-acetylgalactosaminyltransferases IV|beta4GalNAc-T4|beta4GalNAcT4|betaGT4', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=369779, end_position_on_the_genomic_accession=382117, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105372804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372804, CurrentID=0, Status='live', Symbol='LOC105372804', Aliases='', description='uncharacterized LOC105372804', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39349998, end_position_on_the_genomic_accession=39378559, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131556, CurrentID=0, Status='live', Symbol='LOC100131556', Aliases='', description='zinc finger CCHC-type and RNA binding motif 1 pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71314912, end_position_on_the_genomic_accession=71315879, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129926, CurrentID=0, Status='live', Symbol='RPL12P13', Aliases='', description='ribosomal protein L12 pseudogene 13', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26980080, end_position_on_the_genomic_accession=26981023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927108, CurrentID=0, Status='live', Symbol='LOC101927108', Aliases='', description='uncharacterized LOC101927108', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72030060, end_position_on_the_genomic_accession=72090328, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105372808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372808, CurrentID=0, Status='live', Symbol='LOC105372808', Aliases='', description='uncharacterized LOC105372808', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41576141, end_position_on_the_genomic_accession=41581627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338739, CurrentID=0, Status='live', Symbol='CSTF3-AS1', Aliases='', description='CSTF3 antisense RNA 1 (head to head)', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33161657, end_position_on_the_genomic_accession=33191598, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129929, CurrentID=0, Status='live', Symbol='NUP210P2', Aliases='', description='nucleoporin 210kDa pseudogene 2', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11900011, end_position_on_the_genomic_accession=11901245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927193, CurrentID=0, Status='live', Symbol='CHL1-AS1', Aliases='', description='CHL1 antisense RNA 1', other_designations='CHL1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=363373, end_position_on_the_genomic_accession=385795, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129930, CurrentID=0, Status='live', Symbol='RPL10P10', Aliases='RPL10_3_705', description='ribosomal protein L10 pseudogene 10', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54602589, end_position_on_the_genomic_accession=54603200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338751, CurrentID=0, Status='live', Symbol='OR52L1', Aliases='OR11-50', description='olfactory receptor family 52 subfamily L member 1', other_designations='olfactory receptor OR11-50|olfactory receptor, family 52, subfamily L, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5985892, end_position_on_the_genomic_accession=5986985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  338755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338755, CurrentID=0, Status='live', Symbol='OR2AG2', Aliases='OR11-76P, OR2AG2', description='olfactory receptor family 2 subfamily AG member 2', other_designations='olfactory receptor OR11-76 pseudogene|olfactory receptor, family 2, subfamily AG, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6767345, end_position_on_the_genomic_accession=6926426, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  338758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338758, CurrentID=0, Status='live', Symbol='LINC00936', Aliases='', description='long intergenic non-protein coding RNA 936', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89708955, end_position_on_the_genomic_accession=89711952, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  338761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338761, CurrentID=0, Status='live', Symbol='C1QL4', Aliases='C1QTNF11, CTRP11', description='complement component 1, q subcomponent-like 4', other_designations='C1q and tumor necrosis factor-related protein 11|C1q/TNF-related protein 11', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49332416, end_position_on_the_genomic_accession=49337188, orientation='minus', exon_count=3, OMIM=615229, no_hdr0=''),
  105372813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372813, CurrentID=0, Status='live', Symbol='LOC105372813', Aliases='', description='uncharacterized LOC105372813', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41668949, end_position_on_the_genomic_accession=41672645, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338773, CurrentID=0, Status='live', Symbol='TMEM119', Aliases='OBIF', description='transmembrane protein 119', other_designations='osteoblast induction factor', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108589846, end_position_on_the_genomic_accession=108598118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873520, CurrentID=0, Status='live', Symbol='RNA5SP267', Aliases='RN5S267', description='RNA, 5S ribosomal pseudogene 267', other_designations='RNA, 5S ribosomal 267', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=59455885, end_position_on_the_genomic_accession=59456010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129935, CurrentID=0, Status='live', Symbol='LOC100129935', Aliases='', description='lectin, galactoside-binding, soluble, 14 pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39638685, end_position_on_the_genomic_accession=39642401, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  731996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=731996, CurrentID=0, Status='live', Symbol='LOC731996', Aliases='', description='uncharacterized LOC731996', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  338785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338785, CurrentID=0, Status='live', Symbol='KRT79', Aliases='K6L, KRT6L', description='keratin 79, type II', other_designations='CK-79|K79|cytokeratin-79|keratin 6L|keratin-6L|keratin-79|type-II keratin Kb38', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52821447, end_position_on_the_genomic_accession=52834293, orientation='minus', exon_count=9, OMIM=611160, no_hdr0=''),
  105372817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372817, CurrentID=0, Status='live', Symbol='LOC105372817', Aliases='', description='uncharacterized LOC105372817', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42879667, end_position_on_the_genomic_accession=42887737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338797, CurrentID=0, Status='live', Symbol='LOC338797', Aliases='', description='uncharacterized LOC338797', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131347470, end_position_on_the_genomic_accession=131367555, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  338799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338799, CurrentID=0, Status='live', Symbol='LINC01089', Aliases='', description='long intergenic non-protein coding RNA 1089', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121795267, end_position_on_the_genomic_accession=121803484, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372819, CurrentID=0, Status='live', Symbol='LOC105372819', Aliases='', description='uncharacterized LOC105372819', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42947362, end_position_on_the_genomic_accession=42956065, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338805, CurrentID=0, Status='live', Symbol='ST13P8', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 8', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57983795, end_position_on_the_genomic_accession=57984761, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927195, CurrentID=0, Status='live', Symbol='LINC00382', Aliases='TCONS_00021500', description='long intergenic non-protein coding RNA 382', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79872586, end_position_on_the_genomic_accession=79918036, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  338809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338809, CurrentID=0, Status='live', Symbol='C12orf74', Aliases='', description='chromosome 12 open reading frame 74', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92702843, end_position_on_the_genomic_accession=92708549, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338811, CurrentID=0, Status='live', Symbol='FAM19A2', Aliases='TAFA-2, TAFA2', description='family with sequence similarity 19 (chemokine (C-C motif)-like), member A2', other_designations='chemokine-like protein TAFA-2', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61708248, end_position_on_the_genomic_accession=62260064, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105372821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372821, CurrentID=0, Status='live', Symbol='LOC105372821', Aliases='', description='uncharacterized LOC105372821', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43182044, end_position_on_the_genomic_accession=43184143, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338817, CurrentID=0, Status='live', Symbol='LINC01252', Aliases='', description='long intergenic non-protein coding RNA 1252', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11548030, end_position_on_the_genomic_accession=11564403, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  338821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338821, CurrentID=0, Status='live', Symbol='SLCO1B7', Aliases='LST-3, LST-3TM12, LST3, SLC21A21', description='solute carrier organic anion transporter family member 1B7 (non-functional)', other_designations='liver-specific organic anion transporter 3|liver-specific organic anion transporter 3TM12|organic anion transporter LST-3b|solute carrier organic anion transporter family, member 1B7 (non-functional)', map_location='12p12.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21015696, end_position_on_the_genomic_accession=21090106, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100421436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421436, CurrentID=0, Status='live', Symbol='UBE2FP1', Aliases='', description='ubiquitin conjugating enzyme E2F (putative) pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2F (putative) pseudogene 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37141545, end_position_on_the_genomic_accession=37144182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927196, CurrentID=0, Status='live', Symbol='LOC101927196', Aliases='', description='uncharacterized LOC101927196', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185719874, end_position_on_the_genomic_accession=185740477, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100739996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100739996, CurrentID=0, Status='live', Symbol='MRXS17', Aliases='', description='Mental retardation, X-linked, syndromic 17', other_designations='', map_location='Xp21.1-p11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300858, no_hdr0=''),
  105372827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372827, CurrentID=0, Status='live', Symbol='LOC105372827', Aliases='', description='uncharacterized LOC105372827', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43899369, end_position_on_the_genomic_accession=43902738, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421437, CurrentID=0, Status='live', Symbol='LOC100421437', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125344139, end_position_on_the_genomic_accession=125344797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372828, CurrentID=0, Status='live', Symbol='LOC105372828', Aliases='', description='uncharacterized LOC105372828', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43884323, end_position_on_the_genomic_accession=43889917, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338862, CurrentID=0, Status='live', Symbol='LINC00550', Aliases='', description='long intergenic non-protein coding RNA 550', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68861284, end_position_on_the_genomic_accession=68885325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129949, CurrentID=0, Status='live', Symbol='LOC100129949', Aliases='', description='protein RCC2-like', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242882054, end_position_on_the_genomic_accession=242883531, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338864, CurrentID=0, Status='live', Symbol='LINC00347', Aliases='', description='long intergenic non-protein coding RNA 347', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74552843, end_position_on_the_genomic_accession=74557120, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927197, CurrentID=0, Status='live', Symbol='LOC101927197', Aliases='', description='cell death activator CIDE-3-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128869405, end_position_on_the_genomic_accession=128871478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338870, CurrentID=0, Status='live', Symbol='RPS12P23', Aliases='RPS12_11_1301', description='ribosomal protein S12 pseudogene 23', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21094295, end_position_on_the_genomic_accession=21094795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  338872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338872, CurrentID=0, Status='live', Symbol='C1QTNF9', Aliases='AQL1A, CTRP9, C1QTNF9', description='C1q and tumor necrosis factor related protein 9', other_designations='complement C1q and tumor necrosis factor-related protein 9|complement C1q tumor necrosis factor-related protein 9|complement C1q tumor necrosis factor-related protein 9A', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24307166, end_position_on_the_genomic_accession=24322531, orientation='plus', exon_count=6, OMIM=614285, no_hdr0=''),
  338879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338879, CurrentID=0, Status='live', Symbol='RNASE10', Aliases='RAH1, RNASE9', description='ribonuclease, RNase A family, 10 (non-active)', other_designations='ribonuclease A H1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20510472, end_position_on_the_genomic_accession=20511122, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372832, CurrentID=0, Status='live', Symbol='LOC105372832', Aliases='', description='uncharacterized LOC105372832', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44201798, end_position_on_the_genomic_accession=44208199, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100421438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421438, CurrentID=0, Status='live', Symbol='LOC100421438', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C8 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136736496, end_position_on_the_genomic_accession=136737228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927198, CurrentID=0, Status='live', Symbol='LOC101927198', Aliases='', description='uncharacterized LOC101927198', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32827883, end_position_on_the_genomic_accession=32860890, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372835, CurrentID=0, Status='live', Symbol='LOC105372835', Aliases='', description='uncharacterized LOC105372835', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45091704, end_position_on_the_genomic_accession=45101260, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928855, CurrentID=0, Status='live', Symbol='LOC101928855', Aliases='', description='uncharacterized LOC101928855', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80801640, end_position_on_the_genomic_accession=80805632, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372836, CurrentID=0, Status='live', Symbol='LOC105372836', Aliases='', description='uncharacterized LOC105372836', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45055877, end_position_on_the_genomic_accession=45073903, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373312, CurrentID=0, Status='live', Symbol='LOC105373312', Aliases='', description='uncharacterized LOC105373312', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110318268, end_position_on_the_genomic_accession=110333372, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271221, CurrentID=0, Status='live', Symbol='RPL7P49', Aliases='RPL7_20_1575', description='ribosomal protein L7 pseudogene 49', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76430351, end_position_on_the_genomic_accession=76431144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372837, CurrentID=0, Status='live', Symbol='LOC105372837', Aliases='', description='uncharacterized LOC105372837', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45309459, end_position_on_the_genomic_accession=45311610, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338917, CurrentID=0, Status='live', Symbol='VSX2', Aliases='CHX10, HOX10, MCOP2, MCOPCB3, RET1', description='visual system homeobox 2', other_designations='ceh-10 homeo domain containing homolog|ceh-10 homeodomain-containing homolog|homeobox protein CHX10', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74239472, end_position_on_the_genomic_accession=74262738, orientation='plus', exon_count=5, OMIM=142993, no_hdr0=''),
  105372839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372839, CurrentID=0, Status='live', Symbol='LOC105372839', Aliases='', description='uncharacterized LOC105372839', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45347382, end_position_on_the_genomic_accession=45353536, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375170, CurrentID=0, Status='live', Symbol='LOC105375170', Aliases='', description='uncharacterized LOC105375170', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16988840, end_position_on_the_genomic_accession=17021616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129960, CurrentID=0, Status='live', Symbol='SDCBPP2', Aliases='', description='syndecan binding protein (syntenin) pseudogene 2', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69942921, end_position_on_the_genomic_accession=69944286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372841, CurrentID=0, Status='live', Symbol='LOC105372841', Aliases='', description='uncharacterized LOC105372841', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45978494, end_position_on_the_genomic_accession=45981570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129017, CurrentID=0, Status='live', Symbol='RPL24P5', Aliases='RPL24_1_143', description='ribosomal protein L24 pseudogene 5', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197735604, end_position_on_the_genomic_accession=197736149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  338949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338949, CurrentID=0, Status='live', Symbol='TMEM202', Aliases='', description='transmembrane protein 202', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72398101, end_position_on_the_genomic_accession=72408367, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372844, CurrentID=0, Status='live', Symbol='LOC105372844', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927200, CurrentID=0, Status='live', Symbol='LOC101927200', Aliases='', description='uncharacterized LOC101927200', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43727395, end_position_on_the_genomic_accession=43826432, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  338963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=338963, CurrentID=0, Status='live', Symbol='LOC338963', Aliases='', description='epididymal protein pseudogene', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82710471, end_position_on_the_genomic_accession=82713993, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129966, CurrentID=0, Status='live', Symbol='LOC100129966', Aliases='', description='mitochondrial carrier 1 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18594592, end_position_on_the_genomic_accession=18596404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376804, CurrentID=0, Status='live', Symbol='LOC105376804', Aliases='', description='uncharacterized LOC105376804', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373248, CurrentID=0, Status='live', Symbol='LOC105373248', Aliases='', description='uncharacterized LOC105373248', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71337701, end_position_on_the_genomic_accession=71350152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372450, CurrentID=0, Status='live', Symbol='LOC105372450', Aliases='', description='uncharacterized LOC105372450', other_designations='CTD-2525I3.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52225969, end_position_on_the_genomic_accession=52231294, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  732199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732199, CurrentID=0, Status='live', Symbol='RPL29P21', Aliases='RPL29_9_1124', description='ribosomal protein L29 pseudogene 21', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17193444, end_position_on_the_genomic_accession=17194427, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  732210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732210, CurrentID=0, Status='live', Symbol='LOC732210', Aliases='', description='uncharacterized LOC732210', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129972, CurrentID=0, Status='live', Symbol='NPM1P47', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 47', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62082249, end_position_on_the_genomic_accession=62083399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339005, CurrentID=0, Status='live', Symbol='WHAMMP3', Aliases='WHAMML1, WHDC1L1', description='WAS protein homolog associated with actin, golgi membranes and microtubules pseudogene 3', other_designations='WAS protein homolog associated with actin, golgi membranes and microtubules-like 1 (pseudogene)|WAS protein homology region 2 domain containing 1 pseudogene|WAS protein homology region 2 domain containing 1-like 1', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22664739, end_position_on_the_genomic_accession=22685368, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105372853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372853, CurrentID=0, Status='live', Symbol='LOC105372853', Aliases='', description='uncharacterized LOC105372853', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18056787, end_position_on_the_genomic_accession=18073113, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  732226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732226, CurrentID=0, Status='live', Symbol='LOC732226', Aliases='', description='uncharacterized LOC732226', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.7', start_position_on_the_genomic_accession=40021498, end_position_on_the_genomic_accession=40024532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372854, CurrentID=0, Status='live', Symbol='LOC105372854', Aliases='', description='uncharacterized LOC105372854', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372855, CurrentID=0, Status='live', Symbol='LOC105372855', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18996513, end_position_on_the_genomic_accession=18998725, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646471, CurrentID=0, Status='live', Symbol='LOC646471', Aliases='', description='uncharacterized LOC646471', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25819954, end_position_on_the_genomic_accession=25823606, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129020, CurrentID=0, Status='live', Symbol='LOC100129020', Aliases='', description='uncharacterized LOC100129020', other_designations='', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=114491852, end_position_on_the_genomic_accession=114503326, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  732253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732253, CurrentID=0, Status='live', Symbol='TDRG1', Aliases='LINC00532', description='testis development related 1 (non-protein coding)', other_designations='lincRNA-NR_024015|long intergenic non-protein coding RNA 532|testis development related protein 1', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40378424, end_position_on_the_genomic_accession=40379892, orientation='plus', exon_count=2, OMIM=615676, no_hdr0=''),
  339039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339039, CurrentID=0, Status='live', Symbol='PKD1P3', Aliases='HG3', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 3', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14911477, end_position_on_the_genomic_accession=14935708, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060322, CurrentID=0, Status='live', Symbol='LOC101060322', Aliases='', description='telomeric repeat-binding factor 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  339044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339044, CurrentID=0, Status='live', Symbol='PKD1P1', Aliases='HG1, NPIP', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 1', other_designations='KIAA0220-like protein|NPIP-like protein 1|nuclear pore complex-interacting protein', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16317609, end_position_on_the_genomic_accession=16350608, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  732265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732265, CurrentID=0, Status='live', Symbol='LOC732265', Aliases='', description='uncharacterized LOC732265', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74088446, end_position_on_the_genomic_accession=74100467, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  732272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732272, CurrentID=0, Status='live', Symbol='LOC732272', Aliases='', description='uncharacterized LOC732272', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  732275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732275, CurrentID=0, Status='live', Symbol='LINC00917', Aliases='', description='long intergenic non-protein coding RNA 917', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86331850, end_position_on_the_genomic_accession=86345679, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100129982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129982, CurrentID=0, Status='live', Symbol='RPL12P19', Aliases='RPL12_2_212', description='ribosomal protein L12 pseudogene 19', other_designations='60S ribosomal protein L12 pseudogene', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44270533, end_position_on_the_genomic_accession=44271155, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421444, CurrentID=0, Status='live', Symbol='LOC100421444', Aliases='', description='glucosamine-phosphate N-acetyltransferase 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84828576, end_position_on_the_genomic_accession=84829479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372863, CurrentID=0, Status='live', Symbol='LOC105372863', Aliases='', description='uncharacterized LOC105372863', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20226945, end_position_on_the_genomic_accession=20235955, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375175, CurrentID=0, Status='live', Symbol='LOC105375175', Aliases='', description='uncharacterized LOC105375175', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18999723, end_position_on_the_genomic_accession=19106636, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927204, CurrentID=0, Status='live', Symbol='LOC101927204', Aliases='', description='uncharacterized LOC101927204', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59133972, end_position_on_the_genomic_accession=59142808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129989, CurrentID=0, Status='live', Symbol='LOC100129989', Aliases='', description='chromosome 12 open reading frame 29 pseudogene', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109396613, end_position_on_the_genomic_accession=109397867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339105, CurrentID=0, Status='live', Symbol='PRSS53', Aliases='POL3S, UNQ308', description='protease, serine 53', other_designations='EDTP308|polyserase 3|polyserase-3|polyserine protease 3|protease, serine, 53', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31083424, end_position_on_the_genomic_accession=31090035, orientation='minus', exon_count=11, OMIM=610561, no_hdr0=''),
  105372871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372871, CurrentID=0, Status='live', Symbol='LOC105372871', Aliases='', description='uncharacterized LOC105372871', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18634541, end_position_on_the_genomic_accession=18635687, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372872, CurrentID=0, Status='live', Symbol='LOC105372872', Aliases='', description='predicted GPI-anchored protein 58', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  339122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339122, CurrentID=0, Status='live', Symbol='RAB43', Aliases='RAB11B, RAB41', description='RAB43, member RAS oncogene family', other_designations='ras-related protein Rab-41', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129087569, end_position_on_the_genomic_accession=129122150, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  339123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339123, CurrentID=0, Status='live', Symbol='JMJD8', Aliases='C16orf20, PP14397', description='jumonji domain containing 8', other_designations='jumonji domain-containing protein 8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=681667, end_position_on_the_genomic_accession=684441, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100421446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421446, CurrentID=0, Status='live', Symbol='LOC100421446', Aliases='', description='Wolf-Hirschhorn syndrome candidate 1-like 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11998830, end_position_on_the_genomic_accession=11999505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56521, CurrentID=0, Status='live', Symbol='DNAJC12', Aliases='JDP1', description='DnaJ heat shock protein family (Hsp40) member C12', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 12|J domain containing protein 1 (JDP1)|J domain protein 1|j domain-containing protein 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67796665, end_position_on_the_genomic_accession=67838191, orientation='minus', exon_count=6, OMIM=606060, no_hdr0=''),
  646346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646346, CurrentID=0, Status='live', Symbol='RPL7P58', Aliases='', description='ribosomal protein L7 pseudogene 58', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26687363, end_position_on_the_genomic_accession=26688023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216001, CurrentID=0, Status='live', Symbol='LINC00704', Aliases='', description='long intergenic non-protein coding RNA 704', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4650185, end_position_on_the_genomic_accession=4678070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129995, CurrentID=0, Status='live', Symbol='LINC01460', Aliases='', description='long intergenic non-protein coding RNA 1460', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27707503, end_position_on_the_genomic_accession=27716550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339145, CurrentID=0, Status='live', Symbol='FAM92B', Aliases='', description='family with sequence similarity 92 member B', other_designations='family with sequence similarity 92, member B', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85098352, end_position_on_the_genomic_accession=85112508, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129997, CurrentID=0, Status='live', Symbol='LOC100129997', Aliases='', description='spindlin family member 2B pseudogene', other_designations='spindlin family, member 2B pseudogene', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57068749, end_position_on_the_genomic_accession=57069685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372878, CurrentID=0, Status='live', Symbol='LOC105372878', Aliases='', description='uncharacterized LOC105372878', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206841101, end_position_on_the_genomic_accession=206842821, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129999, CurrentID=0, Status='live', Symbol='LOC100129999', Aliases='', description='uncharacterized LOC100129999', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=139014367, end_position_on_the_genomic_accession=139016209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  339166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339166, CurrentID=0, Status='live', Symbol='LOC339166', Aliases='', description='uncharacterized LOC339166', other_designations='', map_location='17p13.3-p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5772234, end_position_on_the_genomic_accession=5930696, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339168, CurrentID=0, Status='live', Symbol='TMEM95', Aliases='UNQ9390', description='transmembrane protein 95', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7355134, end_position_on_the_genomic_accession=7357219, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101927207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927207, CurrentID=0, Status='live', Symbol='LOC101927207', Aliases='', description='uncharacterized LOC101927207', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49457861, end_position_on_the_genomic_accession=49461749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339175, CurrentID=0, Status='live', Symbol='METTL2A', Aliases='METTL2', description='methyltransferase like 2A', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62423885, end_position_on_the_genomic_accession=62450093, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105372882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372882, CurrentID=0, Status='live', Symbol='LOC105372882', Aliases='', description='uncharacterized LOC105372882', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207873449, end_position_on_the_genomic_accession=207880102, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  339184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339184, CurrentID=0, Status='live', Symbol='CCDC144NL', Aliases='', description='coiled-coil domain containing 144 family, N-terminal like', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20847348, end_position_on_the_genomic_accession=20896647, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  339186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339186, CurrentID=0, Status='live', Symbol='KRT17P4', Aliases='', description='keratin 17 pseudogene 4', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16847400, end_position_on_the_genomic_accession=16850552, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372883, CurrentID=0, Status='live', Symbol='LOC105372883', Aliases='', description='uncharacterized LOC105372883', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207962873, end_position_on_the_genomic_accession=207967460, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339192, CurrentID=0, Status='live', Symbol='LOC339192', Aliases='', description='uncharacterized LOC339192', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45219313, end_position_on_the_genomic_accession=45219974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339201, CurrentID=0, Status='live', Symbol='ASB16-AS1', Aliases='C17orf65', description='ASB16 antisense RNA 1', other_designations='CTB-175E5.4', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44175973, end_position_on_the_genomic_accession=44186717, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646358, CurrentID=0, Status='live', Symbol='LOC646358', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B14 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 14 pseudogene', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65770450, end_position_on_the_genomic_accession=65772154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339210, CurrentID=0, Status='live', Symbol='C17orf67', Aliases='', description='chromosome 17 open reading frame 67', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56791913, end_position_on_the_genomic_accession=56834062, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  646359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646359, CurrentID=0, Status='live', Symbol='TERF1P2', Aliases='', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene 2', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14375942, end_position_on_the_genomic_accession=14378757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372888, CurrentID=0, Status='live', Symbol='LOC105372888', Aliases='', description='uncharacterized LOC105372888', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  339221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339221, CurrentID=0, Status='live', Symbol='ENPP7', Aliases='ALK-SMase, E-NPP 7, NPP-7, NPP7', description='ectonucleotide pyrophosphatase/phosphodiesterase 7', other_designations='alkaline sphingomyelin phosphodiesterase|alkaline sphingomyelinase|intestinal alkaline sphingomyelinase', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79730919, end_position_on_the_genomic_accession=79742926, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372889, CurrentID=0, Status='live', Symbol='LOC105372889', Aliases='', description='uncharacterized LOC105372889', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208252744, end_position_on_the_genomic_accession=208254596, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339229, CurrentID=0, Status='live', Symbol='OXLD1', Aliases='C17orf90', description='oxidoreductase-like domain containing 1', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81665036, end_position_on_the_genomic_accession=81666637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339230, CurrentID=0, Status='live', Symbol='CCDC137', Aliases='RaRF', description='coiled-coil domain containing 137', other_designations='hepatocellular carcinoma related protein 2|retinoic acid resistance factor', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81666731, end_position_on_the_genomic_accession=81673907, orientation='plus', exon_count=7, OMIM=614271, no_hdr0=''),
  339231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339231, CurrentID=0, Status='live', Symbol='ARL16', Aliases='', description='ADP ribosylation factor like GTPase 16', other_designations='ADP-ribosylation factor like GTPase 16|ADP-ribosylation factor-like 16', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81681155, end_position_on_the_genomic_accession=81683924, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  732448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732448, CurrentID=0, Status='live', Symbol='CCDC58P3', Aliases='', description='coiled-coil domain containing 58 pseudogene 3', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12211373, end_position_on_the_genomic_accession=12212007, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130011, CurrentID=0, Status='live', Symbol='LOC100130011', Aliases='', description='synaptophysin-like 1 pseudogene', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154651673, end_position_on_the_genomic_accession=154652375, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339240, CurrentID=0, Status='live', Symbol='KRT17P5', Aliases='', description='keratin 17 pseudogene 5', other_designations='keratin 14 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18422181, end_position_on_the_genomic_accession=18425333, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  339241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339241, CurrentID=0, Status='live', Symbol='KRT17P2', Aliases='', description='keratin 17 pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18426806, end_position_on_the_genomic_accession=18430780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339244, CurrentID=0, Status='live', Symbol='KRT16P4', Aliases='', description='keratin 16 pseudogene 4', other_designations='keratin pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18449953, end_position_on_the_genomic_accession=18451359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372893, CurrentID=0, Status='live', Symbol='LOC105372893', Aliases='', description='uncharacterized LOC105372893', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208728668, end_position_on_the_genomic_accession=208736774, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130014, CurrentID=0, Status='live', Symbol='SETP5', Aliases='', description='SET pseudogene 5', other_designations='SET nuclear oncogene pseudogene', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137641011, end_position_on_the_genomic_accession=137641572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339256, CurrentID=0, Status='live', Symbol='NOS2P3', Aliases='', description='nitric oxide synthase 2 pseudogene 3', other_designations='nitric oxide synthase 2, inducible pseudogene|nitric oxide synthase 2A pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20442786, end_position_on_the_genomic_accession=20447249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339257, CurrentID=0, Status='live', Symbol='TNPO1P3', Aliases='', description='transportin 1 pseudogene 3', other_designations='karyopherin (importin) beta 2 pseudogene|transportin 2 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20481060, end_position_on_the_genomic_accession=20484012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339258, CurrentID=0, Status='live', Symbol='KRT17P7', Aliases='', description='keratin 17 pseudogene 7', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20519073, end_position_on_the_genomic_accession=20522228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927210, CurrentID=0, Status='live', Symbol='LOC101927210', Aliases='', description='uncharacterized LOC101927210', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28491864, end_position_on_the_genomic_accession=28632728, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339260, CurrentID=0, Status='live', Symbol='LOC339260', Aliases='', description='uncharacterized LOC339260', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20938520, end_position_on_the_genomic_accession=20982357, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  339263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339263, CurrentID=0, Status='live', Symbol='C17orf51', Aliases='', description='chromosome 17 open reading frame 51', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21439840, end_position_on_the_genomic_accession=21551678, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372896, CurrentID=0, Status='live', Symbol='LOC105372896', Aliases='', description='uncharacterized LOC105372896', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209211769, end_position_on_the_genomic_accession=209266726, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372897, CurrentID=0, Status='live', Symbol='LOC105372897', Aliases='', description='uncharacterized LOC105372897', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209325438, end_position_on_the_genomic_accession=209328532, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421451, CurrentID=0, Status='live', Symbol='HNRNPA3P16', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 16', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50814388, end_position_on_the_genomic_accession=50815539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372898, CurrentID=0, Status='live', Symbol='LOC105372898', Aliases='', description='uncharacterized LOC105372898', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209515969, end_position_on_the_genomic_accession=209542918, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372899, CurrentID=0, Status='live', Symbol='LOC105372899', Aliases='', description='uncharacterized LOC105372899', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209566948, end_position_on_the_genomic_accession=209567802, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339287, CurrentID=0, Status='live', Symbol='MSL1', Aliases='MSL-1, hMSL1', description='male-specific lethal 1 homolog (Drosophila)', other_designations='MSL1-like 1|male-specific lethal 1-like 1|male-specific lethal-1 homolog 1', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40121988, end_position_on_the_genomic_accession=40136794, orientation='plus', exon_count=11, OMIM=614801, no_hdr0=''),
  101927211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927211, CurrentID=0, Status='live', Symbol='LOC101927211', Aliases='', description='uncharacterized LOC101927211', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57114911, end_position_on_the_genomic_accession=57172213, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339290, CurrentID=0, Status='live', Symbol='LINC00667', Aliases='', description='long intergenic non-protein coding RNA 667', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5238100, end_position_on_the_genomic_accession=5246508, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339291, CurrentID=0, Status='live', Symbol='LRRC30', Aliases='', description='leucine rich repeat containing 30', other_designations='', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7231139, end_position_on_the_genomic_accession=7232044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  339298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339298, CurrentID=0, Status='live', Symbol='LOC339298', Aliases='', description='uncharacterized LOC339298', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76122998, end_position_on_the_genomic_accession=76145255, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129029, CurrentID=0, Status='live', Symbol='LOC100129029', Aliases='', description='uncharacterized LOC100129029', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158678656, end_position_on_the_genomic_accession=158735002, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339302, CurrentID=0, Status='live', Symbol='CPLX4', Aliases='CPX-IV, CPXIV', description='complexin 4', other_designations='CPX IV|complexin IV', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59295402, end_position_on_the_genomic_accession=59318649, orientation='minus', exon_count=3, OMIM=609586, no_hdr0=''),
  339318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339318, CurrentID=0, Status='live', Symbol='ZNF181', Aliases='HHZ181', description='zinc finger protein 181', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34733316, end_position_on_the_genomic_accession=34742872, orientation='plus', exon_count=5, OMIM=606741, no_hdr0=''),
  105372905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372905, CurrentID=0, Status='live', Symbol='LOC105372905', Aliases='', description='uncharacterized LOC105372905', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212180146, end_position_on_the_genomic_accession=212182456, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  732538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=732538, CurrentID=0, Status='live', Symbol='FBXO36P1', Aliases='', description='F-box protein 36 pseudogene 1', other_designations='F-box only protein 36 pseudogene', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68095069, end_position_on_the_genomic_accession=68096490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339324, CurrentID=0, Status='live', Symbol='ZNF260', Aliases='OZRF1, PEX1, ZFP260', description='zinc finger protein 260', other_designations='zfp-260', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36510687, end_position_on_the_genomic_accession=36528346, orientation='minus', exon_count=5, OMIM=613749, no_hdr0=''),
  105372906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372906, CurrentID=0, Status='live', Symbol='LOC105372906', Aliases='', description='uncharacterized LOC105372906', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212429962, end_position_on_the_genomic_accession=212432791, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339327, CurrentID=0, Status='live', Symbol='ZNF546', Aliases='ZNF49', description='zinc finger protein 546', other_designations='CTC-471F3.6|zinc finger protein 49', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39995630, end_position_on_the_genomic_accession=40021041, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105372907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372907, CurrentID=0, Status='live', Symbol='LOC105372907', Aliases='', description='uncharacterized LOC105372907', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212466816, end_position_on_the_genomic_accession=212472865, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060324, CurrentID=0, Status='live', Symbol='LOC101060324', Aliases='', description='espin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372909, CurrentID=0, Status='live', Symbol='LOC105372909', Aliases='', description='uncharacterized LOC105372909', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212634181, end_position_on_the_genomic_accession=212638538, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339344, CurrentID=0, Status='live', Symbol='MYPOP', Aliases='P42pop', description='Myb-related transcription factor, partner of profilin', other_designations='myb-related protein p42POP|p42 Myb-related transcription factor, partner of profilin|partner of profilin', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45890023, end_position_on_the_genomic_accession=45902696, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  339345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339345, CurrentID=0, Status='live', Symbol='NANOS2', Aliases='NOS2', description='nanos homolog 2 (Drosophila)', other_designations='NOS-2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45913214, end_position_on_the_genomic_accession=45914778, orientation='minus', exon_count=1, OMIM=608228, no_hdr0=''),
  105372910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372910, CurrentID=0, Status='live', Symbol='LOC105372910', Aliases='', description='uncharacterized LOC105372910', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213492206, end_position_on_the_genomic_accession=213578880, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339352, CurrentID=0, Status='live', Symbol='LOC339352', Aliases='', description='cytosolic thiouridylase subunit 1 homolog (S. pombe) pseudogene', other_designations='Putative ATP-binding domain-containing protein 3-like protein', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51598544, end_position_on_the_genomic_accession=51601706, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646383, CurrentID=0, Status='live', Symbol='LOC646383', Aliases='', description='uncharacterized LOC646383', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130314990, end_position_on_the_genomic_accession=130446204, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100271224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271224, CurrentID=0, Status='live', Symbol='RPL7P43', Aliases='RPL7_23_1260', description='ribosomal protein L7 pseudogene 43', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76721518, end_position_on_the_genomic_accession=76722355, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339366, CurrentID=0, Status='live', Symbol='ADAMTSL5', Aliases='THSD6', description='ADAMTS like 5', other_designations='ADAMTS-like 5|ADAMTSL-5|thrombospondin type-1 domain-containing protein 6|thrombospondin, type I, domain containing 6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1505018, end_position_on_the_genomic_accession=1513189, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105372914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372914, CurrentID=0, Status='live', Symbol='LOC105372914', Aliases='', description='uncharacterized LOC105372914', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214416631, end_position_on_the_genomic_accession=214419468, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130035, CurrentID=0, Status='live', Symbol='LOC100130035', Aliases='', description='biogenesis of lysosomal organelles complex-1, subunit 6, pallidin pseudogene', other_designations='biogenesis of lysosome-related organelles complex 1, subunit 6, pallidin pseudogene', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18838655, end_position_on_the_genomic_accession=18853651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646388, CurrentID=0, Status='live', Symbol='LOC646388', Aliases='', description='zinc finger, CCHC domain containing 9 pseudogene', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=40083713, end_position_on_the_genomic_accession=40084517, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339390, CurrentID=0, Status='live', Symbol='CLEC4G', Aliases='DTTR431, LP2698, LSECtin, UNQ431', description='C-type lectin domain family 4 member G', other_designations='C-type lectin domain family 4, member G|C-type lectin superfamily 4, member G|liver and lymph node sinusoidal endothelial cell C-type lectin', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7728957, end_position_on_the_genomic_accession=7732171, orientation='minus', exon_count=9, OMIM=616256, no_hdr0=''),
  105372917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372917, CurrentID=0, Status='live', Symbol='LOC105372917', Aliases='', description='uncharacterized LOC105372917', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215186277, end_position_on_the_genomic_accession=215195022, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339396, CurrentID=0, Status='live', Symbol='NEFHP1', Aliases='NEFHL', description='neurofilament, heavy polypeptide pseudogene 1', other_designations='neurofilament, heavy polypeptide 200kDa (NEFH) pseudogene|neurofilament, heavy polypeptide-like pseudogene', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116739784, end_position_on_the_genomic_accession=116742652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646390, CurrentID=0, Status='live', Symbol='IRX1P1', Aliases='', description='iroquois homeobox 1 pseudogene 1', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24745266, end_position_on_the_genomic_accession=24746119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339398, CurrentID=0, Status='live', Symbol='LINGO4', Aliases='DAAT9248, LRRN6D, PRO34002', description='leucine rich repeat and Ig domain containing 4', other_designations='leucine-rich repeat neuronal protein 6D', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151800264, end_position_on_the_genomic_accession=151805442, orientation='minus', exon_count=2, OMIM=609794, no_hdr0=''),
  339400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339400, CurrentID=0, Status='live', Symbol='FLG-AS1', Aliases='', description='FLG antisense RNA 1', other_designations='FLG antisense RNA 1 (non-protein coding)', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152313459, end_position_on_the_genomic_accession=152366692, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  339403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339403, CurrentID=0, Status='live', Symbol='RXFP4', Aliases='GPCR142, GPR100, RLN3R2, RXFPR4', description='relaxin/insulin-like family peptide receptor 4', other_designations='G protein-coupled receptor 100|G-protein coupled receptor 100|G-protein coupled receptor GPCR142|RLN3 receptor 2|insulin-like peptide INSL5 receptor|relaxin 3 receptor 2|relaxin family peptide receptor 4', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155941689, end_position_on_the_genomic_accession=155942834, orientation='plus', exon_count=1, OMIM=609043, no_hdr0=''),
  105375186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375186, CurrentID=0, Status='live', Symbol='LOC105375186', Aliases='', description='uncharacterized LOC105375186', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23077790, end_position_on_the_genomic_accession=23099195, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927215, CurrentID=0, Status='live', Symbol='LINC01266', Aliases='', description='long intergenic non-protein coding RNA 1266', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=592105, end_position_on_the_genomic_accession=846015, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929353, CurrentID=0, Status='live', Symbol='LOC101929353', Aliases='', description='uncharacterized LOC101929353', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100190033, end_position_on_the_genomic_accession=100215505, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  339416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339416, CurrentID=0, Status='live', Symbol='ANKRD45', Aliases='CT117', description='ankyrin repeat domain 45', other_designations='cancer/testis antigen 117', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173608336, end_position_on_the_genomic_accession=173705267, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  106479410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479410, CurrentID=0, Status='live', Symbol='RN7SL510P', Aliases='', description='RNA, 7SL, cytoplasmic 510, pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372922, CurrentID=0, Status='live', Symbol='LOC105372922', Aliases='', description='uncharacterized LOC105372922', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217921582, end_position_on_the_genomic_accession=217944812, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130043, CurrentID=0, Status='live', Symbol='SETP17', Aliases='', description='SET pseudogene 17', other_designations='SET nuclear oncogene pseudogene', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86294788, end_position_on_the_genomic_accession=86295578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372852, CurrentID=0, Status='live', Symbol='LOC105372852', Aliases='', description='uncharacterized LOC105372852', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17866020, end_position_on_the_genomic_accession=17868003, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372924, CurrentID=0, Status='live', Symbol='LOC105372924', Aliases='', description='uncharacterized LOC105372924', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218698587, end_position_on_the_genomic_accession=218720622, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927216, CurrentID=0, Status='live', Symbol='LOC101927216', Aliases='', description='uncharacterized LOC101927216', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80041700, end_position_on_the_genomic_accession=80075306, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  339442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339442, CurrentID=0, Status='live', Symbol='LINC01343', Aliases='', description='long intergenic non-protein coding RNA 1343', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38209034, end_position_on_the_genomic_accession=38214767, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339448, CurrentID=0, Status='live', Symbol='C1orf174', Aliases='', description='chromosome 1 open reading frame 174', other_designations='RP13-531C17.2', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3889133, end_position_on_the_genomic_accession=3900293, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  339451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339451, CurrentID=0, Status='live', Symbol='KLHL17', Aliases='', description='kelch like family member 17', other_designations='actinfilin|kelch-like 17|kelch-like family member 17', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=959952, end_position_on_the_genomic_accession=965720, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  339453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339453, CurrentID=0, Status='live', Symbol='TMEM240', Aliases='C1orf70, SCA21', description='transmembrane protein 240', other_designations='transmembrane protein C1orf70', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1534778, end_position_on_the_genomic_accession=1540360, orientation='minus', exon_count=4, OMIM=616101, no_hdr0=''),
  339456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339456, CurrentID=0, Status='live', Symbol='TMEM52', Aliases='', description='transmembrane protein 52', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1917589, end_position_on_the_genomic_accession=1919301, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130048, CurrentID=0, Status='live', Symbol='LOC100130048', Aliases='', description='H1 histone family, member O, oocyte-specific pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76690140, end_position_on_the_genomic_accession=76691106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130049, CurrentID=0, Status='live', Symbol='LOC100130049', Aliases='', description='ribosomal L1 domain-containing protein 1-like', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85934575, end_position_on_the_genomic_accession=85936598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339468, CurrentID=0, Status='live', Symbol='LOC339468', Aliases='', description='uncharacterized LOC339468', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72765016, end_position_on_the_genomic_accession=72794882, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101927217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927217, CurrentID=0, Status='live', Symbol='LINC01473', Aliases='', description='long intergenic non-protein coding RNA 1473', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186033531, end_position_on_the_genomic_accession=186086317, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100216336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216336, CurrentID=0, Status='live', Symbol='RPS24P8', Aliases='RPS24_4_368', description='ribosomal protein S24 pseudogene 8', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45159698, end_position_on_the_genomic_accession=45160214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216337, CurrentID=0, Status='live', Symbol='LOC100216337', Aliases='', description='ADP-ribosylation factor like GTPase 1 pseudogene', other_designations='ADP-ribosylation factor-like 1 pseudogene', map_location='2p', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8756908, end_position_on_the_genomic_accession=8757866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216338, CurrentID=0, Status='live', Symbol='RPS24P13', Aliases='RPS24_7_1030', description='ribosomal protein S24 pseudogene 13', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32231780, end_position_on_the_genomic_accession=32232283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216339, CurrentID=0, Status='live', Symbol='RPS24P6', Aliases='RPS24_2_250', description='ribosomal protein S24 pseudogene 6', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94912411, end_position_on_the_genomic_accession=94912900, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339476, CurrentID=0, Status='live', Symbol='ERVMER61-1', Aliases='C1orf99', description='endogenous retrovirus group MER61 member 1', other_designations='endogenous retrovirus group MER61, member 1', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=185876981, end_position_on_the_genomic_accession=185879618, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100216341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216341, CurrentID=0, Status='live', Symbol='RPS24P17', Aliases='RPS24_7_1485', description='ribosomal protein S24 pseudogene 17', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56906498, end_position_on_the_genomic_accession=56907023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216342, CurrentID=0, Status='live', Symbol='RPS24P7', Aliases='RPS24_1_225', description='ribosomal protein S24 pseudogene 7', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61803026, end_position_on_the_genomic_accession=61803533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339479, CurrentID=0, Status='live', Symbol='BRINP3', Aliases='DBCCR1L, DBCCR1L1, FAM5C', description='bone morphogenetic protein/retinoic acid inducible neural-specific 3', other_designations='DBCCR1-like protein 1|family with sequence similarity 5, member C|protein FAM5C', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=190097662, end_position_on_the_genomic_accession=190478404, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100216344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216344, CurrentID=0, Status='live', Symbol='DYT17', Aliases='', description='dystonia 17', other_designations='', map_location='20p11.2-q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612406, no_hdr0=''),
  100130052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130052, CurrentID=0, Status='live', Symbol='LOC100130052', Aliases='', description='melanoma antigen family B3 pseudogene', other_designations='melanoma antigen family B, 3 pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26223652, end_position_on_the_genomic_accession=26224608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216346, CurrentID=0, Status='live', Symbol='LOC100216346', Aliases='', description='ADP-ribosylation factor like GTPase 6 interacting protein 6 pseudogene', other_designations='ADP-ribosylation factor-like 6 interacting protein 6 pseudogene|ADP-ribosylation-like factor 6 interacting protein 6 pseudogene', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80351116, end_position_on_the_genomic_accession=80351796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339483, CurrentID=0, Status='live', Symbol='MTMR9LP', Aliases='MTMR9L', description='myotubularin related protein 9-like, pseudogene', other_designations='myotubularin related protein 9 pseudogene', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32231660, end_position_on_the_genomic_accession=32241710, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  339487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339487, CurrentID=0, Status='live', Symbol='ZBTB8OS', Aliases='ARCH, ARCH2', description='zinc finger and BTB domain containing 8 opposite strand', other_designations='archease (ARCH)|archease-like protein|zinc finger and BTB domain-containing opposite strand protein 8', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32620820, end_position_on_the_genomic_accession=32651013, orientation='minus', exon_count=8, OMIM=615891, no_hdr0=''),
  339488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339488, CurrentID=0, Status='live', Symbol='TFAP2E', Aliases='AP2E', description='transcription factor AP-2 epsilon (activating enhancer binding protein 2 epsilon)', other_designations='AP2-epsilon|activating enhancer-binding protein 2-epsilon|adaptor-related protein complex 2, epsilon subunit', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35573370, end_position_on_the_genomic_accession=35595328, orientation='plus', exon_count=7, OMIM=614428, no_hdr0=''),
  100216355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216355, CurrentID=0, Status='live', Symbol='LOC100216355', Aliases='', description='methylmalonic aciduria (cobalamin deficiency) cblD type, with homocystinuria pseudogene', other_designations='', map_location='11p', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34334779, end_position_on_the_genomic_accession=34336108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216356, CurrentID=0, Status='live', Symbol='LOC100216356', Aliases='', description='methylmalonic aciduria (cobalamin deficiency) cblD type, with homocystinuria pseudogene', other_designations='', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76222637, end_position_on_the_genomic_accession=76223431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372934, CurrentID=0, Status='live', Symbol='LOC105372934', Aliases='', description='uncharacterized LOC105372934', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221492288, end_position_on_the_genomic_accession=221504181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339500, CurrentID=0, Status='live', Symbol='ZNF678', Aliases='', description='zinc finger protein 678', other_designations='hypothetical protein MGC42493', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227563519, end_position_on_the_genomic_accession=227677443, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  339501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339501, CurrentID=0, Status='live', Symbol='PRSS38', Aliases='MPN2', description='protease, serine 38', other_designations='marapsin 2|marapsin-2|protease, serine, 38|serine protease MPN2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227815692, end_position_on_the_genomic_accession=227860439, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  339505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339505, CurrentID=0, Status='live', Symbol='LINC01141', Aliases='', description='long intergenic non-protein coding RNA 1141', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20359801, end_position_on_the_genomic_accession=20428794, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105372937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372937, CurrentID=0, Status='live', Symbol='LOC105372937', Aliases='', description='uncharacterized LOC105372937', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221556151, end_position_on_the_genomic_accession=221563633, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339512, CurrentID=0, Status='live', Symbol='CCDC190', Aliases='C1orf110', description='coiled-coil domain containing 190', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162846566, end_position_on_the_genomic_accession=162868900, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370825, CurrentID=0, Status='live', Symbol='LOC105370825', Aliases='', description='uncharacterized LOC105370825', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53548749, end_position_on_the_genomic_accession=53554537, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374189, CurrentID=0, Status='live', Symbol='LOC105374189', Aliases='', description='uncharacterized LOC105374189', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164338927, end_position_on_the_genomic_accession=164370676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339521, CurrentID=0, Status='live', Symbol='FAM58BP', Aliases='FAM58B', description='family with sequence similarity 58 member B, pseudogene', other_designations='family with sequence similarity 58, member B, pseudogene|putative cyclin-related protein FAM58B', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200213532, end_position_on_the_genomic_accession=200214774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372939, CurrentID=0, Status='live', Symbol='LOC105372939', Aliases='', description='uncharacterized LOC105372939', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20482709, end_position_on_the_genomic_accession=20489051, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339524, CurrentID=0, Status='live', Symbol='LINC01140', Aliases='', description='long intergenic non-protein coding RNA 1140', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87129765, end_position_on_the_genomic_accession=87169204, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  339529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339529, CurrentID=0, Status='live', Symbol='LOC339529', Aliases='', description='uncharacterized LOC339529', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243917402, end_position_on_the_genomic_accession=244047317, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  339535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339535, CurrentID=0, Status='live', Symbol='LINC01139', Aliases='', description='long intergenic non-protein coding RNA 1139', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238480384, end_position_on_the_genomic_accession=238486017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339539, CurrentID=0, Status='live', Symbol='LOC339539', Aliases='', description='uncharacterized LOC339539', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42857622, end_position_on_the_genomic_accession=42888792, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339541, CurrentID=0, Status='live', Symbol='C1orf228', Aliases='NCRNA00082, p40', description='chromosome 1 open reading frame 228', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44674694, end_position_on_the_genomic_accession=44725591, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105372943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372943, CurrentID=0, Status='live', Symbol='LOC105372943', Aliases='', description='uncharacterized LOC105372943', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372944, CurrentID=0, Status='live', Symbol='LOC105372944', Aliases='', description='uncharacterized LOC105372944', other_designations='KB-1183D5.18', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21354242, end_position_on_the_genomic_accession=21357968, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375191, CurrentID=0, Status='live', Symbol='LOC105375191', Aliases='', description='uncharacterized LOC105375191', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25186954, end_position_on_the_genomic_accession=25193921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339559, CurrentID=0, Status='live', Symbol='ZFP69', Aliases='ZFP69A, ZKSCAN23A, ZNF642, ZSCAN54A', description='ZFP69 zinc finger protein', other_designations='ZFP69 zinc finger protein A|zinc finger protein 642', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40477220, end_position_on_the_genomic_accession=40496343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  391020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391020, CurrentID=0, Status='live', Symbol='LOC391020', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25124893, end_position_on_the_genomic_accession=25125457, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339562, CurrentID=0, Status='live', Symbol='IGKV1OR2-118', Aliases='', description='immunoglobulin kappa variable 1/OR2-118 (pseudogene)', other_designations='IGKV1/OR2-118|immunoglobulin kappa variable 1/OR2-118 pseudogene', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90315092, end_position_on_the_genomic_accession=90316061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372946, CurrentID=0, Status='live', Symbol='LOC105372946', Aliases='', description='uncharacterized LOC105372946', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21488352, end_position_on_the_genomic_accession=21512698, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339568, CurrentID=0, Status='live', Symbol='LOC339568', Aliases='', description='uncharacterized LOC339568', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39213781, end_position_on_the_genomic_accession=39224748, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372947, CurrentID=0, Status='live', Symbol='LOC105372947', Aliases='', description='uncharacterized LOC105372947', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22920012, end_position_on_the_genomic_accession=22924966, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372950, CurrentID=0, Status='live', Symbol='LOC105372950', Aliases='', description='uncharacterized LOC105372950', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221913762, end_position_on_the_genomic_accession=221916141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339593, CurrentID=0, Status='live', Symbol='LOC339593', Aliases='', description='uncharacterized LOC339593', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11266659, end_position_on_the_genomic_accession=11273383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372951, CurrentID=0, Status='live', Symbol='LOC105372951', Aliases='', description='uncharacterized LOC105372951', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22305111, end_position_on_the_genomic_accession=22310219, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372952, CurrentID=0, Status='live', Symbol='LOC105372952', Aliases='', description='uncharacterized LOC105372952', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22289237, end_position_on_the_genomic_accession=22298030, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100216466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216466, CurrentID=0, Status='live', Symbol='DNM1P43', Aliases='DNM1DN11-10, DNM1DN11@', description='dynamin 1 pseudogene 43', other_designations='DNM1 pseudogene 43', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82337477, end_position_on_the_genomic_accession=82339294, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421462, CurrentID=0, Status='live', Symbol='LOC100421462', Aliases='', description='trafficking protein particle complex 2-like pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70361424, end_position_on_the_genomic_accession=70362374, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372953, CurrentID=0, Status='live', Symbol='LOC105372953', Aliases='', description='uncharacterized LOC105372953', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222385641, end_position_on_the_genomic_accession=222387728, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372954, CurrentID=0, Status='live', Symbol='LOC105372954', Aliases='', description='uncharacterized LOC105372954', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23494294, end_position_on_the_genomic_accession=23510197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100216479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216479, CurrentID=0, Status='live', Symbol='FAR2P2', Aliases='', description='fatty acyl-CoA reductase 2 pseudogene 2', other_designations='fatty acyl CoA reductase 2 pseudogene 2', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130416750, end_position_on_the_genomic_accession=130428546, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100216483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216483, CurrentID=0, Status='live', Symbol='DEL18Q', Aliases='', description='Chromosome 18q deletion syndrome', other_designations='', map_location='18q', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601808, no_hdr0=''),
  100216484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216484, CurrentID=0, Status='live', Symbol='NRCLP4', Aliases='', description='Narcolepsy 4', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612417, no_hdr0=''),
  339622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339622, CurrentID=0, Status='live', Symbol='LOC339622', Aliases='', description='uncharacterized LOC339622', other_designations='', map_location='21q21.1-q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=24840550, end_position_on_the_genomic_accession=25057746, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100216487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216487, CurrentID=0, Status='live', Symbol='AKIRIN2P1', Aliases='', description='akirin 2 pseudogene 1', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=149789958, end_position_on_the_genomic_accession=149790966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216488, CurrentID=0, Status='live', Symbol='LOC100216488', Aliases='', description='akirin 2 pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248165599, end_position_on_the_genomic_accession=248172699, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370829, CurrentID=0, Status='live', Symbol='LOC105370829', Aliases='', description='uncharacterized LOC105370829', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55056747, end_position_on_the_genomic_accession=55092173, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372958, CurrentID=0, Status='live', Symbol='LOC105372958', Aliases='', description='glutathione S-transferase theta-1-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100216505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216505, CurrentID=0, Status='live', Symbol='DNM1P24', Aliases='DNM1DN15-1, DNM1DN15@', description='dynamin 1 pseudogene 24', other_designations='DNM1 pseudogene 24', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24177543, end_position_on_the_genomic_accession=24182632, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216506, CurrentID=0, Status='live', Symbol='DNM1P25', Aliases='DNM1DN15-3, DNM1DN15@', description='dynamin 1 pseudogene 25', other_designations='DNM1 pseudogene 25', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24246282, end_position_on_the_genomic_accession=24247878, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216507, CurrentID=0, Status='live', Symbol='DNM1P26', Aliases='DNM1DN16-1, DNM1DN16@', description='dynamin 1 pseudogene 26', other_designations='DNM1 pseudogene 26', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25487507, end_position_on_the_genomic_accession=25492463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216514, CurrentID=0, Status='live', Symbol='DNM1P27', Aliases='DNM1DN16-3, DNM1DN16@', description='dynamin 1 pseudogene 27', other_designations='DNM1 pseudogene 27', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25421953, end_position_on_the_genomic_accession=25423846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216515, CurrentID=0, Status='live', Symbol='DNM1P36', Aliases='DNM1DN8-3, DNM1DN8@', description='dynamin 1 pseudogene 36', other_designations='DNM1 pseudogene 36', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82439500, end_position_on_the_genomic_accession=82441433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216516, CurrentID=0, Status='live', Symbol='DNM1P37', Aliases='DNM1DN10-1, DNM1DN10@', description='dynamin 1 pseudogene 37', other_designations='DNM1 pseudogene 37', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78737131, end_position_on_the_genomic_accession=78741294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216517, CurrentID=0, Status='live', Symbol='DNM1P38', Aliases='DNM1DN10-2, DNM1DN10@, DNM1DN11-8, DNM1DN11@, DNM1P42', description='dynamin 1 pseudogene 38', other_designations='DNM1 pseudogene 38|DNM1 pseudogene 42', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82529853, end_position_on_the_genomic_accession=82532983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216518, CurrentID=0, Status='live', Symbol='DNM1P39', Aliases='DNM1DN11-3, DNM1DN11@', description='dynamin 1 pseudogene 39', other_designations='DNM1 pseudogene 39', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  122145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122145, CurrentID=0, Status='live', Symbol='RIOK3P1', Aliases='', description='RIO kinase 3 pseudogene 1', other_designations='suppressor of bimD6 homolog pseudogene', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74828493, end_position_on_the_genomic_accession=74830288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373317, CurrentID=0, Status='live', Symbol='LOC105373317', Aliases='', description='uncharacterized LOC105373317', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248859164, end_position_on_the_genomic_accession=248865047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370830, CurrentID=0, Status='live', Symbol='LOC105370830', Aliases='', description='uncharacterized LOC105370830', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55192840, end_position_on_the_genomic_accession=55196799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339665, CurrentID=0, Status='live', Symbol='SLC35E4', Aliases='', description='solute carrier family 35 member E4', other_designations='solute carrier family 35, member E4', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30635806, end_position_on_the_genomic_accession=30647875, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  339666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339666, CurrentID=0, Status='live', Symbol='LOC339666', Aliases='', description='uncharacterized LOC339666', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32376664, end_position_on_the_genomic_accession=32384343, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372963, CurrentID=0, Status='live', Symbol='LOC105372963', Aliases='', description='uncharacterized LOC105372963', other_designations='CTA-125H2.2', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25883141, end_position_on_the_genomic_accession=25901020, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339669, CurrentID=0, Status='live', Symbol='TEX33', Aliases='C22orf33, EAN57, cE81G9.2', description='testis expressed 33', other_designations='protein EAN57', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36991119, end_position_on_the_genomic_accession=37007863, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  339674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339674, CurrentID=0, Status='live', Symbol='LINC00634', Aliases='BK250D10.8', description='long intergenic non-protein coding RNA 634', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41952187, end_position_on_the_genomic_accession=41958942, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100216542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216542, CurrentID=0, Status='live', Symbol='DNM1P45', Aliases='DNM1DN12-2, DNM1DN12@', description='dynamin 1 pseudogene 45', other_designations='DNM1 pseudogene 45', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85250779, end_position_on_the_genomic_accession=85253256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100216544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216544, CurrentID=0, Status='live', Symbol='DNM1P47', Aliases='DNM1DN14-3, DNM1DN14@', description='dynamin 1 pseudogene 47', other_designations='DNM1 pseudogene 47', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101764714, end_position_on_the_genomic_accession=101768338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100216545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216545, CurrentID=0, Status='live', Symbol='KMT2E-AS1', Aliases='', description='KMT2E antisense RNA 1 (head to head)', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105010542, end_position_on_the_genomic_accession=105014141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100216546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100216546, CurrentID=0, Status='live', Symbol='LINC01004', Aliases='', description='long intergenic non-protein coding RNA 1004', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104981747, end_position_on_the_genomic_accession=104991165, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372966, CurrentID=0, Status='live', Symbol='LOC105372966', Aliases='', description='uncharacterized LOC105372966', other_designations='CTA-390C10.10', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25476218, end_position_on_the_genomic_accession=25479939, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106479419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479419, CurrentID=0, Status='live', Symbol='RN7SL542P', Aliases='', description='RNA, 7SL, cytoplasmic 542, pseudogene', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129042, CurrentID=0, Status='live', Symbol='LOC100129042', Aliases='', description='uncharacterized LOC100129042', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=15198113, end_position_on_the_genomic_accession=15198835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421465, CurrentID=0, Status='live', Symbol='LOC100421465', Aliases='', description='ubiquitin-like modifier activating enzyme 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87044933, end_position_on_the_genomic_accession=87046008, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372968, CurrentID=0, Status='live', Symbol='LOC105372968', Aliases='', description='uncharacterized LOC105372968', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25282891, end_position_on_the_genomic_accession=25302453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646442, CurrentID=0, Status='live', Symbol='SNX19P3', Aliases='', description='sorting nexin 19 pseudogene 3', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14628383, end_position_on_the_genomic_accession=14633302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372973, CurrentID=0, Status='live', Symbol='LOC105372973', Aliases='', description='uncharacterized LOC105372973', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26352180, end_position_on_the_genomic_accession=26354675, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270866, CurrentID=0, Status='live', Symbol='RPS6P11', Aliases='RPS6_5_956', description='ribosomal protein S6 pseudogene 11', other_designations='', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4633032, end_position_on_the_genomic_accession=4633770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372974, CurrentID=0, Status='live', Symbol='LOC105372974', Aliases='', description='uncharacterized LOC105372974', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26240611, end_position_on_the_genomic_accession=26254059, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  339736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339736, CurrentID=0, Status='live', Symbol='AK2P2', Aliases='', description='adenylate kinase 2 pseudogene 2', other_designations='', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31823410, end_position_on_the_genomic_accession=31823830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339742, CurrentID=0, Status='live', Symbol='FAM201B', Aliases='', description='family with sequence similarity 201 member B', other_designations='family with sequence similarity 201, member B', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132352499, end_position_on_the_genomic_accession=132354283, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  339745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339745, CurrentID=0, Status='live', Symbol='SPOPL', Aliases='BTBD33', description='speckle type POZ protein like', other_designations='HIB homolog 2|roadkill homolog 2', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138501780, end_position_on_the_genomic_accession=138573235, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  339751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339751, CurrentID=0, Status='live', Symbol='MLK7-AS1', Aliases='', description='MLK7 antisense RNA 1', other_designations='MLK7 antisense RNA 1 (non-protein coding)', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173197712, end_position_on_the_genomic_accession=173282036, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372978, CurrentID=0, Status='live', Symbol='LOC105372978', Aliases='', description='uncharacterized LOC105372978', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27221664, end_position_on_the_genomic_accession=27224710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339761, CurrentID=0, Status='live', Symbol='CYP27C1', Aliases='', description='cytochrome P450 family 27 subfamily C member 1', other_designations='cytochrome P450, family 27, subfamily C, polypeptide 1|cytochrome P450, family 27, subfamily C, polypeptide 13', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127183832, end_position_on_the_genomic_accession=127226356, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105372979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372979, CurrentID=0, Status='live', Symbol='LOC105372979', Aliases='', description='uncharacterized LOC105372979', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27320541, end_position_on_the_genomic_accession=27322221, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339766, CurrentID=0, Status='live', Symbol='MROH2A', Aliases='HEATR7B1', description='maestro heat-like repeat family member 2A', other_designations='HEAT repeat containing 7B1|HEAT repeat-containing protein 7B1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233775662, end_position_on_the_genomic_accession=233833423, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  339768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339768, CurrentID=0, Status='live', Symbol='ESPNL', Aliases='', description='espin-like', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238100310, end_position_on_the_genomic_accession=238133287, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101927227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927227, CurrentID=0, Status='live', Symbol='LOC101927227', Aliases='', description='uncharacterized LOC101927227', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12366408, end_position_on_the_genomic_accession=12372582, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130101, CurrentID=0, Status='live', Symbol='LOC100130101', Aliases='', description='uncharacterized LOC100130101', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=97417524, end_position_on_the_genomic_accession=97418748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  339778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339778, CurrentID=0, Status='live', Symbol='C2orf70', Aliases='', description='chromosome 2 open reading frame 70', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26562556, end_position_on_the_genomic_accession=26579529, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  339779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339779, CurrentID=0, Status='live', Symbol='PRR30', Aliases='C2orf53', description='proline rich 30', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27136847, end_position_on_the_genomic_accession=27139464, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339781, CurrentID=0, Status='live', Symbol='KRT18P19', Aliases='', description='keratin 18 pseudogene 19', other_designations='', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189311205, end_position_on_the_genomic_accession=189312548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339788, CurrentID=0, Status='live', Symbol='LINC00298', Aliases='', description='long intergenic non-protein coding RNA 298', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7922425, end_position_on_the_genomic_accession=7976814, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339789, CurrentID=0, Status='live', Symbol='LINC00299', Aliases='C2orf46, NCRNA00299', description='long intergenic non-protein coding RNA 299', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8007771, end_position_on_the_genomic_accession=8328419, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101060327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060327, CurrentID=0, Status='live', Symbol='LOC101060327', Aliases='', description='protein FAM231B-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  339799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339799, CurrentID=0, Status='live', Symbol='EIF3FP3', Aliases='', description='eukaryotic translation initiation factor 3 subunit F pseudogene 3', other_designations='eukaryotic translation initiation factor 3, subunit F pseudogene 3|hCG15200', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=58251429, end_position_on_the_genomic_accession=58252687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339803, CurrentID=0, Status='live', Symbol='LOC339803', Aliases='', description='uncharacterized LOC339803', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61141592, end_position_on_the_genomic_accession=61144975, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339804, CurrentID=0, Status='live', Symbol='C2orf74', Aliases='', description='chromosome 2 open reading frame 74', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61145091, end_position_on_the_genomic_accession=61164830, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372986, CurrentID=0, Status='live', Symbol='LOC105372986', Aliases='', description='uncharacterized LOC105372986', other_designations='CTA-984G1.5', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29260865, end_position_on_the_genomic_accession=29261191, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  339807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339807, CurrentID=0, Status='live', Symbol='LOC339807', Aliases='', description='uncharacterized LOC339807', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64607312, end_position_on_the_genomic_accession=64616482, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271227, CurrentID=0, Status='live', Symbol='RPL9P26', Aliases='RPL9_11_1421', description='ribosomal protein L9 pseudogene 26', other_designations='', map_location='15q11.2-q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26851114, end_position_on_the_genomic_accession=26851610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129046, CurrentID=0, Status='live', Symbol='LOC100129046', Aliases='', description='uncharacterized LOC100129046', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93591968, end_position_on_the_genomic_accession=93600030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379564, CurrentID=0, Status='live', Symbol='LOC105379564', Aliases='', description='uncharacterized LOC105379564', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372988, CurrentID=0, Status='live', Symbol='LOC105372988', Aliases='', description='uncharacterized LOC105372988', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30182935, end_position_on_the_genomic_accession=30207134, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339822, CurrentID=0, Status='live', Symbol='LINC01115', Aliases='', description='long intergenic non-protein coding RNA 1115', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=779840, end_position_on_the_genomic_accession=868426, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372989, CurrentID=0, Status='live', Symbol='LOC105372989', Aliases='', description='uncharacterized LOC105372989', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30421206, end_position_on_the_genomic_accession=30423452, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  339829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339829, CurrentID=0, Status='live', Symbol='CCDC39', Aliases='CILD14, FAP59', description='coiled-coil domain containing 39', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180614008, end_position_on_the_genomic_accession=180679495, orientation='minus', exon_count=20, OMIM=613798, no_hdr0=''),
  339834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339834, CurrentID=0, Status='live', Symbol='CCDC36', Aliases='CT74', description='coiled-coil domain containing 36', other_designations='cancer/testis antigen 74', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49198428, end_position_on_the_genomic_accession=49258157, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100130111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130111, CurrentID=0, Status='live', Symbol='LOC100130111', Aliases='', description='uncharacterized LOC100130111', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29674989, end_position_on_the_genomic_accession=29681183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339843, CurrentID=0, Status='live', Symbol='WWP1P1', Aliases='', description='WW domain containing E3 ubiquitin protein ligase 1 pseudogene 1', other_designations='', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98657539, end_position_on_the_genomic_accession=98660967, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339845, CurrentID=0, Status='live', Symbol='ASS1P7', Aliases='ASSP7', description='argininosuccinate synthetase 1 pseudogene 7', other_designations='argininosuccinate synthetase pseudogene 7', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177296923, end_position_on_the_genomic_accession=177298488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372994, CurrentID=0, Status='live', Symbol='LOC105372994', Aliases='', description='uncharacterized LOC105372994', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30872270, end_position_on_the_genomic_accession=30879797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339855, CurrentID=0, Status='live', Symbol='KY', Aliases='', description='kyphoscoliosis peptidase', other_designations='', map_location='3q22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134599923, end_position_on_the_genomic_accession=134651677, orientation='minus', exon_count=12, OMIM=605739, no_hdr0=''),
  105372995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372995, CurrentID=0, Status='live', Symbol='LOC105372995', Aliases='', description='uncharacterized LOC105372995', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31107798, end_position_on_the_genomic_accession=31122949, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339862, CurrentID=0, Status='live', Symbol='LOC339862', Aliases='', description='uncharacterized LOC339862', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17962572, end_position_on_the_genomic_accession=18268922, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372997, CurrentID=0, Status='live', Symbol='LOC105372997', Aliases='', description='uncharacterized LOC105372997', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31292947, end_position_on_the_genomic_accession=31303297, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339874, CurrentID=0, Status='live', Symbol='LOC339874', Aliases='', description='uncharacterized LOC339874', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131325092, end_position_on_the_genomic_accession=131381475, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  339879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339879, CurrentID=0, Status='live', Symbol='ALG1L6P', Aliases='', description='asparagine-linked glycosylation 1-like 6, pseudogene', other_designations='asparagine-linked glycosylation 1 homolog pseudogene', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75415067, end_position_on_the_genomic_accession=75424731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339881, CurrentID=0, Status='live', Symbol='EIF4BP8', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 8', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122660587, end_position_on_the_genomic_accession=122662120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339883, CurrentID=0, Status='live', Symbol='C3orf35', Aliases='APRG1', description='chromosome 3 open reading frame 35', other_designations='AP20 region protein1|APRG1 tumor suppressor candidate', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37386269, end_position_on_the_genomic_accession=37435497, orientation='plus', exon_count=9, OMIM=611429, no_hdr0=''),
  644173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644173, CurrentID=0, Status='live', Symbol='LOC644173', Aliases='', description='uncharacterized LOC644173', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=56814122, end_position_on_the_genomic_accession=56815581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375202, CurrentID=0, Status='live', Symbol='LOC105375202', Aliases='', description='uncharacterized LOC105375202', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26429028, end_position_on_the_genomic_accession=26430872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  339894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339894, CurrentID=0, Status='live', Symbol='LINC00880', Aliases='', description='long intergenic non-protein coding RNA 880', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157081667, end_position_on_the_genomic_accession=157123002, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130121, CurrentID=0, Status='live', Symbol='LOC100130121', Aliases='', description='solute carrier family 19 (thiamine transporter), member 3 pseudogene', other_designations='solute carrier family 19, member 3 pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55244247, end_position_on_the_genomic_accession=55245681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339896, CurrentID=0, Status='live', Symbol='GADL1', Aliases='ADC, CSADC, HuADC, HuCSADC', description='glutamate decarboxylase like 1', other_designations='aspartate 1-decarboxylase|cysteine sulfinic acid decarboxylase|glutamate decarboxylase-like 1|glutamate decarboxylase-like protein 1', map_location='3p24.1-p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30726200, end_position_on_the_genomic_accession=30894661, orientation='minus', exon_count=15, OMIM=615601, no_hdr0=''),
  339897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339897, CurrentID=0, Status='live', Symbol='CNN2P6', Aliases='', description='calponin 2 pseudogene 6', other_designations='', map_location='3p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31181556, end_position_on_the_genomic_accession=31182650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339902, CurrentID=0, Status='live', Symbol='LOC339902', Aliases='', description='hCG1813818', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59637296, end_position_on_the_genomic_accession=59658560, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370838, CurrentID=0, Status='live', Symbol='LOC105370838', Aliases='', description='uncharacterized LOC105370838', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60029652, end_position_on_the_genomic_accession=60041899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  339906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339906, CurrentID=0, Status='live', Symbol='PRSS42', Aliases='TESSP2', description='protease, serine 42', other_designations='protease, serine, 42|putative testis serine protease 2|testis serine protease 2', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46830404, end_position_on_the_genomic_accession=46834095, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130691, CurrentID=0, Status='live', Symbol='LOC100130691', Aliases='', description='uncharacterized LOC100130691', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177283508, end_position_on_the_genomic_accession=177392691, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  339910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339910, CurrentID=0, Status='live', Symbol='MRPS35P1', Aliases='', description='mitochondrial ribosomal protein S35 pseudogene 1', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5446976, end_position_on_the_genomic_accession=5447103, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646476, CurrentID=0, Status='live', Symbol='RPL7AP39', Aliases='RPL7A_12_828', description='ribosomal protein L7a pseudogene 39', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102755130, end_position_on_the_genomic_accession=102755989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  339926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339926, CurrentID=0, Status='live', Symbol='EHHADH-AS1', Aliases='', description='EHHADH antisense RNA 1', other_designations='EHHADH antisense RNA 1 (non-protein coding)', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185162901, end_position_on_the_genomic_accession=185191955, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  339929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339929, CurrentID=0, Status='live', Symbol='LPP-AS2', Aliases='MYCLo-5, MYCLo-6', description='LPP antisense RNA 2', other_designations='LPP antisense RNA 2 (non-protein coding)', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188151206, end_position_on_the_genomic_accession=188154088, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373007, CurrentID=0, Status='live', Symbol='LOC105373007', Aliases='', description='uncharacterized LOC105373007', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33573542, end_position_on_the_genomic_accession=33598361, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  253264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253264, CurrentID=0, Status='live', Symbol='ZNF503-AS1', Aliases='', description='ZNF503 antisense RNA 1', other_designations='ZNF503 antisense RNA 1 (non-protein coding)', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75296383, end_position_on_the_genomic_accession=75361678, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060328, CurrentID=0, Status='live', Symbol='LOC101060328', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  339942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339942, CurrentID=0, Status='live', Symbol='H1FX-AS1', Aliases='C3orf47', description='H1FX antisense RNA 1', other_designations='H1FX antisense RNA 1 (non-protein coding)', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129316271, end_position_on_the_genomic_accession=129324569, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646482, CurrentID=0, Status='live', Symbol='LOC646482', Aliases='', description='uncharacterized LOC646482', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=25874262, end_position_on_the_genomic_accession=25875879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373011, CurrentID=0, Status='live', Symbol='LOC105373011', Aliases='', description='uncharacterized LOC105373011', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373319, CurrentID=0, Status='live', Symbol='LOC105373319', Aliases='', description='uncharacterized LOC105373319', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194493, end_position_on_the_genomic_accession=209541, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373012, CurrentID=0, Status='live', Symbol='LOC105373012', Aliases='', description='uncharacterized LOC105373012', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34349672, end_position_on_the_genomic_accession=34362313, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370840, CurrentID=0, Status='live', Symbol='LOC105370840', Aliases='', description='uncharacterized LOC105370840', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60305917, end_position_on_the_genomic_accession=60309400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339965, CurrentID=0, Status='live', Symbol='CCDC158', Aliases='', description='coiled-coil domain containing 158', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76313039, end_position_on_the_genomic_accession=76421893, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  339966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339966, CurrentID=0, Status='live', Symbol='LOC339966', Aliases='', description='ubiquitin family domain containing 1 pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77112495, end_position_on_the_genomic_accession=77113443, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339967, CurrentID=0, Status='live', Symbol='TMPRSS11A', Aliases='ECRG1', description='transmembrane protease, serine 11A', other_designations='airway trypsin-like protease 1|epidermal type II transmembrane serine protease|epidermal type-II transmembrane serine protease|esophageal cancer-susceptibility gene 1 protein', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67910301, end_position_on_the_genomic_accession=67963514, orientation='minus', exon_count=10, OMIM=611704, no_hdr0=''),
  339970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339970, CurrentID=0, Status='live', Symbol='GCOM2', Aliases='GLURR2, GRINL1B', description='GRINL1B complex locus 2, pseudogene', other_designations='glutamate receptor, ionotropic, N-methyl D-aspartate-like 1B', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68038416, end_position_on_the_genomic_accession=68042511, orientation='plus', exon_count=0, OMIM=608311, no_hdr0=''),
  100130134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130134, CurrentID=0, Status='live', Symbol='LOC100130134', Aliases='', description='staufen double-stranded RNA binding protein 2 pseudogene', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=90037863, end_position_on_the_genomic_accession=90040928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  339975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339975, CurrentID=0, Status='live', Symbol='LOC339975', Aliases='', description='uncharacterized LOC339975', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187304083, end_position_on_the_genomic_accession=187505613, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  339976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339976, CurrentID=0, Status='live', Symbol='TRIML1', Aliases='RNF209', description='tripartite motif family like 1', other_designations='RING finger protein 209|tripartite motif family-like 1|tripartite motif family-like protein 1', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188138189, end_position_on_the_genomic_accession=188147743, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  339977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339977, CurrentID=0, Status='live', Symbol='LRRC66', Aliases='', description='leucine rich repeat containing 66', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=51993640, end_position_on_the_genomic_accession=52020311, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  339978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339978, CurrentID=0, Status='live', Symbol='LOC339978', Aliases='', description='uncharacterized LOC339978', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54603211, end_position_on_the_genomic_accession=54607147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373015, CurrentID=0, Status='live', Symbol='LOC105373015', Aliases='', description='uncharacterized LOC105373015', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34947086, end_position_on_the_genomic_accession=34952782, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  339983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339983, CurrentID=0, Status='live', Symbol='NAT8L', Aliases='CML3, NACED, NAT8-LIKE', description='N-acetyltransferase 8-like (GCN5-related, putative)', other_designations='NAA synthetase|camello-like protein 3', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2059512, end_position_on_the_genomic_accession=2069089, orientation='plus', exon_count=3, OMIM=610647, no_hdr0=''),
  105375065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375065, CurrentID=0, Status='live', Symbol='LOC105375065', Aliases='', description='uncharacterized LOC105375065', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43391335, end_position_on_the_genomic_accession=43424729, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373016, CurrentID=0, Status='live', Symbol='LOC105373016', Aliases='', description='uncharacterized LOC105373016', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34946390, end_position_on_the_genomic_accession=34968018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  339988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=339988, CurrentID=0, Status='live', Symbol='LOC339988', Aliases='', description='uncharacterized LOC339988', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6687792, end_position_on_the_genomic_accession=6690729, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373017, CurrentID=0, Status='live', Symbol='LOC105373017', Aliases='', description='uncharacterized LOC105373017', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35232641, end_position_on_the_genomic_accession=35262311, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373019, CurrentID=0, Status='live', Symbol='LOC105373019', Aliases='', description='uncharacterized LOC105373019', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36498165, end_position_on_the_genomic_accession=36498931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  340017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340017, CurrentID=0, Status='live', Symbol='LOC340017', Aliases='', description='uncharacterized LOC340017', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157572490, end_position_on_the_genomic_accession=157576154, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340024, CurrentID=0, Status='live', Symbol='SLC6A19', Aliases='B0AT1, HND', description='solute carrier family 6 (neutral amino acid transporter), member 19', other_designations='sodium-dependent amino acid transporter system B0|solute carrier family 6 (neurotransmitter transporter), member 19|solute carrier family 6 member 19|system B(0) neutral amino acid transporter AT1|system B0 neutral amino acid transporter', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1201595, end_position_on_the_genomic_accession=1225117, orientation='plus', exon_count=12, OMIM=608893, no_hdr0=''),
  105373024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373024, CurrentID=0, Status='live', Symbol='LOC105373024', Aliases='', description='uncharacterized LOC105373024', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37290754, end_position_on_the_genomic_accession=37332126, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  340037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340037, CurrentID=0, Status='live', Symbol='PRR7-AS1', Aliases='', description='PRR7 antisense RNA 1', other_designations='PRR7 antisense RNA 1 (non-protein coding)', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177437889, end_position_on_the_genomic_accession=177447944, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373026, CurrentID=0, Status='live', Symbol='LOC105373026', Aliases='', description='uncharacterized LOC105373026', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37972153, end_position_on_the_genomic_accession=38026755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379572, CurrentID=0, Status='live', Symbol='LOC105379572', Aliases='', description='KAT8 regulatory NSL complex subunit 1-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370843, CurrentID=0, Status='live', Symbol='LOC105370843', Aliases='', description='uncharacterized LOC105370843', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60848758, end_position_on_the_genomic_accession=60852954, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373028, CurrentID=0, Status='live', Symbol='LOC105373028', Aliases='', description='uncharacterized LOC105373028', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38075986, end_position_on_the_genomic_accession=38078192, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340061, CurrentID=0, Status='live', Symbol='TMEM173', Aliases='ERIS, MITA, MPYS, NET23, SAVI, STING, hMITA, hSTING', description='transmembrane protein 173', other_designations='N-terminal methionine-proline-tyrosine-serine plasma membrane tetraspanner|endoplasmic reticulum IFN stimulator|endoplasmic reticulum interferon stimulator|mitochondrial mediator of IRF3 activation', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139475528, end_position_on_the_genomic_accession=139482790, orientation='minus', exon_count=9, OMIM=612374, no_hdr0=''),
  340069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340069, CurrentID=0, Status='live', Symbol='FAM170A', Aliases='ZNFD', description='family with sequence similarity 170 member A', other_designations='Znf domain containing protein|family with sequence similarity 170, member A|zinc finger domain-containing protein|zinc finger protein ZNFD', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119629559, end_position_on_the_genomic_accession=119635822, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421907, CurrentID=0, Status='live', Symbol='LOC100421907', Aliases='', description='olfactory receptor, family 2, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340073, CurrentID=0, Status='live', Symbol='C5orf66-AS2', Aliases='', description='C5orf66 antisense RNA 2', other_designations='CTC-349C3.2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135236234, end_position_on_the_genomic_accession=135248177, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  340074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340074, CurrentID=0, Status='live', Symbol='LOC340074', Aliases='', description='uncharacterized LOC340074', other_designations='CTC-321K16.4', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135648683, end_position_on_the_genomic_accession=135653935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340075, CurrentID=0, Status='live', Symbol='ARSI', Aliases='SPG66', description='arylsulfatase family member I', other_designations='ASI|arylsulfatase family, member I', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150296346, end_position_on_the_genomic_accession=150302962, orientation='minus', exon_count=2, OMIM=610009, no_hdr0=''),
  100130152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130152, CurrentID=0, Status='live', Symbol='LOC100130152', Aliases='', description='uncharacterized LOC100130152', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63392887, end_position_on_the_genomic_accession=63398343, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  340085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340085, CurrentID=0, Status='live', Symbol='LOC340085', Aliases='', description='uncharacterized LOC340085', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373033, CurrentID=0, Status='live', Symbol='LOC105373033', Aliases='', description='uncharacterized LOC105373033', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38963259, end_position_on_the_genomic_accession=38969367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340089, CurrentID=0, Status='live', Symbol='LOC340089', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70508136, end_position_on_the_genomic_accession=70508825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340090, CurrentID=0, Status='live', Symbol='LOC340090', Aliases='', description='uncharacterized LOC340090', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73194775, end_position_on_the_genomic_accession=73201969, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373034, CurrentID=0, Status='live', Symbol='LOC105373034', Aliases='', description='uncharacterized LOC105373034', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223506301, end_position_on_the_genomic_accession=223508625, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340094, CurrentID=0, Status='live', Symbol='LINC01020', Aliases='', description='long intergenic non-protein coding RNA 1020', other_designations='CTD-2247C11.3', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5034359, end_position_on_the_genomic_accession=5070004, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  340096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340096, CurrentID=0, Status='live', Symbol='H3.X', Aliases='', description='histone cluster 2, H3d pseudogene', other_designations='H3 histone, family 3B (H3.3B) pseudogene', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17491361, end_position_on_the_genomic_accession=17491768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646507, CurrentID=0, Status='live', Symbol='SPATA31E3P', Aliases='FAM75E3P', description='SPATA31 subfamily E, member 3, pseudogene', other_designations='family with sequence similarity 75, member E3, pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22403038, end_position_on_the_genomic_accession=22407019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130156, CurrentID=0, Status='live', Symbol='LOC100130156', Aliases='', description='uncharacterized LOC100130156', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=55097413, end_position_on_the_genomic_accession=55097855, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  340107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340107, CurrentID=0, Status='live', Symbol='LOC340107', Aliases='', description='uncharacterized LOC340107', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24835280, end_position_on_the_genomic_accession=24840583, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271229, CurrentID=0, Status='live', Symbol='RPL9P33', Aliases='RPL9_13_1663', description='ribosomal protein L9 pseudogene 33', other_designations='', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51621142, end_position_on_the_genomic_accession=51621701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373037, CurrentID=0, Status='live', Symbol='LOC105373037', Aliases='', description='uncharacterized LOC105373037', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40555177, end_position_on_the_genomic_accession=40564466, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  340113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340113, CurrentID=0, Status='live', Symbol='LOC340113', Aliases='', description='uncharacterized LOC340113', other_designations='CTD-2218G20.2', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32947443, end_position_on_the_genomic_accession=32962467, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373038, CurrentID=0, Status='live', Symbol='LOC105373038', Aliases='', description='uncharacterized LOC105373038', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40651886, end_position_on_the_genomic_accession=40655216, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340120, CurrentID=0, Status='live', Symbol='ANKRD34B', Aliases='DP58', description='ankyrin repeat domain 34B', other_designations='cytosolic phosphoprotein DP58|dendritic phosphoprotein 58 kDa', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80556755, end_position_on_the_genomic_accession=80572753, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375210, CurrentID=0, Status='live', Symbol='LOC105375210', Aliases='', description='uncharacterized LOC105375210', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27733053, end_position_on_the_genomic_accession=27740101, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373041, CurrentID=0, Status='live', Symbol='LOC105373041', Aliases='', description='uncharacterized LOC105373041', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223772711, end_position_on_the_genomic_accession=223774827, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421480, CurrentID=0, Status='live', Symbol='TPT1P12', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 12', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9155665, end_position_on_the_genomic_accession=9156527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340146, CurrentID=0, Status='live', Symbol='SLC35D3', Aliases='FRCL1', description='solute carrier family 35 member D3', other_designations='frc, fringe-like 1|fringe connection-like protein 1|solute carrier family 35, member D3', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136922241, end_position_on_the_genomic_accession=136925660, orientation='plus', exon_count=2, OMIM=612519, no_hdr0=''),
  340152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340152, CurrentID=0, Status='live', Symbol='ZC3H12D', Aliases='C6orf95, MCPIP4, TFL, dJ281H8.1, p34', description='zinc finger CCCH-type containing 12D', other_designations='Zinc finger CCCH domain-containing protein 12D|p34|transformed follicular lymphoma|tumor suppressor TFL', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149447630, end_position_on_the_genomic_accession=149485012, orientation='minus', exon_count=6, OMIM=611106, no_hdr0=''),
  340156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340156, CurrentID=0, Status='live', Symbol='MYLK4', Aliases='SgK085', description='myosin light chain kinase family member 4', other_designations='caMLCK like|myosin light chain kinase family, member 4|sugen kinase 85', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2663629, end_position_on_the_genomic_accession=2770123, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105373045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373045, CurrentID=0, Status='live', Symbol='LOC105373045', Aliases='', description='uncharacterized LOC105373045', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41885186, end_position_on_the_genomic_accession=41892439, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340168, CurrentID=0, Status='live', Symbol='DPPA5', Aliases='ESG1', description='developmental pluripotency associated 5', other_designations='embryonal stem cell specific gene 1', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73353062, end_position_on_the_genomic_accession=73354276, orientation='minus', exon_count=3, OMIM=611111, no_hdr0=''),
  105373047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373047, CurrentID=0, Status='live', Symbol='LOC105373047', Aliases='', description='uncharacterized LOC105373047', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42121001, end_position_on_the_genomic_accession=42123590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379576, CurrentID=0, Status='live', Symbol='LOC105379576', Aliases='', description='uncharacterized LOC105379576', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373048, CurrentID=0, Status='live', Symbol='LOC105373048', Aliases='', description='uncharacterized LOC105373048', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42396547, end_position_on_the_genomic_accession=42398204, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340178, CurrentID=0, Status='live', Symbol='LOC340178', Aliases='', description='uncharacterized LOC340178', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130169, CurrentID=0, Status='live', Symbol='LOC100130169', Aliases='', description='uncharacterized LOC100130169', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=141115747, end_position_on_the_genomic_accession=141116169, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  340184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340184, CurrentID=0, Status='live', Symbol='LOC340184', Aliases='', description='uncharacterized LOC340184', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11690060, end_position_on_the_genomic_accession=11723500, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373050, CurrentID=0, Status='live', Symbol='LOC105373050', Aliases='', description='uncharacterized LOC105373050', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43232011, end_position_on_the_genomic_accession=43237084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340192, CurrentID=0, Status='live', Symbol='IQCB2P', Aliases='IQCB1P1, dJ313I6.1', description='IQ motif containing B2 pseudogene', other_designations='IQ calmodulin-binding motif containing protein 1 pseudogene|IQ motif containing B1 pseudogene|IQCB1 pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28010551, end_position_on_the_genomic_accession=28012902, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130171, CurrentID=0, Status='live', Symbol='LOC100130171', Aliases='', description='hematological and neurological expressed 1 pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173431247, end_position_on_the_genomic_accession=173431812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340198, CurrentID=0, Status='live', Symbol='IFITM4P', Aliases='dJ377H14.5', description='interferon induced transmembrane protein 4 pseudogene', other_designations='Interferon-induced transmembrane protein 3 (Interferon-inducible protein 1-8U)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29750807, end_position_on_the_genomic_accession=29751148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373052, CurrentID=0, Status='live', Symbol='LOC105373052', Aliases='', description='uncharacterized LOC105373052', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43210247, end_position_on_the_genomic_accession=43212578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340204, CurrentID=0, Status='live', Symbol='CLPSL1', Aliases='C6orf127, ESP32, dJ510O8.6', description='colipase like 1', other_designations='colipase-like 1|colipase-like protein C6orf127', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35781054, end_position_on_the_genomic_accession=35788064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340205, CurrentID=0, Status='live', Symbol='TREML1', Aliases='GLTL1825, PRO3438, TLT-1, TLT1, dJ238O23.3', description='triggering receptor expressed on myeloid cells like 1', other_designations='triggering receptor expressed on myeloid cells-like 1|triggering receptor expressed on myeloid cells-like protein 1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41149260, end_position_on_the_genomic_accession=41154349, orientation='minus', exon_count=6, OMIM=609714, no_hdr0=''),
  340206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340206, CurrentID=0, Status='live', Symbol='TREML3P', Aliases='TLT3, TREML3', description='triggering receptor expressed on myeloid cells like 3, pseudogene', other_designations='TREM like transcript 3|triggering receptor expressed on myeloid cells-like 3, pseudogene', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41208554, end_position_on_the_genomic_accession=41217947, orientation='minus', exon_count=5, OMIM=609716, no_hdr0=''),
  105373054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373054, CurrentID=0, Status='live', Symbol='LOC105373054', Aliases='', description='uncharacterized LOC105373054', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43399902, end_position_on_the_genomic_accession=43409717, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  340217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340217, CurrentID=0, Status='live', Symbol='NANOGP3', Aliases='', description='Nanog homeobox pseudogene 3', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53418456, end_position_on_the_genomic_accession=53418738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373055, CurrentID=0, Status='live', Symbol='LOC105373055', Aliases='', description='uncharacterized LOC105373055', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43698657, end_position_on_the_genomic_accession=43700111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340221, CurrentID=0, Status='live', Symbol='TNRC18P3', Aliases='TNRC18C', description='trinucleotide repeat containing 18 pseudogene 3', other_designations='trinucleotide repeat containing 18C (pseudogene)', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56995534, end_position_on_the_genomic_accession=57009152, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130176, CurrentID=0, Status='live', Symbol='LOC100130176', Aliases='', description='transmembrane protein 188 pseudogene', other_designations='CTD nuclear envelope phosphatase 1 regulatory subunit 1 pseudogene', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103651733, end_position_on_the_genomic_accession=103652106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373057, CurrentID=0, Status='live', Symbol='LOC105373057', Aliases='', description='DNA-directed RNA polymerase II subunit RPB1-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43836042, end_position_on_the_genomic_accession=43838860, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370849, CurrentID=0, Status='live', Symbol='LOC105370849', Aliases='', description='golgin subfamily A member 2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62249533, end_position_on_the_genomic_accession=62256938, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130178, CurrentID=0, Status='live', Symbol='LSM3P4', Aliases='', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 4', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated pseudogene 4|LSM3 homolog, U6 small nuclear RNA associated pseudogene 4|LSM3 pseudogene 4', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=142317214, end_position_on_the_genomic_accession=142317774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  257152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257152, CurrentID=0, Status='live', Symbol='LOC257152', Aliases='', description='uncharacterized LOC257152', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373059, CurrentID=0, Status='live', Symbol='LOC105373059', Aliases='', description='uncharacterized LOC105373059', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44537968, end_position_on_the_genomic_accession=44549526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373060, CurrentID=0, Status='live', Symbol='LOC105373060', Aliases='', description='uncharacterized LOC105373060', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44624078, end_position_on_the_genomic_accession=44624792, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  340252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340252, CurrentID=0, Status='live', Symbol='ZNF680', Aliases='', description='zinc finger protein 680', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64501290, end_position_on_the_genomic_accession=64563127, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  340260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340260, CurrentID=0, Status='live', Symbol='UNCX', Aliases='UNCX4.1', description='UNC homeobox', other_designations='UNC homeobox protein|Unc4.1 homeobox|homeobox protein Uncx4.1|truncated UNCX homeobox protein', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1233018, end_position_on_the_genomic_accession=1236977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340267, CurrentID=0, Status='live', Symbol='COL28A1', Aliases='COL28', description='collagen, type XXVIII, alpha 1', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7338549, end_position_on_the_genomic_accession=7537277, orientation='minus', exon_count=45, OMIM=609996, no_hdr0=''),
  340268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340268, CurrentID=0, Status='live', Symbol='LOC340268', Aliases='', description='C3 and PZP-like, alpha-2-macroglobulin domain containing 8 pseudogene', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9726100, end_position_on_the_genomic_accession=9728274, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340273, CurrentID=0, Status='live', Symbol='ABCB5', Aliases='ABCB5alphabeta, EST422562, ABCB5', description='ATP binding cassette subfamily B member 5', other_designations='ABCB5 P-gp|ATP-binding cassette protein|ATP-binding cassette, sub-family B (MDR/TAP), member 5|P-glycoprotein ABCB5', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20615622, end_position_on_the_genomic_accession=20757014, orientation='plus', exon_count=31, OMIM=611785, no_hdr0=''),
  340274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340274, CurrentID=0, Status='live', Symbol='ASS1P11', Aliases='ASSP11', description='argininosuccinate synthetase 1 pseudogene 11', other_designations='TCAG_1641141|argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 11', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21219978, end_position_on_the_genomic_accession=21221533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340277, CurrentID=0, Status='live', Symbol='FAM221A', Aliases='C7orf46', description='family with sequence similarity 221 member A', other_designations='family with sequence similarity 221, member A|uncharacterized protein C7orf46', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23679463, end_position_on_the_genomic_accession=23703249, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373066, CurrentID=0, Status='live', Symbol='LOC105373066', Aliases='', description='uncharacterized LOC105373066', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45498290, end_position_on_the_genomic_accession=45500983, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340286, CurrentID=0, Status='live', Symbol='FAM183BP', Aliases='FAM183B, THEG6', description='family with sequence similarity 183 member B, pseudogene', other_designations='family with sequence similarity 183, member B|family with sequence similarity 183, member B, pseudogene|testis highly expressed protein 6', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38685346, end_position_on_the_genomic_accession=38687089, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373067, CurrentID=0, Status='live', Symbol='LOC105373067', Aliases='', description='uncharacterized LOC105373067', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45635993, end_position_on_the_genomic_accession=45641953, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340307, CurrentID=0, Status='live', Symbol='CTAGE6', Aliases='CTAGE6P', description='CTAGE family member 6', other_designations='CTAGE family, member 6|CTAGE family, member 6, pseudogene|cutaneous T-cell lymphoma-associated antigen 6|protein cTAGE-6|putative protein cTAGE-6', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143755089, end_position_on_the_genomic_accession=143757750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373070, CurrentID=0, Status='live', Symbol='LOC105373070', Aliases='', description='uncharacterized LOC105373070', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45773796, end_position_on_the_genomic_accession=45786392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373072, CurrentID=0, Status='live', Symbol='LOC105373072', Aliases='', description='uncharacterized LOC105373072', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46011249, end_position_on_the_genomic_accession=46014442, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370852, CurrentID=0, Status='live', Symbol='LOC105370852', Aliases='', description='uncharacterized LOC105370852', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62257100, end_position_on_the_genomic_accession=62264954, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373073, CurrentID=0, Status='live', Symbol='LOC105373073', Aliases='', description='uncharacterized LOC105373073', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373074, CurrentID=0, Status='live', Symbol='LOC105373074', Aliases='', description='uncharacterized LOC105373074', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46161042, end_position_on_the_genomic_accession=46172349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  340335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340335, CurrentID=0, Status='live', Symbol='LOC340335', Aliases='', description='uncharacterized LOC340335', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340340, CurrentID=0, Status='live', Symbol='LOC340340', Aliases='', description='uncharacterized LOC340340', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373076, CurrentID=0, Status='live', Symbol='LOC105373076', Aliases='', description='uncharacterized LOC105373076', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340348, CurrentID=0, Status='live', Symbol='TSPAN33', Aliases='PEN, PEN., TSPAN-33', description='tetraspanin 33', other_designations='penumbra|proerythroblast new membrane', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129144716, end_position_on_the_genomic_accession=129169694, orientation='plus', exon_count=8, OMIM=610120, no_hdr0=''),
  340351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340351, CurrentID=0, Status='live', Symbol='AGBL3', Aliases='CCP3', description='ATP/GTP binding protein-like 3', other_designations='ATP/GTP-binding protein-like 3', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134986508, end_position_on_the_genomic_accession=135135778, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100421487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421487, CurrentID=0, Status='live', Symbol='FAM58CP', Aliases='FAM58AP1, FAM58Y', description='family with sequence similarity 58 member C, pseudogene', other_designations='family with sequence similarity 58, member C, pseudogene', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26626323, end_position_on_the_genomic_accession=26627368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340357, CurrentID=0, Status='live', Symbol='LOC340357', Aliases='', description='uncharacterized LOC340357', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12766062, end_position_on_the_genomic_accession=12811401, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  340359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340359, CurrentID=0, Status='live', Symbol='KLHL38', Aliases='C8ORFK36', description='kelch like family member 38', other_designations='kelch-like 38|kelch-like family member 38', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123645505, end_position_on_the_genomic_accession=123653866, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379432, CurrentID=0, Status='live', Symbol='LOC105379432', Aliases='', description='uncharacterized LOC105379432', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89647372, end_position_on_the_genomic_accession=89649440, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100652807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652807, CurrentID=0, Status='live', Symbol='LOC100652807', Aliases='', description='putative uncharacterized protein C3orf53', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28575591, end_position_on_the_genomic_accession=28576491, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  340371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340371, CurrentID=0, Status='live', Symbol='NRBP2', Aliases='TRG16, pp9320', description='nuclear receptor binding protein 2', other_designations='transformation-related protein 16', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143830016, end_position_on_the_genomic_accession=143840974, orientation='minus', exon_count=20, OMIM=615563, no_hdr0=''),
  340385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340385, CurrentID=0, Status='live', Symbol='ZNF517', Aliases='', description='zinc finger protein 517', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144798876, end_position_on_the_genomic_accession=144809144, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100130203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130203, CurrentID=0, Status='live', Symbol='LOC100130203', Aliases='', description='cell adhesion associated, oncogene regulated pseudogene', other_designations='Cdon homolog pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89731014, end_position_on_the_genomic_accession=89734783, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340390, CurrentID=0, Status='live', Symbol='WDR97', Aliases='KIAA1875', description='WD repeat domain 97', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144107726, end_position_on_the_genomic_accession=144118315, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  105372449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372449, CurrentID=0, Status='live', Symbol='LOC105372449', Aliases='', description='uncharacterized LOC105372449', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52230529, end_position_on_the_genomic_accession=52264453, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  340393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340393, CurrentID=0, Status='live', Symbol='TMEM249', Aliases='C8ORFK29', description='transmembrane protein 249', other_designations='putative transmembrane protein C8orfK29', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144353226, end_position_on_the_genomic_accession=144355609, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373085, CurrentID=0, Status='live', Symbol='LOC105373085', Aliases='', description='uncharacterized LOC105373085', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49047480, end_position_on_the_genomic_accession=49052405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927248, CurrentID=0, Status='live', Symbol='LOC101927248', Aliases='', description='uncharacterized LOC101927248', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94704744, end_position_on_the_genomic_accession=94708486, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391048, CurrentID=0, Status='live', Symbol='LOC391048', Aliases='', description='cysteine and histidine rich domain containing 1 pseudogene', other_designations='cysteine and histidine-rich domain (CHORD) containing 1 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70530409, end_position_on_the_genomic_accession=70532474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928865, CurrentID=0, Status='live', Symbol='LOC101928865', Aliases='', description='uncharacterized LOC101928865', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78423982, end_position_on_the_genomic_accession=78429836, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373087, CurrentID=0, Status='live', Symbol='LOC105373087', Aliases='', description='uncharacterized LOC105373087', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49497216, end_position_on_the_genomic_accession=49524246, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319129, CurrentID=0, Status='live', Symbol='RPSAP5', Aliases='LAMR1P5, RPSA_22_1412', description='ribosomal protein SA pseudogene 5', other_designations='laminin receptor 1 pseudogene 5', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103373963, end_position_on_the_genomic_accession=103374992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340419, CurrentID=0, Status='live', Symbol='RSPO2', Aliases='CRISTIN2', description='R-spondin 2', other_designations='R-spondin 2 homolog|roof plate-specific spondin-2', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107899316, end_position_on_the_genomic_accession=108083685, orientation='minus', exon_count=7, OMIM=610575, no_hdr0=''),
  100130209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130209, CurrentID=0, Status='live', Symbol='LOC100130209', Aliases='', description='charged multivesicular body protein 4A pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81418615, end_position_on_the_genomic_accession=81419356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375220, CurrentID=0, Status='live', Symbol='LOC105375220', Aliases='', description='uncharacterized LOC105375220', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30697999, end_position_on_the_genomic_accession=30750346, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373091, CurrentID=0, Status='live', Symbol='LOC105373091', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49849635, end_position_on_the_genomic_accession=49851067, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  340441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340441, CurrentID=0, Status='live', Symbol='POTEA', Aliases='A26A1, CT104.3, POTE-8, POTE8', description='POTE ankyrin domain family member A', other_designations='ANKRD26-like family A, member 1|POTE ankyrin domain family, member A|cancer/testis antigen family 104, member 3|prostate, ovary, testis-expressed protein on chromosome 8|protein expressed in prostate, ovary, testis, and placenta 8', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43292370, end_position_on_the_genomic_accession=43366407, orientation='plus', exon_count=14, OMIM=608915, no_hdr0=''),
  340443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340443, CurrentID=0, Status='live', Symbol='RPS5P5', Aliases='RPS5_2_920', description='ribosomal protein S5 pseudogene 5', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80300797, end_position_on_the_genomic_accession=80301536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370856, CurrentID=0, Status='live', Symbol='LOC105370856', Aliases='', description='uncharacterized LOC105370856', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62379226, end_position_on_the_genomic_accession=62390495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340451, CurrentID=0, Status='live', Symbol='CPP', Aliases='', description='ceruloplasmin (ferroxidase) pseudogene', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91157278, end_position_on_the_genomic_accession=91159922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373094, CurrentID=0, Status='live', Symbol='LOC105373094', Aliases='', description='uncharacterized LOC105373094', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41969247, end_position_on_the_genomic_accession=41975904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373095, CurrentID=0, Status='live', Symbol='LOC105373095', Aliases='', description='uncharacterized LOC105373095', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50208461, end_position_on_the_genomic_accession=50210174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  340460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340460, CurrentID=0, Status='live', Symbol='KRT18P24', Aliases='', description='keratin 18 pseudogene 24', other_designations='', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79036353, end_position_on_the_genomic_accession=79037748, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479445, CurrentID=0, Status='live', Symbol='RN7SL612P', Aliases='', description='RNA, 7SL, cytoplasmic 612, pseudogene', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373098, CurrentID=0, Status='live', Symbol='LOC105373098', Aliases='', description='uncharacterized LOC105373098', other_designations='CTA-384D8.35', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50542303, end_position_on_the_genomic_accession=50544261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377687, CurrentID=0, Status='live', Symbol='LOC105377687', Aliases='', description='uncharacterized LOC105377687', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159698588, end_position_on_the_genomic_accession=159705348, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340481, CurrentID=0, Status='live', Symbol='ZDHHC21', Aliases='9130404H11Rik, DHHC-21, DHHC21, DNZ1, HSPC097', description='zinc finger, DHHC-type containing 21', other_designations='probable palmitoyltransferase ZDHHC21|zinc finger DHHC domain-containing protein 21|zinc finger, DHHC domain containing 21', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14555501, end_position_on_the_genomic_accession=14693482, orientation='minus', exon_count=13, OMIM=614605, no_hdr0=''),
  100130219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130219, CurrentID=0, Status='live', Symbol='LOC100130219', Aliases='', description='uncharacterized LOC100130219', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=1084196, end_position_on_the_genomic_accession=1100319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340485, CurrentID=0, Status='live', Symbol='ACER2', Aliases='ALKCDase2, ASAH3L', description='alkaline ceramidase 2', other_designations='alkCDase 2|alkaline CDase 2|ceramide hydrolase|haCER2', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19406011, end_position_on_the_genomic_accession=19452505, orientation='plus', exon_count=12, OMIM=613492, no_hdr0=''),
  105373100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373100, CurrentID=0, Status='live', Symbol='LOC105373100', Aliases='', description='uncharacterized LOC105373100', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50735829, end_position_on_the_genomic_accession=50738169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373101, CurrentID=0, Status='live', Symbol='LOC105373101', Aliases='', description='uncharacterized LOC105373101', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50710350, end_position_on_the_genomic_accession=50711608, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  340499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340499, CurrentID=0, Status='live', Symbol='IMPDH1P1', Aliases='bA537H15.4', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 1', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34132459, end_position_on_the_genomic_accession=34134795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340501, CurrentID=0, Status='live', Symbol='TCEA1P3', Aliases='', description='transcription elongation factor A (SII), 1 pseudogene 3', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38478474, end_position_on_the_genomic_accession=38479347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100217367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100217367, CurrentID=0, Status='live', Symbol='LOC100217367', Aliases='', description='chromosome 11 open reading frame 73 pseudogene', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23350995, end_position_on_the_genomic_accession=23351654, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100217370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100217370, CurrentID=0, Status='live', Symbol='DEL1Q21', Aliases='C1DELq21', description='Chromosome 1q21.1 deletion syndrome', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612474, no_hdr0=''),
  100217371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100217371, CurrentID=0, Status='live', Symbol='DUP1Q21', Aliases='C1DUPq21', description='Chromosome 1q21.1 duplication syndrome', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612475, no_hdr0=''),
  340508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340508, CurrentID=0, Status='live', Symbol='GAS2L1P2', Aliases='', description='growth arrest specific 2 like 1 pseudogene 2', other_designations='growth arrest-specific 2 like 1 pseudogene 2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97075671, end_position_on_the_genomic_accession=97081945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340512, CurrentID=0, Status='live', Symbol='LOC340512', Aliases='', description='uncharacterized LOC340512', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106974833, end_position_on_the_genomic_accession=107102988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373104, CurrentID=0, Status='live', Symbol='LOC105373104', Aliases='', description='uncharacterized LOC105373104', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=1744771, end_position_on_the_genomic_accession=1745632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  340515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340515, CurrentID=0, Status='live', Symbol='LINC01501', Aliases='', description='long intergenic non-protein coding RNA 1501', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90462432, end_position_on_the_genomic_accession=90582746, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130225, CurrentID=0, Status='live', Symbol='RPL13P11', Aliases='RPL13_3_914', description='ribosomal protein L13 pseudogene 11', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70506255, end_position_on_the_genomic_accession=70506818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100217387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100217387, CurrentID=0, Status='live', Symbol='AGA3', Aliases='', description='Alopecia, androgenetic, 3', other_designations='', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612421, no_hdr0=''),
  105373106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373106, CurrentID=0, Status='live', Symbol='LOC105373106', Aliases='', description='uncharacterized LOC105373106', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224701684, end_position_on_the_genomic_accession=224719087, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  340526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340526, CurrentID=0, Status='live', Symbol='RGAG4', Aliases='6430402L03Rik, MAR5, MART5', description='retrotransposon gag domain containing 4', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72127110, end_position_on_the_genomic_accession=72131901, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  340527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340527, CurrentID=0, Status='live', Symbol='NHSL2', Aliases='', description='NHS like 2', other_designations='NHS-like 2', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71911073, end_position_on_the_genomic_accession=72153286, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  340529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340529, CurrentID=0, Status='live', Symbol='PABPC1L2A', Aliases='RBM32A', description='poly(A) binding protein, cytoplasmic 1-like 2A', other_designations='RNA binding motif protein 32A|RNA-binding motif protein 32|RNA-binding protein 32', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73077338, end_position_on_the_genomic_accession=73079512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  340533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340533, CurrentID=0, Status='live', Symbol='KIAA2022', Aliases='MRX98, XPN', description='KIAA2022', other_designations='XLMR protein related to neurite extension|XLMR-related protein, neurite extension', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74732856, end_position_on_the_genomic_accession=74925452, orientation='minus', exon_count=5, OMIM=300524, no_hdr0=''),
  100421493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421493, CurrentID=0, Status='live', Symbol='LOC100421493', Aliases='', description='family with sequence similarity 64, member A pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26873472, end_position_on_the_genomic_accession=26874398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373108, CurrentID=0, Status='live', Symbol='LOC105373108', Aliases='', description='uncharacterized LOC105373108', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340542, CurrentID=0, Status='live', Symbol='BEX5', Aliases='NGFRAP1L1', description='brain expressed X-linked 5', other_designations='BEX family member 5|NGFRAP1-like 1|NGFRAP1-like protein 1|brain expressed, X-linked 5|brain-expressed X-linked protein 5|nerve growth factor receptor-associated protein 2', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102153707, end_position_on_the_genomic_accession=102156014, orientation='minus', exon_count=4, OMIM=300693, no_hdr0=''),
  340543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340543, CurrentID=0, Status='live', Symbol='TCEAL5', Aliases='WEX4', description='transcription elongation factor A (SII)-like 5', other_designations='TCEA-like protein 5|transcription elongation factor S-II protein-like 5', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103273690, end_position_on_the_genomic_accession=103276869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340544, CurrentID=0, Status='live', Symbol='MORF4L2-AS1', Aliases='', description='MORF4L2 antisense RNA 1', other_designations='MORF4L2 antisense RNA 1 (non-protein coding)', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103687284, end_position_on_the_genomic_accession=103692560, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340547, CurrentID=0, Status='live', Symbol='VSIG1', Aliases='1700062D20Rik, GPA34, dJ889N15.1', description='V-set and immunoglobulin domain containing 1', other_designations='cell surface A33 antigen|glycoprotein A34', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108018906, end_position_on_the_genomic_accession=108079184, orientation='plus', exon_count=10, OMIM=300620, no_hdr0=''),
  100652813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652813, CurrentID=0, Status='live', Symbol='LOC100652813', Aliases='', description='zinc finger protein 138 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15254419, end_position_on_the_genomic_accession=15270928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375224, CurrentID=0, Status='live', Symbol='LOC105375224', Aliases='', description='uncharacterized LOC105375224', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31413754, end_position_on_the_genomic_accession=31416904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927253, CurrentID=0, Status='live', Symbol='LOC101927253', Aliases='', description='uncharacterized LOC101927253', other_designations='CTB-22K21.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50627030, end_position_on_the_genomic_accession=50631940, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  340554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340554, CurrentID=0, Status='live', Symbol='ZC3H12B', Aliases='CXorf32, MCPIP2', description='zinc finger CCCH-type containing 12B', other_designations='MCP-induced protein 2|zinc finger CCCH domain-containing protein 12B', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65034799, end_position_on_the_genomic_accession=65507887, orientation='plus', exon_count=14, OMIM=300889, no_hdr0=''),
  646583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646583, CurrentID=0, Status='live', Symbol='RPS4XP18', Aliases='RPS4P18, RPS4X_10_1587', description='ribosomal protein S4X pseudogene 18', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22586404, end_position_on_the_genomic_accession=22587278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130232, CurrentID=0, Status='live', Symbol='LOC100130232', Aliases='', description='LP2209', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=105394762, end_position_on_the_genomic_accession=105395809, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  340562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340562, CurrentID=0, Status='live', Symbol='SATL1', Aliases='', description='spermidine/spermine N1-acetyl transferase-like 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85092286, end_position_on_the_genomic_accession=85116178, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100421494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421494, CurrentID=0, Status='live', Symbol='LOC100421494', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76885599, end_position_on_the_genomic_accession=76886554, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373113, CurrentID=0, Status='live', Symbol='LOC105373113', Aliases='', description='uncharacterized LOC105373113', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226083576, end_position_on_the_genomic_accession=226090312, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390358, CurrentID=0, Status='live', Symbol='LOC390358', Aliases='', description='ret finger protein-like 2 pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107881058, end_position_on_the_genomic_accession=107883707, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340571, CurrentID=0, Status='live', Symbol='SIAH1P1', Aliases='SIAH1L', description='siah E3 ubiquitin protein ligase 1 pseudogene 1', other_designations='seven in absentia homolog 1 pseudogene 1', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35626013, end_position_on_the_genomic_accession=35628561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373114, CurrentID=0, Status='live', Symbol='LOC105373114', Aliases='', description='uncharacterized LOC105373114', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226233494, end_position_on_the_genomic_accession=226265748, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  340578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340578, CurrentID=0, Status='live', Symbol='DCAF12L2', Aliases='WDR40C', description='DDB1 and CUL4 associated factor 12-like 2', other_designations='WD repeat-containing protein 40C', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126163499, end_position_on_the_genomic_accession=126166097, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373115, CurrentID=0, Status='live', Symbol='LOC105373115', Aliases='', description='uncharacterized LOC105373115', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226516083, end_position_on_the_genomic_accession=226520800, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927254, CurrentID=0, Status='live', Symbol='LOC101927254', Aliases='', description='uncharacterized LOC101927254', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31149950, end_position_on_the_genomic_accession=31232298, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  340581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340581, CurrentID=0, Status='live', Symbol='LOC340581', Aliases='', description='uncharacterized LOC340581', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134949550, end_position_on_the_genomic_accession=134953382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373116, CurrentID=0, Status='live', Symbol='LOC105373116', Aliases='', description='uncharacterized LOC105373116', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226535551, end_position_on_the_genomic_accession=226543273, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340591, CurrentID=0, Status='live', Symbol='CA5BP1', Aliases='CA5BL, CA5BP, PRO2325', description='carbonic anhydrase VB pseudogene 1', other_designations='carbonic dehydratase', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15674916, end_position_on_the_genomic_accession=15703351, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379590, CurrentID=0, Status='live', Symbol='LOC105379590', Aliases='', description='uncharacterized LOC105379590', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340595, CurrentID=0, Status='live', Symbol='ZCCHC16', Aliases='Mar4, Mart4', description='zinc finger, CCHC domain containing 16', other_designations='mammalian retrotransposon-derived protein 4', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112083025, end_position_on_the_genomic_accession=112457245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  340596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340596, CurrentID=0, Status='live', Symbol='LHFPL1', Aliases='', description='lipoma HMGIC fusion partner-like 1', other_designations='LHFP-like protein 1', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112625637, end_position_on_the_genomic_accession=112680178, orientation='minus', exon_count=5, OMIM=300566, no_hdr0=''),
  105373118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373118, CurrentID=0, Status='live', Symbol='LOC105373118', Aliases='', description='uncharacterized LOC105373118', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226847144, end_position_on_the_genomic_accession=226857766, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340598, CurrentID=0, Status='live', Symbol='KRT18P48', Aliases='', description='keratin 18 pseudogene 48', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153603221, end_position_on_the_genomic_accession=153605150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340600, CurrentID=0, Status='live', Symbol='ATF4P2', Aliases='', description='activating transcription factor 4 pseudogene 2', other_designations='activating transcription factor 4 (tax-responsive enhancer element B67) pseudogene|cyclic-AMP-dependent transcription factor ATF-4 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154580646, end_position_on_the_genomic_accession=154582086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340602, CurrentID=0, Status='live', Symbol='CXorf67', Aliases='', description='chromosome X open reading frame 67', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51406915, end_position_on_the_genomic_accession=51408838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373119, CurrentID=0, Status='live', Symbol='LOC105373119', Aliases='', description='uncharacterized LOC105373119', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226860172, end_position_on_the_genomic_accession=226870519, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130240, CurrentID=0, Status='live', Symbol='RPS5P2', Aliases='RPS5_4_1727', description='ribosomal protein S5 pseudogene 2', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34135382, end_position_on_the_genomic_accession=34136128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128077, CurrentID=0, Status='live', Symbol='LOC100128077', Aliases='', description='uncharacterized LOC100128077', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340618, CurrentID=0, Status='live', Symbol='FAM41AY1', Aliases='FAM41AY', description='family with sequence similarity 41 member A, Y-linked 1', other_designations='family with sequence similarity 41, member A, Y-linked 1', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17500958, end_position_on_the_genomic_accession=17516742, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105373123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373123, CurrentID=0, Status='live', Symbol='LOC105373123', Aliases='', description='uncharacterized LOC105373123', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3455913, end_position_on_the_genomic_accession=3466739, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644198, CurrentID=0, Status='live', Symbol='RPLP2P2', Aliases='RPLP2_1_633', description='ribosomal protein, large, P2 pseudogene 2', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157931179, end_position_on_the_genomic_accession=157931620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375227, CurrentID=0, Status='live', Symbol='LOC105375227', Aliases='', description='uncharacterized LOC105375227', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33637841, end_position_on_the_genomic_accession=33652754, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129074, CurrentID=0, Status='live', Symbol='PRKRIRP5', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 5', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11514192, end_position_on_the_genomic_accession=11516389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340654, CurrentID=0, Status='live', Symbol='LIPM', Aliases='LIPL3, bA304I5.1', description='lipase, family member M', other_designations='lipase M|lipase-like abhydrolase domain-containing protein 3|lipase-like, ab-hydrolase domain containing 3', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88802730, end_position_on_the_genomic_accession=88822022, orientation='plus', exon_count=11, OMIM=613923, no_hdr0=''),
  105373128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373128, CurrentID=0, Status='live', Symbol='LOC105373128', Aliases='', description='uncharacterized LOC105373128', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11463677, end_position_on_the_genomic_accession=11469150, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373129, CurrentID=0, Status='live', Symbol='LOC105373129', Aliases='', description='uncharacterized LOC105373129', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11425838, end_position_on_the_genomic_accession=11433652, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  340665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340665, CurrentID=0, Status='live', Symbol='CYP26C1', Aliases='FFDD4', description='cytochrome P450 family 26 subfamily C member 1', other_designations='cytochrome P450, family 26, subfamily C, polypeptide 1', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93061264, end_position_on_the_genomic_accession=93068697, orientation='plus', exon_count=6, OMIM=608428, no_hdr0=''),
  100130251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130251, CurrentID=0, Status='live', Symbol='NXNP1', Aliases='', description='nucleoredoxin pseudogene 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218843762, end_position_on_the_genomic_accession=218882551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373133, CurrentID=0, Status='live', Symbol='LOC105373133', Aliases='', description='uncharacterized LOC105373133', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13044544, end_position_on_the_genomic_accession=13140666, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130256, CurrentID=0, Status='live', Symbol='LOC100130256', Aliases='', description='uncharacterized LOC100130256', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170619392, end_position_on_the_genomic_accession=170683059, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  340706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340706, CurrentID=0, Status='live', Symbol='VWA2', Aliases='AMACO, CCSP-2, NET42', description='von Willebrand factor A domain containing 2', other_designations='a domain-containing protein similar to matrilin and collagen|colon cancer diagnostic marker|colon cancer secreted protein 2', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114239137, end_position_on_the_genomic_accession=114294500, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105373137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373137, CurrentID=0, Status='live', Symbol='LOC105373137', Aliases='', description='uncharacterized LOC105373137', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13368909, end_position_on_the_genomic_accession=13403006, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100129852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129852, CurrentID=0, Status='live', Symbol='LOC100129852', Aliases='', description='uncharacterized LOC100129852', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=105384169, end_position_on_the_genomic_accession=105401270, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  340719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340719, CurrentID=0, Status='live', Symbol='NANOS1', Aliases='EC_Rep1a, NOS-1, NOS1, SPGF12', description='nanos homolog 1 (Drosophila)', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119029716, end_position_on_the_genomic_accession=119034342, orientation='plus', exon_count=1, OMIM=608226, no_hdr0=''),
  105373139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373139, CurrentID=0, Status='live', Symbol='LOC105373139', Aliases='', description='uncharacterized LOC105373139', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646612, CurrentID=0, Status='live', Symbol='EEF1A1P25', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 25', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138825063, end_position_on_the_genomic_accession=138826326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927259, CurrentID=0, Status='live', Symbol='LOC101927259', Aliases='', description='uncharacterized LOC101927259', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128552590, end_position_on_the_genomic_accession=128553593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373142, CurrentID=0, Status='live', Symbol='LOC105373142', Aliases='', description='uncharacterized LOC105373142', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17635559, end_position_on_the_genomic_accession=17651719, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340745, CurrentID=0, Status='live', Symbol='LRIT2', Aliases='LRRC22', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 2', other_designations='leucine rich repeat containing 22|leucine-rich repeat-containing protein 22', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84220485, end_position_on_the_genomic_accession=84225589, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  340749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340749, CurrentID=0, Status='live', Symbol='RPL7AP8', Aliases='RPL7A_24_1075, bA141D8.1', description='ribosomal protein L7a pseudogene 8', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86630766, end_position_on_the_genomic_accession=86631628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130264, CurrentID=0, Status='live', Symbol='LOC100130264', Aliases='', description='uncharacterized LOC100130264', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19242302, end_position_on_the_genomic_accession=19284596, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373145, CurrentID=0, Status='live', Symbol='LOC105373145', Aliases='', description='uncharacterized LOC105373145', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18832058, end_position_on_the_genomic_accession=18833960, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373147, CurrentID=0, Status='live', Symbol='LOC105373147', Aliases='', description='uncharacterized LOC105373147', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22225868, end_position_on_the_genomic_accession=22254078, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130268, CurrentID=0, Status='live', Symbol='LOC100130268', Aliases='', description='uncharacterized LOC100130268', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75234724, end_position_on_the_genomic_accession=75298661, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  340780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340780, CurrentID=0, Status='live', Symbol='IMPDH1P5', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 5', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77779595, end_position_on_the_genomic_accession=77781955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340784, CurrentID=0, Status='live', Symbol='HMX3', Aliases='NKX-5.1, NKX5.1, Nkx5-1', description='H6 family homeobox 3', other_designations='homeo box (H6 family) 3|homeobox (H6 family) 3|homeobox protein H6 family member 3|homeobox protein Nkx-5.1', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123136051, end_position_on_the_genomic_accession=123137731, orientation='plus', exon_count=2, OMIM=613380, no_hdr0=''),
  644203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644203, CurrentID=0, Status='live', Symbol='LOC644203', Aliases='', description='glutaryl-CoA dehydrogenase, mitochondrial-like', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103503696, end_position_on_the_genomic_accession=103504933, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130270, CurrentID=0, Status='live', Symbol='LOC100130270', Aliases='', description='NDUFA4, mitochondrial complex associated pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa pseudogene', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64918369, end_position_on_the_genomic_accession=64918841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375232, CurrentID=0, Status='live', Symbol='LOC105375232', Aliases='', description='uncharacterized LOC105375232', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35908452, end_position_on_the_genomic_accession=35912190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373151, CurrentID=0, Status='live', Symbol='LOC105373151', Aliases='', description='uncharacterized LOC105373151', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27983892, end_position_on_the_genomic_accession=28018116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373152, CurrentID=0, Status='live', Symbol='LOC105373152', Aliases='', description='uncharacterized LOC105373152', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=28065865, end_position_on_the_genomic_accession=28073183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421502, CurrentID=0, Status='live', Symbol='MARK2P2', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10012384, end_position_on_the_genomic_accession=10013133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373153, CurrentID=0, Status='live', Symbol='LOC105373153', Aliases='', description='uncharacterized LOC105373153', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=33726356, end_position_on_the_genomic_accession=34076569, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  340811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340811, CurrentID=0, Status='live', Symbol='AKR1C8P', Aliases='AKR1CL1', description='aldo-keto reductase family 1, member C8, pseudogene', other_designations='aldo-keto reductase family 1, member C-like 1|protein RAKc', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5154460, end_position_on_the_genomic_accession=5185187, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  646626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646626, CurrentID=0, Status='live', Symbol='LOC646626', Aliases='', description='uncharacterized LOC646626', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85276358, end_position_on_the_genomic_accession=85278088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130275, CurrentID=0, Status='live', Symbol='TFAP2A-AS1', Aliases='', description='TFAP2A antisense RNA 1', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10412318, end_position_on_the_genomic_accession=10416169, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130276, CurrentID=0, Status='live', Symbol='LOC100130276', Aliases='', description='uncharacterized LOC100130276', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=11065537, end_position_on_the_genomic_accession=11065902, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129080, CurrentID=0, Status='live', Symbol='TDGF1P7', Aliases='CR-7, CRIPTO-7, TDGF7', description='teratocarcinoma-derived growth factor 1 pseudogene 7', other_designations='teratocarcinoma-derived growth factor 7, pseudogene', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39534496, end_position_on_the_genomic_accession=39536333, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130278, CurrentID=0, Status='live', Symbol='LOC100130278', Aliases='', description='uncharacterized LOC100130278', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=46925520, end_position_on_the_genomic_accession=46926031, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873064, CurrentID=0, Status='live', Symbol='AIMP1P2', Aliases='', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 1 pseudogene 2', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172885947, end_position_on_the_genomic_accession=172886258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873065, CurrentID=0, Status='live', Symbol='PTCHD1-AS', Aliases='DDX53-AS1, PTCHD1AS1, PTCHD1AS2', description='PTCHD1 antisense RNA (head to head)', other_designations='DDX53 antisense RNA 1 (non-protein coding)', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=22259797, end_position_on_the_genomic_accession=23293146, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372452, CurrentID=0, Status='live', Symbol='LOC105372452', Aliases='', description='uncharacterized LOC105372452', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52744589, end_position_on_the_genomic_accession=52754098, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373159, CurrentID=0, Status='live', Symbol='LOC105373159', Aliases='', description='uncharacterized LOC105373159', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229257840, end_position_on_the_genomic_accession=229271044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340844, CurrentID=0, Status='live', Symbol='AHCYP1', Aliases='', description='adenosylhomocysteinase pseudogene 1', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46223058, end_position_on_the_genomic_accession=46224617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375234, CurrentID=0, Status='live', Symbol='LOC105375234', Aliases='', description='uncharacterized LOC105375234', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36053248, end_position_on_the_genomic_accession=36105035, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  648570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648570, CurrentID=0, Status='live', Symbol='LOC648570', Aliases='', description='uncharacterized LOC648570', other_designations='', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=16785576, end_position_on_the_genomic_accession=16787433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379599, CurrentID=0, Status='live', Symbol='LOC105379599', Aliases='', description='uncharacterized LOC105379599', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421504, CurrentID=0, Status='live', Symbol='LOC100421504', Aliases='', description='MAP/microtubule affinity-regulating kinase 4 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12860687, end_position_on_the_genomic_accession=12862250, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130283, CurrentID=0, Status='live', Symbol='LOC100130283', Aliases='', description='uncharacterized LOC100130283', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=8850226, end_position_on_the_genomic_accession=8856654, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373165, CurrentID=0, Status='live', Symbol='LOC105373165', Aliases='', description='uncharacterized LOC105373165', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230549802, end_position_on_the_genomic_accession=230551176, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375235, CurrentID=0, Status='live', Symbol='LOC105375235', Aliases='', description='uncharacterized LOC105375235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36883759, end_position_on_the_genomic_accession=36894836, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373166, CurrentID=0, Status='live', Symbol='LOC105373166', Aliases='', description='uncharacterized LOC105373166', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230716765, end_position_on_the_genomic_accession=230745608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  340888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340888, CurrentID=0, Status='live', Symbol='AKR1B10P1', Aliases='', description='aldo-keto reductase family 1, member B10 pseudogene 1', other_designations='aldo-keto reductase family 1, member B10 (aldose reductase) pseudogene', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67750229, end_position_on_the_genomic_accession=67751429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646639, CurrentID=0, Status='live', Symbol='LOC646639', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog (yeast) pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33501944, end_position_on_the_genomic_accession=33503327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340895, CurrentID=0, Status='live', Symbol='MALRD1', Aliases='C10orf112, DIET1, bA265G8.2', description='MAM and LDL receptor class A domain containing 1', other_designations='MAM and LDL-receptor class A domain-containing protein C10orf112|novel MAM domain containing protein', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19046931, end_position_on_the_genomic_accession=19734478, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  105373168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373168, CurrentID=0, Status='live', Symbol='LOC105373168', Aliases='', description='uncharacterized LOC105373168', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230879004, end_position_on_the_genomic_accession=230890459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  340900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340900, CurrentID=0, Status='live', Symbol='EBLN1', Aliases='EBLN-1', description='endogenous Bornavirus-like nucleoprotein 1', other_designations='endogenous Borna-like N element 1|endogenous Borna-like N element-1|endogenous Borna-like N element-containing protein 1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22208814, end_position_on_the_genomic_accession=22209983, orientation='minus', exon_count=1, OMIM=613249, no_hdr0=''),
  646641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646641, CurrentID=0, Status='live', Symbol='RPL7L1P7', Aliases='', description='ribosomal protein L7 like 1 pseudogene 7', other_designations='ribosomal protein L7-like 1 pseudogene 7', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139081660, end_position_on_the_genomic_accession=139082377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692217, CurrentID=0, Status='live', Symbol='SNORD114@', Aliases='14q(II)', description='small nucleolar RNA, C/D box 114 cluster', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646644, CurrentID=0, Status='live', Symbol='LOC646644', Aliases='', description='LLP homolog, long-term synaptic facilitation (Aplysia) pseudogene', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221944164, end_position_on_the_genomic_accession=221944811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373174, CurrentID=0, Status='live', Symbol='LOC105373174', Aliases='', description='uncharacterized LOC105373174', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38779501, end_position_on_the_genomic_accession=38800658, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873165, CurrentID=0, Status='live', Symbol='SRGAP2-AS1', Aliases='', description='SRGAP2 antisense RNA 1', other_designations='SRGAP2 antisense RNA 1 (non-protein coding)', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121360156, end_position_on_the_genomic_accession=121397905, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873166, CurrentID=0, Status='live', Symbol='SETP7', Aliases='', description='SET pseudogene 7', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107271385, end_position_on_the_genomic_accession=107271843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873168, CurrentID=0, Status='live', Symbol='MTCYBP2', Aliases='', description='MT-CYB pseudogene 2', other_designations='MTCYB pseudogene 2', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18872102, end_position_on_the_genomic_accession=18872272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103625681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103625681, CurrentID=0, Status='live', Symbol='LLPH-AS1', Aliases='', description='LLPH antisense RNA 1 (head to head)', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66130751, end_position_on_the_genomic_accession=66134449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873170, CurrentID=0, Status='live', Symbol='MTND1P2', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 2', other_designations='MT-ND1 pseudogene 2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64111755, end_position_on_the_genomic_accession=64112572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  103625683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103625683, CurrentID=0, Status='live', Symbol='LINC00456', Aliases='', description='long intergenic non-protein coding RNA 456', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97172382, end_position_on_the_genomic_accession=97179622, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103625684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103625684, CurrentID=0, Status='live', Symbol='RNU6-2', Aliases='U6-2', description='RNA, U6 small nuclear 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1021522, end_position_on_the_genomic_accession=1021628, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873173, CurrentID=0, Status='live', Symbol='MTND1P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 6', other_designations='MT-ND1 pseudogene 6', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33014081, end_position_on_the_genomic_accession=33015025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873174, CurrentID=0, Status='live', Symbol='MTND1P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 5', other_designations='MT-ND1 pseudogene 5', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103085667, end_position_on_the_genomic_accession=103086593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873175, CurrentID=0, Status='live', Symbol='MTND2P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 4', other_designations='MT-ND2 pseudogene 4', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64110547, end_position_on_the_genomic_accession=64111550, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873176, CurrentID=0, Status='live', Symbol='MTND2P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 5', other_designations='MT-ND2 pseudogene 5', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141802470, end_position_on_the_genomic_accession=141803487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873177, CurrentID=0, Status='live', Symbol='MTND2P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 6', other_designations='MT-ND2 pseudogene 6', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57186414, end_position_on_the_genomic_accession=57187448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873178, CurrentID=0, Status='live', Symbol='MTND2P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 8', other_designations='MT-ND2 pseudogene 8', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78741722, end_position_on_the_genomic_accession=78742744, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873179, CurrentID=0, Status='live', Symbol='MTND2P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 7', other_designations='MT-ND2 pseudogene 7', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133754739, end_position_on_the_genomic_accession=133755449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873180, CurrentID=0, Status='live', Symbol='MTND3P2', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 2', other_designations='MT-ND3 pseudogene 2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64105654, end_position_on_the_genomic_accession=64105995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873181, CurrentID=0, Status='live', Symbol='MTND3P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 4', other_designations='MT-ND3 pseudogene 4', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101793457, end_position_on_the_genomic_accession=101793735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873182, CurrentID=0, Status='live', Symbol='MTND3P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 6', other_designations='MT-ND3 pseudogene 6', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106896477, end_position_on_the_genomic_accession=106896812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873183, CurrentID=0, Status='live', Symbol='MTND3P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 7', other_designations='MT-ND3 pseudogene 7', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166160773, end_position_on_the_genomic_accession=166161114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873184, CurrentID=0, Status='live', Symbol='MTND3P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 9', other_designations='MT-ND3 pseudogene 9', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143098484, end_position_on_the_genomic_accession=143098819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873185, CurrentID=0, Status='live', Symbol='MTND3P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 8', other_designations='MT-ND3 pseudogene 8', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235541757, end_position_on_the_genomic_accession=235542092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873186, CurrentID=0, Status='live', Symbol='MTND3P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 5', other_designations='MT-ND3 pseudogene 5', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25720099, end_position_on_the_genomic_accession=25720431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873187, CurrentID=0, Status='live', Symbol='MTND3P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 3', other_designations='MT-ND3 pseudogene 3', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155456772, end_position_on_the_genomic_accession=155457107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873188, CurrentID=0, Status='live', Symbol='MTND4P2', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 2', other_designations='MT-ND4 pseudogene 2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64104093, end_position_on_the_genomic_accession=64105295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873189, CurrentID=0, Status='live', Symbol='MTND4P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 3', other_designations='MT-ND4 pseudogene 3', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=68266319, end_position_on_the_genomic_accession=68266862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873190, CurrentID=0, Status='live', Symbol='MTND4P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 4', other_designations='MT-ND4 pseudogene 4', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873191, CurrentID=0, Status='live', Symbol='MTND4P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 5', other_designations='MT-ND4 pseudogene 5', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57192695, end_position_on_the_genomic_accession=57194067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873192, CurrentID=0, Status='live', Symbol='MTND4P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 6', other_designations='MT-ND4 pseudogene 6', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117263895, end_position_on_the_genomic_accession=117264394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873193, CurrentID=0, Status='live', Symbol='MTND4P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 9', other_designations='MT-ND4 pseudogene 9', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25718082, end_position_on_the_genomic_accession=25719745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  340970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340970, CurrentID=0, Status='live', Symbol='TRIM53CP', Aliases='', description='tripartite motif containing 53C, pseudogene', other_designations='ring finger protein 18-like pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48985753, end_position_on_the_genomic_accession=48991881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873195, CurrentID=0, Status='live', Symbol='MTND4P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 7', other_designations='MT-ND4 pseudogene 7', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13353397, end_position_on_the_genomic_accession=13353608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873196, CurrentID=0, Status='live', Symbol='MTND5P6', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 6', other_designations='MT-ND5 pseudogene 6', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57170993, end_position_on_the_genomic_accession=57172769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873197, CurrentID=0, Status='live', Symbol='MTND5P7', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 7', other_designations='MT-ND5 pseudogene 7', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57194324, end_position_on_the_genomic_accession=57196046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873198, CurrentID=0, Status='live', Symbol='MTND5P8', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 8', other_designations='MT-ND5 pseudogene 8', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112372647, end_position_on_the_genomic_accession=112373723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873199, CurrentID=0, Status='live', Symbol='MTND5P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 9', other_designations='MT-ND5 pseudogene 9', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155453054, end_position_on_the_genomic_accession=155454814, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873200, CurrentID=0, Status='live', Symbol='MTND5P4', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 4', other_designations='MT-ND5 pseudogene 4', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17061878, end_position_on_the_genomic_accession=17062165, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873201, CurrentID=0, Status='live', Symbol='MTND5P5', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 5', other_designations='MT-ND5 pseudogene 5', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101972422, end_position_on_the_genomic_accession=101972802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873202, CurrentID=0, Status='live', Symbol='MTCO2P2', Aliases='HsT4010', description='MT-CO2 pseudogene 2', other_designations='MTCO2 pseudogene 2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47853230, end_position_on_the_genomic_accession=47853472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873203, CurrentID=0, Status='live', Symbol='MTCO2P3', Aliases='', description='MT-CO2 pseudogene 3', other_designations='MTCO2 pseudogene 3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57204383, end_position_on_the_genomic_accession=57204804, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340980, CurrentID=0, Status='live', Symbol='OR52B6', Aliases='OR11-47', description='olfactory receptor family 52 subfamily B member 6', other_designations='olfactory receptor OR11-47|olfactory receptor, family 52, subfamily B, member 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5580877, end_position_on_the_genomic_accession=5581884, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873205, CurrentID=0, Status='live', Symbol='MTCO1P3', Aliases='', description='MT-CO1 pseudogene 3', other_designations='MTCO1 pseudogene 3', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45376191, end_position_on_the_genomic_accession=45376382, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873206, CurrentID=0, Status='live', Symbol='MTND5P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 1', other_designations='MT-ND5 pseudogene 1|MTND5 pseudogene 1', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44473738, end_position_on_the_genomic_accession=44475073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873207, CurrentID=0, Status='live', Symbol='MTCO3P2', Aliases='', description='MT-CO3 pseudogene 2', other_designations='MTCO3 pseudogene 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23765934, end_position_on_the_genomic_accession=23766104, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873208, CurrentID=0, Status='live', Symbol='MTND3P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 1', other_designations='MT-ND3 pseudogene 1|MTND3 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22113119, end_position_on_the_genomic_accession=22113338, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873209, CurrentID=0, Status='live', Symbol='MTND5P2', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 2', other_designations='MT-ND5 pseudogene 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95692538, end_position_on_the_genomic_accession=95693972, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873210, CurrentID=0, Status='live', Symbol='MTND5P3', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 3', other_designations='MT-ND5 pseudogene 3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84521806, end_position_on_the_genomic_accession=84523906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873211, CurrentID=0, Status='live', Symbol='MTND4P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 1', other_designations='MT-ND4 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84520230, end_position_on_the_genomic_accession=84521519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873212, CurrentID=0, Status='live', Symbol='MTND1P11', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 11', other_designations='MT-ND1 pseudogene 11', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5094100, end_position_on_the_genomic_accession=5095016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873213, CurrentID=0, Status='live', Symbol='MTND1P16', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 16', other_designations='MT-ND1 pseudogene 16', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106901068, end_position_on_the_genomic_accession=106901962, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  340990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340990, CurrentID=0, Status='live', Symbol='OTOG', Aliases='DFNB18B, MLEMP, OTGN', description='otogelin', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17547373, end_position_on_the_genomic_accession=17645944, orientation='plus', exon_count=56, OMIM=604487, no_hdr0=''),
  100873215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873215, CurrentID=0, Status='live', Symbol='MTND1P28', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 28', other_designations='MT-ND1 pseudogene 28', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117023918, end_position_on_the_genomic_accession=117024864, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873216, CurrentID=0, Status='live', Symbol='MTND1P29', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 29', other_designations='MT-ND1 pseudogene 29', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130271786, end_position_on_the_genomic_accession=130272269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873217, CurrentID=0, Status='live', Symbol='MTND1P31', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 31', other_designations='MT-ND1 pseudogene 31', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=62842512, end_position_on_the_genomic_accession=62843505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873218, CurrentID=0, Status='live', Symbol='MTND1P33', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 33', other_designations='MT-ND1 pseudogene 33', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143433042, end_position_on_the_genomic_accession=143433955, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873219, CurrentID=0, Status='live', Symbol='MTND1P34', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 34', other_designations='MT-ND1 pseudogene 34', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47164492, end_position_on_the_genomic_accession=47165872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873220, CurrentID=0, Status='live', Symbol='MTND1P19', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 19', other_designations='MT-ND1 pseudogene 19', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92702345, end_position_on_the_genomic_accession=92702635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873221, CurrentID=0, Status='live', Symbol='MTND1P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 9', other_designations='MT-ND1 pseudogene 9', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57357429, end_position_on_the_genomic_accession=57358369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873222, CurrentID=0, Status='live', Symbol='MTND2P11', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 11', other_designations='MT-ND2 pseudogene 11', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109300216, end_position_on_the_genomic_accession=109301144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873223, CurrentID=0, Status='live', Symbol='MTND2P14', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 14', other_designations='MT-ND2 pseudogene 14', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106902173, end_position_on_the_genomic_accession=106902523, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873224, CurrentID=0, Status='live', Symbol='MTND2P18', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 18', other_designations='MT-ND2 pseudogene 18', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131384461, end_position_on_the_genomic_accession=131385490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873225, CurrentID=0, Status='live', Symbol='MTND2P19', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 19', other_designations='MT-ND2 pseudogene 19', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140219644, end_position_on_the_genomic_accession=140220368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873226, CurrentID=0, Status='live', Symbol='MTND2P20', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 20', other_designations='MT-ND2 pseudogene 20', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155264163, end_position_on_the_genomic_accession=155264823, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873227, CurrentID=0, Status='live', Symbol='MTND2P21', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 21', other_designations='MT-ND2 pseudogene 21', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117025074, end_position_on_the_genomic_accession=117026112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873228, CurrentID=0, Status='live', Symbol='MTND2P22', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 22', other_designations='MT-ND2 pseudogene 22', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130272479, end_position_on_the_genomic_accession=130273508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873229, CurrentID=0, Status='live', Symbol='MTND2P23', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 23', other_designations='MT-ND2 pseudogene 23', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=211777206, end_position_on_the_genomic_accession=211777856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873230, CurrentID=0, Status='live', Symbol='MTND2P24', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 24', other_designations='MT-ND2 pseudogene 24', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55179132, end_position_on_the_genomic_accession=55180163, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873231, CurrentID=0, Status='live', Symbol='MTND2P25', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 25', other_designations='MT-ND2 pseudogene 25', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=62843726, end_position_on_the_genomic_accession=62844270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873232, CurrentID=0, Status='live', Symbol='MTND2P27', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 27', other_designations='MT-ND2 pseudogene 27', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237941458, end_position_on_the_genomic_accession=237942492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873233, CurrentID=0, Status='live', Symbol='MTND2P29', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 29', other_designations='MT-ND2 pseudogene 29', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50017086, end_position_on_the_genomic_accession=50017504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873234, CurrentID=0, Status='live', Symbol='MTND2P30', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 30', other_designations='MT-ND2 pseudogene 30', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81080778, end_position_on_the_genomic_accession=81081176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873235, CurrentID=0, Status='live', Symbol='MTND2P31', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 31', other_designations='MT-ND2 pseudogene 31', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14505864, end_position_on_the_genomic_accession=14506479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873236, CurrentID=0, Status='live', Symbol='MTND2P33', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 33', other_designations='MT-ND2 pseudogene 33', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155461639, end_position_on_the_genomic_accession=155462666, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873237, CurrentID=0, Status='live', Symbol='MTND2P32', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 32', other_designations='MT-ND2 pseudogene 32', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33015269, end_position_on_the_genomic_accession=33017043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873238, CurrentID=0, Status='live', Symbol='MTND2P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 10', other_designations='MT-ND2 pseudogene 10', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103087110, end_position_on_the_genomic_accession=103088033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873239, CurrentID=0, Status='live', Symbol='MTND3P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 3 pseudogene 10', other_designations='MT-ND3 pseudogene 10', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120212947, end_position_on_the_genomic_accession=120213275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873240, CurrentID=0, Status='live', Symbol='MTND4P15', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 15', other_designations='MT-ND4 pseudogene 15', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92110865, end_position_on_the_genomic_accession=92111832, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873241, CurrentID=0, Status='live', Symbol='MTND4P16', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 16', other_designations='MT-ND4 pseudogene 16', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106894351, end_position_on_the_genomic_accession=106895718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873242, CurrentID=0, Status='live', Symbol='MTND4P17', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 17', other_designations='MT-ND4 pseudogene 17', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166160014, end_position_on_the_genomic_accession=166160175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341019, CurrentID=0, Status='live', Symbol='DCDC1', Aliases='', description='doublecortin domain containing 1', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31262624, end_position_on_the_genomic_accession=31369810, orientation='minus', exon_count=9, OMIM=608062, no_hdr0=''),
  359739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359739, CurrentID=0, Status='live', Symbol='MRPL3P1', Aliases='', description='mitochondrial ribosomal protein L3 pseudogene 1', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19304065, end_position_on_the_genomic_accession=19305760, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873245, CurrentID=0, Status='live', Symbol='MTND4P23', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 23', other_designations='MT-ND4 pseudogene 23', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201213293, end_position_on_the_genomic_accession=201214660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873246, CurrentID=0, Status='live', Symbol='MTND4P25', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 25', other_designations='MT-ND4 pseudogene 25', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82814966, end_position_on_the_genomic_accession=82815519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873247, CurrentID=0, Status='live', Symbol='MTND4P26', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 26', other_designations='MT-ND4 pseudogene 26', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120213631, end_position_on_the_genomic_accession=120214989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873248, CurrentID=0, Status='live', Symbol='MTND4P27', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 27', other_designations='MT-ND4 pseudogene 27', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130278736, end_position_on_the_genomic_accession=130280095, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873249, CurrentID=0, Status='live', Symbol='MTND4P28', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 28', other_designations='MT-ND4 pseudogene 28', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155311077, end_position_on_the_genomic_accession=155311418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873250, CurrentID=0, Status='live', Symbol='MTND4P29', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 29', other_designations='MT-ND4 pseudogene 29', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201554384, end_position_on_the_genomic_accession=201556033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873251, CurrentID=0, Status='live', Symbol='MTND4P30', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 30', other_designations='MT-ND4 pseudogene 30', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202618597, end_position_on_the_genomic_accession=202618851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873252, CurrentID=0, Status='live', Symbol='MTND4P31', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 31', other_designations='MT-ND4 pseudogene 31', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70127619, end_position_on_the_genomic_accession=70129096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873253, CurrentID=0, Status='live', Symbol='MTND4P32', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 32', other_designations='MT-ND4 pseudogene 32', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102799966, end_position_on_the_genomic_accession=102801091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873254, CurrentID=0, Status='live', Symbol='MTND4P24', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 24', other_designations='MT-ND4 pseudogene 24', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126472885, end_position_on_the_genomic_accession=126473283, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873255, CurrentID=0, Status='live', Symbol='MTND4P11', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 11', other_designations='MT-ND4 pseudogene 11', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93922575, end_position_on_the_genomic_accession=93923557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341032, CurrentID=0, Status='live', Symbol='C11orf53', Aliases='', description='chromosome 11 open reading frame 53', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111245704, end_position_on_the_genomic_accession=111286406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873257, CurrentID=0, Status='live', Symbol='MTND5P12', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 12', other_designations='MT-ND5 pseudogene 12', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94567458, end_position_on_the_genomic_accession=94568937, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873258, CurrentID=0, Status='live', Symbol='MTND5P14', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 14', other_designations='MT-ND5 pseudogene 14', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5109523, end_position_on_the_genomic_accession=5110754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873259, CurrentID=0, Status='live', Symbol='MTND5P15', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 15', other_designations='MT-ND5 pseudogene 15', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177010419, end_position_on_the_genomic_accession=177011046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873260, CurrentID=0, Status='live', Symbol='MTND5P22', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 22', other_designations='MT-ND5 pseudogene 22', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=124680713, end_position_on_the_genomic_accession=124680889, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873261, CurrentID=0, Status='live', Symbol='MTND5P23', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 23', other_designations='MT-ND5 pseudogene 23', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131370314, end_position_on_the_genomic_accession=131372084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873262, CurrentID=0, Status='live', Symbol='MTND5P24', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 24', other_designations='MT-ND5 pseudogene 24', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143094765, end_position_on_the_genomic_accession=143096552, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873263, CurrentID=0, Status='live', Symbol='MTND5P25', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 25', other_designations='MT-ND5 pseudogene 25', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201212296, end_position_on_the_genomic_accession=201213069, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873264, CurrentID=0, Status='live', Symbol='MTND5P27', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 27', other_designations='MT-ND5 pseudogene 27', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82815751, end_position_on_the_genomic_accession=82817528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873265, CurrentID=0, Status='live', Symbol='MTND5P29', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 29', other_designations='MT-ND5 pseudogene 29', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130280305, end_position_on_the_genomic_accession=130282079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873266, CurrentID=0, Status='live', Symbol='MTND5P30', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 30', other_designations='MT-ND5 pseudogene 30', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155311654, end_position_on_the_genomic_accession=155313427, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873267, CurrentID=0, Status='live', Symbol='MTND5P31', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 31', other_designations='MT-ND5 pseudogene 31', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201556956, end_position_on_the_genomic_accession=201558269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873268, CurrentID=0, Status='live', Symbol='MTND5P26', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 26', other_designations='MT-ND5 pseudogene 26', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102804632, end_position_on_the_genomic_accession=102806120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873269, CurrentID=0, Status='live', Symbol='MTND5P20', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 20', other_designations='MT-ND5 pseudogene 20', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113576751, end_position_on_the_genomic_accession=113576894, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873270, CurrentID=0, Status='live', Symbol='MTND5P19', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 19', other_designations='MT-ND5 pseudogene 19', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235538051, end_position_on_the_genomic_accession=235539812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873271, CurrentID=0, Status='live', Symbol='MTND5P18', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 18', other_designations='MT-ND5 pseudogene 18', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237949742, end_position_on_the_genomic_accession=237951503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873272, CurrentID=0, Status='live', Symbol='MTND5P13', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 13', other_designations='MT-ND5 pseudogene 13', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64609434, end_position_on_the_genomic_accession=64611081, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873273, CurrentID=0, Status='live', Symbol='RNA5SP18', Aliases='RN5S18', description='RNA, 5S ribosomal pseudogene 18', other_designations='RNA, 5S ribosomal 18', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228647912, end_position_on_the_genomic_accession=228648032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873274, CurrentID=0, Status='live', Symbol='RNA5SP19', Aliases='RN5S19', description='RNA, 5S ribosomal pseudogene 19', other_designations='RNA, 5S ribosomal 19', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228555793, end_position_on_the_genomic_accession=228555901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873275, CurrentID=0, Status='live', Symbol='RNA5SP20', Aliases='RN5S20', description='RNA, 5S ribosomal pseudogene 20', other_designations='RNA, 5S ribosomal 20', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77614869, end_position_on_the_genomic_accession=77614952, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873276, CurrentID=0, Status='live', Symbol='RNA5SP21', Aliases='RN5S21', description='RNA, 5S ribosomal pseudogene 21', other_designations='RNA, 5S ribosomal 21', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77779904, end_position_on_the_genomic_accession=77779999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873277, CurrentID=0, Status='live', Symbol='RNA5SP22', Aliases='RN5S22', description='RNA, 5S ribosomal pseudogene 22', other_designations='RNA, 5S ribosomal 22', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78094807, end_position_on_the_genomic_accession=78094915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873278, CurrentID=0, Status='live', Symbol='RNA5SP23', Aliases='RN5S23', description='RNA, 5S ribosomal pseudogene 23', other_designations='RNA, 5S ribosomal 23', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78375164, end_position_on_the_genomic_accession=78375268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873279, CurrentID=0, Status='live', Symbol='RNA5SP40', Aliases='RN5S40', description='RNA, 5S ribosomal pseudogene 40', other_designations='RNA, 5S ribosomal 40', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9437669, end_position_on_the_genomic_accession=9437778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873280, CurrentID=0, Status='live', Symbol='RNA5SP41', Aliases='RN5S41', description='RNA, 5S ribosomal pseudogene 41', other_designations='RNA, 5S ribosomal 41', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13623184, end_position_on_the_genomic_accession=13623284, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873281, CurrentID=0, Status='live', Symbol='RNA5SP42', Aliases='RN5S42', description='RNA, 5S ribosomal pseudogene 42', other_designations='RNA, 5S ribosomal 42', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34112949, end_position_on_the_genomic_accession=34113063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873282, CurrentID=0, Status='live', Symbol='RNA5SP43', Aliases='RN5S43', description='RNA, 5S ribosomal pseudogene 43', other_designations='RNA, 5S ribosomal 43', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37264677, end_position_on_the_genomic_accession=37264786, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873283, CurrentID=0, Status='live', Symbol='RNA5SP44', Aliases='RN5S44', description='RNA, 5S ribosomal pseudogene 44', other_designations='RNA, 5S ribosomal 44', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39154164, end_position_on_the_genomic_accession=39154296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873284, CurrentID=0, Status='live', Symbol='RNA5SP45', Aliases='RN5S45', description='RNA, 5S ribosomal pseudogene 45', other_designations='RNA, 5S ribosomal 45', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41466937, end_position_on_the_genomic_accession=41467028, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873285, CurrentID=0, Status='live', Symbol='RNA5SP46', Aliases='RN5S46', description='RNA, 5S ribosomal pseudogene 46', other_designations='RNA, 5S ribosomal 46', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43196417, end_position_on_the_genomic_accession=43196536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873286, CurrentID=0, Status='live', Symbol='RNA5SP47', Aliases='RN5S47', description='RNA, 5S ribosomal pseudogene 47', other_designations='RNA, 5S ribosomal 47', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44932323, end_position_on_the_genomic_accession=44932431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873287, CurrentID=0, Status='live', Symbol='RNA5SP48', Aliases='RN5S48', description='RNA, 5S ribosomal pseudogene 48', other_designations='RNA, 5S ribosomal 48', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51973410, end_position_on_the_genomic_accession=51973535, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873288, CurrentID=0, Status='live', Symbol='RNA5SP49', Aliases='RN5S49', description='RNA, 5S ribosomal pseudogene 49', other_designations='RNA, 5S ribosomal 49', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63186336, end_position_on_the_genomic_accession=63186450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873289, CurrentID=0, Status='live', Symbol='RNA5SP51', Aliases='RN5S51', description='RNA, 5S ribosomal pseudogene 51', other_designations='RNA, 5S ribosomal 51', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85883680, end_position_on_the_genomic_accession=85883798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873290, CurrentID=0, Status='live', Symbol='RNA5SP52', Aliases='RN5S52', description='RNA, 5S ribosomal pseudogene 52', other_designations='RNA, 5S ribosomal 52', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87453240, end_position_on_the_genomic_accession=87453373, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873291, CurrentID=0, Status='live', Symbol='RNA5SP53', Aliases='RN5S53', description='RNA, 5S ribosomal pseudogene 53', other_designations='RNA, 5S ribosomal 53', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93488333, end_position_on_the_genomic_accession=93488452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873292, CurrentID=0, Status='live', Symbol='RNA5SP54', Aliases='RN5S54', description='RNA, 5S ribosomal pseudogene 54', other_designations='RNA, 5S ribosomal 54', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111041834, end_position_on_the_genomic_accession=111041925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873293, CurrentID=0, Status='live', Symbol='RNA5SP55', Aliases='RN5S55', description='RNA, 5S ribosomal pseudogene 55', other_designations='RNA, 5S ribosomal 55', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116962347, end_position_on_the_genomic_accession=116962471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873294, CurrentID=0, Status='live', Symbol='RNA5SP56', Aliases='RN5S56', description='RNA, 5S ribosomal pseudogene 56', other_designations='RNA, 5S ribosomal 56', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118264372, end_position_on_the_genomic_accession=118264489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479465, CurrentID=0, Status='live', Symbol='RN7SL667P', Aliases='', description='RNA, 7SL, cytoplasmic 667, pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873296, CurrentID=0, Status='live', Symbol='RNA5SP61', Aliases='RN5S61', description='RNA, 5S ribosomal pseudogene 61', other_designations='RNA, 5S ribosomal 61', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162338643, end_position_on_the_genomic_accession=162338742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873297, CurrentID=0, Status='live', Symbol='RNA5SP62', Aliases='RN5S62', description='RNA, 5S ribosomal pseudogene 62', other_designations='RNA, 5S ribosomal 62', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163468496, end_position_on_the_genomic_accession=163468605, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873298, CurrentID=0, Status='live', Symbol='RNA5SP63', Aliases='RN5S63', description='RNA, 5S ribosomal pseudogene 63', other_designations='RNA, 5S ribosomal 63', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163509484, end_position_on_the_genomic_accession=163509595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873299, CurrentID=0, Status='live', Symbol='RNA5SP64', Aliases='RN5S64', description='RNA, 5S ribosomal pseudogene 64', other_designations='RNA, 5S ribosomal 64', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166042244, end_position_on_the_genomic_accession=166042350, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873300, CurrentID=0, Status='live', Symbol='RNA5SP65', Aliases='RN5S65', description='RNA, 5S ribosomal pseudogene 65', other_designations='RNA, 5S ribosomal 65', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167005959, end_position_on_the_genomic_accession=167006077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873301, CurrentID=0, Status='live', Symbol='RNA5SP66', Aliases='RN5S66', description='RNA, 5S ribosomal pseudogene 66', other_designations='RNA, 5S ribosomal 66', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169067264, end_position_on_the_genomic_accession=169067364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873302, CurrentID=0, Status='live', Symbol='RNA5SP67', Aliases='RN5S67', description='RNA, 5S ribosomal pseudogene 67', other_designations='RNA, 5S ribosomal 67', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173921070, end_position_on_the_genomic_accession=173921200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873303, CurrentID=0, Status='live', Symbol='RNA5SP68', Aliases='RN5S68', description='RNA, 5S ribosomal pseudogene 68', other_designations='RNA, 5S ribosomal 68', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173969318, end_position_on_the_genomic_accession=173969436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873304, CurrentID=0, Status='live', Symbol='RNA5SP69', Aliases='RN5S69', description='RNA, 5S ribosomal pseudogene 69', other_designations='RNA, 5S ribosomal 69', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178560913, end_position_on_the_genomic_accession=178561029, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873305, CurrentID=0, Status='live', Symbol='RNA5SP71', Aliases='RN5S71', description='RNA, 5S ribosomal pseudogene 71', other_designations='RNA, 5S ribosomal 71', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182944365, end_position_on_the_genomic_accession=182944490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873306, CurrentID=0, Status='live', Symbol='RNA5SP72', Aliases='RN5S72', description='RNA, 5S ribosomal pseudogene 72', other_designations='RNA, 5S ribosomal 72', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185014951, end_position_on_the_genomic_accession=185015067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873307, CurrentID=0, Status='live', Symbol='RNA5SP73', Aliases='RN5S73', description='RNA, 5S ribosomal pseudogene 73', other_designations='RNA, 5S ribosomal 73', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=189666149, end_position_on_the_genomic_accession=189666256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873308, CurrentID=0, Status='live', Symbol='RNA5SP74', Aliases='RN5S74', description='RNA, 5S ribosomal pseudogene 74', other_designations='RNA, 5S ribosomal 74', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204562413, end_position_on_the_genomic_accession=204562521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873309, CurrentID=0, Status='live', Symbol='RNA5SP75', Aliases='RN5S75', description='RNA, 5S ribosomal pseudogene 75', other_designations='RNA, 5S ribosomal 75', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204707320, end_position_on_the_genomic_accession=204707430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873310, CurrentID=0, Status='live', Symbol='RNA5SP76', Aliases='RN5S76', description='RNA, 5S ribosomal pseudogene 76', other_designations='RNA, 5S ribosomal 76', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219761789, end_position_on_the_genomic_accession=219761875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873311, CurrentID=0, Status='live', Symbol='RNA5SP77', Aliases='RN5S77', description='RNA, 5S ribosomal pseudogene 77', other_designations='RNA, 5S ribosomal 77', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227561181, end_position_on_the_genomic_accession=227561300, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873312, CurrentID=0, Status='live', Symbol='RNA5SP78', Aliases='RN5S78', description='RNA, 5S ribosomal pseudogene 78', other_designations='RNA, 5S ribosomal 78', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229549905, end_position_on_the_genomic_accession=229550022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873313, CurrentID=0, Status='live', Symbol='RNA5SP79', Aliases='RN5S79', description='RNA, 5S ribosomal pseudogene 79', other_designations='RNA, 5S ribosomal 79', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230820250, end_position_on_the_genomic_accession=230820358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873314, CurrentID=0, Status='live', Symbol='RNA5SP80', Aliases='RN5S80', description='RNA, 5S ribosomal pseudogene 80', other_designations='RNA, 5S ribosomal 80', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231281414, end_position_on_the_genomic_accession=231281524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873315, CurrentID=0, Status='live', Symbol='RNA5SP81', Aliases='RN5S81', description='RNA, 5S ribosomal pseudogene 81', other_designations='RNA, 5S ribosomal 81', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242134272, end_position_on_the_genomic_accession=242134381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873316, CurrentID=0, Status='live', Symbol='RNA5SP82', Aliases='RN5S82', description='RNA, 5S ribosomal pseudogene 82', other_designations='RNA, 5S ribosomal 82', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247204570, end_position_on_the_genomic_accession=247204661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873317, CurrentID=0, Status='live', Symbol='RNA5SP84', Aliases='RN5S84', description='RNA, 5S ribosomal pseudogene 84', other_designations='RNA, 5S ribosomal 84', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11517397, end_position_on_the_genomic_accession=11517503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873318, CurrentID=0, Status='live', Symbol='RNA5SP85', Aliases='RN5S85', description='RNA, 5S ribosomal pseudogene 85', other_designations='RNA, 5S ribosomal 85', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11561661, end_position_on_the_genomic_accession=11561779, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873319, CurrentID=0, Status='live', Symbol='RNA5SP86', Aliases='RN5S86', description='RNA, 5S ribosomal pseudogene 86', other_designations='RNA, 5S ribosomal 86', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20401627, end_position_on_the_genomic_accession=20401724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873320, CurrentID=0, Status='live', Symbol='RNA5SP87', Aliases='RN5S87', description='RNA, 5S ribosomal pseudogene 87', other_designations='RNA, 5S ribosomal 87', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22338886, end_position_on_the_genomic_accession=22339003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873321, CurrentID=0, Status='live', Symbol='RNA5SP88', Aliases='RN5S88', description='RNA, 5S ribosomal pseudogene 88', other_designations='RNA, 5S ribosomal 88', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24564630, end_position_on_the_genomic_accession=24564738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873322, CurrentID=0, Status='live', Symbol='RNA5SP89', Aliases='RN5S89', description='RNA, 5S ribosomal pseudogene 89', other_designations='RNA, 5S ribosomal 89', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28683976, end_position_on_the_genomic_accession=28684083, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873323, CurrentID=0, Status='live', Symbol='RNA5SP90', Aliases='RN5S90', description='RNA, 5S ribosomal pseudogene 90', other_designations='RNA, 5S ribosomal 90', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31228312, end_position_on_the_genomic_accession=31228410, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873324, CurrentID=0, Status='live', Symbol='RNA5SP91', Aliases='RN5S91', description='RNA, 5S ribosomal pseudogene 91', other_designations='RNA, 5S ribosomal 91', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33285769, end_position_on_the_genomic_accession=33285887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873325, CurrentID=0, Status='live', Symbol='RNA5SP92', Aliases='RN5S92', description='RNA, 5S ribosomal pseudogene 92', other_designations='RNA, 5S ribosomal 92', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33332898, end_position_on_the_genomic_accession=33333007, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873326, CurrentID=0, Status='live', Symbol='RNA5SP93', Aliases='RN5S93', description='RNA, 5S ribosomal pseudogene 93', other_designations='RNA, 5S ribosomal 93', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56235217, end_position_on_the_genomic_accession=56235330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873327, CurrentID=0, Status='live', Symbol='RNA5SP94', Aliases='RN5S94', description='RNA, 5S ribosomal pseudogene 94', other_designations='RNA, 5S ribosomal 94', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=59694762, end_position_on_the_genomic_accession=59694869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873328, CurrentID=0, Status='live', Symbol='RNA5SP95', Aliases='RN5S95', description='RNA, 5S ribosomal pseudogene 95', other_designations='RNA, 5S ribosomal 95', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60998752, end_position_on_the_genomic_accession=60998888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873329, CurrentID=0, Status='live', Symbol='RNA5SP96', Aliases='RN5S96', description='RNA, 5S ribosomal pseudogene 96', other_designations='RNA, 5S ribosomal 96', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69181897, end_position_on_the_genomic_accession=69182003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873330, CurrentID=0, Status='live', Symbol='RNA5SP97', Aliases='RN5S97', description='RNA, 5S ribosomal pseudogene 97', other_designations='RNA, 5S ribosomal 97', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73968123, end_position_on_the_genomic_accession=73968206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873331, CurrentID=0, Status='live', Symbol='RNA5SP98', Aliases='RN5S98', description='RNA, 5S ribosomal pseudogene 98', other_designations='RNA, 5S ribosomal 98', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76772909, end_position_on_the_genomic_accession=76773045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873332, CurrentID=0, Status='live', Symbol='MTND6P21', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 21', other_designations='MT-ND6 pseudogene 21|MTND6 pseudogene 21', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44472895, end_position_on_the_genomic_accession=44473739, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873333, CurrentID=0, Status='live', Symbol='SEPT2P1', Aliases='', description='septin 2 pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105698030, end_position_on_the_genomic_accession=105698227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873334, CurrentID=0, Status='live', Symbol='RNA5-8SP3', Aliases='RN5-8S3', description='RNA, 5.8S ribosomal pseudogene 3', other_designations='RNA, 5.8S ribosomal 3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108992409, end_position_on_the_genomic_accession=108992569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873335, CurrentID=0, Status='live', Symbol='RNA5-8SP5', Aliases='RN5-8S5', description='RNA, 5.8S ribosomal pseudogene 5', other_designations='RNA, 5.8S ribosomal 5', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132253154, end_position_on_the_genomic_accession=132253305, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873336, CurrentID=0, Status='live', Symbol='RNA5-8SP6', Aliases='RN5-8S6', description='RNA, 5.8S ribosomal pseudogene 6', other_designations='RNA, 5.8S ribosomal 6', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10200155, end_position_on_the_genomic_accession=10200306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873337, CurrentID=0, Status='live', Symbol='MTND1P14', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 14', other_designations='MT-ND1 pseudogene 14', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19602259, end_position_on_the_genomic_accession=19603205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873338, CurrentID=0, Status='live', Symbol='MTND1P18', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 18', other_designations='MT-ND1 pseudogene 18', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37600796, end_position_on_the_genomic_accession=37601368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873339, CurrentID=0, Status='live', Symbol='MTND1P20', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 20', other_designations='MT-ND1 pseudogene 20', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69595487, end_position_on_the_genomic_accession=69596436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341116, CurrentID=0, Status='live', Symbol='MS4A10', Aliases='CD20L7, MS4A9', description='membrane-spanning 4-domains, subfamily A, member 10', other_designations='CD20 antigen-like 7', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60785304, end_position_on_the_genomic_accession=60801305, orientation='plus', exon_count=10, OMIM=608403, no_hdr0=''),
  100873341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873341, CurrentID=0, Status='live', Symbol='MTND1P24', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 24', other_designations='MT-ND1 pseudogene 24', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41699394, end_position_on_the_genomic_accession=41700251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873342, CurrentID=0, Status='live', Symbol='MTND1P17', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 17', other_designations='MT-ND1 pseudogene 17', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16876218, end_position_on_the_genomic_accession=16876373, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873343, CurrentID=0, Status='live', Symbol='MTND1P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 10', other_designations='MT-ND1 pseudogene 10', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36178556, end_position_on_the_genomic_accession=36178758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873344, CurrentID=0, Status='live', Symbol='MTND2P12', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 12', other_designations='MT-ND2 pseudogene 12', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19603845, end_position_on_the_genomic_accession=19605044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873345, CurrentID=0, Status='live', Symbol='MTND2P13', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 13', other_designations='MT-ND2 pseudogene 13', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22525899, end_position_on_the_genomic_accession=22526924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873346, CurrentID=0, Status='live', Symbol='MTND2P15', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 15', other_designations='MT-ND2 pseudogene 15', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69594020, end_position_on_the_genomic_accession=69594932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873347, CurrentID=0, Status='live', Symbol='MTND2P16', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 16', other_designations='MT-ND2 pseudogene 16', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19747822, end_position_on_the_genomic_accession=19748789, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873348, CurrentID=0, Status='live', Symbol='MTND2P17', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 17', other_designations='MT-ND2 pseudogene 17', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41698495, end_position_on_the_genomic_accession=41699182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873349, CurrentID=0, Status='live', Symbol='MTND2P26', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 26', other_designations='MT-ND2 pseudogene 26', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103406219, end_position_on_the_genomic_accession=103407241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873350, CurrentID=0, Status='live', Symbol='MTND4P18', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 18', other_designations='MT-ND4 pseudogene 18', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36434691, end_position_on_the_genomic_accession=36435174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873351, CurrentID=0, Status='live', Symbol='MTND4P19', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 19', other_designations='MT-ND4 pseudogene 19', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94763426, end_position_on_the_genomic_accession=94774628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341128, CurrentID=0, Status='live', Symbol='OR7E1P', Aliases='', description='olfactory receptor family 7 subfamily E member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 1 pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67974234, end_position_on_the_genomic_accession=67975256, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873353, CurrentID=0, Status='live', Symbol='MTND5P17', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 17', other_designations='MT-ND5 pseudogene 17', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36432863, end_position_on_the_genomic_accession=36434478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  78986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78986, CurrentID=0, Status='live', Symbol='DUSP26', Aliases='DSP-4, DUSP24, LDP-4, MKP-8, MKP8, NATA1, NEAP, SKRP3', description='dual specificity phosphatase 26 (putative)', other_designations='MAP kinase phosphatase 8|Novel amplified gene in thyroid anaplastic cancer|dual specificity phosphatase SKRP3|low-molecular-mass dual-specificity phosphatase 4|mitogen-activated protein kinase phosphatase 8|neuroendocrine-associated phosphatase', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33591330, end_position_on_the_genomic_accession=33600028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  78987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78987, CurrentID=0, Status='live', Symbol='CRELD1', Aliases='AVSD2, CIRRIN', description='cysteine rich with EGF-like domains 1', other_designations='cysteine-rich with EGF-like domains 1', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9933788, end_position_on_the_genomic_accession=9945413, orientation='plus', exon_count=12, OMIM=607170, no_hdr0=''),
  78988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78988, CurrentID=0, Status='live', Symbol='MRPL57', Aliases='MRP63, bMRP63', description='mitochondrial ribosomal protein L57', other_designations='hMRP63|mitochondrial ribosomal protein 63|mitochondrial ribosomal protein bMRP63', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21176233, end_position_on_the_genomic_accession=21179084, orientation='plus', exon_count=2, OMIM=611997, no_hdr0=''),
  78989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78989, CurrentID=0, Status='live', Symbol='COLEC11', Aliases='3MC2, CL-K1-I, CL-K1-II, CL-K1-IIa, CL-K1-IIb, CLK1', description='collectin subfamily member 11', other_designations='Collectin K1|collectin kidney protein 1|collectin sub-family member 11', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3594832, end_position_on_the_genomic_accession=3644644, orientation='plus', exon_count=12, OMIM=612502, no_hdr0=''),
  78990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78990, CurrentID=0, Status='live', Symbol='OTUB2', Aliases='C14orf137, OTB2, OTU2', description='OTU deubiquitinase, ubiquitin aldehyde binding 2', other_designations='OTU domain, ubiquitin aldehyde binding 2|OTU domain-containing ubiquitin aldehyde-binding protein 2|deubiquitinating enzyme OTUB2|otubain-2|ubiquitin-specific protease otubain 2|ubiquitin-specific-processing protease OTUB2', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94016095, end_position_on_the_genomic_accession=94048930, orientation='plus', exon_count=11, OMIM=608338, no_hdr0=''),
  78991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78991, CurrentID=0, Status='live', Symbol='PCYOX1L', Aliases='', description='prenylcysteine oxidase 1 like', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149358007, end_position_on_the_genomic_accession=149369658, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78992, CurrentID=0, Status='live', Symbol='YIPF2', Aliases='FinGER2', description='Yip1 domain family member 2', other_designations='YIP1 family member 2|Yip1 domain family, member 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10922768, end_position_on_the_genomic_accession=10929012, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100873361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873361, CurrentID=0, Status='live', Symbol='RNA5SP29', Aliases='RN5S29', description='RNA, 5S ribosomal pseudogene 29', other_designations='RNA, 5S ribosomal 29', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51027747, end_position_on_the_genomic_accession=51027852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  78994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78994, CurrentID=0, Status='live', Symbol='PRR14', Aliases='', description='proline rich 14', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30650717, end_position_on_the_genomic_accession=30656413, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78995, CurrentID=0, Status='live', Symbol='C17orf53', Aliases='', description='chromosome 17 open reading frame 53', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44141906, end_position_on_the_genomic_accession=44162477, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  78996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78996, CurrentID=0, Status='live', Symbol='C7orf49', Aliases='MRI', description='chromosome 7 open reading frame 49', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135165780, end_position_on_the_genomic_accession=135170826, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  78997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78997, CurrentID=0, Status='live', Symbol='GDAP1L1', Aliases='dJ881L22.1, dJ995J12.1.1', description='ganglioside induced differentiation associated protein 1-like 1', other_designations='GDAP1-L1', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44247099, end_position_on_the_genomic_accession=44280917, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  78998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78998, CurrentID=0, Status='live', Symbol='RHPN1-AS1', Aliases='C8orf51', description='RHPN1 antisense RNA 1 (head to head)', other_designations='RHPN1 antisense RNA 1 (non-protein coding)', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143366623, end_position_on_the_genomic_accession=143368635, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  78999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=78999, CurrentID=0, Status='live', Symbol='LRFN4', Aliases='FIGLER6, SALM3, SALM3.', description='leucine rich repeat and fibronectin type III domain containing 4', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 6', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66856480, end_position_on_the_genomic_accession=66860475, orientation='plus', exon_count=3, OMIM=612810, no_hdr0=''),
  79000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79000, CurrentID=0, Status='live', Symbol='AUNIP', Aliases='AIBP, C1orf135', description='aurora kinase A and ninein interacting protein', other_designations='aurora A-binding protein', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25831913, end_position_on_the_genomic_accession=25859458, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79001, CurrentID=0, Status='live', Symbol='VKORC1', Aliases='EDTP308, MST134, MST576, VKCFD2, VKOR', description='vitamin K epoxide reductase complex subunit 1', other_designations='phylloquinone epoxide reductase|vitamin K dependent clotting factors deficiency 2|vitamin K epoxide reductase complex, subunit 1|vitamin K1 2,3-epoxide reductase subunit 1|vitamin K1 epoxide reductase (warfarin-sensitive)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31090854, end_position_on_the_genomic_accession=31096337, orientation='minus', exon_count=5, OMIM=608547, no_hdr0=''),
  79002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79002, CurrentID=0, Status='live', Symbol='C19orf43', Aliases='fSAP18', description='chromosome 19 open reading frame 43', other_designations='My029 protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12730640, end_position_on_the_genomic_accession=12734715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79003, CurrentID=0, Status='live', Symbol='MIS12', Aliases='2510025F08Rik, KNTC2AP, MTW1, hMis12', description='MIS12 kinetochore complex component', other_designations='MIS12 homolog|MIS12, MIND kinetochore complex component, homolog|homolog of yeast Mis12', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5486374, end_position_on_the_genomic_accession=5490814, orientation='plus', exon_count=4, OMIM=609178, no_hdr0=''),
  79004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79004, CurrentID=0, Status='live', Symbol='CUEDC2', Aliases='C10orf66, bA18I14.5', description='CUE domain containing 2', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102423245, end_position_on_the_genomic_accession=102432666, orientation='minus', exon_count=9, OMIM=614142, no_hdr0=''),
  79005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79005, CurrentID=0, Status='live', Symbol='SCNM1', Aliases='', description='sodium channel modifier 1', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151166022, end_position_on_the_genomic_accession=151170297, orientation='plus', exon_count=7, OMIM=608095, no_hdr0=''),
  79006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79006, CurrentID=0, Status='live', Symbol='METRN', Aliases='C16orf23, c380A1.2', description='meteorin, glial cell differentiation regulator', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=715173, end_position_on_the_genomic_accession=717480, orientation='plus', exon_count=4, OMIM=610998, no_hdr0=''),
  79007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79007, CurrentID=0, Status='live', Symbol='DBNDD1', Aliases='', description='dysbindin (dystrobrevin binding protein 1) domain containing 1', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90004865, end_position_on_the_genomic_accession=90020131, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79008, CurrentID=0, Status='live', Symbol='SLX1B', Aliases='GIYD2', description='SLX1 homolog B, structure-specific endonuclease subunit', other_designations='GIY-YIG domain-containing protein 1|SLX1 structure-specific endonuclease subunit homolog B|SLX1A', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29454501, end_position_on_the_genomic_accession=29458224, orientation='plus', exon_count=6, OMIM=615823, no_hdr0=''),
  79009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79009, CurrentID=0, Status='live', Symbol='DDX50', Aliases='GU2, GUB, RH-II/GuB, mcdrh', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 50', other_designations='DEAD box protein 50|RNA helicase II/Gu beta|gu-beta|malignant cell derived RNA helicase|nucleolar protein GU2', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68898972, end_position_on_the_genomic_accession=68946847, orientation='plus', exon_count=16, OMIM=610373, no_hdr0=''),
  100873378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873378, CurrentID=0, Status='live', Symbol='RNA5SP108', Aliases='RN5S108', description='RNA, 5S ribosomal pseudogene 108', other_designations='RNA, 5S ribosomal 108', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161409554, end_position_on_the_genomic_accession=161409670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873379, CurrentID=0, Status='live', Symbol='RNA5SP109', Aliases='RN5S109', description='RNA, 5S ribosomal pseudogene 109', other_designations='RNA, 5S ribosomal 109', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162496936, end_position_on_the_genomic_accession=162497051, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79012, CurrentID=0, Status='live', Symbol='CAMKV', Aliases='1G5, VACAMKL', description='CaM kinase-like vesicle-associated', other_designations='vesicle-associated calmodulin-binding protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49857981, end_position_on_the_genomic_accession=49869955, orientation='minus', exon_count=11, OMIM=614993, no_hdr0=''),
  100873381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873381, CurrentID=0, Status='live', Symbol='RNA5SP111', Aliases='RN5S111', description='RNA, 5S ribosomal pseudogene 111', other_designations='RNA, 5S ribosomal 111', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164895677, end_position_on_the_genomic_accession=164895777, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873382, CurrentID=0, Status='live', Symbol='RNA5SP112', Aliases='RN5S112', description='RNA, 5S ribosomal pseudogene 112', other_designations='RNA, 5S ribosomal 112', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177138699, end_position_on_the_genomic_accession=177138797, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79015, CurrentID=0, Status='live', Symbol='LINC01260', Aliases='', description='long intergenic non-protein coding RNA 1260', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44656451, end_position_on_the_genomic_accession=44671739, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79016, CurrentID=0, Status='live', Symbol='DDA1', Aliases='C19orf58, PCIA1', description='DET1 and DDB1 associated 1', other_designations='PCIA-1|cross-immune reaction antigen PCIA1|placenta cross-immune reaction antigen 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17309528, end_position_on_the_genomic_accession=17323298, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  79017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79017, CurrentID=0, Status='live', Symbol='GGCT', Aliases='C7orf24, CRF21, GCTG, GGC', description='gamma-glutamylcyclotransferase', other_designations='cytochrome c-releasing factor 21|gamma -glutamyl cyclotransferase', map_location='7p15-p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30496621, end_position_on_the_genomic_accession=30504841, orientation='minus', exon_count=5, OMIM=137170, no_hdr0=''),
  79018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79018, CurrentID=0, Status='live', Symbol='GID4', Aliases='C17orf39, VID24', description='GID complex subunit 4 homolog', other_designations='GID complex subunit 4, VID24 homolog|vacuolar import and degradation 24|vacuolar import and degradation protein 24 homolog', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18039297, end_position_on_the_genomic_accession=18068405, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79019, CurrentID=0, Status='live', Symbol='CENPM', Aliases='C22orf18, CENP-M, PANE1', description='centromere protein M', other_designations='interphase centromere complex protein 39|proliferation-associated nuclear element protein 1', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41927763, end_position_on_the_genomic_accession=41947164, orientation='minus', exon_count=9, OMIM=610152, no_hdr0=''),
  79020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79020, CurrentID=0, Status='live', Symbol='C7orf25', Aliases='', description='chromosome 7 open reading frame 25', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42909273, end_position_on_the_genomic_accession=42912090, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873389, CurrentID=0, Status='live', Symbol='RNA5SP119', Aliases='RN5S119', description='RNA, 5S ribosomal pseudogene 119', other_designations='RNA, 5S ribosomal 119', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=211535653, end_position_on_the_genomic_accession=211535769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79022, CurrentID=0, Status='live', Symbol='TMEM106C', Aliases='', description='transmembrane protein 106C', other_designations='endoplasmic reticulum membrane protein overexpressed in cancer', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47963547, end_position_on_the_genomic_accession=47968878, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79023, CurrentID=0, Status='live', Symbol='NUP37', Aliases='p37', description='nucleoporin 37kDa', other_designations='nup107-160 subcomplex subunit Nup37', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102074189, end_position_on_the_genomic_accession=102120139, orientation='minus', exon_count=11, OMIM=609264, no_hdr0=''),
  79024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79024, CurrentID=0, Status='live', Symbol='SMIM2', Aliases='C13orf44', description='small integral membrane protein 2', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44143151, end_position_on_the_genomic_accession=44161257, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79025, CurrentID=0, Status='live', Symbol='C20orf195', Aliases='', description='chromosome 20 open reading frame 195', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63553020, end_position_on_the_genomic_accession=63556708, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79026, CurrentID=0, Status='live', Symbol='AHNAK', Aliases='AHNAKRS', description='AHNAK nucleoprotein', other_designations='AHNAK-related|desmoyokin', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62433542, end_position_on_the_genomic_accession=62546860, orientation='minus', exon_count=9, OMIM=103390, no_hdr0=''),
  79027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79027, CurrentID=0, Status='live', Symbol='ZNF655', Aliases='VIK, VIK-1', description='zinc finger protein 655', other_designations='vav-1 interacting Kruppel-like protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99558422, end_position_on_the_genomic_accession=99576453, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100873396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873396, CurrentID=0, Status='live', Symbol='RNA5SP126', Aliases='RN5S126', description='RNA, 5S ribosomal pseudogene 126', other_designations='RNA, 5S ribosomal 126', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25352284, end_position_on_the_genomic_accession=25352392, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79029, CurrentID=0, Status='live', Symbol='SPATA5L1', Aliases='', description='spermatogenesis associated 5-like 1', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45402273, end_position_on_the_genomic_accession=45421419, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100873398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873398, CurrentID=0, Status='live', Symbol='RNA5SP128', Aliases='RN5S128', description='RNA, 5S ribosomal pseudogene 128', other_designations='RNA, 5S ribosomal 128', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33491983, end_position_on_the_genomic_accession=33492098, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79031, CurrentID=0, Status='live', Symbol='PDCL3', Aliases='HTPHLP, PHLP2A, PHLP3, VIAF, VIAF1', description='phosducin like 3', other_designations='IAP-associated factor VIAF1|VIAF-1|phPL3|phosducin-like 3|viral IAP-associated factor 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100562956, end_position_on_the_genomic_accession=100576739, orientation='plus', exon_count=6, OMIM=611678, no_hdr0=''),
  100873400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873400, CurrentID=0, Status='live', Symbol='RNA5SP131', Aliases='RN5S131', description='RNA, 5S ribosomal pseudogene 131', other_designations='RNA, 5S ribosomal 131', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50419781, end_position_on_the_genomic_accession=50419899, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79033, CurrentID=0, Status='live', Symbol='ERI3', Aliases='PINT1, PRNPIP', description='ERI1 exoribonuclease family member 3', other_designations='enhanced RNAi three prime mRNA exonuclease homolog 3|prion interactor 1|prion protein-interacting protein', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44221070, end_position_on_the_genomic_accession=44355279, orientation='minus', exon_count=10, OMIM=609917, no_hdr0=''),
  79034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79034, CurrentID=0, Status='live', Symbol='C7orf26', Aliases='', description='chromosome 7 open reading frame 26', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6589368, end_position_on_the_genomic_accession=6608726, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79035, CurrentID=0, Status='live', Symbol='NABP2', Aliases='OBFC2B, SOSS-B1, SSB1, hSSB1', description='nucleic acid binding protein 2', other_designations='LP3587|nucleic acid-binding protein 2|oligonucleotide/oligosaccharide-binding fold containing 2B|oligonucleotide/oligosaccharide-binding fold-containing protein 2B|sensor of single-strand DNA complex subunit B1|sensor of ssDNA subunit B1|single strand DNA-binding protein 1|single-stranded DNA-binding protein 1', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56222093, end_position_on_the_genomic_accession=56229854, orientation='plus', exon_count=9, OMIM=612104, no_hdr0=''),
  79036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79036, CurrentID=0, Status='live', Symbol='KXD1', Aliases='C19orf50, KXDL, MST096, MSTP096', description='KxDL motif containing 1', other_designations='UPF0459 protein C19orf50', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18557762, end_position_on_the_genomic_accession=18569387, orientation='plus', exon_count=9, OMIM=615178, no_hdr0=''),
  79037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79037, CurrentID=0, Status='live', Symbol='PVRIG', Aliases='C7orf15', description='poliovirus receptor related immunoglobulin domain containing', other_designations='poliovirus receptor-related immunoglobulin domain-containing protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100216717, end_position_on_the_genomic_accession=100221490, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79038, CurrentID=0, Status='live', Symbol='ZFYVE21', Aliases='HCVP7TP1, ZF21', description='zinc finger, FYVE domain containing 21', other_designations='FYVE domain containing 21|HCV p7-transactivated protein 1', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103715744, end_position_on_the_genomic_accession=103733668, orientation='plus', exon_count=8, OMIM=613504, no_hdr0=''),
  79039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79039, CurrentID=0, Status='live', Symbol='DDX54', Aliases='DP97', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 54', other_designations='ATP-dependent RNA helicase DP97|DEAD box RNA helicase 97 kDa|DEAD box helicase 97 KDa|DEAD box protein 54', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113157173, end_position_on_the_genomic_accession=113185479, orientation='minus', exon_count=20, OMIM=611665, no_hdr0=''),
  100873408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873408, CurrentID=0, Status='live', Symbol='RNA5SP141', Aliases='RN5S141', description='RNA, 5S ribosomal pseudogene 141', other_designations='RNA, 5S ribosomal 141', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134783436, end_position_on_the_genomic_accession=134783555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79041, CurrentID=0, Status='live', Symbol='TMEM38A', Aliases='TRIC-A, TRICA', description='transmembrane protein 38A', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16661127, end_position_on_the_genomic_accession=16689019, orientation='plus', exon_count=6, OMIM=611235, no_hdr0=''),
  79042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79042, CurrentID=0, Status='live', Symbol='TSEN34', Aliases='LENG5, PCH2C, SEN34, SEN34L', description='tRNA splicing endonuclease subunit 34', other_designations='CTD-3093M3.1|TSEN34 tRNA splicing endonuclease subunit|hsSen34|leukocyte receptor cluster (LRC) member 5|leukocyte receptor cluster member 5|tRNA-intron endonuclease Sen34', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54189517, end_position_on_the_genomic_accession=54194536, orientation='plus', exon_count=8, OMIM=608754, no_hdr0=''),
  100873411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873411, CurrentID=0, Status='live', Symbol='RNA5SP144', Aliases='RN5S144', description='RNA, 5S ribosomal pseudogene 144', other_designations='RNA, 5S ribosomal 144', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=144186905, end_position_on_the_genomic_accession=144187020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873412, CurrentID=0, Status='live', Symbol='RNA5SP145', Aliases='RN5S145', description='RNA, 5S ribosomal pseudogene 145', other_designations='RNA, 5S ribosomal 145', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151188099, end_position_on_the_genomic_accession=151188223, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873413, CurrentID=0, Status='live', Symbol='RNA5SP146', Aliases='RN5S146', description='RNA, 5S ribosomal pseudogene 146', other_designations='RNA, 5S ribosomal 146', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157173918, end_position_on_the_genomic_accession=157174040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873414, CurrentID=0, Status='live', Symbol='RNA5SP147', Aliases='RN5S147', description='RNA, 5S ribosomal pseudogene 147', other_designations='RNA, 5S ribosomal 147', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176763233, end_position_on_the_genomic_accession=176763341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79047, CurrentID=0, Status='live', Symbol='KCTD15', Aliases='', description='potassium channel tetramerization domain containing 15', other_designations='potassium channel tetramerisation domain containing 15|potassium channel tetramerization domain-containing protein 15', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33796846, end_position_on_the_genomic_accession=33815761, orientation='plus', exon_count=10, OMIM=615240, no_hdr0=''),
  79048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79048, CurrentID=0, Status='live', Symbol='SECISBP2', Aliases='SBP2', description='SECIS binding protein 2', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89318473, end_position_on_the_genomic_accession=89367117, orientation='plus', exon_count=22, OMIM=607693, no_hdr0=''),
  100873417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873417, CurrentID=0, Status='live', Symbol='RNA5SP151', Aliases='RN5S151', description='RNA, 5S ribosomal pseudogene 151', other_designations='RNA, 5S ribosomal 151', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183258300, end_position_on_the_genomic_accession=183258416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79050, CurrentID=0, Status='live', Symbol='NOC4L', Aliases='NET49, NOC4, UTP19', description='nucleolar complex associated 4 homolog', other_designations='NOC4 protein homolog|NOC4-like protein|nucleolar complex-associated protein 4-like protein', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132144422, end_position_on_the_genomic_accession=132152473, orientation='plus', exon_count=15, OMIM=612819, no_hdr0=''),
  79051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79051, CurrentID=0, Status='live', Symbol='RPY', Aliases='', description='Retinitis pigmentosa, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=400004, no_hdr0=''),
  79052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79052, CurrentID=0, Status='live', Symbol='PTCPRN', Aliases='PRN1', description='Papillary thyroid carcinoma with papillary renal neoplasia', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605642, no_hdr0=''),
  79053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79053, CurrentID=0, Status='live', Symbol='ALG8', Aliases='CDG1H', description='ALG8, alpha-1,3-glucosyltransferase', other_designations='HUSSY-02|asparagine-linked glycosylation 8 alpha-13-glucosyltransferase-like protein|asparagine-linked glycosylation 8 homolog (S. cerevisiae, alpha-1,3-glucosyltransferase)|asparagine-linked glycosylation 8 homolog (yeast, alpha-1,3-glucosyltransferase)|asparagine-linked glycosylation 8, alpha-1,3-glucosyltransferase homolog|asparagine-linked glycosylation protein 8 homolog|dol-P-Glc:Glc(1)Man(9)GlcNAc(2)-PP-dolichyl alpha-1,3-glucosyltransferase|dolichyl pyrophosphate Glc1Man9GlcNAc2 alpha-1,3-glucosyltransferase|dolichyl-P-Glc:Glc(1)Man(9)GlcNAc(2)-PP-dolichol alpha- 1->3-glucosyltransferase|dolichyl-P-Glc:Glc1Man9GlcNAc2-PP-dolichyl glucosyltransferase|dolichyl-P-glucose:Glc1Man9GlcNAc2-PP-dolichyl-alpha-1,3-glucosyltransferase', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78100942, end_position_on_the_genomic_accession=78139660, orientation='minus', exon_count=15, OMIM=608103, no_hdr0=''),
  79054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79054, CurrentID=0, Status='live', Symbol='TRPM8', Aliases='LTRPC6, TRPP8', description='transient receptor potential cation channel, subfamily M, member 8', other_designations='LTrpC-6|long transient receptor potential channel 6|short form of the TRPM8 cationic channel|transient receptor melastatin 8 variant 1|transient receptor melastatin 8 variant 2|transient receptor potential p8|transient receptor potential subfamily M member 8|trp-p8', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233917362, end_position_on_the_genomic_accession=234019522, orientation='plus', exon_count=26, OMIM=606678, no_hdr0=''),
  79055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79055, CurrentID=0, Status='live', Symbol='PSMA2P3', Aliases='PSMA2L', description='proteasome (prosome, macropain) subunit, alpha type, 2 pseudogene 3', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202179543, end_position_on_the_genomic_accession=202179985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79056, CurrentID=0, Status='live', Symbol='PRRG4', Aliases='PRGP4, TMG4', description='proline rich Gla (G-carboxyglutamic acid) 4 (transmembrane)', other_designations='proline rich Gla 4 (transmembrane) isoform 1|proline rich Gla 4 (transmembrane) isoform 2|proline rich Gla 4 (transmembrane) isoform 3|proline-rich Gla protein 4|proline-rich gamma-carboxyglutamic acid protein 4', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32829774, end_position_on_the_genomic_accession=32858123, orientation='plus', exon_count=6, OMIM=611690, no_hdr0=''),
  79057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79057, CurrentID=0, Status='live', Symbol='PRRG3', Aliases='PRGP3, TMG3', description='proline rich Gla (G-carboxyglutamic acid) 3 (transmembrane)', other_designations='proline-rich Gla protein 3|proline-rich gamma-carboxyglutamic acid protein 3', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151694676, end_position_on_the_genomic_accession=151705924, orientation='plus', exon_count=8, OMIM=300685, no_hdr0=''),
  79058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79058, CurrentID=0, Status='live', Symbol='ASPSCR1', Aliases='ASPCR1, ASPL, ASPS, RCC17, TUG, UBXD9, UBXN9', description='alveolar soft part sarcoma chromosome region, candidate 1', other_designations='UBX domain protein 9|UBX domain-containing protein 9|alveolar soft part sarcoma chromosomal region candidate gene 1 protein|alveolar soft part sarcoma locus|renal cell carcinoma gene on chromosome 17|renal cell carcinoma, papillary, 17|renal papillary cell carcinoma protein 17', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81977550, end_position_on_the_genomic_accession=82017563, orientation='plus', exon_count=19, OMIM=606236, no_hdr0=''),
  100873427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873427, CurrentID=0, Status='live', Symbol='RNA5SP162', Aliases='RN5S162', description='RNA, 5S ribosomal pseudogene 162', other_designations='RNA, 5S ribosomal 162', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56331177, end_position_on_the_genomic_accession=56331217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873428, CurrentID=0, Status='live', Symbol='RNA5SP163', Aliases='RN5S163', description='RNA, 5S ribosomal pseudogene 163', other_designations='RNA, 5S ribosomal 163', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71759518, end_position_on_the_genomic_accession=71759616, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873429, CurrentID=0, Status='live', Symbol='RNA5SP164', Aliases='RN5S164', description='RNA, 5S ribosomal pseudogene 164', other_designations='RNA, 5S ribosomal 164', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=93820171, end_position_on_the_genomic_accession=93820280, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873430, CurrentID=0, Status='live', Symbol='RNA5SP165', Aliases='RN5S165', description='RNA, 5S ribosomal pseudogene 165', other_designations='RNA, 5S ribosomal 165', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147745303, end_position_on_the_genomic_accession=147745417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873431, CurrentID=0, Status='live', Symbol='RNA5SP166', Aliases='RN5S166', description='RNA, 5S ribosomal pseudogene 166', other_designations='RNA, 5S ribosomal 166', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148557947, end_position_on_the_genomic_accession=148558048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341208, CurrentID=0, Status='live', Symbol='HEPHL1', Aliases='', description='hephaestin like 1', other_designations='hephaestin-like 1', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94021212, end_position_on_the_genomic_accession=94114208, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  79065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79065, CurrentID=0, Status='live', Symbol='ATG9A', Aliases='APG9L1, MGD3208, mATG9', description='autophagy related 9A', other_designations='APG9 autophagy 9-like 1|APG9-like 1|ATG9 autophagy related 9 homolog A|autophagy 9-like 1 protein', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219219380, end_position_on_the_genomic_accession=219229688, orientation='minus', exon_count=16, OMIM=612204, no_hdr0=''),
  79066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79066, CurrentID=0, Status='live', Symbol='METTL16', Aliases='METT10D', description='methyltransferase like 16', other_designations='methyltransferase 10 domain containing|methyltransferase 10 domain-containing protein|putative methyltransferase METT10D', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2416049, end_position_on_the_genomic_accession=2511906, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100873435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873435, CurrentID=0, Status='live', Symbol='RNA5SP170', Aliases='RN5S170', description='RNA, 5S ribosomal pseudogene 170', other_designations='RNA, 5S ribosomal 170', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=166053187, end_position_on_the_genomic_accession=166053287, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79068, CurrentID=0, Status='live', Symbol='FTO', Aliases='ALKBH9, GDFD', description='fat mass and obesity associated', other_designations='AlkB homolog 9|fat mass and obesity-associated protein', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53703963, end_position_on_the_genomic_accession=54114467, orientation='plus', exon_count=12, OMIM=610966, no_hdr0=''),
  100873437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873437, CurrentID=0, Status='live', Symbol='RNA5SP173', Aliases='RN5S173', description='RNA, 5S ribosomal pseudogene 173', other_designations='RNA, 5S ribosomal 173', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=178406695, end_position_on_the_genomic_accession=178406830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79070, CurrentID=0, Status='live', Symbol='KDELC1', Aliases='EP58, KDEL1', description='KDEL motif containing 1', other_designations='ER protein 58|ERp58|KDEL (Lys-Asp-Glu-Leu) containing 1|KDEL motif-containing 1|endoplasmic reticulum resident protein 58', map_location='13q33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102784281, end_position_on_the_genomic_accession=102799054, orientation='minus', exon_count=12, OMIM=611613, no_hdr0=''),
  79071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79071, CurrentID=0, Status='live', Symbol='ELOVL6', Aliases='FACE, FAE, LCE', description='ELOVL fatty acid elongase 6', other_designations='3-keto acyl-CoA synthase ELOVL6|ELOVL FA elongase 6|ELOVL family member 6, elongation of long chain fatty acids (FEN1/Elo2, SUR4/Elo3-like, yeast)|fatty acid elongase 2|fatty acyl-CoA elongase|hELO2|long-chain fatty-acyl elongase|very-long-chain 3-oxoacyl-CoA synthase 6', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110049073, end_position_on_the_genomic_accession=110199199, orientation='minus', exon_count=8, OMIM=611546, no_hdr0=''),
  79072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79072, CurrentID=0, Status='live', Symbol='FASTKD3', Aliases='', description='FAST kinase domains 3', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7859159, end_position_on_the_genomic_accession=7869037, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79073, CurrentID=0, Status='live', Symbol='TMEM109', Aliases='', description='transmembrane protein 109', other_designations='mg23|mitsugumin-23', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60913899, end_position_on_the_genomic_accession=60923443, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79074, CurrentID=0, Status='live', Symbol='C2orf49', Aliases='asw', description='chromosome 2 open reading frame 49', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105337359, end_position_on_the_genomic_accession=105359940, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79075, CurrentID=0, Status='live', Symbol='DSCC1', Aliases='DCC1', description='DNA replication and sister chromatid cohesion 1', other_designations='defective in sister chromatid cohesion 1 homolog|defective in sister chromatid cohesion protein 1 homolog', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119833941, end_position_on_the_genomic_accession=119855930, orientation='minus', exon_count=9, OMIM=613203, no_hdr0=''),
  100873444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873444, CurrentID=0, Status='live', Symbol='RNA5SP182', Aliases='RN5S182', description='RNA, 5S ribosomal pseudogene 182', other_designations='RNA, 5S ribosomal 182', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51886688, end_position_on_the_genomic_accession=51886769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79077, CurrentID=0, Status='live', Symbol='DCTPP1', Aliases='CDA03, RS21C6, XTP3TPA', description='dCTP pyrophosphatase 1', other_designations='XTP3-transactivated gene A protein|XTP3-transactivated protein A|dCTPase 1|deoxycytidine-triphosphatase 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30423698, end_position_on_the_genomic_accession=30430052, orientation='minus', exon_count=3, OMIM=615840, no_hdr0=''),
  79078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79078, CurrentID=0, Status='live', Symbol='C1orf50', Aliases='', description='chromosome 1 open reading frame 50', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42767245, end_position_on_the_genomic_accession=42775742, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100873447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873447, CurrentID=0, Status='live', Symbol='RNA5SP185', Aliases='RN5S185', description='RNA, 5S ribosomal pseudogene 185', other_designations='RNA, 5S ribosomal 185', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56260333, end_position_on_the_genomic_accession=56260446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79080, CurrentID=0, Status='live', Symbol='CCDC86', Aliases='', description='coiled-coil domain containing 86', other_designations='cyclon|cytokine-induced protein with coiled-coil domain', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60841956, end_position_on_the_genomic_accession=60851088, orientation='plus', exon_count=4, OMIM=611293, no_hdr0=''),
  79081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79081, CurrentID=0, Status='live', Symbol='LBHD1', Aliases='C11orf48', description='LBH domain containing 1', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62662817, end_position_on_the_genomic_accession=62671769, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100873450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873450, CurrentID=0, Status='live', Symbol='RNA5SP188', Aliases='RN5S188', description='RNA, 5S ribosomal pseudogene 188', other_designations='RNA, 5S ribosomal 188', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102131118, end_position_on_the_genomic_accession=102131135, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79083, CurrentID=0, Status='live', Symbol='MLPH', Aliases='SLAC2-A', description='melanophilin', other_designations='exophilin-3|slp homolog lacking C2 domains a|synaptotagmin-like protein 2a', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237486410, end_position_on_the_genomic_accession=237555318, orientation='plus', exon_count=20, OMIM=606526, no_hdr0=''),
  79084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79084, CurrentID=0, Status='live', Symbol='WDR77', Aliases='HKMT1069, MEP-50, MEP50, Nbla10071, p44, p44/Mep50', description='WD repeat domain 77', other_designations='WD repeat-containing protein 77|androgen receptor cofactor p44', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111439890, end_position_on_the_genomic_accession=111449208, orientation='minus', exon_count=10, OMIM=611734, no_hdr0=''),
  79085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79085, CurrentID=0, Status='live', Symbol='SLC25A23', Aliases='APC2, MCSC2, SCaMC-3', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 23', other_designations='mitochondrial ATP-Mg/Pi carrier protein 2|mitochondrial Ca(2+)-dependent solute carrier protein 2|mitochondrial Ca2+-dependent solute carrier protein 2|short calcium-binding mitochondrial carrier 3|small calcium-binding mitochondrial carrier protein 3|solute carrier family 25 member 23', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6436081, end_position_on_the_genomic_accession=6460221, orientation='minus', exon_count=14, OMIM=608746, no_hdr0=''),
  79086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79086, CurrentID=0, Status='live', Symbol='SMIM7', Aliases='C19orf42', description='small integral membrane protein 7', other_designations='UPF0608 protein C19orf42', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16630743, end_position_on_the_genomic_accession=16660157, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79087, CurrentID=0, Status='live', Symbol='ALG12', Aliases='CDG1G, ECM39, PP14673, hALG12', description='ALG12, alpha-1,6-mannosyltransferase', other_designations='asparagine-linked glycosylation 12 homolog (S. cerevisiae, alpha-1,6-mannosyltransferase)|asparagine-linked glycosylation 12 homolog (yeast, alpha-1,6-mannosyltransferase)|asparagine-linked glycosylation 12, alpha-1,6-mannosyltransferase homolog|asparagine-linked glycosylation protein 12 homolog|dol-P-Man dependent alpha-1,6-mannosyltransferase|dolichyl-P-Man:Man(7)GlcNAc(2)-PP-dolichol alpha-1,6-mannosyltransferase|dolichyl-P-Man:Man(7)GlcNAc(2)-PP-dolichyl-alpha-1,6-mannosyltransferase|dolichyl-P-mannose:Man-7-GlcNAc-2-PP-dolichyl-alpha-6-mannosyltransferase|mannosyltransferase ALG12 homolog|membrane protein SB87', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49859147, end_position_on_the_genomic_accession=49918458, orientation='minus', exon_count=13, OMIM=607144, no_hdr0=''),
  79088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79088, CurrentID=0, Status='live', Symbol='ZNF426', Aliases='K-RBP', description='zinc finger protein 426', other_designations='KSHV RTA binding protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9526551, end_position_on_the_genomic_accession=9538645, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79089, CurrentID=0, Status='live', Symbol='TMUB2', Aliases='FP2653', description='transmembrane and ubiquitin-like domain containing 2', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44186969, end_position_on_the_genomic_accession=44191731, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79090, CurrentID=0, Status='live', Symbol='TRAPPC6A', Aliases='TRS33', description='trafficking protein particle complex 6A', other_designations='TRAPP complex subunit 6A', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45162928, end_position_on_the_genomic_accession=45178243, orientation='minus', exon_count=6, OMIM=610396, no_hdr0=''),
  79091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79091, CurrentID=0, Status='live', Symbol='METTL22', Aliases='C16orf68', description='methyltransferase like 22', other_designations='LP8272', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8621640, end_position_on_the_genomic_accession=8661208, orientation='plus', exon_count=14, OMIM=615261, no_hdr0=''),
  79092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79092, CurrentID=0, Status='live', Symbol='CARD14', Aliases='BIMP2, CARMA2, PRP, PSORS2, PSS1', description='caspase recruitment domain family member 14', other_designations='CARD-containing MAGUK protein 2|bcl10-interacting maguk protein 2|caspase recruitment domain family, member 14', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80169992, end_position_on_the_genomic_accession=80209331, orientation='plus', exon_count=27, OMIM=607211, no_hdr0=''),
  100873461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873461, CurrentID=0, Status='live', Symbol='RNA5SP200', Aliases='RN5S200', description='RNA, 5S ribosomal pseudogene 200', other_designations='RNA, 5S ribosomal 200', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173292517, end_position_on_the_genomic_accession=173292619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79094, CurrentID=0, Status='live', Symbol='CHAC1', Aliases='', description='ChaC glutathione-specific gamma-glutamylcyclotransferase 1', other_designations='ChaC, cation transport regulator homolog 1|ChaC, cation transport regulator-like 1|blocks Notch protein|botch|cation transport regulator-like protein 1|gamma-GCG 1|gamma-GCT acting on glutathione homolog 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40953438, end_position_on_the_genomic_accession=40956519, orientation='plus', exon_count=4, OMIM=614587, no_hdr0=''),
  79095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79095, CurrentID=0, Status='live', Symbol='C9orf16', Aliases='EST00098', description='chromosome 9 open reading frame 16', other_designations='', map_location='9q34.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128159425, end_position_on_the_genomic_accession=128163928, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79096, CurrentID=0, Status='live', Symbol='C11orf49', Aliases='', description='chromosome 11 open reading frame 49', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46936528, end_position_on_the_genomic_accession=47164381, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  79097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79097, CurrentID=0, Status='live', Symbol='TRIM48', Aliases='RNF101', description='tripartite motif containing 48', other_designations='RING finger protein 101|tripartite motif-containing 48', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55262182, end_position_on_the_genomic_accession=55271119, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79098, CurrentID=0, Status='live', Symbol='C1orf116', Aliases='SARG', description='chromosome 1 open reading frame 116', other_designations='specifically androgen-regulated protein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207018521, end_position_on_the_genomic_accession=207032761, orientation='minus', exon_count=4, OMIM=611680, no_hdr0=''),
  100873467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873467, CurrentID=0, Status='live', Symbol='RNA5SP207', Aliases='RN5S207', description='RNA, 5S ribosomal pseudogene 207', other_designations='RNA, 5S ribosomal 207', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41239520, end_position_on_the_genomic_accession=41239626, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79100, CurrentID=0, Status='live', Symbol='LINC00626', Aliases='', description='long intergenic non-protein coding RNA 626', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168786941, end_position_on_the_genomic_accession=168792888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79101, CurrentID=0, Status='live', Symbol='TAF1D', Aliases='JOSD3, RAFI41, TAF(I)41, TAFI41', description='TATA-box binding protein associated factor, RNA polymerase I, D', other_designations='Josephin domain containing 3|RNA polymerase I-specific TBP-associated factor 41 kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase I, D, 41kDa|TATA box binding protein associated factor 1D|TATA box-binding protein-associated factor 1D|TBP-associated factor 1D|transcription initiation factor SL1/TIF-IB subunit D', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93735929, end_position_on_the_genomic_accession=93741537, orientation='minus', exon_count=6, OMIM=612823, no_hdr0=''),
  79102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79102, CurrentID=0, Status='live', Symbol='RNF26', Aliases='', description='ring finger protein 26', other_designations='ring finger protein with leucine zipper', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119334500, end_position_on_the_genomic_accession=119337314, orientation='plus', exon_count=1, OMIM=606130, no_hdr0=''),
  100873471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873471, CurrentID=0, Status='live', Symbol='RNA5SP211', Aliases='RN5S211', description='RNA, 5S ribosomal pseudogene 211', other_designations='RNA, 5S ribosomal 211', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106449381, end_position_on_the_genomic_accession=106449497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79104, CurrentID=0, Status='live', Symbol='MEG8', Aliases='Bsr, Irm, LINC00024, NCRNA00024, Rian', description='maternally expressed 8 (non-protein coding)', other_designations='Imprinted RNA near Meg3/Gtl2|RNA Imprinted and Accumulated in Nucleus|long intergenic non-protein coding RNA 24|maternally expressed (in Callipyge) 8', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100894770, end_position_on_the_genomic_accession=100906968, orientation='plus', exon_count=4, OMIM=613648, no_hdr0=''),
  100873473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873473, CurrentID=0, Status='live', Symbol='RNA5SP213', Aliases='RN5S213', description='RNA, 5S ribosomal pseudogene 213', other_designations='RNA, 5S ribosomal 213', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114220681, end_position_on_the_genomic_accession=114220786, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873474, CurrentID=0, Status='live', Symbol='RNA5SP214', Aliases='RN5S214', description='RNA, 5S ribosomal pseudogene 214', other_designations='RNA, 5S ribosomal 214', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117060682, end_position_on_the_genomic_accession=117060799, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873475, CurrentID=0, Status='live', Symbol='RNA5SP215', Aliases='RN5S215', description='RNA, 5S ribosomal pseudogene 215', other_designations='RNA, 5S ribosomal 215', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=120686621, end_position_on_the_genomic_accession=120686742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873476, CurrentID=0, Status='live', Symbol='RNA5SP216', Aliases='RN5S216', description='RNA, 5S ribosomal pseudogene 216', other_designations='RNA, 5S ribosomal 216', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125979812, end_position_on_the_genomic_accession=125979934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79109, CurrentID=0, Status='live', Symbol='MAPKAP1', Aliases='JC310, MIP1, SIN1, SIN1b, SIN1g', description='mitogen-activated protein kinase associated protein 1', other_designations='MEKK2-interacting protein 1|SAPK-interacting protein 1|TORC2 subunit MAPKAP1|mSIN1|mitogen-activated protein kinase 2-associated protein 1|ras inhibitor MGC2745|stress-activated map kinase interacting protein 1|stress-activated map kinase-interacting protein 1|stress-activated protein kinase-interacting 1', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125437394, end_position_on_the_genomic_accession=125707234, orientation='minus', exon_count=16, OMIM=610558, no_hdr0=''),
  100873478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873478, CurrentID=0, Status='live', Symbol='RNA5SP218', Aliases='RN5S218', description='RNA, 5S ribosomal pseudogene 218', other_designations='RNA, 5S ribosomal 218', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134257035, end_position_on_the_genomic_accession=134257143, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873479, CurrentID=0, Status='live', Symbol='RNA5SP219', Aliases='RN5S219', description='RNA, 5S ribosomal pseudogene 219', other_designations='RNA, 5S ribosomal 219', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136630243, end_position_on_the_genomic_accession=136630355, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873480, CurrentID=0, Status='live', Symbol='RNA5SP220', Aliases='RN5S220', description='RNA, 5S ribosomal pseudogene 220', other_designations='RNA, 5S ribosomal 220', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140158591, end_position_on_the_genomic_accession=140158694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873481, CurrentID=0, Status='live', Symbol='RNA5SP222', Aliases='RN5S222', description='RNA, 5S ribosomal pseudogene 222', other_designations='RNA, 5S ribosomal 222', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146340420, end_position_on_the_genomic_accession=146340524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873482, CurrentID=0, Status='live', Symbol='RNA5SP223', Aliases='RN5S223', description='RNA, 5S ribosomal pseudogene 223', other_designations='RNA, 5S ribosomal 223', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152410275, end_position_on_the_genomic_accession=152410394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873483, CurrentID=0, Status='live', Symbol='RNA5SP224', Aliases='RN5S224', description='RNA, 5S ribosomal pseudogene 224', other_designations='RNA, 5S ribosomal 224', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153159220, end_position_on_the_genomic_accession=153159345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873484, CurrentID=0, Status='live', Symbol='RNA5SP225', Aliases='RN5S225', description='RNA, 5S ribosomal pseudogene 225', other_designations='RNA, 5S ribosomal 225', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153420438, end_position_on_the_genomic_accession=153420572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873485, CurrentID=0, Status='live', Symbol='RNA5SP227', Aliases='RN5S227', description='RNA, 5S ribosomal pseudogene 227', other_designations='RNA, 5S ribosomal 227', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22298594, end_position_on_the_genomic_accession=22298706, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873486, CurrentID=0, Status='live', Symbol='RNA5SP228', Aliases='RN5S228', description='RNA, 5S ribosomal pseudogene 228', other_designations='RNA, 5S ribosomal 228', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24132705, end_position_on_the_genomic_accession=24132837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873487, CurrentID=0, Status='live', Symbol='RNA5SP229', Aliases='RN5S229', description='RNA, 5S ribosomal pseudogene 229', other_designations='RNA, 5S ribosomal 229', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33177740, end_position_on_the_genomic_accession=33177838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102577424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102577424, CurrentID=0, Status='live', Symbol='LINC01574', Aliases='HI-LNC12', description='long intergenic non-protein coding RNA 1574', other_designations='Human Islet Long Non Coding RNA 12|TCONS_00009551', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176743205, end_position_on_the_genomic_accession=176743921, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102577425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102577425, CurrentID=0, Status='live', Symbol='ARHGAP23P1', Aliases='', description='Rho GTPase activating protein 23 pseudogene 1', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33905444, end_position_on_the_genomic_accession=33936913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873490, CurrentID=0, Status='live', Symbol='RNA5SP232', Aliases='RN5S232', description='RNA, 5S ribosomal pseudogene 232', other_designations='RNA, 5S ribosomal 232', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71913738, end_position_on_the_genomic_accession=71913856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873491, CurrentID=0, Status='live', Symbol='RNA5SP233', Aliases='RN5S233', description='RNA, 5S ribosomal pseudogene 233', other_designations='RNA, 5S ribosomal 233', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74487428, end_position_on_the_genomic_accession=74487540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873492, CurrentID=0, Status='live', Symbol='RNA5SP234', Aliases='RN5S234', description='RNA, 5S ribosomal pseudogene 234', other_designations='RNA, 5S ribosomal 234', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79654109, end_position_on_the_genomic_accession=79654227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  777774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777774, CurrentID=0, Status='live', Symbol='ACTR3BP4', Aliases='', description='ACTR3B pseudogene 4', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 4', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102962129, end_position_on_the_genomic_accession=102963771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873494, CurrentID=0, Status='live', Symbol='RNA5SP236', Aliases='RN5S236', description='RNA, 5S ribosomal pseudogene 236', other_designations='RNA, 5S ribosomal 236', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106781600, end_position_on_the_genomic_accession=106781715, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873495, CurrentID=0, Status='live', Symbol='RNA5SP237', Aliases='RN5S237', description='RNA, 5S ribosomal pseudogene 237', other_designations='RNA, 5S ribosomal 237', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111953653, end_position_on_the_genomic_accession=111953765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873496, CurrentID=0, Status='live', Symbol='RNA5SP238', Aliases='RN5S238', description='RNA, 5S ribosomal pseudogene 238', other_designations='RNA, 5S ribosomal 238', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=114613787, end_position_on_the_genomic_accession=114613906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873497, CurrentID=0, Status='live', Symbol='RNA5SP239', Aliases='RN5S239', description='RNA, 5S ribosomal pseudogene 239', other_designations='RNA, 5S ribosomal 239', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116944286, end_position_on_the_genomic_accession=116944414, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873498, CurrentID=0, Status='live', Symbol='RNA5SP240', Aliases='RN5S240', description='RNA, 5S ribosomal pseudogene 240', other_designations='RNA, 5S ribosomal 240', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120981426, end_position_on_the_genomic_accession=120981553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873499, CurrentID=0, Status='live', Symbol='RNA5SP241', Aliases='RN5S241', description='RNA, 5S ribosomal pseudogene 241', other_designations='RNA, 5S ribosomal 241', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121083700, end_position_on_the_genomic_accession=121083813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79132, CurrentID=0, Status='live', Symbol='DHX58', Aliases='D11LGP2, D11lgp2e, LGP2, RLR-3', description='DEXH (Asp-Glu-X-His) box polypeptide 58', other_designations='RIG-I-like receptor 3|RIG-I-like receptor LGP2|RLR|RNA helicase LGP2|ortholog of mouse D11lgp2|probable ATP-dependent helicase LGP2|protein D11Lgp2 homolog', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42101404, end_position_on_the_genomic_accession=42112733, orientation='minus', exon_count=14, OMIM=608588, no_hdr0=''),
  79133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79133, CurrentID=0, Status='live', Symbol='NDUFAF5', Aliases='C20orf7, bA526K24.2, dJ842G6.1', description='NADH:ubiquinone oxidoreductase complex assembly factor 5', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 5|probable methyltransferase C20orf7, mitochondrial', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13785026, end_position_on_the_genomic_accession=13818421, orientation='plus', exon_count=16, OMIM=612360, no_hdr0=''),
  79134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79134, CurrentID=0, Status='live', Symbol='TMEM185B', Aliases='FAM11B', description='transmembrane protein 185B', other_designations='ee3_2|family with sequence similarity 11, member B|putative transmembrane protein 185B|transmembrane protein 185B (pseudogene)', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120217471, end_position_on_the_genomic_accession=120223408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79135, CurrentID=0, Status='live', Symbol='APOO', Aliases='FAM121B, MIC26, Mic23, My025', description='apolipoprotein O', other_designations='MICOS complex subunit MIC23|brain my025|family with sequence similarity 121B', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23833348, end_position_on_the_genomic_accession=23907940, orientation='minus', exon_count=11, OMIM=300753, no_hdr0=''),
  79136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79136, CurrentID=0, Status='live', Symbol='LY6G6E', Aliases='C6orf22, G6e', description='lymphocyte antigen 6 complex, locus G6E (pseudogene)', other_designations='lymphocyte antigen-6 G6E', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31711976, end_position_on_the_genomic_accession=31714065, orientation='minus', exon_count=4, OMIM=610437, no_hdr0=''),
  79137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79137, CurrentID=0, Status='live', Symbol='FAM134A', Aliases='C2orf17', description='family with sequence similarity 134 member A', other_designations='family with sequence similarity 134, member A', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219178217, end_position_on_the_genomic_accession=219185475, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100873506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873506, CurrentID=0, Status='live', Symbol='RNA5SP250', Aliases='RN5S250', description='RNA, 5S ribosomal pseudogene 250', other_designations='RNA, 5S ribosomal 250', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152592973, end_position_on_the_genomic_accession=152593091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79139, CurrentID=0, Status='live', Symbol='DERL1', Aliases='DER-1, DER1', description='derlin 1', other_designations='DERtrin-1|Der1-like domain family, member 1|degradation in endoplasmic reticulum protein 1', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123013164, end_position_on_the_genomic_accession=123042423, orientation='minus', exon_count=9, OMIM=608813, no_hdr0=''),
  79140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79140, CurrentID=0, Status='live', Symbol='CCDC28B', Aliases='', description='coiled-coil domain containing 28B', other_designations='', map_location='1p36.11-p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32200386, end_position_on_the_genomic_accession=32205390, orientation='plus', exon_count=6, OMIM=610162, no_hdr0=''),
  100873509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873509, CurrentID=0, Status='live', Symbol='RNA5SP255', Aliases='RN5S255', description='RNA, 5S ribosomal pseudogene 255', other_designations='RNA, 5S ribosomal 255', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13536029, end_position_on_the_genomic_accession=13536135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79142, CurrentID=0, Status='live', Symbol='PHF23', Aliases='hJUNE-1b', description='PHD finger protein 23', other_designations='PDH-containing protein JUNE-1|PHD finger protein 23a', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7235028, end_position_on_the_genomic_accession=7239717, orientation='minus', exon_count=8, OMIM=612910, no_hdr0=''),
  79143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79143, CurrentID=0, Status='live', Symbol='MBOAT7', Aliases='BB1, LENG4, LPIAT, LRC4, MBOA7, OACT7, hMBOA-7', description='membrane bound O-acyltransferase domain containing 7', other_designations='1-acylglycerophosphatidylinositol O-acyltransferase|LPLAT 7|bladder and breast carcinoma-overexpressed gene 1 protein|h-mboa-7|leukocyte receptor cluster (LRC) member 4|lyso-PI acyltransferase|lysophosphatidylinositol acyltransferase|malignant cell expression-enhanced gene/tumor progression-enhanced', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54173412, end_position_on_the_genomic_accession=54189882, orientation='minus', exon_count=9, OMIM=606048, no_hdr0=''),
  79144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79144, CurrentID=0, Status='live', Symbol='PPDPF', Aliases='C20orf149, dJ697K14.9, exdpf', description='pancreatic progenitor cell differentiation and proliferation factor', other_designations='exocrine differentiation and proliferation factor|pancreatic progenitor cell differentiation and proliferation factor homolog', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63520078, end_position_on_the_genomic_accession=63522171, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79145, CurrentID=0, Status='live', Symbol='CHCHD7', Aliases='COX23', description='coiled-coil-helix-coiled-coil-helix domain containing 7', other_designations='COX23 cytochrome c oxidase assembly homolog', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56211756, end_position_on_the_genomic_accession=56218619, orientation='plus', exon_count=6, OMIM=611238, no_hdr0=''),
  100873514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873514, CurrentID=0, Status='live', Symbol='RNA5SP260', Aliases='RN5S260', description='RNA, 5S ribosomal pseudogene 260', other_designations='RNA, 5S ribosomal 260', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29048494, end_position_on_the_genomic_accession=29048604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79147, CurrentID=0, Status='live', Symbol='FKRP', Aliases='LGMD2I, MDC1C, MDDGA5, MDDGB5, MDDGC5', description='fukutin related protein', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46744606, end_position_on_the_genomic_accession=46758575, orientation='plus', exon_count=9, OMIM=606596, no_hdr0=''),
  79148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79148, CurrentID=0, Status='live', Symbol='MMP28', Aliases='EPILYSIN, MM28, MMP-25, MMP-28, MMP25', description='matrix metallopeptidase 28', other_designations='matrix metalloprotease MMP25', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35756249, end_position_on_the_genomic_accession=35795707, orientation='minus', exon_count=13, OMIM=608417, no_hdr0=''),
  79149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79149, CurrentID=0, Status='live', Symbol='ZSCAN5A', Aliases='ZNF495, ZSCAN5', description='zinc finger and SCAN domain containing 5A', other_designations='zinc finger and SCAN domain containing 5|zinc finger protein 495', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56221303, end_position_on_the_genomic_accession=56230399, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79150, CurrentID=0, Status='live', Symbol='MGC4859', Aliases='', description='uncharacterized LOC79150', other_designations='', map_location='7p', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873519, CurrentID=0, Status='live', Symbol='RNA5SP266', Aliases='RN5S266', description='RNA, 5S ribosomal pseudogene 266', other_designations='RNA, 5S ribosomal 266', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57126655, end_position_on_the_genomic_accession=57126762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79152, CurrentID=0, Status='live', Symbol='FA2H', Aliases='FAAH, FAH1, FAXDC1, SCS7, SPG35', description='fatty acid 2-hydroxylase', other_designations='fatty acid alpha-hydroxylase|fatty acid hydroxylase domain containing 1|spastic paraplegia 35 (autosomal recessive)', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74712955, end_position_on_the_genomic_accession=74774831, orientation='minus', exon_count=9, OMIM=611026, no_hdr0=''),
  79153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79153, CurrentID=0, Status='live', Symbol='GDPD3', Aliases='GDE7', description='glycerophosphodiester phosphodiesterase domain containing 3', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30104810, end_position_on_the_genomic_accession=30113943, orientation='minus', exon_count=10, OMIM=616318, no_hdr0=''),
  79154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79154, CurrentID=0, Status='live', Symbol='DHRS11', Aliases='ARPG836, SDR24C1', description='dehydrogenase/reductase (SDR family) member 11', other_designations='short chain dehydrogenase/reductase family 24C, member 1|short-chain dehydrogenase/reductase family 24C member 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36591786, end_position_on_the_genomic_accession=36600804, orientation='plus', exon_count=9, OMIM=616159, no_hdr0=''),
  79155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79155, CurrentID=0, Status='live', Symbol='TNIP2', Aliases='ABIN2, FLIP1, KLIP', description='TNFAIP3 interacting protein 2', other_designations='A20-binding inhibitor of NF-kappaB activation-2|fetal liver LKB1-interacting protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2741660, end_position_on_the_genomic_accession=2756376, orientation='minus', exon_count=6, OMIM=610669, no_hdr0=''),
  79156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79156, CurrentID=0, Status='live', Symbol='PLEKHF1', Aliases='APPD, LAPF, PHAFIN1, ZFYVE15', description='pleckstrin homology and FYVE domain containing 1', other_designations='PH and FYVE domain-containing protein 1|PH domain-containing family F member 1|apoptosis-inducing protein D|lysosome-associated apoptosis-inducing protein containing PH and FYVE domains|phafin 1|phafin-1|pleckstrin homology domain containing, family F (with FYVE domain) member 1|zinc finger FYVE domain-containing protein 15', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29665420, end_position_on_the_genomic_accession=29675477, orientation='plus', exon_count=3, OMIM=615200, no_hdr0=''),
  79157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79157, CurrentID=0, Status='live', Symbol='MFSD11', Aliases='ET', description='major facilitator superfamily domain containing 11', other_designations='major facilitator superfamily domain-containing protein 11|protein ET', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76736565, end_position_on_the_genomic_accession=76804356, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  79158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79158, CurrentID=0, Status='live', Symbol='GNPTAB', Aliases='GNPTA, ICD', description='N-acetylglucosamine-1-phosphate transferase, alpha and beta subunits', other_designations='GlcNAc phosphotransferase|UDP-N-acetylglucosamine-1-phosphotransferase subunits alpha/beta|UDP-N-acetylglucosamine-lysosomal-enzyme N-acetylglucosamine|glcNAc-1-phosphotransferase subunits alpha/beta|glucosamine (UDP-N-acetyl)-lysosomal-enzyme N-acetylglucosamine phosphotransferase|stealth protein GNPTAB', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101745497, end_position_on_the_genomic_accession=101830867, orientation='minus', exon_count=23, OMIM=607840, no_hdr0=''),
  79159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79159, CurrentID=0, Status='live', Symbol='NOL12', Aliases='Nop25, dJ37E16.7', description='nucleolar protein 12', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37686337, end_position_on_the_genomic_accession=37693478, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873528, CurrentID=0, Status='live', Symbol='RNA5SP275', Aliases='RN5S275', description='RNA, 5S ribosomal pseudogene 275', other_designations='RNA, 5S ribosomal 275', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107884494, end_position_on_the_genomic_accession=107884611, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79161, CurrentID=0, Status='live', Symbol='TMEM243', Aliases='C7orf23, MM-TRAG', description='transmembrane protein 243', other_designations='MDR1 and mitochondrial taxol resistance associated|transmembrane protein 243, mitochondrial|transmembrane protein C7orf23', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87196160, end_position_on_the_genomic_accession=87220573, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873530, CurrentID=0, Status='live', Symbol='RNA5SP277', Aliases='RN5S277', description='RNA, 5S ribosomal pseudogene 277', other_designations='RNA, 5S ribosomal 277', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119891132, end_position_on_the_genomic_accession=119891229, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873531, CurrentID=0, Status='live', Symbol='RNA5SP278', Aliases='RN5S278', description='RNA, 5S ribosomal pseudogene 278', other_designations='RNA, 5S ribosomal 278', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140940562, end_position_on_the_genomic_accession=140940663, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873532, CurrentID=0, Status='live', Symbol='RNA5SP279', Aliases='RN5S279', description='RNA, 5S ribosomal pseudogene 279', other_designations='RNA, 5S ribosomal 279', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1333837, end_position_on_the_genomic_accession=1333953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79165, CurrentID=0, Status='live', Symbol='LENG1', Aliases='', description='leukocyte receptor cluster (LRC) member 1', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54155639, end_position_on_the_genomic_accession=54159743, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79166, CurrentID=0, Status='live', Symbol='LILRP2', Aliases='CD85m, ILT10, LILRA5', description='leukocyte immunoglobulin-like receptor pseudogene 2', other_designations='immunoglobulin-like transcript 10|leukocyte immunoglobulin-like receptor, subfamily A (without TM domain), member 5', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54707916, end_position_on_the_genomic_accession=54713452, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79167, CurrentID=0, Status='live', Symbol='LILRP1', Aliases='CD85l, ILT9, LILRA6P', description='leukocyte immunoglobulin-like receptor pseudogene 1', other_designations='immunoglobulin-like transcript 9 pseudogene|leukocyte immunoglobulin-like receptor, subfamily A (without TM domain), member6, pseudogene', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54696327, end_position_on_the_genomic_accession=54698529, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79168, CurrentID=0, Status='live', Symbol='LILRA6', Aliases='CD85b, ILT-8, ILT5, ILT8, LILRB3, LILRB6', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 6', other_designations='immunoglobulin-like transcript 5|immunoglobulin-like transcript 8|leukocyte Ig-like receptor|leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 6', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54236592, end_position_on_the_genomic_accession=54242768, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79169, CurrentID=0, Status='live', Symbol='C1orf35', Aliases='MMTAG2', description='chromosome 1 open reading frame 35', other_designations='multiple myeloma transforming 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228100727, end_position_on_the_genomic_accession=228105411, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79170, CurrentID=0, Status='live', Symbol='PRR15L', Aliases='ATAD4', description='proline rich 15-like', other_designations='ATPase family, AAA domain containing 4', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47951967, end_position_on_the_genomic_accession=47957894, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79171, CurrentID=0, Status='live', Symbol='RBM42', Aliases='', description='RNA binding motif protein 42', other_designations='RNA-binding motif protein 42', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35629078, end_position_on_the_genomic_accession=35637685, orientation='plus', exon_count=10, OMIM=613232, no_hdr0=''),
  79172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79172, CurrentID=0, Status='live', Symbol='CENPO', Aliases='CENP-O, ICEN-36, MCM21R', description='centromere protein O', other_designations='centromeric protein O|interphase centromere complex protein 36', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24793306, end_position_on_the_genomic_accession=24822376, orientation='plus', exon_count=8, OMIM=611504, no_hdr0=''),
  79173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79173, CurrentID=0, Status='live', Symbol='C19orf57', Aliases='', description='chromosome 19 open reading frame 57', other_designations='pre-T/NK cell associated protein (3B3)|pre-T/NK cell-associated protein 3B3', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13882348, end_position_on_the_genomic_accession=13906099, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79174, CurrentID=0, Status='live', Symbol='CRELD2', Aliases='', description='cysteine rich with EGF-like domains 2', other_designations='cysteine-rich with EGF-like domains 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49918630, end_position_on_the_genomic_accession=49927540, orientation='plus', exon_count=12, OMIM=607171, no_hdr0=''),
  79175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79175, CurrentID=0, Status='live', Symbol='ZNF343', Aliases='dJ734P14.5', description='zinc finger protein 343', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2480071, end_position_on_the_genomic_accession=2524522, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79176, CurrentID=0, Status='live', Symbol='FBXL15', Aliases='FBXO37, Fbl15, JET, PSD', description='F-box and leucine-rich repeat protein 15', other_designations='F-box only protein 37|pleckstrin and Sec7 domain protein', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102419061, end_position_on_the_genomic_accession=102423138, orientation='plus', exon_count=7, OMIM=610287, no_hdr0=''),
  79177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79177, CurrentID=0, Status='live', Symbol='ZNF576', Aliases='', description='zinc finger protein 576', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43596392, end_position_on_the_genomic_accession=43600435, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79178, CurrentID=0, Status='live', Symbol='THTPA', Aliases='THTPSE, THTPA', description='thiamine triphosphatase', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23511760, end_position_on_the_genomic_accession=23559581, orientation='plus', exon_count=6, OMIM=611612, no_hdr0=''),
  79179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79179, CurrentID=0, Status='live', Symbol='FHII', Aliases='FHA2', description='Hyperaldosteronism, familial, type II', other_designations='', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605635, no_hdr0=''),
  79180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79180, CurrentID=0, Status='live', Symbol='EFHD2', Aliases='SWS1', description='EF-hand domain family member D2', other_designations='EF hand domain containing 2|EF hand domain family, member D2|EF-hand domain family, member D2|swiprosin 1|swiprosin-1', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15409895, end_position_on_the_genomic_accession=15430343, orientation='plus', exon_count=5, OMIM=616450, no_hdr0=''),
  100873549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873549, CurrentID=0, Status='live', Symbol='RNA5SP296', Aliases='RN5S296', description='RNA, 5S ribosomal pseudogene 296', other_designations='RNA, 5S ribosomal 296', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127841112, end_position_on_the_genomic_accession=127841222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873550, CurrentID=0, Status='live', Symbol='RNA5SP499', Aliases='RN5S499', description='RNA, 5S ribosomal pseudogene 499', other_designations='RNA, 5S ribosomal 499', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9133828, end_position_on_the_genomic_accession=9133946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79183, CurrentID=0, Status='live', Symbol='TTPAL', Aliases='C20orf121', description='tocopherol (alpha) transfer protein-like', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44475883, end_position_on_the_genomic_accession=44494603, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79184, CurrentID=0, Status='live', Symbol='BRCC3', Aliases='BRCC36, C6.1A, CXorf53', description='BRCA1/BRCA2-containing complex subunit 3', other_designations='BRCA1-A complex subunit BRCC36|BRCA1/BRCA2-containing complex subunit 36|BRCA1/BRCA2-containing complex, subunit 3|BRISC complex subunit BRCC36', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155071420, end_position_on_the_genomic_accession=155123074, orientation='plus', exon_count=12, OMIM=300617, no_hdr0=''),
  100873553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873553, CurrentID=0, Status='live', Symbol='RNA5SP502', Aliases='RN5S502', description='RNA, 5S ribosomal pseudogene 502', other_designations='RNA, 5S ribosomal 502', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41292798, end_position_on_the_genomic_accession=41292916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873554, CurrentID=0, Status='live', Symbol='RNA5SP503', Aliases='RN5S503', description='RNA, 5S ribosomal pseudogene 503', other_designations='RNA, 5S ribosomal 503', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48206258, end_position_on_the_genomic_accession=48206374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79187, CurrentID=0, Status='live', Symbol='FSD1', Aliases='GLFND, MIR1', description='fibronectin type III and SPRY domain containing 1', other_designations='MID1-related protein 1|fibronectin type 3 and SPRY (spla, ryanodine) domain containing (with coiled-coil motif) 1|fibronectin type 3 and SPRY domain containing 1|microtubule-associated protein GLFND|midline 1-related protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4304594, end_position_on_the_genomic_accession=4323846, orientation='plus', exon_count=13, OMIM=609828, no_hdr0=''),
  79188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79188, CurrentID=0, Status='live', Symbol='TMEM43', Aliases='ARVC5, ARVD5, EDMD7, LUMA', description='transmembrane protein 43', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14124940, end_position_on_the_genomic_accession=14143680, orientation='plus', exon_count=13, OMIM=612048, no_hdr0=''),
  341333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341333, CurrentID=0, Status='live', Symbol='HNRNPA1P70', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 70', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68035767, end_position_on_the_genomic_accession=68036774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79190, CurrentID=0, Status='live', Symbol='IRX6', Aliases='IRX-3, IRX7, IRXB3', description='iroquois homeobox 6', other_designations='homeodomain protein IRXB3|iroquois homeobox protein 6|iroquois homeobox protein 7', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55323962, end_position_on_the_genomic_accession=55330760, orientation='plus', exon_count=6, OMIM=606196, no_hdr0=''),
  79191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79191, CurrentID=0, Status='live', Symbol='IRX3', Aliases='IRX-1, IRXB1', description='iroquois homeobox 3', other_designations='homeodomain protein IRXB1|iroquois homeobox protein 3|iroquois-class homeodomain protein IRX-1', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54283300, end_position_on_the_genomic_accession=54286856, orientation='minus', exon_count=4, OMIM=612985, no_hdr0=''),
  79192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79192, CurrentID=0, Status='live', Symbol='IRX1', Aliases='IRX-5, IRXA1', description='iroquois homeobox 1', other_designations='homeodomain protein IRXA1|iroquois homeobox protein 1', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3596054, end_position_on_the_genomic_accession=3601403, orientation='plus', exon_count=4, OMIM=606197, no_hdr0=''),
  100873561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873561, CurrentID=0, Status='live', Symbol='RNA5SP510', Aliases='RN5S510', description='RNA, 5S ribosomal pseudogene 510', other_designations='RNA, 5S ribosomal 510', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95582064, end_position_on_the_genomic_accession=95582162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873562, CurrentID=0, Status='live', Symbol='RNA5SP511', Aliases='RN5S511', description='RNA, 5S ribosomal pseudogene 511', other_designations='RNA, 5S ribosomal 511', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103850797, end_position_on_the_genomic_accession=103850911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873563, CurrentID=0, Status='live', Symbol='RNA5SP512', Aliases='RN5S512', description='RNA, 5S ribosomal pseudogene 512', other_designations='RNA, 5S ribosomal 512', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111669829, end_position_on_the_genomic_accession=111669948, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873564, CurrentID=0, Status='live', Symbol='RNA5SP513', Aliases='RN5S513', description='RNA, 5S ribosomal pseudogene 513', other_designations='RNA, 5S ribosomal 513', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129436530, end_position_on_the_genomic_accession=129436647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873565, CurrentID=0, Status='live', Symbol='RNA5SP514', Aliases='RN5S514', description='RNA, 5S ribosomal pseudogene 514', other_designations='RNA, 5S ribosomal 514', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131938846, end_position_on_the_genomic_accession=131938956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873566, CurrentID=0, Status='live', Symbol='RNA5SP515', Aliases='RN5S515', description='RNA, 5S ribosomal pseudogene 515', other_designations='RNA, 5S ribosomal 515', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135428168, end_position_on_the_genomic_accession=135428285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873567, CurrentID=0, Status='live', Symbol='RNA5SP516', Aliases='RN5S516', description='RNA, 5S ribosomal pseudogene 516', other_designations='RNA, 5S ribosomal 516', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142411053, end_position_on_the_genomic_accession=142411171, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873568, CurrentID=0, Status='live', Symbol='RNA5SP517', Aliases='RN5S517', description='RNA, 5S ribosomal pseudogene 517', other_designations='RNA, 5S ribosomal 517', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146427456, end_position_on_the_genomic_accession=146427572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873569, CurrentID=0, Status='live', Symbol='RNA5SP518', Aliases='RN5S518', description='RNA, 5S ribosomal pseudogene 518', other_designations='RNA, 5S ribosomal 518', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10090410, end_position_on_the_genomic_accession=10090528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341346, CurrentID=0, Status='live', Symbol='SMCO2', Aliases='C12orf70', description='single-pass membrane protein with coiled-coil domains 2', other_designations='', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27446744, end_position_on_the_genomic_accession=27502322, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100873571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873571, CurrentID=0, Status='live', Symbol='RNA5-8SP2', Aliases='RN5-8S2', description='RNA, 5.8S ribosomal pseudogene 2', other_designations='RNA, 5.8S ribosomal 2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34162959, end_position_on_the_genomic_accession=34163110, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873572, CurrentID=0, Status='live', Symbol='RNA5-8SP4', Aliases='RN5-8S4', description='RNA, 5.8S ribosomal pseudogene 4', other_designations='RNA, 5.8S ribosomal 4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=24004358, end_position_on_the_genomic_accession=24004507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873573, CurrentID=0, Status='live', Symbol='RNA5-8SP7', Aliases='RN5-8S7', description='RNA, 5.8S ribosomal pseudogene 7', other_designations='RNA, 5.8S ribosomal 7', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5346006, end_position_on_the_genomic_accession=5346160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341350, CurrentID=0, Status='live', Symbol='OVCH1', Aliases='OVCH', description='ovochymase 1', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29422128, end_position_on_the_genomic_accession=29497800, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  100873575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873575, CurrentID=0, Status='live', Symbol='RNA5SP300', Aliases='RN5S300', description='RNA, 5S ribosomal pseudogene 300', other_designations='RNA, 5S ribosomal 300', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12924679, end_position_on_the_genomic_accession=12924785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873576, CurrentID=0, Status='live', Symbol='RNA5SP301', Aliases='RN5S301', description='RNA, 5S ribosomal pseudogene 301', other_designations='RNA, 5S ribosomal 301', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13727796, end_position_on_the_genomic_accession=13727916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873577, CurrentID=0, Status='live', Symbol='RNA5SP303', Aliases='RN5S303', description='RNA, 5S ribosomal pseudogene 303', other_designations='RNA, 5S ribosomal 303', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19633683, end_position_on_the_genomic_accession=19633812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873578, CurrentID=0, Status='live', Symbol='RNA5SP304', Aliases='RN5S304', description='RNA, 5S ribosomal pseudogene 304', other_designations='RNA, 5S ribosomal 304', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22997255, end_position_on_the_genomic_accession=22997358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873579, CurrentID=0, Status='live', Symbol='RNA5SP305', Aliases='RN5S305', description='RNA, 5S ribosomal pseudogene 305', other_designations='RNA, 5S ribosomal 305', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24640166, end_position_on_the_genomic_accession=24640284, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341356, CurrentID=0, Status='live', Symbol='RPL31P50', Aliases='RPL31_16_1224', description='ribosomal protein L31 pseudogene 50', other_designations='hCG1644239', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31597106, end_position_on_the_genomic_accession=31597408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873581, CurrentID=0, Status='live', Symbol='RNA5SP307', Aliases='RN5S307', description='RNA, 5S ribosomal pseudogene 307', other_designations='RNA, 5S ribosomal 307', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26509589, end_position_on_the_genomic_accession=26509704, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873582, CurrentID=0, Status='live', Symbol='RNA5SP308', Aliases='RN5S308', description='RNA, 5S ribosomal pseudogene 308', other_designations='RNA, 5S ribosomal 308', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28874393, end_position_on_the_genomic_accession=28874507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341359, CurrentID=0, Status='live', Symbol='SYT10', Aliases='', description='synaptotagmin 10', other_designations='synaptotagmin X|sytX', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=33375413, end_position_on_the_genomic_accession=33439819, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100873584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873584, CurrentID=0, Status='live', Symbol='RNA5SP318', Aliases='RN5S318', description='RNA, 5S ribosomal pseudogene 318', other_designations='RNA, 5S ribosomal 318', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=53458197, end_position_on_the_genomic_accession=53458288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873585, CurrentID=0, Status='live', Symbol='RNA5SP319', Aliases='RN5S319', description='RNA, 5S ribosomal pseudogene 319', other_designations='RNA, 5S ribosomal 319', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68461407, end_position_on_the_genomic_accession=68461525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873586, CurrentID=0, Status='live', Symbol='RNA5SP320', Aliases='RN5S320', description='RNA, 5S ribosomal pseudogene 320', other_designations='RNA, 5S ribosomal 320', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73706490, end_position_on_the_genomic_accession=73706613, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873587, CurrentID=0, Status='live', Symbol='RNA5SP321', Aliases='RN5S321', description='RNA, 5S ribosomal pseudogene 321', other_designations='RNA, 5S ribosomal 321', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77587049, end_position_on_the_genomic_accession=77587159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873588, CurrentID=0, Status='live', Symbol='RNA5SP322', Aliases='RN5S322', description='RNA, 5S ribosomal pseudogene 322', other_designations='RNA, 5S ribosomal 322', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86052448, end_position_on_the_genomic_accession=86052566, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873589, CurrentID=0, Status='live', Symbol='RNA5SP323', Aliases='RN5S323', description='RNA, 5S ribosomal pseudogene 323', other_designations='RNA, 5S ribosomal 323', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93510548, end_position_on_the_genomic_accession=93510680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873590, CurrentID=0, Status='live', Symbol='RNA5SP324', Aliases='RN5S324', description='RNA, 5S ribosomal pseudogene 324', other_designations='RNA, 5S ribosomal 324', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96619376, end_position_on_the_genomic_accession=96619483, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873591, CurrentID=0, Status='live', Symbol='RNA5SP325', Aliases='RN5S325', description='RNA, 5S ribosomal pseudogene 325', other_designations='RNA, 5S ribosomal 325', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106970021, end_position_on_the_genomic_accession=106970150, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873592, CurrentID=0, Status='live', Symbol='RNA5SP326', Aliases='RN5S326', description='RNA, 5S ribosomal pseudogene 326', other_designations='RNA, 5S ribosomal 326', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=107461560, end_position_on_the_genomic_accession=107461669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873593, CurrentID=0, Status='live', Symbol='RNA5SP327', Aliases='RN5S327', description='RNA, 5S ribosomal pseudogene 327', other_designations='RNA, 5S ribosomal 327', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110761142, end_position_on_the_genomic_accession=110761242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873594, CurrentID=0, Status='live', Symbol='RNA5SP328', Aliases='RN5S328', description='RNA, 5S ribosomal pseudogene 328', other_designations='RNA, 5S ribosomal 328', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126145474, end_position_on_the_genomic_accession=126145582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873595, CurrentID=0, Status='live', Symbol='RNA5SP329', Aliases='RN5S329', description='RNA, 5S ribosomal pseudogene 329', other_designations='RNA, 5S ribosomal 329', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6016996, end_position_on_the_genomic_accession=6017114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79228, CurrentID=0, Status='live', Symbol='THOC6', Aliases='BBIS, WDR58, fSAP35', description='THO complex 6', other_designations='THO complex 6 homolog|WD repeat domain 58|WD repeat-containing protein 58|functional spliceosome-associated protein 35', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3024031, end_position_on_the_genomic_accession=3027755, orientation='plus', exon_count=13, OMIM=615403, no_hdr0=''),
  100873597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873597, CurrentID=0, Status='live', Symbol='RNA5SP331', Aliases='RN5S331', description='RNA, 5S ribosomal pseudogene 331', other_designations='RNA, 5S ribosomal 331', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13907484, end_position_on_the_genomic_accession=13907590, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79230, CurrentID=0, Status='live', Symbol='ZNF557', Aliases='', description='zinc finger protein 557', other_designations='CTB-25J19.9', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7069460, end_position_on_the_genomic_accession=7087968, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100873599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873599, CurrentID=0, Status='live', Symbol='RNA5SP336', Aliases='RN5S336', description='RNA, 5S ribosomal pseudogene 336', other_designations='RNA, 5S ribosomal 336', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=21000881, end_position_on_the_genomic_accession=21000997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873600, CurrentID=0, Status='live', Symbol='RNA5SP337', Aliases='RN5S337', description='RNA, 5S ribosomal pseudogene 337', other_designations='RNA, 5S ribosomal 337', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=21383801, end_position_on_the_genomic_accession=21383909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873601, CurrentID=0, Status='live', Symbol='RNA5SP338', Aliases='RN5S338', description='RNA, 5S ribosomal pseudogene 338', other_designations='RNA, 5S ribosomal 338', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22688351, end_position_on_the_genomic_accession=22688435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873602, CurrentID=0, Status='live', Symbol='RNA5SP339', Aliases='RN5S339', description='RNA, 5S ribosomal pseudogene 339', other_designations='RNA, 5S ribosomal 339', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27521605, end_position_on_the_genomic_accession=27521726, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873603, CurrentID=0, Status='live', Symbol='RNA5SP340', Aliases='RN5S340', description='RNA, 5S ribosomal pseudogene 340', other_designations='RNA, 5S ribosomal 340', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47804296, end_position_on_the_genomic_accession=47804402, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873604, CurrentID=0, Status='live', Symbol='RNA5SP341', Aliases='RN5S341', description='RNA, 5S ribosomal pseudogene 341', other_designations='RNA, 5S ribosomal 341', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57450183, end_position_on_the_genomic_accession=57450299, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873605, CurrentID=0, Status='live', Symbol='RNA5SP343', Aliases='RN5S343', description='RNA, 5S ribosomal pseudogene 343', other_designations='RNA, 5S ribosomal 343', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74198748, end_position_on_the_genomic_accession=74198854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873606, CurrentID=0, Status='live', Symbol='RNA5SP344', Aliases='RN5S344', description='RNA, 5S ribosomal pseudogene 344', other_designations='RNA, 5S ribosomal 344', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75934936, end_position_on_the_genomic_accession=75935039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873607, CurrentID=0, Status='live', Symbol='RNA5SP345', Aliases='RN5S345', description='RNA, 5S ribosomal pseudogene 345', other_designations='RNA, 5S ribosomal 345', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95840017, end_position_on_the_genomic_accession=95840124, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873608, CurrentID=0, Status='live', Symbol='RNA5SP346', Aliases='RN5S346', description='RNA, 5S ribosomal pseudogene 346', other_designations='RNA, 5S ribosomal 346', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96474572, end_position_on_the_genomic_accession=96474692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873609, CurrentID=0, Status='live', Symbol='RNA5SP347', Aliases='RN5S347', description='RNA, 5S ribosomal pseudogene 347', other_designations='RNA, 5S ribosomal 347', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97657464, end_position_on_the_genomic_accession=97657582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873610, CurrentID=0, Status='live', Symbol='RNA5SP348', Aliases='RN5S348', description='RNA, 5S ribosomal pseudogene 348', other_designations='RNA, 5S ribosomal 348', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104252579, end_position_on_the_genomic_accession=104252615, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873611, CurrentID=0, Status='live', Symbol='RNA5SP349', Aliases='RN5S349', description='RNA, 5S ribosomal pseudogene 349', other_designations='RNA, 5S ribosomal 349', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109120878, end_position_on_the_genomic_accession=109120986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873612, CurrentID=0, Status='live', Symbol='RNA5SP352', Aliases='RN5S352', description='RNA, 5S ribosomal pseudogene 352', other_designations='RNA, 5S ribosomal 352', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124636459, end_position_on_the_genomic_accession=124636592, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873613, CurrentID=0, Status='live', Symbol='RNA5SP353', Aliases='RN5S353', description='RNA, 5S ribosomal pseudogene 353', other_designations='RNA, 5S ribosomal 353', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13440884, end_position_on_the_genomic_accession=13441001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873614, CurrentID=0, Status='live', Symbol='RNA5SP354', Aliases='RN5S354', description='RNA, 5S ribosomal pseudogene 354', other_designations='RNA, 5S ribosomal 354', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26373598, end_position_on_the_genomic_accession=26373733, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873615, CurrentID=0, Status='live', Symbol='RNA5SP356', Aliases='RN5S356', description='RNA, 5S ribosomal pseudogene 356', other_designations='RNA, 5S ribosomal 356', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30306431, end_position_on_the_genomic_accession=30306524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341392, CurrentID=0, Status='live', Symbol='ACSM4', Aliases='', description='acyl-CoA synthetase medium-chain family member 4', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7304332, end_position_on_the_genomic_accession=7328373, orientation='plus', exon_count=13, OMIM=614360, no_hdr0=''),
  100873617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873617, CurrentID=0, Status='live', Symbol='RNA5SP358', Aliases='RN5S358', description='RNA, 5S ribosomal pseudogene 358', other_designations='RNA, 5S ribosomal 358', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38161466, end_position_on_the_genomic_accession=38161582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873618, CurrentID=0, Status='live', Symbol='RNA5SP359', Aliases='RN5S359', description='RNA, 5S ribosomal pseudogene 359', other_designations='RNA, 5S ribosomal 359', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38163403, end_position_on_the_genomic_accession=38163515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873619, CurrentID=0, Status='live', Symbol='RNA5SP360', Aliases='RN5S360', description='RNA, 5S ribosomal pseudogene 360', other_designations='RNA, 5S ribosomal 360', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41658676, end_position_on_the_genomic_accession=41658793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873620, CurrentID=0, Status='live', Symbol='RNA5SP361', Aliases='RN5S361', description='RNA, 5S ribosomal pseudogene 361', other_designations='RNA, 5S ribosomal 361', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45117771, end_position_on_the_genomic_accession=45117888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873621, CurrentID=0, Status='live', Symbol='RNA5SP362', Aliases='RN5S362', description='RNA, 5S ribosomal pseudogene 362', other_designations='RNA, 5S ribosomal 362', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66066221, end_position_on_the_genomic_accession=66066338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873622, CurrentID=0, Status='live', Symbol='RNA5SP363', Aliases='RN5S363', description='RNA, 5S ribosomal pseudogene 363', other_designations='RNA, 5S ribosomal 363', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79878305, end_position_on_the_genomic_accession=79878419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873623, CurrentID=0, Status='live', Symbol='RNA5SP364', Aliases='RN5S364', description='RNA, 5S ribosomal pseudogene 364', other_designations='RNA, 5S ribosomal 364', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88051701, end_position_on_the_genomic_accession=88051778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873624, CurrentID=0, Status='live', Symbol='RNA5SP365', Aliases='RN5S365', description='RNA, 5S ribosomal pseudogene 365', other_designations='RNA, 5S ribosomal 365', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89726936, end_position_on_the_genomic_accession=89727042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873625, CurrentID=0, Status='live', Symbol='RNA5SP366', Aliases='RN5S366', description='RNA, 5S ribosomal pseudogene 366', other_designations='RNA, 5S ribosomal 366', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=99083783, end_position_on_the_genomic_accession=99083905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79258, CurrentID=0, Status='live', Symbol='MMEL1', Aliases='MMEL2, NEP2, NEPII, NL1, NL2, SEP', description='membrane metallo-endopeptidase-like 1', other_designations='NEP2(m)|membrane metallo-endopeptidase-like 2|neprilysin II|neprilysin-2|soluble secreted endopeptidase|zinc metallopeptidase', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2590639, end_position_on_the_genomic_accession=2633042, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  100873627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873627, CurrentID=0, Status='live', Symbol='RNA5SP368', Aliases='RN5S368', description='RNA, 5S ribosomal pseudogene 368', other_designations='RNA, 5S ribosomal 368', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101779814, end_position_on_the_genomic_accession=101779929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873628, CurrentID=0, Status='live', Symbol='RNA5SP369', Aliases='RN5S369', description='RNA, 5S ribosomal pseudogene 369', other_designations='RNA, 5S ribosomal 369', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101836820, end_position_on_the_genomic_accession=101836927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341405, CurrentID=0, Status='live', Symbol='ANKRD33', Aliases='C12orf7, PANKY', description='ankyrin repeat domain 33', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51888009, end_position_on_the_genomic_accession=51891721, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873630, CurrentID=0, Status='live', Symbol='RNA5SP372', Aliases='RN5S372', description='RNA, 5S ribosomal pseudogene 372', other_designations='RNA, 5S ribosomal 372', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109059986, end_position_on_the_genomic_accession=109060100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873631, CurrentID=0, Status='live', Symbol='RNA5SP373', Aliases='RN5S373', description='RNA, 5S ribosomal pseudogene 373', other_designations='RNA, 5S ribosomal 373', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111339527, end_position_on_the_genomic_accession=111339621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873632, CurrentID=0, Status='live', Symbol='RNA5SP374', Aliases='RN5S374', description='RNA, 5S ribosomal pseudogene 374', other_designations='RNA, 5S ribosomal 374', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118829681, end_position_on_the_genomic_accession=118829805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873633, CurrentID=0, Status='live', Symbol='RNA5SP375', Aliases='RN5S375', description='RNA, 5S ribosomal pseudogene 375', other_designations='RNA, 5S ribosomal 375', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123282916, end_position_on_the_genomic_accession=123283034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873634, CurrentID=0, Status='live', Symbol='RNA5SP378', Aliases='RN5S378', description='RNA, 5S ribosomal pseudogene 378', other_designations='RNA, 5S ribosomal 378', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131801315, end_position_on_the_genomic_accession=131801434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873635, CurrentID=0, Status='live', Symbol='RNA5SP379', Aliases='RN5S379', description='RNA, 5S ribosomal pseudogene 379', other_designations='RNA, 5S ribosomal 379', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132723945, end_position_on_the_genomic_accession=132724057, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873636, CurrentID=0, Status='live', Symbol='RNA5SP380', Aliases='RN5S380', description='RNA, 5S ribosomal pseudogene 380', other_designations='RNA, 5S ribosomal 380', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20090130, end_position_on_the_genomic_accession=20090227, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79269, CurrentID=0, Status='live', Symbol='DCAF10', Aliases='WDR32', description='DDB1 and CUL4 associated factor 10', other_designations='WD repeat domain 32|WD repeat-containing protein 32', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37800554, end_position_on_the_genomic_accession=37867669, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100873638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873638, CurrentID=0, Status='live', Symbol='RNA5SP383', Aliases='RN5S383', description='RNA, 5S ribosomal pseudogene 383', other_designations='RNA, 5S ribosomal 383', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24158603, end_position_on_the_genomic_accession=24158722, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873639, CurrentID=0, Status='live', Symbol='RNA5SP384', Aliases='RN5S384', description='RNA, 5S ribosomal pseudogene 384', other_designations='RNA, 5S ribosomal 384', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49552633, end_position_on_the_genomic_accession=49552748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341416, CurrentID=0, Status='live', Symbol='OR6C2', Aliases='OR6C67', description='olfactory receptor family 6 subfamily C member 2', other_designations='HSA3|olfactory receptor, family 6, subfamily C, member 2', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55452214, end_position_on_the_genomic_accession=55453152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873641, CurrentID=0, Status='live', Symbol='RNA5SP386', Aliases='RN5S386', description='RNA, 5S ribosomal pseudogene 386', other_designations='RNA, 5S ribosomal 386', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67657105, end_position_on_the_genomic_accession=67657211, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341418, CurrentID=0, Status='live', Symbol='OR6C4', Aliases='OR12-10', description='olfactory receptor family 6 subfamily C member 4', other_designations='olfactory receptor OR12-10|olfactory receptor, family 6, subfamily C, member 4', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55551227, end_position_on_the_genomic_accession=55552156, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873643, CurrentID=0, Status='live', Symbol='RNA5SP390', Aliases='RN5S390', description='RNA, 5S ribosomal pseudogene 390', other_designations='RNA, 5S ribosomal 390', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25738508, end_position_on_the_genomic_accession=25738620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873644, CurrentID=0, Status='live', Symbol='RNA5SP391', Aliases='RN5S391', description='RNA, 5S ribosomal pseudogene 391', other_designations='RNA, 5S ribosomal 391', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27280318, end_position_on_the_genomic_accession=27280436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873645, CurrentID=0, Status='live', Symbol='RNA5SP392', Aliases='RN5S392', description='RNA, 5S ribosomal pseudogene 392', other_designations='RNA, 5S ribosomal 392', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40331537, end_position_on_the_genomic_accession=40331648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873646, CurrentID=0, Status='live', Symbol='RNA5SP393', Aliases='RN5S393', description='RNA, 5S ribosomal pseudogene 393', other_designations='RNA, 5S ribosomal 393', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41851913, end_position_on_the_genomic_accession=41852029, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873647, CurrentID=0, Status='live', Symbol='RNA5SP394', Aliases='RN5S394', description='RNA, 5S ribosomal pseudogene 394', other_designations='RNA, 5S ribosomal 394', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49982443, end_position_on_the_genomic_accession=49982561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873648, CurrentID=0, Status='live', Symbol='RNA5SP395', Aliases='RN5S395', description='RNA, 5S ribosomal pseudogene 395', other_designations='RNA, 5S ribosomal 395', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50463381, end_position_on_the_genomic_accession=50463498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873649, CurrentID=0, Status='live', Symbol='RNA5SP396', Aliases='RN5S396', description='RNA, 5S ribosomal pseudogene 396', other_designations='RNA, 5S ribosomal 396', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59574103, end_position_on_the_genomic_accession=59574222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873650, CurrentID=0, Status='live', Symbol='RNA5SP397', Aliases='RN5S397', description='RNA, 5S ribosomal pseudogene 397', other_designations='RNA, 5S ribosomal 397', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60736769, end_position_on_the_genomic_accession=60736886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79283, CurrentID=0, Status='live', Symbol='OR5P1P', Aliases='PJCG1', description='olfactory receptor family 5 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily P, member 1 pseudogene|olfactory receptor-like protein PJCG1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7772893, end_position_on_the_genomic_accession=7773814, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873652, CurrentID=0, Status='live', Symbol='RNA5SP399', Aliases='RN5S399', description='RNA, 5S ribosomal pseudogene 399', other_designations='RNA, 5S ribosomal 399', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71858570, end_position_on_the_genomic_accession=71858686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873653, CurrentID=0, Status='live', Symbol='RNA5SP400', Aliases='RN5S400', description='RNA, 5S ribosomal pseudogene 400', other_designations='RNA, 5S ribosomal 400', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86138269, end_position_on_the_genomic_accession=86138379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873654, CurrentID=0, Status='live', Symbol='RNA5SP401', Aliases='RN5S401', description='RNA, 5S ribosomal pseudogene 401', other_designations='RNA, 5S ribosomal 401', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97472126, end_position_on_the_genomic_accession=97472252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106479477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479477, CurrentID=0, Status='live', Symbol='RN7SL693P', Aliases='', description='RNA, 7SL, cytoplasmic 693, pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873656, CurrentID=0, Status='live', Symbol='RNA5SP404', Aliases='RN5S404', description='RNA, 5S ribosomal pseudogene 404', other_designations='RNA, 5S ribosomal 404', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9600097, end_position_on_the_genomic_accession=9600228, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79289, CurrentID=0, Status='live', Symbol='OR5H8', Aliases='OR3-7P, OR5H8', description='olfactory receptor family 5 subfamily H member 8 (gene/pseudogene)', other_designations='olfactory receptor OR3-7 pseudogene|olfactory receptor, family 5, subfamily H, member 8 pseudogene', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98311913, end_position_on_the_genomic_accession=98312840, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79290, CurrentID=0, Status='live', Symbol='OR13A1', Aliases='', description='olfactory receptor family 13 subfamily A member 1', other_designations='olfactory receptor OR10-3|olfactory receptor, family 13, subfamily A, member 1|seven transmembrane helix receptor', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45302294, end_position_on_the_genomic_accession=45315608, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79291, CurrentID=0, Status='live', Symbol='OR5H7P', Aliases='OR3-6', description='olfactory receptor family 5 subfamily H member 7 pseudogene', other_designations='olfactory receptor OR3-6 pseudogene|olfactory receptor, family 5, subfamily H, member 7 pseudogene', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98238352, end_position_on_the_genomic_accession=98239278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873660, CurrentID=0, Status='live', Symbol='RNA5SP408', Aliases='RN5S408', description='RNA, 5S ribosomal pseudogene 408', other_designations='RNA, 5S ribosomal 408', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35735372, end_position_on_the_genomic_accession=35735471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873661, CurrentID=0, Status='live', Symbol='RNA5SP409', Aliases='RN5S409', description='RNA, 5S ribosomal pseudogene 409', other_designations='RNA, 5S ribosomal 409', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35735625, end_position_on_the_genomic_accession=35735733, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873662, CurrentID=0, Status='live', Symbol='RNA5SP410', Aliases='RN5S410', description='RNA, 5S ribosomal pseudogene 410', other_designations='RNA, 5S ribosomal 410', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35745989, end_position_on_the_genomic_accession=35746094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79295, CurrentID=0, Status='live', Symbol='OR5H6', Aliases='OR3-11', description='olfactory receptor family 5 subfamily H member 6 (gene/pseudogene)', other_designations='olfactory receptor OR3-11|olfactory receptor, family 5, subfamily H, member 6', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98263260, end_position_on_the_genomic_accession=98265350, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873664, CurrentID=0, Status='live', Symbol='RNA5SP413', Aliases='RN5S413', description='RNA, 5S ribosomal pseudogene 413', other_designations='RNA, 5S ribosomal 413', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35747032, end_position_on_the_genomic_accession=35747140, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79297, CurrentID=0, Status='live', Symbol='OR5H5P', Aliases='OR3-3', description='olfactory receptor family 5 subfamily H member 5 pseudogene', other_designations='olfactory receptor OR3-3 pseudogene|olfactory receptor, family 5, subfamily H, member 14 pseudogene|olfactory receptor, family 5, subfamily H, member 5 pseudogene', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98197211, end_position_on_the_genomic_accession=98198138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873666, CurrentID=0, Status='live', Symbol='RNA5SP415', Aliases='RN5S415', description='RNA, 5S ribosomal pseudogene 415', other_designations='RNA, 5S ribosomal 415', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35750061, end_position_on_the_genomic_accession=35750175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79299, CurrentID=0, Status='live', Symbol='OR5H4P', Aliases='OR3-5', description='olfactory receptor family 5 subfamily H member 4 pseudogene', other_designations='olfactory receptor OR3-5 pseudogene|olfactory receptor, family 5, subfamily H, member 4 pseudogene', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98222039, end_position_on_the_genomic_accession=98222966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873668, CurrentID=0, Status='live', Symbol='RNA5SP417', Aliases='RN5S417', description='RNA, 5S ribosomal pseudogene 417', other_designations='RNA, 5S ribosomal 417', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35751828, end_position_on_the_genomic_accession=35751935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873669, CurrentID=0, Status='live', Symbol='RNA5SP418', Aliases='RN5S418', description='RNA, 5S ribosomal pseudogene 418', other_designations='RNA, 5S ribosomal 418', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35752826, end_position_on_the_genomic_accession=35752942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873670, CurrentID=0, Status='live', Symbol='RNA5SP419', Aliases='RN5S419', description='RNA, 5S ribosomal pseudogene 419', other_designations='RNA, 5S ribosomal 419', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35753796, end_position_on_the_genomic_accession=35753930, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873671, CurrentID=0, Status='live', Symbol='RNA5SP420', Aliases='RN5S420', description='RNA, 5S ribosomal pseudogene 420', other_designations='RNA, 5S ribosomal 420', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35754046, end_position_on_the_genomic_accession=35754182, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873672, CurrentID=0, Status='live', Symbol='RNA5SP421', Aliases='RN5S421', description='RNA, 5S ribosomal pseudogene 421', other_designations='RNA, 5S ribosomal 421', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35754555, end_position_on_the_genomic_accession=35754688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873673, CurrentID=0, Status='live', Symbol='RNA5SP422', Aliases='RN5S422', description='RNA, 5S ribosomal pseudogene 422', other_designations='RNA, 5S ribosomal 422', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35754923, end_position_on_the_genomic_accession=35755033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873674, CurrentID=0, Status='live', Symbol='RNA5SP423', Aliases='RN5S423', description='RNA, 5S ribosomal pseudogene 423', other_designations='RNA, 5S ribosomal 423', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35755177, end_position_on_the_genomic_accession=35755285, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873675, CurrentID=0, Status='live', Symbol='RNA5SP424', Aliases='RN5S424', description='RNA, 5S ribosomal pseudogene 424', other_designations='RNA, 5S ribosomal 424', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47450347, end_position_on_the_genomic_accession=47450461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873676, CurrentID=0, Status='live', Symbol='RNA5SP425', Aliases='RN5S425', description='RNA, 5S ribosomal pseudogene 425', other_designations='RNA, 5S ribosomal 425', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47505432, end_position_on_the_genomic_accession=47505543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79309, CurrentID=0, Status='live', Symbol='OR5H3P', Aliases='', description='olfactory receptor family 5 subfamily H member 3 pseudogene', other_designations='olfactory receptor OR3-4 pseudogene|olfactory receptor, family 5, subfamily H, member 14 pseudogene|olfactory receptor, family 5, subfamily H, member 3 pseudogene', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98207423, end_position_on_the_genomic_accession=98208341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79310, CurrentID=0, Status='live', Symbol='OR5H2', Aliases='OR3-10', description='olfactory receptor family 5 subfamily H member 2', other_designations='olfactory receptor OR3-10|olfactory receptor, family 5, subfamily H, member 2', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98282888, end_position_on_the_genomic_accession=98283832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873679, CurrentID=0, Status='live', Symbol='RNA5SP428', Aliases='RN5S428', description='RNA, 5S ribosomal pseudogene 428', other_designations='RNA, 5S ribosomal 428', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66301809, end_position_on_the_genomic_accession=66301940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873680, CurrentID=0, Status='live', Symbol='RNA5SP430', Aliases='RN5S430', description='RNA, 5S ribosomal pseudogene 430', other_designations='RNA, 5S ribosomal 430', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75861048, end_position_on_the_genomic_accession=75861157, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873681, CurrentID=0, Status='live', Symbol='RNA5SP431', Aliases='RN5S431', description='RNA, 5S ribosomal pseudogene 431', other_designations='RNA, 5S ribosomal 431', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79264454, end_position_on_the_genomic_accession=79264573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873682, CurrentID=0, Status='live', Symbol='RNA5SP432', Aliases='RN5S432', description='RNA, 5S ribosomal pseudogene 432', other_designations='RNA, 5S ribosomal 432', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84039918, end_position_on_the_genomic_accession=84040033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79315, CurrentID=0, Status='live', Symbol='OR7E91P', Aliases='', description='olfactory receptor family 7 subfamily E member 91 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 91 pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71024075, end_position_on_the_genomic_accession=71029931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79316, CurrentID=0, Status='live', Symbol='OR8B9P', Aliases='', description='olfactory receptor family 8 subfamily B member 9 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 9 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124481078, end_position_on_the_genomic_accession=124481563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79317, CurrentID=0, Status='live', Symbol='OR4K5', Aliases='OR14-16', description='olfactory receptor family 4 subfamily K member 5', other_designations='olfactory receptor OR14-16|olfactory receptor, family 4, subfamily K, member 5', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19920607, end_position_on_the_genomic_accession=19921578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873686, CurrentID=0, Status='live', Symbol='RNA5SP437', Aliases='RN5S437', description='RNA, 5S ribosomal pseudogene 437', other_designations='RNA, 5S ribosomal 437', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31963805, end_position_on_the_genomic_accession=31963933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873687, CurrentID=0, Status='live', Symbol='RNA5SP438', Aliases='RN5S438', description='RNA, 5S ribosomal pseudogene 438', other_designations='RNA, 5S ribosomal 438', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34120056, end_position_on_the_genomic_accession=34120167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873688, CurrentID=0, Status='live', Symbol='RNA5SP440', Aliases='RN5S440', description='RNA, 5S ribosomal pseudogene 440', other_designations='RNA, 5S ribosomal 440', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38731802, end_position_on_the_genomic_accession=38731920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873689, CurrentID=0, Status='live', Symbol='RNA5SP441', Aliases='RN5S441', description='RNA, 5S ribosomal pseudogene 441', other_designations='RNA, 5S ribosomal 441', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40374089, end_position_on_the_genomic_accession=40374176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79322, CurrentID=0, Status='live', Symbol='OR4N1P', Aliases='OR14-12, OR14-6', description='olfactory receptor family 4 subfamily N member 1 pseudogene', other_designations='olfactory receptor OR14-12 pseudogene|olfactory receptor OR14-6 pseudogene|olfactory receptor, family 4, subfamily N, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19796254, end_position_on_the_genomic_accession=19797418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79323, CurrentID=0, Status='live', Symbol='OR6E1P', Aliases='', description='olfactory receptor family 6 subfamily E member 1 pseudogene', other_designations='olfactory receptor, family 6, subfamily E, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22702144, end_position_on_the_genomic_accession=22703069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79324, CurrentID=0, Status='live', Symbol='OR51G1', Aliases='OR11-29, OR51G3P', description='olfactory receptor family 51 subfamily G member 1 (gene/pseudogene)', other_designations='olfactory receptor 51G3|olfactory receptor OR11-29|olfactory receptor, family 51, subfamily G, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4923374, end_position_on_the_genomic_accession=4924339, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  341469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341469, CurrentID=0, Status='live', Symbol='MRPS7P2', Aliases='', description='mitochondrial ribosomal protein S7 pseudogene 2', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15803116, end_position_on_the_genomic_accession=15804164, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873694, CurrentID=0, Status='live', Symbol='RNA5SP449', Aliases='RN5S449', description='RNA, 5S ribosomal pseudogene 449', other_designations='RNA, 5S ribosomal 449', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9844717, end_position_on_the_genomic_accession=9844846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873695, CurrentID=0, Status='live', Symbol='RNA5SP450', Aliases='RN5S450', description='RNA, 5S ribosomal pseudogene 450', other_designations='RNA, 5S ribosomal 450', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9923920, end_position_on_the_genomic_accession=9924021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873696, CurrentID=0, Status='live', Symbol='RNA5SP451', Aliases='RN5S451', description='RNA, 5S ribosomal pseudogene 451', other_designations='RNA, 5S ribosomal 451', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21928985, end_position_on_the_genomic_accession=21929094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873697, CurrentID=0, Status='live', Symbol='RNA5SP452', Aliases='RN5S452', description='RNA, 5S ribosomal pseudogene 452', other_designations='RNA, 5S ribosomal 452', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24170505, end_position_on_the_genomic_accession=24170623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873698, CurrentID=0, Status='live', Symbol='RNA5SP453', Aliases='RN5S453', description='RNA, 5S ribosomal pseudogene 453', other_designations='RNA, 5S ribosomal 453', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32176050, end_position_on_the_genomic_accession=32176168, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873699, CurrentID=0, Status='live', Symbol='RNA5SP454', Aliases='RN5S454', description='RNA, 5S ribosomal pseudogene 454', other_designations='RNA, 5S ribosomal 454', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=42270589, end_position_on_the_genomic_accession=42270715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873700, CurrentID=0, Status='live', Symbol='RNA5SP455', Aliases='RN5S455', description='RNA, 5S ribosomal pseudogene 455', other_designations='RNA, 5S ribosomal 455', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44071583, end_position_on_the_genomic_accession=44071701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873701, CurrentID=0, Status='live', Symbol='RNA5SP456', Aliases='RN5S456', description='RNA, 5S ribosomal pseudogene 456', other_designations='RNA, 5S ribosomal 456', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48475487, end_position_on_the_genomic_accession=48475600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79334, CurrentID=0, Status='live', Symbol='OR11H2', Aliases='C14orf15P, OR11H8P, OR11H2', description='olfactory receptor family 11 subfamily H member 2', other_designations='olfactory receptor OR14-1|seven transmembrane helix receptor', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19712904, end_position_on_the_genomic_accession=19714332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  122449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122449, CurrentID=0, Status='live', Symbol='RPS29P1', Aliases='RPS29_5_1384', description='ribosomal protein S29 pseudogene 1', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69023061, end_position_on_the_genomic_accession=69023345, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873704, CurrentID=0, Status='live', Symbol='RNA5SP460', Aliases='RN5S460', description='RNA, 5S ribosomal pseudogene 460', other_designations='RNA, 5S ribosomal 460', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72751952, end_position_on_the_genomic_accession=72752051, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873705, CurrentID=0, Status='live', Symbol='RNA5SP461', Aliases='RN5S461', description='RNA, 5S ribosomal pseudogene 461', other_designations='RNA, 5S ribosomal 461', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77863464, end_position_on_the_genomic_accession=77863567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873706, CurrentID=0, Status='live', Symbol='RNA5SP462', Aliases='RN5S462', description='RNA, 5S ribosomal pseudogene 462', other_designations='RNA, 5S ribosomal 462', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=453134, end_position_on_the_genomic_accession=453245, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79339, CurrentID=0, Status='live', Symbol='OR51B4', Aliases="HOR5'Beta1", description='olfactory receptor family 51 subfamily B member 4', other_designations="odorant receptor HOR5'beta1|olfactory receptor, family 51, subfamily B, member 4", map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5301014, end_position_on_the_genomic_accession=5301946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873708, CurrentID=0, Status='live', Symbol='RNA5SP464', Aliases='RN5S464', description='RNA, 5S ribosomal pseudogene 464', other_designations='RNA, 5S ribosomal 464', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11996809, end_position_on_the_genomic_accession=11996916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873709, CurrentID=0, Status='live', Symbol='RNA5SP465', Aliases='RN5S465', description='RNA, 5S ribosomal pseudogene 465', other_designations='RNA, 5S ribosomal 465', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12027913, end_position_on_the_genomic_accession=12028021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79342, CurrentID=0, Status='live', Symbol='OR51B3P', Aliases='', description='olfactory receptor family 51 subfamily B member 3 pseudogene', other_designations='olfactory receptor, family 51, subfamily B, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5314745, end_position_on_the_genomic_accession=5315693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873711, CurrentID=0, Status='live', Symbol='RNA5SP467', Aliases='RN5S467', description='RNA, 5S ribosomal pseudogene 467', other_designations='RNA, 5S ribosomal 467', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12106497, end_position_on_the_genomic_accession=12106614, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873712, CurrentID=0, Status='live', Symbol='RNA5SP469', Aliases='RN5S469', description='RNA, 5S ribosomal pseudogene 469', other_designations='RNA, 5S ribosomal 469', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21113128, end_position_on_the_genomic_accession=21113236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79345, CurrentID=0, Status='live', Symbol='OR51B2', Aliases="HOR5'Beta3, OR51B1P", description='olfactory receptor family 51 subfamily B member 2 (gene/pseudogene)', other_designations="odorant receptor HOR5'beta3|olfactory receptor 51B1|olfactory receptor, family 51, subfamily B, member 2", map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5323298, end_position_on_the_genomic_accession=5324352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79346, CurrentID=0, Status='live', Symbol='OR4C5', Aliases='OR4C5PQ, OR4C5', description='olfactory receptor family 4 subfamily C member 5 (gene/pseudogene)', other_designations='olfactory receptor, family 4, subfamily C, member 5 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48365397, end_position_on_the_genomic_accession=48366388, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873715, CurrentID=0, Status='live', Symbol='RNA5SP472', Aliases='RN5S472', description='RNA, 5S ribosomal pseudogene 472', other_designations='RNA, 5S ribosomal 472', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32243965, end_position_on_the_genomic_accession=32244083, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873716, CurrentID=0, Status='live', Symbol='RNA5SP474', Aliases='RN5S474', description='RNA, 5S ribosomal pseudogene 474', other_designations='RNA, 5S ribosomal 474', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5098387, end_position_on_the_genomic_accession=5098476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873717, CurrentID=0, Status='live', Symbol='RNA5SP475', Aliases='RN5S475', description='RNA, 5S ribosomal pseudogene 475', other_designations='RNA, 5S ribosomal 475', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=15280764, end_position_on_the_genomic_accession=15280873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873718, CurrentID=0, Status='live', Symbol='RNA5SP476', Aliases='RN5S476', description='RNA, 5S ribosomal pseudogene 476', other_designations='RNA, 5S ribosomal 476', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18433842, end_position_on_the_genomic_accession=18433944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873719, CurrentID=0, Status='live', Symbol='RNA5SP477', Aliases='RN5S477', description='RNA, 5S ribosomal pseudogene 477', other_designations='RNA, 5S ribosomal 477', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21138956, end_position_on_the_genomic_accession=21139074, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873720, CurrentID=0, Status='live', Symbol='RNA5SP478', Aliases='RN5S478', description='RNA, 5S ribosomal pseudogene 478', other_designations='RNA, 5S ribosomal 478', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23160857, end_position_on_the_genomic_accession=23160971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873721, CurrentID=0, Status='live', Symbol='RNA5SP479', Aliases='RN5S479', description='RNA, 5S ribosomal pseudogene 479', other_designations='RNA, 5S ribosomal 479', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23380896, end_position_on_the_genomic_accession=23381012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873722, CurrentID=0, Status='live', Symbol='RNA5SP480', Aliases='RN5S480', description='RNA, 5S ribosomal pseudogene 480', other_designations='RNA, 5S ribosomal 480', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31356870, end_position_on_the_genomic_accession=31356978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873723, CurrentID=0, Status='live', Symbol='RNA5SP481', Aliases='RN5S481', description='RNA, 5S ribosomal pseudogene 481', other_designations='RNA, 5S ribosomal 481', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32008351, end_position_on_the_genomic_accession=32008472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873724, CurrentID=0, Status='live', Symbol='RNA5SP482', Aliases='RN5S482', description='RNA, 5S ribosomal pseudogene 482', other_designations='RNA, 5S ribosomal 482', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873725, CurrentID=0, Status='live', Symbol='RNA5SP483', Aliases='RN5S483', description='RNA, 5S ribosomal pseudogene 483', other_designations='RNA, 5S ribosomal 483', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35195306, end_position_on_the_genomic_accession=35195423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873726, CurrentID=0, Status='live', Symbol='RNA5SP484', Aliases='RN5S484', description='RNA, 5S ribosomal pseudogene 484', other_designations='RNA, 5S ribosomal 484', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40854119, end_position_on_the_genomic_accession=40854229, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873727, CurrentID=0, Status='live', Symbol='RNA5SP485', Aliases='RN5S485', description='RNA, 5S ribosomal pseudogene 485', other_designations='RNA, 5S ribosomal 485', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45559758, end_position_on_the_genomic_accession=45559876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873728, CurrentID=0, Status='live', Symbol='RNA5SP486', Aliases='RN5S486', description='RNA, 5S ribosomal pseudogene 486', other_designations='RNA, 5S ribosomal 486', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47873191, end_position_on_the_genomic_accession=47873273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873729, CurrentID=0, Status='live', Symbol='RNA5SP487', Aliases='RN5S487', description='RNA, 5S ribosomal pseudogene 487', other_designations='RNA, 5S ribosomal 487', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56205052, end_position_on_the_genomic_accession=56205173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873730, CurrentID=0, Status='live', Symbol='RNA5SP488', Aliases='RN5S488', description='RNA, 5S ribosomal pseudogene 488', other_designations='RNA, 5S ribosomal 488', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14070871, end_position_on_the_genomic_accession=14070986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79363, CurrentID=0, Status='live', Symbol='RSG1', Aliases='C1orf89', description='REM2 and RAB-like small GTPase 1', other_designations='Rem/Rab-Similar GTPase 1|miro domain-containing protein C1orf89', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16231687, end_position_on_the_genomic_accession=16237452, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79364, CurrentID=0, Status='live', Symbol='ZXDC', Aliases='ZXDL', description='ZXD family zinc finger C', other_designations='SERH2790|ZXD-like zinc finger protein', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126437601, end_position_on_the_genomic_accession=126475933, orientation='minus', exon_count=11, OMIM=615746, no_hdr0=''),
  79365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79365, CurrentID=0, Status='live', Symbol='BHLHE41', Aliases='BHLHB3, DEC2, SHARP1, hDEC2', description='basic helix-loop-helix family member e41', other_designations='basic helix-loop-helix domain containing, class B, 3|basic helix-loop-helix family, member e41|differentially expressed in chondrocytes protein 2|enhancer-of-split and hairy-related protein 1', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26120026, end_position_on_the_genomic_accession=26125070, orientation='minus', exon_count=5, OMIM=606200, no_hdr0=''),
  79366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79366, CurrentID=0, Status='live', Symbol='HMGN5', Aliases='NBP-45, NSBP1', description='high mobility group nucleosome binding domain 5', other_designations='high-mobility group nucleosome binding domain 5|nucleosomal binding protein 1|nucleosome-binding protein 1', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81113701, end_position_on_the_genomic_accession=81201942, orientation='minus', exon_count=7, OMIM=300385, no_hdr0=''),
  341511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341511, CurrentID=0, Status='live', Symbol='RPL23AP14', Aliases='RPL23A_29_1192', description='ribosomal protein L23a pseudogene 14', other_designations='hCG1642354', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2781894, end_position_on_the_genomic_accession=2782420, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79368, CurrentID=0, Status='live', Symbol='FCRL2', Aliases='CD307b, FCRH2, IFGP4, IRTA4, SPAP1, SPAP1A, SPAP1B, SPAP1C', description='Fc receptor-like 2', other_designations='IFGP family protein 4|SH2 domain containing phosphatase anchor protein 1|fc receptor homolog 2|immune receptor translocation-associated protein 4|immunoglobulin receptor translocation-associated protein 4|immunoglobulin superfamily Fc receptor, gp42', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157745733, end_position_on_the_genomic_accession=157777439, orientation='minus', exon_count=12, OMIM=606509, no_hdr0=''),
  79369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79369, CurrentID=0, Status='live', Symbol='B3GNT4', Aliases='B3GN-T4, beta3Gn-T4', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 4', other_designations='BGnT-4|beta-1,3-Gn-T4|beta-1,3-N-acetylglucosaminyltransferase 4|beta-1,3-N-acetylglucosaminyltransferase bGn-T4', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122203681, end_position_on_the_genomic_accession=122208135, orientation='plus', exon_count=3, OMIM=605864, no_hdr0=''),
  79370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79370, CurrentID=0, Status='live', Symbol='BCL2L14', Aliases='BCLG', description='BCL2-like 14 (apoptosis facilitator)', other_designations='apoptosis regulator BCL-G|bcl2-L-14', map_location='12p13-p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12049880, end_position_on_the_genomic_accession=12099695, orientation='plus', exon_count=11, OMIM=606126, no_hdr0=''),
  100873739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873739, CurrentID=0, Status='live', Symbol='RNA5SP497', Aliases='RN5S497', description='RNA, 5S ribosomal pseudogene 497', other_designations='RNA, 5S ribosomal 497', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32635102, end_position_on_the_genomic_accession=32635198, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873740, CurrentID=0, Status='live', Symbol='RNU6-31P', Aliases='RNU6-31, RNU6-6', description='RNA, U6 small nuclear 31, pseudogene', other_designations='RNA, U6 small nuclear 6', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200830009, end_position_on_the_genomic_accession=200830116, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873741, CurrentID=0, Status='live', Symbol='RNU6-26P', Aliases='RNU6-26, RNU6-8', description='RNA, U6 small nuclear 26, pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98804978, end_position_on_the_genomic_accession=98805084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873742, CurrentID=0, Status='live', Symbol='GOLGA2P4', Aliases='', description='golgin A2 pseudogene 4', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42784459, end_position_on_the_genomic_accession=42784566, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873743, CurrentID=0, Status='live', Symbol='NCSTNP1', Aliases='', description='nicastrin pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27492118, end_position_on_the_genomic_accession=27492261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873744, CurrentID=0, Status='live', Symbol='PFN1P8', Aliases='', description='profilin 1 pseudogene 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146957261, end_position_on_the_genomic_accession=146957668, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873745, CurrentID=0, Status='live', Symbol='RNU6-10P', Aliases='RNU6-10', description='RNA, U6 small nuclear 10, pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92701708, end_position_on_the_genomic_accession=92701814, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873746, CurrentID=0, Status='live', Symbol='RNU6-14P', Aliases='RNU6-14', description='RNA, U6 small nuclear 14, pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15776181, end_position_on_the_genomic_accession=15776286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873747, CurrentID=0, Status='live', Symbol='RNU6-30P', Aliases='RNU6-30', description='RNA, U6 small nuclear 30, pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101634006, end_position_on_the_genomic_accession=101634112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873748, CurrentID=0, Status='live', Symbol='RNU6-33P', Aliases='RNU6-33', description='RNA, U6 small nuclear 33, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88684848, end_position_on_the_genomic_accession=88684954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873749, CurrentID=0, Status='live', Symbol='RNU6-34P', Aliases='RNU6-34', description='RNA, U6 small nuclear 34, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=96152299, end_position_on_the_genomic_accession=96152403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873750, CurrentID=0, Status='live', Symbol='RNU6-35P', Aliases='RNU6-35', description='RNA, U6 small nuclear 35, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109992325, end_position_on_the_genomic_accession=109992431, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873751, CurrentID=0, Status='live', Symbol='PHBP14', Aliases='', description='prohibitin pseudogene 14', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=167648430, end_position_on_the_genomic_accession=167648796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873752, CurrentID=0, Status='live', Symbol='PHBP15', Aliases='', description='prohibitin pseudogene 15', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74741457, end_position_on_the_genomic_accession=74742121, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873753, CurrentID=0, Status='live', Symbol='RNU6-16P', Aliases='RNU6-16', description='RNA, U6 small nuclear 16, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88612805, end_position_on_the_genomic_accession=88612911, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873754, CurrentID=0, Status='live', Symbol='RNU6-19P', Aliases='RNU6-19', description='RNA, U6 small nuclear 19, pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65553081, end_position_on_the_genomic_accession=65553187, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873755, CurrentID=0, Status='live', Symbol='RNU6-23P', Aliases='RNU6-23', description='RNA, U6 small nuclear 23, pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70445296, end_position_on_the_genomic_accession=70445399, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873756, CurrentID=0, Status='live', Symbol='RNU6-28P', Aliases='RNU6-28', description='RNA, U6 small nuclear 28, pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31479385, end_position_on_the_genomic_accession=31479488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873757, CurrentID=0, Status='live', Symbol='RNU6-36P', Aliases='RNU6-36', description='RNA, U6 small nuclear 36, pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97721716, end_position_on_the_genomic_accession=97721822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873758, CurrentID=0, Status='live', Symbol='RNU6-39P', Aliases='RNU6-39', description='RNA, U6 small nuclear 39, pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68858934, end_position_on_the_genomic_accession=68859040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873759, CurrentID=0, Status='live', Symbol='RNU6-45P', Aliases='RNU6-45', description='RNA, U6 small nuclear 45, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63970470, end_position_on_the_genomic_accession=63970576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873760, CurrentID=0, Status='live', Symbol='RNU6-46P', Aliases='RNU6-46', description='RNA, U6 small nuclear 46, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67895631, end_position_on_the_genomic_accession=67895737, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873761, CurrentID=0, Status='live', Symbol='RNU6-52P', Aliases='RNU6-52', description='RNA, U6 small nuclear 52, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19142982, end_position_on_the_genomic_accession=19143077, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873762, CurrentID=0, Status='live', Symbol='RNU6-53P', Aliases='RNU6-53', description='RNA, U6 small nuclear 53, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28703702, end_position_on_the_genomic_accession=28703807, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873763, CurrentID=0, Status='live', Symbol='RNU6-55P', Aliases='RNU6-55', description='RNA, U6 small nuclear 55, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18869185, end_position_on_the_genomic_accession=18869291, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873764, CurrentID=0, Status='live', Symbol='RNU6-56P', Aliases='RNU6-56', description='RNA, U6 small nuclear 56, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38779424, end_position_on_the_genomic_accession=38779516, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873765, CurrentID=0, Status='live', Symbol='RNU6-57P', Aliases='RNU6-57', description='RNA, U6 small nuclear 57, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41354557, end_position_on_the_genomic_accession=41354659, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873766, CurrentID=0, Status='live', Symbol='RNU6-58P', Aliases='RNU6-58', description='RNA, U6 small nuclear 58, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23217432, end_position_on_the_genomic_accession=23217534, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873767, CurrentID=0, Status='live', Symbol='RNU6-59P', Aliases='RNU6-59', description='RNA, U6 small nuclear 59, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21551304, end_position_on_the_genomic_accession=21551410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79400, CurrentID=0, Status='live', Symbol='NOX5', Aliases='', description='NADPH oxidase, EF-hand calcium binding domain 5', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68930500, end_position_on_the_genomic_accession=69057161, orientation='plus', exon_count=18, OMIM=606572, no_hdr0=''),
  106903081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106903081, CurrentID=0, Status='live', Symbol='DUSP8P5', Aliases='', description='dual specificity phosphatase 8 pseudogene 5', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873770, CurrentID=0, Status='live', Symbol='RNU6-66P', Aliases='RNU6-66', description='RNA, U6 small nuclear 66, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73112915, end_position_on_the_genomic_accession=73113018, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873771, CurrentID=0, Status='live', Symbol='RNU6-67P', Aliases='RNU6-67', description='RNA, U6 small nuclear 67, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=83298071, end_position_on_the_genomic_accession=83298167, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873772, CurrentID=0, Status='live', Symbol='RNU6-68P', Aliases='RNU6-68', description='RNA, U6 small nuclear 68, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46410454, end_position_on_the_genomic_accession=46410557, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873773, CurrentID=0, Status='live', Symbol='RNU6-69P', Aliases='RNU6-69', description='RNA, U6 small nuclear 69, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45137862, end_position_on_the_genomic_accession=45137965, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873774, CurrentID=0, Status='live', Symbol='RNU6-71P', Aliases='RNU6-71', description='RNA, U6 small nuclear 71, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36267168, end_position_on_the_genomic_accession=36267274, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873775, CurrentID=0, Status='live', Symbol='RNU6-72P', Aliases='RNU6-72', description='RNA, U6 small nuclear 72, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85584748, end_position_on_the_genomic_accession=85584853, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873776, CurrentID=0, Status='live', Symbol='RNU6-75P', Aliases='RNU6-75', description='RNA, U6 small nuclear 75, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90781766, end_position_on_the_genomic_accession=90781869, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873777, CurrentID=0, Status='live', Symbol='RNU6-76P', Aliases='RNU6-76', description='RNA, U6 small nuclear 76, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18880054, end_position_on_the_genomic_accession=18880157, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873778, CurrentID=0, Status='live', Symbol='RNU6-78P', Aliases='RNU6-78', description='RNA, U6 small nuclear 78, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25517119, end_position_on_the_genomic_accession=25517225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79411, CurrentID=0, Status='live', Symbol='GLB1L', Aliases='', description='galactosidase beta 1 like', other_designations='galactosidase beta 1-like|galactosidase, beta 1-like', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219236598, end_position_on_the_genomic_accession=219245429, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  79412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79412, CurrentID=0, Status='live', Symbol='KREMEN2', Aliases='KRM2', description='kringle containing transmembrane protein 2', other_designations='dickkopf receptor 2|kringle domain-containing transmembrane protein 2|kringle-containing protein marking the eye and the nose', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2964216, end_position_on_the_genomic_accession=2968383, orientation='plus', exon_count=9, OMIM=609899, no_hdr0=''),
  79413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79413, CurrentID=0, Status='live', Symbol='ZBED2', Aliases='', description='zinc finger, BED-type containing 2', other_designations='zinc finger, BED domain containing 2', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111592900, end_position_on_the_genomic_accession=111595335, orientation='minus', exon_count=2, OMIM=615246, no_hdr0=''),
  79414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79414, CurrentID=0, Status='live', Symbol='LRFN3', Aliases='FIGLER1, SALM4', description='leucine rich repeat and fibronectin type III domain containing 3', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 1|synaptic adhesion-like molecule 4', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35937120, end_position_on_the_genomic_accession=35945195, orientation='plus', exon_count=3, OMIM=612809, no_hdr0=''),
  79415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79415, CurrentID=0, Status='live', Symbol='C17orf62', Aliases='', description='chromosome 17 open reading frame 62', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82442586, end_position_on_the_genomic_accession=82450831, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100873784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873784, CurrentID=0, Status='live', Symbol='PFN1P11', Aliases='', description='profilin 1 pseudogene 11', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102845142, end_position_on_the_genomic_accession=102845473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873785, CurrentID=0, Status='live', Symbol='SNX2P2', Aliases='', description='sorting nexin 2 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31941261, end_position_on_the_genomic_accession=31941580, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873786, CurrentID=0, Status='live', Symbol='TMSB10P2', Aliases='', description='thymosin beta 10 pseudogene 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18341216, end_position_on_the_genomic_accession=18341343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873787, CurrentID=0, Status='live', Symbol='GPM6BP3', Aliases='', description='glycoprotein M6B pseudogene 3', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16837513, end_position_on_the_genomic_accession=16837665, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873788, CurrentID=0, Status='live', Symbol='SNX18P27', Aliases='', description='sorting nexin 18 pseudogene 27', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43513643, end_position_on_the_genomic_accession=43514421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873789, CurrentID=0, Status='live', Symbol='AK3P6', Aliases='', description='adenylate kinase 3 pseudogene 6', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128942099, end_position_on_the_genomic_accession=128945293, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873790, CurrentID=0, Status='live', Symbol='FHP1', Aliases='', description='fumarate hydratase pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42255873, end_position_on_the_genomic_accession=42256575, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341567, CurrentID=0, Status='live', Symbol='H1FNT', Aliases='H1.7, H1T2', description='H1 histone family member N, testis specific', other_designations='H1 histone family, member N, testis-specific|HANP1|haploid germ cell-specific nuclear protein 1|histone H1t2', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48328980, end_position_on_the_genomic_accession=48330279, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  341568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341568, CurrentID=0, Status='live', Symbol='OR8S1', Aliases='', description='olfactory receptor family 8 subfamily S member 1', other_designations='olfactory receptor, family 8, subfamily S, member 1', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48525632, end_position_on_the_genomic_accession=48528103, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873793, CurrentID=0, Status='live', Symbol='PHF2P2', Aliases='', description='PHD finger protein 2 pseudogene 2', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18946724, end_position_on_the_genomic_accession=18952892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873794, CurrentID=0, Status='live', Symbol='ASNSP5', Aliases='', description='asparagine synthetase pseudogene 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873795, CurrentID=0, Status='live', Symbol='WDR95P', Aliases='', description='WD repeat domain 95, pseudogene', other_designations='WD40 repeat domain 95, pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31077254, end_position_on_the_genomic_accession=31115699, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873796, CurrentID=0, Status='live', Symbol='RPL35P9', Aliases='', description='ribosomal protein L35 pseudogene 9', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106425757, end_position_on_the_genomic_accession=106426124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873797, CurrentID=0, Status='live', Symbol='RPSAP64', Aliases='', description='ribosomal protein SA pseudogene 64', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38894914, end_position_on_the_genomic_accession=38895195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873798, CurrentID=0, Status='live', Symbol='RPS21P8', Aliases='', description='ribosomal protein S21 pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26977548, end_position_on_the_genomic_accession=26977739, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873799, CurrentID=0, Status='live', Symbol='ZNF90P3', Aliases='', description='zinc finger protein 90 pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63086744, end_position_on_the_genomic_accession=63088278, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873800, CurrentID=0, Status='live', Symbol='RNY4P22', Aliases='', description='RNA, Ro-associated Y4 pseudogene 22', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20192782, end_position_on_the_genomic_accession=20192874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873801, CurrentID=0, Status='live', Symbol='RNY4P23', Aliases='', description='RNA, Ro-associated Y4 pseudogene 23', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70396279, end_position_on_the_genomic_accession=70396375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873802, CurrentID=0, Status='live', Symbol='RNY1P1', Aliases='', description='RNA, Ro-associated Y1 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27402476, end_position_on_the_genomic_accession=27402587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873803, CurrentID=0, Status='live', Symbol='RNY1P3', Aliases='', description='RNA, Ro-associated Y1 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25950704, end_position_on_the_genomic_accession=25950821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873804, CurrentID=0, Status='live', Symbol='RNY1P4', Aliases='', description='RNA, Ro-associated Y1 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32763438, end_position_on_the_genomic_accession=32763552, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873805, CurrentID=0, Status='live', Symbol='RNY1P6', Aliases='', description='RNA, Ro-associated Y1 pseudogene 6', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51586424, end_position_on_the_genomic_accession=51586532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873806, CurrentID=0, Status='live', Symbol='RNY1P7', Aliases='', description='RNA, Ro-associated Y1 pseudogene 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24703556, end_position_on_the_genomic_accession=24703665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873807, CurrentID=0, Status='live', Symbol='RNY1P8', Aliases='', description='RNA, Ro-associated Y1 pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73227147, end_position_on_the_genomic_accession=73227260, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873808, CurrentID=0, Status='live', Symbol='RNY3P4', Aliases='', description='RNA, Ro-associated Y3 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23152586, end_position_on_the_genomic_accession=23152686, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79441, CurrentID=0, Status='live', Symbol='HAUS3', Aliases='C4orf15, IT1, dgt3', description='HAUS augmin like complex subunit 3', other_designations='HAUS augmin-like complex, subunit 3', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2231836, end_position_on_the_genomic_accession=2242164, orientation='minus', exon_count=6, OMIM=613430, no_hdr0=''),
  79442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79442, CurrentID=0, Status='live', Symbol='LRRC2', Aliases='', description='leucine rich repeat containing 2', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46515385, end_position_on_the_genomic_accession=46580060, orientation='minus', exon_count=10, OMIM=607180, no_hdr0=''),
  79443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79443, CurrentID=0, Status='live', Symbol='FYCO1', Aliases='CATC2, CTRCT18, RUFY3, ZFYVE7', description='FYVE and coiled-coil domain containing 1', other_designations='RUN and FYVE domain containing 3|zinc finger FYVE domain-containing protein 7', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45917899, end_position_on_the_genomic_accession=45995824, orientation='minus', exon_count=21, OMIM=607182, no_hdr0=''),
  79444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79444, CurrentID=0, Status='live', Symbol='BIRC7', Aliases='KIAP, LIVIN, ML-IAP, MLIAP, RNF50', description='baculoviral IAP repeat containing 7', other_designations='RING finger protein 50|kidney inhibitor of apoptosis protein|livin inhibitor of apoptosis|melanoma inhibitor of apoptosis protein', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63235883, end_position_on_the_genomic_accession=63240507, orientation='plus', exon_count=7, OMIM=605737, no_hdr0=''),
  100873813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873813, CurrentID=0, Status='live', Symbol='HLFP1', Aliases='', description='hepatic leukemia factor pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132674885, end_position_on_the_genomic_accession=132675383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79446, CurrentID=0, Status='live', Symbol='WDR25', Aliases='C14orf67', description='WD repeat domain 25', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100376417, end_position_on_the_genomic_accession=100530303, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  79447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79447, CurrentID=0, Status='live', Symbol='PAGR1', Aliases='C16orf53, GAS, PA1', description='PAXIP1 associated glutamate-rich protein 1', other_designations='PAXIP1-associated protein 1|PTIP-associated 1 protein|PTIP-associated protein 1|glutamate-rich coactivator associated with SRC1|glutamate-rich coactivator interacting with SRC1/NCOA1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29816207, end_position_on_the_genomic_accession=29822495, orientation='plus', exon_count=3, OMIM=612033, no_hdr0=''),
  100873816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873816, CurrentID=0, Status='live', Symbol='RNU1-20P', Aliases='', description='RNA, U1 small nuclear 20, pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192729555, end_position_on_the_genomic_accession=192729716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873817, CurrentID=0, Status='live', Symbol='RPL7AP73', Aliases='', description='ribosomal protein L7a pseudogene 73', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22803293, end_position_on_the_genomic_accession=22803598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873818, CurrentID=0, Status='live', Symbol='RPS3AP52', Aliases='', description='ribosomal protein S3a pseudogene 52', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68634184, end_position_on_the_genomic_accession=68634486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873819, CurrentID=0, Status='live', Symbol='RNY4P26', Aliases='', description='RNA, Ro-associated Y4 pseudogene 26', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94047038, end_position_on_the_genomic_accession=94047125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873820, CurrentID=0, Status='live', Symbol='RNY3P10', Aliases='', description='RNA, Ro-associated Y3 pseudogene 10', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69802940, end_position_on_the_genomic_accession=69803041, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873821, CurrentID=0, Status='live', Symbol='RNY4P27', Aliases='', description='RNA, Ro-associated Y4 pseudogene 27', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95336106, end_position_on_the_genomic_accession=95336201, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873822, CurrentID=0, Status='live', Symbol='RNY4P28', Aliases='', description='RNA, Ro-associated Y4 pseudogene 28', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60187738, end_position_on_the_genomic_accession=60187830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873823, CurrentID=0, Status='live', Symbol='RNY4P29', Aliases='', description='RNA, Ro-associated Y4 pseudogene 29', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58527655, end_position_on_the_genomic_accession=58527751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873824, CurrentID=0, Status='live', Symbol='RNU2-7P', Aliases='', description='RNA, U2 small nuclear 7, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20612161, end_position_on_the_genomic_accession=20612338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873825, CurrentID=0, Status='live', Symbol='RNU2-6P', Aliases='', description='RNA, U2 small nuclear 6, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46374401, end_position_on_the_genomic_accession=46374590, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873826, CurrentID=0, Status='live', Symbol='RNU5F-2P', Aliases='', description='RNA, U5F small nuclear 2, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179576268, end_position_on_the_genomic_accession=179576349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873827, CurrentID=0, Status='live', Symbol='RNU5E-2P', Aliases='', description='RNA, U5E small nuclear 2, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235863388, end_position_on_the_genomic_accession=235863437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341604, CurrentID=0, Status='live', Symbol='RPL35P7', Aliases='RPL35_2_1335', description='ribosomal protein L35 pseudogene 7', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106433339, end_position_on_the_genomic_accession=106441916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873829, CurrentID=0, Status='live', Symbol='RNU5E-4P', Aliases='', description='RNA, U5E small nuclear 4, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11909808, end_position_on_the_genomic_accession=11909927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873830, CurrentID=0, Status='live', Symbol='RNU5A-2P', Aliases='', description='RNA, U5A small nuclear 2, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81334303, end_position_on_the_genomic_accession=81334418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873831, CurrentID=0, Status='live', Symbol='RNU5A-3P', Aliases='', description='RNA, U5A small nuclear 3, pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30643689, end_position_on_the_genomic_accession=30643802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873832, CurrentID=0, Status='live', Symbol='RNU5E-6P', Aliases='', description='RNA, U5E small nuclear 6, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44819883, end_position_on_the_genomic_accession=44819997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79465, CurrentID=0, Status='live', Symbol='ULBP3', Aliases='N2DL-3, NKG2DL3, RAET1N', description='UL16 binding protein 3', other_designations='ALCAN-gamma|UL16-binding protein 3|retinoic acid early transcript 1N', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150063264, end_position_on_the_genomic_accession=150069213, orientation='minus', exon_count=5, OMIM=605699, no_hdr0=''),
  100873834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873834, CurrentID=0, Status='live', Symbol='RNU5F-7P', Aliases='', description='RNA, U5F small nuclear 7, pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15916189, end_position_on_the_genomic_accession=15916303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873835, CurrentID=0, Status='live', Symbol='RNU5A-5P', Aliases='', description='RNA, U5A small nuclear 5, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231670635, end_position_on_the_genomic_accession=231670750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873836, CurrentID=0, Status='live', Symbol='RNU5E-7P', Aliases='', description='RNA, U5E small nuclear 7, pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15864935, end_position_on_the_genomic_accession=15865051, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79469, CurrentID=0, Status='live', Symbol='DLEU2L', Aliases='BCMSUNL', description='deleted in lymphocytic leukemia 2-like', other_designations='BCMS upstream neighbor-like', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63548980, end_position_on_the_genomic_accession=63550636, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79470, CurrentID=0, Status='live', Symbol='OR51J1', Aliases='OR51J1P, OR51J2', description='olfactory receptor family 51 subfamily J member 1 (gene/pseudogene)', other_designations='olfactory receptor, family 51, subfamily J, member 1 pseudogene|olfactory receptor, family 51, subfamily J, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5402497, end_position_on_the_genomic_accession=5403671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873839, CurrentID=0, Status='live', Symbol='RNU5F-8P', Aliases='', description='RNA, U5F small nuclear 8, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240653367, end_position_on_the_genomic_accession=240653477, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873840, CurrentID=0, Status='live', Symbol='RNU5E-9P', Aliases='', description='RNA, U5E small nuclear 9, pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178142548, end_position_on_the_genomic_accession=178142651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79473, CurrentID=0, Status='live', Symbol='OR52N1', Aliases='OR11-61', description='olfactory receptor family 52 subfamily N member 1', other_designations='olfactory receptor OR11-61|olfactory receptor, family 52, subfamily N, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5787854, end_position_on_the_genomic_accession=5788816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873842, CurrentID=0, Status='live', Symbol='RNU7-87P', Aliases='', description='RNA, U7 small nuclear 87 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66630715, end_position_on_the_genomic_accession=66630776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873843, CurrentID=0, Status='live', Symbol='RNU7-88P', Aliases='', description='RNA, U7 small nuclear 88 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59480070, end_position_on_the_genomic_accession=59480132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873844, CurrentID=0, Status='live', Symbol='RNU7-89P', Aliases='', description='RNA, U7 small nuclear 89 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73168089, end_position_on_the_genomic_accession=73168150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873845, CurrentID=0, Status='live', Symbol='RN7SKP1', Aliases='', description='RNA, 7SK small nuclear pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37166370, end_position_on_the_genomic_accession=37166658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873846, CurrentID=0, Status='live', Symbol='RN7SKP2', Aliases='', description='RNA, 7SK small nuclear pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40407122, end_position_on_the_genomic_accession=40407440, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873847, CurrentID=0, Status='live', Symbol='RN7SKP3', Aliases='', description='RNA, 7SK small nuclear pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45060729, end_position_on_the_genomic_accession=45061086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873848, CurrentID=0, Status='live', Symbol='RN7SKP4', Aliases='', description='RNA, 7SK small nuclear pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45123093, end_position_on_the_genomic_accession=45123414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873849, CurrentID=0, Status='live', Symbol='RN7SKP5', Aliases='', description='RNA, 7SK small nuclear pseudogene 5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46249942, end_position_on_the_genomic_accession=46250213, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79482, CurrentID=0, Status='live', Symbol='OR5AL1', Aliases='OR5AL1P', description='olfactory receptor family 5 subfamily AL member 1 (gene/pseudogene)', other_designations='Olfactory receptor 5AL1|olfactory receptor, family 5, subfamily AL, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56412697, end_position_on_the_genomic_accession=56413681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873851, CurrentID=0, Status='live', Symbol='RN7SKP7', Aliases='', description='RNA, 7SK small nuclear pseudogene 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96949249, end_position_on_the_genomic_accession=96949533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873852, CurrentID=0, Status='live', Symbol='RN7SKP8', Aliases='', description='RNA, 7SK small nuclear pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98203835, end_position_on_the_genomic_accession=98204192, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873853, CurrentID=0, Status='live', Symbol='RN7SKP9', Aliases='', description='RNA, 7SK small nuclear pseudogene 9', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99205708, end_position_on_the_genomic_accession=99206006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873854, CurrentID=0, Status='live', Symbol='MOB1AP2', Aliases='MOBKL1BP2', description='MOB kinase activator 1A pseudogene 2', other_designations='MOB1 Mps One Binder homolog A pseudogene 2|MOB1, Mps One Binder kinase activator-like 1B pseudogene 2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37291205, end_position_on_the_genomic_accession=37291557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873855, CurrentID=0, Status='live', Symbol='RPL36P19', Aliases='', description='ribosomal protein L36 pseudogene 19', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43394527, end_position_on_the_genomic_accession=43394829, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873856, CurrentID=0, Status='live', Symbol='SNORD116-30', Aliases='HBII-85-30', description='small nucleolar RNA, C/D box 116-30', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25108268, end_position_on_the_genomic_accession=25108352, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873857, CurrentID=0, Status='live', Symbol='SNORD115-46', Aliases='HBII-52-46', description='small nucleolar RNA, C/D box 115-46', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25266591, end_position_on_the_genomic_accession=25266661, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79490, CurrentID=0, Status='live', Symbol='OR10G6', Aliases='OR10G6PQ, OR10G6', description='olfactory receptor family 10 subfamily G member 6', other_designations='olfactory receptor OR11-280|olfactory receptor, family 10, subfamily G, member 6 pseudogene|seven transmembrane helix receptor', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123994166, end_position_on_the_genomic_accession=123994924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873859, CurrentID=0, Status='live', Symbol='RPL18AP17', Aliases='', description='ribosomal protein L18a pseudogene 17', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72460126, end_position_on_the_genomic_accession=72460327, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873860, CurrentID=0, Status='live', Symbol='RNU1-16P', Aliases='U1.64', description='RNA, U1 small nuclear 16, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113478915, end_position_on_the_genomic_accession=113479078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873861, CurrentID=0, Status='live', Symbol='RNU1-19P', Aliases='', description='RNA, U1 small nuclear 19, pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86889569, end_position_on_the_genomic_accession=86889733, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873862, CurrentID=0, Status='live', Symbol='RNU1-21P', Aliases='', description='RNA, U1 small nuclear 21, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123763927, end_position_on_the_genomic_accession=123764089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873863, CurrentID=0, Status='live', Symbol='RNU1-22P', Aliases='', description='RNA, U1 small nuclear 22, pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3086143, end_position_on_the_genomic_accession=3086300, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341640, CurrentID=0, Status='live', Symbol='FREM2', Aliases='', description='FRAS1 related extracellular matrix protein 2', other_designations='ECM3 homolog', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38687036, end_position_on_the_genomic_accession=38887131, orientation='plus', exon_count=26, OMIM=608945, no_hdr0=''),
  100873865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873865, CurrentID=0, Status='live', Symbol='RNU1-24P', Aliases='U1.41', description='RNA, U1 small nuclear 24, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101778884, end_position_on_the_genomic_accession=101779048, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873866, CurrentID=0, Status='live', Symbol='RNU5F-3P', Aliases='', description='RNA, U5F small nuclear 3, pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6017723, end_position_on_the_genomic_accession=6017839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873867, CurrentID=0, Status='live', Symbol='RNU5F-4P', Aliases='', description='RNA, U5F small nuclear 4, pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31594093, end_position_on_the_genomic_accession=31594206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873868, CurrentID=0, Status='live', Symbol='RNU5F-5P', Aliases='', description='RNA, U5F small nuclear 5, pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29910788, end_position_on_the_genomic_accession=29910840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79501, CurrentID=0, Status='live', Symbol='OR4F5', Aliases='', description='olfactory receptor family 4 subfamily F member 5', other_designations='olfactory receptor, family 4, subfamily F, member 5', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69091, end_position_on_the_genomic_accession=70008, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873870, CurrentID=0, Status='live', Symbol='RNU5A-4P', Aliases='', description='RNA, U5A small nuclear 4, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33829539, end_position_on_the_genomic_accession=33829649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873871, CurrentID=0, Status='live', Symbol='RNU5A-6P', Aliases='', description='RNA, U5A small nuclear 6, pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24124170, end_position_on_the_genomic_accession=24124285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873872, CurrentID=0, Status='live', Symbol='RNU5A-7P', Aliases='', description='RNA, U5A small nuclear 7, pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64228560, end_position_on_the_genomic_accession=64228675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873873, CurrentID=0, Status='live', Symbol='RN7SKP10', Aliases='', description='RNA, 7SK small nuclear pseudogene 10', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109915096, end_position_on_the_genomic_accession=109915420, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873874, CurrentID=0, Status='live', Symbol='MOB1AP1', Aliases='MOBKL1BP1', description='MOB kinase activator 1A pseudogene 1', other_designations='MOB1 Mps One Binder homolog A pseudogene 1|MOB1, Mps One Binder kinase activator-like 1B pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85968917, end_position_on_the_genomic_accession=85969531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873875, CurrentID=0, Status='live', Symbol='PSPC1P2', Aliases='', description='paraspeckle component 1 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24566843, end_position_on_the_genomic_accession=24570728, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873876, CurrentID=0, Status='live', Symbol='TRV-CAC10-1', Aliases='TRNAV38', description='transfer RNA-Val (CAC) 10-1', other_designations='tRNA-Val-CAC-10-1|transfer RNA valine 38 (anticodon CAC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873877, CurrentID=0, Status='live', Symbol='TMSB4XP3', Aliases='', description='thymosin beta 4, X-linked pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29625864, end_position_on_the_genomic_accession=29626001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873878, CurrentID=0, Status='live', Symbol='MORF4L1P6', Aliases='', description='mortality factor 4 like 1 pseudogene 6', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73474563, end_position_on_the_genomic_accession=73474830, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873879, CurrentID=0, Status='live', Symbol='HMGB3P9', Aliases='', description='high mobility group box 3 pseudogene 9', other_designations='high-mobility group box 3 pseudogene 9', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92646848, end_position_on_the_genomic_accession=92647730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873880, CurrentID=0, Status='live', Symbol='TMEM97P1', Aliases='', description='transmembrane protein 97 pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43783079, end_position_on_the_genomic_accession=43783561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79513, CurrentID=0, Status='live', Symbol='OR10R1P', Aliases='OR1-10', description='olfactory receptor family 10 subfamily R member 1 pseudogene', other_designations='olfactory receptor OR1-10 pseudogene|olfactory receptor, family 10, subfamily R, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158514785, end_position_on_the_genomic_accession=158515925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873882, CurrentID=0, Status='live', Symbol='ANKRD26P4', Aliases='', description='ankyrin repeat domain 26 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31311285, end_position_on_the_genomic_accession=31312928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873883, CurrentID=0, Status='live', Symbol='FGFR1OP2P1', Aliases='', description='FGFR1 oncogene partner 2 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26905488, end_position_on_the_genomic_accession=26905776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873884, CurrentID=0, Status='live', Symbol='SNX3P1X', Aliases='', description='sorting nexin 3 pseudogene 1, X-linked', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=92515326, end_position_on_the_genomic_accession=92515497, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873885, CurrentID=0, Status='live', Symbol='SNX3P1Y', Aliases='', description='sorting nexin 3 pseudogene 1, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5634484, end_position_on_the_genomic_accession=5634655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  399819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399819, CurrentID=0, Status='live', Symbol='FLJ40536', Aliases='', description='FLJ40536 protein', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873887, CurrentID=0, Status='live', Symbol='PGAM1P3', Aliases='', description='phosphoglycerate mutase 1 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40322718, end_position_on_the_genomic_accession=40323271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873888, CurrentID=0, Status='live', Symbol='RNU5E-10P', Aliases='', description='RNA, U5E small nuclear 10, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47576471, end_position_on_the_genomic_accession=47576588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873889, CurrentID=0, Status='live', Symbol='RNU4ATAC2P', Aliases='', description='RNA, U4atac small nuclear 2, pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159318042, end_position_on_the_genomic_accession=159318167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873890, CurrentID=0, Status='live', Symbol='HMGB3P4', Aliases='', description='high mobility group box 3 pseudogene 4', other_designations='high-mobility group box 3 pseudogene 4', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99372173, end_position_on_the_genomic_accession=99372686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873891, CurrentID=0, Status='live', Symbol='HMGB1P23', Aliases='', description='high mobility group box 1 pseudogene 23', other_designations='high-mobility group box 1 pseudogene 23', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29218905, end_position_on_the_genomic_accession=29219472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873892, CurrentID=0, Status='live', Symbol='HMGB1P29', Aliases='', description='high mobility group box 1 pseudogene 29', other_designations='high-mobility group box 1 pseudogene 29', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124220579, end_position_on_the_genomic_accession=124221077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873893, CurrentID=0, Status='live', Symbol='HMGB1P30', Aliases='', description='high mobility group box 1 pseudogene 30', other_designations='high-mobility group box 1 pseudogene 30', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148824782, end_position_on_the_genomic_accession=148825387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873894, CurrentID=0, Status='live', Symbol='HMGB1P31', Aliases='', description='high mobility group box 1 pseudogene 31', other_designations='high-mobility group box 1 pseudogene 31', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54051334, end_position_on_the_genomic_accession=54051760, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  56945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56945, CurrentID=0, Status='live', Symbol='MRPS22', Aliases='C3orf5, COXPD5, GIBT, GK002, MRP-S22, RPMS22', description='mitochondrial ribosomal protein S22', other_designations='S22mt', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139343994, end_position_on_the_genomic_accession=139357129, orientation='plus', exon_count=9, OMIM=605810, no_hdr0=''),
  100873896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873896, CurrentID=0, Status='live', Symbol='HMGB1P41', Aliases='', description='high mobility group box 1 pseudogene 41', other_designations='high-mobility group box 1 pseudogene 41', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80812142, end_position_on_the_genomic_accession=80812581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873897, CurrentID=0, Status='live', Symbol='HMGB1P47', Aliases='', description='high mobility group box 1 pseudogene 47', other_designations='high-mobility group box 1 pseudogene 47', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51409617, end_position_on_the_genomic_accession=51410205, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873898, CurrentID=0, Status='live', Symbol='HMGB3P12', Aliases='', description='high mobility group box 3 pseudogene 12', other_designations='high-mobility group box 3 pseudogene 12', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26216219, end_position_on_the_genomic_accession=26217073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873899, CurrentID=0, Status='live', Symbol='HMGB3P14', Aliases='', description='high mobility group box 3 pseudogene 14', other_designations='high-mobility group box 3 pseudogene 14', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134170390, end_position_on_the_genomic_accession=134171099, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341676, CurrentID=0, Status='live', Symbol='NEK5', Aliases='', description='NIMA-related kinase 5', other_designations='NIMA (never in mitosis gene a)-related kinase 5|never in mitosis A-related kinase 5|nimA-related protein kinase 5', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52036411, end_position_on_the_genomic_accession=52130366, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  100873901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873901, CurrentID=0, Status='live', Symbol='HMGB3P21', Aliases='', description='high mobility group box 3 pseudogene 21', other_designations='high-mobility group box 3 pseudogene 21', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97135015, end_position_on_the_genomic_accession=97135561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873902, CurrentID=0, Status='live', Symbol='GK-AS1', Aliases='', description='GK antisense RNA 1', other_designations='GK antisense RNA 1 (non-protein coding)', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873903, CurrentID=0, Status='live', Symbol='HSPA9P2', Aliases='', description='heat shock protein family A (Hsp70) member 9 pseudogene 2', other_designations='heat shock 70kDa protein 9 pseudogene 2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=29256490, end_position_on_the_genomic_accession=29257074, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873904, CurrentID=0, Status='live', Symbol='MYB-AS1', Aliases='MYB-AS, MYBAS, NCRNA00209', description='MYB antisense RNA 1', other_designations='MYB antisense RNA (non-protein coding)|MYB antisense RNA 1 (non-protein coding)', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873905, CurrentID=0, Status='live', Symbol='HMGB1P33', Aliases='', description='high mobility group box 1 pseudogene 33', other_designations='high-mobility group box 1 pseudogene 33', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56323113, end_position_on_the_genomic_accession=56323604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873906, CurrentID=0, Status='live', Symbol='HMGB1P42', Aliases='', description='high mobility group box 1 pseudogene 42', other_designations='high-mobility group box 1 pseudogene 42', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120936028, end_position_on_the_genomic_accession=120936652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873907, CurrentID=0, Status='live', Symbol='ORAOV1P1', Aliases='', description='oral cancer overexpressed 1 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186170863, end_position_on_the_genomic_accession=186171257, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873908, CurrentID=0, Status='live', Symbol='HMGB3P25', Aliases='', description='high mobility group box 3 pseudogene 25', other_designations='high-mobility group box 3 pseudogene 25', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87480642, end_position_on_the_genomic_accession=87481337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79541, CurrentID=0, Status='live', Symbol='OR2A4', Aliases='OR2A10', description='olfactory receptor family 2 subfamily A member 4', other_designations='olfactory receptor 2A10|olfactory receptor OR6-37|olfactory receptor, family 2, subfamily A, member 10|olfactory receptor, family 2, subfamily A, member 4', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131699644, end_position_on_the_genomic_accession=131701401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873910, CurrentID=0, Status='live', Symbol='DUTP4', Aliases='', description='deoxyuridine triphosphatase pseudogene 4', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149817172, end_position_on_the_genomic_accession=149817648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873911, CurrentID=0, Status='live', Symbol='DUTP5', Aliases='', description='deoxyuridine triphosphatase pseudogene 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85426654, end_position_on_the_genomic_accession=85427138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79544, CurrentID=0, Status='live', Symbol='OR4K1', Aliases='OR14-19', description='olfactory receptor family 4 subfamily K member 1', other_designations='olfactory receptor OR14-19|olfactory receptor, family 4, subfamily K, member 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19920253, end_position_on_the_genomic_accession=19936736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873913, CurrentID=0, Status='live', Symbol='DUTP7', Aliases='', description='deoxyuridine triphosphatase pseudogene 7', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=51865053, end_position_on_the_genomic_accession=51865535, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873914, CurrentID=0, Status='live', Symbol='DUTP8', Aliases='', description='deoxyuridine triphosphatase pseudogene 8', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98048555, end_position_on_the_genomic_accession=98049040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79547, CurrentID=0, Status='live', Symbol='OR2AL1P', Aliases='', description='olfactory receptor family 2 subfamily AL member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AL, member 1 pseudogene', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105194100, end_position_on_the_genomic_accession=105195036, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873916, CurrentID=0, Status='live', Symbol='DMD-AS3', Aliases='', description='DMD antisense RNA 3', other_designations='DMD antisense RNA 3 (non-protein coding)', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873917, CurrentID=0, Status='live', Symbol='LPP-AS1', Aliases='', description='LPP antisense RNA 1', other_designations='LPP antisense RNA 1 (non-protein coding)', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188562237, end_position_on_the_genomic_accession=188568666, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873918, CurrentID=0, Status='live', Symbol='MME-AS1', Aliases='', description='MME antisense RNA 1', other_designations='MME antisense RNA 1 (non-protein coding)', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873919, CurrentID=0, Status='live', Symbol='NDP-AS1', Aliases='', description='NDP antisense RNA 1', other_designations='NDP antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873920, CurrentID=0, Status='live', Symbol='NHS-AS1', Aliases='', description='NHS antisense RNA 1', other_designations='NHS antisense RNA 1 (non-protein coding)', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17552349, end_position_on_the_genomic_accession=17559127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873921, CurrentID=0, Status='live', Symbol='SYP-AS1', Aliases='', description='SYP antisense RNA 1', other_designations='SYP antisense RNA 1 (non-protein coding)', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49198839, end_position_on_the_genomic_accession=49202454, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873922, CurrentID=0, Status='live', Symbol='ZFX-AS1', Aliases='', description='ZFX antisense RNA 1', other_designations='ZFX antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24146225, end_position_on_the_genomic_accession=24149654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873923, CurrentID=0, Status='live', Symbol='HYI-AS1', Aliases='', description='HYI antisense RNA 1', other_designations='HYI antisense RNA 1 (non-protein coding)', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873924, CurrentID=0, Status='live', Symbol='HLX-AS1', Aliases='', description='HLX antisense RNA 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220832763, end_position_on_the_genomic_accession=220880140, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373302, CurrentID=0, Status='live', Symbol='LOC105373302', Aliases='', description='uncharacterized LOC105373302', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103591364, end_position_on_the_genomic_accession=103629550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873926, CurrentID=0, Status='live', Symbol='TAB3-AS2', Aliases='', description='TAB3 antisense RNA 2', other_designations='TAB3 antisense RNA 2 (non-protein coding)', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873928, CurrentID=0, Status='live', Symbol='CASK-AS1', Aliases='', description='CASK antisense RNA 1', other_designations='CASK antisense RNA 1 (non-protein coding)', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873929, CurrentID=0, Status='live', Symbol='VCAN-AS1', Aliases='', description='VCAN antisense RNA 1', other_designations='VCAN antisense RNA 1 (non-protein coding)', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873930, CurrentID=0, Status='live', Symbol='DLG3-AS1', Aliases='', description='DLG3 antisense RNA 1', other_designations='DLG3 antisense RNA 1 (non-protein coding)', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70452956, end_position_on_the_genomic_accession=70455994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873931, CurrentID=0, Status='live', Symbol='DLX6-AS2', Aliases='', description='DLX6 antisense RNA 2', other_designations='DLX6 antisense RNA 2 (non-protein coding)', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873932, CurrentID=0, Status='live', Symbol='DPYD-AS1', Aliases='', description='DPYD antisense RNA 1', other_designations='DPYD antisense RNA 1 (non-protein coding)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97095923, end_position_on_the_genomic_accession=97322955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100873933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873933, CurrentID=0, Status='live', Symbol='DPYD-AS2', Aliases='', description='DPYD antisense RNA 2', other_designations='DPYD antisense RNA 2 (non-protein coding)', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97796921, end_position_on_the_genomic_accession=97798051, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873934, CurrentID=0, Status='live', Symbol='ETV5-AS1', Aliases='', description='ETV5 antisense RNA 1', other_designations='ETV5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79567, CurrentID=0, Status='live', Symbol='FAM65A', Aliases='', description='family with sequence similarity 65 member A', other_designations='family with sequence similarity 65, member A', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67528814, end_position_on_the_genomic_accession=67546788, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  79568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79568, CurrentID=0, Status='live', Symbol='C2orf47', Aliases='', description='chromosome 2 open reading frame 47', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199955317, end_position_on_the_genomic_accession=199964125, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100873937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873937, CurrentID=0, Status='live', Symbol='GRM7-AS1', Aliases='', description='GRM7 antisense RNA 1', other_designations='GRM7 antisense RNA 1 (non-protein coding)', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=7519741, end_position_on_the_genomic_accession=7535284, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79570, CurrentID=0, Status='live', Symbol='NKAIN1', Aliases='FAM77C', description='Na+/K+ transporting ATPase interacting 1', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 1|family with sequence similarity 77, member C', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31179745, end_position_on_the_genomic_accession=31239887, orientation='minus', exon_count=7, OMIM=612871, no_hdr0=''),
  79571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79571, CurrentID=0, Status='live', Symbol='GCC1', Aliases='GCC1P, GCC88', description='GRIP and coiled-coil domain containing 1', other_designations='golgi coiled-coil 1|golgi coiled-coil protein 1|peripheral membrane Golgi protein', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127580628, end_position_on_the_genomic_accession=127585600, orientation='minus', exon_count=2, OMIM=607418, no_hdr0=''),
  79572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79572, CurrentID=0, Status='live', Symbol='ATP13A3', Aliases='AFURS1', description='ATPase type 13A3', other_designations='ATPase family homolog up-regulated in senescence cells 1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194402674, end_position_on_the_genomic_accession=194486654, orientation='minus', exon_count=37, OMIM=610232, no_hdr0=''),
  79573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79573, CurrentID=0, Status='live', Symbol='TTC13', Aliases='', description='tetratricopeptide repeat domain 13', other_designations='TPR repeat protein 13', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230906241, end_position_on_the_genomic_accession=230978872, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  79574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79574, CurrentID=0, Status='live', Symbol='EPS8L3', Aliases='EPS8R3', description='EPS8-like 3', other_designations='EPS8-like protein 3|EPS8-related protein 3|epidermal growth factor receptor pathway substrate 8-related protein 3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109750080, end_position_on_the_genomic_accession=109764022, orientation='minus', exon_count=20, OMIM=614989, no_hdr0=''),
  79575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79575, CurrentID=0, Status='live', Symbol='ABHD8', Aliases='', description='abhydrolase domain containing 8', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17292131, end_position_on_the_genomic_accession=17303473, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79576, CurrentID=0, Status='live', Symbol='NKAP', Aliases='', description='NFKB activating protein', other_designations='NF-kappaB activating protein', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119925050, end_position_on_the_genomic_accession=119943772, orientation='minus', exon_count=9, OMIM=300766, no_hdr0=''),
  79577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79577, CurrentID=0, Status='live', Symbol='CDC73', Aliases='C1orf28, FIHP, HPTJT, HRPT1, HRPT2, HYX', description='cell division cycle 73', other_designations='Paf1/RNA polymerase II complex component|cell division cycle 73 Paf1/RNA polymerase II complex component-like protein|cell division cycle 73, Paf1/RNA polymerase II complex component, homolog|cell division cycle protein 73 homolog|hyperparathyroidism 2 protein', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193121958, end_position_on_the_genomic_accession=193254815, orientation='plus', exon_count=18, OMIM=607393, no_hdr0=''),
  100873946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873946, CurrentID=0, Status='live', Symbol='VAV3-AS1', Aliases='', description='VAV3 antisense RNA 1', other_designations='VAV3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107964443, end_position_on_the_genomic_accession=107994607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873947, CurrentID=0, Status='live', Symbol='LMLN-AS1', Aliases='', description='LMLN antisense RNA 1', other_designations='LMLN antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198038321, end_position_on_the_genomic_accession=198039467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873948, CurrentID=0, Status='live', Symbol='NREP-AS1', Aliases='', description='NREP antisense RNA 1', other_designations='CTD-2192A1.1|NREP antisense RNA 1 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111912508, end_position_on_the_genomic_accession=112017306, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79581, CurrentID=0, Status='live', Symbol='SLC52A2', Aliases='BVVLS2, D15Ertd747e, GPCR41, GPR172A, PAR1, RFT3, RFVT2, hRFT3', description='solute carrier family 52 (riboflavin transporter), member 2', other_designations='G protein-coupled receptor 172A|PERV-A receptor 1|porcine endogenous retrovirus A receptor 1|putative G-protein coupled receptor GPCR41|riboflavin transporter 3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144358557, end_position_on_the_genomic_accession=144361286, orientation='plus', exon_count=7, OMIM=607882, no_hdr0=''),
  79582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79582, CurrentID=0, Status='live', Symbol='SPAG16', Aliases='PF20, WDR29', description='sperm associated antigen 16', other_designations='WD repeat domain 29|pf20 protein homolog|sperm-associated WD repeat protein', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=213284379, end_position_on_the_genomic_accession=214410501, orientation='plus', exon_count=35, OMIM=612173, no_hdr0=''),
  79583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79583, CurrentID=0, Status='live', Symbol='TMEM231', Aliases='ALYE870, JBTS20, MKS11, PRO1886', description='transmembrane protein 231', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75538117, end_position_on_the_genomic_accession=75556286, orientation='minus', exon_count=7, OMIM=614949, no_hdr0=''),
  100873952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873952, CurrentID=0, Status='live', Symbol='SZT2-AS1', Aliases='', description='SZT2 antisense RNA 1', other_designations='SZT2 antisense RNA 1 (non-protein coding)', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79585, CurrentID=0, Status='live', Symbol='CORO7', Aliases='0610011B16Rik, CRN7, POD1', description='coronin 7', other_designations='70 kDa WD repeat tumor rejection antigen homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4354542, end_position_on_the_genomic_accession=4416961, orientation='minus', exon_count=28, OMIM=611668, no_hdr0=''),
  79586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79586, CurrentID=0, Status='live', Symbol='CHPF', Aliases='CHSY2, CSS2', description='chondroitin polymerizing factor', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase II|N-acetylgalactosaminyltransferase 2|chondroitin glucuronyltransferase 2|chondroitin-polymerizing factor|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase II', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219538947, end_position_on_the_genomic_accession=219543765, orientation='minus', exon_count=6, OMIM=610405, no_hdr0=''),
  79587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79587, CurrentID=0, Status='live', Symbol='CARS2', Aliases='cysRS', description='cysteinyl-tRNA synthetase 2, mitochondrial (putative)', other_designations='cysteine tRNA ligase 2, mitochondrial (putative)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110641410, end_position_on_the_genomic_accession=110713601, orientation='minus', exon_count=22, OMIM=612800, no_hdr0=''),
  100873956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873956, CurrentID=0, Status='live', Symbol='EAF1-AS1', Aliases='', description='EAF1 antisense RNA 1', other_designations='EAF1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79589, CurrentID=0, Status='live', Symbol='RNF128', Aliases='GRAIL', description='ring finger protein 128, E3 ubiquitin protein ligase', other_designations='gene related to anergy in lymphocytes protein', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106693838, end_position_on_the_genomic_accession=106797016, orientation='plus', exon_count=8, OMIM=300439, no_hdr0=''),
  79590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79590, CurrentID=0, Status='live', Symbol='MRPL24', Aliases='L24mt, MRP-L18, MRP-L24', description='mitochondrial ribosomal protein L24', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156737302, end_position_on_the_genomic_accession=156741266, orientation='minus', exon_count=7, OMIM=611836, no_hdr0=''),
  79591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79591, CurrentID=0, Status='live', Symbol='C10orf76', Aliases='', description='chromosome 10 open reading frame 76', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101845599, end_position_on_the_genomic_accession=102056232, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  100873960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873960, CurrentID=0, Status='live', Symbol='FTLP19', Aliases='', description='ferritin, light polypeptide pseudogene 19', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16071416, end_position_on_the_genomic_accession=16071809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  319100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319100, CurrentID=0, Status='live', Symbol='TAAR6', Aliases='TA4, TAR4, TAR6, TRAR4, taR-4, taR-6', description='trace amine associated receptor 6', other_designations='trace amine receptor 4|trace amine receptor 6', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132570322, end_position_on_the_genomic_accession=132571359, orientation='plus', exon_count=1, OMIM=608923, no_hdr0=''),
  79594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79594, CurrentID=0, Status='live', Symbol='MUL1', Aliases='C1orf166, GIDE, MAPL, MULAN, RNF218', description='mitochondrial E3 ubiquitin protein ligase 1', other_designations='E3 SUMO-protein ligase MUL1|E3 ubiquitin ligase|E3 ubiquitin-protein ligase MUL1|growth inhibition and death E3 ligase|mitochondria-anchored protein ligase|mitochondrial E3 ubiquitin ligase 1|mitochondrial ubiquitin ligase activator of NF-kB|mitochondrial-anchored protein ligase|putative NF-kappa-B-activating protein 266|ring finger protein 218', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20499448, end_position_on_the_genomic_accession=20508181, orientation='minus', exon_count=5, OMIM=612037, no_hdr0=''),
  79595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79595, CurrentID=0, Status='live', Symbol='SAP130', Aliases='', description='Sin3A associated protein 130kDa', other_designations='130 kDa Sin3-associated polypeptide|Sin3-associated polypeptide p130|Sin3A-associated protein, 130kDa|mSin3A-associated protein 130', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127941217, end_position_on_the_genomic_accession=128028093, orientation='minus', exon_count=22, OMIM=609697, no_hdr0=''),
  79596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79596, CurrentID=0, Status='live', Symbol='RNF219', Aliases='C13orf7', description='ring finger protein 219', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78614286, end_position_on_the_genomic_accession=78659179, orientation='minus', exon_count=8, OMIM=615906, no_hdr0=''),
  100873965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873965, CurrentID=0, Status='live', Symbol='MED4-AS1', Aliases='MED4-AS', description='MED4 antisense RNA 1', other_designations='MED4 antisense RNA (non-protein coding)|MED4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48077137, end_position_on_the_genomic_accession=48079993, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79598, CurrentID=0, Status='live', Symbol='CEP97', Aliases='2810403B08Rik, LRRIQ2', description='centrosomal protein 97kDa', other_designations='leucine-rich repeat and IQ domain-containing protein 2|leucine-rich repeats and IQ motif containing 2', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101724084, end_position_on_the_genomic_accession=101770562, orientation='plus', exon_count=12, OMIM=615864, no_hdr0=''),
  56957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=56957, CurrentID=0, Status='live', Symbol='OTUD7B', Aliases='CEZANNE, ZA20D1', description='OTU deubiquitinase 7B', other_designations='OTU domain containing 7B|cellular zinc finger anti-NF-kappa-B protein|cellular zinc finger anti-NF-kappaB Cezanne|zinc finger A20 domain-containing protein 1|zinc finger protein Cezanne|zinc finger, A20 domain containing 1', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149940330, end_position_on_the_genomic_accession=150010772, orientation='minus', exon_count=15, OMIM=611748, no_hdr0=''),
  79600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79600, CurrentID=0, Status='live', Symbol='TCTN1', Aliases='JBTS13, TECT1', description='tectonic family member 1', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110614027, end_position_on_the_genomic_accession=110649130, orientation='plus', exon_count=19, OMIM=609863, no_hdr0=''),
  100873969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873969, CurrentID=0, Status='live', Symbol='GPC5-AS1', Aliases='', description='GPC5 antisense RNA 1', other_designations='GPC5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92701389, end_position_on_the_genomic_accession=92721614, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79602, CurrentID=0, Status='live', Symbol='ADIPOR2', Aliases='ACDCR2, PAQR2', description='adiponectin receptor 2', other_designations='progestin and adipoQ receptor family member II', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1690981, end_position_on_the_genomic_accession=1788679, orientation='plus', exon_count=9, OMIM=607946, no_hdr0=''),
  79603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79603, CurrentID=0, Status='live', Symbol='CERS4', Aliases='LASS4, Trh1', description='ceramide synthase 4', other_designations='LAG1 homolog, ceramide synthase 4|LAG1 longevity assurance homolog 4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8209329, end_position_on_the_genomic_accession=8262421, orientation='plus', exon_count=16, OMIM=615334, no_hdr0=''),
  100873972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873972, CurrentID=0, Status='live', Symbol='GPC6-AS1', Aliases='', description='GPC6 antisense RNA 1', other_designations='GPC6 antisense RNA 1 (non-protein coding)', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94154193, end_position_on_the_genomic_accession=94187991, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79605, CurrentID=0, Status='live', Symbol='PGBD5', Aliases='', description='piggyBac transposable element derived 5', other_designations='piggyBac domain related protein 5', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230321646, end_position_on_the_genomic_accession=230425928, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100873974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873974, CurrentID=0, Status='live', Symbol='KCNC4-AS1', Aliases='', description='KCNC4 antisense RNA 1 (head to head)', other_designations='KCNC4 antisense RNA 1 (non-protein coding)', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110208450, end_position_on_the_genomic_accession=110209987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79607, CurrentID=0, Status='live', Symbol='FAM118B', Aliases='', description='family with sequence similarity 118 member B', other_designations='family with sequence similarity 118, member B', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126211450, end_position_on_the_genomic_accession=126262984, orientation='plus', exon_count=9, OMIM=616587, no_hdr0=''),
  79608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79608, CurrentID=0, Status='live', Symbol='RIC3', Aliases='AYST720, PRO1385', description='RIC3 acetylcholine receptor chaperone', other_designations='resistance to inhibitors of cholinesterase 3 homolog|resistance to inhibitors of cholinesterase 3-like protein|resistant to inhibitor of cholinesterase 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8092965, end_position_on_the_genomic_accession=8169083, orientation='minus', exon_count=15, OMIM=610509, no_hdr0=''),
  79609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79609, CurrentID=0, Status='live', Symbol='VCPKMT', Aliases='C14orf138, METTL21D, VCP-KMT', description='valosin containing protein lysine (K) methyltransferase', other_designations='VCP lysine methyltransferase|methyltransferase like 21D|methyltransferase-like protein 21D|valosin-containing protein lysine methyltransferase', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50108632, end_position_on_the_genomic_accession=50116579, orientation='minus', exon_count=8, OMIM=615260, no_hdr0=''),
  100873978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873978, CurrentID=0, Status='live', Symbol='RBMS3-AS2', Aliases='', description='RBMS3 antisense RNA 2', other_designations='RBMS3 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29615151, end_position_on_the_genomic_accession=29643080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79611, CurrentID=0, Status='live', Symbol='ACSS3', Aliases='', description='acyl-CoA synthetase short-chain family member 3', other_designations='AMP-binding enzyme, 33217', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81078018, end_position_on_the_genomic_accession=81257795, orientation='plus', exon_count=19, OMIM=614356, no_hdr0=''),
  79612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79612, CurrentID=0, Status='live', Symbol='NAA16', Aliases='NARG1L', description='N(alpha)-acetyltransferase 16, NatA auxiliary subunit', other_designations='NARG1-like protein|NMDA receptor-regulated 1-like protein', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41311205, end_position_on_the_genomic_accession=41377030, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  79613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79613, CurrentID=0, Status='live', Symbol='TANGO6', Aliases='TMCO7', description='transport and golgi organization 6 homolog', other_designations='transmembrane and coiled-coil domain-containing protein 7|transmembrane and coiled-coil domains 7', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68843553, end_position_on_the_genomic_accession=69085182, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100873982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873982, CurrentID=0, Status='live', Symbol='ABCC5-AS1', Aliases='', description='ABCC5 antisense RNA 1', other_designations='ABCC5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184006338, end_position_on_the_genomic_accession=184011419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873983, CurrentID=0, Status='live', Symbol='MAGI1-AS1', Aliases='', description='MAGI1 antisense RNA 1', other_designations='MAGI1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65893816, end_position_on_the_genomic_accession=65925297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79616, CurrentID=0, Status='live', Symbol='CCNJL', Aliases='', description='cyclin J-like', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160251659, end_position_on_the_genomic_accession=160313703, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100873985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873985, CurrentID=0, Status='live', Symbol='MED14OS', Aliases='MED14-AS1', description='MED14 opposite strand', other_designations='MED14 antisense RNA 1', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40735396, end_position_on_the_genomic_accession=40738701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79618, CurrentID=0, Status='live', Symbol='HMBOX1', Aliases='HNF1LA, HOT1, PBHNF', description='homeobox containing 1', other_designations='homeobox telomere-binding protein 1|homeobox-containing protein PBHNF', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28890394, end_position_on_the_genomic_accession=29064939, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100873987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873987, CurrentID=0, Status='live', Symbol='FGF12-AS2', Aliases='', description='FGF12 antisense RNA 2', other_designations='FGF12 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873988, CurrentID=0, Status='live', Symbol='FGF12-AS3', Aliases='', description='FGF12 antisense RNA 3', other_designations='FGF12 antisense RNA 3 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192514844, end_position_on_the_genomic_accession=192521899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79621, CurrentID=0, Status='live', Symbol='RNASEH2B', Aliases='AGS2, DLEU8', description='ribonuclease H2 subunit B', other_designations='Aicardi-Goutieres syndrome 2 protein|RNase H2 subunit B|deleted in lymphocytic leukemia 8|ribonuclease H2, subunit B|ribonuclease HI subunit B', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50909678, end_position_on_the_genomic_accession=50970462, orientation='plus', exon_count=16, OMIM=610326, no_hdr0=''),
  79622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79622, CurrentID=0, Status='live', Symbol='SNRNP25', Aliases='C16orf33', description='small nuclear ribonucleoprotein, U11/U12 25kDa subunit', other_designations='U11/U12 snRNP 25 kDa protein|U11/U12 snRNP 25K|minus-99 protein|small nuclear ribonucleoprotein 25kDa (U11/U12)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53829, end_position_on_the_genomic_accession=57669, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  79623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79623, CurrentID=0, Status='live', Symbol='GALNT14', Aliases='GALNT15, GalNac-T10, GalNac-T14', description='polypeptide N-acetylgalactosaminyltransferase 14', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 14|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase T10|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase T14|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 14 (GalNAc-T14)|polypeptide GalNAc transferase 14|polypeptide N-acetylgalactosaminyltransferase 15|pp-GaNTase 14|protein-UDP acetylgalactosaminyltransferase 14', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30910465, end_position_on_the_genomic_accession=31161688, orientation='minus', exon_count=23, OMIM=608225, no_hdr0=''),
  79624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79624, CurrentID=0, Status='live', Symbol='ARMT1', Aliases='C6orf211', description='acidic residue methyltransferase 1', other_designations='UPF0364 protein C6orf211', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151452258, end_position_on_the_genomic_accession=151470101, orientation='plus', exon_count=5, OMIM=616332, no_hdr0=''),
  79625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79625, CurrentID=0, Status='live', Symbol='NDNF', Aliases='C4orf31, NORD', description='neuron-derived neurotrophic factor', other_designations='fibronectin type-III domain-containing protein C4orf31', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121035627, end_position_on_the_genomic_accession=121072518, orientation='minus', exon_count=4, OMIM=616506, no_hdr0=''),
  79626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79626, CurrentID=0, Status='live', Symbol='TNFAIP8L2', Aliases='TIPE2', description='TNF alpha induced protein 8 like 2', other_designations='TNF alpha-induced protein 8-like protein 2|TNFAIP8-like protein 2|inflammation factor 20|inflammation factor protein 20|tumor necrosis factor, alpha induced protein 8 like 2|tumor necrosis factor, alpha-induced protein 8-like 2|tumor necrosis factor, alpha-induced protein 8-like protein 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151156629, end_position_on_the_genomic_accession=151159749, orientation='plus', exon_count=2, OMIM=612112, no_hdr0=''),
  79627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79627, CurrentID=0, Status='live', Symbol='OGFRL1', Aliases='dJ331H24.1', description='opioid growth factor receptor-like 1', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71288774, end_position_on_the_genomic_accession=71308953, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  79628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79628, CurrentID=0, Status='live', Symbol='SH3TC2', Aliases='CMT4C, MNMN', description='SH3 domain and tetratricopeptide repeats 2', other_designations='SH3 domain and tetratricopeptide repeats-containing protein 2', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148982150, end_position_on_the_genomic_accession=149063174, orientation='minus', exon_count=17, OMIM=608206, no_hdr0=''),
  79629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79629, CurrentID=0, Status='live', Symbol='OCEL1', Aliases='FWP009, S863-9', description='occludin/ELL domain containing 1', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17226204, end_position_on_the_genomic_accession=17229219, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79630, CurrentID=0, Status='live', Symbol='C1orf54', Aliases='', description='chromosome 1 open reading frame 54', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150269004, end_position_on_the_genomic_accession=150280924, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79631, CurrentID=0, Status='live', Symbol='EFTUD1', Aliases='FAM42A, HsT19294, RIA1', description='elongation factor Tu GTP binding domain containing 1', other_designations='elongation factor-like 1|ribosome assembly 1 homolog', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82130220, end_position_on_the_genomic_accession=82262763, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  79632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79632, CurrentID=0, Status='live', Symbol='FAM184A', Aliases='C6orf60', description='family with sequence similarity 184 member A', other_designations='family with sequence similarity 184, member A', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118959829, end_position_on_the_genomic_accession=119149193, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  79633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79633, CurrentID=0, Status='live', Symbol='FAT4', Aliases='CDHF14, CDHR11, FAT-J, FATJ, HKLLS2, NBLA00548, VMLDS2', description='FAT atypical cadherin 4', other_designations='FAT tumor suppressor homolog 4|cadherin family member 14|cadherin-related family member 11|fat-like cadherin protein FAT-J|putative protein product of Nbla00548', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=125315897, end_position_on_the_genomic_accession=125492932, orientation='plus', exon_count=18, OMIM=612411, no_hdr0=''),
  79634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79634, CurrentID=0, Status='live', Symbol='SCRN3', Aliases='SES3', description='secernin 3', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174395729, end_position_on_the_genomic_accession=174429575, orientation='plus', exon_count=9, OMIM=614967, no_hdr0=''),
  79635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79635, CurrentID=0, Status='live', Symbol='CCDC121', Aliases='', description='coiled-coil domain containing 121', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27625638, end_position_on_the_genomic_accession=27629031, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874004, CurrentID=0, Status='live', Symbol='PPEF1-AS1', Aliases='', description='PPEF1 antisense RNA 1', other_designations='PPEF1 antisense RNA 1 (non-protein coding)', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18688643, end_position_on_the_genomic_accession=18692687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79637, CurrentID=0, Status='live', Symbol='ARMC7', Aliases='', description='armadillo repeat containing 7', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75109952, end_position_on_the_genomic_accession=75130265, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874006, CurrentID=0, Status='live', Symbol='TTC3-AS1', Aliases='', description='TTC3 antisense RNA 1', other_designations='TTC3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37187612, end_position_on_the_genomic_accession=37194628, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79639, CurrentID=0, Status='live', Symbol='TMEM53', Aliases='NET4', description='transmembrane protein 53', other_designations='novel DUF829 domain-containing protein', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44653247, end_position_on_the_genomic_accession=44674608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79640, CurrentID=0, Status='live', Symbol='C22orf46', Aliases='', description='chromosome 22 open reading frame 46', other_designations='CTA-216E10.6|putative uncharacterized protein C22orf46', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41689936, end_position_on_the_genomic_accession=41698136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79641, CurrentID=0, Status='live', Symbol='ROGDI', Aliases='KTZS', description='rogdi homolog', other_designations='leucine zipper domain protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4796962, end_position_on_the_genomic_accession=4802950, orientation='minus', exon_count=11, OMIM=614574, no_hdr0=''),
  79642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79642, CurrentID=0, Status='live', Symbol='ARSJ', Aliases='', description='arylsulfatase family member J', other_designations='ASJ|arylsulfatase family, member J', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113900284, end_position_on_the_genomic_accession=113981430, orientation='minus', exon_count=7, OMIM=610010, no_hdr0=''),
  79643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79643, CurrentID=0, Status='live', Symbol='CHMP6', Aliases='VPS20', description='charged multivesicular body protein 6', other_designations='chromatin modifying protein 6|chromatin-modifying protein 6|hVps20|vacuolar protein sorting-associated protein 20', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80991841, end_position_on_the_genomic_accession=81000133, orientation='plus', exon_count=8, OMIM=610901, no_hdr0=''),
  79644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79644, CurrentID=0, Status='live', Symbol='SRD5A3', Aliases='CDG1P, CDG1Q, KRIZI, SRD5A2L, SRD5A2L1', description='steroid 5 alpha-reductase 3', other_designations='3-oxo-5-alpha-steroid 4-dehydrogenase 3|S5AR 3|SR type 3|probable polyprenol reductase', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55346221, end_position_on_the_genomic_accession=55373100, orientation='plus', exon_count=5, OMIM=611715, no_hdr0=''),
  79645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79645, CurrentID=0, Status='live', Symbol='EFCAB1', Aliases='', description='EF-hand calcium binding domain 1', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48714915, end_position_on_the_genomic_accession=48735311, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79646, CurrentID=0, Status='live', Symbol='PANK3', Aliases='', description='pantothenate kinase 3', other_designations='hPanK3|pantothenic acid kinase 3', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168555623, end_position_on_the_genomic_accession=168579609, orientation='minus', exon_count=7, OMIM=606161, no_hdr0=''),
  79647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79647, CurrentID=0, Status='live', Symbol='AKIRIN1', Aliases='C1orf108, STRF2', description='akirin 1', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38991244, end_position_on_the_genomic_accession=39006065, orientation='plus', exon_count=6, OMIM=615164, no_hdr0=''),
  79648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79648, CurrentID=0, Status='live', Symbol='MCPH1', Aliases='BRIT1, MCT', description='microcephalin 1', other_designations='BRCT-repeat inhibitor of TERT expression 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6406592, end_position_on_the_genomic_accession=6643619, orientation='plus', exon_count=19, OMIM=607117, no_hdr0=''),
  79649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79649, CurrentID=0, Status='live', Symbol='MAP7D3', Aliases='MDP3', description='MAP7 domain containing 3', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136213220, end_position_on_the_genomic_accession=136256482, orientation='minus', exon_count=21, OMIM=300930, no_hdr0=''),
  79650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79650, CurrentID=0, Status='live', Symbol='USB1', Aliases='C16orf57, HVSL1, Mpn1, PN, hUsb1', description='U6 snRNA biogenesis 1', other_designations='HVSL motif containing 1|U six biogenesis 1|UPF0406 protein C16orf57|mutated in poikiloderma with neutropenia protein 1|putative U6 snRNA phosphodiesterase', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57999600, end_position_on_the_genomic_accession=58021623, orientation='plus', exon_count=10, OMIM=613276, no_hdr0=''),
  79651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79651, CurrentID=0, Status='live', Symbol='RHBDF2', Aliases='RHBDL5, RHBDL6, TEC, TOC, TOCG, iRhom2', description='rhomboid 5 homolog 2 (Drosophila)', other_designations='rhomboid family member 2|rhomboid veinlet-like protein 5|rhomboid veinlet-like protein 6', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76470893, end_position_on_the_genomic_accession=76501437, orientation='minus', exon_count=21, OMIM=614404, no_hdr0=''),
  79652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79652, CurrentID=0, Status='live', Symbol='TMEM204', Aliases='C16orf30, CLP24', description='transmembrane protein 204', other_designations='claudin-like protein 24|claudin-like protein of 24 kDa', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1528741, end_position_on_the_genomic_accession=1555580, orientation='plus', exon_count=5, OMIM=611002, no_hdr0=''),
  100874021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874021, CurrentID=0, Status='live', Symbol='MAGI2-AS2', Aliases='', description='MAGI2 antisense RNA 2', other_designations='MAGI2 antisense RNA 2 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79008988, end_position_on_the_genomic_accession=79012277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79654, CurrentID=0, Status='live', Symbol='HECTD3', Aliases='', description='HECT domain containing E3 ubiquitin protein ligase 3', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45002546, end_position_on_the_genomic_accession=45011355, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  341799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341799, CurrentID=0, Status='live', Symbol='OR6S1', Aliases='OR14-37Q, OR6S1', description='olfactory receptor family 6 subfamily S member 1', other_designations='olfactory receptor OR14-37|olfactory receptor, family 6, subfamily S, member 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20640696, end_position_on_the_genomic_accession=20641691, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79656, CurrentID=0, Status='live', Symbol='BEND5', Aliases='C1orf165', description='BEN domain containing 5', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48727523, end_position_on_the_genomic_accession=48776969, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79657, CurrentID=0, Status='live', Symbol='RPAP3', Aliases='', description='RNA polymerase II associated protein 3', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47661932, end_position_on_the_genomic_accession=47706061, orientation='minus', exon_count=17, OMIM=611477, no_hdr0=''),
  79658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79658, CurrentID=0, Status='live', Symbol='ARHGAP10', Aliases='GRAF2, PS-GAP, PSGAP', description='Rho GTPase activating protein 10', other_designations='GTPase regulator associated with focal adhesion kinase 2|graf-related protein 2|rho-type GTPase-activating protein 10', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147732063, end_position_on_the_genomic_accession=148072779, orientation='plus', exon_count=25, OMIM=609746, no_hdr0=''),
  79659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79659, CurrentID=0, Status='live', Symbol='DYNC2H1', Aliases='ATD3, DHC1b, DHC2, DNCH2, DYH1B, SRPS2B, SRTD3, hdhc11', description='dynein, cytoplasmic 2, heavy chain 1', other_designations='dynein cytoplasmic heavy chain 2|dynein heavy chain 11|dynein heavy chain isotype 1B|dynein heavy chain, isotype 1B|dynein, cytoplasmic, heavy polypeptide 2', map_location='11q21-q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103109431, end_position_on_the_genomic_accession=103479863, orientation='plus', exon_count=90, OMIM=603297, no_hdr0=''),
  79660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79660, CurrentID=0, Status='live', Symbol='PPP1R3B', Aliases='GL, PPP1R4, PTG', description='protein phosphatase 1 regulatory subunit 3B', other_designations='PP1 subunit R4|hepatic glycogen-targeting protein phosphatase 1 regulatory subunit GL|hepatic glycogen-targeting subunit, G(L)|protein phosphatase 1 regulatory subunit 4|protein phosphatase 1, regulatory (inhibitor) subunit 3B|protein phosphatase 1, regulatory subunit 3B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9136254, end_position_on_the_genomic_accession=9151642, orientation='minus', exon_count=3, OMIM=610541, no_hdr0=''),
  79661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79661, CurrentID=0, Status='live', Symbol='NEIL1', Aliases='FPG1, NEI1, hFPG1', description='nei-like DNA glycosylase 1', other_designations='DNA endonuclease eight-like glycosylase 1|DNA glycosylase/AP lyase Neil1|DNA-(apurinic or apyrimidinic site) lyase Neil1|NEH1|endonuclease VIII|endonuclease VIII-like 1|nei endonuclease VIII-like 1|nei homolog 1|nei-like protein 1', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75346649, end_position_on_the_genomic_accession=75355251, orientation='plus', exon_count=15, OMIM=608844, no_hdr0=''),
  100874030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874030, CurrentID=0, Status='live', Symbol='SIDT1-AS1', Aliases='', description='SIDT1 antisense RNA 1', other_designations='SIDT1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79663, CurrentID=0, Status='live', Symbol='HSPBAP1', Aliases='PASS1', description='HSPB (heat shock 27kDa) associated protein 1', other_designations='27 kDa heat shock protein-associated protein 1|protein associating with small stress protein PASS1', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122739997, end_position_on_the_genomic_accession=122795216, orientation='minus', exon_count=13, OMIM=608263, no_hdr0=''),
  79664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79664, CurrentID=0, Status='live', Symbol='ICE2', Aliases='BRCC1, NARG2', description='interactor of little elongation complex ELL subunit 2', other_designations='NMDA receptor regulated 2|NMDA receptor-regulated gene 2|NMDA receptor-regulated protein 2|breast cancer cell 1|interactor of little elongator complex ELL subunit 2', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60419609, end_position_on_the_genomic_accession=60479160, orientation='minus', exon_count=20, OMIM=610835, no_hdr0=''),
  79665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79665, CurrentID=0, Status='live', Symbol='DHX40', Aliases='ARG147, DDX40, PAD', description='DEAH (Asp-Glu-Ala-His) box polypeptide 40', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 40 (RNA helicase)|DEAH box protein 40', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59565525, end_position_on_the_genomic_accession=59608352, orientation='plus', exon_count=18, OMIM=607570, no_hdr0=''),
  79666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79666, CurrentID=0, Status='live', Symbol='PLEKHF2', Aliases='EAPF, PHAFIN2, ZFYVE18', description='pleckstrin homology and FYVE domain containing 2', other_designations='PH and FYVE domain-containing protein 2|PH domain-containing family F member 2|endoplasmic reticulum-associated apoptosis-involved protein containing PH and FYVE domains|pleckstrin homology domain containing, family F (with FYVE domain) member 2|zinc finger FYVE domain-containing protein 18', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95133721, end_position_on_the_genomic_accession=95156685, orientation='plus', exon_count=2, OMIM=615208, no_hdr0=''),
  79667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79667, CurrentID=0, Status='live', Symbol='KLF3-AS1', Aliases='', description='KLF3 antisense RNA 1', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38612701, end_position_on_the_genomic_accession=38664628, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79668, CurrentID=0, Status='live', Symbol='PARP8', Aliases='ARTD16, pART16', description='poly(ADP-ribose) polymerase family member 8', other_designations='ADP-ribosyltransferase diphtheria toxin-like 16|poly (ADP-ribose) polymerase family, member 8', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=50665899, end_position_on_the_genomic_accession=50846522, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  79669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79669, CurrentID=0, Status='live', Symbol='C3orf52', Aliases='TTMP', description='chromosome 3 open reading frame 52', other_designations='TPA induced trans-membrane protein|novel endoplasmic reticulum transmembrane protein', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112086335, end_position_on_the_genomic_accession=112131095, orientation='plus', exon_count=9, OMIM=611956, no_hdr0=''),
  79670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79670, CurrentID=0, Status='live', Symbol='ZCCHC6', Aliases='PAPD6, TUT7', description='zinc finger, CCHC domain containing 6', other_designations='PAP associated domain containing 6|TUTase 7|zinc finger CCHC domain-containing protein 6', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86287733, end_position_on_the_genomic_accession=86354501, orientation='minus', exon_count=33, OMIM=613467, no_hdr0=''),
  79671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79671, CurrentID=0, Status='live', Symbol='NLRX1', Aliases='CLR11.3, DLNB26, NOD26, NOD5, NOD9', description='NLR family member X1', other_designations='NLR family, X1|NOD-like receptor X1|caterpiller protein 11.3|nucleotide-binding oligomerization domain protein 26|nucleotide-binding oligomerization domain protein 5|nucleotide-binding oligomerization domain protein 9|nucleotide-binding oligomerization domain, leucine rich repeat containing X1', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119168334, end_position_on_the_genomic_accession=119184016, orientation='plus', exon_count=12, OMIM=611947, no_hdr0=''),
  79672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79672, CurrentID=0, Status='live', Symbol='FN3KRP', Aliases='FN3KL', description='fructosamine 3 kinase related protein', other_designations='FN3K-RP|FN3K-related protein|fructosamine-3-kinase-related protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82716706, end_position_on_the_genomic_accession=82728017, orientation='plus', exon_count=7, OMIM=611683, no_hdr0=''),
  79673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79673, CurrentID=0, Status='live', Symbol='ZNF329', Aliases='', description='zinc finger protein 329', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58126328, end_position_on_the_genomic_accession=58155110, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79674, CurrentID=0, Status='live', Symbol='VEPH1', Aliases='MELT, VEPH', description='ventricular zone expressed PH domain containing 1', other_designations='protein melted|ventricular zone expressed PH domain homolog 1|ventricular zone expressed PH domain-containing 1', map_location='3q24-q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157259742, end_position_on_the_genomic_accession=157503626, orientation='minus', exon_count=20, OMIM=609594, no_hdr0=''),
  79675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79675, CurrentID=0, Status='live', Symbol='FASTKD1', Aliases='', description='FAST kinase domains 1', other_designations='', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169527880, end_position_on_the_genomic_accession=169573921, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  79676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79676, CurrentID=0, Status='live', Symbol='OGFOD2', Aliases='', description='2-oxoglutarate and iron-dependent oxygenase domain containing 2', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122974703, end_position_on_the_genomic_accession=122980041, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  79677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79677, CurrentID=0, Status='live', Symbol='SMC6', Aliases='SMC-6L1, hSMC6, SMC6', description='structural maintenance of chromosomes 6', other_designations='SMC protein 6|SMC6 structural maintenance of chromosomes 6-like 1', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17663812, end_position_on_the_genomic_accession=17753829, orientation='minus', exon_count=30, OMIM=609387, no_hdr0=''),
  100874046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874046, CurrentID=0, Status='live', Symbol='LINC00498', Aliases='', description='long intergenic non-protein coding RNA 498', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79679, CurrentID=0, Status='live', Symbol='VTCN1', Aliases='B7-H4, B7H4, B7S1, B7X, B7h.5, PRO1291, VCTN1', description='V-set domain containing T cell activation inhibitor 1', other_designations='B7 family member, H4|B7 homolog 4|B7 superfamily member 1|T cell costimulatory molecule B7x|T-cell costimulatory molecule B7x|immune costimulatory protein B7-H4', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117143587, end_position_on_the_genomic_accession=117270368, orientation='minus', exon_count=10, OMIM=608162, no_hdr0=''),
  79680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79680, CurrentID=0, Status='live', Symbol='C22orf29', Aliases='BOP', description='chromosome 22 open reading frame 29', other_designations='BH3-only protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19846138, end_position_on_the_genomic_accession=19854848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874049, CurrentID=0, Status='live', Symbol='LINC00505', Aliases='', description='long intergenic non-protein coding RNA 505', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79682, CurrentID=0, Status='live', Symbol='CENPU', Aliases='CENP5050, KLIP1, MLF1IP, PBIP1, CENPU', description='centromere protein U', other_designations='KSHV latent nuclear antigen interacting protein 1|KSHV latent nuclear antigen-interacting protein 1|MLF1 interacting protein|MLF1-interacting protein|centromere protein of 50 kDa|interphase centromere complex protein 24|polo-box-interacting protein 1', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184694065, end_position_on_the_genomic_accession=184734132, orientation='minus', exon_count=14, OMIM=611511, no_hdr0=''),
  79683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79683, CurrentID=0, Status='live', Symbol='ZDHHC14', Aliases='NEW1CP', description='zinc finger, DHHC-type containing 14', other_designations='DHHC-14|NEW1 domain containing protein|NEW1 domain-containing protein|zinc finger DHHC domain-containing protein 14|zinc finger, DHHC domain containing 14', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157381525, end_position_on_the_genomic_accession=157678114, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  79684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79684, CurrentID=0, Status='live', Symbol='MSANTD2', Aliases='C11orf61', description='Myb/SANT-like DNA-binding domain containing 2', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124766492, end_position_on_the_genomic_accession=124800436, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79685, CurrentID=0, Status='live', Symbol='SAP30L', Aliases='NS4ATP2', description='SAP30-like', other_designations='HCV non-structural protein 4A-transactivated protein 2|Sin3A associated protein p30-like|sin3 corepressor complex subunit SAP30L|sin3-associated protein p30-like', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154445957, end_position_on_the_genomic_accession=154461053, orientation='plus', exon_count=5, OMIM=610398, no_hdr0=''),
  100874054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874054, CurrentID=0, Status='live', Symbol='FALEC', Aliases='FAL1, LINC00568, ncRNA-a1', description='focally amplified long non-coding RNA in epithelial cancer', other_designations='focally amplified lncRNA on chromosome 1|long intergenic non-protein coding RNA 568|noncoding RNA-activating 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150515757, end_position_on_the_genomic_accession=150518032, orientation='plus', exon_count=2, OMIM=616092, no_hdr0=''),
  100874055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874055, CurrentID=0, Status='live', Symbol='LINC00570', Aliases='ncRNA-a5', description='long intergenic non-protein coding RNA 570', other_designations='noncoding RNA-activating 5', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11393981, end_position_on_the_genomic_accession=11403077, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874056, CurrentID=0, Status='live', Symbol='NLGN4Y-AS1', Aliases='NLGN4Y-AS, NLGN4YAS', description='NLGN4Y antisense RNA 1', other_designations='NLGN4Y antisense RNA (non-protein coding)|NLGN4Y antisense RNA 1 (non-protein coding)', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14793642, end_position_on_the_genomic_accession=14804033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79689, CurrentID=0, Status='live', Symbol='STEAP4', Aliases='STAMP2, TIARP, TNFAIP9', description='STEAP4 metalloreductase', other_designations='STEAP family member 4|six transmembrane prostate protein 2|six-transmembrane epithelial antigen of prostate 4|sixTransMembrane protein of prostate 2|tumor necrosis factor, alpha-induced protein 9|tumor necrosis-alpha-induced adipose-related protein', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88276429, end_position_on_the_genomic_accession=88306913, orientation='minus', exon_count=6, OMIM=611098, no_hdr0=''),
  79690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79690, CurrentID=0, Status='live', Symbol='GAL3ST4', Aliases='GAL3ST-4', description='galactose-3-O-sulfotransferase 4', other_designations="beta-galactose-3-O-sulfotransferase 4|beta-galactose-3-O-sulfotransferase, 4|gal-beta-1,3-GalNAc 3'-sulfotransferase|galbeta1-3GalNAc 3'-sulfotransferase", map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100159242, end_position_on_the_genomic_accession=100168750, orientation='minus', exon_count=4, OMIM=608235, no_hdr0=''),
  79691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79691, CurrentID=0, Status='live', Symbol='QTRTD1', Aliases='', description='queuine tRNA-ribosyltransferase domain containing 1', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114056731, end_position_on_the_genomic_accession=114088422, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  79692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79692, CurrentID=0, Status='live', Symbol='ZNF322', Aliases='HCG12A, ZNF388, ZNF489, ZNF322', description='zinc finger protein 322', other_designations='HLA complex group 12|zinc finger protein 322A|zinc finger protein 388|zinc finger protein 489', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26634383, end_position_on_the_genomic_accession=26659752, orientation='minus', exon_count=5, OMIM=610847, no_hdr0=''),
  79693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79693, CurrentID=0, Status='live', Symbol='YRDC', Aliases='DRIP3, IRIP, SUA5', description='yrdC N(6)-threonylcarbamoyltransferase domain containing', other_designations='dopamine receptor interacting protein 3|dopamine receptor-interacting protein 3|hIRIP|ischemia/reperfusion inducible protein|ischemia/reperfusion-inducible protein homolog|yrdC domain containing', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37802942, end_position_on_the_genomic_accession=37808193, orientation='minus', exon_count=5, OMIM=612276, no_hdr0=''),
  79694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79694, CurrentID=0, Status='live', Symbol='MANEA', Aliases='ENDO, hEndo', description='mannosidase, endo-alpha', other_designations='alpha 1,2-endomannosidase|endo-alpha mannosidase|endomannosidase|mandaselin', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=95577497, end_position_on_the_genomic_accession=95609452, orientation='plus', exon_count=6, OMIM=612327, no_hdr0=''),
  79695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79695, CurrentID=0, Status='live', Symbol='GALNT12', Aliases='CRCS1, GalNAc-T12', description='polypeptide N-acetylgalactosaminyltransferase 12', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 12|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 12 (GalNAc-T12)|colorectal cancer, susceptibility to 1|polypeptide GalNAc transferase 12|pp-GaNTase 12|protein-UDP acetylgalactosaminyltransferase 12', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98807699, end_position_on_the_genomic_accession=98850081, orientation='plus', exon_count=13, OMIM=610290, no_hdr0=''),
  79696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79696, CurrentID=0, Status='live', Symbol='ZC2HC1C', Aliases='C14orf140, FAM164C', description='zinc finger, C2HC-type containing 1C', other_designations='family with sequence similarity 164, member C|protein FAM164C', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75069232, end_position_on_the_genomic_accession=75079987, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79697, CurrentID=0, Status='live', Symbol='C14orf169', Aliases='MAPJD, NO66, ROX, URLC2, hsNO66', description='chromosome 14 open reading frame 169', other_designations='60S ribosomal protein L8 histidine hydroxylase|MYC-associated protein with JmjC domain|histone lysine demethylase NO66|lysine-specific demethylase NO66|nucleolar protein 66|ribosomal oxygenase NO66|up-regulated in lung cancer 2', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73490934, end_position_on_the_genomic_accession=73493401, orientation='plus', exon_count=1, OMIM=611919, no_hdr0=''),
  79698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79698, CurrentID=0, Status='live', Symbol='ZMAT4', Aliases='', description='zinc finger, matrin-type 4', other_designations='zinc finger, matrin type 4', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40530590, end_position_on_the_genomic_accession=40897824, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79699, CurrentID=0, Status='live', Symbol='ZYG11B', Aliases='ZYG11', description='zyg-11 family member B, cell cycle regulator', other_designations='zyg-11 homolog B', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52726459, end_position_on_the_genomic_accession=52827342, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100874068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874068, CurrentID=0, Status='live', Symbol='GRTP1-AS1', Aliases='', description='GRTP1 antisense RNA 1', other_designations='GRTP1 antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113351673, end_position_on_the_genomic_accession=113361868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79701, CurrentID=0, Status='live', Symbol='OGFOD3', Aliases='C17orf101', description='2-oxoglutarate and iron-dependent oxygenase domain containing 3', other_designations='PKHD domain-containing transmembrane protein C17orf101', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82389210, end_position_on_the_genomic_accession=82418637, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100874070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874070, CurrentID=0, Status='live', Symbol='USP12-AS1', Aliases='', description='USP12 antisense RNA 1', other_designations='USP12 antisense RNA 1 (non-protein coding)', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27125533, end_position_on_the_genomic_accession=27169135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79703, CurrentID=0, Status='live', Symbol='C11orf80', Aliases='', description='chromosome 11 open reading frame 80', other_designations='putative uncharacterized protein C11orf80', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66744736, end_position_on_the_genomic_accession=66843516, orientation='plus', exon_count=17, OMIM=616109, no_hdr0=''),
  100874072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874072, CurrentID=0, Status='live', Symbol='SMAD9-IT1', Aliases='SMAD9-AS1', description='SMAD9 intronic transcript 1', other_designations='SMAD9 antisense RNA 1 (non-protein coding)', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79705, CurrentID=0, Status='live', Symbol='LRRK1', Aliases='RIPK6, Roco1', description='leucine-rich repeat kinase 1', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100919215, end_position_on_the_genomic_accession=101070112, orientation='plus', exon_count=37, OMIM=610986, no_hdr0=''),
  79706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79706, CurrentID=0, Status='live', Symbol='PRKRIP1', Aliases='C114, KRBOX3', description='PRKR interacting protein 1 (IL11 inducible)', other_designations='KRAB box domain containing 3|likely ortholog of mouse C114 dsRNA-binding protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102396357, end_position_on_the_genomic_accession=102426682, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79707, CurrentID=0, Status='live', Symbol='NOL9', Aliases='Grc3, NET6', description='nucleolar protein 9', other_designations="polynucleotide 5'-kinase", map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6521347, end_position_on_the_genomic_accession=6554598, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  79709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79709, CurrentID=0, Status='live', Symbol='COLGALT1', Aliases='GLT25D1', description='collagen beta(1-O)galactosyltransferase 1', other_designations='glycosyltransferase 25 domain containing 1|glycosyltransferase 25 family member 1|hydroxylysine galactosyltransferase 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17555702, end_position_on_the_genomic_accession=17583162, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  79710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79710, CurrentID=0, Status='live', Symbol='MORC4', Aliases='ZCW4, ZCWCC2, dJ75H8.2', description='MORC family CW-type zinc finger 4', other_designations='zinc finger CW-type coiled-coil domain protein 2|zinc finger, CW type with coiled-coil domain 2', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106940734, end_position_on_the_genomic_accession=107000244, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  79711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79711, CurrentID=0, Status='live', Symbol='IPO4', Aliases='Imp4', description='importin 4', other_designations='imp4b|importin-4b|ran-binding protein 4|ranBP4', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24180219, end_position_on_the_genomic_accession=24188918, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  79712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79712, CurrentID=0, Status='live', Symbol='GTDC1', Aliases='Hmat-Xa, mat-Xa', description='glycosyltransferase like domain containing 1', other_designations='glycosyltransferase-like 1|glycosyltransferase-like domain containing 1|mannosyltransferase candidate|mannosyltransferase-like protein Xa', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143945787, end_position_on_the_genomic_accession=144332534, orientation='minus', exon_count=26, OMIM=610165, no_hdr0=''),
  79713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79713, CurrentID=0, Status='live', Symbol='IGFLR1', Aliases='TMEM149, U2AF1L4', description='IGF like family receptor 1', other_designations='U2 small nuclear RNA auxiliary factor 1-like 4|U2(RNU2) small nuclear RNA auxiliary factor 1-like 4|transmembrane protein 149', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35739250, end_position_on_the_genomic_accession=35742450, orientation='minus', exon_count=5, OMIM=614143, no_hdr0=''),
  79714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79714, CurrentID=0, Status='live', Symbol='CCDC51', Aliases='', description='coiled-coil domain containing 51', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48432170, end_position_on_the_genomic_accession=48441078, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874083, CurrentID=0, Status='live', Symbol='KCNAB1-AS2', Aliases='', description='KCNAB1 antisense RNA 2', other_designations='KCNAB1 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156215560, end_position_on_the_genomic_accession=156227883, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79716, CurrentID=0, Status='live', Symbol='NPEPL1', Aliases='bA261P9.2', description='aminopeptidase-like 1', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58689131, end_position_on_the_genomic_accession=58715844, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  79717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79717, CurrentID=0, Status='live', Symbol='PPCS', Aliases='', description='phosphopantothenoylcysteine synthetase', other_designations='PPC synthetase', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42456341, end_position_on_the_genomic_accession=42473381, orientation='plus', exon_count=5, OMIM=609853, no_hdr0=''),
  79718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79718, CurrentID=0, Status='live', Symbol='TBL1XR1', Aliases='C21, DC42, IRA1, TBLR1', description='transducin (beta)-like 1 X-linked receptor 1', other_designations='TBL1-related protein 1|nuclear receptor co-repressor/HDAC3 complex subunit|nuclear receptor corepressor/HDAC3 complex subunit TBLR1|transducin beta-like 1X-related protein 1', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177020754, end_position_on_the_genomic_accession=177197260, orientation='minus', exon_count=21, OMIM=608628, no_hdr0=''),
  79719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79719, CurrentID=0, Status='live', Symbol='AAGAB', Aliases='KPPP1, PPKP1, PPKP1A, p34', description='alpha- and gamma-adaptin binding protein', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67200675, end_position_on_the_genomic_accession=67255198, orientation='minus', exon_count=13, OMIM=614888, no_hdr0=''),
  79720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79720, CurrentID=0, Status='live', Symbol='VPS37B', Aliases='', description='vacuolar protein sorting 37 homolog B (S. cerevisiae)', other_designations='ESCRT-I complex subunit VPS37B|hVps37B|vacuolar protein sorting 37B', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122865328, end_position_on_the_genomic_accession=122896220, orientation='minus', exon_count=10, OMIM=610037, no_hdr0=''),
  100874089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874089, CurrentID=0, Status='live', Symbol='RUVBL1-AS1', Aliases='', description='RUVBL1 antisense RNA 1', other_designations='RUVBL1 antisense RNA 1 (non-protein coding)', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128075810, end_position_on_the_genomic_accession=128079056, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79722, CurrentID=0, Status='live', Symbol='ANKRD55', Aliases='', description='ankyrin repeat domain 55', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56099680, end_position_on_the_genomic_accession=56233359, orientation='minus', exon_count=13, OMIM=615189, no_hdr0=''),
  79723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79723, CurrentID=0, Status='live', Symbol='SUV39H2', Aliases='KMT1B', description='suppressor of variegation 3-9 homolog 2', other_designations='H3-K9-HMTase 2|histone H3-K9 methyltransferase 2|lysine N-methyltransferase 1B|su(var)3-9 homolog 2', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14878783, end_position_on_the_genomic_accession=14904315, orientation='plus', exon_count=8, OMIM=606503, no_hdr0=''),
  79724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79724, CurrentID=0, Status='live', Symbol='ZNF768', Aliases='', description='zinc finger protein 768', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30524001, end_position_on_the_genomic_accession=30526589, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79725, CurrentID=0, Status='live', Symbol='THAP9', Aliases='hTh9', description='THAP domain containing 9', other_designations='THAP domain-containing protein 9', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82900684, end_position_on_the_genomic_accession=82920283, orientation='plus', exon_count=7, OMIM=612537, no_hdr0=''),
  79726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79726, CurrentID=0, Status='live', Symbol='WDR59', Aliases='FP977', description='WD repeat domain 59', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74873568, end_position_on_the_genomic_accession=74985123, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  79727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79727, CurrentID=0, Status='live', Symbol='LIN28A', Aliases='CSDD1, LIN-28, LIN28, ZCCHC1, lin-28A', description='lin-28 homolog A (C. elegans)', other_designations='RNA-binding protein LIN-28|zinc finger CCHC domain-containing protein 1|zinc finger, CCHC domain containing 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26410778, end_position_on_the_genomic_accession=26429728, orientation='plus', exon_count=5, OMIM=611043, no_hdr0=''),
  79728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79728, CurrentID=0, Status='live', Symbol='PALB2', Aliases='FANCN, PNCA3', description='partner and localizer of BRCA2', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23603162, end_position_on_the_genomic_accession=23641357, orientation='minus', exon_count=15, OMIM=610355, no_hdr0=''),
  79729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79729, CurrentID=0, Status='live', Symbol='SH3D21', Aliases='C1orf113', description='SH3 domain containing 21', other_designations='SH3 domain-containing protein C1orf113', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36305726, end_position_on_the_genomic_accession=36321962, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  79730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79730, CurrentID=0, Status='live', Symbol='NSUN7', Aliases='', description='NOP2/Sun RNA methyltransferase family member 7', other_designations='NOL1/NOP2/Sun domain family member 7|NOL1/NOP2/Sun domain family, member 7|NOP2/Sun domain family, member 7', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40749897, end_position_on_the_genomic_accession=40809985, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  79731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79731, CurrentID=0, Status='live', Symbol='NARS2', Aliases='DFNB94, SLM5, asnRS', description='asparaginyl-tRNA synthetase 2, mitochondrial (putative)', other_designations='asparagine tRNA ligase 2, mitochondrial (putative)|deafness, autosomal recessive 94|probable asparaginyl-tRNA synthetase, mitochondrial', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78435961, end_position_on_the_genomic_accession=78574864, orientation='minus', exon_count=16, OMIM=612803, no_hdr0=''),
  100874100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874100, CurrentID=0, Status='live', Symbol='STEAP2-AS1', Aliases='', description='STEAP2 antisense RNA 1', other_designations='STEAP2 antisense RNA 1 (non-protein coding)', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=89882353, end_position_on_the_genomic_accession=90211635, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79733, CurrentID=0, Status='live', Symbol='E2F8', Aliases='E2F-8', description='E2F transcription factor 8', other_designations='E2F family member 8', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19224063, end_position_on_the_genomic_accession=19241655, orientation='minus', exon_count=14, OMIM=612047, no_hdr0=''),
  79734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79734, CurrentID=0, Status='live', Symbol='KCTD17', Aliases='', description='potassium channel tetramerization domain containing 17', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37051725, end_position_on_the_genomic_accession=37063390, orientation='plus', exon_count=9, OMIM=616386, no_hdr0=''),
  79735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79735, CurrentID=0, Status='live', Symbol='TBC1D17', Aliases='', description='TBC1 domain family member 17', other_designations='TBC1 domain family, member 17', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49877425, end_position_on_the_genomic_accession=49888750, orientation='plus', exon_count=18, OMIM=616659, no_hdr0=''),
  79736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79736, CurrentID=0, Status='live', Symbol='TEFM', Aliases='C17orf42', description='transcription elongation factor, mitochondrial', other_designations='transcription elongation factor of mitochondria', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30898983, end_position_on_the_genomic_accession=30906775, orientation='minus', exon_count=4, OMIM=616422, no_hdr0=''),
  100130452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130452, CurrentID=0, Status='live', Symbol='LOC100130452', Aliases='', description='uncharacterized LOC100130452', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196700634, end_position_on_the_genomic_accession=196713012, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79738, CurrentID=0, Status='live', Symbol='BBS10', Aliases='C12orf58', description='Bardet-Biedl syndrome 10', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76344486, end_position_on_the_genomic_accession=76348442, orientation='minus', exon_count=2, OMIM=610148, no_hdr0=''),
  79739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79739, CurrentID=0, Status='live', Symbol='TTLL7', Aliases='', description='tubulin tyrosine ligase like 7', other_designations='testis development protein NYD-SP30|tubulin tyrosine ligase-like family member 7|tubulin tyrosine ligase-like family, member 7|tubulin--tyrosine ligase-like protein 7', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83869374, end_position_on_the_genomic_accession=83999167, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  79740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79740, CurrentID=0, Status='live', Symbol='ZBBX', Aliases='', description='zinc finger, B-box domain containing', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167178425, end_position_on_the_genomic_accession=167380405, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  79741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79741, CurrentID=0, Status='live', Symbol='CCDC7', Aliases='BIOT2, BioT2-A, BioT2-B, BioT2-C, C10orf68', description='coiled-coil domain containing 7', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32446082, end_position_on_the_genomic_accession=32882864, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  79742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79742, CurrentID=0, Status='live', Symbol='CXorf36', Aliases='4930578C19Rik, DIA1R, EPQL1862, PRO3743, bA435K1.1', description='chromosome X open reading frame 36', other_designations='UPF0672 protein CXorf36|deleted in autism-1 related protein|hCG1981635', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45148373, end_position_on_the_genomic_accession=45200901, orientation='minus', exon_count=6, OMIM=300959, no_hdr0=''),
  100874111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874111, CurrentID=0, Status='live', Symbol='STEAP3-AS1', Aliases='', description='STEAP3 antisense RNA 1', other_designations='STEAP3 antisense RNA 1 (non-protein coding)', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119244422, end_position_on_the_genomic_accession=119249071, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79744, CurrentID=0, Status='live', Symbol='ZNF419', Aliases='ZAPHIRA, ZNF419', description='zinc finger protein 419', other_designations='zinc finger protein 419A', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57487711, end_position_on_the_genomic_accession=57494680, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  79745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79745, CurrentID=0, Status='live', Symbol='CLIP4', Aliases='RSNL2', description='CAP-Gly domain containing linker protein family member 4', other_designations='CAP-GLY domain containing linker protein family, member 4|restin-like protein 2', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29097676, end_position_on_the_genomic_accession=29183813, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  79746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79746, CurrentID=0, Status='live', Symbol='ECHDC3', Aliases='', description='enoyl-CoA hydratase domain containing 3', other_designations='enoyl CoA hydratase domain containing 3|enoyl Coenzyme A hydratase domain containing 3', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11742357, end_position_on_the_genomic_accession=11764070, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  79747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79747, CurrentID=0, Status='live', Symbol='ADGB', Aliases='C6orf103, CAPN16', description='androglobin', other_designations='calpain-7-like protein', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146598965, end_position_on_the_genomic_accession=146815462, orientation='plus', exon_count=43, OMIM=614630, no_hdr0=''),
  79748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79748, CurrentID=0, Status='live', Symbol='LMAN1L', Aliases='ERGIC-53L, ERGL', description='lectin, mannose-binding, 1 like', other_designations='ERGIC-53-like protein|ERGIC53-like protein|LMAN1-like protein|lectin mannose-binding 1-like', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74812853, end_position_on_the_genomic_accession=74825758, orientation='plus', exon_count=14, OMIM=609548, no_hdr0=''),
  100874117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874117, CurrentID=0, Status='live', Symbol='EGFLAM-AS1', Aliases='', description='EGFLAM antisense RNA 1', other_designations='EGFLAM antisense RNA 1 (non-protein coding)', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79750, CurrentID=0, Status='live', Symbol='ZNF385D', Aliases='ZNF659', description='zinc finger protein 385D', other_designations='zinc finger protein 659', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21412222, end_position_on_the_genomic_accession=22372641, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  79751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79751, CurrentID=0, Status='live', Symbol='SLC25A22', Aliases='EIEE3, GC1, NET44', description='solute carrier family 25 (mitochondrial carrier: glutamate), member 22', other_designations='GC-1|glutamate/H(+) symporter 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=790475, end_position_on_the_genomic_accession=798269, orientation='minus', exon_count=13, OMIM=609302, no_hdr0=''),
  79752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79752, CurrentID=0, Status='live', Symbol='ZFAND1', Aliases='', description='zinc finger, AN1-type domain 1', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81701331, end_position_on_the_genomic_accession=81721304, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79753, CurrentID=0, Status='live', Symbol='SNIP1', Aliases='PMRED', description='Smad nuclear interacting protein 1', other_designations='FHA domain-containing protein SNIP1', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37534449, end_position_on_the_genomic_accession=37554344, orientation='minus', exon_count=5, OMIM=608241, no_hdr0=''),
  79754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79754, CurrentID=0, Status='live', Symbol='ASB13', Aliases='', description='ankyrin repeat and SOCS box containing 13', other_designations='ankyrin repeat and SOCS box-containing 13|ankyrin repeat domain-containing SOCS box protein Asb-13', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5638857, end_position_on_the_genomic_accession=5666595, orientation='minus', exon_count=7, OMIM=615055, no_hdr0=''),
  79755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79755, CurrentID=0, Status='live', Symbol='ZNF750', Aliases='ZFP750', description='zinc finger protein 750', other_designations='protein ZNF750', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82829434, end_position_on_the_genomic_accession=82840055, orientation='minus', exon_count=3, OMIM=610226, no_hdr0=''),
  100874124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874124, CurrentID=0, Status='live', Symbol='RPLP1P13', Aliases='', description='ribosomal protein, large, P1 pseudogene 13', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23373423, end_position_on_the_genomic_accession=23374016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79758, CurrentID=0, Status='live', Symbol='DHRS12', Aliases='SDR40C1', description='dehydrogenase/reductase (SDR family) member 12', other_designations='short chain dehydrogenase/reductase family 40C, member 1|short-chain dehydrogenase/reductase family 40C member 1', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51767993, end_position_on_the_genomic_accession=51804199, orientation='minus', exon_count=12, OMIM=616163, no_hdr0=''),
  79759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79759, CurrentID=0, Status='live', Symbol='ZNF668', Aliases='', description='zinc finger protein 668', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31060843, end_position_on_the_genomic_accession=31074320, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79760, CurrentID=0, Status='live', Symbol='GEMIN7', Aliases='SIP3', description='gem nuclear organelle associated protein 7', other_designations='gem (nuclear organelle) associated protein 7|gemin 7|gemin-7', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45075739, end_position_on_the_genomic_accession=45091524, orientation='plus', exon_count=5, OMIM=607419, no_hdr0=''),
  319128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319128, CurrentID=0, Status='live', Symbol='PPIAP6', Aliases='PPIAP12, PPIP2', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 6', other_designations='peptidylprolyl isomerase (cyclophilin) pseudogene 2|peptidylprolyl isomerase A (cyclophilin A) pseudogene 12', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68421492, end_position_on_the_genomic_accession=68422242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79762, CurrentID=0, Status='live', Symbol='C1orf115', Aliases='', description='chromosome 1 open reading frame 115', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220690286, end_position_on_the_genomic_accession=220699157, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79763, CurrentID=0, Status='live', Symbol='ISOC2', Aliases='', description='isochorismatase domain containing 2', other_designations='isochorismatase domain-containing protein 2, mitochondrial', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55452978, end_position_on_the_genomic_accession=55461682, orientation='minus', exon_count=6, OMIM=612928, no_hdr0=''),
  100874132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874132, CurrentID=0, Status='live', Symbol='SRGAP3-AS4', Aliases='', description='SRGAP3 antisense RNA 4', other_designations='SRGAP3 antisense RNA 4 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874134, CurrentID=0, Status='live', Symbol='LINC00345', Aliases='', description='long intergenic non-protein coding RNA 345', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79767, CurrentID=0, Status='live', Symbol='ELMO3', Aliases='CED-12, CED12, ELMO-3', description='engulfment and cell motility 3', other_designations='ced-12 homolog 3', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67199111, end_position_on_the_genomic_accession=67204029, orientation='plus', exon_count=20, OMIM=606422, no_hdr0=''),
  79768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79768, CurrentID=0, Status='live', Symbol='KATNBL1', Aliases='C15orf29', description='katanin p80 subunit B-like 1', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34140674, end_position_on_the_genomic_accession=34210096, orientation='minus', exon_count=14, OMIM=616235, no_hdr0=''),
  100874137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874137, CurrentID=0, Status='live', Symbol='LINC00351', Aliases='', description='long intergenic non-protein coding RNA 351', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85363603, end_position_on_the_genomic_accession=85544662, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79770, CurrentID=0, Status='live', Symbol='TXNDC15', Aliases='C5orf14, UNQ335', description='thioredoxin domain containing 15', other_designations='2310047H23Rik|disulfide isomerase', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134873770, end_position_on_the_genomic_accession=134901633, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874139, CurrentID=0, Status='live', Symbol='LINC00353', Aliases='', description='long intergenic non-protein coding RNA 353', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89548794, end_position_on_the_genomic_accession=89564413, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79772, CurrentID=0, Status='live', Symbol='MCTP1', Aliases='', description='multiple C2 domains, transmembrane 1', other_designations='multiple C2-domains with two transmembrane regions 1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94703690, end_position_on_the_genomic_accession=95284575, orientation='minus', exon_count=28, OMIM=616296, no_hdr0=''),
  319130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319130, CurrentID=0, Status='live', Symbol='HMGB1P14', Aliases='HMGB1L14, HMGB1P', description='high mobility group box 1 pseudogene 14', other_designations='high-mobility group box 1 pseudogene 14|high-mobility group box 1-like 14', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58284709, end_position_on_the_genomic_accession=58285833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79774, CurrentID=0, Status='live', Symbol='GRTP1', Aliases='TBC1D6', description='growth hormone regulated TBC protein 1', other_designations='TBC1 domain family member 6', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113324164, end_position_on_the_genomic_accession=113364148, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  386734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386734, CurrentID=0, Status='live', Symbol='CDY12P', Aliases='', description='chromodomain protein, Y-linked 12 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22064597, end_position_on_the_genomic_accession=22065805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79776, CurrentID=0, Status='live', Symbol='ZFHX4', Aliases='ZFH4, ZHF4', description='zinc finger homeobox 4', other_designations='zinc-finger homeodomain protein 4', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76681219, end_position_on_the_genomic_accession=76867285, orientation='plus', exon_count=15, OMIM=606940, no_hdr0=''),
  79777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79777, CurrentID=0, Status='live', Symbol='ACBD4', Aliases='HMFT0700', description='acyl-CoA binding domain containing 4', other_designations='acyl-Coenzyme A binding domain containing 4', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45132600, end_position_on_the_genomic_accession=45144181, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  79778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79778, CurrentID=0, Status='live', Symbol='MICALL2', Aliases='JRAB, MICAL-L2', description='MICAL-like 2', other_designations='junctional Rab13-binding protein|molecule interacting with CasL-like 2', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1428465, end_position_on_the_genomic_accession=1459497, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  319131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319131, CurrentID=0, Status='live', Symbol='HMGN1P1', Aliases='HMGN1L1', description='high mobility group nucleosome binding domain 1 pseudogene 1', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 1|high-mobility group nucleosome binding domain 1-like 1', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55394315, end_position_on_the_genomic_accession=55395327, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79780, CurrentID=0, Status='live', Symbol='CCDC82', Aliases='HSPC048', description='coiled-coil domain containing 82', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96352765, end_position_on_the_genomic_accession=96389931, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  79781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79781, CurrentID=0, Status='live', Symbol='IQCA1', Aliases='4930465P12Rik, DRC11, IQCA', description='IQ motif containing with AAA domain 1', other_designations='dynein regulatory complex subunit 11', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236324147, end_position_on_the_genomic_accession=236507552, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  79782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79782, CurrentID=0, Status='live', Symbol='LRRC31', Aliases='HEL-S-293', description='leucine rich repeat containing 31', other_designations='epididymis secretory protein Li 293', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169839241, end_position_on_the_genomic_accession=169869935, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79783, CurrentID=0, Status='live', Symbol='SUGCT', Aliases='C7orf10, DERP13, ORF19', description='succinyl-CoA:glutarate-CoA transferase', other_designations='Russel-Silver syndrome candidate|dermal papilla-derived protein 13|succinylCoA:glutarate-CoA transferase', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40134965, end_position_on_the_genomic_accession=40860767, orientation='plus', exon_count=19, OMIM=609187, no_hdr0=''),
  79784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79784, CurrentID=0, Status='live', Symbol='MYH14', Aliases='DFNA4, DFNA4A, FP17425, MHC16, MYH17, NMHC II-C, NMHC-II-C, PNMHH, myosin', description='myosin, heavy chain 14, non-muscle', other_designations='MYH14 variant protein|myosin heavy chain, non-muscle IIc|myosin, heavy polypeptide 14|non-muscle myosin heavy chain IIc|nonmuscle myosin heavy chain II-C', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50188186, end_position_on_the_genomic_accession=50310544, orientation='plus', exon_count=46, OMIM=608568, no_hdr0=''),
  79785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79785, CurrentID=0, Status='live', Symbol='RERGL', Aliases='', description='RERG/RAS-like', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18080869, end_position_on_the_genomic_accession=18090193, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79786, CurrentID=0, Status='live', Symbol='KLHL36', Aliases='C16orf44', description='kelch like family member 36', other_designations='kelch-like 36|kelch-like family member 36', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84648511, end_position_on_the_genomic_accession=84667686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874155, CurrentID=0, Status='live', Symbol='LINC00392', Aliases='', description='long intergenic non-protein coding RNA 392', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73564244, end_position_on_the_genomic_accession=73587879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79788, CurrentID=0, Status='live', Symbol='ZNF665', Aliases='ZFP160L', description='zinc finger protein 665', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53163299, end_position_on_the_genomic_accession=53193396, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79789, CurrentID=0, Status='live', Symbol='CLMN', Aliases='', description='calmin (calponin-like, transmembrane)', other_designations='calponin like transmembrane domain protein|calponin-like transmembrane domain protein', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95181939, end_position_on_the_genomic_accession=95319908, orientation='minus', exon_count=15, OMIM=611121, no_hdr0=''),
  100874158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874158, CurrentID=0, Status='live', Symbol='LINC00398', Aliases='', description='long intergenic non-protein coding RNA 398', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30803206, end_position_on_the_genomic_accession=30810645, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79791, CurrentID=0, Status='live', Symbol='FBXO31', Aliases='FBX14, FBXO14, Fbx31, MRT45, pp2386', description='F-box protein 31', other_designations='SCF ubiquitin ligase specificity factor|putative breast cancer tumor-suppressor', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87326987, end_position_on_the_genomic_accession=87392107, orientation='minus', exon_count=11, OMIM=609102, no_hdr0=''),
  79792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79792, CurrentID=0, Status='live', Symbol='GSDMD', Aliases='DF5L, DFNA5L, FKSG10C1, GSDMD', description='gasdermin D', other_designations='gasdermin domain containing 1|gasdermin domain-containing protein 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143553387, end_position_on_the_genomic_accession=143563062, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100874161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874161, CurrentID=0, Status='live', Symbol='LINC00411', Aliases='', description='long intergenic non-protein coding RNA 411', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100939520, end_position_on_the_genomic_accession=100944364, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79794, CurrentID=0, Status='live', Symbol='C12orf49', Aliases='', description='chromosome 12 open reading frame 49', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116713320, end_position_on_the_genomic_accession=116738070, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874163, CurrentID=0, Status='live', Symbol='LINC00415', Aliases='', description='long intergenic non-protein coding RNA 415', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79796, CurrentID=0, Status='live', Symbol='ALG9', Aliases='CDG1L, DIBD1, LOH11CR1J', description='ALG9, alpha-1,2-mannosyltransferase', other_designations='asparagine-linked glycosylation 9 alpha-12-mannosyltransferase-like protein|asparagine-linked glycosylation 9 homolog (S. cerevisiae, alpha- 1,2-mannosyltransferase)|asparagine-linked glycosylation 9 homolog (yeast, alpha- 1,2-mannosyltransferase)|asparagine-linked glycosylation 9, alpha-1,2-mannosyltransferase homolog|asparagine-linked glycosylation protein 9 homolog|disrupted in bipolar affective disorder 1|disrupted in bipolar disorder protein 1|dol-P-Man dependent alpha-1,2-mannosyltransferase|dol-P-Man:Man(6)GlcNAc(2)-PP-Dol alpha-1,2-mannosyltransferase|dol-P-Man:Man(8)GlcNAc(2)-PP-Dol alpha-1,2-mannosyltransferase|dolichyl-P-Man:Man(6)GlcNAc(2)-PP-dolichol alpha-1,2-mannosyltransferase|dolichyl-P-Man:Man(8)GlcNAc(2)-PP-dolichol alpha-1,2-mannosyltransferase|loss of heterozygosity, 11, chromosomal region 1 gene J product', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111782195, end_position_on_the_genomic_accession=111871581, orientation='minus', exon_count=22, OMIM=606941, no_hdr0=''),
  79797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79797, CurrentID=0, Status='live', Symbol='ZNF408', Aliases='', description='zinc finger protein 408', other_designations='PR domain zinc finger protein 17', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46700767, end_position_on_the_genomic_accession=46705916, orientation='plus', exon_count=5, OMIM=616454, no_hdr0=''),
  79798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79798, CurrentID=0, Status='live', Symbol='ARMC5', Aliases='AIMAH2', description='armadillo repeat containing 5', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31458273, end_position_on_the_genomic_accession=31467167, orientation='plus', exon_count=8, OMIM=615549, no_hdr0=''),
  79799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79799, CurrentID=0, Status='live', Symbol='UGT2A3', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3', other_designations='UDPGT 2A3', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68928459, end_position_on_the_genomic_accession=68951798, orientation='minus', exon_count=7, OMIM=616382, no_hdr0=''),
  79800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79800, CurrentID=0, Status='live', Symbol='CARF', Aliases='ALS2CR8, NYD-SP24', description='calcium responsive transcription factor', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 8|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 8 protein|calcium-response factor|testis development protein NYD-SP24', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202912218, end_position_on_the_genomic_accession=202987857, orientation='plus', exon_count=22, OMIM=607586, no_hdr0=''),
  79801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79801, CurrentID=0, Status='live', Symbol='SHCBP1', Aliases='PAL', description='SHC SH2-domain binding protein 1', other_designations='Protein expressed in Activated Lymphocytes', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46580554, end_position_on_the_genomic_accession=46622810, orientation='minus', exon_count=15, OMIM=611027, no_hdr0=''),
  79802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79802, CurrentID=0, Status='live', Symbol='HHIPL2', Aliases='KIAA1822L', description='HHIP-like 2', other_designations='hedgehog interacting protein-like 2', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222522260, end_position_on_the_genomic_accession=222548102, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  341947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341947, CurrentID=0, Status='live', Symbol='COX8C', Aliases='COX8-3', description='cytochrome c oxidase subunit VIIIC', other_designations='COX VIII-3|cytochrome c oxidase polypeptide 8|cytochrome c oxidase polypeptide VIII|cytochrome c oxidase subunit 8-3|cytochrome c oxidase subunit VIII', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93347191, end_position_on_the_genomic_accession=93348354, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79804, CurrentID=0, Status='live', Symbol='HAND2-AS1', Aliases='DEIN, NBLA00301', description='HAND2 antisense RNA 1 (head to head)', other_designations='differentially expressed in neuroblastoma|neuroblastoma transcript 301', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173530458, end_position_on_the_genomic_accession=173541830, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79805, CurrentID=0, Status='live', Symbol='VASH2', Aliases='', description='vasohibin 2', other_designations='vasohibin-like protein', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212935543, end_position_on_the_genomic_accession=212991585, orientation='plus', exon_count=25, OMIM=610471, no_hdr0=''),
  100874174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874174, CurrentID=0, Status='live', Symbol='LINC00444', Aliases='', description='long intergenic non-protein coding RNA 444', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79807, CurrentID=0, Status='live', Symbol='GSTCD', Aliases='', description='glutathione S-transferase, C-terminal domain containing', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105708778, end_position_on_the_genomic_accession=105847725, orientation='plus', exon_count=16, OMIM=615912, no_hdr0=''),
  100874176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874176, CurrentID=0, Status='live', Symbol='LINC00448', Aliases='', description='long intergenic non-protein coding RNA 448', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62672285, end_position_on_the_genomic_accession=62807359, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  79809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79809, CurrentID=0, Status='live', Symbol='TTC21B', Aliases='ATD4, IFT139, JBTS11, NPHP12, Nbla10696, SRTD4, THM1', description='tetratricopeptide repeat domain 21B', other_designations='TPR repeat protein 21B|putative protein product of Nbla10696', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165873362, end_position_on_the_genomic_accession=165953838, orientation='minus', exon_count=31, OMIM=612014, no_hdr0=''),
  79810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79810, CurrentID=0, Status='live', Symbol='PTCD2', Aliases='', description='pentatricopeptide repeat domain 2', other_designations='CTC-365E16.1', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72320367, end_position_on_the_genomic_accession=72368397, orientation='plus', exon_count=11, OMIM=615484, no_hdr0=''),
  79811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79811, CurrentID=0, Status='live', Symbol='SLTM', Aliases='Met', description='SAFB-like, transcription modulator', other_designations='modulator of estrogen-induced transcription', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58879045, end_position_on_the_genomic_accession=58934313, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  79812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79812, CurrentID=0, Status='live', Symbol='MMRN2', Aliases='EMILIN-3, EMILIN3, ENDOGLYX-1', description='multimerin 2', other_designations='EMILIN-like protein EndoGlyx-1|elastin microfibril interface located protein 3|elastin microfibril interfacer 3|endoGlyx-1 p125/p140 subunit', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86935540, end_position_on_the_genomic_accession=86969953, orientation='minus', exon_count=8, OMIM=608925, no_hdr0=''),
  79813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79813, CurrentID=0, Status='live', Symbol='EHMT1', Aliases='EUHMTASE1, Eu-HMTase1, FP13812, GLP, GLP1, KMT1D, bA188C12.1', description='euchromatic histone-lysine N-methyltransferase 1', other_designations='G9a-like protein 1|H3-K9-HMTase 5|histone H3-K9 methyltransferase 5|histone-lysine N-methyltransferase, H3 lysine-9 specific 5|lysine N-methyltransferase 1D', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137618992, end_position_on_the_genomic_accession=137836127, orientation='plus', exon_count=37, OMIM=607001, no_hdr0=''),
  79814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79814, CurrentID=0, Status='live', Symbol='AGMAT', Aliases='', description='agmatinase', other_designations='AUH|agmatine ureohydrolase (agmatinase)', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15571699, end_position_on_the_genomic_accession=15585110, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79815, CurrentID=0, Status='live', Symbol='NIPAL2', Aliases='NPAL2', description='NIPA-like domain containing 2', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98189826, end_position_on_the_genomic_accession=98295442, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  79816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79816, CurrentID=0, Status='live', Symbol='TLE6', Aliases='GRG6', description='transducin like enhancer of split 6', other_designations='transducin-like enhancer of split 6 (E(sp1) homolog, Drosophila)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2977415, end_position_on_the_genomic_accession=2995184, orientation='plus', exon_count=21, OMIM=612399, no_hdr0=''),
  79817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79817, CurrentID=0, Status='live', Symbol='MOB3B', Aliases='C9orf35, MOB1D, MOBKL2B', description='MOB kinase activator 3B', other_designations='MOB1, Mps One Binder kinase activator-like 2B|mob1 homolog 2b|monopolar spindle 1 binding, MOB1, domain containing|mps one binder kinase activator-like 2B', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27325209, end_position_on_the_genomic_accession=27529852, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79818, CurrentID=0, Status='live', Symbol='ZNF552', Aliases='', description='zinc finger protein 552', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57807082, end_position_on_the_genomic_accession=57815188, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79819, CurrentID=0, Status='live', Symbol='WDR78', Aliases='DIC4', description='WD repeat domain 78', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66812885, end_position_on_the_genomic_accession=66925095, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  79820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79820, CurrentID=0, Status='live', Symbol='CATSPERB', Aliases='C14orf161, CatSper(beta)', description='catsper channel auxiliary subunit beta', other_designations='catSper-beta|cation channel, sperm associated beta|cation channel, sperm-associated, beta', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91580774, end_position_on_the_genomic_accession=91732086, orientation='minus', exon_count=27, OMIM=611169, no_hdr0=''),
  100874189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874189, CurrentID=0, Status='live', Symbol='NKX2-2-AS1', Aliases='NKX2-2AS', description='NKX2-2 antisense RNA 1', other_designations='NKX2-2 antisense RNA 1 (non-protein coding)', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79822, CurrentID=0, Status='live', Symbol='ARHGAP28', Aliases='', description='Rho GTPase activating protein 28', other_designations='rho-type GTPase-activating protein 28', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6729718, end_position_on_the_genomic_accession=6915716, orientation='plus', exon_count=20, OMIM=610592, no_hdr0=''),
  79823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79823, CurrentID=0, Status='live', Symbol='CAMKMT', Aliases='C2orf34, CLNMT, CaM KMT, Cam, KMT', description='calmodulin-lysine N-methyltransferase', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44361904, end_position_on_the_genomic_accession=44772592, orientation='plus', exon_count=17, OMIM=609559, no_hdr0=''),
  100874192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874192, CurrentID=0, Status='live', Symbol='RANP7', Aliases='', description='RAN, member RAS oncogene family pseudogene 7', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=22904653, end_position_on_the_genomic_accession=22905338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79825, CurrentID=0, Status='live', Symbol='EFCC1', Aliases='C3orf73, CCDC48', description='EF-hand and coiled-coil domain containing 1', other_designations='EF-hand domain-containing protein ENSP00000381169|coiled-coil domain containing 48|coiled-coil domain-containing protein 48', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129001251, end_position_on_the_genomic_accession=129040742, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100874194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874194, CurrentID=0, Status='live', Symbol='CLDN10-AS1', Aliases='', description='CLDN10 antisense RNA 1', other_designations='CLDN10 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95479444, end_position_on_the_genomic_accession=95533910, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79827, CurrentID=0, Status='live', Symbol='CLMP', Aliases='ACAM, ASAM, CSBM, CSBS', description='CXADR-like membrane protein', other_designations='CAR-like membrane protein|adipocyte-specific adhesion molecule|coxsackie- and adenovirus receptor-like membrane protein', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123072312, end_position_on_the_genomic_accession=123195299, orientation='minus', exon_count=7, OMIM=611693, no_hdr0=''),
  79828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79828, CurrentID=0, Status='live', Symbol='METTL8', Aliases='TIP', description='methyltransferase like 8', other_designations='tension-induced/inhibited protein', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171317402, end_position_on_the_genomic_accession=171434802, orientation='minus', exon_count=13, OMIM=609525, no_hdr0=''),
  79829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79829, CurrentID=0, Status='live', Symbol='NAA40', Aliases='NAT11, PATT1', description='N(alpha)-acetyltransferase 40, NatD catalytic subunit', other_designations='N(alpha)-acetyltransferase 40, NatD catalytic subunit, homolog|N-acetyltransferase 11 (GCN5-related, putative)|N-alpha acetyl transferase 40|N-alpha-acetyltransferase 40, NatD catalytic subunit|natD catalytic subunit', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63938970, end_position_on_the_genomic_accession=63957319, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  79830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79830, CurrentID=0, Status='live', Symbol='ZMYM1', Aliases='MYM', description='zinc finger, MYM-type 1', other_designations='zinc finger, MYM domain containing 1', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35059786, end_position_on_the_genomic_accession=35115859, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  79831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79831, CurrentID=0, Status='live', Symbol='KDM8', Aliases='JMJD5', description='lysine (K)-specific demethylase 8', other_designations='jmjC domain-containing protein 5|jumonji domain containing 5|jumonji domain-containing protein 5', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27203486, end_position_on_the_genomic_accession=27222305, orientation='plus', exon_count=10, OMIM=611917, no_hdr0=''),
  79832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79832, CurrentID=0, Status='live', Symbol='QSER1', Aliases='', description='glutamine and serine rich 1', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32893178, end_position_on_the_genomic_accession=32980270, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  79833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79833, CurrentID=0, Status='live', Symbol='GEMIN6', Aliases='', description='gem nuclear organelle associated protein 6', other_designations='SIP2|gem (nuclear organelle) associated protein 6|gemin 6|gemin-6', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38778185, end_position_on_the_genomic_accession=38782456, orientation='plus', exon_count=4, OMIM=607006, no_hdr0=''),
  79834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79834, CurrentID=0, Status='live', Symbol='PEAK1', Aliases='SGK269', description='pseudopodium-enriched atypical kinase 1', other_designations='NKF3 kinase family member|sugen kinase 269|tyrosine-protein kinase SgK269', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77100829, end_position_on_the_genomic_accession=77420104, orientation='minus', exon_count=19, OMIM=614248, no_hdr0=''),
  100874203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874203, CurrentID=0, Status='live', Symbol='COL4A2-AS1', Aliases='', description='COL4A2 antisense RNA 1', other_designations='COL4A2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110502576, end_position_on_the_genomic_accession=110508179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79836, CurrentID=0, Status='live', Symbol='LONRF3', Aliases='RNF127', description='LON peptidase N-terminal domain and ring finger 3', other_designations='ring finger protein 127', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118974614, end_position_on_the_genomic_accession=119018856, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  79837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79837, CurrentID=0, Status='live', Symbol='PIP4K2C', Aliases='PIP5K2C', description='phosphatidylinositol-5-phosphate 4-kinase, type II, gamma', other_designations='PI(5)P 4-kinase type II gamma|PIP4KII-gamma|phosphatidylinositol 4-phosphate 5-kinase|phosphatidylinositol 5-phosphate 4-kinase type II gamma|phosphatidylinositol-4-phosphate 5-kinase, type II, gamma|phosphatidylinositol-5-phosphate 4-kinase type-2 gamma', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57591159, end_position_on_the_genomic_accession=57603428, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  79838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79838, CurrentID=0, Status='live', Symbol='TMC5', Aliases='', description='transmembrane channel like 5', other_designations='transmembrane channel-like 5', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19410522, end_position_on_the_genomic_accession=19499113, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  79839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79839, CurrentID=0, Status='live', Symbol='CCDC102B', Aliases='ACY1L, AN, C18orf14, HsT1731', description='coiled-coil domain containing 102B', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68715254, end_position_on_the_genomic_accession=69171081, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  79840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79840, CurrentID=0, Status='live', Symbol='NHEJ1', Aliases='XLF', description='non-homologous end joining factor 1', other_designations='XRCC4-like factor|nonhomologous end-joining factor 1|protein cernunnos', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219075324, end_position_on_the_genomic_accession=219160865, orientation='minus', exon_count=8, OMIM=611290, no_hdr0=''),
  79841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79841, CurrentID=0, Status='live', Symbol='AGBL2', Aliases='CCP2', description='ATP/GTP binding protein-like 2', other_designations='cytoplasmic carboxypeptidase 2', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47659169, end_position_on_the_genomic_accession=47715382, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  79842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79842, CurrentID=0, Status='live', Symbol='ZBTB3', Aliases='', description='zinc finger and BTB domain containing 3', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62750963, end_position_on_the_genomic_accession=62754184, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79843, CurrentID=0, Status='live', Symbol='FAM124B', Aliases='', description='family with sequence similarity 124 member B', other_designations='family with sequence similarity 124B', map_location='2q36.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224378698, end_position_on_the_genomic_accession=224401994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79844, CurrentID=0, Status='live', Symbol='ZDHHC11', Aliases='ZNF399', description='zinc finger, DHHC-type containing 11', other_designations='DHHC-11|zinc finger DHHC domain-containing protein 11|zinc finger protein 399|zinc finger, DHHC domain containing 11', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=795605, end_position_on_the_genomic_accession=858973, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  79845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79845, CurrentID=0, Status='live', Symbol='RNF122', Aliases='', description='ring finger protein 122', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33547754, end_position_on_the_genomic_accession=33567128, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79846, CurrentID=0, Status='live', Symbol='CFAP69', Aliases='C7orf63, FAP69', description='cilia and flagella associated protein 69', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90245174, end_position_on_the_genomic_accession=90324208, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  79847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79847, CurrentID=0, Status='live', Symbol='MFSD13A', Aliases='C10orf77, TMEM180, bA18I14.8', description='major facilitator superfamily domain containing 13A', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102461381, end_position_on_the_genomic_accession=102480118, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  79848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79848, CurrentID=0, Status='live', Symbol='CSPP1', Aliases='CSPP, JBTS21', description='centrosome and spindle pole associated protein 1', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67062417, end_position_on_the_genomic_accession=67196614, orientation='plus', exon_count=37, OMIM=611654, no_hdr0=''),
  79849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79849, CurrentID=0, Status='live', Symbol='PDZD3', Aliases='IKEPP, NHERF4, PDZK2', description='PDZ domain containing 3', other_designations='NHERF-4|PDZ domain containing 2|PDZ domain-containing protein 2|PDZ domain-containing protein 3|intestinal and kidney enriched PDZ protein|intestinal and kidney-enriched PDZ protein|na/Pi cotransporter C-terminal-associated protein 2|naPi-Cap2|natrium-phosphate cotransporter IIa C-terminal-associated protein 2|sodium-hydrogen exchanger regulatory factor 4', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119185436, end_position_on_the_genomic_accession=119190223, orientation='plus', exon_count=12, OMIM=607146, no_hdr0=''),
  79850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79850, CurrentID=0, Status='live', Symbol='FAM57A', Aliases='CT120', description='family with sequence similarity 57 member A', other_designations='family with sequence similarity 57, member A|membrane protein expressed in epithelial-like lung adenocarcinoma', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=732607, end_position_on_the_genomic_accession=742835, orientation='plus', exon_count=5, OMIM=611627, no_hdr0=''),
  319143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=319143, CurrentID=0, Status='live', Symbol='RPL21P10', Aliases='RPL21_49_1395', description='ribosomal protein L21 pseudogene 10', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77683190, end_position_on_the_genomic_accession=77684027, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79852, CurrentID=0, Status='live', Symbol='EPHX3', Aliases='ABHD9, EH3', description='epoxide hydrolase 3', other_designations='abhydrolase domain containing 9|abhydrolase domain-containing protein 9', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15226919, end_position_on_the_genomic_accession=15233047, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  79853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79853, CurrentID=0, Status='live', Symbol='TM4SF20', Aliases='PRO994, SLI5, TCCE518', description='transmembrane 4 L six family member 20', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227362156, end_position_on_the_genomic_accession=227379306, orientation='minus', exon_count=5, OMIM=615404, no_hdr0=''),
  100874222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874222, CurrentID=0, Status='live', Symbol='RNF219-AS1', Aliases='POU4F1-AS1', description='RNF219 antisense RNA 1', other_designations='POU4F1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78054855, end_position_on_the_genomic_accession=78617325, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874223, CurrentID=0, Status='live', Symbol='FTOP1', Aliases='', description='fat mass and obesity associated pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42797225, end_position_on_the_genomic_accession=42798712, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79856, CurrentID=0, Status='live', Symbol='SNX22', Aliases='', description='sorting nexin 22', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64151698, end_position_on_the_genomic_accession=64157481, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79857, CurrentID=0, Status='live', Symbol='FLJ13224', Aliases='', description='uncharacterized LOC79857', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31324316, end_position_on_the_genomic_accession=31325945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79858, CurrentID=0, Status='live', Symbol='NEK11', Aliases='', description='NIMA-related kinase 11', other_designations='NIMA (never in mitosis gene a)- related kinase 11', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131026850, end_position_on_the_genomic_accession=131350465, orientation='plus', exon_count=30, OMIM=609779, no_hdr0=''),
  100874227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874227, CurrentID=0, Status='live', Symbol='RANP3', Aliases='', description='RAN, member RAS oncogene family pseudogene 3', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74652562, end_position_on_the_genomic_accession=74653111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79860, CurrentID=0, Status='live', Symbol='FLJ21369', Aliases='', description='uncharacterized protein FLJ21369', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=34767283, end_position_on_the_genomic_accession=34769349, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79861, CurrentID=0, Status='live', Symbol='TUBAL3', Aliases='', description='tubulin alpha like 3', other_designations='tubulin, alpha-like 3', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5393098, end_position_on_the_genomic_accession=5404830, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79862, CurrentID=0, Status='live', Symbol='ZNF669', Aliases='', description='zinc finger protein 669', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247099962, end_position_on_the_genomic_accession=247104400, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79863, CurrentID=0, Status='live', Symbol='RBFA', Aliases='C18orf22, HsT169', description='ribosome binding factor A (putative)', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80034346, end_position_on_the_genomic_accession=80050652, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79864, CurrentID=0, Status='live', Symbol='C11orf63', Aliases='', description='chromosome 11 open reading frame 63', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122881639, end_position_on_the_genomic_accession=122959722, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  79865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79865, CurrentID=0, Status='live', Symbol='TREML2', Aliases='C6orf76, TLT-2, TLT2, dJ238O23.1', description='triggering receptor expressed on myeloid cells like 2', other_designations='triggering receptor expressed on myeloid cells-like 2|triggering receptor expressed on myeloid cells-like protein 2', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41189749, end_position_on_the_genomic_accession=41201192, orientation='minus', exon_count=5, OMIM=609715, no_hdr0=''),
  79866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79866, CurrentID=0, Status='live', Symbol='BORA', Aliases='C13orf34', description='bora, aurora kinase A activator', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72727749, end_position_on_the_genomic_accession=72756198, orientation='plus', exon_count=12, OMIM=610510, no_hdr0=''),
  79867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79867, CurrentID=0, Status='live', Symbol='TCTN2', Aliases='C12orf38, MKS8, TECT2', description='tectonic family member 2', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123671113, end_position_on_the_genomic_accession=123708405, orientation='plus', exon_count=18, OMIM=613846, no_hdr0=''),
  79868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79868, CurrentID=0, Status='live', Symbol='ALG13', Aliases='CDG1S, CXorf45, GLT28D1, MDS031, TDRD13, YGL047W', description='ALG13, UDP-N-acetylglucosaminyltransferase subunit', other_designations='N-acetylglucosaminyldiphosphodolichol N-acetylglucosaminyltransferase|UDP-N-acetylglucosamine transferase subunit ALG13 homolog|asparagine-linked glycosylation 13 homolog|glycosyltransferase 28 domain-containing protein 1|hematopoietic stem/progenitor cells protein MDS031|tudor domain containing 13', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111680742, end_position_on_the_genomic_accession=111764164, orientation='plus', exon_count=33, OMIM=300776, no_hdr0=''),
  79869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79869, CurrentID=0, Status='live', Symbol='CPSF7', Aliases='CFIm59', description='cleavage and polyadenylation specific factor 7', other_designations='CPSF 59 kDa subunit|cleavage and polyadenylation specific factor 7, 59kDa|cleavage and polyadenylation specificity factor 59 kDa subunit|cleavage factor Im complex 59 kDa subunit|pre mRNA cleavage factor I, 59 kDa subunit|pre-mRNA cleavage factor I, 59 kDa subunit|pre-mRNA cleavage factor Im 59 kDa subunit', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61402648, end_position_on_the_genomic_accession=61429992, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  79870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79870, CurrentID=0, Status='live', Symbol='BAALC', Aliases='', description='brain and acute leukemia, cytoplasmic', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103140693, end_position_on_the_genomic_accession=103230305, orientation='plus', exon_count=3, OMIM=606602, no_hdr0=''),
  79871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79871, CurrentID=0, Status='live', Symbol='RPAP2', Aliases='C1orf82, Rtr1', description='RNA polymerase II associated protein 2', other_designations='RNA polymerase II-associated protein 2', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92298965, end_position_on_the_genomic_accession=92388175, orientation='plus', exon_count=17, OMIM=611476, no_hdr0=''),
  79872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79872, CurrentID=0, Status='live', Symbol='CBLL1', Aliases='HAKAI, RNF188', description='Cbl proto-oncogene-like 1, E3 ubiquitin protein ligase', other_designations='Cas-Br-M (murine) ecotropic retroviral transforming sequence-like 1|Cbl proto-oncogene, E3 ubiquitin protein ligase-like 1|E-cadherin binding protein E7|RING finger protein 188|c-Cbl-like protein 1|casitas B-lineage lymphoma-transforming sequence-like protein 1', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107743497, end_position_on_the_genomic_accession=107762011, orientation='plus', exon_count=7, OMIM=606872, no_hdr0=''),
  79873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79873, CurrentID=0, Status='live', Symbol='NUDT18', Aliases='MTH3', description='nudix hydrolase 18', other_designations='2-hydroxy-dADP phosphatase|7,8-dihydro-8-oxoguanine phosphatase|mutT homolog 3|mutT human homolog 3|nucleoside diphosphate-linked moiety X motif 18|nudix (nucleoside diphosphate linked moiety X)-type motif 18|nudix motif 18', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22106872, end_position_on_the_genomic_accession=22110355, orientation='minus', exon_count=4, OMIM=615791, no_hdr0=''),
  79874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79874, CurrentID=0, Status='live', Symbol='RABEP2', Aliases='FRA', description='rabaptin, RAB GTPase binding effector protein 2', other_designations='Rabaptin-5 beta|rabaptin-5beta', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28904421, end_position_on_the_genomic_accession=28925211, orientation='minus', exon_count=13, OMIM=611869, no_hdr0=''),
  79875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79875, CurrentID=0, Status='live', Symbol='THSD4', Aliases='ADAMTSL-6, ADAMTSL6, FVSY9334, PRO34005', description='thrombospondin type 1 domain containing 4', other_designations='A disintegrin and metalloproteinase with thrombospondin motifs-like protein 6|ADAMTS-like protein 6|thrombospondin, type I, domain containing 4', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71115628, end_position_on_the_genomic_accession=71783383, orientation='plus', exon_count=22, OMIM=614476, no_hdr0=''),
  79876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79876, CurrentID=0, Status='live', Symbol='UBA5', Aliases='THIFP1, UBE1DC1', description='ubiquitin-like modifier activating enzyme 5', other_designations='UBA5, ubiquitin-activating enzyme E1 homolog|UFM1-activating enzyme|ubiquitin-activating enzyme 5|ubiquitin-activating enzyme E1 domain-containing protein 1|ubiquitin-activating enzyme E1-domain containing 1', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132654446, end_position_on_the_genomic_accession=132678253, orientation='plus', exon_count=15, OMIM=610552, no_hdr0=''),
  79877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79877, CurrentID=0, Status='live', Symbol='DCAKD', Aliases='', description='dephospho-CoA kinase domain containing', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45023338, end_position_on_the_genomic_accession=45061141, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100874246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874246, CurrentID=0, Status='live', Symbol='ARHGAP31-AS1', Aliases='', description='ARHGAP31 antisense RNA 1', other_designations='ARHGAP31 antisense RNA 1 (non-protein coding)', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119314293, end_position_on_the_genomic_accession=119322760, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79879, CurrentID=0, Status='live', Symbol='CCDC134', Aliases='', description='coiled-coil domain containing 134', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41800622, end_position_on_the_genomic_accession=41832164, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100874248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874248, CurrentID=0, Status='live', Symbol='USP24P1', Aliases='', description='ubiquitin specific peptidase 24 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19015487, end_position_on_the_genomic_accession=19019336, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874249, CurrentID=0, Status='live', Symbol='DENND5B-AS1', Aliases='', description='DENND5B antisense RNA 1', other_designations='DENND5B antisense RNA 1 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31589923, end_position_on_the_genomic_accession=31615351, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79882, CurrentID=0, Status='live', Symbol='ZC3H14', Aliases='MSUT-2, NY-REN-37, SUT2, UKp68', description='zinc finger CCCH-type containing 14', other_designations='mammalian suppressor of tau pathology-2|nuclear protein UKp68|renal carcinoma antigen NY-REN-37', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88562909, end_position_on_the_genomic_accession=88613509, orientation='plus', exon_count=20, OMIM=613279, no_hdr0=''),
  79883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79883, CurrentID=0, Status='live', Symbol='PODNL1', Aliases='SLRR5B', description='podocan-like 1', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13931187, end_position_on_the_genomic_accession=13953392, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  79884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79884, CurrentID=0, Status='live', Symbol='MAP9', Aliases='ASAP', description='microtubule associated protein 9', other_designations='aster-associated protein', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155342658, end_position_on_the_genomic_accession=155376970, orientation='minus', exon_count=15, OMIM=610070, no_hdr0=''),
  79885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79885, CurrentID=0, Status='live', Symbol='HDAC11', Aliases='HD11', description='histone deacetylase 11', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13480171, end_position_on_the_genomic_accession=13506424, orientation='plus', exon_count=12, OMIM=607226, no_hdr0=''),
  79886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79886, CurrentID=0, Status='live', Symbol='CAAP1', Aliases='C9orf82, CAAP', description='caspase activity and apoptosis inhibitor 1', other_designations='conserved anti-apoptotic protein', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26840685, end_position_on_the_genomic_accession=26892828, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79887, CurrentID=0, Status='live', Symbol='PLBD1', Aliases='', description='phospholipase B domain containing 1', other_designations='LAMA-like protein 1|PLB homolog 1|lamina ancestor homolog 1|phospholipase B domain-containing protein 1|putative phospholipase B-like 1', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14503663, end_position_on_the_genomic_accession=14567857, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  79888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79888, CurrentID=0, Status='live', Symbol='LPCAT1', Aliases='AYTL2, PFAAP3, lpcat', description='lysophosphatidylcholine acyltransferase 1', other_designations='1-acylglycerophosphocholine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|LPC acyltransferase 1|LPCAT-1|acetyl-CoA:lyso-PAF acetyltransferase|acetyl-CoA:lyso-platelet-activating factor acetyltransferase|acyl-CoA:lysophosphatidylcholine acyltransferase 1|acyltransferase like 2|acyltransferase-like 2|lyso-PAF acetyltransferase|lysoPAFAT|lysoPC acyltransferase 1|phosphonoformate immuno-associated protein 3|regulated by phosphonoformate', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1461427, end_position_on_the_genomic_accession=1525730, orientation='minus', exon_count=18, OMIM=610472, no_hdr0=''),
  100874257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874257, CurrentID=0, Status='live', Symbol='MTAPP1', Aliases='', description='methylthioadenosine phosphorylase pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158609633, end_position_on_the_genomic_accession=158610199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79890, CurrentID=0, Status='live', Symbol='RIN3', Aliases='', description='Ras and Rab interactor 3', other_designations='RAB5 interacting protein 3|ras interaction/interference protein 3', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92513781, end_position_on_the_genomic_accession=92688994, orientation='plus', exon_count=12, OMIM=610223, no_hdr0=''),
  342035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342035, CurrentID=0, Status='live', Symbol='GLDN', Aliases='CLOM, COLM, CRG-L2, CRGL2, UNC-112', description='gliomedin', other_designations='collomin|colmedin', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51341516, end_position_on_the_genomic_accession=51408013, orientation='plus', exon_count=11, OMIM=608603, no_hdr0=''),
  79892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79892, CurrentID=0, Status='live', Symbol='MCMBP', Aliases='C10orf119, MCM-BP', description='minichromosome maintenance complex binding protein', other_designations='MCM-binding protein|minichromosome maintenance complex-binding protein', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119829404, end_position_on_the_genomic_accession=119873628, orientation='minus', exon_count=17, OMIM=610909, no_hdr0=''),
  79893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79893, CurrentID=0, Status='live', Symbol='GGNBP2', Aliases='DIF-3, DIF3, LCRG1, LZK1, ZFP403, ZNF403', description='gametogenetin binding protein 2', other_designations='C3HC4-type zinc finger protein|laryngeal carcinoma related gene 1|laryngeal carcinoma-related protein 1|zinc finger protein 403', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36544888, end_position_on_the_genomic_accession=36589848, orientation='plus', exon_count=14, OMIM=612275, no_hdr0=''),
  79894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79894, CurrentID=0, Status='live', Symbol='ZNF672', Aliases='', description='zinc finger protein 672', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248838178, end_position_on_the_genomic_accession=248849517, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79895, CurrentID=0, Status='live', Symbol='ATP8B4', Aliases='ATPIM', description='ATPase, class I, type 8B, member 4', other_designations='P4-ATPase flippase complex alpha subunit ATP8B4|potential phospholipid-transporting ATPase IM', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49858238, end_position_on_the_genomic_accession=50181793, orientation='minus', exon_count=50, OMIM=609123, no_hdr0=''),
  79896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79896, CurrentID=0, Status='live', Symbol='THNSL1', Aliases='TSH1', description='threonine synthase like 1', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25016579, end_position_on_the_genomic_accession=25026664, orientation='plus', exon_count=3, OMIM=611260, no_hdr0=''),
  79897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79897, CurrentID=0, Status='live', Symbol='RPP21', Aliases='C6orf135, CAT60', description='ribonuclease P/MRP 21kDa subunit', other_designations='RNaseP protein p21|ribonucleoprotein V', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30345129, end_position_on_the_genomic_accession=30346858, orientation='plus', exon_count=5, OMIM=612524, no_hdr0=''),
  79898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79898, CurrentID=0, Status='live', Symbol='ZNF613', Aliases='', description='zinc finger protein 613', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51927147, end_position_on_the_genomic_accession=51945758, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79899, CurrentID=0, Status='live', Symbol='PRR5L', Aliases='PROTOR2', description='proline rich 5 like', other_designations='Protein observed with Rictor-2|protor-2', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36296175, end_position_on_the_genomic_accession=36465204, orientation='plus', exon_count=11, OMIM=611728, no_hdr0=''),
  100874268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874268, CurrentID=0, Status='live', Symbol='E2F3-IT1', Aliases='', description='E2F3 intronic transcript 1', other_designations='E2F3 intronic transcript 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79901, CurrentID=0, Status='live', Symbol='CYBRD1', Aliases='CYB561A2, DCYTB, FRRS3', description='cytochrome b reductase 1', other_designations='cytochrome b561 family, member A2|duodenal cytochrome b|ferric-chelate reductase 3', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171522247, end_position_on_the_genomic_accession=171558133, orientation='plus', exon_count=5, OMIM=605745, no_hdr0=''),
  79902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79902, CurrentID=0, Status='live', Symbol='NUP85', Aliases='FROUNT, Nup75', description='nucleoporin 85kDa', other_designations='nucleoporin Nup75|nucleoporin Nup85', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75205450, end_position_on_the_genomic_accession=75235759, orientation='plus', exon_count=22, OMIM=170285, no_hdr0=''),
  79903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79903, CurrentID=0, Status='live', Symbol='NAA60', Aliases='HAT4, NAT15', description='N(alpha)-acetyltransferase 60, NatF catalytic subunit', other_designations='N-acetyltransferase 15 (GCN5-related, putative)|histone acetyltransferase type B protein 4', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3443654, end_position_on_the_genomic_accession=3486963, orientation='plus', exon_count=10, OMIM=614246, no_hdr0=''),
  100874272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874272, CurrentID=0, Status='live', Symbol='KLF7-IT1', Aliases='', description='KLF7 intronic transcript 1', other_designations='KLF7 intronic transcript 1 (non-protein coding)', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79905, CurrentID=0, Status='live', Symbol='TMC7', Aliases='', description='transmembrane channel like 7', other_designations='transmembrane channel-like 7', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18983934, end_position_on_the_genomic_accession=19063942, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  79906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79906, CurrentID=0, Status='live', Symbol='MORN1', Aliases='', description='MORN repeat containing 1', other_designations='', map_location='1p36.33-p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2321253, end_position_on_the_genomic_accession=2391751, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100874275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874275, CurrentID=0, Status='live', Symbol='LINC01216', Aliases='EMCN-IT3', description='long intergenic non-protein coding RNA 1216', other_designations='EMCN intronic transcript 3 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100660279, end_position_on_the_genomic_accession=100675113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79908, CurrentID=0, Status='live', Symbol='BTNL8', Aliases='BTN9.2', description='butyrophilin-like 8', other_designations='B7-H5 costimulatory molecule', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180899077, end_position_on_the_genomic_accession=180952166, orientation='plus', exon_count=12, OMIM=615606, no_hdr0=''),
  100874277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874277, CurrentID=0, Status='live', Symbol='LINC01218', Aliases='EMCN-IT1', description='long intergenic non-protein coding RNA 1218', other_designations='EMCN intronic transcript 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874278, CurrentID=0, Status='live', Symbol='ERI3-IT1', Aliases='', description='ERI3 intronic transcript 1', other_designations='ERI3 intronic transcript 1 (non-protein coding)', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44222595, end_position_on_the_genomic_accession=44244283, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373361, CurrentID=0, Status='live', Symbol='LOC105373361', Aliases='', description='uncharacterized LOC105373361', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79912, CurrentID=0, Status='live', Symbol='PYROXD1', Aliases='', description='pyridine nucleotide-disulphide oxidoreductase domain 1', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21437604, end_position_on_the_genomic_accession=21471250, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  79913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79913, CurrentID=0, Status='live', Symbol='ACTR5', Aliases='Arp5, INO80M', description='ARP5 actin-related protein 5 homolog (yeast)', other_designations='INO80 complex subunit M|hARP5|sarcoma antigen NY-SAR-16', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38748258, end_position_on_the_genomic_accession=38772446, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100874282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874282, CurrentID=0, Status='live', Symbol='MLIP-IT1', Aliases='', description='MLIP intronic transcript 1', other_designations='MLIP intronic transcript 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53998890, end_position_on_the_genomic_accession=54007152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79915, CurrentID=0, Status='live', Symbol='ATAD5', Aliases='C17orf41, ELG1, FRAG1', description='ATPase family, AAA domain containing 5', other_designations='chromosome fragility associated gene 1|chromosome fragility-associated gene 1 protein|enhanced level of genomic instability 1 homolog', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30831476, end_position_on_the_genomic_accession=30895869, orientation='plus', exon_count=23, OMIM=609534, no_hdr0=''),
  100874284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874284, CurrentID=0, Status='live', Symbol='DNM3-IT1', Aliases='', description='DNM3 intronic transcript 1', other_designations='DNM3 intronic transcript 1 (non-protein coding)', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171864192, end_position_on_the_genomic_accession=171864687, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79917, CurrentID=0, Status='live', Symbol='MAGIX', Aliases='JM10, PDZX', description='MAGI family member, X-linked', other_designations='PDZ domain containing, X chromosome', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49162564, end_position_on_the_genomic_accession=49168157, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  79918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79918, CurrentID=0, Status='live', Symbol='SETD6', Aliases='', description='SET domain containing 6', other_designations='SET domain-containing protein 6', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58515479, end_position_on_the_genomic_accession=58520527, orientation='plus', exon_count=9, OMIM=616424, no_hdr0=''),
  79919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79919, CurrentID=0, Status='live', Symbol='C2orf54', Aliases='', description='chromosome 2 open reading frame 54', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240886048, end_position_on_the_genomic_accession=240896889, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100874288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874288, CurrentID=0, Status='live', Symbol='TSPY25P', Aliases='', description='testis specific protein, Y-linked 25, pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9624183, end_position_on_the_genomic_accession=9626352, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79921, CurrentID=0, Status='live', Symbol='TCEAL4', Aliases='NPD017, WEX7', description='transcription elongation factor A (SII)-like 4', other_designations='TCEA-like protein 4|transcription elongation factor S-II protein-like 4', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103576231, end_position_on_the_genomic_accession=103587736, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79922, CurrentID=0, Status='live', Symbol='MRM1', Aliases='', description='mitochondrial rRNA methyltransferase 1', other_designations="16S rRNA (guanosine(1145)-2'-O)-methyltransferase|16S rRNA [Gm1145] 2'-O-methyltransferase|CTB-75G16.3|mitochondrial large ribosomal RNA ribose methylase|mitochondrial rRNA methyltransferase 1 homolog", map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36601581, end_position_on_the_genomic_accession=36634698, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  79923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79923, CurrentID=0, Status='live', Symbol='NANOG', Aliases='', description='Nanog homeobox', other_designations='homeobox transcription factor Nanog|homeobox transcription factor Nanog-delta 48', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7788426, end_position_on_the_genomic_accession=7796061, orientation='plus', exon_count=5, OMIM=607937, no_hdr0=''),
  79924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79924, CurrentID=0, Status='live', Symbol='ADM2', Aliases='AM2, dJ579N16.4', description='adrenomedullin 2', other_designations='intermedin', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50481556, end_position_on_the_genomic_accession=50486437, orientation='plus', exon_count=3, OMIM=608682, no_hdr0=''),
  79925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79925, CurrentID=0, Status='live', Symbol='SPEF2', Aliases='CT122, KPL2', description='sperm flagellar 2', other_designations='cancer/testis antigen 122', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35617779, end_position_on_the_genomic_accession=35814611, orientation='plus', exon_count=46, OMIM=610172, no_hdr0=''),
  100874294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874294, CurrentID=0, Status='live', Symbol='HLCS-IT1', Aliases='bde1', description='HLCS intronic transcript 1', other_designations='HLCS intronic transcript 1 (non-protein coding)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79927, CurrentID=0, Status='live', Symbol='FAM110D', Aliases='GRRP1', description='family with sequence similarity 110 member D', other_designations='family with sequence similarity 110, member D|glycine/arginine rich protein 1|glycine/arginine-rich protein 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26159008, end_position_on_the_genomic_accession=26162628, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874296, CurrentID=0, Status='live', Symbol='KCNH1-IT1', Aliases='', description='KCNH1 intronic transcript 1', other_designations='KCNH1 intronic transcript 1 (non-protein coding)', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211132281, end_position_on_the_genomic_accession=211133422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79929, CurrentID=0, Status='live', Symbol='MAP6D1', Aliases='MAPO6D1, SL21', description='MAP6 domain containing 1', other_designations='21 kDa STOP-like protein|STOP-Like protein 21 kD', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183815876, end_position_on_the_genomic_accession=183825605, orientation='minus', exon_count=3, OMIM=610593, no_hdr0=''),
  79930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79930, CurrentID=0, Status='live', Symbol='DOK3', Aliases='DOKL', description='docking protein 3', other_designations='Dok-like protein|downstream of tyrosine kinase 3', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177501904, end_position_on_the_genomic_accession=177510522, orientation='minus', exon_count=8, OMIM=611435, no_hdr0=''),
  79931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79931, CurrentID=0, Status='live', Symbol='TNIP3', Aliases='ABIN-3, LIND', description='TNFAIP3 interacting protein 3', other_designations='A20-binding inhibitor of NF-kappa-B activation 3|ABIN-3 beta|Listeria induced|TNFAIP3-interacting protein 3 beta|TNIP3 beta|listeria-induced gene protein', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121131408, end_position_on_the_genomic_accession=121227609, orientation='minus', exon_count=20, OMIM=608019, no_hdr0=''),
  79932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79932, CurrentID=0, Status='live', Symbol='KIAA0319L', Aliases='', description='KIAA0319-like', other_designations='polycystic kidney disease 1-related', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35433490, end_position_on_the_genomic_accession=35557670, orientation='minus', exon_count=24, OMIM=613535, no_hdr0=''),
  79933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79933, CurrentID=0, Status='live', Symbol='SYNPO2L', Aliases='', description='synaptopodin 2-like', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73644881, end_position_on_the_genomic_accession=73656105, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79934, CurrentID=0, Status='live', Symbol='ADCK4', Aliases='NPHS9', description='aarF domain containing kinase 4', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40691529, end_position_on_the_genomic_accession=40716960, orientation='minus', exon_count=20, OMIM=615567, no_hdr0=''),
  79935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79935, CurrentID=0, Status='live', Symbol='CNTD2', Aliases='CCNP', description='cyclin N-terminal domain containing 2', other_designations='cyclin P', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40222208, end_position_on_the_genomic_accession=40227095, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79937, CurrentID=0, Status='live', Symbol='CNTNAP3', Aliases='CASPR3A, CNTNAP3', description='contactin associated protein-like 3', other_designations='cell recognition molecule Caspr3', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39070195, end_position_on_the_genomic_accession=39288303, orientation='minus', exon_count=25, OMIM=610517, no_hdr0=''),
  100874306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874306, CurrentID=0, Status='live', Symbol='ACAP2-IT1', Aliases='', description='ACAP2 intronic transcript 1', other_designations='ACAP2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79939, CurrentID=0, Status='live', Symbol='SLC35E1', Aliases='', description='solute carrier family 35 member E1', other_designations='solute carrier family 35, member E1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16549837, end_position_on_the_genomic_accession=16572382, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79940, CurrentID=0, Status='live', Symbol='LINC00472', Aliases='C6orf155', description='long intergenic non-protein coding RNA 472', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71407864, end_position_on_the_genomic_accession=71420745, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874309, CurrentID=0, Status='live', Symbol='EPN2-IT1', Aliases='', description='EPN2 intronic transcript 1', other_designations='EPN2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19271393, end_position_on_the_genomic_accession=19274388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874310, CurrentID=0, Status='live', Symbol='MAST4-IT1', Aliases='', description='MAST4 intronic transcript 1', other_designations='MAST4 intronic transcript 1 (non-protein coding)', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79943, CurrentID=0, Status='live', Symbol='ZNF696', Aliases='', description='zinc finger protein 696', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143291389, end_position_on_the_genomic_accession=143299952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79944, CurrentID=0, Status='live', Symbol='L2HGDH', Aliases='C14orf160, L2HGA', description='L-2-hydroxyglutarate dehydrogenase', other_designations='2-hydroxyglutarate dehydrogenase|L-alpha-hydroxyglutarate dehydrogenase|alpha-hydroxyglutarate oxidoreductase|alpha-ketoglutarate reductase|duranin', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50242434, end_position_on_the_genomic_accession=50312389, orientation='minus', exon_count=15, OMIM=609584, no_hdr0=''),
  100874313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874313, CurrentID=0, Status='live', Symbol='AGBL4-IT1', Aliases='', description='AGBL4 intronic transcript 1', other_designations='AGBL4 intronic transcript 1 (non-protein coding)', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=49374201, end_position_on_the_genomic_accession=49472085, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79946, CurrentID=0, Status='live', Symbol='C10orf95', Aliases='', description='chromosome 10 open reading frame 95', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102449837, end_position_on_the_genomic_accession=102451543, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79947, CurrentID=0, Status='live', Symbol='DHDDS', Aliases='CIT, CPT, DS, HDS, RP59', description='dehydrodolichyl diphosphate synthase subunit', other_designations='cis-IPTase|cis-isoprenyltransferase|cis-prenyl transferase|dedol-PP synthase|epididymis tissue protein Li 189m', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26432282, end_position_on_the_genomic_accession=26471306, orientation='plus', exon_count=9, OMIM=608172, no_hdr0=''),
  79948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79948, CurrentID=0, Status='live', Symbol='PLPPR3', Aliases='LPPR3, LPR3, PRG-2, PRG2', description='phospholipid phosphatase related 3', other_designations='PAP-2-like protein 2|plasticity-related gene 2 protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=812488, end_position_on_the_genomic_accession=824243, orientation='minus', exon_count=9, OMIM=610391, no_hdr0=''),
  79949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79949, CurrentID=0, Status='live', Symbol='PLEKHS1', Aliases='C10orf81, HEL185, bA211N11.2', description='pleckstrin homology domain containing S1', other_designations='PH domain-containing family S member 1|PH domain-containing protein C10orf81|epididymis luminal protein 185|pleckstrin homology domain containing, family S member 1', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113751262, end_position_on_the_genomic_accession=113782433, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100874319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874319, CurrentID=0, Status='live', Symbol='NAP1L4P2', Aliases='', description='nucleosome assembly protein 1-like 4 pseudogene 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106466780, end_position_on_the_genomic_accession=106467366, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  342096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342096, CurrentID=0, Status='live', Symbol='GOLGA6A', Aliases='GLP, GOLGA6', description='golgin A6 family member A', other_designations='Golgin linked to PML|golgi autoantigen, golgin subfamily a, 6|golgi autoantigen, golgin subfamily a, 6A|golgi autoantigen, golgin subfamily a, member 6|golgin A6 family, member A|golgin-like protein', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74067654, end_position_on_the_genomic_accession=74082747, orientation='minus', exon_count=18, OMIM=610288, no_hdr0=''),
  79953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79953, CurrentID=0, Status='live', Symbol='SYNDIG1', Aliases='C20orf39, DSPC2, IFITMD5, TMEM90B', description='synapse differentiation inducing 1', other_designations='dispanin subfamily C member 2|interferon induced transmembrane protein domain containing 5|synapse differentiation induced gene 1|transmembrane protein 90B', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24469199, end_position_on_the_genomic_accession=24666617, orientation='plus', exon_count=13, OMIM=614311, no_hdr0=''),
  79954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79954, CurrentID=0, Status='live', Symbol='NOL10', Aliases='PQBP5', description='nucleolar protein 10', other_designations='H_NH0074G24.1|polyglutamine binding protein 5', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10570766, end_position_on_the_genomic_accession=10689987, orientation='minus', exon_count=25, OMIM=616197, no_hdr0=''),
  79955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79955, CurrentID=0, Status='live', Symbol='PDZD7', Aliases='PDZK7', description='PDZ domain containing 7', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101007683, end_position_on_the_genomic_accession=101031157, orientation='minus', exon_count=19, OMIM=612971, no_hdr0=''),
  79956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79956, CurrentID=0, Status='live', Symbol='ERMP1', Aliases='FXNA, KIAA1815, bA207C16.3', description='endoplasmic reticulum metallopeptidase 1', other_designations='Felix-ina|aminopeptidase Fxna|bA207C16.3 (novel protein similar to predicted yeast, plant and worm proteins)', map_location='9p24', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5764061, end_position_on_the_genomic_accession=5833081, orientation='minus', exon_count=18, OMIM=611156, no_hdr0=''),
  79957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79957, CurrentID=0, Status='live', Symbol='PAQR6', Aliases='PRdelta', description='progestin and adipoQ receptor family member VI', other_designations='progestin and adipoQ receptor family member VI variant 3|progestin and adipoQ receptor family member VI variant 4|progestin and adipoQ receptor family member VI variant 5', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156243321, end_position_on_the_genomic_accession=156248488, orientation='minus', exon_count=9, OMIM=614579, no_hdr0=''),
  79958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79958, CurrentID=0, Status='live', Symbol='DENND1C', Aliases='FAM31C', description='DENN/MADD domain containing 1C', other_designations='connecdenn 3|family with sequence similarity 31, member C', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6467204, end_position_on_the_genomic_accession=6481808, orientation='minus', exon_count=23, OMIM=613634, no_hdr0=''),
  79959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79959, CurrentID=0, Status='live', Symbol='CEP76', Aliases='C18orf9, HsT1705', description='centrosomal protein 76kDa', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12661750, end_position_on_the_genomic_accession=12702777, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  79960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79960, CurrentID=0, Status='live', Symbol='JADE1', Aliases='PHF17', description='jade family PHD finger 1', other_designations='PHD finger protein 17|PHD protein Jade-1|gene for apoptosis and differentiation in epithelia|jade family PHD finger protein 1', map_location='4q26-q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128809623, end_position_on_the_genomic_accession=128875224, orientation='plus', exon_count=15, OMIM=610514, no_hdr0=''),
  79961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79961, CurrentID=0, Status='live', Symbol='DENND2D', Aliases='', description='DENN/MADD domain containing 2D', other_designations='RP5-1180E21.2', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111185969, end_position_on_the_genomic_accession=111204791, orientation='minus', exon_count=19, OMIM=615111, no_hdr0=''),
  79962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79962, CurrentID=0, Status='live', Symbol='DNAJC22', Aliases='wus', description='DnaJ heat shock protein family (Hsp40) member C22', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 22|wurst homolog', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49346917, end_position_on_the_genomic_accession=49353684, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79963, CurrentID=0, Status='live', Symbol='ABCA11P', Aliases='ABCA11, EST1133530', description='ATP binding cassette subfamily A member 11, pseudogene', other_designations='ATP-binding cassette, sub-family A (ABC1), member 11 (pseudogene)|ATP-binding cassette, sub-family A (ABC1), member 11, pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=425435, end_position_on_the_genomic_accession=474209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874332, CurrentID=0, Status='live', Symbol='OPCML-IT2', Aliases='', description='OPCML intronic transcript 2', other_designations='OPCML intronic transcript 2 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  450234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450234, CurrentID=0, Status='live', Symbol='PHF5GP', Aliases='', description='PHD finger protein 5G pseudogene', other_designations='PHF5A pseudogene 5', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168231330, end_position_on_the_genomic_accession=168232224, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79966, CurrentID=0, Status='live', Symbol='SCD5', Aliases='ACOD4, FADS4, HSCD5, SCD2, SCD4', description='stearoyl-CoA desaturase 5', other_designations='acyl-CoA-desaturase 4|stearoyl-CoA 9-desaturase|stearoyl-CoA desaturase 2|stearoyl-CoA desaturase 4', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82629537, end_position_on_the_genomic_accession=82798857, orientation='minus', exon_count=6, OMIM=608370, no_hdr0=''),
  105375276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375276, CurrentID=0, Status='live', Symbol='LOC105375276', Aliases='', description='uncharacterized LOC105375276', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50781515, end_position_on_the_genomic_accession=50782187, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79968, CurrentID=0, Status='live', Symbol='WDR76', Aliases='CDW14', description='WD repeat domain 76', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43826914, end_position_on_the_genomic_accession=43868419, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  79969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79969, CurrentID=0, Status='live', Symbol='ATAT1', Aliases='C6orf134, MEC17, Nbla00487, TAT', description='alpha tubulin acetyltransferase 1', other_designations='TAT|acetyltransferase mec-17 homolog|alpha-TAT|alpha-TAT1', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30626836, end_position_on_the_genomic_accession=30646823, orientation='plus', exon_count=16, OMIM=615556, no_hdr0=''),
  79970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79970, CurrentID=0, Status='live', Symbol='ZNF767P', Aliases='ZNF767', description='zinc finger family member 767, pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149547154, end_position_on_the_genomic_accession=149624790, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79971, CurrentID=0, Status='live', Symbol='WLS', Aliases='C1orf139, EVI, GPR177, MRP, mig-14', description='wntless Wnt ligand secretion mediator', other_designations='G protein-coupled receptor 177|integral membrane protein GPR177|protein evenness interrupted homolog|putative NF-kappa-B-activating protein 373|putative NFkB activating protein 373', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68098459, end_position_on_the_genomic_accession=68233000, orientation='minus', exon_count=14, OMIM=611514, no_hdr0=''),
  100874340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874340, CurrentID=0, Status='live', Symbol='CAMTA1-IT1', Aliases='', description='CAMTA1 intronic transcript 1', other_designations='CAMTA1 intronic transcript 1 (non-protein coding)', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7368790, end_position_on_the_genomic_accession=7370263, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  79973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79973, CurrentID=0, Status='live', Symbol='ZNF442', Aliases='', description='zinc finger protein 442', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12349371, end_position_on_the_genomic_accession=12365985, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79974, CurrentID=0, Status='live', Symbol='CPED1', Aliases='C7orf58', description='cadherin-like and PC-esterase domain containing 1', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120988697, end_position_on_the_genomic_accession=121297444, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  100874343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874343, CurrentID=0, Status='live', Symbol='HS1BP3-IT1', Aliases='', description='HS1BP3 intronic transcript 1', other_designations='HS1BP3 intronic transcript 1 (non-protein coding)', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20590775, end_position_on_the_genomic_accession=20592548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874344, CurrentID=0, Status='live', Symbol='FBXO36-IT1', Aliases='', description='FBXO36 intronic transcript 1', other_designations='FBXO36 intronic transcript 1 (non-protein coding)', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79977, CurrentID=0, Status='live', Symbol='GRHL2', Aliases='BOM, DFNA28, ECTDS, TFCP2L3', description='grainyhead-like transcription factor 2', other_designations='brother of mammalian grainyhead|grainyhead-like 2|transcription factor CP2-like 3', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101492406, end_position_on_the_genomic_accession=101678098, orientation='plus', exon_count=18, OMIM=608576, no_hdr0=''),
  100874346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874346, CurrentID=0, Status='live', Symbol='NCKAP5-IT1', Aliases='', description='NCKAP5 intronic transcript 1', other_designations='NCKAP5 intronic transcript 1 (non-protein coding)', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  79979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79979, CurrentID=0, Status='live', Symbol='TRMT2B', Aliases='CXorf34, dJ341D10.3', description='tRNA methyltransferase 2 homolog B', other_designations='TRM2 homolog|TRM2 tRNA methyltransferase 2 homolog B|tRNA (uracil-5-)-methyltransferase homolog', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100973644, end_position_on_the_genomic_accession=101052116, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  79980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79980, CurrentID=0, Status='live', Symbol='DSN1', Aliases='C20orf172, KNL3, MIS13, dJ469A13.2, hKNL-3', description='DSN1 homolog, MIS12 kinetochore complex component', other_designations='DSN1, MIND kinetochore complex component, homolog|DSN1, MIS12 kinetochore complex component|kinetochore null 3 homolog', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36751791, end_position_on_the_genomic_accession=36773827, orientation='minus', exon_count=12, OMIM=609175, no_hdr0=''),
  342125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342125, CurrentID=0, Status='live', Symbol='TMC3', Aliases='', description='transmembrane channel like 3', other_designations='transmembrane channel-like 3', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81332419, end_position_on_the_genomic_accession=81374077, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  79982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79982, CurrentID=0, Status='live', Symbol='DNAJB14', Aliases='EGNR9427, PRO34683', description='DnaJ heat shock protein family (Hsp40) member B14', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 14', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99896248, end_position_on_the_genomic_accession=99946726, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  79983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79983, CurrentID=0, Status='live', Symbol='POF1B', Aliases='POF, POF2B', description='premature ovarian failure, 1B', other_designations='premature ovarian failure protein 1B', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85276894, end_position_on_the_genomic_accession=85379743, orientation='minus', exon_count=17, OMIM=300603, no_hdr0=''),
  100874352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874352, CurrentID=0, Status='live', Symbol='AMMECR1-IT1', Aliases='', description='AMMECR1 intronic transcript 1', other_designations='AMMECR1 intronic transcript 1 (non-protein coding)', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874353, CurrentID=0, Status='live', Symbol='RPS6KA2-IT1', Aliases='', description='RPS6KA2 intronic transcript 1', other_designations='RPS6KA2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166460663, end_position_on_the_genomic_accession=166465383, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79986, CurrentID=0, Status='live', Symbol='ZNF702P', Aliases='ZNF702', description='zinc finger protein 702, pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52968251, end_position_on_the_genomic_accession=52993531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  79987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79987, CurrentID=0, Status='live', Symbol='SVEP1', Aliases='C9orf13, CCP22, POLYDOM, SEL-OB, SELOB', description='sushi, von Willebrand factor type A, EGF and pentraxin domain containing 1', other_designations='CCP module-containing protein 22|selectin-like osteoblast-derived protein|selectin-like protein|serologically defined breast cancer antigen NY-BR-38', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110365248, end_position_on_the_genomic_accession=110579880, orientation='minus', exon_count=48, OMIM=611691, no_hdr0=''),
  342132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342132, CurrentID=0, Status='live', Symbol='ZNF774', Aliases='', description='zinc finger protein 774', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90352245, end_position_on_the_genomic_accession=90361483, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79989, CurrentID=0, Status='live', Symbol='TTC26', Aliases='DYF13, dyf-13', description='tetratricopeptide repeat domain 26', other_designations='TPR repeat protein 26|tetratricopeptide repeat protein 26', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139133370, end_position_on_the_genomic_accession=139191986, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  79990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79990, CurrentID=0, Status='live', Symbol='PLEKHH3', Aliases='', description='pleckstrin homology, MyTH4 and FERM domain containing H3', other_designations='PH domain-containing family H member 3|pleckstrin homology domain containing, family H (with MyTH4 domain) member 3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42667914, end_position_on_the_genomic_accession=42677030, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  79991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79991, CurrentID=0, Status='live', Symbol='OBFC1', Aliases='AAF-44, AAF44, RPA-32, STN1, bA541N10.2', description='oligonucleotide/oligosaccharide-binding fold containing 1', other_designations='alpha accessory factor 44|oligonucleotide/oligosaccharide-binding fold-containing protein 1|replication protein A 32 kDa subunit|suppressor of cdc thirteen homolog', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103870258, end_position_on_the_genomic_accession=103918287, orientation='minus', exon_count=12, OMIM=613128, no_hdr0=''),
  79992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79992, CurrentID=0, Status='live', Symbol='AGPAT4-IT1', Aliases='C6orf59, NCRNA00241', description='AGPAT4 intronic transcript 1', other_designations='AGPAT4 intronic transcript 1 (non-protein coding)', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=161160114, end_position_on_the_genomic_accession=161161982, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79993, CurrentID=0, Status='live', Symbol='ELOVL7', Aliases='', description='ELOVL fatty acid elongase 7', other_designations='3-keto acyl-CoA synthase ELOVL7|ELOVL FA elongase 7|ELOVL family member 7, elongation of long chain fatty acids|very-long-chain 3-oxoacyl-CoA synthase 7', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60751789, end_position_on_the_genomic_accession=60844283, orientation='minus', exon_count=13, OMIM=614451, no_hdr0=''),
  100874362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874362, CurrentID=0, Status='live', Symbol='HOXB-AS1', Aliases='', description='HOXB cluster antisense RNA 1', other_designations='HOXB cluster antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48544351, end_position_on_the_genomic_accession=48551241, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100874363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874363, CurrentID=0, Status='live', Symbol='HOXC-AS1', Aliases='', description='HOXC cluster antisense RNA 1', other_designations='HOXC cluster antisense RNA 1 (non-protein coding)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53999022, end_position_on_the_genomic_accession=54000010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874364, CurrentID=0, Status='live', Symbol='HOXC-AS2', Aliases='', description='HOXC cluster antisense RNA 2', other_designations='HOXC cluster antisense RNA 2 (non-protein coding)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53995208, end_position_on_the_genomic_accession=53996785, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874365, CurrentID=0, Status='live', Symbol='HOXC-AS3', Aliases='', description='HOXC cluster antisense RNA 3', other_designations='HOXC cluster antisense RNA 3 (non-protein coding)', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53983951, end_position_on_the_genomic_accession=53985519, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79998, CurrentID=0, Status='live', Symbol='ANKRD53', Aliases='', description='ankyrin repeat domain 53', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70977210, end_position_on_the_genomic_accession=70986674, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100874367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874367, CurrentID=0, Status='live', Symbol='ZDHHC20-IT1', Aliases='', description='ZDHHC20 intronic transcript 1', other_designations='ZDHHC20 intronic transcript 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  80000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80000, CurrentID=0, Status='live', Symbol='GREB1L', Aliases='C18orf6, KIAA1772', description='growth regulation by estrogen in breast cancer-like', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21242242, end_position_on_the_genomic_accession=21525423, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  100874369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874369, CurrentID=0, Status='live', Symbol='CACNA1C-IT2', Aliases='', description='CACNA1C intronic transcript 2', other_designations='CACNA1C intronic transcript 2 (non-protein coding)', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2048352, end_position_on_the_genomic_accession=2049463, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874370, CurrentID=0, Status='live', Symbol='CACNA1C-IT3', Aliases='', description='CACNA1C intronic transcript 3', other_designations='CACNA1C intronic transcript 3 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2269776, end_position_on_the_genomic_accession=2288745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80003, CurrentID=0, Status='live', Symbol='PCNXL2', Aliases='', description='pecanex-like 2 (Drosophila)', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232984101, end_position_on_the_genomic_accession=233295713, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  80004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80004, CurrentID=0, Status='live', Symbol='ESRP2', Aliases='RBM35B', description='epithelial splicing regulatory protein 2', other_designations='RNA binding motif protein 35A|RNA binding motif protein 35B|RNA-binding motif protein 35B|RNA-binding protein 35B', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68228547, end_position_on_the_genomic_accession=68236233, orientation='minus', exon_count=16, OMIM=612960, no_hdr0=''),
  80005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80005, CurrentID=0, Status='live', Symbol='DOCK5', Aliases='', description='dedicator of cytokinesis 5', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25184772, end_position_on_the_genomic_accession=25415716, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  80006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80006, CurrentID=0, Status='live', Symbol='TRAPPC13', Aliases='C5orf44', description='trafficking protein particle complex 13', other_designations='Trs65-related|UPF0533 protein C5orf44', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=65624717, end_position_on_the_genomic_accession=65666127, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  80007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80007, CurrentID=0, Status='live', Symbol='C10orf88', Aliases='', description='chromosome 10 open reading frame 88', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122930903, end_position_on_the_genomic_accession=122954403, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  80008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80008, CurrentID=0, Status='live', Symbol='TMEM156', Aliases='', description='transmembrane protein 156', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38966016, end_position_on_the_genomic_accession=39032669, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100874377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874377, CurrentID=0, Status='live', Symbol='SMIM2-IT1', Aliases='C13orf44-IT1', description='SMIM2 intronic transcript 1', other_designations='SMIM2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44146470, end_position_on_the_genomic_accession=44158222, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  80010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80010, CurrentID=0, Status='live', Symbol='RMI1', Aliases='BLAP75, C9orf76, FAAP75', description='RecQ mediated genome instability 1', other_designations='BLM-associated polypeptide, 75 kDa|BLM-associated protein 75 kDa|RMI1, RecQ mediated genome instability 1, homolog|homolog of yeast RecQ-mediated genome instability 1 (RMI1)', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83980720, end_position_on_the_genomic_accession=84004074, orientation='plus', exon_count=4, OMIM=610404, no_hdr0=''),
  80011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80011, CurrentID=0, Status='live', Symbol='FAM192A', Aliases='C16orf94, CDA018, CDA10, NIP30', description='family with sequence similarity 192 member A', other_designations='NEFA-interacting nuclear protein NIP30|family with sequence similarity 192, member A', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57152466, end_position_on_the_genomic_accession=57186230, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  80012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80012, CurrentID=0, Status='live', Symbol='PHC3', Aliases='EDR3, HPH3', description='polyhomeotic homolog 3 (Drosophila)', other_designations='early development regulator 3|early development regulatory protein 3|homolog of polyhomeotic 3|polyhomeotic like 3', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170086731, end_position_on_the_genomic_accession=170181749, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  80013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80013, CurrentID=0, Status='live', Symbol='FAM188A', Aliases='C10orf97, CARP, DERP5, MST126, MSTP126, my042', description='family with sequence similarity 188 member A', other_designations='CARD-containing protein|caspase recruitment domain containing pro-apoptotic protein|dermal papilla-derived protein 5|family with sequence similarity 188, member A', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15778169, end_position_on_the_genomic_accession=15860520, orientation='minus', exon_count=19, OMIM=611649, no_hdr0=''),
  80014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80014, CurrentID=0, Status='live', Symbol='WWC2', Aliases='BOMB', description='WW and C2 domain containing 2', other_designations='BH-3-only member B|BH3-only member B protein|WW domain-containing protein 2|WW, C2 and coiled-coil domain containing 2', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183098635, end_position_on_the_genomic_accession=183320777, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  100874383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874383, CurrentID=0, Status='live', Symbol='USP9YP22', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 22', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9185599, end_position_on_the_genomic_accession=9185812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874384, CurrentID=0, Status='live', Symbol='USP9YP30', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 30', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25722990, end_position_on_the_genomic_accession=25723148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80017, CurrentID=0, Status='live', Symbol='C14orf159', Aliases='', description='chromosome 14 open reading frame 159', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91114013, end_position_on_the_genomic_accession=91225396, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  80018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80018, CurrentID=0, Status='live', Symbol='NAA25', Aliases='C12orf30, MDM20, NAP1', description='N(alpha)-acetyltransferase 25, NatB auxiliary subunit', other_designations='N-terminal acetyltransferase B complex subunit NAA25|mitochondrial distribution and morphology 20|natB complex subunit MDM20', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112026689, end_position_on_the_genomic_accession=112109022, orientation='minus', exon_count=26, OMIM=612755, no_hdr0=''),
  80019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80019, CurrentID=0, Status='live', Symbol='UBTD1', Aliases='', description='ubiquitin domain containing 1', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97498882, end_position_on_the_genomic_accession=97571210, orientation='plus', exon_count=3, OMIM=616388, no_hdr0=''),
  80020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80020, CurrentID=0, Status='live', Symbol='FOXRED2', Aliases='ERFAD', description='FAD-dependent oxidoreductase domain containing 2', other_designations='endoplasmic reticulum flavoprotein associated with degradation', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36487186, end_position_on_the_genomic_accession=36507101, orientation='minus', exon_count=9, OMIM=613777, no_hdr0=''),
  80021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80021, CurrentID=0, Status='live', Symbol='TMEM62', Aliases='', description='transmembrane protein 62', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43133144, end_position_on_the_genomic_accession=43185147, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  80022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80022, CurrentID=0, Status='live', Symbol='MYO15B', Aliases='MYO15BP', description='myosin XVB', other_designations='myosin XVB pseudogene|unconventional myosin XVBP pseudogene', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75588058, end_position_on_the_genomic_accession=75626849, orientation='plus', exon_count=61, OMIM='', no_hdr0=''),
  80023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80023, CurrentID=0, Status='live', Symbol='NRSN2', Aliases='C20orf98, dJ1103G7.6', description='neurensin 2', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=346726, end_position_on_the_genomic_accession=354868, orientation='plus', exon_count=8, OMIM=610666, no_hdr0=''),
  80024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80024, CurrentID=0, Status='live', Symbol='SLC8B1', Aliases='NCKX6, NCLX, SLC24A6', description='solute carrier family 8 (sodium/lithium/calcium exchanger), member B1', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 6|sodium/calcium exchanger protein, mitochondrial|solute carrier family 24 (sodium/lithium/calcium exchanger), member 6|solute carrier family 24 (sodium/potassium/calcium exchanger), member 6|solute carrier family 24 member 6|solute carrier family 8 member B1', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113298766, end_position_on_the_genomic_accession=113335142, orientation='minus', exon_count=15, OMIM=609841, no_hdr0=''),
  80025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80025, CurrentID=0, Status='live', Symbol='PANK2', Aliases='C20orf48, HARP, HSS, NBIA1, PKAN', description='pantothenate kinase 2', other_designations='Hallervorden-Spatz syndrome|pantothenic acid kinase 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3888839, end_position_on_the_genomic_accession=3923891, orientation='plus', exon_count=13, OMIM=606157, no_hdr0=''),
  100874394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874394, CurrentID=0, Status='live', Symbol='EIF2AP2', Aliases='', description='eukaryotic translation initiation factor 2A pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150603579, end_position_on_the_genomic_accession=150604648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874395, CurrentID=0, Status='live', Symbol='EIF2AP3', Aliases='', description='eukaryotic translation initiation factor 2A pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150676833, end_position_on_the_genomic_accession=150677918, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80028, CurrentID=0, Status='live', Symbol='FBXL18', Aliases='Fbl18', description='F-box and leucine-rich repeat protein 18', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5454465, end_position_on_the_genomic_accession=5513805, orientation='minus', exon_count=8, OMIM=609084, no_hdr0=''),
  100874397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874397, CurrentID=0, Status='live', Symbol='CDKN2AIPNLP3', Aliases='', description='CDKN2A interacting protein N-terminal like pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39817753, end_position_on_the_genomic_accession=39818050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874398, CurrentID=0, Status='live', Symbol='RAC1P8', Aliases='', description='ras-related C3 botulinum toxin substrate 1 pseudogene 8', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61805150, end_position_on_the_genomic_accession=61805676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80031, CurrentID=0, Status='live', Symbol='SEMA6D', Aliases='', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6D', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47184101, end_position_on_the_genomic_accession=47774223, orientation='plus', exon_count=24, OMIM=609295, no_hdr0=''),
  80032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80032, CurrentID=0, Status='live', Symbol='ZNF556', Aliases='', description='zinc finger protein 556', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2867335, end_position_on_the_genomic_accession=2878517, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874401, CurrentID=0, Status='live', Symbol='EIF2AP4', Aliases='', description='eukaryotic translation initiation factor 2A pseudogene 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44806443, end_position_on_the_genomic_accession=44808223, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80034, CurrentID=0, Status='live', Symbol='CSRNP3', Aliases='FAM130A2, PPP1R73, TAIP-2, TAIP2', description='cysteine-serine-rich nuclear protein 3', other_designations='CSRNP-3|TGF beta induced apotosis protein 2|TGF-beta induced apoptosis protein 2|TGF-beta-induced apoptosis protein 2|family with sequence similarity 130, member A2|protein phosphatase 1, regulatory subunit 73', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165469647, end_position_on_the_genomic_accession=165689407, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80035, CurrentID=0, Status='live', Symbol='ANP32A-IT1', Aliases='C15orf28, HsT18971, NCRNA00321', description='ANP32A intronic transcript 1', other_designations='ANP32A intronic transcript 1 (non-protein coding)', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68803821, end_position_on_the_genomic_accession=68807101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80036, CurrentID=0, Status='live', Symbol='TRPM3', Aliases='GON-2, LTRPC3, MLSN2', description='transient receptor potential cation channel, subfamily M, member 3', other_designations='LTrpC-3|long transient receptor potential channel 3|melastatin 2|melastatin-2', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70529063, end_position_on_the_genomic_accession=71446924, orientation='minus', exon_count=29, OMIM=608961, no_hdr0=''),
  450246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450246, CurrentID=0, Status='live', Symbol='PHF5FP', Aliases='', description='PHD finger protein 5F pseudogene', other_designations='PHF5A pseudogene 4', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115659704, end_position_on_the_genomic_accession=115660512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874406, CurrentID=0, Status='live', Symbol='FAM106DP', Aliases='FAM106D', description='family with sequence similarity 106 member D, pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60173322, end_position_on_the_genomic_accession=60173816, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80039, CurrentID=0, Status='live', Symbol='FAM106A', Aliases='', description='family with sequence similarity 106 member A', other_designations='family with sequence similarity 106, member A', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18524566, end_position_on_the_genomic_accession=18526846, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  342184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342184, CurrentID=0, Status='live', Symbol='FMN1', Aliases='FMN, LD', description='formin 1', other_designations='formin (limb deformity)|limb deformity protein homolog', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32765544, end_position_on_the_genomic_accession=33194784, orientation='minus', exon_count=30, OMIM=136535, no_hdr0=''),
  100874409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874409, CurrentID=0, Status='live', Symbol='SNRPGP8', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 8', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227778510, end_position_on_the_genomic_accession=227778735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874410, CurrentID=0, Status='live', Symbol='HMGN1P9', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene', other_designations='high-mobility group nucleosome binding domain 1 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134385197, end_position_on_the_genomic_accession=134385494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874411, CurrentID=0, Status='live', Symbol='SNRPFP2', Aliases='', description='small nuclear ribonucleoprotein polypeptide F pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25887360, end_position_on_the_genomic_accession=25887602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874412, CurrentID=0, Status='live', Symbol='SNRPFP4', Aliases='', description='small nuclear ribonucleoprotein polypeptide F pseudogene 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48161177, end_position_on_the_genomic_accession=48161414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80045, CurrentID=0, Status='live', Symbol='GPR157', Aliases='', description='G protein-coupled receptor 157', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9103391, end_position_on_the_genomic_accession=9129170, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874414, CurrentID=0, Status='live', Symbol='SNRPEP6', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148567227, end_position_on_the_genomic_accession=148567727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874415, CurrentID=0, Status='live', Symbol='SNRPEP7', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27211195, end_position_on_the_genomic_accession=27211481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874416, CurrentID=0, Status='live', Symbol='SNRPEP9', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 9', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24513407, end_position_on_the_genomic_accession=24513682, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874417, CurrentID=0, Status='live', Symbol='TIMM8BP2', Aliases='', description='translocase of inner mitochondrial membrane 8B pseudogene 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134166682, end_position_on_the_genomic_accession=134166932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874418, CurrentID=0, Status='live', Symbol='SNRPGP5', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 5', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13104574, end_position_on_the_genomic_accession=13104793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874419, CurrentID=0, Status='live', Symbol='IPPKP1', Aliases='', description='inositol 1,3,4,5,6-pentakisphosphate 2-kinase pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21103003, end_position_on_the_genomic_accession=21104371, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80052, CurrentID=0, Status='live', Symbol='RPS2P45', Aliases='RPS2_21_1492', description='ribosomal protein S2 pseudogene 45', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69033630, end_position_on_the_genomic_accession=69034561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874421, CurrentID=0, Status='live', Symbol='ARL2BPP6', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 6', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 6|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 6|ADP-ribosylation factor-like 2 binding protein pseudogene 6', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175290731, end_position_on_the_genomic_accession=175291252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80054, CurrentID=0, Status='live', Symbol='CEBPA-AS1', Aliases='', description='CEBPA antisense RNA 1 (head to head)', other_designations='CEBPA antisense RNA 1 (non-protein coding)', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33302857, end_position_on_the_genomic_accession=33305057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  80055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80055, CurrentID=0, Status='live', Symbol='PGAP1', Aliases='Bst1, ISPD3024, MRT42, SPG67', description='post-GPI attachment to proteins 1', other_designations='GPI deacylase|hPGAP1|post-GPI attachment to proteins factor 1', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196833004, end_position_on_the_genomic_accession=196926995, orientation='minus', exon_count=28, OMIM=611655, no_hdr0=''),
  100874424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874424, CurrentID=0, Status='live', Symbol='SNRPEP10', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 10', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223643954, end_position_on_the_genomic_accession=223644394, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874425, CurrentID=0, Status='live', Symbol='TIMM8BP1', Aliases='', description='translocase of inner mitochondrial membrane 8B pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29865959, end_position_on_the_genomic_accession=29866169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874426, CurrentID=0, Status='live', Symbol='SNRPGP11', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 11', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51496020, end_position_on_the_genomic_accession=51496251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80059, CurrentID=0, Status='live', Symbol='LRRTM4', Aliases='', description='leucine rich repeat transmembrane neuronal 4', other_designations='leucine-rich repeat transmembrane neuronal 4 protein', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76747457, end_position_on_the_genomic_accession=77522376, orientation='minus', exon_count=5, OMIM=610870, no_hdr0=''),
  80060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80060, CurrentID=0, Status='live', Symbol='GNL3LP1', Aliases='GNL3LP', description='guanine nucleotide binding protein-like 3 (nucleolar)-like pseudogene 1', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60891676, end_position_on_the_genomic_accession=60893662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874429, CurrentID=0, Status='live', Symbol='HMGN1P5', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 5', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 5', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158266753, end_position_on_the_genomic_accession=158267320, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874430, CurrentID=0, Status='live', Symbol='HMGN1P6', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 6', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 6', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206259628, end_position_on_the_genomic_accession=206259918, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80063, CurrentID=0, Status='live', Symbol='ATF7IP2', Aliases='MCAF2', description='activating transcription factor 7 interacting protein 2', other_designations='ATF7-interacting protein 2|MBD1-containing chromatin associated factor 2|MBD1-containing chromatin-associated factor 2', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10386055, end_position_on_the_genomic_accession=10483638, orientation='plus', exon_count=16, OMIM=613645, no_hdr0=''),
  100874432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874432, CurrentID=0, Status='live', Symbol='TCEAL4P1', Aliases='', description='transcription elongation factor A (SII)-like 4 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36741873, end_position_on_the_genomic_accession=36742426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874433, CurrentID=0, Status='live', Symbol='IQCJ-SCHIP1-AS1', Aliases='uaz98', description='IQCJ-SCHIP1 readthrough antisense RNA 1', other_designations='IQCJ-SCHIP1 readthrough antisense RNA 1 (non-protein coding)', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159765382, end_position_on_the_genomic_accession=159768612, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874434, CurrentID=0, Status='live', Symbol='SNRPGP14', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 14', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49488970, end_position_on_the_genomic_accession=49489364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80067, CurrentID=0, Status='live', Symbol='DCAF17', Aliases='C2orf37', description='DDB1 and CUL4 associated factor 17', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171434173, end_position_on_the_genomic_accession=171491027, orientation='plus', exon_count=17, OMIM=612515, no_hdr0=''),
  106479503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479503, CurrentID=0, Status='live', Symbol='RN7SL776P', Aliases='', description='RNA, 7SL, cytoplasmic 776, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  80069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80069, CurrentID=0, Status='live', Symbol='LINC00574', Aliases='C6orf208, dJ182D15.1', description='long intergenic non-protein coding RNA 574', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169790073, end_position_on_the_genomic_accession=169802873, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80070, CurrentID=0, Status='live', Symbol='ADAMTS20', Aliases='ADAM-TS20, ADAMTS-20, GON-1', description='ADAM metallopeptidase with thrombospondin type 1 motif 20', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 20|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 20', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43352843, end_position_on_the_genomic_accession=43551940, orientation='minus', exon_count=40, OMIM=611681, no_hdr0=''),
  80071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80071, CurrentID=0, Status='live', Symbol='CCDC15', Aliases='', description='coiled-coil domain containing 15', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124954121, end_position_on_the_genomic_accession=125041489, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  80072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80072, CurrentID=0, Status='live', Symbol='HEXA-AS1', Aliases='C15orf34', description='HEXA antisense RNA 1', other_designations='HEXA antisense RNA 1 (non-protein coding)', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72376113, end_position_on_the_genomic_accession=72378788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874441, CurrentID=0, Status='live', Symbol='HMGN1P15', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 15', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 15', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115289036, end_position_on_the_genomic_accession=115289330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874442, CurrentID=0, Status='live', Symbol='HMGN1P16', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 16', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 16', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148221360, end_position_on_the_genomic_accession=148221642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874443, CurrentID=0, Status='live', Symbol='HMGN1P17', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 17', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 17', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56381781, end_position_on_the_genomic_accession=56382075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874444, CurrentID=0, Status='live', Symbol='HMGN1P18', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 18', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 18', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121050927, end_position_on_the_genomic_accession=121051203, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874445, CurrentID=0, Status='live', Symbol='HMGN1P19', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 19', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 19', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46634614, end_position_on_the_genomic_accession=46634946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80078, CurrentID=0, Status='live', Symbol='LCAL1', Aliases='onco-lncRNA-27', description='lung cancer associated lncRNA 1', other_designations='', map_location='6q14', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79307669, end_position_on_the_genomic_accession=79313384, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874447, CurrentID=0, Status='live', Symbol='HMGN1P34', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 34', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 34', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78519593, end_position_on_the_genomic_accession=78520186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874448, CurrentID=0, Status='live', Symbol='HMGN1P35', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 35', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 35', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69174124, end_position_on_the_genomic_accession=69174717, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874449, CurrentID=0, Status='live', Symbol='ARL2BPP10', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 10', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 10|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 10|ADP-ribosylation factor-like 2 binding protein pseudogene 10', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18102457, end_position_on_the_genomic_accession=18103649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874450, CurrentID=0, Status='live', Symbol='HIGD1AP2', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 2', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81272870, end_position_on_the_genomic_accession=81273167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874451, CurrentID=0, Status='live', Symbol='HIGD1AP8', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 8', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 8', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11481363, end_position_on_the_genomic_accession=11481660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874452, CurrentID=0, Status='live', Symbol='HIGD1AP9', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 9', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 9', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49619313, end_position_on_the_genomic_accession=49619612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874453, CurrentID=0, Status='live', Symbol='HIGD1AP11', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 11', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 11', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53073113, end_position_on_the_genomic_accession=53073411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80086, CurrentID=0, Status='live', Symbol='TUBA4B', Aliases='TUBA4', description='tubulin alpha 4b', other_designations='alpha-tubulin 4B|putative tubulin-like protein alpha-4B|tubulin, alpha 4|tubulin, alpha 4b (pseudogene)', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219253243, end_position_on_the_genomic_accession=219272188, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874455, CurrentID=0, Status='live', Symbol='HIGD1AP14', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 14', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 14', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109673822, end_position_on_the_genomic_accession=109674121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874456, CurrentID=0, Status='live', Symbol='HIGD1AP18', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 18', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 18', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77013754, end_position_on_the_genomic_accession=77014052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874457, CurrentID=0, Status='live', Symbol='CASP1P1', Aliases='', description='caspase 1 pseudogene 1', other_designations='caspase 1, apoptosis-related cysteine peptidase pseudogene 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105123221, end_position_on_the_genomic_accession=105123343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874458, CurrentID=0, Status='live', Symbol='HMGN2P14', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene14', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene14', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63792963, end_position_on_the_genomic_accession=63794072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874459, CurrentID=0, Status='live', Symbol='HMGN1P20', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 20', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 20', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19489101, end_position_on_the_genomic_accession=19489400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874460, CurrentID=0, Status='live', Symbol='HMGN1P22', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 22', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 22', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11181808, end_position_on_the_genomic_accession=11182288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874461, CurrentID=0, Status='live', Symbol='HMGN1P23', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 23', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 23', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14192015, end_position_on_the_genomic_accession=14192288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80094, CurrentID=0, Status='live', Symbol='BIN3-IT1', Aliases='', description='BIN3 intronic transcript 1', other_designations='BIN3 intronic transcript 1 (non-protein coding)', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22640371, end_position_on_the_genomic_accession=22642209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80095, CurrentID=0, Status='live', Symbol='ZNF606', Aliases='ZNF328', description='zinc finger protein 606', other_designations='zinc finger protein 328', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57977072, end_position_on_the_genomic_accession=58003346, orientation='minus', exon_count=7, OMIM=613905, no_hdr0=''),
  100874464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874464, CurrentID=0, Status='live', Symbol='HMGN1P28', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 28', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 28', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60851382, end_position_on_the_genomic_accession=60851745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80097, CurrentID=0, Status='live', Symbol='MZT2B', Aliases='FAM128B, MOZART2B', description='mitotic spindle organizing protein 2B', other_designations='family with sequence similarity 128, member B|mitotic-spindle organizing protein associated with a ring of gamma-tubulin 2B', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130181544, end_position_on_the_genomic_accession=130204794, orientation='plus', exon_count=8, OMIM=613450, no_hdr0=''),
  100874466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874466, CurrentID=0, Status='live', Symbol='HMGN2P19', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 19', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 19', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229570532, end_position_on_the_genomic_accession=229570796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80099, CurrentID=0, Status='live', Symbol='C7orf69', Aliases='', description='chromosome 7 open reading frame 69', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47795291, end_position_on_the_genomic_accession=47819847, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874468, CurrentID=0, Status='live', Symbol='HMGN2P22', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 22', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 22', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100591128, end_position_on_the_genomic_accession=100591326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874469, CurrentID=0, Status='live', Symbol='HMGN2P23', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 23', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 23', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112981166, end_position_on_the_genomic_accession=112981434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874470, CurrentID=0, Status='live', Symbol='HMGN2P26', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 26', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 26', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167958744, end_position_on_the_genomic_accession=167959082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874471, CurrentID=0, Status='live', Symbol='HMGN2P27', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 27', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 27', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116218026, end_position_on_the_genomic_accession=116218295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874472, CurrentID=0, Status='live', Symbol='HMGN2P29', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 29', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 29', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121408542, end_position_on_the_genomic_accession=121408815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874473, CurrentID=0, Status='live', Symbol='HMGN2P32', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 32', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 32', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107458038, end_position_on_the_genomic_accession=107458306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874474, CurrentID=0, Status='live', Symbol='HMGN2P33', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 33', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 33', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84078584, end_position_on_the_genomic_accession=84078857, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874475, CurrentID=0, Status='live', Symbol='HMGN2P45', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 45', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 45', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17132993, end_position_on_the_genomic_accession=17133398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80108, CurrentID=0, Status='live', Symbol='ZFP2', Aliases='ZNF751, zfp-2', description='ZFP2 zinc finger protein', other_designations='zinc finger protein 751', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178895915, end_position_on_the_genomic_accession=178933209, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874477, CurrentID=0, Status='live', Symbol='HIGD1AP17', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 17', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 17', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30964326, end_position_on_the_genomic_accession=30964621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80110, CurrentID=0, Status='live', Symbol='ZNF614', Aliases='', description='zinc finger protein 614', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52013324, end_position_on_the_genomic_accession=52028427, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  80111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80111, CurrentID=0, Status='live', Symbol='C3orf36', Aliases='', description='chromosome 3 open reading frame 36', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133928145, end_position_on_the_genomic_accession=133929812, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100874480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874480, CurrentID=0, Status='live', Symbol='HMGN2P36', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 36', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 36', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13610395, end_position_on_the_genomic_accession=13610664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874481, CurrentID=0, Status='live', Symbol='HMGN2P37', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 37', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 37', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45514189, end_position_on_the_genomic_accession=45514453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80114, CurrentID=0, Status='live', Symbol='BICC1', Aliases='BICC, CYSRD', description='BicC family RNA binding protein 1', other_designations='FGFR2-BICC1 fusion kinase protein|bicaudal C homolog 1', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58512916, end_position_on_the_genomic_accession=58831437, orientation='plus', exon_count=28, OMIM=614295, no_hdr0=''),
  80115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80115, CurrentID=0, Status='live', Symbol='BAIAP2L2', Aliases='', description='BAI1-associated protein 2-like 2', other_designations='BAI1-associated protein 2-like protein 2|pinkbar|planar intestinal- and kidney-specific BAR domain protein', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38084883, end_position_on_the_genomic_accession=38111304, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  100874484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874484, CurrentID=0, Status='live', Symbol='HMGN2P40', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 40', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 40', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68254633, end_position_on_the_genomic_accession=68254901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80117, CurrentID=0, Status='live', Symbol='ARL14', Aliases='ARF7', description='ADP ribosylation factor like GTPase 14', other_designations='ADP-ribosylation factor 7|ADP-ribosylation factor like GTPase 14|ADP-ribosylation factor-like 14', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160677160, end_position_on_the_genomic_accession=160678448, orientation='plus', exon_count=1, OMIM=614439, no_hdr0=''),
  100874486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874486, CurrentID=0, Status='live', Symbol='HMGN2P43', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 43', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 43', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76345838, end_position_on_the_genomic_accession=76346081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80119, CurrentID=0, Status='live', Symbol='PIF1', Aliases='C15orf20, PIF', description="PIF1 5'-to-3' DNA helicase", other_designations="DNA repair and recombination helicase PIF1|PIF1 5'-to-3' DNA helicase homolog|PIF1/RRM3 DNA helicase-like protein|petite integration frequency 1", map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64815630, end_position_on_the_genomic_accession=64827154, orientation='minus', exon_count=15, OMIM=610953, no_hdr0=''),
  100874488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874488, CurrentID=0, Status='live', Symbol='NIPA2P5', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2 pseudogene 5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79090765, end_position_on_the_genomic_accession=79091174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874489, CurrentID=0, Status='live', Symbol='DDX3P2', Aliases='DDX3YP2', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, pseudogene 2', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, Y-linked pseudogene 2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74133004, end_position_on_the_genomic_accession=74133114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80122, CurrentID=0, Status='live', Symbol='MAP3K19', Aliases='RCK, YSK4', description='mitogen-activated protein kinase kinase kinase 19', other_designations='SPS1/STE20-related protein kinase YSK4|YSK4 Sps1/Ste20-related kinase homolog|regulated in COPD kinase|regulated in COPD, protein kinase|yeast Sps1/Ste20-related kinase 4', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134964290, end_position_on_the_genomic_accession=135047444, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  100874491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874491, CurrentID=0, Status='live', Symbol='EEF1A1P41', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 41', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2995067, end_position_on_the_genomic_accession=2995273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80124, CurrentID=0, Status='live', Symbol='VCPIP1', Aliases='DUBA3, VCIP135', description='valosin containing protein (p97)/p47 complex interacting protein 1', other_designations='VCP/p47 complex-interacting 135-kDa protein|valosin-containing protein (p97)/p47 complex-interacting protein p135|valosin-containing protein p97/p47 complex-interacting protein 1|valosin-containing protein p97/p47 complex-interacting protein p135', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66630253, end_position_on_the_genomic_accession=66667217, orientation='minus', exon_count=3, OMIM=611745, no_hdr0=''),
  80125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80125, CurrentID=0, Status='live', Symbol='CCDC33', Aliases='CT61', description='coiled-coil domain containing 33', other_designations='cancer/testis antigen 61', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74217197, end_position_on_the_genomic_accession=74336472, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  100874494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874494, CurrentID=0, Status='live', Symbol='CYP46A4P', Aliases='', description='cytochrome P450 family 46 subfamily A member 4, pseudogene', other_designations='CYP46A-se1[12:13:14]|cytochrome P450, family 46, subfamily A, polypeptide 4, pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48089353, end_position_on_the_genomic_accession=48089634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80127, CurrentID=0, Status='live', Symbol='BBOF1', Aliases='C14orf45, CCDC176', description='basal body orientation factor 1', other_designations='coiled-coil domain containing 176|coiled-coil domain-containing protein 176', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74019350, end_position_on_the_genomic_accession=74078351, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  80128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80128, CurrentID=0, Status='live', Symbol='TRIM46', Aliases='GENEY, TRIFIC', description='tripartite motif containing 46', other_designations='gene Y protein|tripartite, fibronectin type-III and C-terminal SPRY motif protein', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155173787, end_position_on_the_genomic_accession=155184971, orientation='plus', exon_count=16, OMIM=600986, no_hdr0=''),
  80129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80129, CurrentID=0, Status='live', Symbol='CCDC170', Aliases='C6orf97, bA282P11.1', description='coiled-coil domain containing 170', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151494040, end_position_on_the_genomic_accession=151621193, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100874498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874498, CurrentID=0, Status='live', Symbol='SPTLC1P2', Aliases='', description='serine palmitoyltransferase, long chain base subunit 1 pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23856563, end_position_on_the_genomic_accession=23856818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80131, CurrentID=0, Status='live', Symbol='LRRC8E', Aliases='', description='leucine-rich repeat containing 8 family member E', other_designations='leucine rich repeat containing 8 family, member E|leucine-rich repeat-containing protein 8E', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7888505, end_position_on_the_genomic_accession=7902023, orientation='plus', exon_count=5, OMIM=612891, no_hdr0=''),
  100874500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874500, CurrentID=0, Status='live', Symbol='SPTLC1P4', Aliases='', description='serine palmitoyltransferase, long chain base subunit 1 pseudogene 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152077952, end_position_on_the_genomic_accession=152078213, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80133, CurrentID=0, Status='live', Symbol='MROH9', Aliases='ARMC11, C1orf129', description='maestro heat-like repeat family member 9', other_designations='armadillo repeat containing 11', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170935471, end_position_on_the_genomic_accession=171065985, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  100874502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874502, CurrentID=0, Status='live', Symbol='ATP6V0E1P1', Aliases='', description='ATPase, H+ transporting, lysosomal 9kDa, V0 subunit e1 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132948493, end_position_on_the_genomic_accession=132948715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80135, CurrentID=0, Status='live', Symbol='RPF1', Aliases='BXDC5', description='ribosome production factor 1 homolog', other_designations='2310066N05Rik|RNA processing factor 1 (RPF1)|brix domain containing 5|brix domain-containing protein 5|ribosome biogenesis protein RPF1', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84479237, end_position_on_the_genomic_accession=84498350, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100874504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874504, CurrentID=0, Status='live', Symbol='ATP6V0E1P4', Aliases='', description='ATPase, H+ transporting, lysosomal 9kDa, V0 subunit e1 pseudogene 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47550196, end_position_on_the_genomic_accession=47550750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874505, CurrentID=0, Status='live', Symbol='CYP2C59P', Aliases='', description='cytochrome P450 family 2 subfamily C member 59, pseudogene', other_designations='CYP2C9-de2c3c|cytochrome P450, family 2, subfamily C, polypeptide 59, pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95007398, end_position_on_the_genomic_accession=95007565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874506, CurrentID=0, Status='live', Symbol='CHCHD2P1', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123785009, end_position_on_the_genomic_accession=123785451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80139, CurrentID=0, Status='live', Symbol='ZNF703', Aliases='NLZ1, ZEPPO1, ZNF503L, ZPO1', description='zinc finger protein 703', other_designations='zinc finger elbow-related proline domain protein 1', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37695751, end_position_on_the_genomic_accession=37700021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874508, CurrentID=0, Status='live', Symbol='CHCHD2P5', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93921269, end_position_on_the_genomic_accession=93921669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874509, CurrentID=0, Status='live', Symbol='CHCHD2P7', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 7', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87785756, end_position_on_the_genomic_accession=87786455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80142, CurrentID=0, Status='live', Symbol='PTGES2', Aliases='C9orf15, GBF-1, GBF1, PGES2, mPGES-2', description='prostaglandin E synthase 2', other_designations='GATE-binding factor 1|gamma-interferon-activated transcriptional element-binding factor 1|mPGE synthase-2|membrane-associated prostaglandin E synthase 2|membrane-associated prostaglandin E synthase-2|microsomal prostaglandin E synthase-2|prostaglandin-H(2) E-isomerase', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128120693, end_position_on_the_genomic_accession=128128462, orientation='minus', exon_count=10, OMIM=608152, no_hdr0=''),
  80143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80143, CurrentID=0, Status='live', Symbol='SIKE1', Aliases='SIKE', description='suppressor of IKBKE 1', other_designations='suppressor of IKK epsilon|suppressor of IKK-epsilon', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114769479, end_position_on_the_genomic_accession=114780687, orientation='minus', exon_count=5, OMIM=611656, no_hdr0=''),
  80144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80144, CurrentID=0, Status='live', Symbol='FRAS1', Aliases='', description='Fraser extracellular matrix complex subunit 1', other_designations='Fraser syndrome 1', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78056968, end_position_on_the_genomic_accession=78544269, orientation='plus', exon_count=74, OMIM=607830, no_hdr0=''),
  80145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80145, CurrentID=0, Status='live', Symbol='THOC7', Aliases='NIF3L1BP1, fSAP24, hTREX30', description='THO complex 7', other_designations='NIF3L1-binding protein 1|Ngg1 interacting factor 3 like 1 binding protein 1|THO complex 7 homolog|functional spliceosome-associated protein 24|ngg1-interacting factor 3-like protein 1-binding protein 1', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63833870, end_position_on_the_genomic_accession=63864399, orientation='minus', exon_count=9, OMIM=611965, no_hdr0=''),
  80146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80146, CurrentID=0, Status='live', Symbol='UXS1', Aliases='SDR6E1, UGD', description='UDP-glucuronate decarboxylase 1', other_designations='UXS-1|short chain dehydrogenase/reductase family 6E, member 12', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106093303, end_position_on_the_genomic_accession=106194339, orientation='minus', exon_count=18, OMIM=609749, no_hdr0=''),
  100874515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874515, CurrentID=0, Status='live', Symbol='POLR2KP1', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide K, 7.0kDa pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49594091, end_position_on_the_genomic_accession=49594258, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80148, CurrentID=0, Status='live', Symbol='PQLC1', Aliases='', description='PQ loop repeat containing 1', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79902420, end_position_on_the_genomic_accession=79951664, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  80149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80149, CurrentID=0, Status='live', Symbol='ZC3H12A', Aliases='MCPIP, MCPIP1, dJ423B22.1', description='zinc finger CCCH-type containing 12A', other_designations='MCP induced protein 1|MCP-1 treatment-induced protein|MCP-induced protein 1|zinc finger CCCH domain-containing protein 12A', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37473455, end_position_on_the_genomic_accession=37484377, orientation='plus', exon_count=7, OMIM=610562, no_hdr0=''),
  80150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80150, CurrentID=0, Status='live', Symbol='ASRGL1', Aliases='ALP, ALP1, CRASH', description='asparaginase like 1', other_designations='L-asparaginase|L-asparagine amidohydrolase|asparaginase-like 1 protein|asparaginase-like protein 1|beta-aspartyl-peptidase|isoaspartyl dipeptidase', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62337302, end_position_on_the_genomic_accession=62401378, orientation='plus', exon_count=10, OMIM=609212, no_hdr0=''),
  100874519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874519, CurrentID=0, Status='live', Symbol='TCEB2P4', Aliases='', description='transcription elongation factor B (SIII), polypeptide 2 (18kDa, elongin B) pseudogene 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34488579, end_position_on_the_genomic_accession=34488878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80152, CurrentID=0, Status='live', Symbol='CENPT', Aliases='C16orf56, CENP-T', description='centromere protein T', other_designations='interphase centromere complex protein 22', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67828157, end_position_on_the_genomic_accession=67847458, orientation='minus', exon_count=16, OMIM=611510, no_hdr0=''),
  80153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80153, CurrentID=0, Status='live', Symbol='EDC3', Aliases='LSM16, YJDC, YJEFN2', description='enhancer of mRNA decapping 3', other_designations='LSM16 homolog (EDC3, S. cerevisiae)|enhancer of mRNA decapping 3 homolog|hYjeF_N2|hYjeF_N2-15q23|yjeF N-terminal domain-containing protein 2|yjeF domain containing|yjeF domain-containing protein 1|yjeF_N2', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74630558, end_position_on_the_genomic_accession=74696045, orientation='minus', exon_count=12, OMIM=609842, no_hdr0=''),
  80154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80154, CurrentID=0, Status='live', Symbol='GOLGA2P10', Aliases='', description='golgin A2 pseudogene 10', other_designations='antigen SK1, colon cancer-associated|golgin subfamily A member 2-like', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82471477, end_position_on_the_genomic_accession=82513999, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80155, CurrentID=0, Status='live', Symbol='NAA15', Aliases='Ga19, NARG1, NAT1P, NATH, TBDN, TBDN100', description='N(alpha)-acetyltransferase 15, NatA auxiliary subunit', other_designations='N-terminal acetyltransferase|NMDA receptor regulated 1|NMDA receptor-regulated protein 1|gastric cancer antigen Ga19|protein tubedown-1|transcriptional coactivator tubedown-100|tubedown-1', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139301467, end_position_on_the_genomic_accession=139390781, orientation='plus', exon_count=20, OMIM=608000, no_hdr0=''),
  100874524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874524, CurrentID=0, Status='live', Symbol='ST6GALNAC4P1', Aliases='', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 4 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19818182, end_position_on_the_genomic_accession=19818611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80157, CurrentID=0, Status='live', Symbol='CWH43', Aliases='CWH43-C', description='cell wall biogenesis 43 C-terminal homolog', other_designations='cell wall biogenesis protein 43 C-terminal homolog', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48986247, end_position_on_the_genomic_accession=49062079, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100874527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874527, CurrentID=0, Status='live', Symbol='C16DELQ22', Aliases='DEL16q22', description='Chromosome 16q22 deletion syndrome', other_designations='', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614541, no_hdr0=''),
  100874528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874528, CurrentID=0, Status='live', Symbol='XXYLT1-AS1', Aliases='', description='XXYLT1 antisense RNA 1', other_designations='XXYLT1 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195094588, end_position_on_the_genomic_accession=195096057, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  80161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80161, CurrentID=0, Status='live', Symbol='ASMTL-AS1', Aliases='ASMTL-AS, ASMTLAS, CXYorf2, NCRNA00105', description='ASMTL antisense RNA 1', other_designations='ASMTL antisense RNA 1 (non-protein coding)', map_location='Xp22.33; Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1400531, end_position_on_the_genomic_accession=1415421, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80162, CurrentID=0, Status='live', Symbol='ATHL1', Aliases='', description='ATH1, acid trehalase-like 1 (yeast)', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=287832, end_position_on_the_genomic_accession=295688, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100874531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874531, CurrentID=0, Status='live', Symbol='LINC00581', Aliases='linc-MBOAT1-4', description='long intergenic non-protein coding RNA 581', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21486061, end_position_on_the_genomic_accession=21511892, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80164, CurrentID=0, Status='live', Symbol='PRR36', Aliases='', description='proline rich 36', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7868719, end_position_on_the_genomic_accession=7874441, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100874533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874533, CurrentID=0, Status='live', Symbol='DUPXQ27.3Q28', Aliases='CXDUPq27.3q28', description='Chromosome Xq27.3-q28 duplication syndrome', other_designations='', map_location='Xq27.3-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300869, no_hdr0=''),
  80167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80167, CurrentID=0, Status='live', Symbol='ABHD18', Aliases='C4orf29', description='abhydrolase domain containing 18', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127965261, end_position_on_the_genomic_accession=128036326, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  80168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80168, CurrentID=0, Status='live', Symbol='MOGAT2', Aliases='DGAT2L5., MGAT2, hDC5', description='monoacylglycerol O-acyltransferase 2', other_designations='acyl-CoA:monoacylglycerol acyltransferase 2|diacylglycerol O-acyltransferase candidate 5|diacylglycerol acyltransferase 2-like protein 5', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75717819, end_position_on_the_genomic_accession=75731286, orientation='plus', exon_count=7, OMIM=610270, no_hdr0=''),
  80169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80169, CurrentID=0, Status='live', Symbol='CTC1', Aliases='AAF-132, AAF132, C17orf68, CRMCC, tmp494178', description='CTS telomere maintenance complex component 1', other_designations='HBV DNAPTP1-transactivated protein B|alpha accessory factor 132|conserved telomere capping protein 1', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8224821, end_position_on_the_genomic_accession=8248095, orientation='minus', exon_count=23, OMIM=613129, no_hdr0=''),
  80173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80173, CurrentID=0, Status='live', Symbol='IFT74', Aliases='CCDC2, CMG-1, CMG1', description='intraflagellar transport 74', other_designations='capillary morphogenesis gene 1 protein|capillary morphogenesis protein 1|coiled-coil domain containing 2|coiled-coil domain-containing protein 2|intraflagellar transport 74 homolog', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26947039, end_position_on_the_genomic_accession=27066129, orientation='plus', exon_count=25, OMIM=608040, no_hdr0=''),
  80174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80174, CurrentID=0, Status='live', Symbol='DBF4B', Aliases='ASKL1, CHIFB, DRF1, ZDBF1B', description='DBF4 zinc finger B', other_designations='ASK-like protein 1|DBF4 homolog B|Dbf4-related factor 1|activator of S-phase kinase-like protein 1|chiffon homolog B|zinc finger, DBF-type containing 1B', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44708608, end_position_on_the_genomic_accession=44758852, orientation='plus', exon_count=19, OMIM=611661, no_hdr0=''),
  646877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646877, CurrentID=0, Status='live', Symbol='GAPDHP34', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 34', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36740041, end_position_on_the_genomic_accession=36740923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80176, CurrentID=0, Status='live', Symbol='SPSB1', Aliases='SSB-1, SSB1', description='splA/ryanodine receptor domain and SOCS box containing 1', other_designations='SPRY domain-containing SOCS box protein SSB-1|novel SPRY domain containing protein', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9292882, end_position_on_the_genomic_accession=9369532, orientation='plus', exon_count=3, OMIM=611657, no_hdr0=''),
  80177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80177, CurrentID=0, Status='live', Symbol='MYCT1', Aliases='MTLC', description='myc target 1', other_designations='myc target in myeloid cells 1|myc target in myeloid cells protein 1', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152697895, end_position_on_the_genomic_accession=152747175, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  80178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80178, CurrentID=0, Status='live', Symbol='C16orf59', Aliases='', description='chromosome 16 open reading frame 59', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2460111, end_position_on_the_genomic_accession=2464963, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  80179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80179, CurrentID=0, Status='live', Symbol='MYO19', Aliases='MYOHD1', description='myosin XIX', other_designations='myosin head domain containing 1|myosin head domain-containing protein 1|myosin-XIX', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36495755, end_position_on_the_genomic_accession=36535457, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  80183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80183, CurrentID=0, Status='live', Symbol='KIAA0226L', Aliases='C13orf18', description='KIAA0226-like', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46342000, end_position_on_the_genomic_accession=46390042, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  80184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80184, CurrentID=0, Status='live', Symbol='CEP290', Aliases='3H11Ag, BBS14, CT87, JBTS5, LCA10, MKS4, NPHP6, POC3, SLSN6, rd16', description='centrosomal protein 290kDa', other_designations='Bardet-Biedl syndrome 14 protein|CTCL tumor antigen se2-2|Meckel syndrome, type 4|POC3 centriolar protein homolog|cancer/testis antigen 87|monoclonal antibody 3H11 antigen|nephrocytsin-6|prostate cancer antigen T21|tumor antigen se2-2', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88049013, end_position_on_the_genomic_accession=88142216, orientation='minus', exon_count=60, OMIM=610142, no_hdr0=''),
  80185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80185, CurrentID=0, Status='live', Symbol='TTI2', Aliases='C8orf41, MRT39', description='TELO2 interacting protein 2', other_designations='Tel2 interacting protein 2 homolog', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33498509, end_position_on_the_genomic_accession=33513185, orientation='minus', exon_count=8, OMIM=614426, no_hdr0=''),
  105373407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373407, CurrentID=0, Status='live', Symbol='LOC105373407', Aliases='', description='uncharacterized LOC105373407', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7476748, end_position_on_the_genomic_accession=7478651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102503429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102503429, CurrentID=0, Status='live', Symbol='LINC01335', Aliases='', description='long intergenic non-protein coding RNA 1335', other_designations='CTD-2275D24.4', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74306410, end_position_on_the_genomic_accession=74308325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122592, CurrentID=0, Status='live', Symbol='KRT18P6', Aliases='', description='keratin 18 pseudogene 6', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35511767, end_position_on_the_genomic_accession=35513117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80194, CurrentID=0, Status='live', Symbol='TMEM134', Aliases='', description='transmembrane protein 134', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67464348, end_position_on_the_genomic_accession=67469277, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80195, CurrentID=0, Status='live', Symbol='TMEM254', Aliases='C10orf57, bA369J21.6', description='transmembrane protein 254', other_designations='transmembrane protein C10orf57', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80078646, end_position_on_the_genomic_accession=80092557, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80196, CurrentID=0, Status='live', Symbol='RNF34', Aliases='CARP-1, CARP1, RFI, RIF, RIFF, hRFI', description='ring finger protein 34, E3 ubiquitin protein ligase', other_designations='FYVE-RING finger protein MOMO|RING finger protein RIFF|caspase regulator CARP1|caspases-8 and -10-associated RING finger protein 1|human RING finger homologous to inhibitor of apoptosis protein', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121400083, end_position_on_the_genomic_accession=121424352, orientation='plus', exon_count=7, OMIM=608299, no_hdr0=''),
  80198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80198, CurrentID=0, Status='live', Symbol='MUS81', Aliases='SLX3', description='MUS81 structure-specific endonuclease subunit', other_designations='MUS81 endonuclease homolog|SLX3 structure-specific endonuclease subunit homolog', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65860244, end_position_on_the_genomic_accession=65866443, orientation='plus', exon_count=17, OMIM=606591, no_hdr0=''),
  80199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80199, CurrentID=0, Status='live', Symbol='FUZ', Aliases='FY, NTD', description='fuzzy planar cell polarity protein', other_designations='fuzzy homolog', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49806866, end_position_on_the_genomic_accession=49817376, orientation='minus', exon_count=17, OMIM=610622, no_hdr0=''),
  80201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80201, CurrentID=0, Status='live', Symbol='HKDC1', Aliases='', description='hexokinase domain containing 1', other_designations='hexokinase domain-containing protein 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69220303, end_position_on_the_genomic_accession=69267559, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  342346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342346, CurrentID=0, Status='live', Symbol='C16orf96', Aliases='', description='chromosome 16 open reading frame 96', other_designations='putative uncharacterized protein C16orf96', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4538709, end_position_on_the_genomic_accession=4600455, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100133495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133495, CurrentID=0, Status='live', Symbol='LOC100133495', Aliases='', description='HERV-K_11q22.1 provirus ancestral Pol protein [Includes: Reverse transcriptase', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  80204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80204, CurrentID=0, Status='live', Symbol='FBXO11', Aliases='FBX11, PRMT9, UBR6, UG063H01, VIT1', description='F-box protein 11', other_designations='protein arginine N-methyltransferase 9|ubiquitin protein ligase E3 component n-recognin 6|vitiligo-associated protein 1|vitiligo-associated protein VIT-1', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47806920, end_position_on_the_genomic_accession=47906535, orientation='minus', exon_count=24, OMIM=607871, no_hdr0=''),
  80205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80205, CurrentID=0, Status='live', Symbol='CHD9', Aliases='AD013, CHD-9, CReMM, KISH2, PRIC320', description='chromodomain helicase DNA binding protein 9', other_designations='ATP-dependent helicase CHD9|PPAR-alpha-interacting complex protein 320 kDa|PPAR{gamma}-interacting cofactor 320 kDa|chromatin remodeling factor CHROM1|chromatin-related mesenchymal modulator|chromatin-remodeling factor CHROM1|ciprofibrate bound protein p240|kismet homolog 2|peroxisomal proliferator-activated receptor A-interacting complex 320 kDa protein|proteinx0008', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53054998, end_position_on_the_genomic_accession=53327502, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  80206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80206, CurrentID=0, Status='live', Symbol='FHOD3', Aliases='FHOS2, Formactin2', description='formin homology 2 domain containing 3', other_designations='formactin-2|formin homolog overexpressed in spleen 2', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36297696, end_position_on_the_genomic_accession=36780059, orientation='plus', exon_count=31, OMIM=609691, no_hdr0=''),
  80207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80207, CurrentID=0, Status='live', Symbol='OPA3', Aliases='MGA3', description='optic atrophy 3 (autosomal recessive, with chorea and spastic paraplegia)', other_designations="Optic atrophy 3 (Iraqi-Jewish 'optic atrophy plus')", map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45526863, end_position_on_the_genomic_accession=45584864, orientation='minus', exon_count=5, OMIM=606580, no_hdr0=''),
  80208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80208, CurrentID=0, Status='live', Symbol='SPG11', Aliases='KIAA1840', description='spastic paraplegia 11 (autosomal recessive)', other_designations='colorectal carcinoma-associated protein|spastic paraplegia 11 protein', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44562696, end_position_on_the_genomic_accession=44663678, orientation='minus', exon_count=40, OMIM=610844, no_hdr0=''),
  80209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80209, CurrentID=0, Status='live', Symbol='PROSER1', Aliases='C13orf23', description='proline and serine rich 1', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39009865, end_position_on_the_genomic_accession=39038095, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  80210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80210, CurrentID=0, Status='live', Symbol='ARMC9', Aliases='ARM, KU-MEL-1, NS21', description='armadillo repeat containing 9', other_designations='armadillo/beta-catenin-like repeats|melanoma/melanocyte specific protein KU-MEL-1|melanoma/melanocyte-specific tumor antigen KU-MEL-1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231198546, end_position_on_the_genomic_accession=231394991, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  100130531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130531, CurrentID=0, Status='live', Symbol='LOC100130531', Aliases='', description='ribosomal protein L39 pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93278768, end_position_on_the_genomic_accession=93279188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80212, CurrentID=0, Status='live', Symbol='CCDC92', Aliases='', description='coiled-coil domain containing 92', other_designations='limkain beta 2|limkain beta-2', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123936409, end_position_on_the_genomic_accession=123972985, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  342357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342357, CurrentID=0, Status='live', Symbol='ZKSCAN2', Aliases='ZNF694, ZSCAN31, ZSCAN34', description='zinc finger with KRAB and SCAN domains 2', other_designations='zinc finger and SCAN domain containing 31|zinc finger protein 694', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25236001, end_position_on_the_genomic_accession=25257889, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  342358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342358, CurrentID=0, Status='live', Symbol='CYCSP39', Aliases='HC1, HCP39', description='cytochrome c, somatic pseudogene 39', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25454339, end_position_on_the_genomic_accession=25454647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80215, CurrentID=0, Status='live', Symbol='RUNX1-IT1', Aliases='C21orf96', description='RUNX1 intronic transcript 1', other_designations='RUNX1 intronic transcript 1 (non-protein coding)', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35037925, end_position_on_the_genomic_accession=35039426, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80216, CurrentID=0, Status='live', Symbol='ALPK1', Aliases='8430410J10Rik, LAK', description='alpha kinase 1', other_designations='alpha-kinase 1|chromosome 4 kinase|lymphocyte alpha-kinase|lymphocyte alpha-protein kinase', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112297336, end_position_on_the_genomic_accession=112442618, orientation='plus', exon_count=17, OMIM=607347, no_hdr0=''),
  80217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80217, CurrentID=0, Status='live', Symbol='CFAP43', Aliases='C10orf79, WDR96, bA373N18.2', description='cilia and flagella associated protein 43', other_designations='WD repeat domain 96|WD repeat-containing protein 96|WD repeat-containing protein C10orf79', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104129888, end_position_on_the_genomic_accession=104232366, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  80218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80218, CurrentID=0, Status='live', Symbol='NAA50', Aliases='MAK3, NAT13, NAT13P, NAT5, NAT5P, SAN', description='N(alpha)-acetyltransferase 50, NatE catalytic subunit', other_designations='N-acetyltransferase 13 (GCN5-related)|N-acetyltransferase 5|N-acetyltransferase san homolog|N-alpha-acetyltransferase 50, NatE catalytic subunit|natE catalytic subunit', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113716460, end_position_on_the_genomic_accession=113746273, orientation='minus', exon_count=5, OMIM=610834, no_hdr0=''),
  80219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80219, CurrentID=0, Status='live', Symbol='COQ10B', Aliases='', description='coenzyme Q10B', other_designations='coenzyme Q10 homolog B', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197453507, end_position_on_the_genomic_accession=197475135, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80221, CurrentID=0, Status='live', Symbol='ACSF2', Aliases='ACSMW, AVYV493', description='acyl-CoA synthetase family member 2', other_designations='PPARG binding, long chain fatty acid acyl Co-A ligase like', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50426158, end_position_on_the_genomic_accession=50474845, orientation='plus', exon_count=19, OMIM=610465, no_hdr0=''),
  80222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80222, CurrentID=0, Status='live', Symbol='TARS2', Aliases='COXPD21, TARSL1, thrRS', description='threonyl-tRNA synthetase 2, mitochondrial (putative)', other_designations='threonyl-tRNA synthetase-like 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150487364, end_position_on_the_genomic_accession=150507609, orientation='plus', exon_count=18, OMIM=612805, no_hdr0=''),
  80223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80223, CurrentID=0, Status='live', Symbol='RAB11FIP1', Aliases='NOEL1A, RCP, rab11-FIP1', description='RAB11 family interacting protein 1 (class I)', other_designations='RAB11 coupling protein|Rab effector protein|Rab-interacting recycling protein', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37858947, end_position_on_the_genomic_accession=37899497, orientation='minus', exon_count=6, OMIM=608737, no_hdr0=''),
  80224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80224, CurrentID=0, Status='live', Symbol='NUBPL', Aliases='C14orf127, IND1, huInd1', description='nucleotide binding protein-like', other_designations='IND1 homolog|iron-sulfur protein required for NADH dehydrogenase', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31561385, end_position_on_the_genomic_accession=31861224, orientation='plus', exon_count=17, OMIM=613621, no_hdr0=''),
  80227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80227, CurrentID=0, Status='live', Symbol='PAAF1', Aliases='PAAF, Rpn14, WDR71', description='proteasomal ATPase associated factor 1', other_designations='WD repeat domain 71|WD repeat-containing protein 71', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73876699, end_position_on_the_genomic_accession=73927736, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  80228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80228, CurrentID=0, Status='live', Symbol='ORAI2', Aliases='C7orf19, CBCIP2, MEM142B, TMEM142B', description='ORAI calcium release-activated calcium modulator 2', other_designations='CAP-binding protein complex interacting protein 2|CAP-binding protein complex-interacting protein 2|H_NH0514P08.8|putative protein ORAI2-2|transmembrane protein 142B', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102433530, end_position_on_the_genomic_accession=102456821, orientation='plus', exon_count=4, OMIM=610929, no_hdr0=''),
  80230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80230, CurrentID=0, Status='live', Symbol='RUFY1', Aliases='RABIP4, ZFYVE12', description='RUN and FYVE domain containing 1', other_designations='FYVE-finger protein EIP1|la-binding protein 1|rab4-interacting protein|zinc finger FYVE domain-containing protein 12', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179550561, end_position_on_the_genomic_accession=179610026, orientation='plus', exon_count=24, OMIM=610327, no_hdr0=''),
  80231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80231, CurrentID=0, Status='live', Symbol='CXorf21', Aliases='', description='chromosome X open reading frame 21', other_designations='5430427O19Rik', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30558824, end_position_on_the_genomic_accession=30577916, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  80232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80232, CurrentID=0, Status='live', Symbol='WDR26', Aliases='CDW2, GID7, MIP2', description='WD repeat domain 26', other_designations='CUL4- and DDB1-associated WDR protein 2|GID complex subunit 7 homolog|myocardial ischemic preconditioning up-regulated protein 2|myocardial ischemic preconditioning upregulated protein 2', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224385143, end_position_on_the_genomic_accession=224434299, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  80233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80233, CurrentID=0, Status='live', Symbol='FAAP100', Aliases='C17orf70', description='Fanconi anemia core complex associated protein 100', other_designations='Fanconi anemia associated protein 100 kDa subunit|Fanconi anemia core complex 100 kDa subunit|Fanconi anemia-associated protein, 100kDa|fanconi anemia-associated protein of 100 kDa', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81539885, end_position_on_the_genomic_accession=81552403, orientation='minus', exon_count=10, OMIM=611301, no_hdr0=''),
  80235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80235, CurrentID=0, Status='live', Symbol='PIGZ', Aliases='GPI-MT-IV, PIG-Z, SMP3', description='phosphatidylinositol glycan anchor biosynthesis class Z', other_designations='GPI mannosyltransferase IV|SMP3 mannosyltransferase|dol-P-Man dependent GPI mannosyltransferase|phosphatidylinositol glycan anchor biosynthesis, class Z|phosphatidylinositol glycan, class Z|phosphatidylinositol-glycan biosynthesis class Z protein', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196946343, end_position_on_the_genomic_accession=196968871, orientation='minus', exon_count=6, OMIM=611671, no_hdr0=''),
  80237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80237, CurrentID=0, Status='live', Symbol='ELL3', Aliases='', description='elongation factor RNA polymerase II-like 3', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43772600, end_position_on_the_genomic_accession=43777304, orientation='minus', exon_count=11, OMIM=609885, no_hdr0=''),
  80243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80243, CurrentID=0, Status='live', Symbol='PREX2', Aliases='6230420N16Rik, DEP.2, DEPDC2, P-REX2, PPP1R129', description='phosphatidylinositol-3,4,5-trisphosphate-dependent Rac exchange factor 2', other_designations='DEP domain-containing protein 2|PtdIns(3,4,5)-dependent Rac exchanger 2|protein phosphatase 1, regulatory subunit 129', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67951983, end_position_on_the_genomic_accession=68231662, orientation='plus', exon_count=41, OMIM=612139, no_hdr0=''),
  105373417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373417, CurrentID=0, Status='live', Symbol='LOC105373417', Aliases='', description='uncharacterized LOC105373417', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9135020, end_position_on_the_genomic_accession=9145910, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  646890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646890, CurrentID=0, Status='live', Symbol='LOC646890', Aliases='', description='axin interactor, dorsalization associated pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158277758, end_position_on_the_genomic_accession=158278673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80254, CurrentID=0, Status='live', Symbol='CEP63', Aliases='SCKL6', description='centrosomal protein 63kDa', other_designations='centrosome protein CEP63', map_location='3q22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134485733, end_position_on_the_genomic_accession=134575013, orientation='plus', exon_count=21, OMIM=614724, no_hdr0=''),
  80255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80255, CurrentID=0, Status='live', Symbol='SLC35F5', Aliases='', description='solute carrier family 35 member F5', other_designations='HCV NS5A-transactivated protein 3|hepatitis C virus NS5A-transactivated protein 3|solute carrier family 35, member F5', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113712792, end_position_on_the_genomic_accession=113756823, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  80256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80256, CurrentID=0, Status='live', Symbol='FAM214B', Aliases='KIAA1539', description='family with sequence similarity 214 member B', other_designations='family with sequence similarity 214, member B', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35104121, end_position_on_the_genomic_accession=35116351, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  341056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341056, CurrentID=0, Status='live', Symbol='LOC341056', Aliases='', description='SUMO1 activating enzyme subunit 1 pseudogene', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123017566, end_position_on_the_genomic_accession=123019611, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  80258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80258, CurrentID=0, Status='live', Symbol='EFHC2', Aliases='MRX74, dJ1158H2.1', description='EF-hand domain (C-terminal) containing 2', other_designations='mental retardation, X-linked 74', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44146254, end_position_on_the_genomic_accession=44343677, orientation='minus', exon_count=17, OMIM=300817, no_hdr0=''),
  646891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646891, CurrentID=0, Status='live', Symbol='SDCCAG3P2', Aliases='', description='serologically defined colon cancer antigen 3 pseudogene 2', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175044429, end_position_on_the_genomic_accession=175045848, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80262, CurrentID=0, Status='live', Symbol='C16orf70', Aliases='C16orf6, LIN10, lin-10', description='chromosome 16 open reading frame 70', other_designations='lin-10 homolog', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67109898, end_position_on_the_genomic_accession=67148539, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  80263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80263, CurrentID=0, Status='live', Symbol='TRIM45', Aliases='RNF99', description='tripartite motif containing 45', other_designations='ring finger protein 99|tripartite motif-containing 45', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117111055, end_position_on_the_genomic_accession=117121789, orientation='minus', exon_count=6, OMIM=609318, no_hdr0=''),
  80264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80264, CurrentID=0, Status='live', Symbol='ZNF430', Aliases='', description='zinc finger protein 430', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21020620, end_position_on_the_genomic_accession=21060055, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373420, CurrentID=0, Status='live', Symbol='LOC105373420', Aliases='', description='uncharacterized LOC105373420', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9752890, end_position_on_the_genomic_accession=9757498, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80267, CurrentID=0, Status='live', Symbol='EDEM3', Aliases='C1orf22', description='ER degradation enhancer, mannosidase alpha-like 3', other_designations='ER degradation-enhancing -mannosidase-like protein 3|ER degradation-enhancing alpha-mannosidase-like 3|alpha-1,2-mannosidase EDEM3', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184690491, end_position_on_the_genomic_accession=184754907, orientation='minus', exon_count=22, OMIM=610214, no_hdr0=''),
  80270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80270, CurrentID=0, Status='live', Symbol='HSD3B7', Aliases='CBAS1, PFIC4, SDR11E3', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 7', other_designations='3 beta-hydroxy-delta 5-C27-steroid oxidoreductase|3 beta-hydroxysteroid dehydrogenase type VII|3-beta-HSD VII|3-beta-hydroxy-Delta(5)-C27 steroid oxidoreductase|C(27)-3BETA-HSD|c(27) 3-beta-HSD|cholest-5-ene-3-beta,7-alpha-diol 3-beta-dehydrogenase|short chain dehydrogenase/reductase family 11E, member 3', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30985110, end_position_on_the_genomic_accession=30989152, orientation='plus', exon_count=7, OMIM=607764, no_hdr0=''),
  80271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80271, CurrentID=0, Status='live', Symbol='ITPKC', Aliases='IP3-3KC, IP3KC', description='inositol-trisphosphate 3-kinase C', other_designations='IP3 3-kinase C|IP3K C|InsP 3 kinase C|inositol 1,4,5-trisphosphate 3-kinase C|insP 3-kinase C', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40716980, end_position_on_the_genomic_accession=40740860, orientation='plus', exon_count=8, OMIM=606476, no_hdr0=''),
  80273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80273, CurrentID=0, Status='live', Symbol='GRPEL1', Aliases='GrpE, HMGE, mt-GrpE#1', description='GrpE-like 1, mitochondrial (E. coli)', other_designations='GrpE-like protein cochaperone', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7060053, end_position_on_the_genomic_accession=7068073, orientation='minus', exon_count=4, OMIM=606173, no_hdr0=''),
  80274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80274, CurrentID=0, Status='live', Symbol='SCUBE1', Aliases='', description='signal peptide, CUB domain, EGF-like 1', other_designations='signal peptide-CUB domain-EGF-related 1', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43203223, end_position_on_the_genomic_accession=43343388, orientation='minus', exon_count=26, OMIM=611746, no_hdr0=''),
  342419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342419, CurrentID=0, Status='live', Symbol='KRT8P22', Aliases='', description='keratin 8 pseudogene 22', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78048237, end_position_on_the_genomic_accession=78049972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130542, CurrentID=0, Status='live', Symbol='LOC100130542', Aliases='', description='uncharacterized LOC100130542', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=69208613, end_position_on_the_genomic_accession=69209297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  80279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80279, CurrentID=0, Status='live', Symbol='CDK5RAP3', Aliases='C53, HSF-27, IC53, LZAP, MST016, OK/SW-cl.114, PP1553', description='CDK5 regulatory subunit associated protein 3', other_designations='CDK5 regulatory subunit associated protein IC53-2|LXXLL/leucine-zipper-containing ARF-binding protein|LXXLL/leucine-zipper-containing ARFbinding protein|ischemic heart CDK5 activator-binding protein C53', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47967904, end_position_on_the_genomic_accession=47981786, orientation='plus', exon_count=16, OMIM=608202, no_hdr0=''),
  342426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342426, CurrentID=0, Status='live', Symbol='ZNF720P1', Aliases='', description='zinc finger protein 720 pseudogene 1', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31602119, end_position_on_the_genomic_accession=31603528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80298, CurrentID=0, Status='live', Symbol='MTERF2', Aliases='MTERFD3, mTERFL', description='mitochondrial transcription termination factor 2', other_designations='MTERF domain containing 3|mTERF domain-containing protein 3, mitochondrial|mitochondrial transcription termination factor-like protein', map_location='12q24.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106977291, end_position_on_the_genomic_accession=106987166, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  342443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342443, CurrentID=0, Status='live', Symbol='LOC342443', Aliases='', description='ADAM metallopeptidase domain 21 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33808760, end_position_on_the_genomic_accession=33810764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80301, CurrentID=0, Status='live', Symbol='PLEKHO2', Aliases='PLEKHQ1, PP1628, pp9099', description='pleckstrin homology domain containing O2', other_designations='PH domain-containing family O member 2|PH domain-containing family Q member 1|PH domain-containing protein|pleckstrin homology domain containing, family O member 2|pleckstrin homology domain containing, family Q member 1|pleckstrin homology domain-containing family Q member 1', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64841883, end_position_on_the_genomic_accession=64868002, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80303, CurrentID=0, Status='live', Symbol='EFHD1', Aliases='MST133, MSTP133, PP3051, SWS2', description='EF-hand domain family member D1', other_designations='EF-hand domain family, member D1|EF-hand domain-containing protein 1|swiprosin-2', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232606057, end_position_on_the_genomic_accession=232682781, orientation='plus', exon_count=6, OMIM=611617, no_hdr0=''),
  80304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80304, CurrentID=0, Status='live', Symbol='C2orf44', Aliases='PP384, WDCP', description='chromosome 2 open reading frame 44', other_designations='WD repeat-containing protein C2orf44', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24029336, end_position_on_the_genomic_accession=24047426, orientation='minus', exon_count=4, OMIM=616234, no_hdr0=''),
  80305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80305, CurrentID=0, Status='live', Symbol='TRABD', Aliases='LP6054, PP2447', description='TraB domain containing', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50185905, end_position_on_the_genomic_accession=50199598, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  80306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80306, CurrentID=0, Status='live', Symbol='MED28', Aliases='1500003D12Rik, EG1, magicin', description='mediator complex subunit 28', other_designations='endothelial-derived gene 1|endothelial-derived protein 1|mediator of RNA polymerase II transcription, subunit 28 homolog|merlin and Grb2-interacting cytoskeletal protein|tumor angiogenesis marker EG-1', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17614628, end_position_on_the_genomic_accession=17625628, orientation='plus', exon_count=4, OMIM=610311, no_hdr0=''),
  80307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80307, CurrentID=0, Status='live', Symbol='FER1L4', Aliases='C20orf124', description='fer-1-like family member 4, pseudogene (functional)', other_designations='fer-1-like 4, pseudogene (functional)', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35558737, end_position_on_the_genomic_accession=35607562, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  80308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80308, CurrentID=0, Status='live', Symbol='FLAD1', Aliases='FAD1, FADS, PP591', description='flavin adenine dinucleotide synthetase 1', other_designations='FAD pyrophosphorylase|FAD-synthetase|FMN adenylyltransferase|Fad1, flavin adenine dinucleotide synthetase, homolog', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154983294, end_position_on_the_genomic_accession=154993111, orientation='plus', exon_count=8, OMIM=610595, no_hdr0=''),
  80309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80309, CurrentID=0, Status='live', Symbol='SPHKAP', Aliases='SKIP', description='SPHK1 interactor, AKAP domain containing', other_designations='SPHK1 (sphingosine kinase type 1) interacting protein|SPHK1-interactor and AKAP domain-containing protein|sphingosine kinase type 1-interacting protein', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227979954, end_position_on_the_genomic_accession=228181962, orientation='minus', exon_count=16, OMIM=611646, no_hdr0=''),
  80310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80310, CurrentID=0, Status='live', Symbol='PDGFD', Aliases='IEGF, MSTP036, SCDGF-B, SCDGFB', description='platelet derived growth factor D', other_designations='PDGF-D|iris-expressed growth factor|spinal cord derived growth factor B|spinal cord-derived growth factor B|spinal cord-derived growth factor-B', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103907186, end_position_on_the_genomic_accession=104164299, orientation='minus', exon_count=7, OMIM=609673, no_hdr0=''),
  80311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80311, CurrentID=0, Status='live', Symbol='KLHL15', Aliases='HEL-S-305', description='kelch like family member 15', other_designations='epididymis secretory protein Li 305|kelch-like 15|kelch-like family member 15', map_location='Xp22.1-p21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23983716, end_position_on_the_genomic_accession=24027186, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80312, CurrentID=0, Status='live', Symbol='TET1', Aliases='CXXC6, LCX, MLL-TET1-MLL, bA119F7.1, TET1', description='tet methylcytosine dioxygenase 1', other_designations='CXXC finger 6|CXXC zinc finger 6|CXXC-type zinc finger protein 6|TET1 splice variant VP_DE4|TET1 splice variant VP_DE456|leukemia-associated protein with a CXXC domain|ten-eleven translocation 1 gene protein|ten-eleven translocation-1|tet oncogene 1', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68560360, end_position_on_the_genomic_accession=68694487, orientation='plus', exon_count=16, OMIM=607790, no_hdr0=''),
  80313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80313, CurrentID=0, Status='live', Symbol='LRRC27', Aliases='', description='leucine rich repeat containing 27', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132330063, end_position_on_the_genomic_accession=132381506, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  80314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80314, CurrentID=0, Status='live', Symbol='EPC1', Aliases='Epl1', description='enhancer of polycomb homolog 1 (Drosophila)', other_designations='', map_location='10p11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32267716, end_position_on_the_genomic_accession=32378798, orientation='minus', exon_count=18, OMIM=610999, no_hdr0=''),
  80315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80315, CurrentID=0, Status='live', Symbol='CPEB4', Aliases='CPE-BP4, hCPEB-4', description='cytoplasmic polyadenylation element binding protein 4', other_designations='CPE-binding protein 4', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173888328, end_position_on_the_genomic_accession=173960310, orientation='plus', exon_count=11, OMIM=610607, no_hdr0=''),
  80316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80316, CurrentID=0, Status='live', Symbol='PPP1R2P9', Aliases='I-4', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 9', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 9|type 1 protein phosphatase inhibitor', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=42777366, end_position_on_the_genomic_accession=42778235, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80317, CurrentID=0, Status='live', Symbol='ZKSCAN3', Aliases='ZF47, ZFP306, ZNF306, ZNF309, ZSCAN13, ZSCAN35, Zfp47, dJ874C20.1, dJ874C20.1., zfp-47', description='zinc finger with KRAB and SCAN domains 3', other_designations='zinc finger and SCAN domain-containing protein 13|zinc finger protein 306|zinc finger protein 309|zinc finger protein 47 homolog|zinc finger protein zfp47', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28349913, end_position_on_the_genomic_accession=28369177, orientation='plus', exon_count=8, OMIM=612791, no_hdr0=''),
  80318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80318, CurrentID=0, Status='live', Symbol='GKAP1', Aliases='FKSG21, GKAP42', description='G kinase anchoring protein 1', other_designations='cGMP-dependent protein kinase anchoring protein 42kDa|cGMP-dependent protein kinase-anchoring protein of 42 kDa|protein kinase anchoring protein GKAP42', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83739421, end_position_on_the_genomic_accession=83817837, orientation='minus', exon_count=16, OMIM=611356, no_hdr0=''),
  80319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80319, CurrentID=0, Status='live', Symbol='CXXC4', Aliases='IDAX', description='CXXC finger protein 4', other_designations='CXXC finger 4|Dvl-binding protein IDAX (inhibition of the Dvl and Axin complex)|inhibition of the Dvl and axin complex protein', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104468306, end_position_on_the_genomic_accession=104494901, orientation='minus', exon_count=4, OMIM=611645, no_hdr0=''),
  80320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80320, CurrentID=0, Status='live', Symbol='SP6', Aliases='EPFN, EPIPROFIN, KLF14', description='Sp6 transcription factor', other_designations='Kruppel-like factor 14|krueppel-like factor 14', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47844908, end_position_on_the_genomic_accession=47876190, orientation='minus', exon_count=4, OMIM=608613, no_hdr0=''),
  80321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80321, CurrentID=0, Status='live', Symbol='CEP70', Aliases='BITE', description='centrosomal protein 70kDa', other_designations='p10-binding protein', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138494339, end_position_on_the_genomic_accession=138594383, orientation='minus', exon_count=23, OMIM=614310, no_hdr0=''),
  645877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645877, CurrentID=0, Status='live', Symbol='LOC645877', Aliases='', description='acyl-CoA binding domain containing 7 pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56846313, end_position_on_the_genomic_accession=56846848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80323, CurrentID=0, Status='live', Symbol='CCDC68', Aliases='SE57-1', description='coiled-coil domain containing 68', other_designations='CTCL tumor antigen se57-1|CTCL-associated antigen se57-1|cutaneous T-cell lymphoma associated antigen|cutaneous T-cell lymphoma-associated antigen se57-1', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54901509, end_position_on_the_genomic_accession=54959508, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  80324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80324, CurrentID=0, Status='live', Symbol='PUS1', Aliases='MLASA1', description='pseudouridylate synthase 1', other_designations='mitochondrial tRNA pseudouridine synthase A|tRNA pseudouridine(38-40) synthase|tRNA pseudouridylate synthase I|tRNA uridine isomerase I|tRNA-uridine isomerase I', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131929200, end_position_on_the_genomic_accession=131943861, orientation='plus', exon_count=9, OMIM=608109, no_hdr0=''),
  80325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80325, CurrentID=0, Status='live', Symbol='ABTB1', Aliases='BPOZ, BTB3, BTBD21, EF1ABP, PP2259', description='ankyrin repeat and BTB (POZ) domain containing 1', other_designations='elongation factor 1A-binding protein', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127672938, end_position_on_the_genomic_accession=127680926, orientation='plus', exon_count=14, OMIM=608308, no_hdr0=''),
  80326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80326, CurrentID=0, Status='live', Symbol='WNT10A', Aliases='OODD, SSPS, STHAG4', description='wingless-type MMTV integration site family member 10A', other_designations='wingless-type MMTV integration site family, member 10A', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218874323, end_position_on_the_genomic_accession=218893931, orientation='plus', exon_count=6, OMIM=606268, no_hdr0=''),
  80327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80327, CurrentID=0, Status='live', Symbol='MAPRE1P1', Aliases='MAPRE1P', description='MAPRE1 pseudogene 1', other_designations='MAPRE1P pseudogene|microtubule-associated protein, RP/EB family, member 1 pseudogene', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=135623584, end_position_on_the_genomic_accession=135626032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80328, CurrentID=0, Status='live', Symbol='ULBP2', Aliases='ALCAN-alpha, N2DL2, NKG2DL2, RAET1H', description='UL16 binding protein 2', other_designations='retinoic acid early transcript 1H', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149941938, end_position_on_the_genomic_accession=149949235, orientation='plus', exon_count=5, OMIM=605698, no_hdr0=''),
  80329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80329, CurrentID=0, Status='live', Symbol='ULBP1', Aliases='N2DL-1, NKG2DL1, RAET1I', description='UL16 binding protein 1', other_designations='UL16-binding protein 1|alcan-beta|retinoic acid early transcript 1I', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149964007, end_position_on_the_genomic_accession=149973710, orientation='plus', exon_count=5, OMIM=605697, no_hdr0=''),
  80331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80331, CurrentID=0, Status='live', Symbol='DNAJC5', Aliases='CLN4, CLN4B, CSPA, NCL, mir-941-2, mir-941-3, mir-941-4, mir-941-5, DNAJC5', description='DnaJ heat shock protein family (Hsp40) member C5', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5|ceroid-lipofuscinosis neuronal protein 4|cysteine string protein alpha', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63895102, end_position_on_the_genomic_accession=63936031, orientation='plus', exon_count=9, OMIM=611203, no_hdr0=''),
  80332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80332, CurrentID=0, Status='live', Symbol='ADAM33', Aliases='C20orf153, DJ964F7.1', description='ADAM metallopeptidase domain 33', other_designations='ADAM 33|a disintegrin and metalloprotease 33|a disintegrin and metalloproteinase domain 33|disintegrin and reprolysin metalloproteinase family protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3667973, end_position_on_the_genomic_accession=3682131, orientation='minus', exon_count=23, OMIM=607114, no_hdr0=''),
  80333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80333, CurrentID=0, Status='live', Symbol='KCNIP4', Aliases='CALP, KCHIP4', description='Kv channel interacting protein 4', other_designations='a-type potassium channel modulatory protein 4|calsenilin-like protein|potassium channel interacting protein 4|potassium channel-interacting protein 4', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20728606, end_position_on_the_genomic_accession=21948801, orientation='minus', exon_count=15, OMIM=608182, no_hdr0=''),
  80335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80335, CurrentID=0, Status='live', Symbol='WDR82', Aliases='MST107, MSTP107, PRO2730, PRO34047, SWD2, TMEM113A, WDR82', description='WD repeat domain 82', other_designations='transmembrane protein 113', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52254422, end_position_on_the_genomic_accession=52287876, orientation='minus', exon_count=11, OMIM=611059, no_hdr0=''),
  80336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80336, CurrentID=0, Status='live', Symbol='PABPC1L', Aliases='C20orf1191, dJ1069P2.3, ePAB, PABPC1L', description='poly(A) binding protein, cytoplasmic 1-like', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44910058, end_position_on_the_genomic_accession=44958801, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105373432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373432, CurrentID=0, Status='live', Symbol='LOC105373432', Aliases='', description='uncharacterized LOC105373432', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12189643, end_position_on_the_genomic_accession=12192849, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80339, CurrentID=0, Status='live', Symbol='PNPLA3', Aliases='ADPN, C22orf20, iPLA(2)epsilon', description='patatin like phospholipase domain containing 3', other_designations='acylglycerol O-acyltransferase|adiponutrin|calcium-independent phospholipase A2-epsilon|iPLA2-epsilon|iPLA2epsilon|patatin-like phospholipase domain containing 3', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43923739, end_position_on_the_genomic_accession=43947568, orientation='plus', exon_count=9, OMIM=609567, no_hdr0=''),
  80341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80341, CurrentID=0, Status='live', Symbol='BPIFB2', Aliases='BPIL1, C20orf184, LPLUNC2, RYSR, dJ726C3.2', description='BPI fold containing family B member 2', other_designations='BPI fold containing family B, member 2|BPI-like 1|bactericidal/permeability-increasing protein-like 1|long palate, lung and nasal epithelium carcinoma-associated protein 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33007578, end_position_on_the_genomic_accession=33023709, orientation='plus', exon_count=16, OMIM=614108, no_hdr0=''),
  80342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80342, CurrentID=0, Status='live', Symbol='TRAF3IP3', Aliases='T3JAM', description='TRAF3 interacting protein 3', other_designations='TRAF3 interacting Jun N terminal kinase (JNK) activating modulator', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209756032, end_position_on_the_genomic_accession=209782323, orientation='plus', exon_count=19, OMIM=608255, no_hdr0=''),
  80343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80343, CurrentID=0, Status='live', Symbol='SEL1L2', Aliases='C20orf50', description='sel-1 suppressor of lin-12-like 2 (C. elegans)', other_designations='sel-1L2|suppressor of lin-12-like protein 2', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13849188, end_position_on_the_genomic_accession=13995333, orientation='minus', exon_count=25, OMIM=614289, no_hdr0=''),
  80344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80344, CurrentID=0, Status='live', Symbol='DCAF11', Aliases='GL014, PRO2389, WDR23', description='DDB1 and CUL4 associated factor 11', other_designations='WD repeat domain 23|WD repeat-containing protein 23', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24114697, end_position_on_the_genomic_accession=24125242, orientation='plus', exon_count=15, OMIM=613317, no_hdr0=''),
  80345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80345, CurrentID=0, Status='live', Symbol='ZSCAN16', Aliases='ZNF392, ZNF435, dJ265C24.3', description='zinc finger and SCAN domain containing 16', other_designations='zinc finger protein 392|zinc finger protein 435', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28124556, end_position_on_the_genomic_accession=28130086, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  80346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80346, CurrentID=0, Status='live', Symbol='REEP4', Aliases='C8orf20, PP432', description='receptor accessory protein 4', other_designations='receptor expression enhancing protein 4', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22138020, end_position_on_the_genomic_accession=22141935, orientation='minus', exon_count=8, OMIM=609349, no_hdr0=''),
  80347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80347, CurrentID=0, Status='live', Symbol='COASY', Aliases='DPCK, NBIA6, NBP, PPAT, UKR1, pOV-2', description='Coenzyme A synthase', other_designations='CoA synthase|bifunctional phosphopantetheine adenylyl transferase/dephospho CoA kinase|nucleotide binding protein|phosphopantetheine adenylyltransferase / dephosphocoenzyme A kinase', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42562074, end_position_on_the_genomic_accession=42566281, orientation='plus', exon_count=9, OMIM=609855, no_hdr0=''),
  80349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80349, CurrentID=0, Status='live', Symbol='WDR61', Aliases='REC14, SKI8', description='WD repeat domain 61', other_designations='SKI8 homolog|meiotic recombination REC14 protein homolog|recombination protein REC14', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78283233, end_position_on_the_genomic_accession=78299726, orientation='minus', exon_count=11, OMIM=609540, no_hdr0=''),
  80350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80350, CurrentID=0, Status='live', Symbol='LPAL2', Aliases='APOA2, APOAL, APOARGC, apo(a)rg-C', description='lipoprotein, Lp(a)-like 2, pseudogene', other_designations='apolipoprotein (a) related gene C|apolipoprotein A-II', map_location='6q26-q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160466555, end_position_on_the_genomic_accession=160511124, orientation='minus', exon_count=11, OMIM=611682, no_hdr0=''),
  80351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80351, CurrentID=0, Status='live', Symbol='TNKS2', Aliases='ARTD6, PARP-5b, PARP-5c, PARP5B, PARP5C, TANK2, TNKL, pART6', description='tankyrase, TRF1-interacting ankyrin-related ADP-ribose polymerase 2', other_designations='ADP-ribosyltransferase diphtheria toxin-like 6|TNKS-2|TRF1-interacting ankyrin-related ADP-ribose polymerase 2|poly [ADP-ribose] polymerase 5B|tankyrase 2|tankyrase II|tankyrase-like protein|tankyrase-related protein', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91798394, end_position_on_the_genomic_accession=91865475, orientation='plus', exon_count=30, OMIM=607128, no_hdr0=''),
  80352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80352, CurrentID=0, Status='live', Symbol='RNF39', Aliases='HZF, HZFW, LIRF', description='ring finger protein 39', other_designations='LTP (long-term potentiation) induced RING finger protein|protein HZFw', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30070266, end_position_on_the_genomic_accession=30075851, orientation='minus', exon_count=4, OMIM=607524, no_hdr0=''),
  105373435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373435, CurrentID=0, Status='live', Symbol='LOC105373435', Aliases='', description='uncharacterized LOC105373435', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13110189, end_position_on_the_genomic_accession=13128540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373336, CurrentID=0, Status='live', Symbol='LOC105373336', Aliases='', description='uncharacterized LOC105373336', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130527284, end_position_on_the_genomic_accession=130530102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  342510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342510, CurrentID=0, Status='live', Symbol='CD300E', Aliases='CD300LE, CLM-2, CLM2, CMRF35-A5, IREM-2, IREM2, PIgR-2, PIgR2', description='CD300e molecule', other_designations='CD300 antigen like family member E|CD300 antigen-like family member E|CD300e antigen|immune receptor expressed on myeloid cells 2|poly-Ig receptor 2|polymeric immunoglobulin receptor 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74609883, end_position_on_the_genomic_accession=74623758, orientation='minus', exon_count=4, OMIM=609801, no_hdr0=''),
  646909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646909, CurrentID=0, Status='live', Symbol='RPL17P33', Aliases='RPL17_19_888', description='ribosomal protein L17 pseudogene 33', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29632056, end_position_on_the_genomic_accession=29632683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646910, CurrentID=0, Status='live', Symbol='HSPD1P9', Aliases='HSPD1-14P', description='heat shock protein family D (Hsp60) member 1 pseudogene 9', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 9', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37992585, end_position_on_the_genomic_accession=37994765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373439, CurrentID=0, Status='live', Symbol='LOC105373439', Aliases='', description='uncharacterized LOC105373439', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=14778929, end_position_on_the_genomic_accession=15137440, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  80380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80380, CurrentID=0, Status='live', Symbol='PDCD1LG2', Aliases='B7DC, Btdc, CD273, PD-L2, PDCD1L2, PDL2, bA574F11.2', description='programmed cell death 1 ligand 2', other_designations='B7 dendritic cell molecule|B7-DC|PD-1 ligand 2|PD-1-ligand 2|PDCD1 ligand 2|butyrophilin B7-DC|programmed death ligand 2', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5510491, end_position_on_the_genomic_accession=5571282, orientation='plus', exon_count=7, OMIM=605723, no_hdr0=''),
  80381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80381, CurrentID=0, Status='live', Symbol='CD276', Aliases='4Ig-B7-H3, B7-H3, B7H3, B7RP-2', description='CD276 molecule', other_designations='B7 homolog 3|costimulatory molecule', map_location='15q23-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73684281, end_position_on_the_genomic_accession=73714518, orientation='plus', exon_count=12, OMIM=605715, no_hdr0=''),
  342527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342527, CurrentID=0, Status='live', Symbol='SMTNL2', Aliases='', description='smoothelin-like 2', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4583981, end_position_on_the_genomic_accession=4608319, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105373440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373440, CurrentID=0, Status='live', Symbol='LOC105373440', Aliases='', description='uncharacterized LOC105373440', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15561426, end_position_on_the_genomic_accession=15573831, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646913, CurrentID=0, Status='live', Symbol='SEPT7P3', Aliases='', description='septin 7 pseudogene 3', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35946290, end_position_on_the_genomic_accession=35974692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  342538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342538, CurrentID=0, Status='live', Symbol='NACA2', Aliases='ANAC, NACAL', description='nascent polypeptide-associated complex alpha subunit 2', other_designations='IgE autoantigen alpha-nascent polypeptide-associated complex|alpha-NAC protein|hom s 2.01 protein|nascent polypeptide-associated complex, alpha polypeptide, 2', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61590433, end_position_on_the_genomic_accession=61591202, orientation='minus', exon_count=1, OMIM=609274, no_hdr0=''),
  105373442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373442, CurrentID=0, Status='live', Symbol='LOC105373442', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5696237, end_position_on_the_genomic_accession=5700950, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479514, CurrentID=0, Status='live', Symbol='RN7SL803P', Aliases='', description='RNA, 7SL, cytoplasmic 803, pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421560, CurrentID=0, Status='live', Symbol='GCNT1P2', Aliases='', description='glucosaminyl (N-acetyl) transferase 1, core 2 pseudogene 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=22580241, end_position_on_the_genomic_accession=22581075, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373444, CurrentID=0, Status='live', Symbol='LOC105373444', Aliases='', description='uncharacterized LOC105373444', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16454995, end_position_on_the_genomic_accession=16478580, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373446, CurrentID=0, Status='live', Symbol='LOC105373446', Aliases='', description='uncharacterized LOC105373446', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16738038, end_position_on_the_genomic_accession=16743667, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373447, CurrentID=0, Status='live', Symbol='LOC105373447', Aliases='', description='uncharacterized LOC105373447', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17376561, end_position_on_the_genomic_accession=17399920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106479515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479515, CurrentID=0, Status='live', Symbol='RN7SL806P', Aliases='', description='RNA, 7SL, cytoplasmic 806, pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  342574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342574, CurrentID=0, Status='live', Symbol='KRT27', Aliases='KRT25C', description='keratin 27, type I', other_designations='CK-27|K25C|K27|cytokeratin-27|keratin 25C|keratin-25C|keratin-27|type I inner root sheath specific keratin 25 irs3|type I inner root sheath-specific keratin-K25irs3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40776808, end_position_on_the_genomic_accession=40782534, orientation='minus', exon_count=8, OMIM=616676, no_hdr0=''),
  100421561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421561, CurrentID=0, Status='live', Symbol='LOC100421561', Aliases='', description='family with sequence similarity 133, member A pseudogene', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61375058, end_position_on_the_genomic_accession=61377032, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130571, CurrentID=0, Status='live', Symbol='LOC100130571', Aliases='', description='coiled-coil domain containing 107 pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20983089, end_position_on_the_genomic_accession=20983565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373452, CurrentID=0, Status='live', Symbol='LOC105373452', Aliases='', description='uncharacterized LOC105373452', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18361590, end_position_on_the_genomic_accession=18385968, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130573, CurrentID=0, Status='live', Symbol='LOC100130573', Aliases='', description='uncharacterized LOC100130573', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=202713959, end_position_on_the_genomic_accession=202714567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130574, CurrentID=0, Status='live', Symbol='LOC100130574', Aliases='', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 1 pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54878283, end_position_on_the_genomic_accession=54878558, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  342618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342618, CurrentID=0, Status='live', Symbol='SLFN14', Aliases='', description='schlafen family member 14', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35545734, end_position_on_the_genomic_accession=35560836, orientation='minus', exon_count=7, OMIM=614958, no_hdr0=''),
  646927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646927, CurrentID=0, Status='live', Symbol='LOC646927', Aliases='', description='uncharacterized LOC646927', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41189751, end_position_on_the_genomic_accession=41190206, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131582, CurrentID=0, Status='live', Symbol='LOC100131582', Aliases='', description='uncharacterized LOC100131582', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=704611, end_position_on_the_genomic_accession=706429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373457, CurrentID=0, Status='live', Symbol='LOC105373457', Aliases='', description='uncharacterized LOC105373457', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19096587, end_position_on_the_genomic_accession=19100476, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  404216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404216, CurrentID=0, Status='live', Symbol='LINC01561', Aliases='C10orf85', description='long intergenic non-protein coding RNA 1561', other_designations='', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120597949, end_position_on_the_genomic_accession=120600123, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060346, CurrentID=0, Status='live', Symbol='LOC101060346', Aliases='', description='T-complex protein 11 X-linked protein 2-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102031123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102031123, CurrentID=0, Status='live', Symbol='LINC01059', Aliases='', description='long intergenic non-protein coding RNA 1059', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373461, CurrentID=0, Status='live', Symbol='LOC105373461', Aliases='', description='uncharacterized LOC105373461', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19711683, end_position_on_the_genomic_accession=19717471, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373337, CurrentID=0, Status='live', Symbol='LOC105373337', Aliases='', description='uncharacterized LOC105373337', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130827896, end_position_on_the_genomic_accession=130836188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928880, CurrentID=0, Status='live', Symbol='LOC101928880', Aliases='LA16c-444G7.2', description='uncharacterized LOC101928880', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88194281, end_position_on_the_genomic_accession=88195217, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130582, CurrentID=0, Status='live', Symbol='LOC100130582', Aliases='', description='myosin light chain 9 pseudogene', other_designations='myosin, light chain 9, regulatory pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7870852, end_position_on_the_genomic_accession=7871372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479518, CurrentID=0, Status='live', Symbol='RN7SL812P', Aliases='', description='RNA, 7SL, cytoplasmic 812, pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  342666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342666, CurrentID=0, Status='live', Symbol='LRRC37A11P', Aliases='', description='leucine-rich repeat containing 37 member A11, pseudogene', other_designations='leucine rich repeat containing 37, member A11, pseudogene|leucine rich repeat containing 37, member A3 pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39029906, end_position_on_the_genomic_accession=39053205, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  342667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342667, CurrentID=0, Status='live', Symbol='STAC2', Aliases='24b2, 24b2/STAC2', description='SH3 and cysteine rich domain 2', other_designations='SRC homology 3 and cysteine-rich domain-containing protein 2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39210536, end_position_on_the_genomic_accession=39225787, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  646936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646936, CurrentID=0, Status='live', Symbol='CTBP2P5', Aliases='', description='C-terminal binding protein 2 pseudogene 5', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48915265, end_position_on_the_genomic_accession=48916587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373465, CurrentID=0, Status='live', Symbol='LOC105373465', Aliases='', description='uncharacterized LOC105373465', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20405084, end_position_on_the_genomic_accession=20424964, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873546, CurrentID=0, Status='live', Symbol='RNA5SP293', Aliases='RN5S293', description='RNA, 5S ribosomal pseudogene 293', other_designations='RNA, 5S ribosomal 293', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107918866, end_position_on_the_genomic_accession=107918978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373466, CurrentID=0, Status='live', Symbol='LOC105373466', Aliases='', description='uncharacterized LOC105373466', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373467, CurrentID=0, Status='live', Symbol='LOC105373467', Aliases='', description='uncharacterized LOC105373467', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144412509, end_position_on_the_genomic_accession=144414069, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373468, CurrentID=0, Status='live', Symbol='LOC105373468', Aliases='', description='uncharacterized LOC105373468', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146938328, end_position_on_the_genomic_accession=146960004, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373469, CurrentID=0, Status='live', Symbol='LOC105373469', Aliases='', description='uncharacterized LOC105373469', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145281216, end_position_on_the_genomic_accession=145282608, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  342705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342705, CurrentID=0, Status='live', Symbol='PGDP1', Aliases='PGDL1', description='phosphogluconate dehydrogenase pseudogene 1', other_designations='phosphogluconate dehydrogenase-like 1', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31962129, end_position_on_the_genomic_accession=31963716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373470, CurrentID=0, Status='live', Symbol='LOC105373470', Aliases='', description='uncharacterized LOC105373470', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151922278, end_position_on_the_genomic_accession=151953555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130591, CurrentID=0, Status='live', Symbol='LOC100130591', Aliases='', description='uncharacterized LOC100130591', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=151448019, end_position_on_the_genomic_accession=151450365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104413890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104413890, CurrentID=0, Status='live', Symbol='ZBTB20-AS3', Aliases='', description='ZBTB20 antisense RNA 3', other_designations='ZBTB20 antisense RNA 3 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114873033, end_position_on_the_genomic_accession=114876514, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104413891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104413891, CurrentID=0, Status='live', Symbol='SAPCD1-AS1', Aliases='C6orf26-AS1', description='SAPCD1 antisense RNA 1', other_designations='SAPCD1 antisense RNA 1 (non-protein coding)|XXbac-BPG32J3.18', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31764135, end_position_on_the_genomic_accession=31765588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104413892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104413892, CurrentID=0, Status='live', Symbol='F10-AS1', Aliases='', description='F10 antisense RNA 1', other_designations='F10 antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113128155, end_position_on_the_genomic_accession=113130046, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  342732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342732, CurrentID=0, Status='live', Symbol='KRT8P5', Aliases='', description='keratin 8 pseudogene 5', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44320560, end_position_on_the_genomic_accession=44322288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  341112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341112, CurrentID=0, Status='live', Symbol='WARSP1', Aliases='', description='tryptophanyl-tRNA synthetase pseudogene 1', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59255468, end_position_on_the_genomic_accession=59258657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373477, CurrentID=0, Status='live', Symbol='LOC105373477', Aliases='', description='protein GVQW1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149983141, end_position_on_the_genomic_accession=149990952, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873340, CurrentID=0, Status='live', Symbol='MTND1P21', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 21', other_designations='MT-ND1 pseudogene 21', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20804281, end_position_on_the_genomic_accession=20804580, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373479, CurrentID=0, Status='live', Symbol='LOC105373479', Aliases='', description='uncharacterized LOC105373479', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=689814, end_position_on_the_genomic_accession=692453, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130600, CurrentID=0, Status='live', Symbol='CICP14', Aliases='', description='capicua transcriptional repressor pseudogene 14', other_designations='capicua homolog pseudogene 14', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128655051, end_position_on_the_genomic_accession=128658830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373481, CurrentID=0, Status='live', Symbol='LOC105373481', Aliases='', description='uncharacterized LOC105373481', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=905595, end_position_on_the_genomic_accession=916572, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  342776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342776, CurrentID=0, Status='live', Symbol='MRPS17P7', Aliases='MRPS17P4, MRPS17P8', description='mitochondrial ribosomal protein S17 pseudogene 7', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57763720, end_position_on_the_genomic_accession=57764218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253738, CurrentID=0, Status='live', Symbol='EBF3', Aliases='COE3, EBF-3, O/E-2, OE-2', description='early B-cell factor 3', other_designations='olf-1/EBF-like 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129835232, end_position_on_the_genomic_accession=129964281, orientation='minus', exon_count=20, OMIM=607407, no_hdr0=''),
  342784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342784, CurrentID=0, Status='live', Symbol='LOC342784', Aliases='', description='general transcription factor IIH subunit 1 pseudogene', other_designations='general transcription factor IIH, polypeptide 1, 62kDa pseudogene', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60196554, end_position_on_the_genomic_accession=60198173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130603, CurrentID=0, Status='live', Symbol='RBM22P13', Aliases='', description='RNA binding motif protein 22 pseudogene 13', other_designations='RNA binding motif protein 22 pseudogene 12', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31863550, end_position_on_the_genomic_accession=31864789, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373484, CurrentID=0, Status='live', Symbol='LOC105373484', Aliases='', description='uncharacterized LOC105373484', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13000448, end_position_on_the_genomic_accession=13213628, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375299, CurrentID=0, Status='live', Symbol='LOC105375299', Aliases='', description='uncharacterized LOC105375299', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57401997, end_position_on_the_genomic_accession=57404717, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373486, CurrentID=0, Status='live', Symbol='LOC105373486', Aliases='', description='uncharacterized LOC105373486', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94651639, end_position_on_the_genomic_accession=94668792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373338, CurrentID=0, Status='live', Symbol='LOC105373338', Aliases='', description='uncharacterized LOC105373338', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130802636, end_position_on_the_genomic_accession=130811799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  342808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342808, CurrentID=0, Status='live', Symbol='RPS2P6', Aliases='RPS2_22_1612', description='ribosomal protein S2 pseudogene 6', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70430210, end_position_on_the_genomic_accession=70431161, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130607, CurrentID=0, Status='live', Symbol='LOC100130607', Aliases='', description='WD repeat domain 4 pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135464316, end_position_on_the_genomic_accession=135465484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379664, CurrentID=0, Status='live', Symbol='LOC105379664', Aliases='', description='uncharacterized LOC105379664', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373488, CurrentID=0, Status='live', Symbol='LOC105373488', Aliases='', description='uncharacterized LOC105373488', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95074832, end_position_on_the_genomic_accession=95076293, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373489, CurrentID=0, Status='live', Symbol='LOC105373489', Aliases='', description='uncharacterized LOC105373489', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95207503, end_position_on_the_genomic_accession=95222682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373490, CurrentID=0, Status='live', Symbol='LOC105373490', Aliases='', description='uncharacterized LOC105373490', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95637090, end_position_on_the_genomic_accession=95641984, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873352, CurrentID=0, Status='live', Symbol='MTND4P20', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 20', other_designations='MT-ND4 pseudogene 20', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94940304, end_position_on_the_genomic_accession=94941495, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375084, CurrentID=0, Status='live', Symbol='LOC105375084', Aliases='', description='uncharacterized LOC105375084', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50797933, end_position_on_the_genomic_accession=50802822, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130612, CurrentID=0, Status='live', Symbol='LOC100130612', Aliases='', description='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27838497, end_position_on_the_genomic_accession=27840412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80700, CurrentID=0, Status='live', Symbol='UBXN6', Aliases='UBXD1, UBXDC2', description='UBX domain protein 6', other_designations='UBX domain containing 1|UBX domain-containing 1|UBX domain-containing 2|UBX domain-containing protein 1', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4445006, end_position_on_the_genomic_accession=4457794, orientation='minus', exon_count=12, OMIM=611946, no_hdr0=''),
  80701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80701, CurrentID=0, Status='live', Symbol='AKR1D1P1', Aliases='AKR1D1P, SRD5BP1', description='aldo-keto reductase family 1, member D1 pseudogene 1', other_designations='aldo-keto reductase family 1, member D1 (delta 4-3-ketosteroid-5-beta-reductase) pseudogene|steroid-5-beta-reductase, beta polypeptide pseudogene 1', map_location='1q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167519550, end_position_on_the_genomic_accession=167536087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379665, CurrentID=0, Status='live', Symbol='LOC105379665', Aliases='', description='DLA class II histocompatibility antigen, DR-1 beta chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373493, CurrentID=0, Status='live', Symbol='LOC105373493', Aliases='', description='uncharacterized LOC105373493', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96035328, end_position_on_the_genomic_accession=96038215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80704, CurrentID=0, Status='live', Symbol='SLC19A3', Aliases='BBGD, THMD2, THTR2', description='solute carrier family 19 (thiamine transporter), member 3', other_designations='solute carrier family 19, member 3|thTr-2', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227685210, end_position_on_the_genomic_accession=227718029, orientation='minus', exon_count=9, OMIM=606152, no_hdr0=''),
  80705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80705, CurrentID=0, Status='live', Symbol='TSGA10', Aliases='CEP4L, CT79', description='testis specific 10', other_designations='cancer/testis antigen 79|testis development protein NYD-SP7|testis specific, 10', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98997261, end_position_on_the_genomic_accession=99154920, orientation='minus', exon_count=28, OMIM=607166, no_hdr0=''),
  342850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342850, CurrentID=0, Status='live', Symbol='ANKRD62', Aliases='DKFZp779B1634', description='ankyrin repeat domain 62', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12072533, end_position_on_the_genomic_accession=12129764, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100873355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873355, CurrentID=0, Status='live', Symbol='TEKT4P3', Aliases='', description='tektin 4 pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130313940, end_position_on_the_genomic_accession=130314411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80709, CurrentID=0, Status='live', Symbol='AKNA', Aliases='', description='AT-hook transcription factor', other_designations='AKNA transcript F2|AT-hook transcription factor AKNA', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114334149, end_position_on_the_genomic_accession=114398551, orientation='minus', exon_count=33, OMIM=605729, no_hdr0=''),
  80712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80712, CurrentID=0, Status='live', Symbol='ESX1', Aliases='ESX1L, ESXR1', description='ESX homeobox 1', other_designations='ESX1-related protein|extraembryonic, spermatogenesis, homeobox 1 homolog', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104250038, end_position_on_the_genomic_accession=104254918, orientation='minus', exon_count=4, OMIM=300154, no_hdr0=''),
  80713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80713, CurrentID=0, Status='live', Symbol='MTIF2P1', Aliases='', description='mitochondrial translational initiation factor 2 pseudogene 1', other_designations='', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121502325, end_position_on_the_genomic_accession=121503917, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80714, CurrentID=0, Status='live', Symbol='PBX4', Aliases='', description='pre-B-cell leukemia homeobox 4', other_designations='homeobox protein PBX4', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19561707, end_position_on_the_genomic_accession=19618916, orientation='minus', exon_count=10, OMIM=608127, no_hdr0=''),
  100873357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873357, CurrentID=0, Status='live', Symbol='RNA5SP25', Aliases='RN5S25', description='RNA, 5S ribosomal pseudogene 25', other_designations='RNA, 5S ribosomal 25', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21369790, end_position_on_the_genomic_accession=21369899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  342865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342865, CurrentID=0, Status='live', Symbol='VSTM2B', Aliases='', description='V-set and transmembrane domain containing 2B', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29525431, end_position_on_the_genomic_accession=29564551, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  80723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80723, CurrentID=0, Status='live', Symbol='SLC35G2', Aliases='TMEM22', description='solute carrier family 35 member G2', other_designations='solute carrier family 35, member G2|transmembrane protein 22', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136819019, end_position_on_the_genomic_accession=136855892, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  80724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80724, CurrentID=0, Status='live', Symbol='ACAD10', Aliases='', description='acyl-CoA dehydrogenase family member 10', other_designations='ACAD-10|acyl-CoA dehydrogenase family, member 10|acyl-Coenzyme A dehydrogenase family, member 10', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111686053, end_position_on_the_genomic_accession=111757107, orientation='plus', exon_count=22, OMIM=611181, no_hdr0=''),
  80725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80725, CurrentID=0, Status='live', Symbol='SRCIN1', Aliases='P140, SNIP', description='SRC kinase signaling inhibitor 1', other_designations='P130Cas-associated protein|SNAP-25-interacting protein|SNAP25-interacting protein|p140Cap', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38530024, end_position_on_the_genomic_accession=38605930, orientation='minus', exon_count=19, OMIM=610786, no_hdr0=''),
  80726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80726, CurrentID=0, Status='live', Symbol='KIAA1683', Aliases='', description='KIAA1683', other_designations='', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18257096, end_position_on_the_genomic_accession=18274509, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80727, CurrentID=0, Status='live', Symbol='TTYH3', Aliases='', description='tweety family member 3', other_designations='hTTY3|tweety homolog 3', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2631969, end_position_on_the_genomic_accession=2664802, orientation='plus', exon_count=16, OMIM=608919, no_hdr0=''),
  80728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80728, CurrentID=0, Status='live', Symbol='ARHGAP39', Aliases='CrGAP, Vilse', description='Rho GTPase activating protein 39', other_designations='RhoGAP93B homolog|crossGAP homolog', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144529179, end_position_on_the_genomic_accession=144700520, orientation='minus', exon_count=16, OMIM=615880, no_hdr0=''),
  80731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80731, CurrentID=0, Status='live', Symbol='THSD7B', Aliases='', description='thrombospondin type 1 domain containing 7B', other_designations='thrombospondin, type I, domain containing 7B', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136765562, end_position_on_the_genomic_accession=137677717, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  105373498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373498, CurrentID=0, Status='live', Symbol='LOC105373498', Aliases='', description='uncharacterized LOC105373498', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97438098, end_position_on_the_genomic_accession=97439092, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80736, CurrentID=0, Status='live', Symbol='SLC44A4', Aliases='C6orf29, CTL4, NG22, TPPT', description='solute carrier family 44 member 4', other_designations='solute carrier family 44, member 4', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31863192, end_position_on_the_genomic_accession=31879046, orientation='minus', exon_count=22, OMIM=606107, no_hdr0=''),
  80737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80737, CurrentID=0, Status='live', Symbol='VWA7', Aliases='C6orf27, G7c, NG37', description='von Willebrand factor A domain containing 7', other_designations='protein G7c', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31765401, end_position_on_the_genomic_accession=31777331, orientation='minus', exon_count=18, OMIM=609693, no_hdr0=''),
  80739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80739, CurrentID=0, Status='live', Symbol='C6orf25', Aliases='G6b, NG31', description='chromosome 6 open reading frame 25', other_designations='immunoglobulin receptor', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31720808, end_position_on_the_genomic_accession=31726710, orientation='plus', exon_count=7, OMIM=606520, no_hdr0=''),
  80740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80740, CurrentID=0, Status='live', Symbol='LY6G6C', Aliases='C6orf24, G6c, NG24', description='lymphocyte antigen 6 complex, locus G6C', other_designations='lymphocyte antigen-6 G6C', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31718648, end_position_on_the_genomic_accession=31721734, orientation='minus', exon_count=3, OMIM=610435, no_hdr0=''),
  80741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80741, CurrentID=0, Status='live', Symbol='LY6G5C', Aliases='C6orf20, G5CA, LY6G5CB, NG33, LY6G5C', description='lymphocyte antigen 6 complex, locus G5C', other_designations='lymphocyte antigen-6 G5C', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31676684, end_position_on_the_genomic_accession=31680373, orientation='minus', exon_count=3, OMIM=610434, no_hdr0=''),
  80742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80742, CurrentID=0, Status='live', Symbol='PRR3', Aliases='CAT56', description='proline rich 3', other_designations='MHC class I region proline-rich protein CAT56|proline-rich polypeptide 3', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30556709, end_position_on_the_genomic_accession=30564696, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  80745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80745, CurrentID=0, Status='live', Symbol='THUMPD2', Aliases='C2orf8', description='THUMP domain containing 2', other_designations='SAM-dependent methyltransferase', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39736060, end_position_on_the_genomic_accession=39779276, orientation='minus', exon_count=13, OMIM=611751, no_hdr0=''),
  80746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80746, CurrentID=0, Status='live', Symbol='TSEN2', Aliases='PCH2B, SEN2, SEN2L', description='tRNA splicing endonuclease subunit 2', other_designations='TSEN2 tRNA splicing endonuclease subunit|hsSen2|tRNA-intron endonuclease Sen2', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12484432, end_position_on_the_genomic_accession=12539622, orientation='plus', exon_count=17, OMIM=608753, no_hdr0=''),
  342892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342892, CurrentID=0, Status='live', Symbol='ZNF850', Aliases='ZNF850P', description='zinc finger protein 850', other_designations='putative zinc finger protein ENSP00000330994', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36743478, end_position_on_the_genomic_accession=36772814, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873362, CurrentID=0, Status='live', Symbol='RNA5SP30', Aliases='RN5S30', description='RNA, 5S ribosomal pseudogene 30', other_designations='RNA, 5S ribosomal 30', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57991350, end_position_on_the_genomic_accession=57991464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373501, CurrentID=0, Status='live', Symbol='LOC105373501', Aliases='', description='uncharacterized LOC105373501', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98045461, end_position_on_the_genomic_accession=98052372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  342897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342897, CurrentID=0, Status='live', Symbol='NCCRP1', Aliases='FBXO50, NCCRP-1', description='non-specific cytotoxic cell receptor protein 1 homolog (zebrafish)', other_designations='NCC receptor protein 1 homolog|nonspecific cytotoxic cell receptor protein 1 homolog', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39196964, end_position_on_the_genomic_accession=39201882, orientation='plus', exon_count=7, OMIM=615901, no_hdr0=''),
  342898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342898, CurrentID=0, Status='live', Symbol='SYCN', Aliases='INSSA1, SYL', description='syncollin', other_designations='insulin synthesis associated 1|insulin synthesis-associated protein 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39202922, end_position_on_the_genomic_accession=39204266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  80755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80755, CurrentID=0, Status='live', Symbol='AARSD1', Aliases='', description='alanyl-tRNA synthetase domain containing 1', other_designations='alanyl-tRNA synthetase domain-containing protein 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42950526, end_position_on_the_genomic_accession=42964498, orientation='minus', exon_count=12, OMIM=613212, no_hdr0=''),
  342900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342900, CurrentID=0, Status='live', Symbol='LEUTX', Aliases='', description='leucine twenty homeobox', other_designations='arginine-fifty homeobox-like pseudogene|putative leucine-twenty homeobox', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39776594, end_position_on_the_genomic_accession=39786135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80757, CurrentID=0, Status='live', Symbol='TMEM121', Aliases='hole', description='transmembrane protein 121', other_designations='hole', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105526031, end_position_on_the_genomic_accession=105530202, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  80758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80758, CurrentID=0, Status='live', Symbol='PRR7', Aliases='', description='proline rich 7 (synaptic)', other_designations='synaptic proline-rich membrane protein', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177446344, end_position_on_the_genomic_accession=177456701, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80759, CurrentID=0, Status='live', Symbol='KHDC1', Aliases='C6orf147, C6orf148, Em:AC019205.8, NDG1, bA257K9.4', description='KH homology domain containing 1', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73241314, end_position_on_the_genomic_accession=73310365, orientation='minus', exon_count=9, OMIM=611688, no_hdr0=''),
  80760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80760, CurrentID=0, Status='live', Symbol='ITIH5', Aliases='ITI-HC5, PP14776', description='inter-alpha-trypsin inhibitor heavy chain family member 5', other_designations='ITI heavy chain H5|inter-alpha (globulin) inhibitor H5|inter-alpha-inhibitor heavy chain 5|inter-alpha-trypsin inhibitor heavy chain family, member 5', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7559270, end_position_on_the_genomic_accession=7666998, orientation='minus', exon_count=17, OMIM=609783, no_hdr0=''),
  80761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80761, CurrentID=0, Status='live', Symbol='UPK3B', Aliases='P35, UP3B, UPIIIB', description='uroplakin 3B', other_designations='uroplakin IIIb', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611887, no_hdr0=''),
  80762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80762, CurrentID=0, Status='live', Symbol='NDFIP1', Aliases='N4WBP5', description='Nedd4 family interacting protein 1', other_designations='Nedd4 WW domain-binding protein 5|breast cancer-associated protein SGA-1M|putative MAPK-activating protein PM13|putative NF-kappa-B-activating protein 164|putative NFKB and MAPK-activating protein', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142108759, end_position_on_the_genomic_accession=142154443, orientation='plus', exon_count=8, OMIM=612050, no_hdr0=''),
  80763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80763, CurrentID=0, Status='live', Symbol='SPX', Aliases='C12orf39', description='spexin hormone', other_designations='NPQ|neuropeptide Q', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21526322, end_position_on_the_genomic_accession=21531276, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80764, CurrentID=0, Status='live', Symbol='THAP7', Aliases='', description='THAP domain containing 7', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20999772, end_position_on_the_genomic_accession=21002115, orientation='minus', exon_count=5, OMIM=609518, no_hdr0=''),
  342909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342909, CurrentID=0, Status='live', Symbol='ZNF284', Aliases='ZNF284L', description='zinc finger protein 284', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44072144, end_position_on_the_genomic_accession=44087470, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80767, CurrentID=0, Status='live', Symbol='MMDFS', Aliases='', description='Multiple mitochondrial dysfunctions syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605711, no_hdr0=''),
  80768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80768, CurrentID=0, Status='live', Symbol='HMNJ', Aliases='MNDJ', description='Distal hereditary motor neuropathy, Jerash type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605726, no_hdr0=''),
  646976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646976, CurrentID=0, Status='live', Symbol='LOC646976', Aliases='', description='uncharacterized LOC646976', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178722503, end_position_on_the_genomic_accession=178726285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80770, CurrentID=0, Status='live', Symbol='CTRCT25', Aliases='CCPSO, CCSSO', description='Cataract, central pouch-like, with sutural opacities', other_designations='', map_location='15q21-q22', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605728, no_hdr0=''),
  80771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80771, CurrentID=0, Status='live', Symbol='MCOPCB2', Aliases='MOPCB2', description='Microphthalmia, colobomatous, 2', other_designations='', map_location='15q12-q15', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605738, no_hdr0=''),
  80772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80772, CurrentID=0, Status='live', Symbol='CPTP', Aliases='GLTPD1', description='ceramide-1-phosphate transfer protein', other_designations='GLTP domain-containing protein 1|glycolipid transfer protein domain containing 1|glycolipid transfer protein domain-containing protein 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1324756, end_position_on_the_genomic_accession=1328896, orientation='plus', exon_count=5, OMIM=615467, no_hdr0=''),
  100873366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873366, CurrentID=0, Status='live', Symbol='RNA5SP34', Aliases='RN5S34', description='RNA, 5S ribosomal pseudogene 34', other_designations='RNA, 5S ribosomal 34', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90135712, end_position_on_the_genomic_accession=90135841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80774, CurrentID=0, Status='live', Symbol='LIMD2', Aliases='', description='LIM domain containing 2', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63695889, end_position_on_the_genomic_accession=63701769, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80775, CurrentID=0, Status='live', Symbol='TMEM177', Aliases='', description='transmembrane protein 177', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119679167, end_position_on_the_genomic_accession=119682118, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80776, CurrentID=0, Status='live', Symbol='B9D2', Aliases='ICIS-1, MKS10, MKSR2', description='B9 protein domain 2', other_designations='MKS1-related protein 2|involved in cIlia stability-1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41354417, end_position_on_the_genomic_accession=41364540, orientation='minus', exon_count=5, OMIM=611951, no_hdr0=''),
  80777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80777, CurrentID=0, Status='live', Symbol='CYB5B', Aliases='CYB5-M, CYPB5M, OMB5', description='cytochrome b5 type B (outer mitochondrial membrane)', other_designations='cytochrome b5 outer mitochondrial membrane isoform|outer mitochondrial membrane cytochrome b5|type 2 cyt-b5', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69424595, end_position_on_the_genomic_accession=69466264, orientation='plus', exon_count=5, OMIM=611964, no_hdr0=''),
  80778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80778, CurrentID=0, Status='live', Symbol='ZNF34', Aliases='KOX32', description='zinc finger protein 34', other_designations='zinc finger protein 34 (KOX 32)|zinc finger protein KOX32', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144772224, end_position_on_the_genomic_accession=144787359, orientation='minus', exon_count=11, OMIM=194526, no_hdr0=''),
  100873367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873367, CurrentID=0, Status='live', Symbol='RNA5SP35', Aliases='RN5S35', description='RNA, 5S ribosomal pseudogene 35', other_designations='RNA, 5S ribosomal 35', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94025470, end_position_on_the_genomic_accession=94025598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80781, CurrentID=0, Status='live', Symbol='COL18A1', Aliases='KNO, KNO1, KS', description='collagen, type XVIII, alpha 1', other_designations='antiangiogenic agent|endostatin|multi-functional protein MFP', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45405137, end_position_on_the_genomic_accession=45513720, orientation='plus', exon_count=43, OMIM=120328, no_hdr0=''),
  342926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342926, CurrentID=0, Status='live', Symbol='ZNF677', Aliases='', description='zinc finger protein 677', other_designations='hypothetical protein MGC48625', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53235385, end_position_on_the_genomic_accession=53254858, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100873368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873368, CurrentID=0, Status='live', Symbol='RNA5SP36', Aliases='RN5S36', description='RNA, 5S ribosomal pseudogene 36', other_designations='RNA, 5S ribosomal 36', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94651397, end_position_on_the_genomic_accession=94651529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  342931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342931, CurrentID=0, Status='live', Symbol='RFPL4A', Aliases='RFPL4, RNF210', description='ret finger protein-like 4A', other_designations='RING finger protein 210|ret finger protein-like 4', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55757430, end_position_on_the_genomic_accession=55763419, orientation='plus', exon_count=4, OMIM=612601, no_hdr0=''),
  80789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80789, CurrentID=0, Status='live', Symbol='INTS5', Aliases='INT5, KIAA1698', description='integrator complex subunit 5', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62646848, end_position_on_the_genomic_accession=62653302, orientation='minus', exon_count=2, OMIM=611349, no_hdr0=''),
  80790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80790, CurrentID=0, Status='live', Symbol='CMIP', Aliases='TCMIP', description='c-Maf inducing protein', other_designations='tc-Mip', map_location='16q23', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81445170, end_position_on_the_genomic_accession=81711762, orientation='plus', exon_count=26, OMIM=610112, no_hdr0=''),
  100873369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873369, CurrentID=0, Status='live', Symbol='RNA5SP37', Aliases='RN5S37', description='RNA, 5S ribosomal pseudogene 37', other_designations='RNA, 5S ribosomal 37', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97362972, end_position_on_the_genomic_accession=97363101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379668, CurrentID=0, Status='live', Symbol='LOC105379668', Aliases='', description='uncharacterized LOC105379668', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873370, CurrentID=0, Status='live', Symbol='RNA5SP38', Aliases='RN5S38', description='RNA, 5S ribosomal pseudogene 38', other_designations='RNA, 5S ribosomal 38', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106155372, end_position_on_the_genomic_accession=106155490, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373509, CurrentID=0, Status='live', Symbol='LOC105373509', Aliases='', description='BBSome-interacting protein 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101068155, end_position_on_the_genomic_accession=101068752, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  342945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342945, CurrentID=0, Status='live', Symbol='ZSCAN22', Aliases='HKR2, ZNF50', description='zinc finger and SCAN domain containing 22', other_designations='GLI-Kruppel family member HKR2|krueppel-related zinc finger protein 2|zinc finger protein 50', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58327019, end_position_on_the_genomic_accession=58342349, orientation='plus', exon_count=6, OMIM=165260, no_hdr0=''),
  100873371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873371, CurrentID=0, Status='live', Symbol='RNA5SP39', Aliases='RN5S39', description='RNA, 5S ribosomal pseudogene 39', other_designations='RNA, 5S ribosomal 39', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108301212, end_position_on_the_genomic_accession=108301331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130630, CurrentID=0, Status='live', Symbol='RPL23P3', Aliases='RPL23_1_65', description='ribosomal protein L23 pseudogene 3', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78666271, end_position_on_the_genomic_accession=78666697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  341148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341148, CurrentID=0, Status='live', Symbol='CYCSP27', Aliases='HCP27', description='cytochrome c, somatic pseudogene 27', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74482250, end_position_on_the_genomic_accession=74482547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130631, CurrentID=0, Status='live', Symbol='EEF1B2P4', Aliases='', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 4', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106901238, end_position_on_the_genomic_accession=106902398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873373, CurrentID=0, Status='live', Symbol='RNA5SP102', Aliases='RN5S102', description='RNA, 5S ribosomal pseudogene 102', other_designations='RNA, 5S ribosomal 102', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=124932915, end_position_on_the_genomic_accession=124933020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80816, CurrentID=0, Status='live', Symbol='ASXL3', Aliases='BRPS, KIAA1713', description='additional sex combs like 3, transcriptional regulator', other_designations='additional sex combs like 3|additional sex combs like transcriptional regulator 3|additional sex combs-like protein 3', map_location='18q11', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=33578577, end_position_on_the_genomic_accession=33751195, orientation='plus', exon_count=20, OMIM=615115, no_hdr0=''),
  80817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80817, CurrentID=0, Status='live', Symbol='CEP44', Aliases='KIAA1712, PS1TP3', description='centrosomal protein 44kDa', other_designations='HBV PreS1-transactivated protein 3', map_location='4q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174283677, end_position_on_the_genomic_accession=174334355, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  80818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80818, CurrentID=0, Status='live', Symbol='ZNF436', Aliases='ZNF, Zfp46', description='zinc finger protein 436', other_designations='DNA-binding protein', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23359448, end_position_on_the_genomic_accession=23369864, orientation='minus', exon_count=4, OMIM=611703, no_hdr0=''),
  80820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80820, CurrentID=0, Status='live', Symbol='EEPD1', Aliases='HSPC107', description='endonuclease/exonuclease/phosphatase family domain containing 1', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36153227, end_position_on_the_genomic_accession=36301543, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  80821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80821, CurrentID=0, Status='live', Symbol='DDHD1', Aliases='PA-PLA1, PAPLA1, SPG28', description='DDHD domain containing 1', other_designations='phosphatidic acid-preferring phospholipase A1 homolog|phosphatidic acid-preferring phospholipase A1-like protein|spastic paraplegia 28 (autosomal recessive)', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53036740, end_position_on_the_genomic_accession=53153328, orientation='minus', exon_count=15, OMIM=614603, no_hdr0=''),
  80823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80823, CurrentID=0, Status='live', Symbol='BHLHB9', Aliases='GASP3, p60TRP', description='basic helix-loop-helix domain containing, class B, 9', other_designations='p60-like protein|transcription regulator of 60 kDa', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102720714, end_position_on_the_genomic_accession=102752441, orientation='plus', exon_count=5, OMIM=300921, no_hdr0=''),
  80824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80824, CurrentID=0, Status='live', Symbol='DUSP16', Aliases='MKP-7, MKP7', description='dual specificity phosphatase 16', other_designations='MAP kinase phosphatase 7|MAPK phosphatase-7|mitogen-activated protein kinase phosphatase 7', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12473282, end_position_on_the_genomic_accession=12562839, orientation='minus', exon_count=9, OMIM=607175, no_hdr0=''),
  100873375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873375, CurrentID=0, Status='live', Symbol='RNA5SP104', Aliases='RN5S104', description='RNA, 5S ribosomal pseudogene 104', other_designations='RNA, 5S ribosomal 104', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134231195, end_position_on_the_genomic_accession=134231312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80829, CurrentID=0, Status='live', Symbol='ZFP91', Aliases='DMS-8, DSM-8, FKSG11, PZF, ZFP-91, ZNF757', description='ZFP91 zinc finger protein', other_designations='drug resistance-associated sequence in melanoma 8|penta zinc finger protein|zinc finger protein 757|zinc finger protein 91 homolog|zinc finger protein homologous to Zfp91 in mouse', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58579114, end_position_on_the_genomic_accession=58621550, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  80830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80830, CurrentID=0, Status='live', Symbol='APOL6', Aliases='APOL-VI, APOLVI', description='apolipoprotein L6', other_designations='apolipoprotein L, 6|apolipoprotein L-VI', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35648360, end_position_on_the_genomic_accession=35668409, orientation='plus', exon_count=4, OMIM=607256, no_hdr0=''),
  80831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80831, CurrentID=0, Status='live', Symbol='APOL5', Aliases='APOL-V, APOLV', description='apolipoprotein L5', other_designations='apolipoprotein L, 5|apolipoprotein L-V', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35708424, end_position_on_the_genomic_accession=35730325, orientation='plus', exon_count=8, OMIM=607255, no_hdr0=''),
  80832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80832, CurrentID=0, Status='live', Symbol='APOL4', Aliases='APOL-IV, APOLIV', description='apolipoprotein L4', other_designations='apolipoprotein L, 4|apolipoprotein L-IV', map_location='22q11.2-q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36189128, end_position_on_the_genomic_accession=36204833, orientation='minus', exon_count=6, OMIM=607254, no_hdr0=''),
  342977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342977, CurrentID=0, Status='live', Symbol='NANOS3', Aliases='NANOS1L, NOS3', description='nanos homolog 3 (Drosophila)', other_designations='', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13876406, end_position_on_the_genomic_accession=13880757, orientation='plus', exon_count=2, OMIM=608229, no_hdr0=''),
  80834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80834, CurrentID=0, Status='live', Symbol='TAS1R2', Aliases='GPR71, T1R2, TR2', description='taste 1 receptor member 2', other_designations='G protein-coupled receptor 71|G-protein coupled receptor 71|sweet taste receptor T1R2|taste receptor, type 1, member 2', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18839599, end_position_on_the_genomic_accession=18859661, orientation='minus', exon_count=6, OMIM=606226, no_hdr0=''),
  342979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342979, CurrentID=0, Status='live', Symbol='PALM3', Aliases='', description='paralemmin 3', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14053367, end_position_on_the_genomic_accession=14062111, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100873377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873377, CurrentID=0, Status='live', Symbol='RNA5SP107', Aliases='RN5S107', description='RNA, 5S ribosomal pseudogene 107', other_designations='RNA, 5S ribosomal 107', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154590407, end_position_on_the_genomic_accession=154590528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373516, CurrentID=0, Status='live', Symbol='LOC105373516', Aliases='', description='uncharacterized LOC105373516', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102179734, end_position_on_the_genomic_accession=102182108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373517, CurrentID=0, Status='live', Symbol='LOC105373517', Aliases='', description='uncharacterized LOC105373517', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102519988, end_position_on_the_genomic_accession=102522382, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  342994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342994, CurrentID=0, Status='live', Symbol='RPL34P34', Aliases='RPL34_15_1639', description='ribosomal protein L34 pseudogene 34', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22368584, end_position_on_the_genomic_accession=22369094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80851, CurrentID=0, Status='live', Symbol='SH3BP5L', Aliases='', description='SH3-binding domain protein 5-like', other_designations='SH3BP-5-like', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248810446, end_position_on_the_genomic_accession=248825955, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80852, CurrentID=0, Status='live', Symbol='GRIP2', Aliases='', description='glutamate receptor interacting protein 2', other_designations='', map_location='3p24-p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14489111, end_position_on_the_genomic_accession=14602996, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  80853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80853, CurrentID=0, Status='live', Symbol='KDM7A', Aliases='JHDM1D', description='lysine (K)-specific demethylase 7A', other_designations='histone lysine demethylase JHDM1D|jmjC domain-containing histone demethylation protein 1D|jumonji C domain containing histone demethylase 1 homolog D|jumonji C domain-containing histone demethylase 1 homolog D|lysine-specific demethylase 7', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140084746, end_position_on_the_genomic_accession=140176956, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  80854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80854, CurrentID=0, Status='live', Symbol='SETD7', Aliases='KMT7, SET7, SET7/9, SET9', description='SET domain containing (lysine methyltransferase) 7', other_designations='H3-K4-HMTase SETD7|SET domain-containing protein 7|histone H3-K4 methyltransferase SETD7|histone H3-lysine 4-specific methyltransferase|lysine N-methyltransferase 7', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139495660, end_position_on_the_genomic_accession=139557157, orientation='minus', exon_count=9, OMIM=606594, no_hdr0=''),
  80856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80856, CurrentID=0, Status='live', Symbol='KIAA1715', Aliases='LNP, LNP1, Ul, ulnaless', description='KIAA1715', other_designations='2310011O18Rik|4921514L11Rik|Lunapark|limb and neural patterns', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175923862, end_position_on_the_genomic_accession=176005388, orientation='minus', exon_count=19, OMIM=610236, no_hdr0=''),
  100873380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873380, CurrentID=0, Status='live', Symbol='RNA5SP110', Aliases='RN5S110', description='RNA, 5S ribosomal pseudogene 110', other_designations='RNA, 5S ribosomal 110', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164858432, end_position_on_the_genomic_accession=164858546, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130639, CurrentID=0, Status='live', Symbol='LOC100130639', Aliases='', description='leucine rich repeat containing 3C pseudogene', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69074916, end_position_on_the_genomic_accession=69075678, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  80862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80862, CurrentID=0, Status='live', Symbol='ZNRD1-AS1', Aliases='C6orf12, HTEX4, NCRNA00171, TCTEX4, ZNRD1-AS, ZNRD1AS, ZNRD1AS1', description='ZNRD1 antisense RNA 1', other_designations='ZNRD1 antisense RNA (non-protein coding)|ZNRD1 antisense RNA 1 (non-protein coding)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30001011, end_position_on_the_genomic_accession=30061184, orientation='minus', exon_count=6, OMIM=615714, no_hdr0=''),
  80863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80863, CurrentID=0, Status='live', Symbol='PRRT1', Aliases='C6orf31, DSPD1, IFITMD7, NG5', description='proline-rich transmembrane protein 1', other_designations='dispanin subfamily D member 1|interferon induced transmembrane protein domain containing 7', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32148363, end_position_on_the_genomic_accession=32151943, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80864, CurrentID=0, Status='live', Symbol='EGFL8', Aliases='C6orf8, NG3', description='EGF-like-domain, multiple 8', other_designations='EGF-like protein 8|VE-statin-2|vascular endothelial statin-2', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32164605, end_position_on_the_genomic_accession=32168285, orientation='plus', exon_count=9, OMIM=609897, no_hdr0=''),
  100130640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130640, CurrentID=0, Status='live', Symbol='YWHAZP7', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 7', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64612150, end_position_on_the_genomic_accession=64613418, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80868, CurrentID=0, Status='live', Symbol='HCG4B', Aliases='HCG4P6, HCGIV-06, HCGIV-6, HCGIV.5, bCX67J3.3, bPG309N1.1, bQB90C11.3', description='HLA complex group 4B (non-protein coding)', other_designations='HCGIV-6 pseudogene|HLA complex group 4 pseudogene 6', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29924592, end_position_on_the_genomic_accession=29927215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80870, CurrentID=0, Status='live', Symbol='DKBI', Aliases='', description='Dyskeratosis, hereditary benign intraepithelial', other_designations='', map_location='4q35', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=127600, no_hdr0=''),
  80871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80871, CurrentID=0, Status='live', Symbol='ANMA', Aliases='', description='Anisomastia', other_designations='', map_location='16q13-q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605746, no_hdr0=''),
  100873383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873383, CurrentID=0, Status='live', Symbol='RNA5SP113', Aliases='RN5S113', description='RNA, 5S ribosomal pseudogene 113', other_designations='RNA, 5S ribosomal 113', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182048822, end_position_on_the_genomic_accession=182048905, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373522, CurrentID=0, Status='live', Symbol='LOC105373522', Aliases='', description='uncharacterized LOC105373522', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104144117, end_position_on_the_genomic_accession=104148317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873384, CurrentID=0, Status='live', Symbol='RNA5SP114', Aliases='RN5S114', description='RNA, 5S ribosomal pseudogene 114', other_designations='RNA, 5S ribosomal 114', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188276732, end_position_on_the_genomic_accession=188276844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370942, CurrentID=0, Status='live', Symbol='LOC105370942', Aliases='', description='gametogenetin-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873385, CurrentID=0, Status='live', Symbol='RNA5SP115', Aliases='RN5S115', description='RNA, 5S ribosomal pseudogene 115', other_designations='RNA, 5S ribosomal 115', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200849038, end_position_on_the_genomic_accession=200849144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373524, CurrentID=0, Status='live', Symbol='LOC105373524', Aliases='', description='uncharacterized LOC105373524', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104580821, end_position_on_the_genomic_accession=104581930, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  343035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343035, CurrentID=0, Status='live', Symbol='RD3', Aliases='C1orf36, LCA12', description='retinal degeneration 3', other_designations='retinal degeneration protein 3', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211476522, end_position_on_the_genomic_accession=211492917, orientation='minus', exon_count=3, OMIM=180040, no_hdr0=''),
  100873386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873386, CurrentID=0, Status='live', Symbol='RNA5SP116', Aliases='RN5S116', description='RNA, 5S ribosomal pseudogene 116', other_designations='RNA, 5S ribosomal 116', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208054161, end_position_on_the_genomic_accession=208054258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  80895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80895, CurrentID=0, Status='live', Symbol='ILKAP', Aliases='ILKAP23, PP2C-DELTA, ILKAP', description='ILK associated serine/threonine phosphatase', other_designations='integrin-linked kinase associated phosphatase|protein phosphatase 2c, delta isozyme', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238170401, end_position_on_the_genomic_accession=238203702, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  80896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80896, CurrentID=0, Status='live', Symbol='NPL', Aliases='C112, C1orf13, NAL1, NPL', description='N-acetylneuraminate pyruvate lyase (dihydrodipicolinate synthase)', other_designations='N-acetylneuraminic acid aldolase|NALase|dihydrodipicolinate synthetase homolog 1|sialate-pyruvate lyase|sialic acid aldolase', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182789449, end_position_on_the_genomic_accession=182830384, orientation='plus', exon_count=13, OMIM=611412, no_hdr0=''),
  105375307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375307, CurrentID=0, Status='live', Symbol='LOC105375307', Aliases='', description='uncharacterized LOC105375307', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873387, CurrentID=0, Status='live', Symbol='RNA5SP117', Aliases='RN5S117', description='RNA, 5S ribosomal pseudogene 117', other_designations='RNA, 5S ribosomal 117', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208920547, end_position_on_the_genomic_accession=208920664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343045, CurrentID=0, Status='live', Symbol='CYCSP51', Aliases='HCP1', description='cytochrome c, somatic pseudogene 51', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150903896, end_position_on_the_genomic_accession=150904213, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873388, CurrentID=0, Status='live', Symbol='RNA5SP118', Aliases='RN5S118', description='RNA, 5S ribosomal pseudogene 118', other_designations='RNA, 5S ribosomal 118', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209721913, end_position_on_the_genomic_accession=209722021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373527, CurrentID=0, Status='live', Symbol='LOC105373527', Aliases='', description='uncharacterized LOC105373527', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105103354, end_position_on_the_genomic_accession=105106155, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  343052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343052, CurrentID=0, Status='live', Symbol='LOC343052', Aliases='', description='immunoglobulin superfamily, DCC subclass, member 3 pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153793947, end_position_on_the_genomic_accession=153796302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373528, CurrentID=0, Status='live', Symbol='LOC105373528', Aliases='', description='uncharacterized LOC105373528', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105194218, end_position_on_the_genomic_accession=105197397, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873390, CurrentID=0, Status='live', Symbol='RNA5SP120', Aliases='RN5S120', description='RNA, 5S ribosomal pseudogene 120', other_designations='RNA, 5S ribosomal 120', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216793489, end_position_on_the_genomic_accession=216793600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373529, CurrentID=0, Status='live', Symbol='LOC105373529', Aliases='', description='uncharacterized LOC105373529', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105450327, end_position_on_the_genomic_accession=105454691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  343066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343066, CurrentID=0, Status='live', Symbol='AADACL4', Aliases='', description='arylacetamide deacetylase-like 4', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12644547, end_position_on_the_genomic_accession=12667086, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100873391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873391, CurrentID=0, Status='live', Symbol='RNA5SP121', Aliases='RN5S121', description='RNA, 5S ribosomal pseudogene 121', other_designations='RNA, 5S ribosomal 121', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227763052, end_position_on_the_genomic_accession=227763152, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343068, CurrentID=0, Status='live', Symbol='PRAMEF5', Aliases='PRAMEF23L, PRAMEF5', description='PRAME family member 5', other_designations='PRAME family member 23', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13254198, end_position_on_the_genomic_accession=13263435, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  343069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343069, CurrentID=0, Status='live', Symbol='HNRNPCL1', Aliases='HNRPCL1', description='heterogeneous nuclear ribonucleoprotein C-like 1', other_designations='hnRNP C-like-1|hnRNP core protein C-like 1', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12847377, end_position_on_the_genomic_accession=12848756, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  343070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343070, CurrentID=0, Status='live', Symbol='PRAMEF9', Aliases='', description='PRAME family member 9', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13175281, end_position_on_the_genomic_accession=13179132, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  343071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343071, CurrentID=0, Status='live', Symbol='PRAMEF10', Aliases='', description='PRAME family member 10', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12892896, end_position_on_the_genomic_accession=12898270, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873392, CurrentID=0, Status='live', Symbol='RNA5SP122', Aliases='RN5S122', description='RNA, 5S ribosomal pseudogene 122', other_designations='RNA, 5S ribosomal 122', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241584968, end_position_on_the_genomic_accession=241585075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474148, CurrentID=0, Status='live', Symbol='TTTY3B', Aliases='LNCRNA00122, NCRNA00122', description='testis-specific transcript, Y-linked 3B (non-protein coding)', other_designations='long intergenic non-protein coding RNA 122|testis-specific transcript, Y-linked 3, centromeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23936727, end_position_on_the_genomic_accession=23941622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  474149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474149, CurrentID=0, Status='live', Symbol='TTTY4B', Aliases='LINC00124, NCRNA00124', description='testis-specific transcript, Y-linked 4B (non-protein coding)', other_designations='long intergenic non-protein coding RNA 124|testis-specific transcript, Y-linked 4, middle', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24570202, end_position_on_the_genomic_accession=24607025, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  474150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474150, CurrentID=0, Status='live', Symbol='TTTY4C', Aliases='LINC00125, NCRNA00125', description='testis-specific transcript, Y-linked 4C (non-protein coding)', other_designations='long intergenic non-protein coding RNA 125|testis-specific transcript, Y-linked 4, telomeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25063083, end_position_on_the_genomic_accession=25099892, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  474151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474151, CurrentID=0, Status='live', Symbol='TTTY17B', Aliases='NCRNA00141', description='testis-specific transcript, Y-linked 17B (non-protein coding)', other_designations='testis-specific transcript, Y-linked 17, middle', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24485332, end_position_on_the_genomic_accession=24486463, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  474152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474152, CurrentID=0, Status='live', Symbol='TTTY17C', Aliases='NCRNA00142', description='testis-specific transcript, Y-linked 17C (non-protein coding)', other_designations='testis-specific transcript, Y-linked 17, telomeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25183643, end_position_on_the_genomic_accession=25184773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129155, CurrentID=0, Status='live', Symbol='LOC100129155', Aliases='', description='phosphatase, orphan 1 pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110645762, end_position_on_the_genomic_accession=110646555, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873394, CurrentID=0, Status='live', Symbol='RNA5SP124', Aliases='RN5S124', description='RNA, 5S ribosomal pseudogene 124', other_designations='RNA, 5S ribosomal 124', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14394648, end_position_on_the_genomic_accession=14394757, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474159, CurrentID=0, Status='live', Symbol='DYX9', Aliases='DXY9', description='dyslexia susceptibility 9', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300509, no_hdr0=''),
  100873395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873395, CurrentID=0, Status='live', Symbol='RNA5SP125', Aliases='RN5S125', description='RNA, 5S ribosomal pseudogene 125', other_designations='RNA, 5S ribosomal 125', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25025007, end_position_on_the_genomic_accession=25025126, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372467, CurrentID=0, Status='live', Symbol='LOC105372467', Aliases='', description='uncharacterized LOC105372467', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56347036, end_position_on_the_genomic_accession=56368062, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373534, CurrentID=0, Status='live', Symbol='LOC105373534', Aliases='', description='uncharacterized LOC105373534', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106468418, end_position_on_the_genomic_accession=106470142, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  474167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474167, CurrentID=0, Status='live', Symbol='HDGFP1', Aliases='HDGFP', description='hepatoma-derived growth factor pseudogene 1', other_designations='hepatoma-derived growth factor (high-mobility group protein 1-like) pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131646640, end_position_on_the_genomic_accession=131648237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474168, CurrentID=0, Status='live', Symbol='WG', Aliases='', description='Wegener granulomatosis', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608710, no_hdr0=''),
  474170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474170, CurrentID=0, Status='live', Symbol='LRRC37A2', Aliases='LRRC37', description='leucine-rich repeat containing 37 member A2', other_designations='c114 SLIT-like testicular protein|leucine rich repeat containing 37, member A2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46511690, end_position_on_the_genomic_accession=46555650, orientation='plus', exon_count=15, OMIM=616556, no_hdr0=''),
  474171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474171, CurrentID=0, Status='live', Symbol='STGC3', Aliases='', description='uncharacterized STGC3', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=49297518, end_position_on_the_genomic_accession=49298749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474174, CurrentID=0, Status='live', Symbol='SGCEP', Aliases='', description='sarcoglycan epsilon pseudogene', other_designations='sarcoglycan, epsilon pseudogene|sarcoglycan, epsilon, pseudogene', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144754509, end_position_on_the_genomic_accession=144756143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873397, CurrentID=0, Status='live', Symbol='RNA5SP127', Aliases='RN5S127', description='RNA, 5S ribosomal pseudogene 127', other_designations='RNA, 5S ribosomal 127', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31228715, end_position_on_the_genomic_accession=31228847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373340, CurrentID=0, Status='live', Symbol='LOC105373340', Aliases='', description='uncharacterized LOC105373340', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135007751, end_position_on_the_genomic_accession=135020204, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927400, CurrentID=0, Status='live', Symbol='LOC101927400', Aliases='', description='uncharacterized LOC101927400', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113325091, end_position_on_the_genomic_accession=113427915, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105373537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373537, CurrentID=0, Status='live', Symbol='LOC105373537', Aliases='', description='uncharacterized LOC105373537', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107529378, end_position_on_the_genomic_accession=107555805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  474185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474185, CurrentID=0, Status='live', Symbol='OA29', Aliases='', description='Osteoarthritis QTL 29', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474186, CurrentID=0, Status='live', Symbol='OA30', Aliases='', description='Osteoarthritis QTL 29', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474187, CurrentID=0, Status='live', Symbol='BP35', Aliases='', description='Blood pressure QTL 35', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474188, CurrentID=0, Status='live', Symbol='OA24', Aliases='', description='Osteoarthritis QTL 24', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474189, CurrentID=0, Status='live', Symbol='BP24', Aliases='', description='Blood pressure QTL 24', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474190, CurrentID=0, Status='live', Symbol='BP30', Aliases='', description='Blood pressure QTL 30', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  80975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80975, CurrentID=0, Status='live', Symbol='TMPRSS5', Aliases='SPINESIN', description='transmembrane protease, serine 5', other_designations='', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113687546, end_position_on_the_genomic_accession=113714727, orientation='minus', exon_count=15, OMIM=606751, no_hdr0=''),
  474192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474192, CurrentID=0, Status='live', Symbol='BP32', Aliases='', description='Blood pressure QTL 32', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474193, CurrentID=0, Status='live', Symbol='BP33', Aliases='', description='Blood pressure QTL 33', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474194, CurrentID=0, Status='live', Symbol='BP34', Aliases='', description='Blood pressure QTL 34', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474195, CurrentID=0, Status='live', Symbol='OA16', Aliases='', description='Osteoarthritis QTL 16', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474196, CurrentID=0, Status='live', Symbol='OA25', Aliases='', description='Osteoarthritis QTL 25', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474197, CurrentID=0, Status='live', Symbol='BP22', Aliases='', description='Blood pressure QTL 22', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474198, CurrentID=0, Status='live', Symbol='BP23', Aliases='', description='Blood pressure QTL 23', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474199, CurrentID=0, Status='live', Symbol='BP9', Aliases='', description='Blood pressure QTL 9', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474200, CurrentID=0, Status='live', Symbol='BW22', Aliases='', description='Body weight QTL 22', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474201, CurrentID=0, Status='live', Symbol='BP45', Aliases='', description='Blood pressure QTL 45', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474202, CurrentID=0, Status='live', Symbol='BP46', Aliases='', description='Blood pressure QTL 46', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474203, CurrentID=0, Status='live', Symbol='BP47', Aliases='', description='Blood pressure QTL 47', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474204, CurrentID=0, Status='live', Symbol='BP48', Aliases='', description='Blood pressure QTL 48', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474205, CurrentID=0, Status='live', Symbol='BP49', Aliases='', description='Blood pressure QTL 49', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474206, CurrentID=0, Status='live', Symbol='BP51', Aliases='', description='Blood pressure QTL 51', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474207, CurrentID=0, Status='live', Symbol='BP50', Aliases='', description='Blood pressure QTL 50', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474208, CurrentID=0, Status='live', Symbol='BP52', Aliases='', description='Blood pressure QTL 52', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474209, CurrentID=0, Status='live', Symbol='BP53', Aliases='', description='Blood pressure QTL 53', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474210, CurrentID=0, Status='live', Symbol='BP54', Aliases='', description='Blood pressure QTL 54', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474211, CurrentID=0, Status='live', Symbol='BP55', Aliases='', description='Blood pressure QTL 55', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474212, CurrentID=0, Status='live', Symbol='BP17', Aliases='', description='Blood pressure QTL 17', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474213, CurrentID=0, Status='live', Symbol='BP18', Aliases='', description='Blood pressure QTL 18', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474214, CurrentID=0, Status='live', Symbol='BP19', Aliases='', description='Blood pressure QTL 19', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474215, CurrentID=0, Status='live', Symbol='BP21', Aliases='', description='Blood pressure QTL 21', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474216, CurrentID=0, Status='live', Symbol='BP20', Aliases='', description='Blood pressure QTL 20', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474217, CurrentID=0, Status='live', Symbol='BP36', Aliases='', description='Blood pressure QTL 36', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474218, CurrentID=0, Status='live', Symbol='BP37', Aliases='', description='Blood pressure QTL 37', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474219, CurrentID=0, Status='live', Symbol='BP39', Aliases='', description='Blood pressure QTL 39', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474220, CurrentID=0, Status='live', Symbol='BP38', Aliases='', description='Blood pressure QTL 38', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474221, CurrentID=0, Status='live', Symbol='RA1', Aliases='', description='Rheumatoid arthritis QTL 01', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474222, CurrentID=0, Status='live', Symbol='RA2', Aliases='', description='Rheumatoid arthritis QTL 01', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474223, CurrentID=0, Status='live', Symbol='RA3', Aliases='', description='Rheumatoid arthritis QTL 02', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474224, CurrentID=0, Status='live', Symbol='RA4', Aliases='', description='Rheumatoid arthritis QTL 04', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474225, CurrentID=0, Status='live', Symbol='RA5', Aliases='', description='Rheumatoid arthritis QTL 01', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474226, CurrentID=0, Status='live', Symbol='RA6', Aliases='', description='Rheumatoid arthritis QTL 06', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474227, CurrentID=0, Status='live', Symbol='RA7', Aliases='', description='Rheumatoid arthritis QTL 07', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474228, CurrentID=0, Status='live', Symbol='RA8', Aliases='', description='Rheumatoid arthritis QTL 08', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474229, CurrentID=0, Status='live', Symbol='RA9', Aliases='', description='Rheumatoid arthritis QTL 09', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474230, CurrentID=0, Status='live', Symbol='RA10', Aliases='', description='Rheumatoid arthritis QTL 10', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474231, CurrentID=0, Status='live', Symbol='RA11', Aliases='', description='Rheumatoid arthritis QTL 11', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474232, CurrentID=0, Status='live', Symbol='RA12', Aliases='', description='Rheumatoid arthritis QTL 12', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474233, CurrentID=0, Status='live', Symbol='RA13', Aliases='', description='Rheumatoid arthritis QTL 13', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474234, CurrentID=0, Status='live', Symbol='RA14', Aliases='', description='Rheumatoid arthritis QTL 14', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474235, CurrentID=0, Status='live', Symbol='RA15', Aliases='', description='Rheumatoid arthritis QTL 15', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474236, CurrentID=0, Status='live', Symbol='RA16', Aliases='', description='Rheumatoid arthritis QTL 16', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474237, CurrentID=0, Status='live', Symbol='RA17', Aliases='', description='Rheumatoid arthritis QTL 17', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474238, CurrentID=0, Status='live', Symbol='RA31', Aliases='', description='Rheumatoid arthritis QTL 31', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474239, CurrentID=0, Status='live', Symbol='RA18', Aliases='', description='Rheumatoid arthritis QTL 18', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474240, CurrentID=0, Status='live', Symbol='RA19', Aliases='', description='Rheumatoid arthritis QTL 19', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81025, CurrentID=0, Status='live', Symbol='GJA9', Aliases='CX58, CX59, GJA10', description='gap junction protein alpha 9', other_designations='connexin 59|connexin-58|gap junction alpha 10|gap junction protein, alpha 9, 59kDa', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38874067, end_position_on_the_genomic_accession=38881626, orientation='minus', exon_count=2, OMIM=611923, no_hdr0=''),
  343170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343170, CurrentID=0, Status='live', Symbol='OR14K1', Aliases='OR1-39, OR1.5.9, OR5AY1', description='olfactory receptor family 14 subfamily K member 1', other_designations='olfactory receptor OR1-39|olfactory receptor, family 14, subfamily K, member 1|olfactory receptor, family 5, subfamily AY, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247738615, end_position_on_the_genomic_accession=247739559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81027, CurrentID=0, Status='live', Symbol='TUBB1', Aliases='', description='tubulin beta 1 class VI', other_designations='tubulin, beta 1 class VI', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59019254, end_position_on_the_genomic_accession=59026654, orientation='plus', exon_count=4, OMIM=612901, no_hdr0=''),
  343172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343172, CurrentID=0, Status='live', Symbol='OR2T8', Aliases='OR2T8P', description='olfactory receptor family 2 subfamily T member 8', other_designations='olfactory receptor, family 2, subfamily T, member 8 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247912253, end_position_on_the_genomic_accession=247923152, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  81029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81029, CurrentID=0, Status='live', Symbol='WNT5B', Aliases='', description='wingless-type MMTV integration site family member 5B', other_designations='WNT-5B protein|wingless-type MMTV integration site family, member 5B', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1529891, end_position_on_the_genomic_accession=1647213, orientation='plus', exon_count=7, OMIM=606361, no_hdr0=''),
  81030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81030, CurrentID=0, Status='live', Symbol='ZBP1', Aliases='C20orf183, DAI, DLM-1, DLM1', description='Z-DNA binding protein 1', other_designations='DNA-dependent activator of IFN-regulatory factors|DNA-dependent activator of IRFs|DNA-dependent activator of interferon regulatory factors|tumor stroma and activated macrophage protein DLM-1', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57603846, end_position_on_the_genomic_accession=57620576, orientation='minus', exon_count=12, OMIM=606750, no_hdr0=''),
  81031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81031, CurrentID=0, Status='live', Symbol='SLC2A10', Aliases='ATS, GLUT10', description='solute carrier family 2 (facilitated glucose transporter), member 10', other_designations='GLUT-10|glucose transporter type 10', map_location='20q13.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46708347, end_position_on_the_genomic_accession=46736347, orientation='plus', exon_count=8, OMIM=606145, no_hdr0=''),
  474248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474248, CurrentID=0, Status='live', Symbol='RA27', Aliases='', description='Rheumatoid arthritis QTL 27', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81033, CurrentID=0, Status='live', Symbol='KCNH6', Aliases='ERG-2, ERG2, HERG2, Kv11.2, hERG-2', description='potassium channel, voltage gated eag related subfamily H, member 6', other_designations='eag-related gene member 2|ether-a-go-go-related protein 2|potassium voltage-gated channel, subfamily H (eag-related), member 6|voltage-gated potassium channel subunit Kv11.2', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63523334, end_position_on_the_genomic_accession=63548977, orientation='plus', exon_count=16, OMIM=608168, no_hdr0=''),
  81034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81034, CurrentID=0, Status='live', Symbol='SLC25A32', Aliases='MFT, MFTC', description='solute carrier family 25 (mitochondrial folate carrier), member 32', other_designations='solute carrier family 25 (mitochondrial folate carrier) , member 32|solute carrier family 25, member 32', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103398638, end_position_on_the_genomic_accession=103415335, orientation='minus', exon_count=8, OMIM=610815, no_hdr0=''),
  81035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81035, CurrentID=0, Status='live', Symbol='COLEC12', Aliases='CLP1, NSR2, SCARA4, SRCL', description='collectin subfamily member 12', other_designations='collectin placenta protein 1|collectin sub-family member 12|hCL-P1|nurse cell scavenger receptor 2|scavenger receptor class A, member 4|scavenger receptor with C-type lectin', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=319355, end_position_on_the_genomic_accession=500729, orientation='minus', exon_count=10, OMIM=607621, no_hdr0=''),
  474252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474252, CurrentID=0, Status='live', Symbol='BP10', Aliases='', description='Blood pressure QTL 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81037, CurrentID=0, Status='live', Symbol='CLPTM1L', Aliases='CRR9', description='CLPTM1-like', other_designations='CLPTM1-like protein|cisplatin resistance related protein CRR9p|cisplatin resistance-related protein 9', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1317755, end_position_on_the_genomic_accession=1345065, orientation='minus', exon_count=17, OMIM=612585, no_hdr0=''),
  474254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474254, CurrentID=0, Status='live', Symbol='BP12', Aliases='', description='Blood pressure QTL 12', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474255, CurrentID=0, Status='live', Symbol='BP13', Aliases='', description='Blood pressure QTL 13', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474256, CurrentID=0, Status='live', Symbol='BP1', Aliases='', description='Blood pressure QTL 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474257, CurrentID=0, Status='live', Symbol='BP2', Aliases='', description='Blood pressure QTL 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474258, CurrentID=0, Status='live', Symbol='BP4', Aliases='', description='Blood pressure QTL 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474259, CurrentID=0, Status='live', Symbol='BP3', Aliases='', description='Blood pressure QTL 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474260, CurrentID=0, Status='live', Symbol='BP14', Aliases='', description='Blood pressure QTL 14', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474261, CurrentID=0, Status='live', Symbol='BP15', Aliases='', description='Blood pressure QTL 15', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101735302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101735302, CurrentID=0, Status='live', Symbol='STAG3L5P', Aliases='', description='stromal antigen 3-like 5 pseudogene', other_designations='stromal antigen 3 pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100336079, end_position_on_the_genomic_accession=100341328, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  474263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474263, CurrentID=0, Status='live', Symbol='BW8', Aliases='', description='Body weight QTL 8', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474264, CurrentID=0, Status='live', Symbol='BW9', Aliases='', description='Body weight QTL 9', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474265, CurrentID=0, Status='live', Symbol='BW10', Aliases='', description='Body weight QTL 10', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81050, CurrentID=0, Status='live', Symbol='OR5AC2', Aliases='HSA1', description='olfactory receptor family 5 subfamily AC member 2', other_designations='olfactory receptor, family 5, subfamily AC, member 2', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98087173, end_position_on_the_genomic_accession=98088102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474267, CurrentID=0, Status='live', Symbol='BP26', Aliases='', description='Blood pressure QTL 26', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474268, CurrentID=0, Status='live', Symbol='BP28', Aliases='', description='Blood pressure QTL 28', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474269, CurrentID=0, Status='live', Symbol='BP27', Aliases='', description='Blood pressure QTL 27', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474270, CurrentID=0, Status='live', Symbol='BW2', Aliases='', description='Body weight QTL 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474271, CurrentID=0, Status='live', Symbol='BW3', Aliases='', description='Body weight QTL 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474272, CurrentID=0, Status='live', Symbol='BW4', Aliases='', description='Body weight QTL 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474273, CurrentID=0, Status='live', Symbol='BW5', Aliases='', description='Body weight QTL 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474274, CurrentID=0, Status='live', Symbol='BW7', Aliases='', description='Body weight QTL 7', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474275, CurrentID=0, Status='live', Symbol='BW6', Aliases='', description='Body weight QTL 6', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474276, CurrentID=0, Status='live', Symbol='BW23', Aliases='', description='Body weight QTL 23', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81061, CurrentID=0, Status='live', Symbol='OR11H1', Aliases='OR22-1', description='olfactory receptor family 11 subfamily H member 1', other_designations='olfactory receptor OR22-1|olfactory receptor, family 11, subfamily H, member 1|seven transmembrane helix receptor', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15528159, end_position_on_the_genomic_accession=15529139, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474278, CurrentID=0, Status='live', Symbol='BW25', Aliases='', description='Body weight QTL 25', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474279, CurrentID=0, Status='live', Symbol='BW26', Aliases='', description='Body weight QTL 26', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474280, CurrentID=0, Status='live', Symbol='BW27', Aliases='', description='Body weight QTL 27', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474281, CurrentID=0, Status='live', Symbol='BW28', Aliases='', description='Body weight QTL 28', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474282, CurrentID=0, Status='live', Symbol='BW29', Aliases='', description='Body weight QTL 29', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474283, CurrentID=0, Status='live', Symbol='BW30', Aliases='', description='Body weight QTL 30', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474284, CurrentID=0, Status='live', Symbol='BW31', Aliases='', description='Body weight QTL 31', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474285, CurrentID=0, Status='live', Symbol='OA1', Aliases='', description='Osteoarthritis QTL 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474286, CurrentID=0, Status='live', Symbol='OA2', Aliases='', description='Osteoarthritis QTL 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474287, CurrentID=0, Status='live', Symbol='OSTEAR3_H', Aliases='OA3_H', description='Osteoarthritis QTL 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474288, CurrentID=0, Status='live', Symbol='OA4', Aliases='', description='Osteoarthritis QTL 4', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474289, CurrentID=0, Status='live', Symbol='OA5', Aliases='', description='Osteoarthritis QTL 5', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474290, CurrentID=0, Status='live', Symbol='OA6', Aliases='', description='Osteoarthritis QTL 6', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474291, CurrentID=0, Status='live', Symbol='OA7', Aliases='', description='Osteoarthritis QTL 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474292, CurrentID=0, Status='live', Symbol='OA8', Aliases='', description='Osteoarthritis QTL 8', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474293, CurrentID=0, Status='live', Symbol='OA9', Aliases='', description='Osteoarthritis QTL 9', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474294, CurrentID=0, Status='live', Symbol='BW19', Aliases='', description='Body weight QTL 19', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474295, CurrentID=0, Status='live', Symbol='BW20', Aliases='', description='Body weight QTL 20', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474296, CurrentID=0, Status='live', Symbol='BFD1', Aliases='', description='Body fluid distribution QTL 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474297, CurrentID=0, Status='live', Symbol='BFD2', Aliases='', description='Body fluid distribution QTL 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474298, CurrentID=0, Status='live', Symbol='BFD3', Aliases='', description='Body fluid distribution QTL 3', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474299, CurrentID=0, Status='live', Symbol='BFD4', Aliases='', description='Body fluid distribution QTL 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474300, CurrentID=0, Status='live', Symbol='BFD5', Aliases='', description='Body fluid distribution QTL 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474301, CurrentID=0, Status='live', Symbol='BP5', Aliases='', description='Blood pressure QTL 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474302, CurrentID=0, Status='live', Symbol='BP6', Aliases='', description='Blood pressure QTL 6', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57205, CurrentID=0, Status='live', Symbol='ATP10D', Aliases='ATPVD', description='ATPase, class V, type 10D', other_designations='P4-ATPase flippase complex alpha subunit ATP10D|type IV aminophospholipid transporting ATPase', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47485271, end_position_on_the_genomic_accession=47593486, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  474304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474304, CurrentID=0, Status='live', Symbol='BW33', Aliases='', description='Body weight QTL 33', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474305, CurrentID=0, Status='live', Symbol='BW32', Aliases='', description='Body weight QTL 32', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474306, CurrentID=0, Status='live', Symbol='BW34', Aliases='', description='Body weight QTL 34', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474307, CurrentID=0, Status='live', Symbol='BW18', Aliases='', description='Body weight QTL 18', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474308, CurrentID=0, Status='live', Symbol='UAE1', Aliases='', description='Urinary albumin excretion QTL 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474309, CurrentID=0, Status='live', Symbol='UAE2', Aliases='', description='Urinary albumin excretion QTL 2', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474310, CurrentID=0, Status='live', Symbol='OA26', Aliases='', description='Osteoarthritis QTL 26', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474311, CurrentID=0, Status='live', Symbol='BW15', Aliases='', description='Body weight QTL 15', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474312, CurrentID=0, Status='live', Symbol='BW12', Aliases='', description='Body weight QTL 12', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474313, CurrentID=0, Status='live', Symbol='BW13', Aliases='', description='Body weight QTL 13', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474314, CurrentID=0, Status='live', Symbol='BW14', Aliases='', description='Body weight QTL 14', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81099, CurrentID=0, Status='live', Symbol='OR4F17', Aliases='OR4F11P, OR4F18, OR4F19', description='olfactory receptor family 4 subfamily F member 17', other_designations='olfactory receptor 4F11|olfactory receptor 4F18|olfactory receptor 4F19|olfactory receptor, family 4, subfamily F, member 11 pseudogene|olfactory receptor, family 4, subfamily F, member 17|olfactory receptor, family 4, subfamily F, member 19', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=107152, end_position_on_the_genomic_accession=111690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873421, CurrentID=0, Status='live', Symbol='RNA5SP155', Aliases='RN5S155', description='RNA, 5S ribosomal pseudogene 155', other_designations='RNA, 5S ribosomal 155', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10115756, end_position_on_the_genomic_accession=10115884, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81104, CurrentID=0, Status='live', Symbol='OR4Q1P', Aliases='', description='olfactory receptor family 4 subfamily Q member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily Q, member 1 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22044417, end_position_on_the_genomic_accession=22045397, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130680, CurrentID=0, Status='live', Symbol='LOC100130680', Aliases='', description='ankyrin repeat domain 49 pseudogene', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117439985, end_position_on_the_genomic_accession=117440773, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873422, CurrentID=0, Status='live', Symbol='RNA5SP156', Aliases='RN5S156', description='RNA, 5S ribosomal pseudogene 156', other_designations='RNA, 5S ribosomal 156', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11414051, end_position_on_the_genomic_accession=11414160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81110, CurrentID=0, Status='live', Symbol='OR11J2P', Aliases='', description='olfactory receptor family 11 subfamily J member 2 pseudogene', other_designations='olfactory receptor, family 11, subfamily J, member 2 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20960662, end_position_on_the_genomic_accession=20961607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81111, CurrentID=0, Status='live', Symbol='OR11J1P', Aliases='', description='olfactory receptor family 11 subfamily J member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily J, member 1 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21542998, end_position_on_the_genomic_accession=21543957, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373341, CurrentID=0, Status='live', Symbol='LOC105373341', Aliases='', description='uncharacterized LOC105373341', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135034406, end_position_on_the_genomic_accession=135040358, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81113, CurrentID=0, Status='live', Symbol='OR11H3P', Aliases='', description='olfactory receptor family 11 subfamily H member 3 pseudogene', other_designations='olfactory receptor, family 11, subfamily H, member 3 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22009549, end_position_on_the_genomic_accession=22010495, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81114, CurrentID=0, Status='live', Symbol='OR7K1P', Aliases='', description='olfactory receptor family 7 subfamily K member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily K, member 1 pseudogene', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25357765, end_position_on_the_genomic_accession=25358768, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81115, CurrentID=0, Status='live', Symbol='OR7E106P', Aliases='OR14-NP2, OR7E34P, OR7E40P, OST011, OST215', description='olfactory receptor family 7 subfamily E member 106 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 106 pseudogene|olfactory receptor, family 7, subfamily E, member 34 pseudogene|olfactory receptor, family 7, subfamily E, member 40 pseudogene', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51763048, end_position_on_the_genomic_accession=51764074, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81116, CurrentID=0, Status='live', Symbol='OR7E105P', Aliases='OR14-NP1', description='olfactory receptor family 7 subfamily E member 105 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 105 pseudogene', map_location='14q21', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51756991, end_position_on_the_genomic_accession=51757986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474333, CurrentID=0, Status='live', Symbol='GEVQ2', Aliases='', description='gene expression, variation in, quantitative trait locus', other_designations='', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608878, no_hdr0=''),
  474334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474334, CurrentID=0, Status='live', Symbol='GEVQ1', Aliases='', description='gene expression, variation in, quantitative trait locus', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608875, no_hdr0=''),
  343263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343263, CurrentID=0, Status='live', Symbol='MYBPHL', Aliases='', description='myosin binding protein H-like', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109292365, end_position_on_the_genomic_accession=109307041, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  474337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474337, CurrentID=0, Status='live', Symbol='SUMO1P2', Aliases='bA375F2.1', description='SUMO1 pseudogene 2', other_designations='SMT3 suppressor of mif two 3 homolog 1 pseudogene', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168898097, end_position_on_the_genomic_accession=168899044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474338, CurrentID=0, Status='live', Symbol='SUMO1P3', Aliases='', description='SUMO1 pseudogene 3 (functional)', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160317265, end_position_on_the_genomic_accession=160318474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130683, CurrentID=0, Status='live', Symbol='LOC100130683', Aliases='', description='chromatin assembly factor 1 subunit B pseudogene', other_designations='chromatin assembly factor 1, subunit B (p60) pseudogene', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105666223, end_position_on_the_genomic_accession=105668395, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474341, CurrentID=0, Status='live', Symbol='SUMO1P5', Aliases='', description='SUMO1 pseudogene 5', other_designations='', map_location='5q23', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123235957, end_position_on_the_genomic_accession=123243587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81126, CurrentID=0, Status='live', Symbol='OR4K16P', Aliases='OR14-24', description='olfactory receptor family 4 subfamily K member 16 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 16 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19956547, end_position_on_the_genomic_accession=19957554, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474343, CurrentID=0, Status='live', Symbol='SPIN2B', Aliases='SPIN-2, SPIN-2B, SPIN2_duplicate, dJ323P24.2', description='spindlin family member 2B', other_designations='spindlin family, member 2 duplicate|spindlin family, member 2B|spindlin-like protein 2B|telomeric SPIN2 copy', map_location='Xp11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57119682, end_position_on_the_genomic_accession=57121556, orientation='minus', exon_count=3, OMIM=300517, no_hdr0=''),
  474344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474344, CurrentID=0, Status='live', Symbol='GIMAP6', Aliases='IAN-2, IAN-6, IAN2, IAN6', description='GTPase, IMAP family member 6', other_designations='immune associated nucleotide 2|immune associated nucleotide 6|immune-associated nucleotide-binding protein 6', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150625375, end_position_on_the_genomic_accession=150632648, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  474345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474345, CurrentID=0, Status='live', Symbol='GIMAP3P', Aliases='GIMAP3, IAN4P', description='GTPase, IMAP family member 3 pseudogene', other_designations='GTPase, IMAP family member 5 pseudogene|immune-associated nucleotide-binding protein 4 pseudogene|immunity associated protein 3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150746925, end_position_on_the_genomic_accession=150747731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873426, CurrentID=0, Status='live', Symbol='RNA5SP160', Aliases='RN5S160', description='RNA, 5S ribosomal pseudogene 160', other_designations='RNA, 5S ribosomal 160', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40990154, end_position_on_the_genomic_accession=40990273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81136, CurrentID=0, Status='live', Symbol='OR7E111P', Aliases='', description='olfactory receptor family 7 subfamily E member 111 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 111 pseudogene', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67902244, end_position_on_the_genomic_accession=67903224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81137, CurrentID=0, Status='live', Symbol='OR7E104P', Aliases='', description='olfactory receptor family 7 subfamily E member 104 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 104 pseudogene', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63836928, end_position_on_the_genomic_accession=63837673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  474354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474354, CurrentID=0, Status='live', Symbol='LRRC18', Aliases='UNQ933, UNQ9338, VKGE9338', description='leucine rich repeat containing 18', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48883696, end_position_on_the_genomic_accession=48940126, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  81140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81140, CurrentID=0, Status='live', Symbol='OR9R1P', Aliases='', description='olfactory receptor family 9 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 9, subfamily R, member 1 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55158848, end_position_on_the_genomic_accession=55159801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130687, CurrentID=0, Status='live', Symbol='FABP5P6', Aliases='FABP5L6', description='fatty acid binding protein 5 pseudogene 6', other_designations='fatty acid binding protein 5-like 6', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119555248, end_position_on_the_genomic_accession=119555658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343296, CurrentID=0, Status='live', Symbol='ADH5P2', Aliases='', description='alcohol dehydrogenase 5 (class III), chi polypeptide, pseudogene 2', other_designations='alcohol dehydrogenase 5 (class III), chi polypeptide pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=79521069, end_position_on_the_genomic_accession=79522467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81153, CurrentID=0, Status='live', Symbol='OR9M1P', Aliases='OR5BG1P', description='olfactory receptor family 9 subfamily M member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BG, member 1 pseudogene|olfactory receptor, family 9, subfamily M, member 1, pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55855592, end_position_on_the_genomic_accession=55856521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81155, CurrentID=0, Status='live', Symbol='OR9I3P', Aliases='OST714', description='olfactory receptor family 9 subfamily I member 3 pseudogene', other_designations='olfactory receptor, family 9, subfamily I, member 3 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58108720, end_position_on_the_genomic_accession=58109661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81156, CurrentID=0, Status='live', Symbol='OR9I2P', Aliases='', description='olfactory receptor family 9 subfamily I member 2 pseudogene', other_designations='olfactory receptor, family 9, subfamily I, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58144849, end_position_on_the_genomic_accession=58145334, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  343301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343301, CurrentID=0, Status='live', Symbol='RPL10AP4', Aliases='RPL10A_1_67', description='ribosomal protein L10a pseudogene 4', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81208535, end_position_on_the_genomic_accession=81209249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81159, CurrentID=0, Status='live', Symbol='OR9G3P', Aliases='', description='olfactory receptor family 9 subfamily G member 3 pseudogene', other_designations='olfactory receptor, family 9, subfamily G, member 3 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56740180, end_position_on_the_genomic_accession=56741093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81160, CurrentID=0, Status='live', Symbol='OR9G2P', Aliases='OR9G6', description='olfactory receptor family 9 subfamily G member 2 pseudogene', other_designations='olfactory receptor, family 9, subfamily G, member 2 pseudogene|olfactory receptor, family 9, subfamily G, member 6', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56750997, end_position_on_the_genomic_accession=56751984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81162, CurrentID=0, Status='live', Symbol='OR8R1P', Aliases='', description='olfactory receptor family 8 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily R, member 1 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73249316, end_position_on_the_genomic_accession=73249869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81163, CurrentID=0, Status='live', Symbol='OR8Q1P', Aliases='', description='olfactory receptor family 8 subfamily Q member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily Q, member 1 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124601603, end_position_on_the_genomic_accession=124602259, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81164, CurrentID=0, Status='live', Symbol='OR8L1P', Aliases='OR11-115', description='olfactory receptor family 8 subfamily L member 1 pseudogene', other_designations='olfactory receptor pseudogene|olfactory receptor, family 8, subfamily L, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56381636, end_position_on_the_genomic_accession=56382590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  474381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474381, CurrentID=0, Status='live', Symbol='H2AFB2', Aliases='H2A.Bbd', description='H2A histone family member B2', other_designations='H2A Barr body deficient|H2A Barr body-deficient|H2A histone family, member B2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155380706, end_position_on_the_genomic_accession=155381299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474382, CurrentID=0, Status='live', Symbol='H2AFB1', Aliases='H2A.Bbd', description='H2A histone family member B1', other_designations='H2A Barr body-deficient|H2A histone family, member B1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154885042, end_position_on_the_genomic_accession=154885558, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474383, CurrentID=0, Status='live', Symbol='F8A2', Aliases='', description='coagulation factor VIII-associated 2', other_designations='coagulation factor VIII-associated (intronic transcript) 2|cpG island protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155382115, end_position_on_the_genomic_accession=155383230, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  474384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474384, CurrentID=0, Status='live', Symbol='F8A3', Aliases='', description='coagulation factor VIII-associated 3', other_designations='coagulation factor VIII-associated (intronic transcript) 3|cpG island protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155457485, end_position_on_the_genomic_accession=155458600, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81169, CurrentID=0, Status='live', Symbol='OR8J2', Aliases='OR8J2P', description='olfactory receptor family 8 subfamily J member 2 (gene/pseudogene)', other_designations='Olfactory receptor 8J2|olfactory receptor, family 8, subfamily J, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56210857, end_position_on_the_genomic_accession=56211821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  474387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474387, CurrentID=0, Status='live', Symbol='OFC4', Aliases='', description='Orofacial cleft 4', other_designations='', map_location='4q21-q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608371, no_hdr0=''),
  474388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474388, CurrentID=0, Status='live', Symbol='CMM4', Aliases='', description='Melanoma, cutaneous malignant, 4', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608035, no_hdr0=''),
  100873433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873433, CurrentID=0, Status='live', Symbol='RNA5SP168', Aliases='RN5S168', description='RNA, 5S ribosomal pseudogene 168', other_designations='RNA, 5S ribosomal 168', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=150970671, end_position_on_the_genomic_accession=150970782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81176, CurrentID=0, Status='live', Symbol='OR7E126P', Aliases='OR11-1, hg500', description='olfactory receptor family 7 subfamily E member 126 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 126 pseudogene|olfactory receptor, family 7, subfamily E, member 126 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71902904, end_position_on_the_genomic_accession=71904322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373572, CurrentID=0, Status='live', Symbol='LOC105373572', Aliases='', description='uncharacterized LOC105373572', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115738897, end_position_on_the_genomic_accession=115756822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873434, CurrentID=0, Status='live', Symbol='RNA5SP169', Aliases='RN5S169', description='RNA, 5S ribosomal pseudogene 169', other_designations='RNA, 5S ribosomal 169', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151971497, end_position_on_the_genomic_accession=151971613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343326, CurrentID=0, Status='live', Symbol='KRT18P28', Aliases='', description='keratin 18 pseudogene 28', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182959019, end_position_on_the_genomic_accession=182960412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130693, CurrentID=0, Status='live', Symbol='ARHGAP42P2', Aliases='', description='Rho GTPase activating protein 42 pseudogene 2', other_designations='rho GTPase-activating protein 42-like', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130006200, end_position_on_the_genomic_accession=130009198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81185, CurrentID=0, Status='live', Symbol='OR5P4P', Aliases='OST730', description='olfactory receptor family 5 subfamily P member 4 pseudogene', other_designations='olfactory receptor, family 5, subfamily P, member 4 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7745991, end_position_on_the_genomic_accession=7746910, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81187, CurrentID=0, Status='live', Symbol='OR5M13P', Aliases='', description='olfactory receptor family 5 subfamily M member 13 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 13 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56597507, end_position_on_the_genomic_accession=56597994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343332, CurrentID=0, Status='live', Symbol='FDPSP1', Aliases='CHR39A, FDPSL1, FPSL1', description='farnesyl diphosphate synthase pseudogene 1', other_designations='farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene|farnesyl diphosphate synthase-like 1 (farnesyl pyrophosphate synthetase-like 1, cholesterol-repressible protein 39A)', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187563027, end_position_on_the_genomic_accession=187564175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373574, CurrentID=0, Status='live', Symbol='LOC105373574', Aliases='', description='uncharacterized LOC105373574', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115112255, end_position_on_the_genomic_accession=115115265, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81191, CurrentID=0, Status='live', Symbol='OR5G5P', Aliases='', description='olfactory receptor family 5 subfamily G member 5 pseudogene', other_designations='olfactory receptor, family 5, subfamily G, member 5 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56801859, end_position_on_the_genomic_accession=56802794, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81192, CurrentID=0, Status='live', Symbol='OR5G4P', Aliases='', description='olfactory receptor family 5 subfamily G member 4 pseudogene', other_designations='olfactory receptor, family 5, subfamily G, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56790121, end_position_on_the_genomic_accession=56791101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81193, CurrentID=0, Status='live', Symbol='OR5G3', Aliases='OR5G3P, OR5G6P', description='olfactory receptor family 5 subfamily G member 3 (gene/pseudogene)', other_designations='Olfactory receptor 5G3|olfactory receptor, family 5, subfamily G, member 3 pseudogene|olfactory receptor, family 5, subfamily G, member 6 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56819479, end_position_on_the_genomic_accession=56820646, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343338, CurrentID=0, Status='live', Symbol='GAPDHP20', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 20', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32401899, end_position_on_the_genomic_accession=32403183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373575, CurrentID=0, Status='live', Symbol='LOC105373575', Aliases='', description='uncharacterized LOC105373575', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115075161, end_position_on_the_genomic_accession=115099027, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  81196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81196, CurrentID=0, Status='live', Symbol='OR5D17P', Aliases='', description='olfactory receptor family 5 subfamily D member 17 pseudogene', other_designations='olfactory receptor, family 5, subfamily D, member 17 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55754977, end_position_on_the_genomic_accession=55755954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81198, CurrentID=0, Status='live', Symbol='OR5D15P', Aliases='', description='olfactory receptor family 5 subfamily D member 15 pseudogene', other_designations='olfactory receptor, family 5, subfamily D, member 15 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55786965, end_position_on_the_genomic_accession=55787906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81201, CurrentID=0, Status='live', Symbol='OR5BR1P', Aliases='', description='olfactory receptor family 5 subfamily BR member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BR, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59332756, end_position_on_the_genomic_accession=59334085, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81202, CurrentID=0, Status='live', Symbol='OR5BQ1P', Aliases='OR5BQ2P', description='olfactory receptor family 5 subfamily BQ member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BQ, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57029388, end_position_on_the_genomic_accession=57029723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81203, CurrentID=0, Status='live', Symbol='OR5BP1P', Aliases='', description='olfactory receptor family 5 subfamily BP member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BP, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57055701, end_position_on_the_genomic_accession=57056134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81204, CurrentID=0, Status='live', Symbol='OR5BN1P', Aliases='OR11-161', description='olfactory receptor family 5 subfamily BN member 1 pseudogene', other_designations='olfactory receptor OR11-161 pseudogene|olfactory receptor, family 5, subfamily BN, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56132665, end_position_on_the_genomic_accession=56133493, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873438, CurrentID=0, Status='live', Symbol='RNA5SP176', Aliases='RN5S176', description='RNA, 5S ribosomal pseudogene 176', other_designations='RNA, 5S ribosomal 176', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7047227, end_position_on_the_genomic_accession=7047336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81207, CurrentID=0, Status='live', Symbol='OR5BE1P', Aliases='', description='olfactory receptor family 5 subfamily BE member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BE, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56082801, end_position_on_the_genomic_accession=56083736, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81208, CurrentID=0, Status='live', Symbol='OR5BC1P', Aliases='', description='olfactory receptor family 5 subfamily BC member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BC, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58317640, end_position_on_the_genomic_accession=58318473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81209, CurrentID=0, Status='live', Symbol='OR5BB1P', Aliases='', description='olfactory receptor family 5 subfamily BB member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BB, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59391353, end_position_on_the_genomic_accession=59392291, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81210, CurrentID=0, Status='live', Symbol='OR5BA1P', Aliases='', description='olfactory receptor family 5 subfamily BA member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BA, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57866302, end_position_on_the_genomic_accession=57867280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873439, CurrentID=0, Status='live', Symbol='RNA5SP177', Aliases='RN5S177', description='RNA, 5S ribosomal pseudogene 177', other_designations='RNA, 5S ribosomal 177', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9760907, end_position_on_the_genomic_accession=9761013, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373578, CurrentID=0, Status='live', Symbol='LOC105373578', Aliases='', description='uncharacterized LOC105373578', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118187838, end_position_on_the_genomic_accession=118224425, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130714, CurrentID=0, Status='live', Symbol='LOC100130714', Aliases='', description='metaxin 3 pseudogene', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44187939, end_position_on_the_genomic_accession=44189099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81215, CurrentID=0, Status='live', Symbol='OR5B15P', Aliases='', description='olfactory receptor family 5 subfamily B member 15 pseudogene', other_designations='olfactory receptor, family 5, subfamily B, member 15 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58387571, end_position_on_the_genomic_accession=58388486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873440, CurrentID=0, Status='live', Symbol='RNA5SP178', Aliases='RN5S178', description='RNA, 5S ribosomal pseudogene 178', other_designations='RNA, 5S ribosomal 178', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16033229, end_position_on_the_genomic_accession=16033356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647051, CurrentID=0, Status='live', Symbol='YWHABP1', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta pseudogene 1', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta polypeptide pseudogene', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38643039, end_position_on_the_genomic_accession=38643770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81221, CurrentID=0, Status='live', Symbol='OR5AQ1P', Aliases='OR11-165', description='olfactory receptor family 5 subfamily AQ member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AQ, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56054365, end_position_on_the_genomic_accession=56055300, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873441, CurrentID=0, Status='live', Symbol='RNA5SP179', Aliases='RN5S179', description='RNA, 5S ribosomal pseudogene 179', other_designations='RNA, 5S ribosomal 179', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16853994, end_position_on_the_genomic_accession=16854114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81224, CurrentID=0, Status='live', Symbol='OR5AN2P', Aliases='', description='olfactory receptor family 5 subfamily AN member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily AN, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59309642, end_position_on_the_genomic_accession=59310588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373580, CurrentID=0, Status='live', Symbol='LOC105373580', Aliases='', description='uncharacterized LOC105373580', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118949305, end_position_on_the_genomic_accession=118953065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81226, CurrentID=0, Status='live', Symbol='OR5AM1P', Aliases='', description='olfactory receptor family 5 subfamily AM member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AM, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56619746, end_position_on_the_genomic_accession=56620810, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81228, CurrentID=0, Status='live', Symbol='OR5AK3P', Aliases='OR5AK3', description='olfactory receptor family 5 subfamily AK member 3 pseudogene', other_designations='Olfactory receptor 5AK3|olfactory receptor, family 5, subfamily AK, member 3 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56971051, end_position_on_the_genomic_accession=56971977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873442, CurrentID=0, Status='live', Symbol='RNA5SP180', Aliases='RN5S180', description='RNA, 5S ribosomal pseudogene 180', other_designations='RNA, 5S ribosomal 180', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17156971, end_position_on_the_genomic_accession=17157071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81230, CurrentID=0, Status='live', Symbol='OR5AK1P', Aliases='OR5AK5P', description='olfactory receptor family 5 subfamily AK member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AK, member 1 pseudogene|olfactory receptor, family 5, subfamily AK, member 5 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57018092, end_position_on_the_genomic_accession=57019019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130701, CurrentID=0, Status='live', Symbol='LOC100130701', Aliases='', description='uncharacterized LOC100130701', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120366539, end_position_on_the_genomic_accession=120384909, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105463123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463123, CurrentID=0, Status='live', Symbol='XACT', Aliases='', description='X active specific transcript (non-protein coding)', other_designations='X active coating transcript', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300901, no_hdr0=''),
  81236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81236, CurrentID=0, Status='live', Symbol='OR56A7P', Aliases='', description='olfactory receptor family 56 subfamily A member 7 pseudogene', other_designations='olfactory receptor, family 56, subfamily A, member 7 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6045809, end_position_on_the_genomic_accession=6046746, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105463125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463125, CurrentID=0, Status='live', Symbol='ATOD8', Aliases='', description='Dermatitis, atopic, 8', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613518, no_hdr0=''),
  105463126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463126, CurrentID=0, Status='live', Symbol='ATOD9', Aliases='', description='Dermatitis, atopic, 9', other_designations='', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613519, no_hdr0=''),
  105463127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463127, CurrentID=0, Status='live', Symbol='PFHB2', Aliases='', description='Progressive familial heart block, type II', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=140400, no_hdr0=''),
  105463128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463128, CurrentID=0, Status='live', Symbol='MRXSBWB', Aliases='', description='Brooks-Wisniewski-Brown syndrome', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300612, no_hdr0=''),
  105463129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105463129, CurrentID=0, Status='live', Symbol='MRX88', Aliases='', description='Mental retardation, X-linked 88', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300852, no_hdr0=''),
  81242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81242, CurrentID=0, Status='live', Symbol='OR52V1P', Aliases='', description='olfactory receptor family 52 subfamily V member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily V, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5527223, end_position_on_the_genomic_accession=5528126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81243, CurrentID=0, Status='live', Symbol='OR52U1P', Aliases='', description='olfactory receptor family 52 subfamily U member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily U, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5719290, end_position_on_the_genomic_accession=5720266, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81244, CurrentID=0, Status='live', Symbol='OR52T1P', Aliases='', description='olfactory receptor family 52 subfamily T member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily T, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5566650, end_position_on_the_genomic_accession=5567615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81246, CurrentID=0, Status='live', Symbol='OR52Q1P', Aliases='', description='olfactory receptor family 52 subfamily Q member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily Q, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5903636, end_position_on_the_genomic_accession=5904479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873445, CurrentID=0, Status='live', Symbol='RNA5SP183', Aliases='RN5S183', description='RNA, 5S ribosomal pseudogene 183', other_designations='RNA, 5S ribosomal 183', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55777469, end_position_on_the_genomic_accession=55777587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81248, CurrentID=0, Status='live', Symbol='OR52P1P', Aliases='OR11-66, OR52P1', description='olfactory receptor family 52 subfamily P member 1 pseudogene', other_designations='Olfactory receptor 52P1|olfactory receptor OR11-66 pseudogene|olfactory receptor, family 52, subfamily P, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5726501, end_position_on_the_genomic_accession=5727460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130704, CurrentID=0, Status='live', Symbol='LOC100130704', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 1 pseudogene', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131892612, end_position_on_the_genomic_accession=131893503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81251, CurrentID=0, Status='live', Symbol='OR52N3P', Aliases='OR11-60', description='olfactory receptor family 52 subfamily N member 3 pseudogene', other_designations='olfactory receptor OR11-60 pseudogene|olfactory receptor, family 52, subfamily N, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5800364, end_position_on_the_genomic_accession=5801321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873446, CurrentID=0, Status='live', Symbol='RNA5SP184', Aliases='RN5S184', description='RNA, 5S ribosomal pseudogene 184', other_designations='RNA, 5S ribosomal 184', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56146739, end_position_on_the_genomic_accession=56146846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130705, CurrentID=0, Status='live', Symbol='LOC100130705', Aliases='', description='uncharacterized LOC100130705', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128866410, end_position_on_the_genomic_accession=128872047, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100652908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652908, CurrentID=0, Status='live', Symbol='LOC100652908', Aliases='', description='GRB2-related adaptor protein pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209173009, end_position_on_the_genomic_accession=209173372, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375319, CurrentID=0, Status='live', Symbol='LOC105375319', Aliases='', description='uncharacterized LOC105375319', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64225751, end_position_on_the_genomic_accession=64227621, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  81261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81261, CurrentID=0, Status='live', Symbol='OR52H2P', Aliases='', description='olfactory receptor family 52 subfamily H member 2 pseudogene', other_designations='olfactory receptor, family 52, subfamily H, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5551665, end_position_on_the_genomic_accession=5552596, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343406, CurrentID=0, Status='live', Symbol='OR10R2', Aliases='OR1-8Q, OR10R2', description='olfactory receptor family 10 subfamily R member 2', other_designations='olfactory receptor OR1-8|olfactory receptor, family 10, subfamily R, member 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158479878, end_position_on_the_genomic_accession=158480885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81264, CurrentID=0, Status='live', Symbol='OR52E7P', Aliases='', description='olfactory receptor family 52 subfamily E member 7 pseudogene', other_designations='olfactory receptor, family 52, subfamily E, member 7 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5873960, end_position_on_the_genomic_accession=5874888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873448, CurrentID=0, Status='live', Symbol='RNA5SP186', Aliases='RN5S186', description='RNA, 5S ribosomal pseudogene 186', other_designations='RNA, 5S ribosomal 186', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76091708, end_position_on_the_genomic_accession=76091801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343413, CurrentID=0, Status='live', Symbol='FCRL6', Aliases='FcRH6', description='Fc receptor-like 6', other_designations='Fc receptor-like protein 7|IFGP6|IgSF type I transmembrane receptor|fc receptor homolog 6|fcR-like protein 6|leukocyte receptor', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159800511, end_position_on_the_genomic_accession=159816257, orientation='plus', exon_count=11, OMIM=613562, no_hdr0=''),
  81270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81270, CurrentID=0, Status='live', Symbol='OR52B5P', Aliases='', description='olfactory receptor family 52 subfamily B member 5 pseudogene', other_designations='olfactory receptor, family 52, subfamily B, member 5 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5560967, end_position_on_the_genomic_accession=5561911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873449, CurrentID=0, Status='live', Symbol='RNA5SP187', Aliases='RN5S187', description='RNA, 5S ribosomal pseudogene 187', other_designations='RNA, 5S ribosomal 187', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88274258, end_position_on_the_genomic_accession=88274376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130708, CurrentID=0, Status='live', Symbol='ANP32BP2', Aliases='', description='acidic nuclear phosphoprotein 32 family member B pseudogene 2', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member B pseudogene 2', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44651835, end_position_on_the_genomic_accession=44652542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81274, CurrentID=0, Status='live', Symbol='OR52B1P', Aliases='PJCG4', description='olfactory receptor family 52 subfamily B member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily B, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6151595, end_position_on_the_genomic_accession=6152549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81278, CurrentID=0, Status='live', Symbol='OR51K1P', Aliases='', description='olfactory receptor family 51 subfamily K member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily K, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5430656, end_position_on_the_genomic_accession=5431602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373589, CurrentID=0, Status='live', Symbol='LOC105373589', Aliases='', description='uncharacterized LOC105373589', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121562038, end_position_on_the_genomic_accession=121574097, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81282, CurrentID=0, Status='live', Symbol='OR51G2', Aliases='OR11-28', description='olfactory receptor family 51 subfamily G member 2', other_designations='olfactory receptor OR11-28|olfactory receptor, family 51, subfamily G, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4914719, end_position_on_the_genomic_accession=4915663, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873451, CurrentID=0, Status='live', Symbol='RNA5SP189', Aliases='RN5S189', description='RNA, 5S ribosomal pseudogene 189', other_designations='RNA, 5S ribosomal 189', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=105922994, end_position_on_the_genomic_accession=105923108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81285, CurrentID=0, Status='live', Symbol='OR51E2', Aliases='OR51E3P, OR52A2, PSGR', description='olfactory receptor family 51 subfamily E member 2', other_designations='HPRAJ|olfactory receptor OR11-16|olfactory receptor, family 51, subfamily E, member 2|olfactory receptor, family 51, subfamily E, member 3 pseudogene|olfactory receptor, family 52, subfamily A, member 2|prostate specific G-protein coupled receptor|prostate-specific G-protein coupled receptor', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4680171, end_position_on_the_genomic_accession=4697846, orientation='minus', exon_count=2, OMIM=611268, no_hdr0=''),
  100873452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873452, CurrentID=0, Status='live', Symbol='RNA5SP190', Aliases='RN5S190', description='RNA, 5S ribosomal pseudogene 190', other_designations='RNA, 5S ribosomal 190', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119445716, end_position_on_the_genomic_accession=119445833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873453, CurrentID=0, Status='live', Symbol='RNA5SP191', Aliases='RN5S191', description='RNA, 5S ribosomal pseudogene 191', other_designations='RNA, 5S ribosomal 191', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130114409, end_position_on_the_genomic_accession=130114510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81298, CurrentID=0, Status='live', Symbol='OR4R3P', Aliases='', description='olfactory receptor family 4 subfamily R member 3 pseudogene', other_designations='olfactory receptor, family 4, subfamily R, member 3 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49923050, end_position_on_the_genomic_accession=49923887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81299, CurrentID=0, Status='live', Symbol='OR4R2P', Aliases='', description='olfactory receptor family 4 subfamily R member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily R, member 2 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54657097, end_position_on_the_genomic_accession=54657953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81300, CurrentID=0, Status='live', Symbol='OR4P4', Aliases='OR4P3P', description='olfactory receptor family 4 subfamily P member 4', other_designations='olfactory receptor 4P3|olfactory receptor, family 4, subfamily P, member 3 pseudogene|olfactory receptor, family 4, subfamily P, member 4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55638358, end_position_on_the_genomic_accession=55639296, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  341230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341230, CurrentID=0, Status='live', Symbol='ASS1P13', Aliases='ASSP13', description='argininosuccinate synthetase 1 pseudogene 13', other_designations='argininosuccinate synthetase pseudogene 13', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107176212, end_position_on_the_genomic_accession=107177767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379685, CurrentID=0, Status='live', Symbol='LOC105379685', Aliases='', description='uncharacterized LOC105379685', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81304, CurrentID=0, Status='live', Symbol='OR4D7P', Aliases='OST724', description='olfactory receptor family 4 subfamily D member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily D, member 7 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59531698, end_position_on_the_genomic_accession=59532711, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  343450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343450, CurrentID=0, Status='live', Symbol='KCNT2', Aliases='KCa4.2, SLICK, SLO2.1', description='potassium channel, sodium activated subfamily T, member 2', other_designations='potassium channel, subfamily T, member 2|sequence like an intermediate conductance potassium channel subunit|sodium and chloride-activated ATP-sensitive potassium channel Slo2.1|sodium-and chloride-activated ATP-sensitive potassium channel (SLICK)', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196225779, end_position_on_the_genomic_accession=196608560, orientation='minus', exon_count=31, OMIM=610044, no_hdr0=''),
  100873455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873455, CurrentID=0, Status='live', Symbol='RNA5SP193', Aliases='RN5S193', description='RNA, 5S ribosomal pseudogene 193', other_designations='RNA, 5S ribosomal 193', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136969200, end_position_on_the_genomic_accession=136969306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81309, CurrentID=0, Status='live', Symbol='OR4C15', Aliases='OR11-127, OR11-134', description='olfactory receptor family 4 subfamily C member 15', other_designations='olfactory receptor OR11-127|olfactory receptor OR11-134|olfactory receptor, family 4, subfamily C, member 15', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55554307, end_position_on_the_genomic_accession=55555419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81310, CurrentID=0, Status='live', Symbol='OR4C14P', Aliases='', description='olfactory receptor family 4 subfamily C member 14 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 14 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55537003, end_position_on_the_genomic_accession=55537925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873456, CurrentID=0, Status='live', Symbol='RNA5SP194', Aliases='RN5S194', description='RNA, 5S ribosomal pseudogene 194', other_designations='RNA, 5S ribosomal 194', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139012329, end_position_on_the_genomic_accession=139012441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81315, CurrentID=0, Status='live', Symbol='OR4A8', Aliases='OR4A8P', description='olfactory receptor family 4 subfamily A member 8 (gene/pseudogene)', other_designations='Olfactory receptor 4A8|olfactory receptor, family 4, subfamily A, member 8 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54682876, end_position_on_the_genomic_accession=54683820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81316, CurrentID=0, Status='live', Symbol='OR4A7P', Aliases='', description='olfactory receptor family 4 subfamily A member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 7 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54692451, end_position_on_the_genomic_accession=54693360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81318, CurrentID=0, Status='live', Symbol='OR4A5', Aliases='OR11-111', description='olfactory receptor family 4 subfamily A member 5', other_designations='olfactory receptor OR11-111|olfactory receptor, family 4, subfamily A, member 5|seven transmembrane helix receptor', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54706832, end_position_on_the_genomic_accession=54707902, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873457, CurrentID=0, Status='live', Symbol='RNA5SP195', Aliases='RN5S195', description='RNA, 5S ribosomal pseudogene 195', other_designations='RNA, 5S ribosomal 195', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139299852, end_position_on_the_genomic_accession=139299968, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  399879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=399879, CurrentID=0, Status='live', Symbol='LINC00610', Aliases='C11orf55', description='long intergenic non-protein coding RNA 610', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=36247410, end_position_on_the_genomic_accession=36251485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81322, CurrentID=0, Status='live', Symbol='OR4A21P', Aliases='', description='olfactory receptor family 4 subfamily A member 21 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 21 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55491157, end_position_on_the_genomic_accession=55492146, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81324, CurrentID=0, Status='live', Symbol='OR4A19P', Aliases='', description='olfactory receptor family 4 subfamily A member 19 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 19 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49920311, end_position_on_the_genomic_accession=49921413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873458, CurrentID=0, Status='live', Symbol='RNA5SP196', Aliases='RN5S196', description='RNA, 5S ribosomal pseudogene 196', other_designations='RNA, 5S ribosomal 196', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146586099, end_position_on_the_genomic_accession=146586200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81327, CurrentID=0, Status='live', Symbol='OR4A16', Aliases='OR11-117Q, OR4A16', description='olfactory receptor family 4 subfamily A member 16', other_designations='olfactory receptor OR11-117|olfactory receptor, family 4, subfamily A, member 16', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55323661, end_position_on_the_genomic_accession=55344225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  343472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343472, CurrentID=0, Status='live', Symbol='BARHL2', Aliases='', description='BarH-like homeobox 2', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90712022, end_position_on_the_genomic_accession=90717237, orientation='minus', exon_count=3, OMIM=605212, no_hdr0=''),
  81329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81329, CurrentID=0, Status='live', Symbol='OR4A14P', Aliases='', description='olfactory receptor family 4 subfamily A member 14 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 14 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55477552, end_position_on_the_genomic_accession=55478510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81330, CurrentID=0, Status='live', Symbol='OR4A13P', Aliases='', description='olfactory receptor family 4 subfamily A member 13 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 13 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55466772, end_position_on_the_genomic_accession=55467713, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873459, CurrentID=0, Status='live', Symbol='RNA5SP197', Aliases='RN5S197', description='RNA, 5S ribosomal pseudogene 197', other_designations='RNA, 5S ribosomal 197', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151477459, end_position_on_the_genomic_accession=151477577, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343477, CurrentID=0, Status='live', Symbol='HSP90B3P', Aliases='GRP94C, TRA1P2, TRAP2', description='heat shock protein 90kDa beta family member 3, pseudogene', other_designations='heat shock protein 90kDa beta (Grp94), member 3, pseudogene|heat shock protein 94c|tumor rejection antigen (gp96) 1 pseudogene 2', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91642518, end_position_on_the_genomic_accession=91644341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81336, CurrentID=0, Status='live', Symbol='OR2AH1P', Aliases='', description='olfactory receptor family 2 subfamily AH member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AH, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56669020, end_position_on_the_genomic_accession=56669944, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873460, CurrentID=0, Status='live', Symbol='RNA5SP198', Aliases='RN5S198', description='RNA, 5S ribosomal pseudogene 198', other_designations='RNA, 5S ribosomal 198', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151875566, end_position_on_the_genomic_accession=151875692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189199, CurrentID=0, Status='live', Symbol='TRL-CAA4-1', Aliases='TRNAL19', description='transfer RNA-Leu (CAA) 4-1', other_designations='transfer RNA leucine 19 (anticodon CAA)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248873855, end_position_on_the_genomic_accession=248873960, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373599, CurrentID=0, Status='live', Symbol='LOC105373599', Aliases='', description='uncharacterized LOC105373599', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=125822158, end_position_on_the_genomic_accession=125832794, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  81340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81340, CurrentID=0, Status='live', Symbol='OR10Y1P', Aliases='', description='olfactory receptor family 10 subfamily Y member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily Y, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59728520, end_position_on_the_genomic_accession=59729493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81341, CurrentID=0, Status='live', Symbol='OR10W1', Aliases='OR10W1PQ, OR11-236, UNQ6469, OR10W1', description='olfactory receptor family 10 subfamily W member 1', other_designations='olfactory receptor OR11-236|olfactory receptor, family 10, subfamily W, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58266792, end_position_on_the_genomic_accession=58268260, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81342, CurrentID=0, Status='live', Symbol='OR10V3P', Aliases='', description='olfactory receptor family 10 subfamily V member 3 pseudogene', other_designations='olfactory receptor, family 10, subfamily V, member 3 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59741326, end_position_on_the_genomic_accession=59742050, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81343, CurrentID=0, Status='live', Symbol='OR10V2P', Aliases='', description='olfactory receptor family 10 subfamily V member 2 pseudogene', other_designations='olfactory receptor, family 10, subfamily V, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59748812, end_position_on_the_genomic_accession=59749581, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81345, CurrentID=0, Status='live', Symbol='OR10Q2P', Aliases='', description='olfactory receptor family 10 subfamily Q member 2 pseudogene', other_designations='olfactory receptor, family 10, subfamily Q, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58291829, end_position_on_the_genomic_accession=58292771, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652911, CurrentID=0, Status='live', Symbol='LOC100652911', Aliases='', description='uncharacterized LOC100652911', other_designations='CTC-523E23.6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34905286, end_position_on_the_genomic_accession=34908191, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873462, CurrentID=0, Status='live', Symbol='RNA5SP201', Aliases='RN5S201', description='RNA, 5S ribosomal pseudogene 201', other_designations='RNA, 5S ribosomal 201', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3088686, end_position_on_the_genomic_accession=3088809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343495, CurrentID=0, Status='live', Symbol='RPL6P2', Aliases='RPL6_1_93', description='ribosomal protein L6 pseudogene 2', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119219293, end_position_on_the_genomic_accession=119220117, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81352, CurrentID=0, Status='live', Symbol='OR7M1P', Aliases='', description='olfactory receptor family 7 subfamily M member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily M, member 1 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133481222, end_position_on_the_genomic_accession=133481771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81353, CurrentID=0, Status='live', Symbol='OR7E115P', Aliases='OR10-2, OST704', description='olfactory receptor family 7 subfamily E member 115 pseudogene', other_designations='olfactory receptor OR10-2 pseudogene|olfactory receptor, family 7, subfamily E, member 115 pseudogene', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15007788, end_position_on_the_genomic_accession=15008726, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81354, CurrentID=0, Status='live', Symbol='OR7E110P', Aliases='OR7E68P, OR7E71P, OR7E72P, OR7E73P, OR7E74P, OR7E75P, OR912-108, OR912-109, OR912-110, OR912-46, hg523, hg674', description='olfactory receptor family 7 subfamily E member 110 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 110 pseudogene', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14986876, end_position_on_the_genomic_accession=14987471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873463, CurrentID=0, Status='live', Symbol='RNA5SP202', Aliases='RN5S202', description='RNA, 5S ribosomal pseudogene 202', other_designations='RNA, 5S ribosomal 202', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4427963, end_position_on_the_genomic_accession=4428081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81356, CurrentID=0, Status='live', Symbol='OR6L2P', Aliases='', description='olfactory receptor family 6 subfamily L member 2 pseudogene', other_designations='olfactory receptor pseudogene|olfactory receptor, family 6, subfamily L, member 2 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133430394, end_position_on_the_genomic_accession=133431318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373602, CurrentID=0, Status='live', Symbol='LOC105373602', Aliases='', description='uncharacterized LOC105373602', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126535789, end_position_on_the_genomic_accession=126640247, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  81358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81358, CurrentID=0, Status='live', Symbol='OR7E116P', Aliases='OR7E112P, OST733', description='olfactory receptor family 7 subfamily E member 116 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 112 pseudogene|olfactory receptor, family 7, subfamily E, member 116 pseudogene', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90232099, end_position_on_the_genomic_accession=90233124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  343505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343505, CurrentID=0, Status='live', Symbol='NBPF7', Aliases='', description='neuroblastoma breakpoint family member 7', other_designations='neuroblastoma breakpoint family, member 7', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119833730, end_position_on_the_genomic_accession=119844880, orientation='minus', exon_count=10, OMIM=613997, no_hdr0=''),
  81363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81363, CurrentID=0, Status='live', Symbol='OR7E108P', Aliases='OST726', description='olfactory receptor family 7 subfamily E member 108 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 108 pseudogene', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90751218, end_position_on_the_genomic_accession=90752240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343508, CurrentID=0, Status='live', Symbol='LOC343508', Aliases='', description='aconitase 2, mitochondrial pseudogene', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236111020, end_position_on_the_genomic_accession=236112369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81366, CurrentID=0, Status='live', Symbol='OR2AM1P', Aliases='OR37E', description='olfactory receptor family 2 subfamily AM member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AM, member 1 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36021712, end_position_on_the_genomic_accession=36022116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873465, CurrentID=0, Status='live', Symbol='RNA5SP204', Aliases='RN5S204', description='RNA, 5S ribosomal pseudogene 204', other_designations='RNA, 5S ribosomal 204', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17721786, end_position_on_the_genomic_accession=17721903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373604, CurrentID=0, Status='live', Symbol='LOC105373604', Aliases='', description='uncharacterized LOC105373604', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127023810, end_position_on_the_genomic_accession=127035755, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  343515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343515, CurrentID=0, Status='live', Symbol='ATP6V1E1P1', Aliases='ATP6EP1, ATP6V1EP1', description='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E1 pseudogene 1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) 31kD pseudogene 1|ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E pseudogene 1', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42903114, end_position_on_the_genomic_accession=42904112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873466, CurrentID=0, Status='live', Symbol='RNA5SP205', Aliases='RN5S205', description='RNA, 5S ribosomal pseudogene 205', other_designations='RNA, 5S ribosomal 205', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19438283, end_position_on_the_genomic_accession=19438418, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373605, CurrentID=0, Status='live', Symbol='LOC105373605', Aliases='', description='uncharacterized LOC105373605', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127107499, end_position_on_the_genomic_accession=127136514, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  343521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343521, CurrentID=0, Status='live', Symbol='TCTEX1D4', Aliases='', description='Tctex1 domain containing 4', other_designations='Tctex2 beta|novel Tctex-1 family domain-containing protein|protein N22.1|tctex-2-beta', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44805908, end_position_on_the_genomic_accession=44807442, orientation='minus', exon_count=3, OMIM=611713, no_hdr0=''),
  81382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81382, CurrentID=0, Status='live', Symbol='OR9P1P', Aliases='OR9P2P', description='olfactory receptor family 9 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 9, subfamily P, member 1 pseudogene|olfactory receptor, family 9, subfamily P, member 2 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143047067, end_position_on_the_genomic_accession=143047772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81383, CurrentID=0, Status='live', Symbol='OR9N1P', Aliases='', description='olfactory receptor family 9 subfamily N member 1 pseudogene', other_designations='olfactory receptor, family 9, subfamily N, member 1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141911402, end_position_on_the_genomic_accession=141912022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873468, CurrentID=0, Status='live', Symbol='RNA5SP208', Aliases='RN5S208', description='RNA, 5S ribosomal pseudogene 208', other_designations='RNA, 5S ribosomal 208', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67467231, end_position_on_the_genomic_accession=67467308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81390, CurrentID=0, Status='live', Symbol='OR4F7P', Aliases='OR4F10', description='olfactory receptor family 4 subfamily F member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 10|olfactory receptor, family 4, subfamily F, member 7 pseudogene', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170639606, end_position_on_the_genomic_accession=170640543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873469, CurrentID=0, Status='live', Symbol='RNA5SP209', Aliases='RN5S209', description='RNA, 5S ribosomal pseudogene 209', other_designations='RNA, 5S ribosomal 209', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75865239, end_position_on_the_genomic_accession=75865335, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81392, CurrentID=0, Status='live', Symbol='OR2AE1', Aliases='OR2AE2', description='olfactory receptor family 2 subfamily AE member 1', other_designations='olfactory receptor 2AE2|olfactory receptor, family 2, subfamily AE, member 1|olfactory receptor, family 2, subfamily AE, member 2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99876062, end_position_on_the_genomic_accession=99877033, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379688, CurrentID=0, Status='live', Symbol='LOC105379688', Aliases='', description='uncharacterized LOC105379688', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873470, CurrentID=0, Status='live', Symbol='RNA5SP210', Aliases='RN5S210', description='RNA, 5S ribosomal pseudogene 210', other_designations='RNA, 5S ribosomal 210', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81622200, end_position_on_the_genomic_accession=81622319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81399, CurrentID=0, Status='live', Symbol='OR4F16', Aliases='OR1-1, OR7-21', description='olfactory receptor family 4 subfamily F member 16', other_designations='olfactory receptor OR1-1|olfactory receptor OR4F16|olfactory receptor OR7-21|olfactory receptor, family 4, subfamily F, member 16', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=685716, end_position_on_the_genomic_accession=686654, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81406, CurrentID=0, Status='live', Symbol='OR2W6P', Aliases='OR2W7P', description='olfactory receptor family 2 subfamily W member 6 pseudogene', other_designations='olfactory receptor, family 2, subfamily W, member 6 pseudogene|olfactory receptor, family 2, subfamily W, member 7 pseudogene|seven transmembrane helix receptor', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27937404, end_position_on_the_genomic_accession=27938401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81407, CurrentID=0, Status='live', Symbol='OR2W4P', Aliases='', description='olfactory receptor family 2 subfamily W member 4 pseudogene', other_designations='olfactory receptor, family 2, subfamily W, member 4 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27977151, end_position_on_the_genomic_accession=27978073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873472, CurrentID=0, Status='live', Symbol='RNA5SP212', Aliases='RN5S212', description='RNA, 5S ribosomal pseudogene 212', other_designations='RNA, 5S ribosomal 212', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108252401, end_position_on_the_genomic_accession=108252507, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130731, CurrentID=0, Status='live', Symbol='LOC100130731', Aliases='', description='proteasome subunit beta 1 pseudogene', other_designations='proteasome (prosome, macropain) subunit, beta type, 1 pseudogene', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7736439, end_position_on_the_genomic_accession=7737095, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373612, CurrentID=0, Status='live', Symbol='LOC105373612', Aliases='', description='uncharacterized LOC105373612', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129243286, end_position_on_the_genomic_accession=129289156, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  343563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343563, CurrentID=0, Status='live', Symbol='OR2T29', Aliases='', description='olfactory receptor family 2 subfamily T member 29', other_designations='olfactory receptor, family 2, subfamily T, member 29', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248558544, end_position_on_the_genomic_accession=248559491, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81421, CurrentID=0, Status='live', Symbol='OR5M14P', Aliases='OR5M15P', description='olfactory receptor family 5 subfamily M member 14 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 14 pseudogene', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41723210, end_position_on_the_genomic_accession=41723563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647085, CurrentID=0, Status='live', Symbol='PPIAP25', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 25', other_designations='peptidylprolyl isomerase A pseudogene', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45436127, end_position_on_the_genomic_accession=45436844, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81428, CurrentID=0, Status='live', Symbol='OR7E122P', Aliases='OST719', description='olfactory receptor family 7 subfamily E member 122 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 122 pseudogene|olfactory receptor, family 7, subfamily E, member 122 pseudogene', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8688234, end_position_on_the_genomic_accession=8689154, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647086, CurrentID=0, Status='live', Symbol='LOC647086', Aliases='', description='chromosome 20 open reading frame 27 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30821072, end_position_on_the_genomic_accession=30821520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81430, CurrentID=0, Status='live', Symbol='OR7E100P', Aliases='OR7E135P', description='olfactory receptor family 7 subfamily E member 100 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 100 pseudogene|olfactory receptor, family 7, subfamily E, member 135 pseudogene', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112524186, end_position_on_the_genomic_accession=112525200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81431, CurrentID=0, Status='live', Symbol='OR5AC1', Aliases='OR5AC1P', description='olfactory receptor family 5 subfamily AC member 1 (gene/pseudogene)', other_designations='Olfactory receptor 5AC1|olfactory receptor, family 5, subfamily AC, member 1 pseudogene', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98064471, end_position_on_the_genomic_accession=98065392, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  343578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343578, CurrentID=0, Status='live', Symbol='ARHGAP40', Aliases='C20orf95, dJ1100H13.4', description='Rho GTPase activating protein 40', other_designations='rho-type GTPase-activating protein 40', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38601934, end_position_on_the_genomic_accession=38650652, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105373615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373615, CurrentID=0, Status='live', Symbol='LOC105373615', Aliases='', description='uncharacterized LOC105373615', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129862281, end_position_on_the_genomic_accession=129877563, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  81438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81438, CurrentID=0, Status='live', Symbol='OR4G6P', Aliases='OR11-302', description='olfactory receptor family 4 subfamily G member 6 pseudogene', other_designations='olfactory receptor OR11-302 pseudogene|olfactory receptor, family 4, subfamily G, member 6 pseudogene', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101937549, end_position_on_the_genomic_accession=101938492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81439, CurrentID=0, Status='live', Symbol='OR9H1P', Aliases='UNQ9373', description='olfactory receptor family 9 subfamily H member 1 pseudogene', other_designations='VNFT9373|olfactory receptor, family 9, subfamily H, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247774911, end_position_on_the_genomic_accession=247775834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81440, CurrentID=0, Status='live', Symbol='OR6R1P', Aliases='', description='olfactory receptor family 6 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 6, subfamily R, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247833317, end_position_on_the_genomic_accession=247834245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81442, CurrentID=0, Status='live', Symbol='OR6N2', Aliases='OR1-23', description='olfactory receptor family 6 subfamily N member 2', other_designations='olfactory receptor OR1-23|olfactory receptor, family 6, subfamily N, member 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158776682, end_position_on_the_genomic_accession=158777635, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  347509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347509, CurrentID=0, Status='live', Symbol='RPL7L1P11', Aliases='', description='ribosomal protein L7 like 1 pseudogene 11', other_designations='ribosomal protein L7-like 1 pseudogene 11', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=148464685, end_position_on_the_genomic_accession=148465437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81448, CurrentID=0, Status='live', Symbol='OR6K2', Aliases='OR1-17', description='olfactory receptor family 6 subfamily K member 2', other_designations='olfactory receptor OR1-17|olfactory receptor, family 6, subfamily K, member 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158699678, end_position_on_the_genomic_accession=158700652, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81449, CurrentID=0, Status='live', Symbol='OR6K1P', Aliases='', description='olfactory receptor family 6 subfamily K member 1 pseudogene', other_designations='olfactory receptor, family 6, subfamily K, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158694539, end_position_on_the_genomic_accession=158695509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81458, CurrentID=0, Status='live', Symbol='OR2T7', Aliases='OR2T7P, OST723', description='olfactory receptor family 2 subfamily T member 7', other_designations='Olfactory receptor 2T7|olfactory receptor OR1-44|olfactory receptor, family 2, subfamily T, member 7 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248441182, end_position_on_the_genomic_accession=248442133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419645, CurrentID=0, Status='live', Symbol='LOC100419645', Aliases='', description='ring finger protein 14 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135128446, end_position_on_the_genomic_accession=135129689, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81462, CurrentID=0, Status='live', Symbol='OR2L9P', Aliases='', description='olfactory receptor family 2 subfamily L member 9 pseudogene', other_designations='olfactory receptor, family 2, subfamily L, member 9 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247974743, end_position_on_the_genomic_accession=247975676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81465, CurrentID=0, Status='live', Symbol='OR2L6P', Aliases='', description='olfactory receptor family 2 subfamily L member 6 pseudogene', other_designations='olfactory receptor, family 2, subfamily L, member 6 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248003131, end_position_on_the_genomic_accession=248004069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81466, CurrentID=0, Status='live', Symbol='OR2L5', Aliases='OR2L11P, OR2L5', description='olfactory receptor family 2 subfamily L member 5', other_designations='olfactory receptor 2L11|olfactory receptor OR1-53|olfactory receptor, family 2, subfamily L, member 11|olfactory receptor, family 2, subfamily L, member 5 pseudogene|seven transmembrane helix receptor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248021948, end_position_on_the_genomic_accession=248022886, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81469, CurrentID=0, Status='live', Symbol='OR2G3', Aliases='OR1-33', description='olfactory receptor family 2 subfamily G member 3', other_designations='olfactory receptor OR1-33|olfactory receptor, family 2, subfamily G, member 3', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247605586, end_position_on_the_genomic_accession=247606515, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81470, CurrentID=0, Status='live', Symbol='OR2G2', Aliases='OR1-32', description='olfactory receptor family 2 subfamily G member 2', other_designations='olfactory receptor OR1-32|olfactory receptor, family 2, subfamily G, member 2 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247588360, end_position_on_the_genomic_accession=247589313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130741, CurrentID=0, Status='live', Symbol='LOC100130741', Aliases='', description='uncharacterized LOC100130741', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=157550018, end_position_on_the_genomic_accession=157550413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81472, CurrentID=0, Status='live', Symbol='OR2C3', Aliases='OR2C4, OR2C5P, OST742', description='olfactory receptor family 2 subfamily C member 3', other_designations='olfactory receptor 2C4|olfactory receptor 2C5|olfactory receptor OR1-30|olfactory receptor, family 2, subfamily C, member 3|olfactory receptor, family 2, subfamily C, member 4|olfactory receptor, family 2, subfamily C, member 5 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247530132, end_position_on_the_genomic_accession=247533839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81473, CurrentID=0, Status='live', Symbol='OR2AS1P', Aliases='', description='olfactory receptor family 2 subfamily AS member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AS, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248549422, end_position_on_the_genomic_accession=248549903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81474, CurrentID=0, Status='live', Symbol='OR2AQ1P', Aliases='', description='olfactory receptor family 2 subfamily AQ member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AQ, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158795998, end_position_on_the_genomic_accession=158796238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647094, CurrentID=0, Status='live', Symbol='RPLP0P11', Aliases='RPLP0_4_1600', description='ribosomal protein, large, P0 pseudogene 11', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50457814, end_position_on_the_genomic_accession=50458859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81479, CurrentID=0, Status='live', Symbol='OR11I1P', Aliases='OR11I2P', description='olfactory receptor family 11 subfamily I member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily I, member 1 pseudogene|olfactory receptor, family 11, subfamily I, member 2 pseudogene', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110854099, end_position_on_the_genomic_accession=110855092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379691, CurrentID=0, Status='live', Symbol='LOC105379691', Aliases='', description='uncharacterized LOC105379691', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81486, CurrentID=0, Status='live', Symbol='OR10AE1P', Aliases='OR10AE2P', description='olfactory receptor family 10 subfamily AE member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily AE, member 1 pseudogene|olfactory receptor, family 10, subfamily AE, member 2 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159581623, end_position_on_the_genomic_accession=159582220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81487, CurrentID=0, Status='live', Symbol='OR10AA1P', Aliases='', description='olfactory receptor family 10 subfamily AA member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily AA, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158808402, end_position_on_the_genomic_accession=158809335, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81488, CurrentID=0, Status='live', Symbol='POLR2M', Aliases='GCOM1, GRINL1A, Gdown, Gdown1', description='polymerase (RNA) II (DNA directed) polypeptide M', other_designations='DNA-directed RNA polymerase II subunit M, isoforms 4/5|GRINL1A downstream protein Gdown4|glutamate receptor, ionotropic, N-methyl D-aspartate-like 1A|protein GRINL1A, isoforms 4/5', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57706521, end_position_on_the_genomic_accession=57717557, orientation='plus', exon_count=4, OMIM=606485, no_hdr0=''),
  100130744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130744, CurrentID=0, Status='live', Symbol='LOC100130744', Aliases='', description='uncharacterized LOC100130744', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14712694, end_position_on_the_genomic_accession=14716529, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81490, CurrentID=0, Status='live', Symbol='PTDSS2', Aliases='PSS2', description='phosphatidylserine synthase 2', other_designations='PSS-2|ptdSer synthase 2|serine-exchange enzyme II', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=448271, end_position_on_the_genomic_accession=491399, orientation='plus', exon_count=15, OMIM=612793, no_hdr0=''),
  81491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81491, CurrentID=0, Status='live', Symbol='GPR63', Aliases='PSP24(beta), PSP24B', description='G protein-coupled receptor 63', other_designations='PSP24-2|PSP24-beta|brain expressed G-protein-coupled receptor PSP24 beta', map_location='6q16.1-q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96794122, end_position_on_the_genomic_accession=96837492, orientation='minus', exon_count=10, OMIM=606915, no_hdr0=''),
  81492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81492, CurrentID=0, Status='live', Symbol='RSPH6A', Aliases='RSHL1, RSP4, RSP6, RSPH4B', description='radial spoke head 6 homolog A (Chlamydomonas)', other_designations='ortholog of mouse radial spokehead-like 1|radial spoke head-like protein 1|radial spokehead-like 1', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45795710, end_position_on_the_genomic_accession=45815347, orientation='minus', exon_count=6, OMIM=607548, no_hdr0=''),
  343637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343637, CurrentID=0, Status='live', Symbol='RSPO4', Aliases='C20orf182, CRISTIN4', description='R-spondin 4', other_designations='R-spondin family, member 4|hRspo4|roof plate-specific spondin-4', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=958452, end_position_on_the_genomic_accession=1002284, orientation='minus', exon_count=8, OMIM=610573, no_hdr0=''),
  81494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81494, CurrentID=0, Status='live', Symbol='CFHR5', Aliases='CFHL5D, FHR-5, FHR5, CFHR5', description='complement factor H-related 5', other_designations='factor H-related protein 5', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196975037, end_position_on_the_genomic_accession=197009678, orientation='plus', exon_count=11, OMIM=608593, no_hdr0=''),
  105373625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373625, CurrentID=0, Status='live', Symbol='LOC105373625', Aliases='', description='uncharacterized LOC105373625', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132673312, end_position_on_the_genomic_accession=132675855, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  343641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343641, CurrentID=0, Status='live', Symbol='TGM6', Aliases='SCA35, TG6, TGM3L, TGY, dJ734P14.3', description='transglutaminase 6', other_designations='TGase Y|TGase-6|transglutaminase y', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2380908, end_position_on_the_genomic_accession=2432753, orientation='plus', exon_count=13, OMIM=613900, no_hdr0=''),
  81501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81501, CurrentID=0, Status='live', Symbol='DCSTAMP', Aliases='FIND, TM7SF4, hDC-STAMP', description='dendrocyte expressed seven transmembrane protein', other_designations='DC-specific transmembrane protein|Dendritic cells (DC)-specific transmembrane protein|IL-4-induced protein|IL-Four (IL-4) induced|IL-Four INDuced|IL-four-induced protein|dendrocyte-expressed seven transmembrane protein|transmembrane 7 superfamily member 4', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104339087, end_position_on_the_genomic_accession=104356689, orientation='plus', exon_count=6, OMIM=605933, no_hdr0=''),
  81502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81502, CurrentID=0, Status='live', Symbol='HM13', Aliases='H13, IMP1, IMPAS, IMPAS-1, MSTP086, PSENL3, PSL3, SPP, SPPL1', description='histocompatibility (minor) 13', other_designations='histocompatibility 13|intramembrane protease 1|minor histocompatibility antigen 13|presenilin-like protein 3|signal peptide peptidase beta|signal peptide peptidase like 1', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31514410, end_position_on_the_genomic_accession=31569567, orientation='plus', exon_count=13, OMIM=607106, no_hdr0=''),
  100873488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873488, CurrentID=0, Status='live', Symbol='RNA5SP230', Aliases='RN5S230', description='RNA, 5S ribosomal pseudogene 230', other_designations='RNA, 5S ribosomal 230', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44125752, end_position_on_the_genomic_accession=44125873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373627, CurrentID=0, Status='live', Symbol='LOC105373627', Aliases='', description='uncharacterized LOC105373627', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=133952986, end_position_on_the_genomic_accession=133993284, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100873489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873489, CurrentID=0, Status='live', Symbol='RNA5SP231', Aliases='RN5S231', description='RNA, 5S ribosomal pseudogene 231', other_designations='RNA, 5S ribosomal 231', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=68723911, end_position_on_the_genomic_accession=68724048, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130748, CurrentID=0, Status='live', Symbol='LOC100130748', Aliases='', description='3-hydroxyisobutyryl-Coenzyme A hydrolase pseudogene', other_designations='3-hydroxyisobutyryl-CoA hydrolase pseudogene', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2787251, end_position_on_the_genomic_accession=2788608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370963, CurrentID=0, Status='live', Symbol='LOC105370963', Aliases='', description='uncharacterized LOC105370963', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89461158, end_position_on_the_genomic_accession=89469729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102577426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102577426, CurrentID=0, Status='live', Symbol='LOC102577426', Aliases='', description='uncharacterized LOC102577426', other_designations='CTC-338M12.2', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181191046, end_position_on_the_genomic_accession=181191908, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101269108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101269108, CurrentID=0, Status='live', Symbol='LOC101269108', Aliases='', description='zinc finger, FYVE domain containing 9 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22689714, end_position_on_the_genomic_accession=22691251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647102, CurrentID=0, Status='live', Symbol='ELK1P1', Aliases='', description='ELK1, member of ETS oncogene family pseudogene 1', other_designations='ELK1 pseudogene 1', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45391624, end_position_on_the_genomic_accession=45392252, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81532, CurrentID=0, Status='live', Symbol='MOB2', Aliases='HCCA2', description='MOB kinase activator 2', other_designations='MOB2 Mps One Binder homolog|Mps one binder kinase activator-like 2', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1469448, end_position_on_the_genomic_accession=1486779, orientation='minus', exon_count=5, OMIM=611969, no_hdr0=''),
  81533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81533, CurrentID=0, Status='live', Symbol='ITFG1', Aliases='2310047C21Rik, CDA08, TIP', description='integrin alpha FG-GAP repeat containing 1', other_designations='LINKIN|integrin-alpha FG-GAP repeat-containing protein 1', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47154387, end_position_on_the_genomic_accession=47461274, orientation='minus', exon_count=18, OMIM=611803, no_hdr0=''),
  100873493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873493, CurrentID=0, Status='live', Symbol='RNA5SP235', Aliases='RN5S235', description='RNA, 5S ribosomal pseudogene 235', other_designations='RNA, 5S ribosomal 235', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83017898, end_position_on_the_genomic_accession=83018011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81537, CurrentID=0, Status='live', Symbol='SGPP1', Aliases='SPPase1', description='sphingosine-1-phosphate phosphatase 1', other_designations='hSPP1|hSPPase1|sphingosine-1-phosphatase 1|spp1', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63684216, end_position_on_the_genomic_accession=63728095, orientation='minus', exon_count=3, OMIM=612826, no_hdr0=''),
  81539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81539, CurrentID=0, Status='live', Symbol='SLC38A1', Aliases='ATA1, NAT2, SAT1, SNAT1', description='solute carrier family 38 member 1', other_designations='N-system amino acid transporter 2|amino acid transporter A1|amino acid transporter system A1|solute carrier family 38, member 1|system A amino acid transporter 1|system N amino acid transporter 1', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46183055, end_position_on_the_genomic_accession=46269425, orientation='minus', exon_count=22, OMIM=608490, no_hdr0=''),
  81542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81542, CurrentID=0, Status='live', Symbol='TMX1', Aliases='PDIA11, TMX, TXNDC, TXNDC1', description='thioredoxin-related transmembrane protein 1', other_designations='protein disulfide isomerase family A, member 11|thioredoxin domain containing 1|thioredoxin domain-containing protein 1|transmembrane Trx-related protein', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51240168, end_position_on_the_genomic_accession=51257655, orientation='plus', exon_count=8, OMIM=610527, no_hdr0=''),
  81543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81543, CurrentID=0, Status='live', Symbol='LRRC3', Aliases='C21orf102', description='leucine rich repeat containing 3', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44455486, end_position_on_the_genomic_accession=44458856, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81544, CurrentID=0, Status='live', Symbol='GDPD5', Aliases='GDE2, PP1665', description='glycerophosphodiester phosphodiesterase domain containing 5', other_designations='glycerophosphodiester phosphodiesterase 2', map_location='11q13.4-q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75434640, end_position_on_the_genomic_accession=75525920, orientation='minus', exon_count=23, OMIM=609632, no_hdr0=''),
  81545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81545, CurrentID=0, Status='live', Symbol='FBXO38', Aliases='Fbx38, HMN2D, MOKA, SP329', description='F-box protein 38', other_designations='modulator of KLF7 activity homolog', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148383935, end_position_on_the_genomic_accession=148442836, orientation='plus', exon_count=24, OMIM=608533, no_hdr0=''),
  81550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81550, CurrentID=0, Status='live', Symbol='TDRD3', Aliases='', description='tudor domain containing 3', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60396457, end_position_on_the_genomic_accession=60573879, orientation='plus', exon_count=18, OMIM=614392, no_hdr0=''),
  81551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81551, CurrentID=0, Status='live', Symbol='STMN4', Aliases='RB3', description='stathmin 4', other_designations='stathmin-like 4|stathmin-like protein B3|stathmin-like-protein RB3', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27235323, end_position_on_the_genomic_accession=27258865, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  81552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81552, CurrentID=0, Status='live', Symbol='VOPP1', Aliases='ECOP, GASP', description='vesicular, overexpressed in cancer, prosurvival protein 1', other_designations='EGFR-coamplified and overexpressed protein|Glioblastoma-amplified secreted protein|hypothetical protein DKFZp564K0822|putative NF-kappa-B-activating protein 055N', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55434406, end_position_on_the_genomic_accession=55573355, orientation='minus', exon_count=14, OMIM=611915, no_hdr0=''),
  81553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81553, CurrentID=0, Status='live', Symbol='FAM49A', Aliases='', description='family with sequence similarity 49 member A', other_designations='family with sequence similarity 49, member A', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16549459, end_position_on_the_genomic_accession=16665866, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  81554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81554, CurrentID=0, Status='live', Symbol='WBSCR16', Aliases='', description='Williams-Beuren syndrome chromosome region 16', other_designations='RCC1-like G exchanging factor-like protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75027119, end_position_on_the_genomic_accession=75073881, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  81555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81555, CurrentID=0, Status='live', Symbol='YIPF5', Aliases='FinGER5, SB140, SMAP-5, SMAP5, YIP1A', description='Yip1 domain family member 5', other_designations='YIP1 family member 5|YPT-interacting protein 1 A|Yip1 domain family, member 5|five-pass transmembrane protein localizing in the Golgi apparatus and the endoplasmic reticulum 5|golgi membrane protein SB140|smooth muscle cell associated protein 5|smooth muscle cell-associated protein 5', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=144158159, end_position_on_the_genomic_accession=144170714, orientation='minus', exon_count=6, OMIM=611483, no_hdr0=''),
  81556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81556, CurrentID=0, Status='live', Symbol='VWA9', Aliases='C15orf44', description='von Willebrand factor A domain containing 9', other_designations='UPF0464 protein C15orf44', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65578757, end_position_on_the_genomic_accession=65611289, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  81557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81557, CurrentID=0, Status='live', Symbol='MAGED4B', Aliases='', description='melanoma antigen family D4B', other_designations='MAGE-D4 antigen|MAGE-E1 antigen|melanoma antigen family D, 4B', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52061827, end_position_on_the_genomic_accession=52069272, orientation='minus', exon_count=15, OMIM=300765, no_hdr0=''),
  343702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343702, CurrentID=0, Status='live', Symbol='XKR7', Aliases='C20orf159, dJ310O13.4', description='X-linked Kx blood group related 7', other_designations='X Kell blood group precursor-related family, member 7|XK, Kell blood group complex subunit-related family, member 7', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31968002, end_position_on_the_genomic_accession=31998453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  81559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81559, CurrentID=0, Status='live', Symbol='TRIM11', Aliases='BIA1, RNF92', description='tripartite motif containing 11', other_designations='RING finger protein 92|tripartite motif-containing 11|tripartite motif-containing protein 11', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228393672, end_position_on_the_genomic_accession=228406877, orientation='minus', exon_count=6, OMIM=607868, no_hdr0=''),
  105373636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373636, CurrentID=0, Status='live', Symbol='LOC105373636', Aliases='', description='uncharacterized LOC105373636', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138046743, end_position_on_the_genomic_accession=138060251, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  81562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81562, CurrentID=0, Status='live', Symbol='LMAN2L', Aliases='VIPL', description='lectin, mannose-binding 2-like', other_designations='LMAN2-like protein', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96705929, end_position_on_the_genomic_accession=96740082, orientation='minus', exon_count=10, OMIM=609552, no_hdr0=''),
  81563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81563, CurrentID=0, Status='live', Symbol='C1orf21', Aliases='PIG13', description='chromosome 1 open reading frame 21', other_designations='cell proliferation-inducing gene 13 protein|proliferation-inducing protein 13', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184387016, end_position_on_the_genomic_accession=184629021, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  81565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81565, CurrentID=0, Status='live', Symbol='NDEL1', Aliases='EOPA, MITAP1, NDE1L1, NDE2, NUDEL', description='nudE neurodevelopment protein 1-like 1', other_designations='endooligopeptidase A|mitosin-associated protein 1|mitosin-associated protein MITAP1|nudE nuclear distribution E homolog-like 1|nudE nuclear distribution gene E homolog-like 1|protein Nudel', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8435852, end_position_on_the_genomic_accession=8471708, orientation='plus', exon_count=12, OMIM=607538, no_hdr0=''),
  81566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81566, CurrentID=0, Status='live', Symbol='CSRNP2', Aliases='C12orf2, C12orf22, FAM130A1, PPP1R72, TAIP-12', description='cysteine-serine-rich nuclear protein 2', other_designations='CSRNP-2|TGF-beta-induced apoptosis protein 12|family with sequence similarity 130, member A1|protein phosphatase 1, regulatory subunit 72', map_location='12q13.11-q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51061205, end_position_on_the_genomic_accession=51083671, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  81567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81567, CurrentID=0, Status='live', Symbol='TXNDC5', Aliases='ENDOPDI, ERP46, HCC-2, HCC2, PDIA15, STRF8, UNQ364', description='thioredoxin domain containing 5 (endoplasmic reticulum)', other_designations='ER protein 46|endoplasmic reticulum protein ERp46|endoplasmic reticulum resident protein 46|endothelial protein disulphide isomerase|protein disulfide isomerase family A, member 15|thioredoxin related protein|thioredoxin-like protein p46', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7881250, end_position_on_the_genomic_accession=7910814, orientation='minus', exon_count=11, OMIM=616412, no_hdr0=''),
  81569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81569, CurrentID=0, Status='live', Symbol='ACTL8', Aliases='CT57', description='actin like 8', other_designations='actin like protein|actin-like 8|cancer/testis antigen 57', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17712000, end_position_on_the_genomic_accession=17827063, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  81570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81570, CurrentID=0, Status='live', Symbol='CLPB', Aliases='ANKCLB, HSP78, MEGCANN, MGCA7, SKD3', description='ClpB homolog, mitochondrial AAA ATPase chaperonin', other_designations='ClpB caseinolytic peptidase B homolog|ankyrin-repeat containing bacterial clp fusion|suppressor of potassium transport defect 3', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72292425, end_position_on_the_genomic_accession=72434680, orientation='minus', exon_count=19, OMIM=616254, no_hdr0=''),
  81571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81571, CurrentID=0, Status='live', Symbol='MIR600HG', Aliases='C9orf45, GL012, NCRNA00287', description='MIR600 host gene', other_designations='MIR600 host gene (non-protein coding)', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123109494, end_position_on_the_genomic_accession=123115477, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81572, CurrentID=0, Status='live', Symbol='PDRG1', Aliases='C20orf126, PDRG', description='p53 and DNA damage regulated 1', other_designations='p53 and DNA-damage regulated 1', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31944955, end_position_on_the_genomic_accession=31968710, orientation='minus', exon_count=6, OMIM=610789, no_hdr0=''),
  81573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81573, CurrentID=0, Status='live', Symbol='ANKRD13C', Aliases='dJ677H15.3', description='ankyrin repeat domain 13C', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70259002, end_position_on_the_genomic_accession=70354788, orientation='minus', exon_count=14, OMIM=615125, no_hdr0=''),
  81575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81575, CurrentID=0, Status='live', Symbol='APOLD1', Aliases='VERGE', description='apolipoprotein L domain containing 1', other_designations='vascular early response gene protein', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12725917, end_position_on_the_genomic_accession=12791466, orientation='plus', exon_count=3, OMIM=612456, no_hdr0=''),
  81576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81576, CurrentID=0, Status='live', Symbol='CCDC130', Aliases='', description='coiled-coil domain containing 130', other_designations='9 kDa protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13731763, end_position_on_the_genomic_accession=13763292, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  341276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341276, CurrentID=0, Status='live', Symbol='OR10A2', Aliases='OR10A2P, OR11-82, OR11-86, OST363', description='olfactory receptor family 10 subfamily A member 2', other_designations='hP4 olfactory receptor|olfactory receptor OR11-82 pseudogene|olfactory receptor OR11-86|olfactory receptor, family 10, subfamily A, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6869755, end_position_on_the_genomic_accession=6870666, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81578, CurrentID=0, Status='live', Symbol='COL21A1', Aliases='COLA1L, FP633, dJ682J15.1, dJ708F5.1', description='collagen, type XXI, alpha 1', other_designations='alpha 1 chain-like collagen|alpha 1 type XXI collagen', map_location='6p12.3-p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56056590, end_position_on_the_genomic_accession=56394122, orientation='minus', exon_count=36, OMIM=610002, no_hdr0=''),
  81579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81579, CurrentID=0, Status='live', Symbol='PLA2G12A', Aliases='GXII, PLA2G12, ROSSY', description='phospholipase A2 group XIIA', other_designations='GXII sPLA2|group XII secreted phospholipase A2|group XIIA secreted phospholipase A2|phosphatidylcholine 2-acylhydrolase 12A|phospholipase A2, group XII|phospholipase A2, group XIIA|sPLA2-XII', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109709989, end_position_on_the_genomic_accession=109730086, orientation='minus', exon_count=4, OMIM=611652, no_hdr0=''),
  341277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341277, CurrentID=0, Status='live', Symbol='OVCH2', Aliases='OVTN', description='ovochymase 2 (gene/pseudogene)', other_designations='oviductin protease', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7678692, end_position_on_the_genomic_accession=7706454, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105373640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373640, CurrentID=0, Status='live', Symbol='LOC105373640', Aliases='', description='uncharacterized LOC105373640', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138917225, end_position_on_the_genomic_accession=138975280, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873502, CurrentID=0, Status='live', Symbol='RNA5SP246', Aliases='RN5S246', description='RNA, 5S ribosomal pseudogene 246', other_designations='RNA, 5S ribosomal 246', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130602795, end_position_on_the_genomic_accession=130602913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373641, CurrentID=0, Status='live', Symbol='LOC105373641', Aliases='', description='uncharacterized LOC105373641', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139259920, end_position_on_the_genomic_accession=139264152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873503, CurrentID=0, Status='live', Symbol='RNA5SP247', Aliases='RN5S247', description='RNA, 5S ribosomal pseudogene 247', other_designations='RNA, 5S ribosomal 247', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140370441, end_position_on_the_genomic_accession=140370545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873504, CurrentID=0, Status='live', Symbol='RNA5SP248', Aliases='RN5S248', description='RNA, 5S ribosomal pseudogene 248', other_designations='RNA, 5S ribosomal 248', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140386781, end_position_on_the_genomic_accession=140386907, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81602, CurrentID=0, Status='live', Symbol='CDADC1', Aliases='NYD-SP15, bA103J18.1', description='cytidine and dCMP deaminase domain containing 1', other_designations='protein kinase NYD-SP15|testis development protein NYD-SP15', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49247875, end_position_on_the_genomic_accession=49293486, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  81603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81603, CurrentID=0, Status='live', Symbol='TRIM8', Aliases='GERP, RNF27', description='tripartite motif containing 8', other_designations='glioblastoma expressed ring finger protein|glioblastoma-expressed RING finger protein|ring finger protein 27|tripartite motif protein TRIM8|tripartite motif-containing 8|tripartite motif-containing protein 8', map_location='10q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102642903, end_position_on_the_genomic_accession=102658319, orientation='plus', exon_count=6, OMIM=606125, no_hdr0=''),
  81605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81605, CurrentID=0, Status='live', Symbol='URM1', Aliases='C9orf74', description='ubiquitin related modifier 1', other_designations='ubiquitin related modifier 1 homolog|ubiquitin-related modifier 1 homolog', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128371319, end_position_on_the_genomic_accession=128392016, orientation='plus', exon_count=3, OMIM=612693, no_hdr0=''),
  81606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81606, CurrentID=0, Status='live', Symbol='LBH', Aliases='', description='limb bud and heart development', other_designations='hLBH|limb bud and heart development homolog', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30231531, end_position_on_the_genomic_accession=30260033, orientation='plus', exon_count=3, OMIM=611763, no_hdr0=''),
  81607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81607, CurrentID=0, Status='live', Symbol='PVRL4', Aliases='EDSS1, LNIR, PRR4, nectin-4', description='poliovirus receptor-related 4', other_designations='Ig superfamily receptor LNIR|nectin 4|poliovirus receptor-related protein 4', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161070991, end_position_on_the_genomic_accession=161089595, orientation='minus', exon_count=9, OMIM=609607, no_hdr0=''),
  81608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81608, CurrentID=0, Status='live', Symbol='FIP1L1', Aliases='FIP1, Rhe, hFip1', description='factor interacting with PAPOLA and CPSF1', other_designations='FIP1 like 1|FIP1-like 1 protein|FIP1L1 cleavage and polyadenylation specific factor subunit|factor interacting with PAP|rearranged in hypereosinophilia', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53377646, end_position_on_the_genomic_accession=53462611, orientation='plus', exon_count=19, OMIM=607686, no_hdr0=''),
  81609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81609, CurrentID=0, Status='live', Symbol='SNX27', Aliases='MRT1, MY014', description='sorting nexin family member 27', other_designations='methamphetamine-responsive transcript 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151611984, end_position_on_the_genomic_accession=151699083, orientation='plus', exon_count=13, OMIM=611541, no_hdr0=''),
  81610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81610, CurrentID=0, Status='live', Symbol='FAM83D', Aliases='C20orf129, CHICA, dJ616B8.3', description='family with sequence similarity 83 member D', other_designations='family with sequence similarity 83, member D|spindle protein CHICA', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38925078, end_position_on_the_genomic_accession=38953060, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  81611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81611, CurrentID=0, Status='live', Symbol='ANP32E', Aliases='LANP-L, LANPL', description='acidic nuclear phosphoprotein 32 family member E', other_designations='LANP-like protein|acidic (leucine-rich) nuclear phosphoprotein 32 family, member E|leucine-rich acidic nuclear protein like', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150218417, end_position_on_the_genomic_accession=150236156, orientation='minus', exon_count=8, OMIM=609611, no_hdr0=''),
  81614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81614, CurrentID=0, Status='live', Symbol='NIPA2', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2', other_designations='non-imprinted in Prader-Willi/Angelman syndrome region protein 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22838641, end_position_on_the_genomic_accession=22868384, orientation='plus', exon_count=11, OMIM=608146, no_hdr0=''),
  81615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81615, CurrentID=0, Status='live', Symbol='TMEM163', Aliases='DC29, SV31', description='transmembrane protein 163', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134455759, end_position_on_the_genomic_accession=134736424, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  81616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81616, CurrentID=0, Status='live', Symbol='ACSBG2', Aliases='BGR, BRGL, PRTD-NY3, PRTDNY3', description='acyl-CoA synthetase bubblegum family member 2', other_designations='bubblegum-related protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6135633, end_position_on_the_genomic_accession=6193103, orientation='plus', exon_count=16, OMIM=614363, no_hdr0=''),
  81617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81617, CurrentID=0, Status='live', Symbol='CAB39L', Aliases='MO25-BETA, MO2L, bA103J18.3', description='calcium binding protein 39 like', other_designations='MO25beta|U937-associated antigen|antigen MLAA-34|calcium binding protein 39-like|mo25-like protein|sarcoma antigen NY-SAR-79', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49308647, end_position_on_the_genomic_accession=49444134, orientation='minus', exon_count=16, OMIM=612175, no_hdr0=''),
  81618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81618, CurrentID=0, Status='live', Symbol='ITM2C', Aliases='BRI3, BRICD2C, E25, E25C, ITM3', description='integral membrane protein 2C', other_designations='BRICHOS domain containing 2C|cerebral protein 14|integral membrane protein 3|transmembrane protein BRI3', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230864598, end_position_on_the_genomic_accession=230879254, orientation='plus', exon_count=7, OMIM=609554, no_hdr0=''),
  81619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81619, CurrentID=0, Status='live', Symbol='TSPAN14', Aliases='DC-TM4F2, TM4SF14', description='tetraspanin 14', other_designations='tetraspanin similar to TM4SF9|transmembrane 4 superfamily member 14|tspan-14', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80454278, end_position_on_the_genomic_accession=80522635, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  81620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81620, CurrentID=0, Status='live', Symbol='CDT1', Aliases='DUP, RIS2', description='chromatin licensing and DNA replication factor 1', other_designations='Double parked, Drosophila, homolog of', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88803778, end_position_on_the_genomic_accession=88809258, orientation='plus', exon_count=10, OMIM=605525, no_hdr0=''),
  81621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81621, CurrentID=0, Status='live', Symbol='KAZALD1', Aliases='BONO1, FKSG28, FKSG40, IGFBP-rP10', description='Kazal-type serine peptidase inhibitor domain 1', other_designations='IGFBP-related protein 10|Kazal-type serine protease inhibitor domain 1|bone- and odontoblast-expressed gene 1|novel kazal-type serine protease inhibitor domain and immunoglobulin domain containing protein', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101061141, end_position_on_the_genomic_accession=101068131, orientation='plus', exon_count=6, OMIM=609208, no_hdr0=''),
  81622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81622, CurrentID=0, Status='live', Symbol='UNC93B1', Aliases='IIAE1, UNC93, UNC93B, Unc-93B1', description='unc-93 homolog B1 (C. elegans)', other_designations='unc-93 related protein|unc93 homolog B1', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67991100, end_position_on_the_genomic_accession=68004125, orientation='minus', exon_count=11, OMIM=608204, no_hdr0=''),
  81623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81623, CurrentID=0, Status='live', Symbol='DEFB126', Aliases='C20orf8, DEFB-26, DEFB26, bA530N10.1, hBD-26', description='defensin beta 126', other_designations='defensin, beta 126|epididymal secretory protein ESP13.2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=142553, end_position_on_the_genomic_accession=145751, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81624, CurrentID=0, Status='live', Symbol='DIAPH3', Aliases='AN, AUNA1, DIA2, DRF3, NSDAN, diap3, mDia2', description='diaphanous related formin 3', other_designations='diaphanous homolog 3|diaphanous-related formin 3|diaphanous-related formin-3', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59665583, end_position_on_the_genomic_accession=60163985, orientation='minus', exon_count=39, OMIM=614567, no_hdr0=''),
  100873508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873508, CurrentID=0, Status='live', Symbol='RNA5SP252', Aliases='RN5S252', description='RNA, 5S ribosomal pseudogene 252', other_designations='RNA, 5S ribosomal 252', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10686569, end_position_on_the_genomic_accession=10686687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81626, CurrentID=0, Status='live', Symbol='SHCBP1L', Aliases='C1orf14, GE36', description='SHC SH2-domain binding protein 1-like', other_designations='GE36', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182899865, end_position_on_the_genomic_accession=182953418, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  81627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81627, CurrentID=0, Status='live', Symbol='TRMT1L', Aliases='C1orf25, MST070, MSTP070, TRM1L, bG120K12.3', description='tRNA methyltransferase 1 like', other_designations='TRM1 tRNA methyltransferase 1-like|TRM1-like protein|tRNA methyltransferase 1 homolog (S. cerevisiae)-like|tRNA methyltransferase 1-like', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185118085, end_position_on_the_genomic_accession=185157098, orientation='minus', exon_count=16, OMIM=611673, no_hdr0=''),
  81628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81628, CurrentID=0, Status='live', Symbol='TSC22D4', Aliases='THG-1, THG1, TILZ2', description='TSC22 domain family member 4', other_designations='TSC22 domain family, member 4|TSC22-related-inducible leucine zipper protein 2|tsc-22-like protein THG-1', map_location='7p21-p15', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100466519, end_position_on_the_genomic_accession=100479279, orientation='minus', exon_count=5, OMIM=611914, no_hdr0=''),
  81629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81629, CurrentID=0, Status='live', Symbol='TSSK3', Aliases='SPOGA3, STK22C, STK22D, TSK3', description='testis-specific serine kinase 3', other_designations='TSK-3|TSSK-3|serine/threonine-protein kinase 22C|spermiogenesis associated 3|testis-specific kinase 3|testis-specific serine/threonine kinase 22C', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32362261, end_position_on_the_genomic_accession=32364325, orientation='plus', exon_count=2, OMIM=607660, no_hdr0=''),
  81631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81631, CurrentID=0, Status='live', Symbol='MAP1LC3B', Aliases='ATG8F, LC3B, MAP1A/1BLC3-a, MAP1LC3B', description='microtubule associated protein 1 light chain 3 beta', other_designations='MAP1 light chain 3-like protein 2|MAP1A/MAP1B LC3 B|MAP1A/MAP1B light chain 3 B|autophagy-related ubiquitin-like modifier LC3 B|microtubule-associated protein 1 light chain 3 beta', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87392195, end_position_on_the_genomic_accession=87404774, orientation='plus', exon_count=4, OMIM=609604, no_hdr0=''),
  100130768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130768, CurrentID=0, Status='live', Symbol='LOC100130768', Aliases='', description='uncharacterized LOC100130768', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=128852395, end_position_on_the_genomic_accession=128853091, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873510, CurrentID=0, Status='live', Symbol='RNA5SP256', Aliases='RN5S256', description='RNA, 5S ribosomal pseudogene 256', other_designations='RNA, 5S ribosomal 256', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17748118, end_position_on_the_genomic_accession=17748230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373649, CurrentID=0, Status='live', Symbol='LOC105373649', Aliases='', description='uncharacterized LOC105373649', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=141459801, end_position_on_the_genomic_accession=141475120, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100270902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270902, CurrentID=0, Status='live', Symbol='RPS26P14', Aliases='RPS26_1_49', description='ribosomal protein S26 pseudogene 14', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51980446, end_position_on_the_genomic_accession=51980887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873511, CurrentID=0, Status='live', Symbol='RNA5SP257', Aliases='RN5S257', description='RNA, 5S ribosomal pseudogene 257', other_designations='RNA, 5S ribosomal 257', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20290174, end_position_on_the_genomic_accession=20290289, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644303, CurrentID=0, Status='live', Symbol='LOC644303', Aliases='', description='ATP-dependent RNA helicase DDX24-like', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118452516, end_position_on_the_genomic_accession=118454991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873512, CurrentID=0, Status='live', Symbol='RNA5SP258', Aliases='RN5S258', description='RNA, 5S ribosomal pseudogene 258', other_designations='RNA, 5S ribosomal 258', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26136880, end_position_on_the_genomic_accession=26136998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647123, CurrentID=0, Status='live', Symbol='UQCRFS1P2', Aliases='', description='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide 1 pseudogene 2', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138701659, end_position_on_the_genomic_accession=138703067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873513, CurrentID=0, Status='live', Symbol='RNA5SP259', Aliases='RN5S259', description='RNA, 5S ribosomal pseudogene 259', other_designations='RNA, 5S ribosomal 259', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28530411, end_position_on_the_genomic_accession=28530519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373653, CurrentID=0, Status='live', Symbol='LOC105373653', Aliases='', description='uncharacterized LOC105373653', other_designations='CTD-2252P21.1', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143737983, end_position_on_the_genomic_accession=143776114, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370968, CurrentID=0, Status='live', Symbol='LOC105370968', Aliases='', description='uncharacterized LOC105370968', other_designations='CTD-3065B20.2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90593762, end_position_on_the_genomic_accession=90604378, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873515, CurrentID=0, Status='live', Symbol='RNA5SP261', Aliases='RN5S261', description='RNA, 5S ribosomal pseudogene 261', other_designations='RNA, 5S ribosomal 261', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31598099, end_position_on_the_genomic_accession=31598204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81669, CurrentID=0, Status='live', Symbol='CCNL2', Aliases='ANIA-6B, CCNM, CCNS, HCLA-ISO, HLA-ISO, PCEE, SB138', description='cyclin L2', other_designations='cyclin M|cyclin S|paneth cell-enhanced expression protein', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1385711, end_position_on_the_genomic_accession=1399338, orientation='minus', exon_count=15, OMIM=613482, no_hdr0=''),
  81671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81671, CurrentID=0, Status='live', Symbol='VMP1', Aliases='EPG3, TANGO5, TMEM49', description='vacuole membrane protein 1', other_designations='ectopic P-granules autophagy protein 3 homolog|transmembrane protein 49|transport and golgi organization 5 homolog', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59707502, end_position_on_the_genomic_accession=59840593, orientation='plus', exon_count=12, OMIM=611753, no_hdr0=''),
  100873516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873516, CurrentID=0, Status='live', Symbol='RNA5SP262', Aliases='RN5S262', description='RNA, 5S ribosomal pseudogene 262', other_designations='RNA, 5S ribosomal 262', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=32192028, end_position_on_the_genomic_accession=32192153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343819, CurrentID=0, Status='live', Symbol='MRPL51P2', Aliases='', description='mitochondrial ribosomal protein L51 pseudogene 2', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43115216, end_position_on_the_genomic_accession=43115847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873517, CurrentID=0, Status='live', Symbol='RNA5SP264', Aliases='RN5S264', description='RNA, 5S ribosomal pseudogene 264', other_designations='RNA, 5S ribosomal 264', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36764445, end_position_on_the_genomic_accession=36764551, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927362, CurrentID=0, Status='live', Symbol='LOC101927362', Aliases='', description='uncharacterized LOC101927362', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=545799, end_position_on_the_genomic_accession=546667, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373656, CurrentID=0, Status='live', Symbol='LOC105373656', Aliases='', description='uncharacterized LOC105373656', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144523541, end_position_on_the_genomic_accession=144524580, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873518, CurrentID=0, Status='live', Symbol='RNA5SP265', Aliases='RN5S265', description='RNA, 5S ribosomal pseudogene 265', other_designations='RNA, 5S ribosomal 265', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55745232, end_position_on_the_genomic_accession=55745342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81688, CurrentID=0, Status='live', Symbol='C6orf62', Aliases='Nbla00237, XTP12, dJ30M3.2', description='chromosome 6 open reading frame 62', other_designations='HBV X-transactivated gene 12 protein|HBV X-transactivated protein 12|HBV XAg-transactivated protein 12', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24704861, end_position_on_the_genomic_accession=24720836, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  81689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81689, CurrentID=0, Status='live', Symbol='ISCA1', Aliases='HBLD2, ISA1, hIscA', description='iron-sulfur cluster assembly 1', other_designations='HESB like domain containing 2|HESB-like domain-containing protein 2|iron sulfur assembly protein IscA|iron-sulfur assembly protein IscA', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86264546, end_position_on_the_genomic_accession=86282575, orientation='minus', exon_count=4, OMIM=611006, no_hdr0=''),
  81691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81691, CurrentID=0, Status='live', Symbol='LOC81691', Aliases='', description='exonuclease NEF-sp', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20806429, end_position_on_the_genomic_accession=20849668, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  81693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81693, CurrentID=0, Status='live', Symbol='AMN', Aliases='PRO1028, amnionless', description='amnion associated transmembrane protein', other_designations='amnionless homolog|visceral endoderm-specific type 1 transmembrane protein', map_location='14q32.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102922548, end_position_on_the_genomic_accession=102930847, orientation='plus', exon_count=13, OMIM=605799, no_hdr0=''),
  81694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81694, CurrentID=0, Status='live', Symbol='OR2W2P', Aliases='hs6M1-30P', description='olfactory receptor family 2 subfamily W member 2 pseudogene', other_designations='olfactory receptor, family 2, subfamily W, member 2 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28033915, end_position_on_the_genomic_accession=28034862, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81695, CurrentID=0, Status='live', Symbol='OR2B7P', Aliases='hs6M1-31P', description='olfactory receptor family 2 subfamily B member 7 pseudogene', other_designations='olfactory receptor, family 2, subfamily B, member 7 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28046435, end_position_on_the_genomic_accession=28047369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81696, CurrentID=0, Status='live', Symbol='OR5V1', Aliases='6M1-21, hs6M1-21', description='olfactory receptor family 5 subfamily V member 1', other_designations='olfactory receptor OR6-26|olfactory receptor, family 5, subfamily V, member 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29355230, end_position_on_the_genomic_accession=29457067, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  81697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81697, CurrentID=0, Status='live', Symbol='OR2B2', Aliases='OR2B2Q, OR2B9, OR6-1, dJ193B12.4, hs6M1-10', description='olfactory receptor family 2 subfamily B member 2', other_designations='olfactory receptor 2B9|olfactory receptor 6-1|olfactory receptor OR6-2|olfactory receptor, family 2, subfamily B, member 2|olfactory receptor, family 2, subfamily B, member 9', map_location='6p22.3-p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27911185, end_position_on_the_genomic_accession=27912396, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81698, CurrentID=0, Status='live', Symbol='LINC00597', Aliases='C15orf5', description='long intergenic non-protein coding RNA 597', other_designations='', map_location='15q23-q24', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77223893, end_position_on_the_genomic_accession=77225404, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873521, CurrentID=0, Status='live', Symbol='RNA5SP268', Aliases='RN5S268', description='RNA, 5S ribosomal pseudogene 268', other_designations='RNA, 5S ribosomal 268', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67099302, end_position_on_the_genomic_accession=67099425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81704, CurrentID=0, Status='live', Symbol='DOCK8', Aliases='HEL-205, MRD2, ZIR8', description='dedicator of cytokinesis 8', other_designations='1200017A24Rik|epididymis luminal protein 205', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=214865, end_position_on_the_genomic_accession=465259, orientation='plus', exon_count=53, OMIM=611432, no_hdr0=''),
  100133545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133545, CurrentID=0, Status='live', Symbol='MRPL23-AS1', Aliases='', description='MRPL23 antisense RNA 1', other_designations='MRPL23 antisense RNA 1 (non-protein coding)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1983209, end_position_on_the_genomic_accession=1989920, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  81706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81706, CurrentID=0, Status='live', Symbol='PPP1R14C', Aliases='CPI17-like, KEPI, NY-BR-81', description='protein phosphatase 1 regulatory inhibitor subunit 14C', other_designations='PKC-potentiated PP1 inhibitory protein|kinase C-enhanced PP1 inhibitor|kinase-enhanced PP1 inhibitor|protein phosphatase 1, regulatory (inhibitor) subunit 14C|serologically defined breast cancer antigen NY-BR-81', map_location='6q24.3-q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150143052, end_position_on_the_genomic_accession=150250392, orientation='plus', exon_count=5, OMIM=613242, no_hdr0=''),
  343851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343851, CurrentID=0, Status='live', Symbol='RPS10P30', Aliases='RPS10_14_1741', description='ribosomal protein S10 pseudogene 30', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23322478, end_position_on_the_genomic_accession=23322928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873522, CurrentID=0, Status='live', Symbol='RNA5SP269', Aliases='RN5S269', description='RNA, 5S ribosomal pseudogene 269', other_designations='RNA, 5S ribosomal 269', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68695643, end_position_on_the_genomic_accession=68695779, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873523, CurrentID=0, Status='live', Symbol='RNA5SP270', Aliases='RN5S270', description='RNA, 5S ribosomal pseudogene 270', other_designations='RNA, 5S ribosomal 270', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69929905, end_position_on_the_genomic_accession=69930022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373662, CurrentID=0, Status='live', Symbol='LOC105373662', Aliases='', description='uncharacterized LOC105373662', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144877864, end_position_on_the_genomic_accession=144882011, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271254, CurrentID=0, Status='live', Symbol='RPS3AP30', Aliases='RPS3A_11_871', description='ribosomal protein S3a pseudogene 30', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7122848, end_position_on_the_genomic_accession=7123696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873524, CurrentID=0, Status='live', Symbol='RNA5SP271', Aliases='RN5S271', description='RNA, 5S ribosomal pseudogene 271', other_designations='RNA, 5S ribosomal 271', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72357728, end_position_on_the_genomic_accession=72357815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130783, CurrentID=0, Status='live', Symbol='LOC100130783', Aliases='', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 2 pseudogene', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130285592, end_position_on_the_genomic_accession=130285912, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873525, CurrentID=0, Status='live', Symbol='RNA5SP272', Aliases='RN5S272', description='RNA, 5S ribosomal pseudogene 272', other_designations='RNA, 5S ribosomal 272', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88396437, end_position_on_the_genomic_accession=88396547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873526, CurrentID=0, Status='live', Symbol='RNA5SP273', Aliases='RN5S273', description='RNA, 5S ribosomal pseudogene 273', other_designations='RNA, 5S ribosomal 273', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90511128, end_position_on_the_genomic_accession=90511240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373665, CurrentID=0, Status='live', Symbol='LOC105373665', Aliases='', description='uncharacterized LOC105373665', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145856245, end_position_on_the_genomic_accession=145869832, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100873527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873527, CurrentID=0, Status='live', Symbol='RNA5SP274', Aliases='RN5S274', description='RNA, 5S ribosomal pseudogene 274', other_designations='RNA, 5S ribosomal 274', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93441277, end_position_on_the_genomic_accession=93441384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373666, CurrentID=0, Status='live', Symbol='LOC105373666', Aliases='', description='uncharacterized LOC105373666', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145872226, end_position_on_the_genomic_accession=145881611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79160, CurrentID=0, Status='live', Symbol='LOC79160', Aliases='', description='uncharacterized LOC79160', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58634772, end_position_on_the_genomic_accession=58635738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873529, CurrentID=0, Status='live', Symbol='RNA5SP276', Aliases='RN5S276', description='RNA, 5S ribosomal pseudogene 276', other_designations='RNA, 5S ribosomal 276', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116022686, end_position_on_the_genomic_accession=116022779, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343901, CurrentID=0, Status='live', Symbol='MRPS18CP6', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene 6', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44483442, end_position_on_the_genomic_accession=44483723, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652925, CurrentID=0, Status='live', Symbol='LOC100652925', Aliases='', description='uncharacterized LOC100652925', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=31497862, end_position_on_the_genomic_accession=31499885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873533, CurrentID=0, Status='live', Symbol='RNA5SP280', Aliases='RN5S280', description='RNA, 5S ribosomal pseudogene 280', other_designations='RNA, 5S ribosomal 280', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27198287, end_position_on_the_genomic_accession=27198401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647144, CurrentID=0, Status='live', Symbol='ABHD17AP8', Aliases='FAM108A9P', description='abhydrolase domain containing 17A pseudogene 8', other_designations='family with sequence similarity 108, member A9, pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32199889, end_position_on_the_genomic_accession=32202822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376837, CurrentID=0, Status='live', Symbol='LOC105376837', Aliases='', description='uncharacterized LOC105376837', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41965519, end_position_on_the_genomic_accession=41966617, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873534, CurrentID=0, Status='live', Symbol='RNA5SP281', Aliases='RN5S281', description='RNA, 5S ribosomal pseudogene 281', other_designations='RNA, 5S ribosomal 281', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32293558, end_position_on_the_genomic_accession=32293692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373673, CurrentID=0, Status='live', Symbol='LOC105373673', Aliases='', description='uncharacterized LOC105373673', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148021851, end_position_on_the_genomic_accession=148040045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370972, CurrentID=0, Status='live', Symbol='LOC105370972', Aliases='', description='uncharacterized LOC105370972', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91238995, end_position_on_the_genomic_accession=91254887, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421606, CurrentID=0, Status='live', Symbol='LOC100421606', Aliases='', description='family with sequence similarity 133, member A pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227598430, end_position_on_the_genomic_accession=227599133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81786, CurrentID=0, Status='live', Symbol='TRIM7', Aliases='GNIP, RNF90', description='tripartite motif containing 7', other_designations='RING finger protein 90|glycogenin-interacting protein|tripartite motif protein TRIM7', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181193924, end_position_on_the_genomic_accession=181205293, orientation='minus', exon_count=8, OMIM=609315, no_hdr0=''),
  100873535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873535, CurrentID=0, Status='live', Symbol='RNA5SP282', Aliases='RN5S282', description='RNA, 5S ribosomal pseudogene 282', other_designations='RNA, 5S ribosomal 282', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34206134, end_position_on_the_genomic_accession=34206241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81788, CurrentID=0, Status='live', Symbol='NUAK2', Aliases='SNARK', description='NUAK family, SNF1-like kinase, 2', other_designations='SNF1/AMP activated protein kinase|SNF1/AMP kinase-related kinase|omphalocele kinase 2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205302059, end_position_on_the_genomic_accession=205321791, orientation='minus', exon_count=8, OMIM=608131, no_hdr0=''),
  81789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81789, CurrentID=0, Status='live', Symbol='TIGD6', Aliases='', description='tigger transposable element derived 6', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149993123, end_position_on_the_genomic_accession=150001167, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81790, CurrentID=0, Status='live', Symbol='RNF170', Aliases='ADSA, SNAX1', description='ring finger protein 170', other_designations='putative LAG1-interacting protein', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42849637, end_position_on_the_genomic_accession=42897931, orientation='minus', exon_count=11, OMIM=614649, no_hdr0=''),
  81792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81792, CurrentID=0, Status='live', Symbol='ADAMTS12', Aliases='PRO4389', description='ADAM metallopeptidase with thrombospondin type 1 motif 12', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 12|ADAM-TS 12|ADAM-TS12|ADAMTS-12|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 12', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33523535, end_position_on_the_genomic_accession=33892194, orientation='minus', exon_count=28, OMIM=606184, no_hdr0=''),
  81793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81793, CurrentID=0, Status='live', Symbol='TLR10', Aliases='CD290', description='toll-like receptor 10', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38772238, end_position_on_the_genomic_accession=38782990, orientation='minus', exon_count=5, OMIM=606270, no_hdr0=''),
  81794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81794, CurrentID=0, Status='live', Symbol='ADAMTS10', Aliases='ADAM-TS10, ADAMTS-10, WMS, WMS1', description='ADAM metallopeptidase with thrombospondin type 1 motif 10', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 10|ADAM-TS 10|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 10|zinc metalloendopeptidase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8580240, end_position_on_the_genomic_accession=8610753, orientation='minus', exon_count=26, OMIM=608990, no_hdr0=''),
  105373675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373675, CurrentID=0, Status='live', Symbol='LOC105373675', Aliases='', description='uncharacterized LOC105373675', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148520106, end_position_on_the_genomic_accession=148525052, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  81796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81796, CurrentID=0, Status='live', Symbol='SLCO5A1', Aliases='OATP-J, OATP-RP4, OATP5A1, OATPJ, OATPRP4, SLC21A15', description='solute carrier organic anion transporter family member 5A1', other_designations='organic anion transporter polypeptide-related protein 4|solute carrier family 21 (organic anion transporter), member 15|solute carrier family 21 member 15|solute carrier organic anion transporter family, member 5A1', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69667046, end_position_on_the_genomic_accession=69835064, orientation='minus', exon_count=10, OMIM=613543, no_hdr0=''),
  81797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81797, CurrentID=0, Status='live', Symbol='OR12D3', Aliases='hs6M1-27', description='olfactory receptor family 12 subfamily D member 3', other_designations='olfactory receptor OR6-27|olfactory receptor, family 12, subfamily D, member 3', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29373423, end_position_on_the_genomic_accession=29375291, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873537, CurrentID=0, Status='live', Symbol='RNA5SP284', Aliases='RN5S284', description='RNA, 5S ribosomal pseudogene 284', other_designations='RNA, 5S ribosomal 284', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63813190, end_position_on_the_genomic_accession=63813307, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873538, CurrentID=0, Status='live', Symbol='RNA5SP285', Aliases='RN5S285', description='RNA, 5S ribosomal pseudogene 285', other_designations='RNA, 5S ribosomal 285', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72299720, end_position_on_the_genomic_accession=72299838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373677, CurrentID=0, Status='live', Symbol='LOC105373677', Aliases='', description='uncharacterized LOC105373677', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149174738, end_position_on_the_genomic_accession=149184798, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  341315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341315, CurrentID=0, Status='live', Symbol='PABPC1P4', Aliases='PABPCP5, PABPL2', description='poly(A) binding protein, cytoplasmic 1 pseudogene 4', other_designations='poly(A) binding protein, cytoplasmic pseudogene 5|poly(A) binding protein-like 2', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63821546, end_position_on_the_genomic_accession=63824353, orientation='plus', exon_count=0, OMIM=604681, no_hdr0=''),
  105373678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373678, CurrentID=0, Status='live', Symbol='LOC105373678', Aliases='', description='uncharacterized LOC105373678', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149170628, end_position_on_the_genomic_accession=149174748, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873540, CurrentID=0, Status='live', Symbol='RNA5SP287', Aliases='RN5S287', description='RNA, 5S ribosomal pseudogene 287', other_designations='RNA, 5S ribosomal 287', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81885496, end_position_on_the_genomic_accession=81885598, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130799, CurrentID=0, Status='live', Symbol='SRSF10P1', Aliases='FUSIP1P, HsT2687, SFRS13AP1', description='serine/arginine-rich splicing factor 10 pseudogene 1', other_designations='FUS interacting protein (serine-arginine rich) 1 pseudogene|splicing factor, arginine/serine-rich 13A 1 pseudogene', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51158202, end_position_on_the_genomic_accession=51159443, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873541, CurrentID=0, Status='live', Symbol='RNA5SP288', Aliases='RN5S288', description='RNA, 5S ribosomal pseudogene 288', other_designations='RNA, 5S ribosomal 288', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95142449, end_position_on_the_genomic_accession=95142541, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873542, CurrentID=0, Status='live', Symbol='RNA5SP289', Aliases='RN5S289', description='RNA, 5S ribosomal pseudogene 289', other_designations='RNA, 5S ribosomal 289', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95905005, end_position_on_the_genomic_accession=95905141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81831, CurrentID=0, Status='live', Symbol='NETO2', Aliases='BTCL2, NEOT2', description='neuropilin and tolloid like 2', other_designations='brain-specific transmembrane protein containing 2 CUB and 1 LDL-receptor class A domains protein 2|neuropilin (NRP) and tolloid (TLL)-like 2', map_location='16q11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47081520, end_position_on_the_genomic_accession=47144025, orientation='minus', exon_count=10, OMIM=607974, no_hdr0=''),
  81832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81832, CurrentID=0, Status='live', Symbol='NETO1', Aliases='BCTL1, BTCL1', description='neuropilin and tolloid like 1', other_designations='brain-specific transmembrane protein containing 2 CUB and 1 LDL-receptor class A domains protein 1|neuropilin (NRP) and tolloid (TLL)-like 1', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72742314, end_position_on_the_genomic_accession=72867575, orientation='minus', exon_count=18, OMIM=607973, no_hdr0=''),
  81833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81833, CurrentID=0, Status='live', Symbol='SPACA1', Aliases='SAMP32', description='sperm acrosome associated 1', other_designations='sperm acrosomal membrane-associated protein 32', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88047035, end_position_on_the_genomic_accession=88066832, orientation='plus', exon_count=8, OMIM=612739, no_hdr0=''),
  100873543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873543, CurrentID=0, Status='live', Symbol='RNA5SP290', Aliases='RN5S290', description='RNA, 5S ribosomal pseudogene 290', other_designations='RNA, 5S ribosomal 290', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99145840, end_position_on_the_genomic_accession=99145966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  343981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343981, CurrentID=0, Status='live', Symbol='GLULP6', Aliases='', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 6', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194129355, end_position_on_the_genomic_accession=194130375, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81839, CurrentID=0, Status='live', Symbol='VANGL1', Aliases='KITENIN, LPP2, STB2, STBM2', description='VANGL planar cell polarity protein 1', other_designations='KAI1 C-terminal interacting tetraspanin|loop-tail protein 2 homolog|strabismus 2|van Gogh-like protein 1|vang-like 1 (van gogh, Drosophila)', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115641953, end_position_on_the_genomic_accession=115698224, orientation='plus', exon_count=9, OMIM=610132, no_hdr0=''),
  100873544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873544, CurrentID=0, Status='live', Symbol='RNA5SP291', Aliases='RN5S291', description='RNA, 5S ribosomal pseudogene 291', other_designations='RNA, 5S ribosomal 291', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103689473, end_position_on_the_genomic_accession=103689577, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130803, CurrentID=0, Status='live', Symbol='UBL5P1', Aliases='', description='ubiquitin like 5 pseudogene 1', other_designations='ubiquitin-like 5 pseudogene 1', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29600666, end_position_on_the_genomic_accession=29600855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81844, CurrentID=0, Status='live', Symbol='TRIM56', Aliases='RNF109', description='tripartite motif containing 56', other_designations='RING finger protein 109|tripartite motif-containing protein 56', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101085424, end_position_on_the_genomic_accession=101091738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  343990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343990, CurrentID=0, Status='live', Symbol='KIAA1211L', Aliases='C2orf55', description='KIAA1211-like', other_designations='uncharacterized protein C2orf55', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98793846, end_position_on_the_genomic_accession=98936221, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  81847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81847, CurrentID=0, Status='live', Symbol='RNF146', Aliases='', description='ring finger protein 146', other_designations='dactylidin|iduna', map_location='6q22.1-q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127266340, end_position_on_the_genomic_accession=127288562, orientation='plus', exon_count=7, OMIM=612137, no_hdr0=''),
  81848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81848, CurrentID=0, Status='live', Symbol='SPRY4', Aliases='HH17', description='sprouty RTK signaling antagonist 4', other_designations='sprouty homolog 4', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142310427, end_position_on_the_genomic_accession=142325055, orientation='minus', exon_count=5, OMIM=607984, no_hdr0=''),
  81849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81849, CurrentID=0, Status='live', Symbol='ST6GALNAC5', Aliases='SIAT7E, ST6GalNAcV', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 5', other_designations='GD1 alpha synthase|GalNAc alpha-2,6-sialyltransferase V|SIAT7-E|ST6 neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 5|ST6GalNAc V|alpha-N-acetylgalactosaminide alpha-2,6-sialyltransferase V|alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase E|sialyltransferase 7 ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) E|sialyltransferase 7 (alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) E|sialyltransferase 7E', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76867441, end_position_on_the_genomic_accession=77069167, orientation='plus', exon_count=5, OMIM=610134, no_hdr0=''),
  81850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81850, CurrentID=0, Status='live', Symbol='KRTAP1-3', Aliases='KAP1.2, KAP1.3, KAP1.6, KAP1.8A, KAP1.8B, KAP1.9, KRTAP1.3', description='keratin associated protein 1-3', other_designations='keratin associated protein 1.6|keratin-associated protein 1.8|keratin-associated protein 1.9', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41033884, end_position_on_the_genomic_accession=41034855, orientation='minus', exon_count=1, OMIM=608820, no_hdr0=''),
  81851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81851, CurrentID=0, Status='live', Symbol='KRTAP1-1', Aliases='HB2A, KAP1.1, KAP1.1A, KAP1.1B, KAP1.6, KAP1.7, KRTAP1.1, KRTAP1A, hKAP1.7', description='keratin associated protein 1-1', other_designations='high sulfur keratin-associated protein 1.1|keratin associated protein 1.7|keratin associated protein 1A|keratin-associated protein 1.1|keratin-associated protein 1.6|keratin-associated protein 1.7', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41040559, end_position_on_the_genomic_accession=41041461, orientation='minus', exon_count=1, OMIM=608819, no_hdr0=''),
  81853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81853, CurrentID=0, Status='live', Symbol='TMEM14B', Aliases='', description='transmembrane protein 14B', other_designations='', map_location='6p25.1-p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10747759, end_position_on_the_genomic_accession=10759774, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  81854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81854, CurrentID=0, Status='live', Symbol='ZNF205-AS1', Aliases='', description='ZNF205 antisense RNA 1', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3110460, end_position_on_the_genomic_accession=3115598, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81855, CurrentID=0, Status='live', Symbol='SFXN3', Aliases='BA108L7.2, SFX3', description='sideroflexin 3', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101031229, end_position_on_the_genomic_accession=101041241, orientation='plus', exon_count=12, OMIM=615571, no_hdr0=''),
  81856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81856, CurrentID=0, Status='live', Symbol='ZNF611', Aliases='', description='zinc finger protein 611', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52702813, end_position_on_the_genomic_accession=52735054, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  81857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81857, CurrentID=0, Status='live', Symbol='MED25', Aliases='ACID1, ARC92, BVSYS, CMT2B2, P78, PTOV2, TCBAP0758', description='mediator complex subunit 25', other_designations='ARC/mediator transcriptional coactivator subunit|activator interaction domain-containing protein 1|activator-recruited cofactor 92 kDa component|mediator of RNA polymerase II transcription, subunit 25 homolog|prostate-derived protein 78 kDa', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49818279, end_position_on_the_genomic_accession=49840384, orientation='plus', exon_count=19, OMIM=610197, no_hdr0=''),
  81858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81858, CurrentID=0, Status='live', Symbol='SHARPIN', Aliases='SIPL1', description='SHANK associated RH domain interactor', other_designations='SHANK-associated RH domain interactor|hSIPL1|shank-associated RH domain-interacting protein|shank-interacting protein-like 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144098633, end_position_on_the_genomic_accession=144104237, orientation='minus', exon_count=9, OMIM=611885, no_hdr0=''),
  100873547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873547, CurrentID=0, Status='live', Symbol='RNA5SP294', Aliases='RN5S294', description='RNA, 5S ribosomal pseudogene 294', other_designations='RNA, 5S ribosomal 294', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111411644, end_position_on_the_genomic_accession=111411742, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81861, CurrentID=0, Status='live', Symbol='LPRS', Aliases='LPRS1', description='Leprosy, susceptibility to 1', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609888, no_hdr0=''),
  81863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81863, CurrentID=0, Status='live', Symbol='CTPL1', Aliases='CAAR', description='cataract, pulverulent (autosomal recessive, early-onset)', other_designations='', map_location='9q13-q22', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605749, no_hdr0=''),
  81864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81864, CurrentID=0, Status='live', Symbol='EVR3', Aliases='', description='exudative vitreoretinopathy 3', other_designations='', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605750, no_hdr0=''),
  81865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81865, CurrentID=0, Status='live', Symbol='BFIC2', Aliases='', description='Benign familial infantile convulsions-2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605751, no_hdr0=''),
  81866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81866, CurrentID=0, Status='live', Symbol='NDIC', Aliases='', description='Nail dysplasia, isolated congenital', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605779, no_hdr0=''),
  105373687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373687, CurrentID=0, Status='live', Symbol='LOC105373687', Aliases='', description='uncharacterized LOC105373687', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151298348, end_position_on_the_genomic_accession=151303452, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81870, CurrentID=0, Status='live', Symbol='KRTAP9-9', Aliases='KAP9.5, KAP9.9, KRTAP9-5, KRTAP9.9', description='keratin associated protein 9-9', other_designations='keratin-associated protein 9-5|keratin-associated protein 9.5|ultrahigh sulfur keratin-associated protein 9.9', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41255384, end_position_on_the_genomic_accession=41256364, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81871, CurrentID=0, Status='live', Symbol='KRTAP4-6', Aliases='KAP4.15, KAP4.6, KRTAP4-15, KRTAP4.15', description='keratin associated protein 4-6', other_designations='keratin associated protein 4-15|keratin-associated protein 4.6|ultrahigh sulfur keratin-associated protein 4.15', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41139433, end_position_on_the_genomic_accession=41140487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81872, CurrentID=0, Status='live', Symbol='KRTAP2-1', Aliases='KAP2.1A, KRTAP2.1A', description='keratin associated protein 2-1', other_designations='high sulfur keratin-associated protein 2.1|keratin associated protein KRTAP2.1A|keratin-associated protein 2.1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41046541, end_position_on_the_genomic_accession=41047316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81873, CurrentID=0, Status='live', Symbol='ARPC5L', Aliases='ARC16-2', description='actin related protein 2/3 complex subunit 5-like', other_designations='arp2/3 complex 16 kDa subunit 2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124862135, end_position_on_the_genomic_accession=124877733, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  344018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344018, CurrentID=0, Status='live', Symbol='FIGLA', Aliases='BHLHC8, FIGALPHA, POF6', description='folliculogenesis specific bHLH transcription factor', other_designations='class C basic helix-loop-helix protein 8|folliculogenesis specific basic helix-loop-helix|folliculogenesis-specific basic helix-loop-helix protein|transcription factor FIGa', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70777310, end_position_on_the_genomic_accession=70790643, orientation='minus', exon_count=5, OMIM=608697, no_hdr0=''),
  81875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81875, CurrentID=0, Status='live', Symbol='ISG20L2', Aliases='HSD38', description='interferon stimulated exonuclease gene 20kDa like 2', other_designations='interferon stimulated exonuclease gene 20kDa-like 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156721891, end_position_on_the_genomic_accession=156728472, orientation='minus', exon_count=5, OMIM=611930, no_hdr0=''),
  81876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81876, CurrentID=0, Status='live', Symbol='RAB1B', Aliases='', description='RAB1B, member RAS oncogene family', other_designations='small GTP-binding protein', map_location='11q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66268585, end_position_on_the_genomic_accession=66277492, orientation='plus', exon_count=6, OMIM=612565, no_hdr0=''),
  344022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344022, CurrentID=0, Status='live', Symbol='NOTO', Aliases='', description='notochord homeobox', other_designations='notochord homolog', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73202258, end_position_on_the_genomic_accession=73211212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130540, CurrentID=0, Status='live', Symbol='LOC100130540', Aliases='', description='uncharacterized LOC100130540', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=130185059, end_position_on_the_genomic_accession=130185829, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100873551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873551, CurrentID=0, Status='live', Symbol='RNA5SP500', Aliases='RN5S500', description='RNA, 5S ribosomal pseudogene 500', other_designations='RNA, 5S ribosomal 500', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=28982852, end_position_on_the_genomic_accession=28982970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373690, CurrentID=0, Status='live', Symbol='LOC105373690', Aliases='', description='uncharacterized LOC105373690', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152634933, end_position_on_the_genomic_accession=152646102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81887, CurrentID=0, Status='live', Symbol='LAS1L', Aliases='Las1-like, dJ475B7.2', description='LAS1-like, ribosome biogenesis factor', other_designations='protein LAS1 homolog', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65512582, end_position_on_the_genomic_accession=65534806, orientation='minus', exon_count=16, OMIM=300964, no_hdr0=''),
  81888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81888, CurrentID=0, Status='live', Symbol='HYI', Aliases='HT036', description='hydroxypyruvate isomerase (putative)', other_designations='endothelial cell apoptosis protein E-CE1|hydroxypyruvate isomerase homolog', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43450607, end_position_on_the_genomic_accession=43455882, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  81889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81889, CurrentID=0, Status='live', Symbol='FAHD1', Aliases='C16orf36, YISKL', description='fumarylacetoacetate hydrolase domain containing 1', other_designations='FAH domain-containing protein 1|OAA decarboxylase|YISK like/RJD15|acylpyruvate hydrolase|fumarylacetoacetate hydrolase domain-containing protein 1|oxaloacetate decarboxylase|yisK-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1827224, end_position_on_the_genomic_accession=1840207, orientation='plus', exon_count=3, OMIM=616320, no_hdr0=''),
  81890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81890, CurrentID=0, Status='live', Symbol='QTRT1', Aliases='FP3235, TGT, TGUT', description='queuine tRNA-ribosyltransferase 1', other_designations='TGT, 43-KD subunit|TGT, catalytic subunit|guanine insertion enzyme|tRNA-guanine transglycosylase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10701436, end_position_on_the_genomic_accession=10713367, orientation='plus', exon_count=10, OMIM=609615, no_hdr0=''),
  81892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81892, CurrentID=0, Status='live', Symbol='SLIRP', Aliases='C14orf156, DC50, PD04872', description='SRA stem-loop interacting RNA binding protein', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77708071, end_position_on_the_genomic_accession=77717598, orientation='plus', exon_count=5, OMIM=610211, no_hdr0=''),
  81893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81893, CurrentID=0, Status='live', Symbol='SLC7A5P1', Aliases='DC49, LAT1-3TM, MLAS, hLAT1-3TM', description='solute carrier family 7 (amino acid transporter light chain, L system), member 5 pseudogene 1', other_designations='LAT1-3TM protein 1|SLC7A5 pseudogene|solute carrier family 7 (cationic amino acid transporter, y+ system), member 5 pseudogene 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29613103, end_position_on_the_genomic_accession=29613717, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81894, CurrentID=0, Status='live', Symbol='SLC25A28', Aliases='MFRN2, MRS3/4, MRS4L, NPD016', description='solute carrier family 25 (mitochondrial iron transporter), member 28', other_designations='hMRS3/4|mitochondrial RNA splicing protein 3/4|mitochondrial RNA-splicing protein 3/4 homolog|mitochondrial iron transporter 2|mitoferrin 2|putative mitochondrial solute carrier|solute carrier family 25, member 28', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99610518, end_position_on_the_genomic_accession=99659339, orientation='minus', exon_count=12, OMIM=609767, no_hdr0=''),
  105373692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373692, CurrentID=0, Status='live', Symbol='LOC105373692', Aliases='', description='uncharacterized LOC105373692', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154469634, end_position_on_the_genomic_accession=154471577, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873555, CurrentID=0, Status='live', Symbol='RNA5SP504', Aliases='RN5S504', description='RNA, 5S ribosomal pseudogene 504', other_designations='RNA, 5S ribosomal 504', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52665231, end_position_on_the_genomic_accession=52665347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373694, CurrentID=0, Status='live', Symbol='LOC105373694', Aliases='', description='uncharacterized LOC105373694', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154471596, end_position_on_the_genomic_accession=154488929, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873556, CurrentID=0, Status='live', Symbol='RNA5SP505', Aliases='RN5S505', description='RNA, 5S ribosomal pseudogene 505', other_designations='RNA, 5S ribosomal 505', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373695, CurrentID=0, Status='live', Symbol='LOC105373695', Aliases='', description='uncharacterized LOC105373695', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154935462, end_position_on_the_genomic_accession=154945130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100652930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652930, CurrentID=0, Status='live', Symbol='LOC100652930', Aliases='', description='uncharacterized LOC100652930', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58717364, end_position_on_the_genomic_accession=58719231, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873557, CurrentID=0, Status='live', Symbol='RNA5SP506', Aliases='RN5S506', description='RNA, 5S ribosomal pseudogene 506', other_designations='RNA, 5S ribosomal 506', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69672479, end_position_on_the_genomic_accession=69672597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344065, CurrentID=0, Status='live', Symbol='LOC344065', Aliases='', description='zinc finger protein 184 pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95214332, end_position_on_the_genomic_accession=95215609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873558, CurrentID=0, Status='live', Symbol='RNA5SP507', Aliases='RN5S507', description='RNA, 5S ribosomal pseudogene 507', other_designations='RNA, 5S ribosomal 507', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70253042, end_position_on_the_genomic_accession=70253162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81926, CurrentID=0, Status='live', Symbol='ABHD17A', Aliases='C19orf27, FAM108A1', description='abhydrolase domain containing 17A', other_designations='abhydrolase domain-containing protein FAM108A1|family with sequence similarity 108, member A1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1876976, end_position_on_the_genomic_accession=1885519, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105373697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373697, CurrentID=0, Status='live', Symbol='LOC105373697', Aliases='', description='uncharacterized LOC105373697', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155075485, end_position_on_the_genomic_accession=155225935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81928, CurrentID=0, Status='live', Symbol='CABLES2', Aliases='C20orf150, dJ908M14.2, ik3-2', description='Cdk5 and Abl enzyme substrate 2', other_designations='interactor with CDK3 2', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62388630, end_position_on_the_genomic_accession=62407283, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  81929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81929, CurrentID=0, Status='live', Symbol='SEH1L', Aliases='SEC13L, SEH1A, SEH1B, Seh1', description='SEH1 like nucleoporin', other_designations='SEH1-like nucleoporin|nup107-160 subcomplex subunit SEH1|sec13 like protein|sec13-like protein', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12947984, end_position_on_the_genomic_accession=12987537, orientation='plus', exon_count=12, OMIM=609263, no_hdr0=''),
  81930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81930, CurrentID=0, Status='live', Symbol='KIF18A', Aliases='MS-KIF18A, PPP1R99', description='kinesin family member 18A', other_designations='protein phosphatase 1, regulatory subunit 99', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28020616, end_position_on_the_genomic_accession=28108199, orientation='minus', exon_count=17, OMIM=611271, no_hdr0=''),
  81931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81931, CurrentID=0, Status='live', Symbol='ZNF93', Aliases='HPF34, HTF34, TF34, ZNF505', description='zinc finger protein 93', other_designations='zinc finger protein 505|zinc finger protein HTF34', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19900913, end_position_on_the_genomic_accession=19935998, orientation='plus', exon_count=5, OMIM=603975, no_hdr0=''),
  81932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81932, CurrentID=0, Status='live', Symbol='HDHD3', Aliases='2810435D12Rik, C9orf158', description='haloacid dehalogenase like hydrolase domain containing 3', other_designations='haloacid dehalogenase-like hydrolase domain containing 3', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113373416, end_position_on_the_genomic_accession=113377009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421611, CurrentID=0, Status='live', Symbol='LOC100421611', Aliases='', description='family with sequence similarity 71, member B pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=84998278, end_position_on_the_genomic_accession=84999293, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873560, CurrentID=0, Status='live', Symbol='RNA5SP509', Aliases='RN5S509', description='RNA, 5S ribosomal pseudogene 509', other_designations='RNA, 5S ribosomal 509', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77066709, end_position_on_the_genomic_accession=77066827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373700, CurrentID=0, Status='live', Symbol='LOC105373700', Aliases='', description='uncharacterized LOC105373700', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155473246, end_position_on_the_genomic_accession=155489338, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373701, CurrentID=0, Status='live', Symbol='LOC105373701', Aliases='', description='uncharacterized LOC105373701', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155759646, end_position_on_the_genomic_accession=155814139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373702, CurrentID=0, Status='live', Symbol='LOC105373702', Aliases='', description='uncharacterized LOC105373702', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155838364, end_position_on_the_genomic_accession=155840975, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647175, CurrentID=0, Status='live', Symbol='GSTA11P', Aliases='GSTAP3', description='glutathione S-transferase alpha 11, pseudogene', other_designations='glutathione S-transferase alpha 1 pseudogene|glutathione S-transferase alpha 3 pseudogene', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52847744, end_position_on_the_genomic_accession=52870358, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373704, CurrentID=0, Status='live', Symbol='LOC105373704', Aliases='', description='uncharacterized LOC105373704', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156160950, end_position_on_the_genomic_accession=156318728, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373705, CurrentID=0, Status='live', Symbol='LOC105373705', Aliases='', description='uncharacterized LOC105373705', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155979693, end_position_on_the_genomic_accession=155981040, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375343, CurrentID=0, Status='live', Symbol='LOC105375343', Aliases='', description='uncharacterized LOC105375343', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69115870, end_position_on_the_genomic_accession=69150203, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373706, CurrentID=0, Status='live', Symbol='LOC105373706', Aliases='', description='uncharacterized LOC105373706', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156407650, end_position_on_the_genomic_accession=156433598, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373707, CurrentID=0, Status='live', Symbol='LOC105373707', Aliases='', description='uncharacterized LOC105373707', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156610724, end_position_on_the_genomic_accession=156613737, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81993, CurrentID=0, Status='live', Symbol='COX17P1', Aliases='COX17P', description='COX17, cytochrome c oxidase copper chaperone pseudogene 1', other_designations='COX17 cytochrome c oxidase assembly homolog pseudogene 1|COX17 cytochrome c oxidase copper chaperone pseudogene 1|COX17 pseudogene|cytochrome c oxidase assembly homolog 17 pseudogene 1', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46490857, end_position_on_the_genomic_accession=46491259, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873570, CurrentID=0, Status='live', Symbol='RNA5SP519', Aliases='RN5S519', description='RNA, 5S ribosomal pseudogene 519', other_designations='RNA, 5S ribosomal 519', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10092875, end_position_on_the_genomic_accession=10092993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373709, CurrentID=0, Status='live', Symbol='LOC105373709', Aliases='', description='uncharacterized LOC105373709', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157031461, end_position_on_the_genomic_accession=157049719, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  344148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344148, CurrentID=0, Status='live', Symbol='NCKAP5', Aliases='ERIH1, ERIH2, NAP5', description='NCK-associated protein 5', other_designations='NAP-5|Nck associated protein 5|peripheral clock protein 2', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132671788, end_position_on_the_genomic_accession=133674923, orientation='minus', exon_count=30, OMIM=608789, no_hdr0=''),
  606293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606293, CurrentID=0, Status='live', Symbol='KLKP1', Aliases='KARMA, KLK31P, KRSP1, PsiKLK1, YKLK1', description='kallikrein pseudogene 1', other_designations='kallikrein 31 pseudogene|kallikrein-related protein|kallikrein-related sequence from prostate', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50882096, end_position_on_the_genomic_accession=50896398, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375344, CurrentID=0, Status='live', Symbol='LOC105375344', Aliases='', description='uncharacterized LOC105375344', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69162421, end_position_on_the_genomic_accession=69172832, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271256, CurrentID=0, Status='live', Symbol='RPS3AP31', Aliases='RPS3A_12_873', description='ribosomal protein S3a pseudogene 31', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8095922, end_position_on_the_genomic_accession=8096770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  144742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144742, CurrentID=0, Status='live', Symbol='LINC00934', Aliases='', description='long intergenic non-protein coding RNA 934', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119283825, end_position_on_the_genomic_accession=119303380, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  344167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344167, CurrentID=0, Status='live', Symbol='FOXI3', Aliases='', description='forkhead box I3', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88448207, end_position_on_the_genomic_accession=88452535, orientation='minus', exon_count=2, OMIM=612351, no_hdr0=''),
  100421614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421614, CurrentID=0, Status='live', Symbol='LOC100421614', Aliases='', description='ARP2 actin-related protein 2 homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24686923, end_position_on_the_genomic_accession=24688088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373714, CurrentID=0, Status='live', Symbol='LOC105373714', Aliases='', description='uncharacterized LOC105373714', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157876566, end_position_on_the_genomic_accession=157918379, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  344178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344178, CurrentID=0, Status='live', Symbol='RPL5P7', Aliases='RPL5_2_294', description='ribosomal protein L5 pseudogene 7', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173871572, end_position_on_the_genomic_accession=173872567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373715, CurrentID=0, Status='live', Symbol='LOC105373715', Aliases='', description='uncharacterized LOC105373715', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158577399, end_position_on_the_genomic_accession=158589945, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  344191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344191, CurrentID=0, Status='live', Symbol='EVX2', Aliases='EVX-2', description='even-skipped homeobox 2', other_designations='eve, even-skipped homeo box homolog 2|even-skipped homeo box 2 (homolog of Drosophila eve)', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176076021, end_position_on_the_genomic_accession=176085797, orientation='minus', exon_count=5, OMIM=142991, no_hdr0=''),
  100873464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873464, CurrentID=0, Status='live', Symbol='RNA5SP203', Aliases='RN5S203', description='RNA, 5S ribosomal pseudogene 203', other_designations='RNA, 5S ribosomal 203', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10752791, end_position_on_the_genomic_accession=10752926, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373718, CurrentID=0, Status='live', Symbol='LOC105373718', Aliases='', description='uncharacterized LOC105373718', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160555163, end_position_on_the_genomic_accession=160648509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873580, CurrentID=0, Status='live', Symbol='RNA5SP306', Aliases='RN5S306', description='RNA, 5S ribosomal pseudogene 306', other_designations='RNA, 5S ribosomal 306', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25813811, end_position_on_the_genomic_accession=25813917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130841, CurrentID=0, Status='live', Symbol='LOC100130841', Aliases='', description='MDM2 oncogene, E3 ubiquitin protein ligase pseudogene', other_designations='Mdm2 p53 binding protein homolog (mouse) pseudogene', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22188333, end_position_on_the_genomic_accession=22188847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391175, CurrentID=0, Status='live', Symbol='RPL23AP15', Aliases='RPL23A_4_173', description='ribosomal protein L23a pseudogene 15', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228449168, end_position_on_the_genomic_accession=228449621, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873583, CurrentID=0, Status='live', Symbol='RNA5SP309', Aliases='RN5S309', description='RNA, 5S ribosomal pseudogene 309', other_designations='RNA, 5S ribosomal 309', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31259599, end_position_on_the_genomic_accession=31259704, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130842, CurrentID=0, Status='live', Symbol='LOC100130842', Aliases='', description='small integral membrane protein 7 pseudogene', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=35219380, end_position_on_the_genomic_accession=35219604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373723, CurrentID=0, Status='live', Symbol='LOC105373723', Aliases='', description='uncharacterized LOC105373723', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162073485, end_position_on_the_genomic_accession=162077941, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373724, CurrentID=0, Status='live', Symbol='LOC105373724', Aliases='', description='uncharacterized LOC105373724', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162246132, end_position_on_the_genomic_accession=162250062, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375347, CurrentID=0, Status='live', Symbol='LOC105375347', Aliases='', description='uncharacterized LOC105375347', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=70744014, end_position_on_the_genomic_accession=70746136, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373726, CurrentID=0, Status='live', Symbol='LOC105373726', Aliases='', description='uncharacterized LOC105373726', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162490351, end_position_on_the_genomic_accession=162495479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100885777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885777, CurrentID=0, Status='live', Symbol='PCCA-AS1', Aliases='', description='PCCA antisense RNA 1', other_designations='PCCA antisense RNA 1 (non-protein coding)', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100464437, end_position_on_the_genomic_accession=100481015, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373728, CurrentID=0, Status='live', Symbol='LOC105373728', Aliases='', description='uncharacterized LOC105373728', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=163450806, end_position_on_the_genomic_accession=163456673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379712, CurrentID=0, Status='live', Symbol='LOC105379712', Aliases='', description='uncharacterized LOC105379712', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421617, CurrentID=0, Status='live', Symbol='LOC100421617', Aliases='', description='family with sequence similarity 133, member A pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25435588, end_position_on_the_genomic_accession=25436297, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130849, CurrentID=0, Status='live', Symbol='LOC100130849', Aliases='', description='phosphorylase kinase, gamma 1 (muscle) pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56875385, end_position_on_the_genomic_accession=56882146, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373730, CurrentID=0, Status='live', Symbol='LOC105373730', Aliases='', description='uncharacterized LOC105373730', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165821977, end_position_on_the_genomic_accession=165848186, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375348, CurrentID=0, Status='live', Symbol='LOC105375348', Aliases='', description='uncharacterized LOC105375348', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=70595123, end_position_on_the_genomic_accession=70596799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372475, CurrentID=0, Status='live', Symbol='LOC105372475', Aliases='', description='uncharacterized LOC105372475', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57074600, end_position_on_the_genomic_accession=57078880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130854, CurrentID=0, Status='live', Symbol='LOC100130854', Aliases='', description='zinc finger protein 562 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9724008, end_position_on_the_genomic_accession=9725086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873596, CurrentID=0, Status='live', Symbol='RNA5SP330', Aliases='RN5S330', description='RNA, 5S ribosomal pseudogene 330', other_designations='RNA, 5S ribosomal 330', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8845263, end_position_on_the_genomic_accession=8845358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373735, CurrentID=0, Status='live', Symbol='LOC105373735', Aliases='', description='uncharacterized LOC105373735', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170242361, end_position_on_the_genomic_accession=170249706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373736, CurrentID=0, Status='live', Symbol='LOC105373736', Aliases='', description='uncharacterized LOC105373736', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170968877, end_position_on_the_genomic_accession=170973258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  144766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144766, CurrentID=0, Status='live', Symbol='LINC00355', Aliases='', description='long intergenic non-protein coding RNA 355', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373737, CurrentID=0, Status='live', Symbol='LOC105373737', Aliases='', description='uncharacterized LOC105373737', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171244626, end_position_on_the_genomic_accession=171301140, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106479573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479573, CurrentID=0, Status='live', Symbol='RNU4-36P', Aliases='', description='RNA, U4 small nuclear 36, pseudogene', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129196, CurrentID=0, Status='live', Symbol='MATN1-AS1', Aliases='', description='MATN1 antisense RNA 1', other_designations='MATN1 antisense RNA 1 (non-protein coding)', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30718772, end_position_on_the_genomic_accession=30726746, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  344320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344320, CurrentID=0, Status='live', Symbol='KRT8P10', Aliases='', description='keratin 8 pseudogene 10', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183070796, end_position_on_the_genomic_accession=183072562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79232, CurrentID=0, Status='live', Symbol='ZNF37CP', Aliases='BA775A3.1, ZNF37C', description='zinc finger protein 37C, pseudogene', other_designations='KRAB box zinc finger protein pseudogene', map_location='10p11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37883620, end_position_on_the_genomic_accession=37884098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647211, CurrentID=0, Status='live', Symbol='LOC647211', Aliases='', description='rhophilin-2-like', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  343169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343169, CurrentID=0, Status='live', Symbol='OR6F1', Aliases='OR1-34, OR1-38, OST731', description='olfactory receptor family 6 subfamily F member 1', other_designations='olfactory receptor OR1-34 pseudogene|olfactory receptor OR1-38|olfactory receptor, family 6, subfamily F, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247711829, end_position_on_the_genomic_accession=247712755, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  344332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344332, CurrentID=0, Status='live', Symbol='FABP5P10', Aliases='FABP5L10', description='fatty acid binding protein 5 pseudogene 10', other_designations='fatty acid binding protein 5-like 10', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151186161, end_position_on_the_genomic_accession=151186793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375350, CurrentID=0, Status='live', Symbol='LOC105375350', Aliases='', description='uncharacterized LOC105375350', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73703002, end_position_on_the_genomic_accession=73705160, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927379, CurrentID=0, Status='live', Symbol='LOC101927379', Aliases='', description='uncharacterized LOC101927379', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122628952, end_position_on_the_genomic_accession=122730560, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100130861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130861, CurrentID=0, Status='live', Symbol='LOC100130861', Aliases='', description='zinc finger protein 736 pseudogene', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46784791, end_position_on_the_genomic_accession=46786046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479574, CurrentID=0, Status='live', Symbol='RNU4-38P', Aliases='', description='RNA, U4 small nuclear 38, pseudogene', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373743, CurrentID=0, Status='live', Symbol='LOC105373743', Aliases='', description='uncharacterized LOC105373743', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173291826, end_position_on_the_genomic_accession=173296884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  606495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606495, CurrentID=0, Status='live', Symbol='CYB5RL', Aliases='', description='cytochrome b5 reductase-like', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54172336, end_position_on_the_genomic_accession=54200073, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  606500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606500, CurrentID=0, Status='live', Symbol='SNORD68', Aliases='HBII-202', description='small nucleolar RNA, C/D box 68', other_designations='HBII-202 small nucleolar RNA|HBII-202 snoRNA', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89561430, end_position_on_the_genomic_accession=89561501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130865, CurrentID=0, Status='live', Symbol='LOC100130865', Aliases='', description='uncharacterized LOC100130865', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=2235350, end_position_on_the_genomic_accession=2235778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375351, CurrentID=0, Status='live', Symbol='LOC105375351', Aliases='', description='uncharacterized LOC105375351', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74763935, end_position_on_the_genomic_accession=74764574, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373746, CurrentID=0, Status='live', Symbol='LOC105373746', Aliases='', description='uncharacterized LOC105373746', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174060753, end_position_on_the_genomic_accession=174062516, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144776, CurrentID=0, Status='live', Symbol='LINC00410', Aliases='', description='long intergenic non-protein coding RNA 410', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90890954, end_position_on_the_genomic_accession=90926597, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  344371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344371, CurrentID=0, Status='live', Symbol='SLC25A5P2', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 2', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33839522, end_position_on_the_genomic_accession=33840742, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379716, CurrentID=0, Status='live', Symbol='LOC105379716', Aliases='', description='uncharacterized LOC105379716', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  606523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606523, CurrentID=0, Status='live', Symbol='DFNB57', Aliases='', description='deafness, autosomal recessive 57', other_designations='', map_location='10q23.1-q26.11', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  344382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344382, CurrentID=0, Status='live', Symbol='LOC344382', Aliases='', description='serine/threonine kinase receptor associated protein pseudogene', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37816725, end_position_on_the_genomic_accession=37818158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344387, CurrentID=0, Status='live', Symbol='CDKL4', Aliases='', description='cyclin-dependent kinase-like 4', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39174937, end_position_on_the_genomic_accession=39246810, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105373751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373751, CurrentID=0, Status='live', Symbol='LOC105373751', Aliases='', description='uncharacterized LOC105373751', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175166467, end_position_on_the_genomic_accession=175226558, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373752, CurrentID=0, Status='live', Symbol='LOC105373752', Aliases='', description='uncharacterized LOC105373752', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175454606, end_position_on_the_genomic_accession=175471257, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  344405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344405, CurrentID=0, Status='live', Symbol='PRORSD1P', Aliases='NCRNA00117, PRDXDD1P, Prdxdd1, Ybakd1', description='prolyl-tRNA synthetase associated domain containing 1, pseudogene', other_designations='PrdX deacylase domain containing 1, pseudogene|YBak domain containing 1|prolyl-tRNA synthetase domain containing 1, pseudogene', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55282319, end_position_on_the_genomic_accession=55284472, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  606551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606551, CurrentID=0, Status='live', Symbol='UBE2MP1', Aliases='', description='ubiquitin conjugating enzyme E2M pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2M pseudogene 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35169431, end_position_on_the_genomic_accession=35170391, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379717, CurrentID=0, Status='live', Symbol='LOC105379717', Aliases='', description='uncharacterized LOC105379717', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  606553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606553, CurrentID=0, Status='live', Symbol='C8orf49', Aliases='G4DM', description='chromosome 8 open reading frame 49', other_designations='GATA4 downstream membrane protein', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11761256, end_position_on_the_genomic_accession=11763223, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370730, CurrentID=0, Status='live', Symbol='LOC105370730', Aliases='', description='uncharacterized LOC105370730', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23527058, end_position_on_the_genomic_accession=23541404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873616, CurrentID=0, Status='live', Symbol='RNA5SP357', Aliases='RN5S357', description='RNA, 5S ribosomal pseudogene 357', other_designations='RNA, 5S ribosomal 357', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34205699, end_position_on_the_genomic_accession=34205817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344423, CurrentID=0, Status='live', Symbol='RPS12P3', Aliases='RPS12_2_223', description='ribosomal protein S12 pseudogene 3', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60938177, end_position_on_the_genomic_accession=60938685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130876, CurrentID=0, Status='live', Symbol='LOC100130876', Aliases='', description='uncharacterized LOC100130876', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=940349, end_position_on_the_genomic_accession=943585, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373757, CurrentID=0, Status='live', Symbol='LOC105373757', Aliases='', description='uncharacterized LOC105373757', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176830839, end_position_on_the_genomic_accession=176859854, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105373758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373758, CurrentID=0, Status='live', Symbol='LOC105373758', Aliases='', description='uncharacterized LOC105373758', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176892118, end_position_on_the_genomic_accession=176966102, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373759, CurrentID=0, Status='live', Symbol='LOC105373759', Aliases='', description='uncharacterized LOC105373759', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177211394, end_position_on_the_genomic_accession=177226835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373760, CurrentID=0, Status='live', Symbol='LOC105373760', Aliases='', description='uncharacterized LOC105373760', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177064319, end_position_on_the_genomic_accession=177164600, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100421303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421303, CurrentID=0, Status='live', Symbol='LOC100421303', Aliases='', description='DEAH (Asp-Glu-Ala-His) box helicase 9 pseudogene', other_designations='DEAH (Asp-Glu-Ala-His) box polypeptide 9 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85452804, end_position_on_the_genomic_accession=85453771, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375354, CurrentID=0, Status='live', Symbol='LOC105375354', Aliases='', description='postmeiotic segregation increased 2-like protein 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76524383, end_position_on_the_genomic_accession=76544090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  344454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344454, CurrentID=0, Status='live', Symbol='AOX2P', Aliases='AOH2, AOX2', description='aldehyde oxidase 2 pseudogene', other_designations='AOX3/AOX3L1-A|AOX3/AOX3L1-B', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200695723, end_position_on_the_genomic_accession=200794223, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  105373761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373761, CurrentID=0, Status='live', Symbol='LOC105373761', Aliases='', description='uncharacterized LOC105373761', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177032670, end_position_on_the_genomic_accession=177035022, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  344462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344462, CurrentID=0, Status='live', Symbol='KRT18P39', Aliases='', description='keratin 18 pseudogene 39', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203764233, end_position_on_the_genomic_accession=203765632, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271258, CurrentID=0, Status='live', Symbol='RPS3AP32', Aliases='RPS3A_13_917', description='ribosomal protein S3a pseudogene 32', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74055858, end_position_on_the_genomic_accession=74056690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370990, CurrentID=0, Status='live', Symbol='LOC105370990', Aliases='', description='uncharacterized LOC105370990', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95132612, end_position_on_the_genomic_accession=95150549, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  344471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344471, CurrentID=0, Status='live', Symbol='RPL12P17', Aliases='RPL12_2_319', description='ribosomal protein L12 pseudogene 17', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208057202, end_position_on_the_genomic_accession=208057833, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373764, CurrentID=0, Status='live', Symbol='LOC105373764', Aliases='', description='uncharacterized LOC105373764', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177702404, end_position_on_the_genomic_accession=177728138, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873626, CurrentID=0, Status='live', Symbol='RNA5SP367', Aliases='RN5S367', description='RNA, 5S ribosomal pseudogene 367', other_designations='RNA, 5S ribosomal 367', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101472339, end_position_on_the_genomic_accession=101472456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375355, CurrentID=0, Status='live', Symbol='LOC105375355', Aliases='', description='uroplakin-3b', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76510102, end_position_on_the_genomic_accession=76524424, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373766, CurrentID=0, Status='live', Symbol='LOC105373766', Aliases='', description='uncharacterized LOC105373766', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178827303, end_position_on_the_genomic_accession=178922667, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373767, CurrentID=0, Status='live', Symbol='LOC105373767', Aliases='', description='uncharacterized LOC105373767', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180116044, end_position_on_the_genomic_accession=180125194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873629, CurrentID=0, Status='live', Symbol='RNA5SP371', Aliases='RN5S371', description='RNA, 5S ribosomal pseudogene 371', other_designations='RNA, 5S ribosomal 371', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107486511, end_position_on_the_genomic_accession=107486635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122953, CurrentID=0, Status='live', Symbol='JDP2', Aliases='JUNDM2', description='Jun dimerization protein 2', other_designations='progesterone receptor co-activator', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75427806, end_position_on_the_genomic_accession=75472701, orientation='plus', exon_count=8, OMIM=608657, no_hdr0=''),
  105373770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373770, CurrentID=0, Status='live', Symbol='LOC105373770', Aliases='', description='uncharacterized LOC105373770', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180566859, end_position_on_the_genomic_accession=180571819, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375356, CurrentID=0, Status='live', Symbol='LOC105375356', Aliases='', description='uncharacterized LOC105375356', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76650797, end_position_on_the_genomic_accession=76728561, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130891, CurrentID=0, Status='live', Symbol='LOC100130891', Aliases='', description='PDZ and LIM domain 7 (enigma) pseudogene', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28238430, end_position_on_the_genomic_accession=28239000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254028, CurrentID=0, Status='live', Symbol='LOC254028', Aliases='', description='uncharacterized LOC254028', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20588224, end_position_on_the_genomic_accession=20607221, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373774, CurrentID=0, Status='live', Symbol='LOC105373774', Aliases='', description='uncharacterized LOC105373774', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183211206, end_position_on_the_genomic_accession=183215400, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873637, CurrentID=0, Status='live', Symbol='RNA5SP382', Aliases='RN5S382', description='RNA, 5S ribosomal pseudogene 382', other_designations='RNA, 5S ribosomal 382', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20414987, end_position_on_the_genomic_accession=20415098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647249, CurrentID=0, Status='live', Symbol='GAPDHP36', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 36', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180212190, end_position_on_the_genomic_accession=180213213, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373778, CurrentID=0, Status='live', Symbol='LOC105373778', Aliases='', description='uncharacterized LOC105373778', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184431247, end_position_on_the_genomic_accession=184436490, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  344558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344558, CurrentID=0, Status='live', Symbol='SH3RF3', Aliases='POSH2, SH3MD4', description='SH3 domain containing ring finger 3', other_designations='SH3 multiple domains 4|plenty of SH3s 2|plenty of SH3s-2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109129498, end_position_on_the_genomic_accession=109504636, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  344561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344561, CurrentID=0, Status='live', Symbol='GPR148', Aliases='BTR, PGR6', description='G protein-coupled receptor 148', other_designations='G protein-coupled receptor PGR6|G-protein coupled receptor PGR6|brain and testis restricted GPCR', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130728156, end_position_on_the_genomic_accession=130730544, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130899, CurrentID=0, Status='live', Symbol='LOC100130899', Aliases='', description='uncharacterized LOC100130899', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40032332, end_position_on_the_genomic_accession=40036577, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  144809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144809, CurrentID=0, Status='live', Symbol='FAM216B', Aliases='C13orf30', description='family with sequence similarity 216 member B', other_designations='family with sequence similarity 216, member B', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42781550, end_position_on_the_genomic_accession=42791549, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79274, CurrentID=0, Status='live', Symbol='OR52L2P', Aliases='OR52L2', description='olfactory receptor family 52 subfamily L member 2 pseudogene', other_designations='Olfactory receptor 52L2|olfactory receptor OR11-74|olfactory receptor, family 52, subfamily L, member 2 pseudogene|seven transmembrane helix receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6057219, end_position_on_the_genomic_accession=6058330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927387, CurrentID=0, Status='live', Symbol='LOC101927387', Aliases='', description='uncharacterized LOC101927387', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61755710, end_position_on_the_genomic_accession=61756647, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  606719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606719, CurrentID=0, Status='live', Symbol='DFNB66', Aliases='', description='deafness, autosomal recessive 66', other_designations='', map_location='6p22.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610212, no_hdr0=''),
  144811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144811, CurrentID=0, Status='live', Symbol='LACC1', Aliases='C13orf31', description='laccase (multicopper oxidoreductase) domain containing 1', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43879284, end_position_on_the_genomic_accession=43893933, orientation='plus', exon_count=9, OMIM=613409, no_hdr0=''),
  606724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=606724, CurrentID=0, Status='live', Symbol='LOC606724', Aliases='', description='coronin, actin binding protein, 1A pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29449345, end_position_on_the_genomic_accession=29449915, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372477, CurrentID=0, Status='live', Symbol='LOC105372477', Aliases='', description='uncharacterized LOC105372477', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57582897, end_position_on_the_genomic_accession=57584944, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  344593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344593, CurrentID=0, Status='live', Symbol='LOC344593', Aliases='', description='protein tyrosine phosphatase, non-receptor type 11 pseudogene', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107105314, end_position_on_the_genomic_accession=107107199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724965, CurrentID=0, Status='live', Symbol='LOC102724965', Aliases='', description='uncharacterized LOC102724965', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45648733, end_position_on_the_genomic_accession=45650986, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  344595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344595, CurrentID=0, Status='live', Symbol='DUBR', Aliases='DUM, LINC00883', description='DPPA2 upstream binding RNA', other_designations='long intergenic non-protein coding RNA 883', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107240692, end_position_on_the_genomic_accession=107326964, orientation='plus', exon_count=5, OMIM=616619, no_hdr0=''),
  105374303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374303, CurrentID=0, Status='live', Symbol='LOC105374303', Aliases='', description='uncharacterized LOC105374303', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196192785, end_position_on_the_genomic_accession=196193534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375359, CurrentID=0, Status='live', Symbol='LOC105375359', Aliases='', description='uncharacterized LOC105375359', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77221196, end_position_on_the_genomic_accession=77246728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927388, CurrentID=0, Status='live', Symbol='LOC101927388', Aliases='', description='uncharacterized LOC101927388', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=101452675, end_position_on_the_genomic_accession=101454295, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254042, CurrentID=0, Status='live', Symbol='METAP1D', Aliases='MAP 1D, MAP1D, MetAP 1D, Metap1l', description='methionyl aminopeptidase type 1D (mitochondrial)', other_designations='CDS of metAP-3 within PCR fragment|MAP 1D|metAP 1D|methionyl aminopeptidase type 1D, mitochondrial|peptidase M 1D', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171999897, end_position_on_the_genomic_accession=172082430, orientation='plus', exon_count=13, OMIM=610267, no_hdr0=''),
  101928893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928893, CurrentID=0, Status='live', Symbol='LOC101928893', Aliases='', description='uncharacterized LOC101928893', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78716076, end_position_on_the_genomic_accession=78743807, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271259, CurrentID=0, Status='live', Symbol='RPS3AP33', Aliases='RPS3A_14_872', description='ribosomal protein S3a pseudogene 33', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7190739, end_position_on_the_genomic_accession=7191587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  144817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144817, CurrentID=0, Status='live', Symbol='LINC00330', Aliases='NCRNA00330', description='long intergenic non-protein coding RNA 330', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44799504, end_position_on_the_genomic_accession=44809630, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79282, CurrentID=0, Status='live', Symbol='OR7E93P', Aliases='OR7E131P', description='olfactory receptor family 7 subfamily E member 93 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 131 pseudogene|olfactory receptor, family 7, subfamily E, member 93 pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125724434, end_position_on_the_genomic_accession=125725604, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130909, CurrentID=0, Status='live', Symbol='LOC100130909', Aliases='', description='ribonuclease T2 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56175538, end_position_on_the_genomic_accession=56176491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873651, CurrentID=0, Status='live', Symbol='RNA5SP398', Aliases='RN5S398', description='RNA, 5S ribosomal pseudogene 398', other_designations='RNA, 5S ribosomal 398', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69367585, end_position_on_the_genomic_accession=69367681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929766, CurrentID=0, Status='live', Symbol='LOC101929766', Aliases='', description='endogenous retrovirus group V member 2 Env polyprotein-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=189266081, end_position_on_the_genomic_accession=189335533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873559, CurrentID=0, Status='live', Symbol='RNA5SP508', Aliases='RN5S508', description='RNA, 5S ribosomal pseudogene 508', other_designations='RNA, 5S ribosomal 508', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76655009, end_position_on_the_genomic_accession=76655127, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647264, CurrentID=0, Status='live', Symbol='LOC647264', Aliases='', description='keratin-associated protein 6-2', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63746116, end_position_on_the_genomic_accession=63747788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  82500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=82500, CurrentID=0, Status='live', Symbol='PKP2P1', Aliases='', description='plakophilin 2 pseudogene 1', other_designations='plakophilin-2 pseudogene 1', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6477639, end_position_on_the_genomic_accession=6482189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373793, CurrentID=0, Status='live', Symbol='LOC105373793', Aliases='', description='uncharacterized LOC105373793', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189673120, end_position_on_the_genomic_accession=189674826, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873655, CurrentID=0, Status='live', Symbol='RNA5SP402', Aliases='RN5S402', description='RNA, 5S ribosomal pseudogene 402', other_designations='RNA, 5S ribosomal 402', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100074081, end_position_on_the_genomic_accession=100074190, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  344653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344653, CurrentID=0, Status='live', Symbol='ABCF2P1', Aliases='', description='ATP binding cassette subfamily F member 2 pseudogene 1', other_designations='ATP-binding cassette, sub-family F (GCN20), member 2 pseudogene 1', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88317091, end_position_on_the_genomic_accession=88319021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344657, CurrentID=0, Status='live', Symbol='LRRIQ4', Aliases='LRRC64', description='leucine-rich repeats and IQ motif containing 4', other_designations='leucine rich repeat containing 64|leucine-rich repeat-containing protein 64', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169812947, end_position_on_the_genomic_accession=169838007, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  344658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344658, CurrentID=0, Status='live', Symbol='SAMD7', Aliases='', description='sterile alpha motif domain containing 7', other_designations='SAM domain-containing protein 7', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169905106, end_position_on_the_genomic_accession=169939175, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100873657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873657, CurrentID=0, Status='live', Symbol='RNA5SP405', Aliases='RN5S405', description='RNA, 5S ribosomal pseudogene 405', other_designations='RNA, 5S ribosomal 405', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26028912, end_position_on_the_genomic_accession=26029025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373796, CurrentID=0, Status='live', Symbol='LOC105373796', Aliases='', description='uncharacterized LOC105373796', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189852771, end_position_on_the_genomic_accession=189879498, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873658, CurrentID=0, Status='live', Symbol='RNA5SP406', Aliases='RN5S406', description='RNA, 5S ribosomal pseudogene 406', other_designations='RNA, 5S ribosomal 406', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35734393, end_position_on_the_genomic_accession=35734504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873659, CurrentID=0, Status='live', Symbol='RNA5SP407', Aliases='RN5S407', description='RNA, 5S ribosomal pseudogene 407', other_designations='RNA, 5S ribosomal 407', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35734639, end_position_on_the_genomic_accession=35734748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373800, CurrentID=0, Status='live', Symbol='LOC105373800', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190342271, end_position_on_the_genomic_accession=190343960, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79294, CurrentID=0, Status='live', Symbol='OR4C7P', Aliases='', description='olfactory receptor family 4 subfamily C member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 7 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54634929, end_position_on_the_genomic_accession=54636053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927391, CurrentID=0, Status='live', Symbol='LOC101927391', Aliases='', description='uncharacterized LOC101927391', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7550104, end_position_on_the_genomic_accession=7566065, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100885780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885780, CurrentID=0, Status='live', Symbol='VWC2L-IT1', Aliases='', description='VWC2L intronic transcript 1', other_designations='VWC2L intronic transcript 1 (non-protein coding)', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214510182, end_position_on_the_genomic_accession=214536890, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  344697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344697, CurrentID=0, Status='live', Symbol='HNRNPA1P21', Aliases='HNRNPA1P', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 21', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39334915, end_position_on_the_genomic_accession=39336077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873663, CurrentID=0, Status='live', Symbol='RNA5SP412', Aliases='RN5S412', description='RNA, 5S ribosomal pseudogene 412', other_designations='RNA, 5S ribosomal 412', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35746772, end_position_on_the_genomic_accession=35746894, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373802, CurrentID=0, Status='live', Symbol='LOC105373802', Aliases='', description='uncharacterized LOC105373802', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190438859, end_position_on_the_genomic_accession=190446113, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  144832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=144832, CurrentID=0, Status='live', Symbol='ESRRAP2', Aliases='ESTRRA', description='estrogen-related receptor alpha pseudogene 2', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21260091, end_position_on_the_genomic_accession=21262362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647275, CurrentID=0, Status='live', Symbol='LOC647275', Aliases='', description='purine nucleoside phosphorylase pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76257562, end_position_on_the_genomic_accession=76258927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379727, CurrentID=0, Status='live', Symbol='LOC105379727', Aliases='', description='uncharacterized LOC105379727', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  344709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344709, CurrentID=0, Status='live', Symbol='CCDC75P1', Aliases='', description='coiled-coil domain containing 75 pseudogene 1', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73182168, end_position_on_the_genomic_accession=73183407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873665, CurrentID=0, Status='live', Symbol='RNA5SP414', Aliases='RN5S414', description='RNA, 5S ribosomal pseudogene 414', other_designations='RNA, 5S ribosomal 414', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35748520, end_position_on_the_genomic_accession=35748655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130924, CurrentID=0, Status='live', Symbol='LOC100130924', Aliases='', description='uncharacterized LOC100130924', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=11588717, end_position_on_the_genomic_accession=11589452, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373805, CurrentID=0, Status='live', Symbol='LOC105373805', Aliases='', description='uncharacterized LOC105373805', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191030972, end_position_on_the_genomic_accession=191032314, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873667, CurrentID=0, Status='live', Symbol='RNA5SP416', Aliases='RN5S416', description='RNA, 5S ribosomal pseudogene 416', other_designations='RNA, 5S ribosomal 416', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35751574, end_position_on_the_genomic_accession=35751684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344729, CurrentID=0, Status='live', Symbol='OR7E140P', Aliases='', description='olfactory receptor family 7 subfamily E member 140 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 140 pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8414779, end_position_on_the_genomic_accession=8415774, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130927, CurrentID=0, Status='live', Symbol='LOC100130927', Aliases='', description='uncharacterized LOC100130927', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=3093291, end_position_on_the_genomic_accession=3094693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79301, CurrentID=0, Status='live', Symbol='OR9L1P', Aliases='OR9L2P', description='olfactory receptor family 9 subfamily L member 1 pseudogene', other_designations='olfactory receptor, family 9, subfamily L, member 1 pseudogene|olfactory receptor, family 9, subfamily L, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58077350, end_position_on_the_genomic_accession=58078606, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  344741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344741, CurrentID=0, Status='live', Symbol='HNRNPA1P20', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 20', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148166139, end_position_on_the_genomic_accession=148167381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130929, CurrentID=0, Status='live', Symbol='RPL5P33', Aliases='RPL5_18_1557', description='ribosomal protein L5 pseudogene 33', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50847934, end_position_on_the_genomic_accession=50849653, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130930, CurrentID=0, Status='live', Symbol='LOC100130930', Aliases='', description='uncharacterized LOC100130930', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=4396003, end_position_on_the_genomic_accession=4401236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644335, CurrentID=0, Status='live', Symbol='LOC644335', Aliases='', description='SMG5 nonsense mediated mRNA decay factor pseudogene', other_designations='smg-5 homolog, nonsense mediated mRNA decay factor pseudogene', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109777938, end_position_on_the_genomic_accession=109782261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344752, CurrentID=0, Status='live', Symbol='AADACL2', Aliases='', description='arylacetamide deacetylase-like 2', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151733916, end_position_on_the_genomic_accession=151758016, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130931, CurrentID=0, Status='live', Symbol='LOC100130931', Aliases='', description='uncharacterized LOC100130931', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=27926394, end_position_on_the_genomic_accession=27927243, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  344758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344758, CurrentID=0, Status='live', Symbol='GPR149', Aliases='IEDA, PGR10', description='G protein-coupled receptor 149', other_designations='G protein-coupled receptor PGR10|G-protein coupled receptor PGR10', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154337672, end_position_on_the_genomic_accession=154429715, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  344760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344760, CurrentID=0, Status='live', Symbol='VN2R1P', Aliases='CASRL1, GPRC2B', description='vomeronasal 2 receptor 1 pseudogene', other_designations='calcium-sensing receptor like 1|seven transmembrane helix receptor', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156033582, end_position_on_the_genomic_accession=156037647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373812, CurrentID=0, Status='live', Symbol='LOC105373812', Aliases='', description='uncharacterized LOC105373812', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191779245, end_position_on_the_genomic_accession=191792523, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79306, CurrentID=0, Status='live', Symbol='OR4K7P', Aliases='OR4K10P', description='olfactory receptor family 4 subfamily K member 7 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 10 pseudogene|olfactory receptor, family 4, subfamily K, member 7 pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14570744, end_position_on_the_genomic_accession=14571716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647285, CurrentID=0, Status='live', Symbol='RPL29P9', Aliases='RPL29_4_447', description='ribosomal protein L29 pseudogene 9', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186922152, end_position_on_the_genomic_accession=186922800, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130934, CurrentID=0, Status='live', Symbol='ZNF663P', Aliases='ZNF663', description='zinc finger protein 663, pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46456493, end_position_on_the_genomic_accession=46457643, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79308, CurrentID=0, Status='live', Symbol='OR4P1P', Aliases='OR4P2P', description='olfactory receptor family 4 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily P, member 1 pseudogene|olfactory receptor, family 4, subfamily P, member 2 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55683240, end_position_on_the_genomic_accession=55684375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373815, CurrentID=0, Status='live', Symbol='LOC105373815', Aliases='', description='uncharacterized LOC105373815', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=193949681, end_position_on_the_genomic_accession=193972977, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873677, CurrentID=0, Status='live', Symbol='RNA5SP426', Aliases='RN5S426', description='RNA, 5S ribosomal pseudogene 426', other_designations='RNA, 5S ribosomal 426', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50869455, end_position_on_the_genomic_accession=50869575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373816, CurrentID=0, Status='live', Symbol='LOC105373816', Aliases='', description='uncharacterized LOC105373816', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194379601, end_position_on_the_genomic_accession=194386835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  344787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344787, CurrentID=0, Status='live', Symbol='ZNF860', Aliases='', description='zinc finger protein 860', other_designations='', map_location='3p23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31981774, end_position_on_the_genomic_accession=31991736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873678, CurrentID=0, Status='live', Symbol='RNA5SP427', Aliases='RN5S427', description='RNA, 5S ribosomal pseudogene 427', other_designations='RNA, 5S ribosomal 427', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53337453, end_position_on_the_genomic_accession=53337571, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347549, CurrentID=0, Status='live', Symbol='CENPVP3', Aliases='', description='centromere protein V pseudogene 3', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51618097, end_position_on_the_genomic_accession=51618873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  406991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=406991, CurrentID=0, Status='live', Symbol='MIR21', Aliases='MIRN21, hsa-mir-21, miR-21, miRNA21', description='microRNA 21', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59841266, end_position_on_the_genomic_accession=59841337, orientation='plus', exon_count=1, OMIM=611020, no_hdr0=''),
  79312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79312, CurrentID=0, Status='live', Symbol='OR1AA1P', Aliases='', description='olfactory receptor family 1 subfamily AA member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily AA, member 1 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131424302, end_position_on_the_genomic_accession=131425242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344805, CurrentID=0, Status='live', Symbol='TMPRSS7', Aliases='', description='transmembrane protease, serine 7', other_designations='matriptase-3|transmembrane serine protease 7|type II transmembrane serine protease 7', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112039067, end_position_on_the_genomic_accession=112081269, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  344807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344807, CurrentID=0, Status='live', Symbol='CD200R1L', Aliases='CD200R2, CD200RLa', description='CD200 receptor 1 like', other_designations='CD200 cell surface glycoprotein receptor 2|CD200 cell surface glycoprotein receptor-like 2|CD200 cell surface glycoprotein receptor-like a|CD200 receptor 1-like|CD200 receptor 2|CD200 receptor-like 2|cell surface glycoprotein CD200 receptor 1-like|cell surface glycoprotein OX2 receptor 2|huCD200R2', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112812239, end_position_on_the_genomic_accession=112846889, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  344813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344813, CurrentID=0, Status='live', Symbol='GSTO3P', Aliases='GSTO3P1', description='glutathione S-transferase omega 3, pseudogene', other_designations='glutathione S-transferase omega 1 pseudogene|glutathione S-transferase omega 3 pseudogene 1|glutathione-S-transferase omega 1 pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130827638, end_position_on_the_genomic_accession=130828433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128105, CurrentID=0, Status='live', Symbol='LOC100128105', Aliases='', description='uncharacterized LOC100128105', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=78914218, end_position_on_the_genomic_accession=78915833, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873683, CurrentID=0, Status='live', Symbol='RNA5SP433', Aliases='RN5S433', description='RNA, 5S ribosomal pseudogene 433', other_designations='RNA, 5S ribosomal 433', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84262815, end_position_on_the_genomic_accession=84262928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873684, CurrentID=0, Status='live', Symbol='RNA5SP434', Aliases='RN5S434', description='RNA, 5S ribosomal pseudogene 434', other_designations='RNA, 5S ribosomal 434', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4056756, end_position_on_the_genomic_accession=4056876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379731, CurrentID=0, Status='live', Symbol='LOC105379731', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421636, CurrentID=0, Status='live', Symbol='LOC100421636', Aliases='', description='poly(ADP-ribose) polymerase family member 14 pseudogene', other_designations='poly (ADP-ribose) polymerase family, member 14 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122594770, end_position_on_the_genomic_accession=122603627, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873685, CurrentID=0, Status='live', Symbol='RNA5SP436', Aliases='RN5S436', description='RNA, 5S ribosomal pseudogene 436', other_designations='RNA, 5S ribosomal 436', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15782341, end_position_on_the_genomic_accession=15782455, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  79318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79318, CurrentID=0, Status='live', Symbol='OR4K4P', Aliases='', description='olfactory receptor family 4 subfamily K member 4 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 4 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19905712, end_position_on_the_genomic_accession=19906644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344838, CurrentID=0, Status='live', Symbol='PAQR9', Aliases='', description='progestin and adipoQ receptor family member IX', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142961231, end_position_on_the_genomic_accession=142963336, orientation='minus', exon_count=1, OMIM=614580, no_hdr0=''),
  105375367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375367, CurrentID=0, Status='live', Symbol='LOC105375367', Aliases='', description='uncharacterized LOC105375367', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=78185459, end_position_on_the_genomic_accession=78215014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373826, CurrentID=0, Status='live', Symbol='LOC105373826', Aliases='', description='uncharacterized LOC105373826', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197215837, end_position_on_the_genomic_accession=197216865, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  647299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647299, CurrentID=0, Status='live', Symbol='LOC647299', Aliases='', description='enolase-phosphatase 1 pseudogene', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188989185, end_position_on_the_genomic_accession=188990314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79321, CurrentID=0, Status='live', Symbol='OR6M3P', Aliases='', description='olfactory receptor family 6 subfamily M member 3 pseudogene', other_designations='olfactory receptor, family 6, subfamily M, member 3 pseudogene', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123861681, end_position_on_the_genomic_accession=123862523, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421637, CurrentID=0, Status='live', Symbol='LOC100421637', Aliases='', description='acyl-CoA synthetase long-chain family member 3 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133760300, end_position_on_the_genomic_accession=133762363, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873690, CurrentID=0, Status='live', Symbol='RNA5SP444', Aliases='RN5S444', description='RNA, 5S ribosomal pseudogene 444', other_designations='RNA, 5S ribosomal 444', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66267743, end_position_on_the_genomic_accession=66267848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130724, CurrentID=0, Status='live', Symbol='LOC100130724', Aliases='', description='cyclin Y like 1 pseudogene', other_designations='cyclin Y-like 1 pseudogene|cyclin Y-like pseudogene', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65819443, end_position_on_the_genomic_accession=65822482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  344866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344866, CurrentID=0, Status='live', Symbol='KRT18P17', Aliases='', description='keratin 18 pseudogene 17', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12787335, end_position_on_the_genomic_accession=12788673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873691, CurrentID=0, Status='live', Symbol='RNA5SP445', Aliases='RN5S445', description='RNA, 5S ribosomal pseudogene 445', other_designations='RNA, 5S ribosomal 445', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66639269, end_position_on_the_genomic_accession=66639383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379462, CurrentID=0, Status='live', Symbol='LOC105379462', Aliases='', description='uncharacterized LOC105379462', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27926080, end_position_on_the_genomic_accession=27929559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130950, CurrentID=0, Status='live', Symbol='LOC100130950', Aliases='', description='uncharacterized LOC100130950', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5192084, end_position_on_the_genomic_accession=5235636, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100873692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873692, CurrentID=0, Status='live', Symbol='RNA5SP446', Aliases='RN5S446', description='RNA, 5S ribosomal pseudogene 446', other_designations='RNA, 5S ribosomal 446', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66895491, end_position_on_the_genomic_accession=66895572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375368, CurrentID=0, Status='live', Symbol='LOC105375368', Aliases='', description='uncharacterized LOC105375368', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77990389, end_position_on_the_genomic_accession=77996366, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  344875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344875, CurrentID=0, Status='live', Symbol='COL6A4P1', Aliases='COL6A4, COL6A4P, DIVA, DVWA, VWA6', description='collagen, type VI, alpha 4 pseudogene 1', other_designations='collagen VI alpha 4 pseudogene 1|collagen, type VI, alpha 4, pseudogene|dual intracellular von Willebrand factor domain A|dual von Willebrand factor A domains|von Willebrand factor A domain containing 6', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15165362, end_position_on_the_genomic_accession=15205959, orientation='minus', exon_count=5, OMIM=612397, no_hdr0=''),
  100131597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131597, CurrentID=0, Status='live', Symbol='LOC100131597', Aliases='', description='phosphorylated adaptor for RNA export pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6843478, end_position_on_the_genomic_accession=6844666, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873693, CurrentID=0, Status='live', Symbol='RNA5SP447', Aliases='RN5S447', description='RNA, 5S ribosomal pseudogene 447', other_designations='RNA, 5S ribosomal 447', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67268110, end_position_on_the_genomic_accession=67268228, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373832, CurrentID=0, Status='live', Symbol='LOC105373832', Aliases='', description='uncharacterized LOC105373832', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199812116, end_position_on_the_genomic_accession=199817697, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  344887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344887, CurrentID=0, Status='live', Symbol='LOC344887', Aliases='', description='NmrA-like family domain containing 1 pseudogene', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185959969, end_position_on_the_genomic_accession=185980877, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  344892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344892, CurrentID=0, Status='live', Symbol='RTP2', Aliases='Z3CXXC2', description='receptor (chemosensory) transporter protein 2', other_designations='3CxxC-type zinc finger protein 2|receptor transporter protein 2|receptor transporting protein 2|zinc finger, 3CxxC-type 2', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187698259, end_position_on_the_genomic_accession=187716720, orientation='minus', exon_count=7, OMIM=609138, no_hdr0=''),
  105373834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373834, CurrentID=0, Status='live', Symbol='LOC105373834', Aliases='', description='uncharacterized LOC105373834', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199970638, end_position_on_the_genomic_accession=200174746, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373835, CurrentID=0, Status='live', Symbol='LOC105373835', Aliases='', description='uncharacterized LOC105373835', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200963490, end_position_on_the_genomic_accession=201016352, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  344901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344901, CurrentID=0, Status='live', Symbol='OSTN', Aliases='MUSCLIN', description='osteocrin', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191199142, end_position_on_the_genomic_accession=191263338, orientation='plus', exon_count=5, OMIM=610280, no_hdr0=''),
  344905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344905, CurrentID=0, Status='live', Symbol='ATP13A5', Aliases='', description='ATPase type 13A5', other_designations='P5-ATPase', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193274789, end_position_on_the_genomic_accession=193378824, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  101928895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928895, CurrentID=0, Status='live', Symbol='LOC101928895', Aliases='', description='uncharacterized LOC101928895', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1274002, end_position_on_the_genomic_accession=1279447, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373837, CurrentID=0, Status='live', Symbol='LOC105373837', Aliases='', description='uncharacterized LOC105373837', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201359683, end_position_on_the_genomic_accession=201364217, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271261, CurrentID=0, Status='live', Symbol='RPS3AP35', Aliases='RPS3A_16_877', description='ribosomal protein S3a pseudogene 35', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12628432, end_position_on_the_genomic_accession=12629280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79331, CurrentID=0, Status='live', Symbol='OR2AF1P', Aliases='OR2AF2P', description='olfactory receptor family 2 subfamily AF member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AF, member 1 pseudogene|olfactory receptor, family 2, subfamily AF, member 2 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131619047, end_position_on_the_genomic_accession=131619936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647310, CurrentID=0, Status='live', Symbol='TEX22', Aliases='', description='testis expressed 22', other_designations='testis expressed gene 22', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105398583, end_position_on_the_genomic_accession=105413859, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106182118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106182118, CurrentID=0, Status='live', Symbol='LIMS1-AS1', Aliases='', description='LIMS1 antisense RNA 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375370, CurrentID=0, Status='live', Symbol='LOC105375370', Aliases='', description='uncharacterized LOC105375370', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79696924, end_position_on_the_genomic_accession=79716240, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873703, CurrentID=0, Status='live', Symbol='RNA5SP459', Aliases='RN5S459', description='RNA, 5S ribosomal pseudogene 459', other_designations='RNA, 5S ribosomal 459', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55146546, end_position_on_the_genomic_accession=55146645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373842, CurrentID=0, Status='live', Symbol='LOC105373842', Aliases='', description='uncharacterized LOC105373842', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202178660, end_position_on_the_genomic_accession=202193756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79336, CurrentID=0, Status='live', Symbol='OR5AH1P', Aliases='', description='olfactory receptor family 5 subfamily AH member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AH, member 1 pseudogene', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56760865, end_position_on_the_genomic_accession=56761233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130963, CurrentID=0, Status='live', Symbol='RPL38P4', Aliases='', description='ribosomal protein L38 pseudogene 4', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98525016, end_position_on_the_genomic_accession=98525349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373844, CurrentID=0, Status='live', Symbol='LOC105373844', Aliases='', description='uncharacterized LOC105373844', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203847449, end_position_on_the_genomic_accession=203865146, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373845, CurrentID=0, Status='live', Symbol='LOC105373845', Aliases='', description='uncharacterized LOC105373845', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204473812, end_position_on_the_genomic_accession=204475910, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873707, CurrentID=0, Status='live', Symbol='RNA5SP463', Aliases='RN5S463', description='RNA, 5S ribosomal pseudogene 463', other_designations='RNA, 5S ribosomal 463', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7886977, end_position_on_the_genomic_accession=7887096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373846, CurrentID=0, Status='live', Symbol='LOC105373846', Aliases='', description='uncharacterized LOC105373846', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204661813, end_position_on_the_genomic_accession=204663666, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  344967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344967, CurrentID=0, Status='live', Symbol='LOC344967', Aliases='', description='acyl-CoA thioesterase 7 pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40042917, end_position_on_the_genomic_accession=40057199, orientation='minus', exon_count=3, OMIM=611963, no_hdr0=''),
  647319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647319, CurrentID=0, Status='live', Symbol='VEZF1P1', Aliases='VEZF1L1, ZNF161L1', description='vascular endothelial zinc finger 1 pseudogene 1', other_designations='vascular endothelial zinc finger 1-like 1|zinc finger protein 161-like 1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193153537, end_position_on_the_genomic_accession=193157299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373848, CurrentID=0, Status='live', Symbol='LOC105373848', Aliases='', description='uncharacterized LOC105373848', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206094918, end_position_on_the_genomic_accession=206106004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  344978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344978, CurrentID=0, Status='live', Symbol='ACTN4P1', Aliases='', description='actinin, alpha 4 pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116598161, end_position_on_the_genomic_accession=116599608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873710, CurrentID=0, Status='live', Symbol='RNA5SP466', Aliases='RN5S466', description='RNA, 5S ribosomal pseudogene 466', other_designations='RNA, 5S ribosomal 466', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12070164, end_position_on_the_genomic_accession=12070271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  344988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344988, CurrentID=0, Status='live', Symbol='SAR1AP3', Aliases='SAR1P3', description='SAR1 homolog A (S. cerevisiae) pseudogene 3', other_designations='SAR1 gene homolog pseudogene 3|SAR1 homolog pseudogene 3', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120341980, end_position_on_the_genomic_accession=120344860, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373850, CurrentID=0, Status='live', Symbol='LOC105373850', Aliases='', description='uncharacterized LOC105373850', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206840962, end_position_on_the_genomic_accession=206869493, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373851, CurrentID=0, Status='live', Symbol='LOC105373851', Aliases='', description='uncharacterized LOC105373851', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207334516, end_position_on_the_genomic_accession=207351877, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873713, CurrentID=0, Status='live', Symbol='RNA5SP470', Aliases='RN5S470', description='RNA, 5S ribosomal pseudogene 470', other_designations='RNA, 5S ribosomal 470', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29001670, end_position_on_the_genomic_accession=29001775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130972, CurrentID=0, Status='live', Symbol='PPP1R2P6', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 6', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 6', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140292556, end_position_on_the_genomic_accession=140293538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873714, CurrentID=0, Status='live', Symbol='RNA5SP471', Aliases='RN5S471', description='RNA, 5S ribosomal pseudogene 471', other_designations='RNA, 5S ribosomal 471', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31655356, end_position_on_the_genomic_accession=31655461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  345016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345016, CurrentID=0, Status='live', Symbol='PES1P1', Aliases='', description='pescadillo ribosomal biogenesis factor 1 pseudogene 1', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134326993, end_position_on_the_genomic_accession=134328990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130976, CurrentID=0, Status='live', Symbol='LOC100130976', Aliases='', description='uncharacterized LOC100130976', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=94632284, end_position_on_the_genomic_accession=94632664, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373857, CurrentID=0, Status='live', Symbol='LOC105373857', Aliases='', description='uncharacterized LOC105373857', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208757110, end_position_on_the_genomic_accession=208768944, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  345041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345041, CurrentID=0, Status='live', Symbol='HSPD1P5', Aliases='HSPD1-8P, HSPDP5', description='heat shock protein family D (Hsp60) member 1 pseudogene 5', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 5', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144845565, end_position_on_the_genomic_accession=144847660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373860, CurrentID=0, Status='live', Symbol='LOC105373860', Aliases='', description='uncharacterized LOC105373860', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212706806, end_position_on_the_genomic_accession=212709454, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  345051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345051, CurrentID=0, Status='live', Symbol='LOC345051', Aliases='', description='hCG38984', other_designations='', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146215157, end_position_on_the_genomic_accession=146230875, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100418600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418600, CurrentID=0, Status='live', Symbol='LOC100418600', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  345062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345062, CurrentID=0, Status='live', Symbol='PRSS48', Aliases='ESSPL', description='protease, serine 48', other_designations='epidermis-specific serine protease-like protein|protease, serine, 48', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151276779, end_position_on_the_genomic_accession=151291611, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271262, CurrentID=0, Status='live', Symbol='RPS6P20', Aliases='RPS6_10_1296', description='ribosomal protein S6 pseudogene 20', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131292585, end_position_on_the_genomic_accession=131293176, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373863, CurrentID=0, Status='live', Symbol='LOC105373863', Aliases='', description='uncharacterized LOC105373863', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=213155804, end_position_on_the_genomic_accession=213167704, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  345079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345079, CurrentID=0, Status='live', Symbol='SOWAHB', Aliases='ANKRD56', description='sosondowah ankyrin repeat domain family member B', other_designations='ankyrin repeat domain 56|ankyrin repeat domain-containing protein 56|protein sosondowah homolog B', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76894929, end_position_on_the_genomic_accession=76897849, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373866, CurrentID=0, Status='live', Symbol='LOC105373866', Aliases='', description='uncharacterized LOC105373866', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214336954, end_position_on_the_genomic_accession=214472229, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130987, CurrentID=0, Status='live', Symbol='LOC100130987', Aliases='', description='uncharacterized LOC100130987', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67317839, end_position_on_the_genomic_accession=67391687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373868, CurrentID=0, Status='live', Symbol='LOC105373868', Aliases='', description='uncharacterized LOC105373868', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215436488, end_position_on_the_genomic_accession=215437119, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873731, CurrentID=0, Status='live', Symbol='RNA5SP489', Aliases='RN5S489', description='RNA, 5S ribosomal pseudogene 489', other_designations='RNA, 5S ribosomal 489', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25202207, end_position_on_the_genomic_accession=25202315, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130990, CurrentID=0, Status='live', Symbol='LOC100130990', Aliases='', description='zinc finger protein 561-like', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54857163, end_position_on_the_genomic_accession=54858491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873732, CurrentID=0, Status='live', Symbol='RNA5SP490', Aliases='RN5S490', description='RNA, 5S ribosomal pseudogene 490', other_designations='RNA, 5S ribosomal 490', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32563075, end_position_on_the_genomic_accession=32563210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  126506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126506, CurrentID=0, Status='live', Symbol='SNX6P1', Aliases='', description='sorting nexin 6 pseudogene 1', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23125841, end_position_on_the_genomic_accession=23127804, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128107, CurrentID=0, Status='live', Symbol='LOC100128107', Aliases='', description='uncharacterized LOC100128107', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=51585560, end_position_on_the_genomic_accession=51586648, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873733, CurrentID=0, Status='live', Symbol='RNA5SP491', Aliases='RN5S491', description='RNA, 5S ribosomal pseudogene 491', other_designations='RNA, 5S ribosomal 491', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36851911, end_position_on_the_genomic_accession=36852028, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130992, CurrentID=0, Status='live', Symbol='LOC100130992', Aliases='', description='uncharacterized LOC100130992', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22252072, end_position_on_the_genomic_accession=22258548, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100873734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873734, CurrentID=0, Status='live', Symbol='RNA5SP492', Aliases='RN5S492', description='RNA, 5S ribosomal pseudogene 492', other_designations='RNA, 5S ribosomal 492', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42221173, end_position_on_the_genomic_accession=42221286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371012, CurrentID=0, Status='live', Symbol='LOC105371012', Aliases='', description='uncharacterized LOC105371012', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98520258, end_position_on_the_genomic_accession=98553223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873735, CurrentID=0, Status='live', Symbol='RNA5SP493', Aliases='RN5S493', description='RNA, 5S ribosomal pseudogene 493', other_designations='RNA, 5S ribosomal 493', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21792434, end_position_on_the_genomic_accession=21792524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373874, CurrentID=0, Status='live', Symbol='LOC105373874', Aliases='', description='uncharacterized LOC105373874', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217035119, end_position_on_the_genomic_accession=217039287, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873736, CurrentID=0, Status='live', Symbol='RNA5SP494', Aliases='RN5S494', description='RNA, 5S ribosomal pseudogene 494', other_designations='RNA, 5S ribosomal 494', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25715041, end_position_on_the_genomic_accession=25715169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373875, CurrentID=0, Status='live', Symbol='LOC105373875', Aliases='', description='uncharacterized LOC105373875', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216887926, end_position_on_the_genomic_accession=216889525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873737, CurrentID=0, Status='live', Symbol='RNA5SP495', Aliases='RN5S495', description='RNA, 5S ribosomal pseudogene 495', other_designations='RNA, 5S ribosomal 495', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26390667, end_position_on_the_genomic_accession=26390776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373876, CurrentID=0, Status='live', Symbol='LOC105373876', Aliases='', description='uncharacterized LOC105373876', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216866332, end_position_on_the_genomic_accession=216867813, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100873738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873738, CurrentID=0, Status='live', Symbol='RNA5SP496', Aliases='RN5S496', description='RNA, 5S ribosomal pseudogene 496', other_designations='RNA, 5S ribosomal 496', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31305404, end_position_on_the_genomic_accession=31305512, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131000, CurrentID=0, Status='live', Symbol='LOC100131000', Aliases='', description='uncharacterized LOC100131000', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=74351729, end_position_on_the_genomic_accession=74352966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131599, CurrentID=0, Status='live', Symbol='LOC100131599', Aliases='', description='uncharacterized LOC100131599', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=8799594, end_position_on_the_genomic_accession=8800377, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  345193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345193, CurrentID=0, Status='live', Symbol='LRIT3', Aliases='CSNB1F, FIGLER4', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 3', other_designations='fibronectin type III, immunoglobulin and leucine rich repeat domains 4', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109848184, end_position_on_the_genomic_accession=109872315, orientation='plus', exon_count=5, OMIM=615004, no_hdr0=''),
  101060363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060363, CurrentID=0, Status='live', Symbol='LOC101060363', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22322825, end_position_on_the_genomic_accession=22323331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372481, CurrentID=0, Status='live', Symbol='LOC105372481', Aliases='', description='uncharacterized LOC105372481', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58305337, end_position_on_the_genomic_accession=58315653, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421304, CurrentID=0, Status='live', Symbol='LOC100421304', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369241, CurrentID=0, Status='live', Symbol='LOC105369241', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271263, CurrentID=0, Status='live', Symbol='RPS6P21', Aliases='RPS6_11_1297', description='ribosomal protein S6 pseudogene 21', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131667480, end_position_on_the_genomic_accession=131668281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371015, CurrentID=0, Status='live', Symbol='LOC105371015', Aliases='', description='uncharacterized LOC105371015', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98772629, end_position_on_the_genomic_accession=98780294, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288721, CurrentID=0, Status='live', Symbol='LOC100288721', Aliases='', description='uncharacterized LOC100288721', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=108895291, end_position_on_the_genomic_accession=108896213, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  345222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345222, CurrentID=0, Status='live', Symbol='MSANTD1', Aliases='C4orf44', description='Myb/SANT-like DNA-binding domain containing 1', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3241598, end_position_on_the_genomic_accession=3265085, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100131011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131011, CurrentID=0, Status='live', Symbol='LOC100131011', Aliases='', description='myotubularin related protein 14 pseudogene', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67830673, end_position_on_the_genomic_accession=67831943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373892, CurrentID=0, Status='live', Symbol='LOC105373892', Aliases='', description='uncharacterized LOC105373892', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=220067468, end_position_on_the_genomic_accession=220073310, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104940698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104940698, CurrentID=0, Status='live', Symbol='DALIR', Aliases='DALI', description='DNMT1-associated long intergenic non-coding RNA', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131013, CurrentID=0, Status='live', Symbol='LOC100131013', Aliases='', description='ribosomal protein L7-like 1 pseudogene', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54696393, end_position_on_the_genomic_accession=54697123, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371016, CurrentID=0, Status='live', Symbol='LOC105371016', Aliases='', description='uncharacterized LOC105371016', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99090904, end_position_on_the_genomic_accession=99107995, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  345258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345258, CurrentID=0, Status='live', Symbol='BIN2P1', Aliases='', description='bridging integrator 2 pseudogene 1', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82274849, end_position_on_the_genomic_accession=82276800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375381, CurrentID=0, Status='live', Symbol='LOC105375381', Aliases='', description='uncharacterized LOC105375381', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=86594339, end_position_on_the_genomic_accession=86614750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373897, CurrentID=0, Status='live', Symbol='LOC105373897', Aliases='', description='uncharacterized LOC105373897', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221587411, end_position_on_the_genomic_accession=221610365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  345274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345274, CurrentID=0, Status='live', Symbol='SLC10A6', Aliases='SOAT', description='solute carrier family 10 (sodium/bile acid cotransporter), member 6', other_designations='sodium-dependent organic anion transporter|solute carrier family 10 (sodium/bile acid cotransporter family), member 6', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86787607, end_position_on_the_genomic_accession=86849360, orientation='minus', exon_count=8, OMIM=613366, no_hdr0=''),
  345275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345275, CurrentID=0, Status='live', Symbol='HSD17B13', Aliases='HMFN0376, NIIL497, SCDR9, SDR16C3', description='hydroxysteroid (17-beta) dehydrogenase 13', other_designations='17-beta hydroxysteroid dehydrogenase|17-beta-HSD 13|short chain dehydrogenase/reductase family 16C member 3|short chain dehydrogenase/reductase family 16C, member 3|short-chain dehydrogenase/reductase 9', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87303789, end_position_on_the_genomic_accession=87322906, orientation='minus', exon_count=7, OMIM=612127, no_hdr0=''),
  105373900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373900, CurrentID=0, Status='live', Symbol='LOC105373900', Aliases='', description='uncharacterized LOC105373900', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221906474, end_position_on_the_genomic_accession=221960848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373902, CurrentID=0, Status='live', Symbol='LOC105373902', Aliases='', description='uncharacterized LOC105373902', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222644912, end_position_on_the_genomic_accession=222650900, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104940763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104940763, CurrentID=0, Status='live', Symbol='LOC104940763', Aliases='PR-lncRNA-1', description='p53-regulated lncRNA 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100873768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873768, CurrentID=0, Status='live', Symbol='RNU6-63P', Aliases='RNU6-63', description='RNA, U6 small nuclear 63, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27488138, end_position_on_the_genomic_accession=27488232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373907, CurrentID=0, Status='live', Symbol='LOC105373907', Aliases='', description='uncharacterized LOC105373907', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223489113, end_position_on_the_genomic_accession=223506153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873769, CurrentID=0, Status='live', Symbol='RNU6-64P', Aliases='RNU6-64', description='RNA, U6 small nuclear 64, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30281765, end_position_on_the_genomic_accession=30281869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379748, CurrentID=0, Status='live', Symbol='LOC105379748', Aliases='', description='uncharacterized LOC105379748', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371019, CurrentID=0, Status='live', Symbol='LOC105371019', Aliases='', description='uncharacterized LOC105371019', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99498492, end_position_on_the_genomic_accession=99509820, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131029, CurrentID=0, Status='live', Symbol='LOC100131029', Aliases='', description='transportin 3 pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114389135, end_position_on_the_genomic_accession=114389956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373911, CurrentID=0, Status='live', Symbol='LOC105373911', Aliases='', description='uncharacterized LOC105373911', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224739008, end_position_on_the_genomic_accession=224742296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373912, CurrentID=0, Status='live', Symbol='LOC105373912', Aliases='', description='uncharacterized LOC105373912', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224936467, end_position_on_the_genomic_accession=224941418, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373913, CurrentID=0, Status='live', Symbol='LOC105373913', Aliases='', description='uncharacterized LOC105373913', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226284256, end_position_on_the_genomic_accession=226285872, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373914, CurrentID=0, Status='live', Symbol='LOC105373914', Aliases='', description='uncharacterized LOC105373914', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=225680561, end_position_on_the_genomic_accession=225748583, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929771, CurrentID=0, Status='live', Symbol='LOC101929771', Aliases='', description='uncharacterized LOC101929771', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221827666, end_position_on_the_genomic_accession=221840666, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373915, CurrentID=0, Status='live', Symbol='LOC105373915', Aliases='', description='uncharacterized LOC105373915', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226454782, end_position_on_the_genomic_accession=226466774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373916, CurrentID=0, Status='live', Symbol='LOC105373916', Aliases='', description='uncharacterized LOC105373916', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226799905, end_position_on_the_genomic_accession=226812425, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373917, CurrentID=0, Status='live', Symbol='LOC105373917', Aliases='', description='uncharacterized LOC105373917', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227172055, end_position_on_the_genomic_accession=227176410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873779, CurrentID=0, Status='live', Symbol='RNU6-79P', Aliases='RNU6-79', description='RNA, U6 small nuclear 79, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72881108, end_position_on_the_genomic_accession=72881214, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131038, CurrentID=0, Status='live', Symbol='LOC100131038', Aliases='', description='GC-rich promoter binding protein 1 pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158469587, end_position_on_the_genomic_accession=158470670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873780, CurrentID=0, Status='live', Symbol='RNU6-81P', Aliases='RNU6-81', description='RNA, U6 small nuclear 81, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63783801, end_position_on_the_genomic_accession=63783904, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83259, CurrentID=0, Status='live', Symbol='PCDH11Y', Aliases='PCDH-PC, PCDH22, PCDHX, PCDHY', description='protocadherin 11 Y-linked', other_designations='protocadherin 22|protocadherin on the Y chromosome|protocadherin prostate cancer|protocadherin-PC', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5000226, end_position_on_the_genomic_accession=5742228, orientation='plus', exon_count=13, OMIM=400022, no_hdr0=''),
  100873781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873781, CurrentID=0, Status='live', Symbol='RNU6-82P', Aliases='RNU6-82', description='RNA, U6 small nuclear 82, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28232501, end_position_on_the_genomic_accession=28232607, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646700, CurrentID=0, Status='live', Symbol='LOC646700', Aliases='', description='UPF0607 protein ENSP00000381418-like pseudogene', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28148011, end_position_on_the_genomic_accession=28149236, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873782, CurrentID=0, Status='live', Symbol='RNU6-83P', Aliases='RNU6-83', description='RNA, U6 small nuclear 83, pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99025234, end_position_on_the_genomic_accession=99025345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131041, CurrentID=0, Status='live', Symbol='LOC100131041', Aliases='', description='FK506 binding protein 7 pseudogene', other_designations='', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144036501, end_position_on_the_genomic_accession=144037765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873783, CurrentID=0, Status='live', Symbol='UBBP5', Aliases='', description='ubiquitin B pseudogene 5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87219004, end_position_on_the_genomic_accession=87219269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131043, CurrentID=0, Status='live', Symbol='LOC100131043', Aliases='', description='uncharacterized LOC100131043', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=36831299, end_position_on_the_genomic_accession=36831869, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  345430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345430, CurrentID=0, Status='live', Symbol='KRT18P41', Aliases='', description='keratin 18 pseudogene 41', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170140604, end_position_on_the_genomic_accession=170141946, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652976, CurrentID=0, Status='live', Symbol='LOC100652976', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49849615, end_position_on_the_genomic_accession=49852499, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373926, CurrentID=0, Status='live', Symbol='LOC105373926', Aliases='', description='putative HERV-K_Xq28 provirus ancestral Pol protein', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231401613, end_position_on_the_genomic_accession=231408342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373927, CurrentID=0, Status='live', Symbol='LOC105373927', Aliases='', description='uncharacterized LOC105373927', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231412453, end_position_on_the_genomic_accession=231426537, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  345456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345456, CurrentID=0, Status='live', Symbol='PFN3', Aliases='', description='profilin 3', other_designations='profilin III', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177400107, end_position_on_the_genomic_accession=177400636, orientation='minus', exon_count=1, OMIM=612812, no_hdr0=''),
  100131048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131048, CurrentID=0, Status='live', Symbol='LOC100131048', Aliases='', description='uncharacterized LOC100131048', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=130280209, end_position_on_the_genomic_accession=130280826, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  345462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345462, CurrentID=0, Status='live', Symbol='ZNF879', Aliases='DKFZp686E2433', description='zinc finger protein 879', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179023732, end_position_on_the_genomic_accession=179034387, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100873791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873791, CurrentID=0, Status='live', Symbol='CHEK2P3', Aliases='', description='checkpoint kinase 2 pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91963523, end_position_on_the_genomic_accession=91963612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  345471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345471, CurrentID=0, Status='live', Symbol='LOC345471', Aliases='', description='cell division cycle 37 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157986113, end_position_on_the_genomic_accession=157987716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79424, CurrentID=0, Status='live', Symbol='CYP3AP2', Aliases='CYP3A5P2', description='cytochrome P450, family 3, subfamily A, polypeptide 5 pseudogene 2', other_designations='cytochrome P450, subfamily IIIA, pseudogene 2', map_location='7q21.3-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99748249, end_position_on_the_genomic_accession=99752491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131051, CurrentID=0, Status='live', Symbol='MED28P3', Aliases='', description='mediator complex subunit 28 pseudogene 3', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186364598, end_position_on_the_genomic_accession=186365130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131053, CurrentID=0, Status='live', Symbol='LOC100131053', Aliases='', description='uncharacterized LOC100131053', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=19531483, end_position_on_the_genomic_accession=19532393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132474, CurrentID=0, Status='live', Symbol='LOC100132474', Aliases='', description='uncharacterized LOC100132474', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=9812452, end_position_on_the_genomic_accession=9813718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373935, CurrentID=0, Status='live', Symbol='LOC105373935', Aliases='', description='uncharacterized LOC105373935', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234438328, end_position_on_the_genomic_accession=234463212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373356, CurrentID=0, Status='live', Symbol='LOC105373356', Aliases='', description='uncharacterized LOC105373356', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373940, CurrentID=0, Status='live', Symbol='LOC105373940', Aliases='', description='uncharacterized LOC105373940', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235270617, end_position_on_the_genomic_accession=235274228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83394, CurrentID=0, Status='live', Symbol='PITPNM3', Aliases='ACKR6, CORD5, NIR1, RDGBA3', description='PITPNM family member 3', other_designations='NIR-1|PITPnm 3|PYK2 N-terminal domain-interacting receptor 1|atypical chemokine receptor 6|cone rod dystrophy 5|phosphatidylinositol transfer protein, membrane-associated 3|retinal degeneration B alpha 3', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6450344, end_position_on_the_genomic_accession=6556557, orientation='minus', exon_count=23, OMIM=608921, no_hdr0=''),
  83401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83401, CurrentID=0, Status='live', Symbol='ELOVL3', Aliases='CIG-30, CIG30', description='ELOVL fatty acid elongase 3', other_designations='3-keto acyl-CoA synthase ELOVL3|ELOVL FA elongase 3|cold-inducible glycoprotein of 30 kDa|elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 3|very-long-chain 3-oxoacyl-CoA synthase 3', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102225358, end_position_on_the_genomic_accession=102229589, orientation='plus', exon_count=5, OMIM=611815, no_hdr0=''),
  105373943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373943, CurrentID=0, Status='live', Symbol='LOC105373943', Aliases='', description='uncharacterized LOC105373943', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235675982, end_position_on_the_genomic_accession=235685021, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  345557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345557, CurrentID=0, Status='live', Symbol='PLCXD3', Aliases='', description='phosphatidylinositol-specific phospholipase C, X domain containing 3', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41306946, end_position_on_the_genomic_accession=41510628, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373945, CurrentID=0, Status='live', Symbol='LOC105373945', Aliases='', description='uncharacterized LOC105373945', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236337818, end_position_on_the_genomic_accession=236356131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  83416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83416, CurrentID=0, Status='live', Symbol='FCRL5', Aliases='BXMAS1, CD307, CD307e, FCRH5, IRTA2, PRO820', description='Fc receptor-like 5', other_designations='fc receptor homolog 5|fcR-like protein 5|immune receptor translocation-associated protein 2|immunoglobulin superfamily receptor translocation associated 2 (IRTA2)', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157513377, end_position_on_the_genomic_accession=157552520, orientation='minus', exon_count=17, OMIM=605877, no_hdr0=''),
  83417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83417, CurrentID=0, Status='live', Symbol='FCRL4', Aliases='CD307d, FCRH4, IGFP2, IRTA1', description='Fc receptor-like 4', other_designations='IFGP family protein 2|fc receptor homolog 4|fcR-like protein 4|hIFGP2|immune receptor translocation-associated protein 1|immunoglobulin superfamily Fc receptor, gp42|immunoglobulin superfamily receptor translocation associated 1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157573749, end_position_on_the_genomic_accession=157598080, orientation='minus', exon_count=12, OMIM=605876, no_hdr0=''),
  100131066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131066, CurrentID=0, Status='live', Symbol='LOC100131066', Aliases='', description='uncharacterized LOC100131066', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=92961925, end_position_on_the_genomic_accession=92962326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  345571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345571, CurrentID=0, Status='live', Symbol='LOC345571', Aliases='', description='RUN and FYVE domain containing 3 pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=106416099, end_position_on_the_genomic_accession=106417453, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873809, CurrentID=0, Status='live', Symbol='RNY3P5', Aliases='', description='RNA, Ro-associated Y3 pseudogene 5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60789460, end_position_on_the_genomic_accession=60789761, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  345576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345576, CurrentID=0, Status='live', Symbol='LOC345576', Aliases='', description='ubiquitin specific peptidase 7 (herpes virus-associated) pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=107724800, end_position_on_the_genomic_accession=107727126, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373948, CurrentID=0, Status='live', Symbol='LOC105373948', Aliases='', description='uncharacterized LOC105373948', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236731523, end_position_on_the_genomic_accession=236743517, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100873810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873810, CurrentID=0, Status='live', Symbol='RNY3P6', Aliases='', description='RNA, Ro-associated Y3 pseudogene 6', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99536319, end_position_on_the_genomic_accession=99536396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83439, CurrentID=0, Status='live', Symbol='TCF7L1', Aliases='TCF-3, TCF3', description='transcription factor 7-like 1 (T-cell specific, HMG-box)', other_designations='HMG box transcription factor 3', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85133460, end_position_on_the_genomic_accession=85310388, orientation='plus', exon_count=12, OMIM=604652, no_hdr0=''),
  83440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83440, CurrentID=0, Status='live', Symbol='ADPGK', Aliases='2610017G09Rik, ADP-GK', description='ADP-dependent glucokinase', other_designations='ATP-dependent glucokinase|rbBP-35', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72751294, end_position_on_the_genomic_accession=72783785, orientation='minus', exon_count=8, OMIM=611861, no_hdr0=''),
  83442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83442, CurrentID=0, Status='live', Symbol='SH3BGRL3', Aliases='HEL-S-297, SH3BP-1, TIP-B1', description='SH3 domain binding glutamate-rich protein like 3', other_designations='SH3 domain binding glutamic acid-rich protein like 3|SH3 domain-binding protein 1|SH3BGRL3-like protein|TNF inhibitory protein|epididymis secretory protein Li 297', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26279722, end_position_on_the_genomic_accession=26281522, orientation='plus', exon_count=3, OMIM=615679, no_hdr0=''),
  83443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83443, CurrentID=0, Status='live', Symbol='SF3B5', Aliases='SF3b10, Ysf3', description='splicing factor 3b subunit 5', other_designations='pre-mRNA-splicing factor SF3b 10 kDa subunit|splicing factor 3b, subunit 5, 10kDa', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144094881, end_position_on_the_genomic_accession=144095617, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83444, CurrentID=0, Status='live', Symbol='INO80B', Aliases='HMGA1L4, HMGIYL4, IES2, PAP-1BP, PAPA-1, ZNHIT4, hIes2', description='INO80 complex subunit B', other_designations='IES2 homolog|PAP (Pim-1 associated protein) associated protein 1|PAP-1 binding protein|PAP-1-associated protein 1|high mobility group AT-hook 1-like 4|zinc finger HIT domain-containing protein 4|zinc finger, HIT type 4', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74455023, end_position_on_the_genomic_accession=74457960, orientation='plus', exon_count=5, OMIM=616456, no_hdr0=''),
  83445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83445, CurrentID=0, Status='live', Symbol='GSG1', Aliases='', description='germ cell associated 1', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13083537, end_position_on_the_genomic_accession=13103720, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  83446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83446, CurrentID=0, Status='live', Symbol='CCDC70', Aliases='', description='coiled-coil domain containing 70', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51861981, end_position_on_the_genomic_accession=51866236, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83447, CurrentID=0, Status='live', Symbol='SLC25A31', Aliases='AAC4, ANT4, SFEC35kDa', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 31', other_designations='ADP,ATP carrier protein 4|ANT 4|adenine nucleotide translocase 4|adenine nucleotide translocator 4|solute carrier family 25 member 31|sperm flagellar energy carrier protein', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127730375, end_position_on_the_genomic_accession=127774299, orientation='plus', exon_count=7, OMIM=610796, no_hdr0=''),
  83448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83448, CurrentID=0, Status='live', Symbol='PUS7L', Aliases='', description='pseudouridylate synthase 7-like', other_designations='pseudouridylate synthase 7 homolog (S. cerevisiae)-like', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43728025, end_position_on_the_genomic_accession=43758817, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  83449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83449, CurrentID=0, Status='live', Symbol='PMFBP1', Aliases='', description='polyamine modulated factor 1 binding protein 1', other_designations='PMF-1 binding protein', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72116698, end_position_on_the_genomic_accession=72177002, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  83450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83450, CurrentID=0, Status='live', Symbol='DRC3', Aliases='LRRC48', description='dynein regulatory complex subunit 3', other_designations='leucine rich repeat containing 48', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17972780, end_position_on_the_genomic_accession=18016889, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  83451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83451, CurrentID=0, Status='live', Symbol='ABHD11', Aliases='PP1226, WBSCR21', description='abhydrolase domain containing 11', other_designations='Williams Beuren syndrome chromosome region 21|abhydrolase domain-containing protein 11|williams-Beuren syndrome chromosomal region 21 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73736094, end_position_on_the_genomic_accession=73738885, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83452, CurrentID=0, Status='live', Symbol='RAB33B', Aliases='SMC2', description='RAB33B, member RAS oncogene family', other_designations='', map_location='4q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139453232, end_position_on_the_genomic_accession=139475916, orientation='plus', exon_count=3, OMIM=605950, no_hdr0=''),
  100131072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131072, CurrentID=0, Status='live', Symbol='LOC100131072', Aliases='', description='karyopherin (importin) beta 1 pseudogene', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=97976317, end_position_on_the_genomic_accession=97978642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83460, CurrentID=0, Status='live', Symbol='EMC6', Aliases='TMEM93', description='ER membrane protein complex subunit 6', other_designations='transmembrane protein 93', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3668723, end_position_on_the_genomic_accession=3669669, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83461, CurrentID=0, Status='live', Symbol='CDCA3', Aliases='GRCC8, TOME-1', description='cell division cycle associated 3', other_designations='gene rich cluster, C8|gene-rich cluster protein C8|trigger of mitotic entry 1|trigger of mitotic entry protein 1', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6844798, end_position_on_the_genomic_accession=6852066, orientation='minus', exon_count=8, OMIM=607749, no_hdr0=''),
  83463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83463, CurrentID=0, Status='live', Symbol='MXD3', Aliases='BHLHC13, MAD3, MYX', description='MAX dimerization protein 3', other_designations='Max-associated protein 3|Max-interacting transcriptional repressor MAD3|class C basic helix-loop-helix protein 13|max dimerizer 3', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177305500, end_position_on_the_genomic_accession=177312291, orientation='minus', exon_count=7, OMIM=609450, no_hdr0=''),
  83464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83464, CurrentID=0, Status='live', Symbol='APH1B', Aliases='APH-1B, PRO1328, PSFL, TAAV688', description='aph-1 homolog B, gamma secretase subunit', other_designations='APH1B gamma secretase subunit|anterior pharynx defective 1 homolog B|aph-1beta', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63277519, end_position_on_the_genomic_accession=63309126, orientation='plus', exon_count=7, OMIM=607630, no_hdr0=''),
  345611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345611, CurrentID=0, Status='live', Symbol='IRGM', Aliases='IFI11, LRG-47, LRG47, IRGM', description='immunity-related GTPase family, M', other_designations='LPS-stimulated RAW 264.7 macrophage protein 47 homolog|LRG-47-like protein|immunity-related GTPase family, M1|interferon-inducible protein 1', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150846523, end_position_on_the_genomic_accession=150902400, orientation='plus', exon_count=5, OMIM=608212, no_hdr0=''),
  83468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83468, CurrentID=0, Status='live', Symbol='GLT8D2', Aliases='', description='glycosyltransferase 8 domain containing 2', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103988984, end_position_on_the_genomic_accession=104064183, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  83473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83473, CurrentID=0, Status='live', Symbol='KATNAL2', Aliases='', description='katanin p60 subunit A-like 2', other_designations='p60 katanin-like 2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46917602, end_position_on_the_genomic_accession=47102243, orientation='plus', exon_count=21, OMIM=614697, no_hdr0=''),
  83475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83475, CurrentID=0, Status='live', Symbol='DOHH', Aliases='HLRC1, hDOHH', description='deoxyhypusine hydroxylase/monooxygenase', other_designations='HEAT-like (PBS lyase) repeat containing 1|HEAT-like repeat-containing protein 1|deoxyhypusine dioxygenase|deoxyhypusine monooxygenase', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3490821, end_position_on_the_genomic_accession=3505918, orientation='minus', exon_count=7, OMIM=611262, no_hdr0=''),
  83478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83478, CurrentID=0, Status='live', Symbol='ARHGAP24', Aliases='FILGAP, RC-GAP72, RCGAP72, p73, p73RhoGAP', description='Rho GTPase activating protein 24', other_designations='RAC1- and CDC42-specific GTPase-activating protein of 72 kDa|filamin-A-associated RhoGAP|rho-type GTPase-activating protein 24|rhoGAP of 73 kDa|sarcoma antigen NY-SAR-88', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=85147742, end_position_on_the_genomic_accession=86002670, orientation='plus', exon_count=15, OMIM=610586, no_hdr0=''),
  83479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83479, CurrentID=0, Status='live', Symbol='DDX59', Aliases='OFD5, ZNHIT5', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 59', other_designations='DEAD box protein 59|zinc finger HIT domain-containing protein 5', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200640807, end_position_on_the_genomic_accession=200669998, orientation='minus', exon_count=11, OMIM=615464, no_hdr0=''),
  83480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83480, CurrentID=0, Status='live', Symbol='PUS3', Aliases='2610020J05Rik, FKSG32', description='pseudouridylate synthase 3', other_designations='tRNA pseudouridine synthase 3|tRNA pseudouridylate synthase 3|tRNA-uridine isomerase 3', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125893485, end_position_on_the_genomic_accession=125903369, orientation='minus', exon_count=4, OMIM=616283, no_hdr0=''),
  83481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83481, CurrentID=0, Status='live', Symbol='EPPK1', Aliases='EPIPL, EPIPL1', description='epiplakin 1', other_designations='450 kDa epidermal antigen|epidermal autoantigen 450K', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143857319, end_position_on_the_genomic_accession=143878464, orientation='minus', exon_count=3, OMIM=607553, no_hdr0=''),
  83482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83482, CurrentID=0, Status='live', Symbol='SCRT1', Aliases='SCRT, ZNF898', description='scratch family zinc finger 1', other_designations='scratch homolog 1, zinc finger protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144330565, end_position_on_the_genomic_accession=144336281, orientation='minus', exon_count=2, OMIM=605858, no_hdr0=''),
  83483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83483, CurrentID=0, Status='live', Symbol='PLVAP', Aliases='FELS, PV-1, PV1, gp68', description='plasmalemma vesicle associated protein', other_designations='fenestrated endothelial-linked structure protein|fenestrated-endothelial linked structure protein; PV-1 protein|plasmalemma vesicle protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17351448, end_position_on_the_genomic_accession=17377349, orientation='minus', exon_count=6, OMIM=607647, no_hdr0=''),
  345630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345630, CurrentID=0, Status='live', Symbol='FBLL1', Aliases='', description='fibrillarin-like 1', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168529577, end_position_on_the_genomic_accession=168530634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373957, CurrentID=0, Status='live', Symbol='LOC105373957', Aliases='', description='uncharacterized LOC105373957', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237485700, end_position_on_the_genomic_accession=237487338, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  574532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574532, CurrentID=0, Status='live', Symbol='NUDT15P1', Aliases='', description='nudix hydrolase 15 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 15 pseudogene 1', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=59556699, end_position_on_the_genomic_accession=59557936, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129240, CurrentID=0, Status='live', Symbol='LOC100129240', Aliases='', description='uncharacterized LOC100129240', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=186145552, end_position_on_the_genomic_accession=186146269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131078, CurrentID=0, Status='live', Symbol='LOC100131078', Aliases='', description='MOB family member 4, phocein pseudogene', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86821143, end_position_on_the_genomic_accession=86821816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  345643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345643, CurrentID=0, Status='live', Symbol='MCIDAS', Aliases='IDAS, MCI, MCIN', description='multiciliate differentiation and DNA synthesis associated cell cycle protein', other_designations='multiciliate cell differentiation 1|multiciliate differentiation and DNA synthesis-associated cell cycle protein|protein Idas', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55219597, end_position_on_the_genomic_accession=55227315, orientation='minus', exon_count=7, OMIM=614086, no_hdr0=''),
  345645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345645, CurrentID=0, Status='live', Symbol='PSMC1P4', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 4', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56275076, end_position_on_the_genomic_accession=56276673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373960, CurrentID=0, Status='live', Symbol='LOC105373960', Aliases='', description='uncharacterized LOC105373960', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237922333, end_position_on_the_genomic_accession=237934197, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  345651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345651, CurrentID=0, Status='live', Symbol='ACTBL2', Aliases='ACT', description='actin, beta-like 2', other_designations='actin-like protein|kappa-actin', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57480016, end_position_on_the_genomic_accession=57482809, orientation='minus', exon_count=1, OMIM=614835, no_hdr0=''),
  574533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574533, CurrentID=0, Status='live', Symbol='NUDT15P2', Aliases='', description='nudix hydrolase 15 pseudogene 2', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 15 pseudogene 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32428087, end_position_on_the_genomic_accession=32435429, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373963, CurrentID=0, Status='live', Symbol='LOC105373963', Aliases='', description='uncharacterized LOC105373963', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238707848, end_position_on_the_genomic_accession=238722612, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421664, CurrentID=0, Status='live', Symbol='LOC100421664', Aliases='', description='Spi-C transcription factor (Spi-1/PU.1 related) pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2324468, end_position_on_the_genomic_accession=2325086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83528, CurrentID=0, Status='live', Symbol='CYP51A1P1', Aliases='CYP51P1', description='cytochrome P450 family 51 subfamily A member 1 pseudogene 1', other_designations='cytochrome P450, family 51, pseudogene 1|cytochrome P450, family 51, subfamily A, polypeptide 1 pseudogene 1|cytochrome P450, subfamily 51 pseudogene 1', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=82805011, end_position_on_the_genomic_accession=82808156, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647436, CurrentID=0, Status='live', Symbol='RPL5P1', Aliases='RPL5_12_1419', description='ribosomal protein L5 pseudogene 1', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24908864, end_position_on_the_genomic_accession=24909900, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83530, CurrentID=0, Status='live', Symbol='CYP51A1P2', Aliases='CYP51P2', description='cytochrome P450 family 51 subfamily A member 1 pseudogene 2', other_designations='cytochrome P450, family 51, pseudogene 2|cytochrome P450, family 51, subfamily A, polypeptide 1 pseudogene 2|cytochrome P450, subfamily 51 pseudogene 2', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28720864, end_position_on_the_genomic_accession=28724159, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373965, CurrentID=0, Status='live', Symbol='LOC105373965', Aliases='', description='uncharacterized LOC105373965', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239299121, end_position_on_the_genomic_accession=239303601, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83538, CurrentID=0, Status='live', Symbol='TTC25', Aliases='', description='tetratricopeptide repeat domain 25', other_designations='TPR repeat protein 25', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41930617, end_position_on_the_genomic_accession=41965651, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  83539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83539, CurrentID=0, Status='live', Symbol='CHST9', Aliases='GALNAC4ST-2, GalNAc4ST2', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 9', other_designations='GalNAc-4-sulfotransferase 2|N-acetylgalactosamine 4-O-sulfotransferase 2|galNAc-4-O-sulfotransferase 2', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26908754, end_position_on_the_genomic_accession=27185338, orientation='minus', exon_count=10, OMIM=610191, no_hdr0=''),
  83540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83540, CurrentID=0, Status='live', Symbol='NUF2', Aliases='CDCA1, CT106R, NUF2', description='NUF2, NDC80 kinetochore complex component', other_designations='NUF2, NDC80 kinetochore complex component, homolog|cancer/testis antigen 106|cell division cycle associated 1|cell division cycle-associated protein 1|hNuf2|hNuf2R|hsNuf2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163321933, end_position_on_the_genomic_accession=163355764, orientation='plus', exon_count=14, OMIM=611772, no_hdr0=''),
  83541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83541, CurrentID=0, Status='live', Symbol='FAM110A', Aliases='C20orf55, F10, bA371L19.3', description='family with sequence similarity 110 member A', other_designations='family with sequence similarity 110, member A', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=833697, end_position_on_the_genomic_accession=846279, orientation='plus', exon_count=7, OMIM=611393, no_hdr0=''),
  83543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83543, CurrentID=0, Status='live', Symbol='AIF1L', Aliases='C9orf58, IBA2', description='allograft inflammatory factor 1-like', other_designations='ionized calcium binding adapter molecule 2|ionized calcium-binding adapter molecule 2', map_location='9q34.13-q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131096476, end_position_on_the_genomic_accession=131123152, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  83544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83544, CurrentID=0, Status='live', Symbol='DNAL1', Aliases='C14orf168, CILD16', description='dynein, axonemal, light chain 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73644875, end_position_on_the_genomic_accession=73703728, orientation='plus', exon_count=10, OMIM=610062, no_hdr0=''),
  100873828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873828, CurrentID=0, Status='live', Symbol='RNU5E-3P', Aliases='', description='RNA, U5E small nuclear 3, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48574453, end_position_on_the_genomic_accession=48574520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83546, CurrentID=0, Status='live', Symbol='RTBDN', Aliases='', description='retbindin', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12825477, end_position_on_the_genomic_accession=12838267, orientation='minus', exon_count=14, OMIM=609553, no_hdr0=''),
  83547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83547, CurrentID=0, Status='live', Symbol='RILP', Aliases='PP10141', description='Rab interacting lysosomal protein', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1646150, end_position_on_the_genomic_accession=1650098, orientation='minus', exon_count=8, OMIM=607848, no_hdr0=''),
  83548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83548, CurrentID=0, Status='live', Symbol='COG3', Aliases='SEC34', description='component of oligomeric golgi complex 3', other_designations='COG complex subunit 3|p94|tethering factor SEC34|vesicle-docking protein SEC34 homolog', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45464895, end_position_on_the_genomic_accession=45536698, orientation='plus', exon_count=24, OMIM=606975, no_hdr0=''),
  83549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83549, CurrentID=0, Status='live', Symbol='UCK1', Aliases='URK1', description='uridine-cytidine kinase 1', other_designations='UCK 1|cytidine monophosphokinase 1|uridine monophosphokinase 1', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131523796, end_position_on_the_genomic_accession=131531275, orientation='minus', exon_count=8, OMIM=609328, no_hdr0=''),
  83550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83550, CurrentID=0, Status='live', Symbol='GPR101', Aliases='GPCR6, PAGH2', description='G protein-coupled receptor 101', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=137030148, end_position_on_the_genomic_accession=137031674, orientation='minus', exon_count=1, OMIM=300393, no_hdr0=''),
  83551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83551, CurrentID=0, Status='live', Symbol='TAAR8', Aliases='GPR102, TA5, TAR5, TRAR5, TaR-5, TaR-8', description='trace amine associated receptor 8', other_designations='G protein-coupled receptor 102|G-protein coupled receptor 102|trace amine receptor 5|trace amine receptor 8', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132552693, end_position_on_the_genomic_accession=132553721, orientation='plus', exon_count=1, OMIM=606927, no_hdr0=''),
  83552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83552, CurrentID=0, Status='live', Symbol='MFRP', Aliases='MCOP5, NNO2, RD6', description='membrane frizzled-related protein', other_designations='membrane-type frizzled-related protein', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119338934, end_position_on_the_genomic_accession=119346673, orientation='minus', exon_count=15, OMIM=606227, no_hdr0=''),
  100131089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131089, CurrentID=0, Status='live', Symbol='SRP14-AS1', Aliases='', description='SRP14 antisense RNA1 (head to head)', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40039311, end_position_on_the_genomic_accession=40067509, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873833, CurrentID=0, Status='live', Symbol='RNU5F-6P', Aliases='', description='RNA, U5F small nuclear 6, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164351273, end_position_on_the_genomic_accession=164351388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131092, CurrentID=0, Status='live', Symbol='ISCA1P3', Aliases='', description='iron-sulfur cluster assembly 1 pseudogene 3', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 3', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54477707, end_position_on_the_genomic_accession=54479680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373974, CurrentID=0, Status='live', Symbol='LOC105373974', Aliases='', description='uncharacterized LOC105373974', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241546658, end_position_on_the_genomic_accession=241549970, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  83590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83590, CurrentID=0, Status='live', Symbol='TMUB1', Aliases='C7orf21, DULP, SB144', description='transmembrane and ubiquitin-like domain containing 1', other_designations='dendritic cell-derived ubiquitin-like protein|hepatocyte odd protein shuttling protein|ubiquitin-like protein DULP|ubiquitin-like protein SB144', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151081085, end_position_on_the_genomic_accession=151083533, orientation='minus', exon_count=4, OMIM=614792, no_hdr0=''),
  83591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83591, CurrentID=0, Status='live', Symbol='THAP2', Aliases='', description='THAP domain containing, apoptosis associated protein 2', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71663897, end_position_on_the_genomic_accession=71680648, orientation='plus', exon_count=3, OMIM=612531, no_hdr0=''),
  83592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83592, CurrentID=0, Status='live', Symbol='AKR1E2', Aliases='AKR1CL2, AKRDC1, LoopADR, hTSP, htAKR', description='aldo-keto reductase family 1, member E2', other_designations='AF reductase|aldo-keto reductase family 1 member C-like protein 2|aldo-keto reductase family 1, member C-like 2|aldo-keto reductase loopADR|testis-specific protein', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4824560, end_position_on_the_genomic_accession=4854131, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  83593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83593, CurrentID=0, Status='live', Symbol='RASSF5', Aliases='Maxp1, NORE1, NORE1A, NORE1B, RAPL, RASSF3', description='Ras association (RalGDS/AF-6) domain family member 5', other_designations='Rap1-binding protein|Ras association (RalGDS/AF-6) domain family 5|Ras effector-like protein|new ras effector 1|novel Ras effector 1|regulator for cell adhesion and polarization enriched in lymphoid tissue|tumor suppressor RASSF3', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206507530, end_position_on_the_genomic_accession=206589284, orientation='plus', exon_count=7, OMIM=607020, no_hdr0=''),
  83594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83594, CurrentID=0, Status='live', Symbol='NUDT12', Aliases='', description='nudix hydrolase 12', other_designations='nucleoside diphosphate linked moiety X-type motif 12|nucleoside diphosphate-linked moiety X motif 12|nudix (nucleoside diphosphate linked moiety X)-type motif 12|nudix motif 12', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103548855, end_position_on_the_genomic_accession=103562804, orientation='minus', exon_count=7, OMIM=609232, no_hdr0=''),
  83595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83595, CurrentID=0, Status='live', Symbol='SOX7', Aliases='', description='SRY-box 7', other_designations='SRY (sex determining region Y)-box 7|SRY box 7', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10723768, end_position_on_the_genomic_accession=10730574, orientation='minus', exon_count=2, OMIM=612202, no_hdr0=''),
  83596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83596, CurrentID=0, Status='live', Symbol='BCL2L12', Aliases='', description='BCL2-like 12 (proline rich)', other_designations='Bcl-2 related proline-rich protein', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49665142, end_position_on_the_genomic_accession=49673916, orientation='plus', exon_count=7, OMIM=610837, no_hdr0=''),
  83597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83597, CurrentID=0, Status='live', Symbol='RTP3', Aliases='LTM1, TMEM7, Z3CXXC3', description='receptor (chemosensory) transporter protein 3', other_designations='3CxxC-type zinc finger protein 3|receptor transporter protein 3|receptor transporting protein 3|transmembrane protein 7|zinc finger, 3CxxC-type 3', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46497995, end_position_on_the_genomic_accession=46500949, orientation='plus', exon_count=2, OMIM=607181, no_hdr0=''),
  83598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83598, CurrentID=0, Status='live', Symbol='LRRC2-AS1', Aliases='LUZP3, LUZP3P, LUZPP1', description='LRRC2 antisense RNA 1', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46557398, end_position_on_the_genomic_accession=46559688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873837, CurrentID=0, Status='live', Symbol='RNU5E-8P', Aliases='', description='RNA, U5E small nuclear 8, pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116965112, end_position_on_the_genomic_accession=116965227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373976, CurrentID=0, Status='live', Symbol='LOC105373976', Aliases='', description='uncharacterized LOC105373976', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241769784, end_position_on_the_genomic_accession=241772524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83604, CurrentID=0, Status='live', Symbol='TMEM47', Aliases='BCMP1, TM4SF10', description='transmembrane protein 47', other_designations='brain cell membrane protein 1|transmembrane 4 superfamily member 10', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=34627064, end_position_on_the_genomic_accession=34657288, orientation='minus', exon_count=3, OMIM=300698, no_hdr0=''),
  83605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83605, CurrentID=0, Status='live', Symbol='CCM2', Aliases='C7orf22, OSM, PP10187', description='cerebral cavernous malformation 2', other_designations='malcavernin', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44999746, end_position_on_the_genomic_accession=45076470, orientation='plus', exon_count=13, OMIM=607929, no_hdr0=''),
  83606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83606, CurrentID=0, Status='live', Symbol='GUCD1', Aliases='C22orf13, LLN4', description='guanylyl cyclase domain containing 1', other_designations='CG13760 gene product [Drosophila melanogaster] homolog|guanylyl cyclase domain-containing protein 1', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24540423, end_position_on_the_genomic_accession=24555935, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83607, CurrentID=0, Status='live', Symbol='AMMECR1L', Aliases='', description='AMMECR1 like', other_designations='AMME chromosomal region gene 1-like|AMMECR1-like', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127861630, end_position_on_the_genomic_accession=127888570, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  83608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83608, CurrentID=0, Status='live', Symbol='C18orf21', Aliases='HsT3108, PNAS-124, PNAS-131, XTP13', description='chromosome 18 open reading frame 21', other_designations='HBV X-transactivated gene 13 protein|HBV X-transactivated protein 13|HBV XAg-transactivated protein 13', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35972625, end_position_on_the_genomic_accession=35979287, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  345757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345757, CurrentID=0, Status='live', Symbol='FAM174A', Aliases='HGS_RE408, TMEM157, UNQ1912', description='family with sequence similarity 174 member A', other_designations='HCV NS5A-transactivated protein 6|family with sequence similarity 174, member A|hepatitis C virus NS5A-transactivated protein 6|transmembrane protein 157', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100535305, end_position_on_the_genomic_accession=100586741, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373979, CurrentID=0, Status='live', Symbol='LOC105373979', Aliases='', description='uncharacterized LOC105373979', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242047626, end_position_on_the_genomic_accession=242084592, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100873841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873841, CurrentID=0, Status='live', Symbol='RNU5D-2P', Aliases='', description='RNA, U5D small nuclear 2, pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132421641, end_position_on_the_genomic_accession=132421690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375398, CurrentID=0, Status='live', Symbol='LOC105375398', Aliases='', description='uncharacterized LOC105375398', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92651141, end_position_on_the_genomic_accession=92659435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  345778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345778, CurrentID=0, Status='live', Symbol='MTX3', Aliases='', description='metaxin 3', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79976716, end_position_on_the_genomic_accession=79991265, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83636, CurrentID=0, Status='live', Symbol='C19orf12', Aliases='NBIA3, NBIA4, SPG43', description='chromosome 19 open reading frame 12', other_designations='neurodegeneration with brain iron accumulation 3', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29698886, end_position_on_the_genomic_accession=29715789, orientation='minus', exon_count=6, OMIM=614297, no_hdr0=''),
  83637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83637, CurrentID=0, Status='live', Symbol='ZMIZ2', Aliases='NET27, TRAFIP20, ZIMP7, hZIMP7', description='zinc finger, MIZ-type containing 2', other_designations='PIAS-like protein Zimp7', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44748547, end_position_on_the_genomic_accession=44769881, orientation='plus', exon_count=24, OMIM=611196, no_hdr0=''),
  83638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83638, CurrentID=0, Status='live', Symbol='C11orf68', Aliases='BLES03, P5326', description='chromosome 11 open reading frame 68', other_designations='basophilic leukemia expressed protein BLES03|basophilic leukemia-expressed protein Bles03|protein p5326', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65916810, end_position_on_the_genomic_accession=65919060, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  83639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83639, CurrentID=0, Status='live', Symbol='TEX101', Aliases='CT131, GTPR867, NYD-SP8, PRO1884, SGRG, SPATA44, TES101RP', description='testis expressed 101', other_designations='cancer/testis antigen 131|cell surface receptor NYD-SP8|scleroderma-associated autoantigen|spermatogenesis associated 44|spermatogenesis-related gene protein|testis expressed sequence 101|testis-expressed protein 101|testis-specific protein TES101RP', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43388611, end_position_on_the_genomic_accession=43418615, orientation='plus', exon_count=10, OMIM=612665, no_hdr0=''),
  83640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83640, CurrentID=0, Status='live', Symbol='FAM103A1', Aliases='C15orf18, HsT19360, RAM', description='family with sequence similarity 103 member A1', other_designations='family with sequence similarity 103, member A1|protein FAM103A1', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82986203, end_position_on_the_genomic_accession=82991057, orientation='plus', exon_count=4, OMIM=614547, no_hdr0=''),
  83641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83641, CurrentID=0, Status='live', Symbol='FAM107B', Aliases='C10orf45', description='family with sequence similarity 107 member B', other_designations='family with sequence similarity 107, member B', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14518557, end_position_on_the_genomic_accession=14774897, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  83642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83642, CurrentID=0, Status='live', Symbol='SELO', Aliases='', description='selenoprotein O', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50200979, end_position_on_the_genomic_accession=50217616, orientation='plus', exon_count=9, OMIM=607917, no_hdr0=''),
  83643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83643, CurrentID=0, Status='live', Symbol='CCDC3', Aliases='', description='coiled-coil domain containing 3', other_designations='fat/vessel-derived secretory protein|favine', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12896625, end_position_on_the_genomic_accession=13099773, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  83647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83647, CurrentID=0, Status='live', Symbol='LINC00529', Aliases='C8orf8', description='long intergenic non-protein coding RNA 529', other_designations='', map_location='8p23-p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  83648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83648, CurrentID=0, Status='live', Symbol='FAM167A', Aliases='C8orf13, D8S265', description='family with sequence similarity 167 member A', other_designations='family with sequence similarity 167, member A', map_location='8p23-p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11421464, end_position_on_the_genomic_accession=11475915, orientation='minus', exon_count=9, OMIM=610085, no_hdr0=''),
  105373984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373984, CurrentID=0, Status='live', Symbol='LOC105373984', Aliases='', description='anther-specific proline-rich protein APG-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105142926, end_position_on_the_genomic_accession=105145520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83650, CurrentID=0, Status='live', Symbol='SLC35G5', Aliases='AMAC, AMAC1L2', description='solute carrier family 35 member G5', other_designations='acyl-malonyl condensing enzyme 1-like 2|acyl-malonyl-condensing enzyme 1-like protein 2|protein AMAC1L2|solute carrier family 35, member G5', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11330986, end_position_on_the_genomic_accession=11332186, orientation='plus', exon_count=1, OMIM=615199, no_hdr0=''),
  83655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83655, CurrentID=0, Status='live', Symbol='LINC00208', Aliases='C8orf14, NCRNA00208, VIR35', description='long intergenic non-protein coding RNA 208', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11576535, end_position_on_the_genomic_accession=11581341, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83656, CurrentID=0, Status='live', Symbol='FAM167A-AS1', Aliases='C8orf12', description='FAM167A antisense RNA 1', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11368402, end_position_on_the_genomic_accession=11438657, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  83657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83657, CurrentID=0, Status='live', Symbol='DYNLRB2', Aliases='DNCL2B, DNLC2B, ROBLD2', description='dynein, light chain, roadblock-type 2', other_designations='bithoraxoid-like protein|dynein light chain 2B, cytoplasmic|dynein, cytoplasmic, light polypeptide 2B|roadblock domain-containing protein 2', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80540258, end_position_on_the_genomic_accession=80550811, orientation='plus', exon_count=9, OMIM=607168, no_hdr0=''),
  83658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83658, CurrentID=0, Status='live', Symbol='DYNLRB1', Aliases='BITH, BLP, DNCL2A, DNLC2A, ROBLD1', description='dynein, light chain, roadblock-type 1', other_designations='ROBL/LC7-like 1|bithoraxoid-like protein|dynein, cytoplasmic, light polypeptide 2A|dynein-associated protein Km23|roadblock domain-containing protein 1', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34516355, end_position_on_the_genomic_accession=34540958, orientation='plus', exon_count=7, OMIM=607167, no_hdr0=''),
  83659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83659, CurrentID=0, Status='live', Symbol='TEKT1', Aliases='', description='tektin 1', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6797993, end_position_on_the_genomic_accession=6831741, orientation='minus', exon_count=8, OMIM=609002, no_hdr0=''),
  83660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83660, CurrentID=0, Status='live', Symbol='TLN2', Aliases='ILWEQ', description='talin 2', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62390575, end_position_on_the_genomic_accession=62844631, orientation='plus', exon_count=67, OMIM=607349, no_hdr0=''),
  83661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83661, CurrentID=0, Status='live', Symbol='MS4A8', Aliases='4SPAN4, CD20L5, MS4A4B, MS4A8', description='membrane-spanning 4-domains, subfamily A, member 8', other_designations='four-span transmembrane protein 4|membrane-spanning 4-domains, subfamily A, member 8B', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60699574, end_position_on_the_genomic_accession=60715811, orientation='plus', exon_count=7, OMIM=606549, no_hdr0=''),
  83666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83666, CurrentID=0, Status='live', Symbol='PARP9', Aliases='ARTD9, BAL, BAL1, MGC:7868', description='poly(ADP-ribose) polymerase family member 9', other_designations='ADP-ribosyltransferase diphtheria toxin-like 9|PARP-9|b aggressive lymphoma protein|poly (ADP-ribose) polymerase 9|poly (ADP-ribose) polymerase family, member 9', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122527910, end_position_on_the_genomic_accession=122564676, orientation='minus', exon_count=14, OMIM=612065, no_hdr0=''),
  83667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83667, CurrentID=0, Status='live', Symbol='SESN2', Aliases='HI95, SES2, SEST2', description='sestrin 2', other_designations='hypoxia induced gene 95', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28259452, end_position_on_the_genomic_accession=28282491, orientation='plus', exon_count=10, OMIM=607767, no_hdr0=''),
  574538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=574538, CurrentID=0, Status='live', Symbol='LOC574538', Aliases='', description='uncharacterized LOC574538', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=137411, end_position_on_the_genomic_accession=149169, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100271267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271267, CurrentID=0, Status='live', Symbol='RPSAP51', Aliases='RPSA_25_1205', description='ribosomal protein SA pseudogene 51', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8742351, end_position_on_the_genomic_accession=8743076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100873850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873850, CurrentID=0, Status='live', Symbol='RN7SKP6', Aliases='', description='RNA, 7SK small nuclear pseudogene 6', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=56885284, end_position_on_the_genomic_accession=56885602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79483, CurrentID=0, Status='live', Symbol='OR5BD1P', Aliases='', description='olfactory receptor family 5 subfamily BD member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BD, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57945497, end_position_on_the_genomic_accession=57946523, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  345829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345829, CurrentID=0, Status='live', Symbol='BTF3P10', Aliases='RAET1J', description='basic transcription factor 3 pseudogene 10', other_designations='basic transcription factor 3, pseudogene 6|retinoic acid early transcript 1J pseudogene', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149977826, end_position_on_the_genomic_accession=149978689, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130331, CurrentID=0, Status='live', Symbol='LOC100130331', Aliases='', description='POTE ankyrin domain family, member F pseudogene', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237862175, end_position_on_the_genomic_accession=237928321, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  83690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83690, CurrentID=0, Status='live', Symbol='CRISPLD1', Aliases='CRISP-10, CRISP10, LCRISP1', description='cysteine-rich secretory protein LCCL domain containing 1', other_designations='CocoaCrisp|LCCL domain containing cysteine-rich secretory protein 1|LCCL domain-containing cysteine-rich secretory protein 1|cysteine-rich secretory protein 10|trypsin inhibitor Hl', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74984473, end_position_on_the_genomic_accession=75034558, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105375400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375400, CurrentID=0, Status='live', Symbol='LOC105375400', Aliases='', description='uncharacterized LOC105375400', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93533247, end_position_on_the_genomic_accession=93559530, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  83692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83692, CurrentID=0, Status='live', Symbol='CD99L2', Aliases='CD99B, MIC2L1', description='CD99 molecule like 2', other_designations='CD99 molecule-like 2|MIC2 like 1|MIC2-like protein 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150766336, end_position_on_the_genomic_accession=150898816, orientation='minus', exon_count=13, OMIM=300846, no_hdr0=''),
  83693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83693, CurrentID=0, Status='live', Symbol='HSDL1', Aliases='SDR12C3', description='hydroxysteroid dehydrogenase like 1', other_designations='short chain dehydrogenase/reductase family 12C member 3|short chain dehydrogenase/reductase family 12C, member 3|steroid dehydrogenase-like protein', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84122139, end_position_on_the_genomic_accession=84145195, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83694, CurrentID=0, Status='live', Symbol='RPS6KL1', Aliases='RSKL2', description='ribosomal protein S6 kinase like 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74902849, end_position_on_the_genomic_accession=74923282, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  83695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83695, CurrentID=0, Status='live', Symbol='RHNO1', Aliases='C12orf32, HKMT1188, RHINO', description='RAD9-HUS1-RAD1 interacting nuclear orphan 1', other_designations='Rad9, Rad1, Hus1 interacting nuclear orphan|rAD9, RAD1, HUS1-interacting nuclear orphan protein', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2876258, end_position_on_the_genomic_accession=2889525, orientation='plus', exon_count=5, OMIM=614085, no_hdr0=''),
  83696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83696, CurrentID=0, Status='live', Symbol='TRAPPC9', Aliases='IBP, IKBKBBP, MRT13, NIBP, T1, TRS120', description='trafficking protein particle complex 9', other_designations='IKK2 binding protein|NIK and IKK-beta binding protein|NIK- and IKBKB-binding protein|TRAPP 120 kDa subunit|tularik gene 1 protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=139728081, end_position_on_the_genomic_accession=140458579, orientation='minus', exon_count=30, OMIM=611966, no_hdr0=''),
  83697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83697, CurrentID=0, Status='live', Symbol='SLC4A9', Aliases='AE4', description='solute carrier family 4, sodium bicarbonate cotransporter, member 9', other_designations='anion exchanger 4|sodium bicarbonate cotransporter 5|solute carrier family 4, sodium bicarbonate cotransporter-like, member 9', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140358559, end_position_on_the_genomic_accession=140375137, orientation='plus', exon_count=24, OMIM=610207, no_hdr0=''),
  83698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83698, CurrentID=0, Status='live', Symbol='CALN1', Aliases='CABP8', description='calneuron 1', other_designations='calcium-binding protein CABP8|calneuron I|calneuron-1', map_location='7q11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71779491, end_position_on_the_genomic_accession=72447151, orientation='minus', exon_count=14, OMIM=607176, no_hdr0=''),
  83699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83699, CurrentID=0, Status='live', Symbol='SH3BGRL2', Aliases='', description='SH3 domain binding glutamate-rich protein like 2', other_designations='SH3 domain binding glutamic acid-rich protein like 2|fovea-associated SH3 domain-binding protein', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79631283, end_position_on_the_genomic_accession=79703670, orientation='plus', exon_count=5, OMIM=615678, no_hdr0=''),
  83700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83700, CurrentID=0, Status='live', Symbol='JAM3', Aliases='JAM-2, JAM-3, JAM-C, JAMC', description='junctional adhesion molecule 3', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134068925, end_position_on_the_genomic_accession=134151757, orientation='plus', exon_count=9, OMIM=606871, no_hdr0=''),
  83706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83706, CurrentID=0, Status='live', Symbol='FERMT3', Aliases='KIND3, MIG-2, MIG2B, UNC112C, URP2, URP2SF', description='fermitin family member 3', other_designations='MIG2-like protein|UNC-112 related protein 2|kindlin 3|unc-112-related protein 2', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64206655, end_position_on_the_genomic_accession=64223891, orientation='plus', exon_count=16, OMIM=607901, no_hdr0=''),
  83707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83707, CurrentID=0, Status='live', Symbol='TRPT1', Aliases='', description='tRNA phosphotransferase 1', other_designations="tRNA splicing 2' phosphotransferase 1", map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64223799, end_position_on_the_genomic_accession=64226254, orientation='minus', exon_count=8, OMIM=610470, no_hdr0=''),
  100131114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131114, CurrentID=0, Status='live', Symbol='LOC100131114', Aliases='', description='serine/arginine-rich splicing factor 3 pseudogene', other_designations='', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106253219, end_position_on_the_genomic_accession=106253825, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79488, CurrentID=0, Status='live', Symbol='OR10D5P', Aliases='', description='olfactory receptor family 10 subfamily D member 5 pseudogene', other_designations='olfactory receptor, family 10, subfamily D, member 5 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124054793, end_position_on_the_genomic_accession=124055726, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83714, CurrentID=0, Status='live', Symbol='NRIP2', Aliases='', description='nuclear receptor interacting protein 2', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2824882, end_position_on_the_genomic_accession=2835536, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83715, CurrentID=0, Status='live', Symbol='ESPN', Aliases='DFNB36, LP2654', description='espin', other_designations='autosomal recessive deafness type 36 protein|ectoplasmic specialization protein', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6424776, end_position_on_the_genomic_accession=6460944, orientation='plus', exon_count=21, OMIM=606351, no_hdr0=''),
  83716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83716, CurrentID=0, Status='live', Symbol='CRISPLD2', Aliases='CRISP11, LCRISP2', description='cysteine-rich secretory protein LCCL domain containing 2', other_designations='CRISP-11|LCCL domain containing cysteine-rich secretory protein 2|LCCL domain-containing cysteine-rich secretory protein 2|cysteine-rich secretory protein 11|trypsin inhibitor', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84819981, end_position_on_the_genomic_accession=84909510, orientation='plus', exon_count=16, OMIM=612434, no_hdr0=''),
  83719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83719, CurrentID=0, Status='live', Symbol='YPEL3', Aliases='', description='yippee like 3', other_designations='yippee-like 3', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30092314, end_position_on_the_genomic_accession=30096216, orientation='minus', exon_count=7, OMIM=609724, no_hdr0=''),
  83723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83723, CurrentID=0, Status='live', Symbol='FAM57B', Aliases='FP1188', description='family with sequence similarity 57 member B', other_designations='family with sequence similarity 57, member B', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30024427, end_position_on_the_genomic_accession=30053012, orientation='minus', exon_count=13, OMIM=615175, no_hdr0=''),
  100873858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873858, CurrentID=0, Status='live', Symbol='SDAD1P3', Aliases='', description='SDA1 domain containing 1 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33771963, end_position_on_the_genomic_accession=33772295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373997, CurrentID=0, Status='live', Symbol='LOC105373997', Aliases='', description='uncharacterized LOC105373997', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98191574, end_position_on_the_genomic_accession=98196523, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83729, CurrentID=0, Status='live', Symbol='INHBE', Aliases='', description='inhibin beta E', other_designations='activin beta E|activin beta-E chain|inhibin, beta E', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57455313, end_position_on_the_genomic_accession=57458008, orientation='plus', exon_count=2, OMIM=612031, no_hdr0=''),
  79491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79491, CurrentID=0, Status='live', Symbol='OR5BM1P', Aliases='', description='olfactory receptor family 5 subfamily BM member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BM, member 1 pseudogene', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98052893, end_position_on_the_genomic_accession=98053818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83732, CurrentID=0, Status='live', Symbol='RIOK1', Aliases='AD034, RRP10, bA288G3.1', description='RIO kinase 1', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7389419, end_position_on_the_genomic_accession=7418037, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  83733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83733, CurrentID=0, Status='live', Symbol='SLC25A18', Aliases='GC2', description='solute carrier family 25 (glutamate carrier), member 18', other_designations='glutamate/H(+) symporter 2|solute carrier family 25 (mitochondrial carrier), member 18', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17563165, end_position_on_the_genomic_accession=17590890, orientation='plus', exon_count=12, OMIM=609303, no_hdr0=''),
  83734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83734, CurrentID=0, Status='live', Symbol='ATG10', Aliases='APG10, APG10L, pp12616', description='autophagy related 10', other_designations='ATG10 autophagy related 10 homolog|autophagy-related protein 10', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81972025, end_position_on_the_genomic_accession=82258502, orientation='plus', exon_count=13, OMIM=610800, no_hdr0=''),
  83737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83737, CurrentID=0, Status='live', Symbol='ITCH', Aliases='ADMFD, AIF4, AIP4, NAPP1', description='itchy E3 ubiquitin protein ligase', other_designations='NFE2-associated polypeptide 1|atrophin-1 interacting protein 4|itchy E3 ubiquitin protein ligase homolog', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34363235, end_position_on_the_genomic_accession=34511393, orientation='plus', exon_count=28, OMIM=606409, no_hdr0=''),
  83740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83740, CurrentID=0, Status='live', Symbol='H2AFB3', Aliases='H2ABBD, H2AFB', description='H2A histone family member B3', other_designations='H2A Barr body-deficient|H2A histone family, member B3|H2A.Bbd|histone variant H2A, Barr-body deficient', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155459419, end_position_on_the_genomic_accession=155459935, orientation='minus', exon_count=1, OMIM=300445, no_hdr0=''),
  83741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83741, CurrentID=0, Status='live', Symbol='TFAP2D', Aliases='TFAP2BL1', description='transcription factor AP-2 delta (activating enhancer binding protein 2 delta)', other_designations='AP-2 like|AP2-delta|activating enhancer binding protein 2 beta-like 1|activating enhancer-binding protein 2-delta|transcription factor AP-2 beta (activating enhancer binding protein 2 beta)-like 1|transcription factor AP-2 beta (activating enhancer-binding protein 2 beta)-like 1|transcription factor AP-2-beta-like 1', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50713544, end_position_on_the_genomic_accession=50773033, orientation='plus', exon_count=8, OMIM=610161, no_hdr0=''),
  83742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83742, CurrentID=0, Status='live', Symbol='MARVELD1', Aliases='GB14, MARVD1, MRVLDC1, bA548K23.8', description='MARVEL domain containing 1', other_designations='MARVEL (membrane-associating) domain containing 1|putative MARVEL domain-containing protein 1', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97713708, end_position_on_the_genomic_accession=97718152, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83743, CurrentID=0, Status='live', Symbol='GRWD1', Aliases='CDW4, GRWD, RRB1, WDR28', description='glutamate-rich WD repeat containing 1', other_designations='CUL4- and DDB1-associated WDR protein 4|WD repeat domain 28|glutamate rich WD repeat protein GRWD|regulator of ribosome biogenesis 1 homolog', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48445773, end_position_on_the_genomic_accession=48453907, orientation='plus', exon_count=7, OMIM=610597, no_hdr0=''),
  83744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83744, CurrentID=0, Status='live', Symbol='ZNF484', Aliases='BA526D8.4', description='zinc finger protein 484', other_designations='KRAB box containing C2H2 type zinc finger bA526D8.4', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92842573, end_position_on_the_genomic_accession=92878058, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100131120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131120, CurrentID=0, Status='live', Symbol='LOC100131120', Aliases='', description='PRO1828', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=137560313, end_position_on_the_genomic_accession=137562138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83746, CurrentID=0, Status='live', Symbol='L3MBTL2', Aliases='H-l(3)mbt-l, L3MBT', description='l(3)mbt-like 2 (Drosophila)', other_designations='H-l(3)mbt-like protein 2|L(3)mbt-like protein 2|lmbt-like 2 (Drosophila) isoform 1|lmbt-like 2 (Drosophila) isoform 2', map_location='22q13.31-q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41205232, end_position_on_the_genomic_accession=41231272, orientation='plus', exon_count=17, OMIM=611865, no_hdr0=''),
  83747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83747, CurrentID=0, Status='live', Symbol='C20orf57', Aliases='bA243J16.5', description='chromosome 20 open reading frame 57', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31847644, end_position_on_the_genomic_accession=31861400, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  79494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79494, CurrentID=0, Status='live', Symbol='OR8K2P', Aliases='', description='olfactory receptor family 8 subfamily K member 2 pseudogene', other_designations='olfactory receptor, family 8, subfamily K, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56335143, end_position_on_the_genomic_accession=56336069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  345895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345895, CurrentID=0, Status='live', Symbol='RSPH4A', Aliases='CILD11, RSHL3, RSPH6B, dJ412I7.1', description='radial spoke head 4 homolog A (Chlamydomonas)', other_designations='radial spoke head-like protein 3|radial spokehead-like 3', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116593900, end_position_on_the_genomic_accession=116633051, orientation='plus', exon_count=9, OMIM=612647, no_hdr0=''),
  83752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83752, CurrentID=0, Status='live', Symbol='LONP2', Aliases='LONP, LONPL', description='lon peptidase 2, peroxisomal', other_designations='lon protease 2|lon protease-like protein 2|peroxisomal LON protease like|peroxisomal Lon protease homolog 2', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48244167, end_position_on_the_genomic_accession=48353979, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  83755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83755, CurrentID=0, Status='live', Symbol='KRTAP4-12', Aliases='KAP4.12, KRTAP4.12', description='keratin associated protein 4-12', other_designations='keratin-associated protein 4.12|ultrahigh sulfur keratin-associated protein 4.12', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41123091, end_position_on_the_genomic_accession=41124167, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83756, CurrentID=0, Status='live', Symbol='TAS1R3', Aliases='T1R3', description='taste 1 receptor member 3', other_designations='sweet taste receptor T1R3|taste receptor, type 1, member 3', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1328999, end_position_on_the_genomic_accession=1335320, orientation='plus', exon_count=6, OMIM=605865, no_hdr0=''),
  83757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83757, CurrentID=0, Status='live', Symbol='CYP2G2P', Aliases='CYP2G2, CYP2GP2', description='cytochrome P450 family 2 subfamily G member 2, pseudogene', other_designations='cytochrome P450, family 2, subfamily G, polypeptide 2 pseudogene|cytochrome P450, subfamily IIG, polypeptide 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41050452, end_position_on_the_genomic_accession=41062444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  83758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83758, CurrentID=0, Status='live', Symbol='RBP5', Aliases='CRBP-III, CRBP3, CRBPIII', description='retinol binding protein 5', other_designations='HRBPiso|cellular retinol-binding protein III|putative cellular retinol-binding protein CRBP III|retinol binding protein 5, cellular|retinol-binding protein 5, cellular', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7119074, end_position_on_the_genomic_accession=7130236, orientation='minus', exon_count=7, OMIM=611866, no_hdr0=''),
  83759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83759, CurrentID=0, Status='live', Symbol='RBM4B', Aliases='RBM30, RBM4L, ZCCHC15, ZCCHC21B, ZCRB3B', description='RNA binding motif protein 4B', other_designations='RNA-binding motif protein 30|zinc finger CCHC-type and RNA binding motif 3B', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66664994, end_position_on_the_genomic_accession=66677921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100873864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873864, CurrentID=0, Status='live', Symbol='LAP3P1', Aliases='', description='leucine aminopeptidase 3 pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82924827, end_position_on_the_genomic_accession=82925122, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374004, CurrentID=0, Status='live', Symbol='LOC105374004', Aliases='', description='uncharacterized LOC105374004', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99329838, end_position_on_the_genomic_accession=99337655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79498, CurrentID=0, Status='live', Symbol='OR8I1P', Aliases='', description='olfactory receptor family 8 subfamily I member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily I, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56296540, end_position_on_the_genomic_accession=56297470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79499, CurrentID=0, Status='live', Symbol='OR7L1P', Aliases='', description='olfactory receptor family 7 subfamily L member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily L, member 1 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131596122, end_position_on_the_genomic_accession=131597204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131126, CurrentID=0, Status='live', Symbol='LOC100131126', Aliases='', description='leucine rich repeat containing 59 pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78055827, end_position_on_the_genomic_accession=78056455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131604, CurrentID=0, Status='live', Symbol='LOC100131604', Aliases='', description='uncharacterized LOC100131604', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=162638018, end_position_on_the_genomic_accession=162639995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  345930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=345930, CurrentID=0, Status='live', Symbol='ECT2L', Aliases='ARHGEF32, C6orf91, FBXO49, LFDH, dJ509I19.2, dJ509I19.3, dJ509I19.5', description='epithelial cell transforming 2 like', other_designations='ECT2-like|F-box protein 49|epithelial cell transforming sequence 2 oncogene-like|lung specific F-box and DH domain containing protein|lung-specific F-box and DH domain-containing protein|putative guanine nucleotide exchange factor LFDH', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138796051, end_position_on_the_genomic_accession=138904070, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  83787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83787, CurrentID=0, Status='live', Symbol='ARMC10', Aliases='PNAS-112, PNAS112, PSEC0198, SVH', description='armadillo repeat containing 10', other_designations='specific splicing variant involved in hepatocarcinogenesis|splicing variant involved in hepatocarcinogenesis protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103074727, end_position_on_the_genomic_accession=103099763, orientation='plus', exon_count=9, OMIM=611864, no_hdr0=''),
  100873869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873869, CurrentID=0, Status='live', Symbol='RNU5E-5P', Aliases='', description='RNA, U5E small nuclear 5, pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101466605, end_position_on_the_genomic_accession=101466920, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83795, CurrentID=0, Status='live', Symbol='KCNK16', Aliases='K2p16.1, TALK-1, TALK1', description='potassium channel, two pore domain subfamily K, member 16', other_designations='2P domain potassium channel Talk-1|TWIK-related alkaline pH-activated K(+) channel 1|pancreatic potassium channel Talk-1|potassium channel, subfamily K, member 16', map_location='6p21.2-p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39314698, end_position_on_the_genomic_accession=39322985, orientation='minus', exon_count=8, OMIM=607369, no_hdr0=''),
  105371039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371039, CurrentID=0, Status='live', Symbol='LOC105371039', Aliases='', description='uncharacterized LOC105371039', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=664986, end_position_on_the_genomic_accession=665575, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421673, CurrentID=0, Status='live', Symbol='LOC100421673', Aliases='', description='small nuclear ribonucleoprotein polypeptide B pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13909232, end_position_on_the_genomic_accession=13909784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374009, CurrentID=0, Status='live', Symbol='LOC105374009', Aliases='', description='uncharacterized LOC105374009', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100120058, end_position_on_the_genomic_accession=100160369, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372486, CurrentID=0, Status='live', Symbol='LOC105372486', Aliases='', description='uncharacterized LOC105372486', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58514264, end_position_on_the_genomic_accession=58516426, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374010, CurrentID=0, Status='live', Symbol='LOC105374010', Aliases='', description='uncharacterized LOC105374010', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100023272, end_position_on_the_genomic_accession=100033236, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79504, CurrentID=0, Status='live', Symbol='OR4G4P', Aliases='', description='olfactory receptor family 4 subfamily G member 4 pseudogene', other_designations='olfactory receptor, family 4, subfamily G, member 4 pseudogene', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52453, end_position_on_the_genomic_accession=53396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374011, CurrentID=0, Status='live', Symbol='LOC105374011', Aliases='', description='uncharacterized LOC105374011', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99835482, end_position_on_the_genomic_accession=99855814, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374012, CurrentID=0, Status='live', Symbol='LOC105374012', Aliases='', description='uncharacterized LOC105374012', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100765962, end_position_on_the_genomic_accession=100792684, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371040, CurrentID=0, Status='live', Symbol='LOC105371040', Aliases='', description='serine/arginine repetitive matrix protein 2-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=902230, end_position_on_the_genomic_accession=904612, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  79508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79508, CurrentID=0, Status='live', Symbol='OR5M7P', Aliases='', description='olfactory receptor family 5 subfamily M member 7 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 7 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56500299, end_position_on_the_genomic_accession=56501270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374015, CurrentID=0, Status='live', Symbol='LOC105374015', Aliases='', description='uncharacterized LOC105374015', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102573964, end_position_on_the_genomic_accession=102602123, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374016, CurrentID=0, Status='live', Symbol='LOC105374016', Aliases='', description='uncharacterized LOC105374016', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102660879, end_position_on_the_genomic_accession=102710772, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  83844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83844, CurrentID=0, Status='live', Symbol='USP26', Aliases='', description='ubiquitin specific peptidase 26', other_designations='deubiquitinating enzyme 26|ubiquitin specific protease 26|ubiquitin thioesterase 26|ubiquitin thiolesterase 26|ubiquitin-specific processing protease 26|ubiquitin-specific-processing protease 26', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133090718, end_position_on_the_genomic_accession=133097304, orientation='minus', exon_count=5, OMIM=300309, no_hdr0=''),
  105374017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374017, CurrentID=0, Status='live', Symbol='LOC105374017', Aliases='', description='uncharacterized LOC105374017', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=102710844, end_position_on_the_genomic_accession=102727620, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  83849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83849, CurrentID=0, Status='live', Symbol='SYT15', Aliases='CHR10SYT, sytXV', description='synaptotagmin 15', other_designations='chr10 synaptotagmin|synaptotagmin XV|synaptotagmin XV-a', map_location='10q11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46578434, end_position_on_the_genomic_accession=46595194, orientation='plus', exon_count=12, OMIM=608081, no_hdr0=''),
  83850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83850, CurrentID=0, Status='live', Symbol='ESYT3', Aliases='CHR3SYT, E-Syt3, FAM62C', description='extended synaptotagmin-like protein 3', other_designations='chr3 synaptotagmin|family with sequence similarity 62 (C2 domain containing), member C', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138434573, end_position_on_the_genomic_accession=138479942, orientation='plus', exon_count=25, OMIM=616692, no_hdr0=''),
  83851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83851, CurrentID=0, Status='live', Symbol='SYT16', Aliases='CHR14SYT, SYT14L, Strep14, syt14r, yt14r', description='synaptotagmin 16', other_designations='CTD-2277K2.1|chr14 synaptotagmin|synaptotagmin 14-like protein|synaptotagmin XIV-related protein|synaptotagmin XVI', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61812729, end_position_on_the_genomic_accession=62117043, orientation='plus', exon_count=22, OMIM=610950, no_hdr0=''),
  83852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83852, CurrentID=0, Status='live', Symbol='SETDB2', Aliases='C13orf4, CLLD8, CLLL8, KMT1F', description='SET domain, bifurcated 2', other_designations='chronic lymphocytic leukemia deletion region 8|chronic lymphocytic leukemia deletion region gene 8 protein|lysine N-methyltransferase 1F', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49444056, end_position_on_the_genomic_accession=49495003, orientation='plus', exon_count=17, OMIM=607865, no_hdr0=''),
  83853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83853, CurrentID=0, Status='live', Symbol='ROPN1L', Aliases='ASP, RSPH11', description='rhophilin associated tail protein 1-like', other_designations='AKAP-associated sperm protein|ROPN1-like protein|radial spoke head 11 homolog', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10441862, end_position_on_the_genomic_accession=10482298, orientation='plus', exon_count=8, OMIM=611756, no_hdr0=''),
  83854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83854, CurrentID=0, Status='live', Symbol='ANGPTL6', Aliases='AGF, ARP5', description='angiopoietin like 6', other_designations='angiopoietin-like 6|angiopoietin-like protein 6|angiopoietin-related growth factor|angiopoietin-related protein 5', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10092337, end_position_on_the_genomic_accession=10106209, orientation='minus', exon_count=9, OMIM=609336, no_hdr0=''),
  83855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83855, CurrentID=0, Status='live', Symbol='KLF16', Aliases='BTEB4, DRRF, NSLP2', description='Kruppel-like factor 16', other_designations='BTE-binding protein 4|basic transcription element binding protein 4|basic transcription element-binding protein 4|dopamine receptor regulating factor|novel Sp1-like zinc finger transcription factor 2|transcription factor BTEB4|transcription factor NSLP2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1852399, end_position_on_the_genomic_accession=1863565, orientation='minus', exon_count=2, OMIM=606139, no_hdr0=''),
  83856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83856, CurrentID=0, Status='live', Symbol='FSD1L', Aliases='CCDC10, CSDUFD1, FSD1CL, FSD1NL, MIR1', description='fibronectin type III and SPRY domain containing 1-like', other_designations='FSD1 C-terminal like|FSD1 N-terminal like|coiled-coil domain containing 10|cystatin and DUF19 domain containing 1', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105442242, end_position_on_the_genomic_accession=105552433, orientation='plus', exon_count=18, OMIM=609829, no_hdr0=''),
  83857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83857, CurrentID=0, Status='live', Symbol='TMTC1', Aliases='ARG99, OLFA, TMTC1', description='transmembrane and tetratricopeptide repeat containing 1', other_designations='transmembrane and tetratricopeptide repeat containing 1A', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29500813, end_position_on_the_genomic_accession=29784759, orientation='minus', exon_count=23, OMIM=615855, no_hdr0=''),
  83858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83858, CurrentID=0, Status='live', Symbol='ATAD3B', Aliases='AAA-TOB3, TOB3', description='ATPase family, AAA domain containing 3B', other_designations='AAA-ATPase TOB3', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1471732, end_position_on_the_genomic_accession=1509479, orientation='plus', exon_count=20, OMIM=612317, no_hdr0=''),
  83860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83860, CurrentID=0, Status='live', Symbol='TAF3', Aliases='TAF140, TAFII-140, TAFII140', description='TATA-box binding protein associated factor 3', other_designations='140 kDa TATA box-binding protein-associated factor|RNA polymerase II transcription factor TAFII140|TAF(II)140|TATA box binding protein associated factor 3|TBP-associated factor 3|transcription initiation factor TFIID 140 kDa subunit', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7818477, end_position_on_the_genomic_accession=8016631, orientation='plus', exon_count=7, OMIM=606576, no_hdr0=''),
  83861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83861, CurrentID=0, Status='live', Symbol='RSPH3', Aliases='RSHL2, RSP3, dJ111C20.1', description='radial spoke 3 homolog (Chlamydomonas)', other_designations='A-kinase anchor protein RSPH3|radial spoke head 3 homolog|radial spoke head-like protein 2|radial spoke protein 3|radial spokehead-like 2', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158976739, end_position_on_the_genomic_accession=159000432, orientation='minus', exon_count=9, OMIM=615876, no_hdr0=''),
  83862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83862, CurrentID=0, Status='live', Symbol='TMEM120A', Aliases='NET29, TMPIT', description='transmembrane protein 120A', other_designations='transmembrane protein induced by tumor necrosis factor alpha', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75986837, end_position_on_the_genomic_accession=75994674, orientation='minus', exon_count=12, OMIM=616550, no_hdr0=''),
  346007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346007, CurrentID=0, Status='live', Symbol='EYS', Aliases='C6orf178, C6orf179, C6orf180, EGFL10, EGFL11, RP25, SPAM, bA166P24.2, bA307F22.3, bA74E24.1, dJ1018A4.2, dJ22I17.2, dJ303F19.1', description='eyes shut homolog (Drosophila)', other_designations='EGF-like-domain, multiple 10|EGF-like-domain, multiple 11|epidermal growth factor-like protein 10|epidermal growth factor-like protein 11|protein spacemaker homolog', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63719980, end_position_on_the_genomic_accession=65707225, orientation='minus', exon_count=46, OMIM=612424, no_hdr0=''),
  83864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83864, CurrentID=0, Status='live', Symbol='TTTY9A', Aliases='NCRNA00131, TTTY9, TTY9', description='testis-specific transcript, Y-linked 9A (non-protein coding)', other_designations='testis transcript Y 9', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18729882, end_position_on_the_genomic_accession=18739197, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  83866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83866, CurrentID=0, Status='live', Symbol='TTTY11', Aliases='NCRNA00134, TTY11', description='testis-specific transcript, Y-linked 11 (non-protein coding)', other_designations='testis transcript Y 11|transcript Y 11', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8783313, end_position_on_the_genomic_accession=8817382, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  83867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83867, CurrentID=0, Status='live', Symbol='TTTY12', Aliases='NCRNA00135, TTY11, TTY12', description='testis-specific transcript, Y-linked 12 (non-protein coding)', other_designations='testis transcript Y 12|transcript Y 12', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7804924, end_position_on_the_genomic_accession=7810683, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83868, CurrentID=0, Status='live', Symbol='TTTY13', Aliases='NCRNA00136, TTY13', description='testis-specific transcript, Y-linked 13 (non-protein coding)', other_designations='testis transcript Y 13|transcript Y 13', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21583600, end_position_on_the_genomic_accession=21594666, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  83869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83869, CurrentID=0, Status='live', Symbol='TTTY14', Aliases='CYorf14, NCRNA00137, NCRNA00185, PRO2834, TTY14', description='testis-specific transcript, Y-linked 14 (non-protein coding)', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18872501, end_position_on_the_genomic_accession=19077547, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  83871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83871, CurrentID=0, Status='live', Symbol='RAB34', Aliases='RAB39, RAH', description='RAB34, member RAS oncogene family', other_designations='ras-related protein Rab-39|ras-related protein Rah', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28714281, end_position_on_the_genomic_accession=28718268, orientation='minus', exon_count=11, OMIM=610917, no_hdr0=''),
  83872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83872, CurrentID=0, Status='live', Symbol='HMCN1', Aliases='ARMD1, FBLN6, FIBL-6, FIBL6', description='hemicentin 1', other_designations='fibulin-6', map_location='1q25.3-q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185734381, end_position_on_the_genomic_accession=186190953, orientation='plus', exon_count=107, OMIM=608548, no_hdr0=''),
  83873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83873, CurrentID=0, Status='live', Symbol='GPR61', Aliases='BALGR, GPCR3', description='G protein-coupled receptor 61', other_designations='biogenic amine receptor-like G-protein coupled receptor|biogenic amine receptor-like GPCR', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109539872, end_position_on_the_genomic_accession=109547260, orientation='plus', exon_count=4, OMIM=606916, no_hdr0=''),
  83874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83874, CurrentID=0, Status='live', Symbol='TBC1D10A', Aliases='EPI64, TBC1D10, dJ130H16.1, dJ130H16.2', description='TBC1 domain family member 10A', other_designations='EBP50-PDX interactor of 64 kDa|EBP50-PDZ interactor of 64 kD|TBC1 domain family, member 10A|rab27A-GAP-alpha', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30291990, end_position_on_the_genomic_accession=30326966, orientation='minus', exon_count=9, OMIM=610020, no_hdr0=''),
  83875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83875, CurrentID=0, Status='live', Symbol='BCO2', Aliases='B-DIOX-II, BCDO2', description='beta-carotene oxygenase 2', other_designations="b,b-carotene-9',10'-dioxygenase|beta,beta-carotene 9',10'-dioxygenase variant 1|beta,beta-carotene 9',10'-dioxygenase variant 2|beta,beta-carotene 9',10'-dioxygenase variant 3|beta-carotene 9',10' oxygenase|beta-carotene dioxygenase 2|carotenoid-9',10'-cleaving dioxygenase", map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112175485, end_position_on_the_genomic_accession=112218929, orientation='plus', exon_count=13, OMIM=611740, no_hdr0=''),
  83876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83876, CurrentID=0, Status='live', Symbol='MRO', Aliases='B29, C18orf3', description='maestro', other_designations='beside the Ma29 deletion|male-specific transcription in the developing reproductive organs', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50795120, end_position_on_the_genomic_accession=50825421, orientation='minus', exon_count=9, OMIM=608080, no_hdr0=''),
  100616101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616101, CurrentID=0, Status='live', Symbol='ERVK-10', Aliases='PR, Protease, Proteinase', description='endogenous retrovirus group K member 10', other_designations='Gag polyprotein|Gag-Pro-Pol protein|HERV-K_5q33.3 provirus Np9 protein|HERV-K_5q33.3 provirus ancestral Gag polyprotein|HERV-K_5q33.3 provirus ancestral Pol protein|HERV-K_5q33.3 provirus ancestral Pro protein|endogenous retrovirus group K, member 10', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  83878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83878, CurrentID=0, Status='live', Symbol='USHBP1', Aliases='AIEBP, MCC2', description='Usher syndrome 1C binding protein 1', other_designations='AIE-75-binding protein|USH1C-binding protein 1|mutated in colon cancer protein 2', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17250021, end_position_on_the_genomic_accession=17264796, orientation='minus', exon_count=14, OMIM=611810, no_hdr0=''),
  83879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83879, CurrentID=0, Status='live', Symbol='CDCA7', Aliases='JPO1', description='cell division cycle associated 7', other_designations='c-Myc target JPO1|cell division cycle associated protein 7', map_location='2q31', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173354017, end_position_on_the_genomic_accession=173368990, orientation='plus', exon_count=11, OMIM=609937, no_hdr0=''),
  100616104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616104, CurrentID=0, Status='live', Symbol='ERVFH21-1', Aliases='', description='endogenous retrovirus group FH21 member 1', other_designations='HERV-F(XA34)|endogenous retrovirus group FH21, member 1', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616105, CurrentID=0, Status='live', Symbol='ERVE-4', Aliases='', description='endogenous retrovirus group E member 4', other_designations='CT-RCC HERV-E|endogenous retrovirus group E, member 4', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  83882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83882, CurrentID=0, Status='live', Symbol='TSPAN10', Aliases='OCSP', description='tetraspanin 10', other_designations='oculospanin|tspan-10', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81637171, end_position_on_the_genomic_accession=81648755, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100616107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616107, CurrentID=0, Status='live', Symbol='ERVK3-6', Aliases='HML6-c19A', description='endogenous retrovirus group K3 member 6', other_designations='endogenous retrovirus group K3, member 6', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  83884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83884, CurrentID=0, Status='live', Symbol='SLC25A2', Aliases='ORC2, ORNT2', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 2', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141302629, end_position_on_the_genomic_accession=141304063, orientation='minus', exon_count=1, OMIM=608157, no_hdr0=''),
  100616109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616109, CurrentID=0, Status='live', Symbol='MIR4464', Aliases='mir-4464', description='microRNA 4464', other_designations='hsa-mir-4464', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90312742, end_position_on_the_genomic_accession=90312833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83886, CurrentID=0, Status='live', Symbol='PRSS27', Aliases='CAPH2, MPN', description='protease, serine 27', other_designations='channel-activating protease 2|marapsin|pancreasin', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2712418, end_position_on_the_genomic_accession=2720578, orientation='minus', exon_count=7, OMIM=608018, no_hdr0=''),
  83887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83887, CurrentID=0, Status='live', Symbol='TTLL2', Aliases='C6orf104, NYD-TSPG, dJ366N23.3', description='tubulin tyrosine ligase like 2', other_designations='testis-specific protein NYD-TSPG|tubulin tyrosine ligase-like family member 2|tubulin tyrosine ligase-like family, member 2|tubulin--tyrosine ligase-like protein 2', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167325086, end_position_on_the_genomic_accession=167342689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83888, CurrentID=0, Status='live', Symbol='FGFBP2', Aliases='HBP17RP, KSP37', description='fibroblast growth factor binding protein 2', other_designations='37 kDa killer-specific secretory protein|FGF-BP2|FGF-binding protein 2|FGFBP-2|HBp17-RP|HBp17-related protein|killer-specific secretory protein of 37 kDa', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15960240, end_position_on_the_genomic_accession=15963236, orientation='minus', exon_count=2, OMIM=607713, no_hdr0=''),
  83889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83889, CurrentID=0, Status='live', Symbol='WDR87', Aliases='NYD-SP11', description='WD repeat domain 87', other_designations='testis development protein NYD-SP11', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37884573, end_position_on_the_genomic_accession=37906706, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83890, CurrentID=0, Status='live', Symbol='SPATA9', Aliases='NYD-SP16', description='spermatogenesis associated 9', other_designations='testis development protein NYD-SP16', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95652181, end_position_on_the_genomic_accession=95683010, orientation='minus', exon_count=12, OMIM=608039, no_hdr0=''),
  83891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83891, CurrentID=0, Status='live', Symbol='SNX25', Aliases='MSTP043, SBBI31', description='sorting nexin 25', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185204242, end_position_on_the_genomic_accession=185363969, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  83892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83892, CurrentID=0, Status='live', Symbol='KCTD10', Aliases='BTBD28, MSTP028, ULRO61, hBACURD3', description='potassium channel tetramerization domain containing 10', other_designations='BTB/POZ domain-containing protein KCTD10|potassium channel tetramerisation domain containing 10|potassium channel tetramerization domain-containing protein 10', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109448655, end_position_on_the_genomic_accession=109477350, orientation='minus', exon_count=8, OMIM=613421, no_hdr0=''),
  83893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83893, CurrentID=0, Status='live', Symbol='SPATA16', Aliases='NYD-SP12, SPGF6', description='spermatogenesis associated 16', other_designations='testis development protein NYD-SP12|testis-specific Golgi protein', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172889357, end_position_on_the_genomic_accession=173141268, orientation='minus', exon_count=14, OMIM=609856, no_hdr0=''),
  83894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83894, CurrentID=0, Status='live', Symbol='TTC29', Aliases='NYD-SP14, TBPP2A', description='tetratricopeptide repeat domain 29', other_designations='TPR repeat protein 29|testis development protein NYD-SP14', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146707027, end_position_on_the_genomic_accession=146945882, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  83895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83895, CurrentID=0, Status='live', Symbol='KRTAP1-5', Aliases='KAP1.5, KRTAP1.5', description='keratin associated protein 1-5', other_designations='high sulfur keratin-associated protein 1.5|keratin associated protein 1.5|keratin-associated protein 1.5', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41026026, end_position_on_the_genomic_accession=41027202, orientation='minus', exon_count=1, OMIM=608822, no_hdr0=''),
  83896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83896, CurrentID=0, Status='live', Symbol='KRTAP3-1', Aliases='KAP3.1, KRTAP3.1', description='keratin associated protein 3-1', other_designations='high sulfur keratin-associated protein 3.1|keratin-associated protein 3.1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41008521, end_position_on_the_genomic_accession=41009114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83897, CurrentID=0, Status='live', Symbol='KRTAP3-2', Aliases='KAP3.2, KRTAP3.2', description='keratin associated protein 3-2', other_designations='high sulfur keratin-associated protein 3.2|keratin associated protein 3.2|keratin-associated protein 3.2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40999193, end_position_on_the_genomic_accession=40999886, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616122, CurrentID=0, Status='live', Symbol='MIR4523', Aliases='', description='microRNA 4523', other_designations='hsa-mir-4523', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29390662, end_position_on_the_genomic_accession=29390730, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83899, CurrentID=0, Status='live', Symbol='KRTAP9-2', Aliases='KAP9.2, KRTAP9.2', description='keratin associated protein 9-2', other_designations='keratin-associated protein 9.2|ultrahigh sulfur keratin-associated protein 9.2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41226648, end_position_on_the_genomic_accession=41227652, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83900, CurrentID=0, Status='live', Symbol='KRTAP9-3', Aliases='KAP9.3, KRTAP9.3', description='keratin associated protein 9-3', other_designations='keratin associated protein 9.3|keratin-associated protein 9.3|ultrahigh sulfur keratin-associated protein 9.3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41232463, end_position_on_the_genomic_accession=41233454, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83901, CurrentID=0, Status='live', Symbol='KRTAP9-8', Aliases='KAP9.8, KRTAP9.8', description='keratin associated protein 9-8', other_designations='keratin associated protein 9.8|keratin-associated protein 9.8|ultrahigh sulfur keratin-associated protein 9.8', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41238018, end_position_on_the_genomic_accession=41239004, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83902, CurrentID=0, Status='live', Symbol='KRTAP17-1', Aliases='KAP17.1, KRTAP16.1, KRTAP17.1', description='keratin associated protein 17-1', other_designations='keratin associated protein 16.1|keratin-associated protein 16.1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41314917, end_position_on_the_genomic_accession=41315695, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83903, CurrentID=0, Status='live', Symbol='GSG2', Aliases='HASPIN', description='germ cell associated 2 (haspin)', other_designations='H-haspin|germ cell-specific gene 2 protein|haploid germ cell-specific nuclear protein kinase', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3723903, end_position_on_the_genomic_accession=3726699, orientation='plus', exon_count=1, OMIM=609240, no_hdr0=''),
  100616128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616128, CurrentID=0, Status='live', Symbol='MIR4679-1', Aliases='', description='microRNA 4679-1', other_designations='hsa-mir-4679-1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89063336, end_position_on_the_genomic_accession=89063410, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616129, CurrentID=0, Status='live', Symbol='MIR4445', Aliases='', description='microRNA 4445', other_designations='hsa-mir-4445', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109602828, end_position_on_the_genomic_accession=109602897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616130, CurrentID=0, Status='live', Symbol='MIR4526', Aliases='mir-4526', description='microRNA 4526', other_designations='hsa-mir-4526', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13611114, end_position_on_the_genomic_accession=13611200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616131, CurrentID=0, Status='live', Symbol='MIR3689D1', Aliases='', description='microRNA 3689d-1', other_designations='hsa-mir-3689d-1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134849609, end_position_on_the_genomic_accession=134849682, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616132, CurrentID=0, Status='live', Symbol='MIR4728', Aliases='', description='microRNA 4728', other_designations='hsa-mir-4728', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39726495, end_position_on_the_genomic_accession=39726561, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616133, CurrentID=0, Status='live', Symbol='MIR4699', Aliases='', description='microRNA 4699', other_designations='hsa-mir-4699', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81158388, end_position_on_the_genomic_accession=81158461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616134, CurrentID=0, Status='live', Symbol='MIR499B', Aliases='MIR499A, hsa-mir-499b', description='microRNA 499b', other_designations='hsa-mir-499a|microRNA 499a', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34990400, end_position_on_the_genomic_accession=34990472, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616135, CurrentID=0, Status='live', Symbol='MIR4507', Aliases='', description='microRNA 4507', other_designations='hsa-mir-4507', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105858124, end_position_on_the_genomic_accession=105858175, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616136, CurrentID=0, Status='live', Symbol='MIR4430', Aliases='', description='microRNA 4430', other_designations='hsa-mir-4430', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33418516, end_position_on_the_genomic_accession=33418564, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616137, CurrentID=0, Status='live', Symbol='MIR4501', Aliases='', description='microRNA 4501', other_designations='hsa-mir-4501', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96427229, end_position_on_the_genomic_accession=96427292, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616138, CurrentID=0, Status='live', Symbol='MIR4787', Aliases='mir-4787', description='microRNA 4787', other_designations='hsa-mir-4787', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50675080, end_position_on_the_genomic_accession=50675163, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616139, CurrentID=0, Status='live', Symbol='MIR4741', Aliases='mir-4741', description='microRNA 4741', other_designations='hsa-mir-4741', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22933349, end_position_on_the_genomic_accession=22933438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616140, CurrentID=0, Status='live', Symbol='MIR4506', Aliases='', description='microRNA 4506', other_designations='hsa-mir-4506', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93948226, end_position_on_the_genomic_accession=93948302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616141, CurrentID=0, Status='live', Symbol='MIR4428', Aliases='', description='microRNA 4428', other_designations='hsa-mir-4428', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=237471119, end_position_on_the_genomic_accession=237471191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616142, CurrentID=0, Status='live', Symbol='MIR4458', Aliases='mir-4458', description='microRNA 4458', other_designations='hsa-mir-4458', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8460925, end_position_on_the_genomic_accession=8460999, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616143, CurrentID=0, Status='live', Symbol='MIR4763', Aliases='', description='microRNA 4763', other_designations='hsa-mir-4763', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46113566, end_position_on_the_genomic_accession=46113657, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616144, CurrentID=0, Status='live', Symbol='MIR548AN', Aliases='', description='microRNA 548an', other_designations='hsa-mir-548an', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106639814, end_position_on_the_genomic_accession=106639896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616145, CurrentID=0, Status='live', Symbol='MIR1273G', Aliases='', description='microRNA 1273g', other_designations='hsa-mir-1273g', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52940314, end_position_on_the_genomic_accession=52940413, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616146, CurrentID=0, Status='live', Symbol='MIR4534', Aliases='', description='microRNA 4534', other_designations='hsa-mir-4534', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37988794, end_position_on_the_genomic_accession=37988853, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616147, CurrentID=0, Status='live', Symbol='MIR4769', Aliases='', description='microRNA 4769', other_designations='hsa-mir-4769', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47587429, end_position_on_the_genomic_accession=47587505, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616148, CurrentID=0, Status='live', Symbol='MIR4760', Aliases='', description='microRNA 4760', other_designations='hsa-mir-4760', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40212352, end_position_on_the_genomic_accession=40212431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616149, CurrentID=0, Status='live', Symbol='MIR4512', Aliases='', description='microRNA 4512', other_designations='hsa-mir-4512', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66496958, end_position_on_the_genomic_accession=66497034, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616150, CurrentID=0, Status='live', Symbol='MIR4720', Aliases='', description='microRNA 4720', other_designations='hsa-mir-4720', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81385018, end_position_on_the_genomic_accession=81385093, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616151, CurrentID=0, Status='live', Symbol='MIR4480', Aliases='', description='microRNA 4480', other_designations='hsa-mir-4480', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12578753, end_position_on_the_genomic_accession=12578823, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616152, CurrentID=0, Status='live', Symbol='MIR4804', Aliases='', description='microRNA 4804', other_designations='hsa-mir-4804', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72878591, end_position_on_the_genomic_accession=72878663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616153, CurrentID=0, Status='live', Symbol='MIR4726', Aliases='', description='microRNA 4726', other_designations='hsa-mir-4726', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38719691, end_position_on_the_genomic_accession=38719748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83930, CurrentID=0, Status='live', Symbol='STARD3NL', Aliases='MENTHO', description='STARD3 N-terminal like', other_designations='STARD3 N-terminal-like protein', map_location='7p14-p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38178206, end_position_on_the_genomic_accession=38230671, orientation='plus', exon_count=12, OMIM=611759, no_hdr0=''),
  83931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83931, CurrentID=0, Status='live', Symbol='STK40', Aliases='SHIK, SgK495', description='serine/threonine kinase 40', other_designations='SINK-homologous serine/threonine-protein kinase|Ser/Thr-like kinase|sugen kinase 495', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36339619, end_position_on_the_genomic_accession=36385927, orientation='minus', exon_count=12, OMIM=609437, no_hdr0=''),
  83932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83932, CurrentID=0, Status='live', Symbol='SPRTN', Aliases='C1orf124, DVC1, PRO4323, spartan', description='SprT-like N-terminal domain', other_designations='DNA damage protein targeting VCP|DNA damage-targeting VCP (p97) adaptor|zinc finger RAD18 domain-containing protein', map_location='1q42.12-q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231337936, end_position_on_the_genomic_accession=231370551, orientation='plus', exon_count=7, OMIM=616086, no_hdr0=''),
  83933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83933, CurrentID=0, Status='live', Symbol='HDAC10', Aliases='HD10', description='histone deacetylase 10', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50245183, end_position_on_the_genomic_accession=50251405, orientation='minus', exon_count=20, OMIM=608544, no_hdr0=''),
  100616158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616158, CurrentID=0, Status='live', Symbol='MIR4505', Aliases='mir-4505', description='microRNA 4505', other_designations='hsa-mir-4505', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73758747, end_position_on_the_genomic_accession=73758819, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83935, CurrentID=0, Status='live', Symbol='TMEM133', Aliases='AD031', description='transmembrane protein 133', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100992080, end_position_on_the_genomic_accession=100993935, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616160, CurrentID=0, Status='live', Symbol='MIR4655', Aliases='', description='microRNA 4655', other_designations='hsa-mir-4655', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1844180, end_position_on_the_genomic_accession=1844253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83937, CurrentID=0, Status='live', Symbol='RASSF4', Aliases='AD037', description='Ras association (RalGDS/AF-6) domain family member 4', other_designations='Ras association (RalGDS/AF-6) domain family 4|Ras association domain family 4|tumor suppressor RASSF4', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44959750, end_position_on_the_genomic_accession=44994724, orientation='plus', exon_count=12, OMIM=610559, no_hdr0=''),
  83938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83938, CurrentID=0, Status='live', Symbol='C10orf11', Aliases='CDA017', description='chromosome 10 open reading frame 11', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75431458, end_position_on_the_genomic_accession=76557375, orientation='plus', exon_count=11, OMIM=614537, no_hdr0=''),
  83939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83939, CurrentID=0, Status='live', Symbol='EIF2A', Aliases='CDA02, EIF-2A, MST089, MSTP004, MSTP089', description='eukaryotic translation initiation factor 2A', other_designations='65 kDa eukaryotic translation initiation factor 2A|eukaryotic translation initiation factor 2A, 65kDa', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150546690, end_position_on_the_genomic_accession=150586016, orientation='plus', exon_count=14, OMIM=609234, no_hdr0=''),
  83940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83940, CurrentID=0, Status='live', Symbol='TATDN1', Aliases='CDA11', description='TatD DNase domain containing 1', other_designations='hepatocarcinoma high expression protein', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124488494, end_position_on_the_genomic_accession=124539088, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101926885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926885, CurrentID=0, Status='live', Symbol='OCA5', Aliases='OCA7', description='oculocutaneous albinism 5 (autosomal recessive)', other_designations='Albinism, oculocutaneous, type VII', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615312, no_hdr0=''),
  83942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83942, CurrentID=0, Status='live', Symbol='TSSK1B', Aliases='FKSG81, SPOGA4, STK22D, TSK1, TSSK1', description='testis-specific serine kinase 1B', other_designations='TSK-1|TSSK-1|serine/threonine kinase 22D (spermiogenesis associated)|serine/threonine kinase FKSG81|serine/threonine-protein kinase 22A|spermiogenesis associated 4|testis-specific kinase 1|testis-specific serine kinase 1', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=113432554, end_position_on_the_genomic_accession=113435031, orientation='minus', exon_count=1, OMIM=610709, no_hdr0=''),
  83943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83943, CurrentID=0, Status='live', Symbol='IMMP2L', Aliases='IMMP2L-IT1, IMP2, IMP2-LIKE', description='inner mitochondrial membrane peptidase subunit 2', other_designations='IMP2 inner mitochondrial membrane peptidase-like|inner mitochondrial membrane peptidase 2 like', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110662645, end_position_on_the_genomic_accession=111562531, orientation='minus', exon_count=18, OMIM=605977, no_hdr0=''),
  100616168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616168, CurrentID=0, Status='live', Symbol='MIR4754', Aliases='mir-4754', description='microRNA 4754', other_designations='hsa-mir-4754', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58386770, end_position_on_the_genomic_accession=58386858, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616169, CurrentID=0, Status='live', Symbol='MIR378D2', Aliases='mir-378d-2', description='microRNA 378d-2', other_designations='hsa-mir-378d-2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93916022, end_position_on_the_genomic_accession=93916119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616170, CurrentID=0, Status='live', Symbol='MIR4739', Aliases='mir-4739', description='microRNA 4739', other_designations='hsa-mir-4739', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79707176, end_position_on_the_genomic_accession=79707249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616171, CurrentID=0, Status='live', Symbol='MIR4752', Aliases='', description='microRNA 4752', other_designations='hsa-mir-4752', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54282109, end_position_on_the_genomic_accession=54282180, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616172, CurrentID=0, Status='live', Symbol='MIR4719', Aliases='', description='microRNA 4719', other_designations='hsa-mir-4719', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76868936, end_position_on_the_genomic_accession=76869019, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616173, CurrentID=0, Status='live', Symbol='MIR203B', Aliases='MIR3545, hsa-mir-203b', description='microRNA 203b', other_designations='hsa-mir-3545|microRNA 3545', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104117418, end_position_on_the_genomic_accession=104117503, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616174, CurrentID=0, Status='live', Symbol='MIR4643', Aliases='', description='microRNA 4643', other_designations='hsa-mir-4643', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=91521660, end_position_on_the_genomic_accession=91521737, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616175, CurrentID=0, Status='live', Symbol='MIR4633', Aliases='', description='microRNA 4633', other_designations='hsa-mir-4633', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129097688, end_position_on_the_genomic_accession=129097766, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616176, CurrentID=0, Status='live', Symbol='MIR4708', Aliases='mir-4708', description='microRNA 4708', other_designations='hsa-mir-4708', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65335117, end_position_on_the_genomic_accession=65335183, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83953, CurrentID=0, Status='live', Symbol='FCAMR', Aliases='CD351, FCA/MR, FKSG87', description='Fc receptor, IgA, IgM, high affinity', other_designations='Fc alpha/mu receptor|immunity related factor|receptor for Fc fragment of IgA and IgM', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206957647, end_position_on_the_genomic_accession=206970659, orientation='minus', exon_count=10, OMIM=605484, no_hdr0=''),
  100616178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616178, CurrentID=0, Status='live', Symbol='MIR4641', Aliases='', description='microRNA 4641', other_designations='hsa-mir-4641', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41598723, end_position_on_the_genomic_accession=41598788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83955, CurrentID=0, Status='live', Symbol='NACAP1', Aliases='FKSG17', description='nascent polypeptide associated complex alpha subunit pseudogene 1', other_designations='FKSG17|nascent-polypeptide-associated complex alpha polypeptide pseudogene 1', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101368893, end_position_on_the_genomic_accession=101369595, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616180, CurrentID=0, Status='live', Symbol='MIR4465', Aliases='', description='microRNA 4465', other_designations='hsa-mir-4465', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140683814, end_position_on_the_genomic_accession=140683883, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616181, CurrentID=0, Status='live', Symbol='MIR4514', Aliases='', description='microRNA 4514', other_designations='hsa-mir-4514', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80997417, end_position_on_the_genomic_accession=80997473, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616182, CurrentID=0, Status='live', Symbol='MIR4500', Aliases='', description='microRNA 4500', other_designations='hsa-mir-4500', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87618665, end_position_on_the_genomic_accession=87618740, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83959, CurrentID=0, Status='live', Symbol='SLC4A11', Aliases='BTR1, CDPD1, CHED2, NABC1, dJ794I6.2', description='solute carrier family 4, sodium borate transporter, member 11', other_designations='bicarbonate transporter related protein 1|sodium-coupled borate cotransporter 1|solute carrier family 4, sodium bicarbonate transporter-like, member 11', map_location='20p12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3227417, end_position_on_the_genomic_accession=3241482, orientation='minus', exon_count=24, OMIM=610206, no_hdr0=''),
  100616184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616184, CurrentID=0, Status='live', Symbol='MIR4477A', Aliases='', description='microRNA 4477a', other_designations='hsa-mir-4477a', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41233755, end_position_on_the_genomic_accession=41233835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616185, CurrentID=0, Status='live', Symbol='MIR371B', Aliases='', description='microRNA 371b', other_designations='hsa-mir-371b', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53787677, end_position_on_the_genomic_accession=53787742, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616186, CurrentID=0, Status='live', Symbol='MIR4490', Aliases='', description='microRNA 4490', other_designations='hsa-mir-4490', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90555774, end_position_on_the_genomic_accession=90555857, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616187, CurrentID=0, Status='live', Symbol='MIR4783', Aliases='', description='microRNA 4783', other_designations='hsa-mir-4783', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127423537, end_position_on_the_genomic_accession=127423618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616188, CurrentID=0, Status='live', Symbol='MIR3972', Aliases='', description='microRNA 3972', other_designations='hsa-mir-3972', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17277889, end_position_on_the_genomic_accession=17277975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616189, CurrentID=0, Status='live', Symbol='MIR4421', Aliases='', description='microRNA 4421', other_designations='hsa-mir-4421', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51059837, end_position_on_the_genomic_accession=51059905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616190, CurrentID=0, Status='live', Symbol='MIR548O2', Aliases='', description='microRNA 548o-2', other_designations='hsa-mir-548o-2', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38516563, end_position_on_the_genomic_accession=38516632, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616191, CurrentID=0, Status='live', Symbol='MIR548AJ1', Aliases='', description='microRNA 548aj-1', other_designations='hsa-mir-548aj-1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132115192, end_position_on_the_genomic_accession=132115263, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616192, CurrentID=0, Status='live', Symbol='MIR4679-2', Aliases='', description='microRNA 4679-2', other_designations='hsa-mir-4679-2', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89063335, end_position_on_the_genomic_accession=89063411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616193, CurrentID=0, Status='live', Symbol='MIR4453', Aliases='', description='microRNA 4453', other_designations='hsa-mir-4453', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152536428, end_position_on_the_genomic_accession=152536516, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616194, CurrentID=0, Status='live', Symbol='MIR4477B', Aliases='', description='microRNA 4477b', other_designations='hsa-mir-4477b', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63819574, end_position_on_the_genomic_accession=63819654, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616195, CurrentID=0, Status='live', Symbol='MIR4718', Aliases='', description='microRNA 4718', other_designations='hsa-mir-4718', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12720321, end_position_on_the_genomic_accession=12720371, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616196, CurrentID=0, Status='live', Symbol='MIR4525', Aliases='', description='microRNA 4525', other_designations='hsa-mir-4525', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82668233, end_position_on_the_genomic_accession=82668307, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616197, CurrentID=0, Status='live', Symbol='ERVK-23', Aliases='c21_A', description='endogenous retrovirus group K member 23', other_designations='endogenous retrovirus group K, member 23', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616198, CurrentID=0, Status='live', Symbol='ERVW-6', Aliases='', description='endogenous retrovirus group W member 6', other_designations='endogenous retrovirus group W, member 6', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616199, CurrentID=0, Status='live', Symbol='ERVK3-7', Aliases='HML6-c19B', description='endogenous retrovirus group K3 member 7', other_designations='endogenous retrovirus group K3, member 7', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616200, CurrentID=0, Status='live', Symbol='ERVS71-2', Aliases='HERV-HC2', description='endogenous retrovirus group S71 member 2', other_designations='endogenous retrovirus group S71, member 2', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616201, CurrentID=0, Status='live', Symbol='MIR378D1', Aliases='mir-378d-1', description='microRNA 378d-1', other_designations='hsa-mir-378d-1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5923275, end_position_on_the_genomic_accession=5923328, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616202, CurrentID=0, Status='live', Symbol='MIR4634', Aliases='', description='microRNA 4634', other_designations='hsa-mir-4634', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174751734, end_position_on_the_genomic_accession=174751787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616203, CurrentID=0, Status='live', Symbol='MIR4734', Aliases='mir-4734', description='microRNA 4734', other_designations='hsa-mir-4734', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38702262, end_position_on_the_genomic_accession=38702331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616204, CurrentID=0, Status='live', Symbol='MIR4729', Aliases='', description='microRNA 4729', other_designations='hsa-mir-4729', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59366083, end_position_on_the_genomic_accession=59366154, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616205, CurrentID=0, Status='live', Symbol='MIR4704', Aliases='', description='microRNA 4704', other_designations='hsa-mir-4704', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66218250, end_position_on_the_genomic_accession=66218324, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83982, CurrentID=0, Status='live', Symbol='IFI27L2', Aliases='FAM14A, ISG12B, TLH29', description='interferon, alpha-inducible protein 27-like 2', other_designations='ISG12(b) protein|family with sequence similarity 14, member A|interferon-stimulated gene 12b protein|pIFI27-like protein', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94127781, end_position_on_the_genomic_accession=94129620, orientation='minus', exon_count=4, OMIM=611319, no_hdr0=''),
  83983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83983, CurrentID=0, Status='live', Symbol='TSSK6', Aliases='CT72, FKSG82, SSTK, TSSK4', description='testis-specific serine kinase 6', other_designations='TSK-6|cancer/testis antigen 72|serine/threonine protein kinase SSTK|small serine/threonine kinase', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19514219, end_position_on_the_genomic_accession=19515660, orientation='minus', exon_count=1, OMIM=610712, no_hdr0=''),
  100616208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616208, CurrentID=0, Status='live', Symbol='MIR4782', Aliases='', description='microRNA 4782', other_designations='hsa-mir-4782', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113721290, end_position_on_the_genomic_accession=113721368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  83985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83985, CurrentID=0, Status='live', Symbol='SPNS1', Aliases='HSpin1, LAT, PP2030, SPIN1, SPINL, nrs', description='spinster homolog 1 (Drosophila)', other_designations='spinster-like protein 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28974771, end_position_on_the_genomic_accession=28984762, orientation='plus', exon_count=14, OMIM=612583, no_hdr0=''),
  83986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83986, CurrentID=0, Status='live', Symbol='FAM234A', Aliases='C16orf9, ITFG3, gs19', description='family with sequence similarity 234 member A', other_designations='family with sequence similarity 234, member A|integrin alpha FG-GAP repeat containing 3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=234546, end_position_on_the_genomic_accession=269943, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  83987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83987, CurrentID=0, Status='live', Symbol='CCDC8', Aliases='3M3, PPP1R20, p90', description='coiled-coil domain containing 8', other_designations='protein phosphatase 1, regulatory subunit 20', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46410329, end_position_on_the_genomic_accession=46413662, orientation='minus', exon_count=1, OMIM=614145, no_hdr0=''),
  83988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83988, CurrentID=0, Status='live', Symbol='NCALD', Aliases='', description='neurocalcin delta', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101686542, end_position_on_the_genomic_accession=102124907, orientation='minus', exon_count=14, OMIM=606722, no_hdr0=''),
  83989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83989, CurrentID=0, Status='live', Symbol='FAM172A', Aliases='C5orf21', description='family with sequence similarity 172 member A', other_designations='family with sequence similarity 172, member A', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93617725, end_position_on_the_genomic_accession=94111699, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  83990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83990, CurrentID=0, Status='live', Symbol='BRIP1', Aliases='BACH1, FANCJ, OF', description='BRCA1 interacting protein C-terminal helicase 1', other_designations='ATP-dependent RNA helicase BRIP1|BRCA1-associated C-terminal helicase 1|BRCA1-binding helicase-like protein BACH1|BRCA1/BRCA2-associated helicase 1', map_location='17q22.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61679186, end_position_on_the_genomic_accession=61863937, orientation='minus', exon_count=24, OMIM=605882, no_hdr0=''),
  100616215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616215, CurrentID=0, Status='live', Symbol='MIR548AL', Aliases='', description='microRNA 548al', other_designations='hsa-mir-548al', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74399237, end_position_on_the_genomic_accession=74399333, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83992, CurrentID=0, Status='live', Symbol='CTTNBP2', Aliases='C7orf8, CORTBP2, Orf4', description='cortactin binding protein 2', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117710648, end_position_on_the_genomic_accession=117873507, orientation='minus', exon_count=27, OMIM=609772, no_hdr0=''),
  100616217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616217, CurrentID=0, Status='live', Symbol='MIR1254-2', Aliases='', description='microRNA 1254-2', other_designations='hsa-mir-1254-2', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23393405, end_position_on_the_genomic_accession=23393467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616218, CurrentID=0, Status='live', Symbol='MIR3135B', Aliases='', description='microRNA 3135b', other_designations='hsa-mir-3135b', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32749912, end_position_on_the_genomic_accession=32749979, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616219, CurrentID=0, Status='live', Symbol='MIR4765', Aliases='', description='microRNA 4765', other_designations='hsa-mir-4765', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32635255, end_position_on_the_genomic_accession=32635331, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616220, CurrentID=0, Status='live', Symbol='MIR4736', Aliases='', description='microRNA 4736', other_designations='hsa-mir-4736', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58335976, end_position_on_the_genomic_accession=58336022, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616221, CurrentID=0, Status='live', Symbol='MIR4662A', Aliases='', description='microRNA 4662a', other_designations='hsa-mir-4662a', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124821985, end_position_on_the_genomic_accession=124822051, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  83998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83998, CurrentID=0, Status='live', Symbol='REG4', Aliases='GISP, REG-IV, RELP', description='regenerating family member 4', other_designations='REG-4|REG-like protein|gastrointestinal secretory protein|reg IV|regenerating gene type IV|regenerating islet-derived family, member 4|regenerating islet-derived protein IV', map_location='1p13.1-p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119794018, end_position_on_the_genomic_accession=119811580, orientation='minus', exon_count=7, OMIM=609846, no_hdr0=''),
  83999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83999, CurrentID=0, Status='live', Symbol='KREMEN1', Aliases='KREMEN, KRM1', description='kringle containing transmembrane protein 1', other_designations='dickkopf receptor|kringle domain-containing transmembrane protein 1|kringle-coding gene marking the eye and the nose|kringle-containing protein marking the eye and the nose', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29073078, end_position_on_the_genomic_accession=29168333, orientation='plus', exon_count=12, OMIM=609898, no_hdr0=''),
  84000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84000, CurrentID=0, Status='live', Symbol='TMPRSS13', Aliases='MSP, MSPL, MSPS, TMPRSS11', description='transmembrane protease, serine 13', other_designations='membrane-type mosaic serine protease|transmembrane protease, serine 11', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117900641, end_position_on_the_genomic_accession=117929453, orientation='minus', exon_count=14, OMIM=610050, no_hdr0=''),
  100616225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616225, CurrentID=0, Status='live', Symbol='MIR4756', Aliases='', description='microRNA 4756', other_designations='hsa-mir-4756', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=54068408, end_position_on_the_genomic_accession=54068485, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84002, CurrentID=0, Status='live', Symbol='B3GNT5', Aliases='B3GN-T5, beta3Gn-T5', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 5', other_designations='BGnT-5|UDP-GlcNAc:beta-Gal beta-1,3-N-acetylglucosaminyltransferase 5|beta 1,3 N-acetyglucosaminyltransferase Lc3 synthase|beta-1,3-Gn-T5|beta-1,3-N-acetylglucosaminyltransferase 5|beta-1,3-N-acetylglucosaminyltransferase bGnT-5|lactotriaosylceramide synthase|lc(3)Cer synthase|lc3 synthase', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183253226, end_position_on_the_genomic_accession=183273391, orientation='plus', exon_count=7, OMIM=615333, no_hdr0=''),
  101926947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926947, CurrentID=0, Status='live', Symbol='LINC01617', Aliases='C8orf23, NCRNA00249, XLOC_006844', description='long intergenic non-protein coding RNA 1617', other_designations='uncharacterized NCRNA00249', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73879005, end_position_on_the_genomic_accession=73904734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616228, CurrentID=0, Status='live', Symbol='MIR4509-2', Aliases='', description='microRNA 4509-2', other_designations='hsa-mir-4509-2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28426491, end_position_on_the_genomic_accession=28426584, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616229, CurrentID=0, Status='live', Symbol='MIR4473', Aliases='', description='microRNA 4473', other_designations='hsa-mir-4473', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20411148, end_position_on_the_genomic_accession=20411238, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616230, CurrentID=0, Status='live', Symbol='MIR4646', Aliases='mir-4646', description='microRNA 4646', other_designations='hsa-mir-4646', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31701029, end_position_on_the_genomic_accession=31701091, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616231, CurrentID=0, Status='live', Symbol='MIR4519', Aliases='', description='microRNA 4519', other_designations='hsa-mir-4519', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30875266, end_position_on_the_genomic_accession=30875323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616232, CurrentID=0, Status='live', Symbol='MIR4528', Aliases='', description='microRNA 4528', other_designations='hsa-mir-4528', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=53237101, end_position_on_the_genomic_accession=53237190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616233, CurrentID=0, Status='live', Symbol='MIR4459', Aliases='', description='microRNA 4459', other_designations='hsa-mir-4459', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54075518, end_position_on_the_genomic_accession=54075583, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616234, CurrentID=0, Status='live', Symbol='MIR4454', Aliases='', description='microRNA 4454', other_designations='hsa-mir-4454', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163093574, end_position_on_the_genomic_accession=163093628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616235, CurrentID=0, Status='live', Symbol='MIR4457', Aliases='mir-4457', description='microRNA 4457', other_designations='hsa-mir-4457', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1309310, end_position_on_the_genomic_accession=1309377, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616236, CurrentID=0, Status='live', Symbol='MIR4669', Aliases='', description='microRNA 4669', other_designations='hsa-mir-4669', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134379411, end_position_on_the_genomic_accession=134379472, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  346157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346157, CurrentID=0, Status='live', Symbol='ZNF391', Aliases='dJ153G14.3', description='zinc finger protein 391', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27374938, end_position_on_the_genomic_accession=27403908, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100616238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616238, CurrentID=0, Status='live', Symbol='MIR3529', Aliases='mir-3529', description='microRNA 3529', other_designations='hsa-mir-3529', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88611847, end_position_on_the_genomic_accession=88611924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616239, CurrentID=0, Status='live', Symbol='MIR4705', Aliases='', description='microRNA 4705', other_designations='hsa-mir-4705', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102045934, end_position_on_the_genomic_accession=102046004, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616240, CurrentID=0, Status='live', Symbol='MIR4496', Aliases='', description='microRNA 4496', other_designations='hsa-mir-4496', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108635810, end_position_on_the_genomic_accession=108635870, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616241, CurrentID=0, Status='live', Symbol='MIR4717', Aliases='', description='microRNA 4717', other_designations='hsa-mir-4717', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2274620, end_position_on_the_genomic_accession=2274691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616242, CurrentID=0, Status='live', Symbol='MIR4673', Aliases='', description='microRNA 4673', other_designations='hsa-mir-4673', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136519568, end_position_on_the_genomic_accession=136519626, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616243, CurrentID=0, Status='live', Symbol='MIR4759', Aliases='', description='microRNA 4759', other_designations='hsa-mir-4759', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26953961, end_position_on_the_genomic_accession=26954043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616244, CurrentID=0, Status='live', Symbol='MIR3976', Aliases='', description='microRNA 3976', other_designations='hsa-mir-3976', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5840695, end_position_on_the_genomic_accession=5840833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616245, CurrentID=0, Status='live', Symbol='MIR4661', Aliases='', description='microRNA 4661', other_designations='hsa-mir-4661', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91205485, end_position_on_the_genomic_accession=91205559, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616246, CurrentID=0, Status='live', Symbol='MIR4799', Aliases='', description='microRNA 4799', other_designations='hsa-mir-4799', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147782595, end_position_on_the_genomic_accession=147782668, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616247, CurrentID=0, Status='live', Symbol='MIR151B', Aliases='', description='microRNA 151b', other_designations='hsa-mir-151b', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100109419, end_position_on_the_genomic_accession=100109514, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616248, CurrentID=0, Status='live', Symbol='MIR4724', Aliases='mir-4724', description='microRNA 4724', other_designations='hsa-mir-4724', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31534883, end_position_on_the_genomic_accession=31534971, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616249, CurrentID=0, Status='live', Symbol='MIR4768', Aliases='', description='microRNA 4768', other_designations='hsa-mir-4768', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17425881, end_position_on_the_genomic_accession=17425954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616250, CurrentID=0, Status='live', Symbol='MIR3960', Aliases='', description='microRNA 3960', other_designations='hsa-mir-3960', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127785833, end_position_on_the_genomic_accession=127785923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  346171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346171, CurrentID=0, Status='live', Symbol='ZFP57', Aliases='C6orf40, TNDM1, ZNF698, bA145L22, bA145L22.2', description='ZFP57 zinc finger protein', other_designations='zfp-57|zinc finger protein 698', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29672392, end_position_on_the_genomic_accession=29681146, orientation='minus', exon_count=5, OMIM=612192, no_hdr0=''),
  100616252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616252, CurrentID=0, Status='live', Symbol='MIR548AJ2', Aliases='', description='microRNA 548aj-2', other_designations='hsa-mir-548aj-2', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38023895, end_position_on_the_genomic_accession=38023986, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616253, CurrentID=0, Status='live', Symbol='MIR4762', Aliases='', description='microRNA 4762', other_designations='hsa-mir-4762', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45760524, end_position_on_the_genomic_accession=45760598, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616254, CurrentID=0, Status='live', Symbol='MIR548AH', Aliases='', description='microRNA 548ah', other_designations='hsa-mir-548ah', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76575551, end_position_on_the_genomic_accession=76575626, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616255, CurrentID=0, Status='live', Symbol='MIR4662B', Aliases='', description='microRNA 4662b', other_designations='hsa-mir-4662b', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124821978, end_position_on_the_genomic_accession=124822058, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616256, CurrentID=0, Status='live', Symbol='MIR4721', Aliases='', description='microRNA 4721', other_designations='hsa-mir-4721', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28843919, end_position_on_the_genomic_accession=28844007, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84033, CurrentID=0, Status='live', Symbol='OBSCN', Aliases='ARHGEF30, UNC89', description='obscurin, cytoskeletal calmodulin and titin-interacting RhoGEF', other_designations='obscurin, myosin light chain kinase|obscurin-MLCK', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228208072, end_position_on_the_genomic_accession=228378876, orientation='plus', exon_count=122, OMIM=608616, no_hdr0=''),
  84034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84034, CurrentID=0, Status='live', Symbol='EMILIN2', Aliases='EMILIN-2, FOAP-10', description='elastin microfibril interfacer 2', other_designations='elastin microfibril interface-located protein 2|extracellular glycoprotein EMILIN-2', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2846231, end_position_on_the_genomic_accession=2914092, orientation='plus', exon_count=11, OMIM=608928, no_hdr0=''),
  100616259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616259, CurrentID=0, Status='live', Symbol='MIR378I', Aliases='', description='microRNA 378i', other_designations='hsa-mir-378i', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41923222, end_position_on_the_genomic_accession=41923297, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616260, CurrentID=0, Status='live', Symbol='MIR4663', Aliases='', description='microRNA 4663', other_designations='hsa-mir-4663', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123215788, end_position_on_the_genomic_accession=123215863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616261, CurrentID=0, Status='live', Symbol='MIR4504', Aliases='', description='microRNA 4504', other_designations='hsa-mir-4504', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50299855, end_position_on_the_genomic_accession=50299946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616262, CurrentID=0, Status='live', Symbol='MIR4701', Aliases='', description='microRNA 4701', other_designations='hsa-mir-4701', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48771975, end_position_on_the_genomic_accession=48772037, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616263, CurrentID=0, Status='live', Symbol='MIR4485', Aliases='mir-4485', description='microRNA 4485', other_designations='hsa-mir-4485', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10508270, end_position_on_the_genomic_accession=10508326, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616264, CurrentID=0, Status='live', Symbol='MIR4527', Aliases='', description='microRNA 4527', other_designations='hsa-mir-4527', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47380496, end_position_on_the_genomic_accession=47380565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616265, CurrentID=0, Status='live', Symbol='MIR4433A', Aliases='MIR4433, hsa-mir-4433a', description='microRNA 4433a', other_designations='hsa-mir-4433|microRNA 4433', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64340759, end_position_on_the_genomic_accession=64340839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616266, CurrentID=0, Status='live', Symbol='MIR4733', Aliases='', description='microRNA 4733', other_designations='hsa-mir-4733', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31094350, end_position_on_the_genomic_accession=31094425, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616267, CurrentID=0, Status='live', Symbol='MIR4776-1', Aliases='mir-4776-1', description='microRNA 4776-1', other_designations='hsa-mir-4776-1', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212926257, end_position_on_the_genomic_accession=212926336, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616268, CurrentID=0, Status='live', Symbol='MIR4472-1', Aliases='', description='microRNA 4472-1', other_designations='hsa-mir-4472-1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142176339, end_position_on_the_genomic_accession=142176418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616269, CurrentID=0, Status='live', Symbol='MIR4639', Aliases='', description='microRNA 4639', other_designations='hsa-mir-4639', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16141556, end_position_on_the_genomic_accession=16141624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616270, CurrentID=0, Status='live', Symbol='MIR4651', Aliases='mir-4651', description='microRNA 4651', other_designations='hsa-mir-4651', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75915197, end_position_on_the_genomic_accession=75915269, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616271, CurrentID=0, Status='live', Symbol='MIR4637', Aliases='', description='microRNA 4637', other_designations='hsa-mir-4637', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14825929, end_position_on_the_genomic_accession=14826012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616272, CurrentID=0, Status='live', Symbol='MIR4422', Aliases='mir-4422', description='microRNA 4422', other_designations='hsa-mir-4422', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55225641, end_position_on_the_genomic_accession=55225723, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616273, CurrentID=0, Status='live', Symbol='MIR451B', Aliases='', description='microRNA 451b', other_designations='hsa-mir-451b', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28861371, end_position_on_the_genomic_accession=28861438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616274, CurrentID=0, Status='live', Symbol='MIR4802', Aliases='', description='microRNA 4802', other_designations='hsa-mir-4802', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40502040, end_position_on_the_genomic_accession=40502119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616275, CurrentID=0, Status='live', Symbol='MIR4508', Aliases='mir-4508', description='microRNA 4508', other_designations='hsa-mir-4508', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23562062, end_position_on_the_genomic_accession=23562131, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616276, CurrentID=0, Status='live', Symbol='MIR4538', Aliases='', description='microRNA 4538', other_designations='hsa-mir-4538', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105858165, end_position_on_the_genomic_accession=105858242, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616277, CurrentID=0, Status='live', Symbol='MIR4522', Aliases='', description='microRNA 4522', other_designations='hsa-mir-4522', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27293910, end_position_on_the_genomic_accession=27293996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84054, CurrentID=0, Status='live', Symbol='PCDHB19P', Aliases='PCDH-PSI5, PCDHB19', description='protocadherin beta 19 pseudogene', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141240121, end_position_on_the_genomic_accession=141244752, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84055, CurrentID=0, Status='live', Symbol='PCDHGB9P', Aliases='PSI6', description='protocadherin gamma subfamily B, 9 pseudogene', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141440123, end_position_on_the_genomic_accession=141442426, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84056, CurrentID=0, Status='live', Symbol='KATNAL1', Aliases='', description='katanin p60 subunit A-like 1', other_designations='p60 katanin-like 1', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30202630, end_position_on_the_genomic_accession=30307487, orientation='minus', exon_count=14, OMIM=614764, no_hdr0=''),
  84057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84057, CurrentID=0, Status='live', Symbol='MND1', Aliases='GAJ', description='meiotic nuclear divisions 1', other_designations='homolog of yeast MND1|meiotic nuclear divisions 1 homolog', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153344649, end_position_on_the_genomic_accession=153415095, orientation='plus', exon_count=10, OMIM=611422, no_hdr0=''),
  84058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84058, CurrentID=0, Status='live', Symbol='WDR54', Aliases='', description='WD repeat domain 54', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74421733, end_position_on_the_genomic_accession=74425755, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84059, CurrentID=0, Status='live', Symbol='ADGRV1', Aliases='FEB4, GPR98, MASS1, USH2B, USH2C, VLGR1, VLGR1b', description='adhesion G protein-coupled receptor V1', other_designations='G protein-coupled receptor 98|monogenic audiogenic seizure susceptibility protein 1 homolog|monogenic, audiogenic seizure susceptibility 1 homolog|usher syndrome type-2C protein|very large G protein-coupled receptor 1|very large G-protein coupled receptor 1', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90558800, end_position_on_the_genomic_accession=91164223, orientation='plus', exon_count=91, OMIM=602851, no_hdr0=''),
  84060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84060, CurrentID=0, Status='live', Symbol='RBM48', Aliases='C7orf64, HSPC304, MGC16142', description='RNA binding motif protein 48', other_designations='UPF0712 protein C7orf64', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92538475, end_position_on_the_genomic_accession=92539765, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84061, CurrentID=0, Status='live', Symbol='MAGT1', Aliases='IAP, MRX95, OST3B, PRO0756, XMEN, bA217H1.1', description='magnesium transporter 1', other_designations='implantation-associated protein|oligosaccharyltransferase 3 homolog B', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77826364, end_position_on_the_genomic_accession=77895568, orientation='minus', exon_count=10, OMIM=300715, no_hdr0=''),
  84062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84062, CurrentID=0, Status='live', Symbol='DTNBP1', Aliases='BLOC1S8, DBND, HPS7, My031, SDY', description='dystrobrevin binding protein 1', other_designations='BLOC-1 subunit 8|Hermansky-Pudlak syndrome 7 protein|biogenesis of lysosomal organelles complex-1, subunit 8|biogenesis of lysosome-related organelles complex 1 subunit 8|dysbindin-1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15522801, end_position_on_the_genomic_accession=15663058, orientation='minus', exon_count=12, OMIM=607145, no_hdr0=''),
  84063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84063, CurrentID=0, Status='live', Symbol='KIRREL2', Aliases='FILTRIN, NEPH3, NLG1', description='kin of IRRE like 2 (Drosophila)', other_designations='kin of irregular chiasm-like protein 2|nephrin-like gene 1|nephrin-like protein 3', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35851635, end_position_on_the_genomic_accession=35867146, orientation='plus', exon_count=16, OMIM=607762, no_hdr0=''),
  84064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84064, CurrentID=0, Status='live', Symbol='HDHD2', Aliases='3110052N05Rik, HEL-S-301', description='haloacid dehalogenase like hydrolase domain containing 2', other_designations='epididymis secretory protein Li 301|haloacid dehalogenase-like hydrolase domain containing 2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47107410, end_position_on_the_genomic_accession=47150515, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  84065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84065, CurrentID=0, Status='live', Symbol='TMEM222', Aliases='C1orf160', description='transmembrane protein 222', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27322145, end_position_on_the_genomic_accession=27336400, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84066, CurrentID=0, Status='live', Symbol='TEX35', Aliases='C1orf49, TSC24', description='testis expressed 35', other_designations='Testis-Specific Conserved gene 24kDa', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178513077, end_position_on_the_genomic_accession=178523500, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84067, CurrentID=0, Status='live', Symbol='FAM160A2', Aliases='C11orf56', description='family with sequence similarity 160 member A2', other_designations='FHIP|family with sequence similarity 160, member A2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6211334, end_position_on_the_genomic_accession=6234711, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  84068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84068, CurrentID=0, Status='live', Symbol='SLC10A7', Aliases='C4orf13, P7', description='solute carrier family 10 member 7', other_designations='Na(+)/bile acid cotransporter 7|SBF-domain containing protein|solute carrier family 10 (sodium/bile acid cotransporter family), member 7|solute carrier family 10, member 7', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146253981, end_position_on_the_genomic_accession=146521971, orientation='minus', exon_count=14, OMIM=611459, no_hdr0=''),
  84069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84069, CurrentID=0, Status='live', Symbol='PLEKHN1', Aliases='', description='pleckstrin homology domain containing N1', other_designations='PH domain-containing family N member 1|pleckstrin homology domain containing, family N member 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=966497, end_position_on_the_genomic_accession=975108, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  84070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84070, CurrentID=0, Status='live', Symbol='FAM186B', Aliases='C12orf25', description='family with sequence similarity 186 member B', other_designations='family with sequence similarity 186, member B', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49582875, end_position_on_the_genomic_accession=49605650, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84071, CurrentID=0, Status='live', Symbol='ARMC2', Aliases='bA787I22.1', description='armadillo repeat containing 2', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108848414, end_position_on_the_genomic_accession=108998506, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  84072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84072, CurrentID=0, Status='live', Symbol='HORMAD1', Aliases='CT46, NOHMA', description='HORMA domain containing 1', other_designations='cancer/testis antigen 46|newborn ovary HORMA protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150698059, end_position_on_the_genomic_accession=150720909, orientation='minus', exon_count=17, OMIM=609824, no_hdr0=''),
  84073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84073, CurrentID=0, Status='live', Symbol='MYCBPAP', Aliases='AMAP-1, AMAP1', description='MYCBP associated protein', other_designations='AMAM-1|AMY-1-binding protein 1|AMY1-associating protein 1', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50508384, end_position_on_the_genomic_accession=50531502, orientation='plus', exon_count=20, OMIM=609835, no_hdr0=''),
  84074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84074, CurrentID=0, Status='live', Symbol='QRICH2', Aliases='', description='glutamine rich 2', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76274049, end_position_on_the_genomic_accession=76308690, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  84075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84075, CurrentID=0, Status='live', Symbol='FSCB', Aliases='C14orf155', description='fibrous sheath CABYR binding protein', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44504149, end_position_on_the_genomic_accession=44507296, orientation='minus', exon_count=1, OMIM=611779, no_hdr0=''),
  84076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84076, CurrentID=0, Status='live', Symbol='TKTL2', Aliases='', description='transketolase-like 2', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163471095, end_position_on_the_genomic_accession=163473895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84077, CurrentID=0, Status='live', Symbol='C3orf20', Aliases='', description='chromosome 3 open reading frame 20', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14674225, end_position_on_the_genomic_accession=14773036, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  84078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84078, CurrentID=0, Status='live', Symbol='KBTBD7', Aliases='', description='kelch repeat and BTB (POZ) domain containing 7', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41191575, end_position_on_the_genomic_accession=41194566, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84079, CurrentID=0, Status='live', Symbol='ANKRD27', Aliases='PP12899, VARP', description='ankyrin repeat domain 27', other_designations='VPS9 domain-containing protein|VPS9-ankyrin-repeat protein|Vps9 domain and ankyrin-repeat-containing protein|ankyrin repeat domain 27 (VPS9 domain)', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32597001, end_position_on_the_genomic_accession=32675196, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  84080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84080, CurrentID=0, Status='live', Symbol='ENKD1', Aliases='C16orf48, DAKV6410', description='enkurin domain containing 1', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67662947, end_position_on_the_genomic_accession=67666725, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84081, CurrentID=0, Status='live', Symbol='NSRP1', Aliases='CCDC55, HSPC095, NSrp70', description='nuclear speckle splicing regulatory protein 1', other_designations='coiled-coil domain containing 55|coiled-coil domain-containing protein 55|nuclear speckle-related protein 70', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30116807, end_position_on_the_genomic_accession=30186475, orientation='plus', exon_count=8, OMIM=616173, no_hdr0=''),
  100616306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616306, CurrentID=0, Status='live', Symbol='MIR378H', Aliases='', description='microRNA 378h', other_designations='hsa-mir-378h', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154829458, end_position_on_the_genomic_accession=154829540, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84083, CurrentID=0, Status='live', Symbol='ZRANB3', Aliases='4933425L19Rik, AH2', description='zinc finger, RAN-binding domain containing 3', other_designations='annealing helicase 2|zinc finger Ran-binding domain-containing protein 3', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135196969, end_position_on_the_genomic_accession=135531322, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  84084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84084, CurrentID=0, Status='live', Symbol='RAB6C', Aliases='WTH3', description='RAB6C, member RAS oncogene family', other_designations='rab6-like protein WTH3', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129979662, end_position_on_the_genomic_accession=129982738, orientation='plus', exon_count=1, OMIM=612909, no_hdr0=''),
  84085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84085, CurrentID=0, Status='live', Symbol='FBXO30', Aliases='Fbx30', description='F-box protein 30', other_designations='F-box domain protein|F-box only protein, helicase, 18', map_location='6q24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145798121, end_position_on_the_genomic_accession=145814786, orientation='minus', exon_count=4, OMIM=609101, no_hdr0=''),
  100616310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616310, CurrentID=0, Status='live', Symbol='MIR4650-1', Aliases='', description='microRNA 4650-1', other_designations='hsa-mir-4650-1', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67114322, end_position_on_the_genomic_accession=67114397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616311, CurrentID=0, Status='live', Symbol='MIR3973', Aliases='', description='microRNA 3973', other_designations='hsa-mir-3973', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36010098, end_position_on_the_genomic_accession=36010204, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616312, CurrentID=0, Status='live', Symbol='MIR4478', Aliases='mir-4478', description='microRNA 4478', other_designations='hsa-mir-4478', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122120082, end_position_on_the_genomic_accession=122120167, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84089, CurrentID=0, Status='live', Symbol='IGLVVII-41-1', Aliases='lambdavg1', description='immunoglobulin lambda variable (VII)-41-1 (pseudogene)', other_designations='IGLV(VII)-41-1|immunoglobulin lambda variable (VII)-41-1 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22398284, end_position_on_the_genomic_accession=22398582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616314, CurrentID=0, Status='live', Symbol='MIR4750', Aliases='mir-4750', description='microRNA 4750', other_designations='hsa-mir-4750', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49888175, end_position_on_the_genomic_accession=49888230, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84091, CurrentID=0, Status='live', Symbol='IGLVVI-22-1', Aliases='Lvg2', description='immunoglobulin lambda variable (VI)-22-1 (pseudogene)', other_designations='IGLV(VI)-22-1|immunoglobulin lambda variable (VI)-22-1 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22700752, end_position_on_the_genomic_accession=22701041, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616316, CurrentID=0, Status='live', Symbol='MIR4524A', Aliases='MIR4524, hsa-mir-4524a, mir-4524a', description='microRNA 4524a', other_designations='hsa-mir-4524|microRNA 4524', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69099564, end_position_on_the_genomic_accession=69099632, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616317, CurrentID=0, Status='live', Symbol='MIR4777', Aliases='', description='microRNA 4777', other_designations='hsa-mir-4777', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231362708, end_position_on_the_genomic_accession=231362793, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616318, CurrentID=0, Status='live', Symbol='MIR4664', Aliases='', description='microRNA 4664', other_designations='hsa-mir-4664', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143733083, end_position_on_the_genomic_accession=143733153, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616319, CurrentID=0, Status='live', Symbol='MIR4493', Aliases='', description='microRNA 4493', other_designations='hsa-mir-4493', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123381440, end_position_on_the_genomic_accession=123381512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616320, CurrentID=0, Status='live', Symbol='MIR4481', Aliases='', description='microRNA 4481', other_designations='hsa-mir-4481', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12653138, end_position_on_the_genomic_accession=12653197, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616321, CurrentID=0, Status='live', Symbol='MIR378G', Aliases='', description='microRNA 378g', other_designations='hsa-mir-378g', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94745860, end_position_on_the_genomic_accession=94745900, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616322, CurrentID=0, Status='live', Symbol='MIR4682', Aliases='', description='microRNA 4682', other_designations='hsa-mir-4682', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119958513, end_position_on_the_genomic_accession=119958592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616323, CurrentID=0, Status='live', Symbol='MIR4482', Aliases='MIR4482-1, hsa-mir-4482-1', description='microRNA 4482', other_designations='hsa-mir-4482|microRNA 4482-1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104268336, end_position_on_the_genomic_accession=104268405, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84100, CurrentID=0, Status='live', Symbol='ARL6', Aliases='BBS3, RP55', description='ADP ribosylation factor like GTPase 6', other_designations='ADP-ribosylation factor like GTPase 6|ADP-ribosylation factor-like 6|Bardet-Biedl syndrome 3 protein', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97764521, end_position_on_the_genomic_accession=97801242, orientation='plus', exon_count=14, OMIM=608845, no_hdr0=''),
  84101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84101, CurrentID=0, Status='live', Symbol='USP44', Aliases='', description='ubiquitin specific peptidase 44', other_designations='deubiquitinating enzyme 44|ubiquitin specific protease 44|ubiquitin thioesterase 44|ubiquitin thiolesterase 44|ubiquitin-specific-processing protease 44', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95516560, end_position_on_the_genomic_accession=95551529, orientation='minus', exon_count=14, OMIM=610993, no_hdr0=''),
  84102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84102, CurrentID=0, Status='live', Symbol='SLC41A2', Aliases='SLC41A1-L1', description='solute carrier family 41 (magnesium transporter), member 2', other_designations='SLC41A1-like 1|solute carrier family 41, member 2', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104802553, end_position_on_the_genomic_accession=104958740, orientation='minus', exon_count=17, OMIM=610802, no_hdr0=''),
  84103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84103, CurrentID=0, Status='live', Symbol='C4orf17', Aliases='', description='chromosome 4 open reading frame 17', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99511004, end_position_on_the_genomic_accession=99542303, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100616328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616328, CurrentID=0, Status='live', Symbol='MIR4424', Aliases='', description='microRNA 4424', other_designations='hsa-mir-4424', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178677749, end_position_on_the_genomic_accession=178677834, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84105, CurrentID=0, Status='live', Symbol='PCBD2', Aliases='DCOH2, DCOHM, PHS2', description='pterin-4 alpha-carbinolamine dehydratase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1) 2', other_designations='4-alpha-hydroxy-tetrahydropterin dehydratase 2|6-pyruvoyl-tetrahydropterin synthase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1) 2|DcoH-like protein DCoHm|HNF-1-alpha dimerization cofactor|PHS 2|dimerization cofactor of hepatocyte nuclear factor 1 (HNF1) from muscle|dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1)|dimerization cofactor of hepatocyte nuclear factor 1 from muscle|pterin-4 alpha-carbinolamine dehydratase 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134905120, end_position_on_the_genomic_accession=134962646, orientation='plus', exon_count=4, OMIM=609836, no_hdr0=''),
  84106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84106, CurrentID=0, Status='live', Symbol='PRAM1', Aliases='PML-RAR, PRAM-1', description='PML-RARA regulated adaptor molecule 1', other_designations='PML-RARA target gene encoding an Adaptor Molecule-1|PRAM', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8490055, end_position_on_the_genomic_accession=8502654, orientation='minus', exon_count=11, OMIM=606466, no_hdr0=''),
  84107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84107, CurrentID=0, Status='live', Symbol='ZIC4', Aliases='', description='Zic family member 4', other_designations='zinc family member 4 protein HZIC4|zinc finger protein of the cerebellum 4', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147386048, end_position_on_the_genomic_accession=147407476, orientation='minus', exon_count=8, OMIM=608948, no_hdr0=''),
  84108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84108, CurrentID=0, Status='live', Symbol='PCGF6', Aliases='MBLR, RNF134', description='polycomb group ring finger 6', other_designations='Mel18 and Bmi1-like RING finger protein|ring finger protein 134', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103302796, end_position_on_the_genomic_accession=103351134, orientation='minus', exon_count=10, OMIM=607816, no_hdr0=''),
  84109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84109, CurrentID=0, Status='live', Symbol='QRFPR', Aliases='AQ27, GPR103, SP9155', description='pyroglutamylated RFamide peptide receptor', other_designations='G protein-coupled receptor 103|G-protein coupled receptor 103|QRFP receptor|orexigenic neuropeptide QRFP receptor|peptide P518 receptor', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121327211, end_position_on_the_genomic_accession=121381059, orientation='minus', exon_count=8, OMIM=606925, no_hdr0=''),
  100616334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616334, CurrentID=0, Status='live', Symbol='MIR4790', Aliases='', description='microRNA 4790', other_designations='hsa-mir-4790', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5250177, end_position_on_the_genomic_accession=5250255, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=57709, CurrentID=0, Status='live', Symbol='SLC7A14', Aliases='PPP1R142', description='solute carrier family 7 member 14', other_designations='protein phosphatase 1, regulatory subunit 142|solute carrier family 7 (cationic amino acid transporter, y+ system), member 14|solute carrier family 7 (orphan transporter), member 14|solute carrier family 7, member 14', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170459554, end_position_on_the_genomic_accession=170586074, orientation='minus', exon_count=9, OMIM=615720, no_hdr0=''),
  100616336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616336, CurrentID=0, Status='live', Symbol='MIR548AB', Aliases='', description='microRNA 548ab', other_designations='hsa-mir-548ab', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=103524033, end_position_on_the_genomic_accession=103524116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616337, CurrentID=0, Status='live', Symbol='MIR4747', Aliases='mir-4747', description='microRNA 4747', other_designations='hsa-mir-4747', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4932687, end_position_on_the_genomic_accession=4932740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616338, CurrentID=0, Status='live', Symbol='MIR4794', Aliases='', description='microRNA 4794', other_designations='hsa-mir-4794', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64579847, end_position_on_the_genomic_accession=64579923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616339, CurrentID=0, Status='live', Symbol='MIR548AE2', Aliases='', description='microRNA 548ae-2', other_designations='hsa-mir-548ae-2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58530043, end_position_on_the_genomic_accession=58530109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616340, CurrentID=0, Status='live', Symbol='MIR4758', Aliases='', description='microRNA 4758', other_designations='hsa-mir-4758', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62332487, end_position_on_the_genomic_accession=62332557, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616341, CurrentID=0, Status='live', Symbol='MIR4435-2', Aliases='mir-4435-2', description='microRNA 4435-2', other_designations='hsa-mir-4435-2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111321013, end_position_on_the_genomic_accession=111321086, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616342, CurrentID=0, Status='live', Symbol='MIR4638', Aliases='mir-4638', description='microRNA 4638', other_designations='hsa-mir-4638', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181222566, end_position_on_the_genomic_accession=181222633, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616343, CurrentID=0, Status='live', Symbol='MIR4677', Aliases='mir-4677', description='microRNA 4677', other_designations='hsa-mir-4677', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243346176, end_position_on_the_genomic_accession=243346255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616344, CurrentID=0, Status='live', Symbol='MIR3689D2', Aliases='', description='microRNA 3689d-2', other_designations='hsa-mir-3689d-2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134850277, end_position_on_the_genomic_accession=134850356, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616345, CurrentID=0, Status='live', Symbol='MIR4426', Aliases='', description='microRNA 4426', other_designations='hsa-mir-4426', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192716328, end_position_on_the_genomic_accession=192716390, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927066, CurrentID=0, Status='live', Symbol='LOC101927066', Aliases='', description='uncharacterized LOC101927066', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96951864, end_position_on_the_genomic_accession=97446497, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616347, CurrentID=0, Status='live', Symbol='MIR548AI', Aliases='', description='microRNA 548ai', other_designations='hsa-mir-548ai', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99124609, end_position_on_the_genomic_accession=99124696, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84124, CurrentID=0, Status='live', Symbol='ZNF394', Aliases='ZKSCAN14, ZSCAN46', description='zinc finger protein 394', other_designations='truncated zinc finger protein isoform|zinc finger protein 99|zinc finger protein with KRAB and SCAN domains 14', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99493231, end_position_on_the_genomic_accession=99500254, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84125, CurrentID=0, Status='live', Symbol='LRRIQ1', Aliases='', description='leucine-rich repeats and IQ motif containing 1', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=85036321, end_position_on_the_genomic_accession=85272867, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  84126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84126, CurrentID=0, Status='live', Symbol='ATRIP', Aliases='', description='ATR interacting protein', other_designations='ATM and Rad3-related-interacting protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48446710, end_position_on_the_genomic_accession=48465655, orientation='plus', exon_count=15, OMIM=606605, no_hdr0=''),
  100616351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616351, CurrentID=0, Status='live', Symbol='MIR4670', Aliases='', description='microRNA 4670', other_designations='hsa-mir-4670', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92527984, end_position_on_the_genomic_accession=92528058, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84128, CurrentID=0, Status='live', Symbol='WDR75', Aliases='NET16, UTP17', description='WD repeat domain 75', other_designations='UTP17, small subunit (SSU) processome component, homolog', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189441433, end_position_on_the_genomic_accession=189475562, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  84129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84129, CurrentID=0, Status='live', Symbol='ACAD11', Aliases='ACAD-11', description='acyl-CoA dehydrogenase family member 11', other_designations='acyl-CoA dehydrogenase family, member 11|acyl-Coenzyme A dehydrogenase family, member 11', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132558138, end_position_on_the_genomic_accession=132660131, orientation='minus', exon_count=20, OMIM=614288, no_hdr0=''),
  100616354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616354, CurrentID=0, Status='live', Symbol='MIR550A3', Aliases='mir-550a-3', description='microRNA 550a-3', other_designations='hsa-mir-550a-3', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29680734, end_position_on_the_genomic_accession=29680828, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84131, CurrentID=0, Status='live', Symbol='CEP78', Aliases='C9orf81, IP63', description='centrosomal protein 78kDa', other_designations='', map_location='9q21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78236062, end_position_on_the_genomic_accession=78271883, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84132, CurrentID=0, Status='live', Symbol='USP42', Aliases='', description='ubiquitin specific peptidase 42', other_designations='deubiquitinating enzyme 42|ubiquitin specific protease 42|ubiquitin thioesterase 42|ubiquitin thiolesterase 42|ubiquitin-specific-processing protease 42', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6078743, end_position_on_the_genomic_accession=6161564, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  84133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84133, CurrentID=0, Status='live', Symbol='ZNRF3', Aliases='BK747E2.3, RNF203', description='zinc and ring finger 3', other_designations='CTA-292E10.6|RING finger protein 203|novel C3HC4 type Zinc finger (ring finger)|zinc/RING finger protein 3', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28800654, end_position_on_the_genomic_accession=29057488, orientation='plus', exon_count=19, OMIM=612062, no_hdr0=''),
  84134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84134, CurrentID=0, Status='live', Symbol='TOMM40L', Aliases='TOMM40B', description='translocase of outer mitochondrial membrane 40 homolog (yeast)-like', other_designations='protein TOMM40-like|translocase of outer mitochondrial membrane 40 homolog-like', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161225939, end_position_on_the_genomic_accession=161230746, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84135, CurrentID=0, Status='live', Symbol='UTP15', Aliases='NET21', description='UTP15, U3 small nucleolar ribonucleoprotein, homolog (S. cerevisiae)', other_designations='Src-associated protein SAW', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73565741, end_position_on_the_genomic_accession=73583377, orientation='plus', exon_count=13, OMIM=616194, no_hdr0=''),
  100616360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616360, CurrentID=0, Status='live', Symbol='MIR2467', Aliases='mir-2467', description='microRNA 2467', other_designations='hsa-mir-2467', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239351724, end_position_on_the_genomic_accession=239351804, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616361, CurrentID=0, Status='live', Symbol='MIR4775', Aliases='', description='microRNA 4775', other_designations='hsa-mir-4775', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207754807, end_position_on_the_genomic_accession=207754881, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84138, CurrentID=0, Status='live', Symbol='SLC7A6OS', Aliases='', description='solute carrier family 7 member 6 opposite strand', other_designations='ADAMS proteinase-related protein|protein SLC7A6OS|solute carrier family 7 member 6 opposite strand transcript|solute carrier family 7, member 6 opposite strand', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68300615, end_position_on_the_genomic_accession=68310965, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616363, CurrentID=0, Status='live', Symbol='MIR4735', Aliases='', description='microRNA 4735', other_designations='hsa-mir-4735', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196582413, end_position_on_the_genomic_accession=196582481, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84140, CurrentID=0, Status='live', Symbol='FAM161A', Aliases='RP28', description='family with sequence similarity 161 member A', other_designations='family with sequence similarity 161, member A|retinitis pigmentosa 28 (autosomal recessive)', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61744556, end_position_on_the_genomic_accession=61855402, orientation='minus', exon_count=12, OMIM=613596, no_hdr0=''),
  84141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84141, CurrentID=0, Status='live', Symbol='EVA1A', Aliases='FAM176A, TMEM166', description='eva-1 homolog A (C. elegans)', other_designations='family with sequence similarity 176, member A|protein FAM176A|transmembrane protein 166', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75492318, end_position_on_the_genomic_accession=75569722, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84142, CurrentID=0, Status='live', Symbol='FAM175A', Aliases='ABRA1, CCDC98', description='family with sequence similarity 175 member A', other_designations='abraxas protein|coiled-coil domain containing 98|coiled-coil domain-containing protein 98|family with sequence similarity 175, member A', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83460939, end_position_on_the_genomic_accession=83485163, orientation='minus', exon_count=10, OMIM=611143, no_hdr0=''),
  100616367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616367, CurrentID=0, Status='live', Symbol='MIR4467', Aliases='', description='microRNA 4467', other_designations='hsa-mir-4467', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102471469, end_position_on_the_genomic_accession=102471531, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84144, CurrentID=0, Status='live', Symbol='SYDE2', Aliases='', description='synapse defective 1, Rho GTPase, homolog 2 (C. elegans)', other_designations='protein syd-1 homolog 2|synapse defective protein 1 homolog 2', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85157673, end_position_on_the_genomic_accession=85201704, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100616369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616369, CurrentID=0, Status='live', Symbol='MIR4713', Aliases='', description='microRNA 4713', other_designations='hsa-mir-4713', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51242190, end_position_on_the_genomic_accession=51242264, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84146, CurrentID=0, Status='live', Symbol='ZNF644', Aliases='BM-005, MYP21, NatF, ZEP-2', description='zinc finger protein 644', other_designations='zinc finger motif enhancer binding protein 2', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90915298, end_position_on_the_genomic_accession=91022255, orientation='minus', exon_count=11, OMIM=614159, no_hdr0=''),
  100616371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616371, CurrentID=0, Status='live', Symbol='MIR4746', Aliases='mir-4746', description='microRNA 4746', other_designations='hsa-mir-4746', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4445978, end_position_on_the_genomic_accession=4446048, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84148, CurrentID=0, Status='live', Symbol='KAT8', Aliases='MOF, MYST1, ZC2HC8, hMOF', description='K(lysine) acetyltransferase 8', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 1|MYST histone acetyltransferase 1|MYST-1|histone acetyltransferase MYST1|lysine acetyltransferase 8|ortholog of Drosophila males absent on the first (MOF)|probable histone acetyltransferase MYST1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31117664, end_position_on_the_genomic_accession=31131393, orientation='plus', exon_count=12, OMIM=609912, no_hdr0=''),
  100616373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616373, CurrentID=0, Status='live', Symbol='MIR4770', Aliases='', description='microRNA 4770', other_designations='hsa-mir-4770', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=6383906, end_position_on_the_genomic_accession=6383963, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616374, CurrentID=0, Status='live', Symbol='MIR4539', Aliases='', description='microRNA 4539', other_designations='hsa-mir-4539', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105857513, end_position_on_the_genomic_accession=105857572, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616375, CurrentID=0, Status='live', Symbol='MIR4438', Aliases='', description='microRNA 4438', other_designations='hsa-mir-4438', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=213758067, end_position_on_the_genomic_accession=213758159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84152, CurrentID=0, Status='live', Symbol='PPP1R1B', Aliases='DARPP-32, DARPP32', description='protein phosphatase 1 regulatory inhibitor subunit 1B', other_designations='dopamine and cAMP-regulated neuronal phosphoprotein 32', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39626208, end_position_on_the_genomic_accession=39636625, orientation='plus', exon_count=9, OMIM=604399, no_hdr0=''),
  84153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84153, CurrentID=0, Status='live', Symbol='RNASEH2C', Aliases='AGS3, AYP1', description='ribonuclease H2 subunit C', other_designations='RNase H1 small subunit|RNase H2 subunit C|aicardi-Goutieres syndrome 3 protein|ribonuclease H2, subunit C|ribonuclease HI subunit C', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65717673, end_position_on_the_genomic_accession=65720938, orientation='minus', exon_count=4, OMIM=610330, no_hdr0=''),
  84154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84154, CurrentID=0, Status='live', Symbol='RPF2', Aliases='BXDC1, bA397G5.4', description='ribosome production factor 2 homolog', other_designations='brix domain containing 1|brix domain-containing protein 1|homolog of Rpf2|ribosomal processing factor 2 homolog|ribosome biogenesis protein RPF2 homolog', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110982017, end_position_on_the_genomic_accession=111028263, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100616379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616379, CurrentID=0, Status='live', Symbol='MIR4511', Aliases='mir-4511', description='microRNA 4511', other_designations='hsa-mir-4511', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65719246, end_position_on_the_genomic_accession=65719332, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616380, CurrentID=0, Status='live', Symbol='MIR4671', Aliases='', description='microRNA 4671', other_designations='hsa-mir-4671', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234306467, end_position_on_the_genomic_accession=234306539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616381, CurrentID=0, Status='live', Symbol='MIR4456', Aliases='', description='microRNA 4456', other_designations='hsa-mir-4456', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=535840, end_position_on_the_genomic_accession=535882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616382, CurrentID=0, Status='live', Symbol='MIR4509-3', Aliases='', description='microRNA 4509-3', other_designations='hsa-mir-4509-3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28490752, end_position_on_the_genomic_accession=28490845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84159, CurrentID=0, Status='live', Symbol='ARID5B', Aliases='DESRT, MRF-2, MRF2', description='AT-rich interaction domain 5B', other_designations='ARID domain-containing protein 5B|AT rich interactive domain 5B (MRF1-like)|AT-rich interactive domain 5B (MRF1-like)|MRF1-like protein|modulator recognition factor 2 (MRF2)', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61901254, end_position_on_the_genomic_accession=62096948, orientation='plus', exon_count=11, OMIM=608538, no_hdr0=''),
  100616384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616384, CurrentID=0, Status='live', Symbol='MIR548AC', Aliases='', description='microRNA 548ac', other_designations='hsa-mir-548ac', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116560024, end_position_on_the_genomic_accession=116560111, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616385, CurrentID=0, Status='live', Symbol='MIR4732', Aliases='mir-4732', description='microRNA 4732', other_designations='hsa-mir-4732', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28861655, end_position_on_the_genomic_accession=28861730, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84162, CurrentID=0, Status='live', Symbol='KIAA1109', Aliases='FSA, Tweek', description='KIAA1109', other_designations='fragile site-associated protein', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122152333, end_position_on_the_genomic_accession=122362759, orientation='plus', exon_count=93, OMIM=611565, no_hdr0=''),
  84163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84163, CurrentID=0, Status='live', Symbol='GTF2IRD2', Aliases='FP630 alpha, GTF2IRD2A, GTF2IRD2', description='GTF2I repeat domain containing 2', other_designations='GTF2I repeat domain-containing protein 2, alpha|GTF2I repeat domain-containing protein 2A|general transcription factor II i repeat domain 2 alpha|transcription factor GTF2IRD2-alpha', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74796144, end_position_on_the_genomic_accession=74851792, orientation='minus', exon_count=17, OMIM=608899, no_hdr0=''),
  84164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84164, CurrentID=0, Status='live', Symbol='ASCC2', Aliases='ASC1p100, p100', description='activating signal cointegrator 1 complex subunit 2', other_designations='ASC-1 complex subunit P100|trip4 complex subunit p100', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29788608, end_position_on_the_genomic_accession=29838304, orientation='minus', exon_count=23, OMIM=614216, no_hdr0=''),
  100616389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616389, CurrentID=0, Status='live', Symbol='MIR4463', Aliases='', description='microRNA 4463', other_designations='hsa-mir-4463', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75428407, end_position_on_the_genomic_accession=75428473, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84166, CurrentID=0, Status='live', Symbol='NLRC5', Aliases='CLR16.1, NOD27, NOD4', description='NLR family, CARD domain containing 5', other_designations='NOD-like receptor C5|caterpiller protein 16.1|nucleotide-binding oligomerization domain protein 4|nucleotide-binding oligomerization domain, leucine rich repeat and CARD domain containing 5|nucleotide-binding oligomerization domains 27', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56989547, end_position_on_the_genomic_accession=57083526, orientation='plus', exon_count=55, OMIM=613537, no_hdr0=''),
  84167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84167, CurrentID=0, Status='live', Symbol='C19orf44', Aliases='', description='chromosome 19 open reading frame 44', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16496311, end_position_on_the_genomic_accession=16521369, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84168, CurrentID=0, Status='live', Symbol='ANTXR1', Aliases='ATR, GAPO, TEM8', description='anthrax toxin receptor 1', other_designations='2310008J16Rik|2810405N18Rik|tumor endothelial marker 8', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69013144, end_position_on_the_genomic_accession=69249327, orientation='plus', exon_count=22, OMIM=606410, no_hdr0=''),
  100616393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616393, CurrentID=0, Status='live', Symbol='MIR4657', Aliases='mir-4657', description='microRNA 4657', other_designations='hsa-mir-4657', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44881748, end_position_on_the_genomic_accession=44881800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616394, CurrentID=0, Status='live', Symbol='MIR4444-1', Aliases='MIR4444, hsa-mir-4444-1', description='microRNA 4444-1', other_designations='hsa-mir-4444|microRNA 4444', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177212726, end_position_on_the_genomic_accession=177212799, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84171, CurrentID=0, Status='live', Symbol='LOXL4', Aliases='LOXC', description='lysyl oxidase like 4', other_designations='lysyl oxidase related C|lysyl oxidase-like 4 pseudogene|lysyl oxidase-like protein 4|lysyl oxidase-related protein C', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98247686, end_position_on_the_genomic_accession=98268250, orientation='minus', exon_count=16, OMIM=607318, no_hdr0=''),
  84172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84172, CurrentID=0, Status='live', Symbol='POLR1B', Aliases='RPA135, RPA2, Rpo1-2', description='polymerase (RNA) I polypeptide B', other_designations='DNA-directed RNA polymerase I 135 kDa polypeptide|DNA-directed RNA polymerase I 135kDa polypeptide|RNA polymerase I subunit 2|polymerase (RNA) I polypeptide B, 128kDa', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112541915, end_position_on_the_genomic_accession=112577153, orientation='plus', exon_count=16, OMIM=602000, no_hdr0=''),
  84173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84173, CurrentID=0, Status='live', Symbol='ELMOD3', Aliases='DFNB88, LST3, RBED1, RBM29', description='ELMO/CED-12 domain containing 3', other_designations='C330008I15Rik|RNA binding motif and ELMO domain 1|RNA binding motif and ELMO/CED-12 domain 1|RNA binding motif protein 29|RNA-binding motif and ELMO domain-containing protein 1|RNA-binding motif protein 29|RNA-binding protein 29|deafness, autosomal recessive 88|liver-specific organic anion transporter 3TM12|organic anion transporter LST-3b', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85354720, end_position_on_the_genomic_accession=85391752, orientation='plus', exon_count=17, OMIM=615427, no_hdr0=''),
  84174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84174, CurrentID=0, Status='live', Symbol='SLA2', Aliases='C20orf156, MARS, SLAP-2, SLAP2', description='Src-like-adaptor 2', other_designations='Src-like adapter protein-2|modulator of antigen receptor signaling', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36612519, end_position_on_the_genomic_accession=36646216, orientation='minus', exon_count=9, OMIM=606577, no_hdr0=''),
  100616399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616399, CurrentID=0, Status='live', Symbol='MIR4436A', Aliases='mir-4436a', description='microRNA 4436a', other_designations='hsa-mir-4436a', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88812370, end_position_on_the_genomic_accession=88812454, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84176, CurrentID=0, Status='live', Symbol='MYH16', Aliases='MHC20P, MYH5, MYH16', description='myosin, heavy chain 16 pseudogene', other_designations='myosin, heavy polypeptide 16|myosin, heavy polypeptide 5|sarcomeric myosin', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99273301, end_position_on_the_genomic_accession=99297972, orientation='plus', exon_count=16, OMIM=608580, no_hdr0=''),
  100616401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616401, CurrentID=0, Status='live', Symbol='MIR4520-1', Aliases='MIR4520A, hsa-mir-4520-1, mir-4520-1', description='microRNA 4520-1', other_designations='hsa-mir-4520a|microRNA 4520a', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6655440, end_position_on_the_genomic_accession=6655509, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616402, CurrentID=0, Status='live', Symbol='MIR4696', Aliases='', description='microRNA 4696', other_designations='hsa-mir-4696', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74720268, end_position_on_the_genomic_accession=74720337, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84179, CurrentID=0, Status='live', Symbol='MFSD7', Aliases='LP2561', description='major facilitator superfamily domain containing 7', other_designations='MYL5|myosin light polypeptide 5 regulatory protein', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=674690, end_position_on_the_genomic_accession=689468, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100616404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616404, CurrentID=0, Status='live', Symbol='MIR4515', Aliases='mir-4515', description='microRNA 4515', other_designations='hsa-mir-4515', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83067335, end_position_on_the_genomic_accession=83067415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927125, CurrentID=0, Status='live', Symbol='ELFN1-AS1', Aliases='MYCLo-2', description='ELFN1 antisense RNA 1', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1738630, end_position_on_the_genomic_accession=1742310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84182, CurrentID=0, Status='live', Symbol='FAM188B', Aliases='C7orf67', description='family with sequence similarity 188 member B', other_designations='family with sequence similarity 188, member B', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30771417, end_position_on_the_genomic_accession=30892387, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100616407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616407, CurrentID=0, Status='live', Symbol='MIR4443', Aliases='mir-4443', description='microRNA 4443', other_designations='hsa-mir-4443', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48196564, end_position_on_the_genomic_accession=48196616, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616408, CurrentID=0, Status='live', Symbol='MIR5047', Aliases='', description='microRNA 5047', other_designations='hsa-mir-5047', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64501214, end_position_on_the_genomic_accession=64501313, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616409, CurrentID=0, Status='live', Symbol='MIR4711', Aliases='', description='microRNA 4711', other_designations='hsa-mir-4711', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59733227, end_position_on_the_genomic_accession=59733296, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84186, CurrentID=0, Status='live', Symbol='ZCCHC7', Aliases='AIR1, HSPC086', description='zinc finger, CCHC domain containing 7', other_designations='TRAMP-like complex RNA-binding factor ZCCHC7', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37120169, end_position_on_the_genomic_accession=37358149, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  84187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84187, CurrentID=0, Status='live', Symbol='TMEM164', Aliases='bB360B22.3', description='transmembrane protein 164', other_designations='RP13-360B22.2', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110002413, end_position_on_the_genomic_accession=110184038, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  84188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84188, CurrentID=0, Status='live', Symbol='FAR1', Aliases='MLSTD2, PFCRD, SDR10E1', description='fatty acyl-CoA reductase 1', other_designations='fatty acyl CoA reductase 1|male sterility domain-containing protein 2|short chain dehydrogenase/reductase family 10E, member 1', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13668659, end_position_on_the_genomic_accession=13732346, orientation='plus', exon_count=13, OMIM=616107, no_hdr0=''),
  84189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84189, CurrentID=0, Status='live', Symbol='SLITRK6', Aliases='DFNMYP', description='SLIT and NTRK like family member 6', other_designations='4832410J21Rik|SLIT and NTRK-like family, member 6|slit and trk like gene 6', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85792787, end_position_on_the_genomic_accession=85799348, orientation='minus', exon_count=2, OMIM=609681, no_hdr0=''),
  84190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84190, CurrentID=0, Status='live', Symbol='METTL25', Aliases='C12orf26', description='methyltransferase like 25', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82358406, end_position_on_the_genomic_accession=82479239, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84191, CurrentID=0, Status='live', Symbol='FAM96A', Aliases='', description='family with sequence similarity 96 member A', other_designations='family with sequence similarity 96, member A', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64072559, end_position_on_the_genomic_accession=64094008, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616416, CurrentID=0, Status='live', Symbol='MIR4727', Aliases='mir-4727', description='microRNA 4727', other_designations='hsa-mir-4727', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38825838, end_position_on_the_genomic_accession=38825892, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84193, CurrentID=0, Status='live', Symbol='SETD3', Aliases='C14orf154', description='SET domain containing 3', other_designations='SET domain-containing protein 3', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99397746, end_position_on_the_genomic_accession=99481418, orientation='minus', exon_count=16, OMIM=615671, no_hdr0=''),
  100616418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616418, CurrentID=0, Status='live', Symbol='MIR4773-2', Aliases='', description='microRNA 4773-2', other_designations='hsa-mir-4773-2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151368334, end_position_on_the_genomic_accession=151368411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616419, CurrentID=0, Status='live', Symbol='MIR4434', Aliases='', description='microRNA 4434', other_designations='hsa-mir-4434', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64525513, end_position_on_the_genomic_accession=64525565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84196, CurrentID=0, Status='live', Symbol='USP48', Aliases='RAP1GA1, USP31', description='ubiquitin specific peptidase 48', other_designations='deubiquitinating enzyme 48|ubiquitin specific protease 31|ubiquitin thioesterase 48|ubiquitin thiolesterase 48|ubiquitin-specific-processing protease 48', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21678298, end_position_on_the_genomic_accession=21783199, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  84197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84197, CurrentID=0, Status='live', Symbol='POMK', Aliases='MDDGA12, MDDGC12, SGK196', description='protein-O-mannose kinase', other_designations='Sugen kinase 196|probable inactive protein kinase-like protein SgK196|protein kinase-like protein SgK196', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43093506, end_position_on_the_genomic_accession=43123440, orientation='plus', exon_count=5, OMIM=615247, no_hdr0=''),
  100616422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616422, CurrentID=0, Status='live', Symbol='MIR4537', Aliases='', description='microRNA 4537', other_designations='hsa-mir-4537', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105859484, end_position_on_the_genomic_accession=105859553, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616423, CurrentID=0, Status='live', Symbol='MIR4703', Aliases='', description='microRNA 4703', other_designations='hsa-mir-4703', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51552589, end_position_on_the_genomic_accession=51552667, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616424, CurrentID=0, Status='live', Symbol='MIR4707', Aliases='mir-4707', description='microRNA 4707', other_designations='hsa-mir-4707', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22956950, end_position_on_the_genomic_accession=22957029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616425, CurrentID=0, Status='live', Symbol='MIR4748', Aliases='mir-4748', description='microRNA 4748', other_designations='hsa-mir-4748', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10780254, end_position_on_the_genomic_accession=10780335, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616426, CurrentID=0, Status='live', Symbol='MIR4694', Aliases='', description='microRNA 4694', other_designations='hsa-mir-4694', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19760004, end_position_on_the_genomic_accession=19760083, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84203, CurrentID=0, Status='live', Symbol='TXNDC2', Aliases='SPTRX, SPTRX1', description='thioredoxin domain containing 2 (spermatozoa)', other_designations='sperm-specific thioredoxin 1|spermatid-specific thioredoxin-1|sptrx-1|thioredoxin domain-containing 2 (spermatozoa)', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9885726, end_position_on_the_genomic_accession=9888159, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616428, CurrentID=0, Status='live', Symbol='MIR548AM', Aliases='', description='microRNA 548am', other_designations='hsa-mir-548am', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16627012, end_position_on_the_genomic_accession=16627085, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616429, CurrentID=0, Status='live', Symbol='MIR4672', Aliases='', description='microRNA 4672', other_designations='hsa-mir-4672', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127869415, end_position_on_the_genomic_accession=127869495, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84206, CurrentID=0, Status='live', Symbol='MEX3B', Aliases='MEX-3B, RKHD3, RNF195', description='mex-3 RNA binding family member B', other_designations='RING finger protein 195|ring finger and KH domain containing 3', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82041778, end_position_on_the_genomic_accession=82046143, orientation='minus', exon_count=2, OMIM=611008, no_hdr0=''),
  100616431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616431, CurrentID=0, Status='live', Symbol='MIR4431', Aliases='', description='microRNA 4431', other_designations='hsa-mir-4431', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52702522, end_position_on_the_genomic_accession=52702615, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616432, CurrentID=0, Status='live', Symbol='MIR4714', Aliases='', description='microRNA 4714', other_designations='hsa-mir-4714', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98784426, end_position_on_the_genomic_accession=98784502, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616433, CurrentID=0, Status='live', Symbol='MIR4418', Aliases='', description='microRNA 4418', other_designations='hsa-mir-4418', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22266239, end_position_on_the_genomic_accession=22266300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84210, CurrentID=0, Status='live', Symbol='ANKRD20A1', Aliases='ANKRD20A', description='ankyrin repeat domain 20 family member A1', other_designations='ankyrin repeat domain 20 family, member A1|ankyrin repeat domain 20A', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67859286, end_position_on_the_genomic_accession=67902394, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100616435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616435, CurrentID=0, Status='live', Symbol='MIR4801', Aliases='', description='microRNA 4801', other_designations='hsa-mir-4801', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37241910, end_position_on_the_genomic_accession=37241991, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616436, CurrentID=0, Status='live', Symbol='MIR4449', Aliases='', description='microRNA 4449', other_designations='hsa-mir-4449', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52712682, end_position_on_the_genomic_accession=52712747, orientation='plus', exon_count=1, OMIM=614627, no_hdr0=''),
  100616437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616437, CurrentID=0, Status='live', Symbol='MIR1343', Aliases='', description='microRNA 1343', other_designations='hsa-mir-1343', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34941837, end_position_on_the_genomic_accession=34941920, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616438, CurrentID=0, Status='live', Symbol='MIR4632', Aliases='', description='microRNA 4632', other_designations='hsa-mir-4632', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12191713, end_position_on_the_genomic_accession=12191773, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84215, CurrentID=0, Status='live', Symbol='ZNF541', Aliases='', description='zinc finger protein 541', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47520685, end_position_on_the_genomic_accession=47573238, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  84216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84216, CurrentID=0, Status='live', Symbol='TMEM117', Aliases='', description='transmembrane protein 117', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43795764, end_position_on_the_genomic_accession=44437808, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  84217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84217, CurrentID=0, Status='live', Symbol='ZMYND12', Aliases='', description='zinc finger, MYND-type containing 12', other_designations='zinc finger, MYND domain containing 12', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42430329, end_position_on_the_genomic_accession=42456267, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100616442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616442, CurrentID=0, Status='live', Symbol='ERVK-12', Aliases='', description='endogenous retrovirus group K member 12', other_designations='endogenous retrovirus group K, member 12|env protein', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84219, CurrentID=0, Status='live', Symbol='WDR24', Aliases='C16orf21, JFP7', description='WD repeat domain 24', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=684622, end_position_on_the_genomic_accession=690649, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  84220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84220, CurrentID=0, Status='live', Symbol='RGPD5', Aliases='BS-63, BS63, HEL161, RGP5', description='RANBP2-like and GRIP domain containing 5', other_designations='RANBP2-like and GRIP domain-containing protein 5|epididymis luminal protein 161|ran-binding protein 2-like 1|ran-binding protein 2-like 1/2|ranBP2-like 1|ranBP2-like 1/2|ranBP2L1|sperm membrane protein BS-63', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109760607, end_position_on_the_genomic_accession=109857691, orientation='plus', exon_count=31, OMIM=612708, no_hdr0=''),
  84221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84221, CurrentID=0, Status='live', Symbol='SPATC1L', Aliases='C21orf56', description='spermatogenesis and centriole associated 1-like', other_designations='spermatogenesis and centriole-associated protein 1-like protein', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46161148, end_position_on_the_genomic_accession=46185149, orientation='minus', exon_count=6, OMIM=612412, no_hdr0=''),
  100616446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616446, CurrentID=0, Status='live', Symbol='ERVH-1', Aliases='', description='endogenous retrovirus group H member 1', other_designations='HERV-H4p15.2|endogenous retrovirus group H, member 1', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84223, CurrentID=0, Status='live', Symbol='IQCG', Aliases='CFAP122, DRC9', description='IQ motif containing G', other_designations='dynein regulatory complex subunit 9', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197887208, end_position_on_the_genomic_accession=197960015, orientation='minus', exon_count=13, OMIM=612477, no_hdr0=''),
  84224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84224, CurrentID=0, Status='live', Symbol='NBPF3', Aliases='AE2', description='neuroblastoma breakpoint family member 3', other_designations='neuroblastoma breakpoint family, member 3|protein SHIIIa4', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21435318, end_position_on_the_genomic_accession=21484900, orientation='plus', exon_count=24, OMIM=612992, no_hdr0=''),
  84225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84225, CurrentID=0, Status='live', Symbol='ZMYND15', Aliases='SPGF14', description='zinc finger, MYND-type containing 15', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4740015, end_position_on_the_genomic_accession=4746119, orientation='plus', exon_count=14, OMIM=614312, no_hdr0=''),
  84226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84226, CurrentID=0, Status='live', Symbol='C2orf16', Aliases='', description='chromosome 2 open reading frame 16', other_designations='P-S-E-R-S-H-H-S repeats containing', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27576522, end_position_on_the_genomic_accession=27582722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616451, CurrentID=0, Status='live', Symbol='MIR4471', Aliases='', description='microRNA 4471', other_designations='hsa-mir-4471', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100382763, end_position_on_the_genomic_accession=100382845, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616452, CurrentID=0, Status='live', Symbol='MIR2682', Aliases='mir-2682', description='microRNA 2682', other_designations='hsa-mir-2682', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98045242, end_position_on_the_genomic_accession=98045351, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84229, CurrentID=0, Status='live', Symbol='DRC7', Aliases='C16orf50, CCDC135, CFAP50, FAP50', description='dynein regulatory complex subunit 7', other_designations='coiled-coil domain containing 135|coiled-coil domain-containing protein 135|coiled-coil domain-containing protein lobo homolog', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57694801, end_position_on_the_genomic_accession=57731455, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  84230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84230, CurrentID=0, Status='live', Symbol='LRRC8C', Aliases='AD158, FAD158', description='leucine-rich repeat containing 8 family member C', other_designations='factor for adipocyte differentiation 158|hypothetical protein AD158|leucine rich repeat containing 8 family, member C|leucine rich repeat containing 8C|leucine-rich repeat-containing protein 8C', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89615873, end_position_on_the_genomic_accession=89719535, orientation='plus', exon_count=6, OMIM=612889, no_hdr0=''),
  84231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84231, CurrentID=0, Status='live', Symbol='TRAF7', Aliases='RFWD1, RNF119', description='TNF receptor associated factor 7', other_designations='RING finger and WD repeat-containing protein 1|RING finger protein 119|TNF receptor-associated factor 7, E3 ubiquitin protein ligase|ring finger and WD repeat domain 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2155753, end_position_on_the_genomic_accession=2178129, orientation='plus', exon_count=21, OMIM=606692, no_hdr0=''),
  84232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84232, CurrentID=0, Status='live', Symbol='MAF1', Aliases='', description='MAF1 homolog, negative regulator of RNA polymerase III', other_designations='MAF1 negative regulator of RNA polymerase III|homolog of yeast MAF1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144104402, end_position_on_the_genomic_accession=144107612, orientation='plus', exon_count=8, OMIM=610210, no_hdr0=''),
  84233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84233, CurrentID=0, Status='live', Symbol='TMEM126A', Aliases='OPA7', description='transmembrane protein 126A', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85647919, end_position_on_the_genomic_accession=85656553, orientation='plus', exon_count=5, OMIM=612988, no_hdr0=''),
  100616458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616458, CurrentID=0, Status='live', Symbol='MIR5095', Aliases='', description='microRNA 5095', other_designations='hsa-mir-5095', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52934930, end_position_on_the_genomic_accession=52935017, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616459, CurrentID=0, Status='live', Symbol='MIR4745', Aliases='mir-4745', description='microRNA 4745', other_designations='hsa-mir-4745', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=804940, end_position_on_the_genomic_accession=805001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84236, CurrentID=0, Status='live', Symbol='RHBDD1', Aliases='RRP4', description='rhomboid domain containing 1', other_designations='rhomboid domain-containing protein 1|rhomboid-like protein 4', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226835936, end_position_on_the_genomic_accession=226999210, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100616461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616461, CurrentID=0, Status='live', Symbol='ERVH-2', Aliases='HERV-HX', description='endogenous retrovirus group H member 2', other_designations='endogenous retrovirus group H, member 2', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616462, CurrentID=0, Status='live', Symbol='ERVK3-3', Aliases='HML6-c5', description='endogenous retrovirus group K3 member 3', other_designations='endogenous retrovirus group K3, member 3', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84239, CurrentID=0, Status='live', Symbol='ATP13A4', Aliases='', description='ATPase type 13A4', other_designations='P5-ATPase|cation-transporting P5-ATPase', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193398967, end_position_on_the_genomic_accession=193555122, orientation='minus', exon_count=32, OMIM=609556, no_hdr0=''),
  84240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84240, CurrentID=0, Status='live', Symbol='ZCCHC9', Aliases='PPP1R41', description='zinc finger, CCHC domain containing 9', other_designations='protein phosphatase 1, regulatory subunit 41', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81301583, end_position_on_the_genomic_accession=81313344, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100616465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616465, CurrentID=0, Status='live', Symbol='MIR4656', Aliases='', description='microRNA 4656', other_designations='hsa-mir-4656', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4788565, end_position_on_the_genomic_accession=4788639, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616466, CurrentID=0, Status='live', Symbol='MIR4520-2', Aliases='MIR4520B, hsa-mir-4520-2, mir-4520-2', description='microRNA 4520-2', other_designations='hsa-mir-4520b|microRNA 4520b', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6655449, end_position_on_the_genomic_accession=6655502, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84243, CurrentID=0, Status='live', Symbol='ZDHHC18', Aliases='DHHC-18, DHHC18', description='zinc finger, DHHC-type containing 18', other_designations='zinc finger DHHC domain-containing protein 18|zinc finger, DHHC domain containing 18', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26826710, end_position_on_the_genomic_accession=26855720, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100616468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616468, CurrentID=0, Status='live', Symbol='MIR4742', Aliases='mir-4742', description='microRNA 4742', other_designations='hsa-mir-4742', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224398227, end_position_on_the_genomic_accession=224398311, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84245, CurrentID=0, Status='live', Symbol='MRI1', Aliases='MRDI, MTNA, Ypr118w', description='methylthioribose-1-phosphate isomerase 1', other_designations='M1Pi|MTR-1-P isomerase|S-methyl-5-thioribose-1-phosphate isomerase 1|mediator of RhoA-dependent invasion|methylthioribose-1-phosphate isomerase homolog|translation initiation factor eIF-2B subunit alpha/beta/delta-like protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13764492, end_position_on_the_genomic_accession=13774282, orientation='plus', exon_count=7, OMIM=615105, no_hdr0=''),
  84246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84246, CurrentID=0, Status='live', Symbol='MED10', Aliases='L6, NUT2, TRG20', description='mediator complex subunit 10', other_designations='TRG-17|TRG-20|mediator of RNA polymerase II transcription, subunit 10|transformation-related gene 17 protein|transformation-related gene 20 protein', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6371926, end_position_on_the_genomic_accession=6378526, orientation='minus', exon_count=4, OMIM=612382, no_hdr0=''),
  84247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84247, CurrentID=0, Status='live', Symbol='LDOC1L', Aliases='Mar6, Mart6, dJ1033E15.2', description='leucine zipper, down-regulated in cancer 1-like', other_designations='leucine zipper protein down-regulated in cancer cells-like|mammalian retrotransposon-derived protein 6', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44492570, end_position_on_the_genomic_accession=44498125, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84248, CurrentID=0, Status='live', Symbol='FYTTD1', Aliases='UIF', description='forty-two-three domain containing 1', other_designations='forty-two-three domain-containing protein 1|protein 40-2-3', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197749553, end_position_on_the_genomic_accession=197784446, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  84249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84249, CurrentID=0, Status='live', Symbol='PSD2', Aliases='EFA6C', description='pleckstrin and Sec7 domain containing 2', other_designations='6330404E20Rik|pleckstrin homology and SEC7 domain-containing protein 2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139721903, end_position_on_the_genomic_accession=139844466, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84250, CurrentID=0, Status='live', Symbol='SLF1', Aliases='ANKRD32, BRCTD1, BRCTx', description='SMC5-SMC6 complex localization factor 1', other_designations='BRCT domain containing 1|BRCT domain-containing protein 1|ankyrin repeat domain 32|ankyrin repeat domain-containing protein 32|smc5/6 localization factor 1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94615858, end_position_on_the_genomic_accession=94714918, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  84251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84251, CurrentID=0, Status='live', Symbol='SGIP1', Aliases='', description='SH3-domain GRB2-like (endophilin) interacting protein 1', other_designations='endophilin-3-interacting protein', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66533311, end_position_on_the_genomic_accession=66748999, orientation='plus', exon_count=29, OMIM=611540, no_hdr0=''),
  100616476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616476, CurrentID=0, Status='live', Symbol='MIR4446', Aliases='mir-4446', description='microRNA 4446', other_designations='hsa-mir-4446', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113594876, end_position_on_the_genomic_accession=113594942, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84253, CurrentID=0, Status='live', Symbol='GARNL3', Aliases='bA356B19.1', description='GTPase activating Rap/RanGAP domain-like 3', other_designations='GTPase activating RANGAP domain-like 3', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127224471, end_position_on_the_genomic_accession=127393549, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  84254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84254, CurrentID=0, Status='live', Symbol='CAMKK1', Aliases='CAMKKA', description='calcium/calmodulin-dependent protein kinase kinase 1, alpha', other_designations='CAMKK alpha protein|caM-KK 1|caM-KK alpha|caM-kinase IV kinase|caM-kinase kinase 1|caM-kinase kinase alpha|caMKK 1|calcium/calmodulin-dependent protein kinase kinase alpha', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3860315, end_position_on_the_genomic_accession=3894891, orientation='minus', exon_count=20, OMIM=611411, no_hdr0=''),
  84255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84255, CurrentID=0, Status='live', Symbol='SLC37A3', Aliases='', description='solute carrier family 37 member 3', other_designations='solute carrier family 37 (glycerol-3-phosphate transporter), member 3|solute carrier family 37, member 3', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140333752, end_position_on_the_genomic_accession=140398639, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  84256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84256, CurrentID=0, Status='live', Symbol='FLYWCH1', Aliases='', description='FLYWCH-type zinc finger 1', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2911921, end_position_on_the_genomic_accession=2951208, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100616481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616481, CurrentID=0, Status='live', Symbol='MIR4423', Aliases='mir-4423', description='microRNA 4423', other_designations='hsa-mir-4423', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85133794, end_position_on_the_genomic_accession=85133873, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84258, CurrentID=0, Status='live', Symbol='SYT3', Aliases='SytIII', description='synaptotagmin 3', other_designations='synaptotagmin III', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50621860, end_position_on_the_genomic_accession=50658127, orientation='minus', exon_count=12, OMIM=600327, no_hdr0=''),
  84259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84259, CurrentID=0, Status='live', Symbol='DCUN1D5', Aliases='SCCRO5', description='DCN1, defective in cullin neddylation 1, domain containing 5', other_designations='DCUN1 domain-containing protein 5|defective in cullin neddylation protein 1-like protein 5', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103050684, end_position_on_the_genomic_accession=103092215, orientation='minus', exon_count=8, OMIM=616522, no_hdr0=''),
  84260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84260, CurrentID=0, Status='live', Symbol='TCHP', Aliases='TpMs', description='trichoplein, keratin filament binding', other_designations='mitochondrial protein with oncostatic activity|mitostatin|tumor suppressor protein', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109880680, end_position_on_the_genomic_accession=109918069, orientation='plus', exon_count=15, OMIM=612654, no_hdr0=''),
  84261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84261, CurrentID=0, Status='live', Symbol='FBXW9', Aliases='Fbw9', description='F-box and WD repeat domain containing 9', other_designations='F-box and WD-40 domain protein 9|F-box and WD-40 domain-containing protein 9|F-box and WD-40 repeat containing protein 9|specificity factor for SCF ubiquitin ligase', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12688916, end_position_on_the_genomic_accession=12696868, orientation='minus', exon_count=10, OMIM=609074, no_hdr0=''),
  84262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84262, CurrentID=0, Status='live', Symbol='PSMG3', Aliases='C7orf48, PAC3', description='proteasome (prosome, macropain) assembly chaperone 3', other_designations='PAC-3|hPAC3|proteasome assembling chaperone 3', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1567332, end_position_on_the_genomic_accession=1570493, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84263, CurrentID=0, Status='live', Symbol='HSDL2', Aliases='C9orf99, SDR13C1', description='hydroxysteroid dehydrogenase like 2', other_designations='short chain dehydrogenase/reductase family 13C member 1|short chain dehydrogenase/reductase family 13C, member 1', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112379909, end_position_on_the_genomic_accession=112472405, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  84264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84264, CurrentID=0, Status='live', Symbol='HAGHL', Aliases='', description='hydroxyacylglutathione hydrolase-like', other_designations='GLO2-like/ RJD12', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=726936, end_position_on_the_genomic_accession=729715, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  84265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84265, CurrentID=0, Status='live', Symbol='POLR3GL', Aliases='RPC32HOM, flj32422', description='polymerase (RNA) III (DNA directed) polypeptide G (32kD)-like', other_designations='DNA-directed RNA polymerase III subunit G-like|RNA polymerase III subunit C7-like|RPC32-like protein|alternative RNA polymerase III subunit 32', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145959007, end_position_on_the_genomic_accession=145978848, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  84266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84266, CurrentID=0, Status='live', Symbol='ALKBH7', Aliases='ABH7, SPATA11, UNQ6002', description='alkB homolog 7', other_designations='alkB, alkylation repair homolog 7|alkylated DNA repair protein alkB homolog 7|probable alpha-ketoglutarate-dependent dioxygenase ABH7|spermatogenesis associated 11|spermatogenesis cell proliferation related protein|spermatogenesis cell proliferation-related protein|spermatogenesis-associated protein 11', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6372433, end_position_on_the_genomic_accession=6375250, orientation='plus', exon_count=5, OMIM=613305, no_hdr0=''),
  84267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84267, CurrentID=0, Status='live', Symbol='C9orf64', Aliases='', description='chromosome 9 open reading frame 64', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83938307, end_position_on_the_genomic_accession=83957067, orientation='minus', exon_count=6, OMIM=611342, no_hdr0=''),
  84268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84268, CurrentID=0, Status='live', Symbol='RPAIN', Aliases='HRIP, RIP', description='RPA interacting protein', other_designations='RAP interaction protein|nuclear transporter', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5419641, end_position_on_the_genomic_accession=5433020, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84269, CurrentID=0, Status='live', Symbol='CHCHD5', Aliases='C2orf9, MIC14', description='coiled-coil-helix-coiled-coil-helix domain containing 5', other_designations='mitochondrial intermembrane space cysteine motif protein of 14 kDa homolog', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112583606, end_position_on_the_genomic_accession=112589040, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84270, CurrentID=0, Status='live', Symbol='CARD19', Aliases='BinCARD, C9orf89', description='caspase recruitment domain family member 19', other_designations='Bcl10-interacting protein with CARD|caspase recruitment domain family, member 19', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93096129, end_position_on_the_genomic_accession=93113283, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84271, CurrentID=0, Status='live', Symbol='POLDIP3', Aliases='PDIP46, SKAR', description='polymerase (DNA-directed), delta interacting protein 3', other_designations='46 kDa DNA polymerase delta interaction protein|RNA-binding protein P46|S6K1 Aly/REF-like target', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42583721, end_position_on_the_genomic_accession=42614962, orientation='minus', exon_count=10, OMIM=611520, no_hdr0=''),
  84272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84272, CurrentID=0, Status='live', Symbol='YIPF4', Aliases='FinGER4, Nbla11189', description='Yip1 domain family member 4', other_designations='YIP1 family member 4|Yip1 domain family, member 4|putative protein product of Nbla11189', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32277889, end_position_on_the_genomic_accession=32307364, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100616497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616497, CurrentID=0, Status='live', Symbol='ERV9-1', Aliases='pTR2', description='endogenous retrovirus group 9 member 1', other_designations='endogenous retrovirus ERV9|endogenous retrovirus group 9, member 1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84274, CurrentID=0, Status='live', Symbol='COQ5', Aliases='', description='coenzyme Q5, methyltransferase', other_designations='coenzyme Q5 homolog, methyltransferase|ubiquinone biosynthesis methyltransferase COQ5, mitochondrial', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120503279, end_position_on_the_genomic_accession=120529161, orientation='minus', exon_count=8, OMIM=616359, no_hdr0=''),
  84275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84275, CurrentID=0, Status='live', Symbol='SLC25A33', Aliases='BMSC-MCP, PNC1', description='solute carrier family 25 (pyrimidine nucleotide carrier), member 33', other_designations='PNC1 protein|bone marrow stromal cell mitochondrial carrier protein|huBMSC-MCP|mitochondrial carrier protein|novel mitochondrial carrier protein|solute carrier family 25, member 33', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9539469, end_position_on_the_genomic_accession=9582773, orientation='plus', exon_count=9, OMIM=610816, no_hdr0=''),
  84276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84276, CurrentID=0, Status='live', Symbol='NICN1', Aliases='', description='nicolin 1', other_designations='NPCEDRG|PGs5|tubulin polyglutamylase complex subunit 5', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49422333, end_position_on_the_genomic_accession=49429324, orientation='minus', exon_count=6, OMIM=611516, no_hdr0=''),
  84277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84277, CurrentID=0, Status='live', Symbol='DNAJC30', Aliases='WBSCR18', description='DnaJ heat shock protein family (Hsp40) member C30', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 30|Williams Beuren syndrome chromosome region 18|williams-Beuren syndrome chromosomal region 18 protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73680918, end_position_on_the_genomic_accession=73683451, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84278, CurrentID=0, Status='live', Symbol='MFSD14C', Aliases='HIATL2', description='major facilitator superfamily domain containing 14C', other_designations='hippocampus abundant gene transcript-like 2|hippocampus abundant transcript-like 2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96946045, end_position_on_the_genomic_accession=97013580, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84279, CurrentID=0, Status='live', Symbol='PRADC1', Aliases='C2orf7, PAP21', description='protease-associated domain containing 1', other_designations='hPAP21|protease-associated domain-containing glycoprotein 21 kDa|protease-associated domain-containing protein of 21 kDa', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73228006, end_position_on_the_genomic_accession=73233228, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84280, CurrentID=0, Status='live', Symbol='BTBD10', Aliases='GMRP-1, GMRP1', description='BTB (POZ) domain containing 10', other_designations='BTB (broad-complex tramtrack and bric-a-brac/pox virus and zinc finger) domain-containing protein 10|K+ channel tetramerization protein|glucose metabolism-related protein 1', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13387996, end_position_on_the_genomic_accession=13463297, orientation='minus', exon_count=13, OMIM=615933, no_hdr0=''),
  84281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84281, CurrentID=0, Status='live', Symbol='C2orf88', Aliases='smAKAP', description='chromosome 2 open reading frame 88', other_designations='small A-kinase anchoring protein|small membrane AKAP', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189879553, end_position_on_the_genomic_accession=190203484, orientation='plus', exon_count=12, OMIM=615117, no_hdr0=''),
  84282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84282, CurrentID=0, Status='live', Symbol='RNF135', Aliases='L13, MMFD, REUL, Riplet', description='ring finger protein 135', other_designations='RIG-I E3 ubiquitin ligase', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30958921, end_position_on_the_genomic_accession=30999911, orientation='plus', exon_count=8, OMIM=611358, no_hdr0=''),
  84283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84283, CurrentID=0, Status='live', Symbol='TMEM79', Aliases='MATT', description='transmembrane protein 79', other_designations='mattrin', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156282913, end_position_on_the_genomic_accession=156292443, orientation='plus', exon_count=5, OMIM=615531, no_hdr0=''),
  84284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84284, CurrentID=0, Status='live', Symbol='NTPCR', Aliases='C1orf57, HCR-NTPase', description='nucleoside-triphosphatase, cancer-related', other_designations='NTPase|human cancer-related NTPase|nucleoside triphosphate phosphohydrolase|nucleoside-triphosphatase C1orf57', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232950536, end_position_on_the_genomic_accession=232978799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84285, CurrentID=0, Status='live', Symbol='EIF1AD', Aliases='haponin', description='eukaryotic translation initiation factor 1A domain containing', other_designations='eukaryotic translation initiation factor 1A domain-containing protein', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65996545, end_position_on_the_genomic_accession=66002166, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84286, CurrentID=0, Status='live', Symbol='TMEM175', Aliases='hTMEM175', description='transmembrane protein 175', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=932387, end_position_on_the_genomic_accession=958656, orientation='plus', exon_count=12, OMIM=616660, no_hdr0=''),
  84287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84287, CurrentID=0, Status='live', Symbol='ZDHHC16', Aliases='APH2, DHHC-16', description='zinc finger, DHHC-type containing 16', other_designations='Abl-philin 2|zinc finger DHHC domain-containing protein 16|zinc finger, DHHC domain containing 16', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97446131, end_position_on_the_genomic_accession=97457370, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  84288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84288, CurrentID=0, Status='live', Symbol='EFCAB2', Aliases='CFAP200, DRC8', description='EF-hand calcium binding domain 2', other_designations='dynein regulatory complex subunit 8', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244969739, end_position_on_the_genomic_accession=245125228, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84289, CurrentID=0, Status='live', Symbol='ING5', Aliases='p28ING5', description='inhibitor of growth family member 5', other_designations='inhibitor of growth family, member 5', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241700593, end_position_on_the_genomic_accession=241729481, orientation='plus', exon_count=13, OMIM=608525, no_hdr0=''),
  84290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84290, CurrentID=0, Status='live', Symbol='CAPNS2', Aliases='', description='calpain, small subunit 2', other_designations='CSS2|calcium-dependent protease small subunit 2', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55566672, end_position_on_the_genomic_accession=55567680, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927235, CurrentID=0, Status='live', Symbol='LOC101927235', Aliases='', description='uncharacterized LOC101927235', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57701361, end_position_on_the_genomic_accession=57713180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84292, CurrentID=0, Status='live', Symbol='WDR83', Aliases='MORG1', description='WD repeat domain 83', other_designations='MAPK organizer 1|mitogen-activated protein kinase organizer 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12666804, end_position_on_the_genomic_accession=12675832, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84293, CurrentID=0, Status='live', Symbol='FAM213A', Aliases='C10orf58, PAMM', description='family with sequence similarity 213 member A', other_designations='UPF0765 protein C10orf58|family with sequence similarity 213, member A|peroxiredoxin (PRX)-like 2 activated in M-CSF stimulated monocytes|peroxiredoxin-like 2 activated in M-CSF stimulated monocytes|redox-regulatory protein PAMM', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80407829, end_position_on_the_genomic_accession=80432997, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84294, CurrentID=0, Status='live', Symbol='UTP23', Aliases='C8orf53', description='UTP23, small subunit processome component', other_designations='UTP23, small subunit (SSU) processome component, homolog', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116766503, end_position_on_the_genomic_accession=116848959, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84295, CurrentID=0, Status='live', Symbol='PHF6', Aliases='BFLS, BORJ, CENP-31', description='PHD finger protein 6', other_designations='PHD-like zinc finger protein|centromere protein 31', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134373312, end_position_on_the_genomic_accession=134428792, orientation='plus', exon_count=10, OMIM=300414, no_hdr0=''),
  84296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84296, CurrentID=0, Status='live', Symbol='GINS4', Aliases='SLD5', description='GINS complex subunit 4 (Sld5 homolog)', other_designations='SLD5 homolog', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41529206, end_position_on_the_genomic_accession=41545046, orientation='plus', exon_count=8, OMIM=610611, no_hdr0=''),
  101927241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927241, CurrentID=0, Status='live', Symbol='LOC101927241', Aliases='', description='uncharacterized LOC101927241', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48813762, end_position_on_the_genomic_accession=48814547, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84298, CurrentID=0, Status='live', Symbol='LLPH', Aliases='C12orf31, hLLP', description='LLP homolog, long-term synaptic facilitation (Aplysia)', other_designations='human LAPS18-like protein', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66123069, end_position_on_the_genomic_accession=66130753, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84299, CurrentID=0, Status='live', Symbol='MIEN1', Aliases='C17orf37, C35, ORB3, RDX12, XTP4', description='migration and invasion enhancer 1', other_designations='HBV X-transactivated gene 4 protein|HBV XAg-transactivated protein 4|protein C17orf37', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39729156, end_position_on_the_genomic_accession=39730541, orientation='minus', exon_count=4, OMIM=611802, no_hdr0=''),
  84300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84300, CurrentID=0, Status='live', Symbol='UQCC2', Aliases='C6orf125, Cbp6, M19, MNF1, bA6B20.2', description='ubiquinol-cytochrome c reductase complex assembly factor 2', other_designations='breast cancer-associated protein SGA-81M|cytochrome B protein synthesis 6 homolog|mitochondrial nucleoid factor 1|mitochondrial protein M19', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33696761, end_position_on_the_genomic_accession=33711751, orientation='minus', exon_count=4, OMIM=614461, no_hdr0=''),
  84301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84301, CurrentID=0, Status='live', Symbol='DDI2', Aliases='', description='DNA damage inducible 1 homolog 2', other_designations='DDI1, DNA-damage inducible 1, homolog 2|DNA-damage inducible 1 homolog 2|DNA-damage inducible protein 2 (DDI2)', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15617458, end_position_on_the_genomic_accession=15661057, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84302, CurrentID=0, Status='live', Symbol='TMEM246', Aliases='C9orf125', description='transmembrane protein 246', other_designations='transmembrane protein C9orf125', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101473158, end_position_on_the_genomic_accession=101487221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84303, CurrentID=0, Status='live', Symbol='CHCHD6', Aliases='CHCM1, Mic25, PPP1R23', description='coiled-coil-helix-coiled-coil-helix domain containing 6', other_designations='coiled coil helix cristae morphology 1|coiled coil helix cristae morphology protein 1|coiled-coil-helix cristae morphology 1|coiled-coil-helix cristae morphology protein 1|coiled-coil-helix-coiled-coil-helix domain-containing protein 6, mitochondrial|protein phosphatase 1, regulatory subunit 23', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126704220, end_position_on_the_genomic_accession=126960420, orientation='plus', exon_count=16, OMIM=615634, no_hdr0=''),
  84304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84304, CurrentID=0, Status='live', Symbol='NUDT22', Aliases='', description='nudix hydrolase 22', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 22|nudix motif 22', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64226258, end_position_on_the_genomic_accession=64230269, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84305, CurrentID=0, Status='live', Symbol='PYM1', Aliases='PYM, WIBG', description='PYM homolog 1, exon junction complex associated factor', other_designations='protein wibg homolog|within bgcn homolog', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55901413, end_position_on_the_genomic_accession=55927913, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84306, CurrentID=0, Status='live', Symbol='PDCD2L', Aliases='', description='programmed cell death 2-like', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34404398, end_position_on_the_genomic_accession=34426168, orientation='plus', exon_count=7, OMIM=615661, no_hdr0=''),
  84307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84307, CurrentID=0, Status='live', Symbol='ZNF397', Aliases='ZNF47, ZSCAN15', description='zinc finger protein 397', other_designations='zinc finger and SCAN domain-containing protein 15|zinc finger protein 47', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35240718, end_position_on_the_genomic_accession=35258433, orientation='plus', exon_count=7, OMIM=609601, no_hdr0=''),
  103106900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103106900, CurrentID=0, Status='live', Symbol='TRBVC', Aliases='', description='T cell receptor beta variable C', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84309, CurrentID=0, Status='live', Symbol='NUDT16L1', Aliases='SDOS', description='nudix hydrolase 16 like 1', other_designations='NUDT16-like protein 1|nudix (nucleoside diphosphate linked moiety X)-type motif 16-like 1|syndesmos', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4693558, end_position_on_the_genomic_accession=4695859, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84310, CurrentID=0, Status='live', Symbol='C7orf50', Aliases='YCR016W', description='chromosome 7 open reading frame 50', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=978498, end_position_on_the_genomic_accession=1138275, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  84311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84311, CurrentID=0, Status='live', Symbol='MRPL45', Aliases='L45mt, MRP-L45', description='mitochondrial ribosomal protein L45', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38297023, end_position_on_the_genomic_accession=38323218, orientation='plus', exon_count=8, OMIM=611850, no_hdr0=''),
  84312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84312, CurrentID=0, Status='live', Symbol='BRMS1L', Aliases='BRMS1', description='breast cancer metastasis-suppressor 1-like', other_designations='BRMS1-homolog protein p40|BRMS1-like protein p40|breast cancer metastasis-suppressor 1', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35826318, end_position_on_the_genomic_accession=35871963, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84313, CurrentID=0, Status='live', Symbol='VPS25', Aliases='DERP9, EAP20, FAP20', description='vacuolar protein sorting 25 homolog (S. cerevisiae)', other_designations='ELL-associated protein of 20 kDa|ESCRT-II complex subunit VPS25|dermal papilla-derived protein 9', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42773436, end_position_on_the_genomic_accession=42779600, orientation='plus', exon_count=6, OMIM=610907, no_hdr0=''),
  84314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84314, CurrentID=0, Status='live', Symbol='TMEM107', Aliases='GRVS638, PRO1268', description='transmembrane protein 107', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8172978, end_position_on_the_genomic_accession=8176396, orientation='minus', exon_count=5, OMIM=616183, no_hdr0=''),
  84315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84315, CurrentID=0, Status='live', Symbol='MON1A', Aliases='SAND1', description='MON1 homolog A, secretory trafficking associated', other_designations='MON1 homolog A|MON1 secretory trafficking family member A', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49908869, end_position_on_the_genomic_accession=49930357, orientation='minus', exon_count=8, OMIM=611464, no_hdr0=''),
  84316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84316, CurrentID=0, Status='live', Symbol='NAA38', Aliases='LSMD1, PFAAP2', description='N(alpha)-acetyltransferase 38, NatC auxiliary subunit', other_designations='LSM domain containing 1|LSM domain-containing protein 1|N-alpha-acetyltransferase 38, NatC auxiliary subunit|phosphonoformate immuno-associated protein 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7856685, end_position_on_the_genomic_accession=7885283, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84317, CurrentID=0, Status='live', Symbol='CCDC115', Aliases='ccp1', description='coiled-coil domain containing 115', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130337933, end_position_on_the_genomic_accession=130342810, orientation='minus', exon_count=6, OMIM=613734, no_hdr0=''),
  84318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84318, CurrentID=0, Status='live', Symbol='CCDC77', Aliases='', description='coiled-coil domain containing 77', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=389350, end_position_on_the_genomic_accession=442642, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  84319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84319, CurrentID=0, Status='live', Symbol='CMSS1', Aliases='C3orf26', description='cms1 ribosomal small subunit homolog (yeast)', other_designations='', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99817834, end_position_on_the_genomic_accession=100178632, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84320, CurrentID=0, Status='live', Symbol='ACBD6', Aliases='', description='acyl-CoA binding domain containing 6', other_designations='acyl-Coenzyme A binding domain containing 6', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180287385, end_position_on_the_genomic_accession=180502967, orientation='minus', exon_count=9, OMIM=616352, no_hdr0=''),
  84321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84321, CurrentID=0, Status='live', Symbol='THOC3', Aliases='THO3, hTREX45', description='THO complex 3', other_designations='TEX1 homolog', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175938810, end_position_on_the_genomic_accession=175968537, orientation='minus', exon_count=10, OMIM=606929, no_hdr0=''),
  84322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84322, CurrentID=0, Status='live', Symbol='C18orf12', Aliases='HEIL1, HsT2508', description='chromosome 18 open reading frame 12', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48252296, end_position_on_the_genomic_accession=48252837, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927267, CurrentID=0, Status='live', Symbol='LOC101927267', Aliases='', description='uncharacterized LOC101927267', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49292631, end_position_on_the_genomic_accession=49324576, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84324, CurrentID=0, Status='live', Symbol='SARNP', Aliases='CIP29, HCC1, HSPC316, THO1', description='SAP domain containing ribonucleoprotein', other_designations='cytokine induced protein 29 kDa|cytokine-induced protein of 29 kDa|hepatocellular carcinoma 1|nuclear protein Hcc-1|proliferation associated cytokine-inducible protein CIP29|proliferation-associated cytokine-inducible protein CIP29', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55752463, end_position_on_the_genomic_accession=55817756, orientation='minus', exon_count=12, OMIM=610049, no_hdr0=''),
  100873958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873958, CurrentID=0, Status='live', Symbol='FABP5P15', Aliases='FABP5L15', description='fatty acid binding protein 5 pseudogene 15', other_designations='fatty acid binding protein 5-like 15 (pseudogene)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77727845, end_position_on_the_genomic_accession=77728413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84326, CurrentID=0, Status='live', Symbol='C16orf13', Aliases='JFP2', description='chromosome 16 open reading frame 13', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=634427, end_position_on_the_genomic_accession=636366, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84327, CurrentID=0, Status='live', Symbol='ZBED3', Aliases='', description='zinc finger, BED-type containing 3', other_designations='2610005H11Rik|axin-interacting protein|zinc finger, BED domain containing 3', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77076707, end_position_on_the_genomic_accession=77087205, orientation='minus', exon_count=3, OMIM=615250, no_hdr0=''),
  84328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84328, CurrentID=0, Status='live', Symbol='LZIC', Aliases='', description='leucine zipper and CTNNBIP1 domain containing', other_designations='leucine zipper and CTNNBIP1 domain-containing protein|leucine zipper and ICAT homologous domain-containing protein|leucine zipper domain and ICAT homologous domain containing', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9922110, end_position_on_the_genomic_accession=9943402, orientation='minus', exon_count=10, OMIM=610458, no_hdr0=''),
  84329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84329, CurrentID=0, Status='live', Symbol='HVCN1', Aliases='HV1, VSOP', description='hydrogen voltage gated channel 1', other_designations='hydrogen voltage-gated channel 1|voltage sensor domain-only protein', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110648686, end_position_on_the_genomic_accession=110704950, orientation='minus', exon_count=16, OMIM=611227, no_hdr0=''),
  84330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84330, CurrentID=0, Status='live', Symbol='ZNF414', Aliases='ZFP414', description='zinc finger protein 414', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8510578, end_position_on_the_genomic_accession=8514164, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84331, CurrentID=0, Status='live', Symbol='FAM195A', Aliases='C16orf14, c349E10.1', description='family with sequence similarity 195 member A', other_designations='family with sequence similarity 195, member A', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=641849, end_position_on_the_genomic_accession=648474, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84332, CurrentID=0, Status='live', Symbol='DYDC2', Aliases='', description='DPY30 domain containing 2', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80344745, end_position_on_the_genomic_accession=80368118, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84333, CurrentID=0, Status='live', Symbol='PCGF5', Aliases='RNF159', description='polycomb group ring finger 5', other_designations='RING finger protein 159|ring finger protein (C3HC4 type) 159', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91163012, end_position_on_the_genomic_accession=91284331, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  84334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84334, CurrentID=0, Status='live', Symbol='APOPT1', Aliases='APOP, APOP1, C14orf153', description='apoptogenic 1, mitochondrial', other_designations='UPF0671 protein C14orf153', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103562957, end_position_on_the_genomic_accession=103592187, orientation='plus', exon_count=8, OMIM=616003, no_hdr0=''),
  84335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84335, CurrentID=0, Status='live', Symbol='AKT1S1', Aliases='Lobe, PRAS40', description='AKT1 substrate 1 (proline rich)', other_designations='40 kDa proline-rich AKT substrate|AKT1 substrate 1 (proline-rich)', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49869033, end_position_on_the_genomic_accession=49878356, orientation='minus', exon_count=7, OMIM=610221, no_hdr0=''),
  84336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84336, CurrentID=0, Status='live', Symbol='TMEM101', Aliases='', description='transmembrane protein 101', other_designations='putative NF-kappa-B-activating protein 130', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44011188, end_position_on_the_genomic_accession=44024260, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84337, CurrentID=0, Status='live', Symbol='ELOF1', Aliases='ELF1', description='ELF1 homolog, elongation factor 1', other_designations='elongation factor 1 homolog (ELF1, S. cerevisiae)', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11553043, end_position_on_the_genomic_accession=11559236, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927282, CurrentID=0, Status='live', Symbol='LOC101927282', Aliases='', description='uncharacterized LOC101927282', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129724171, end_position_on_the_genomic_accession=129771478, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374099, CurrentID=0, Status='live', Symbol='LOC105374099', Aliases='', description='uncharacterized LOC105374099', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128161729, end_position_on_the_genomic_accession=128162994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84340, CurrentID=0, Status='live', Symbol='GFM2', Aliases='EF-G2mt, EFG2, MRRF2, MST027, MSTP027, RRF2, RRF2mt, hEFG2, mEF-G 2', description='G elongation factor, mitochondrial 2', other_designations='EF-G2mt|RRF2mt|mEF-G 2|mitochondrial elongation factor G2|mitochondrial ribosome recycling factor 2', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74721204, end_position_on_the_genomic_accession=74767371, orientation='minus', exon_count=24, OMIM=606544, no_hdr0=''),
  101927285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927285, CurrentID=0, Status='live', Symbol='LOC101927285', Aliases='', description='uncharacterized LOC101927285', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=59217708, end_position_on_the_genomic_accession=59279400, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84342, CurrentID=0, Status='live', Symbol='COG8', Aliases='CDG2H, DOR1', description='component of oligomeric golgi complex 8', other_designations='COG complex subunit 8|conserved oligomeric golgi complex component 8|dependent on RIC1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69328621, end_position_on_the_genomic_accession=69339623, orientation='minus', exon_count=6, OMIM=606979, no_hdr0=''),
  84343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84343, CurrentID=0, Status='live', Symbol='HPS3', Aliases='SUTAL', description='Hermansky-Pudlak syndrome 3', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149129584, end_position_on_the_genomic_accession=149173518, orientation='plus', exon_count=17, OMIM=606118, no_hdr0=''),
  101927288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927288, CurrentID=0, Status='live', Symbol='MANEA-AS1', Aliases='', description='MANEA antisense RNA 1 (head to head)', other_designations='MANEA antisense RNA 1 (non-protein coding)', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=95560096, end_position_on_the_genomic_accession=95577451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927289, CurrentID=0, Status='live', Symbol='LINC01237', Aliases='', description='long intergenic non-protein coding RNA 1237', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241881363, end_position_on_the_genomic_accession=242078722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927292, CurrentID=0, Status='live', Symbol='LOC101927292', Aliases='', description='uncharacterized LOC101927292', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49911953, end_position_on_the_genomic_accession=49926339, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873962, CurrentID=0, Status='live', Symbol='LINC00278', Aliases='NCRNA00278', description='long intergenic non-protein coding RNA 278', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3002996, end_position_on_the_genomic_accession=3102272, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927295, CurrentID=0, Status='live', Symbol='LINC01360', Aliases='', description='long intergenic non-protein coding RNA 1360', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=73306170, end_position_on_the_genomic_accession=73338877, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927296, CurrentID=0, Status='live', Symbol='LOC101927296', Aliases='', description='uncharacterized LOC101927296', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73521322, end_position_on_the_genomic_accession=73569774, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927297, CurrentID=0, Status='live', Symbol='LOC101927297', Aliases='', description='uncharacterized LOC101927297', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70678677, end_position_on_the_genomic_accession=70686816, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100873963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873963, CurrentID=0, Status='live', Symbol='LINC00279', Aliases='NCRNA00279', description='long intergenic non-protein coding RNA 279', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8638294, end_position_on_the_genomic_accession=8644842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927300, CurrentID=0, Status='live', Symbol='LOC101927300', Aliases='', description='uncharacterized LOC101927300', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99433225, end_position_on_the_genomic_accession=99461216, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927304, CurrentID=0, Status='live', Symbol='LOC101927304', Aliases='', description='uncharacterized LOC101927304', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73594208, end_position_on_the_genomic_accession=73608354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873964, CurrentID=0, Status='live', Symbol='LINC00280', Aliases='NCRNA00280', description='long intergenic non-protein coding RNA 280', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6357219, end_position_on_the_genomic_accession=6361413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374103, CurrentID=0, Status='live', Symbol='LOC105374103', Aliases='', description='uncharacterized LOC105374103', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130143281, end_position_on_the_genomic_accession=130145044, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84364, CurrentID=0, Status='live', Symbol='ARFGAP2', Aliases='IRZ, NBLA10535, ZFP289, ZNF289', description='ADP ribosylation factor GTPase activating protein 2', other_designations='ADP-ribosylation factor GTPase activating protein 2|GTPase-activating protein ZNF289|zinc finger protein 289, ID1 regulated', map_location='11p11.2-p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47164298, end_position_on_the_genomic_accession=47177503, orientation='minus', exon_count=17, OMIM=606908, no_hdr0=''),
  84365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84365, CurrentID=0, Status='live', Symbol='NIFK', Aliases='MKI67IP, Nopp34', description='nucleolar protein interacting with the FHA domain of MKI67', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein|hNIFK|nucleolar phosphoprotein Nopp34|nucleolar protein interacting with the FHA domain of pKi-67', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121726945, end_position_on_the_genomic_accession=121736927, orientation='minus', exon_count=7, OMIM=611970, no_hdr0=''),
  84366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84366, CurrentID=0, Status='live', Symbol='PRAC1', Aliases='C17orf92, PRAC', description='prostate cancer susceptibility candidate 1', other_designations='prostate cancer susceptibility candidate protein 1|prostate, rectum and colon expressed gene protein|small nuclear protein PRAC', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48721719, end_position_on_the_genomic_accession=48722520, orientation='minus', exon_count=2, OMIM=609819, no_hdr0=''),
  101927311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927311, CurrentID=0, Status='live', Symbol='LOC101927311', Aliases='', description='uncharacterized LOC101927311', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14009276, end_position_on_the_genomic_accession=14016018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927313, CurrentID=0, Status='live', Symbol='LOC101927313', Aliases='', description='uncharacterized LOC101927313', other_designations='CTD-2034I21.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50856613, end_position_on_the_genomic_accession=50879720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927314, CurrentID=0, Status='live', Symbol='LOC101927314', Aliases='', description='uncharacterized LOC101927314', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=97305586, end_position_on_the_genomic_accession=97708917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  346517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346517, CurrentID=0, Status='live', Symbol='OR6V1', Aliases='GPR138', description='olfactory receptor family 6 subfamily V member 1', other_designations='olfactory receptor OR7-3|olfactory receptor, family 6, subfamily V, member 1', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143052341, end_position_on_the_genomic_accession=143053282, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927318, CurrentID=0, Status='live', Symbol='LOC101927318', Aliases='', description='uncharacterized LOC101927318', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49951512, end_position_on_the_genomic_accession=49962924, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374105, CurrentID=0, Status='live', Symbol='LOC105374105', Aliases='', description='uncharacterized LOC105374105', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130088881, end_position_on_the_genomic_accession=130095090, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84376, CurrentID=0, Status='live', Symbol='HOOK3', Aliases='HK3', description='hook microtubule-tethering protein 3', other_designations='h-hook3|hHK3|hook homolog 3', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42896890, end_position_on_the_genomic_accession=43030539, orientation='plus', exon_count=22, OMIM=607825, no_hdr0=''),
  101927321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927321, CurrentID=0, Status='live', Symbol='LOC101927321', Aliases='', description='uncharacterized LOC101927321', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73406212, end_position_on_the_genomic_accession=73421024, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927322, CurrentID=0, Status='live', Symbol='LOC101927322', Aliases='', description='uncharacterized LOC101927322', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58670009, end_position_on_the_genomic_accession=58671873, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927323, CurrentID=0, Status='live', Symbol='PPP4R1-AS1', Aliases='', description='PPP4R1 antisense RNA 1', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9614807, end_position_on_the_genomic_accession=9619363, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927324, CurrentID=0, Status='live', Symbol='LINC01475', Aliases='', description='long intergenic non-protein coding RNA 1475', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99526350, end_position_on_the_genomic_accession=99531177, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  346525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346525, CurrentID=0, Status='live', Symbol='OR2A12', Aliases='OR2A12P, OR2A16P', description='olfactory receptor family 2 subfamily A member 12', other_designations='olfactory receptor OR7-10|olfactory receptor, family 2, subfamily A, member 12 pseudogene|olfactory receptor, family 2, subfamily A, member 16 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144095108, end_position_on_the_genomic_accession=144096040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927327, CurrentID=0, Status='live', Symbol='SUMO2P20', Aliases='', description='SUMO2 pseudogene 20', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69987362, end_position_on_the_genomic_accession=69987754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  346528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346528, CurrentID=0, Status='live', Symbol='OR2A1', Aliases='', description='olfactory receptor family 2 subfamily A member 1', other_designations='olfactory receptor OR7-16|olfactory receptor OR7-19|olfactory receptor, family 2, subfamily A, member 1', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144311946, end_position_on_the_genomic_accession=144321105, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927329, CurrentID=0, Status='live', Symbol='LOC101927329', Aliases='', description='uncharacterized LOC101927329', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74123039, end_position_on_the_genomic_accession=74275363, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927330, CurrentID=0, Status='live', Symbol='LOC101927330', Aliases='', description='uncharacterized LOC101927330', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112583053, end_position_on_the_genomic_accession=112585698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927331, CurrentID=0, Status='live', Symbol='LOC101927331', Aliases='', description='uncharacterized LOC101927331', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104412152, end_position_on_the_genomic_accession=104414079, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927332, CurrentID=0, Status='live', Symbol='LINC01582', Aliases='', description='long intergenic non-protein coding RNA 1582', other_designations='TCONS_00023827', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98082979, end_position_on_the_genomic_accession=98088753, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927333, CurrentID=0, Status='live', Symbol='CH17-125A10.2', Aliases='', description='uncharacterized LOC101927333', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144324093, end_position_on_the_genomic_accession=144340912, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927334, CurrentID=0, Status='live', Symbol='LOC101927334', Aliases='', description='uncharacterized LOC101927334', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51017758, end_position_on_the_genomic_accession=51035777, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927335, CurrentID=0, Status='live', Symbol='LOC101927335', Aliases='', description='uncharacterized LOC101927335', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=97816605, end_position_on_the_genomic_accession=98398963, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101927337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927337, CurrentID=0, Status='live', Symbol='LOC101927337', Aliases='', description='uncharacterized LOC101927337', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53088571, end_position_on_the_genomic_accession=53106388, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927339, CurrentID=0, Status='live', Symbol='LOC101927339', Aliases='', description='uncharacterized LOC101927339', other_designations='CTB-186H2.3', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35984459, end_position_on_the_genomic_accession=35990270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927340, CurrentID=0, Status='live', Symbol='LOC101927340', Aliases='', description='uncharacterized LOC101927340', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50143423, end_position_on_the_genomic_accession=50165623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873970, CurrentID=0, Status='live', Symbol='GPC5-AS2', Aliases='', description='GPC5 antisense RNA 2', other_designations='GPC5 antisense RNA 2 (non-protein coding)', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92339794, end_position_on_the_genomic_accession=92352207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927342, CurrentID=0, Status='live', Symbol='LOC101927342', Aliases='', description='uncharacterized LOC101927342', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76011324, end_position_on_the_genomic_accession=76019356, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374109, CurrentID=0, Status='live', Symbol='LOC105374109', Aliases='', description='uncharacterized LOC105374109', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130899411, end_position_on_the_genomic_accession=130915056, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927344, CurrentID=0, Status='live', Symbol='LOC101927344', Aliases='', description='uncharacterized LOC101927344', other_designations='CTA-126B4.7', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42438023, end_position_on_the_genomic_accession=42450776, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927345, CurrentID=0, Status='live', Symbol='LOC101927345', Aliases='', description='putative ankyrin repeat domain-containing protein 20A12 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9357431, end_position_on_the_genomic_accession=9377128, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927346, CurrentID=0, Status='live', Symbol='LOC101927346', Aliases='', description='uncharacterized LOC101927346', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73999805, end_position_on_the_genomic_accession=74005609, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  346547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346547, CurrentID=0, Status='live', Symbol='ERVFC1-1', Aliases='envF(c)2', description='endogenous retrovirus group FC1 member 1', other_designations='Envelope polyprotein|Fc2deltaenv|HERV-F(c)2_7q36.2 provirus ancestral Env polyprotein|TCONS_l2_00027524|endogenous retrovirus group FC1, member 1', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=152713947, end_position_on_the_genomic_accession=152776963, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101927348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927348, CurrentID=0, Status='live', Symbol='LOC101927348', Aliases='', description='uncharacterized LOC101927348', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14018880, end_position_on_the_genomic_accession=14021077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374110, CurrentID=0, Status='live', Symbol='LOC105374110', Aliases='', description='uncharacterized LOC105374110', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131431975, end_position_on_the_genomic_accession=131442423, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927350, CurrentID=0, Status='live', Symbol='LINC01254', Aliases='', description='long intergenic non-protein coding RNA 1254', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10405133, end_position_on_the_genomic_accession=10414370, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927351, CurrentID=0, Status='live', Symbol='LOC101927351', Aliases='', description='uncharacterized LOC101927351', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44390679, end_position_on_the_genomic_accession=44798431, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  101927352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927352, CurrentID=0, Status='live', Symbol='PLS3-AS1', Aliases='', description='PLS3 antisense RNA 1', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115518189, end_position_on_the_genomic_accession=115562731, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927353, CurrentID=0, Status='live', Symbol='LOC101927353', Aliases='', description='uncharacterized LOC101927353', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=906470, end_position_on_the_genomic_accession=915491, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927354, CurrentID=0, Status='live', Symbol='LOC101927354', Aliases='', description='uncharacterized LOC101927354', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7255154, end_position_on_the_genomic_accession=7278069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927355, CurrentID=0, Status='live', Symbol='LOC101927355', Aliases='', description='uncharacterized LOC101927355', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46900386, end_position_on_the_genomic_accession=46901915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927356, CurrentID=0, Status='live', Symbol='LOC101927356', Aliases='', description='uncharacterized LOC101927356', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82009177, end_position_on_the_genomic_accession=82029955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927357, CurrentID=0, Status='live', Symbol='LOC101927357', Aliases='', description='uncharacterized LOC101927357', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122581499, end_position_on_the_genomic_accession=122584611, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927358, CurrentID=0, Status='live', Symbol='LOC101927358', Aliases='', description='uncharacterized LOC101927358', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74371531, end_position_on_the_genomic_accession=74383530, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873973, CurrentID=0, Status='live', Symbol='GPC6-AS2', Aliases='', description='GPC6 antisense RNA 2', other_designations='GPC6 antisense RNA 2 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93818424, end_position_on_the_genomic_accession=93858473, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927360, CurrentID=0, Status='live', Symbol='LINC01103', Aliases='', description='long intergenic non-protein coding RNA 1103', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104488456, end_position_on_the_genomic_accession=104510509, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84417, CurrentID=0, Status='live', Symbol='C2orf40', Aliases='ECRG4', description='chromosome 2 open reading frame 40', other_designations='esophageal cancer related gene 4 protein|esophageal cancer-related gene 4 protein', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106063294, end_position_on_the_genomic_accession=106078155, orientation='plus', exon_count=6, OMIM=611752, no_hdr0=''),
  84418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84418, CurrentID=0, Status='live', Symbol='CYSTM1', Aliases='C5orf32, ORF1-FL49', description='cysteine-rich transmembrane module containing 1', other_designations='UPF0467 protein C5orf32|putative nuclear protein ORF1-FL49', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140175068, end_position_on_the_genomic_accession=140243789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84419, CurrentID=0, Status='live', Symbol='C15orf48', Aliases='FOAP-11, NMES1', description='chromosome 15 open reading frame 48', other_designations='normal mucosa of esophagus specific 1', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45430529, end_position_on_the_genomic_accession=45433449, orientation='plus', exon_count=5, OMIM=608409, no_hdr0=''),
  101927364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927364, CurrentID=0, Status='live', Symbol='LINC01571', Aliases='', description='long intergenic non-protein coding RNA 1571', other_designations='TCONS_00024399', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51762519, end_position_on_the_genomic_accession=51772646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927365, CurrentID=0, Status='live', Symbol='LOC101927365', Aliases='', description='uncharacterized LOC101927365', other_designations='', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99424871, end_position_on_the_genomic_accession=99431315, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927366, CurrentID=0, Status='live', Symbol='LOC101927366', Aliases='', description='uncharacterized LOC101927366', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=192046376, end_position_on_the_genomic_accession=192057724, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927367, CurrentID=0, Status='live', Symbol='LOC101927367', Aliases='', description='uncharacterized LOC101927367', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55449863, end_position_on_the_genomic_accession=55458532, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271272, CurrentID=0, Status='live', Symbol='RPL11P4', Aliases='RPL11_2_1080', description='ribosomal protein L11 pseudogene 4', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92596859, end_position_on_the_genomic_accession=92597289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100873975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873975, CurrentID=0, Status='live', Symbol='CNTN4-AS1', Aliases='', description='CNTN4 antisense RNA 1', other_designations='CNTN4 antisense RNA 1 (non-protein coding)', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3039661, end_position_on_the_genomic_accession=3061145, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927372, CurrentID=0, Status='live', Symbol='LOC101927372', Aliases='', description='uncharacterized LOC101927372', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42501899, end_position_on_the_genomic_accession=42505859, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374114, CurrentID=0, Status='live', Symbol='LOC105374114', Aliases='', description='uncharacterized LOC105374114', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131501938, end_position_on_the_genomic_accession=131689554, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101927374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927374, CurrentID=0, Status='live', Symbol='LOC101927374', Aliases='', description='uncharacterized LOC101927374', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=79411727, end_position_on_the_genomic_accession=79435227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927375, CurrentID=0, Status='live', Symbol='LOC101927375', Aliases='', description='ankyrin repeat domain-containing protein 26-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24541663, end_position_on_the_genomic_accession=24567321, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84432, CurrentID=0, Status='live', Symbol='PROK1', Aliases='EGVEGF, PK1, PRK1', description='prokineticin 1', other_designations='EG-VEGF|black mamba toxin-related protein|endocrine-gland-derived vascular endothelial growth factor|mambakine', map_location='1p21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110451166, end_position_on_the_genomic_accession=110457354, orientation='plus', exon_count=3, OMIM=606233, no_hdr0=''),
  84433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84433, CurrentID=0, Status='live', Symbol='CARD11', Aliases='BENTA, BIMP3, CARMA1, IMD11, PPBL', description='caspase recruitment domain family member 11', other_designations='CARD-containing MAGUK protein 1|bcl10-interacting maguk protein 3|carma 1|caspase recruitment domain family, member 11', map_location='7p22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2906075, end_position_on_the_genomic_accession=3043945, orientation='minus', exon_count=26, OMIM=607210, no_hdr0=''),
  84434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84434, CurrentID=0, Status='live', Symbol='THCYTX', Aliases='FTX', description='Thrombocytosis, familial X-linked', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300331, no_hdr0=''),
  84435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84435, CurrentID=0, Status='live', Symbol='ADGRA1', Aliases='GPR123', description='adhesion G protein-coupled receptor A1', other_designations='G protein-coupled receptor 123|G-protein coupled receptor 123|probable G-protein coupled receptor 123', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133070929, end_position_on_the_genomic_accession=133131675, orientation='plus', exon_count=19, OMIM=612302, no_hdr0=''),
  84436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84436, CurrentID=0, Status='live', Symbol='ZNF528', Aliases='', description='zinc finger protein 528', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52397868, end_position_on_the_genomic_accession=52418408, orientation='plus', exon_count=9, OMIM=615580, no_hdr0=''),
  84437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84437, CurrentID=0, Status='live', Symbol='MSANTD4', Aliases='KIAA1826', description='Myb/SANT-like DNA-binding domain containing 4 with coiled-coils', other_designations='', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106007902, end_position_on_the_genomic_accession=106022254, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84439, CurrentID=0, Status='live', Symbol='HHIPL1', Aliases='KIAA1822, UNQ9245', description='HHIP-like 1', other_designations='ARAR9245', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99604438, end_position_on_the_genomic_accession=99680569, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  84440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84440, CurrentID=0, Status='live', Symbol='RAB11FIP4', Aliases='RAB11-FIP4', description='RAB11 family interacting protein 4 (class II)', other_designations='FIP4-Rab11|arfophilin-2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31391624, end_position_on_the_genomic_accession=31538217, orientation='plus', exon_count=16, OMIM=611999, no_hdr0=''),
  84441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84441, CurrentID=0, Status='live', Symbol='MAML2', Aliases='MAM-3, MAM2, MAM3, MLL-MAML2', description='mastermind like transcriptional coactivator 2', other_designations='mam-2|mastermind-like 2|mastermind-like transcriptional coactivator 2', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95976204, end_position_on_the_genomic_accession=96343245, orientation='minus', exon_count=7, OMIM=607537, no_hdr0=''),
  84443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84443, CurrentID=0, Status='live', Symbol='FRMPD3', Aliases='', description='FERM and PDZ domain containing 3', other_designations='RP5-1070B1.1', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107522450, end_position_on_the_genomic_accession=107605244, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  84444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84444, CurrentID=0, Status='live', Symbol='DOT1L', Aliases='DOT1, KMT4', description='DOT1-like histone H3K79 methyltransferase', other_designations='DOT1-like protein|DOT1-like, histone H3 methyltransferase|H3-K79-HMTase|histone H3-K79 methyltransferase|histone methyltransferase DOT1L|lysine N-methyltransferase 4', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2164127, end_position_on_the_genomic_accession=2232578, orientation='plus', exon_count=31, OMIM=607375, no_hdr0=''),
  84445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84445, CurrentID=0, Status='live', Symbol='LZTS2', Aliases='LAPSER1', description='leucine zipper, putative tumor suppressor 2', other_designations='protein LAPSER1', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100997107, end_position_on_the_genomic_accession=101007836, orientation='plus', exon_count=8, OMIM=610454, no_hdr0=''),
  84446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84446, CurrentID=0, Status='live', Symbol='BRSK1', Aliases='hSAD1', description='BR serine/threonine kinase 1', other_designations='BR serine/threonine-protein kinase 1|SAD1 homolog|SAD1 kinase|SadB kinase short isoform|brain-selective kinase 1|brain-specific serine/threonine-protein kinase 1|protein kinase SAD1A|serine/threonine-protein kinase SAD-B|synapses of Amphids Defective homolog 1', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55284166, end_position_on_the_genomic_accession=55312562, orientation='plus', exon_count=20, OMIM=609235, no_hdr0=''),
  84447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84447, CurrentID=0, Status='live', Symbol='SYVN1', Aliases='DER3, HRD1', description='synoviolin 1', other_designations='HMG-coA reductase degradation 1 homolog|synovial apoptosis inhibitor 1, synoviolin', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65127279, end_position_on_the_genomic_accession=65135178, orientation='minus', exon_count=16, OMIM=608046, no_hdr0=''),
  84448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84448, CurrentID=0, Status='live', Symbol='ABLIM2', Aliases='', description='actin binding LIM protein family member 2', other_designations='abLIM-2|actin binding LIM protein 2|actin binding LIM protein family, member 2|actin-binding LIM protein family member 2', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7965310, end_position_on_the_genomic_accession=8158832, orientation='minus', exon_count=26, OMIM=612544, no_hdr0=''),
  84449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84449, CurrentID=0, Status='live', Symbol='ZNF333', Aliases='', description='zinc finger protein 333', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14689787, end_position_on_the_genomic_accession=14733746, orientation='plus', exon_count=18, OMIM=611811, no_hdr0=''),
  84450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84450, CurrentID=0, Status='live', Symbol='ZNF512', Aliases='', description='zinc finger protein 512', other_designations='', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27582969, end_position_on_the_genomic_accession=27623215, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  84451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84451, CurrentID=0, Status='live', Symbol='KIAA1804', Aliases='MLK4, dJ862P8.3', description='mixed lineage kinase 4', other_designations='MLK4', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233327743, end_position_on_the_genomic_accession=233385148, orientation='plus', exon_count=10, OMIM=614793, no_hdr0=''),
  101927396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927396, CurrentID=0, Status='live', Symbol='LINC01514', Aliases='', description='long intergenic non-protein coding RNA 1514', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101176322, end_position_on_the_genomic_accession=101194147, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927397, CurrentID=0, Status='live', Symbol='LOC101927397', Aliases='', description='uncharacterized LOC101927397', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124131194, end_position_on_the_genomic_accession=124137316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84455, CurrentID=0, Status='live', Symbol='EFCAB7', Aliases='', description='EF-hand calcium binding domain 7', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63523301, end_position_on_the_genomic_accession=63585365, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84456, CurrentID=0, Status='live', Symbol='L3MBTL3', Aliases='MBT-1, MBT1', description='l(3)mbt-like 3 (Drosophila)', other_designations='H-l(3)mbt-like protein 3|l(3)mbt-like protein 3', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129978804, end_position_on_the_genomic_accession=130141449, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  84457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84457, CurrentID=0, Status='live', Symbol='PHYHIPL', Aliases='', description='phytanoyl-CoA 2-hydroxylase interacting protein-like', other_designations='', map_location='10q11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59173371, end_position_on_the_genomic_accession=59247774, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84458, CurrentID=0, Status='live', Symbol='LCOR', Aliases='MLR2', description='ligand dependent nuclear receptor corepressor', other_designations='mblk1-related protein 2', map_location='10q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96832260, end_position_on_the_genomic_accession=96964441, orientation='plus', exon_count=10, OMIM=607698, no_hdr0=''),
  101927403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927403, CurrentID=0, Status='live', Symbol='LOC101927403', Aliases='', description='uncharacterized LOC101927403', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33012924, end_position_on_the_genomic_accession=33016442, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84460, CurrentID=0, Status='live', Symbol='ZMAT1', Aliases='', description='zinc finger, matrin-type 1', other_designations='', map_location='Xq21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101882288, end_position_on_the_genomic_accession=101932135, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  84461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84461, CurrentID=0, Status='live', Symbol='NEURL4', Aliases='', description='neuralized E3 ubiquitin protein ligase 4', other_designations='neuralized homolog 4', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7315632, end_position_on_the_genomic_accession=7329319, orientation='minus', exon_count=29, OMIM=615865, no_hdr0=''),
  346606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346606, CurrentID=0, Status='live', Symbol='MOGAT3', Aliases='DC7, DGAT2L7, MGAT3', description='monoacylglycerol O-acyltransferase 3', other_designations='acyl coenzyme A:monoacylglycerol acyltransferase 3|acyl-CoA:monoacylglycerol acyltransferase 3|diacylglycerol O-acyltransferase candidate 7|diacylglycerol acyltransferase 2-like protein 7', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101195729, end_position_on_the_genomic_accession=101201060, orientation='minus', exon_count=7, OMIM=610184, no_hdr0=''),
  341757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341757, CurrentID=0, Status='live', Symbol='MRPS31P2', Aliases='', description='mitochondrial ribosomal protein S31 pseudogene 2', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19552570, end_position_on_the_genomic_accession=19568822, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  84464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84464, CurrentID=0, Status='live', Symbol='SLX4', Aliases='BTBD12, FANCP, MUS312', description='SLX4 structure-specific endonuclease subunit', other_designations='BTB (POZ) domain containing 12|BTB/POZ domain-containing protein 12|SLX4 structure-specific endonuclease subunit homolog', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3581165, end_position_on_the_genomic_accession=3611598, orientation='minus', exon_count=15, OMIM=613278, no_hdr0=''),
  84465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84465, CurrentID=0, Status='live', Symbol='MEGF11', Aliases='', description='multiple EGF-like-domains 11', other_designations='multiple EGF-like domains protein 11', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65895296, end_position_on_the_genomic_accession=66253737, orientation='minus', exon_count=28, OMIM=612454, no_hdr0=''),
  84466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84466, CurrentID=0, Status='live', Symbol='MEGF10', Aliases='EMARDD', description='multiple EGF-like-domains 10', other_designations='', map_location='5q33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127228947, end_position_on_the_genomic_accession=127461222, orientation='plus', exon_count=30, OMIM=612453, no_hdr0=''),
  84467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84467, CurrentID=0, Status='live', Symbol='FBN3', Aliases='', description='fibrillin 3', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8065402, end_position_on_the_genomic_accession=8149012, orientation='minus', exon_count=64, OMIM=608529, no_hdr0=''),
  101927412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927412, CurrentID=0, Status='live', Symbol='LOC101927412', Aliases='', description='uncharacterized LOC101927412', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=80535755, end_position_on_the_genomic_accession=80646788, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927413, CurrentID=0, Status='live', Symbol='LOC101927413', Aliases='', description='uncharacterized LOC101927413', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108894892, end_position_on_the_genomic_accession=108970266, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927414, CurrentID=0, Status='live', Symbol='LOC101927414', Aliases='', description='uncharacterized LOC101927414', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137603714, end_position_on_the_genomic_accession=137701055, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101927415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927415, CurrentID=0, Status='live', Symbol='LOC101927415', Aliases='', description='uncharacterized LOC101927415', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123581066, end_position_on_the_genomic_accession=123584327, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927416, CurrentID=0, Status='live', Symbol='LOC101927416', Aliases='', description='uncharacterized LOC101927416', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9192493, end_position_on_the_genomic_accession=9194453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927417, CurrentID=0, Status='live', Symbol='LOC101927417', Aliases='', description='uncharacterized LOC101927417', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15328743, end_position_on_the_genomic_accession=15343980, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927418, CurrentID=0, Status='live', Symbol='LOC101927418', Aliases='', description='uncharacterized LOC101927418', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44898908, end_position_on_the_genomic_accession=44911917, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927419, CurrentID=0, Status='live', Symbol='LOC101927419', Aliases='', description='uncharacterized LOC101927419', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101252821, end_position_on_the_genomic_accession=101263550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927420, CurrentID=0, Status='live', Symbol='LOC101927420', Aliases='', description='uncharacterized LOC101927420', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87151423, end_position_on_the_genomic_accession=87156676, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927421, CurrentID=0, Status='live', Symbol='LOC101927421', Aliases='', description='uncharacterized LOC101927421', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125036831, end_position_on_the_genomic_accession=125367734, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927422, CurrentID=0, Status='live', Symbol='GNA14-AS1', Aliases='', description='GNA14 antisense RNA 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77447649, end_position_on_the_genomic_accession=77526762, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927423, CurrentID=0, Status='live', Symbol='PANO1', Aliases='PANO', description='proapoptotic nucleolar protein 1', other_designations='pre-mRNA-splicing factor cwc22-like', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=797511, end_position_on_the_genomic_accession=799185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927424, CurrentID=0, Status='live', Symbol='LOC101927424', Aliases='', description='uncharacterized LOC101927424', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113512451, end_position_on_the_genomic_accession=113515482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873984, CurrentID=0, Status='live', Symbol='CHRM3-AS1', Aliases='', description='CHRM3 antisense RNA 1', other_designations='CHRM3 antisense RNA 1 (non-protein coding)', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239898019, end_position_on_the_genomic_accession=239899872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927429, CurrentID=0, Status='live', Symbol='LOC101927429', Aliases='', description='uncharacterized LOC101927429', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144663793, end_position_on_the_genomic_accession=144676283, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927430, CurrentID=0, Status='live', Symbol='LOC101927430', Aliases='', description='uncharacterized LOC101927430', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67808507, end_position_on_the_genomic_accession=67814822, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927431, CurrentID=0, Status='live', Symbol='LOC101927431', Aliases='', description='uncharacterized LOC101927431', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194730595, end_position_on_the_genomic_accession=194761435, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927432, CurrentID=0, Status='live', Symbol='LOC101927432', Aliases='', description='uncharacterized LOC101927432', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133333577, end_position_on_the_genomic_accession=133386339, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101927433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927433, CurrentID=0, Status='live', Symbol='LINC01255', Aliases='', description='long intergenic non-protein coding RNA 1255', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11488570, end_position_on_the_genomic_accession=11506983, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927434, CurrentID=0, Status='live', Symbol='LOC101927434', Aliases='', description='uncharacterized LOC101927434', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81513880, end_position_on_the_genomic_accession=81557702, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927435, CurrentID=0, Status='live', Symbol='LOC101927435', Aliases='', description='uncharacterized LOC101927435', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130205332, end_position_on_the_genomic_accession=130496873, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  101927436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927436, CurrentID=0, Status='live', Symbol='LOC101927436', Aliases='', description='uncharacterized LOC101927436', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124605057, end_position_on_the_genomic_accession=124609077, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100873986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873986, CurrentID=0, Status='live', Symbol='FGF12-AS1', Aliases='', description='FGF12 antisense RNA 1', other_designations='FGF12 antisense RNA 1 (non-protein coding)', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192238630, end_position_on_the_genomic_accession=192283097, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927438, CurrentID=0, Status='live', Symbol='LOC101927438', Aliases='', description='uncharacterized LOC101927438', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64846130, end_position_on_the_genomic_accession=64863631, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927439, CurrentID=0, Status='live', Symbol='LOC101927439', Aliases='', description='uncharacterized LOC101927439', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83218094, end_position_on_the_genomic_accession=83273324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927440, CurrentID=0, Status='live', Symbol='LOC101927440', Aliases='', description='uncharacterized LOC101927440', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9249724, end_position_on_the_genomic_accession=9257604, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927441, CurrentID=0, Status='live', Symbol='LOC101927441', Aliases='', description='uncharacterized LOC101927441', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15403771, end_position_on_the_genomic_accession=15415522, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84498, CurrentID=0, Status='live', Symbol='FAM120B', Aliases='CCPG, KIAA1838, PGCC1, dJ894D12.1', description='family with sequence similarity 120B', other_designations='PPARG constitutive coactivator 1|PPARgamma constitutive coactivator 1|constitutive coactivator of PPAR-gamma', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170290703, end_position_on_the_genomic_accession=170407071, orientation='plus', exon_count=13, OMIM=612266, no_hdr0=''),
  84501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84501, CurrentID=0, Status='live', Symbol='SPIRE2', Aliases='Spir-2', description='spire-type actin nucleation factor 2', other_designations='spire actin nucleation factor 2|spire family actin nucleation factor 2|spire homolog 2', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89817477, end_position_on_the_genomic_accession=89871319, orientation='plus', exon_count=21, OMIM=609217, no_hdr0=''),
  84502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84502, CurrentID=0, Status='live', Symbol='JPH4', Aliases='JP4, JPHL1', description='junctophilin 4', other_designations='JP-4|junctophilin like 1|junctophilin-like 1 protein', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23568035, end_position_on_the_genomic_accession=23578800, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84503, CurrentID=0, Status='live', Symbol='ZNF527', Aliases='', description='zinc finger protein 527', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37371048, end_position_on_the_genomic_accession=37393066, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84504, CurrentID=0, Status='live', Symbol='NKX6-2', Aliases='GTX, NKX6.2, NKX6B', description='NK6 homeobox 2', other_designations='NK homeobox family 6, B|NK6 transcription factor related, locus 2|glial and testis-specific homeobox protein|homeobox 6B|homeobox protein NK-6 homolog B', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132784816, end_position_on_the_genomic_accession=132786033, orientation='minus', exon_count=3, OMIM=605955, no_hdr0=''),
  101927449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927449, CurrentID=0, Status='live', Symbol='LINC01282', Aliases='', description='long intergenic non-protein coding RNA 1282', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39367285, end_position_on_the_genomic_accession=39391774, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927450, CurrentID=0, Status='live', Symbol='LOC101927450', Aliases='', description='uncharacterized LOC101927450', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79135422, end_position_on_the_genomic_accession=79145666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927451, CurrentID=0, Status='live', Symbol='LOC101927451', Aliases='', description='uncharacterized LOC101927451', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139413719, end_position_on_the_genomic_accession=139454187, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927452, CurrentID=0, Status='live', Symbol='LOC101927452', Aliases='', description='uncharacterized LOC101927452', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143498954, end_position_on_the_genomic_accession=143515940, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  346653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346653, CurrentID=0, Status='live', Symbol='FAM71F2', Aliases='FAM137B', description='family with sequence similarity 71 member F2', other_designations='family with sequence similarity 137, member B|family with sequence similarity 71, member F2', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128671636, end_position_on_the_genomic_accession=128702262, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  101927454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927454, CurrentID=0, Status='live', Symbol='LOC101927454', Aliases='', description='uncharacterized LOC101927454', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196171539, end_position_on_the_genomic_accession=196172849, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927455, CurrentID=0, Status='live', Symbol='TMEM108-AS1', Aliases='', description='TMEM108 antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133246448, end_position_on_the_genomic_accession=133257121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84513, CurrentID=0, Status='live', Symbol='PLPP5', Aliases='DPPL1, HTPAP, PPAPDC1B', description='phospholipid phosphatase 5', other_designations='diacylglycerol pyrophosphate like 1|phosphatidic acid phosphatase type 2 domain containing 1B|phosphatidic acid phosphatase type 2 domain-containing protein 1B', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38261790, end_position_on_the_genomic_accession=38269437, orientation='minus', exon_count=9, OMIM=610626, no_hdr0=''),
  84514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84514, CurrentID=0, Status='live', Symbol='GHDC', Aliases='D11LGP1, LGP1', description='GH3 domain containing', other_designations='homolog of mouse LGP1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42189087, end_position_on_the_genomic_accession=42194532, orientation='minus', exon_count=10, OMIM=608587, no_hdr0=''),
  84515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84515, CurrentID=0, Status='live', Symbol='MCM8', Aliases='C20orf154, POF10, dJ967N21.5', description='minichromosome maintenance 8 homologous recombination repair factor', other_designations='DNA replication licensing factor MCM8|MCM8 minichromosome maintenance deficient 8|REC homolog|minichromosome maintenance complex component 8', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5950652, end_position_on_the_genomic_accession=5995206, orientation='plus', exon_count=21, OMIM=608187, no_hdr0=''),
  84516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84516, CurrentID=0, Status='live', Symbol='DCTN5', Aliases='', description='dynactin subunit 5', other_designations='dynactin 4|dynactin 5 (p25)|dynactin subunit p25', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23641366, end_position_on_the_genomic_accession=23673747, orientation='plus', exon_count=8, OMIM=612962, no_hdr0=''),
  84517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84517, CurrentID=0, Status='live', Symbol='ACTRT3', Aliases='ARP-T3, ARPM1', description='actin-related protein T3', other_designations='actin related protein M1|actin-related protein M1', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169766921, end_position_on_the_genomic_accession=169769895, orientation='minus', exon_count=2, OMIM=608534, no_hdr0=''),
  84518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84518, CurrentID=0, Status='live', Symbol='CNFN', Aliases='PLAC8L2', description='cornifelin', other_designations='cornefied envelope protein cornefilin', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42387019, end_position_on_the_genomic_accession=42390317, orientation='minus', exon_count=5, OMIM=611764, no_hdr0=''),
  84519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84519, CurrentID=0, Status='live', Symbol='ACRBP', Aliases='CT23, OY-TES-1, SP32', description='acrosin binding protein', other_designations='cancer/testis antigen 23|cancer/testis antigen OY-TES-1|proacrosin binding protein sp32', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6638076, end_position_on_the_genomic_accession=6647414, orientation='minus', exon_count=10, OMIM=608352, no_hdr0=''),
  84520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84520, CurrentID=0, Status='live', Symbol='C14orf142', Aliases='PNAS-127', description='chromosome 14 open reading frame 142', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93202892, end_position_on_the_genomic_accession=93207114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927465, CurrentID=0, Status='live', Symbol='CLYBL-AS1', Aliases='', description='CLYBL antisense RNA 1', other_designations='CLYBL antisense RNA 1 (non-protein coding)', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99725668, end_position_on_the_genomic_accession=99729001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84522, CurrentID=0, Status='live', Symbol='JAGN1', Aliases='GL009, SCN6', description='jagunal homolog 1', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9890587, end_position_on_the_genomic_accession=9894349, orientation='plus', exon_count=2, OMIM=616012, no_hdr0=''),
  100873991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873991, CurrentID=0, Status='live', Symbol='IGBP1-AS1', Aliases='', description='IGBP1 antisense RNA 1', other_designations='IGBP1 antisense RNA 1 (non-protein coding)', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84524, CurrentID=0, Status='live', Symbol='ZC3H8', Aliases='Fliz1, ZC3HDC8', description='zinc finger CCCH-type containing 8', other_designations='zinc finger CCCH type domain containing 8|zinc finger CCCH-type domain containing 8', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112215677, end_position_on_the_genomic_accession=112255101, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  84525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84525, CurrentID=0, Status='live', Symbol='HOPX', Aliases='CAMEO, HOD, HOP, LAGY, NECC1, OB1, SMAP31, TOTO', description='HOP homeobox', other_designations='lung cancer-associated Y protein|not expressed in choriocarcinoma clone 1|not expressed in choriocarcinoma protein 1|odd homeobox 1 protein|odd homeobox protein 1', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56647988, end_position_on_the_genomic_accession=56681706, orientation='minus', exon_count=9, OMIM=607275, no_hdr0=''),
  84527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84527, CurrentID=0, Status='live', Symbol='ZNF559', Aliases='NBLA00121', description='zinc finger protein 559', other_designations='putative protein product of Nbla00121', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9323772, end_position_on_the_genomic_accession=9343845, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84528, CurrentID=0, Status='live', Symbol='RHOXF2', Aliases='CT107, PEPP-2, PEPP2, THG1', description='Rhox homeobox family member 2', other_designations='PEPP subfamily gene 2|Rhox homeobox family, member 2|cancer/testis antigen 107|homeobox protein from AL590526|paired-like homeobox protein PEPP-2|testis homeobox gene 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120158539, end_position_on_the_genomic_accession=120164027, orientation='plus', exon_count=4, OMIM=300447, no_hdr0=''),
  84529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84529, CurrentID=0, Status='live', Symbol='C15orf41', Aliases='HH114', description='chromosome 15 open reading frame 41', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36579603, end_position_on_the_genomic_accession=36810260, orientation='plus', exon_count=19, OMIM=615626, no_hdr0=''),
  84530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84530, CurrentID=0, Status='live', Symbol='SRRM4', Aliases='KIAA1853, MU-MB-2.76, nSR100', description='serine/arginine repetitive matrix 4', other_designations='medulloblastoma antigen MU-MB-2.76|neural-specific SR-related protein of 100 kDa|neural-specific serine/arginine repetitive splicing factor of 100 kDa', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118981486, end_position_on_the_genomic_accession=119163053, orientation='plus', exon_count=13, OMIM=613103, no_hdr0=''),
  84532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84532, CurrentID=0, Status='live', Symbol='ACSS1', Aliases='ACAS2L, ACECS1, AceCS2L', description='acyl-CoA synthetase short-chain family member 1', other_designations='acetate--CoA ligase 2', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25006230, end_position_on_the_genomic_accession=25058182, orientation='minus', exon_count=15, OMIM=614355, no_hdr0=''),
  101927477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927477, CurrentID=0, Status='live', Symbol='LINC01507', Aliases='', description='long intergenic non-protein coding RNA 1507', other_designations='', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79824530, end_position_on_the_genomic_accession=80034555, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927478, CurrentID=0, Status='live', Symbol='LOC101927478', Aliases='', description='uncharacterized LOC101927478', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229508430, end_position_on_the_genomic_accession=229514087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873993, CurrentID=0, Status='live', Symbol='ITIH4-AS1', Aliases='', description='ITIH4 antisense RNA 1', other_designations='ITIH4 antisense RNA 1 (non-protein coding)', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52823935, end_position_on_the_genomic_accession=52825314, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84536, CurrentID=0, Status='live', Symbol='LINC01547', Aliases='C21orf67, C21orf69, PRED54', description='long intergenic non-protein coding RNA 1547', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44933284, end_position_on_the_genomic_accession=44939913, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927481, CurrentID=0, Status='live', Symbol='LOC101927481', Aliases='', description='uncharacterized LOC101927481', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70335439, end_position_on_the_genomic_accession=70352459, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927482, CurrentID=0, Status='live', Symbol='LOC101927482', Aliases='', description='uncharacterized LOC101927482', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196260024, end_position_on_the_genomic_accession=196264204, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84539, CurrentID=0, Status='live', Symbol='MCHR2', Aliases='GPR145, GPRv17, MCH-2R, MCH-R2, MCH2, MCH2R, MCHR-2, SLT', description='melanin concentrating hormone receptor 2', other_designations='G protein-coupled receptor 145|G protein-coupled receptor slt|G-protein coupled receptor 145|MCH receptor 2|melanin-concentrating hormone 2', map_location='6q16', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99919910, end_position_on_the_genomic_accession=99994238, orientation='minus', exon_count=6, OMIM=606111, no_hdr0=''),
  101927484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927484, CurrentID=0, Status='live', Symbol='LOC101927484', Aliases='', description='uncharacterized LOC101927484', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55023488, end_position_on_the_genomic_accession=55035740, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84541, CurrentID=0, Status='live', Symbol='KBTBD8', Aliases='TA-KRP, TAKRP', description='kelch repeat and BTB (POZ) domain containing 8', other_designations='T-cell activation kelch repeat protein', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66998303, end_position_on_the_genomic_accession=67011211, orientation='plus', exon_count=4, OMIM=616607, no_hdr0=''),
  84542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84542, CurrentID=0, Status='live', Symbol='KIAA1841', Aliases='', description='KIAA1841', other_designations='', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61065870, end_position_on_the_genomic_accession=61157942, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  101927487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927487, CurrentID=0, Status='live', Symbol='LINC01609', Aliases='', description='long intergenic non-protein coding RNA 1609', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111098961, end_position_on_the_genomic_accession=111236203, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927488, CurrentID=0, Status='live', Symbol='LOC101927488', Aliases='', description='uncharacterized LOC101927488', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126272517, end_position_on_the_genomic_accession=126285175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  346689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346689, CurrentID=0, Status='live', Symbol='KLRG2', Aliases='CLEC15B', description='killer cell lectin-like receptor subfamily G, member 2', other_designations='C-type lectin domain family 15 member B|C-type lectin domain family 15, member B', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139426933, end_position_on_the_genomic_accession=139483711, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84546, CurrentID=0, Status='live', Symbol='SNORD35B', Aliases='RNU35B, U35B', description='small nucleolar RNA, C/D box 35B', other_designations='RNA, U35B small nucleolar|RNA, small nucleolar|snoRNA', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49497719, end_position_on_the_genomic_accession=49497806, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84547, CurrentID=0, Status='live', Symbol='PGBD1', Aliases='HUCEP-4, SCAND4, dJ874C20.4', description='piggyBac transposable element derived 1', other_designations='cerebral protein 4|cerebral protein-4', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28281537, end_position_on_the_genomic_accession=28302549, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84548, CurrentID=0, Status='live', Symbol='TMEM185A', Aliases='CXorf13, FAM11A, FRAXF, ee3', description='transmembrane protein 185A', other_designations='family with sequence similarity 11, member A|fragile site, folic acid type, rare, fra(X)(q28) F', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149596556, end_position_on_the_genomic_accession=149631912, orientation='minus', exon_count=7, OMIM=300031, no_hdr0=''),
  84549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84549, CurrentID=0, Status='live', Symbol='MAK16', Aliases='MAK16L, RBM13', description='MAK16 homolog', other_designations='NNP78|RNA binding motif protein 13|RNA binding protein', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33485167, end_position_on_the_genomic_accession=33501260, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101927494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927494, CurrentID=0, Status='live', Symbol='LOC101927494', Aliases='', description='uncharacterized LOC101927494', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86481943, end_position_on_the_genomic_accession=86496996, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927495, CurrentID=0, Status='live', Symbol='LOC101927495', Aliases='', description='uncharacterized LOC101927495', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61588494, end_position_on_the_genomic_accession=61607546, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84552, CurrentID=0, Status='live', Symbol='PARD6G', Aliases='PAR-6G, PAR6gamma', description='par-6 family cell polarity regulator gamma', other_designations='PAR-6 gamma protein|PAR6D|par-6 partitioning defective 6 homolog gamma', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80157232, end_position_on_the_genomic_accession=80247514, orientation='minus', exon_count=3, OMIM=608976, no_hdr0=''),
  84553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84553, CurrentID=0, Status='live', Symbol='FAXC', Aliases='C6orf168, dJ273F20', description='failed axon connections homolog', other_designations='', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99272915, end_position_on_the_genomic_accession=99350122, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927498, CurrentID=0, Status='live', Symbol='LINC01361', Aliases='', description='long intergenic non-protein coding RNA 1361', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=82973882, end_position_on_the_genomic_accession=82986208, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374135, CurrentID=0, Status='live', Symbol='LOC105374135', Aliases='', description='uncharacterized LOC105374135', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141932514, end_position_on_the_genomic_accession=141938356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84557, CurrentID=0, Status='live', Symbol='MAP1LC3A', Aliases='ATG8E, LC3, LC3A, MAP1ALC3, MAP1BLC3', description='microtubule associated protein 1 light chain 3 alpha', other_designations='MAP1 light chain 3-like protein 1|MAP1A/1B light chain 3 A|MAP1A/MAP1B LC3 A|MAP1A/MAP1B light chain 3 A|autophagy-related ubiquitin-like modifier LC3 A|microtubule-associated protein 1 light chain 3 alpha|microtubule-associated proteins 1A/1B light chain 3', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34546830, end_position_on_the_genomic_accession=34560345, orientation='plus', exon_count=7, OMIM=601242, no_hdr0=''),
  346702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346702, CurrentID=0, Status='live', Symbol='PRSS51', Aliases='', description='protease, serine 51', other_designations='protease, serine, 51', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=10353836, end_position_on_the_genomic_accession=10405089, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84559, CurrentID=0, Status='live', Symbol='GOLGA2P2Y', Aliases='GOLGA2LY, GOLGA2LY1, GOLGA2P2', description='golgin A2 pseudogene 2, Y-linked', other_designations='golgi autoantigen, golgin subfamily a, 2-like, Y-linked 1', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24209967, end_position_on_the_genomic_accession=24214831, orientation='minus', exon_count=9, OMIM=400035, no_hdr0=''),
  84560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84560, CurrentID=0, Status='live', Symbol='MT4', Aliases='MT-4, MT-IV, MTIV', description='metallothionein 4', other_designations='metallothionein IV|metallothionein-IV', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56565049, end_position_on_the_genomic_accession=56568957, orientation='plus', exon_count=3, OMIM=606206, no_hdr0=''),
  84561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84561, CurrentID=0, Status='live', Symbol='SLC12A8', Aliases='CCC9', description='solute carrier family 12, member 8', other_designations='cation-chloride cotransporter 9|solute carrier family 12 (sodium/potassium/chloride transporters), member 8', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125082636, end_position_on_the_genomic_accession=125212765, orientation='minus', exon_count=14, OMIM=611316, no_hdr0=''),
  101927506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927506, CurrentID=0, Status='live', Symbol='LOC101927506', Aliases='', description='uncharacterized LOC101927506', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63713, end_position_on_the_genomic_accession=112953, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  346708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346708, CurrentID=0, Status='live', Symbol='OR7E8P', Aliases='OR11-11a', description='olfactory receptor family 7 subfamily E member 8 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 8 pseudogene|seven transmembrane helix receptor', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12684140, end_position_on_the_genomic_accession=12685165, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  84565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84565, CurrentID=0, Status='live', Symbol='LCS1', Aliases='CHLS, CHSL, LCS', description='lymphedema-cholestasis syndrome 1', other_designations='', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=214900, no_hdr0=''),
  101927510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927510, CurrentID=0, Status='live', Symbol='LOC101927510', Aliases='', description='uncharacterized LOC101927510', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57252769, end_position_on_the_genomic_accession=57257037, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  346711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346711, CurrentID=0, Status='live', Symbol='MRPL49P2', Aliases='', description='mitochondrial ribosomal protein L49 pseudogene 2', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16374619, end_position_on_the_genomic_accession=16375086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927512, CurrentID=0, Status='live', Symbol='LOC101927512', Aliases='', description='uncharacterized LOC101927512', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50162769, end_position_on_the_genomic_accession=50166247, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84569, CurrentID=0, Status='live', Symbol='LYZL1', Aliases='KAAG648, LYC2, PRO1278, bA534G20.1', description='lysozyme like 1', other_designations='lysozyme-like 1', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29289051, end_position_on_the_genomic_accession=29318328, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84570, CurrentID=0, Status='live', Symbol='COL25A1', Aliases='AMY, CFEOM5, CLAC, CLAC-P, CLACP', description='collagen, type XXV, alpha 1', other_designations='alzheimer disease amyloid-associated protein|collagen-like Alzheimer amyloid plaque component|collagenous Alzheimer amyloid plaque component', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108808725, end_position_on_the_genomic_accession=109302752, orientation='minus', exon_count=52, OMIM=610004, no_hdr0=''),
  84572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84572, CurrentID=0, Status='live', Symbol='GNPTG', Aliases='C16orf27, GNPTAG, LP2537, RJD9', description='N-acetylglucosamine-1-phosphate transferase, gamma subunit', other_designations='UDP-N-acetylglucosamine-1-phosphotransferase subunit gamma|glcNAc-1-phosphotransferase subunit gamma', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1351899, end_position_on_the_genomic_accession=1363351, orientation='plus', exon_count=11, OMIM=607838, no_hdr0=''),
  105374138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374138, CurrentID=0, Status='live', Symbol='LOC105374138', Aliases='', description='uncharacterized LOC105374138', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143084002, end_position_on_the_genomic_accession=143108884, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927518, CurrentID=0, Status='live', Symbol='LOC101927518', Aliases='', description='uncharacterized LOC101927518', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86659067, end_position_on_the_genomic_accession=86737260, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271273, CurrentID=0, Status='live', Symbol='RPL12P28', Aliases='RPL12_7_1019', description='ribosomal protein L12 pseudogene 28', other_designations='', map_location='10p', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5862193, end_position_on_the_genomic_accession=5862594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927521, CurrentID=0, Status='live', Symbol='LOC101927521', Aliases='', description='uncharacterized LOC101927521', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21363298, end_position_on_the_genomic_accession=21451054, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927522, CurrentID=0, Status='live', Symbol='LOC101927522', Aliases='', description='uncharacterized LOC101927522', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35418954, end_position_on_the_genomic_accession=35434707, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927523, CurrentID=0, Status='live', Symbol='LOC101927523', Aliases='', description='uncharacterized LOC101927523', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104474939, end_position_on_the_genomic_accession=104480275, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927524, CurrentID=0, Status='live', Symbol='MEOX2-AS1', Aliases='', description='MEOX2 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15688378, end_position_on_the_genomic_accession=15696892, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927525, CurrentID=0, Status='live', Symbol='LOC101927525', Aliases='', description='uncharacterized LOC101927525', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94277389, end_position_on_the_genomic_accession=94397735, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927526, CurrentID=0, Status='live', Symbol='LOC101927526', Aliases='', description='uncharacterized LOC101927526', other_designations='CTA-397C4.2', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44443327, end_position_on_the_genomic_accession=44444788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874001, CurrentID=0, Status='live', Symbol='NCBP2-AS1', Aliases='', description='NCBP2 antisense RNA 1', other_designations='NCBP2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927528, CurrentID=0, Status='live', Symbol='LINC01204', Aliases='', description='long intergenic non-protein coding RNA 1204', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45505388, end_position_on_the_genomic_accession=45527239, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927529, CurrentID=0, Status='live', Symbol='LOC101927529', Aliases='', description='uncharacterized LOC101927529', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83888148, end_position_on_the_genomic_accession=83891288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104548970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104548970, CurrentID=0, Status='live', Symbol='UBE2V2P2', Aliases='', description='ubiquitin conjugating enzyme E2 variant 2 pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2 variant 2 pseudgene 2|ubiquitin-conjugating enzyme E2 variant 2 pseudogene 2', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78066265, end_position_on_the_genomic_accession=78066647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104548971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104548971, CurrentID=0, Status='live', Symbol='SALRNA1', Aliases='SAL-RNA1, XLOC_023166', description='senescence associated long non-coding RNA 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60639216, end_position_on_the_genomic_accession=60640416, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104548972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104548972, CurrentID=0, Status='live', Symbol='SALRNA3', Aliases='SAL-RNA3, XLOC_025918', description='senescence associated long non-coding RNA 3', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70615547, end_position_on_the_genomic_accession=70616567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104548973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104548973, CurrentID=0, Status='live', Symbol='SALRNA2', Aliases='SAL-RNA2, XLOC_025931', description='senescence associated long non-coding RNA 2', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70635249, end_position_on_the_genomic_accession=70637314, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927534, CurrentID=0, Status='live', Symbol='LOC101927534', Aliases='', description='uncharacterized LOC101927534', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=2870558, end_position_on_the_genomic_accession=2871231, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374141, CurrentID=0, Status='live', Symbol='LOC105374141', Aliases='', description='uncharacterized LOC105374141', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=144884609, end_position_on_the_genomic_accession=144896335, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927536, CurrentID=0, Status='live', Symbol='LOC101927536', Aliases='', description='uncharacterized LOC101927536', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56662257, end_position_on_the_genomic_accession=56666864, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927538, CurrentID=0, Status='live', Symbol='LOC101927538', Aliases='', description='uncharacterized LOC101927538', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109334199, end_position_on_the_genomic_accession=109351724, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874003, CurrentID=0, Status='live', Symbol='PLCL2-AS1', Aliases='', description='PLCL2 antisense RNA 1', other_designations='PLCL2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927540, CurrentID=0, Status='live', Symbol='TOMM20P2', Aliases='', description='TOMM20 pseudogene 2', other_designations='mitochondrial import receptor subunit TOM20 homolog|translocase of outer mitochondrial membrane 20 homolog (yeast) pseudogene 2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35514396, end_position_on_the_genomic_accession=35522379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927541, CurrentID=0, Status='live', Symbol='LINC01273', Aliases='', description='long intergenic non-protein coding RNA 1273', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50172570, end_position_on_the_genomic_accession=50176671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927542, CurrentID=0, Status='live', Symbol='LOC101927542', Aliases='', description='uncharacterized LOC101927542', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83445876, end_position_on_the_genomic_accession=83483441, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927543, CurrentID=0, Status='live', Symbol='LOC101927543', Aliases='', description='uncharacterized LOC101927543', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120761253, end_position_on_the_genomic_accession=120776832, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927546, CurrentID=0, Status='live', Symbol='ANO7P1', Aliases='ANO7L1, C1orf224, TMEM16M', description='anoctamin 7 pseudogene 1', other_designations='anoctamin 7-like 1|anoctamin-4-like|transmembrane protein 16M', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16215907, end_position_on_the_genomic_accession=16227036, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927547, CurrentID=0, Status='live', Symbol='ANKRD44-IT1', Aliases='', description='ANKRD44 intronic transcript 1', other_designations='ANKRD44 intronic transcript 1 (non-protein coding)', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197250858, end_position_on_the_genomic_accession=197302519, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927548, CurrentID=0, Status='live', Symbol='LOC101927548', Aliases='', description='uncharacterized LOC101927548', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22297805, end_position_on_the_genomic_accession=22312015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927549, CurrentID=0, Status='live', Symbol='LOC101927549', Aliases='', description='uncharacterized LOC101927549', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105673610, end_position_on_the_genomic_accession=105820333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927550, CurrentID=0, Status='live', Symbol='LOC101927550', Aliases='', description='uncharacterized LOC101927550', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99013165, end_position_on_the_genomic_accession=99036492, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874005, CurrentID=0, Status='live', Symbol='SIAH2-AS1', Aliases='', description='SIAH2 antisense RNA 1', other_designations='SIAH2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927552, CurrentID=0, Status='live', Symbol='LOC101927552', Aliases='', description='uncharacterized LOC101927552', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83817666, end_position_on_the_genomic_accession=83859339, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  101927553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927553, CurrentID=0, Status='live', Symbol='LOC101927553', Aliases='', description='uncharacterized LOC101927553', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230592660, end_position_on_the_genomic_accession=230595583, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927554, CurrentID=0, Status='live', Symbol='LINC01250', Aliases='', description='long intergenic non-protein coding RNA 1250', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=2895048, end_position_on_the_genomic_accession=3126026, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927555, CurrentID=0, Status='live', Symbol='LOC101927555', Aliases='', description='uncharacterized LOC101927555', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84612, CurrentID=0, Status='live', Symbol='PARD6B', Aliases='PAR6B', description='par-6 family cell polarity regulator beta', other_designations='PAR-6 beta|par-6 partitioning defective 6 homolog beta', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50731544, end_position_on_the_genomic_accession=50753741, orientation='plus', exon_count=3, OMIM=608975, no_hdr0=''),
  101927557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927557, CurrentID=0, Status='live', Symbol='LOC101927557', Aliases='', description='uncharacterized LOC101927557', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57522248, end_position_on_the_genomic_accession=57523597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84614, CurrentID=0, Status='live', Symbol='ZBTB37', Aliases='D430004I08Rik, ZNF908', description='zinc finger and BTB domain containing 37', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173868082, end_position_on_the_genomic_accession=173887458, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927559, CurrentID=0, Status='live', Symbol='CEBPB-AS1', Aliases='', description='CEBPB antisense RNA 1', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50184603, end_position_on_the_genomic_accession=50192069, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84616, CurrentID=0, Status='live', Symbol='KRTAP4-4', Aliases='KAP4.13, KAP4.4, KRTAP4-13, KRTAP4.13, KRTAP4.4', description='keratin associated protein 4-4', other_designations='keratin associated protein 4-13|keratin associated protein 4.4|keratin-associated protein 4-13|keratin-associated protein 4.13|keratin-associated protein 4.4|ultrahigh sulfur keratin-associated protein 4.13|ultrahigh sulfur keratin-associated protein 4.4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41159653, end_position_on_the_genomic_accession=41160731, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84617, CurrentID=0, Status='live', Symbol='TUBB6', Aliases='HsT1601, TUBB-5', description='tubulin beta 6 class V', other_designations='class V beta-tubulin|tubulin beta MGC4083|tubulin beta class V|tubulin, beta 6 class V', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12307669, end_position_on_the_genomic_accession=12329826, orientation='plus', exon_count=7, OMIM=615103, no_hdr0=''),
  84618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84618, CurrentID=0, Status='live', Symbol='NT5C1A', Aliases='CN-I, CN-IA, CN1, CN1A, CNI', description="5'-nucleotidase, cytosolic IA", other_designations="AMP-specific 5'-NT|cytosolic 5' nucleotidase, type 1A|cytosolic 5'-nucleotidase IA", map_location='1p34.3-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39659121, end_position_on_the_genomic_accession=39672038, orientation='minus', exon_count=6, OMIM=610525, no_hdr0=''),
  84619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84619, CurrentID=0, Status='live', Symbol='ZGPAT', Aliases='GPATC6, GPATCH6, KIAA1847, ZC3H9, ZC3HDC9, ZIP', description='zinc finger, CCCH-type with G-patch domain', other_designations='g patch domain-containing protein 6|zinc finger CCCH domain-containing protein 9|zinc finger CCCH-type with G patch domain|zinc finger and G patch domain-containing protein|zinc finger, CCCH-type with G patch domain', map_location='20q13.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63707442, end_position_on_the_genomic_accession=63736142, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84620, CurrentID=0, Status='live', Symbol='ST6GAL2', Aliases='SIAT2, ST6GalII', description='ST6 beta-galactosamide alpha-2,6-sialyltranferase 2', other_designations='CMP-N-acetylneuraminate-beta-galactosamide-alpha-2,6-sialyltransferase 2|alpha 2,6-ST 2|beta-galactoside alpha-2,6-sialyltransferase II|sialyltransferase 2 (monosialoganglioside sialyltransferase)', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106794128, end_position_on_the_genomic_accession=106887265, orientation='minus', exon_count=15, OMIM=608472, no_hdr0=''),
  101927565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927565, CurrentID=0, Status='live', Symbol='LINC01267', Aliases='', description='long intergenic non-protein coding RNA 1267', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14348451, end_position_on_the_genomic_accession=14352568, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84622, CurrentID=0, Status='live', Symbol='ZNF594', Aliases='', description='zinc finger protein 594', other_designations='zinc finger protein HZF18', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5179535, end_position_on_the_genomic_accession=5191883, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84623, CurrentID=0, Status='live', Symbol='KIRREL3', Aliases='KIRRE, MRD4, NEPH2, PRO4502', description='kin of IRRE like 3 (Drosophila)', other_designations='kin of irregular chiasm-like protein 3|nephrin-like 2|nephrin-like protein 2', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126423493, end_position_on_the_genomic_accession=127003460, orientation='minus', exon_count=21, OMIM=607761, no_hdr0=''),
  84624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84624, CurrentID=0, Status='live', Symbol='FNDC1', Aliases='AGS8, FNDC2, MEL4B3, bA243O10.1, dJ322A24.1', description='fibronectin type III domain containing 1', other_designations='activation-associated cDNA protein|expressed in synovial lining protein|fibronectin type III domain containing 2', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159169397, end_position_on_the_genomic_accession=159272108, orientation='plus', exon_count=23, OMIM=609991, no_hdr0=''),
  100874008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874008, CurrentID=0, Status='live', Symbol='WNT5A-AS1', Aliases='', description='WNT5A antisense RNA 1', other_designations='WNT5A antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84626, CurrentID=0, Status='live', Symbol='KRBA1', Aliases='', description='KRAB-A domain containing 1', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149714969, end_position_on_the_genomic_accession=149734575, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  84627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84627, CurrentID=0, Status='live', Symbol='ZNF469', Aliases='BCS, BCS1', description='zinc finger protein 469', other_designations='', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88233015, end_position_on_the_genomic_accession=88440757, orientation='plus', exon_count=7, OMIM=612078, no_hdr0=''),
  84628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84628, CurrentID=0, Status='live', Symbol='NTNG2', Aliases='LHLL9381, Lmnt2, NTNG1, bA479K20.1', description='netrin G2', other_designations='bA479K20.1 (novel protein)|laminet 2|laminet-2|netrin G1', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132161652, end_position_on_the_genomic_accession=132244534, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  84629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84629, CurrentID=0, Status='live', Symbol='TNRC18', Aliases='CAGL79A, TNRC18', description='trinucleotide repeat containing 18', other_designations='long CAG trinucleotide repeat-containing gene 79 protein', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5306790, end_position_on_the_genomic_accession=5423546, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  84630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84630, CurrentID=0, Status='live', Symbol='TTBK1', Aliases='BDTK', description='tau tubulin kinase 1', other_designations='brain-derived tau kinase', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43243484, end_position_on_the_genomic_accession=43288259, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84631, CurrentID=0, Status='live', Symbol='SLITRK2', Aliases='CXorf2, SLITL1', description='SLIT and NTRK like family member 2', other_designations='SLIT and NTRK-like family, member 2|slit and trk like gene 2|slit-like 1', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145817829, end_position_on_the_genomic_accession=145825842, orientation='plus', exon_count=5, OMIM=300561, no_hdr0=''),
  84632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84632, CurrentID=0, Status='live', Symbol='AFAP1L2', Aliases='CTB-1144G6.4, KIAA1914, XB130', description='actin filament associated protein 1 like 2', other_designations='AFAP1-like protein 2|CTB-1144G6.6|actin filament associated protein 1-like 2', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114280730, end_position_on_the_genomic_accession=114405143, orientation='minus', exon_count=24, OMIM=612420, no_hdr0=''),
  105374148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374148, CurrentID=0, Status='live', Symbol='LOC105374148', Aliases='', description='uncharacterized LOC105374148', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148312643, end_position_on_the_genomic_accession=148339933, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84634, CurrentID=0, Status='live', Symbol='KISS1R', Aliases='AXOR12, CPPB1, GPR54, HH8, HOT7T175, KISS-1R', description='KISS1 receptor', other_designations='G protein-coupled receptor 54|G-protein coupled receptor OT7T175|hypogonadotropin-1|kisspeptins receptor|metastin receptor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=917342, end_position_on_the_genomic_accession=921015, orientation='plus', exon_count=5, OMIM=604161, no_hdr0=''),
  101927579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927579, CurrentID=0, Status='live', Symbol='LOC101927579', Aliases='', description='dynamin-1-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30539463, end_position_on_the_genomic_accession=30625778, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  84636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84636, CurrentID=0, Status='live', Symbol='GPR174', Aliases='FKSG79, GPCR17, LYPSR3', description='G protein-coupled receptor 174', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79170972, end_position_on_the_genomic_accession=79172229, orientation='plus', exon_count=1, OMIM=300903, no_hdr0=''),
  101927581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927581, CurrentID=0, Status='live', Symbol='CCDC182', Aliases='', description='coiled-coil domain containing 182', other_designations='coiled-coil domain-containing protein ENSP00000299415', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57744481, end_position_on_the_genomic_accession=57745329, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84639, CurrentID=0, Status='live', Symbol='IL1F10', Aliases='FIL1-theta, FKSG75, IL-1HY2, IL-38, IL1-theta, IL1HY2', description='interleukin 1 family member 10 (theta)', other_designations='FIL1 theta|IL-1 theta|IL-1F10 (canonical form IL-1F10a)|family of interleukin 1-theta|interleukin 1 family, member 10 (theta)|interleukin-1 HY2|interleukin-1 receptor antagonist FKSG75|interleukin-1 receptor antagonist-like FIL1 theta|interleukin-38', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113067970, end_position_on_the_genomic_accession=113075850, orientation='plus', exon_count=5, OMIM=615296, no_hdr0=''),
  84640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84640, CurrentID=0, Status='live', Symbol='USP38', Aliases='HP43.8KD', description='ubiquitin specific peptidase 38', other_designations='deubiquitinating enzyme 38|ubiquitin specific protease 38|ubiquitin thioesterase 38|ubiquitin thiolesterase 38|ubiquitin-specific-processing protease 38', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143184917, end_position_on_the_genomic_accession=143223874, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84641, CurrentID=0, Status='live', Symbol='MFSD14B', Aliases='HIATL1', description='major facilitator superfamily domain containing 14B', other_designations='hippocampus abundant transcript-like 1', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94374322, end_position_on_the_genomic_accession=94460920, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101927586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927586, CurrentID=0, Status='live', Symbol='LINC01271', Aliases='', description='long intergenic non-protein coding RNA 1271', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50310711, end_position_on_the_genomic_accession=50321342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84643, CurrentID=0, Status='live', Symbol='KIF2B', Aliases='', description='kinesin family member 2B', other_designations='kinesin protein', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53822878, end_position_on_the_genomic_accession=53825212, orientation='plus', exon_count=1, OMIM=615142, no_hdr0=''),
  101927588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927588, CurrentID=0, Status='live', Symbol='LOC101927588', Aliases='', description='uncharacterized LOC101927588', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124192691, end_position_on_the_genomic_accession=124247398, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84645, CurrentID=0, Status='live', Symbol='C22orf23', Aliases='EVG1, dJ1039K5.6', description='chromosome 22 open reading frame 23', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37943050, end_position_on_the_genomic_accession=37953669, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927590, CurrentID=0, Status='live', Symbol='LINC01166', Aliases='', description='long intergenic non-protein coding RNA 1166', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132943967, end_position_on_the_genomic_accession=132965289, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84647, CurrentID=0, Status='live', Symbol='PLA2G12B', Aliases='FKSG71, GXIIB, GXIIIsPLA2, PLA2G13, sPLA2-GXIIB', description='phospholipase A2 group XIIB', other_designations='group XIII secreted phospholipase A2|phospholipase A2, group XIII', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72935180, end_position_on_the_genomic_accession=72954883, orientation='minus', exon_count=4, OMIM=611653, no_hdr0=''),
  84648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84648, CurrentID=0, Status='live', Symbol='LCE3D', Aliases='LEP16, SPRL6A, SPRL6B', description='late cornified envelope 3D', other_designations='late envelope protein 16|small proline rich-like (epidermal differentiation complex) 6A|small proline rich-like (epidermal differentiation complex) 6B|small proline-rich-like epidermal differentiation complex protein 6A|small proline-rich-like epidermal differentiation complex protein 6B', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152579384, end_position_on_the_genomic_accession=152580504, orientation='minus', exon_count=2, OMIM=612616, no_hdr0=''),
  84649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84649, CurrentID=0, Status='live', Symbol='DGAT2', Aliases='ARAT, GS1999FULL, HMFN1045', description='diacylglycerol O-acyltransferase 2', other_designations='acyl-CoA retinol O-fatty-acyltransferase|diacylglycerol O-acyltransferase homolog 2|diacylglycerol O-acyltransferase-like protein 2|diglyceride acyltransferase 2|retinol O-fatty-acyltransferase', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75768733, end_position_on_the_genomic_accession=75801536, orientation='plus', exon_count=9, OMIM=606983, no_hdr0=''),
  84650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84650, CurrentID=0, Status='live', Symbol='EBPL', Aliases='EBRP', description='emopamil binding protein-like', other_designations='emopamil binding related protein, delta8-delta7 sterol isomerase related protein|emopamil-binding-related protein', map_location='13q12-q13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49660676, end_position_on_the_genomic_accession=49691487, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84651, CurrentID=0, Status='live', Symbol='SPINK7', Aliases='ECG2, ECRG2', description='serine peptidase inhibitor, Kazal type 7 (putative)', other_designations='ECRG-2|esophagus cancer related gene 2|esophagus cancer-related gene 2 protein|esophagus cancer-related gene-2', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148312427, end_position_on_the_genomic_accession=148315919, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927596, CurrentID=0, Status='live', Symbol='LOC101927596', Aliases='', description='uncharacterized LOC101927596', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197197934, end_position_on_the_genomic_accession=197199273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927597, CurrentID=0, Status='live', Symbol='NCK1-AS1', Aliases='SLC35G2-AS1', description='NCK1 antisense RNA 1 (head to head)', other_designations='SLC35G2 antisense RNA 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136841976, end_position_on_the_genomic_accession=136862063, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84654, CurrentID=0, Status='live', Symbol='SPZ1', Aliases='NYD-TSP1, PPP1R148', description='spermatogenic leucine zipper 1', other_designations='protein phosphatase 1, regulatory subunit 148|spermatogenic zip 1|testis-specific protein 1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80319971, end_position_on_the_genomic_accession=80321842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927599, CurrentID=0, Status='live', Symbol='ZNF529-AS1', Aliases='', description='ZNF529 antisense RNA 1', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36573070, end_position_on_the_genomic_accession=36594716, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84656, CurrentID=0, Status='live', Symbol='GLYR1', Aliases='BM045, HIBDL, N-PAC, NP60', description='glyoxylate reductase 1 homolog (Arabidopsis)', other_designations='3-hydroxyisobutyrate dehydrogenase-like protein|cytokine-like nuclear factor n-pac|nuclear protein 60 kDa|nuclear protein 60kDa|nuclear protein NP60|nuclear protein of 60 kDa', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4803203, end_position_on_the_genomic_accession=4847333, orientation='minus', exon_count=19, OMIM=610660, no_hdr0=''),
  101927601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927601, CurrentID=0, Status='live', Symbol='GOLGA6L19', Aliases='', description='golgin A6 family-like 19', other_designations='Putative golgin subfamily A member 6-like protein 19', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84658, CurrentID=0, Status='live', Symbol='ADGRE3', Aliases='EMR3', description='adhesion G protein-coupled receptor E3', other_designations='EGF-like module receptor 3|EGF-like module-containing mucin-like hormone receptor-like 3|egf-like module containing, mucin-like, hormone receptor-like 3|egf-like module-containing mucin-like receptor 3', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14619117, end_position_on_the_genomic_accession=14674918, orientation='minus', exon_count=16, OMIM=606101, no_hdr0=''),
  84659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84659, CurrentID=0, Status='live', Symbol='RNASE7', Aliases='', description='ribonuclease, RNase A family, 7', other_designations='RNase 7|SAP-2|skin-derived antimicrobial protein 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21042226, end_position_on_the_genomic_accession=21044233, orientation='plus', exon_count=2, OMIM=612484, no_hdr0=''),
  84660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84660, CurrentID=0, Status='live', Symbol='CCDC62', Aliases='CT109, ERAP75, TSP-NY', description='coiled-coil domain containing 62', other_designations='aaa-protein|cancer/testis antigen 109|coiled-coil domain-containing protein 60|testis-specific protein TSP-NY', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122774509, end_position_on_the_genomic_accession=122827385, orientation='plus', exon_count=14, OMIM=613481, no_hdr0=''),
  84661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84661, CurrentID=0, Status='live', Symbol='DPY30', Aliases='Cps25, HDPY-30, Saf19', description='dpy-30, histone methyltransferase complex regulatory subunit', other_designations='dpy-30 histone methyltransferase complex regulatory subunit|dpy-30 homolog|dpy-30-like protein|dpy-30L', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32021080, end_position_on_the_genomic_accession=32039840, orientation='minus', exon_count=8, OMIM=612032, no_hdr0=''),
  84662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84662, CurrentID=0, Status='live', Symbol='GLIS2', Aliases='NKL, NPHP7', description='GLIS family zinc finger 2', other_designations='GLI-similar 2|GLI-similar protein 2|Kruppel-like zinc finger protein GLIS2|nephrocystin-7|neuronal Krueppel-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4316181, end_position_on_the_genomic_accession=4339597, orientation='plus', exon_count=8, OMIM=608539, no_hdr0=''),
  84664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84664, CurrentID=0, Status='live', Symbol='CSPG4P2Y', Aliases='CSPG4LYP2, CSPG4P2, CSPG4PY2', description='chondroitin sulfate proteoglycan 4 pseudogene 2, Y-linked', other_designations='chondroitin sulfate proteoglycan 4-like, Y-linked pseudogene 2|chondroitin sulfate proteoglycan 4-like, Y-linked, centromeric', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24183433, end_position_on_the_genomic_accession=24187230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  84665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84665, CurrentID=0, Status='live', Symbol='MYPN', Aliases='CMD1DD, CMH22, MYOP, RCM4', description='myopalladin', other_designations='sarcomeric protein myopalladin, 145 kDa (MYOP)', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68106117, end_position_on_the_genomic_accession=68212017, orientation='plus', exon_count=26, OMIM=608517, no_hdr0=''),
  84666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84666, CurrentID=0, Status='live', Symbol='RETNLB', Aliases='FIZZ1, FIZZ2, HXCP2, RELM-beta, RELMb, RELMbeta, XCP2', description='resistin like beta', other_designations='C/EBP-epsilon regulated myeloid-specific secreted cysteine-rich protein precursor 2|colon and small intestine-specific cysteine-rich protein|colon carcinoma-related gene protein|cysteine-rich secreted A12-alpha-like protein 1|cysteine-rich secreted protein A12-alpha-like 1|cysteine-rich secreted protein FIZZ2|found in inflammatory zone 1', map_location='3q13.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108755639, end_position_on_the_genomic_accession=108757283, orientation='minus', exon_count=3, OMIM=605645, no_hdr0=''),
  84667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84667, CurrentID=0, Status='live', Symbol='HES7', Aliases='SCDO4, bHLHb37', description='hes family bHLH transcription factor 7', other_designations='bHLH factor Hes7|class B basic helix-loop-helix protein 37|hHes7|hairy and enhancer of split 7', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8120590, end_position_on_the_genomic_accession=8125739, orientation='minus', exon_count=7, OMIM=608059, no_hdr0=''),
  84668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84668, CurrentID=0, Status='live', Symbol='FAM126A', Aliases='DRCTNNB1A, HCC, HLD5, HYCC1', description='family with sequence similarity 126 member A', other_designations='down regulated by Ctnnb1, a|down-regulated by CTNNB1 protein A|family with sequence similarity 126, member A', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22895848, end_position_on_the_genomic_accession=23014195, orientation='minus', exon_count=14, OMIM=610531, no_hdr0=''),
  84669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84669, CurrentID=0, Status='live', Symbol='USP32', Aliases='NY-REN-60, USP10', description='ubiquitin specific peptidase 32', other_designations='deubiquitinating enzyme 32|renal carcinoma antigen NY-REN-60|ubiquitin specific protease 32|ubiquitin thioesterase 32|ubiquitin thiolesterase 32|ubiquitin-specific-processing protease 32', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60177231, end_position_on_the_genomic_accession=60422470, orientation='minus', exon_count=38, OMIM=607740, no_hdr0=''),
  84671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84671, CurrentID=0, Status='live', Symbol='ZNF347', Aliases='ZNF1111', description='zinc finger protein 347', other_designations='zinc finger 1111|zinc finger protein 1111', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53138300, end_position_on_the_genomic_accession=53163938, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84672, CurrentID=0, Status='live', Symbol='TTTY6', Aliases='LINC00127, NCRNA00127A, TTY6, TTTY6', description='testis-specific transcript, Y-linked 6 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 127|testis transcript Y 6', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22439593, end_position_on_the_genomic_accession=22441459, orientation='minus', exon_count=3, OMIM=400039, no_hdr0=''),
  100874016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874016, CurrentID=0, Status='live', Symbol='SYNPR-AS1', Aliases='', description='SYNPR antisense RNA 1', other_designations='SYNPR antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63423596, end_position_on_the_genomic_accession=63550051, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84674, CurrentID=0, Status='live', Symbol='CARD6', Aliases='CINCIN1', description='caspase recruitment domain family member 6', other_designations='CARD-containing inhibitor of Nod1 and Cardiak-induced NF-kB activation|caspase recruitment domain family, member 6|caspase recruitment domain protein 6', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40841308, end_position_on_the_genomic_accession=40855354, orientation='plus', exon_count=3, OMIM=609986, no_hdr0=''),
  84675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84675, CurrentID=0, Status='live', Symbol='TRIM55', Aliases='MURF-2, RNF29, muRF2', description='tripartite motif containing 55', other_designations='muscle specific ring finger 2|muscle-specific RING finger protein 2|ring finger protein 29|tripartite motif-containing 55', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66112667, end_position_on_the_genomic_accession=66175489, orientation='plus', exon_count=15, OMIM=606469, no_hdr0=''),
  84676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84676, CurrentID=0, Status='live', Symbol='TRIM63', Aliases='IRF, MURF1, MURF2, RNF28, SMRZ', description='tripartite motif containing 63, E3 ubiquitin protein ligase', other_designations='iris ring finger protein|muscle specific ring finger protein 2|muscle-specific RING finger protein 1|ring finger protein 28|striated muscle RING zinc finger protein|tripartite motif-containing protein 63', map_location='1p34-p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26051304, end_position_on_the_genomic_accession=26067634, orientation='minus', exon_count=9, OMIM=606131, no_hdr0=''),
  84677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84677, CurrentID=0, Status='live', Symbol='DSCR8', Aliases='C21orf65, CT25.1a, CT25.1b, MMA-1, MMA-1a, MMA-1b, MMA1, MTAG2', description='Down syndrome critical region 8', other_designations='Down syndrome critical region gene 8|cancer/testis antigen family 25, member 1a|cancer/testis antigen family 25, member 1b|malignant melanoma associated protein 1|malignant melanoma-associated 1|melanoma-testis-associated protein 2', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38121451, end_position_on_the_genomic_accession=38156511, orientation='plus', exon_count=4, OMIM=613396, no_hdr0=''),
  84678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84678, CurrentID=0, Status='live', Symbol='KDM2B', Aliases='CXXC2, FBXL10, Fbl10, JHDM1B, PCCX2', description='lysine (K)-specific demethylase 2B', other_designations='CXXC-type zinc finger protein 2|F-box and leucine-rich repeat protein 10|F-box protein FBL10|F-box/LRR-repeat protein 10|JEMMA (Jumonji domain, EMSY-interactor, methyltransferase motif) protein|[Histone-H3]-lysine-36 demethylase 1B|jmjC domain-containing histone demethylation protein 1B|jumonji C domain-containing histone demethylase 1B|protein-containing CXXC domain 2', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121402501, end_position_on_the_genomic_accession=121581113, orientation='minus', exon_count=34, OMIM=609078, no_hdr0=''),
  84679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84679, CurrentID=0, Status='live', Symbol='SLC9A7', Aliases='NHE-7, NHE7, SLC9A6', description='solute carrier family 9, subfamily A (NHE7, cation proton antiporter 7), member 7', other_designations='Na(+)/H(+) exchanger 7|nonselective sodium potassium/proton exchanger|solute carrier family 9 (sodium/hydrogen exchanger)', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46599246, end_position_on_the_genomic_accession=46759172, orientation='minus', exon_count=20, OMIM=300368, no_hdr0=''),
  84680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84680, CurrentID=0, Status='live', Symbol='ACCS', Aliases='ACS, PHACS', description='1-aminocyclopropane-1-carboxylate synthase homolog (inactive)', other_designations='1-aminocyclopropane-1-carboxylate synthase-like protein (non-functional) isoform 1|1-aminocyclopropane-1-carboxylate synthase-like protein (non-functional) isoform 2|1-aminocyclopropane-1-carboxylate synthase-like protein (non-functional) isoform 4|ACC synthase-like protein 1', map_location='11p11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44066179, end_position_on_the_genomic_accession=44084237, orientation='plus', exon_count=18, OMIM=608405, no_hdr0=''),
  84681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84681, CurrentID=0, Status='live', Symbol='HINT2', Aliases='HIT-17', description='histidine triad nucleotide binding protein 2', other_designations='HINT-2|HINT-3|HIT-17kDa|PKCI-1-related HIT protein|protein kinase C inhibitor-2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35812960, end_position_on_the_genomic_accession=35815479, orientation='minus', exon_count=6, OMIM=609997, no_hdr0=''),
  101927627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927627, CurrentID=0, Status='live', Symbol='LOC101927627', Aliases='', description='uncharacterized LOC101927627', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109590320, end_position_on_the_genomic_accession=109598832, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84684, CurrentID=0, Status='live', Symbol='INSM2', Aliases='IA-6, IA6, mlt1', description='insulinoma-associated 2', other_designations='INSM2 variant 1|insulinoma-associated protein IA-6|zinc finger protein IA-6', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35534042, end_position_on_the_genomic_accession=35537054, orientation='plus', exon_count=1, OMIM=614027, no_hdr0=''),
  100874018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874018, CurrentID=0, Status='live', Symbol='EPN2-AS1', Aliases='', description='EPN2 antisense RNA 1', other_designations='EPN2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19296596, end_position_on_the_genomic_accession=19306261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927630, CurrentID=0, Status='live', Symbol='LOC101927630', Aliases='', description='uncharacterized LOC101927630', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17463445, end_position_on_the_genomic_accession=17558909, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84687, CurrentID=0, Status='live', Symbol='PPP1R9B', Aliases='PPP1R6, PPP1R9, SPINO, Spn', description='protein phosphatase 1 regulatory subunit 9B', other_designations='neurabin-II|protein phosphatase 1, regulatory (inhibitor) subunit 9B|protein phosphatase 1, regulatory subunit 9B|protein phosphatase 1, regulatory subunit 9B, spinophilin', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50133735, end_position_on_the_genomic_accession=50151022, orientation='minus', exon_count=11, OMIM=603325, no_hdr0=''),
  84688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84688, CurrentID=0, Status='live', Symbol='C9orf24', Aliases='CBE1, NYD-SP22, SMRP1, bA573M23.4', description='chromosome 9 open reading frame 24', other_designations='ciliated bronchial epithelial protein 1|ciliated bronchial epithelium 1|testis development protein NYD-SP22', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34379019, end_position_on_the_genomic_accession=34398577, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84689, CurrentID=0, Status='live', Symbol='MS4A14', Aliases='MS4A16, NYD-SP21', description='membrane-spanning 4-domains, subfamily A, member 14', other_designations='MS4A13 protein|membrane-spanning 4-domains, subfamily A, member 16|testes development-related NYD-SP21|testis development protein NYD-SP21', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60396014, end_position_on_the_genomic_accession=60417756, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84690, CurrentID=0, Status='live', Symbol='SPATA22', Aliases='NYD-SP20, NYDSP20', description='spermatogenesis associated 22', other_designations='testes development-related NYD-SP20|testis development protein NYD-SP20', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3440011, end_position_on_the_genomic_accession=3513852, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  84691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84691, CurrentID=0, Status='live', Symbol='FAM71F1', Aliases='FAM137A, NYD-SP18', description='family with sequence similarity 71 member F1', other_designations='family with sequence similarity 137, member A|family with sequence similarity 71, member F1|testes development-related NYD-SP18|testis development protein NYD-SP18', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128709061, end_position_on_the_genomic_accession=128731743, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84692, CurrentID=0, Status='live', Symbol='CCDC54', Aliases='NYD-SP17, SP17', description='coiled-coil domain containing 54', other_designations='sperm protein 17|testes development-related NYD-SP17|testis development protein NYD-SP17', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107377341, end_position_on_the_genomic_accession=107378634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84693, CurrentID=0, Status='live', Symbol='MCEE', Aliases='GLOD2', description='methylmalonyl-CoA epimerase', other_designations='DL-methylmalonyl-CoA racemase|glyoxalase domain containing 2|methylmalonyl CoA epimerase', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71109676, end_position_on_the_genomic_accession=71130288, orientation='minus', exon_count=4, OMIM=608419, no_hdr0=''),
  84694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84694, CurrentID=0, Status='live', Symbol='GJA10', Aliases='CX62', description='gap junction protein alpha 10', other_designations='connexin 62|connexin-62|gap junction protein, alpha 10, 62kDa', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89894469, end_position_on_the_genomic_accession=89896100, orientation='plus', exon_count=1, OMIM=611924, no_hdr0=''),
  84695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84695, CurrentID=0, Status='live', Symbol='LOXL3', Aliases='LOXL', description='lysyl oxidase like 3', other_designations='lysyl oxidase-like 3|lysyl oxidase-like protein 3', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74532258, end_position_on_the_genomic_accession=74555887, orientation='minus', exon_count=15, OMIM=607163, no_hdr0=''),
  84696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84696, CurrentID=0, Status='live', Symbol='ABHD1', Aliases='LABH1', description='abhydrolase domain containing 1', other_designations='lung alpha/beta hydrolase 1', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27123780, end_position_on_the_genomic_accession=27130991, orientation='plus', exon_count=9, OMIM=612195, no_hdr0=''),
  100874020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874020, CurrentID=0, Status='live', Symbol='MAGI2-AS1', Aliases='', description='MAGI2 antisense RNA 1', other_designations='MAGI2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84698, CurrentID=0, Status='live', Symbol='CAPS2', Aliases='UG0636c06', description='calcyphosine 2', other_designations='calcyphosphine 2', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75257036, end_position_on_the_genomic_accession=75390928, orientation='minus', exon_count=25, OMIM=607724, no_hdr0=''),
  84699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84699, CurrentID=0, Status='live', Symbol='CREB3L3', Aliases='CREB-H, CREBH, HYST1481', description='cAMP responsive element binding protein 3-like 3', other_designations='CREB/ATF family transcription factor', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4153601, end_position_on_the_genomic_accession=4173054, orientation='plus', exon_count=10, OMIM=611998, no_hdr0=''),
  84700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84700, CurrentID=0, Status='live', Symbol='MYO18B', Aliases='', description='myosin XVIIIB', other_designations='myosin 18B|myosin-XVIIIb', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25742147, end_position_on_the_genomic_accession=26058788, orientation='plus', exon_count=52, OMIM=607295, no_hdr0=''),
  84701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84701, CurrentID=0, Status='live', Symbol='COX4I2', Aliases='COX4, COX4-2, COX4B, COX4L2, COXIV-2, dJ857M17.2', description='cytochrome c oxidase subunit IV isoform 2 (lung)', other_designations='COX IV-2|cytochrome c oxidase subunit IV-like 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31637879, end_position_on_the_genomic_accession=31645009, orientation='plus', exon_count=7, OMIM=607976, no_hdr0=''),
  101927646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927646, CurrentID=0, Status='live', Symbol='LOC101927646', Aliases='', description='uncharacterized LOC101927646', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110036195, end_position_on_the_genomic_accession=110056149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927647, CurrentID=0, Status='live', Symbol='LOC101927647', Aliases='', description='uncharacterized LOC101927647', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15739388, end_position_on_the_genomic_accession=15766274, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927648, CurrentID=0, Status='live', Symbol='LOC101927648', Aliases='', description='uncharacterized LOC101927648', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84705, CurrentID=0, Status='live', Symbol='GTPBP3', Aliases='COXPD23, GTPBG3, MSS1, MTGP1, THDF1', description='GTP binding protein 3 (mitochondrial)', other_designations='mitochondrial GTP-binding protein 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17334982, end_position_on_the_genomic_accession=17342731, orientation='plus', exon_count=10, OMIM=608536, no_hdr0=''),
  84706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84706, CurrentID=0, Status='live', Symbol='GPT2', Aliases='ALT2, GPT 2, MRT49', description='glutamic pyruvate transaminase (alanine aminotransferase) 2', other_designations='glutamate pyruvate transaminase 2|glutamic--alanine transaminase 2|glutamic--pyruvic transaminase 2', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46884396, end_position_on_the_genomic_accession=46931289, orientation='plus', exon_count=14, OMIM=138210, no_hdr0=''),
  84707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84707, CurrentID=0, Status='live', Symbol='BEX2', Aliases='BEX1, DJ79P11.1', description='brain expressed X-linked 2', other_designations='brain-expressed X-linked protein 2|hBex2', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103309346, end_position_on_the_genomic_accession=103311046, orientation='minus', exon_count=3, OMIM=300691, no_hdr0=''),
  84708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84708, CurrentID=0, Status='live', Symbol='LNX1', Aliases='LNX, MPDZ, PDZRN2', description='ligand of numb-protein X 1, E3 ubiquitin protein ligase', other_designations='PDZ domain-containing ring finger protein 2|multi-PDZ-domain-containing protein, E3 ubiquitin-protein ligase LNX|numb-binding protein 1', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53460270, end_position_on_the_genomic_accession=53652514, orientation='minus', exon_count=14, OMIM=609732, no_hdr0=''),
  84709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84709, CurrentID=0, Status='live', Symbol='MGARP', Aliases='C4orf49, CESP-1, HUMMR, OSAP', description='mitochondria localized glutamic acid rich protein', other_designations='corneal endothelium-specific protein 1|hypoxia up-regulated mitochondrial movement regulator protein|mitochondria-localized glutamic acid-rich protein|ovary-specific acidic protein', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139266163, end_position_on_the_genomic_accession=139280338, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927654, CurrentID=0, Status='live', Symbol='LOC101927654', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113482040, end_position_on_the_genomic_accession=113483302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927655, CurrentID=0, Status='live', Symbol='ZASP', Aliases='', description='ZO-2 associated speckle protein', other_designations='zinc finger protein 469-like', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100574035, end_position_on_the_genomic_accession=100574733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927657, CurrentID=0, Status='live', Symbol='LOC101927657', Aliases='', description='uncharacterized LOC101927657', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126325495, end_position_on_the_genomic_accession=126329535, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927659, CurrentID=0, Status='live', Symbol='SMAD1-AS2', Aliases='', description='SMAD1 antisense RNA 2', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145497049, end_position_on_the_genomic_accession=145502980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927660, CurrentID=0, Status='live', Symbol='LOC101927660', Aliases='', description='uncharacterized LOC101927660', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232174935, end_position_on_the_genomic_accession=232180113, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84717, CurrentID=0, Status='live', Symbol='HDGFRP2', Aliases='HDGF-2, HDGF2, HRP-2', description='hepatoma-derived growth factor-related protein 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4472196, end_position_on_the_genomic_accession=4502220, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  84719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84719, CurrentID=0, Status='live', Symbol='LINC00260', Aliases='C1orf217, NCRNA00260', description='long intergenic non-protein coding RNA 260', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203730577, end_position_on_the_genomic_accession=203731853, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84720, CurrentID=0, Status='live', Symbol='PIGO', Aliases='HPMRS2', description='phosphatidylinositol glycan anchor biosynthesis class O', other_designations='phosphatidylinositol glycan anchor biosynthesis, class O|phosphatidylinositol-glycan biosynthesis class O protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35088688, end_position_on_the_genomic_accession=35096601, orientation='minus', exon_count=14, OMIM=614730, no_hdr0=''),
  100874024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874024, CurrentID=0, Status='live', Symbol='TRPC7-AS1', Aliases='', description='TRPC7 antisense RNA 1', other_designations='TRPC7 antisense RNA 1 (non-protein coding)', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84722, CurrentID=0, Status='live', Symbol='PSRC1', Aliases='DDA3, FP3214', description='proline/serine-rich coiled-coil 1', other_designations='differential display and activated by p53|p53-regulated DDA3', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109279554, end_position_on_the_genomic_accession=109283168, orientation='minus', exon_count=8, OMIM=613126, no_hdr0=''),
  101927667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927667, CurrentID=0, Status='live', Symbol='LINC01535', Aliases='', description='long intergenic non-protein coding RNA 1535', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37251885, end_position_on_the_genomic_accession=37265535, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101927668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927668, CurrentID=0, Status='live', Symbol='LOC101927668', Aliases='', description='uncharacterized LOC101927668', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19918981, end_position_on_the_genomic_accession=20140426, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84725, CurrentID=0, Status='live', Symbol='PLEKHA8', Aliases='FAPP2', description='pleckstrin homology domain containing A8', other_designations='PH domain-containing family A member 8|phosphatidylinositol-four-phosphate adapter protein 2|phosphoinositol 4-phosphate adapter protein 2|pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 8|serologically defined breast cancer antigen NY-BR-86', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30028361, end_position_on_the_genomic_accession=30130483, orientation='plus', exon_count=19, OMIM=608639, no_hdr0=''),
  84726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84726, CurrentID=0, Status='live', Symbol='PRRC2B', Aliases='BAT2L, BAT2L1, KIAA0515, LQFBS-1', description='proline-rich coiled-coil 2B', other_designations='HLA-B associated transcript 2-like 1|HLA-B-associated transcript 2-like 1|proline-rich coiled-coil protein 2B|protein BAT2-like 1', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131430090, end_position_on_the_genomic_accession=131500191, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  84727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84727, CurrentID=0, Status='live', Symbol='SPSB2', Aliases='GRCC9, SSB2', description='splA/ryanodine receptor domain and SOCS box containing 2', other_designations='SPRY domain-containing SOCS box protein SSB-2|gene-rich cluster protein C9', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6870935, end_position_on_the_genomic_accession=6873471, orientation='minus', exon_count=4, OMIM=611658, no_hdr0=''),
  101927673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927673, CurrentID=0, Status='live', Symbol='LOC101927673', Aliases='', description='uncharacterized LOC101927673', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64000351, end_position_on_the_genomic_accession=64008379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927676, CurrentID=0, Status='live', Symbol='LOC101927676', Aliases='', description='uncharacterized LOC101927676', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66127806, end_position_on_the_genomic_accession=66134038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84733, CurrentID=0, Status='live', Symbol='CBX2', Aliases='CDCA6, M33, SRXY5', description='chromobox 2', other_designations='Pc class homolog|cell division cycle associated 6|chromobox homolog 2 (Pc class homolog, Drosophila)|modifier 3', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79777188, end_position_on_the_genomic_accession=79787650, orientation='plus', exon_count=7, OMIM=602770, no_hdr0=''),
  84734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84734, CurrentID=0, Status='live', Symbol='FAM167B', Aliases='C1orf90', description='family with sequence similarity 167 member B', other_designations='family with sequence similarity 167, member B', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32247217, end_position_on_the_genomic_accession=32248860, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84735, CurrentID=0, Status='live', Symbol='CNDP1', Aliases='CN1, CPGL2, HsT2308', description='carnosine dipeptidase 1 (metallopeptidase M20 family)', other_designations='CNDP dipeptidase 1|carnosinase 1|glutamate carboxypeptidase-like protein 2|serum carnosinase', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74534457, end_position_on_the_genomic_accession=74585025, orientation='plus', exon_count=12, OMIM=609064, no_hdr0=''),
  101927680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927680, CurrentID=0, Status='live', Symbol='LOC101927680', Aliases='', description='uncharacterized LOC101927680', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88031753, end_position_on_the_genomic_accession=88066151, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927681, CurrentID=0, Status='live', Symbol='LOC101927681', Aliases='', description='5E5 antigen-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145480860, end_position_on_the_genomic_accession=145482697, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927682, CurrentID=0, Status='live', Symbol='CD81-AS1', Aliases='', description='CD81 antisense RNA 1', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2328749, end_position_on_the_genomic_accession=2377992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874027, CurrentID=0, Status='live', Symbol='TPRG1-AS2', Aliases='', description='TPRG1 antisense RNA 2', other_designations='TPRG1 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189238686, end_position_on_the_genomic_accession=189240594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84740, CurrentID=0, Status='live', Symbol='AFAP1-AS1', Aliases='AFAP1-AS, AFAP1AS', description='AFAP1 antisense RNA 1', other_designations='AFAP1 antisense RNA 1 (non-protein coding)', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7754090, end_position_on_the_genomic_accession=7778928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927685, CurrentID=0, Status='live', Symbol='LOC101927685', Aliases='', description='heat shock transcription factor, X-linked-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149929624, end_position_on_the_genomic_accession=149931821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927686, CurrentID=0, Status='live', Symbol='LOC101927686', Aliases='', description='uncharacterized LOC101927686', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113623535, end_position_on_the_genomic_accession=113650075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873574, CurrentID=0, Status='live', Symbol='RNA5SP299', Aliases='RN5S299', description='RNA, 5S ribosomal pseudogene 299', other_designations='RNA, 5S ribosomal 299', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8656716, end_position_on_the_genomic_accession=8656831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927688, CurrentID=0, Status='live', Symbol='SEPT4-AS1', Aliases='', description='SEPT4 antisense RNA 1', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58519837, end_position_on_the_genomic_accession=58556925, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874028, CurrentID=0, Status='live', Symbol='SGOL1-AS1', Aliases='', description='SGOL1 antisense RNA 1', other_designations='SGOL1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20174244, end_position_on_the_genomic_accession=20186427, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927690, CurrentID=0, Status='live', Symbol='LOC101927690', Aliases='', description='uncharacterized LOC101927690', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56514351, end_position_on_the_genomic_accession=56551309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84747, CurrentID=0, Status='live', Symbol='UNC119B', Aliases='POC7B', description='unc-119 lipid binding chaperone B', other_designations='POC7 centriolar protein homolog B|unc-119 homolog B', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120710435, end_position_on_the_genomic_accession=120723640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927692, CurrentID=0, Status='live', Symbol='LOC101927692', Aliases='', description='uncharacterized LOC101927692', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114764788, end_position_on_the_genomic_accession=114779903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84749, CurrentID=0, Status='live', Symbol='USP30', Aliases='', description='ubiquitin specific peptidase 30', other_designations='deubiquitinating enzyme 30|ub-specific protease 30|ubiquitin specific protease 30|ubiquitin thioesterase 30|ubiquitin thiolesterase 30|ubiquitin-specific protease 30|ubiquitin-specific-processing protease 30', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109023089, end_position_on_the_genomic_accession=109088026, orientation='plus', exon_count=19, OMIM=612492, no_hdr0=''),
  84750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84750, CurrentID=0, Status='live', Symbol='FUT10', Aliases='', description='fucosyltransferase 10 (alpha (1,3) fucosyltransferase)', other_designations='alpha 1,3-fucosyl transferase|fuc-TX|fucT-X|fucosyltransferase X|galactoside 3-L-fucosyltransferase 10', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33308611, end_position_on_the_genomic_accession=33473478, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100874029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874029, CurrentID=0, Status='live', Symbol='CFAP44-AS1', Aliases='WDR52-AS1', description='CFAP44 antisense RNA 1', other_designations='WDR52 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113403998, end_position_on_the_genomic_accession=113433992, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84752, CurrentID=0, Status='live', Symbol='B3GNT9', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 9', other_designations='BGnT-9|beta-1,3-Gn-T9|beta-1,3-N-acetylglucosaminyltransferase 9|beta-1,3-galactosyltransferase-related protein|beta3Gn-T9', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67148102, end_position_on_the_genomic_accession=67150999, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927697, CurrentID=0, Status='live', Symbol='LOC101927697', Aliases='', description='uncharacterized LOC101927697', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158320683, end_position_on_the_genomic_accession=158409773, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927698, CurrentID=0, Status='live', Symbol='DSCAS', Aliases='', description='DSC1/DSC2 antisense RNA', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31101588, end_position_on_the_genomic_accession=31162789, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927699, CurrentID=0, Status='live', Symbol='LOC101927699', Aliases='', description='uncharacterized LOC101927699', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46582163, end_position_on_the_genomic_accession=46598145, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927700, CurrentID=0, Status='live', Symbol='LINC01524', Aliases='', description='long intergenic non-protein coding RNA 1524', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52620838, end_position_on_the_genomic_accession=52650426, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84757, CurrentID=0, Status='live', Symbol='MGC10814', Aliases='', description='uncharacterized protein MGC10814', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927702, CurrentID=0, Status='live', Symbol='LOC101927702', Aliases='', description='uncharacterized LOC101927702', other_designations='CTD-2184C24.2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60240415, end_position_on_the_genomic_accession=60245670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927703, CurrentID=0, Status='live', Symbol='FOXCUT', Aliases='LINC01379, TCONS_00011636', description='FOXC1 upstream transcript (non-protein coding)', other_designations='FOXC1 promoter upstream transcript|long intergenic non-protein coding RNA 1379', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1605531, end_position_on_the_genomic_accession=1607358, orientation='plus', exon_count=2, OMIM=615976, no_hdr0=''),
  101927704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927704, CurrentID=0, Status='live', Symbol='MIR3663HG', Aliases='', description='MIR3663 host gene', other_designations='MIR3663 host gene (non-protein coding)', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117153521, end_position_on_the_genomic_accession=117169055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874031, CurrentID=0, Status='live', Symbol='AGBL5-AS1', Aliases='', description='AGBL5 antisense RNA 1', other_designations='AGBL5 antisense RNA 1 (non-protein coding)', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27049683, end_position_on_the_genomic_accession=27050264, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927708, CurrentID=0, Status='live', Symbol='LOC101927708', Aliases='', description='uncharacterized LOC101927708', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3507980, end_position_on_the_genomic_accession=3571987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84765, CurrentID=0, Status='live', Symbol='ZNF577', Aliases='', description='zinc finger protein 577', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51855803, end_position_on_the_genomic_accession=51890916, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  84766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84766, CurrentID=0, Status='live', Symbol='CRACR2A', Aliases='EFCAB4B', description='calcium release activated channel regulator 2A', other_designations='CRAC channel regulator 2A|CRAC regulator 2A|Ca2+ release-activated Ca2+ (CRAC) channel regulator 2A|EF-hand calcium binding domain 4B|calcium release-activated calcium channel regulator 2A|calcium release-activated channel regulator 2A', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3612930, end_position_on_the_genomic_accession=3753200, orientation='minus', exon_count=23, OMIM=614178, no_hdr0=''),
  84767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84767, CurrentID=0, Status='live', Symbol='TRIM51', Aliases='SPRYD5A, TRIM51', description='tripartite motif-containing 51', other_designations='SPRY domain containing 5|SPRY domain-containing protein 5', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55883297, end_position_on_the_genomic_accession=55891810, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927712, CurrentID=0, Status='live', Symbol='LOC101927712', Aliases='', description='uncharacterized LOC101927712', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110131507, end_position_on_the_genomic_accession=110140218, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84769, CurrentID=0, Status='live', Symbol='MPV17L2', Aliases='FKSG24', description='MPV17 mitochondrial membrane protein-like 2', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18193230, end_position_on_the_genomic_accession=18196742, orientation='plus', exon_count=5, OMIM=616133, no_hdr0=''),
  254427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254427, CurrentID=0, Status='live', Symbol='PROSER2', Aliases='C10orf47', description='proline and serine rich 2', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11823349, end_position_on_the_genomic_accession=11872277, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927717, CurrentID=0, Status='live', Symbol='GTF2IP8', Aliases='', description='general transcription factor IIi pseudogene 8', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14642719, end_position_on_the_genomic_accession=14706605, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927718, CurrentID=0, Status='live', Symbol='DSG1-AS1', Aliases='', description='DSG1 antisense RNA 1', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31343368, end_position_on_the_genomic_accession=31426988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84775, CurrentID=0, Status='live', Symbol='ZNF607', Aliases='', description='zinc finger protein 607', other_designations='', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37696363, end_position_on_the_genomic_accession=37719790, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927720, CurrentID=0, Status='live', Symbol='ZNF793-AS1', Aliases='', description='ZNF793 antisense RNA 1 (head to head)', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37497159, end_position_on_the_genomic_accession=37507087, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84777, CurrentID=0, Status='live', Symbol='DLGAP1-AS2', Aliases='', description='DLGAP1 antisense RNA 2', other_designations='DLGAP1 antisense RNA 2 (non-protein coding)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3603737, end_position_on_the_genomic_accession=3610089, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927722, CurrentID=0, Status='live', Symbol='LL22NC03-75H12.2', Aliases='', description='uncharacterized LOC101927722', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47461299, end_position_on_the_genomic_accession=47487111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84779, CurrentID=0, Status='live', Symbol='NAA11', Aliases='ARD1B, ARD2, hARD2', description='N(alpha)-acetyltransferase 11, NatA catalytic subunit', other_designations='=N-terminal acetyltransferase complex ARD1 subunit homolog B|ARD1 homolog B|N-alpha-acetyltransferase 11, NatA catalytic subunit|N-terminal acetyltransferase complex ARD1 subunit homolog B|human arrest defective 2|natA catalytic subunit Naa11', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79317118, end_position_on_the_genomic_accession=79326017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874034, CurrentID=0, Status='live', Symbol='UBXN7-AS1', Aliases='', description='UBXN7 antisense RNA 1', other_designations='UBXN7 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196431239, end_position_on_the_genomic_accession=196433034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927726, CurrentID=0, Status='live', Symbol='BEAN1-AS1', Aliases='CTD-2258A20.5', description='BEAN1 antisense RNA 1', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66469796, end_position_on_the_genomic_accession=66476402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927727, CurrentID=0, Status='live', Symbol='LOC101927727', Aliases='', description='uncharacterized LOC101927727', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=909632, end_position_on_the_genomic_accession=911230, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927728, CurrentID=0, Status='live', Symbol='LINC01476', Aliases='', description='long intergenic non-protein coding RNA 1476', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59430869, end_position_on_the_genomic_accession=59526857, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927729, CurrentID=0, Status='live', Symbol='LOC101927729', Aliases='', description='uncharacterized LOC101927729', other_designations='CTD-2568P8.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60305562, end_position_on_the_genomic_accession=60323907, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927730, CurrentID=0, Status='live', Symbol='LOC101927730', Aliases='', description='uncharacterized LOC101927730', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2883988, end_position_on_the_genomic_accession=2900676, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84787, CurrentID=0, Status='live', Symbol='KMT5C', Aliases='SUV420H2', description='lysine (K)-specific methyltransferase 5C', other_designations='lysine N-methyltransferase 5C|su(var)4-20 homolog 2|suppressor of variegation 4-20 homolog 2|suv4-20h2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55339210, end_position_on_the_genomic_accession=55348121, orientation='plus', exon_count=11, OMIM=613198, no_hdr0=''),
  84789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84789, CurrentID=0, Status='live', Symbol='MGC2889', Aliases='', description='uncharacterized protein MGC2889', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193241779, end_position_on_the_genomic_accession=193243972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84790, CurrentID=0, Status='live', Symbol='TUBA1C', Aliases='TUBA6, bcm948', description='tubulin alpha 1c', other_designations='alpha-tubulin 6|tubulin alpha-6 chain|tubulin, alpha 1c|tubulin, alpha 6', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49227926, end_position_on_the_genomic_accession=49273338, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84791, CurrentID=0, Status='live', Symbol='LINC00467', Aliases='C1orf97', description='long intergenic non-protein coding RNA 467', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211382755, end_position_on_the_genomic_accession=211432535, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84792, CurrentID=0, Status='live', Symbol='FAM220A', Aliases='C7orf70, SIPAR', description='family with sequence similarity 220 member A', other_designations='STAT3-interacting protein as a repressor|family with sequence similarity 220, member A', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6329409, end_position_on_the_genomic_accession=6348959, orientation='minus', exon_count=2, OMIM=616628, no_hdr0=''),
  84793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84793, CurrentID=0, Status='live', Symbol='FOXD2-AS1', Aliases='', description='FOXD2 antisense RNA 1 (head to head)', other_designations='FOXD2 antisense RNA 1 (non-protein coding)|uncharacterized protein MGC12982', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47432133, end_position_on_the_genomic_accession=47434641, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84795, CurrentID=0, Status='live', Symbol='PYROXD2', Aliases='C10orf33, FP3420', description='pyridine nucleotide-disulphide oxidoreductase domain 2', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98383565, end_position_on_the_genomic_accession=98415221, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  101927740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927740, CurrentID=0, Status='live', Symbol='LOC101927740', Aliases='', description='uncharacterized LOC101927740', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159100483, end_position_on_the_genomic_accession=159117478, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927741, CurrentID=0, Status='live', Symbol='LOC101927741', Aliases='', description='uncharacterized LOC101927741', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200394903, end_position_on_the_genomic_accession=200477636, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84798, CurrentID=0, Status='live', Symbol='C19orf48', Aliases='', description='chromosome 19 open reading frame 48', other_designations='multidrug resistance-related protein', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50797693, end_position_on_the_genomic_accession=50804853, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100874037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874037, CurrentID=0, Status='live', Symbol='CADM2-AS2', Aliases='', description='CADM2 antisense RNA 2', other_designations='CADM2 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=85799987, end_position_on_the_genomic_accession=85828050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927745, CurrentID=0, Status='live', Symbol='LOC101927745', Aliases='', description='uncharacterized LOC101927745', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15368447, end_position_on_the_genomic_accession=15444674, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101927746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927746, CurrentID=0, Status='live', Symbol='LOC101927746', Aliases='', description='uncharacterized LOC101927746', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101308340, end_position_on_the_genomic_accession=101321149, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84803, CurrentID=0, Status='live', Symbol='GPAT3', Aliases='AGPAT 10, AGPAT8, AGPAT9, HMFN0839, LPAAT-theta, MAG1', description='glycerol-3-phosphate acyltransferase 3', other_designations='1-AGP acyltransferase 9|1-AGPAT 9|1-acyl-sn-glycerol-3-phosphate O-acyltransferase 10|1-acyl-sn-glycerol-3-phosphate O-acyltransferase 9|1-acylglycerol-3-phosphate O-acyltransferase 8|GPAT-3|MAG-1|acyl-CoA:glycerol-3-phosphate acyltransferase 3|endoplasmic reticulum associated GPAT|hGPAT3|lung cancer metastasis-associated protein 1|lysophosphatidic acid acyltransferase theta|lysophosphatidic acid acyltransferase, theta', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83535914, end_position_on_the_genomic_accession=83605875, orientation='plus', exon_count=16, OMIM=610958, no_hdr0=''),
  84804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84804, CurrentID=0, Status='live', Symbol='MFSD9', Aliases='', description='major facilitator superfamily domain containing 9', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102714630, end_position_on_the_genomic_accession=102736882, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100874038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874038, CurrentID=0, Status='live', Symbol='CADM2-AS1', Aliases='', description='CADM2 antisense RNA 1', other_designations='CADM2 antisense RNA 1 (non-protein coding)', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=85992183, end_position_on_the_genomic_accession=86028007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  346950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346950, CurrentID=0, Status='live', Symbol='RPL37P6', Aliases='RPL37_5_905', description='ribosomal protein L37 pseudogene 6', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56588392, end_position_on_the_genomic_accession=56588781, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84807, CurrentID=0, Status='live', Symbol='NFKBID', Aliases='IkappaBNS, TA-NFKBH', description='nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, delta', other_designations='I-kappa-B-delta|T-cell activation NFKB-like protein|ikB-delta|ikappaBdelta', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35887129, end_position_on_the_genomic_accession=35902334, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  84808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84808, CurrentID=0, Status='live', Symbol='PERM1', Aliases='C1orf170', description='PPARGC1 and ESRR induced regulator, muscle 1', other_designations='PGC-1 and ERR-induced regulator in muscle 1|PPARGC1 and ESRR-induced regulator in muscle 1|pGC-1 and ERR-induced regulator in muscle protein 1|peroxisome proliferator-activated receptor gamma coactivator 1 and estrogen-related receptor-induced regulator in muscle 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=975199, end_position_on_the_genomic_accession=982117, orientation='minus', exon_count=5, OMIM=615921, no_hdr0=''),
  84809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84809, CurrentID=0, Status='live', Symbol='CROCCP2', Aliases='CROCCL1', description='ciliary rootlet coiled-coil, rootletin pseudogene 2', other_designations='ciliary rootlet coiled-coil, rootletin-like 1', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16618256, end_position_on_the_genomic_accession=16630906, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84811, CurrentID=0, Status='live', Symbol='BUD13', Aliases='Cwc26, fSAP71', description='BUD13 homolog', other_designations='functional spliceosome-associated protein 71', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116748170, end_position_on_the_genomic_accession=116772998, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  84812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84812, CurrentID=0, Status='live', Symbol='PLCD4', Aliases='', description='phospholipase C delta 4', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase delta-4|PLC delta4|phosphoinositide phospholipase C-delta-4|phospholipase C, delta 4', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218607765, end_position_on_the_genomic_accession=218637186, orientation='plus', exon_count=17, OMIM=605939, no_hdr0=''),
  101927757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927757, CurrentID=0, Status='live', Symbol='LOC101927757', Aliases='', description='uncharacterized LOC101927757', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34907056, end_position_on_the_genomic_accession=34943680, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84814, CurrentID=0, Status='live', Symbol='PLPP7', Aliases='C9orf67, NET39, PPAPDC3', description='phospholipid phosphatase 7 (inactive)', other_designations='phosphatidic acid phosphatase type 2 domain containing 3|phosphatidic acid phosphatase type 2 domain-containing protein 3', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131289682, end_position_on_the_genomic_accession=131309262, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84815, CurrentID=0, Status='live', Symbol='MGC12916', Aliases='', description='uncharacterized protein MGC12916', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14303740, end_position_on_the_genomic_accession=14305745, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84816, CurrentID=0, Status='live', Symbol='RTN4IP1', Aliases='NIMP', description='reticulon 4 interacting protein 1', other_designations='NOGO-interacting mitochondrial protein', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106571028, end_position_on_the_genomic_accession=106630500, orientation='minus', exon_count=11, OMIM=610502, no_hdr0=''),
  84817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84817, CurrentID=0, Status='live', Symbol='TXNDC17', Aliases='TRP14, TXNL5', description='thioredoxin domain containing 17', other_designations='14 kDa thioredoxin-related protein|protein 42-9-9|thioredoxin (Trx)-related protein, 14 kDa|thioredoxin-like 5|thioredoxin-like protein 5', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6640902, end_position_on_the_genomic_accession=6644541, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84818, CurrentID=0, Status='live', Symbol='IL17RC', Aliases='CANDF9, IL17-RL, IL17RL', description='interleukin 17 receptor C', other_designations='IL-17 receptor C|IL-17RL|IL17F receptor|IL17Rhom|interleukin-17 receptor homolog|interleukin-17 receptor-like protein|zcytoR14', map_location='3p25.3-p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9917074, end_position_on_the_genomic_accession=9933621, orientation='plus', exon_count=19, OMIM=610925, no_hdr0=''),
  101927763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927763, CurrentID=0, Status='live', Symbol='OLFM5P', Aliases='', description='olfactomedin family member 5, pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5514393, end_position_on_the_genomic_accession=5521635, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84820, CurrentID=0, Status='live', Symbol='POLR2J4', Aliases='RPB11-phi', description='polymerase (RNA) II (DNA directed) polypeptide J4, pseudogene', other_designations='polymerase (RNA) II (DNA directed) polypeptide J, 13.3kDa pseudogene', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43940895, end_position_on_the_genomic_accession=44019194, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  101927765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927765, CurrentID=0, Status='live', Symbol='LOC101927765', Aliases='', description='uncharacterized LOC101927765', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234372807, end_position_on_the_genomic_accession=234373593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927766, CurrentID=0, Status='live', Symbol='LOC101927766', Aliases='', description='uncharacterized LOC101927766', other_designations='', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159776772, end_position_on_the_genomic_accession=159871384, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84823, CurrentID=0, Status='live', Symbol='LMNB2', Aliases='LAMB2, LMN2', description='lamin B2', other_designations='lamin B3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2427907, end_position_on_the_genomic_accession=2456968, orientation='minus', exon_count=14, OMIM=150341, no_hdr0=''),
  84824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84824, CurrentID=0, Status='live', Symbol='FCRLA', Aliases='FCRL, FCRL1, FCRLM1, FCRLX, FCRLb, FCRLc1, FCRLc2, FCRLd, FCRLe, FCRX, FREB', description='Fc receptor-like A', other_designations='Fc receptor homolog expressed in B cells (FREB)|Fc receptor related protein X|Fc receptor-like and mucin-like 1|fc receptor homolog expressed in B-cells|fc receptor-like and mucin-like protein 1|fc receptor-like protein|fc receptor-related protein X', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161706972, end_position_on_the_genomic_accession=161714352, orientation='plus', exon_count=6, OMIM=606891, no_hdr0=''),
  101927769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927769, CurrentID=0, Status='live', Symbol='LOC101927769', Aliases='', description='uncharacterized LOC101927769', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20296708, end_position_on_the_genomic_accession=20311712, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84826, CurrentID=0, Status='live', Symbol='SFT2D3', Aliases='', description='SFT2 domain containing 3', other_designations='SFT2 domain-containing protein 3', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127701023, end_position_on_the_genomic_accession=127703833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874042, CurrentID=0, Status='live', Symbol='HM13-AS1', Aliases='', description='HM13 antisense RNA 1', other_designations='HM13 antisense RNA 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31567707, end_position_on_the_genomic_accession=31573263, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84830, CurrentID=0, Status='live', Symbol='ADTRP', Aliases='AIG1L, C6orf105, dJ413H6.1', description='androgen-dependent TFPI-regulating protein', other_designations='androgen-dependent TPF1-regulating protein', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11712895, end_position_on_the_genomic_accession=11779628, orientation='minus', exon_count=13, OMIM=614348, no_hdr0=''),
  84832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84832, CurrentID=0, Status='live', Symbol='ANKRD36BP1', Aliases='ANKRD26L1, ANKRD36BL1', description='ankyrin repeat domain 36B pseudogene 1', other_designations='ankyrin repeat domain 26-like 1|ankyrin repeat domain 36B-like 1 (pseudogene)', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168245565, end_position_on_the_genomic_accession=168247430, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84833, CurrentID=0, Status='live', Symbol='USMG5', Aliases='DAPIT, HCVFTP2, bA792D24.4', description='up-regulated during skeletal muscle growth 5 homolog (mouse)', other_designations='HCV F-transactivated protein 2|diabetes-associated protein in insulin-sensitive tissues|upregulated during skeletal muscle growth 5 homolog', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103389050, end_position_on_the_genomic_accession=103396513, orientation='minus', exon_count=6, OMIM=615204, no_hdr0=''),
  100874043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874043, CurrentID=0, Status='live', Symbol='TPRG1-AS1', Aliases='', description='TPRG1 antisense RNA 1', other_designations='TPRG1 antisense RNA 1 (non-protein coding)', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188941715, end_position_on_the_genomic_accession=188947639, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84836, CurrentID=0, Status='live', Symbol='ABHD14B', Aliases='CIB, HEL-S-299', description='abhydrolase domain containing 14B', other_designations='CCG1-interacting factor B|abhydrolase domain-containing protein 14B|cell cycle gene 1-interacting factor B|epididymis secretory protein Li 299', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51968510, end_position_on_the_genomic_accession=51974630, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84837, CurrentID=0, Status='live', Symbol='ARHGAP5-AS1', Aliases='C14orf128', description='ARHGAP5 antisense RNA 1 (head to head)', other_designations='ARHGAP5 antisense RNA 1 (non-protein coding)', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32075419, end_position_on_the_genomic_accession=32076699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84838, CurrentID=0, Status='live', Symbol='ZNF496', Aliases='NIZP1, ZFP496, ZKSCAN17, ZSCAN49', description='zinc finger protein 496', other_designations='NSD1 (nuclear receptor binding SET-domain containing 1)-interacting zinc finger protein 1|zinc finger protein with KRAB and SCAN domains 17', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247297413, end_position_on_the_genomic_accession=247333376, orientation='minus', exon_count=16, OMIM=613911, no_hdr0=''),
  84839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84839, CurrentID=0, Status='live', Symbol='RAX2', Aliases='ARMD6, CORD11, QRX, RAXL1', description='retina and anterior neural fold homeobox 2', other_designations='Q50-type retinal homeobox protein|retina and anterior neural fold homeobox like 1|retina and anterior neural fold homeobox-like protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3769089, end_position_on_the_genomic_accession=3772221, orientation='minus', exon_count=3, OMIM=610362, no_hdr0=''),
  100874044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874044, CurrentID=0, Status='live', Symbol='HHATL-AS1', Aliases='', description='HHATL antisense RNA 1', other_designations='HHATL antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84842, CurrentID=0, Status='live', Symbol='HPDL', Aliases='4-HPPD-L, GLOXD1', description='4-hydroxyphenylpyruvate dioxygenase-like', other_designations='glyoxalase domain containing 1|glyoxalase domain-containing protein 1', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45326873, end_position_on_the_genomic_accession=45328675, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131303, CurrentID=0, Status='live', Symbol='LOC100131303', Aliases='', description='uncharacterized LOC100131303', other_designations='CTC-479C5.6', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68013436, end_position_on_the_genomic_accession=68015911, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84844, CurrentID=0, Status='live', Symbol='PHF5A', Aliases='INI, Rds3, SAP14b, SF3B7, SF3b14b, bK223H9.2', description='PHD finger protein 5A', other_designations='PHD finger-like domain protein 5A|PHD-finger 5a|splicing factor 3B associated 14 kDa protein|splicing factor 3B-associated 14 kDa protein|splicing factor 3b, subunit 7', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41459717, end_position_on_the_genomic_accession=41468704, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927789, CurrentID=0, Status='live', Symbol='LOC101927789', Aliases='', description='putative ubiquitin-like protein FUBI-like protein ENSP00000310146', other_designations='CMB9-22P13.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65454421, end_position_on_the_genomic_accession=65466720, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  346990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346990, CurrentID=0, Status='live', Symbol='H2AFZP2', Aliases='', description='H2A histone family member Z pseudogene 2', other_designations='H2A histone family, member Z pseudogene 2', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70103412, end_position_on_the_genomic_accession=70104275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874045, CurrentID=0, Status='live', Symbol='MEAF6P1', Aliases='', description='MYST/Esa1-associated factor 6 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209181834, end_position_on_the_genomic_accession=209182070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84848, CurrentID=0, Status='live', Symbol='MIR503HG', Aliases='H19X', description='MIR503 host gene', other_designations='H19 X-linked co-expressed lncRNA|MIR503 host gene (non-protein coding)|uncharacterized protein MGC16121', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134543377, end_position_on_the_genomic_accession=134546630, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927793, CurrentID=0, Status='live', Symbol='LOC101927793', Aliases='', description='uncharacterized LOC101927793', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88883396, end_position_on_the_genomic_accession=88888453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84850, CurrentID=0, Status='live', Symbol='GLIS3-AS1', Aliases='C9orf70', description='GLIS3 antisense RNA 1', other_designations='GLIS3 antisense RNA 1 (non-protein coding)', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3898646, end_position_on_the_genomic_accession=3901248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84851, CurrentID=0, Status='live', Symbol='TRIM52', Aliases='RNF102', description='tripartite motif containing 52', other_designations='RING finger protein 102|tripartite motif-containing 52', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181256327, end_position_on_the_genomic_accession=181261160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927796, CurrentID=0, Status='live', Symbol='LINC01441', Aliases='', description='long intergenic non-protein coding RNA 1441', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55420336, end_position_on_the_genomic_accession=55427197, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927797, CurrentID=0, Status='live', Symbol='MIR548XHG', Aliases='', description='MIR548X host gene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=18561265, end_position_on_the_genomic_accession=18759812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927798, CurrentID=0, Status='live', Symbol='LOC101927798', Aliases='', description='uncharacterized LOC101927798', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133775551, end_position_on_the_genomic_accession=133777352, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84856, CurrentID=0, Status='live', Symbol='LINC00839', Aliases='', description='long intergenic non-protein coding RNA 839', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42475491, end_position_on_the_genomic_accession=42495337, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927801, CurrentID=0, Status='live', Symbol='LOC101927801', Aliases='', description='uncharacterized LOC101927801', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121790444, end_position_on_the_genomic_accession=121809945, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84858, CurrentID=0, Status='live', Symbol='ZNF503', Aliases='NOLZ-1, NOLZ1, Nlz2', description='zinc finger protein 503', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75279726, end_position_on_the_genomic_accession=75401894, orientation='minus', exon_count=5, OMIM=613902, no_hdr0=''),
  84859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84859, CurrentID=0, Status='live', Symbol='LRCH3', Aliases='', description='leucine-rich repeats and calponin homology (CH) domain containing 3', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197791226, end_position_on_the_genomic_accession=197884310, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  101927804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927804, CurrentID=0, Status='live', Symbol='LOC101927804', Aliases='', description='uncharacterized LOC101927804', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42842679, end_position_on_the_genomic_accession=42845860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84861, CurrentID=0, Status='live', Symbol='KLHL22', Aliases='KELCHL', description='kelch like family member 22', other_designations='kelch-like 22|kelch-like family member 22', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20441519, end_position_on_the_genomic_accession=20495883, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927806, CurrentID=0, Status='live', Symbol='LOC101927806', Aliases='', description='uncharacterized LOC101927806', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16871824, end_position_on_the_genomic_accession=16874037, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927807, CurrentID=0, Status='live', Symbol='LOC101927807', Aliases='', description='uncharacterized LOC101927807', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79357542, end_position_on_the_genomic_accession=79358234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84864, CurrentID=0, Status='live', Symbol='MINA', Aliases='MDIG53, NO52, ROX, MINA', description='MYC induced nuclear antigen', other_designations='60S ribosomal protein L27a histidine hydroxylase|histone lysine demethylase MINA|mineral dust induced gene protein|mineral dust-induced gene protein|myc-induced nuclear antigen, 53 kDa|nucleolar protein 52|ribosomal oxygenase MINA', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97941817, end_position_on_the_genomic_accession=97972451, orientation='minus', exon_count=12, OMIM=612049, no_hdr0=''),
  84865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84865, CurrentID=0, Status='live', Symbol='CCDC142', Aliases='', description='coiled-coil domain containing 142', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74472832, end_position_on_the_genomic_accession=74483230, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84866, CurrentID=0, Status='live', Symbol='TMEM25', Aliases='', description='transmembrane protein 25', other_designations='0610039J01Rik', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118531088, end_position_on_the_genomic_accession=118546598, orientation='plus', exon_count=10, OMIM=613934, no_hdr0=''),
  84867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84867, CurrentID=0, Status='live', Symbol='PTPN5', Aliases='PTPSTEP, STEP', description='protein tyrosine phosphatase, non-receptor type 5 (striatum-enriched)', other_designations='neural-specific protein-tyrosine phosphatase|protein-tyrosine phosphatase striatum-enriched|striatum-enriched protein-tyrosine phosphatase', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18727928, end_position_on_the_genomic_accession=18792721, orientation='minus', exon_count=16, OMIM=176879, no_hdr0=''),
  84868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84868, CurrentID=0, Status='live', Symbol='HAVCR2', Aliases='CD366, HAVcr-2, KIM-3, TIM3, TIMD-3, TIMD3, Tim-3', description='hepatitis A virus cellular receptor 2', other_designations='T cell immunoglobulin mucin 3|T cell immunoglobulin mucin-3|T-cell immunoglobulin and mucin domain-containing protein 3|T-cell immunoglobulin mucin family member 3|T-cell immunoglobulin mucin receptor 3|T-cell membrane protein 3|kidney injury molecule-3', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157085832, end_position_on_the_genomic_accession=157109237, orientation='minus', exon_count=7, OMIM=606652, no_hdr0=''),
  84869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84869, CurrentID=0, Status='live', Symbol='CBR4', Aliases='SDR45C1', description='carbonyl reductase 4', other_designations='3-oxoacyl-[acyl-carrier-protein] reductase|carbonic reductase 4|quinone reductase CBR4|short chain dehydrogenase/reductase family 45C member 1|short chain dehydrogenase/reductase family 45C, member 1', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168894623, end_position_on_the_genomic_accession=169010317, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84870, CurrentID=0, Status='live', Symbol='RSPO3', Aliases='CRISTIN1, PWTSR, THSD2', description='R-spondin 3', other_designations='R-spondin 3 homolog|protein with TSP type-1 repeat|roof plate-specific spondin-3|thrombospondin type-1 domain-containing protein 2|thrombospondin, type I, domain containing 2', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127118903, end_position_on_the_genomic_accession=127199481, orientation='plus', exon_count=5, OMIM=610574, no_hdr0=''),
  84871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84871, CurrentID=0, Status='live', Symbol='AGBL4', Aliases='CCP6', description='ATP/GTP binding protein-like 4', other_designations='ATP/GTP-binding protein-like 4', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48532834, end_position_on_the_genomic_accession=50023954, orientation='minus', exon_count=17, OMIM=616476, no_hdr0=''),
  84872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84872, CurrentID=0, Status='live', Symbol='ZC3H10', Aliases='ZC3HDC10', description='zinc finger CCCH-type containing 10', other_designations='zinc finger CCCH-type domain containing 10', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56118220, end_position_on_the_genomic_accession=56122497, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84873, CurrentID=0, Status='live', Symbol='ADGRG7', Aliases='GPR128', description='adhesion G protein-coupled receptor G7', other_designations='G protein-coupled receptor 128|G-protein coupled receptor 128', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100609589, end_position_on_the_genomic_accession=100695479, orientation='plus', exon_count=16, OMIM=612307, no_hdr0=''),
  84874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84874, CurrentID=0, Status='live', Symbol='ZNF514', Aliases='', description='zinc finger protein 514', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95130110, end_position_on_the_genomic_accession=95165187, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  84875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84875, CurrentID=0, Status='live', Symbol='PARP10', Aliases='ARTD10', description='poly(ADP-ribose) polymerase family member 10', other_designations='ADP-ribosyltransferase diphtheria toxin-like 10|poly (ADP-ribose) polymerase family, member 10', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143977152, end_position_on_the_genomic_accession=143986467, orientation='minus', exon_count=11, OMIM=609564, no_hdr0=''),
  84876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84876, CurrentID=0, Status='live', Symbol='ORAI1', Aliases='CRACM1, IMD9, ORAT1, TAM2, TMEM142A', description='ORAI calcium release-activated calcium modulator 1', other_designations='calcium release-activated calcium modulator 1|protein orai-1|transmembrane protein 142A', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121626550, end_position_on_the_genomic_accession=121642040, orientation='plus', exon_count=2, OMIM=610277, no_hdr0=''),
  101927821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927821, CurrentID=0, Status='live', Symbol='LINC01425', Aliases='', description='long intergenic non-protein coding RNA 1425', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=21746985, end_position_on_the_genomic_accession=21797415, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84878, CurrentID=0, Status='live', Symbol='ZBTB45', Aliases='ZNF499', description='zinc finger and BTB domain containing 45', other_designations='zinc finger protein 499', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58513530, end_position_on_the_genomic_accession=58520823, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84879, CurrentID=0, Status='live', Symbol='MFSD2A', Aliases='MFSD2, NLS1', description='major facilitator superfamily domain containing 2A', other_designations='major facilitator superfamily domain-containing protein 2A|sodium-dependent LPC symporter 1', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39955112, end_position_on_the_genomic_accession=39969968, orientation='plus', exon_count=14, OMIM=614397, no_hdr0=''),
  101927824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927824, CurrentID=0, Status='live', Symbol='LOC101927824', Aliases='', description='uncharacterized LOC101927824', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3052510, end_position_on_the_genomic_accession=3069273, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84881, CurrentID=0, Status='live', Symbol='RPUSD4', Aliases='', description='RNA pseudouridylate synthase domain containing 4', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126202094, end_position_on_the_genomic_accession=126211692, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84883, CurrentID=0, Status='live', Symbol='AIFM2', Aliases='AMID, PRG3', description='apoptosis inducing factor, mitochondria associated 2', other_designations='apoptosis-inducing factor (AIF)-homologous mitochondrion-associated inducer of death|apoptosis-inducing factor (AIF)-like mitochondrion-associated inducer of death|apoptosis-inducing factor, mitochondrion-associated, 2|p53-responsive gene 3 protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70112267, end_position_on_the_genomic_accession=70132934, orientation='minus', exon_count=9, OMIM=605159, no_hdr0=''),
  101927828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927828, CurrentID=0, Status='live', Symbol='OVOL1-AS1', Aliases='', description='OVOL1 antisense RNA 1', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65789001, end_position_on_the_genomic_accession=65790868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84885, CurrentID=0, Status='live', Symbol='ZDHHC12', Aliases='DHHC-12, ZNF400', description='zinc finger, DHHC-type containing 12', other_designations='zinc finger DHHC domain-containing protein 12|zinc finger protein 400', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128720869, end_position_on_the_genomic_accession=128724141, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84886, CurrentID=0, Status='live', Symbol='C1orf198', Aliases='', description='chromosome 1 open reading frame 198', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230837119, end_position_on_the_genomic_accession=230869589, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84888, CurrentID=0, Status='live', Symbol='SPPL2A', Aliases='IMP3, PSL2', description='signal peptide peptidase like 2A', other_designations='IMP-3|SPP-like 2A|intramembrane cleaving protease|intramembrane protease 3|presenilin-like protein 2', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50707533, end_position_on_the_genomic_accession=50766158, orientation='minus', exon_count=17, OMIM=608238, no_hdr0=''),
  84889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84889, CurrentID=0, Status='live', Symbol='SLC7A3', Aliases='ATRC3, CAT-3, CAT3', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 3', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70925579, end_position_on_the_genomic_accession=70931125, orientation='minus', exon_count=12, OMIM=300443, no_hdr0=''),
  84890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84890, CurrentID=0, Status='live', Symbol='ADO', Aliases='C10orf22', description='2-aminoethanethiol (cysteamine) dioxygenase', other_designations='cysteamine (2-aminoethanethiol) dioxygenase (ADO)|cysteamine dioxygenase', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62804756, end_position_on_the_genomic_accession=62808479, orientation='plus', exon_count=1, OMIM=611392, no_hdr0=''),
  84891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84891, CurrentID=0, Status='live', Symbol='ZSCAN10', Aliases='ZNF206', description='zinc finger and SCAN domain containing 10', other_designations='zinc finger protein 206', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3088890, end_position_on_the_genomic_accession=3099317, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84892, CurrentID=0, Status='live', Symbol='POMGNT2', Aliases='AGO61, C3orf39, GTDC2, MDDGA8', description='protein O-linked mannose N-acetylglucosaminyltransferase 2 (beta 1,4-)', other_designations='glycosyltransferase-like domain containing 2|glycosyltransferase-like domain-containing protein 2', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43079229, end_position_on_the_genomic_accession=43106083, orientation='minus', exon_count=4, OMIM=614828, no_hdr0=''),
  84893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84893, CurrentID=0, Status='live', Symbol='FBXO18', Aliases='FBH1, Fbx18, hFBH1', description='F-box protein, helicase, 18', other_designations='F-box only protein 18|F-box only protein, helicase, 18', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5889572, end_position_on_the_genomic_accession=5937595, orientation='plus', exon_count=26, OMIM=607222, no_hdr0=''),
  84894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84894, CurrentID=0, Status='live', Symbol='LINGO1', Aliases='LERN1, LRRN6A, UNQ201', description='leucine rich repeat and Ig domain containing 1', other_designations='leucine rich repeat neuronal 6A|leucine-rich repeat and immunoglobulin domain-containing protein 1|leucine-rich repeat neuronal protein 1', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77613024, end_position_on_the_genomic_accession=77820900, orientation='minus', exon_count=16, OMIM=609791, no_hdr0=''),
  84895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84895, CurrentID=0, Status='live', Symbol='FAM73B', Aliases='C9orf54', description='family with sequence similarity 73 member B', other_designations='family with sequence similarity 73, member B', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129036605, end_position_on_the_genomic_accession=129072082, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  84896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84896, CurrentID=0, Status='live', Symbol='ATAD1', Aliases='AFDC1, FNP001, THORASE', description='ATPase family, AAA domain containing 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87751851, end_position_on_the_genomic_accession=87841349, orientation='minus', exon_count=16, OMIM=614452, no_hdr0=''),
  84897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84897, CurrentID=0, Status='live', Symbol='TBRG1', Aliases='NIAM, TB-5', description='transforming growth factor beta regulator 1', other_designations='nuclear interactor of ARF and MDM2', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124622599, end_position_on_the_genomic_accession=124635926, orientation='plus', exon_count=10, OMIM=610614, no_hdr0=''),
  84898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84898, CurrentID=0, Status='live', Symbol='PLXDC2', Aliases='TEM7R', description='plexin domain containing 2', other_designations='1200007L24Rik|tumor endothelial marker 7-related protein', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19816227, end_position_on_the_genomic_accession=20289856, orientation='plus', exon_count=15, OMIM=606827, no_hdr0=''),
  84899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84899, CurrentID=0, Status='live', Symbol='TMTC4', Aliases='', description='transmembrane and tetratricopeptide repeat containing 4', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100603836, end_position_on_the_genomic_accession=100675089, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  84900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84900, CurrentID=0, Status='live', Symbol='RNFT2', Aliases='TMEM118', description='ring finger protein, transmembrane 2', other_designations='transmembrane protein 118', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116738291, end_position_on_the_genomic_accession=116853631, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  84901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84901, CurrentID=0, Status='live', Symbol='NFATC2IP', Aliases='ESC2, NIP45, RAD60', description='nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 2 interacting protein', other_designations='45 kDa NF-AT-interacting protein|45 kDa NFAT-interacting protein|nuclear factor of activated T-cells, cytoplasmic 2-interacting protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28950988, end_position_on_the_genomic_accession=28966465, orientation='plus', exon_count=8, OMIM=614525, no_hdr0=''),
  84902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84902, CurrentID=0, Status='live', Symbol='CEP89', Aliases='CCDC123, CEP123', description='centrosomal protein 89kDa', other_designations='centrosomal protein 123|coiled-coil domain containing 123|coiled-coil domain-containing protein 123, mitochondrial', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32878998, end_position_on_the_genomic_accession=32972029, orientation='minus', exon_count=21, OMIM=615470, no_hdr0=''),
  101927847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927847, CurrentID=0, Status='live', Symbol='LOC101927847', Aliases='', description='uncharacterized LOC101927847', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89969416, end_position_on_the_genomic_accession=89987486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84904, CurrentID=0, Status='live', Symbol='ARHGEF39', Aliases='C9orf100', description='Rho guanine nucleotide exchange factor 39', other_designations='Rho guanine nucleotide exchange factor (GEF) 39|vav-like protein C9orf100', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35658875, end_position_on_the_genomic_accession=35675738, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  84905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84905, CurrentID=0, Status='live', Symbol='ZNF341', Aliases='', description='zinc finger protein 341', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33731760, end_position_on_the_genomic_accession=33792269, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  347051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347051, CurrentID=0, Status='live', Symbol='SLC10A5', Aliases='P5', description='solute carrier family 10 member 5', other_designations='Na(+)/bile acid cotransporter 5|solute carrier family 10 (sodium/bile acid cotransporter family), member 5|solute carrier family 10, member 5', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81693656, end_position_on_the_genomic_accession=81694972, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84908, CurrentID=0, Status='live', Symbol='FAM136A', Aliases='', description='family with sequence similarity 136 member A', other_designations='family with sequence similarity 136, member A', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70295976, end_position_on_the_genomic_accession=70302088, orientation='minus', exon_count=4, OMIM=616275, no_hdr0=''),
  84909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84909, CurrentID=0, Status='live', Symbol='C9orf3', Aliases='AOPEP, AP-O, APO, C90RF3, ONPEP', description='chromosome 9 open reading frame 3', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94726604, end_position_on_the_genomic_accession=95147311, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  84910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84910, CurrentID=0, Status='live', Symbol='TMEM87B', Aliases='', description='transmembrane protein 87B', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112055223, end_position_on_the_genomic_accession=112119318, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  84911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84911, CurrentID=0, Status='live', Symbol='ZNF382', Aliases='KS1', description='zinc finger protein 382', other_designations='KRAB/zinc finger suppressor protein 1|multiple zinc finger and krueppel-associated box protein KS1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36605305, end_position_on_the_genomic_accession=36628597, orientation='plus', exon_count=5, OMIM=609516, no_hdr0=''),
  84912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84912, CurrentID=0, Status='live', Symbol='SLC35B4', Aliases='YEA, YEA4', description='solute carrier family 35 (UDP-xylose/UDP-N-acetylglucosamine transporter), member B4', other_designations='UDP-Xylose/N-Acetylglucosamine transporter|YEA4 homolog|solute carrier family 35, member B4', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134289332, end_position_on_the_genomic_accession=134320314, orientation='minus', exon_count=11, OMIM=610923, no_hdr0=''),
  84913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84913, CurrentID=0, Status='live', Symbol='ATOH8', Aliases='HATH6, bHLHa21', description='atonal bHLH transcription factor 8', other_designations='atonal homolog 6|atonal homolog 8|atonal homolog bHLH transcription factor 8|basic helix loop helix transcription factor 6|basic helix-loop-helix transcription factor 6|class A basic helix-loop-helix protein 21|helix-loop-helix protein hATH-6', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85753786, end_position_on_the_genomic_accession=85791383, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  84914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84914, CurrentID=0, Status='live', Symbol='ZNF587', Aliases='ZF6', description='zinc finger protein 587', other_designations='zinc finger protein zfp6', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57849813, end_position_on_the_genomic_accession=57865123, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84915, CurrentID=0, Status='live', Symbol='FAM222A', Aliases='C12orf34', description='family with sequence similarity 222 member A', other_designations='family with sequence similarity 222, member A', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109714382, end_position_on_the_genomic_accession=109770509, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84916, CurrentID=0, Status='live', Symbol='UTP4', Aliases='CIRH1A, CIRHIN, NAIC, TEX292', description='UTP4 small subunit processome component', other_designations='UTP4 small subunit (SSU) processome component|UTP4, small subunit (SSU) processome component, homolog|cirrhosis, autosomal recessive 1A (cirhin)|testis expressed gene 292', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69132596, end_position_on_the_genomic_accession=69169034, orientation='plus', exon_count=18, OMIM=607456, no_hdr0=''),
  101927861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927861, CurrentID=0, Status='live', Symbol='LINC01293', Aliases='', description='long intergenic non-protein coding RNA 1293', other_designations='TCONS_00002938', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74939389, end_position_on_the_genomic_accession=74942670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84918, CurrentID=0, Status='live', Symbol='LRP11', Aliases='MANSC3, bA350J20.3', description='LDL receptor related protein 11', other_designations='LRP-11|low density lipoprotein receptor-related protein 11', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149795188, end_position_on_the_genomic_accession=149864348, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  84919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84919, CurrentID=0, Status='live', Symbol='PPP1R15B', Aliases='CREP', description='protein phosphatase 1 regulatory subunit 15B', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 15B|protein phosphatase 1, regulatory subunit 15B', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204400275, end_position_on_the_genomic_accession=204411817, orientation='minus', exon_count=3, OMIM=613257, no_hdr0=''),
  84920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84920, CurrentID=0, Status='live', Symbol='ALG10', Aliases='ALG10A, DIE2, KCR1', description='ALG10, alpha-1,2-glucosyltransferase', other_designations='alpha-1,2-glucosyltransferase ALG10-A|alpha-2-glucosyltransferase ALG10-A|alpha2-glucosyltransferase|asparagine-linked glycosylation 10 homolog (yeast, alpha-1,2-glucosyltransferase)|asparagine-linked glycosylation 10, alpha-1,2-glucosyltransferase homolog|asparagine-linked glycosylation protein 10 homolog A|derepression of ITR1 expression 2 homolog|dolichyl-P-Glc:Glc(2)Man(9)GlcNAc(2)-PP-dolichol alpha-1,2- glucosyltransferase|potassium channel regulator 1', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34022281, end_position_on_the_genomic_accession=34028302, orientation='plus', exon_count=3, OMIM=603313, no_hdr0=''),
  101927865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927865, CurrentID=0, Status='live', Symbol='LOC101927865', Aliases='', description='uncharacterized LOC101927865', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207186690, end_position_on_the_genomic_accession=207254316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84922, CurrentID=0, Status='live', Symbol='FIZ1', Aliases='ZNF798', description='FLT3-interacting zinc finger 1', other_designations='zinc finger protein 798', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55591371, end_position_on_the_genomic_accession=55599696, orientation='minus', exon_count=5, OMIM=609133, no_hdr0=''),
  84923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84923, CurrentID=0, Status='live', Symbol='FAM104A', Aliases='', description='family with sequence similarity 104 member A', other_designations='family with sequence similarity 104, member A', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73207353, end_position_on_the_genomic_accession=73232394, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84924, CurrentID=0, Status='live', Symbol='ZNF566', Aliases='', description='zinc finger protein 566', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36436034, end_position_on_the_genomic_accession=36489902, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  84925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84925, CurrentID=0, Status='live', Symbol='DIRC2', Aliases='RCC4', description='disrupted in renal carcinoma 2', other_designations='disrupted in renal cancer protein 2|renal cell carcinoma 4', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122795054, end_position_on_the_genomic_accession=122881139, orientation='plus', exon_count=9, OMIM=602773, no_hdr0=''),
  84926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84926, CurrentID=0, Status='live', Symbol='SPRYD3', Aliases='', description='SPRY domain containing 3', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53064316, end_position_on_the_genomic_accession=53079420, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101927871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927871, CurrentID=0, Status='live', Symbol='LOC101927871', Aliases='', description='uncharacterized LOC101927871', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87959420, end_position_on_the_genomic_accession=87976899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84928, CurrentID=0, Status='live', Symbol='TMEM209', Aliases='NET31', description='transmembrane protein 209', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130164713, end_position_on_the_genomic_accession=130206197, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  84929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84929, CurrentID=0, Status='live', Symbol='FIBCD1', Aliases='', description='fibrinogen C domain containing 1', other_designations='', map_location='9q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130902438, end_position_on_the_genomic_accession=130939068, orientation='minus', exon_count=8, OMIM=613357, no_hdr0=''),
  84930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84930, CurrentID=0, Status='live', Symbol='MASTL', Aliases='GREATWALL, GW, GWL, MAST-L, THC2, hGWL', description='microtubule associated serine/threonine kinase like', other_designations='greatwall kinase homolog|greatwall protein kinase|microtubule associated serine/threonine kinase-like|microtubule-associated serine/threonine-protein kinase-like', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27154824, end_position_on_the_genomic_accession=27186920, orientation='plus', exon_count=14, OMIM=608221, no_hdr0=''),
  100874059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874059, CurrentID=0, Status='live', Symbol='LINC00297', Aliases='NCRNA00297', description='long intergenic non-protein coding RNA 297', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29872613, end_position_on_the_genomic_accession=29888405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84932, CurrentID=0, Status='live', Symbol='RAB2B', Aliases='', description='RAB2B, member RAS oncogene family', other_designations='GTP-binding protein RAB2B|RAS family, member RAB2B', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21459020, end_position_on_the_genomic_accession=21476973, orientation='minus', exon_count=11, OMIM=607466, no_hdr0=''),
  84933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84933, CurrentID=0, Status='live', Symbol='C8orf76', Aliases='', description='chromosome 8 open reading frame 76', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123219956, end_position_on_the_genomic_accession=123241398, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84934, CurrentID=0, Status='live', Symbol='RITA1', Aliases='C12orf52, RITA', description='RBPJ interacting and tubulin associated 1', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113185526, end_position_on_the_genomic_accession=113192368, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84935, CurrentID=0, Status='live', Symbol='MEDAG', Aliases='AWMS3, C13orf33, MEDA-4, MEDA4, hAWMS3', description='mesenteric estrogen-dependent adipogenesis', other_designations='activated in W/Wv mouse stomach 3 homolog|mesenteric estrogen-dependent adipose 4', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30906175, end_position_on_the_genomic_accession=30925572, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84936, CurrentID=0, Status='live', Symbol='ZFYVE19', Aliases='ANCHR, MPFYVE', description='zinc finger, FYVE domain containing 19', other_designations='MLL partner containing FYVE domain|zinc finger FYVE domain-containing protein 19', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40807076, end_position_on_the_genomic_accession=40814569, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  84937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84937, CurrentID=0, Status='live', Symbol='ZNRF1', Aliases='NIN283', description='zinc and ring finger 1, E3 ubiquitin protein ligase', other_designations='nerve injury gene 283|nerve injury-induced gene 283 protein|zinc and ring finger protein 1|zinc/RING finger protein 1', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74999017, end_position_on_the_genomic_accession=75110994, orientation='plus', exon_count=10, OMIM=612060, no_hdr0=''),
  84938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84938, CurrentID=0, Status='live', Symbol='ATG4C', Aliases='APG4-C, APG4C, AUTL1, AUTL3', description='autophagy related 4C, cysteine peptidase', other_designations='APG4 autophagy 4 homolog C|ATG4 autophagy related 4 homolog C|AUT-like 1, cysteine endopeptidase|AUT-like 3 cysteine endopeptidase|autophagin-3|autophagy-related cysteine endopeptidase 3|autophagy-related protein 4 homolog C', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62784106, end_position_on_the_genomic_accession=62865270, orientation='plus', exon_count=13, OMIM=611339, no_hdr0=''),
  84939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84939, CurrentID=0, Status='live', Symbol='MUM1', Aliases='EXPAND1, HSPC211, MUM-1', description='melanoma associated antigen (mutated) 1', other_designations='melanoma ubiquitous mutated protein|mutated melanoma-associated antigen 1|protein expandere', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1354977, end_position_on_the_genomic_accession=1378431, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  84940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84940, CurrentID=0, Status='live', Symbol='CORO6', Aliases='', description='coronin 6', other_designations='clipin-E|coronin, actin binding protein 6|coronin-like protein E', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29614756, end_position_on_the_genomic_accession=29622908, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  84941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84941, CurrentID=0, Status='live', Symbol='HSH2D', Aliases='ALX, HSH2', description='hematopoietic SH2 domain containing', other_designations='adaptor in lymphocytes of unknown function X', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16134028, end_position_on_the_genomic_accession=16158575, orientation='plus', exon_count=10, OMIM=608349, no_hdr0=''),
  84942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84942, CurrentID=0, Status='live', Symbol='WDR73', Aliases='GAMOS, HSPC264', description='WD repeat domain 73', other_designations='FLJ00296 protein', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84641495, end_position_on_the_genomic_accession=84654343, orientation='minus', exon_count=9, OMIM=616144, no_hdr0=''),
  100874061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874061, CurrentID=0, Status='live', Symbol='WASF3-AS1', Aliases='', description='WASF3 antisense RNA 1', other_designations='WASF3 antisense RNA 1 (non-protein coding)', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347088, CurrentID=0, Status='live', Symbol='ADGRD2', Aliases='GPR144, PGR24', description='adhesion G protein-coupled receptor D2', other_designations='G protein-coupled receptor 144|G protein-coupled receptor PGR24|G-protein coupled receptor 144|G-protein coupled receptor GPR144|G-protein coupled receptor PGR24', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124451087, end_position_on_the_genomic_accession=124477771, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  84945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84945, CurrentID=0, Status='live', Symbol='ABHD13', Aliases='BEM46L1, C13orf6, bA153I24.2', description='abhydrolase domain containing 13', other_designations='abhydrolase domain-containing protein 13', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108218415, end_position_on_the_genomic_accession=108234255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84946, CurrentID=0, Status='live', Symbol='LTV1', Aliases='C6orf93, dJ468K18.4', description='LTV1 ribosome biogenesis factor', other_designations='LTV1 homolog', map_location='6q24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143843371, end_position_on_the_genomic_accession=143863806, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  84947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84947, CurrentID=0, Status='live', Symbol='SERAC1', Aliases='', description='serine active site containing 1', other_designations='serine active site-containing protein 1', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158109504, end_position_on_the_genomic_accession=158168280, orientation='minus', exon_count=19, OMIM=614725, no_hdr0=''),
  84948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84948, CurrentID=0, Status='live', Symbol='TIGD5', Aliases='', description='tigger transposable element derived 5', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143597904, end_position_on_the_genomic_accession=143600315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  84950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84950, CurrentID=0, Status='live', Symbol='PRPF38A', Aliases='Prp38', description='pre-mRNA processing factor 38A', other_designations='PRP38 pre-mRNA processing factor 38 domain containing A', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52404547, end_position_on_the_genomic_accession=52418320, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84951, CurrentID=0, Status='live', Symbol='TNS4', Aliases='CTEN, PP14434', description='tensin 4', other_designations='C terminal tensin like|C-terminal tensin-like protein', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40475828, end_position_on_the_genomic_accession=40501681, orientation='minus', exon_count=13, OMIM=608385, no_hdr0=''),
  84952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84952, CurrentID=0, Status='live', Symbol='CGNL1', Aliases='JACOP', description='cingulin-like 1', other_designations='junction-associated coiled-coil protein|paracingulin', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57376505, end_position_on_the_genomic_accession=57550727, orientation='plus', exon_count=24, OMIM=607856, no_hdr0=''),
  84953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84953, CurrentID=0, Status='live', Symbol='MICALCL', Aliases='Ebitein1', description='MICAL C-terminal like', other_designations='ERK2-binding testicular protein 1|ebitein-1', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12286900, end_position_on_the_genomic_accession=12359144, orientation='plus', exon_count=9, OMIM=612355, no_hdr0=''),
  84954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84954, CurrentID=0, Status='live', Symbol='MPND', Aliases='', description='MPN domain containing', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4343527, end_position_on_the_genomic_accession=4360086, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  84955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84955, CurrentID=0, Status='live', Symbol='NUDCD1', Aliases='CML66, OVA66', description='NudC domain containing 1', other_designations='chronic myelogenous leukemia tumor antigen 66|tumor antigen CML66', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109240919, end_position_on_the_genomic_accession=109334121, orientation='minus', exon_count=12, OMIM=606109, no_hdr0=''),
  101927900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927900, CurrentID=0, Status='live', Symbol='LINC01477', Aliases='', description='long intergenic non-protein coding RNA 1477', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=40066286, end_position_on_the_genomic_accession=40099233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84957, CurrentID=0, Status='live', Symbol='RELT', Aliases='TNFRSF19L, TRLT', description='RELT tumor necrosis factor receptor', other_designations='receptor expressed in lymphoid tissues', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73376360, end_position_on_the_genomic_accession=73397474, orientation='plus', exon_count=13, OMIM=611211, no_hdr0=''),
  84958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84958, CurrentID=0, Status='live', Symbol='SYTL1', Aliases='JFC1, SLP1', description='synaptotagmin like 1', other_designations='NADPH oxidase-related, C2 domain-containing protein|exophilin-7|synaptotagmin-like 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27341992, end_position_on_the_genomic_accession=27353932, orientation='plus', exon_count=15, OMIM=608042, no_hdr0=''),
  84959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84959, CurrentID=0, Status='live', Symbol='UBASH3B', Aliases='STS-1, STS1, TULA-2, TULA2, p70', description='ubiquitin associated and SH3 domain containing B', other_designations='Cbl-interacting protein Sts-1|SH3 domain-containing 70 kDa protein, suppressor of T-cell receptor signaling 1, nm23-phosphorylated unknown substrate|T-cell ubiquitin ligand 2|cbl-interacting protein p70|nm23-phosphorylated unknown substrate|suppressor of T-cell receptor signaling 1|tyrosine-protein phosphatase STS1/TULA2', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122655690, end_position_on_the_genomic_accession=122814479, orientation='plus', exon_count=15, OMIM=609201, no_hdr0=''),
  84960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84960, CurrentID=0, Status='live', Symbol='CCDC183', Aliases='KIAA1984, PARF, bA216L13.7', description='coiled-coil domain containing 183', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136796338, end_position_on_the_genomic_accession=136807741, orientation='plus', exon_count=14, OMIM=615955, no_hdr0=''),
  84961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84961, CurrentID=0, Status='live', Symbol='FBXL20', Aliases='Fbl2, Fbl20', description='F-box and leucine-rich repeat protein 20', other_designations='F-box protein FBL2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39252644, end_position_on_the_genomic_accession=39402661, orientation='minus', exon_count=18, OMIM=609086, no_hdr0=''),
  84962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84962, CurrentID=0, Status='live', Symbol='AJUBA', Aliases='JUB', description='ajuba LIM protein', other_designations='jub, ajuba homolog', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22971174, end_position_on_the_genomic_accession=22982642, orientation='minus', exon_count=9, OMIM=609066, no_hdr0=''),
  101927907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927907, CurrentID=0, Status='live', Symbol='LOC101927907', Aliases='', description='uncharacterized LOC101927907', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76985965, end_position_on_the_genomic_accession=77009788, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84964, CurrentID=0, Status='live', Symbol='ALKBH6', Aliases='ABH6', description='alkB homolog 6', other_designations='alkB, alkylation repair homolog 6|alkylated DNA repair protein alkB homolog 6|alkylation repair homolog 6|probable alpha-ketoglutarate-dependent dioxygenase ABH6', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36009116, end_position_on_the_genomic_accession=36014243, orientation='minus', exon_count=8, OMIM=613304, no_hdr0=''),
  84966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84966, CurrentID=0, Status='live', Symbol='IGSF21', Aliases='', description='immunoglobin superfamily member 21', other_designations='immunoglobin superfamily, member 21', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18107746, end_position_on_the_genomic_accession=18378484, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  84967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84967, CurrentID=0, Status='live', Symbol='LSM10', Aliases='MST074, MSTP074', description='LSM10, U7 small nuclear RNA associated', other_designations='U7 snRNP-specific Sm-like protein LSM10', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36393420, end_position_on_the_genomic_accession=36397959, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84968, CurrentID=0, Status='live', Symbol='PNMA6A', Aliases='MA6, PNMA6, PNMA6C', description='paraneoplastic Ma antigen family member 6A', other_designations='paraneoplastic Ma antigen family member 6C|paraneoplastic antigen-like 6A related protein|paraneoplastic antigen-like protein 6C', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153072414, end_position_on_the_genomic_accession=153075019, orientation='plus', exon_count=2, OMIM=300917, no_hdr0=''),
  84969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84969, CurrentID=0, Status='live', Symbol='TOX2', Aliases='C20orf100, GCX-1, GCX1, dJ1108D11.2, dJ495O3.1', description='TOX high mobility group box family member 2', other_designations='granulosa cell HMG box 1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43914852, end_position_on_the_genomic_accession=44069616, orientation='plus', exon_count=12, OMIM=611163, no_hdr0=''),
  84970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84970, CurrentID=0, Status='live', Symbol='C1orf94', Aliases='', description='chromosome 1 open reading frame 94', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34166883, end_position_on_the_genomic_accession=34219131, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84971, CurrentID=0, Status='live', Symbol='ATG4D', Aliases='APG4-D, APG4D, AUTL4', description='autophagy related 4D, cysteine peptidase', other_designations='APG4 autophagy 4 homolog D|ATG4 autophagy related 4 homolog D|AUT-like 4 cysteine endopeptidase|autophagin-4|autophagy-related cysteine endopeptidase 4|autophagy-related protein 4 homolog D|cysteine protease involved in autophagy', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10543894, end_position_on_the_genomic_accession=10553423, orientation='plus', exon_count=10, OMIM=611340, no_hdr0=''),
  101927916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927916, CurrentID=0, Status='live', Symbol='LOC101927916', Aliases='', description='uncharacterized LOC101927916', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91484278, end_position_on_the_genomic_accession=91487307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  84973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84973, CurrentID=0, Status='live', Symbol='SNHG7', Aliases='NCRNA00061', description='small nucleolar RNA host gene 7', other_designations='small nucleolar RNA host gene (non-protein coding) 7|small nucleolar RNA host gene 7 (non-protein coding)', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136724594, end_position_on_the_genomic_accession=136728184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84975, CurrentID=0, Status='live', Symbol='MFSD5', Aliases='hsMOT2', description='major facilitator superfamily domain containing 5', other_designations='major facilitator superfamily domain-containing protein 5|molybdate transporter 2 homolog', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53251251, end_position_on_the_genomic_accession=53254406, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84976, CurrentID=0, Status='live', Symbol='DISP1', Aliases='DISPA', description='dispatched homolog 1 (Drosophila)', other_designations='dispatched A', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222814514, end_position_on_the_genomic_accession=223005995, orientation='plus', exon_count=22, OMIM=607502, no_hdr0=''),
  101927921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927921, CurrentID=0, Status='live', Symbol='LINC01478', Aliases='', description='long intergenic non-protein coding RNA 1478', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44323435, end_position_on_the_genomic_accession=44531697, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  84978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84978, CurrentID=0, Status='live', Symbol='FRMD5', Aliases='', description='FERM domain containing 5', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43870761, end_position_on_the_genomic_accession=44195577, orientation='minus', exon_count=18, OMIM=616309, no_hdr0=''),
  100874067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874067, CurrentID=0, Status='live', Symbol='MYO16-AS2', Aliases='', description='MYO16 antisense RNA 2', other_designations='MYO16 antisense RNA 2 (non-protein coding)', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927924, CurrentID=0, Status='live', Symbol='LOC101927924', Aliases='', description='uncharacterized LOC101927924', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129923177, end_position_on_the_genomic_accession=129946703, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84981, CurrentID=0, Status='live', Symbol='MIR22HG', Aliases='C17orf91', description='MIR22 host gene', other_designations='MIR22 host gene (non-protein coding)', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1711504, end_position_on_the_genomic_accession=1716272, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927926, CurrentID=0, Status='live', Symbol='LOC101927926', Aliases='', description='uncharacterized LOC101927926', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77915934, end_position_on_the_genomic_accession=77918009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84983, CurrentID=0, Status='live', Symbol='FAM222A-AS1', Aliases='', description='FAM222A antisense RNA 1', other_designations='FAM222A antisense RNA 1 (non-protein coding)', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109734209, end_position_on_the_genomic_accession=109773487, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84984, CurrentID=0, Status='live', Symbol='CEP19', Aliases='C3orf34, MOSPGF', description='centrosomal protein 19kDa', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196706273, end_position_on_the_genomic_accession=196712294, orientation='minus', exon_count=3, OMIM=615586, no_hdr0=''),
  84985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84985, CurrentID=0, Status='live', Symbol='FAM83A', Aliases='BJ-TSA-9', description='family with sequence similarity 83 member A', other_designations='family with sequence similarity 83, member A|tumor antigen BJ-TSA-9|tumor-specific gene expressed in prostate protein', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123179047, end_position_on_the_genomic_accession=123210079, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  84986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84986, CurrentID=0, Status='live', Symbol='ARHGAP19', Aliases='', description='Rho GTPase activating protein 19', other_designations='putative RhoGAP protein|rho-type GTPase-activating protein 19', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97222173, end_position_on_the_genomic_accession=97292673, orientation='minus', exon_count=13, OMIM=611587, no_hdr0=''),
  84987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84987, CurrentID=0, Status='live', Symbol='COX14', Aliases='C12orf62', description='COX14, cytochrome c oxidase assembly factor', other_designations='COX14 cytochrome c oxidase assembly factor|COX14 cytochrome c oxidase assembly homolog|cytochrome c oxidase assembly homolog 14', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50109635, end_position_on_the_genomic_accession=50120457, orientation='plus', exon_count=4, OMIM=614478, no_hdr0=''),
  84988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84988, CurrentID=0, Status='live', Symbol='PPP1R16A', Aliases='MYPT3', description='protein phosphatase 1 regulatory subunit 16A', other_designations='likley ortholog of mouse myosin phosphatase targeting subunit 3|myosin phosphatase-targeting subunit 3|protein phosphatase 1, regulatory (inhibitor) subunit 16A|protein phosphatase 1, regulatory subunit 16A', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144477987, end_position_on_the_genomic_accession=144502121, orientation='plus', exon_count=17, OMIM=609172, no_hdr0=''),
  84989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84989, CurrentID=0, Status='live', Symbol='JMJD1C-AS1', Aliases='', description='JMJD1C antisense RNA 1', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63465229, end_position_on_the_genomic_accession=63466563, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927934, CurrentID=0, Status='live', Symbol='LOC101927934', Aliases='', description='uncharacterized LOC101927934', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134506552, end_position_on_the_genomic_accession=134509230, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84991, CurrentID=0, Status='live', Symbol='RBM17', Aliases='SPF45', description='RNA binding motif protein 17', other_designations='45 kDa-splicing factor|RNA-binding motif protein 17|splicing factor 45kDa', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6088986, end_position_on_the_genomic_accession=6117459, orientation='plus', exon_count=13, OMIM=606935, no_hdr0=''),
  84992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84992, CurrentID=0, Status='live', Symbol='PIGY', Aliases='PIG-Y', description='phosphatidylinositol glycan anchor biosynthesis class Y', other_designations='phosphatidylinositol glycan anchor biosynthesis, class Y|phosphatidylinositol-glycan biosynthesis class Y protein', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88520978, end_position_on_the_genomic_accession=88523801, orientation='minus', exon_count=2, OMIM=610662, no_hdr0=''),
  84993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84993, CurrentID=0, Status='live', Symbol='UBL7', Aliases='BMSC-UbP, TCBA1', description='ubiquitin like 7', other_designations='ubiquitin-like 7 (bone marrow stromal cell-derived)|ubiquitin-like protein SB132', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74445977, end_position_on_the_genomic_accession=74461188, orientation='minus', exon_count=13, OMIM=609748, no_hdr0=''),
  84996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84996, CurrentID=0, Status='live', Symbol='URB1-AS1', Aliases='C21orf119, PRED84', description='URB1 antisense RNA 1 (head to head)', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32393133, end_position_on_the_genomic_accession=32393960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927942, CurrentID=0, Status='live', Symbol='LOC101927942', Aliases='', description='uncharacterized LOC101927942', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149224290, end_position_on_the_genomic_accession=149226353, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927943, CurrentID=0, Status='live', Symbol='LOC101927943', Aliases='', description='uncharacterized LOC101927943', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44676927, end_position_on_the_genomic_accession=44679911, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927944, CurrentID=0, Status='live', Symbol='FAM53B-AS1', Aliases='', description='FAM53B antisense RNA 1', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124704028, end_position_on_the_genomic_accession=124714774, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85001, CurrentID=0, Status='live', Symbol='MGC16275', Aliases='', description='uncharacterized protein MGC16275', other_designations='CTD-2514K5.2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74209980, end_position_on_the_genomic_accession=74213321, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85002, CurrentID=0, Status='live', Symbol='FAM86B1', Aliases='', description='family with sequence similarity 86 member B1', other_designations='family with sequence similarity 86, member B1|protein FAM86B1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12182104, end_position_on_the_genomic_accession=12194149, orientation='minus', exon_count=9, OMIM=616122, no_hdr0=''),
  85003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85003, CurrentID=0, Status='live', Symbol='BFSP2-AS1', Aliases='', description='BFSP2 antisense RNA 1', other_designations='uncharacterized protein MGC2848', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133429286, end_position_on_the_genomic_accession=133456237, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  85004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85004, CurrentID=0, Status='live', Symbol='RERG', Aliases='', description='RAS-like, estrogen-regulated, growth inhibitor', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15107782, end_position_on_the_genomic_accession=15221477, orientation='minus', exon_count=6, OMIM=612664, no_hdr0=''),
  101927949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927949, CurrentID=0, Status='live', Symbol='LOC101927949', Aliases='', description='uncharacterized LOC101927949', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152047296, end_position_on_the_genomic_accession=152049878, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927950, CurrentID=0, Status='live', Symbol='LOC101927950', Aliases='', description='uncharacterized LOC101927950', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5664985, end_position_on_the_genomic_accession=5695272, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  85007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85007, CurrentID=0, Status='live', Symbol='PHYKPL', Aliases='AGXT2L2, PHLU', description='5-phosphohydroxy-L-lysine phospho-lyase', other_designations='5-phosphonooxy-L-lysine phospho-lyase|alanine--glyoxylate aminotransferase 2-like 2|alanine-glyoxylate aminotransferase 2-like 2', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178208474, end_position_on_the_genomic_accession=178232822, orientation='minus', exon_count=16, OMIM=614683, no_hdr0=''),
  101927952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927952, CurrentID=0, Status='live', Symbol='LOC101927952', Aliases='', description='uncharacterized LOC101927952', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89737971, end_position_on_the_genomic_accession=89764523, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  85009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85009, CurrentID=0, Status='live', Symbol='MGC16025', Aliases='', description='uncharacterized LOC85009', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239193331, end_position_on_the_genomic_accession=239195457, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927954, CurrentID=0, Status='live', Symbol='LOC101927954', Aliases='', description='uncharacterized LOC101927954', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93008115, end_position_on_the_genomic_accession=93014944, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927955, CurrentID=0, Status='live', Symbol='LOC101927955', Aliases='', description='uncharacterized LOC101927955', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9704617, end_position_on_the_genomic_accession=9706902, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85012, CurrentID=0, Status='live', Symbol='TCEAL3', Aliases='WEX8', description='transcription elongation factor A (SII)-like 3', other_designations='TCEA-like protein 3|transcription elongation factor S-II protein-like 3', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103607906, end_position_on_the_genomic_accession=103609927, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85013, CurrentID=0, Status='live', Symbol='TMEM128', Aliases='', description='transmembrane protein 128', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4235542, end_position_on_the_genomic_accession=4248242, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  85014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85014, CurrentID=0, Status='live', Symbol='TMEM141', Aliases='', description='transmembrane protein 141', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136791325, end_position_on_the_genomic_accession=136793317, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85015, CurrentID=0, Status='live', Symbol='USP45', Aliases='', description='ubiquitin specific peptidase 45', other_designations='deubiquitinating enzyme 45|ubiquitin specific protease 45|ubiquitin thioesterase 45|ubiquitin thiolesterase 45|ubiquitin-specific-processing protease 45', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99432306, end_position_on_the_genomic_accession=99515777, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  85016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85016, CurrentID=0, Status='live', Symbol='C11orf70', Aliases='', description='chromosome 11 open reading frame 70', other_designations='', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102047411, end_position_on_the_genomic_accession=102084560, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927961, CurrentID=0, Status='live', Symbol='LOC101927961', Aliases='', description='uncharacterized LOC101927961', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45168543, end_position_on_the_genomic_accession=45181394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  85019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85019, CurrentID=0, Status='live', Symbol='TMEM241', Aliases='C18orf45, hVVT', description='transmembrane protein 241', other_designations='putative vertebrate vrg4-like nucleotide-sugar transporter truncated variant2|putative vertebrate vrg4-like nucleotide-sugar transporter variant1|transmembrane protein C18orf45', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23295781, end_position_on_the_genomic_accession=23437969, orientation='minus', exon_count=23, OMIM=615430, no_hdr0=''),
  101927964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927964, CurrentID=0, Status='live', Symbol='LOC101927964', Aliases='', description='uncharacterized LOC101927964', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4051726, end_position_on_the_genomic_accession=4089013, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85021, CurrentID=0, Status='live', Symbol='REPS1', Aliases='RALBP1', description='RALBP1 associated Eps domain containing 1', other_designations='ralBP1-interacting protein 1', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138904015, end_position_on_the_genomic_accession=138988282, orientation='minus', exon_count=21, OMIM=614825, no_hdr0=''),
  101927966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927966, CurrentID=0, Status='live', Symbol='LOC101927966', Aliases='', description='40S ribosomal protein S24-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8319327, end_position_on_the_genomic_accession=8369597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927967, CurrentID=0, Status='live', Symbol='LOC101927967', Aliases='', description='uncharacterized LOC101927967', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77743696, end_position_on_the_genomic_accession=78290731, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  347168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347168, CurrentID=0, Status='live', Symbol='OR1J1', Aliases='OR9-18, hg32', description='olfactory receptor family 1 subfamily J member 1', other_designations='olfactory receptor OR9-18|olfactory receptor, family 1, subfamily J, member 1', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122476958, end_position_on_the_genomic_accession=122477926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  347169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347169, CurrentID=0, Status='live', Symbol='OR1B1', Aliases='OR9-26, OR9-B', description='olfactory receptor family 1 subfamily B member 1 (gene/pseudogene)', other_designations='olfactory receptor 9-B|olfactory receptor OR9-26|olfactory receptor, family 1, subfamily B, member 1', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122628579, end_position_on_the_genomic_accession=122629535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85026, CurrentID=0, Status='live', Symbol='ARRDC1-AS1', Aliases='C9orf37', description='ARRDC1 antisense RNA 1', other_designations='AD038', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137615332, end_position_on_the_genomic_accession=137618895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  85027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85027, CurrentID=0, Status='live', Symbol='SMIM3', Aliases='C5orf62, MST150, NID67', description='small integral membrane protein 3', other_designations='NGF-induced differentiation clone 67 protein|small membrane protein NID67', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150777946, end_position_on_the_genomic_accession=150796736, orientation='plus', exon_count=2, OMIM=608324, no_hdr0=''),
  85028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85028, CurrentID=0, Status='live', Symbol='SNHG12', Aliases='ASLNC04080, C1orf79, LINC00100, NCRNA00100', description='small nucleolar RNA host gene 12', other_designations='PNAS-123|long intergenic non-protein coding RNA 100|small nucleolar RNA host gene 12 (non-protein coding)', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28578538, end_position_on_the_genomic_accession=28581854, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927973, CurrentID=0, Status='live', Symbol='LOC101927973', Aliases='', description='uncharacterized LOC101927973', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28116094, end_position_on_the_genomic_accession=28228667, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927974, CurrentID=0, Status='live', Symbol='LOC101927974', Aliases='', description='uncharacterized LOC101927974', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107742751, end_position_on_the_genomic_accession=107744051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927975, CurrentID=0, Status='live', Symbol='LOC101927975', Aliases='', description='uncharacterized LOC101927975', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89786313, end_position_on_the_genomic_accession=89790492, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927978, CurrentID=0, Status='live', Symbol='LOC101927978', Aliases='', description='uncharacterized LOC101927978', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72747101, end_position_on_the_genomic_accession=72787246, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927979, CurrentID=0, Status='live', Symbol='LOC101927979', Aliases='', description='uncharacterized LOC101927979', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4987682, end_position_on_the_genomic_accession=4988778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927980, CurrentID=0, Status='live', Symbol='SLC14A2-AS1', Aliases='', description='SLC14A2 antisense RNA 1', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45438183, end_position_on_the_genomic_accession=45507036, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927982, CurrentID=0, Status='live', Symbol='LOC101927982', Aliases='', description='uncharacterized LOC101927982', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=159092714, end_position_on_the_genomic_accession=159096034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927983, CurrentID=0, Status='live', Symbol='EDRF1-AS1', Aliases='', description='EDRF1 antisense RNA 1', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125744727, end_position_on_the_genomic_accession=125752060, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927984, CurrentID=0, Status='live', Symbol='LOC101927984', Aliases='', description='uncharacterized LOC101927984', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91495539, end_position_on_the_genomic_accession=91496243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927987, CurrentID=0, Status='live', Symbol='LOC101927987', Aliases='', description='uncharacterized LOC101927987', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79493716, end_position_on_the_genomic_accession=79500844, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927988, CurrentID=0, Status='live', Symbol='LINC01527', Aliases='lnc-SPRR2D-1', description='long intergenic non-protein coding RNA 1527', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152930040, end_position_on_the_genomic_accession=152949249, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874078, CurrentID=0, Status='live', Symbol='EIF1AX-AS1', Aliases='', description='EIF1AX antisense RNA 1', other_designations='EIF1AX antisense RNA 1 (non-protein coding)', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20139968, end_position_on_the_genomic_accession=20140444, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927990, CurrentID=0, Status='live', Symbol='HRAT13', Aliases='', description='heart tissue-associated transcript 13', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123439604, end_position_on_the_genomic_accession=123471756, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927992, CurrentID=0, Status='live', Symbol='LINC01213', Aliases='', description='long intergenic non-protein coding RNA 1213', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150238519, end_position_on_the_genomic_accession=150240206, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347193, CurrentID=0, Status='live', Symbol='PES1P2', Aliases='', description='pescadillo ribosomal biogenesis factor 1 pseudogene 2', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13986119, end_position_on_the_genomic_accession=13988111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927994, CurrentID=0, Status='live', Symbol='LINC01087', Aliases='', description='long intergenic non-protein coding RNA 1087', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131637025, end_position_on_the_genomic_accession=131649615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874079, CurrentID=0, Status='live', Symbol='EIF2B5-AS1', Aliases='', description='EIF2B5 antisense RNA 1', other_designations='EIF2B5 antisense RNA 1 (non-protein coding)', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927996, CurrentID=0, Status='live', Symbol='LOC101927996', Aliases='', description='spidroin-2-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374218, CurrentID=0, Status='live', Symbol='LOC105374218', Aliases='', description='uncharacterized LOC105374218', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171900511, end_position_on_the_genomic_accession=171938643, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927998, CurrentID=0, Status='live', Symbol='LOC101927998', Aliases='', description='uncharacterized LOC101927998', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73609525, end_position_on_the_genomic_accession=73811427, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927999, CurrentID=0, Status='live', Symbol='LOC101927999', Aliases='', description='putative uncharacterized protein encoded by LINC00174', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90165270, end_position_on_the_genomic_accession=90168604, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928000, CurrentID=0, Status='live', Symbol='LOC101928000', Aliases='', description='uncharacterized LOC101928000', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5111936, end_position_on_the_genomic_accession=5114397, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928001, CurrentID=0, Status='live', Symbol='PRKCA-AS1', Aliases='', description='PRKCA antisense RNA 1', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=66398069, end_position_on_the_genomic_accession=66416854, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928002, CurrentID=0, Status='live', Symbol='LOC101928002', Aliases='', description='uncharacterized LOC101928002', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69713633, end_position_on_the_genomic_accession=69738568, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374219, CurrentID=0, Status='live', Symbol='LOC105374219', Aliases='', description='uncharacterized LOC105374219', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172590774, end_position_on_the_genomic_accession=172595594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928004, CurrentID=0, Status='live', Symbol='LOC101928004', Aliases='', description='uncharacterized LOC101928004', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6694799, end_position_on_the_genomic_accession=6799335, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101928005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928005, CurrentID=0, Status='live', Symbol='LOC101928005', Aliases='', description='uncharacterized LOC101928005', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137858442, end_position_on_the_genomic_accession=137889236, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874081, CurrentID=0, Status='live', Symbol='FGF14-AS1', Aliases='', description='FGF14 antisense RNA 1', other_designations='FGF14 antisense RNA 1 (non-protein coding)', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102367530, end_position_on_the_genomic_accession=102373666, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928008, CurrentID=0, Status='live', Symbol='LOC101928008', Aliases='', description='uncharacterized LOC101928008', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9839142, end_position_on_the_genomic_accession=9929263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928009, CurrentID=0, Status='live', Symbol='LOC101928009', Aliases='', description='uncharacterized LOC101928009', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153174518, end_position_on_the_genomic_accession=153191676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928010, CurrentID=0, Status='live', Symbol='LOC101928010', Aliases='', description='histone deacetylase complex subunit SAP25 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70234237, end_position_on_the_genomic_accession=70235153, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928012, CurrentID=0, Status='live', Symbol='LOC101928012', Aliases='', description='uncharacterized LOC101928012', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112622381, end_position_on_the_genomic_accession=112708072, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100874082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874082, CurrentID=0, Status='live', Symbol='HTR2A-AS1', Aliases='', description='HTR2A antisense RNA 1', other_designations='HTR2A antisense RNA 1 (non-protein coding)', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46852140, end_position_on_the_genomic_accession=46856299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928014, CurrentID=0, Status='live', Symbol='LOC101928014', Aliases='', description='uncharacterized LOC101928014', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93808375, end_position_on_the_genomic_accession=93858327, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928015, CurrentID=0, Status='live', Symbol='LOC101928015', Aliases='', description='uncharacterized LOC101928015', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30518758, end_position_on_the_genomic_accession=30527193, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928016, CurrentID=0, Status='live', Symbol='LOC101928016', Aliases='', description='adenylate kinase isoenzyme 6 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6473636, end_position_on_the_genomic_accession=6476687, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928017, CurrentID=0, Status='live', Symbol='UBXN10-AS1', Aliases='', description='UBXN10 antisense RNA 1', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20184242, end_position_on_the_genomic_accession=20186486, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928020, CurrentID=0, Status='live', Symbol='LOC101928020', Aliases='', description='uncharacterized LOC101928020', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210029922, end_position_on_the_genomic_accession=210064356, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928021, CurrentID=0, Status='live', Symbol='LOC101928021', Aliases='', description='formin-like protein 18', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67243959, end_position_on_the_genomic_accession=67272108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928022, CurrentID=0, Status='live', Symbol='LINC01214', Aliases='', description='long intergenic non-protein coding RNA 1214', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150265407, end_position_on_the_genomic_accession=150323747, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874084, CurrentID=0, Status='live', Symbol='KCNAB1-AS1', Aliases='', description='KCNAB1 antisense RNA 1', other_designations='KCNAB1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156441157, end_position_on_the_genomic_accession=156446905, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928026, CurrentID=0, Status='live', Symbol='LOC101928026', Aliases='', description='uncharacterized LOC101928026', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60470713, end_position_on_the_genomic_accession=60482972, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128362, CurrentID=0, Status='live', Symbol='LOC100128362', Aliases='', description='protein phosphatase 2, catalytic subunit, alpha isozyme pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40938102, end_position_on_the_genomic_accession=40939760, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129293, CurrentID=0, Status='live', Symbol='LOC100129293', Aliases='', description='IgLON family member 5 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64567870, end_position_on_the_genomic_accession=64570437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928030, CurrentID=0, Status='live', Symbol='LOC101928030', Aliases='', description='uncharacterized LOC101928030', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9371782, end_position_on_the_genomic_accession=9381422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874085, CurrentID=0, Status='live', Symbol='NUCB1-AS1', Aliases='', description='NUCB1 antisense RNA 1', other_designations='NUCB1 antisense RNA 1 (non-protein coding)', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48910930, end_position_on_the_genomic_accession=48918891, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928032, CurrentID=0, Status='live', Symbol='LOC101928032', Aliases='', description='uncharacterized LOC101928032', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1337148, end_position_on_the_genomic_accession=1382456, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  101928033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928033, CurrentID=0, Status='live', Symbol='CTD-2270F17.1', Aliases='', description='uncharacterized LOC101928033', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170389493, end_position_on_the_genomic_accession=170422844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928034, CurrentID=0, Status='live', Symbol='LOC101928034', Aliases='', description='uncharacterized LOC101928034', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153545742, end_position_on_the_genomic_accession=153551769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928035, CurrentID=0, Status='live', Symbol='LOC101928035', Aliases='', description='uncharacterized LOC101928035', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74192392, end_position_on_the_genomic_accession=74215521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928036, CurrentID=0, Status='live', Symbol='HRAT17', Aliases='', description='heart tissue-associated transcript 17', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112954635, end_position_on_the_genomic_accession=112995643, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100189224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189224, CurrentID=0, Status='live', Symbol='TRP-TGG2-1', Aliases='TRNAP14', description='transfer RNA-Pro (TGG) 2-1', other_designations='tRNA-Pro-TGG-2-1|transfer RNA proline 14 (anticodon UGG)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76235825, end_position_on_the_genomic_accession=76235896, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928039, CurrentID=0, Status='live', Symbol='LOC101928039', Aliases='', description='uncharacterized LOC101928039', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22194638, end_position_on_the_genomic_accession=22258112, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  347240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347240, CurrentID=0, Status='live', Symbol='KIF24', Aliases='C9orf48, bA571F15.4', description='kinesin family member 24', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34252380, end_position_on_the_genomic_accession=34359914, orientation='minus', exon_count=17, OMIM=613747, no_hdr0=''),
  101928041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928041, CurrentID=0, Status='live', Symbol='LOC101928041', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11816503, end_position_on_the_genomic_accession=11891666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653147, CurrentID=0, Status='live', Symbol='RPL26P30', Aliases='RPL26_10_1113', description='ribosomal protein L26 pseudogene 30', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2335311, end_position_on_the_genomic_accession=2335764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874087, CurrentID=0, Status='live', Symbol='PCDH9-AS4', Aliases='', description='PCDH9 antisense RNA 4', other_designations='PCDH9 antisense RNA 4 (non-protein coding)', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=66990886, end_position_on_the_genomic_accession=67002000, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928044, CurrentID=0, Status='live', Symbol='LOC101928044', Aliases='', description='zinc finger CCHC domain-containing protein 9 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5511843, end_position_on_the_genomic_accession=5513323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928045, CurrentID=0, Status='live', Symbol='LOC101928045', Aliases='', description='uncharacterized LOC101928045', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67675542, end_position_on_the_genomic_accession=67717775, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928046, CurrentID=0, Status='live', Symbol='LOC101928046', Aliases='', description='uncharacterized LOC101928046', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70256152, end_position_on_the_genomic_accession=70315231, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928047, CurrentID=0, Status='live', Symbol='LOC101928047', Aliases='', description='uncharacterized LOC101928047', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6710979, end_position_on_the_genomic_accession=6730254, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928048, CurrentID=0, Status='live', Symbol='LOC101928048', Aliases='', description='uncharacterized LOC101928048', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60469609, end_position_on_the_genomic_accession=60472976, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928049, CurrentID=0, Status='live', Symbol='LOC101928049', Aliases='', description='uncharacterized LOC101928049', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25327100, end_position_on_the_genomic_accession=25349903, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374227, CurrentID=0, Status='live', Symbol='LOC105374227', Aliases='', description='uncharacterized LOC105374227', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175823132, end_position_on_the_genomic_accession=175826959, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  347252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347252, CurrentID=0, Status='live', Symbol='IGFBPL1', Aliases='IGFBP-RP4, IGFBPRP4, bA113O24.1', description='insulin like growth factor binding protein-like 1', other_designations='IGFBP-related protein 10|insulin-like growth factor binding protein related protein 4|insulin-like growth factor binding protein-like 1|insulin-like growth factor-binding-related protein 4', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38406528, end_position_on_the_genomic_accession=38424447, orientation='minus', exon_count=5, OMIM=610413, no_hdr0=''),
  101928053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928053, CurrentID=0, Status='live', Symbol='LOC101928053', Aliases='', description='uncharacterized LOC101928053', other_designations='', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10809297, end_position_on_the_genomic_accession=10822931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928055, CurrentID=0, Status='live', Symbol='LOC101928055', Aliases='', description='uncharacterized LOC101928055', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=1337148, end_position_on_the_genomic_accession=1382456, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100131348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131348, CurrentID=0, Status='live', Symbol='LOC100131348', Aliases='', description='adaptor-related protein complex 2, beta 1 subunit pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103926567, end_position_on_the_genomic_accession=103927824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928058, CurrentID=0, Status='live', Symbol='LOC101928058', Aliases='', description='uncharacterized LOC101928058', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1972618, end_position_on_the_genomic_accession=1973870, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928059, CurrentID=0, Status='live', Symbol='LOC101928059', Aliases='', description='uncharacterized LOC101928059', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153923347, end_position_on_the_genomic_accession=153934964, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874090, CurrentID=0, Status='live', Symbol='SLC6A1-AS1', Aliases='', description='SLC6A1 antisense RNA 1', other_designations='SLC6A1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11006098, end_position_on_the_genomic_accession=11019224, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928062, CurrentID=0, Status='live', Symbol='LINC01481', Aliases='', description='long intergenic non-protein coding RNA 1481', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70222190, end_position_on_the_genomic_accession=70243360, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928063, CurrentID=0, Status='live', Symbol='LOC101928063', Aliases='', description='uncharacterized LOC101928063', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43996896, end_position_on_the_genomic_accession=44002870, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928064, CurrentID=0, Status='live', Symbol='LOC101928064', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105685379, end_position_on_the_genomic_accession=105686323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347265, CurrentID=0, Status='live', Symbol='KRT8P11', Aliases='KRT8L1', description='keratin 8 pseudogene 11', other_designations='KRT8 pseudogene|keratin 8-like 1', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99305084, end_position_on_the_genomic_accession=99306851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928066, CurrentID=0, Status='live', Symbol='LOC101928066', Aliases='', description='uncharacterized LOC101928066', other_designations='', map_location='', chromosome='1, Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874091, CurrentID=0, Status='live', Symbol='TM4SF1-AS1', Aliases='', description='TM4SF1 antisense RNA 1', other_designations='TM4SF1 antisense RNA 1 (non-protein coding)', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149377778, end_position_on_the_genomic_accession=149386583, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928068, CurrentID=0, Status='live', Symbol='LOC101928068', Aliases='', description='uncharacterized LOC101928068', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244965204, end_position_on_the_genomic_accession=244970869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928069, CurrentID=0, Status='live', Symbol='LOC101928069', Aliases='', description='uncharacterized LOC101928069', other_designations='CTD-3074O7.5', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66473490, end_position_on_the_genomic_accession=66480249, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928070, CurrentID=0, Status='live', Symbol='LOC101928070', Aliases='', description='uncharacterized LOC101928070', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2320642, end_position_on_the_genomic_accession=2337998, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928071, CurrentID=0, Status='live', Symbol='LOC101928071', Aliases='', description='M-phase phosphoprotein 6 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22067609, end_position_on_the_genomic_accession=22069174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928072, CurrentID=0, Status='live', Symbol='LOC101928072', Aliases='', description='uncharacterized LOC101928072', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347273, CurrentID=0, Status='live', Symbol='MURC', Aliases='CAVIN4, cavin-4', description='muscle-related coiled-coil protein', other_designations='muscle-restricted coiled-coil protein', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100576867, end_position_on_the_genomic_accession=100588402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347275, CurrentID=0, Status='live', Symbol='BAATP1', Aliases='ACNATP', description='bile acid-CoA: amino acid N-acyltransferase pseudogene 1', other_designations='acid-CoA: amino acid N-acyltransferase, pseudogene|acyl-coenzyme A amino acid N-acyltransferase pseudogene|bile acid CoA: amino acid N-acyltransferase pseudogene 1', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101334092, end_position_on_the_genomic_accession=101339347, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928076, CurrentID=0, Status='live', Symbol='LOC101928076', Aliases='', description='uncharacterized LOC101928076', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7540114, end_position_on_the_genomic_accession=7543038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928077, CurrentID=0, Status='live', Symbol='LOC101928077', Aliases='', description='uncharacterized LOC101928077', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26598937, end_position_on_the_genomic_accession=26623381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928078, CurrentID=0, Status='live', Symbol='LOC101928078', Aliases='', description='histone demethylase UTY-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32454182, end_position_on_the_genomic_accession=32457573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928079, CurrentID=0, Status='live', Symbol='LINC01057', Aliases='XLOC_000303', description='long intergenic non-protein coding RNA 1057', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94657533, end_position_on_the_genomic_accession=94820281, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928080, CurrentID=0, Status='live', Symbol='LOC101928080', Aliases='', description='uncharacterized LOC101928080', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6121798, end_position_on_the_genomic_accession=6141157, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  347281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347281, CurrentID=0, Status='live', Symbol='TOPORSLP1', Aliases='TOPORSLP', description='topoisomerase I binding, arginine/serine-rich like, pseudogene 1', other_designations='topoisomerase I binding, arginine/serine-rich pseudogene|topoisomerase I binding, arginine/serine-rich, E3 ubiquitin protein ligase pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104275322, end_position_on_the_genomic_accession=104276931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928084, CurrentID=0, Status='live', Symbol='LOC101928084', Aliases='', description='uncharacterized LOC101928084', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214084846, end_position_on_the_genomic_accession=214102978, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928085, CurrentID=0, Status='live', Symbol='ERICH6-AS1', Aliases='', description='ERICH6 antisense RNA 1', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150703564, end_position_on_the_genomic_accession=150720146, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928087, CurrentID=0, Status='live', Symbol='LOC101928087', Aliases='', description='uncharacterized LOC101928087', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142638492, end_position_on_the_genomic_accession=142640648, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928088, CurrentID=0, Status='live', Symbol='LOC101928088', Aliases='', description='LYR motif-containing protein 5-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33014813, end_position_on_the_genomic_accession=33015461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928089, CurrentID=0, Status='live', Symbol='LOC101928089', Aliases='', description='uncharacterized LOC101928089', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246108299, end_position_on_the_genomic_accession=246113869, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874095, CurrentID=0, Status='live', Symbol='ADIPOQ-AS1', Aliases='', description='ADIPOQ antisense RNA 1', other_designations='ADIPOQ antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186851886, end_position_on_the_genomic_accession=186856123, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  347292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347292, CurrentID=0, Status='live', Symbol='RPL36P14', Aliases='RPL36_5_1005', description='ribosomal protein L36 pseudogene 14', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108626766, end_position_on_the_genomic_accession=108627199, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928093, CurrentID=0, Status='live', Symbol='LOC101928093', Aliases='', description='uncharacterized LOC101928093', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172755503, end_position_on_the_genomic_accession=172762565, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928094, CurrentID=0, Status='live', Symbol='SMG1P6', Aliases='', description='SMG1 pseudogene 6', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29403884, end_position_on_the_genomic_accession=29443688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928095, CurrentID=0, Status='live', Symbol='LOC101928095', Aliases='', description='40S ribosomal protein S24 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7113126, end_position_on_the_genomic_accession=7227955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928096, CurrentID=0, Status='live', Symbol='LOC101928096', Aliases='', description='uncharacterized LOC101928096', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126128942, end_position_on_the_genomic_accession=126200587, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874096, CurrentID=0, Status='live', Symbol='DOCK9-AS1', Aliases='', description='DOCK9 antisense RNA 1', other_designations='DOCK9 antisense RNA 1 (non-protein coding)', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98832084, end_position_on_the_genomic_accession=98834629, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928098, CurrentID=0, Status='live', Symbol='LOC101928098', Aliases='', description='uncharacterized LOC101928098', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95061550, end_position_on_the_genomic_accession=95067545, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928099, CurrentID=0, Status='live', Symbol='PCAT7', Aliases='PCAN-R2', description='prostate cancer associated transcript 7 (non-protein coding)', other_designations='prostate cancer-associated noncoding RNA 2', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94555045, end_position_on_the_genomic_accession=94568141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928100, CurrentID=0, Status='live', Symbol='LOC101928100', Aliases='', description='uncharacterized LOC101928100', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10363769, end_position_on_the_genomic_accession=10398506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928101, CurrentID=0, Status='live', Symbol='LOC101928101', Aliases='', description='uncharacterized LOC101928101', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154402321, end_position_on_the_genomic_accession=154408834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874097, CurrentID=0, Status='live', Symbol='UBE2Q1-AS1', Aliases='', description='UBE2Q1 antisense RNA 1', other_designations='UBE2Q1 antisense RNA 1 (non-protein coding)', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154553609, end_position_on_the_genomic_accession=154555028, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928104, CurrentID=0, Status='live', Symbol='LINC01482', Aliases='', description='long intergenic non-protein coding RNA 1482', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68628154, end_position_on_the_genomic_accession=68679606, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928105, CurrentID=0, Status='live', Symbol='LOC101928105', Aliases='', description='uncharacterized LOC101928105', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150734471, end_position_on_the_genomic_accession=150738985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928106, CurrentID=0, Status='live', Symbol='LOC101928106', Aliases='', description='uncharacterized LOC101928106', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10465978, end_position_on_the_genomic_accession=10481946, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928107, CurrentID=0, Status='live', Symbol='LOC101928107', Aliases='', description='uncharacterized LOC101928107', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33111704, end_position_on_the_genomic_accession=33123700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928108, CurrentID=0, Status='live', Symbol='LOC101928108', Aliases='', description='lymphocyte antigen 6A-2/6E-1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143080043, end_position_on_the_genomic_accession=143083478, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874098, CurrentID=0, Status='live', Symbol='ARPP21-AS1', Aliases='', description='ARPP21 antisense RNA 1', other_designations='ARPP21 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928110, CurrentID=0, Status='live', Symbol='LOC101928110', Aliases='', description='uncharacterized LOC101928110', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13412426, end_position_on_the_genomic_accession=13427412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928111, CurrentID=0, Status='live', Symbol='LOC101928111', Aliases='', description='uncharacterized LOC101928111', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239401720, end_position_on_the_genomic_accession=239402361, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928112, CurrentID=0, Status='live', Symbol='LOC101928112', Aliases='', description='suppressyn-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246635318, end_position_on_the_genomic_accession=246641742, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928113, CurrentID=0, Status='live', Symbol='ST3GAL5-AS1', Aliases='', description='ST3GAL5 antisense RNA 1 (head to head)', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85889280, end_position_on_the_genomic_accession=85892252, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928114, CurrentID=0, Status='live', Symbol='LOC101928114', Aliases='', description='uncharacterized LOC101928114', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34159165, end_position_on_the_genomic_accession=34436072, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105374238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374238, CurrentID=0, Status='live', Symbol='LOC105374238', Aliases='', description='uncharacterized LOC105374238', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179571523, end_position_on_the_genomic_accession=179576837, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928118, CurrentID=0, Status='live', Symbol='LOC101928118', Aliases='', description='uncharacterized LOC101928118', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95163219, end_position_on_the_genomic_accession=95233982, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928119, CurrentID=0, Status='live', Symbol='LOC101928119', Aliases='', description='uncharacterized LOC101928119', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94920339, end_position_on_the_genomic_accession=94937841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928120, CurrentID=0, Status='live', Symbol='LOC101928120', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154961560, end_position_on_the_genomic_accession=154962625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928122, CurrentID=0, Status='live', Symbol='LINC01483', Aliases='', description='long intergenic non-protein coding RNA 1483', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69593987, end_position_on_the_genomic_accession=69903000, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928123, CurrentID=0, Status='live', Symbol='LOC101928123', Aliases='', description='uncharacterized LOC101928123', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73244018, end_position_on_the_genomic_accession=73246001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928125, CurrentID=0, Status='live', Symbol='LOC101928125', Aliases='', description='uncharacterized LOC101928125', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122672892, end_position_on_the_genomic_accession=122679470, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928126, CurrentID=0, Status='live', Symbol='LOC101928126', Aliases='', description='uncharacterized LOC101928126', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33975186, end_position_on_the_genomic_accession=33978885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874101, CurrentID=0, Status='live', Symbol='HDAC11-AS1', Aliases='', description='HDAC11 antisense RNA 1', other_designations='HDAC11 antisense RNA 1 (non-protein coding)', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13476987, end_position_on_the_genomic_accession=13480053, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928128, CurrentID=0, Status='live', Symbol='LOC101928128', Aliases='', description='uncharacterized LOC101928128', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85210706, end_position_on_the_genomic_accession=85219289, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131360, CurrentID=0, Status='live', Symbol='LOC100131360', Aliases='', description='serine/threonine-protein phosphatase 4 regulatory subunit 2-like', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195996668, end_position_on_the_genomic_accession=195998370, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928131, CurrentID=0, Status='live', Symbol='LOC101928131', Aliases='', description='uncharacterized LOC101928131', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165730779, end_position_on_the_genomic_accession=165743071, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928132, CurrentID=0, Status='live', Symbol='LOC101928132', Aliases='', description='uncharacterized LOC101928132', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13921450, end_position_on_the_genomic_accession=13924863, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874102, CurrentID=0, Status='live', Symbol='HS6ST2-AS1', Aliases='', description='HS6ST2 antisense RNA 1', other_designations='HS6ST2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132667642, end_position_on_the_genomic_accession=132669888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928134, CurrentID=0, Status='live', Symbol='LOC101928134', Aliases='', description='uncharacterized LOC101928134', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33303657, end_position_on_the_genomic_accession=33310659, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928135, CurrentID=0, Status='live', Symbol='LOC101928135', Aliases='', description='uncharacterized LOC101928135', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34875795, end_position_on_the_genomic_accession=35394023, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928136, CurrentID=0, Status='live', Symbol='LINC01484', Aliases='', description='long intergenic non-protein coding RNA 1484', other_designations='CTB-43E15.3', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173707599, end_position_on_the_genomic_accession=173746209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928137, CurrentID=0, Status='live', Symbol='LOC101928137', Aliases='', description='uncharacterized LOC101928137', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=73159190, end_position_on_the_genomic_accession=73208317, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928138, CurrentID=0, Status='live', Symbol='LINC01486', Aliases='', description='long intergenic non-protein coding RNA 1486', other_designations='TCONS_00021407', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109354187, end_position_on_the_genomic_accession=109359488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874103, CurrentID=0, Status='live', Symbol='SRGAP3-AS1', Aliases='', description='SRGAP3 antisense RNA 1', other_designations='SRGAP3 antisense RNA 1 (non-protein coding)', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928140, CurrentID=0, Status='live', Symbol='LOC101928140', Aliases='', description='uncharacterized LOC101928140', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128027865, end_position_on_the_genomic_accession=128086293, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928141, CurrentID=0, Status='live', Symbol='LINC01028', Aliases='TCONS_00025127', description='long intergenic non-protein coding RNA 1028', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70051271, end_position_on_the_genomic_accession=70068095, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928142, CurrentID=0, Status='live', Symbol='AADACL2-AS1', Aliases='', description='AADACL2 antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151751179, end_position_on_the_genomic_accession=151928175, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928143, CurrentID=0, Status='live', Symbol='LOC101928143', Aliases='', description='uncharacterized LOC101928143', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73458991, end_position_on_the_genomic_accession=73463642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  347344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347344, CurrentID=0, Status='live', Symbol='ZNF81', Aliases='HFZ20, MRX45, dJ54B20.6', description='zinc finger protein 81', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47836871, end_position_on_the_genomic_accession=47925624, orientation='plus', exon_count=8, OMIM=314998, no_hdr0=''),
  341880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341880, CurrentID=0, Status='live', Symbol='SLC35F4', Aliases='C14orf36, c14_5373', description='solute carrier family 35 member F4', other_designations='solute carrier family 35, member F4', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57563922, end_position_on_the_genomic_accession=57982274, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  101928146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928146, CurrentID=0, Status='live', Symbol='LOC101928146', Aliases='', description='uncharacterized LOC101928146', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10508663, end_position_on_the_genomic_accession=10521255, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928147, CurrentID=0, Status='live', Symbol='C21orf140', Aliases='', description='chromosome 21 open reading frame 140', other_designations='uncharacterized protein ENSP00000386791', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34400112, end_position_on_the_genomic_accession=34401072, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928149, CurrentID=0, Status='live', Symbol='LOC101928149', Aliases='', description='uncharacterized LOC101928149', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17514391, end_position_on_the_genomic_accession=17514849, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928150, CurrentID=0, Status='live', Symbol='LOC101928150', Aliases='', description='uncharacterized LOC101928150', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6618716, end_position_on_the_genomic_accession=6639781, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928151, CurrentID=0, Status='live', Symbol='LINC01179', Aliases='', description='long intergenic non-protein coding RNA 1179', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165684639, end_position_on_the_genomic_accession=165762778, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928152, CurrentID=0, Status='live', Symbol='LOC101928152', Aliases='', description='uncharacterized LOC101928152', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87470096, end_position_on_the_genomic_accession=87486429, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928153, CurrentID=0, Status='live', Symbol='ITGA9-AS1', Aliases='', description='ITGA9 antisense RNA 1', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37753689, end_position_on_the_genomic_accession=37861780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928154, CurrentID=0, Status='live', Symbol='LINC01485', Aliases='', description='long intergenic non-protein coding RNA 1485', other_designations='CTB-43E15.4', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173786790, end_position_on_the_genomic_accession=173790942, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928156, CurrentID=0, Status='live', Symbol='LOC101928156', Aliases='', description='uncharacterized LOC101928156', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216519182, end_position_on_the_genomic_accession=216529579, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874106, CurrentID=0, Status='live', Symbol='MTUS2-AS2', Aliases='', description='MTUS2 antisense RNA 2', other_designations='MTUS2 antisense RNA 2 (non-protein coding)', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928158, CurrentID=0, Status='live', Symbol='LAMA5-AS1', Aliases='', description='LAMA5 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62353010, end_position_on_the_genomic_accession=62356480, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347359, CurrentID=0, Status='live', Symbol='BMP2KL', Aliases='', description='BMP2 inducible kinase-like, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74185925, end_position_on_the_genomic_accession=74187308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928160, CurrentID=0, Status='live', Symbol='LOC101928160', Aliases='', description='uncharacterized LOC101928160', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143707017, end_position_on_the_genomic_accession=143713898, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928161, CurrentID=0, Status='live', Symbol='LOC101928161', Aliases='', description='uncharacterized LOC101928161', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=133266195, end_position_on_the_genomic_accession=133284758, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928162, CurrentID=0, Status='live', Symbol='LOC101928162', Aliases='', description='uncharacterized LOC101928162', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10750234, end_position_on_the_genomic_accession=10777451, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  347363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347363, CurrentID=0, Status='live', Symbol='KIF4CP', Aliases='KIF4P1', description='kinesin family member 4C, pseudogene', other_designations='chromosome-associated kinesin KIF4A pseudogene|kinesin family member 4C pseudogene', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=79323299, end_position_on_the_genomic_accession=79324785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347365, CurrentID=0, Status='live', Symbol='ITIH6', Aliases='ITIH5L, UNQ6369, dJ14O9.1', description='inter-alpha-trypsin inhibitor heavy chain family member 6', other_designations='ITI-like protein|inter-alpha inhibitor H5-like protein|inter-alpha-trypsin inhibitor heavy chain H5-like protein|inter-alpha-trypsin inhibitor heavy chain family, member 6', map_location='Xp11.22-p11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54748899, end_position_on_the_genomic_accession=54798240, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101928166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928166, CurrentID=0, Status='live', Symbol='LOC101928166', Aliases='', description='uncharacterized LOC101928166', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152157473, end_position_on_the_genomic_accession=152205440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928167, CurrentID=0, Status='live', Symbol='LOC101928167', Aliases='', description='uncharacterized LOC101928167', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=53579623, end_position_on_the_genomic_accession=53581059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928168, CurrentID=0, Status='live', Symbol='LOC101928168', Aliases='', description='uncharacterized LOC101928168', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29125033, end_position_on_the_genomic_accession=29128172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874108, CurrentID=0, Status='live', Symbol='A2ML1-AS1', Aliases='', description='A2ML1 antisense RNA 1', other_designations='A2ML1 antisense RNA 1 (non-protein coding)', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928173, CurrentID=0, Status='live', Symbol='LOC101928173', Aliases='', description='uncharacterized LOC101928173', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87439523, end_position_on_the_genomic_accession=87465886, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928174, CurrentID=0, Status='live', Symbol='LOC101928174', Aliases='', description='uncharacterized LOC101928174', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34755084, end_position_on_the_genomic_accession=34812923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874109, CurrentID=0, Status='live', Symbol='CNOT10-AS1', Aliases='', description='CNOT10 antisense RNA 1', other_designations='CNOT10 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928176, CurrentID=0, Status='live', Symbol='LINC01411', Aliases='', description='long intergenic non-protein coding RNA 1411', other_designations='RP11-267A15.1', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174336354, end_position_on_the_genomic_accession=174527139, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928177, CurrentID=0, Status='live', Symbol='LOC101928177', Aliases='', description='uncharacterized LOC101928177', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156614742, end_position_on_the_genomic_accession=156621928, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85235, CurrentID=0, Status='live', Symbol='HIST1H2AH', Aliases='H2A/S, H2AFALii, H2AH, dJ86C11.1', description='histone cluster 1, H2ah', other_designations='H2A histone family member|histone H2A/s', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27147082, end_position_on_the_genomic_accession=27147562, orientation='plus', exon_count=1, OMIM=615013, no_hdr0=''),
  85236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85236, CurrentID=0, Status='live', Symbol='HIST1H2BK', Aliases='H2B/S, H2BFAiii, H2BFT, H2BK', description='histone cluster 1, H2bk', other_designations='H2B K|H2B histone family, member T|HIRA-interacting protein 1|histone 1, H2bk|histone family member', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27138293, end_position_on_the_genomic_accession=27146858, orientation='minus', exon_count=2, OMIM=615045, no_hdr0=''),
  100874110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874110, CurrentID=0, Status='live', Symbol='GLYCTK-AS1', Aliases='', description='GLYCTK antisense RNA 1', other_designations='GLYCTK antisense RNA 1 (non-protein coding)', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52288580, end_position_on_the_genomic_accession=52289733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928182, CurrentID=0, Status='live', Symbol='LOC101928182', Aliases='', description='uncharacterized LOC101928182', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34419638, end_position_on_the_genomic_accession=34425353, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131369, CurrentID=0, Status='live', Symbol='DUX4L27', Aliases='', description='double homeobox 4 like 27', other_designations='double homeobox 4 like 4 pseudogene|double homeobox, 4-like pseudogene', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34208423, end_position_on_the_genomic_accession=34209801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928185, CurrentID=0, Status='live', Symbol='LOC101928185', Aliases='', description='uncharacterized LOC101928185', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132912428, end_position_on_the_genomic_accession=132931704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928188, CurrentID=0, Status='live', Symbol='LOC101928188', Aliases='', description='uncharacterized LOC101928188', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29212568, end_position_on_the_genomic_accession=29221006, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374250, CurrentID=0, Status='live', Symbol='LOC105374250', Aliases='', description='uncharacterized LOC105374250', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184395351, end_position_on_the_genomic_accession=184451071, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928190, CurrentID=0, Status='live', Symbol='LINC01487', Aliases='', description='long intergenic non-protein coding RNA 1487', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155240945, end_position_on_the_genomic_accession=155243124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928191, CurrentID=0, Status='live', Symbol='LOC101928191', Aliases='', description='uncharacterized LOC101928191', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10932317, end_position_on_the_genomic_accession=10954089, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874112, CurrentID=0, Status='live', Symbol='CCDC39-AS1', Aliases='', description='CCDC39 antisense RNA 1', other_designations='CCDC39 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928195, CurrentID=0, Status='live', Symbol='LOC101928195', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65820966, end_position_on_the_genomic_accession=65878507, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101928196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928196, CurrentID=0, Status='live', Symbol='LOC101928196', Aliases='', description='uncharacterized LOC101928196', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19398653, end_position_on_the_genomic_accession=19468911, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874113, CurrentID=0, Status='live', Symbol='YEATS2-AS1', Aliases='', description='YEATS2 antisense RNA 1', other_designations='YEATS2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928200, CurrentID=0, Status='live', Symbol='MRGPRF-AS1', Aliases='', description='MRGPRF antisense RNA 1', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69012354, end_position_on_the_genomic_accession=69018447, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928201, CurrentID=0, Status='live', Symbol='LOC101928201', Aliases='', description='uncharacterized LOC101928201', other_designations='', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=4627200, end_position_on_the_genomic_accession=4633572, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  347402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347402, CurrentID=0, Status='live', Symbol='NLRP3P', Aliases='NOD13', description='NLR family, pyrin domain containing 3 pseudogene', other_designations='NOD13 pseudogene', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107973894, end_position_on_the_genomic_accession=107984991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928203, CurrentID=0, Status='live', Symbol='LOC101928203', Aliases='', description='uncharacterized LOC101928203', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76634998, end_position_on_the_genomic_accession=76658478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347404, CurrentID=0, Status='live', Symbol='LANCL3', Aliases='', description='LanC lantibiotic synthetase component C-like 3 (bacterial)', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37571569, end_position_on_the_genomic_accession=37677499, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100874114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874114, CurrentID=0, Status='live', Symbol='CCDC13-AS1', Aliases='', description='CCDC13 antisense RNA 1', other_designations='CCDC13 antisense RNA 1 (non-protein coding)', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42732575, end_position_on_the_genomic_accession=42745158, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928207, CurrentID=0, Status='live', Symbol='LOC101928207', Aliases='', description='uncharacterized LOC101928207', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77777348, end_position_on_the_genomic_accession=77783581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928208, CurrentID=0, Status='live', Symbol='LOC101928208', Aliases='', description='probable C-mannosyltransferase DPY19L2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29720497, end_position_on_the_genomic_accession=29742604, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  347411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347411, CurrentID=0, Status='live', Symbol='MPC1L', Aliases='', description='mitochondrial pyruvate carrier 1-like', other_designations='brain protein 44-like protein 2', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40617505, end_position_on_the_genomic_accession=40624139, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928212, CurrentID=0, Status='live', Symbol='LOC101928212', Aliases='', description='uncharacterized LOC101928212', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42571425, end_position_on_the_genomic_accession=42573810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374254, CurrentID=0, Status='live', Symbol='LOC105374254', Aliases='', description='uncharacterized LOC105374254', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185499140, end_position_on_the_genomic_accession=185504493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928214, CurrentID=0, Status='live', Symbol='LOC101928214', Aliases='', description='uncharacterized LOC101928214', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87659371, end_position_on_the_genomic_accession=87686729, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101928215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928215, CurrentID=0, Status='live', Symbol='LOC101928215', Aliases='', description='uncharacterized LOC101928215', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29104510, end_position_on_the_genomic_accession=29109208, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928216, CurrentID=0, Status='live', Symbol='LOC101928216', Aliases='', description='uncharacterized LOC101928216', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26006534, end_position_on_the_genomic_accession=26036470, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100874116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874116, CurrentID=0, Status='live', Symbol='EGFLAM-AS3', Aliases='tbp1', description='EGFLAM antisense RNA 3', other_designations='EGFLAM antisense RNA 3 (non-protein coding)', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928219, CurrentID=0, Status='live', Symbol='LOC101928219', Aliases='', description='uncharacterized LOC101928219', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95625450, end_position_on_the_genomic_accession=95815917, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928222, CurrentID=0, Status='live', Symbol='LOC101928222', Aliases='', description='uncharacterized LOC101928222', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19990204, end_position_on_the_genomic_accession=20004806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928223, CurrentID=0, Status='live', Symbol='LINC01612', Aliases='', description='long intergenic non-protein coding RNA 1612', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170226591, end_position_on_the_genomic_accession=170283723, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85280, CurrentID=0, Status='live', Symbol='KRTAP9-4', Aliases='KAP9.4, KRTAP9.4', description='keratin associated protein 9-4', other_designations='keratin-associated protein 9.4|ultrahigh sulfur keratin-associated protein 9.4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41249687, end_position_on_the_genomic_accession=41250653, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928225, CurrentID=0, Status='live', Symbol='LOC101928225', Aliases='', description='uncharacterized LOC101928225', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115362930, end_position_on_the_genomic_accession=115364745, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928226, CurrentID=0, Status='live', Symbol='LOC101928226', Aliases='', description='uncharacterized LOC101928226', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248697062, end_position_on_the_genomic_accession=248699110, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928227, CurrentID=0, Status='live', Symbol='LOC101928227', Aliases='', description='uncharacterized LOC101928227', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38040393, end_position_on_the_genomic_accession=38062340, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  400011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400011, CurrentID=0, Status='live', Symbol='SUPT16HP1', Aliases='SUPT16HP, bcm670', description='SPT16 homolog, facilitates chromatin remodeling subunit pseudogene 1', other_designations='suppressor of Ty 16 homolog (S. cerevisiae) pseudogene 1', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16398875, end_position_on_the_genomic_accession=16402509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  85285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85285, CurrentID=0, Status='live', Symbol='KRTAP4-1', Aliases='KAP4.1, KAP4.100, KRTAP4.10, KRTAP4-1', description='keratin associated protein 4-1', other_designations='keratin associated protein 4-10|keratin-associated protein 4.1|ultrahigh sulfur keratin-associated protein 4.10', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41184100, end_position_on_the_genomic_accession=41184895, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928230, CurrentID=0, Status='live', Symbol='LOC101928230', Aliases='', description='uncharacterized LOC101928230', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79600903, end_position_on_the_genomic_accession=79612426, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928231, CurrentID=0, Status='live', Symbol='LOC101928231', Aliases='', description='uncharacterized LOC101928231', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134429030, end_position_on_the_genomic_accession=134478897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928232, CurrentID=0, Status='live', Symbol='LINC01280', Aliases='', description='long intergenic non-protein coding RNA 1280', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216589993, end_position_on_the_genomic_accession=216606938, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  85289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85289, CurrentID=0, Status='live', Symbol='KRTAP4-5', Aliases='KAP4.5, KRTAP4.5', description='keratin associated protein 4-5', other_designations='keratin-associated protein 4.5|ultrahigh sulfur keratin-associated protein 4.5', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41148924, end_position_on_the_genomic_accession=41149802, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85290, CurrentID=0, Status='live', Symbol='KRTAP4-3', Aliases='KAP4.3', description='keratin associated protein 4-3', other_designations='keratin-associated protein 4.3|ultrahigh sulfur keratin-associated protein 4.3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41167231, end_position_on_the_genomic_accession=41168172, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85291, CurrentID=0, Status='live', Symbol='KRTAP4-2', Aliases='KAP4.2, KRTAP4.2', description='keratin associated protein 4-2', other_designations='keratin-associated protein 4.2|ultrahigh sulfur keratin-associated protein 4.2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41177446, end_position_on_the_genomic_accession=41178208, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928236, CurrentID=0, Status='live', Symbol='LOC101928236', Aliases='', description='uncharacterized LOC101928236', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157163566, end_position_on_the_genomic_accession=157381196, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  85293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85293, CurrentID=0, Status='live', Symbol='KRTAP3-3', Aliases='KAP3.3, KRTAP3.3', description='keratin associated protein 3-3', other_designations='high sulfur keratin-associated protein 3.3|keratin-associated protein 3.3', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40993430, end_position_on_the_genomic_accession=40994133, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85294, CurrentID=0, Status='live', Symbol='KRTAP2-4', Aliases='KAP2.1B, KAP2.4, KRTAP2.4', description='keratin associated protein 2-4', other_designations='high sulfur keratin-associated protein 2.4|keratin associated protein 2.1B|keratin-associated protein 2.4', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41065116, end_position_on_the_genomic_accession=41065879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480204, CurrentID=0, Status='live', Symbol='RNU2-17P', Aliases='', description='RNA, U2 small nuclear 17, pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347442, CurrentID=0, Status='live', Symbol='DCAF8L2', Aliases='WDR42C', description='DDB1 and CUL4 associated factor 8-like 2', other_designations='WD repeat domain 42C|WD repeat-containing protein 42C', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27746809, end_position_on_the_genomic_accession=27748821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928243, CurrentID=0, Status='live', Symbol='DARS-AS1', Aliases='', description='DARS antisense RNA 1', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135985176, end_position_on_the_genomic_accession=136007542, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85300, CurrentID=0, Status='live', Symbol='ATCAY', Aliases='BNIP-H, CLAC', description='ataxia, cerebellar, Cayman type', other_designations='BNIP-2-homolgy|BNIP-2-homology|Cayman ataxia|ataxia cayman type protein|ataxia cerebellar Cayman type', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3880620, end_position_on_the_genomic_accession=3928082, orientation='plus', exon_count=13, OMIM=608179, no_hdr0=''),
  85301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85301, CurrentID=0, Status='live', Symbol='COL27A1', Aliases='STLS', description='collagen, type XXVII, alpha 1', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114130331, end_position_on_the_genomic_accession=114312516, orientation='plus', exon_count=67, OMIM=608461, no_hdr0=''),
  85302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85302, CurrentID=0, Status='live', Symbol='FBF1', Aliases='Alb, FBF-1', description='Fas (TNFRSF6) binding factor 1', other_designations='Fas binding protein 1|albatross|protein albatross', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75910537, end_position_on_the_genomic_accession=75941038, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  85303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85303, CurrentID=0, Status='live', Symbol='HIST1H2APS2', Aliases='H2A/T, H2AFTP, dJ139G21.2', description='histone cluster 1, H2a, pseudogene 2', other_designations='H2A histone family, member T, pseudogene|histone 1, H2a, pseudogene 2|histone family member pseudogene', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25881926, end_position_on_the_genomic_accession=25882416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928248, CurrentID=0, Status='live', Symbol='LINC01229', Aliases='', description='long intergenic non-protein coding RNA 1229', other_designations='TCONS_00024480', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79770471, end_position_on_the_genomic_accession=79801343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928249, CurrentID=0, Status='live', Symbol='LOC101928249', Aliases='', description='uncharacterized LOC101928249', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134498114, end_position_on_the_genomic_accession=134524516, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101928250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928250, CurrentID=0, Status='live', Symbol='LOC101928250', Aliases='', description='uncharacterized LOC101928250', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216712380, end_position_on_the_genomic_accession=216810997, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928251, CurrentID=0, Status='live', Symbol='LOC101928251', Aliases='', description='uncharacterized LOC101928251', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73513636, end_position_on_the_genomic_accession=73518770, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874122, CurrentID=0, Status='live', Symbol='FRMPD3-AS1', Aliases='', description='FRMPD3 antisense RNA 1', other_designations='FRMPD3 antisense RNA 1 (non-protein coding)', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107512983, end_position_on_the_genomic_accession=107545821, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  347454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347454, CurrentID=0, Status='live', Symbol='SOWAHD', Aliases='ANKRD58', description='sosondowah ankyrin repeat domain family member D', other_designations='ankyrin repeat domain 58|ankyrin repeat domain-containing protein 58|protein sosondowah homolog D', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119758613, end_position_on_the_genomic_accession=119760202, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374261, CurrentID=0, Status='live', Symbol='LOC105374261', Aliases='', description='uncharacterized LOC105374261', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187447041, end_position_on_the_genomic_accession=187449953, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928256, CurrentID=0, Status='live', Symbol='UBE2Q2P10', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 10', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 10', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86660366, end_position_on_the_genomic_accession=86663665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  85313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85313, CurrentID=0, Status='live', Symbol='PPIL4', Aliases='HDCME13P', description='peptidylprolyl isomerase like 4', other_designations='PPIase|cyclophilin-like protein PPIL4|cyclophilin-type peptidyl-prolyl cis-trans isomerase|peptidylprolyl isomerase (cyclophilin)-like 4|rotamase PPIL4|serologically defined breast cancer antigen NY-BR-18', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149504495, end_position_on_the_genomic_accession=149546102, orientation='minus', exon_count=13, OMIM=607609, no_hdr0=''),
  85314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85314, CurrentID=0, Status='live', Symbol='ERVE-1', Aliases='ERVE1, HERV-E1, HERVE1', description='endogenous retrovirus group E member 1', other_designations='TCONS_00025599|endogenous retroviral sequence E, 1|endogenous retrovirus group E, member 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606601, no_hdr0=''),
  85315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85315, CurrentID=0, Status='live', Symbol='PAQR8', Aliases='C6orf33, LMPB1, MPRB', description='progestin and adipoQ receptor family member VIII', other_designations='lysosomal membrane protein in brain 1|lysosomal membrane protein in brain-1|mPR beta|progestin and adipoQ receptor family member 8', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52362128, end_position_on_the_genomic_accession=52407777, orientation='plus', exon_count=2, OMIM=607780, no_hdr0=''),
  85316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85316, CurrentID=0, Status='live', Symbol='BAGE5', Aliases='CT2.5', description='B melanoma antigen family member 5', other_designations='B melanoma antigen family, member 5|cancer/testis antigen 2.5|cancer/testis antigen family 2, member 5', map_location='13cen', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85317, CurrentID=0, Status='live', Symbol='BAGE4', Aliases='CT2.4, MLL3P', description='B melanoma antigen family member 4', other_designations='B melanoma antigen family, member 4|MLL3 pseudogene|cancer/testis antigen 2.4|cancer/testis antigen family 2, member 4', map_location='21p11.1', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85318, CurrentID=0, Status='live', Symbol='BAGE3', Aliases='CT2.3', description='B melanoma antigen family member 3', other_designations='B melanoma antigen family, member 3|cancer/testis antigen 2.3|cancer/testis antigen family 2, member 3', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=11020842, end_position_on_the_genomic_accession=11058323, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  85319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85319, CurrentID=0, Status='live', Symbol='BAGE2', Aliases='CT2.2', description='B melanoma antigen family member 2', other_designations='B melanoma antigen family, member 2|cancer/testis antigen 2.2|cancer/testis antigen family 2, member 2', map_location='21p', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10413532, end_position_on_the_genomic_accession=10491615, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  85320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85320, CurrentID=0, Status='live', Symbol='ABCC11', Aliases='EWWD, MRP8, WW', description='ATP binding cassette subfamily C member 11', other_designations='ATP-binding cassette protein C11|ATP-binding cassette transporter MRP8|ATP-binding cassette transporter sub-family C member 11|ATP-binding cassette, sub-family C (CFTR/MRP), member 11|multi-resistance protein 8|multidrug resistance-associated protein 8', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48166910, end_position_on_the_genomic_accession=48235177, orientation='minus', exon_count=33, OMIM=607040, no_hdr0=''),
  101928266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928266, CurrentID=0, Status='live', Symbol='LOC101928266', Aliases='', description='uncharacterized LOC101928266', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9171068, end_position_on_the_genomic_accession=9179118, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  347468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347468, CurrentID=0, Status='live', Symbol='OR13H1', Aliases='ORX1', description='olfactory receptor family 13 subfamily H member 1', other_designations='olfactory receptor ORX-1|olfactory receptor ORX1|olfactory receptor, family 13, subfamily H, member 1', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131544074, end_position_on_the_genomic_accession=131545000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928269, CurrentID=0, Status='live', Symbol='LOC101928269', Aliases='', description='uncharacterized LOC101928269', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35954660, end_position_on_the_genomic_accession=36004667, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928270, CurrentID=0, Status='live', Symbol='LOC101928270', Aliases='', description='uncharacterized LOC101928270', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99472420, end_position_on_the_genomic_accession=99487804, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928271, CurrentID=0, Status='live', Symbol='LOC101928271', Aliases='', description='uncharacterized LOC101928271', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20859771, end_position_on_the_genomic_accession=20861661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928272, CurrentID=0, Status='live', Symbol='LOC101928272', Aliases='', description='uncharacterized LOC101928272', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9197366, end_position_on_the_genomic_accession=9295593, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  85329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85329, CurrentID=0, Status='live', Symbol='LGALS12', Aliases='GAL12, GRIP1', description='lectin, galactoside-binding, soluble, 12', other_designations='galectin 12|galectin-related inhibitor of proliferation', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63506052, end_position_on_the_genomic_accession=63516774, orientation='plus', exon_count=11, OMIM=606096, no_hdr0=''),
  101928274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928274, CurrentID=0, Status='live', Symbol='LOC101928274', Aliases='', description='uncharacterized LOC101928274', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118117622, end_position_on_the_genomic_accession=118135946, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  347475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347475, CurrentID=0, Status='live', Symbol='CCDC160', Aliases='', description='coiled-coil domain containing 160', other_designations='UPF0625 coiled-coil domain-containing protein ENSP00000359845', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134237047, end_position_on_the_genomic_accession=134246207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928276, CurrentID=0, Status='live', Symbol='LINC01227', Aliases='', description='long intergenic non-protein coding RNA 1227', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80567103, end_position_on_the_genomic_accession=80572808, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874126, CurrentID=0, Status='live', Symbol='LINC00331', Aliases='NCRNA00331', description='long intergenic non-protein coding RNA 331', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78787319, end_position_on_the_genomic_accession=78840050, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928278, CurrentID=0, Status='live', Symbol='LOC101928278', Aliases='', description='uncharacterized LOC101928278', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216819340, end_position_on_the_genomic_accession=217250031, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  101928279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928279, CurrentID=0, Status='live', Symbol='LOC101928279', Aliases='', description='uncharacterized LOC101928279', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4761760, end_position_on_the_genomic_accession=4787374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928280, CurrentID=0, Status='live', Symbol='LOC101928280', Aliases='', description='uncharacterized LOC101928280', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68226970, end_position_on_the_genomic_accession=68329899, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874127, CurrentID=0, Status='live', Symbol='LINC00332', Aliases='NCRNA00332', description='long intergenic non-protein coding RNA 332', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40181809, end_position_on_the_genomic_accession=40189030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85340, CurrentID=0, Status='live', Symbol='KRT90P', Aliases='HBA, KRT124P, KRTHBP1', description='keratin 90 pseudogene', other_designations='keratin 124 pseudogene|keratin 8 pseudogene|keratin, hair, basic pseudogene 1|psihHbA hair keratin pseudogene', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52413861, end_position_on_the_genomic_accession=52419240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928286, CurrentID=0, Status='live', Symbol='LOC101928286', Aliases='', description='uncharacterized LOC101928286', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103766869, end_position_on_the_genomic_accession=103797864, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  347487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347487, CurrentID=0, Status='live', Symbol='CXorf66', Aliases='SGPX', description='chromosome X open reading frame 66', other_designations='RP11-35F15.2|secreted glycoprotein, X-linked', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=139955725, end_position_on_the_genomic_accession=139965518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85344, CurrentID=0, Status='live', Symbol='KRT89P', Aliases='HBB, KRT123P, KRTHBP2', description='keratin 89 pseudogene', other_designations='keratin 123 pseudogene|keratin, hair, basic pseudogene 2|psihHbB hair keratin pseudogene', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52340997, end_position_on_the_genomic_accession=52354463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  85345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85345, CurrentID=0, Status='live', Symbol='KRTAP3-4P', Aliases='KAP3A, KRTAP3P1', description='keratin associated protein 3-4, pseudogene', other_designations='keratin associated protein 3 pseudogene 1|keratin-associated protein 3A pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41004520, end_position_on_the_genomic_accession=41004759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928290, CurrentID=0, Status='live', Symbol='PLBD1-AS1', Aliases='', description='PLBD1 antisense RNA 1', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14567732, end_position_on_the_genomic_accession=14619755, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  85347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85347, CurrentID=0, Status='live', Symbol='KRTAP9-11P', Aliases='KAP9A, KRTAP9P1', description='keratin associated protein 9-11, pseudogene', other_designations='keratin associated protein 9 pseudogene 1|keratin-associated protein 9A pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41271311, end_position_on_the_genomic_accession=41271835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  85348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85348, CurrentID=0, Status='live', Symbol='KRT88P', Aliases='HBC, KRT122P, KRTHBP3', description='keratin 88 pseudogene', other_designations='keratin 122 pseudogene|keratin, hair, basic pseudogene 3|psihHbC hair keratin pseudogene', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52263781, end_position_on_the_genomic_accession=52268497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  85349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85349, CurrentID=0, Status='live', Symbol='KRT87P', Aliases='HBD, KRT121P, KRTHBP4', description='keratin 87 pseudogene', other_designations='Keratin-81-like protein|keratin 121 pseudogene|keratin, hair, basic pseudogene 4|psihHbD hair keratin pseudogene', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52252026, end_position_on_the_genomic_accession=52258521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874129, CurrentID=0, Status='live', Symbol='ENOX1-AS1', Aliases='', description='ENOX1 antisense RNA 1', other_designations='ENOX1 antisense RNA 1 (non-protein coding)', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85352, CurrentID=0, Status='live', Symbol='KIAA1644', Aliases='', description='KIAA1644', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44243665, end_position_on_the_genomic_accession=44331695, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374268, CurrentID=0, Status='live', Symbol='LOC105374268', Aliases='', description='uncharacterized LOC105374268', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=188072888, end_position_on_the_genomic_accession=188081047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85354, CurrentID=0, Status='live', Symbol='KRTAP4-16P', Aliases='KAP4A, KRTAP4P1', description='keratin associated protein 4-16, pseudogene', other_designations='keratin associated protein 4 pseudogene 1|keratin-associated protein 4A pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41101402, end_position_on_the_genomic_accession=41102309, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928300, CurrentID=0, Status='live', Symbol='LOC101928300', Aliases='', description='uncharacterized LOC101928300', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120629611, end_position_on_the_genomic_accession=120640501, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874130, CurrentID=0, Status='live', Symbol='ENOX1-AS2', Aliases='', description='ENOX1 antisense RNA 2', other_designations='ENOX1 antisense RNA 2 (non-protein coding)', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43448852, end_position_on_the_genomic_accession=43459488, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85358, CurrentID=0, Status='live', Symbol='SHANK3', Aliases='DEL22q13.3, PROSAP2, PSAP2, SCZD15, SPANK-2', description='SH3 and multiple ankyrin repeat domains 3', other_designations='proline rich synapse associated protein 2|shank postsynaptic density protein', map_location='22q13.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50674642, end_position_on_the_genomic_accession=50733212, orientation='plus', exon_count=22, OMIM=606230, no_hdr0=''),
  85359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85359, CurrentID=0, Status='live', Symbol='DGCR6L', Aliases='DGCR6', description='DiGeorge syndrome critical region gene 6-like', other_designations='DiGeorge syndrome critical region gene 6 like|diGeorge syndrome critical region 6-like protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20314238, end_position_on_the_genomic_accession=20320105, orientation='minus', exon_count=5, OMIM=609459, no_hdr0=''),
  85360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85360, CurrentID=0, Status='live', Symbol='SYDE1', Aliases='7h3', description='synapse defective 1, Rho GTPase, homolog 1 (C. elegans)', other_designations='protein syd-1 homolog 1|synapse defective protein 1 homolog 1', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15107364, end_position_on_the_genomic_accession=15114988, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928306, CurrentID=0, Status='live', Symbol='LOC101928306', Aliases='', description='uncharacterized LOC101928306', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4920808, end_position_on_the_genomic_accession=4988671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85363, CurrentID=0, Status='live', Symbol='TRIM5', Aliases='RNF88alpha, TRIM5', description='tripartite motif containing 5', other_designations='ring finger protein 88|tripartite motif containing 5 transcript variant iota|tripartite motif containing 5 transcript variant kappa|tripartite motif protein TRIM5', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5590982, end_position_on_the_genomic_accession=5685109, orientation='minus', exon_count=14, OMIM=608487, no_hdr0=''),
  85364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85364, CurrentID=0, Status='live', Symbol='ZCCHC3', Aliases='C20orf99', description='zinc finger, CCHC domain containing 3', other_designations='', map_location='20p13-p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=297563, end_position_on_the_genomic_accession=300321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  85365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85365, CurrentID=0, Status='live', Symbol='ALG2', Aliases='CDGIi, CMS14, CMSTA3, NET38, hALPG2', description='ALG2, alpha-1,3/1,6-mannosyltransferase', other_designations='GDP-Man:Man(1)GlcNAc(2)-PP-Dol alpha-1,3-mannosyltransferase|GDP-Man:Man(1)GlcNAc(2)-PP-dolichol mannosyltransferase|GDP-Man:Man(2)GlcNAc(2)-PP-Dol alpha-1,6-mannosyltransferase|alpha-1,3-mannosyltransferase ALG2|asparagine-linked glycosylation 2 homolog (S. cerevisiae, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation 2 homolog (yeast, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation 2, alpha-1,3-mannosyltransferase homolog|asparagine-linked glycosylation protein 2 homolog|homolog of yeast ALG2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99216425, end_position_on_the_genomic_accession=99221964, orientation='minus', exon_count=3, OMIM=607905, no_hdr0=''),
  85366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85366, CurrentID=0, Status='live', Symbol='MYLK2', Aliases='KMLC, MLCK, MLCK2, skMLCK', description='myosin light chain kinase 2', other_designations='myosin light chain kinase 2, skeletal muscle', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31819312, end_position_on_the_genomic_accession=31834697, orientation='plus', exon_count=14, OMIM=606566, no_hdr0=''),
  85367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85367, CurrentID=0, Status='live', Symbol='KIAA1652', Aliases='', description='KIAA1652 protein', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19889773, end_position_on_the_genomic_accession=19892873, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85368, CurrentID=0, Status='live', Symbol='KIAA1654', Aliases='', description='KIAA1654 protein', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85369, CurrentID=0, Status='live', Symbol='STRIP1', Aliases='FAM40A, FAR11A', description='striatin interacting protein 1', other_designations='FAR11 factor arrest 11 homolog A|family with sequence similarity 40, member A|homolog of yeast FAR11 protein 1|protein FAM40A', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110031577, end_position_on_the_genomic_accession=110054641, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  101928314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928314, CurrentID=0, Status='live', Symbol='LOC101928314', Aliases='', description='uncharacterized LOC101928314', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=172629932, end_position_on_the_genomic_accession=172726878, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  85371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85371, CurrentID=0, Status='live', Symbol='KIAA1656', Aliases='', description='KIAA1656 protein', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30368811, end_position_on_the_genomic_accession=30377905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  347516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347516, CurrentID=0, Status='live', Symbol='DGAT2L6', Aliases='DC3', description='diacylglycerol O-acyltransferase 2-like 6', other_designations='diacylglycerol O-acyltransferase candidate 3', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70177483, end_position_on_the_genomic_accession=70205705, orientation='plus', exon_count=8, OMIM=300926, no_hdr0=''),
  347517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347517, CurrentID=0, Status='live', Symbol='RAB41', Aliases='', description='RAB41, member RAS oncogene family', other_designations='RAB41, member RAS homolog family', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70281391, end_position_on_the_genomic_accession=70285800, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928318, CurrentID=0, Status='live', Symbol='LOC101928318', Aliases='', description='nucleolar protein 58-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85375, CurrentID=0, Status='live', Symbol='KIAA1661', Aliases='', description='KIAA1661 protein', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85376, CurrentID=0, Status='live', Symbol='RIMBP3', Aliases='RIM-BP3, RIM-BP3.1, RIM-BP3.A.1, RIMBP3A, RIMBP3', description='RIMS binding protein 3', other_designations='RIMS binding protein 3.1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18606124, end_position_on_the_genomic_accession=18611919, orientation='minus', exon_count=1, OMIM=612699, no_hdr0=''),
  85377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85377, CurrentID=0, Status='live', Symbol='MICALL1', Aliases='MICAL-L1, MIRAB13', description='MICAL like 1', other_designations='MICAL-like 1|molecule interacting with Rab13', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37906148, end_position_on_the_genomic_accession=37942458, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  85378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85378, CurrentID=0, Status='live', Symbol='TUBGCP6', Aliases='GCP-6, GCP6, MCCRP, MCCRP1, MCPHCR', description='tubulin gamma complex associated protein 6', other_designations='tubulin, gamma complex associated protein 6', map_location='22q13.31-q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50217689, end_position_on_the_genomic_accession=50245024, orientation='minus', exon_count=25, OMIM=610053, no_hdr0=''),
  85379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85379, CurrentID=0, Status='live', Symbol='KIAA1671', Aliases='', description='KIAA1671', other_designations='CTA-221G9.5', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24952702, end_position_on_the_genomic_accession=25197448, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  101928324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928324, CurrentID=0, Status='live', Symbol='LOC101928324', Aliases='', description='uncharacterized LOC101928324', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26462756, end_position_on_the_genomic_accession=26467537, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928326, CurrentID=0, Status='live', Symbol='LOC101928326', Aliases='', description='uncharacterized LOC101928326', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135807145, end_position_on_the_genomic_accession=135808473, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  347527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347527, CurrentID=0, Status='live', Symbol='ARSH', Aliases='sulfatase', description='arylsulfatase family member H', other_designations='ASH|arylsulfatase family, member H', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3006613, end_position_on_the_genomic_accession=3033385, orientation='plus', exon_count=9, OMIM=300586, no_hdr0=''),
  100874135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874135, CurrentID=0, Status='live', Symbol='LINC00349', Aliases='', description='long intergenic non-protein coding RNA 349', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85388, CurrentID=0, Status='live', Symbol='SNORD14B', Aliases='RNU14B, U14, U14B', description='small nucleolar RNA, C/D box 14B', other_designations='RNA, U14B small nucleolar|snoRNA', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17075778, end_position_on_the_genomic_accession=17075868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85389, CurrentID=0, Status='live', Symbol='SNORD14C', Aliases='', description='small nucleolar RNA, C/D box 14C', other_designations='RNA, small nucleolar|snoRNA', map_location='11q23.3-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123059335, end_position_on_the_genomic_accession=123059422, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85390, CurrentID=0, Status='live', Symbol='SNORD14D', Aliases='', description='small nucleolar RNA, C/D box 14D', other_designations='RNA, small nucleolar|snoRNA', map_location='11q23.3-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123058909, end_position_on_the_genomic_accession=123058995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85391, CurrentID=0, Status='live', Symbol='SNORD14E', Aliases='', description='small nucleolar RNA, C/D box 14E', other_designations='RNA, small nucleolar|snoRNA', map_location='11q23.3-q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123058077, end_position_on_the_genomic_accession=123058156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928336, CurrentID=0, Status='live', Symbol='LOC101928336', Aliases='', description='uncharacterized LOC101928336', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119291529, end_position_on_the_genomic_accession=119335610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928337, CurrentID=0, Status='live', Symbol='PTCSC2', Aliases='', description='papillary thyroid carcinoma susceptibility candidate 2 (non-protein coding)', other_designations='', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97709920, end_position_on_the_genomic_accession=97853093, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928338, CurrentID=0, Status='live', Symbol='LOC101928338', Aliases='', description='uncharacterized LOC101928338', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30741421, end_position_on_the_genomic_accession=30794447, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85395, CurrentID=0, Status='live', Symbol='FAM207A', Aliases='C21orf70, PRED56', description='family with sequence similarity 207 member A', other_designations='family with sequence similarity 207, member A', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44939998, end_position_on_the_genomic_accession=44976976, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928340, CurrentID=0, Status='live', Symbol='LINC01489', Aliases='', description='long intergenic non-protein coding RNA 1489', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15001833, end_position_on_the_genomic_accession=15006683, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85397, CurrentID=0, Status='live', Symbol='RGS8', Aliases='', description='regulator of G-protein signaling 8', other_designations='regulator of G-protein signalling 8', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182643341, end_position_on_the_genomic_accession=182753307, orientation='minus', exon_count=15, OMIM=607189, no_hdr0=''),
  101928343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928343, CurrentID=0, Status='live', Symbol='LOC101928343', Aliases='', description='uncharacterized LOC101928343', other_designations='CTD-2006K23.1', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74599840, end_position_on_the_genomic_accession=74625910, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  347544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347544, CurrentID=0, Status='live', Symbol='RPL18AP16', Aliases='RPL18A_10_1821', description='ribosomal protein L18a pseudogene 16', other_designations='hCG2040224', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153662127, end_position_on_the_genomic_accession=153662741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928345, CurrentID=0, Status='live', Symbol='PABPC1L2B-AS1', Aliases='', description='PABPC1L2B antisense RNA 1 (head to head)', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72998388, end_position_on_the_genomic_accession=73002856, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928346, CurrentID=0, Status='live', Symbol='LOC101928346', Aliases='', description='uncharacterized LOC101928346', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121581176, end_position_on_the_genomic_accession=121596000, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85403, CurrentID=0, Status='live', Symbol='EAF1', Aliases='', description='ELL associated factor 1', other_designations='ELL (eleven nineteen lysine-rich leukemia gene)-associated factor 1', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15427557, end_position_on_the_genomic_accession=15442613, orientation='plus', exon_count=7, OMIM=608315, no_hdr0=''),
  102578074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102578074, CurrentID=0, Status='live', Symbol='PCAT5', Aliases='LINC01452, TPCAT-10-36067', description='prostate cancer associated transcript 5 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1452', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35778302, end_position_on_the_genomic_accession=35800920, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  85406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85406, CurrentID=0, Status='live', Symbol='DNAJC14', Aliases='DNAJ, DRIP78, HDJ3, LIP6', description='DnaJ heat shock protein family (Hsp40) member C14', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 14|LYST-interacting protein LIP6|dnaJ protein homolog 3|dopamine receptor interacting protein|dopamine receptor-interacting protein of 78 kDa|hDj-3|human DnaJ protein 3', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55820960, end_position_on_the_genomic_accession=55829636, orientation='minus', exon_count=7, OMIM=606092, no_hdr0=''),
  85407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85407, CurrentID=0, Status='live', Symbol='NKD1', Aliases='Naked1', description='naked cuticle homolog 1 (Drosophila)', other_designations='Dvl-binding protein|naked cuticle-1|naked-1', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50548330, end_position_on_the_genomic_accession=50649249, orientation='plus', exon_count=14, OMIM=607851, no_hdr0=''),
  101928352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928352, CurrentID=0, Status='live', Symbol='LOC101928352', Aliases='', description='uncharacterized LOC101928352', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74613185, end_position_on_the_genomic_accession=74614048, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928353, CurrentID=0, Status='live', Symbol='EVADR', Aliases='', description='uncharacterized LOC101928353', other_designations='endogenous retroviral-associated adenocarcinoma RNA', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70394202, end_position_on_the_genomic_accession=70399417, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928354, CurrentID=0, Status='live', Symbol='LOC101928354', Aliases='', description='uncharacterized LOC101928354', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14597513, end_position_on_the_genomic_accession=14717497, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928355, CurrentID=0, Status='live', Symbol='LOC101928355', Aliases='', description='uncharacterized LOC101928355', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85413, CurrentID=0, Status='live', Symbol='SLC22A16', Aliases='CT2, FLIPT2, HEL-S-18, OAT6, OCT6, OKB1, dJ261K5.1', description='solute carrier family 22 (organic cation/carnitine transporter), member 16', other_designations='WUGSC:RG331P03.1|carnitine transporter 2|epididymis secretory protein Li 18|flipt 2|fly-like putative organic ion transporter 2|fly-like putative transporter 2|organic cation transporter 6|organic cation transporter OKB1|organic cation/carnitine transporter 6|solute carrier family 22 (organic cation transporter), member 16', map_location='6q21-q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110424685, end_position_on_the_genomic_accession=110476641, orientation='minus', exon_count=11, OMIM=608276, no_hdr0=''),
  85414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85414, CurrentID=0, Status='live', Symbol='SLC45A3', Aliases='IPCA-2, IPCA-6, IPCA-8, IPCA6, PCANAP2, PCANAP6, PCANAP8, PRST', description='solute carrier family 45 member 3', other_designations='prostate cancer associated protein 2|prostate cancer associated protein 6|prostate cancer associated protein 8|prostate cancer-associated gene 2|prostate cancer-associated gene 6|prostate cancer-associated gene 8|prostate cancer-associated protein 6|prostein|solute carrier family 45, member 3', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205657851, end_position_on_the_genomic_accession=205680502, orientation='minus', exon_count=9, OMIM=605097, no_hdr0=''),
  85415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85415, CurrentID=0, Status='live', Symbol='RHPN2', Aliases='P76RBE, RHOBP', description='rhophilin, Rho GTPase binding protein 2', other_designations='76 kDa RhoB effector protein|GTP-Rho-binding protein 2|rhophilin 2|rhophilin-like Rho-GTPase binding protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32978592, end_position_on_the_genomic_accession=33064918, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  85416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85416, CurrentID=0, Status='live', Symbol='ZIC5', Aliases='', description='Zic family member 5', other_designations='Zic family member 5 (odd-paired homolog, Drosophila)|zinc family member 5 protein|zinc finger protein of the cerebellum 5', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99963021, end_position_on_the_genomic_accession=99971924, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85417, CurrentID=0, Status='live', Symbol='CCNB3', Aliases='', description='cyclin B3', other_designations='', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50281421, end_position_on_the_genomic_accession=50351914, orientation='plus', exon_count=13, OMIM=300456, no_hdr0=''),
  101928362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928362, CurrentID=0, Status='live', Symbol='LOC101928362', Aliases='', description='uncharacterized LOC101928362', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16268301, end_position_on_the_genomic_accession=16332337, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101928363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928363, CurrentID=0, Status='live', Symbol='EHD4-AS1', Aliases='', description='EHD4 antisense RNA 1', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41921417, end_position_on_the_genomic_accession=41929286, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928364, CurrentID=0, Status='live', Symbol='LOC101928364', Aliases='', description='uncharacterized LOC101928364', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82545424, end_position_on_the_genomic_accession=82553242, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928368, CurrentID=0, Status='live', Symbol='LOC101928368', Aliases='', description='uncharacterized LOC101928368', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37751555, end_position_on_the_genomic_accession=37761121, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928369, CurrentID=0, Status='live', Symbol='LOC101928369', Aliases='', description='uncharacterized LOC101928369', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43107721, end_position_on_the_genomic_accession=43144669, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928370, CurrentID=0, Status='live', Symbol='LOC101928370', Aliases='', description='uncharacterized LOC101928370', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101234873, end_position_on_the_genomic_accession=101236528, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928371, CurrentID=0, Status='live', Symbol='LOC101928371', Aliases='', description='uncharacterized LOC101928371', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88538720, end_position_on_the_genomic_accession=88575610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928372, CurrentID=0, Status='live', Symbol='LOC101928372', Aliases='', description='uncharacterized LOC101928372', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160932465, end_position_on_the_genomic_accession=160949922, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928373, CurrentID=0, Status='live', Symbol='LOC101928373', Aliases='', description='uncharacterized LOC101928373', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135853867, end_position_on_the_genomic_accession=135922229, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928376, CurrentID=0, Status='live', Symbol='IL12A-AS1', Aliases='', description='IL12A antisense RNA 1', other_designations='CTD-2049J23.2', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159913400, end_position_on_the_genomic_accession=160207092, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101928377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928377, CurrentID=0, Status='live', Symbol='LOC101928377', Aliases='', description='uncharacterized LOC101928377', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75425467, end_position_on_the_genomic_accession=75427690, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928378, CurrentID=0, Status='live', Symbol='PTOV1-AS2', Aliases='', description='PTOV1 antisense RNA 2', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49856465, end_position_on_the_genomic_accession=49859289, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874143, CurrentID=0, Status='live', Symbol='LINC00358', Aliases='', description='long intergenic non-protein coding RNA 358', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62003525, end_position_on_the_genomic_accession=62029548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928380, CurrentID=0, Status='live', Symbol='LOC101928380', Aliases='', description='uncharacterized LOC101928380', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73079995, end_position_on_the_genomic_accession=73087708, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85437, CurrentID=0, Status='live', Symbol='ZCRB1', Aliases='MADP-1, MADP1, RBM36, SNRNP31, ZCCHC19', description='zinc finger CCHC-type and RNA binding motif 1', other_designations='U11/U12 small nuclear ribonucleoprotein 31 kDa protein|U11/U12 snRNP 31 kDa protein|U11/U12 snRNP 31K|U11/U12-31K', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42312086, end_position_on_the_genomic_accession=42326130, orientation='minus', exon_count=8, OMIM=610750, no_hdr0=''),
  85438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85438, CurrentID=0, Status='live', Symbol='CABS1', Aliases='C4orf35, CLPH, NYD-SP26', description='calcium-binding protein, spermatid-specific 1', other_designations='calcium-binding protein, sperm-specific 1|casein-like phosphoprotein|testis development protein NYD-SP26', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70334954, end_position_on_the_genomic_accession=70337116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85439, CurrentID=0, Status='live', Symbol='STON2', Aliases='STN2, STNB, STNB2', description='stonin 2', other_designations='homolog of stoned B|stoned B homolog 2', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81260650, end_position_on_the_genomic_accession=81427404, orientation='minus', exon_count=9, OMIM=608467, no_hdr0=''),
  85440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85440, CurrentID=0, Status='live', Symbol='DOCK7', Aliases='EIEE23, ZIR2', description='dedicator of cytokinesis 7', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62454726, end_position_on_the_genomic_accession=62688368, orientation='minus', exon_count=53, OMIM=615730, no_hdr0=''),
  85441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85441, CurrentID=0, Status='live', Symbol='HELZ2', Aliases='PDIP-1, PRIC285', description='helicase with zinc finger 2, transcriptional coactivator', other_designations='ATP-dependent helicase PRIC285|PDIP1|PPAR gamma DBD-interacting protein 1|PPAR-alpha-interacting complex protein 285|PPAR-gamma DBD-interacting protein 1|PPAR-gamma DNA-binding domain-interacting protein 1|PPARG-DBD-interacting protein 1|PPARgamma-DNA-binding domain-interacting protein1|PRIC complex|peroxisomal proliferator activated receptor A interacting complex 285|peroxisomal proliferator-activated receptor A interacting complex 285|peroxisomal proliferator-activated receptor A-interacting complex 285 kDa protein|peroxisomal proliferator-activated receptor alpha-interacting cofactor complex, 285 kD subunit', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63558086, end_position_on_the_genomic_accession=63574239, orientation='minus', exon_count=21, OMIM=611265, no_hdr0=''),
  85442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85442, CurrentID=0, Status='live', Symbol='KNDC1', Aliases='C10orf23, RASGEF2, bB439H18.3', description='kinase non-catalytic C-lobe domain (KIND) containing 1', other_designations='KIND domain-containing protein 1|RasGEF domain family, member 2|cerebral protein 9|kinase non-catalytic C-lobe domain-containing protein 1|ras-GEF domain-containing family member 2', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133160467, end_position_on_the_genomic_accession=133226412, orientation='plus', exon_count=30, OMIM=616237, no_hdr0=''),
  85443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85443, CurrentID=0, Status='live', Symbol='DCLK3', Aliases='CLR, DCAMKL3, DCDC3C, DCK3', description='doublecortin like kinase 3', other_designations='CLICK-I,II-related|doublecortin and CaM kinase-like 3|doublecortin domain-containing protein 3C|doublecortin-like and CAM kinase-like 3|doublecortin-like kinase 3', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36712422, end_position_on_the_genomic_accession=36764204, orientation='minus', exon_count=7, OMIM=613167, no_hdr0=''),
  85444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85444, CurrentID=0, Status='live', Symbol='LRRCC1', Aliases='CLERC, SAP2, VFL1', description='leucine rich repeat and coiled-coil centrosomal protein 1', other_designations='centrosomal leucine-rich repeat and coiled-coil containing protein|centrosomal leucine-rich repeat and coiled-coil domain-containing protein|leucine rich repeat and coiled-coil domain containing 1|sodium channel associated protein 2|variable number of flagella 1 homolog', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85107088, end_position_on_the_genomic_accession=85146080, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  85445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85445, CurrentID=0, Status='live', Symbol='CNTNAP4', Aliases='CASPR4', description='contactin associated protein-like 4', other_designations='cell recognition molecule Caspr4|cell recognition protein CASPR4', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76277278, end_position_on_the_genomic_accession=76560884, orientation='plus', exon_count=25, OMIM=610518, no_hdr0=''),
  85446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85446, CurrentID=0, Status='live', Symbol='ZFHX2', Aliases='ZFH-5, ZNF409', description='zinc finger homeobox 2', other_designations='ZFH-2|zinc finger homeodomain protein 2|zinc finger protein 409', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23520855, end_position_on_the_genomic_accession=23555939, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100874145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874145, CurrentID=0, Status='live', Symbol='LINC00364', Aliases='', description='long intergenic non-protein coding RNA 364', other_designations='', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67372387, end_position_on_the_genomic_accession=67379994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928392, CurrentID=0, Status='live', Symbol='LOC101928392', Aliases='', description='uncharacterized LOC101928392', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82545096, end_position_on_the_genomic_accession=82575461, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  85449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85449, CurrentID=0, Status='live', Symbol='KIAA1755', Aliases='', description='KIAA1755', other_designations='RP5-1054A22.3', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38210505, end_position_on_the_genomic_accession=38260772, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  85450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85450, CurrentID=0, Status='live', Symbol='ITPRIP', Aliases='DANGER, KIAA1754, bA127L20, bA127L20.2', description='inositol 1,4,5-trisphosphate receptor interacting protein', other_designations='inositol 1,4,5-triphosphate receptor-interacting protein', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104309696, end_position_on_the_genomic_accession=104338493, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  85451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85451, CurrentID=0, Status='live', Symbol='UNK', Aliases='UNKEMPT, ZC3H5, ZC3HDC5', description='unkempt family zinc finger', other_designations='unkempt homolog|zinc finger CCCH domain-containing protein 5|zinc finger CCCH type domain containing 5|zinc finger CCCH-type containing 5|zinc finger CCCH-type domain containing 5', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75784839, end_position_on_the_genomic_accession=75825805, orientation='plus', exon_count=22, OMIM=616375, no_hdr0=''),
  85452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85452, CurrentID=0, Status='live', Symbol='CFAP74', Aliases='C1orf222, KIAA1751', description='cilia and flagella associated protein 74', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1921951, end_position_on_the_genomic_accession=2003837, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  85453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85453, CurrentID=0, Status='live', Symbol='TSPYL5', Aliases='', description='TSPY-like 5', other_designations='TSPY-like protein 5', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97273486, end_position_on_the_genomic_accession=97277948, orientation='minus', exon_count=1, OMIM=614721, no_hdr0=''),
  347598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347598, CurrentID=0, Status='live', Symbol='RBMY2DP', Aliases='RBM, RBMY2', description='RNA binding motif protein, Y-linked, family 2, member D pseudogene', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26121585, end_position_on_the_genomic_accession=26133714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928399, CurrentID=0, Status='live', Symbol='TCONS_00029157', Aliases='', description='uncharacterized LOC101928399', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43358055, end_position_on_the_genomic_accession=43362688, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85456, CurrentID=0, Status='live', Symbol='TNKS1BP1', Aliases='TAB182', description='tankyrase 1 binding protein 1', other_designations='tankyrase 1 binding protein 1, 182kDa|tankyrase 1-binding protein of 182 kDa', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57299629, end_position_on_the_genomic_accession=57324952, orientation='minus', exon_count=15, OMIM=607104, no_hdr0=''),
  85457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85457, CurrentID=0, Status='live', Symbol='CIPC', Aliases='KIAA1737', description='CLOCK-interacting pacemaker', other_designations='CLOCK-interacting circadian protein|CLOCK-interacting protein, circadian', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77098235, end_position_on_the_genomic_accession=77117287, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85458, CurrentID=0, Status='live', Symbol='DIXDC1', Aliases='CCD1', description='DIX domain containing 1', other_designations='coiled-coil protein DIX1', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111927144, end_position_on_the_genomic_accession=112022650, orientation='plus', exon_count=26, OMIM=610493, no_hdr0=''),
  85459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85459, CurrentID=0, Status='live', Symbol='CEP295', Aliases='KIAA1731', description='centrosomal protein 295kDa', other_designations='centriolar protein CEP295', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93661650, end_position_on_the_genomic_accession=93730358, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  85460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85460, CurrentID=0, Status='live', Symbol='ZNF518B', Aliases='', description='zinc finger protein 518B', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10439873, end_position_on_the_genomic_accession=10457408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85461, CurrentID=0, Status='live', Symbol='TANC1', Aliases='ROLSB, TANC', description='tetratricopeptide repeat, ankyrin repeat and coiled-coil containing 1', other_designations='TPR domain, ankyrin-repeat and coiled-coil-containing|rolling pebbles homolog B|tetratricopeptide repeat, ankyrin repeat and coiled-coil domain-containing protein 1', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158968634, end_position_on_the_genomic_accession=159232663, orientation='plus', exon_count=34, OMIM=611397, no_hdr0=''),
  85462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85462, CurrentID=0, Status='live', Symbol='FHDC1', Aliases='', description='FH2 domain containing 1', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152911570, end_position_on_the_genomic_accession=152979696, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  85463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85463, CurrentID=0, Status='live', Symbol='ZC3H12C', Aliases='MCPIP3', description='zinc finger CCCH-type containing 12C', other_designations='MCP induced protein 3|MCP-induced protein 3|zinc finger CCCH domain-containing protein 12C', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110088412, end_position_on_the_genomic_accession=110171841, orientation='plus', exon_count=9, OMIM=615001, no_hdr0=''),
  85464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85464, CurrentID=0, Status='live', Symbol='SSH2', Aliases='SSH-2, SSH-2L', description='slingshot protein phosphatase 2', other_designations='SSH-like protein 2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29625938, end_position_on_the_genomic_accession=29930239, orientation='minus', exon_count=24, OMIM=606779, no_hdr0=''),
  85465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85465, CurrentID=0, Status='live', Symbol='EPT1', Aliases='SELI, SEPI', description='ethanolaminephosphotransferase 1', other_designations='ethanolaminephosphotransferase 1 (CDP-ethanolamine-specific)|hEPT1|selenoprotein I', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26346086, end_position_on_the_genomic_accession=26395891, orientation='plus', exon_count=10, OMIM=607915, no_hdr0=''),
  347613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347613, CurrentID=0, Status='live', Symbol='PARP4P1', Aliases='ADPRTL1P, PARP4P, PARP4PY1', description='poly(ADP-ribose) polymerase family member 4 pseudogene 1', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 1 Y-linked pseudogene|poly (ADP-ribose) polymerase family, member 4 pseudogene 1|poly (ADP-ribose) polymerase family, member 4 pseudogene, Y-linked 1|poly(ADP-ribose) polymerase family, member 4 pseudogene 1', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26594668, end_position_on_the_genomic_accession=26634655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928414, CurrentID=0, Status='live', Symbol='LOC101928414', Aliases='', description='uncharacterized LOC101928414', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45251581, end_position_on_the_genomic_accession=45279222, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928415, CurrentID=0, Status='live', Symbol='LOC101928415', Aliases='', description='uncharacterized LOC101928415', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18984202, end_position_on_the_genomic_accession=19083978, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928416, CurrentID=0, Status='live', Symbol='LOC101928416', Aliases='', description='uncharacterized LOC101928416', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132329850, end_position_on_the_genomic_accession=132331575, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928417, CurrentID=0, Status='live', Symbol='LOC101928417', Aliases='', description='uncharacterized LOC101928417', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82953230, end_position_on_the_genomic_accession=82990298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85474, CurrentID=0, Status='live', Symbol='LBX2', Aliases='LP3727', description='ladybird homeobox 2', other_designations='lady bird-like homeobox 2|ladybird homeobox homolog 2|ladybird homeobox protein homolog 2|ladybird-like homeobox 2|ladybird-like homeodomain protein 2', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74497517, end_position_on_the_genomic_accession=74503316, orientation='minus', exon_count=3, OMIM=607164, no_hdr0=''),
  101928419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928419, CurrentID=0, Status='live', Symbol='LOC101928419', Aliases='', description='uncharacterized LOC101928419', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56124145, end_position_on_the_genomic_accession=56127275, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85476, CurrentID=0, Status='live', Symbol='GFM1', Aliases='COXPD1, EFG, EFG1, EFGM, EGF1, GFM, hEFG1', description='G elongation factor, mitochondrial 1', other_designations='G elongation factor mitochondrial 1|G translation elongation factor, mitochondrial|mitochondrial elongation factor G|mitochondrial elongation factor G1', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158644497, end_position_on_the_genomic_accession=158692572, orientation='plus', exon_count=19, OMIM=606639, no_hdr0=''),
  85477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85477, CurrentID=0, Status='live', Symbol='SCIN', Aliases='', description='scinderin', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12570349, end_position_on_the_genomic_accession=12653603, orientation='plus', exon_count=17, OMIM=613416, no_hdr0=''),
  85478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85478, CurrentID=0, Status='live', Symbol='CCDC65', Aliases='CILD27, FAP250, NYD-SP28', description='coiled-coil domain containing 65', other_designations='testis development protein NYD-SP28', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48904110, end_position_on_the_genomic_accession=48921576, orientation='plus', exon_count=8, OMIM=611088, no_hdr0=''),
  85479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85479, CurrentID=0, Status='live', Symbol='DNAJC5B', Aliases='CSP-beta', description='DnaJ heat shock protein family (Hsp40) member C5 beta', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5 beta|beta cysteine string protein|beta-CSP|cysteine string protein beta|dnaJ homolog subfamily C member X', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66015042, end_position_on_the_genomic_accession=66100520, orientation='plus', exon_count=9, OMIM=613945, no_hdr0=''),
  85480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85480, CurrentID=0, Status='live', Symbol='TSLP', Aliases='', description='thymic stromal lymphopoietin', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111070080, end_position_on_the_genomic_accession=111078024, orientation='plus', exon_count=6, OMIM=607003, no_hdr0=''),
  85481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85481, CurrentID=0, Status='live', Symbol='PSKH2', Aliases='', description='protein serine kinase H2', other_designations='PSK-H2|serine/threonine kinase PSKH2', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86047109, end_position_on_the_genomic_accession=86069622, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874151, CurrentID=0, Status='live', Symbol='LINC00381', Aliases='', description='long intergenic non-protein coding RNA 381', other_designations='hCG1820717', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74419173, end_position_on_the_genomic_accession=74435159, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928429, CurrentID=0, Status='live', Symbol='LOC101928429', Aliases='', description='uncharacterized LOC101928429', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136784116, end_position_on_the_genomic_accession=136786025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85486, CurrentID=0, Status='live', Symbol='RPS19P1', Aliases='RPS19_3_1688, bA189K21.7', description='ribosomal protein S19 pseudogene 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18504581, end_position_on_the_genomic_accession=18505104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  85487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85487, CurrentID=0, Status='live', Symbol='RPS19P2', Aliases='RPS19_1_86', description='ribosomal protein S19 pseudogene 2', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113168989, end_position_on_the_genomic_accession=113169429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874152, CurrentID=0, Status='live', Symbol='LINC00387', Aliases='', description='long intergenic non-protein coding RNA 387', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928435, CurrentID=0, Status='live', Symbol='LOC101928435', Aliases='', description='uncharacterized LOC101928435', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38974429, end_position_on_the_genomic_accession=38977774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928436, CurrentID=0, Status='live', Symbol='LOC101928436', Aliases='', description='uncharacterized LOC101928436', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103418079, end_position_on_the_genomic_accession=103525483, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928437, CurrentID=0, Status='live', Symbol='LOC101928437', Aliases='', description='uncharacterized LOC101928437', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=113042727, end_position_on_the_genomic_accession=113520614, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927489, CurrentID=0, Status='live', Symbol='TCERG1L-AS1', Aliases='', description='TCERG1L antisense RNA 1', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131095218, end_position_on_the_genomic_accession=131095777, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85495, CurrentID=0, Status='live', Symbol='RPPH1', Aliases='H1RNA-1, RPPH1', description='ribonuclease P RNA component H1', other_designations='H1 RNA', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20343071, end_position_on_the_genomic_accession=20343411, orientation='minus', exon_count=1, OMIM=608513, no_hdr0=''),
  101928440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928440, CurrentID=0, Status='live', Symbol='FBXO3-AS1', Aliases='', description='FBXO3 antisense RNA 1 (head to head)', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33774699, end_position_on_the_genomic_accession=33775670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928441, CurrentID=0, Status='live', Symbol='LOC101928441', Aliases='', description='uncharacterized LOC101928441', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=23175636, end_position_on_the_genomic_accession=23191587, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928442, CurrentID=0, Status='live', Symbol='LINC01491', Aliases='', description='long intergenic non-protein coding RNA 1491', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47803384, end_position_on_the_genomic_accession=47846236, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928443, CurrentID=0, Status='live', Symbol='LOC101928443', Aliases='', description='uncharacterized LOC101928443', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70056230, end_position_on_the_genomic_accession=70065371, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928444, CurrentID=0, Status='live', Symbol='LOC101928444', Aliases='', description='histone H2B type 1-F/J/L-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24632246, end_position_on_the_genomic_accession=24632601, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874154, CurrentID=0, Status='live', Symbol='LINC00391', Aliases='', description='long intergenic non-protein coding RNA 391', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85502, CurrentID=0, Status='live', Symbol='SCAR3', Aliases='SCABD', description='spinocerebellar ataxia, autosomal recessive 3', other_designations='spinocerebellar ataxia with blindness and deafness', map_location='6p23-p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=271250, no_hdr0=''),
  85503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85503, CurrentID=0, Status='live', Symbol='MCOPCB1', Aliases='MOPCB, MOPCB1', description='Microphthalmia, colobomatous, 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300345, no_hdr0=''),
  85504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85504, CurrentID=0, Status='live', Symbol='BDET', Aliases='', description='Bleeding disorder, east Texas type', other_designations='', map_location='1q23', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605913, no_hdr0=''),
  85505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85505, CurrentID=0, Status='live', Symbol='ACRPS', Aliases='', description='Acropectoral syndrome', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605967, no_hdr0=''),
  101928450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928450, CurrentID=0, Status='live', Symbol='LOC101928450', Aliases='', description='uncharacterized LOC101928450', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=781002, end_position_on_the_genomic_accession=782290, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85507, CurrentID=0, Status='live', Symbol='DFNA18', Aliases='', description='deafness, autosomal dominant 18', other_designations='', map_location='3q22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606012, no_hdr0=''),
  85508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85508, CurrentID=0, Status='live', Symbol='SCRT2', Aliases='ZNF898B', description='scratch family zinc finger 2', other_designations='scratch 2 protein|scratch homolog 2 zinc finger protein|scratch homolog 2, zinc finger protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=661596, end_position_on_the_genomic_accession=676179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  85509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85509, CurrentID=0, Status='live', Symbol='MBD3L1', Aliases='MBD3L', description='methyl-CpG binding domain protein 3-like 1', other_designations='MBD3-like protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8842593, end_position_on_the_genomic_accession=8843340, orientation='plus', exon_count=1, OMIM=607963, no_hdr0=''),
  101928455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928455, CurrentID=0, Status='live', Symbol='LINC01412', Aliases='', description='long intergenic non-protein coding RNA 1412', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144521868, end_position_on_the_genomic_accession=144579434, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874156, CurrentID=0, Status='live', Symbol='LINC00393', Aliases='', description='long intergenic non-protein coding RNA 393', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73456899, end_position_on_the_genomic_accession=73662021, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105374295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374295, CurrentID=0, Status='live', Symbol='LOC105374295', Aliases='', description='uncharacterized LOC105374295', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195005769, end_position_on_the_genomic_accession=195014874, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928460, CurrentID=0, Status='live', Symbol='LOC101928460', Aliases='', description='uncharacterized LOC101928460', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=29329620, end_position_on_the_genomic_accession=29350114, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928461, CurrentID=0, Status='live', Symbol='LOC101928461', Aliases='', description='uncharacterized LOC101928461', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136629114, end_position_on_the_genomic_accession=136648198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928462, CurrentID=0, Status='live', Symbol='LOC101928462', Aliases='', description='uncharacterized LOC101928462', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81052778, end_position_on_the_genomic_accession=81172005, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100874157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874157, CurrentID=0, Status='live', Symbol='LINC00395', Aliases='', description='long intergenic non-protein coding RNA 395', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63667682, end_position_on_the_genomic_accession=63738018, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928464, CurrentID=0, Status='live', Symbol='LOC101928464', Aliases='', description='uncharacterized LOC101928464', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11987603, end_position_on_the_genomic_accession=12046276, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105374296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374296, CurrentID=0, Status='live', Symbol='LOC105374296', Aliases='', description='uncharacterized LOC105374296', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195148459, end_position_on_the_genomic_accession=195151262, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928466, CurrentID=0, Status='live', Symbol='LOC101928466', Aliases='', description='uncharacterized LOC101928466', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143588289, end_position_on_the_genomic_accession=143620919, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928467, CurrentID=0, Status='live', Symbol='LOC101928467', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13116502, end_position_on_the_genomic_accession=13120807, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928468, CurrentID=0, Status='live', Symbol='LOC101928468', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56813396, end_position_on_the_genomic_accession=56814205, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928469, CurrentID=0, Status='live', Symbol='MIR325HG', Aliases='MIR384HG', description='MIR325 host gene', other_designations='MIR325 host gene (non-protein coding)', map_location='Xq13.3-q21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76657798, end_position_on_the_genomic_accession=77014532, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928470, CurrentID=0, Status='live', Symbol='TMEM246-AS1', Aliases='', description='TMEM246 antisense RNA 1', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101468439, end_position_on_the_genomic_accession=101481506, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928471, CurrentID=0, Status='live', Symbol='LOC101928471', Aliases='', description='uncharacterized LOC101928471', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24223271, end_position_on_the_genomic_accession=24237965, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928472, CurrentID=0, Status='live', Symbol='LOC101928472', Aliases='', description='uncharacterized LOC101928472', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48782907, end_position_on_the_genomic_accession=48787102, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347673, CurrentID=0, Status='live', Symbol='EIF2S2P2', Aliases='bcm2636', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 2', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 2', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185245166, end_position_on_the_genomic_accession=185246567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928474, CurrentID=0, Status='live', Symbol='LOC101928474', Aliases='', description='uncharacterized LOC101928474', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84916186, end_position_on_the_genomic_accession=84924966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874159, CurrentID=0, Status='live', Symbol='LINC00400', Aliases='', description='long intergenic non-protein coding RNA 400', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43113165, end_position_on_the_genomic_accession=43159466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928476, CurrentID=0, Status='live', Symbol='LOC101928476', Aliases='', description='uncharacterized LOC101928476', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105589694, end_position_on_the_genomic_accession=105618935, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374298, CurrentID=0, Status='live', Symbol='LOC105374298', Aliases='', description='uncharacterized LOC105374298', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195636885, end_position_on_the_genomic_accession=195640642, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928478, CurrentID=0, Status='live', Symbol='LOC101928478', Aliases='', description='uncharacterized LOC101928478', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173695511, end_position_on_the_genomic_accession=173929717, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100874160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874160, CurrentID=0, Status='live', Symbol='LINC00404', Aliases='', description='long intergenic non-protein coding RNA 404', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928483, CurrentID=0, Status='live', Symbol='NALT1', Aliases='LINC01573, MIR4674HG, NALT', description='NOTCH1 associated lncRNA in T-cell acute lymphoblastic leukemia 1', other_designations='MIR4674 host gene (non-protein coding)|NOTCH1 associated lncRNA in T cell acute lymphoblastic leukemia 1|NOTCH1 associated long non-coding RNA in pediatric T-ALL|TCONS_l2_00029132|long intergenic non-protein coding RNA 1573', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136547842, end_position_on_the_genomic_accession=136549744, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928484, CurrentID=0, Status='live', Symbol='LINC01363', Aliases='RBSG4', description='long intergenic non-protein coding RNA 1363', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167175362, end_position_on_the_genomic_accession=167195805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  347686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347686, CurrentID=0, Status='live', Symbol='SNORD64', Aliases='HBII-13', description='small nucleolar RNA, C/D box 64', other_designations='HBII-13 snoRNA', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24985100, end_position_on_the_genomic_accession=24985166, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  347687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347687, CurrentID=0, Status='live', Symbol='GNG5P2', Aliases='GNG5ps, dJ364I1.1', description='guanine nucleotide binding protein (G protein), gamma 5 pseudogene 2', other_designations='G protein gamma 5-like subunit|dJ364I1.1 (G Protein (Guanine Nucleotide-binding protein) Gamma 5 subunit (GNGT5, GNG5) LIKE protein)', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110346455, end_position_on_the_genomic_accession=110346973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347688, CurrentID=0, Status='live', Symbol='TUBB8', Aliases='bA631M21.2', description='tubulin beta 8 class VIII', other_designations='HSA10p15 beta-tubulin 4Q|class VIII beta-tubulin|tubulin, beta 8 class VIII', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46638, end_position_on_the_genomic_accession=76455, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  347689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347689, CurrentID=0, Status='live', Symbol='SOX2-OT', Aliases='NCRNA00043, SOX2OT', description='SOX2 overlapping transcript', other_designations='SOX2 overlapping transcript (non-coding RNA)|SOX2 overlapping transcript (non-protein coding)', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181056680, end_position_on_the_genomic_accession=181742228, orientation='plus', exon_count=10, OMIM=616338, no_hdr0=''),
  101928491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928491, CurrentID=0, Status='live', Symbol='LOC101928491', Aliases='', description='uncharacterized LOC101928491', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17501360, end_position_on_the_genomic_accession=17511265, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928492, CurrentID=0, Status='live', Symbol='OR2A1-AS1', Aliases='', description='OR2A1 antisense RNA 1', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144238573, end_position_on_the_genomic_accession=144355718, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  347694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347694, CurrentID=0, Status='live', Symbol='ECEL1P2', Aliases='ECEL2', description='endothelin converting enzyme-like 1, pseudogene 2', other_designations='ECEL2 (XCE) pseudogene 2|ecel1 pseudogene 2', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232385750, end_position_on_the_genomic_accession=232387044, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928495, CurrentID=0, Status='live', Symbol='LOC101928495', Aliases='', description='uncharacterized LOC101928495', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126109762, end_position_on_the_genomic_accession=126115562, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928496, CurrentID=0, Status='live', Symbol='LINC01492', Aliases='', description='long intergenic non-protein coding RNA 1492', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103140528, end_position_on_the_genomic_accession=103325033, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101928499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928499, CurrentID=0, Status='live', Symbol='LOC101928499', Aliases='', description='uncharacterized LOC101928499', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52800166, end_position_on_the_genomic_accession=52806418, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928500, CurrentID=0, Status='live', Symbol='LOC101928500', Aliases='', description='uncharacterized LOC101928500', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=70449353, end_position_on_the_genomic_accession=70460391, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347701, CurrentID=0, Status='live', Symbol='S100A11P2', Aliases='', description='S100 calcium binding protein A11 pseudogene 2', other_designations='H_NH0456N16.1|WUGSC:H_NH0456N16.1', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35169175, end_position_on_the_genomic_accession=35169705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347702, CurrentID=0, Status='live', Symbol='MRPS36P4', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 4', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112208597, end_position_on_the_genomic_accession=112208898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347703, CurrentID=0, Status='live', Symbol='MRPS36P5', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 5', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=43054083, end_position_on_the_genomic_accession=43054384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347704, CurrentID=0, Status='live', Symbol='MRPS36P3', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 3', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122091470, end_position_on_the_genomic_accession=122091602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347705, CurrentID=0, Status='live', Symbol='MRPS36P1', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 1', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6773037, end_position_on_the_genomic_accession=6773346, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347706, CurrentID=0, Status='live', Symbol='MRPS36P2', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 2', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186900739, end_position_on_the_genomic_accession=186900841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347707, CurrentID=0, Status='live', Symbol='MRPS36P6', Aliases='', description='mitochondrial ribosomal protein S36 pseudogene 6', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13405850, end_position_on_the_genomic_accession=13406070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928508, CurrentID=0, Status='live', Symbol='LOC101928508', Aliases='', description='zinc finger Ran-binding domain-containing protein 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96070659, end_position_on_the_genomic_accession=96072868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928509, CurrentID=0, Status='live', Symbol='LOC101928509', Aliases='', description='uncharacterized LOC101928509', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174094660, end_position_on_the_genomic_accession=174220398, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928510, CurrentID=0, Status='live', Symbol='LOC101928510', Aliases='', description='uncharacterized LOC101928510', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36063283, end_position_on_the_genomic_accession=36078277, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371098, CurrentID=0, Status='live', Symbol='LOC105371098', Aliases='', description='uncharacterized LOC105371098', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16380296, end_position_on_the_genomic_accession=16398504, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85569, CurrentID=0, Status='live', Symbol='GALP', Aliases='', description='galanin like peptide', other_designations='alarin|gal-like peptide', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56176020, end_position_on_the_genomic_accession=56185775, orientation='plus', exon_count=6, OMIM=611178, no_hdr0=''),
  101928514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928514, CurrentID=0, Status='live', Symbol='LOC101928514', Aliases='', description='uncharacterized LOC101928514', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76799044, end_position_on_the_genomic_accession=76807102, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347715, CurrentID=0, Status='live', Symbol='LOC347715', Aliases='', description='FLJ11800 related gene', other_designations='', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347716, CurrentID=0, Status='live', Symbol='USP32P3', Aliases='', description='ubiquitin specific peptidase 32 pseudogene 3', other_designations='TL132 pseudogene', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20424591, end_position_on_the_genomic_accession=20433075, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  347717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347717, CurrentID=0, Status='live', Symbol='SRP68P3', Aliases='SRPSMCR', description='signal recognition particle 68kDa pseudogene 3', other_designations='signal recognition particle 68kD pseudogene', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20415791, end_position_on_the_genomic_accession=20416695, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928519, CurrentID=0, Status='live', Symbol='LOC101928519', Aliases='', description='uncharacterized LOC101928519', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19068543, end_position_on_the_genomic_accession=19180480, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  347720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347720, CurrentID=0, Status='live', Symbol='COTL1P2', Aliases='CLPSMCR', description='coactosin-like F-actin binding protein 1 pseudogene 2', other_designations='COTL1 pseudogene 2|coactosin-like 1 pseudogene 2', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20564464, end_position_on_the_genomic_accession=20566681, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928521, CurrentID=0, Status='live', Symbol='LOC101928521', Aliases='', description='uncharacterized LOC101928521', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=652909, end_position_on_the_genomic_accession=655946, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928522, CurrentID=0, Status='live', Symbol='LOC101928522', Aliases='', description='uncharacterized LOC101928522', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129072907, end_position_on_the_genomic_accession=129128364, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874167, CurrentID=0, Status='live', Symbol='LINC00423', Aliases='', description='long intergenic non-protein coding RNA 423', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32809190, end_position_on_the_genomic_accession=32911652, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928524, CurrentID=0, Status='live', Symbol='LOC101928524', Aliases='', description='uncharacterized LOC101928524', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13632291, end_position_on_the_genomic_accession=13655929, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928525, CurrentID=0, Status='live', Symbol='LOC101928525', Aliases='', description='uncharacterized LOC101928525', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135503273, end_position_on_the_genomic_accession=135506447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928526, CurrentID=0, Status='live', Symbol='LOC101928526', Aliases='', description='uncharacterized LOC101928526', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148855086, end_position_on_the_genomic_accession=148859502, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928527, CurrentID=0, Status='live', Symbol='PIGBOS1', Aliases='HP06981', description='PIGB opposite strand 1', other_designations='PIGB opposite strand protein', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55317184, end_position_on_the_genomic_accession=55319189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927492, CurrentID=0, Status='live', Symbol='UTAT33', Aliases='', description='uncharacterized LOC101927492', other_designations='uterine endometrial carcinoma-associated transcript 33', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105097052, end_position_on_the_genomic_accession=105102944, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874168, CurrentID=0, Status='live', Symbol='LINC00433', Aliases='', description='long intergenic non-protein coding RNA 433', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88540829, end_position_on_the_genomic_accession=88545509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347730, CurrentID=0, Status='live', Symbol='LRRTM1', Aliases='', description='leucine rich repeat transmembrane neuronal 1', other_designations='leucine-rich repeat transmembrane neuronal 1 protein', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=80288280, end_position_on_the_genomic_accession=80304593, orientation='minus', exon_count=2, OMIM=610867, no_hdr0=''),
  347731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347731, CurrentID=0, Status='live', Symbol='LRRTM3', Aliases='', description='leucine rich repeat transmembrane neuronal 3', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=66926034, end_position_on_the_genomic_accession=67101551, orientation='plus', exon_count=5, OMIM=610869, no_hdr0=''),
  347732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347732, CurrentID=0, Status='live', Symbol='CATSPER3', Aliases='CACRC', description='cation channel, sperm associated 3', other_designations='ca(v)-like protein|calcium channel repeat containing 1|one-repeat calcium channel-like protein|putative ion channel CatSper3', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134967906, end_position_on_the_genomic_accession=135011707, orientation='plus', exon_count=8, OMIM=609120, no_hdr0=''),
  347733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347733, CurrentID=0, Status='live', Symbol='TUBB2B', Aliases='PMGYSA, bA506K6.1', description='tubulin beta 2B class IIb', other_designations='class II beta-tubulin isotype|class IIb beta-tubulin|tubulin, beta 2B class IIb|tubulin, beta polypeptide paralog', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3224261, end_position_on_the_genomic_accession=3231665, orientation='minus', exon_count=5, OMIM=612850, no_hdr0=''),
  347734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347734, CurrentID=0, Status='live', Symbol='SLC35B2', Aliases='PAPST1, SLL, UGTrel4', description="solute carrier family 35 (adenosine 3'-phospho 5'-phosphosulfate transporter), member B2", other_designations="3'-phosphoadenosine 5'-phosphosulfate transporter|PAPS transporter 1|putative MAPK-activating protein PM15|putative NF-kappa-B-activating protein 48|solute carrier family 35 member B2 variant 2|solute carrier family 35, member B2", map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44254101, end_position_on_the_genomic_accession=44257890, orientation='minus', exon_count=6, OMIM=610788, no_hdr0=''),
  347735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347735, CurrentID=0, Status='live', Symbol='SERINC2', Aliases='FKSG84, PRO0899, TDE2, TDE2L', description='serine incorporator 2', other_designations='tumor differentially expressed protein 2', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31409565, end_position_on_the_genomic_accession=31434680, orientation='plus', exon_count=14, OMIM=614549, no_hdr0=''),
  347736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347736, CurrentID=0, Status='live', Symbol='NME9', Aliases='NM23-H9, TXL-2, TXL2, TXNDC6', description='NME/NM23 family member 9', other_designations='NME gene family member 9|thioredoxin-like protein 2', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138261437, end_position_on_the_genomic_accession=138330149, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  347737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347737, CurrentID=0, Status='live', Symbol='DFNA31', Aliases='', description='deafness, autosomal dominant 31', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608645, no_hdr0=''),
  101928539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928539, CurrentID=0, Status='live', Symbol='LOC101928539', Aliases='', description='uncharacterized LOC101928539', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57890331, end_position_on_the_genomic_accession=57899162, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928540, CurrentID=0, Status='live', Symbol='LOC101928540', Aliases='', description='uncharacterized LOC101928540', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75383186, end_position_on_the_genomic_accession=75458900, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  347741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347741, CurrentID=0, Status='live', Symbol='OTOP3', Aliases='', description='otopetrin 3', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74935802, end_position_on_the_genomic_accession=74949992, orientation='plus', exon_count=7, OMIM=607828, no_hdr0=''),
  100131429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131429, CurrentID=0, Status='live', Symbol='LOC100131429', Aliases='', description='armadillo repeat containing 1 pseudogene', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138477233, end_position_on_the_genomic_accession=138482307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347744, CurrentID=0, Status='live', Symbol='C6orf52', Aliases='', description='chromosome 6 open reading frame 52', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10671418, end_position_on_the_genomic_accession=10694797, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  347745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347745, CurrentID=0, Status='live', Symbol='PWAR4', Aliases='PAR-4, PAR4', description='Prader Willi/Angelman region RNA 4', other_designations='Prader-Willi/Angelman region gene 4', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25211692, end_position_on_the_genomic_accession=25212033, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  347746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347746, CurrentID=0, Status='live', Symbol='PWARSN', Aliases='PAR-SN, PARSN', description='Prader Willi/Angelman region RNA, SNRPN neighbor', other_designations='paternally expressed transcript PAR-SN|paternally expressed transcript adjacent to snRPN', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24981994, end_position_on_the_genomic_accession=24983790, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  79803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79803, CurrentID=0, Status='live', Symbol='HPS6', Aliases='', description='Hermansky-Pudlak syndrome 6', other_designations='Hermansky-Pudlak syndrome-6 protein (HPS6)|ruby-eye protein homolog', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102065367, end_position_on_the_genomic_accession=102068038, orientation='plus', exon_count=1, OMIM=607522, no_hdr0=''),
  101928548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928548, CurrentID=0, Status='live', Symbol='LOC101928548', Aliases='', description='uncharacterized LOC101928548', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1200379, end_position_on_the_genomic_accession=1201262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928550, CurrentID=0, Status='live', Symbol='SMC2-AS1', Aliases='', description='SMC2 antisense RNA 1 (head to head)', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104080024, end_position_on_the_genomic_accession=104092474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928551, CurrentID=0, Status='live', Symbol='LOC101928551', Aliases='', description='uncharacterized LOC101928551', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174831728, end_position_on_the_genomic_accession=174875964, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928552, CurrentID=0, Status='live', Symbol='LOC101928552', Aliases='', description='mucin-6-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874172, CurrentID=0, Status='live', Symbol='LINC00440', Aliases='', description='long intergenic non-protein coding RNA 440', other_designations='', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89235291, end_position_on_the_genomic_accession=89280240, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928554, CurrentID=0, Status='live', Symbol='LOC101928554', Aliases='', description='uncharacterized LOC101928554', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26210842, end_position_on_the_genomic_accession=26219348, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928555, CurrentID=0, Status='live', Symbol='LINC01537', Aliases='', description='long intergenic non-protein coding RNA 1537', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72570656, end_position_on_the_genomic_accession=72573222, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928556, CurrentID=0, Status='live', Symbol='LOC101928556', Aliases='', description='uncharacterized LOC101928556', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29469795, end_position_on_the_genomic_accession=29505466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928557, CurrentID=0, Status='live', Symbol='LOC101928557', Aliases='', description='uncharacterized LOC101928557', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85792415, end_position_on_the_genomic_accession=85792933, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874173, CurrentID=0, Status='live', Symbol='LINC00443', Aliases='', description='long intergenic non-protein coding RNA 443', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106653880, end_position_on_the_genomic_accession=106672180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928560, CurrentID=0, Status='live', Symbol='LOC101928560', Aliases='', description='uncharacterized LOC101928560', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109546610, end_position_on_the_genomic_accession=109548538, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928561, CurrentID=0, Status='live', Symbol='LOC101928561', Aliases='', description='uncharacterized LOC101928561', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108499062, end_position_on_the_genomic_accession=108517723, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928563, CurrentID=0, Status='live', Symbol='LOC101928563', Aliases='', description='uncharacterized LOC101928563', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38868368, end_position_on_the_genomic_accession=38885214, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928565, CurrentID=0, Status='live', Symbol='LOC101928565', Aliases='', description='uncharacterized LOC101928565', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168464114, end_position_on_the_genomic_accession=168495644, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928566, CurrentID=0, Status='live', Symbol='TBC1D3P6', Aliases='', description='TBC1 domain family member 3 pseudogene 6', other_designations='TBC1 domain family, member 3 pseudogene 6', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15976131, end_position_on_the_genomic_accession=15997206, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928567, CurrentID=0, Status='live', Symbol='LOC101928567', Aliases='', description='uncharacterized LOC101928567', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15806242, end_position_on_the_genomic_accession=15817742, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928569, CurrentID=0, Status='live', Symbol='LOC101928569', Aliases='', description='uncharacterized LOC101928569', other_designations='CTD-2037L6.2', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58107632, end_position_on_the_genomic_accession=58122341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928570, CurrentID=0, Status='live', Symbol='LOC101928570', Aliases='', description='uncharacterized LOC101928570', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77112751, end_position_on_the_genomic_accession=77316924, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928571, CurrentID=0, Status='live', Symbol='ZNF649-AS1', Aliases='', description='ZNF649 antisense RNA 1', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51888025, end_position_on_the_genomic_accession=51900464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928572, CurrentID=0, Status='live', Symbol='LOC101928572', Aliases='', description='uncharacterized LOC101928572', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2302315, end_position_on_the_genomic_accession=2304225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374314, CurrentID=0, Status='live', Symbol='LOC105374314', Aliases='', description='uncharacterized LOC105374314', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160032259, end_position_on_the_genomic_accession=160040364, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928574, CurrentID=0, Status='live', Symbol='LOC101928574', Aliases='', description='uncharacterized LOC101928574', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12207435, end_position_on_the_genomic_accession=12242746, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928576, CurrentID=0, Status='live', Symbol='LOC101928576', Aliases='', description='uncharacterized LOC101928576', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5101665, end_position_on_the_genomic_accession=5103184, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928578, CurrentID=0, Status='live', Symbol='RAP2C-AS1', Aliases='', description='RAP2C antisense RNA 1', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132218507, end_position_on_the_genomic_accession=132432811, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928580, CurrentID=0, Status='live', Symbol='LOC101928580', Aliases='', description='uncharacterized LOC101928580', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74311362, end_position_on_the_genomic_accession=74324705, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928582, CurrentID=0, Status='live', Symbol='LOC101928582', Aliases='', description='uncharacterized LOC101928582', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86196419, end_position_on_the_genomic_accession=86201314, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928583, CurrentID=0, Status='live', Symbol='LOC101928583', Aliases='', description='uncharacterized LOC101928583', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170656562, end_position_on_the_genomic_accession=170662123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928584, CurrentID=0, Status='live', Symbol='LOC101928584', Aliases='', description='uncharacterized LOC101928584', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91171198, end_position_on_the_genomic_accession=91178107, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928587, CurrentID=0, Status='live', Symbol='LOC101928587', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56826850, end_position_on_the_genomic_accession=56827843, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928588, CurrentID=0, Status='live', Symbol='LOC101928588', Aliases='', description='uncharacterized LOC101928588', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1234725, end_position_on_the_genomic_accession=1237856, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928589, CurrentID=0, Status='live', Symbol='LOC101928589', Aliases='', description='uncharacterized LOC101928589', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110175773, end_position_on_the_genomic_accession=110177788, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928590, CurrentID=0, Status='live', Symbol='LOC101928590', Aliases='', description='uncharacterized LOC101928590', other_designations='', map_location='4q34.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175790307, end_position_on_the_genomic_accession=175812189, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928591, CurrentID=0, Status='live', Symbol='LOC101928591', Aliases='', description='uncharacterized LOC101928591', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=42001722, end_position_on_the_genomic_accession=42085460, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101928592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928592, CurrentID=0, Status='live', Symbol='LOC101928592', Aliases='', description='zinc finger protein OZF-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44680777, end_position_on_the_genomic_accession=44685782, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874179, CurrentID=0, Status='live', Symbol='LINC00457', Aliases='', description='long intergenic non-protein coding RNA 457', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34435450, end_position_on_the_genomic_accession=34640685, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928596, CurrentID=0, Status='live', Symbol='LOC101928596', Aliases='', description='probable tumor suppressor protein MN1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169103700, end_position_on_the_genomic_accession=169106473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928597, CurrentID=0, Status='live', Symbol='LOC101928597', Aliases='', description='uncharacterized LOC101928597', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133032659, end_position_on_the_genomic_accession=133037222, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928599, CurrentID=0, Status='live', Symbol='LOC101928599', Aliases='', description='uncharacterized LOC101928599', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=82642574, end_position_on_the_genomic_accession=82746724, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928600, CurrentID=0, Status='live', Symbol='LOC101928600', Aliases='', description='uncharacterized LOC101928600', other_designations='CTD-2117L12.1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58541567, end_position_on_the_genomic_accession=58558243, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928601, CurrentID=0, Status='live', Symbol='MEI4', Aliases='', description='meiotic double-stranded break formation protein 4', other_designations='meiosis-specific 4 homolog', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77650233, end_position_on_the_genomic_accession=77927045, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928602, CurrentID=0, Status='live', Symbol='LOC101928602', Aliases='', description='uncharacterized LOC101928602', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2458937, end_position_on_the_genomic_accession=2462185, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928603, CurrentID=0, Status='live', Symbol='C6orf229', Aliases='', description='chromosome 6 open reading frame 229', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24797321, end_position_on_the_genomic_accession=24798896, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928604, CurrentID=0, Status='live', Symbol='ZBTB46-AS1', Aliases='', description='ZBTB46 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63808076, end_position_on_the_genomic_accession=63823015, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928605, CurrentID=0, Status='live', Symbol='LOC101928605', Aliases='', description='uncharacterized LOC101928605', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144195321, end_position_on_the_genomic_accession=144299589, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928607, CurrentID=0, Status='live', Symbol='TRPM2-AS', Aliases='TRPM2-AS1', description='TRPM2 antisense RNA', other_designations='transient receptor potential cation channel, subfamily M, member 2', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44414590, end_position_on_the_genomic_accession=44425563, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928608, CurrentID=0, Status='live', Symbol='LOC101928608', Aliases='', description='uncharacterized LOC101928608', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67832801, end_position_on_the_genomic_accession=67853880, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105374320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374320, CurrentID=0, Status='live', Symbol='LOC105374320', Aliases='', description='uncharacterized LOC105374320', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21951656, end_position_on_the_genomic_accession=21980652, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928610, CurrentID=0, Status='live', Symbol='LOC101928610', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47939639, end_position_on_the_genomic_accession=47940314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928611, CurrentID=0, Status='live', Symbol='LINC01413', Aliases='', description='long intergenic non-protein coding RNA 1413', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57319138, end_position_on_the_genomic_accession=57325039, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928612, CurrentID=0, Status='live', Symbol='HSPC324', Aliases='', description='uncharacterized LOC101928612', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136648610, end_position_on_the_genomic_accession=136660463, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874182, CurrentID=0, Status='live', Symbol='LINC00424', Aliases='', description='long intergenic non-protein coding RNA 424', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21872796, end_position_on_the_genomic_accession=21878160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928614, CurrentID=0, Status='live', Symbol='LOC101928614', Aliases='', description='uncharacterized LOC101928614', other_designations='CTD-2015G9.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86720564, end_position_on_the_genomic_accession=86722010, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131441, CurrentID=0, Status='live', Symbol='LOC100131441', Aliases='', description='coiled-coil domain containing 65 pseudogene', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62489446, end_position_on_the_genomic_accession=62490217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928616, CurrentID=0, Status='live', Symbol='LINC00354', Aliases='', description='long intergenic non-protein coding RNA 354', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111893378, end_position_on_the_genomic_accession=111901176, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928617, CurrentID=0, Status='live', Symbol='LOC101928617', Aliases='', description='uncharacterized LOC101928617', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92146085, end_position_on_the_genomic_accession=92167560, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928618, CurrentID=0, Status='live', Symbol='LOC101928618', Aliases='', description='uncharacterized LOC101928618', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36095310, end_position_on_the_genomic_accession=36100653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874183, CurrentID=0, Status='live', Symbol='NCOR1P3', Aliases='', description='nuclear receptor corepressor 1 pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57599794, end_position_on_the_genomic_accession=57610092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374322, CurrentID=0, Status='live', Symbol='LOC105374322', Aliases='', description='uncharacterized LOC105374322', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22390134, end_position_on_the_genomic_accession=22544763, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928622, CurrentID=0, Status='live', Symbol='LOC101928622', Aliases='', description='uncharacterized LOC101928622', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33896339, end_position_on_the_genomic_accession=34039893, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928624, CurrentID=0, Status='live', Symbol='LOC101928624', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43466101, end_position_on_the_genomic_accession=43467124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874184, CurrentID=0, Status='live', Symbol='LINC00502', Aliases='', description='long intergenic non-protein coding RNA 502', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91045808, end_position_on_the_genomic_accession=91062159, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928626, CurrentID=0, Status='live', Symbol='LOC101928626', Aliases='', description='uncharacterized LOC101928626', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=627380, end_position_on_the_genomic_accession=629009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928627, CurrentID=0, Status='live', Symbol='LOC101928627', Aliases='', description='uncharacterized LOC101928627', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36365626, end_position_on_the_genomic_accession=36440292, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928628, CurrentID=0, Status='live', Symbol='LOC101928628', Aliases='', description='uncharacterized LOC101928628', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169762914, end_position_on_the_genomic_accession=169764740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928630, CurrentID=0, Status='live', Symbol='CTC-436P18.1', Aliases='', description='uncharacterized LOC101928630', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61162316, end_position_on_the_genomic_accession=61232080, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874185, CurrentID=0, Status='live', Symbol='FRMD6-AS2', Aliases='', description='FRMD6 antisense RNA 2', other_designations='FRMD6 antisense RNA 2 (non-protein coding)', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51454512, end_position_on_the_genomic_accession=51599952, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100093624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093624, CurrentID=0, Status='live', Symbol='CDH12P3', Aliases='', description='cadherin 12 (N-cadherin 2) pseudogene 3', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70132675, end_position_on_the_genomic_accession=70132976, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100093625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093625, CurrentID=0, Status='live', Symbol='CDH12P1', Aliases='', description='cadherin 12 (N-cadherin 2) pseudogene 1', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70860283, end_position_on_the_genomic_accession=70860584, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100093626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093626, CurrentID=0, Status='live', Symbol='CDH12P4', Aliases='', description='cadherin 12 (N-cadherin 2) pseudogene 4', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71132989, end_position_on_the_genomic_accession=71133290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928635, CurrentID=0, Status='live', Symbol='LOC101928635', Aliases='', description='uncharacterized LOC101928635', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58213535, end_position_on_the_genomic_accession=58332001, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  101928636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928636, CurrentID=0, Status='live', Symbol='LOC101928636', Aliases='', description='uncharacterized LOC101928636', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45199970, end_position_on_the_genomic_accession=45216773, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874186, CurrentID=0, Status='live', Symbol='PLCB2-AS1', Aliases='', description='PLCB2 antisense RNA 1', other_designations='PLCB2 antisense RNA 1 (non-protein coding)', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928638, CurrentID=0, Status='live', Symbol='PRR31', Aliases='C9orf141', description='proline rich 31', other_designations='putative uncharacterized protein C9orf141-like', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136968886, end_position_on_the_genomic_accession=136976820, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100093631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093631, CurrentID=0, Status='live', Symbol='GTF2IP4', Aliases='', description='general transcription factor IIi, pseudogene 4', other_designations='general transcription factor II, i, pseudogene|general transcription factor IIi pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73154924, end_position_on_the_genomic_accession=73207296, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  100093632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093632, CurrentID=0, Status='live', Symbol='DFNA57', Aliases='', description='deafness, autosomal dominant 57', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874187, CurrentID=0, Status='live', Symbol='LINC00559', Aliases='', description='long intergenic non-protein coding RNA 559', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90060247, end_position_on_the_genomic_accession=90119717, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928645, CurrentID=0, Status='live', Symbol='LOC101928645', Aliases='', description='uncharacterized LOC101928645', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136117031, end_position_on_the_genomic_accession=136121839, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928646, CurrentID=0, Status='live', Symbol='ARNTL2-AS1', Aliases='', description='ARNTL2 antisense RNA 1', other_designations='', map_location='12p11.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27389789, end_position_on_the_genomic_accession=27446634, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874188, CurrentID=0, Status='live', Symbol='LINC00571', Aliases='', description='long intergenic non-protein coding RNA 571', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38050817, end_position_on_the_genomic_accession=38143232, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928650, CurrentID=0, Status='live', Symbol='LOC101928650', Aliases='', description='uncharacterized LOC101928650', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170461317, end_position_on_the_genomic_accession=170532609, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928651, CurrentID=0, Status='live', Symbol='LOC101928651', Aliases='', description='uncharacterized LOC101928651', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61663157, end_position_on_the_genomic_accession=61698432, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928652, CurrentID=0, Status='live', Symbol='LOC101928652', Aliases='', description='uncharacterized LOC101928652', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144386497, end_position_on_the_genomic_accession=144390962, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  347853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347853, CurrentID=0, Status='live', Symbol='TBX10', Aliases='TBX13, TBX7', description='T-box 10', other_designations='T-box 7|T-box protein 10|T-box-containing transcriptional activator', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67631303, end_position_on_the_genomic_accession=67639560, orientation='minus', exon_count=8, OMIM=604648, no_hdr0=''),
  341965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341965, CurrentID=0, Status='live', Symbol='VDAC3P1', Aliases='VDAC3P', description='voltage-dependent anion channel 3 pseudogene 1', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99964464, end_position_on_the_genomic_accession=99965575, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928656, CurrentID=0, Status='live', Symbol='LINC01099', Aliases='', description='long intergenic non-protein coding RNA 1099', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177812046, end_position_on_the_genomic_accession=177907936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131448, CurrentID=0, Status='live', Symbol='LOC100131448', Aliases='', description='CTD small phosphatase like 2 pseudogene', other_designations='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase like 2 pseudogene', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106070481, end_position_on_the_genomic_accession=106073199, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  134997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=134997, CurrentID=0, Status='live', Symbol='PPIAP29', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 29', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24976377, end_position_on_the_genomic_accession=24977187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928659, CurrentID=0, Status='live', Symbol='LOC101928659', Aliases='', description='uncharacterized LOC101928659', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87317509, end_position_on_the_genomic_accession=87318038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874190, CurrentID=0, Status='live', Symbol='SAMSN1-AS1', Aliases='', description='SAMSN1 antisense RNA 1', other_designations='SAMSN1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14582202, end_position_on_the_genomic_accession=14598303, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  347862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347862, CurrentID=0, Status='live', Symbol='PDDC1', Aliases='', description='Parkinson disease 7 domain containing 1', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=767217, end_position_on_the_genomic_accession=777501, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928663, CurrentID=0, Status='live', Symbol='LOC101928663', Aliases='', description='uncharacterized LOC101928663', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25244916, end_position_on_the_genomic_accession=25261407, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874191, CurrentID=0, Status='live', Symbol='RANP2', Aliases='', description='RAN, member RAS oncogene family pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147167340, end_position_on_the_genomic_accession=147167776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928668, CurrentID=0, Status='live', Symbol='LOC101928668', Aliases='', description='uncharacterized LOC101928668', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58521311, end_position_on_the_genomic_accession=58523371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928669, CurrentID=0, Status='live', Symbol='LOC101928669', Aliases='', description='uncharacterized LOC101928669', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928671, CurrentID=0, Status='live', Symbol='LOC101928671', Aliases='', description='uncharacterized LOC101928671', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928673, CurrentID=0, Status='live', Symbol='LOC101928673', Aliases='', description='uncharacterized LOC101928673', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173417789, end_position_on_the_genomic_accession=173461362, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928674, CurrentID=0, Status='live', Symbol='LOC101928674', Aliases='', description='uncharacterized LOC101928674', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78360448, end_position_on_the_genomic_accession=78373909, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105374331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374331, CurrentID=0, Status='live', Symbol='LOC105374331', Aliases='', description='uncharacterized LOC105374331', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25685019, end_position_on_the_genomic_accession=25690713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928677, CurrentID=0, Status='live', Symbol='LOC101928677', Aliases='', description='uncharacterized LOC101928677', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135252012, end_position_on_the_genomic_accession=135254496, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928678, CurrentID=0, Status='live', Symbol='LOC101928678', Aliases='', description='uncharacterized LOC101928678', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15737773, end_position_on_the_genomic_accession=15850013, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874193, CurrentID=0, Status='live', Symbol='RANP9', Aliases='', description='RAN, member RAS oncogene family pseudogene 9', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33039523, end_position_on_the_genomic_accession=33039793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928682, CurrentID=0, Status='live', Symbol='LOC101928682', Aliases='', description='uncharacterized LOC101928682', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87271906, end_position_on_the_genomic_accession=87292452, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928684, CurrentID=0, Status='live', Symbol='LINC01209', Aliases='', description='long intergenic non-protein coding RNA 1209', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176814155, end_position_on_the_genomic_accession=176817001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928685, CurrentID=0, Status='live', Symbol='NACAP3', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 3', other_designations='nascent-polypeptide-associated complex alpha polypeptide pseudogene 3', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93123947, end_position_on_the_genomic_accession=93128532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928687, CurrentID=0, Status='live', Symbol='LINC01468', Aliases='lnc-MBL2-4', description='long intergenic non-protein coding RNA 1468', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52450877, end_position_on_the_genomic_accession=52470533, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928688, CurrentID=0, Status='live', Symbol='LOC101928688', Aliases='', description='uncharacterized LOC101928688', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39816958, end_position_on_the_genomic_accession=39899185, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101928689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928689, CurrentID=0, Status='live', Symbol='LINC01424', Aliases='', description='long intergenic non-protein coding RNA 1424', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44802241, end_position_on_the_genomic_accession=44804717, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874195, CurrentID=0, Status='live', Symbol='DIAPH3-AS1', Aliases='', description='DIAPH3 antisense RNA 1', other_designations='DIAPH3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60012718, end_position_on_the_genomic_accession=60044357, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928692, CurrentID=0, Status='live', Symbol='LOC101928692', Aliases='', description='uncharacterized LOC101928692', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55488960, end_position_on_the_genomic_accession=55491513, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374334, CurrentID=0, Status='live', Symbol='LOC105374334', Aliases='', description='uncharacterized LOC105374334', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26298656, end_position_on_the_genomic_accession=26310607, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928694, CurrentID=0, Status='live', Symbol='LOC101928694', Aliases='', description='uncharacterized LOC101928694', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58434901, end_position_on_the_genomic_accession=58498735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928696, CurrentID=0, Status='live', Symbol='LOC101928696', Aliases='', description='uncharacterized LOC101928696', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174934947, end_position_on_the_genomic_accession=174954261, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874196, CurrentID=0, Status='live', Symbol='DIAPH3-AS2', Aliases='DIPAH3-AS2', description='DIAPH3 antisense RNA 2', other_designations='DIAPH3 antisense RNA 2 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60144698, end_position_on_the_genomic_accession=60153505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928698, CurrentID=0, Status='live', Symbol='LINC01070', Aliases='', description='long intergenic non-protein coding RNA 1070', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112197333, end_position_on_the_genomic_accession=112201000, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928700, CurrentID=0, Status='live', Symbol='LOC101928700', Aliases='', description='uncharacterized LOC101928700', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147080938, end_position_on_the_genomic_accession=147097609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928701, CurrentID=0, Status='live', Symbol='LOC101928701', Aliases='', description='uncharacterized LOC101928701', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16524830, end_position_on_the_genomic_accession=16539768, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347902, CurrentID=0, Status='live', Symbol='AMIGO2', Aliases='ALI1, AMIGO-2, DEGA', description='adhesion molecule with Ig-like domain 2', other_designations='alivin 1|amphoterin induced gene 2|differentially expressed in gastric adenocarcinoma|differentially expressed in gastric adenocarcinomas|transmembrane protein AMIGO2', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47075707, end_position_on_the_genomic_accession=47080826, orientation='minus', exon_count=3, OMIM=615690, no_hdr0=''),
  100874197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874197, CurrentID=0, Status='live', Symbol='EFCAB6-AS1', Aliases='', description='EFCAB6 antisense RNA 1', other_designations='CITF22-123F2.1|EFCAB6 antisense RNA 1 (non-protein coding)', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43516254, end_position_on_the_genomic_accession=43536605, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928704, CurrentID=0, Status='live', Symbol='LOC101928704', Aliases='', description='uncharacterized LOC101928704', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43556696, end_position_on_the_genomic_accession=43674197, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928705, CurrentID=0, Status='live', Symbol='LOC101928705', Aliases='', description='uncharacterized LOC101928705', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28566432, end_position_on_the_genomic_accession=28581505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100093698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093698, CurrentID=0, Status='live', Symbol='GS1-600G8.3', Aliases='', description='unknown transcript', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13310652, end_position_on_the_genomic_accession=13319933, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  101928707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928707, CurrentID=0, Status='live', Symbol='LOC101928707', Aliases='', description='uncharacterized LOC101928707', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30355395, end_position_on_the_genomic_accession=30357108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928708, CurrentID=0, Status='live', Symbol='LOC101928708', Aliases='', description='uncharacterized LOC101928708', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87212115, end_position_on_the_genomic_accession=87226430, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874198, CurrentID=0, Status='live', Symbol='SHANK2-AS1', Aliases='', description='SHANK2 antisense RNA 1', other_designations='SHANK2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70631094, end_position_on_the_genomic_accession=70635490, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928710, CurrentID=0, Status='live', Symbol='LOC101928710', Aliases='', description='uncharacterized LOC101928710', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78617384, end_position_on_the_genomic_accession=78632057, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374337, CurrentID=0, Status='live', Symbol='LOC105374337', Aliases='', description='uncharacterized LOC105374337', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26950441, end_position_on_the_genomic_accession=26965072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928714, CurrentID=0, Status='live', Symbol='LOC101928714', Aliases='', description='uncharacterized LOC101928714', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80453157, end_position_on_the_genomic_accession=80469088, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928715, CurrentID=0, Status='live', Symbol='TPT1P10', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 10', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58211967, end_position_on_the_genomic_accession=58213185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928716, CurrentID=0, Status='live', Symbol='LOC101928716', Aliases='', description='uncharacterized LOC101928716', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40538006, end_position_on_the_genomic_accession=40547158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  347918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347918, CurrentID=0, Status='live', Symbol='EP400NL', Aliases='', description='EP400 N-terminal like', other_designations='E1A binding protein p400 pseudogene', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132084283, end_position_on_the_genomic_accession=132126340, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101928719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928719, CurrentID=0, Status='live', Symbol='LOC101928719', Aliases='', description='uncharacterized LOC101928719', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50199088, end_position_on_the_genomic_accession=50200884, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928720, CurrentID=0, Status='live', Symbol='LINC01509', Aliases='PRCAT37', description='long intergenic non-protein coding RNA 1509', other_designations='prostate cancer-associated transcript 37', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107420284, end_position_on_the_genomic_accession=107466585, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100874200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874200, CurrentID=0, Status='live', Symbol='ARHGEF3-AS1', Aliases='', description='ARHGEF3 antisense RNA 1', other_designations='ARHGEF3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56940040, end_position_on_the_genomic_accession=56960854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928722, CurrentID=0, Status='live', Symbol='LOC101928722', Aliases='', description='nascent polypeptide-associated complex subunit alpha, muscle-specific form-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136252193, end_position_on_the_genomic_accession=136266525, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374339, CurrentID=0, Status='live', Symbol='LOC105374339', Aliases='', description='uncharacterized LOC105374339', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=787341, end_position_on_the_genomic_accession=796440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100093716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093716, CurrentID=0, Status='live', Symbol='DFNB75', Aliases='', description='deafness, autosomal recessive 75', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928725, CurrentID=0, Status='live', Symbol='LOC101928725', Aliases='', description='uncharacterized LOC101928725', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58768072, end_position_on_the_genomic_accession=58771114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928726, CurrentID=0, Status='live', Symbol='LOC101928726', Aliases='', description='uncharacterized LOC101928726', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177108798, end_position_on_the_genomic_accession=177121834, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874201, CurrentID=0, Status='live', Symbol='B4GALT4-AS1', Aliases='', description='B4GALT4 antisense RNA 1', other_designations='B4GALT4 antisense RNA 1 (non-protein coding)', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119226486, end_position_on_the_genomic_accession=119235141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928728, CurrentID=0, Status='live', Symbol='LOC101928728', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26692063, end_position_on_the_genomic_accession=26696405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928730, CurrentID=0, Status='live', Symbol='LOC101928730', Aliases='', description='uncharacterized LOC101928730', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112256725, end_position_on_the_genomic_accession=112273000, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928731, CurrentID=0, Status='live', Symbol='LOC101928731', Aliases='', description='uncharacterized LOC101928731', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93707791, end_position_on_the_genomic_accession=93737823, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  347933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347933, CurrentID=0, Status='live', Symbol='NLRP9P', Aliases='NOD25', description='NLR family, pyrin domain containing 9 pseudogene', other_designations='NOD25 pseudogene', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129013284, end_position_on_the_genomic_accession=129017166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928734, CurrentID=0, Status='live', Symbol='WWC2-AS1', Aliases='', description='WWC2 antisense RNA 1', other_designations='WWC2 antisense RNA 1 (non-protein coding)', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183233628, end_position_on_the_genomic_accession=183240634, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928735, CurrentID=0, Status='live', Symbol='LOC101928735', Aliases='', description='uncharacterized LOC101928735', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29040670, end_position_on_the_genomic_accession=29056711, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928736, CurrentID=0, Status='live', Symbol='MIR762HG', Aliases='', description='MIR762 host gene', other_designations='MIR762 host gene (non-protein coding)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30875460, end_position_on_the_genomic_accession=30895220, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928737, CurrentID=0, Status='live', Symbol='LOC101928737', Aliases='', description='uncharacterized LOC101928737', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87494187, end_position_on_the_genomic_accession=87515630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928738, CurrentID=0, Status='live', Symbol='LOC101928738', Aliases='', description='uncharacterized LOC101928738', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79919357, end_position_on_the_genomic_accession=79924010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928739, CurrentID=0, Status='live', Symbol='LOC101928739', Aliases='', description='uncharacterized LOC101928739', other_designations='', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179101371, end_position_on_the_genomic_accession=179147973, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928741, CurrentID=0, Status='live', Symbol='LOC101928741', Aliases='', description='uncharacterized LOC101928741', other_designations='CTD-2016O11.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66205123, end_position_on_the_genomic_accession=66209906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928742, CurrentID=0, Status='live', Symbol='LOC101928742', Aliases='', description='uncharacterized LOC101928742', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81491439, end_position_on_the_genomic_accession=81513904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928743, CurrentID=0, Status='live', Symbol='LOC101928743', Aliases='', description='uncharacterized LOC101928743', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26324698, end_position_on_the_genomic_accession=26332124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928744, CurrentID=0, Status='live', Symbol='LINC01450', Aliases='', description='long intergenic non-protein coding RNA 1450', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40964678, end_position_on_the_genomic_accession=40979939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928745, CurrentID=0, Status='live', Symbol='LL21NC02-21A1.1', Aliases='', description='uncharacterized LOC101928745', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45403806, end_position_on_the_genomic_accession=45405442, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928747, CurrentID=0, Status='live', Symbol='LOC101928747', Aliases='', description='uncharacterized LOC101928747', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136874661, end_position_on_the_genomic_accession=136879312, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928748, CurrentID=0, Status='live', Symbol='LOC101928748', Aliases='', description='uncharacterized LOC101928748', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115119539, end_position_on_the_genomic_accession=115138409, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874205, CurrentID=0, Status='live', Symbol='ATP11A-AS1', Aliases='', description='ATP11A antisense RNA 1', other_designations='ATP11A antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112745449, end_position_on_the_genomic_accession=112754731, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928752, CurrentID=0, Status='live', Symbol='LINC01043', Aliases='', description='long intergenic non-protein coding RNA 1043', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112313058, end_position_on_the_genomic_accession=112321766, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371106, CurrentID=0, Status='live', Symbol='LOC105371106', Aliases='', description='uncharacterized LOC105371106', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17471011, end_position_on_the_genomic_accession=17483616, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928755, CurrentID=0, Status='live', Symbol='LOC101928755', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55743709, end_position_on_the_genomic_accession=55746589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874206, CurrentID=0, Status='live', Symbol='OSBPL10-AS1', Aliases='', description='OSBPL10 antisense RNA 1', other_designations='OSBPL10 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31704213, end_position_on_the_genomic_accession=31721586, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131465, CurrentID=0, Status='live', Symbol='LOC100131465', Aliases='', description='peroxisomal biogenesis factor 5 pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247183816, end_position_on_the_genomic_accession=247185478, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928760, CurrentID=0, Status='live', Symbol='LOC101928760', Aliases='', description='uncharacterized LOC101928760', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56229634, end_position_on_the_genomic_accession=56246446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928762, CurrentID=0, Status='live', Symbol='LOC101928762', Aliases='', description='uncharacterized LOC101928762', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31041932, end_position_on_the_genomic_accession=31049816, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874207, CurrentID=0, Status='live', Symbol='ST3GAL6-AS1', Aliases='', description='ST3GAL6 antisense RNA 1', other_designations='ST3GAL6 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98714333, end_position_on_the_genomic_accession=98732651, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928764, CurrentID=0, Status='live', Symbol='LOC101928764', Aliases='', description='uncharacterized LOC101928764', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21287321, end_position_on_the_genomic_accession=21298115, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928765, CurrentID=0, Status='live', Symbol='LOC101928765', Aliases='', description='uncharacterized LOC101928765', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=228475164, end_position_on_the_genomic_accession=228744681, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928766, CurrentID=0, Status='live', Symbol='LOC101928766', Aliases='', description='uncharacterized LOC101928766', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79915252, end_position_on_the_genomic_accession=79926725, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928767, CurrentID=0, Status='live', Symbol='LOC101928767', Aliases='', description='uncharacterized LOC101928767', other_designations='CTD-2341M24.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85934678, end_position_on_the_genomic_accession=86129778, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85824, CurrentID=0, Status='live', Symbol='RPL12P12', Aliases='RPL12L3, RPL12_16_1689, dJ999L4.1', description='ribosomal protein L12 pseudogene 12', other_designations='ribosomal protein L12-like 3|ribosomal protein S12 pseudogene 12', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19823407, end_position_on_the_genomic_accession=19824031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874208, CurrentID=0, Status='live', Symbol='NDFIP2-AS1', Aliases='', description='NDFIP2 antisense RNA 1', other_designations='NDFIP2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79477364, end_position_on_the_genomic_accession=79481231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928770, CurrentID=0, Status='live', Symbol='LINC01526', Aliases='Lnc-FAM46A-1', description='long intergenic non-protein coding RNA 1526', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81813286, end_position_on_the_genomic_accession=81814192, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374347, CurrentID=0, Status='live', Symbol='LOC105374347', Aliases='', description='uncharacterized LOC105374347', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1285614, end_position_on_the_genomic_accession=1289704, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928773, CurrentID=0, Status='live', Symbol='LINC01449', Aliases='', description='long intergenic non-protein coding RNA 1449', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41101604, end_position_on_the_genomic_accession=41133501, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928775, CurrentID=0, Status='live', Symbol='LOC101928775', Aliases='', description='uncharacterized LOC101928775', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115739670, end_position_on_the_genomic_accession=115744239, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928776, CurrentID=0, Status='live', Symbol='LINC01258', Aliases='', description='long intergenic non-protein coding RNA 1258', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38420662, end_position_on_the_genomic_accession=38523180, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928777, CurrentID=0, Status='live', Symbol='LOC101928777', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132080125, end_position_on_the_genomic_accession=132085413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928778, CurrentID=0, Status='live', Symbol='LOC101928778', Aliases='', description='uncharacterized LOC101928778', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177700524, end_position_on_the_genomic_accession=177710330, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928780, CurrentID=0, Status='live', Symbol='LOC101928780', Aliases='', description='uncharacterized LOC101928780', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54119511, end_position_on_the_genomic_accession=54125343, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928781, CurrentID=0, Status='live', Symbol='TMEM26-AS1', Aliases='', description='TMEM26 antisense RNA 1', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61452639, end_position_on_the_genomic_accession=61493433, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928782, CurrentID=0, Status='live', Symbol='LOC101928782', Aliases='', description='uncharacterized LOC101928782', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131910220, end_position_on_the_genomic_accession=131948953, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374349, CurrentID=0, Status='live', Symbol='LOC105374349', Aliases='', description='uncharacterized LOC105374349', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1627370, end_position_on_the_genomic_accession=1629063, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928784, CurrentID=0, Status='live', Symbol='RORA-AS1', Aliases='', description='RORA antisense RNA 1', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60479178, end_position_on_the_genomic_accession=60630639, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101928786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928786, CurrentID=0, Status='live', Symbol='LOC101928786', Aliases='', description='uncharacterized LOC101928786', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138040503, end_position_on_the_genomic_accession=138045071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874211, CurrentID=0, Status='live', Symbol='MID1IP1-AS1', Aliases='', description='MID1IP1 antisense RNA 1', other_designations='MID1IP1 antisense RNA 1 (non-protein coding)', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38801248, end_position_on_the_genomic_accession=38803883, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928788, CurrentID=0, Status='live', Symbol='LOC101928788', Aliases='', description='ADP-ribosylation factor-like protein 4A', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150934732, end_position_on_the_genomic_accession=150935753, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131470, CurrentID=0, Status='live', Symbol='LOC100131470', Aliases='', description='uncharacterized LOC100131470', other_designations='', map_location='4q33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=171139473, end_position_on_the_genomic_accession=171151532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928790, CurrentID=0, Status='live', Symbol='PEX5L-AS2', Aliases='', description='PEX5L antisense RNA 2', other_designations='PEX5L antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179898194, end_position_on_the_genomic_accession=179921899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928791, CurrentID=0, Status='live', Symbol='LOC101928791', Aliases='', description='uncharacterized LOC101928791', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88018605, end_position_on_the_genomic_accession=88026312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928792, CurrentID=0, Status='live', Symbol='LOC101928792', Aliases='', description='uncharacterized LOC101928792', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24164163, end_position_on_the_genomic_accession=24357981, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  100874212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874212, CurrentID=0, Status='live', Symbol='MYCBP2-AS1', Aliases='', description='MYCBP2 antisense RNA 1', other_designations='MYCBP2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77075514, end_position_on_the_genomic_accession=77087776, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928794, CurrentID=0, Status='live', Symbol='LOC101928794', Aliases='', description='uncharacterized LOC101928794', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67001618, end_position_on_the_genomic_accession=67004154, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928795, CurrentID=0, Status='live', Symbol='LINC01448', Aliases='', description='long intergenic non-protein coding RNA 1448', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42661726, end_position_on_the_genomic_accession=42706447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928796, CurrentID=0, Status='live', Symbol='LOC101928796', Aliases='', description='uncharacterized LOC101928796', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45972970, end_position_on_the_genomic_accession=45974953, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928797, CurrentID=0, Status='live', Symbol='LOC101928797', Aliases='', description='uncharacterized LOC101928797', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117648606, end_position_on_the_genomic_accession=117657027, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874213, CurrentID=0, Status='live', Symbol='MYCBP2-AS2', Aliases='', description='MYCBP2 antisense RNA 2', other_designations='MYCBP2 antisense RNA 2 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131472, CurrentID=0, Status='live', Symbol='LOC100131472', Aliases='', description='uncharacterized LOC100131472', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10702672, end_position_on_the_genomic_accession=10707538, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928804, CurrentID=0, Status='live', Symbol='LOC101928804', Aliases='', description='uncharacterized LOC101928804', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54769422, end_position_on_the_genomic_accession=54771064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874214, CurrentID=0, Status='live', Symbol='TM4SF19-AS1', Aliases='', description='TM4SF19 antisense RNA 1', other_designations='TM4SF19 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196318332, end_position_on_the_genomic_accession=196325570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928807, CurrentID=0, Status='live', Symbol='LOC101928807', Aliases='', description='uncharacterized LOC101928807', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132260827, end_position_on_the_genomic_accession=132264224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928808, CurrentID=0, Status='live', Symbol='LOC101928808', Aliases='', description='uncharacterized LOC101928808', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146394682, end_position_on_the_genomic_accession=146401858, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85865, CurrentID=0, Status='live', Symbol='GTPBP10', Aliases='ObgH2, UG0751c10', description='GTP-binding protein 10 (putative)', other_designations='protein obg homolog 2', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90346665, end_position_on_the_genomic_accession=90391455, orientation='plus', exon_count=10, OMIM=610920, no_hdr0=''),
  101928810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928810, CurrentID=0, Status='live', Symbol='LOC101928810', Aliases='', description='uncharacterized LOC101928810', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184894149, end_position_on_the_genomic_accession=184963549, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928811, CurrentID=0, Status='live', Symbol='LOC101928811', Aliases='', description='uncharacterized LOC101928811', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7616946, end_position_on_the_genomic_accession=7620622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928812, CurrentID=0, Status='live', Symbol='LOC101928812', Aliases='', description='uncharacterized LOC101928812', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45355371, end_position_on_the_genomic_accession=45356657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  348013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348013, CurrentID=0, Status='live', Symbol='TMEM255B', Aliases='FAM70B', description='transmembrane protein 255B', other_designations='family with sequence similarity 70, member B|protein FAM70B', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113759210, end_position_on_the_genomic_accession=113813797, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  101928814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928814, CurrentID=0, Status='live', Symbol='LOC101928814', Aliases='', description='uncharacterized LOC101928814', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48996585, end_position_on_the_genomic_accession=48999783, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928815, CurrentID=0, Status='live', Symbol='LOC101928815', Aliases='', description='uncharacterized LOC101928815', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56653359, end_position_on_the_genomic_accession=56665018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928816, CurrentID=0, Status='live', Symbol='LOC101928816', Aliases='', description='uncharacterized LOC101928816', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230443107, end_position_on_the_genomic_accession=230521299, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874216, CurrentID=0, Status='live', Symbol='ZNF385D-AS1', Aliases='', description='ZNF385D antisense RNA 1', other_designations='ZNF385D antisense RNA 1 (non-protein coding)', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21542789, end_position_on_the_genomic_accession=21579959, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928818, CurrentID=0, Status='live', Symbol='LOC101928818', Aliases='', description='uncharacterized LOC101928818', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24358024, end_position_on_the_genomic_accession=24412219, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101928819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928819, CurrentID=0, Status='live', Symbol='LOC101928819', Aliases='', description='uncharacterized LOC101928819', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67463082, end_position_on_the_genomic_accession=67492538, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928820, CurrentID=0, Status='live', Symbol='LOC101928820', Aliases='', description='uncharacterized LOC101928820', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85387219, end_position_on_the_genomic_accession=85390364, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  348021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348021, CurrentID=0, Status='live', Symbol='LINC00442', Aliases='', description='long intergenic non-protein coding RNA 442', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19008259, end_position_on_the_genomic_accession=19012634, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874217, CurrentID=0, Status='live', Symbol='TBL1XR1-AS1', Aliases='', description='TBL1XR1 antisense RNA 1', other_designations='TBL1XR1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928824, CurrentID=0, Status='live', Symbol='LOC101928824', Aliases='', description='uncharacterized LOC101928824', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20495966, end_position_on_the_genomic_accession=20500835, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928827, CurrentID=0, Status='live', Symbol='TCEB1P33', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 33', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39932502, end_position_on_the_genomic_accession=39933591, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927502, CurrentID=0, Status='live', Symbol='LOC101927502', Aliases='', description='uncharacterized LOC101927502', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81689713, end_position_on_the_genomic_accession=81776900, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874218, CurrentID=0, Status='live', Symbol='ATP13A5-AS1', Aliases='', description='ATP13A5 antisense RNA 1', other_designations='ATP13A5 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193307244, end_position_on_the_genomic_accession=193314521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928832, CurrentID=0, Status='live', Symbol='LOC101928832', Aliases='', description='uncharacterized LOC101928832', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146941756, end_position_on_the_genomic_accession=146954196, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928833, CurrentID=0, Status='live', Symbol='LOC101928833', Aliases='', description='uncharacterized LOC101928833', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140681106, end_position_on_the_genomic_accession=140687834, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928834, CurrentID=0, Status='live', Symbol='LOC101928834', Aliases='', description='uncharacterized LOC101928834', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19710335, end_position_on_the_genomic_accession=19728550, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  348035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348035, CurrentID=0, Status='live', Symbol='MGC40069', Aliases='', description='uncharacterized protein MGC40069', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928836, CurrentID=0, Status='live', Symbol='LOC101928836', Aliases='', description='uncharacterized LOC101928836', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7440567, end_position_on_the_genomic_accession=7441592, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374358, CurrentID=0, Status='live', Symbol='LOC105374358', Aliases='', description='uncharacterized LOC105374358', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3956372, end_position_on_the_genomic_accession=3963889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928840, CurrentID=0, Status='live', Symbol='PWRN3', Aliases='', description='Prader-Willi region non-protein coding RNA 3', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24441127, end_position_on_the_genomic_accession=24447967, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928841, CurrentID=0, Status='live', Symbol='LOC101928841', Aliases='', description='collagen alpha-1(II) chain-like', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113399610, end_position_on_the_genomic_accession=113420858, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928842, CurrentID=0, Status='live', Symbol='LOC101928842', Aliases='', description='uncharacterized LOC101928842', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85941210, end_position_on_the_genomic_accession=85998671, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105374359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374359, CurrentID=0, Status='live', Symbol='LOC105374359', Aliases='', description='uncharacterized LOC105374359', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3988581, end_position_on_the_genomic_accession=4006568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928844, CurrentID=0, Status='live', Symbol='LOC101928844', Aliases='', description='uncharacterized LOC101928844', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5847467, end_position_on_the_genomic_accession=5858239, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928845, CurrentID=0, Status='live', Symbol='LOC101928845', Aliases='', description='uncharacterized LOC101928845', other_designations='CTC-548K16.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14333742, end_position_on_the_genomic_accession=14343913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928846, CurrentID=0, Status='live', Symbol='HIPK1-AS1', Aliases='', description='HIPK1 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113924001, end_position_on_the_genomic_accession=113929525, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874221, CurrentID=0, Status='live', Symbol='ZNF385D-AS2', Aliases='', description='ZNF385D antisense RNA 2', other_designations='ZNF385D antisense RNA 2 (non-protein coding)', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21957139, end_position_on_the_genomic_accession=21979828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928848, CurrentID=0, Status='live', Symbol='LOC101928848', Aliases='', description='uncharacterized LOC101928848', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49512843, end_position_on_the_genomic_accession=49536075, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928849, CurrentID=0, Status='live', Symbol='LOC101928849', Aliases='', description='uncharacterized LOC101928849', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118852061, end_position_on_the_genomic_accession=118980425, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928850, CurrentID=0, Status='live', Symbol='LOC101928850', Aliases='', description='uncharacterized LOC101928850', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61598514, end_position_on_the_genomic_accession=61605241, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928851, CurrentID=0, Status='live', Symbol='LOC101928851', Aliases='', description='uncharacterized LOC101928851', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57425872, end_position_on_the_genomic_accession=57465986, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  79854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79854, CurrentID=0, Status='live', Symbol='LINC00115', Aliases='NCRNA00115', description='long intergenic non-protein coding RNA 115', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=826206, end_position_on_the_genomic_accession=827522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374361, CurrentID=0, Status='live', Symbol='LOC105374361', Aliases='', description='uncharacterized LOC105374361', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5004561, end_position_on_the_genomic_accession=5034820, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928856, CurrentID=0, Status='live', Symbol='LOC101928856', Aliases='', description='uncharacterized LOC101928856', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180456643, end_position_on_the_genomic_accession=180602102, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  101928857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928857, CurrentID=0, Status='live', Symbol='CTD-3080P12.3', Aliases='', description='uncharacterized LOC101928857', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1173096, end_position_on_the_genomic_accession=1178605, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928858, CurrentID=0, Status='live', Symbol='LOC101928858', Aliases='', description='uncharacterized LOC101928858', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67793213, end_position_on_the_genomic_accession=67801058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928859, CurrentID=0, Status='live', Symbol='LOC101928859', Aliases='', description='uncharacterized LOC101928859', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63859646, end_position_on_the_genomic_accession=63887311, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928860, CurrentID=0, Status='live', Symbol='LOC101928860', Aliases='', description='uncharacterized LOC101928860', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43664137, end_position_on_the_genomic_accession=43667674, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928861, CurrentID=0, Status='live', Symbol='LOC101928861', Aliases='', description='uncharacterized LOC101928861', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133800493, end_position_on_the_genomic_accession=133824888, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928863, CurrentID=0, Status='live', Symbol='LOC101928863', Aliases='', description='uncharacterized LOC101928863', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147280989, end_position_on_the_genomic_accession=147286475, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928864, CurrentID=0, Status='live', Symbol='LOC101928864', Aliases='', description='uncharacterized LOC101928864', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77154775, end_position_on_the_genomic_accession=77156667, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874224, CurrentID=0, Status='live', Symbol='KRTAP10-13P', Aliases='', description='keratin associated protein 10-13, pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44702034, end_position_on_the_genomic_accession=44702801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928866, CurrentID=0, Status='live', Symbol='LOC101928866', Aliases='', description='uncharacterized LOC101928866', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178506136, end_position_on_the_genomic_accession=178513028, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130157, CurrentID=0, Status='live', Symbol='LOC100130157', Aliases='', description='uncharacterized LOC100130157', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=44074071, end_position_on_the_genomic_accession=44077220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928868, CurrentID=0, Status='live', Symbol='LOC101928868', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154510097, end_position_on_the_genomic_accession=154522300, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928869, CurrentID=0, Status='live', Symbol='GABRG3-AS1', Aliases='', description='GABRG3 antisense RNA 1', other_designations='', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27157735, end_position_on_the_genomic_accession=27161319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928870, CurrentID=0, Status='live', Symbol='LOC101928870', Aliases='', description='classical arabinogalactan protein 9-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87152459, end_position_on_the_genomic_accession=87155589, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874225, CurrentID=0, Status='live', Symbol='RPP40P1', Aliases='', description='ribonuclease P/MRP 40kDa subunit pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=25591422, end_position_on_the_genomic_accession=25591817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928872, CurrentID=0, Status='live', Symbol='LOC101928872', Aliases='', description='uncharacterized LOC101928872', other_designations='CTC-232P5.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6066876, end_position_on_the_genomic_accession=6077118, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928875, CurrentID=0, Status='live', Symbol='LOC101928875', Aliases='', description='uncharacterized LOC101928875', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114158509, end_position_on_the_genomic_accession=114174468, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928876, CurrentID=0, Status='live', Symbol='LOC101928876', Aliases='', description='uncharacterized LOC101928876', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121453966, end_position_on_the_genomic_accession=121463330, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874226, CurrentID=0, Status='live', Symbol='ARHGEF7-AS1', Aliases='', description='ARHGEF7 antisense RNA 1', other_designations='ARHGEF7 antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111144305, end_position_on_the_genomic_accession=111148565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928879, CurrentID=0, Status='live', Symbol='LOC101928879', Aliases='', description='COMM domain-containing protein 5-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53575693, end_position_on_the_genomic_accession=53576543, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374246, CurrentID=0, Status='live', Symbol='LOC105374246', Aliases='', description='uncharacterized LOC105374246', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182980942, end_position_on_the_genomic_accession=182986729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928881, CurrentID=0, Status='live', Symbol='LOC101928881', Aliases='', description='uncharacterized LOC101928881', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233012624, end_position_on_the_genomic_accession=233015885, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928882, CurrentID=0, Status='live', Symbol='LOC101928882', Aliases='', description='uncharacterized LOC101928882', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180707589, end_position_on_the_genomic_accession=180870178, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101928885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928885, CurrentID=0, Status='live', Symbol='LOC101928885', Aliases='', description='uncharacterized LOC101928885', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68967741, end_position_on_the_genomic_accession=69029927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928886, CurrentID=0, Status='live', Symbol='LOC101928886', Aliases='', description='uncharacterized LOC101928886', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56066684, end_position_on_the_genomic_accession=56078801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928887, CurrentID=0, Status='live', Symbol='LOC101928887', Aliases='', description='uncharacterized LOC101928887', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64901136, end_position_on_the_genomic_accession=64924518, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928890, CurrentID=0, Status='live', Symbol='LOC101928890', Aliases='', description='uncharacterized LOC101928890', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114401443, end_position_on_the_genomic_accession=114411092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928891, CurrentID=0, Status='live', Symbol='LOC101928891', Aliases='', description='uncharacterized LOC101928891', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20957092, end_position_on_the_genomic_accession=20964680, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928892, CurrentID=0, Status='live', Symbol='LOC101928892', Aliases='', description='nascent polypeptide-associated complex subunit alpha, muscle-specific form-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  348093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348093, CurrentID=0, Status='live', Symbol='RBPMS2', Aliases='', description='RNA binding protein with multiple splicing 2', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64739894, end_position_on_the_genomic_accession=64775571, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  348094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348094, CurrentID=0, Status='live', Symbol='ANKDD1A', Aliases='', description='ankyrin repeat and death domain containing 1A', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64911902, end_position_on_the_genomic_accession=64958700, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100874229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874229, CurrentID=0, Status='live', Symbol='DCUN1D2-AS', Aliases='DCUN1D2-AS2', description='DCUN1D2 antisense RNA', other_designations='DCUN1D2 antisense RNA 2 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928896, CurrentID=0, Status='live', Symbol='LOC101928896', Aliases='', description='uncharacterized LOC101928896', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=78533176, end_position_on_the_genomic_accession=78558565, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129322, CurrentID=0, Status='live', Symbol='LOC100129322', Aliases='', description='uncharacterized LOC100129322', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=11459202, end_position_on_the_genomic_accession=11459636, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928902, CurrentID=0, Status='live', Symbol='LOC101928902', Aliases='', description='uncharacterized LOC101928902', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144368937, end_position_on_the_genomic_accession=144372465, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374369, CurrentID=0, Status='live', Symbol='LOC105374369', Aliases='', description='uncharacterized LOC105374369', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27942455, end_position_on_the_genomic_accession=27943504, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928904, CurrentID=0, Status='live', Symbol='LOC101928904', Aliases='', description='nucleophosmin-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21112775, end_position_on_the_genomic_accession=21113703, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928906, CurrentID=0, Status='live', Symbol='LOC101928906', Aliases='', description='uncharacterized LOC101928906', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41651666, end_position_on_the_genomic_accession=41655197, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874231, CurrentID=0, Status='live', Symbol='SPATA13-AS1', Aliases='', description='SPATA13 antisense RNA 1', other_designations='SPATA13 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24252749, end_position_on_the_genomic_accession=24254439, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928908, CurrentID=0, Status='live', Symbol='LOC101928908', Aliases='', description='60S ribosomal protein L31-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172523242, end_position_on_the_genomic_accession=172555223, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928909, CurrentID=0, Status='live', Symbol='LOC101928909', Aliases='', description='uncharacterized LOC101928909', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90642629, end_position_on_the_genomic_accession=90648891, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  348110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348110, CurrentID=0, Status='live', Symbol='ARPIN', Aliases='C15orf38', description='actin-related protein 2/3 complex inhibitor', other_designations='UPF0552 protein C15orf38|arp2/3 inhibition protein', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89896532, end_position_on_the_genomic_accession=89912990, orientation='minus', exon_count=7, OMIM=615543, no_hdr0=''),
  101928911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928911, CurrentID=0, Status='live', Symbol='LOC101928911', Aliases='', description='uncharacterized LOC101928911', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87784860, end_position_on_the_genomic_accession=87911731, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928912, CurrentID=0, Status='live', Symbol='LOC101928912', Aliases='', description='uncharacterized LOC101928912', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97229811, end_position_on_the_genomic_accession=97257412, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928913, CurrentID=0, Status='live', Symbol='LINC01515', Aliases='', description='long intergenic non-protein coding RNA 1515', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65571425, end_position_on_the_genomic_accession=65766541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928915, CurrentID=0, Status='live', Symbol='LOC101928915', Aliases='', description='uncharacterized LOC101928915', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135774948, end_position_on_the_genomic_accession=135932476, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928917, CurrentID=0, Status='live', Symbol='LOC101928917', Aliases='', description='heat shock transcription factor, X-linked-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149547679, end_position_on_the_genomic_accession=149549870, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874233, CurrentID=0, Status='live', Symbol='ARHGEF19-AS1', Aliases='', description='ARHGEF19 antisense RNA 1', other_designations='ARHGEF19 antisense RNA 1 (non-protein coding)', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  348120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348120, CurrentID=0, Status='live', Symbol='LINC01193', Aliases='CT60', description='long intergenic non-protein coding RNA 1193', other_designations='cancer/testis antigen 60 (non-protein coding)', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20940438, end_position_on_the_genomic_accession=20993304, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131492, CurrentID=0, Status='live', Symbol='LOC100131492', Aliases='', description='PP12901', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=128206448, end_position_on_the_genomic_accession=128211783, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928922, CurrentID=0, Status='live', Symbol='LINC00463', Aliases='AMER2-AS1', description='long intergenic non-protein coding RNA 463', other_designations='AMER2 antisense RNA 1 (head to head)|TCONS_l2_00007499', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25172824, end_position_on_the_genomic_accession=25180001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928923, CurrentID=0, Status='live', Symbol='LOC101928923', Aliases='', description='uncharacterized LOC101928923', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155808725, end_position_on_the_genomic_accession=156373758, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928924, CurrentID=0, Status='live', Symbol='LOC101928924', Aliases='', description='uncharacterized LOC101928924', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69477484, end_position_on_the_genomic_accession=69485612, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874234, CurrentID=0, Status='live', Symbol='CACNA1C-AS4', Aliases='', description='CACNA1C antisense RNA 4', other_designations='CACNA1C antisense RNA 4 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2220537, end_position_on_the_genomic_accession=2223481, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928926, CurrentID=0, Status='live', Symbol='LOC101928926', Aliases='', description='uncharacterized LOC101928926', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19763517, end_position_on_the_genomic_accession=19789213, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376865, CurrentID=0, Status='live', Symbol='LOC105376865', Aliases='', description='uncharacterized LOC105376865', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24096053, end_position_on_the_genomic_accession=24097921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928929, CurrentID=0, Status='live', Symbol='LOC101928929', Aliases='', description='uncharacterized LOC101928929', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185902942, end_position_on_the_genomic_accession=185919458, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928931, CurrentID=0, Status='live', Symbol='SEMA3B-AS1', Aliases='', description='SEMA3B antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50266641, end_position_on_the_genomic_accession=50267371, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379841, CurrentID=0, Status='live', Symbol='LOC105379841', Aliases='', description='uncharacterized LOC105379841', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928933, CurrentID=0, Status='live', Symbol='LOC101928933', Aliases='', description='uncharacterized LOC101928933', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179926390, end_position_on_the_genomic_accession=179954671, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928936, CurrentID=0, Status='live', Symbol='LOC101928936', Aliases='', description='uncharacterized LOC101928936', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88525924, end_position_on_the_genomic_accession=88539369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874236, CurrentID=0, Status='live', Symbol='COL18A1-AS2', Aliases='', description='COL18A1 antisense RNA 2', other_designations='COL18A1 antisense RNA 2 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45407386, end_position_on_the_genomic_accession=45410065, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131495, CurrentID=0, Status='live', Symbol='LOC100131495', Aliases='', description='nitrilase family member 2 pseudogene', other_designations='nitrilase family, member 2 pseudogene', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78043187, end_position_on_the_genomic_accession=78044221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928940, CurrentID=0, Status='live', Symbol='LOC101928940', Aliases='', description='uncharacterized LOC101928940', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114360635, end_position_on_the_genomic_accession=114380039, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928941, CurrentID=0, Status='live', Symbol='RHOXF1P1', Aliases='TCONS_00017086', description='Rhox homeobox family member 1 pseudogene 1', other_designations='Rhox homeobox family, member 1 pseudogene 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119986530, end_position_on_the_genomic_accession=120015622, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928942, CurrentID=0, Status='live', Symbol='LOC101928942', Aliases='', description='uncharacterized LOC101928942', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81164940, end_position_on_the_genomic_accession=81193395, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874237, CurrentID=0, Status='live', Symbol='CACNA2D3-AS1', Aliases='', description='CACNA2D3 antisense RNA 1', other_designations='CACNA2D3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54874605, end_position_on_the_genomic_accession=54901255, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928944, CurrentID=0, Status='live', Symbol='LOC101928944', Aliases='', description='uncharacterized LOC101928944', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80751200, end_position_on_the_genomic_accession=80762803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374376, CurrentID=0, Status='live', Symbol='LOC105374376', Aliases='', description='uncharacterized LOC105374376', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8436316, end_position_on_the_genomic_accession=8439065, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928946, CurrentID=0, Status='live', Symbol='LOC101928946', Aliases='', description='uncharacterized LOC101928946', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157309727, end_position_on_the_genomic_accession=157313979, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928947, CurrentID=0, Status='live', Symbol='LOC101928947', Aliases='', description='uncharacterized LOC101928947', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928948, CurrentID=0, Status='live', Symbol='LOC101928948', Aliases='', description='scavenger receptor class F member 2-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9532033, end_position_on_the_genomic_accession=9660852, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874238, CurrentID=0, Status='live', Symbol='ARHGEF7-AS2', Aliases='', description='ARHGEF7 antisense RNA 2', other_designations='ARHGEF7 antisense RNA 2 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131497, CurrentID=0, Status='live', Symbol='LOC100131497', Aliases='', description='uncharacterized LOC100131497', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=78278151, end_position_on_the_genomic_accession=78278327, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928952, CurrentID=0, Status='live', Symbol='LOC101928952', Aliases='', description='serine palmitoyltransferase small subunit A pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115916477, end_position_on_the_genomic_accession=115916781, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928953, CurrentID=0, Status='live', Symbol='LOC101928953', Aliases='', description='uncharacterized LOC101928953', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=145719473, end_position_on_the_genomic_accession=145725663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874239, CurrentID=0, Status='live', Symbol='ARHGAP26-AS1', Aliases='', description='ARHGAP26 antisense RNA 1', other_designations='ARHGAP26 antisense RNA 1 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142859604, end_position_on_the_genomic_accession=142868910, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  348156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348156, CurrentID=0, Status='live', Symbol='PKD1P5', Aliases='HG5', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 5', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18374521, end_position_on_the_genomic_accession=18402014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928957, CurrentID=0, Status='live', Symbol='LOC101928957', Aliases='', description='uncharacterized LOC101928957', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91520516, end_position_on_the_genomic_accession=91577822, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  348158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348158, CurrentID=0, Status='live', Symbol='ACSM2B', Aliases='ACSM2, HXMA, HYST1046', description='acyl-CoA synthetase medium-chain family member 2B', other_designations='acyl-CoA synthetase medium-chain family member 2|butyrate--CoA ligase 2B|butyryl-coenzyme A synthetase 2B|middle-chain acyl-CoA synthetase 2B|xenobiotic/medium-chain fatty acid-CoA ligase HXM-A|xenobiotic/medium-chain fatty acid:CoA ligase', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20536261, end_position_on_the_genomic_accession=20576373, orientation='minus', exon_count=16, OMIM=614359, no_hdr0=''),
  101928959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928959, CurrentID=0, Status='live', Symbol='LOC101928959', Aliases='', description='probable ribosome biogenesis protein RLP24 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98794471, end_position_on_the_genomic_accession=98794991, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928960, CurrentID=0, Status='live', Symbol='LOC101928960', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20033834, end_position_on_the_genomic_accession=20034526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928961, CurrentID=0, Status='live', Symbol='LOC101928961', Aliases='', description='uncharacterized LOC101928961', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=66894110, end_position_on_the_genomic_accession=67012153, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379842, CurrentID=0, Status='live', Symbol='LOC105379842', Aliases='', description='uncharacterized LOC105379842', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928963, CurrentID=0, Status='live', Symbol='LOC101928963', Aliases='', description='uncharacterized LOC101928963', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82374079, end_position_on_the_genomic_accession=82384027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928964, CurrentID=0, Status='live', Symbol='LOC101928964', Aliases='', description='uncharacterized LOC101928964', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80458975, end_position_on_the_genomic_accession=81252971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874241, CurrentID=0, Status='live', Symbol='STARD13-AS', Aliases='STARD13-AS2', description='STARD13 antisense RNA', other_designations='STARD13 antisense RNA 2 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33277554, end_position_on_the_genomic_accession=33281334, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928969, CurrentID=0, Status='live', Symbol='RHOXF1-AS1', Aliases='', description='RHOXF1 antisense RNA 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120119707, end_position_on_the_genomic_accession=120146855, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928971, CurrentID=0, Status='live', Symbol='LINC01262', Aliases='TCONS_l2_00021807', description='long intergenic non-protein coding RNA 1262', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189659606, end_position_on_the_genomic_accession=189661486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928972, CurrentID=0, Status='live', Symbol='LOC101928972', Aliases='', description='uncharacterized LOC101928972', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63178034, end_position_on_the_genomic_accession=63189457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874242, CurrentID=0, Status='live', Symbol='PRICKLE2-AS2', Aliases='', description='PRICKLE2 antisense RNA 2', other_designations='PRICKLE2 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64103470, end_position_on_the_genomic_accession=64106358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  348174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348174, CurrentID=0, Status='live', Symbol='CLEC18A', Aliases='MRCL, MRCL1, MRLP2', description='C-type lectin domain family 18 member A', other_designations='C-type lectin domain family 18, member A|mannose receptor-like 1|mannose receptor-like protein 2|mannose receptor-like protein 3', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69950394, end_position_on_the_genomic_accession=69964347, orientation='plus', exon_count=15, OMIM=616571, no_hdr0=''),
  101928977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928977, CurrentID=0, Status='live', Symbol='LOC101928977', Aliases='', description='uncharacterized LOC101928977', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115922573, end_position_on_the_genomic_accession=115977276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928978, CurrentID=0, Status='live', Symbol='LOC101928978', Aliases='', description='uncharacterized LOC101928978', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83968082, end_position_on_the_genomic_accession=84299169, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100874243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874243, CurrentID=0, Status='live', Symbol='PRICKLE2-AS3', Aliases='', description='PRICKLE2 antisense RNA 3', other_designations='PRICKLE2 antisense RNA 3 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64187544, end_position_on_the_genomic_accession=64201561, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  348180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348180, CurrentID=0, Status='live', Symbol='CTU2', Aliases='C16orf84, NCS2, UPF0432', description='cytosolic thiouridylase subunit 2 homolog (S. pombe)', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88706483, end_position_on_the_genomic_accession=88715386, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101928981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928981, CurrentID=0, Status='live', Symbol='LOC101928981', Aliases='', description='uncharacterized LOC101928981', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101730593, end_position_on_the_genomic_accession=101743753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928982, CurrentID=0, Status='live', Symbol='LOC101928982', Aliases='', description='uncharacterized LOC101928982', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56536158, end_position_on_the_genomic_accession=56538750, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928984, CurrentID=0, Status='live', Symbol='LOC101928984', Aliases='', description='uncharacterized LOC101928984', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139641662, end_position_on_the_genomic_accession=139657842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874244, CurrentID=0, Status='live', Symbol='NAALADL2-AS2', Aliases='', description='NAALADL2 antisense RNA 2', other_designations='NAALADL2 antisense RNA 2 (non-protein coding)', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=175235092, end_position_on_the_genomic_accession=175271096, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928988, CurrentID=0, Status='live', Symbol='LOC101928988', Aliases='', description='uncharacterized LOC101928988', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63928274, end_position_on_the_genomic_accession=63935953, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928989, CurrentID=0, Status='live', Symbol='MIR4300HG', Aliases='', description='MIR4300 host gene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=81879851, end_position_on_the_genomic_accession=82403913, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101928991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928991, CurrentID=0, Status='live', Symbol='CCDC92B', Aliases='', description='coiled-coil domain containing 92B', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2724411, end_position_on_the_genomic_accession=2749562, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928992, CurrentID=0, Status='live', Symbol='LOC101928992', Aliases='', description='uncharacterized LOC101928992', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184756389, end_position_on_the_genomic_accession=184773155, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928993, CurrentID=0, Status='live', Symbol='LOC101928993', Aliases='', description='uncharacterized LOC101928993', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69844637, end_position_on_the_genomic_accession=69879530, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928994, CurrentID=0, Status='live', Symbol='LOC101928994', Aliases='', description='uncharacterized LOC101928994', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69215333, end_position_on_the_genomic_accession=69232490, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928995, CurrentID=0, Status='live', Symbol='LOC101928995', Aliases='', description='uncharacterized LOC101928995', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115919376, end_position_on_the_genomic_accession=115925908, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928996, CurrentID=0, Status='live', Symbol='LOC101928996', Aliases='', description='histone H3.3-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=123514464, end_position_on_the_genomic_accession=123514924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928998, CurrentID=0, Status='live', Symbol='LOC101928998', Aliases='', description='uncharacterized LOC101928998', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12509563, end_position_on_the_genomic_accession=12527742, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928999, CurrentID=0, Status='live', Symbol='IQCF5-AS1', Aliases='', description='IQCF5 antisense RNA 1', other_designations='IQCF5 antisense RNA 1 (non-protein coding)', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51873596, end_position_on_the_genomic_accession=51875767, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929000, CurrentID=0, Status='live', Symbol='LOC101929000', Aliases='', description='golgin subfamily A member 6-like protein 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53361874, end_position_on_the_genomic_accession=53367364, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929002, CurrentID=0, Status='live', Symbol='LOC101929002', Aliases='', description='uncharacterized LOC101929002', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92905697, end_position_on_the_genomic_accession=92907548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874247, CurrentID=0, Status='live', Symbol='ANKRD33B-AS1', Aliases='', description='ANKRD33B antisense RNA 1', other_designations='ANKRD33B antisense RNA 1 (non-protein coding)', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929004, CurrentID=0, Status='live', Symbol='LOC101929004', Aliases='', description='uncharacterized LOC101929004', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89131481, end_position_on_the_genomic_accession=89147723, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644450, CurrentID=0, Status='live', Symbol='LOC644450', Aliases='', description='uncharacterized LOC644450', other_designations='', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=141815040, end_position_on_the_genomic_accession=141832633, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929006, CurrentID=0, Status='live', Symbol='LOC101929006', Aliases='', description='uncharacterized LOC101929006', other_designations='XXbac-BPG308J9.3', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29223885, end_position_on_the_genomic_accession=29290440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929007, CurrentID=0, Status='live', Symbol='LOC101929007', Aliases='', description='uncharacterized LOC101929007', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21587405, end_position_on_the_genomic_accession=21594699, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929008, CurrentID=0, Status='live', Symbol='LOC101929008', Aliases='', description='uncharacterized LOC101929008', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45769085, end_position_on_the_genomic_accession=45805139, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101929010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929010, CurrentID=0, Status='live', Symbol='SIRPG-AS1', Aliases='', description='SIRPG antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1633508, end_position_on_the_genomic_accession=1648473, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929011, CurrentID=0, Status='live', Symbol='LOC101929011', Aliases='', description='uncharacterized LOC101929011', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116639422, end_position_on_the_genomic_accession=116658252, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369245, CurrentID=0, Status='live', Symbol='LOC105369245', Aliases='', description='FERM and PDZ domain-containing protein 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929013, CurrentID=0, Status='live', Symbol='LOC101929013', Aliases='', description='proline-rich protein 23D1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7792627, end_position_on_the_genomic_accession=7799468, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929014, CurrentID=0, Status='live', Symbol='LOC101929014', Aliases='', description='uncharacterized LOC101929014', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125410223, end_position_on_the_genomic_accession=125412453, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131508, CurrentID=0, Status='live', Symbol='LOC100131508', Aliases='', description='PRO2122', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=82319026, end_position_on_the_genomic_accession=82319256, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929018, CurrentID=0, Status='live', Symbol='LOC101929018', Aliases='', description='uncharacterized LOC101929018', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185475814, end_position_on_the_genomic_accession=185481259, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100271283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271283, CurrentID=0, Status='live', Symbol='RPL15P14', Aliases='RPL15_7_1062', description='ribosomal protein L15 pseudogene 14', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72189947, end_position_on_the_genomic_accession=72190576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874250, CurrentID=0, Status='live', Symbol='KIRREL3-AS1', Aliases='', description='KIRREL3 antisense RNA 1', other_designations='KIRREL3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929022, CurrentID=0, Status='live', Symbol='LOC101929022', Aliases='', description='uncharacterized LOC101929022', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63394754, end_position_on_the_genomic_accession=63395439, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929023, CurrentID=0, Status='live', Symbol='LOC101929023', Aliases='', description='uncharacterized LOC101929023', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116423724, end_position_on_the_genomic_accession=116478826, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929025, CurrentID=0, Status='live', Symbol='LINC01516', Aliases='', description='long intergenic non-protein coding RNA 1516', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25113058, end_position_on_the_genomic_accession=25161233, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929026, CurrentID=0, Status='live', Symbol='ZACNP1', Aliases='', description='zinc activated ligand-gated ion channel pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108396902, end_position_on_the_genomic_accession=108416642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874251, CurrentID=0, Status='live', Symbol='KIRREL3-AS2', Aliases='', description='KIRREL3 antisense RNA 2', other_designations='KIRREL3 antisense RNA 2 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126940746, end_position_on_the_genomic_accession=126945090, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929028, CurrentID=0, Status='live', Symbol='LOC101929028', Aliases='', description='uncharacterized LOC101929028', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16367445, end_position_on_the_genomic_accession=16685849, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105374390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374390, CurrentID=0, Status='live', Symbol='LOC105374390', Aliases='', description='uncharacterized LOC105374390', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33565172, end_position_on_the_genomic_accession=33571580, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929031, CurrentID=0, Status='live', Symbol='LOC101929031', Aliases='', description='uncharacterized LOC101929031', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33562875, end_position_on_the_genomic_accession=33571660, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874252, CurrentID=0, Status='live', Symbol='USP32P4', Aliases='', description='ubiquitin specific peptidase 32 pseudogene 4', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62282693, end_position_on_the_genomic_accession=62288789, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929034, CurrentID=0, Status='live', Symbol='LOC101929034', Aliases='', description='uncharacterized LOC101929034', other_designations='CTD-2194D22.2', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1855970, end_position_on_the_genomic_accession=1856568, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  348235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348235, CurrentID=0, Status='live', Symbol='SKA2', Aliases='FAM33A', description='spindle and kinetochore associated complex subunit 2', other_designations='family with sequence similarity 33, member A|spindle and KT (kinetochore) associated 2', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59109947, end_position_on_the_genomic_accession=59155439, orientation='minus', exon_count=4, OMIM=616674, no_hdr0=''),
  101929036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929036, CurrentID=0, Status='live', Symbol='LOC101929036', Aliases='', description='uncharacterized LOC101929036', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102927667, end_position_on_the_genomic_accession=102951231, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100874253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874253, CurrentID=0, Status='live', Symbol='LINC00853', Aliases='PDZK1IP1-AS1, ncRNA-a4', description='long intergenic non-protein coding RNA 853', other_designations='PDZK1IP1 antisense RNA 1 (non-protein coding)|PDZK1IP1 antisense RNA 1 (tail to tail)|noncoding RNA-activating 4', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47179250, end_position_on_the_genomic_accession=47180339, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929043, CurrentID=0, Status='live', Symbol='LOC101929043', Aliases='', description='uncharacterized LOC101929043', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83637653, end_position_on_the_genomic_accession=83665596, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100874254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874254, CurrentID=0, Status='live', Symbol='JTBP1', Aliases='', description='jumping translocation breakpoint pseudogene 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40798039, end_position_on_the_genomic_accession=40798285, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929047, CurrentID=0, Status='live', Symbol='LOC101929047', Aliases='', description='uncharacterized LOC101929047', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28589492, end_position_on_the_genomic_accession=28618260, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929048, CurrentID=0, Status='live', Symbol='LOC101929048', Aliases='', description='uncharacterized LOC101929048', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13764508, end_position_on_the_genomic_accession=13841125, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  348249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348249, CurrentID=0, Status='live', Symbol='CCL15-CCL14', Aliases='', description='CCL15-CCL14 readthrough (NMD candidate)', other_designations='CCL15-CCL14 readthrough (non-protein coding)', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35983656, end_position_on_the_genomic_accession=36002038, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929050, CurrentID=0, Status='live', Symbol='LOC101929050', Aliases='', description='uncharacterized LOC101929050', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63378118, end_position_on_the_genomic_accession=63394193, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100874255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874255, CurrentID=0, Status='live', Symbol='RNASEH2B-AS1', Aliases='', description='RNASEH2B antisense RNA 1', other_designations='RNASEH2B antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50882378, end_position_on_the_genomic_accession=50910712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929052, CurrentID=0, Status='live', Symbol='LINC00836', Aliases='', description='long intergenic non-protein coding RNA 836', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25651712, end_position_on_the_genomic_accession=25732935, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371116, CurrentID=0, Status='live', Symbol='LOC105371116', Aliases='', description='uncharacterized LOC105371116', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19923592, end_position_on_the_genomic_accession=19929225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  348254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348254, CurrentID=0, Status='live', Symbol='CCDC144CP', Aliases='CCDC144C', description='coiled-coil domain containing 144C, pseudogene', other_designations='coiled-coil domain containing 144 pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20321174, end_position_on_the_genomic_accession=20402191, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  101929056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929056, CurrentID=0, Status='live', Symbol='LOC101929056', Aliases='', description='uncharacterized LOC101929056', other_designations='CTD-2194D22.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1883383, end_position_on_the_genomic_accession=1884649, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874256, CurrentID=0, Status='live', Symbol='TNR-IT1', Aliases='', description='TNR intronic transcript 1', other_designations='TNR intronic transcript 1 (non-protein coding)', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929058, CurrentID=0, Status='live', Symbol='LOC101929058', Aliases='', description='uncharacterized LOC101929058', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103151842, end_position_on_the_genomic_accession=103168309, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929059, CurrentID=0, Status='live', Symbol='ZIM2-AS1', Aliases='ZIM2as', description='ZIM2 antisense RNA 1', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56765322, end_position_on_the_genomic_accession=56799646, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929060, CurrentID=0, Status='live', Symbol='LOC101929060', Aliases='', description='uncharacterized LOC101929060', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=301437, end_position_on_the_genomic_accession=304505, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  348262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348262, CurrentID=0, Status='live', Symbol='FAM195B', Aliases='MCRIP1', description='family with sequence similarity 195 member B', other_designations='MAPK-regulated co-repressor interacting protein 1|MCRIP1 protein|family with sequence similarity 195, member B|protein FAM195B', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81822361, end_position_on_the_genomic_accession=81833294, orientation='minus', exon_count=7, OMIM=616514, no_hdr0=''),
  101929064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929064, CurrentID=0, Status='live', Symbol='LOC101929064', Aliases='', description='uncharacterized LOC101929064', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86119806, end_position_on_the_genomic_accession=86219926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929066, CurrentID=0, Status='live', Symbol='LOC101929066', Aliases='', description='uncharacterized LOC101929066', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18084868, end_position_on_the_genomic_accession=18097651, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  86123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=86123, CurrentID=0, Status='live', Symbol='YWHAZP5', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 5', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105686308, end_position_on_the_genomic_accession=105689724, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929069, CurrentID=0, Status='live', Symbol='GS1-120K12.4', Aliases='', description='uncharacterized LOC101929069', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185001527, end_position_on_the_genomic_accession=185008678, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929070, CurrentID=0, Status='live', Symbol='LOC101929070', Aliases='', description='uncharacterized LOC101929070', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929071, CurrentID=0, Status='live', Symbol='LINC01182', Aliases='', description='long intergenic non-protein coding RNA 1182', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13655179, end_position_on_the_genomic_accession=13931228, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929072, CurrentID=0, Status='live', Symbol='LOC101929072', Aliases='', description='uncharacterized LOC101929072', other_designations='XXbac-BPG181B23.7', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31394291, end_position_on_the_genomic_accession=31398485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929073, CurrentID=0, Status='live', Symbol='LOC101929073', Aliases='', description='uncharacterized LOC101929073', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25924378, end_position_on_the_genomic_accession=25933710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929074, CurrentID=0, Status='live', Symbol='PIK3CD-AS2', Aliases='', description='PIK3CD antisense RNA 2', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9672426, end_position_on_the_genomic_accession=9687574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79891, CurrentID=0, Status='live', Symbol='ZNF671', Aliases='', description='zinc finger protein 671', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57719747, end_position_on_the_genomic_accession=57727627, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929076, CurrentID=0, Status='live', Symbol='LOC101929076', Aliases='', description='uncharacterized LOC101929076', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67834308, end_position_on_the_genomic_accession=67838879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929077, CurrentID=0, Status='live', Symbol='LOC101929077', Aliases='', description='uncharacterized LOC101929077', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37934572, end_position_on_the_genomic_accession=38065564, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929079, CurrentID=0, Status='live', Symbol='LOC101929079', Aliases='', description='uncharacterized LOC101929079', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186985753, end_position_on_the_genomic_accession=186992175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929080, CurrentID=0, Status='live', Symbol='LOC101929080', Aliases='', description='uncharacterized LOC101929080', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95329584, end_position_on_the_genomic_accession=95335504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929081, CurrentID=0, Status='live', Symbol='CTD-2194D22.4', Aliases='', description='uncharacterized LOC101929081', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1887331, end_position_on_the_genomic_accession=1900493, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929082, CurrentID=0, Status='live', Symbol='LINC01333', Aliases='', description='long intergenic non-protein coding RNA 1333', other_designations='CTD-2275D24.2', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74322486, end_position_on_the_genomic_accession=74328357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929083, CurrentID=0, Status='live', Symbol='CASC6', Aliases='', description='cancer susceptibility candidate 6 (non-protein coding)', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=91629040, end_position_on_the_genomic_accession=91690428, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929084, CurrentID=0, Status='live', Symbol='LOC101929084', Aliases='', description='uncharacterized LOC101929084', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103547794, end_position_on_the_genomic_accession=103559815, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929085, CurrentID=0, Status='live', Symbol='LOC101929085', Aliases='', description='uncharacterized LOC101929085', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71564886, end_position_on_the_genomic_accession=71597654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929086, CurrentID=0, Status='live', Symbol='LINC01447', Aliases='', description='long intergenic non-protein coding RNA 1447', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47621939, end_position_on_the_genomic_accession=47629893, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874261, CurrentID=0, Status='live', Symbol='LINC00854', Aliases='TMEM106A-AS1', description='long intergenic non-protein coding RNA 854', other_designations='TMEM106A antisense RNA 1 (non-protein coding)|TMEM106A antisense RNA 1 (tail to tail)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43221417, end_position_on_the_genomic_accession=43231269, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929088, CurrentID=0, Status='live', Symbol='LOC101929088', Aliases='', description='uncharacterized LOC101929088', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63354293, end_position_on_the_genomic_accession=63380868, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929089, CurrentID=0, Status='live', Symbol='LOC101929089', Aliases='', description='uncharacterized LOC101929089', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118511918, end_position_on_the_genomic_accession=118531094, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929090, CurrentID=0, Status='live', Symbol='LOC101929090', Aliases='', description='uncharacterized LOC101929090', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87460811, end_position_on_the_genomic_accession=87462765, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929091, CurrentID=0, Status='live', Symbol='LOC101929091', Aliases='', description='enoyl-CoA hydratase domain containing 2 pseudogene', other_designations='enoyl CoA hydratase domain containing 2 pseudogene|enoyl-CoA hydratase domain-containing protein 2, mitochondrial pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26800823, end_position_on_the_genomic_accession=26801570, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929093, CurrentID=0, Status='live', Symbol='LINC01350', Aliases='GS1-204I12.1', description='long intergenic non-protein coding RNA 1350', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185558379, end_position_on_the_genomic_accession=185628488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929095, CurrentID=0, Status='live', Symbol='LOC101929095', Aliases='', description='uncharacterized LOC101929095', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15004942, end_position_on_the_genomic_accession=15427914, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929097, CurrentID=0, Status='live', Symbol='LOC101929097', Aliases='', description='uncharacterized LOC101929097', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2511218, end_position_on_the_genomic_accession=2513571, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929098, CurrentID=0, Status='live', Symbol='LOC101929098', Aliases='', description='uncharacterized LOC101929098', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3068760, end_position_on_the_genomic_accession=3072278, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874263, CurrentID=0, Status='live', Symbol='AKT3-IT1', Aliases='', description='AKT3 intronic transcript 1', other_designations='AKT3 intronic transcript 1 (non-protein coding)', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929101, CurrentID=0, Status='live', Symbol='LOC101929101', Aliases='', description='carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 2 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  348303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348303, CurrentID=0, Status='live', Symbol='SELV', Aliases='', description='selenoprotein V', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39515113, end_position_on_the_genomic_accession=39520686, orientation='plus', exon_count=6, OMIM=607919, no_hdr0=''),
  101929104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929104, CurrentID=0, Status='live', Symbol='LOC101929104', Aliases='', description='mitochondrial import receptor subunit TOM22 homolog pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87258653, end_position_on_the_genomic_accession=87259284, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874264, CurrentID=0, Status='live', Symbol='AOAH-IT1', Aliases='', description='AOAH intronic transcript 1', other_designations='AOAH intronic transcript 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36597834, end_position_on_the_genomic_accession=36600120, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929106, CurrentID=0, Status='live', Symbol='LOC101929106', Aliases='', description='uncharacterized LOC101929106', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187197090, end_position_on_the_genomic_accession=187207629, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929107, CurrentID=0, Status='live', Symbol='LOC101929107', Aliases='', description='uncharacterized LOC101929107', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95515386, end_position_on_the_genomic_accession=95518599, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929109, CurrentID=0, Status='live', Symbol='LOC101929109', Aliases='', description='uncharacterized LOC101929109', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76606608, end_position_on_the_genomic_accession=76609016, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929110, CurrentID=0, Status='live', Symbol='CASC18', Aliases='', description='cancer susceptibility candidate 18 (non-protein coding)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105704203, end_position_on_the_genomic_accession=105744063, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929111, CurrentID=0, Status='live', Symbol='LINC01149', Aliases='', description='long intergenic non-protein coding RNA 1149', other_designations='XXbac-BPG181B23.4', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31441667, end_position_on_the_genomic_accession=31445803, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379847, CurrentID=0, Status='live', Symbol='LOC105379847', Aliases='', description='uncharacterized LOC105379847', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130747, CurrentID=0, Status='live', Symbol='LOC100130747', Aliases='', description='N-acetylneuraminic acid phosphatase pseudogene', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=25140264, end_position_on_the_genomic_accession=25141201, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929114, CurrentID=0, Status='live', Symbol='LOC101929114', Aliases='', description='uncharacterized LOC101929114', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63352675, end_position_on_the_genomic_accession=63353360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929116, CurrentID=0, Status='live', Symbol='LOC101929116', Aliases='', description='uncharacterized LOC101929116', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126270124, end_position_on_the_genomic_accession=126275915, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874266, CurrentID=0, Status='live', Symbol='LYST-AS1', Aliases='LYST-IT2', description='LYST antisense RNA 1', other_designations='LYST intronic transcript 2 (non-protein coding)', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725001, CurrentID=0, Status='live', Symbol='LOC102725001', Aliases='', description='putative ankyrin repeat domain-containing protein 20A5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929120, CurrentID=0, Status='live', Symbol='LINC01351', Aliases='', description='long intergenic non-protein coding RNA 1351', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=190478260, end_position_on_the_genomic_accession=190481394, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929121, CurrentID=0, Status='live', Symbol='LINC00437', Aliases='TCONS_00021639', description='long intergenic non-protein coding RNA 437', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38532000, end_position_on_the_genomic_accession=38686675, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929122, CurrentID=0, Status='live', Symbol='LOC101929122', Aliases='', description='uncharacterized LOC101929122', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159165899, end_position_on_the_genomic_accession=159170007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874267, CurrentID=0, Status='live', Symbol='DPYD-IT1', Aliases='', description='DPYD intronic transcript 1', other_designations='DPYD intronic transcript 1 (non-protein coding)', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929124, CurrentID=0, Status='live', Symbol='LOC101929124', Aliases='', description='uncharacterized LOC101929124', other_designations='CTC-457E21.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22520995, end_position_on_the_genomic_accession=22527949, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374406, CurrentID=0, Status='live', Symbol='LOC105374406', Aliases='', description='uncharacterized LOC105374406', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37761112, end_position_on_the_genomic_accession=37781030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  348327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348327, CurrentID=0, Status='live', Symbol='ZNF530', Aliases='', description='zinc finger protein 530', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57599885, end_position_on_the_genomic_accession=57612722, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929128, CurrentID=0, Status='live', Symbol='LOC101929128', Aliases='', description='uncharacterized LOC101929128', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9188999, end_position_on_the_genomic_accession=9202858, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929129, CurrentID=0, Status='live', Symbol='LOC101929129', Aliases='', description='uncharacterized LOC101929129', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69972477, end_position_on_the_genomic_accession=70025507, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929130, CurrentID=0, Status='live', Symbol='LOC101929130', Aliases='', description='uncharacterized LOC101929130', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187291272, end_position_on_the_genomic_accession=187297933, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131527, CurrentID=0, Status='live', Symbol='RPL19P7', Aliases='RPL19_4_388', description='ribosomal protein L19 pseudogene 7', other_designations='', map_location='3q12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98620076, end_position_on_the_genomic_accession=98620696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929132, CurrentID=0, Status='live', Symbol='LOC101929132', Aliases='', description='uncharacterized LOC101929132', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106248237, end_position_on_the_genomic_accession=106252213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929133, CurrentID=0, Status='live', Symbol='LOC101929133', Aliases='', description='uncharacterized LOC101929133', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118886127, end_position_on_the_genomic_accession=118891007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929134, CurrentID=0, Status='live', Symbol='LOC101929134', Aliases='', description='uncharacterized LOC101929134', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88523844, end_position_on_the_genomic_accession=88524991, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874269, CurrentID=0, Status='live', Symbol='AFF2-IT1', Aliases='', description='AFF2 intronic transcript 1', other_designations='AFF2 intronic transcript 1 (non-protein coding)', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131528, CurrentID=0, Status='live', Symbol='APOOP3', Aliases='', description='apolipoprotein O pseudogene 3', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65134661, end_position_on_the_genomic_accession=65135560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929138, CurrentID=0, Status='live', Symbol='LOC101929138', Aliases='', description='uncharacterized LOC101929138', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53679285, end_position_on_the_genomic_accession=53681274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929140, CurrentID=0, Status='live', Symbol='LOC101929140', Aliases='', description='uncharacterized LOC101929140', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41132928, end_position_on_the_genomic_accession=41236686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874270, CurrentID=0, Status='live', Symbol='HAO2-IT1', Aliases='', description='HAO2 intronic transcript 1', other_designations='HAO2 intronic transcript 1 (non-protein coding)', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119368946, end_position_on_the_genomic_accession=119370331, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929142, CurrentID=0, Status='live', Symbol='LOC101929142', Aliases='', description='uncharacterized LOC101929142', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159463280, end_position_on_the_genomic_accession=159468861, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371119, CurrentID=0, Status='live', Symbol='LOC105371119', Aliases='', description='uncharacterized LOC105371119', other_designations='CTD-2194A8.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20440697, end_position_on_the_genomic_accession=20447517, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929144, CurrentID=0, Status='live', Symbol='LOC101929144', Aliases='', description='uncharacterized LOC101929144', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23054823, end_position_on_the_genomic_accession=23061265, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929145, CurrentID=0, Status='live', Symbol='PPP3CB-AS1', Aliases='USP54-AS1', description='PPP3CB antisense RNA 1 (head to head)', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73495752, end_position_on_the_genomic_accession=73507295, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929147, CurrentID=0, Status='live', Symbol='LOC101929147', Aliases='', description='uncharacterized LOC101929147', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119140396, end_position_on_the_genomic_accession=119275973, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929148, CurrentID=0, Status='live', Symbol='LOC101929148', Aliases='', description='uncharacterized LOC101929148', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22101692, end_position_on_the_genomic_accession=22147484, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101929151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929151, CurrentID=0, Status='live', Symbol='LOC101929151', Aliases='', description='uncharacterized LOC101929151', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70570958, end_position_on_the_genomic_accession=70586644, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929152, CurrentID=0, Status='live', Symbol='P3H2-AS1', Aliases='LEPREL1-AS1', description='P3H2 antisense RNA 1', other_designations='LEPREL1 antisense RNA 1 (non-protein coding)', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190120964, end_position_on_the_genomic_accession=190144846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79904, CurrentID=0, Status='live', Symbol='FLJ11710', Aliases='', description='uncharacterized protein FLJ11710', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  7235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7235, CurrentID=0, Status='live', Symbol='TRU-TCA2-1', Aliases='TRNAU2, TRSP2, TRSPP1', description='transfer RNA-SeC (TCA) 2-1', other_designations='tRNA-SeC-TCA-2-1|transfer RNA selenocysteine 2 (anticodon UCA)|transfer RNA selenocysteine/phosphoserine (opal suppressor) 2', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44150657, end_position_on_the_genomic_accession=44150742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929156, CurrentID=0, Status='live', Symbol='LOC101929156', Aliases='', description='uncharacterized LOC101929156', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119751961, end_position_on_the_genomic_accession=119770074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929157, CurrentID=0, Status='live', Symbol='PGM5P3', Aliases='', description='phosphoglucomutase 5 pseudogene 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=45108, end_position_on_the_genomic_accession=73177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874273, CurrentID=0, Status='live', Symbol='TGFA-IT1', Aliases='', description='TGFA intronic transcript 1', other_designations='TGFA intronic transcript 1 (non-protein coding)', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70467385, end_position_on_the_genomic_accession=70481640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929160, CurrentID=0, Status='live', Symbol='LOC101929160', Aliases='', description='uncharacterized LOC101929160', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193122338, end_position_on_the_genomic_accession=193130659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374412, CurrentID=0, Status='live', Symbol='LOC105374412', Aliases='', description='uncharacterized LOC105374412', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38748752, end_position_on_the_genomic_accession=38753468, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929162, CurrentID=0, Status='live', Symbol='LOC101929162', Aliases='', description='uncharacterized LOC101929162', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107736555, end_position_on_the_genomic_accession=107759968, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929163, CurrentID=0, Status='live', Symbol='LOC101929163', Aliases='', description='uncharacterized LOC101929163', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32255717, end_position_on_the_genomic_accession=32407822, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929164, CurrentID=0, Status='live', Symbol='LOC101929164', Aliases='', description='uncharacterized LOC101929164', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23015075, end_position_on_the_genomic_accession=23023697, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874274, CurrentID=0, Status='live', Symbol='THRB-IT1', Aliases='', description='THRB intronic transcript 1', other_designations='THRB intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131533, CurrentID=0, Status='live', Symbol='LOC100131533', Aliases='', description='dpy-19-like 4 (C. elegans) pseudogene', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248771641, end_position_on_the_genomic_accession=248773785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929170, CurrentID=0, Status='live', Symbol='LOC101929170', Aliases='', description='keratin, type I cytoskeletal 18 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929172, CurrentID=0, Status='live', Symbol='LOC101929172', Aliases='', description='uncharacterized LOC101929172', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21298239, end_position_on_the_genomic_accession=21309449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929173, CurrentID=0, Status='live', Symbol='THSD4-AS2', Aliases='', description='THSD4 antisense RNA 2', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71688701, end_position_on_the_genomic_accession=71691539, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929174, CurrentID=0, Status='live', Symbol='LOC101929174', Aliases='', description='uncharacterized LOC101929174', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88337873, end_position_on_the_genomic_accession=88428192, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929176, CurrentID=0, Status='live', Symbol='CTD-2297D10.2', Aliases='', description='uncharacterized LOC101929176', other_designations='', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5132085, end_position_on_the_genomic_accession=5140054, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874276, CurrentID=0, Status='live', Symbol='LINC01217', Aliases='EMCN-IT2', description='long intergenic non-protein coding RNA 1217', other_designations='EMCN intronic transcript 2 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  348378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348378, CurrentID=0, Status='live', Symbol='FAM159A', Aliases='PRO7171, WWLS2783', description='family with sequence similarity 159 member A', other_designations='family with sequence similarity 159, member A', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52631439, end_position_on_the_genomic_accession=52669120, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929179, CurrentID=0, Status='live', Symbol='LINC01388', Aliases='', description='long intergenic non-protein coding RNA 1388', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100994, end_position_on_the_genomic_accession=114247, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929180, CurrentID=0, Status='live', Symbol='LOC101929180', Aliases='', description='uncharacterized LOC101929180', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27343436, end_position_on_the_genomic_accession=27344917, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929181, CurrentID=0, Status='live', Symbol='LOC101929181', Aliases='', description='uncharacterized LOC101929181', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11609531, end_position_on_the_genomic_accession=11613355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929184, CurrentID=0, Status='live', Symbol='LINC01031', Aliases='TCONS_00000361', description='long intergenic non-protein coding RNA 1031', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=193304745, end_position_on_the_genomic_accession=193365953, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929185, CurrentID=0, Status='live', Symbol='LOC101929185', Aliases='', description='putative glycine-rich cell wall structural protein 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57988110, end_position_on_the_genomic_accession=57989295, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929188, CurrentID=0, Status='live', Symbol='LOC101929188', Aliases='', description='uncharacterized LOC101929188', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33631244, end_position_on_the_genomic_accession=33633706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929190, CurrentID=0, Status='live', Symbol='LOC101929190', Aliases='', description='helicase SRCAP-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63746834, end_position_on_the_genomic_accession=63752784, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929191, CurrentID=0, Status='live', Symbol='LINCR-0001', Aliases='', description='uncharacterized LINCR-0001', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10474565, end_position_on_the_genomic_accession=10481974, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929194, CurrentID=0, Status='live', Symbol='LOC101929194', Aliases='', description='uncharacterized LOC101929194', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92262498, end_position_on_the_genomic_accession=92277302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929196, CurrentID=0, Status='live', Symbol='THSD4-AS1', Aliases='', description='THSD4 antisense RNA 1', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71166812, end_position_on_the_genomic_accession=71189059, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929198, CurrentID=0, Status='live', Symbol='ATP13A4-AS1', Aliases='', description='ATP13A4 antisense RNA 1', other_designations='ATP13A4 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193553213, end_position_on_the_genomic_accession=193555584, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929199, CurrentID=0, Status='live', Symbol='LOC101929199', Aliases='', description='uncharacterized LOC101929199', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27217479, end_position_on_the_genomic_accession=27282225, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929200, CurrentID=0, Status='live', Symbol='LOC101929200', Aliases='', description='uncharacterized LOC101929200', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5390103, end_position_on_the_genomic_accession=5422123, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929201, CurrentID=0, Status='live', Symbol='LOC101929201', Aliases='', description='uncharacterized LOC101929201', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79349353, end_position_on_the_genomic_accession=79371475, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929202, CurrentID=0, Status='live', Symbol='MKX-AS1', Aliases='', description='MKX antisense RNA 1', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27744615, end_position_on_the_genomic_accession=27767799, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374419, CurrentID=0, Status='live', Symbol='LOC105374419', Aliases='', description='uncharacterized LOC105374419', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39638987, end_position_on_the_genomic_accession=39666653, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929204, CurrentID=0, Status='live', Symbol='LOC101929204', Aliases='', description='uncharacterized LOC101929204', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108802375, end_position_on_the_genomic_accession=108817031, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929206, CurrentID=0, Status='live', Symbol='LOC101929206', Aliases='', description='zinc finger protein 708-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23832189, end_position_on_the_genomic_accession=23871182, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874281, CurrentID=0, Status='live', Symbol='NTM-IT', Aliases='NTM-IT3', description='NTM intronic transcript', other_designations='NTM intronic transcript (non-protein coding)|NTM intronic transcript 3 (non-protein coding)', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=132284302, end_position_on_the_genomic_accession=132288459, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929208, CurrentID=0, Status='live', Symbol='LOC101929208', Aliases='', description='uncharacterized LOC101929208', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120940464, end_position_on_the_genomic_accession=120958061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374420, CurrentID=0, Status='live', Symbol='LOC105374420', Aliases='', description='uncharacterized LOC105374420', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40553558, end_position_on_the_genomic_accession=40558355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102060282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102060282, CurrentID=0, Status='live', Symbol='RASSF1-AS1', Aliases='ANRASSF1', description='RASSF1 antisense RNA 1', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50337511, end_position_on_the_genomic_accession=50338300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929211, CurrentID=0, Status='live', Symbol='LOC101929211', Aliases='', description='proline-rich protein 2-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22603133, end_position_on_the_genomic_accession=22605404, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929212, CurrentID=0, Status='live', Symbol='SMIM2-AS1', Aliases='C13orf44-AS1, C3orf81-AS1', description='SMIM2 antisense RNA 1', other_designations='SMIM2 antisense RNA 1 (non-protein coding)', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44110549, end_position_on_the_genomic_accession=44154079, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929214, CurrentID=0, Status='live', Symbol='LOC101929214', Aliases='', description='uncharacterized LOC101929214', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5932076, end_position_on_the_genomic_accession=5932884, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929215, CurrentID=0, Status='live', Symbol='CTD-2201I18.1', Aliases='', description='uncharacterized LOC101929215', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80066349, end_position_on_the_genomic_accession=80083665, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929216, CurrentID=0, Status='live', Symbol='LOC101929216', Aliases='', description='uncharacterized LOC101929216', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47210953, end_position_on_the_genomic_accession=47231647, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929217, CurrentID=0, Status='live', Symbol='LOC101929217', Aliases='', description='uncharacterized LOC101929217', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48620403, end_position_on_the_genomic_accession=48623895, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929218, CurrentID=0, Status='live', Symbol='LINC01517', Aliases='', description='long intergenic non-protein coding RNA 1517', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28743650, end_position_on_the_genomic_accession=28808219, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874283, CurrentID=0, Status='live', Symbol='UPP2-IT1', Aliases='', description='UPP2 intronic transcript 1', other_designations='UPP2 intronic transcript 1 (non-protein coding)', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929221, CurrentID=0, Status='live', Symbol='C15orf59-AS1', Aliases='', description='C15orf59 antisense RNA 1', other_designations='TCONS_00023459', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73768086, end_position_on_the_genomic_accession=73769802, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929222, CurrentID=0, Status='live', Symbol='DISC1FP1', Aliases='Boymaw', description='DISC1 fusion partner 1 (non-protein coding)', other_designations='DISC1/FP1 fusion', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90251232, end_position_on_the_genomic_accession=90915052, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  474241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474241, CurrentID=0, Status='live', Symbol='RA20', Aliases='', description='Rheumatoid arthritis QTL 20', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929225, CurrentID=0, Status='live', Symbol='MCM8-AS1', Aliases='', description='MCM8 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5990795, end_position_on_the_genomic_accession=6005821, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929226, CurrentID=0, Status='live', Symbol='LOC101929226', Aliases='', description='histone H3.3-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52943534, end_position_on_the_genomic_accession=52943931, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390760, CurrentID=0, Status='live', Symbol='LOC390760', Aliases='', description='protein phosphatase inhibitor 2-like', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=7233771, end_position_on_the_genomic_accession=7238462, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929229, CurrentID=0, Status='live', Symbol='LOC101929229', Aliases='', description='uncharacterized LOC101929229', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10840026, end_position_on_the_genomic_accession=10846509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929230, CurrentID=0, Status='live', Symbol='LOC101929230', Aliases='', description='uncharacterized LOC101929230', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21607483, end_position_on_the_genomic_accession=21649563, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  101929231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929231, CurrentID=0, Status='live', Symbol='LOC101929231', Aliases='', description='uncharacterized LOC101929231', other_designations='', map_location='2q23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149587358, end_position_on_the_genomic_accession=149848234, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929232, CurrentID=0, Status='live', Symbol='LOC101929232', Aliases='', description='uncharacterized LOC101929232', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7289757, end_position_on_the_genomic_accession=7348955, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100421756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421756, CurrentID=0, Status='live', Symbol='EEF1B2P7', Aliases='', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 7', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232729478, end_position_on_the_genomic_accession=232730221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929234, CurrentID=0, Status='live', Symbol='LOC101929234', Aliases='', description='uncharacterized LOC101929234', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75409157, end_position_on_the_genomic_accession=75411324, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929235, CurrentID=0, Status='live', Symbol='LOC101929235', Aliases='', description='putative ubiquitin-conjugating enzyme E2Q2-like protein', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25463984, end_position_on_the_genomic_accession=25473624, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929236, CurrentID=0, Status='live', Symbol='LOC101929236', Aliases='', description='uncharacterized LOC101929236', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28897902, end_position_on_the_genomic_accession=28934486, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100874286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874286, CurrentID=0, Status='live', Symbol='NRG1-IT3', Aliases='', description='NRG1 intronic transcript 3', other_designations='NRG1 intronic transcript 3 (non-protein coding)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=32440746, end_position_on_the_genomic_accession=32442725, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929238, CurrentID=0, Status='live', Symbol='C3orf67-AS1', Aliases='', description='C3orf67 antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58824471, end_position_on_the_genomic_accession=59019093, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929239, CurrentID=0, Status='live', Symbol='PACRG-AS2', Aliases='', description='PACRG antisense RNA 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163042980, end_position_on_the_genomic_accession=163054161, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929240, CurrentID=0, Status='live', Symbol='LOC101929240', Aliases='', description='U5 small nuclear ribonucleoprotein 200 kDa helicase pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64750402, end_position_on_the_genomic_accession=64751508, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929241, CurrentID=0, Status='live', Symbol='LOC101929241', Aliases='', description='uncharacterized LOC101929241', other_designations='CTD-2506J14.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97458816, end_position_on_the_genomic_accession=97581383, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929242, CurrentID=0, Status='live', Symbol='LOC101929242', Aliases='', description='S-phase kinase-associated protein 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80746201, end_position_on_the_genomic_accession=80747368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874287, CurrentID=0, Status='live', Symbol='TSPY24P', Aliases='', description='testis specific protein, Y-linked 24, pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8280198, end_position_on_the_genomic_accession=8282209, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929244, CurrentID=0, Status='live', Symbol='CASC20', Aliases='', description='cancer susceptibility candidate 20 (non-protein coding)', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6426732, end_position_on_the_genomic_accession=6528459, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374426, CurrentID=0, Status='live', Symbol='LOC105374426', Aliases='', description='uncharacterized LOC105374426', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41892449, end_position_on_the_genomic_accession=41935081, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929247, CurrentID=0, Status='live', Symbol='LINC01231', Aliases='', description='long intergenic non-protein coding RNA 1231', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3181589, end_position_on_the_genomic_accession=3200500, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929248, CurrentID=0, Status='live', Symbol='LOC101929248', Aliases='', description='uncharacterized LOC101929248', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10850647, end_position_on_the_genomic_accession=10859974, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929249, CurrentID=0, Status='live', Symbol='LOC101929249', Aliases='', description='uncharacterized LOC101929249', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128392009, end_position_on_the_genomic_accession=128395466, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929250, CurrentID=0, Status='live', Symbol='LOC101929250', Aliases='', description='uncharacterized LOC101929250', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95356160, end_position_on_the_genomic_accession=95358396, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929251, CurrentID=0, Status='live', Symbol='LOC101929251', Aliases='', description='uncharacterized LOC101929251', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23018975, end_position_on_the_genomic_accession=23199104, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101929254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929254, CurrentID=0, Status='live', Symbol='LRRC37A12P', Aliases='', description='leucine-rich repeat containing 37 member A12, pseudogene', other_designations='leucine rich repeat containing 37, member A12, pseudogene|leucine-rich repeat-containing protein 37B-like', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32424895, end_position_on_the_genomic_accession=32427579, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874289, CurrentID=0, Status='live', Symbol='VDAC1P12', Aliases='', description='voltage-dependent anion channel 1 pseudogene 12', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34082777, end_position_on_the_genomic_accession=34083310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929256, CurrentID=0, Status='live', Symbol='LOC101929256', Aliases='', description='uncharacterized LOC101929256', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30049323, end_position_on_the_genomic_accession=30111844, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374428, CurrentID=0, Status='live', Symbol='LOC105374428', Aliases='', description='uncharacterized LOC105374428', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42299012, end_position_on_the_genomic_accession=42391326, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929258, CurrentID=0, Status='live', Symbol='LOC101929258', Aliases='', description='uncharacterized LOC101929258', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23711902, end_position_on_the_genomic_accession=23726988, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929259, CurrentID=0, Status='live', Symbol='LOC101929259', Aliases='', description='uncharacterized LOC101929259', other_designations='', map_location='13q14.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45106279, end_position_on_the_genomic_accession=45113206, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929260, CurrentID=0, Status='live', Symbol='LOC101929260', Aliases='', description='uncharacterized LOC101929260', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150552532, end_position_on_the_genomic_accession=150572221, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874290, CurrentID=0, Status='live', Symbol='RERG-IT1', Aliases='', description='RERG intronic transcript 1', other_designations='RERG intronic transcript 1 (non-protein coding)', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374429, CurrentID=0, Status='live', Symbol='LOC105374429', Aliases='', description='uncharacterized LOC105374429', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42487578, end_position_on_the_genomic_accession=42500137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929265, CurrentID=0, Status='live', Symbol='LINC01428', Aliases='', description='long intergenic non-protein coding RNA 1428', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7146467, end_position_on_the_genomic_accession=7254202, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929266, CurrentID=0, Status='live', Symbol='LOC101929266', Aliases='', description='uncharacterized LOC101929266', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874291, CurrentID=0, Status='live', Symbol='ALMS1-IT1', Aliases='', description='ALMS1 intronic transcript 1', other_designations='ALMS1 intronic transcript 1 (non-protein coding)', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73457103, end_position_on_the_genomic_accession=73459482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929268, CurrentID=0, Status='live', Symbol='LOC101929268', Aliases='', description='uncharacterized LOC101929268', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48551567, end_position_on_the_genomic_accession=48698510, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374430, CurrentID=0, Status='live', Symbol='LOC105374430', Aliases='', description='uncharacterized LOC105374430', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42675211, end_position_on_the_genomic_accession=42704271, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929270, CurrentID=0, Status='live', Symbol='LOC101929270', Aliases='', description='uncharacterized LOC101929270', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128528657, end_position_on_the_genomic_accession=128543478, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929272, CurrentID=0, Status='live', Symbol='LOC101929272', Aliases='', description='uncharacterized LOC101929272', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25420730, end_position_on_the_genomic_accession=25441645, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100874292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874292, CurrentID=0, Status='live', Symbol='CDC42-IT1', Aliases='', description='CDC42 intronic transcript 1', other_designations='CDC42 intronic transcript 1 (non-protein coding)', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374431, CurrentID=0, Status='live', Symbol='LOC105374431', Aliases='', description='uncharacterized LOC105374431', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42786003, end_position_on_the_genomic_accession=42828168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929278, CurrentID=0, Status='live', Symbol='LOC101929278', Aliases='', description='uncharacterized LOC101929278', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97287984, end_position_on_the_genomic_accession=97439566, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  400046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400046, CurrentID=0, Status='live', Symbol='FLJ41278', Aliases='', description='uncharacterized LOC400046', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64883774, end_position_on_the_genomic_accession=64977522, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929280, CurrentID=0, Status='live', Symbol='LOC101929280', Aliases='', description='uncharacterized LOC101929280', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=594997, end_position_on_the_genomic_accession=597964, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929282, CurrentID=0, Status='live', Symbol='LOC101929282', Aliases='', description='uncharacterized LOC101929282', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150628897, end_position_on_the_genomic_accession=150635357, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  474243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474243, CurrentID=0, Status='live', Symbol='RA22', Aliases='', description='Rheumatoid arthritis QTL 22', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929285, CurrentID=0, Status='live', Symbol='LOC101929285', Aliases='', description='uncharacterized LOC101929285', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35224649, end_position_on_the_genomic_accession=35259398, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101929286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929286, CurrentID=0, Status='live', Symbol='KCNMA1-AS3', Aliases='', description='KCNMA1 antisense RNA 3', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77350872, end_position_on_the_genomic_accession=77376613, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  348487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348487, CurrentID=0, Status='live', Symbol='FAM131C', Aliases='C1orf117', description='family with sequence similarity 131 member C', other_designations='family with sequence similarity 131, member C', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16057769, end_position_on_the_genomic_accession=16073632, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929288, CurrentID=0, Status='live', Symbol='LOC101929288', Aliases='', description='uncharacterized LOC101929288', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7302089, end_position_on_the_genomic_accession=7307507, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929289, CurrentID=0, Status='live', Symbol='LOC101929289', Aliases='', description='uncharacterized LOC101929289', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123062467, end_position_on_the_genomic_accession=123084713, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929290, CurrentID=0, Status='live', Symbol='LOC101929290', Aliases='', description='uncharacterized LOC101929290', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11320857, end_position_on_the_genomic_accession=11329186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874295, CurrentID=0, Status='live', Symbol='KCND3-IT1', Aliases='', description='KCND3 intronic transcript 1', other_designations='KCND3 intronic transcript 1 (non-protein coding)', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111853762, end_position_on_the_genomic_accession=111856905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379853, CurrentID=0, Status='live', Symbol='LOC105379853', Aliases='', description='uncharacterized LOC105379853', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929293, CurrentID=0, Status='live', Symbol='PTGES3P2', Aliases='', description='prostaglandin E synthase 3 (cytosolic) pseudogene 2', other_designations='prostaglandin E synthase 3 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25822175, end_position_on_the_genomic_accession=25823837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929294, CurrentID=0, Status='live', Symbol='LOC101929294', Aliases='', description='uncharacterized LOC101929294', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24295814, end_position_on_the_genomic_accession=24548618, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929295, CurrentID=0, Status='live', Symbol='LOC101929295', Aliases='', description='uncharacterized LOC101929295', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95150539, end_position_on_the_genomic_accession=95159148, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929297, CurrentID=0, Status='live', Symbol='LOC101929297', Aliases='', description='uncharacterized LOC101929297', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166236969, end_position_on_the_genomic_accession=166256947, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929298, CurrentID=0, Status='live', Symbol='LOC101929298', Aliases='', description='uncharacterized LOC101929298', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97800805, end_position_on_the_genomic_accession=97821776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929301, CurrentID=0, Status='live', Symbol='LOC101929301', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44870946, end_position_on_the_genomic_accession=44872158, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929302, CurrentID=0, Status='live', Symbol='RFX3-AS1', Aliases='CVAT7', description='RFX3 antisense RNA 1', other_designations='cervical cancer-associated transcript 7', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3668086, end_position_on_the_genomic_accession=3691756, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929305, CurrentID=0, Status='live', Symbol='LOC101929305', Aliases='', description='uncharacterized LOC101929305', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201023936, end_position_on_the_genomic_accession=201043069, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929307, CurrentID=0, Status='live', Symbol='LOC101929307', Aliases='', description='uncharacterized LOC101929307', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8758080, end_position_on_the_genomic_accession=8840527, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929308, CurrentID=0, Status='live', Symbol='LOC101929308', Aliases='', description='protein FAM173B-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42069933, end_position_on_the_genomic_accession=42087763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874298, CurrentID=0, Status='live', Symbol='RASA2-IT1', Aliases='', description='RASA2 intronic transcript 1', other_designations='RASA2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929310, CurrentID=0, Status='live', Symbol='KCNMA1-AS2', Aliases='', description='KCNMA1 antisense RNA 2', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77140831, end_position_on_the_genomic_accession=77150100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131557, CurrentID=0, Status='live', Symbol='LOC100131557', Aliases='', description='programmed cell death 2-like pseudogene', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23370116, end_position_on_the_genomic_accession=23377701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929312, CurrentID=0, Status='live', Symbol='LOC101929312', Aliases='', description='uncharacterized LOC101929312', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7347433, end_position_on_the_genomic_accession=7368501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929314, CurrentID=0, Status='live', Symbol='LOC101929314', Aliases='', description='uncharacterized LOC101929314', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128993654, end_position_on_the_genomic_accession=129006564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874299, CurrentID=0, Status='live', Symbol='KDM5C-IT1', Aliases='', description='KDM5C intronic transcript 1', other_designations='KDM5C intronic transcript 1 (non-protein coding)', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929316, CurrentID=0, Status='live', Symbol='LOC101929316', Aliases='', description='uncharacterized LOC101929316', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64445231, end_position_on_the_genomic_accession=64456070, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929319, CurrentID=0, Status='live', Symbol='LOC101929319', Aliases='', description='uncharacterized LOC101929319', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151337826, end_position_on_the_genomic_accession=151379371, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929320, CurrentID=0, Status='live', Symbol='LOC101929320', Aliases='', description='uncharacterized LOC101929320', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98235525, end_position_on_the_genomic_accession=98237373, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874300, CurrentID=0, Status='live', Symbol='SYNJ2-IT1', Aliases='', description='SYNJ2 intronic transcript 1', other_designations='SYNJ2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158001107, end_position_on_the_genomic_accession=158002383, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929322, CurrentID=0, Status='live', Symbol='LOC101929322', Aliases='', description='integrator complex subunit 4-like protein 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65499303, end_position_on_the_genomic_accession=65555350, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929325, CurrentID=0, Status='live', Symbol='LOC101929325', Aliases='', description='uncharacterized LOC101929325', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148293278, end_position_on_the_genomic_accession=148297556, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929326, CurrentID=0, Status='live', Symbol='LOC101929326', Aliases='', description='uncharacterized LOC101929326', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201355226, end_position_on_the_genomic_accession=201356658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874301, CurrentID=0, Status='live', Symbol='TSSC1-IT1', Aliases='', description='TSSC1 intronic transcript 1', other_designations='TSSC1 intronic transcript 1 (non-protein coding)', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929328, CurrentID=0, Status='live', Symbol='KCNMA1-AS1', Aliases='', description='KCNMA1 antisense RNA 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=76888044, end_position_on_the_genomic_accession=76978593, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929329, CurrentID=0, Status='live', Symbol='LAMP5-AS1', Aliases='', description='LAMP5 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9505180, end_position_on_the_genomic_accession=9514998, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374249, CurrentID=0, Status='live', Symbol='LOC105374249', Aliases='', description='uncharacterized LOC105374249', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184132942, end_position_on_the_genomic_accession=184134799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929331, CurrentID=0, Status='live', Symbol='LOC101929331', Aliases='', description='uncharacterized LOC101929331', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129282416, end_position_on_the_genomic_accession=129285719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874302, CurrentID=0, Status='live', Symbol='ITCH-IT1', Aliases='', description='ITCH intronic transcript 1', other_designations='ITCH intronic transcript 1 (non-protein coding)', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929335, CurrentID=0, Status='live', Symbol='ADAMTS9-AS1', Aliases='', description='ADAMTS9 antisense RNA 1', other_designations='ADAMTS9 antisense RNA 1 (non-protein coding)', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64561346, end_position_on_the_genomic_accession=64590084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929337, CurrentID=0, Status='live', Symbol='LOC101929337', Aliases='', description='uncharacterized LOC101929337', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194247648, end_position_on_the_genomic_accession=194257772, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929338, CurrentID=0, Status='live', Symbol='CTD-2201E9.1', Aliases='', description='uncharacterized LOC101929338', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9511333, end_position_on_the_genomic_accession=9518086, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874303, CurrentID=0, Status='live', Symbol='ALG9-IT1', Aliases='', description='ALG9 intronic transcript 1', other_designations='ALG9 intronic transcript 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929340, CurrentID=0, Status='live', Symbol='LOC101929340', Aliases='', description='uncharacterized LOC101929340', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124759134, end_position_on_the_genomic_accession=124765925, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929341, CurrentID=0, Status='live', Symbol='LOC101929341', Aliases='', description='uncharacterized LOC101929341', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52194320, end_position_on_the_genomic_accession=52199552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102060414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102060414, CurrentID=0, Status='live', Symbol='C2-AS1', Aliases='', description='C2 antisense RNA 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31934474, end_position_on_the_genomic_accession=31941724, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929343, CurrentID=0, Status='live', Symbol='LOC101929343', Aliases='', description='uncharacterized LOC101929343', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201399633, end_position_on_the_genomic_accession=201401564, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  474245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474245, CurrentID=0, Status='live', Symbol='RA24', Aliases='', description='Rheumatoid arthritis QTL 24', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347674, CurrentID=0, Status='live', Symbol='LOC347674', Aliases='', description='armadillo repeat containing, X-linked 2 pseudogene', other_designations='ALEX2 pseudogene', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130822459, end_position_on_the_genomic_accession=130823225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929347, CurrentID=0, Status='live', Symbol='LOC101929347', Aliases='', description='uncharacterized LOC101929347', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77889599, end_position_on_the_genomic_accession=77900567, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929348, CurrentID=0, Status='live', Symbol='LOC101929348', Aliases='', description='uncharacterized LOC101929348', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9619724, end_position_on_the_genomic_accession=9626898, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929349, CurrentID=0, Status='live', Symbol='LOC101929349', Aliases='', description='activator of 90 kDa heat shock protein ATPase homolog 2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46735522, end_position_on_the_genomic_accession=46743132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929350, CurrentID=0, Status='live', Symbol='LOC101929350', Aliases='', description='uncharacterized LOC101929350', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16746869, end_position_on_the_genomic_accession=16748633, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929351, CurrentID=0, Status='live', Symbol='CDC37L1-AS1', Aliases='', description='CDC37L1 antisense RNA 1 (head to head)', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4676600, end_position_on_the_genomic_accession=4679480, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929352, CurrentID=0, Status='live', Symbol='LOC101929352', Aliases='', description='uncharacterized LOC101929352', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31187710, end_position_on_the_genomic_accession=31261552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  254700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254700, CurrentID=0, Status='live', Symbol='USP12PX', Aliases='USP12P2', description='ubiquitin specific peptidase 12 pseudogene, X-linked', other_designations='ubiquitin specific peptidase 12 pseudogene 2|ubiquitin specific protease 12 like 1 pseudogene|ubiquitin specific protease 12 pseudogene 2', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=90112446, end_position_on_the_genomic_accession=90114729, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929355, CurrentID=0, Status='live', Symbol='LOC101929355', Aliases='', description='uncharacterized protein C17orf105-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114172978, end_position_on_the_genomic_accession=114223074, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101929356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929356, CurrentID=0, Status='live', Symbol='LOC101929356', Aliases='', description='uncharacterized LOC101929356', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151776347, end_position_on_the_genomic_accession=151792045, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929357, CurrentID=0, Status='live', Symbol='HP09053', Aliases='', description='uncharacterized LOC101929357', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99816632, end_position_on_the_genomic_accession=99823865, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374445, CurrentID=0, Status='live', Symbol='LOC105374445', Aliases='', description='uncharacterized LOC105374445', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48269972, end_position_on_the_genomic_accession=48288942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929360, CurrentID=0, Status='live', Symbol='LOC101929360', Aliases='', description='uncharacterized LOC101929360', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65110630, end_position_on_the_genomic_accession=65123062, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874307, CurrentID=0, Status='live', Symbol='RSF1-IT2', Aliases='', description='RSF1 intronic transcript 2', other_designations='RSF1 intronic transcript 2 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929364, CurrentID=0, Status='live', Symbol='LOC101929364', Aliases='', description='uncharacterized LOC101929364', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62432453, end_position_on_the_genomic_accession=62444701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  348567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348567, CurrentID=0, Status='live', Symbol='TERF1P1', Aliases='TERF1P', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13776086, end_position_on_the_genomic_accession=13778812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929369, CurrentID=0, Status='live', Symbol='SNHG24', Aliases='', description='small nucleolar RNA host gene 24', other_designations='small nucleolar RNA host gene 24 (non-protein coding)', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100967652, end_position_on_the_genomic_accession=100986650, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  101929371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929371, CurrentID=0, Status='live', Symbol='LOC101929371', Aliases='', description='uncharacterized LOC101929371', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=9986333, end_position_on_the_genomic_accession=10007116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929372, CurrentID=0, Status='live', Symbol='LOC101929372', Aliases='', description='uncharacterized LOC101929372', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17580157, end_position_on_the_genomic_accession=17589192, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929373, CurrentID=0, Status='live', Symbol='LOC101929373', Aliases='', description='putative protein NHEDC1-like 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42242499, end_position_on_the_genomic_accession=42277067, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929374, CurrentID=0, Status='live', Symbol='LOC101929374', Aliases='', description='uncharacterized LOC101929374', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23433564, end_position_on_the_genomic_accession=23435071, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  735301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=735301, CurrentID=0, Status='live', Symbol='SNHG9', Aliases='NCRNA00062', description='small nucleolar RNA host gene 9', other_designations='small nucleolar RNA host gene (non-protein coding) 9|small nucleolar RNA host gene 9 (non-protein coding)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1964996, end_position_on_the_genomic_accession=1965504, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  347675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347675, CurrentID=0, Status='live', Symbol='EIF2S2P5', Aliases='', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 5', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 5', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114468118, end_position_on_the_genomic_accession=114469644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374448, CurrentID=0, Status='live', Symbol='LOC105374448', Aliases='', description='uncharacterized LOC105374448', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30898690, end_position_on_the_genomic_accession=30899514, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929378, CurrentID=0, Status='live', Symbol='LOC101929378', Aliases='', description='uncharacterized LOC101929378', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156020535, end_position_on_the_genomic_accession=156254931, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929380, CurrentID=0, Status='live', Symbol='LOC101929380', Aliases='', description='uncharacterized LOC101929380', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87120148, end_position_on_the_genomic_accession=87248021, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929384, CurrentID=0, Status='live', Symbol='LOC101929384', Aliases='', description='uncharacterized LOC101929384', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=203645, end_position_on_the_genomic_accession=205093, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929386, CurrentID=0, Status='live', Symbol='LOC101929386', Aliases='', description='uncharacterized LOC101929386', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29319191, end_position_on_the_genomic_accession=29352202, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  101929387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929387, CurrentID=0, Status='live', Symbol='LMF1-AS1', Aliases='LA16c-366D3.1', description='LMF1 antisense RNA 1', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=921033, end_position_on_the_genomic_accession=934524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929388, CurrentID=0, Status='live', Symbol='LOC101929388', Aliases='', description='uncharacterized LOC101929388', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202227318, end_position_on_the_genomic_accession=202238957, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929390, CurrentID=0, Status='live', Symbol='LINC01247', Aliases='', description='long intergenic non-protein coding RNA 1247', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6366010, end_position_on_the_genomic_accession=6375422, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874312, CurrentID=0, Status='live', Symbol='FNDC1-IT1', Aliases='', description='FNDC1 intronic transcript 1', other_designations='FNDC1 intronic transcript 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929395, CurrentID=0, Status='live', Symbol='LOC101929395', Aliases='', description='uncharacterized LOC101929395', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10753090, end_position_on_the_genomic_accession=10754027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929397, CurrentID=0, Status='live', Symbol='LINC01518', Aliases='', description='long intergenic non-protein coding RNA 1518', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42673013, end_position_on_the_genomic_accession=42691759, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929398, CurrentID=0, Status='live', Symbol='LOC101929398', Aliases='', description='uncharacterized LOC101929398', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929400, CurrentID=0, Status='live', Symbol='LOC101929400', Aliases='', description='uncharacterized LOC101929400', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7676601, end_position_on_the_genomic_accession=7687580, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374452, CurrentID=0, Status='live', Symbol='LOC105374452', Aliases='', description='uncharacterized LOC105374452', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32625002, end_position_on_the_genomic_accession=32626671, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929402, CurrentID=0, Status='live', Symbol='EXTL3-AS1', Aliases='C8orf50', description='EXTL3 antisense RNA 1', other_designations='EXTL3 antisense RNA 1 (non-protein coding)', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28698403, end_position_on_the_genomic_accession=28701464, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874314, CurrentID=0, Status='live', Symbol='TTLL7-IT1', Aliases='', description='TTLL7 intronic transcript 1', other_designations='TTLL7 intronic transcript 1 (non-protein coding)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929406, CurrentID=0, Status='live', Symbol='LOC101929406', Aliases='', description='uncharacterized LOC101929406', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30013952, end_position_on_the_genomic_accession=30037612, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929407, CurrentID=0, Status='live', Symbol='PTPRD-AS1', Aliases='', description='PTPRD antisense RNA 1', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=8858018, end_position_on_the_genomic_accession=8861727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929408, CurrentID=0, Status='live', Symbol='LOC101929408', Aliases='', description='uncharacterized LOC101929408', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75758348, end_position_on_the_genomic_accession=75763322, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929410, CurrentID=0, Status='live', Symbol='SUCLA2-AS1', Aliases='', description='SUCLA2 antisense RNA 1', other_designations='SUCLA2 antisense RNA 1 (non-protein coding)', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48001389, end_position_on_the_genomic_accession=48002614, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874315, CurrentID=0, Status='live', Symbol='SIK3-IT1', Aliases='', description='SIK3 intronic transcript 1', other_designations='SIK3 intronic transcript 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929412, CurrentID=0, Status='live', Symbol='LOC101929412', Aliases='', description='uncharacterized LOC101929412', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10493527, end_position_on_the_genomic_accession=10502728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929413, CurrentID=0, Status='live', Symbol='LOC101929413', Aliases='', description='uncharacterized LOC101929413', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10875333, end_position_on_the_genomic_accession=10909272, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929415, CurrentID=0, Status='live', Symbol='LOC101929415', Aliases='', description='uncharacterized LOC101929415', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56445807, end_position_on_the_genomic_accession=56552067, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929418, CurrentID=0, Status='live', Symbol='LOC101929418', Aliases='', description='uncharacterized LOC101929418', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29926232, end_position_on_the_genomic_accession=30146654, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  101929419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929419, CurrentID=0, Status='live', Symbol='LOC101929419', Aliases='', description='uncharacterized LOC101929419', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29288790, end_position_on_the_genomic_accession=29320461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929420, CurrentID=0, Status='live', Symbol='LOC101929420', Aliases='', description='uncharacterized LOC101929420', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168242938, end_position_on_the_genomic_accession=168262578, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929421, CurrentID=0, Status='live', Symbol='LOC101929421', Aliases='', description='cyclin-dependent kinase 2-associated protein 1 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120484171, end_position_on_the_genomic_accession=120484527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929422, CurrentID=0, Status='live', Symbol='LOC101929422', Aliases='', description='uncharacterized LOC101929422', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101120762, end_position_on_the_genomic_accession=101123545, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929423, CurrentID=0, Status='live', Symbol='MEF2C-AS1', Aliases='CTC-454M9.1', description='MEF2C antisense RNA 1', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88883330, end_position_on_the_genomic_accession=89032379, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929425, CurrentID=0, Status='live', Symbol='LOC101929425', Aliases='', description='uncharacterized LOC101929425', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38475148, end_position_on_the_genomic_accession=38500230, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929426, CurrentID=0, Status='live', Symbol='LOC101929426', Aliases='', description='uncharacterized LOC101929426', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18092899, end_position_on_the_genomic_accession=18094449, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929427, CurrentID=0, Status='live', Symbol='LOC101929427', Aliases='', description='uncharacterized LOC101929427', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126652778, end_position_on_the_genomic_accession=126682104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929428, CurrentID=0, Status='live', Symbol='PTPRD-AS2', Aliases='', description='PTPRD antisense RNA 2 (head to head)', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=10613206, end_position_on_the_genomic_accession=10620420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374457, CurrentID=0, Status='live', Symbol='LOC105374457', Aliases='', description='uncharacterized LOC105374457', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34306288, end_position_on_the_genomic_accession=34343658, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929432, CurrentID=0, Status='live', Symbol='LOC101929432', Aliases='', description='uncharacterized LOC101929432', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=689795, end_position_on_the_genomic_accession=711006, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929433, CurrentID=0, Status='live', Symbol='LOC101929433', Aliases='', description='uncharacterized LOC101929433', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65981868, end_position_on_the_genomic_accession=65988356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929435, CurrentID=0, Status='live', Symbol='LOC101929435', Aliases='', description='uncharacterized LOC101929435', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=103098229, end_position_on_the_genomic_accession=103146788, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929437, CurrentID=0, Status='live', Symbol='PRRX2-AS1', Aliases='', description='PRRX2 antisense RNA 1', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129712891, end_position_on_the_genomic_accession=129719224, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929439, CurrentID=0, Status='live', Symbol='LOC101929439', Aliases='', description='uncharacterized LOC101929439', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76174891, end_position_on_the_genomic_accession=76181486, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929440, CurrentID=0, Status='live', Symbol='LOC101929440', Aliases='LA16c-312E8.2', description='uncharacterized LOC101929440', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1408404, end_position_on_the_genomic_accession=1412251, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929441, CurrentID=0, Status='live', Symbol='LOC101929441', Aliases='', description='uncharacterized LOC101929441', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204141408, end_position_on_the_genomic_accession=204143009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374459, CurrentID=0, Status='live', Symbol='LOC105374459', Aliases='', description='uncharacterized LOC105374459', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34598070, end_position_on_the_genomic_accession=34618578, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929444, CurrentID=0, Status='live', Symbol='LOC101929444', Aliases='', description='uncharacterized LOC101929444', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31669811, end_position_on_the_genomic_accession=31671067, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  348645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348645, CurrentID=0, Status='live', Symbol='C22orf34', Aliases='', description='chromosome 22 open reading frame 34', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49414524, end_position_on_the_genomic_accession=49657542, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101929446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929446, CurrentID=0, Status='live', Symbol='LOC101929446', Aliases='', description='uncharacterized LOC101929446', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11826182, end_position_on_the_genomic_accession=11835013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874321, CurrentID=0, Status='live', Symbol='RASA3-IT1', Aliases='', description='RASA3 intronic transcript 1', other_designations='RASA3 intronic transcript 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929448, CurrentID=0, Status='live', Symbol='LOC101929448', Aliases='', description='uncharacterized LOC101929448', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103425042, end_position_on_the_genomic_accession=103439728, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929450, CurrentID=0, Status='live', Symbol='LOC101929450', Aliases='', description='uncharacterized LOC101929450', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29748309, end_position_on_the_genomic_accession=29798492, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929451, CurrentID=0, Status='live', Symbol='LOC101929451', Aliases='', description='dynamin-1 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85203818, end_position_on_the_genomic_accession=85204893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929452, CurrentID=0, Status='live', Symbol='LOC101929452', Aliases='', description='uncharacterized LOC101929452', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7062727, end_position_on_the_genomic_accession=7077880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874322, CurrentID=0, Status='live', Symbol='BACH1-IT2', Aliases='', description='BACH1 intronic transcript 2', other_designations='BACH1 intronic transcript 2 (non-protein coding)', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29372501, end_position_on_the_genomic_accession=29373980, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  348654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348654, CurrentID=0, Status='live', Symbol='GEN1', Aliases='Gen', description="GEN1, Holliday junction 5' flap endonuclease", other_designations="GEN1 Holliday junction 5' flap endonuclease|Gen endonuclease homolog 1|Gen homolog 1, endonuclease|Holliday junction resolvase", map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17753627, end_position_on_the_genomic_accession=17787104, orientation='plus', exon_count=16, OMIM=612449, no_hdr0=''),
  105374461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374461, CurrentID=0, Status='live', Symbol='LOC105374461', Aliases='', description='uncharacterized LOC105374461', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34998267, end_position_on_the_genomic_accession=35068525, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929457, CurrentID=0, Status='live', Symbol='LOC101929457', Aliases='', description='uncharacterized LOC101929457', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77525926, end_position_on_the_genomic_accession=77534104, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874323, CurrentID=0, Status='live', Symbol='HOXA10-AS', Aliases='HOXA-AS4', description='HOXA10 antisense RNA', other_designations='HOXA cluster antisense RNA 4 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27168899, end_position_on_the_genomic_accession=27171915, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929460, CurrentID=0, Status='live', Symbol='LOC101929460', Aliases='', description='uncharacterized LOC101929460', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169034171, end_position_on_the_genomic_accession=169035643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929461, CurrentID=0, Status='live', Symbol='LOC101929461', Aliases='', description='uncharacterized LOC101929461', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159392376, end_position_on_the_genomic_accession=159404632, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929463, CurrentID=0, Status='live', Symbol='LOC101929463', Aliases='', description='uncharacterized LOC101929463', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66435969, end_position_on_the_genomic_accession=66491427, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101929464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929464, CurrentID=0, Status='live', Symbol='LOC101929464', Aliases='', description='uncharacterized LOC101929464', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33350352, end_position_on_the_genomic_accession=33363245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874324, CurrentID=0, Status='live', Symbol='ATP2B2-IT1', Aliases='', description='ATP2B2 intronic transcript 1', other_designations='ATP2B2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373377, CurrentID=0, Status='live', Symbol='LOC105373377', Aliases='', description='putative paraneoplastic antigen-like protein 6B-like protein LOC649238', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153317683, end_position_on_the_genomic_accession=153320123, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929467, CurrentID=0, Status='live', Symbol='LURAP1L-AS1', Aliases='', description='LURAP1L antisense RNA 1', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12699999, end_position_on_the_genomic_accession=12814389, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929468, CurrentID=0, Status='live', Symbol='LOC101929468', Aliases='', description='uncharacterized LOC101929468', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104490965, end_position_on_the_genomic_accession=104697592, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101929469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929469, CurrentID=0, Status='live', Symbol='LOC101929469', Aliases='', description='uncharacterized protein ENSP00000372125', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2848243, end_position_on_the_genomic_accession=2857215, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929470, CurrentID=0, Status='live', Symbol='LOC101929470', Aliases='', description='uncharacterized LOC101929470', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29815004, end_position_on_the_genomic_accession=29828504, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929471, CurrentID=0, Status='live', Symbol='HINT2P1', Aliases='', description='histidine triad nucleotide binding protein 2 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=2362258, end_position_on_the_genomic_accession=2362867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929472, CurrentID=0, Status='live', Symbol='CTD-2350J17.1', Aliases='', description='uncharacterized LOC101929472', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15602189, end_position_on_the_genomic_accession=15615002, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929473, CurrentID=0, Status='live', Symbol='LOC101929473', Aliases='', description='uncharacterized LOC101929473', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=127067002, end_position_on_the_genomic_accession=127101306, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929475, CurrentID=0, Status='live', Symbol='LOC101929475', Aliases='', description='uncharacterized LOC101929475', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32982457, end_position_on_the_genomic_accession=33082102, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929476, CurrentID=0, Status='live', Symbol='LOC101929476', Aliases='', description='uncharacterized LOC101929476', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209372042, end_position_on_the_genomic_accession=209379690, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874326, CurrentID=0, Status='live', Symbol='DSCAM-IT1', Aliases='zag4', description='DSCAM intronic transcript 1', other_designations='DSCAM intronic transcript 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40618507, end_position_on_the_genomic_accession=40630767, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929478, CurrentID=0, Status='live', Symbol='LINGO1-AS2', Aliases='', description='LINGO1 antisense RNA 2', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77660051, end_position_on_the_genomic_accession=77668074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929479, CurrentID=0, Status='live', Symbol='LOC101929479', Aliases='', description='golgin subfamily A member 2-like protein 3', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85205389, end_position_on_the_genomic_accession=85234795, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929480, CurrentID=0, Status='live', Symbol='LOC101929480', Aliases='LA16c-329F2.1', description='uncharacterized LOC101929480', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1713529, end_position_on_the_genomic_accession=1715612, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929482, CurrentID=0, Status='live', Symbol='LOC101929482', Aliases='', description='uncharacterized LOC101929482', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=242140, end_position_on_the_genomic_accession=244984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874327, CurrentID=0, Status='live', Symbol='DSCR4-IT1', Aliases='', description='DSCR4 intronic transcript 1', other_designations='DSCR4 intronic transcript 1 (non-protein coding)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929484, CurrentID=0, Status='live', Symbol='LINC01615', Aliases='', description='long intergenic non-protein coding RNA 1615', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169157119, end_position_on_the_genomic_accession=169163523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929485, CurrentID=0, Status='live', Symbol='LOC101929485', Aliases='', description='uncharacterized LOC101929485', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106378119, end_position_on_the_genomic_accession=106631881, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  101929486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929486, CurrentID=0, Status='live', Symbol='LOC101929486', Aliases='', description='uncharacterized LOC101929486', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12865204, end_position_on_the_genomic_accession=12952519, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  101929488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929488, CurrentID=0, Status='live', Symbol='LOC101929488', Aliases='', description='uncharacterized LOC101929488', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57343787, end_position_on_the_genomic_accession=57364857, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100874328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874328, CurrentID=0, Status='live', Symbol='EIF2B5-IT1', Aliases='', description='EIF2B5 intronic transcript 1', other_designations='EIF2B5 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929490, CurrentID=0, Status='live', Symbol='LOC101929490', Aliases='', description='uncharacterized LOC101929490', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11677071, end_position_on_the_genomic_accession=11682039, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929492, CurrentID=0, Status='live', Symbol='LOC101929492', Aliases='', description='uncharacterized LOC101929492', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31275583, end_position_on_the_genomic_accession=31401108, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929494, CurrentID=0, Status='live', Symbol='LOC101929494', Aliases='', description='uncharacterized LOC101929494', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38450679, end_position_on_the_genomic_accession=38452425, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929495, CurrentID=0, Status='live', Symbol='LINC01339', Aliases='', description='long intergenic non-protein coding RNA 1339', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90158339, end_position_on_the_genomic_accession=90290071, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929497, CurrentID=0, Status='live', Symbol='LOC101929497', Aliases='', description='uncharacterized LOC101929497', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=127271070, end_position_on_the_genomic_accession=127337033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929498, CurrentID=0, Status='live', Symbol='LOC101929498', Aliases='', description='uncharacterized LOC101929498', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37297755, end_position_on_the_genomic_accession=37306964, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874330, CurrentID=0, Status='live', Symbol='LINC01014', Aliases='KCNMB2-IT1', description='long intergenic non-protein coding RNA 1014', other_designations='KCNMB2 intronic transcript 1 (non-protein coding)|hCG1646471', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178419201, end_position_on_the_genomic_accession=178457305, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929504, CurrentID=0, Status='live', Symbol='LOC101929504', Aliases='', description='uncharacterized LOC101929504', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169175304, end_position_on_the_genomic_accession=169182740, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929505, CurrentID=0, Status='live', Symbol='LOC101929505', Aliases='', description='uncharacterized LOC101929505', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16373361, end_position_on_the_genomic_accession=16441102, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929506, CurrentID=0, Status='live', Symbol='LOC101929506', Aliases='', description='uncharacterized LOC101929506', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79669474, end_position_on_the_genomic_accession=79679237, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874331, CurrentID=0, Status='live', Symbol='LARGE-IT1', Aliases='', description='LARGE intronic transcript 1', other_designations='LARGE intronic transcript 1 (non-protein coding)', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929510, CurrentID=0, Status='live', Symbol='LOC101929510', Aliases='', description='uncharacterized LOC101929510', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7661864, end_position_on_the_genomic_accession=7675004, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929511, CurrentID=0, Status='live', Symbol='LOC101929511', Aliases='', description='uncharacterized LOC101929511', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82290046, end_position_on_the_genomic_accession=82293210, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929512, CurrentID=0, Status='live', Symbol='LOC101929512', Aliases='', description='uncharacterized LOC101929512', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161223258, end_position_on_the_genomic_accession=161254643, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929516, CurrentID=0, Status='live', Symbol='LOC101929516', Aliases='', description='uncharacterized LOC101929516', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39565070, end_position_on_the_genomic_accession=39573203, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929517, CurrentID=0, Status='live', Symbol='LOC101929517', Aliases='', description='uncharacterized LOC101929517', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128526142, end_position_on_the_genomic_accession=128530383, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929518, CurrentID=0, Status='live', Symbol='LOC101929518', Aliases='', description='uncharacterized LOC101929518', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874333, CurrentID=0, Status='live', Symbol='OPCML-IT1', Aliases='', description='OPCML intronic transcript 1', other_designations='OPCML intronic transcript 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929520, CurrentID=0, Status='live', Symbol='LINC00993', Aliases='', description='long intergenic non-protein coding RNA 993', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37309186, end_position_on_the_genomic_accession=37347029, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929521, CurrentID=0, Status='live', Symbol='LOC101929521', Aliases='', description='uncharacterized LOC101929521', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36561411, end_position_on_the_genomic_accession=36575291, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929523, CurrentID=0, Status='live', Symbol='LOC101929523', Aliases='', description='uncharacterized LOC101929523', other_designations='XXyac-YX65C7_A.2', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169213212, end_position_on_the_genomic_accession=169239565, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929524, CurrentID=0, Status='live', Symbol='LOC101929524', Aliases='', description='uncharacterized LOC101929524', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16615926, end_position_on_the_genomic_accession=16629969, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874334, CurrentID=0, Status='live', Symbol='TCF7L1-IT1', Aliases='', description='TCF7L1 intronic transcript 1', other_designations='TCF7L1 intronic transcript 1 (non-protein coding)', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929526, CurrentID=0, Status='live', Symbol='LOC101929526', Aliases='', description='uncharacterized LOC101929526', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18674395, end_position_on_the_genomic_accession=18698708, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374473, CurrentID=0, Status='live', Symbol='LOC105374473', Aliases='', description='uncharacterized LOC105374473', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40617529, end_position_on_the_genomic_accession=40620321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929528, CurrentID=0, Status='live', Symbol='LOC101929528', Aliases='', description='uncharacterized LOC101929528', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58255771, end_position_on_the_genomic_accession=58272101, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929529, CurrentID=0, Status='live', Symbol='LOC101929529', Aliases='', description='uncharacterized LOC101929529', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105815145, end_position_on_the_genomic_accession=105827172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929530, CurrentID=0, Status='live', Symbol='FANCD2P2', Aliases='', description='Fanconi anemia complementation group D2 pseudogene 2', other_designations='Fanconi anemia, complementation group D2 pseudogene 2', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11870884, end_position_on_the_genomic_accession=11901607, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101929531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929531, CurrentID=0, Status='live', Symbol='LOC101929531', Aliases='', description='uncharacterized LOC101929531', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7723730, end_position_on_the_genomic_accession=7730724, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379861, CurrentID=0, Status='live', Symbol='LOC105379861', Aliases='', description='neuropeptide Y receptor type 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374474, CurrentID=0, Status='live', Symbol='LOC105374474', Aliases='', description='uncharacterized LOC105374474', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8847235, end_position_on_the_genomic_accession=8853617, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929534, CurrentID=0, Status='live', Symbol='LOC101929534', Aliases='', description='uncharacterized LOC101929534', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106810641, end_position_on_the_genomic_accession=106838386, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929535, CurrentID=0, Status='live', Symbol='LOC101929535', Aliases='', description='uncharacterized LOC101929535', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40712821, end_position_on_the_genomic_accession=40715369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929536, CurrentID=0, Status='live', Symbol='LOC101929536', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39788982, end_position_on_the_genomic_accession=39790171, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874336, CurrentID=0, Status='live', Symbol='ARID4B-IT1', Aliases='', description='ARID4B intronic transcript 1', other_designations='ARID4B intronic transcript 1 (non-protein coding)', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  348738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348738, CurrentID=0, Status='live', Symbol='C2orf48', Aliases='', description='chromosome 2 open reading frame 48', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10141382, end_position_on_the_genomic_accession=10211730, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929539, CurrentID=0, Status='live', Symbol='LINC01422', Aliases='', description='long intergenic non-protein coding RNA 1422', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26903292, end_position_on_the_genomic_accession=26920611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929540, CurrentID=0, Status='live', Symbol='LOC101929540', Aliases='', description='uncharacterized LOC101929540', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38403166, end_position_on_the_genomic_accession=38435338, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  101929541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929541, CurrentID=0, Status='live', Symbol='LOC101929541', Aliases='', description='uncharacterized LOC101929541', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212213002, end_position_on_the_genomic_accession=212285021, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929542, CurrentID=0, Status='live', Symbol='LOC101929542', Aliases='', description='uncharacterized LOC101929542', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37324879, end_position_on_the_genomic_accession=37326797, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874337, CurrentID=0, Status='live', Symbol='PLCB1-IT1', Aliases='', description='PLCB1 intronic transcript 1', other_designations='PLCB1 intronic transcript 1 (non-protein coding)', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929544, CurrentID=0, Status='live', Symbol='LOC101929544', Aliases='', description='uncharacterized LOC101929544', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17404096, end_position_on_the_genomic_accession=17441674, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929547, CurrentID=0, Status='live', Symbol='LOC101929547', Aliases='', description='uncharacterized LOC101929547', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19208994, end_position_on_the_genomic_accession=19210497, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929548, CurrentID=0, Status='live', Symbol='FOCAD-AS1', Aliases='', description='FOCAD antisense RNA 1', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20683302, end_position_on_the_genomic_accession=20684750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874338, CurrentID=0, Status='live', Symbol='TRIM36-IT1', Aliases='', description='TRIM36 intronic transcript 1', other_designations='TRIM36 intronic transcript 1 (non-protein coding)', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929550, CurrentID=0, Status='live', Symbol='LOC101929550', Aliases='', description='uncharacterized LOC101929550', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=35672163, end_position_on_the_genomic_accession=35710320, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  348751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348751, CurrentID=0, Status='live', Symbol='FTCDNL1', Aliases='FONG', description='formiminotransferase cyclodeaminase N-terminal like', other_designations='formiminotransferase N-terminal sub-domain containing', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199760536, end_position_on_the_genomic_accession=199851173, orientation='minus', exon_count=4, OMIM=614308, no_hdr0=''),
  101929552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929552, CurrentID=0, Status='live', Symbol='LOC101929552', Aliases='', description='uncharacterized LOC101929552', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82713905, end_position_on_the_genomic_accession=82716473, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929554, CurrentID=0, Status='live', Symbol='LOC101929554', Aliases='', description='suppressyn-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88233290, end_position_on_the_genomic_accession=88233894, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874339, CurrentID=0, Status='live', Symbol='ZNRF3-IT1', Aliases='', description='ZNRF3 intronic transcript 1', other_designations='ZNRF3 intronic transcript 1 (non-protein coding)', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929556, CurrentID=0, Status='live', Symbol='LOC101929556', Aliases='', description='uncharacterized LOC101929556', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134467806, end_position_on_the_genomic_accession=134492448, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929557, CurrentID=0, Status='live', Symbol='LINC01395', Aliases='', description='long intergenic non-protein coding RNA 1395', other_designations='TCONS_00019178', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129612116, end_position_on_the_genomic_accession=129617269, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  86614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=86614, CurrentID=0, Status='live', Symbol='HSFY1', Aliases='HSF2L, HSFY', description='heat shock transcription factor, Y-linked 1', other_designations='heat shock transcription factor 2-like protein', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18546685, end_position_on_the_genomic_accession=18588963, orientation='plus', exon_count=8, OMIM=400029, no_hdr0=''),
  101929559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929559, CurrentID=0, Status='live', Symbol='LINC00211', Aliases='NCRNA00211', description='long intergenic non-protein coding RNA 211', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37826247, end_position_on_the_genomic_accession=37875863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929560, CurrentID=0, Status='live', Symbol='LOC101929560', Aliases='', description='uncharacterized LOC101929560', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80344853, end_position_on_the_genomic_accession=80362464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  348761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348761, CurrentID=0, Status='live', Symbol='SPATA3-AS1', Aliases='', description='SPATA3 antisense RNA 1 (head to head)', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230984368, end_position_on_the_genomic_accession=230996032, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379862, CurrentID=0, Status='live', Symbol='LOC105379862', Aliases='', description='uncharacterized LOC105379862', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929563, CurrentID=0, Status='live', Symbol='LOC101929563', Aliases='', description='uncharacterized LOC101929563', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23500691, end_position_on_the_genomic_accession=23672399, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  101929565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929565, CurrentID=0, Status='live', Symbol='LOC101929565', Aliases='', description='uncharacterized LOC101929565', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212545694, end_position_on_the_genomic_accession=212556066, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929566, CurrentID=0, Status='live', Symbol='LOC101929566', Aliases='LA16c-325D7.2', description='uncharacterized LOC101929566', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2865837, end_position_on_the_genomic_accession=2868236, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929567, CurrentID=0, Status='live', Symbol='LOC101929567', Aliases='', description='uncharacterized LOC101929567', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8559833, end_position_on_the_genomic_accession=8583792, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929570, CurrentID=0, Status='live', Symbol='LOC101929570', Aliases='', description='uncharacterized LOC101929570', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162768936, end_position_on_the_genomic_accession=162797972, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929572, CurrentID=0, Status='live', Symbol='LOC101929572', Aliases='', description='uncharacterized LOC101929572', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19421026, end_position_on_the_genomic_accession=19424584, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874342, CurrentID=0, Status='live', Symbol='RNF216-IT1', Aliases='', description='RNF216 intronic transcript 1', other_designations='RNF216 intronic transcript 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5662432, end_position_on_the_genomic_accession=5680461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929574, CurrentID=0, Status='live', Symbol='LOC101929574', Aliases='', description='uncharacterized LOC101929574', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80529597, end_position_on_the_genomic_accession=80535942, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929575, CurrentID=0, Status='live', Symbol='LOC101929575', Aliases='', description='uncharacterized LOC101929575', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20254830, end_position_on_the_genomic_accession=20267873, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929576, CurrentID=0, Status='live', Symbol='LOC101929576', Aliases='', description='uncharacterized LOC101929576', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60884892, end_position_on_the_genomic_accession=60904825, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929577, CurrentID=0, Status='live', Symbol='LOC101929577', Aliases='', description='uncharacterized LOC101929577', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106003317, end_position_on_the_genomic_accession=106022478, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929578, CurrentID=0, Status='live', Symbol='LOC101929578', Aliases='', description='transcription factor CP2-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39237786, end_position_on_the_genomic_accession=39251019, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101929579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929579, CurrentID=0, Status='live', Symbol='LOC101929579', Aliases='', description='uncharacterized LOC101929579', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107380476, end_position_on_the_genomic_accession=107382159, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929580, CurrentID=0, Status='live', Symbol='LOC101929580', Aliases='', description='uncharacterized LOC101929580', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66847163, end_position_on_the_genomic_accession=66896593, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  101929581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929581, CurrentID=0, Status='live', Symbol='MZT1P2', Aliases='', description='mitotic spindle organizing protein 1 pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139227616, end_position_on_the_genomic_accession=139228660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929582, CurrentID=0, Status='live', Symbol='LINC01241', Aliases='', description='long intergenic non-protein coding RNA 1241', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25780056, end_position_on_the_genomic_accession=25812968, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929583, CurrentID=0, Status='live', Symbol='LOC101929583', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64638745, end_position_on_the_genomic_accession=64711065, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929584, CurrentID=0, Status='live', Symbol='LOC101929584', Aliases='', description='uncharacterized LOC101929584', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5233996, end_position_on_the_genomic_accession=5243151, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929586, CurrentID=0, Status='live', Symbol='LOC101929586', Aliases='', description='uncharacterized LOC101929586', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=80554609, end_position_on_the_genomic_accession=80562944, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929589, CurrentID=0, Status='live', Symbol='RPL7P61', Aliases='', description='ribosomal protein L7 pseudogene 61', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=163152225, end_position_on_the_genomic_accession=163153307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929590, CurrentID=0, Status='live', Symbol='NRG3-AS1', Aliases='C10orf100, Em:AC010157.1', description='NRG3 antisense RNA 1', other_designations='NRG3 antisense RNA 1 (non-protein coding)', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=82229014, end_position_on_the_genomic_accession=82232920, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874345, CurrentID=0, Status='live', Symbol='ZBTB40-IT1', Aliases='', description='ZBTB40 intronic transcript 1', other_designations='ZBTB40 intronic transcript 1 (non-protein coding)', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929592, CurrentID=0, Status='live', Symbol='LOC101929592', Aliases='', description='uncharacterized LOC101929592', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43709392, end_position_on_the_genomic_accession=43727343, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  348793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348793, CurrentID=0, Status='live', Symbol='WDR53', Aliases='', description='WD repeat domain 53', other_designations='WD domain, G-beta repeat-containing protein', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196554183, end_position_on_the_genomic_accession=196568610, orientation='minus', exon_count=5, OMIM=615110, no_hdr0=''),
  101929594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929594, CurrentID=0, Status='live', Symbol='LOC101929594', Aliases='', description='uncharacterized LOC101929594', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28514560, end_position_on_the_genomic_accession=28517252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929595, CurrentID=0, Status='live', Symbol='LOC101929595', Aliases='', description='uncharacterized LOC101929595', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107863479, end_position_on_the_genomic_accession=107932119, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929596, CurrentID=0, Status='live', Symbol='LOC101929596', Aliases='', description='uncharacterized LOC101929596', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38458637, end_position_on_the_genomic_accession=38515740, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929598, CurrentID=0, Status='live', Symbol='LOC101929598', Aliases='', description='zinc finger CCHC domain-containing protein 10 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181866309, end_position_on_the_genomic_accession=181867154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929599, CurrentID=0, Status='live', Symbol='LOC101929599', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70774097, end_position_on_the_genomic_accession=70778812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  348801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348801, CurrentID=0, Status='live', Symbol='LNP1', Aliases='NP3', description='leukemia NUP98 fusion partner 1', other_designations='', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100401193, end_position_on_the_genomic_accession=100456326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929602, CurrentID=0, Status='live', Symbol='IFT74-AS1', Aliases='', description='IFT74 antisense RNA 1', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26955780, end_position_on_the_genomic_accession=26956307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929605, CurrentID=0, Status='live', Symbol='LOC101929605', Aliases='', description='uncharacterized LOC101929605', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33495801, end_position_on_the_genomic_accession=33527959, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  348807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348807, CurrentID=0, Status='live', Symbol='CFAP100', Aliases='CCDC37, MIA1', description='cilia and flagella associated protein 100', other_designations='coiled-coil domain containing 37', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126394893, end_position_on_the_genomic_accession=126436556, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  348808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348808, CurrentID=0, Status='live', Symbol='NPHP3-AS1', Aliases='NCRNA00119', description='NPHP3 antisense RNA 1', other_designations='NPHP3 antisense RNA 1 (non-protein coding)', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132722342, end_position_on_the_genomic_accession=132874206, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100874348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874348, CurrentID=0, Status='live', Symbol='ZDHHC20P3', Aliases='', description='zinc finger, DHHC-type containing 20 pseudogene 3', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75228314, end_position_on_the_genomic_accession=75228652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929612, CurrentID=0, Status='live', Symbol='LOC101929612', Aliases='', description='uncharacterized LOC101929612', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215481830, end_position_on_the_genomic_accession=215499834, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929613, CurrentID=0, Status='live', Symbol='LOC101929613', Aliases='LA16c-380H5.3', description='uncharacterized LOC101929613', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2998955, end_position_on_the_genomic_accession=3000813, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929614, CurrentID=0, Status='live', Symbol='LOC101929614', Aliases='', description='uncharacterized LOC101929614', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170147068, end_position_on_the_genomic_accession=170152997, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79981, CurrentID=0, Status='live', Symbol='FRMD1', Aliases='bA164L23.1', description='FERM domain containing 1', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168055745, end_position_on_the_genomic_accession=168101514, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105374488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374488, CurrentID=0, Status='live', Symbol='LOC105374488', Aliases='', description='uncharacterized LOC105374488', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11259800, end_position_on_the_genomic_accession=11280653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929620, CurrentID=0, Status='live', Symbol='LINC01243', Aliases='', description='long intergenic non-protein coding RNA 1243', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31371611, end_position_on_the_genomic_accession=31381490, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874350, CurrentID=0, Status='live', Symbol='HOXB-AS2', Aliases='', description='HOXB cluster antisense RNA 2', other_designations='HOXB cluster antisense RNA 2 (non-protein coding)', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929622, CurrentID=0, Status='live', Symbol='LOC101929622', Aliases='', description='uncharacterized LOC101929622', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37717579, end_position_on_the_genomic_accession=37733493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929623, CurrentID=0, Status='live', Symbol='LINC01215', Aliases='', description='long intergenic non-protein coding RNA 1215', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108125064, end_position_on_the_genomic_accession=108138610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929624, CurrentID=0, Status='live', Symbol='LINC01519', Aliases='', description='long intergenic non-protein coding RNA 1519', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85193421, end_position_on_the_genomic_accession=85198938, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  348825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348825, CurrentID=0, Status='live', Symbol='TPRXL', Aliases='', description='tetra-peptide repeat homeobox-like', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13937307, end_position_on_the_genomic_accession=14065984, orientation='plus', exon_count=3, OMIM=611167, no_hdr0=''),
  101929626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929626, CurrentID=0, Status='live', Symbol='LOC101929626', Aliases='', description='uncharacterized LOC101929626', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46133371, end_position_on_the_genomic_accession=46139081, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874351, CurrentID=0, Status='live', Symbol='HOXB-AS4', Aliases='', description='HOXB cluster antisense RNA 4', other_designations='HOXB cluster antisense RNA 4 (non-protein coding)', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929628, CurrentID=0, Status='live', Symbol='LOC101929628', Aliases='', description='uncharacterized LOC101929628', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61785072, end_position_on_the_genomic_accession=62052364, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  101929631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929631, CurrentID=0, Status='live', Symbol='LOC101929631', Aliases='', description='uncharacterized LOC101929631', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218043505, end_position_on_the_genomic_accession=218059140, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929633, CurrentID=0, Status='live', Symbol='LOC101929633', Aliases='', description='uncharacterized LOC101929633', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=164840749, end_position_on_the_genomic_accession=164849396, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929634, CurrentID=0, Status='live', Symbol='LOC101929634', Aliases='', description='keratin-associated protein 5-1-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104661120, end_position_on_the_genomic_accession=104665618, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374491, CurrentID=0, Status='live', Symbol='LOC105374491', Aliases='', description='uncharacterized LOC105374491', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40717276, end_position_on_the_genomic_accession=40732352, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929637, CurrentID=0, Status='live', Symbol='NTM-AS1', Aliases='C11orf39', description='NTM antisense RNA 1', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131662129, end_position_on_the_genomic_accession=131663556, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929638, CurrentID=0, Status='live', Symbol='LOC101929638', Aliases='', description='uncharacterized LOC101929638', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29180623, end_position_on_the_genomic_accession=29205822, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929639, CurrentID=0, Status='live', Symbol='B4GALT1-AS1', Aliases='', description='B4GALT1 antisense RNA 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33166948, end_position_on_the_genomic_accession=33179983, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  348840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348840, CurrentID=0, Status='live', Symbol='ANKRD18DP', Aliases='', description='ankyrin repeat domain 18D, pseudogene', other_designations='ankyrin repeat domain 18A pseudogene', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198057531, end_position_on_the_genomic_accession=198080671, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101929643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929643, CurrentID=0, Status='live', Symbol='LOC101929643', Aliases='', description='uncharacterized LOC101929643', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9565641, end_position_on_the_genomic_accession=9575508, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929644, CurrentID=0, Status='live', Symbol='LOC101929644', Aliases='', description='uncharacterized LOC101929644', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108652808, end_position_on_the_genomic_accession=108668570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874354, CurrentID=0, Status='live', Symbol='ST6GAL2-IT1', Aliases='', description='ST6GAL2 intronic transcript 1', other_designations='ST6GAL2 intronic transcript 1 (non-protein coding)', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929646, CurrentID=0, Status='live', Symbol='LOC101929646', Aliases='', description='uncharacterized LOC101929646', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85431943, end_position_on_the_genomic_accession=85432772, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131613, CurrentID=0, Status='live', Symbol='LOC100131613', Aliases='', description='PRO1454', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=69001745, end_position_on_the_genomic_accession=69002101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929650, CurrentID=0, Status='live', Symbol='LOC101929650', Aliases='', description='uncharacterized LOC101929650', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83143915, end_position_on_the_genomic_accession=83145259, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929651, CurrentID=0, Status='live', Symbol='LINC01398', Aliases='', description='long intergenic non-protein coding RNA 1398', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46446673, end_position_on_the_genomic_accession=46449704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929653, CurrentID=0, Status='live', Symbol='LOC101929653', Aliases='', description='uncharacterized LOC101929653', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131533857, end_position_on_the_genomic_accession=131540867, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101929654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929654, CurrentID=0, Status='live', Symbol='LOC101929654', Aliases='', description='uncharacterized LOC101929654', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112516807, end_position_on_the_genomic_accession=112518219, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929655, CurrentID=0, Status='live', Symbol='TMC3-AS1', Aliases='', description='TMC3 antisense RNA 1', other_designations='', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81324333, end_position_on_the_genomic_accession=81443076, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100874356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874356, CurrentID=0, Status='live', Symbol='TSPAN9-IT1', Aliases='', description='TSPAN9 intronic transcript 1', other_designations='TSPAN9 intronic transcript 1 (non-protein coding)', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929660, CurrentID=0, Status='live', Symbol='LOC101929660', Aliases='', description='uncharacterized LOC101929660', other_designations='CTD-2118P12.1', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29143503, end_position_on_the_genomic_accession=29172975, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101929662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929662, CurrentID=0, Status='live', Symbol='LOC101929662', Aliases='', description='uncharacterized LOC101929662', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85432863, end_position_on_the_genomic_accession=85448956, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929663, CurrentID=0, Status='live', Symbol='LINC01427', Aliases='', description='long intergenic non-protein coding RNA 1427', other_designations='', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22280228, end_position_on_the_genomic_accession=22282801, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929664, CurrentID=0, Status='live', Symbol='HORMAD2-AS1', Aliases='MTMR3-AS1', description='HORMAD2 antisense RNA 1', other_designations='NONHSAG033653', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30008742, end_position_on_the_genomic_accession=30080480, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929665, CurrentID=0, Status='live', Symbol='UBE2R2-AS1', Aliases='', description='UBE2R2 antisense RNA 1', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33782070, end_position_on_the_genomic_accession=33819644, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929667, CurrentID=0, Status='live', Symbol='LOC101929667', Aliases='', description='uncharacterized LOC101929667', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40510764, end_position_on_the_genomic_accession=40545583, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929669, CurrentID=0, Status='live', Symbol='GEMIN2P1', Aliases='', description='gem nuclear organelle associated protein 2 pseudogene 1', other_designations='gem (nuclear organelle) associated protein 2 pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153698257, end_position_on_the_genomic_accession=153718306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929670, CurrentID=0, Status='live', Symbol='LOC101929670', Aliases='', description='uncharacterized LOC101929670', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170395611, end_position_on_the_genomic_accession=170400256, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131617, CurrentID=0, Status='live', Symbol='LOC100131617', Aliases='', description='v-ets avian erythroblastosis virus E26 oncogene homolog 2 pseudogene', other_designations='v-ets erythroblastosis virus E26 oncogene homolog 2 pseudogene', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135584279, end_position_on_the_genomic_accession=135585160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874359, CurrentID=0, Status='live', Symbol='TMEM212-IT1', Aliases='', description='TMEM212 intronic transcript 1', other_designations='TMEM212 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929679, CurrentID=0, Status='live', Symbol='LOC101929679', Aliases='', description='uncharacterized LOC101929679', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85754941, end_position_on_the_genomic_accession=85758725, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929680, CurrentID=0, Status='live', Symbol='LOC101929680', Aliases='', description='uncharacterized LOC101929680', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=166081531, end_position_on_the_genomic_accession=166301784, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100874360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874360, CurrentID=0, Status='live', Symbol='GRPEL2P1', Aliases='', description='GrpE-like 2, mitochondrial (E. coli) pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88978349, end_position_on_the_genomic_accession=88978764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929682, CurrentID=0, Status='live', Symbol='LOC101929682', Aliases='', description='40S ribosomal protein S3a-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78708710, end_position_on_the_genomic_accession=78716080, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929684, CurrentID=0, Status='live', Symbol='LINC01520', Aliases='', description='long intergenic non-protein coding RNA 1520', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85449592, end_position_on_the_genomic_accession=85492001, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929685, CurrentID=0, Status='live', Symbol='LINC01384', Aliases='', description='long intergenic non-protein coding RNA 1384', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22587522, end_position_on_the_genomic_accession=22607517, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874361, CurrentID=0, Status='live', Symbol='PPP2R2B-IT1', Aliases='', description='PPP2R2B intronic transcript 1', other_designations='PPP2R2B intronic transcript 1 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146914207, end_position_on_the_genomic_accession=146919506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102191832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102191832, CurrentID=0, Status='live', Symbol='DEAR', Aliases='DEspR', description='dual endothelin-1(VEGFsp)/angiotensin II receptor pseudogene', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152337655, end_position_on_the_genomic_accession=152338096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929690, CurrentID=0, Status='live', Symbol='LINC01583', Aliases='', description='long intergenic non-protein coding RNA 1583', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82088594, end_position_on_the_genomic_accession=82097694, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929691, CurrentID=0, Status='live', Symbol='LOC101929691', Aliases='', description='uncharacterized LOC101929691', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10288460, end_position_on_the_genomic_accession=10302449, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929692, CurrentID=0, Status='live', Symbol='LOC101929692', Aliases='', description='uncharacterized LOC101929692', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170712352, end_position_on_the_genomic_accession=170736419, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929693, CurrentID=0, Status='live', Symbol='RARA-AS1', Aliases='', description='RARA antisense RNA 1', other_designations='CTD-2267D19.2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40340867, end_position_on_the_genomic_accession=40343136, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929694, CurrentID=0, Status='live', Symbol='LOC101929694', Aliases='', description='uncharacterized LOC101929694', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112736449, end_position_on_the_genomic_accession=112749319, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929696, CurrentID=0, Status='live', Symbol='LOC101929696', Aliases='', description='uncharacterized LOC101929696', other_designations='CTB-35F21.2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139741347, end_position_on_the_genomic_accession=139746223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929697, CurrentID=0, Status='live', Symbol='LOC101929697', Aliases='', description='uncharacterized LOC101929697', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195836226, end_position_on_the_genomic_accession=195854303, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929698, CurrentID=0, Status='live', Symbol='LOC101929698', Aliases='', description='uncharacterized LOC101929698', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32509959, end_position_on_the_genomic_accession=32520285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927531, CurrentID=0, Status='live', Symbol='LOC101927531', Aliases='', description='uncharacterized LOC101927531', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929700, CurrentID=0, Status='live', Symbol='LOC101929700', Aliases='', description='uncharacterized LOC101929700', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40746553, end_position_on_the_genomic_accession=40823415, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929701, CurrentID=0, Status='live', Symbol='LINC01584', Aliases='', description='long intergenic non-protein coding RNA 1584', other_designations='TCONS_00024001', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86083345, end_position_on_the_genomic_accession=86116717, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929703, CurrentID=0, Status='live', Symbol='LOC101929703', Aliases='', description='uncharacterized LOC101929703', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155570570, end_position_on_the_genomic_accession=155571589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929704, CurrentID=0, Status='live', Symbol='LOC101929704', Aliases='', description='uncharacterized LOC101929704', other_designations='CTD-2517O10.6', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34837714, end_position_on_the_genomic_accession=34839275, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929705, CurrentID=0, Status='live', Symbol='LOC101929705', Aliases='', description='uncharacterized LOC101929705', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43995723, end_position_on_the_genomic_accession=44074652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929707, CurrentID=0, Status='live', Symbol='LOC101929707', Aliases='', description='uncharacterized LOC101929707', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22995690, end_position_on_the_genomic_accession=23030820, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101929708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929708, CurrentID=0, Status='live', Symbol='LOC101929708', Aliases='', description='Ig kappa chain V-I region Walker-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929709, CurrentID=0, Status='live', Symbol='LOC101929709', Aliases='', description='uncharacterized LOC101929709', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89717399, end_position_on_the_genomic_accession=89757711, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  348910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348910, CurrentID=0, Status='live', Symbol='RAF1P1', Aliases='', description='RAF1 pseudogene 1', other_designations='v-raf-1 murine leukemia viral oncogene homolog 1 pseudogene 1', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10254926, end_position_on_the_genomic_accession=10256659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131624, CurrentID=0, Status='live', Symbol='WASF1P1', Aliases='', description='WASF1 pseudogene 1', other_designations='WAS protein family, member 1 pseudogene', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108276644, end_position_on_the_genomic_accession=108277943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929715, CurrentID=0, Status='live', Symbol='LOC101929715', Aliases='', description='uncharacterized LOC101929715', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10449728, end_position_on_the_genomic_accession=10455552, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929716, CurrentID=0, Status='live', Symbol='ARMC2-AS1', Aliases='', description='ARMC2 antisense RNA 1', other_designations='ARMC2 antisense RNA 1 (non-protein coding)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108908140, end_position_on_the_genomic_accession=108924103, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874366, CurrentID=0, Status='live', Symbol='HOXC13-AS', Aliases='HOXC-AS5', description='HOXC13 antisense RNA', other_designations='HOXC cluster antisense RNA 5 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53935328, end_position_on_the_genomic_accession=53939643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929718, CurrentID=0, Status='live', Symbol='LOC101929718', Aliases='', description='uncharacterized LOC101929718', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20995837, end_position_on_the_genomic_accession=20999163, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929719, CurrentID=0, Status='live', Symbol='LOC101929719', Aliases='', description='uncharacterized LOC101929719', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140157319, end_position_on_the_genomic_accession=140168785, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929721, CurrentID=0, Status='live', Symbol='LOC101929721', Aliases='', description='uncharacterized LOC101929721', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=49257411, end_position_on_the_genomic_accession=49269294, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929722, CurrentID=0, Status='live', Symbol='LOC101929722', Aliases='', description='uncharacterized LOC101929722', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66139117, end_position_on_the_genomic_accession=66141674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79999, CurrentID=0, Status='live', Symbol='LOC79999', Aliases='', description='uncharacterized LOC79999', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19154524, end_position_on_the_genomic_accession=19158872, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  86781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=86781, CurrentID=0, Status='live', Symbol='RPS2P7', Aliases='RPS2L1, RPS2_27_1706, dJ688G8.1', description='ribosomal protein S2 pseudogene 7', other_designations='ribosomal protein S2-like 1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45640451, end_position_on_the_genomic_accession=45641393, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  348926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348926, CurrentID=0, Status='live', Symbol='FAM86EP', Aliases='', description='family with sequence similarity 86 member E, pseudogene', other_designations='family with sequence similarity 86, member E, pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3941760, end_position_on_the_genomic_accession=3955428, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929727, CurrentID=0, Status='live', Symbol='LOC101929727', Aliases='', description='uncharacterized LOC101929727', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88113382, end_position_on_the_genomic_accession=88380366, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101929728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929728, CurrentID=0, Status='live', Symbol='LOC101929728', Aliases='', description='uncharacterized LOC101929728', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30667582, end_position_on_the_genomic_accession=30669016, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874368, CurrentID=0, Status='live', Symbol='CACNA1C-IT1', Aliases='', description='CACNA1C intronic transcript 1', other_designations='CACNA1C intronic transcript 1 (non-protein coding)', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929730, CurrentID=0, Status='live', Symbol='LINC01352', Aliases='', description='long intergenic non-protein coding RNA 1352', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220829255, end_position_on_the_genomic_accession=220832429, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374507, CurrentID=0, Status='live', Symbol='LOC105374507', Aliases='', description='uncharacterized LOC105374507', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17073859, end_position_on_the_genomic_accession=17082615, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  348932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348932, CurrentID=0, Status='live', Symbol='SLC6A18', Aliases='Xtrp2', description='solute carrier family 6 (neutral amino acid transporter), member 18', other_designations='sodium channel-like protein|sodium- and chloride-dependent transporter XTRP2|solute carrier family 6 (neurotransmitter transporter), member 18|solute carrier family 6, member 18|system B(0) neutral amino acid transporter AT3', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1225355, end_position_on_the_genomic_accession=1246189, orientation='plus', exon_count=12, OMIM=610300, no_hdr0=''),
  101929733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929733, CurrentID=0, Status='live', Symbol='LOC101929733', Aliases='', description='uncharacterized LOC101929733', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10878267, end_position_on_the_genomic_accession=10885118, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929735, CurrentID=0, Status='live', Symbol='LOC101929735', Aliases='', description='uncharacterized LOC101929735', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=22849082, end_position_on_the_genomic_accession=22951948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929736, CurrentID=0, Status='live', Symbol='LINC01372', Aliases='', description='long intergenic non-protein coding RNA 1372', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67335976, end_position_on_the_genomic_accession=67340024, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  348938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348938, CurrentID=0, Status='live', Symbol='NIPAL4', Aliases='ARCI6, ICHTHYIN, ICHYN', description='NIPA-like domain containing 4', other_designations='NIPA-like protein 4|non-imprinted in Prader-Willi/Angelman syndrome region protein 4', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157460019, end_position_on_the_genomic_accession=157474722, orientation='plus', exon_count=6, OMIM=609383, no_hdr0=''),
  101929741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929741, CurrentID=0, Status='live', Symbol='LOC101929741', Aliases='', description='uncharacterized LOC101929741', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118664087, end_position_on_the_genomic_accession=118685341, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929743, CurrentID=0, Status='live', Symbol='LINC01586', Aliases='', description='long intergenic non-protein coding RNA 1586', other_designations='TCONS_00023252', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88585571, end_position_on_the_genomic_accession=88605110, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929744, CurrentID=0, Status='live', Symbol='LOC101929744', Aliases='', description='phosphopantothenate--cysteine ligase pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136419847, end_position_on_the_genomic_accession=136420424, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929745, CurrentID=0, Status='live', Symbol='LOC101929745', Aliases='', description='uncharacterized LOC101929745', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38148480, end_position_on_the_genomic_accession=38153715, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929746, CurrentID=0, Status='live', Symbol='ZNF341-AS1', Aliases='', description='ZNF341 antisense RNA 1', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33787373, end_position_on_the_genomic_accession=33811099, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929747, CurrentID=0, Status='live', Symbol='LOC101929747', Aliases='', description='uncharacterized LOC101929747', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96934080, end_position_on_the_genomic_accession=96936094, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929748, CurrentID=0, Status='live', Symbol='LOC101929748', Aliases='', description='transcription initiation factor TFIID subunit 4-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92805385, end_position_on_the_genomic_accession=92808644, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929749, CurrentID=0, Status='live', Symbol='LOC101929749', Aliases='', description='carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 2 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929750, CurrentID=0, Status='live', Symbol='LOC101929750', Aliases='', description='uncharacterized LOC101929750', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221073718, end_position_on_the_genomic_accession=221143011, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101929752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929752, CurrentID=0, Status='live', Symbol='LOC101929752', Aliases='', description='uncharacterized LOC101929752', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11719840, end_position_on_the_genomic_accession=11723554, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874372, CurrentID=0, Status='live', Symbol='ARHGAP26-IT1', Aliases='', description='ARHGAP26 intronic transcript 1', other_designations='ARHGAP26 intronic transcript 1 (non-protein coding)', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143192500, end_position_on_the_genomic_accession=143194166, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929754, CurrentID=0, Status='live', Symbol='LOC101929754', Aliases='', description='uncharacterized LOC101929754', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114453593, end_position_on_the_genomic_accession=114520132, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647983, CurrentID=0, Status='live', Symbol='LOC647983', Aliases='', description='uncharacterized LOC647983', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=14981272, end_position_on_the_genomic_accession=14981748, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929756, CurrentID=0, Status='live', Symbol='LOC101929756', Aliases='', description='uncharacterized LOC101929756', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77056, end_position_on_the_genomic_accession=83368, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  348958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348958, CurrentID=0, Status='live', Symbol='LOC348958', Aliases='', description='mitochondrial ribosomal protein L10 pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119680172, end_position_on_the_genomic_accession=119682008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874373, CurrentID=0, Status='live', Symbol='STARD13-IT1', Aliases='', description='STARD13 intronic transcript 1', other_designations='STARD13 intronic transcript 1 (non-protein coding)', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929760, CurrentID=0, Status='live', Symbol='PIK3IP1-AS1', Aliases='', description='PIK3IP1 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31335349, end_position_on_the_genomic_accession=31338021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929762, CurrentID=0, Status='live', Symbol='LOC101929762', Aliases='', description='uncharacterized LOC101929762', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119192773, end_position_on_the_genomic_accession=119212644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929763, CurrentID=0, Status='live', Symbol='LOC101929763', Aliases='', description='uncharacterized LOC101929763', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43171749, end_position_on_the_genomic_accession=43197822, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929765, CurrentID=0, Status='live', Symbol='LINC01585', Aliases='', description='long intergenic non-protein coding RNA 1585', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90660234, end_position_on_the_genomic_accession=90664954, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373379, CurrentID=0, Status='live', Symbol='LOC105373379', Aliases='', description='uncharacterized LOC105373379', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153401583, end_position_on_the_genomic_accession=153406606, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929767, CurrentID=0, Status='live', Symbol='LOC101929767', Aliases='', description='uncharacterized LOC101929767', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43148366, end_position_on_the_genomic_accession=43170403, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929768, CurrentID=0, Status='live', Symbol='OSMR-AS1', Aliases='CTD-2127H9.1', description='OSMR antisense RNA 1 (head to head)', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38693213, end_position_on_the_genomic_accession=38845829, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929769, CurrentID=0, Status='live', Symbol='LINC01063', Aliases='', description='long intergenic non-protein coding RNA 1063', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196631498, end_position_on_the_genomic_accession=196632696, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929770, CurrentID=0, Status='live', Symbol='LOC101929770', Aliases='', description='uncharacterized LOC101929770', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44728010, end_position_on_the_genomic_accession=44829669, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874375, CurrentID=0, Status='live', Symbol='FAM155A-IT1', Aliases='', description='FAM155A intronic transcript 1', other_designations='FAM155A intronic transcript 1 (non-protein coding)', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107787361, end_position_on_the_genomic_accession=107835458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929772, CurrentID=0, Status='live', Symbol='LOC101929772', Aliases='', description='uncharacterized LOC101929772', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929773, CurrentID=0, Status='live', Symbol='LOC101929773', Aliases='', description='UDP-glucuronosyltransferase 2B10-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929774, CurrentID=0, Status='live', Symbol='LOC101929774', Aliases='', description='uncharacterized LOC101929774', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929775, CurrentID=0, Status='live', Symbol='LOC101929775', Aliases='', description='C-Jun-amino-terminal kinase-interacting protein 1-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929777, CurrentID=0, Status='live', Symbol='LOC101929777', Aliases='', description='uncharacterized LOC101929777', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46830595, end_position_on_the_genomic_accession=46833154, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374515, CurrentID=0, Status='live', Symbol='LOC105374515', Aliases='', description='uncharacterized LOC105374515', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20658446, end_position_on_the_genomic_accession=20674272, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  348980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348980, CurrentID=0, Status='live', Symbol='HCN1', Aliases='BCNG-1, BCNG1, EIEE24, HAC-2', description='hyperpolarization activated cyclic nucleotide gated potassium channel 1', other_designations='brain cyclic nucleotide-gated channel 1', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=45254950, end_position_on_the_genomic_accession=45696118, orientation='minus', exon_count=8, OMIM=602780, no_hdr0=''),
  101929784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929784, CurrentID=0, Status='live', Symbol='LOC101929784', Aliases='', description='gem-associated protein 2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929787, CurrentID=0, Status='live', Symbol='LOC101929787', Aliases='', description='uncharacterized LOC101929787', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148865454, end_position_on_the_genomic_accession=148871271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929788, CurrentID=0, Status='live', Symbol='LOC101929788', Aliases='', description='uncharacterized LOC101929788', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120393545, end_position_on_the_genomic_accession=120411356, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874378, CurrentID=0, Status='live', Symbol='PTGER4P3', Aliases='', description='prostaglandin E receptor 4 pseudogene 3', other_designations='prostaglandin E receptor 4 (subtype EP4) pseudogene 3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63775896, end_position_on_the_genomic_accession=63776277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929790, CurrentID=0, Status='live', Symbol='LOC101929790', Aliases='', description='uncharacterized LOC101929790', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149107223, end_position_on_the_genomic_accession=149124744, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374517, CurrentID=0, Status='live', Symbol='LOC105374517', Aliases='', description='uncharacterized LOC105374517', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41983356, end_position_on_the_genomic_accession=41987312, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  348995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348995, CurrentID=0, Status='live', Symbol='NUP43', Aliases='bA350J20.1, p42', description='nucleoporin 43kDa', other_designations='nup107-160 subcomplex subunit Nup43', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149724315, end_position_on_the_genomic_accession=149746572, orientation='minus', exon_count=12, OMIM=608141, no_hdr0=''),
  101929796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929796, CurrentID=0, Status='live', Symbol='LOC101929796', Aliases='', description='notch homolog 2 N-terminal-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120723955, end_position_on_the_genomic_accession=120805451, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374518, CurrentID=0, Status='live', Symbol='LOC105374518', Aliases='', description='uncharacterized LOC105374518', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21345028, end_position_on_the_genomic_accession=21365258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929798, CurrentID=0, Status='live', Symbol='LOC101929798', Aliases='', description='myomegalin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121028075, end_position_on_the_genomic_accession=121053518, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101929800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929800, CurrentID=0, Status='live', Symbol='LOC101929800', Aliases='', description='uncharacterized LOC101929800', other_designations='', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68321050, end_position_on_the_genomic_accession=68330436, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100874380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874380, CurrentID=0, Status='live', Symbol='DPPA5P2', Aliases='', description='developmental pluripotency associated 5 pseudogene 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45566774, end_position_on_the_genomic_accession=45567043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929802, CurrentID=0, Status='live', Symbol='LOC101929802', Aliases='', description='uncharacterized LOC101929802', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929804, CurrentID=0, Status='live', Symbol='LOC101929804', Aliases='', description='histone-lysine N-methyltransferase 2C-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929805, CurrentID=0, Status='live', Symbol='LOC101929805', Aliases='', description='myomegalin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145135059, end_position_on_the_genomic_accession=145149891, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101929806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929806, CurrentID=0, Status='live', Symbol='LOC101929806', Aliases='', description='zonadhesin-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131884018, end_position_on_the_genomic_accession=131889796, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100874381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874381, CurrentID=0, Status='live', Symbol='NPIPP1', Aliases='', description='nuclear pore complex interacting protein pseudogene 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15104298, end_position_on_the_genomic_accession=15123498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374520, CurrentID=0, Status='live', Symbol='LOC105374520', Aliases='', description='uncharacterized LOC105374520', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22635675, end_position_on_the_genomic_accession=22686904, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929812, CurrentID=0, Status='live', Symbol='LOC101929812', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874382, CurrentID=0, Status='live', Symbol='USP9YP12', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 12', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23946965, end_position_on_the_genomic_accession=23947123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929814, CurrentID=0, Status='live', Symbol='LOC101929814', Aliases='', description='uncharacterized LOC101929814', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143300730, end_position_on_the_genomic_accession=143308909, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374521, CurrentID=0, Status='live', Symbol='LOC105374521', Aliases='', description='uncharacterized LOC105374521', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22733068, end_position_on_the_genomic_accession=22744294, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929816, CurrentID=0, Status='live', Symbol='LOC101929816', Aliases='', description='ankyrin repeat domain-containing protein 30A-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  132386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132386, CurrentID=0, Status='live', Symbol='LOC132386', Aliases='', description='epithelial cell adhesion molecule pseudogene', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181522660, end_position_on_the_genomic_accession=181523421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929818, CurrentID=0, Status='live', Symbol='LOC101929818', Aliases='', description='killer cell immunoglobulin-like receptor 3DL1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54515664, end_position_on_the_genomic_accession=54527372, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929819, CurrentID=0, Status='live', Symbol='LOC101929819', Aliases='', description='uncharacterized LOC101929819', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929820, CurrentID=0, Status='live', Symbol='LOC101929820', Aliases='', description='uncharacterized LOC101929820', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144099901, end_position_on_the_genomic_accession=144100607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374522, CurrentID=0, Status='live', Symbol='LOC105374522', Aliases='', description='uncharacterized LOC105374522', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23332413, end_position_on_the_genomic_accession=23409367, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929823, CurrentID=0, Status='live', Symbol='LOC101929823', Aliases='', description='uncharacterized LOC101929823', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68098, end_position_on_the_genomic_accession=108364, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  101929826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929826, CurrentID=0, Status='live', Symbol='LOC101929826', Aliases='', description='uncharacterized LOC101929826', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144625606, end_position_on_the_genomic_accession=144642375, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376871, CurrentID=0, Status='live', Symbol='LOC105376871', Aliases='', description='uncharacterized LOC105376871', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24305928, end_position_on_the_genomic_accession=24321888, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929828, CurrentID=0, Status='live', Symbol='LOC101929828', Aliases='', description='uncharacterized LOC101929828', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=109546, end_position_on_the_genomic_accession=111566, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929829, CurrentID=0, Status='live', Symbol='LOC101929829', Aliases='', description='cytochrome P450 family 2 subfamily D member 6 pseudogene', other_designations='cytochrome P450 2D6-like|cytochrome P450, family 2, subfamily D, polypeptide 6 pseudogene', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929830, CurrentID=0, Status='live', Symbol='LOC101929830', Aliases='', description='uncharacterized LOC101929830', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874385, CurrentID=0, Status='live', Symbol='HNRNPKP5', Aliases='', description='heterogeneous nuclear ribonucleoprotein K pseudogene 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14877662, end_position_on_the_genomic_accession=14877917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647996, CurrentID=0, Status='live', Symbol='LOC647996', Aliases='', description='chromosome 9 open reading frame 172 pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131827545, end_position_on_the_genomic_accession=131830136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929835, CurrentID=0, Status='live', Symbol='LOC101929835', Aliases='', description='putative protein FAM90A15P', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874386, CurrentID=0, Status='live', Symbol='OSTCP5', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45069980, end_position_on_the_genomic_accession=45070429, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929839, CurrentID=0, Status='live', Symbol='LOC101929839', Aliases='', description='uncharacterized LOC101929839', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15819006, end_position_on_the_genomic_accession=15821252, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377744, CurrentID=0, Status='live', Symbol='LOC105377744', Aliases='', description='uncharacterized LOC105377744', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175824372, end_position_on_the_genomic_accession=175827049, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874387, CurrentID=0, Status='live', Symbol='OSTCP6', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56975607, end_position_on_the_genomic_accession=56976063, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929846, CurrentID=0, Status='live', Symbol='LOC101929846', Aliases='', description='uncharacterized LOC101929846', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63630677, end_position_on_the_genomic_accession=63654745, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874388, CurrentID=0, Status='live', Symbol='OSTCP8', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 8', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33968077, end_position_on_the_genomic_accession=33968517, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374527, CurrentID=0, Status='live', Symbol='LOC105374527', Aliases='', description='uncharacterized LOC105374527', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23769094, end_position_on_the_genomic_accession=23782448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874389, CurrentID=0, Status='live', Symbol='CNOT4P1', Aliases='', description='CCR4-NOT transcription complex subunit 4 pseudogene 1', other_designations='CCR4-NOT transcription complex, subunit 4 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20921664, end_position_on_the_genomic_accession=20923233, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374528, CurrentID=0, Status='live', Symbol='LOC105374528', Aliases='', description='uncharacterized LOC105374528', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23718578, end_position_on_the_genomic_accession=23769046, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101929860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929860, CurrentID=0, Status='live', Symbol='LOC101929860', Aliases='', description='uncharacterized LOC101929860', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41731957, end_position_on_the_genomic_accession=41733513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874390, CurrentID=0, Status='live', Symbol='RAC1P7', Aliases='', description='ras-related C3 botulinum toxin substrate 1 pseudogene 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234026873, end_position_on_the_genomic_accession=234027227, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929862, CurrentID=0, Status='live', Symbol='LOC101929862', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929863, CurrentID=0, Status='live', Symbol='LOC101929863', Aliases='', description='uncharacterized LOC101929863', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45137459, end_position_on_the_genomic_accession=45139104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929866, CurrentID=0, Status='live', Symbol='LOC101929866', Aliases='', description='uncharacterized LOC101929866', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45179754, end_position_on_the_genomic_accession=45191664, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874391, CurrentID=0, Status='live', Symbol='OSTCP7', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 7', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35005300, end_position_on_the_genomic_accession=35005741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374530, CurrentID=0, Status='live', Symbol='LOC105374530', Aliases='', description='uncharacterized LOC105374530', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24589191, end_position_on_the_genomic_accession=24598366, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929870, CurrentID=0, Status='live', Symbol='LOC101929870', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929872, CurrentID=0, Status='live', Symbol='LOC101929872', Aliases='', description='uncharacterized LOC101929872', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100750225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750225, CurrentID=0, Status='live', Symbol='PCAT1', Aliases='PCA1, PCAT-1', description='prostate cancer associated transcript 1 (non-protein coding)', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127013154, end_position_on_the_genomic_accession=127021014, orientation='plus', exon_count=2, OMIM=616043, no_hdr0=''),
  349075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349075, CurrentID=0, Status='live', Symbol='ZNF713', Aliases='', description='zinc finger protein 713', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55887277, end_position_on_the_genomic_accession=55942225, orientation='plus', exon_count=7, OMIM=616181, no_hdr0=''),
  101929876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929876, CurrentID=0, Status='live', Symbol='LOC101929876', Aliases='', description='40S ribosomal protein S26', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929878, CurrentID=0, Status='live', Symbol='LOC101929878', Aliases='', description='splicing factor 45-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61463460, end_position_on_the_genomic_accession=61465050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874393, CurrentID=0, Status='live', Symbol='EIF2AP1', Aliases='', description='eukaryotic translation initiation factor 2A pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150484059, end_position_on_the_genomic_accession=150485144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929881, CurrentID=0, Status='live', Symbol='LOC101929881', Aliases='', description='uncharacterized LOC101929881', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929882, CurrentID=0, Status='live', Symbol='LOC101929882', Aliases='', description='uncharacterized LOC101929882', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10034216, end_position_on_the_genomic_accession=10043441, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100750238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750238, CurrentID=0, Status='live', Symbol='EJM9', Aliases='', description='Epilepsy, juvenile myoclonic, susceptiblity to, 9', other_designations='', map_location='2q33-q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614280, no_hdr0=''),
  101929891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929891, CurrentID=0, Status='live', Symbol='LOC101929891', Aliases='', description='uncharacterized LOC101929891', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929892, CurrentID=0, Status='live', Symbol='LOC101929892', Aliases='', description='uncharacterized LOC101929892', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100750245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750245, CurrentID=0, Status='live', Symbol='SNAPC5P1', Aliases='', description='small nuclear RNA activating complex polypeptide 5 pseudogene 1', other_designations='small nuclear RNA activating complex, polypeptide 5, 19kDa pseudogene 1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6347087, end_position_on_the_genomic_accession=6347367, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100750246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750246, CurrentID=0, Status='live', Symbol='HIF1A-AS1', Aliases="5'aHIF-1A, 5'aHIF1alpha", description='HIF1A antisense RNA 1', other_designations='HIF1A antisense RNA 1 (non-protein coding)', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61681041, end_position_on_the_genomic_accession=61695823, orientation='minus', exon_count=2, OMIM=614528, no_hdr0=''),
  100750247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750247, CurrentID=0, Status='live', Symbol='HIF1A-AS2', Aliases="3'aHIF-1A, aHIF", description='HIF1A antisense RNA 2', other_designations='HIF1A antisense RNA 2 (non-protein coding)', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61747039, end_position_on_the_genomic_accession=61749089, orientation='minus', exon_count=1, OMIM=614529, no_hdr0=''),
  100874396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874396, CurrentID=0, Status='live', Symbol='HNRNPA1P24', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 24', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149999645, end_position_on_the_genomic_accession=150000690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929898, CurrentID=0, Status='live', Symbol='LOC101929898', Aliases='', description='uncharacterized LOC101929898', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=693527, end_position_on_the_genomic_accession=695809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374535, CurrentID=0, Status='live', Symbol='LOC105374535', Aliases='', description='uncharacterized LOC105374535', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25190212, end_position_on_the_genomic_accession=25233864, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101929901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929901, CurrentID=0, Status='live', Symbol='LOC101929901', Aliases='', description='uncharacterized LOC101929901', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41375004, end_position_on_the_genomic_accession=41380923, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929902, CurrentID=0, Status='live', Symbol='LOC101929902', Aliases='', description='uncharacterized LOC101929902', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131656, CurrentID=0, Status='live', Symbol='SDCBPP3', Aliases='', description='syndecan binding protein (syntenin) pseudogene 3', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40890647, end_position_on_the_genomic_accession=40891537, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929908, CurrentID=0, Status='live', Symbol='LOC101929908', Aliases='', description='uncharacterized LOC101929908', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117921189, end_position_on_the_genomic_accession=117935488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  349114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349114, CurrentID=0, Status='live', Symbol='LINC00265', Aliases='NCRNA00265', description='long intergenic non-protein coding RNA 265', other_designations='NCRNA00265-1', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39733568, end_position_on_the_genomic_accession=39794623, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100874399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874399, CurrentID=0, Status='live', Symbol='IMPDH1P10', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201137319, end_position_on_the_genomic_accession=201139003, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374538, CurrentID=0, Status='live', Symbol='LOC105374538', Aliases='', description='uncharacterized LOC105374538', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25900814, end_position_on_the_genomic_accession=25914093, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929918, CurrentID=0, Status='live', Symbol='LOC101929918', Aliases='', description='uncharacterized LOC101929918', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34044872, end_position_on_the_genomic_accession=34052509, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100271289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271289, CurrentID=0, Status='live', Symbol='RPL22P17', Aliases='RPL22_7_1091', description='ribosomal protein L22 pseudogene 17', other_designations='', map_location='10q24.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102887842, end_position_on_the_genomic_accession=102888214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874400, CurrentID=0, Status='live', Symbol='HNRNPA1P18', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 18', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68301605, end_position_on_the_genomic_accession=68303302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929922, CurrentID=0, Status='live', Symbol='LOC101929922', Aliases='', description='uncharacterized LOC101929922', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929923, CurrentID=0, Status='live', Symbol='LOC101929923', Aliases='', description='uncharacterized LOC101929923', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=36101116, end_position_on_the_genomic_accession=36102205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481961, CurrentID=0, Status='live', Symbol='RNU2-55P', Aliases='', description='RNA, U2 small nuclear 55, pseudogene', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929926, CurrentID=0, Status='live', Symbol='LOC101929926', Aliases='', description='uncharacterized LOC101929926', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126898878, end_position_on_the_genomic_accession=126902097, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101929927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929927, CurrentID=0, Status='live', Symbol='LOC101929927', Aliases='', description='putative protein FAM90A16P/FAM90A17P', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929928, CurrentID=0, Status='live', Symbol='LOC101929928', Aliases='', description='protein SPHAR-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  260403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260403, CurrentID=0, Status='live', Symbol='MMEDF', Aliases='', description='Macrocephaly with multiple epiphyseal dysplasia and distinctive facies', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607131, no_hdr0=''),
  101929932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929932, CurrentID=0, Status='live', Symbol='LOC101929932', Aliases='', description='AP-3 complex subunit sigma-1 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874402, CurrentID=0, Status='live', Symbol='HNRNPA1P19', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 19', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20771066, end_position_on_the_genomic_accession=20771423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374541, CurrentID=0, Status='live', Symbol='LOC105374541', Aliases='', description='uncharacterized LOC105374541', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26151965, end_position_on_the_genomic_accession=26158684, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  349136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349136, CurrentID=0, Status='live', Symbol='WDR86', Aliases='', description='WD repeat domain 86', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151378541, end_position_on_the_genomic_accession=151410705, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101929937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929937, CurrentID=0, Status='live', Symbol='LOC101929937', Aliases='', description='uncharacterized LOC101929937', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=346266, end_position_on_the_genomic_accession=347981, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874403, CurrentID=0, Status='live', Symbol='ERHP2', Aliases='', description='enhancer of rudimentary homolog (Drosophila) pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54016484, end_position_on_the_genomic_accession=54016777, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929941, CurrentID=0, Status='live', Symbol='LOC101929941', Aliases='', description='neural Wiskott-Aldrich syndrome protein-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929943, CurrentID=0, Status='live', Symbol='LOC101929943', Aliases='', description='tropomyosin alpha-3 chain-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929944, CurrentID=0, Status='live', Symbol='LOC101929944', Aliases='', description='uncharacterized LOC101929944', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874404, CurrentID=0, Status='live', Symbol='LINC00328-2P', Aliases='NCRNA00328-2P, NCRNA00328B', description='long intergenic non-protein coding RNA 328-2, pseudogene', other_designations='non-protein coding RNA 328-2, pseudogene|non-protein coding RNA 328B', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18809830, end_position_on_the_genomic_accession=18810048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929946, CurrentID=0, Status='live', Symbol='ZEB1-AS1', Aliases='', description='ZEB1 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31307719, end_position_on_the_genomic_accession=31320696, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929947, CurrentID=0, Status='live', Symbol='LOC101929947', Aliases='', description='uncharacterized LOC101929947', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  349149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349149, CurrentID=0, Status='live', Symbol='GJC3', Aliases='CX29, CX30.2, CX31.3, GJE1', description='gap junction protein gamma 3', other_designations='connexin 29|connexin-30.2|connexin-31.3|gap junction epsilon-1 protein|gap junction protein, gamma 3, 30.2kDa', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99923269, end_position_on_the_genomic_accession=99935091, orientation='minus', exon_count=3, OMIM=611925, no_hdr0=''),
  101929950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929950, CurrentID=0, Status='live', Symbol='LOC101929950', Aliases='', description='puromycin-sensitive aminopeptidase-like protein', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38078236, end_position_on_the_genomic_accession=38118280, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874405, CurrentID=0, Status='live', Symbol='PPIAP27', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 27', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20956176, end_position_on_the_genomic_accession=20956437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  349152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349152, CurrentID=0, Status='live', Symbol='DPY19L2P2', Aliases='', description='DPY19L2 pseudogene 2', other_designations='dpy-19-like 2 pseudogene 2', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103175013, end_position_on_the_genomic_accession=103280466, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  101929955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929955, CurrentID=0, Status='live', Symbol='LOC101929955', Aliases='', description='uncharacterized LOC101929955', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  349158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349158, CurrentID=0, Status='live', Symbol='CYCSP20', Aliases='HCP20', description='cytochrome c, somatic pseudogene 20', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129117513, end_position_on_the_genomic_accession=129117827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929959, CurrentID=0, Status='live', Symbol='LOC101929959', Aliases='', description='putative BMS1-like protein ENSP00000382360', other_designations='ribosome biogenesis protein BMS1 homolog', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40463754, end_position_on_the_genomic_accession=40584232, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  349160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349160, CurrentID=0, Status='live', Symbol='LOC349160', Aliases='', description='uncharacterized LOC349160', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136898773, end_position_on_the_genomic_accession=137164341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874407, CurrentID=0, Status='live', Symbol='SNRPGP3', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129477789, end_position_on_the_genomic_accession=129478222, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929964, CurrentID=0, Status='live', Symbol='LOC101929964', Aliases='', description='uncharacterized LOC101929964', other_designations='CTC-228N24.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128002346, end_position_on_the_genomic_accession=128004112, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374546, CurrentID=0, Status='live', Symbol='LOC105374546', Aliases='', description='uncharacterized LOC105374546', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26859806, end_position_on_the_genomic_accession=26860655, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929967, CurrentID=0, Status='live', Symbol='LOC101929967', Aliases='', description='uncharacterized LOC101929967', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74611828, end_position_on_the_genomic_accession=74621059, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874408, CurrentID=0, Status='live', Symbol='SNRPGP7', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 7', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28460256, end_position_on_the_genomic_accession=28460361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929970, CurrentID=0, Status='live', Symbol='LOC101929970', Aliases='', description='CHRNA7-FAM7A fusion protein', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100750325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750325, CurrentID=0, Status='live', Symbol='RCAN3AS', Aliases='', description='RCAN3 antisense', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24496333, end_position_on_the_genomic_accession=24502360, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100750326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750326, CurrentID=0, Status='live', Symbol='HTT-AS', Aliases='HTT-AS1, HTTAS', description='HTT antisense RNA (head to head)', other_designations='HTT antisense RNA 1 (head to head)|HTT antisense RNA 1 (non-protein coding)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3063246, end_position_on_the_genomic_accession=3074514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100750328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750328, CurrentID=0, Status='live', Symbol='ANIB11', Aliases='', description='Aneurysm, intracranial berry, 11', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614252, no_hdr0=''),
  100750329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100750329, CurrentID=0, Status='live', Symbol='TET18P', Aliases='', description='Tetrasomy 18p', other_designations='', map_location='18p', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614290, no_hdr0=''),
  100129358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129358, CurrentID=0, Status='live', Symbol='LOC100129358', Aliases='', description='kelch like family member 5 pseudogene', other_designations='kelch-like family member 5 pseudogene', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23926922, end_position_on_the_genomic_accession=23932056, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929981, CurrentID=0, Status='live', Symbol='LOC101929981', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929983, CurrentID=0, Status='live', Symbol='PRAMEF27', Aliases='', description='PRAME family member 27', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13049808, end_position_on_the_genomic_accession=13053659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929984, CurrentID=0, Status='live', Symbol='LOC101929984', Aliases='', description='uncharacterized LOC101929984', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128844182, end_position_on_the_genomic_accession=128848474, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100189237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100189237, CurrentID=0, Status='live', Symbol='TRL-CAA1-2', Aliases='TRNAL22', description='transfer RNA-Leu (CAA) 1-2', other_designations='transfer RNA leucine 22 (anticodon CAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28941053, end_position_on_the_genomic_accession=28941157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929988, CurrentID=0, Status='live', Symbol='LOC101929988', Aliases='', description='uncharacterized LOC101929988', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33852064, end_position_on_the_genomic_accession=33865133, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929989, CurrentID=0, Status='live', Symbol='LOC101929989', Aliases='', description='lipopolysaccharide-induced tumor necrosis factor-alpha factor-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8882334, end_position_on_the_genomic_accession=8892650, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929990, CurrentID=0, Status='live', Symbol='LOC101929990', Aliases='', description='collagen alpha-1(XXVII) chain-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929991, CurrentID=0, Status='live', Symbol='LOC101929991', Aliases='', description='zinc finger protein 93-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374551, CurrentID=0, Status='live', Symbol='LOC105374551', Aliases='', description='uncharacterized LOC105374551', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27967270, end_position_on_the_genomic_accession=27977542, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  349196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349196, CurrentID=0, Status='live', Symbol='LINC00965', Aliases='', description='long intergenic non-protein coding RNA 965', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7260619, end_position_on_the_genomic_accession=7263492, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929997, CurrentID=0, Status='live', Symbol='LOC101929997', Aliases='', description='uncharacterized LOC101929997', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22331202, end_position_on_the_genomic_accession=22333141, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  349198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349198, CurrentID=0, Status='live', Symbol='CYCSP3', Aliases='HCP21, HS7', description='cytochrome c, somatic pseudogene 3', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33969561, end_position_on_the_genomic_accession=33969878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874413, CurrentID=0, Status='live', Symbol='SNRPEP5', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 5', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8603057, end_position_on_the_genomic_accession=8603504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374552, CurrentID=0, Status='live', Symbol='LOC105374552', Aliases='', description='uncharacterized LOC105374552', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27902568, end_position_on_the_genomic_accession=28003072, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131284, CurrentID=0, Status='live', Symbol='LOC100131284', Aliases='', description='proteasome 26S subunit, non-ATPase 14 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 14 pseudogene', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=123762207, end_position_on_the_genomic_accession=123763072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930006, CurrentID=0, Status='live', Symbol='LOC101930006', Aliases='', description='putative protein FRMPD2-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930009, CurrentID=0, Status='live', Symbol='LOC101930009', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930010, CurrentID=0, Status='live', Symbol='LOC101930010', Aliases='', description='uncharacterized LOC101930010', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56843986, end_position_on_the_genomic_accession=56864338, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930012, CurrentID=0, Status='live', Symbol='LOC101930012', Aliases='', description='uncharacterized LOC101930012', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930013, CurrentID=0, Status='live', Symbol='LOC101930013', Aliases='', description='polycystin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  341883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341883, CurrentID=0, Status='live', Symbol='LRRC9', Aliases='', description='leucine rich repeat containing 9', other_designations='leucine-rich repeat-containing 9', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59919713, end_position_on_the_genomic_accession=60063559, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  101930023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930023, CurrentID=0, Status='live', Symbol='LOC101930023', Aliases='', description='uncharacterized LOC101930023', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92146489, end_position_on_the_genomic_accession=92363858, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100131676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131676, CurrentID=0, Status='live', Symbol='LOC100131676', Aliases='', description='clathrin, light chain A pseudogene', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7606835, end_position_on_the_genomic_accession=7607621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930026, CurrentID=0, Status='live', Symbol='LOC101930026', Aliases='', description='uncharacterized LOC101930026', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61917497, end_position_on_the_genomic_accession=61973057, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930027, CurrentID=0, Status='live', Symbol='LOC101930027', Aliases='', description='FERM and PDZ domain-containing protein 2-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930028, CurrentID=0, Status='live', Symbol='LOC101930028', Aliases='', description='uncharacterized LOC101930028', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188768999, end_position_on_the_genomic_accession=188818810, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101930031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930031, CurrentID=0, Status='live', Symbol='LOC101930031', Aliases='', description='uncharacterized LOC101930031', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930032, CurrentID=0, Status='live', Symbol='LOC101930032', Aliases='', description='uncharacterized LOC101930032', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374558, CurrentID=0, Status='live', Symbol='LOC105374558', Aliases='', description='uncharacterized LOC105374558', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28522448, end_position_on_the_genomic_accession=28585703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874420, CurrentID=0, Status='live', Symbol='ARL2BPP5', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 5', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 5|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 5|ADP-ribosylation factor-like 2 binding protein pseudogene 5', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=162761833, end_position_on_the_genomic_accession=162762313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930042, CurrentID=0, Status='live', Symbol='LOC101930042', Aliases='', description='Krueppel-like factor 2 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372508, CurrentID=0, Status='live', Symbol='LOC105372508', Aliases='', description='uncharacterized LOC105372508', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3417036, end_position_on_the_genomic_accession=3430842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101930048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930048, CurrentID=0, Status='live', Symbol='LOC101930048', Aliases='', description='uncharacterized LOC101930048', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2494265, end_position_on_the_genomic_accession=2507110, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374560, CurrentID=0, Status='live', Symbol='LOC105374560', Aliases='', description='uncharacterized LOC105374560', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29118545, end_position_on_the_genomic_accession=29156591, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874422, CurrentID=0, Status='live', Symbol='SNRPFP3', Aliases='', description='small nuclear ribonucleoprotein polypeptide F pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69101832, end_position_on_the_genomic_accession=69102056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374561, CurrentID=0, Status='live', Symbol='LOC105374561', Aliases='', description='uncharacterized LOC105374561', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29214286, end_position_on_the_genomic_accession=29292295, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100874423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874423, CurrentID=0, Status='live', Symbol='SNRPEP8', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 8', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=53038285, end_position_on_the_genomic_accession=53038540, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930061, CurrentID=0, Status='live', Symbol='LOC101930061', Aliases='', description='uncharacterized LOC101930061', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930062, CurrentID=0, Status='live', Symbol='LOC101930062', Aliases='', description='uncharacterized LOC101930062', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131158, CurrentID=0, Status='live', Symbol='LOC100131158', Aliases='', description='eukaryotic translation initiation factor 2 subunit gamma pseudogene', other_designations='eukaryotic translation initiation factor 2, subunit 3 gamma, 52kDa pseudogene', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112797050, end_position_on_the_genomic_accession=112798296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930067, CurrentID=0, Status='live', Symbol='LOC101930067', Aliases='', description='uncharacterized LOC101930067', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930071, CurrentID=0, Status='live', Symbol='LOC101930071', Aliases='', description='uncharacterized LOC101930071', other_designations='CTB-50E14.6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42397128, end_position_on_the_genomic_accession=42408452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930072, CurrentID=0, Status='live', Symbol='FP248', Aliases='', description='uncharacterized LOC101930072', other_designations='CTD-3064M3.7', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141437479, end_position_on_the_genomic_accession=141437954, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930073, CurrentID=0, Status='live', Symbol='LOC101930073', Aliases='', description='putative uncharacterized protein ENSP00000383309', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930075, CurrentID=0, Status='live', Symbol='LOC101930075', Aliases='', description='polycystin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930078, CurrentID=0, Status='live', Symbol='LOC101930078', Aliases='', description='double homeobox protein 4-like protein 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131685, CurrentID=0, Status='live', Symbol='LOC100131685', Aliases='', description='enabled homolog (Drosophila) pseudogene', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190830965, end_position_on_the_genomic_accession=190832173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930080, CurrentID=0, Status='live', Symbol='LOC101930080', Aliases='', description='GPS, PLAT and transmembrane domain-containing protein FLJ00285-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930081, CurrentID=0, Status='live', Symbol='LOC101930081', Aliases='', description='uncharacterized LOC101930081', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874427, CurrentID=0, Status='live', Symbol='SNRPGP12', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 12', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110139214, end_position_on_the_genomic_accession=110139426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930085, CurrentID=0, Status='live', Symbol='LOC101930085', Aliases='', description='uncharacterized LOC101930085', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35694329, end_position_on_the_genomic_accession=35699413, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874428, CurrentID=0, Status='live', Symbol='SNRPGP13', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 13', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38502445, end_position_on_the_genomic_accession=38502621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930090, CurrentID=0, Status='live', Symbol='LOC101930090', Aliases='', description='contactin-associated protein-like 3B', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61063582, end_position_on_the_genomic_accession=61114015, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101930091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930091, CurrentID=0, Status='live', Symbol='LOC101930091', Aliases='', description='uncharacterized LOC101930091', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64519143, end_position_on_the_genomic_accession=64637840, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101930094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930094, CurrentID=0, Status='live', Symbol='LOC101930094', Aliases='', description='uncharacterized LOC101930094', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42496452, end_position_on_the_genomic_accession=42497443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101930095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930095, CurrentID=0, Status='live', Symbol='LOC101930095', Aliases='', description='histone H3.3-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374568, CurrentID=0, Status='live', Symbol='LOC105374568', Aliases='', description='uncharacterized LOC105374568', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42920435, end_position_on_the_genomic_accession=42922943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101930100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930100, CurrentID=0, Status='live', Symbol='LOC101930100', Aliases='', description='uncharacterized LOC101930100', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9325037, end_position_on_the_genomic_accession=9367242, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374569, CurrentID=0, Status='live', Symbol='LOC105374569', Aliases='', description='uncharacterized LOC105374569', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42972722, end_position_on_the_genomic_accession=43022358, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101930105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930105, CurrentID=0, Status='live', Symbol='LOC101930105', Aliases='', description='zinc finger protein 839-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3854393, end_position_on_the_genomic_accession=3886550, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874431, CurrentID=0, Status='live', Symbol='HMGN1P8', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 8', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 8', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167694355, end_position_on_the_genomic_accession=167694642, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930109, CurrentID=0, Status='live', Symbol='LOC101930109', Aliases='', description='uncharacterized LOC101930109', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56391248, end_position_on_the_genomic_accession=56408798, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101930111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930111, CurrentID=0, Status='live', Symbol='LOC101930111', Aliases='', description='putative exonuclease GOR-like protein', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930112, CurrentID=0, Status='live', Symbol='LOC101930112', Aliases='', description='uncharacterized LOC101930112', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89490955, end_position_on_the_genomic_accession=89498516, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930114, CurrentID=0, Status='live', Symbol='LOC101930114', Aliases='', description='uncharacterized LOC101930114', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209661359, end_position_on_the_genomic_accession=209757599, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100131691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131691, CurrentID=0, Status='live', Symbol='MZF1-AS1', Aliases='', description='MZF1 antisense RNA 1', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58559186, end_position_on_the_genomic_accession=58574797, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101930116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930116, CurrentID=0, Status='live', Symbol='LOC101930116', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1, A2/B1 homolog', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930117, CurrentID=0, Status='live', Symbol='LOC101930117', Aliases='', description='uncharacterized LOC101930117', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930119, CurrentID=0, Status='live', Symbol='LOC101930119', Aliases='', description='uncharacterized LOC101930119', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930120, CurrentID=0, Status='live', Symbol='LOC101930120', Aliases='', description='uncharacterized LOC101930120', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99913197, end_position_on_the_genomic_accession=100037998, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  87178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=87178, CurrentID=0, Status='live', Symbol='PNPT1', Aliases='COXPD13, DFNB70, OLD35, PNPASE, old-35', description='polyribonucleotide nucleotidyltransferase 1', other_designations="3'-5' RNA exonuclease OLD35|PNPase 1|PNPase old-35|polynucleotide phosphorylase 1|polynucleotide phosphorylase-like protein", map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55634063, end_position_on_the_genomic_accession=55693910, orientation='minus', exon_count=29, OMIM=610316, no_hdr0=''),
  101930123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930123, CurrentID=0, Status='live', Symbol='LOC101930123', Aliases='', description='eukaryotic elongation factor 2 kinase', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930125, CurrentID=0, Status='live', Symbol='LOC101930125', Aliases='', description='uncharacterized LOC101930125', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930126, CurrentID=0, Status='live', Symbol='LOC101930126', Aliases='', description='uncharacterized LOC101930126', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374573, CurrentID=0, Status='live', Symbol='LOC105374573', Aliases='', description='uncharacterized LOC105374573', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44816972, end_position_on_the_genomic_accession=44822357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101930129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930129, CurrentID=0, Status='live', Symbol='LOC101930129', Aliases='', description='uncharacterized LOC101930129', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129363, CurrentID=0, Status='live', Symbol='LOC100129363', Aliases='', description='uncharacterized LOC100129363', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=86996922, end_position_on_the_genomic_accession=86997488, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874435, CurrentID=0, Status='live', Symbol='HIGD1AP4', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 4', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223398265, end_position_on_the_genomic_accession=223398544, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930132, CurrentID=0, Status='live', Symbol='LOC101930132', Aliases='', description='pollen-specific leucine-rich repeat extensin-like protein 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  349334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349334, CurrentID=0, Status='live', Symbol='FOXD4L4', Aliases='FOXD4L2, FOXD4b, bA460E7.2', description='forkhead box D4-like 4', other_designations='FOXD4-like 4|forkhead box D4-like 2|forkhead box protein D4b|myeloid factor-gamma|winged helix factor-1|winged helix transcription factor beta', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65737146, end_position_on_the_genomic_accession=65738396, orientation='plus', exon_count=1, OMIM=611085, no_hdr0=''),
  100874436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874436, CurrentID=0, Status='live', Symbol='HIGD1AP6', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 6', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 6', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=75302296, end_position_on_the_genomic_accession=75302574, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374575, CurrentID=0, Status='live', Symbol='LOC105374575', Aliases='', description='uncharacterized LOC105374575', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45304938, end_position_on_the_genomic_accession=45311582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874437, CurrentID=0, Status='live', Symbol='HIGD1AP7', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 7', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 7', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40115662, end_position_on_the_genomic_accession=40115943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930145, CurrentID=0, Status='live', Symbol='LOC101930145', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874438, CurrentID=0, Status='live', Symbol='HMGN1P10', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 10', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 10', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136609050, end_position_on_the_genomic_accession=136609602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930150, CurrentID=0, Status='live', Symbol='NF1P9', Aliases='', description='neurofibromin 1 pseudogene 9', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21499792, end_position_on_the_genomic_accession=21508548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930151, CurrentID=0, Status='live', Symbol='LOC101930151', Aliases='', description='uncharacterized LOC101930151', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874439, CurrentID=0, Status='live', Symbol='HMGN1P13', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 13', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 13', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111572102, end_position_on_the_genomic_accession=111572239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930160, CurrentID=0, Status='live', Symbol='LOC101930160', Aliases='', description='uncharacterized LOC101930160', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874440, CurrentID=0, Status='live', Symbol='HMGN1P14', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 14', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 14', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111846050, end_position_on_the_genomic_accession=111846566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930162, CurrentID=0, Status='live', Symbol='LOC101930162', Aliases='', description='uncharacterized LOC101930162', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3367486, end_position_on_the_genomic_accession=3371641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101930164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930164, CurrentID=0, Status='live', Symbol='LOC101930164', Aliases='', description='uncharacterized LOC101930164', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40141294, end_position_on_the_genomic_accession=40158559, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101930165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930165, CurrentID=0, Status='live', Symbol='LOC101930165', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930168, CurrentID=0, Status='live', Symbol='LOC101930168', Aliases='', description='putative aquaporin-7-like protein 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930169, CurrentID=0, Status='live', Symbol='LOC101930169', Aliases='', description='putative PRAME family member 26', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101930171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930171, CurrentID=0, Status='live', Symbol='LOC101930171', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47021726, end_position_on_the_genomic_accession=47024244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  349372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349372, CurrentID=0, Status='live', Symbol='NANOGP10', Aliases='NANOGP5', description='Nanog homeobox pseudogene 10', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43407042, end_position_on_the_genomic_accession=43408746, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379492, CurrentID=0, Status='live', Symbol='LOC105379492', Aliases='', description='uncharacterized LOC105379492', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6311244, end_position_on_the_genomic_accession=6314001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131702, CurrentID=0, Status='live', Symbol='LOC100131702', Aliases='', description='uncharacterized LOC100131702', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=22577062, end_position_on_the_genomic_accession=22577454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371636, CurrentID=0, Status='live', Symbol='LOC105371636', Aliases='', description='uncharacterized LOC105371636', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180152608, end_position_on_the_genomic_accession=180154439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  349386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349386, CurrentID=0, Status='live', Symbol='NANOGP9', Aliases='NANOGP6', description='Nanog homeobox pseudogene 9', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65771763, end_position_on_the_genomic_accession=65773832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  349391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349391, CurrentID=0, Status='live', Symbol='CYCSP44', Aliases='HCP44', description='cytochrome c, somatic pseudogene 44', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145176582, end_position_on_the_genomic_accession=145176901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874446, CurrentID=0, Status='live', Symbol='HMGN1P33', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 33', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 33', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9865093, end_position_on_the_genomic_accession=9865543, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060391, CurrentID=0, Status='live', Symbol='LOC101060391', Aliases='', description='zinc finger CCCH domain-containing protein 18-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=949634, end_position_on_the_genomic_accession=954998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  349408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349408, CurrentID=0, Status='live', Symbol='TLR8-AS1', Aliases='GS1-324M7.6', description='TLR8 antisense RNA 1', other_designations='TLR8 antisense RNA 1 (non-protein coding)', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12902817, end_position_on_the_genomic_accession=12943300, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374588, CurrentID=0, Status='live', Symbol='LOC105374588', Aliases='', description='uncharacterized LOC105374588', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47218130, end_position_on_the_genomic_accession=47221349, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374589, CurrentID=0, Status='live', Symbol='LOC105374589', Aliases='', description='uncharacterized LOC105374589', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47755588, end_position_on_the_genomic_accession=47760444, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106481963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481963, CurrentID=0, Status='live', Symbol='DPPA3P3', Aliases='', description='developmental pluripotency associated 3 pseudogene 3', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131711, CurrentID=0, Status='live', Symbol='RFC3P1', Aliases='', description='replication factor C 3 pseudogene 1', other_designations='replication factor C (activator 1) 3, 38kDa pseudogene 1', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36169761, end_position_on_the_genomic_accession=36170448, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874454, CurrentID=0, Status='live', Symbol='HIGD1AP13', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 13', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 13', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78648984, end_position_on_the_genomic_accession=78649251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374593, CurrentID=0, Status='live', Symbol='LOC105374593', Aliases='', description='uncharacterized LOC105374593', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48108948, end_position_on_the_genomic_accession=48314501, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  349448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349448, CurrentID=0, Status='live', Symbol='TUBB1P1', Aliases='', description='tubulin beta 1 class VI pseudogene 1', other_designations='HSAYq11 beta-tubulin 4Q (TUBB4Q) pseudogene|tubulin, beta 1 class VI pseudogene 1|tubulin, beta 1 pseudogene 1|tubulin, beta polypeptide 4, member Q pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18386581, end_position_on_the_genomic_accession=18389416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930249, CurrentID=0, Status='live', Symbol='LOC101930249', Aliases='', description='uncharacterized LOC101930249', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42363414, end_position_on_the_genomic_accession=42365104, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930275, CurrentID=0, Status='live', Symbol='LOC101930275', Aliases='', description='uncharacterized LOC101930275', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17801463, end_position_on_the_genomic_accession=17822430, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930276, CurrentID=0, Status='live', Symbol='LOC101930276', Aliases='', description='uncharacterized LOC101930276', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103880661, end_position_on_the_genomic_accession=103891392, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128365, CurrentID=0, Status='live', Symbol='LOC100128365', Aliases='', description='interferon stimulated exonuclease gene 20kDa like 2 pseudogene', other_designations='interferon stimulated exonuclease gene 20kDa-like 2 pseudogene', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23971766, end_position_on_the_genomic_accession=23973351, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131719, CurrentID=0, Status='live', Symbol='LOC100131719', Aliases='', description='uncharacterized LOC100131719', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=124939910, end_position_on_the_genomic_accession=124940323, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100874462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874462, CurrentID=0, Status='live', Symbol='HMGN1P24', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 24', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 24', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95809310, end_position_on_the_genomic_accession=95809607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930294, CurrentID=0, Status='live', Symbol='LOC101930294', Aliases='', description='uncharacterized LOC101930294', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90450753, end_position_on_the_genomic_accession=90455219, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375357, CurrentID=0, Status='live', Symbol='LOC105375357', Aliases='', description='uncharacterized LOC105375357', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76552955, end_position_on_the_genomic_accession=76578809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874463, CurrentID=0, Status='live', Symbol='HMGN1P27', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 27', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 27', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77204923, end_position_on_the_genomic_accession=77205415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374603, CurrentID=0, Status='live', Symbol='LOC105374603', Aliases='', description='uncharacterized LOC105374603', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53818165, end_position_on_the_genomic_accession=53838667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874465, CurrentID=0, Status='live', Symbol='HMGN1P32', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 32', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 32', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53767971, end_position_on_the_genomic_accession=53768553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930319, CurrentID=0, Status='live', Symbol='LOC101930319', Aliases='', description='histone H1.4-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874467, CurrentID=0, Status='live', Symbol='HMGN2P21', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 21', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 21', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70803157, end_position_on_the_genomic_accession=70803426, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374606, CurrentID=0, Status='live', Symbol='LOC105374606', Aliases='', description='uncharacterized LOC105374606', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=555061, end_position_on_the_genomic_accession=555868, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374607, CurrentID=0, Status='live', Symbol='LOC105374607', Aliases='', description='uncharacterized LOC105374607', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=571044, end_position_on_the_genomic_accession=577073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374609, CurrentID=0, Status='live', Symbol='LOC105374609', Aliases='', description='uncharacterized LOC105374609', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=900501, end_position_on_the_genomic_accession=912593, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101930347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930347, CurrentID=0, Status='live', Symbol='LOC101930347', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=93760919, end_position_on_the_genomic_accession=93763920, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374610, CurrentID=0, Status='live', Symbol='LOC105374610', Aliases='', description='uncharacterized LOC105374610', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54269193, end_position_on_the_genomic_accession=54278557, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  349565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349565, CurrentID=0, Status='live', Symbol='NMNAT3', Aliases='FKSG76, PNAT-3, PNAT3', description='nicotinamide nucleotide adenylyltransferase 3', other_designations='NMN adenylyltransferase 3|NMN/NaMN adenylyltransferase 3|NaMN adenylyltransferase 3|nicotinamide mononucleotide adenylyltransferase 3|nicotinamide-nucleotide adenylyltransferase 3|nicotinate-nucleotide adenylyltransferase 3|pyridine nucleotide adenylyltransferase 3', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139560180, end_position_on_the_genomic_accession=139678043, orientation='minus', exon_count=11, OMIM=608702, no_hdr0=''),
  105373383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373383, CurrentID=0, Status='live', Symbol='LOC105373383', Aliases='', description='uncharacterized LOC105373383', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153599347, end_position_on_the_genomic_accession=153599897, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374613, CurrentID=0, Status='live', Symbol='LOC105374613', Aliases='', description='uncharacterized LOC105374613', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1252008, end_position_on_the_genomic_accession=1255382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929287, CurrentID=0, Status='live', Symbol='LOC101929287', Aliases='', description='uncharacterized LOC101929287', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46756445, end_position_on_the_genomic_accession=46759740, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930370, CurrentID=0, Status='live', Symbol='LOC101930370', Aliases='', description='uncharacterized LOC101930370', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173094672, end_position_on_the_genomic_accession=173168961, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100874476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874476, CurrentID=0, Status='live', Symbol='HIGD1AP16', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 16', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 16', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35878434, end_position_on_the_genomic_accession=35878730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131736, CurrentID=0, Status='live', Symbol='LOC100131736', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161048609, end_position_on_the_genomic_accession=161050546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874478, CurrentID=0, Status='live', Symbol='HMGN2P34', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 34', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 34', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72636497, end_position_on_the_genomic_accession=72636765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874479, CurrentID=0, Status='live', Symbol='HMGN2P35', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 35', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 35', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96915633, end_position_on_the_genomic_accession=96915902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374619, CurrentID=0, Status='live', Symbol='LOC105374619', Aliases='', description='uncharacterized LOC105374619', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2840562, end_position_on_the_genomic_accession=2850913, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374620, CurrentID=0, Status='live', Symbol='LOC105374620', Aliases='', description='uncharacterized LOC105374620', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2917838, end_position_on_the_genomic_accession=2965527, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100874482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874482, CurrentID=0, Status='live', Symbol='HMGN2P38', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 38', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 38', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73580146, end_position_on_the_genomic_accession=73580556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419685, CurrentID=0, Status='live', Symbol='LOC100419685', Aliases='', description='zinc finger protein 555 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204065691, end_position_on_the_genomic_accession=204067275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874483, CurrentID=0, Status='live', Symbol='HMGN2P39', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 39', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 39', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59010218, end_position_on_the_genomic_accession=59010635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930420, CurrentID=0, Status='live', Symbol='LOC101930420', Aliases='', description='DNA primase large subunit-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=91373565, end_position_on_the_genomic_accession=91516097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930421, CurrentID=0, Status='live', Symbol='LOC101930421', Aliases='', description='uncharacterized LOC101930421', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=624991, end_position_on_the_genomic_accession=631255, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131743, CurrentID=0, Status='live', Symbol='GAPDHP40', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 40', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159950291, end_position_on_the_genomic_accession=159951569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874485, CurrentID=0, Status='live', Symbol='HMGN2P42', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 42', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 42', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60435133, end_position_on_the_genomic_accession=60435536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  349633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349633, CurrentID=0, Status='live', Symbol='PLET1', Aliases='C11orf34', description='placenta expressed transcript 1', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112248153, end_position_on_the_genomic_accession=112260860, orientation='minus', exon_count=4, OMIM=611904, no_hdr0=''),
  105371162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371162, CurrentID=0, Status='live', Symbol='LOC105371162', Aliases='', description='uncharacterized LOC105371162', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29130314, end_position_on_the_genomic_accession=29132441, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874487, CurrentID=0, Status='live', Symbol='HMGN2P44', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 44', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 44', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32032445, end_position_on_the_genomic_accession=32034545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644257, CurrentID=0, Status='live', Symbol='LOC644257', Aliases='', description='uncharacterized LOC644257', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375527, CurrentID=0, Status='live', Symbol='LOC105375527', Aliases='', description='uncharacterized LOC105375527', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137324589, end_position_on_the_genomic_accession=137326027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104326051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326051, CurrentID=0, Status='live', Symbol='MAFA-AS1', Aliases='TCONS_00014882', description='MAFA antisense RNA 1', other_designations='RP11-909N17.3', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143417679, end_position_on_the_genomic_accession=143419150, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930452, CurrentID=0, Status='live', Symbol='LOC101930452', Aliases='', description='uncharacterized LOC101930452', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9367464, end_position_on_the_genomic_accession=9397617, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374628, CurrentID=0, Status='live', Symbol='LOC105374628', Aliases='', description='uncharacterized LOC105374628', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=4486706, end_position_on_the_genomic_accession=4487963, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874490, CurrentID=0, Status='live', Symbol='EEF1A1P40', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 40', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107032248, end_position_on_the_genomic_accession=107032820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374629, CurrentID=0, Status='live', Symbol='LOC105374629', Aliases='', description='uncharacterized LOC105374629', other_designations='CTD-2046I8.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=4866490, end_position_on_the_genomic_accession=4868215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  349667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349667, CurrentID=0, Status='live', Symbol='RTN4RL2', Aliases='NGRH1, NgR2', description='reticulon 4 receptor-like 2', other_designations='Nogo-66 receptor homolog 1|nogo receptor-like 3|nogo-66 receptor-related protein 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57460537, end_position_on_the_genomic_accession=57477539, orientation='plus', exon_count=3, OMIM=610462, no_hdr0=''),
  105374630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374630, CurrentID=0, Status='live', Symbol='LOC105374630', Aliases='', description='uncharacterized LOC105374630', other_designations='CTD-2247C11.1|CTD-2247C11.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=4967772, end_position_on_the_genomic_accession=4975243, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  260421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260421, CurrentID=0, Status='live', Symbol='LOC260421', Aliases='', description='actin related protein 2/3 complex subunit 1A pseudogene', other_designations='actin related protein 2/3 complex, subunit 1A, 41kDa pseudogene', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74798894, end_position_on_the_genomic_accession=74800406, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874492, CurrentID=0, Status='live', Symbol='BRK1P2', Aliases='', description='BRICK1, SCAR/WAVE actin-nucleating complex subunit pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91023743, end_position_on_the_genomic_accession=91023951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  254887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254887, CurrentID=0, Status='live', Symbol='ZDHHC23', Aliases='NIDD', description='zinc finger, DHHC-type containing 23', other_designations='DHHC-23|probable palmitoyltransferase ZDHHC23|zinc finger DHHC domain-containing protein 23|zinc finger, DHHC domain containing 23', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113947901, end_position_on_the_genomic_accession=113992167, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100874493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874493, CurrentID=0, Status='live', Symbol='PIEZO1P2', Aliases='', description='piezo-type mechanosensitive ion channel component 1 pseudogene 2', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58741189, end_position_on_the_genomic_accession=58766333, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374632, CurrentID=0, Status='live', Symbol='LOC105374632', Aliases='', description='uncharacterized LOC105374632', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54581659, end_position_on_the_genomic_accession=54584999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927181, CurrentID=0, Status='live', Symbol='LOC101927181', Aliases='', description='uncharacterized LOC101927181', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2437763, end_position_on_the_genomic_accession=2447850, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129375, CurrentID=0, Status='live', Symbol='LOC100129375', Aliases='', description='ring finger protein 14 pseudogene', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138412069, end_position_on_the_genomic_accession=138413318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874495, CurrentID=0, Status='live', Symbol='CYP4A43P', Aliases='', description='cytochrome P450 family 4 subfamily A member 43, pseudogene', other_designations='CYP4A-se2[1]|cytochrome P450, family 4, subfamily A, polypeptide 43, pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374634, CurrentID=0, Status='live', Symbol='LOC105374634', Aliases='', description='uncharacterized LOC105374634', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5688424, end_position_on_the_genomic_accession=5699846, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421798, CurrentID=0, Status='live', Symbol='LOC100421798', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132942929, end_position_on_the_genomic_accession=132944076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930496, CurrentID=0, Status='live', Symbol='LOC101930496', Aliases='', description='uncharacterized LOC101930496', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83110233, end_position_on_the_genomic_accession=83119423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374635, CurrentID=0, Status='live', Symbol='LOC105374635', Aliases='', description='uncharacterized LOC105374635', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5781120, end_position_on_the_genomic_accession=5789183, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100874497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874497, CurrentID=0, Status='live', Symbol='RBMY1KP', Aliases='RBMY1K', description='RNA binding motif protein, Y-linked, family 1, member K, pseudogene', other_designations='RBMY1A1 pseudogene', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22209331, end_position_on_the_genomic_accession=22216580, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374636, CurrentID=0, Status='live', Symbol='LOC105374636', Aliases='', description='uncharacterized LOC105374636', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5789465, end_position_on_the_genomic_accession=5796226, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369247, CurrentID=0, Status='live', Symbol='LOC105369247', Aliases='', description='nicotinate-nucleotide pyrophosphorylase [carboxylating]-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928927, CurrentID=0, Status='live', Symbol='LOC101928927', Aliases='', description='uncharacterized LOC101928927', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44965092, end_position_on_the_genomic_accession=44985875, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874499, CurrentID=0, Status='live', Symbol='SPTLC1P3', Aliases='', description='serine palmitoyltransferase, long chain base subunit 1 pseudogene 3', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63227483, end_position_on_the_genomic_accession=63227743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930517, CurrentID=0, Status='live', Symbol='LOC101930517', Aliases='', description='uncharacterized LOC101930517', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374639, CurrentID=0, Status='live', Symbol='LOC105374639', Aliases='', description='uncharacterized LOC105374639', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6454465, end_position_on_the_genomic_accession=6459317, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371165, CurrentID=0, Status='live', Symbol='LOC105371165', Aliases='', description='uncharacterized LOC105371165', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28492368, end_position_on_the_genomic_accession=28493602, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421799, CurrentID=0, Status='live', Symbol='LONRF2P2', Aliases='', description='LON peptidase N-terminal domain and ring finger 2 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18609046, end_position_on_the_genomic_accession=18609705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874501, CurrentID=0, Status='live', Symbol='NME1P1', Aliases='', description='NME/NM23 nucleoside diphosphate kinase 1 pseudogene 1', other_designations='non-metastatic cells 1, protein (NM23A) expressed in, pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22099163, end_position_on_the_genomic_accession=22099553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  254896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254896, CurrentID=0, Status='live', Symbol='LOC254896', Aliases='', description='uncharacterized LOC254896', other_designations='', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23084355, end_position_on_the_genomic_accession=23103558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930531, CurrentID=0, Status='live', Symbol='LOC101930531', Aliases='', description='uncharacterized LOC101930531', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=12602292, end_position_on_the_genomic_accession=12629455, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374641, CurrentID=0, Status='live', Symbol='LOC105374641', Aliases='', description='uncharacterized LOC105374641', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6829727, end_position_on_the_genomic_accession=6833157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874503, CurrentID=0, Status='live', Symbol='ATP6V0E1P2', Aliases='', description='ATPase, H+ transporting, lysosomal 9kDa, V0 subunit e1 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41859353, end_position_on_the_genomic_accession=41859592, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391359, CurrentID=0, Status='live', Symbol='LOC391359', Aliases='', description='epithelial membrane protein 2 pseudogene', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26140255, end_position_on_the_genomic_accession=26141264, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374643, CurrentID=0, Status='live', Symbol='LOC105374643', Aliases='', description='uncharacterized LOC105374643', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7272809, end_position_on_the_genomic_accession=7277788, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101930552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930552, CurrentID=0, Status='live', Symbol='LOC101930552', Aliases='', description='zinc finger protein 717-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374644, CurrentID=0, Status='live', Symbol='LOC105374644', Aliases='', description='uncharacterized LOC105374644', other_designations='CTD-2296D1.5', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7373399, end_position_on_the_genomic_accession=7391983, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100874507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874507, CurrentID=0, Status='live', Symbol='CHCHD2P4', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134393142, end_position_on_the_genomic_accession=134393509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374646, CurrentID=0, Status='live', Symbol='LOC105374646', Aliases='', description='uncharacterized LOC105374646', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7417774, end_position_on_the_genomic_accession=7419472, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  132391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=132391, CurrentID=0, Status='live', Symbol='KRT18P21', Aliases='', description='keratin 18 pseudogene 21', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=115920657, end_position_on_the_genomic_accession=115922033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131768, CurrentID=0, Status='live', Symbol='LOC100131768', Aliases='', description='uncharacterized LOC100131768', other_designations='', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=130495722, end_position_on_the_genomic_accession=130505665, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100874510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874510, CurrentID=0, Status='live', Symbol='PSMA6P3', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 6 pseudogene 3', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46072085, end_position_on_the_genomic_accession=46072248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874511, CurrentID=0, Status='live', Symbol='SPTLC1P1', Aliases='', description='serine palmitoyltransferase, long chain base subunit 1 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31360955, end_position_on_the_genomic_accession=31361215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930589, CurrentID=0, Status='live', Symbol='LOC101930589', Aliases='', description='cell division cycle protein 27 homolog pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377749, CurrentID=0, Status='live', Symbol='LOC105377749', Aliases='', description='uncharacterized LOC105377749', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177085847, end_position_on_the_genomic_accession=177086638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100874512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874512, CurrentID=0, Status='live', Symbol='SPTLC1P5', Aliases='', description='serine palmitoyltransferase, long chain base subunit 1 pseudogene 5', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77659413, end_position_on_the_genomic_accession=77659673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874513, CurrentID=0, Status='live', Symbol='CYP2C115P', Aliases='CYP2C9-de1b', description='cytochrome P450 family 2 subfamily C member 115, pseudogene', other_designations='cytochrome P450, family 2, subfamily C, polypeptide 115, pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94905898, end_position_on_the_genomic_accession=94906133, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874514, CurrentID=0, Status='live', Symbol='ATP6V1G1P7', Aliases='', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=103444020, end_position_on_the_genomic_accession=103445568, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374654, CurrentID=0, Status='live', Symbol='LOC105374654', Aliases='', description='uncharacterized LOC105374654', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10964705, end_position_on_the_genomic_accession=10987571, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  349814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349814, CurrentID=0, Status='live', Symbol='VENTXP3', Aliases='VENTX2P3', description='VENT homeobox pseudogene 3', other_designations='VENT homeobox homolog pseudogene|VENT-like homeobox 2 pseudogene 3', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74292313, end_position_on_the_genomic_accession=74294332, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371168, CurrentID=0, Status='live', Symbol='LOC105371168', Aliases='', description='uncharacterized LOC105371168', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29597791, end_position_on_the_genomic_accession=29603655, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100874516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874516, CurrentID=0, Status='live', Symbol='LAMTOR3P2', Aliases='', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 3 pseudogene 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119132685, end_position_on_the_genomic_accession=119132902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374655, CurrentID=0, Status='live', Symbol='LOC105374655', Aliases='', description='uncharacterized LOC105374655', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12362202, end_position_on_the_genomic_accession=12574539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874517, CurrentID=0, Status='live', Symbol='TOMM22P4', Aliases='', description='TOMM22 pseudogene 4', other_designations='translocase of outer mitochondrial membrane 22 homolog pseudogene 4', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=162521702, end_position_on_the_genomic_accession=162522041, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374007, CurrentID=0, Status='live', Symbol='LOC105374007', Aliases='', description='uncharacterized LOC105374007', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99548907, end_position_on_the_genomic_accession=99597972, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100874518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874518, CurrentID=0, Status='live', Symbol='ATP8A2P3', Aliases='', description='ATPase, aminophospholipid transporter, class I, type 8A, member 2 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26114409, end_position_on_the_genomic_accession=26154175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374657, CurrentID=0, Status='live', Symbol='LOC105374657', Aliases='', description='uncharacterized LOC105374657', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12739663, end_position_on_the_genomic_accession=12862122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  87688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=87688, CurrentID=0, Status='live', Symbol='RPL7AP50', Aliases='RPL7A_23_1048', description='ribosomal protein L7a pseudogene 50', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63902385, end_position_on_the_genomic_accession=63903265, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874520, CurrentID=0, Status='live', Symbol='PCMTD1P1', Aliases='', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 1 pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10173844, end_position_on_the_genomic_accession=10174204, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  349842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=349842, CurrentID=0, Status='live', Symbol='MRPS18CP4', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene 4', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14237261, end_position_on_the_genomic_accession=14237324, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131779, CurrentID=0, Status='live', Symbol='LOC100131779', Aliases='', description='translocase of outer mitochondrial membrane 6 homolog (yeast) pseudogene', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125056360, end_position_on_the_genomic_accession=125056581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100874521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874521, CurrentID=0, Status='live', Symbol='ST13P22', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 22', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101773114, end_position_on_the_genomic_accession=101773421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372512, CurrentID=0, Status='live', Symbol='LOC105372512', Aliases='', description='uncharacterized LOC105372512', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5059229, end_position_on_the_genomic_accession=5069347, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100874522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874522, CurrentID=0, Status='live', Symbol='SLC25A15P2', Aliases='', description='solute carrier family 25 (mitochondrial carrier; ornithine transporter) member 15 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19415268, end_position_on_the_genomic_accession=19415432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874523, CurrentID=0, Status='live', Symbol='HIF1AP1', Aliases='', description='hypoxia inducible factor 1, alpha subunit (basic helix-loop-helix transcription factor) pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75116426, end_position_on_the_genomic_accession=75116608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930665, CurrentID=0, Status='live', Symbol='LOC101930665', Aliases='', description='uncharacterized LOC101930665', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21612849, end_position_on_the_genomic_accession=21642127, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374664, CurrentID=0, Status='live', Symbol='LOC105374664', Aliases='', description='uncharacterized LOC105374664', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17164723, end_position_on_the_genomic_accession=17185213, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131785, CurrentID=0, Status='live', Symbol='LOC100131785', Aliases='', description='ring finger protein 181 pseudogene', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105530189, end_position_on_the_genomic_accession=105530745, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374258, CurrentID=0, Status='live', Symbol='LOC105374258', Aliases='', description='uncharacterized LOC105374258', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186641153, end_position_on_the_genomic_accession=186715028, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374667, CurrentID=0, Status='live', Symbol='LOC105374667', Aliases='', description='uncharacterized LOC105374667', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54747103, end_position_on_the_genomic_accession=54750515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874529, CurrentID=0, Status='live', Symbol='MPRIP-AS1', Aliases='', description='MPRIP antisense RNA 1', other_designations='MPRIP antisense RNA 1 (non-protein coding)', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374668, CurrentID=0, Status='live', Symbol='LOC105374668', Aliases='', description='uncharacterized LOC105374668', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17810771, end_position_on_the_genomic_accession=17820745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100874530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874530, CurrentID=0, Status='live', Symbol='APOBEC3B-AS1', Aliases='', description='APOBEC3B antisense RNA 1', other_designations='APOBEC3B antisense RNA 1 (non-protein coding)|CTA-150C2.16', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38991559, end_position_on_the_genomic_accession=38998220, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374669, CurrentID=0, Status='live', Symbol='LOC105374669', Aliases='', description='uncharacterized LOC105374669', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=18972408, end_position_on_the_genomic_accession=18990438, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  87769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=87769, CurrentID=0, Status='live', Symbol='GGACT', Aliases='A2LD1', description='gamma-glutamylamine cyclotransferase', other_designations='AIG2-like domain 1|AIG2-like domain-containing protein 1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100530164, end_position_on_the_genomic_accession=100588792, orientation='minus', exon_count=5, OMIM=613378, no_hdr0=''),
  100419687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419687, CurrentID=0, Status='live', Symbol='RNF152P1', Aliases='', description='ring finger protein 152 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5418371, end_position_on_the_genomic_accession=5418976, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374672, CurrentID=0, Status='live', Symbol='LOC105374672', Aliases='', description='uncharacterized LOC105374672', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=19771367, end_position_on_the_genomic_accession=19778750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374675, CurrentID=0, Status='live', Symbol='LOC105374675', Aliases='', description='uncharacterized LOC105374675', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20751700, end_position_on_the_genomic_accession=20752543, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101930744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930744, CurrentID=0, Status='live', Symbol='SERPINA15P', Aliases='', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 15, pseudogene', other_designations='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin) pseudogene|serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 6 pseudogene', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111380603, end_position_on_the_genomic_accession=111380795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930745, CurrentID=0, Status='live', Symbol='SERPINE4P', Aliases='', description='serpin peptidase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 4 pseudogene', other_designations='', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27340526, end_position_on_the_genomic_accession=27341242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930746, CurrentID=0, Status='live', Symbol='LINC00945', Aliases='', description='long intergenic non-protein coding RNA 945', other_designations='TCONS_00028834', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33057829, end_position_on_the_genomic_accession=33065079, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101930747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930747, CurrentID=0, Status='live', Symbol='LINC00380', Aliases='', description='long intergenic non-protein coding RNA 380', other_designations='TCONS_00021685', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91087254, end_position_on_the_genomic_accession=91091439, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101930748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930748, CurrentID=0, Status='live', Symbol='LINC00367', Aliases='', description='long intergenic non-protein coding RNA 367', other_designations='TCONS_l2_00007488', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20938564, end_position_on_the_genomic_accession=20949428, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101930749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930749, CurrentID=0, Status='live', Symbol='LINC00397', Aliases='', description='long intergenic non-protein coding RNA 397', other_designations='TCONS_00021582', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87801043, end_position_on_the_genomic_accession=87810520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374677, CurrentID=0, Status='live', Symbol='LOC105374677', Aliases='', description='uncharacterized LOC105374677', other_designations='CTD-2333M24.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20607455, end_position_on_the_genomic_accession=20616320, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  123862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123862, CurrentID=0, Status='live', Symbol='LOC123862', Aliases='', description='interferon induced transmembrane protein pseudogene', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84920430, end_position_on_the_genomic_accession=84921038, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129384, CurrentID=0, Status='live', Symbol='LOC100129384', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48900278, end_position_on_the_genomic_accession=48902181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374680, CurrentID=0, Status='live', Symbol='LOC105374680', Aliases='', description='uncharacterized LOC105374680', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21275672, end_position_on_the_genomic_accession=21282672, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374681, CurrentID=0, Status='live', Symbol='LOC105374681', Aliases='', description='uncharacterized LOC105374681', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=22574914, end_position_on_the_genomic_accession=22589298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131802, CurrentID=0, Status='live', Symbol='LOC100131802', Aliases='', description='uncharacterized LOC100131802', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=36340583, end_position_on_the_genomic_accession=36340954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927183, CurrentID=0, Status='live', Symbol='LOC101927183', Aliases='', description='uncharacterized LOC101927183', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38424275, end_position_on_the_genomic_accession=38432007, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374685, CurrentID=0, Status='live', Symbol='LOC105374685', Aliases='', description='uncharacterized LOC105374685', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21616247, end_position_on_the_genomic_accession=21676412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375539, CurrentID=0, Status='live', Symbol='LOC105375539', Aliases='', description='uncharacterized LOC105375539', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142321556, end_position_on_the_genomic_accession=142336648, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  342293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342293, CurrentID=0, Status='live', Symbol='LOC342293', Aliases='', description='ATP binding cassette subfamily A member 3 pseudogene', other_designations='ATP-binding cassette, sub-family A (ABC1), member 3 pseudogene', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21938865, end_position_on_the_genomic_accession=21940466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374688, CurrentID=0, Status='live', Symbol='LOC105374688', Aliases='', description='uncharacterized LOC105374688', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23344658, end_position_on_the_genomic_accession=23349272, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374689, CurrentID=0, Status='live', Symbol='LOC105374689', Aliases='', description='uncharacterized LOC105374689', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24291916, end_position_on_the_genomic_accession=24323113, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421809, CurrentID=0, Status='live', Symbol='LOC100421809', Aliases='', description='LON peptidase N-terminal domain and ring finger 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20775462, end_position_on_the_genomic_accession=20776116, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374690, CurrentID=0, Status='live', Symbol='LOC105374690', Aliases='', description='uncharacterized LOC105374690', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55945824, end_position_on_the_genomic_accession=56177552, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374259, CurrentID=0, Status='live', Symbol='LOC105374259', Aliases='', description='uncharacterized LOC105374259', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186753227, end_position_on_the_genomic_accession=186761048, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254948, CurrentID=0, Status='live', Symbol='RPL9P8', Aliases='RPL9P9, RPL9_14_1458, RPL9_15_1459', description='ribosomal protein L9 pseudogene 8', other_designations='ribosomal protein L9 pseudogene 9', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82372211, end_position_on_the_genomic_accession=82372914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131814, CurrentID=0, Status='live', Symbol='LINC00271', Aliases='C6orf217, NCRNA00271', description='long intergenic non-protein coding RNA 271', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135497801, end_position_on_the_genomic_accession=135690838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374695, CurrentID=0, Status='live', Symbol='LOC105374695', Aliases='', description='uncharacterized LOC105374695', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=27406530, end_position_on_the_genomic_accession=27413642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374696, CurrentID=0, Status='live', Symbol='LOC105374696', Aliases='', description='uncharacterized LOC105374696', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=27772331, end_position_on_the_genomic_accession=27775475, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372785, CurrentID=0, Status='live', Symbol='LOC105372785', Aliases='', description='uncharacterized LOC105372785', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33035539, end_position_on_the_genomic_accession=33037078, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  442272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442272, CurrentID=0, Status='live', Symbol='LOC442272', Aliases='', description='YKT6 v-SNARE homolog (S. cerevisiae) pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158921591, end_position_on_the_genomic_accession=158922150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372459, CurrentID=0, Status='live', Symbol='LOC105372459', Aliases='', description='uncharacterized LOC105372459', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54210790, end_position_on_the_genomic_accession=54216055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374699, CurrentID=0, Status='live', Symbol='LOC105374699', Aliases='', description='uncharacterized LOC105374699', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29077276, end_position_on_the_genomic_accession=29143424, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374700, CurrentID=0, Status='live', Symbol='LOC105374700', Aliases='', description='uncharacterized LOC105374700', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=28809035, end_position_on_the_genomic_accession=28898807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131821, CurrentID=0, Status='live', Symbol='LOC100131821', Aliases='', description='uncharacterized LOC100131821', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=35759010, end_position_on_the_genomic_accession=35759408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374702, CurrentID=0, Status='live', Symbol='LOC105374702', Aliases='', description='uncharacterized LOC105374702', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29472719, end_position_on_the_genomic_accession=29485424, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374703, CurrentID=0, Status='live', Symbol='LOC105374703', Aliases='', description='uncharacterized LOC105374703', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29793721, end_position_on_the_genomic_accession=29814358, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374705, CurrentID=0, Status='live', Symbol='LOC105374705', Aliases='', description='uncharacterized LOC105374705', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29931556, end_position_on_the_genomic_accession=29964943, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  552859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=552859, CurrentID=0, Status='live', Symbol='LINC00251', Aliases='C8orf25, NCRNA00251', description='long intergenic non-protein coding RNA 251', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65161145, end_position_on_the_genomic_accession=65180340, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374706, CurrentID=0, Status='live', Symbol='LOC105374706', Aliases='', description='uncharacterized LOC105374706', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29989131, end_position_on_the_genomic_accession=30002215, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374707, CurrentID=0, Status='live', Symbol='LOC105374707', Aliases='', description='uncharacterized LOC105374707', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=30099013, end_position_on_the_genomic_accession=30102387, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374708, CurrentID=0, Status='live', Symbol='LOC105374708', Aliases='', description='uncharacterized LOC105374708', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31164034, end_position_on_the_genomic_accession=31175969, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131829, CurrentID=0, Status='live', Symbol='LOC100131829', Aliases='', description='uncharacterized LOC100131829', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=102446155, end_position_on_the_genomic_accession=102447501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371179, CurrentID=0, Status='live', Symbol='LOC105371179', Aliases='', description='proline-rich receptor-like protein kinase PERK2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32636694, end_position_on_the_genomic_accession=32650869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131831, CurrentID=0, Status='live', Symbol='LOC100131831', Aliases='', description='uncharacterized LOC100131831', other_designations='', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=13950526, end_position_on_the_genomic_accession=13951343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131832, CurrentID=0, Status='live', Symbol='LOC100131832', Aliases='', description='ZFP91 zinc finger protein pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127840506, end_position_on_the_genomic_accession=127841865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373387, CurrentID=0, Status='live', Symbol='LOC105373387', Aliases='', description='uncharacterized LOC105373387', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154653688, end_position_on_the_genomic_accession=154655374, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106481968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481968, CurrentID=0, Status='live', Symbol='B3GNTL1P1', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like 1 pseudogene 1', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631311, CurrentID=0, Status='live', Symbol='LOC100631311', Aliases='', description='Obg-like ATPase 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374717, CurrentID=0, Status='live', Symbol='LOC105374717', Aliases='', description='uncharacterized LOC105374717', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33516019, end_position_on_the_genomic_accession=33522505, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  80213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80213, CurrentID=0, Status='live', Symbol='TM2D3', Aliases='BLP2', description='TM2 domain containing 3', other_designations='BBP-like protein 2|beta-amyloid-binding protein-like protein 2', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101633226, end_position_on_the_genomic_accession=101652391, orientation='minus', exon_count=7, OMIM=610014, no_hdr0=''),
  100131840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131840, CurrentID=0, Status='live', Symbol='LOC100131840', Aliases='', description='uncharacterized LOC100131840', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=49211221, end_position_on_the_genomic_accession=49212300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927575, CurrentID=0, Status='live', Symbol='LOC101927575', Aliases='', description='uncharacterized LOC101927575', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84063443, end_position_on_the_genomic_accession=84094601, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374722, CurrentID=0, Status='live', Symbol='LOC105374722', Aliases='', description='uncharacterized LOC105374722', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34798347, end_position_on_the_genomic_accession=34814576, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374723, CurrentID=0, Status='live', Symbol='LOC105374723', Aliases='', description='uncharacterized LOC105374723', other_designations='CTD-2024P10.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34651468, end_position_on_the_genomic_accession=34656253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129393, CurrentID=0, Status='live', Symbol='LOC100129393', Aliases='', description='uncharacterized LOC100129393', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=38311672, end_position_on_the_genomic_accession=38312097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421816, CurrentID=0, Status='live', Symbol='ENPP7P4', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125848223, end_position_on_the_genomic_accession=125909414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928817, CurrentID=0, Status='live', Symbol='LOC101928817', Aliases='', description='uncharacterized LOC101928817', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89350089, end_position_on_the_genomic_accession=89365008, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374725, CurrentID=0, Status='live', Symbol='LOC105374725', Aliases='', description='uncharacterized LOC105374725', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36324527, end_position_on_the_genomic_accession=36338168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374726, CurrentID=0, Status='live', Symbol='LOC105374726', Aliases='', description='uncharacterized LOC105374726', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57916338, end_position_on_the_genomic_accession=58047296, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100131849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131849, CurrentID=0, Status='live', Symbol='LOC100131849', Aliases='', description='coiled-coil domain containing 53 pseudogene', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98413315, end_position_on_the_genomic_accession=98413865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421817, CurrentID=0, Status='live', Symbol='LOC100421817', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149494660, end_position_on_the_genomic_accession=149496003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401825, CurrentID=0, Status='live', Symbol='NIFKP1', Aliases='MKI67IPP1', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 1', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 1', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30380132, end_position_on_the_genomic_accession=30381121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131852, CurrentID=0, Status='live', Symbol='LOC100131852', Aliases='', description='coiled-coil domain containing 12 pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51926632, end_position_on_the_genomic_accession=51927348, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  342371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342371, CurrentID=0, Status='live', Symbol='ATXN1L', Aliases='BOAT, BOAT1', description='ataxin 1-like', other_designations='brother of ATXN1|brother of ataxin-1', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71845991, end_position_on_the_genomic_accession=71857333, orientation='plus', exon_count=3, OMIM=614301, no_hdr0=''),
  342372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342372, CurrentID=0, Status='live', Symbol='PKD1L3', Aliases='', description='polycystic kidney disease 1-like 3', other_designations='PC1-like 3 protein|polycystin-1L3', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71929538, end_position_on_the_genomic_accession=71999978, orientation='minus', exon_count=30, OMIM=607895, no_hdr0=''),
  100131855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131855, CurrentID=0, Status='live', Symbol='C2orf27AP2', Aliases='', description='chromosome 2 open reading frame 27A pseudogene 2', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30670987, end_position_on_the_genomic_accession=30671582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374736, CurrentID=0, Status='live', Symbol='LOC105374736', Aliases='', description='uncharacterized LOC105374736', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40307687, end_position_on_the_genomic_accession=40346555, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  342374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342374, CurrentID=0, Status='live', Symbol='KRT18P18', Aliases='', description='keratin 18 pseudogene 18', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72728603, end_position_on_the_genomic_accession=72729963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106479773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479773, CurrentID=0, Status='live', Symbol='RNU6-470P', Aliases='', description='RNA, U6 small nuclear 470, pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374739, CurrentID=0, Status='live', Symbol='LOC105374739', Aliases='', description='uncharacterized LOC105374739', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41071432, end_position_on_the_genomic_accession=41201549, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374740, CurrentID=0, Status='live', Symbol='LOC105374740', Aliases='', description='uncharacterized LOC105374740', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41385123, end_position_on_the_genomic_accession=41409162, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106780802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106780802, CurrentID=0, Status='live', Symbol='KCNK4-TEX40', Aliases='', description='KCNK4-TEX40 readthrough', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374743, CurrentID=0, Status='live', Symbol='LOC105374743', Aliases='', description='uncharacterized LOC105374743', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42889625, end_position_on_the_genomic_accession=42892218, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374744, CurrentID=0, Status='live', Symbol='LOC105374744', Aliases='', description='uncharacterized LOC105374744', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=58276112, end_position_on_the_genomic_accession=58296797, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  219285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219285, CurrentID=0, Status='live', Symbol='SAMD9L', Aliases='C7orf6, DRIF2, UEF1', description='sterile alpha motif domain containing 9-like', other_designations='SAM domain-containing protein 9-like', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93130054, end_position_on_the_genomic_accession=93148393, orientation='minus', exon_count=7, OMIM=611170, no_hdr0=''),
  219287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219287, CurrentID=0, Status='live', Symbol='AMER2', Aliases='FAM123A', description='APC membrane recruitment protein 2', other_designations='family with sequence similarity 123A|protein FAM123A', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25161679, end_position_on_the_genomic_accession=25172324, orientation='minus', exon_count=3, OMIM=614659, no_hdr0=''),
  106780825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106780825, CurrentID=0, Status='live', Symbol='LOC106780825', Aliases='', description='FOXM1-regulated, gastric cancer associated-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219293, CurrentID=0, Status='live', Symbol='ATAD3C', Aliases='', description='ATPase family, AAA domain containing 3C', other_designations='ATPase family, AAA domain containing 3A', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1441301, end_position_on_the_genomic_accession=1470158, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100420563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420563, CurrentID=0, Status='live', Symbol='LOC100420563', Aliases='', description='ADAM metallopeptidase domain 20 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70556871, end_position_on_the_genomic_accession=70557494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  350383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=350383, CurrentID=0, Status='live', Symbol='GPR142', Aliases='PGR2', description='G protein-coupled receptor 142', other_designations='G-protein coupled receptor PGR2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74367407, end_position_on_the_genomic_accession=74373008, orientation='plus', exon_count=4, OMIM=609046, no_hdr0=''),
  105371187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371187, CurrentID=0, Status='live', Symbol='LOC105371187', Aliases='', description='putative short-chain dehydrogenase/reductase family 42E member 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219333, CurrentID=0, Status='live', Symbol='USP12', Aliases='UBH1L1, USP12', description='ubiquitin specific peptidase 12', other_designations='deubiquitinating enzyme 12|ubiquitin specific protease 12 like 1|ubiquitin thioesterase 12|ubiquitin thiolesterase 12|ubiquitin-hydrolyzing enzyme 1|ubiquitin-specific-processing protease 12', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27066150, end_position_on_the_genomic_accession=27171896, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374754, CurrentID=0, Status='live', Symbol='LOC105374754', Aliases='', description='uncharacterized LOC105374754', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=59238394, end_position_on_the_genomic_accession=59388349, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100421822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421822, CurrentID=0, Status='live', Symbol='LOC100421822', Aliases='', description='protein tyrosine phosphatase, non-receptor type 11 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58424993, end_position_on_the_genomic_accession=58426885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  219347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219347, CurrentID=0, Status='live', Symbol='TMEM254-AS1', Aliases='', description='TMEM254 antisense RNA 1', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80046233, end_position_on_the_genomic_accession=80079193, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  219348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219348, CurrentID=0, Status='live', Symbol='PLAC9', Aliases='', description='placenta specific 9', other_designations='placenta-specific 9', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80132502, end_position_on_the_genomic_accession=80145028, orientation='plus', exon_count=4, OMIM=612857, no_hdr0=''),
  105375553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375553, CurrentID=0, Status='live', Symbol='LOC105375553', Aliases='', description='uncharacterized LOC105375553', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145675839, end_position_on_the_genomic_accession=145681460, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131877, CurrentID=0, Status='live', Symbol='LOC100131877', Aliases='', description='small EDRK-rich factor 2 pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114491501, end_position_on_the_genomic_accession=114492013, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131878, CurrentID=0, Status='live', Symbol='LOC100131878', Aliases='', description='uncharacterized LOC100131878', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=89809932, end_position_on_the_genomic_accession=89810432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374759, CurrentID=0, Status='live', Symbol='LOC105374759', Aliases='', description='uncharacterized LOC105374759', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61129405, end_position_on_the_genomic_accession=61140184, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060398, CurrentID=0, Status='live', Symbol='LOC101060398', Aliases='', description='uncharacterized LOC101060398', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149090901, end_position_on_the_genomic_accession=149104180, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374761, CurrentID=0, Status='live', Symbol='LOC105374761', Aliases='', description='uncharacterized LOC105374761', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62264243, end_position_on_the_genomic_accession=62275531, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644525, CurrentID=0, Status='live', Symbol='LOC644525', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15893130, end_position_on_the_genomic_accession=15893714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374053, CurrentID=0, Status='live', Symbol='LOC105374053', Aliases='', description='uncharacterized LOC105374053', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115870593, end_position_on_the_genomic_accession=115871344, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374763, CurrentID=0, Status='live', Symbol='LOC105374763', Aliases='', description='uncharacterized LOC105374763', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62540906, end_position_on_the_genomic_accession=62662834, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  101928932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928932, CurrentID=0, Status='live', Symbol='LOC101928932', Aliases='', description='uncharacterized LOC101928932', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138199933, end_position_on_the_genomic_accession=138203453, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131884, CurrentID=0, Status='live', Symbol='LOC100131884', Aliases='', description='capicua transcriptional repressor pseudogene', other_designations='capicua homolog pseudogene', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119403081, end_position_on_the_genomic_accession=119405582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  219402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219402, CurrentID=0, Status='live', Symbol='MTIF3', Aliases='IF3mt', description='mitochondrial translational initiation factor 3', other_designations='IF-3(Mt)', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27435639, end_position_on_the_genomic_accession=27450854, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105374765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374765, CurrentID=0, Status='live', Symbol='LOC105374765', Aliases='', description='uncharacterized LOC105374765', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63881420, end_position_on_the_genomic_accession=63884684, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  219409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219409, CurrentID=0, Status='live', Symbol='GSX1', Aliases='GSH1, Gsh-1', description='GS homeobox 1', other_designations='GS homeo box protein 1|genomic screened homeo box 1|homeobox protein Gsh-1', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27792643, end_position_on_the_genomic_accession=27793952, orientation='plus', exon_count=2, OMIM=616542, no_hdr0=''),
  104326190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326190, CurrentID=0, Status='live', Symbol='LINC01309', Aliases='TCONS_l2_00007459', description='long intergenic non-protein coding RNA 1309', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=103425200, end_position_on_the_genomic_accession=103427685, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369248, CurrentID=0, Status='live', Symbol='LOC105369248', Aliases='', description='nuclear pore complex-interacting protein family member B5-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219417, CurrentID=0, Status='live', Symbol='OR8U1', Aliases='', description='olfactory receptor family 8 subfamily U member 1', other_designations='olfactory receptor, family 8, subfamily U, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56375624, end_position_on_the_genomic_accession=56376553, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374767, CurrentID=0, Status='live', Symbol='LOC105374767', Aliases='', description='uncharacterized LOC105374767', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63975789, end_position_on_the_genomic_accession=63992698, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374768, CurrentID=0, Status='live', Symbol='LOC105374768', Aliases='', description='uncharacterized LOC105374768', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64302869, end_position_on_the_genomic_accession=64341706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219428, CurrentID=0, Status='live', Symbol='OR4C16', Aliases='OR11-135', description='olfactory receptor family 4 subfamily C member 16 (gene/pseudogene)', other_designations='olfactory receptor OR11-135|olfactory receptor, family 4, subfamily C, member 16', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55572128, end_position_on_the_genomic_accession=55573060, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219429, CurrentID=0, Status='live', Symbol='OR4C11', Aliases='OR11-136P, OR4C11', description='olfactory receptor family 4 subfamily C member 11', other_designations='olfactory receptor OR11-136|olfactory receptor, family 4, subfamily C, member 11 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55603354, end_position_on_the_genomic_accession=55604398, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219431, CurrentID=0, Status='live', Symbol='OR4S2', Aliases='OR11-137P, OST725, OR4S2', description='olfactory receptor family 4 subfamily S member 2', other_designations='olfactory receptor OR11-137|olfactory receptor, family 4, subfamily S, member 2 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55650904, end_position_on_the_genomic_accession=55651839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219432, CurrentID=0, Status='live', Symbol='OR4C6', Aliases='OR11-138', description='olfactory receptor family 4 subfamily C member 6', other_designations='olfactory receptor OR11-138|olfactory receptor, family 4, subfamily C, member 6', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55658587, end_position_on_the_genomic_accession=55666189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  219436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219436, CurrentID=0, Status='live', Symbol='OR5D14', Aliases='OR11-141, OR11-150', description='olfactory receptor family 5 subfamily D member 14', other_designations='olfactory receptor OR11-141|olfactory receptor OR11-150|olfactory receptor, family 5, subfamily D, member 14', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55795556, end_position_on_the_genomic_accession=55796500, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219437, CurrentID=0, Status='live', Symbol='OR5L1', Aliases='OR11-151, OST262', description='olfactory receptor family 5 subfamily L member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-151|olfactory receptor, family 5, subfamily L, member 1', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55811467, end_position_on_the_genomic_accession=55812402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219438, CurrentID=0, Status='live', Symbol='OR5D18', Aliases='OR11-143', description='olfactory receptor family 5 subfamily D member 18', other_designations='olfactory receptor OR11-143|olfactory receptor OR11-152|olfactory receptor, family 5, subfamily D, member 18', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55818023, end_position_on_the_genomic_accession=55820591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374771, CurrentID=0, Status='live', Symbol='LOC105374771', Aliases='', description='uncharacterized LOC105374771', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64395432, end_position_on_the_genomic_accession=64437171, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  219445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219445, CurrentID=0, Status='live', Symbol='OR7E5P', Aliases='OR11-12, OR11-156, OR7F5P', description='olfactory receptor family 7 subfamily E member 5 pseudogene', other_designations='olfactory receptor OR11-156 pseudogene|olfactory receptor, family 7, subfamily E, member 5 pseudogene|olfactory receptor, family 7, subfamily F, member 5 pseudogene|seven transmembrane helix receptor', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55978703, end_position_on_the_genomic_accession=55986405, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  219447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219447, CurrentID=0, Status='live', Symbol='OR5AS1', Aliases='OR11-168', description='olfactory receptor family 5 subfamily AS member 1', other_designations='olfactory receptor OR11-168|olfactory receptor, family 5, subfamily AS, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56030419, end_position_on_the_genomic_accession=56031393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375556, CurrentID=0, Status='live', Symbol='LOC105375556', Aliases='', description='uncharacterized LOC105375556', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=147671699, end_position_on_the_genomic_accession=147673144, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219453, CurrentID=0, Status='live', Symbol='OR8K5', Aliases='OR11-174', description='olfactory receptor family 8 subfamily K member 5', other_designations='olfactory receptor OR11-174|olfactory receptor, family 8, subfamily K, member 5', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56159394, end_position_on_the_genomic_accession=56160317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374773, CurrentID=0, Status='live', Symbol='LOC105374773', Aliases='', description='uncharacterized LOC105374773', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64537400, end_position_on_the_genomic_accession=64557556, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374774, CurrentID=0, Status='live', Symbol='LOC105374774', Aliases='', description='uncharacterized LOC105374774', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64484674, end_position_on_the_genomic_accession=64501651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371192, CurrentID=0, Status='live', Symbol='LOC105371192', Aliases='', description='uncharacterized LOC105371192', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33027234, end_position_on_the_genomic_accession=33062726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  219464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219464, CurrentID=0, Status='live', Symbol='OR5T2', Aliases='OR11-177', description='olfactory receptor family 5 subfamily T member 2', other_designations='olfactory receptor OR11-177|olfactory receptor, family 5, subfamily T, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56232106, end_position_on_the_genomic_accession=56233185, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374775, CurrentID=0, Status='live', Symbol='LOC105374775', Aliases='', description='uncharacterized LOC105374775', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64644435, end_position_on_the_genomic_accession=64647690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377754, CurrentID=0, Status='live', Symbol='LOC105377754', Aliases='', description='uncharacterized LOC105377754', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178088666, end_position_on_the_genomic_accession=178093386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219469, CurrentID=0, Status='live', Symbol='OR8H1', Aliases='OR11-180', description='olfactory receptor family 8 subfamily H member 1', other_designations='olfactory receptor OR11-180|olfactory receptor, family 8, subfamily H, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56290127, end_position_on_the_genomic_accession=56291062, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219473, CurrentID=0, Status='live', Symbol='OR8K3', Aliases='OR11-181', description='olfactory receptor family 8 subfamily K member 3 (gene/pseudogene)', other_designations='olfactory receptor OR11-181|olfactory receptor, family 8, subfamily K, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56318307, end_position_on_the_genomic_accession=56319245, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219477, CurrentID=0, Status='live', Symbol='OR8J1', Aliases='OR11-183', description='olfactory receptor family 8 subfamily J member 1', other_designations='olfactory receptor OR11-183|olfactory receptor, family 8, subfamily J, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56360215, end_position_on_the_genomic_accession=56361288, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219479, CurrentID=0, Status='live', Symbol='OR5R1', Aliases='OR11-185P, OR5R1', description='olfactory receptor family 5 subfamily R member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-185|olfactory receptor, family 5, subfamily R, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56417258, end_position_on_the_genomic_accession=56418232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219482, CurrentID=0, Status='live', Symbol='OR5M3', Aliases='OR11-191', description='olfactory receptor family 5 subfamily M member 3', other_designations='olfactory receptor OR11-191|olfactory receptor, family 5, subfamily M, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56469488, end_position_on_the_genomic_accession=56470538, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219484, CurrentID=0, Status='live', Symbol='OR5M8', Aliases='OR11-194', description='olfactory receptor family 5 subfamily M member 8', other_designations='olfactory receptor OR11-194|olfactory receptor, family 5, subfamily M, member 8', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56490435, end_position_on_the_genomic_accession=56491370, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219487, CurrentID=0, Status='live', Symbol='OR5M11', Aliases='OR11-199', description='olfactory receptor family 5 subfamily M member 11', other_designations='olfactory receptor OR11-199 pseudogene|olfactory receptor, family 5, subfamily M, member 11', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56542340, end_position_on_the_genomic_accession=56543257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374779, CurrentID=0, Status='live', Symbol='LOC105374779', Aliases='', description='uncharacterized LOC105374779', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65070701, end_position_on_the_genomic_accession=65082529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  219493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219493, CurrentID=0, Status='live', Symbol='OR5AR1', Aliases='OR11-209', description='olfactory receptor family 5 subfamily AR member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-209|olfactory receptor, family 5, subfamily AR, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56663686, end_position_on_the_genomic_accession=56664618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927587, CurrentID=0, Status='live', Symbol='LOC101927587', Aliases='', description='uncharacterized LOC101927587', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83575787, end_position_on_the_genomic_accession=83860996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374783, CurrentID=0, Status='live', Symbol='LOC105374783', Aliases='', description='uncharacterized LOC105374783', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66336464, end_position_on_the_genomic_accession=66360141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100415894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100415894, CurrentID=0, Status='live', Symbol='GLM5', Aliases='', description='Glioma susceptibility 5', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613030, no_hdr0=''),
  219527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219527, CurrentID=0, Status='live', Symbol='LRRC55', Aliases='', description='leucine rich repeat containing 55', other_designations='BK channel auxiliary gamma subunit LRRC55|BK channel auxilliary gamma snit LRRC55', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57181747, end_position_on_the_genomic_accession=57191714, orientation='plus', exon_count=2, OMIM=615213, no_hdr0=''),
  255042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255042, CurrentID=0, Status='live', Symbol='MGC16723', Aliases='', description='uncharacterized protein MGC16723', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219537, CurrentID=0, Status='live', Symbol='SMTNL1', Aliases='CHASM', description='smoothelin-like 1', other_designations='calponin homology-associated smooth muscle protein', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57537552, end_position_on_the_genomic_accession=57550274, orientation='plus', exon_count=8, OMIM=613664, no_hdr0=''),
  219539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219539, CurrentID=0, Status='live', Symbol='YPEL4', Aliases='', description='yippee like 4', other_designations='yippee-like 4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57645087, end_position_on_the_genomic_accession=57649944, orientation='minus', exon_count=6, OMIM=609725, no_hdr0=''),
  219541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219541, CurrentID=0, Status='live', Symbol='MED19', Aliases='DT2P1G7, LCMR1AS, MED19', description='mediator complex subunit 19', other_designations='lung cancer metastasis-related protein 1|mediator of RNA polymerase II transcription, subunit 19 homolog|putative mediator subunit MED19AS protein', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57703709, end_position_on_the_genomic_accession=57712438, orientation='minus', exon_count=5, OMIM=612385, no_hdr0=''),
  105373390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373390, CurrentID=0, Status='live', Symbol='LOC105373390', Aliases='', description='uncharacterized LOC105373390', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=2831396, end_position_on_the_genomic_accession=2839440, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371195, CurrentID=0, Status='live', Symbol='LOC105371195', Aliases='', description='uncharacterized LOC105371195', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35025592, end_position_on_the_genomic_accession=35030242, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219557, CurrentID=0, Status='live', Symbol='C7orf62', Aliases='', description='chromosome 7 open reading frame 62', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88794106, end_position_on_the_genomic_accession=88795773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374793, CurrentID=0, Status='live', Symbol='LOC105374793', Aliases='', description='uncharacterized LOC105374793', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70418959, end_position_on_the_genomic_accession=70441467, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  219578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219578, CurrentID=0, Status='live', Symbol='ZNF804B', Aliases='', description='zinc finger protein 804B', other_designations='zinc finger 804B', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88759368, end_position_on_the_genomic_accession=89337057, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219595, CurrentID=0, Status='live', Symbol='FOLH1B', Aliases='FOLH2, FOLHP, GCP3, GCPIII, PSM, PSMA-LIKE, PSMAL', description='folate hydrolase 1B', other_designations='N-acetylated-alpha-linked-acidic dipeptidase|NAALADase|cell growth-inhibiting gene 26 protein|folate hydrolase 2|folate hydrolase pseudogene|glutamate carboxypeptidase III|prostate-specific membrane antigen-like protein|putative folate hydrolase 1B', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89637520, end_position_on_the_genomic_accession=89698718, orientation='plus', exon_count=19, OMIM=609020, no_hdr0=''),
  100128144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128144, CurrentID=0, Status='live', Symbol='LOC100128144', Aliases='', description='ribosomal protein L32 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6072823, end_position_on_the_genomic_accession=6073528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131918, CurrentID=0, Status='live', Symbol='LOC100131918', Aliases='', description='KIAA1191 pseudogene', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229689000, end_position_on_the_genomic_accession=229689904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129408, CurrentID=0, Status='live', Symbol='LOC100129408', Aliases='', description='uncharacterized LOC100129408', other_designations='', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=53475558, end_position_on_the_genomic_accession=53476272, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371197, CurrentID=0, Status='live', Symbol='LOC105371197', Aliases='', description='uncharacterized LOC105371197', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35139993, end_position_on_the_genomic_accession=35145951, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219621, CurrentID=0, Status='live', Symbol='C10orf107', Aliases='bA63A2.1', description='chromosome 10 open reading frame 107', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61662952, end_position_on_the_genomic_accession=61766766, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  219623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219623, CurrentID=0, Status='live', Symbol='TMEM26', Aliases='', description='transmembrane protein 26', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61406642, end_position_on_the_genomic_accession=61453450, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371867, CurrentID=0, Status='live', Symbol='LOC105371867', Aliases='', description='uncharacterized LOC105371867', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65879310, end_position_on_the_genomic_accession=65926424, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374804, CurrentID=0, Status='live', Symbol='LOC105374804', Aliases='', description='uncharacterized LOC105374804', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73352367, end_position_on_the_genomic_accession=73385681, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374805, CurrentID=0, Status='live', Symbol='LOC105374805', Aliases='', description='uncharacterized LOC105374805', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73613839, end_position_on_the_genomic_accession=73626560, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219654, CurrentID=0, Status='live', Symbol='ZCCHC24', Aliases='C10orf56, Z3CXXC8', description='zinc finger, CCHC domain containing 24', other_designations='zinc finger, 3CxxC-type 8', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79382325, end_position_on_the_genomic_accession=79445627, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643043, CurrentID=0, Status='live', Symbol='LOC643043', Aliases='', description='protein shisa-5 pseudogene', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41322925, end_position_on_the_genomic_accession=41323505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  219670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219670, CurrentID=0, Status='live', Symbol='ENKUR', Aliases='C10orf63, CFAP106', description='enkurin, TRPC channel interacting protein', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24981979, end_position_on_the_genomic_accession=25062279, orientation='minus', exon_count=9, OMIM=611025, no_hdr0=''),
  100131929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131929, CurrentID=0, Status='live', Symbol='LOC100131929', Aliases='', description='uncharacterized LOC100131929', other_designations='', map_location='7q22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=104428514, end_position_on_the_genomic_accession=104429712, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219681, CurrentID=0, Status='live', Symbol='ARMC3', Aliases='CT81, KU-CT-1', description='armadillo repeat containing 3', other_designations='beta-catenin-like protein|cancer/testis antigen 81', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22928024, end_position_on_the_genomic_accession=23038523, orientation='plus', exon_count=22, OMIM=611226, no_hdr0=''),
  105374811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374811, CurrentID=0, Status='live', Symbol='LOC105374811', Aliases='', description='uncharacterized LOC105374811', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75146353, end_position_on_the_genomic_accession=75279139, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  219688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219688, CurrentID=0, Status='live', Symbol='LOC219688', Aliases='', description='uncharacterized LOC219688', other_designations='', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374812, CurrentID=0, Status='live', Symbol='LOC105374812', Aliases='', description='uncharacterized LOC105374812', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75153943, end_position_on_the_genomic_accession=75189220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219690, CurrentID=0, Status='live', Symbol='LOC219690', Aliases='', description='uncharacterized LOC219690', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219699, CurrentID=0, Status='live', Symbol='UNC5B', Aliases='UNC5H2, p53RDL1', description='unc-5 netrin receptor B', other_designations='p53-regulated receptor for death and life protein 1|protein unc-5 homolog 2|protein unc-5 homolog B|transmembrane receptor Unc5H2|unc-5 homolog 2|unc-5 homolog B', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71212535, end_position_on_the_genomic_accession=71302879, orientation='plus', exon_count=19, OMIM=607870, no_hdr0=''),
  105374814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374814, CurrentID=0, Status='live', Symbol='LOC105374814', Aliases='', description='uncharacterized LOC105374814', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76506881, end_position_on_the_genomic_accession=76589939, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374815, CurrentID=0, Status='live', Symbol='LOC105374815', Aliases='', description='uncharacterized LOC105374815', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=76854860, end_position_on_the_genomic_accession=76859183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400123, CurrentID=0, Status='live', Symbol='LINC00548', Aliases='', description='long intergenic non-protein coding RNA 548', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40194509, end_position_on_the_genomic_accession=40220502, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372810, CurrentID=0, Status='live', Symbol='LOC105372810', Aliases='', description='putative zinc finger protein 355P', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13096329, end_position_on_the_genomic_accession=13097778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646535, CurrentID=0, Status='live', Symbol='LOC646535', Aliases='', description='60S ribosomal protein L7a-like', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44539779, end_position_on_the_genomic_accession=44540693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374818, CurrentID=0, Status='live', Symbol='LOC105374818', Aliases='', description='uncharacterized LOC105374818', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78409494, end_position_on_the_genomic_accession=78542064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219731, CurrentID=0, Status='live', Symbol='PROSER2-AS1', Aliases='', description='PROSER2 antisense RNA 1', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11849608, end_position_on_the_genomic_accession=11894710, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  219736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219736, CurrentID=0, Status='live', Symbol='STOX1', Aliases='C10orf24, PEE4', description='storkhead box 1', other_designations='winged-helix domain-containing protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68827537, end_position_on_the_genomic_accession=68895453, orientation='plus', exon_count=6, OMIM=609397, no_hdr0=''),
  105374820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374820, CurrentID=0, Status='live', Symbol='LOC105374820', Aliases='', description='uncharacterized LOC105374820', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78542135, end_position_on_the_genomic_accession=78566088, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219738, CurrentID=0, Status='live', Symbol='C10orf35', Aliases='', description='chromosome 10 open reading frame 35', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69630243, end_position_on_the_genomic_accession=69633599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  219743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219743, CurrentID=0, Status='live', Symbol='TYSND1', Aliases='NET41', description='trypsin domain containing 1', other_designations='peroxisomal cysteine endopeptidase|trypsin domain-containing protein 1', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70137981, end_position_on_the_genomic_accession=70146740, orientation='minus', exon_count=6, OMIM=611017, no_hdr0=''),
  219749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219749, CurrentID=0, Status='live', Symbol='ZNF25', Aliases='KOX19, Zfp9', description='zinc finger protein 25', other_designations='zinc finger protein 25 (KOX 19)|zinc finger protein KOX19', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37949572, end_position_on_the_genomic_accession=37976980, orientation='minus', exon_count=11, OMIM=194528, no_hdr0=''),
  219752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219752, CurrentID=0, Status='live', Symbol='TACC1P1', Aliases='', description='transforming, acidic coiled-coil containing protein 1 pseudogene 1', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37523010, end_position_on_the_genomic_accession=37530482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374823, CurrentID=0, Status='live', Symbol='LOC105374823', Aliases='', description='uncharacterized LOC105374823', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79269925, end_position_on_the_genomic_accession=79292856, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100131945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131945, CurrentID=0, Status='live', Symbol='RPL35AP13', Aliases='RPL35A_5_575', description='ribosomal protein L35a pseudogene 13', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72878824, end_position_on_the_genomic_accession=72879217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  219770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219770, CurrentID=0, Status='live', Symbol='GJD4', Aliases='CX40.1', description='gap junction protein delta 4', other_designations='connexin 40.1|connexin-40.1|connexin40.1|gap junction protein, delta 4, 40.1kDa', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35605410, end_position_on_the_genomic_accession=35608935, orientation='plus', exon_count=2, OMIM=611922, no_hdr0=''),
  219771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219771, CurrentID=0, Status='live', Symbol='CCNY', Aliases='C10orf9, CBCP1, CCNX, CFP1', description='cyclin Y', other_designations='cyc-Y|cyclin box protein 1|cyclin fold protein 1|cyclin-X|cyclin-box carrying protein 1', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35246981, end_position_on_the_genomic_accession=35572670, orientation='plus', exon_count=23, OMIM=612786, no_hdr0=''),
  105374826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374826, CurrentID=0, Status='live', Symbol='LOC105374826', Aliases='', description='uncharacterized LOC105374826', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79762194, end_position_on_the_genomic_accession=79778096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374827, CurrentID=0, Status='live', Symbol='LOC105374827', Aliases='', description='uncharacterized LOC105374827', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=80912464, end_position_on_the_genomic_accession=80955011, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373613, CurrentID=0, Status='live', Symbol='LOC105373613', Aliases='', description='uncharacterized LOC105373613', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129402519, end_position_on_the_genomic_accession=129462087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  219790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219790, CurrentID=0, Status='live', Symbol='RTKN2', Aliases='PLEKHK1, bA531F24.1', description='rhotekin 2', other_designations='PH domain-containing family K member 1|pleckstrin homology domain containing, family K member 1|pleckstrin homology domain-containing family K member 1', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62183035, end_position_on_the_genomic_accession=62268863, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  219793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219793, CurrentID=0, Status='live', Symbol='TBATA', Aliases='C10orf27, SPATIAL', description='thymus, brain and testes associated', other_designations='stromal protein associated with thymii and lymph node homolog|thymus, brain and testes-associated protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70771238, end_position_on_the_genomic_accession=70785401, orientation='minus', exon_count=18, OMIM=612640, no_hdr0=''),
  100131950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131950, CurrentID=0, Status='live', Symbol='LOC100131950', Aliases='', description='methyltransferase like 5 pseudogene', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74873629, end_position_on_the_genomic_accession=74874254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374531, CurrentID=0, Status='live', Symbol='LOC105374531', Aliases='', description='uncharacterized LOC105374531', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42083474, end_position_on_the_genomic_accession=42097793, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374833, CurrentID=0, Status='live', Symbol='LOC105374833', Aliases='', description='uncharacterized LOC105374833', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=83256589, end_position_on_the_genomic_accession=83293509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  88745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=88745, CurrentID=0, Status='live', Symbol='RRP36', Aliases='C6orf153, dJ20C7.4', description='ribosomal RNA processing 36', other_designations='RNA processing factor|ribosomal RNA processing 36 homolog', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43021636, end_position_on_the_genomic_accession=43029599, orientation='plus', exon_count=7, OMIM=613475, no_hdr0=''),
  613037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613037, CurrentID=0, Status='live', Symbol='LOC613037', Aliases='', description='nuclear pore complex interacting protein pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30223029, end_position_on_the_genomic_accession=30245611, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  613038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613038, CurrentID=0, Status='live', Symbol='LOC613038', Aliases='', description='SAGA complex associated factor 29 pseudogene', other_designations='SAGA-associated factor 29 homolog|coiled-coil domain containing 101 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30204316, end_position_on_the_genomic_accession=30206927, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928903, CurrentID=0, Status='live', Symbol='LOC101928903', Aliases='', description='N-alpha-acetyltransferase 20 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132850014, end_position_on_the_genomic_accession=132907583, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  219833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219833, CurrentID=0, Status='live', Symbol='C11orf45', Aliases='', description='chromosome 11 open reading frame 45', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128899565, end_position_on_the_genomic_accession=128909717, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  219844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219844, CurrentID=0, Status='live', Symbol='HYLS1', Aliases='HLS', description='hydrolethalus syndrome 1', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125883614, end_position_on_the_genomic_accession=125900646, orientation='plus', exon_count=5, OMIM=610693, no_hdr0=''),
  105374838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374838, CurrentID=0, Status='live', Symbol='LOC105374838', Aliases='', description='uncharacterized LOC105374838', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84926031, end_position_on_the_genomic_accession=84964294, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219854, CurrentID=0, Status='live', Symbol='TMEM218', Aliases='', description='transmembrane protein 218', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125094370, end_position_on_the_genomic_accession=125111763, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  219855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219855, CurrentID=0, Status='live', Symbol='SLC37A2', Aliases='pp11662', description='solute carrier family 37 (glucose-6-phosphate transporter), member 2', other_designations='solute carrier family 37 (glycerol-3-phosphate transporter), member 2|solute carrier family 37 member 2', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125063067, end_position_on_the_genomic_accession=125090516, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  219858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219858, CurrentID=0, Status='live', Symbol='OR8B12', Aliases='OR11-317', description='olfactory receptor family 8 subfamily B member 12', other_designations='olfactory receptor OR11-317|olfactory receptor, family 8, subfamily B, member 12', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124542722, end_position_on_the_genomic_accession=124543654, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374841, CurrentID=0, Status='live', Symbol='LOC105374841', Aliases='', description='uncharacterized LOC105374841', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85413992, end_position_on_the_genomic_accession=85416576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  219865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219865, CurrentID=0, Status='live', Symbol='OR8G5', Aliases='OR11-298P, OR8G6, OR8G5', description='olfactory receptor family 8 subfamily G member 5', other_designations='olfactory receptor 8G6|olfactory receptor OR11-298|olfactory receptor, family 8, subfamily G, member 5 pseudogene|olfactory receptor, family 8, subfamily G, member 6', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124241158, end_position_on_the_genomic_accession=124323224, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  219869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219869, CurrentID=0, Status='live', Symbol='OR10G8', Aliases='OR11-274, OR11-282', description='olfactory receptor family 10 subfamily G member 8', other_designations='olfactory receptor OR11-274 pseudogene|olfactory receptor OR11-282|olfactory receptor, family 10, subfamily G, member 8', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124029623, end_position_on_the_genomic_accession=124030558, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219870, CurrentID=0, Status='live', Symbol='OR10G9', Aliases='OR10G10P', description='olfactory receptor family 10 subfamily G member 9', other_designations='olfactory receptor 10G10|olfactory receptor OR11-272|olfactory receptor, family 10, subfamily G, member 9', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124023013, end_position_on_the_genomic_accession=124023948, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219873, CurrentID=0, Status='live', Symbol='OR10S1', Aliases='OR11-279', description='olfactory receptor family 10 subfamily S member 1', other_designations='olfactory receptor OR11-279|olfactory receptor, family 10, subfamily S, member 1', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123976696, end_position_on_the_genomic_accession=123977691, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219874, CurrentID=0, Status='live', Symbol='OR6T1', Aliases='OR11-277', description='olfactory receptor family 6 subfamily T member 1', other_designations='olfactory receptor OR11-277|olfactory receptor, family 6, subfamily T, member 1', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123942432, end_position_on_the_genomic_accession=123993780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219875, CurrentID=0, Status='live', Symbol='OR4D5', Aliases='OR11-276', description='olfactory receptor family 4 subfamily D member 5', other_designations='olfactory receptor OR11-276|olfactory receptor, family 4, subfamily D, member 5', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123939617, end_position_on_the_genomic_accession=123940573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374844, CurrentID=0, Status='live', Symbol='LOC105374844', Aliases='', description='uncharacterized LOC105374844', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86409955, end_position_on_the_genomic_accession=86440973, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421840, CurrentID=0, Status='live', Symbol='LOC100421840', Aliases='', description='eukaryotic translation elongation factor 1 alpha 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199387141, end_position_on_the_genomic_accession=199388469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374845, CurrentID=0, Status='live', Symbol='LOC105374845', Aliases='', description='uncharacterized LOC105374845', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86562068, end_position_on_the_genomic_accession=86620173, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219899, CurrentID=0, Status='live', Symbol='TBCEL', Aliases='El, LRRC35', description='tubulin folding cofactor E-like', other_designations='E-like|catastrophin|leucine rich repeat containing 35|leucine-rich repeat-containing protein 35|tubulin-specific chaperone e-like', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121024094, end_position_on_the_genomic_accession=121089647, orientation='plus', exon_count=8, OMIM=610451, no_hdr0=''),
  219902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219902, CurrentID=0, Status='live', Symbol='TMEM136', Aliases='', description='transmembrane protein 136', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120324630, end_position_on_the_genomic_accession=120333688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374848, CurrentID=0, Status='live', Symbol='LOC105374848', Aliases='', description='uncharacterized LOC105374848', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87319419, end_position_on_the_genomic_accession=87323434, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  613126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613126, CurrentID=0, Status='live', Symbol='CYP51A1-AS1', Aliases='LRRD1-AS1', description='CYP51A1 antisense RNA 1', other_designations='CTB-161K23.1|CYP51A1 antisense RNA 1 (head to head)', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92134563, end_position_on_the_genomic_accession=92180725, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102725065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725065, CurrentID=0, Status='live', Symbol='LOC102725065', Aliases='', description='uncharacterized LOC102725065', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44208134, end_position_on_the_genomic_accession=44210115, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371207, CurrentID=0, Status='live', Symbol='LOC105371207', Aliases='', description='uncharacterized LOC105371207', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144148427, end_position_on_the_genomic_accession=144152410, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374851, CurrentID=0, Status='live', Symbol='LOC105374851', Aliases='', description='uncharacterized LOC105374851', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87990654, end_position_on_the_genomic_accession=87991283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219927, CurrentID=0, Status='live', Symbol='MRPL21', Aliases='L21mt, MRP-L21', description='mitochondrial ribosomal protein L21', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68891272, end_position_on_the_genomic_accession=68903835, orientation='minus', exon_count=7, OMIM=611834, no_hdr0=''),
  219931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219931, CurrentID=0, Status='live', Symbol='TPCN2', Aliases='SHEP10, TPC2', description='two pore segment channel 2', other_designations='two-pore calcium channel protein 2|voltage-dependent calcium channel protein TPC2', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69048882, end_position_on_the_genomic_accession=69160413, orientation='plus', exon_count=30, OMIM=612163, no_hdr0=''),
  219938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219938, CurrentID=0, Status='live', Symbol='SPATA19', Aliases='CT132, SPAS1, spergen1', description='spermatogenesis associated 19', other_designations='cancer/testis antigen 132|spergen-1|spermatogenic cell-specific gene 1 protein|spermatogenic specific-gene1', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133835930, end_position_on_the_genomic_accession=133845538, orientation='minus', exon_count=8, OMIM=609805, no_hdr0=''),
  105374855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374855, CurrentID=0, Status='live', Symbol='LOC105374855', Aliases='', description='uncharacterized LOC105374855', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88810391, end_position_on_the_genomic_accession=88825340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379503, CurrentID=0, Status='live', Symbol='LOC105379503', Aliases='', description='uncharacterized LOC105379503', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7421931, end_position_on_the_genomic_accession=7423788, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219952, CurrentID=0, Status='live', Symbol='OR6Q1', Aliases='OR11-226', description='olfactory receptor family 6 subfamily Q member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-226|olfactory receptor, family 6, subfamily Q, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58030953, end_position_on_the_genomic_accession=58031906, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219954, CurrentID=0, Status='live', Symbol='OR9I1', Aliases='OR11-228', description='olfactory receptor family 9 subfamily I member 1', other_designations='olfactory receptor OR11-228|olfactory receptor, family 9, subfamily I, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58118500, end_position_on_the_genomic_accession=58119444, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219956, CurrentID=0, Status='live', Symbol='OR9Q1', Aliases='', description='olfactory receptor family 9 subfamily Q member 1', other_designations='olfactory receptor, family 9, subfamily Q, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58023881, end_position_on_the_genomic_accession=58181587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  219957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219957, CurrentID=0, Status='live', Symbol='OR9Q2', Aliases='OR9Q2P', description='olfactory receptor family 9 subfamily Q member 2', other_designations='olfactory receptor, family 9, subfamily Q, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58190434, end_position_on_the_genomic_accession=58191518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219958, CurrentID=0, Status='live', Symbol='OR1S2', Aliases='OR11-231', description='olfactory receptor family 1 subfamily S member 2', other_designations='olfactory receptor OR11-231|olfactory receptor, family 1, subfamily S, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58203204, end_position_on_the_genomic_accession=58204181, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219959, CurrentID=0, Status='live', Symbol='OR1S1', Aliases='OR11-232, OST034', description='olfactory receptor family 1 subfamily S member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-232|olfactory receptor, family 1, subfamily S, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58214745, end_position_on_the_genomic_accession=58215722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219960, CurrentID=0, Status='live', Symbol='OR10Q1', Aliases='OR11-233', description='olfactory receptor family 10 subfamily Q member 1', other_designations='olfactory receptor OR11-233|olfactory receptor, family 10, subfamily Q, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58227882, end_position_on_the_genomic_accession=58228918, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219965, CurrentID=0, Status='live', Symbol='OR5B17', Aliases='OR11-237, OR5B20P', description='olfactory receptor family 5 subfamily B member 17', other_designations='olfactory receptor 5B20|olfactory receptor OR11-237|olfactory receptor, family 5, subfamily B, member 17|olfactory receptor, family 5, subfamily B, member 20 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58358125, end_position_on_the_genomic_accession=58359069, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219968, CurrentID=0, Status='live', Symbol='OR5B21', Aliases='', description='olfactory receptor family 5 subfamily B member 21', other_designations='olfactory receptor, family 5, subfamily B, member 21', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58507176, end_position_on_the_genomic_accession=58508105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219970, CurrentID=0, Status='live', Symbol='GLYATL2', Aliases='BXMAS2-10, GATF-B', description='glycine-N-acyltransferase like 2', other_designations='acyl-CoA:glycine N-acyltransferase-like protein 2|glycine acyltransferase family-B|glycine-N-acyltransferase-like 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58834065, end_position_on_the_genomic_accession=58844808, orientation='minus', exon_count=6, OMIM=614762, no_hdr0=''),
  219972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219972, CurrentID=0, Status='live', Symbol='MPEG1', Aliases='MPG1, MPS1, Mpg-1', description='macrophage expressed 1', other_designations='macrophage expressed gene 1|macrophage gene 1 protein|perforin 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59208510, end_position_on_the_genomic_accession=59213021, orientation='minus', exon_count=1, OMIM=610390, no_hdr0=''),
  124044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124044, CurrentID=0, Status='live', Symbol='SPATA2L', Aliases='C16orf76, tamo', description='spermatogenesis associated 2-like', other_designations='SPATA2-like protein', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89696357, end_position_on_the_genomic_accession=89701713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  219981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219981, CurrentID=0, Status='live', Symbol='OR5A2', Aliases='', description='olfactory receptor family 5 subfamily A member 2', other_designations='olfactory receptor OR11-248|olfactory receptor, family 5, subfamily A, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59421979, end_position_on_the_genomic_accession=59422953, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  219982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219982, CurrentID=0, Status='live', Symbol='OR5A1', Aliases='OR11-249P, OST181, OR5A1', description='olfactory receptor family 5 subfamily A member 1', other_designations='olfactory receptor OR11-249|olfactory receptor, family 5, subfamily A, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59438057, end_position_on_the_genomic_accession=59444188, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219983, CurrentID=0, Status='live', Symbol='OR4D6', Aliases='OR11-250', description='olfactory receptor family 4 subfamily D member 6', other_designations='olfactory receptor OR11-250|olfactory receptor, family 4, subfamily D, member 6', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59456961, end_position_on_the_genomic_accession=59457905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219986, CurrentID=0, Status='live', Symbol='OR4D11', Aliases='OR4D11P', description='olfactory receptor family 4 subfamily D member 11', other_designations='olfactory receptor OR11-252 pseudogene|olfactory receptor, family 4, subfamily D, member 11 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59503576, end_position_on_the_genomic_accession=59504511, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219988, CurrentID=0, Status='live', Symbol='PATL1', Aliases='Pat1b, hPat1b', description='protein associated with topoisomerase II homolog 1 (yeast)', other_designations='PAT1-like protein 1|protein PAT1 homolog b', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59636716, end_position_on_the_genomic_accession=59669040, orientation='minus', exon_count=21, OMIM=614660, no_hdr0=''),
  219990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219990, CurrentID=0, Status='live', Symbol='OOSP2', Aliases='PLAC1L, TMEM122', description='oocyte secreted protein 2', other_designations='placenta-specific 1-like protein|transmembrane protein 122', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60040275, end_position_on_the_genomic_accession=60048044, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103504727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504727, CurrentID=0, Status='live', Symbol='MIR6724-2', Aliases='hsa-mir-6724-2', description='microRNA 6724-2', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8249505, end_position_on_the_genomic_accession=8249596, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103504728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504728, CurrentID=0, Status='live', Symbol='MIR3687-2', Aliases='hsa-mir-3687-2', description='microRNA 3687-2', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8987370, end_position_on_the_genomic_accession=8987430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  613209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613209, CurrentID=0, Status='live', Symbol='DEFB135', Aliases='DEFB136', description='defensin beta 135', other_designations='beta-defensin 136|defensin, beta 135', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11982321, end_position_on_the_genomic_accession=11984590, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  613210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613210, CurrentID=0, Status='live', Symbol='DEFB136', Aliases='DEFB137', description='defensin beta 136', other_designations='beta-defensin 137|defensin, beta 136', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11973937, end_position_on_the_genomic_accession=11974599, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  219995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=219995, CurrentID=0, Status='live', Symbol='MS4A15', Aliases='', description='membrane-spanning 4-domains, subfamily A, member 15', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60756867, end_position_on_the_genomic_accession=60776733, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  613212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613212, CurrentID=0, Status='live', Symbol='CTXN3', Aliases='KABE', description='cortexin 3', other_designations='kidney and brain-expressed protein', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127649021, end_position_on_the_genomic_accession=127658630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103504733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504733, CurrentID=0, Status='live', Symbol='MIR6724-4', Aliases='hsa-mir-6724-4', description='microRNA 6724-4', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8432530, end_position_on_the_genomic_accession=8432621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103504734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504734, CurrentID=0, Status='live', Symbol='MIR1244-4', Aliases='hsa-mir-1244-4', description='microRNA 1244-4', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12111952, end_position_on_the_genomic_accession=12112036, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103504735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504735, CurrentID=0, Status='live', Symbol='MIR548BB', Aliases='MI0029321, hsa-mir-548bb', description='microRNA 548bb', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60617805, end_position_on_the_genomic_accession=60617870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  103504736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504736, CurrentID=0, Status='live', Symbol='MIR8069-2', Aliases='hsa-mir-8069-2', description='microRNA 8069-2', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13724189, end_position_on_the_genomic_accession=13724274, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  220001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220001, CurrentID=0, Status='live', Symbol='VWCE', Aliases='URG11, VWC1', description='von Willebrand factor C and EGF domains', other_designations='HBV X protein up-regulated gene 11 protein|HBxAg up-regulated gene 11 protein', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61258286, end_position_on_the_genomic_accession=61295316, orientation='minus', exon_count=21, OMIM=611115, no_hdr0=''),
  220002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220002, CurrentID=0, Status='live', Symbol='CYB561A3', Aliases='CYBASC3, LCYTB', description='cytochrome b561 family member A3', other_designations='cytochrome b, ascorbate dependent 3|cytochrome b561 family, member A3|lysosomal cytochrome b', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61348748, end_position_on_the_genomic_accession=61362299, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  103504739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504739, CurrentID=0, Status='live', Symbol='MIR6724-3', Aliases='hsa-mir-6724-3', description='microRNA 6724-3', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8388362, end_position_on_the_genomic_accession=8388453, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  220004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220004, CurrentID=0, Status='live', Symbol='PPP1R32', Aliases='C11orf66, IIIG9', description='protein phosphatase 1 regulatory subunit 32', other_designations='protein phosphatase 1, regulatory subunit 32', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61481113, end_position_on_the_genomic_accession=61490928, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  103504741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504741, CurrentID=0, Status='live', Symbol='MIR5701-3', Aliases='hsa-mir-5701-3', description='microRNA 5701-3', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21951242, end_position_on_the_genomic_accession=21951323, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  613227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613227, CurrentID=0, Status='live', Symbol='HIGD1C', Aliases='GM921', description='HIG1 hypoxia inducible domain family member 1C', other_designations='HIG1 domain family, member 1C|HIG1 hypoxia inducible domain family, member 1C', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50936223, end_position_on_the_genomic_accession=50972571, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129422, CurrentID=0, Status='live', Symbol='LOC100129422', Aliases='', description='LSM14A, SCD6 homolog A (S. cerevisiae) pseudogene', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121682925, end_position_on_the_genomic_accession=121684266, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101669760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669760, CurrentID=0, Status='live', Symbol='PAURT1', Aliases='', description='Preauricular tag, isolated, autosomal dominant, 1', other_designations='', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610420, no_hdr0=''),
  101669762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669762, CurrentID=0, Status='live', Symbol='BLACAT1', Aliases='LINC00912, linc-UBC1, onco-lncRNA-30', description='bladder cancer associated transcript 1 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 912|long intergenic noncoding RNA up-regulated in bladder cancer', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205434886, end_position_on_the_genomic_accession=205456086, orientation='minus', exon_count=2, OMIM=615480, no_hdr0=''),
  101669763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669763, CurrentID=0, Status='live', Symbol='JADRR', Aliases='LINC00915, lncRNA-JADE', description='JADE1 adjacent regulatory RNA', other_designations='long intergenic non-protein coding RNA 915', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101669764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669764, CurrentID=0, Status='live', Symbol='GPR1-AS', Aliases='GPR1-AS1, GPR1AS', description='GPR1 antisense RNA', other_designations='G protein-coupled receptor 1 antisense RNA|GPR1 antisense RNA 1', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206203376, end_position_on_the_genomic_accession=206266243, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  613253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613253, CurrentID=0, Status='live', Symbol='DEFA1A3', Aliases='DEFA1, DEFA3, DEFT1P', description='defensin alpha 1 and alpha 3, variable copy number locus', other_designations='defensin, alpha 1 and alpha 3, variable copy number locus', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101669766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669766, CurrentID=0, Status='live', Symbol='ZNF350-AS1', Aliases='HCCAT3, uc002pyc', description='ZNF350 antisense RNA 1', other_designations='CTC-429C10.4|hepatocellular carcinoma associated transcript 3 (non-protein coding)', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51949134, end_position_on_the_genomic_accession=51981367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101669767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101669767, CurrentID=0, Status='live', Symbol='SCHLAP1', Aliases='LINC00913, PCAT11, PCAT114', description='SWI/SNF complex antagonist associated with prostate cancer 1 (non-protein coding)', other_designations='Second Chromosome Locus Associated with Prostate 1|long intergenic non-protein coding RNA 913|prostate cancer associated transcript 11', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180692104, end_position_on_the_genomic_accession=180916939, orientation='plus', exon_count=7, OMIM=615568, no_hdr0=''),
  220042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220042, CurrentID=0, Status='live', Symbol='DDIAS', Aliases='C11orf82, noxin', description='DNA damage-induced apoptosis suppressor', other_designations='nitric oxide-inducible gene protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82901695, end_position_on_the_genomic_accession=82934659, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  648343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648343, CurrentID=0, Status='live', Symbol='RPS29P9', Aliases='RPS29_6_317', description='ribosomal protein S29 pseudogene 9', other_designations='40S ribosomal protein S29 pseudogene', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207654751, end_position_on_the_genomic_accession=207655049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  220047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220047, CurrentID=0, Status='live', Symbol='CCDC83', Aliases='CT148, HSD9', description='coiled-coil domain containing 83', other_designations='', map_location='11q14.1-q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85855055, end_position_on_the_genomic_accession=85920021, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105374872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374872, CurrentID=0, Status='live', Symbol='LOC105374872', Aliases='', description='uncharacterized LOC105374872', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=356608, end_position_on_the_genomic_accession=363395, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  613266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=613266, CurrentID=0, Status='live', Symbol='LOC613266', Aliases='', description='uncharacterized LOC613266', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=15892333, end_position_on_the_genomic_accession=15985882, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374874, CurrentID=0, Status='live', Symbol='LOC105374874', Aliases='', description='uncharacterized LOC105374874', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=597631, end_position_on_the_genomic_accession=602734, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  220064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220064, CurrentID=0, Status='live', Symbol='ORAOV1', Aliases='TAOS1', description='oral cancer overexpressed 1', other_designations='oral cancer overexpressed protein 1-A|tumor-amplified and overexpressed sequence 1', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69653420, end_position_on_the_genomic_accession=69675397, orientation='minus', exon_count=6, OMIM=607224, no_hdr0=''),
  105374875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374875, CurrentID=0, Status='live', Symbol='LOC105374875', Aliases='', description='uncharacterized LOC105374875', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=451012, end_position_on_the_genomic_accession=486528, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  220070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220070, CurrentID=0, Status='live', Symbol='SHANK2-AS3', Aliases='C11orf76', description='SHANK2 antisense RNA 3', other_designations='SHANK2 antisense RNA 3 (non-protein coding)', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70862790, end_position_on_the_genomic_accession=70865115, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  351143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=351143, CurrentID=0, Status='live', Symbol='MRPL50P1', Aliases='', description='mitochondrial ribosomal protein L50 pseudogene 1', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=35724759, end_position_on_the_genomic_accession=35725156, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  220074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220074, CurrentID=0, Status='live', Symbol='LRTOMT', Aliases='CFAP111, DFNB63, LRRC51', description='leucine rich transmembrane and O-methyltransferase domain containing', other_designations='leucine rich transmembrane and 0-methyltransferase domain containing', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72080331, end_position_on_the_genomic_accession=72110782, orientation='plus', exon_count=13, OMIM=612414, no_hdr0=''),
  220077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220077, CurrentID=0, Status='live', Symbol='LOC220077', Aliases='', description='dedicator of cytokinesis 1 pseudogene', other_designations='docking protein 1 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72249420, end_position_on_the_genomic_accession=72251308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374877, CurrentID=0, Status='live', Symbol='LOC105374877', Aliases='', description='uncharacterized LOC105374877', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1026056, end_position_on_the_genomic_accession=1029557, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  220081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220081, CurrentID=0, Status='live', Symbol='ERICH6B', Aliases='FAM194B', description='glutamate rich 6B', other_designations='family with sequence similarity 194, member B|glutamate-rich 6B|protein FAM194B', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45541241, end_position_on_the_genomic_accession=45615739, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  220082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220082, CurrentID=0, Status='live', Symbol='SPERT', Aliases='CBY2, NURIT', description='spermatid associated', other_designations='novel leucine zipper testicular protein|protein chibby homolog 2|spermatid flower-like structure protein|testis-specific leucine zipper protein nurit', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45702292, end_position_on_the_genomic_accession=45714559, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374878, CurrentID=0, Status='live', Symbol='LOC105374878', Aliases='', description='uncharacterized LOC105374878', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1169643, end_position_on_the_genomic_accession=1172029, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374881, CurrentID=0, Status='live', Symbol='LOC105374881', Aliases='', description='uncharacterized LOC105374881', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1321236, end_position_on_the_genomic_accession=1335651, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  220107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220107, CurrentID=0, Status='live', Symbol='DLEU7', Aliases='', description='deleted in lymphocytic leukemia, 7', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50695550, end_position_on_the_genomic_accession=50843933, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  220108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220108, CurrentID=0, Status='live', Symbol='FAM124A', Aliases='', description='family with sequence similarity 124 member A', other_designations='family with sequence similarity 124A', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51222334, end_position_on_the_genomic_accession=51284241, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  220112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220112, CurrentID=0, Status='live', Symbol='CTAGE3P', Aliases='CTAGE-3, CTAGE3', description='CTAGE family member 3, pseudogene', other_designations='CTAGE family, member 3, pseudogene|cutaneous T-cell lymphoma-associated antigen 3', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51907820, end_position_on_the_genomic_accession=51910559, orientation='minus', exon_count=0, OMIM=608857, no_hdr0=''),
  220115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220115, CurrentID=0, Status='live', Symbol='TPTE2P3', Aliases='TPTEps1', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 3', other_designations='TPTE and PTEN homologous inositol lipid phosphatase pseudogene', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52488993, end_position_on_the_genomic_accession=52587090, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105374884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374884, CurrentID=0, Status='live', Symbol='LOC105374884', Aliases='', description='uncharacterized LOC105374884', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1855239, end_position_on_the_genomic_accession=1858263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374885, CurrentID=0, Status='live', Symbol='LOC105374885', Aliases='', description='uncharacterized LOC105374885', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2629267, end_position_on_the_genomic_accession=2642796, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372521, CurrentID=0, Status='live', Symbol='LOC105372521', Aliases='', description='uncharacterized LOC105372521', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8624469, end_position_on_the_genomic_accession=8657173, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374886, CurrentID=0, Status='live', Symbol='LOC105374886', Aliases='', description='uncharacterized LOC105374886', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2790732, end_position_on_the_genomic_accession=2805235, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  220134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220134, CurrentID=0, Status='live', Symbol='SKA1', Aliases='C18orf24', description='spindle and kinetochore associated complex subunit 1', other_designations='spindle and KT (kinetochore) associated 1', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50375022, end_position_on_the_genomic_accession=50394168, orientation='plus', exon_count=7, OMIM=616673, no_hdr0=''),
  220136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220136, CurrentID=0, Status='live', Symbol='CFAP53', Aliases='CCDC11, HTX6', description='cilia and flagella associated protein 53', other_designations='coiled-coil domain containing 11', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50227193, end_position_on_the_genomic_accession=50266495, orientation='minus', exon_count=8, OMIM=614759, no_hdr0=''),
  105374887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374887, CurrentID=0, Status='live', Symbol='LOC105374887', Aliases='', description='uncharacterized LOC105374887', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2940398, end_position_on_the_genomic_accession=2941406, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373394, CurrentID=0, Status='live', Symbol='LOC105373394', Aliases='', description='uncharacterized LOC105373394', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3974161, end_position_on_the_genomic_accession=3975170, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  220147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220147, CurrentID=0, Status='live', Symbol='CTBP2P3', Aliases='', description='C-terminal binding protein 2 pseudogene 3', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60663039, end_position_on_the_genomic_accession=60664370, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132009, CurrentID=0, Status='live', Symbol='LOC100132009', Aliases='', description='pentatricopeptide repeat domain 3 pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57175231, end_position_on_the_genomic_accession=57177919, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  220158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220158, CurrentID=0, Status='live', Symbol='GTSCR1', Aliases='', description='Gilles de la Tourette syndrome chromosome region, candidate 1 (non-protein coding)', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70630519, end_position_on_the_genomic_accession=70670875, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  220164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220164, CurrentID=0, Status='live', Symbol='DOK6', Aliases='DOK5L, HsT3226', description='docking protein 6', other_designations='downstream of tyrosine kinase 6', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=69401048, end_position_on_the_genomic_accession=69849087, orientation='plus', exon_count=11, OMIM=611402, no_hdr0=''),
  105374894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374894, CurrentID=0, Status='live', Symbol='LOC105374894', Aliases='', description='uncharacterized LOC105374894', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4497945, end_position_on_the_genomic_accession=4583834, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374895, CurrentID=0, Status='live', Symbol='LOC105374895', Aliases='', description='uncharacterized LOC105374895', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4488809, end_position_on_the_genomic_accession=4494613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  89122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89122, CurrentID=0, Status='live', Symbol='TRIM4', Aliases='RNF87', description='tripartite motif containing 4', other_designations='RING finger protein 87|tripartite motif protein 4|tripartite motif-containing 4', map_location='7q22-q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99890407, end_position_on_the_genomic_accession=99919600, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927610, CurrentID=0, Status='live', Symbol='LOC101927610', Aliases='', description='uncharacterized LOC101927610', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99919647, end_position_on_the_genomic_accession=99925310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  220202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220202, CurrentID=0, Status='live', Symbol='ATOH7', Aliases='Math5, NCRNA, PHPVAR, RNANC, bHLHa13', description='atonal bHLH transcription factor 7', other_designations='atonal homolog 7|atonal homolog bHLH transcription factor 7|class A basic helix-loop-helix protein 13|helix-loop-helix protein hATH-5', map_location='10q21.3-q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68230595, end_position_on_the_genomic_accession=68232113, orientation='minus', exon_count=1, OMIM=609875, no_hdr0=''),
  220213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220213, CurrentID=0, Status='live', Symbol='OTUD1', Aliases='DUBA7, OTDC1', description='OTU deubiquitinase 1', other_designations='DUBA-7|OTU domain containing 1', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23439269, end_position_on_the_genomic_accession=23442381, orientation='plus', exon_count=1, OMIM=612022, no_hdr0=''),
  100422315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422315, CurrentID=0, Status='live', Symbol='LOC100422315', Aliases='', description='poly(A) binding protein, cytoplasmic 4 (inducible form) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104860455, end_position_on_the_genomic_accession=104861344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  342541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342541, CurrentID=0, Status='live', Symbol='LOC342541', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63471595, end_position_on_the_genomic_accession=63472302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374905, CurrentID=0, Status='live', Symbol='LOC105374905', Aliases='', description='uncharacterized LOC105374905', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7255405, end_position_on_the_genomic_accession=7276293, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927612, CurrentID=0, Status='live', Symbol='RNF139-AS1', Aliases='', description='RNF139 antisense RNA 1 (head to head)', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124462493, end_position_on_the_genomic_accession=124474563, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374909, CurrentID=0, Status='live', Symbol='LOC105374909', Aliases='', description='uncharacterized LOC105374909', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8102697, end_position_on_the_genomic_accession=8153333, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371219, CurrentID=0, Status='live', Symbol='LOC105371219', Aliases='', description='uncharacterized LOC105371219', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145128366, end_position_on_the_genomic_accession=145135724, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374910, CurrentID=0, Status='live', Symbol='LOC105374910', Aliases='', description='uncharacterized LOC105374910', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8158179, end_position_on_the_genomic_accession=8260628, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132031, CurrentID=0, Status='live', Symbol='DNM1P28', Aliases='DNM1DN3-1, DNM1DN3-3, DNM1DN3.1, DNM1DN3.3, DNM1DN3@', description='dynamin 1 pseudogene 28', other_designations='DNM1 pseudogene 28|DNM1DN3.1 duplicon|DNM1DN3.3 duplicon', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30098251, end_position_on_the_genomic_accession=30102101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927613, CurrentID=0, Status='live', Symbol='LOC101927613', Aliases='', description='uncharacterized LOC101927613', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=142565937, end_position_on_the_genomic_accession=142614111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  220296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220296, CurrentID=0, Status='live', Symbol='HEPACAM', Aliases='GlialCAM, MLC2A, MLC2B', description='hepatic and glial cell adhesion molecule', other_designations='protein hepaCAM', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124919193, end_position_on_the_genomic_accession=124936412, orientation='minus', exon_count=7, OMIM=611642, no_hdr0=''),
  106481976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481976, CurrentID=0, Status='live', Symbol='RN7SL567P', Aliases='', description='RNA, 7SL, cytoplasmic 567, pseudogene', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374916, CurrentID=0, Status='live', Symbol='LOC105374916', Aliases='', description='uncharacterized LOC105374916', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=9335916, end_position_on_the_genomic_accession=9344588, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374917, CurrentID=0, Status='live', Symbol='LOC105374917', Aliases='', description='uncharacterized LOC105374917', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=9697235, end_position_on_the_genomic_accession=9698141, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  220323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220323, CurrentID=0, Status='live', Symbol='OAF', Aliases='NS5ATP13TP2', description='out at first homolog', other_designations='HCV NS5A-transactivated protein 13 target protein 2|OAF homolog', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120211039, end_position_on_the_genomic_accession=120229941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132039, CurrentID=0, Status='live', Symbol='LOC100132039', Aliases='', description='spermine synthase pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76078997, end_position_on_the_genomic_accession=76079765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371221, CurrentID=0, Status='live', Symbol='LOC105371221', Aliases='', description='uncharacterized LOC105371221', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148949651, end_position_on_the_genomic_accession=148952093, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374920, CurrentID=0, Status='live', Symbol='LOC105374920', Aliases='', description='uncharacterized LOC105374920', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10325210, end_position_on_the_genomic_accession=10329501, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374921, CurrentID=0, Status='live', Symbol='LOC105374921', Aliases='', description='uncharacterized LOC105374921', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10347963, end_position_on_the_genomic_accession=10352022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  252862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252862, CurrentID=0, Status='live', Symbol='LOC252862', Aliases='', description='CDC26 subunit of anaphase promoting complex-related gene', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  220359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220359, CurrentID=0, Status='live', Symbol='TIGD3', Aliases='', description='tigger transposable element derived 3', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65354811, end_position_on_the_genomic_accession=65357613, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374924, CurrentID=0, Status='live', Symbol='LOC105374924', Aliases='', description='uncharacterized LOC105374924', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10659564, end_position_on_the_genomic_accession=10684185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371222, CurrentID=0, Status='live', Symbol='LOC105371222', Aliases='', description='uncharacterized LOC105371222', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148733395, end_position_on_the_genomic_accession=148735995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  351449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=351449, CurrentID=0, Status='live', Symbol='MRPS31P1', Aliases='', description='mitochondrial ribosomal protein S31 pseudogene 1', other_designations='', map_location='3p21.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41189030, end_position_on_the_genomic_accession=41189242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374927, CurrentID=0, Status='live', Symbol='LOC105374927', Aliases='', description='uncharacterized LOC105374927', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11391875, end_position_on_the_genomic_accession=11394860, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927616, CurrentID=0, Status='live', Symbol='LOC101927616', Aliases='', description='uncharacterized LOC101927616', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127324155, end_position_on_the_genomic_accession=127340072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  220382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220382, CurrentID=0, Status='live', Symbol='FAM181B', Aliases='', description='family with sequence similarity 181 member B', other_designations='family with sequence similarity 181, member B', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82732004, end_position_on_the_genomic_accession=82733864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100359393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359393, CurrentID=0, Status='live', Symbol='SCA30', Aliases='', description='spinocerebellar ataxia 30', other_designations='', map_location='4q34.3-q35.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613371, no_hdr0=''),
  100359394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359394, CurrentID=0, Status='live', Symbol='LINC00102', Aliases='NCRNA00102', description='long intergenic non-protein coding RNA 102', other_designations='OTTHUMT00000055623', map_location='Xp22.33 and Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2612991, end_position_on_the_genomic_accession=2615347, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100359395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359395, CurrentID=0, Status='live', Symbol='DFNA58', Aliases='', description='deafness, autosomal dominant 58', other_designations='', map_location='2p21-p12', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615654, no_hdr0=''),
  220388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220388, CurrentID=0, Status='live', Symbol='CCDC89', Aliases='', description='coiled-coil domain containing 89', other_designations='bc8 orange-interacting protein', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85683844, end_position_on_the_genomic_accession=85686277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100359397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359397, CurrentID=0, Status='live', Symbol='HSPD1P18', Aliases='HSPD1-1P', description='heat shock protein family D (Hsp60) member 1 pseudogene 18', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 18', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136381016, end_position_on_the_genomic_accession=136381511, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100359399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359399, CurrentID=0, Status='live', Symbol='DFNB85', Aliases='', description='deafness, autosomal recessive 85', other_designations='', map_location='17p12-q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613392, no_hdr0=''),
  100359400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359400, CurrentID=0, Status='live', Symbol='DFNB83', Aliases='', description='deafness, autosomal recessive 83', other_designations='', map_location='9p23-p21.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613685, no_hdr0=''),
  100359401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359401, CurrentID=0, Status='live', Symbol='MYP17', Aliases='MYP4', description='myopia 17', other_designations='', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608367, no_hdr0=''),
  100359402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359402, CurrentID=0, Status='live', Symbol='SPG41', Aliases='', description='spastic paraplegia 41 (autosomal dominant)', other_designations='', map_location='11p14.1-p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613364, no_hdr0=''),
  100359403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359403, CurrentID=0, Status='live', Symbol='PARK16', Aliases='', description='Parkinson disease 16 (susceptibility)', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613164, no_hdr0=''),
  100359406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359406, CurrentID=0, Status='live', Symbol='MYP18', Aliases='', description='myopia 18 (high grade, autosomal recessive)', other_designations='', map_location='14q22.1-q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=255500, no_hdr0=''),
  100359407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100359407, CurrentID=0, Status='live', Symbol='MYP14', Aliases='', description='myopia 14', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610320, no_hdr0=''),
  105374931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374931, CurrentID=0, Status='live', Symbol='LOC105374931', Aliases='', description='uncharacterized LOC105374931', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12395474, end_position_on_the_genomic_accession=12407661, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375588, CurrentID=0, Status='live', Symbol='LOC105375588', Aliases='', description='uncharacterized LOC105375588', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155262893, end_position_on_the_genomic_accession=155266679, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927617, CurrentID=0, Status='live', Symbol='LOC101927617', Aliases='', description='uncharacterized LOC101927617', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62545768, end_position_on_the_genomic_accession=62546423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132053, CurrentID=0, Status='live', Symbol='RPL30P8', Aliases='RPL30_4_741', description='ribosomal protein L30 pseudogene 8', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113829822, end_position_on_the_genomic_accession=113834765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  220416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220416, CurrentID=0, Status='live', Symbol='LRRC63', Aliases='', description='leucine rich repeat containing 63', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46211911, end_position_on_the_genomic_accession=46276818, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  220429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220429, CurrentID=0, Status='live', Symbol='CTAGE10P', Aliases='', description='CTAGE family member 10, pseudogene', other_designations='CTAGE family, member 10, pseudogene|CTAGE family, member 5 pseudogene', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49890409, end_position_on_the_genomic_accession=49893380, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  220433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220433, CurrentID=0, Status='live', Symbol='RPS4XP16', Aliases='RPS4P16, RPS4X_8_1313', description='ribosomal protein S4X pseudogene 16', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51460495, end_position_on_the_genomic_accession=51461380, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132057, CurrentID=0, Status='live', Symbol='LOC100132057', Aliases='', description='phosphodiesterase 4D interacting protein pseudogene', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148402483, end_position_on_the_genomic_accession=148432545, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  220441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220441, CurrentID=0, Status='live', Symbol='RNF152', Aliases='', description='ring finger protein 152', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61808067, end_position_on_the_genomic_accession=61894432, orientation='minus', exon_count=8, OMIM=616512, no_hdr0=''),
  105374938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374938, CurrentID=0, Status='live', Symbol='LOC105374938', Aliases='', description='uncharacterized LOC105374938', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13599162, end_position_on_the_genomic_accession=13611819, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374940, CurrentID=0, Status='live', Symbol='LOC105374940', Aliases='', description='uncharacterized LOC105374940', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14194449, end_position_on_the_genomic_accession=14201140, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374941, CurrentID=0, Status='live', Symbol='LOC105374941', Aliases='', description='uncharacterized LOC105374941', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14460481, end_position_on_the_genomic_accession=14463522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374942, CurrentID=0, Status='live', Symbol='LOC105374942', Aliases='', description='uncharacterized LOC105374942', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14424257, end_position_on_the_genomic_accession=14460449, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372741, CurrentID=0, Status='live', Symbol='LOC105372741', Aliases='', description='uncharacterized LOC105372741', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17743227, end_position_on_the_genomic_accession=17767158, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  692106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692106, CurrentID=0, Status='live', Symbol='SNORD65', Aliases='HBII-135', description='small nucleolar RNA, C/D box 65', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16441226, end_position_on_the_genomic_accession=16441298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374946, CurrentID=0, Status='live', Symbol='LOC105374946', Aliases='', description='uncharacterized LOC105374946', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15143785, end_position_on_the_genomic_accession=15176913, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927620, CurrentID=0, Status='live', Symbol='LOC101927620', Aliases='', description='uncharacterized LOC101927620', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53153354, end_position_on_the_genomic_accession=53157528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132069, CurrentID=0, Status='live', Symbol='LOC100132069', Aliases='', description='uncharacterized LOC100132069', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=52898658, end_position_on_the_genomic_accession=52899372, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374950, CurrentID=0, Status='live', Symbol='LOC105374950', Aliases='', description='uncharacterized LOC105374950', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16901790, end_position_on_the_genomic_accession=16953272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132071, CurrentID=0, Status='live', Symbol='LOC100132071', Aliases='', description='uncharacterized LOC100132071', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=11175336, end_position_on_the_genomic_accession=11176169, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374952, CurrentID=0, Status='live', Symbol='LOC105374952', Aliases='', description='uncharacterized LOC105374952', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17706527, end_position_on_the_genomic_accession=17711027, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374953, CurrentID=0, Status='live', Symbol='LOC105374953', Aliases='', description='uncharacterized LOC105374953', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17850434, end_position_on_the_genomic_accession=17853690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374954, CurrentID=0, Status='live', Symbol='LOC105374954', Aliases='', description='uncharacterized LOC105374954', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17754042, end_position_on_the_genomic_accession=17760051, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132076, CurrentID=0, Status='live', Symbol='LOC100132076', Aliases='', description='radixin pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27517883, end_position_on_the_genomic_accession=27519767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374957, CurrentID=0, Status='live', Symbol='LOC105374957', Aliases='', description='uncharacterized LOC105374957', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18746782, end_position_on_the_genomic_accession=18757290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374958, CurrentID=0, Status='live', Symbol='LOC105374958', Aliases='', description='uncharacterized LOC105374958', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18977113, end_position_on_the_genomic_accession=19129915, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105374959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374959, CurrentID=0, Status='live', Symbol='LOC105374959', Aliases='', description='uncharacterized LOC105374959', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19639212, end_position_on_the_genomic_accession=19691070, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372524, CurrentID=0, Status='live', Symbol='LOC105372524', Aliases='', description='uncharacterized LOC105372524', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10172431, end_position_on_the_genomic_accession=10200528, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132081, CurrentID=0, Status='live', Symbol='LOC100132081', Aliases='', description='ladinin 1 pseudogene', other_designations='', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27896512, end_position_on_the_genomic_accession=27897433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927623, CurrentID=0, Status='live', Symbol='LOC101927623', Aliases='', description='uncharacterized LOC101927623', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86127540, end_position_on_the_genomic_accession=86157086, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100187725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100187725, CurrentID=0, Status='live', Symbol='LOC100187725', Aliases='', description='5-azacytidine induced 2 pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44969814, end_position_on_the_genomic_accession=44971314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  220594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220594, CurrentID=0, Status='live', Symbol='USP32P2', Aliases='TL132', description='ubiquitin specific peptidase 32 pseudogene 2', other_designations='TL132 protein|ubiquitin specific peptidase 6 (Tre-2 oncogene) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18511262, end_position_on_the_genomic_accession=18521252, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374964, CurrentID=0, Status='live', Symbol='LOC105374964', Aliases='', description='uncharacterized LOC105374964', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20048826, end_position_on_the_genomic_accession=20055559, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124149, CurrentID=0, Status='live', Symbol='ANKRD26P1', Aliases='', description='ankyrin repeat domain 26 pseudogene 1', other_designations='', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46469337, end_position_on_the_genomic_accession=46569097, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105374270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374270, CurrentID=0, Status='live', Symbol='LOC105374270', Aliases='', description='uncharacterized LOC105374270', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189461881, end_position_on_the_genomic_accession=189478191, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132086, CurrentID=0, Status='live', Symbol='AK6P1', Aliases='', description='adenylate kinase 6 pseudogene 1', other_designations='adenylate kinase isoenzyme 6 pseudogene', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34230379, end_position_on_the_genomic_accession=34250272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927624, CurrentID=0, Status='live', Symbol='LINC01150', Aliases='2G7, TCONS_00019134', description='long intergenic non-protein coding RNA 1150', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1896759, end_position_on_the_genomic_accession=1908359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374968, CurrentID=0, Status='live', Symbol='LOC105374968', Aliases='', description='uncharacterized LOC105374968', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20597732, end_position_on_the_genomic_accession=20600617, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374969, CurrentID=0, Status='live', Symbol='LOC105374969', Aliases='', description='uncharacterized LOC105374969', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22147034, end_position_on_the_genomic_accession=22149137, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132090, CurrentID=0, Status='live', Symbol='LOC100132090', Aliases='', description='survival of motor neuron 1, telomeric pseudogene', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183028600, end_position_on_the_genomic_accession=183029071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132093, CurrentID=0, Status='live', Symbol='LOC100132093', Aliases='', description='uncharacterized LOC100132093', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=1812776, end_position_on_the_genomic_accession=1814911, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374975, CurrentID=0, Status='live', Symbol='LOC105374975', Aliases='', description='uncharacterized LOC105374975', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23415990, end_position_on_the_genomic_accession=23455137, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374979, CurrentID=0, Status='live', Symbol='LOC105374979', Aliases='', description='uncharacterized LOC105374979', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24792909, end_position_on_the_genomic_accession=24796298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420789, CurrentID=0, Status='live', Symbol='LOC100420789', Aliases='', description='chromosome 2 open reading frame 27A pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66253455, end_position_on_the_genomic_accession=66254040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374980, CurrentID=0, Status='live', Symbol='LOC105374980', Aliases='', description='uncharacterized LOC105374980', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24796743, end_position_on_the_genomic_accession=24797435, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  351773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=351773, CurrentID=0, Status='live', Symbol='MRPL51P1', Aliases='', description='mitochondrial ribosomal protein L51 pseudogene 1', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27422933, end_position_on_the_genomic_accession=27423195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374981, CurrentID=0, Status='live', Symbol='LOC105374981', Aliases='', description='uncharacterized LOC105374981', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24933921, end_position_on_the_genomic_accession=24950491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132103, CurrentID=0, Status='live', Symbol='FAM66E', Aliases='', description='family with sequence similarity 66 member E', other_designations='family with sequence similarity 66, member E', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7955013, end_position_on_the_genomic_accession=8008755, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  220717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220717, CurrentID=0, Status='live', Symbol='RPLP0P6', Aliases='RPLP0_2_209', description='ribosomal protein, large, P0 pseudogene 6', other_designations='', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38481775, end_position_on_the_genomic_accession=38482875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374984, CurrentID=0, Status='live', Symbol='LOC105374984', Aliases='', description='uncharacterized LOC105374984', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26226480, end_position_on_the_genomic_accession=26227473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  220729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220729, CurrentID=0, Status='live', Symbol='LOC220729', Aliases='', description='succinate dehydrogenase complex subunit A, flavoprotein (Fp) pseudogene', other_designations='succinate dehydrogenase complex, subunit A, flavoprotein (Fp) pseudogene', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197614027, end_position_on_the_genomic_accession=197627881, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372525, CurrentID=0, Status='live', Symbol='LOC105372525', Aliases='', description='uncharacterized LOC105372525', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10843776, end_position_on_the_genomic_accession=10852803, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132108, CurrentID=0, Status='live', Symbol='LOC100132108', Aliases='', description='proteasome 26S subunit, non-ATPase 7 pseudogene', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 pseudogene', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155845340, end_position_on_the_genomic_accession=155847270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374991, CurrentID=0, Status='live', Symbol='LOC105374991', Aliases='', description='uncharacterized LOC105374991', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26698075, end_position_on_the_genomic_accession=26722150, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723539, CurrentID=0, Status='live', Symbol='LOC102723539', Aliases='', description='uncharacterized LOC102723539', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34154778, end_position_on_the_genomic_accession=34159974, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371236, CurrentID=0, Status='live', Symbol='LOC105371236', Aliases='', description='uncharacterized LOC105371236', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47744170, end_position_on_the_genomic_accession=47753802, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  220786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220786, CurrentID=0, Status='live', Symbol='CDK7PS', Aliases='', description='cyclin-dependent kinase 7 pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157210336, end_position_on_the_genomic_accession=157211525, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374995, CurrentID=0, Status='live', Symbol='LOC105374995', Aliases='', description='uncharacterized LOC105374995', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27125897, end_position_on_the_genomic_accession=27130449, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  255252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255252, CurrentID=0, Status='live', Symbol='LRRC57', Aliases='', description='leucine rich repeat containing 57', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42528444, end_position_on_the_genomic_accession=42548804, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375000, CurrentID=0, Status='live', Symbol='LOC105375000', Aliases='', description='uncharacterized LOC105375000', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27791721, end_position_on_the_genomic_accession=27796867, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375001, CurrentID=0, Status='live', Symbol='LOC105375001', Aliases='', description='uncharacterized LOC105375001', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27473191, end_position_on_the_genomic_accession=27480803, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375002, CurrentID=0, Status='live', Symbol='LOC105375002', Aliases='', description='uncharacterized LOC105375002', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29022007, end_position_on_the_genomic_accession=29035914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  89758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89758, CurrentID=0, Status='live', Symbol='LINC01621', Aliases='C6orf7', description='long intergenic non-protein coding RNA 1621', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79803582, end_position_on_the_genomic_accession=79878494, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  89760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89760, CurrentID=0, Status='live', Symbol='MRX75', Aliases='', description='mental retardation, X-linked 75', other_designations='', map_location='Xq24-q26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100884129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100884129, CurrentID=0, Status='live', Symbol='DUP17Q12', Aliases='C17DUPq12', description='Chromosome 17q12 duplication syndrome', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614526, no_hdr0=''),
  100884130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100884130, CurrentID=0, Status='live', Symbol='DEL17Q12', Aliases='C17DELq12', description='Chromosome 17q12 deletion syndrome', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614527, no_hdr0=''),
  89765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89765, CurrentID=0, Status='live', Symbol='RSPH1', Aliases='CT79, RSP440A, TSA2, TSGA2, RSPH1', description='radial spoke head 1 homolog (Chlamydomonas)', other_designations='cancer/testis antigen 79|male meiotic metaphase chromosome-associated acidic protein|meichroacidin|testes specific gene A2 homolog|testis specific A2 homolog', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42472486, end_position_on_the_genomic_accession=42496354, orientation='minus', exon_count=9, OMIM=609314, no_hdr0=''),
  89766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89766, CurrentID=0, Status='live', Symbol='UMODL1', Aliases='', description='uromodulin like 1', other_designations='olfactorin', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42031463, end_position_on_the_genomic_accession=42142996, orientation='plus', exon_count=32, OMIM=613859, no_hdr0=''),
  105375004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375004, CurrentID=0, Status='live', Symbol='LOC105375004', Aliases='', description='uncharacterized LOC105375004', other_designations='XXbac-BCX196D17.5', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29123980, end_position_on_the_genomic_accession=29128974, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  89770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89770, CurrentID=0, Status='live', Symbol='IGHV3-30-5', Aliases='IGHV3-3, IGHV3305', description='immunoglobulin heavy variable 3-30-5', other_designations='', map_location='14q32.33 not on reference assembly', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371238, CurrentID=0, Status='live', Symbol='LOC105371238', Aliases='', description='uncharacterized LOC105371238', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47965570, end_position_on_the_genomic_accession=47971685, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  89777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89777, CurrentID=0, Status='live', Symbol='SERPINB12', Aliases='YUKOPIN', description='serpin peptidase inhibitor, clade B (ovalbumin), member 12', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 12', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63519090, end_position_on_the_genomic_accession=63569329, orientation='plus', exon_count=12, OMIM=615662, no_hdr0=''),
  89778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89778, CurrentID=0, Status='live', Symbol='SERPINB11', Aliases='EPIPIN, SERPIN11', description='serpin peptidase inhibitor, clade B (ovalbumin), member 11 (gene/pseudogene)', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 11', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63702304, end_position_on_the_genomic_accession=63723893, orientation='plus', exon_count=10, OMIM=615682, no_hdr0=''),
  89780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89780, CurrentID=0, Status='live', Symbol='WNT3A', Aliases='', description='wingless-type MMTV integration site family member 3A', other_designations='wingless-type MMTV integration site family, member 3A', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228007022, end_position_on_the_genomic_accession=228066711, orientation='plus', exon_count=6, OMIM=606359, no_hdr0=''),
  89781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89781, CurrentID=0, Status='live', Symbol='HPS4', Aliases='LE', description='Hermansky-Pudlak syndrome 4', other_designations='light-ear protein homolog', map_location='22cen-q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26443613, end_position_on_the_genomic_accession=26483863, orientation='minus', exon_count=20, OMIM=606682, no_hdr0=''),
  89782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89782, CurrentID=0, Status='live', Symbol='LMLN', Aliases='GP63, INV, IX14, MSP', description='leishmanolysin-like (metallopeptidase M8 family)', other_designations='invadolysin|leishmanolysin-2', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197960200, end_position_on_the_genomic_accession=198043720, orientation='plus', exon_count=18, OMIM=609380, no_hdr0=''),
  653303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653303, CurrentID=0, Status='live', Symbol='LOC653303', Aliases='', description='proprotein convertase subtilisin/kexin type 7 pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117135529, end_position_on_the_genomic_accession=117138858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100884155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100884155, CurrentID=0, Status='live', Symbol='MTND2P9', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 9', other_designations='MT-ND2 pseudogene 9', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80565035, end_position_on_the_genomic_accession=80565966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  89790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89790, CurrentID=0, Status='live', Symbol='SIGLEC10', Aliases='PRO940, SIGLEC-10, SLG2', description='sialic acid binding Ig-like lectin 10', other_designations='sialic acid binding Ig-like lectin 10 Ig-like lectin 7|siglec-like gene 2|siglec-like protein 2', map_location='19q13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51410021, end_position_on_the_genomic_accession=51417803, orientation='minus', exon_count=11, OMIM=606091, no_hdr0=''),
  89792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89792, CurrentID=0, Status='live', Symbol='GAL3ST3', Aliases='GAL3ST-3, GAL3ST2', description='galactose-3-O-sulfotransferase 3', other_designations="beta-galactose-3-O-sulfotransferase 3|gal-beta-1, 3-GalNAc 3'-sulfotransferase 3|galactose 3'-sulfotransferase|galbeta1-3GalNAc 3'-sulfotransferase 3", map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66040765, end_position_on_the_genomic_accession=66049180, orientation='minus', exon_count=3, OMIM=608234, no_hdr0=''),
  89795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89795, CurrentID=0, Status='live', Symbol='NAV3', Aliases='POMFIL1, STEERIN3, unc53H3', description='neuron navigator 3', other_designations='pore membrane and/or filament interacting like protein 1|unc-53 homolog 3', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77571862, end_position_on_the_genomic_accession=78213012, orientation='plus', exon_count=42, OMIM=611629, no_hdr0=''),
  89796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89796, CurrentID=0, Status='live', Symbol='NAV1', Aliases='POMFIL3, STEERIN1, UNC53H1', description='neuron navigator 1', other_designations='neuron navigator-1|pore membrane and/or filament interacting like protein 3|pore membrane and/or filament-interacting-like protein 3|unc-53 homolog 1', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201539102, end_position_on_the_genomic_accession=201826974, orientation='plus', exon_count=36, OMIM=611628, no_hdr0=''),
  220869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220869, CurrentID=0, Status='live', Symbol='CBWD5', Aliases='DC36', description='COBW domain containing 5', other_designations='COBW-like placental protein|cobalamin synthase W domain-containing protein 5|cobalamin synthetase W domain-containing protein 5|dopamine responsive protein', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65675834, end_position_on_the_genomic_accession=65734041, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  89801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89801, CurrentID=0, Status='live', Symbol='PPP1R3F', Aliases='HB2E, R3F', description='protein phosphatase 1 regulatory subunit 3F', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 3F|protein phosphatase 1, regulatory subunit 3F', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49269829, end_position_on_the_genomic_accession=49288095, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372526, CurrentID=0, Status='live', Symbol='LOC105372526', Aliases='', description='uncharacterized LOC105372526', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10659560, end_position_on_the_genomic_accession=10663956, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  220885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220885, CurrentID=0, Status='live', Symbol='RPSAP15', Aliases='LAMR1P15, LAMRL5, RPSA_25_1791', description='ribosomal protein SA pseudogene 15', other_designations='laminin receptor 1 pseudogene 15|laminin receptor-like protein LAMRL5 pseudogene', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=87703273, end_position_on_the_genomic_accession=87704306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375012, CurrentID=0, Status='live', Symbol='LOC105375012', Aliases='', description='uncharacterized LOC105375012', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30460914, end_position_on_the_genomic_accession=30463663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  89822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89822, CurrentID=0, Status='live', Symbol='KCNK17', Aliases='K2p17.1, TALK-2, TALK2, TASK-4, TASK4', description='potassium channel, two pore domain subfamily K, member 17', other_designations='2P domain potassium channel Talk-2|TWIK-related acid-sensitive K(+) channel 4|TWIK-related alkaline pH-activated K(+) channel 2|acid-sensitive potassium channel protein TASK-4', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39299001, end_position_on_the_genomic_accession=39314461, orientation='minus', exon_count=7, OMIM=607370, no_hdr0=''),
  728815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728815, CurrentID=0, Status='live', Symbol='LOC728815', Aliases='', description='cytokine receptor-like factor 3 pseudogene', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=103494241, end_position_on_the_genomic_accession=103496367, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  89832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89832, CurrentID=0, Status='live', Symbol='CHRFAM7A', Aliases='CHRNA7, CHRNA7-DR1, D-10', description='CHRNA7 (cholinergic receptor, nicotinic, alpha 7, exons 5-10) and FAM7A (family with sequence similarity 7A, exons A-E) fusion', other_designations='CHRNA7 (cholinergic receptor, nicotinic, alpha polypeptide 7, exons 5-10) and FAM7A (family with sequence similarity 7A, exons A-E) fusion|alpha 7 neuronal nicotinic acetylcholine receptor-FAM7A hybrid|alpha-7 nicotinic cholinergic receptor subunit', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30357770, end_position_on_the_genomic_accession=30393849, orientation='minus', exon_count=12, OMIM=609756, no_hdr0=''),
  220906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220906, CurrentID=0, Status='live', Symbol='WAC-AS1', Aliases='', description='WAC antisense RNA 1 (head to head)', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28519917, end_position_on_the_genomic_accession=28532354, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375015, CurrentID=0, Status='live', Symbol='LOC105375015', Aliases='', description='uncharacterized LOC105375015', other_designations='XXbac-BPG248L24.13', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31296818, end_position_on_the_genomic_accession=31308579, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  89837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89837, CurrentID=0, Status='live', Symbol='ULK4P3', Aliases='D-X, FAM7A3', description='ULK4 pseudogene 3', other_designations='family with sequence similarity 7, member A3|unc-51-like kinase 4 (ULK4) pseudogene|unc-51-like kinase 4 pseudogene 3', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30103732, end_position_on_the_genomic_accession=30131748, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  89838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89838, CurrentID=0, Status='live', Symbol='ULK4P1', Aliases='D-X, FAM7A1', description='ULK4 pseudogene 1', other_designations='family with sequence similarity 7, member A1 (non-protein coding)|unc-51-like kinase 4 (ULK4) pseudogene|unc-51-like kinase 4 pseudogene 1', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32406611, end_position_on_the_genomic_accession=32435049, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  89839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89839, CurrentID=0, Status='live', Symbol='ARHGAP11B', Aliases="B'-T, FAM7B1", description='Rho GTPase activating protein 11B', other_designations='GAP (1-8)|family with sequence similarity 7, member B1|rho-type GTPase-activating protein 11B', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30626131, end_position_on_the_genomic_accession=30638810, orientation='plus', exon_count=8, OMIM=616310, no_hdr0=''),
  105374272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374272, CurrentID=0, Status='live', Symbol='LOC105374272', Aliases='', description='uncharacterized LOC105374272', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190696278, end_position_on_the_genomic_accession=190699253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  89844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89844, CurrentID=0, Status='live', Symbol='LOC89844', Aliases='', description='mitochondrial RNA polymerase pseudogene', other_designations='polymerase (RNA) mitochondrial (DNA directed) pseudogene', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15902272, end_position_on_the_genomic_accession=15903239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  89845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89845, CurrentID=0, Status='live', Symbol='ABCC10', Aliases='EST182763, MRP7, SIMRP7', description='ATP binding cassette subfamily C member 10', other_designations='ATP-binding cassette sub-family C member 10|ATP-binding cassette, sub-family C (CFTR/MRP), member 10', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43427546, end_position_on_the_genomic_accession=43451980, orientation='plus', exon_count=26, OMIM=612509, no_hdr0=''),
  89846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89846, CurrentID=0, Status='live', Symbol='FGD3', Aliases='ZFYVE5', description='FYVE, RhoGEF and PH domain containing 3', other_designations='FGD1 family, member 3|faciogenital dysplasia 3|zinc finger FYVE domain-containing protein 5', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92947319, end_position_on_the_genomic_accession=93036236, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  89848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89848, CurrentID=0, Status='live', Symbol='FCHSD1', Aliases='NWK2', description='FCH and double SH3 domains 1', other_designations='FCH and double SH3 domains protein 1|nervous wreck homolog 2', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141639302, end_position_on_the_genomic_accession=141651428, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  89849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89849, CurrentID=0, Status='live', Symbol='ATG16L2', Aliases='ATG16B, WDR80', description='autophagy related 16 like 2', other_designations='APG16-like 2|ATG16 autophagy related 16-like 2|WD repeat-containing protein 80|autophagy related 16-like 2', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72814406, end_position_on_the_genomic_accession=72842705, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  89853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89853, CurrentID=0, Status='live', Symbol='MVB12B', Aliases='C9orf28, FAM125B', description='multivesicular body subunit 12B', other_designations='ESCRT-I complex subunit MVB12B|family with sequence similarity 125, member B', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126326844, end_position_on_the_genomic_accession=126507041, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  89857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89857, CurrentID=0, Status='live', Symbol='KLHL6', Aliases='', description='kelch like family member 6', other_designations='kelch-like 6|kelch-like family member 6|kelch-like protein KLHL6', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183487531, end_position_on_the_genomic_accession=183555712, orientation='minus', exon_count=5, OMIM=614214, no_hdr0=''),
  220930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220930, CurrentID=0, Status='live', Symbol='ZEB1-AS1', Aliases='', description='ZEB1 antisense RNA 1', other_designations='ZEB1 antisense RNA 1 (non-protein coding)', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31316528, end_position_on_the_genomic_accession=31319095, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129452, CurrentID=0, Status='live', Symbol='GAMTP2', Aliases='', description='guanidinoacetate N-methyltransferase pseudogene 2', other_designations='', map_location='13q13.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34574202, end_position_on_the_genomic_accession=34574749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371241, CurrentID=0, Status='live', Symbol='LOC105371241', Aliases='', description='uncharacterized LOC105371241', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49075573, end_position_on_the_genomic_accession=49122735, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  89866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89866, CurrentID=0, Status='live', Symbol='SEC16B', Aliases='LZTR2, PGPR-p117, RGPR, SEC16S', description='SEC16 homolog B, endoplasmic reticulum export factor', other_designations='RGPR-p117|SEC16 homolog B|leucine zipper transcription regulator 2|protein SEC16 homolog B|regucalcin gene promoter region-related protein p117|regucalcin gene promotor region related protein', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177929107, end_position_on_the_genomic_accession=177969915, orientation='minus', exon_count=26, OMIM=612855, no_hdr0=''),
  89869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89869, CurrentID=0, Status='live', Symbol='PLCZ1', Aliases='NYD-SP27, PLC-zeta-1, PLCzeta', description='phospholipase C zeta 1', other_designations='PI-phospholipase C zeta 1|PLC-zeta-1|phosphoinositide phospholipase C-zeta-1|phospholipase C, zeta 1|testis-development related NYD-SP27', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18641990, end_position_on_the_genomic_accession=18738059, orientation='minus', exon_count=22, OMIM=608075, no_hdr0=''),
  89870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89870, CurrentID=0, Status='live', Symbol='TRIM15', Aliases='RNF93, ZNF178, ZNFB7', description='tripartite motif containing 15', other_designations='RING finger protein 93|tripartite motif-containing 15|zinc finger protein 178|zinc finger protein B7', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30163206, end_position_on_the_genomic_accession=30172696, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  89872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89872, CurrentID=0, Status='live', Symbol='AQP10', Aliases='AQPA_HUMAN', description='aquaporin 10', other_designations='AQP-10|aquaglyceroporin-10|small intestine aquaporin', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154321067, end_position_on_the_genomic_accession=154325325, orientation='plus', exon_count=6, OMIM=606578, no_hdr0=''),
  89873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89873, CurrentID=0, Status='live', Symbol='PPH2', Aliases='', description='primary pulmonary hypertension 2', other_designations='', map_location='2q31-q32', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  89874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89874, CurrentID=0, Status='live', Symbol='SLC25A21', Aliases='ODC, ODC1', description='solute carrier family 25 (mitochondrial oxoadipate carrier), member 21', other_designations='oxodicarboxylate carrier|solute carrier family 25 (mitochondrial oxodicarboxylate carrier), member 21|solute carrier family 25 member 21', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36677318, end_position_on_the_genomic_accession=37172660, orientation='minus', exon_count=13, OMIM=607571, no_hdr0=''),
  89876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89876, CurrentID=0, Status='live', Symbol='MAATS1', Aliases='AAT1, AAT1alpha, C3orf15, CaM-IP2, SPATA26', description='MYCBP-associated, testis expressed 1', other_designations='AAT-1|AAT1-alpha|AMY-1-associating protein expressed in testis 1|MYCBP-binding protein|MYCBP/AMY-1-associated testis-expressed protein 1|MYCBP/AMY-1-associated, testis expressed 1|spermatogenesis associated 26', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119703022, end_position_on_the_genomic_accession=119767102, orientation='plus', exon_count=19, OMIM=609910, no_hdr0=''),
  105375022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375022, CurrentID=0, Status='live', Symbol='LOC105375022', Aliases='', description='uncharacterized LOC105375022', other_designations='XXbac-BPG157A10.21', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33246075, end_position_on_the_genomic_accession=33247086, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  89882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89882, CurrentID=0, Status='live', Symbol='TPD52L3', Aliases='NYDSP25, hD55', description='tumor protein D52-like 3', other_designations='protein kinase NYD-SP25|testis development protein NYD-SP25', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6328349, end_position_on_the_genomic_accession=6331900, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  89883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89883, CurrentID=0, Status='live', Symbol='OR6W1P', Aliases='OR6W1, sdolf', description='olfactory receptor family 6 subfamily W member 1 pseudogene', other_designations='olfactory receptor sdolf|olfactory receptor, family 6, subfamily W, member 1 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143062284, end_position_on_the_genomic_accession=143063785, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  89884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89884, CurrentID=0, Status='live', Symbol='LHX4', Aliases='CPHD4', description='LIM homeobox 4', other_designations='LIM homeobox protein 4', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180228803, end_position_on_the_genomic_accession=180275053, orientation='plus', exon_count=10, OMIM=602146, no_hdr0=''),
  89885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89885, CurrentID=0, Status='live', Symbol='FATE1', Aliases='CT43, FATE', description='fetal and adult testis expressed 1', other_designations='BJ-HCC-2 antigen|cancer/testis antigen 43|fetal and adult testis expressed transcript protein|tumor antigen BJ-HCC-2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151716036, end_position_on_the_genomic_accession=151723194, orientation='plus', exon_count=5, OMIM=300450, no_hdr0=''),
  89886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89886, CurrentID=0, Status='live', Symbol='SLAMF9', Aliases='CD2F-10, CD2F10, CD84-H1, CD84H1, SF2001', description='SLAM family member 9', other_designations='CD2 family member 10|CD84 homolog 1|cluster of differentiation 2 antigen family member 10|cluster of differentiation 84 homolog 1|signaling lymphocytic activation molecule family member 2001|signaling lymphocytic activation molecule family member 9', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159951492, end_position_on_the_genomic_accession=159954254, orientation='minus', exon_count=4, OMIM=608589, no_hdr0=''),
  89887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89887, CurrentID=0, Status='live', Symbol='ZNF628', Aliases='ZEC, Zfp628', description='zinc finger protein 628', other_designations='zinc finger protein Zec', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55476332, end_position_on_the_genomic_accession=55484487, orientation='plus', exon_count=3, OMIM=610671, no_hdr0=''),
  89890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89890, CurrentID=0, Status='live', Symbol='KBTBD6', Aliases='', description='kelch repeat and BTB (POZ) domain containing 6', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41127573, end_position_on_the_genomic_accession=41132800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  89891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89891, CurrentID=0, Status='live', Symbol='WDR34', Aliases='DIC5, FAP133, SRTD11, bA216B9.3', description='WD repeat domain 34', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128633661, end_position_on_the_genomic_accession=128656850, orientation='minus', exon_count=9, OMIM=613363, no_hdr0=''),
  220965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220965, CurrentID=0, Status='live', Symbol='FAM13C', Aliases='FAM13C1', description='family with sequence similarity 13 member C', other_designations='family with sequence similarity 13, member C|family with sequence similarity 13, member C1', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59246129, end_position_on_the_genomic_accession=59363181, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  89894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89894, CurrentID=0, Status='live', Symbol='TMEM116', Aliases='', description='transmembrane protein 116', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111894930, end_position_on_the_genomic_accession=112013219, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  220972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220972, CurrentID=0, Status='live', Symbol='MARCH8', Aliases='CMIR, MARCH-VIII, MIR, RNF178, c-MIR', description='membrane associated ring-CH-type finger 8', other_designations='RING finger protein 178|c-mir, cellular modulator of immune recognition|cellular modulator of immune recognition (c-MIR)|membrane associated ring finger 8|membrane-associated RING finger protein 8|membrane-associated RING-CH protein VIII|membrane-associated ring finger (C3HC4) 8, E3 ubiquitin protein ligase', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45454585, end_position_on_the_genomic_accession=45594906, orientation='minus', exon_count=15, OMIM=613335, no_hdr0=''),
  105375026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375026, CurrentID=0, Status='live', Symbol='LOC105375026', Aliases='', description='uncharacterized LOC105375026', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33893107, end_position_on_the_genomic_accession=33914794, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  220974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220974, CurrentID=0, Status='live', Symbol='OR6D1P', Aliases='OR6D2P', description='olfactory receptor family 6 subfamily D member 1 pseudogene', other_designations='olfactory receptor, family 6, subfamily D, member 1 pseudogene|olfactory receptor, family 6, subfamily D, member 2 pseudogene', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45258309, end_position_on_the_genomic_accession=45259121, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  220979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220979, CurrentID=0, Status='live', Symbol='C10orf25', Aliases='', description='chromosome 10 open reading frame 25', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44993932, end_position_on_the_genomic_accession=45000910, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  220980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220980, CurrentID=0, Status='live', Symbol='TMEM72-AS1', Aliases='', description='TMEM72 antisense RNA 1', other_designations='TMEM72 antisense RNA 1 (non-protein coding)', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44811024, end_position_on_the_genomic_accession=44959689, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  89910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89910, CurrentID=0, Status='live', Symbol='UBE3B', Aliases='BPIDS, KOS', description='ubiquitin protein ligase E3B', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109477410, end_position_on_the_genomic_accession=109543628, orientation='plus', exon_count=33, OMIM=608047, no_hdr0=''),
  220988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220988, CurrentID=0, Status='live', Symbol='HNRNPA3', Aliases='2610510D13Rik, D10S102, FBRNP, HNRPA3', description='heterogeneous nuclear ribonucleoprotein A3', other_designations='hnRNP A3', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177212314, end_position_on_the_genomic_accession=177223959, orientation='plus', exon_count=12, OMIM=605372, no_hdr0=''),
  220992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=220992, CurrentID=0, Status='live', Symbol='ZNF485', Aliases='', description='zinc finger protein 485', other_designations='Zinc finger protein 93 (Zinc finger protein HTF34)', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43606407, end_position_on_the_genomic_accession=43617904, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375030, CurrentID=0, Status='live', Symbol='LOC105375030', Aliases='', description='uncharacterized LOC105375030', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34880063, end_position_on_the_genomic_accession=34885341, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  89927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89927, CurrentID=0, Status='live', Symbol='C16orf45', Aliases='', description='chromosome 16 open reading frame 45', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15434468, end_position_on_the_genomic_accession=15594689, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  221002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221002, CurrentID=0, Status='live', Symbol='RASGEF1A', Aliases='CG4853', description='RasGEF domain family member 1A', other_designations='CG4853 gene product|RasGEF domain family, member 1A', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43194533, end_position_on_the_genomic_accession=43266919, orientation='minus', exon_count=16, OMIM=614531, no_hdr0=''),
  89932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89932, CurrentID=0, Status='live', Symbol='PAPLN', Aliases='', description='papilin, proteoglycan-like sulfated glycoprotein', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73235843, end_position_on_the_genomic_accession=73274640, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  89941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89941, CurrentID=0, Status='live', Symbol='RHOT2', Aliases='ARHT2, C16orf39, MIRO-2, MIRO2, RASL', description='ras homolog family member T2', other_designations='mitochondrial Rho (MIRO) GTPase 2|ras homolog gene family, member T2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=668083, end_position_on_the_genomic_accession=674174, orientation='plus', exon_count=19, OMIM=613889, no_hdr0=''),
  105375033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375033, CurrentID=0, Status='live', Symbol='LOC105375033', Aliases='', description='uncharacterized LOC105375033', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35600388, end_position_on_the_genomic_accession=35618431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  89944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89944, CurrentID=0, Status='live', Symbol='GLB1L2', Aliases='MST114, MSTP114', description='galactosidase beta 1 like 2', other_designations='galactosidase beta 1-like 2|galactosidase, beta 1-like 2', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134331874, end_position_on_the_genomic_accession=134376328, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  692196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692196, CurrentID=0, Status='live', Symbol='SNORD76', Aliases='U76', description='small nucleolar RNA, C/D box 76', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173866635, end_position_on_the_genomic_accession=173866715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375034, CurrentID=0, Status='live', Symbol='LOC105375034', Aliases='', description='uncharacterized LOC105375034', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35997884, end_position_on_the_genomic_accession=36012225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  89953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89953, CurrentID=0, Status='live', Symbol='KLC4', Aliases='KNSL8, bA387M24.3', description='kinesin light chain 4', other_designations='kinesin-like protein 8', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43059594, end_position_on_the_genomic_accession=43075099, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  89958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89958, CurrentID=0, Status='live', Symbol='SAPCD2', Aliases='C9orf140, p42.3', description='suppressor APC domain containing 2', other_designations='2010317E24Rik|TS/MDEP|protein C9orf140|tumor specificity and mitosis phase-dependent expression protein', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137062124, end_position_on_the_genomic_accession=137070576, orientation='minus', exon_count=7, OMIM=612057, no_hdr0=''),
  105375036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375036, CurrentID=0, Status='live', Symbol='LOC105375036', Aliases='', description='uncharacterized LOC105375036', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36241391, end_position_on_the_genomic_accession=36248340, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  221035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221035, CurrentID=0, Status='live', Symbol='REEP3', Aliases='C10orf74', description='receptor accessory protein 3', other_designations='receptor expression enhancing protein 3', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63521353, end_position_on_the_genomic_accession=63625123, orientation='plus', exon_count=9, OMIM=609348, no_hdr0=''),
  221037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221037, CurrentID=0, Status='live', Symbol='JMJD1C', Aliases='TRIP-8, TRIP8', description='jumonji domain containing 1C', other_designations='TR-interacting protein 8|thyroid hormone receptor interactor 8|thyroid receptor-interacting protein 8', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63167221, end_position_on_the_genomic_accession=63466001, orientation='minus', exon_count=28, OMIM=604503, no_hdr0=''),
  105375037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375037, CurrentID=0, Status='live', Symbol='LOC105375037', Aliases='', description='uncharacterized LOC105375037', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36386395, end_position_on_the_genomic_accession=36391823, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  89970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89970, CurrentID=0, Status='live', Symbol='RSPRY1', Aliases='SEMDFA', description='ring finger and SPRY domain containing 1', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57186291, end_position_on_the_genomic_accession=57240473, orientation='plus', exon_count=17, OMIM=616585, no_hdr0=''),
  221044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221044, CurrentID=0, Status='live', Symbol='UCMA', Aliases='C10orf49, GRP, GRP/UCMA', description='upper zone of growth plate and cartilage matrix associated', other_designations='Gla-rich protein', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13221766, end_position_on_the_genomic_accession=13234328, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375038, CurrentID=0, Status='live', Symbol='LOC105375038', Aliases='', description='uncharacterized LOC105375038', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36639111, end_position_on_the_genomic_accession=36642010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928943, CurrentID=0, Status='live', Symbol='LOC101928943', Aliases='', description='uncharacterized LOC101928943', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47270657, end_position_on_the_genomic_accession=47272110, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  89978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=89978, CurrentID=0, Status='live', Symbol='DPH6', Aliases='ATPBD4', description='diphthamine biosynthesis 6', other_designations='ATP binding domain 4|ATP-binding domain-containing protein 4|DPH6 homolog|diphthamide synthase|diphthamide synthetase|protein DPH6 homolog', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35304893, end_position_on_the_genomic_accession=35546222, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105375039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375039, CurrentID=0, Status='live', Symbol='LOC105375039', Aliases='', description='uncharacterized LOC105375039', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36678946, end_position_on_the_genomic_accession=36681307, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221060, CurrentID=0, Status='live', Symbol='C10orf111', Aliases='bA455B2.4', description='chromosome 10 open reading frame 111', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15095385, end_position_on_the_genomic_accession=15097319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  221061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221061, CurrentID=0, Status='live', Symbol='FAM171A1', Aliases='C10orf38', description='family with sequence similarity 171 member A1', other_designations='astroprincin|family with sequence similarity 171, member A1', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15211643, end_position_on_the_genomic_accession=15371059, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105374273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374273, CurrentID=0, Status='live', Symbol='LOC105374273', Aliases='', description='uncharacterized LOC105374273', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190782315, end_position_on_the_genomic_accession=190787221, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81558, CurrentID=0, Status='live', Symbol='FAM117A', Aliases='', description='family with sequence similarity 117 member A', other_designations='C/EBP induced protein|C/EBP-induced protein|family with sequence similarity 117, member A', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49710325, end_position_on_the_genomic_accession=49764156, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  221074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221074, CurrentID=0, Status='live', Symbol='SLC39A12', Aliases='LZT-Hs8, ZIP-12, bA570F3.1', description='solute carrier family 39 (zinc transporter), member 12', other_designations='LIV-1 subfamily of ZIP zinc transporter 8|solute carrier family 39 (metal ion transporter), member 12|solute carrier family 39 member 12|zrt- and Irt-like protein 12', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17951839, end_position_on_the_genomic_accession=18043292, orientation='plus', exon_count=13, OMIM=608734, no_hdr0=''),
  105375043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375043, CurrentID=0, Status='live', Symbol='LOC105375043', Aliases='', description='uncharacterized LOC105375043', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38133756, end_position_on_the_genomic_accession=38139938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221078, CurrentID=0, Status='live', Symbol='NSUN6', Aliases='4933414E04Rik, ARL5B-AS1, NOPD1', description='NOP2/Sun RNA methyltransferase family member 6', other_designations='ARL5B antisense RNA 1|NOL1/NOP2/Sun and PUA domain-containing protein 1|NOL1/NOP2/Sun domain family, member 6|NOP2/Sun domain family, member 6|nucleolar protein (NOL1/NOP2/sun) and PUA domains 1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18545207, end_position_on_the_genomic_accession=18659309, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  221079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221079, CurrentID=0, Status='live', Symbol='ARL5B', Aliases='ARL8', description='ADP ribosylation factor like GTPase 5B', other_designations='ADP-ribosylation factor like GTPase 5B|ADP-ribosylation factor-like 5B|ADP-ribosylation factor-like 8|ADP-ribosylation factor-like protein 8|ADP-ribosylation-like factor 8', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18659335, end_position_on_the_genomic_accession=18681639, orientation='plus', exon_count=7, OMIM=608909, no_hdr0=''),
  100129457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129457, CurrentID=0, Status='live', Symbol='LOC100129457', Aliases='', description='coiled-coil domain containing 86 pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178840097, end_position_on_the_genomic_accession=178841136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  90011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90011, CurrentID=0, Status='live', Symbol='KIR3DX1', Aliases='KIR3DL0, LENG12', description='killer cell immunoglobulin-like receptor, three domains, X1', other_designations='immunoglobulin-like receptor KIR3DL0|leukocyte receptor cluster (LRC) member 12', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54532692, end_position_on_the_genomic_accession=54545768, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  221091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221091, CurrentID=0, Status='live', Symbol='LRRN4CL', Aliases='', description='LRRN4 C-terminal like', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62686402, end_position_on_the_genomic_accession=62689899, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  221092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221092, CurrentID=0, Status='live', Symbol='HNRNPUL2', Aliases='HNRPUL2, SAF-A2', description='heterogeneous nuclear ribonucleoprotein U-like 2', other_designations='scaffold-attachment factor A2', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62712625, end_position_on_the_genomic_accession=62727385, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105375046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375046, CurrentID=0, Status='live', Symbol='LOC105375046', Aliases='', description='uncharacterized LOC105375046', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39077293, end_position_on_the_genomic_accession=39078956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90024, CurrentID=0, Status='live', Symbol='FLJ20021', Aliases='', description='uncharacterized LOC90024', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101347777, end_position_on_the_genomic_accession=101348883, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  90025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90025, CurrentID=0, Status='live', Symbol='UBE3D', Aliases='C6orf157, H10BH, UBE2CBP, YJR141W', description='ubiquitin protein ligase E3D', other_designations='UBCH10 binding protein with a hect-like domain|ubcH10-binding protein with a HECT-like domain|ubiquitin-conjugating enzyme E2C binding protein|ubiquitin-conjugating enzyme E2C-binding protein', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82892398, end_position_on_the_genomic_accession=83065841, orientation='minus', exon_count=16, OMIM=612495, no_hdr0=''),
  105375047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375047, CurrentID=0, Status='live', Symbol='LOC105375047', Aliases='', description='uncharacterized LOC105375047', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39314153, end_position_on_the_genomic_accession=39342752, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128154, CurrentID=0, Status='live', Symbol='LOC100128154', Aliases='', description='KIAA1919 pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52579561, end_position_on_the_genomic_accession=52581106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  648947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648947, CurrentID=0, Status='live', Symbol='AKR1C7P', Aliases='', description='aldo-keto reductase family 1, member C7, pseudogene', other_designations='aldo-keto reductase family 1, member C pseudogene|aldo-keto reductase family 1, member C1 (dihydrodiol dehydrogenase 1; 20-alpha (3-alpha)-hydroxysteroid dehydrogenase) pseudogene', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5275206, end_position_on_the_genomic_accession=5285925, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  221120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221120, CurrentID=0, Status='live', Symbol='ALKBH3', Aliases='ABH3, DEPC-1, DEPC1, PCA1', description='alkB homolog 3, alpha-ketoglutarate-dependent dioxygenase', other_designations='alkB, alkylation repair homolog 3|alkylated DNA repair protein alkB homolog 3|prostate cancer antigen 1|prostate cancer antigen-1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43880807, end_position_on_the_genomic_accession=43920275, orientation='plus', exon_count=10, OMIM=610603, no_hdr0=''),
  90050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90050, CurrentID=0, Status='live', Symbol='FAM181A', Aliases='C14orf152', description='family with sequence similarity 181 member A', other_designations='family with sequence similarity 181, member A', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93918894, end_position_on_the_genomic_accession=93929608, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375052, CurrentID=0, Status='live', Symbol='LOC105375052', Aliases='', description='uncharacterized LOC105375052', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40715540, end_position_on_the_genomic_accession=40726072, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90060, CurrentID=0, Status='live', Symbol='CCDC120', Aliases='JM11', description='coiled-coil domain containing 120', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49053432, end_position_on_the_genomic_accession=49069858, orientation='plus', exon_count=11, OMIM=300947, no_hdr0=''),
  105375053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375053, CurrentID=0, Status='live', Symbol='LOC105375053', Aliases='', description='uncharacterized LOC105375053', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40794741, end_position_on_the_genomic_accession=40825326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129459, CurrentID=0, Status='live', Symbol='PWWP2AP1', Aliases='', description='PWWP domain containing 2A pseudogene 1', other_designations='PWWP domain-containing protein 2A pseudogene', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80654755, end_position_on_the_genomic_accession=80657002, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  90070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90070, CurrentID=0, Status='live', Symbol='LACRT', Aliases='', description='lacritin', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54630839, end_position_on_the_genomic_accession=54634879, orientation='minus', exon_count=5, OMIM=607360, no_hdr0=''),
  221143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221143, CurrentID=0, Status='live', Symbol='N6AMT2', Aliases='ESP13', description='N-6 adenine-specific DNA methyltransferase 2 (putative)', other_designations='n(6)-adenine-specific DNA methyltransferase 2', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20728931, end_position_on_the_genomic_accession=20773956, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  90075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90075, CurrentID=0, Status='live', Symbol='ZNF30', Aliases='KOX28', description='zinc finger protein 30', other_designations='zinc finger protein 30 (KOX 28)|zinc finger protein KOX28', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34926903, end_position_on_the_genomic_accession=34945172, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  692218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692218, CurrentID=0, Status='live', Symbol='SNORD115@', Aliases='HBII-52', description='small nucleolar RNA, C/D box 115 cluster', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25170723, end_position_on_the_genomic_accession=25269858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  221150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221150, CurrentID=0, Status='live', Symbol='SKA3', Aliases='C13orf3, RAMA1', description='spindle and kinetochore associated complex subunit 3', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21153595, end_position_on_the_genomic_accession=21176602, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105379511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379511, CurrentID=0, Status='live', Symbol='LOC105379511', Aliases='', description='uncharacterized LOC105379511', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9529922, end_position_on_the_genomic_accession=9559264, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  221154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221154, CurrentID=0, Status='live', Symbol='MICU2', Aliases='1110008L20Rik, EFHA1', description='mitochondrial calcium uptake 2', other_designations='EF hand domain family A1|EF hand domain family, member A1|EF-hand domain family member A1|EF-hand domain family, member A1|EF-hand domain-containing family member A1|Smhs2 homolog', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21492689, end_position_on_the_genomic_accession=21604216, orientation='minus', exon_count=12, OMIM=610632, no_hdr0=''),
  100306939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306939, CurrentID=0, Status='live', Symbol='NBLST6', Aliases='', description='Neuroblastoma, susceptibility to, 6', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613017, no_hdr0=''),
  105375057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375057, CurrentID=0, Status='live', Symbol='LOC105375057', Aliases='', description='triggering receptor expressed on myeloid cells 1-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41250077, end_position_on_the_genomic_accession=41253515, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100306940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100306940, CurrentID=0, Status='live', Symbol='FL1', Aliases='', description='Follicular lymphoma, susceptibility to, 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613024, no_hdr0=''),
  100132179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132179, CurrentID=0, Status='live', Symbol='LOC100132179', Aliases='', description='solute carrier family 35 member E3 pseudogene', other_designations='solute carrier family 35, member E3 pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221549788, end_position_on_the_genomic_accession=221550649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90102, CurrentID=0, Status='live', Symbol='PHLDB2', Aliases='LL5b, LL5beta', description='pleckstrin homology-like domain, family B, member 2', other_designations='LL5 beta|protein LL5-beta', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111732480, end_position_on_the_genomic_accession=111976517, orientation='plus', exon_count=21, OMIM=610298, no_hdr0=''),
  105375060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375060, CurrentID=0, Status='live', Symbol='LOC105375060', Aliases='', description='uncharacterized LOC105375060', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41996902, end_position_on_the_genomic_accession=42013865, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  221178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221178, CurrentID=0, Status='live', Symbol='SPATA13', Aliases='ARHGEF29, ASEF2', description='spermatogenesis associated 13', other_designations='APC-stimulated guanine nucleotide exchange factor 2|adenomatous polyposis coli stimulated exchange factor 2', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23979626, end_position_on_the_genomic_accession=24307074, orientation='plus', exon_count=20, OMIM=613324, no_hdr0=''),
  221180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221180, CurrentID=0, Status='live', Symbol='CYCSP33', Aliases='HCP33', description='cytochrome c, somatic pseudogene 33', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24400712, end_position_on_the_genomic_accession=24401029, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375061, CurrentID=0, Status='live', Symbol='LOC105375061', Aliases='', description='uncharacterized LOC105375061', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42234458, end_position_on_the_genomic_accession=42240949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  221184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221184, CurrentID=0, Status='live', Symbol='CPNE2', Aliases='COPN2, CPN2', description='copine II', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57092543, end_position_on_the_genomic_accession=57148369, orientation='plus', exon_count=16, OMIM=604206, no_hdr0=''),
  90113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90113, CurrentID=0, Status='live', Symbol='VWA5B2', Aliases='', description='von Willebrand factor A domain containing 5B2', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184229588, end_position_on_the_genomic_accession=184242330, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  221188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221188, CurrentID=0, Status='live', Symbol='ADGRG5', Aliases='GPR114, PGR27', description='adhesion G protein-coupled receptor G5', other_designations='G protein-coupled receptor 114|G protein-coupled receptor PGR27|G-protein coupled receptor 114|G-protein coupled receptor PGR27|probable G-protein coupled receptor 114', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57529333, end_position_on_the_genomic_accession=57577210, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  221191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221191, CurrentID=0, Status='live', Symbol='PRSS54', Aliases='CT67, KLKBL4', description='protease, serine 54', other_designations='cancer/testis antigen 67|plasma kallikrein-like protein 4|protease, serine, 54', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58279997, end_position_on_the_genomic_accession=58295047, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  90120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90120, CurrentID=0, Status='live', Symbol='C9orf69', Aliases='', description='chromosome 9 open reading frame 69', other_designations='herpes virus UL25-binding protein', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136110220, end_position_on_the_genomic_accession=136118885, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  90121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90121, CurrentID=0, Status='live', Symbol='TSR2', Aliases='DBA14, DT1P1A10, WGG1', description='TSR2, 20S rRNA accumulation, homolog (S. cerevisiae)', other_designations='WGG motif containing 1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54440420, end_position_on_the_genomic_accession=54445297, orientation='plus', exon_count=5, OMIM=300945, no_hdr0=''),
  105373401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373401, CurrentID=0, Status='live', Symbol='LOC105373401', Aliases='', description='uncharacterized LOC105373401', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6291565, end_position_on_the_genomic_accession=6313142, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375063, CurrentID=0, Status='live', Symbol='LOC105375063', Aliases='', description='uncharacterized LOC105375063', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42894450, end_position_on_the_genomic_accession=42903035, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060275, CurrentID=0, Status='live', Symbol='LOC101060275', Aliases='', description='nuclear pore complex-interacting protein family member B4-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  90133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90133, CurrentID=0, Status='live', Symbol='KRT8P12', Aliases='KRT8L2', description='keratin 8 pseudogene 12', other_designations='keratin 8-like 2', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160567779, end_position_on_the_genomic_accession=160569472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  90134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90134, CurrentID=0, Status='live', Symbol='KCNH7', Aliases='ERG3, HERG3, Kv11.3', description='potassium channel, voltage gated eag related subfamily H, member 7', other_designations='ERG-3|eag-related protein 3|ether-a-go-go-related gene potassium channel 3|ether-a-go-go-related protein 3|potassium channel subunit HERG-3|voltage-gated potassium channel subunit Kv11.3', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162371407, end_position_on_the_genomic_accession=162838747, orientation='minus', exon_count=17, OMIM=608169, no_hdr0=''),
  90135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90135, CurrentID=0, Status='live', Symbol='BTBD6', Aliases='BDPL', description='BTB (POZ) domain containing 6', other_designations='BTB domain protein BDPL|glucocorticoid receptor AF-1 coactivator-1|lens BTB domain protein', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105248542, end_position_on_the_genomic_accession=105251093, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90139, CurrentID=0, Status='live', Symbol='TSPAN18', Aliases='TSPAN', description='tetraspanin 18', other_designations='tspan-18', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44727130, end_position_on_the_genomic_accession=44932427, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  90141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90141, CurrentID=0, Status='live', Symbol='EFCAB11', Aliases='C14orf143', description='EF-hand calcium binding domain 11', other_designations='EF-hand domain-containing protein C14orf143', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89794984, end_position_on_the_genomic_accession=89954777, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  352290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352290, CurrentID=0, Status='live', Symbol='MRPS23P1', Aliases='', description='mitochondrial ribosomal protein S23 pseudogene 1', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44919587, end_position_on_the_genomic_accession=44920490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  221223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221223, CurrentID=0, Status='live', Symbol='CES5A', Aliases='CAUXIN, CES4C1, CES5, CES7, HEL126', description='carboxylesterase 5A', other_designations='carboxylesterase 5|carboxylesterase 7|carboxylesterase-like urinary excreted protein homolog|epididymis luminal protein 126', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55846154, end_position_on_the_genomic_accession=55956031, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105375068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375068, CurrentID=0, Status='live', Symbol='LOC105375068', Aliases='', description='uncharacterized LOC105375068', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43862667, end_position_on_the_genomic_accession=43870149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  692231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=692231, CurrentID=0, Status='live', Symbol='RHEBP3', Aliases='', description='Ras-homolog enriched in brain pseudogene 3', other_designations='Ras homolog enriched in brain pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18019206, end_position_on_the_genomic_accession=18020604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90161, CurrentID=0, Status='live', Symbol='HS6ST2', Aliases='', description='heparan sulfate 6-O-sulfotransferase 2', other_designations='HS6ST-2', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132626010, end_position_on_the_genomic_accession=132961395, orientation='minus', exon_count=10, OMIM=300545, no_hdr0=''),
  90167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90167, CurrentID=0, Status='live', Symbol='FRMD7', Aliases='NYS, NYS1, XIPAN', description='FERM domain containing 7', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132076993, end_position_on_the_genomic_accession=132128022, orientation='minus', exon_count=12, OMIM=300628, no_hdr0=''),
  221241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221241, CurrentID=0, Status='live', Symbol='LINC00305', Aliases='C18orf20, HsT1235, NCRNA00305', description='long intergenic non-protein coding RNA 305', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64080009, end_position_on_the_genomic_accession=64149026, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  90187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90187, CurrentID=0, Status='live', Symbol='EMILIN3', Aliases='C20orf130, EMILIN5, dJ620E11.4', description='elastin microfibril interfacer 3', other_designations='EMILIN-5|elastin microfibril interface-located protein 3|elastin microfibril interface-located protein 5|elastin microfibril interfacer 5', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41359966, end_position_on_the_genomic_accession=41366858, orientation='minus', exon_count=4, OMIM=608929, no_hdr0=''),
  105375074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375074, CurrentID=0, Status='live', Symbol='LOC105375074', Aliases='', description='uncharacterized LOC105375074', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44510223, end_position_on_the_genomic_accession=44527546, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  221262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221262, CurrentID=0, Status='live', Symbol='CCDC162P', Aliases='C6orf184, C6orf185, CCDC162, bA425D10.3, bA425D10.7', description='coiled-coil domain containing 162, pseudogene', other_designations='coiled-coil domain containing 162 pseudogene', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109294303, end_position_on_the_genomic_accession=109308220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  221264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221264, CurrentID=0, Status='live', Symbol='AK9', Aliases='AK 9, AKD1, AKD2, C6orf199, C6orf224, dJ70A9.1', description='adenylate kinase 9', other_designations='adenylate kinase domain containing 1|adenylate kinase domain containing 2', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109492855, end_position_on_the_genomic_accession=109691212, orientation='minus', exon_count=48, OMIM=615358, no_hdr0=''),
  90193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90193, CurrentID=0, Status='live', Symbol='RPL7P6', Aliases='RPL7_20_1216', description='ribosomal protein L7 pseudogene 6', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19089131, end_position_on_the_genomic_accession=19089951, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375075, CurrentID=0, Status='live', Symbol='LOC105375075', Aliases='', description='uncharacterized LOC105375075', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44546166, end_position_on_the_genomic_accession=44551972, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  90196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90196, CurrentID=0, Status='live', Symbol='SYS1', Aliases='C20orf169, dJ453C12.4, dJ453C12.4.1', description='Sys1 golgi trafficking protein', other_designations='SYS1 Golgi-localized integral membrane protein homolog', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45361937, end_position_on_the_genomic_accession=45376802, orientation='plus', exon_count=6, OMIM=612979, no_hdr0=''),
  90199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90199, CurrentID=0, Status='live', Symbol='WFDC8', Aliases='C20orf170, HEL-S-292, WAP8, dJ461P17.1', description='WAP four-disulfide core domain 8', other_designations='WAP motif protein 1|epididymis secretory protein Li 292|protease inhibitor WAP8|putative protease inhibitor WAP8', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45551152, end_position_on_the_genomic_accession=45579350, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  221272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221272, CurrentID=0, Status='live', Symbol='LOC221272', Aliases='', description='uncharacterized LOC221272', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  90203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90203, CurrentID=0, Status='live', Symbol='SNX21', Aliases='C20orf161, PP3993, SNX-L, dJ337O18.4', description='sorting nexin family member 21', other_designations='sorting nexin 21|sorting nexin L', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45833622, end_position_on_the_genomic_accession=45843275, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  90204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90204, CurrentID=0, Status='live', Symbol='ZSWIM1', Aliases='C20orf162', description='zinc finger, SWIM-type containing 1', other_designations='zinc finger, SWIM domain containing 1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45880942, end_position_on_the_genomic_accession=45885266, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132197, CurrentID=0, Status='live', Symbol='LOC100132197', Aliases='', description='uncharacterized LOC100132197', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=20391097, end_position_on_the_genomic_accession=20391666, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375079, CurrentID=0, Status='live', Symbol='LOC105375079', Aliases='', description='uncharacterized LOC105375079', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46214661, end_position_on_the_genomic_accession=46267419, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  221294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221294, CurrentID=0, Status='live', Symbol='NT5DC1', Aliases='C6orf200, LP2642, NT5C2L1', description="5'-nucleotidase domain containing 1", other_designations="5'-nucleotidase, cytosolic II-like 1 protein", map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116100823, end_position_on_the_genomic_accession=116245690, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  90226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90226, CurrentID=0, Status='live', Symbol='UCN2', Aliases='SRP, UCN-II, UCNI, UR, URP', description='urocortin 2', other_designations='prepro-urocortin 2|stresscopin-related peptide|ucn II|urocortin II|urocortin-related peptide', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48561718, end_position_on_the_genomic_accession=48563768, orientation='minus', exon_count=2, OMIM=605902, no_hdr0=''),
  221301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221301, CurrentID=0, Status='live', Symbol='FAM26D', Aliases='C6orf78', description='family with sequence similarity 26 member D', other_designations='family with sequence similarity 26, member D', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116528999, end_position_on_the_genomic_accession=116561123, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  221302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221302, CurrentID=0, Status='live', Symbol='ZUFSP', Aliases='C6orf113, dJ412I7.3', description='zinc finger with UFM1-specific peptidase domain', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116635618, end_position_on_the_genomic_accession=116668810, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  90231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90231, CurrentID=0, Status='live', Symbol='KIAA2013', Aliases='', description='KIAA2013', other_designations='RP5-1077B9.1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11919588, end_position_on_the_genomic_accession=11926428, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90233, CurrentID=0, Status='live', Symbol='ZNF551', Aliases='', description='zinc finger protein 551', other_designations='KOX 23 protein (56 AA)|zinc finger protein KOX23', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57681969, end_position_on_the_genomic_accession=57689811, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375082, CurrentID=0, Status='live', Symbol='LOC105375082', Aliases='', description='uncharacterized LOC105375082', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47377368, end_position_on_the_genomic_accession=47397171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928809, CurrentID=0, Status='live', Symbol='LOC101928809', Aliases='', description='uncharacterized LOC101928809', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75980790, end_position_on_the_genomic_accession=76005942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927199, CurrentID=0, Status='live', Symbol='LOC101927199', Aliases='', description='uncharacterized LOC101927199', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36194205, end_position_on_the_genomic_accession=36235362, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  90246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90246, CurrentID=0, Status='live', Symbol='LOC90246', Aliases='', description='uncharacterized LOC90246', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128507835, end_position_on_the_genomic_accession=128510595, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  90249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90249, CurrentID=0, Status='live', Symbol='UNC5A', Aliases='UNC5H1', description='unc-5 netrin receptor A', other_designations='netrin receptor Unc5h1|protein unc-5 homolog 1|protein unc-5 homolog A|unc-5 homolog 1|unc-5 homolog A|unc5 (C.elegans homolog) a', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176810559, end_position_on_the_genomic_accession=176880898, orientation='plus', exon_count=17, OMIM=607869, no_hdr0=''),
  221322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221322, CurrentID=0, Status='live', Symbol='TBC1D32', Aliases='BROMI, C6orf170, C6orf171, bA301B7.2, bA57L9.1, dJ310J6.1', description='TBC1 domain family member 32', other_designations='TBC1 domain family, member 32|bA301B7.2 (novel protein)|broad-minded homolog|dJ310J6.1 (novel protein)', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121079494, end_position_on_the_genomic_accession=121334956, orientation='minus', exon_count=47, OMIM=615867, no_hdr0=''),
  221336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221336, CurrentID=0, Status='live', Symbol='BEND6', Aliases='C6orf65, bA203B9.1', description='BEN domain containing 6', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56954975, end_position_on_the_genomic_accession=57027652, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100132207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132207, CurrentID=0, Status='live', Symbol='LOC100132207', Aliases='', description='uncharacterized LOC100132207', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=130809897, end_position_on_the_genomic_accession=130810421, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  90268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90268, CurrentID=0, Status='live', Symbol='OTULIN', Aliases='FAM105B, GUM', description='OTU deubiquitinase with linear linkage specificity', other_designations='OTU domain-containing deubiquitinase with linear linkage specificity|deubiquitinating enzyme otulin|family with sequence similarity 105, member B|gumby|protein FAM105B|ubiquitin thioesterase Gumby', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14664674, end_position_on_the_genomic_accession=14712867, orientation='plus', exon_count=11, OMIM=615712, no_hdr0=''),
  90271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90271, CurrentID=0, Status='live', Symbol='OLMALINC', Aliases='C10orf75, HI-LNC80, LINC00263, NCRNA00263AS, OLMALINC', description='oligodendrocyte maturation-associated long intergenic non-coding RNA', other_designations='Human Islet Long Non Coding RNA 80|long intergenic non-protein coding RNA 263|oligodendrocyte maturation-associated long intervening non-coding RNA', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100373576, end_position_on_the_genomic_accession=100388354, orientation='plus', exon_count=3, OMIM=616240, no_hdr0=''),
  90273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90273, CurrentID=0, Status='live', Symbol='CEACAM21', Aliases='CEACAM3, R29124_1', description='carcinoembryonic antigen-related cell adhesion molecule 21', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41549250, end_position_on_the_genomic_accession=41586844, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  221357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221357, CurrentID=0, Status='live', Symbol='GSTA5', Aliases='', description='glutathione S-transferase alpha 5', other_designations='GST class-alpha member 5|glutathione S-transferase A5-5|glutathione transferase A5', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52831742, end_position_on_the_genomic_accession=52846095, orientation='minus', exon_count=7, OMIM=607605, no_hdr0=''),
  90288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90288, CurrentID=0, Status='live', Symbol='EFCAB12', Aliases='C3orf25', description='EF-hand calcium binding domain 12', other_designations='EF-hand domain-containing protein C3orf25', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129401321, end_position_on_the_genomic_accession=129428651, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105374275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374275, CurrentID=0, Status='live', Symbol='LOC105374275', Aliases='', description='uncharacterized LOC105374275', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191748620, end_position_on_the_genomic_accession=191759658, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  90293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90293, CurrentID=0, Status='live', Symbol='KLHL13', Aliases='BKLHD2', description='kelch like family member 13', other_designations='BTB and kelch domain containing 2|kelch-like 13|kelch-like family member 13', map_location='Xq23-q24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=117897813, end_position_on_the_genomic_accession=118117340, orientation='minus', exon_count=14, OMIM=300655, no_hdr0=''),
  105375093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375093, CurrentID=0, Status='live', Symbol='LOC105375093', Aliases='', description='uncharacterized LOC105375093', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53101119, end_position_on_the_genomic_accession=53102749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  90313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90313, CurrentID=0, Status='live', Symbol='TP53I13', Aliases='DSCP1', description='tumor protein p53 inducible protein 13', other_designations='damage-stimulated cytoplasmic protein 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29567180, end_position_on_the_genomic_accession=29580820, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100132215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132215, CurrentID=0, Status='live', Symbol='LOC100132215', Aliases='', description='uncharacterized LOC100132215', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63043922, end_position_on_the_genomic_accession=63048521, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  90316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90316, CurrentID=0, Status='live', Symbol='TGIF2LX', Aliases='TGIFLX', description='TGFB-induced factor homeobox 2-like, X-linked', other_designations='TGF-beta-induced transcription factor 2-like protein|TGFB-induced factor 2-like protein, X-linked|TGFB-induced factor 2-like, X-linked|TGIF-like on the X', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=89921941, end_position_on_the_genomic_accession=89922883, orientation='plus', exon_count=2, OMIM=300411, no_hdr0=''),
  90317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90317, CurrentID=0, Status='live', Symbol='ZNF616', Aliases='', description='zinc finger protein 616', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52114400, end_position_on_the_genomic_accession=52139938, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  221391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221391, CurrentID=0, Status='live', Symbol='OPN5', Aliases='GPR136, GRP136, PGR12, TMEM13', description='opsin 5', other_designations='G-protein coupled receptor 136|neuropsin|transmembrane protein 13', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47780302, end_position_on_the_genomic_accession=47857840, orientation='plus', exon_count=20, OMIM=609042, no_hdr0=''),
  221393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221393, CurrentID=0, Status='live', Symbol='ADGRF4', Aliases='GPR115, PGR18', description='adhesion G protein-coupled receptor F4', other_designations='G protein-coupled receptor 115|G-protein coupled receptor 115|G-protein coupled receptor PGR18|probable G-protein coupled receptor 115|seven transmembrane helix receptor', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47698553, end_position_on_the_genomic_accession=47722021, orientation='plus', exon_count=10, OMIM=614268, no_hdr0=''),
  221395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221395, CurrentID=0, Status='live', Symbol='ADGRF5', Aliases='GPR116, KPG_001', description='adhesion G protein-coupled receptor F5', other_designations='G protein-coupled receptor 116|G-protein coupled receptor 116|Ig-Hepta homolog|probable G-protein coupled receptor 116', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46852505, end_position_on_the_genomic_accession=46954954, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  90324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90324, CurrentID=0, Status='live', Symbol='CCDC97', Aliases='', description='coiled-coil domain containing 97', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41310189, end_position_on_the_genomic_accession=41324883, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  90326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90326, CurrentID=0, Status='live', Symbol='THAP3', Aliases='', description='THAP domain containing, apoptosis associated protein 3', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6624865, end_position_on_the_genomic_accession=6635586, orientation='plus', exon_count=8, OMIM=612532, no_hdr0=''),
  105375097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375097, CurrentID=0, Status='live', Symbol='LOC105375097', Aliases='', description='uncharacterized LOC105375097', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54146386, end_position_on_the_genomic_accession=54190432, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  221400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221400, CurrentID=0, Status='live', Symbol='TDRD6', Aliases='CT41.2, NY-CO-45, SPATA36, TDR2, bA446F17.4', description='tudor domain containing 6', other_designations='antigen NY-CO-45|cancer/testis antigen 41.2|spermatogenesis associated 36|tudor repeat 2', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46687875, end_position_on_the_genomic_accession=46704319, orientation='plus', exon_count=4, OMIM=611200, no_hdr0=''),
  90332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90332, CurrentID=0, Status='live', Symbol='EXOC3L2', Aliases='XTP7', description='exocyst complex component 3-like 2', other_designations='HBV X-transactivated gene 7 protein|HBV XAg-transactivated protein 7|protein 7 transactivated by hepatitis B virus X antigen (HBxAg)', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45212621, end_position_on_the_genomic_accession=45234211, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  90333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90333, CurrentID=0, Status='live', Symbol='ZNF468', Aliases='', description='zinc finger protein 468', other_designations='zinc finger protein ZNF468', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52838532, end_position_on_the_genomic_accession=52857649, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  221409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221409, CurrentID=0, Status='live', Symbol='SPATS1', Aliases='DDIP, SPATA8, SRSP1', description='spermatogenesis associated, serine-rich 1', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44342660, end_position_on_the_genomic_accession=44377167, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  90338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90338, CurrentID=0, Status='live', Symbol='ZNF160', Aliases='F11, HKr18, HZF5, KR18', description='zinc finger protein 160', other_designations='KRAB zinc finger protein KR18|zinc finger protein 5|zinc finger protein HZF5|zinc finger protein Kr18', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53064882, end_position_on_the_genomic_accession=53103471, orientation='minus', exon_count=10, OMIM=600398, no_hdr0=''),
  90342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90342, CurrentID=0, Status='live', Symbol='FER1L5', Aliases='', description='fer-1-like family member 5', other_designations='fer-1-like 5', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96642711, end_position_on_the_genomic_accession=96706220, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  221416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221416, CurrentID=0, Status='live', Symbol='C6orf223', Aliases='', description='chromosome 6 open reading frame 223', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44000597, end_position_on_the_genomic_accession=44005958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  221421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221421, CurrentID=0, Status='live', Symbol='RSPH9', Aliases='C6orf206, CILD12, MRPS18AL1', description='radial spoke head 9 homolog (Chlamydomonas)', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43645030, end_position_on_the_genomic_accession=43671013, orientation='plus', exon_count=8, OMIM=612648, no_hdr0=''),
  221424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221424, CurrentID=0, Status='live', Symbol='LRRC73', Aliases='C6orf154', description='leucine rich repeat containing 73', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43506965, end_position_on_the_genomic_accession=43510343, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  90353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90353, CurrentID=0, Status='live', Symbol='CTU1', Aliases='ATPBD3, NCS6', description='cytosolic thiouridylase subunit 1', other_designations='ATP binding domain 3|ATP-binding domain-containing protein 3|cancer-associated gene protein|cancer-associated protein|cytoplasmic tRNA adenylyltransferase 1|cytosolic thiouridylase subunit 1 homolog', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51097606, end_position_on_the_genomic_accession=51108390, orientation='minus', exon_count=3, OMIM=612694, no_hdr0=''),
  90355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90355, CurrentID=0, Status='live', Symbol='C5orf30', Aliases='', description='chromosome 5 open reading frame 30', other_designations='UPF0684 protein C5orf30', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103258741, end_position_on_the_genomic_accession=103278660, orientation='plus', exon_count=6, OMIM=616608, no_hdr0=''),
  90362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90362, CurrentID=0, Status='live', Symbol='FAM110B', Aliases='C8orf72', description='family with sequence similarity 110 member B', other_designations='family with sequence similarity 110, member B', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57994509, end_position_on_the_genomic_accession=58149718, orientation='plus', exon_count=6, OMIM=611394, no_hdr0=''),
  100132223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132223, CurrentID=0, Status='live', Symbol='LOC100132223', Aliases='', description='actin, beta pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180820460, end_position_on_the_genomic_accession=180821815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  221438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221438, CurrentID=0, Status='live', Symbol='TREML5P', Aliases='TLT5, TREML2P, TREML2P1', description='triggering receptor expressed on myeloid cells like 5, pseudogene', other_designations='TREM like transcript 5|TREM-like transcript 5|triggering receptor expressed on myeloid cells-like 2 pseudogene 1|triggering receptor expressed on myeloid cells-like 5, pseudogene', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41249377, end_position_on_the_genomic_accession=41249589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  221442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221442, CurrentID=0, Status='live', Symbol='ADCY10P1', Aliases='', description='adenylate cyclase 10 (soluble) pseudogene 1', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41101034, end_position_on_the_genomic_accession=41140835, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  221443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221443, CurrentID=0, Status='live', Symbol='OARD1', Aliases='C6orf130, TARG1, dJ34B21.3', description='O-acyl-ADP-ribose deacylase 1', other_designations='O-acetyl-ADP-ribose deacetylase C6orf130|terminal ADP-ribose protein glycohydrolase 1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41066792, end_position_on_the_genomic_accession=41072449, orientation='minus', exon_count=6, OMIM=614393, no_hdr0=''),
  90378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90378, CurrentID=0, Status='live', Symbol='SAMD1', Aliases='', description='sterile alpha motif domain containing 1', other_designations='SAM domain containing 1|sterile alpha motif domain-containing protein 1', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14088417, end_position_on_the_genomic_accession=14090420, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  90379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90379, CurrentID=0, Status='live', Symbol='DCAF15', Aliases='C19orf72', description='DDB1 and CUL4 associated factor 15', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13952507, end_position_on_the_genomic_accession=13961444, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  90381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90381, CurrentID=0, Status='live', Symbol='TICRR', Aliases='C15orf42, SLD3, Treslin', description='TOPBP1-interacting checkpoint and replication regulator', other_designations='topBP1-interacting replication-stimulating protein|topBP1-interacting, replication-stimulating protein', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89575438, end_position_on_the_genomic_accession=89628023, orientation='plus', exon_count=22, OMIM=613298, no_hdr0=''),
  221458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221458, CurrentID=0, Status='live', Symbol='KIF6', Aliases='C6orf102, dJ1043E3.1, dJ137F1.4, dJ188D3.1', description='kinesin family member 6', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39335100, end_position_on_the_genomic_accession=39725472, orientation='minus', exon_count=27, OMIM=613919, no_hdr0=''),
  90390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90390, CurrentID=0, Status='live', Symbol='MED30', Aliases='MED30S, THRAP6, TRAP25', description='mediator complex subunit 30', other_designations='TRAP/Mediator complex component TRAP25|putative mediator of RNA polymerase II transcription subunit 30|thyroid hormone receptor associated protein 6|thyroid hormone receptor-associated protein 6|thyroid hormone receptor-associated protein complex 25 kDa component', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117520713, end_position_on_the_genomic_accession=117540262, orientation='plus', exon_count=5, OMIM=610237, no_hdr0=''),
  221468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221468, CurrentID=0, Status='live', Symbol='TMEM217', Aliases='C6orf128, dJ355M6.2', description='transmembrane protein 217', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37212178, end_position_on_the_genomic_accession=37259005, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  221472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221472, CurrentID=0, Status='live', Symbol='FGD2', Aliases='ZFYVE4', description='FYVE, RhoGEF and PH domain containing 2', other_designations='FGD1 family, member 2|FLJ00276 protein|zinc finger FYVE domain-containing protein 4', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37005647, end_position_on_the_genomic_accession=37029070, orientation='plus', exon_count=19, OMIM=605091, no_hdr0=''),
  221476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221476, CurrentID=0, Status='live', Symbol='PI16', Aliases='CD364, CRISP9, MSMBBP, PSPBP', description='peptidase inhibitor 16', other_designations='PSP94-binding protein|cysteine-rich secretory protein 9|microseminoprotein, beta-binding protein|protease inhibitor 16', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36948263, end_position_on_the_genomic_accession=36964837, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  221477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221477, CurrentID=0, Status='live', Symbol='C6orf89', Aliases='BRAP', description='chromosome 6 open reading frame 89', other_designations='amfion', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36871846, end_position_on_the_genomic_accession=36928964, orientation='plus', exon_count=15, OMIM=616642, no_hdr0=''),
  90407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90407, CurrentID=0, Status='live', Symbol='TMEM41A', Aliases='2900010K02Rik', description='transmembrane protein 41A', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185482671, end_position_on_the_genomic_accession=185499057, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  221481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221481, CurrentID=0, Status='live', Symbol='ARMC12', Aliases='C6orf81', description='armadillo repeat containing 12', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35733805, end_position_on_the_genomic_accession=35748913, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  90410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90410, CurrentID=0, Status='live', Symbol='IFT20', Aliases='', description='intraflagellar transport 20', other_designations='intraflagellar transport 20 homolog|intraflagellar transport protein IFT20', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28328325, end_position_on_the_genomic_accession=28335489, orientation='minus', exon_count=7, OMIM=614394, no_hdr0=''),
  90411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90411, CurrentID=0, Status='live', Symbol='MCFD2', Aliases='F5F8D, F5F8D2, LMAN1IP, SDNSF', description='multiple coagulation factor deficiency 2', other_designations='neural stem cell-derived neuronal survival protein', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46901870, end_position_on_the_genomic_accession=46941855, orientation='minus', exon_count=7, OMIM=607788, no_hdr0=''),
  90416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90416, CurrentID=0, Status='live', Symbol='C15orf57', Aliases='CCDC32', description='chromosome 15 open reading frame 57', other_designations='coiled-coil domain containing 32', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40531341, end_position_on_the_genomic_accession=40565057, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  90417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90417, CurrentID=0, Status='live', Symbol='KNSTRN', Aliases='C15orf23, HSD11, SKAP, TRAF4AF1', description='kinetochore-localized astrin/SPAG5 binding protein', other_designations='TRAF4 associated factor 1|TRAF4-associated factor 1|kinastrin|kinetochore-localized astrin-binding protein|kinetochore-localized astrin/SPAG5-binding protein|putative TRAF4-associated factor 1|small kinetochore associated protein|small kinetochore-associated protein, kinetochore-localized astrin-binding protein, TRAF4 associated factor 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40382721, end_position_on_the_genomic_accession=40394288, orientation='plus', exon_count=9, OMIM=614718, no_hdr0=''),
  105375624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375624, CurrentID=0, Status='live', Symbol='LOC105375624', Aliases='', description='uncharacterized LOC105375624', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86333274, end_position_on_the_genomic_accession=86345460, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  221491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221491, CurrentID=0, Status='live', Symbol='C6orf1', Aliases='LBH', description='chromosome 6 open reading frame 1', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34246380, end_position_on_the_genomic_accession=34249108, orientation='minus', exon_count=5, OMIM=611419, no_hdr0=''),
  105373403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373403, CurrentID=0, Status='live', Symbol='LOC105373403', Aliases='', description='uncharacterized LOC105373403', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6428679, end_position_on_the_genomic_accession=6431854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  90423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90423, CurrentID=0, Status='live', Symbol='ATP6V1E2', Aliases='ATP6E1, ATP6EL2, ATP6V1EL2, VMA4', description='ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E2', other_designations='V-ATPase subunit E 2|vacuolar proton pump subunit E 2|vacuolar-type proton-translocating ATPase subunit E1', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46489355, end_position_on_the_genomic_accession=46542412, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  221496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221496, CurrentID=0, Status='live', Symbol='LEMD2', Aliases='NET25, dJ482C21.1', description='LEM domain containing 2', other_designations='hLEM2', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33771213, end_position_on_the_genomic_accession=33789163, orientation='minus', exon_count=9, OMIM=616312, no_hdr0=''),
  90427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90427, CurrentID=0, Status='live', Symbol='BMF', Aliases='', description='Bcl2 modifying factor', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40087890, end_position_on_the_genomic_accession=40108884, orientation='minus', exon_count=7, OMIM=606266, no_hdr0=''),
  105375114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375114, CurrentID=0, Status='live', Symbol='LOC105375114', Aliases='', description='uncharacterized LOC105375114', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112878, end_position_on_the_genomic_accession=119313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  221504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221504, CurrentID=0, Status='live', Symbol='ZBTB9', Aliases='ZNF919', description='zinc finger and BTB domain containing 9', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33454579, end_position_on_the_genomic_accession=33457544, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  90441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90441, CurrentID=0, Status='live', Symbol='ZNF622', Aliases='ZPR9', description='zinc finger protein 622', other_designations='zinc finger-like protein 9', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16451519, end_position_on_the_genomic_accession=16470327, orientation='minus', exon_count=7, OMIM=608694, no_hdr0=''),
  105375118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375118, CurrentID=0, Status='live', Symbol='LOC105375118', Aliases='', description='uncharacterized LOC105375118', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=330339, end_position_on_the_genomic_accession=333249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  221527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221527, CurrentID=0, Status='live', Symbol='ZBTB12', Aliases='Bat9, C6orf46, D6S59E, G10, NG35', description='zinc finger and BTB domain containing 12', other_designations='HLA-B-associated transcript 9', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31899204, end_position_on_the_genomic_accession=31902367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90459, CurrentID=0, Status='live', Symbol='ERI1', Aliases="3'HEXO, HEXO, THEX1", description='exoribonuclease 1', other_designations="3' exoribonuclease|3'-5' exonuclease ERI1|Eri-1 homolog|enhanced RNAi three prime mRNA exonuclease homolog 1|histone mRNA 3' end-specific exonuclease|histone mRNA 3'-end-specific exoribonuclease|histone mRNA 3'-exonuclease 1|protein 3'hExo|three prime histone mRNA exonuclease 1|three prime mRNA exonuclease 1", map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9002804, end_position_on_the_genomic_accession=9100256, orientation='plus', exon_count=12, OMIM=608739, no_hdr0=''),
  105375120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375120, CurrentID=0, Status='live', Symbol='LOC105375120', Aliases='', description='uncharacterized LOC105375120', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1074248, end_position_on_the_genomic_accession=1078875, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  221545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221545, CurrentID=0, Status='live', Symbol='C6orf136', Aliases='', description='chromosome 6 open reading frame 136', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30647039, end_position_on_the_genomic_accession=30653210, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  221547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221547, CurrentID=0, Status='live', Symbol='RANP1', Aliases='Ras-like, TC4', description='RAN, member RAS oncogene family pseudogene 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30485885, end_position_on_the_genomic_accession=30486976, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132242, CurrentID=0, Status='live', Symbol='LOC100132242', Aliases='', description='exosome component 8 pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43705856, end_position_on_the_genomic_accession=43706757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90480, CurrentID=0, Status='live', Symbol='GADD45GIP1', Aliases='CKBBP2, CKbetaBP2, CRIF1, MRP-L59, PLINP, PLINP-1, PRG6, Plinp1', description='GADD45G interacting protein 1', other_designations='39S ribosomal protein L59, mitochondrial|CKII beta binding protein 2|CKII beta-associating protein|CR6 interacting factor 1|growth arrest and DNA-damage-inducible, gamma interacting protein 1|growth arrest- and DNA damage-inducible GADD45G-interacting protein|p53-responsive gene 6 protein|papillomavirus L2 interacting nuclear protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12954156, end_position_on_the_genomic_accession=12957254, orientation='minus', exon_count=2, OMIM=605162, no_hdr0=''),
  105376895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376895, CurrentID=0, Status='live', Symbol='LOC105376895', Aliases='', description='uncharacterized LOC105376895', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28240974, end_position_on_the_genomic_accession=28241376, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  90485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90485, CurrentID=0, Status='live', Symbol='ZNF835', Aliases='BC37295_3', description='zinc finger protein 835', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56662652, end_position_on_the_genomic_accession=56671783, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  90488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90488, CurrentID=0, Status='live', Symbol='TMEM263', Aliases='C12orf23', description='transmembrane protein 263', other_designations='UPF0444 transmembrane protein C12orf23', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106953415, end_position_on_the_genomic_accession=106974035, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90499, CurrentID=0, Status='live', Symbol='FAM95A', Aliases='', description='family with sequence similarity 95 member A', other_designations='family with sequence similarity 95, member A', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94755326, end_position_on_the_genomic_accession=94761184, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90506, CurrentID=0, Status='live', Symbol='LRRC46', Aliases='', description='leucine rich repeat containing 46', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47831627, end_position_on_the_genomic_accession=47837713, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  90507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90507, CurrentID=0, Status='live', Symbol='SCRN2', Aliases='Ses2', description='secernin 2', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47837681, end_position_on_the_genomic_accession=47841333, orientation='minus', exon_count=7, OMIM=614966, no_hdr0=''),
  221584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221584, CurrentID=0, Status='live', Symbol='ZSCAN12P1', Aliases='ZNF187p1, ZNF305P1, ZNF96L1, ZNF96P1, ZSCAN12L1, dJ313I6.7', description='zinc finger and SCAN domain containing 12 pseudogene 1', other_designations='zinc finger and SCAN domain containing 12-like 1|zinc finger protein 305 pseudogene 1|zinc finger protein 96 pseudogene 1|zinc finger protein 96-like 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28090807, end_position_on_the_genomic_accession=28095715, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422324, CurrentID=0, Status='live', Symbol='LOC100422324', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143979094, end_position_on_the_genomic_accession=143979629, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375129, CurrentID=0, Status='live', Symbol='LOC105375129', Aliases='', description='uncharacterized LOC105375129', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3041844, end_position_on_the_genomic_accession=3042662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  90522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90522, CurrentID=0, Status='live', Symbol='YIF1B', Aliases='FinGER8', description='Yip1 interacting factor homolog B (S. cerevisiae)', other_designations='YIP1-interacting factor homolog B', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38303560, end_position_on_the_genomic_accession=38315976, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  90523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90523, CurrentID=0, Status='live', Symbol='MLIP', Aliases='C6orf142', description='muscular LMNA-interacting protein', other_designations='muscle-enriched A-type lamin interacting protein|muscular-enriched A-type laminin-interacting protein', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54010979, end_position_on_the_genomic_accession=54266280, orientation='plus', exon_count=21, OMIM=614106, no_hdr0=''),
  90525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90525, CurrentID=0, Status='live', Symbol='SHF', Aliases='', description='Src homology 2 domain containing F', other_designations='CTD-2651B20.7', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45167214, end_position_on_the_genomic_accession=45201329, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  90527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90527, CurrentID=0, Status='live', Symbol='DUOXA1', Aliases='NIP, NUMBIP, mol', description='dual oxidase maturation factor 1', other_designations='dual oxidase activator 1|dual oxidase maturation factor 1 alpha|dual oxidase maturation factor 1 beta|dual oxidase maturation factor 1 delta|dual oxidase maturation factor 1 gamma|homolog of Drosophila Numb-interacting protein', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45117366, end_position_on_the_genomic_accession=45129938, orientation='minus', exon_count=12, OMIM=612771, no_hdr0=''),
  90529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90529, CurrentID=0, Status='live', Symbol='STPG1', Aliases='C1orf201, MAPO2', description='sperm-tail PG-rich repeat containing 1', other_designations='O6-methylguanine-induced apoptosis 2|UPF0490 protein C1orf201|sperm-tail PG-rich repeat-containing protein 1', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24356999, end_position_on_the_genomic_accession=24415497, orientation='minus', exon_count=12, OMIM=615826, no_hdr0=''),
  105375131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375131, CurrentID=0, Status='live', Symbol='LOC105375131', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4126653, end_position_on_the_genomic_accession=4137174, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101409188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409188, CurrentID=0, Status='live', Symbol='FAME4', Aliases='FCMTE4', description='Epilepsy, familial adult myoclonic, 4', other_designations='', map_location='3q26.32-q28', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615127, no_hdr0=''),
  221613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221613, CurrentID=0, Status='live', Symbol='HIST1H2AA', Aliases='H2AA, H2AFR, bA317E16.2', description='histone cluster 1, H2aa', other_designations='H2A histone family, member R|histone 1, H2aa|histone H2A/r', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25726063, end_position_on_the_genomic_accession=25726562, orientation='minus', exon_count=1, OMIM=613499, no_hdr0=''),
  90550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90550, CurrentID=0, Status='live', Symbol='MCU', Aliases='C10orf42, CCDC109A', description='mitochondrial calcium uniporter', other_designations='coiled-coil domain-containing protein 109A', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72692131, end_position_on_the_genomic_accession=72887694, orientation='plus', exon_count=12, OMIM=614197, no_hdr0=''),
  100421898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421898, CurrentID=0, Status='live', Symbol='LOC100421898', Aliases='', description='olfactory receptor, family 2, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  90557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90557, CurrentID=0, Status='live', Symbol='CCDC74A', Aliases='', description='coiled-coil domain containing 74A', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131522641, end_position_on_the_genomic_accession=131533670, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375137, CurrentID=0, Status='live', Symbol='LOC105375137', Aliases='', description='uncharacterized LOC105375137', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6950450, end_position_on_the_genomic_accession=6955417, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129844, CurrentID=0, Status='live', Symbol='LOC100129844', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 7 pseudogene', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139338182, end_position_on_the_genomic_accession=139339642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375138, CurrentID=0, Status='live', Symbol='LOC105375138', Aliases='', description='uncharacterized LOC105375138', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7094962, end_position_on_the_genomic_accession=7109629, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90576, CurrentID=0, Status='live', Symbol='ZNF799', Aliases='HIT-40, ZNF842', description='zinc finger protein 799', other_designations='protein for MGC71805|zinc finger protein 14|zinc finger protein 842|zinc finger protein HIT-40', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12390014, end_position_on_the_genomic_accession=12413983, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  90580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90580, CurrentID=0, Status='live', Symbol='C19orf52', Aliases='', description='chromosome 19 open reading frame 52', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10928748, end_position_on_the_genomic_accession=10930241, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221656, CurrentID=0, Status='live', Symbol='KDM1B', Aliases='AOF1, C6orf193, LSD2, bA204B7.3, dJ298J15.2', description='lysine (K)-specific demethylase 1B', other_designations='amine oxidase (flavin containing) domain 1|amine oxidase, flavin containing 1|flavin-containing amine oxidase domain-containing protein 1|lysine-specific histone demethylase 2', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18155388, end_position_on_the_genomic_accession=18223854, orientation='plus', exon_count=22, OMIM=613081, no_hdr0=''),
  100132260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132260, CurrentID=0, Status='live', Symbol='LOC100132260', Aliases='', description='chromodomain protein, Y-like pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16692748, end_position_on_the_genomic_accession=16693432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90586, CurrentID=0, Status='live', Symbol='AOC4P', Aliases='AOC4', description='amine oxidase, copper containing 4, pseudogene', other_designations='amine oxidase pseudogene|amine oxidase, copper containing 3 (vascular adhesion protein 1) pseudogene', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42867145, end_position_on_the_genomic_accession=42869217, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  90589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90589, CurrentID=0, Status='live', Symbol='ZNF625', Aliases='', description='zinc finger protein 625', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12144890, end_position_on_the_genomic_accession=12156731, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  221662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221662, CurrentID=0, Status='live', Symbol='RBM24', Aliases='RNPC6, dJ259A10.1', description='RNA binding motif protein 24', other_designations='RNA-binding motif protein 24|RNA-binding region (RNP1, RRM) containing 6|RNA-binding region-containing protein 6', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17281492, end_position_on_the_genomic_accession=17293875, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  90592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90592, CurrentID=0, Status='live', Symbol='ZNF700', Aliases='', description='zinc finger protein 700', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11925068, end_position_on_the_genomic_accession=11950773, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130682, CurrentID=0, Status='live', Symbol='LOC100130682', Aliases='', description='ubiquitin specific peptidase 12 pseudogene', other_designations='', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148295559, end_position_on_the_genomic_accession=148296679, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90594, CurrentID=0, Status='live', Symbol='ZNF439', Aliases='', description='zinc finger protein 439', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11848530, end_position_on_the_genomic_accession=11869491, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101409253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409253, CurrentID=0, Status='live', Symbol='LINC00676', Aliases='', description='long intergenic non-protein coding RNA 676', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109728274, end_position_on_the_genomic_accession=109730034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101409254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409254, CurrentID=0, Status='live', Symbol='LINC00681', Aliases='', description='long intergenic non-protein coding RNA 681', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12794243, end_position_on_the_genomic_accession=12818291, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101409255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409255, CurrentID=0, Status='live', Symbol='LOC101409255', Aliases='', description='cell division cycle 42 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55877919, end_position_on_the_genomic_accession=55880747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101409256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409256, CurrentID=0, Status='live', Symbol='LOC101409256', Aliases='', description='cell division cycle 42 pseudogene', other_designations='', map_location='7q21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90396292, end_position_on_the_genomic_accession=90398236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101409257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409257, CurrentID=0, Status='live', Symbol='EZR-AS1', Aliases='', description='EZR antisense RNA 1', other_designations='EZR antisense RNA 1 (non-protein coding)', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158817979, end_position_on_the_genomic_accession=158822239, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101409258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409258, CurrentID=0, Status='live', Symbol='OSGEPL1-AS1', Aliases='', description='OSGEPL1 antisense RNA 1', other_designations='OSGEPL1 antisense RNA 1 (non-protein coding)', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189762780, end_position_on_the_genomic_accession=189765556, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101409261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409261, CurrentID=0, Status='live', Symbol='OGFR-AS1', Aliases='', description='OGFR antisense RNA 1', other_designations='OGFR antisense RNA 1 (non-protein coding)', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62800627, end_position_on_the_genomic_accession=62805587, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101409262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409262, CurrentID=0, Status='live', Symbol='LOC101409262', Aliases='', description='cell division cycle 42 pseudogene', other_designations='', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38268022, end_position_on_the_genomic_accession=38270881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101409265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409265, CurrentID=0, Status='live', Symbol='GLM7', Aliases='', description='Glioma susceptibility 7', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613032, no_hdr0=''),
  101409266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409266, CurrentID=0, Status='live', Symbol='MRT35', Aliases='', description='Mental retardation, autosomal recessive 35', other_designations='', map_location='17q21.31-q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615162, no_hdr0=''),
  101409267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101409267, CurrentID=0, Status='live', Symbol='CORD17', Aliases='', description='cone rod dystrophy 17 (autosomal dominant)', other_designations='Cone-rod dystrophy 17', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615163, no_hdr0=''),
  221687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221687, CurrentID=0, Status='live', Symbol='RNF182', Aliases='', description='ring finger protein 182', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13924446, end_position_on_the_genomic_accession=13980009, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  221692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221692, CurrentID=0, Status='live', Symbol='PHACTR1', Aliases='RPEL, RPEL1, dJ257A7.2', description='phosphatase and actin regulator 1', other_designations='RPEL repeat containing 1', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12716606, end_position_on_the_genomic_accession=13290244, orientation='plus', exon_count=20, OMIM=608723, no_hdr0=''),
  90624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90624, CurrentID=0, Status='live', Symbol='LYRM7', Aliases='C5orf31, MC3DN8, MZM1L', description='LYR motif containing 7', other_designations='LYR motif-containing protein 7|Lyrm7 homolog', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131170914, end_position_on_the_genomic_accession=131205426, orientation='plus', exon_count=5, OMIM=615831, no_hdr0=''),
  90625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90625, CurrentID=0, Status='live', Symbol='ERVH48-1', Aliases='C21orf105, NDUFV3-AS1', description='endogenous retrovirus group 48 member 1', other_designations='NDUFV3 antisense RNA 1 (non-protein coding)|endogenous retrovirus group 48, member 1|endogenous retrovirus group Fb member 1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42916803, end_position_on_the_genomic_accession=42919021, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  90627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90627, CurrentID=0, Status='live', Symbol='STARD13', Aliases='ARHGAP37, DLC2, GT650, LINC00464', description='StAR related lipid transfer domain containing 13', other_designations='Rho GTPase activating protein on chromosome 13q12|StAR-related lipid transfer (START) domain containing 13|StAR-related lipid transfer domain containing 13|deleted in liver cancer 2 protein|long intergenic non-protein coding RNA 464', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33103135, end_position_on_the_genomic_accession=33676835, orientation='minus', exon_count=25, OMIM=609866, no_hdr0=''),
  90632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90632, CurrentID=0, Status='live', Symbol='LINC00473', Aliases='C6orf176, bA142J11.1', description='long intergenic non-protein coding RNA 473', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165924048, end_position_on_the_genomic_accession=165988039, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375148, CurrentID=0, Status='live', Symbol='LOC105375148', Aliases='', description='uncharacterized LOC105375148', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9734270, end_position_on_the_genomic_accession=9769513, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  90634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90634, CurrentID=0, Status='live', Symbol='N4BP2L1', Aliases='CG018', description='NEDD4 binding protein 2-like 1', other_designations='hypothetical gene CG018', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32400723, end_position_on_the_genomic_accession=32429614, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  90637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90637, CurrentID=0, Status='live', Symbol='ZFAND2A', Aliases='AIRAP', description='zinc finger, AN1-type domain 2A', other_designations='arsenite inducible RNA associated protein', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1152907, end_position_on_the_genomic_accession=1160219, orientation='minus', exon_count=5, OMIM=610699, no_hdr0=''),
  221710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221710, CurrentID=0, Status='live', Symbol='SMIM13', Aliases='C6orf228', description='small integral membrane protein 13', other_designations='UPF0766 protein C6orf228', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11094033, end_position_on_the_genomic_accession=11138738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221711, CurrentID=0, Status='live', Symbol='SYCP2L', Aliases='C6orf177, NO145, dJ62D2.1', description='synaptonemal complex protein 2 like', other_designations='145 kDa nucleolar protein homolog|SCP-2-like', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10886831, end_position_on_the_genomic_accession=10974309, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  221717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221717, CurrentID=0, Status='live', Symbol='MRPL48P1', Aliases='MRPL48L1, dJ290I10.4', description='mitochondrial ribosomal protein L48 pseudogene 1', other_designations='mitochondrial ribosomal protein L48-like 1', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10459677, end_position_on_the_genomic_accession=10460589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  221718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221718, CurrentID=0, Status='live', Symbol='LINC00518', Aliases='C6orf218', description='long intergenic non-protein coding RNA 518', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10427785, end_position_on_the_genomic_accession=10434822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90649, CurrentID=0, Status='live', Symbol='ZNF486', Aliases='KRBO2', description='zinc finger protein 486', other_designations='KRAB box only protein 2|KRAB domain only 2|KRAB domain only protein 2', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20167214, end_position_on_the_genomic_accession=20200490, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  90655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90655, CurrentID=0, Status='live', Symbol='TGIF2LY', Aliases='TGIFLY', description='TGFB-induced factor homeobox 2-like, Y-linked', other_designations='TGF-beta-induced transcription factor 2-like protein|TGFB-induced factor 2-like protein, Y-linked|TGFB-induced factor 2-like, Y-linked|TGIF-like on the Y', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3579085, end_position_on_the_genomic_accession=3580041, orientation='plus', exon_count=2, OMIM=400025, no_hdr0=''),
  100132272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132272, CurrentID=0, Status='live', Symbol='LOC100132272', Aliases='', description='uncharacterized LOC100132272', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=42746927, end_position_on_the_genomic_accession=42749125, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375153, CurrentID=0, Status='live', Symbol='LOC105375153', Aliases='', description='uncharacterized LOC105375153', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11320754, end_position_on_the_genomic_accession=11385609, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  90665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90665, CurrentID=0, Status='live', Symbol='TBL1Y', Aliases='TBL1', description='transducin (beta)-like 1, Y-linked', other_designations='transducin beta-like 1|transducin beta-like protein 1Y|transducin-beta-like protein 1, Y-linked', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6910686, end_position_on_the_genomic_accession=7091749, orientation='plus', exon_count=19, OMIM=400033, no_hdr0=''),
  90668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90668, CurrentID=0, Status='live', Symbol='LRRC16B', Aliases='C14orf121, CARMIL3, crml-1', description='leucine rich repeat containing 16B', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24051997, end_position_on_the_genomic_accession=24069728, orientation='plus', exon_count=41, OMIM=614716, no_hdr0=''),
  90673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90673, CurrentID=0, Status='live', Symbol='PPP1R3E', Aliases='', description='protein phosphatase 1 regulatory subunit 3E', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 3E|protein phosphatase 1, regulatory subunit 3E', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23295921, end_position_on_the_genomic_accession=23302848, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  255411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255411, CurrentID=0, Status='live', Symbol='LINC00935', Aliases='', description='long intergenic non-protein coding RNA 935', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48727435, end_position_on_the_genomic_accession=48765786, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  221749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221749, CurrentID=0, Status='live', Symbol='PXDC1', Aliases='C6orf145', description='PX domain containing 1', other_designations='PX domain-containing protein C6orf145', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3722602, end_position_on_the_genomic_accession=3752012, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  90678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90678, CurrentID=0, Status='live', Symbol='LRSAM1', Aliases='CMT2P, RIFLE, TAL', description='leucine rich repeat and sterile alpha motif containing 1', other_designations='Tsg101-associated ligase', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127451486, end_position_on_the_genomic_accession=127503501, orientation='plus', exon_count=26, OMIM=610933, no_hdr0=''),
  102725031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725031, CurrentID=0, Status='live', Symbol='LOC102725031', Aliases='', description='leukocyte immunoglobulin-like receptor subfamily B member 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  221756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221756, CurrentID=0, Status='live', Symbol='SERPINB9P1', Aliases='SERPINB9P', description='serpin peptidase inhibitor, clade B (ovalbumin), member 9, pseudogene 1', other_designations='serpin peptidase inhibitor, clade B (ovalbumin), member 9 pseudogene', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2854657, end_position_on_the_genomic_accession=2876510, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375633, CurrentID=0, Status='live', Symbol='LOC105375633', Aliases='', description='uncharacterized LOC105375633', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90690441, end_position_on_the_genomic_accession=90791632, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  90693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90693, CurrentID=0, Status='live', Symbol='CCDC126', Aliases='', description='coiled-coil domain containing 126', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23597379, end_position_on_the_genomic_accession=23644708, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927202, CurrentID=0, Status='live', Symbol='LOC101927202', Aliases='', description='uncharacterized LOC101927202', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39090731, end_position_on_the_genomic_accession=39092851, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  90701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90701, CurrentID=0, Status='live', Symbol='SEC11C', Aliases='SEC11L3, SPC21, SPCS4C', description='SEC11 homolog C, signal peptidase complex subunit', other_designations='SEC11 homolog C|SEC11-like 3|SEC11-like protein 3|SPase 21 kDa subunit|microsomal signal peptidase 21 kDa subunit', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59139884, end_position_on_the_genomic_accession=59158837, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100132280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132280, CurrentID=0, Status='live', Symbol='LOC100132280', Aliases='', description='ataxin 3 pseudogene', other_designations='', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=110555387, end_position_on_the_genomic_accession=110556525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  352852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352852, CurrentID=0, Status='live', Symbol='CYCSP45', Aliases='HCP45', description='cytochrome c, somatic pseudogene 45', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153841351, end_position_on_the_genomic_accession=153841468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  221785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221785, CurrentID=0, Status='live', Symbol='ZSCAN25', Aliases='ZNF498', description='zinc finger and SCAN domain containing 25', other_designations='zinc finger protein 498', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99616611, end_position_on_the_genomic_accession=99738530, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  221786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221786, CurrentID=0, Status='live', Symbol='FAM200A', Aliases='C7orf38', description='family with sequence similarity 200 member A', other_designations='family with sequence similarity 200, member A', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99546300, end_position_on_the_genomic_accession=99559395, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  221806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221806, CurrentID=0, Status='live', Symbol='VWDE', Aliases='', description='von Willebrand factor D and EGF domains', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12330883, end_position_on_the_genomic_accession=12404226, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  105375165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375165, CurrentID=0, Status='live', Symbol='LOC105375165', Aliases='', description='uncharacterized LOC105375165', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15237171, end_position_on_the_genomic_accession=15312750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  90736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90736, CurrentID=0, Status='live', Symbol='FAM104B', Aliases='CXorf44', description='family with sequence similarity 104 member B', other_designations='family with sequence similarity 104, member B', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55143102, end_position_on_the_genomic_accession=55161195, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  90737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90737, CurrentID=0, Status='live', Symbol='PAGE5', Aliases='CT16, CT16.1, CT16.2, GAGEE1, PAGE-5', description='PAGE family member 5', other_designations='P antigen family, member 5 (prostate associated)|cancer/testis antigen 16.1|cancer/testis antigen family 16, member 1|cancer/testis antigen family 16, member 2|g antigen family E member 1|prostate-associated gene 5 protein', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55220347, end_position_on_the_genomic_accession=55224108, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375166, CurrentID=0, Status='live', Symbol='LOC105375166', Aliases='', description='uncharacterized LOC105375166', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15603767, end_position_on_the_genomic_accession=15681980, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  221814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221814, CurrentID=0, Status='live', Symbol='LOC221814', Aliases='', description='uncharacterized LOC221814', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  352887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352887, CurrentID=0, Status='live', Symbol='GYG2P1', Aliases='GYG2P', description='glycogenin 2 pseudogene 1', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12406116, end_position_on_the_genomic_accession=12421590, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375167, CurrentID=0, Status='live', Symbol='LOC105375167', Aliases='', description='uncharacterized LOC105375167', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15735489, end_position_on_the_genomic_accession=15802995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  221823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221823, CurrentID=0, Status='live', Symbol='PRPS1L1', Aliases='PRPS1, PRPS3, PRPSL, PRS-III', description='phosphoribosyl pyrophosphate synthetase 1-like 1', other_designations='PRPS1-like 1|phosphoribosyl pyrophosphate synthase 1-like 1|phosphoribosyl pyrophosphate synthase III|phosphoribosylpyrophosphate synthetase subunit III|ribose-phosphate diphosphokinase catalytic chain III|ribose-phosphate pyrophosphokinase III', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18026774, end_position_on_the_genomic_accession=18027863, orientation='minus', exon_count=1, OMIM=611566, no_hdr0=''),
  352896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352896, CurrentID=0, Status='live', Symbol='GPR143P', Aliases='OA1P', description='G protein-coupled receptor 143 pseudogene', other_designations='ocular albinism 1 (Nettleship-Falls) Y-linked pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7100398, end_position_on_the_genomic_accession=7106505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375168, CurrentID=0, Status='live', Symbol='LOC105375168', Aliases='', description='uncharacterized LOC105375168', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16460200, end_position_on_the_genomic_accession=16471809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221830, CurrentID=0, Status='live', Symbol='TWISTNB', Aliases='', description='TWIST neighbor', other_designations='twist neighbor protein', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19695462, end_position_on_the_genomic_accession=19709037, orientation='minus', exon_count=4, OMIM=608312, no_hdr0=''),
  105375169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375169, CurrentID=0, Status='live', Symbol='LOC105375169', Aliases='', description='uncharacterized LOC105375169', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16604824, end_position_on_the_genomic_accession=16628475, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  221833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221833, CurrentID=0, Status='live', Symbol='SP8', Aliases='BTD', description='Sp8 transcription factor', other_designations='specificity protein 8', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20781467, end_position_on_the_genomic_accession=20786889, orientation='minus', exon_count=4, OMIM=608306, no_hdr0=''),
  352909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352909, CurrentID=0, Status='live', Symbol='DNAAF3', Aliases='C19orf51, CILD2, DAB1, PCD, PF22', description='dynein, axonemal, assembly factor 3', other_designations='UPF0470 protein C19orf51', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55158661, end_position_on_the_genomic_accession=55166722, orientation='minus', exon_count=12, OMIM=614566, no_hdr0=''),
  221838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221838, CurrentID=0, Status='live', Symbol='EEF1A1P6', Aliases='EEF1AL4', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 6', other_designations='eukaryotic translation elongation factor 1 alpha-like 4', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22510305, end_position_on_the_genomic_accession=22512050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90768, CurrentID=0, Status='live', Symbol='LOC90768', Aliases='', description='uncharacterized LOC90768', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182138658, end_position_on_the_genomic_accession=182144515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375171, CurrentID=0, Status='live', Symbol='LOC105375171', Aliases='', description='uncharacterized LOC105375171', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17362445, end_position_on_the_genomic_accession=17371846, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128159, CurrentID=0, Status='live', Symbol='ATF1P1', Aliases='', description='activating transcription factor 1 pseudogene 1', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92887267, end_position_on_the_genomic_accession=92888058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90780, CurrentID=0, Status='live', Symbol='PYGO2', Aliases='1190004M21Rik', description='pygopus family PHD finger 2', other_designations='pygopus 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154957026, end_position_on_the_genomic_accession=154961782, orientation='minus', exon_count=3, OMIM=606903, no_hdr0=''),
  90784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90784, CurrentID=0, Status='live', Symbol='LOC90784', Aliases='', description='uncharacterized LOC90784', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86020216, end_position_on_the_genomic_accession=86023868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132295, CurrentID=0, Status='live', Symbol='PPP1R26P4', Aliases='KIAA0649P4', description='protein phosphatase 1 regulatory subunit 26 pseudogene 4', other_designations='protein phosphatase 1, regulatory subunit 26 pseudogene 4', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18715620, end_position_on_the_genomic_accession=18719643, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  90799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90799, CurrentID=0, Status='live', Symbol='CEP95', Aliases='CCDC45', description='centrosomal protein 95kDa', other_designations='coiled-coil domain containing 45|coiled-coil domain-containing protein 45', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64506762, end_position_on_the_genomic_accession=64537951, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  90806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90806, CurrentID=0, Status='live', Symbol='ANGEL2', Aliases='Ccr4d, KIAA0759L', description='angel homolog 2 (Drosophila)', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212992182, end_position_on_the_genomic_accession=213015875, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  90809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90809, CurrentID=0, Status='live', Symbol='TMEM55B', Aliases='C14orf9', description='transmembrane protein 55B', other_designations='ptdIns-4,5-P(2) 4-phosphatase type I|ptdIns-4,5-P2 4-Ptase I|type 1 PtdIns-4,5-P2 4-Ptase', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20457853, end_position_on_the_genomic_accession=20461612, orientation='minus', exon_count=7, OMIM=609865, no_hdr0=''),
  352954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352954, CurrentID=0, Status='live', Symbol='GATS', Aliases='STAG3OS', description='GATS, stromal antigen 3 opposite strand', other_designations='STAG3 opposite strand transcript protein|opposite strand transcription unit to STAG3|stromal antigen 3 opposite strand', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100200653, end_position_on_the_genomic_accession=100272232, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  221883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221883, CurrentID=0, Status='live', Symbol='HOXA11-AS', Aliases='HOXA-AS51, HOXA11AS, HOXA11S, NCRNA00076, HOXA11-AS', description='HOXA11 antisense RNA', other_designations='HOXA cluster antisense RNA 5 (non-protein coding)|HOXA11 antisense RNA (non-protein coding)|HOXA11 antisense RNA 1 (non-protein coding)|homeo box A11, antisense|homeo box A11, opposite strand transcript|homeobox A11 antisense', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27185408, end_position_on_the_genomic_accession=27189293, orientation='plus', exon_count=2, OMIM=607530, no_hdr0=''),
  352956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352956, CurrentID=0, Status='live', Symbol='MMVP2', Aliases='', description='myxomatous mitral valve prolapse 2', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607829, no_hdr0=''),
  352957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352957, CurrentID=0, Status='live', Symbol='MICF', Aliases='', description='MHC class I polypeptide-related sequence F (pseudogene)', other_designations='MHC class I polypeptide-related sequence F pseudogene|MHC class I polypeptide-related sequence F, pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29852187, end_position_on_the_genomic_accession=29854052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  352961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352961, CurrentID=0, Status='live', Symbol='HCG26', Aliases='3.8-1, NCRNA00191, bCX205D4.4, bPG181B23.4', description='HLA complex group 26 (non-protein coding)', other_designations='MHC class I mRNA fragment 3.8-1', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31471229, end_position_on_the_genomic_accession=31472408, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  352962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352962, CurrentID=0, Status='live', Symbol='HLA-V', Aliases='HLA-75, dJ377H14.4', description='major histocompatibility complex, class I, V (pseudogene)', other_designations='HLA-75 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29791753, end_position_on_the_genomic_accession=29792750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  352963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352963, CurrentID=0, Status='live', Symbol='HLA-P', Aliases='C6orf101, HLA-90, dJ377H14.3', description='major histocompatibility complex, class I, P (pseudogene)', other_designations='HLA-90 pseudogene', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29800044, end_position_on_the_genomic_accession=29803079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  352964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352964, CurrentID=0, Status='live', Symbol='HLA-T', Aliases='HLA-16', description='major histocompatibility complex, class I, T (pseudogene)', other_designations='HLA-16 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29896443, end_position_on_the_genomic_accession=29898947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  352965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352965, CurrentID=0, Status='live', Symbol='HLA-U', Aliases='HLA-21', description='major histocompatibility complex, class I, U (pseudogene)', other_designations='HLA-21 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29933764, end_position_on_the_genomic_accession=29934880, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  352966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352966, CurrentID=0, Status='live', Symbol='HLA-W', Aliases='HLA-80', description='major histocompatibility complex, class I, W (pseudogene)', other_designations='HLA-80 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29955834, end_position_on_the_genomic_accession=29959058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  221895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221895, CurrentID=0, Status='live', Symbol='JAZF1', Aliases='TIP27, ZNF802', description='JAZF zinc finger 1', other_designations='TAK1-interacting protein 27|juxtaposed with another zinc finger gene 1|zinc finger protein 802', map_location='7p15.2-p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27830573, end_position_on_the_genomic_accession=28180818, orientation='minus', exon_count=7, OMIM=606246, no_hdr0=''),
  105371273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371273, CurrentID=0, Status='live', Symbol='LOC105371273', Aliases='', description='uncharacterized LOC105371273', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54545728, end_position_on_the_genomic_accession=54559078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375180, CurrentID=0, Status='live', Symbol='LOC105375180', Aliases='', description='uncharacterized LOC105375180', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19589165, end_position_on_the_genomic_accession=19668277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  90826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90826, CurrentID=0, Status='live', Symbol='PRMT9', Aliases='PRMT10', description='protein arginine methyltransferase 9', other_designations='protein arginine methyltransferase 10 (putative)|putative protein arginine N-methyltransferase 10', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147637785, end_position_on_the_genomic_accession=147684230, orientation='minus', exon_count=12, OMIM=616125, no_hdr0=''),
  90827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90827, CurrentID=0, Status='live', Symbol='ZNF479', Aliases='HKr19, KR19', description='zinc finger protein 479', other_designations='KRAB zinc finger protein KR19|zinc finger protein Kr19', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57119574, end_position_on_the_genomic_accession=57142003, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  90835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90835, CurrentID=0, Status='live', Symbol='CCDC189', Aliases='C16orf93', description='coiled-coil domain containing 189', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30757423, end_position_on_the_genomic_accession=30762277, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  221908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221908, CurrentID=0, Status='live', Symbol='PPP1R35', Aliases='C7orf47', description='protein phosphatase 1 regulatory subunit 35', other_designations='UPF0683 protein C7orf47|protein phosphatase 1, regulatory subunit 35', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100435277, end_position_on_the_genomic_accession=100436615, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  221914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221914, CurrentID=0, Status='live', Symbol='GPC2', Aliases='', description='glypican 2', other_designations='cerebroglycan proteoglycan|glypican proteoglycan 2, cerebroglycan proteoglycan', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100169606, end_position_on_the_genomic_accession=100177426, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  90843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90843, CurrentID=0, Status='live', Symbol='TCEAL8', Aliases='WEX3', description='transcription elongation factor A (SII)-like 8', other_designations='TCEA-like protein 8|transcription elongation factor S-II protein-like 8', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103252995, end_position_on_the_genomic_accession=103255193, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  352990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352990, CurrentID=0, Status='live', Symbol='HCP5B', Aliases='HCP5P10, P5-10, P5.9', description='HLA complex P5B (non-protein coding)', other_designations='HLA complex P5 pseudogene 10|P5-1 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375184, CurrentID=0, Status='live', Symbol='LOC105375184', Aliases='', description='uncharacterized LOC105375184', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22214000, end_position_on_the_genomic_accession=22215603, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  90850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90850, CurrentID=0, Status='live', Symbol='ZNF598', Aliases='', description='zinc finger protein 598', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1997652, end_position_on_the_genomic_accession=2009821, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  90853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90853, CurrentID=0, Status='live', Symbol='SPOCD1', Aliases='PPP1R146', description='SPOC domain containing 1', other_designations='protein phosphatase 1, regulatory subunit 146', map_location='1p35.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31790422, end_position_on_the_genomic_accession=31816051, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  352999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=352999, CurrentID=0, Status='live', Symbol='C6orf58', Aliases='LEG1', description='chromosome 6 open reading frame 58', other_designations='UPF0762 protein C6orf58', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127577174, end_position_on_the_genomic_accession=127591818, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  353000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353000, CurrentID=0, Status='live', Symbol='HCGVIII-2', Aliases='', description='HCGVIII-2 pseudogene', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29833588, end_position_on_the_genomic_accession=29835128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353001, CurrentID=0, Status='live', Symbol='HCG4P3', Aliases='HCGIV-03, HCGIV-3, HCGIV.2', description='HLA complex group 4 pseudogene 3', other_designations='HCGIV-03 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30004845, end_position_on_the_genomic_accession=30005828, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353002, CurrentID=0, Status='live', Symbol='HCG4P4', Aliases='HCGIV-04, HCGIV-4, HCGIV.3', description='HLA complex group 4 pseudogene 4', other_designations='HCGIV-04 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29955205, end_position_on_the_genomic_accession=29955633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90859, CurrentID=0, Status='live', Symbol='NT5C3AP1', Aliases='NT5C3P1, UMPH1P', description="5'-nucleotidase, cytosolic IIIA pseudogene 1", other_designations="5'-nucleotidase, cytosolic III pseudogene 1|uridine 5'-monophosphate hydrolase 1 pseudogene", map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117574489, end_position_on_the_genomic_accession=117576186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353004, CurrentID=0, Status='live', Symbol='HCG4P7', Aliases='HCGIV-07, HCGIV-7, HCGIV.C', description='HLA complex group 4 pseudogene 7', other_designations='HCGIV-07 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29886110, end_position_on_the_genomic_accession=29887007, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90861, CurrentID=0, Status='live', Symbol='HN1L', Aliases='C16orf34, L11', description='hematological and neurological expressed 1-like', other_designations='CRAMP_1 like|HN1 like|HN1-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1678277, end_position_on_the_genomic_accession=1702072, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  221935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221935, CurrentID=0, Status='live', Symbol='SDK1', Aliases='', description='sidekick cell adhesion molecule 1', other_designations='sidekick homolog 1, cell adhesion molecule', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3301448, end_position_on_the_genomic_accession=4269000, orientation='plus', exon_count=47, OMIM=607216, no_hdr0=''),
  90864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90864, CurrentID=0, Status='live', Symbol='SPSB3', Aliases='C16orf31, SSB3', description='splA/ryanodine receptor domain and SOCS box containing 3', other_designations='SPRY domain-containing SOCS box protein SSB-3', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1776712, end_position_on_the_genomic_accession=1782605, orientation='minus', exon_count=9, OMIM=611659, no_hdr0=''),
  221937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221937, CurrentID=0, Status='live', Symbol='FOXK1', Aliases='FOXK1L', description='forkhead box K1', other_designations='MNF|myocyte nuclear factor', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4644691, end_position_on_the_genomic_accession=4771443, orientation='plus', exon_count=12, OMIM=616302, no_hdr0=''),
  221938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221938, CurrentID=0, Status='live', Symbol='MMD2', Aliases='PAQR10', description='monocyte to macrophage differentiation-associated 2', other_designations='monocyte-to-macrophage differentiation factor 2|monocyte-to-macrophage differentiation-associated protein 2|progestin and adipoQ receptor family member 10|progestin and adipoQ receptor family member X', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4892245, end_position_on_the_genomic_accession=4959213, orientation='minus', exon_count=10, OMIM=614581, no_hdr0=''),
  353013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353013, CurrentID=0, Status='live', Symbol='RPL7AP7', Aliases='RPL7A_13_677, RPL7B, dJ377H14.2', description='ribosomal protein L7a pseudogene 7', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29803133, end_position_on_the_genomic_accession=29804020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353014, CurrentID=0, Status='live', Symbol='HCG4P9', Aliases='HCGIV-09, HCGIV-9', description='HLA complex group 4 pseudogene 9', other_designations='HCGIV-09 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29798424, end_position_on_the_genomic_accession=29800126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  90871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90871, CurrentID=0, Status='live', Symbol='TMEM261', Aliases='C9orf123', description='transmembrane protein 261', other_designations='transmembrane protein C9orf123', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7796490, end_position_on_the_genomic_accession=7799817, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375188, CurrentID=0, Status='live', Symbol='LOC105375188', Aliases='', description='uncharacterized LOC105375188', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24017156, end_position_on_the_genomic_accession=24027698, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  90874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90874, CurrentID=0, Status='live', Symbol='ZNF697', Aliases='', description='zinc finger protein 697', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119619377, end_position_on_the_genomic_accession=119648170, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  353019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353019, CurrentID=0, Status='live', Symbol='HCG9P5', Aliases='HCGIX-5, dJ377H14.6', description='HLA complex group 9 pseudogene 5', other_designations='HCGIX-5 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29748254, end_position_on_the_genomic_accession=29748648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353020, CurrentID=0, Status='live', Symbol='HCG4P11', Aliases='HCGIV-11, HCGIV.10', description='HLA complex group 4 pseudogene 11', other_designations='HCGIV-11 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29721178, end_position_on_the_genomic_accession=29722167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929159, CurrentID=0, Status='live', Symbol='LOC101929159', Aliases='', description='uncharacterized LOC101929159', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57693068, end_position_on_the_genomic_accession=57696651, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221955, CurrentID=0, Status='live', Symbol='DAGLB', Aliases='DAGLBETA, KCCR13L', description='diacylglycerol lipase, beta', other_designations='DGL-beta', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6409116, end_position_on_the_genomic_accession=6448012, orientation='minus', exon_count=15, OMIM=614016, no_hdr0=''),
  105375190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375190, CurrentID=0, Status='live', Symbol='LOC105375190', Aliases='', description='uncharacterized LOC105375190', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24760272, end_position_on_the_genomic_accession=24762168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  221960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221960, CurrentID=0, Status='live', Symbol='CCZ1B', Aliases='C7orf28B, H_NH0577018.2', description='CCZ1 homolog B, vacuolar protein trafficking and biogenesis associated', other_designations='CCZ1 homolog, vacuolar protein trafficking and biogenesis associated B|CCZ1 vacuolar protein trafficking and biogenesis associated B|CCZ1 vacuolar protein trafficking and biogenesis associated homolog B|H_DJ1163J12.2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6798935, end_position_on_the_genomic_accession=6826295, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  106783499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106783499, CurrentID=0, Status='live', Symbol='OPA8', Aliases='', description='optic atrophy 8 (autosomal dominant)', other_designations='Optic atroph 8', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616648, no_hdr0=''),
  105372743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372743, CurrentID=0, Status='live', Symbol='LOC105372743', Aliases='', description='uncharacterized LOC105372743', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17803428, end_position_on_the_genomic_accession=17812311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  221981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221981, CurrentID=0, Status='live', Symbol='THSD7A', Aliases='', description='thrombospondin type 1 domain containing 7A', other_designations='thrombospondin, type I, domain containing 7A', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11370435, end_position_on_the_genomic_accession=11832198, orientation='minus', exon_count=30, OMIM=612249, no_hdr0=''),
  101927670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927670, CurrentID=0, Status='live', Symbol='LOC101927670', Aliases='', description='uncharacterized LOC101927670', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131450055, end_position_on_the_genomic_accession=131466441, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  90933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90933, CurrentID=0, Status='live', Symbol='TRIM41', Aliases='RINCK', description='tripartite motif containing 41', other_designations='RING finger-interacting protein with C kinase|RING-finger protein that interacts with C kinase', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181223263, end_position_on_the_genomic_accession=181235808, orientation='plus', exon_count=6, OMIM=610530, no_hdr0=''),
  222008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222008, CurrentID=0, Status='live', Symbol='VSTM2A', Aliases='VSTM2', description='V-set and transmembrane domain containing 2A', other_designations='V-set and transmembrane domain containing 2|hypothetical protein MGC33530', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54542325, end_position_on_the_genomic_accession=54571080, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375199, CurrentID=0, Status='live', Symbol='LOC105375199', Aliases='', description='uncharacterized LOC105375199', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25939087, end_position_on_the_genomic_accession=26130930, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  353088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353088, CurrentID=0, Status='live', Symbol='ZNF429', Aliases='', description='zinc finger protein 429', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21505635, end_position_on_the_genomic_accession=21538277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353089, CurrentID=0, Status='live', Symbol='RAET1F', Aliases='', description='retinoic acid early transcript 1F pseudogene', other_designations='', map_location='6q24.1-q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149910028, end_position_on_the_genomic_accession=149910490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353091, CurrentID=0, Status='live', Symbol='RAET1G', Aliases='ULBP5', description='retinoic acid early transcript 1G', other_designations='UL-16 binding protein 5', map_location='6q24.1-q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149916878, end_position_on_the_genomic_accession=149923152, orientation='minus', exon_count=5, OMIM=609244, no_hdr0=''),
  353094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353094, CurrentID=0, Status='live', Symbol='RAET1M', Aliases='', description='retinoic acid early transcript 1M pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150032769, end_position_on_the_genomic_accession=150033566, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375201, CurrentID=0, Status='live', Symbol='LOC105375201', Aliases='', description='uncharacterized LOC105375201', other_designations='CTB-119C2.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26171511, end_position_on_the_genomic_accession=26185286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  90952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90952, CurrentID=0, Status='live', Symbol='ESAM', Aliases='W117m', description='endothelial cell adhesion molecule', other_designations='2310008D05Rik|HUEL (C4orf1)-interacting protein|LP4791 protein', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124753123, end_position_on_the_genomic_accession=124762327, orientation='minus', exon_count=7, OMIM=614281, no_hdr0=''),
  90956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90956, CurrentID=0, Status='live', Symbol='ADCK2', Aliases='AARF', description='aarF domain containing kinase 2', other_designations='putative ubiquinone biosynthesis protein AarF', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140672925, end_position_on_the_genomic_accession=140695111, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  90957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=90957, CurrentID=0, Status='live', Symbol='DHX57', Aliases='DDX57', description='DEAH (Asp-Glu-Ala-Asp/His) box polypeptide 57', other_designations='DEAH box protein 57|DEAH-box RNA/DNA helicase AAM73547', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38797729, end_position_on_the_genomic_accession=38875883, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  105375203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375203, CurrentID=0, Status='live', Symbol='LOC105375203', Aliases='', description='uncharacterized LOC105375203', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26585561, end_position_on_the_genomic_accession=26598242, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375204, CurrentID=0, Status='live', Symbol='LOC105375204', Aliases='', description='uncharacterized LOC105375204', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26805511, end_position_on_the_genomic_accession=26858152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353116, CurrentID=0, Status='live', Symbol='RILPL1', Aliases='GOSPEL, RLP1', description='Rab interacting lysosomal protein-like 1', other_designations='rab-interacting lysosomal-like protein 1', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123471362, end_position_on_the_genomic_accession=123533718, orientation='minus', exon_count=15, OMIM=614092, no_hdr0=''),
  105375205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375205, CurrentID=0, Status='live', Symbol='LOC105375205', Aliases='', description='uncharacterized LOC105375205', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27114130, end_position_on_the_genomic_accession=27116665, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  353124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353124, CurrentID=0, Status='live', Symbol='EIG2', Aliases='EGI14', description='Epilepsy, idiopathic generalized, susceptibility to 2', other_designations='', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606972, no_hdr0=''),
  353125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353125, CurrentID=0, Status='live', Symbol='HDLC3', Aliases='', description='High density lipoprotein cholesterol, low serum, 3', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607687, no_hdr0=''),
  353126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353126, CurrentID=0, Status='live', Symbol='BMIQ10', Aliases='OB10Q', description='Obesity, susceptibility to, on chromosome 10q', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607514, no_hdr0=''),
  353127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353127, CurrentID=0, Status='live', Symbol='HDLCQ2', Aliases='HDLC2', description='High density lipoprotein cholesterol level QTL on chromosome 8', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607053, no_hdr0=''),
  353128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353128, CurrentID=0, Status='live', Symbol='AD8', Aliases='', description='Alzheimer disease 8', other_designations='', map_location='20p', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607116, no_hdr0=''),
  353131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353131, CurrentID=0, Status='live', Symbol='LCE1A', Aliases='LEP1', description='late cornified envelope 1A', other_designations='late envelope protein 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152827473, end_position_on_the_genomic_accession=152828097, orientation='plus', exon_count=1, OMIM=612603, no_hdr0=''),
  353132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353132, CurrentID=0, Status='live', Symbol='LCE1B', Aliases='LEP2, SPRL2A', description='late cornified envelope 1B', other_designations='late envelope protein 2|small proline rich-like (epidermal differentiation complex) 2A|small proline-rich-like epidermal differentiation complex protein 2A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152811971, end_position_on_the_genomic_accession=152813109, orientation='plus', exon_count=1, OMIM=612604, no_hdr0=''),
  353133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353133, CurrentID=0, Status='live', Symbol='LCE1C', Aliases='LEP3', description='late cornified envelope 1C', other_designations='late envelope protein 3', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152804835, end_position_on_the_genomic_accession=152806631, orientation='minus', exon_count=3, OMIM=612605, no_hdr0=''),
  353134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353134, CurrentID=0, Status='live', Symbol='LCE1D', Aliases='LEP4', description='late cornified envelope 1D', other_designations='late envelope protein 4', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152796751, end_position_on_the_genomic_accession=152798181, orientation='plus', exon_count=2, OMIM=612606, no_hdr0=''),
  353135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353135, CurrentID=0, Status='live', Symbol='LCE1E', Aliases='LEP5', description='late cornified envelope 1E', other_designations='late envelope protein 5', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152786277, end_position_on_the_genomic_accession=152788426, orientation='plus', exon_count=2, OMIM=612607, no_hdr0=''),
  353137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353137, CurrentID=0, Status='live', Symbol='LCE1F', Aliases='LEP6', description='late cornified envelope 1F', other_designations='late envelope protein 6', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152776372, end_position_on_the_genomic_accession=152776969, orientation='plus', exon_count=1, OMIM=612608, no_hdr0=''),
  353138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353138, CurrentID=0, Status='live', Symbol='LEP7', Aliases='', description='late envelope protein 7', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  353139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353139, CurrentID=0, Status='live', Symbol='LCE2A', Aliases='LEP9', description='late cornified envelope 2A', other_designations='late envelope protein 9', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152698364, end_position_on_the_genomic_accession=152699442, orientation='plus', exon_count=2, OMIM=612609, no_hdr0=''),
  353140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353140, CurrentID=0, Status='live', Symbol='LCE2C', Aliases='LEP11', description='late cornified envelope 2C', other_designations='late envelope protein 11', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152675315, end_position_on_the_genomic_accession=152676574, orientation='plus', exon_count=2, OMIM=612611, no_hdr0=''),
  353141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353141, CurrentID=0, Status='live', Symbol='LCE2D', Aliases='LEP12, SPRL1A', description='late cornified envelope 2D', other_designations='late envelope protein 12|small proline rich-like (epidermal differentiation complex) 1A|small proline-rich-like epidermal differentiation complex protein 1A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152663411, end_position_on_the_genomic_accession=152664659, orientation='plus', exon_count=2, OMIM=612612, no_hdr0=''),
  353142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353142, CurrentID=0, Status='live', Symbol='LCE3A', Aliases='LEP13', description='late cornified envelope 3A', other_designations='late envelope protein 13', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152622834, end_position_on_the_genomic_accession=152623103, orientation='minus', exon_count=1, OMIM=612613, no_hdr0=''),
  353143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353143, CurrentID=0, Status='live', Symbol='LCE3B', Aliases='LEP14', description='late cornified envelope 3B', other_designations='late envelope protein 14', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152613811, end_position_on_the_genomic_accession=152614098, orientation='plus', exon_count=1, OMIM=612614, no_hdr0=''),
  353144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353144, CurrentID=0, Status='live', Symbol='LCE3C', Aliases='LEP15, SPRL3A', description='late cornified envelope 3C', other_designations='late envelope protein 15|small proline rich-like (epidermal differentiation complex) 3A|small proline-rich-like epidermal differentiation complex protein 3A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152600662, end_position_on_the_genomic_accession=152601086, orientation='plus', exon_count=1, OMIM=612615, no_hdr0=''),
  353145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353145, CurrentID=0, Status='live', Symbol='LCE3E', Aliases='LEP17', description='late cornified envelope 3E', other_designations='late envelope protein 17', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152565654, end_position_on_the_genomic_accession=152566770, orientation='minus', exon_count=2, OMIM=612617, no_hdr0=''),
  353147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353147, CurrentID=0, Status='live', Symbol='POROK4', Aliases='DSAP2', description='disseminated superficial actinic porokeratosis 2', other_designations='', map_location='15q25.1-q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607728, no_hdr0=''),
  100421913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421913, CurrentID=0, Status='live', Symbol='LOC100421913', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  353149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353149, CurrentID=0, Status='live', Symbol='TBC1D26', Aliases='', description='TBC1 domain family member 26', other_designations='TBC1 domain family, member 26', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15732277, end_position_on_the_genomic_accession=15744784, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  91010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91010, CurrentID=0, Status='live', Symbol='FMNL3', Aliases='FHOD3, FRL2, WBP-3, WBP3', description='formin like 3', other_designations='WW domain binding protein 3|formin homology 2 domain-containing protein 3|formin-like 3', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49637941, end_position_on_the_genomic_accession=49707414, orientation='minus', exon_count=32, OMIM=616288, no_hdr0=''),
  105375211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375211, CurrentID=0, Status='live', Symbol='LOC105375211', Aliases='', description='uncharacterized LOC105375211', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27647093, end_position_on_the_genomic_accession=27723002, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91012, CurrentID=0, Status='live', Symbol='CERS5', Aliases='LASS5, Trh4', description='ceramide synthase 5', other_designations='LAG1 homolog, ceramide synthase 5|LAG1 longevity assurance homolog 5', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50129306, end_position_on_the_genomic_accession=50167533, orientation='minus', exon_count=13, OMIM=615335, no_hdr0=''),
  105375212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375212, CurrentID=0, Status='live', Symbol='LOC105375212', Aliases='', description='uncharacterized LOC105375212', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28186841, end_position_on_the_genomic_accession=28193003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  353163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353163, CurrentID=0, Status='live', Symbol='CAP2P1', Aliases='', description='CAP, adenylate cyclase-associated protein, 2 (yeast) pseudogene 1', other_designations='adenylyl cyclase-associated protein 2 pseudogene 1', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89289518, end_position_on_the_genomic_accession=89291731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353164, CurrentID=0, Status='live', Symbol='TAS2R42', Aliases='T2R24, T2R42, T2R55, TAS2R55', description='taste 2 receptor member 42', other_designations='candidate taste receptor hT2R55|taste receptor type 2 member 55|taste receptor, type 2, member 42', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11185993, end_position_on_the_genomic_accession=11186937, orientation='minus', exon_count=1, OMIM=613966, no_hdr0=''),
  353173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353173, CurrentID=0, Status='live', Symbol='ZLS', Aliases='', description='Zimmerman-Laband Syndrome', other_designations='Laband Syndrome', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=135500, no_hdr0=''),
  353174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353174, CurrentID=0, Status='live', Symbol='ZACN', Aliases='L2, LGICZ, LGICZ1, ZAC, ZAC1', description='zinc activated ligand-gated ion channel', other_designations='ligand-gated ion channel subunit|ligand-gated ion channel zinc-activated 1|ligand-gated ion channel, zinc activated 1|ligand-gated ion-channel receptor L2|zinc activated ligand-gated ion channel 1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76079182, end_position_on_the_genomic_accession=76082804, orientation='plus', exon_count=9, OMIM=610935, no_hdr0=''),
  91039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91039, CurrentID=0, Status='live', Symbol='DPP9', Aliases='DP9, DPLP9, DPRP-2, DPRP2', description='dipeptidyl-peptidase 9', other_designations='DPP IX|dipeptidyl peptidase IV-related protein 2|dipeptidyl peptidase IV-related protein-2|dipeptidyl peptidase IX|dipeptidyl peptidase-like protein 9|dipeptidylpeptidase 9', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4675226, end_position_on_the_genomic_accession=4723843, orientation='minus', exon_count=27, OMIM=608258, no_hdr0=''),
  353189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353189, CurrentID=0, Status='live', Symbol='SLCO4C1', Aliases='OATP-H, OATP-M1, OATP4C1, OATPX, PRO2176, SLC21A20', description='solute carrier organic anion transporter family member 4C1', other_designations='organic anion transporter M1|solute carrier family 21 member 20|solute carrier organic anion transporter family, member 4C1', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102233986, end_position_on_the_genomic_accession=102296549, orientation='minus', exon_count=16, OMIM=609013, no_hdr0=''),
  105375217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375217, CurrentID=0, Status='live', Symbol='LOC105375217', Aliases='', description='uncharacterized LOC105375217', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30162495, end_position_on_the_genomic_accession=30183210, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91050, CurrentID=0, Status='live', Symbol='CCDC149', Aliases='', description='coiled-coil domain containing 149', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24806117, end_position_on_the_genomic_accession=24980204, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  353196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353196, CurrentID=0, Status='live', Symbol='KRT16P5', Aliases='', description='keratin 16 pseudogene 5', other_designations='keratin pseudogene', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20493041, end_position_on_the_genomic_accession=20494447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375218, CurrentID=0, Status='live', Symbol='LOC105375218', Aliases='', description='uncharacterized LOC105375218', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30243598, end_position_on_the_genomic_accession=30282382, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  91056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91056, CurrentID=0, Status='live', Symbol='AP5B1', Aliases='AP-5, PP1030', description='adaptor related protein complex 5, beta 1 subunit', other_designations='AP-5 complex subunit beta|adaptor-related protein complex 5 beta subunit|adaptor-related protein complex 5, beta 1 subunit|beta5', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65773898, end_position_on_the_genomic_accession=65780591, orientation='minus', exon_count=2, OMIM=614367, no_hdr0=''),
  91057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91057, CurrentID=0, Status='live', Symbol='CCDC34', Aliases='L15, NY-REN-41, RAMA3', description='coiled-coil domain containing 34', other_designations='NY-REN-41 antigen|renal carcinoma antigen NY-REN-41', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27338514, end_position_on_the_genomic_accession=27363248, orientation='minus', exon_count=6, OMIM=612324, no_hdr0=''),
  105375219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375219, CurrentID=0, Status='live', Symbol='LOC105375219', Aliases='', description='uncharacterized LOC105375219', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30717588, end_position_on_the_genomic_accession=30727274, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132340, CurrentID=0, Status='live', Symbol='CYP4F59P', Aliases='', description='cytochrome P450 family 4 subfamily F member 59, pseudogene', other_designations='CYP4F-se14[6:8]|cytochrome P450, family 4, subfamily F, polypeptide 59, pseudogene', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66050577, end_position_on_the_genomic_accession=66050859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375221, CurrentID=0, Status='live', Symbol='LOC105375221', Aliases='', description='uncharacterized LOC105375221', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30994090, end_position_on_the_genomic_accession=31001326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91074, CurrentID=0, Status='live', Symbol='ANKRD30A', Aliases='NY-BR-1', description='ankyrin repeat domain 30A', other_designations='breast cancer antigen NY-BR-1|serologically defined breast cancer antigen NY-BR-1', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37125742, end_position_on_the_genomic_accession=37235831, orientation='plus', exon_count=46, OMIM=610856, no_hdr0=''),
  353219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353219, CurrentID=0, Status='live', Symbol='KAAG1', Aliases='RU2AS', description='kidney associated antigen 1', other_designations='RU2 antisense gene protein', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24356903, end_position_on_the_genomic_accession=24358284, orientation='plus', exon_count=1, OMIM=608211, no_hdr0=''),
  105375646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375646, CurrentID=0, Status='live', Symbol='LOC105375646', Aliases='', description='uncharacterized LOC105375646', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93899862, end_position_on_the_genomic_accession=93904743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  222161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222161, CurrentID=0, Status='live', Symbol='DKFZP586I1420', Aliases='', description='uncharacterized protein DKFZp586I1420', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 5 pseudogene', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30370050, end_position_on_the_genomic_accession=30372809, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  222166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222166, CurrentID=0, Status='live', Symbol='MTURN', Aliases='C7orf41, Ells1', description='maturin, neural progenitor differentiation regulator homolog (Xenopus)', other_designations='UPF0452 protein C7orf41|maturin neural progenitor differentiation regulator protein homolog|protein Ells1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30134936, end_position_on_the_genomic_accession=30162765, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  222171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222171, CurrentID=0, Status='live', Symbol='PRR15', Aliases='', description='proline rich 15', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29563811, end_position_on_the_genomic_accession=29567300, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375226, CurrentID=0, Status='live', Symbol='LOC105375226', Aliases='', description='uncharacterized LOC105375226', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33464935, end_position_on_the_genomic_accession=33467207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91107, CurrentID=0, Status='live', Symbol='TRIM47', Aliases='GOA, RNF100', description='tripartite motif containing 47', other_designations='RING finger protein 100|gene overexpressed in astrocytoma protein|tripartite motif-containing 47', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75874164, end_position_on_the_genomic_accession=75878600, orientation='minus', exon_count=7, OMIM=611041, no_hdr0=''),
  222183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222183, CurrentID=0, Status='live', Symbol='SRRM3', Aliases='', description='serine/arginine repetitive matrix 3', other_designations='SRRM2-like protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76166675, end_position_on_the_genomic_accession=76287292, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  105375228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375228, CurrentID=0, Status='live', Symbol='LOC105375228', Aliases='', description='uncharacterized LOC105375228', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35035034, end_position_on_the_genomic_accession=35036050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91120, CurrentID=0, Status='live', Symbol='ZNF682', Aliases='BC39498_3', description='zinc finger protein 682', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20004418, end_position_on_the_genomic_accession=20039526, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  222194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222194, CurrentID=0, Status='live', Symbol='RSBN1L', Aliases='', description='round spermatid basic protein 1-like', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77696426, end_position_on_the_genomic_accession=77779803, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  353267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353267, CurrentID=0, Status='live', Symbol='FAM103A2P', Aliases='C6orf119, dJ427A4.2', description='family with sequence similarity 103 member A2, pseudogene', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166585518, end_position_on_the_genomic_accession=166586685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353274, CurrentID=0, Status='live', Symbol='ZNF445', Aliases='ZKSCAN15, ZNF168, ZSCAN47', description='zinc finger protein 445', other_designations='zinc finger protein 168|zinc finger protein with KRAB and SCAN domains 15', map_location='3p21.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44439770, end_position_on_the_genomic_accession=44477670, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  91133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91133, CurrentID=0, Status='live', Symbol='L3MBTL4', Aliases='HsT1031', description='l(3)mbt-like 4 (Drosophila)', other_designations='H-l(3)mbt-like protein 4|l(3)mbt-like protein 4', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5954703, end_position_on_the_genomic_accession=6414911, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  91137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91137, CurrentID=0, Status='live', Symbol='SLC25A46', Aliases='', description='solute carrier family 25 member 46', other_designations='solute carrier family 25, member 46', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110738136, end_position_on_the_genomic_accession=110765161, orientation='plus', exon_count=10, OMIM=610826, no_hdr0=''),
  400206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400206, CurrentID=0, Status='live', Symbol='DPPA3P2', Aliases='STELLAR', description='developmental pluripotency associated 3 pseudogene 2', other_designations='germ and embryonic stem cell enriched protein STELLA', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36371318, end_position_on_the_genomic_accession=36372369, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353288, CurrentID=0, Status='live', Symbol='KRT26', Aliases='CK26, K25IRS2, K26, KRT25B', description='keratin 26, type I', other_designations='CK-26|K25B|cytokeratin-26|keratin 25B|keratin-25B|keratin-26|type I inner root sheath specific keratin 25 irs2|type I inner root sheath-specific keratin-K25irs2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40766238, end_position_on_the_genomic_accession=40772159, orientation='minus', exon_count=8, OMIM=616675, no_hdr0=''),
  91147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91147, CurrentID=0, Status='live', Symbol='TMEM67', Aliases='JBTS6, MECKELIN, MKS3, NPHP11, TNEM67', description='transmembrane protein 67', other_designations='meckel syndrome type 3 protein', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93754844, end_position_on_the_genomic_accession=93832649, orientation='plus', exon_count=36, OMIM=609884, no_hdr0=''),
  353293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353293, CurrentID=0, Status='live', Symbol='SPG23', Aliases='', description='spastic paraplegia 23 (autosomal recessive)', other_designations='', map_location='1q24-q32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=270750, no_hdr0=''),
  646066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646066, CurrentID=0, Status='live', Symbol='LOC646066', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17525135, end_position_on_the_genomic_accession=17525931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  222223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222223, CurrentID=0, Status='live', Symbol='KIAA1324L', Aliases='EIG121L', description='KIAA1324-like', other_designations='EIG121-like|estrogen-induced gene 121-like protein', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=86876906, end_position_on_the_genomic_accession=87059698, orientation='minus', exon_count=28, OMIM=614048, no_hdr0=''),
  353299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353299, CurrentID=0, Status='live', Symbol='RGSL1', Aliases='RGSL, RGSL2', description='regulator of G-protein signaling like 1', other_designations='regulator of G-protein signaling like 2', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182449217, end_position_on_the_genomic_accession=182560597, orientation='plus', exon_count=34, OMIM=611012, no_hdr0=''),
  91156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91156, CurrentID=0, Status='live', Symbol='IGFN1', Aliases='EEF1A2BP1', description='immunoglobulin-like and fibronectin type III domain containing 1', other_designations='EEF1A2-binding protein 1|KY-interacting protein 1|eEF1A2 binding protein', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201190806, end_position_on_the_genomic_accession=201228952, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  222229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222229, CurrentID=0, Status='live', Symbol='LRWD1', Aliases='CENP-33, ORCA', description='leucine-rich repeats and WD repeat domain containing 1', other_designations='ORC-associated protein|centromere protein 33|origin recognition complex-associated protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102464943, end_position_on_the_genomic_accession=102473168, orientation='plus', exon_count=15, OMIM=615167, no_hdr0=''),
  105375236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375236, CurrentID=0, Status='live', Symbol='LOC105375236', Aliases='', description='uncharacterized LOC105375236', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38003229, end_position_on_the_genomic_accession=38044111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  222234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222234, CurrentID=0, Status='live', Symbol='FAM185A', Aliases='', description='family with sequence similarity 185 member A', other_designations='family with sequence similarity 185, member A', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102748944, end_position_on_the_genomic_accession=102912918, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  222235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222235, CurrentID=0, Status='live', Symbol='FBXL13', Aliases='Fbl13', description='F-box and leucine-rich repeat protein 13', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102700062, end_position_on_the_genomic_accession=103074841, orientation='minus', exon_count=33, OMIM=609080, no_hdr0=''),
  222236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222236, CurrentID=0, Status='live', Symbol='NAPEPLD', Aliases='FMP30, NAPE-PLD', description='N-acyl phosphatidylethanolamine phospholipase D', other_designations='NAPE-hydrolyzing phospholipase D', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103099576, end_position_on_the_genomic_accession=103149322, orientation='minus', exon_count=8, OMIM=612334, no_hdr0=''),
  101901829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101901829, CurrentID=0, Status='live', Symbol='LOC101901829', Aliases='', description='dynein, light chain, roadblock-type 1 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5583761, end_position_on_the_genomic_accession=5584445, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375238, CurrentID=0, Status='live', Symbol='LOC105375238', Aliases='', description='uncharacterized LOC105375238', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39416277, end_position_on_the_genomic_accession=39474451, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  353322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353322, CurrentID=0, Status='live', Symbol='ANKRD37', Aliases='Lrp2bp', description='ankyrin repeat domain 37', other_designations='low density lipoprotein receptor-related protein binding protein|low-density lipoprotein receptor-related protein 2-binding protein', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185396686, end_position_on_the_genomic_accession=185400243, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  91179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91179, CurrentID=0, Status='live', Symbol='SCARF2', Aliases='NSR1, SREC-II, SREC2, SRECRP-1, VDEGS', description='scavenger receptor class F member 2', other_designations='scavenger receptor class F, member 2|scavenger receptor expressed by endothelial cells 2 protein', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20424584, end_position_on_the_genomic_accession=20437859, orientation='minus', exon_count=11, OMIM=613619, no_hdr0=''),
  353324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353324, CurrentID=0, Status='live', Symbol='SPATA12', Aliases='SRG5', description='spermatogenesis associated 12', other_designations='spermatogenesis-related protein 5', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57060441, end_position_on_the_genomic_accession=57076068, orientation='plus', exon_count=3, OMIM=609869, no_hdr0=''),
  91181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91181, CurrentID=0, Status='live', Symbol='NUP210L', Aliases='', description='nucleoporin 210kDa like', other_designations='nuclear pore membrane glycoprotein 210-like (LOC91181)|nucleoporin 210kDa-like|nucleoporin Nup210-like', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153992690, end_position_on_the_genomic_accession=154155116, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  222255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222255, CurrentID=0, Status='live', Symbol='ATXN7L1', Aliases='ATXN7L4', description='ataxin 7-like 1', other_designations='ataxin 7-like 4|ataxin-7-like protein 4', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105604774, end_position_on_the_genomic_accession=105877232, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  222256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222256, CurrentID=0, Status='live', Symbol='CDHR3', Aliases='CDH28', description='cadherin related family member 3', other_designations='cadherin-like protein 28', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105877200, end_position_on_the_genomic_accession=106036431, orientation='plus', exon_count=22, OMIM=615610, no_hdr0=''),
  353329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353329, CurrentID=0, Status='live', Symbol='PSIP1P1', Aliases='PSIP1P', description='PC4 and SFRS1 interacting protein 1 pseudogene 1', other_designations='PC4 and SFRS1 interacting protein pseudogene', map_location='9p22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15054967, end_position_on_the_genomic_accession=15056600, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353332, CurrentID=0, Status='live', Symbol='KRTAP12-1', Aliases='KAP12.1, KRTAP12.1', description='keratin associated protein 12-1', other_designations='high sulfur keratin-associated protein 12.1|keratin-associated protein 12.1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44681576, end_position_on_the_genomic_accession=44682163, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353333, CurrentID=0, Status='live', Symbol='KRTAP10-10', Aliases='KAP10.10, KAP18.10, KRTAP18-10, KRTAP18.10', description='keratin associated protein 10-10', other_designations='high sulfur keratin-associated protein 10.10|keratin-associated protein 18-10', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44637356, end_position_on_the_genomic_accession=44638455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  353345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353345, CurrentID=0, Status='live', Symbol='GPR141', Aliases='PGR13', description='G protein-coupled receptor 141', other_designations='G-protein coupled receptor PGR13', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37683625, end_position_on_the_genomic_accession=37742310, orientation='plus', exon_count=11, OMIM=609045, no_hdr0=''),
  353347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353347, CurrentID=0, Status='live', Symbol='DFNA53', Aliases='', description='deafness, autosomal dominant 53', other_designations='', map_location='14q11-q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609965, no_hdr0=''),
  100129497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129497, CurrentID=0, Status='live', Symbol='LOC100129497', Aliases='', description='valosin containing protein lysine (K) methyltransferase pseudogene', other_designations='methyltransferase like 21D pseudogene', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37782216, end_position_on_the_genomic_accession=37782885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353355, CurrentID=0, Status='live', Symbol='ZNF233', Aliases='', description='zinc finger protein 233', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44259880, end_position_on_the_genomic_accession=44275317, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371286, CurrentID=0, Status='live', Symbol='LOC105371286', Aliases='', description='uncharacterized LOC105371286', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56708899, end_position_on_the_genomic_accession=56730015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375245, CurrentID=0, Status='live', Symbol='LOC105375245', Aliases='', description='uncharacterized LOC105375245', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40239916, end_position_on_the_genomic_accession=40297221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23515, CurrentID=0, Status='live', Symbol='MORC3', Aliases='NXP2, ZCW5, ZCWCC3', description='MORC family CW-type zinc finger 3', other_designations='nuclear matrix protein NXP2|zinc finger CW-type coiled-coil domain protein 3|zinc finger, CW type with coiled-coil domain 3|zinc finger, CW-type with coiled-coil domain 3', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36320189, end_position_on_the_genomic_accession=36376646, orientation='plus', exon_count=20, OMIM=610078, no_hdr0=''),
  91219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91219, CurrentID=0, Status='live', Symbol='ABHD17AP5', Aliases='FAM108A6, FAM108A6P', description='abhydrolase domain containing 17A pseudogene 5', other_designations='family with sequence similarity 108, member A1 pseudogene|family with sequence similarity 108, member A6, pseudogene', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22117481, end_position_on_the_genomic_accession=22118005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  91227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91227, CurrentID=0, Status='live', Symbol='GGTLC2', Aliases='GGTL4', description='gamma-glutamyltransferase light chain 2', other_designations='gamma-glutamyltransferase-like 4|gamma-glutamyltransferase-like protein 4', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22644620, end_position_on_the_genomic_accession=22647898, orientation='plus', exon_count=6, OMIM=612339, no_hdr0=''),
  400209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400209, CurrentID=0, Status='live', Symbol='YTHDF2P1', Aliases='HGRG8, NY-REN-2, YTHDF2P', description='YTH domain family member 2 pseudogene 1', other_designations='YTH domain family 2 pseudogene|YTH domain family, member 2 pseudogene 1|high-glucose-regulated protein 8 pseudogene', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39215064, end_position_on_the_genomic_accession=39217159, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  353376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353376, CurrentID=0, Status='live', Symbol='TICAM2', Aliases='MyD88-4, TICAM-2, TIRAP3, TIRP, TRAM', description='toll-like receptor adaptor molecule 2', other_designations='NF-kappa-B-activating protein 502|TRIF-related adaptor molecule|cytoplasmic adaptor|putative NF-kappa-B-activating protein 502|toll-like receptor adaptor protein 3|toll/interleukin-1 receptor (TIR) domain-containing adapter protein', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115578496, end_position_on_the_genomic_accession=115602479, orientation='minus', exon_count=2, OMIM=608321, no_hdr0=''),
  100132368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132368, CurrentID=0, Status='live', Symbol='LOC100132368', Aliases='', description='uncharacterized LOC100132368', other_designations='', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=13017606, end_position_on_the_genomic_accession=13018335, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100420583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420583, CurrentID=0, Status='live', Symbol='LOC100420583', Aliases='', description='taste receptor, type 2, member 7 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11196014, end_position_on_the_genomic_accession=11197018, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129498, CurrentID=0, Status='live', Symbol='LOC100129498', Aliases='', description='uncharacterized LOC100129498', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=117537769, end_position_on_the_genomic_accession=117538443, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  91252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91252, CurrentID=0, Status='live', Symbol='SLC39A13', Aliases='', description='solute carrier family 39 (zinc transporter), member 13', other_designations='LIV-1 subfamily of ZIP zinc transporter 9|LZT-Hs9|solute carrier family 39 (metal ion transporter), member 13|zrt- and Irt-like protein 13', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47407132, end_position_on_the_genomic_accession=47416500, orientation='plus', exon_count=13, OMIM=608735, no_hdr0=''),
  105375252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375252, CurrentID=0, Status='live', Symbol='LOC105375252', Aliases='', description='serine-threonine kinase receptor-associated protein pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42919050, end_position_on_the_genomic_accession=42920871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392539, CurrentID=0, Status='live', Symbol='LOC392539', Aliases='', description='tight junction associated protein 1 pseudogene', other_designations='tight junction associated protein 1 (peripheral) pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129040524, end_position_on_the_genomic_accession=129043412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  91272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91272, CurrentID=0, Status='live', Symbol='BOD1', Aliases='FAM44B', description='biorientation of chromosomes in cell division 1', other_designations='biorientation defective 1|family with sequence similarity 44, member B', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173607145, end_position_on_the_genomic_accession=173617077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132375, CurrentID=0, Status='live', Symbol='LOC100132375', Aliases='', description='uncharacterized LOC100132375', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=96841719, end_position_on_the_genomic_accession=96842192, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379519, CurrentID=0, Status='live', Symbol='LOC105379519', Aliases='', description='putative uncharacterized protein ENSP00000383309', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18361818, end_position_on_the_genomic_accession=18369146, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132376, CurrentID=0, Status='live', Symbol='RPS29P8', Aliases='RPS29_5_283', description='ribosomal protein S29 pseudogene 8', other_designations='', map_location='2q23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148595127, end_position_on_the_genomic_accession=148595415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91283, CurrentID=0, Status='live', Symbol='MSANTD3', Aliases='C9orf30, L8', description='Myb/SANT-like DNA-binding domain containing 3', other_designations='UPF0439 protein C9orf30', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100427213, end_position_on_the_genomic_accession=100451734, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375257, CurrentID=0, Status='live', Symbol='LOC105375257', Aliases='', description='uncharacterized LOC105375257', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44172375, end_position_on_the_genomic_accession=44184141, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91289, CurrentID=0, Status='live', Symbol='LMF2', Aliases='TMEM112B, TMEM153', description='lipase maturation factor 2', other_designations='transmembrane protein 112B|transmembrane protein 153', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50502947, end_position_on_the_genomic_accession=50507840, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  91298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91298, CurrentID=0, Status='live', Symbol='C12orf29', Aliases='', description='chromosome 12 open reading frame 29', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88035491, end_position_on_the_genomic_accession=88050160, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  91300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91300, CurrentID=0, Status='live', Symbol='R3HDM4', Aliases='C19orf22', description='R3H domain containing 4', other_designations='R3H domain-containing protein C19orf22', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=896503, end_position_on_the_genomic_accession=913225, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  91304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91304, CurrentID=0, Status='live', Symbol='TMEM259', Aliases='ASBABP1, C19orf6, MBRL, MEMBRALIN, R32184_3', description='transmembrane protein 259', other_designations='aspecific BCL2 ARE-binding protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1009651, end_position_on_the_genomic_accession=1021169, orientation='minus', exon_count=11, OMIM=611011, no_hdr0=''),
  105375260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375260, CurrentID=0, Status='live', Symbol='LOC105375260', Aliases='', description='uncharacterized LOC105375260', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44821994, end_position_on_the_genomic_accession=44824842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91316, CurrentID=0, Status='live', Symbol='GUSBP11', Aliases='', description='glucuronidase, beta pseudogene 11', other_designations='glucuronidase, beta/immunoglobulin lambda-like polypeptide 1 pseudogene', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23638487, end_position_on_the_genomic_accession=23717423, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  222389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222389, CurrentID=0, Status='live', Symbol='BEND7', Aliases='C10orf30', description='BEN domain containing 7', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13438484, end_position_on_the_genomic_accession=13528597, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  91319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91319, CurrentID=0, Status='live', Symbol='DERL3', Aliases='C22orf14, IZP6, LLN2, derlin-3', description='derlin 3', other_designations='DERtrin 3|DERtrin-3|Der1-like domain family, member 3|degradation in endoplasmic reticulum protein 3|der1-like protein 3', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23834503, end_position_on_the_genomic_accession=23839213, orientation='minus', exon_count=8, OMIM=610305, no_hdr0=''),
  105373409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373409, CurrentID=0, Status='live', Symbol='LOC105373409', Aliases='', description='uncharacterized LOC105373409', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7881307, end_position_on_the_genomic_accession=7892182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375264, CurrentID=0, Status='live', Symbol='LOC105375264', Aliases='', description='uncharacterized LOC105375264', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45992866, end_position_on_the_genomic_accession=46024296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371290, CurrentID=0, Status='live', Symbol='LOC105371290', Aliases='', description='uncharacterized LOC105371290', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57255511, end_position_on_the_genomic_accession=57256164, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375265, CurrentID=0, Status='live', Symbol='LOC105375265', Aliases='', description='uncharacterized LOC105375265', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46302128, end_position_on_the_genomic_accession=46353154, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387119, CurrentID=0, Status='live', Symbol='CEP85L', Aliases='C6orf204, NY-BR-15, bA57K17.2', description='centrosomal protein 85kDa-like', other_designations='serologically defined breast cancer antigen NY-BR-15', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118460772, end_position_on_the_genomic_accession=118710082, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  645680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645680, CurrentID=0, Status='live', Symbol='LOC645680', Aliases='', description='karyopherin alpha 2 pseudogene', other_designations='karyopherin alpha 2 (RAG cohort 1, importin alpha 1) pseudogene', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88564468, end_position_on_the_genomic_accession=88566406, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375266, CurrentID=0, Status='live', Symbol='LOC105375266', Aliases='', description='uncharacterized LOC105375266', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46476457, end_position_on_the_genomic_accession=46481587, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  91351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91351, CurrentID=0, Status='live', Symbol='DDX60L', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 60-like', other_designations='DEAD box protein 60-like|putative ATP-dependent RNA helicase DDX60', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168356735, end_position_on_the_genomic_accession=168480515, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  91353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91353, CurrentID=0, Status='live', Symbol='IGLL3P', Aliases='16.1, IGLL3', description='immunoglobulin lambda-like polypeptide 3, pseudogene', other_designations='omega protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25318257, end_position_on_the_genomic_accession=25320227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91355, CurrentID=0, Status='live', Symbol='LRP5L', Aliases='', description='LDL receptor related protein 5 like', other_designations='LRP-5-like|low density lipoprotein receptor-related protein 5-like', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25351418, end_position_on_the_genomic_accession=25381989, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  353500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353500, CurrentID=0, Status='live', Symbol='BMP8A', Aliases='', description='bone morphogenetic protein 8a', other_designations='BMP-8A', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39491634, end_position_on_the_genomic_accession=39529869, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375269, CurrentID=0, Status='live', Symbol='LOC105375269', Aliases='', description='uncharacterized LOC105375269', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47265300, end_position_on_the_genomic_accession=47268847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375270, CurrentID=0, Status='live', Symbol='LOC105375270', Aliases='', description='uncharacterized LOC105375270', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47982589, end_position_on_the_genomic_accession=47986467, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  353511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353511, CurrentID=0, Status='live', Symbol='PKD1P6', Aliases='HG6', description='polycystic kidney disease 1 (autosomal dominant) pseudogene 6', other_designations='GPS, PLAT and transmembrane domain-containing protein', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15104723, end_position_on_the_genomic_accession=15131601, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  91368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91368, CurrentID=0, Status='live', Symbol='CDKN2AIPNL', Aliases='', description='CDKN2A interacting protein N-terminal like', other_designations='CDKN2A-interacting protein N-terminal-like protein', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134402065, end_position_on_the_genomic_accession=134414603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  91369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91369, CurrentID=0, Status='live', Symbol='ANKRD40', Aliases='', description='ankyrin repeat domain 40', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50693190, end_position_on_the_genomic_accession=50707909, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  353514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353514, CurrentID=0, Status='live', Symbol='LILRA5', Aliases='CD85, CD85F, ILT-11, ILT11, LILRB7, LIR-9, LIR9', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 5', other_designations='CD85 antigen-like family member F|immunoglobulin-like transcript 11 protein|leukocyte Ig-like receptor 9|leukocyte immunoglobulin-like receptor 9|leukocyte immunoglobulin-like receptor subfamily A member 5 soluble|leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 7', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54307070, end_position_on_the_genomic_accession=54313139, orientation='minus', exon_count=7, OMIM=606047, no_hdr0=''),
  353515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353515, CurrentID=0, Status='live', Symbol='XKRY2', Aliases='XKRYP7', description='X-linked Kx blood group related, Y-linked 2', other_designations='X Kell blood group precursor-related, Y-linked 2|X Kell blood group precursor-related, Y-linked pseudogene 7|XK, Kell blood group complex subunit-related, Y-linked 2', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18135449, end_position_on_the_genomic_accession=18137029, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  91373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91373, CurrentID=0, Status='live', Symbol='UAP1L1', Aliases='', description='UDP-N-acetylglucosamine pyrophosphorylase 1 like 1', other_designations='UDP-N-acteylglucosamine pyrophosphorylase 1-like 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137077465, end_position_on_the_genomic_accession=137084538, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80765, CurrentID=0, Status='live', Symbol='STARD5', Aliases='', description='StAR related lipid transfer domain containing 5', other_designations='START domain containing 5|START domain-containing protein 5|StAR-related lipid transfer (START) domain containing 5|StAR-related lipid transfer domain containing 5', map_location='15q26', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81312666, end_position_on_the_genomic_accession=81324183, orientation='minus', exon_count=6, OMIM=607050, no_hdr0=''),
  105375272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375272, CurrentID=0, Status='live', Symbol='LOC105375272', Aliases='', description='uncharacterized LOC105375272', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48659486, end_position_on_the_genomic_accession=48669491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91380, CurrentID=0, Status='live', Symbol='SNORD107', Aliases='HBII-436', description='small nucleolar RNA, C/D box 107', other_designations='HBII-436 C/D box snoRNA', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24981994, end_position_on_the_genomic_accession=24982068, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  91392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91392, CurrentID=0, Status='live', Symbol='ZNF502', Aliases='', description='zinc finger protein 502', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44712643, end_position_on_the_genomic_accession=44723831, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132396, CurrentID=0, Status='live', Symbol='ZNF705B', Aliases='', description='zinc finger protein 705B', other_designations='Putative zinc finger protein 705D-like protein LOC100132396|putative zinc finger protein 705B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7926337, end_position_on_the_genomic_accession=7952413, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  91404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91404, CurrentID=0, Status='live', Symbol='SESTD1', Aliases='SOLO', description='SEC14 and spectrin domain containing 1', other_designations='SEC14 and spectrin domains 1|huntingtin-interacting protein-like protein|protein Solo', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=179101692, end_position_on_the_genomic_accession=179264623, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  100885773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885773, CurrentID=0, Status='live', Symbol='DYT21', Aliases='', description='dystonia 21, torsion (autosomal dominant)', other_designations='', map_location='2q14.3-q21.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614588, no_hdr0=''),
  100885774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885774, CurrentID=0, Status='live', Symbol='PDCOS', Aliases='', description='Podoconiosis, susceptibility to', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614590, no_hdr0=''),
  100885775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885775, CurrentID=0, Status='live', Symbol='BANCR', Aliases='LINC00586', description='BRAF-activated non-protein coding RNA', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69296681, end_position_on_the_genomic_accession=69307056, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  91408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91408, CurrentID=0, Status='live', Symbol='BTF3L4', Aliases='', description='basic transcription factor 3-like 4', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52056185, end_position_on_the_genomic_accession=52090716, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  91409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91409, CurrentID=0, Status='live', Symbol='CCDC74B', Aliases='', description='coiled-coil domain containing 74B', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130139283, end_position_on_the_genomic_accession=130145134, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100885778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885778, CurrentID=0, Status='live', Symbol='NALCN-AS1', Aliases='', description='NALCN antisense RNA 1', other_designations='NALCN antisense RNA 1 (non-protein coding)', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100708325, end_position_on_the_genomic_accession=101059286, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100885779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885779, CurrentID=0, Status='live', Symbol='LINC-ROR', Aliases='ROR, lincRNA-RoR', description='long intergenic non-protein coding RNA, regulator of reprogramming', other_designations='CTD-2526M8.1|lincRNA-ST8SIA3', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57054571, end_position_on_the_genomic_accession=57072119, orientation='minus', exon_count=4, OMIM=615173, no_hdr0=''),
  222484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222484, CurrentID=0, Status='live', Symbol='LNX2', Aliases='PDZRN1', description='ligand of numb-protein X 2', other_designations='PDZ domain containing ring finger 1|PDZ domain-containing RING finger protein 1|numb-binding protein 2', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27545911, end_position_on_the_genomic_accession=27620583, orientation='minus', exon_count=10, OMIM=609733, no_hdr0=''),
  100885781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885781, CurrentID=0, Status='live', Symbol='LINC00348', Aliases='', description='long intergenic non-protein coding RNA 348', other_designations='', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=71015141, end_position_on_the_genomic_accession=71168417, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100885782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885782, CurrentID=0, Status='live', Symbol='MYO16-AS1', Aliases='', description='MYO16 antisense RNA 1', other_designations='MYO16 antisense RNA 1 (non-protein coding)', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109163905, end_position_on_the_genomic_accession=109167303, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  222487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222487, CurrentID=0, Status='live', Symbol='ADGRG3', Aliases='GPR97, PB99, PGR26', description='adhesion G protein-coupled receptor G3', other_designations='G protein-coupled receptor 97|G-protein coupled receptor 97|G-protein coupled receptor PGR26|probable G-protein coupled receptor 97', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57668192, end_position_on_the_genomic_accession=57689378, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100885786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885786, CurrentID=0, Status='live', Symbol='NMLFS', Aliases='C8DELq22.1, DEL8q22.1', description='Nablus mask-like facial syndrome', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608156, no_hdr0=''),
  386607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386607, CurrentID=0, Status='live', Symbol='ZFP91-CNTF', Aliases='', description='ZFP91-CNTF readthrough (NMD candidate)', other_designations='ZFP91-CNTF readthrough (non-protein coding)|zinc finger protein 91 homolog, ciliary neurotrophic factor transcription unit', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58579114, end_position_on_the_genomic_accession=58625733, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100885788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885788, CurrentID=0, Status='live', Symbol='OTDD', Aliases='C11DELq13, DEL11q13', description='Otodental dysplasia chromosome deletion syndrome', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=166750, no_hdr0=''),
  100885789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885789, CurrentID=0, Status='live', Symbol='IFNG-AS1', Aliases='GS1-410F4.2, NEST, Tmevpg1', description='IFNG antisense RNA 1', other_designations="IFNG antisense RNA 1 (non-protein coding)|LincR-Ifng-3'AS|Theiler's murine encephalomyelitis virus persistence candidate gene 1", map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67989445, end_position_on_the_genomic_accession=68021327, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  386608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386608, CurrentID=0, Status='live', Symbol='RNASEH2CP1', Aliases='AYP1p1', description='ribonuclease H2 subunit C pseudogene 1', other_designations='AYP1 pseudogene 1|T6 pseudogene|ribonuclease H2, subunit C pseudogene 1', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2789826, end_position_on_the_genomic_accession=2790396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100885795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885795, CurrentID=0, Status='live', Symbol='LARS2-AS1', Aliases='', description='LARS2 antisense RNA 1', other_designations='LARS2 antisense RNA 1 (non-protein coding)', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45482695, end_position_on_the_genomic_accession=45509545, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100885796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885796, CurrentID=0, Status='live', Symbol='SLC9A9-AS1', Aliases='', description='SLC9A9 antisense RNA 1', other_designations='SLC9A9 antisense RNA 1 (non-protein coding)', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143342246, end_position_on_the_genomic_accession=143347071, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  342918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342918, CurrentID=0, Status='live', Symbol='C19orf81', Aliases='', description='chromosome 19 open reading frame 81', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50632767, end_position_on_the_genomic_accession=50659310, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100885798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885798, CurrentID=0, Status='live', Symbol='LINC00210', Aliases='NCRNA00210', description='long intergenic non-protein coding RNA 210', other_designations='RP11-72L13.1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217892900, end_position_on_the_genomic_accession=217920804, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100885799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885799, CurrentID=0, Status='live', Symbol='ENO1P4', Aliases='', description='enolase 1, (alpha) pseudogene 4', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201621646, end_position_on_the_genomic_accession=201623430, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100885800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885800, CurrentID=0, Status='live', Symbol='SNAP47-AS1', Aliases='SNAP47-IT1', description='SNAP47 antisense RNA 1', other_designations='SNAP47 antisense RNA 1 (non-protein coding)|SNAP47 intronic transcript 1 (non-protein coding)', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  91433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91433, CurrentID=0, Status='live', Symbol='RCCD1', Aliases='', description='RCC1 domain containing 1', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90954695, end_position_on_the_genomic_accession=90963543, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100885802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885802, CurrentID=0, Status='live', Symbol='KTCN6', Aliases='', description='Keratoconus 6', other_designations='', map_location='9q34', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614623, no_hdr0=''),
  100885803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885803, CurrentID=0, Status='live', Symbol='KTCN8', Aliases='', description='Keratoconus 8', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614628, no_hdr0=''),
  386610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386610, CurrentID=0, Status='live', Symbol='E2F6P1', Aliases='', description='E2F transcription factor 6 pseudogene 1', other_designations='E2F6 pseudogene 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18833419, end_position_on_the_genomic_accession=18835626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375658, CurrentID=0, Status='live', Symbol='LOC105375658', Aliases='', description='uncharacterized LOC105375658', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97865361, end_position_on_the_genomic_accession=97869408, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  91442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91442, CurrentID=0, Status='live', Symbol='FAAP24', Aliases='C19orf40', description='Fanconi anemia core complex associated protein 24', other_designations='Fanconi anemia-associated protein of 24 kDa', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32972217, end_position_on_the_genomic_accession=32977495, orientation='plus', exon_count=5, OMIM=610884, no_hdr0=''),
  105375283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375283, CurrentID=0, Status='live', Symbol='GS1-278J22.2', Aliases='', description='uncharacterized LOC105375283', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53564198, end_position_on_the_genomic_accession=53574528, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91445, CurrentID=0, Status='live', Symbol='RNF185', Aliases='', description='ring finger protein 185', other_designations='BSK65-MONO1|BSK65-MONO2|BSK65-PANC1|BSK65-PANC2|BSK65-TEST1|BSK65-TEST2|BSK65-TEST3', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31160152, end_position_on_the_genomic_accession=31207019, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  91450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91450, CurrentID=0, Status='live', Symbol='LOC91450', Aliases='', description='uncharacterized LOC91450', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77993233, end_position_on_the_genomic_accession=77994225, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  91452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91452, CurrentID=0, Status='live', Symbol='ACBD5', Aliases='', description='acyl-CoA binding domain containing 5', other_designations='acyl-Coenzyme A binding domain containing 5|endozepine-related protein|membrane-associated diazepam binding inhibitor', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27195214, end_position_on_the_genomic_accession=27242139, orientation='minus', exon_count=15, OMIM=616618, no_hdr0=''),
  91461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91461, CurrentID=0, Status='live', Symbol='PKDCC', Aliases='SGK493, Vlk', description='protein kinase domain containing, cytoplasmic', other_designations='protein kinase domain containing, cytoplasmic homolog|protein kinase domain-containing protein, cytoplasmic|protein kinase-like protein SgK493|sugen kinase 493|vertebrate lonesome kinase', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42048021, end_position_on_the_genomic_accession=42058528, orientation='plus', exon_count=7, OMIM=614150, no_hdr0=''),
  91464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91464, CurrentID=0, Status='live', Symbol='ISX', Aliases='Pix-1, RAXLX', description='intestine-specific homeobox', other_designations='RAX-like homeobox|pancreas-intestine homeodomain transcription factor', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35066136, end_position_on_the_genomic_accession=35087387, orientation='plus', exon_count=5, OMIM=612019, no_hdr0=''),
  222537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222537, CurrentID=0, Status='live', Symbol='HS3ST5', Aliases='3-OST-5, 3OST5, HS3OST5, NBLA04021', description='heparan sulfate-glucosamine 3-sulfotransferase 5', other_designations='h3-OST-5|heparan sulfate (glucosamine) 3-O-sulfotransferase 5|heparan sulfate 3-O-sulfotransferase 5|heparan sulfate 3-OST-5|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 5', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114055586, end_position_on_the_genomic_accession=114342433, orientation='minus', exon_count=8, OMIM=609407, no_hdr0=''),
  222545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222545, CurrentID=0, Status='live', Symbol='GPRC6A', Aliases='GPCR, bA86F4.3', description='G protein-coupled receptor, class C, group 6, member A', other_designations='G protein-coupled receptor, family C, group 6, member A|G-protein coupled receptor GPCR33|predicted with SOSUI analysis|seven transmembrane helix receptor', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116792085, end_position_on_the_genomic_accession=116829057, orientation='minus', exon_count=6, OMIM=613572, no_hdr0=''),
  222546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222546, CurrentID=0, Status='live', Symbol='RFX6', Aliases='MTCHRS, MTFS, RFXDC1, dJ955L16.1', description='regulatory factor X6', other_designations='regulatory factor X domain-containing protein 1|regulatory factor X, 6', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116877213, end_position_on_the_genomic_accession=116932163, orientation='plus', exon_count=19, OMIM=612659, no_hdr0=''),
  105375289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375289, CurrentID=0, Status='live', Symbol='LOC105375289', Aliases='', description='uncharacterized LOC105375289', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56667613, end_position_on_the_genomic_accession=56669561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100885848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885848, CurrentID=0, Status='live', Symbol='PTGES3L', Aliases='', description='prostaglandin E synthase 3 (cytosolic)-like', other_designations='prostaglandin E synthase 3-like', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42968088, end_position_on_the_genomic_accession=42980528, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  222553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222553, CurrentID=0, Status='live', Symbol='SLC35F1', Aliases='C6orf169, dJ230I3.1', description='solute carrier family 35 member F1', other_designations='solute carrier family 35, member F1', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117907231, end_position_on_the_genomic_accession=118317676, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100885850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885850, CurrentID=0, Status='live', Symbol='PTGES3L-AARSD1', Aliases='', description='PTGES3L-AARSD1 readthrough', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42950526, end_position_on_the_genomic_accession=42980528, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105375290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375290, CurrentID=0, Status='live', Symbol='LOC105375290', Aliases='', description='uncharacterized LOC105375290', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56651011, end_position_on_the_genomic_accession=56658811, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375291, CurrentID=0, Status='live', Symbol='LOC105375291', Aliases='', description='uncharacterized LOC105375291', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56576608, end_position_on_the_genomic_accession=56617967, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100885864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885864, CurrentID=0, Status='live', Symbol='RNA5SP244', Aliases='RN5S244', description='RNA, 5S ribosomal pseudogene 244', other_designations='RNA, 5S ribosomal 244', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129756266, end_position_on_the_genomic_accession=129756377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100885865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885865, CurrentID=0, Status='live', Symbol='HMGN1P29', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 29', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 29', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43144956, end_position_on_the_genomic_accession=43145255, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100885866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885866, CurrentID=0, Status='live', Symbol='MRPS31P4', Aliases='', description='mitochondrial ribosomal protein S31 pseudogene 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52617558, end_position_on_the_genomic_accession=52637446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375293, CurrentID=0, Status='live', Symbol='LOC105375293', Aliases='', description='uncharacterized LOC105375293', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56373987, end_position_on_the_genomic_accession=56374802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  222584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222584, CurrentID=0, Status='live', Symbol='FAM83B', Aliases='C6orf143', description='family with sequence similarity 83 member B', other_designations='family with sequence similarity 83, member B', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54846643, end_position_on_the_genomic_accession=54945099, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375295, CurrentID=0, Status='live', Symbol='LOC105375295', Aliases='', description='uncharacterized LOC105375295', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56858000, end_position_on_the_genomic_accession=56862321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  342933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=342933, CurrentID=0, Status='live', Symbol='ZSCAN5B', Aliases='ZNF371, ZNF495B', description='zinc finger and SCAN domain containing 5B', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56189689, end_position_on_the_genomic_accession=56208038, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  91522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91522, CurrentID=0, Status='live', Symbol='COL23A1', Aliases='', description='collagen, type XXIII, alpha 1', other_designations='procollagen, type XXIII, alpha 1', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178237616, end_position_on_the_genomic_accession=178590555, orientation='minus', exon_count=31, OMIM=610043, no_hdr0=''),
  91523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91523, CurrentID=0, Status='live', Symbol='PCED1B', Aliases='FAM113B', description='PC-esterase domain containing 1B', other_designations='family with sequence similarity 113, member B', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47079603, end_position_on_the_genomic_accession=47236663, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  91526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91526, CurrentID=0, Status='live', Symbol='ANKRD44', Aliases='PP6-ARS-B', description='ankyrin repeat domain 44', other_designations='ankyrin repeat domain-containing protein 44|protein phosphatase 6 ankyrin repeat subunit B|serine/threonine-protein phosphatase 6 regulatory subunit ARS-B', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196962617, end_position_on_the_genomic_accession=197310797, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  105375297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375297, CurrentID=0, Status='live', Symbol='LOC105375297', Aliases='', description='uncharacterized LOC105375297', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57209802, end_position_on_the_genomic_accession=57222887, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375661, CurrentID=0, Status='live', Symbol='LOC105375661', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98923917, end_position_on_the_genomic_accession=98942358, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375298, CurrentID=0, Status='live', Symbol='LOC105375298', Aliases='', description='uncharacterized LOC105375298', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57242679, end_position_on_the_genomic_accession=57245963, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  222611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222611, CurrentID=0, Status='live', Symbol='ADGRF2', Aliases='GPR111, PGR20, hGPCR35', description='adhesion G protein-coupled receptor F2', other_designations='G protein-coupled receptor 111|probable G-protein coupled receptor 111', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47656590, end_position_on_the_genomic_accession=47697797, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  91543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91543, CurrentID=0, Status='live', Symbol='RSAD2', Aliases='2510004L01Rik, cig33, cig5, vig1', description='radical S-adenosyl methionine domain containing 2', other_designations='cytomegalovirus-induced gene 5 protein|viperin|virus inhibitory protein, endoplasmic reticulum-associated, interferon-inducible', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6865840, end_position_on_the_genomic_accession=6898232, orientation='plus', exon_count=7, OMIM=607810, no_hdr0=''),
  91544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91544, CurrentID=0, Status='live', Symbol='UBXN11', Aliases='COA-1, PP2243, SOC, SOCI, UBXD5', description='UBX domain protein 11', other_designations='UBX domain-containing protein 5|colorectal tumor-associated antigen COA-1|colorectal tumor-associated antigen-1|socius', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26282282, end_position_on_the_genomic_accession=26318265, orientation='minus', exon_count=17, OMIM=609151, no_hdr0=''),
  105375300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375300, CurrentID=0, Status='live', Symbol='LOC105375300', Aliases='', description='uncharacterized LOC105375300', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57518013, end_position_on_the_genomic_accession=57520665, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91548, CurrentID=0, Status='live', Symbol='LOC91548', Aliases='', description='uncharacterized LOC91548', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211675722, end_position_on_the_genomic_accession=211725694, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375301, CurrentID=0, Status='live', Symbol='LOC105375301', Aliases='', description='uncharacterized LOC105375301', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57540792, end_position_on_the_genomic_accession=57549935, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  648774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648774, CurrentID=0, Status='live', Symbol='LOC648774', Aliases='', description='polo-like kinase 1 (Drosophila) pseudogene', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84119217, end_position_on_the_genomic_accession=84120436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91561, CurrentID=0, Status='live', Symbol='RPS2P40', Aliases='RPS2_17_1125', description='ribosomal protein S2 pseudogene 40', other_designations='hCG1992539', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17228251, end_position_on_the_genomic_accession=17229198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  222642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222642, CurrentID=0, Status='live', Symbol='TSPO2', Aliases='BZRPL1', description='translocator protein 2', other_designations='benzodiazapine receptor (peripheral)-like 1|peripheral-type benzodiazepine receptor-like protein 1', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41041370, end_position_on_the_genomic_accession=41044337, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  222643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222643, CurrentID=0, Status='live', Symbol='UNC5CL', Aliases='MUXA, ZUD', description='unc-5 family C-terminal like', other_designations='ZU5 and death domain containing|protein unc-5 homolog C-like|unc-5 homolog C (C. elegans)-like', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41026901, end_position_on_the_genomic_accession=41039199, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  91574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91574, CurrentID=0, Status='live', Symbol='C12orf65', Aliases='COXPD7, SPG55', description='chromosome 12 open reading frame 65', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123233297, end_position_on_the_genomic_accession=123257959, orientation='plus', exon_count=7, OMIM=613541, no_hdr0=''),
  91582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91582, CurrentID=0, Status='live', Symbol='RPS19BP1', Aliases='AROS, S19BP', description='ribosomal protein S19 binding protein 1', other_designations='40S ribosomal protein S19-binding protein 1|RPS19-binding protein 1|homolog of mouse S19 binding protein', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39529093, end_position_on_the_genomic_accession=39532855, orientation='minus', exon_count=5, OMIM=610225, no_hdr0=''),
  91584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91584, CurrentID=0, Status='live', Symbol='PLXNA4', Aliases='FAYV2820, PLEXA4A, PLXNA4B, PRO34003, PLXNA4', description='plexin A4', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132123332, end_position_on_the_genomic_accession=132648688, orientation='minus', exon_count=40, OMIM=604280, no_hdr0=''),
  222658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222658, CurrentID=0, Status='live', Symbol='KCTD20', Aliases='C6orf69, dJ108K11.3', description='potassium channel tetramerization domain containing 20', other_designations='potassium channel tetramerisation domain containing 20', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36442767, end_position_on_the_genomic_accession=36491143, orientation='plus', exon_count=11, OMIM=615932, no_hdr0=''),
  222659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222659, CurrentID=0, Status='live', Symbol='PXT1', Aliases='STEPP', description='peroxisomal, testis specific 1', other_designations='small testis-specific peroxisomal protein', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36390551, end_position_on_the_genomic_accession=36442889, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375663, CurrentID=0, Status='live', Symbol='LOC105375663', Aliases='', description='uncharacterized LOC105375663', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99718973, end_position_on_the_genomic_accession=99784307, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  222662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222662, CurrentID=0, Status='live', Symbol='LHFPL5', Aliases='DFNB67, TMHS, dJ510O8.8', description='lipoma HMGIC fusion partner-like 5', other_designations='LHFP-like protein 5|lipoma HMGIC fusion partner-like 5 protein', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35797420, end_position_on_the_genomic_accession=35824075, orientation='plus', exon_count=5, OMIM=609427, no_hdr0=''),
  222663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222663, CurrentID=0, Status='live', Symbol='SCUBE3', Aliases='CEGF3', description='signal peptide, CUB domain, EGF-like 3', other_designations='CUB domain and EGF-like repeat containing 3|signal peptide, CUB and EGF-like domain containing protein 3|signal peptide-, CUB domain-, and EGF-like domains-containing protein 3', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35214047, end_position_on_the_genomic_accession=35253079, orientation='plus', exon_count=25, OMIM=614708, no_hdr0=''),
  91603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91603, CurrentID=0, Status='live', Symbol='ZNF830', Aliases='CCDC16, OMCG1', description='zinc finger protein 830', other_designations='coiled-coil domain containing 16|coiled-coil domain-containing protein 16|hypothetical protein MGC20398|orphan maintenance of genome 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34961530, end_position_on_the_genomic_accession=34963186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  91607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91607, CurrentID=0, Status='live', Symbol='SLFN11', Aliases='SLFN8/9', description='schlafen family member 11', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35350305, end_position_on_the_genomic_accession=35373853, orientation='minus', exon_count=7, OMIM=614953, no_hdr0=''),
  91608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91608, CurrentID=0, Status='live', Symbol='RASL10B', Aliases='RRP17, VTS58635', description='RAS-like family 10 member B', other_designations='RAS-like, family 10, member B|Ras-related protein 17|ras-like protein VTS58635', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35731660, end_position_on_the_genomic_accession=35743525, orientation='plus', exon_count=4, OMIM=612128, no_hdr0=''),
  91612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91612, CurrentID=0, Status='live', Symbol='CHURC1', Aliases='C14orf52, My015, chch', description='churchill domain containing 1', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64914361, end_position_on_the_genomic_accession=64935368, orientation='plus', exon_count=4, OMIM=608577, no_hdr0=''),
  91614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91614, CurrentID=0, Status='live', Symbol='DEPDC7', Aliases='TR2, dJ85M6.4', description='DEP domain containing 7', other_designations='dJ85M6.4 (novel 58.3 KDA protein)|novel 58.3 KDA protein', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33015864, end_position_on_the_genomic_accession=33033582, orientation='plus', exon_count=10, OMIM=612294, no_hdr0=''),
  105375312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375312, CurrentID=0, Status='live', Symbol='LOC105375312', Aliases='', description='Ig heavy chain V region 914-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  222696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222696, CurrentID=0, Status='live', Symbol='ZSCAN23', Aliases='ZNF390, ZNF453, dJ29K1.3, dJ29K1.3.1', description='zinc finger and SCAN domain containing 23', other_designations='zinc finger protein 390|zinc finger protein 453', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28425531, end_position_on_the_genomic_accession=28443974, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  222698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222698, CurrentID=0, Status='live', Symbol='NKAPL', Aliases='C6orf194, bA424I5.1', description='NFKB activating protein like', other_designations='NFKB activating protein-like', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28259297, end_position_on_the_genomic_accession=28260958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  222699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222699, CurrentID=0, Status='live', Symbol='TOB2P1', Aliases='TOB2P, TOB4p, p373c6.3', description='transducer of ERBB2, 2 pseudogene 1', other_designations='CRA_a protein', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28215338, end_position_on_the_genomic_accession=28218929, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  222701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222701, CurrentID=0, Status='live', Symbol='ZNF192P2', Aliases='', description='zinc finger protein 192 pseudogene 2', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28187938, end_position_on_the_genomic_accession=28189411, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421490, CurrentID=0, Status='live', Symbol='LOC100421490', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C9 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 9 pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10613239, end_position_on_the_genomic_accession=10614173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421934, CurrentID=0, Status='live', Symbol='LOC100421934', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421951, CurrentID=0, Status='live', Symbol='LOC100421951', Aliases='', description='olfactory receptor, family 5, subfamily P, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  91646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91646, CurrentID=0, Status='live', Symbol='TDRD12', Aliases='ECAT8', description='tudor domain containing 12', other_designations='ES cell associated transcript 8|ES cell-associated transcript 8 protein|tudor domain-containing protein 12', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32719773, end_position_on_the_genomic_accession=32827415, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  91647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91647, CurrentID=0, Status='live', Symbol='ATPAF2', Aliases='ATP12, ATP12p, LP3663, MC5DN1', description='ATP synthase mitochondrial F1 complex assembly factor 2', other_designations='ATP12 homolog', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17977781, end_position_on_the_genomic_accession=18039169, orientation='minus', exon_count=15, OMIM=608918, no_hdr0=''),
  101927694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927694, CurrentID=0, Status='live', Symbol='LOC101927694', Aliases='', description='uncharacterized LOC101927694', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128118217, end_position_on_the_genomic_accession=128121952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91653, CurrentID=0, Status='live', Symbol='BOC', Aliases='CDON2', description='BOC cell adhesion associated, oncogene regulated', other_designations='Boc homolog|brother of CDON|cell adhesion associated, oncogene regulated 2', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113211003, end_position_on_the_genomic_accession=113287463, orientation='plus', exon_count=23, OMIM=608708, no_hdr0=''),
  100132439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132439, CurrentID=0, Status='live', Symbol='FAM27E4', Aliases='FAM27E4P', description='family with sequence similarity 27 member E4', other_designations='family with sequence similarity 27, member E4|family with sequence similarity 27, member E4, pseudogene|protein FAM27E3-like', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=44998147, end_position_on_the_genomic_accession=44998488, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  91661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91661, CurrentID=0, Status='live', Symbol='ZNF765', Aliases='', description='zinc finger protein 765', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53387860, end_position_on_the_genomic_accession=53412009, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  91662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91662, CurrentID=0, Status='live', Symbol='NLRP12', Aliases='CLR19.3, FCAS2, NALP12, PAN6, PYPAF7, RNO, RNO2', description='NLR family, pyrin domain containing 12', other_designations='PYRIN-containing APAF1-like protein 7|monarch 1|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 12|regulated by nitric oxide', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53793584, end_position_on_the_genomic_accession=53824403, orientation='minus', exon_count=10, OMIM=609648, no_hdr0=''),
  91663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91663, CurrentID=0, Status='live', Symbol='MYADM', Aliases='SB135', description='myeloid-associated differentiation marker', other_designations='myeloid upregulated protein', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53865584, end_position_on_the_genomic_accession=53876435, orientation='plus', exon_count=8, OMIM=609959, no_hdr0=''),
  91664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91664, CurrentID=0, Status='live', Symbol='ZNF845', Aliases='', description='zinc finger protein 845', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53333714, end_position_on_the_genomic_accession=53354869, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375321, CurrentID=0, Status='live', Symbol='LOC105375321', Aliases='', description='uncharacterized LOC105375321', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64371319, end_position_on_the_genomic_accession=64374770, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91683, CurrentID=0, Status='live', Symbol='SYT12', Aliases='SYT11, sytXII', description='synaptotagmin 12', other_designations='synaptotagmin XII|synaptotagmin-XII', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67006773, end_position_on_the_genomic_accession=67050863, orientation='plus', exon_count=14, OMIM=606436, no_hdr0=''),
  91687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91687, CurrentID=0, Status='live', Symbol='CENPL', Aliases='C1orf155, CENP-L, dJ383J4.3', description='centromere protein L', other_designations='interphase centromere complex protein 33', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173799550, end_position_on_the_genomic_accession=173824639, orientation='minus', exon_count=7, OMIM=611503, no_hdr0=''),
  91689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91689, CurrentID=0, Status='live', Symbol='SMDT1', Aliases='C22orf32, DDDD, EMRE, dJ186O1.1', description='single-pass membrane protein with aspartate-rich tail 1', other_designations='UPF0466 protein C22orf32, mitochondrial|single-pass membrane protein with aspartate-rich tail 1, mitochondrial', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42079691, end_position_on_the_genomic_accession=42084284, orientation='plus', exon_count=3, OMIM=615588, no_hdr0=''),
  91694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91694, CurrentID=0, Status='live', Symbol='LONRF1', Aliases='RNF191', description='LON peptidase N-terminal domain and ring finger 1', other_designations='RING finger protein 191', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12721897, end_position_on_the_genomic_accession=12755485, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  91695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91695, CurrentID=0, Status='live', Symbol='RRP7BP', Aliases='RRP7B, dJ222E13.2', description='ribosomal RNA processing 7 homolog B, pseudogene', other_designations='ribosomal RNA processing 7 homolog B (S. cerevisiae), pseudogene', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42573244, end_position_on_the_genomic_accession=42582011, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375326, CurrentID=0, Status='live', Symbol='LOC105375326', Aliases='', description='uncharacterized LOC105375326', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64633090, end_position_on_the_genomic_accession=64636078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91703, CurrentID=0, Status='live', Symbol='ACY3', Aliases='ACY-3, HCBP1', description='aminoacylase 3', other_designations='HCV core-binding protein 1|acylase III|aminoacylase-3|aspartoacylase (aminoacylase) 3|aspartoacylase (aminocyclase) 3|aspartoacylase-2|hepatitis C virus core-binding protein 1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67642555, end_position_on_the_genomic_accession=67650807, orientation='minus', exon_count=9, OMIM=614413, no_hdr0=''),
  105375328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375328, CurrentID=0, Status='live', Symbol='LOC105375328', Aliases='', description='uncharacterized LOC105375328', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64947546, end_position_on_the_genomic_accession=64950664, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375330, CurrentID=0, Status='live', Symbol='LOC105375330', Aliases='', description='uncharacterized LOC105375330', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65008674, end_position_on_the_genomic_accession=65038261, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375331, CurrentID=0, Status='live', Symbol='LOC105375331', Aliases='', description='uncharacterized LOC105375331', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65059353, end_position_on_the_genomic_accession=65081281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91734, CurrentID=0, Status='live', Symbol='IDI2', Aliases='IPPI2', description='isopentenyl-diphosphate delta isomerase 2', other_designations='IPP isomerase 2|diphosphate dimethylallyl diphosphate isomerase 2|isopentenyl pyrophosphate isomerase 2|isopentenyl-diphosphate Delta-isomerase 2', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1018907, end_position_on_the_genomic_accession=1025859, orientation='minus', exon_count=5, OMIM=615389, no_hdr0=''),
  91746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91746, CurrentID=0, Status='live', Symbol='YTHDC1', Aliases='YT521, YT521-B', description='YTH domain containing 1', other_designations='putative splicing factor YT521|splicing factor YT521-B', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68310387, end_position_on_the_genomic_accession=68350106, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  91748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91748, CurrentID=0, Status='live', Symbol='ELMSAN1', Aliases='C14orf117, C14orf43, LSR68, MIDEAS, c14_5541', description='ELM2 and Myb/SANT-like domain containing 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73715122, end_position_on_the_genomic_accession=73790285, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  91749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91749, CurrentID=0, Status='live', Symbol='KIAA1919', Aliases='MFSD4B, NaGLT1', description='KIAA1919', other_designations='', map_location='6q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111259279, end_position_on_the_genomic_accession=111269060, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91750, CurrentID=0, Status='live', Symbol='LIN52', Aliases='C14orf46, c14_5549', description='lin-52 DREAM MuvB core complex component', other_designations='lin-52 homolog', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74084836, end_position_on_the_genomic_accession=74200414, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  91752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91752, CurrentID=0, Status='live', Symbol='ZNF804A', Aliases='C2orf10', description='zinc finger protein 804A', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184598366, end_position_on_the_genomic_accession=184939487, orientation='plus', exon_count=4, OMIM=612282, no_hdr0=''),
  91754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91754, CurrentID=0, Status='live', Symbol='NEK9', Aliases='NERCC, NERCC1, Nek8', description='NIMA-related kinase 9', other_designations='NIMA (never in mitosis gene a)- related kinase 9|NIMA-related kinase Nek8|nercc1 kinase|never in mitosis A-related kinase 9|nimA-related kinase 8|nimA-related protein kinase 9', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75082115, end_position_on_the_genomic_accession=75127075, orientation='minus', exon_count=22, OMIM=609798, no_hdr0=''),
  105375335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375335, CurrentID=0, Status='live', Symbol='LOC105375335', Aliases='', description='uncharacterized LOC105375335', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65828902, end_position_on_the_genomic_accession=65863735, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375712, CurrentID=0, Status='live', Symbol='LOC105375712', Aliases='', description='uncharacterized LOC105375712', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116604519, end_position_on_the_genomic_accession=116615299, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375336, CurrentID=0, Status='live', Symbol='LOC105375336', Aliases='', description='uncharacterized LOC105375336', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66000864, end_position_on_the_genomic_accession=66038095, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  386665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386665, CurrentID=0, Status='live', Symbol='AURKAPS2', Aliases='STK6LP', description='aurora kinase A pseudogene 2', other_designations='serine/threonine kinase 6-like pseudogene', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114246000, end_position_on_the_genomic_accession=114247197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91768, CurrentID=0, Status='live', Symbol='CABLES1', Aliases='CABL1, CABLES, HsT2563, IK3-1', description='Cdk5 and Abl enzyme substrate 1', other_designations='interactor with CDK3 1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23134564, end_position_on_the_genomic_accession=23260470, orientation='plus', exon_count=12, OMIM=609194, no_hdr0=''),
  91775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91775, CurrentID=0, Status='live', Symbol='NXPE3', Aliases='FAM55C, MST115, MSTP115', description='neurexophilin and PC-esterase domain family member 3', other_designations='family with sequence similarity 55, member C|neurexophilin and PC-esterase domain family, member 3|protein FAM55C', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101779185, end_position_on_the_genomic_accession=101828231, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105373412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373412, CurrentID=0, Status='live', Symbol='LOC105373412', Aliases='', description='uncharacterized LOC105373412', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8679242, end_position_on_the_genomic_accession=8681930, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91782, CurrentID=0, Status='live', Symbol='CHMP7', Aliases='', description='charged multivesicular body protein 7', other_designations='CHMP family, member 7|chromatin-modifying protein 7', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23243637, end_position_on_the_genomic_accession=23261999, orientation='plus', exon_count=12, OMIM=611130, no_hdr0=''),
  80833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80833, CurrentID=0, Status='live', Symbol='APOL3', Aliases='APOLIII, CG121, CG12_1, apoL-III', description='apolipoprotein L3', other_designations='TNF-inducible protein CG12-1|apolipoprotein L, 3|apolipoprotein L-III', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36140323, end_position_on_the_genomic_accession=36166177, orientation='minus', exon_count=10, OMIM=607253, no_hdr0=''),
  285176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285176, CurrentID=0, Status='live', Symbol='RPL10P6', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214847137, end_position_on_the_genomic_accession=214847502, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  222865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222865, CurrentID=0, Status='live', Symbol='TMEM130', Aliases='', description='transmembrane protein 130', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98846488, end_position_on_the_genomic_accession=98870050, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=80835, CurrentID=0, Status='live', Symbol='TAS1R1', Aliases='GM148, GPR70, T1R1, TR1', description='taste 1 receptor member 1', other_designations='G protein-coupled receptor 70|G-protein coupled receptor 70|seven transmembrane helix receptor|taste receptor, type 1, member 1', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6555278, end_position_on_the_genomic_accession=6579757, orientation='plus', exon_count=9, OMIM=606225, no_hdr0=''),
  91801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91801, CurrentID=0, Status='live', Symbol='ALKBH8', Aliases='ABH8, TRM9, TRMT9', description='alkB homolog 8, tRNA methyltransferase', other_designations='AlkB homologue 8|S-adenosyl-L-methionine-dependent tRNA methyltransferase ABH8|alkB, alkylation repair homolog 8|probable alpha-ketoglutarate-dependent dioxygenase ABH8|tRNA (carboxymethyluridine(34)-5-O)-methyltransferase ABH8|tRNA methyltransferase 9 homolog', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107502726, end_position_on_the_genomic_accession=107565737, orientation='minus', exon_count=13, OMIM=613306, no_hdr0=''),
  386671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386671, CurrentID=0, Status='live', Symbol='FLJ44715', Aliases='', description='uncharacterized LOC386671', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375158, CurrentID=0, Status='live', Symbol='LOC105375158', Aliases='', description='uncharacterized LOC105375158', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12931340, end_position_on_the_genomic_accession=13061535, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91807, CurrentID=0, Status='live', Symbol='MYLK3', Aliases='MLCK, MLCK2, caMLCK', description='myosin light chain kinase 3', other_designations='MLC kinase|cardiac-MLCK|cardiac-MyBP-C associated Ca/CaM kinase|cardiac-MyBP-C-associated Ca/CaM kinase|putative myosin light chain kinase 3', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46702282, end_position_on_the_genomic_accession=46763289, orientation='minus', exon_count=16, OMIM=612147, no_hdr0=''),
  100132464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132464, CurrentID=0, Status='live', Symbol='FAM99B', Aliases='', description='family with sequence similarity 99 member B (non-protein coding)', other_designations='family with sequence similarity 99, member B (non-protein coding)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1683270, end_position_on_the_genomic_accession=1685629, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375346, CurrentID=0, Status='live', Symbol='LOC105375346', Aliases='', description='uncharacterized LOC105375346', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69544025, end_position_on_the_genomic_accession=69580727, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  222894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222894, CurrentID=0, Status='live', Symbol='FERD3L', Aliases='N-TWIST, NATO3, NTWIST, bHLHa31', description='Fer3-like bHLH transcription factor', other_designations='basic helix-loop-helix protein N-twist|class A basic helix-loop-helix protein 31|nephew of atonal 3|neuronal twist', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19144782, end_position_on_the_genomic_accession=19145421, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132467, CurrentID=0, Status='live', Symbol='RNF2P1', Aliases='RNF2P', description='ring finger protein 2 pseudogene 1', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6669691, end_position_on_the_genomic_accession=6670633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  91828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91828, CurrentID=0, Status='live', Symbol='EXOC3L4', Aliases='C14orf73', description='exocyst complex component 3-like 4', other_designations='SEC6-like protein C14orf73', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103094410, end_position_on_the_genomic_accession=103110562, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  222901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222901, CurrentID=0, Status='live', Symbol='RPL23P8', Aliases='RPL23_4_790', description='ribosomal protein L23 pseudogene 8', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20827298, end_position_on_the_genomic_accession=20827820, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  91833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91833, CurrentID=0, Status='live', Symbol='WDR20', Aliases='DMR', description='WD repeat domain 20', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102139842, end_position_on_the_genomic_accession=102224839, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105375349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375349, CurrentID=0, Status='live', Symbol='LOC105375349', Aliases='', description='uncharacterized LOC105375349', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73210536, end_position_on_the_genomic_accession=73211198, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  648822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648822, CurrentID=0, Status='live', Symbol='HMGN2P18', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 18', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 18', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155148471, end_position_on_the_genomic_accession=155150278, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  91851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91851, CurrentID=0, Status='live', Symbol='CHRDL1', Aliases='CHL, MGC1, MGCN, NRLN1, VOPT, dA141H5.1', description='chordin-like 1', other_designations='neuralin 1|neuralin-1|neurogenesin-1|ventroptin', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110673856, end_position_on_the_genomic_accession=110796058, orientation='minus', exon_count=12, OMIM=300350, no_hdr0=''),
  105375352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375352, CurrentID=0, Status='live', Symbol='LOC105375352', Aliases='', description='uncharacterized LOC105375352', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74947726, end_position_on_the_genomic_accession=74954289, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  91860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91860, CurrentID=0, Status='live', Symbol='CALML4', Aliases='NY-BR-20', description='calmodulin like 4', other_designations='calmodulin-like 4|serologically defined breast cancer antigen NY-BR-20', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68190705, end_position_on_the_genomic_accession=68206110, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  91862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91862, CurrentID=0, Status='live', Symbol='MARVELD3', Aliases='MARVD3, MRVLDC3', description='MARVEL domain containing 3', other_designations='MARVEL (membrane-associating) domain containing 3', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71626138, end_position_on_the_genomic_accession=71641965, orientation='plus', exon_count=5, OMIM=614094, no_hdr0=''),
  105375353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375353, CurrentID=0, Status='live', Symbol='LOC105375353', Aliases='', description='uncharacterized LOC105375353', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75181459, end_position_on_the_genomic_accession=75182121, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  392543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392543, CurrentID=0, Status='live', Symbol='LOC392543', Aliases='', description='acylglycerol kinase pseudogene', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=132130089, end_position_on_the_genomic_accession=132131352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386682, CurrentID=0, Status='live', Symbol='KRTAP10-3', Aliases='KAP10.3, KAP18-3, KAP18.3, KRTAP10.3, KRTAP18-3, KRTAP18.3', description='keratin associated protein 10-3', other_designations='high sulfur keratin-associated protein 10.3|keratin-associated protein 10.3|keratin-associated protein 18-3|keratin-associated protein 18.3', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44558023, end_position_on_the_genomic_accession=44558760, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  91875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91875, CurrentID=0, Status='live', Symbol='TTC5', Aliases='Strap', description='tetratricopeptide repeat domain 5', other_designations='TPR repeat protein 5|stress-responsive activator of p300', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20289131, end_position_on_the_genomic_accession=20305994, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101410533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410533, CurrentID=0, Status='live', Symbol='YTHDF3-AS1', Aliases='', description='YTHDF3 antisense RNA 1 (head to head)', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63167725, end_position_on_the_genomic_accession=63168442, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  222950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222950, CurrentID=0, Status='live', Symbol='NYAP1', Aliases='C7orf51', description='neuronal tyrosine-phosphorylated phosphoinositide-3-kinase adaptor 1', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100483862, end_position_on_the_genomic_accession=100494802, orientation='plus', exon_count=8, OMIM=615477, no_hdr0=''),
  101410535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410535, CurrentID=0, Status='live', Symbol='LAMTOR5-AS1', Aliases='', description='LAMTOR5 antisense RNA 1', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110407809, end_position_on_the_genomic_accession=110416274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101410536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410536, CurrentID=0, Status='live', Symbol='EVX1-AS', Aliases='EVX1-AS1', description='EVX1 antisense RNA', other_designations='EVX1 antisense RNA 1 (non-protein coding)', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27241429, end_position_on_the_genomic_accession=27247229, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  386684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386684, CurrentID=0, Status='live', Symbol='KRTAP12-4', Aliases='KRTAP12.4', description='keratin associated protein 12-4', other_designations='high sulfur keratin-associated protein 12.4', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44654213, end_position_on_the_genomic_accession=44654659, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101410538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410538, CurrentID=0, Status='live', Symbol='MMP24-AS1', Aliases='', description='MMP24 antisense RNA 1', other_designations='', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35276333, end_position_on_the_genomic_accession=35278157, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101410540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410540, CurrentID=0, Status='live', Symbol='FANK1-AS1', Aliases='', description='FANK1 antisense RNA 1', other_designations='FANK1 antisense RNA 1 (non-protein coding)', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125972188, end_position_on_the_genomic_accession=125973273, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101410541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410541, CurrentID=0, Status='live', Symbol='LINC00678', Aliases='', description='long intergenic non-protein coding RNA 678', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27617626, end_position_on_the_genomic_accession=27634627, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101410542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410542, CurrentID=0, Status='live', Symbol='UCHL1-AS1', Aliases='', description='UCHL1 antisense RNA 1 (head to head)', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41220074, end_position_on_the_genomic_accession=41256727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101410543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410543, CurrentID=0, Status='live', Symbol='XXYLT1-AS2', Aliases='', description='XXYLT1 antisense RNA 2', other_designations='XXYLT1 antisense RNA 2 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195147871, end_position_on_the_genomic_accession=195152856, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101410544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410544, CurrentID=0, Status='live', Symbol='RMDN2-AS1', Aliases='', description='RMDN2 antisense RNA 1', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37950334, end_position_on_the_genomic_accession=38036341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101410545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101410545, CurrentID=0, Status='live', Symbol='STPG2-AS1', Aliases='', description='STPG2 antisense RNA 1', other_designations='STPG2 antisense RNA 1 (non-protein coding)', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=97366926, end_position_on_the_genomic_accession=97490164, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  222962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222962, CurrentID=0, Status='live', Symbol='SLC29A4', Aliases='ENT4, PMAT', description='solute carrier family 29 (equilibrative nucleoside transporter), member 4', other_designations='plasma membrane monoamine transporter|solute carrier family 29 (nucleoside transporters), member 4|solute carrier family 29 member 4', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5274374, end_position_on_the_genomic_accession=5304073, orientation='plus', exon_count=12, OMIM=609149, no_hdr0=''),
  91893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91893, CurrentID=0, Status='live', Symbol='FDXACB1', Aliases='hCG_2033039', description='ferredoxin-fold anticodon binding domain containing 1', other_designations='FDX-ACDB domain-containing protein 1|hypothetical protein BC006136', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111874056, end_position_on_the_genomic_accession=111879457, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  91894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91894, CurrentID=0, Status='live', Symbol='C11orf52', Aliases='', description='chromosome 11 open reading frame 52', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111918877, end_position_on_the_genomic_accession=111926871, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  222967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222967, CurrentID=0, Status='live', Symbol='RSPH10B', Aliases='', description='radial spoke head 10 homolog B (Chlamydomonas)', other_designations='radial spoke head 10 homolog B2', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5926136, end_position_on_the_genomic_accession=5970684, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100132479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132479, CurrentID=0, Status='live', Symbol='RPL22P5', Aliases='RPL22_2_96', description='ribosomal protein L22 pseudogene 5', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143929993, end_position_on_the_genomic_accession=143930382, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375360, CurrentID=0, Status='live', Symbol='LOC105375360', Aliases='', description='uncharacterized LOC105375360', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77125977, end_position_on_the_genomic_accession=77165747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375361, CurrentID=0, Status='live', Symbol='LOC105375361', Aliases='', description='uncharacterized LOC105375361', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76971188, end_position_on_the_genomic_accession=76977221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375362, CurrentID=0, Status='live', Symbol='LOC105375362', Aliases='', description='uncharacterized LOC105375362', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77446745, end_position_on_the_genomic_accession=77460536, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375363, CurrentID=0, Status='live', Symbol='LOC105375363', Aliases='', description='uncharacterized LOC105375363', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77653366, end_position_on_the_genomic_accession=77655912, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255620, CurrentID=0, Status='live', Symbol='UNC93B6', Aliases='', description='unc-93 homolog B6 pseudogene (C. elegans)', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71603164, end_position_on_the_genomic_accession=71608044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375365, CurrentID=0, Status='live', Symbol='LOC105375365', Aliases='', description='uncharacterized LOC105375365', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77911039, end_position_on_the_genomic_accession=77920260, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91937, CurrentID=0, Status='live', Symbol='TIMD4', Aliases='SMUCKLER, TIM4', description='T-cell immunoglobulin and mucin domain containing 4', other_designations='T-cell immunoglobulin mucin receptor 4|T-cell membrane protein 4|TIM-4|TIMD-4', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156919282, end_position_on_the_genomic_accession=156972310, orientation='minus', exon_count=11, OMIM=610096, no_hdr0=''),
  105375366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375366, CurrentID=0, Status='live', Symbol='LOC105375366', Aliases='', description='uncharacterized LOC105375366', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79139789, end_position_on_the_genomic_accession=79177270, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91942, CurrentID=0, Status='live', Symbol='NDUFAF2', Aliases='B17.2L, MMTN, NDUFA12L, mimitin', description='NADH:ubiquinone oxidoreductase complex assembly factor 2', other_designations='Myc-induced mitochondrial protein|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 2|NADH dehydrogenase (ubiquinone) complex I, assembly factor 2|NADH dehydrogenase 1 alpha subcomplex assembly factor 2|NDUFA12-like protein', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60945129, end_position_on_the_genomic_accession=61153037, orientation='plus', exon_count=4, OMIM=609653, no_hdr0=''),
  91947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91947, CurrentID=0, Status='live', Symbol='ARRDC4', Aliases='', description='arrestin domain containing 4', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97960703, end_position_on_the_genomic_accession=97973838, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  91948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91948, CurrentID=0, Status='live', Symbol='LINC00923', Aliases='', description='long intergenic non-protein coding RNA 923', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97742616, end_position_on_the_genomic_accession=97874429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  91949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91949, CurrentID=0, Status='live', Symbol='COG7', Aliases='CDG2E', description='component of oligomeric golgi complex 7', other_designations='COG complex subunit 7', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23388493, end_position_on_the_genomic_accession=23453210, orientation='minus', exon_count=18, OMIM=606978, no_hdr0=''),
  100132488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132488, CurrentID=0, Status='live', Symbol='RPS27P6', Aliases='RPS27_2_83', description='ribosomal protein S27 pseudogene 6', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109030032, end_position_on_the_genomic_accession=109030376, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375369, CurrentID=0, Status='live', Symbol='LOC105375369', Aliases='', description='uncharacterized LOC105375369', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79634687, end_position_on_the_genomic_accession=79670974, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  386698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386698, CurrentID=0, Status='live', Symbol='OFD1P14Y', Aliases='OFD1P14, OFD1PY14, OFDYP14', description='OFD1 pseudogene 14, Y-linked', other_designations='OFD1 pseudogene, Y-linked 14|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25871831, end_position_on_the_genomic_accession=25887981, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  91966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91966, CurrentID=0, Status='live', Symbol='CXorf40A', Aliases='CXorf40, EOLA1', description='chromosome X open reading frame 40A', other_designations='endothelial-overexpressed lipopolysaccharide-associated factor 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149540715, end_position_on_the_genomic_accession=149555340, orientation='plus', exon_count=9, OMIM=300954, no_hdr0=''),
  105375371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375371, CurrentID=0, Status='live', Symbol='LOC105375371', Aliases='', description='uncharacterized LOC105375371', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79768025, end_position_on_the_genomic_accession=79839246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  91975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91975, CurrentID=0, Status='live', Symbol='ZNF300', Aliases='', description='zinc finger protein 300', other_designations='kruppel-like zinc finger protein', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150894383, end_position_on_the_genomic_accession=150904983, orientation='minus', exon_count=8, OMIM=612429, no_hdr0=''),
  91977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91977, CurrentID=0, Status='live', Symbol='MYOZ3', Aliases='CS-3, CS3, FRP3', description='myozenin 3', other_designations='FATZ-related protein 3|calsarcin-3', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150660841, end_position_on_the_genomic_accession=150679700, orientation='plus', exon_count=7, OMIM=610735, no_hdr0=''),
  91978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91978, CurrentID=0, Status='live', Symbol='TPGS1', Aliases='C19orf20, GTRGEO22, PGs1', description='tubulin polyglutamylase complex subunit 1', other_designations='gene trap ROSA b-geo 22', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=507497, end_position_on_the_genomic_accession=519654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375375, CurrentID=0, Status='live', Symbol='LOC105375375', Aliases='', description='uncharacterized LOC105375375', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82311743, end_position_on_the_genomic_accession=82313714, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375376, CurrentID=0, Status='live', Symbol='LOC105375376', Aliases='', description='uncharacterized LOC105375376', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82690719, end_position_on_the_genomic_accession=82695876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  92002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92002, CurrentID=0, Status='live', Symbol='FAM58A', Aliases='STAR', description='family with sequence similarity 58 member A', other_designations='CDK10-activating cyclin|cyclin M|family with sequence similarity 58, member A', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153587925, end_position_on_the_genomic_accession=153599177, orientation='minus', exon_count=7, OMIM=300708, no_hdr0=''),
  223075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=223075, CurrentID=0, Status='live', Symbol='CCDC129', Aliases='', description='coiled-coil domain containing 129', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31514071, end_position_on_the_genomic_accession=31658720, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  223082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=223082, CurrentID=0, Status='live', Symbol='ZNRF2', Aliases='RNF202', description='zinc and ring finger 2, E3 ubiquitin protein ligase', other_designations='RING finger protein 202|protein Ells2|zinc finger/RING finger 2|zinc/RING finger protein 2', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30284307, end_position_on_the_genomic_accession=30367692, orientation='plus', exon_count=9, OMIM=612061, no_hdr0=''),
  100132498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132498, CurrentID=0, Status='live', Symbol='LOC100132498', Aliases='', description='centromere protein N pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87221132, end_position_on_the_genomic_accession=87222488, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92014, CurrentID=0, Status='live', Symbol='SLC25A51', Aliases='CG7943, MCART1', description='solute carrier family 25 member 51', other_designations='mitochondrial carrier triple repeat 1|mitochondrial carrier triple repeat protein 1|solute carrier family 25, member 51', map_location='9p13.3-p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37877575, end_position_on_the_genomic_accession=37904353, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  92017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92017, CurrentID=0, Status='live', Symbol='SNX29', Aliases='A-388D4.1, RUNDC2A', description='sorting nexin 29', other_designations='RUN domain containing 2A|RUN domain-containing protein 2A', map_location='16p13.13-p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12570192, end_position_on_the_genomic_accession=12574287, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  105375379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375379, CurrentID=0, Status='live', Symbol='LOC105375379', Aliases='', description='uncharacterized LOC105375379', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84648536, end_position_on_the_genomic_accession=84880517, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132501, CurrentID=0, Status='live', Symbol='LINC01415', Aliases='', description='long intergenic non-protein coding RNA 1415', other_designations='TCONS_00026208', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55774601, end_position_on_the_genomic_accession=55781721, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  223117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=223117, CurrentID=0, Status='live', Symbol='SEMA3D', Aliases='Sema-Z2, coll-2', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3D', other_designations='collapsin 2', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84995192, end_position_on_the_genomic_accession=85186855, orientation='minus', exon_count=21, OMIM=609907, no_hdr0=''),
  105375384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375384, CurrentID=0, Status='live', Symbol='LOC105375384', Aliases='', description='uncharacterized LOC105375384', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87199302, end_position_on_the_genomic_accession=87202754, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372541, CurrentID=0, Status='live', Symbol='LOC105372541', Aliases='', description='uncharacterized LOC105372541', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16589897, end_position_on_the_genomic_accession=16605499, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375386, CurrentID=0, Status='live', Symbol='LOC105375386', Aliases='', description='uncharacterized LOC105375386', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88140085, end_position_on_the_genomic_accession=88157816, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  92070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92070, CurrentID=0, Status='live', Symbol='CTBP1-AS2', Aliases='C4orf42, CTBP1-AS1', description='CTBP1 antisense RNA 2 (head to head)', other_designations='CTBP1 antisense RNA 1 (head to head)|CTBP1 antisense RNA 1 (non-protein coding)', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1249440, end_position_on_the_genomic_accession=1253007, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375389, CurrentID=0, Status='live', Symbol='LOC105375389', Aliases='', description='uncharacterized LOC105375389', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90235143, end_position_on_the_genomic_accession=90243164, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129526, CurrentID=0, Status='live', Symbol='LOC100129526', Aliases='', description='protein tyrosine phosphatase, receptor type, D pseudogene', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117717325, end_position_on_the_genomic_accession=117722097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375390, CurrentID=0, Status='live', Symbol='LOC105375390', Aliases='', description='uncharacterized LOC105375390', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90355336, end_position_on_the_genomic_accession=90418856, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92086, CurrentID=0, Status='live', Symbol='GGTLC1', Aliases='GGTL6, GGTLA3, GGTLA4, dJ831C21.1, dJ831C21.2', description='gamma-glutamyltransferase light chain 1', other_designations='gamma-glutamyl transpeptidase|gamma-glutamyltransferase-like activity 3|gamma-glutamyltransferase-like activity 4|gamma-glutamyltransferase-like protein 6', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23985053, end_position_on_the_genomic_accession=23988779, orientation='minus', exon_count=9, OMIM=612338, no_hdr0=''),
  92092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92092, CurrentID=0, Status='live', Symbol='ZC3HAV1L', Aliases='C7orf39', description='zinc finger CCCH-type, antiviral 1-like', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139021498, end_position_on_the_genomic_accession=139036489, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375393, CurrentID=0, Status='live', Symbol='LOC105375393', Aliases='', description='uncharacterized LOC105375393', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91515330, end_position_on_the_genomic_accession=91556282, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92104, CurrentID=0, Status='live', Symbol='TTC30A', Aliases='', description='tetratricopeptide repeat domain 30A', other_designations='TPR repeat protein 30A', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177614298, end_position_on_the_genomic_accession=177618966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  92105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92105, CurrentID=0, Status='live', Symbol='INTS4', Aliases='INT4, MST093', description='integrator complex subunit 4', other_designations='MSTP093', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77878720, end_position_on_the_genomic_accession=77994842, orientation='minus', exon_count=26, OMIM=611348, no_hdr0=''),
  92106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92106, CurrentID=0, Status='live', Symbol='OXNAD1', Aliases='', description='oxidoreductase NAD-binding domain containing 1', other_designations='', map_location='3p25-p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16265160, end_position_on_the_genomic_accession=16386979, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100129527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129527, CurrentID=0, Status='live', Symbol='LOC100129527', Aliases='', description='ESF1, nucleolar pre-rRNA processing protein, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72429621, end_position_on_the_genomic_accession=72430122, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375396, CurrentID=0, Status='live', Symbol='LOC105375396', Aliases='', description='bax inhibitor 1-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92412550, end_position_on_the_genomic_accession=92437456, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  92126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92126, CurrentID=0, Status='live', Symbol='DSEL', Aliases='C18orf4', description='dermatan sulfate epimerase-like', other_designations='NCAG1', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67506582, end_position_on_the_genomic_accession=67516730, orientation='minus', exon_count=2, OMIM=611125, no_hdr0=''),
  105375397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375397, CurrentID=0, Status='live', Symbol='LOC105375397', Aliases='', description='uncharacterized LOC105375397', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92591933, end_position_on_the_genomic_accession=92600603, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92129, CurrentID=0, Status='live', Symbol='RIPPLY1', Aliases='', description='ripply transcriptional repressor 1', other_designations='ripply1 homolog', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106900063, end_position_on_the_genomic_accession=106903331, orientation='minus', exon_count=4, OMIM=300575, no_hdr0=''),
  255654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255654, CurrentID=0, Status='live', Symbol='LOC255654', Aliases='', description='uncharacterized LOC255654', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246516039, end_position_on_the_genomic_accession=246524287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  386727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386727, CurrentID=0, Status='live', Symbol='FAM8A9P', Aliases='', description='family with sequence similarity 8 member A9, pseudogene', other_designations='family with sequence similarity 8, member A9 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20610114, end_position_on_the_genomic_accession=20612062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  92140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92140, CurrentID=0, Status='live', Symbol='MTDH', Aliases='3D3, AEG-1, AEG1, LYRIC, LYRIC/3D3', description='metadherin', other_designations='3D3/LYRIC|astrocyte elevated gene 1|astrocyte elevated gene-1 protein|lysine-rich CEACAM1 co-isolated protein|metastasis adhesion protein', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97644147, end_position_on_the_genomic_accession=97730260, orientation='plus', exon_count=13, OMIM=610323, no_hdr0=''),
  100132520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132520, CurrentID=0, Status='live', Symbol='MTRF1LP1', Aliases='', description='mitochondrial translational release factor 1 like pseudogene 1', other_designations='mitochondrial translational release factor 1-like pseudogene 1', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113711730, end_position_on_the_genomic_accession=113712399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375401, CurrentID=0, Status='live', Symbol='LOC105375401', Aliases='', description='uncharacterized LOC105375401', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93890844, end_position_on_the_genomic_accession=93893588, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  92154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92154, CurrentID=0, Status='live', Symbol='MTSS1L', Aliases='ABBA-1', description='metastasis suppressor 1-like', other_designations='actin-bundling protein with BAIAP2 homology|actin-bundling with BAIAP2 homology protein 1', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70661204, end_position_on_the_genomic_accession=70686051, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105375402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375402, CurrentID=0, Status='live', Symbol='LOC105375402', Aliases='', description='uncharacterized LOC105375402', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93914500, end_position_on_the_genomic_accession=93937436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92162, CurrentID=0, Status='live', Symbol='TMEM88', Aliases='', description='transmembrane protein 88', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7854896, end_position_on_the_genomic_accession=7856099, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375403, CurrentID=0, Status='live', Symbol='LOC105375403', Aliases='', description='uncharacterized LOC105375403', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93966033, end_position_on_the_genomic_accession=93993694, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375404, CurrentID=0, Status='live', Symbol='LOC105375404', Aliases='', description='uncharacterized LOC105375404', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94477910, end_position_on_the_genomic_accession=94564586, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  92170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92170, CurrentID=0, Status='live', Symbol='MTG1', Aliases='GTP, GTPBP7', description='mitochondrial ribosome-associated GTPase 1', other_designations='GTP-binding protein 7 (putative)|mitochondrial GTPase 1 homolog', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133394117, end_position_on_the_genomic_accession=133420670, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  92181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92181, CurrentID=0, Status='live', Symbol='UBTD2', Aliases='DCUBP', description='ubiquitin domain containing 2', other_designations='ubiquitin-like protein SB72', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172209645, end_position_on_the_genomic_accession=172283791, orientation='minus', exon_count=4, OMIM=610174, no_hdr0=''),
  105375407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375407, CurrentID=0, Status='live', Symbol='LOC105375407', Aliases='', description='uncharacterized LOC105375407', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95553004, end_position_on_the_genomic_accession=95562440, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375408, CurrentID=0, Status='live', Symbol='LOC105375408', Aliases='', description='uncharacterized LOC105375408', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95679148, end_position_on_the_genomic_accession=95682153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  92196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92196, CurrentID=0, Status='live', Symbol='DAPL1', Aliases='', description='death associated protein like 1', other_designations='death associated protein-like 1|early epithelial differentiation-associated protein', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158795317, end_position_on_the_genomic_accession=158827457, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132529, CurrentID=0, Status='live', Symbol='TAT-AS1', Aliases='', description='TAT antisense RNA 1', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71565016, end_position_on_the_genomic_accession=71572438, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375410, CurrentID=0, Status='live', Symbol='LOC105375410', Aliases='', description='uncharacterized LOC105375410', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96358526, end_position_on_the_genomic_accession=96445087, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100421953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421953, CurrentID=0, Status='live', Symbol='LOC100421953', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92211, CurrentID=0, Status='live', Symbol='CDHR1', Aliases='CORD15, PCDH21, PRCAD, RP65', description='cadherin related family member 1', other_designations='MT-protocadherin|photoreceptor cadherin|protocadherin-21', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84194635, end_position_on_the_genomic_accession=84219621, orientation='plus', exon_count=19, OMIM=609502, no_hdr0=''),
  101060424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060424, CurrentID=0, Status='live', Symbol='LOC101060424', Aliases='', description='otoancorin-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375412, CurrentID=0, Status='live', Symbol='LOC105375412', Aliases='', description='uncharacterized LOC105375412', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96509238, end_position_on_the_genomic_accession=96531299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375413, CurrentID=0, Status='live', Symbol='LOC105375413', Aliases='', description='uncharacterized LOC105375413', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96703966, end_position_on_the_genomic_accession=96709009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373415, CurrentID=0, Status='live', Symbol='LOC105373415', Aliases='', description='uncharacterized LOC105373415', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9115766, end_position_on_the_genomic_accession=9127143, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132534, CurrentID=0, Status='live', Symbol='PDHA1P1', Aliases='', description='pyruvate dehydrogenase (lipoamide) alpha 1 pseudogene 1', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57800054, end_position_on_the_genomic_accession=57801449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92235, CurrentID=0, Status='live', Symbol='DUSP27', Aliases='', description='dual specificity phosphatase 27 (putative)', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167094077, end_position_on_the_genomic_accession=167129897, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  92241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92241, CurrentID=0, Status='live', Symbol='RCSD1', Aliases='CAPZIP, MK2S4', description='RCSD domain containing 1', other_designations='CapZ interacting protein|RCSD domain-containing protein 1|protein kinase substrate CapZIP|protein kinase substrate MK2S4', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167630237, end_position_on_the_genomic_accession=167708696, orientation='plus', exon_count=7, OMIM=610579, no_hdr0=''),
  101929831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929831, CurrentID=0, Status='live', Symbol='LOC101929831', Aliases='', description='uncharacterized LOC101929831', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79632706, end_position_on_the_genomic_accession=79659866, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132537, CurrentID=0, Status='live', Symbol='DOCK11P1', Aliases='', description='dedicator of cytokinesis 11 pseudogene 1', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44859958, end_position_on_the_genomic_accession=44866240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372437, CurrentID=0, Status='live', Symbol='LOC105372437', Aliases='', description='uncharacterized LOC105372437', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50329710, end_position_on_the_genomic_accession=50353028, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  92249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92249, CurrentID=0, Status='live', Symbol='LINC01278', Aliases='', description='long intergenic non-protein coding RNA 1278', other_designations='', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63426559, end_position_on_the_genomic_accession=63560993, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  124602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124602, CurrentID=0, Status='live', Symbol='KIF19', Aliases='KIF19A', description='kinesin family member 19', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74326212, end_position_on_the_genomic_accession=74355820, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  92255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92255, CurrentID=0, Status='live', Symbol='LMBRD2', Aliases='', description='LMBR1 domain containing 2', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36098419, end_position_on_the_genomic_accession=36151913, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  92259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92259, CurrentID=0, Status='live', Symbol='MRPS36', Aliases='DC47, MRP-S36', description='mitochondrial ribosomal protein S36', other_designations='S36mt', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69217746, end_position_on_the_genomic_accession=69230158, orientation='plus', exon_count=4, OMIM=611996, no_hdr0=''),
  92270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92270, CurrentID=0, Status='live', Symbol='ATP6AP1L', Aliases='', description='ATPase, H+ transporting, lysosomal accessory protein 1-like', other_designations='vacuolar proton pump subunit S1-like protein', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82278665, end_position_on_the_genomic_accession=82344260, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105375421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375421, CurrentID=0, Status='live', Symbol='LOC105375421', Aliases='', description='uncharacterized LOC105375421', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99278271, end_position_on_the_genomic_accession=99325799, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927715, CurrentID=0, Status='live', Symbol='LINC01029', Aliases='', description='long intergenic non-protein coding RNA 1029', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77971294, end_position_on_the_genomic_accession=77993727, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  92283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92283, CurrentID=0, Status='live', Symbol='ZNF461', Aliases='GIOT-1, GIOT1, HZF28', description='zinc finger protein 461', other_designations='gonadotropin inducible transcription repressor 1|gonadotropin-inducible ovary transcription repressor 1|zinc finger protein 28', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36636683, end_position_on_the_genomic_accession=36667131, orientation='minus', exon_count=8, OMIM=608640, no_hdr0=''),
  92285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92285, CurrentID=0, Status='live', Symbol='ZNF585B', Aliases='SZFP41', description='zinc finger protein 585B', other_designations='zinc finger protein 41-like protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37181579, end_position_on_the_genomic_accession=37210549, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  92291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92291, CurrentID=0, Status='live', Symbol='CAPN13', Aliases='', description='calpain 13', other_designations='CANP 13|calcium-activated neutral proteinase 13', map_location='2p22-p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30722771, end_position_on_the_genomic_accession=30807484, orientation='minus', exon_count=29, OMIM=610228, no_hdr0=''),
  92292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92292, CurrentID=0, Status='live', Symbol='GLYATL1', Aliases='GATF-C, GNAT', description='glycine-N-acyltransferase like 1', other_designations='acyl-CoA:glycine N-acyltransferase-like protein 1|glutamine N-acyltransferase|glycine-N-acyltransferase-like 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58927629, end_position_on_the_genomic_accession=58957074, orientation='plus', exon_count=11, OMIM=614761, no_hdr0=''),
  92293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92293, CurrentID=0, Status='live', Symbol='TMEM132C', Aliases='PPP1R152', description='transmembrane protein 132C', other_designations='protein phosphatase 1, regulatory subunit 152', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128267403, end_position_on_the_genomic_accession=128707915, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  92304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92304, CurrentID=0, Status='live', Symbol='SCGB3A1', Aliases='HIN-1, HIN1, LU105, PnSP-2, UGRP2', description='secretoglobin, family 3A, member 1', other_designations='cytokine HIN-1|cytokine high in normal-1|high in normal 1|pneumo secretory protein 2|uteroglobin-related protein 2', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180590105, end_position_on_the_genomic_accession=180591487, orientation='minus', exon_count=3, OMIM=606500, no_hdr0=''),
  92305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92305, CurrentID=0, Status='live', Symbol='TMEM129', Aliases='D4S2561E', description='transmembrane protein 129', other_designations='transmembrane protein 129, E3 ubiquitin protein ligase', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1715952, end_position_on_the_genomic_accession=1721357, orientation='minus', exon_count=6, OMIM=615975, no_hdr0=''),
  105375427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375427, CurrentID=0, Status='live', Symbol='LOC105375427', Aliases='', description='uncharacterized LOC105375427', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100233681, end_position_on_the_genomic_accession=100234760, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375687, CurrentID=0, Status='live', Symbol='LOC105375687', Aliases='', description='uncharacterized LOC105375687', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103444698, end_position_on_the_genomic_accession=103445577, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  92312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92312, CurrentID=0, Status='live', Symbol='MEX3A', Aliases='MEX-3A, RKHD4', description='mex-3 RNA binding family member A', other_designations='RING finger and KH domain-containing protein 4|ring finger and KH domain containing 4|ring finger and KH domain containing protein', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156072013, end_position_on_the_genomic_accession=156081998, orientation='minus', exon_count=2, OMIM=611007, no_hdr0=''),
  105375428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375428, CurrentID=0, Status='live', Symbol='LOC105375428', Aliases='', description='uncharacterized LOC105375428', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100622122, end_position_on_the_genomic_accession=100633005, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375429, CurrentID=0, Status='live', Symbol='LOC105375429', Aliases='', description='uncharacterized LOC105375429', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100655815, end_position_on_the_genomic_accession=100666793, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375430, CurrentID=0, Status='live', Symbol='LOC105375430', Aliases='', description='uncharacterized LOC105375430', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100836894, end_position_on_the_genomic_accession=100852637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421957, CurrentID=0, Status='live', Symbol='LOC100421957', Aliases='', description='olfactory receptor, family 8, subfamily D, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92335, CurrentID=0, Status='live', Symbol='STRADA', Aliases='LYK5, NY-BR-96, PMSE, STRAD, Stlk', description='STE20-related kinase adaptor alpha', other_designations='STE20-like pseudokinase|STRAD alpha|protein kinase LYK5|serologically defined breast cancer antigen NY-BR-96', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63702832, end_position_on_the_genomic_accession=63741970, orientation='minus', exon_count=15, OMIM=608626, no_hdr0=''),
  92340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92340, CurrentID=0, Status='live', Symbol='PRR29', Aliases='C17orf72', description='proline rich 29', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63998344, end_position_on_the_genomic_accession=64004305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  92342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92342, CurrentID=0, Status='live', Symbol='METTL18', Aliases='AsTP2, C1orf156, HPM1', description='methyltransferase like 18', other_designations='arsenic-transactivated protein 2|methyltransferase-like protein 18', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169792529, end_position_on_the_genomic_accession=169794959, orientation='minus', exon_count=4, OMIM=615255, no_hdr0=''),
  92344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92344, CurrentID=0, Status='live', Symbol='GORAB', Aliases='GO, NTKLBP1, SCYL1BP1', description='golgin, RAB6-interacting', other_designations='N-terminal kinase-like-binding protein 1|NTKL-binding protein 1|SCY1-like 1 binding protein 1|SCY1-like 1-binding protein 1|SCYL1-BP1|SCYL1-binding protein 1|hNTKL-BP1', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170532122, end_position_on_the_genomic_accession=170553834, orientation='plus', exon_count=7, OMIM=607983, no_hdr0=''),
  92345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92345, CurrentID=0, Status='live', Symbol='NAF1', Aliases='', description='nuclear assembly factor 1 ribonucleoprotein', other_designations='nuclear assembly factor 1 homolog', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163109133, end_position_on_the_genomic_accession=163166921, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  92346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92346, CurrentID=0, Status='live', Symbol='C1orf105', Aliases='', description='chromosome 1 open reading frame 105', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172420515, end_position_on_the_genomic_accession=172468829, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101927213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927213, CurrentID=0, Status='live', Symbol='LOC101927213', Aliases='', description='uncharacterized LOC101927213', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56523059, end_position_on_the_genomic_accession=56595913, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  649024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649024, CurrentID=0, Status='live', Symbol='LOC649024', Aliases='', description='citrate synthase pseudogene', other_designations='', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=107327833, end_position_on_the_genomic_accession=107329180, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92359, CurrentID=0, Status='live', Symbol='CRB3', Aliases='', description='crumbs family member 3', other_designations='crumbs homolog 3|crumbs protein homolog 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6463777, end_position_on_the_genomic_accession=6467221, orientation='plus', exon_count=5, OMIM=609737, no_hdr0=''),
  92369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92369, CurrentID=0, Status='live', Symbol='SPSB4', Aliases='SSB-4, SSB4', description='splA/ryanodine receptor domain and SOCS box containing 4', other_designations='SPRY domain-containing SOCS box protein SSB-4', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141050795, end_position_on_the_genomic_accession=141148611, orientation='plus', exon_count=6, OMIM=611660, no_hdr0=''),
  92370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92370, CurrentID=0, Status='live', Symbol='PXYLP1', Aliases='ACPL2, HEL124, XYLP', description='2-phosphoxylose phosphatase 1', other_designations='acid phosphatase-like 2|acid phosphatase-like protein 2|epididymis luminal protein 124|xylosyl phosphatase', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141231825, end_position_on_the_genomic_accession=141294903, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101929161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929161, CurrentID=0, Status='live', Symbol='LOC101929161', Aliases='', description='uncharacterized LOC101929161', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25517390, end_position_on_the_genomic_accession=25569394, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100132562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132562, CurrentID=0, Status='live', Symbol='YPEL5P2', Aliases='', description='YPEL5 pseudogene 2', other_designations='yippee-like 5 pseudogene 2', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47841241, end_position_on_the_genomic_accession=47841687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  92399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92399, CurrentID=0, Status='live', Symbol='MRRF', Aliases='MRFF, MTRRF, RRF', description='mitochondrial ribosome recycling factor', other_designations='ribosome-releasing factor, mitochondrial', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122264603, end_position_on_the_genomic_accession=122325297, orientation='plus', exon_count=10, OMIM=604602, no_hdr0=''),
  92400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92400, CurrentID=0, Status='live', Symbol='RBM18', Aliases='', description='RNA binding motif protein 18', other_designations='RNA-binding motif protein 18', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122239555, end_position_on_the_genomic_accession=122264864, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375445, CurrentID=0, Status='live', Symbol='LOC105375445', Aliases='', description='uncharacterized LOC105375445', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107870924, end_position_on_the_genomic_accession=107874273, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92421, CurrentID=0, Status='live', Symbol='CHMP4C', Aliases='SNF7-3, Shax3, VPS32C', description='charged multivesicular body protein 4C', other_designations='SNF7 homolog associated with Alix 3|Snf7 homologue associated with Alix 3|chromatin modifying protein 4C|chromatin-modifying protein 4c|hSnf7-3|hVps32-3|vacuolar protein sorting-associated protein 32-3|vps32-3', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81732453, end_position_on_the_genomic_accession=81759515, orientation='plus', exon_count=5, OMIM=610899, no_hdr0=''),
  92427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92427, CurrentID=0, Status='live', Symbol='RPS2P4', Aliases='RPS2_20_1415', description='ribosomal protein S2 pseudogene 4', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104835944, end_position_on_the_genomic_accession=104836907, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128170, CurrentID=0, Status='live', Symbol='LOC100128170', Aliases='', description='uncharacterized LOC100128170', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=42016401, end_position_on_the_genomic_accession=42017109, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132570, CurrentID=0, Status='live', Symbol='LOC100132570', Aliases='', description='chromosome 17 open reading frame 58 pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46266927, end_position_on_the_genomic_accession=46269810, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132571, CurrentID=0, Status='live', Symbol='LOC100132571', Aliases='', description='short coiled-coil protein pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150780349, end_position_on_the_genomic_accession=150784473, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375453, CurrentID=0, Status='live', Symbol='LOC105375453', Aliases='', description='uncharacterized LOC105375453', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110591466, end_position_on_the_genomic_accession=110622056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  648927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648927, CurrentID=0, Status='live', Symbol='LOC648927', Aliases='', description='K(lysine) acetyltransferase 7 pseudogene', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=95973334, end_position_on_the_genomic_accession=95974314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371328, CurrentID=0, Status='live', Symbol='LOC105371328', Aliases='', description='uncharacterized LOC105371328', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70188316, end_position_on_the_genomic_accession=70189006, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725043, CurrentID=0, Status='live', Symbol='LOC102725043', Aliases='', description='uncharacterized LOC102725043', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133770224, end_position_on_the_genomic_accession=133778345, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  92482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92482, CurrentID=0, Status='live', Symbol='BBIP1', Aliases='BBIP10, BBS18, NCRNA00081, bA348N5.3', description='BBSome interacting protein 1', other_designations='BBSome-interacting protein of 10 kDa|UPF0604 protein', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110898730, end_position_on_the_genomic_accession=110919366, orientation='minus', exon_count=6, OMIM=613605, no_hdr0=''),
  92483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92483, CurrentID=0, Status='live', Symbol='LDHAL6B', Aliases='LDH6B, LDHAL6, LDHL', description='lactate dehydrogenase A-like 6B', other_designations='lactate dehydrogenase A-like 6', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59206816, end_position_on_the_genomic_accession=59208588, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375457, CurrentID=0, Status='live', Symbol='LOC105375457', Aliases='', description='uncharacterized LOC105375457', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112399042, end_position_on_the_genomic_accession=112445306, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375460, CurrentID=0, Status='live', Symbol='LOC105375460', Aliases='', description='uncharacterized LOC105375460', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=113055668, end_position_on_the_genomic_accession=113061751, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  92507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92507, CurrentID=0, Status='live', Symbol='RPS4XP2', Aliases='RPS4L2, RPS4P2, RPS4Y_2_1680, dJ189G13.2', description='ribosomal protein S4X pseudogene 2', other_designations='ribosomal protein S4 pseudogene 2|ribosomal protein S4, Y-linked 1 pseudogene|ribosomal protein S4-like 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4629423, end_position_on_the_genomic_accession=4630222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372544, CurrentID=0, Status='live', Symbol='LOC105372544', Aliases='', description='uncharacterized LOC105372544', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17122178, end_position_on_the_genomic_accession=17196938, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  343099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343099, CurrentID=0, Status='live', Symbol='CCDC18', Aliases='NY-SAR-41', description='coiled-coil domain containing 18', other_designations='sarcoma antigen NY-SAR-24|sarcoma antigen NY-SAR-41', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93180330, end_position_on_the_genomic_accession=93279037, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  648934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648934, CurrentID=0, Status='live', Symbol='LOC648934', Aliases='', description='citrate synthase, mitochondrial-like', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80555807, end_position_on_the_genomic_accession=80557189, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92521, CurrentID=0, Status='live', Symbol='SPECC1', Aliases='CYTSB, HCMOGT-1, HCMOGT1, NSP', description='sperm antigen with calponin homology and coiled-coil domains 1', other_designations='NSP5|cytokinesis and spindle organization B|nuclear structure protein 5|sperm antigen HCMOGT-1|sperm antigen with calponin-like and coiled coil domains 1|structure protein NSP5a3a|structure protein NSP5a3b|structure protein NSP5b3a|structure protein NSP5b3b', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20009336, end_position_on_the_genomic_accession=20314759, orientation='plus', exon_count=19, OMIM=608793, no_hdr0=''),
  100421964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421964, CurrentID=0, Status='live', Symbol='LOC100421964', Aliases='', description='olfactory receptor, family 9, subfamily G, member 9 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285181, CurrentID=0, Status='live', Symbol='LOC285181', Aliases='', description='uncharacterized LOC285181', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92552, CurrentID=0, Status='live', Symbol='ATXN3L', Aliases='MJDL', description='ataxin 3-like', other_designations='machado-Joseph disease protein 1-like', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13318647, end_position_on_the_genomic_accession=13320399, orientation='minus', exon_count=1, OMIM=300920, no_hdr0=''),
  92558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92558, CurrentID=0, Status='live', Symbol='CCDC64', Aliases='BICDR-1, H_267D11.1', description='coiled-coil domain containing 64', other_designations='BICD-related protein 1|WUGSC:H_267D11.1|coiled-coil domain-containing protein 64A', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119989799, end_position_on_the_genomic_accession=120094495, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105375469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375469, CurrentID=0, Status='live', Symbol='LOC105375469', Aliases='', description='uncharacterized LOC105375469', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117715739, end_position_on_the_genomic_accession=117730525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  92565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92565, CurrentID=0, Status='live', Symbol='FANK1', Aliases='HSD13', description='fibronectin type III and ankyrin repeat domains 1', other_designations='1700007B22Rik|fibronectin type 3 and ankyrin repeat domains 1', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125896224, end_position_on_the_genomic_accession=126009592, orientation='plus', exon_count=21, OMIM=611640, no_hdr0=''),
  105375471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375471, CurrentID=0, Status='live', Symbol='LOC105375471', Aliases='', description='uncharacterized LOC105375471', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118427040, end_position_on_the_genomic_accession=118444086, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  92579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92579, CurrentID=0, Status='live', Symbol='G6PC3', Aliases='SCN4, UGRP', description='glucose 6 phosphatase, catalytic, 3', other_designations='G-6-Pase 3|G6Pase 3|G6Pase-beta|glucose-6-phosphatase catalytic subunit 3|ubiquitous glucose-6-phosphatase catalytic subunit-related protein|ubiquitously expressed G6Pase catalytic subunit-related protein', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44070730, end_position_on_the_genomic_accession=44076344, orientation='plus', exon_count=7, OMIM=611045, no_hdr0=''),
  105375473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375473, CurrentID=0, Status='live', Symbol='LOC105375473', Aliases='', description='uncharacterized LOC105375473', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118503018, end_position_on_the_genomic_accession=118545686, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375474, CurrentID=0, Status='live', Symbol='LOC105375474', Aliases='', description='uncharacterized LOC105375474', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118859070, end_position_on_the_genomic_accession=118862193, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  92591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92591, CurrentID=0, Status='live', Symbol='ASB16', Aliases='', description='ankyrin repeat and SOCS box containing 16', other_designations='ankyrin repeat and SOCS box-containing 16|ankyrin repeat and SOCS box-containing protein 16', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44170706, end_position_on_the_genomic_accession=44179083, orientation='plus', exon_count=5, OMIM=615056, no_hdr0=''),
  92595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92595, CurrentID=0, Status='live', Symbol='ZNF764', Aliases='', description='zinc finger protein 764', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30553764, end_position_on_the_genomic_accession=30558321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100886964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100886964, CurrentID=0, Status='live', Symbol='PTCSC3', Aliases='', description='papillary thyroid carcinoma susceptibility candidate 3 (non-protein coding)', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36135710, end_position_on_the_genomic_accession=36176651, orientation='minus', exon_count=4, OMIM=614821, no_hdr0=''),
  92597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92597, CurrentID=0, Status='live', Symbol='MOB1B', Aliases='MATS2, MOB4A, MOBKL1A', description='MOB kinase activator 1B', other_designations='MOB1 Mps One Binder homolog B|MOB1, Mps One Binder kinase activator-like 1A|mob1 homolog 1A|mob1A|mps one binder kinase activator-like 1A', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70901974, end_position_on_the_genomic_accession=70988174, orientation='plus', exon_count=9, OMIM=609282, no_hdr0=''),
  151477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151477, CurrentID=0, Status='live', Symbol='LINC00471', Aliases='C2orf52', description='long intergenic non-protein coding RNA 471', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231508426, end_position_on_the_genomic_accession=231514339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375477, CurrentID=0, Status='live', Symbol='LOC105375477', Aliases='', description='uncharacterized LOC105375477', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=120216560, end_position_on_the_genomic_accession=120227118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92609, CurrentID=0, Status='live', Symbol='TIMM50', Aliases='TIM50, TIM50L', description='translocase of inner mitochondrial membrane 50 homolog (S. cerevisiae)', other_designations='Tim50-like protein|homolog of yeast Tim50', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39480412, end_position_on_the_genomic_accession=39493018, orientation='plus', exon_count=12, OMIM=607381, no_hdr0=''),
  92610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92610, CurrentID=0, Status='live', Symbol='TIFA', Aliases='T2BP, T6BPA, TIFA', description='TRAF-interacting protein with forkhead-associated domain', other_designations='TRAF-interacting protein with a forkhead-associated domain|TRAF2 binding protein|TRAF2-binding protein|TRAF6 binding protein|putative MAPK-activating protein PM14|putative NF-kappa-B-activating protein 20', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112275597, end_position_on_the_genomic_accession=112285903, orientation='minus', exon_count=3, OMIM=609028, no_hdr0=''),
  100129917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129917, CurrentID=0, Status='live', Symbol='LOC100129917', Aliases='', description='uncharacterized LOC100129917', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=780149, end_position_on_the_genomic_accession=781848, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375478, CurrentID=0, Status='live', Symbol='LOC105375478', Aliases='', description='uncharacterized LOC105375478', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121486317, end_position_on_the_genomic_accession=121505159, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132599, CurrentID=0, Status='live', Symbol='LOC100132599', Aliases='', description='zinc finger protein 114 pseudogene', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66042621, end_position_on_the_genomic_accession=66043329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421967, CurrentID=0, Status='live', Symbol='LOC100421967', Aliases='', description='olfactory receptor, family 9, subfamily I, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375481, CurrentID=0, Status='live', Symbol='LOC105375481', Aliases='', description='uncharacterized LOC105375481', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122527732, end_position_on_the_genomic_accession=122563437, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  474191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474191, CurrentID=0, Status='live', Symbol='BP31', Aliases='', description='Blood pressure QTL 31', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375482, CurrentID=0, Status='live', Symbol='LOC105375482', Aliases='', description='uncharacterized LOC105375482', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122915558, end_position_on_the_genomic_accession=122967210, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100533787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533787, CurrentID=0, Status='live', Symbol='LOC100533787', Aliases='', description='HERV-FRD provirus ancestral Env polyprotein pseudogene', other_designations='endogenous retrovirus group FRD, member 1 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32241523, end_position_on_the_genomic_accession=32243093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375485, CurrentID=0, Status='live', Symbol='LOC105375485', Aliases='', description='uncharacterized LOC105375485', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124048766, end_position_on_the_genomic_accession=124090520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421968, CurrentID=0, Status='live', Symbol='LOC100421968', Aliases='', description='olfactory receptor, family 5, subfamily B, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92659, CurrentID=0, Status='live', Symbol='MAFG-AS1', Aliases='', description='MAFG antisense RNA 1 (head to head)', other_designations='MAFG antisense RNA 1 (non-protein coding)', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81927829, end_position_on_the_genomic_accession=81930753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375486, CurrentID=0, Status='live', Symbol='LOC105375486', Aliases='', description='uncharacterized LOC105375486', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125151164, end_position_on_the_genomic_accession=125160038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  92667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92667, CurrentID=0, Status='live', Symbol='MGME1', Aliases='C20orf72, DDK1, MTDPS11, bA504H3.4', description='mitochondrial genome maintenance exonuclease 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17968853, end_position_on_the_genomic_accession=17991122, orientation='plus', exon_count=8, OMIM=615076, no_hdr0=''),
  92675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92675, CurrentID=0, Status='live', Symbol='DTD1', Aliases='C20orf88, DUE-B, DUEB, HARS2, pqn-68', description='D-tyrosyl-tRNA deacylase 1', other_designations='D-tyrosyl-tRNA deacylase 1 homolog|DNA-unwinding element-binding protein B|histidyl-tRNA synthase-related|histidyl-tRNA synthetase 2', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18587912, end_position_on_the_genomic_accession=18763917, orientation='plus', exon_count=8, OMIM=610996, no_hdr0=''),
  100132609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132609, CurrentID=0, Status='live', Symbol='LOC100132609', Aliases='', description='programmed cell death 2 pseudogene', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113850490, end_position_on_the_genomic_accession=113851635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92682, CurrentID=0, Status='live', Symbol='RPS15AP1', Aliases='RPS15A_24_1692, dJ872K7.6', description='ribosomal protein S15a pseudogene 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21166165, end_position_on_the_genomic_accession=21166614, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92689, CurrentID=0, Status='live', Symbol='FAM114A1', Aliases='Noxp20', description='family with sequence similarity 114 member A1', other_designations='family with sequence similarity 114, member A1|nervous system over-expressed protein 20|nervous system overexpressed protein 20', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38867733, end_position_on_the_genomic_accession=38945744, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  92691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92691, CurrentID=0, Status='live', Symbol='TMEM169', Aliases='', description='transmembrane protein 169', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216081866, end_position_on_the_genomic_accession=216102783, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375492, CurrentID=0, Status='live', Symbol='LOC105375492', Aliases='', description='uncharacterized LOC105375492', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127938214, end_position_on_the_genomic_accession=128008045, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100887067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887067, CurrentID=0, Status='live', Symbol='KRT8P48', Aliases='', description='keratin 8 pseudogene 48', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146705946, end_position_on_the_genomic_accession=146707674, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887068, CurrentID=0, Status='live', Symbol='LOC100887068', Aliases='', description='keratin 8 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202924609, end_position_on_the_genomic_accession=202925630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887070, CurrentID=0, Status='live', Symbol='LOC100887070', Aliases='', description='zinc finger protein 729 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92703, CurrentID=0, Status='live', Symbol='TMEM183A', Aliases='C1orf37', description='transmembrane protein 183A', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203007383, end_position_on_the_genomic_accession=203024848, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100887072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887072, CurrentID=0, Status='live', Symbol='LOC100887072', Aliases='', description='zinc finger protein 470 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58319277, end_position_on_the_genomic_accession=58320489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887073, CurrentID=0, Status='live', Symbol='DUX4L31', Aliases='', description='double homeobox 4 like 31, pseudogene', other_designations='double homeobox 4 like 6 pseudogene', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10171614, end_position_on_the_genomic_accession=10172653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887074, CurrentID=0, Status='live', Symbol='DUX4L45', Aliases='', description='double homeobox 4 like 45, pseudogene', other_designations='double homeobox 4 like 2 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34135739, end_position_on_the_genomic_accession=34136881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887075, CurrentID=0, Status='live', Symbol='ARHGAP42P1', Aliases='', description='Rho GTPase activating protein 42 pseudogene 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131328176, end_position_on_the_genomic_accession=131331094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887076, CurrentID=0, Status='live', Symbol='LOC100887076', Aliases='', description='chromosome 9 open reading frame 172 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14997356, end_position_on_the_genomic_accession=14999954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887077, CurrentID=0, Status='live', Symbol='LOC100887077', Aliases='', description='endogenous retrovirus group 3, member 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134468323, end_position_on_the_genomic_accession=134469377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887078, CurrentID=0, Status='live', Symbol='LOC100887078', Aliases='', description='endogenous retrovirus group 3, member 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246635462, end_position_on_the_genomic_accession=246636503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100887079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887079, CurrentID=0, Status='live', Symbol='LOC100887079', Aliases='', description='ubiquitin specific peptidase 17-like family member 22 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100887080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887080, CurrentID=0, Status='live', Symbol='LOC100887080', Aliases='', description='acyl-CoA synthetase medium-chain family member 4 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20711427, end_position_on_the_genomic_accession=20732990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887081, CurrentID=0, Status='live', Symbol='LOC100887081', Aliases='', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 6 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92714, CurrentID=0, Status='live', Symbol='ARRDC1', Aliases='', description='arrestin domain containing 1', other_designations='alpha-arrestin 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137605640, end_position_on_the_genomic_accession=137615360, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  92715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92715, CurrentID=0, Status='live', Symbol='DPH7', Aliases='C9orf112, RRT2, WDR85', description='diphthamide biosynthesis 7', other_designations='WD repeat domain 85|WD repeat-containing protein 85|diphthamide biosynthesis protein 7', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137554899, end_position_on_the_genomic_accession=137578935, orientation='minus', exon_count=14, OMIM=613210, no_hdr0=''),
  105375497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375497, CurrentID=0, Status='live', Symbol='LOC105375497', Aliases='', description='uncharacterized LOC105375497', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128891229, end_position_on_the_genomic_accession=128900919, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375498, CurrentID=0, Status='live', Symbol='LOC105375498', Aliases='', description='uncharacterized LOC105375498', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129413689, end_position_on_the_genomic_accession=129434311, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  92736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92736, CurrentID=0, Status='live', Symbol='OTOP2', Aliases='', description='otopetrin 2', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74922963, end_position_on_the_genomic_accession=74933913, orientation='plus', exon_count=8, OMIM=607827, no_hdr0=''),
  92737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92737, CurrentID=0, Status='live', Symbol='DNER', Aliases='UNQ26, bet', description='delta/notch like EGF repeat containing', other_designations='H_NH0150O02.1|WUGSC:H_NH0150O02.1|delta-notch-like EGF repeat-containing transmembrane|delta/notch-like EGF repeat containing', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229357629, end_position_on_the_genomic_accession=229714570, orientation='minus', exon_count=13, OMIM=607299, no_hdr0=''),
  105375499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375499, CurrentID=0, Status='live', Symbol='LOC105375499', Aliases='', description='uncharacterized LOC105375499', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129363159, end_position_on_the_genomic_accession=129367502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  92745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92745, CurrentID=0, Status='live', Symbol='SLC38A5', Aliases='JM24, SN2, SNAT5, pp7194', description='solute carrier family 38 member 5', other_designations='solute carrier family 38 (amino acid transporter), member 5|solute carrier family 38, member 5|system N transporter 2|transport system N, protein 2', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48458537, end_position_on_the_genomic_accession=48470256, orientation='minus', exon_count=20, OMIM=300649, no_hdr0=''),
  92747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92747, CurrentID=0, Status='live', Symbol='BPIFB1', Aliases='C20orf114, LPLUNC1', description='BPI fold containing family B member 1', other_designations='BPI fold containing family B, member 1|VEMSGP|long palate, lung and nasal epithelium carcinoma associated 1|long palate, lung and nasal epithelium carcinoma-associated protein 1|von Ebner minor salivary gland protein', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33283135, end_position_on_the_genomic_accession=33309878, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  92749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92749, CurrentID=0, Status='live', Symbol='DRC1', Aliases='C2orf39, CCDC164, CILD21', description='dynein regulatory complex subunit 1', other_designations='coiled-coil domain-containing protein 164', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26401916, end_position_on_the_genomic_accession=26456713, orientation='plus', exon_count=17, OMIM=615288, no_hdr0=''),
  105375501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375501, CurrentID=0, Status='live', Symbol='LOC105375501', Aliases='', description='uncharacterized LOC105375501', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129763058, end_position_on_the_genomic_accession=129769186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92755, CurrentID=0, Status='live', Symbol='TUBBP1', Aliases='', description='tubulin beta pseudogene 1', other_designations='tubulin, beta polypeptide pseudogene 1|tubulin, beta pseudogene 1', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30351857, end_position_on_the_genomic_accession=30353521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375502, CurrentID=0, Status='live', Symbol='LOC105375502', Aliases='', description='uncharacterized LOC105375502', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130170474, end_position_on_the_genomic_accession=130200917, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132626, CurrentID=0, Status='live', Symbol='LOC100132626', Aliases='', description='family with sequence similarity 103, member A1 pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221133868, end_position_on_the_genomic_accession=221134221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375508, CurrentID=0, Status='live', Symbol='LOC105375508', Aliases='', description='uncharacterized LOC105375508', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130734252, end_position_on_the_genomic_accession=130862158, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  92797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92797, CurrentID=0, Status='live', Symbol='HELB', Aliases='hDHB', description='helicase (DNA) B', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66302555, end_position_on_the_genomic_accession=66343643, orientation='plus', exon_count=14, OMIM=614539, no_hdr0=''),
  92799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92799, CurrentID=0, Status='live', Symbol='SHKBP1', Aliases='PP203, Sb1', description='SH3KBP1 binding protein 1', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40576851, end_position_on_the_genomic_accession=40591399, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105375510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375510, CurrentID=0, Status='live', Symbol='LOC105375510', Aliases='', description='uncharacterized LOC105375510', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132421842, end_position_on_the_genomic_accession=132448248, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  92806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92806, CurrentID=0, Status='live', Symbol='CENPBD1', Aliases='', description='CENPB DNA-binding domain containing 1', other_designations='CENPB DNA-binding domains containing 1', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89969775, end_position_on_the_genomic_accession=89972832, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  92815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92815, CurrentID=0, Status='live', Symbol='HIST3H2A', Aliases='', description='histone cluster 3, H2a', other_designations='histone 3, H2a', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228457364, end_position_on_the_genomic_accession=228457859, orientation='minus', exon_count=1, OMIM=615015, no_hdr0=''),
  92822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92822, CurrentID=0, Status='live', Symbol='ZNF276', Aliases='CENP-Z, CENPZ, ZADT, ZFP276, ZNF477', description='zinc finger protein 276', other_designations='centromere protein Z|zfp-276|zinc finger protein 276 homolog|zinc finger protein 477|zinc finger, AD-type', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89720195, end_position_on_the_genomic_accession=89740925, orientation='plus', exon_count=14, OMIM=608460, no_hdr0=''),
  105373419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373419, CurrentID=0, Status='live', Symbol='LOC105373419', Aliases='', description='uncharacterized LOC105373419', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9660378, end_position_on_the_genomic_accession=9661283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375515, CurrentID=0, Status='live', Symbol='LOC105375515', Aliases='', description='uncharacterized LOC105375515', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134065797, end_position_on_the_genomic_accession=134072990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  92840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92840, CurrentID=0, Status='live', Symbol='REEP6', Aliases='C19orf32, DP1L1, TB2L1', description='receptor accessory protein 6', other_designations='deleted in polyposis 1-like 1|polyposis locus protein 1-like 1|receptor expression enhancing protein 6', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1491166, end_position_on_the_genomic_accession=1497925, orientation='plus', exon_count=5, OMIM=609346, no_hdr0=''),
  92856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92856, CurrentID=0, Status='live', Symbol='IMP4', Aliases='BXDC4', description='IMP4 homolog, U3 small nucleolar ribonucleoprotein', other_designations='IMP4, U3 small nucleolar ribonucleoprotein|IMP4, U3 small nucleolar ribonucleoprotein, homolog|U3 snoRNP protein 4 homolog|U3 snoRNP protein IMP4|brix domain-containing protein 4', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130342897, end_position_on_the_genomic_accession=130346625, orientation='plus', exon_count=9, OMIM=612981, no_hdr0=''),
  105375521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375521, CurrentID=0, Status='live', Symbol='LOC105375521', Aliases='', description='uncharacterized LOC105375521', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135687024, end_position_on_the_genomic_accession=135708104, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132644, CurrentID=0, Status='live', Symbol='LOC100132644', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59406904, end_position_on_the_genomic_accession=59408181, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375525, CurrentID=0, Status='live', Symbol='LOC105375525', Aliases='', description='uncharacterized LOC105375525', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137196764, end_position_on_the_genomic_accession=137204824, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371342, CurrentID=0, Status='live', Symbol='LOC105371342', Aliases='', description='uncharacterized LOC105371342', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73815650, end_position_on_the_genomic_accession=73816011, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  92906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92906, CurrentID=0, Status='live', Symbol='HNRNPLL', Aliases='HNRPLL, SRRF', description='heterogeneous nuclear ribonucleoprotein L-like', other_designations='stromal RNA regulating factor|stromal RNA-regulating factor', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38561978, end_position_on_the_genomic_accession=38603036, orientation='minus', exon_count=16, OMIM=611208, no_hdr0=''),
  648999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648999, CurrentID=0, Status='live', Symbol='LOC648999', Aliases='', description='uncharacterized LOC648999', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  343165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343165, CurrentID=0, Status='live', Symbol='LOC343165', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene', other_designations='heat shock 70kDa protein 8 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247229852, end_position_on_the_genomic_accession=247231335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  92912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92912, CurrentID=0, Status='live', Symbol='UBE2Q2', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2', other_designations='E2 ubiquitin-conjugating enzyme Q2|ubiquitin carrier protein Q2|ubiquitin-conjugating enzyme E2Q family member 2|ubiquitin-protein ligase Q2', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75843286, end_position_on_the_genomic_accession=75901047, orientation='plus', exon_count=15, OMIM=612501, no_hdr0=''),
  92922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92922, CurrentID=0, Status='live', Symbol='CCDC102A', Aliases='', description='coiled-coil domain containing 102A', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57512178, end_position_on_the_genomic_accession=57537130, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105371343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371343, CurrentID=0, Status='live', Symbol='LOC105371343', Aliases='', description='uncharacterized LOC105371343', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74843105, end_position_on_the_genomic_accession=74845444, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132651, CurrentID=0, Status='live', Symbol='LOC100132651', Aliases='', description='UBX domain protein 2A pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68615510, end_position_on_the_genomic_accession=68616137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92935, CurrentID=0, Status='live', Symbol='MARS2', Aliases='COXPD25, MetRS, mtMetRS', description='methionyl-tRNA synthetase 2, mitochondrial', other_designations='methionine tRNA ligase 2, mitochondrial|methionine--tRNA ligase 2|methionine-tRNA synthetase 2, mitochondrial', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197705304, end_position_on_the_genomic_accession=197708390, orientation='plus', exon_count=1, OMIM=609728, no_hdr0=''),
  105375708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375708, CurrentID=0, Status='live', Symbol='LOC105375708', Aliases='', description='uncharacterized LOC105375708', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=112188842, end_position_on_the_genomic_accession=112194453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  474242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474242, CurrentID=0, Status='live', Symbol='RA21', Aliases='', description='Rheumatoid arthritis QTL 21', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  343171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343171, CurrentID=0, Status='live', Symbol='OR2W3', Aliases='OR2W3P, OR2W8P, OST718', description='olfactory receptor family 2 subfamily W member 3', other_designations='olfactory receptor 2W8|olfactory receptor OR1-49|olfactory receptor, family 2, subfamily W, member 3 pseudogene|olfactory receptor, family 2, subfamily W, member 8 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247895587, end_position_on_the_genomic_accession=247896531, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  92949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92949, CurrentID=0, Status='live', Symbol='ADAMTSL1', Aliases='ADAMTSL-1, ADAMTSR1, C9orf94, PUNCTIN', description='ADAMTS like 1', other_designations='ADAM-TS related protein 1|ADAMTS-like 1|punctin-1', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17906603, end_position_on_the_genomic_accession=18910950, orientation='plus', exon_count=36, OMIM=609198, no_hdr0=''),
  474244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474244, CurrentID=0, Status='live', Symbol='RA23', Aliases='', description='Rheumatoid arthritis QTL 23', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  343173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343173, CurrentID=0, Status='live', Symbol='OR2T3', Aliases='', description='olfactory receptor family 2 subfamily T member 3', other_designations='olfactory receptor, family 2, subfamily T, member 3|seven transmembrane helix receptor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248473351, end_position_on_the_genomic_accession=248474307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  92960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92960, CurrentID=0, Status='live', Symbol='PEX11G', Aliases='', description='peroxisomal biogenesis factor 11 gamma', other_designations='PEX11-gamma|Pex11pgamma|peroxin Pex11p gamma|peroxin-11C|peroxisomal biogenesis factor 11C|peroxisomal biogenesis factor 11G|protein PEX11 homolog gamma', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7476870, end_position_on_the_genomic_accession=7497485, orientation='minus', exon_count=13, OMIM=607583, no_hdr0=''),
  105375536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375536, CurrentID=0, Status='live', Symbol='LOC105375536', Aliases='', description='uncharacterized LOC105375536', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140925086, end_position_on_the_genomic_accession=141014544, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  474246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474246, CurrentID=0, Status='live', Symbol='RA25', Aliases='', description='Rheumatoid arthritis QTL 25', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474247, CurrentID=0, Status='live', Symbol='RA26', Aliases='', description='Rheumatoid arthritis QTL 26', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  92973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92973, CurrentID=0, Status='live', Symbol='LINC00950', Aliases='FP588', description='long intergenic non-protein coding RNA 950', other_designations='Putative uncharacterized protein FP588', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35860274, end_position_on_the_genomic_accession=35865518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  92979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92979, CurrentID=0, Status='live', Symbol='MARCH9', Aliases='MARCH-IX, RNF179', description='membrane associated ring-CH-type finger 9', other_designations='RING finger protein 179|membrane associated ring finger 9|membrane-associated RING finger protein 9|membrane-associated RING-CH protein IX|membrane-associated ring finger (C3HC4) 9', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57753622, end_position_on_the_genomic_accession=57760410, orientation='plus', exon_count=5, OMIM=613336, no_hdr0=''),
  474249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474249, CurrentID=0, Status='live', Symbol='RA28', Aliases='', description='Rheumatoid arthritis QTL 28', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  474250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474250, CurrentID=0, Status='live', Symbol='RA29', Aliases='', description='Rheumatoid arthritis QTL 29', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132661, CurrentID=0, Status='live', Symbol='LOC100132661', Aliases='', description='PRO0419', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=145418952, end_position_on_the_genomic_accession=145419767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  474251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474251, CurrentID=0, Status='live', Symbol='RA30', Aliases='', description='Rheumatoid arthritis QTL 30', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375542, CurrentID=0, Status='live', Symbol='LOC105375542', Aliases='', description='uncharacterized LOC105375542', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142935727, end_position_on_the_genomic_accession=142936950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644681, CurrentID=0, Status='live', Symbol='LOC644681', Aliases='', description='synovial sarcoma, X breakpoint 2 interacting protein pseudogene', other_designations='', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=103538574, end_position_on_the_genomic_accession=103540576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  92999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=92999, CurrentID=0, Status='live', Symbol='ZBTB47', Aliases='ZNF651', description='zinc finger and BTB domain containing 47', other_designations='zinc finger protein 651', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42653684, end_position_on_the_genomic_accession=42667580, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  474253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474253, CurrentID=0, Status='live', Symbol='BP11', Aliases='', description='Blood pressure QTL 11', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100307118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100307118, CurrentID=0, Status='live', Symbol='BCC1', Aliases='', description='Basal cell carcinoma, susceptibility to, 1', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605462, no_hdr0=''),
  347894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347894, CurrentID=0, Status='live', Symbol='MRPL2P1', Aliases='', description='mitochondrial ribosomal protein L2 pseudogene 1', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89752909, end_position_on_the_genomic_accession=89753867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  93010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93010, CurrentID=0, Status='live', Symbol='B3GNT7', Aliases='beta3GnT7', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 7', other_designations='BGnT-7|beta 1,3-N-acetylglucosaminyltransferase 7|beta-1,3-Gn-T7|beta-1,3-N-acetylglucosaminyltransferase 7|beta3Gn-T7', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231395624, end_position_on_the_genomic_accession=231408801, orientation='plus', exon_count=4, OMIM=615313, no_hdr0=''),
  343184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343184, CurrentID=0, Status='live', Symbol='RPS2P11', Aliases='RPS2_3_47', description='ribosomal protein S2 pseudogene 11', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50326112, end_position_on_the_genomic_accession=50327049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375711, CurrentID=0, Status='live', Symbol='LOC105375711', Aliases='', description='uncharacterized LOC105375711', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116454153, end_position_on_the_genomic_accession=116474403, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93034, CurrentID=0, Status='live', Symbol='NT5C1B', Aliases='AIRP, CN-IB, CN1B', description="5'-nucleotidase, cytosolic IB", other_designations='autoimmune infertility-related protein', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18562871, end_position_on_the_genomic_accession=18589580, orientation='minus', exon_count=10, OMIM=610526, no_hdr0=''),
  93035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93035, CurrentID=0, Status='live', Symbol='PKHD1L1', Aliases='PKHDL1', description='polycystic kidney and hepatic disease 1 (autosomal recessive)-like 1', other_designations='PKHD1-like protein 1|fibrocystin L|polycystic kidney and hepatic disease 1-like protein 1|polycystic kidney and hepatic disease-like 1', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109362477, end_position_on_the_genomic_accession=109532686, orientation='plus', exon_count=82, OMIM=607843, no_hdr0=''),
  100132669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132669, CurrentID=0, Status='live', Symbol='LOC100132669', Aliases='', description='carboxymethylenebutenolidase homolog (Pseudomonas) pseudogene', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204379095, end_position_on_the_genomic_accession=204379792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375552, CurrentID=0, Status='live', Symbol='LOC105375552', Aliases='', description='uncharacterized LOC105375552', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144976901, end_position_on_the_genomic_accession=144979545, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93058, CurrentID=0, Status='live', Symbol='COQ10A', Aliases='', description='coenzyme Q10A', other_designations='coenzyme Q10 homolog A', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56266858, end_position_on_the_genomic_accession=56270966, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  474262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474262, CurrentID=0, Status='live', Symbol='BP16', Aliases='', description='Blood pressure QTL 16', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132673, CurrentID=0, Status='live', Symbol='RPS2P28', Aliases='RPS2_13_694', description='ribosomal protein S2 pseudogene 28', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43363415, end_position_on_the_genomic_accession=43364219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81047, CurrentID=0, Status='live', Symbol='OR5BJ1P', Aliases='OST740', description='olfactory receptor family 5 subfamily BJ member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BJ, member 1 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48394482, end_position_on_the_genomic_accession=48395385, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374183, CurrentID=0, Status='live', Symbol='LOC105374183', Aliases='', description='uncharacterized LOC105374183', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159825279, end_position_on_the_genomic_accession=159862747, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375555, CurrentID=0, Status='live', Symbol='LOC105375555', Aliases='', description='uncharacterized LOC105375555', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148103201, end_position_on_the_genomic_accession=148155390, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  93081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93081, CurrentID=0, Status='live', Symbol='TEX30', Aliases='C13orf27', description='testis expressed 30', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102765888, end_position_on_the_genomic_accession=102773821, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  93082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93082, CurrentID=0, Status='live', Symbol='NEURL3', Aliases='LINCR', description='neuralized E3 ubiquitin protein ligase 3', other_designations='lung-inducible neuralized-related C3CH4 RING domain protein|neuralized homolog 3 pseudogene|neuralized-like protein 3', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96497643, end_position_on_the_genomic_accession=96508136, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  474266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474266, CurrentID=0, Status='live', Symbol='BP25', Aliases='', description='Blood pressure QTL 25', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375557, CurrentID=0, Status='live', Symbol='LOC105375557', Aliases='', description='uncharacterized LOC105375557', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148543561, end_position_on_the_genomic_accession=148625612, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375558, CurrentID=0, Status='live', Symbol='LOC105375558', Aliases='', description='uncharacterized LOC105375558', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148802093, end_position_on_the_genomic_accession=148824109, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93099, CurrentID=0, Status='live', Symbol='DMKN', Aliases='UNQ729, ZD52F10', description='dermokine', other_designations='epidermis-specific secreted protein SK30/SK89', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35497217, end_position_on_the_genomic_accession=35513668, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  93100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93100, CurrentID=0, Status='live', Symbol='NAPRT', Aliases='NAPRT1, PP3856', description='nicotinate phosphoribosyltransferase', other_designations='FHA-HIT-interacting protein|NAPRTase|nicotinate phosphoribosyltransferase domain containing 1|nicotinate phosphoribosyltransferase domain-containing protein 1|nicotinic acid phosphoribosyltransferase', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143574664, end_position_on_the_genomic_accession=143579626, orientation='minus', exon_count=13, OMIM=611552, no_hdr0=''),
  93107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93107, CurrentID=0, Status='live', Symbol='KCNG4', Aliases='KV6.3, KV6.4', description='potassium channel, voltage gated modifier subfamily G, member 4', other_designations='potassium voltage-gated channel, subfamily G, member 4|voltage-gated potassium channel Kv6.3|voltage-gated potassium channel subunit Kv6.4', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84221135, end_position_on_the_genomic_accession=84239750, orientation='minus', exon_count=3, OMIM=607603, no_hdr0=''),
  93109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93109, CurrentID=0, Status='live', Symbol='TMEM44', Aliases='', description='transmembrane protein 44', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194587673, end_position_on_the_genomic_accession=194633780, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105375561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375561, CurrentID=0, Status='live', Symbol='LOC105375561', Aliases='', description='uncharacterized LOC105375561', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150076554, end_position_on_the_genomic_accession=150092263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375563, CurrentID=0, Status='live', Symbol='LOC105375563', Aliases='', description='uncharacterized LOC105375563', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150362699, end_position_on_the_genomic_accession=150368844, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93129, CurrentID=0, Status='live', Symbol='ORAI3', Aliases='TMEM142C', description='ORAI calcium release-activated calcium modulator 3', other_designations='transmembrane protein 142C', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30949084, end_position_on_the_genomic_accession=30954938, orientation='plus', exon_count=2, OMIM=610930, no_hdr0=''),
  93134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93134, CurrentID=0, Status='live', Symbol='ZNF561', Aliases='', description='zinc finger protein 561', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9607326, end_position_on_the_genomic_accession=9621240, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105371350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371350, CurrentID=0, Status='live', Symbol='LOC105371350', Aliases='', description='uncharacterized LOC105371350', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77590782, end_position_on_the_genomic_accession=77626235, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132686, CurrentID=0, Status='live', Symbol='LOC100132686', Aliases='', description='uncharacterized LOC100132686', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=112131195, end_position_on_the_genomic_accession=112131624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93144, CurrentID=0, Status='live', Symbol='RPS10P5', Aliases='RPS10L, RPS10_13_1677, bA371L19.2', description='ribosomal protein S10 pseudogene 5', other_designations='Putative 40S ribosomal protein S10-like protein', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=839401, end_position_on_the_genomic_accession=840020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93145, CurrentID=0, Status='live', Symbol='OLFM2', Aliases='NOE2, NOELIN2, NOELIN2_V1, OlfC', description='olfactomedin 2', other_designations='neuronal olfactomedin related ER localized protein 2|olfactomedin-2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9853718, end_position_on_the_genomic_accession=9936552, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375567, CurrentID=0, Status='live', Symbol='LOC105375567', Aliases='', description='uncharacterized LOC105375567', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150823033, end_position_on_the_genomic_accession=150877309, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372746, CurrentID=0, Status='live', Symbol='LOC105372746', Aliases='', description='uncharacterized LOC105372746', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19885221, end_position_on_the_genomic_accession=19899790, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  474277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474277, CurrentID=0, Status='live', Symbol='BW24', Aliases='', description='Body weight QTL 24', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  93164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93164, CurrentID=0, Status='live', Symbol='HTR7P1', Aliases='HTR7P', description='5-hydroxytryptamine (serotonin) receptor 7 pseudogene 1', other_designations='serotonin-7 receptor pseudogene', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13000442, end_position_on_the_genomic_accession=13004830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93166, CurrentID=0, Status='live', Symbol='PRDM6', Aliases='', description='PR domain containing 6', other_designations='PR domain zinc finger protein 6|PR domain-containing protein 6|PR-domain zinc finger protein 6', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123035985, end_position_on_the_genomic_accession=123194266, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100101111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101111, CurrentID=0, Status='live', Symbol='GXYLT1P1', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 1', other_designations='glycosyltransferase 8 domain containing 3 pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18723720, end_position_on_the_genomic_accession=18724702, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100101112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101112, CurrentID=0, Status='live', Symbol='CDC14BL', Aliases='', description='CDC14 cell division cycle 14 C-like', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48846402, end_position_on_the_genomic_accession=48848081, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100101115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101115, CurrentID=0, Status='live', Symbol='TTTY21B', Aliases='', description='testis-specific transcript, Y-linked 21B (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6443434, end_position_on_the_genomic_accession=6447077, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100101116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101116, CurrentID=0, Status='live', Symbol='TTTY1B', Aliases='TTTY1', description='testis-specific transcript, Y-linked 1B (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6390401, end_position_on_the_genomic_accession=6411564, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100101117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101117, CurrentID=0, Status='live', Symbol='TTTY2B', Aliases='', description='testis-specific transcript, Y-linked 2B (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6406244, end_position_on_the_genomic_accession=6428444, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100101118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101118, CurrentID=0, Status='live', Symbol='TTTY8B', Aliases='NCRNA00217, TTTY8', description='testis-specific transcript, Y-linked 8B (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6471031, end_position_on_the_genomic_accession=6473630, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93183, CurrentID=0, Status='live', Symbol='PIGM', Aliases='GPI-MT-I', description='phosphatidylinositol glycan anchor biosynthesis class M', other_designations='DPM:GlcN-(acyl-)PI mannosyltransferase|GPI mannosyltransferase I|PIG-M mannosyltransferase|dol-P-Man dependent GPI mannosyltransferase|phosphatidylinositol glycan anchor biosynthesis, class M|phosphatidylinositol-glycan biosynthesis class M protein', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160027672, end_position_on_the_genomic_accession=160031993, orientation='minus', exon_count=1, OMIM=610273, no_hdr0=''),
  100101120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101120, CurrentID=0, Status='live', Symbol='TTTY7B', Aliases='NCRNA00215', description='testis-specific transcript, Y-linked 7B (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9706824, end_position_on_the_genomic_accession=9715262, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  93185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93185, CurrentID=0, Status='live', Symbol='IGSF8', Aliases='CD316, CD81P3, EWI-2, EWI2, KCT-4, LIR-D1, PGRL', description='immunoglobulin superfamily member 8', other_designations='CD81 partner 3|glu-Trp-Ile EWI motif-containing protein 2|immunoglobulin superfamily, member 8|keratinocytes-associated transmembrane protein 4|prostaglandin regulatory-like protein', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160091339, end_position_on_the_genomic_accession=160098860, orientation='minus', exon_count=10, OMIM=606644, no_hdr0=''),
  93190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93190, CurrentID=0, Status='live', Symbol='C1orf158', Aliases='', description='chromosome 1 open reading frame 158', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12746186, end_position_on_the_genomic_accession=12761169, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100101127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101127, CurrentID=0, Status='live', Symbol='LOC100101127', Aliases='', description='thioredoxin-related transmembrane protein 1 pseudogene', other_designations='', map_location='8q', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69794168, end_position_on_the_genomic_accession=69800242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060075, CurrentID=0, Status='live', Symbol='LOC101060075', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9233013, end_position_on_the_genomic_accession=9233451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375575, CurrentID=0, Status='live', Symbol='LOC105375575', Aliases='', description='uncharacterized LOC105375575', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153129527, end_position_on_the_genomic_accession=153144071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93202, CurrentID=0, Status='live', Symbol='TXNP1', Aliases='', description='thioredoxin pseudogene 1', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119682983, end_position_on_the_genomic_accession=119683864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93210, CurrentID=0, Status='live', Symbol='PGAP3', Aliases='AGLA546, CAB2, PERLD1, PP1498, hCOS16', description='post-GPI attachment to proteins 3', other_designations='COS16 homolog|gene coamplified with ERBB2 protein|per1-like domain containing 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39671122, end_position_on_the_genomic_accession=39688070, orientation='minus', exon_count=9, OMIM=611801, no_hdr0=''),
  100101148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101148, CurrentID=0, Status='live', Symbol='LOC100101148', Aliases='', description='FK506 binding protein 6, 36kDa pseudogene', other_designations='', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72969682, end_position_on_the_genomic_accession=72973143, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132698, CurrentID=0, Status='live', Symbol='LOC100132698', Aliases='', description='single stranded DNA binding protein 4 pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18645045, end_position_on_the_genomic_accession=18646221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375579, CurrentID=0, Status='live', Symbol='LOC105375579', Aliases='', description='uncharacterized LOC105375579', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153507091, end_position_on_the_genomic_accession=153508337, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  93233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93233, CurrentID=0, Status='live', Symbol='CCDC114', Aliases='CILD20', description='coiled-coil domain containing 114', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48296452, end_position_on_the_genomic_accession=48322795, orientation='minus', exon_count=20, OMIM=615038, no_hdr0=''),
  81075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81075, CurrentID=0, Status='live', Symbol='OR5BH1P', Aliases='OR5BH2P', description='olfactory receptor family 5 subfamily BH member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BH, member 1 pseudogene|olfactory receptor, family 5, subfamily BH, member 2 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131398049, end_position_on_the_genomic_accession=131398957, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81076, CurrentID=0, Status='live', Symbol='OR5AW1P', Aliases='', description='olfactory receptor family 5 subfamily AW member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AW, member 1 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131341529, end_position_on_the_genomic_accession=131342452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928813, CurrentID=0, Status='live', Symbol='LOC101928813', Aliases='', description='uncharacterized LOC101928813', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76608151, end_position_on_the_genomic_accession=76630363, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  81077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81077, CurrentID=0, Status='live', Symbol='OR4W1P', Aliases='', description='olfactory receptor family 4 subfamily W member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily W, member 1 pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126502871, end_position_on_the_genomic_accession=126503218, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81078, CurrentID=0, Status='live', Symbol='OR4K12P', Aliases='OR21-2', description='olfactory receptor family 4 subfamily K member 12 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 12 pseudogene', map_location='21q11.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13581072, end_position_on_the_genomic_accession=13582004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81079, CurrentID=0, Status='live', Symbol='OR4K11P', Aliases='OR21-1', description='olfactory receptor family 4 subfamily K member 11 pseudogene', other_designations='olfactory receptor, family 4, subfamily K, member 11 pseudogene', map_location='21q11.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13544231, end_position_on_the_genomic_accession=13545185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375587, CurrentID=0, Status='live', Symbol='LOC105375587', Aliases='', description='uncharacterized LOC105375587', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155238658, end_position_on_the_genomic_accession=155254119, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  81081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81081, CurrentID=0, Status='live', Symbol='OR7A18P', Aliases='', description='olfactory receptor family 7 subfamily A member 18 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 18 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14886846, end_position_on_the_genomic_accession=14887272, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93273, CurrentID=0, Status='live', Symbol='LEMD1', Aliases='CT50, LEMP-1', description='LEM domain containing 1', other_designations='LEM domain protein 1|cancer/testis antigen 50', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205381378, end_position_on_the_genomic_accession=205449950, orientation='minus', exon_count=11, OMIM=610480, no_hdr0=''),
  105375589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375589, CurrentID=0, Status='live', Symbol='LOC105375589', Aliases='', description='uncharacterized LOC105375589', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155267147, end_position_on_the_genomic_accession=155268074, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375592, CurrentID=0, Status='live', Symbol='LOC105375592', Aliases='', description='uncharacterized LOC105375592', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155311336, end_position_on_the_genomic_accession=155337461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  474303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=474303, CurrentID=0, Status='live', Symbol='BW11', Aliases='', description='Body weight QTL 11', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100101246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101246, CurrentID=0, Status='live', Symbol='LOC100101246', Aliases='', description='interferon induced transmembrane protein 3 pseudogene', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111581294, end_position_on_the_genomic_accession=111581945, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100101247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101247, CurrentID=0, Status='live', Symbol='LOC100101247', Aliases='', description='interferon induced transmembrane protein pseudogene', other_designations='', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34576101, end_position_on_the_genomic_accession=34576757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93323, CurrentID=0, Status='live', Symbol='HAUS8', Aliases='DGT4, HICE1, NY-SAR-48', description='HAUS augmin like complex subunit 8', other_designations='HAUS augmin-like complex, subunit 8|HEC1/NDC80 interacting, centrosome associated 1|HEC1/NDC80-interacting centrosome-associated protein 1|Hec1-interacting and centrosome-associated 1|sarcoma antigen NY-SAR-48', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17049761, end_position_on_the_genomic_accession=17075608, orientation='minus', exon_count=11, OMIM=613434, no_hdr0=''),
  81090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81090, CurrentID=0, Status='live', Symbol='OR1AB1P', Aliases='', description='olfactory receptor family 1 subfamily AB member 1 pseudogene', other_designations='olfactory receptor, family 1, subfamily AB, member 1 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16051978, end_position_on_the_genomic_accession=16052716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100101266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101266, CurrentID=0, Status='live', Symbol='HAVCR1P1', Aliases='', description='hepatitis A virus cellular receptor 1 pseudogene 1', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=24162193, end_position_on_the_genomic_accession=24163447, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100101267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101267, CurrentID=0, Status='live', Symbol='POM121C', Aliases='POM121-2', description='POM121 transmembrane nucleoporin C', other_designations='POM121 membrane glycoprotein C|nuclear pore membrane protein 121-2|pore membrane protein of 121 kDa C', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75416781, end_position_on_the_genomic_accession=75486288, orientation='minus', exon_count=16, OMIM=615754, no_hdr0=''),
  100101268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101268, CurrentID=0, Status='live', Symbol='SPDYE12P', Aliases='', description='speedy/RINGO cell cycle regulator family member E12, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74904301, end_position_on_the_genomic_accession=74912323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375599, CurrentID=0, Status='live', Symbol='LOC105375599', Aliases='', description='uncharacterized LOC105375599', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156432367, end_position_on_the_genomic_accession=156433006, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  93343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93343, CurrentID=0, Status='live', Symbol='MVB12A', Aliases='CFBP, FAM125A', description='multivesicular body subunit 12A', other_designations='CIN85/CD2AP family binding protein|ESCRT-I complex subunit MVB12A|family with sequence similarity 125, member A', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17405686, end_position_on_the_genomic_accession=17425339, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  93349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93349, CurrentID=0, Status='live', Symbol='SP140L', Aliases='', description='SP140 nuclear body protein like', other_designations='SP140 nuclear body protein-like|SP140L-1 protein|SP140L-2 protein|SP140L-3 protein', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230327166, end_position_on_the_genomic_accession=230403730, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105375602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375602, CurrentID=0, Status='live', Symbol='LOC105375602', Aliases='', description='uncharacterized LOC105375602', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156146078, end_position_on_the_genomic_accession=156158167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132723, CurrentID=0, Status='live', Symbol='LOC100132723', Aliases='', description='single stranded DNA binding protein 4 pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15572087, end_position_on_the_genomic_accession=15573263, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375604, CurrentID=0, Status='live', Symbol='LOC105375604', Aliases='', description='uncharacterized threonine-rich GPI-anchored glycoprotein PJ4664.02-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156937417, end_position_on_the_genomic_accession=156939681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100887740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887740, CurrentID=0, Status='live', Symbol='RNA5SP411', Aliases='RN5S411', description='RNA, 5S ribosomal pseudogene 411', other_designations='RNA, 5S ribosomal 411', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35746523, end_position_on_the_genomic_accession=35746636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887743, CurrentID=0, Status='live', Symbol='DELXQ21', Aliases='CXDELq21', description='Choroideremia, deafness, and mental retardation', other_designations='', map_location='Xq21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=303110, no_hdr0=''),
  100887744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887744, CurrentID=0, Status='live', Symbol='RNU6-48P', Aliases='RNU6-48', description='RNA, U6 small nuclear 48, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27325219, end_position_on_the_genomic_accession=27325325, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93377, CurrentID=0, Status='live', Symbol='OPALIN', Aliases='HTMP10, TMEM10, TMP10', description='oligodendrocytic myelin paranodal and inner loop protein', other_designations='transmembrane protein 10|transmembrane protein TMP10', map_location='10q23-q24', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96343216, end_position_on_the_genomic_accession=96359365, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100887746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887746, CurrentID=0, Status='live', Symbol='UNC93B7', Aliases='', description='unc-93 homolog B7 pseudogene (C. elegans)', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9493735, end_position_on_the_genomic_accession=9498505, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100887747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887747, CurrentID=0, Status='live', Symbol='EEF1A1P18', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 18', other_designations='', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65025385, end_position_on_the_genomic_accession=65026925, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93380, CurrentID=0, Status='live', Symbol='MMGT1', Aliases='EMC5, TMEM32', description='membrane magnesium transporter 1', other_designations='ER membrane protein complex subunit 5|transmembrane protein 32', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135962072, end_position_on_the_genomic_accession=135973975, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375606, CurrentID=0, Status='live', Symbol='MNX1-AS2', Aliases='', description='MNX1 antisense RNA 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157005881, end_position_on_the_genomic_accession=157007120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100887750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887750, CurrentID=0, Status='live', Symbol='MRPS31P5', Aliases='MRPS31P3', description='mitochondrial ribosomal protein S31 pseudogene 5', other_designations='mitochondrial ribosomal protein S31 pseudogene 3', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52167709, end_position_on_the_genomic_accession=52194467, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100887751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887751, CurrentID=0, Status='live', Symbol='MPVQTL6', Aliases='', description='Mean platelet volume QTL6', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614646, no_hdr0=''),
  100887752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887752, CurrentID=0, Status='live', Symbol='MPVQTL5', Aliases='', description='Mean platelet volume QTL5', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614645, no_hdr0=''),
  100887753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887753, CurrentID=0, Status='live', Symbol='MPVQTL4', Aliases='', description='Mean platelet volume QTL4', other_designations='', map_location='6p21.3-p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614644, no_hdr0=''),
  100887754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887754, CurrentID=0, Status='live', Symbol='LINC00359', Aliases='', description='long intergenic non-protein coding RNA 359', other_designations='', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96941281, end_position_on_the_genomic_accession=96984172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100887755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887755, CurrentID=0, Status='live', Symbol='TRERNA1', Aliases='LINC00651, treRNA', description='translation regulatory long non-coding RNA 1', other_designations='long intergenic non-protein coding RNA 651|ncRNA-a7|noncoding RNA-activating 7|translational regulatory long non-coding RNA', map_location='20q13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50040707, end_position_on_the_genomic_accession=50041629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100526741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526741, CurrentID=0, Status='live', Symbol='DEL8Q13', Aliases='C8DELq13', description='Mesomelia-synostoses syndrome', other_designations='', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600383, no_hdr0=''),
  101927220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927220, CurrentID=0, Status='live', Symbol='DEPDC1-AS1', Aliases='', description='DEPDC1 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68496676, end_position_on_the_genomic_accession=68538627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93408, CurrentID=0, Status='live', Symbol='MYL10', Aliases='MYLC2PL, PLRLC', description='myosin light chain 10', other_designations='myosin light chain 2, lymphocyte-specific|myosin light chain 2, precursor lymphocyte-specific|myosin, light chain 10, regulatory|precursor lymphocyte-specific regulatory light chain', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101613325, end_position_on_the_genomic_accession=101629296, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375611, CurrentID=0, Status='live', Symbol='LOC105375611', Aliases='', description='uncharacterized LOC105375611', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158537495, end_position_on_the_genomic_accession=158551601, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375613, CurrentID=0, Status='live', Symbol='LOC105375613', Aliases='', description='uncharacterized LOC105375613', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157987158, end_position_on_the_genomic_accession=157989196, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93426, CurrentID=0, Status='live', Symbol='SYCE1', Aliases='C10orf94, CT76', description='synaptonemal complex central element protein 1', other_designations='cancer/testis antigen 76', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133553900, end_position_on_the_genomic_accession=133568291, orientation='minus', exon_count=15, OMIM=611486, no_hdr0=''),
  93429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93429, CurrentID=0, Status='live', Symbol='DKFZp434J0226', Aliases='', description='uncharacterized LOC93429', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46210242, end_position_on_the_genomic_accession=46214838, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  93432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93432, CurrentID=0, Status='live', Symbol='MGAM2', Aliases='', description='maltase-glucoamylase 2 (putative)', other_designations='Putative maltase-glucoamylase-like protein FLJ16351|maltase-glucoamylase (alpha-glucosidase)', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142111722, end_position_on_the_genomic_accession=142223533, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  105375615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375615, CurrentID=0, Status='live', Symbol='LOC105375615', Aliases='', description='uncharacterized LOC105375615', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157609875, end_position_on_the_genomic_accession=157613149, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  93436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93436, CurrentID=0, Status='live', Symbol='ARMC6', Aliases='R30923_1', description='armadillo repeat containing 6', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19033578, end_position_on_the_genomic_accession=19058178, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100421994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421994, CurrentID=0, Status='live', Symbol='LOC100421994', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81109, CurrentID=0, Status='live', Symbol='OR11K1P', Aliases='', description='olfactory receptor family 11 subfamily K member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily K, member 1 pseudogene|seven transmembrane helix receptor', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22030902, end_position_on_the_genomic_accession=22031830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132736, CurrentID=0, Status='live', Symbol='LOC100132736', Aliases='', description='family with sequence similarity 64, member A pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89814633, end_position_on_the_genomic_accession=89816114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  93444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93444, CurrentID=0, Status='live', Symbol='LOC93444', Aliases='', description='uncharacterized LOC93444', other_designations='', map_location='15q22.32', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375617, CurrentID=0, Status='live', Symbol='LOC105375617', Aliases='', description='uncharacterized LOC105375617', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158608249, end_position_on_the_genomic_accession=158610454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100887822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887822, CurrentID=0, Status='live', Symbol='KTCN7', Aliases='', description='Keratoconus 7', other_designations='', map_location='13q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614629, no_hdr0=''),
  100887824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100887824, CurrentID=0, Status='live', Symbol='DDX11L11', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 11', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 11|DEAD/H box polypeptide 11 like 11', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  93463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93463, CurrentID=0, Status='live', Symbol='LOC93463', Aliases='', description='uncharacterized LOC93463', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237048599, end_position_on_the_genomic_accession=237056167, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421995, CurrentID=0, Status='live', Symbol='LOC100421995', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100101405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101405, CurrentID=0, Status='live', Symbol='TMA16P1', Aliases='', description='translation machinery associated 16 homolog pseudogene 1', other_designations='FLJ11184 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58795203, end_position_on_the_genomic_accession=58796955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100101407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101407, CurrentID=0, Status='live', Symbol='TMA16P2', Aliases='', description='translation machinery associated 16 homolog pseudogene 2', other_designations='FLJ11184 pseudogene', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45845338, end_position_on_the_genomic_accession=45847719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93474, CurrentID=0, Status='live', Symbol='ZNF670', Aliases='', description='zinc finger protein 670', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247034638, end_position_on_the_genomic_accession=247078813, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375622, CurrentID=0, Status='live', Symbol='LOC105375622', Aliases='', description='uncharacterized LOC105375622', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85851630, end_position_on_the_genomic_accession=85861603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375726, CurrentID=0, Status='live', Symbol='LOC105375726', Aliases='', description='uncharacterized LOC105375726', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119214623, end_position_on_the_genomic_accession=119247004, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927755, CurrentID=0, Status='live', Symbol='LOC101927755', Aliases='', description='uncharacterized LOC101927755', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59965096, end_position_on_the_genomic_accession=59996972, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100101423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101423, CurrentID=0, Status='live', Symbol='MRT8', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 8', other_designations='', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611094, no_hdr0=''),
  100101424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101424, CurrentID=0, Status='live', Symbol='MRT9', Aliases='MRT26', description='mental retardation, non-syndromic, autosomal recessive, 9', other_designations='', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611095, no_hdr0=''),
  100101425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101425, CurrentID=0, Status='live', Symbol='MRT10', Aliases='MRT20', description='mental retardation, non-syndromic, autosomal recessive, 10', other_designations='', map_location='16p12-q12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611096, no_hdr0=''),
  100101426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101426, CurrentID=0, Status='live', Symbol='MRT11', Aliases='', description='mental retardation, non-syndromic, autosomal recessive, 11', other_designations='', map_location='19q13.2-q13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611097, no_hdr0=''),
  93492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93492, CurrentID=0, Status='live', Symbol='TPTE2', Aliases='TPIP', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2', other_designations='PTEN-like inositol lipid phosphatase|TPIP lipid phosphatase|TPTE and PTEN homologous inositol lipid phosphatase|lipid phosphatase TPIP|phosphatidylinositol-3,4,5-trisphosphate 3-phosphatase TPTE2', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19422877, end_position_on_the_genomic_accession=19561574, orientation='minus', exon_count=26, OMIM=606791, no_hdr0=''),
  105371362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371362, CurrentID=0, Status='live', Symbol='LOC105371362', Aliases='', description='uncharacterized LOC105371362', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81629894, end_position_on_the_genomic_accession=81648564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100101440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101440, CurrentID=0, Status='live', Symbol='PMS2P7', Aliases='PMS2L15, PMS2LP2, PMSR7', description='PMS1 homolog 2, mismatch repair system component pseudogene 7', other_designations='postmeiotic segregation increased 2 pseudogene 7|postmeiotic segregation increased 2-like 15|postmeiotic segregation increased 2-like pseudogene 2', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73006080, end_position_on_the_genomic_accession=73016375, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375727, CurrentID=0, Status='live', Symbol='LOC105375727', Aliases='', description='uncharacterized LOC105375727', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119671950, end_position_on_the_genomic_accession=119675343, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93517, CurrentID=0, Status='live', Symbol='SDR42E1', Aliases='HSPC105', description='short chain dehydrogenase/reductase family 42E, member 1', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81997616, end_position_on_the_genomic_accession=82011488, orientation='minus', exon_count=6, OMIM=616164, no_hdr0=''),
  100421997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421997, CurrentID=0, Status='live', Symbol='LOC100421997', Aliases='', description='olfactory receptor, family 4, subfamily A, member 5 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100101467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101467, CurrentID=0, Status='live', Symbol='ZSCAN30', Aliases='ZNF-WYM, ZNF397OS, ZNF917', description='zinc finger and SCAN domain containing 30', other_designations='ZNF397 opposite strand|zinc finger protein 397 opposite strand|zinc finger protein 397OS', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35251058, end_position_on_the_genomic_accession=35290259, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375728, CurrentID=0, Status='live', Symbol='LOC105375728', Aliases='', description='uncharacterized LOC105375728', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119832951, end_position_on_the_genomic_accession=119855597, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100101478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101478, CurrentID=0, Status='live', Symbol='LOC100101478', Aliases='', description='H2B histone family, member M pseudogene', other_designations='', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=104060504, end_position_on_the_genomic_accession=104062938, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132753, CurrentID=0, Status='live', Symbol='RFKP2', Aliases='', description='riboflavin kinase pseudogene 2', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112028903, end_position_on_the_genomic_accession=112031676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81127, CurrentID=0, Status='live', Symbol='OR4K15', Aliases='OR14-20Q, OR4K15', description='olfactory receptor family 4 subfamily K member 15', other_designations='olfactory receptor OR14-20|olfactory receptor, family 4, subfamily K, member 15', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19975519, end_position_on_the_genomic_accession=19976565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93550, CurrentID=0, Status='live', Symbol='ZFAND4', Aliases='ANUBL1', description='zinc finger, AN1-type domain 4', other_designations='AN1, ubiquitin-like, homolog|AN1-type zinc finger and ubiquitin domain-containing protein 1|AN1-type zinc finger and ubiquitin domain-containing protein-like 1', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45614567, end_position_on_the_genomic_accession=45672962, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100101490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101490, CurrentID=0, Status='live', Symbol='MCTS2P', Aliases='MCTS2, PSIMCT-1', description='malignant T-cell amplified sequence 2, pseudogene', other_designations='malignant T cell amplified sequence 1 pseudogene|malignant T cell amplified sequence 2, pseudogene', map_location='20q11.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31547382, end_position_on_the_genomic_accession=31548220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93556, CurrentID=0, Status='live', Symbol='EGFEM1P', Aliases='C3orf50, NCRNA00259', description='EGF-like and EMI domain containing 1, pseudogene', other_designations='multiple EGF-like-domains 6 pseudogene', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168249522, end_position_on_the_genomic_accession=168830599, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100421998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421998, CurrentID=0, Status='live', Symbol='LOC100421998', Aliases='', description='olfactory receptor, family 4, subfamily C, member 6 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375729, CurrentID=0, Status='live', Symbol='LOC105375729', Aliases='', description='uncharacterized LOC105375729', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120072405, end_position_on_the_genomic_accession=120080747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375639, CurrentID=0, Status='live', Symbol='LOC105375639', Aliases='', description='uncharacterized LOC105375639', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92444501, end_position_on_the_genomic_accession=92509815, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  93587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93587, CurrentID=0, Status='live', Symbol='TRMT10A', Aliases='HEL-S-88, MSSGM, RG9MTD2, TRM10', description='tRNA methyltransferase 10A', other_designations='RNA (guanine-9-)-methyltransferase domain-containing protein 2|epididymis secretory protein Li 88', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99546707, end_position_on_the_genomic_accession=99564057, orientation='minus', exon_count=9, OMIM=616013, no_hdr0=''),
  93589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93589, CurrentID=0, Status='live', Symbol='CACNA2D4', Aliases='RCD4', description='calcium channel, voltage-dependent, alpha 2/delta subunit 4', other_designations='voltage-gated calcium channel alpha(2)delta-4 subunit|voltage-gated calcium channel subunit alpha-2/delta-4', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1791957, end_position_on_the_genomic_accession=1920865, orientation='minus', exon_count=38, OMIM=608171, no_hdr0=''),
  105375641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375641, CurrentID=0, Status='live', Symbol='LOC105375641', Aliases='', description='uncharacterized LOC105375641', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93495348, end_position_on_the_genomic_accession=93506835, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93594, CurrentID=0, Status='live', Symbol='TBC1D31', Aliases='Gm85, WDR67', description='TBC1 domain family member 31', other_designations='TBC1 domain family, member 31|WD repeat domain 67|WD repeat-containing protein 67', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123072680, end_position_on_the_genomic_accession=123165173, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  100132762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132762, CurrentID=0, Status='live', Symbol='LOC100132762', Aliases='', description='chromosome 17 open reading frame 80 pseudogene', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161568448, end_position_on_the_genomic_accession=161570356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391559, CurrentID=0, Status='live', Symbol='VTI1BP1', Aliases='', description='vesicle transport through interaction with t-SNAREs 1B pseudogene 1', other_designations='vesicle transport through interaction with t-SNAREs homolog 1B pseudogene 1', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100225250, end_position_on_the_genomic_accession=100226270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132764, CurrentID=0, Status='live', Symbol='LOC100132764', Aliases='', description='uncharacterized LOC100132764', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=114548539, end_position_on_the_genomic_accession=114549343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  93611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93611, CurrentID=0, Status='live', Symbol='FBXO44', Aliases='FBG3, FBX30, FBX6A, Fbx44, Fbxo6a', description='F-box protein 44', other_designations='F-box gene 3|F-box protein FBX30|F-box/G-domain protein 3', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11654375, end_position_on_the_genomic_accession=11663327, orientation='plus', exon_count=9, OMIM=609111, no_hdr0=''),
  81138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81138, CurrentID=0, Status='live', Symbol='OR7E101P', Aliases='', description='olfactory receptor family 7 subfamily E member 101 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 101 pseudogene', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46460536, end_position_on_the_genomic_accession=46461209, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101805488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101805488, CurrentID=0, Status='live', Symbol='CCAT2', Aliases='LINC00873, NCCP1', description='colon cancer associated transcript 2 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 873|non-coding RNA involved in cancer predisposition 1', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127400399, end_position_on_the_genomic_accession=127402150, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643276, CurrentID=0, Status='live', Symbol='LOC643276', Aliases='', description='NOC2-like nucleolar associated transcriptional repressor pseudogene', other_designations='nucleolar complex associated 2 homolog (S. cerevisiae) pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130229320, end_position_on_the_genomic_accession=130232086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100101554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101554, CurrentID=0, Status='live', Symbol='BRI3P3', Aliases='', description='brain protein I3 pseudogene 3', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17981189, end_position_on_the_genomic_accession=17981570, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101805491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101805491, CurrentID=0, Status='live', Symbol='LOC101805491', Aliases='', description='uncharacterized LOC101805491', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46429190, end_position_on_the_genomic_accession=46441833, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101805492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101805492, CurrentID=0, Status='live', Symbol='CASC9', Aliases='ESCCAL-1, ESSCAL1, LINC00981', description='cancer susceptibility candidate 9 (non-protein coding)', other_designations='ESCC-associated lncRNA|esophageal squamous cell carcinoma associated lncRNA-1|esophageal squamous cell carcinoma sssociated lncRNA-1|long intergenic non-protein coding RNA 981', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=75223117, end_position_on_the_genomic_accession=75278889, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93621, CurrentID=0, Status='live', Symbol='MRFAP1', Aliases='PAM14, PGR1', description='Morf4 family associated protein 1', other_designations='Mof4 family associated protein 1|T-cell activation protein|protein associated with MRG of 14 kDa', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6640091, end_position_on_the_genomic_accession=6642745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93622, CurrentID=0, Status='live', Symbol='LOC93622', Aliases='', description='Morf4 family associated protein 1-like 1 pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6674094, end_position_on_the_genomic_accession=6676047, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  93624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93624, CurrentID=0, Status='live', Symbol='TADA2B', Aliases='ADA2(beta), ADA2B', description='transcriptional adaptor 2B', other_designations='ADA2-beta|ADA2-like protein beta|transcriptional adaptor 2 (ADA2 homolog, yeast)-beta', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7043429, end_position_on_the_genomic_accession=7057952, orientation='plus', exon_count=3, OMIM=608790, no_hdr0=''),
  93627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93627, CurrentID=0, Status='live', Symbol='TBCK', Aliases='HSPC302L, TBCK', description='TBC1 domain containing kinase', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106044317, end_position_on_the_genomic_accession=106316704, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  391560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391560, CurrentID=0, Status='live', Symbol='RPL32P7', Aliases='RPL32_1_390', description='ribosomal protein L32 pseudogene 7', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101599239, end_position_on_the_genomic_accession=101599644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93643, CurrentID=0, Status='live', Symbol='TJAP1', Aliases='PILT, TJP4', description='tight junction associated protein 1 (peripheral)', other_designations='protein incorporated later into tight junctions|tight junction protein 4 (peripheral)', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43477523, end_position_on_the_genomic_accession=43506556, orientation='plus', exon_count=15, OMIM=612658, no_hdr0=''),
  101929261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929261, CurrentID=0, Status='live', Symbol='LOC101929261', Aliases='', description='uncharacterized LOC101929261', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7346986, end_position_on_the_genomic_accession=7347977, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  93649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93649, CurrentID=0, Status='live', Symbol='MYOCD', Aliases='MYCD', description='myocardin', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12665890, end_position_on_the_genomic_accession=12768949, orientation='plus', exon_count=15, OMIM=606127, no_hdr0=''),
  93650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93650, CurrentID=0, Status='live', Symbol='ACPT', Aliases='', description='acid phosphatase, testicular', other_designations='', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50790415, end_position_on_the_genomic_accession=50795224, orientation='plus', exon_count=11, OMIM=606362, no_hdr0=''),
  93653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93653, CurrentID=0, Status='live', Symbol='ST7-AS1', Aliases='ST7AS1, ST7OT1', description='ST7 antisense RNA 1', other_designations='ST7 antisense RNA 1 (non-protein coding)|ST7 overlapping transcript 1 (antisense non-coding RNA)|ST7 overlapping transcript 1 (non-protein coding)|ST7 overlapping transcript antisense 1', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116952446, end_position_on_the_genomic_accession=116954334, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  93654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93654, CurrentID=0, Status='live', Symbol='ST7-AS2', Aliases='ST7AS2, ST7OT2', description='ST7 antisense RNA 2', other_designations='ST7 antisense RNA 2 (non-protein coding)|ST7 overlapping transcript 2 (antisense non-coding RNA)|ST7 overlapping transcript 2 (non-protein coding)', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117072072, end_position_on_the_genomic_accession=117145592, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  93655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93655, CurrentID=0, Status='live', Symbol='ST7-OT3', Aliases='NCRNA00026, ST7, ST7OT3', description='ST7 overlapping transcript 3', other_designations='ST7 overlapping transcript 3 (non-coding RNA)|ST7 overlapping transcript 3 (non-protein coding)|suppression of tumorigenicity 7', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117182681, end_position_on_the_genomic_accession=117209937, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375652, CurrentID=0, Status='live', Symbol='LOC105375652', Aliases='', description='uncharacterized LOC105375652', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96264673, end_position_on_the_genomic_accession=96284001, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93659, CurrentID=0, Status='live', Symbol='CGB5', Aliases='CGB, HCG, hCGB', description='chorionic gonadotropin, beta polypeptide 5', other_designations='chorionic gonadotropin beta 5 subunit|chorionic gonadotropin beta subunit', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49043845, end_position_on_the_genomic_accession=49045311, orientation='plus', exon_count=3, OMIM=608825, no_hdr0=''),
  93661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93661, CurrentID=0, Status='live', Symbol='CAPZA3', Aliases='CAPPA3, Gsg3, HEL-S-86', description='capping protein (actin filament) muscle Z-line, alpha 3', other_designations='CP-alpha-3|CapZ alpha-3|F-actin capping protein alpha-3 subunit|epididymis secretory protein Li 86|germ cell-specific protein 3', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18738111, end_position_on_the_genomic_accession=18739188, orientation='plus', exon_count=1, OMIM=608722, no_hdr0=''),
  93663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93663, CurrentID=0, Status='live', Symbol='ARHGAP18', Aliases='MacGAP, SENEX, bA307O14.2', description='Rho GTPase activating protein 18', other_designations='rho-type GTPase-activating protein 18', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129576145, end_position_on_the_genomic_accession=129710225, orientation='minus', exon_count=16, OMIM=613351, no_hdr0=''),
  93664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93664, CurrentID=0, Status='live', Symbol='CADPS2', Aliases='', description='Ca++-dependent secretion activator 2', other_designations='CAPS-2|Ca2+-dependent activator protein for secretion 2|calcium-dependent activator protein for secretion 2', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122318424, end_position_on_the_genomic_accession=122886759, orientation='minus', exon_count=32, OMIM=609978, no_hdr0=''),
  93668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93668, CurrentID=0, Status='live', Symbol='HPYR1', Aliases='HPRG1, LINC00027, NCRNA00027', description='Helicobacter pylori responsive 1 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 27', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132560498, end_position_on_the_genomic_accession=132561479, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375654, CurrentID=0, Status='live', Symbol='LOC105375654', Aliases='', description='uncharacterized LOC105375654', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96774987, end_position_on_the_genomic_accession=96793342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378662, CurrentID=0, Status='live', Symbol='LOC105378662', Aliases='', description='uncharacterized LOC105378662', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38915225, end_position_on_the_genomic_accession=38919471, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422002, CurrentID=0, Status='live', Symbol='LOC100422002', Aliases='', description='olfactory receptor, family 5, subfamily J, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927762, CurrentID=0, Status='live', Symbol='LOC101927762', Aliases='', description='uncharacterized LOC101927762', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=933026, end_position_on_the_genomic_accession=942743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100101629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101629, CurrentID=0, Status='live', Symbol='GAGE8', Aliases='CT4.8, GAGE-8', description='G antigen 8', other_designations='cancer/testis antigen 4.8|cancer/testis antigen family 4, member 8', map_location='Xp11 not on reference assembly', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=49207160, end_position_on_the_genomic_accession=49214420, orientation='plus', exon_count=5, OMIM=300638, no_hdr0=''),
  649133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649133, CurrentID=0, Status='live', Symbol='LOC649133', Aliases='', description='uncharacterized LOC649133', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120249759, end_position_on_the_genomic_accession=120265932, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102723099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723099, CurrentID=0, Status='live', Symbol='GHET1', Aliases='lncRNA-GHET1', description='gastric carcinoma proliferation enhancing transcript 1', other_designations='gastric carcinoma high expressed transcript 1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148987527, end_position_on_the_genomic_accession=148989429, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723100, CurrentID=0, Status='live', Symbol='OTSC10', Aliases='', description='otosclerosis 10', other_designations='', map_location='1q41-q44', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615589, no_hdr0=''),
  105375665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375665, CurrentID=0, Status='live', Symbol='LOC105375665', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99003235, end_position_on_the_genomic_accession=99003847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422004, CurrentID=0, Status='live', Symbol='LOC100422004', Aliases='', description='olfactory receptor, family 5, subfamily R, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375666, CurrentID=0, Status='live', Symbol='LOC105375666', Aliases='', description='uncharacterized LOC105375666', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98996760, end_position_on_the_genomic_accession=99013022, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375735, CurrentID=0, Status='live', Symbol='LOC105375735', Aliases='', description='uncharacterized LOC105375735', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122977030, end_position_on_the_genomic_accession=123000154, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927764, CurrentID=0, Status='live', Symbol='LOC101927764', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119710868, end_position_on_the_genomic_accession=119760452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100132789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132789, CurrentID=0, Status='live', Symbol='LOC100132789', Aliases='', description='transmembrane protein 126A pseudogene', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38763813, end_position_on_the_genomic_accession=38764413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375670, CurrentID=0, Status='live', Symbol='LOC105375670', Aliases='', description='uncharacterized LOC105375670', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100431067, end_position_on_the_genomic_accession=100500705, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100422005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422005, CurrentID=0, Status='live', Symbol='LOC100422005', Aliases='', description='olfactory receptor, family 8, subfamily H, member 3 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131032, CurrentID=0, Status='live', Symbol='LOC100131032', Aliases='', description='uncharacterized LOC100131032', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=208697157, end_position_on_the_genomic_accession=208697934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375672, CurrentID=0, Status='live', Symbol='LOC105375672', Aliases='', description='uncharacterized LOC105375672', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100663612, end_position_on_the_genomic_accession=100669096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81166, CurrentID=0, Status='live', Symbol='OR8K4P', Aliases='', description='olfactory receptor family 8 subfamily K member 4 pseudogene', other_designations='olfactory receptor, family 8, subfamily K, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56142515, end_position_on_the_genomic_accession=56143386, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375673, CurrentID=0, Status='live', Symbol='LOC105375673', Aliases='', description='uncharacterized LOC105375673', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100723080, end_position_on_the_genomic_accession=100745851, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723165, CurrentID=0, Status='live', Symbol='LINC01169', Aliases='', description='long intergenic non-protein coding RNA 1169', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66582190, end_position_on_the_genomic_accession=66685798, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723166, CurrentID=0, Status='live', Symbol='STAM-AS1', Aliases='locus3182', description='STAM antisense RNA 1 (head to head)', other_designations='locus 3182', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17641284, end_position_on_the_genomic_accession=17643918, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723167, CurrentID=0, Status='live', Symbol='MIR133A1HG', Aliases='MIR1-2HG', description='MIR133A1 host gene (non-protein coding)', other_designations='MIR1-2 host gene (non-protein coding)', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21825479, end_position_on_the_genomic_accession=21831406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723168, CurrentID=0, Status='live', Symbol='IGHV2-70D', Aliases='', description='immunoglobulin heavy variable 2-70D', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81168, CurrentID=0, Status='live', Symbol='OR8J3', Aliases='OR11-173', description='olfactory receptor family 8 subfamily J member 3', other_designations='olfactory receptor OR11-173|olfactory receptor, family 8, subfamily J, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56136771, end_position_on_the_genomic_accession=56137718, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723170, CurrentID=0, Status='live', Symbol='IGHV3-64D', Aliases='', description='immunoglobulin heavy variable 3-64D', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375675, CurrentID=0, Status='live', Symbol='LOC105375675', Aliases='', description='uncharacterized LOC105375675', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101128996, end_position_on_the_genomic_accession=101139903, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101060145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060145, CurrentID=0, Status='live', Symbol='LOC101060145', Aliases='', description='putative gamma-glutamyltranspeptidase 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18253360, end_position_on_the_genomic_accession=18309655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375676, CurrentID=0, Status='live', Symbol='LOC105375676', Aliases='', description='uncharacterized LOC105375676', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101144092, end_position_on_the_genomic_accession=101148321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132797, CurrentID=0, Status='live', Symbol='LOC100132797', Aliases='', description='fragile X mental retardation, autosomal homolog 1 pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42430819, end_position_on_the_genomic_accession=42433356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645773, CurrentID=0, Status='live', Symbol='LOC645773', Aliases='', description='tubulin, alpha 4a pseudogene', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124435673, end_position_on_the_genomic_accession=124437185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369934, CurrentID=0, Status='live', Symbol='LOC105369934', Aliases='', description='uncharacterized LOC105369934', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=99131035, end_position_on_the_genomic_accession=99135125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375680, CurrentID=0, Status='live', Symbol='LOC105375680', Aliases='', description='uncharacterized LOC105375680', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101387319, end_position_on_the_genomic_accession=101393173, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371373, CurrentID=0, Status='live', Symbol='LOC105371373', Aliases='', description='uncharacterized LOC105371373', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83915609, end_position_on_the_genomic_accession=83920952, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  124865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124865, CurrentID=0, Status='live', Symbol='LOC124865', Aliases='', description='MRT4 homolog, ribosome maturation factor pseudogene', other_designations='mRNA turnover 4 homolog (S. cerevisiae) pseudogene', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49939122, end_position_on_the_genomic_accession=49939821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81175, CurrentID=0, Status='live', Symbol='OR7E128P', Aliases='', description='olfactory receptor family 7 subfamily E member 128 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 128 pseudogene|olfactory receptor, family 7, subfamily E, member 128 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71893408, end_position_on_the_genomic_accession=71894439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927767, CurrentID=0, Status='live', Symbol='LOC101927767', Aliases='', description='uncharacterized LOC101927767', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149792523, end_position_on_the_genomic_accession=149793847, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375683, CurrentID=0, Status='live', Symbol='LOC105375683', Aliases='', description='uncharacterized LOC105375683', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102412942, end_position_on_the_genomic_accession=102468670, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  124871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124871, CurrentID=0, Status='live', Symbol='FLJ40194', Aliases='', description='uncharacterized FLJ40194', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49248243, end_position_on_the_genomic_accession=49258665, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375739, CurrentID=0, Status='live', Symbol='LOC105375739', Aliases='', description='uncharacterized LOC105375739', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123719810, end_position_on_the_genomic_accession=123766246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132808, CurrentID=0, Status='live', Symbol='RBMX2P2', Aliases='', description='RNA binding motif protein, X-linked 2 pseudogene 2', other_designations='', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141263343, end_position_on_the_genomic_accession=141264247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81182, CurrentID=0, Status='live', Symbol='OR5W1P', Aliases='OR11-144', description='olfactory receptor family 5 subfamily W member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily W, member 1 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55903343, end_position_on_the_genomic_accession=55904145, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422009, CurrentID=0, Status='live', Symbol='LOC100422009', Aliases='', description='olfactory receptor, family 4, subfamily D, member 9 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375692, CurrentID=0, Status='live', Symbol='LOC105375692', Aliases='', description='uncharacterized LOC105375692', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104428129, end_position_on_the_genomic_accession=104431295, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372747, CurrentID=0, Status='live', Symbol='LOC105372747', Aliases='', description='uncharacterized LOC105372747', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23408637, end_position_on_the_genomic_accession=23434254, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375693, CurrentID=0, Status='live', Symbol='LOC105375693', Aliases='', description='uncharacterized LOC105375693', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104880926, end_position_on_the_genomic_accession=104886052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649166, CurrentID=0, Status='live', Symbol='LOC649166', Aliases='', description='eukaryotic translation initiation factor 3 subunit J pseudogene', other_designations='eukaryotic translation initiation factor 3, subunit J pseudogene', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87741223, end_position_on_the_genomic_accession=87743569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81188, CurrentID=0, Status='live', Symbol='OR5M12P', Aliases='', description='olfactory receptor family 5 subfamily M member 12 pseudogene', other_designations='olfactory receptor, family 5, subfamily M, member 12 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56628933, end_position_on_the_genomic_accession=56629900, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375695, CurrentID=0, Status='live', Symbol='LOC105375695', Aliases='', description='uncharacterized LOC105375695', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105117396, end_position_on_the_genomic_accession=105140172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375697, CurrentID=0, Status='live', Symbol='LOC105375697', Aliases='', description='uncharacterized LOC105375697', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105534462, end_position_on_the_genomic_accession=105543267, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723305, CurrentID=0, Status='live', Symbol='LOC102723305', Aliases='', description='uncharacterized LOC102723305', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29409570, end_position_on_the_genomic_accession=29415438, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723306, CurrentID=0, Status='live', Symbol='LOC102723306', Aliases='', description='uncharacterized LOC102723306', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34541234, end_position_on_the_genomic_accession=34543429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723307, CurrentID=0, Status='live', Symbol='LOC102723307', Aliases='', description='uncharacterized LOC102723307', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44583383, end_position_on_the_genomic_accession=44606467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723309, CurrentID=0, Status='live', Symbol='LOC102723309', Aliases='', description='uncharacterized LOC102723309', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120541605, end_position_on_the_genomic_accession=120546266, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723311, CurrentID=0, Status='live', Symbol='LOC102723311', Aliases='', description='uncharacterized LOC102723311', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30007304, end_position_on_the_genomic_accession=30010299, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723312, CurrentID=0, Status='live', Symbol='LOC102723312', Aliases='', description='angiogenic factor with G patch and FHA domains 1 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723313, CurrentID=0, Status='live', Symbol='LOC102723313', Aliases='', description='uncharacterized LOC102723313', other_designations='CTD-2135J3.3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10728358, end_position_on_the_genomic_accession=10780112, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723314, CurrentID=0, Status='live', Symbol='PRELID3BP11', Aliases='SLMO2P11', description='PRELI domain containing 3B pseudogene 11', other_designations='SLMO2 pseudogene 11|slowmo homolog 2 (Drosophila) pseudogene 11', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6726269, end_position_on_the_genomic_accession=6735412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723316, CurrentID=0, Status='live', Symbol='FRG1DP', Aliases='', description='FSHD region gene 1 family member D, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29080251, end_position_on_the_genomic_accession=29100567, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102723318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723318, CurrentID=0, Status='live', Symbol='LINC01053', Aliases='', description='long intergenic non-protein coding RNA 1053', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25181332, end_position_on_the_genomic_accession=25190657, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723319, CurrentID=0, Status='live', Symbol='LOC102723319', Aliases='', description='uncharacterized LOC102723319', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34553934, end_position_on_the_genomic_accession=34562381, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723320, CurrentID=0, Status='live', Symbol='CERS3-AS1', Aliases='', description='CERS3 antisense RNA 1', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100372939, end_position_on_the_genomic_accession=100437914, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723321, CurrentID=0, Status='live', Symbol='LOC102723321', Aliases='', description='uncharacterized LOC102723321', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147700376, end_position_on_the_genomic_accession=147810024, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723322, CurrentID=0, Status='live', Symbol='LOC102723322', Aliases='', description='uncharacterized LOC102723322', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85172438, end_position_on_the_genomic_accession=85177041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723323, CurrentID=0, Status='live', Symbol='LOC102723323', Aliases='', description='uncharacterized LOC102723323', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51387701, end_position_on_the_genomic_accession=51525613, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  93948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93948, CurrentID=0, Status='live', Symbol='RPS23P1', Aliases='RPS23_3_930', description='ribosomal protein S23 pseudogene 1', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97865025, end_position_on_the_genomic_accession=97865529, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81194, CurrentID=0, Status='live', Symbol='OR5F2P', Aliases='', description='olfactory receptor family 5 subfamily F member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily F, member 2 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56015016, end_position_on_the_genomic_accession=56015956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93953, CurrentID=0, Status='live', Symbol='ACRC', Aliases='NAAR1', description='acidic repeat containing', other_designations='putative nuclear protein', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71578024, end_position_on_the_genomic_accession=71613583, orientation='plus', exon_count=17, OMIM=300369, no_hdr0=''),
  102723330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723330, CurrentID=0, Status='live', Symbol='LOC102723330', Aliases='', description='uncharacterized LOC102723330', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17695267, end_position_on_the_genomic_accession=17697526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723331, CurrentID=0, Status='live', Symbol='LOC102723331', Aliases='', description='uncharacterized LOC102723331', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168848184, end_position_on_the_genomic_accession=168865953, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93956, CurrentID=0, Status='live', Symbol='AOCH', Aliases='', description='Acromegaloid features, overgrowth, cleft palate, and hernia', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606049, no_hdr0=''),
  102723333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723333, CurrentID=0, Status='live', Symbol='LOC102723333', Aliases='', description='retinoic acid receptor responder protein 2 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723334, CurrentID=0, Status='live', Symbol='LOC102723334', Aliases='', description='uncharacterized LOC102723334', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44992636, end_position_on_the_genomic_accession=44999810, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723335, CurrentID=0, Status='live', Symbol='LOC102723335', Aliases='', description='uncharacterized LOC102723335', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100558677, end_position_on_the_genomic_accession=100559798, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723337, CurrentID=0, Status='live', Symbol='LOC102723337', Aliases='', description='uncharacterized LOC102723337', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133011764, end_position_on_the_genomic_accession=133015765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723338, CurrentID=0, Status='live', Symbol='LOC102723338', Aliases='', description='uncharacterized LOC102723338', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84526820, end_position_on_the_genomic_accession=84583155, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  102723339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723339, CurrentID=0, Status='live', Symbol='LOC102723339', Aliases='', description='uncharacterized LOC102723339', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8066944, end_position_on_the_genomic_accession=8071825, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723340, CurrentID=0, Status='live', Symbol='LOC102723340', Aliases='', description='uncharacterized LOC102723340', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95996446, end_position_on_the_genomic_accession=96011704, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723341, CurrentID=0, Status='live', Symbol='LOC102723341', Aliases='', description='uncharacterized LOC102723341', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125370211, end_position_on_the_genomic_accession=125445210, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723342, CurrentID=0, Status='live', Symbol='LOC102723342', Aliases='', description='uncharacterized LOC102723342', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104767967, end_position_on_the_genomic_accession=104770123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723344, CurrentID=0, Status='live', Symbol='LOC102723344', Aliases='', description='uncharacterized LOC102723344', other_designations='', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63390230, end_position_on_the_genomic_accession=63437536, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375704, CurrentID=0, Status='live', Symbol='LOC105375704', Aliases='', description='uncharacterized LOC105375704', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108267859, end_position_on_the_genomic_accession=108443467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723346, CurrentID=0, Status='live', Symbol='HCG21', Aliases='', description='HLA complex group 21 (non-protein coding)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30952446, end_position_on_the_genomic_accession=30955720, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723347, CurrentID=0, Status='live', Symbol='LOC102723347', Aliases='', description='uncharacterized LOC102723347', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723348, CurrentID=0, Status='live', Symbol='LOC102723348', Aliases='', description='uncharacterized LOC102723348', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149127872, end_position_on_the_genomic_accession=149153274, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  93973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93973, CurrentID=0, Status='live', Symbol='ACTR8', Aliases='ARP8, INO80N, hArp8', description='ARP8 actin-related protein 8 homolog (yeast)', other_designations='INO80 complex subunit N', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53864931, end_position_on_the_genomic_accession=53882202, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  93974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93974, CurrentID=0, Status='live', Symbol='ATPIF1', Aliases='ATPI, ATPIP, IP', description='ATPase inhibitory factor 1', other_designations='ATP synthase inhibitor protein|ATPase inhibitor protein|IF(1)|IF1|inhibitor of F(1)F(o)-ATPase', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28236091, end_position_on_the_genomic_accession=28238105, orientation='plus', exon_count=3, OMIM=614981, no_hdr0=''),
  93978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93978, CurrentID=0, Status='live', Symbol='CLEC6A', Aliases='CLEC4N, CLECSF10', description='C-type lectin domain family 6 member A', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 10|C-type lectin domain family 6, member A|C-type lectin superfamily member 10|DC-associated C-type lectin 2|dectin 2|dectin-2|dendritic cell-associated C-type lectin 2', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8455995, end_position_on_the_genomic_accession=8478330, orientation='plus', exon_count=6, OMIM=613579, no_hdr0=''),
  93979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93979, CurrentID=0, Status='live', Symbol='CPA5', Aliases='', description='carboxypeptidase A5', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130344790, end_position_on_the_genomic_accession=130374709, orientation='plus', exon_count=19, OMIM=609561, no_hdr0=''),
  102723356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723356, CurrentID=0, Status='live', Symbol='ZFPM2-AS1', Aliases='', description='ZFPM2 antisense RNA 1', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105780410, end_position_on_the_genomic_accession=106060503, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100101917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101917, CurrentID=0, Status='live', Symbol='ELF2P1', Aliases='', description='E74-like factor 2 pseudogene 1', other_designations='E74-like factor 2 (ets domain transcription factor) pseudogene|ELF2P1 pseudogene', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103908424, end_position_on_the_genomic_accession=103911366, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93983, CurrentID=0, Status='live', Symbol='DYT13', Aliases='', description='dystonia 13, torsion', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607671, no_hdr0=''),
  102723360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723360, CurrentID=0, Status='live', Symbol='LOC102723360', Aliases='', description='uncharacterized LOC102723360', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7669397, end_position_on_the_genomic_accession=7692185, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93985, CurrentID=0, Status='live', Symbol='FGS3', Aliases='', description='FG syndrome 3', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300406, no_hdr0=''),
  93986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93986, CurrentID=0, Status='live', Symbol='FOXP2', Aliases='CAGH44, SPCH1, TNRC10', description='forkhead box P2', other_designations='CAG repeat protein 44|forkhead/winged-helix transcription factor|trinucleotide repeat containing 10|trinucleotide repeat-containing gene 10 protein', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=114086310, end_position_on_the_genomic_accession=114693772, orientation='plus', exon_count=24, OMIM=605317, no_hdr0=''),
  105375707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375707, CurrentID=0, Status='live', Symbol='LOC105375707', Aliases='', description='uncharacterized LOC105375707', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111741954, end_position_on_the_genomic_accession=111758305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723364, CurrentID=0, Status='live', Symbol='LOC102723364', Aliases='', description='uncharacterized LOC102723364', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86134721, end_position_on_the_genomic_accession=86267390, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102723365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723365, CurrentID=0, Status='live', Symbol='LOC102723365', Aliases='', description='uncharacterized LOC102723365', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115185952, end_position_on_the_genomic_accession=115196416, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93991, CurrentID=0, Status='live', Symbol='MRX77', Aliases='', description='mental retardation, X-linked 77', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300454, no_hdr0=''),
  102723368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723368, CurrentID=0, Status='live', Symbol='LOC102723368', Aliases='', description='uncharacterized LOC102723368', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108756964, end_position_on_the_genomic_accession=108770865, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93993, CurrentID=0, Status='live', Symbol='NDUFB3P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B3 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa pseudogene 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109092467, end_position_on_the_genomic_accession=109093040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93994, CurrentID=0, Status='live', Symbol='NDUFB3P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B3 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa pseudogene 2', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125178514, end_position_on_the_genomic_accession=125178817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  93995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93995, CurrentID=0, Status='live', Symbol='NDUFB3P4', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B3 pseudogene 4', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa pseudogene 4', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100366588, end_position_on_the_genomic_accession=100367043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  93996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93996, CurrentID=0, Status='live', Symbol='NDUFB3P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B3 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa pseudogene 3', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53011929, end_position_on_the_genomic_accession=53012398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  93997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93997, CurrentID=0, Status='live', Symbol='NDUFB3P5', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B3 pseudogene 5', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 3, 12kDa pseudogene 5', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141380777, end_position_on_the_genomic_accession=141381237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  93999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=93999, CurrentID=0, Status='live', Symbol='OTSC2', Aliases='', description='otosclerosis 2', other_designations='', map_location='7q34-q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605727, no_hdr0=''),
  105371681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371681, CurrentID=0, Status='live', Symbol='LOC105371681', Aliases='', description='uncharacterized LOC105371681', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199775675, end_position_on_the_genomic_accession=199783747, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723377, CurrentID=0, Status='live', Symbol='LOC102723377', Aliases='', description='uncharacterized LOC102723377', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71508153, end_position_on_the_genomic_accession=71511920, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100101938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100101938, CurrentID=0, Status='live', Symbol='ANKRD26P3', Aliases='LINC00414', description='ankyrin repeat domain 26 pseudogene 3', other_designations='long intergenic non-protein coding RNA 414', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19262800, end_position_on_the_genomic_accession=19344973, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  94003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94003, CurrentID=0, Status='live', Symbol='PDB4', Aliases='', description='Paget disease of bone 4', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606263, no_hdr0=''),
  102723380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723380, CurrentID=0, Status='live', Symbol='LOC102723380', Aliases='', description='uncharacterized LOC102723380', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43465133, end_position_on_the_genomic_accession=43467509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94005, CurrentID=0, Status='live', Symbol='PIGS', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class S', other_designations='GPI transamidase subunit|phosphatidylinositol glycan anchor biosynthesis, class S|phosphatidylinositol glycan, class S|phosphatidylinositol-glycan biosynthesis class S protein', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28553388, end_position_on_the_genomic_accession=28571869, orientation='minus', exon_count=12, OMIM=610271, no_hdr0=''),
  102723382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723382, CurrentID=0, Status='live', Symbol='LOC102723382', Aliases='', description='ankyrin repeat domain-containing protein 20B-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723383, CurrentID=0, Status='live', Symbol='LOC102723383', Aliases='', description='protein FAM231A/C', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16733807, end_position_on_the_genomic_accession=16734539, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  94008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94008, CurrentID=0, Status='live', Symbol='SCA18', Aliases='SMNA', description='spinocerebellar ataxia 18 (sensory with neurogenic muscular atrophy)', other_designations='', map_location='7q22-q32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607458, no_hdr0=''),
  94009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94009, CurrentID=0, Status='live', Symbol='SERHL', Aliases='BK126B4.1, BK126B4.2, HS126B42, dJ222E13.1', description='serine hydrolase-like (pseudogene)', other_designations='kraken-like', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42500579, end_position_on_the_genomic_accession=42512560, orientation='plus', exon_count=11, OMIM=607979, no_hdr0=''),
  387039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387039, CurrentID=0, Status='live', Symbol='HTATSF1P1', Aliases='HTATSF1P, TAT-SF1-L, dJ1033B10.6', description='HIV-1 Tat specific factor 1 pseudogene 1', other_designations='HIV TAT specific factor 1 pseudogene', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723389, CurrentID=0, Status='live', Symbol='LOC102723389', Aliases='', description='uncharacterized LOC102723389', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23206067, end_position_on_the_genomic_accession=23208507, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  94014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94014, CurrentID=0, Status='live', Symbol='TRICY1', Aliases='', description='trichilemmal cyst 1', other_designations='pilar cyst', map_location='3p24.1-p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609649, no_hdr0=''),
  94015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94015, CurrentID=0, Status='live', Symbol='TTYH2', Aliases='C17orf29', description='tweety family member 2', other_designations='hTTY2|tweety homolog 2', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74213557, end_position_on_the_genomic_accession=74262020, orientation='plus', exon_count=15, OMIM=608855, no_hdr0=''),
  102723392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723392, CurrentID=0, Status='live', Symbol='LOC102723392', Aliases='', description='uncharacterized LOC102723392', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31052213, end_position_on_the_genomic_accession=31056765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94017, CurrentID=0, Status='live', Symbol='TRS-TGA2-1', Aliases='TRNAS1, TRS1', description='transfer RNA-Ser (TGA) 2-1', other_designations='tRNA serine (UGA) 1|tRNA serine 1 (anticodon UGA)|tRNA-Ser-TGA-2-1|transfer RNA serine 1 (anticodon UGA)', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27545689, end_position_on_the_genomic_accession=27545770, orientation='plus', exon_count=0, OMIM=606171, no_hdr0=''),
  102723394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723394, CurrentID=0, Status='live', Symbol='LOC102723394', Aliases='', description='uncharacterized LOC102723394', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101581253, end_position_on_the_genomic_accession=101596474, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723395, CurrentID=0, Status='live', Symbol='LOC102723395', Aliases='', description='uncharacterized LOC102723395', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25836383, end_position_on_the_genomic_accession=25873700, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723396, CurrentID=0, Status='live', Symbol='LOC102723396', Aliases='', description='uncharacterized LOC102723396', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20220737, end_position_on_the_genomic_accession=20232533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723400, CurrentID=0, Status='live', Symbol='LOC102723400', Aliases='', description='C-terminal-binding protein 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7625708, end_position_on_the_genomic_accession=7627977, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  94025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94025, CurrentID=0, Status='live', Symbol='MUC16', Aliases='CA125', description='mucin 16, cell surface associated', other_designations='CA125 ovarian cancer antigen|ovarian cancer-related tumor marker CA125|ovarian carcinoma antigen CA125', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8848844, end_position_on_the_genomic_accession=8981342, orientation='minus', exon_count=84, OMIM=606154, no_hdr0=''),
  94026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94026, CurrentID=0, Status='live', Symbol='POM121L2', Aliases='POM121-L, POM121L', description='POM121 transmembrane nucleoporin like 2', other_designations='OM121 membrane glycoprotein (rat homolog)-like 2|POM121 membrane glycoprotein (rat homolog)-like 2|POM121 membrane glycoprotein-like 2|POM121 transmembrane nucleoporin-like 2', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27309063, end_position_on_the_genomic_accession=27312232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  94027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94027, CurrentID=0, Status='live', Symbol='CGB7', Aliases='CG-beta-a, CGB6', description='chorionic gonadotropin, beta polypeptide 7', other_designations='chorionic gonadotropin beta 7 subunit', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49054274, end_position_on_the_genomic_accession=49057788, orientation='minus', exon_count=5, OMIM=608826, no_hdr0=''),
  105375714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375714, CurrentID=0, Status='live', Symbol='LOC105375714', Aliases='', description='uncharacterized LOC105375714', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116755899, end_position_on_the_genomic_accession=116766531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  94030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94030, CurrentID=0, Status='live', Symbol='LRRC4B', Aliases='HSM, LRIG4', description='leucine rich repeat containing 4B', other_designations='NGL-3|leucine-rich repeats and immunoglobulin-like domains 4|netrin-G3 ligand', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50516892, end_position_on_the_genomic_accession=50568219, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  94031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94031, CurrentID=0, Status='live', Symbol='HTRA3', Aliases='Prsp, Tasp', description='HtrA serine peptidase 3', other_designations='high-temperature requirement factor A3|pregnancy-related serine protease|probable serine protease HTRA3', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8269754, end_position_on_the_genomic_accession=8307111, orientation='plus', exon_count=10, OMIM=608785, no_hdr0=''),
  94032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94032, CurrentID=0, Status='live', Symbol='CAMK2N2', Aliases='CAM-KIIN, CAMKIIN', description='calcium/calmodulin-dependent protein kinase II inhibitor 2', other_designations='CaM-KII inhibitory protein', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184259215, end_position_on_the_genomic_accession=184261463, orientation='minus', exon_count=2, OMIM=608721, no_hdr0=''),
  94033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94033, CurrentID=0, Status='live', Symbol='FTMT', Aliases='MTF', description='ferritin mitochondrial', other_designations='ferritin H subunit|mitochondrial ferritin', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121851955, end_position_on_the_genomic_accession=121852833, orientation='plus', exon_count=1, OMIM=608847, no_hdr0=''),
  105375715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375715, CurrentID=0, Status='live', Symbol='LOC105375715', Aliases='', description='uncharacterized LOC105375715', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116944395, end_position_on_the_genomic_accession=116948719, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723412, CurrentID=0, Status='live', Symbol='LOC102723412', Aliases='', description='uncharacterized LOC102723412', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66332780, end_position_on_the_genomic_accession=66348269, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723413, CurrentID=0, Status='live', Symbol='LOC102723413', Aliases='', description='uncharacterized LOC102723413', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118044566, end_position_on_the_genomic_accession=118058126, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723414, CurrentID=0, Status='live', Symbol='LOC102723414', Aliases='', description='uncharacterized LOC102723414', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36349124, end_position_on_the_genomic_accession=36361576, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  94039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94039, CurrentID=0, Status='live', Symbol='ZNF101', Aliases='HZF12', description='zinc finger protein 101', other_designations='zinc finger protein 101 (Y2)|zinc finger protein 12|zinc finger protein HZF12', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19668153, end_position_on_the_genomic_accession=19683506, orientation='plus', exon_count=5, OMIM=603983, no_hdr0=''),
  102723416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723416, CurrentID=0, Status='live', Symbol='CDC27P3', Aliases='', description='cell division cycle 27 pseudogene 3', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29746181, end_position_on_the_genomic_accession=29748094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375716, CurrentID=0, Status='live', Symbol='LOC105375716', Aliases='', description='uncharacterized LOC105375716', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117084525, end_position_on_the_genomic_accession=117520695, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  102723418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723418, CurrentID=0, Status='live', Symbol='LOC102723418', Aliases='', description='uncharacterized LOC102723418', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36273320, end_position_on_the_genomic_accession=36476196, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  102723421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723421, CurrentID=0, Status='live', Symbol='LOC102723421', Aliases='', description='uncharacterized LOC102723421', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218622885, end_position_on_the_genomic_accession=218636563, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723422, CurrentID=0, Status='live', Symbol='LOC102723422', Aliases='', description='foot protein 1 variant 1-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1038454, end_position_on_the_genomic_accession=1040858, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723423, CurrentID=0, Status='live', Symbol='DUX4L25', Aliases='', description='double homeobox 4 like 25', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133667151, end_position_on_the_genomic_accession=133668675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723424, CurrentID=0, Status='live', Symbol='LOC102723424', Aliases='', description='keratin, type I cytoskeletal 18 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7631304, end_position_on_the_genomic_accession=7636051, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723426, CurrentID=0, Status='live', Symbol='LINC01496', Aliases='', description='long intergenic non-protein coding RNA 1496', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51499909, end_position_on_the_genomic_accession=51507441, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  81211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81211, CurrentID=0, Status='live', Symbol='OR5B19P', Aliases='', description='olfactory receptor family 5 subfamily B member 19 pseudogene', other_designations='olfactory receptor, family 5, subfamily B, member 19 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58343813, end_position_on_the_genomic_accession=58344670, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375718, CurrentID=0, Status='live', Symbol='LOC105375718', Aliases='', description='uncharacterized LOC105375718', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117108456, end_position_on_the_genomic_accession=117114078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723430, CurrentID=0, Status='live', Symbol='LOC102723430', Aliases='', description='uncharacterized LOC102723430', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113157926, end_position_on_the_genomic_accession=113174954, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94056, CurrentID=0, Status='live', Symbol='SYAP1', Aliases='PRO3113', description='synapse associated protein 1', other_designations='SAP47 homolog|synapse associated protein 1, SAP47 homolog', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16719584, end_position_on_the_genomic_accession=16762684, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  94059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94059, CurrentID=0, Status='live', Symbol='LENG9', Aliases='', description='leukocyte receptor cluster (LRC) member 9', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54461796, end_position_on_the_genomic_accession=54463711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723436, CurrentID=0, Status='live', Symbol='LOC102723436', Aliases='', description='uncharacterized LOC102723436', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91518637, end_position_on_the_genomic_accession=91569826, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723438, CurrentID=0, Status='live', Symbol='LOC102723438', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40687166, end_position_on_the_genomic_accession=40692542, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723439, CurrentID=0, Status='live', Symbol='LOC102723439', Aliases='', description='uncharacterized LOC102723439', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74506498, end_position_on_the_genomic_accession=74529324, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723440, CurrentID=0, Status='live', Symbol='LOC102723440', Aliases='', description='uncharacterized LOC102723440', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173040008, end_position_on_the_genomic_accession=173073050, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102723441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723441, CurrentID=0, Status='live', Symbol='CTD-2154B17.1', Aliases='', description='uncharacterized CTD-2154B17.1', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10774934, end_position_on_the_genomic_accession=10787035, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723442, CurrentID=0, Status='live', Symbol='LOC102723442', Aliases='', description='uncharacterized LOC102723442', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47384739, end_position_on_the_genomic_accession=47412256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723444, CurrentID=0, Status='live', Symbol='LOC102723444', Aliases='', description='uncharacterized LOC102723444', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119345856, end_position_on_the_genomic_accession=119355805, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723445, CurrentID=0, Status='live', Symbol='LOC102723445', Aliases='', description='uncharacterized LOC102723445', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131084992, end_position_on_the_genomic_accession=131129831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723446, CurrentID=0, Status='live', Symbol='LOC102723446', Aliases='', description='uncharacterized LOC102723446', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45940560, end_position_on_the_genomic_accession=46000627, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375721, CurrentID=0, Status='live', Symbol='LOC105375721', Aliases='', description='uncharacterized LOC105375721', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117679813, end_position_on_the_genomic_accession=117713902, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723448, CurrentID=0, Status='live', Symbol='LOC102723448', Aliases='', description='uncharacterized LOC102723448', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11751, end_position_on_the_genomic_accession=24501, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723449, CurrentID=0, Status='live', Symbol='DUX4L24', Aliases='', description='double homeobox 4 like 24', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133670450, end_position_on_the_genomic_accession=133671974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723451, CurrentID=0, Status='live', Symbol='LOC102723451', Aliases='', description='uncharacterized protein C21orf140', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7769592, end_position_on_the_genomic_accession=7771151, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723456, CurrentID=0, Status='live', Symbol='LOC102723456', Aliases='', description='classical arabinogalactan protein 4-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36861674, end_position_on_the_genomic_accession=36863150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  94081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94081, CurrentID=0, Status='live', Symbol='SFXN1', Aliases='', description='sideroflexin 1', other_designations='TCC|tricarboxylate carrier protein', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175478511, end_position_on_the_genomic_accession=175529742, orientation='plus', exon_count=11, OMIM=615569, no_hdr0=''),
  105375746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375746, CurrentID=0, Status='live', Symbol='LOC105375746', Aliases='', description='uncharacterized LOC105375746', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125648155, end_position_on_the_genomic_accession=125680038, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723461, CurrentID=0, Status='live', Symbol='LOC102723461', Aliases='', description='uncharacterized LOC102723461', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94086, CurrentID=0, Status='live', Symbol='HSPB9', Aliases='CT51', description='heat shock protein family B (small) member 9', other_designations='cancer/testis antigen 51|heat shock protein, alpha-crystallin-related, B9|small heat shock protein B9', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42122738, end_position_on_the_genomic_accession=42123353, orientation='plus', exon_count=1, OMIM=608344, no_hdr0=''),
  102723463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723463, CurrentID=0, Status='live', Symbol='LOC102723463', Aliases='', description='uncharacterized LOC102723463', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134398707, end_position_on_the_genomic_accession=134406650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723464, CurrentID=0, Status='live', Symbol='LOC102723464', Aliases='', description='uncharacterized LOC102723464', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65617958, end_position_on_the_genomic_accession=65622769, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723465, CurrentID=0, Status='live', Symbol='LOC102723465', Aliases='', description='uncharacterized LOC102723465', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202000095, end_position_on_the_genomic_accession=202010046, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723466, CurrentID=0, Status='live', Symbol='LOC102723466', Aliases='', description='uncharacterized LOC102723466', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56280231, end_position_on_the_genomic_accession=56302899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723468, CurrentID=0, Status='live', Symbol='LOC102723468', Aliases='', description='uncharacterized LOC102723468', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125537797, end_position_on_the_genomic_accession=125557920, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102723471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723471, CurrentID=0, Status='live', Symbol='LOC102723471', Aliases='', description='uncharacterized LOC102723471', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36932105, end_position_on_the_genomic_accession=36936708, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723472, CurrentID=0, Status='live', Symbol='DUX4L23', Aliases='', description='double homeobox 4 like 23', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133673749, end_position_on_the_genomic_accession=133675273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94097, CurrentID=0, Status='live', Symbol='SFXN5', Aliases='BBG-TCC', description='sideroflexin 5', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72942036, end_position_on_the_genomic_accession=73071836, orientation='minus', exon_count=19, OMIM=615572, no_hdr0=''),
  102723475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723475, CurrentID=0, Status='live', Symbol='KCNE1B', Aliases='', description='potassium channel, voltage gated subfamily E regulatory beta subunit 1B', other_designations='putative uncharacterized protein LOC388820', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7789189, end_position_on_the_genomic_accession=7829632, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  94101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94101, CurrentID=0, Status='live', Symbol='ORMDL1', Aliases='', description='ORMDL sphingolipid biosynthesis regulator 1', other_designations='adoplin-1', map_location='2q32', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189767220, end_position_on_the_genomic_accession=189784371, orientation='minus', exon_count=8, OMIM=610073, no_hdr0=''),
  102723478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723478, CurrentID=0, Status='live', Symbol='LOC102723478', Aliases='', description='coxsackievirus and adenovirus receptor pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94103, CurrentID=0, Status='live', Symbol='ORMDL3', Aliases='', description='ORMDL sphingolipid biosynthesis regulator 3', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39921042, end_position_on_the_genomic_accession=39927631, orientation='minus', exon_count=5, OMIM=610075, no_hdr0=''),
  94104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94104, CurrentID=0, Status='live', Symbol='PAXBP1', Aliases='BM020, C21orf66, FSAP105, GCFC, GCFC1', description='PAX3 and PAX7 binding protein 1', other_designations='GC-rich sequence DNA-binding factor 1|GC-rich sequence DNA-binding factor candidate|functional spliceosome-associated protein 105', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32733899, end_position_on_the_genomic_accession=32771858, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  102723481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723481, CurrentID=0, Status='live', Symbol='LOC102723481', Aliases='', description='uncharacterized LOC102723481', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66931539, end_position_on_the_genomic_accession=66942154, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  94107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94107, CurrentID=0, Status='live', Symbol='TMEM203', Aliases='HBEBP1', description='transmembrane protein 203', other_designations='HBeAg-binding protein 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137204082, end_position_on_the_genomic_accession=137205638, orientation='minus', exon_count=1, OMIM=616499, no_hdr0=''),
  102723484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723484, CurrentID=0, Status='live', Symbol='LOC102723484', Aliases='', description='uncharacterized LOC102723484', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723487, CurrentID=0, Status='live', Symbol='LINC01497', Aliases='', description='long intergenic non-protein coding RNA 1497', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69961707, end_position_on_the_genomic_accession=69983544, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723488, CurrentID=0, Status='live', Symbol='LOC102723488', Aliases='', description='uncharacterized LOC102723488', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88760642, end_position_on_the_genomic_accession=88784082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723489, CurrentID=0, Status='live', Symbol='LINC00205', Aliases='C21orf86, NCRNA00205', description='long intergenic non-protein coding RNA 205', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45290197, end_position_on_the_genomic_accession=45296026, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723490, CurrentID=0, Status='live', Symbol='LOC102723490', Aliases='', description='uncharacterized LOC102723490', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36664681, end_position_on_the_genomic_accession=36789186, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  94115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94115, CurrentID=0, Status='live', Symbol='CGB8', Aliases='', description='chorionic gonadotropin, beta polypeptide 8', other_designations='chorionic gonadotropin beta 8 subunit', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49047638, end_position_on_the_genomic_accession=49049111, orientation='minus', exon_count=3, OMIM=608827, no_hdr0=''),
  102723493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723493, CurrentID=0, Status='live', Symbol='LOC102723493', Aliases='', description='uncharacterized LOC102723493', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66984106, end_position_on_the_genomic_accession=67064182, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  102723495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723495, CurrentID=0, Status='live', Symbol='DUX4L22', Aliases='', description='double homeobox 4 like 22', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133677059, end_position_on_the_genomic_accession=133678583, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94120, CurrentID=0, Status='live', Symbol='SYTL3', Aliases='SLP3', description='synaptotagmin like 3', other_designations='exophilin-6|synaptotagmin-like 3', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158644899, end_position_on_the_genomic_accession=158764876, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  94121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94121, CurrentID=0, Status='live', Symbol='SYTL4', Aliases='SLP4', description='synaptotagmin like 4', other_designations='exophilin-2|granuphilin-a', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100674491, end_position_on_the_genomic_accession=100732148, orientation='minus', exon_count=20, OMIM=300723, no_hdr0=''),
  94122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94122, CurrentID=0, Status='live', Symbol='SYTL5', Aliases='slp5', description='synaptotagmin like 5', other_designations='exophilin 9|synaptotagmin-like 5', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37906068, end_position_on_the_genomic_accession=38129294, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  81223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81223, CurrentID=0, Status='live', Symbol='OR5AP1P', Aliases='', description='olfactory receptor family 5 subfamily AP member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily AP, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56633163, end_position_on_the_genomic_accession=56634102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132850, CurrentID=0, Status='live', Symbol='LOC100132850', Aliases='', description='uncharacterized LOC100132850', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=43087788, end_position_on_the_genomic_accession=43088201, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723502, CurrentID=0, Status='live', Symbol='LOC102723502', Aliases='', description='POTE ankyrin domain family member B-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371383, CurrentID=0, Status='live', Symbol='LOC105371383', Aliases='', description='uncharacterized LOC105371383', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85634228, end_position_on_the_genomic_accession=85639847, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422017, CurrentID=0, Status='live', Symbol='LOC100422017', Aliases='', description='olfactory receptor, family 4, subfamily F, member 5 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723505, CurrentID=0, Status='live', Symbol='LOC102723505', Aliases='', description='uncharacterized LOC102723505', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72021851, end_position_on_the_genomic_accession=72025979, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723506, CurrentID=0, Status='live', Symbol='LOC102723506', Aliases='', description='uncharacterized LOC102723506', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79515804, end_position_on_the_genomic_accession=79527402, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723507, CurrentID=0, Status='live', Symbol='LOC102723507', Aliases='', description='uncharacterized LOC102723507', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36121182, end_position_on_the_genomic_accession=36136398, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723508, CurrentID=0, Status='live', Symbol='KANTR', Aliases='LINC01155, Spasm', description='KDM5C adjacent non-coding transcript', other_designations='long intergenic non-protein coding RNA 1155', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53094157, end_position_on_the_genomic_accession=53144072, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723509, CurrentID=0, Status='live', Symbol='LOC102723509', Aliases='', description='uncharacterized LOC102723509', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94134, CurrentID=0, Status='live', Symbol='ARHGAP12', Aliases='', description='Rho GTPase activating protein 12', other_designations='rho-type GTPase-activating protein 12', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31805398, end_position_on_the_genomic_accession=31928876, orientation='minus', exon_count=20, OMIM=610577, no_hdr0=''),
  102723512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723512, CurrentID=0, Status='live', Symbol='LOC102723512', Aliases='', description='uncharacterized LOC102723512', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4001909, end_position_on_the_genomic_accession=4028064, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94137, CurrentID=0, Status='live', Symbol='RP1L1', Aliases='DCDC4B', description='retinitis pigmentosa 1-like 1', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10606349, end_position_on_the_genomic_accession=10655107, orientation='minus', exon_count=4, OMIM=608581, no_hdr0=''),
  100233146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233146, CurrentID=0, Status='live', Symbol='HINT1P2', Aliases='', description='histidine triad nucleotide binding protein 1 pseudogene 2', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95018000, end_position_on_the_genomic_accession=95018438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100233147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233147, CurrentID=0, Status='live', Symbol='ARHI1', Aliases='', description='Age-related hearing impairment 1', other_designations='', map_location='8q24.13-q24.22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612448, no_hdr0=''),
  94140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94140, CurrentID=0, Status='live', Symbol='DFNA35', Aliases='', description='deafness, autosomal dominant 35', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723517, CurrentID=0, Status='live', Symbol='LOC102723517', Aliases='', description='uncharacterized LOC102723517', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72034107, end_position_on_the_genomic_accession=72057734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723518, CurrentID=0, Status='live', Symbol='DUX4L21', Aliases='', description='double homeobox 4 like 21', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133680369, end_position_on_the_genomic_accession=133681893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723519, CurrentID=0, Status='live', Symbol='LOC102723519', Aliases='', description='putative uncharacterized protein ENSP00000344348', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30613672, end_position_on_the_genomic_accession=30664683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  94144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94144, CurrentID=0, Status='live', Symbol='CDRT2', Aliases='', description='CMT1A duplicated region transcript 2', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94145, CurrentID=0, Status='live', Symbol='CDRT3', Aliases='', description='CMT1A duplicated region transcript 3', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94148, CurrentID=0, Status='live', Symbol='CDRT5', Aliases='', description='CMT1A duplicated region transcript 5', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100233157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233157, CurrentID=0, Status='live', Symbol='BMIQ13', Aliases='', description='Body mass index QTL13', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612459, no_hdr0=''),
  100233158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233158, CurrentID=0, Status='live', Symbol='BMIQ14', Aliases='', description='Body mass index QTL14', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612460, no_hdr0=''),
  100233159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233159, CurrentID=0, Status='live', Symbol='WAGRO', Aliases='DEL11p14p12', description='WAGRO syndrome', other_designations='', map_location='11p13-p12', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612469, no_hdr0=''),
  102723528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723528, CurrentID=0, Status='live', Symbol='LOC102723528', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94153, CurrentID=0, Status='live', Symbol='CDRT10', Aliases='', description='CMT1A duplicated region transcript 10', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94154, CurrentID=0, Status='live', Symbol='CDRT11', Aliases='', description='CMT1A duplicated region transcript 11', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94155, CurrentID=0, Status='live', Symbol='CDRT12', Aliases='', description='CMT1A duplicated region transcript 12', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94156, CurrentID=0, Status='live', Symbol='CDRT13', Aliases='', description='CMT1A duplicated region transcript 13', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723533, CurrentID=0, Status='live', Symbol='LOC102723533', Aliases='', description='uncharacterized LOC102723533', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51395080, end_position_on_the_genomic_accession=51402733, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  94158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94158, CurrentID=0, Status='live', Symbol='CDRT15P1', Aliases='CDRT15P', description='CMT1A duplicated region transcript 15 pseudogene 1', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14024498, end_position_on_the_genomic_accession=14025598, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723535, CurrentID=0, Status='live', Symbol='LOC102723535', Aliases='', description='uncharacterized LOC102723535', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30554416, end_position_on_the_genomic_accession=30578043, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94160, CurrentID=0, Status='live', Symbol='ABCC12', Aliases='MRP9', description='ATP binding cassette subfamily C member 12', other_designations='ATP-binding cassette sub-family C member 12|ATP-binding cassette transporter sub-family C member 12|ATP-binding cassette, sub-family C (CFTR/MRP), member 12', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48082973, end_position_on_the_genomic_accession=48146770, orientation='minus', exon_count=29, OMIM=607041, no_hdr0=''),
  94161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94161, CurrentID=0, Status='live', Symbol='SNORD46', Aliases='RNU40, RNU46, U40, U46', description='small nucleolar RNA, C/D box 46', other_designations='RNA, U40 small nuclear|RNA, U40 small nucleolar|RNA, U46 small nuclear|U40 small nucleolar RNA|U40 snoRNA', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44776492, end_position_on_the_genomic_accession=44776589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723538, CurrentID=0, Status='live', Symbol='DUX4L20', Aliases='', description='double homeobox 4 like 20', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133683668, end_position_on_the_genomic_accession=133685192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94163, CurrentID=0, Status='live', Symbol='SNORD38B', Aliases='RNU38B, U38B', description='small nucleolar RNA, C/D box 38B', other_designations='RNA, U38B small nucleolar|U38B small nucleolar RNA|U38B snoRNA', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44778390, end_position_on_the_genomic_accession=44778458, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723540, CurrentID=0, Status='live', Symbol='LOC102723540', Aliases='', description='uncharacterized LOC102723540', other_designations='CTC-523E23.11', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34849048, end_position_on_the_genomic_accession=34860269, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375737, CurrentID=0, Status='live', Symbol='LOC105375737', Aliases='', description='uncharacterized LOC105375737', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123403576, end_position_on_the_genomic_accession=123416423, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723544, CurrentID=0, Status='live', Symbol='LOC102723544', Aliases='', description='uncharacterized LOC102723544', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=253442, end_position_on_the_genomic_accession=257299, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101180897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180897, CurrentID=0, Status='live', Symbol='LOC101180897', Aliases='', description='Rab geranylgeranyltransferase, beta subunit pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113360243, end_position_on_the_genomic_accession=113363025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723546, CurrentID=0, Status='live', Symbol='LOC102723546', Aliases='', description='uncharacterized LOC102723546', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131361194, end_position_on_the_genomic_accession=131399832, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723547, CurrentID=0, Status='live', Symbol='CSAG2', Aliases='CSAG3B, CT24.2, TRAG-3, TRAG3', description='CSAG family, member 2', other_designations='CSAG family, member 3B|cancer/testis antigen 24.2|cancer/testis antigen family 24, member 2|chondrosarcoma-associated gene 2/3 protein-like|taxol resistance associated gene 3|taxol-resistant-associated gene 3 protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152708261, end_position_on_the_genomic_accession=152714012, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723548, CurrentID=0, Status='live', Symbol='LOC102723548', Aliases='', description='uncharacterized LOC102723548', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69262912, end_position_on_the_genomic_accession=69298808, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375738, CurrentID=0, Status='live', Symbol='LOC105375738', Aliases='', description='uncharacterized LOC105375738', other_designations='CTD-2552K11.2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123657958, end_position_on_the_genomic_accession=123667312, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723550, CurrentID=0, Status='live', Symbol='LOC102723550', Aliases='', description='uncharacterized LOC102723550', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=62239210, end_position_on_the_genomic_accession=62248068, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723552, CurrentID=0, Status='live', Symbol='LOC102723552', Aliases='', description='ankyrin repeat domain-containing protein 20A2', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30681105, end_position_on_the_genomic_accession=30723938, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  102723553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723553, CurrentID=0, Status='live', Symbol='SMIM11B', Aliases='', description='small integral membrane protein 11B', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7744419, end_position_on_the_genomic_accession=7772253, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723555, CurrentID=0, Status='live', Symbol='SPDYE16', Aliases='', description='speedy/RINGO cell cycle regulator family member E16', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76533538, end_position_on_the_genomic_accession=76541484, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723557, CurrentID=0, Status='live', Symbol='LOC102723557', Aliases='', description='uncharacterized LOC102723557', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126752002, end_position_on_the_genomic_accession=126777046, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723560, CurrentID=0, Status='live', Symbol='LOC102723560', Aliases='', description='uncharacterized LOC102723560', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=62725618, end_position_on_the_genomic_accession=62835682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723561, CurrentID=0, Status='live', Symbol='LOC102723561', Aliases='', description='uncharacterized LOC102723561', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723562, CurrentID=0, Status='live', Symbol='LINC01498', Aliases='', description='long intergenic non-protein coding RNA 1498', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108459425, end_position_on_the_genomic_accession=108473689, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371385, CurrentID=0, Status='live', Symbol='LOC105371385', Aliases='', description='uncharacterized LOC105371385', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85520673, end_position_on_the_genomic_accession=85522866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422019, CurrentID=0, Status='live', Symbol='EFTUD1P2', Aliases='', description='elongation factor Tu GTP binding domain containing 1 pseudogene 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65142703, end_position_on_the_genomic_accession=65145351, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100233198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233198, CurrentID=0, Status='live', Symbol='TMX2P1', Aliases='', description='thioredoxin-related transmembrane protein 2 pseudogene 1', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37885683, end_position_on_the_genomic_accession=37887573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375741, CurrentID=0, Status='live', Symbol='LOC105375741', Aliases='', description='uncharacterized LOC105375741', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124728618, end_position_on_the_genomic_accession=124757131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723568, CurrentID=0, Status='live', Symbol='LOC102723568', Aliases='', description='uncharacterized LOC102723568', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34692784, end_position_on_the_genomic_accession=34877212, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102723570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723570, CurrentID=0, Status='live', Symbol='CDC27P9', Aliases='', description='cell division cycle 27 pseudogene 9', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8240739, end_position_on_the_genomic_accession=8243081, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723573, CurrentID=0, Status='live', Symbol='LOC102723573', Aliases='', description='40S ribosomal protein S8 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723574, CurrentID=0, Status='live', Symbol='LOC102723574', Aliases='', description='HAUS augmin-like complex subunit 6 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53862235, end_position_on_the_genomic_accession=53863518, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723575, CurrentID=0, Status='live', Symbol='LOC102723575', Aliases='', description='uncharacterized LOC102723575', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59537532, end_position_on_the_genomic_accession=59560746, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723576, CurrentID=0, Status='live', Symbol='LOC102723576', Aliases='', description='uncharacterized LOC102723576', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99378855, end_position_on_the_genomic_accession=99404606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100233209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233209, CurrentID=0, Status='live', Symbol='PCED1B-AS1', Aliases='', description='PCED1B antisense RNA 1', other_designations='PCED1B antisense RNA 1 (non-protein coding)', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47208420, end_position_on_the_genomic_accession=47216443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723578, CurrentID=0, Status='live', Symbol='LINC01440', Aliases='', description='long intergenic non-protein coding RNA 1440', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55423043, end_position_on_the_genomic_accession=55426791, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387071, CurrentID=0, Status='live', Symbol='FAM136BP', Aliases='C6orf87, FAM136B, dJ40E16.3', description='family with sequence similarity 136 member B, pseudogene', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3045384, end_position_on_the_genomic_accession=3045800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723580, CurrentID=0, Status='live', Symbol='LOC102723580', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723582, CurrentID=0, Status='live', Symbol='LOC102723582', Aliases='', description='uncharacterized LOC102723582', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122416206, end_position_on_the_genomic_accession=122443247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  343381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343381, CurrentID=0, Status='live', Symbol='NBPF2P', Aliases='', description='neuroblastoma breakpoint family member 2, pseudogene', other_designations='neuroblastoma breakpoint family, member 2 (pseudogene)|neuroblastoma breakpoint family, member 2, pseudogene|neuroblastoma breakpoint family, member 3 pseudogene', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21423108, end_position_on_the_genomic_accession=21427960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723584, CurrentID=0, Status='live', Symbol='CDC27P10', Aliases='', description='cell division cycle 27 pseudogene 10', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8423784, end_position_on_the_genomic_accession=8426058, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723585, CurrentID=0, Status='live', Symbol='LOC102723585', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52747522, end_position_on_the_genomic_accession=52749357, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723589, CurrentID=0, Status='live', Symbol='LOC102723589', Aliases='', description='uncharacterized LOC102723589', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197577153, end_position_on_the_genomic_accession=197579243, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723590, CurrentID=0, Status='live', Symbol='BMP7-AS1', Aliases='', description='BMP7 antisense RNA 1', other_designations='', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57213696, end_position_on_the_genomic_accession=57215866, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723591, CurrentID=0, Status='live', Symbol='LOC102723591', Aliases='', description='uncharacterized LOC102723591', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153761629, end_position_on_the_genomic_accession=153766466, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723592, CurrentID=0, Status='live', Symbol='LOC102723592', Aliases='', description='neurofibromin-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100233226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233226, CurrentID=0, Status='live', Symbol='MCI2', Aliases='', description='Myocardial infarction, susceptiblity to, 2', other_designations='', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608557, no_hdr0=''),
  100233227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233227, CurrentID=0, Status='live', Symbol='SLI4', Aliases='', description='Specific language impairment 4', other_designations='', map_location='7q35-q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612514, no_hdr0=''),
  102723596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723596, CurrentID=0, Status='live', Symbol='LOC102723596', Aliases='', description='uncharacterized LOC102723596', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5962732, end_position_on_the_genomic_accession=5985254, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387074, CurrentID=0, Status='live', Symbol='RPL35AP3', Aliases='RPL35A_7_760, bA55K22.4', description='ribosomal protein L35a pseudogene 3', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136973930, end_position_on_the_genomic_accession=136974259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723600, CurrentID=0, Status='live', Symbol='LOC102723600', Aliases='', description='uncharacterized LOC102723600', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81240, CurrentID=0, Status='live', Symbol='OR52Y1P', Aliases='OR52Y2P', description='olfactory receptor family 52 subfamily Y member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily Y, member 1 pseudogene|olfactory receptor, family 52, subfamily Y, member 2 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4793768, end_position_on_the_genomic_accession=4794261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723603, CurrentID=0, Status='live', Symbol='CDC27P11', Aliases='', description='cell division cycle 27 pseudogene 11', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8466830, end_position_on_the_genomic_accession=8470663, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  107048980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=107048980, CurrentID=0, Status='live', Symbol='DSPS', Aliases='', description='Delayed sleep phase syndrome, susceptibility to', other_designations='', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614163, no_hdr0=''),
  102723605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723605, CurrentID=0, Status='live', Symbol='LINC01445', Aliases='', description='long intergenic non-protein coding RNA 1445', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54330697, end_position_on_the_genomic_accession=54349845, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375751, CurrentID=0, Status='live', Symbol='LOC105375751', Aliases='', description='uncharacterized LOC105375751', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126557485, end_position_on_the_genomic_accession=126634564, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  94233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94233, CurrentID=0, Status='live', Symbol='OPN4', Aliases='MOP', description='opsin 4', other_designations='opsin-4', map_location='10q22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86654557, end_position_on_the_genomic_accession=86666460, orientation='plus', exon_count=11, OMIM=606665, no_hdr0=''),
  94234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94234, CurrentID=0, Status='live', Symbol='FOXQ1', Aliases='HFH1', description='forkhead box Q1', other_designations='HFH-1|HNF-3/forkhead-like protein 1|hepatocyte nuclear factor 3 forkhead homolog 1|winged helix/forkhead transcription factor', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1312440, end_position_on_the_genomic_accession=1314758, orientation='plus', exon_count=1, OMIM=612788, no_hdr0=''),
  94235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94235, CurrentID=0, Status='live', Symbol='GNG8', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 8', other_designations='gamma-9', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46633934, end_position_on_the_genomic_accession=46635534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94236, CurrentID=0, Status='live', Symbol='DNAJA1P5', Aliases='HEJ1', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 5', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 5|DnaJ pseudogene', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101872012, end_position_on_the_genomic_accession=101894743, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723613, CurrentID=0, Status='live', Symbol='LOC102723613', Aliases='', description='olfactory receptor 11G2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94239, CurrentID=0, Status='live', Symbol='H2AFV', Aliases='H2A.Z-2, H2AV', description='H2A histone family member V', other_designations='H2A histone family, member V|H2A.F/Z|histone H2A.F/Z|purine-rich binding element protein B', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44826889, end_position_on_the_genomic_accession=44848126, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  94240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94240, CurrentID=0, Status='live', Symbol='EPSTI1', Aliases='BRESI1', description='epithelial stromal interaction 1 (breast)', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42886388, end_position_on_the_genomic_accession=42992271, orientation='minus', exon_count=16, OMIM=607441, no_hdr0=''),
  94241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94241, CurrentID=0, Status='live', Symbol='TP53INP1', Aliases='SIP, TP53DINP1A, TP53INP1B, Teap, p53DINP1, TP53INP1', description='tumor protein p53 inducible nuclear protein 1', other_designations='p53-dependent damage-inducible nuclear protein 1|p53-inducible p53DINP1|stress-induced protein', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94925972, end_position_on_the_genomic_accession=94949387, orientation='minus', exon_count=6, OMIM=606185, no_hdr0=''),
  102723618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723618, CurrentID=0, Status='live', Symbol='LOC102723618', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30486088, end_position_on_the_genomic_accession=30491303, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723619, CurrentID=0, Status='live', Symbol='LOC102723619', Aliases='', description='uncharacterized LOC102723619', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111599498, end_position_on_the_genomic_accession=111601280, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375750, CurrentID=0, Status='live', Symbol='LOC105375750', Aliases='', description='uncharacterized LOC105375750', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126506359, end_position_on_the_genomic_accession=126522310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723622, CurrentID=0, Status='live', Symbol='LOC102723622', Aliases='', description='uncharacterized LOC102723622', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55029389, end_position_on_the_genomic_accession=55045239, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723623, CurrentID=0, Status='live', Symbol='LOC102723623', Aliases='', description='golgin subfamily A member 6-like protein 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23325739, end_position_on_the_genomic_accession=23334275, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102723624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723624, CurrentID=0, Status='live', Symbol='LOC102723624', Aliases='', description='uncharacterized LOC102723624', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68341960, end_position_on_the_genomic_accession=68351098, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723626, CurrentID=0, Status='live', Symbol='TUBB8P8', Aliases='', description='tubulin beta 8 class VIII pseudogene 8', other_designations='tubulin beta-8 chain-like|tubulin, beta 8 class VIII pseudogene 8', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198119655, end_position_on_the_genomic_accession=198120914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132871, CurrentID=0, Status='live', Symbol='RAB28P4', Aliases='', description='RAB28, member RAS oncogene family pseudogene 4', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66640484, end_position_on_the_genomic_accession=66641662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723629, CurrentID=0, Status='live', Symbol='LOC102723629', Aliases='', description='uncharacterized LOC102723629', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5616618, end_position_on_the_genomic_accession=5632478, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723630, CurrentID=0, Status='live', Symbol='LOC102723630', Aliases='', description='uncharacterized LOC102723630', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723631, CurrentID=0, Status='live', Symbol='CT45A10', Aliases='', description='cancer/testis antigen family 45, member A10', other_designations='Cancer/testis antigen 45A10|Cancer/testis antigen family 45 member A10|cancer/testis antigen family 45 member A-like', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135880752, end_position_on_the_genomic_accession=135893577, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102723632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723632, CurrentID=0, Status='live', Symbol='LOC102723632', Aliases='', description='olfactory receptor 4S2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375752, CurrentID=0, Status='live', Symbol='LOC105375752', Aliases='', description='uncharacterized LOC105375752', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127140084, end_position_on_the_genomic_accession=127144150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723634, CurrentID=0, Status='live', Symbol='LOC102723634', Aliases='', description='uncharacterized LOC102723634', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23278823, end_position_on_the_genomic_accession=23281372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723635, CurrentID=0, Status='live', Symbol='LOC102723635', Aliases='', description='uncharacterized LOC102723635', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133211426, end_position_on_the_genomic_accession=133212987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723636, CurrentID=0, Status='live', Symbol='LOC102723636', Aliases='', description='uncharacterized LOC102723636', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30537621, end_position_on_the_genomic_accession=30546376, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723637, CurrentID=0, Status='live', Symbol='LOC102723637', Aliases='', description='uncharacterized LOC102723637', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=30459086, end_position_on_the_genomic_accession=30472313, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723638, CurrentID=0, Status='live', Symbol='LOC102723638', Aliases='', description='uncharacterized LOC102723638', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129034886, end_position_on_the_genomic_accession=129088184, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102723639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723639, CurrentID=0, Status='live', Symbol='LOC102723639', Aliases='', description='uncharacterized LOC102723639', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115013187, end_position_on_the_genomic_accession=115105119, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102723640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723640, CurrentID=0, Status='live', Symbol='LOC102723640', Aliases='', description='uncharacterized LOC102723640', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72599673, end_position_on_the_genomic_accession=72636941, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723641, CurrentID=0, Status='live', Symbol='HID1-AS1', Aliases='', description='HID1 antisense RNA 1', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74970689, end_position_on_the_genomic_accession=74975728, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723643, CurrentID=0, Status='live', Symbol='LOC102723643', Aliases='', description='uncharacterized LOC102723643', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198185965, end_position_on_the_genomic_accession=198189923, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723644, CurrentID=0, Status='live', Symbol='LINC01499', Aliases='', description='long intergenic non-protein coding RNA 1499', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=41714568, end_position_on_the_genomic_accession=41836442, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132874, CurrentID=0, Status='live', Symbol='LOC100132874', Aliases='', description='uncharacterized LOC100132874', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61982796, end_position_on_the_genomic_accession=61986927, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723647, CurrentID=0, Status='live', Symbol='RPL23AP97', Aliases='', description='ribosomal protein L23a pseudogene 97', other_designations='60S ribosomal protein L23a pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114338798, end_position_on_the_genomic_accession=114347209, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723648, CurrentID=0, Status='live', Symbol='LOC102723648', Aliases='', description='uncharacterized LOC102723648', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36146708, end_position_on_the_genomic_accession=36196427, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102723649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723649, CurrentID=0, Status='live', Symbol='LOC102723649', Aliases='', description='uncharacterized LOC102723649', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137665646, end_position_on_the_genomic_accession=137674554, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  94274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94274, CurrentID=0, Status='live', Symbol='PPP1R14A', Aliases='CPI-17, CPI17, PPP1INL', description='protein phosphatase 1 regulatory inhibitor subunit 14A', other_designations='17 kDa PKC-potentiated inhibitory protein of PP1|17-KDa protein|17-kDa PKC-potentiated inhibitory protein of PP1|PKC-potentiated inhibitory protein of PP1|protein kinase C-potentiated inhibitor protein of 17 kDa|protein phosphatase 1, regulatory (inhibitor) subunit 14A', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38251237, end_position_on_the_genomic_accession=38256591, orientation='minus', exon_count=4, OMIM=608153, no_hdr0=''),
  105375755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375755, CurrentID=0, Status='live', Symbol='LOC105375755', Aliases='', description='uncharacterized LOC105375755', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128150986, end_position_on_the_genomic_accession=128186746, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723654, CurrentID=0, Status='live', Symbol='LOC102723654', Aliases='', description='uncharacterized LOC102723654', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129152890, end_position_on_the_genomic_accession=129393969, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102723655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723655, CurrentID=0, Status='live', Symbol='LOC102723655', Aliases='', description='TP53-target gene 3 protein', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723656, CurrentID=0, Status='live', Symbol='LOC102723656', Aliases='', description='uncharacterized LOC102723656', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55750747, end_position_on_the_genomic_accession=55758569, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723657, CurrentID=0, Status='live', Symbol='LOC102723657', Aliases='', description='uncharacterized LOC102723657', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73864592, end_position_on_the_genomic_accession=73873526, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102723658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723658, CurrentID=0, Status='live', Symbol='LOC102723658', Aliases='', description='uncharacterized LOC102723658', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132029049, end_position_on_the_genomic_accession=132038252, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723660, CurrentID=0, Status='live', Symbol='LOC102723660', Aliases='', description='ankyrin repeat domain-containing protein SOWAHC-like', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30725496, end_position_on_the_genomic_accession=30730173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723661, CurrentID=0, Status='live', Symbol='LOC102723661', Aliases='', description='uncharacterized LOC102723661', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95992068, end_position_on_the_genomic_accession=96022880, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375757, CurrentID=0, Status='live', Symbol='LOC105375757', Aliases='', description='uncharacterized LOC105375757', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130300388, end_position_on_the_genomic_accession=130333346, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723665, CurrentID=0, Status='live', Symbol='LOC102723665', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96830632, end_position_on_the_genomic_accession=96832393, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102723668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723668, CurrentID=0, Status='live', Symbol='LOC102723668', Aliases='', description='protein FAM182A-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9578638, end_position_on_the_genomic_accession=9591727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100526771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526771, CurrentID=0, Status='live', Symbol='TMPRSS4-AS1', Aliases='', description='TMPRSS4 antisense RNA 1', other_designations='TMPRSS4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118015772, end_position_on_the_genomic_accession=118086793, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723670, CurrentID=0, Status='live', Symbol='LOC102723670', Aliases='', description='uncharacterized LOC102723670', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55943650, end_position_on_the_genomic_accession=55962935, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723672, CurrentID=0, Status='live', Symbol='LOC102723672', Aliases='', description='uncharacterized LOC102723672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144444, end_position_on_the_genomic_accession=149455, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  94297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94297, CurrentID=0, Status='live', Symbol='PPP1R11P1', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 11 pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 11 pseudogene 1', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21397993, end_position_on_the_genomic_accession=21398364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94298, CurrentID=0, Status='live', Symbol='PPP1R11P2', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 11 pseudogene 2', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 11 pseudogene 2', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57229034, end_position_on_the_genomic_accession=57229415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94299, CurrentID=0, Status='live', Symbol='TP53BP2P1', Aliases='PPP1R13AP', description='tumor protein p53 binding protein 2 pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 13A pseudogene|tumor protein p53 binding protein, 2 pseudogene 1', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  222344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222344, CurrentID=0, Status='live', Symbol='LOC222344', Aliases='', description='TNF receptor-associated factor 6', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  94302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94302, CurrentID=0, Status='live', Symbol='PPP1R14BP4', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 14B pseudogene 4', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 14B pseudogene 4', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35970950, end_position_on_the_genomic_accession=35971762, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723680, CurrentID=0, Status='live', Symbol='CT45A9', Aliases='', description='cancer/testis antigen family 45, member A9', other_designations='cancer/testis antigen 45A9', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135863776, end_position_on_the_genomic_accession=135881166, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375760, CurrentID=0, Status='live', Symbol='LOC105375760', Aliases='', description='uncharacterized LOC105375760', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=131039522, end_position_on_the_genomic_accession=131129953, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723682, CurrentID=0, Status='live', Symbol='LOC102723682', Aliases='', description='uncharacterized LOC102723682', other_designations='CTD-2588E21.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56117316, end_position_on_the_genomic_accession=56118631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723684, CurrentID=0, Status='live', Symbol='LOC102723684', Aliases='', description='uncharacterized LOC102723684', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56659157, end_position_on_the_genomic_accession=56675101, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  94310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94310, CurrentID=0, Status='live', Symbol='PPP1R2P8', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 8', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 8', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20997164, end_position_on_the_genomic_accession=20998748, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723688, CurrentID=0, Status='live', Symbol='LOC102723688', Aliases='', description='trithorax group protein osa-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11252926, end_position_on_the_genomic_accession=11256506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  94313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94313, CurrentID=0, Status='live', Symbol='AUTS5', Aliases='', description='Autism, susceptibility to, 5', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606053, no_hdr0=''),
  102723690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723690, CurrentID=0, Status='live', Symbol='LOC102723690', Aliases='', description='uncharacterized LOC102723690', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139250258, end_position_on_the_genomic_accession=139302539, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723691, CurrentID=0, Status='live', Symbol='LOC102723691', Aliases='', description='uncharacterized LOC102723691', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723692, CurrentID=0, Status='live', Symbol='LOC102723692', Aliases='', description='uncharacterized LOC102723692', other_designations='CTD-2576D5.4', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17134469, end_position_on_the_genomic_accession=17138732, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375762, CurrentID=0, Status='live', Symbol='LOC105375762', Aliases='', description='uncharacterized LOC105375762', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130850137, end_position_on_the_genomic_accession=130884544, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723694, CurrentID=0, Status='live', Symbol='LOC102723694', Aliases='', description='uncharacterized LOC102723694', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134792012, end_position_on_the_genomic_accession=134832031, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102723696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723696, CurrentID=0, Status='live', Symbol='LOC102723696', Aliases='', description='uncharacterized LOC102723696', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125908519, end_position_on_the_genomic_accession=125916486, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375754, CurrentID=0, Status='live', Symbol='LOC105375754', Aliases='', description='uncharacterized LOC105375754', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127664552, end_position_on_the_genomic_accession=127669024, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723699, CurrentID=0, Status='live', Symbol='LOC102723699', Aliases='', description='uncharacterized LOC102723699', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56317459, end_position_on_the_genomic_accession=56342597, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102723701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723701, CurrentID=0, Status='live', Symbol='LOC102723701', Aliases='', description='uncharacterized LOC102723701', other_designations='', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37734761, end_position_on_the_genomic_accession=37737426, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390311, CurrentID=0, Status='live', Symbol='PHBP18', Aliases='', description='prohibitin pseudogene 18', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48168850, end_position_on_the_genomic_accession=48169659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723704, CurrentID=0, Status='live', Symbol='LOC102723704', Aliases='', description='uncharacterized LOC102723704', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102777037, end_position_on_the_genomic_accession=102799575, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132884, CurrentID=0, Status='live', Symbol='LOC100132884', Aliases='', description='uncharacterized LOC100132884', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=148394137, end_position_on_the_genomic_accession=148405726, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723706, CurrentID=0, Status='live', Symbol='LOC102723706', Aliases='', description='uncharacterized LOC102723706', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41986086, end_position_on_the_genomic_accession=41998659, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723708, CurrentID=0, Status='live', Symbol='LOC102723708', Aliases='', description='uncharacterized LOC102723708', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29225594, end_position_on_the_genomic_accession=29231787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723709, CurrentID=0, Status='live', Symbol='LOC102723709', Aliases='', description='uncharacterized LOC102723709', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67711677, end_position_on_the_genomic_accession=67713733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129601, CurrentID=0, Status='live', Symbol='LOC100129601', Aliases='', description='cell division cycle associated 7 pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75401892, end_position_on_the_genomic_accession=75402992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271338, CurrentID=0, Status='live', Symbol='RPL37AP4', Aliases='RPL37A_2_776', description='ribosomal protein L37a pseudogene 4', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163363198, end_position_on_the_genomic_accession=163363474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723712, CurrentID=0, Status='live', Symbol='LOC102723712', Aliases='', description='uncharacterized LOC102723712', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119173960, end_position_on_the_genomic_accession=119177318, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=94337, CurrentID=0, Status='live', Symbol='MNG3', Aliases='', description='Multinodular goiter-3', other_designations='', map_location='3q26.1-q26.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606082, no_hdr0=''),
  102723714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723714, CurrentID=0, Status='live', Symbol='LOC102723714', Aliases='', description='uncharacterized LOC102723714', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74503129, end_position_on_the_genomic_accession=74512628, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723716, CurrentID=0, Status='live', Symbol='LOC102723716', Aliases='', description='uncharacterized LOC102723716', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38400416, end_position_on_the_genomic_accession=38403940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649238, CurrentID=0, Status='live', Symbol='LOC649238', Aliases='', description='putative paraneoplastic antigen-like protein 6B-like', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=152662364, end_position_on_the_genomic_accession=152666148, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723718, CurrentID=0, Status='live', Symbol='LOC102723718', Aliases='', description='uncharacterized LOC102723718', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134491391, end_position_on_the_genomic_accession=134497246, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723720, CurrentID=0, Status='live', Symbol='LOC102723720', Aliases='', description='uncharacterized LOC102723720', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61943099, end_position_on_the_genomic_accession=61945588, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723721, CurrentID=0, Status='live', Symbol='LOC102723721', Aliases='', description='uncharacterized LOC102723721', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6736971, end_position_on_the_genomic_accession=6738943, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723722, CurrentID=0, Status='live', Symbol='LOC102723722', Aliases='', description='uncharacterized LOC102723722', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42129871, end_position_on_the_genomic_accession=42132085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723723, CurrentID=0, Status='live', Symbol='LOC102723723', Aliases='', description='uncharacterized LOC102723723', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119939398, end_position_on_the_genomic_accession=119975189, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723724, CurrentID=0, Status='live', Symbol='LOC102723724', Aliases='', description='uncharacterized LOC102723724', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140845958, end_position_on_the_genomic_accession=140898412, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723725, CurrentID=0, Status='live', Symbol='LOC102723725', Aliases='', description='uncharacterized LOC102723725', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58748750, end_position_on_the_genomic_accession=58750153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723727, CurrentID=0, Status='live', Symbol='LOC102723727', Aliases='', description='mitogen-activated protein kinase 7-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211829718, end_position_on_the_genomic_accession=211852415, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102723728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723728, CurrentID=0, Status='live', Symbol='LOC102723728', Aliases='', description='nodal modulator 3-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18417325, end_position_on_the_genomic_accession=18434787, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102723729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723729, CurrentID=0, Status='live', Symbol='LOC102723729', Aliases='', description='uncharacterized LOC102723729', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41534305, end_position_on_the_genomic_accession=41578368, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723730, CurrentID=0, Status='live', Symbol='LOC102723730', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1548525, end_position_on_the_genomic_accession=1556488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723733, CurrentID=0, Status='live', Symbol='LOC102723733', Aliases='', description='uncharacterized LOC102723733', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25860607, end_position_on_the_genomic_accession=25905684, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375769, CurrentID=0, Status='live', Symbol='LOC105375769', Aliases='', description='uncharacterized LOC105375769', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133143699, end_position_on_the_genomic_accession=133166333, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723736, CurrentID=0, Status='live', Symbol='LOC102723736', Aliases='', description='uncharacterized LOC102723736', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37546155, end_position_on_the_genomic_accession=37550856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723737, CurrentID=0, Status='live', Symbol='CT45A8', Aliases='', description='cancer/testis antigen family 45, member A8', other_designations='cancer/testis antigen 45A8', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135846497, end_position_on_the_genomic_accession=135863879, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723739, CurrentID=0, Status='live', Symbol='LOC102723739', Aliases='', description='uncharacterized LOC102723739', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38184360, end_position_on_the_genomic_accession=38239635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723740, CurrentID=0, Status='live', Symbol='LOC102723740', Aliases='', description='uncharacterized LOC102723740', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41284750, end_position_on_the_genomic_accession=41294928, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723741, CurrentID=0, Status='live', Symbol='LOC102723741', Aliases='', description='uncharacterized LOC102723741', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132422664, end_position_on_the_genomic_accession=132432102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723742, CurrentID=0, Status='live', Symbol='LINC01500', Aliases='', description='long intergenic non-protein coding RNA 1500', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58828288, end_position_on_the_genomic_accession=59017328, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723743, CurrentID=0, Status='live', Symbol='LOC102723743', Aliases='', description='E3 ubiquitin-protein ligase RNF12-B-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=846962, end_position_on_the_genomic_accession=848411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723746, CurrentID=0, Status='live', Symbol='LOC102723746', Aliases='', description='uncharacterized LOC102723746', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93286635, end_position_on_the_genomic_accession=93385970, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375771, CurrentID=0, Status='live', Symbol='LOC105375771', Aliases='', description='uncharacterized LOC105375771', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133361735, end_position_on_the_genomic_accession=133377649, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723748, CurrentID=0, Status='live', Symbol='LOC102723748', Aliases='', description='40S ribosomal protein SA-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723749, CurrentID=0, Status='live', Symbol='LOC102723749', Aliases='', description='uncharacterized LOC102723749', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24340746, end_position_on_the_genomic_accession=24343720, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723750, CurrentID=0, Status='live', Symbol='LOC102723750', Aliases='', description='uncharacterized LOC102723750', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74598919, end_position_on_the_genomic_accession=74615079, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  343409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343409, CurrentID=0, Status='live', Symbol='OR10J8P', Aliases='', description='olfactory receptor family 10 subfamily J member 8 pseudogene', other_designations='olfactory receptor, family 10, subfamily J, member 8 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159366161, end_position_on_the_genomic_accession=159367086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723752, CurrentID=0, Status='live', Symbol='LOC102723752', Aliases='', description='uncharacterized LOC102723752', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41868671, end_position_on_the_genomic_accession=41880167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723753, CurrentID=0, Status='live', Symbol='LOC102723753', Aliases='', description='putative HERC2-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723754, CurrentID=0, Status='live', Symbol='LOC102723754', Aliases='', description='uncharacterized LOC102723754', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22509553, end_position_on_the_genomic_accession=22523186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723755, CurrentID=0, Status='live', Symbol='LOC102723755', Aliases='', description='uncharacterized LOC102723755', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2254409, end_position_on_the_genomic_accession=2255148, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928679, CurrentID=0, Status='live', Symbol='LOC101928679', Aliases='', description='uncharacterized LOC101928679', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181260864, end_position_on_the_genomic_accession=181265093, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723757, CurrentID=0, Status='live', Symbol='LOC102723757', Aliases='', description='uncharacterized LOC102723757', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46372783, end_position_on_the_genomic_accession=46373778, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723758, CurrentID=0, Status='live', Symbol='LOC102723758', Aliases='', description='uncharacterized LOC102723758', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1124102, end_position_on_the_genomic_accession=1130589, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375773, CurrentID=0, Status='live', Symbol='LOC105375773', Aliases='', description='uncharacterized LOC105375773', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133663761, end_position_on_the_genomic_accession=133684096, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723760, CurrentID=0, Status='live', Symbol='LOC102723760', Aliases='', description='uncharacterized LOC102723760', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27157556, end_position_on_the_genomic_accession=27160453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723761, CurrentID=0, Status='live', Symbol='LOC102723761', Aliases='', description='uncharacterized LOC102723761', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212632369, end_position_on_the_genomic_accession=212665649, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723763, CurrentID=0, Status='live', Symbol='LOC102723763', Aliases='', description='uncharacterized LOC102723763', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2761384, end_position_on_the_genomic_accession=2775232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723764, CurrentID=0, Status='live', Symbol='LOC102723764', Aliases='', description='uncharacterized LOC102723764', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28996517, end_position_on_the_genomic_accession=29017255, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102723765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723765, CurrentID=0, Status='live', Symbol='LOC102723765', Aliases='', description='uncharacterized LOC102723765', other_designations='CTD-2531D15.4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62410575, end_position_on_the_genomic_accession=62427799, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723766, CurrentID=0, Status='live', Symbol='LOC102723766', Aliases='', description='uncharacterized LOC102723766', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184365561, end_position_on_the_genomic_accession=184382306, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723768, CurrentID=0, Status='live', Symbol='LOC102723768', Aliases='', description='uncharacterized LOC102723768', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42314897, end_position_on_the_genomic_accession=42374443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723769, CurrentID=0, Status='live', Symbol='LOC102723769', Aliases='', description='uncharacterized LOC102723769', other_designations='', map_location='22p11', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11897406, end_position_on_the_genomic_accession=11956534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375775, CurrentID=0, Status='live', Symbol='LOC105375775', Aliases='', description='uncharacterized LOC105375775', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=136237261, end_position_on_the_genomic_accession=136238931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723775, CurrentID=0, Status='live', Symbol='LINC01315', Aliases='lnc-C22orf32-1', description='long intergenic non-protein coding RNA 1315', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42364400, end_position_on_the_genomic_accession=42369208, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723776, CurrentID=0, Status='live', Symbol='LOC102723776', Aliases='', description='protein FAM182A-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723777, CurrentID=0, Status='live', Symbol='LOC102723777', Aliases='', description='uncharacterized LOC102723777', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65380, end_position_on_the_genomic_accession=68072, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723778, CurrentID=0, Status='live', Symbol='LOC102723778', Aliases='', description='uncharacterized LOC102723778', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=31171144, end_position_on_the_genomic_accession=31211675, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723779, CurrentID=0, Status='live', Symbol='LOC102723779', Aliases='', description='uncharacterized LOC102723779', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184474802, end_position_on_the_genomic_accession=184489551, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723780, CurrentID=0, Status='live', Symbol='LOC102723780', Aliases='', description='protein FRG1-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10936830, end_position_on_the_genomic_accession=10961529, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  102723782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723782, CurrentID=0, Status='live', Symbol='LOC102723782', Aliases='', description='uncharacterized LOC102723782', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87168767, end_position_on_the_genomic_accession=87172649, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375777, CurrentID=0, Status='live', Symbol='LOC105375777', Aliases='', description='uncharacterized LOC105375777', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=139702853, end_position_on_the_genomic_accession=139707127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723784, CurrentID=0, Status='live', Symbol='LOC102723784', Aliases='', description='uncharacterized LOC102723784', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101296753, end_position_on_the_genomic_accession=101309187, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723786, CurrentID=0, Status='live', Symbol='LOC102723786', Aliases='', description='uncharacterized LOC102723786', other_designations='CTD-2520I13.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71288808, end_position_on_the_genomic_accession=71309963, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723788, CurrentID=0, Status='live', Symbol='LOC102723788', Aliases='', description='uncharacterized LOC102723788', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63370753, end_position_on_the_genomic_accession=63378135, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723789, CurrentID=0, Status='live', Symbol='LOC102723789', Aliases='', description='uncharacterized LOC102723789', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39934569, end_position_on_the_genomic_accession=39954301, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723791, CurrentID=0, Status='live', Symbol='LOC102723791', Aliases='', description='uncharacterized LOC102723791', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77252133, end_position_on_the_genomic_accession=77278397, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723794, CurrentID=0, Status='live', Symbol='LOC102723794', Aliases='', description='uncharacterized LOC102723794', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87430292, end_position_on_the_genomic_accession=87499379, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375779, CurrentID=0, Status='live', Symbol='LOC105375779', Aliases='', description='uncharacterized LOC105375779', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140262853, end_position_on_the_genomic_accession=140282281, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723796, CurrentID=0, Status='live', Symbol='LOC102723796', Aliases='', description='transducin-like enhancer protein 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71429996, end_position_on_the_genomic_accession=71442046, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  102723798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723798, CurrentID=0, Status='live', Symbol='LOC102723798', Aliases='', description='uncharacterized LOC102723798', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1162445, end_position_on_the_genomic_accession=1168687, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723799, CurrentID=0, Status='live', Symbol='LOC102723799', Aliases='', description='uncharacterized LOC102723799', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723802, CurrentID=0, Status='live', Symbol='LOC102723802', Aliases='', description='uncharacterized LOC102723802', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214256636, end_position_on_the_genomic_accession=214279385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723803, CurrentID=0, Status='live', Symbol='LOC102723803', Aliases='', description='uncharacterized LOC102723803', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1298290, end_position_on_the_genomic_accession=1489023, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  102723805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723805, CurrentID=0, Status='live', Symbol='LOC102723805', Aliases='', description='uncharacterized LOC102723805', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723806, CurrentID=0, Status='live', Symbol='LOC102723806', Aliases='', description='uncharacterized LOC102723806', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82395800, end_position_on_the_genomic_accession=82398189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723809, CurrentID=0, Status='live', Symbol='LOC102723809', Aliases='', description='uncharacterized LOC102723809', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64514151, end_position_on_the_genomic_accession=64540368, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723811, CurrentID=0, Status='live', Symbol='LOC102723811', Aliases='', description='uncharacterized LOC102723811', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1246268, end_position_on_the_genomic_accession=1248504, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723814, CurrentID=0, Status='live', Symbol='LOC102723814', Aliases='', description='uncharacterized LOC102723814', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63517262, end_position_on_the_genomic_accession=63519417, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723815, CurrentID=0, Status='live', Symbol='LOC102723815', Aliases='', description='disintegrin and metalloproteinase domain-containing protein 29-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723817, CurrentID=0, Status='live', Symbol='LOC102723817', Aliases='', description='uncharacterized LOC102723817', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224610608, end_position_on_the_genomic_accession=224616279, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723818, CurrentID=0, Status='live', Symbol='LINC01248', Aliases='', description='long intergenic non-protein coding RNA 1248', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5634141, end_position_on_the_genomic_accession=5691118, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927787, CurrentID=0, Status='live', Symbol='LOC101927787', Aliases='', description='uncharacterized LOC101927787', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234646289, end_position_on_the_genomic_accession=234660924, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723822, CurrentID=0, Status='live', Symbol='LOC102723822', Aliases='', description='nucleolar GTP-binding protein 1-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723824, CurrentID=0, Status='live', Symbol='LOC102723824', Aliases='', description='uncharacterized LOC102723824', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42142435, end_position_on_the_genomic_accession=42170231, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723825, CurrentID=0, Status='live', Symbol='LOC102723825', Aliases='', description='uncharacterized LOC102723825', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240384284, end_position_on_the_genomic_accession=240391219, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723828, CurrentID=0, Status='live', Symbol='LOC102723828', Aliases='', description='uncharacterized LOC102723828', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=31997379, end_position_on_the_genomic_accession=32155406, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723831, CurrentID=0, Status='live', Symbol='LOC102723831', Aliases='', description='uncharacterized LOC102723831', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151196673, end_position_on_the_genomic_accession=151228457, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723832, CurrentID=0, Status='live', Symbol='LOC102723832', Aliases='', description='uncharacterized LOC102723832', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30860590, end_position_on_the_genomic_accession=30862399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723833, CurrentID=0, Status='live', Symbol='LOC102723833', Aliases='', description='uncharacterized LOC102723833', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=216072465, end_position_on_the_genomic_accession=216086917, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723834, CurrentID=0, Status='live', Symbol='LOC102723834', Aliases='', description='uncharacterized LOC102723834', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225700693, end_position_on_the_genomic_accession=225716434, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375786, CurrentID=0, Status='live', Symbol='LOC105375786', Aliases='', description='uncharacterized LOC105375786', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141288852, end_position_on_the_genomic_accession=141290819, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723838, CurrentID=0, Status='live', Symbol='LOC102723838', Aliases='', description='uncharacterized LOC102723838', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102493780, end_position_on_the_genomic_accession=102498801, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723839, CurrentID=0, Status='live', Symbol='BRCAT107', Aliases='', description='breast cancer-associated transcript 107', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=44495144, end_position_on_the_genomic_accession=44687182, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723840, CurrentID=0, Status='live', Symbol='LOC102723840', Aliases='', description='uncharacterized LOC102723840', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=430265, end_position_on_the_genomic_accession=470745, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102723841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723841, CurrentID=0, Status='live', Symbol='LOC102723841', Aliases='', description='uncharacterized LOC102723841', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723842, CurrentID=0, Status='live', Symbol='LOC102723842', Aliases='', description='uncharacterized LOC102723842', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57866002, end_position_on_the_genomic_accession=57868127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723846, CurrentID=0, Status='live', Symbol='LOC102723846', Aliases='', description='uncharacterized LOC102723846', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=32225079, end_position_on_the_genomic_accession=32238547, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723847, CurrentID=0, Status='live', Symbol='LOC102723847', Aliases='', description='uncharacterized LOC102723847', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723849, CurrentID=0, Status='live', Symbol='SPDYE17', Aliases='', description='speedy/RINGO cell cycle regulator family member E17', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77023688, end_position_on_the_genomic_accession=77031631, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723850, CurrentID=0, Status='live', Symbol='LOC102723850', Aliases='', description='uncharacterized LOC102723850', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73452882, end_position_on_the_genomic_accession=73464022, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723851, CurrentID=0, Status='live', Symbol='LOC102723851', Aliases='', description='uncharacterized LOC102723851', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38127033, end_position_on_the_genomic_accession=38128066, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723852, CurrentID=0, Status='live', Symbol='TPT1P7', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23811141, end_position_on_the_genomic_accession=23811572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723854, CurrentID=0, Status='live', Symbol='LOC102723854', Aliases='', description='uncharacterized LOC102723854', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43027853, end_position_on_the_genomic_accession=43039543, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375789, CurrentID=0, Status='live', Symbol='LOC105375789', Aliases='', description='uncharacterized LOC105375789', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141446707, end_position_on_the_genomic_accession=141464934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928973, CurrentID=0, Status='live', Symbol='LOC101928973', Aliases='', description='uncharacterized LOC101928973', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181174484, end_position_on_the_genomic_accession=181182208, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723859, CurrentID=0, Status='live', Symbol='TBC1D3E', Aliases='TBC1DE3', description='TBC1 domain family member 3E', other_designations='TBC1 domain family member-like|TBC1 domain family, member 3E', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38119293, end_position_on_the_genomic_accession=38138868, orientation='plus', exon_count=18, OMIM=610808, no_hdr0=''),
  105375790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375790, CurrentID=0, Status='live', Symbol='LOC105375790', Aliases='', description='uncharacterized LOC105375790', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141388118, end_position_on_the_genomic_accession=141391805, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723862, CurrentID=0, Status='live', Symbol='LOC102723862', Aliases='', description='uncharacterized LOC102723862', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103838797, end_position_on_the_genomic_accession=103849540, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723864, CurrentID=0, Status='live', Symbol='LOC102723864', Aliases='', description='uncharacterized LOC102723864', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89634146, end_position_on_the_genomic_accession=89636340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132911, CurrentID=0, Status='live', Symbol='DPH3P1', Aliases='C20orf143, DPH3B, ZCSL1', description='diphthamide biosynthesis 3 pseudogene 1', other_designations='CSL-type zinc finger-containing protein 1|DPH3 homolog B (KTI11, S. cerevisiae)|zinc finger, CSL domain containing 1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62844566, end_position_on_the_genomic_accession=62846191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723869, CurrentID=0, Status='live', Symbol='LOC102723869', Aliases='', description='uncharacterized LOC102723869', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723870, CurrentID=0, Status='live', Symbol='LOC102723870', Aliases='', description='uncharacterized LOC102723870', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32609298, end_position_on_the_genomic_accession=32619941, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723872, CurrentID=0, Status='live', Symbol='LOC102723872', Aliases='', description='putative protein FAM157B', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12842, end_position_on_the_genomic_accession=31446, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  649264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649264, CurrentID=0, Status='live', Symbol='CES5AP1', Aliases='', description='carboxylesterase 5A pseudogene 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23359606, end_position_on_the_genomic_accession=23382126, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723875, CurrentID=0, Status='live', Symbol='LINC01065', Aliases='', description='long intergenic non-protein coding RNA 1065', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53131746, end_position_on_the_genomic_accession=53151900, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723876, CurrentID=0, Status='live', Symbol='LOC102723876', Aliases='', description='uncharacterized LOC102723876', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73764154, end_position_on_the_genomic_accession=73775520, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723878, CurrentID=0, Status='live', Symbol='LOC102723878', Aliases='', description='uncharacterized LOC102723878', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64325050, end_position_on_the_genomic_accession=64340625, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723879, CurrentID=0, Status='live', Symbol='LOC102723879', Aliases='', description='uncharacterized LOC102723879', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104194445, end_position_on_the_genomic_accession=104358989, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101937451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101937451, CurrentID=0, Status='live', Symbol='LINC01030', Aliases='', description='long intergenic non-protein coding RNA 1030', other_designations='TCONS_00015085', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90592775, end_position_on_the_genomic_accession=90606063, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723885, CurrentID=0, Status='live', Symbol='LOC102723885', Aliases='', description='uncharacterized LOC102723885', other_designations='', map_location='7q21.12', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88216660, end_position_on_the_genomic_accession=88229768, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723886, CurrentID=0, Status='live', Symbol='LOC102723886', Aliases='', description='uncharacterized LOC102723886', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219409459, end_position_on_the_genomic_accession=219557280, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102723888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723888, CurrentID=0, Status='live', Symbol='LOC102723888', Aliases='', description='uncharacterized LOC102723888', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186416044, end_position_on_the_genomic_accession=186487637, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102723890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723890, CurrentID=0, Status='live', Symbol='LOC102723890', Aliases='', description='uncharacterized LOC102723890', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375795, CurrentID=0, Status='live', Symbol='LOC105375795', Aliases='', description='uncharacterized protein FLJ40521-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143043082, end_position_on_the_genomic_accession=143073956, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723895, CurrentID=0, Status='live', Symbol='LOC102723895', Aliases='', description='uncharacterized LOC102723895', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104568489, end_position_on_the_genomic_accession=104609302, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723896, CurrentID=0, Status='live', Symbol='LOC102723896', Aliases='', description='uncharacterized LOC102723896', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112880305, end_position_on_the_genomic_accession=112886201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375796, CurrentID=0, Status='live', Symbol='LOC105375796', Aliases='', description='uncharacterized LOC105375796', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143551963, end_position_on_the_genomic_accession=143553407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723899, CurrentID=0, Status='live', Symbol='LOC102723899', Aliases='', description='uncharacterized LOC102723899', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90509145, end_position_on_the_genomic_accession=90513397, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375797, CurrentID=0, Status='live', Symbol='LOC105375797', Aliases='', description='uncharacterized LOC105375797', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143592466, end_position_on_the_genomic_accession=143594965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723904, CurrentID=0, Status='live', Symbol='LOC102723904', Aliases='', description='uncharacterized LOC102723904', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44161184, end_position_on_the_genomic_accession=44168467, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723906, CurrentID=0, Status='live', Symbol='LOC102723906', Aliases='', description='uncharacterized LOC102723906', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186840663, end_position_on_the_genomic_accession=187061215, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723908, CurrentID=0, Status='live', Symbol='LOC102723908', Aliases='', description='uncharacterized LOC102723908', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375798, CurrentID=0, Status='live', Symbol='LOC105375798', Aliases='', description='uncharacterized LOC105375798', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143620439, end_position_on_the_genomic_accession=143635068, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723910, CurrentID=0, Status='live', Symbol='LOC102723910', Aliases='', description='uncharacterized LOC102723910', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61178641, end_position_on_the_genomic_accession=61186738, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102723911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723911, CurrentID=0, Status='live', Symbol='LOC102723911', Aliases='', description='uncharacterized LOC102723911', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69031043, end_position_on_the_genomic_accession=69040276, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723912, CurrentID=0, Status='live', Symbol='LOC102723912', Aliases='', description='sodium/hydrogen exchanger 9B1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723913, CurrentID=0, Status='live', Symbol='LOC102723913', Aliases='', description='uncharacterized LOC102723913', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83119828, end_position_on_the_genomic_accession=83123431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723914, CurrentID=0, Status='live', Symbol='LOC102723914', Aliases='', description='uncharacterized LOC102723914', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117815020, end_position_on_the_genomic_accession=117828156, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723915, CurrentID=0, Status='live', Symbol='LOC102723915', Aliases='', description='uncharacterized LOC102723915', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141618260, end_position_on_the_genomic_accession=141625965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723917, CurrentID=0, Status='live', Symbol='LOC102723917', Aliases='', description='uncharacterized LOC102723917', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=268133, end_position_on_the_genomic_accession=297530, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723919, CurrentID=0, Status='live', Symbol='LOC102723919', Aliases='', description='uncharacterized LOC102723919', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113131186, end_position_on_the_genomic_accession=113132260, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723920, CurrentID=0, Status='live', Symbol='LOC102723920', Aliases='', description='uncharacterized LOC102723920', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187408776, end_position_on_the_genomic_accession=187415697, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375800, CurrentID=0, Status='live', Symbol='LOC105375800', Aliases='', description='UPF0454 protein C12orf49 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143757749, end_position_on_the_genomic_accession=143760508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723922, CurrentID=0, Status='live', Symbol='LOC102723922', Aliases='', description='uncharacterized LOC102723922', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=266184, end_position_on_the_genomic_accession=292170, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723924, CurrentID=0, Status='live', Symbol='LOC102723924', Aliases='', description='uncharacterized LOC102723924', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41242335, end_position_on_the_genomic_accession=41334558, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  102723927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723927, CurrentID=0, Status='live', Symbol='LOC102723927', Aliases='', description='uncharacterized LOC102723927', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241970683, end_position_on_the_genomic_accession=241977276, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723928, CurrentID=0, Status='live', Symbol='LOC102723928', Aliases='', description='uncharacterized LOC102723928', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118635788, end_position_on_the_genomic_accession=118679805, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723930, CurrentID=0, Status='live', Symbol='LOC102723930', Aliases='', description='uncharacterized LOC102723930', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81295, CurrentID=0, Status='live', Symbol='OR51A10P', Aliases='OR51A11P, OR51A13', description='olfactory receptor family 51 subfamily A member 10 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 10 pseudogene|olfactory receptor, family 51, subfamily A, member 11 pseudogene|olfactory receptor, family 51, subfamily A, member 13', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5468504, end_position_on_the_genomic_accession=5469359, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723932, CurrentID=0, Status='live', Symbol='LOC102723932', Aliases='', description='uncharacterized LOC102723932', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79517873, end_position_on_the_genomic_accession=79567765, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375802, CurrentID=0, Status='live', Symbol='LOC105375802', Aliases='', description='uncharacterized LOC105375802', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144497357, end_position_on_the_genomic_accession=144505078, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723934, CurrentID=0, Status='live', Symbol='LOC102723934', Aliases='', description='BCL-6 corepressor pseudogene', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19481616, end_position_on_the_genomic_accession=19483933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723935, CurrentID=0, Status='live', Symbol='LOC102723935', Aliases='', description='uncharacterized LOC102723935', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229242988, end_position_on_the_genomic_accession=229256698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102723936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723936, CurrentID=0, Status='live', Symbol='LOC102723936', Aliases='', description='acidic phospholipase A2 PA-1G-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375803, CurrentID=0, Status='live', Symbol='LOC105375803', Aliases='', description='uncharacterized LOC105375803', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144494204, end_position_on_the_genomic_accession=144496461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723943, CurrentID=0, Status='live', Symbol='LOC102723943', Aliases='', description='uncharacterized LOC102723943', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38192649, end_position_on_the_genomic_accession=38193477, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723944, CurrentID=0, Status='live', Symbol='LOC102723944', Aliases='', description='uncharacterized LOC102723944', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1453286, end_position_on_the_genomic_accession=1555226, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105375804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375804, CurrentID=0, Status='live', Symbol='LOC105375804', Aliases='', description='uncharacterized LOC105375804', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144711850, end_position_on_the_genomic_accession=144714845, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723948, CurrentID=0, Status='live', Symbol='LOC102723948', Aliases='', description='uncharacterized LOC102723948', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135162461, end_position_on_the_genomic_accession=135167123, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375805, CurrentID=0, Status='live', Symbol='LOC105375805', Aliases='', description='uncharacterized LOC105375805', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144765393, end_position_on_the_genomic_accession=144770252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723952, CurrentID=0, Status='live', Symbol='LOC102723952', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=122010971, end_position_on_the_genomic_accession=122030384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723955, CurrentID=0, Status='live', Symbol='LOC102723955', Aliases='', description='neuroblastoma breakpoint family member 6-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149719896, end_position_on_the_genomic_accession=149749184, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  102723956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723956, CurrentID=0, Status='live', Symbol='LOC102723956', Aliases='', description='uncharacterized LOC102723956', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119011173, end_position_on_the_genomic_accession=119026619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723961, CurrentID=0, Status='live', Symbol='LOC102723961', Aliases='', description='uncharacterized LOC102723961', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79815944, end_position_on_the_genomic_accession=79823284, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102723963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723963, CurrentID=0, Status='live', Symbol='LOC102723963', Aliases='', description='uncharacterized LOC102723963', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49976931, end_position_on_the_genomic_accession=49978419, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723964, CurrentID=0, Status='live', Symbol='LOC102723964', Aliases='', description='alpha/beta hydrolase domain-containing protein 17A-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723965, CurrentID=0, Status='live', Symbol='LOC102723965', Aliases='', description='uncharacterized LOC102723965', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6793859, end_position_on_the_genomic_accession=6797028, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723966, CurrentID=0, Status='live', Symbol='LOC102723966', Aliases='', description='uncharacterized LOC102723966', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111097272, end_position_on_the_genomic_accession=111106961, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723967, CurrentID=0, Status='live', Symbol='LOC102723967', Aliases='', description='uncharacterized LOC102723967', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119100132, end_position_on_the_genomic_accession=119133575, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102723968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723968, CurrentID=0, Status='live', Symbol='LOC102723968', Aliases='', description='uncharacterized LOC102723968', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63828840, end_position_on_the_genomic_accession=63844125, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102723971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723971, CurrentID=0, Status='live', Symbol='LOC102723971', Aliases='', description='beta-lactoglobulin-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135614222, end_position_on_the_genomic_accession=135619893, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102723973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723973, CurrentID=0, Status='live', Symbol='LOC102723973', Aliases='', description='uncharacterized LOC102723973', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189779352, end_position_on_the_genomic_accession=189787641, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375809, CurrentID=0, Status='live', Symbol='LOC105375809', Aliases='', description='platelet endothelial aggregation receptor 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102723977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723977, CurrentID=0, Status='live', Symbol='LOC102723977', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201823832, end_position_on_the_genomic_accession=201824859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102723981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723981, CurrentID=0, Status='live', Symbol='RPL23AP84', Aliases='', description='ribosomal protein L23a pseudogene 84', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190184881, end_position_on_the_genomic_accession=190195825, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  285389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285389, CurrentID=0, Status='live', Symbol='LOC285389', Aliases='', description='uncharacterized LOC285389', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194005207, end_position_on_the_genomic_accession=194071044, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  102723983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723983, CurrentID=0, Status='live', Symbol='LOC102723983', Aliases='', description='uncharacterized LOC102723983', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51270256, end_position_on_the_genomic_accession=51271362, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371399, CurrentID=0, Status='live', Symbol='LOC105371399', Aliases='', description='uncharacterized LOC105371399', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87766256, end_position_on_the_genomic_accession=87775382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102723985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723985, CurrentID=0, Status='live', Symbol='LOC102723985', Aliases='', description='uncharacterized LOC102723985', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81614940, end_position_on_the_genomic_accession=81663201, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375811, CurrentID=0, Status='live', Symbol='LOC105375811', Aliases='', description='uncharacterized LOC105375811', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46641340, end_position_on_the_genomic_accession=46656765, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372558, CurrentID=0, Status='live', Symbol='LOC105372558', Aliases='', description='uncharacterized LOC105372558', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21397337, end_position_on_the_genomic_accession=21441456, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102723991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723991, CurrentID=0, Status='live', Symbol='LOC102723991', Aliases='', description='uncharacterized LOC102723991', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22467198, end_position_on_the_genomic_accession=22475287, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  225689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=225689, CurrentID=0, Status='live', Symbol='MAPK15', Aliases='ERK7, ERK8', description='mitogen-activated protein kinase 15', other_designations='ERK-7|ERK-8|MAP kinase 15|MAPK 15|extracellular regulated kinase 8 delta|extracellular signal regulated kinase 8|extracellular signal-regulated kinase 7|extracellular signal-regulated kinase 8', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143716302, end_position_on_the_genomic_accession=143722463, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  102723994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723994, CurrentID=0, Status='live', Symbol='LOC102723994', Aliases='', description='uncharacterized LOC102723994', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7343891, end_position_on_the_genomic_accession=7346337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102723996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723996, CurrentID=0, Status='live', Symbol='LOC102723996', Aliases='', description='ICOS ligand', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5022044, end_position_on_the_genomic_accession=5046692, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375813, CurrentID=0, Status='live', Symbol='LOC105375813', Aliases='', description='uncharacterized LOC105375813', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46876957, end_position_on_the_genomic_accession=46880014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724000, CurrentID=0, Status='live', Symbol='LINC01626', Aliases='', description='long intergenic non-protein coding RNA 1626', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71421027, end_position_on_the_genomic_accession=71458872, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724001, CurrentID=0, Status='live', Symbol='LOC102724001', Aliases='', description='uncharacterized LOC102724001', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81914890, end_position_on_the_genomic_accession=81967179, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724002, CurrentID=0, Status='live', Symbol='LOC102724002', Aliases='', description='uncharacterized LOC102724002', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373431, CurrentID=0, Status='live', Symbol='LOC105373431', Aliases='', description='uncharacterized LOC105373431', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12445240, end_position_on_the_genomic_accession=12463279, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724004, CurrentID=0, Status='live', Symbol='LOC102724004', Aliases='', description='uncharacterized LOC102724004', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25516913, end_position_on_the_genomic_accession=25527799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724006, CurrentID=0, Status='live', Symbol='LOC102724006', Aliases='', description='uncharacterized LOC102724006', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99474559, end_position_on_the_genomic_accession=99479725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724008, CurrentID=0, Status='live', Symbol='LOC102724008', Aliases='', description='uncharacterized LOC102724008', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48253473, end_position_on_the_genomic_accession=48286444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724009, CurrentID=0, Status='live', Symbol='LOC102724009', Aliases='', description='uncharacterized LOC102724009', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5019214, end_position_on_the_genomic_accession=5020093, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129611, CurrentID=0, Status='live', Symbol='PRAMEF31P', Aliases='', description='PRAME family member 31, pseudogene', other_designations='PRAME family member 17 pseudogene', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13245837, end_position_on_the_genomic_accession=13248814, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130299, CurrentID=0, Status='live', Symbol='LOC100130299', Aliases='', description='isopentenyl-diphosphate delta isomerase 1 pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48294027, end_position_on_the_genomic_accession=48294789, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724012, CurrentID=0, Status='live', Symbol='LOC102724012', Aliases='', description='uncharacterized LOC102724012', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150629700, end_position_on_the_genomic_accession=150643018, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724014, CurrentID=0, Status='live', Symbol='LOC102724014', Aliases='', description='spidroin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724015, CurrentID=0, Status='live', Symbol='LOC102724015', Aliases='', description='uncharacterized LOC102724015', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71292715, end_position_on_the_genomic_accession=71322339, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375816, CurrentID=0, Status='live', Symbol='LOC105375816', Aliases='', description='putative coiled-coil domain-containing protein 144C', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47088584, end_position_on_the_genomic_accession=47092604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724018, CurrentID=0, Status='live', Symbol='LOC102724018', Aliases='', description='uncharacterized LOC102724018', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31562665, end_position_on_the_genomic_accession=31596887, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724019, CurrentID=0, Status='live', Symbol='LOC102724019', Aliases='', description='uncharacterized LOC102724019', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135139826, end_position_on_the_genomic_accession=135147801, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724020, CurrentID=0, Status='live', Symbol='LOC102724020', Aliases='', description='uncharacterized LOC102724020', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10015352, end_position_on_the_genomic_accession=10030606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724023, CurrentID=0, Status='live', Symbol='LOC102724023', Aliases='', description='ES1 protein homolog, mitochondrial', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5116341, end_position_on_the_genomic_accession=5128451, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724025, CurrentID=0, Status='live', Symbol='LOC102724025', Aliases='', description='60S ribosomal protein L29-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100187453, end_position_on_the_genomic_accession=100188255, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724027, CurrentID=0, Status='live', Symbol='LOC102724027', Aliases='', description='uncharacterized LOC102724027', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11143880, end_position_on_the_genomic_accession=11195441, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375765, CurrentID=0, Status='live', Symbol='LOC105375765', Aliases='', description='uncharacterized LOC105375765', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132487345, end_position_on_the_genomic_accession=132503418, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375818, CurrentID=0, Status='live', Symbol='LOC105375818', Aliases='', description='uncharacterized LOC105375818', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47941435, end_position_on_the_genomic_accession=47945779, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724030, CurrentID=0, Status='live', Symbol='LOC102724030', Aliases='', description='uncharacterized LOC102724030', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54150219, end_position_on_the_genomic_accession=54168595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724031, CurrentID=0, Status='live', Symbol='LOC102724031', Aliases='', description='ribosome biogenesis protein BMS1 homolog', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724034, CurrentID=0, Status='live', Symbol='LOC102724034', Aliases='', description='uncharacterized LOC102724034', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82443332, end_position_on_the_genomic_accession=82444869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724035, CurrentID=0, Status='live', Symbol='LOC102724035', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 5-like', map_location='8q22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104565963, end_position_on_the_genomic_accession=104567105, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724036, CurrentID=0, Status='live', Symbol='LOC102724036', Aliases='', description='uncharacterized LOC102724036', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84409801, end_position_on_the_genomic_accession=84656485, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  102724039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724039, CurrentID=0, Status='live', Symbol='LOC102724039', Aliases='', description='uncharacterized LOC102724039', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16276917, end_position_on_the_genomic_accession=16295772, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724040, CurrentID=0, Status='live', Symbol='LOC102724040', Aliases='', description='uncharacterized LOC102724040', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139583466, end_position_on_the_genomic_accession=139600368, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132940, CurrentID=0, Status='live', Symbol='PRELID3BP6', Aliases='', description='PRELI domain containing 3B pseudogene 6', other_designations='slowmo homolog 2 (Drosophila) pseudogene', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42917414, end_position_on_the_genomic_accession=42917973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724042, CurrentID=0, Status='live', Symbol='LOC102724042', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63486564, end_position_on_the_genomic_accession=63491940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371401, CurrentID=0, Status='live', Symbol='LOC105371401', Aliases='', description='uncharacterized LOC105371401', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88177298, end_position_on_the_genomic_accession=88186920, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724046, CurrentID=0, Status='live', Symbol='LOC102724046', Aliases='', description='uncharacterized LOC102724046', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223092928, end_position_on_the_genomic_accession=223103308, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724047, CurrentID=0, Status='live', Symbol='LOC102724047', Aliases='', description='uncharacterized LOC102724047', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78230187, end_position_on_the_genomic_accession=78241459, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724048, CurrentID=0, Status='live', Symbol='LOC102724048', Aliases='', description='uncharacterized LOC102724048', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34543796, end_position_on_the_genomic_accession=34563139, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724050, CurrentID=0, Status='live', Symbol='LOC102724050', Aliases='', description='uncharacterized LOC102724050', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54353691, end_position_on_the_genomic_accession=54467030, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724052, CurrentID=0, Status='live', Symbol='LOC102724052', Aliases='', description='spidroin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375822, CurrentID=0, Status='live', Symbol='LOC105375822', Aliases='', description='uncharacterized LOC105375822', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48515852, end_position_on_the_genomic_accession=48518036, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724055, CurrentID=0, Status='live', Symbol='LOC102724055', Aliases='', description='uncharacterized LOC102724055', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30322577, end_position_on_the_genomic_accession=30365882, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724057, CurrentID=0, Status='live', Symbol='LOC102724057', Aliases='', description='uncharacterized LOC102724057', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85756136, end_position_on_the_genomic_accession=85805341, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  102724058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724058, CurrentID=0, Status='live', Symbol='LOC102724058', Aliases='', description='uncharacterized LOC102724058', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165957418, end_position_on_the_genomic_accession=166036400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724062, CurrentID=0, Status='live', Symbol='LOC102724062', Aliases='', description='protein FAM153B-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32600341, end_position_on_the_genomic_accession=32637511, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  102724063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724063, CurrentID=0, Status='live', Symbol='LOC102724063', Aliases='', description='uncharacterized LOC102724063', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54417489, end_position_on_the_genomic_accession=54433863, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724064, CurrentID=0, Status='live', Symbol='LOC102724064', Aliases='', description='proline-rich receptor-like protein kinase PERK13', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66347544, end_position_on_the_genomic_accession=66363764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724066, CurrentID=0, Status='live', Symbol='LOC102724066', Aliases='', description='uncharacterized LOC102724066', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724068, CurrentID=0, Status='live', Symbol='LOC102724068', Aliases='', description='uncharacterized LOC102724068', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140581877, end_position_on_the_genomic_accession=140677957, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724069, CurrentID=0, Status='live', Symbol='LOC102724069', Aliases='', description='uncharacterized LOC102724069', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119552585, end_position_on_the_genomic_accession=119574167, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724070, CurrentID=0, Status='live', Symbol='LOC102724070', Aliases='', description='NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 5 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96463969, end_position_on_the_genomic_accession=96494800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724072, CurrentID=0, Status='live', Symbol='LOC102724072', Aliases='', description='uncharacterized LOC102724072', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54433945, end_position_on_the_genomic_accession=54447396, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371402, CurrentID=0, Status='live', Symbol='LOC105371402', Aliases='', description='uncharacterized LOC105371402', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724076, CurrentID=0, Status='live', Symbol='LINC01038', Aliases='', description='long intergenic non-protein coding RNA 1038', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79804386, end_position_on_the_genomic_accession=79805682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724077, CurrentID=0, Status='live', Symbol='LINC01389', Aliases='', description='long intergenic non-protein coding RNA 1389', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47380796, end_position_on_the_genomic_accession=47437317, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724078, CurrentID=0, Status='live', Symbol='LOC102724078', Aliases='', description='uncharacterized LOC102724078', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32164871, end_position_on_the_genomic_accession=32207981, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724080, CurrentID=0, Status='live', Symbol='LOC102724080', Aliases='', description='uncharacterized LOC102724080', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86414220, end_position_on_the_genomic_accession=86452686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81320, CurrentID=0, Status='live', Symbol='OR4A3P', Aliases='', description='olfactory receptor family 4 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 3 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54662581, end_position_on_the_genomic_accession=54663451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724082, CurrentID=0, Status='live', Symbol='LOC102724082', Aliases='', description='uncharacterized LOC102724082', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19439493, end_position_on_the_genomic_accession=19457598, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375827, CurrentID=0, Status='live', Symbol='LOC105375827', Aliases='', description='uncharacterized LOC105375827', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49224972, end_position_on_the_genomic_accession=49229278, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724084, CurrentID=0, Status='live', Symbol='LOC102724084', Aliases='', description='uncharacterized LOC102724084', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80155958, end_position_on_the_genomic_accession=80563135, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724086, CurrentID=0, Status='live', Symbol='LOC102724086', Aliases='', description='uncharacterized LOC102724086', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90177446, end_position_on_the_genomic_accession=90186789, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724087, CurrentID=0, Status='live', Symbol='LOC102724087', Aliases='', description='uncharacterized LOC102724087', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160926322, end_position_on_the_genomic_accession=160981614, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375828, CurrentID=0, Status='live', Symbol='LOC105375828', Aliases='', description='uncharacterized LOC105375828', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50364446, end_position_on_the_genomic_accession=50377000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724090, CurrentID=0, Status='live', Symbol='LOC102724090', Aliases='', description='uncharacterized LOC102724090', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86948028, end_position_on_the_genomic_accession=86948519, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724092, CurrentID=0, Status='live', Symbol='LOC102724092', Aliases='', description='uncharacterized LOC102724092', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15232234, end_position_on_the_genomic_accession=15236715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724093, CurrentID=0, Status='live', Symbol='LOC102724093', Aliases='', description='putative golgin subfamily A member 6-like protein 4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84350347, end_position_on_the_genomic_accession=84359422, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724094, CurrentID=0, Status='live', Symbol='LOC102724094', Aliases='', description='uncharacterized LOC102724094', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101014320, end_position_on_the_genomic_accession=101017608, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132949, CurrentID=0, Status='live', Symbol='RBM17P3', Aliases='', description='RNA binding motif protein 17 pseudogene 3', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41974309, end_position_on_the_genomic_accession=41975993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724096, CurrentID=0, Status='live', Symbol='LOC102724096', Aliases='', description='uncharacterized LOC102724096', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4186320, end_position_on_the_genomic_accession=4190029, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724097, CurrentID=0, Status='live', Symbol='TTC39A-AS1', Aliases='', description='TTC39A antisense RNA 1', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51329654, end_position_on_the_genomic_accession=51331281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724100, CurrentID=0, Status='live', Symbol='LOC102724100', Aliases='', description='uncharacterized LOC102724100', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234270403, end_position_on_the_genomic_accession=234273025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724101, CurrentID=0, Status='live', Symbol='TP53TG3E', Aliases='', description='TP53 target 3 family member E', other_designations='TP53-target gene 3 protein', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33304368, end_position_on_the_genomic_accession=33306938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724102, CurrentID=0, Status='live', Symbol='LOC102724102', Aliases='', description='uncharacterized LOC102724102', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18657652, end_position_on_the_genomic_accession=18694384, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724103, CurrentID=0, Status='live', Symbol='LOC102724103', Aliases='', description='uncharacterized LOC102724103', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39993915, end_position_on_the_genomic_accession=39997323, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724104, CurrentID=0, Status='live', Symbol='LOC102724104', Aliases='', description='uncharacterized LOC102724104', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39232520, end_position_on_the_genomic_accession=39259464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724105, CurrentID=0, Status='live', Symbol='ZNF528-AS1', Aliases='', description='ZNF528 antisense RNA 1', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52388839, end_position_on_the_genomic_accession=52397757, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724109, CurrentID=0, Status='live', Symbol='LOC102724109', Aliases='', description='uncharacterized LOC102724109', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79421984, end_position_on_the_genomic_accession=79489077, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724110, CurrentID=0, Status='live', Symbol='LOC102724110', Aliases='', description='uncharacterized LOC102724110', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64070574, end_position_on_the_genomic_accession=64075852, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81325, CurrentID=0, Status='live', Symbol='OR4A18P', Aliases='', description='olfactory receptor family 4 subfamily A member 18 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 18 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49917390, end_position_on_the_genomic_accession=49918477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724112, CurrentID=0, Status='live', Symbol='LOC102724112', Aliases='', description='uncharacterized LOC102724112', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55913746, end_position_on_the_genomic_accession=55923482, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724116, CurrentID=0, Status='live', Symbol='LOC102724116', Aliases='', description='uncharacterized LOC102724116', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=58087363, end_position_on_the_genomic_accession=58088585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724117, CurrentID=0, Status='live', Symbol='LOC102724117', Aliases='', description='putative golgin subfamily A member 6-like protein 4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84428359, end_position_on_the_genomic_accession=84437807, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375833, CurrentID=0, Status='live', Symbol='LOC105375833', Aliases='', description='uncharacterized LOC105375833', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51756745, end_position_on_the_genomic_accession=51779953, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724120, CurrentID=0, Status='live', Symbol='LOC102724120', Aliases='', description='uncharacterized LOC102724120', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143633876, end_position_on_the_genomic_accession=143657759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724122, CurrentID=0, Status='live', Symbol='LOC102724122', Aliases='', description='uncharacterized LOC102724122', other_designations='CTD-2023N9.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57483017, end_position_on_the_genomic_accession=57512753, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724123, CurrentID=0, Status='live', Symbol='LOC102724123', Aliases='', description='immunoglobulin superfamily member 3-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18794436, end_position_on_the_genomic_accession=18805215, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927229, CurrentID=0, Status='live', Symbol='LOC101927229', Aliases='', description='uncharacterized LOC101927229', other_designations='CTD-2171N6.1', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55105904, end_position_on_the_genomic_accession=55124306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724125, CurrentID=0, Status='live', Symbol='LOC102724125', Aliases='', description='uncharacterized LOC102724125', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53413151, end_position_on_the_genomic_accession=53428069, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724127, CurrentID=0, Status='live', Symbol='TP53TG3F', Aliases='', description='TP53 target 3 family member F', other_designations='TP53-target gene 3 protein', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33459674, end_position_on_the_genomic_accession=33462249, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81328, CurrentID=0, Status='live', Symbol='OR4A15', Aliases='OR11-118', description='olfactory receptor family 4 subfamily A member 15', other_designations='olfactory receptor OR11-118|olfactory receptor, family 4, subfamily A, member 15', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55367884, end_position_on_the_genomic_accession=55368918, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724130, CurrentID=0, Status='live', Symbol='LOC102724130', Aliases='', description='uncharacterized LOC102724130', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11147, end_position_on_the_genomic_accession=16350, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724133, CurrentID=0, Status='live', Symbol='LOC102724133', Aliases='', description='uncharacterized LOC102724133', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724134, CurrentID=0, Status='live', Symbol='LOC102724134', Aliases='', description='uncharacterized LOC102724134', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=61005069, end_position_on_the_genomic_accession=61006535, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724135, CurrentID=0, Status='live', Symbol='LOC102724135', Aliases='', description='golgin subfamily A member 2-like protein 3', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84442434, end_position_on_the_genomic_accession=84472780, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724139, CurrentID=0, Status='live', Symbol='LOC102724139', Aliases='', description='uncharacterized LOC102724139', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80896349, end_position_on_the_genomic_accession=80902474, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724140, CurrentID=0, Status='live', Symbol='LOC102724140', Aliases='', description='uncharacterized LOC102724140', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56436681, end_position_on_the_genomic_accession=56445707, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724142, CurrentID=0, Status='live', Symbol='LOC102724142', Aliases='', description='uncharacterized LOC102724142', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60394850, end_position_on_the_genomic_accession=60402128, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724143, CurrentID=0, Status='live', Symbol='LOC102724143', Aliases='', description='uncharacterized LOC102724143', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22110511, end_position_on_the_genomic_accession=22112603, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724145, CurrentID=0, Status='live', Symbol='LOC102724145', Aliases='', description='uncharacterized LOC102724145', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147736359, end_position_on_the_genomic_accession=147758678, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724146, CurrentID=0, Status='live', Symbol='LOC102724146', Aliases='', description='uncharacterized LOC102724146', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=15817069, end_position_on_the_genomic_accession=15882348, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  81331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81331, CurrentID=0, Status='live', Symbol='OR4A12P', Aliases='', description='olfactory receptor family 4 subfamily A member 12 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 12 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55325757, end_position_on_the_genomic_accession=55326701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724148, CurrentID=0, Status='live', Symbol='LOC102724148', Aliases='', description='uncharacterized LOC102724148', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18408127, end_position_on_the_genomic_accession=18422796, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724149, CurrentID=0, Status='live', Symbol='LOC102724149', Aliases='', description='uncharacterized LOC102724149', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95300632, end_position_on_the_genomic_accession=95310874, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724150, CurrentID=0, Status='live', Symbol='LOC102724150', Aliases='', description='uncharacterized LOC102724150', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=89403134, end_position_on_the_genomic_accession=89455254, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724151, CurrentID=0, Status='live', Symbol='LOC102724151', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724152, CurrentID=0, Status='live', Symbol='LOC102724152', Aliases='', description='uncharacterized LOC102724152', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163664366, end_position_on_the_genomic_accession=163671533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724153, CurrentID=0, Status='live', Symbol='LOC102724153', Aliases='', description='uncharacterized LOC102724153', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75574901, end_position_on_the_genomic_accession=75579588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724156, CurrentID=0, Status='live', Symbol='LOC102724156', Aliases='', description='uncharacterized LOC102724156', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88066435, end_position_on_the_genomic_accession=88072796, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724158, CurrentID=0, Status='live', Symbol='LOC102724158', Aliases='', description='uncharacterized LOC102724158', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121950980, end_position_on_the_genomic_accession=121989617, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724159, CurrentID=0, Status='live', Symbol='LOC102724159', Aliases='', description='periodic tryptophan protein 2 homolog', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5130871, end_position_on_the_genomic_accession=5154767, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  100129616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129616, CurrentID=0, Status='live', Symbol='LOC100129616', Aliases='', description='putative UPF0607 protein ENSP00000332738-like', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23854437, end_position_on_the_genomic_accession=23855457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724162, CurrentID=0, Status='live', Symbol='LOC102724162', Aliases='', description='uncharacterized LOC102724162', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22572728, end_position_on_the_genomic_accession=22577080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724163, CurrentID=0, Status='live', Symbol='LOC102724163', Aliases='', description='uncharacterized LOC102724163', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83797236, end_position_on_the_genomic_accession=83803976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724164, CurrentID=0, Status='live', Symbol='LOC102724164', Aliases='', description='uncharacterized LOC102724164', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7582170, end_position_on_the_genomic_accession=7583814, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724165, CurrentID=0, Status='live', Symbol='LOC102724165', Aliases='', description='uncharacterized LOC102724165', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115919626, end_position_on_the_genomic_accession=115946880, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102724167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724167, CurrentID=0, Status='live', Symbol='PABPC5-AS1', Aliases='CXorf46', description='PABPC5 antisense RNA 1', other_designations='PABPC5 antisense RNA 1 (non-protein coding)', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91414903, end_position_on_the_genomic_accession=91434999, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724168, CurrentID=0, Status='live', Symbol='LOC102724168', Aliases='', description='uncharacterized LOC102724168', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81843775, end_position_on_the_genomic_accession=82085889, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102724169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724169, CurrentID=0, Status='live', Symbol='LOC102724169', Aliases='', description='uncharacterized LOC102724169', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166331776, end_position_on_the_genomic_accession=166341191, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374305, CurrentID=0, Status='live', Symbol='LOC105374305', Aliases='', description='uncharacterized LOC105374305', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196455875, end_position_on_the_genomic_accession=196460736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724174, CurrentID=0, Status='live', Symbol='LOC102724174', Aliases='', description='uncharacterized LOC102724174', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239127261, end_position_on_the_genomic_accession=239208303, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724175, CurrentID=0, Status='live', Symbol='LOC102724175', Aliases='', description='membrane primary amine oxidase-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42871312, end_position_on_the_genomic_accession=42874494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724176, CurrentID=0, Status='live', Symbol='LOC102724176', Aliases='', description='uncharacterized LOC102724176', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124025295, end_position_on_the_genomic_accession=124028069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724178, CurrentID=0, Status='live', Symbol='LOC102724178', Aliases='', description='filamin-B-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51892389, end_position_on_the_genomic_accession=51901434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375843, CurrentID=0, Status='live', Symbol='LOC105375843', Aliases='', description='uncharacterized LOC105375843', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55514455, end_position_on_the_genomic_accession=55521012, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724180, CurrentID=0, Status='live', Symbol='LOC102724180', Aliases='', description='uncharacterized LOC102724180', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156511684, end_position_on_the_genomic_accession=156512509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724181, CurrentID=0, Status='live', Symbol='LOC102724181', Aliases='', description='rhophilin-2-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33571798, end_position_on_the_genomic_accession=33623371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724183, CurrentID=0, Status='live', Symbol='LOC102724183', Aliases='', description='uncharacterized LOC102724183', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43929879, end_position_on_the_genomic_accession=43943989, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724184, CurrentID=0, Status='live', Symbol='LOC102724184', Aliases='', description='DNA (cytosine-5)-methyltransferase 3-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5012324, end_position_on_the_genomic_accession=5017158, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375844, CurrentID=0, Status='live', Symbol='LOC105375844', Aliases='', description='uncharacterized LOC105375844', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55161476, end_position_on_the_genomic_accession=55164451, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724188, CurrentID=0, Status='live', Symbol='LOC102724188', Aliases='', description='uncharacterized LOC102724188', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13546033, end_position_on_the_genomic_accession=13558461, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724190, CurrentID=0, Status='live', Symbol='LOC102724190', Aliases='', description='uncharacterized LOC102724190', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77069180, end_position_on_the_genomic_accession=77076191, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375845, CurrentID=0, Status='live', Symbol='LOC105375845', Aliases='', description='uncharacterized LOC105375845', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55682471, end_position_on_the_genomic_accession=55683021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724193, CurrentID=0, Status='live', Symbol='LOC102724193', Aliases='', description='uncharacterized LOC102724193', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136649001, end_position_on_the_genomic_accession=136653923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724194, CurrentID=0, Status='live', Symbol='LOC102724194', Aliases='', description='uncharacterized LOC102724194', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174329340, end_position_on_the_genomic_accession=174334716, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724196, CurrentID=0, Status='live', Symbol='LOC102724196', Aliases='', description='uncharacterized LOC102724196', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16567192, end_position_on_the_genomic_accession=16575230, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724197, CurrentID=0, Status='live', Symbol='LOC102724197', Aliases='', description='inactive gamma-glutamyltranspeptidase 2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724200, CurrentID=0, Status='live', Symbol='LOC102724200', Aliases='', description='trafficking protein particle complex subunit 10-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5155499, end_position_on_the_genomic_accession=5165464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724201, CurrentID=0, Status='live', Symbol='TBX18-AS1', Aliases='', description='TBX18 antisense RNA 1', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84689458, end_position_on_the_genomic_accession=84709534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724202, CurrentID=0, Status='live', Symbol='LOC102724202', Aliases='', description='uncharacterized LOC102724202', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26635871, end_position_on_the_genomic_accession=26637719, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375847, CurrentID=0, Status='live', Symbol='LOC105375847', Aliases='', description='uncharacterized LOC105375847', other_designations='CTA-397H3.3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56074592, end_position_on_the_genomic_accession=56081915, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724205, CurrentID=0, Status='live', Symbol='LOC102724205', Aliases='', description='centromere protein J-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167183455, end_position_on_the_genomic_accession=167196978, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724207, CurrentID=0, Status='live', Symbol='LOC102724207', Aliases='', description='ankyrin repeat domain-containing protein 36C-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33721294, end_position_on_the_genomic_accession=33763756, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  102724208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724208, CurrentID=0, Status='live', Symbol='LOC102724208', Aliases='', description='uncharacterized LOC102724208', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21975281, end_position_on_the_genomic_accession=22043478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724209, CurrentID=0, Status='live', Symbol='LOC102724209', Aliases='', description='uncharacterized LOC102724209', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68272184, end_position_on_the_genomic_accession=68275036, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724210, CurrentID=0, Status='live', Symbol='LOC102724210', Aliases='', description='uncharacterized LOC102724210', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127073776, end_position_on_the_genomic_accession=127352572, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724211, CurrentID=0, Status='live', Symbol='LOC102724211', Aliases='', description='uncharacterized LOC102724211', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57873873, end_position_on_the_genomic_accession=57891558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724212, CurrentID=0, Status='live', Symbol='LOC102724212', Aliases='', description='uncharacterized LOC102724212', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101663713, end_position_on_the_genomic_accession=101671221, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724214, CurrentID=0, Status='live', Symbol='LOC102724214', Aliases='', description='uncharacterized LOC102724214', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36106915, end_position_on_the_genomic_accession=36180989, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724216, CurrentID=0, Status='live', Symbol='LOC102724216', Aliases='', description='uncharacterized LOC102724216', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33730175, end_position_on_the_genomic_accession=33748743, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724219, CurrentID=0, Status='live', Symbol='LOC102724219', Aliases='', description='uncharacterized LOC102724219', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5553666, end_position_on_the_genomic_accession=5592004, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724220, CurrentID=0, Status='live', Symbol='LOC102724220', Aliases='', description='uncharacterized LOC102724220', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167237518, end_position_on_the_genomic_accession=167241938, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724222, CurrentID=0, Status='live', Symbol='LOC102724222', Aliases='', description='uncharacterized LOC102724222', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157050052, end_position_on_the_genomic_accession=157051683, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724223, CurrentID=0, Status='live', Symbol='LOC102724223', Aliases='', description='sodium- and chloride-dependent creatine transporter 1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33976001, end_position_on_the_genomic_accession=33984174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724224, CurrentID=0, Status='live', Symbol='LINC01117', Aliases='', description='long intergenic non-protein coding RNA 1117', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176637754, end_position_on_the_genomic_accession=176655958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724225, CurrentID=0, Status='live', Symbol='LOC102724225', Aliases='', description='uncharacterized LOC102724225', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100729562, end_position_on_the_genomic_accession=100732024, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724227, CurrentID=0, Status='live', Symbol='LOC102724227', Aliases='', description='uncharacterized LOC102724227', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18753526, end_position_on_the_genomic_accession=18800108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724229, CurrentID=0, Status='live', Symbol='LOC102724229', Aliases='', description='uncharacterized LOC102724229', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102573648, end_position_on_the_genomic_accession=102580440, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724231, CurrentID=0, Status='live', Symbol='LOC102724231', Aliases='', description='uncharacterized LOC102724231', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44421132, end_position_on_the_genomic_accession=44424058, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375852, CurrentID=0, Status='live', Symbol='LOC105375852', Aliases='', description='uncharacterized LOC105375852', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56683232, end_position_on_the_genomic_accession=56715513, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724234, CurrentID=0, Status='live', Symbol='LOC102724234', Aliases='', description='uncharacterized LOC102724234', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7426847, end_position_on_the_genomic_accession=7452752, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724235, CurrentID=0, Status='live', Symbol='LOC102724235', Aliases='', description='protein FAM231D-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145164645, end_position_on_the_genomic_accession=145166132, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724236, CurrentID=0, Status='live', Symbol='LOC102724236', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243054512, end_position_on_the_genomic_accession=243055920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724237, CurrentID=0, Status='live', Symbol='LOC102724237', Aliases='', description='putative uncharacterized protein UNQ6494/PRO21346', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89664915, end_position_on_the_genomic_accession=89672876, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724238, CurrentID=0, Status='live', Symbol='LOC102724238', Aliases='', description='uncharacterized LOC102724238', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62375328, end_position_on_the_genomic_accession=62376870, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375853, CurrentID=0, Status='live', Symbol='LOC105375853', Aliases='', description='uncharacterized LOC105375853', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56998037, end_position_on_the_genomic_accession=57007957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724241, CurrentID=0, Status='live', Symbol='LOC102724241', Aliases='', description='uncharacterized LOC102724241', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46829505, end_position_on_the_genomic_accession=46833878, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724245, CurrentID=0, Status='live', Symbol='LOC102724245', Aliases='', description='uncharacterized LOC102724245', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27297807, end_position_on_the_genomic_accession=27307419, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724246, CurrentID=0, Status='live', Symbol='TTC39C-AS1', Aliases='', description='TTC39C antisense RNA 1', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23994213, end_position_on_the_genomic_accession=24015460, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724247, CurrentID=0, Status='live', Symbol='LOC102724247', Aliases='', description='uncharacterized LOC102724247', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20234853, end_position_on_the_genomic_accession=20264157, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724250, CurrentID=0, Status='live', Symbol='LOC102724250', Aliases='', description='neuroblastoma breakpoint family member 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724252, CurrentID=0, Status='live', Symbol='LOC102724252', Aliases='', description='uncharacterized LOC102724252', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102909431, end_position_on_the_genomic_accession=102964523, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724253, CurrentID=0, Status='live', Symbol='LOC102724253', Aliases='', description='uncharacterized LOC102724253', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38754950, end_position_on_the_genomic_accession=38757380, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724258, CurrentID=0, Status='live', Symbol='LOC102724258', Aliases='', description='WAS/WASL-interacting protein family member 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65432010, end_position_on_the_genomic_accession=65434846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724261, CurrentID=0, Status='live', Symbol='LOC102724261', Aliases='', description='uncharacterized LOC102724261', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=21727123, end_position_on_the_genomic_accession=21757855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724262, CurrentID=0, Status='live', Symbol='LOC102724262', Aliases='', description='uncharacterized LOC102724262', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8506497, end_position_on_the_genomic_accession=8513798, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132977, CurrentID=0, Status='live', Symbol='LOC100132977', Aliases='', description='chromosome 2 open reading frame 27B pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20555002, end_position_on_the_genomic_accession=20555572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724264, CurrentID=0, Status='live', Symbol='LOC102724264', Aliases='', description='uncharacterized LOC102724264', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43777562, end_position_on_the_genomic_accession=43778862, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724265, CurrentID=0, Status='live', Symbol='LOC102724265', Aliases='', description='uncharacterized LOC102724265', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69425671, end_position_on_the_genomic_accession=69429624, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132978, CurrentID=0, Status='live', Symbol='ATP6V0CP1', Aliases='ATCD1', description='ATPase, H+ transporting, lysosomal 16kDa, V0 subunit c pseudogene 1', other_designations='ATP synthase subunit C domain containing 1 pseudogene|ATPase, H+ transporting, lysosomal 16kDa, V0 subunit c (ATP6VOC) pseudogene', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4856967, end_position_on_the_genomic_accession=4857399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724271, CurrentID=0, Status='live', Symbol='LOC102724271', Aliases='', description='uncharacterized LOC102724271', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90973804, end_position_on_the_genomic_accession=91000414, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724273, CurrentID=0, Status='live', Symbol='LOC102724273', Aliases='', description='uncharacterized LOC102724273', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54146664, end_position_on_the_genomic_accession=54152219, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375859, CurrentID=0, Status='live', Symbol='LOC105375859', Aliases='', description='uncharacterized LOC105375859', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=59601009, end_position_on_the_genomic_accession=59620198, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  619189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619189, CurrentID=0, Status='live', Symbol='SERINC4', Aliases='', description='serine incorporator 4', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43794174, end_position_on_the_genomic_accession=43800097, orientation='minus', exon_count=12, OMIM=614550, no_hdr0=''),
  619190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619190, CurrentID=0, Status='live', Symbol='FDPSP2', Aliases='FDPSL2, FDPSL2A, FPSL2', description='farnesyl diphosphate synthase pseudogene 2', other_designations='TCAG_1641456|farnesyl diphosphate synthase-like 2 (farnesyl pyrophosphate synthetase-like 2)|farnesyl diphosphate synthase-like 2A (farnesyl pyrophosphate synthetase-like 2A)', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76470020, end_position_on_the_genomic_accession=76475000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724279, CurrentID=0, Status='live', Symbol='LOC102724279', Aliases='', description='uncharacterized LOC102724279', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15851993, end_position_on_the_genomic_accession=15864904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724281, CurrentID=0, Status='live', Symbol='LOC102724281', Aliases='', description='uncharacterized LOC102724281', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61096801, end_position_on_the_genomic_accession=61133282, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724282, CurrentID=0, Status='live', Symbol='LOC102724282', Aliases='', description='uncharacterized LOC102724282', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120724448, end_position_on_the_genomic_accession=120746675, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724283, CurrentID=0, Status='live', Symbol='LOC102724283', Aliases='', description='uncharacterized LOC102724283', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39383779, end_position_on_the_genomic_accession=39394120, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724284, CurrentID=0, Status='live', Symbol='LOC102724284', Aliases='', description='uncharacterized LOC102724284', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104822183, end_position_on_the_genomic_accession=104826508, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724288, CurrentID=0, Status='live', Symbol='LOC102724288', Aliases='', description='uncharacterized LOC102724288', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182051960, end_position_on_the_genomic_accession=182056212, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724289, CurrentID=0, Status='live', Symbol='LOC102724289', Aliases='', description='uncharacterized LOC102724289', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152650369, end_position_on_the_genomic_accession=152737779, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724290, CurrentID=0, Status='live', Symbol='LOC102724290', Aliases='', description='uncharacterized LOC102724290', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54671429, end_position_on_the_genomic_accession=54672828, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724291, CurrentID=0, Status='live', Symbol='LOC102724291', Aliases='', description='keratin, type I cytoskeletal 18 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5688427, end_position_on_the_genomic_accession=5693174, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724292, CurrentID=0, Status='live', Symbol='LOC102724292', Aliases='', description='uncharacterized LOC102724292', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70719772, end_position_on_the_genomic_accession=70755443, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724294, CurrentID=0, Status='live', Symbol='LOC102724294', Aliases='', description='uncharacterized LOC102724294', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81734402, end_position_on_the_genomic_accession=81743753, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619207, CurrentID=0, Status='live', Symbol='SCART1', Aliases='', description='scavenger receptor protein family member', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133453928, end_position_on_the_genomic_accession=133468445, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  619208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619208, CurrentID=0, Status='live', Symbol='FAM229B', Aliases='C6orf225', description='family with sequence similarity 229 member B', other_designations='UPF0731 protein C6orf225|family with sequence similarity 229, member B', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112087471, end_position_on_the_genomic_accession=112102790, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724297, CurrentID=0, Status='live', Symbol='LOC102724297', Aliases='', description='uncharacterized LOC102724297', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46364660, end_position_on_the_genomic_accession=46407059, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619211, CurrentID=0, Status='live', Symbol='DFNB58', Aliases='', description='deafness, autosomal recessive 58', other_designations='', map_location='2q14.2-q14.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619212, CurrentID=0, Status='live', Symbol='PA2G4P3', Aliases='HsT2542, PA2G4L4', description='proliferation-associated 2G4 pseudogene 3', other_designations='proliferation-associated 2G4, 38kDa pseudogene', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27452238, end_position_on_the_genomic_accession=27453816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724301, CurrentID=0, Status='live', Symbol='LOC102724301', Aliases='', description='uncharacterized LOC102724301', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119729583, end_position_on_the_genomic_accession=119739623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724305, CurrentID=0, Status='live', Symbol='LOC102724305', Aliases='', description='uncharacterized LOC102724305', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46468135, end_position_on_the_genomic_accession=46485234, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724307, CurrentID=0, Status='live', Symbol='CYP17A1-AS1', Aliases='CYP17A1OS, bA753C18.3', description='CYP17A1 antisense RNA 1', other_designations='CYP17A1 antisense RNA 1 (non-protein coding)|cytochrome P450, family 17, subfamily A, polypeptide 1 opposite strand', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102832667, end_position_on_the_genomic_accession=102834512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724309, CurrentID=0, Status='live', Symbol='LOC102724309', Aliases='', description='uncharacterized LOC102724309', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128653753, end_position_on_the_genomic_accession=128662528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724310, CurrentID=0, Status='live', Symbol='LOC102724310', Aliases='', description='C-terminal-binding protein 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5696501, end_position_on_the_genomic_accession=5698770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724312, CurrentID=0, Status='live', Symbol='LOC102724312', Aliases='', description='uncharacterized LOC102724312', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1430157, end_position_on_the_genomic_accession=1434573, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724316, CurrentID=0, Status='live', Symbol='SVIL-AS1', Aliases='', description='SVIL antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29409534, end_position_on_the_genomic_accession=29487856, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375866, CurrentID=0, Status='live', Symbol='LOC105375866', Aliases='', description='uncharacterized LOC105375866', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60468225, end_position_on_the_genomic_accession=60516851, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724318, CurrentID=0, Status='live', Symbol='LOC102724318', Aliases='', description='uncharacterized LOC102724318', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62470067, end_position_on_the_genomic_accession=62480285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724319, CurrentID=0, Status='live', Symbol='LOC102724319', Aliases='', description='uncharacterized LOC102724319', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63487786, end_position_on_the_genomic_accession=63517794, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724321, CurrentID=0, Status='live', Symbol='LINC01628', Aliases='', description='long intergenic non-protein coding RNA 1628', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66921510, end_position_on_the_genomic_accession=66922490, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724322, CurrentID=0, Status='live', Symbol='LOC102724322', Aliases='', description='uncharacterized LOC102724322', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36257845, end_position_on_the_genomic_accession=36299529, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375867, CurrentID=0, Status='live', Symbol='LOC105375867', Aliases='', description='uncharacterized LOC105375867', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60464577, end_position_on_the_genomic_accession=60465928, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724326, CurrentID=0, Status='live', Symbol='LOC102724326', Aliases='', description='uncharacterized LOC102724326', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129324823, end_position_on_the_genomic_accession=129327578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724327, CurrentID=0, Status='live', Symbol='LOC102724327', Aliases='', description='uncharacterized LOC102724327', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52201491, end_position_on_the_genomic_accession=52208240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724328, CurrentID=0, Status='live', Symbol='LOC102724328', Aliases='', description='uncharacterized LOC102724328', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40073791, end_position_on_the_genomic_accession=40079701, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873642, CurrentID=0, Status='live', Symbol='RNA5SP388', Aliases='RN5S388', description='RNA, 5S ribosomal pseudogene 388', other_designations='RNA, 5S ribosomal 388', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78177803, end_position_on_the_genomic_accession=78177921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724330, CurrentID=0, Status='live', Symbol='LOC102724330', Aliases='', description='uncharacterized LOC102724330', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51899282, end_position_on_the_genomic_accession=52022948, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724332, CurrentID=0, Status='live', Symbol='LOC102724332', Aliases='', description='uncharacterized LOC102724332', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90190207, end_position_on_the_genomic_accession=90195883, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724333, CurrentID=0, Status='live', Symbol='LOC102724333', Aliases='', description='uncharacterized LOC102724333', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29838722, end_position_on_the_genomic_accession=29839746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724334, CurrentID=0, Status='live', Symbol='LOC102724334', Aliases='', description='histone H2B type F-S-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5972932, end_position_on_the_genomic_accession=5973359, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724337, CurrentID=0, Status='live', Symbol='LOC102724337', Aliases='', description='uncharacterized LOC102724337', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63008743, end_position_on_the_genomic_accession=63022906, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724338, CurrentID=0, Status='live', Symbol='LOC102724338', Aliases='', description='uncharacterized LOC102724338', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=83903458, end_position_on_the_genomic_accession=83915360, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724339, CurrentID=0, Status='live', Symbol='CTB-30L5.1', Aliases='', description='uncharacterized LOC102724339', other_designations='CTB-111H14.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106570755, end_position_on_the_genomic_accession=106615753, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724340, CurrentID=0, Status='live', Symbol='LOC102724340', Aliases='', description='uncharacterized LOC102724340', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184190270, end_position_on_the_genomic_accession=184334893, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724341, CurrentID=0, Status='live', Symbol='NEURL1-AS1', Aliases='', description='NEURL1 antisense RNA 1', other_designations='', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103479603, end_position_on_the_genomic_accession=103517442, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724344, CurrentID=0, Status='live', Symbol='LOC102724344', Aliases='', description='uncharacterized LOC102724344', other_designations='CTD-2015G9.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86716913, end_position_on_the_genomic_accession=86742083, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724345, CurrentID=0, Status='live', Symbol='LOC102724345', Aliases='', description='ADP-ribosylation factor like GTPase 17A pseudogene', other_designations='ADP-ribosylation factor-like 17A pseudogene', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45525850, end_position_on_the_genomic_accession=45535971, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724347, CurrentID=0, Status='live', Symbol='LOC102724347', Aliases='', description='uncharacterized LOC102724347', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108279361, end_position_on_the_genomic_accession=108292727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724349, CurrentID=0, Status='live', Symbol='FSIP2-AS1', Aliases='', description='FSIP2 antisense RNA 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185756219, end_position_on_the_genomic_accession=185788201, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724350, CurrentID=0, Status='live', Symbol='LOC102724350', Aliases='', description='uncharacterized LOC102724350', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46629291, end_position_on_the_genomic_accession=46633159, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724351, CurrentID=0, Status='live', Symbol='LOC102724351', Aliases='', description='uncharacterized LOC102724351', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103937113, end_position_on_the_genomic_accession=103967053, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724352, CurrentID=0, Status='live', Symbol='LOC102724352', Aliases='', description='uncharacterized LOC102724352', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55175311, end_position_on_the_genomic_accession=55177540, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375872, CurrentID=0, Status='live', Symbol='LOC105375872', Aliases='', description='uncharacterized LOC105375872', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62990978, end_position_on_the_genomic_accession=62993185, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724354, CurrentID=0, Status='live', Symbol='LOC102724354', Aliases='', description='uncharacterized LOC102724354', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6057767, end_position_on_the_genomic_accession=6066898, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724355, CurrentID=0, Status='live', Symbol='LOC102724355', Aliases='', description='uncharacterized LOC102724355', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27173543, end_position_on_the_genomic_accession=27351332, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724356, CurrentID=0, Status='live', Symbol='LOC102724356', Aliases='', description='uncharacterized LOC102724356', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69979017, end_position_on_the_genomic_accession=70473520, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724357, CurrentID=0, Status='live', Symbol='LOC102724357', Aliases='', description='uncharacterized LOC102724357', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169067786, end_position_on_the_genomic_accession=169069416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724358, CurrentID=0, Status='live', Symbol='LOC102724358', Aliases='', description='uncharacterized LOC102724358', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162552151, end_position_on_the_genomic_accession=162561489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375873, CurrentID=0, Status='live', Symbol='LOC105375873', Aliases='', description='uncharacterized LOC105375873', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63243403, end_position_on_the_genomic_accession=63308374, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724360, CurrentID=0, Status='live', Symbol='LOC102724360', Aliases='', description='uncharacterized LOC102724360', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18946128, end_position_on_the_genomic_accession=18946836, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724362, CurrentID=0, Status='live', Symbol='LOC102724362', Aliases='', description='uncharacterized LOC102724362', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40835993, end_position_on_the_genomic_accession=40844268, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724363, CurrentID=0, Status='live', Symbol='LOC102724363', Aliases='', description='uncharacterized LOC102724363', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108164564, end_position_on_the_genomic_accession=108177126, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724364, CurrentID=0, Status='live', Symbol='LOC102724364', Aliases='', description='vesicle-trafficking protein SEC22b-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146316891, end_position_on_the_genomic_accession=146376833, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  256130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256130, CurrentID=0, Status='live', Symbol='TMEM196', Aliases='', description='transmembrane protein 196', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19719315, end_position_on_the_genomic_accession=19773576, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  619279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619279, CurrentID=0, Status='live', Symbol='ZNF704', Aliases='Gig1', description='zinc finger protein 704', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80628451, end_position_on_the_genomic_accession=80874781, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102724368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724368, CurrentID=0, Status='live', Symbol='CELSR3-AS1', Aliases='', description='CELSR3 antisense RNA 1 (head to head)', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48663768, end_position_on_the_genomic_accession=48669174, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724370, CurrentID=0, Status='live', Symbol='LOC102724370', Aliases='', description='uncharacterized LOC102724370', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6070758, end_position_on_the_genomic_accession=6073132, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375875, CurrentID=0, Status='live', Symbol='LOC105375875', Aliases='', description='uncharacterized LOC105375875', other_designations='CTD-3046C4.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63384839, end_position_on_the_genomic_accession=63470197, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724373, CurrentID=0, Status='live', Symbol='LOC102724373', Aliases='', description='uncharacterized LOC102724373', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67564318, end_position_on_the_genomic_accession=67598767, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100422046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422046, CurrentID=0, Status='live', Symbol='LOC100422046', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724378, CurrentID=0, Status='live', Symbol='LOC102724378', Aliases='', description='uncharacterized LOC102724378', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37729416, end_position_on_the_genomic_accession=37747434, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724379, CurrentID=0, Status='live', Symbol='LOC102724379', Aliases='', description='uncharacterized LOC102724379', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11862808, end_position_on_the_genomic_accession=11868586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724380, CurrentID=0, Status='live', Symbol='LOC102724380', Aliases='', description='uncharacterized LOC102724380', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=86007099, end_position_on_the_genomic_accession=86062957, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724382, CurrentID=0, Status='live', Symbol='LOC102724382', Aliases='', description='uncharacterized LOC102724382', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246682071, end_position_on_the_genomic_accession=246692537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375877, CurrentID=0, Status='live', Symbol='LOC105375877', Aliases='', description='uncharacterized LOC105375877', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64354506, end_position_on_the_genomic_accession=64356300, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724385, CurrentID=0, Status='live', Symbol='LOC102724385', Aliases='', description='uncharacterized LOC102724385', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6066958, end_position_on_the_genomic_accession=6100022, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724386, CurrentID=0, Status='live', Symbol='LINC01393', Aliases='', description='long intergenic non-protein coding RNA 1393', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115078958, end_position_on_the_genomic_accession=115126314, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724389, CurrentID=0, Status='live', Symbol='LOC102724389', Aliases='', description='uncharacterized LOC102724389', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68247218, end_position_on_the_genomic_accession=68346020, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724390, CurrentID=0, Status='live', Symbol='LOC102724390', Aliases='', description='D(1B) dopamine receptor-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89610044, end_position_on_the_genomic_accession=89611840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724391, CurrentID=0, Status='live', Symbol='LOC102724391', Aliases='', description='uncharacterized LOC102724391', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46034951, end_position_on_the_genomic_accession=46035770, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724392, CurrentID=0, Status='live', Symbol='LOC102724392', Aliases='', description='uncharacterized LOC102724392', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71313576, end_position_on_the_genomic_accession=71448004, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724393, CurrentID=0, Status='live', Symbol='LOC102724393', Aliases='', description='uncharacterized LOC102724393', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85349344, end_position_on_the_genomic_accession=85373931, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724394, CurrentID=0, Status='live', Symbol='LOC102724394', Aliases='', description='uncharacterized LOC102724394', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138273125, end_position_on_the_genomic_accession=138275506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724397, CurrentID=0, Status='live', Symbol='LOC102724397', Aliases='', description='uncharacterized LOC102724397', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9318622, end_position_on_the_genomic_accession=9334475, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724398, CurrentID=0, Status='live', Symbol='CH507-42P11.6', Aliases='', description='putative uncharacterized protein C21orf125', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6084373, end_position_on_the_genomic_accession=6088242, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724400, CurrentID=0, Status='live', Symbol='LOC102724400', Aliases='', description='uncharacterized LOC102724400', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49299395, end_position_on_the_genomic_accession=49301465, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375880, CurrentID=0, Status='live', Symbol='LOC105375880', Aliases='', description='uncharacterized LOC105375880', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65006900, end_position_on_the_genomic_accession=65017822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724404, CurrentID=0, Status='live', Symbol='LOC102724404', Aliases='', description='uncharacterized LOC102724404', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157361616, end_position_on_the_genomic_accession=157460056, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724406, CurrentID=0, Status='live', Symbol='LOC102724406', Aliases='', description='uncharacterized LOC102724406', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100393887, end_position_on_the_genomic_accession=100445425, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724407, CurrentID=0, Status='live', Symbol='LOC102724407', Aliases='', description='uncharacterized LOC102724407', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115789912, end_position_on_the_genomic_accession=115791323, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724410, CurrentID=0, Status='live', Symbol='LOC102724410', Aliases='', description='uncharacterized LOC102724410', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124218136, end_position_on_the_genomic_accession=124228128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724416, CurrentID=0, Status='live', Symbol='LOC102724416', Aliases='', description='uncharacterized LOC102724416', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65277547, end_position_on_the_genomic_accession=65309457, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724417, CurrentID=0, Status='live', Symbol='LOC102724417', Aliases='', description='uncharacterized LOC102724417', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37722772, end_position_on_the_genomic_accession=37725864, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724418, CurrentID=0, Status='live', Symbol='LOC102724418', Aliases='', description='uncharacterized LOC102724418', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105489342, end_position_on_the_genomic_accession=105642760, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724419, CurrentID=0, Status='live', Symbol='LOC102724419', Aliases='', description='uncharacterized LOC102724419', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164029276, end_position_on_the_genomic_accession=164030685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724420, CurrentID=0, Status='live', Symbol='LOC102724420', Aliases='', description='uncharacterized LOC102724420', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724421, CurrentID=0, Status='live', Symbol='LOC102724421', Aliases='', description='uncharacterized LOC102724421', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67077626, end_position_on_the_genomic_accession=67096284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724422, CurrentID=0, Status='live', Symbol='LOC102724422', Aliases='', description='uncharacterized LOC102724422', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100881408, end_position_on_the_genomic_accession=101167758, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  81377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81377, CurrentID=0, Status='live', Symbol='OR13C7', Aliases='OR13C7P, OR37C, OST706', description='olfactory receptor family 13 subfamily C member 7 (gene/pseudogene)', other_designations='olfactory receptor, family 13, subfamily C, member 7 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36002912, end_position_on_the_genomic_accession=36003867, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371688, CurrentID=0, Status='live', Symbol='LOC105371688', Aliases='', description='uncharacterized LOC105371688', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203510493, end_position_on_the_genomic_accession=203513771, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724427, CurrentID=0, Status='live', Symbol='LOC102724427', Aliases='', description='uncharacterized LOC102724427', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20247268, end_position_on_the_genomic_accession=20257291, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724428, CurrentID=0, Status='live', Symbol='LOC102724428', Aliases='', description='serine/threonine-protein kinase SIK1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6111135, end_position_on_the_genomic_accession=6123746, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  102724429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724429, CurrentID=0, Status='live', Symbol='LOC102724429', Aliases='', description='uncharacterized LOC102724429', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4998037, end_position_on_the_genomic_accession=5016708, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724430, CurrentID=0, Status='live', Symbol='KRT8P50', Aliases='', description='keratin 8 pseudogene 50', other_designations='keratin, type II cytoskeletal 8 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42800608, end_position_on_the_genomic_accession=42802508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  619343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619343, CurrentID=0, Status='live', Symbol='ERICH1-AS1', Aliases='C8orf68', description='ERICH1 antisense RNA 1', other_designations='ERICH1 antisense RNA 1 (non-protein coding)', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=737605, end_position_on_the_genomic_accession=1137777, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  619344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619344, CurrentID=0, Status='live', Symbol='LINC01298', Aliases='C8orf69', description='long intergenic non-protein coding RNA 1298', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95207007, end_position_on_the_genomic_accession=95216374, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724433, CurrentID=0, Status='live', Symbol='LOC102724433', Aliases='', description='uncharacterized LOC102724433', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67440998, end_position_on_the_genomic_accession=67442952, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724434, CurrentID=0, Status='live', Symbol='LOC102724434', Aliases='', description='uncharacterized LOC102724434', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116275606, end_position_on_the_genomic_accession=116286727, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724436, CurrentID=0, Status='live', Symbol='LOC102724436', Aliases='', description='putative zinc finger protein 66', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20397440, end_position_on_the_genomic_accession=20398579, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619349, CurrentID=0, Status='live', Symbol='SETP3', Aliases='SETP13', description='SET pseudogene 3', other_designations='SET pseudogene 13|SET translocation, pseudogene 3', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58475991, end_position_on_the_genomic_accession=58477102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724438, CurrentID=0, Status='live', Symbol='LOC102724438', Aliases='', description='uncharacterized LOC102724438', other_designations='CTD-2330K9.2|CTD-2330K9.3', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49899191, end_position_on_the_genomic_accession=49925038, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  619351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619351, CurrentID=0, Status='live', Symbol='LINC00589', Aliases='C8orf75', description='long intergenic non-protein coding RNA 589', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29721259, end_position_on_the_genomic_accession=29748109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724441, CurrentID=0, Status='live', Symbol='LOC102724441', Aliases='', description='uncharacterized LOC102724441', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6228966, end_position_on_the_genomic_accession=6267304, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724442, CurrentID=0, Status='live', Symbol='LOC102724442', Aliases='', description='uncharacterized LOC102724442', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110309095, end_position_on_the_genomic_accession=110315426, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724443, CurrentID=0, Status='live', Symbol='LOC102724443', Aliases='', description='uncharacterized LOC102724443', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104336527, end_position_on_the_genomic_accession=104359776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724444, CurrentID=0, Status='live', Symbol='LOC102724444', Aliases='', description='uncharacterized LOC102724444', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90613393, end_position_on_the_genomic_accession=90630842, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724446, CurrentID=0, Status='live', Symbol='LOC102724446', Aliases='', description='uncharacterized LOC102724446', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247565639, end_position_on_the_genomic_accession=247641191, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724447, CurrentID=0, Status='live', Symbol='LOC102724447', Aliases='', description='protein capicua homolog pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138282007, end_position_on_the_genomic_accession=138284727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375888, CurrentID=0, Status='live', Symbol='LOC105375888', Aliases='', description='uncharacterized LOC105375888', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68082051, end_position_on_the_genomic_accession=68095113, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724450, CurrentID=0, Status='live', Symbol='LOC102724450', Aliases='', description='uncharacterized LOC102724450', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6204840, end_position_on_the_genomic_accession=6205780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724452, CurrentID=0, Status='live', Symbol='LOC102724452', Aliases='', description='uncharacterized LOC102724452', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86938797, end_position_on_the_genomic_accession=86939766, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373434, CurrentID=0, Status='live', Symbol='LOC105373434', Aliases='', description='uncharacterized LOC105373434', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724455, CurrentID=0, Status='live', Symbol='LOC102724455', Aliases='', description='uncharacterized LOC102724455', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48036675, end_position_on_the_genomic_accession=48042170, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724457, CurrentID=0, Status='live', Symbol='LOC102724457', Aliases='', description='uncharacterized LOC102724457', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124740135, end_position_on_the_genomic_accession=124745998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619370, CurrentID=0, Status='live', Symbol='MCDR4', Aliases='', description='macular dystrophy, retinal, 4 (North Carolina type with progressive sensorineural hearing loss)', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724459, CurrentID=0, Status='live', Symbol='LOC102724459', Aliases='', description='otopetrin-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619373, CurrentID=0, Status='live', Symbol='MBOAT4', Aliases='FKSG89, GOAT, OACT4', description='membrane bound O-acyltransferase domain containing 4', other_designations='O-acyltransferase (membrane bound) domain containing 4|O-acyltransferase domain-containing protein 4|membrane-bound O-acyltransferase domain-containing protein 4', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30131454, end_position_on_the_genomic_accession=30158699, orientation='minus', exon_count=6, OMIM=611940, no_hdr0=''),
  102724463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724463, CurrentID=0, Status='live', Symbol='LOC102724463', Aliases='', description='uncharacterized LOC102724463', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14661124, end_position_on_the_genomic_accession=14665219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724465, CurrentID=0, Status='live', Symbol='LOC102724465', Aliases='', description='uncharacterized LOC102724465', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=87429904, end_position_on_the_genomic_accession=87501807, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  619379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619379, CurrentID=0, Status='live', Symbol='SPG29', Aliases='', description='spastic paraplegia 29 (autosomal dominant)', other_designations='', map_location='1p31.1-p21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609727, no_hdr0=''),
  102724468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724468, CurrentID=0, Status='live', Symbol='LOC102724468', Aliases='', description='olfactory receptor 4F3/4F16/4F29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138319283, end_position_on_the_genomic_accession=138321334, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724470, CurrentID=0, Status='live', Symbol='LOC102724470', Aliases='', description='uncharacterized LOC102724470', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10612677, end_position_on_the_genomic_accession=10629072, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  619383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619383, CurrentID=0, Status='live', Symbol='SCARNA9', Aliases='Z32, mgU2-19/30', description='small Cajal body-specific RNA 9', other_designations='mgU2-19/30 scaRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93721514, end_position_on_the_genomic_accession=93721866, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724473, CurrentID=0, Status='live', Symbol='GAGE10', Aliases='GAGE-10', description='G antigen 10', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49303646, end_position_on_the_genomic_accession=49319844, orientation='plus', exon_count=5, OMIM=300737, no_hdr0=''),
  102724474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724474, CurrentID=0, Status='live', Symbol='LOC102724474', Aliases='', description='WAS/WASL-interacting protein family member 2-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113149633, end_position_on_the_genomic_accession=113153986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724475, CurrentID=0, Status='live', Symbol='LOC102724475', Aliases='', description='uncharacterized LOC102724475', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169973375, end_position_on_the_genomic_accession=170023239, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724478, CurrentID=0, Status='live', Symbol='LOC102724478', Aliases='', description='uncharacterized LOC102724478', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95550539, end_position_on_the_genomic_accession=95565938, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724479, CurrentID=0, Status='live', Symbol='LOC102724479', Aliases='', description='uncharacterized LOC102724479', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171460554, end_position_on_the_genomic_accession=171469177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927809, CurrentID=0, Status='live', Symbol='LOC101927809', Aliases='', description='uncharacterized LOC101927809', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36180001, end_position_on_the_genomic_accession=36187438, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724481, CurrentID=0, Status='live', Symbol='LOC102724481', Aliases='', description='uncharacterized LOC102724481', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67522311, end_position_on_the_genomic_accession=67532308, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  619394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619394, CurrentID=0, Status='live', Symbol='SPOAN', Aliases='', description='spastic paraplegia, optic atrophy, and neuropathy', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609541, no_hdr0=''),
  102724484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724484, CurrentID=0, Status='live', Symbol='LOC102724484', Aliases='', description='uncharacterized LOC102724484', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29199040, end_position_on_the_genomic_accession=29208970, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  619397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619397, CurrentID=0, Status='live', Symbol='FEB6', Aliases='', description='febrile convulsions 6', other_designations='', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609253, no_hdr0=''),
  619398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619398, CurrentID=0, Status='live', Symbol='FEB5', Aliases='', description='febrile convulsions 5', other_designations='', map_location='6q22-q24', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609255, no_hdr0=''),
  619399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619399, CurrentID=0, Status='live', Symbol='ETL2', Aliases='FTLE', description='epilepsy, familial temporal lobe', other_designations='', map_location='12q22-q23.3', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608096, no_hdr0=''),
  102724488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724488, CurrentID=0, Status='live', Symbol='LOC102724488', Aliases='', description='synaptotagmin-15', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47746653, end_position_on_the_genomic_accession=47763592, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  102724489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724489, CurrentID=0, Status='live', Symbol='LOC102724489', Aliases='', description='uncharacterized LOC102724489', other_designations='LOF:RP11-65D24.2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111669647, end_position_on_the_genomic_accession=111672566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  619402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619402, CurrentID=0, Status='live', Symbol='HPC5', Aliases='', description='prostate cancer, hereditary, 5', other_designations='', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609299, no_hdr0=''),
  102724491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724491, CurrentID=0, Status='live', Symbol='LOC102724491', Aliases='', description='uncharacterized LOC102724491', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=113970746, end_position_on_the_genomic_accession=113977143, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  619404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619404, CurrentID=0, Status='live', Symbol='CTRCT35', Aliases='CATCN1', description='cataract, congenital nuclear, autosomal recessive', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609376, no_hdr0=''),
  619405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619405, CurrentID=0, Status='live', Symbol='AIS4', Aliases='VAMAS5', description='autoimmune disease, susceptibility to, 4', other_designations='', map_location='4q13-q21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609400, no_hdr0=''),
  619406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619406, CurrentID=0, Status='live', Symbol='HPE8', Aliases='', description='holoprosencephaly 8', other_designations='', map_location='14q13', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609408, no_hdr0=''),
  619407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619407, CurrentID=0, Status='live', Symbol='MSSD', Aliases='', description='syndactyly, mesoaxial synostotic, with phalangeal reduction', other_designations='syntactyly, type IX', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609432, no_hdr0=''),
  619408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619408, CurrentID=0, Status='live', Symbol='PSNP2', Aliases='', description='supranuclear palsy, progressive, 2', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609454, no_hdr0=''),
  619409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619409, CurrentID=0, Status='live', Symbol='MDCMP', Aliases='', description='muscular dystrophy, congenital, merosin-positive', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609456, no_hdr0=''),
  619410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619410, CurrentID=0, Status='live', Symbol='NEDE', Aliases='', description='nephropathy, progressive, with deafness', other_designations='', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609469, no_hdr0=''),
  619411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619411, CurrentID=0, Status='live', Symbol='LVNC2', Aliases='', description='noncompaction of left ventricular myocardium, familial isolated, autosomal dominant 2', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609470, no_hdr0=''),
  102724502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724502, CurrentID=0, Status='live', Symbol='LOC102724502', Aliases='', description='uncharacterized LOC102724502', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32958888, end_position_on_the_genomic_accession=32960566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  619418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619418, CurrentID=0, Status='live', Symbol='FAM90A21P', Aliases='', description='family with sequence similarity 90 member A21, pseudogene', other_designations='family with sequence similarity 90, member A21, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7563788, end_position_on_the_genomic_accession=7569019, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724507, CurrentID=0, Status='live', Symbol='LOC102724507', Aliases='', description='uncharacterized LOC102724507', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65132444, end_position_on_the_genomic_accession=65144175, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  102724508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724508, CurrentID=0, Status='live', Symbol='THCAT158', Aliases='', description='thyroid cancer-associated transcript 158', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47303443, end_position_on_the_genomic_accession=47323901, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724509, CurrentID=0, Status='live', Symbol='LOC102724509', Aliases='', description='uncharacterized LOC102724509', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37235996, end_position_on_the_genomic_accession=37251719, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724510, CurrentID=0, Status='live', Symbol='LOC102724510', Aliases='', description='uncharacterized LOC102724510', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111754561, end_position_on_the_genomic_accession=111757470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  619423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619423, CurrentID=0, Status='live', Symbol='FAM85A', Aliases='', description='family with sequence similarity 85 member A', other_designations='family with sequence similarity 85, member A', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12186897, end_position_on_the_genomic_accession=12204842, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724513, CurrentID=0, Status='live', Symbol='LOC102724513', Aliases='', description='uncharacterized LOC102724513', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88546506, end_position_on_the_genomic_accession=88547315, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724514, CurrentID=0, Status='live', Symbol='LOC102724514', Aliases='', description='uncharacterized LOC102724514', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121179967, end_position_on_the_genomic_accession=121266708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724515, CurrentID=0, Status='live', Symbol='LINC01291', Aliases='', description='long intergenic non-protein coding RNA 1291', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74909131, end_position_on_the_genomic_accession=74938418, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724517, CurrentID=0, Status='live', Symbol='LOC102724517', Aliases='', description='uncharacterized LOC102724517', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28043340, end_position_on_the_genomic_accession=28048926, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  619431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619431, CurrentID=0, Status='live', Symbol='FAM85B', Aliases='', description='family with sequence similarity 85 member B', other_designations='family with sequence similarity 85, member B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724520, CurrentID=0, Status='live', Symbol='LOC102724520', Aliases='', description='uncharacterized LOC102724520', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15665458, end_position_on_the_genomic_accession=15758240, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724521, CurrentID=0, Status='live', Symbol='LOC102724521', Aliases='', description='uncharacterized LOC102724521', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=430265, end_position_on_the_genomic_accession=470745, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  619434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619434, CurrentID=0, Status='live', Symbol='LINC00051', Aliases='C8orf43, NCRNA00051', description='long intergenic non-protein coding RNA 51', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142198356, end_position_on_the_genomic_accession=142209003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724523, CurrentID=0, Status='live', Symbol='LOC102724523', Aliases='', description='uncharacterized LOC102724523', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34431899, end_position_on_the_genomic_accession=34433114, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371417, CurrentID=0, Status='live', Symbol='LOC105371417', Aliases='', description='homeobox protein cut-like 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89701491, end_position_on_the_genomic_accession=89703524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724525, CurrentID=0, Status='live', Symbol='LOC102724525', Aliases='', description='carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6322697, end_position_on_the_genomic_accession=6324836, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724526, CurrentID=0, Status='live', Symbol='LOC102724526', Aliases='', description='protein phosphatase 1 regulatory subunit 26-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724527, CurrentID=0, Status='live', Symbol='LOC102724527', Aliases='', description='uncharacterized LOC102724527', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122192419, end_position_on_the_genomic_accession=122219265, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724528, CurrentID=0, Status='live', Symbol='LOC102724528', Aliases='', description='uncharacterized LOC102724528', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172393133, end_position_on_the_genomic_accession=172456471, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724529, CurrentID=0, Status='live', Symbol='LOC102724529', Aliases='', description='uncharacterized LOC102724529', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1413625, end_position_on_the_genomic_accession=1424572, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724530, CurrentID=0, Status='live', Symbol='LOC102724530', Aliases='', description='uncharacterized LOC102724530', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79317009, end_position_on_the_genomic_accession=79348008, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619443, CurrentID=0, Status='live', Symbol='RPS26P7', Aliases='HsT24280, RPS26_15_908', description='ribosomal protein S26 pseudogene 7', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58588335, end_position_on_the_genomic_accession=58588788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724532, CurrentID=0, Status='live', Symbol='LOC102724532', Aliases='', description='uncharacterized LOC102724532', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47891229, end_position_on_the_genomic_accession=47895811, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  619445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619445, CurrentID=0, Status='live', Symbol='RPL10P3', Aliases='bA567J2.2', description='ribosomal protein L10 pseudogene 3', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117180644, end_position_on_the_genomic_accession=117181010, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724534, CurrentID=0, Status='live', Symbol='LOC102724534', Aliases='', description='uncharacterized LOC102724534', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11695144, end_position_on_the_genomic_accession=11714338, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724535, CurrentID=0, Status='live', Symbol='LOC102724535', Aliases='', description='uncharacterized LOC102724535', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51668599, end_position_on_the_genomic_accession=51671097, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724536, CurrentID=0, Status='live', Symbol='PRR33', Aliases='C11orf89', description='proline rich 33', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1887586, end_position_on_the_genomic_accession=1917665, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  619449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619449, CurrentID=0, Status='live', Symbol='RPIAP1', Aliases='', description='ribose 5-phosphate isomerase A pseudogene 1', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61957094, end_position_on_the_genomic_accession=61957553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724539, CurrentID=0, Status='live', Symbol='LOC102724539', Aliases='', description='uncharacterized LOC102724539', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8208672, end_position_on_the_genomic_accession=8215210, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  619452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619452, CurrentID=0, Status='live', Symbol='RPS29P2', Aliases='HsT23291, RPS29_12_897', description='ribosomal protein S29 pseudogene 2', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41271404, end_position_on_the_genomic_accession=41271659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724541, CurrentID=0, Status='live', Symbol='LOC102724541', Aliases='', description='uncharacterized LOC102724541', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172369728, end_position_on_the_genomic_accession=172384910, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724542, CurrentID=0, Status='live', Symbol='LOC102724542', Aliases='', description='uncharacterized LOC102724542', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81481750, end_position_on_the_genomic_accession=81850643, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  619455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619455, CurrentID=0, Status='live', Symbol='SPANXA2-OT1', Aliases='CXorf18', description='SPANXA2 overlapping transcript 1', other_designations='SPANXA2 overlapping transcript 1 (non-protein coding)', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141502849, end_position_on_the_genomic_accession=141649939, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  619457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619457, CurrentID=0, Status='live', Symbol='UBE2WP1', Aliases='UBE2WP', description='ubiquitin conjugating enzyme E2W pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2W pseudogene 1', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96418573, end_position_on_the_genomic_accession=96418922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724547, CurrentID=0, Status='live', Symbol='LOC102724547', Aliases='', description='uncharacterized LOC102724547', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724548, CurrentID=0, Status='live', Symbol='LOC102724548', Aliases='', description='uncharacterized LOC102724548', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70355885, end_position_on_the_genomic_accession=70361007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724550, CurrentID=0, Status='live', Symbol='KCCAT211', Aliases='', description='renal clear cell carcinoma-associated transcript 211', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177816865, end_position_on_the_genomic_accession=177899224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724551, CurrentID=0, Status='live', Symbol='LOC102724551', Aliases='', description='uncharacterized LOC102724551', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174075178, end_position_on_the_genomic_accession=174086601, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  619464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619464, CurrentID=0, Status='live', Symbol='RPL31P9', Aliases='', description='ribosomal protein L31 pseudogene 9', other_designations='ribosomal protein L31 pseudogene 8', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67137611, end_position_on_the_genomic_accession=67137758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  619465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619465, CurrentID=0, Status='live', Symbol='ERVK-8', Aliases='ERVK8, HERV-K115, envK6', description='endogenous retrovirus group K, member 8', other_designations='HERV-K_8p23.1 provirus Rec protein|HERV-K_8p23.1 provirus ancestral Env polyprotein|HERV-K_8p23.1 provirus ancestral Gag polyprotein|HERV-K_8p23.1 provirus ancestral Pol protein|HERV-K_8p23.1 provirus ancestral Pro protein|endogenous retroviral sequence K, 8', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724555, CurrentID=0, Status='live', Symbol='LOC102724555', Aliases='', description='uncharacterized LOC102724555', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123614161, end_position_on_the_genomic_accession=123624957, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  619468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619468, CurrentID=0, Status='live', Symbol='RCM2', Aliases='', description='cardiomyopathy, familial restrictive, 2', other_designations='', map_location='10q23.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609578, no_hdr0=''),
  619469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619469, CurrentID=0, Status='live', Symbol='DELYQ11', Aliases='CYDELq11, SCO, SPGFY1', description='Sertoli cell-only syndrome, Y-linked', other_designations='', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=400042, no_hdr0=''),
  102724558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724558, CurrentID=0, Status='live', Symbol='CH17-408M7.1', Aliases='', description='uncharacterized LOC102724558', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146486332, end_position_on_the_genomic_accession=146525289, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  619471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619471, CurrentID=0, Status='live', Symbol='PPR3', Aliases='', description='photoparoxysmal response 3', other_designations='', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609573, no_hdr0=''),
  102724560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724560, CurrentID=0, Status='live', Symbol='CBSL', Aliases='', description='cystathionine-beta-synthase like', other_designations='cystathionine beta-synthase', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6444869, end_position_on_the_genomic_accession=6468507, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  102724562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724562, CurrentID=0, Status='live', Symbol='LOC102724562', Aliases='', description='hepatocyte growth factor-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619475, CurrentID=0, Status='live', Symbol='PFM3', Aliases='', description='parietal foramina 3', other_designations='', map_location='4q21-q23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609566, no_hdr0=''),
  619477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619477, CurrentID=0, Status='live', Symbol='GCCD3', Aliases='FGD2, FGD3, GCCD2', description='glucocorticoid deficiency 3', other_designations='', map_location='8q11.2-q13.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609197, no_hdr0=''),
  619478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619478, CurrentID=0, Status='live', Symbol='ATFB2', Aliases='', description='familial atrial fibrillation 2', other_designations='', map_location='6q14-q16', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608988, no_hdr0=''),
  619479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619479, CurrentID=0, Status='live', Symbol='STQTL5', Aliases='', description='stature quantitative trait locus 5', other_designations='', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608982, no_hdr0=''),
  619480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619480, CurrentID=0, Status='live', Symbol='MDBS2', Aliases='', description='butterfly-shaped pigmentary macular dystrophy 2', other_designations='', map_location='5q21.2-q33.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608970, no_hdr0=''),
  105375783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375783, CurrentID=0, Status='live', Symbol='LOC105375783', Aliases='', description='uncharacterized LOC105375783', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141023547, end_position_on_the_genomic_accession=141033014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724571, CurrentID=0, Status='live', Symbol='LOC102724571', Aliases='', description='uncharacterized LOC102724571', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9181422, end_position_on_the_genomic_accession=9192086, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724572, CurrentID=0, Status='live', Symbol='LOC102724572', Aliases='', description='uncharacterized LOC102724572', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70812548, end_position_on_the_genomic_accession=70855460, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375909, CurrentID=0, Status='live', Symbol='LOC105375909', Aliases='', description='uncharacterized LOC105375909', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77378702, end_position_on_the_genomic_accession=77418744, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  619488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619488, CurrentID=0, Status='live', Symbol='SCZD12', Aliases='', description='schizophrenia 12', other_designations='schizophrenia susceptibility locus, chromosome 1p-related', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608543, no_hdr0=''),
  102724577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724577, CurrentID=0, Status='live', Symbol='LOC102724577', Aliases='', description='uncharacterized LOC102724577', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124388434, end_position_on_the_genomic_accession=124392640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724579, CurrentID=0, Status='live', Symbol='LOC102724579', Aliases='', description='uncharacterized LOC102724579', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85213610, end_position_on_the_genomic_accession=85218191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724580, CurrentID=0, Status='live', Symbol='LOC102724580', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40497865, end_position_on_the_genomic_accession=40635468, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  619493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619493, CurrentID=0, Status='live', Symbol='PAND2', Aliases='', description='panic disorder 2', other_designations='panic disorder susceptibility locus, chromosome 9q-related', map_location='9q31', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607853, no_hdr0=''),
  619496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619496, CurrentID=0, Status='live', Symbol='CMT2H', Aliases='', description='Charcot-Marie-Tooth disease, axonal, type 2H', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607731, no_hdr0=''),
  100422053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422053, CurrentID=0, Status='live', Symbol='LOC100422053', Aliases='', description='intraflagellar transport 57 homolog (Chlamydomonas) pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51687533, end_position_on_the_genomic_accession=51688209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  619498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619498, CurrentID=0, Status='live', Symbol='SNORD74', Aliases='U74, Z18', description='small nucleolar RNA, C/D box 74', other_designations='C/D box snoRNA U74|RNA, U74 small nucleolar', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173867674, end_position_on_the_genomic_accession=173867745, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619499, CurrentID=0, Status='live', Symbol='SNORA27', Aliases='ACA27', description='small nucleolar RNA, H/ACA box 27', other_designations='ACA27 small nucleolar RNA|ACA27 snoRNA', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27255401, end_position_on_the_genomic_accession=27255526, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619501, CurrentID=0, Status='live', Symbol='HCC', Aliases='', description='thyroid carcinoma, Hurthle cell', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607464, no_hdr0=''),
  619503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619503, CurrentID=0, Status='live', Symbol='FIQTL1', Aliases='FSI', description='fasting glucose and specific insulin levels', other_designations='', map_location='6q22-q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606035, no_hdr0=''),
  619505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619505, CurrentID=0, Status='live', Symbol='SNORA21', Aliases='ACA21', description='small nucleolar RNA, H/ACA box 21', other_designations='ACA21 small nucleolar RNA|ACA21 snoRNA', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38852863, end_position_on_the_genomic_accession=38852995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724594, CurrentID=0, Status='live', Symbol='U2AF1L5', Aliases='', description='U2 small nuclear RNA auxiliary factor 1-like 5', other_designations='splicing factor U2AF 35 kDa subunit', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6484623, end_position_on_the_genomic_accession=6499284, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102724596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724596, CurrentID=0, Status='live', Symbol='LOC102724596', Aliases='', description='uncharacterized LOC102724596', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49361165, end_position_on_the_genomic_accession=49380094, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  619509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619509, CurrentID=0, Status='live', Symbol='WTRS', Aliases='', description='Wittwer syndrome', other_designations='', map_location='Xp22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300421, no_hdr0=''),
  619510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619510, CurrentID=0, Status='live', Symbol='AMCX5', Aliases='', description='arthrogryposis, X-linked, type 5', other_designations='', map_location='Xq23-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300158, no_hdr0=''),
  619511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619511, CurrentID=0, Status='live', Symbol='MHB', Aliases='', description='myopathy, hyaline body, autosomal recessive', other_designations='', map_location='3p22.2-p21.32', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=255160, no_hdr0=''),
  102724601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724601, CurrentID=0, Status='live', Symbol='LOC102724601', Aliases='', description='uncharacterized LOC102724601', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174121638, end_position_on_the_genomic_accession=174159287, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724602, CurrentID=0, Status='live', Symbol='LOC102724602', Aliases='', description='COMM domain-containing protein 6-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161411463, end_position_on_the_genomic_accession=161418078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724603, CurrentID=0, Status='live', Symbol='LOC102724603', Aliases='', description='uncharacterized LOC102724603', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47708508, end_position_on_the_genomic_accession=47712375, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724604, CurrentID=0, Status='live', Symbol='LOC102724604', Aliases='', description='uncharacterized LOC102724604', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181421163, end_position_on_the_genomic_accession=181442486, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724605, CurrentID=0, Status='live', Symbol='UBE2V1P10', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 10', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene|ubiquitin-conjugating enzyme E2 variant 1 pseudogene 10', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85156267, end_position_on_the_genomic_accession=85160026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  619518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619518, CurrentID=0, Status='live', Symbol='SSBP3-AS1', Aliases='C1orf191, MST128, MSTP128', description='SSBP3 antisense RNA 1', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54236443, end_position_on_the_genomic_accession=54239063, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619519, CurrentID=0, Status='live', Symbol='TWF1P1', Aliases='A6ps1, PTK9P, PTK9P1', description='twinfilin 1 pseudogene 1', other_designations='PTK9 protein tyrosine kinase 9 pseudogene|protein tyrosine kinase 9 pseudogene 1|twinfilin, actin-binding protein, homolog 1 pseudogene', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29201560, end_position_on_the_genomic_accession=29204545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724608, CurrentID=0, Status='live', Symbol='LOC102724608', Aliases='', description='uncharacterized LOC102724608', other_designations='CTA-384D8.34', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50541414, end_position_on_the_genomic_accession=50543013, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724610, CurrentID=0, Status='live', Symbol='LOC102724610', Aliases='', description='putative vomeronasal receptor-like protein 4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21464753, end_position_on_the_genomic_accession=21498922, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724612, CurrentID=0, Status='live', Symbol='LOC102724612', Aliases='', description='uncharacterized LOC102724612', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63465849, end_position_on_the_genomic_accession=63475482, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724615, CurrentID=0, Status='live', Symbol='LOC102724615', Aliases='', description='uncharacterized LOC102724615', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13542043, end_position_on_the_genomic_accession=13543623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375916, CurrentID=0, Status='live', Symbol='LOC105375916', Aliases='', description='uncharacterized LOC105375916', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79584170, end_position_on_the_genomic_accession=79638166, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724618, CurrentID=0, Status='live', Symbol='LOC102724618', Aliases='', description='uncharacterized LOC102724618', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31996895, end_position_on_the_genomic_accession=32017487, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724621, CurrentID=0, Status='live', Symbol='LOC102724621', Aliases='', description='cell division cycle protein 27 homolog pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724623, CurrentID=0, Status='live', Symbol='LOC102724623', Aliases='', description='uncharacterized LOC102724623', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63856943, end_position_on_the_genomic_accession=64368558, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  619536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619536, CurrentID=0, Status='live', Symbol='BPPV', Aliases='', description='vestibulopathy, familial', other_designations='benign paroxysmal positional vertigo', map_location='6q', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=193007, no_hdr0=''),
  619537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619537, CurrentID=0, Status='live', Symbol='VV', Aliases='', description='varicose veins', other_designations='', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=192200, no_hdr0=''),
  619538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619538, CurrentID=0, Status='live', Symbol='OMS', Aliases='COME/ROM', description='otitis media, susceptibility to', other_designations='chronic/recurrent otitis media', map_location='10q26.3;19q13.42-q13.43;3p25.3', chromosome='10, 19, 3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=166760, no_hdr0=''),
  619539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619539, CurrentID=0, Status='live', Symbol='LOC619539', Aliases='', description='lentiginosis, inherited patterned', other_designations='familial generalized lentiginosis', map_location='4q21.1-q22.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151001, no_hdr0=''),
  619540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619540, CurrentID=0, Status='live', Symbol='LOC619540', Aliases='', description='keratosis palmoplantaris papulosa', other_designations='punctate palmoplantar keratoderma type I', map_location='15q22-q24', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=148600, no_hdr0=''),
  105375918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375918, CurrentID=0, Status='live', Symbol='LOC105375918', Aliases='', description='uncharacterized LOC105375918', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80082073, end_position_on_the_genomic_accession=80103863, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724631, CurrentID=0, Status='live', Symbol='POTEB3', Aliases='POTE-15, POTEB', description='POTE ankyrin domain family member B3', other_designations='ANKRD26-like family B member 1|POTE ankyrin domain family, member B-like|POTE ankyrin domain family, member B3|prostate, ovary, testis-expressed protein on chromosome 15', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21408964, end_position_on_the_genomic_accession=21440429, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  102724632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724632, CurrentID=0, Status='live', Symbol='LOC102724632', Aliases='', description='uncharacterized LOC102724632', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88903697, end_position_on_the_genomic_accession=88908108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724634, CurrentID=0, Status='live', Symbol='LOC102724634', Aliases='', description='uncharacterized LOC102724634', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13471715, end_position_on_the_genomic_accession=13484018, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375919, CurrentID=0, Status='live', Symbol='LOC105375919', Aliases='', description='uncharacterized LOC105375919', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80079666, end_position_on_the_genomic_accession=80081867, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724636, CurrentID=0, Status='live', Symbol='LOC102724636', Aliases='', description='uncharacterized LOC102724636', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6499399, end_position_on_the_genomic_accession=6536357, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724637, CurrentID=0, Status='live', Symbol='LOC102724637', Aliases='', description='uncharacterized LOC102724637', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89900664, end_position_on_the_genomic_accession=89963978, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724638, CurrentID=0, Status='live', Symbol='LOC102724638', Aliases='', description='uncharacterized LOC102724638', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22877536, end_position_on_the_genomic_accession=22881125, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724639, CurrentID=0, Status='live', Symbol='LOC102724639', Aliases='', description='keratin, type I cytoskeletal 18 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  619552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619552, CurrentID=0, Status='live', Symbol='MIR483', Aliases='MIRN483, hsa-mir-483, mir-483', description='microRNA 483', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2134134, end_position_on_the_genomic_accession=2134209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619553, CurrentID=0, Status='live', Symbol='MIR484', Aliases='MIRN484, hsa-mir-484, mir-484', description='microRNA 484', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15643294, end_position_on_the_genomic_accession=15643372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619554, CurrentID=0, Status='live', Symbol='MIR486-1', Aliases='MIR486, MIRN486, hsa-mir-486, hsa-mir-486-1, mir-486-1', description='microRNA 486-1', other_designations='microRNA 486', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41660441, end_position_on_the_genomic_accession=41660508, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619555, CurrentID=0, Status='live', Symbol='MIR487A', Aliases='MIRN487, MIRN487A, hsa-mir-487, mir-487a', description='microRNA 487a', other_designations='hsa-mir-487a', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101052446, end_position_on_the_genomic_accession=101052525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619556, CurrentID=0, Status='live', Symbol='MIR455', Aliases='MIRN455, hsa-mir-455, mir-455', description='microRNA 455', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114209434, end_position_on_the_genomic_accession=114209529, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724645, CurrentID=0, Status='live', Symbol='LOC102724645', Aliases='', description='uncharacterized LOC102724645', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724646, CurrentID=0, Status='live', Symbol='LOC102724646', Aliases='', description='smoothelin-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111872732, end_position_on_the_genomic_accession=111877015, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724648, CurrentID=0, Status='live', Symbol='LOC102724648', Aliases='', description='uncharacterized LOC102724648', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56156520, end_position_on_the_genomic_accession=56160704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724649, CurrentID=0, Status='live', Symbol='LOC102724649', Aliases='', description='uncharacterized LOC102724649', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176521216, end_position_on_the_genomic_accession=176552695, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619562, CurrentID=0, Status='live', Symbol='SNORA3A', Aliases='ACA3, SNORA3, SNORA45A', description='small nucleolar RNA, H/ACA box 3A', other_designations='ACA3 small nucleolar RNA|ACA3 snoRNA|small nucleolar RNA, H/ACA box 3|small nucleolar RNA, H/ACA box 45A', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8684227, end_position_on_the_genomic_accession=8684356, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619563, CurrentID=0, Status='live', Symbol='SNORA7A', Aliases='ACA7', description='small nucleolar RNA, H/ACA box 7A', other_designations='ACA7 small nucleolar RNA|ACA7 snoRNA', map_location='3p25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12840312, end_position_on_the_genomic_accession=12840450, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619564, CurrentID=0, Status='live', Symbol='SNORD72', Aliases='HBII-240', description='small nucleolar RNA, C/D box 72', other_designations='C/D box snoRNA HBII-240|HBII-240 small nucleolar RNA', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40832656, end_position_on_the_genomic_accession=40832735, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619565, CurrentID=0, Status='live', Symbol='SNORA52', Aliases='ACA52', description='small nucleolar RNA, H/ACA box 52', other_designations='ACA52 small nucleolar RNA|ACA52 snoRNA', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=811681, end_position_on_the_genomic_accession=811814, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724654, CurrentID=0, Status='live', Symbol='LOC102724654', Aliases='', description='uncharacterized LOC102724654', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=21852638, end_position_on_the_genomic_accession=21855921, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  619567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619567, CurrentID=0, Status='live', Symbol='SNORD2', Aliases='R39B, SNR39B', description='small nucleolar RNA, C/D box 2', other_designations='snR39B small nucleolar RNA|snR39B snoRNA', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186784796, end_position_on_the_genomic_accession=186784865, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619568, CurrentID=0, Status='live', Symbol='SNORA4', Aliases='ACA4', description='small nucleolar RNA, H/ACA box 4', other_designations='ACA4 small nucleolar RNA|ACA4 snoRNA', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186787613, end_position_on_the_genomic_accession=186787749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619569, CurrentID=0, Status='live', Symbol='SNORA41', Aliases='ACA41', description='small nucleolar RNA, H/ACA box 41', other_designations='ACA41 small nucleolar RNA|ACA41 snoRNA', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206162228, end_position_on_the_genomic_accession=206162359, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  619570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619570, CurrentID=0, Status='live', Symbol='SNORD95', Aliases='U95', description='small nucleolar RNA, C/D box 95', other_designations='C/D box snoRNA U95', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181243314, end_position_on_the_genomic_accession=181243376, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=619571, CurrentID=0, Status='live', Symbol='SNORD96A', Aliases='U96A', description='small nucleolar RNA, C/D box 96A', other_designations='C/D box snoRNA U96a', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181241818, end_position_on_the_genomic_accession=181241889, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724660, CurrentID=0, Status='live', Symbol='LOC102724660', Aliases='', description='uncharacterized LOC102724660', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128450785, end_position_on_the_genomic_accession=128469400, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724661, CurrentID=0, Status='live', Symbol='LOC102724661', Aliases='', description='uncharacterized LOC102724661', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177351586, end_position_on_the_genomic_accession=177366458, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724663, CurrentID=0, Status='live', Symbol='LOC102724663', Aliases='', description='uncharacterized LOC102724663', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81279189, end_position_on_the_genomic_accession=81312422, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375924, CurrentID=0, Status='live', Symbol='LOC105375924', Aliases='', description='uncharacterized LOC105375924', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81153968, end_position_on_the_genomic_accession=81226927, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724668, CurrentID=0, Status='live', Symbol='DPY19L1P2', Aliases='THC1072510', description='DPY19L1 pseudogene 2', other_designations='dpy-19-like 1 pseudogene 2|probable C-mannosyltransferase DPY19L1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32834792, end_position_on_the_genomic_accession=32838704, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102724669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724669, CurrentID=0, Status='live', Symbol='LOC102724669', Aliases='', description='uncharacterized LOC102724669', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89157890, end_position_on_the_genomic_accession=89158152, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724670, CurrentID=0, Status='live', Symbol='LOC102724670', Aliases='', description='uncharacterized LOC102724670', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90647715, end_position_on_the_genomic_accession=90687843, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724672, CurrentID=0, Status='live', Symbol='LOC102724672', Aliases='', description='uncharacterized LOC102724672', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148219636, end_position_on_the_genomic_accession=148230208, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81419, CurrentID=0, Status='live', Symbol='OR7E99P', Aliases='', description='olfactory receptor family 7 subfamily E member 99 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 99 pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4156479, end_position_on_the_genomic_accession=4157503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724676, CurrentID=0, Status='live', Symbol='LOC102724676', Aliases='', description='uncharacterized LOC102724676', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=15182242, end_position_on_the_genomic_accession=15197810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375926, CurrentID=0, Status='live', Symbol='LOC105375926', Aliases='', description='uncharacterized LOC105375926', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81357097, end_position_on_the_genomic_accession=81360953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724678, CurrentID=0, Status='live', Symbol='LINC01423', Aliases='', description='long intergenic non-protein coding RNA 1423', other_designations='', map_location='21q22', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38326359, end_position_on_the_genomic_accession=38346076, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724679, CurrentID=0, Status='live', Symbol='LOC102724679', Aliases='', description='uncharacterized LOC102724679', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23326133, end_position_on_the_genomic_accession=23342583, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724680, CurrentID=0, Status='live', Symbol='LOC102724680', Aliases='', description='uncharacterized LOC102724680', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=84912837, end_position_on_the_genomic_accession=84992723, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724682, CurrentID=0, Status='live', Symbol='LOC102724682', Aliases='', description='uncharacterized LOC102724682', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99940998, end_position_on_the_genomic_accession=99949285, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375927, CurrentID=0, Status='live', Symbol='LOC105375927', Aliases='', description='uncharacterized LOC105375927', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81399720, end_position_on_the_genomic_accession=81406241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724684, CurrentID=0, Status='live', Symbol='LOC102724684', Aliases='', description='uncharacterized LOC102724684', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98868988, end_position_on_the_genomic_accession=98877335, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724687, CurrentID=0, Status='live', Symbol='LOC102724687', Aliases='', description='uncharacterized LOC102724687', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66203725, end_position_on_the_genomic_accession=66432418, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724689, CurrentID=0, Status='live', Symbol='LOC102724689', Aliases='', description='uncharacterized LOC102724689', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87344930, end_position_on_the_genomic_accession=87348776, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724690, CurrentID=0, Status='live', Symbol='LOC102724690', Aliases='', description='uncharacterized LOC102724690', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41649665, end_position_on_the_genomic_accession=41653021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724691, CurrentID=0, Status='live', Symbol='LOC102724691', Aliases='', description='uncharacterized LOC102724691', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104936241, end_position_on_the_genomic_accession=105038496, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724694, CurrentID=0, Status='live', Symbol='LOC102724694', Aliases='', description='cyclin-Y-like protein 2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375929, CurrentID=0, Status='live', Symbol='LOC105375929', Aliases='', description='uncharacterized LOC105375929', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=82108567, end_position_on_the_genomic_accession=82160534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724698, CurrentID=0, Status='live', Symbol='LOC102724698', Aliases='', description='uncharacterized LOC102724698', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56063167, end_position_on_the_genomic_accession=56097451, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102724699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724699, CurrentID=0, Status='live', Symbol='LOC102724699', Aliases='', description='uncharacterized LOC102724699', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186810880, end_position_on_the_genomic_accession=186825521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724700, CurrentID=0, Status='live', Symbol='LOC102724700', Aliases='', description='uncharacterized LOC102724700', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151887543, end_position_on_the_genomic_accession=151910184, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724701, CurrentID=0, Status='live', Symbol='LOC102724701', Aliases='CH507-145C22.1, CH507-154B10.1', description='uncharacterized LOC102724701', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6630190, end_position_on_the_genomic_accession=6670705, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  102724702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724702, CurrentID=0, Status='live', Symbol='LOC102724702', Aliases='', description='uncharacterized LOC102724702', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9511774, end_position_on_the_genomic_accession=9517795, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724703, CurrentID=0, Status='live', Symbol='LOC102724703', Aliases='', description='uncharacterized LOC102724703', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724706, CurrentID=0, Status='live', Symbol='LOC102724706', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40942175, end_position_on_the_genomic_accession=40945143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375931, CurrentID=0, Status='live', Symbol='LOC105375931', Aliases='', description='uncharacterized LOC105375931', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=82442832, end_position_on_the_genomic_accession=82632962, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724708, CurrentID=0, Status='live', Symbol='LOC102724708', Aliases='', description='spidroin-1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67343647, end_position_on_the_genomic_accession=67375213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724710, CurrentID=0, Status='live', Symbol='LOC102724710', Aliases='', description='uncharacterized LOC102724710', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92565443, end_position_on_the_genomic_accession=92655494, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724711, CurrentID=0, Status='live', Symbol='LOC102724711', Aliases='', description='uncharacterized LOC102724711', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6958639, end_position_on_the_genomic_accession=6994175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724712, CurrentID=0, Status='live', Symbol='LOC102724712', Aliases='', description='uncharacterized LOC102724712', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83286466, end_position_on_the_genomic_accession=83292150, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724714, CurrentID=0, Status='live', Symbol='LOC102724714', Aliases='', description='uncharacterized LOC102724714', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207662384, end_position_on_the_genomic_accession=207678855, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724715, CurrentID=0, Status='live', Symbol='LOC102724715', Aliases='', description='uncharacterized LOC102724715', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32877333, end_position_on_the_genomic_accession=32906609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375933, CurrentID=0, Status='live', Symbol='LOC105375933', Aliases='', description='uncharacterized LOC105375933', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85086561, end_position_on_the_genomic_accession=85109574, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724720, CurrentID=0, Status='live', Symbol='LOC102724720', Aliases='', description='uncharacterized LOC102724720', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95841753, end_position_on_the_genomic_accession=95849274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724723, CurrentID=0, Status='live', Symbol='LOC102724723', Aliases='', description='uncharacterized LOC102724723', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34939374, end_position_on_the_genomic_accession=34948608, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  256190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256190, CurrentID=0, Status='live', Symbol='OR4C50P', Aliases='', description='olfactory receptor family 4 subfamily C member 50 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 50 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54591480, end_position_on_the_genomic_accession=54592406, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724726, CurrentID=0, Status='live', Symbol='LOC102724726', Aliases='', description='coxsackievirus and adenovirus receptor-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6789094, end_position_on_the_genomic_accession=6818263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724727, CurrentID=0, Status='live', Symbol='LOC102724727', Aliases='', description='protein FRG1B-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724728, CurrentID=0, Status='live', Symbol='LOC102724728', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724730, CurrentID=0, Status='live', Symbol='LOC102724730', Aliases='', description='uncharacterized LOC102724730', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89833423, end_position_on_the_genomic_accession=89835956, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375935, CurrentID=0, Status='live', Symbol='LOC105375935', Aliases='', description='uncharacterized LOC105375935', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85243666, end_position_on_the_genomic_accession=85255919, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724732, CurrentID=0, Status='live', Symbol='LOC102724732', Aliases='', description='uncharacterized LOC102724732', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55580597, end_position_on_the_genomic_accession=55589913, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724733, CurrentID=0, Status='live', Symbol='LOC102724733', Aliases='', description='uncharacterized LOC102724733', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57735552, end_position_on_the_genomic_accession=57738674, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724734, CurrentID=0, Status='live', Symbol='LOC102724734', Aliases='', description='uncharacterized LOC102724734', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=19047399, end_position_on_the_genomic_accession=19050391, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724736, CurrentID=0, Status='live', Symbol='LOC102724736', Aliases='', description='uncharacterized LOC102724736', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22662395, end_position_on_the_genomic_accession=22674065, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724737, CurrentID=0, Status='live', Symbol='LOC102724737', Aliases='', description='40S ribosomal protein S8 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21709595, end_position_on_the_genomic_accession=21710272, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724738, CurrentID=0, Status='live', Symbol='LOC102724738', Aliases='', description='tricarboxylate transport protein, mitochondrial pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63346800, end_position_on_the_genomic_accession=63350912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724739, CurrentID=0, Status='live', Symbol='LOC102724739', Aliases='', description='uncharacterized LOC102724739', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41787105, end_position_on_the_genomic_accession=41822008, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724740, CurrentID=0, Status='live', Symbol='LOC102724740', Aliases='', description='uncharacterized LOC102724740', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38988389, end_position_on_the_genomic_accession=39019256, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724742, CurrentID=0, Status='live', Symbol='PRAMEF28', Aliases='', description='PRAME family member 28', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13030809, end_position_on_the_genomic_accession=13032199, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724744, CurrentID=0, Status='live', Symbol='LOC102724744', Aliases='', description='uncharacterized LOC102724744', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106701690, end_position_on_the_genomic_accession=106781111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724745, CurrentID=0, Status='live', Symbol='LOC102724745', Aliases='', description='uncharacterized LOC102724745', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58658588, end_position_on_the_genomic_accession=58669329, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724746, CurrentID=0, Status='live', Symbol='LOC102724746', Aliases='', description='uncharacterized LOC102724746', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121996345, end_position_on_the_genomic_accession=121998491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724748, CurrentID=0, Status='live', Symbol='LOC102724748', Aliases='', description='uncharacterized LOC102724748', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96776558, end_position_on_the_genomic_accession=96785791, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375938, CurrentID=0, Status='live', Symbol='LOC105375938', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60909752, end_position_on_the_genomic_accession=60910855, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724750, CurrentID=0, Status='live', Symbol='LOC102724750', Aliases='', description='uncharacterized protein C2orf27-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724751, CurrentID=0, Status='live', Symbol='LOC102724751', Aliases='', description='putative BCR-like protein 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724754, CurrentID=0, Status='live', Symbol='LOC102724754', Aliases='', description='uncharacterized LOC102724754', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178936757, end_position_on_the_genomic_accession=178938904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375939, CurrentID=0, Status='live', Symbol='LOC105375939', Aliases='', description='uncharacterized LOC105375939', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724757, CurrentID=0, Status='live', Symbol='LOC102724757', Aliases='', description='uncharacterized LOC102724757', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39406639, end_position_on_the_genomic_accession=39428408, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724760, CurrentID=0, Status='live', Symbol='LOC102724760', Aliases='', description='Ig heavy chain V-I region V35-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21717665, end_position_on_the_genomic_accession=21718311, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724761, CurrentID=0, Status='live', Symbol='LOC102724761', Aliases='', description='uncharacterized LOC102724761', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104906209, end_position_on_the_genomic_accession=104922762, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724762, CurrentID=0, Status='live', Symbol='LOC102724762', Aliases='', description='uncharacterized LOC102724762', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107903963, end_position_on_the_genomic_accession=107904243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724763, CurrentID=0, Status='live', Symbol='LOC102724763', Aliases='', description='uncharacterized LOC102724763', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65174977, end_position_on_the_genomic_accession=65193497, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724765, CurrentID=0, Status='live', Symbol='LOC102724765', Aliases='', description='uncharacterized LOC102724765', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25014971, end_position_on_the_genomic_accession=25036418, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724766, CurrentID=0, Status='live', Symbol='LOC102724766', Aliases='', description='uncharacterized LOC102724766', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58343021, end_position_on_the_genomic_accession=58431352, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724768, CurrentID=0, Status='live', Symbol='LOC102724768', Aliases='', description='uncharacterized LOC102724768', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59013680, end_position_on_the_genomic_accession=59065845, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724770, CurrentID=0, Status='live', Symbol='LOC102724770', Aliases='', description='protein DGCR6', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374309, CurrentID=0, Status='live', Symbol='LOC105374309', Aliases='', description='uncharacterized LOC105374309', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197477986, end_position_on_the_genomic_accession=197479651, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375942, CurrentID=0, Status='live', Symbol='LOC105375942', Aliases='', description='uncharacterized LOC105375942', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131973, end_position_on_the_genomic_accession=133872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724774, CurrentID=0, Status='live', Symbol='LINC01594', Aliases='TCONS_l2_00015893', description='long intergenic non-protein coding RNA 1594', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108048498, end_position_on_the_genomic_accession=108232868, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  102724775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724775, CurrentID=0, Status='live', Symbol='LOC102724775', Aliases='', description='uncharacterized LOC102724775', other_designations='CTD-2005H7.2', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86763060, end_position_on_the_genomic_accession=86765464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724776, CurrentID=0, Status='live', Symbol='LOC102724776', Aliases='', description='uncharacterized LOC102724776', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155354216, end_position_on_the_genomic_accession=155360461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724777, CurrentID=0, Status='live', Symbol='LOC102724777', Aliases='', description='putative uncharacterized protein FLJ92257', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63763951, end_position_on_the_genomic_accession=63764555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724778, CurrentID=0, Status='live', Symbol='LOC102724778', Aliases='', description='uncharacterized LOC102724778', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59322135, end_position_on_the_genomic_accession=59332607, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375943, CurrentID=0, Status='live', Symbol='LOC105375943', Aliases='', description='uncharacterized LOC105375943', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=190714, end_position_on_the_genomic_accession=199142, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724780, CurrentID=0, Status='live', Symbol='LOC102724780', Aliases='', description='uncharacterized LOC102724780', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67786812, end_position_on_the_genomic_accession=67802860, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724782, CurrentID=0, Status='live', Symbol='LOC102724782', Aliases='', description='uncharacterized LOC102724782', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89525664, end_position_on_the_genomic_accession=89598063, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102724784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724784, CurrentID=0, Status='live', Symbol='LOC102724784', Aliases='', description='uncharacterized LOC102724784', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8768778, end_position_on_the_genomic_accession=8810276, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375944, CurrentID=0, Status='live', Symbol='LOC105375944', Aliases='', description='uncharacterized LOC105375944', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=453775, end_position_on_the_genomic_accession=470124, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102724787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724787, CurrentID=0, Status='live', Symbol='LOC102724787', Aliases='', description='unconventional myosin-Vb pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724788, CurrentID=0, Status='live', Symbol='LOC102724788', Aliases='', description='proline dehydrogenase 1, mitochondrial', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724790, CurrentID=0, Status='live', Symbol='LOC102724790', Aliases='', description='uncharacterized LOC102724790', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61181126, end_position_on_the_genomic_accession=61212432, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375945, CurrentID=0, Status='live', Symbol='LOC105375945', Aliases='', description='uncharacterized LOC105375945', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=355801, end_position_on_the_genomic_accession=364395, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724793, CurrentID=0, Status='live', Symbol='LOC102724793', Aliases='', description='uncharacterized LOC102724793', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125584486, end_position_on_the_genomic_accession=125603252, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724800, CurrentID=0, Status='live', Symbol='LOC102724800', Aliases='', description='uncharacterized LOC102724800', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41767186, end_position_on_the_genomic_accession=41774864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724801, CurrentID=0, Status='live', Symbol='LOC102724801', Aliases='', description='uncharacterized LOC102724801', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26108552, end_position_on_the_genomic_accession=26154901, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724802, CurrentID=0, Status='live', Symbol='LOC102724802', Aliases='', description='uncharacterized LOC102724802', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60856535, end_position_on_the_genomic_accession=60879565, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375947, CurrentID=0, Status='live', Symbol='LOC105375947', Aliases='', description='uncharacterized LOC105375947', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=500662, end_position_on_the_genomic_accession=504242, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724804, CurrentID=0, Status='live', Symbol='LOC102724804', Aliases='', description='uncharacterized LOC102724804', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96371926, end_position_on_the_genomic_accession=96387438, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724805, CurrentID=0, Status='live', Symbol='LOC102724805', Aliases='', description='uncharacterized LOC102724805', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88634296, end_position_on_the_genomic_accession=88691476, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724808, CurrentID=0, Status='live', Symbol='LOC102724808', Aliases='', description='uncharacterized LOC102724808', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193647195, end_position_on_the_genomic_accession=193683261, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375948, CurrentID=0, Status='live', Symbol='LOC105375948', Aliases='', description='uncharacterized LOC105375948', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=746996, end_position_on_the_genomic_accession=763189, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724810, CurrentID=0, Status='live', Symbol='LOC102724810', Aliases='', description='uncharacterized LOC102724810', other_designations='CTD-2007H13.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98969475, end_position_on_the_genomic_accession=99003936, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724813, CurrentID=0, Status='live', Symbol='LOC102724813', Aliases='', description='protein FRG1B', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724814, CurrentID=0, Status='live', Symbol='LOC102724814', Aliases='', description='uncharacterized LOC102724814', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23561097, end_position_on_the_genomic_accession=23568076, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724816, CurrentID=0, Status='live', Symbol='LOC102724816', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129787424, end_position_on_the_genomic_accession=129791847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724817, CurrentID=0, Status='live', Symbol='LOC102724817', Aliases='', description='uncharacterized LOC102724817', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72614228, end_position_on_the_genomic_accession=72616367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724818, CurrentID=0, Status='live', Symbol='LOC102724818', Aliases='', description='uncharacterized LOC102724818', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42834175, end_position_on_the_genomic_accession=42840920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724819, CurrentID=0, Status='live', Symbol='LOC102724819', Aliases='', description='uncharacterized LOC102724819', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109260659, end_position_on_the_genomic_accession=109272527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724820, CurrentID=0, Status='live', Symbol='LANCL1-AS1', Aliases='LANCLOT', description='LANCL1 antisense RNA 1', other_designations='LANCL1 opposite transcript', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210324712, end_position_on_the_genomic_accession=210470333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724821, CurrentID=0, Status='live', Symbol='LOC102724821', Aliases='', description='uncharacterized LOC102724821', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71183701, end_position_on_the_genomic_accession=71186523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724822, CurrentID=0, Status='live', Symbol='LOC102724822', Aliases='', description='carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 2 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6991360, end_position_on_the_genomic_accession=6993499, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724823, CurrentID=0, Status='live', Symbol='LOC102724823', Aliases='', description='inactive gamma-glutamyltranspeptidase 2', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371427, CurrentID=0, Status='live', Symbol='LOC105371427', Aliases='', description='uncharacterized LOC105371427', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150048513, end_position_on_the_genomic_accession=150067605, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724826, CurrentID=0, Status='live', Symbol='ZNF337-AS1', Aliases='', description='ZNF337 antisense RNA 1', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25624045, end_position_on_the_genomic_accession=25678074, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102724827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724827, CurrentID=0, Status='live', Symbol='MIATNB', Aliases='', description='MIAT neighbor (non-protein coding)', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26672843, end_position_on_the_genomic_accession=26780893, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724828, CurrentID=0, Status='live', Symbol='LOC102724828', Aliases='', description='39S ribosomal protein L23, mitochondrial-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724830, CurrentID=0, Status='live', Symbol='LOC102724830', Aliases='', description='uncharacterized LOC102724830', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184112570, end_position_on_the_genomic_accession=184122538, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724832, CurrentID=0, Status='live', Symbol='LOC102724832', Aliases='', description='uncharacterized LOC102724832', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73258895, end_position_on_the_genomic_accession=73369941, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102724834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724834, CurrentID=0, Status='live', Symbol='LOC102724834', Aliases='', description='uncharacterized LOC102724834', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90675299, end_position_on_the_genomic_accession=90680690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724835, CurrentID=0, Status='live', Symbol='LOC102724835', Aliases='', description='uncharacterized LOC102724835', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22178234, end_position_on_the_genomic_accession=22178602, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724838, CurrentID=0, Status='live', Symbol='LOC102724838', Aliases='', description='uncharacterized LOC102724838', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17345014, end_position_on_the_genomic_accession=17382440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375953, CurrentID=0, Status='live', Symbol='LOC105375953', Aliases='', description='uncharacterized LOC105375953', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1912498, end_position_on_the_genomic_accession=1918715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724843, CurrentID=0, Status='live', Symbol='LOC102724843', Aliases='', description='uncharacterized LOC102724843', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7048888, end_position_on_the_genomic_accession=7087230, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724844, CurrentID=0, Status='live', Symbol='LOC102724844', Aliases='', description='immunoglobulin superfamily member 3-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724845, CurrentID=0, Status='live', Symbol='LINC00596', Aliases='', description='long intergenic non-protein coding RNA 596', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23838474, end_position_on_the_genomic_accession=23933924, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724846, CurrentID=0, Status='live', Symbol='LOC102724846', Aliases='', description='uncharacterized LOC102724846', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184208211, end_position_on_the_genomic_accession=184213972, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724847, CurrentID=0, Status='live', Symbol='LOC102724847', Aliases='', description='uncharacterized LOC102724847', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49162683, end_position_on_the_genomic_accession=49170015, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724848, CurrentID=0, Status='live', Symbol='LOC102724848', Aliases='', description='uncharacterized LOC102724848', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109859494, end_position_on_the_genomic_accession=109877007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724849, CurrentID=0, Status='live', Symbol='LOC102724849', Aliases='', description='uncharacterized LOC102724849', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215453707, end_position_on_the_genomic_accession=215463871, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  401296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401296, CurrentID=0, Status='live', Symbol='LOC401296', Aliases='', description='uncharacterized LOC401296', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1692112, end_position_on_the_genomic_accession=1694437, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724851, CurrentID=0, Status='live', Symbol='LOC102724851', Aliases='', description='uncharacterized LOC102724851', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26670973, end_position_on_the_genomic_accession=26687841, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102724852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724852, CurrentID=0, Status='live', Symbol='LOC102724852', Aliases='', description='uncharacterized LOC102724852', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724853, CurrentID=0, Status='live', Symbol='NDUFA5P10', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 10', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 10|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 5-like', map_location='1p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109809560, end_position_on_the_genomic_accession=109811014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724855, CurrentID=0, Status='live', Symbol='CTD-2151A2.1', Aliases='', description='uncharacterized LOC102724855', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=99533990, end_position_on_the_genomic_accession=99577954, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724856, CurrentID=0, Status='live', Symbol='LOC102724856', Aliases='', description='uncharacterized LOC102724856', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13487577, end_position_on_the_genomic_accession=13493138, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724857, CurrentID=0, Status='live', Symbol='LOC102724857', Aliases='', description='uncharacterized LOC102724857', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184538781, end_position_on_the_genomic_accession=184542649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724858, CurrentID=0, Status='live', Symbol='LOC102724858', Aliases='', description='uncharacterized LOC102724858', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76140903, end_position_on_the_genomic_accession=76308493, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724859, CurrentID=0, Status='live', Symbol='LOC102724859', Aliases='', description='uncharacterized LOC102724859', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49282051, end_position_on_the_genomic_accession=49290362, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724861, CurrentID=0, Status='live', Symbol='LOC102724861', Aliases='', description='uncharacterized LOC102724861', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215615344, end_position_on_the_genomic_accession=215690922, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102724862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724862, CurrentID=0, Status='live', Symbol='TBC1D3I', Aliases='', description='TBC1 domain family member 3I', other_designations='TBC1 domain family member-like|TBC1 domain family, member 3I', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36253639, end_position_on_the_genomic_accession=36264603, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  102724863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724863, CurrentID=0, Status='live', Symbol='LOC102724863', Aliases='', description='uncharacterized LOC102724863', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64933244, end_position_on_the_genomic_accession=64938712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724865, CurrentID=0, Status='live', Symbol='LOC102724865', Aliases='', description='transcription factor E2F6 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724866, CurrentID=0, Status='live', Symbol='LOC102724866', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 5-like', map_location='16q12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47597693, end_position_on_the_genomic_accession=47599023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724868, CurrentID=0, Status='live', Symbol='LOC102724868', Aliases='', description='translationally-controlled tumor protein pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48632665, end_position_on_the_genomic_accession=48633482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724872, CurrentID=0, Status='live', Symbol='LOC102724872', Aliases='', description='uncharacterized LOC102724872', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96341202, end_position_on_the_genomic_accession=96353520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724874, CurrentID=0, Status='live', Symbol='LOC102724874', Aliases='', description='uncharacterized LOC102724874', other_designations='', map_location='8q21.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77450697, end_position_on_the_genomic_accession=77476331, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375959, CurrentID=0, Status='live', Symbol='LOC105375959', Aliases='', description='uncharacterized LOC105375959', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3194304, end_position_on_the_genomic_accession=3196066, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724877, CurrentID=0, Status='live', Symbol='LOC102724877', Aliases='', description='uncharacterized LOC102724877', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194070145, end_position_on_the_genomic_accession=194123474, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102724880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724880, CurrentID=0, Status='live', Symbol='LOC102724880', Aliases='', description='uncharacterized LOC102724880', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9151732, end_position_on_the_genomic_accession=9169478, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375960, CurrentID=0, Status='live', Symbol='LOC105375960', Aliases='', description='uncharacterized LOC105375960', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3663427, end_position_on_the_genomic_accession=3665653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724883, CurrentID=0, Status='live', Symbol='LOC102724883', Aliases='', description='uncharacterized LOC102724883', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130111219, end_position_on_the_genomic_accession=130120721, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724885, CurrentID=0, Status='live', Symbol='LOC102724885', Aliases='', description='uncharacterized LOC102724885', other_designations='CTD-2001C12.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100450912, end_position_on_the_genomic_accession=100535277, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724886, CurrentID=0, Status='live', Symbol='LOC102724886', Aliases='', description='uncharacterized LOC102724886', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724889, CurrentID=0, Status='live', Symbol='LOC102724889', Aliases='', description='uncharacterized LOC102724889', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28078790, end_position_on_the_genomic_accession=28081174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724890, CurrentID=0, Status='live', Symbol='LOC102724890', Aliases='', description='uncharacterized LOC102724890', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26809348, end_position_on_the_genomic_accession=26820703, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724892, CurrentID=0, Status='live', Symbol='LOC102724892', Aliases='', description='uncharacterized LOC102724892', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84781006, end_position_on_the_genomic_accession=84797644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724895, CurrentID=0, Status='live', Symbol='LOC102724895', Aliases='', description='uncharacterized LOC102724895', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70335508, end_position_on_the_genomic_accession=70337711, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724896, CurrentID=0, Status='live', Symbol='LOC102724896', Aliases='', description='uncharacterized LOC102724896', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118027414, end_position_on_the_genomic_accession=118045914, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724897, CurrentID=0, Status='live', Symbol='LOC102724897', Aliases='', description='uncharacterized LOC102724897', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185370666, end_position_on_the_genomic_accession=185375028, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724898, CurrentID=0, Status='live', Symbol='LOC102724898', Aliases='', description='uncharacterized LOC102724898', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18560969, end_position_on_the_genomic_accession=18563908, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375794, CurrentID=0, Status='live', Symbol='LOC105375794', Aliases='', description='uncharacterized LOC105375794', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142925293, end_position_on_the_genomic_accession=142927173, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724900, CurrentID=0, Status='live', Symbol='LOC102724900', Aliases='', description='uncharacterized LOC102724900', other_designations='CTA-929C8.8', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27331704, end_position_on_the_genomic_accession=27357550, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724902, CurrentID=0, Status='live', Symbol='LOC102724902', Aliases='', description='uncharacterized LOC102724902', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724903, CurrentID=0, Status='live', Symbol='LOC102724903', Aliases='', description='uncharacterized LOC102724903', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43297316, end_position_on_the_genomic_accession=43301360, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724904, CurrentID=0, Status='live', Symbol='LOC102724904', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=43126349, end_position_on_the_genomic_accession=43153730, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102724906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724906, CurrentID=0, Status='live', Symbol='LOC102724906', Aliases='', description='uncharacterized LOC102724906', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132409176, end_position_on_the_genomic_accession=132414183, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724907, CurrentID=0, Status='live', Symbol='LOC102724907', Aliases='', description='uncharacterized LOC102724907', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50734702, end_position_on_the_genomic_accession=50742815, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724908, CurrentID=0, Status='live', Symbol='LOC102724908', Aliases='', description='uncharacterized LOC102724908', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28606688, end_position_on_the_genomic_accession=28615229, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724909, CurrentID=0, Status='live', Symbol='LOC102724909', Aliases='', description='keratin, type I cytoskeletal 18 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7120580, end_position_on_the_genomic_accession=7125327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724913, CurrentID=0, Status='live', Symbol='LOC102724913', Aliases='', description='uncharacterized LOC102724913', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71732615, end_position_on_the_genomic_accession=71782226, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724916, CurrentID=0, Status='live', Symbol='LOC102724916', Aliases='', description='uncharacterized LOC102724916', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77918053, end_position_on_the_genomic_accession=77952510, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375966, CurrentID=0, Status='live', Symbol='LOC105375966', Aliases='', description='uncharacterized LOC105375966', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4540097, end_position_on_the_genomic_accession=4544846, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102724919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724919, CurrentID=0, Status='live', Symbol='LOC102724919', Aliases='', description='uncharacterized LOC102724919', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=186435161, end_position_on_the_genomic_accession=186470291, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724922, CurrentID=0, Status='live', Symbol='LOC102724922', Aliases='', description='putative aquaporin-7-like protein 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=43109811, end_position_on_the_genomic_accession=43127504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133087, CurrentID=0, Status='live', Symbol='VN1R18P', Aliases='', description='vomeronasal 1 receptor 18 pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47989625, end_position_on_the_genomic_accession=47990561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724926, CurrentID=0, Status='live', Symbol='LOC102724926', Aliases='', description='uncharacterized LOC102724926', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20525777, end_position_on_the_genomic_accession=20533886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102724927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724927, CurrentID=0, Status='live', Symbol='LOC102724927', Aliases='', description='uncharacterized LOC102724927', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3947609, end_position_on_the_genomic_accession=3950444, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724928, CurrentID=0, Status='live', Symbol='LOC102724928', Aliases='', description='putative cyclin-Y-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35030347, end_position_on_the_genomic_accession=35060345, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  102724929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724929, CurrentID=0, Status='live', Symbol='LOC102724929', Aliases='', description='uncharacterized LOC102724929', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118644338, end_position_on_the_genomic_accession=118732791, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  693131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693131, CurrentID=0, Status='live', Symbol='MIR548D2', Aliases='MIRN548D2', description='microRNA 548d-2', other_designations='hsa-mir-548d-2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67471489, end_position_on_the_genomic_accession=67471585, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102724933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724933, CurrentID=0, Status='live', Symbol='LOC102724933', Aliases='', description='uncharacterized LOC102724933', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93174366, end_position_on_the_genomic_accession=93181832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724934, CurrentID=0, Status='live', Symbol='LOC102724934', Aliases='', description='uncharacterized LOC102724934', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=29211071, end_position_on_the_genomic_accession=29274989, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724937, CurrentID=0, Status='live', Symbol='LOC102724937', Aliases='', description='angiogenic factor with G patch and FHA domains 1 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724938, CurrentID=0, Status='live', Symbol='LOC102724938', Aliases='', description='uncharacterized LOC102724938', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62267323, end_position_on_the_genomic_accession=62278860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724940, CurrentID=0, Status='live', Symbol='LOC102724940', Aliases='', description='uncharacterized LOC102724940', other_designations='CTD-2034I21.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50880173, end_position_on_the_genomic_accession=50901061, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724943, CurrentID=0, Status='live', Symbol='LOC102724943', Aliases='', description='uncharacterized LOC102724943', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6765910, end_position_on_the_genomic_accession=6777775, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371431, CurrentID=0, Status='live', Symbol='LOC105371431', Aliases='', description='uncharacterized LOC105371431', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150012435, end_position_on_the_genomic_accession=150058301, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724946, CurrentID=0, Status='live', Symbol='LOC102724946', Aliases='', description='uncharacterized LOC102724946', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43752482, end_position_on_the_genomic_accession=43839019, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724947, CurrentID=0, Status='live', Symbol='LOC102724947', Aliases='', description='uncharacterized LOC102724947', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134660776, end_position_on_the_genomic_accession=135004613, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  102724948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724948, CurrentID=0, Status='live', Symbol='LOC102724948', Aliases='', description='uncharacterized LOC102724948', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20524722, end_position_on_the_genomic_accession=20540137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724949, CurrentID=0, Status='live', Symbol='LOC102724949', Aliases='', description='uncharacterized LOC102724949', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=76709613, end_position_on_the_genomic_accession=76742522, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724950, CurrentID=0, Status='live', Symbol='LOC102724950', Aliases='', description='uncharacterized LOC102724950', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39655323, end_position_on_the_genomic_accession=39661014, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724951, CurrentID=0, Status='live', Symbol='LOC102724951', Aliases='', description='uncharacterized LOC102724951', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7430666, end_position_on_the_genomic_accession=7469008, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100133092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133092, CurrentID=0, Status='live', Symbol='LOC100133092', Aliases='', description='PAGE family member 4 pseudogene', other_designations='P antigen family, member 4 (prostate associated) pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64566796, end_position_on_the_genomic_accession=64567473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724954, CurrentID=0, Status='live', Symbol='LINC01032', Aliases='', description='long intergenic non-protein coding RNA 1032', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192935744, end_position_on_the_genomic_accession=192948257, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102724955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724955, CurrentID=0, Status='live', Symbol='LOC102724955', Aliases='', description='uncharacterized LOC102724955', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112674321, end_position_on_the_genomic_accession=112681167, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102724956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724956, CurrentID=0, Status='live', Symbol='LOC102724956', Aliases='', description='uncharacterized LOC102724956', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36264619, end_position_on_the_genomic_accession=36265447, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724957, CurrentID=0, Status='live', Symbol='LOC102724957', Aliases='', description='uncharacterized LOC102724957', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15623774, end_position_on_the_genomic_accession=15705368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102724958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724958, CurrentID=0, Status='live', Symbol='LOC102724958', Aliases='', description='uncharacterized LOC102724958', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28968468, end_position_on_the_genomic_accession=28970672, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375973, CurrentID=0, Status='live', Symbol='LOC105375973', Aliases='', description='uncharacterized LOC105375973', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=10685373, end_position_on_the_genomic_accession=10696693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102724960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724960, CurrentID=0, Status='live', Symbol='LOC102724960', Aliases='', description='uncharacterized LOC102724960', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94472309, end_position_on_the_genomic_accession=94539214, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724961, CurrentID=0, Status='live', Symbol='LOC102724961', Aliases='', description='double homeobox protein 4-like protein 4', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724962, CurrentID=0, Status='live', Symbol='LOC102724962', Aliases='', description='angiogenic factor with G patch and FHA domains 1-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724964, CurrentID=0, Status='live', Symbol='LOC102724964', Aliases='', description='uncharacterized LOC102724964', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44100837, end_position_on_the_genomic_accession=44104135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375974, CurrentID=0, Status='live', Symbol='LOC105375974', Aliases='', description='uncharacterized LOC105375974', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11253969, end_position_on_the_genomic_accession=11502598, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102724966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724966, CurrentID=0, Status='live', Symbol='LOC102724966', Aliases='', description='uncharacterized LOC102724966', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120048192, end_position_on_the_genomic_accession=120066999, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102724968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724968, CurrentID=0, Status='live', Symbol='LOC102724968', Aliases='', description='uncharacterized LOC102724968', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40573977, end_position_on_the_genomic_accession=40619185, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102724970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724970, CurrentID=0, Status='live', Symbol='LOC102724970', Aliases='', description='putative uncharacterized protein FLJ35883', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16535773, end_position_on_the_genomic_accession=16536956, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102724971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724971, CurrentID=0, Status='live', Symbol='LOC102724971', Aliases='', description='putative V-set and immunoglobulin domain-containing-like protein IGHV4OR15-8', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21742069, end_position_on_the_genomic_accession=21742898, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  693138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693138, CurrentID=0, Status='live', Symbol='MIR553', Aliases='MIRN553, hsa-mir-553', description='microRNA 553', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100281241, end_position_on_the_genomic_accession=100281308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102724977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724977, CurrentID=0, Status='live', Symbol='LOC102724977', Aliases='', description='uncharacterized LOC102724977', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724978, CurrentID=0, Status='live', Symbol='LOC102724978', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724979, CurrentID=0, Status='live', Symbol='LOC102724979', Aliases='', description='uncharacterized LOC102724979', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724980, CurrentID=0, Status='live', Symbol='LOC102724980', Aliases='', description='uncharacterized LOC102724980', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724981, CurrentID=0, Status='live', Symbol='LOC102724981', Aliases='', description='uncharacterized LOC102724981', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724983, CurrentID=0, Status='live', Symbol='LOC102724983', Aliases='', description='uncharacterized LOC102724983', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724984, CurrentID=0, Status='live', Symbol='LOC102724984', Aliases='', description='uncharacterized protein KIAA2013', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724985, CurrentID=0, Status='live', Symbol='LOC102724985', Aliases='', description='pyridoxal-dependent decarboxylase domain-containing protein 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724986, CurrentID=0, Status='live', Symbol='LOC102724986', Aliases='', description='uncharacterized LOC102724986', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724988, CurrentID=0, Status='live', Symbol='LOC102724988', Aliases='', description='uncharacterized LOC102724988', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724992, CurrentID=0, Status='live', Symbol='LOC102724992', Aliases='', description='metacaspase-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724993, CurrentID=0, Status='live', Symbol='LOC102724993', Aliases='', description='NPIP-like protein 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724994, CurrentID=0, Status='live', Symbol='LOC102724994', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724995, CurrentID=0, Status='live', Symbol='LOC102724995', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724996, CurrentID=0, Status='live', Symbol='LOC102724996', Aliases='', description='uncharacterized LOC102724996', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724998, CurrentID=0, Status='live', Symbol='LOC102724998', Aliases='', description='set1/Ash2 histone methyltransferase complex subunit ASH2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102724999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724999, CurrentID=0, Status='live', Symbol='LOC102724999', Aliases='', description='Ig lambda chain V-I region BL2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725000, CurrentID=0, Status='live', Symbol='LOC102725000', Aliases='', description='collagen alpha-2(V) chain-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133100, CurrentID=0, Status='live', Symbol='RPL6P8', Aliases='RPL6_8_430', description='ribosomal protein L6 pseudogene 8', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160579267, end_position_on_the_genomic_accession=160585787, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102725003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725003, CurrentID=0, Status='live', Symbol='LOC102725003', Aliases='', description='Ig lambda chain V-I region BL2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725004, CurrentID=0, Status='live', Symbol='LOC102725004', Aliases='', description='uncharacterized protein FLJ40521-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725009, CurrentID=0, Status='live', Symbol='LOC102725009', Aliases='', description='putative beta-glucuronidase-like protein MGC156142', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725013, CurrentID=0, Status='live', Symbol='LOC102725013', Aliases='', description='uncharacterized LOC102725013', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725015, CurrentID=0, Status='live', Symbol='LOC102725015', Aliases='', description='leukocyte immunoglobulin-like receptor subfamily B member 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725016, CurrentID=0, Status='live', Symbol='LOC102725016', Aliases='', description="serine/threonine-protein phosphatase 2A regulatory subunit B'' subunit beta", other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725017, CurrentID=0, Status='live', Symbol='LOC102725017', Aliases='', description='uncharacterized LOC102725017', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725018, CurrentID=0, Status='live', Symbol='LOC102725018', Aliases='', description='uncharacterized LOC102725018', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725019, CurrentID=0, Status='live', Symbol='LOC102725019', Aliases='', description='uncharacterized LOC102725019', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32718016, end_position_on_the_genomic_accession=32718827, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102725021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725021, CurrentID=0, Status='live', Symbol='LOC102725021', Aliases='', description='uncharacterized LOC102725021', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725022, CurrentID=0, Status='live', Symbol='LOC102725022', Aliases='', description='uncharacterized LOC102725022', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31214222, end_position_on_the_genomic_accession=31219598, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102725023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725023, CurrentID=0, Status='live', Symbol='LOC102725023', Aliases='', description='killer cell immunoglobulin-like receptor 2DS3 allele 0020101', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725024, CurrentID=0, Status='live', Symbol='LOC102725024', Aliases='', description='mucin-6-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375984, CurrentID=0, Status='live', Symbol='LOC105375984', Aliases='', description='uncharacterized LOC105375984', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18904789, end_position_on_the_genomic_accession=18926403, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102725028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725028, CurrentID=0, Status='live', Symbol='LOC102725028', Aliases='', description='uncharacterized LOC102725028', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725029, CurrentID=0, Status='live', Symbol='LOC102725029', Aliases='', description='leukocyte immunoglobulin-like receptor subfamily B member 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375985, CurrentID=0, Status='live', Symbol='LOC105375985', Aliases='', description='uncharacterized LOC105375985', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18889761, end_position_on_the_genomic_accession=18901103, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102725032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725032, CurrentID=0, Status='live', Symbol='LOC102725032', Aliases='', description='uncharacterized LOC102725032', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725034, CurrentID=0, Status='live', Symbol='LOC102725034', Aliases='', description='leukocyte immunoglobulin-like receptor subfamily B member 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725035, CurrentID=0, Status='live', Symbol='LOC102725035', Aliases='', description='leukocyte immunoglobulin-like receptor subfamily B member 3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  649458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649458, CurrentID=0, Status='live', Symbol='MPTX1', Aliases='MPTX', description='mucosal pentraxin 1 (pseudogene)', other_designations='amyloid P component, serum pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159276507, end_position_on_the_genomic_accession=159277182, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375987, CurrentID=0, Status='live', Symbol='LOC105375987', Aliases='', description='uncharacterized LOC105375987', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17902147, end_position_on_the_genomic_accession=17907280, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102725044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725044, CurrentID=0, Status='live', Symbol='LOC102725044', Aliases='', description='uncharacterized LOC102725044', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24271228, end_position_on_the_genomic_accession=24298653, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102725045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725045, CurrentID=0, Status='live', Symbol='LOC102725045', Aliases='', description='uncharacterized LOC102725045', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26598405, end_position_on_the_genomic_accession=26806980, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  102725046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725046, CurrentID=0, Status='live', Symbol='LOC102725046', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725048, CurrentID=0, Status='live', Symbol='LOC102725048', Aliases='', description='uncharacterized LOC102725048', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167241622, end_position_on_the_genomic_accession=167247465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133108, CurrentID=0, Status='live', Symbol='RPL13P10', Aliases='RPL13_2_647', description='ribosomal protein L13 pseudogene 10', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181065155, end_position_on_the_genomic_accession=181065680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102725051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725051, CurrentID=0, Status='live', Symbol='LOC102725051', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='1, Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375989, CurrentID=0, Status='live', Symbol='LOC105375989', Aliases='', description='uncharacterized LOC105375989', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21030488, end_position_on_the_genomic_accession=21054169, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  95681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=95681, CurrentID=0, Status='live', Symbol='CEP41', Aliases='JBTS15, TSGA14', description='centrosomal protein 41kDa', other_designations='centrosomal protein 41 kDa|testis specific protein A14|testis specific, 14|testis-specific gene A14 protein', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130393771, end_position_on_the_genomic_accession=130441741, orientation='minus', exon_count=15, OMIM=610523, no_hdr0=''),
  102725060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725060, CurrentID=0, Status='live', Symbol='LOC102725060', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371435, CurrentID=0, Status='live', Symbol='LOC105371435', Aliases='', description='uncharacterized LOC105371435', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150507605, end_position_on_the_genomic_accession=150510869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375991, CurrentID=0, Status='live', Symbol='LOC105375991', Aliases='', description='uncharacterized LOC105375991', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22956846, end_position_on_the_genomic_accession=22962233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725068, CurrentID=0, Status='live', Symbol='LOC102725068', Aliases='', description='uncharacterized LOC102725068', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31479918, end_position_on_the_genomic_accession=31494794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102725069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725069, CurrentID=0, Status='live', Symbol='LOC102725069', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725070, CurrentID=0, Status='live', Symbol='LOC102725070', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725072, CurrentID=0, Status='live', Symbol='LOC102725072', Aliases='', description='uncharacterized LOC102725072', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18850161, end_position_on_the_genomic_accession=18855879, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375992, CurrentID=0, Status='live', Symbol='LOC105375992', Aliases='', description='uncharacterized LOC105375992', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23826428, end_position_on_the_genomic_accession=23830193, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725078, CurrentID=0, Status='live', Symbol='LOC102725078', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375993, CurrentID=0, Status='live', Symbol='LOC105375993', Aliases='', description='uncharacterized LOC105375993', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23851160, end_position_on_the_genomic_accession=23949656, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102725080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725080, CurrentID=0, Status='live', Symbol='LOC102725080', Aliases='', description='uncharacterized LOC102725080', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13629635, end_position_on_the_genomic_accession=13631936, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725082, CurrentID=0, Status='live', Symbol='LOC102725082', Aliases='', description='uncharacterized LOC102725082', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212926386, end_position_on_the_genomic_accession=212996462, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375994, CurrentID=0, Status='live', Symbol='LOC105375994', Aliases='', description='uncharacterized LOC105375994', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=24531281, end_position_on_the_genomic_accession=24543670, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102725086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725086, CurrentID=0, Status='live', Symbol='LOC102725086', Aliases='', description='uncharacterized LOC102725086', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  751071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751071, CurrentID=0, Status='live', Symbol='METTL12', Aliases='U99HG', description='methyltransferase like 12', other_designations='Putative methyltransferase LOC751071, mitochondrial', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62665263, end_position_on_the_genomic_accession=62667451, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725089, CurrentID=0, Status='live', Symbol='LOC102725089', Aliases='', description='uncharacterized LOC102725089', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725090, CurrentID=0, Status='live', Symbol='LOC102725090', Aliases='', description='uncharacterized LOC102725090', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375995, CurrentID=0, Status='live', Symbol='LOC105375995', Aliases='', description='uncharacterized LOC105375995', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25412622, end_position_on_the_genomic_accession=25419763, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725092, CurrentID=0, Status='live', Symbol='LOC102725092', Aliases='', description='uncharacterized LOC102725092', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725101, CurrentID=0, Status='live', Symbol='LOC102725101', Aliases='', description='Ig heavy chain V-III region VH26-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32051787, end_position_on_the_genomic_accession=32053547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375997, CurrentID=0, Status='live', Symbol='LOC105375997', Aliases='', description='uncharacterized LOC105375997', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25581177, end_position_on_the_genomic_accession=25588569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725104, CurrentID=0, Status='live', Symbol='LOC102725104', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725105, CurrentID=0, Status='live', Symbol='LOC102725105', Aliases='', description='uncharacterized LOC102725105', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20305551, end_position_on_the_genomic_accession=20332446, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102725106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725106, CurrentID=0, Status='live', Symbol='LOC102725106', Aliases='', description='uncharacterized LOC102725106', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725108, CurrentID=0, Status='live', Symbol='LOC102725108', Aliases='', description='uncharacterized LOC102725108', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375801, CurrentID=0, Status='live', Symbol='LOC105375801', Aliases='', description='uncharacterized LOC105375801', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143986055, end_position_on_the_genomic_accession=143989509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927830, CurrentID=0, Status='live', Symbol='LOC101927830', Aliases='', description='uncharacterized LOC101927830', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155347137, end_position_on_the_genomic_accession=155374174, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102725112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725112, CurrentID=0, Status='live', Symbol='LOC102725112', Aliases='', description='uncharacterized LOC102725112', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20845375, end_position_on_the_genomic_accession=20884942, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102725114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725114, CurrentID=0, Status='live', Symbol='LOC102725114', Aliases='', description='uncharacterized LOC102725114', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725115, CurrentID=0, Status='live', Symbol='LOC102725115', Aliases='', description='proline-rich protein 23D1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725116, CurrentID=0, Status='live', Symbol='LOC102725116', Aliases='', description='uncharacterized LOC102725116', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56108956, end_position_on_the_genomic_accession=56136743, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102725117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725117, CurrentID=0, Status='live', Symbol='LOC102725117', Aliases='', description='rho guanine nucleotide exchange factor 5-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725118, CurrentID=0, Status='live', Symbol='LOC102725118', Aliases='', description='proline-rich protein 23D1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  81493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81493, CurrentID=0, Status='live', Symbol='SYNC', Aliases='SYNC1OILIN, SYNC', description='syncoilin, intermediate filament protein', other_designations='intermediate filament protein syncoilin|syncoilin intermediate filament 1|syncoilin, intermediate filament 1|syncoilin-1', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32679906, end_position_on_the_genomic_accession=32702760, orientation='minus', exon_count=5, OMIM=611750, no_hdr0=''),
  105376000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376000, CurrentID=0, Status='live', Symbol='LOC105376000', Aliases='', description='uncharacterized LOC105376000', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26721263, end_position_on_the_genomic_accession=26724931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725126, CurrentID=0, Status='live', Symbol='LOC102725126', Aliases='', description='uncharacterized LOC102725126', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376001, CurrentID=0, Status='live', Symbol='LOC105376001', Aliases='', description='uncharacterized LOC105376001', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27496255, end_position_on_the_genomic_accession=27498430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723503, CurrentID=0, Status='live', Symbol='LOC102723503', Aliases='', description='uncharacterized LOC102723503', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38256229, end_position_on_the_genomic_accession=38386634, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102725131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725131, CurrentID=0, Status='live', Symbol='LOC102725131', Aliases='', description='putative uncharacterized protein ENSP00000383309', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376002, CurrentID=0, Status='live', Symbol='LOC105376002', Aliases='', description='uncharacterized LOC105376002', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27664332, end_position_on_the_genomic_accession=27684227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725134, CurrentID=0, Status='live', Symbol='LOC102725134', Aliases='', description='uncharacterized LOC102725134', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725138, CurrentID=0, Status='live', Symbol='LOC102725138', Aliases='', description='uncharacterized protein C16orf52 homolog B', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376003, CurrentID=0, Status='live', Symbol='LOC105376003', Aliases='', description='uncharacterized LOC105376003', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28619948, end_position_on_the_genomic_accession=28657726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725140, CurrentID=0, Status='live', Symbol='LOC102725140', Aliases='', description='uncharacterized LOC102725140', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165215855, end_position_on_the_genomic_accession=165219047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725141, CurrentID=0, Status='live', Symbol='LOC102725141', Aliases='', description='uncharacterized LOC102725141', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725143, CurrentID=0, Status='live', Symbol='LOC102725143', Aliases='', description='uncharacterized LOC102725143', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725144, CurrentID=0, Status='live', Symbol='LOC102725144', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376004, CurrentID=0, Status='live', Symbol='LOC105376004', Aliases='', description='uncharacterized LOC105376004', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28539715, end_position_on_the_genomic_accession=28601495, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102725148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725148, CurrentID=0, Status='live', Symbol='LOC102725148', Aliases='', description='uncharacterized LOC102725148', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71810532, end_position_on_the_genomic_accession=71839298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725151, CurrentID=0, Status='live', Symbol='LOC102725151', Aliases='', description='uncharacterized LOC102725151', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725155, CurrentID=0, Status='live', Symbol='LOC102725155', Aliases='', description='uncharacterized LOC102725155', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725156, CurrentID=0, Status='live', Symbol='LOC102725156', Aliases='', description='uncharacterized LOC102725156', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725159, CurrentID=0, Status='live', Symbol='LOC102725159', Aliases='', description='uncharacterized LOC102725159', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43843792, end_position_on_the_genomic_accession=43844520, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102725160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725160, CurrentID=0, Status='live', Symbol='LOC102725160', Aliases='', description='uncharacterized LOC102725160', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46096309, end_position_on_the_genomic_accession=46129866, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102725161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725161, CurrentID=0, Status='live', Symbol='LOC102725161', Aliases='', description='uncharacterized LOC102725161', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133127, CurrentID=0, Status='live', Symbol='LOC100133127', Aliases='', description='geminin, DNA replication inhibitor pseudogene 1', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16541223, end_position_on_the_genomic_accession=16541835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102725165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725165, CurrentID=0, Status='live', Symbol='LOC102725165', Aliases='', description='uncharacterized LOC102725165', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24376765, end_position_on_the_genomic_accession=24385452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725167, CurrentID=0, Status='live', Symbol='LOC102725167', Aliases='', description='uncharacterized LOC102725167', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725168, CurrentID=0, Status='live', Symbol='LOC102725168', Aliases='', description='uncharacterized LOC102725168', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71326163, end_position_on_the_genomic_accession=71328977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376008, CurrentID=0, Status='live', Symbol='LOC105376008', Aliases='', description='uncharacterized LOC105376008', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30271682, end_position_on_the_genomic_accession=30349696, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927832, CurrentID=0, Status='live', Symbol='PAQR9-AS1', Aliases='', description='PAQR9 antisense RNA 1', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142964058, end_position_on_the_genomic_accession=143001090, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102725176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725176, CurrentID=0, Status='live', Symbol='LINC01595', Aliases='', description='long intergenic non-protein coding RNA 1595', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47863251, end_position_on_the_genomic_accession=47869359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102725179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725179, CurrentID=0, Status='live', Symbol='LOC102725179', Aliases='', description='tyrosine-protein phosphatase non-receptor type 20-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725180, CurrentID=0, Status='live', Symbol='LOC102725180', Aliases='', description='uncharacterized LOC102725180', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1168955, end_position_on_the_genomic_accession=1174735, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376010, CurrentID=0, Status='live', Symbol='LOC105376010', Aliases='', description='uncharacterized LOC105376010', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31578618, end_position_on_the_genomic_accession=31606119, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102725182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725182, CurrentID=0, Status='live', Symbol='LOC102725182', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725186, CurrentID=0, Status='live', Symbol='LOC102725186', Aliases='', description='uncharacterized LOC102725186', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133131, CurrentID=0, Status='live', Symbol='LOC100133131', Aliases='', description='uncharacterized LOC100133131', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=6744683, end_position_on_the_genomic_accession=6745090, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102725188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725188, CurrentID=0, Status='live', Symbol='LOC102725188', Aliases='', description='uncharacterized LOC102725188', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725191, CurrentID=0, Status='live', Symbol='LOC102725191', Aliases='', description='uncharacterized LOC102725191', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12496425, end_position_on_the_genomic_accession=12544195, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102725193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725193, CurrentID=0, Status='live', Symbol='LOC102725193', Aliases='', description='uncharacterized LOC102725193', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7388183, end_position_on_the_genomic_accession=7389923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725194, CurrentID=0, Status='live', Symbol='LOC102725194', Aliases='', description='uncharacterized LOC102725194', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725195, CurrentID=0, Status='live', Symbol='LOC102725195', Aliases='', description='uncharacterized LOC102725195', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134092572, end_position_on_the_genomic_accession=134095196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102725200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725200, CurrentID=0, Status='live', Symbol='LOC102725200', Aliases='', description='uncharacterized LOC102725200', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=661209, end_position_on_the_genomic_accession=664172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102725201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725201, CurrentID=0, Status='live', Symbol='LOC102725201', Aliases='', description='uncharacterized LOC102725201', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35476374, end_position_on_the_genomic_accession=35491205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102725203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725203, CurrentID=0, Status='live', Symbol='LOC102725203', Aliases='', description='uncharacterized LOC102725203', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725204, CurrentID=0, Status='live', Symbol='LOC102725204', Aliases='', description='glutathione S-transferase theta-1-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24006088, end_position_on_the_genomic_accession=24008824, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102725205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725205, CurrentID=0, Status='live', Symbol='LOC102725205', Aliases='', description='uncharacterized LOC102725205', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725207, CurrentID=0, Status='live', Symbol='LOC102725207', Aliases='', description='annexin A8-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725208, CurrentID=0, Status='live', Symbol='LOC102725208', Aliases='', description='uncharacterized LOC102725208', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93619691, end_position_on_the_genomic_accession=93620970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725209, CurrentID=0, Status='live', Symbol='LOC102725209', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725210, CurrentID=0, Status='live', Symbol='LOC102725210', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369258, CurrentID=0, Status='live', Symbol='LOC105369258', Aliases='', description='arf-GAP with GTPase, ANK repeat and PH domain-containing protein 4', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725220, CurrentID=0, Status='live', Symbol='LOC102725220', Aliases='', description='uncharacterized LOC102725220', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107286612, end_position_on_the_genomic_accession=107304537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376017, CurrentID=0, Status='live', Symbol='LOC105376017', Aliases='', description='uncharacterized LOC105376017', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32679859, end_position_on_the_genomic_accession=32783241, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102725227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725227, CurrentID=0, Status='live', Symbol='LOC102725227', Aliases='', description='uncharacterized LOC102725227', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26654484, end_position_on_the_genomic_accession=26689436, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102725228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725228, CurrentID=0, Status='live', Symbol='LOC102725228', Aliases='', description='uncharacterized LOC102725228', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112686769, end_position_on_the_genomic_accession=112689917, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102725231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725231, CurrentID=0, Status='live', Symbol='LOC102725231', Aliases='', description='uncharacterized LOC102725231', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180466989, end_position_on_the_genomic_accession=180493801, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102725234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725234, CurrentID=0, Status='live', Symbol='LOC102725234', Aliases='', description='uncharacterized LOC102725234', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133139, CurrentID=0, Status='live', Symbol='RPL12P41', Aliases='RPL12_21_1660', description='ribosomal protein L12 pseudogene 41', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46219335, end_position_on_the_genomic_accession=46219966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102725236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725236, CurrentID=0, Status='live', Symbol='LOC102725236', Aliases='', description='histone-lysine N-methyltransferase 2C-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725237, CurrentID=0, Status='live', Symbol='LOC102725237', Aliases='', description='uncharacterized LOC102725237', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725238, CurrentID=0, Status='live', Symbol='LOC102725238', Aliases='', description='uncharacterized LOC102725238', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42423187, end_position_on_the_genomic_accession=42440922, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102725239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725239, CurrentID=0, Status='live', Symbol='LOC102725239', Aliases='', description='uncharacterized LOC102725239', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725241, CurrentID=0, Status='live', Symbol='LOC102725241', Aliases='', description='putative neutral ceramidase C', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725247, CurrentID=0, Status='live', Symbol='LOC102725247', Aliases='', description='uncharacterized LOC102725247', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725250, CurrentID=0, Status='live', Symbol='LOC102725250', Aliases='', description='putative uncharacterized protein DDB_G0291608', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376022, CurrentID=0, Status='live', Symbol='LOC105376022', Aliases='', description='uncharacterized LOC105376022', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34416930, end_position_on_the_genomic_accession=34424860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725254, CurrentID=0, Status='live', Symbol='LOC102725254', Aliases='', description='uncharacterized LOC102725254', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18204739, end_position_on_the_genomic_accession=18205569, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  693185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=693185, CurrentID=0, Status='live', Symbol='MIR600', Aliases='MIRN600, hsa-mir-600', description='microRNA 600', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123111546, end_position_on_the_genomic_accession=123111643, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102725258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725258, CurrentID=0, Status='live', Symbol='LOC102725258', Aliases='', description='uncharacterized LOC102725258', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48197002, end_position_on_the_genomic_accession=48252461, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376024, CurrentID=0, Status='live', Symbol='LOC105376024', Aliases='', description='uncharacterized LOC105376024', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34667387, end_position_on_the_genomic_accession=34668804, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376025, CurrentID=0, Status='live', Symbol='LOC105376025', Aliases='', description='uncharacterized LOC105376025', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35013291, end_position_on_the_genomic_accession=35024751, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133147, CurrentID=0, Status='live', Symbol='LOC100133147', Aliases='', description='glucosamine-phosphate N-acetyltransferase 1 pseudogene', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121113359, end_position_on_the_genomic_accession=121113920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376029, CurrentID=0, Status='live', Symbol='LOC105376029', Aliases='', description='uncharacterized LOC105376029', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36392188, end_position_on_the_genomic_accession=36397037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133150, CurrentID=0, Status='live', Symbol='LOC100133150', Aliases='', description='uncharacterized LOC100133150', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198215137, end_position_on_the_genomic_accession=198219697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371443, CurrentID=0, Status='live', Symbol='LOC105371443', Aliases='', description='uncharacterized LOC105371443', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376031, CurrentID=0, Status='live', Symbol='LOC105376031', Aliases='', description='uncharacterized LOC105376031', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37027764, end_position_on_the_genomic_accession=37031333, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376033, CurrentID=0, Status='live', Symbol='LOC105376033', Aliases='', description='uncharacterized LOC105376033', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37161531, end_position_on_the_genomic_accession=37187276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927837, CurrentID=0, Status='live', Symbol='LOC101927837', Aliases='', description='uncharacterized LOC101927837', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67481320, end_position_on_the_genomic_accession=67505134, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376034, CurrentID=0, Status='live', Symbol='LOC105376034', Aliases='', description='uncharacterized LOC105376034', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37110958, end_position_on_the_genomic_accession=37120088, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376035, CurrentID=0, Status='live', Symbol='LOC105376035', Aliases='', description='uncharacterized LOC105376035', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37365826, end_position_on_the_genomic_accession=37421411, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376036, CurrentID=0, Status='live', Symbol='LOC105376036', Aliases='', description='uncharacterized LOC105376036', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37380680, end_position_on_the_genomic_accession=37395211, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  653509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653509, CurrentID=0, Status='live', Symbol='SFTPA1', Aliases='COLEC4, PSAP, PSP-A, PSPA, SFTP1B, SP-A, SP-A1, SPA, SPA1, SFTPA1', description='surfactant protein A1', other_designations="35 kDa pulmonary surfactant-associated protein|alveolar proteinosis protein|collectin-4|surfactant protein A1 variant AB'D' 6A|surfactant protein A1 variant AB'D' 6A2|surfactant protein A1 variant AB'D' 6A3|surfactant protein A1 variant AB'D' 6A4|surfactant protein A1 variant ACD' 6A|surfactant protein A1 variant ACD' 6A2|surfactant protein A1 variant ACD' 6A3|surfactant protein A1 variant ACD' 6A4|surfactant protein A1 variant AD' 6A|surfactant protein A1 variant AD' 6A2|surfactant protein A1 variant AD' 6A3|surfactant protein A1 variant AD' 6A4|surfactant protein A1B|surfactant, pulmonary-associated protein A1A|surfactant, pulmonary-associated protein A1B", map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79610939, end_position_on_the_genomic_accession=79615443, orientation='plus', exon_count=8, OMIM=178630, no_hdr0=''),
  100132741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132741, CurrentID=0, Status='live', Symbol='LOC100132741', Aliases='', description='uncharacterized LOC100132741', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71778169, end_position_on_the_genomic_accession=71784378, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271349, CurrentID=0, Status='live', Symbol='RPS10P18', Aliases='RPS10_6_1108', description='ribosomal protein S10 pseudogene 18', other_designations='', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124490192, end_position_on_the_genomic_accession=124490615, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133161, CurrentID=0, Status='live', Symbol='LINC01001', Aliases='', description='long intergenic non-protein coding RNA 1001', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126987, end_position_on_the_genomic_accession=131920, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376042, CurrentID=0, Status='live', Symbol='LOC105376042', Aliases='', description='uncharacterized LOC105376042', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38502039, end_position_on_the_genomic_accession=38504347, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376043, CurrentID=0, Status='live', Symbol='LOC105376043', Aliases='', description='uncharacterized LOC105376043', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38848587, end_position_on_the_genomic_accession=38871999, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376044, CurrentID=0, Status='live', Symbol='LOC105376044', Aliases='', description='uncharacterized LOC105376044', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38901175, end_position_on_the_genomic_accession=38903556, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376045, CurrentID=0, Status='live', Symbol='LOC105376045', Aliases='', description='uncharacterized LOC105376045', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39321342, end_position_on_the_genomic_accession=39322285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376046, CurrentID=0, Status='live', Symbol='LOC105376046', Aliases='', description='uncharacterized LOC105376046', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39674635, end_position_on_the_genomic_accession=39680966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419742, CurrentID=0, Status='live', Symbol='R3HDM2P1', Aliases='', description='R3H domain containing 2 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=133124104, end_position_on_the_genomic_accession=133126009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376048, CurrentID=0, Status='live', Symbol='LOC105376048', Aliases='', description='uncharacterized LOC105376048', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40044873, end_position_on_the_genomic_accession=40048713, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376050, CurrentID=0, Status='live', Symbol='LOC105376050', Aliases='', description='uncharacterized LOC105376050', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39897458, end_position_on_the_genomic_accession=40001428, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376051, CurrentID=0, Status='live', Symbol='LOC105376051', Aliases='', description='uncharacterized LOC105376051', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39858100, end_position_on_the_genomic_accession=39863825, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401661, CurrentID=0, Status='live', Symbol='OR51C1P', Aliases='OR51C2P, OR51C3P, OST734', description='olfactory receptor family 51 subfamily C member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily C, member 1 pseudogene|olfactory receptor, family 51, subfamily C, member 2 pseudogene|olfactory receptor, family 51, subfamily C, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4690760, end_position_on_the_genomic_accession=4691701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376052, CurrentID=0, Status='live', Symbol='LOC105376052', Aliases='', description='uncharacterized LOC105376052', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40415934, end_position_on_the_genomic_accession=40432901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  341378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341378, CurrentID=0, Status='live', Symbol='LOC341378', Aliases='', description='thyroid hormone receptor interactor 11 pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106170406, end_position_on_the_genomic_accession=106172434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376053, CurrentID=0, Status='live', Symbol='LOC105376053', Aliases='', description='uncharacterized LOC105376053', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40375670, end_position_on_the_genomic_accession=40387219, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375812, CurrentID=0, Status='live', Symbol='LOC105375812', Aliases='', description='uncharacterized LOC105375812', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46816163, end_position_on_the_genomic_accession=46818525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421184, CurrentID=0, Status='live', Symbol='LOC100421184', Aliases='', description='zinc finger and BTB domain containing 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15226364, end_position_on_the_genomic_accession=15227243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376055, CurrentID=0, Status='live', Symbol='LOC105376055', Aliases='', description='uncharacterized LOC105376055', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40348754, end_position_on_the_genomic_accession=40417393, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100628307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100628307, CurrentID=0, Status='live', Symbol='DIO2-AS1', Aliases='', description='DIO2 antisense RNA 1', other_designations='CTD-2540C19.1|DIO2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80211419, end_position_on_the_genomic_accession=80455467, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102725067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725067, CurrentID=0, Status='live', Symbol='LOC102725067', Aliases='', description='uncharacterized LOC102725067', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133177, CurrentID=0, Status='live', Symbol='LOC100133177', Aliases='', description='uncharacterized LOC100133177', other_designations='', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51387554, end_position_on_the_genomic_accession=51389620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100628314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100628314, CurrentID=0, Status='live', Symbol='LOC100628314', Aliases='', description='uncharacterized LOC100628314', other_designations='', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100628315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100628315, CurrentID=0, Status='live', Symbol='DNM3OS', Aliases='DNM3-AS1, MIR199A2HG', description='DNM3 opposite strand/antisense RNA', other_designations='DNM3 antisense RNA 1 (non-protein coding)|DNM3 opposite strand/antisense RNA (non-protein coding)|MIR199A2 host gene (non-protein coding)|dynamin 3, opposite strand', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172136879, end_position_on_the_genomic_accession=172144835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376058, CurrentID=0, Status='live', Symbol='LOC105376058', Aliases='', description='uncharacterized LOC105376058', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41080489, end_position_on_the_genomic_accession=41089629, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376060, CurrentID=0, Status='live', Symbol='LOC105376060', Aliases='', description='uncharacterized LOC105376060', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41264101, end_position_on_the_genomic_accession=41274051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371449, CurrentID=0, Status='live', Symbol='LOC105371449', Aliases='', description='uncharacterized LOC105371449', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154919506, end_position_on_the_genomic_accession=154924183, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133182, CurrentID=0, Status='live', Symbol='LOC100133182', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927843, CurrentID=0, Status='live', Symbol='LOC101927843', Aliases='', description='uncharacterized LOC101927843', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=22009158, end_position_on_the_genomic_accession=22098459, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102725509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725509, CurrentID=0, Status='live', Symbol='LINC01232', Aliases='TCONS_00021520', description='long intergenic non-protein coding RNA 1232', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99486962, end_position_on_the_genomic_accession=99499306, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  139431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139431, CurrentID=0, Status='live', Symbol='FTH1P14', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene 14', other_designations='ferritin, heavy polypeptide-like 17 pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=34147040, end_position_on_the_genomic_accession=34147747, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102725532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725532, CurrentID=0, Status='live', Symbol='LOC102725532', Aliases='', description='uncharacterized LOC102725532', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102725541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725541, CurrentID=0, Status='live', Symbol='ELDR', Aliases='Fabl, LINC01156', description='EGFR long non-coding downstream RNA', other_designations='long intergenic non-protein coding RNA 1156', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55238436, end_position_on_the_genomic_accession=55255635, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133193, CurrentID=0, Status='live', Symbol='RPL32P28', Aliases='RPL32_18_1316', description='ribosomal protein L32 pseudogene 28', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62327884, end_position_on_the_genomic_accession=62328396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  751580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751580, CurrentID=0, Status='live', Symbol='LINC00106', Aliases='CXYorf8, NCRNA00106', description='long intergenic non-protein coding RNA 106', other_designations='', map_location='Xp22.33; Yp11.32', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1397025, end_position_on_the_genomic_accession=1399412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376080, CurrentID=0, Status='live', Symbol='LOC105376080', Aliases='', description='uncharacterized LOC105376080', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72478111, end_position_on_the_genomic_accession=72485301, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  751599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751599, CurrentID=0, Status='live', Symbol='TPM3P7', Aliases='', description='tropomyosin 3 pseudogene 7', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25808688, end_position_on_the_genomic_accession=25810727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  751602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751602, CurrentID=0, Status='live', Symbol='LOC751602', Aliases='', description='tropomyosin 3 pseudogene', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104156409, end_position_on_the_genomic_accession=104157545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  751603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751603, CurrentID=0, Status='live', Symbol='LOC751603', Aliases='', description='tropomyosin 3 pseudogene', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36613404, end_position_on_the_genomic_accession=36615474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  81577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81577, CurrentID=0, Status='live', Symbol='GFOD2', Aliases='', description='glucose-fructose oxidoreductase domain containing 2', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67674533, end_position_on_the_genomic_accession=67719370, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  751609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751609, CurrentID=0, Status='live', Symbol='DFNB68', Aliases='', description='deafness, autosomal recessive 68', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610419, no_hdr0=''),
  105376085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376085, CurrentID=0, Status='live', Symbol='LOC105376085', Aliases='', description='uncharacterized LOC105376085', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73573615, end_position_on_the_genomic_accession=73579780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422088, CurrentID=0, Status='live', Symbol='LOC100422088', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila) pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22366366, end_position_on_the_genomic_accession=22368423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376086, CurrentID=0, Status='live', Symbol='LOC105376086', Aliases='', description='uncharacterized LOC105376086', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73924035, end_position_on_the_genomic_accession=73938958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133207, CurrentID=0, Status='live', Symbol='LOC100133207', Aliases='', description='uncharacterized LOC100133207', other_designations='', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=96181820, end_position_on_the_genomic_accession=96223227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  343727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343727, CurrentID=0, Status='live', Symbol='CYCSP42', Aliases='CYC1LP5, HCP42', description='cytochrome c, somatic pseudogene 42', other_designations='cytochrome C pseudogene|cytochrome c-1-like pseudogene 5', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15490524, end_position_on_the_genomic_accession=15490782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376090, CurrentID=0, Status='live', Symbol='LOC105376090', Aliases='', description='uncharacterized LOC105376090', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75151635, end_position_on_the_genomic_accession=75180672, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133211, CurrentID=0, Status='live', Symbol='LOC100133211', Aliases='', description='related RAS viral (r-ras) oncogene homolog 2 pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52991782, end_position_on_the_genomic_accession=52993703, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376092, CurrentID=0, Status='live', Symbol='LOC105376092', Aliases='', description='uncharacterized LOC105376092', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76218619, end_position_on_the_genomic_accession=76233440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372473, CurrentID=0, Status='live', Symbol='LOC105372473', Aliases='', description='uncharacterized LOC105372473', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56672325, end_position_on_the_genomic_accession=56709836, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375820, CurrentID=0, Status='live', Symbol='LOC105375820', Aliases='', description='uncharacterized LOC105375820', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48309221, end_position_on_the_genomic_accession=48313182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129667, CurrentID=0, Status='live', Symbol='LOC100129667', Aliases='', description='chromosome 5 open reading frame 13 pseudogene', other_designations='neuronal regeneration related protein homolog (rat) pseudogene', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53712697, end_position_on_the_genomic_accession=53712896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100526839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526839, CurrentID=0, Status='live', Symbol='LOC100526839', Aliases='', description='chromosome 18 open reading frame 32 pseudogene', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183587140, end_position_on_the_genomic_accession=183588580, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100628560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100628560, CurrentID=0, Status='live', Symbol='MIR3155B', Aliases='hsa-mir-3155b', description='microRNA 3155b', other_designations='microRNA mir-3155b', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6152207, end_position_on_the_genomic_accession=6152262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376099, CurrentID=0, Status='live', Symbol='LOC105376099', Aliases='', description='uncharacterized LOC105376099', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79985220, end_position_on_the_genomic_accession=79987899, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133220, CurrentID=0, Status='live', Symbol='GOLGA6L3', Aliases='', description='golgin A6 family-like 3', other_designations='golgi autoantigen, golgin subfamily a, 6-like 3', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85240410, end_position_on_the_genomic_accession=85247187, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105371457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371457, CurrentID=0, Status='live', Symbol='LOC105371457', Aliases='', description='uncharacterized LOC105371457', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157653129, end_position_on_the_genomic_accession=157660145, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129069, CurrentID=0, Status='live', Symbol='LOC100129069', Aliases='', description='uncharacterized LOC100129069', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=124135279, end_position_on_the_genomic_accession=124135713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376102, CurrentID=0, Status='live', Symbol='LOC105376102', Aliases='', description='uncharacterized LOC105376102', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80269529, end_position_on_the_genomic_accession=80304971, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133225, CurrentID=0, Status='live', Symbol='LOC100133225', Aliases='', description='zinc finger, DHHC-type containing 7 pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51194120, end_position_on_the_genomic_accession=51195555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644794, CurrentID=0, Status='live', Symbol='LOC644794', Aliases='', description='uncharacterized LOC644794', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66903864, end_position_on_the_genomic_accession=66906297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376108, CurrentID=0, Status='live', Symbol='LOC105376108', Aliases='', description='uncharacterized LOC105376108', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81961102, end_position_on_the_genomic_accession=81977698, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  751781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751781, CurrentID=0, Status='live', Symbol='GLC1L', Aliases='', description='glaucoma 1, open angle, L (adult-onset)', other_designations='', map_location='3p22-p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  751783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751783, CurrentID=0, Status='live', Symbol='DFNB69', Aliases='', description='deafness, autosomal recessive 69', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376114, CurrentID=0, Status='live', Symbol='LOC105376114', Aliases='', description='histone-lysine N-methyltransferase SETMAR-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83707835, end_position_on_the_genomic_accession=83713496, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271352, CurrentID=0, Status='live', Symbol='RPS12P16', Aliases='RPS12_7_1034', description='ribosomal protein S12 pseudogene 16', other_designations='', map_location='10p11.22-p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34675613, end_position_on_the_genomic_accession=34676128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  751798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751798, CurrentID=0, Status='live', Symbol='AUNX1', Aliases='DFNX5', description='auditory neuropathy, X-linked recessive 1', other_designations='', map_location='Xq23-q27.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300614, no_hdr0=''),
  256374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256374, CurrentID=0, Status='live', Symbol='LOC256374', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138643309, end_position_on_the_genomic_accession=138644721, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  96459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=96459, CurrentID=0, Status='live', Symbol='FNIP1', Aliases='', description='folliculin interacting protein 1', other_designations='', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131641714, end_position_on_the_genomic_accession=131797063, orientation='minus', exon_count=18, OMIM=610594, no_hdr0=''),
  100129672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129672, CurrentID=0, Status='live', Symbol='LOC100129672', Aliases='', description='cell adhesion associated, oncogene regulated pseudogene', other_designations='Cdon homolog pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49098161, end_position_on_the_genomic_accession=49101926, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  751837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=751837, CurrentID=0, Status='live', Symbol='LOC751837', Aliases='', description='small nuclear ribonucleoprotein, U5 40kDa subunit pseudogene', other_designations='WD repeat domain 57 (U5 snRNP specific) pseudogene|small nuclear ribonucleoprotein 40kDa (U5) pseudogene', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156175181, end_position_on_the_genomic_accession=156176923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927855, CurrentID=0, Status='live', Symbol='LOC101927855', Aliases='', description='uncharacterized LOC101927855', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61361594, end_position_on_the_genomic_accession=61383553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376124, CurrentID=0, Status='live', Symbol='LOC105376124', Aliases='', description='uncharacterized LOC105376124', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86669272, end_position_on_the_genomic_accession=86708305, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928505, CurrentID=0, Status='live', Symbol='LOC101928505', Aliases='', description='uncharacterized LOC101928505', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57650660, end_position_on_the_genomic_accession=57670989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400385, CurrentID=0, Status='live', Symbol='RPL5P3', Aliases='HsT18216, RPL5_13_1449', description='ribosomal protein L5 pseudogene 3', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71063038, end_position_on_the_genomic_accession=71064065, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376129, CurrentID=0, Status='live', Symbol='LOC105376129', Aliases='', description='uncharacterized LOC105376129', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87824673, end_position_on_the_genomic_accession=87833420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376131, CurrentID=0, Status='live', Symbol='LOC105376131', Aliases='', description='uncharacterized LOC105376131', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87974081, end_position_on_the_genomic_accession=87994235, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133252, CurrentID=0, Status='live', Symbol='LOC100133252', Aliases='', description='zinc finger protein 131 pseudogene', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130840009, end_position_on_the_genomic_accession=130843740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376135, CurrentID=0, Status='live', Symbol='LOC105376135', Aliases='', description='uncharacterized LOC105376135', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88652027, end_position_on_the_genomic_accession=88652623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376138, CurrentID=0, Status='live', Symbol='LOC105376138', Aliases='', description='uncharacterized LOC105376138', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89565907, end_position_on_the_genomic_accession=89574565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375829, CurrentID=0, Status='live', Symbol='LOC105375829', Aliases='', description='uncharacterized LOC105375829', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50214984, end_position_on_the_genomic_accession=50237236, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376139, CurrentID=0, Status='live', Symbol='LOC105376139', Aliases='', description='uncharacterized LOC105376139', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89598783, end_position_on_the_genomic_accession=89604896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376141, CurrentID=0, Status='live', Symbol='LOC105376141', Aliases='', description='uncharacterized LOC105376141', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89654416, end_position_on_the_genomic_accession=89662115, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  96597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=96597, CurrentID=0, Status='live', Symbol='TBC1D27', Aliases='', description='TBC1 domain family member 27', other_designations='TBC1 domain family, member 27', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16922923, end_position_on_the_genomic_accession=16929592, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  96610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=96610, CurrentID=0, Status='live', Symbol='BMS1P20', Aliases='IGL, IGLV, IGLV1', description='BMS1, ribosome biogenesis factor pseudogene 20', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 20|BMS1 ribosome biogenesis factor pseudogene 20|anti-HIV-1 V3 immunoglobulin light chain|immunoglobulin lambda-1 variable region|immunoglobulin light chain variable region', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22298098, end_position_on_the_genomic_accession=22322968, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  131578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131578, CurrentID=0, Status='live', Symbol='LRRC15', Aliases='LIB', description='leucine rich repeat containing 15', other_designations='hLib|leucine-rich repeat protein induced by beta amyloid|leucine-rich repeat protein induced by beta-amyloid homolog', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194355247, end_position_on_the_genomic_accession=194369743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376145, CurrentID=0, Status='live', Symbol='LOC105376145', Aliases='', description='uncharacterized LOC105376145', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91119142, end_position_on_the_genomic_accession=91125134, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422100, CurrentID=0, Status='live', Symbol='LOC100422100', Aliases='', description='olfactory receptor, family 6, subfamily C, member 65 pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  96626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=96626, CurrentID=0, Status='live', Symbol='LIMS3', Aliases='PINCH-3', description='LIM and senescent cell antigen-like domains 3', other_designations='particularly interesting new Cys-His protein 3|pinch 2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109898432, end_position_on_the_genomic_accession=109925625, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105376147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376147, CurrentID=0, Status='live', Symbol='LOC105376147', Aliases='', description='uncharacterized LOC105376147', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91691893, end_position_on_the_genomic_accession=91718922, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  649620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649620, CurrentID=0, Status='live', Symbol='CDC27P6', Aliases='', description='cell division cycle 27 pseudogene 6', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=16047787, end_position_on_the_genomic_accession=16049736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422101, CurrentID=0, Status='live', Symbol='LOC100422101', Aliases='', description='olfactory receptor, family 4, subfamily F, member 21 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  752014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=752014, CurrentID=0, Status='live', Symbol='CEMP1', Aliases='CP-23, CP23', description='cementum protein 1', other_designations='cementum protein 23|cementum protein-23', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2530035, end_position_on_the_genomic_accession=2531408, orientation='minus', exon_count=1, OMIM=611113, no_hdr0=''),
  105376152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376152, CurrentID=0, Status='live', Symbol='LOC105376152', Aliases='', description='uncharacterized LOC105376152', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94144669, end_position_on_the_genomic_accession=94147468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376153, CurrentID=0, Status='live', Symbol='LOC105376153', Aliases='', description='uncharacterized LOC105376153', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94361251, end_position_on_the_genomic_accession=94374815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376158, CurrentID=0, Status='live', Symbol='LOC105376158', Aliases='', description='uncharacterized LOC105376158', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95599493, end_position_on_the_genomic_accession=95603968, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376163, CurrentID=0, Status='live', Symbol='LOC105376163', Aliases='', description='uncharacterized LOC105376163', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96599431, end_position_on_the_genomic_accession=96614270, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387492, CurrentID=0, Status='live', Symbol='DSTNP1', Aliases='', description='destrin (actin depolymerizing factor) pseudogene 1', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46653527, end_position_on_the_genomic_accession=46654022, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376165, CurrentID=0, Status='live', Symbol='LOC105376165', Aliases='', description='uncharacterized LOC105376165', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97222572, end_position_on_the_genomic_accession=97234983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422104, CurrentID=0, Status='live', Symbol='LOC100422104', Aliases='', description='olfactory receptor, family 7, subfamily A, member 17 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133286, CurrentID=0, Status='live', Symbol='LOC100133286', Aliases='', description='uncharacterized LOC100133286', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36069642, end_position_on_the_genomic_accession=36126640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376167, CurrentID=0, Status='live', Symbol='LOC105376167', Aliases='', description='uncharacterized LOC105376167', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97396118, end_position_on_the_genomic_accession=97409352, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376168, CurrentID=0, Status='live', Symbol='LOC105376168', Aliases='', description='uncharacterized LOC105376168', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97513032, end_position_on_the_genomic_accession=97516842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  96764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=96764, CurrentID=0, Status='live', Symbol='TGS1', Aliases='NCOA6IP, PIMT, PIPMT', description='trimethylguanosine synthase 1', other_designations='CLL-associated antigen KW-2|PRIP-interacting protein with methyltransferase motif|cap-specific guanine-N2 methyltransferase|hepatocellular carcinoma-associated antigen 137|nuclear receptor coactivator 6 interacting protein|nuclear receptor coactivator 6-interacting protein', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55773232, end_position_on_the_genomic_accession=55825448, orientation='plus', exon_count=13, OMIM=606461, no_hdr0=''),
  105376174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376174, CurrentID=0, Status='live', Symbol='LOC105376174', Aliases='', description='uncharacterized LOC105376174', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99184323, end_position_on_the_genomic_accession=99185536, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376175, CurrentID=0, Status='live', Symbol='LOC105376175', Aliases='', description='uncharacterized LOC105376175', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99665596, end_position_on_the_genomic_accession=99750922, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129683, CurrentID=0, Status='live', Symbol='LOC100129683', Aliases='', description='WEE1 homolog (S. pombe) pseudogene', other_designations='', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22704119, end_position_on_the_genomic_accession=22705636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376176, CurrentID=0, Status='live', Symbol='LOC105376176', Aliases='', description='uncharacterized LOC105376176', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100016604, end_position_on_the_genomic_accession=100033095, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376177, CurrentID=0, Status='live', Symbol='LOC105376177', Aliases='', description='uncharacterized LOC105376177', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100353036, end_position_on_the_genomic_accession=100390489, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  387506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387506, CurrentID=0, Status='live', Symbol='HCG9P2', Aliases='HCGIX-2', description='HLA complex group 9 pseudogene 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387507, CurrentID=0, Status='live', Symbol='HCG9P3', Aliases='HCGIX-3.2', description='HLA complex group 9 pseudogene 3', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387509, CurrentID=0, Status='live', Symbol='GPR153', Aliases='PGR1', description='G protein-coupled receptor 153', other_designations='G-protein coupled receptor PGR1', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6247346, end_position_on_the_genomic_accession=6260990, orientation='minus', exon_count=6, OMIM=614269, no_hdr0=''),
  100133306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133306, CurrentID=0, Status='live', Symbol='LOC100133306', Aliases='', description='uncharacterized LOC100133306', other_designations='', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=70675431, end_position_on_the_genomic_accession=70675967, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100133307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133307, CurrentID=0, Status='live', Symbol='CYP1D1P', Aliases='CYP1A8P, CYP1D1', description='cytochrome P450 family 1 subfamily D member 1, pseudogene', other_designations='cytochrome P450, family 1, subfamily D pseudogene|cytochrome P450, family 1, subfamily D, polypeptide 1, pseudogene', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73042637, end_position_on_the_genomic_accession=73056597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376188, CurrentID=0, Status='live', Symbol='LOC105376188', Aliases='', description='uncharacterized LOC105376188', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102216901, end_position_on_the_genomic_accession=102236462, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376189, CurrentID=0, Status='live', Symbol='LOC105376189', Aliases='', description='uncharacterized LOC105376189', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102236532, end_position_on_the_genomic_accession=102259725, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376190, CurrentID=0, Status='live', Symbol='LOC105376190', Aliases='', description='uncharacterized LOC105376190', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102503404, end_position_on_the_genomic_accession=102514607, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371475, CurrentID=0, Status='live', Symbol='LOC105371475', Aliases='', description='uncharacterized LOC105371475', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162105454, end_position_on_the_genomic_accession=162175134, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376191, CurrentID=0, Status='live', Symbol='LOC105376191', Aliases='', description='uncharacterized LOC105376191', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103126741, end_position_on_the_genomic_accession=103129179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374319, CurrentID=0, Status='live', Symbol='LOC105374319', Aliases='', description='uncharacterized LOC105374319', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21845475, end_position_on_the_genomic_accession=21913178, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100629121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100629121, CurrentID=0, Status='live', Symbol='LOC100629121', Aliases='', description='endonuclease domain containing 1 pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30824841, end_position_on_the_genomic_accession=30825201, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376193, CurrentID=0, Status='live', Symbol='LOC105376193', Aliases='', description='uncharacterized LOC105376193', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103956315, end_position_on_the_genomic_accession=103999643, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  101927869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927869, CurrentID=0, Status='live', Symbol='LOC101927869', Aliases='', description='uncharacterized LOC101927869', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=24304550, end_position_on_the_genomic_accession=24321377, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133314, CurrentID=0, Status='live', Symbol='SETP1', Aliases='SETP9, SETP1', description='SET pseudogene 1', other_designations='SET pseudogene 19', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20528113, end_position_on_the_genomic_accession=20528960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376195, CurrentID=0, Status='live', Symbol='LOC105376195', Aliases='', description='uncharacterized LOC105376195', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104392445, end_position_on_the_genomic_accession=104463583, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105371476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371476, CurrentID=0, Status='live', Symbol='LOC105371476', Aliases='', description='uncharacterized LOC105371476', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=573264, end_position_on_the_genomic_accession=588903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133316, CurrentID=0, Status='live', Symbol='LOC100133316', Aliases='', description='required for meiotic nuclear division 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74935176, end_position_on_the_genomic_accession=74935690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375841, CurrentID=0, Status='live', Symbol='LOC105375841', Aliases='', description='uncharacterized LOC105375841', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54426896, end_position_on_the_genomic_accession=54454838, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100510709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100510709, CurrentID=0, Status='live', Symbol='LOC100510709', Aliases='', description='ADP/ATP translocase 2-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376199, CurrentID=0, Status='live', Symbol='LOC105376199', Aliases='', description='uncharacterized LOC105376199', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105554035, end_position_on_the_genomic_accession=105558280, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371477, CurrentID=0, Status='live', Symbol='LOC105371477', Aliases='', description='uncharacterized LOC105371477', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=523140, end_position_on_the_genomic_accession=541832, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376201, CurrentID=0, Status='live', Symbol='LOC105376201', Aliases='', description='uncharacterized LOC105376201', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106025473, end_position_on_the_genomic_accession=106041350, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133326, CurrentID=0, Status='live', Symbol='LOC100133326', Aliases='', description='serine/threonine-protein phosphatase 4 regulatory subunit 2-like', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195937173, end_position_on_the_genomic_accession=195938876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376207, CurrentID=0, Status='live', Symbol='LOC105376207', Aliases='', description='uncharacterized LOC105376207', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107704389, end_position_on_the_genomic_accession=107717309, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376210, CurrentID=0, Status='live', Symbol='LOC105376210', Aliases='', description='uncharacterized LOC105376210', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107734800, end_position_on_the_genomic_accession=107766225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376211, CurrentID=0, Status='live', Symbol='LOC105376211', Aliases='', description='uncharacterized LOC105376211', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107951186, end_position_on_the_genomic_accession=107999710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  544431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544431, CurrentID=0, Status='live', Symbol='COPD15', Aliases='', description='Chronic obstructive pulmonary disease QTL 15', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=55605, CurrentID=0, Status='live', Symbol='KIF21A', Aliases='CFEOM1, FEOM1, FEOM3A', description='kinesin family member 21A', other_designations='kinesin-like protein KIF2|renal carcinoma antigen NY-REN-62', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39293228, end_position_on_the_genomic_accession=39443390, orientation='minus', exon_count=38, OMIM=608283, no_hdr0=''),
  105376215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376215, CurrentID=0, Status='live', Symbol='LOC105376215', Aliases='', description='uncharacterized LOC105376215', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109021800, end_position_on_the_genomic_accession=109080553, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376216, CurrentID=0, Status='live', Symbol='LOC105376216', Aliases='', description='uncharacterized LOC105376216', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108984443, end_position_on_the_genomic_accession=109006950, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  391674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391674, CurrentID=0, Status='live', Symbol='LOC391674', Aliases='', description='aminomethyltransferase pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83247179, end_position_on_the_genomic_accession=83247944, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106003269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106003269, CurrentID=0, Status='live', Symbol='SRXY10', Aliases='', description='46XY sex reversal 10', other_designations='', map_location='17q24', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616425, no_hdr0=''),
  105376225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376225, CurrentID=0, Status='live', Symbol='LOC105376225', Aliases='', description='uncharacterized LOC105376225', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114144277, end_position_on_the_genomic_accession=114152681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422021, CurrentID=0, Status='live', Symbol='LOC100422021', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B4 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69215598, end_position_on_the_genomic_accession=69216529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376226, CurrentID=0, Status='live', Symbol='LOC105376226', Aliases='', description='uncharacterized LOC105376226', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114058604, end_position_on_the_genomic_accession=114064348, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129947, CurrentID=0, Status='live', Symbol='LOC100129947', Aliases='', description='zinc finger protein 674 pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128098001, end_position_on_the_genomic_accession=128098530, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376231, CurrentID=0, Status='live', Symbol='LOC105376231', Aliases='', description='uncharacterized LOC105376231', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114697068, end_position_on_the_genomic_accession=114700173, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376232, CurrentID=0, Status='live', Symbol='LOC105376232', Aliases='', description='uncharacterized LOC105376232', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115353979, end_position_on_the_genomic_accession=115367921, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376234, CurrentID=0, Status='live', Symbol='LOC105376234', Aliases='', description='uncharacterized LOC105376234', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115744014, end_position_on_the_genomic_accession=115826737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371702, CurrentID=0, Status='live', Symbol='LOC105371702', Aliases='', description='uncharacterized LOC105371702', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205775571, end_position_on_the_genomic_accession=205783623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  780779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780779, CurrentID=0, Status='live', Symbol='USP10P1', Aliases='', description='ubiquitin specific peptidase 10 pseudogene 1', other_designations='ubiquitin specific protease 10 pseudogene 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19677750, end_position_on_the_genomic_accession=19680947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780780, CurrentID=0, Status='live', Symbol='LOC780780', Aliases='', description='ubiquitin specific peptidase 10 pseudogene', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15561169, end_position_on_the_genomic_accession=15564363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376241, CurrentID=0, Status='live', Symbol='LOC105376241', Aliases='', description='uncharacterized LOC105376241', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116288604, end_position_on_the_genomic_accession=116319945, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376242, CurrentID=0, Status='live', Symbol='LOC105376242', Aliases='', description='uncharacterized LOC105376242', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117431359, end_position_on_the_genomic_accession=117466163, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376244, CurrentID=0, Status='live', Symbol='LOC105376244', Aliases='', description='uncharacterized LOC105376244', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117759345, end_position_on_the_genomic_accession=117871225, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  780789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780789, CurrentID=0, Status='live', Symbol='C1orf196', Aliases='', description='chromosome 1 open reading frame 196', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129697, CurrentID=0, Status='live', Symbol='LOC100129697', Aliases='', description='uncharacterized LOC100129697', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88940039, end_position_on_the_genomic_accession=88951524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376246, CurrentID=0, Status='live', Symbol='LOC105376246', Aliases='', description='uncharacterized LOC105376246', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117997648, end_position_on_the_genomic_accession=118024353, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376247, CurrentID=0, Status='live', Symbol='LOC105376247', Aliases='', description='uncharacterized LOC105376247', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118057110, end_position_on_the_genomic_accession=118070509, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376248, CurrentID=0, Status='live', Symbol='LOC105376248', Aliases='', description='uncharacterized LOC105376248', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118514718, end_position_on_the_genomic_accession=118523497, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927880, CurrentID=0, Status='live', Symbol='LOC101927880', Aliases='', description='uncharacterized LOC101927880', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3433735, end_position_on_the_genomic_accession=3502860, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376249, CurrentID=0, Status='live', Symbol='LOC105376249', Aliases='', description='uncharacterized LOC105376249', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118593061, end_position_on_the_genomic_accession=118597172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376250, CurrentID=0, Status='live', Symbol='LOC105376250', Aliases='', description='uncharacterized LOC105376250', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=119424012, end_position_on_the_genomic_accession=119524055, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371487, CurrentID=0, Status='live', Symbol='LOC105371487', Aliases='', description='uncharacterized LOC105371487', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2174367, end_position_on_the_genomic_accession=2215826, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376251, CurrentID=0, Status='live', Symbol='LOC105376251', Aliases='', description='uncharacterized LOC105376251', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120120674, end_position_on_the_genomic_accession=120127022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376253, CurrentID=0, Status='live', Symbol='LOC105376253', Aliases='', description='uncharacterized LOC105376253', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120163025, end_position_on_the_genomic_accession=120384779, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129948, CurrentID=0, Status='live', Symbol='LOC100129948', Aliases='', description='riboflavin kinase pseudogene', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212647270, end_position_on_the_genomic_accession=212650326, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376254, CurrentID=0, Status='live', Symbol='LOC105376254', Aliases='', description='uncharacterized LOC105376254', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120581228, end_position_on_the_genomic_accession=120583640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376255, CurrentID=0, Status='live', Symbol='LOC105376255', Aliases='', description='uncharacterized LOC105376255', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121052514, end_position_on_the_genomic_accession=121060104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376256, CurrentID=0, Status='live', Symbol='LOC105376256', Aliases='', description='uncharacterized LOC105376256', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121292216, end_position_on_the_genomic_accession=121299157, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376257, CurrentID=0, Status='live', Symbol='LOC105376257', Aliases='', description='uncharacterized LOC105376257', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121731744, end_position_on_the_genomic_accession=121737106, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  282569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282569, CurrentID=0, Status='live', Symbol='BACE2-IT1', Aliases='C21orf75, NCRNA00228, PRED43', description='BACE2 intronic transcript 1', other_designations='BACE2 intronic transcript 1 (non-protein coding)', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376262, CurrentID=0, Status='live', Symbol='LOC105376262', Aliases='', description='uncharacterized LOC105376262', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122406570, end_position_on_the_genomic_accession=122437964, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376266, CurrentID=0, Status='live', Symbol='LOC105376266', Aliases='', description='uncharacterized LOC105376266', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123567251, end_position_on_the_genomic_accession=123573554, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376267, CurrentID=0, Status='live', Symbol='LOC105376267', Aliases='', description='uncharacterized LOC105376267', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123996149, end_position_on_the_genomic_accession=123999377, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376268, CurrentID=0, Status='live', Symbol='LOC105376268', Aliases='', description='uncharacterized LOC105376268', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124031427, end_position_on_the_genomic_accession=124032423, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  780813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780813, CurrentID=0, Status='live', Symbol='PAICSP4', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 4', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=4787136, end_position_on_the_genomic_accession=4841905, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376270, CurrentID=0, Status='live', Symbol='LOC105376270', Aliases='', description='uncharacterized LOC105376270', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124446670, end_position_on_the_genomic_accession=124452364, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376271, CurrentID=0, Status='live', Symbol='LOC105376271', Aliases='', description='uncharacterized LOC105376271', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125199529, end_position_on_the_genomic_accession=125200396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376272, CurrentID=0, Status='live', Symbol='LOC105376272', Aliases='', description='uncharacterized LOC105376272', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125567658, end_position_on_the_genomic_accession=125573156, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376273, CurrentID=0, Status='live', Symbol='LOC105376273', Aliases='', description='uncharacterized LOC105376273', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126051988, end_position_on_the_genomic_accession=126052986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  348040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348040, CurrentID=0, Status='live', Symbol='LOC348040', Aliases='', description='uncharacterized LOC348040', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422196, CurrentID=0, Status='live', Symbol='LOC100422196', Aliases='', description='olfactory receptor, family 4, subfamily F, member 15 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133398, CurrentID=0, Status='live', Symbol='LOC100133398', Aliases='', description='selenoprotein K pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927886, CurrentID=0, Status='live', Symbol='GBAT2', Aliases='', description='glioblastoma multiforme-associated transcript 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151346956, end_position_on_the_genomic_accession=151347891, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376280, CurrentID=0, Status='live', Symbol='LOC105376280', Aliases='', description='uncharacterized LOC105376280', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127243162, end_position_on_the_genomic_accession=127245177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375858, CurrentID=0, Status='live', Symbol='LOC105375858', Aliases='', description='uncharacterized LOC105375858', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58739526, end_position_on_the_genomic_accession=58750170, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376284, CurrentID=0, Status='live', Symbol='LOC105376284', Aliases='', description='uncharacterized LOC105376284', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127939712, end_position_on_the_genomic_accession=127941509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372228, CurrentID=0, Status='live', Symbol='LOC105372228', Aliases='', description='uncharacterized LOC105372228', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79610658, end_position_on_the_genomic_accession=79615136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422128, CurrentID=0, Status='live', Symbol='LOC100422128', Aliases='', description='olfactory receptor, family 52, subfamily H, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376287, CurrentID=0, Status='live', Symbol='LOC105376287', Aliases='', description='uncharacterized LOC105376287', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128663131, end_position_on_the_genomic_accession=128685777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  544446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544446, CurrentID=0, Status='live', Symbol='AASTH2', Aliases='', description='Allergic/atopic asthma related QTL 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927888, CurrentID=0, Status='live', Symbol='LOC101927888', Aliases='', description='uncharacterized LOC101927888', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4343047, end_position_on_the_genomic_accession=4347168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271359, CurrentID=0, Status='live', Symbol='RPS15AP26', Aliases='RPS15A_7_928', description='ribosomal protein S15a pseudogene 26', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94688238, end_position_on_the_genomic_accession=94688610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376292, CurrentID=0, Status='live', Symbol='LOC105376292', Aliases='', description='uncharacterized LOC105376292', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129575417, end_position_on_the_genomic_accession=129582686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  343930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=343930, CurrentID=0, Status='live', Symbol='MSGN1', Aliases='MSOG', description='mesogenin 1', other_designations='pMesogenin1|pMsgn1|paraxial mesoderm-specific mesogenin1|paraxial mesogenin', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17816519, end_position_on_the_genomic_accession=17817100, orientation='plus', exon_count=1, OMIM=612209, no_hdr0=''),
  81787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81787, CurrentID=0, Status='live', Symbol='DKFZP547L112', Aliases='', description='uncharacterized protein DKFZp547L112', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100498661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100498661, CurrentID=0, Status='live', Symbol='OSBPL9P4', Aliases='', description='oxysterol binding protein like 9 pseudogene 4', other_designations='oxysterol binding protein-like 9 pseudogene 4', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66563254, end_position_on_the_genomic_accession=66569541, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129707, CurrentID=0, Status='live', Symbol='LOC100129707', Aliases='', description='presenilin associated, rhomboid-like pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120028713, end_position_on_the_genomic_accession=120030058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927890, CurrentID=0, Status='live', Symbol='LOC101927890', Aliases='', description='uncharacterized LOC101927890', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23214008, end_position_on_the_genomic_accession=23239759, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376300, CurrentID=0, Status='live', Symbol='LOC105376300', Aliases='', description='uncharacterized LOC105376300', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131346986, end_position_on_the_genomic_accession=131350546, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371497, CurrentID=0, Status='live', Symbol='LOC105371497', Aliases='', description='uncharacterized LOC105371497', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162467315, end_position_on_the_genomic_accession=162497713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376303, CurrentID=0, Status='live', Symbol='LOC105376303', Aliases='', description='uncharacterized LOC105376303', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132374798, end_position_on_the_genomic_accession=132375522, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  359710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359710, CurrentID=0, Status='live', Symbol='BPIFB3', Aliases='C20orf185, LPLUNC3, RYA3', description='BPI fold containing family B member 3', other_designations='BPI fold containing family B, member 3|ligand-binding protein RYA3|long palate, lung and nasal epithelium carcinoma-associated protein 3', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33055331, end_position_on_the_genomic_accession=33073637, orientation='plus', exon_count=15, OMIM=615717, no_hdr0=''),
  105376304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376304, CurrentID=0, Status='live', Symbol='LOC105376304', Aliases='', description='uncharacterized LOC105376304', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132411325, end_position_on_the_genomic_accession=132415957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  359724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359724, CurrentID=0, Status='live', Symbol='RNA18SP2', Aliases='RN18S2P', description='RNA, 18S ribosomal pseudogene 2', other_designations='18S ribosomal RNA pseudogene|RNA, 18S ribosomal 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10197257, end_position_on_the_genomic_accession=10199104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105419997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105419997, CurrentID=0, Status='live', Symbol='SFTPD-AS1', Aliases='', description='SFTPD antisense RNA 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102464817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464817, CurrentID=0, Status='live', Symbol='MIR5787', Aliases='hsa-mir-5787', description='microRNA 5787', other_designations='microRNA mir-5787', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50227436, end_position_on_the_genomic_accession=50227490, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  359730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359730, CurrentID=0, Status='live', Symbol='MRPL57P3', Aliases='MRP63P3', description='mitochondrial ribosomal protein L57 pseudogene 3', other_designations='mitochondrial ribosomal protein 63 pseudogene 3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47028844, end_position_on_the_genomic_accession=47029039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780851, CurrentID=0, Status='live', Symbol='SNORD3A', Aliases='RNU3, U3', description='small nucleolar RNA, C/D box 3A', other_designations='RNA, U3 small nucleolar', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19188016, end_position_on_the_genomic_accession=19188232, orientation='plus', exon_count=1, OMIM=180710, no_hdr0=''),
  359732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359732, CurrentID=0, Status='live', Symbol='MRPL57P8', Aliases='MRP63P8', description='mitochondrial ribosomal protein L57 pseudogene 8', other_designations='mitochondrial ribosomal protein 63 pseudogene 8', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35219020, end_position_on_the_genomic_accession=35219294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359733, CurrentID=0, Status='live', Symbol='MRPL57P9', Aliases='MRP63P9', description='mitochondrial ribosomal protein L57 pseudogene 9', other_designations='mitochondrial ribosomal protein 63 pseudogene 9', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50917657, end_position_on_the_genomic_accession=50917851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359734, CurrentID=0, Status='live', Symbol='MRPL11P3', Aliases='', description='mitochondrial ribosomal protein L11 pseudogene 3', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78473314, end_position_on_the_genomic_accession=78473423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102464823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464823, CurrentID=0, Status='live', Symbol='MIR6068', Aliases='hsa-mir-6068', description='microRNA 6068', other_designations='microRNA mir-6068', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63326925, end_position_on_the_genomic_accession=63326984, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102464824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464824, CurrentID=0, Status='live', Symbol='MIR6069', Aliases='hsa-mir-6069', description='microRNA 6069', other_designations='microRNA mir-6069', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35336721, end_position_on_the_genomic_accession=35336799, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102464825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464825, CurrentID=0, Status='live', Symbol='MIR6070', Aliases='hsa-mir-6070', description='microRNA 6070', other_designations='microRNA mir-6070', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43609887, end_position_on_the_genomic_accession=43609989, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102464826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464826, CurrentID=0, Status='live', Symbol='MIR6073', Aliases='hsa-mir-6073', description='microRNA 6073', other_designations='microRNA mir-6073', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15969533, end_position_on_the_genomic_accession=15969621, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102464827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464827, CurrentID=0, Status='live', Symbol='MIR6074', Aliases='hsa-mir-6074', description='microRNA 6074', other_designations='microRNA mir-6074', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66023620, end_position_on_the_genomic_accession=66023726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102464828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464828, CurrentID=0, Status='live', Symbol='MIR6076', Aliases='hsa-mir-6076', description='microRNA 6076', other_designations='microRNA mir-6076', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49966399, end_position_on_the_genomic_accession=49966511, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464829, CurrentID=0, Status='live', Symbol='MIR6078', Aliases='hsa-mir-6078', description='microRNA 6078', other_designations='microRNA mir-6078', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3991160, end_position_on_the_genomic_accession=3991259, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464830, CurrentID=0, Status='live', Symbol='MIR6079', Aliases='hsa-mir-6079', description='microRNA 6079', other_designations='microRNA mir-6079', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43838622, end_position_on_the_genomic_accession=43838683, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464831, CurrentID=0, Status='live', Symbol='MIR6080', Aliases='hsa-mir-6080', description='microRNA 6080', other_designations='microRNA mir-6080', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64780759, end_position_on_the_genomic_accession=64780824, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  359744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359744, CurrentID=0, Status='live', Symbol='MRPL35P3', Aliases='', description='mitochondrial ribosomal protein L35 pseudogene 3', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74527584, end_position_on_the_genomic_accession=74527953, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102464833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464833, CurrentID=0, Status='live', Symbol='MIR6084', Aliases='hsa-mir-6084', description='microRNA 6084', other_designations='microRNA mir-6084', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20633679, end_position_on_the_genomic_accession=20633788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464834, CurrentID=0, Status='live', Symbol='MIR6085', Aliases='hsa-mir-6085', description='microRNA 6085', other_designations='microRNA mir-6085', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62343029, end_position_on_the_genomic_accession=62343138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464835, CurrentID=0, Status='live', Symbol='MIR6087', Aliases='hsa-mir-6087', description='microRNA 6087', other_designations='microRNA mir-6087', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109054542, end_position_on_the_genomic_accession=109054590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464836, CurrentID=0, Status='live', Symbol='MIR6088', Aliases='hsa-mir-6088', description='microRNA 6088', other_designations='microRNA mir-6088', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45436654, end_position_on_the_genomic_accession=45436704, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102464837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102464837, CurrentID=0, Status='live', Symbol='MIR6089', Aliases='MIR6089-1-2, hsa-mir-6089-1, hsa-mir-6089-2, MIR6089', description='microRNA 6089', other_designations='microRNA 6089-1|microRNA 6089-2|microRNA mir-6089-1|microRNA mir-6089-2', map_location='Xp22.3 and Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2609191, end_position_on_the_genomic_accession=2609254, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  359750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359750, CurrentID=0, Status='live', Symbol='MRPL50P3', Aliases='', description='mitochondrial ribosomal protein L50 pseudogene 3', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178344843, end_position_on_the_genomic_accession=178345319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359751, CurrentID=0, Status='live', Symbol='MRPL50P4', Aliases='', description='mitochondrial ribosomal protein L50 pseudogene 4', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59551399, end_position_on_the_genomic_accession=59551878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359752, CurrentID=0, Status='live', Symbol='MRPL53P1', Aliases='', description='mitochondrial ribosomal protein L53 pseudogene 1', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112625872, end_position_on_the_genomic_accession=112626225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359753, CurrentID=0, Status='live', Symbol='MRPS10P1', Aliases='', description='mitochondrial ribosomal protein S10 pseudogene 1', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169990075, end_position_on_the_genomic_accession=169990205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359754, CurrentID=0, Status='live', Symbol='MRPS10P5', Aliases='', description='mitochondrial ribosomal protein S10 pseudogene 5', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37835821, end_position_on_the_genomic_accession=37836399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359755, CurrentID=0, Status='live', Symbol='MRPS15P1', Aliases='', description='mitochondrial ribosomal protein S15 pseudogene 1', other_designations='', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73483190, end_position_on_the_genomic_accession=73483945, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359756, CurrentID=0, Status='live', Symbol='MRPS15P2', Aliases='', description='mitochondrial ribosomal protein S15 pseudogene 2', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44290638, end_position_on_the_genomic_accession=44290930, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359757, CurrentID=0, Status='live', Symbol='MRPS17P3', Aliases='', description='mitochondrial ribosomal protein S17 pseudogene 3', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=78525103, end_position_on_the_genomic_accession=78525491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359758, CurrentID=0, Status='live', Symbol='MRPS17P5', Aliases='', description='mitochondrial ribosomal protein S17 pseudogene 5', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127909831, end_position_on_the_genomic_accession=127910016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359759, CurrentID=0, Status='live', Symbol='MRPS17P6', Aliases='', description='mitochondrial ribosomal protein S17 pseudogene 6', other_designations='', map_location='14p11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21291239, end_position_on_the_genomic_accession=21291583, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359760, CurrentID=0, Status='live', Symbol='MRPS17P9', Aliases='', description='mitochondrial ribosomal protein S17 pseudogene 9', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119379687, end_position_on_the_genomic_accession=119380272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100526648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526648, CurrentID=0, Status='live', Symbol='MIR1273E', Aliases='hsa-mir-1273e', description='microRNA 1273e', other_designations='microRNA mir-1273e', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64425069, end_position_on_the_genomic_accession=64425170, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  359762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359762, CurrentID=0, Status='live', Symbol='MRPS18BP1', Aliases='', description='mitochondrial ribosomal protein S18B pseudogene 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=216201058, end_position_on_the_genomic_accession=216201826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359763, CurrentID=0, Status='live', Symbol='MRPS18BP2', Aliases='', description='mitochondrial ribosomal protein S18B pseudogene 2', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139668535, end_position_on_the_genomic_accession=139669310, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359764, CurrentID=0, Status='live', Symbol='MRPS18CP3', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene 3', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18086882, end_position_on_the_genomic_accession=18087636, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359765, CurrentID=0, Status='live', Symbol='MRPS18CP5', Aliases='', description='mitochondrial ribosomal protein S18C pseudogene 5', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25928046, end_position_on_the_genomic_accession=25928595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359766, CurrentID=0, Status='live', Symbol='MRPS21P2', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 2', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157861139, end_position_on_the_genomic_accession=157861616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359767, CurrentID=0, Status='live', Symbol='MRPS21P3', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 3', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197363784, end_position_on_the_genomic_accession=197364212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359768, CurrentID=0, Status='live', Symbol='MRPS21P4', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 4', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36410490, end_position_on_the_genomic_accession=36410634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359769, CurrentID=0, Status='live', Symbol='MRPS21P5', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 5', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28004666, end_position_on_the_genomic_accession=28004935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359770, CurrentID=0, Status='live', Symbol='MRPS21P6', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 6', other_designations='', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125166619, end_position_on_the_genomic_accession=125167096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359771, CurrentID=0, Status='live', Symbol='MRPS21P7', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 7', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49708001, end_position_on_the_genomic_accession=49708260, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359772, CurrentID=0, Status='live', Symbol='MRPS21P9', Aliases='', description='mitochondrial ribosomal protein S21 pseudogene 9', other_designations='', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50450469, end_position_on_the_genomic_accession=50450735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359773, CurrentID=0, Status='live', Symbol='MRPS22P1', Aliases='', description='mitochondrial ribosomal protein S22 pseudogene 1', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=87809282, end_position_on_the_genomic_accession=87809738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359774, CurrentID=0, Status='live', Symbol='DAP3P2', Aliases='MRPS29P2', description='death associated protein 3 pseudogene 2', other_designations='mitochondrial ribosomal protein S29 pseudogene 2', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171491417, end_position_on_the_genomic_accession=171491931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359775, CurrentID=0, Status='live', Symbol='MRPS33P2', Aliases='', description='mitochondrial ribosomal protein S33 pseudogene 2', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38006784, end_position_on_the_genomic_accession=38007105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359776, CurrentID=0, Status='live', Symbol='MRPS33P3', Aliases='', description='mitochondrial ribosomal protein S33 pseudogene 3', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=115142931, end_position_on_the_genomic_accession=115143497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359777, CurrentID=0, Status='live', Symbol='MRPS33P4', Aliases='', description='mitochondrial ribosomal protein S33 pseudogene 4', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52511299, end_position_on_the_genomic_accession=52511550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359778, CurrentID=0, Status='live', Symbol='MRPS35P2', Aliases='', description='mitochondrial ribosomal protein S35 pseudogene 2', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98402543, end_position_on_the_genomic_accession=98402721, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359779, CurrentID=0, Status='live', Symbol='MRPS35P3', Aliases='', description='mitochondrial ribosomal protein S35 pseudogene 3', other_designations='', map_location='10q21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=57982473, end_position_on_the_genomic_accession=57983011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359780, CurrentID=0, Status='live', Symbol='MRPS6P1', Aliases='MRPS6P3', description='mitochondrial ribosomal protein S6 pseudogene 1', other_designations='', map_location='1p35', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26263035, end_position_on_the_genomic_accession=26263398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359781, CurrentID=0, Status='live', Symbol='MRPS6P2', Aliases='', description='mitochondrial ribosomal protein S6 pseudogene 2', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50846469, end_position_on_the_genomic_accession=50846842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359782, CurrentID=0, Status='live', Symbol='MRPS6P4', Aliases='', description='mitochondrial ribosomal protein S6 pseudogene 4', other_designations='', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89319299, end_position_on_the_genomic_accession=89319606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359783, CurrentID=0, Status='live', Symbol='MRPS7P1', Aliases='', description='mitochondrial ribosomal protein S7 pseudogene 1', other_designations='', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36795263, end_position_on_the_genomic_accession=36795967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376316, CurrentID=0, Status='live', Symbol='LOC105376316', Aliases='', description='uncharacterized LOC105376316', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135577426, end_position_on_the_genomic_accession=135585270, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  359787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359787, CurrentID=0, Status='live', Symbol='DPPA3', Aliases='STELLA', description='developmental pluripotency associated 3', other_designations='stella-related protein', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7711454, end_position_on_the_genomic_accession=7717559, orientation='plus', exon_count=4, OMIM=608408, no_hdr0=''),
  359791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359791, CurrentID=0, Status='live', Symbol='MED14P1', Aliases='CRSP2P, CXorf4P, MED14P', description='mediator complex subunit 14 pseudogene 1', other_designations='cofactor required for Sp1 transcriptional activation, subunit 2, 150kDa pseudogene|cofactor required for Sp1 transcriptional activation, subunit 2, pseudogene', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12618985, end_position_on_the_genomic_accession=12634401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359792, CurrentID=0, Status='live', Symbol='EIF4A1P2', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 2', other_designations='eukaryotic translation initiation factor 4A, isoform 1 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5337603, end_position_on_the_genomic_accession=5338980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359793, CurrentID=0, Status='live', Symbol='AGKP1', Aliases='AGKY1', description='acylglycerol kinase pseudogene 1', other_designations='acylglycerol kinase Y-linked pseudogene 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14639044, end_position_on_the_genomic_accession=14641313, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359794, CurrentID=0, Status='live', Symbol='TUSC2P1', Aliases='TUSC2P', description='tumor suppressor candidate 2 pseudogene 1', other_designations='lung cancer candidate FUS1 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5887504, end_position_on_the_genomic_accession=5888740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359795, CurrentID=0, Status='live', Symbol='NEFLP1', Aliases='', description='neurofilament, light polypeptide pseudogene 1', other_designations='neurofilament, light polypeptide 68kDa pseudogene', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21221351, end_position_on_the_genomic_accession=21221782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359796, CurrentID=0, Status='live', Symbol='KDM5DP1', Aliases='SMCYP', description='lysine (K)-specific demethylase 5D pseudogene 1', other_designations='Smcy homolog, Y chromosome pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20569690, end_position_on_the_genomic_accession=20570564, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359797, CurrentID=0, Status='live', Symbol='MRPS5P4', Aliases='', description='mitochondrial ribosomal protein S5 pseudogene 4', other_designations='', map_location='18q21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60519834, end_position_on_the_genomic_accession=60520171, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359798, CurrentID=0, Status='live', Symbol='UBE2V1P3', Aliases='', description='ubiquitin conjugating enzyme E2 variant 1 pseudogene 3', other_designations='ubiquitin-conjugating enzyme E2 variant 1 pseudogene 3', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3864609, end_position_on_the_genomic_accession=3866779, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359799, CurrentID=0, Status='live', Symbol='USP12PY', Aliases='USP12P3', description='ubiquitin specific peptidase 12 pseudogene, Y-linked', other_designations='ubiquitin specific peptidase 12 pseudogene 3|ubiquitin specific protease 12 pseudogene 3|ubiquitin specific protease 12-like 1 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3682558, end_position_on_the_genomic_accession=3684842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359800, CurrentID=0, Status='live', Symbol='VDAC1P6', Aliases='', description='voltage-dependent anion channel 1 pseudogene 6', other_designations='voltage-dependent anion channel 2 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5206367, end_position_on_the_genomic_accession=5208078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359802, CurrentID=0, Status='live', Symbol='MRPS10P2', Aliases='MRPS10P3, MRPS10P4', description='mitochondrial ribosomal protein S10 pseudogene 2', other_designations='', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4414560, end_position_on_the_genomic_accession=4414963, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359803, CurrentID=0, Status='live', Symbol='MRPS17P1', Aliases='MRPS17P2', description='mitochondrial ribosomal protein S17 pseudogene 1', other_designations='', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44991677, end_position_on_the_genomic_accession=44992253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359804, CurrentID=0, Status='live', Symbol='DAP3P1', Aliases='DAP3P, MRPS29P1', description='death associated protein 3 pseudogene 1', other_designations='DAP3 pseudogene|mitochondrial ribosomal protein S29 pseudogene 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155598813, end_position_on_the_genomic_accession=155602197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359805, CurrentID=0, Status='live', Symbol='PCNAP2', Aliases='pF2PCNA', description='proliferating cell nuclear antigen pseudogene 2', other_designations='PCNA pseudogene pF2PCNA', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99162554, end_position_on_the_genomic_accession=99162690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359806, CurrentID=0, Status='live', Symbol='PCNAP1', Aliases='p1PCNA', description='proliferating cell nuclear antigen pseudogene 1', other_designations='PCNA pseudogene p1PCNA', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99160593, end_position_on_the_genomic_accession=99161647, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376320, CurrentID=0, Status='live', Symbol='LOC105376320', Aliases='', description='uncharacterized LOC105376320', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135708265, end_position_on_the_genomic_accession=135714313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371501, CurrentID=0, Status='live', Symbol='LOC105371501', Aliases='', description='uncharacterized LOC105371501', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5030806, end_position_on_the_genomic_accession=5053374, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  359817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359817, CurrentID=0, Status='live', Symbol='MRPL30P2', Aliases='MRPL30P3', description='mitochondrial ribosomal protein L30 pseudogene 2', other_designations='', map_location='12p11.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31477476, end_position_on_the_genomic_accession=31477857, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359819, CurrentID=0, Status='live', Symbol='LOC359819', Aliases='', description='mitochondrial ribosomal protein L39 pseudogene', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67689960, end_position_on_the_genomic_accession=67690439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359821, CurrentID=0, Status='live', Symbol='MRPL42P5', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 5', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40531884, end_position_on_the_genomic_accession=40532550, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  359822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359822, CurrentID=0, Status='live', Symbol='KCNIP4-IT1', Aliases='NCRNA00099, UM9(5), UM9-5', description='KCNIP4 intronic transcript 1', other_designations='KCNIP4 intronic transcript 1 (non-protein coding)|non coding transcript UM9(5)|non-coding transcript UM9(5)', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21843341, end_position_on_the_genomic_accession=21853188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  359825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359825, CurrentID=0, Status='live', Symbol='PSORS9', Aliases='', description='psoriasis susceptibility 9', other_designations='', map_location='4q31-q34', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607857, no_hdr0=''),
  105376323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376323, CurrentID=0, Status='live', Symbol='LOC105376323', Aliases='', description='uncharacterized LOC105376323', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136107829, end_position_on_the_genomic_accession=136109787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927895, CurrentID=0, Status='live', Symbol='LOC101927895', Aliases='', description='uncharacterized LOC101927895', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19210365, end_position_on_the_genomic_accession=19240704, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  359844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359844, CurrentID=0, Status='live', Symbol='PRDX2P1', Aliases='', description='peroxiredoxin 2 pseudogene 1', other_designations='', map_location='13p12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30296068, end_position_on_the_genomic_accession=30296987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359845, CurrentID=0, Status='live', Symbol='FAM101B', Aliases='CFM1', description='family with sequence similarity 101 member B', other_designations='family with sequence similarity 101, member B|protein FAM101B|refilinB|regulator of filamin protein B', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=439978, end_position_on_the_genomic_accession=445940, orientation='minus', exon_count=2, OMIM=615928, no_hdr0=''),
  100129078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129078, CurrentID=0, Status='live', Symbol='LOC100129078', Aliases='', description='tetratricopeptide repeat domain 5 pseudogene', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41911256, end_position_on_the_genomic_accession=41912398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100498859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100498859, CurrentID=0, Status='live', Symbol='LINC00881', Aliases='', description='long intergenic non-protein coding RNA 881', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157089881, end_position_on_the_genomic_accession=157101135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376328, CurrentID=0, Status='live', Symbol='LOC105376328', Aliases='', description='uncharacterized LOC105376328', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137155022, end_position_on_the_genomic_accession=137156661, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376330, CurrentID=0, Status='live', Symbol='LOC105376330', Aliases='', description='uncharacterized LOC105376330', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137814482, end_position_on_the_genomic_accession=137834342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422137, CurrentID=0, Status='live', Symbol='LOC100422137', Aliases='', description='olfactory receptor, family 4, subfamily A, member 15 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375868, CurrentID=0, Status='live', Symbol='LOC105375868', Aliases='', description='uncharacterized LOC105375868', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60799244, end_position_on_the_genomic_accession=60808632, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927897, CurrentID=0, Status='live', Symbol='LOC101927897', Aliases='', description='uncharacterized LOC101927897', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79456753, end_position_on_the_genomic_accession=79467434, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422138, CurrentID=0, Status='live', Symbol='LOC100422138', Aliases='', description='olfactory receptor, family 4, subfamily C, member 12 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376341, CurrentID=0, Status='live', Symbol='LOC105376341', Aliases='', description='uncharacterized LOC105376341', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=951571, end_position_on_the_genomic_accession=956386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  359940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359940, CurrentID=0, Status='live', Symbol='CDK8PS', Aliases='', description='cyclin-dependent kinase 8 pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  359941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359941, CurrentID=0, Status='live', Symbol='CDK4PS', Aliases='', description='cyclin-dependent kinase 4 pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105433994, end_position_on_the_genomic_accession=105434821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644841, CurrentID=0, Status='live', Symbol='LOC644841', Aliases='', description='uncharacterized LOC644841', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=154994350, end_position_on_the_genomic_accession=154999070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  359948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359948, CurrentID=0, Status='live', Symbol='IRF2BP2', Aliases='', description='interferon regulatory factor 2 binding protein 2', other_designations='IRF-2-binding protein 2', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234604269, end_position_on_the_genomic_accession=234609525, orientation='minus', exon_count=2, OMIM=615332, no_hdr0=''),
  105375870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375870, CurrentID=0, Status='live', Symbol='LOC105375870', Aliases='', description='uncharacterized LOC105375870', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61350176, end_position_on_the_genomic_accession=61379509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376347, CurrentID=0, Status='live', Symbol='LOC105376347', Aliases='', description='uncharacterized LOC105376347', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2071798, end_position_on_the_genomic_accession=2072729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376351, CurrentID=0, Status='live', Symbol='LOC105376351', Aliases='', description='uncharacterized LOC105376351', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2422114, end_position_on_the_genomic_accession=2501968, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  359996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359996, CurrentID=0, Status='live', Symbol='SERBP1P2', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene 2', other_designations='PAI 1 mRNA binding protein pseudogene|PAI-1 mRNA-binding protein pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=4801010, end_position_on_the_genomic_accession=4802929, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359997, CurrentID=0, Status='live', Symbol='ATP5JP1', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F6 pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F6 pseudogene 1|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F6 pseudogene 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6901134, end_position_on_the_genomic_accession=6901468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  359998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359998, CurrentID=0, Status='live', Symbol='CHRFAM7AP2', Aliases='', description='CHRFAM7A pseudogene 2', other_designations='CHRNA7 (cholinergic receptor, nicotinic, alpha polypeptide 7, exons 5-10) and FAM7A (family with sequence similarity 7A, exons A-E) fusion pseudogene, centromeric', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24205464, end_position_on_the_genomic_accession=24206069, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  359999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=359999, CurrentID=0, Status='live', Symbol='CHRFAM7AP1', Aliases='', description='CHRFAM7A pseudogene 1', other_designations='', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25464069, end_position_on_the_genomic_accession=25464674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360001, CurrentID=0, Status='live', Symbol='IFNWP18', Aliases='LEIF-M', description='interferon, omega 1 pseudogene 18', other_designations='interferon, omega 18 (pseudogene)|pseudo-IFN-alpha-m processed pseudogene', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21209192, end_position_on_the_genomic_accession=21209899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360002, CurrentID=0, Status='live', Symbol='TAF9P1', Aliases='', description='TATA-box binding protein associated factor 9 pseudogene 1', other_designations='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene 1|TAF9 pseudogene 1', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17628413, end_position_on_the_genomic_accession=17629792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360003, CurrentID=0, Status='live', Symbol='TAF9P2', Aliases='', description='TATA-box binding protein associated factor 9 pseudogene 2', other_designations='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene 2|TAF9 pseudogene 2', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18276280, end_position_on_the_genomic_accession=18277495, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  81846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=81846, CurrentID=0, Status='live', Symbol='SBF2', Aliases='CMT4B2, DENND7B, MTMR13', description='SET binding factor 2', other_designations='DENN/MADD domain containing 7B', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9778667, end_position_on_the_genomic_accession=10294207, orientation='minus', exon_count=43, OMIM=607697, no_hdr0=''),
  105376353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376353, CurrentID=0, Status='live', Symbol='LOC105376353', Aliases='', description='mucin-17-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3243059, end_position_on_the_genomic_accession=3253291, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  360008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360008, CurrentID=0, Status='live', Symbol='GAPDHP17', Aliases='GAPDHP18', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 17', other_designations='actin, gamma 1 pseudogene|glyceraldehyde 3 phosphate dehydrogenase pseudogene 18', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20861379, end_position_on_the_genomic_accession=20862541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360009, CurrentID=0, Status='live', Symbol='CYCSP49', Aliases='CYCSP54, HCP49', description='cytochrome c, somatic pseudogene 49', other_designations='cytochrome c, somatic pseudogene 54', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26549364, end_position_on_the_genomic_accession=26550366, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376354, CurrentID=0, Status='live', Symbol='LOC105376354', Aliases='', description='uncharacterized LOC105376354', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3238087, end_position_on_the_genomic_accession=3258678, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  360014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360014, CurrentID=0, Status='live', Symbol='DLGAP5P1', Aliases='', description='discs, large (Drosophila) homolog-associated protein 5 pseudogene 1', other_designations='discs, large homolog 7 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6158798, end_position_on_the_genomic_accession=6159833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360015, CurrentID=0, Status='live', Symbol='RFTN1P1', Aliases='', description='raftlin, lipid raft linker 1 pseudogene 1', other_designations='raft-linking protein pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7712985, end_position_on_the_genomic_accession=7776705, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360016, CurrentID=0, Status='live', Symbol='CICP2', Aliases='', description='capicua transcriptional repressor pseudogene 2', other_designations='capicua homolog pseudogene 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24277326, end_position_on_the_genomic_accession=24280818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360017, CurrentID=0, Status='live', Symbol='CICP1', Aliases='', description='capicua transcriptional repressor pseudogene 1', other_designations='capicua homolog pseudogene 1', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25389329, end_position_on_the_genomic_accession=25392822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360018, CurrentID=0, Status='live', Symbol='PRRC2CP1', Aliases='BAT2D1P1, BAT2L2P1', description='proline-rich coiled-coil 2C pseudogene 1', other_designations='BAT2 domain containing 1 pseudogene 1|HBxAg transactivated protein 2 pseudogene|HLA-B associated transcript 2-like 2 pseudogene 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3549813, end_position_on_the_genomic_accession=3550585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360019, CurrentID=0, Status='live', Symbol='KRT18P10', Aliases='', description='keratin 18 pseudogene 10', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=5573113, end_position_on_the_genomic_accession=5574488, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360020, CurrentID=0, Status='live', Symbol='PPP1R12BP2', Aliases='', description='protein phosphatase 1 regulatory subunit 12B pseudogene 2', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 12B Y-linked pseudogene 2|protein phosphatase 1, regulatory (inhibitor) subunit 12B pseudogene|protein phosphatase 1, regulatory subunit 12B Y-linked pseudogene 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23379382, end_position_on_the_genomic_accession=23392797, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360021, CurrentID=0, Status='live', Symbol='PPP1R12BP1', Aliases='PPP1R12BP, PPP1R12BPY1', description='protein phosphatase 1 regulatory subunit 12B pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 12B Y-linked pseudogene 1|protein phosphatase 1, regulatory (inhibitor) subunit 12B pseudogene, Y-linked 1|protein phosphatase 1, regulatory subunit 12B pseudogene, Y-linked 1', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26277823, end_position_on_the_genomic_accession=26348678, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360022, CurrentID=0, Status='live', Symbol='PSIP1P2', Aliases='', description='PC4 and SFRS1 interacting protein 1 pseudogene 2', other_designations='PC4 and SFRS1 interacting protein 2 pseudogene', map_location='Yq11', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=13718065, end_position_on_the_genomic_accession=13718837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360023, CurrentID=0, Status='live', Symbol='ZBTB41', Aliases='FRBZ1, ZNF924', description='zinc finger and BTB domain containing 41', other_designations='FRBZ1 protein (FRBZ1)', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197153684, end_position_on_the_genomic_accession=197200542, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  360024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360024, CurrentID=0, Status='live', Symbol='GOLGA6L11P', Aliases='GOLGA6L15P', description='golgin A6 family-like 11, pseudogene', other_designations='golgin A6 family-like 15, pseudogene|intersectin 2 pseudogene', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24169911, end_position_on_the_genomic_accession=24172669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360025, CurrentID=0, Status='live', Symbol='GOLGA6L12P', Aliases='', description='golgin A6 family-like 12, pseudogene', other_designations='intersectin 2 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24248478, end_position_on_the_genomic_accession=24249860, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360026, CurrentID=0, Status='live', Symbol='GOLGA6L13P', Aliases='', description='golgin A6 family-like 13, pseudogene', other_designations='intersectin 2 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25420283, end_position_on_the_genomic_accession=25421665, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360027, CurrentID=0, Status='live', Symbol='GOLGA6L14P', Aliases='', description='golgin A6 family-like 14, pseudogene', other_designations='intersectin 2 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25492674, end_position_on_the_genomic_accession=25495599, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360028, CurrentID=0, Status='live', Symbol='TMEM167AP1', Aliases='', description='transmembrane protein 167A pseudogene 1', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20901853, end_position_on_the_genomic_accession=20907612, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360029, CurrentID=0, Status='live', Symbol='FAM199YP', Aliases='FAM199Y', description='family with sequence similarity 199, Y-linked, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6950426, end_position_on_the_genomic_accession=6950616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360030, CurrentID=0, Status='live', Symbol='NANOGNB', Aliases='', description='NANOG neighbor homeobox', other_designations='homeobox C14|homeobox protein C14', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7765216, end_position_on_the_genomic_accession=7774121, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  780901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780901, CurrentID=0, Status='live', Symbol='NYS5', Aliases='', description='Nystagmus 5, infantile periodic alternating', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300589, no_hdr0=''),
  105376358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376358, CurrentID=0, Status='live', Symbol='LOC105376358', Aliases='', description='uncharacterized LOC105376358', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3394079, end_position_on_the_genomic_accession=3397455, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376359, CurrentID=0, Status='live', Symbol='LOC105376359', Aliases='', description='uncharacterized LOC105376359', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3485183, end_position_on_the_genomic_accession=3486998, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465133, CurrentID=0, Status='live', Symbol='MIR6125', Aliases='hsa-mir-6125, mir-6125', description='microRNA 6125', other_designations='microRNA mir-6125', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62260359, end_position_on_the_genomic_accession=62260454, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465134, CurrentID=0, Status='live', Symbol='MIR6126', Aliases='hsa-mir-6126', description='microRNA 6126', other_designations='microRNA mir-6126', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3485381, end_position_on_the_genomic_accession=3485469, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465135, CurrentID=0, Status='live', Symbol='MIR6128', Aliases='hsa-mir-6128', description='microRNA 6128', other_designations='microRNA mir-6128', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56743873, end_position_on_the_genomic_accession=56743981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465136, CurrentID=0, Status='live', Symbol='MIR378J', Aliases='hsa-mir-378j', description='microRNA 378j', other_designations='microRNA mir-378j', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37614931, end_position_on_the_genomic_accession=37615039, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465137, CurrentID=0, Status='live', Symbol='MIR6129', Aliases='hsa-mir-6129', description='microRNA 6129', other_designations='microRNA mir-6129', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49288346, end_position_on_the_genomic_accession=49288454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465138, CurrentID=0, Status='live', Symbol='MIR6131', Aliases='hsa-mir-6131', description='microRNA 6131', other_designations='microRNA mir-6131', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10478037, end_position_on_the_genomic_accession=10478145, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465139, CurrentID=0, Status='live', Symbol='MIR6133', Aliases='hsa-mir-6133', description='microRNA 6133', other_designations='microRNA mir-6133', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133290881, end_position_on_the_genomic_accession=133290988, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465140, CurrentID=0, Status='live', Symbol='MIR6134', Aliases='hsa-mir-6134', description='microRNA 6134', other_designations='microRNA mir-6134', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=28495555, end_position_on_the_genomic_accession=28495663, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465141, CurrentID=0, Status='live', Symbol='MIR6165', Aliases='hsa-mir-6165', description='microRNA 6165', other_designations='microRNA mir-6165', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49510817, end_position_on_the_genomic_accession=49510900, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  388780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388780, CurrentID=0, Status='live', Symbol='LOC388780', Aliases='', description='uncharacterized LOC388780', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2206468, end_position_on_the_genomic_accession=2258031, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376361, CurrentID=0, Status='live', Symbol='LOC105376361', Aliases='', description='uncharacterized LOC105376361', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3750852, end_position_on_the_genomic_accession=3756540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376362, CurrentID=0, Status='live', Symbol='LOC105376362', Aliases='', description='uncharacterized LOC105376362', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3756546, end_position_on_the_genomic_accession=3758096, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375874, CurrentID=0, Status='live', Symbol='LOC105375874', Aliases='', description='uncharacterized LOC105375874', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63435761, end_position_on_the_genomic_accession=63457186, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376364, CurrentID=0, Status='live', Symbol='LOC105376364', Aliases='', description='uncharacterized LOC105376364', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3806414, end_position_on_the_genomic_accession=3808517, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376365, CurrentID=0, Status='live', Symbol='LOC105376365', Aliases='', description='uncharacterized LOC105376365', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3834008, end_position_on_the_genomic_accession=3834728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  780910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780910, CurrentID=0, Status='live', Symbol='PPR2', Aliases='', description='Photoparoxysmal response 2', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609572, no_hdr0=''),
  387695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387695, CurrentID=0, Status='live', Symbol='C10orf99', Aliases='AP-57, CSBF, UNQ1833', description='chromosome 10 open reading frame 99', other_designations='RLLV1833|anti-microbial peptide with 57 amino acid residues|antimicrobial peptide-57|colon-derived SUSD2 binding factor', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84173764, end_position_on_the_genomic_accession=84185294, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376368, CurrentID=0, Status='live', Symbol='LOC105376368', Aliases='', description='uncharacterized LOC105376368', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4150239, end_position_on_the_genomic_accession=4174518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376369, CurrentID=0, Status='live', Symbol='LOC105376369', Aliases='', description='uncharacterized LOC105376369', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4247273, end_position_on_the_genomic_accession=4256851, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376370, CurrentID=0, Status='live', Symbol='LOC105376370', Aliases='', description='uncharacterized LOC105376370', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4334396, end_position_on_the_genomic_accession=4336838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376371, CurrentID=0, Status='live', Symbol='LOC105376371', Aliases='', description='uncharacterized LOC105376371', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4369184, end_position_on_the_genomic_accession=4370565, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422145, CurrentID=0, Status='live', Symbol='LOC100422145', Aliases='', description='olfactory receptor, family 2, subfamily AT, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387700, CurrentID=0, Status='live', Symbol='SLC16A12', Aliases='CJMG, CRT2, MCT12', description='solute carrier family 16 member 12', other_designations='creatine transporter 2|monocarboxylic acid transporter 12|solute carrier family 16, member 12 (monocarboxylic acid transporter 12)', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89430294, end_position_on_the_genomic_accession=89535556, orientation='minus', exon_count=8, OMIM=611910, no_hdr0=''),
  105376373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376373, CurrentID=0, Status='live', Symbol='LOC105376373', Aliases='', description='uncharacterized LOC105376373', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4669904, end_position_on_the_genomic_accession=4679654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  360132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360132, CurrentID=0, Status='live', Symbol='FKBP9P1', Aliases='FKBP9L', description='FK506 binding protein 9 pseudogene 1', other_designations='FK506 binding protein 9-like', map_location='7p11.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55681074, end_position_on_the_genomic_accession=55704567, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376945, CurrentID=0, Status='live', Symbol='LOC105376945', Aliases='', description='uncharacterized LOC105376945', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6828631, end_position_on_the_genomic_accession=6847774, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387703, CurrentID=0, Status='live', Symbol='XRCC6P1', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 6 pseudogene 1', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93206655, end_position_on_the_genomic_accession=93209164, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376376, CurrentID=0, Status='live', Symbol='LOC105376376', Aliases='', description='uncharacterized LOC105376376', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5065263, end_position_on_the_genomic_accession=5069616, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376377, CurrentID=0, Status='live', Symbol='LOC105376377', Aliases='', description='uncharacterized LOC105376377', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5120191, end_position_on_the_genomic_accession=5123956, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  360155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360155, CurrentID=0, Status='live', Symbol='CYCSP52', Aliases='HC6, HCP2', description='cytochrome c, somatic pseudogene 52', other_designations='somatic cytochrome c (HC6) processed pseudogene', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157128362, end_position_on_the_genomic_accession=157128671, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  360156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360156, CurrentID=0, Status='live', Symbol='CYCSP53', Aliases='HCP3', description='cytochrome c, somatic pseudogene 53', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171444684, end_position_on_the_genomic_accession=171444963, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499165, CurrentID=0, Status='live', Symbol='FGQTL1', Aliases='', description='Fasting plasma glucose level QTL 1', other_designations='', map_location='2q24-q32', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612108, no_hdr0=''),
  100499166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499166, CurrentID=0, Status='live', Symbol='LPRS6', Aliases='', description='Leprosy, susceptiblity to, 6', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613407, no_hdr0=''),
  100499167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499167, CurrentID=0, Status='live', Symbol='EOE2', Aliases='', description='Esophagitis, eosinophilic, 2', other_designations='', map_location='5q22', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613412, no_hdr0=''),
  100499168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499168, CurrentID=0, Status='live', Symbol='FGQTL6', Aliases='BWQTL3, PGQTL1', description='Fasting plasma glucose level QTL 6', other_designations='', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613460, no_hdr0=''),
  100499169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499169, CurrentID=0, Status='live', Symbol='FGQTL4', Aliases='', description='Fasting plasma glucose level QTL 4', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613462, no_hdr0=''),
  102465250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465250, CurrentID=0, Status='live', Symbol='MIR6503', Aliases='hsa-mir-6503', description='microRNA 6503', other_designations='microRNA mir-6503', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60209071, end_position_on_the_genomic_accession=60209156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  780923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780923, CurrentID=0, Status='live', Symbol='IHPS2', Aliases='', description='Pyloric stenosis, infantile hypertrophic, 2', other_designations='', map_location='16p13-p12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610260, no_hdr0=''),
  102465252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465252, CurrentID=0, Status='live', Symbol='MIR6506', Aliases='hsa-mir-6506', description='microRNA 6506', other_designations='microRNA mir-6506', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15611030, end_position_on_the_genomic_accession=15611095, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465253, CurrentID=0, Status='live', Symbol='MIR6507', Aliases='hsa-mir-6507', description='microRNA 6507', other_designations='microRNA mir-6507', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98924499, end_position_on_the_genomic_accession=98924568, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465254, CurrentID=0, Status='live', Symbol='MIR6509', Aliases='hsa-mir-6509', description='microRNA 6509', other_designations='microRNA mir-6509', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135206994, end_position_on_the_genomic_accession=135207078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465255, CurrentID=0, Status='live', Symbol='MIR6512', Aliases='hsa-mir-6512, mir-6512', description='microRNA 6512', other_designations='microRNA mir-6512', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177313806, end_position_on_the_genomic_accession=177313882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465256, CurrentID=0, Status='live', Symbol='MIR6513', Aliases='hsa-mir-6513, mir-6513', description='microRNA 6513', other_designations='microRNA mir-6513', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218280125, end_position_on_the_genomic_accession=218280188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100499177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499177, CurrentID=0, Status='live', Symbol='THAP9-AS1', Aliases='', description='THAP9 antisense RNA 1', other_designations='THAP9 antisense RNA 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82893452, end_position_on_the_genomic_accession=82900916, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  360170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360170, CurrentID=0, Status='live', Symbol='CYCSP14', Aliases='HCP14', description='cytochrome c, somatic pseudogene 14', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130022587, end_position_on_the_genomic_accession=130022883, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360171, CurrentID=0, Status='live', Symbol='CYCSP16', Aliases='HCP16', description='cytochrome c, somatic pseudogene 16', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89002730, end_position_on_the_genomic_accession=89002900, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360172, CurrentID=0, Status='live', Symbol='CYCSP17', Aliases='HCP17', description='cytochrome c, somatic pseudogene 17', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=95504182, end_position_on_the_genomic_accession=95504496, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360173, CurrentID=0, Status='live', Symbol='CYCSP18', Aliases='HCP18', description='cytochrome c, somatic pseudogene 18', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96127827, end_position_on_the_genomic_accession=96128062, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360174, CurrentID=0, Status='live', Symbol='CYCSP19', Aliases='HCP19', description='cytochrome c, somatic pseudogene 19', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121398454, end_position_on_the_genomic_accession=121398770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499183, CurrentID=0, Status='live', Symbol='BAALC-AS1', Aliases='', description='BAALC antisense RNA 1', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103156990, end_position_on_the_genomic_accession=103172140, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  360176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360176, CurrentID=0, Status='live', Symbol='CYCSP23', Aliases='HCP23', description='cytochrome c, somatic pseudogene 23', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119618156, end_position_on_the_genomic_accession=119618474, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360177, CurrentID=0, Status='live', Symbol='CYCSP24', Aliases='HC3, HCP24, HCP31', description='cytochrome c, somatic pseudogene 24', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94038030, end_position_on_the_genomic_accession=94038345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360178, CurrentID=0, Status='live', Symbol='CYCSP26', Aliases='HCP26', description='cytochrome c, somatic pseudogene 26', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58005386, end_position_on_the_genomic_accession=58005710, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360179, CurrentID=0, Status='live', Symbol='CYCSP28', Aliases='HCP28', description='cytochrome c, somatic pseudogene 28', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83328926, end_position_on_the_genomic_accession=83329251, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360180, CurrentID=0, Status='live', Symbol='CYCSP29', Aliases='HCP29', description='cytochrome c, somatic pseudogene 29', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=108822333, end_position_on_the_genomic_accession=108822650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360181, CurrentID=0, Status='live', Symbol='CYCSP32', Aliases='HCP32', description='cytochrome c, somatic pseudogene 32', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19458401, end_position_on_the_genomic_accession=19458718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360182, CurrentID=0, Status='live', Symbol='CYCSP34', Aliases='HCP34', description='cytochrome c, somatic pseudogene 34', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40863594, end_position_on_the_genomic_accession=40863906, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360183, CurrentID=0, Status='live', Symbol='CYCSP2', Aliases='HCP37', description='cytochrome c, somatic pseudogene 2', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41108865, end_position_on_the_genomic_accession=41109177, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360184, CurrentID=0, Status='live', Symbol='CYCSP38', Aliases='HC2, HCP38', description='cytochrome c, somatic pseudogene 38', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60558080, end_position_on_the_genomic_accession=60558353, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360185, CurrentID=0, Status='live', Symbol='CYCSP40', Aliases='HC7, HCP40', description='cytochrome c, somatic pseudogene 40', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77199301, end_position_on_the_genomic_accession=77199543, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360186, CurrentID=0, Status='live', Symbol='CYCSP41', Aliases='CYC1LP4, HCP41', description='cytochrome c, somatic pseudogene 41', other_designations='cytochrome c pseudogene|cytochrome c-1-like pseudogene 4', map_location='21p11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10576290, end_position_on_the_genomic_accession=10576482, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360187, CurrentID=0, Status='live', Symbol='CYCSP43', Aliases='HCP43', description='cytochrome c, somatic pseudogene 43', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69485341, end_position_on_the_genomic_accession=69485652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  360188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360188, CurrentID=0, Status='live', Symbol='CYCSP48', Aliases='HCP48', description='cytochrome c, somatic pseudogene 48', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26400954, end_position_on_the_genomic_accession=26401230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391707, CurrentID=0, Status='live', Symbol='LOC391707', Aliases='', description='chromatin accessibility complex 1 pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157204465, end_position_on_the_genomic_accession=157204855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376384, CurrentID=0, Status='live', Symbol='LOC105376384', Aliases='', description='uncharacterized LOC105376384', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5934267, end_position_on_the_genomic_accession=5946581, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  360200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360200, CurrentID=0, Status='live', Symbol='TMPRSS9', Aliases='', description='transmembrane protease, serine 9', other_designations='polyserase 1|polyserase-1|polyserase-I|polyserine protease 1|transmembrane serine protease 9', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2385235, end_position_on_the_genomic_accession=2426261, orientation='plus', exon_count=21, OMIM=610477, no_hdr0=''),
  360203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360203, CurrentID=0, Status='live', Symbol='GLT6D1', Aliases='GLTDC1, GT6M7', description='glycosyltransferase 6 domain containing 1', other_designations='galactosyltransferase family 6 domain containing 1|galactosyltransferase family 6 domain-containing 1', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135623656, end_position_on_the_genomic_accession=135641216, orientation='minus', exon_count=10, OMIM=613699, no_hdr0=''),
  360205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360205, CurrentID=0, Status='live', Symbol='PRAC2', Aliases='C17orf93, HOXB-AS5, HOXB13-AS1, NCRNA00253', description='prostate cancer susceptibility candidate 2', other_designations='HOXB cluster antisense RNA 5|prostate, rectum and colon 2|prostate, rectum and colon expressed gene 2 protein|prostate, rectum and colon expressed gene protein 2|small nuclear protein PRAC2', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48720809, end_position_on_the_genomic_accession=48724758, orientation='plus', exon_count=5, OMIM=610787, no_hdr0=''),
  100422148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422148, CurrentID=0, Status='live', Symbol='LOC100422148', Aliases='', description='olfactory receptor, family 52, subfamily B, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  256643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256643, CurrentID=0, Status='live', Symbol='CXorf23', Aliases='', description='chromosome X open reading frame 23', other_designations='', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19912860, end_position_on_the_genomic_accession=19990934, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100499223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499223, CurrentID=0, Status='live', Symbol='LOC100499223', Aliases='', description='beta-1,3-N-acetylgalactosaminyltransferase 1 (globoside blood group) pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179220480, end_position_on_the_genomic_accession=179222232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  360219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360219, CurrentID=0, Status='live', Symbol='GCRG224', Aliases='GCRG-P224', description='gastric cancer-related gene GCRG224', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610888, no_hdr0=''),
  360226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360226, CurrentID=0, Status='live', Symbol='PRSS41', Aliases='TESSP1', description='protease, serine 41', other_designations='protease, serine, 41|testis-specific serine protease 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2798476, end_position_on_the_genomic_accession=2805284, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  360227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=360227, CurrentID=0, Status='live', Symbol='OTUD4P1', Aliases='HIN1L', description='OTUD4 pseudogene 1', other_designations='HIN1-like cysteine protease|OTU domain containing 4 pseudogene', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3839459, end_position_on_the_genomic_accession=3846276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387718, CurrentID=0, Status='live', Symbol='TEX36', Aliases='C10orf122, bA383C5.1', description='testis expressed 36', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125576513, end_position_on_the_genomic_accession=125683154, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271363, CurrentID=0, Status='live', Symbol='RPS17P15', Aliases='RPS17_7_1173', description='ribosomal protein S17 pseudogene 15', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111105568, end_position_on_the_genomic_accession=111105895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422149, CurrentID=0, Status='live', Symbol='LOC100422149', Aliases='', description='olfactory receptor, family 51, subfamily G, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130684, CurrentID=0, Status='live', Symbol='LOC100130684', Aliases='', description='fibrous sheath-interacting protein 2-like', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130445085, end_position_on_the_genomic_accession=130475614, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376392, CurrentID=0, Status='live', Symbol='LOC105376392', Aliases='', description='uncharacterized LOC105376392', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7799196, end_position_on_the_genomic_accession=7807724, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100499251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499251, CurrentID=0, Status='live', Symbol='GMFBP1', Aliases='', description='glia maturation factor, beta pseudogene 1', other_designations='', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100665050, end_position_on_the_genomic_accession=100665476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499252, CurrentID=0, Status='live', Symbol='ZBTB12BP', Aliases='ZBTB12B, ZBTB12P1, ZNF920', description='zinc finger and BTB domain containing 12B, pseudogene', other_designations='zinc finger and BTB domain containing 12 pseudogene 1', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39769761, end_position_on_the_genomic_accession=39771621, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499258, CurrentID=0, Status='live', Symbol='SNRPBP1', Aliases='', description='small nuclear ribonucleoprotein polypeptides B and B1 pseudogene 1', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80377554, end_position_on_the_genomic_accession=80377835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499259, CurrentID=0, Status='live', Symbol='CKS1BP4', Aliases='', description='CDC28 protein kinase regulatory subunit 1B pseudogene 4', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83209313, end_position_on_the_genomic_accession=83210119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499260, CurrentID=0, Status='live', Symbol='AMMEC', Aliases='CXDELq22.3, DELXq22.3', description='Alport syndrome, mental retardation, midface hypoplasia, and elliptocytosis', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300194, no_hdr0=''),
  100499261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499261, CurrentID=0, Status='live', Symbol='BTF3P6', Aliases='', description='basic transcription factor 3 pseudogene 6', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33518610, end_position_on_the_genomic_accession=33519108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376419, CurrentID=0, Status='live', Symbol='LOC105376419', Aliases='', description='uncharacterized LOC105376419', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13383049, end_position_on_the_genomic_accession=13409240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376395, CurrentID=0, Status='live', Symbol='LOC105376395', Aliases='', description='uncharacterized LOC105376395', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8094295, end_position_on_the_genomic_accession=8097227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101154642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154642, CurrentID=0, Status='live', Symbol='CHDT3', Aliases='', description='Congenital heart defects, multiple types, 3', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614954, no_hdr0=''),
  101154643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154643, CurrentID=0, Status='live', Symbol='LOC101154643', Aliases='', description='cyclin-dependent kinase 2 associated protein 2 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10476400, end_position_on_the_genomic_accession=10477346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101154644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154644, CurrentID=0, Status='live', Symbol='LINCMD1', Aliases='LINC-MD1, MIR133BHG', description='long intergenic non-protein coding RNA, muscle differentiation 1', other_designations='MIR133B host gene (non-protein coding)|long noncoding RNA, muscle differentiation 1', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614933, no_hdr0=''),
  101154649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154649, CurrentID=0, Status='live', Symbol='MBNP', Aliases='', description='?Membranous nephropathy, susceptibility to', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614692, no_hdr0=''),
  105376399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376399, CurrentID=0, Status='live', Symbol='LOC105376399', Aliases='', description='uncharacterized LOC105376399', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8532214, end_position_on_the_genomic_accession=8594352, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376400, CurrentID=0, Status='live', Symbol='LOC105376400', Aliases='', description='uncharacterized LOC105376400', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8658696, end_position_on_the_genomic_accession=9065394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129728, CurrentID=0, Status='live', Symbol='LOC100129728', Aliases='', description='nucleic acid binding protein 2 pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53927193, end_position_on_the_genomic_accession=53928150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376401, CurrentID=0, Status='live', Symbol='LOC105376401', Aliases='', description='uncharacterized LOC105376401', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9296074, end_position_on_the_genomic_accession=9312349, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422151, CurrentID=0, Status='live', Symbol='LOC100422151', Aliases='', description='olfactory receptor, family 51, subfamily A, member 7 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376402, CurrentID=0, Status='live', Symbol='LOC105376402', Aliases='', description='uncharacterized LOC105376402', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9443281, end_position_on_the_genomic_accession=9878080, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105376403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376403, CurrentID=0, Status='live', Symbol='LOC105376403', Aliases='', description='uncharacterized LOC105376403', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9926726, end_position_on_the_genomic_accession=9940724, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376404, CurrentID=0, Status='live', Symbol='LOC105376404', Aliases='', description='uncharacterized LOC105376404', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11008016, end_position_on_the_genomic_accession=11012745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101154686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154686, CurrentID=0, Status='live', Symbol='MROH4P', Aliases='', description='maestro heat-like repeat family member 4, pseudogene', other_designations='LOC101154686 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142567260, end_position_on_the_genomic_accession=142575462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101154687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154687, CurrentID=0, Status='live', Symbol='LINC01081', Aliases='TCONS00024764, TCONS_00024764', description='long intergenic non-protein coding RNA 1081', other_designations='long non-coding RNA TCONS_00024764', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86225580, end_position_on_the_genomic_accession=86286247, orientation='minus', exon_count=6, OMIM=614977, no_hdr0=''),
  100422375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422375, CurrentID=0, Status='live', Symbol='LOC100422375', Aliases='', description='phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit alpha pseudogene', other_designations='phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit alpha pseudogene', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16572019, end_position_on_the_genomic_accession=16574645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376406, CurrentID=0, Status='live', Symbol='LOC105376406', Aliases='', description='uncharacterized LOC105376406', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10798187, end_position_on_the_genomic_accession=10846839, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376407, CurrentID=0, Status='live', Symbol='LOC105376407', Aliases='', description='uncharacterized LOC105376407', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10796845, end_position_on_the_genomic_accession=10922316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465427, CurrentID=0, Status='live', Symbol='MIR6715B', Aliases='hsa-mir-6715b', description='microRNA 6715b', other_designations='microRNA mir-6715b', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112299612, end_position_on_the_genomic_accession=112299688, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465428, CurrentID=0, Status='live', Symbol='MIR6717', Aliases='hsa-mir-6717', description='microRNA 6717', other_designations='microRNA mir-6717', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21023314, end_position_on_the_genomic_accession=21023386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465429, CurrentID=0, Status='live', Symbol='MIR6511B1', Aliases='MIR6511B-1, hsa-mir-6511b-1', description='microRNA 6511b-1', other_designations='microRNA mir-6511b-1|mir-6511b-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2106669, end_position_on_the_genomic_accession=2106753, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465430, CurrentID=0, Status='live', Symbol='MIR6718', Aliases='hsa-mir-6718', description='microRNA 6718', other_designations='microRNA mir-6718', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3885353, end_position_on_the_genomic_accession=3885432, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465431, CurrentID=0, Status='live', Symbol='MIR6722', Aliases='hsa-mir-6722', description='microRNA 6722', other_designations='microRNA mir-6722', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136746893, end_position_on_the_genomic_accession=136746970, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465432, CurrentID=0, Status='live', Symbol='MIR6723', Aliases='hsa-mir-6723', description='microRNA 6723', other_designations='microRNA mir-6723', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=632325, end_position_on_the_genomic_accession=632413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465433, CurrentID=0, Status='live', Symbol='MIR6724-1', Aliases='MIR6724, hsa-mir-6724, hsa-mir-6724-1', description='microRNA 6724-1', other_designations='microRNA mir-6724', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8205315, end_position_on_the_genomic_accession=8205406, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465434, CurrentID=0, Status='live', Symbol='MIR6726', Aliases='hsa-mir-6726', description='microRNA 6726', other_designations='microRNA mir-6726', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1296110, end_position_on_the_genomic_accession=1296170, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465435, CurrentID=0, Status='live', Symbol='MIR6727', Aliases='hsa-mir-6727', description='microRNA 6727', other_designations='microRNA mir-6727', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1312502, end_position_on_the_genomic_accession=1312566, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465436, CurrentID=0, Status='live', Symbol='MIR6728', Aliases='hsa-mir-6728', description='microRNA 6728', other_designations='microRNA mir-6728', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8866502, end_position_on_the_genomic_accession=8866590, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465437, CurrentID=0, Status='live', Symbol='MIR6731', Aliases='hsa-mir-6731', description='microRNA 6731', other_designations='microRNA mir-6731', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24919345, end_position_on_the_genomic_accession=24919416, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465438, CurrentID=0, Status='live', Symbol='MIR6732', Aliases='hsa-mir-6732', description='microRNA 6732', other_designations='microRNA mir-6732', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37480230, end_position_on_the_genomic_accession=37480289, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465439, CurrentID=0, Status='live', Symbol='MIR6733', Aliases='hsa-mir-6733', description='microRNA 6733', other_designations='microRNA mir-6733', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43171652, end_position_on_the_genomic_accession=43171712, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465440, CurrentID=0, Status='live', Symbol='MIR6735', Aliases='hsa-mir-6735', description='microRNA 6735', other_designations='microRNA mir-6735', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43448539, end_position_on_the_genomic_accession=43448611, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465441, CurrentID=0, Status='live', Symbol='MIR6737', Aliases='hsa-mir-6737', description='microRNA 6737', other_designations='microRNA mir-6737', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153962351, end_position_on_the_genomic_accession=153962420, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465442, CurrentID=0, Status='live', Symbol='MIR6738', Aliases='hsa-mir-6738', description='microRNA 6738', other_designations='microRNA mir-6738', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155951273, end_position_on_the_genomic_accession=155951336, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465443, CurrentID=0, Status='live', Symbol='MIR6740', Aliases='hsa-mir-6740', description='microRNA 6740', other_designations='microRNA mir-6740', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202003124, end_position_on_the_genomic_accession=202003236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465444, CurrentID=0, Status='live', Symbol='MIR6742', Aliases='hsa-mir-6742', description='microRNA 6742', other_designations='microRNA mir-6742', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228397048, end_position_on_the_genomic_accession=228397109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465445, CurrentID=0, Status='live', Symbol='MIR6743', Aliases='hsa-mir-6743', description='microRNA 6743', other_designations='microRNA mir-6743', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=209336, end_position_on_the_genomic_accession=209406, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465446, CurrentID=0, Status='live', Symbol='MIR6746', Aliases='hsa-mir-6746', description='microRNA 6746', other_designations='microRNA mir-6746', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61878216, end_position_on_the_genomic_accession=61878278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465447, CurrentID=0, Status='live', Symbol='MIR6747', Aliases='hsa-mir-6747', description='microRNA 6747', other_designations='microRNA mir-6747', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62567011, end_position_on_the_genomic_accession=62567071, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465448, CurrentID=0, Status='live', Symbol='MIR6748', Aliases='hsa-mir-6748', description='microRNA 6748', other_designations='microRNA mir-6748', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62789815, end_position_on_the_genomic_accession=62789885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465449, CurrentID=0, Status='live', Symbol='MIR6751', Aliases='hsa-mir-6751', description='microRNA 6751', other_designations='microRNA mir-6751', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65129916, end_position_on_the_genomic_accession=65129978, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465450, CurrentID=0, Status='live', Symbol='MIR6752', Aliases='hsa-mir-6752', description='microRNA 6752', other_designations='microRNA mir-6752', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67490245, end_position_on_the_genomic_accession=67490315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465451, CurrentID=0, Status='live', Symbol='MIR6753', Aliases='hsa-mir-6753', description='microRNA 6753', other_designations='microRNA mir-6753', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68044794, end_position_on_the_genomic_accession=68044957, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465452, CurrentID=0, Status='live', Symbol='MIR6755', Aliases='hsa-mir-6755', description='microRNA 6755', other_designations='microRNA mir-6755', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86278333, end_position_on_the_genomic_accession=86278398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465453, CurrentID=0, Status='live', Symbol='MIR6756', Aliases='hsa-mir-6756', description='microRNA 6756', other_designations='microRNA mir-6756', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119312950, end_position_on_the_genomic_accession=119313012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465454, CurrentID=0, Status='live', Symbol='MIR6758', Aliases='hsa-mir-6758', description='microRNA 6758', other_designations='microRNA mir-6758', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57512688, end_position_on_the_genomic_accession=57512750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465455, CurrentID=0, Status='live', Symbol='MIR6760', Aliases='hsa-mir-6760', description='microRNA 6760', other_designations='microRNA mir-6760', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111304142, end_position_on_the_genomic_accession=111304209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465456, CurrentID=0, Status='live', Symbol='MIR6761', Aliases='hsa-mir-6761', description='microRNA 6761', other_designations='microRNA mir-6761', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111799834, end_position_on_the_genomic_accession=111799905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465457, CurrentID=0, Status='live', Symbol='MIR6762', Aliases='hsa-mir-6762', description='microRNA 6762', other_designations='microRNA mir-6762', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113291523, end_position_on_the_genomic_accession=113291608, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465458, CurrentID=0, Status='live', Symbol='MIR6765', Aliases='hsa-mir-6765', description='microRNA 6765', other_designations='microRNA mir-6765', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105150778, end_position_on_the_genomic_accession=105150864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465459, CurrentID=0, Status='live', Symbol='MIR6767', Aliases='hsa-mir-6767', description='microRNA 6767', other_designations='microRNA mir-6767', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2445392, end_position_on_the_genomic_accession=2445457, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465460, CurrentID=0, Status='live', Symbol='MIR6768', Aliases='hsa-mir-6768', description='microRNA 6768', other_designations='microRNA mir-6768', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2463967, end_position_on_the_genomic_accession=2464038, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465461, CurrentID=0, Status='live', Symbol='MIR6770-1', Aliases='hsa-mir-6770-1', description='microRNA 6770-1', other_designations='microRNA mir-6770-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14930820, end_position_on_the_genomic_accession=14930879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465462, CurrentID=0, Status='live', Symbol='MIR6771', Aliases='hsa-mir-6771', description='microRNA 6771', other_designations='microRNA mir-6771', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50292616, end_position_on_the_genomic_accession=50292675, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465463, CurrentID=0, Status='live', Symbol='MIR6772', Aliases='hsa-mir-6772', description='microRNA 6772', other_designations='microRNA mir-6772', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57772289, end_position_on_the_genomic_accession=57772352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465464, CurrentID=0, Status='live', Symbol='MIR6775', Aliases='hsa-mir-6775', description='microRNA 6775', other_designations='microRNA mir-6775', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87834592, end_position_on_the_genomic_accession=87834660, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465465, CurrentID=0, Status='live', Symbol='MIR6776', Aliases='hsa-mir-6776', description='microRNA 6776', other_designations='microRNA mir-6776', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2692861, end_position_on_the_genomic_accession=2692919, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465466, CurrentID=0, Status='live', Symbol='MIR6777', Aliases='hsa-mir-6777', description='microRNA 6777', other_designations='microRNA mir-6777', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17813480, end_position_on_the_genomic_accession=17813545, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465467, CurrentID=0, Status='live', Symbol='MIR6779', Aliases='hsa-mir-6779', description='microRNA 6779', other_designations='microRNA mir-6779', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38914979, end_position_on_the_genomic_accession=38915042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465468, CurrentID=0, Status='live', Symbol='MIR6781', Aliases='hsa-mir-6781', description='microRNA 6781', other_designations='microRNA mir-6781', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42823880, end_position_on_the_genomic_accession=42823943, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465469, CurrentID=0, Status='live', Symbol='MIR6782', Aliases='hsa-mir-6782', description='microRNA 6782', other_designations='microRNA mir-6782', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44207771, end_position_on_the_genomic_accession=44207839, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465470, CurrentID=0, Status='live', Symbol='MIR6784', Aliases='hsa-mir-6784', description='microRNA 6784', other_designations='microRNA mir-6784', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45114367, end_position_on_the_genomic_accession=45114433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465471, CurrentID=0, Status='live', Symbol='MIR6786', Aliases='hsa-mir-6786', description='microRNA 6786', other_designations='microRNA mir-6786', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81693757, end_position_on_the_genomic_accession=81693869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101154752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154752, CurrentID=0, Status='live', Symbol='PPKP1B', Aliases='', description='Keratoderma, palmoplantar, punctate type IB', other_designations='', map_location='8q24.13-q24.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614936, no_hdr0=''),
  101154753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101154753, CurrentID=0, Status='live', Symbol='PANDAR', Aliases='PANDA', description='promoter of CDKN1A antisense DNA damage activated RNA', other_designations='p21 associated ncRNA DNA damage activated', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36673621, end_position_on_the_genomic_accession=36675126, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465474, CurrentID=0, Status='live', Symbol='MIR6791', Aliases='hsa-mir-6791', description='microRNA 6791', other_designations='microRNA mir-6791', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6736712, end_position_on_the_genomic_accession=6736778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465475, CurrentID=0, Status='live', Symbol='MIR6792', Aliases='hsa-mir-6792', description='microRNA 6792', other_designations='microRNA mir-6792', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7617439, end_position_on_the_genomic_accession=7617505, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465476, CurrentID=0, Status='live', Symbol='MIR6795', Aliases='hsa-mir-6795', description='microRNA 6795', other_designations='microRNA mir-6795', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15179283, end_position_on_the_genomic_accession=15179350, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100499397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499397, CurrentID=0, Status='live', Symbol='DYZ1L1', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100499398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499398, CurrentID=0, Status='live', Symbol='DYZ1L4', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465479, CurrentID=0, Status='live', Symbol='MIR6799', Aliases='hsa-mir-6799', description='microRNA 6799', other_designations='microRNA mir-6799', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49791866, end_position_on_the_genomic_accession=49791934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465480, CurrentID=0, Status='live', Symbol='MIR6800', Aliases='hsa-mir-6800', description='microRNA 6800', other_designations='microRNA mir-6800', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49832018, end_position_on_the_genomic_accession=49832099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465481, CurrentID=0, Status='live', Symbol='MIR6802', Aliases='hsa-mir-6802', description='microRNA 6802', other_designations='microRNA mir-6802', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55239912, end_position_on_the_genomic_accession=55239976, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465482, CurrentID=0, Status='live', Symbol='MIR6804', Aliases='hsa-mir-6804', description='microRNA 6804', other_designations='microRNA mir-6804', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55230885, end_position_on_the_genomic_accession=55230952, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465483, CurrentID=0, Status='live', Symbol='MIR6805', Aliases='hsa-mir-6805', description='microRNA 6805', other_designations='microRNA mir-6805', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55388181, end_position_on_the_genomic_accession=55388242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465484, CurrentID=0, Status='live', Symbol='MIR6806', Aliases='hsa-mir-6806', description='microRNA 6806', other_designations='microRNA mir-6806', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58334688, end_position_on_the_genomic_accession=58334751, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465485, CurrentID=0, Status='live', Symbol='MIR6809', Aliases='hsa-mir-6809', description='microRNA 6809', other_designations='microRNA mir-6809', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217900513, end_position_on_the_genomic_accession=217900628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465486, CurrentID=0, Status='live', Symbol='MIR6811', Aliases='hsa-mir-6811', description='microRNA 6811', other_designations='microRNA mir-6811', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237510931, end_position_on_the_genomic_accession=237510988, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465487, CurrentID=0, Status='live', Symbol='MIR6812', Aliases='hsa-mir-6812', description='microRNA 6812', other_designations='microRNA mir-6812', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45425510, end_position_on_the_genomic_accession=45425573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465488, CurrentID=0, Status='live', Symbol='MIR6814', Aliases='hsa-mir-6814', description='microRNA 6814', other_designations='microRNA mir-6814', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41746772, end_position_on_the_genomic_accession=41746841, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100499409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499409, CurrentID=0, Status='live', Symbol='DYZ1L6', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465490, CurrentID=0, Status='live', Symbol='MIR6816', Aliases='hsa-mir-6816', description='microRNA 6816', other_designations='microRNA mir-6816', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20114686, end_position_on_the_genomic_accession=20114751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100499411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499411, CurrentID=0, Status='live', Symbol='DYZ1L7', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100499412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499412, CurrentID=0, Status='live', Symbol='DYZ1L16', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465493, CurrentID=0, Status='live', Symbol='MIR6821', Aliases='hsa-mir-6821', description='microRNA 6821', other_designations='microRNA mir-6821', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49962866, end_position_on_the_genomic_accession=49962939, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465494, CurrentID=0, Status='live', Symbol='MIR6823', Aliases='hsa-mir-6823', description='microRNA 6823', other_designations='microRNA mir-6823', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48549961, end_position_on_the_genomic_accession=48550021, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465495, CurrentID=0, Status='live', Symbol='MIR6824', Aliases='hsa-mir-6824', description='microRNA 6824', other_designations='microRNA mir-6824', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48633636, end_position_on_the_genomic_accession=48633698, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465496, CurrentID=0, Status='live', Symbol='MIR6826', Aliases='hsa-mir-6826', description='microRNA 6826', other_designations='microRNA mir-6826', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129272146, end_position_on_the_genomic_accession=129272243, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465497, CurrentID=0, Status='live', Symbol='MIR6828', Aliases='hsa-mir-6828', description='microRNA 6828', other_designations='microRNA mir-6828', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170423103, end_position_on_the_genomic_accession=170423162, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465498, CurrentID=0, Status='live', Symbol='MIR6830', Aliases='hsa-mir-6830', description='microRNA 6830', other_designations='microRNA mir-6830', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132217849, end_position_on_the_genomic_accession=132217918, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465499, CurrentID=0, Status='live', Symbol='MIR6831', Aliases='hsa-mir-6831', description='microRNA 6831', other_designations='microRNA mir-6831', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140563671, end_position_on_the_genomic_accession=140563751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465500, CurrentID=0, Status='live', Symbol='MIR6833', Aliases='hsa-mir-6833', description='microRNA 6833', other_designations='microRNA mir-6833', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32179816, end_position_on_the_genomic_accession=32179876, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465501, CurrentID=0, Status='live', Symbol='MIR6834', Aliases='hsa-mir-6834', description='microRNA 6834', other_designations='microRNA mir-6834', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33290245, end_position_on_the_genomic_accession=33290325, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465502, CurrentID=0, Status='live', Symbol='MIR6835', Aliases='hsa-mir-6835', description='microRNA 6835', other_designations='microRNA mir-6835', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34240673, end_position_on_the_genomic_accession=34240736, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465503, CurrentID=0, Status='live', Symbol='MIR6836', Aliases='hsa-mir-6836', description='microRNA 6836', other_designations='microRNA mir-6836', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2257515, end_position_on_the_genomic_accession=2257577, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465504, CurrentID=0, Status='live', Symbol='MIR6838', Aliases='hsa-mir-6838', description='microRNA 6838', other_designations='microRNA mir-6838', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44073378, end_position_on_the_genomic_accession=44073433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465505, CurrentID=0, Status='live', Symbol='MIR6839', Aliases='hsa-mir-6839', description='microRNA 6839', other_designations='microRNA mir-6839', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64679064, end_position_on_the_genomic_accession=64679176, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465506, CurrentID=0, Status='live', Symbol='MIR6841', Aliases='hsa-mir-6841', description='microRNA 6841', other_designations='microRNA mir-6841', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24953796, end_position_on_the_genomic_accession=24953867, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465507, CurrentID=0, Status='live', Symbol='MIR6842', Aliases='hsa-mir-6842', description='microRNA 6842', other_designations='microRNA mir-6842', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27433370, end_position_on_the_genomic_accession=27433434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465508, CurrentID=0, Status='live', Symbol='MIR6843', Aliases='hsa-mir-6843', description='microRNA 6843', other_designations='microRNA mir-6843', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27610601, end_position_on_the_genomic_accession=27610751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465509, CurrentID=0, Status='live', Symbol='MIR6846', Aliases='hsa-mir-6846', description='microRNA 6846', other_designations='microRNA mir-6846', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144057321, end_position_on_the_genomic_accession=144057380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465510, CurrentID=0, Status='live', Symbol='MIR6847', Aliases='hsa-mir-6847', description='microRNA 6847', other_designations='microRNA mir-6847', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144079874, end_position_on_the_genomic_accession=144079942, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465511, CurrentID=0, Status='live', Symbol='MIR6848', Aliases='hsa-mir-6848', description='microRNA 6848', other_designations='microRNA mir-6848', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144317246, end_position_on_the_genomic_accession=144317315, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465512, CurrentID=0, Status='live', Symbol='MIR6851', Aliases='hsa-mir-6851', description='microRNA 6851', other_designations='microRNA mir-6851', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33467869, end_position_on_the_genomic_accession=33467935, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465513, CurrentID=0, Status='live', Symbol='MIR6852', Aliases='hsa-mir-6852', description='microRNA 6852', other_designations='microRNA mir-6852', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35710676, end_position_on_the_genomic_accession=35710741, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465514, CurrentID=0, Status='live', Symbol='MIR6854', Aliases='hsa-mir-6854', description='microRNA 6854', other_designations='microRNA mir-6854', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98229149, end_position_on_the_genomic_accession=98229217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465515, CurrentID=0, Status='live', Symbol='MIR6856', Aliases='hsa-mir-6856', description='microRNA 6856', other_designations='microRNA mir-6856', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130626297, end_position_on_the_genomic_accession=130626363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465516, CurrentID=0, Status='live', Symbol='MIR6857', Aliases='hsa-mir-6857', description='microRNA 6857', other_designations='microRNA mir-6857', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53405673, end_position_on_the_genomic_accession=53405765, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465517, CurrentID=0, Status='live', Symbol='MIR6858', Aliases='hsa-mir-6858', description='microRNA 6858', other_designations='microRNA mir-6858', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154450320, end_position_on_the_genomic_accession=154450386, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100499438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499438, CurrentID=0, Status='live', Symbol='DYZ1L20', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465519, CurrentID=0, Status='live', Symbol='MIR6861', Aliases='hsa-mir-6861', description='microRNA 6861', other_designations='microRNA mir-6861', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112163258, end_position_on_the_genomic_accession=112163321, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465520, CurrentID=0, Status='live', Symbol='MIR6862-1', Aliases='hsa-mir-6862-1', description='microRNA 6862-1', other_designations='microRNA mir-6862-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28390982, end_position_on_the_genomic_accession=28391051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465521, CurrentID=0, Status='live', Symbol='MIR6864', Aliases='hsa-mir-6864', description='microRNA 6864', other_designations='microRNA mir-6864', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4969702, end_position_on_the_genomic_accession=4969771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100499442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499442, CurrentID=0, Status='live', Symbol='DYZ1L23', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100499443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499443, CurrentID=0, Status='live', Symbol='DYZ1L27', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465524, CurrentID=0, Status='live', Symbol='MIR6869', Aliases='hsa-mir-6869', description='microRNA 6869', other_designations='microRNA mir-6869', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1392900, end_position_on_the_genomic_accession=1392961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465525, CurrentID=0, Status='live', Symbol='MIR6870', Aliases='hsa-mir-6870', description='microRNA 6870', other_designations='microRNA mir-6870', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10649636, end_position_on_the_genomic_accession=10649695, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465526, CurrentID=0, Status='live', Symbol='MIR6872', Aliases='hsa-mir-6872', description='microRNA 6872', other_designations='microRNA mir-6872', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50273236, end_position_on_the_genomic_accession=50273297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387753, CurrentID=0, Status='live', Symbol='RPL21P97', Aliases='RPL21_40_1120', description='ribosomal protein L21 pseudogene 97', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9660395, end_position_on_the_genomic_accession=9660956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102465528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465528, CurrentID=0, Status='live', Symbol='MIR6877', Aliases='hsa-mir-6877', description='microRNA 6877', other_designations='microRNA mir-6877', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133051996, end_position_on_the_genomic_accession=133052059, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465529, CurrentID=0, Status='live', Symbol='MIR6878', Aliases='hsa-mir-6878', description='microRNA 6878', other_designations='microRNA mir-6878', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150492345, end_position_on_the_genomic_accession=150492410, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465530, CurrentID=0, Status='live', Symbol='MIR6881', Aliases='hsa-mir-6881', description='microRNA 6881', other_designations='microRNA mir-6881', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74411357, end_position_on_the_genomic_accession=74411432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465531, CurrentID=0, Status='live', Symbol='MIR6882', Aliases='hsa-mir-6882', description='microRNA 6882', other_designations='microRNA mir-6882', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74840642, end_position_on_the_genomic_accession=74840707, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465532, CurrentID=0, Status='live', Symbol='MIR6883', Aliases='hsa-mir-6883', description='microRNA 6883', other_designations='microRNA mir-6883', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8144994, end_position_on_the_genomic_accession=8145071, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465533, CurrentID=0, Status='live', Symbol='MIR6885', Aliases='hsa-mir-6885', description='microRNA 6885', other_designations='microRNA mir-6885', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6389638, end_position_on_the_genomic_accession=6389703, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465534, CurrentID=0, Status='live', Symbol='MIR6886', Aliases='hsa-mir-6886', description='microRNA 6886', other_designations='microRNA mir-6886', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11113474, end_position_on_the_genomic_accession=11113534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465535, CurrentID=0, Status='live', Symbol='MIR6888', Aliases='hsa-mir-6888', description='microRNA 6888', other_designations='microRNA mir-6888', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159186835, end_position_on_the_genomic_accession=159186901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465536, CurrentID=0, Status='live', Symbol='MIR6890', Aliases='hsa-mir-6890', description='microRNA 6890', other_designations='microRNA mir-6890', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49099854, end_position_on_the_genomic_accession=49099914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465537, CurrentID=0, Status='live', Symbol='MIR6891', Aliases='hsa-mir-6891', description='microRNA 6891', other_designations='microRNA mir-6891', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31355224, end_position_on_the_genomic_accession=31355316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465538, CurrentID=0, Status='live', Symbol='MIR6892', Aliases='hsa-mir-6892', description='microRNA 6892', other_designations='microRNA mir-6892', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143382686, end_position_on_the_genomic_accession=143382800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465539, CurrentID=0, Status='live', Symbol='MIR6895', Aliases='hsa-mir-6895', description='microRNA 6895', other_designations='microRNA mir-6895', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53195411, end_position_on_the_genomic_accession=53195488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391713, CurrentID=0, Status='live', Symbol='LOC391713', Aliases='', description='tripartite motif-containing 75 pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165010459, end_position_on_the_genomic_accession=165011836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499466, CurrentID=0, Status='live', Symbol='LINC00674', Aliases='', description='long intergenic non-protein coding RNA 674', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68101555, end_position_on_the_genomic_accession=68135929, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100499467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499467, CurrentID=0, Status='live', Symbol='LINC00673', Aliases='HI-LNC75', description='long intergenic non-protein coding RNA 673', other_designations='Human Islet Long Non Coding RNA 75', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72403322, end_position_on_the_genomic_accession=72592804, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100499468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499468, CurrentID=0, Status='live', Symbol='MTND1P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 1', other_designations='MT-ND1 pseudogene 1|MTND1 pseudogene 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8371663, end_position_on_the_genomic_accession=8372030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499469, CurrentID=0, Status='live', Symbol='MTND6P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 1', other_designations='MT-ND6 pseudogene 1|MTND6 pseudogene 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8363536, end_position_on_the_genomic_accession=8363959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499471, CurrentID=0, Status='live', Symbol='LOC100499471', Aliases='', description='leucine rich repeat containing 14 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102259043, end_position_on_the_genomic_accession=102261692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  192343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=192343, CurrentID=0, Status='live', Symbol='NEWENTRY', Aliases='', description='Record to support submission of GeneRIFs for a gene not in Gene (human; man).', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376430, CurrentID=0, Status='live', Symbol='LOC105376430', Aliases='', description='uncharacterized LOC105376430', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14986795, end_position_on_the_genomic_accession=14988047, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129734, CurrentID=0, Status='live', Symbol='LOC100129734', Aliases='', description='zinc finger, AN1-type domain 1 pseudogene', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176272464, end_position_on_the_genomic_accession=176274431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499481, CurrentID=0, Status='live', Symbol='MED13P1', Aliases='', description='mediator complex subunit 13 pseudogene 1', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14907896, end_position_on_the_genomic_accession=14908068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499483, CurrentID=0, Status='live', Symbol='CCDC180', Aliases='BDAG1, C9orf174, KIAA1529', description='coiled-coil domain containing 180', other_designations="Behcet's Disease Associated Gene 1", map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97307628, end_position_on_the_genomic_accession=97377295, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  100499484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499484, CurrentID=0, Status='live', Symbol='LOC100499484', Aliases='', description='SUGT1-1300002K09Rik pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97238426, end_position_on_the_genomic_accession=97297314, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100499485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499485, CurrentID=0, Status='live', Symbol='SRIP2', Aliases='', description='sorcin pseudogene 2', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=89367466, end_position_on_the_genomic_accession=89367697, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499489, CurrentID=0, Status='live', Symbol='LOC100499489', Aliases='', description='uncharacterized LOC100499489', other_designations='', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22435425, end_position_on_the_genomic_accession=22437929, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376433, CurrentID=0, Status='live', Symbol='LOC105376433', Aliases='', description='uncharacterized LOC105376433', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15207898, end_position_on_the_genomic_accession=15237848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100499496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499496, CurrentID=0, Status='live', Symbol='LOC100499496', Aliases='', description='matrin 3 pseudogene', other_designations='', map_location='Xq13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72660457, end_position_on_the_genomic_accession=72661737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100499497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499497, CurrentID=0, Status='live', Symbol='LOC100499497', Aliases='', description='matrin 3 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105889537, end_position_on_the_genomic_accession=105891470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419862, CurrentID=0, Status='live', Symbol='LOC100419862', Aliases='', description='dynein, axonemal, assembly factor 5 pseudodgene', other_designations='HEAT repeat containing 2 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67718999, end_position_on_the_genomic_accession=67721241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387763, CurrentID=0, Status='live', Symbol='C11orf96', Aliases='AG2', description='chromosome 11 open reading frame 96', other_designations='protein Ag2 homolog', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43942556, end_position_on_the_genomic_accession=43943883, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376436, CurrentID=0, Status='live', Symbol='LOC105376436', Aliases='', description='uncharacterized LOC105376436', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17437290, end_position_on_the_genomic_accession=17453546, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422158, CurrentID=0, Status='live', Symbol='LOC100422158', Aliases='', description='olfactory receptor, family 52, subfamily B, member 4 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376437, CurrentID=0, Status='live', Symbol='LOC105376437', Aliases='', description='uncharacterized LOC105376437', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17471639, end_position_on_the_genomic_accession=17472987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372583, CurrentID=0, Status='live', Symbol='LOC105372583', Aliases='', description='uncharacterized LOC105372583', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25738901, end_position_on_the_genomic_accession=25750488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100499532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499532, CurrentID=0, Status='live', Symbol='ATPLS', Aliases='', description='antiphospholipid syndrome, familial', other_designations='lupus anticoagulant, familial', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=107320, no_hdr0=''),
  105376440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376440, CurrentID=0, Status='live', Symbol='LOC105376440', Aliases='', description='uncharacterized LOC105376440', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18710294, end_position_on_the_genomic_accession=18836471, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371525, CurrentID=0, Status='live', Symbol='LOC105371525', Aliases='', description='uncharacterized LOC105371525', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8670612, end_position_on_the_genomic_accession=8702516, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387770, CurrentID=0, Status='live', Symbol='LOC387770', Aliases='', description='tripartite motif containing 49D1 pseudogene', other_designations='ring finger protein 18-like pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49843962, end_position_on_the_genomic_accession=49861219, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100499563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100499563, CurrentID=0, Status='live', Symbol='PTP4A1P2', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 2', other_designations='', map_location='12q24', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133032863, end_position_on_the_genomic_accession=133034170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371526, CurrentID=0, Status='live', Symbol='LOC105371526', Aliases='', description='uncharacterized LOC105371526', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8714568, end_position_on_the_genomic_accession=8717443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100526783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526783, CurrentID=0, Status='live', Symbol='C15orf38-AP3S2', Aliases='', description='C15orf38-AP3S2 readthrough', other_designations='', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89830599, end_position_on_the_genomic_accession=89912990, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  102465665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465665, CurrentID=0, Status='live', Symbol='MIR7107', Aliases='hsa-mir-7107', description='microRNA 7107', other_designations='microRNA mir-7107', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121444273, end_position_on_the_genomic_accession=121444352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465666, CurrentID=0, Status='live', Symbol='MIR7109', Aliases='hsa-mir-7109', description='microRNA 7109', other_designations='microRNA mir-7109', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31621467, end_position_on_the_genomic_accession=31621531, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465667, CurrentID=0, Status='live', Symbol='MIR7110', Aliases='hsa-mir-7110', description='microRNA 7110', other_designations='microRNA mir-7110', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123161794, end_position_on_the_genomic_accession=123161879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465668, CurrentID=0, Status='live', Symbol='MIR7111', Aliases='hsa-mir-7111', description='microRNA 7111', other_designations='microRNA mir-7111', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35470508, end_position_on_the_genomic_accession=35470579, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465669, CurrentID=0, Status='live', Symbol='MIR7113', Aliases='hsa-mir-7113', description='microRNA 7113', other_designations='microRNA mir-7113', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68032864, end_position_on_the_genomic_accession=68032922, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376450, CurrentID=0, Status='live', Symbol='LOC105376450', Aliases='', description='uncharacterized LOC105376450', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22471451, end_position_on_the_genomic_accession=22472577, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465683, CurrentID=0, Status='live', Symbol='MIR6511A3', Aliases='MIR6511A-3, hsa-mir-6511a-3', description='microRNA 6511a-3', other_designations='microRNA mir-6511a-3', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16368876, end_position_on_the_genomic_accession=16368942, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465684, CurrentID=0, Status='live', Symbol='MIR6511A4', Aliases='MIR6511A-4, hsa-mir-6511a-4', description='microRNA 6511a-4', other_designations='microRNA mir-6511a-4', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18344013, end_position_on_the_genomic_accession=18344079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465688, CurrentID=0, Status='live', Symbol='MIR7150', Aliases='hsa-mir-7150', description='microRNA 7150', other_designations='microRNA mir-7150', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123485529, end_position_on_the_genomic_accession=123485622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465689, CurrentID=0, Status='live', Symbol='MIR7152', Aliases='hsa-mir-7152', description='microRNA 7152', other_designations='microRNA mir-7152', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71790747, end_position_on_the_genomic_accession=71790800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465690, CurrentID=0, Status='live', Symbol='MIR7153', Aliases='hsa-mir-7153, mir-7153', description='microRNA 7153', other_designations='microRNA mir-7153', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11654885, end_position_on_the_genomic_accession=11654941, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465691, CurrentID=0, Status='live', Symbol='MIR7154', Aliases='hsa-mir-7154', description='microRNA 7154', other_designations='microRNA mir-7154', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45691704, end_position_on_the_genomic_accession=45691776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465692, CurrentID=0, Status='live', Symbol='MIR7157', Aliases='hsa-mir-7157', description='microRNA 7157', other_designations='microRNA mir-7157', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140586626, end_position_on_the_genomic_accession=140586685, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465693, CurrentID=0, Status='live', Symbol='MIR7158', Aliases='hsa-mir-7158', description='microRNA 7158', other_designations='microRNA mir-7158', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5974662, end_position_on_the_genomic_accession=5974732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465694, CurrentID=0, Status='live', Symbol='MIR7161', Aliases='hsa-mir-7161', description='microRNA 7161', other_designations='microRNA mir-7161', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158609707, end_position_on_the_genomic_accession=158609790, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465695, CurrentID=0, Status='live', Symbol='MIR7160', Aliases='hsa-mir-7160', description='microRNA 7160', other_designations='microRNA mir-7160', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2076589, end_position_on_the_genomic_accession=2076640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465696, CurrentID=0, Status='live', Symbol='MIR486-2', Aliases='hsa-mir-486-2, mir-486-2', description='microRNA 486-2', other_designations='microRNA mir-486-2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41660444, end_position_on_the_genomic_accession=41660507, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375892, CurrentID=0, Status='live', Symbol='LOC105375892', Aliases='', description='uncharacterized LOC105375892', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71549720, end_position_on_the_genomic_accession=71554714, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131536, CurrentID=0, Status='live', Symbol='LOC100131536', Aliases='', description='tRNA methyltransferase 11 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35501629, end_position_on_the_genomic_accession=35503667, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100526790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526790, CurrentID=0, Status='live', Symbol='TST1', Aliases='', description='Tuberculin skin test reactivity, absence of', other_designations='', map_location='11p14', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613636, no_hdr0=''),
  105376455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376455, CurrentID=0, Status='live', Symbol='LOC105376455', Aliases='', description='uncharacterized LOC105376455', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24457008, end_position_on_the_genomic_accession=24466437, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376457, CurrentID=0, Status='live', Symbol='LOC105376457', Aliases='', description='uncharacterized LOC105376457', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25813120, end_position_on_the_genomic_accession=25882732, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100526794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526794, CurrentID=0, Status='live', Symbol='NT5C1B-RDH14', Aliases='', description='NT5C1B-RDH14 readthrough', other_designations='NT5C1B-RDH14 read-through transcript', map_location='2p', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18554723, end_position_on_the_genomic_accession=18589580, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105376459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376459, CurrentID=0, Status='live', Symbol='LOC105376459', Aliases='', description='uncharacterized LOC105376459', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26147588, end_position_on_the_genomic_accession=26168658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104169670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104169670, CurrentID=0, Status='live', Symbol='TMEM5-AS1', Aliases='', description='TMEM5 antisense RNA 1', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63808845, end_position_on_the_genomic_accession=63822156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104169671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104169671, CurrentID=0, Status='live', Symbol='LINC01036', Aliases='', description='long intergenic non-protein coding RNA 1036', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187290056, end_position_on_the_genomic_accession=187360244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376460, CurrentID=0, Status='live', Symbol='LOC105376460', Aliases='', description='uncharacterized LOC105376460', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25905897, end_position_on_the_genomic_accession=25912012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465752, CurrentID=0, Status='live', Symbol='MIR7641-1', Aliases='hsa-mir-7641-1', description='microRNA 7641-1', other_designations='microRNA mir-7641-1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104252591, end_position_on_the_genomic_accession=104252651, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465753, CurrentID=0, Status='live', Symbol='MIR7641-2', Aliases='B-ATF, hsa-mir-7641-2', description='microRNA 7641-2', other_designations='RPS24a-like|microRNA mir-7641-2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75604209, end_position_on_the_genomic_accession=75604261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376463, CurrentID=0, Status='live', Symbol='LOC105376463', Aliases='', description='uncharacterized LOC105376463', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27566805, end_position_on_the_genomic_accession=27584417, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376467, CurrentID=0, Status='live', Symbol='LOC105376467', Aliases='', description='uncharacterized LOC105376467', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28326012, end_position_on_the_genomic_accession=28335354, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102465800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465800, CurrentID=0, Status='live', Symbol='MIR7702', Aliases='hsa-mir-7702', description='microRNA 7702', other_designations='microRNA mir-7702', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111271156, end_position_on_the_genomic_accession=111271214, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465801, CurrentID=0, Status='live', Symbol='MIR7703', Aliases='hsa-mir-7703', description='microRNA 7703', other_designations='microRNA mir-7703', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24143489, end_position_on_the_genomic_accession=24143565, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465802, CurrentID=0, Status='live', Symbol='MIR7704', Aliases='hsa-mir-7704', description='microRNA 7704', other_designations='microRNA mir-7704', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176188843, end_position_on_the_genomic_accession=176188901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465803, CurrentID=0, Status='live', Symbol='MIR7706', Aliases='hsa-mir-7706', description='microRNA 7706', other_designations='microRNA mir-7706', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85380596, end_position_on_the_genomic_accession=85380662, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371531, CurrentID=0, Status='live', Symbol='LOC105371531', Aliases='', description='uncharacterized LOC105371531', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9243887, end_position_on_the_genomic_accession=9250114, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376474, CurrentID=0, Status='live', Symbol='LOC105376474', Aliases='', description='uncharacterized LOC105376474', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29405881, end_position_on_the_genomic_accession=29408473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102465832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465832, CurrentID=0, Status='live', Symbol='MIR7843', Aliases='hsa-mir-7843', description='microRNA 7843', other_designations='microRNA mir-7843', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72516820, end_position_on_the_genomic_accession=72516898, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465833, CurrentID=0, Status='live', Symbol='MIR4433B', Aliases='hsa-mir-4433b', description='microRNA 4433b', other_designations='microRNA mir-4433b', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64340747, end_position_on_the_genomic_accession=64340848, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465834, CurrentID=0, Status='live', Symbol='MIR7844', Aliases='hsa-mir-7844', description='microRNA 7844', other_designations='microRNA mir-7844', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94571231, end_position_on_the_genomic_accession=94571352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465835, CurrentID=0, Status='live', Symbol='MIR7845', Aliases='hsa-mir-7845', description='microRNA 7845', other_designations='microRNA mir-7845', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207166400, end_position_on_the_genomic_accession=207166498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465836, CurrentID=0, Status='live', Symbol='MIR7846', Aliases='hsa-mir-7846', description='microRNA 7846', other_designations='microRNA mir-7846', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12166943, end_position_on_the_genomic_accession=12167038, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465837, CurrentID=0, Status='live', Symbol='MIR7849', Aliases='hsa-mir-7849', description='microRNA 7849', other_designations='microRNA mir-7849', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146408583, end_position_on_the_genomic_accession=146408688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465838, CurrentID=0, Status='live', Symbol='MIR7850', Aliases='hsa-mir-7850', description='microRNA 7850', other_designations='microRNA mir-7850', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2630715, end_position_on_the_genomic_accession=2630793, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465839, CurrentID=0, Status='live', Symbol='MIR7852', Aliases='hsa-mir-7852', description='microRNA 7852', other_designations='microRNA mir-7852', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=107897223, end_position_on_the_genomic_accession=107897304, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465840, CurrentID=0, Status='live', Symbol='MIR7854', Aliases='hsa-mir-7854', description='microRNA 7854', other_designations='microRNA mir-7854', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81533902, end_position_on_the_genomic_accession=81533966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465841, CurrentID=0, Status='live', Symbol='MIR7855', Aliases='hsa-mir-7855', description='microRNA 7855', other_designations='microRNA mir-7855', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64785626, end_position_on_the_genomic_accession=64785686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465842, CurrentID=0, Status='live', Symbol='MIR7856', Aliases='hsa-mir-7856', description='microRNA 7856', other_designations='microRNA mir-7856', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86357632, end_position_on_the_genomic_accession=86357687, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465854, CurrentID=0, Status='live', Symbol='MIR548BA', Aliases='hsa-mir-548ba', description='microRNA 548ba', other_designations='microRNA mir-548ba', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=49059603, end_position_on_the_genomic_accession=49059658, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465855, CurrentID=0, Status='live', Symbol='MIR7973-2', Aliases='hsa-mir-7973-2', description='microRNA 7973-2', other_designations='microRNA mir-7973-2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51314032, end_position_on_the_genomic_accession=51314107, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465856, CurrentID=0, Status='live', Symbol='MIR7974', Aliases='hsa-mir-7974', description='microRNA 7974', other_designations='microRNA mir-7974', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11495544, end_position_on_the_genomic_accession=11495622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465857, CurrentID=0, Status='live', Symbol='MIR7976', Aliases='hsa-mir-7976', description='microRNA 7976', other_designations='microRNA mir-7976', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127587111, end_position_on_the_genomic_accession=127587176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465858, CurrentID=0, Status='live', Symbol='MIR7977', Aliases='hsa-mir-7977', description='microRNA 7977', other_designations='microRNA mir-7977', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176515103, end_position_on_the_genomic_accession=176515151, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465859, CurrentID=0, Status='live', Symbol='MIR7978', Aliases='hsa-mir-7978', description='microRNA 7978', other_designations='microRNA mir-7978', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21464700, end_position_on_the_genomic_accession=21464758, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465860, CurrentID=0, Status='live', Symbol='MIR8054', Aliases='hsa-mir-8054', description='microRNA 8054', other_designations='microRNA mir-8054', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23419105, end_position_on_the_genomic_accession=23419190, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465861, CurrentID=0, Status='live', Symbol='MIR8055', Aliases='hsa-mir-8055', description='microRNA 8055', other_designations='microRNA mir-8055', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6622124, end_position_on_the_genomic_accession=6622220, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465862, CurrentID=0, Status='live', Symbol='MIR8056', Aliases='hsa-mir-8056', description='microRNA 8056', other_designations='microRNA mir-8056', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173347455, end_position_on_the_genomic_accession=173347536, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465863, CurrentID=0, Status='live', Symbol='MIR8058', Aliases='hsa-mir-8058', description='microRNA 8058', other_designations='microRNA mir-8058', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82688931, end_position_on_the_genomic_accession=82689019, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465864, CurrentID=0, Status='live', Symbol='MIR8060', Aliases='hsa-mir-8060', description='microRNA 8060', other_designations='microRNA mir-8060', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96359964, end_position_on_the_genomic_accession=96360039, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465865, CurrentID=0, Status='live', Symbol='MIR8062', Aliases='hsa-mir-8062', description='microRNA 8062', other_designations='microRNA mir-8062', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7371608, end_position_on_the_genomic_accession=7371692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465866, CurrentID=0, Status='live', Symbol='MIR8064', Aliases='hsa-mir-8064', description='microRNA 8064', other_designations='microRNA mir-8064', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52846463, end_position_on_the_genomic_accession=52846552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465867, CurrentID=0, Status='live', Symbol='MIR8065', Aliases='hsa-mir-8065', description='microRNA 8065', other_designations='microRNA mir-8065', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5632467, end_position_on_the_genomic_accession=5632566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465868, CurrentID=0, Status='live', Symbol='MIR8066', Aliases='hsa-mir-8066', description='microRNA 8066', other_designations='microRNA mir-8066', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101240795, end_position_on_the_genomic_accession=101240872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465869, CurrentID=0, Status='live', Symbol='MIR8067', Aliases='hsa-mir-8067', description='microRNA 8067', other_designations='microRNA mir-8067', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62304658, end_position_on_the_genomic_accession=62304734, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465870, CurrentID=0, Status='live', Symbol='MIR8070', Aliases='hsa-mir-8070', description='microRNA 8070', other_designations='microRNA mir-8070', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11783135, end_position_on_the_genomic_accession=11783222, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465871, CurrentID=0, Status='live', Symbol='MIR8071-1', Aliases='hsa-mir-8071-1', description='microRNA 8071-1', other_designations='microRNA mir-8071-1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105621116, end_position_on_the_genomic_accession=105621180, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465872, CurrentID=0, Status='live', Symbol='MIR8073', Aliases='hsa-mir-8073', description='microRNA 8073', other_designations='microRNA mir-8073', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110340958, end_position_on_the_genomic_accession=110341029, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465873, CurrentID=0, Status='live', Symbol='MIR8074', Aliases='hsa-mir-8074', description='microRNA 8074', other_designations='microRNA mir-8074', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51206929, end_position_on_the_genomic_accession=51207009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465874, CurrentID=0, Status='live', Symbol='MIR8075', Aliases='hsa-mir-8075', description='microRNA 8075', other_designations='microRNA mir-8075', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113262920, end_position_on_the_genomic_accession=113262999, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465875, CurrentID=0, Status='live', Symbol='MIR8077', Aliases='hsa-mir-8077', description='microRNA 8077', other_designations='microRNA mir-8077', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42351131, end_position_on_the_genomic_accession=42351205, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465876, CurrentID=0, Status='live', Symbol='MIR8079', Aliases='hsa-mir-8079', description='microRNA 8079', other_designations='microRNA mir-8079', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44196129, end_position_on_the_genomic_accession=44196200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465877, CurrentID=0, Status='live', Symbol='MIR8080', Aliases='hsa-mir-8080', description='microRNA 8080', other_designations='microRNA mir-8080', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79866495, end_position_on_the_genomic_accession=79866583, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465878, CurrentID=0, Status='live', Symbol='MIR8082', Aliases='hsa-mir-8082', description='microRNA 8082', other_designations='microRNA mir-8082', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113152282, end_position_on_the_genomic_accession=113152362, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465879, CurrentID=0, Status='live', Symbol='MIR8085', Aliases='hsa-mir-8085', description='microRNA 8085', other_designations='microRNA mir-8085', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44758657, end_position_on_the_genomic_accession=44758721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465880, CurrentID=0, Status='live', Symbol='MIR8086', Aliases='hsa-mir-8086', description='microRNA 8086', other_designations='microRNA mir-8086', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28289258, end_position_on_the_genomic_accession=28289350, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465881, CurrentID=0, Status='live', Symbol='MIR8087', Aliases='hsa-mir-8087', description='microRNA 8087', other_designations='microRNA mir-8087', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27514970, end_position_on_the_genomic_accession=27515047, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465882, CurrentID=0, Status='live', Symbol='MIR8089', Aliases='hsa-mir-8089', description='microRNA 8089', other_designations='microRNA mir-8089', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181043403, end_position_on_the_genomic_accession=181043484, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422168, CurrentID=0, Status='live', Symbol='LOC100422168', Aliases='', description='olfactory receptor, family 4, subfamily C, member 46 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387815, CurrentID=0, Status='live', Symbol='OR8G3P', Aliases='', description='olfactory receptor family 8 subfamily G member 3 pseudogene', other_designations='olfactory receptor, family 8, subfamily G, member 3 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124215220, end_position_on_the_genomic_accession=124215784, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102465906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465906, CurrentID=0, Status='live', Symbol='MIR7112', Aliases='MIR7112-1-2, hsa-mir-7112, hsa-mir-7112-1, hsa-mir-7112-2, MIR7112', description='microRNA 7112', other_designations='microRNA 7112-1|microRNA 7112-2|microRNA mir-7112-1|microRNA mir-7112-2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144262673, end_position_on_the_genomic_accession=144262737, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465907, CurrentID=0, Status='live', Symbol='MIR6862-2', Aliases='hsa-mir-6862-2', description='microRNA 6862-2', other_designations='microRNA mir-6862-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28724252, end_position_on_the_genomic_accession=28724321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465908, CurrentID=0, Status='live', Symbol='MIR6770-2', Aliases='hsa-mir-6770-2', description='microRNA 6770-2', other_designations='microRNA mir-6770-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16329305, end_position_on_the_genomic_accession=16329364, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465909, CurrentID=0, Status='live', Symbol='MIR6859-2', Aliases='hsa-mir-6859-2', description='microRNA 6859-2', other_designations='microRNA mir-6859-2', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187891, end_position_on_the_genomic_accession=187958, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465910, CurrentID=0, Status='live', Symbol='MIR6859-3', Aliases='hsa-mir-6859-3', description='microRNA 6859-3', other_designations='microRNA mir-6859-3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101973524, end_position_on_the_genomic_accession=101973591, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376489, CurrentID=0, Status='live', Symbol='LOC105376489', Aliases='', description='uncharacterized LOC105376489', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33530768, end_position_on_the_genomic_accession=33537149, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100630918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100630918, CurrentID=0, Status='live', Symbol='PARTICL', Aliases='PARTICLE', description='promoter of MAT2A antisense radiation-induced circulating long non-coding RNA', other_designations='promoter of MAT2A-antisense radiation-induced circulating lncRNA', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85537467, end_position_on_the_genomic_accession=85538886, orientation='minus', exon_count=1, OMIM=616350, no_hdr0=''),
  100630919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100630919, CurrentID=0, Status='live', Symbol='LOC100630919', Aliases='', description='thymosin beta 4, X-linked pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33523096, end_position_on_the_genomic_accession=33523673, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100630920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100630920, CurrentID=0, Status='live', Symbol='LOC100630920', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33540194, end_position_on_the_genomic_accession=33540879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387820, CurrentID=0, Status='live', Symbol='LOC387820', Aliases='', description='DnaJ heat shock protein family (Hsp40) member B7 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7 pseudogene', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=127940984, end_position_on_the_genomic_accession=127941655, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100630923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100630923, CurrentID=0, Status='live', Symbol='LOC100630923', Aliases='', description='LOC100289561-PRKRIP1 readthrough', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102363861, end_position_on_the_genomic_accession=102426682, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376494, CurrentID=0, Status='live', Symbol='LOC105376494', Aliases='', description='uncharacterized LOC105376494', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35643802, end_position_on_the_genomic_accession=35647824, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102465944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465944, CurrentID=0, Status='live', Symbol='MIR6072', Aliases='hsa-mir-6072', description='microRNA 6072', other_designations='microRNA mir-6072', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2076019, end_position_on_the_genomic_accession=2076089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376495, CurrentID=0, Status='live', Symbol='LOC105376495', Aliases='', description='uncharacterized LOC105376495', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35897206, end_position_on_the_genomic_accession=35898849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133616, CurrentID=0, Status='live', Symbol='LOC100133616', Aliases='', description='uncharacterized LOC100133616', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422170, CurrentID=0, Status='live', Symbol='LOC100422170', Aliases='', description='olfactory receptor, family 4, subfamily A, member 16 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928331, CurrentID=0, Status='live', Symbol='LOC101928331', Aliases='', description='uncharacterized LOC101928331', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14431940, end_position_on_the_genomic_accession=14516086, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376497, CurrentID=0, Status='live', Symbol='LOC105376497', Aliases='', description='uncharacterized LOC105376497', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37437593, end_position_on_the_genomic_accession=37438734, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376499, CurrentID=0, Status='live', Symbol='LOC105376499', Aliases='', description='uncharacterized LOC105376499', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37992380, end_position_on_the_genomic_accession=38004541, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102465974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465974, CurrentID=0, Status='live', Symbol='MIR6719', Aliases='hsa-mir-6719', description='microRNA 6719', other_designations='microRNA mir-6719', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39829716, end_position_on_the_genomic_accession=39829802, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465975, CurrentID=0, Status='live', Symbol='MIR6763', Aliases='hsa-mir-6763', description='microRNA 6763', other_designations='microRNA mir-6763', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132581997, end_position_on_the_genomic_accession=132582061, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465976, CurrentID=0, Status='live', Symbol='MIR6807', Aliases='hsa-mir-6807', description='microRNA 6807', other_designations='microRNA mir-6807', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58550285, end_position_on_the_genomic_accession=58550376, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465977, CurrentID=0, Status='live', Symbol='MIR6829', Aliases='hsa-mir-6829', description='microRNA 6829', other_designations='microRNA mir-6829', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195882329, end_position_on_the_genomic_accession=195882395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102465978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465978, CurrentID=0, Status='live', Symbol='MIR6850', Aliases='hsa-mir-6850', description='microRNA 6850', other_designations='microRNA mir-6850', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144791931, end_position_on_the_genomic_accession=144791991, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376501, CurrentID=0, Status='live', Symbol='LOC105376501', Aliases='', description='uncharacterized LOC105376501', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3532040, end_position_on_the_genomic_accession=3556794, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422171, CurrentID=0, Status='live', Symbol='LOC100422171', Aliases='', description='olfactory receptor, family 4, subfamily A, member 15 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102465985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465985, CurrentID=0, Status='live', Symbol='MIR6511B2', Aliases='MIR6511B-2, hsa-mir-6511b-2', description='microRNA 6511b-2', other_designations='microRNA mir-6511b-2|mir-6511b-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15134075, end_position_on_the_genomic_accession=15134145, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100526838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526838, CurrentID=0, Status='live', Symbol='LOC100526838', Aliases='', description='chromosome 18 open reading frame 32 pseudogene', other_designations='', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48470274, end_position_on_the_genomic_accession=48471522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102465993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465993, CurrentID=0, Status='live', Symbol='MIR7847', Aliases='hsa-mir-7847', description='microRNA 7847', other_designations='microRNA mir-7847', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1880045, end_position_on_the_genomic_accession=1880147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102465995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465995, CurrentID=0, Status='live', Symbol='MIR8081', Aliases='hsa-mir-8081', description='microRNA 8081', other_designations='microRNA mir-8081', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106600928, end_position_on_the_genomic_accession=106601022, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375902, CurrentID=0, Status='live', Symbol='LOC105375902', Aliases='', description='uncharacterized LOC105375902', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73653660, end_position_on_the_genomic_accession=73664104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100526840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526840, CurrentID=0, Status='live', Symbol='LOC100526840', Aliases='', description='chromosome 18 open reading frame 32 pseudogene', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134816338, end_position_on_the_genomic_accession=134817259, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100526841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526841, CurrentID=0, Status='live', Symbol='LOC100526841', Aliases='', description='chromosome 18 open reading frame 32 pseudogene', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202995162, end_position_on_the_genomic_accession=202997021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376506, CurrentID=0, Status='live', Symbol='LOC105376506', Aliases='', description='uncharacterized LOC105376506', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=432108, end_position_on_the_genomic_accession=440672, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  256764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256764, CurrentID=0, Status='live', Symbol='WDR72', Aliases='AI2A3', description='WD repeat domain 72', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53513741, end_position_on_the_genomic_accession=53762878, orientation='minus', exon_count=25, OMIM=613214, no_hdr0=''),
  105376509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376509, CurrentID=0, Status='live', Symbol='LOC105376509', Aliases='', description='uncharacterized LOC105376509', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=736676, end_position_on_the_genomic_accession=747315, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422173, CurrentID=0, Status='live', Symbol='LOC100422173', Aliases='', description='olfactory receptor, family 5, subfamily D, member 13 (gene/pseudogene) pseudogene', other_designations='olfactory receptor, family 5, subfamily D, member 13 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372586, CurrentID=0, Status='live', Symbol='LOC105372586', Aliases='', description='uncharacterized LOC105372586', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30292608, end_position_on_the_genomic_accession=30297883, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376513, CurrentID=0, Status='live', Symbol='LOC105376513', Aliases='', description='uncharacterized LOC105376513', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1350315, end_position_on_the_genomic_accession=1365193, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375904, CurrentID=0, Status='live', Symbol='LOC105375904', Aliases='', description='uncharacterized LOC105375904', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74589626, end_position_on_the_genomic_accession=74633263, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100422174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422174, CurrentID=0, Status='live', Symbol='LOC100422174', Aliases='', description='olfactory receptor, family 5, subfamily AK, member 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376517, CurrentID=0, Status='live', Symbol='LOC105376517', Aliases='', description='uncharacterized LOC105376517', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1771396, end_position_on_the_genomic_accession=1778593, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102466081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466081, CurrentID=0, Status='live', Symbol='MIR5739', Aliases='hsa-mir-5739', description='microRNA 5739', other_designations='microRNA mir-5739', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28459869, end_position_on_the_genomic_accession=28459948, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376520, CurrentID=0, Status='live', Symbol='LOC105376520', Aliases='', description='uncharacterized LOC105376520', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2373193, end_position_on_the_genomic_accession=2375686, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102466103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466103, CurrentID=0, Status='live', Symbol='MIR6075', Aliases='hsa-mir-6075', description='microRNA 6075', other_designations='microRNA mir-6075', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1510762, end_position_on_the_genomic_accession=1510856, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466104, CurrentID=0, Status='live', Symbol='MIR6090', Aliases='hsa-mir-6090', description='microRNA 6090', other_designations='microRNA mir-6090', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128522390, end_position_on_the_genomic_accession=128522449, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927935, CurrentID=0, Status='live', Symbol='LOC101927935', Aliases='', description='uncharacterized LOC101927935', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91851929, end_position_on_the_genomic_accession=91853248, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387853, CurrentID=0, Status='live', Symbol='OR7A19P', Aliases='', description='olfactory receptor family 7 subfamily A member 19 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 19 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46592574, end_position_on_the_genomic_accession=46593306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376528, CurrentID=0, Status='live', Symbol='LOC105376528', Aliases='', description='uncharacterized LOC105376528', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4666625, end_position_on_the_genomic_accession=4673496, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102725253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102725253, CurrentID=0, Status='live', Symbol='LOC102725253', Aliases='', description='uncharacterized LOC102725253', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927936, CurrentID=0, Status='live', Symbol='LOC101927936', Aliases='', description='uncharacterized LOC101927936', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8169082, end_position_on_the_genomic_accession=8180247, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102466162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466162, CurrentID=0, Status='live', Symbol='MIR548AZ', Aliases='hsa-mir-548az', description='microRNA 548az', other_designations='microRNA mir-548az', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119325171, end_position_on_the_genomic_accession=119325265, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422177, CurrentID=0, Status='live', Symbol='LOC100422177', Aliases='', description='olfactory receptor, family 10, subfamily Q, member 1 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376535, CurrentID=0, Status='live', Symbol='LOC105376535', Aliases='', description='uncharacterized LOC105376535', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7571274, end_position_on_the_genomic_accession=7597180, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102466189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466189, CurrentID=0, Status='live', Symbol='MIR6715A', Aliases='hsa-mir-6715a', description='microRNA 6715a', other_designations='microRNA mir-6715a', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112299612, end_position_on_the_genomic_accession=112299690, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466190, CurrentID=0, Status='live', Symbol='MIR6721', Aliases='hsa-mir-6721, mir-6721', description='microRNA 6721', other_designations='microRNA mir-6721', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32170030, end_position_on_the_genomic_accession=32170116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466191, CurrentID=0, Status='live', Symbol='MIR6736', Aliases='hsa-mir-6736', description='microRNA 6736', other_designations='microRNA mir-6736', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145850587, end_position_on_the_genomic_accession=145850645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466192, CurrentID=0, Status='live', Symbol='MIR6750', Aliases='hsa-mir-6750', description='microRNA 6750', other_designations='microRNA mir-6750', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64898363, end_position_on_the_genomic_accession=64898437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466193, CurrentID=0, Status='live', Symbol='MIR6757', Aliases='hsa-mir-6757', description='microRNA 6757', other_designations='microRNA mir-6757', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53056944, end_position_on_the_genomic_accession=53057012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466194, CurrentID=0, Status='live', Symbol='MIR6773', Aliases='hsa-mir-6773', description='microRNA 6773', other_designations='microRNA mir-6773', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68233426, end_position_on_the_genomic_accession=68233499, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466195, CurrentID=0, Status='live', Symbol='MIR6780A', Aliases='hsa-mir-6780a', description='microRNA 6780a', other_designations='microRNA mir-6780a', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42708084, end_position_on_the_genomic_accession=42708151, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466196, CurrentID=0, Status='live', Symbol='MIR6794', Aliases='hsa-mir-6794', description='microRNA 6794', other_designations='microRNA mir-6794', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12852260, end_position_on_the_genomic_accession=12852327, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466197, CurrentID=0, Status='live', Symbol='MIR6810', Aliases='hsa-mir-6810', description='microRNA 6810', other_designations='microRNA mir-6810', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218341911, end_position_on_the_genomic_accession=218341980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466198, CurrentID=0, Status='live', Symbol='MIR6817', Aliases='hsa-mir-6817', description='microRNA 6817', other_designations='microRNA mir-6817', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25455646, end_position_on_the_genomic_accession=25455711, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466199, CurrentID=0, Status='live', Symbol='MIR6825', Aliases='hsa-mir-6825', description='microRNA 6825', other_designations='microRNA mir-6825', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127575266, end_position_on_the_genomic_accession=127575331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466200, CurrentID=0, Status='live', Symbol='MIR6844', Aliases='hsa-mir-6844', description='microRNA 6844', other_designations='microRNA mir-6844', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124508515, end_position_on_the_genomic_accession=124508576, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466201, CurrentID=0, Status='live', Symbol='MIR6853', Aliases='hsa-mir-6853', description='microRNA 6853', other_designations='microRNA mir-6853', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35732922, end_position_on_the_genomic_accession=35732995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466202, CurrentID=0, Status='live', Symbol='MIR6769B', Aliases='hsa-mir-6769b', description='microRNA 6769b', other_designations='microRNA mir-6769b', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206474803, end_position_on_the_genomic_accession=206474864, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466203, CurrentID=0, Status='live', Symbol='MIR6874', Aliases='hsa-mir-6874', description='microRNA 6874', other_designations='microRNA mir-6874', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5711840, end_position_on_the_genomic_accession=5711910, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466204, CurrentID=0, Status='live', Symbol='MIR6880', Aliases='hsa-mir-6880', description='microRNA 6880', other_designations='microRNA mir-6880', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124337181, end_position_on_the_genomic_accession=124337242, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466205, CurrentID=0, Status='live', Symbol='MIR6887', Aliases='hsa-mir-6887', description='microRNA 6887', other_designations='microRNA mir-6887', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35122700, end_position_on_the_genomic_accession=35122764, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376539, CurrentID=0, Status='live', Symbol='LOC105376539', Aliases='', description='uncharacterized LOC105376539', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8766783, end_position_on_the_genomic_accession=8779535, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102466222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466222, CurrentID=0, Status='live', Symbol='MIR7106', Aliases='hsa-mir-7106', description='microRNA 7106', other_designations='microRNA mir-7106', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113159113, end_position_on_the_genomic_accession=113159177, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466223, CurrentID=0, Status='live', Symbol='MIR7114', Aliases='hsa-mir-7114', description='microRNA 7114', other_designations='microRNA mir-7114', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137450026, end_position_on_the_genomic_accession=137450086, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466225, CurrentID=0, Status='live', Symbol='MIR6077', Aliases='MIR6077-1-2, hsa-mir-6077, hsa-mir-6077-1, hsa-mir-6077-2, MIR6077', description='microRNA 6077', other_designations='microRNA 6077-1|microRNA 6077-2|microRNA mir-6077-1|microRNA mir-6077-2', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148388282, end_position_on_the_genomic_accession=148388363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466227, CurrentID=0, Status='live', Symbol='MIR7162', Aliases='hsa-mir-7162', description='microRNA 7162', other_designations='microRNA mir-7162', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30368597, end_position_on_the_genomic_accession=30368665, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466235, CurrentID=0, Status='live', Symbol='MIR7515', Aliases='hsa-mir-7515', description='microRNA 7515', other_designations='microRNA mir-7515', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6650373, end_position_on_the_genomic_accession=6650439, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466247, CurrentID=0, Status='live', Symbol='MIR1273H', Aliases='hsa-mir-1273h', description='microRNA 1273h', other_designations='microRNA mir-1273h', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24203116, end_position_on_the_genomic_accession=24203231, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466250, CurrentID=0, Status='live', Symbol='MIR7973-1', Aliases='hsa-mir-7973-1', description='microRNA 7973-1', other_designations='microRNA mir-7973-1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51314034, end_position_on_the_genomic_accession=51314109, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466251, CurrentID=0, Status='live', Symbol='MIR8061', Aliases='hsa-mir-8061', description='microRNA 8061', other_designations='microRNA mir-8061', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54645309, end_position_on_the_genomic_accession=54645383, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466252, CurrentID=0, Status='live', Symbol='MIR8069-1', Aliases='MIR8069, hsa-mir-8069, hsa-mir-8069-1', description='microRNA 8069-1', other_designations='microRNA 8069|microRNA mir-8069', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6859171, end_position_on_the_genomic_accession=6859256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466253, CurrentID=0, Status='live', Symbol='MIR8076', Aliases='hsa-mir-8076', description='microRNA 8076', other_designations='microRNA mir-8076', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113432118, end_position_on_the_genomic_accession=113432200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100631247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631247, CurrentID=0, Status='live', Symbol='GRAMD4P8', Aliases='', description='GRAM domain containing 4 pseudogene 8', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131646704, end_position_on_the_genomic_accession=131648383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100631248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631248, CurrentID=0, Status='live', Symbol='GRAMD4P3', Aliases='', description='GRAM domain containing 4 pseudogene 3', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19371817, end_position_on_the_genomic_accession=19373496, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631249, CurrentID=0, Status='live', Symbol='GRAMD4P5', Aliases='', description='GRAM domain containing 4 pseudogene 5', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20813290, end_position_on_the_genomic_accession=20814973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631250, CurrentID=0, Status='live', Symbol='GRAMD4P6', Aliases='', description='GRAM domain containing 4 pseudogene 6', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21382383, end_position_on_the_genomic_accession=21384066, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102466259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466259, CurrentID=0, Status='live', Symbol='MIR6770-3', Aliases='hsa-mir-6770-3', description='microRNA 6770-3', other_designations='microRNA mir-6770-3', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18379351, end_position_on_the_genomic_accession=18379410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100631252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631252, CurrentID=0, Status='live', Symbol='GRAMD4P1', Aliases='', description='GRAM domain containing 4 pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13665867, end_position_on_the_genomic_accession=13667557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100631253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631253, CurrentID=0, Status='live', Symbol='TBC1D3P4', Aliases='USP6P2', description='TBC1 domain family member 3 pseudogene 4', other_designations='TBC1 domain family, member 3 pseudogene 4|ubiquitin specific peptidase 6 pseudogene 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18394899, end_position_on_the_genomic_accession=18401264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631254, CurrentID=0, Status='live', Symbol='CSPG4P11', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 11', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84186755, end_position_on_the_genomic_accession=84192178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631255, CurrentID=0, Status='live', Symbol='LOC100631255', Aliases='', description='arginine/serine-rich coiled-coil 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89875998, end_position_on_the_genomic_accession=89877290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376548, CurrentID=0, Status='live', Symbol='LOC105376548', Aliases='', description='uncharacterized LOC105376548', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11152396, end_position_on_the_genomic_accession=11156532, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100631258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631258, CurrentID=0, Status='live', Symbol='LOC100631258', Aliases='', description='isocitrate dehydrogenase 3 (NAD+) gamma pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=21283781, end_position_on_the_genomic_accession=21284939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102466268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466268, CurrentID=0, Status='live', Symbol='MIR6511A1', Aliases='MIR6511A-1, hsa-mir-6511a-1', description='microRNA 6511a-1', other_designations='microRNA mir-6511a-1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14925937, end_position_on_the_genomic_accession=14926003, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100631261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631261, CurrentID=0, Status='live', Symbol='MARK2P14', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 14', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11906572, end_position_on_the_genomic_accession=11913429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102466270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466270, CurrentID=0, Status='live', Symbol='MIR6741', Aliases='hsa-mir-6741', description='microRNA 6741', other_designations='microRNA mir-6741', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225922080, end_position_on_the_genomic_accession=225922142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466271, CurrentID=0, Status='live', Symbol='MIR6893', Aliases='hsa-mir-6893', description='microRNA 6893', other_designations='microRNA mir-6893', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144435551, end_position_on_the_genomic_accession=144435619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100631265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631265, CurrentID=0, Status='live', Symbol='LOC100631265', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20546501, end_position_on_the_genomic_accession=20547689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100631266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631266, CurrentID=0, Status='live', Symbol='LOC100631266', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22453033, end_position_on_the_genomic_accession=22454221, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631267, CurrentID=0, Status='live', Symbol='LOC100631267', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23141478, end_position_on_the_genomic_accession=23142666, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100631268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631268, CurrentID=0, Status='live', Symbol='LOC100631268', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28547668, end_position_on_the_genomic_accession=28548856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631269, CurrentID=0, Status='live', Symbol='SLC10A5P1', Aliases='', description='SLC10A5 pseudogene 1', other_designations='solute carrier family 10 (sodium/bile acid cotransporter family), member 5 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81634355, end_position_on_the_genomic_accession=81635442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422181, CurrentID=0, Status='live', Symbol='LOC100422181', Aliases='', description='olfactory receptor, family 7, subfamily E, member 24 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  650024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650024, CurrentID=0, Status='live', Symbol='LOC650024', Aliases='', description='cancer/testis antigen 55 pseudogene', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135201732, end_position_on_the_genomic_accession=135215229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631287, CurrentID=0, Status='live', Symbol='LOC100631287', Aliases='', description='keratin 18 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631288, CurrentID=0, Status='live', Symbol='LOC100631288', Aliases='', description='keratin 18 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631289, CurrentID=0, Status='live', Symbol='LOC100631289', Aliases='', description='keratin 18 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631290, CurrentID=0, Status='live', Symbol='LOC100631290', Aliases='', description='keratin 18 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631291, CurrentID=0, Status='live', Symbol='LOC100631291', Aliases='', description='keratin 38 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631292, CurrentID=0, Status='live', Symbol='LOC100631292', Aliases='', description='keratin 18 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660605, CurrentID=0, Status='live', Symbol='MIR2117HG', Aliases='', description='MIR2117 host gene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660606, CurrentID=0, Status='live', Symbol='LOC106660606', Aliases='', description='uncharacterized LOC106660606', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660607, CurrentID=0, Status='live', Symbol='DUX4L42', Aliases='', description='double homeobox 4 like 42, pseudogene', other_designations='', map_location='22p11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660609, CurrentID=0, Status='live', Symbol='MIR4432HG', Aliases='', description='MIR4432 host gene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660610, CurrentID=0, Status='live', Symbol='MIR5689HG', Aliases='', description='MIR5689 host gene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660611, CurrentID=0, Status='live', Symbol='RNU2-22P', Aliases='', description='RNA, U2 small nuclear 22, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660612, CurrentID=0, Status='live', Symbol='LINC00680', Aliases='', description='long intergenic non-protein coding RNA 680', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660613, CurrentID=0, Status='live', Symbol='LINC00680-GUSBP4', Aliases='', description='LINC00680-GUSBP4 readthrough', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631302, CurrentID=0, Status='live', Symbol='LOC100631302', Aliases='', description='tenascin XB pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=492295, CurrentID=0, Status='live', Symbol='RAC1P3', Aliases='Psi3Rac1', description='ras-related C3 botulinum toxin substrate 1 pseudogene 3', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41287485, end_position_on_the_genomic_accession=41288439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106660616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660616, CurrentID=0, Status='live', Symbol='SAP18P3', Aliases='', description='SAP18 pseudogene 3', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660617, CurrentID=0, Status='live', Symbol='DUX4L43', Aliases='', description='double homeobox 4 like 43, pseudogene', other_designations='', map_location='22p11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660618, CurrentID=0, Status='live', Symbol='MCDR5', Aliases='', description='macular dystrophy 5', other_designations='', map_location='19q13.31-q13.32', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660619, CurrentID=0, Status='live', Symbol='TP53TG3GP', Aliases='', description='TP53 target 3 family member G, pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660620, CurrentID=0, Status='live', Symbol='RARRES2P9', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 9', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631309, CurrentID=0, Status='live', Symbol='LOC100631309', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631310, CurrentID=0, Status='live', Symbol='LOC100631310', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=492303, CurrentID=0, Status='live', Symbol='GEMIN8P4', Aliases='', description='gem nuclear organelle associated protein 8 pseudogene 4', other_designations='FAM51A1 pseudogene|gem (nuclear organelle) associated protein 8 pseudogene 4', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89993265, end_position_on_the_genomic_accession=89994966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  492307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=492307, CurrentID=0, Status='live', Symbol='C8orf22', Aliases='', description='chromosome 8 open reading frame 22', other_designations='exocrine differentiation and proliferation factor-like protein', map_location='8q11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49054336, end_position_on_the_genomic_accession=49076083, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100631318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631318, CurrentID=0, Status='live', Symbol='LOC100631318', Aliases='', description='zinc finger protein 85 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=492311, CurrentID=0, Status='live', Symbol='IGIP', Aliases='C5orf53', description='IgA-inducing protein', other_designations='IgA-inducing protein homolog (Bos taurus)', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140125936, end_position_on_the_genomic_accession=140128806, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100631320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631320, CurrentID=0, Status='live', Symbol='LOC100631320', Aliases='', description='zinc finger protein 486 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106660633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106660633, CurrentID=0, Status='live', Symbol='MTATP6P6', Aliases='', description='mitochondrially encoded ATP synthase 6 pseudogene 6', other_designations='', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631322, CurrentID=0, Status='live', Symbol='LOC100631322', Aliases='', description='MAM domain containing 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376559, CurrentID=0, Status='live', Symbol='LOC105376559', Aliases='', description='uncharacterized LOC105376559', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13360016, end_position_on_the_genomic_accession=13381154, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100631330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631330, CurrentID=0, Status='live', Symbol='LOC100631330', Aliases='', description='polycomb group ring finger 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631337, CurrentID=0, Status='live', Symbol='LOC100631337', Aliases='', description='nitric oxide synthase trafficking pseudogene', other_designations='nitric oxide synthase trafficker pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376562, CurrentID=0, Status='live', Symbol='LOC105376562', Aliases='', description='uncharacterized LOC105376562', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13781379, end_position_on_the_genomic_accession=13847956, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100631343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631343, CurrentID=0, Status='live', Symbol='LOC100631343', Aliases='', description='diffuse panbronchiolitis critical region 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631344, CurrentID=0, Status='live', Symbol='LOC100631344', Aliases='', description='diffuse panbronchiolitis critical region 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631345, CurrentID=0, Status='live', Symbol='LOC100631345', Aliases='', description='diffuse panbronchiolitis critical region 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631346, CurrentID=0, Status='live', Symbol='LOC100631346', Aliases='', description='diffuse panbronchiolitis critical region 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631354, CurrentID=0, Status='live', Symbol='LOC100631354', Aliases='', description='SMAD specific E3 ubiquitin protein ligase 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421500, CurrentID=0, Status='live', Symbol='MARK2P6', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 6', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128851620, end_position_on_the_genomic_accession=128853008, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376565, CurrentID=0, Status='live', Symbol='LOC105376565', Aliases='', description='uncharacterized LOC105376565', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14905082, end_position_on_the_genomic_accession=14915499, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100631370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631370, CurrentID=0, Status='live', Symbol='LOC100631370', Aliases='', description='immunoglobulin superfamily, DCC subclass, member 3 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376567, CurrentID=0, Status='live', Symbol='LOC105376567', Aliases='', description='uncharacterized LOC105376567', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15700972, end_position_on_the_genomic_accession=15768476, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100631372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631372, CurrentID=0, Status='live', Symbol='LOC100631372', Aliases='', description='ATP binding cassette subfamily B member 10 pseudogene', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631373, CurrentID=0, Status='live', Symbol='LOC100631373', Aliases='', description='ATP binding cassette subfamily B member 10 pseudogene', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631374, CurrentID=0, Status='live', Symbol='LOC100631374', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631375, CurrentID=0, Status='live', Symbol='LOC100631375', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631376, CurrentID=0, Status='live', Symbol='LOC100631376', Aliases='', description='sphingomyelin phosphodiesterase 3, neutral membrane (neutral sphingomyelinase II) pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  256824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256824, CurrentID=0, Status='live', Symbol='LOC256824', Aliases='', description='cryptic gene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100631378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631378, CurrentID=0, Status='live', Symbol='LOC100631378', Aliases='', description='uncharacterized 100631378', other_designations='', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37823722, end_position_on_the_genomic_accession=37855196, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100631380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631380, CurrentID=0, Status='live', Symbol='LOC100631380', Aliases='', description='protein phosphatase 6, regulatory subunit 2 pseudogene 1', other_designations='', map_location='18p11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14143597, end_position_on_the_genomic_accession=14151139, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100631381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631381, CurrentID=0, Status='live', Symbol='PPP6R2P1', Aliases='', description='protein phosphatase 6 regulatory subunit 2 pseudogene 1', other_designations='protein phosphatase 6, regulatory subunit 2 pseudogene 1', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14007135, end_position_on_the_genomic_accession=14014842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631382, CurrentID=0, Status='live', Symbol='ZFAND6P1', Aliases='', description='zinc finger, AN1-type domain 6 pseudogene 1', other_designations='', map_location='9p', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35449640, end_position_on_the_genomic_accession=35451108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100631383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100631383, CurrentID=0, Status='live', Symbol='FAM47E-STBD1', Aliases='', description='FAM47E-STBD1 readthrough', other_designations='', map_location='4q', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76251700, end_position_on_the_genomic_accession=76311130, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100616165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616165, CurrentID=0, Status='live', Symbol='MIR4771-2', Aliases='', description='microRNA 4771-2', other_designations='hsa-mir-4771-2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111771061, end_position_on_the_genomic_accession=111771134, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480139, CurrentID=0, Status='live', Symbol='RNU6-1313P', Aliases='', description='RNA, U6 small nuclear 1313, pseudogene', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422185, CurrentID=0, Status='live', Symbol='LOC100422185', Aliases='', description='olfactory receptor, family 4, subfamily K, member 15 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375917, CurrentID=0, Status='live', Symbol='LOC105375917', Aliases='', description='uncharacterized LOC105375917', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79892070, end_position_on_the_genomic_accession=79931062, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422382, CurrentID=0, Status='live', Symbol='LOC100422382', Aliases='', description='nocturnin pseudogene', other_designations='CCR4 carbon catabolite repression 4-like (S. cerevisiae) pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114453074, end_position_on_the_genomic_accession=114454125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376583, CurrentID=0, Status='live', Symbol='LOC105376583', Aliases='', description='uncharacterized LOC105376583', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20187503, end_position_on_the_genomic_accession=20199929, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376586, CurrentID=0, Status='live', Symbol='LOC105376586', Aliases='', description='uncharacterized LOC105376586', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20670354, end_position_on_the_genomic_accession=20672328, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376587, CurrentID=0, Status='live', Symbol='LOC105376587', Aliases='', description='uncharacterized LOC105376587', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=21912503, end_position_on_the_genomic_accession=21932735, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927948, CurrentID=0, Status='live', Symbol='LOC101927948', Aliases='', description='uncharacterized LOC101927948', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78088730, end_position_on_the_genomic_accession=78127806, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102466515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466515, CurrentID=0, Status='live', Symbol='MIR1199', Aliases='hsa-mir-1199', description='microRNA 1199', other_designations='microRNA mir-1199', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14073361, end_position_on_the_genomic_accession=14073479, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466516, CurrentID=0, Status='live', Symbol='MIR6071', Aliases='hsa-mir-6071', description='microRNA 6071', other_designations='microRNA mir-6071', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85783600, end_position_on_the_genomic_accession=85783677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466518, CurrentID=0, Status='live', Symbol='MIR6081', Aliases='hsa-mir-6081', description='microRNA 6081', other_designations='microRNA mir-6081', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95065350, end_position_on_the_genomic_accession=95065446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466519, CurrentID=0, Status='live', Symbol='MIR6086', Aliases='hsa-mir-6086', description='microRNA 6086', other_designations='microRNA mir-6086', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13590292, end_position_on_the_genomic_accession=13590346, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371555, CurrentID=0, Status='live', Symbol='LOC105371555', Aliases='', description='uncharacterized LOC105371555', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16486552, end_position_on_the_genomic_accession=16503709, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376593, CurrentID=0, Status='live', Symbol='LOC105376593', Aliases='', description='uncharacterized LOC105376593', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23730588, end_position_on_the_genomic_accession=23749918, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375920, CurrentID=0, Status='live', Symbol='LOC105375920', Aliases='', description='uncharacterized LOC105375920', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80177711, end_position_on_the_genomic_accession=80231553, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376595, CurrentID=0, Status='live', Symbol='LOC105376595', Aliases='', description='uncharacterized LOC105376595', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=24310689, end_position_on_the_genomic_accession=24338799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376596, CurrentID=0, Status='live', Symbol='LOC105376596', Aliases='', description='uncharacterized LOC105376596', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=25734786, end_position_on_the_genomic_accession=25848603, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100422190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422190, CurrentID=0, Status='live', Symbol='LOC100422190', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B7 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69387613, end_position_on_the_genomic_accession=69388312, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270838, CurrentID=0, Status='live', Symbol='RPL5P21', Aliases='RPL5_7_753', description='ribosomal protein L5 pseudogene 21', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129756333, end_position_on_the_genomic_accession=129757112, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129089, CurrentID=0, Status='live', Symbol='LOC100129089', Aliases='', description='uncharacterized LOC100129089', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=72846945, end_position_on_the_genomic_accession=72847790, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  644893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644893, CurrentID=0, Status='live', Symbol='LOC644893', Aliases='', description='MAGE family member E1 pseudogene', other_designations='melanoma antigen family E, 1 pseudogene|melanoma antigen family E1 pseudogene', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55508685, end_position_on_the_genomic_accession=55523885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376605, CurrentID=0, Status='live', Symbol='LOC105376605', Aliases='', description='uncharacterized LOC105376605', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29335713, end_position_on_the_genomic_accession=29531385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376606, CurrentID=0, Status='live', Symbol='LOC105376606', Aliases='', description='uncharacterized LOC105376606', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29618802, end_position_on_the_genomic_accession=29630766, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102466615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466615, CurrentID=0, Status='live', Symbol='MIR6127', Aliases='hsa-mir-6127', description='microRNA 6127', other_designations='microRNA mir-6127', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22633258, end_position_on_the_genomic_accession=22633366, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466616, CurrentID=0, Status='live', Symbol='MIR6132', Aliases='hsa-mir-6132', description='microRNA 6132', other_designations='microRNA mir-6132', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117020211, end_position_on_the_genomic_accession=117020319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376607, CurrentID=0, Status='live', Symbol='LOC105376607', Aliases='', description='uncharacterized LOC105376607', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30042722, end_position_on_the_genomic_accession=30322980, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288644, CurrentID=0, Status='live', Symbol='LOC100288644', Aliases='', description='uncharacterized LOC100288644', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=186346203, end_position_on_the_genomic_accession=186348049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375923, CurrentID=0, Status='live', Symbol='LOC105375923', Aliases='', description='uncharacterized LOC105375923', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80558182, end_position_on_the_genomic_accession=80580165, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480147, CurrentID=0, Status='live', Symbol='RNU6-1331P', Aliases='', description='RNA, U6 small nuclear 1331, pseudogene', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927250, CurrentID=0, Status='live', Symbol='LOC101927250', Aliases='', description='uncharacterized LOC101927250', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50100478, end_position_on_the_genomic_accession=50153014, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376611, CurrentID=0, Status='live', Symbol='LOC105376611', Aliases='', description='uncharacterized LOC105376611', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30868523, end_position_on_the_genomic_accession=30889948, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102466656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466656, CurrentID=0, Status='live', Symbol='MIR6500', Aliases='hsa-mir-6500', description='microRNA 6500', other_designations='microRNA mir-6500', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51060018, end_position_on_the_genomic_accession=51060103, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466657, CurrentID=0, Status='live', Symbol='MIR6505', Aliases='hsa-mir-6505, mir-6505', description='microRNA 6505', other_designations='microRNA mir-6505', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48132797, end_position_on_the_genomic_accession=48132867, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466658, CurrentID=0, Status='live', Symbol='MIR6510', Aliases='hsa-mir-6510, mir-6510', description='microRNA 6510', other_designations='microRNA mir-6510', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41517164, end_position_on_the_genomic_accession=41517217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466659, CurrentID=0, Status='live', Symbol='MIR6515', Aliases='hsa-mir-6515, mir-6515', description='microRNA 6515', other_designations='microRNA mir-6515', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12940484, end_position_on_the_genomic_accession=12940540, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129254, CurrentID=0, Status='live', Symbol='LOC100129254', Aliases='', description='chromosome 3 open reading frame 38 pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19630429, end_position_on_the_genomic_accession=19632169, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376619, CurrentID=0, Status='live', Symbol='LOC105376619', Aliases='', description='uncharacterized LOC105376619', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33804920, end_position_on_the_genomic_accession=33810357, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100422195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422195, CurrentID=0, Status='live', Symbol='LOC100422195', Aliases='', description='olfactory receptor, family 4, subfamily F, member 21 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102466719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466719, CurrentID=0, Status='live', Symbol='MIR6716', Aliases='hsa-mir-6716', description='microRNA 6716', other_designations='microRNA mir-6716', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118644000, end_position_on_the_genomic_accession=118644079, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466720, CurrentID=0, Status='live', Symbol='MIR6720', Aliases='hsa-mir-6720, mir-6720', description='microRNA 6720', other_designations='microRNA mir-6720', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1390314, end_position_on_the_genomic_accession=1390411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466721, CurrentID=0, Status='live', Symbol='MIR892C', Aliases='hsa-mir-892c', description='microRNA 892c', other_designations='microRNA mir-892c', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145992750, end_position_on_the_genomic_accession=145992826, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466722, CurrentID=0, Status='live', Symbol='MIR6730', Aliases='hsa-mir-6730', description='microRNA 6730', other_designations='microRNA mir-6730', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12578957, end_position_on_the_genomic_accession=12579023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466723, CurrentID=0, Status='live', Symbol='MIR6734', Aliases='hsa-mir-6734', description='microRNA 6734', other_designations='microRNA mir-6734', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43364648, end_position_on_the_genomic_accession=43364715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466724, CurrentID=0, Status='live', Symbol='MIR6739', Aliases='hsa-mir-6739', description='microRNA 6739', other_designations='microRNA mir-6739', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201863373, end_position_on_the_genomic_accession=201863447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466725, CurrentID=0, Status='live', Symbol='MIR6744', Aliases='hsa-mir-6744', description='microRNA 6744', other_designations='microRNA mir-6744', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1256605, end_position_on_the_genomic_accession=1256670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466726, CurrentID=0, Status='live', Symbol='MIR6745', Aliases='hsa-mir-6745', description='microRNA 6745', other_designations='microRNA mir-6745', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47179611, end_position_on_the_genomic_accession=47179737, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466727, CurrentID=0, Status='live', Symbol='MIR6749', Aliases='hsa-mir-6749', description='microRNA 6749', other_designations='microRNA mir-6749', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64902387, end_position_on_the_genomic_accession=64902455, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466728, CurrentID=0, Status='live', Symbol='MIR6754', Aliases='hsa-mir-6754', description='microRNA 6754', other_designations='microRNA mir-6754', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71473503, end_position_on_the_genomic_accession=71473568, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466729, CurrentID=0, Status='live', Symbol='MIR6759', Aliases='hsa-mir-6759', description='microRNA 6759', other_designations='microRNA mir-6759', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57748618, end_position_on_the_genomic_accession=57748682, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466730, CurrentID=0, Status='live', Symbol='MIR6764', Aliases='hsa-mir-6764', description='microRNA 6764', other_designations='microRNA mir-6764', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100277357, end_position_on_the_genomic_accession=100277417, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466731, CurrentID=0, Status='live', Symbol='MIR6769A', Aliases='hsa-mir-6769a', description='microRNA 6769a', other_designations='microRNA mir-6769a', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4671318, end_position_on_the_genomic_accession=4671390, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466732, CurrentID=0, Status='live', Symbol='MIR6774', Aliases='hsa-mir-6774', description='microRNA 6774', other_designations='microRNA mir-6774', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85918347, end_position_on_the_genomic_accession=85918416, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466733, CurrentID=0, Status='live', Symbol='MIR6778', Aliases='hsa-mir-6778', description='microRNA 6778', other_designations='microRNA mir-6778', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18340814, end_position_on_the_genomic_accession=18340886, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466734, CurrentID=0, Status='live', Symbol='MIR6783', Aliases='hsa-mir-6783', description='microRNA 6783', other_designations='microRNA mir-6783', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44934618, end_position_on_the_genomic_accession=44934681, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466735, CurrentID=0, Status='live', Symbol='MIR6788', Aliases='hsa-mir-6788', description='microRNA 6788', other_designations='microRNA mir-6788', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10759584, end_position_on_the_genomic_accession=10759649, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466736, CurrentID=0, Status='live', Symbol='MIR6789', Aliases='hsa-mir-6789', description='microRNA 6789', other_designations='microRNA mir-6789', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2235829, end_position_on_the_genomic_accession=2235926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466737, CurrentID=0, Status='live', Symbol='MIR6793', Aliases='hsa-mir-6793', description='microRNA 6793', other_designations='microRNA mir-6793', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10828973, end_position_on_the_genomic_accession=10829035, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466738, CurrentID=0, Status='live', Symbol='MIR6798', Aliases='hsa-mir-6798', description='microRNA 6798', other_designations='microRNA mir-6798', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49009906, end_position_on_the_genomic_accession=49009972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466739, CurrentID=0, Status='live', Symbol='MIR6803', Aliases='hsa-mir-6803', description='microRNA 6803', other_designations='microRNA mir-6803', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55245186, end_position_on_the_genomic_accession=55245250, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466740, CurrentID=0, Status='live', Symbol='MIR6808', Aliases='hsa-mir-6808', description='microRNA 6808', other_designations='microRNA mir-6808', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1339650, end_position_on_the_genomic_accession=1339708, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466741, CurrentID=0, Status='live', Symbol='MIR6813', Aliases='hsa-mir-6813', description='microRNA 6813', other_designations='microRNA mir-6813', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64076955, end_position_on_the_genomic_accession=64077010, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466742, CurrentID=0, Status='live', Symbol='MIR6818', Aliases='hsa-mir-6818', description='microRNA 6818', other_designations='microRNA mir-6818', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30007049, end_position_on_the_genomic_accession=30007113, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466743, CurrentID=0, Status='live', Symbol='MIR6822', Aliases='hsa-mir-6822', description='microRNA 6822', other_designations='microRNA mir-6822', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39138206, end_position_on_the_genomic_accession=39138266, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466744, CurrentID=0, Status='live', Symbol='MIR6827', Aliases='hsa-mir-6827', description='microRNA 6827', other_designations='microRNA mir-6827', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134367804, end_position_on_the_genomic_accession=134367862, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466745, CurrentID=0, Status='live', Symbol='MIR6832', Aliases='hsa-mir-6832', description='microRNA 6832', other_designations='microRNA mir-6832', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31633787, end_position_on_the_genomic_accession=31633858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466746, CurrentID=0, Status='live', Symbol='MIR6780B', Aliases='hsa-mir-6780b', description='microRNA 6780b', other_designations='microRNA mir-6780b', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43434542, end_position_on_the_genomic_accession=43434620, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466747, CurrentID=0, Status='live', Symbol='MIR6840', Aliases='hsa-mir-6840', description='microRNA 6840', other_designations='microRNA mir-6840', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100356651, end_position_on_the_genomic_accession=100356721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466748, CurrentID=0, Status='live', Symbol='MIR6845', Aliases='hsa-mir-6845', description='microRNA 6845', other_designations='microRNA mir-6845', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143837756, end_position_on_the_genomic_accession=143837816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466749, CurrentID=0, Status='live', Symbol='MIR6849', Aliases='hsa-mir-6849', description='microRNA 6849', other_designations='microRNA mir-6849', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144400277, end_position_on_the_genomic_accession=144400345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466750, CurrentID=0, Status='live', Symbol='MIR6855', Aliases='hsa-mir-6855', description='microRNA 6855', other_designations='microRNA mir-6855', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129869605, end_position_on_the_genomic_accession=129869671, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466751, CurrentID=0, Status='live', Symbol='MIR6859-1', Aliases='hsa-mir-6859-1', description='microRNA 6859-1', other_designations='microRNA mir-6859-1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17369, end_position_on_the_genomic_accession=17436, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466752, CurrentID=0, Status='live', Symbol='MIR6863', Aliases='hsa-mir-6863', description='microRNA 6863', other_designations='microRNA mir-6863', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56904264, end_position_on_the_genomic_accession=56904353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466753, CurrentID=0, Status='live', Symbol='MIR6868', Aliases='hsa-mir-6868', description='microRNA 6868', other_designations='microRNA mir-6868', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76098019, end_position_on_the_genomic_accession=76098076, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466754, CurrentID=0, Status='live', Symbol='MIR6873', Aliases='hsa-mir-6873', description='microRNA 6873', other_designations='microRNA mir-6873', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33287227, end_position_on_the_genomic_accession=33287289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466755, CurrentID=0, Status='live', Symbol='MIR6875', Aliases='hsa-mir-6875', description='microRNA 6875', other_designations='microRNA mir-6875', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100868036, end_position_on_the_genomic_accession=100868107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466756, CurrentID=0, Status='live', Symbol='MIR6879', Aliases='hsa-mir-6879', description='microRNA 6879', other_designations='microRNA mir-6879', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65018505, end_position_on_the_genomic_accession=65018570, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466757, CurrentID=0, Status='live', Symbol='MIR6884', Aliases='hsa-mir-6884', description='microRNA 6884', other_designations='microRNA mir-6884', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40026332, end_position_on_the_genomic_accession=40026409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466758, CurrentID=0, Status='live', Symbol='MIR6889', Aliases='hsa-mir-6889', description='microRNA 6889', other_designations='microRNA mir-6889', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41252992, end_position_on_the_genomic_accession=41253050, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466759, CurrentID=0, Status='live', Symbol='MIR6894', Aliases='hsa-mir-6894', description='microRNA 6894', other_designations='microRNA mir-6894', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53198889, end_position_on_the_genomic_accession=53198945, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480151, CurrentID=0, Status='live', Symbol='RNU6-1340P', Aliases='', description='RNA, U6 small nuclear 1340, pseudogene', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129774, CurrentID=0, Status='live', Symbol='LOC100129774', Aliases='', description='ATPase family, AAA domain containing 1 pseudogene', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1362443, end_position_on_the_genomic_accession=1363873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326289, CurrentID=0, Status='live', Symbol='RPL18AP1', Aliases='RPL18A_3_1401', description='ribosomal protein L18a pseudogene 1', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93223549, end_position_on_the_genomic_accession=93224175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644899, CurrentID=0, Status='live', Symbol='TERF2IPP1', Aliases='', description='telomeric repeat binding factor 2, interacting protein pseudogene 1', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17418645, end_position_on_the_genomic_accession=17420678, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927957, CurrentID=0, Status='live', Symbol='LINC01572', Aliases='', description='long intergenic non-protein coding RNA 1572', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72283301, end_position_on_the_genomic_accession=72665009, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105376636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376636, CurrentID=0, Status='live', Symbol='LOC105376636', Aliases='', description='uncharacterized LOC105376636', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=39542646, end_position_on_the_genomic_accession=39549877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480299, CurrentID=0, Status='live', Symbol='CTHRC1P1', Aliases='', description='collagen triple helix repeat containing 1 pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100500719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500719, CurrentID=0, Status='live', Symbol='LOC100500719', Aliases='', description='Hermansky-Pudlak syndrome 1 pseudogene', other_designations='', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36536652, end_position_on_the_genomic_accession=36537714, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102466806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466806, CurrentID=0, Status='live', Symbol='MIR7108', Aliases='hsa-mir-7108', description='microRNA 7108', other_designations='microRNA mir-7108', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2434914, end_position_on_the_genomic_accession=2435000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927958, CurrentID=0, Status='live', Symbol='LOC101927958', Aliases='', description='uncharacterized LOC101927958', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238235931, end_position_on_the_genomic_accession=238237206, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102466812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466812, CurrentID=0, Status='live', Symbol='MIR6511A2', Aliases='MIR6511A-2, hsa-mir-6511a-2', description='microRNA 6511a-2', other_designations='microRNA mir-6511a-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16324588, end_position_on_the_genomic_accession=16324654, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103908605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103908605, CurrentID=0, Status='live', Symbol='LOC103908605', Aliases='', description='uncharacterized LOC103908605', other_designations='', map_location='9p11.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42567859, end_position_on_the_genomic_accession=42569349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102466814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466814, CurrentID=0, Status='live', Symbol='MIR7151', Aliases='hsa-mir-7151', description='microRNA 7151', other_designations='microRNA mir-7151', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67403351, end_position_on_the_genomic_accession=67403410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466815, CurrentID=0, Status='live', Symbol='MIR7155', Aliases='hsa-mir-7155', description='microRNA 7155', other_designations='microRNA mir-7155', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64341849, end_position_on_the_genomic_accession=64341904, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466816, CurrentID=0, Status='live', Symbol='MIR7159', Aliases='hsa-mir-7159', description='microRNA 7159', other_designations='microRNA mir-7159', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33899135, end_position_on_the_genomic_accession=33899200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500737, CurrentID=0, Status='live', Symbol='PPIAP20', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 20', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12508129, end_position_on_the_genomic_accession=12508263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371565, CurrentID=0, Status='live', Symbol='LOC105371565', Aliases='', description='uncharacterized LOC105371565', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165535531, end_position_on_the_genomic_accession=165535915, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100500773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500773, CurrentID=0, Status='live', Symbol='LOC100500773', Aliases='', description='serine/arginine-rich splicing factor 3 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95946984, end_position_on_the_genomic_accession=95948348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466854, CurrentID=0, Status='live', Symbol='MIR7705', Aliases='hsa-mir-7705', description='microRNA 7705', other_designations='microRNA mir-7705', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100702968, end_position_on_the_genomic_accession=100703024, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466864, CurrentID=0, Status='live', Symbol='MIR6516', Aliases='hsa-mir-6516', description='microRNA 6516', other_designations='microRNA mir-6516', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77089417, end_position_on_the_genomic_accession=77089497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466865, CurrentID=0, Status='live', Symbol='MIR7848', Aliases='hsa-mir-7848', description='microRNA 7848', other_designations='microRNA mir-7848', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133046481, end_position_on_the_genomic_accession=133046581, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466866, CurrentID=0, Status='live', Symbol='MIR7853', Aliases='hsa-mir-7853', description='microRNA 7853', other_designations='microRNA mir-7853', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6169304, end_position_on_the_genomic_accession=6169435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927960, CurrentID=0, Status='live', Symbol='LOC101927960', Aliases='', description='uncharacterized LOC101927960', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208542842, end_position_on_the_genomic_accession=208854414, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105376649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376649, CurrentID=0, Status='live', Symbol='LOC105376649', Aliases='', description='uncharacterized LOC105376649', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44972769, end_position_on_the_genomic_accession=44973719, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102466872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466872, CurrentID=0, Status='live', Symbol='MIR7975', Aliases='hsa-mir-7975', description='microRNA 7975', other_designations='microRNA mir-7975', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55123225, end_position_on_the_genomic_accession=55123292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466873, CurrentID=0, Status='live', Symbol='MIR8052', Aliases='hsa-mir-8052', description='microRNA 8052', other_designations='microRNA mir-8052', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130666735, end_position_on_the_genomic_accession=130666803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466874, CurrentID=0, Status='live', Symbol='MIR8057', Aliases='hsa-mir-8057', description='microRNA 8057', other_designations='microRNA mir-8057', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26591467, end_position_on_the_genomic_accession=26591535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466875, CurrentID=0, Status='live', Symbol='MIR8063', Aliases='hsa-mir-8063', description='microRNA 8063', other_designations='microRNA mir-8063', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36972821, end_position_on_the_genomic_accession=36972901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466876, CurrentID=0, Status='live', Symbol='MIR8068', Aliases='hsa-mir-8068', description='microRNA 8068', other_designations='microRNA mir-8068', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28477481, end_position_on_the_genomic_accession=28477548, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466877, CurrentID=0, Status='live', Symbol='MIR8072', Aliases='hsa-mir-8072', description='microRNA 8072', other_designations='microRNA mir-8072', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123364764, end_position_on_the_genomic_accession=123364843, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466878, CurrentID=0, Status='live', Symbol='MIR8078', Aliases='hsa-mir-8078', description='microRNA 8078', other_designations='microRNA mir-8078', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=112256, end_position_on_the_genomic_accession=112339, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466879, CurrentID=0, Status='live', Symbol='MIR8083', Aliases='hsa-mir-8083', description='microRNA 8083', other_designations='microRNA mir-8083', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153689705, end_position_on_the_genomic_accession=153689793, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102466880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466880, CurrentID=0, Status='live', Symbol='MIR8088', Aliases='hsa-mir-8088', description='microRNA 8088', other_designations='microRNA mir-8088', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52336557, end_position_on_the_genomic_accession=52336642, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500801, CurrentID=0, Status='live', Symbol='MIR3659', Aliases='mir-3659', description='microRNA 3659', other_designations='hsa-mir-3659', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38089231, end_position_on_the_genomic_accession=38089329, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500802, CurrentID=0, Status='live', Symbol='MIR3685', Aliases='', description='microRNA 3685', other_designations='hsa-mir-3685', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95309923, end_position_on_the_genomic_accession=95309984, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500803, CurrentID=0, Status='live', Symbol='MIR3919', Aliases='', description='microRNA 3919', other_designations='hsa-mir-3919', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159282646, end_position_on_the_genomic_accession=159282734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500804, CurrentID=0, Status='live', Symbol='MIR3654', Aliases='', description='microRNA 3654', other_designations='hsa-mir-3654', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133034860, end_position_on_the_genomic_accession=133034915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500805, CurrentID=0, Status='live', Symbol='MIR3607', Aliases='mir-3607', description='microRNA 3607', other_designations='hsa-mir-3607', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86620497, end_position_on_the_genomic_accession=86620575, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500807, CurrentID=0, Status='live', Symbol='MIR374C', Aliases='mir-374c', description='microRNA 374c', other_designations='hsa-mir-374c', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74218549, end_position_on_the_genomic_accession=74218618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500808, CurrentID=0, Status='live', Symbol='MIR3917', Aliases='mir-3917', description='microRNA 3917', other_designations='hsa-mir-3917', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25906362, end_position_on_the_genomic_accession=25906454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500809, CurrentID=0, Status='live', Symbol='MIR23C', Aliases='hsa-mir-23c, mir-23c', description='microRNA 23c', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20017088, end_position_on_the_genomic_accession=20017187, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500810, CurrentID=0, Status='live', Symbol='MIR3620', Aliases='mir-3620', description='microRNA 3620', other_designations='hsa-mir-3620', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228097263, end_position_on_the_genomic_accession=228097341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500811, CurrentID=0, Status='live', Symbol='MIR3621', Aliases='', description='microRNA 3621', other_designations='hsa-mir-3621', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137169186, end_position_on_the_genomic_accession=137169270, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500812, CurrentID=0, Status='live', Symbol='MIR3677', Aliases='mir-3677', description='microRNA 3677', other_designations='hsa-mir-3677', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2270713, end_position_on_the_genomic_accession=2270772, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500813, CurrentID=0, Status='live', Symbol='MIR3646', Aliases='', description='microRNA 3646', other_designations='hsa-mir-3646', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44408120, end_position_on_the_genomic_accession=44408203, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500814, CurrentID=0, Status='live', Symbol='MIR3616', Aliases='mir-3616', description='microRNA 3616', other_designations='hsa-mir-3616', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47166967, end_position_on_the_genomic_accession=47167058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500815, CurrentID=0, Status='live', Symbol='MIR3687-1', Aliases='MIR3687, hsa-mir-3687-1', description='microRNA 3687-1', other_designations='hsa-mir-3687|microRNA 3687', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8208844, end_position_on_the_genomic_accession=8208904, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500816, CurrentID=0, Status='live', Symbol='MIR3649', Aliases='', description='microRNA 3649', other_designations='hsa-mir-3649', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1660315, end_position_on_the_genomic_accession=1660380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500817, CurrentID=0, Status='live', Symbol='MIR3612', Aliases='', description='microRNA 3612', other_designations='hsa-mir-3612', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128294092, end_position_on_the_genomic_accession=128294178, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500818, CurrentID=0, Status='live', Symbol='MIR3945', Aliases='', description='microRNA 3945', other_designations='hsa-mir-3945', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184851013, end_position_on_the_genomic_accession=184851110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500819, CurrentID=0, Status='live', Symbol='MIR3609', Aliases='mir-3609', description='microRNA 3609', other_designations='hsa-mir-3609', map_location='7q21.2-q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98881650, end_position_on_the_genomic_accession=98881729, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500820, CurrentID=0, Status='live', Symbol='MIR3655', Aliases='', description='microRNA 3655', other_designations='hsa-mir-3655', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140647844, end_position_on_the_genomic_accession=140647926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500821, CurrentID=0, Status='live', Symbol='MIR3910-1', Aliases='mir-3910-1', description='microRNA 3910-1', other_designations='hsa-mir-3910-1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91636251, end_position_on_the_genomic_accession=91636361, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500822, CurrentID=0, Status='live', Symbol='MIR3937', Aliases='', description='microRNA 3937', other_designations='hsa-mir-3937', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39661216, end_position_on_the_genomic_accession=39661321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500823, CurrentID=0, Status='live', Symbol='MIR3920', Aliases='', description='microRNA 3920', other_designations='hsa-mir-3920', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101519820, end_position_on_the_genomic_accession=101519905, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500824, CurrentID=0, Status='live', Symbol='MIR3650', Aliases='', description='microRNA 3650', other_designations='hsa-mir-3650', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38557502, end_position_on_the_genomic_accession=38557561, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500825, CurrentID=0, Status='live', Symbol='MIR3660', Aliases='mir-3660', description='microRNA 3660', other_designations='hsa-mir-3660', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90016621, end_position_on_the_genomic_accession=90016720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500826, CurrentID=0, Status='live', Symbol='MIR3909', Aliases='mir-3909', description='microRNA 3909', other_designations='hsa-mir-3909', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35335640, end_position_on_the_genomic_accession=35335758, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500827, CurrentID=0, Status='live', Symbol='MIR3614', Aliases='mir-3614', description='microRNA 3614', other_designations='hsa-mir-3614', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56891270, end_position_on_the_genomic_accession=56891355, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500828, CurrentID=0, Status='live', Symbol='MIR3619', Aliases='', description='microRNA 3619', other_designations='hsa-mir-3619', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46091044, end_position_on_the_genomic_accession=46091126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500829, CurrentID=0, Status='live', Symbol='MIR3943', Aliases='', description='microRNA 3943', other_designations='hsa-mir-3943', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43150895, end_position_on_the_genomic_accession=43150994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500830, CurrentID=0, Status='live', Symbol='MIR550B2', Aliases='', description='microRNA 550b-2', other_designations='hsa-mir-550b-2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32732981, end_position_on_the_genomic_accession=32733077, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500831, CurrentID=0, Status='live', Symbol='MIR3912', Aliases='mir-3912', description='microRNA 3912', other_designations='hsa-mir-3912', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171386656, end_position_on_the_genomic_accession=171386760, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500832, CurrentID=0, Status='live', Symbol='MIR3658', Aliases='', description='microRNA 3658', other_designations='hsa-mir-3658', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165907921, end_position_on_the_genomic_accession=165907976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500833, CurrentID=0, Status='live', Symbol='MIR3653', Aliases='mir-3653', description='microRNA 3653', other_designations='hsa-mir-3653', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29333158, end_position_on_the_genomic_accession=29333267, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500834, CurrentID=0, Status='live', Symbol='MIR3924', Aliases='', description='microRNA 3924', other_designations='hsa-mir-3924', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=57304479, end_position_on_the_genomic_accession=57304559, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500835, CurrentID=0, Status='live', Symbol='MIR3907', Aliases='', description='microRNA 3907', other_designations='hsa-mir-3907', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151433489, end_position_on_the_genomic_accession=151433639, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500836, CurrentID=0, Status='live', Symbol='MIR3914-1', Aliases='', description='microRNA 3914-1', other_designations='hsa-mir-3914-1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71307672, end_position_on_the_genomic_accession=71307770, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500837, CurrentID=0, Status='live', Symbol='MIR3606', Aliases='mir-3606', description='microRNA 3606', other_designations='hsa-mir-3606', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188995630, end_position_on_the_genomic_accession=188995692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500838, CurrentID=0, Status='live', Symbol='MIR3926-2', Aliases='mir-3926-2', description='microRNA 3926-2', other_designations='hsa-mir-3926-2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12727237, end_position_on_the_genomic_accession=12727299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500839, CurrentID=0, Status='live', Symbol='MIR3686', Aliases='', description='microRNA 3686', other_designations='hsa-mir-3686', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129484057, end_position_on_the_genomic_accession=129484142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500840, CurrentID=0, Status='live', Symbol='MIR3656', Aliases='mir-3656', description='microRNA 3656', other_designations='hsa-mir-3656', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119018944, end_position_on_the_genomic_accession=119019012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500841, CurrentID=0, Status='live', Symbol='MIR3678', Aliases='mir-3678', description='microRNA 3678', other_designations='hsa-mir-3678', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75406069, end_position_on_the_genomic_accession=75406162, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500842, CurrentID=0, Status='live', Symbol='MIR3652', Aliases='', description='microRNA 3652', other_designations='hsa-mir-3652', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103930425, end_position_on_the_genomic_accession=103930555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500843, CurrentID=0, Status='live', Symbol='MIR3922', Aliases='', description='microRNA 3922', other_designations='hsa-mir-3922', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104591633, end_position_on_the_genomic_accession=104591716, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500844, CurrentID=0, Status='live', Symbol='MIR3664', Aliases='mir-3664', description='microRNA 3664', other_designations='hsa-mir-3664', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70872270, end_position_on_the_genomic_accession=70872368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500845, CurrentID=0, Status='live', Symbol='MIR642B', Aliases='mir-642b', description='microRNA 642b', other_designations='hsa-mir-642b', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45674932, end_position_on_the_genomic_accession=45675008, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500846, CurrentID=0, Status='live', Symbol='MIR3689A', Aliases='', description='microRNA 3689a', other_designations='hsa-mir-3689a', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134849487, end_position_on_the_genomic_accession=134849564, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500847, CurrentID=0, Status='live', Symbol='MIR3615', Aliases='mir-3615', description='microRNA 3615', other_designations='hsa-mir-3615', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74748613, end_position_on_the_genomic_accession=74748699, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500849, CurrentID=0, Status='live', Symbol='MIR3916', Aliases='mir-3916', description='microRNA 3916', other_designations='hsa-mir-3916', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247201967, end_position_on_the_genomic_accession=247202060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500850, CurrentID=0, Status='live', Symbol='MIR3682', Aliases='mir-3682', description='microRNA 3682', other_designations='hsa-mir-3682', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53849122, end_position_on_the_genomic_accession=53849205, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500851, CurrentID=0, Status='live', Symbol='MIR3918', Aliases='mir-3918', description='microRNA 3918', other_designations='hsa-mir-3918', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158764661, end_position_on_the_genomic_accession=158764753, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500852, CurrentID=0, Status='live', Symbol='MIR3180-4', Aliases='mir-3180-4', description='microRNA 3180-4', other_designations='hsa-mir-3180-4', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15154850, end_position_on_the_genomic_accession=15155002, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500853, CurrentID=0, Status='live', Symbol='MIR3605', Aliases='mir-3605', description='microRNA 3605', other_designations='hsa-mir-3605', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33332393, end_position_on_the_genomic_accession=33332492, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500854, CurrentID=0, Status='live', Symbol='MIR3671', Aliases='', description='microRNA 3671', other_designations='hsa-mir-3671', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65057755, end_position_on_the_genomic_accession=65057842, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500855, CurrentID=0, Status='live', Symbol='MIR3713', Aliases='', description='microRNA 3713', other_designations='hsa-mir-3713', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76586647, end_position_on_the_genomic_accession=76586691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500856, CurrentID=0, Status='live', Symbol='MIR548Z', Aliases='', description='microRNA 548z', other_designations='hsa-mir-548z', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64622509, end_position_on_the_genomic_accession=64622605, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500857, CurrentID=0, Status='live', Symbol='MIR3939', Aliases='mir-3939', description='microRNA 3939', other_designations='hsa-mir-3939', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166997807, end_position_on_the_genomic_accession=166997912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500858, CurrentID=0, Status='live', Symbol='MIR3622A', Aliases='mir-3622a', description='microRNA 3622a', other_designations='hsa-mir-3622a', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27701677, end_position_on_the_genomic_accession=27701759, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500859, CurrentID=0, Status='live', Symbol='MIR3921', Aliases='', description='microRNA 3921', other_designations='hsa-mir-3921', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99964314, end_position_on_the_genomic_accession=99964398, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500860, CurrentID=0, Status='live', Symbol='MIR3618', Aliases='', description='microRNA 3618', other_designations='hsa-mir-3618', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20085746, end_position_on_the_genomic_accession=20085833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500861, CurrentID=0, Status='live', Symbol='MIR3665', Aliases='', description='microRNA 3665', other_designations='hsa-mir-3665', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77698012, end_position_on_the_genomic_accession=77698116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500862, CurrentID=0, Status='live', Symbol='MIR3648-1', Aliases='MIR3648, hsa-mir-3648-1', description='microRNA 3648-1', other_designations='hsa-mir-3648|microRNA 3648', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8208473, end_position_on_the_genomic_accession=8208652, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500863, CurrentID=0, Status='live', Symbol='MIR548AA1', Aliases='', description='microRNA 548aa-1', other_designations='hsa-mir-548aa-1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123348034, end_position_on_the_genomic_accession=123348130, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500864, CurrentID=0, Status='live', Symbol='MIR3929', Aliases='', description='microRNA 3929', other_designations='hsa-mir-3929', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35934088, end_position_on_the_genomic_accession=35934142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500865, CurrentID=0, Status='live', Symbol='MIR3936', Aliases='mir-3936', description='microRNA 3936', other_designations='hsa-mir-3936', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132365490, end_position_on_the_genomic_accession=132365599, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500866, CurrentID=0, Status='live', Symbol='MIR3941', Aliases='', description='microRNA 3941', other_designations='hsa-mir-3941', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122416965, end_position_on_the_genomic_accession=122417067, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500867, CurrentID=0, Status='live', Symbol='MIR3684', Aliases='', description='microRNA 3684', other_designations='hsa-mir-3684', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98997387, end_position_on_the_genomic_accession=98997460, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500868, CurrentID=0, Status='live', Symbol='MIR3913-2', Aliases='mir-3913-2', description='microRNA 3913-2', other_designations='hsa-mir-3913-2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69584723, end_position_on_the_genomic_accession=69584822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500869, CurrentID=0, Status='live', Symbol='MIR3672', Aliases='', description='microRNA 3672', other_designations='hsa-mir-3672', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121370972, end_position_on_the_genomic_accession=121371053, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500870, CurrentID=0, Status='live', Symbol='MIR3926-1', Aliases='mir-3926-1', description='microRNA 3926-1', other_designations='hsa-mir-3926-1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12727232, end_position_on_the_genomic_accession=12727304, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500871, CurrentID=0, Status='live', Symbol='MIR3622B', Aliases='', description='microRNA 3622b', other_designations='hsa-mir-3622b', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27701673, end_position_on_the_genomic_accession=27701767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500872, CurrentID=0, Status='live', Symbol='MIR3911', Aliases='mir-3911', description='microRNA 3911', other_designations='hsa-mir-3911', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127690687, end_position_on_the_genomic_accession=127690795, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500873, CurrentID=0, Status='live', Symbol='MIR3934', Aliases='', description='microRNA 3934', other_designations='hsa-mir-3934', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33698128, end_position_on_the_genomic_accession=33698234, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500875, CurrentID=0, Status='live', Symbol='MIR3938', Aliases='', description='microRNA 3938', other_designations='hsa-mir-3938', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55852492, end_position_on_the_genomic_accession=55852594, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500876, CurrentID=0, Status='live', Symbol='MIR3675', Aliases='', description='microRNA 3675', other_designations='hsa-mir-3675', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16858949, end_position_on_the_genomic_accession=16859021, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500877, CurrentID=0, Status='live', Symbol='MIR3923', Aliases='', description='microRNA 3923', other_designations='hsa-mir-3923', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=79507887, end_position_on_the_genomic_accession=79507969, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500878, CurrentID=0, Status='live', Symbol='MIR3679', Aliases='mir-3679', description='microRNA 3679', other_designations='hsa-mir-3679', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134127125, end_position_on_the_genomic_accession=134127192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500879, CurrentID=0, Status='live', Symbol='MIR3668', Aliases='', description='microRNA 3668', other_designations='hsa-mir-3668', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140205252, end_position_on_the_genomic_accession=140205326, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500880, CurrentID=0, Status='live', Symbol='MIR3662', Aliases='mir-3662', description='microRNA 3662', other_designations='hsa-mir-3662', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134979338, end_position_on_the_genomic_accession=134979432, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500881, CurrentID=0, Status='live', Symbol='MIR3688-1', Aliases='MIR3688, mir-3688-1', description='microRNA 3688-1', other_designations='hsa-mir-3688|hsa-mir-3688-1|microRNA 3688', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159128802, end_position_on_the_genomic_accession=159128894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500882, CurrentID=0, Status='live', Symbol='MIR3667', Aliases='', description='microRNA 3667', other_designations='hsa-mir-3667', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49543393, end_position_on_the_genomic_accession=49543466, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500883, CurrentID=0, Status='live', Symbol='MIR550B1', Aliases='', description='microRNA 550b-1', other_designations='hsa-mir-550b-1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30289794, end_position_on_the_genomic_accession=30289890, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500884, CurrentID=0, Status='live', Symbol='MIR3681', Aliases='', description='microRNA 3681', other_designations='hsa-mir-3681', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12199130, end_position_on_the_genomic_accession=12199201, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500885, CurrentID=0, Status='live', Symbol='MIR3925', Aliases='', description='microRNA 3925', other_designations='hsa-mir-3925', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36622436, end_position_on_the_genomic_accession=36622512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500886, CurrentID=0, Status='live', Symbol='MIR3683', Aliases='', description='microRNA 3683', other_designations='hsa-mir-3683', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7066964, end_position_on_the_genomic_accession=7067045, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500887, CurrentID=0, Status='live', Symbol='MIR676', Aliases='hsa-mir-676, mir-676', description='microRNA 676', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70022857, end_position_on_the_genomic_accession=70022923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500888, CurrentID=0, Status='live', Symbol='MIR3940', Aliases='mir-3940', description='microRNA 3940', other_designations='hsa-mir-3940', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6416410, end_position_on_the_genomic_accession=6416511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500889, CurrentID=0, Status='live', Symbol='MIR3657', Aliases='', description='microRNA 3657', other_designations='hsa-mir-3657', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112037599, end_position_on_the_genomic_accession=112037715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500890, CurrentID=0, Status='live', Symbol='MIR3611', Aliases='', description='microRNA 3611', other_designations='hsa-mir-3611', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35079598, end_position_on_the_genomic_accession=35079680, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500891, CurrentID=0, Status='live', Symbol='MIR3935', Aliases='', description='microRNA 3935', other_designations='hsa-mir-3935', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56245520, end_position_on_the_genomic_accession=56245623, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102466972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102466972, CurrentID=0, Status='live', Symbol='MIR6508', Aliases='hsa-mir-6508', description='microRNA 6508', other_designations='microRNA mir-6508', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39447010, end_position_on_the_genomic_accession=39447069, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500893, CurrentID=0, Status='live', Symbol='MIR3663', Aliases='', description='microRNA 3663', other_designations='hsa-mir-3663', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117167678, end_position_on_the_genomic_accession=117167774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500894, CurrentID=0, Status='live', Symbol='MIR3690', Aliases='MIR3690-1-2, hsa-mir-3690-1, hsa-mir-3690-2, mir-3690-1, mir-3690-2, MIR3690', description='microRNA 3690', other_designations='hsa-mir-3690|microRNA 3690-1|microRNA 3690-2', map_location='Xp22.32 and Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1293918, end_position_on_the_genomic_accession=1293992, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500895, CurrentID=0, Status='live', Symbol='MIR548AA2', Aliases='', description='microRNA 548aa-2', other_designations='hsa-mir-548aa-2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67471489, end_position_on_the_genomic_accession=67471585, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500896, CurrentID=0, Status='live', Symbol='MIR3666', Aliases='', description='microRNA 3666', other_designations='hsa-mir-3666', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=114653345, end_position_on_the_genomic_accession=114653455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500897, CurrentID=0, Status='live', Symbol='MIR3617', Aliases='', description='microRNA 3617', other_designations='hsa-mir-3617', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45705102, end_position_on_the_genomic_accession=45705180, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500898, CurrentID=0, Status='live', Symbol='MIR3927', Aliases='', description='microRNA 3927', other_designations='hsa-mir-3927', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109511475, end_position_on_the_genomic_accession=109511545, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500899, CurrentID=0, Status='live', Symbol='MIR3692', Aliases='', description='microRNA 3692', other_designations='hsa-mir-3692', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157529132, end_position_on_the_genomic_accession=157529200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500900, CurrentID=0, Status='live', Symbol='MIR3691', Aliases='mir-3691', description='microRNA 3691', other_designations='hsa-mir-3691', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5148233, end_position_on_the_genomic_accession=5148322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500901, CurrentID=0, Status='live', Symbol='MIR3928', Aliases='mir-3928', description='microRNA 3928', other_designations='hsa-mir-3928', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31160062, end_position_on_the_genomic_accession=31160119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500902, CurrentID=0, Status='live', Symbol='MIR3910-2', Aliases='mir-3910-2', description='microRNA 3910-2', other_designations='hsa-mir-3910-2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91636264, end_position_on_the_genomic_accession=91636345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500903, CurrentID=0, Status='live', Symbol='MIR3913-1', Aliases='mir-3913-1', description='microRNA 3913-1', other_designations='hsa-mir-3913-1', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69584722, end_position_on_the_genomic_accession=69584823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500904, CurrentID=0, Status='live', Symbol='MIR3942', Aliases='', description='microRNA 3942', other_designations='hsa-mir-3942', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35372256, end_position_on_the_genomic_accession=35372364, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500905, CurrentID=0, Status='live', Symbol='MIR3661', Aliases='mir-3661', description='microRNA 3661', other_designations='hsa-mir-3661', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134225757, end_position_on_the_genomic_accession=134225852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500906, CurrentID=0, Status='live', Symbol='MIR3689B', Aliases='mir-3689b', description='microRNA 3689b', other_designations='hsa-mir-3689b', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134850125, end_position_on_the_genomic_accession=134850272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500907, CurrentID=0, Status='live', Symbol='MIR3150B', Aliases='mir-3150b', description='microRNA 3150b', other_designations='hsa-mir-3150b', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95072911, end_position_on_the_genomic_accession=95072996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500908, CurrentID=0, Status='live', Symbol='MIR3613', Aliases='', description='microRNA 3613', other_designations='hsa-mir-3613', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49996415, end_position_on_the_genomic_accession=49996501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500909, CurrentID=0, Status='live', Symbol='MIR3908', Aliases='', description='microRNA 3908', other_designations='hsa-mir-3908', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123536409, end_position_on_the_genomic_accession=123536534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500910, CurrentID=0, Status='live', Symbol='MIR3670-1', Aliases='MIR3670, hsa-mir-3670-1', description='microRNA 3670-1', other_designations='hsa-mir-3670|microRNA 3670', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14907717, end_position_on_the_genomic_accession=14907781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500911, CurrentID=0, Status='live', Symbol='MIR3944', Aliases='', description='microRNA 3944', other_designations='hsa-mir-3944', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133371556, end_position_on_the_genomic_accession=133371663, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500912, CurrentID=0, Status='live', Symbol='MIR3674', Aliases='', description='microRNA 3674', other_designations='hsa-mir-3674', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1801125, end_position_on_the_genomic_accession=1801192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500913, CurrentID=0, Status='live', Symbol='MIR3714', Aliases='', description='microRNA 3714', other_designations='hsa-mir-3714', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16933196, end_position_on_the_genomic_accession=16933260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100500914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500914, CurrentID=0, Status='live', Symbol='MIR3610', Aliases='mir-3610', description='microRNA 3610', other_designations='hsa-mir-3610', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116874728, end_position_on_the_genomic_accession=116874800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500915, CurrentID=0, Status='live', Symbol='MIR3915', Aliases='', description='microRNA 3915', other_designations='hsa-mir-3915', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=32583656, end_position_on_the_genomic_accession=32583752, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500916, CurrentID=0, Status='live', Symbol='MIR3180-5', Aliases='mir-3180-5', description='microRNA 3180-5', other_designations='hsa-mir-3180-5', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2135977, end_position_on_the_genomic_accession=2136129, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500917, CurrentID=0, Status='live', Symbol='MIR3680-1', Aliases='MIR3680, hsa-mir-3680-1', description='microRNA 3680-1', other_designations='hsa-mir-3680|microRNA 3680', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21506049, end_position_on_the_genomic_accession=21506135, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500918, CurrentID=0, Status='live', Symbol='MIR3651', Aliases='mir-3651', description='microRNA 3651', other_designations='hsa-mir-3651', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92292458, end_position_on_the_genomic_accession=92292547, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500919, CurrentID=0, Status='live', Symbol='MIR548Y', Aliases='', description='microRNA 548y', other_designations='hsa-mir-548y', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=47760995, end_position_on_the_genomic_accession=47761104, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100500920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500920, CurrentID=0, Status='live', Symbol='MIR3914-2', Aliases='', description='microRNA 3914-2', other_designations='hsa-mir-3914-2', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=71307674, end_position_on_the_genomic_accession=71307768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102467003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467003, CurrentID=0, Status='live', Symbol='MIR7851', Aliases='hsa-mir-7851', description='microRNA 7851', other_designations='microRNA mir-7851', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42323700, end_position_on_the_genomic_accession=42323859, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102467004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467004, CurrentID=0, Status='live', Symbol='MIR8053', Aliases='hsa-mir-8053', description='microRNA 8053', other_designations='microRNA mir-8053', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47652669, end_position_on_the_genomic_accession=47652743, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102467005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467005, CurrentID=0, Status='live', Symbol='MIR8084', Aliases='hsa-mir-8084', description='microRNA 8084', other_designations='microRNA mir-8084', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93029751, end_position_on_the_genomic_accession=93029839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376421, CurrentID=0, Status='live', Symbol='LOC105376421', Aliases='', description='uncharacterized LOC105376421', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13528604, end_position_on_the_genomic_accession=13530264, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100500934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500934, CurrentID=0, Status='live', Symbol='LOC100500934', Aliases='', description='ring-box 1, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='5q14', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80019603, end_position_on_the_genomic_accession=80020777, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376673, CurrentID=0, Status='live', Symbol='LOC105376673', Aliases='', description='uncharacterized LOC105376673', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3995104, end_position_on_the_genomic_accession=3998855, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100500938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100500938, CurrentID=0, Status='live', Symbol='CCDC179', Aliases='', description='coiled-coil domain containing 179', other_designations='', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22846922, end_position_on_the_genomic_accession=22860486, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375936, CurrentID=0, Status='live', Symbol='LOC105375936', Aliases='', description='uncharacterized LOC105375936', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85221680, end_position_on_the_genomic_accession=85243485, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376677, CurrentID=0, Status='live', Symbol='LOC105376677', Aliases='', description='uncharacterized LOC105376677', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34028664, end_position_on_the_genomic_accession=34035296, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100421513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421513, CurrentID=0, Status='live', Symbol='LOC100421513', Aliases='', description='transducin (beta)-like 1 X-linked receptor 1 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127659693, end_position_on_the_genomic_accession=127660579, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376679, CurrentID=0, Status='live', Symbol='LOC105376679', Aliases='', description='uncharacterized LOC105376679', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4571514, end_position_on_the_genomic_accession=4594998, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  104564224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104564224, CurrentID=0, Status='live', Symbol='LUNAR1', Aliases='', description='leukemia-associated non-coding IGF1R activator RNA 1', other_designations='leukemia-induced noncoding activator RNA', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99014704, end_position_on_the_genomic_accession=99031050, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104564225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104564225, CurrentID=0, Status='live', Symbol='MHRT', Aliases='Myheart', description='myosin heavy chain-associated RNA transcript', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23415450, end_position_on_the_genomic_accession=23417595, orientation='plus', exon_count=6, OMIM=616096, no_hdr0=''),
  102467074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467074, CurrentID=0, Status='live', Symbol='LINC01377', Aliases='', description='long intergenic non-protein coding RNA 1377', other_designations='CTD-2029E14.1', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3177835, end_position_on_the_genomic_accession=3181232, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102467075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467075, CurrentID=0, Status='live', Symbol='LINC01017', Aliases='', description='long intergenic non-protein coding RNA 1017', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3496372, end_position_on_the_genomic_accession=3504004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102467077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467077, CurrentID=0, Status='live', Symbol='LINC00603', Aliases='', description='long intergenic non-protein coding RNA 603', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40052291, end_position_on_the_genomic_accession=40053324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102467079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467079, CurrentID=0, Status='live', Symbol='LOC102467079', Aliases='', description='uncharacterized LOC102467079', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52085281, end_position_on_the_genomic_accession=52099067, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102467080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467080, CurrentID=0, Status='live', Symbol='LOC102467080', Aliases='', description='uncharacterized LOC102467080', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54660158, end_position_on_the_genomic_accession=54744258, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102467081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467081, CurrentID=0, Status='live', Symbol='LOC102467081', Aliases='', description='uncharacterized LOC102467081', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55021299, end_position_on_the_genomic_accession=55024169, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643118, CurrentID=0, Status='live', Symbol='LOC643118', Aliases='', description='serine/arginine-rich splicing factor 2 pseudogene', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95037516, end_position_on_the_genomic_accession=95038162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376685, CurrentID=0, Status='live', Symbol='LOC105376685', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94712307, end_position_on_the_genomic_accession=94713712, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129785, CurrentID=0, Status='live', Symbol='LOC100129785', Aliases='', description='uncharacterized LOC100129785', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=131920872, end_position_on_the_genomic_accession=131921691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376687, CurrentID=0, Status='live', Symbol='LOC105376687', Aliases='', description='uncharacterized LOC105376687', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6131763, end_position_on_the_genomic_accession=6132692, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  344328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=344328, CurrentID=0, Status='live', Symbol='ST13P2', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 2', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187825341, end_position_on_the_genomic_accession=187826446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271375, CurrentID=0, Status='live', Symbol='RPS27AP8', Aliases='RPS27A_2_437', description='ribosomal protein S27a pseudogene 8', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172225428, end_position_on_the_genomic_accession=172225869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376691, CurrentID=0, Status='live', Symbol='LOC105376691', Aliases='', description='uncharacterized LOC105376691', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7503177, end_position_on_the_genomic_accession=7509205, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  79530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79530, CurrentID=0, Status='live', Symbol='OR4C9P', Aliases='', description='olfactory receptor family 4 subfamily C member 9 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 9 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48464053, end_position_on_the_genomic_accession=48464977, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256949, CurrentID=0, Status='live', Symbol='KANK3', Aliases='ANKRD47', description='KN motif and ankyrin repeat domains 3', other_designations='ankyrin repeat domain 47|ankyrin repeat domain-containing protein 47|kidney ankyrin repeat-containing protein 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8322582, end_position_on_the_genomic_accession=8343262, orientation='minus', exon_count=14, OMIM=614611, no_hdr0=''),
  101927969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927969, CurrentID=0, Status='live', Symbol='LOC101927969', Aliases='', description='uncharacterized LOC101927969', other_designations='CTC-558O2.1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168706927, end_position_on_the_genomic_accession=168720884, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102467146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467146, CurrentID=0, Status='live', Symbol='MAFTRR', Aliases='linc-MAF-4, lincMAF4', description='MAF transcriptional regulator RNA', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79721312, end_position_on_the_genomic_accession=79770532, orientation='minus', exon_count=6, OMIM=616264, no_hdr0=''),
  102467147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467147, CurrentID=0, Status='live', Symbol='LOC102467147', Aliases='', description='uncharacterized LOC102467147', other_designations='CTC-236F12.4', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56457795, end_position_on_the_genomic_accession=56481769, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  400499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400499, CurrentID=0, Status='live', Symbol='LOC400499', Aliases='', description='uncharacterized LOC400499', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11372601, end_position_on_the_genomic_accession=11523533, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  105376700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376700, CurrentID=0, Status='live', Symbol='LOC105376700', Aliases='', description='uncharacterized LOC105376700', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24158946, end_position_on_the_genomic_accession=24164138, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102467212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467212, CurrentID=0, Status='live', Symbol='LOC102467212', Aliases='', description='uncharacterized LOC102467212', other_designations='CTD-2193G5.1', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103528434, end_position_on_the_genomic_accession=103541985, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102467213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467213, CurrentID=0, Status='live', Symbol='LOC102467213', Aliases='', description='uncharacterized LOC102467213', other_designations='CTC-254B4.1', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=106815197, end_position_on_the_genomic_accession=107011014, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102467214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467214, CurrentID=0, Status='live', Symbol='LOC102467214', Aliases='', description='uncharacterized LOC102467214', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112628436, end_position_on_the_genomic_accession=112630921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102467216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467216, CurrentID=0, Status='live', Symbol='LOC102467216', Aliases='', description='uncharacterized LOC102467216', other_designations='', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112656428, end_position_on_the_genomic_accession=112682888, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102467217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467217, CurrentID=0, Status='live', Symbol='LOC102467217', Aliases='', description='uncharacterized LOC102467217', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115642125, end_position_on_the_genomic_accession=115672652, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102467222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467222, CurrentID=0, Status='live', Symbol='LOC102467222', Aliases='', description='uncharacterized LOC102467222', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20953633, end_position_on_the_genomic_accession=20968938, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102467223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467223, CurrentID=0, Status='live', Symbol='LOC102467223', Aliases='', description='uncharacterized LOC102467223', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116743302, end_position_on_the_genomic_accession=116762209, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102467224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467224, CurrentID=0, Status='live', Symbol='LOC102467224', Aliases='', description='uncharacterized LOC102467224', other_designations='CTD-3179P9.1', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117730361, end_position_on_the_genomic_accession=118266062, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102467225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467225, CurrentID=0, Status='live', Symbol='LOC102467225', Aliases='', description='uncharacterized LOC102467225', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118596188, end_position_on_the_genomic_accession=118628092, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102467226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467226, CurrentID=0, Status='live', Symbol='LOC102467226', Aliases='', description='uncharacterized LOC102467226', other_designations='CTC-546K23.1', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121322550, end_position_on_the_genomic_accession=121325837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376709, CurrentID=0, Status='live', Symbol='LOC105376709', Aliases='', description='uncharacterized LOC105376709', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32356371, end_position_on_the_genomic_accession=32358631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060476, CurrentID=0, Status='live', Symbol='LOC101060476', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480168, CurrentID=0, Status='live', Symbol='RNU1-64P', Aliases='', description='RNA, U1 small nuclear 64, pseudogene', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376723, CurrentID=0, Status='live', Symbol='LOC105376723', Aliases='', description='uncharacterized LOC105376723', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84235491, end_position_on_the_genomic_accession=84236712, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326307, CurrentID=0, Status='live', Symbol='RPL3P4', Aliases='', description='ribosomal protein L3 pseudogene 4', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98972823, end_position_on_the_genomic_accession=98973301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652833, CurrentID=0, Status='live', Symbol='LOC100652833', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=99522014, end_position_on_the_genomic_accession=99526049, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376727, CurrentID=0, Status='live', Symbol='LOC105376727', Aliases='', description='uncharacterized LOC105376727', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85336832, end_position_on_the_genomic_accession=85359284, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376731, CurrentID=0, Status='live', Symbol='LOC105376731', Aliases='', description='taste receptor cell protein 1-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75225896, end_position_on_the_genomic_accession=75232485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376735, CurrentID=0, Status='live', Symbol='LOC105376735', Aliases='', description='uncharacterized LOC105376735', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133857, CurrentID=0, Status='live', Symbol='LOC100133857', Aliases='', description='uncharacterized LOC100133857', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271377, CurrentID=0, Status='live', Symbol='RPS27AP10', Aliases='RPS27A_3_621', description='ribosomal protein S27a pseudogene 10', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141588938, end_position_on_the_genomic_accession=141589396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376741, CurrentID=0, Status='live', Symbol='LOC105376741', Aliases='', description='uncharacterized LOC105376741', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11485876, end_position_on_the_genomic_accession=11499470, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376744, CurrentID=0, Status='live', Symbol='LOC105376744', Aliases='', description='uncharacterized LOC105376744', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12525280, end_position_on_the_genomic_accession=12528868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  414760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414760, CurrentID=0, Status='live', Symbol='HCG14', Aliases='dJ111M5.4', description='HLA complex group 14 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28896530, end_position_on_the_genomic_accession=28897320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100873374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873374, CurrentID=0, Status='live', Symbol='RNA5SP103', Aliases='RN5S103', description='RNA, 5S ribosomal pseudogene 103', other_designations='RNA, 5S ribosomal 103', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128445081, end_position_on_the_genomic_accession=128445198, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376747, CurrentID=0, Status='live', Symbol='LOC105376747', Aliases='', description='collagen alpha-3(IX) chain-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14734887, end_position_on_the_genomic_accession=14742760, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376748, CurrentID=0, Status='live', Symbol='LOC105376748', Aliases='', description='uncharacterized LOC105376748', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376749, CurrentID=0, Status='live', Symbol='LOC105376749', Aliases='', description='uncharacterized LOC105376749', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14748192, end_position_on_the_genomic_accession=14769985, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376753, CurrentID=0, Status='live', Symbol='LOC105376753', Aliases='', description='uncharacterized LOC105376753', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18313388, end_position_on_the_genomic_accession=18331611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376754, CurrentID=0, Status='live', Symbol='LOC105376754', Aliases='', description='uncharacterized LOC105376754', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376763, CurrentID=0, Status='live', Symbol='LOC105376763', Aliases='', description='uncharacterized LOC105376763', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644925, CurrentID=0, Status='live', Symbol='LOC644925', Aliases='', description='uncharacterized LOC644925', other_designations='', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=55032836, end_position_on_the_genomic_accession=55034393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375954, CurrentID=0, Status='live', Symbol='LOC105375954', Aliases='', description='uncharacterized LOC105375954', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2243775, end_position_on_the_genomic_accession=2246216, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376764, CurrentID=0, Status='live', Symbol='LOC105376764', Aliases='', description='uncharacterized LOC105376764', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14278127, end_position_on_the_genomic_accession=14304452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  400512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400512, CurrentID=0, Status='live', Symbol='FLJ21408', Aliases='', description='uncharacterized LOC400512', other_designations='CTD-3203P2.2', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27268205, end_position_on_the_genomic_accession=27290468, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376766, CurrentID=0, Status='live', Symbol='LOC105376766', Aliases='', description='uncharacterized LOC105376766', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376767, CurrentID=0, Status='live', Symbol='LOC105376767', Aliases='', description='uncharacterized LOC105376767', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15461665, end_position_on_the_genomic_accession=15464159, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326316, CurrentID=0, Status='live', Symbol='RPS15AP2', Aliases='RPS15A_11_1364', description='ribosomal protein S15a pseudogene 2', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50053236, end_position_on_the_genomic_accession=50053597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376769, CurrentID=0, Status='live', Symbol='LOC105376769', Aliases='', description='uncharacterized LOC105376769', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418898, CurrentID=0, Status='live', Symbol='LOC100418898', Aliases='', description='transportin 3 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50897618, end_position_on_the_genomic_accession=50898246, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371591, CurrentID=0, Status='live', Symbol='LOC105371591', Aliases='', description='uncharacterized LOC105371591', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2706477, end_position_on_the_genomic_accession=2708492, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376774, CurrentID=0, Status='live', Symbol='LOC105376774', Aliases='', description='uncharacterized LOC105376774', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76771020, end_position_on_the_genomic_accession=76772931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376775, CurrentID=0, Status='live', Symbol='LOC105376775', Aliases='', description='uncharacterized LOC105376775', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77462910, end_position_on_the_genomic_accession=77516073, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388104, CurrentID=0, Status='live', Symbol='LOC388104', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30881919, end_position_on_the_genomic_accession=30883290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371592, CurrentID=0, Status='live', Symbol='LOC105371592', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2708663, end_position_on_the_genomic_accession=2713146, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129096, CurrentID=0, Status='live', Symbol='LOC100129096', Aliases='', description='tropomyosin 1 (alpha) pseudogene', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69129711, end_position_on_the_genomic_accession=69130336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421517, CurrentID=0, Status='live', Symbol='LOC100421517', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C17 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 17 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=47597151, end_position_on_the_genomic_accession=47597943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376778, CurrentID=0, Status='live', Symbol='LOC105376778', Aliases='', description='uncharacterized LOC105376778', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86049455, end_position_on_the_genomic_accession=86060206, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376779, CurrentID=0, Status='live', Symbol='LOC105376779', Aliases='', description='uncharacterized LOC105376779', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88386927, end_position_on_the_genomic_accession=88391933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102467655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102467655, CurrentID=0, Status='live', Symbol='LOC102467655', Aliases='', description='uncharacterized LOC102467655', other_designations='', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68189876, end_position_on_the_genomic_accession=68198407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376781, CurrentID=0, Status='live', Symbol='LOC105376781', Aliases='', description='uncharacterized LOC105376781', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90158128, end_position_on_the_genomic_accession=90159806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422227, CurrentID=0, Status='live', Symbol='LOC100422227', Aliases='', description='small ubiquitin-like modifier 1 pseudogene', other_designations='SMT3 suppressor of mif two 3 homolog 1 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27949096, end_position_on_the_genomic_accession=27949707, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376784, CurrentID=0, Status='live', Symbol='LOC105376784', Aliases='', description='uncharacterized LOC105376784', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90122393, end_position_on_the_genomic_accession=90125950, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376787, CurrentID=0, Status='live', Symbol='LOC105376787', Aliases='', description='uncharacterized LOC105376787', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  257044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257044, CurrentID=0, Status='live', Symbol='C1orf101', Aliases='', description='chromosome 1 open reading frame 101', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244452100, end_position_on_the_genomic_accession=244640362, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  105376793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376793, CurrentID=0, Status='live', Symbol='LOC105376793', Aliases='', description='uncharacterized LOC105376793', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376794, CurrentID=0, Status='live', Symbol='LOC105376794', Aliases='', description='uncharacterized LOC105376794', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388125, CurrentID=0, Status='live', Symbol='C2CD4B', Aliases='FAM148B, NLF2', description='C2 calcium-dependent domain containing 4B', other_designations='family with sequence similarity 148, member B|nuclear localized factor 2|nuclear-localized factor 2', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62163538, end_position_on_the_genomic_accession=62165283, orientation='minus', exon_count=2, OMIM=610344, no_hdr0=''),
  100419772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419772, CurrentID=0, Status='live', Symbol='LOC100419772', Aliases='', description='CGRP receptor component pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66025181, end_position_on_the_genomic_accession=66028657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  493753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493753, CurrentID=0, Status='live', Symbol='COA5', Aliases='6330578E17Rik, C2orf64, Pet191', description='cytochrome c oxidase assembly factor 5', other_designations='protein C2orf64', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98599323, end_position_on_the_genomic_accession=98608518, orientation='minus', exon_count=3, OMIM=613920, no_hdr0=''),
  493754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493754, CurrentID=0, Status='live', Symbol='GS1-124K5.11', Aliases='', description='RAB guanine nucleotide exchange factor (GEF) 1 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66530305, end_position_on_the_genomic_accession=66592407, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  149935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149935, CurrentID=0, Status='live', Symbol='LOC149935', Aliases='', description='CDK5 regulatory subunit associated protein 3 pseudogene', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31285049, end_position_on_the_genomic_accession=31286865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133920, CurrentID=0, Status='live', Symbol='LOC100133920', Aliases='', description='uncharacterized LOC100133920', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94575976, end_position_on_the_genomic_accession=94587183, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388132, CurrentID=0, Status='live', Symbol='RPL17P5', Aliases='RPL17_2_69', description='ribosomal protein L17 pseudogene 5', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86784878, end_position_on_the_genomic_accession=86785502, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  138649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=138649, CurrentID=0, Status='live', Symbol='ANKRD19P', Aliases='ANKRD19, bA526D8.2', description='ankyrin repeat domain 19, pseudogene', other_designations='ankyrin repeat domain 19 pseudogene', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92809611, end_position_on_the_genomic_accession=92838457, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105376806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376806, CurrentID=0, Status='live', Symbol='LOC105376806', Aliases='', description='uncharacterized LOC105376806', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16976197, end_position_on_the_genomic_accession=16977795, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130843, CurrentID=0, Status='live', Symbol='LOC100130843', Aliases='', description='SPRY domain containing 7 pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=77374212, end_position_on_the_genomic_accession=77381329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376807, CurrentID=0, Status='live', Symbol='LOC105376807', Aliases='', description='uncharacterized LOC105376807', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17111776, end_position_on_the_genomic_accession=17119273, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  493812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493812, CurrentID=0, Status='live', Symbol='HCG11', Aliases='CTA-14H9.3, bK14H9.3', description='HLA complex group 11 (non-protein coding)', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26521706, end_position_on_the_genomic_accession=26527393, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  493813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493813, CurrentID=0, Status='live', Symbol='HCG13P', Aliases='dJ193B12.5', description='HLA complex group 13 pseudogene', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  493816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493816, CurrentID=0, Status='live', Symbol='TUBBP4', Aliases='bA537H15.5', description='tubulin beta pseudogene 4', other_designations='tubulin, beta polypeptide pseudogene 4|tubulin, beta pseudogene 4', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34109199, end_position_on_the_genomic_accession=34109453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  493818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493818, CurrentID=0, Status='live', Symbol='MYMY3', Aliases='', description='moyamoya disease 3', other_designations='', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608796, no_hdr0=''),
  493821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493821, CurrentID=0, Status='live', Symbol='ZNF603P', Aliases='p373c6.7', description='zinc finger protein 603, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28176188, end_position_on_the_genomic_accession=28176674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  493824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493824, CurrentID=0, Status='live', Symbol='ZNF601P', Aliases='D6S2711, PRD51p, dJ313I6.8', description='zinc finger protein 601, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28097111, end_position_on_the_genomic_accession=28097369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  493825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493825, CurrentID=0, Status='live', Symbol='POLR2LP', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide L pseudogene', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31140727, end_position_on_the_genomic_accession=31140913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  493829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493829, CurrentID=0, Status='live', Symbol='TRIM72', Aliases='MG53', description='tripartite motif containing 72, E3 ubiquitin protein ligase', other_designations='mitsugumin 53|mitsugumin-53|tripartite motif-containing 72', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31214021, end_position_on_the_genomic_accession=31225189, orientation='plus', exon_count=7, OMIM=613288, no_hdr0=''),
  105376812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376812, CurrentID=0, Status='live', Symbol='LOC105376812', Aliases='', description='uncharacterized LOC105376812', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18044182, end_position_on_the_genomic_accession=18052715, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376813, CurrentID=0, Status='live', Symbol='LOC105376813', Aliases='', description='uncharacterized LOC105376813', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18262235, end_position_on_the_genomic_accession=18264921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287965, CurrentID=0, Status='live', Symbol='LOC100287965', Aliases='', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene', other_designations='telomeric repeat binding factor (NIMA-interacting) pseudogene', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189917254, end_position_on_the_genomic_accession=189917758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927993, CurrentID=0, Status='live', Symbol='LOC101927993', Aliases='', description='uncharacterized LOC101927993', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93091406, end_position_on_the_genomic_accession=93095543, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  493856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493856, CurrentID=0, Status='live', Symbol='CISD2', Aliases='ERIS, Miner1, NAF-1, WFS2, ZCD2', description='CDGSH iron sulfur domain 2', other_designations='endoplasmic reticulum intermembrane small protein|mitoNEET-related 1 protein|nutrient-deprivation autophagy factor-1|zinc finger, CDGSH-type domain 2', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102868978, end_position_on_the_genomic_accession=102892807, orientation='plus', exon_count=3, OMIM=611507, no_hdr0=''),
  493860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493860, CurrentID=0, Status='live', Symbol='CCDC73', Aliases='NY-SAR-79', description='coiled-coil domain containing 73', other_designations='sarcoma antigen NY-SAR-79', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32602080, end_position_on_the_genomic_accession=32794658, orientation='minus', exon_count=19, OMIM=612328, no_hdr0=''),
  493861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493861, CurrentID=0, Status='live', Symbol='EID3', Aliases='NS4EB, NSE4B, NSMCE4B', description='EP300 interacting inhibitor of differentiation 3', other_designations='E1A-like inhibitor of differentiation 3|EID-1-like inhibitor of differentiation 3|EID-3|non-SMC element 4 homolog B|non-structural maintenance of chromosomes element 4 homolog B', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104303732, end_position_on_the_genomic_accession=104305205, orientation='plus', exon_count=1, OMIM=612986, no_hdr0=''),
  493862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493862, CurrentID=0, Status='live', Symbol='CTRCT28', Aliases='ARCC1', description='Cataract 28', other_designations='', map_location='6p12-q12', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609026, no_hdr0=''),
  493869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493869, CurrentID=0, Status='live', Symbol='GPX8', Aliases='EPLA847, GPx-8, GSHPx-8, UNQ847', description='glutathione peroxidase 8 (putative)', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55160156, end_position_on_the_genomic_accession=55167301, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376819, CurrentID=0, Status='live', Symbol='LOC105376819', Aliases='', description='uncharacterized LOC105376819', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19500221, end_position_on_the_genomic_accession=19531778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  650293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650293, CurrentID=0, Status='live', Symbol='LOC650293', Aliases='', description='seven transmembrane helix receptor', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8949751, end_position_on_the_genomic_accession=8950401, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376822, CurrentID=0, Status='live', Symbol='LOC105376822', Aliases='', description='uncharacterized LOC105376822', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19566744, end_position_on_the_genomic_accession=19582847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  493900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493900, CurrentID=0, Status='live', Symbol='TMEM9B-AS1', Aliases='C11orf18', description='TMEM9B antisense RNA 1', other_designations='TMEM9B antisense RNA 1 (non-protein coding)', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8964675, end_position_on_the_genomic_accession=8976283, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  493901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493901, CurrentID=0, Status='live', Symbol='RNASE12', Aliases='HEL-S-85p, RAI1', description='ribonuclease, RNase A family, 12 (non-active)', other_designations='epididymis secretory sperm binding protein Li 85p|ribonuclease A I1|ribonuclease-like protein 12', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20590080, end_position_on_the_genomic_accession=20590823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927995, CurrentID=0, Status='live', Symbol='LOC101927995', Aliases='', description='uncharacterized LOC101927995', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30349724, end_position_on_the_genomic_accession=30469224, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  493911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493911, CurrentID=0, Status='live', Symbol='PHOSPHO2', Aliases='', description='phosphatase, orphan 2', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169694454, end_position_on_the_genomic_accession=169701708, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  493913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493913, CurrentID=0, Status='live', Symbol='PAPPA-AS1', Aliases='DIPAS, NCRNA00156, PAPPA-AS, PAPPAAS, PAPPAS, TAF2C2, TAF4B, TAFII105', description='PAPPA antisense RNA 1', other_designations='DIPLA1-antisense expressed|PAPPA antisense RNA (non-protein coding)|PAPPA antisense RNA 1 (non-protein coding)|PAPPA antisense RNA 1 (tail to tail)', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116398157, end_position_on_the_genomic_accession=116400606, orientation='minus', exon_count=1, OMIM=610689, no_hdr0=''),
  100129813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129813, CurrentID=0, Status='live', Symbol='CT45A11P', Aliases='', description='cancer/testis antigen family 45, member A11, pseudogene', other_designations='cancer/testis antigen family 45 member A pseudogene', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135942487, end_position_on_the_genomic_accession=135947151, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375967, CurrentID=0, Status='live', Symbol='LOC105375967', Aliases='', description='uncharacterized LOC105375967', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5577021, end_position_on_the_genomic_accession=5628911, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376830, CurrentID=0, Status='live', Symbol='LOC105376830', Aliases='', description='uncharacterized LOC105376830', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20903705, end_position_on_the_genomic_accession=20919273, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376831, CurrentID=0, Status='live', Symbol='LOC105376831', Aliases='', description='uncharacterized LOC105376831', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371603, CurrentID=0, Status='live', Symbol='LOC105371603', Aliases='', description='uncharacterized LOC105371603', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168317512, end_position_on_the_genomic_accession=168417970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370853, CurrentID=0, Status='live', Symbol='LOC105370853', Aliases='', description='uncharacterized LOC105370853', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376833, CurrentID=0, Status='live', Symbol='LOC105376833', Aliases='', description='uncharacterized LOC105376833', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21543560, end_position_on_the_genomic_accession=21551254, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927269, CurrentID=0, Status='live', Symbol='LOC101927269', Aliases='', description='uncharacterized LOC101927269', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80371517, end_position_on_the_genomic_accession=80374439, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133957, CurrentID=0, Status='live', Symbol='UXT-AS1', Aliases='', description='UXT antisense RNA 1', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47658833, end_position_on_the_genomic_accession=47660111, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  493987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=493987, CurrentID=0, Status='live', Symbol='MT1P1', Aliases='bA435O5.3', description='metallothionein 1 pseudogene 1', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95413117, end_position_on_the_genomic_accession=95413515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287966, CurrentID=0, Status='live', Symbol='LOC100287966', Aliases='', description='ATP synthase-coupling factor 6, mitochondrial-like', other_designations='', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22988261, end_position_on_the_genomic_accession=22988612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  494028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494028, CurrentID=0, Status='live', Symbol='CVMRF', Aliases='', description='cubitus valgus with mental retardation and unusual facies', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300471, no_hdr0=''),
  494029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494029, CurrentID=0, Status='live', Symbol='MENOQ1', Aliases='', description='menopause, natural, age at, QTL', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300488, no_hdr0=''),
  105376845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376845, CurrentID=0, Status='live', Symbol='LOC105376845', Aliases='', description='uncharacterized LOC105376845', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22140953, end_position_on_the_genomic_accession=22157385, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100419774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419774, CurrentID=0, Status='live', Symbol='LOC100419774', Aliases='', description='zinc finger protein 383 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35933888, end_position_on_the_genomic_accession=35934569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375971, CurrentID=0, Status='live', Symbol='LOC105375971', Aliases='', description='uncharacterized LOC105375971', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7914543, end_position_on_the_genomic_accession=7961227, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376850, CurrentID=0, Status='live', Symbol='LOC105376850', Aliases='', description='uncharacterized LOC105376850', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22163384, end_position_on_the_genomic_accession=22172505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129818, CurrentID=0, Status='live', Symbol='LOC100129818', Aliases='', description='tubulin, alpha 1a pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115424045, end_position_on_the_genomic_accession=115425694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129099, CurrentID=0, Status='live', Symbol='LOC100129099', Aliases='', description='proline rich nuclear receptor coactivator 2 pseudogene', other_designations='proline-rich nuclear receptor coactivator 2 pseudogene', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111155247, end_position_on_the_genomic_accession=111155967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388181, CurrentID=0, Status='live', Symbol='FAM149B1P1', Aliases='FAM149B2, KIAA0974P', description='family with sequence similarity 149 member B1 pseudogene 1', other_designations='family with sequence similarity 149, member B1 pseudogene 1', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96756986, end_position_on_the_genomic_accession=96758593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376854, CurrentID=0, Status='live', Symbol='LOC105376854', Aliases='', description='eukaryotic translation initiation factor 3 subunit A-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=508396, end_position_on_the_genomic_accession=531912, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422242, CurrentID=0, Status='live', Symbol='KCTD9P4', Aliases='', description='potassium channel tetramerization domain containing 9 pseudogene 4', other_designations='potassium channel tetramerisation domain containing 9 pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112180773, end_position_on_the_genomic_accession=112181922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  494115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494115, CurrentID=0, Status='live', Symbol='RBMXL1', Aliases='RBM1', description='RNA binding motif protein, X-linked-like 1', other_designations='heterogeneous nuclear ribonucleoprotein G-like 1', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88979456, end_position_on_the_genomic_accession=88992960, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  494118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494118, CurrentID=0, Status='live', Symbol='SPANXN1', Aliases='CT11.6', description='SPANX family member N1', other_designations='SPANX family, member N1|cancer/testis antigen family 11, member 6|nuclear-associated protein SPAN-Xn1', map_location='Xq27', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145247587, end_position_on_the_genomic_accession=145256208, orientation='plus', exon_count=2, OMIM=300664, no_hdr0=''),
  494119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494119, CurrentID=0, Status='live', Symbol='SPANXN2', Aliases='CT11.7, SPANX-N2', description='SPANX family member N2', other_designations='SPANX family, member N2|cancer/testis antigen family 11, member 7|nuclear-associated protein SPAN-Xn2', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143712035, end_position_on_the_genomic_accession=143720668, orientation='minus', exon_count=2, OMIM=300665, no_hdr0=''),
  494127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494127, CurrentID=0, Status='live', Symbol='LOC494127', Aliases='', description='NFYC pseudogene', other_designations='nuclear transcription factor Y, gamma pseudogene', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87083886, end_position_on_the_genomic_accession=87085225, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100917, CurrentID=0, Status='live', Symbol='VWSM', Aliases='', description='Van der Woude syndrome modifier', other_designations='', map_location='17p11.2-p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604547, no_hdr0=''),
  100918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100918, CurrentID=0, Status='live', Symbol='NRCLP2', Aliases='NRCLP', description='Narcolepsy, HLA-associated', other_designations='', map_location='4p13-q21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605841, no_hdr0=''),
  494141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494141, CurrentID=0, Status='live', Symbol='LOC494141', Aliases='', description='solute carrier family 25 member 51 pseudogene', other_designations='mitochondrial carrier triple repeat 1 pseudogene|solute carrier family 25, member 51 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18209138, end_position_on_the_genomic_accession=18213578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  494142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494142, CurrentID=0, Status='live', Symbol='DFNM2', Aliases='', description='deafness (mitochondrial) modifier 2', other_designations='', map_location='8p23', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  494143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494143, CurrentID=0, Status='live', Symbol='CHAC2', Aliases='', description='ChaC, cation transport regulator homolog 2 (E. coli)', other_designations='ChaC, cation transport regulator-like 2|cation transport regulator-like protein 2|gamma-GCG 2|gamma-GCT acting on glutathione homolog 2', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53767792, end_position_on_the_genomic_accession=53775150, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  494148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494148, CurrentID=0, Status='live', Symbol='DFNB55', Aliases='', description='deafness, autosomal recessive 55', other_designations='', map_location='4q12-q13.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609952, no_hdr0=''),
  494149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494149, CurrentID=0, Status='live', Symbol='DFNB56', Aliases='', description='deafness, autosomal recessive 56', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  494150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494150, CurrentID=0, Status='live', Symbol='PHBP19', Aliases='', description='prohibitin pseudogene 19', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51124109, end_position_on_the_genomic_accession=51125384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133985, CurrentID=0, Status='live', Symbol='LOC100133985', Aliases='', description='uncharacterized LOC100133985', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70124036, end_position_on_the_genomic_accession=70125316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645744, CurrentID=0, Status='live', Symbol='TAF5LP1', Aliases='HsT33855, TAF5LP', description='TAF5-like RNA polymerase II, p300/CBP-associated factor (PCAF)-associated factor, 65kDa pseudogene 1', other_designations='TAF5-like RNA polymerase II pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35498216, end_position_on_the_genomic_accession=35499181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376868, CurrentID=0, Status='live', Symbol='LOC105376868', Aliases='', description='uncharacterized LOC105376868', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44577963, end_position_on_the_genomic_accession=44579937, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133991, CurrentID=0, Status='live', Symbol='MAP3K14-AS1', Aliases='', description='MAP3K14 antisense RNA 1', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45247925, end_position_on_the_genomic_accession=45268630, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  494188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494188, CurrentID=0, Status='live', Symbol='FBXO47', Aliases='', description='F-box protein 47', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38936278, end_position_on_the_genomic_accession=38967486, orientation='minus', exon_count=11, OMIM=609498, no_hdr0=''),
  494197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494197, CurrentID=0, Status='live', Symbol='SPANXN5', Aliases='CT11.10, SPANX-N5', description='SPANX family member N5', other_designations='SPANX family, member N5|cancer/testis antigen family 11, member 10|nuclear-associated protein SPAN-Xn5', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52796146, end_position_on_the_genomic_accession=52797348, orientation='minus', exon_count=2, OMIM=300668, no_hdr0=''),
  105371612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371612, CurrentID=0, Status='live', Symbol='LOC105371612', Aliases='', description='uncharacterized LOC105371612', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171601081, end_position_on_the_genomic_accession=171626916, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129100, CurrentID=0, Status='live', Symbol='LOC100129100', Aliases='', description='uncharacterized protein C12orf29 homolog', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88420307, end_position_on_the_genomic_accession=88486094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100271364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271364, CurrentID=0, Status='live', Symbol='RPS18P13', Aliases='RPS18_8_1634', description='ribosomal protein S18 pseudogene 13', other_designations='', map_location='19p13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18048629, end_position_on_the_genomic_accession=18049134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376884, CurrentID=0, Status='live', Symbol='LOC105376884', Aliases='', description='uncharacterized LOC105376884', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25887968, end_position_on_the_genomic_accession=25897101, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928007, CurrentID=0, Status='live', Symbol='LOC101928007', Aliases='', description='uncharacterized LOC101928007', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158179052, end_position_on_the_genomic_accession=158182737, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376885, CurrentID=0, Status='live', Symbol='LOC105376885', Aliases='', description='uncharacterized LOC105376885', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25906550, end_position_on_the_genomic_accession=25908243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  348162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348162, CurrentID=0, Status='live', Symbol='LOC348162', Aliases='', description='uncharacterized protein 348162', other_designations='PI-3-kinase-related kinase SMG-1 isoform 1homolog', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376886, CurrentID=0, Status='live', Symbol='LOC105376886', Aliases='', description='uncharacterized LOC105376886', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26118278, end_position_on_the_genomic_accession=26127752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376888, CurrentID=0, Status='live', Symbol='LOC105376888', Aliases='', description='uncharacterized LOC105376888', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26682574, end_position_on_the_genomic_accession=26690812, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  544566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544566, CurrentID=0, Status='live', Symbol='AASTH40', Aliases='', description='Allergic/atopic asthma related QTL 40', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376890, CurrentID=0, Status='live', Symbol='LOC105376890', Aliases='', description='uncharacterized LOC105376890', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27229401, end_position_on_the_genomic_accession=27234494, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376891, CurrentID=0, Status='live', Symbol='LOC105376891', Aliases='', description='uncharacterized LOC105376891', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  544567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544567, CurrentID=0, Status='live', Symbol='AASTH14', Aliases='', description='Allergic/atopic asthma related QTL 14', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  494323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494323, CurrentID=0, Status='live', Symbol='MIR361', Aliases='MIRN361, hsa-mir-361, mir-361', description='microRNA 361', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=85903636, end_position_on_the_genomic_accession=85903707, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  494324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494324, CurrentID=0, Status='live', Symbol='MIR375', Aliases='MIRN375, hsa-mir-375, miRNA375, mir-375', description='microRNA 375', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219001645, end_position_on_the_genomic_accession=219001708, orientation='minus', exon_count=1, OMIM=611173, no_hdr0=''),
  494325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494325, CurrentID=0, Status='live', Symbol='MIR376A1', Aliases='MIR376A-1, MIRN376A, MIRN376A-1, MIRN376A1, mir-376a-1', description='microRNA 376a-1', other_designations='hsa-mir-376a|hsa-mir-376a-1', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101040782, end_position_on_the_genomic_accession=101040849, orientation='plus', exon_count=1, OMIM=610959, no_hdr0=''),
  494326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494326, CurrentID=0, Status='live', Symbol='MIR377', Aliases='MIRN377, hsa-mir-377, mir-377', description='microRNA 377', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101062050, end_position_on_the_genomic_accession=101062118, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  494327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494327, CurrentID=0, Status='live', Symbol='MIR378A', Aliases='MIR378, MIRN378, hsa-mir-378, hsa-mir-378a, miRNA378', description='microRNA 378a', other_designations='microRNA 378', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149732825, end_position_on_the_genomic_accession=149732890, orientation='plus', exon_count=1, OMIM=611957, no_hdr0=''),
  494328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494328, CurrentID=0, Status='live', Symbol='MIR379', Aliases='MIRN379, hsa-mir-379, mir-379', description='microRNA 379', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101022066, end_position_on_the_genomic_accession=101022132, orientation='plus', exon_count=1, OMIM=616358, no_hdr0=''),
  494329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494329, CurrentID=0, Status='live', Symbol='MIR380', Aliases='MIR380-3p, MIRN380, hsa-mir-380, mir-380', description='microRNA 380', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101025017, end_position_on_the_genomic_accession=101025077, orientation='plus', exon_count=1, OMIM=613654, no_hdr0=''),
  494330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494330, CurrentID=0, Status='live', Symbol='MIR381', Aliases='MIRN381, hsa-mir-381, mir-381', description='microRNA 381', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101045920, end_position_on_the_genomic_accession=101045994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  494331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494331, CurrentID=0, Status='live', Symbol='MIR382', Aliases='MIRN382, hsa-mir-382, mir-382', description='microRNA 382', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101054306, end_position_on_the_genomic_accession=101054381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  494332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494332, CurrentID=0, Status='live', Symbol='MIR383', Aliases='MIRN383, hsa-mir-383, mir-383', description='microRNA 383', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14853438, end_position_on_the_genomic_accession=14853510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  494333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494333, CurrentID=0, Status='live', Symbol='MIR384', Aliases='MIRN384, hsa-mir-384', description='microRNA 384', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76919273, end_position_on_the_genomic_accession=76919360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  494334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494334, CurrentID=0, Status='live', Symbol='MIR422A', Aliases='MIRN422A', description='microRNA 422a', other_designations='hsa-mir-422a|miR-422a', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63870930, end_position_on_the_genomic_accession=63871019, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  494335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494335, CurrentID=0, Status='live', Symbol='MIR423', Aliases='MIRN423, hsa-mir-423, mir-423', description='microRNA 423', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30117079, end_position_on_the_genomic_accession=30117172, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  494336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494336, CurrentID=0, Status='live', Symbol='MIR424', Aliases='MIR322, MIRN424, hsa-mir-424, miRNA424, mir-424', description='microRNA 424', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134546614, end_position_on_the_genomic_accession=134546711, orientation='minus', exon_count=1, OMIM=300682, no_hdr0=''),
  494337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494337, CurrentID=0, Status='live', Symbol='MIR425', Aliases='MIRN425, hsa-mir-425, mir-425', description='microRNA 425', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49020148, end_position_on_the_genomic_accession=49020234, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376900, CurrentID=0, Status='live', Symbol='LOC105376900', Aliases='', description='uncharacterized LOC105376900', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36266808, end_position_on_the_genomic_accession=36269037, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100652840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652840, CurrentID=0, Status='live', Symbol='LOC100652840', Aliases='', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein) pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30507372, end_position_on_the_genomic_accession=30512796, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376904, CurrentID=0, Status='live', Symbol='LOC105376904', Aliases='', description='uncharacterized LOC105376904', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54271187, end_position_on_the_genomic_accession=54271747, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422252, CurrentID=0, Status='live', Symbol='LOC100422252', Aliases='', description='olfactory receptor, family 11, subfamily H, member 12 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375983, CurrentID=0, Status='live', Symbol='LOC105375983', Aliases='', description='uncharacterized LOC105375983', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=17014078, end_position_on_the_genomic_accession=17112857, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927272, CurrentID=0, Status='live', Symbol='LOC101927272', Aliases='', description='uncharacterized LOC101927272', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50668855, end_position_on_the_genomic_accession=50671639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372602, CurrentID=0, Status='live', Symbol='LOC105372602', Aliases='', description='uncharacterized LOC105372602', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36213337, end_position_on_the_genomic_accession=36236461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376913, CurrentID=0, Status='live', Symbol='LOC105376913', Aliases='', description='uncharacterized LOC105376913', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=93537, end_position_on_the_genomic_accession=95391, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388242, CurrentID=0, Status='live', Symbol='LOC388242', Aliases='', description='SAGA complex associated factor 29 pseudogene', other_designations='coiled-coil domain containing 101 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29464967, end_position_on_the_genomic_accession=29467578, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  494446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494446, CurrentID=0, Status='live', Symbol='NRCLP3', Aliases='', description='Narcolepsy 3', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609039, no_hdr0=''),
  494447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494447, CurrentID=0, Status='live', Symbol='COILP2', Aliases='COILP-2', description='coilin pseudogene 2', other_designations='', map_location='4q27-q28', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123023734, end_position_on_the_genomic_accession=123026090, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653602, CurrentID=0, Status='live', Symbol='LOC653602', Aliases='', description='uncharacterized LOC653602', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=14635091, end_position_on_the_genomic_accession=14650819, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422254, CurrentID=0, Status='live', Symbol='EBAG9P1', Aliases='', description='estrogen receptor binding site associated, antigen, 9 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99697339, end_position_on_the_genomic_accession=99697949, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  494470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494470, CurrentID=0, Status='live', Symbol='RNF165', Aliases='ARKL2, RNF111L2', description='ring finger protein 165', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46334224, end_position_on_the_genomic_accession=46461412, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105376920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376920, CurrentID=0, Status='live', Symbol='LOC105376920', Aliases='', description='uncharacterized LOC105376920', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60177, end_position_on_the_genomic_accession=77689, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376921, CurrentID=0, Status='live', Symbol='LOC105376921', Aliases='', description='uncharacterized LOC105376921', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=54198, end_position_on_the_genomic_accession=67717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131034, CurrentID=0, Status='live', Symbol='LOC100131034', Aliases='', description='uncharacterized LOC100131034', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=103893175, end_position_on_the_genomic_accession=103896272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375986, CurrentID=0, Status='live', Symbol='LOC105375986', Aliases='', description='uncharacterized LOC105375986', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18797058, end_position_on_the_genomic_accession=18800642, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  494513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494513, CurrentID=0, Status='live', Symbol='DFNB59', Aliases='PJVK', description='deafness, autosomal recessive 59', other_designations='autosomal recessive deafness type 59 protein', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178450592, end_position_on_the_genomic_accession=178461390, orientation='plus', exon_count=10, OMIM=610219, no_hdr0=''),
  494514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494514, CurrentID=0, Status='live', Symbol='TYMSOS', Aliases='C18orf56', description='TYMS opposite strand', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=649620, end_position_on_the_genomic_accession=658340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480210, CurrentID=0, Status='live', Symbol='RNU2-31P', Aliases='', description='RNA, U2 small nuclear 31, pseudogene', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376926, CurrentID=0, Status='live', Symbol='LOC105376926', Aliases='', description='uncharacterized LOC105376926', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=2810851, end_position_on_the_genomic_accession=2853757, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376927, CurrentID=0, Status='live', Symbol='LOC105376927', Aliases='', description='uncharacterized LOC105376927', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  494540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494540, CurrentID=0, Status='live', Symbol='TBCAP1', Aliases='', description='tubulin folding cofactor A pseudogene 1', other_designations='tubulin-specific chaperone a pseudogene 1', map_location='Xp21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=33041291, end_position_on_the_genomic_accession=33041898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  494541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494541, CurrentID=0, Status='live', Symbol='TBCAP2', Aliases='', description='tubulin folding cofactor A pseudogene 2', other_designations='tubulin-specific chaperone a pseudogene 2', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14691937, end_position_on_the_genomic_accession=14692400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  494551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494551, CurrentID=0, Status='live', Symbol='WEE2', Aliases='WEE1B', description='WEE1 homolog 2 (S. pombe)', other_designations='wee1-like protein kinase 1B|wee1B kinase', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141708353, end_position_on_the_genomic_accession=141731271, orientation='plus', exon_count=12, OMIM=614084, no_hdr0=''),
  494555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494555, CurrentID=0, Status='live', Symbol='TWF1P2', Aliases='A6ps2, PTK9P2', description='twinfilin 1 pseudogene 2', other_designations='protein tyrosine kinase 9 pseudogene 2|twinfilin, actin-binding protein, homolog 1 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155288174, end_position_on_the_genomic_accession=155291419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  494558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=494558, CurrentID=0, Status='live', Symbol='C9orf135-AS1', Aliases='', description='C9orf135 antisense RNA 1 (head to head)', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69819405, end_position_on_the_genomic_accession=69820739, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375988, CurrentID=0, Status='live', Symbol='LOC105375988', Aliases='', description='uncharacterized LOC105375988', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19470455, end_position_on_the_genomic_accession=19562872, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376935, CurrentID=0, Status='live', Symbol='LOC105376935', Aliases='', description='uncharacterized LOC105376935', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5156839, end_position_on_the_genomic_accession=5187376, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100286973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100286973, CurrentID=0, Status='live', Symbol='LOC100286973', Aliases='', description='protein phosphatase 1, catalytic subunit, alpha isozyme pseudogene', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96672157, end_position_on_the_genomic_accession=96672810, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376938, CurrentID=0, Status='live', Symbol='LOC105376938', Aliases='', description='uncharacterized LOC105376938', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5691054, end_position_on_the_genomic_accession=5699455, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102724973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724973, CurrentID=0, Status='live', Symbol='LOC102724973', Aliases='', description='uncharacterized LOC102724973', other_designations='CTD-3076O17.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99928452, end_position_on_the_genomic_accession=99974003, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376939, CurrentID=0, Status='live', Symbol='LOC105376939', Aliases='', description='uncharacterized LOC105376939', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5684336, end_position_on_the_genomic_accession=5697634, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100291323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100291323, CurrentID=0, Status='live', Symbol='LOC100291323', Aliases='', description='uncharacterized LOC100291323', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376944, CurrentID=0, Status='live', Symbol='LOC105376944', Aliases='', description='uncharacterized LOC105376944', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6694754, end_position_on_the_genomic_accession=6735810, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  400548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400548, CurrentID=0, Status='live', Symbol='LOC400548', Aliases='', description='uncharacterized LOC400548', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85137150, end_position_on_the_genomic_accession=85149443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100502567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100502567, CurrentID=0, Status='live', Symbol='DEL15Q24', Aliases='CHR15DELq24', description='Chromosome 15q24 deletion syndrome', other_designations='', map_location='15q24', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613406, no_hdr0=''),
  100502568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100502568, CurrentID=0, Status='live', Symbol='STQTL21', Aliases='', description='Stature QTL 21', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613440, no_hdr0=''),
  100502569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100502569, CurrentID=0, Status='live', Symbol='BWQTL2', Aliases='', description='Birth weight QTL 2', other_designations='', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613459, no_hdr0=''),
  100502572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100502572, CurrentID=0, Status='live', Symbol='LOC100502572', Aliases='', description='chromosome 5 open reading frame 60 pseudogene', other_designations='', map_location='5q35', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179651290, end_position_on_the_genomic_accession=179653226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376946, CurrentID=0, Status='live', Symbol='GRM7-AS2', Aliases='', description='GRM7 antisense RNA 2', other_designations='GRM7 antisense RNA 2 (non-protein coding)', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6883470, end_position_on_the_genomic_accession=6895322, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391343, CurrentID=0, Status='live', Symbol='LOC391343', Aliases='', description='uncharacterized LOC391343', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388275, CurrentID=0, Status='live', Symbol='HNRNPA1P7', Aliases='HNRPA1L3', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 7', other_designations='heterogeneous nuclear ribonucleoprotein A1-like 3', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32411506, end_position_on_the_genomic_accession=32413237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376948, CurrentID=0, Status='live', Symbol='LOC105376948', Aliases='', description='uncharacterized LOC105376948', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376949, CurrentID=0, Status='live', Symbol='LOC105376949', Aliases='', description='uncharacterized LOC105376949', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9598421, end_position_on_the_genomic_accession=9600687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422261, CurrentID=0, Status='live', Symbol='NPM1P37', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 37', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=69705090, end_position_on_the_genomic_accession=69706257, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376954, CurrentID=0, Status='live', Symbol='LOC105376954', Aliases='', description='uncharacterized LOC105376954', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13007888, end_position_on_the_genomic_accession=13010674, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376956, CurrentID=0, Status='live', Symbol='LOC105376956', Aliases='', description='uncharacterized LOC105376956', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12876344, end_position_on_the_genomic_accession=12942962, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376961, CurrentID=0, Status='live', Symbol='LOC105376961', Aliases='', description='uncharacterized LOC105376961', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14587952, end_position_on_the_genomic_accession=14594765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422263, CurrentID=0, Status='live', Symbol='LOC100422263', Aliases='', description='serine palmitoyltransferase, long chain base subunit 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167767083, end_position_on_the_genomic_accession=167768585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376962, CurrentID=0, Status='live', Symbol='LOC105376962', Aliases='', description='uncharacterized LOC105376962', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14655640, end_position_on_the_genomic_accession=14665801, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376963, CurrentID=0, Status='live', Symbol='LOC105376963', Aliases='', description='uncharacterized LOC105376963', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14920444, end_position_on_the_genomic_accession=14922404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376967, CurrentID=0, Status='live', Symbol='LOC105376967', Aliases='', description='uncharacterized LOC105376967', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16125425, end_position_on_the_genomic_accession=16130614, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422265, CurrentID=0, Status='live', Symbol='SUMO2P19', Aliases='', description='SUMO2 pseudogene 19', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene|small ubiquitin-like modifier 2 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102242038, end_position_on_the_genomic_accession=102242618, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131069, CurrentID=0, Status='live', Symbol='LOC100131069', Aliases='', description='coiled-coil domain containing 53 pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70857065, end_position_on_the_genomic_accession=70857778, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376975, CurrentID=0, Status='live', Symbol='LOC105376975', Aliases='', description='uncharacterized LOC105376975', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17742878, end_position_on_the_genomic_accession=17745665, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480220, CurrentID=0, Status='live', Symbol='RNU2-53P', Aliases='', description='RNA, U2 small nuclear 53, pseudogene', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129843, CurrentID=0, Status='live', Symbol='LOC100129843', Aliases='', description='pyruvate dehydrogenase kinase, isozyme 1 pseudogene', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81600033, end_position_on_the_genomic_accession=81601301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376979, CurrentID=0, Status='live', Symbol='LOC105376979', Aliases='', description='uncharacterized LOC105376979', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47831871, end_position_on_the_genomic_accession=47834983, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376980, CurrentID=0, Status='live', Symbol='LOC105376980', Aliases='', description='uncharacterized LOC105376980', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47894362, end_position_on_the_genomic_accession=47899523, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480221, CurrentID=0, Status='live', Symbol='RNU2-56P', Aliases='', description='RNA, U2 small nuclear 56, pseudogene', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376981, CurrentID=0, Status='live', Symbol='LOC105376981', Aliases='', description='uncharacterized LOC105376981', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19003665, end_position_on_the_genomic_accession=19012232, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101027084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101027084, CurrentID=0, Status='live', Symbol='LOC101027084', Aliases='', description='MAX pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135650715, end_position_on_the_genomic_accession=135653415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376984, CurrentID=0, Status='live', Symbol='LOC105376984', Aliases='', description='uncharacterized LOC105376984', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19703639, end_position_on_the_genomic_accession=19728957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126170, CurrentID=0, Status='live', Symbol='LOC126170', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29921135, end_position_on_the_genomic_accession=29922171, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372605, CurrentID=0, Status='live', Symbol='LOC105372605', Aliases='', description='uncharacterized LOC105372605', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37281184, end_position_on_the_genomic_accession=37289076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  391828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391828, CurrentID=0, Status='live', Symbol='RPL28P3', Aliases='RPL28_3_609', description='ribosomal protein L28 pseudogene 3', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124808826, end_position_on_the_genomic_accession=124809231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929021, CurrentID=0, Status='live', Symbol='LOC101929021', Aliases='', description='uncharacterized LOC101929021', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69572980, end_position_on_the_genomic_accession=69577099, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480223, CurrentID=0, Status='live', Symbol='RNU2-59P', Aliases='', description='RNA, U2 small nuclear 59, pseudogene', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376991, CurrentID=0, Status='live', Symbol='LOC105376991', Aliases='', description='uncharacterized LOC105376991', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23417436, end_position_on_the_genomic_accession=23419769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100240702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240702, CurrentID=0, Status='live', Symbol='PLSA1', Aliases='PLSA', description='Primary lateral sclerosis, adult, 1', other_designations='', map_location='4p16', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611637, no_hdr0=''),
  645749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645749, CurrentID=0, Status='live', Symbol='LOC645749', Aliases='', description='zinc finger protein 131 pseudogene', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145488934, end_position_on_the_genomic_accession=145492106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100240707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240707, CurrentID=0, Status='live', Symbol='GAPDHP42', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 42', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=69745800, end_position_on_the_genomic_accession=69747055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100240708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240708, CurrentID=0, Status='live', Symbol='GAPDHP43', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 43', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44354608, end_position_on_the_genomic_accession=44355902, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100240709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240709, CurrentID=0, Status='live', Symbol='GAPDHP44', Aliases='GAPDH', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 44', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62755153, end_position_on_the_genomic_accession=62756433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100240710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240710, CurrentID=0, Status='live', Symbol='GAPDHP45', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 45', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15093061, end_position_on_the_genomic_accession=15093613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100240711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240711, CurrentID=0, Status='live', Symbol='GAPDHP46', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 46', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92114639, end_position_on_the_genomic_accession=92115849, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100240712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240712, CurrentID=0, Status='live', Symbol='GAPDHP47', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 47', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143503074, end_position_on_the_genomic_accession=143504344, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100240713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240713, CurrentID=0, Status='live', Symbol='GAPDHP48', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 48', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3687784, end_position_on_the_genomic_accession=3689059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100240714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240714, CurrentID=0, Status='live', Symbol='LOC100240714', Aliases='', description='multimerin 2 pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149905638, end_position_on_the_genomic_accession=149907006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376995, CurrentID=0, Status='live', Symbol='LOC105376995', Aliases='', description='uncharacterized LOC105376995', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62534032, end_position_on_the_genomic_accession=62536728, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100240728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240728, CurrentID=0, Status='live', Symbol='LOC100240728', Aliases='', description='uncharacterized LOC100240728', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56482105, end_position_on_the_genomic_accession=56483319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480224, CurrentID=0, Status='live', Symbol='RNU2-62P', Aliases='', description='RNA, U2 small nuclear 62, pseudogene', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100240730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240730, CurrentID=0, Status='live', Symbol='DUP7Q11.23', Aliases='C7DUPq11.23', description='Chromosome 7q11.23 duplication syndrome', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609757, no_hdr0=''),
  100240731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240731, CurrentID=0, Status='live', Symbol='DEL15Q15.3', Aliases='C15DELq15.3', description='Deafness and male infertility', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611102, no_hdr0=''),
  100240732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240732, CurrentID=0, Status='live', Symbol='IBD22', Aliases='', description='Inflammatory bowel disease-22', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612380, no_hdr0=''),
  100240733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240733, CurrentID=0, Status='live', Symbol='IBD23', Aliases='', description='Inflammatory bowel disease-23', other_designations='', map_location='1q32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612381, no_hdr0=''),
  100240734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240734, CurrentID=0, Status='live', Symbol='LOC100240734', Aliases='', description='uncharacterized LOC100240734', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54102209, end_position_on_the_genomic_accession=54102699, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100240735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240735, CurrentID=0, Status='live', Symbol='LOC100240735', Aliases='', description='uncharacterized LOC100240735', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54078839, end_position_on_the_genomic_accession=54081823, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100240736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240736, CurrentID=0, Status='live', Symbol='DEL11P15P14', Aliases='C11DELp15p14', description='Chromosome 11p15-p14 deletion syndrome', other_designations='', map_location='11p15-p14', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606528, no_hdr0=''),
  100240737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240737, CurrentID=0, Status='live', Symbol='DEL1P36', Aliases='C1DELp36', description='Chromosome 1p36 deletion syndrome', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607872, no_hdr0=''),
  100240738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240738, CurrentID=0, Status='live', Symbol='DUP22Q11.2', Aliases='', description='Chromosome 22q11.2 microduplication syndrome', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608363, no_hdr0=''),
  100240739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240739, CurrentID=0, Status='live', Symbol='DUP18PDEL18Q', Aliases='DUP18qDEL18p, DUP81qDEL18p', description='Chromosome 18 pericentric inversion', other_designations='', map_location='18q22', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609334, no_hdr0=''),
  100240740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240740, CurrentID=0, Status='live', Symbol='DEL2P16.1-P15', Aliases='C2DELp161-p15', description='Chromosome 2p16.1-p15 deletion syndrome', other_designations='', map_location='2p16.1-p15', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612513, no_hdr0=''),
  100240747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240747, CurrentID=0, Status='live', Symbol='DEL18P', Aliases='C18DELp', description='Chromosome 18p deletion syndrome', other_designations='', map_location='18p', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=146390, no_hdr0=''),
  100240748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240748, CurrentID=0, Status='live', Symbol='DEL9P', Aliases='C9DELp', description='Chromosome 9p deletion syndrome', other_designations='', map_location='9p', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=158170, no_hdr0=''),
  100240749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100240749, CurrentID=0, Status='live', Symbol='LOC100240749', Aliases='', description='tubulin, epsilon 1 pseudogene', other_designations='', map_location='5p34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170896866, end_position_on_the_genomic_accession=170905084, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480225, CurrentID=0, Status='live', Symbol='RNU2-63P', Aliases='', description='RNA, U2 small nuclear 63, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377002, CurrentID=0, Status='live', Symbol='LOC105377002', Aliases='', description='uncharacterized LOC105377002', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26143650, end_position_on_the_genomic_accession=26208261, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377005, CurrentID=0, Status='live', Symbol='LOC105377005', Aliases='', description='uncharacterized LOC105377005', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27512614, end_position_on_the_genomic_accession=27532654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377007, CurrentID=0, Status='live', Symbol='LOC105377007', Aliases='', description='uncharacterized LOC105377007', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27797588, end_position_on_the_genomic_accession=27860325, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377008, CurrentID=0, Status='live', Symbol='LOC105377008', Aliases='', description='uncharacterized LOC105377008', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28089251, end_position_on_the_genomic_accession=28137033, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505501, CurrentID=0, Status='live', Symbol='LOC100505501', Aliases='', description='uncharacterized LOC100505501', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=59119843, end_position_on_the_genomic_accession=59121347, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377011, CurrentID=0, Status='live', Symbol='LOC105377011', Aliases='', description='uncharacterized LOC105377011', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29260787, end_position_on_the_genomic_accession=29261760, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129847, CurrentID=0, Status='live', Symbol='LOC100129847', Aliases='', description='ras homolog family member A pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=91815843, end_position_on_the_genomic_accession=91816407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377013, CurrentID=0, Status='live', Symbol='LOC105377013', Aliases='', description='uncharacterized LOC105377013', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30422266, end_position_on_the_genomic_accession=30467207, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271388, CurrentID=0, Status='live', Symbol='RPL17P49', Aliases='RPL17_20_1764', description='ribosomal protein L17 pseudogene 49', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=12395757, end_position_on_the_genomic_accession=12396167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  3391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=3391, CurrentID=0, Status='live', Symbol='ICR4', Aliases='', description='ichthyosis congenita IV, ichthyosis-prematurity syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929895, CurrentID=0, Status='live', Symbol='LOC101929895', Aliases='', description='eukaryotic translation initiation factor 3 subunit A-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=665244, end_position_on_the_genomic_accession=668743, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377020, CurrentID=0, Status='live', Symbol='LOC105377020', Aliases='', description='uncharacterized LOC105377020', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31533327, end_position_on_the_genomic_accession=31537413, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377021, CurrentID=0, Status='live', Symbol='LOC105377021', Aliases='', description='putative uncharacterized protein UNQ6490/PRO21339', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32814869, end_position_on_the_genomic_accession=32817341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422275, CurrentID=0, Status='live', Symbol='LOC100422275', Aliases='', description='proteasome subunit beta 5 pseudogene', other_designations='proteasome (prosome, macropain) subunit, beta type, 5 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47262898, end_position_on_the_genomic_accession=47263720, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377022, CurrentID=0, Status='live', Symbol='LOC105377022', Aliases='', description='uncharacterized LOC105377022', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32961108, end_position_on_the_genomic_accession=32975839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377024, CurrentID=0, Status='live', Symbol='LOC105377024', Aliases='', description='uncharacterized LOC105377024', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33878854, end_position_on_the_genomic_accession=33880213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287066, CurrentID=0, Status='live', Symbol='LOC100287066', Aliases='', description='defensin beta 131 pseudogene', other_designations='defensin, beta 131 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12290068, end_position_on_the_genomic_accession=12296189, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101027378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101027378, CurrentID=0, Status='live', Symbol='ETM3', Aliases='', description='Tremor, hereditary essential, 3', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611456, no_hdr0=''),
  105221694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105221694, CurrentID=0, Status='live', Symbol='BISPR', Aliases='lncBST2', description='BST2 interferon stimulated positive regulator (non-protein coding)', other_designations='BST2 IFN-stimulated positive regulator|CTD-2521M24.9', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17405686, end_position_on_the_genomic_accession=17415736, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377035, CurrentID=0, Status='live', Symbol='LOC105377035', Aliases='', description='uncharacterized LOC105377035', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39177764, end_position_on_the_genomic_accession=39181521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377036, CurrentID=0, Status='live', Symbol='LOC105377036', Aliases='', description='uncharacterized LOC105377036', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39201800, end_position_on_the_genomic_accession=39210654, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100287074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287074, CurrentID=0, Status='live', Symbol='LOC100287074', Aliases='', description='ARP1 actin-related protein 1 homolog A, centractin alpha (yeast) pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110184056, end_position_on_the_genomic_accession=110185178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372265, CurrentID=0, Status='live', Symbol='LOC105372265', Aliases='', description='uncharacterized LOC105372265', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7925688, end_position_on_the_genomic_accession=7926793, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373480, CurrentID=0, Status='live', Symbol='LOC105373480', Aliases='', description='uncharacterized LOC105373480', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=731290, end_position_on_the_genomic_accession=779111, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927123, CurrentID=0, Status='live', Symbol='LOC101927123', Aliases='', description='uncharacterized LOC101927123', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127322307, end_position_on_the_genomic_accession=127390670, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377044, CurrentID=0, Status='live', Symbol='LOC105377044', Aliases='', description='uncharacterized LOC105377044', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41041226, end_position_on_the_genomic_accession=41043973, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  400568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400568, CurrentID=0, Status='live', Symbol='LOC400568', Aliases='', description='uncharacterized LOC400568', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377045, CurrentID=0, Status='live', Symbol='LOC105377045', Aliases='', description='uncharacterized LOC105377045', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41050537, end_position_on_the_genomic_accession=41058139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377047, CurrentID=0, Status='live', Symbol='LOC105377047', Aliases='', description='uncharacterized LOC105377047', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42128140, end_position_on_the_genomic_accession=42131437, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928040, CurrentID=0, Status='live', Symbol='BARX1-AS1', Aliases='', description='BARX1 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93955834, end_position_on_the_genomic_accession=93957922, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377051, CurrentID=0, Status='live', Symbol='LOC105377051', Aliases='', description='uncharacterized LOC105377051', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43151579, end_position_on_the_genomic_accession=43153124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377053, CurrentID=0, Status='live', Symbol='LOC105377053', Aliases='', description='uncharacterized LOC105377053', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43305750, end_position_on_the_genomic_accession=43322716, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377057, CurrentID=0, Status='live', Symbol='LOC105377057', Aliases='', description='uncharacterized LOC105377057', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44937901, end_position_on_the_genomic_accession=44943802, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377058, CurrentID=0, Status='live', Symbol='LOC105377058', Aliases='', description='uncharacterized LOC105377058', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18945739, end_position_on_the_genomic_accession=18946078, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377059, CurrentID=0, Status='live', Symbol='LOC105377059', Aliases='', description='uncharacterized LOC105377059', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45068106, end_position_on_the_genomic_accession=45082343, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377061, CurrentID=0, Status='live', Symbol='LOC105377061', Aliases='', description='uncharacterized LOC105377061', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45307889, end_position_on_the_genomic_accession=45313277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377064, CurrentID=0, Status='live', Symbol='LOC105377064', Aliases='', description='aspartate, glycine, lysine and serine-rich protein-like', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3123606, end_position_on_the_genomic_accession=3127210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377066, CurrentID=0, Status='live', Symbol='LOC105377066', Aliases='', description='uncharacterized LOC105377066', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46061083, end_position_on_the_genomic_accession=46085851, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929897, CurrentID=0, Status='live', Symbol='LOC101929897', Aliases='', description='uncharacterized LOC101929897', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42255320, end_position_on_the_genomic_accession=42270961, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317647, CurrentID=0, Status='live', Symbol='FLJ11857', Aliases='', description='uncharacterized protein FLJ11857', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377073, CurrentID=0, Status='live', Symbol='LOC105377073', Aliases='', description='uncharacterized LOC105377073', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47476216, end_position_on_the_genomic_accession=47482470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376100, CurrentID=0, Status='live', Symbol='LOC105376100', Aliases='', description='uncharacterized LOC105376100', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79887960, end_position_on_the_genomic_accession=79892009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100419783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419783, CurrentID=0, Status='live', Symbol='LOC100419783', Aliases='', description='zinc finger protein 831 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17897705, end_position_on_the_genomic_accession=17900269, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  401320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401320, CurrentID=0, Status='live', Symbol='LOC401320', Aliases='', description='uncharacterized LOC401320', other_designations='', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30548357, end_position_on_the_genomic_accession=30577779, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  388403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388403, CurrentID=0, Status='live', Symbol='YPEL2', Aliases='FKSG4', description='yippee like 2', other_designations='DiGeorge syndrome-related protein|yippee-like 2', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59331692, end_position_on_the_genomic_accession=59401734, orientation='plus', exon_count=5, OMIM=609723, no_hdr0=''),
  100422286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422286, CurrentID=0, Status='live', Symbol='LOC100422286', Aliases='', description='proteasome subunit alpha 1 pseudogene', other_designations='proteasome (prosome, macropain) subunit, alpha type, 1 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80709821, end_position_on_the_genomic_accession=80710879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129108, CurrentID=0, Status='live', Symbol='LOC100129108', Aliases='', description='tripartite motif containing 64B pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49015483, end_position_on_the_genomic_accession=49015894, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723362, CurrentID=0, Status='live', Symbol='LOC102723362', Aliases='', description='uncharacterized LOC102723362', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22536479, end_position_on_the_genomic_accession=22538288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388406, CurrentID=0, Status='live', Symbol='LOC388406', Aliases='', description='uncharacterized LOC388406', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60564546, end_position_on_the_genomic_accession=60586620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377079, CurrentID=0, Status='live', Symbol='LOC105377079', Aliases='', description='uncharacterized LOC105377079', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48985485, end_position_on_the_genomic_accession=48987083, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505570, CurrentID=0, Status='live', Symbol='LOC100505570', Aliases='', description='uncharacterized LOC100505570', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1687506, end_position_on_the_genomic_accession=1689058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422287, CurrentID=0, Status='live', Symbol='LOC100422287', Aliases='', description='proteasome subunit alpha 3 pseudogene', other_designations='proteasome (prosome, macropain) subunit, alpha type, 3 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33511891, end_position_on_the_genomic_accession=33512784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  650560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650560, CurrentID=0, Status='live', Symbol='DDX3P3', Aliases='DDX3YP3', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, pseudogene 3', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, Y-linked pseudogene 3', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103572059, end_position_on_the_genomic_accession=103576532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369869, CurrentID=0, Status='live', Symbol='LOC105369869', Aliases='', description='uncharacterized LOC105369869', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80936093, end_position_on_the_genomic_accession=80937633, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377089, CurrentID=0, Status='live', Symbol='LOC105377089', Aliases='', description='uncharacterized LOC105377089', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52046148, end_position_on_the_genomic_accession=52050405, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376019, CurrentID=0, Status='live', Symbol='LOC105376019', Aliases='', description='uncharacterized LOC105376019', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33290610, end_position_on_the_genomic_accession=33294413, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373482, CurrentID=0, Status='live', Symbol='LOC105373482', Aliases='', description='uncharacterized LOC105373482', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1443513, end_position_on_the_genomic_accession=1445676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271391, CurrentID=0, Status='live', Symbol='RPL21P133', Aliases='RPL21_63_1815', description='ribosomal protein L21 pseudogene 133', other_designations='', map_location='Xq26', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134607112, end_position_on_the_genomic_accession=134607671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377091, CurrentID=0, Status='live', Symbol='LOC105377091', Aliases='', description='uncharacterized LOC105377091', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52443076, end_position_on_the_genomic_accession=52448421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388248, CurrentID=0, Status='live', Symbol='LOC388248', Aliases='', description='Rho GTPase activating protein 23 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31975802, end_position_on_the_genomic_accession=32002846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377098, CurrentID=0, Status='live', Symbol='LOC105377098', Aliases='', description='uncharacterized LOC105377098', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55286324, end_position_on_the_genomic_accession=55299743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377100, CurrentID=0, Status='live', Symbol='LOC105377100', Aliases='', description='uncharacterized LOC105377100', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55609467, end_position_on_the_genomic_accession=55613097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480846, CurrentID=0, Status='live', Symbol='RN7SKP42', Aliases='', description='RNA, 7SK small nuclear pseudogene 42', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371657, CurrentID=0, Status='live', Symbol='LOC105371657', Aliases='', description='uncharacterized LOC105371657', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377102, CurrentID=0, Status='live', Symbol='LOC105377102', Aliases='', description='uncharacterized LOC105377102', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57276349, end_position_on_the_genomic_accession=57278129, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723363, CurrentID=0, Status='live', Symbol='LOC102723363', Aliases='', description='uncharacterized LOC102723363', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928051, CurrentID=0, Status='live', Symbol='LOC101928051', Aliases='', description='uncharacterized LOC101928051', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4995491, end_position_on_the_genomic_accession=4997777, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422292, CurrentID=0, Status='live', Symbol='LOC100422292', Aliases='', description='ral guanine nucleotide dissociation stimulator-like 4 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32386668, end_position_on_the_genomic_accession=32387316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388436, CurrentID=0, Status='live', Symbol='LOC388436', Aliases='', description='uncharacterized protein ENSP00000382042', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19219382, end_position_on_the_genomic_accession=19223730, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377109, CurrentID=0, Status='live', Symbol='LOC105377109', Aliases='', description='uncharacterized LOC105377109', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58663142, end_position_on_the_genomic_accession=58706070, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388439, CurrentID=0, Status='live', Symbol='FLJ12120', Aliases='', description='uncharacterized LOC388439', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377114, CurrentID=0, Status='live', Symbol='LOC105377114', Aliases='', description='uncharacterized LOC105377114', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=61284401, end_position_on_the_genomic_accession=61428677, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377115, CurrentID=0, Status='live', Symbol='LOC105377115', Aliases='', description='uncharacterized LOC105377115', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=61251598, end_position_on_the_genomic_accession=61315055, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373483, CurrentID=0, Status='live', Symbol='LOC105373483', Aliases='', description='uncharacterized LOC105373483', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7720340, end_position_on_the_genomic_accession=7723681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100134237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134237, CurrentID=0, Status='live', Symbol='LOC100134237', Aliases='', description='uncharacterized LOC100134237', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  652990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652990, CurrentID=0, Status='live', Symbol='LOC652990', Aliases='', description='uncharacterized LOC652990', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=44059750, end_position_on_the_genomic_accession=44061875, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126491, CurrentID=0, Status='live', Symbol='TRNAS-UGA', Aliases='', description='transfer RNA serine (anticodon UGA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100134240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134240, CurrentID=0, Status='live', Symbol='LOC100134240', Aliases='', description='uncharacterized LOC100134240', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377122, CurrentID=0, Status='live', Symbol='LOC105377122', Aliases='', description='uncharacterized LOC105377122', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64089957, end_position_on_the_genomic_accession=64091318, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377124, CurrentID=0, Status='live', Symbol='LOC105377124', Aliases='', description='uncharacterized LOC105377124', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64876395, end_position_on_the_genomic_accession=64988817, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  348210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=348210, CurrentID=0, Status='live', Symbol='LOC348210', Aliases='', description='mitochondrial ribosomal protein L57 pseudogene', other_designations='mitochondrial ribosomal protein 63 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34120489, end_position_on_the_genomic_accession=34122998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377128, CurrentID=0, Status='live', Symbol='LOC105377128', Aliases='', description='uncharacterized LOC105377128', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66038720, end_position_on_the_genomic_accession=66041999, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377132, CurrentID=0, Status='live', Symbol='LOC105377132', Aliases='', description='uncharacterized LOC105377132', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37175864, end_position_on_the_genomic_accession=37196722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377137, CurrentID=0, Status='live', Symbol='LOC105377137', Aliases='', description='uncharacterized LOC105377137', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43388599, end_position_on_the_genomic_accession=43392291, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131004, CurrentID=0, Status='live', Symbol='LOC100131004', Aliases='', description='uncharacterized LOC100131004', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=60217436, end_position_on_the_genomic_accession=60218137, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377139, CurrentID=0, Status='live', Symbol='LOC105377139', Aliases='', description='uncharacterized LOC105377139', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44172147, end_position_on_the_genomic_accession=44194634, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  257396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257396, CurrentID=0, Status='live', Symbol='LOC257396', Aliases='', description='uncharacterized LOC257396', other_designations='CTD-2366F13.1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53109842, end_position_on_the_genomic_accession=53115126, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  148170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148170, CurrentID=0, Status='live', Symbol='CDC42EP5', Aliases='Borg3, CEP5', description='CDC42 effector protein (Rho GTPase binding) 5', other_designations='binder of Rho GTPases 3', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54465025, end_position_on_the_genomic_accession=54473290, orientation='minus', exon_count=3, OMIM=609171, no_hdr0=''),
  106480253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480253, CurrentID=0, Status='live', Symbol='KLF2P1', Aliases='', description='Kruppel-like factor 2 pseudogene 1', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645978, CurrentID=0, Status='live', Symbol='LOC645978', Aliases='', description='alkaline ceramidase 2 pseudogene', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206410087, end_position_on_the_genomic_accession=206410909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377142, CurrentID=0, Status='live', Symbol='LOC105377142', Aliases='', description='uncharacterized LOC105377142', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66701054, end_position_on_the_genomic_accession=66718224, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377144, CurrentID=0, Status='live', Symbol='LOC105377144', Aliases='', description='uncharacterized LOC105377144', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66780077, end_position_on_the_genomic_accession=66972393, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131174, CurrentID=0, Status='live', Symbol='LOC100131174', Aliases='', description='uncharacterized LOC100131174', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=61172797, end_position_on_the_genomic_accession=61174714, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480254, CurrentID=0, Status='live', Symbol='HSPE1P6', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 6', other_designations='heat shock 10kDa protein 1 pseudogene 6', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377150, CurrentID=0, Status='live', Symbol='LOC105377150', Aliases='', description='uncharacterized LOC105377150', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70191367, end_position_on_the_genomic_accession=70249668, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  650623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650623, CurrentID=0, Status='live', Symbol='BEND3P3', Aliases='', description='BEN domain containing 3 pseudogene 3', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79682975, end_position_on_the_genomic_accession=79688897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480256, CurrentID=0, Status='live', Symbol='HSPE1P13', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 13', other_designations='heat shock 10kDa protein 1 pseudogene 13', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377159, CurrentID=0, Status='live', Symbol='LOC105377159', Aliases='', description='uncharacterized LOC105377159', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72312365, end_position_on_the_genomic_accession=72316968, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377165, CurrentID=0, Status='live', Symbol='LOC105377165', Aliases='', description='uncharacterized LOC105377165', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73813626, end_position_on_the_genomic_accession=73903392, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100271394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271394, CurrentID=0, Status='live', Symbol='RPL36P18', Aliases='RPL36_10_1804', description='ribosomal protein L36 pseudogene 18', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115198399, end_position_on_the_genomic_accession=115198800, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371670, CurrentID=0, Status='live', Symbol='LOC105371670', Aliases='', description='uncharacterized LOC105371670', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=194291217, end_position_on_the_genomic_accession=194299946, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377169, CurrentID=0, Status='live', Symbol='LOC105377169', Aliases='', description='uncharacterized LOC105377169', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75383653, end_position_on_the_genomic_accession=75393267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646629, CurrentID=0, Status='live', Symbol='LOC646629', Aliases='', description='zinc finger protein 470 pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24506864, end_position_on_the_genomic_accession=24508772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377171, CurrentID=0, Status='live', Symbol='LOC105377171', Aliases='', description='uncharacterized LOC105377171', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=77846644, end_position_on_the_genomic_accession=78029866, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100287212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287212, CurrentID=0, Status='live', Symbol='LOC100287212', Aliases='', description='chromosome 4 open reading frame 3 pseudogene', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86132624, end_position_on_the_genomic_accession=86135008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377175, CurrentID=0, Status='live', Symbol='LOC105377175', Aliases='', description='uncharacterized LOC105377175', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80764842, end_position_on_the_genomic_accession=80789378, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129112, CurrentID=0, Status='live', Symbol='LOC100129112', Aliases='', description='uncharacterized LOC100129112', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=57438250, end_position_on_the_genomic_accession=57438889, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377179, CurrentID=0, Status='live', Symbol='LOC105377179', Aliases='', description='uncharacterized LOC105377179', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=81922854, end_position_on_the_genomic_accession=81953255, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100190987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100190987, CurrentID=0, Status='live', Symbol='FDPSP10', Aliases='', description='farnesyl diphosphate synthase pseudogene 10', other_designations='farnesyl diphosphate synthase (farnesyl pyrophosphate synthetase, dimethylallyltranstransferase, geranyltranstransferase) pseudogene', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43140093, end_position_on_the_genomic_accession=43141267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377180, CurrentID=0, Status='live', Symbol='LOC105377180', Aliases='', description='uncharacterized LOC105377180', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=81986183, end_position_on_the_genomic_accession=82204637, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100134301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134301, CurrentID=0, Status='live', Symbol='LOC100134301', Aliases='', description='uncharacterized LOC100134301', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377183, CurrentID=0, Status='live', Symbol='LOC105377183', Aliases='', description='uncharacterized LOC105377183', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83412543, end_position_on_the_genomic_accession=83451661, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100287221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287221, CurrentID=0, Status='live', Symbol='LOC100287221', Aliases='', description='uncharacterized LOC100287221', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=88212051, end_position_on_the_genomic_accession=88212574, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377184, CurrentID=0, Status='live', Symbol='LOC105377184', Aliases='', description='uncharacterized LOC105377184', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19136209, end_position_on_the_genomic_accession=19136435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388514, CurrentID=0, Status='live', Symbol='CYP4F24P', Aliases='', description='cytochrome P450 family 4 subfamily F member 24, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 24 pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 24, pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15759334, end_position_on_the_genomic_accession=15779988, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105377190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377190, CurrentID=0, Status='live', Symbol='LOC105377190', Aliases='', description='uncharacterized LOC105377190', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21359596, end_position_on_the_genomic_accession=21360724, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377191, CurrentID=0, Status='live', Symbol='LOC105377191', Aliases='', description='uncharacterized LOC105377191', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21420300, end_position_on_the_genomic_accession=21442749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101290498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290498, CurrentID=0, Status='live', Symbol='AD17', Aliases='', description='Alzheimer disease 17', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615080, no_hdr0=''),
  101290499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290499, CurrentID=0, Status='live', Symbol='OPHLC', Aliases='C1DUPp31.3, DUP1p31.3', description='Omphalocele due to duplication of 1p31.3', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=164750, no_hdr0=''),
  101290500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290500, CurrentID=0, Status='live', Symbol='SNAP23P', Aliases='dJ155G6.3', description='synaptosome associated protein 23kDa pseudogene', other_designations='synaptosomal-associated protein, 23kD pseudogene|synaptosomal-associated protein, 23kDa pseudogene', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49038358, end_position_on_the_genomic_accession=49038609, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101290501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290501, CurrentID=0, Status='live', Symbol='TBCAP3', Aliases='', description='tubulin folding cofactor A pseudogene 3', other_designations='tubulin-specific chaperone a pseudogene 3', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98909381, end_position_on_the_genomic_accession=98909903, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101290502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290502, CurrentID=0, Status='live', Symbol='SUMO1P4', Aliases='', description='SUMO1 pseudogene 4', other_designations='', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49275026, end_position_on_the_genomic_accession=49276499, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101290503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290503, CurrentID=0, Status='live', Symbol='BANF1P5', Aliases='', description='barrier to autointegration factor 1 pseudogene 5', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107642388, end_position_on_the_genomic_accession=107643079, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101290504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290504, CurrentID=0, Status='live', Symbol='YWHAQP6', Aliases='', description='YWHAQ pseudogene 6', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141599244, end_position_on_the_genomic_accession=141600693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377196, CurrentID=0, Status='live', Symbol='LOC105377196', Aliases='', description='uncharacterized LOC105377196', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86904084, end_position_on_the_genomic_accession=86937149, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101290506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290506, CurrentID=0, Status='live', Symbol='LRRFIP1P1', Aliases='', description='leucine rich repeat (in FLII) interacting protein 1 pseudogene 1', other_designations='', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179257260, end_position_on_the_genomic_accession=179261567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101290508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101290508, CurrentID=0, Status='live', Symbol='YWHAQP9', Aliases='', description='YWHAQ pseudogene 9', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238106454, end_position_on_the_genomic_accession=238108590, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287234, CurrentID=0, Status='live', Symbol='RAB9AP5', Aliases='', description='RAB9A, member RAS oncogene family pseudogene 5', other_designations='RAB9B, member RAS oncogene family pseudogene 5', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23956569, end_position_on_the_genomic_accession=23960215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100134317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134317, CurrentID=0, Status='live', Symbol='LOC100134317', Aliases='', description='uncharacterized LOC100134317', other_designations='CTD-3162L10.1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36311343, end_position_on_the_genomic_accession=36312668, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377198, CurrentID=0, Status='live', Symbol='LOC105377198', Aliases='', description='uncharacterized LOC105377198', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87711673, end_position_on_the_genomic_accession=87740949, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376041, CurrentID=0, Status='live', Symbol='LOC105376041', Aliases='', description='uncharacterized LOC105376041', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38424611, end_position_on_the_genomic_accession=38432384, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129888, CurrentID=0, Status='live', Symbol='LOC100129888', Aliases='', description='uncharacterized LOC100129888', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=201476898, end_position_on_the_genomic_accession=201478120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377203, CurrentID=0, Status='live', Symbol='LOC105377203', Aliases='', description='uncharacterized LOC105377203', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42145451, end_position_on_the_genomic_accession=42149500, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377204, CurrentID=0, Status='live', Symbol='LOC105377204', Aliases='', description='uncharacterized LOC105377204', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45607784, end_position_on_the_genomic_accession=45609092, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377205, CurrentID=0, Status='live', Symbol='LOC105377205', Aliases='', description='uncharacterized LOC105377205', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49832412, end_position_on_the_genomic_accession=49857660, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377207, CurrentID=0, Status='live', Symbol='LOC105377207', Aliases='', description='uncharacterized LOC105377207', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75281933, end_position_on_the_genomic_accession=75342180, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377208, CurrentID=0, Status='live', Symbol='LOC105377208', Aliases='', description='uncharacterized LOC105377208', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52054057, end_position_on_the_genomic_accession=52058594, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377209, CurrentID=0, Status='live', Symbol='LOC105377209', Aliases='', description='uncharacterized LOC105377209', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52195501, end_position_on_the_genomic_accession=52265905, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927284, CurrentID=0, Status='live', Symbol='LOC101927284', Aliases='', description='uncharacterized LOC101927284', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94760941, end_position_on_the_genomic_accession=94935410, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377212, CurrentID=0, Status='live', Symbol='LOC105377212', Aliases='', description='uncharacterized LOC105377212', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63237522, end_position_on_the_genomic_accession=63242881, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377218, CurrentID=0, Status='live', Symbol='LOC105377218', Aliases='', description='uncharacterized LOC105377218', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17688837, end_position_on_the_genomic_accession=17689263, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377222, CurrentID=0, Status='live', Symbol='LOC105377222', Aliases='', description='uncharacterized LOC105377222', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18802540, end_position_on_the_genomic_accession=18836212, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  148252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148252, CurrentID=0, Status='live', Symbol='DIRAS1', Aliases='Di-Ras1, GBTS1, RIG', description='DIRAS family, GTP-binding RAS-like 1', other_designations='distinct subgroup of the Ras family member 1|ras-related inhibitor of cell growth|small GTP-binding tumor suppressor 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2714567, end_position_on_the_genomic_accession=2721392, orientation='minus', exon_count=3, OMIM=607862, no_hdr0=''),
  105377223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377223, CurrentID=0, Status='live', Symbol='LOC105377223', Aliases='', description='uncharacterized LOC105377223', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19553779, end_position_on_the_genomic_accession=19567251, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377224, CurrentID=0, Status='live', Symbol='LOC105377224', Aliases='', description='uncharacterized LOC105377224', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19744788, end_position_on_the_genomic_accession=19756924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377225, CurrentID=0, Status='live', Symbol='LOC105377225', Aliases='', description='uncharacterized LOC105377225', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21138588, end_position_on_the_genomic_accession=21170218, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377226, CurrentID=0, Status='live', Symbol='LOC105377226', Aliases='', description='uncharacterized LOC105377226', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3111310, end_position_on_the_genomic_accession=3115972, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480270, CurrentID=0, Status='live', Symbol='SNRPGP16', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 16', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  541472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541472, CurrentID=0, Status='live', Symbol='LOC541472', Aliases='', description='uncharacterized LOC541472', other_designations='', map_location='7p21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=22765014, end_position_on_the_genomic_accession=22767235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377227, CurrentID=0, Status='live', Symbol='LOC105377227', Aliases='', description='uncharacterized LOC105377227', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3803911, end_position_on_the_genomic_accession=3836025, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377228, CurrentID=0, Status='live', Symbol='LOC105377228', Aliases='', description='uncharacterized LOC105377228', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6262178, end_position_on_the_genomic_accession=6265007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377230, CurrentID=0, Status='live', Symbol='LOC105377230', Aliases='', description='uncharacterized LOC105377230', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8094272, end_position_on_the_genomic_accession=8126268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377231, CurrentID=0, Status='live', Symbol='LOC105377231', Aliases='', description='uncharacterized LOC105377231', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8395047, end_position_on_the_genomic_accession=8407314, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377232, CurrentID=0, Status='live', Symbol='LOC105377232', Aliases='', description='uncharacterized LOC105377232', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8549057, end_position_on_the_genomic_accession=8551268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377233, CurrentID=0, Status='live', Symbol='LOC105377233', Aliases='', description='uncharacterized LOC105377233', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8673875, end_position_on_the_genomic_accession=8677525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377236, CurrentID=0, Status='live', Symbol='LOC105377236', Aliases='', description='uncharacterized LOC105377236', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21809153, end_position_on_the_genomic_accession=21823816, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377238, CurrentID=0, Status='live', Symbol='LOC105377238', Aliases='', description='uncharacterized LOC105377238', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23106397, end_position_on_the_genomic_accession=23109726, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377239, CurrentID=0, Status='live', Symbol='LOC105377239', Aliases='', description='uncharacterized LOC105377239', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23310985, end_position_on_the_genomic_accession=23314314, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377241, CurrentID=0, Status='live', Symbol='LOC105377241', Aliases='', description='uncharacterized LOC105377241', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24740119, end_position_on_the_genomic_accession=24743448, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100134362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134362, CurrentID=0, Status='live', Symbol='LOC100134362', Aliases='', description='60S ribosomal protein L19-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131892862, end_position_on_the_genomic_accession=131895762, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377243, CurrentID=0, Status='live', Symbol='LOC105377243', Aliases='', description='uncharacterized LOC105377243', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26196211, end_position_on_the_genomic_accession=26235345, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388572, CurrentID=0, Status='live', Symbol='LOC388572', Aliases='', description='uncharacterized LOC388572', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133725, end_position_on_the_genomic_accession=185265, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  105377246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377246, CurrentID=0, Status='live', Symbol='LOC105377246', Aliases='', description='uncharacterized LOC105377246', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=58976071, end_position_on_the_genomic_accession=58984194, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371686, CurrentID=0, Status='live', Symbol='LOC105371686', Aliases='', description='uncharacterized LOC105371686', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202596977, end_position_on_the_genomic_accession=202602000, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100134368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134368, CurrentID=0, Status='live', Symbol='LOC100134368', Aliases='', description='uncharacterized LOC100134368', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=382241, end_position_on_the_genomic_accession=392960, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377249, CurrentID=0, Status='live', Symbol='LOC105377249', Aliases='', description='uncharacterized LOC105377249', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=61143688, end_position_on_the_genomic_accession=61173212, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377251, CurrentID=0, Status='live', Symbol='LOC105377251', Aliases='', description='uncharacterized LOC105377251', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62346532, end_position_on_the_genomic_accession=62355105, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377253, CurrentID=0, Status='live', Symbol='LOC105377253', Aliases='', description='uncharacterized LOC105377253', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=63465832, end_position_on_the_genomic_accession=63513435, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100271222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271222, CurrentID=0, Status='live', Symbol='RPL7P41', Aliases='RPL7_21_1238', description='ribosomal protein L7 pseudogene 41', other_designations='', map_location='12q13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52885759, end_position_on_the_genomic_accession=52886409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480276, CurrentID=0, Status='live', Symbol='UBE2Q2P9', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 9', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 9', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377257, CurrentID=0, Status='live', Symbol='LOC105377257', Aliases='', description='uncharacterized LOC105377257', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65700365, end_position_on_the_genomic_accession=65708264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377260, CurrentID=0, Status='live', Symbol='LOC105377260', Aliases='', description='uncharacterized LOC105377260', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65944939, end_position_on_the_genomic_accession=65948261, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480324, CurrentID=0, Status='live', Symbol='BOLA3P4', Aliases='', description='bolA family member 3 pseudogene 4', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377263, CurrentID=0, Status='live', Symbol='LOC105377263', Aliases='', description='uncharacterized LOC105377263', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68077467, end_position_on_the_genomic_accession=68080981, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271398, CurrentID=0, Status='live', Symbol='RPL7AP44', Aliases='RPL7A_21_1001', description='ribosomal protein L7a pseudogene 44', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106978740, end_position_on_the_genomic_accession=106979576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377267, CurrentID=0, Status='live', Symbol='LOC105377267', Aliases='', description='uncharacterized LOC105377267', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69194353, end_position_on_the_genomic_accession=69200539, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100134391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134391, CurrentID=0, Status='live', Symbol='LOC100134391', Aliases='', description='uncharacterized LOC100134391', other_designations='CTD-2532D12.4', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73737854, end_position_on_the_genomic_accession=73756831, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106480279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480279, CurrentID=0, Status='live', Symbol='CRB3P1', Aliases='', description='crumbs family member 3 pseudogene 1', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  650747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650747, CurrentID=0, Status='live', Symbol='IMPA1P', Aliases='', description='inositol(myo)-1(or 4)-monophosphatase 1 pseudogene', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81604561, end_position_on_the_genomic_accession=81626869, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377277, CurrentID=0, Status='live', Symbol='LOC105377277', Aliases='', description='uncharacterized LOC105377277', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74115444, end_position_on_the_genomic_accession=74138287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377278, CurrentID=0, Status='live', Symbol='LOC105377278', Aliases='', description='uncharacterized LOC105377278', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74831127, end_position_on_the_genomic_accession=74835652, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377282, CurrentID=0, Status='live', Symbol='LOC105377282', Aliases='', description='uncharacterized LOC105377282', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74841556, end_position_on_the_genomic_accession=74881197, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  442517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442517, CurrentID=0, Status='live', Symbol='LOC442517', Aliases='', description='ATP binding cassette subfamily E member 1 pseudogene', other_designations='ATP-binding cassette, sub-family E (OABP), member 1 pseudogene', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23562464, end_position_on_the_genomic_accession=23564260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377283, CurrentID=0, Status='live', Symbol='LOC105377283', Aliases='', description='uncharacterized LOC105377283', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75081925, end_position_on_the_genomic_accession=75127205, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377285, CurrentID=0, Status='live', Symbol='LOC105377285', Aliases='', description='uncharacterized LOC105377285', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75878149, end_position_on_the_genomic_accession=75888466, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271399, CurrentID=0, Status='live', Symbol='RPL7AP71', Aliases='RPL7A_33_1780', description='ribosomal protein L7a pseudogene 71', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54223303, end_position_on_the_genomic_accession=54224069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377293, CurrentID=0, Status='live', Symbol='LOC105377293', Aliases='', description='uncharacterized LOC105377293', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76971148, end_position_on_the_genomic_accession=76987802, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  400618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400618, CurrentID=0, Status='live', Symbol='SOX9-AS1', Aliases='', description='SOX9 antisense RNA 1', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72071042, end_position_on_the_genomic_accession=72120793, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100287335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287335, CurrentID=0, Status='live', Symbol='LOC100287335', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70858871, end_position_on_the_genomic_accession=70859425, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  450232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=450232, CurrentID=0, Status='live', Symbol='PHF5CP', Aliases='PHF5AP1', description='PHD finger protein 5C pseudogene', other_designations='PHD finger protein 5A pseudogene 1|PHF5A pseudogene 1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19587606, end_position_on_the_genomic_accession=19588439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377299, CurrentID=0, Status='live', Symbol='LOC105377299', Aliases='', description='uncharacterized LOC105377299', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78787005, end_position_on_the_genomic_accession=78811325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377302, CurrentID=0, Status='live', Symbol='LOC105377302', Aliases='', description='uncharacterized LOC105377302', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79696907, end_position_on_the_genomic_accession=79744315, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100134423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134423, CurrentID=0, Status='live', Symbol='LOC100134423', Aliases='', description='uncharacterized LOC100134423', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377305, CurrentID=0, Status='live', Symbol='LOC105377305', Aliases='', description='uncharacterized LOC105377305', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80488785, end_position_on_the_genomic_accession=80491550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480286, CurrentID=0, Status='live', Symbol='BNIP3P11', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 11', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129085, CurrentID=0, Status='live', Symbol='LOC100129085', Aliases='', description='selenoprotein K pseudogene', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37693031, end_position_on_the_genomic_accession=37693591, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372618, CurrentID=0, Status='live', Symbol='LOC105372618', Aliases='', description='uncharacterized LOC105372618', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40458151, end_position_on_the_genomic_accession=40466842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377316, CurrentID=0, Status='live', Symbol='LOC105377316', Aliases='', description='uncharacterized LOC105377316', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84371614, end_position_on_the_genomic_accession=84374900, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377318, CurrentID=0, Status='live', Symbol='LOC105377318', Aliases='', description='uncharacterized LOC105377318', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84480883, end_position_on_the_genomic_accession=84482987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105747689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105747689, CurrentID=0, Status='live', Symbol='LOC105747689', Aliases='', description='uncharacterized LOC105747689', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287357, CurrentID=0, Status='live', Symbol='LOC100287357', Aliases='', description='coxsackie virus and adenovirus receptor pseudogene', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20803791, end_position_on_the_genomic_accession=20804886, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377321, CurrentID=0, Status='live', Symbol='LOC105377321', Aliases='', description='uncharacterized LOC105377321', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87462991, end_position_on_the_genomic_accession=87468224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  650794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650794, CurrentID=0, Status='live', Symbol='MIPEPP3', Aliases='', description='mitochondrial intermediate peptidase pseudogene 3', other_designations='long intergenic non-protein coding RNA 539', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21298125, end_position_on_the_genomic_accession=21392923, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377324, CurrentID=0, Status='live', Symbol='LOC105377324', Aliases='', description='uncharacterized LOC105377324', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87729837, end_position_on_the_genomic_accession=87741413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377325, CurrentID=0, Status='live', Symbol='LOC105377325', Aliases='', description='uncharacterized LOC105377325', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87999366, end_position_on_the_genomic_accession=88003887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377327, CurrentID=0, Status='live', Symbol='LOC105377327', Aliases='', description='uncharacterized LOC105377327', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88929976, end_position_on_the_genomic_accession=88957527, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480291, CurrentID=0, Status='live', Symbol='SNRPCP16', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 16', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  541578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=541578, CurrentID=0, Status='live', Symbol='CXorf40B', Aliases='', description='chromosome X open reading frame 40B', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149924157, end_position_on_the_genomic_accession=149938850, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377334, CurrentID=0, Status='live', Symbol='LOC105377334', Aliases='', description='uncharacterized LOC105377334', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92635149, end_position_on_the_genomic_accession=92673772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126520, CurrentID=0, Status='live', Symbol='PLK5', Aliases='PLK-5P, SgK384ps, PLK5', description='polo-like kinase 5', other_designations='polo-like kinase 5, pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1524074, end_position_on_the_genomic_accession=1535456, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105372619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372619, CurrentID=0, Status='live', Symbol='LOC105372619', Aliases='', description='uncharacterized LOC105372619', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40531464, end_position_on_the_genomic_accession=40543923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377339, CurrentID=0, Status='live', Symbol='LOC105377339', Aliases='', description='uncharacterized LOC105377339', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=96850387, end_position_on_the_genomic_accession=96854590, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927605, CurrentID=0, Status='live', Symbol='LOC101927605', Aliases='', description='uncharacterized LOC101927605', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=60359782, end_position_on_the_genomic_accession=60499355, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377342, CurrentID=0, Status='live', Symbol='LOC105377342', Aliases='', description='uncharacterized LOC105377342', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98658736, end_position_on_the_genomic_accession=98678619, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376070, CurrentID=0, Status='live', Symbol='LOC105376070', Aliases='', description='uncharacterized LOC105376070', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68371722, end_position_on_the_genomic_accession=68406622, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  282786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282786, CurrentID=0, Status='live', Symbol='OR8G7P', Aliases='', description='olfactory receptor family 8 subfamily G member 7 pseudogene', other_designations='olfactory receptor, family 8, subfamily G, member 7 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124238422, end_position_on_the_genomic_accession=124239577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100270859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270859, CurrentID=0, Status='live', Symbol='RPS2P19', Aliases='RPS2_4_428', description='ribosomal protein S2 pseudogene 19', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159386756, end_position_on_the_genomic_accession=159387433, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480294, CurrentID=0, Status='live', Symbol='NENFP1', Aliases='', description='neudesin neurotrophic factor pseudogene 1', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270844, CurrentID=0, Status='live', Symbol='RPL6P5', Aliases='RPL6_4_280', description='ribosomal protein L6 pseudogene 5', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145337233, end_position_on_the_genomic_accession=145338063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  497048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497048, CurrentID=0, Status='live', Symbol='KU-MEL-3', Aliases='', description='uncharacterized LOC497048', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4610565, end_position_on_the_genomic_accession=4611919, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  497049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497049, CurrentID=0, Status='live', Symbol='FLJ25758', Aliases='', description='MAP/microtubule affinity-regulating kinase 1 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7004953, end_position_on_the_genomic_accession=7006415, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377349, CurrentID=0, Status='live', Symbol='LOC105377349', Aliases='', description='uncharacterized LOC105377349', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=103961620, end_position_on_the_genomic_accession=104036522, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480295, CurrentID=0, Status='live', Symbol='PDCL2P2', Aliases='', description='phosducin-like 2 pseudogene 2', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377352, CurrentID=0, Status='live', Symbol='LOC105377352', Aliases='', description='uncharacterized LOC105377352', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105537443, end_position_on_the_genomic_accession=105552487, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480296, CurrentID=0, Status='live', Symbol='MTCO2P4', Aliases='', description='MT-CO2 pseudogene 4', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388685, CurrentID=0, Status='live', Symbol='LINC01138', Aliases='LINC00875', description='long intergenic non-protein coding RNA 1138', other_designations='long intergenic non-protein coding RNA 875', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148432959, end_position_on_the_genomic_accession=148459920, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376073, CurrentID=0, Status='live', Symbol='LOC105376073', Aliases='', description='uncharacterized LOC105376073', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68823926, end_position_on_the_genomic_accession=68835766, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103910, CurrentID=0, Status='live', Symbol='MYL12B', Aliases='MLC-B, MRLC2', description='myosin light chain 12B', other_designations='MLC-2|MLC-2A|MLC20|SHUJUN-1|myosin regulatory light chain 2|myosin regulatory light chain 2-B, smooth muscle isoform|myosin regulatory light chain 20 kDa|myosin regulatory light chain MRLC2|myosin, light chain 12B, regulatory', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3262113, end_position_on_the_genomic_accession=3278284, orientation='plus', exon_count=5, OMIM=609211, no_hdr0=''),
  103911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103911, CurrentID=0, Status='live', Symbol='RPS27AP1', Aliases='RPS27AP, RPS27A_6_1513', description='ribosomal protein S27a pseudogene 1', other_designations='ubiquitin and ribosomal protein S27a precursor pseudogene|ubiquitin carboxyl extension protein 80 pseudogene', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10258713, end_position_on_the_genomic_accession=10259262, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480297, CurrentID=0, Status='live', Symbol='THEM7P', Aliases='', description='thioesterase superfamily member 7, pseudogene', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480328, CurrentID=0, Status='live', Symbol='TRIM80P', Aliases='', description='tripartite motif containing 80, pseudogene', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377363, CurrentID=0, Status='live', Symbol='LOC105377363', Aliases='', description='uncharacterized LOC105377363', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110607975, end_position_on_the_genomic_accession=110613503, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377364, CurrentID=0, Status='live', Symbol='LOC105377364', Aliases='', description='uncharacterized LOC105377364', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110986190, end_position_on_the_genomic_accession=111013982, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287402, CurrentID=0, Status='live', Symbol='LOC100287402', Aliases='', description='uncharacterized LOC100287402', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186596964, end_position_on_the_genomic_accession=186597281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377365, CurrentID=0, Status='live', Symbol='LOC105377365', Aliases='', description='uncharacterized LOC105377365', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111632225, end_position_on_the_genomic_accession=111640477, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377367, CurrentID=0, Status='live', Symbol='LOC105377367', Aliases='', description='uncharacterized LOC105377367', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111705581, end_position_on_the_genomic_accession=111715305, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105616916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105616916, CurrentID=0, Status='live', Symbol='LINC01455', Aliases='MDONA, TCONS_00009433', description='long intergenic non-protein coding RNA 1455', other_designations='MTX3 downstream antisense RNA', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377369, CurrentID=0, Status='live', Symbol='LOC105377369', Aliases='', description='uncharacterized LOC105377369', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111846142, end_position_on_the_genomic_accession=111930877, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  282791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282791, CurrentID=0, Status='live', Symbol='OR8T1P', Aliases='', description='olfactory receptor family 8 subfamily T member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily T, member 1 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48442032, end_position_on_the_genomic_accession=48442946, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414765, CurrentID=0, Status='live', Symbol='HCG25', Aliases='dJ1033B10.16', description='HLA complex group 25 (non-protein coding)', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33249536, end_position_on_the_genomic_accession=33254890, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377371, CurrentID=0, Status='live', Symbol='LOC105377371', Aliases='', description='uncharacterized LOC105377371', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112229561, end_position_on_the_genomic_accession=112232381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  497188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497188, CurrentID=0, Status='live', Symbol='GLC1I', Aliases='', description='glaucoma 1, open angle, I', other_designations='', map_location='15q11-q13', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609745, no_hdr0=''),
  497189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497189, CurrentID=0, Status='live', Symbol='TIFAB', Aliases='', description='TRAF-interacting protein with forkhead-associated domain, family member B', other_designations='TIFA-like protein|TIFA-related protein TIFAB', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135448868, end_position_on_the_genomic_accession=135452399, orientation='minus', exon_count=2, OMIM=612663, no_hdr0=''),
  497190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497190, CurrentID=0, Status='live', Symbol='CLEC18B', Aliases='MRCL2', description='C-type lectin domain family 18 member B', other_designations='C-type lectin domain family 18, member B|mannose receptor-like 2|mannose receptor-like protein 1|secretory protein LOC497190', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74408631, end_position_on_the_genomic_accession=74422248, orientation='minus', exon_count=15, OMIM=616572, no_hdr0=''),
  257629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257629, CurrentID=0, Status='live', Symbol='ANKS4B', Aliases='HARP', description='ankyrin repeat and sterile alpha motif domain containing 4B', other_designations='harmonin-interacting ankyrin repeat-containing protein|harmonin-interacting ankyrin-repeat containing protein', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21233695, end_position_on_the_genomic_accession=21252429, orientation='plus', exon_count=2, OMIM=609901, no_hdr0=''),
  100287413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287413, CurrentID=0, Status='live', Symbol='LOC100287413', Aliases='', description='uncharacterized LOC100287413', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58905310, end_position_on_the_genomic_accession=58908598, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480300, CurrentID=0, Status='live', Symbol='LLPHP2', Aliases='', description='LLP homolog, long-term synaptic facilitation (Aplysia) pseudogene 2', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129923, CurrentID=0, Status='live', Symbol='LOC100129923', Aliases='', description='myoneurin pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21184341, end_position_on_the_genomic_accession=21184960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  497231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497231, CurrentID=0, Status='live', Symbol='CLN9', Aliases='', description='ceroid-lipofuscinosis, neuronal 9', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609055, no_hdr0=''),
  105377379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377379, CurrentID=0, Status='live', Symbol='LOC105377379', Aliases='', description='uncharacterized LOC105377379', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114458112, end_position_on_the_genomic_accession=114491099, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105616981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105616981, CurrentID=0, Status='live', Symbol='LOC105616981', Aliases='', description='uncharacterized LOC105616981', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105616982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105616982, CurrentID=0, Status='live', Symbol='LINC01047', Aliases='', description='long intergenic non-protein coding RNA 1047', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378731, CurrentID=0, Status='live', Symbol='LOC105378731', Aliases='', description='uncharacterized LOC105378731', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53329815, end_position_on_the_genomic_accession=53346230, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100287419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287419, CurrentID=0, Status='live', Symbol='LOC100287419', Aliases='', description='cyclin-dependent kinase 8 pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112932625, end_position_on_the_genomic_accession=112933799, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  497256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497256, CurrentID=0, Status='live', Symbol='LOC497256', Aliases='', description='uncharacterized LOC497256', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87844517, end_position_on_the_genomic_accession=87847764, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  497258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497258, CurrentID=0, Status='live', Symbol='BDNF-AS', Aliases='ANTI-BDNF, BDNF1, BDNFOS, NCRNA00049, BDNF-AS', description='BDNF antisense RNA', other_designations='BDNF antisense RNA (non-protein coding)|BDNF antisense RNA 1 (non-protein coding)|BDNF opposite strand (non-protein coding)|brain-derived neurotrophic factor opposite strand', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27506852, end_position_on_the_genomic_accession=27698171, orientation='plus', exon_count=10, OMIM=611468, no_hdr0=''),
  497259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497259, CurrentID=0, Status='live', Symbol='C18orf61', Aliases='', description='uncharacterized LOC497259', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12200779, end_position_on_the_genomic_accession=12224711, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418493, CurrentID=0, Status='live', Symbol='LOC100418493', Aliases='', description='thymopoietin pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377385, CurrentID=0, Status='live', Symbol='LOC105377385', Aliases='', description='uncharacterized LOC105377385', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116663938, end_position_on_the_genomic_accession=116667507, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371748, CurrentID=0, Status='live', Symbol='LOC105371748', Aliases='', description='uncharacterized LOC105371748', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36483242, end_position_on_the_genomic_accession=36486382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377386, CurrentID=0, Status='live', Symbol='LOC105377386', Aliases='', description='uncharacterized LOC105377386', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116770919, end_position_on_the_genomic_accession=116774231, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377387, CurrentID=0, Status='live', Symbol='LOC105377387', Aliases='', description='uncharacterized LOC105377387', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117049110, end_position_on_the_genomic_accession=117073107, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100287425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287425, CurrentID=0, Status='live', Symbol='LOC100287425', Aliases='', description='chromosome 7 open reading frame 73 pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202239459, end_position_on_the_genomic_accession=202240027, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418498, CurrentID=0, Status='live', Symbol='LOC100418498', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287427, CurrentID=0, Status='live', Symbol='LOC100287427', Aliases='', description='chromosome 18 open reading frame 21 pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83377263, end_position_on_the_genomic_accession=83378143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282795, CurrentID=0, Status='live', Symbol='OR11P1P', Aliases='', description='olfactory receptor family 11 subfamily P member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily P, member 1 pseudogene', map_location='14q11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20183925, end_position_on_the_genomic_accession=20184298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377391, CurrentID=0, Status='live', Symbol='LOC105377391', Aliases='', description='uncharacterized LOC105377391', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117834400, end_position_on_the_genomic_accession=117869948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287430, CurrentID=0, Status='live', Symbol='LOC100287430', Aliases='', description='ribosomal protein S5 pseudogene', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116258445, end_position_on_the_genomic_accession=116259187, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418503, CurrentID=0, Status='live', Symbol='LOC100418503', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377394, CurrentID=0, Status='live', Symbol='LOC105377394', Aliases='', description='uncharacterized LOC105377394', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118961071, end_position_on_the_genomic_accession=118984749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418504, CurrentID=0, Status='live', Symbol='LOC100418504', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375240, CurrentID=0, Status='live', Symbol='LOC105375240', Aliases='', description='uncharacterized LOC105375240', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39551638, end_position_on_the_genomic_accession=39566353, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377397, CurrentID=0, Status='live', Symbol='LOC105377397', Aliases='', description='uncharacterized LOC105377397', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119689433, end_position_on_the_genomic_accession=119693463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418510, CurrentID=0, Status='live', Symbol='LOC100418510', Aliases='', description='MAS1 proto-oncogene like, G protein-coupled receptor pseudogene', other_designations='MAS1 oncogene-like pseudogene', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377402, CurrentID=0, Status='live', Symbol='LOC105377402', Aliases='', description='uncharacterized LOC105377402', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121642879, end_position_on_the_genomic_accession=121650651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418513, CurrentID=0, Status='live', Symbol='LOC100418513', Aliases='', description='lysophospholipase II pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388734, CurrentID=0, Status='live', Symbol='LOC388734', Aliases='', description='PRELI domain containing 1 pseudogene', other_designations='PX19 pseudogene', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211207064, end_position_on_the_genomic_accession=211208286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480306, CurrentID=0, Status='live', Symbol='AKIRIN1P2', Aliases='', description='akirin 1 pseudogene 2', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376083, CurrentID=0, Status='live', Symbol='LOC105376083', Aliases='', description='uncharacterized LOC105376083', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73411654, end_position_on_the_genomic_accession=73430915, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130657, CurrentID=0, Status='live', Symbol='LOC100130657', Aliases='', description='uncharacterized LOC100130657', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=63677569, end_position_on_the_genomic_accession=63677958, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106480307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480307, CurrentID=0, Status='live', Symbol='CCND3P2', Aliases='', description='cyclin D3 pseudogene 2', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  340569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=340569, CurrentID=0, Status='live', Symbol='LOC340569', Aliases='', description='origin recognition complex subunit 1 pseudogene', other_designations='origin recognition complex, subunit 1 pseudogene|origin recognition complex, subunit 1-like (yeast) pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27893592, end_position_on_the_genomic_accession=27894028, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505381, CurrentID=0, Status='live', Symbol='MMD2', Aliases='', description='Miyoshi muscular dystrophy 2', other_designations='', map_location='10p', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613318, no_hdr0=''),
  100505383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505383, CurrentID=0, Status='live', Symbol='BPIFB5P', Aliases='', description='BPI fold containing family B member 5, pseudogene', other_designations='BPI fold containing family B, member 5, pseudogene|long palate lung and nasal epithelium protein 5 pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33316869, end_position_on_the_genomic_accession=33329562, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376084, CurrentID=0, Status='live', Symbol='LOC105376084', Aliases='', description='uncharacterized LOC105376084', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73411629, end_position_on_the_genomic_accession=73452886, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100505385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505385, CurrentID=0, Status='live', Symbol='IQCJ-SCHIP1', Aliases='', description='IQCJ-SCHIP1 readthrough', other_designations='IQ motif containing J-schwannomin interacting protein 1 read-through transcript', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159069252, end_position_on_the_genomic_accession=159897366, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105377415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377415, CurrentID=0, Status='live', Symbol='LOC105377415', Aliases='', description='uncharacterized LOC105377415', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127827438, end_position_on_the_genomic_accession=127829797, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505389, CurrentID=0, Status='live', Symbol='DELXP21', Aliases='CXDELp21', description='Chromosome Xp21 deletion syndrome', other_designations='', map_location='Xp21', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300679, no_hdr0=''),
  100505390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505390, CurrentID=0, Status='live', Symbol='AAA2', Aliases='', description='Aortic aneurysm, familial abdominal 2', other_designations='', map_location='4q31', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609782, no_hdr0=''),
  100505391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505391, CurrentID=0, Status='live', Symbol='DEL6Q24Q25', Aliases='C6DELq25q25', description='Chromosome 6q25-q25 deletion syndrome', other_designations='', map_location='6q24-q25', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612863, no_hdr0=''),
  100505392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505392, CurrentID=0, Status='live', Symbol='DEL14Q11Q22', Aliases='C14DELq11q22', description='Chromosome 14q11-q22 deletion syndrome', other_designations='', map_location='14q11-q22', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613457, no_hdr0=''),
  100505393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505393, CurrentID=0, Status='live', Symbol='DUP16P13.3', Aliases='C16DUPq13.3', description='Chromosome 16p13.3 duplication syndrome', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613458, no_hdr0=''),
  100505394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505394, CurrentID=0, Status='live', Symbol='EDSS2', Aliases='', description='Ectodermal dysplasia-syndactyly syndrome 2', other_designations='', map_location='7p21.2-p14.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613576, no_hdr0=''),
  100129931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129931, CurrentID=0, Status='live', Symbol='LOC100129931', Aliases='', description='uncharacterized LOC100129931', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7030554, end_position_on_the_genomic_accession=7046231, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100270670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270670, CurrentID=0, Status='live', Symbol='LOC100270670', Aliases='', description='chloride channel, nucleotide-sensitive, 1A pseudogene', other_designations='', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151777906, end_position_on_the_genomic_accession=151778630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418533, CurrentID=0, Status='live', Symbol='LOC100418533', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377425, CurrentID=0, Status='live', Symbol='LOC105377425', Aliases='', description='uncharacterized LOC105377425', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131804134, end_position_on_the_genomic_accession=131837555, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100287463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287463, CurrentID=0, Status='live', Symbol='LOC100287463', Aliases='', description='uncharacterized LOC100287463', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9009656, end_position_on_the_genomic_accession=9020058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377429, CurrentID=0, Status='live', Symbol='LOC105377429', Aliases='', description='uncharacterized LOC105377429', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132531090, end_position_on_the_genomic_accession=132547319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505474, CurrentID=0, Status='live', Symbol='LINC01539', Aliases='', description='long intergenic non-protein coding RNA 1539', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56083356, end_position_on_the_genomic_accession=56137536, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100505475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505475, CurrentID=0, Status='live', Symbol='SYNE1-AS1', Aliases='', description='SYNE1 antisense RNA 1', other_designations='SYNE1 antisense RNA 1 (non-protein coding)', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152380530, end_position_on_the_genomic_accession=152381713, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505477, CurrentID=0, Status='live', Symbol='LINC01602', Aliases='T1560', description='long intergenic non-protein coding RNA 1602', other_designations='uncharacterized gene T1560', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57978358, end_position_on_the_genomic_accession=57984126, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505478, CurrentID=0, Status='live', Symbol='LOC100505478', Aliases='', description='uncharacterized LOC100505478', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114666434, end_position_on_the_genomic_accession=114682089, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100418540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418540, CurrentID=0, Status='live', Symbol='LOC100418540', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505483, CurrentID=0, Status='live', Symbol='PRKAG2-AS1', Aliases='', description='PRKAG2 antisense RNA 1', other_designations='PRKAG2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151877042, end_position_on_the_genomic_accession=151879223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480311, CurrentID=0, Status='live', Symbol='RNY1P14', Aliases='', description='RNA, Ro-associated Y1 pseudogene 14', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418542, CurrentID=0, Status='live', Symbol='LOC100418542', Aliases='', description='ribosomal protein L13 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505494, CurrentID=0, Status='live', Symbol='ANKRD10-IT1', Aliases='', description='ANKRD10 intronic transcript 1', other_designations='ANKRD10 intronic transcript 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=111551120, end_position_on_the_genomic_accession=111554449, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505495, CurrentID=0, Status='live', Symbol='PCAT19', Aliases='LINC01190', description='prostate cancer associated transcript 19 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1190', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41454169, end_position_on_the_genomic_accession=41500644, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505498, CurrentID=0, Status='live', Symbol='LOC100505498', Aliases='', description='uncharacterized LOC100505498', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145023074, end_position_on_the_genomic_accession=145182645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377434, CurrentID=0, Status='live', Symbol='LOC105377434', Aliases='', description='uncharacterized LOC105377434', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134382712, end_position_on_the_genomic_accession=134457847, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505502, CurrentID=0, Status='live', Symbol='LOC100505502', Aliases='', description='putative UPF0607 protein FLJ37424', other_designations='putative UPF0607 protein FLJ37424-like', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31362778, end_position_on_the_genomic_accession=31364077, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100505504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505504, CurrentID=0, Status='live', Symbol='LOC100505504', Aliases='', description='uncharacterized LOC100505504', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66409274, end_position_on_the_genomic_accession=66417137, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418544, CurrentID=0, Status='live', Symbol='LOC100418544', Aliases='', description='ribosomal protein L7a pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505506, CurrentID=0, Status='live', Symbol='BANF1P3', Aliases='', description='barrier to autointegration factor 1 pseudogene 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230724752, end_position_on_the_genomic_accession=230725479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927293, CurrentID=0, Status='live', Symbol='LOC101927293', Aliases='', description='uncharacterized LOC101927293', other_designations='XXbac-BPG55C20.7', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57961548, end_position_on_the_genomic_accession=58096528, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100505515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505515, CurrentID=0, Status='live', Symbol='LOC100505515', Aliases='', description='uncharacterized LOC100505515', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12950340, end_position_on_the_genomic_accession=12951627, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418546, CurrentID=0, Status='live', Symbol='LOC100418546', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505518, CurrentID=0, Status='live', Symbol='EDNRB-AS1', Aliases='', description='EDNRB antisense RNA 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77818937, end_position_on_the_genomic_accession=77908442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100505524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505524, CurrentID=0, Status='live', Symbol='LOC100505524', Aliases='', description='chromosome 5 open reading frame 63 pseudogene', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66454089, end_position_on_the_genomic_accession=66454740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505530, CurrentID=0, Status='live', Symbol='LOC100505530', Aliases='', description='uncharacterized LOC100505530', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37507348, end_position_on_the_genomic_accession=37536280, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505532, CurrentID=0, Status='live', Symbol='LINC01301', Aliases='', description='long intergenic non-protein coding RNA 1301', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60402171, end_position_on_the_genomic_accession=60413818, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505534, CurrentID=0, Status='live', Symbol='LOC100505534', Aliases='', description='uncharacterized LOC100505534', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39921025, end_position_on_the_genomic_accession=39925880, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418549, CurrentID=0, Status='live', Symbol='LOC100418549', Aliases='', description='zinc finger protein 93 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505536, CurrentID=0, Status='live', Symbol='ISM1-AS1', Aliases='', description='ISM1 antisense RNA 1', other_designations='ISM1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13237420, end_position_on_the_genomic_accession=13239687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505538, CurrentID=0, Status='live', Symbol='RBM26-AS1', Aliases='', description='RBM26 antisense RNA 1', other_designations='RBM26 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79406309, end_position_on_the_genomic_accession=79424336, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100505540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505540, CurrentID=0, Status='live', Symbol='SLIT1-AS1', Aliases='', description='SLIT1 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97102756, end_position_on_the_genomic_accession=97103747, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418550, CurrentID=0, Status='live', Symbol='LOC100418550', Aliases='', description='zinc finger protein 93 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377441, CurrentID=0, Status='live', Symbol='LOC105377441', Aliases='', description='uncharacterized LOC105377441', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=136796675, end_position_on_the_genomic_accession=136967318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505545, CurrentID=0, Status='live', Symbol='LINC01095', Aliases='', description='long intergenic non-protein coding RNA 1095', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146109455, end_position_on_the_genomic_accession=146121913, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129936, CurrentID=0, Status='live', Symbol='LOC100129936', Aliases='', description='uncharacterized LOC100129936', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=25397864, end_position_on_the_genomic_accession=25402443, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100505547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505547, CurrentID=0, Status='live', Symbol='LINC01208', Aliases='', description='long intergenic non-protein coding RNA 1208', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176604148, end_position_on_the_genomic_accession=176635532, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505549, CurrentID=0, Status='live', Symbol='LOC100505549', Aliases='', description='uncharacterized LOC100505549', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57630302, end_position_on_the_genomic_accession=57669276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372761, CurrentID=0, Status='live', Symbol='LOC105372761', Aliases='', description='uncharacterized LOC105372761', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27375371, end_position_on_the_genomic_accession=27395627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505555, CurrentID=0, Status='live', Symbol='LOC100505555', Aliases='', description='uncharacterized LOC100505555', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9787406, end_position_on_the_genomic_accession=9793381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505557, CurrentID=0, Status='live', Symbol='GSTM5P1', Aliases='GST1L, GSTM1L', description='glutathione S-transferase mu 5 pseudogene 1', other_designations='glutathione S-transferase M1-like|glutathione S-transferase mu 1-like', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12257376, end_position_on_the_genomic_accession=12258482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505561, CurrentID=0, Status='live', Symbol='LINC00866', Aliases='', description='long intergenic non-protein coding RNA 866', other_designations='', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97828478, end_position_on_the_genomic_accession=97849798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505564, CurrentID=0, Status='live', Symbol='LOC100505564', Aliases='', description='uncharacterized LOC100505564', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10940719, end_position_on_the_genomic_accession=10943021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418554, CurrentID=0, Status='live', Symbol='LOC100418554', Aliases='', description='zinc finger protein 93 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505566, CurrentID=0, Status='live', Symbol='LINC00578', Aliases='', description='long intergenic non-protein coding RNA 578', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177441921, end_position_on_the_genomic_accession=177752704, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505567, CurrentID=0, Status='live', Symbol='BTBD9-AS1', Aliases='', description='BTBD9 antisense RNA 1', other_designations='BTBD9 antisense RNA 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=38449372, end_position_on_the_genomic_accession=38450303, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  497634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497634, CurrentID=0, Status='live', Symbol='LINC00293', Aliases='BEYLA, NCRNA00293', description='long intergenic non-protein coding RNA 293', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46840886, end_position_on_the_genomic_accession=46855785, orientation='plus', exon_count=12, OMIM=609543, no_hdr0=''),
  100418555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418555, CurrentID=0, Status='live', Symbol='LOC100418555', Aliases='', description='zinc finger protein 70 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505573, CurrentID=0, Status='live', Symbol='INAFM2', Aliases='LINC00984, OGU1', description='InaF-motif containing 2', other_designations='inaF-motif-containing protein 2|osteogenesis up-regulated transcript 1|osteogenesis upregulating transcript 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40324037, end_position_on_the_genomic_accession=40326715, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100505576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505576, CurrentID=0, Status='live', Symbol='LINC00672', Aliases='', description='long intergenic non-protein coding RNA 672', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38925168, end_position_on_the_genomic_accession=38929384, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505583, CurrentID=0, Status='live', Symbol='LINC00838', Aliases='', description='long intergenic non-protein coding RNA 838', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33759713, end_position_on_the_genomic_accession=33772680, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505585, CurrentID=0, Status='live', Symbol='LOC100505585', Aliases='', description='uncharacterized LOC100505585', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41928820, end_position_on_the_genomic_accession=41930136, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100419798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419798, CurrentID=0, Status='live', Symbol='LOC100419798', Aliases='', description='gon-4-like (C. elegans) pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155615569, end_position_on_the_genomic_accession=155646675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505588, CurrentID=0, Status='live', Symbol='LOC100505588', Aliases='', description='uncharacterized LOC100505588', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124007675, end_position_on_the_genomic_accession=124009396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505591, CurrentID=0, Status='live', Symbol='LRRC3C', Aliases='', description='leucine rich repeat containing 3C', other_designations='putative LRRC3-like protein ENSP00000367157', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39927722, end_position_on_the_genomic_accession=39945034, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100505592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505592, CurrentID=0, Status='live', Symbol='GAPLINC', Aliases='LINC01540', description='gastric adenocarcinoma associated, positive CD44 regulator, long intergenic non-coding RNA', other_designations='TCONS_00026238|gastric adenocarcinoma predictive long intergenic noncoding RNA|long intergenic non-protein coding RNA 1540', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3466250, end_position_on_the_genomic_accession=3478978, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  497657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497657, CurrentID=0, Status='live', Symbol='ANIB3', Aliases='AIB3', description='aneurysm, intracranial berry 3', other_designations='', map_location='1p36.13-p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609122, no_hdr0=''),
  100418559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418559, CurrentID=0, Status='live', Symbol='LOC100418559', Aliases='', description='dihydrofolate reductase pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  497661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497661, CurrentID=0, Status='live', Symbol='C18orf32', Aliases='', description='chromosome 18 open reading frame 32', other_designations='putative NF-kappa-B-activating protein 200|putative NFkB activating protein', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49481178, end_position_on_the_genomic_accession=49487274, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  497662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=497662, CurrentID=0, Status='live', Symbol='RRIS', Aliases='', description='Respiratory rhythmicity in sleep', other_designations='', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609116, no_hdr0=''),
  100505601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505601, CurrentID=0, Status='live', Symbol='PARD3-AS1', Aliases='', description='PARD3 antisense RNA 1', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34815768, end_position_on_the_genomic_accession=34816422, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287489, CurrentID=0, Status='live', Symbol='LOC100287489', Aliases='', description='neurotrophin 4 pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49050020, end_position_on_the_genomic_accession=49050873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505609, CurrentID=0, Status='live', Symbol='LOC100505609', Aliases='', description='uncharacterized LOC100505609', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180414173, end_position_on_the_genomic_accession=180422493, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377453, CurrentID=0, Status='live', Symbol='LOC105377453', Aliases='', description='uncharacterized LOC105377453', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139782146, end_position_on_the_genomic_accession=139798525, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505616, CurrentID=0, Status='live', Symbol='SNHG21', Aliases='FSD2-AS1, SCARNA15HG', description='small nucleolar RNA host gene 21', other_designations='FSD2 antisense RNA 1|RP11-752G15.6|small nucleolar RNA host gene 21 (non-protein coding)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82750564, end_position_on_the_genomic_accession=82757206, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100505619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505619, CurrentID=0, Status='live', Symbol='LINC00919', Aliases='', description='long intergenic non-protein coding RNA 919', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52083065, end_position_on_the_genomic_accession=52085109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505620, CurrentID=0, Status='live', Symbol='LOC100505620', Aliases='', description='uncharacterized LOC100505620', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=38148442, end_position_on_the_genomic_accession=38150200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100505621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505621, CurrentID=0, Status='live', Symbol='C11orf72', Aliases='', description='chromosome 11 open reading frame 72', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67602880, end_position_on_the_genomic_accession=67606706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505622, CurrentID=0, Status='live', Symbol='LOC100505622', Aliases='', description='uncharacterized LOC100505622', other_designations='CTC-378H22.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42132629, end_position_on_the_genomic_accession=42137099, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505624, CurrentID=0, Status='live', Symbol='GTF3C2-AS1', Aliases='', description='GTF3C2 antisense RNA 1', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27335542, end_position_on_the_genomic_accession=27337803, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505625, CurrentID=0, Status='live', Symbol='LOC100505625', Aliases='', description='uncharacterized LOC100505625', other_designations='CTD-2044J15.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6686325, end_position_on_the_genomic_accession=6707711, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505633, CurrentID=0, Status='live', Symbol='LINC01133', Aliases='', description='long intergenic non-protein coding RNA 1133', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159961224, end_position_on_the_genomic_accession=159979086, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505635, CurrentID=0, Status='live', Symbol='LOC100505635', Aliases='', description='uncharacterized LOC100505635', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39888790, end_position_on_the_genomic_accession=39897380, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505636, CurrentID=0, Status='live', Symbol='LINC01024', Aliases='MA-linc1', description='long intergenic non-protein coding RNA 1024', other_designations='CTB-131B5.4|mitosis-associated long intergenic non-coding RNA 1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140102922, end_position_on_the_genomic_accession=140108012, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505641, CurrentID=0, Status='live', Symbol='FGD5-AS1', Aliases='', description='FGD5 antisense RNA 1', other_designations='FGD5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14942779, end_position_on_the_genomic_accession=14948441, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505642, CurrentID=0, Status='live', Symbol='LOC100505642', Aliases='', description='chromosome 9 open reading frame 85 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182955238, end_position_on_the_genomic_accession=182955597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505648, CurrentID=0, Status='live', Symbol='RAD51-AS1', Aliases='TODRA', description='RAD51 antisense RNA 1 (head to head)', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40693734, end_position_on_the_genomic_accession=40695105, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505650, CurrentID=0, Status='live', Symbol='LOC100505650', Aliases='', description='uncharacterized LOC100505650', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505658, CurrentID=0, Status='live', Symbol='LOC100505658', Aliases='', description='uncharacterized LOC100505658', other_designations='CTD-2024I7.13', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141558311, end_position_on_the_genomic_accession=141565260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505659, CurrentID=0, Status='live', Symbol='LINC00967', Aliases='', description='long intergenic non-protein coding RNA 967', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66192114, end_position_on_the_genomic_accession=66197319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287498, CurrentID=0, Status='live', Symbol='LOC100287498', Aliases='', description='NDUFA4, mitochondrial complex associated pseudogene', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa pseudogene', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203672178, end_position_on_the_genomic_accession=203672405, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505663, CurrentID=0, Status='live', Symbol='LINC01370', Aliases='HI-LNC25', description='long intergenic non-protein coding RNA 1370', other_designations='Human Islet Long Non Coding RNA 25', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40004461, end_position_on_the_genomic_accession=40008529, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505664, CurrentID=0, Status='live', Symbol='LOC100505664', Aliases='', description='uncharacterized LOC100505664', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23159611, end_position_on_the_genomic_accession=23190307, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100505666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505666, CurrentID=0, Status='live', Symbol='LOC100505666', Aliases='', description='uncharacterized LOC100505666', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155045191, end_position_on_the_genomic_accession=155063991, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371729, CurrentID=0, Status='live', Symbol='LOC105371729', Aliases='', description='uncharacterized LOC105371729', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156001153, end_position_on_the_genomic_accession=156033225, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100505676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505676, CurrentID=0, Status='live', Symbol='RRS1-AS1', Aliases='', description='RRS1 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66419587, end_position_on_the_genomic_accession=66428977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505678, CurrentID=0, Status='live', Symbol='STARD4-AS1', Aliases='', description='STARD4 antisense RNA 1', other_designations='CTC-426B10.1|STARD4 antisense RNA 1 (non-protein coding)', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111512226, end_position_on_the_genomic_accession=111739726, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100505679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505679, CurrentID=0, Status='live', Symbol='UBE2Q2L', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2-like', other_designations='ubiquitin-conjugating enzyme E2Q family member 2-like', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84172490, end_position_on_the_genomic_accession=84182233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505681, CurrentID=0, Status='live', Symbol='NAPA-AS1', Aliases='', description='NAPA antisense RNA 1', other_designations='NAPA antisense RNA 1 (non-protein coding)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47484282, end_position_on_the_genomic_accession=47501597, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505683, CurrentID=0, Status='live', Symbol='LINC01431', Aliases='', description='long intergenic non-protein coding RNA 1431', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23356594, end_position_on_the_genomic_accession=23358127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505685, CurrentID=0, Status='live', Symbol='LOC100505685', Aliases='', description='uncharacterized LOC100505685', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151955965, end_position_on_the_genomic_accession=151967587, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505687, CurrentID=0, Status='live', Symbol='LINC00888', Aliases='', description='long intergenic non-protein coding RNA 888', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183447608, end_position_on_the_genomic_accession=183456013, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505690, CurrentID=0, Status='live', Symbol='LOC100505690', Aliases='', description='thioredoxin 2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84085675, end_position_on_the_genomic_accession=84086967, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505692, CurrentID=0, Status='live', Symbol='FTO-IT1', Aliases='', description='FTO intronic transcript 1', other_designations='FTO intronic transcript 1 (non-protein coding)', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54039393, end_position_on_the_genomic_accession=54040373, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927606, CurrentID=0, Status='live', Symbol='LOC101927606', Aliases='', description='uncharacterized LOC101927606', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74409448, end_position_on_the_genomic_accession=74413209, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505694, CurrentID=0, Status='live', Symbol='LOC100505694', Aliases='', description='family with sequence similarity 133, member A pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28015762, end_position_on_the_genomic_accession=28016482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480318, CurrentID=0, Status='live', Symbol='MTCO1P5', Aliases='', description='MT-CO1 pseudogene 5', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505696, CurrentID=0, Status='live', Symbol='SH3BP5-AS1', Aliases='', description='SH3BP5 antisense RNA 1', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15254184, end_position_on_the_genomic_accession=15264498, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377467, CurrentID=0, Status='live', Symbol='LOC105377467', Aliases='', description='uncharacterized LOC105377467', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145599063, end_position_on_the_genomic_accession=145602281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505702, CurrentID=0, Status='live', Symbol='LINC01094', Aliases='', description='long intergenic non-protein coding RNA 1094', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78645994, end_position_on_the_genomic_accession=78684501, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418577, CurrentID=0, Status='live', Symbol='LOC100418577', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505704, CurrentID=0, Status='live', Symbol='ATP5G1P6', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 6', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 6', map_location='18q21.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63496942, end_position_on_the_genomic_accession=63497498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505705, CurrentID=0, Status='live', Symbol='SDIM1', Aliases='', description='stress responsive DNAJB4 interacting membrane protein 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=166307330, end_position_on_the_genomic_accession=166309725, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100505711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505711, CurrentID=0, Status='live', Symbol='LOC100505711', Aliases='', description='uncharacterized LOC100505711', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41381392, end_position_on_the_genomic_accession=41383128, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505715, CurrentID=0, Status='live', Symbol='LOC100505715', Aliases='', description='uncharacterized LOC100505715', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43891804, end_position_on_the_genomic_accession=43901803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505716, CurrentID=0, Status='live', Symbol='LOC100505716', Aliases='', description='uncharacterized LOC100505716', other_designations='', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28307691, end_position_on_the_genomic_accession=28310459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505718, CurrentID=0, Status='live', Symbol='LINC01592', Aliases='', description='long intergenic non-protein coding RNA 1592', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68911803, end_position_on_the_genomic_accession=69104190, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375243, CurrentID=0, Status='live', Symbol='LOC105375243', Aliases='', description='uncharacterized LOC105375243', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40708084, end_position_on_the_genomic_accession=40709013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377471, CurrentID=0, Status='live', Symbol='LOC105377471', Aliases='', description='uncharacterized LOC105377471', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146199982, end_position_on_the_genomic_accession=146204436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505724, CurrentID=0, Status='live', Symbol='KRTAP9-7', Aliases='KAP9.7, KRTAP9L1', description='keratin associated protein 9-7', other_designations='keratin associated protein 9-like 1|putative keratin-associated protein 9-2-like 1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41275698, end_position_on_the_genomic_accession=41276207, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480319, CurrentID=0, Status='live', Symbol='NAMPTP3', Aliases='', description='nicotinamide phosphoribosyltransferase pseudogene 3', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129942, CurrentID=0, Status='live', Symbol='LOC100129942', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 2 pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88989522, end_position_on_the_genomic_accession=88990246, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505727, CurrentID=0, Status='live', Symbol='LOC100505727', Aliases='', description='uncharacterized LOC100505727', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44485007, end_position_on_the_genomic_accession=44493034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505728, CurrentID=0, Status='live', Symbol='LOC100505728', Aliases='', description='uncharacterized LOC100505728', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155745112, end_position_on_the_genomic_accession=155748218, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270845, CurrentID=0, Status='live', Symbol='RPL6P6', Aliases='RPL6_5_321', description='ribosomal protein L6 pseudogene 6', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210088226, end_position_on_the_genomic_accession=210089064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505736, CurrentID=0, Status='live', Symbol='LOC100505736', Aliases='', description='uncharacterized LOC100505736', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28148555, end_position_on_the_genomic_accession=28206961, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505738, CurrentID=0, Status='live', Symbol='MIR4458HG', Aliases='', description='MIR4458 host gene', other_designations='MIR4458 host gene (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8457687, end_position_on_the_genomic_accession=8463096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505739, CurrentID=0, Status='live', Symbol='LINC01603', Aliases='', description='long intergenic non-protein coding RNA 1603', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69424870, end_position_on_the_genomic_accession=69448244, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505741, CurrentID=0, Status='live', Symbol='SPATA1', Aliases='SP-2, SPAP1', description='spermatogenesis associated 1', other_designations='sperm-specific protein SP-2|spermatogenesis associated protein 1', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84506301, end_position_on_the_genomic_accession=84556495, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100418584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418584, CurrentID=0, Status='live', Symbol='LOC100418584', Aliases='', description='mucin 4, cell surface associated pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505746, CurrentID=0, Status='live', Symbol='ITGB2-AS1', Aliases='', description='ITGB2 antisense RNA 1', other_designations='ITGB2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44921035, end_position_on_the_genomic_accession=44929680, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418585, CurrentID=0, Status='live', Symbol='LOC100418585', Aliases='', description='mucin 4, cell surface associated pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505753, CurrentID=0, Status='live', Symbol='KRTAP16-1', Aliases='KAP16.1', description='keratin associated protein 16-1', other_designations='putative keratin-associated protein 10-like ENSP00000375147', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41307700, end_position_on_the_genomic_accession=41309253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100505755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505755, CurrentID=0, Status='live', Symbol='MTND5P28', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 28', other_designations='MT-ND5 pseudogene 28', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120215338, end_position_on_the_genomic_accession=120217212, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480320, CurrentID=0, Status='live', Symbol='TMF1P1', Aliases='', description='TATA element modulatory factor 1 pseudogene 1', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505758, CurrentID=0, Status='live', Symbol='PRMT5-AS1', Aliases='', description='PRMT5 antisense RNA 1', other_designations='PRMT5 antisense RNA 1 (non-protein coding)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22919456, end_position_on_the_genomic_accession=22923407, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377477, CurrentID=0, Status='live', Symbol='LOC105377477', Aliases='', description='uncharacterized LOC105377477', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147569353, end_position_on_the_genomic_accession=147594586, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100505761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505761, CurrentID=0, Status='live', Symbol='RPARP-AS1', Aliases='', description='RPARP antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102449817, end_position_on_the_genomic_accession=102456293, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100505767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505767, CurrentID=0, Status='live', Symbol='SPDYE18', Aliases='', description='speedy/RINGO cell cycle regulator family member E18', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77051689, end_position_on_the_genomic_accession=77060538, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100505768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505768, CurrentID=0, Status='live', Symbol='LINC01364', Aliases='', description='long intergenic non-protein coding RNA 1364', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87353524, end_position_on_the_genomic_accession=87371655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505771, CurrentID=0, Status='live', Symbol='LOC100505771', Aliases='', description='uncharacterized LOC100505771', other_designations='CTD-3184A7.4', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63627243, end_position_on_the_genomic_accession=63628757, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505774, CurrentID=0, Status='live', Symbol='LOC100505774', Aliases='', description='uncharacterized LOC100505774', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28709050, end_position_on_the_genomic_accession=28751731, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100505776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505776, CurrentID=0, Status='live', Symbol='LINC01541', Aliases='UTAT39', description='long intergenic non-protein coding RNA 1541', other_designations='uterine endometrial carcinoma-associated transcript 39', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71519964, end_position_on_the_genomic_accession=71578956, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100129324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129324, CurrentID=0, Status='live', Symbol='LOC100129324', Aliases='', description='uncharacterized LOC100129324', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=66958729, end_position_on_the_genomic_accession=66960140, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100505782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505782, CurrentID=0, Status='live', Symbol='LOC100505782', Aliases='', description='uncharacterized LOC100505782', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41402416, end_position_on_the_genomic_accession=41412588, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505783, CurrentID=0, Status='live', Symbol='OSER1-AS1', Aliases='', description='OSER1 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44210960, end_position_on_the_genomic_accession=44226027, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505784, CurrentID=0, Status='live', Symbol='LOC100505784', Aliases='', description='uncharacterized LOC100505784', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=156642531, end_position_on_the_genomic_accession=156643861, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505785, CurrentID=0, Status='live', Symbol='KIRREL-IT1', Aliases='', description='KIRREL intronic transcript 1', other_designations='KIRREL intronic transcript 1 (non-protein coding)', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=157995345, end_position_on_the_genomic_accession=158001611, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505788, CurrentID=0, Status='live', Symbol='LOC100505788', Aliases='', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4, 9kDa pseudogene', other_designations='NDUFA4, mitochondrial complex associated pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89661053, end_position_on_the_genomic_accession=89661533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100287520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287520, CurrentID=0, Status='live', Symbol='GLYATL1P3', Aliases='', description='glycine-N-acyltransferase like 1 pseudogene 3', other_designations='glycine-N-acyltransferase-like 1 pseudogene 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59084937, end_position_on_the_genomic_accession=59095841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505795, CurrentID=0, Status='live', Symbol='LOC100505795', Aliases='', description='uncharacterized LOC100505795', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164769116, end_position_on_the_genomic_accession=164774641, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505796, CurrentID=0, Status='live', Symbol='LOC100505796', Aliases='', description='IMPACT pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79000216, end_position_on_the_genomic_accession=79001120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505797, CurrentID=0, Status='live', Symbol='LOC100505797', Aliases='', description='myosin heavy chain IB-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72867267, end_position_on_the_genomic_accession=72881399, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100505798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505798, CurrentID=0, Status='live', Symbol='LOC100505798', Aliases='', description='uncharacterized LOC100505798', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140574988, end_position_on_the_genomic_accession=140594956, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105377484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377484, CurrentID=0, Status='live', Symbol='LOC105377484', Aliases='', description='uncharacterized LOC105377484', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151256005, end_position_on_the_genomic_accession=151325320, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505806, CurrentID=0, Status='live', Symbol='SNHG18', Aliases='', description='small nucleolar RNA host gene 18', other_designations='CTD-2001E22.2|small nucleolar RNA host gene 18 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9546200, end_position_on_the_genomic_accession=9550298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388813, CurrentID=0, Status='live', Symbol='LOC388813', Aliases='', description='uncharacterized protein ENSP00000383407-like', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14591392, end_position_on_the_genomic_accession=14643107, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100505811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505811, CurrentID=0, Status='live', Symbol='HNCAT21', Aliases='', description='head and neck cancer-associated transcript 21', other_designations='CTD-3179P9.2', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118282574, end_position_on_the_genomic_accession=118284777, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505812, CurrentID=0, Status='live', Symbol='CARD8-AS1', Aliases='', description='CARD8 antisense RNA 1', other_designations='CTC-241F20.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48255675, end_position_on_the_genomic_accession=48258194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505817, CurrentID=0, Status='live', Symbol='LOC100505817', Aliases='', description='uncharacterized LOC100505817', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=73324941, end_position_on_the_genomic_accession=73349889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377487, CurrentID=0, Status='live', Symbol='LOC105377487', Aliases='', description='uncharacterized LOC105377487', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151799449, end_position_on_the_genomic_accession=151828603, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418597, CurrentID=0, Status='live', Symbol='LOC100418597', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505824, CurrentID=0, Status='live', Symbol='LOC100505824', Aliases='', description='uncharacterized LOC100505824', other_designations='XXyac-YX155B6.5', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148162738, end_position_on_the_genomic_accession=148163413, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505826, CurrentID=0, Status='live', Symbol='STK4-AS1', Aliases='', description='STK4 antisense RNA 1 (head to head)', other_designations='STK4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44963794, end_position_on_the_genomic_accession=44966458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377489, CurrentID=0, Status='live', Symbol='LOC105377489', Aliases='', description='uncharacterized LOC105377489', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152179490, end_position_on_the_genomic_accession=152182360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505832, CurrentID=0, Status='live', Symbol='PROX1-AS1', Aliases='', description='PROX1 antisense RNA 1', other_designations='PROX1 antisense RNA 1 (non-protein coding)', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213819641, end_position_on_the_genomic_accession=213986153, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100505835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505835, CurrentID=0, Status='live', Symbol='LINC01532', Aliases='', description='long intergenic non-protein coding RNA 1532', other_designations='CTD-2081K17.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29002555, end_position_on_the_genomic_accession=29013955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377490, CurrentID=0, Status='live', Symbol='LOC105377490', Aliases='', description='uncharacterized LOC105377490', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152206983, end_position_on_the_genomic_accession=152225943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505839, CurrentID=0, Status='live', Symbol='SH3PXD2A-AS1', Aliases='', description='SH3PXD2A antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103746779, end_position_on_the_genomic_accession=103755409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505841, CurrentID=0, Status='live', Symbol='LOC100505841', Aliases='', description='zinc finger protein 474-like', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122129546, end_position_on_the_genomic_accession=122182663, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100505845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505845, CurrentID=0, Status='live', Symbol='ROPN1L-AS1', Aliases='', description='ROPN1L antisense RNA 1', other_designations='ROPN1L antisense RNA 1 (non-protein coding)', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10440735, end_position_on_the_genomic_accession=10441851, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723543, CurrentID=0, Status='live', Symbol='LOC102723543', Aliases='', description='uncharacterized LOC102723543', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203737201, end_position_on_the_genomic_accession=203744286, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505848, CurrentID=0, Status='live', Symbol='LOC100505848', Aliases='', description='uncharacterized LOC100505848', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=31027307, end_position_on_the_genomic_accession=31263299, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  388820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388820, CurrentID=0, Status='live', Symbol='LOC388820', Aliases='', description='putative uncharacterized protein LOC388820', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505851, CurrentID=0, Status='live', Symbol='LOC100505851', Aliases='', description='uncharacterized LOC100505851', other_designations='CTB-92J24.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23937829, end_position_on_the_genomic_accession=23957930, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505852, CurrentID=0, Status='live', Symbol='ZNF355P', Aliases='PRED65, ZNF834', description='zinc finger protein 355, pseudogene', other_designations='zinc finger protein 834', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13095785, end_position_on_the_genomic_accession=13157061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505853, CurrentID=0, Status='live', Symbol='LOC100505853', Aliases='', description='uncharacterized LOC100505853', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75696083, end_position_on_the_genomic_accession=75712403, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505854, CurrentID=0, Status='live', Symbol='APTR', Aliases='RSBN1L-AS1', description='Alu-mediated CDKN1A/p21 transcriptional regulator (non-protein coding)', other_designations='RSBN1L antisense RNA 1 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77657660, end_position_on_the_genomic_accession=77661301, orientation='minus', exon_count=5, OMIM=616048, no_hdr0=''),
  100505855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505855, CurrentID=0, Status='live', Symbol='EIF1P1', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=659986, end_position_on_the_genomic_accession=660323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418603, CurrentID=0, Status='live', Symbol='LOC100418603', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505865, CurrentID=0, Status='live', Symbol='LINC00920', Aliases='', description='long intergenic non-protein coding RNA 920', other_designations='CTD-2258A20.4', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66408524, end_position_on_the_genomic_accession=66410901, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377495, CurrentID=0, Status='live', Symbol='LOC105377495', Aliases='', description='uncharacterized LOC105377495', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152680120, end_position_on_the_genomic_accession=152681656, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505869, CurrentID=0, Status='live', Symbol='CFAP58-AS1', Aliases='CCDC147-AS1', description='CFAP58 antisense RNA 1 (head to head)', other_designations='CCDC147 antisense RNA 1 (head to head)', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104351301, end_position_on_the_genomic_accession=104353575, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505870, CurrentID=0, Status='live', Symbol='LOC100505870', Aliases='', description='uncharacterized LOC100505870', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=3010975, end_position_on_the_genomic_accession=3012574, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100505874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505874, CurrentID=0, Status='live', Symbol='LOC100505874', Aliases='', description='uncharacterized LOC100505874', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505875, CurrentID=0, Status='live', Symbol='LINC01088', Aliases='', description='long intergenic non-protein coding RNA 1088', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78971748, end_position_on_the_genomic_accession=79308799, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100505876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505876, CurrentID=0, Status='live', Symbol='CEBPZOS', Aliases='CEBPZ-AS1', description='CEBPZ opposite strand', other_designations='CEBPZ antisense RNA 1', map_location='2p22', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37196492, end_position_on_the_genomic_accession=37204743, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100505877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505877, CurrentID=0, Status='live', Symbol='UBE2E2-AS1', Aliases='', description='UBE2E2 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23195070, end_position_on_the_genomic_accession=23203045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505878, CurrentID=0, Status='live', Symbol='LOC100505878', Aliases='', description='uncharacterized LOC100505878', other_designations='CTC-493L21.1', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86746818, end_position_on_the_genomic_accession=86749772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505879, CurrentID=0, Status='live', Symbol='KIF25-AS1', Aliases='C6orf54, HGC6.1.1, NCRNA00300', description='KIF25 antisense RNA 1', other_designations='KIF25 antisense RNA 1 (non-protein coding)', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167994187, end_position_on_the_genomic_accession=167997077, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505881, CurrentID=0, Status='live', Symbol='MAGI2-AS3', Aliases='', description='MAGI2 antisense RNA 3', other_designations='MAGI2 antisense RNA 3 (non-protein coding)', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79452957, end_position_on_the_genomic_accession=79471208, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100505887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505887, CurrentID=0, Status='live', Symbol='LOC100505887', Aliases='', description='uncharacterized LOC100505887', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6724745, end_position_on_the_genomic_accession=6730012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505890, CurrentID=0, Status='live', Symbol='SORCS3-AS1', Aliases='', description='SORCS3 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104664608, end_position_on_the_genomic_accession=104666216, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505892, CurrentID=0, Status='live', Symbol='CCNB1IP1P2', Aliases='', description='cyclin B1 interacting protein 1 pseudogene 2', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45567804, end_position_on_the_genomic_accession=45569007, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505893, CurrentID=0, Status='live', Symbol='SLIT2-IT1', Aliases='', description='SLIT2 intronic transcript 1', other_designations='SLIT2 intronic transcript 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20392189, end_position_on_the_genomic_accession=20394856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505894, CurrentID=0, Status='live', Symbol='TMEM161B-AS1', Aliases='linc-POLR3G-8', description='TMEM161B antisense RNA 1', other_designations='CTC-358I24.1|TMEM161B antisense RNA 1 (non-protein coding)', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88268882, end_position_on_the_genomic_accession=88436674, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106480325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480325, CurrentID=0, Status='live', Symbol='WBP1LP8', Aliases='', description='WW domain binding protein 1-like pseudogene 8', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287539, CurrentID=0, Status='live', Symbol='LOC100287539', Aliases='', description='epsin 2 pseudogene', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31966960, end_position_on_the_genomic_accession=31967545, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505909, CurrentID=0, Status='live', Symbol='LOC100505909', Aliases='', description='uncharacterized LOC100505909', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83220527, end_position_on_the_genomic_accession=83227932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505911, CurrentID=0, Status='live', Symbol='SOS1-IT1', Aliases='', description='SOS1 intronic transcript 1', other_designations='SOS1 intronic transcript 1 (non-protein coding)', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=39219420, end_position_on_the_genomic_accession=39220432, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505912, CurrentID=0, Status='live', Symbol='LOC100505912', Aliases='', description='uncharacterized LOC100505912', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22327354, end_position_on_the_genomic_accession=22339666, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505915, CurrentID=0, Status='live', Symbol='LOC100505915', Aliases='', description='uncharacterized LOC100505915', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15094411, end_position_on_the_genomic_accession=15109197, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505918, CurrentID=0, Status='live', Symbol='LOC100505918', Aliases='', description='uncharacterized LOC100505918', other_designations='', map_location='1q23-q24', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168400188, end_position_on_the_genomic_accession=168422656, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505919, CurrentID=0, Status='live', Symbol='CFL1P7', Aliases='', description='cofilin 1 (non-muscle) pseudogene 7', other_designations='', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25324119, end_position_on_the_genomic_accession=25325777, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505920, CurrentID=0, Status='live', Symbol='LOC100505920', Aliases='', description='uncharacterized LOC100505920', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194203016, end_position_on_the_genomic_accession=194250153, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505921, CurrentID=0, Status='live', Symbol='LOC100505921', Aliases='', description='uncharacterized LOC100505921', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7949839, end_position_on_the_genomic_accession=7968752, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100505923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505923, CurrentID=0, Status='live', Symbol='LOC100505923', Aliases='', description='uncharacterized LOC100505923', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=48759915, end_position_on_the_genomic_accession=48761738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100287542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287542, CurrentID=0, Status='live', Symbol='RAB9AP2', Aliases='', description='RAB9A, member RAS oncogene family pseudogene 2', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25709911, end_position_on_the_genomic_accession=25713557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505929, CurrentID=0, Status='live', Symbol='LINC01549', Aliases='C21orf37', description='long intergenic non-protein coding RNA 1549', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17438890, end_position_on_the_genomic_accession=17449185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100505932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505932, CurrentID=0, Status='live', Symbol='LOC100505932', Aliases='', description='uncharacterized LOC100505932', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=86780739, end_position_on_the_genomic_accession=86781656, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505933, CurrentID=0, Status='live', Symbol='ADD3-AS1', Aliases='', description='ADD3 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109945559, end_position_on_the_genomic_accession=110008381, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480326, CurrentID=0, Status='live', Symbol='WBP1LP10', Aliases='', description='WW domain binding protein 1-like pseudogene 10', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418616, CurrentID=0, Status='live', Symbol='LOC100418616', Aliases='', description='family with sequence similarity 8, member A1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505938, CurrentID=0, Status='live', Symbol='LOC100505938', Aliases='', description='uncharacterized LOC100505938', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8262225, end_position_on_the_genomic_accession=8343281, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100505942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505942, CurrentID=0, Status='live', Symbol='LOC100505942', Aliases='', description='uncharacterized LOC100505942', other_designations='CTD-2012K14.8', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67517706, end_position_on_the_genomic_accession=67528745, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505945, CurrentID=0, Status='live', Symbol='BTF3L4P1', Aliases='', description='basic transcription factor 3-like 4 pseudogene 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17516934, end_position_on_the_genomic_accession=17519061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505948, CurrentID=0, Status='live', Symbol='LOC100505948', Aliases='', description='uncharacterized LOC100505948', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34978722, end_position_on_the_genomic_accession=34979937, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505951, CurrentID=0, Status='live', Symbol='LOC100505951', Aliases='', description='uncharacterized LOC100505951', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=86788836, end_position_on_the_genomic_accession=86790759, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505953, CurrentID=0, Status='live', Symbol='CICP23', Aliases='', description='capicua transcriptional repressor pseudogene 23', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123272, end_position_on_the_genomic_accession=127050, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  650983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650983, CurrentID=0, Status='live', Symbol='RCC2P1', Aliases='', description='regulator of chromosome condensation 2 pseudogene 1', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11781056, end_position_on_the_genomic_accession=11782527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505964, CurrentID=0, Status='live', Symbol='LINC01304', Aliases='', description='long intergenic non-protein coding RNA 1304', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3957655, end_position_on_the_genomic_accession=3974032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100505967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505967, CurrentID=0, Status='live', Symbol='LINC00645', Aliases='', description='long intergenic non-protein coding RNA 645', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27612588, end_position_on_the_genomic_accession=27639636, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100505970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505970, CurrentID=0, Status='live', Symbol='HEXDC-IT1', Aliases='', description='HEXDC intronic transcript 1', other_designations='HEXDC intronic transcript 1 (non-protein coding)', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=80383374, end_position_on_the_genomic_accession=80385186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100505975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505975, CurrentID=0, Status='live', Symbol='ENO1-AS1', Aliases='', description='ENO1 antisense RNA 1', other_designations='ENO1 antisense RNA 1 (non-protein coding)', map_location='1p36.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8878835, end_position_on_the_genomic_accession=8879885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505978, CurrentID=0, Status='live', Symbol='LOC100505978', Aliases='', description='uncharacterized LOC100505978', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106680758, end_position_on_the_genomic_accession=106684700, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100418623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418623, CurrentID=0, Status='live', Symbol='LOC100418623', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505982, CurrentID=0, Status='live', Symbol='LOC100505982', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214612961, end_position_on_the_genomic_accession=214613397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100505984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505984, CurrentID=0, Status='live', Symbol='LOC100505984', Aliases='', description='uncharacterized LOC100505984', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160257719, end_position_on_the_genomic_accession=160271895, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100505985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505985, CurrentID=0, Status='live', Symbol='LOC100505985', Aliases='', description='uncharacterized LOC100505985', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50093434, end_position_on_the_genomic_accession=50099283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100505989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505989, CurrentID=0, Status='live', Symbol='LINC01207', Aliases='', description='long intergenic non-protein coding RNA 1207', other_designations='', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164754131, end_position_on_the_genomic_accession=164803795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100505991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505991, CurrentID=0, Status='live', Symbol='LOC100505991', Aliases='', description='uncharacterized LOC100505991', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=172338868, end_position_on_the_genomic_accession=172384514, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100505993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505993, CurrentID=0, Status='live', Symbol='TMEM210', Aliases='', description='transmembrane protein 210', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137170928, end_position_on_the_genomic_accession=137172046, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100505994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505994, CurrentID=0, Status='live', Symbol='LUCAT1', Aliases='SCAL1', description='lung cancer associated transcript 1 (non-protein coding)', other_designations='smoke and cancer associated lncRNA 1', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91302986, end_position_on_the_genomic_accession=91314402, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100505995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505995, CurrentID=0, Status='live', Symbol='LOC100505995', Aliases='', description='apoptosis-inducing, TAF9-like domain 1 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12685545, end_position_on_the_genomic_accession=12686316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100505996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100505996, CurrentID=0, Status='live', Symbol='LINC00403', Aliases='', description='long intergenic non-protein coding RNA 403', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111972310, end_position_on_the_genomic_accession=112108015, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506000, CurrentID=0, Status='live', Symbol='HSPE1P5', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 5', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 5', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68686379, end_position_on_the_genomic_accession=68686683, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506001, CurrentID=0, Status='live', Symbol='NARF-IT1', Aliases='NARF-OT1', description='NARF intronic transcript 1', other_designations='NARF intronic transcript 1 (non-protein coding)|NARF overlapping transcript 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=80439974, end_position_on_the_genomic_accession=80441264, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506003, CurrentID=0, Status='live', Symbol='ZFY-AS1', Aliases='ZFY-AS, ZFYAS', description='ZFY antisense RNA 1', other_designations='ZFY antisense RNA (non-protein coding)|ZFY antisense RNA 1 (non-protein coding)', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=2834873, end_position_on_the_genomic_accession=2870667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506007, CurrentID=0, Status='live', Symbol='RTCA-AS1', Aliases='', description='RTCA antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100264742, end_position_on_the_genomic_accession=100266174, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418628, CurrentID=0, Status='live', Symbol='LOC100418628', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506012, CurrentID=0, Status='live', Symbol='PPP5D1', Aliases='', description='PPP5 tetratricopeptide repeat domain containing 1', other_designations='PPP5 TPR repeat domain-containing protein 1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46480788, end_position_on_the_genomic_accession=46601200, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506013, CurrentID=0, Status='live', Symbol='APELA', Aliases='ELA, Ende, tdl', description='apelin receptor early endogenous ligand', other_designations='ELABELA|protein Elabela|protein Toddler|toddler', map_location='4q32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164876988, end_position_on_the_genomic_accession=164897527, orientation='plus', exon_count=3, OMIM=615594, no_hdr0=''),
  100506016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506016, CurrentID=0, Status='live', Symbol='LOC100506016', Aliases='', description='uncharacterized LOC100506016', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112155157, end_position_on_the_genomic_accession=112160710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506020, CurrentID=0, Status='live', Symbol='ARAP1-AS2', Aliases='', description='ARAP1 antisense RNA 2', other_designations='ARAP1 antisense RNA 2 (non-protein coding)', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=72411519, end_position_on_the_genomic_accession=72416889, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418630, CurrentID=0, Status='live', Symbol='LOC100418630', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506022, CurrentID=0, Status='live', Symbol='LOC100506022', Aliases='', description='uncharacterized LOC100506022', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9425094, end_position_on_the_genomic_accession=9428996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506023, CurrentID=0, Status='live', Symbol='LOC100506023', Aliases='', description='uncharacterized LOC100506023', other_designations='', map_location='1q23-q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173235060, end_position_on_the_genomic_accession=173477155, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506025, CurrentID=0, Status='live', Symbol='ISPD-AS1', Aliases='', description='ISPD antisense RNA 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16210486, end_position_on_the_genomic_accession=16270604, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  103695431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103695431, CurrentID=0, Status='live', Symbol='LINC01048', Aliases='', description='long intergenic non-protein coding RNA 1048', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37481464, end_position_on_the_genomic_accession=37484784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506031, CurrentID=0, Status='live', Symbol='LOC100506031', Aliases='', description='uncharacterized LOC100506031', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=69359469, end_position_on_the_genomic_accession=69376357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506033, CurrentID=0, Status='live', Symbol='PTOV1-AS1', Aliases='', description='PTOV1 antisense RNA 1', other_designations='PTOV1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49838645, end_position_on_the_genomic_accession=49851676, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506035, CurrentID=0, Status='live', Symbol='LINC00989', Aliases='', description='long intergenic non-protein coding RNA 989', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79492593, end_position_on_the_genomic_accession=79576460, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506036, CurrentID=0, Status='live', Symbol='LOC100506036', Aliases='', description='uncharacterized LOC100506036', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96812239, end_position_on_the_genomic_accession=96814072, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506038, CurrentID=0, Status='live', Symbol='LOC100506038', Aliases='', description='uncharacterized LOC100506038', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149977212, end_position_on_the_genomic_accession=149999828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506039, CurrentID=0, Status='live', Symbol='HMGB3P8', Aliases='', description='high mobility group box 3 pseudogene 8', other_designations='high-mobility group box 3 pseudogene 8', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116439205, end_position_on_the_genomic_accession=116440067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418634, CurrentID=0, Status='live', Symbol='LOC100418634', Aliases='', description='olfactory receptor, family 2, subfamily G, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506046, CurrentID=0, Status='live', Symbol='GAS5-AS1', Aliases='', description='GAS5 antisense RNA 1', other_designations='GAS5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173863248, end_position_on_the_genomic_accession=173863941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506047, CurrentID=0, Status='live', Symbol='LOC100506047', Aliases='', description='uncharacterized LOC100506047', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43069833, end_position_on_the_genomic_accession=43158220, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100506049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506049, CurrentID=0, Status='live', Symbol='LRRC72', Aliases='', description='leucine rich repeat containing 72', other_designations='leucine-rich repeat-containing protein ENSP00000371558', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16526827, end_position_on_the_genomic_accession=16581554, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100506050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506050, CurrentID=0, Status='live', Symbol='LINC01005', Aliases='', description='long intergenic non-protein coding RNA 1005', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64024416, end_position_on_the_genomic_accession=64030102, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100418635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418635, CurrentID=0, Status='live', Symbol='LOC100418635', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506054, CurrentID=0, Status='live', Symbol='RNASEH1-AS1', Aliases='', description='RNASEH1 antisense RNA 1', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3558386, end_position_on_the_genomic_accession=3561750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506055, CurrentID=0, Status='live', Symbol='LOC100506055', Aliases='', description='matrix-remodelling associated 7 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161339503, end_position_on_the_genomic_accession=161341385, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418636, CurrentID=0, Status='live', Symbol='LOC100418636', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506058, CurrentID=0, Status='live', Symbol='EIF5A2P1', Aliases='', description='eukaryotic translation initiation factor 5A2 pseudogene 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72041177, end_position_on_the_genomic_accession=72041781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506060, CurrentID=0, Status='live', Symbol='SMG1P7', Aliases='', description='SMG1 pseudogene 7', other_designations='SMG1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70219581, end_position_on_the_genomic_accession=70226033, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506061, CurrentID=0, Status='live', Symbol='BTF3L4P4', Aliases='', description='basic transcription factor 3-like 4 pseudogene 4', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168598115, end_position_on_the_genomic_accession=168598826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506068, CurrentID=0, Status='live', Symbol='DACT3-AS1', Aliases='', description='DACT3 antisense RNA 1', other_designations='DACT3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46660364, end_position_on_the_genomic_accession=46677456, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506070, CurrentID=0, Status='live', Symbol='RBFADN', Aliases='', description='RBFA downstream neighbor (non-protein coding)', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80067256, end_position_on_the_genomic_accession=80081304, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506071, CurrentID=0, Status='live', Symbol='LOC100506071', Aliases='', description='uncharacterized LOC100506071', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30876179, end_position_on_the_genomic_accession=30889808, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100418639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418639, CurrentID=0, Status='live', Symbol='LOC100418639', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506076, CurrentID=0, Status='live', Symbol='LOC100506076', Aliases='', description='uncharacterized LOC100506076', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97278717, end_position_on_the_genomic_accession=97291878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506077, CurrentID=0, Status='live', Symbol='LOC100506077', Aliases='', description='uncharacterized LOC100506077', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=162949939, end_position_on_the_genomic_accession=162951376, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506080, CurrentID=0, Status='live', Symbol='HMGA1P4', Aliases='', description='high mobility group AT-hook 1 pseudogene 4', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128663134, end_position_on_the_genomic_accession=128663578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506082, CurrentID=0, Status='live', Symbol='LOC100506082', Aliases='', description='uncharacterized LOC100506082', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4187152, end_position_on_the_genomic_accession=4202653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506083, CurrentID=0, Status='live', Symbol='LOC100506083', Aliases='', description='uncharacterized LOC100506083', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70315640, end_position_on_the_genomic_accession=70346747, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506084, CurrentID=0, Status='live', Symbol='ARL17B', Aliases='ARL17, ARL17A', description='ADP ribosylation factor like GTPase 17B', other_designations='ADP-ribosylation factor 7|ADP-ribosylation factor like GTPase 17B|ADP-ribosylation factor-like 17B', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46274184, end_position_on_the_genomic_accession=46362064, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100506085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506085, CurrentID=0, Status='live', Symbol='LOC100506085', Aliases='', description='uncharacterized LOC100506085', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169917761, end_position_on_the_genomic_accession=169975902, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100418641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418641, CurrentID=0, Status='live', Symbol='LOC100418641', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105748977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105748977, CurrentID=0, Status='live', Symbol='LOC105748977', Aliases='', description='uncharacterized LOC105748977', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506098, CurrentID=0, Status='live', Symbol='LOC100506098', Aliases='', description='uncharacterized LOC100506098', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20213995, end_position_on_the_genomic_accession=20221700, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506099, CurrentID=0, Status='live', Symbol='LOC100506099', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47341614, end_position_on_the_genomic_accession=47342089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506100, CurrentID=0, Status='live', Symbol='LOC100506100', Aliases='', description='uncharacterized LOC100506100', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128724462, end_position_on_the_genomic_accession=128733243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506103, CurrentID=0, Status='live', Symbol='LOC100506103', Aliases='', description='6-pyruvoyltetrahydropterin synthase pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=60927094, end_position_on_the_genomic_accession=60927428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506107, CurrentID=0, Status='live', Symbol='LOC100506107', Aliases='', description='uncharacterized LOC100506107', other_designations='', map_location='4q32-q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170742595, end_position_on_the_genomic_accession=170743749, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506108, CurrentID=0, Status='live', Symbol='SIX3-AS1', Aliases='', description='SIX3 antisense RNA 1', other_designations='SIX3 antisense RNA 1 (non-protein coding)', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44940154, end_position_on_the_genomic_accession=44941878, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418645, CurrentID=0, Status='live', Symbol='LOC100418645', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506113, CurrentID=0, Status='live', Symbol='LOC100506113', Aliases='', description='uncharacterized LOC100506113', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75800758, end_position_on_the_genomic_accession=75815187, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100506114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506114, CurrentID=0, Status='live', Symbol='LOC100506114', Aliases='', description='uncharacterized LOC100506114', other_designations='CTD-2545M3.8', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50486560, end_position_on_the_genomic_accession=50488086, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506115, CurrentID=0, Status='live', Symbol='LINC01272', Aliases='', description='long intergenic non-protein coding RNA 1272', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50267483, end_position_on_the_genomic_accession=50279795, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129955, CurrentID=0, Status='live', Symbol='LOC100129955', Aliases='', description='RNA polymerase I-specific transcription initiation factor RRN3-like', other_designations='', map_location='2q22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142854051, end_position_on_the_genomic_accession=142856202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506119, CurrentID=0, Status='live', Symbol='LINC01503', Aliases='', description='long intergenic non-protein coding RNA 1503', other_designations='lnc-PPP2R4-5', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129336923, end_position_on_the_genomic_accession=129347477, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506122, CurrentID=0, Status='live', Symbol='LOC100506122', Aliases='', description='uncharacterized LOC100506122', other_designations='', map_location='4q32-q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171040602, end_position_on_the_genomic_accession=171059160, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506123, CurrentID=0, Status='live', Symbol='LOC100506123', Aliases='', description='uncharacterized LOC100506123', other_designations='', map_location='2q11', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97416165, end_position_on_the_genomic_accession=97424027, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506124, CurrentID=0, Status='live', Symbol='LOC100506124', Aliases='', description='uncharacterized LOC100506124', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165857475, end_position_on_the_genomic_accession=165871941, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506125, CurrentID=0, Status='live', Symbol='LOC100506125', Aliases='', description='uncharacterized LOC100506125', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48789147, end_position_on_the_genomic_accession=48790535, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506126, CurrentID=0, Status='live', Symbol='LINC00867', Aliases='', description='long intergenic non-protein coding RNA 867', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118357108, end_position_on_the_genomic_accession=118359203, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506127, CurrentID=0, Status='live', Symbol='LOC100506127', Aliases='', description='putative uncharacterized protein FLJ37770-like', other_designations='', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76381325, end_position_on_the_genomic_accession=76408132, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418648, CurrentID=0, Status='live', Symbol='LOC100418648', Aliases='', description='olfactory receptor, family 2, subfamily Y, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506133, CurrentID=0, Status='live', Symbol='LOC100506133', Aliases='', description='ubiquitin conjugating enzyme E2N pseudogene', other_designations='ubiquitin-conjugating enzyme E2N pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=172380031, end_position_on_the_genomic_accession=172380487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506134, CurrentID=0, Status='live', Symbol='TTC21B-AS1', Aliases='', description='TTC21B antisense RNA 1', other_designations='', map_location='2q24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=165933857, end_position_on_the_genomic_accession=165948321, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418649, CurrentID=0, Status='live', Symbol='LOC100418649', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506136, CurrentID=0, Status='live', Symbol='LOC100506136', Aliases='', description='uncharacterized LOC100506136', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96621657, end_position_on_the_genomic_accession=96664338, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506142, CurrentID=0, Status='live', Symbol='LOC100506142', Aliases='', description='uncharacterized LOC100506142', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46568256, end_position_on_the_genomic_accession=46580238, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506144, CurrentID=0, Status='live', Symbol='ZMYM6NB', Aliases='', description='ZMYM6 neighbor', other_designations='ZMYM6 neighbor protein', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34981526, end_position_on_the_genomic_accession=34985347, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418651, CurrentID=0, Status='live', Symbol='LOC100418651', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377543, CurrentID=0, Status='live', Symbol='LOC105377543', Aliases='', description='uncharacterized LOC105377543', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173909737, end_position_on_the_genomic_accession=173937358, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100506157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506157, CurrentID=0, Status='live', Symbol='SRP54-AS1', Aliases='', description='SRP54 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34916379, end_position_on_the_genomic_accession=34982609, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105374373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374373, CurrentID=0, Status='live', Symbol='LOC105374373', Aliases='', description='uncharacterized LOC105374373', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8268608, end_position_on_the_genomic_accession=8269606, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506159, CurrentID=0, Status='live', Symbol='LOC100506159', Aliases='', description='uncharacterized LOC100506159', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9936579, end_position_on_the_genomic_accession=9943495, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506161, CurrentID=0, Status='live', Symbol='TAF1A-AS1', Aliases='', description='TAF1A antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222589920, end_position_on_the_genomic_accession=222592633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506164, CurrentID=0, Status='live', Symbol='HSFX1', Aliases='LW-1', description='heat shock transcription factor family, X-linked 1', other_designations='heat shock transcription factor family, X linked 1', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149774068, end_position_on_the_genomic_accession=149776870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506169, CurrentID=0, Status='live', Symbol='MTND5P11', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 11', other_designations='MT-ND5 pseudogene 11', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134924648, end_position_on_the_genomic_accession=134926459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506172, CurrentID=0, Status='live', Symbol='LINC01568', Aliases='', description='long intergenic non-protein coding RNA 1568', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73386805, end_position_on_the_genomic_accession=73421396, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506173, CurrentID=0, Status='live', Symbol='1060P11.3', Aliases='', description='killer cell immunoglobulin-like receptor, three domains, pseudogene', other_designations='1060P11.3 (killer inhibitory receptor 2-2-2 (KIR222) (KIRX))', map_location='', chromosome='19', genomic_nucleotide_accession_version='NT_113949.1', start_position_on_the_genomic_accession=38479, end_position_on_the_genomic_accession=51882, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506175, CurrentID=0, Status='live', Symbol='LOC100506175', Aliases='', description='uncharacterized LOC100506175', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50645471, end_position_on_the_genomic_accession=50660291, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129957, CurrentID=0, Status='live', Symbol='LOC100129957', Aliases='', description='PC4 and SFRS1 interacting protein 1 pseudogene', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175286615, end_position_on_the_genomic_accession=175287531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506178, CurrentID=0, Status='live', Symbol='LOC100506178', Aliases='', description='uncharacterized LOC100506178', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22563337, end_position_on_the_genomic_accession=22573998, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377547, CurrentID=0, Status='live', Symbol='LOC105377547', Aliases='', description='uncharacterized LOC105377547', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174354729, end_position_on_the_genomic_accession=174396697, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100419802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419802, CurrentID=0, Status='live', Symbol='LOC100419802', Aliases='', description='zinc finger protein 184 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35510077, end_position_on_the_genomic_accession=35511341, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506188, CurrentID=0, Status='live', Symbol='LOC100506188', Aliases='', description='uncharacterized LOC100506188', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57260524, end_position_on_the_genomic_accession=57263346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506190, CurrentID=0, Status='live', Symbol='LINC00963', Aliases='', description='long intergenic non-protein coding RNA 963', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129488660, end_position_on_the_genomic_accession=129513686, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506191, CurrentID=0, Status='live', Symbol='LOC100506191', Aliases='', description='uncharacterized LOC100506191', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=91832852, end_position_on_the_genomic_accession=91835376, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506195, CurrentID=0, Status='live', Symbol='LARGE-AS1', Aliases='', description='LARGE antisense RNA 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33724986, end_position_on_the_genomic_accession=33750817, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100506205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506205, CurrentID=0, Status='live', Symbol='TMEM185AP1', Aliases='FAM11C, TMEM185AL', description='transmembrane protein 185A pseudogene 1', other_designations='family with sequence similarity 11, member C', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149768883, end_position_on_the_genomic_accession=149772826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506207, CurrentID=0, Status='live', Symbol='LOC100506207', Aliases='', description='uncharacterized LOC100506207', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8435623, end_position_on_the_genomic_accession=8785445, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506211, CurrentID=0, Status='live', Symbol='MIR210HG', Aliases='', description='MIR210 host gene', other_designations='MIR210 host gene (non-protein coding)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=565657, end_position_on_the_genomic_accession=568457, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723381, CurrentID=0, Status='live', Symbol='LOC102723381', Aliases='', description='uncharacterized LOC102723381', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30234956, end_position_on_the_genomic_accession=30243660, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506215, CurrentID=0, Status='live', Symbol='PAXBP1-AS1', Aliases='GCFC1-AS1', description='PAXBP1 antisense RNA 1', other_designations='GCFC1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32728114, end_position_on_the_genomic_accession=32743122, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506217, CurrentID=0, Status='live', Symbol='LOC100506217', Aliases='', description='putative POM121-like protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21481307, end_position_on_the_genomic_accession=21484694, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418663, CurrentID=0, Status='live', Symbol='LOC100418663', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377554, CurrentID=0, Status='live', Symbol='LOC105377554', Aliases='', description='uncharacterized LOC105377554', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176380828, end_position_on_the_genomic_accession=176521706, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506224, CurrentID=0, Status='live', Symbol='LOC100506224', Aliases='', description='translocase of outer mitochondrial membrane 5 homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108372775, end_position_on_the_genomic_accession=108372927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418664, CurrentID=0, Status='live', Symbol='LOC100418664', Aliases='', description='olfactory receptor, family 2, subfamily B, member 6 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377555, CurrentID=0, Status='live', Symbol='LOC105377555', Aliases='', description='uncharacterized LOC105377555', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176654157, end_position_on_the_genomic_accession=176706143, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506229, CurrentID=0, Status='live', Symbol='LINC01093', Aliases='', description='long intergenic non-protein coding RNA 1093', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184893000, end_position_on_the_genomic_accession=184899461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506233, CurrentID=0, Status='live', Symbol='RAB30-AS1', Aliases='', description='RAB30 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83072066, end_position_on_the_genomic_accession=83073712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506235, CurrentID=0, Status='live', Symbol='LOC100506235', Aliases='', description='uncharacterized LOC100506235', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47906596, end_position_on_the_genomic_accession=47908167, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506236, CurrentID=0, Status='live', Symbol='LOC100506236', Aliases='', description='uncharacterized LOC100506236', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25831291, end_position_on_the_genomic_accession=25862265, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506237, CurrentID=0, Status='live', Symbol='NKX2-1-AS1', Aliases='', description='NKX2-1 antisense RNA 1', other_designations='NKX2-1 antisense RNA 1 (non-protein coding)|Putative uncharacterized protein CS0DJ013YG01', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36519278, end_position_on_the_genomic_accession=36523016, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377557, CurrentID=0, Status='live', Symbol='LOC105377557', Aliases='', description='uncharacterized LOC105377557', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177220431, end_position_on_the_genomic_accession=177226776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506241, CurrentID=0, Status='live', Symbol='LL22NC01-81G9.3', Aliases='', description='uncharacterized protein FLJ39582-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36965248, end_position_on_the_genomic_accession=36968172, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506242, CurrentID=0, Status='live', Symbol='LOC100506242', Aliases='', description='GLLV3061', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=180138806, end_position_on_the_genomic_accession=180139219, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506243, CurrentID=0, Status='live', Symbol='KRBOX1', Aliases='', description='KRAB box domain containing 1', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42936342, end_position_on_the_genomic_accession=42942792, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506244, CurrentID=0, Status='live', Symbol='SUMO2P8', Aliases='', description='SUMO2 pseudogene 8', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 8|small ubiquitin-related modifier 2 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108617401, end_position_on_the_genomic_accession=108618411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377558, CurrentID=0, Status='live', Symbol='LOC105377558', Aliases='', description='uncharacterized LOC105377558', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177342348, end_position_on_the_genomic_accession=177381354, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506246, CurrentID=0, Status='live', Symbol='REXO1L9P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 9, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 9, pseudogene|REX1, RNA exonuclease 1 homolog-like 1 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85762855, end_position_on_the_genomic_accession=85764884, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506248, CurrentID=0, Status='live', Symbol='LOC100506248', Aliases='', description='prothymosin alpha', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12111002, end_position_on_the_genomic_accession=12111688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418668, CurrentID=0, Status='live', Symbol='LOC100418668', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506253, CurrentID=0, Status='live', Symbol='LOC100506253', Aliases='', description='uncharacterized LOC100506253', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75178628, end_position_on_the_genomic_accession=75181227, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376113, CurrentID=0, Status='live', Symbol='LOC105376113', Aliases='', description='uncharacterized LOC105376113', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83573852, end_position_on_the_genomic_accession=83585798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506257, CurrentID=0, Status='live', Symbol='LOC100506257', Aliases='', description='uncharacterized LOC100506257', other_designations='', map_location='1p21-p13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=110772526, end_position_on_the_genomic_accession=110774042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506258, CurrentID=0, Status='live', Symbol='LOC100506258', Aliases='', description='uncharacterized LOC100506258', other_designations='CTD-2516F10.2', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7437627, end_position_on_the_genomic_accession=7465835, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506271, CurrentID=0, Status='live', Symbol='LOC100506271', Aliases='', description='uncharacterized LOC100506271', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37352190, end_position_on_the_genomic_accession=37354839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506272, CurrentID=0, Status='live', Symbol='LOC100506272', Aliases='', description='uncharacterized LOC100506272', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187532878, end_position_on_the_genomic_accession=187672641, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506274, CurrentID=0, Status='live', Symbol='LOC100506274', Aliases='', description='uncharacterized LOC100506274', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7421261, end_position_on_the_genomic_accession=7450254, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506275, CurrentID=0, Status='live', Symbol='KRBOX1-AS1', Aliases='', description='KRBOX1 antisense RNA 1', other_designations='KRBOX1 antisense RNA 1 (non-protein coding)', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42934244, end_position_on_the_genomic_accession=42936785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506276, CurrentID=0, Status='live', Symbol='LOC100506276', Aliases='', description='uncharacterized LOC100506276', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=109197562, end_position_on_the_genomic_accession=109201487, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929178, CurrentID=0, Status='live', Symbol='LOC101929178', Aliases='', description='uncharacterized LOC101929178', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120069837, end_position_on_the_genomic_accession=120077091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506281, CurrentID=0, Status='live', Symbol='LOC100506281', Aliases='', description='uncharacterized LOC100506281', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75226383, end_position_on_the_genomic_accession=75228203, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506282, CurrentID=0, Status='live', Symbol='LOC100506282', Aliases='', description='uncharacterized LOC100506282', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83190906, end_position_on_the_genomic_accession=83193794, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506286, CurrentID=0, Status='live', Symbol='LOC100506286', Aliases='', description='uncharacterized LOC100506286', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101151583, end_position_on_the_genomic_accession=101155444, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506289, CurrentID=0, Status='live', Symbol='LOC100506289', Aliases='', description='uncharacterized LOC100506289', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26551839, end_position_on_the_genomic_accession=26557579, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506290, CurrentID=0, Status='live', Symbol='LOC100506290', Aliases='', description='uncharacterized LOC100506290', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=65360937, end_position_on_the_genomic_accession=65362009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418675, CurrentID=0, Status='live', Symbol='LOC100418675', Aliases='', description='olfactory receptor, family 2, subfamily B, member 2 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506299, CurrentID=0, Status='live', Symbol='ID2-AS1', Aliases='', description='ID2 antisense RNA 1 (head to head)', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8670984, end_position_on_the_genomic_accession=8681863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506301, CurrentID=0, Status='live', Symbol='LINC00694', Aliases='', description='long intergenic non-protein coding RNA 694', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=44462622, end_position_on_the_genomic_accession=44465517, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506302, CurrentID=0, Status='live', Symbol='LOC100506302', Aliases='', description='uncharacterized LOC100506302', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155641916, end_position_on_the_genomic_accession=155644406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506305, CurrentID=0, Status='live', Symbol='LINC00958', Aliases='', description='long intergenic non-protein coding RNA 958', other_designations='CTC-497E21.4', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12979534, end_position_on_the_genomic_accession=12989548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929917, CurrentID=0, Status='live', Symbol='LOC101929917', Aliases='', description='uncharacterized LOC101929917', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100418678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418678, CurrentID=0, Status='live', Symbol='LOC100418678', Aliases='', description='olfactory receptor, family 2, subfamily W, member 1 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506311, CurrentID=0, Status='live', Symbol='HOTAIRM1', Aliases='HOXA-AS1, HOXA1-AS1, NCRNA00179', description='HOXA transcript antisense RNA, myeloid-specific 1', other_designations='HOX antisense intergenic RNA myeloid 1|HOXA cluster antisense RNA 1 (non-protein coding)|HOXA transcript antisense RNA, myeloid-specific 1 (non-protein coding)', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27096094, end_position_on_the_genomic_accession=27100258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645086, CurrentID=0, Status='live', Symbol='LOC645086', Aliases='', description='chromosome 11 open reading frame 58 pseudogene', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44119418, end_position_on_the_genomic_accession=44120871, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506314, CurrentID=0, Status='live', Symbol='LOC100506314', Aliases='', description='uncharacterized LOC100506314', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12979837, end_position_on_the_genomic_accession=12984642, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376115, CurrentID=0, Status='live', Symbol='LOC105376115', Aliases='', description='uncharacterized LOC105376115', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83719180, end_position_on_the_genomic_accession=83734087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928144, CurrentID=0, Status='live', Symbol='LOC101928144', Aliases='', description='uncharacterized LOC101928144', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49499320, end_position_on_the_genomic_accession=49527494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506317, CurrentID=0, Status='live', Symbol='LOC100506317', Aliases='', description='uncharacterized LOC100506317', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9204574, end_position_on_the_genomic_accession=9206506, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506319, CurrentID=0, Status='live', Symbol='LOC100506319', Aliases='', description='uncharacterized LOC100506319', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44424141, end_position_on_the_genomic_accession=44429529, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506321, CurrentID=0, Status='live', Symbol='LOC100506321', Aliases='', description='uncharacterized LOC100506321', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65089918, end_position_on_the_genomic_accession=65094311, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377571, CurrentID=0, Status='live', Symbol='LOC105377571', Aliases='', description='uncharacterized LOC105377571', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182762148, end_position_on_the_genomic_accession=182767986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506325, CurrentID=0, Status='live', Symbol='SP2-AS1', Aliases='', description='SP2 antisense RNA 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47898386, end_position_on_the_genomic_accession=47941404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506326, CurrentID=0, Status='live', Symbol='LOC100506326', Aliases='', description='uncharacterized LOC100506326', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=39063590, end_position_on_the_genomic_accession=39077775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506328, CurrentID=0, Status='live', Symbol='LINC01127', Aliases='', description='long intergenic non-protein coding RNA 1127', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101983465, end_position_on_the_genomic_accession=101985545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506330, CurrentID=0, Status='live', Symbol='WBP1LP2', Aliases='', description='WW domain binding protein 1-like pseudogene 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107628466, end_position_on_the_genomic_accession=107629547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506331, CurrentID=0, Status='live', Symbol='LOC100506331', Aliases='', description='uncharacterized LOC100506331', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15887292, end_position_on_the_genomic_accession=15892688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287610, CurrentID=0, Status='live', Symbol='LOC100287610', Aliases='', description='zinc finger protein 73-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506334, CurrentID=0, Status='live', Symbol='LINC00649', Aliases='', description='long intergenic non-protein coding RNA 649', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33947934, end_position_on_the_genomic_accession=33965164, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506343, CurrentID=0, Status='live', Symbol='FAM212B-AS1', Aliases='', description='FAM212B antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111739841, end_position_on_the_genomic_accession=111747798, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287613, CurrentID=0, Status='live', Symbol='LOC100287613', Aliases='', description='Rho GTPase activating protein 20 pseudogene', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16849991, end_position_on_the_genomic_accession=16851001, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506354, CurrentID=0, Status='live', Symbol='LOC100506354', Aliases='', description='uncharacterized LOC100506354', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=224622614, end_position_on_the_genomic_accession=224709707, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506356, CurrentID=0, Status='live', Symbol='ARSD-AS1', Aliases='', description='ARSD antisense RNA 1', other_designations='ARSD antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=2822945, end_position_on_the_genomic_accession=2824188, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506357, CurrentID=0, Status='live', Symbol='LOC100506357', Aliases='', description='uncharacterized LOC100506357', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=9632432, end_position_on_the_genomic_accession=9633421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506358, CurrentID=0, Status='live', Symbol='LOC100506358', Aliases='', description='uncharacterized LOC100506358', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69811440, end_position_on_the_genomic_accession=69815967, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506365, CurrentID=0, Status='live', Symbol='OTUD6B-AS1', Aliases='GS1-251I9.4', description='OTUD6B antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91059906, end_position_on_the_genomic_accession=91070189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506368, CurrentID=0, Status='live', Symbol='LOC100506368', Aliases='', description='uncharacterized LOC100506368', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86955621, end_position_on_the_genomic_accession=87000950, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418688, CurrentID=0, Status='live', Symbol='LOC100418688', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506371, CurrentID=0, Status='live', Symbol='LOC100506371', Aliases='', description='uncharacterized LOC100506371', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=183824, end_position_on_the_genomic_accession=187251, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506374, CurrentID=0, Status='live', Symbol='LOC100506374', Aliases='', description='uncharacterized LOC100506374', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56264260, end_position_on_the_genomic_accession=56310454, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100506376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506376, CurrentID=0, Status='live', Symbol='TTLL10-AS1', Aliases='', description='TTLL10 antisense RNA 1', other_designations='TTLL10 antisense RNA 1 (non-protein coding)', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1173055, end_position_on_the_genomic_accession=1179555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506377, CurrentID=0, Status='live', Symbol='LINC00973', Aliases='', description='long intergenic non-protein coding RNA 973', other_designations='CTD-2021J15.2', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=98699902, end_position_on_the_genomic_accession=98701876, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506380, CurrentID=0, Status='live', Symbol='LINC01006', Aliases='C7orf13, MY040', description='long intergenic non-protein coding RNA 1006', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156472099, end_position_on_the_genomic_accession=156603195, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377581, CurrentID=0, Status='live', Symbol='LOC105377581', Aliases='', description='uncharacterized LOC105377581', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183725789, end_position_on_the_genomic_accession=183733808, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506384, CurrentID=0, Status='live', Symbol='LOC100506384', Aliases='', description='uncharacterized LOC100506384', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59626464, end_position_on_the_genomic_accession=59628289, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506385, CurrentID=0, Status='live', Symbol='LINC01426', Aliases='lincRNA-uc002yug.2', description='long intergenic non-protein coding RNA 1426', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34745825, end_position_on_the_genomic_accession=34784886, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100418691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418691, CurrentID=0, Status='live', Symbol='LOC100418691', Aliases='', description='N-ethylmaleimide-sensitive factor attachment protein, gamma pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506388, CurrentID=0, Status='live', Symbol='LOC100506388', Aliases='', description='uncharacterized LOC100506388', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=331205, end_position_on_the_genomic_accession=333488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377582, CurrentID=0, Status='live', Symbol='LOC105377582', Aliases='', description='uncharacterized LOC105377582', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184262609, end_position_on_the_genomic_accession=184293682, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100506390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506390, CurrentID=0, Status='live', Symbol='RAP1BP1', Aliases='', description='RAP1B, member of RAS oncogene family pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18718111, end_position_on_the_genomic_accession=18718527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506392, CurrentID=0, Status='live', Symbol='SLC16A1-AS1', Aliases='', description='SLC16A1 antisense RNA 1', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112956415, end_position_on_the_genomic_accession=112964072, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506393, CurrentID=0, Status='live', Symbol='LOC100506393', Aliases='', description='uncharacterized LOC100506393', other_designations='', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20014685, end_position_on_the_genomic_accession=20098868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506394, CurrentID=0, Status='live', Symbol='GAS6-AS2', Aliases='', description='GAS6 antisense RNA 2 (head to head)', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113864177, end_position_on_the_genomic_accession=113866832, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377583, CurrentID=0, Status='live', Symbol='LOC105377583', Aliases='', description='uncharacterized LOC105377583', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184287690, end_position_on_the_genomic_accession=184296524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506399, CurrentID=0, Status='live', Symbol='LOC100506399', Aliases='', description='FLJ00363 protein', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=3751387, end_position_on_the_genomic_accession=3751836, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506400, CurrentID=0, Status='live', Symbol='LOC100506400', Aliases='', description='transmembrane protein 258 pseudogene', other_designations='UPF0197 transmembrane protein C11orf10 homolog', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99925649, end_position_on_the_genomic_accession=99925890, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506403, CurrentID=0, Status='live', Symbol='LOC100506403', Aliases='', description='uncharacterized LOC100506403', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35372507, end_position_on_the_genomic_accession=35580764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506405, CurrentID=0, Status='live', Symbol='LOC100506405', Aliases='', description='uncharacterized LOC100506405', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11681460, end_position_on_the_genomic_accession=11683323, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506406, CurrentID=0, Status='live', Symbol='LOC100506406', Aliases='', description='uncharacterized LOC100506406', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35938801, end_position_on_the_genomic_accession=35939993, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506409, CurrentID=0, Status='live', Symbol='ELOVL2-AS1', Aliases='', description='ELOVL2 antisense RNA 1', other_designations='ELOVL2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11043758, end_position_on_the_genomic_accession=11079144, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506411, CurrentID=0, Status='live', Symbol='LOC100506411', Aliases='', description='uncharacterized LOC100506411', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70809199, end_position_on_the_genomic_accession=70816048, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506412, CurrentID=0, Status='live', Symbol='LINC00871', Aliases='', description='long intergenic non-protein coding RNA 871', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46064159, end_position_on_the_genomic_accession=46501903, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100506413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506413, CurrentID=0, Status='live', Symbol='DOCK4-AS1', Aliases='', description='DOCK4 antisense RNA 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111808505, end_position_on_the_genomic_accession=111821779, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506417, CurrentID=0, Status='live', Symbol='LOC100506417', Aliases='', description='uncharacterized LOC100506417', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=84091428, end_position_on_the_genomic_accession=84092795, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506421, CurrentID=0, Status='live', Symbol='LINC01158', Aliases='PANTR1, linc-Brn1a, linc-POU3F3', description='long intergenic non-protein coding RNA 1158', other_designations='', map_location='2q12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104805425, end_position_on_the_genomic_accession=104851476, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506422, CurrentID=0, Status='live', Symbol='LOC100506422', Aliases='', description="putative deoxyuridine 5'-triphosphate nucleotidohydrolase-like protein FLJ16323", other_designations='dUTPase-like protein', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26066675, end_position_on_the_genomic_accession=26118408, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100506424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506424, CurrentID=0, Status='live', Symbol='UBE2L2', Aliases='', description='ubiquitin conjugating enzyme E2L 2 (pseudogene)', other_designations='ubiquitin-conjugating enzyme E2L 2 (pseudogene)', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20449438, end_position_on_the_genomic_accession=20450133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506428, CurrentID=0, Status='live', Symbol='CBR3-AS1', Aliases='PlncRNA-1, PlncRNA1', description='CBR3 antisense RNA 1', other_designations='CBR3 antisense RNA 1 (non-protein coding)|prostate cancer-up-regulated long noncoding RNA 1', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36131767, end_position_on_the_genomic_accession=36156308, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506431, CurrentID=0, Status='live', Symbol='LOC100506431', Aliases='', description='B2 RNA', other_designations='', map_location='6p25-p24', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=11198541, end_position_on_the_genomic_accession=11199708, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=317751, CurrentID=0, Status='live', Symbol='MESTIT1', Aliases='MEST-AS1, MEST-IT, MEST-IT1, NCRNA00040, PEG1-AS', description='MEST intronic transcript 1, antisense RNA', other_designations='MEST intronic transcript 1 (non-protein coding)|paternally expressed gene, antisense transcript', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130487057, end_position_on_the_genomic_accession=130491172, orientation='minus', exon_count=2, OMIM=607794, no_hdr0=''),
  100506436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506436, CurrentID=0, Status='live', Symbol='SPAG5-AS1', Aliases='', description='SPAG5 antisense RNA 1', other_designations='SPAG5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28599133, end_position_on_the_genomic_accession=28617377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506443, CurrentID=0, Status='live', Symbol='ITPKB-IT1', Aliases='', description='ITPKB intronic transcript 1', other_designations='ITPKB intronic transcript 1 (non-protein coding)', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226656080, end_position_on_the_genomic_accession=226675068, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506444, CurrentID=0, Status='live', Symbol='LOC100506444', Aliases='', description='uncharacterized LOC100506444', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53695901, end_position_on_the_genomic_accession=53734358, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506446, CurrentID=0, Status='live', Symbol='LOC100506446', Aliases='', description='uncharacterized LOC100506446', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49919814, end_position_on_the_genomic_accession=49961960, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506447, CurrentID=0, Status='live', Symbol='LOC100506447', Aliases='', description='uncharacterized LOC100506447', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100506451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506451, CurrentID=0, Status='live', Symbol='RASSF8-AS1', Aliases='', description='RASSF8 atnisense RNA 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25954654, end_position_on_the_genomic_accession=25959442, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377593, CurrentID=0, Status='live', Symbol='LOC105377593', Aliases='', description='uncharacterized LOC105377593', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185990125, end_position_on_the_genomic_accession=186022965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506457, CurrentID=0, Status='live', Symbol='MIR3681HG', Aliases='', description='MIR3681 host gene', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12007116, end_position_on_the_genomic_accession=12578348, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100506458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506458, CurrentID=0, Status='live', Symbol='LOC100506458', Aliases='', description='putative uncharacterized protein LOC65996-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=157056843, end_position_on_the_genomic_accession=157059007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506459, CurrentID=0, Status='live', Symbol='LOC100506459', Aliases='', description='uncharacterized LOC100506459', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=117035638, end_position_on_the_genomic_accession=117041834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506462, CurrentID=0, Status='live', Symbol='SRD5A3-AS1', Aliases='', description='SRD5A3 antisense RNA 1', other_designations='SRD5A3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55366601, end_position_on_the_genomic_accession=55385580, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506465, CurrentID=0, Status='live', Symbol='LINC01234', Aliases='LCAL84, onco-lncRNA-32', description='long intergenic non-protein coding RNA 1234', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113744577, end_position_on_the_genomic_accession=113773683, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506469, CurrentID=0, Status='live', Symbol='TMEM147-AS1', Aliases='', description='TMEM147 antisense RNA 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35540738, end_position_on_the_genomic_accession=35546029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506470, CurrentID=0, Status='live', Symbol='LOC100506470', Aliases='', description='uncharacterized LOC100506470', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61079064, end_position_on_the_genomic_accession=61080179, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506472, CurrentID=0, Status='live', Symbol='LOC100506472', Aliases='', description='uncharacterized LOC100506472', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39432160, end_position_on_the_genomic_accession=39437128, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506473, CurrentID=0, Status='live', Symbol='LOC100506473', Aliases='', description='uncharacterized LOC100506473', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105334034, end_position_on_the_genomic_accession=105337278, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506474, CurrentID=0, Status='live', Symbol='LOC100506474', Aliases='', description='uncharacterized LOC100506474', other_designations='', map_location='2p24', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12966782, end_position_on_the_genomic_accession=13007013, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506475, CurrentID=0, Status='live', Symbol='EGFLAM-AS2', Aliases='', description='EGFLAM antisense RNA 2', other_designations='CTD-2108O9.2|EGFLAM antisense RNA 2 (non-protein coding)', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38399814, end_position_on_the_genomic_accession=38403471, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506476, CurrentID=0, Status='live', Symbol='LOC100506476', Aliases='', description='uncharacterized LOC100506476', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73787360, end_position_on_the_genomic_accession=73803628, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377597, CurrentID=0, Status='live', Symbol='LOC105377597', Aliases='', description='uncharacterized LOC105377597', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186666515, end_position_on_the_genomic_accession=186673420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418707, CurrentID=0, Status='live', Symbol='LOC100418707', Aliases='', description='septin 7 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=119126757, end_position_on_the_genomic_accession=119128444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506486, CurrentID=0, Status='live', Symbol='LOC100506486', Aliases='', description='uncharacterized LOC100506486', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=56287451, end_position_on_the_genomic_accession=56289843, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506489, CurrentID=0, Status='live', Symbol='LOC100506489', Aliases='', description='uncharacterized LOC100506489', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115679345, end_position_on_the_genomic_accession=115682618, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377599, CurrentID=0, Status='live', Symbol='LOC105377599', Aliases='', description='uncharacterized LOC105377599', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186891259, end_position_on_the_genomic_accession=186892983, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506492, CurrentID=0, Status='live', Symbol='DSCAM-AS1', Aliases='M41', description='DSCAM antisense RNA 1', other_designations='DSCAM antisense RNA 1 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40383083, end_position_on_the_genomic_accession=40385358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506495, CurrentID=0, Status='live', Symbol='LIFR-AS1', Aliases='', description='LIFR antisense RNA 1', other_designations='CTD-2196P11.1|LIFR antisense RNA 1 (non-protein coding)', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38556786, end_position_on_the_genomic_accession=38671216, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100506497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506497, CurrentID=0, Status='live', Symbol='LOC100506497', Aliases='', description='uncharacterized LOC100506497', other_designations='CTB-113D17.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28979967, end_position_on_the_genomic_accession=29013369, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506498, CurrentID=0, Status='live', Symbol='LOC100506498', Aliases='', description='uncharacterized LOC100506498', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73822559, end_position_on_the_genomic_accession=73830285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506504, CurrentID=0, Status='live', Symbol='LOC100506504', Aliases='', description='uncharacterized LOC100506504', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2181811, end_position_on_the_genomic_accession=2184360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506506, CurrentID=0, Status='live', Symbol='MORC1-AS1', Aliases='', description='MORC1 antisense RNA 1', other_designations='MORC1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=108820303, end_position_on_the_genomic_accession=108829187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418711, CurrentID=0, Status='live', Symbol='CLUHP4', Aliases='KIAA0664P4', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190061079, end_position_on_the_genomic_accession=190062996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422391, CurrentID=0, Status='live', Symbol='POLR3GP2', Aliases='', description='polymerase (RNA) III (DNA directed) polypeptide G (32kD) pseudogene 2', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48478518, end_position_on_the_genomic_accession=48479422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506513, CurrentID=0, Status='live', Symbol='CATX-1', Aliases='', description='uncharacterized LOC100506513', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=106499480, end_position_on_the_genomic_accession=106499838, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506514, CurrentID=0, Status='live', Symbol='LOC100506514', Aliases='', description='uncharacterized LOC100506514', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56408789, end_position_on_the_genomic_accession=56429005, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377603, CurrentID=0, Status='live', Symbol='LOC105377603', Aliases='', description='uncharacterized LOC105377603', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187613340, end_position_on_the_genomic_accession=187712233, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100506516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506516, CurrentID=0, Status='live', Symbol='LINC01176', Aliases='', description='long intergenic non-protein coding RNA 1176', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30390885, end_position_on_the_genomic_accession=30404055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418714, CurrentID=0, Status='live', Symbol='LOC100418714', Aliases='', description='septin 7 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57067949, end_position_on_the_genomic_accession=57074683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506526, CurrentID=0, Status='live', Symbol='LOC100506526', Aliases='', description='uncharacterized LOC100506526', other_designations='CTD-2170G1.2', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61732790, end_position_on_the_genomic_accession=61735698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506527, CurrentID=0, Status='live', Symbol='LINC01007', Aliases='', description='long intergenic non-protein coding RNA 1007', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101562755, end_position_on_the_genomic_accession=101569006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506530, CurrentID=0, Status='live', Symbol='LOC100506530', Aliases='', description='uncharacterized LOC100506530', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61729850, end_position_on_the_genomic_accession=61731389, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506532, CurrentID=0, Status='live', Symbol='LOC100506532', Aliases='', description='uncharacterized LOC100506532', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134486281, end_position_on_the_genomic_accession=134545232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377607, CurrentID=0, Status='live', Symbol='LOC105377607', Aliases='', description='uncharacterized LOC105377607', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188175272, end_position_on_the_genomic_accession=188181120, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506540, CurrentID=0, Status='live', Symbol='SPTY2D1-AS1', Aliases='', description='SPTY2D1 antisense RNA 1', other_designations='SPTY2D1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18599804, end_position_on_the_genomic_accession=18610255, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506542, CurrentID=0, Status='live', Symbol='LINC01082', Aliases='TCONS_00024492', description='long intergenic non-protein coding RNA 1082', other_designations='lncRNA TCONS_00024492|long noncoding RNA TCONS_00024492', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=86196181, end_position_on_the_genomic_accession=86199720, orientation='plus', exon_count=2, OMIM=614978, no_hdr0=''),
  100506544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506544, CurrentID=0, Status='live', Symbol='LOC100506544', Aliases='', description='uncharacterized LOC100506544', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=41605126, end_position_on_the_genomic_accession=41613631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377608, CurrentID=0, Status='live', Symbol='LOC105377608', Aliases='', description='uncharacterized LOC105377608', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188376283, end_position_on_the_genomic_accession=188378118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506548, CurrentID=0, Status='live', Symbol='LOC100506548', Aliases='', description='uncharacterized LOC100506548', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40825262, end_position_on_the_genomic_accession=40829142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  148638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=148638, CurrentID=0, Status='live', Symbol='LOC148638', Aliases='', description='uncharacterized LOC148638', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377609, CurrentID=0, Status='live', Symbol='LOC105377609', Aliases='', description='uncharacterized LOC105377609', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188433216, end_position_on_the_genomic_accession=188460145, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506555, CurrentID=0, Status='live', Symbol='PVRL3-AS1', Aliases='', description='PVRL3 antisense RNA 1', other_designations='PVRL3 antisense RNA 1 (non-protein coding)', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111045315, end_position_on_the_genomic_accession=111069959, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506557, CurrentID=0, Status='live', Symbol='LOC100506557', Aliases='', description='uncharacterized LOC100506557', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157525740, end_position_on_the_genomic_accession=157527319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506562, CurrentID=0, Status='live', Symbol='LOC100506562', Aliases='', description='protein LLP homolog', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=42072552, end_position_on_the_genomic_accession=42073252, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100527947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527947, CurrentID=0, Status='live', Symbol='PDCD5P2', Aliases='', description='programmed cell death 5 pseudogene 2', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76280916, end_position_on_the_genomic_accession=76281401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506564, CurrentID=0, Status='live', Symbol='THEGL', Aliases='', description='theg spermatid protein-like', other_designations='theg spermatid-like protein', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56530575, end_position_on_the_genomic_accession=56641550, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100418721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418721, CurrentID=0, Status='live', Symbol='LOC100418721', Aliases='', description='formin 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26496936, end_position_on_the_genomic_accession=26505603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377612, CurrentID=0, Status='live', Symbol='LOC105377612', Aliases='', description='uncharacterized LOC105377612', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189058981, end_position_on_the_genomic_accession=189096944, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506571, CurrentID=0, Status='live', Symbol='LOC100506571', Aliases='', description='uncharacterized LOC100506571', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228073909, end_position_on_the_genomic_accession=228078006, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506574, CurrentID=0, Status='live', Symbol='SNRPGP9', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 9', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109251334, end_position_on_the_genomic_accession=109251839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377613, CurrentID=0, Status='live', Symbol='LOC105377613', Aliases='', description='uncharacterized LOC105377613', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189114216, end_position_on_the_genomic_accession=189126175, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506576, CurrentID=0, Status='live', Symbol='LOC100506576', Aliases='', description='ubiquitin-conjugating enzyme E2L 3 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76013644, end_position_on_the_genomic_accession=76014334, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506578, CurrentID=0, Status='live', Symbol='LOC100506578', Aliases='', description='uncharacterized protein C12orf71-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27590445, end_position_on_the_genomic_accession=27595622, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129262, CurrentID=0, Status='live', Symbol='LOC100129262', Aliases='', description='ribosomal protein L28 pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19055841, end_position_on_the_genomic_accession=19056187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506581, CurrentID=0, Status='live', Symbol='C16orf95', Aliases='', description='chromosome 16 open reading frame 95', other_designations='', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87302798, end_position_on_the_genomic_accession=87317420, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100506585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506585, CurrentID=0, Status='live', Symbol='LOC100506585', Aliases='', description='uncharacterized LOC100506585', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157854585, end_position_on_the_genomic_accession=157866092, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100113373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113373, CurrentID=0, Status='live', Symbol='HMGN2P17', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 17', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 17', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8893411, end_position_on_the_genomic_accession=8894603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100113374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113374, CurrentID=0, Status='live', Symbol='LOC100113374', Aliases='', description='tRNA nucleotidyl transferase, CCA-adding, 1 pseudogene', other_designations='', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163237206, end_position_on_the_genomic_accession=163239987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100287653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287653, CurrentID=0, Status='live', Symbol='LOC100287653', Aliases='', description='PDGFA associated protein 1 pseudogene', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119194850, end_position_on_the_genomic_accession=119195391, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100113378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113378, CurrentID=0, Status='live', Symbol='SNORD119', Aliases='', description='small nucleolar RNA, C/D box 119', other_designations='SNORD119 snoRNA', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2462952, end_position_on_the_genomic_accession=2463047, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100113379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113379, CurrentID=0, Status='live', Symbol='SNORD121A', Aliases='', description='small nucleolar RNA, C/D box 121A', other_designations='SNORD121A snoRNA', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33952764, end_position_on_the_genomic_accession=33952854, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100113380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113380, CurrentID=0, Status='live', Symbol='SNORD125', Aliases='', description='small nucleolar RNA, C/D box 125', other_designations='SNORD125 snoRNA', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29333163, end_position_on_the_genomic_accession=29333258, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100113381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113381, CurrentID=0, Status='live', Symbol='SNORD19B', Aliases='', description='small nucleolar RNA, C/D box 19B', other_designations='SNORD117 snoRNA', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52690738, end_position_on_the_genomic_accession=52690830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113382, CurrentID=0, Status='live', Symbol='SNORD105B', Aliases='', description='small nucleolar RNA, C/D box 105B', other_designations='SNORD105B snoRNA', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10109749, end_position_on_the_genomic_accession=10109840, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113383, CurrentID=0, Status='live', Symbol='PCBP3-OT1', Aliases='PCBP3OT', description='PCBP3 overlapping transcript 1', other_designations='PCBP3 overlapping transcript (non-protein coding)|poly(rC) binding protein 3 overlapping transcript', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100113384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113384, CurrentID=0, Status='live', Symbol='SNORD123', Aliases='', description='small nucleolar RNA, C/D box 123', other_designations='SNORD123 snoRNA', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9548827, end_position_on_the_genomic_accession=9548914, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113386, CurrentID=0, Status='live', Symbol='UCKL1-AS1', Aliases='UCKL1-AS, UCKL1AS, UCKL1OS', description='UCKL1 antisense RNA 1', other_designations='UCKL1 antisense RNA (non-protein coding)|UCKL1 antisense RNA 1 (non-protein coding)|Uncharacterized protein UCKL1AS|Uridine/cytidine kinase-like 1 opposite strand transcript protein|uridine-cytidine kinase 1-like 1 opposite strand', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63953384, end_position_on_the_genomic_accession=63956985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506603, CurrentID=0, Status='live', Symbol='LOC100506603', Aliases='', description='uncharacterized LOC100506603', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76781733, end_position_on_the_genomic_accession=76786724, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100113389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113389, CurrentID=0, Status='live', Symbol='SNORD127', Aliases='', description='small nucleolar RNA, C/D box 127', other_designations='SNORD118 snoRNA', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45110875, end_position_on_the_genomic_accession=45110975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506606, CurrentID=0, Status='live', Symbol='LOC100506606', Aliases='', description='uncharacterized LOC100506606', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29280415, end_position_on_the_genomic_accession=29317848, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100113391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113391, CurrentID=0, Status='live', Symbol='SNORD126', Aliases='MIR1201, MIRN1201', description='small nucleolar RNA, C/D box 126', other_designations='SNORD126 snoRNA|hsa-mir-1201|microRNA 1201', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20326441, end_position_on_the_genomic_accession=20326539, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100113392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113392, CurrentID=0, Status='live', Symbol='SNORD11B', Aliases='', description='small nucleolar RNA, C/D box 11B', other_designations='SNORD11B snoRNA', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202291317, end_position_on_the_genomic_accession=202291428, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113393, CurrentID=0, Status='live', Symbol='SNORD12B', Aliases='', description='small nucleolar RNA, C/D box 12B', other_designations='SNORD12B snoRNA', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49280313, end_position_on_the_genomic_accession=49280415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113397, CurrentID=0, Status='live', Symbol='LOC100113397', Aliases='', description='vezatin, adherens junctions transmembrane protein pseudogene', other_designations='', map_location='Xp', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46453460, end_position_on_the_genomic_accession=46457244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377620, CurrentID=0, Status='live', Symbol='LOC105377620', Aliases='', description='uncharacterized LOC105377620', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189808190, end_position_on_the_genomic_accession=189817564, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100113402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113402, CurrentID=0, Status='live', Symbol='SNORD111B', Aliases='MIR3647', description='small nucleolar RNA, C/D box 111B', other_designations='SNORD111B snoRNA|hsa-mir-3647|microRNA 3647', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70529499, end_position_on_the_genomic_accession=70529599, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100113403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113403, CurrentID=0, Status='live', Symbol='LINC00577', Aliases='C6orf220, dJ439I14.1', description='long intergenic non-protein coding RNA 577', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104936294, end_position_on_the_genomic_accession=104940527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100113404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113404, CurrentID=0, Status='live', Symbol='LINC00583', Aliases='C9orf146', description='long intergenic non-protein coding RNA 583', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13927971, end_position_on_the_genomic_accession=13945607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506621, CurrentID=0, Status='live', Symbol='LINC01279', Aliases='', description='long intergenic non-protein coding RNA 1279', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112596794, end_position_on_the_genomic_accession=112601969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506622, CurrentID=0, Status='live', Symbol='LINC00540', Aliases='', description='long intergenic non-protein coding RNA 540', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22210285, end_position_on_the_genomic_accession=22276521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100113407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113407, CurrentID=0, Status='live', Symbol='TMEM170B', Aliases='', description='transmembrane protein 170B', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11538196, end_position_on_the_genomic_accession=11583524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506627, CurrentID=0, Status='live', Symbol='DCDC5', Aliases='', description='doublecortin domain containing 5', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30863603, end_position_on_the_genomic_accession=30992686, orientation='minus', exon_count=20, OMIM=612321, no_hdr0=''),
  105371761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371761, CurrentID=0, Status='live', Symbol='LOC105371761', Aliases='', description='uncharacterized LOC105371761', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38564330, end_position_on_the_genomic_accession=38565391, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506633, CurrentID=0, Status='live', Symbol='LOC100506633', Aliases='', description='ubiquitin-fold modifier 1 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100164330, end_position_on_the_genomic_accession=100165543, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506634, CurrentID=0, Status='live', Symbol='LOC100506634', Aliases='', description='hepatocellular carcinoma-associated antigen HCA25a', other_designations='CTD-2619J13.14', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58403204, end_position_on_the_genomic_accession=58408457, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377623, CurrentID=0, Status='live', Symbol='LOC105377623', Aliases='', description='uncharacterized LOC105377623', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143177737, end_position_on_the_genomic_accession=143185297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506637, CurrentID=0, Status='live', Symbol='PRKAR2A-AS1', Aliases='', description='PRKAR2A antisense RNA 1', other_designations='PRKAR2A antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48847937, end_position_on_the_genomic_accession=48851982, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506639, CurrentID=0, Status='live', Symbol='LOC100506639', Aliases='', description='uncharacterized LOC100506639', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43067023, end_position_on_the_genomic_accession=43093558, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377624, CurrentID=0, Status='live', Symbol='LOC105377624', Aliases='', description='uncharacterized LOC105377624', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185570102, end_position_on_the_genomic_accession=185580930, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728023, CurrentID=0, Status='live', Symbol='H2AFZP3', Aliases='', description='H2A histone family member Z pseudogene 3', other_designations='H2A histone family pseudogene|H2A histone family, member Z pseudogene 3', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99215172, end_position_on_the_genomic_accession=99215862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506649, CurrentID=0, Status='live', Symbol='PXN-AS1', Aliases='EyeLinc4', description='PXN antisense RNA 1', other_designations='PXN antisense RNA 1 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120201291, end_position_on_the_genomic_accession=120212828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506650, CurrentID=0, Status='live', Symbol='C17orf112', Aliases='', description='chromosome 17 open reading frame 112', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=52983991, end_position_on_the_genomic_accession=52987658, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506654, CurrentID=0, Status='live', Symbol='LOC100506654', Aliases='', description='gem nuclear organelle associated protein 7 pseudogene', other_designations='gem (nuclear organelle) associated protein 7 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41472261, end_position_on_the_genomic_accession=41472653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506655, CurrentID=0, Status='live', Symbol='LCMT1-AS2', Aliases='', description='LCMT1 antisense RNA 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25140571, end_position_on_the_genomic_accession=25149032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287664, CurrentID=0, Status='live', Symbol='TSRM', Aliases='', description='zinc finger domain-related protein TSRM', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=113091130, end_position_on_the_genomic_accession=113091456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506658, CurrentID=0, Status='live', Symbol='OCLN', Aliases='BLCPMG, PPP1R115', description='occludin', other_designations='phosphatase 1, regulatory subunit 115|tight junction protein occludin', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69492292, end_position_on_the_genomic_accession=69558104, orientation='plus', exon_count=10, OMIM=602876, no_hdr0=''),
  126811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126811, CurrentID=0, Status='live', Symbol='KRT8P21', Aliases='', description='keratin 8 pseudogene 21', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=73104779, end_position_on_the_genomic_accession=73106484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506660, CurrentID=0, Status='live', Symbol='DDX11-AS1', Aliases='', description='DDX11 antisense RNA 1', other_designations='DDX11 antisense RNA 1 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31020763, end_position_on_the_genomic_accession=31073847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506664, CurrentID=0, Status='live', Symbol='LOC100506664', Aliases='', description='uncharacterized LOC100506664', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125437855, end_position_on_the_genomic_accession=125466310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377628, CurrentID=0, Status='live', Symbol='LOC105377628', Aliases='', description='uncharacterized LOC105377628', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57782225, end_position_on_the_genomic_accession=57785883, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506667, CurrentID=0, Status='live', Symbol='LOC100506667', Aliases='', description='uncharacterized LOC100506667', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=98532606, end_position_on_the_genomic_accession=98534956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506668, CurrentID=0, Status='live', Symbol='NRAV', Aliases='DYNLL1-AS1, DYNLL1AS1', description='negative regulator of antiviral response (non-protein coding)', other_designations='DYNLL1 antisense RNA 1', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120490338, end_position_on_the_genomic_accession=120495946, orientation='minus', exon_count=2, OMIM=616207, no_hdr0=''),
  100506671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506671, CurrentID=0, Status='live', Symbol='PHKA1-AS1', Aliases='', description='PHKA1 antisense RNA 1', other_designations='PHKA1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=71908799, end_position_on_the_genomic_accession=71932190, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506672, CurrentID=0, Status='live', Symbol='LOC100506672', Aliases='', description='cytochrome c oxidase subunit VIIc pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202767095, end_position_on_the_genomic_accession=202767508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506673, CurrentID=0, Status='live', Symbol='ZBTB20-AS2', Aliases='', description='ZBTB20 antisense RNA 2', other_designations='ZBTB20 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=114403427, end_position_on_the_genomic_accession=114406469, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506674, CurrentID=0, Status='live', Symbol='BRCAT54', Aliases='', description='breast cancer-associated transcript 54', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=44744328, end_position_on_the_genomic_accession=44808793, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506675, CurrentID=0, Status='live', Symbol='LOC100506675', Aliases='', description='uncharacterized LOC100506675', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32035979, end_position_on_the_genomic_accession=32041318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506677, CurrentID=0, Status='live', Symbol='AA06', Aliases='AA02, AA04, AA05', description='uncharacterized LOC100506677', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=33529787, end_position_on_the_genomic_accession=33533760, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506679, CurrentID=0, Status='live', Symbol='LOC100506679', Aliases='', description='uncharacterized LOC100506679', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43038585, end_position_on_the_genomic_accession=43052366, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506680, CurrentID=0, Status='live', Symbol='SACS-AS1', Aliases='', description='SACS antisense RNA 1', other_designations='SACS antisense RNA 1 (non-protein coding)', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23418971, end_position_on_the_genomic_accession=23428869, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506681, CurrentID=0, Status='live', Symbol='JARID2-AS1', Aliases='', description='JARID2 antisense RNA 1', other_designations='JARID2 antisense RNA 1 (non-protein coding)', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15247815, end_position_on_the_genomic_accession=15248696, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506682, CurrentID=0, Status='live', Symbol='LOC100506682', Aliases='', description='uncharacterized LOC100506682', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127476883, end_position_on_the_genomic_accession=127485804, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506685, CurrentID=0, Status='live', Symbol='STMN1P1', Aliases='', description='stathmin 1 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31715144, end_position_on_the_genomic_accession=31715878, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506686, CurrentID=0, Status='live', Symbol='IQCH-AS1', Aliases='', description='IQCH antisense RNA 1', other_designations='IQCH antisense RNA 1 (non-protein coding)', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67403611, end_position_on_the_genomic_accession=67521844, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129974, CurrentID=0, Status='live', Symbol='LOC100129974', Aliases='', description='uncharacterized LOC100129974', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=387874, end_position_on_the_genomic_accession=388251, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506688, CurrentID=0, Status='live', Symbol='LOC100506688', Aliases='', description='uncharacterized LOC100506688', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=987177, end_position_on_the_genomic_accession=997340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377632, CurrentID=0, Status='live', Symbol='LOC105377632', Aliases='', description='uncharacterized LOC105377632', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57289876, end_position_on_the_genomic_accession=57886242, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100506691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506691, CurrentID=0, Status='live', Symbol='LOC100506691', Aliases='', description='uncharacterized LOC100506691', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122063253, end_position_on_the_genomic_accession=122068560, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  102997063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102997063, CurrentID=0, Status='live', Symbol='DEL15Q11.2', Aliases='C15DELq11.2', description='Chromosome 15q11.2 deletion syndrome', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615656, no_hdr0=''),
  100506696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506696, CurrentID=0, Status='live', Symbol='PCAT6', Aliases='KDM5B-AS1, KDM5BAS1, PCAN-R1, ncRNA-a2, onco-lncRNA-96', description='prostate cancer associated transcript 6 (non-protein coding)', other_designations='KDM5B antisense RNA 1 (head to head)|KDM5B antisense RNA 1 (non-protein coding)|noncoding RNA-activating 2|prostate cancer-associated noncoding RNA 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202810946, end_position_on_the_genomic_accession=202811913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102997065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102997065, CurrentID=0, Status='live', Symbol='DDD3', Aliases='', description='Dowling-Degos disease 3', other_designations='', map_location='17q21.3-q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615674, no_hdr0=''),
  105370885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370885, CurrentID=0, Status='live', Symbol='LOC105370885', Aliases='', description='uncharacterized LOC105370885', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71226285, end_position_on_the_genomic_accession=71241439, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506699, CurrentID=0, Status='live', Symbol='LOC100506699', Aliases='', description='uncharacterized LOC100506699', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=33972436, end_position_on_the_genomic_accession=33972933, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506700, CurrentID=0, Status='live', Symbol='LOC100506700', Aliases='', description='uncharacterized LOC100506700', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81441987, end_position_on_the_genomic_accession=81450157, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377634, CurrentID=0, Status='live', Symbol='LOC105377634', Aliases='', description='uncharacterized LOC105377634', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64488748, end_position_on_the_genomic_accession=64506810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506706, CurrentID=0, Status='live', Symbol='CICP19', Aliases='', description='capicua transcriptional repressor pseudogene 19', other_designations='capicua homolog pseudogene', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=193198, end_position_on_the_genomic_accession=196968, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377635, CurrentID=0, Status='live', Symbol='LOC105377635', Aliases='', description='uncharacterized LOC105377635', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87349152, end_position_on_the_genomic_accession=87353057, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506708, CurrentID=0, Status='live', Symbol='LSAMP-AS1', Aliases='', description='LSAMP antisense RNA 1', other_designations='LSAMP antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=116078871, end_position_on_the_genomic_accession=116088932, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506710, CurrentID=0, Status='live', Symbol='EBLN3', Aliases='', description='endogenous Bornavirus-like nucleoprotein 3', other_designations='endogenous Bornavirus-like nucleoprotein 2 pseudogene', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37079896, end_position_on_the_genomic_accession=37090401, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506712, CurrentID=0, Status='live', Symbol='HNRNPA1P5', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 5', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67627768, end_position_on_the_genomic_accession=67629469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506713, CurrentID=0, Status='live', Symbol='ALOX12-AS1', Aliases='', description='ALOX12 antisense RNA 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6985123, end_position_on_the_genomic_accession=7012334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506714, CurrentID=0, Status='live', Symbol='NUP50-AS1', Aliases='', description='NUP50 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45133758, end_position_on_the_genomic_accession=45163781, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506715, CurrentID=0, Status='live', Symbol='PNKDP1', Aliases='', description='PNKD pseudogene 1', other_designations='paroxysmal nonkinesigenic dyskinesia pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131803129, end_position_on_the_genomic_accession=131803756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506718, CurrentID=0, Status='live', Symbol='LOC100506718', Aliases='', description='uncharacterized LOC100506718', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377637, CurrentID=0, Status='live', Symbol='LOC105377637', Aliases='', description='transcription initiation factor TFIID subunit 4-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23382827, end_position_on_the_genomic_accession=23402338, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100506721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506721, CurrentID=0, Status='live', Symbol='MTND1P36', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 36', other_designations='MT-ND1 pseudogene 36', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103407441, end_position_on_the_genomic_accession=103408376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506723, CurrentID=0, Status='live', Symbol='TMSB10P1', Aliases='', description='thymosin beta 10 pseudogene 1', other_designations='', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235882117, end_position_on_the_genomic_accession=235882527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506724, CurrentID=0, Status='live', Symbol='LINC00901', Aliases='LSAMP-AS4, TCONS_00005428', description='long intergenic non-protein coding RNA 901', other_designations='LSAMP antisense RNA 4 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116921431, end_position_on_the_genomic_accession=116932238, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506725, CurrentID=0, Status='live', Symbol='LOC100506725', Aliases='', description='uncharacterized LOC100506725', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35716437, end_position_on_the_genomic_accession=35719712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170513, CurrentID=0, Status='live', Symbol='FAT1P1', Aliases='dJ697P8.1', description='FAT atypical cadherin 1 pseudogene 1', other_designations='FAT tumor suppressor homolog 1 pseudogene 1|FAT tumor suppressor homolog pseudogene 1', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10888568, end_position_on_the_genomic_accession=10889090, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506730, CurrentID=0, Status='live', Symbol='LOC100506730', Aliases='', description='uncharacterized LOC100506730', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19293247, end_position_on_the_genomic_accession=19295736, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377639, CurrentID=0, Status='live', Symbol='LOC105377639', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87411884, end_position_on_the_genomic_accession=87422883, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506733, CurrentID=0, Status='live', Symbol='FAM170B-AS1', Aliases='', description='FAM170B antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49121839, end_position_on_the_genomic_accession=49151547, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506736, CurrentID=0, Status='live', Symbol='SLFN12L', Aliases='', description='schlafen family member 12-like', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35474923, end_position_on_the_genomic_accession=35491208, orientation='minus', exon_count=6, OMIM=614956, no_hdr0=''),
  100506737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506737, CurrentID=0, Status='live', Symbol='LINC01589', Aliases='TCONS_00029353', description='long intergenic non-protein coding RNA 1589', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45589135, end_position_on_the_genomic_accession=45606015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  653631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653631, CurrentID=0, Status='live', Symbol='LOC653631', Aliases='', description='axin interactor, dorsalization associated pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77809768, end_position_on_the_genomic_accession=77811935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506742, CurrentID=0, Status='live', Symbol='CASP12', Aliases='CASP-12P1, CASP12', description='caspase 12 (gene/pseudogene)', other_designations='caspase 12 pseudogene 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104885718, end_position_on_the_genomic_accession=104898670, orientation='minus', exon_count=9, OMIM=608633, no_hdr0=''),
  105377641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377641, CurrentID=0, Status='live', Symbol='LOC105377641', Aliases='', description='uncharacterized LOC105377641', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16724265, end_position_on_the_genomic_accession=16772204, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506746, CurrentID=0, Status='live', Symbol='LOC100506746', Aliases='', description='uncharacterized LOC100506746', other_designations='', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86924894, end_position_on_the_genomic_accession=86934850, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418751, CurrentID=0, Status='live', Symbol='GOLGA2P1', Aliases='', description='golgin A2 pseudogene 1', other_designations='GOLGA2 pseudogene 1', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86657233, end_position_on_the_genomic_accession=86660335, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506749, CurrentID=0, Status='live', Symbol='AGAP1-IT1', Aliases='', description='AGAP1 intronic transcript 1', other_designations='AGAP1 intronic transcript 1 (non-protein coding)', map_location='2q37.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235504628, end_position_on_the_genomic_accession=235507566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371765, CurrentID=0, Status='live', Symbol='LOC105371765', Aliases='', description='uncharacterized LOC105371765', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38841520, end_position_on_the_genomic_accession=38848290, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506753, CurrentID=0, Status='live', Symbol='AZIN1-AS1', Aliases='', description='AZIN1 antisense RNA 1', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102864271, end_position_on_the_genomic_accession=102977876, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506755, CurrentID=0, Status='live', Symbol='MIR497HG', Aliases='MIR195HG', description='mir-497-195 cluster host gene', other_designations='mir-497-195 cluster host gene (non-protein coding)', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7015818, end_position_on_the_genomic_accession=7019654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506757, CurrentID=0, Status='live', Symbol='LINC00629', Aliases='', description='long intergenic non-protein coding RNA 629', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134550024, end_position_on_the_genomic_accession=134560398, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506759, CurrentID=0, Status='live', Symbol='OR7E37P', Aliases='OR7E48P, OST193, hg533', description='olfactory receptor family 7 subfamily E member 37 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 37 pseudogene|olfactory receptor, family 7, subfamily E, member 48 pseudogene', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41442564, end_position_on_the_genomic_accession=41443766, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100506765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506765, CurrentID=0, Status='live', Symbol='IGSF11-AS1', Aliases='', description='IGSF11 antisense RNA 1', other_designations='IGSF11 antisense RNA 1 (non-protein coding)', map_location='3q13.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118943076, end_position_on_the_genomic_accession=118948241, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506766, CurrentID=0, Status='live', Symbol='LOC100506766', Aliases='', description='uncharacterized LOC100506766', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=1054090, end_position_on_the_genomic_accession=1055690, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506769, CurrentID=0, Status='live', Symbol='C10orf71-AS1', Aliases='', description='C10orf71 antisense RNA 1', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49296283, end_position_on_the_genomic_accession=49298893, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506775, CurrentID=0, Status='live', Symbol='LINC01353', Aliases='', description='long intergenic non-protein coding RNA 1353', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203287152, end_position_on_the_genomic_accession=203288798, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506776, CurrentID=0, Status='live', Symbol='TRG-AS1', Aliases='TCRGV', description='T cell receptor gamma locus antisense RNA 1', other_designations='TRG antisense RNA 1', map_location='7p14', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=38341577, end_position_on_the_genomic_accession=38378637, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506779, CurrentID=0, Status='live', Symbol='BZRAP1-AS1', Aliases='', description='BZRAP1 antisense RNA 1', other_designations='BZRAP1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58325450, end_position_on_the_genomic_accession=58353727, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100506783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506783, CurrentID=0, Status='live', Symbol='HOXD-AS2', Aliases='', description='HOXD cluster antisense RNA 2', other_designations='HOXD cluster antisense RNA 2 (non-protein coding)', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176134841, end_position_on_the_genomic_accession=176137098, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506787, CurrentID=0, Status='live', Symbol='LINC01543', Aliases='', description='long intergenic non-protein coding RNA 1543', other_designations='TCONS_00026302', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26422995, end_position_on_the_genomic_accession=26438044, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506790, CurrentID=0, Status='live', Symbol='LOC100506790', Aliases='', description='uncharacterized LOC100506790', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135396429, end_position_on_the_genomic_accession=135397749, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506791, CurrentID=0, Status='live', Symbol='LINC01511', Aliases='', description='long intergenic non-protein coding RNA 1511', other_designations='RP11-325I22.2', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1363582, end_position_on_the_genomic_accession=1380073, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506795, CurrentID=0, Status='live', Symbol='LINC01354', Aliases='', description='long intergenic non-protein coding RNA 1354', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234527891, end_position_on_the_genomic_accession=234531779, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506797, CurrentID=0, Status='live', Symbol='LOC100506797', Aliases='', description='uncharacterized LOC100506797', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118182922, end_position_on_the_genomic_accession=118186448, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506801, CurrentID=0, Status='live', Symbol='LOC100506801', Aliases='', description='uncharacterized LOC100506801', other_designations='', map_location='1p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21293290, end_position_on_the_genomic_accession=21299869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506804, CurrentID=0, Status='live', Symbol='LOC100506804', Aliases='', description='uncharacterized LOC100506804', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75285014, end_position_on_the_genomic_accession=75291864, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376996, CurrentID=0, Status='live', Symbol='LOC105376996', Aliases='', description='uncharacterized LOC105376996', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63117066, end_position_on_the_genomic_accession=63130287, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506810, CurrentID=0, Status='live', Symbol='LINC01132', Aliases='', description='long intergenic non-protein coding RNA 1132', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234724042, end_position_on_the_genomic_accession=234731643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170527, CurrentID=0, Status='live', Symbol='KRT18P3', Aliases='bA359G22.1', description='keratin 18 pseudogene 3', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22732733, end_position_on_the_genomic_accession=22734091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506813, CurrentID=0, Status='live', Symbol='LOC100506813', Aliases='', description='uncharacterized LOC100506813', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=46937392, end_position_on_the_genomic_accession=46944094, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506815, CurrentID=0, Status='live', Symbol='FOXP1-IT1', Aliases='', description='FOXP1 intronic transcript 1', other_designations='FOXP1 intronic transcript 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=71619425, end_position_on_the_genomic_accession=71623608, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506826, CurrentID=0, Status='live', Symbol='MYLK-AS1', Aliases='', description='MYLK antisense RNA 1', other_designations='MYLK antisense RNA 1 (non-protein coding)', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123585513, end_position_on_the_genomic_accession=123630821, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506827, CurrentID=0, Status='live', Symbol='LVCAT1', Aliases='', description='liver cancer-associated transcript 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44016913, end_position_on_the_genomic_accession=44022061, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506834, CurrentID=0, Status='live', Symbol='GAS1RR', Aliases='LncRNA-Hh', description='GAS1 adjacent regulatory RNA', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86948698, end_position_on_the_genomic_accession=87002033, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100506835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506835, CurrentID=0, Status='live', Symbol='LINC00840', Aliases='', description='long intergenic non-protein coding RNA 840', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43859411, end_position_on_the_genomic_accession=43895433, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100506840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506840, CurrentID=0, Status='live', Symbol='IQCF4', Aliases='', description='IQ motif containing F5 pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51817603, end_position_on_the_genomic_accession=51819635, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506844, CurrentID=0, Status='live', Symbol='LOC100506844', Aliases='', description='uncharacterized LOC100506844', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57931449, end_position_on_the_genomic_accession=57936164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370886, CurrentID=0, Status='live', Symbol='LOC105370886', Aliases='', description='uncharacterized LOC105370886', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71091407, end_position_on_the_genomic_accession=71093078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506851, CurrentID=0, Status='live', Symbol='LOC100506851', Aliases='', description='uncharacterized LOC100506851', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79419879, end_position_on_the_genomic_accession=79421659, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418769, CurrentID=0, Status='live', Symbol='KRT18P52', Aliases='', description='keratin 18 pseudogene 52', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31822536, end_position_on_the_genomic_accession=31823436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372446, CurrentID=0, Status='live', Symbol='LOC105372446', Aliases='', description='uncharacterized LOC105372446', other_designations='CTD-2616J11.14', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51394484, end_position_on_the_genomic_accession=51414415, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100506860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506860, CurrentID=0, Status='live', Symbol='LOC100506860', Aliases='', description='uncharacterized LOC100506860', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130913464, end_position_on_the_genomic_accession=130921946, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506866, CurrentID=0, Status='live', Symbol='TTN-AS1', Aliases='', description='TTN antisense RNA 1', other_designations='TTN antisense RNA 1 (non-protein coding)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178522827, end_position_on_the_genomic_accession=178620217, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100506869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506869, CurrentID=0, Status='live', Symbol='LOC100506869', Aliases='', description='uncharacterized LOC100506869', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58591702, end_position_on_the_genomic_accession=58812669, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  170537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170537, CurrentID=0, Status='live', Symbol='PPIAP2', Aliases='bA384D7.2', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 2', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25368067, end_position_on_the_genomic_accession=25368808, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506872, CurrentID=0, Status='live', Symbol='RNF7P1', Aliases='', description='ring finger protein 7 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57519672, end_position_on_the_genomic_accession=57520010, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378873, CurrentID=0, Status='live', Symbol='LOC105378873', Aliases='', description='uncharacterized LOC105378873', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=102387211, end_position_on_the_genomic_accession=102388621, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506874, CurrentID=0, Status='live', Symbol='LINC00933', Aliases='', description='long intergenic non-protein coding RNA 933', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84570649, end_position_on_the_genomic_accession=84580181, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326295, CurrentID=0, Status='live', Symbol='EIF4EBP1P1', Aliases='EIF4EBP1P', description='eukaryotic translation initiation factor 4E binding protein 1 pseudogene 1', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21419600, end_position_on_the_genomic_accession=21420407, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506881, CurrentID=0, Status='live', Symbol='MKLN1-AS', Aliases='MKLN1-AS2', description='MKLN1 antisense RNA', other_designations='MKLN1 antisense RNA 2', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131309744, end_position_on_the_genomic_accession=131327779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  170539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170539, CurrentID=0, Status='live', Symbol='PPIAP3', Aliases='bA425M5.2', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 3', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37530003, end_position_on_the_genomic_accession=37530744, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506885, CurrentID=0, Status='live', Symbol='LOC100506885', Aliases='', description='uncharacterized LOC100506885', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19803706, end_position_on_the_genomic_accession=19804793, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506888, CurrentID=0, Status='live', Symbol='TCEB3CL2', Aliases='', description='transcription elongation factor B polypeptide 3C-like 2', other_designations='eloA3-like-2|elongin-A3-like-2|transcription elongation factor B polypeptide 3C-like-2', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46962768, end_position_on_the_genomic_accession=46964644, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506895, CurrentID=0, Status='live', Symbol='LOC100506895', Aliases='', description='uncharacterized LOC100506895', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43508728, end_position_on_the_genomic_accession=43522542, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506897, CurrentID=0, Status='live', Symbol='DAPK1-IT1', Aliases='', description='DAPK1 intronic transcript 1', other_designations='DAPK1 intronic transcript 1 (non-protein coding)', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=90168369, end_position_on_the_genomic_accession=90169374, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506898, CurrentID=0, Status='live', Symbol='MAGOH2P', Aliases='MAGOH2', description='mago homolog 2, pseudogene', other_designations='mago-nashi homolog 2 (pseudogene)|mago-nashi homolog 2, proliferation-associated|mago-nashi homolog 2, pseudogene|mago-nashi homolog, proliferation-associated pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10715943, end_position_on_the_genomic_accession=10716582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100506901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506901, CurrentID=0, Status='live', Symbol='LL0XNC01-250H12.3', Aliases='', description='uncharacterized LOC100506901', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103404777, end_position_on_the_genomic_accession=103519242, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  388996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388996, CurrentID=0, Status='live', Symbol='LOC388996', Aliases='', description='otopetrin 1 pseudogene', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91589464, end_position_on_the_genomic_accession=91621424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506906, CurrentID=0, Status='live', Symbol='FLNB-AS1', Aliases='', description='FLNB antisense RNA 1', other_designations='FLNB antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=58148274, end_position_on_the_genomic_accession=58156345, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506907, CurrentID=0, Status='live', Symbol='CCDC37-AS1', Aliases='', description='CCDC37 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126393032, end_position_on_the_genomic_accession=126394798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506911, CurrentID=0, Status='live', Symbol='LOC100506911', Aliases='', description='chromosome 6 open reading frame 35 pseudogene', other_designations='transmembrane protein 242 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63002450, end_position_on_the_genomic_accession=63002806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506915, CurrentID=0, Status='live', Symbol='CHRM3-AS2', Aliases='', description='CHRM3 antisense RNA 2', other_designations='CHRM3 antisense RNA 2 (non-protein coding)', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239707126, end_position_on_the_genomic_accession=239719119, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506916, CurrentID=0, Status='live', Symbol='CYP4F9P', Aliases='', description='cytochrome P450 family 4 subfamily F member 9, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 9, pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377670, CurrentID=0, Status='live', Symbol='LOC105377670', Aliases='', description='uncharacterized LOC105377670', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57849857, end_position_on_the_genomic_accession=57854653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506922, CurrentID=0, Status='live', Symbol='LOC100506922', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127378128, end_position_on_the_genomic_accession=127389411, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377671, CurrentID=0, Status='live', Symbol='LOC105377671', Aliases='', description='uncharacterized LOC105377671', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=58524515, end_position_on_the_genomic_accession=58536379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506925, CurrentID=0, Status='live', Symbol='HNRNPA1P31', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 31', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80618736, end_position_on_the_genomic_accession=80620126, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418781, CurrentID=0, Status='live', Symbol='KRT8P39', Aliases='', description='keratin 8 pseudogene 39', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68705556, end_position_on_the_genomic_accession=68707265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506928, CurrentID=0, Status='live', Symbol='LOC100506928', Aliases='', description='uncharacterized LOC100506928', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389000, CurrentID=0, Status='live', Symbol='SLC9B1P2', Aliases='NHEDC1P2', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1 pseudogene 2', other_designations='Na+/H+ exchanger domain containing 1 pseudogene 2|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1 pseudogene 2', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91882906, end_position_on_the_genomic_accession=91931714, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506930, CurrentID=0, Status='live', Symbol='LINC00665', Aliases='', description='long intergenic non-protein coding RNA 665', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36313061, end_position_on_the_genomic_accession=36331718, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100506937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506937, CurrentID=0, Status='live', Symbol='LOC100506937', Aliases='', description='uncharacterized LOC100506937', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132759007, end_position_on_the_genomic_accession=132760632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506938, CurrentID=0, Status='live', Symbol='S100A11P4', Aliases='', description='S100 calcium binding protein A11 pseudogene 4', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63651578, end_position_on_the_genomic_accession=63652108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506939, CurrentID=0, Status='live', Symbol='PRKG1-AS1', Aliases='', description='PRKG1 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52296848, end_position_on_the_genomic_accession=52314128, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100506944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506944, CurrentID=0, Status='live', Symbol='LOC100506944', Aliases='', description='uncharacterized LOC100506944', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88338295, end_position_on_the_genomic_accession=88468050, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100506946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506946, CurrentID=0, Status='live', Symbol='LOC100506946', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=60690016, end_position_on_the_genomic_accession=60690731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377675, CurrentID=0, Status='live', Symbol='LOC105377675', Aliases='', description='uncharacterized LOC105377675', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157652245, end_position_on_the_genomic_accession=157659054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  170550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170550, CurrentID=0, Status='live', Symbol='SNRPFP1', Aliases='dJ1041C10.2', description='small nuclear ribonucleoprotein polypeptide F pseudogene 1', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49729572, end_position_on_the_genomic_accession=49729939, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506953, CurrentID=0, Status='live', Symbol='FABP7P1', Aliases='', description='fatty acid binding protein 7, brain pseudogene 1', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243624382, end_position_on_the_genomic_accession=243625118, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376997, CurrentID=0, Status='live', Symbol='LOC105376997', Aliases='', description='uncharacterized LOC105376997', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25383110, end_position_on_the_genomic_accession=25384942, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100652871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652871, CurrentID=0, Status='live', Symbol='LOC100652871', Aliases='', description='glutathione S-transferase theta-4-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23935477, end_position_on_the_genomic_accession=23942861, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506963, CurrentID=0, Status='live', Symbol='TCEB3-AS1', Aliases='', description='TCEB3 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23760382, end_position_on_the_genomic_accession=23778297, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506965, CurrentID=0, Status='live', Symbol='PWAR6', Aliases='HBT8, PAR-6', description='Prader Willi/Angelman region RNA 6', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=25272883, end_position_on_the_genomic_accession=25281637, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506974, CurrentID=0, Status='live', Symbol='LOC100506974', Aliases='', description='uncharacterized LOC100506974', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13776628, end_position_on_the_genomic_accession=13892841, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377680, CurrentID=0, Status='live', Symbol='LOC105377680', Aliases='', description='uncharacterized LOC105377680', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158317269, end_position_on_the_genomic_accession=158323197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170555, CurrentID=0, Status='live', Symbol='TPM3P2', Aliases='TPM5P, dJ1125A11.2', description='tropomyosin 3 pseudogene 2', other_designations='tropomyosin 5, pseudogene', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33910723, end_position_on_the_genomic_accession=33913423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377681, CurrentID=0, Status='live', Symbol='LOC105377681', Aliases='', description='uncharacterized LOC105377681', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158470884, end_position_on_the_genomic_accession=158515080, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100506985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506985, CurrentID=0, Status='live', Symbol='LOC100506985', Aliases='', description='uncharacterized LOC100506985', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24539274, end_position_on_the_genomic_accession=24556024, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377682, CurrentID=0, Status='live', Symbol='LOC105377682', Aliases='', description='uncharacterized LOC105377682', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159227715, end_position_on_the_genomic_accession=159245157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506990, CurrentID=0, Status='live', Symbol='LOC100506990', Aliases='', description='uncharacterized LOC100506990', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12437013, end_position_on_the_genomic_accession=12566845, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100506994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506994, CurrentID=0, Status='live', Symbol='PTPRG-AS1', Aliases='', description='PTPRG antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62261819, end_position_on_the_genomic_accession=62318947, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377683, CurrentID=0, Status='live', Symbol='LOC105377683', Aliases='', description='uncharacterized LOC105377683', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159310859, end_position_on_the_genomic_accession=159315252, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506999, CurrentID=0, Status='live', Symbol='LOC100506999', Aliases='', description='uncharacterized LOC100506999', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95573554, end_position_on_the_genomic_accession=95582449, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100528020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528020, CurrentID=0, Status='live', Symbol='FAM187A', Aliases='', description='family with sequence similarity 187 member A', other_designations='family with sequence similarity 187, member A', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100507002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507002, CurrentID=0, Status='live', Symbol='LOC100507002', Aliases='', description='uncharacterized LOC100507002', other_designations='', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65100812, end_position_on_the_genomic_accession=65111095, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507003, CurrentID=0, Status='live', Symbol='GFY', Aliases='Goofy', description='golgi associated olfactory signaling regulator', other_designations='Golgi protein in olfactory neurons|golgi-associated olfactory signaling regulator|golgi-associated, olfactory signaling regulator|protein Goofy', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49421318, end_position_on_the_genomic_accession=49428977, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507006, CurrentID=0, Status='live', Symbol='LOC100507006', Aliases='', description='uncharacterized LOC100507006', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64228401, end_position_on_the_genomic_accession=64252531, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377685, CurrentID=0, Status='live', Symbol='LOC105377685', Aliases='', description='uncharacterized LOC105377685', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159512508, end_position_on_the_genomic_accession=159520471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507008, CurrentID=0, Status='live', Symbol='LINC00844', Aliases='', description='long intergenic non-protein coding RNA 844', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58999518, end_position_on_the_genomic_accession=59001617, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507012, CurrentID=0, Status='live', Symbol='BMPR1B-AS1', Aliases='', description='BMPR1B antisense RNA 1 (head to head)', other_designations='TCONS_00007797', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94743668, end_position_on_the_genomic_accession=94757533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377686, CurrentID=0, Status='live', Symbol='LOC105377686', Aliases='', description='uncharacterized LOC105377686', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159504782, end_position_on_the_genomic_accession=159511892, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100528023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528023, CurrentID=0, Status='live', Symbol='PPR1', Aliases='PPR', description='Photoparoxysmal response 1', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=132100, no_hdr0=''),
  100507022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507022, CurrentID=0, Status='live', Symbol='LOC100507022', Aliases='', description='FLJ00291 protein', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=5077612, end_position_on_the_genomic_accession=5078357, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377688, CurrentID=0, Status='live', Symbol='LOC105377688', Aliases='', description='uncharacterized LOC105377688', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159708085, end_position_on_the_genomic_accession=159761045, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507027, CurrentID=0, Status='live', Symbol='MRLN', Aliases='LINC00948, M1, MLN, MUSER1', description='myoregulin', other_designations='long intergenic non-protein coding RNA 948|muscle enriched RNA 1|muscle specific 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59736990, end_position_on_the_genomic_accession=59753471, orientation='minus', exon_count=4, OMIM=616246, no_hdr0=''),
  105377689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377689, CurrentID=0, Status='live', Symbol='LOC105377689', Aliases='', description='uncharacterized LOC105377689', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159865080, end_position_on_the_genomic_accession=159906757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507032, CurrentID=0, Status='live', Symbol='TMCC1-AS1', Aliases='', description='TMCC1 antisense RNA 1 (head to head)', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129893871, end_position_on_the_genomic_accession=129908912, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507034, CurrentID=0, Status='live', Symbol='PITRM1-AS1', Aliases='', description='PITRM1 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3141601, end_position_on_the_genomic_accession=3148629, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377690, CurrentID=0, Status='live', Symbol='LOC105377690', Aliases='', description='uncharacterized LOC105377690', other_designations='CTB-127C13.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160195696, end_position_on_the_genomic_accession=160210577, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507040, CurrentID=0, Status='live', Symbol='LINC00427', Aliases='', description='long intergenic non-protein coding RNA 427', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=30890497, end_position_on_the_genomic_accession=30894040, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507043, CurrentID=0, Status='live', Symbol='TUNAR', Aliases='HI-LNC78, LINC00617, TUNA', description='TCL1 upstream neural differentiation-associated RNA', other_designations='Human Islet Long Non Coding RNA 78|Tcl1 upstream neuron-associated lincRNA|long intergenic non-protein coding RNA 617', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95876772, end_position_on_the_genomic_accession=95925571, orientation='plus', exon_count=2, OMIM=615719, no_hdr0=''),
  100507044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507044, CurrentID=0, Status='live', Symbol='NBPF5P', Aliases='DKFZp434D177-like, NBPF5', description='neuroblastoma breakpoint family member 5, pseudogene', other_designations='neuroblastoma breakpoint family, member 5, pseudogene', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108375971, end_position_on_the_genomic_accession=108383614, orientation='plus', exon_count=0, OMIM=613995, no_hdr0=''),
  100507046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507046, CurrentID=0, Status='live', Symbol='HSPE1P3', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 3', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90634725, end_position_on_the_genomic_accession=90635030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507050, CurrentID=0, Status='live', Symbol='TPBGL', Aliases='', description='trophoblast glycoprotein-like', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75240905, end_position_on_the_genomic_accession=75243704, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507053, CurrentID=0, Status='live', Symbol='LOC100507053', Aliases='', description='uncharacterized LOC100507053', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99088857, end_position_on_the_genomic_accession=99301356, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100507055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507055, CurrentID=0, Status='live', Symbol='LRCOL1', Aliases='', description='leucine-rich colipase like 1', other_designations='leucine rich colipase-like 1', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132602224, end_position_on_the_genomic_accession=132610451, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100507056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507056, CurrentID=0, Status='live', Symbol='CCAT1', Aliases='CARLo-5, onco-lncRNA-40', description='colon cancer associated transcript 1 (non-protein coding)', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127207382, end_position_on_the_genomic_accession=127219268, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507057, CurrentID=0, Status='live', Symbol='MFI2-AS1', Aliases='', description='MFI2 antisense RNA 1', other_designations='MFI2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197002906, end_position_on_the_genomic_accession=197004744, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507058, CurrentID=0, Status='live', Symbol='LINC00845', Aliases='', description='long intergenic non-protein coding RNA 845', other_designations='', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61016275, end_position_on_the_genomic_accession=61026417, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507059, CurrentID=0, Status='live', Symbol='LINC00703', Aliases='', description='long intergenic non-protein coding RNA 703', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4384246, end_position_on_the_genomic_accession=4410612, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377694, CurrentID=0, Status='live', Symbol='LOC105377694', Aliases='', description='uncharacterized LOC105377694', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161369998, end_position_on_the_genomic_accession=161407224, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507062, CurrentID=0, Status='live', Symbol='PSMD6-AS2', Aliases='', description='PSMD6 antisense RNA 2', other_designations='PSMD6 antisense RNA 2 (non-protein coding)', map_location='3p14', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64004022, end_position_on_the_genomic_accession=64012243, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507064, CurrentID=0, Status='live', Symbol='TEX26-AS1', Aliases='LINC00447', description='TEX26 antisense RNA 1', other_designations='TEX26 antisense RNA 1 (non-protein coding)|long intergenic non-protein coding RNA 447', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30882835, end_position_on_the_genomic_accession=30932608, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507065, CurrentID=0, Status='live', Symbol='LOC100507065', Aliases='', description='uncharacterized LOC100507065', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65466817, end_position_on_the_genomic_accession=65642372, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100507066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507066, CurrentID=0, Status='live', Symbol='CLIP1-AS1', Aliases='', description='CLIP1 antisense RNA 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122395542, end_position_on_the_genomic_accession=122399944, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100528031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528031, CurrentID=0, Status='live', Symbol='FKBP1A-SDCBP2', Aliases='', description='FKBP1A-SDCBP2 readthrough (NMD candidate)', other_designations='FKBP1A-SDCBP2 readthrough (non-protein coding)', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1309909, end_position_on_the_genomic_accession=1393172, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100507071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507071, CurrentID=0, Status='live', Symbol='LOC100507071', Aliases='', description='uncharacterized LOC100507071', other_designations='CTD-2530N21.4', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22254576, end_position_on_the_genomic_accession=22275233, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507073, CurrentID=0, Status='live', Symbol='LOC100507073', Aliases='', description='uncharacterized LOC100507073', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66574030, end_position_on_the_genomic_accession=66695588, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507079, CurrentID=0, Status='live', Symbol='LOC100507079', Aliases='', description='uncharacterized LOC100507079', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=91116791, end_position_on_the_genomic_accession=91118583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507083, CurrentID=0, Status='live', Symbol='LOC100507083', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9) pseudogene', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9) pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130293319, end_position_on_the_genomic_accession=130293759, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507086, CurrentID=0, Status='live', Symbol='DLG1-AS1', Aliases='', description='DLG1 antisense RNA 1', other_designations='DLG1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197298247, end_position_on_the_genomic_accession=197303751, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507091, CurrentID=0, Status='live', Symbol='LOC100507091', Aliases='', description='uncharacterized LOC100507091', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123081384, end_position_on_the_genomic_accession=123084744, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507096, CurrentID=0, Status='live', Symbol='GIMD1', Aliases='', description='GIMAP family P-loop NTPase domain containing 1', other_designations='GIMAP family P-loop NTPase domain-containing protein 1', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106357335, end_position_on_the_genomic_accession=106368788, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507097, CurrentID=0, Status='live', Symbol='ATP5G2P3', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9) pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9) pseudogene 3', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197263057, end_position_on_the_genomic_accession=197263679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507098, CurrentID=0, Status='live', Symbol='ADAMTS9-AS2', Aliases='', description='ADAMTS9 antisense RNA 2', other_designations='ADAMTS9 antisense RNA 2 (non-protein coding)|CTD-2230D16.1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64684870, end_position_on_the_genomic_accession=65011468, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507099, CurrentID=0, Status='live', Symbol='FRY-AS1', Aliases='', description='FRY antisense RNA 1', other_designations='FRY antisense RNA 1 (non-protein coding)', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32024046, end_position_on_the_genomic_accession=32031639, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507102, CurrentID=0, Status='live', Symbol='COX20P1', Aliases='FAM36B, bA119F7.2', description='COX20 cytochrome c oxidase assembly factor pseudogene 1', other_designations='COX20 Cox2 chaperone homolog pseudogene 1|COX20 pseudogene 1|family with sequence similarity 36, member A pseudogene|family with sequence similarity 36, member B (pseudogene)', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68632337, end_position_on_the_genomic_accession=68632836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507103, CurrentID=0, Status='live', Symbol='LOC100507103', Aliases='', description='uncharacterized LOC100507103', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91104957, end_position_on_the_genomic_accession=91165723, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100418811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418811, CurrentID=0, Status='live', Symbol='KRT8P34', Aliases='', description='keratin 8 pseudogene 34', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39834724, end_position_on_the_genomic_accession=39836493, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100134822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134822, CurrentID=0, Status='live', Symbol='LOC100134822', Aliases='', description='uncharacterized LOC100134822', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=497259, end_position_on_the_genomic_accession=532779, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100507113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507113, CurrentID=0, Status='live', Symbol='ZNF638-IT1', Aliases='', description='ZNF638 intronic transcript 1', other_designations='ZNF638 intronic transcript 1 (non-protein coding)', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71373941, end_position_on_the_genomic_accession=71376320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507114, CurrentID=0, Status='live', Symbol='LINC00445', Aliases='', description='long intergenic non-protein coding RNA 445', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35697538, end_position_on_the_genomic_accession=35699001, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507117, CurrentID=0, Status='live', Symbol='ASAP1-IT2', Aliases='', description='ASAP1 intronic transcript 2', other_designations='ASAP1 intronic transcript 2 (non-protein coding)', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130082738, end_position_on_the_genomic_accession=130084768, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507118, CurrentID=0, Status='live', Symbol='PRC1-AS1', Aliases='', description='PRC1 antisense RNA 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90966369, end_position_on_the_genomic_accession=90988624, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507124, CurrentID=0, Status='live', Symbol='HMGN1P25', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 25', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 25', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100480881, end_position_on_the_genomic_accession=100482078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507125, CurrentID=0, Status='live', Symbol='ERHP1', Aliases='', description='enhancer of rudimentary homolog (Drosophila) pseudogene 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139534194, end_position_on_the_genomic_accession=139534961, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376142, CurrentID=0, Status='live', Symbol='LOC105376142', Aliases='', description='uncharacterized LOC105376142', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90158810, end_position_on_the_genomic_accession=90161228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507127, CurrentID=0, Status='live', Symbol='LINC00707', Aliases='', description='long intergenic non-protein coding RNA 707', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6779598, end_position_on_the_genomic_accession=6842906, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507131, CurrentID=0, Status='live', Symbol='LOC100507131', Aliases='', description='uncharacterized LOC100507131', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17858227, end_position_on_the_genomic_accession=17860803, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507135, CurrentID=0, Status='live', Symbol='SPG20-AS1', Aliases='C13orf43, SPG20OS', description='SPG20 antisense RNA 1', other_designations='SPG20 opposite strand', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36346431, end_position_on_the_genomic_accession=36369735, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507137, CurrentID=0, Status='live', Symbol='FP2234', Aliases='', description='uncharacterized LOC100507137', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=93968457, end_position_on_the_genomic_accession=93972589, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507140, CurrentID=0, Status='live', Symbol='LOC100507140', Aliases='', description='uncharacterized LOC100507140', other_designations='', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200712305, end_position_on_the_genomic_accession=200735177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507143, CurrentID=0, Status='live', Symbol='LINC00708', Aliases='GS1-756B1.2', description='long intergenic non-protein coding RNA 708', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8259331, end_position_on_the_genomic_accession=8268305, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507144, CurrentID=0, Status='live', Symbol='LOC100507144', Aliases='', description='uncharacterized LOC100507144', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35132650, end_position_on_the_genomic_accession=35138032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377708, CurrentID=0, Status='live', Symbol='LOC105377708', Aliases='', description='uncharacterized LOC105377708', other_designations='CTB-178M22.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168181377, end_position_on_the_genomic_accession=168186178, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507156, CurrentID=0, Status='live', Symbol='LOC100507156', Aliases='', description='uncharacterized LOC100507156', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23336208, end_position_on_the_genomic_accession=23366125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507157, CurrentID=0, Status='live', Symbol='MTND1P30', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 30', other_designations='MT-ND1 pseudogene 30', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55180371, end_position_on_the_genomic_accession=55181307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507163, CurrentID=0, Status='live', Symbol='LINC00709', Aliases='', description='long intergenic non-protein coding RNA 709', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9275613, end_position_on_the_genomic_accession=9287057, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507170, CurrentID=0, Status='live', Symbol='CT47A12', Aliases='CT47, CT47.12', description='cancer/testis antigen family 47, member A12', other_designations='cancer/testis CT47 family, member 12|cancer/testis antigen 47|cancer/testis antigen 47A family member', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120877490, end_position_on_the_genomic_accession=120932297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507171, CurrentID=0, Status='live', Symbol='BOLA3-AS1', Aliases='', description='BOLA3 antisense RNA 1 (head to head)', other_designations='BOLA3 antisense RNA 1 (non-protein coding)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74147981, end_position_on_the_genomic_accession=74152389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371779, CurrentID=0, Status='live', Symbol='LOC105371779', Aliases='', description='uncharacterized LOC105371779', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41893181, end_position_on_the_genomic_accession=41897742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507173, CurrentID=0, Status='live', Symbol='LINC01012', Aliases='', description='long intergenic non-protein coding RNA 1012', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27694035, end_position_on_the_genomic_accession=27710222, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507175, CurrentID=0, Status='live', Symbol='LOC100507175', Aliases='', description='uncharacterized LOC100507175', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67520082, end_position_on_the_genomic_accession=67567131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507178, CurrentID=0, Status='live', Symbol='SLFNL1-AS1', Aliases='', description='SLFNL1 antisense RNA 1', other_designations='SLFNL1 antisense RNA 1 (non-protein coding)', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41014590, end_position_on_the_genomic_accession=41043890, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377715, CurrentID=0, Status='live', Symbol='LOC105377715', Aliases='', description='uncharacterized LOC105377715', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169460336, end_position_on_the_genomic_accession=169561759, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507191, CurrentID=0, Status='live', Symbol='LOC100507191', Aliases='', description='uncharacterized LOC100507191', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75821143, end_position_on_the_genomic_accession=75821990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927608, CurrentID=0, Status='live', Symbol='LOC101927608', Aliases='', description='uncharacterized LOC101927608', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57893268, end_position_on_the_genomic_accession=57896846, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507194, CurrentID=0, Status='live', Symbol='LOC100507194', Aliases='', description='uncharacterized LOC100507194', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3182817, end_position_on_the_genomic_accession=3195767, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507195, CurrentID=0, Status='live', Symbol='LOC100507195', Aliases='', description='uncharacterized LOC100507195', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68431842, end_position_on_the_genomic_accession=68451484, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507201, CurrentID=0, Status='live', Symbol='LOC100507201', Aliases='', description='uncharacterized LOC100507201', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81461359, end_position_on_the_genomic_accession=81466946, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422414, CurrentID=0, Status='live', Symbol='LOC100422414', Aliases='', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 7 pseudogene', other_designations='proteasome 26S subunit, non-ATPase 7 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185280844, end_position_on_the_genomic_accession=185281817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507203, CurrentID=0, Status='live', Symbol='SMLR1', Aliases='', description='small leucine-rich protein 1', other_designations='', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130827405, end_position_on_the_genomic_accession=130837135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507205, CurrentID=0, Status='live', Symbol='LOC100507205', Aliases='', description='uncharacterized LOC100507205', other_designations='CTD-2572N17.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=42187743, end_position_on_the_genomic_accession=42253690, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507206, CurrentID=0, Status='live', Symbol='LINC00943', Aliases='', description='long intergenic non-protein coding RNA 943', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126737007, end_position_on_the_genomic_accession=126746256, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507209, CurrentID=0, Status='live', Symbol='LOC100507209', Aliases='', description='uncharacterized LOC100507209', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=114647555, end_position_on_the_genomic_accession=114649312, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507210, CurrentID=0, Status='live', Symbol='LOC100507210', Aliases='', description='uncharacterized LOC100507210', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134055253, end_position_on_the_genomic_accession=134057734, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287757, CurrentID=0, Status='live', Symbol='LOC100287757', Aliases='', description='SRA stem-loop interacting RNA binding protein pseudogene', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147800972, end_position_on_the_genomic_accession=147801353, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507217, CurrentID=0, Status='live', Symbol='LINC01578', Aliases='', description='long intergenic non-protein coding RNA 1578', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92882843, end_position_on_the_genomic_accession=92898747, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100507218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507218, CurrentID=0, Status='live', Symbol='RNF157-AS1', Aliases='', description='RNF157 antisense RNA 1', other_designations='RNF157 antisense RNA 1 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76140556, end_position_on_the_genomic_accession=76154648, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100507221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507221, CurrentID=0, Status='live', Symbol='LOC100507221', Aliases='', description='uncharacterized LOC100507221', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32262898, end_position_on_the_genomic_accession=32281946, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100507231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507231, CurrentID=0, Status='live', Symbol='LOC100507231', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57717849, end_position_on_the_genomic_accession=57718566, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507236, CurrentID=0, Status='live', Symbol='LOC100507236', Aliases='', description='uncharacterized LOC100507236', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25941229, end_position_on_the_genomic_accession=25943863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507240, CurrentID=0, Status='live', Symbol='VWA8-AS1', Aliases='', description='VWA8 antisense RNA 1 (head to head)', other_designations='VWA8 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41961169, end_position_on_the_genomic_accession=41981565, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507242, CurrentID=0, Status='live', Symbol='SNHG23', Aliases='', description='small nucleolar RNA host gene 23', other_designations='small nucleolar RNA host gene 23 (non-protein coding)', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100910281, end_position_on_the_genomic_accession=100959190, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100507244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507244, CurrentID=0, Status='live', Symbol='MAMDC2-AS1', Aliases='', description='MAMDC2 antisense RNA 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70153134, end_position_on_the_genomic_accession=70175888, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100418834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418834, CurrentID=0, Status='live', Symbol='LOC100418834', Aliases='', description='metadherin pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21709536, end_position_on_the_genomic_accession=21710166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507246, CurrentID=0, Status='live', Symbol='SNHG16', Aliases='ncRAN', description='small nucleolar RNA host gene 16', other_designations='non-coding RNA expressed in aggressive neuroblastoma|small nucleolar RNA host gene 16 (non-protein coding)', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76557764, end_position_on_the_genomic_accession=76565348, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377725, CurrentID=0, Status='live', Symbol='LOC105377725', Aliases='', description='uncharacterized LOC105377725', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171794052, end_position_on_the_genomic_accession=171848089, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100507249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507249, CurrentID=0, Status='live', Symbol='C8orf17', Aliases='MOST-1', description='chromosome 8 open reading frame 17', other_designations='MOLT-4 sequence tag-1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=140943416, end_position_on_the_genomic_accession=140946188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507250, CurrentID=0, Status='live', Symbol='LOC100507250', Aliases='', description='uncharacterized LOC100507250', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68674371, end_position_on_the_genomic_accession=68686859, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507251, CurrentID=0, Status='live', Symbol='CICP6', Aliases='', description='capicua transcriptional repressor pseudogene 6', other_designations='capicua homolog pseudogene 6', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198222576, end_position_on_the_genomic_accession=198226350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377726, CurrentID=0, Status='live', Symbol='LOC105377726', Aliases='', description='uncharacterized LOC105377726', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171792593, end_position_on_the_genomic_accession=171796883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507254, CurrentID=0, Status='live', Symbol='LINC01013', Aliases='', description='long intergenic non-protein coding RNA 1013', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132133978, end_position_on_the_genomic_accession=132169374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376999, CurrentID=0, Status='live', Symbol='LOC105376999', Aliases='', description='uncharacterized LOC105376999', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25783639, end_position_on_the_genomic_accession=25785159, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507257, CurrentID=0, Status='live', Symbol='MEG9', Aliases='LINC00584', description='maternally expressed 9 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 584', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101069911, end_position_on_the_genomic_accession=101072937, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507258, CurrentID=0, Status='live', Symbol='LOC100507258', Aliases='', description='uncharacterized LOC100507258', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85220509, end_position_on_the_genomic_accession=85240654, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100528063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528063, CurrentID=0, Status='live', Symbol='LOC100528063', Aliases='', description='gamma-glutamylcyclotransferase pseudogene', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52473471, end_position_on_the_genomic_accession=52474352, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507261, CurrentID=0, Status='live', Symbol='MIR670HG', Aliases='', description='MIR670 host gene', other_designations='MIR670 host gene (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43568401, end_position_on_the_genomic_accession=43572643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287765, CurrentID=0, Status='live', Symbol='LINC00630', Aliases='LL0XNC01-157D4.1, LL0XNC01-237H1.2', description='long intergenic non-protein coding RNA 630', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102769167, end_position_on_the_genomic_accession=102885410, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377728, CurrentID=0, Status='live', Symbol='CTB-54I1.1', Aliases='', description='uncharacterized LOC105377728', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172556725, end_position_on_the_genomic_accession=172559153, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507266, CurrentID=0, Status='live', Symbol='STX18-AS1', Aliases='', description='STX18 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4542131, end_position_on_the_genomic_accession=4710938, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507267, CurrentID=0, Status='live', Symbol='LINC01187', Aliases='CTB-27N1.1, KIDR', description='long intergenic non-protein coding RNA 1187', other_designations='kidney-specific RNA', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170191647, end_position_on_the_genomic_accession=170199141, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507274, CurrentID=0, Status='live', Symbol='LINC01210', Aliases='', description='long intergenic non-protein coding RNA 1210', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137771911, end_position_on_the_genomic_accession=137780878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377730, CurrentID=0, Status='live', Symbol='LOC105377730', Aliases='', description='uncharacterized LOC105377730', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172771409, end_position_on_the_genomic_accession=172783198, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507281, CurrentID=0, Status='live', Symbol='LOC100507281', Aliases='', description='uncharacterized LOC100507281', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=14607109, end_position_on_the_genomic_accession=14608832, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507283, CurrentID=0, Status='live', Symbol='LOC100507283', Aliases='', description='uncharacterized LOC100507283', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124800428, end_position_on_the_genomic_accession=124834487, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507290, CurrentID=0, Status='live', Symbol='ZNF865', Aliases='', description='zinc finger protein 865', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55613593, end_position_on_the_genomic_accession=55618545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507291, CurrentID=0, Status='live', Symbol='LOC100507291', Aliases='', description='uncharacterized LOC100507291', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139389803, end_position_on_the_genomic_accession=139583319, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100422417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422417, CurrentID=0, Status='live', Symbol='LOC100422417', Aliases='', description='solute carrier family 6 (amino acid transporter), member 14 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126311408, end_position_on_the_genomic_accession=126311953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377733, CurrentID=0, Status='live', Symbol='LOC105377733', Aliases='', description='uncharacterized LOC105377733', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173543294, end_position_on_the_genomic_accession=173555535, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507297, CurrentID=0, Status='live', Symbol='TMEM44-AS1', Aliases='', description='TMEM44 antisense RNA 1', other_designations='TMEM44 antisense RNA 1 (non-protein coding)', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194584268, end_position_on_the_genomic_accession=194590832, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507299, CurrentID=0, Status='live', Symbol='SMC5-AS1', Aliases='MAMDC2-AS2', description='SMC5 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70216059, end_position_on_the_genomic_accession=70258874, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507300, CurrentID=0, Status='live', Symbol='ALKBH3-AS1', Aliases='', description='ALKBH3 antisense RNA 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43909289, end_position_on_the_genomic_accession=43920944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507303, CurrentID=0, Status='live', Symbol='SNHG19', Aliases='SNORD60HG', description='small nucleolar RNA host gene 19', other_designations='small nucleolar RNA host gene 19 (non-protein coding)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2154802, end_position_on_the_genomic_accession=2155375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376148, CurrentID=0, Status='live', Symbol='LOC105376148', Aliases='', description='uncharacterized LOC105376148', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91966546, end_position_on_the_genomic_accession=91986056, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377735, CurrentID=0, Status='live', Symbol='LOC105377735', Aliases='', description='uncharacterized LOC105377735', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173581542, end_position_on_the_genomic_accession=173585551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507308, CurrentID=0, Status='live', Symbol='TARID', Aliases='EYA4-AS1', description='TCF21 antisense RNA inducing promoter demethylation', other_designations='EYA4 antisense RNA 1', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=133502252, end_position_on_the_genomic_accession=133889006, orientation='minus', exon_count=9, OMIM=616058, no_hdr0=''),
  100507311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507311, CurrentID=0, Status='live', Symbol='LOC100507311', Aliases='', description='uncharacterized LOC100507311', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=96696644, end_position_on_the_genomic_accession=96702853, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507316, CurrentID=0, Status='live', Symbol='MINCR', Aliases='LINC01604', description='MYC-induced long noncoding RNA', other_designations='TCONS_00015189|long intergenic non-protein coding RNA 1604', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143280155, end_position_on_the_genomic_accession=143281700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507321, CurrentID=0, Status='live', Symbol='ERVK13-1', Aliases='', description='endogenous retrovirus group K13 member 1', other_designations='HERV-K_16p3.3 provirus ancestral Env polyprotein|endogenous retrovirus group K13, member 1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2658389, end_position_on_the_genomic_accession=2673439, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372635, CurrentID=0, Status='live', Symbol='LOC105372635', Aliases='', description='uncharacterized LOC105372635', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47651283, end_position_on_the_genomic_accession=47652327, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  440082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440082, CurrentID=0, Status='live', Symbol='LOC440082', Aliases='', description='uncharacterized LOC440082', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=11005926, end_position_on_the_genomic_accession=11009863, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507330, CurrentID=0, Status='live', Symbol='LOC100507330', Aliases='', description='uncharacterized LOC100507330', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=70921460, end_position_on_the_genomic_accession=70932859, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507331, CurrentID=0, Status='live', Symbol='ZSWIM8-AS1', Aliases='', description='ZSWIM8 antisense RNA 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73796514, end_position_on_the_genomic_accession=73801399, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507334, CurrentID=0, Status='live', Symbol='LOC100507334', Aliases='', description='two pore channel 3 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110245638, end_position_on_the_genomic_accession=110266559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507336, CurrentID=0, Status='live', Symbol='LOC100507336', Aliases='', description='uncharacterized LOC100507336', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3588984, end_position_on_the_genomic_accession=3720074, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100507341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507341, CurrentID=0, Status='live', Symbol='SMIM18', Aliases='', description='small integral membrane protein 18', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30638600, end_position_on_the_genomic_accession=30645956, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507346, CurrentID=0, Status='live', Symbol='LOC100507346', Aliases='', description='uncharacterized LOC100507346', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95463609, end_position_on_the_genomic_accession=95470019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507347, CurrentID=0, Status='live', Symbol='VIM-AS1', Aliases='', description='VIM antisense RNA 1', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17214239, end_position_on_the_genomic_accession=17229985, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129996, CurrentID=0, Status='live', Symbol='LOC100129996', Aliases='', description='chromosome 6 open reading frame 203 pseudogene', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154549488, end_position_on_the_genomic_accession=154550249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507351, CurrentID=0, Status='live', Symbol='LOC100507351', Aliases='', description='uncharacterized LOC100507351', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77546941, end_position_on_the_genomic_accession=77565024, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371785, CurrentID=0, Status='live', Symbol='LOC105371785', Aliases='', description='uncharacterized LOC105371785', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43368644, end_position_on_the_genomic_accession=43373809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507354, CurrentID=0, Status='live', Symbol='CDC42P5', Aliases='', description='cell division cycle 42 pseudogene 5', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173083861, end_position_on_the_genomic_accession=173086288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507357, CurrentID=0, Status='live', Symbol='RPL23AP86', Aliases='', description='ribosomal protein L23a pseudogene 86', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2922280, end_position_on_the_genomic_accession=2923230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507361, CurrentID=0, Status='live', Symbol='PCNPP5', Aliases='', description='PEST containing nuclear protein pseudogene 5', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48326375, end_position_on_the_genomic_accession=48328574, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507362, CurrentID=0, Status='live', Symbol='LINC01015', Aliases='', description='long intergenic non-protein coding RNA 1015', other_designations='DAQB-12N14.5', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29529406, end_position_on_the_genomic_accession=29533568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377745, CurrentID=0, Status='live', Symbol='LOC105377745', Aliases='', description='uncharacterized LOC105377745', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175972270, end_position_on_the_genomic_accession=175979402, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646652, CurrentID=0, Status='live', Symbol='LOC646652', Aliases='', description='integral membrane glycoprotein-like', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34189521, end_position_on_the_genomic_accession=34193736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507373, CurrentID=0, Status='live', Symbol='LOC100507373', Aliases='', description='uncharacterized LOC100507373', other_designations='CTB-55O6.12', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14137152, end_position_on_the_genomic_accession=14171264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507377, CurrentID=0, Status='live', Symbol='LOC100507377', Aliases='', description='uncharacterized LOC100507377', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74133173, end_position_on_the_genomic_accession=74292631, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377747, CurrentID=0, Status='live', Symbol='LOC105377747', Aliases='', description='uncharacterized LOC105377747', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176040326, end_position_on_the_genomic_accession=176042971, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507381, CurrentID=0, Status='live', Symbol='KCNQ5-IT1', Aliases='', description='KCNQ5 intronic transcript 1', other_designations='KCNQ5 intronic transcript 1 (non-protein coding)', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72630495, end_position_on_the_genomic_accession=72678558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507384, CurrentID=0, Status='live', Symbol='LOC100507384', Aliases='', description='uncharacterized LOC100507384', other_designations='CTD-2210P24.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45722308, end_position_on_the_genomic_accession=45724555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377748, CurrentID=0, Status='live', Symbol='LOC105377748', Aliases='', description='uncharacterized LOC105377748', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176135102, end_position_on_the_genomic_accession=176171585, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507387, CurrentID=0, Status='live', Symbol='LOC100507387', Aliases='', description='uncharacterized LOC100507387', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176119549, end_position_on_the_genomic_accession=176125165, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100507388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507388, CurrentID=0, Status='live', Symbol='LOC100507388', Aliases='', description='uncharacterized LOC100507388', other_designations='', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74955974, end_position_on_the_genomic_accession=74970362, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507389, CurrentID=0, Status='live', Symbol='LOC100507389', Aliases='', description='uncharacterized LOC100507389', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142926675, end_position_on_the_genomic_accession=142942536, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507391, CurrentID=0, Status='live', Symbol='LOC100507391', Aliases='', description='uncharacterized LOC100507391', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194708421, end_position_on_the_genomic_accession=194782168, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507392, CurrentID=0, Status='live', Symbol='SENCR', Aliases='FLI1-AS1, lncRNA9', description='smooth muscle and endothelial cell enriched migration/differentiation-associated long non-coding RNA', other_designations='FLI1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128691672, end_position_on_the_genomic_accession=128696023, orientation='minus', exon_count=3, OMIM=615815, no_hdr0=''),
  100507393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507393, CurrentID=0, Status='live', Symbol='LOC100507393', Aliases='', description='putative uncharacterized protein FLJ44672-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56821258, end_position_on_the_genomic_accession=56822459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507395, CurrentID=0, Status='live', Symbol='LOC100507395', Aliases='', description='uncharacterized LOC100507395', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928180, CurrentID=0, Status='live', Symbol='LOC101928180', Aliases='', description='uncharacterized LOC101928180', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216533704, end_position_on_the_genomic_accession=216579185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507398, CurrentID=0, Status='live', Symbol='INTS6-AS1', Aliases='', description='INTS6 antisense RNA 1', other_designations='INTS6 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51453345, end_position_on_the_genomic_accession=51454716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507399, CurrentID=0, Status='live', Symbol='HCG8', Aliases='HCGVIII, HCGVIII-1', description='HLA complex group 8', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30012101, end_position_on_the_genomic_accession=30013922, orientation='minus', exon_count=1, OMIM=615801, no_hdr0=''),
  100507401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507401, CurrentID=0, Status='live', Symbol='LRP4-AS1', Aliases='', description='LRP4 antisense RNA 1', other_designations='LRP4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46846411, end_position_on_the_genomic_accession=46874416, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507403, CurrentID=0, Status='live', Symbol='LOC100507403', Aliases='', description='uncharacterized LOC100507403', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37326957, end_position_on_the_genomic_accession=37332008, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507404, CurrentID=0, Status='live', Symbol='TMLHE-AS1', Aliases='', description='TMLHE antisense RNA 1', other_designations='TMLHE antisense RNA 1 (non-protein coding)', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155466540, end_position_on_the_genomic_accession=155494110, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507406, CurrentID=0, Status='live', Symbol='LOC100507406', Aliases='', description='uncharacterized LOC100507406', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137730170, end_position_on_the_genomic_accession=137739037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507410, CurrentID=0, Status='live', Symbol='C1QTNF1-AS1', Aliases='', description='C1QTNF1 antisense RNA 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79019209, end_position_on_the_genomic_accession=79027655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507412, CurrentID=0, Status='live', Symbol='LOC100507412', Aliases='', description='uncharacterized LOC100507412', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129135, CurrentID=0, Status='live', Symbol='LOC100129135', Aliases='', description='selenocysteine insertion sequence-binding protein 2-like', other_designations='', map_location='18q22.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67788272, end_position_on_the_genomic_accession=67791033, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507415, CurrentID=0, Status='live', Symbol='FAM220CP', Aliases='', description='family with sequence similarity 220 member C, pseudogene', other_designations='family with sequence similarity 220, member C, pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97120401, end_position_on_the_genomic_accession=97123015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507419, CurrentID=0, Status='live', Symbol='MMP25-AS1', Aliases='', description='MMP25 antisense RNA 1', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3052125, end_position_on_the_genomic_accession=3059370, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507420, CurrentID=0, Status='live', Symbol='LINC01605', Aliases='', description='long intergenic non-protein coding RNA 1605', other_designations='TCONS_00014973', map_location='8p11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37406464, end_position_on_the_genomic_accession=37521386, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100507421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507421, CurrentID=0, Status='live', Symbol='TMEM178B', Aliases='', description='transmembrane protein 178B', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141073311, end_position_on_the_genomic_accession=141497881, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100507422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507422, CurrentID=0, Status='live', Symbol='LOC100507422', Aliases='', description='ubiquitin conjugating enzyme E2N pseudogene', other_designations='ubiquitin-conjugating enzyme E2N pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47592147, end_position_on_the_genomic_accession=47592608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507423, CurrentID=0, Status='live', Symbol='MKNK1-AS1', Aliases='', description='MKNK1 antisense RNA 1', other_designations='MKNK1 antisense RNA 1 (non-protein coding)', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46538696, end_position_on_the_genomic_accession=46570255, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100507424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507424, CurrentID=0, Status='live', Symbol='LOC100507424', Aliases='', description='uncharacterized LOC100507424', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2836816, end_position_on_the_genomic_accession=2859795, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507428, CurrentID=0, Status='live', Symbol='LINC00458', Aliases='LncRNA-ES3', description='long intergenic non-protein coding RNA 458', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54124324, end_position_on_the_genomic_accession=54132871, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507431, CurrentID=0, Status='live', Symbol='LOC100507431', Aliases='', description='uncharacterized LOC100507431', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130844193, end_position_on_the_genomic_accession=130862110, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507433, CurrentID=0, Status='live', Symbol='ZNF571-AS1', Aliases='', description='ZNF571 antisense RNA 1', other_designations='CTD-3064H18.6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37548949, end_position_on_the_genomic_accession=37587348, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100507435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507435, CurrentID=0, Status='live', Symbol='DLGAP2-AS1', Aliases='', description='DLGAP2 antisense RNA 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1565509, end_position_on_the_genomic_accession=1621664, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507436, CurrentID=0, Status='live', Symbol='MICA', Aliases='MIC-A, PERB11.1', description='MHC class I polypeptide-related sequence A', other_designations='HLA class I antigen|MHC class I chain-related protein A|MHC class I related sequence A|stress inducible class I homolog', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31399784, end_position_on_the_genomic_accession=31415315, orientation='plus', exon_count=7, OMIM=600169, no_hdr0=''),
  100507437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507437, CurrentID=0, Status='live', Symbol='LOC100507437', Aliases='', description='uncharacterized LOC100507437', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105417581, end_position_on_the_genomic_accession=105419739, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507443, CurrentID=0, Status='live', Symbol='LOC100507443', Aliases='', description='uncharacterized LOC100507443', other_designations='', map_location='2q33-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208119129, end_position_on_the_genomic_accession=208156762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507444, CurrentID=0, Status='live', Symbol='PPP1R2P1', Aliases='IPP-2P, PPP1R2P', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32876475, end_position_on_the_genomic_accession=32880074, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507445, CurrentID=0, Status='live', Symbol='LOC100507445', Aliases='', description='uncharacterized LOC100507445', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=22199117, end_position_on_the_genomic_accession=22199482, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130881, CurrentID=0, Status='live', Symbol='LOC100130881', Aliases='', description='zinc finger protein 12 pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43523241, end_position_on_the_genomic_accession=43525104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507458, CurrentID=0, Status='live', Symbol='ZNF213-AS1', Aliases='', description='ZNF213 antisense RNA 1 (head to head)', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3125529, end_position_on_the_genomic_accession=3134882, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507459, CurrentID=0, Status='live', Symbol='NRSN2-AS1', Aliases='', description='NRSN2 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=319687, end_position_on_the_genomic_accession=325252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507460, CurrentID=0, Status='live', Symbol='LOC100507460', Aliases='', description='uncharacterized LOC100507460', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131682485, end_position_on_the_genomic_accession=131699863, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507461, CurrentID=0, Status='live', Symbol='LOC100507461', Aliases='', description='uncharacterized LOC100507461', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148078127, end_position_on_the_genomic_accession=148088029, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507462, CurrentID=0, Status='live', Symbol='GVQW2', Aliases='', description='GVQW motif containing 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138725211, end_position_on_the_genomic_accession=138773679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507463, CurrentID=0, Status='live', Symbol='PSMB8-AS1', Aliases='TAP1-AS1, TAPSAR1', description='PSMB8 antisense RNA 1 (head to head)', other_designations='TAP1 and PSMB8 antisense RNA 1|XXbac-BPG246D15.8', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32844086, end_position_on_the_genomic_accession=32846500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507464, CurrentID=0, Status='live', Symbol='LOC100507464', Aliases='', description='uncharacterized LOC100507464', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49496763, end_position_on_the_genomic_accession=49512234, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507466, CurrentID=0, Status='live', Symbol='DPH6-AS1', Aliases='ATPBD4-AS1', description='DPH6 antisense RNA 1 (head to head)', other_designations='ATPBD4 antisense RNA 1 (head to head)|ATPBD4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35546195, end_position_on_the_genomic_accession=35859001, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418871, CurrentID=0, Status='live', Symbol='LOC100418871', Aliases='', description='cytohesin 1 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21939479, end_position_on_the_genomic_accession=21940137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507468, CurrentID=0, Status='live', Symbol='LOC100507468', Aliases='', description='uncharacterized LOC100507468', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69596137, end_position_on_the_genomic_accession=69597495, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377762, CurrentID=0, Status='live', Symbol='LOC105377762', Aliases='', description='uncharacterized LOC105377762', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179534545, end_position_on_the_genomic_accession=179550505, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507470, CurrentID=0, Status='live', Symbol='GRID1-AS1', Aliases='', description='GRID1 antisense RNA 1', other_designations='GRID1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85577731, end_position_on_the_genomic_accession=85607215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507472, CurrentID=0, Status='live', Symbol='LOC100507472', Aliases='', description='uncharacterized LOC100507472', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101307251, end_position_on_the_genomic_accession=101309304, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507475, CurrentID=0, Status='live', Symbol='IDH1-AS1', Aliases='', description='IDH1 antisense RNA 1', other_designations='IDH1 antisense RNA 1 (non-protein coding)', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208255233, end_position_on_the_genomic_accession=208256194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507477, CurrentID=0, Status='live', Symbol='LOC100507477', Aliases='', description='uncharacterized LOC100507477', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139976319, end_position_on_the_genomic_accession=140093721, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507480, CurrentID=0, Status='live', Symbol='LOC100507480', Aliases='', description='uncharacterized LOC100507480', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36045900, end_position_on_the_genomic_accession=36049529, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507483, CurrentID=0, Status='live', Symbol='LOC100507483', Aliases='', description='uncharacterized LOC100507483', other_designations='3-AS protein', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=16130552, end_position_on_the_genomic_accession=16131094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507487, CurrentID=0, Status='live', Symbol='LOC100507487', Aliases='', description='uncharacterized LOC100507487', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=128428016, end_position_on_the_genomic_accession=128519398, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100507488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507488, CurrentID=0, Status='live', Symbol='DNAPTP3', Aliases='', description='histone demethylase UTY-like', other_designations='DNA polymerase-transactivated protein 3', map_location='2q34-q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=216224632, end_position_on_the_genomic_accession=216226762, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507489, CurrentID=0, Status='live', Symbol='LINC01277', Aliases='', description='long intergenic non-protein coding RNA 1277', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142966421, end_position_on_the_genomic_accession=143037582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507490, CurrentID=0, Status='live', Symbol='ZNF859P', Aliases='', description='zinc finger protein 859, pseudogene', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=49844652, end_position_on_the_genomic_accession=49846489, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507494, CurrentID=0, Status='live', Symbol='LOC100507494', Aliases='', description='uncharacterized LOC100507494', other_designations='FLJ00403 protein', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100507495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507495, CurrentID=0, Status='live', Symbol='SDCBP2-AS1', Aliases='', description='SDCBP2 antisense RNA 1', other_designations='SDCBP2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1325343, end_position_on_the_genomic_accession=1378735, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271418, CurrentID=0, Status='live', Symbol='RPL17P42', Aliases='RPL17_25_1540', description='ribosomal protein L17 pseudogene 42', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35719907, end_position_on_the_genomic_accession=35720426, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130001, CurrentID=0, Status='live', Symbol='LOC100130001', Aliases='', description='small nuclear ribonucleoprotein polypeptide B pseudogene', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56975596, end_position_on_the_genomic_accession=56976224, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507500, CurrentID=0, Status='live', Symbol='EGFR-AS1', Aliases='', description='EGFR antisense RNA 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55179750, end_position_on_the_genomic_accession=55188949, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507501, CurrentID=0, Status='live', Symbol='LINC01569', Aliases='', description='long intergenic non-protein coding RNA 1569', other_designations='TCONS_00024567', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4245825, end_position_on_the_genomic_accession=4253789, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507506, CurrentID=0, Status='live', Symbol='LOC100507506', Aliases='', description='uncharacterized LOC100507506', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4136072, end_position_on_the_genomic_accession=4157385, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507507, CurrentID=0, Status='live', Symbol='LOC100507507', Aliases='', description='uncharacterized LOC100507507', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143379155, end_position_on_the_genomic_accession=143380795, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507513, CurrentID=0, Status='live', Symbol='LOC100507513', Aliases='', description='uncharacterized LOC100507513', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20870237, end_position_on_the_genomic_accession=20936309, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507516, CurrentID=0, Status='live', Symbol='LOC100507516', Aliases='', description='uncharacterized LOC100507516', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53511779, end_position_on_the_genomic_accession=53524084, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100126517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126517, CurrentID=0, Status='live', Symbol='TRNAE-UUC', Aliases='', description='transfer RNA glutamic acid (anticodon UUC)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16535279, end_position_on_the_genomic_accession=16535350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507521, CurrentID=0, Status='live', Symbol='LOC100507521', Aliases='', description='uncharacterized LOC100507521', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62050913, end_position_on_the_genomic_accession=62082393, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507524, CurrentID=0, Status='live', Symbol='ARHGEF26-AS1', Aliases='', description='ARHGEF26 antisense RNA 1', other_designations='ARHGEF26 antisense RNA 1 (non-protein coding)', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154024401, end_position_on_the_genomic_accession=154121210, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507526, CurrentID=0, Status='live', Symbol='SMG1P4', Aliases='', description='SMG1 pseudogene 4', other_designations='SMG1 phosphatidylinositol 3-kinase-related kinase pseudogene|smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21879337, end_position_on_the_genomic_accession=21916024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507527, CurrentID=0, Status='live', Symbol='THEG5', Aliases='', description='testis highly expressed protein 5', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31588189, end_position_on_the_genomic_accession=31593550, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507528, CurrentID=0, Status='live', Symbol='LINC00613', Aliases='', description='long intergenic non-protein coding RNA 613', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135866983, end_position_on_the_genomic_accession=135913680, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507530, CurrentID=0, Status='live', Symbol='MCPH1-AS1', Aliases='', description='MCPH1 antisense RNA 1', other_designations='CTD-2541M15.1|CTD-2541M15.3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6615603, end_position_on_the_genomic_accession=6708209, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371791, CurrentID=0, Status='live', Symbol='LOC105371791', Aliases='', description='uncharacterized LOC105371791', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44536276, end_position_on_the_genomic_accession=44537857, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507533, CurrentID=0, Status='live', Symbol='SOX21-AS1', Aliases='', description='SOX21 antisense RNA 1 (head to head)', other_designations='SOX21 antisense RNA 1 (non-protein coding)', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94712716, end_position_on_the_genomic_accession=94715945, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507534, CurrentID=0, Status='live', Symbol='LOC100507534', Aliases='', description='uncharacterized LOC100507534', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47849314, end_position_on_the_genomic_accession=47887130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377773, CurrentID=0, Status='live', Symbol='LOC105377773', Aliases='', description='uncharacterized LOC105377773', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=308407, end_position_on_the_genomic_accession=311282, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507537, CurrentID=0, Status='live', Symbol='LOC100507537', Aliases='', description='uncharacterized LOC100507537', other_designations='', map_location='3q25', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155290232, end_position_on_the_genomic_accession=155293700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507540, CurrentID=0, Status='live', Symbol='LINC01504', Aliases='', description='long intergenic non-protein coding RNA 1504', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72305436, end_position_on_the_genomic_accession=72343210, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507546, CurrentID=0, Status='live', Symbol='LOC100507546', Aliases='', description='uncharacterized LOC100507546', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=180625202, end_position_on_the_genomic_accession=180630117, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507547, CurrentID=0, Status='live', Symbol='LOC100507547', Aliases='', description='uncharacterized LOC100507547', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32152802, end_position_on_the_genomic_accession=32154365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507548, CurrentID=0, Status='live', Symbol='LOC100507548', Aliases='', description='uncharacterized LOC100507548', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134985284, end_position_on_the_genomic_accession=134986939, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507549, CurrentID=0, Status='live', Symbol='HMGB3P32', Aliases='', description='high mobility group box 3 pseudogene 32', other_designations='high-mobility group box 3 pseudogene 32', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57603098, end_position_on_the_genomic_accession=57605538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507551, CurrentID=0, Status='live', Symbol='LOC100507551', Aliases='', description='uncharacterized LOC100507551', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=17611431, end_position_on_the_genomic_accession=17613273, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507554, CurrentID=0, Status='live', Symbol='LOC100507554', Aliases='', description='uncharacterized LOC100507554', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216483897, end_position_on_the_genomic_accession=216499263, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100507556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507556, CurrentID=0, Status='live', Symbol='LOC100507556', Aliases='', description='putative uncharacterized protein NCRNA00269-like', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=155459931, end_position_on_the_genomic_accession=155461515, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507557, CurrentID=0, Status='live', Symbol='LOC100507557', Aliases='', description='uncharacterized LOC100507557', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145734869, end_position_on_the_genomic_accession=145886585, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  651249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651249, CurrentID=0, Status='live', Symbol='RPL34P27', Aliases='RPL34_11_1303', description='ribosomal protein L34 pseudogene 27', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24988533, end_position_on_the_genomic_accession=24988956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507560, CurrentID=0, Status='live', Symbol='LOC100507560', Aliases='', description='uncharacterized LOC100507560', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5017685, end_position_on_the_genomic_accession=5025594, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100507561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507561, CurrentID=0, Status='live', Symbol='LOC100507561', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55872381, end_position_on_the_genomic_accession=55874589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371792, CurrentID=0, Status='live', Symbol='LOC105371792', Aliases='', description='uncharacterized LOC105371792', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44678962, end_position_on_the_genomic_accession=44688218, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507564, CurrentID=0, Status='live', Symbol='LOC100507564', Aliases='', description='uncharacterized LOC100507564', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53238610, end_position_on_the_genomic_accession=53242783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377778, CurrentID=0, Status='live', Symbol='LOC105377778', Aliases='', description='uncharacterized LOC105377778', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1748302, end_position_on_the_genomic_accession=1755682, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507567, CurrentID=0, Status='live', Symbol='RAB11B-AS1', Aliases='', description='RAB11B antisense RNA 1', other_designations='RAB11B antisense RNA 1 (non-protein coding)', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8374376, end_position_on_the_genomic_accession=8390691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418888, CurrentID=0, Status='live', Symbol='LOC100418888', Aliases='', description='peptidase D pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55560897, end_position_on_the_genomic_accession=55561577, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418889, CurrentID=0, Status='live', Symbol='STX18P1', Aliases='', description='syntaxin 18 pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30068632, end_position_on_the_genomic_accession=30069631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507577, CurrentID=0, Status='live', Symbol='LOC100507577', Aliases='', description='uncharacterized LOC100507577', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48355685, end_position_on_the_genomic_accession=48362999, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507582, CurrentID=0, Status='live', Symbol='BHLHE40-AS1', Aliases='', description='BHLHE40 antisense RNA 1', other_designations='', map_location='3p26', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4896809, end_position_on_the_genomic_accession=4979961, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507584, CurrentID=0, Status='live', Symbol='LINC01016', Aliases='', description='long intergenic non-protein coding RNA 1016', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33889511, end_position_on_the_genomic_accession=33896907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507586, CurrentID=0, Status='live', Symbol='LOC100507586', Aliases='', description='cytochrome c oxidase subunit VIIb pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54089793, end_position_on_the_genomic_accession=54090220, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507588, CurrentID=0, Status='live', Symbol='TGFBR3L', Aliases='TGFR-3L', description='transforming growth factor beta receptor III like', other_designations='TGF-beta receptor type III-like protein|TGF-beta receptor type-3-like protein|transforming growth factor, beta receptor III-like|transforming growth factor-beta receptor type III-like protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7915455, end_position_on_the_genomic_accession=7919097, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507589, CurrentID=0, Status='live', Symbol='NAGPA-AS1', Aliases='', description='NAGPA antisense RNA 1', other_designations='NAGPA antisense RNA 1 (non-protein coding)', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5033702, end_position_on_the_genomic_accession=5042971, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507599, CurrentID=0, Status='live', Symbol='LOC100507599', Aliases='', description='uncharacterized LOC100507599', other_designations='CTA-445C9.14', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26512510, end_position_on_the_genomic_accession=26514568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507600, CurrentID=0, Status='live', Symbol='LOC100507600', Aliases='', description='uncharacterized LOC100507600', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135820191, end_position_on_the_genomic_accession=135823087, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377784, CurrentID=0, Status='live', Symbol='LOC105377784', Aliases='', description='uncharacterized LOC105377784', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2467499, end_position_on_the_genomic_accession=2478505, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507602, CurrentID=0, Status='live', Symbol='TRIM52-AS1', Aliases='', description='TRIM52 antisense RNA 1 (head to head)', other_designations='CTC-338M12.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181261212, end_position_on_the_genomic_accession=181272307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100507605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507605, CurrentID=0, Status='live', Symbol='LINC00837', Aliases='', description='long intergenic non-protein coding RNA 837', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28789311, end_position_on_the_genomic_accession=28796050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507606, CurrentID=0, Status='live', Symbol='LOC100507606', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49048268, end_position_on_the_genomic_accession=49050159, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507607, CurrentID=0, Status='live', Symbol='NPIPB9', Aliases='', description='nuclear pore complex interacting protein family member B9', other_designations='Nuclear pore complex-interacting protein family member B9|nuclear pore complex interacting protein family, member B9', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28742728, end_position_on_the_genomic_accession=28772850, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100507608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507608, CurrentID=0, Status='live', Symbol='KRTAP9-6', Aliases='KAP9.6, KRTAP9.6, KRTAP9L2', description='keratin associated protein 9-6', other_designations='keratin associated protein 9-like 2|putative keratin-associated protein 9-2-like 2', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41265378, end_position_on_the_genomic_accession=41265860, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507612, CurrentID=0, Status='live', Symbol='LINC00402', Aliases='', description='long intergenic non-protein coding RNA 402', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=74805581, end_position_on_the_genomic_accession=74834113, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  651258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651258, CurrentID=0, Status='live', Symbol='GAPDHP37', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 37', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40673289, end_position_on_the_genomic_accession=40674522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507616, CurrentID=0, Status='live', Symbol='LOC100507616', Aliases='', description='uncharacterized LOC100507616', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92466431, end_position_on_the_genomic_accession=92486360, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507617, CurrentID=0, Status='live', Symbol='PPP1R14BP3', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 14B pseudogene 3', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 14B pseudogene 3', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139114733, end_position_on_the_genomic_accession=139115539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102724041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102724041, CurrentID=0, Status='live', Symbol='LOC102724041', Aliases='', description='uncharacterized LOC102724041', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231461477, end_position_on_the_genomic_accession=231520998, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100287825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287825, CurrentID=0, Status='live', Symbol='LOC100287825', Aliases='', description='golgin A8 family member A pseudogene', other_designations='golgin A8 family, member A pseudogene', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30523144, end_position_on_the_genomic_accession=30532220, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507629, CurrentID=0, Status='live', Symbol='LINC00658', Aliases='', description='long intergenic non-protein coding RNA 658', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5431957, end_position_on_the_genomic_accession=5445748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100507630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507630, CurrentID=0, Status='live', Symbol='LOC100507630', Aliases='', description='uncharacterized LOC100507630', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=24447506, end_position_on_the_genomic_accession=24449402, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507632, CurrentID=0, Status='live', Symbol='LINC00968', Aliases='', description='long intergenic non-protein coding RNA 968', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56518318, end_position_on_the_genomic_accession=56559823, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507633, CurrentID=0, Status='live', Symbol='TNKS2-AS1', Aliases='', description='TNKS2 antisense RNA 1 (head to head)', other_designations='', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91782839, end_position_on_the_genomic_accession=91798291, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507634, CurrentID=0, Status='live', Symbol='LOC100507634', Aliases='', description='uncharacterized LOC100507634', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55215408, end_position_on_the_genomic_accession=55217455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377790, CurrentID=0, Status='live', Symbol='LOC105377790', Aliases='', description='uncharacterized LOC105377790', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=3861263, end_position_on_the_genomic_accession=3869294, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100507639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507639, CurrentID=0, Status='live', Symbol='LOC100507639', Aliases='', description='uncharacterized LOC100507639', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=141321124, end_position_on_the_genomic_accession=141332617, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507642, CurrentID=0, Status='live', Symbol='LOC100507642', Aliases='', description='uncharacterized LOC100507642', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149718, end_position_on_the_genomic_accession=155465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377791, CurrentID=0, Status='live', Symbol='LOC105377791', Aliases='', description='uncharacterized LOC105377791', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=3409128, end_position_on_the_genomic_accession=3426587, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507646, CurrentID=0, Status='live', Symbol='LOC100507646', Aliases='', description='uncharacterized LOC100507646', other_designations='NINP6167', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=40779396, end_position_on_the_genomic_accession=40781458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507648, CurrentID=0, Status='live', Symbol='CICP15', Aliases='', description='capicua transcriptional repressor pseudogene 15', other_designations='capicua homolog pseudogene 15', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181328132, end_position_on_the_genomic_accession=181331896, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377792, CurrentID=0, Status='live', Symbol='LOC105377792', Aliases='', description='protein app1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=3001061, end_position_on_the_genomic_accession=3024227, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507650, CurrentID=0, Status='live', Symbol='RNF212B', Aliases='C14orf164', description='ring finger protein 212B', other_designations='RING finger protein C14orf164|RING finger protein C14orf164-like', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23185334, end_position_on_the_genomic_accession=23274285, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100507651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507651, CurrentID=0, Status='live', Symbol='LINC01606', Aliases='', description='long intergenic non-protein coding RNA 1606', other_designations='TCONS_l2_00027778', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57218276, end_position_on_the_genomic_accession=57232954, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100507654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507654, CurrentID=0, Status='live', Symbol='LOC100507654', Aliases='', description='uncharacterized LOC100507654', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=101187995, end_position_on_the_genomic_accession=101232367, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507657, CurrentID=0, Status='live', Symbol='LOC100507657', Aliases='', description='uncharacterized LOC100507657', other_designations='CTA-929C8.6', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27311743, end_position_on_the_genomic_accession=27318444, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100418903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418903, CurrentID=0, Status='live', Symbol='NBEAP5', Aliases='', description='neurobeachin pseudogene 5', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19130636, end_position_on_the_genomic_accession=19154329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100507660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507660, CurrentID=0, Status='live', Symbol='SRGAP3-AS2', Aliases='', description='SRGAP3 antisense RNA 2', other_designations='SRGAP3 antisense RNA 2 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=9234219, end_position_on_the_genomic_accession=9236137, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507661, CurrentID=0, Status='live', Symbol='LOC100507661', Aliases='', description='uncharacterized LOC100507661', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168901945, end_position_on_the_genomic_accession=168921996, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100507662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507662, CurrentID=0, Status='live', Symbol='LOC100507662', Aliases='', description='uncharacterized LOC100507662', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=151408775, end_position_on_the_genomic_accession=151473355, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100507663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507663, CurrentID=0, Status='live', Symbol='LOC100507663', Aliases='', description='uncharacterized LOC100507663', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=31310000, end_position_on_the_genomic_accession=31311746, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507664, CurrentID=0, Status='live', Symbol='LOC100507664', Aliases='', description='uncharacterized LOC100507664', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=63073437, end_position_on_the_genomic_accession=63075255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418904, CurrentID=0, Status='live', Symbol='NBEAP6', Aliases='', description='neurobeachin pseudogene 6', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19245668, end_position_on_the_genomic_accession=19269380, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377795, CurrentID=0, Status='live', Symbol='LOC105377795', Aliases='', description='uncharacterized LOC105377795', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5858274, end_position_on_the_genomic_accession=6004226, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100507670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507670, CurrentID=0, Status='live', Symbol='LOC100507670', Aliases='', description='uncharacterized LOC100507670', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151279896, end_position_on_the_genomic_accession=151281929, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375256, CurrentID=0, Status='live', Symbol='LOC105375256', Aliases='', description='uncharacterized LOC105375256', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44065428, end_position_on_the_genomic_accession=44071671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377796, CurrentID=0, Status='live', Symbol='LOC105377796', Aliases='', description='uncharacterized LOC105377796', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6033655, end_position_on_the_genomic_accession=6036938, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507674, CurrentID=0, Status='live', Symbol='MARK2P9', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 9', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92418661, end_position_on_the_genomic_accession=92419606, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418906, CurrentID=0, Status='live', Symbol='LOC100418906', Aliases='', description='sideroflexin 4 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142947364, end_position_on_the_genomic_accession=142948174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100507679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507679, CurrentID=0, Status='live', Symbol='MUC22', Aliases='PBMUCL1', description='mucin 22', other_designations='panbronchiolitis related mucin-like 1|panbronchiolitis-related mucin-like protein 1', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31005952, end_position_on_the_genomic_accession=31035570, orientation='plus', exon_count=5, OMIM=613917, no_hdr0=''),
  100270673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270673, CurrentID=0, Status='live', Symbol='FABP5P12', Aliases='FABP5L12', description='fatty acid binding protein 5 pseudogene 12', other_designations='fatty acid binding protein 5 (psoriasis-associated) pseudogene|fatty acid binding protein 5-like 12', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159026061, end_position_on_the_genomic_accession=159026724, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377799, CurrentID=0, Status='live', Symbol='LOC105377799', Aliases='', description='uncharacterized LOC105377799', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7072864, end_position_on_the_genomic_accession=7073294, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100507703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507703, CurrentID=0, Status='live', Symbol='LOC100507703', Aliases='', description='HLA class I histocompatibility antigen, A-69 alpha chain-like', other_designations='MHC class I antigen|major histocompatibility complex, class I, A', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  126987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126987, CurrentID=0, Status='live', Symbol='LOC126987', Aliases='', description='tight junction protein 3 (zona occludens 3) pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106080760, end_position_on_the_genomic_accession=106081379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421558, CurrentID=0, Status='live', Symbol='LOC100421558', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29391525, end_position_on_the_genomic_accession=29393846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377806, CurrentID=0, Status='live', Symbol='LOC105377806', Aliases='', description='uncharacterized LOC105377806', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113936588, end_position_on_the_genomic_accession=113945367, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100507739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507739, CurrentID=0, Status='live', Symbol='ERVK3-1', Aliases='', description='endogenous retrovirus group K3 member 1', other_designations='HERV-K(HML6-1)|endogenous retrovirus group K3, member 1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=58823740, end_position_on_the_genomic_accession=58827023, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377808, CurrentID=0, Status='live', Symbol='LOC105377808', Aliases='', description='LWamide neuropeptides-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114007613, end_position_on_the_genomic_accession=114009124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507747, CurrentID=0, Status='live', Symbol='LOC100507747', Aliases='', description='uncharacterized LOC100507747', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113956426, end_position_on_the_genomic_accession=113974083, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377809, CurrentID=0, Status='live', Symbol='LOC105377809', Aliases='', description='uncharacterized LOC105377809', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113946084, end_position_on_the_genomic_accession=113954544, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170685, CurrentID=0, Status='live', Symbol='NUDT10', Aliases='APS2, DIPP3-alpha, DIPP3a', description='nudix hydrolase 10', other_designations="diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 3-alpha|diadenosine hexaphosphate hydrolase (AMP-forming)|nucleoside diphosphate-linked moiety X motif 10|nudix (nucleoside diphosphate linked moiety X)-type motif 10|nudix motif 10", map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51332231, end_position_on_the_genomic_accession=51337525, orientation='plus', exon_count=3, OMIM=300527, no_hdr0=''),
  105377813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377813, CurrentID=0, Status='live', Symbol='LOC105377813', Aliases='', description='endogenous retrovirus group K member 11 Pol protein', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377814, CurrentID=0, Status='live', Symbol='LOC105377814', Aliases='', description='uncharacterized LOC105377814', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37233424, end_position_on_the_genomic_accession=37238742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287852, CurrentID=0, Status='live', Symbol='LOC100287852', Aliases='', description='protein FAM136A pseudogene', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101414203, end_position_on_the_genomic_accession=101414648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376164, CurrentID=0, Status='live', Symbol='LOC105376164', Aliases='', description='uncharacterized LOC105376164', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96620028, end_position_on_the_genomic_accession=96621416, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377816, CurrentID=0, Status='live', Symbol='LOC105377816', Aliases='', description='uncharacterized LOC105377816', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73523599, end_position_on_the_genomic_accession=73526034, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377817, CurrentID=0, Status='live', Symbol='LOC105377817', Aliases='', description='uncharacterized LOC105377817', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371800, CurrentID=0, Status='live', Symbol='LOC105371800', Aliases='', description='uncharacterized LOC105371800', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45922506, end_position_on_the_genomic_accession=45935950, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100134938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134938, CurrentID=0, Status='live', Symbol='UPK3BL', Aliases='UPLP', description='uroplakin 3B-like', other_designations='uroplakin 3C protein|uroplakin-3-like protein|uroplakin-like protein', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102637025, end_position_on_the_genomic_accession=102642791, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100287856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287856, CurrentID=0, Status='live', Symbol='LOC100287856', Aliases='', description='cornichon family AMPA receptor auxiliary protein 4 pseudogene', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127416379, end_position_on_the_genomic_accession=127416949, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377819, CurrentID=0, Status='live', Symbol='LOC105377819', Aliases='', description='uncharacterized LOC105377819', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18962129, end_position_on_the_genomic_accession=19019869, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100287859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287859, CurrentID=0, Status='live', Symbol='FAM197Y6', Aliases='', description='family with sequence similarity 197 Y-linked member 6', other_designations='family with sequence similarity 197, Y-linked, member 6', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9231475, end_position_on_the_genomic_accession=9233681, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377823, CurrentID=0, Status='live', Symbol='LOC105377823', Aliases='', description='uncharacterized LOC105377823', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22379256, end_position_on_the_genomic_accession=22380521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377824, CurrentID=0, Status='live', Symbol='LOC105377824', Aliases='', description='uncharacterized LOC105377824', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22375280, end_position_on_the_genomic_accession=22378151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376166, CurrentID=0, Status='live', Symbol='LOC105376166', Aliases='', description='uncharacterized LOC105376166', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97128242, end_position_on_the_genomic_accession=97130679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127011, CurrentID=0, Status='live', Symbol='LOC127011', Aliases='', description='ATPase, H+ transporting, lysosomal accessory protein 2 pseudogene', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197222222, end_position_on_the_genomic_accession=197223252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377828, CurrentID=0, Status='live', Symbol='LOC105377828', Aliases='', description='uncharacterized LOC105377828', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421559, CurrentID=0, Status='live', Symbol='LOC100421559', Aliases='', description='adaptor-related protein complex 5, mu 1 subunit pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=39873782, end_position_on_the_genomic_accession=39874388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  651302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651302, CurrentID=0, Status='live', Symbol='ZNF192P1', Aliases='ZNF389, ZNF389P, dJ265C24.4', description='zinc finger protein 192 pseudogene 1', other_designations='zinc finger protein 389, pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28161761, end_position_on_the_genomic_accession=28169595, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100418940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418940, CurrentID=0, Status='live', Symbol='LOC100418940', Aliases='', description='ATM interactor pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55580377, end_position_on_the_genomic_accession=55581459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377832, CurrentID=0, Status='live', Symbol='LOC105377832', Aliases='', description='uncharacterized LOC105377832', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377833, CurrentID=0, Status='live', Symbol='LOC105377833', Aliases='', description='uncharacterized LOC105377833', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377834, CurrentID=0, Status='live', Symbol='LOC105377834', Aliases='', description='uncharacterized LOC105377834', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645139, CurrentID=0, Status='live', Symbol='LOC645139', Aliases='', description='poly(A) binding protein interacting protein 1 pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18650196, end_position_on_the_genomic_accession=18651539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372415, CurrentID=0, Status='live', Symbol='LOC105372415', Aliases='', description='uncharacterized LOC105372415', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44194511, end_position_on_the_genomic_accession=44198475, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287877, CurrentID=0, Status='live', Symbol='LOC100287877', Aliases='', description='uncharacterized LOC100287877', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5705568, end_position_on_the_genomic_accession=5713772, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377840, CurrentID=0, Status='live', Symbol='LOC105377840', Aliases='', description='uncharacterized LOC105377840', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66050908, end_position_on_the_genomic_accession=66060428, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100287879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287879, CurrentID=0, Status='live', Symbol='LINC00994', Aliases='', description='long intergenic non-protein coding RNA 994', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64078361, end_position_on_the_genomic_accession=64087363, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377842, CurrentID=0, Status='live', Symbol='LOC105377842', Aliases='', description='uncharacterized LOC105377842', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67300675, end_position_on_the_genomic_accession=67352437, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104472714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472714, CurrentID=0, Status='live', Symbol='DISC1-IT1', Aliases='', description='DISC1 intronic transcript 1', other_designations='DISC1 intronic transcript 1 (non-protein coding)', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231925834, end_position_on_the_genomic_accession=231945205, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  79272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79272, CurrentID=0, Status='live', Symbol='OR52X1P', Aliases='', description='olfactory receptor family 52 subfamily X member 1 pseudogene', other_designations='olfactory receptor, family 52, subfamily X, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6066713, end_position_on_the_genomic_accession=6067892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422440, CurrentID=0, Status='live', Symbol='LOC100422440', Aliases='', description='general transcription factor IIE subunit 2 pseudogene', other_designations='general transcription factor IIE, polypeptide 2, beta 34kDa pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94992914, end_position_on_the_genomic_accession=94993503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377849, CurrentID=0, Status='live', Symbol='LOC105377849', Aliases='', description='uncharacterized LOC105377849', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70412828, end_position_on_the_genomic_accession=70414036, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100418959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418959, CurrentID=0, Status='live', Symbol='LOC100418959', Aliases='', description='aminoacylase 1 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68992719, end_position_on_the_genomic_accession=68993372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377851, CurrentID=0, Status='live', Symbol='LOC105377851', Aliases='', description='uncharacterized LOC105377851', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71224923, end_position_on_the_genomic_accession=71234722, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100418961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418961, CurrentID=0, Status='live', Symbol='LOC100418961', Aliases='', description='aurora kinase A pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128306708, end_position_on_the_genomic_accession=128307585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377852, CurrentID=0, Status='live', Symbol='LOC105377852', Aliases='', description='uncharacterized LOC105377852', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71293618, end_position_on_the_genomic_accession=71295464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422441, CurrentID=0, Status='live', Symbol='LOC100422441', Aliases='', description='heat shock protein 90kDa alpha family class B member 1 pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=78041880, end_position_on_the_genomic_accession=78044136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104472716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104472716, CurrentID=0, Status='live', Symbol='WARS2-IT1', Aliases='', description='WARS2 intronic transcript 1', other_designations='WARS2 intronic transcript 1 (non-protein coding)', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119047405, end_position_on_the_genomic_accession=119063287, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100508046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508046, CurrentID=0, Status='live', Symbol='LOC100508046', Aliases='', description='uncharacterized LOC100508046', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18976748, end_position_on_the_genomic_accession=18980793, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100287896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287896, CurrentID=0, Status='live', Symbol='LOC100287896', Aliases='', description='uncharacterized LOC100287896', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74493440, end_position_on_the_genomic_accession=74498242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928202, CurrentID=0, Status='live', Symbol='LOC101928202', Aliases='', description='SLAM family member 5', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157284342, end_position_on_the_genomic_accession=157311377, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100287900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287900, CurrentID=0, Status='live', Symbol='GM2AP2', Aliases='', description='GM2 ganglioside activator pseudogene 2', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171392248, end_position_on_the_genomic_accession=171392791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418974, CurrentID=0, Status='live', Symbol='TUBB4BP2', Aliases='', description='tubulin beta 4B class IVb pseudogene 2', other_designations='tubulin, beta 2B class IIb pseudogene|tubulin, beta 4B class IVb pseudogene 2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34109174, end_position_on_the_genomic_accession=34109967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  651337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651337, CurrentID=0, Status='live', Symbol='LOC651337', Aliases='', description='uncharacterized LOC651337', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=140679527, end_position_on_the_genomic_accession=140679967, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373513, CurrentID=0, Status='live', Symbol='LOC105373513', Aliases='', description='uncharacterized LOC105373513', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101325593, end_position_on_the_genomic_accession=101371755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100418977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418977, CurrentID=0, Status='live', Symbol='SNX18P5', Aliases='SNX18P6', description='sorting nexin 18 pseudogene 5', other_designations='sorting nexin 18 pseudogene 6', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42922701, end_position_on_the_genomic_accession=42924096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377869, CurrentID=0, Status='live', Symbol='LOC105377869', Aliases='', description='uncharacterized LOC105377869', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80530627, end_position_on_the_genomic_accession=80548624, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=79273, CurrentID=0, Status='live', Symbol='OR7E94P', Aliases='', description='olfactory receptor family 7 subfamily E member 94 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 94 pseudogene', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79587754, end_position_on_the_genomic_accession=79588083, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100114903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100114903, CurrentID=0, Status='live', Symbol='NRBF2P2', Aliases='', description='nuclear receptor binding factor 2 pseudogene 2', other_designations='', map_location='3p', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46023112, end_position_on_the_genomic_accession=46024911, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100508120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508120, CurrentID=0, Status='live', Symbol='GMDS-AS1', Aliases='', description='GMDS antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2245753, end_position_on_the_genomic_accession=2413591, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100287908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287908, CurrentID=0, Status='live', Symbol='ZNF519P2', Aliases='', description='zinc finger protein 519 pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20900539, end_position_on_the_genomic_accession=20902075, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377871, CurrentID=0, Status='live', Symbol='LOC105377871', Aliases='', description='uncharacterized LOC105377871', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81526712, end_position_on_the_genomic_accession=81544045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100287912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287912, CurrentID=0, Status='live', Symbol='LOC100287912', Aliases='', description='ankyrin repeat domain 11 pseudogene', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81194139, end_position_on_the_genomic_accession=81201179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389203, CurrentID=0, Status='live', Symbol='SMIM20', Aliases='C4orf52', description='small integral membrane protein 20', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25914192, end_position_on_the_genomic_accession=25929879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928205, CurrentID=0, Status='live', Symbol='LOC101928205', Aliases='', description='uncharacterized LOC101928205', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72072333, end_position_on_the_genomic_accession=72093481, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130137, CurrentID=0, Status='live', Symbol='LOC100130137', Aliases='', description='integrin beta 1 binding protein 1 pseudogene', other_designations='', map_location='1q31.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192800572, end_position_on_the_genomic_accession=192801173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  127062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127062, CurrentID=0, Status='live', Symbol='OR2M3', Aliases='OR1-54P, OR2M6, OST003, OR2M3', description='olfactory receptor family 2 subfamily M member 3', other_designations='novel 7 transmembrane receptor (rhodopsin family) protein|olfactory receptor 2M6|olfactory receptor OR1-54|olfactory receptor, family 2, subfamily M, member 3 pseudogene|olfactory receptor, family 2, subfamily M, member 6', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248203068, end_position_on_the_genomic_accession=248204006, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377878, CurrentID=0, Status='live', Symbol='LOC105377878', Aliases='', description='uncharacterized LOC105377878', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82880833, end_position_on_the_genomic_accession=82882194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377879, CurrentID=0, Status='live', Symbol='LOC105377879', Aliases='', description='uncharacterized LOC105377879', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83722331, end_position_on_the_genomic_accession=83727629, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127064, CurrentID=0, Status='live', Symbol='OR2T12', Aliases='OR1-57', description='olfactory receptor family 2 subfamily T member 12', other_designations='novel 7 transmembrane receptor (rhodopsin family) protein|olfactory receptor OR1-57|olfactory receptor, family 2, subfamily T, member 12', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248294616, end_position_on_the_genomic_accession=248295578, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100287919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287919, CurrentID=0, Status='live', Symbol='VN1R30P', Aliases='', description='vomeronasal 1 receptor 30 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63345840, end_position_on_the_genomic_accession=63346732, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377882, CurrentID=0, Status='live', Symbol='LOC105377882', Aliases='', description='uncharacterized LOC105377882', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87930924, end_position_on_the_genomic_accession=87942511, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377884, CurrentID=0, Status='live', Symbol='LOC105377884', Aliases='', description='uncharacterized LOC105377884', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88277591, end_position_on_the_genomic_accession=88299584, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377885, CurrentID=0, Status='live', Symbol='LOC105377885', Aliases='', description='uncharacterized LOC105377885', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88299739, end_position_on_the_genomic_accession=88442913, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377886, CurrentID=0, Status='live', Symbol='LOC105377886', Aliases='', description='small EDRK-rich factor 2 pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89001391, end_position_on_the_genomic_accession=89003572, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100431177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100431177, CurrentID=0, Status='live', Symbol='BHLHB9P1', Aliases='', description='basic helix-loop-helix domain containing, class B, 9 pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80981589, end_position_on_the_genomic_accession=80985795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100508226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508226, CurrentID=0, Status='live', Symbol='LOC100508226', Aliases='', description='HHSL751', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=65994445, end_position_on_the_genomic_accession=65994875, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100508227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508227, CurrentID=0, Status='live', Symbol='LOC100508227', Aliases='', description='uncharacterized LOC100508227', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=186650391, end_position_on_the_genomic_accession=186650802, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372641, CurrentID=0, Status='live', Symbol='LOC105372641', Aliases='', description='uncharacterized LOC105372641', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48192338, end_position_on_the_genomic_accession=48202756, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377889, CurrentID=0, Status='live', Symbol='LOC105377889', Aliases='', description='uncharacterized LOC105377889', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89508178, end_position_on_the_genomic_accession=89515600, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377890, CurrentID=0, Status='live', Symbol='LOC105377890', Aliases='', description='uncharacterized LOC105377890', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90183915, end_position_on_the_genomic_accession=90185520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287928, CurrentID=0, Status='live', Symbol='LOC100287928', Aliases='', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), beta subunit pseudogene', other_designations='', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=5305514, end_position_on_the_genomic_accession=5307184, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929057, CurrentID=0, Status='live', Symbol='LOC101929057', Aliases='', description='uncharacterized LOC101929057', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89638532, end_position_on_the_genomic_accession=89689481, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130026, CurrentID=0, Status='live', Symbol='LOC100130026', Aliases='', description='ribosomal protein L38 pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201162434, end_position_on_the_genomic_accession=201162756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377893, CurrentID=0, Status='live', Symbol='LOC105377893', Aliases='', description='uncharacterized LOC105377893', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90641749, end_position_on_the_genomic_accession=90642684, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287931, CurrentID=0, Status='live', Symbol='LOC100287931', Aliases='', description='uncharacterized LOC100287931', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=123966, end_position_on_the_genomic_accession=125449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100419004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419004, CurrentID=0, Status='live', Symbol='SNX18P24', Aliases='', description='sorting nexin 18 pseudogene 24', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49560650, end_position_on_the_genomic_accession=49562149, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377895, CurrentID=0, Status='live', Symbol='LOC105377895', Aliases='', description='uncharacterized LOC105377895', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92208756, end_position_on_the_genomic_accession=92246265, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287934, CurrentID=0, Status='live', Symbol='LOC100287934', Aliases='', description='uncharacterized LOC100287934', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=785940, end_position_on_the_genomic_accession=787186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130027, CurrentID=0, Status='live', Symbol='LOC100130027', Aliases='', description='uncharacterized LOC100130027', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376181, CurrentID=0, Status='live', Symbol='LOC105376181', Aliases='', description='uncharacterized LOC105376181', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100901512, end_position_on_the_genomic_accession=100906823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105751182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105751182, CurrentID=0, Status='live', Symbol='THCAT126', Aliases='', description='thyroid cancer-associated transcript 126', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105751187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105751187, CurrentID=0, Status='live', Symbol='LRP1-AS', Aliases='', description='LRP1 antisense RNA', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377902, CurrentID=0, Status='live', Symbol='LOC105377902', Aliases='', description='uncharacterized LOC105377902', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=94607878, end_position_on_the_genomic_accession=94627356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928211, CurrentID=0, Status='live', Symbol='LOC101928211', Aliases='', description='uncharacterized LOC101928211', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124337380, end_position_on_the_genomic_accession=124352096, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100419016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419016, CurrentID=0, Status='live', Symbol='LOC100419016', Aliases='', description='proline rich 21 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32931768, end_position_on_the_genomic_accession=32932569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370896, CurrentID=0, Status='live', Symbol='LOC105370896', Aliases='', description='uncharacterized LOC105370896', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74797566, end_position_on_the_genomic_accession=74798234, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419020, CurrentID=0, Status='live', Symbol='SNX18P16', Aliases='', description='sorting nexin 18 pseudogene 16', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419021, CurrentID=0, Status='live', Symbol='SNX18P18', Aliases='', description='sorting nexin 18 pseudogene 18', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377912, CurrentID=0, Status='live', Symbol='LOC105377912', Aliases='', description='uncharacterized LOC105377912', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=101454359, end_position_on_the_genomic_accession=101463831, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100287951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287951, CurrentID=0, Status='live', Symbol='LOC100287951', Aliases='', description='dynein, light chain, Tctex-type 1 pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10410924, end_position_on_the_genomic_accession=10411994, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377915, CurrentID=0, Status='live', Symbol='LOC105377915', Aliases='', description='uncharacterized LOC105377915', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=103463681, end_position_on_the_genomic_accession=103534905, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100287954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287954, CurrentID=0, Status='live', Symbol='LOC100287954', Aliases='', description='double homeobox protein 4-like protein 4', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103752583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752583, CurrentID=0, Status='live', Symbol='LINC01075', Aliases='', description='long intergenic non-protein coding RNA 1075', other_designations='long intergenic non-protein coding RNA 1065', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62212282, end_position_on_the_genomic_accession=62249951, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377918, CurrentID=0, Status='live', Symbol='LOC105377918', Aliases='', description='uncharacterized LOC105377918', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104500831, end_position_on_the_genomic_accession=104527391, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377919, CurrentID=0, Status='live', Symbol='LOC105377919', Aliases='', description='uncharacterized LOC105377919', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105309552, end_position_on_the_genomic_accession=105328856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389249, CurrentID=0, Status='live', Symbol='LOC389249', Aliases='', description='peroxiredoxin 6 pseudogene', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187970273, end_position_on_the_genomic_accession=187971281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106144524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144524, CurrentID=0, Status='live', Symbol='HUNK-AS1', Aliases='', description='HUNK antisense RNA 1', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144526, CurrentID=0, Status='live', Symbol='PLA2G4C-AS1', Aliases='', description='PLA2G4C antisense RNA 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144527, CurrentID=0, Status='live', Symbol='LINC01438', Aliases='', description='long intergenic non-protein coding RNA 1438', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144528, CurrentID=0, Status='live', Symbol='DENND6A-AS1', Aliases='', description='DENND6A antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144529, CurrentID=0, Status='live', Symbol='GRPEL2-AS1', Aliases='', description='GRPEL2 antisense RNA 1', other_designations='GRPEL2 atnisense RNA 1', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144530, CurrentID=0, Status='live', Symbol='LINC01332', Aliases='', description='long intergenic non-protein coding RNA 1332', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144531, CurrentID=0, Status='live', Symbol='LINC00374', Aliases='', description='long intergenic non-protein coding RNA 374', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144532, CurrentID=0, Status='live', Symbol='ARF4-AS1', Aliases='', description='ARF4 antisense RNA 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144533, CurrentID=0, Status='live', Symbol='LINC00430', Aliases='', description='long intergenic non-protein coding RNA 430', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144534, CurrentID=0, Status='live', Symbol='GLTSCR1-AS1', Aliases='', description='GLTSCR1 antisense RNA 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144535, CurrentID=0, Status='live', Symbol='LINC01037', Aliases='', description='long intergenic non-protein coding RNA 1037', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144536, CurrentID=0, Status='live', Symbol='LINC01078', Aliases='', description='long intergenic non-protein coding RNA 1078', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100462724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100462724, CurrentID=0, Status='live', Symbol='OFC9', Aliases='', description='Orofacial cleft 9', other_designations='', map_location='13q33.1-q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610361, no_hdr0=''),
  106144538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144538, CurrentID=0, Status='live', Symbol='KCNK15-AS1', Aliases='', description='KCNK15 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144539, CurrentID=0, Status='live', Symbol='FOXD1-AS1', Aliases='', description='FOXD1 antisense RNA 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144540, CurrentID=0, Status='live', Symbol='LINC01326', Aliases='', description='long intergenic non-protein coding RNA 1326', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144541, CurrentID=0, Status='live', Symbol='LINC00407', Aliases='', description='long intergenic non-protein coding RNA 407', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144542, CurrentID=0, Status='live', Symbol='LINC01385', Aliases='', description='long intergenic non-protein coding RNA 1385', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144543, CurrentID=0, Status='live', Symbol='WISP1-OT1', Aliases='WISP1-UT1', description='WISP1 overlapping transcript 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144544, CurrentID=0, Status='live', Symbol='LINC01066', Aliases='', description='long intergenic non-protein coding RNA 1066', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419036, CurrentID=0, Status='live', Symbol='LOC100419036', Aliases='', description='tubulin, alpha 3e pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43231830, end_position_on_the_genomic_accession=43233173, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106799914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106799914, CurrentID=0, Status='live', Symbol='YUHAL', Aliases='', description='Yuan-Harel-Lupski syndrome', other_designations='', map_location='17p12-p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616652, no_hdr0=''),
  106144555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144555, CurrentID=0, Status='live', Symbol='ERVK-29', Aliases='', description='endogenous retrovirus group K member 29', other_designations='endogenous retrovirus group K, member 29', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144559, CurrentID=0, Status='live', Symbol='ERVW-8', Aliases='', description='endogenous retrovirus group W member 8', other_designations='endogenous retrovirus group W, member 8', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144560, CurrentID=0, Status='live', Symbol='ERVW-11', Aliases='', description='endogenous retrovirus group W member 11', other_designations='endogenous retrovirus group W, member 11', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144561, CurrentID=0, Status='live', Symbol='ERVW-14', Aliases='', description='endogenous retrovirus group W member 14', other_designations='endogenous retrovirus group W, member 14', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144562, CurrentID=0, Status='live', Symbol='ERVW-18', Aliases='', description='endogenous retrovirus group W member 18', other_designations='endogenous retrovirus group W, member 18', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144563, CurrentID=0, Status='live', Symbol='ERVW-22', Aliases='', description='endogenous retrovirus group W member 22', other_designations='endogenous retrovirus group W, member 22', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144564, CurrentID=0, Status='live', Symbol='ERVW-24', Aliases='', description='endogenous retrovirus group W member 24', other_designations='endogenous retrovirus group W, member 24', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144565, CurrentID=0, Status='live', Symbol='ERVW-27', Aliases='', description='endogenous retrovirus group W member 27', other_designations='endogenous retrovirus group W, member 27', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144566, CurrentID=0, Status='live', Symbol='ERVK-27', Aliases='', description='endogenous retrovirus group K member 27', other_designations='endogenous retrovirus group K, member 27', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144567, CurrentID=0, Status='live', Symbol='ERVW-17', Aliases='', description='endogenous retrovirus group W member 17', other_designations='endogenous retrovirus group W, member 17', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144568, CurrentID=0, Status='live', Symbol='LINC01381', Aliases='DADOT', description='long intergenic non-protein coding RNA 1381', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144571, CurrentID=0, Status='live', Symbol='ERVW-19', Aliases='', description='endogenous retrovirus group W member 19', other_designations='endogenous retrovirus group W, member 19', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144572, CurrentID=0, Status='live', Symbol='ERVW-26', Aliases='', description='endogenous retrovirus group W member 26', other_designations='endogenous retrovirus group W, member 26', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376187, CurrentID=0, Status='live', Symbol='LOC105376187', Aliases='', description='uncharacterized LOC105376187', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101822504, end_position_on_the_genomic_accession=101848699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106144575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144575, CurrentID=0, Status='live', Symbol='LINC01154', Aliases='', description='long intergenic non-protein coding RNA 1154', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144576, CurrentID=0, Status='live', Symbol='ERVW-13', Aliases='', description='endogenous retrovirus group W member 13', other_designations='endogenous retrovirus group W, member 13', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144578, CurrentID=0, Status='live', Symbol='LINC00604', Aliases='', description='long intergenic non-protein coding RNA 604', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144579, CurrentID=0, Status='live', Symbol='ERVW-10', Aliases='', description='endogenous retrovirus group W member 10', other_designations='endogenous retrovirus group W, member 10', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144580, CurrentID=0, Status='live', Symbol='ERVW-12', Aliases='', description='endogenous retrovirus group W member 12', other_designations='endogenous retrovirus group W, member 12', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144581, CurrentID=0, Status='live', Symbol='ERVW-16', Aliases='', description='endogenous retrovirus group W member 16', other_designations='endogenous retrovirus group W, member 16', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144583, CurrentID=0, Status='live', Symbol='ERVW-29', Aliases='', description='endogenous retrovirus group W member 29', other_designations='endogenous retrovirus group W, member 29', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144584, CurrentID=0, Status='live', Symbol='LINC01290', Aliases='', description='long intergenic non-protein coding RNA 1290', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144585, CurrentID=0, Status='live', Symbol='LINC01318', Aliases='', description='long intergenic non-protein coding RNA 1318', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144586, CurrentID=0, Status='live', Symbol='LINC01394', Aliases='', description='long intergenic non-protein coding RNA 1394', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144587, CurrentID=0, Status='live', Symbol='ERVW-21', Aliases='', description='endogenous retrovirus group W member 21', other_designations='endogenous retrovirus group W, member 21', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144589, CurrentID=0, Status='live', Symbol='ERVW-7', Aliases='', description='endogenous retrovirus group W member 7', other_designations='endogenous retrovirus group W, member 7', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144590, CurrentID=0, Status='live', Symbol='ERVK-31', Aliases='', description='endogenous retrovirus group K member 31', other_designations='endogenous retrovirus group K, member 31', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144591, CurrentID=0, Status='live', Symbol='ERVK-30', Aliases='', description='endogenous retrovirus group K member 30', other_designations='endogenous retrovirus group K, member 30', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144592, CurrentID=0, Status='live', Symbol='LINC01035', Aliases='', description='long intergenic non-protein coding RNA 1035', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144593, CurrentID=0, Status='live', Symbol='GLTSCR2-AS1', Aliases='', description='GLTSCR2 antisense RNA 1', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144595, CurrentID=0, Status='live', Symbol='LINC01374', Aliases='', description='long intergenic non-protein coding RNA 1374', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144597, CurrentID=0, Status='live', Symbol='ERVW-15', Aliases='', description='endogenous retrovirus group W member 15', other_designations='endogenous retrovirus group W, member 15', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144598, CurrentID=0, Status='live', Symbol='FARSA-AS1', Aliases='', description='FARSA antisense RNA 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144599, CurrentID=0, Status='live', Symbol='LINC01259', Aliases='', description='long intergenic non-protein coding RNA 1259', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144601, CurrentID=0, Status='live', Symbol='LINC01034', Aliases='', description='long intergenic non-protein coding RNA 1034', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144602, CurrentID=0, Status='live', Symbol='LINC01076', Aliases='', description='long intergenic non-protein coding RNA 1076', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144603, CurrentID=0, Status='live', Symbol='ERVW-23', Aliases='', description='endogenous retrovirus group W member 23', other_designations='endogenous retrovirus group W, member 23', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144604, CurrentID=0, Status='live', Symbol='ERVW-20', Aliases='', description='endogenous retrovirus group W member 20', other_designations='endogenous retrovirus group W, member 20', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144605, CurrentID=0, Status='live', Symbol='ERVW-25', Aliases='', description='endogenous retrovirus group W member 25', other_designations='endogenous retrovirus group W, member 25', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144606, CurrentID=0, Status='live', Symbol='CCND2-AS2', Aliases='', description='CCND2 antisense RNA 2', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144607, CurrentID=0, Status='live', Symbol='LINC01074', Aliases='', description='long intergenic non-protein coding RNA 1074', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106144608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144608, CurrentID=0, Status='live', Symbol='LINC00976', Aliases='', description='long intergenic non-protein coding RNA 976', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371824, CurrentID=0, Status='live', Symbol='LOC105371824', Aliases='', description='uncharacterized LOC105371824', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50537294, end_position_on_the_genomic_accession=50538197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129090, CurrentID=0, Status='live', Symbol='RPS5P4', Aliases='RPS5_1_162', description='ribosomal protein S5 pseudogene 4', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212823957, end_position_on_the_genomic_accession=212824507, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377940, CurrentID=0, Status='live', Symbol='LOC105377940', Aliases='', description='uncharacterized LOC105377940', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110530699, end_position_on_the_genomic_accession=110549402, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100287979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287979, CurrentID=0, Status='live', Symbol='TTTY13C', Aliases='', description='testis-specific transcript, Y-linked 13C', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=8869469, end_position_on_the_genomic_accession=8872853, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100506433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506433, CurrentID=0, Status='live', Symbol='LINC00648', Aliases='', description='long intergenic non-protein coding RNA 648', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=47764953, end_position_on_the_genomic_accession=47795014, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377943, CurrentID=0, Status='live', Symbol='LOC105377943', Aliases='', description='uncharacterized LOC105377943', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111276114, end_position_on_the_genomic_accession=111343553, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377944, CurrentID=0, Status='live', Symbol='LOC105377944', Aliases='', description='uncharacterized LOC105377944', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111642803, end_position_on_the_genomic_accession=111655549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377945, CurrentID=0, Status='live', Symbol='LOC105377945', Aliases='', description='uncharacterized LOC105377945', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111681398, end_position_on_the_genomic_accession=111694367, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375262, CurrentID=0, Status='live', Symbol='LOC105375262', Aliases='', description='uncharacterized LOC105375262', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44904771, end_position_on_the_genomic_accession=44908780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377948, CurrentID=0, Status='live', Symbol='LOC105377948', Aliases='', description='uncharacterized LOC105377948', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111996861, end_position_on_the_genomic_accession=112012253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419059, CurrentID=0, Status='live', Symbol='LOC100419059', Aliases='', description='tubulin, epsilon 1 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170896866, end_position_on_the_genomic_accession=170904656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103171574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103171574, CurrentID=0, Status='live', Symbol='LOC103171574', Aliases='', description='uncharacterized LOC103171574', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84422519, end_position_on_the_genomic_accession=84425882, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645163, CurrentID=0, Status='live', Symbol='LOC645163', Aliases='', description='phosphatidylinositol 4-kinase type 2 alpha pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53434545, end_position_on_the_genomic_accession=53437359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422384, CurrentID=0, Status='live', Symbol='POLR3DP1', Aliases='', description='polymerase (RNA) III (DNA directed) polypeptide D, 44kDa pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74654951, end_position_on_the_genomic_accession=74656327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  440031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=440031, CurrentID=0, Status='live', Symbol='LOC440031', Aliases='', description='uncharacterized LOC440031', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377958, CurrentID=0, Status='live', Symbol='LOC105377958', Aliases='', description='uncharacterized LOC105377958', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113871442, end_position_on_the_genomic_accession=113880812, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377961, CurrentID=0, Status='live', Symbol='LOC105377961', Aliases='', description='uncharacterized LOC105377961', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=115472502, end_position_on_the_genomic_accession=115486379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422463, CurrentID=0, Status='live', Symbol='LOC100422463', Aliases='', description='calcium channel, voltage-dependent, L type, alpha 1C subunit pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2742058, end_position_on_the_genomic_accession=2742806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377963, CurrentID=0, Status='live', Symbol='LOC105377963', Aliases='', description='uncharacterized LOC105377963', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116766354, end_position_on_the_genomic_accession=116768033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288001, CurrentID=0, Status='live', Symbol='LOC100288001', Aliases='', description='pre-mRNA processing factor 3 pseudogene', other_designations='PRP3 pre-mRNA processing factor 3 homolog pseudogene', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66983922, end_position_on_the_genomic_accession=66986508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377965, CurrentID=0, Status='live', Symbol='LOC105377965', Aliases='', description='uncharacterized LOC105377965', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117526822, end_position_on_the_genomic_accession=117537178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376194, CurrentID=0, Status='live', Symbol='LOC105376194', Aliases='', description='uncharacterized LOC105376194', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104270224, end_position_on_the_genomic_accession=104281266, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100508689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100508689, CurrentID=0, Status='live', Symbol='LOC100508689', Aliases='', description='mucin', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371830, CurrentID=0, Status='live', Symbol='LOC105371830', Aliases='', description='uncharacterized LOC105371830', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=52532531, end_position_on_the_genomic_accession=52535722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  653895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653895, CurrentID=0, Status='live', Symbol='PGGT1BP1', Aliases='', description='protein geranylgeranyltransferase type I, beta subunit pseudogene 1', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50726945, end_position_on_the_genomic_accession=50727969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419081, CurrentID=0, Status='live', Symbol='DUXAP12', Aliases='', description='double homeobox A pseudogene 12', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70593911, end_position_on_the_genomic_accession=70594428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371831, CurrentID=0, Status='live', Symbol='LOC105371831', Aliases='', description='uncharacterized LOC105371831', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=52862364, end_position_on_the_genomic_accession=52899583, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100419085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419085, CurrentID=0, Status='live', Symbol='LOC100419085', Aliases='', description='spermidine synthase pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128139535, end_position_on_the_genomic_accession=128140661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480997, CurrentID=0, Status='live', Symbol='RN7SL275P', Aliases='', description='RNA, 7SL, cytoplasmic 275, pseudogene', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377980, CurrentID=0, Status='live', Symbol='LOC105377980', Aliases='', description='uncharacterized LOC105377980', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123105910, end_position_on_the_genomic_accession=123107312, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130487, CurrentID=0, Status='live', Symbol='LOC100130487', Aliases='', description='WD repeat domain 89 pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13151635, end_position_on_the_genomic_accession=13152688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130044, CurrentID=0, Status='live', Symbol='LOC100130044', Aliases='', description='cilia and flagella associated protein 69 pseudogene', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57013054, end_position_on_the_genomic_accession=57014113, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422467, CurrentID=0, Status='live', Symbol='NUDT21P1', Aliases='', description='nudix hydrolase 21 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 21 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87765607, end_position_on_the_genomic_accession=87766196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419098, CurrentID=0, Status='live', Symbol='ENO1P2', Aliases='', description='enolase 1, (alpha) pseudogene 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92422134, end_position_on_the_genomic_accession=92422787, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928228, CurrentID=0, Status='live', Symbol='LOC101928228', Aliases='', description='uncharacterized LOC101928228', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=5719171, end_position_on_the_genomic_accession=5726143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  550643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=550643, CurrentID=0, Status='live', Symbol='LINC01420', Aliases='', description='long intergenic non-protein coding RNA 1420', other_designations='', map_location='Xp11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56729285, end_position_on_the_genomic_accession=56817571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377992, CurrentID=0, Status='live', Symbol='LOC105377992', Aliases='', description='uncharacterized LOC105377992', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126705140, end_position_on_the_genomic_accession=126706594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377994, CurrentID=0, Status='live', Symbol='LOC105377994', Aliases='', description='uncharacterized LOC105377994', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127296987, end_position_on_the_genomic_accession=127341875, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372940, CurrentID=0, Status='live', Symbol='LOC105372940', Aliases='', description='uncharacterized LOC105372940', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20460650, end_position_on_the_genomic_accession=20475576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  400759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400759, CurrentID=0, Status='live', Symbol='GBP1P1', Aliases='', description='guanylate binding protein 1, interferon-inducible pseudogene 1', other_designations='interferon-induced guanylate-binding protein 1 pseudogene', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89407679, end_position_on_the_genomic_accession=89424934, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100419110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419110, CurrentID=0, Status='live', Symbol='LOC100419110', Aliases='', description='ribosomal protein L4 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197029248, end_position_on_the_genomic_accession=197030092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422471, CurrentID=0, Status='live', Symbol='LOC100422471', Aliases='', description='solute carrier family 9, subfamily A (NHE2, cation proton antiporter 2), member 2 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=81408939, end_position_on_the_genomic_accession=81413647, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378004, CurrentID=0, Status='live', Symbol='LOC105378004', Aliases='', description='uncharacterized LOC105378004', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131063420, end_position_on_the_genomic_accession=131074841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378005, CurrentID=0, Status='live', Symbol='LOC105378005', Aliases='', description='uncharacterized LOC105378005', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131439987, end_position_on_the_genomic_accession=131443792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130018, CurrentID=0, Status='live', Symbol='LOC100130018', Aliases='', description='prostaglandin reductase 1 pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147319082, end_position_on_the_genomic_accession=147320377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370027, CurrentID=0, Status='live', Symbol='LOC105370027', Aliases='', description='uncharacterized LOC105370027', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119387991, end_position_on_the_genomic_accession=119668006, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378008, CurrentID=0, Status='live', Symbol='LOC105378008', Aliases='', description='uncharacterized LOC105378008', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132936567, end_position_on_the_genomic_accession=132941103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378009, CurrentID=0, Status='live', Symbol='LOC105378009', Aliases='', description='uncharacterized LOC105378009', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134335846, end_position_on_the_genomic_accession=134351644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378014, CurrentID=0, Status='live', Symbol='LOC105378014', Aliases='', description='uncharacterized LOC105378014', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136294986, end_position_on_the_genomic_accession=136306064, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928233, CurrentID=0, Status='live', Symbol='LOC101928233', Aliases='', description='uncharacterized LOC101928233', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42599498, end_position_on_the_genomic_accession=42615095, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130051, CurrentID=0, Status='live', Symbol='LOC100130051', Aliases='', description='uncharacterized LOC100130051', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=954681, end_position_on_the_genomic_accession=956451, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101929935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929935, CurrentID=0, Status='live', Symbol='LOC101929935', Aliases='', description='uncharacterized LOC101929935', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56627762, end_position_on_the_genomic_accession=56645313, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928234, CurrentID=0, Status='live', Symbol='LOC101928234', Aliases='', description='uncharacterized LOC101928234', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38156046, end_position_on_the_genomic_accession=38165478, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646663, CurrentID=0, Status='live', Symbol='RFPL4AP1', Aliases='', description='ret finger protein-like 4A pseudogene 1', other_designations='ret finger protein-like 1 pseudogene', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55777564, end_position_on_the_genomic_accession=55778957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371841, CurrentID=0, Status='live', Symbol='LOC105371841', Aliases='', description='uncharacterized LOC105371841', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58219980, end_position_on_the_genomic_accession=58233553, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378023, CurrentID=0, Status='live', Symbol='LOC105378023', Aliases='', description='uncharacterized LOC105378023', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139525504, end_position_on_the_genomic_accession=139527587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378024, CurrentID=0, Status='live', Symbol='LOC105378024', Aliases='', description='uncharacterized LOC105378024', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139546271, end_position_on_the_genomic_accession=139547690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378029, CurrentID=0, Status='live', Symbol='LOC105378029', Aliases='', description='uncharacterized LOC105378029', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=141403268, end_position_on_the_genomic_accession=141451006, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100509072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509072, CurrentID=0, Status='live', Symbol='LOC100509072', Aliases='', description='uncharacterized LOC100509072', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288069, CurrentID=0, Status='live', Symbol='LOC100288069', Aliases='', description='uncharacterized LOC100288069', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=764865, end_position_on_the_genomic_accession=778688, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378032, CurrentID=0, Status='live', Symbol='LOC105378032', Aliases='', description='uncharacterized LOC105378032', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142242314, end_position_on_the_genomic_accession=142259717, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100509091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509091, CurrentID=0, Status='live', Symbol='LOC100509091', Aliases='', description='extracellular matrix protein 2-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=152744793, end_position_on_the_genomic_accession=152752546, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105378034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378034, CurrentID=0, Status='live', Symbol='LOC105378034', Aliases='', description='uncharacterized LOC105378034', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143505416, end_position_on_the_genomic_accession=143506213, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928237, CurrentID=0, Status='live', Symbol='LOC101928237', Aliases='', description='uncharacterized LOC101928237', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4560368, end_position_on_the_genomic_accession=4583317, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100528061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528061, CurrentID=0, Status='live', Symbol='LOC100528061', Aliases='', description='gamma-glutamylcyclotransferase pseudogene', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95833955, end_position_on_the_genomic_accession=95834568, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288077, CurrentID=0, Status='live', Symbol='WTAPP1', Aliases='', description='Wilms tumor 1 associated protein pseudogene 1', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102783676, end_position_on_the_genomic_accession=102836766, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928238, CurrentID=0, Status='live', Symbol='LOC101928238', Aliases='', description='uncharacterized LOC101928238', other_designations='CTC-543D15.8', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9538739, end_position_on_the_genomic_accession=9539728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373520, CurrentID=0, Status='live', Symbol='LOC105373520', Aliases='', description='uncharacterized LOC105373520', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=103812111, end_position_on_the_genomic_accession=103820168, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100288079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288079, CurrentID=0, Status='live', Symbol='GS1-279B7.1', Aliases='', description='microtubule associated protein 1 light chain 3 beta pseudogene', other_designations='microtubule-associated protein 1 light chain 3 beta pseudogene', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185323846, end_position_on_the_genomic_accession=185335039, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371845, CurrentID=0, Status='live', Symbol='LOC105371845', Aliases='', description='uncharacterized LOC105371845', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59500392, end_position_on_the_genomic_accession=59506857, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130057, CurrentID=0, Status='live', Symbol='LOC100130057', Aliases='', description='uncharacterized LOC100130057', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=49146773, end_position_on_the_genomic_accession=49147605, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371846, CurrentID=0, Status='live', Symbol='LOC105371846', Aliases='', description='uncharacterized LOC105371846', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59425037, end_position_on_the_genomic_accession=59499435, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378048, CurrentID=0, Status='live', Symbol='LOC105378048', Aliases='', description='uncharacterized LOC105378048', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148926277, end_position_on_the_genomic_accession=148937730, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928309, CurrentID=0, Status='live', Symbol='LOC101928309', Aliases='', description='uncharacterized LOC101928309', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12582768, end_position_on_the_genomic_accession=12585590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102723401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723401, CurrentID=0, Status='live', Symbol='LOC102723401', Aliases='', description='uncharacterized LOC102723401', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23616499, end_position_on_the_genomic_accession=23618144, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378055, CurrentID=0, Status='live', Symbol='LOC105378055', Aliases='', description='uncharacterized LOC105378055', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150336271, end_position_on_the_genomic_accession=150345620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928241, CurrentID=0, Status='live', Symbol='LOC101928241', Aliases='', description='uncharacterized LOC101928241', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96254069, end_position_on_the_genomic_accession=96374125, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378056, CurrentID=0, Status='live', Symbol='LOC105378056', Aliases='', description='uncharacterized LOC105378056', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150600272, end_position_on_the_genomic_accession=150601061, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378058, CurrentID=0, Status='live', Symbol='LOC105378058', Aliases='', description='uncharacterized LOC105378058', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151626575, end_position_on_the_genomic_accession=151662953, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  285407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285407, CurrentID=0, Status='live', Symbol='ALG1L9P', Aliases='', description='asparagine-linked glycosylation 1-like 9, pseudogene', other_designations='asparagine-linked glycosylation 1 homolog pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71794363, end_position_on_the_genomic_accession=71813859, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928242, CurrentID=0, Status='live', Symbol='LOC101928242', Aliases='', description='cell cycle control protein 50B-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118769312, end_position_on_the_genomic_accession=118770201, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104310350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104310350, CurrentID=0, Status='live', Symbol='LINC01327', Aliases='TCONS_00005644', description='long intergenic non-protein coding RNA 1327', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167392796, end_position_on_the_genomic_accession=167408519, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104310351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104310351, CurrentID=0, Status='live', Symbol='LINC01331', Aliases='', description='long intergenic non-protein coding RNA 1331', other_designations='CTC-419K13.1', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74369376, end_position_on_the_genomic_accession=74536976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104310353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104310353, CurrentID=0, Status='live', Symbol='LINC01399', Aliases='', description='long intergenic non-protein coding RNA 1399', other_designations='CTA-714B7.5', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35119824, end_position_on_the_genomic_accession=35231056, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378063, CurrentID=0, Status='live', Symbol='LOC105378063', Aliases='', description='uncharacterized LOC105378063', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152838747, end_position_on_the_genomic_accession=152867545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378065, CurrentID=0, Status='live', Symbol='LOC105378065', Aliases='', description='uncharacterized LOC105378065', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152806226, end_position_on_the_genomic_accession=152875814, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102800316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800316, CurrentID=0, Status='live', Symbol='LINC01222', Aliases='TCONS_00000112', description='long intergenic non-protein coding RNA 1222', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198992589, end_position_on_the_genomic_accession=199018964, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378067, CurrentID=0, Status='live', Symbol='LOC105378067', Aliases='', description='uncharacterized LOC105378067', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154360787, end_position_on_the_genomic_accession=154362977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100509303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509303, CurrentID=0, Status='live', Symbol='LOC100509303', Aliases='', description='uncharacterized LOC100509303', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209147267, end_position_on_the_genomic_accession=209154609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378068, CurrentID=0, Status='live', Symbol='LOC105378068', Aliases='', description='uncharacterized LOC105378068', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155514862, end_position_on_the_genomic_accession=155526018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102606462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102606462, CurrentID=0, Status='live', Symbol='PRELID2P1', Aliases='', description='PRELI domain containing 2 pseudogene 1', other_designations='PRELI domain containing 2 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68956516, end_position_on_the_genomic_accession=68957680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102606463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102606463, CurrentID=0, Status='live', Symbol='LINC01152', Aliases='CMPD, TCONS_00025128', description='long intergenic non-protein coding RNA 1152', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72030654, end_position_on_the_genomic_accession=72041297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102606465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102606465, CurrentID=0, Status='live', Symbol='LOC102606465', Aliases='', description='uncharacterized LOC102606465', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101025846, end_position_on_the_genomic_accession=101087379, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102606466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102606466, CurrentID=0, Status='live', Symbol='LOC102606466', Aliases='', description='uncharacterized LOC102606466', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12934877, end_position_on_the_genomic_accession=12936963, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378070, CurrentID=0, Status='live', Symbol='LOC105378070', Aliases='', description='uncharacterized LOC105378070', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156352747, end_position_on_the_genomic_accession=156371021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378071, CurrentID=0, Status='live', Symbol='LOC105378071', Aliases='', description='uncharacterized LOC105378071', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156310941, end_position_on_the_genomic_accession=156352705, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646665, CurrentID=0, Status='live', Symbol='LOC646665', Aliases='', description='golgin A6 family member A pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 6 pseudogene|golgin A6 family, member A pseudogene', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72605178, end_position_on_the_genomic_accession=72612324, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378072, CurrentID=0, Status='live', Symbol='LOC105378072', Aliases='', description='uncharacterized LOC105378072', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=155793341, end_position_on_the_genomic_accession=155915268, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100506563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506563, CurrentID=0, Status='live', Symbol='LOC100506563', Aliases='', description='uncharacterized LOC100506563', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109758799, end_position_on_the_genomic_accession=109760383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378076, CurrentID=0, Status='live', Symbol='LOC105378076', Aliases='', description='uncharacterized LOC105378076', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157326565, end_position_on_the_genomic_accession=157332523, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389405, CurrentID=0, Status='live', Symbol='LOC389405', Aliases='', description='notch 5-like', other_designations='EGF-like-domain, multiple 10', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=64529823, end_position_on_the_genomic_accession=64939784, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100130063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130063, CurrentID=0, Status='live', Symbol='LOC100130063', Aliases='', description='C-x(9)-C motif containing 2 pseudogene', other_designations='COX assembly mitochondrial protein 2 homolog pseudogene', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7288261, end_position_on_the_genomic_accession=7288477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378078, CurrentID=0, Status='live', Symbol='LOC105378078', Aliases='', description='uncharacterized LOC105378078', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157885114, end_position_on_the_genomic_accession=157893132, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506569, CurrentID=0, Status='live', Symbol='LOC100506569', Aliases='', description='serine/arginine-rich splicing factor 3 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18664930, end_position_on_the_genomic_accession=18666484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100509370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509370, CurrentID=0, Status='live', Symbol='LOC100509370', Aliases='', description='39S ribosomal protein L21, mitochondrial pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68142579, end_position_on_the_genomic_accession=68143695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645188, CurrentID=0, Status='live', Symbol='LOC645188', Aliases='', description='uncharacterized LOC645188', other_designations='', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=141625868, end_position_on_the_genomic_accession=141626733, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422385, CurrentID=0, Status='live', Symbol='LOC100422385', Aliases='', description='DIM1 dimethyladenosine transferase 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=109701490, end_position_on_the_genomic_accession=109702295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421292, CurrentID=0, Status='live', Symbol='LOC100421292', Aliases='', description='cAMP responsive element binding protein 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95616574, end_position_on_the_genomic_accession=95617130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481446, CurrentID=0, Status='live', Symbol='RNU6-839P', Aliases='', description='RNA, U6 small nuclear 839, pseudogene', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928247, CurrentID=0, Status='live', Symbol='LOC101928247', Aliases='', description='uncharacterized LOC101928247', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177782255, end_position_on_the_genomic_accession=177794396, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100288125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288125, CurrentID=0, Status='live', Symbol='LOC100288125', Aliases='', description='NME/NM23 nucleoside diphosphate kinase 2 pseudogene', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98890429, end_position_on_the_genomic_accession=98892076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371854, CurrentID=0, Status='live', Symbol='LOC105371854', Aliases='', description='uncharacterized LOC105371854', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62375075, end_position_on_the_genomic_accession=62380502, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378088, CurrentID=0, Status='live', Symbol='LOC105378088', Aliases='', description='uncharacterized LOC105378088', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160263288, end_position_on_the_genomic_accession=160276240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288129, CurrentID=0, Status='live', Symbol='LOC100288129', Aliases='', description='acireductone dioxygenase 1 pseudogene', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47463087, end_position_on_the_genomic_accession=47463621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378093, CurrentID=0, Status='live', Symbol='LOC105378093', Aliases='', description='uncharacterized LOC105378093', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160899973, end_position_on_the_genomic_accession=160926209, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  387684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387684, CurrentID=0, Status='live', Symbol='ARL4AP1', Aliases='ARL4B, ARL4P', description='ADP ribosylation factor like GTPase 4A pseudogene 1', other_designations='ADP-ribosylation factor like GTPase 4A pseudogene 1|ADP-ribosylation factor-like 4 pseudogene|ADP-ribosylation factor-like 4A pseudogene 1|ADP-ribosylation factor-like 4B', map_location='10q21.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60684231, end_position_on_the_genomic_accession=60685396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374395, CurrentID=0, Status='live', Symbol='LOC105374395', Aliases='', description='uncharacterized LOC105374395', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=35022232, end_position_on_the_genomic_accession=35028289, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378094, CurrentID=0, Status='live', Symbol='LOC105378094', Aliases='', description='uncharacterized LOC105378094', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=161336250, end_position_on_the_genomic_accession=161338180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100509780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509780, CurrentID=0, Status='live', Symbol='LOC100509780', Aliases='', description='uncharacterized LOC100509780', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=55761428, end_position_on_the_genomic_accession=55761784, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378096, CurrentID=0, Status='live', Symbol='LOC105378096', Aliases='', description='uncharacterized LOC105378096', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162900285, end_position_on_the_genomic_accession=162901279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646666, CurrentID=0, Status='live', Symbol='LOC646666', Aliases='', description='apolipoprotein A-I binding protein pseudogene', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18040331, end_position_on_the_genomic_accession=18041144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375268, CurrentID=0, Status='live', Symbol='LOC105375268', Aliases='', description='uncharacterized LOC105375268', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47000620, end_position_on_the_genomic_accession=47079809, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378097, CurrentID=0, Status='live', Symbol='LOC105378097', Aliases='', description='uncharacterized LOC105378097', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162234816, end_position_on_the_genomic_accession=162239913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371856, CurrentID=0, Status='live', Symbol='LOC105371856', Aliases='', description='uncharacterized LOC105371856', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63380158, end_position_on_the_genomic_accession=63415011, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378100, CurrentID=0, Status='live', Symbol='LOC105378100', Aliases='', description='uncharacterized LOC105378100', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163387586, end_position_on_the_genomic_accession=163393639, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100113375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113375, CurrentID=0, Status='live', Symbol='LOC100113375', Aliases='', description='tRNA nucleotidyl transferase, CCA-adding, 1 pseudogene', other_designations='', map_location='22q', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45791983, end_position_on_the_genomic_accession=45793697, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422036, CurrentID=0, Status='live', Symbol='LOC100422036', Aliases='', description='olfactory receptor, family 6, subfamily C, member 2 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480445, CurrentID=0, Status='live', Symbol='MED28P2', Aliases='', description='mediator complex subunit 28 pseudogene 2', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371857, CurrentID=0, Status='live', Symbol='LOC105371857', Aliases='', description='uncharacterized LOC105371857', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63320175, end_position_on_the_genomic_accession=63354776, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378104, CurrentID=0, Status='live', Symbol='LOC105378104', Aliases='', description='uncharacterized LOC105378104', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163715106, end_position_on_the_genomic_accession=163720235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100506595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506595, CurrentID=0, Status='live', Symbol='HMGB1P40', Aliases='', description='high mobility group box 1 pseudogene 40', other_designations='high-mobility group box 1 pseudogene 40', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20574182, end_position_on_the_genomic_accession=20575363, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  149134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149134, CurrentID=0, Status='live', Symbol='LINC01341', Aliases='', description='long intergenic non-protein coding RNA 1341', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246789617, end_position_on_the_genomic_accession=246791486, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100506596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506596, CurrentID=0, Status='live', Symbol='LOC100506596', Aliases='', description='fibril-forming collagen alpha chain', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32178956, end_position_on_the_genomic_accession=32180069, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378106, CurrentID=0, Status='live', Symbol='LOC105378106', Aliases='', description='uncharacterized LOC105378106', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164083987, end_position_on_the_genomic_accession=164088302, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378107, CurrentID=0, Status='live', Symbol='LOC105378107', Aliases='', description='uncharacterized LOC105378107', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164184366, end_position_on_the_genomic_accession=164337805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  341152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=341152, CurrentID=0, Status='live', Symbol='OR2AT4', Aliases='OR11-265', description='olfactory receptor family 2 subfamily AT member 4', other_designations='olfactory receptor OR11-265 pseudogene|olfactory receptor, family 2, subfamily AT, member 4', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75088751, end_position_on_the_genomic_accession=75089713, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422492, CurrentID=0, Status='live', Symbol='LOC100422492', Aliases='', description='amyloid beta (A4) precursor protein-binding, family A, member 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34555870, end_position_on_the_genomic_accession=34556831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100506599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506599, CurrentID=0, Status='live', Symbol='PPP1R26-AS1', Aliases='', description='PPP1R26 antisense RNA 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135462718, end_position_on_the_genomic_accession=135472249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928877, CurrentID=0, Status='live', Symbol='LOC101928877', Aliases='', description='uncharacterized LOC101928877', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185051876, end_position_on_the_genomic_accession=185107262, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100506600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506600, CurrentID=0, Status='live', Symbol='MAST4-AS1', Aliases='', description='MAST4 antisense RNA 1', other_designations='MAST4 antisense RNA 1 (non-protein coding)', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=66297446, end_position_on_the_genomic_accession=66299781, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7974, CurrentID=0, Status='live', Symbol='HRX', Aliases='', description='Hyperreflexia', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145290, no_hdr0=''),
  105378110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378110, CurrentID=0, Status='live', Symbol='LOC105378110', Aliases='', description='uncharacterized LOC105378110', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164184277, end_position_on_the_genomic_accession=164190941, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127295, CurrentID=0, Status='live', Symbol='RPL36P5', Aliases='RPL36_1_21', description='ribosomal protein L36 pseudogene 5', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24007881, end_position_on_the_genomic_accession=24008519, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422493, CurrentID=0, Status='live', Symbol='LOC100422493', Aliases='', description='required for meiotic nuclear division 1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73068107, end_position_on_the_genomic_accession=73068618, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645195, CurrentID=0, Status='live', Symbol='LOC645195', Aliases='', description='uncharacterized LOC645195', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=65713847, end_position_on_the_genomic_accession=65721850, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928253, CurrentID=0, Status='live', Symbol='LOC101928253', Aliases='', description='uncharacterized LOC101928253', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11990343, end_position_on_the_genomic_accession=12001244, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480448, CurrentID=0, Status='live', Symbol='RNU7-104P', Aliases='', description='RNA, U7 small nuclear 104 pseudogene', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271432, CurrentID=0, Status='live', Symbol='RPL21P102', Aliases='RPL21_43_1219', description='ribosomal protein L21 pseudogene 102', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24642653, end_position_on_the_genomic_accession=24643198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371860, CurrentID=0, Status='live', Symbol='LOC105371860', Aliases='', description='uncharacterized LOC105371860', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64844396, end_position_on_the_genomic_accession=64847381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929939, CurrentID=0, Status='live', Symbol='LOC101929939', Aliases='', description='uncharacterized LOC101929939', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104613533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104613533, CurrentID=0, Status='live', Symbol='LOC104613533', Aliases='', description='uncharacterized LOC104613533', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92470233, end_position_on_the_genomic_accession=92471546, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100509620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509620, CurrentID=0, Status='live', Symbol='LOC100509620', Aliases='', description='putative aquaporin-7-like protein 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94587234, end_position_on_the_genomic_accession=94619833, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  106480449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480449, CurrentID=0, Status='live', Symbol='RNU7-121P', Aliases='', description='RNA, U7 small nuclear 121 pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288160, CurrentID=0, Status='live', Symbol='LOC100288160', Aliases='', description='esophagus cancer-related gene-2 interaction susceptibility protein', other_designations='ECRG2 interaction protein', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=101359010, end_position_on_the_genomic_accession=101359615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422495, CurrentID=0, Status='live', Symbol='LOC100422495', Aliases='', description='protein tyrosine phosphatase, non-receptor type 4 (megakaryocyte) pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6718395, end_position_on_the_genomic_accession=6721010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928255, CurrentID=0, Status='live', Symbol='LOC101928255', Aliases='', description='uncharacterized LOC101928255', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42648294, end_position_on_the_genomic_accession=42653442, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480450, CurrentID=0, Status='live', Symbol='RNU7-129P', Aliases='', description='RNA, U7 small nuclear 129 pseudogene', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371862, CurrentID=0, Status='live', Symbol='LOC105371862', Aliases='', description='uncharacterized LOC105371862', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64926998, end_position_on_the_genomic_accession=64930092, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378128, CurrentID=0, Status='live', Symbol='LOC105378128', Aliases='', description='uncharacterized LOC105378128', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167481678, end_position_on_the_genomic_accession=167489224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378130, CurrentID=0, Status='live', Symbol='LOC105378130', Aliases='', description='uncharacterized LOC105378130', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167565510, end_position_on_the_genomic_accession=167573780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480451, CurrentID=0, Status='live', Symbol='RNU7-147P', Aliases='', description='RNA, U7 small nuclear 147 pseudogene', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422497, CurrentID=0, Status='live', Symbol='LOC100422497', Aliases='', description='UTP18, small subunit (SSU) processome component, homolog (yeast) pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46028182, end_position_on_the_genomic_accession=46029832, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928257, CurrentID=0, Status='live', Symbol='LOC101928257', Aliases='', description='uncharacterized LOC101928257', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=52771390, end_position_on_the_genomic_accession=52774861, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389465, CurrentID=0, Status='live', Symbol='LOC389465', Aliases='', description='Sjogren syndrome antigen B pseudogene', other_designations='Sjogren syndrome antigen B (autoantigen La) pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12471342, end_position_on_the_genomic_accession=12472826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326590, CurrentID=0, Status='live', Symbol='SSXP7', Aliases='psiSSX7', description='SSX family pseudogene 7', other_designations='SSX7 pseudogene', map_location='Xp11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48441519, end_position_on_the_genomic_accession=48442977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288179, CurrentID=0, Status='live', Symbol='TSEN15P1', Aliases='', description='tRNA splicing endonuclease subunit 15 pseudogene 1', other_designations='TSEN15 tRNA splicing endonuclease subunit pseudogene 1|tRNA splicing endonuclease 15 homolog pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17456397, end_position_on_the_genomic_accession=17457423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480453, CurrentID=0, Status='live', Symbol='RNU7-170P', Aliases='', description='RNA, U7 small nuclear 170 pseudogene', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271433, CurrentID=0, Status='live', Symbol='RPL21P103', Aliases='RPL21_44_1241', description='ribosomal protein L21 pseudogene 103', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58244344, end_position_on_the_genomic_accession=58244904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371865, CurrentID=0, Status='live', Symbol='LOC105371865', Aliases='', description='uncharacterized LOC105371865', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378144, CurrentID=0, Status='live', Symbol='LOC105378144', Aliases='', description='uncharacterized LOC105378144', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168860822, end_position_on_the_genomic_accession=168862366, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378145, CurrentID=0, Status='live', Symbol='LOC105378145', Aliases='', description='uncharacterized LOC105378145', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168964400, end_position_on_the_genomic_accession=169063791, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928259, CurrentID=0, Status='live', Symbol='LOC101928259', Aliases='', description='uncharacterized LOC101928259', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72688949, end_position_on_the_genomic_accession=72712348, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100113421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100113421, CurrentID=0, Status='live', Symbol='LOC100113421', Aliases='', description='fibroblast growth factor 7 pseudogene', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62888535, end_position_on_the_genomic_accession=62892746, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480454, CurrentID=0, Status='live', Symbol='RNU7-183P', Aliases='', description='RNA, U7 small nuclear 183 pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419257, CurrentID=0, Status='live', Symbol='LOC100419257', Aliases='', description='ribosomal protein S2 pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32986931, end_position_on_the_genomic_accession=32988247, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377016, CurrentID=0, Status='live', Symbol='LOC105377016', Aliases='', description='uncharacterized LOC105377016', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30653482, end_position_on_the_genomic_accession=30663487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480455, CurrentID=0, Status='live', Symbol='RNU7-197P', Aliases='', description='RNA, U7 small nuclear 197 pseudogene', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100509814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509814, CurrentID=0, Status='live', Symbol='LOC100509814', Aliases='', description='uncharacterized LOC100509814', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378156, CurrentID=0, Status='live', Symbol='LOC105378156', Aliases='', description='uncharacterized LOC105378156', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170358326, end_position_on_the_genomic_accession=170388280, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480456, CurrentID=0, Status='live', Symbol='RN7SKP15', Aliases='', description='RNA, 7SK small nuclear pseudogene 15', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378159, CurrentID=0, Status='live', Symbol='LOC105378159', Aliases='', description='uncharacterized LOC105378159', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130287921, end_position_on_the_genomic_accession=130383090, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100509857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509857, CurrentID=0, Status='live', Symbol='KRT8P4', Aliases='', description='keratin 8 pseudogene 4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=89412513, end_position_on_the_genomic_accession=89415071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131770, CurrentID=0, Status='live', Symbol='LOC100131770', Aliases='', description='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32kDa pseudogene', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66481027, end_position_on_the_genomic_accession=66481486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480457, CurrentID=0, Status='live', Symbol='RN7SKP29', Aliases='', description='RNA, 7SK small nuclear pseudogene 29', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644269, CurrentID=0, Status='live', Symbol='LOC644269', Aliases='', description='G protein pathway suppressor 1 pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89886059, end_position_on_the_genomic_accession=89887843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378165, CurrentID=0, Status='live', Symbol='LOC105378165', Aliases='', description='uncharacterized LOC105378165', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928263, CurrentID=0, Status='live', Symbol='LOC101928263', Aliases='', description='uncharacterized LOC101928263', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39150903, end_position_on_the_genomic_accession=39173071, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100509894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509894, CurrentID=0, Status='live', Symbol='CPB2-AS1', Aliases='', description='CPB2 antisense RNA 1', other_designations='CPB2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46052848, end_position_on_the_genomic_accession=46101347, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480458, CurrentID=0, Status='live', Symbol='RN7SKP36', Aliases='', description='RNA, 7SK small nuclear pseudogene 36', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371870, CurrentID=0, Status='live', Symbol='LOC105371870', Aliases='', description='uncharacterized LOC105371870', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68166437, end_position_on_the_genomic_accession=68186866, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376235, CurrentID=0, Status='live', Symbol='LOC105376235', Aliases='', description='uncharacterized LOC105376235', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115596658, end_position_on_the_genomic_accession=115672726, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101928264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928264, CurrentID=0, Status='live', Symbol='LOC101928264', Aliases='', description='uncharacterized LOC101928264', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159776472, end_position_on_the_genomic_accession=159780820, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100509927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100509927, CurrentID=0, Status='live', Symbol='HIST1H2APS6', Aliases='', description='histone cluster 1, H2a, pseudogene 6', other_designations='HIST1H2A pseudogene 6|histone cluster 1, H2a pseudogene', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36490050, end_position_on_the_genomic_accession=36490441, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378172, CurrentID=0, Status='live', Symbol='LOC105378172', Aliases='', description='uncharacterized LOC105378172', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19454758, end_position_on_the_genomic_accession=19481801, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106480459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480459, CurrentID=0, Status='live', Symbol='RN7SKP50', Aliases='', description='RNA, 7SK small nuclear pseudogene 50', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480460, CurrentID=0, Status='live', Symbol='RN7SKP62', Aliases='', description='RNA, 7SK small nuclear pseudogene 62', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874144, CurrentID=0, Status='live', Symbol='LINC00362', Aliases='', description='long intergenic non-protein coding RNA 362', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480461, CurrentID=0, Status='live', Symbol='RN7SKP69', Aliases='', description='RNA, 7SK small nuclear pseudogene 69', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371873, CurrentID=0, Status='live', Symbol='LOC105371873', Aliases='', description='uncharacterized LOC105371873', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68558393, end_position_on_the_genomic_accession=68581418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379636, CurrentID=0, Status='live', Symbol='LOC105379636', Aliases='', description='uncharacterized LOC105379636', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378186, CurrentID=0, Status='live', Symbol='LOC105378186', Aliases='', description='uncharacterized LOC105378186', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105648459, end_position_on_the_genomic_accession=105676881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480462, CurrentID=0, Status='live', Symbol='RN7SKP82', Aliases='', description='RNA, 7SK small nuclear pseudogene 82', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  439911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=439911, CurrentID=0, Status='live', Symbol='LOC439911', Aliases='', description='uncharacterized LOC439911', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288227, CurrentID=0, Status='live', Symbol='MTND1P27', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 1 pseudogene 27', other_designations='MT-ND1 pseudogene 27', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140220563, end_position_on_the_genomic_accession=140221473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928268, CurrentID=0, Status='live', Symbol='LOC101928268', Aliases='', description='uncharacterized LOC101928268', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30460051, end_position_on_the_genomic_accession=30469422, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378191, CurrentID=0, Status='live', Symbol='LOC105378191', Aliases='', description='uncharacterized LOC105378191', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105481769, end_position_on_the_genomic_accession=105486591, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480463, CurrentID=0, Status='live', Symbol='RN7SKP97', Aliases='', description='RNA, 7SK small nuclear pseudogene 97', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271435, CurrentID=0, Status='live', Symbol='RPL21P104', Aliases='RPL21_45_1243', description='ribosomal protein L21 pseudogene 104', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61775882, end_position_on_the_genomic_accession=61776431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106146149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106146149, CurrentID=0, Status='live', Symbol='LINC01613', Aliases='n341773', description='long intergenic non-protein coding RNA 1613', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106146153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106146153, CurrentID=0, Status='live', Symbol='LOC106146153', Aliases='lnc-RP11-211G3.3.1-1', description='uncharacterized lncRNA LOC106146153', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378196, CurrentID=0, Status='live', Symbol='LOC105378196', Aliases='', description='uncharacterized LOC105378196', other_designations='CTB-35F21.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139741655, end_position_on_the_genomic_accession=139745010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102800447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102800447, CurrentID=0, Status='live', Symbol='LOC102800447', Aliases='', description='uncharacterized LOC102800447', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67086446, end_position_on_the_genomic_accession=67289244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378197, CurrentID=0, Status='live', Symbol='LOC105378197', Aliases='', description='uncharacterized LOC105378197', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140149614, end_position_on_the_genomic_accession=140157785, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480464, CurrentID=0, Status='live', Symbol='RN7SKP107', Aliases='', description='RNA, 7SK small nuclear pseudogene 107', other_designations='', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100419307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419307, CurrentID=0, Status='live', Symbol='LOC100419307', Aliases='', description='ribosomal protein L5 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23125621, end_position_on_the_genomic_accession=23128601, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378201, CurrentID=0, Status='live', Symbol='LOC105378201', Aliases='', description='uncharacterized LOC105378201', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141705512, end_position_on_the_genomic_accession=141717470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480465, CurrentID=0, Status='live', Symbol='RN7SKP113', Aliases='', description='RNA, 7SK small nuclear pseudogene 113', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130088, CurrentID=0, Status='live', Symbol='LOC100130088', Aliases='', description='oligonucleotide/oligosaccharide-binding fold containing 1 pseudogene', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=155397937, end_position_on_the_genomic_accession=155398805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288241, CurrentID=0, Status='live', Symbol='LOC100288241', Aliases='', description='protein (peptidylprolyl cis/trans isomerase) NIMA-interacting, 4 (parvulin) pseudogene', other_designations='peptidylprolyl cis/trans isomerase, NIMA-interacting 4 pseudogene', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82010215, end_position_on_the_genomic_accession=82010637, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421574, CurrentID=0, Status='live', Symbol='LOC100421574', Aliases='', description='family with sequence similarity 47, member C pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69160753, end_position_on_the_genomic_accession=69161905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102997064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102997064, CurrentID=0, Status='live', Symbol='DEL5Q12', Aliases='C5DELq12', description='Chromosome 5q12 deletion sydrome', other_designations='', map_location='5q12', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615668, no_hdr0=''),
  105373313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373313, CurrentID=0, Status='live', Symbol='LOC105373313', Aliases='', description='uncharacterized LOC105373313', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114713926, end_position_on_the_genomic_accession=114810776, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100506697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506697, CurrentID=0, Status='live', Symbol='LINC00327', Aliases='NCRNA00327', description='long intergenic non-protein coding RNA 327', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23469512, end_position_on_the_genomic_accession=23487464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378207, CurrentID=0, Status='live', Symbol='LOC105378207', Aliases='', description='uncharacterized LOC105378207', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143462113, end_position_on_the_genomic_accession=143494009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480466, CurrentID=0, Status='live', Symbol='RN7SKP127', Aliases='', description='RNA, 7SK small nuclear pseudogene 127', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378209, CurrentID=0, Status='live', Symbol='LOC105378209', Aliases='', description='uncharacterized LOC105378209', other_designations='CTD-2050E21.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143561313, end_position_on_the_genomic_accession=143598434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378210, CurrentID=0, Status='live', Symbol='LOC105378210', Aliases='', description='uncharacterized LOC105378210', other_designations='CTB-57H20.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143662893, end_position_on_the_genomic_accession=143672726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=145, CurrentID=0, Status='live', Symbol='PARP1P2', Aliases='ADPRTP2, PPOLP2', description='poly(ADP-ribose) polymerase family member 1 pseudogene 2', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase) pseudogene 2|poly (ADP-ribose) polymerase family, member 1 pseudogene 2|poly(ADP-ribose) polymerase family, member 1 pseudogene 2', map_location='14q22', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63123001, end_position_on_the_genomic_accession=63123935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378211, CurrentID=0, Status='live', Symbol='LOC105378211', Aliases='', description='uncharacterized LOC105378211', other_designations='CTB-73N10.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145429604, end_position_on_the_genomic_accession=145480465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419321, CurrentID=0, Status='live', Symbol='LOC100419321', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125862650, end_position_on_the_genomic_accession=125863286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378212, CurrentID=0, Status='live', Symbol='LOC105378212', Aliases='', description='uncharacterized LOC105378212', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146098324, end_position_on_the_genomic_accession=146104568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480467, CurrentID=0, Status='live', Symbol='RN7SKP148', Aliases='', description='RNA, 7SK small nuclear pseudogene 148', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378213, CurrentID=0, Status='live', Symbol='LOC105378213', Aliases='', description='uncharacterized LOC105378213', other_designations='CTC-359M8.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146370385, end_position_on_the_genomic_accession=146376959, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372654, CurrentID=0, Status='live', Symbol='LOC105372654', Aliases='', description='uncharacterized LOC105372654', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50029091, end_position_on_the_genomic_accession=50030477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419324, CurrentID=0, Status='live', Symbol='BCL9P1', Aliases='', description='B-cell CLL/lymphoma 9 pseudogene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67636382, end_position_on_the_genomic_accession=67638942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379132, CurrentID=0, Status='live', Symbol='SNORA70G', Aliases='U70G', description='small nucleolar RNA, H/ACA box 70G', other_designations='small nucleolar RNA, H/ACA box 70G (retrotransposed)', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68627234, end_position_on_the_genomic_accession=68627375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100379141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379141, CurrentID=0, Status='live', Symbol='SNORA70D', Aliases='U70D', description='small nucleolar RNA, H/ACA box 70D', other_designations='small nucleolar RNA, H/ACA box 70D (retrotransposed)', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71698567, end_position_on_the_genomic_accession=71698701, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100379142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379142, CurrentID=0, Status='live', Symbol='NCLP2', Aliases='bA85L21.1', description='nucleolin pseudogene 2', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135075136, end_position_on_the_genomic_accession=135076597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376245, CurrentID=0, Status='live', Symbol='LOC105376245', Aliases='', description='uncharacterized LOC105376245', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117768394, end_position_on_the_genomic_accession=117773510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378221, CurrentID=0, Status='live', Symbol='LOC105378221', Aliases='', description='uncharacterized LOC105378221', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148550232, end_position_on_the_genomic_accession=148559987, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378222, CurrentID=0, Status='live', Symbol='LOC105378222', Aliases='', description='uncharacterized LOC105378222', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149060564, end_position_on_the_genomic_accession=149061324, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480469, CurrentID=0, Status='live', Symbol='RN7SKP172', Aliases='', description='RNA, 7SK small nuclear pseudogene 172', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100379174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379174, CurrentID=0, Status='live', Symbol='MACROD2-AS1', Aliases='BA318C17.1, NCRNA00186', description='MACROD2 antisense RNA 1', other_designations='MACROD2 antisense RNA 1 (non-protein coding)', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14884253, end_position_on_the_genomic_accession=14929518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100379198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379198, CurrentID=0, Status='live', Symbol='ECT', Aliases='BECTS', description='centralopathic epilepsy', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=117100, no_hdr0=''),
  100379199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379199, CurrentID=0, Status='live', Symbol='HRM2', Aliases='', description='hair, curly', other_designations='hair curvature, variation in|hair morphology 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=139450, no_hdr0=''),
  100379200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379200, CurrentID=0, Status='live', Symbol='SLEB15', Aliases='', description='Systemic lupus erthematosus, susceptibility to, 15', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300809, no_hdr0=''),
  100379201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379201, CurrentID=0, Status='live', Symbol='SPAX3', Aliases='ARSAL', description='Ataxia, spastic, 3, autosomal recessive', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611390, no_hdr0=''),
  100379202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379202, CurrentID=0, Status='live', Symbol='DUP5P13', Aliases='C5DUPp13', description='Chromosome 5p13 duplication syndrome', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613174, no_hdr0=''),
  100379203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379203, CurrentID=0, Status='live', Symbol='DUP17P13.3', Aliases='C17DUPp13.3', description='Chromosome 17p13.3 duplication syndrome', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613215, no_hdr0=''),
  105378230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378230, CurrentID=0, Status='live', Symbol='LOC105378230', Aliases='', description='uncharacterized LOC105378230', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151211768, end_position_on_the_genomic_accession=151223916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100379215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379215, CurrentID=0, Status='live', Symbol='QTV', Aliases='', description='QT interval, variation in', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610141, no_hdr0=''),
  100379216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379216, CurrentID=0, Status='live', Symbol='FGQTL2', Aliases='BWQTL1', description='Fasting plasma glucose level QTL2', other_designations='', map_location='7p15-p13', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613219, no_hdr0=''),
  100379217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379217, CurrentID=0, Status='live', Symbol='FGQTL3', Aliases='', description='Fasting plasma glucose level QTL3', other_designations='', map_location='11q21-q22', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613233, no_hdr0=''),
  100379218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379218, CurrentID=0, Status='live', Symbol='SPDA3', Aliases='', description='Spondylarthropathy, susceptibility to, 3', other_designations='', map_location='2q36.1-q36.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613238, no_hdr0=''),
  100379220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379220, CurrentID=0, Status='live', Symbol='TMED11P', Aliases='', description='transmembrane p24 trafficking protein 11, pseudogene', other_designations='transmembrane emp24 protein transport domain containing 11, pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1115197, end_position_on_the_genomic_accession=1123164, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371883, CurrentID=0, Status='live', Symbol='LOC105371883', Aliases='', description='uncharacterized LOC105371883', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70885135, end_position_on_the_genomic_accession=70900858, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100379224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379224, CurrentID=0, Status='live', Symbol='LOC100379224', Aliases='', description='uncharacterized LOC100379224', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44105339, end_position_on_the_genomic_accession=44113183, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283523, CurrentID=0, Status='live', Symbol='TERF1P5', Aliases='', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene 5', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18680495, end_position_on_the_genomic_accession=18683300, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928277, CurrentID=0, Status='live', Symbol='LOC101928277', Aliases='', description='uncharacterized LOC101928277', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134673918, end_position_on_the_genomic_accession=134904643, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100379249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379249, CurrentID=0, Status='live', Symbol='BOK-AS1', Aliases='BOK-AS, BOKAS, NAToB, NCRNA00151', description='BOK antisense RNA 1', other_designations='BOK antisense RNA (non-protein coding)|BOK antisense RNA 1 (non-protein coding)', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241544384, end_position_on_the_genomic_accession=241559143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100379250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379250, CurrentID=0, Status='live', Symbol='SNORA70E', Aliases='U70E', description='small nucleolar RNA, H/ACA box 70E', other_designations='small nucleolar RNA, H/ACA box 70E (retrotransposed)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83041464, end_position_on_the_genomic_accession=83041598, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100379251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379251, CurrentID=0, Status='live', Symbol='NPPA-AS1', Aliases='NPPA-AS, NPPAAS', description='NPPA antisense RNA 1', other_designations='NPPA antisense RNA (non-protein coding)|NPPA antisense RNA 1 (non-protein coding)', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11840319, end_position_on_the_genomic_accession=11847616, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378239, CurrentID=0, Status='live', Symbol='LOC105378239', Aliases='', description='uncharacterized LOC105378239', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=155115726, end_position_on_the_genomic_accession=155119567, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100506731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506731, CurrentID=0, Status='live', Symbol='LOC100506731', Aliases='', description='uncharacterized LOC100506731', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85524432, end_position_on_the_genomic_accession=85530319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378242, CurrentID=0, Status='live', Symbol='LOC105378242', Aliases='', description='zinc finger protein 705A-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371885, CurrentID=0, Status='live', Symbol='LOC105371885', Aliases='', description='uncharacterized LOC105371885', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=71045051, end_position_on_the_genomic_accession=71047895, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100379287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379287, CurrentID=0, Status='live', Symbol='RNY4P6', Aliases='', description='RNA, Ro-associated Y4 pseudogene 6', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117015897, end_position_on_the_genomic_accession=117015991, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379288, CurrentID=0, Status='live', Symbol='RNY4P11', Aliases='', description='RNA, Ro-associated Y4 pseudogene 11', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18329009, end_position_on_the_genomic_accession=18329102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378244, CurrentID=0, Status='live', Symbol='LOC105378244', Aliases='', description='uncharacterized LOC105378244', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100379290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379290, CurrentID=0, Status='live', Symbol='LOC100379290', Aliases='', description='transmembrane and coiled-coil domain family 1 pseudogene', other_designations='', map_location='2p13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68627061, end_position_on_the_genomic_accession=68628155, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379291, CurrentID=0, Status='live', Symbol='RNY4P15', Aliases='', description='RNA, Ro-associated Y4 pseudogene 15', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88229572, end_position_on_the_genomic_accession=88229664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379295, CurrentID=0, Status='live', Symbol='RNY4P8', Aliases='', description='RNA, Ro-associated Y4 pseudogene 8', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40243223, end_position_on_the_genomic_accession=40243317, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379296, CurrentID=0, Status='live', Symbol='RNY4P13', Aliases='', description='RNA, Ro-associated Y4 pseudogene 13', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30059054, end_position_on_the_genomic_accession=30059147, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379297, CurrentID=0, Status='live', Symbol='RNY4P4', Aliases='', description='RNA, Ro-associated Y4 pseudogene 4', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136588209, end_position_on_the_genomic_accession=136588311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379298, CurrentID=0, Status='live', Symbol='RNY4P18', Aliases='', description='RNA, Ro-associated Y4 pseudogene 18', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111097327, end_position_on_the_genomic_accession=111097413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379299, CurrentID=0, Status='live', Symbol='RNY4P5', Aliases='', description='RNA, Ro-associated Y4 pseudogene 5', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123044717, end_position_on_the_genomic_accession=123044811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379300, CurrentID=0, Status='live', Symbol='RNY4P9', Aliases='', description='RNA, Ro-associated Y4 pseudogene 9', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49908637, end_position_on_the_genomic_accession=49908730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379323, CurrentID=0, Status='live', Symbol='ERVW-2', Aliases='ERVWE2, NTrenv', description='endogenous retrovirus group W member 2', other_designations='endogenous retroviral family W, env, member 2|endogenous retrovirus group W, member 2', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378251, CurrentID=0, Status='live', Symbol='LOC105378251', Aliases='', description='uncharacterized LOC105378251', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68185586, end_position_on_the_genomic_accession=68196507, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480475, CurrentID=0, Status='live', Symbol='RN7SKP233', Aliases='', description='RNA, 7SK small nuclear pseudogene 233', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371887, CurrentID=0, Status='live', Symbol='LOC105371887', Aliases='', description='uncharacterized LOC105371887', other_designations='CTD-3010D24.3', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72640675, end_position_on_the_genomic_accession=72644729, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100379345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379345, CurrentID=0, Status='live', Symbol='MIR181A2HG', Aliases='', description='MIR181A2 host gene', other_designations='MIR181A2 host gene (non-protein coding)', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124658436, end_position_on_the_genomic_accession=124698628, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102723521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723521, CurrentID=0, Status='live', Symbol='LOC102723521', Aliases='', description='uncharacterized LOC102723521', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34273202, end_position_on_the_genomic_accession=34282446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506747, CurrentID=0, Status='live', Symbol='LOC100506747', Aliases='', description='alpha-1,3-mannosyl-glycoprotein 4-beta-N-acetylglucosaminyltransferase-like protein LOC641515 homolog', other_designations='4933406M09Rik pseudogene', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202986452, end_position_on_the_genomic_accession=203007265, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371888, CurrentID=0, Status='live', Symbol='LOC105371888', Aliases='', description='uncharacterized LOC105371888', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73941852, end_position_on_the_genomic_accession=73950860, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100379378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379378, CurrentID=0, Status='live', Symbol='RNY4P20', Aliases='', description='RNA, Ro-associated Y4 pseudogene 20', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151298842, end_position_on_the_genomic_accession=151298933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130100, CurrentID=0, Status='live', Symbol='LOC100130100', Aliases='', description='Ig kappa chain V-I region Walker-like', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113386972, end_position_on_the_genomic_accession=113414730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371889, CurrentID=0, Status='live', Symbol='LOC105371889', Aliases='', description='uncharacterized LOC105371889', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73958733, end_position_on_the_genomic_accession=73979508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132647, CurrentID=0, Status='live', Symbol='LOC100132647', Aliases='', description='signal peptide peptidase like 3 pseudogene', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71225697, end_position_on_the_genomic_accession=71226718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928283, CurrentID=0, Status='live', Symbol='LOC101928283', Aliases='', description='uncharacterized LOC101928283', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125184569, end_position_on_the_genomic_accession=125379321, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100271438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271438, CurrentID=0, Status='live', Symbol='RPL21P110', Aliases='RPL21_50_1323', description='ribosomal protein L21 pseudogene 110', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=72703653, end_position_on_the_genomic_accession=72704170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422524, CurrentID=0, Status='live', Symbol='LOC100422524', Aliases='', description='solute carrier family 4, sodium bicarbonate cotransporter, member 8 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21580133, end_position_on_the_genomic_accession=21580670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378271, CurrentID=0, Status='live', Symbol='LOC105378271', Aliases='', description='uncharacterized LOC105378271', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43318488, end_position_on_the_genomic_accession=43350340, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378272, CurrentID=0, Status='live', Symbol='LOC105378272', Aliases='', description='uncharacterized LOC105378272', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43420738, end_position_on_the_genomic_accession=43422148, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130102, CurrentID=0, Status='live', Symbol='LOC100130102', Aliases='', description='WW domain binding protein 11 pseudogene', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127247703, end_position_on_the_genomic_accession=127250284, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  389602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389602, CurrentID=0, Status='live', Symbol='LOC389602', Aliases='', description='uncharacterized LOC389602', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155962632, end_position_on_the_genomic_accession=155966343, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378275, CurrentID=0, Status='live', Symbol='LOC105378275', Aliases='', description='uncharacterized LOC105378275', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43778961, end_position_on_the_genomic_accession=43818803, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378276, CurrentID=0, Status='live', Symbol='LOC105378276', Aliases='', description='uncharacterized LOC105378276', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43912405, end_position_on_the_genomic_accession=43914496, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288315, CurrentID=0, Status='live', Symbol='LOC100288315', Aliases='', description='uncharacterized LOC100288315', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=28357497, end_position_on_the_genomic_accession=28358955, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100288319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288319, CurrentID=0, Status='live', Symbol='AK3P5', Aliases='', description='adenylate kinase 3 pseudogene 5', other_designations='', map_location='10p11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32944417, end_position_on_the_genomic_accession=32945001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378282, CurrentID=0, Status='live', Symbol='LOC105378282', Aliases='', description='uncharacterized LOC105378282', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45215738, end_position_on_the_genomic_accession=45224468, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378283, CurrentID=0, Status='live', Symbol='LOC105378283', Aliases='', description='putative ankyrin repeat domain-containing protein ENSP00000383069', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45180356, end_position_on_the_genomic_accession=45194262, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928287, CurrentID=0, Status='live', Symbol='LINC01285', Aliases='', description='long intergenic non-protein coding RNA 1285', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118839556, end_position_on_the_genomic_accession=118882015, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378287, CurrentID=0, Status='live', Symbol='LOC105378287', Aliases='', description='uncharacterized LOC105378287', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46030741, end_position_on_the_genomic_accession=46038059, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130105, CurrentID=0, Status='live', Symbol='TRIM51CP', Aliases='', description='tripartite motif-containing 51C, pseudogene', other_designations='SPRY domain containing 5 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48945684, end_position_on_the_genomic_accession=48952068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378288, CurrentID=0, Status='live', Symbol='LOC105378288', Aliases='', description='uncharacterized LOC105378288', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46090397, end_position_on_the_genomic_accession=46103699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422528, CurrentID=0, Status='live', Symbol='LOC100422528', Aliases='', description='pleckstrin homology domain containing B2 pseudogene', other_designations='pleckstrin homology domain containing, family B (evectins) member 2 pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19668545, end_position_on_the_genomic_accession=19669052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372267, CurrentID=0, Status='live', Symbol='LOC105372267', Aliases='', description='nuclear factor interleukin-3-regulated protein-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8630645, end_position_on_the_genomic_accession=8677993, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100379564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379564, CurrentID=0, Status='live', Symbol='RNY4P17', Aliases='', description='RNA, Ro-associated Y4 pseudogene 17', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169005249, end_position_on_the_genomic_accession=169005343, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378290, CurrentID=0, Status='live', Symbol='LOC105378290', Aliases='', description='uncharacterized LOC105378290', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46906750, end_position_on_the_genomic_accession=46911302, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928288, CurrentID=0, Status='live', Symbol='LOC101928288', Aliases='', description='uncharacterized LOC101928288', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171221414, end_position_on_the_genomic_accession=171298146, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100379571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379571, CurrentID=0, Status='live', Symbol='BACE1-AS', Aliases='BACE1-AS1, BACE1AS, NCRNA00177', description='BACE1 antisense RNA', other_designations='BACE1 antisense RNA (non-protein coding)', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117291346, end_position_on_the_genomic_accession=117292170, orientation='plus', exon_count=1, OMIM=614263, no_hdr0=''),
  83786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83786, CurrentID=0, Status='live', Symbol='FRMD8', Aliases='FKSG44', description='FERM domain containing 8', other_designations='FKSG44', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65386570, end_position_on_the_genomic_accession=65413525, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100379584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379584, CurrentID=0, Status='live', Symbol='RNY4P14', Aliases='', description='RNA, Ro-associated Y4 pseudogene 14', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39785419, end_position_on_the_genomic_accession=39785510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379592, CurrentID=0, Status='live', Symbol='RNY4P10', Aliases='', description='RNA, Ro-associated Y4 pseudogene 10', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33199601, end_position_on_the_genomic_accession=33199692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646674, CurrentID=0, Status='live', Symbol='LOC646674', Aliases='', description='single stranded DNA binding protein 4 pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129992889, end_position_on_the_genomic_accession=129994037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378297, CurrentID=0, Status='live', Symbol='LOC105378297', Aliases='', description='uncharacterized LOC105378297', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48511685, end_position_on_the_genomic_accession=48519778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378298, CurrentID=0, Status='live', Symbol='LOC105378298', Aliases='', description='uncharacterized LOC105378298', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48976986, end_position_on_the_genomic_accession=48984429, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100288336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288336, CurrentID=0, Status='live', Symbol='LOC100288336', Aliases='', description='Putative uncharacterized protein ENSP00000383251', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=44561072, end_position_on_the_genomic_accession=44582387, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100379623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379623, CurrentID=0, Status='live', Symbol='LARP1P1', Aliases='HCG10P, dJ221C16.7', description='La ribonucleoprotein domain family member 1 pseudogene 1', other_designations='HLA complex group 10 pseudogene|La ribonucleoprotein domain family, member 1 pseudogene 1', map_location='6p22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26164033, end_position_on_the_genomic_accession=26168176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100510707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100510707, CurrentID=0, Status='live', Symbol='TBC1D3J', Aliases='', description='TBC1 domain family member 3J', other_designations='TBC1 domain family, member 3J', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36449777, end_position_on_the_genomic_accession=36456035, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378302, CurrentID=0, Status='live', Symbol='LOC105378302', Aliases='', description='uncharacterized LOC105378302', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50666542, end_position_on_the_genomic_accession=50674958, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928311, CurrentID=0, Status='live', Symbol='LOC101928311', Aliases='', description='uncharacterized LOC101928311', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42777813, end_position_on_the_genomic_accession=42780091, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378303, CurrentID=0, Status='live', Symbol='LOC105378303', Aliases='', description='uncharacterized LOC105378303', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51748350, end_position_on_the_genomic_accession=51763302, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378305, CurrentID=0, Status='live', Symbol='LOC105378305', Aliases='', description='uncharacterized LOC105378305', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52556961, end_position_on_the_genomic_accession=52755564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100379658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379658, CurrentID=0, Status='live', Symbol='RNY5P5', Aliases='', description='RNA, Ro-associated Y5 pseudogene 5', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25192413, end_position_on_the_genomic_accession=25192486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928291, CurrentID=0, Status='live', Symbol='LOC101928291', Aliases='', description='uncharacterized LOC101928291', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100379661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379661, CurrentID=0, Status='live', Symbol='GRIK1-AS2', Aliases='C21orf41, NCRNA00258', description='GRIK1 antisense RNA 2', other_designations='GRIK1 antisense RNA 2 (non-protein coding)', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29596040, end_position_on_the_genomic_accession=29630747, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100379663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379663, CurrentID=0, Status='live', Symbol='RNY5P6', Aliases='', description='RNA, Ro-associated Y5 pseudogene 6', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43152976, end_position_on_the_genomic_accession=43153051, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100379664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379664, CurrentID=0, Status='live', Symbol='RNY5P3', Aliases='', description='RNA, Ro-associated Y5 pseudogene 3', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170842098, end_position_on_the_genomic_accession=170842172, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379666, CurrentID=0, Status='live', Symbol='RNY5P1', Aliases='', description='RNA, Ro-associated Y5 pseudogene 1', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35427810, end_position_on_the_genomic_accession=35427892, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379667, CurrentID=0, Status='live', Symbol='NEK2P2', Aliases='', description='NEK2 pseudogene 2', other_designations='NIMA (never in mitosis gene a)-related kinase 2 pseudogene', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15611559, end_position_on_the_genomic_accession=15613237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370912, CurrentID=0, Status='live', Symbol='LOC105370912', Aliases='', description='uncharacterized LOC105370912', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78375312, end_position_on_the_genomic_accession=78409612, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928292, CurrentID=0, Status='live', Symbol='LINC01488', Aliases='BRCAT8', description='long intergenic non-protein coding RNA 1488', other_designations='breast cancer-associated transcript 8', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69485561, end_position_on_the_genomic_accession=69493543, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378311, CurrentID=0, Status='live', Symbol='LOC105378311', Aliases='', description='uncharacterized LOC105378311', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=54486230, end_position_on_the_genomic_accession=54656051, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378313, CurrentID=0, Status='live', Symbol='LOC105378313', Aliases='', description='uncharacterized LOC105378313', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=57322249, end_position_on_the_genomic_accession=57416411, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378315, CurrentID=0, Status='live', Symbol='LOC105378315', Aliases='', description='uncharacterized LOC105378315', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58346287, end_position_on_the_genomic_accession=58359032, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378316, CurrentID=0, Status='live', Symbol='LOC105378316', Aliases='', description='uncharacterized LOC105378316', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58433930, end_position_on_the_genomic_accession=58503479, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100271440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271440, CurrentID=0, Status='live', Symbol='RPL21P113', Aliases='RPL21_51_1437', description='ribosomal protein L21 pseudogene 113', other_designations='', map_location='15q21', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59049399, end_position_on_the_genomic_accession=59049733, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506811, CurrentID=0, Status='live', Symbol='LOC100506811', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=28659065, end_position_on_the_genomic_accession=28659469, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378321, CurrentID=0, Status='live', Symbol='LOC105378321', Aliases='', description='uncharacterized LOC105378321', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60967387, end_position_on_the_genomic_accession=60975802, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  149351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149351, CurrentID=0, Status='live', Symbol='LOC149351', Aliases='', description='uncharacterized LOC149351', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90830079, end_position_on_the_genomic_accession=90836161, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422535, CurrentID=0, Status='live', Symbol='FCF1P6', Aliases='', description='FCF1 pseudogene 6', other_designations='FCF1 small subunit processome component homolog pseudogene 6', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50405432, end_position_on_the_genomic_accession=50406024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378324, CurrentID=0, Status='live', Symbol='LOC105378324', Aliases='', description='uncharacterized LOC105378324', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61997017, end_position_on_the_genomic_accession=62000562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928295, CurrentID=0, Status='live', Symbol='LOC101928295', Aliases='', description='uncharacterized LOC101928295', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49368705, end_position_on_the_genomic_accession=49388081, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64756, CurrentID=0, Status='live', Symbol='ATPAF1', Aliases='ATP11, ATP11p', description='ATP synthase mitochondrial F1 complex assembly factor 1', other_designations='homolog of yeast ATP11', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46635039, end_position_on_the_genomic_accession=46668427, orientation='minus', exon_count=11, OMIM=608917, no_hdr0=''),
  105378327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378327, CurrentID=0, Status='live', Symbol='LOC105378327', Aliases='', description='uncharacterized LOC105378327', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62624528, end_position_on_the_genomic_accession=62655506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652897, CurrentID=0, Status='live', Symbol='LOC100652897', Aliases='', description='density-regulated protein pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76057430, end_position_on_the_genomic_accession=76058318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101559451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101559451, CurrentID=0, Status='live', Symbol='LOC101559451', Aliases='', description='uncharacterized LOC101559451', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4704230, end_position_on_the_genomic_accession=4705529, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378330, CurrentID=0, Status='live', Symbol='LOC105378330', Aliases='', description='uncharacterized LOC105378330', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63430575, end_position_on_the_genomic_accession=63431111, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928296, CurrentID=0, Status='live', Symbol='LOC101928296', Aliases='', description='ATP synthase-coupling factor 6, mitochondrial-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98628962, end_position_on_the_genomic_accession=98634216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130114, CurrentID=0, Status='live', Symbol='LOC100130114', Aliases='', description='Spi-C transcription factor (Spi-1/PU.1 related) pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35190935, end_position_on_the_genomic_accession=35191897, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378334, CurrentID=0, Status='live', Symbol='LOC105378334', Aliases='', description='uncharacterized LOC105378334', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64301520, end_position_on_the_genomic_accession=64693763, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928297, CurrentID=0, Status='live', Symbol='LOC101928297', Aliases='', description='tubulin beta-8 chain-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16853796, end_position_on_the_genomic_accession=16855280, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419447, CurrentID=0, Status='live', Symbol='LOC100419447', Aliases='', description='CSRP2 binding protein pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=88512029, end_position_on_the_genomic_accession=88513187, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378340, CurrentID=0, Status='live', Symbol='LOC105378340', Aliases='', description='uncharacterized LOC105378340', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=66079231, end_position_on_the_genomic_accession=66275772, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105376269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376269, CurrentID=0, Status='live', Symbol='LOC105376269', Aliases='', description='uncharacterized LOC105376269', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124282102, end_position_on_the_genomic_accession=124284039, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100419451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419451, CurrentID=0, Status='live', Symbol='LOC100419451', Aliases='', description='zinc finger protein 3 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99540527, end_position_on_the_genomic_accession=99541041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378343, CurrentID=0, Status='live', Symbol='LOC105378343', Aliases='', description='uncharacterized LOC105378343', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69405092, end_position_on_the_genomic_accession=69410996, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378344, CurrentID=0, Status='live', Symbol='LOC105378344', Aliases='', description='uncharacterized LOC105378344', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69547155, end_position_on_the_genomic_accession=69550352, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378346, CurrentID=0, Status='live', Symbol='LOC105378346', Aliases='', description='uncharacterized LOC105378346', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69684953, end_position_on_the_genomic_accession=69691179, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100873501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873501, CurrentID=0, Status='live', Symbol='RNA5SP245', Aliases='RN5S245', description='RNA, 5S ribosomal pseudogene 245', other_designations='RNA, 5S ribosomal 245', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130027277, end_position_on_the_genomic_accession=130027397, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378348, CurrentID=0, Status='live', Symbol='LOC105378348', Aliases='', description='uncharacterized LOC105378348', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70945915, end_position_on_the_genomic_accession=70947168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378349, CurrentID=0, Status='live', Symbol='LOC105378349', Aliases='', description='uncharacterized LOC105378349', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70939060, end_position_on_the_genomic_accession=70944144, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101060825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060825, CurrentID=0, Status='live', Symbol='LOC101060825', Aliases='', description='serine/arginine-rich splicing factor 11-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378350, CurrentID=0, Status='live', Symbol='LOC105378350', Aliases='', description='uncharacterized LOC105378350', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=70978335, end_position_on_the_genomic_accession=70999102, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130490, CurrentID=0, Status='live', Symbol='LOC100130490', Aliases='', description='tumor protein p63 regulated 1-like pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57664292, end_position_on_the_genomic_accession=57665036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378356, CurrentID=0, Status='live', Symbol='LOC105378356', Aliases='', description='uncharacterized LOC105378356', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71472884, end_position_on_the_genomic_accession=71474153, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130119, CurrentID=0, Status='live', Symbol='LAMTOR3P1', Aliases='', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 3 pseudogene 1', other_designations='MAPK scaffold protein 1 pseudogene', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36922443, end_position_on_the_genomic_accession=36923690, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378360, CurrentID=0, Status='live', Symbol='LOC105378360', Aliases='', description='uncharacterized LOC105378360', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73654129, end_position_on_the_genomic_accession=73669176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479495, CurrentID=0, Status='live', Symbol='RN7SL751P', Aliases='', description='RNA, 7SL, cytoplasmic 751, pseudogene', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378361, CurrentID=0, Status='live', Symbol='LOC105378361', Aliases='', description='uncharacterized LOC105378361', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75026427, end_position_on_the_genomic_accession=75034790, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378362, CurrentID=0, Status='live', Symbol='LOC105378362', Aliases='', description='uncharacterized LOC105378362', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74976338, end_position_on_the_genomic_accession=75004753, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060542, CurrentID=0, Status='live', Symbol='LOC101060542', Aliases='', description='uncharacterized LOC101060542', other_designations='', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70380140, end_position_on_the_genomic_accession=70384591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378364, CurrentID=0, Status='live', Symbol='LOC105378364', Aliases='', description='uncharacterized LOC105378364', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=76560354, end_position_on_the_genomic_accession=76561407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100506858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506858, CurrentID=0, Status='live', Symbol='LOC100506858', Aliases='', description='uncharacterized LOC100506858', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2303720, end_position_on_the_genomic_accession=2312201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100642175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100642175, CurrentID=0, Status='live', Symbol='SPRY4-IT1', Aliases='', description='SPRY4 intronic transcript 1', other_designations='SPRY4 intronic transcript 1 (non-protein coding)', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100642176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100642176, CurrentID=0, Status='live', Symbol='LINC00266-4P', Aliases='NCRNA00266-4P, NCRNA00266D, NCRNA00266Y2', description='long intergenic non-protein coding RNA 266-4, pseudogene', other_designations='non-protein coding RNA 266-4, pseudogene|non-protein coding RNA 266D', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25378671, end_position_on_the_genomic_accession=25391610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371910, CurrentID=0, Status='live', Symbol='LOC105371910', Aliases='', description='uncharacterized LOC105371910', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78248227, end_position_on_the_genomic_accession=78252491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83863, CurrentID=0, Status='live', Symbol='TTTY5', Aliases='LINC00126, NCRNA00126, TTY5', description='testis-specific transcript, Y-linked 5 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 126|transcript Y 5', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22296798, end_position_on_the_genomic_accession=22298876, orientation='minus', exon_count=2, OMIM=400038, no_hdr0=''),
  101928304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928304, CurrentID=0, Status='live', Symbol='LOC101928304', Aliases='', description='uncharacterized LOC101928304', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134525318, end_position_on_the_genomic_accession=134540005, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378371, CurrentID=0, Status='live', Symbol='LOC105378371', Aliases='', description='uncharacterized LOC105378371', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77045566, end_position_on_the_genomic_accession=77061082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375279, CurrentID=0, Status='live', Symbol='LOC105375279', Aliases='', description='uncharacterized LOC105375279', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=52022309, end_position_on_the_genomic_accession=52046314, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480499, CurrentID=0, Status='live', Symbol='RN7SL301P', Aliases='', description='RNA, 7SL, cytoplasmic 301, pseudogene', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371911, CurrentID=0, Status='live', Symbol='LOC105371911', Aliases='', description='uncharacterized LOC105371911', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78279631, end_position_on_the_genomic_accession=78306116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389705, CurrentID=0, Status='live', Symbol='LOC389705', Aliases='', description='chromosome 4 open reading frame 27 pseudogene', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14993327, end_position_on_the_genomic_accession=15019724, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616100, CurrentID=0, Status='live', Symbol='ERVK-2', Aliases='c3_A', description='endogenous retrovirus group K member 2', other_designations='endogenous retrovirus group K, member 2|env protein', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371913, CurrentID=0, Status='live', Symbol='LOC105371913', Aliases='', description='uncharacterized LOC105371913', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78670609, end_position_on_the_genomic_accession=78675631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83877, CurrentID=0, Status='live', Symbol='TM2D2', Aliases='BLP1', description='TM2 domain containing 2', other_designations='BBP-like protein 1|beta-amyloid-binding protein-like protein 1', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38988808, end_position_on_the_genomic_accession=38996816, orientation='minus', exon_count=5, OMIM=610081, no_hdr0=''),
  105370915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370915, CurrentID=0, Status='live', Symbol='LOC105370915', Aliases='', description='uncharacterized LOC105370915', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78860863, end_position_on_the_genomic_accession=78863611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616102, CurrentID=0, Status='live', Symbol='ERVK-9', Aliases='polymerase', description='endogenous retrovirus group K, member 9', other_designations='HERV-K_6q14.1 provirus ancestral Env polyprotein|HERV-K_6q14.1 provirus ancestral Gag polyprotein|HERV-K_6q14.1 provirus ancestral Gag-Pol polyprotein|HERV-K_6q14.1 provirus ancestral Pro protein', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616103, CurrentID=0, Status='live', Symbol='ERVK-21', Aliases='Endo-Rev2, c12_A, envK1', description='endogenous retrovirus group K member 21', other_designations='Envelope polyprotein|HERV-K_12q14.1 provirus Rec protein|HERV-K_12q14.1 provirus ancestral Env polyprotein|HERV-K_12q14.1 provirus ancestral Gag polyprotein|HERV-K_12q14.1 provirus ancestral Pro protein|endogenous retrovirus group K, member 21', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  83880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83880, CurrentID=0, Status='live', Symbol='EIF3FP2', Aliases='IFP38', description='eukaryotic translation initiation factor 3 subunit F pseudogene 2', other_designations='eukaryotic translation initiation factor 3, subunit 5 epsilon pseudogene|eukaryotic translation initiation factor 3, subunit F pseudogene 2', map_location='13p13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18206420, end_position_on_the_genomic_accession=18207468, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378387, CurrentID=0, Status='live', Symbol='LOC105378387', Aliases='', description='uncharacterized LOC105378387', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80629686, end_position_on_the_genomic_accession=80649343, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  83881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83881, CurrentID=0, Status='live', Symbol='MIXL1', Aliases='MILD1, MIX, MIXL', description='Mix paired-like homeobox', other_designations='MIX1 homeobox-like protein 1|Mix-like homeobox protein 1|homeodomain protein MIX|mix.1 homeobox-like protein', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226223549, end_position_on_the_genomic_accession=226227056, orientation='plus', exon_count=2, OMIM=609852, no_hdr0=''),
  105371914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371914, CurrentID=0, Status='live', Symbol='LOC105371914', Aliases='', description='uncharacterized LOC105371914', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78903339, end_position_on_the_genomic_accession=78913431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616106, CurrentID=0, Status='live', Symbol='ERVK3-4', Aliases='HML6-c11', description='endogenous retrovirus group K3 member 4', other_designations='endogenous retrovirus group K3, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616108, CurrentID=0, Status='live', Symbol='ERVK3-8', Aliases='', description='endogenous retrovirus group K3 member 8', other_designations='endogenous retrovirus group K3, member 8', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  102477328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102477328, CurrentID=0, Status='live', Symbol='LOC102477328', Aliases='', description='uncharacterized LOC102477328', other_designations='CTC-347C20.2', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72574075, end_position_on_the_genomic_accession=72660669, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616110, CurrentID=0, Status='live', Symbol='MIR2681', Aliases='', description='microRNA 2681', other_designations='hsa-mir-2681', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101967642, end_position_on_the_genomic_accession=101967746, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616111, CurrentID=0, Status='live', Symbol='MIR4455', Aliases='', description='microRNA 4455', other_designations='hsa-mir-4455', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184938383, end_position_on_the_genomic_accession=184938440, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616112, CurrentID=0, Status='live', Symbol='MIR4793', Aliases='', description='microRNA 4793', other_designations='hsa-mir-4793', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48644194, end_position_on_the_genomic_accession=48644280, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378395, CurrentID=0, Status='live', Symbol='LOC105378395', Aliases='', description='uncharacterized LOC105378395', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83560958, end_position_on_the_genomic_accession=83565143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616113, CurrentID=0, Status='live', Symbol='MIR4680', Aliases='', description='microRNA 4680', other_designations='hsa-mir-4680', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110898090, end_position_on_the_genomic_accession=110898155, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616114, CurrentID=0, Status='live', Symbol='MIR4668', Aliases='mir-4668', description='microRNA 4668', other_designations='hsa-mir-4668', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111932100, end_position_on_the_genomic_accession=111932169, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480504, CurrentID=0, Status='live', Symbol='RN7SL363P', Aliases='', description='RNA, 7SL, cytoplasmic 363, pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616115, CurrentID=0, Status='live', Symbol='MIR4469', Aliases='mir-4469', description='microRNA 4469', other_designations='hsa-mir-4469', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42896197, end_position_on_the_genomic_accession=42896275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616116, CurrentID=0, Status='live', Symbol='MIR4648', Aliases='', description='microRNA 4648', other_designations='hsa-mir-4648', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2527074, end_position_on_the_genomic_accession=2527145, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389727, CurrentID=0, Status='live', Symbol='FKBP4P2', Aliases='', description='FK506 binding protein 4 pseudogene 2', other_designations='FK506 binding protein 4, 59kDa pseudogene 2', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39630902, end_position_on_the_genomic_accession=39632510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616117, CurrentID=0, Status='live', Symbol='MIR4653', Aliases='', description='microRNA 4653', other_designations='hsa-mir-4653', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101159473, end_position_on_the_genomic_accession=101159555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616118, CurrentID=0, Status='live', Symbol='MIR4486', Aliases='', description='microRNA 4486', other_designations='hsa-mir-4486', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19575310, end_position_on_the_genomic_accession=19575372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616119, CurrentID=0, Status='live', Symbol='MIR4697', Aliases='', description='microRNA 4697', other_designations='hsa-mir-4697', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133898504, end_position_on_the_genomic_accession=133898581, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378402, CurrentID=0, Status='live', Symbol='LOC105378402', Aliases='', description='uncharacterized LOC105378402', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85199067, end_position_on_the_genomic_accession=85212802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616120, CurrentID=0, Status='live', Symbol='MIR4695', Aliases='', description='microRNA 4695', other_designations='hsa-mir-4695', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18883202, end_position_on_the_genomic_accession=18883275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616121, CurrentID=0, Status='live', Symbol='MIR1268B', Aliases='', description='microRNA 1268b', other_designations='hsa-mir-1268b', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80098828, end_position_on_the_genomic_accession=80098877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616123, CurrentID=0, Status='live', Symbol='MIR4436B1', Aliases='MIR4436B, hsa-mir-4436b-1', description='microRNA 4436b-1', other_designations='hsa-mir-4436b|microRNA 4436b', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110086433, end_position_on_the_genomic_accession=110086523, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616124, CurrentID=0, Status='live', Symbol='MIR4647', Aliases='mir-4647', description='microRNA 4647', other_designations='hsa-mir-4647', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44254206, end_position_on_the_genomic_accession=44254285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616125, CurrentID=0, Status='live', Symbol='MIR4731', Aliases='', description='microRNA 4731', other_designations='hsa-mir-4731', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15251627, end_position_on_the_genomic_accession=15251696, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100380272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380272, CurrentID=0, Status='live', Symbol='NDUFB1P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B1 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 1 pseudogene 2|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 1, 7kDa pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222945725, end_position_on_the_genomic_accession=222945900, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616126, CurrentID=0, Status='live', Symbol='MIR4686', Aliases='', description='microRNA 4686', other_designations='hsa-mir-4686', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2173063, end_position_on_the_genomic_accession=2173138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378409, CurrentID=0, Status='live', Symbol='LOC105378409', Aliases='', description='uncharacterized LOC105378409', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86656996, end_position_on_the_genomic_accession=86666318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370916, CurrentID=0, Status='live', Symbol='LOC105370916', Aliases='', description='uncharacterized LOC105370916', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78920937, end_position_on_the_genomic_accession=78921993, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616127, CurrentID=0, Status='live', Symbol='MIR4448', Aliases='', description='microRNA 4448', other_designations='hsa-mir-4448', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183886800, end_position_on_the_genomic_accession=183886885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378412, CurrentID=0, Status='live', Symbol='LOC105378412', Aliases='', description='uncharacterized LOC105378412', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87596449, end_position_on_the_genomic_accession=87659279, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378413, CurrentID=0, Status='live', Symbol='LOC105378413', Aliases='', description='uncharacterized LOC105378413', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87769963, end_position_on_the_genomic_accession=87802099, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644060, CurrentID=0, Status='live', Symbol='LOC644060', Aliases='', description='IGF-II mRNA-binding protein 2-like', other_designations='insulin-like growth factor 2 mRNA binding protein 2 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147050817, end_position_on_the_genomic_accession=147052478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100419525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419525, CurrentID=0, Status='live', Symbol='LOC100419525', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31264214, end_position_on_the_genomic_accession=31264785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378418, CurrentID=0, Status='live', Symbol='LOC105378418', Aliases='', description='uncharacterized LOC105378418', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89179302, end_position_on_the_genomic_accession=89194754, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378419, CurrentID=0, Status='live', Symbol='LOC105378419', Aliases='', description='uncharacterized LOC105378419', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89278360, end_position_on_the_genomic_accession=89314581, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375281, CurrentID=0, Status='live', Symbol='LOC105375281', Aliases='', description='uncharacterized LOC105375281', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=52839565, end_position_on_the_genomic_accession=52843577, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  503496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503496, CurrentID=0, Status='live', Symbol='MRX86', Aliases='', description='mental retardation, X-linked 86', other_designations='', map_location='Xp11.3-q13.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  503497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503497, CurrentID=0, Status='live', Symbol='MS4A13', Aliases='', description='membrane-spanning 4-domains, subfamily A, member 13', other_designations='testis-expressed transmembrane protein 4|testis-expressed transmembrane-4 protein', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60515376, end_position_on_the_genomic_accession=60543424, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378424, CurrentID=0, Status='live', Symbol='LOC105378424', Aliases='', description='uncharacterized LOC105378424', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89823519, end_position_on_the_genomic_accession=89830304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  503519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503519, CurrentID=0, Status='live', Symbol='LINC00929', Aliases='', description='long intergenic non-protein coding RNA 929', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26115813, end_position_on_the_genomic_accession=26133037, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100288465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288465, CurrentID=0, Status='live', Symbol='FAM197Y10', Aliases='', description='family with sequence similarity 197 Y-linked member 10', other_designations='family with sequence similarity 197, Y-linked, member 10', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=23625546, end_position_on_the_genomic_accession=23627679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378428, CurrentID=0, Status='live', Symbol='LOC105378428', Aliases='', description='uncharacterized LOC105378428', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91186676, end_position_on_the_genomic_accession=91199596, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  503538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503538, CurrentID=0, Status='live', Symbol='A1BG-AS1', Aliases='A1BG-AS, A1BGAS, NCRNA00181', description='A1BG antisense RNA 1', other_designations='A1BG antisense RNA (non-protein coding)|A1BG antisense RNA 1 (non-protein coding)', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58351970, end_position_on_the_genomic_accession=58355183, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  503540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503540, CurrentID=0, Status='live', Symbol='LOC503540', Aliases='', description='AKT interacting protein pseudogene', other_designations='fused toes homolog pseudogene', map_location='5p14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24170190, end_position_on_the_genomic_accession=24172277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503542, CurrentID=0, Status='live', Symbol='SPRN', Aliases='SHADOO, SHO, bA108K14.1', description='shadow of prion protein homolog (zebrafish)', other_designations='hypothetical protein BC004409|protein shadoo', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133420666, end_position_on_the_genomic_accession=133424617, orientation='minus', exon_count=2, OMIM=610447, no_hdr0=''),
  503543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503543, CurrentID=0, Status='live', Symbol='BTF3P9', Aliases='', description='basic transcription factor 3 pseudogene 9', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227433913, end_position_on_the_genomic_accession=227434606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928316, CurrentID=0, Status='live', Symbol='LOC101928316', Aliases='', description='uncharacterized LOC101928316', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30686553, end_position_on_the_genomic_accession=30735039, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100419542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419542, CurrentID=0, Status='live', Symbol='LOC100419542', Aliases='', description='ribosomal protein L37 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75642645, end_position_on_the_genomic_accession=75645089, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  127617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127617, CurrentID=0, Status='live', Symbol='OR14L1P', Aliases='OR1-35, OR1-36, OR5AV1, OR5AV1P', description='olfactory receptor family 14 subfamily L member 1 pseudogene', other_designations='olfactory receptor OR1-35 pseudogene|olfactory receptor OR1-36|olfactory receptor, family 14, subfamily L, member 1 pseudogene|olfactory receptor, family 5, subfamily AV, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247619681, end_position_on_the_genomic_accession=247620559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378434, CurrentID=0, Status='live', Symbol='LOC105378434', Aliases='', description='uncharacterized LOC105378434', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92555433, end_position_on_the_genomic_accession=92573911, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  503569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503569, CurrentID=0, Status='live', Symbol='RGMB-AS1', Aliases='', description='RGMB antisense RNA 1', other_designations='CTC-463N11.3', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98769618, end_position_on_the_genomic_accession=98773084, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378435, CurrentID=0, Status='live', Symbol='LOC105378435', Aliases='', description='uncharacterized LOC105378435', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93059297, end_position_on_the_genomic_accession=93060500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928317, CurrentID=0, Status='live', Symbol='LOC101928317', Aliases='', description='uncharacterized LOC101928317', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14529831, end_position_on_the_genomic_accession=14535673, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616154, CurrentID=0, Status='live', Symbol='MIR4466', Aliases='mir-4466', description='microRNA 4466', other_designations='hsa-mir-4466', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156779678, end_position_on_the_genomic_accession=156779731, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503582, CurrentID=0, Status='live', Symbol='ARGFX', Aliases='', description='arginine-fifty homeobox', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121562303, end_position_on_the_genomic_accession=121590622, orientation='plus', exon_count=6, OMIM=611164, no_hdr0=''),
  503583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503583, CurrentID=0, Status='live', Symbol='ARGFXP1', Aliases='', description='arginine-fifty homeobox pseudogene 1', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122675700, end_position_on_the_genomic_accession=122680538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616155, CurrentID=0, Status='live', Symbol='MIR4536-1', Aliases='MIR4536, hsa-mir-4536-1', description='microRNA 4536-1', other_designations='hsa-mir-4536|microRNA 4536', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55451495, end_position_on_the_genomic_accession=55451582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378438, CurrentID=0, Status='live', Symbol='LOC105378438', Aliases='', description='uncharacterized LOC105378438', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93987724, end_position_on_the_genomic_accession=93993737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422558, CurrentID=0, Status='live', Symbol='LOC100422558', Aliases='', description='SMG1 phosphatidylinositol 3-kinase-related kinase pseudogene', other_designations='smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29426639, end_position_on_the_genomic_accession=29427339, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100506929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506929, CurrentID=0, Status='live', Symbol='LOC100506929', Aliases='', description='uncharacterized LOC100506929', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240739142, end_position_on_the_genomic_accession=240743679, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616156, CurrentID=0, Status='live', Symbol='MIR1273F', Aliases='', description='microRNA 1273f', other_designations='hsa-mir-1273f', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52928674, end_position_on_the_genomic_accession=52928772, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378439, CurrentID=0, Status='live', Symbol='LOC105378439', Aliases='', description='uncharacterized LOC105378439', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96042611, end_position_on_the_genomic_accession=96044341, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616157, CurrentID=0, Status='live', Symbol='MIR4772', Aliases='mir-4772', description='microRNA 4772', other_designations='hsa-mir-4772', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102432289, end_position_on_the_genomic_accession=102432366, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389768, CurrentID=0, Status='live', Symbol='LOC389768', Aliases='', description='potassium channel tetramerization domain containing 1 pseudogene', other_designations='potassium channel tetramerisation domain containing 1 pseudogene', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88180657, end_position_on_the_genomic_accession=88182557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503611, CurrentID=0, Status='live', Symbol='TAAR7P', Aliases='', description='trace amine associated receptor 7, pseudogene', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132559024, end_position_on_the_genomic_accession=132559233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503612, CurrentID=0, Status='live', Symbol='TAAR4P', Aliases='', description='trace amine associated receptor 4, pseudogene', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132594389, end_position_on_the_genomic_accession=132595439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503613, CurrentID=0, Status='live', Symbol='PSORS10', Aliases='', description='psoriasis susceptibility 10', other_designations='', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612410, no_hdr0=''),
  503614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503614, CurrentID=0, Status='live', Symbol='DEFB107B', Aliases='HsT21816', description='defensin beta 107B', other_designations='beta-defensin 7|defensin, beta 107|defensin, beta 107B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7495846, end_position_on_the_genomic_accession=7509311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  503618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503618, CurrentID=0, Status='live', Symbol='DEFB104B', Aliases='BD-4, DEFB-4, hBD-4', description='defensin beta 104B', other_designations='beta-defensin 4|defensin, beta 104|defensin, beta 104B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7470308, end_position_on_the_genomic_accession=7475082, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422559, CurrentID=0, Status='live', Symbol='LOC100422559', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 2-like 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36656322, end_position_on_the_genomic_accession=36656941, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616161, CurrentID=0, Status='live', Symbol='MIR4795', Aliases='', description='microRNA 4795', other_designations='hsa-mir-4795', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87226189, end_position_on_the_genomic_accession=87226277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503627, CurrentID=0, Status='live', Symbol='TPRX2P', Aliases='TPRX2P1, TPRXP1', description='tetra-peptide repeat homeobox 2 pseudogene', other_designations='tetra-peptide repeat homeobox 1 pseudogene|tetra-peptide repeat homeobox pseudogene 1', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47860182, end_position_on_the_genomic_accession=47862000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503628, CurrentID=0, Status='live', Symbol='TPRX1P1', Aliases='TPRXP2', description='tetra-peptide repeat homeobox 1 pseudogene 1', other_designations='tetra-peptide repeat homeobox pseudogene 2', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79499088, end_position_on_the_genomic_accession=79500012, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616162, CurrentID=0, Status='live', Symbol='MIR4483', Aliases='', description='microRNA 4483', other_designations='hsa-mir-4483', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113777993, end_position_on_the_genomic_accession=113778054, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503630, CurrentID=0, Status='live', Symbol='DUXAP1', Aliases='', description='double homeobox A pseudogene 1', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84750727, end_position_on_the_genomic_accession=84751261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503631, CurrentID=0, Status='live', Symbol='DUXAP2', Aliases='', description='double homeobox A pseudogene 2', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101366338, end_position_on_the_genomic_accession=101367047, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503632, CurrentID=0, Status='live', Symbol='DUXAP3', Aliases='', description='double homeobox A pseudogene 3', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42745272, end_position_on_the_genomic_accession=42748704, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503633, CurrentID=0, Status='live', Symbol='DUXAP4', Aliases='', description='double homeobox A pseudogene 4', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45055614, end_position_on_the_genomic_accession=45057226, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503634, CurrentID=0, Status='live', Symbol='DUXAP5', Aliases='', description='double homeobox A pseudogene 5', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119444237, end_position_on_the_genomic_accession=119445554, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616163, CurrentID=0, Status='live', Symbol='MIR4530', Aliases='', description='microRNA 4530', other_designations='hsa-mir-4530', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39409623, end_position_on_the_genomic_accession=39409678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503636, CurrentID=0, Status='live', Symbol='DUXAP7', Aliases='', description='double homeobox A pseudogene 7', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18642310, end_position_on_the_genomic_accession=18643647, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503637, CurrentID=0, Status='live', Symbol='DUXAP8', Aliases='', description='double homeobox A pseudogene 8', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15784954, end_position_on_the_genomic_accession=15827434, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  503638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503638, CurrentID=0, Status='live', Symbol='DUXAP9', Aliases='', description='double homeobox A pseudogene 9', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19104058, end_position_on_the_genomic_accession=19106298, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503639, CurrentID=0, Status='live', Symbol='DUXAP10', Aliases='', description='double homeobox A pseudogene 10', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19294785, end_position_on_the_genomic_accession=19337674, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  503640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503640, CurrentID=0, Status='live', Symbol='ARGFXP2', Aliases='', description='arginine-fifty homeobox pseudogene 2', other_designations='arginine-fifty homeobox pseudogene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32150357, end_position_on_the_genomic_accession=32151571, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616164, CurrentID=0, Status='live', Symbol='MIR4420', Aliases='mir-4420', description='microRNA 4420', other_designations='hsa-mir-4420', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30739156, end_position_on_the_genomic_accession=30739232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503643, CurrentID=0, Status='live', Symbol='DPRXP2', Aliases='', description='divergent-paired related homeobox pseudogene 2', other_designations='divergent paired-related homeobox pseudogene 2', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35989490, end_position_on_the_genomic_accession=35990492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503644, CurrentID=0, Status='live', Symbol='DPRXP3', Aliases='', description='divergent-paired related homeobox pseudogene 3', other_designations='divergent paired-related homeobox pseudogene 3', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35222127, end_position_on_the_genomic_accession=35222765, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503645, CurrentID=0, Status='live', Symbol='DPRXP4', Aliases='', description='divergent-paired related homeobox pseudogene 4', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30975335, end_position_on_the_genomic_accession=30976004, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  503646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503646, CurrentID=0, Status='live', Symbol='DPRXP5', Aliases='', description='divergent-paired related homeobox pseudogene 5', other_designations='', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36943160, end_position_on_the_genomic_accession=36944254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  83941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83941, CurrentID=0, Status='live', Symbol='TM2D1', Aliases='BBP', description='TM2 domain containing 1', other_designations='beta-amyloid-binding protein|hBBP', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61681046, end_position_on_the_genomic_accession=61725423, orientation='minus', exon_count=8, OMIM=610080, no_hdr0=''),
  503648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503648, CurrentID=0, Status='live', Symbol='DPRXP7', Aliases='', description='divergent-paired related homeobox pseudogene 7', other_designations='divergent paired-related homeobox pseudogene 7', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112319868, end_position_on_the_genomic_accession=112320521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288486, CurrentID=0, Status='live', Symbol='DDX11L9', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 9', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 9|DEAD/H box polypeptide 11 like 9', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101976558, end_position_on_the_genomic_accession=101979093, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419838, CurrentID=0, Status='live', Symbol='LOC100419838', Aliases='', description='zinc finger protein 860 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52861816, end_position_on_the_genomic_accession=52864159, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616167, CurrentID=0, Status='live', Symbol='MIR4722', Aliases='', description='microRNA 4722', other_designations='hsa-mir-4722', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88716278, end_position_on_the_genomic_accession=88716337, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926888, CurrentID=0, Status='live', Symbol='RALY-AS1', Aliases='', description='RALY antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33992488, end_position_on_the_genomic_accession=33994367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100506255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506255, CurrentID=0, Status='live', Symbol='LOC100506255', Aliases='', description='uncharacterized LOC100506255', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=126178353, end_position_on_the_genomic_accession=126181688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101926889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926889, CurrentID=0, Status='live', Symbol='LOC101926889', Aliases='', description='uncharacterized LOC101926889', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25184672, end_position_on_the_genomic_accession=25196892, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  121520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121520, CurrentID=0, Status='live', Symbol='GOT2P4', Aliases='GOT2L4', description='glutamic-oxaloacetic transaminase 2 pseudogene 4', other_designations='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene|glutamic-oxaloacetic transaminase 2-like pseudogene', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16513340, end_position_on_the_genomic_accession=16514870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926890, CurrentID=0, Status='live', Symbol='DAB1-AS1', Aliases='', description='DAB1 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=57860543, end_position_on_the_genomic_accession=57863114, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926892, CurrentID=0, Status='live', Symbol='LOC101926892', Aliases='', description='uncharacterized LOC101926892', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70471134, end_position_on_the_genomic_accession=70485687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101822346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101822346, CurrentID=0, Status='live', Symbol='TRN-GTT5-1', Aliases='TRNAN-GUU, TRNAN37', description='transfer RNA-Asn (GTT) 5-1', other_designations='tRNA-Asn-GTT-5-1|transfer RNA asparagine (anticodon GUU)|transfer RNA asparagine 37 (anticodon GUU)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378455, CurrentID=0, Status='live', Symbol='LOC105378455', Aliases='', description='uncharacterized LOC105378455', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101702065, end_position_on_the_genomic_accession=101704752, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645263, CurrentID=0, Status='live', Symbol='LOC645263', Aliases='', description='hCG1793639', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=45613312, end_position_on_the_genomic_accession=45613584, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  503693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503693, CurrentID=0, Status='live', Symbol='LOH12CR2', Aliases='LOH2CR12', description='loss of heterozygosity, 12, chromosomal region 2 (non-protein coding)', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12355408, end_position_on_the_genomic_accession=12357067, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  503694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503694, CurrentID=0, Status='live', Symbol='DEFB108P1', Aliases='DEFB-8, DEFB108, DEFB108A', description='defensin beta 108 pseudogene 1', other_designations='defensin, beta 108, pseudogene 1', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7934412, end_position_on_the_genomic_accession=7938770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926893, CurrentID=0, Status='live', Symbol='LOC101926893', Aliases='', description='uncharacterized LOC101926893', other_designations='CTD-2215E18.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97073744, end_position_on_the_genomic_accession=97183118, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  101926895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926895, CurrentID=0, Status='live', Symbol='CRTC3-AS1', Aliases='', description='CRTC3 antisense RNA 1', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90620009, end_position_on_the_genomic_accession=90717140, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422562, CurrentID=0, Status='live', Symbol='LOC100422562', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 2-like 1 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=92821966, end_position_on_the_genomic_accession=92823013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926896, CurrentID=0, Status='live', Symbol='GLIS2-AS1', Aliases='', description='GLIS2 antisense RNA 1', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4324648, end_position_on_the_genomic_accession=4328340, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378459, CurrentID=0, Status='live', Symbol='LOC105378459', Aliases='', description='uncharacterized LOC105378459', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102120240, end_position_on_the_genomic_accession=102124115, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616177, CurrentID=0, Status='live', Symbol='MIR4419A', Aliases='', description='microRNA 4419a', other_designations='hsa-mir-4419a', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23057858, end_position_on_the_genomic_accession=23057934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926898, CurrentID=0, Status='live', Symbol='LOC101926898', Aliases='', description='uncharacterized LOC101926898', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46492052, end_position_on_the_genomic_accession=46533486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616179, CurrentID=0, Status='live', Symbol='MIR4498', Aliases='mir-4498', description='microRNA 4498', other_designations='hsa-mir-4498', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120155434, end_position_on_the_genomic_accession=120155499, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926900, CurrentID=0, Status='live', Symbol='DNAJB5-AS1', Aliases='', description='DNAJB5 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34985410, end_position_on_the_genomic_accession=34989379, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371929, CurrentID=0, Status='live', Symbol='LOC105371929', Aliases='', description='uncharacterized LOC105371929', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81733569, end_position_on_the_genomic_accession=81739548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=83957, CurrentID=0, Status='live', Symbol='FRMD8P1', Aliases='', description='FERM domain containing 8 pseudogene 1', other_designations='FERM domain-containing protein 8 pseudogene|FKSG43', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65550622, end_position_on_the_genomic_accession=65552421, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372664, CurrentID=0, Status='live', Symbol='LOC105372664', Aliases='', description='uncharacterized LOC105372664', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52080871, end_position_on_the_genomic_accession=52096481, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378464, CurrentID=0, Status='live', Symbol='LOC105378464', Aliases='', description='uncharacterized LOC105378464', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104566946, end_position_on_the_genomic_accession=104624743, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100506955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100506955, CurrentID=0, Status='live', Symbol='LOC100506955', Aliases='', description='cancer/testis antigen 83 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116692973, end_position_on_the_genomic_accession=116693239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616183, CurrentID=0, Status='live', Symbol='MIR4513', Aliases='', description='microRNA 4513', other_designations='hsa-mir-4513', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74788672, end_position_on_the_genomic_accession=74788757, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928323, CurrentID=0, Status='live', Symbol='LOC101928323', Aliases='', description='uncharacterized LOC101928323', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42612310, end_position_on_the_genomic_accession=42654388, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369474, CurrentID=0, Status='live', Symbol='LOC105369474', Aliases='', description='uncharacterized LOC105369474', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106367231, end_position_on_the_genomic_accession=106409184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926904, CurrentID=0, Status='live', Symbol='LINC01265', Aliases='', description='long intergenic non-protein coding RNA 1265', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38710367, end_position_on_the_genomic_accession=38720273, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100271446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271446, CurrentID=0, Status='live', Symbol='RPL21P118', Aliases='RPL21_57_1497', description='ribosomal protein L21 pseudogene 118', other_designations='', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74313036, end_position_on_the_genomic_accession=74313383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926905, CurrentID=0, Status='live', Symbol='LOC101926905', Aliases='', description='uncharacterized LOC101926905', other_designations='CTC-270D5.1', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141046205, end_position_on_the_genomic_accession=141078031, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371930, CurrentID=0, Status='live', Symbol='LOC105371930', Aliases='', description='uncharacterized LOC105371930', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28222783, end_position_on_the_genomic_accession=28228461, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422564, CurrentID=0, Status='live', Symbol='LOC100422564', Aliases='', description='glucosaminyl (N-acetyl) transferase 2, I-branching enzyme (I blood group) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10514185, end_position_on_the_genomic_accession=10514946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926906, CurrentID=0, Status='live', Symbol='SLC16A12-AS1', Aliases='', description='SLC16A12 antisense RNA 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89455991, end_position_on_the_genomic_accession=89468151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378469, CurrentID=0, Status='live', Symbol='LOC105378469', Aliases='', description='uncharacterized LOC105378469', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106095794, end_position_on_the_genomic_accession=106140098, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101926907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926907, CurrentID=0, Status='live', Symbol='LOC101926907', Aliases='', description='uncharacterized LOC101926907', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58805296, end_position_on_the_genomic_accession=58816042, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926908, CurrentID=0, Status='live', Symbol='LOC101926908', Aliases='', description='uncharacterized LOC101926908', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72732045, end_position_on_the_genomic_accession=72751843, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926910, CurrentID=0, Status='live', Symbol='LOC101926910', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30349719, end_position_on_the_genomic_accession=30350632, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926911, CurrentID=0, Status='live', Symbol='LOC101926911', Aliases='', description='uncharacterized LOC101926911', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91022622, end_position_on_the_genomic_accession=91031139, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926913, CurrentID=0, Status='live', Symbol='LOC101926913', Aliases='', description='uncharacterized LOC101926913', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170700368, end_position_on_the_genomic_accession=170770766, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101926915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926915, CurrentID=0, Status='live', Symbol='LOC101926915', Aliases='', description='uncharacterized LOC101926915', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46276324, end_position_on_the_genomic_accession=46365508, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  346116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346116, CurrentID=0, Status='live', Symbol='MRPL42P2', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 2', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16171606, end_position_on_the_genomic_accession=16172034, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503834, CurrentID=0, Status='live', Symbol='DPRX', Aliases='', description='divergent-paired related homeobox', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53601119, end_position_on_the_genomic_accession=53637038, orientation='plus', exon_count=5, OMIM=611165, no_hdr0=''),
  503835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503835, CurrentID=0, Status='live', Symbol='DUXA', Aliases='', description='double homeobox A', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57151726, end_position_on_the_genomic_accession=57167488, orientation='minus', exon_count=6, OMIM=611168, no_hdr0=''),
  503837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503837, CurrentID=0, Status='live', Symbol='DEFB108P2', Aliases='', description='defensin beta 108 pseudogene 2', other_designations='defensin, beta 108, pseudogene 2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7373144, end_position_on_the_genomic_accession=7377500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  503838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503838, CurrentID=0, Status='live', Symbol='DEFB109P3', Aliases='', description='defensin beta 109 pseudogene 3', other_designations='defensin, beta 109, pseudogene 3|defensin, beta 9 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12150963, end_position_on_the_genomic_accession=12158034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503839, CurrentID=0, Status='live', Symbol='DEFB109P2', Aliases='', description='defensin beta 109 pseudogene 2', other_designations='defensin, beta 109, pseudogene 2|defensin, beta 9 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12437655, end_position_on_the_genomic_accession=12444737, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503840, CurrentID=0, Status='live', Symbol='DEFB108P4', Aliases='', description='defensin beta 108 pseudogene 4', other_designations='defensin, beta 108, pseudogene 4', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12341424, end_position_on_the_genomic_accession=12345778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  503841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503841, CurrentID=0, Status='live', Symbol='DEFB106B', Aliases='BD-6, DEFB-6', description='defensin beta 106B', other_designations='beta-defensin 6|defensin, beta 106|defensin, beta 106B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7482504, end_position_on_the_genomic_accession=7486387, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  503842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503842, CurrentID=0, Status='live', Symbol='DFNB60', Aliases='', description='deafness, autosomal recessive 60', other_designations='', map_location='5q22-q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101926918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926918, CurrentID=0, Status='live', Symbol='LOC101926918', Aliases='', description='uncharacterized LOC101926918', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119412502, end_position_on_the_genomic_accession=119454651, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  105378481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378481, CurrentID=0, Status='live', Symbol='LOC105378481', Aliases='', description='uncharacterized LOC105378481', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110474348, end_position_on_the_genomic_accession=110494740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371933, CurrentID=0, Status='live', Symbol='LOC105371933', Aliases='', description='uncharacterized LOC105371933', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35372833, end_position_on_the_genomic_accession=35374437, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378484, CurrentID=0, Status='live', Symbol='LOC105378484', Aliases='', description='uncharacterized LOC105378484', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111400281, end_position_on_the_genomic_accession=111410597, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378485, CurrentID=0, Status='live', Symbol='LOC105378485', Aliases='', description='uncharacterized LOC105378485', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111439300, end_position_on_the_genomic_accession=111442826, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926923, CurrentID=0, Status='live', Symbol='LOC101926923', Aliases='', description='uncharacterized LOC101926923', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197333688, end_position_on_the_genomic_accession=197341152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378486, CurrentID=0, Status='live', Symbol='LOC105378486', Aliases='', description='uncharacterized LOC105378486', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111635101, end_position_on_the_genomic_accession=112083383, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101926924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926924, CurrentID=0, Status='live', Symbol='LINC01375', Aliases='', description='long intergenic non-protein coding RNA 1375', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89915489, end_position_on_the_genomic_accession=89957373, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  651959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651959, CurrentID=0, Status='live', Symbol='LOC651959', Aliases='', description='FSHD region gene 2-like', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101926925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926925, CurrentID=0, Status='live', Symbol='LINC01358', Aliases='', description='long intergenic non-protein coding RNA 1358', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59020476, end_position_on_the_genomic_accession=59088247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378488, CurrentID=0, Status='live', Symbol='LOC105378488', Aliases='', description='uncharacterized LOC105378488', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112950493, end_position_on_the_genomic_accession=112957171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616206, CurrentID=0, Status='live', Symbol='MIR4652', Aliases='', description='microRNA 4652', other_designations='hsa-mir-4652', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93716928, end_position_on_the_genomic_accession=93717005, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616207, CurrentID=0, Status='live', Symbol='MIR4439', Aliases='', description='microRNA 4439', other_designations='hsa-mir-4439', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=225010461, end_position_on_the_genomic_accession=225010540, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926928, CurrentID=0, Status='live', Symbol='CRAT37', Aliases='', description='cervical cancer-associated transcript 37', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91463339, end_position_on_the_genomic_accession=91494850, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616209, CurrentID=0, Status='live', Symbol='MIR4461', Aliases='mir-4461', description='microRNA 4461', other_designations='hsa-mir-4461', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134928039, end_position_on_the_genomic_accession=134928112, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480523, CurrentID=0, Status='live', Symbol='RN7SL650P', Aliases='', description='RNA, 7SL, cytoplasmic 650, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616210, CurrentID=0, Status='live', Symbol='MIR4737', Aliases='', description='microRNA 4737', other_designations='hsa-mir-4737', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60043025, end_position_on_the_genomic_accession=60043105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371935, CurrentID=0, Status='live', Symbol='LOC105371935', Aliases='', description='collagen alpha-4(IV) chain-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60025994, end_position_on_the_genomic_accession=60040475, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616211, CurrentID=0, Status='live', Symbol='MIR4709', Aliases='mir-4709', description='microRNA 4709', other_designations='hsa-mir-4709', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74480133, end_position_on_the_genomic_accession=74480204, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616212, CurrentID=0, Status='live', Symbol='MIR3689F', Aliases='mir-3689f', description='microRNA 3689f', other_designations='hsa-mir-3689f', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134850742, end_position_on_the_genomic_accession=134850807, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616213, CurrentID=0, Status='live', Symbol='MIR4437', Aliases='mir-4437', description='microRNA 4437', other_designations='hsa-mir-4437', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181305593, end_position_on_the_genomic_accession=181305652, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616214, CurrentID=0, Status='live', Symbol='MIR4667', Aliases='mir-4667', description='microRNA 4667', other_designations='hsa-mir-4667', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35608094, end_position_on_the_genomic_accession=35608159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101926935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926935, CurrentID=0, Status='live', Symbol='LOC101926935', Aliases='', description='uncharacterized LOC101926935', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25844667, end_position_on_the_genomic_accession=25854021, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616216, CurrentID=0, Status='live', Symbol='MIR4797', Aliases='', description='microRNA 4797', other_designations='hsa-mir-4797', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197293878, end_position_on_the_genomic_accession=197293948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378499, CurrentID=0, Status='live', Symbol='LOC105378499', Aliases='', description='uncharacterized LOC105378499', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116670186, end_position_on_the_genomic_accession=116672738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926938, CurrentID=0, Status='live', Symbol='RPL12P48', Aliases='', description='ribosomal protein L12 pseudogene 48', other_designations='60S ribosomal protein L12 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37983014, end_position_on_the_genomic_accession=37983838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926940, CurrentID=0, Status='live', Symbol='LOC101926940', Aliases='', description='uncharacterized LOC101926940', other_designations='CTD-2078B5.2', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39520431, end_position_on_the_genomic_accession=39524708, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378503, CurrentID=0, Status='live', Symbol='EMX2OS', Aliases='', description='EMX2 opposite strand/antisense RNA', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117473215, end_position_on_the_genomic_accession=117479160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101926941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926941, CurrentID=0, Status='live', Symbol='LOC101926941', Aliases='', description='uncharacterized LOC101926941', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142325293, end_position_on_the_genomic_accession=142464054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616222, CurrentID=0, Status='live', Symbol='MIR4487', Aliases='mir-4487', description='microRNA 4487', other_designations='hsa-mir-4487', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47400970, end_position_on_the_genomic_accession=47401042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616223, CurrentID=0, Status='live', Symbol='MIR4509-1', Aliases='', description='microRNA 4509-1', other_designations='hsa-mir-4509-1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23197827, end_position_on_the_genomic_accession=23197920, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  400860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400860, CurrentID=0, Status='live', Symbol='FLJ42200', Aliases='', description='FLJ42200 protein', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378506, CurrentID=0, Status='live', Symbol='LOC105378506', Aliases='', description='uncharacterized LOC105378506', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118297667, end_position_on_the_genomic_accession=118302782, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616224, CurrentID=0, Status='live', Symbol='MIR4753', Aliases='mir-4753', description='microRNA 4753', other_designations='hsa-mir-4753', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235190034, end_position_on_the_genomic_accession=235190116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926945, CurrentID=0, Status='live', Symbol='LOC101926945', Aliases='', description='NXPE family member 2-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114512168, end_position_on_the_genomic_accession=114517191, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100380873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380873, CurrentID=0, Status='live', Symbol='FECD3', Aliases='FCD2', description='Corneal dystrophy, Fuchs endothelial, 3', other_designations='', map_location='18q21.2-q21.3', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613267, no_hdr0=''),
  100380874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380874, CurrentID=0, Status='live', Symbol='FECD5', Aliases='FCD3', description='Corneal dystrophy, Fuchs endothelial, 5', other_designations='', map_location='5q33.1-q35.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613269, no_hdr0=''),
  100380875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380875, CurrentID=0, Status='live', Symbol='FECD7', Aliases='FCD4', description='Corneal dystrophy, Fuchs endothelial, 7', other_designations='', map_location='9p24.1-p22.1', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613271, no_hdr0=''),
  100380876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380876, CurrentID=0, Status='live', Symbol='NAFLD1', Aliases='FLD1', description='Fatty liver disease 1, susceptiblity to', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613282, no_hdr0=''),
  100616226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616226, CurrentID=0, Status='live', Symbol='MIR4468', Aliases='', description='microRNA 4468', other_designations='hsa-mir-4468', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138123758, end_position_on_the_genomic_accession=138123821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  104968398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104968398, CurrentID=0, Status='live', Symbol='LOC104968398', Aliases='ARA', description='adriamycin resistance-associated', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104968399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104968399, CurrentID=0, Status='live', Symbol='LOC104968399', Aliases='', description='ncRNA', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3255436, end_position_on_the_genomic_accession=3262359, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616227, CurrentID=0, Status='live', Symbol='MIR4502', Aliases='', description='microRNA 4502', other_designations='hsa-mir-4502', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114273828, end_position_on_the_genomic_accession=114273908, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479501, CurrentID=0, Status='live', Symbol='RN7SL766P', Aliases='', description='RNA, 7SL, cytoplasmic 766, pseudogene', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101926948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926948, CurrentID=0, Status='live', Symbol='LOC101926948', Aliases='', description='uncharacterized LOC101926948', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35646120, end_position_on_the_genomic_accession=35647476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926949, CurrentID=0, Status='live', Symbol='LOC101926949', Aliases='', description='keratinocyte-associated protein 2-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44996437, end_position_on_the_genomic_accession=44996888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378512, CurrentID=0, Status='live', Symbol='LOC105378512', Aliases='', description='uncharacterized LOC105378512', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119595564, end_position_on_the_genomic_accession=119599806, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926950, CurrentID=0, Status='live', Symbol='LINC01570', Aliases='', description='long intergenic non-protein coding RNA 1570', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5601169, end_position_on_the_genomic_accession=5616250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926951, CurrentID=0, Status='live', Symbol='LOC101926951', Aliases='', description='uncharacterized LOC101926951', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61424689, end_position_on_the_genomic_accession=61427946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378514, CurrentID=0, Status='live', Symbol='LOC105378514', Aliases='', description='uncharacterized LOC105378514', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119724999, end_position_on_the_genomic_accession=119726193, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101926952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926952, CurrentID=0, Status='live', Symbol='LOC101926952', Aliases='', description='uncharacterized LOC101926952', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=14991292, end_position_on_the_genomic_accession=14998208, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926953, CurrentID=0, Status='live', Symbol='LOC101926953', Aliases='', description='uncharacterized LOC101926953', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117719900, end_position_on_the_genomic_accession=117793622, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  389844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389844, CurrentID=0, Status='live', Symbol='FTH1P19', Aliases='FTHL19', description='ferritin, heavy polypeptide 1 pseudogene 19', other_designations='ferritin, heavy polypeptide-like 19', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37491957, end_position_on_the_genomic_accession=37492714, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  171391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171391, CurrentID=0, Status='live', Symbol='LOC171391', Aliases='NS3BP', description='uncharacterized LOC171391', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=777578, end_position_on_the_genomic_accession=784297, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271448, CurrentID=0, Status='live', Symbol='RPL21P123', Aliases='RPL21_59_1535', description='ribosomal protein L21 pseudogene 123', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29716280, end_position_on_the_genomic_accession=29716725, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926955, CurrentID=0, Status='live', Symbol='LOC101926955', Aliases='', description='uncharacterized LOC101926955', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25955812, end_position_on_the_genomic_accession=25969288, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371940, CurrentID=0, Status='live', Symbol='LOC105371940', Aliases='', description='uncharacterized LOC105371940', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82425894, end_position_on_the_genomic_accession=82430428, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101926956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926956, CurrentID=0, Status='live', Symbol='LOC101926956', Aliases='', description='uncharacterized LOC101926956', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92668656, end_position_on_the_genomic_accession=92683445, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378519, CurrentID=0, Status='live', Symbol='LOC105378519', Aliases='', description='uncharacterized LOC105378519', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120902268, end_position_on_the_genomic_accession=120980750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616237, CurrentID=0, Status='live', Symbol='MIR4640', Aliases='', description='microRNA 4640', other_designations='hsa-mir-4640', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30890883, end_position_on_the_genomic_accession=30890972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646616, CurrentID=0, Status='live', Symbol='LOC646616', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32522740, end_position_on_the_genomic_accession=32524086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100380945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100380945, CurrentID=0, Status='live', Symbol='ZNF848P', Aliases='', description='zinc finger protein 848, pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65176428, end_position_on_the_genomic_accession=65177524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926958, CurrentID=0, Status='live', Symbol='LOC101926958', Aliases='', description='uncharacterized LOC101926958', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21386700, end_position_on_the_genomic_accession=21404921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378521, CurrentID=0, Status='live', Symbol='LOC105378521', Aliases='', description='uncharacterized LOC105378521', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121037805, end_position_on_the_genomic_accession=121077465, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101926959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926959, CurrentID=0, Status='live', Symbol='LOC101926959', Aliases='', description='uncharacterized LOC101926959', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95665651, end_position_on_the_genomic_accession=95668727, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926960, CurrentID=0, Status='live', Symbol='LOC101926960', Aliases='', description='uncharacterized LOC101926960', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42155833, end_position_on_the_genomic_accession=42175345, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378523, CurrentID=0, Status='live', Symbol='LOC105378523', Aliases='', description='uncharacterized LOC105378523', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121291288, end_position_on_the_genomic_accession=121322945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371941, CurrentID=0, Status='live', Symbol='LOC105371941', Aliases='', description='uncharacterized LOC105371941', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82454374, end_position_on_the_genomic_accession=82458035, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378524, CurrentID=0, Status='live', Symbol='LOC105378524', Aliases='', description='uncharacterized LOC105378524', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122417975, end_position_on_the_genomic_accession=122423094, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926962, CurrentID=0, Status='live', Symbol='LOC101926962', Aliases='', description='uncharacterized LOC101926962', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46903471, end_position_on_the_genomic_accession=46910029, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926963, CurrentID=0, Status='live', Symbol='LOC101926963', Aliases='', description='uncharacterized LOC101926963', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=602845, end_position_on_the_genomic_accession=608502, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101926964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926964, CurrentID=0, Status='live', Symbol='LOC101926964', Aliases='', description='uncharacterized LOC101926964', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60659631, end_position_on_the_genomic_accession=60825584, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101926965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926965, CurrentID=0, Status='live', Symbol='LOC101926965', Aliases='', description='uncharacterized LOC101926965', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74267170, end_position_on_the_genomic_accession=74269687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378528, CurrentID=0, Status='live', Symbol='LOC105378528', Aliases='', description='uncharacterized LOC105378528', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122588147, end_position_on_the_genomic_accession=122600983, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101926966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926966, CurrentID=0, Status='live', Symbol='LOC101926966', Aliases='', description='uncharacterized LOC101926966', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15690782, end_position_on_the_genomic_accession=15718961, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378529, CurrentID=0, Status='live', Symbol='LOC105378529', Aliases='', description='uncharacterized LOC105378529', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123558625, end_position_on_the_genomic_accession=123562220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926967, CurrentID=0, Status='live', Symbol='LOC101926967', Aliases='', description='uncharacterized LOC101926967', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44198843, end_position_on_the_genomic_accession=44223707, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  373071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373071, CurrentID=0, Status='live', Symbol='AIS1', Aliases='', description='autoimmune susceptibility 1', other_designations='', map_location='1p31', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607836, no_hdr0=''),
  373073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373073, CurrentID=0, Status='live', Symbol='CLAM', Aliases='PCH3', description='cerebellar atrophy with progressive microcephaly', other_designations='', map_location='7q11-q21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608027, no_hdr0=''),
  101926969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926969, CurrentID=0, Status='live', Symbol='LOC101926969', Aliases='', description='uncharacterized LOC101926969', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91571575, end_position_on_the_genomic_accession=91580863, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378532, CurrentID=0, Status='live', Symbol='LOC105378532', Aliases='', description='uncharacterized LOC105378532', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123702976, end_position_on_the_genomic_accession=123713482, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100288570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288570, CurrentID=0, Status='live', Symbol='LOC100288570', Aliases='', description='glycosylphosphatidylinositol anchor attachment protein 1 homolog (yeast) pseudogene', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110402917, end_position_on_the_genomic_accession=110423816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100616251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616251, CurrentID=0, Status='live', Symbol='MIR1587', Aliases='', description='microRNA 1587', other_designations='hsa-mir-1587', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39837561, end_position_on_the_genomic_accession=39837613, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100419644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419644, CurrentID=0, Status='live', Symbol='LOC100419644', Aliases='', description='sorbitol dehydrogenase pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108720608, end_position_on_the_genomic_accession=108721603, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  504175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504175, CurrentID=0, Status='live', Symbol='OR3D1P', Aliases='', description='olfactory receptor family 3 subfamily D member 1 pseudogene', other_designations='olfactor receptor, family 3, subfamily D, member 1 pseudogene|olfactory receptor, family 3, subfamily D, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247666882, end_position_on_the_genomic_accession=247667817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  504180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504180, CurrentID=0, Status='live', Symbol='DEFB105B', Aliases='BD-5, DEFB-5', description='defensin beta 105B', other_designations='beta-defensin 5|defensin, beta 105|defensin, beta 105B', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7487668, end_position_on_the_genomic_accession=7490779, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  504181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504181, CurrentID=0, Status='live', Symbol='OR5BN2P', Aliases='', description='olfactory receptor family 5 subfamily BN member 2 pseudogene', other_designations='olfactory receptor, family 5, subfamily BN, member 2 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56116638, end_position_on_the_genomic_accession=56117255, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  504185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504185, CurrentID=0, Status='live', Symbol='DEFB108P3', Aliases='', description='defensin beta 108 pseudogene 3', other_designations='defensin, beta 108, pseudogene 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12095176, end_position_on_the_genomic_accession=12099535, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926975, CurrentID=0, Status='live', Symbol='LOC101926975', Aliases='', description='uncharacterized LOC101926975', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142745600, end_position_on_the_genomic_accession=142760993, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  504189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504189, CurrentID=0, Status='live', Symbol='OR8U8', Aliases='', description='olfactory receptor family 8 subfamily U member 8', other_designations='olfactory receptor, family 8, subfamily U, member 8', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  504190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504190, CurrentID=0, Status='live', Symbol='OR8U9', Aliases='', description='olfactory receptor family 8 subfamily U member 9', other_designations='olfactory receptor, family 8, subfamily U, member 9', map_location='11q1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  504191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=504191, CurrentID=0, Status='live', Symbol='OR9G9', Aliases='', description='olfactory receptor family 9 subfamily G member 9', other_designations='olfactory receptor, family 9, subfamily G, member 9', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616257, CurrentID=0, Status='live', Symbol='MIR3975', Aliases='', description='microRNA 3975', other_designations='hsa-mir-3975', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35591737, end_position_on_the_genomic_accession=35591806, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616258, CurrentID=0, Status='live', Symbol='MIR4516', Aliases='mir-4516', description='microRNA 4516', other_designations='hsa-mir-4516', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2133119, end_position_on_the_genomic_accession=2133204, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373540, CurrentID=0, Status='live', Symbol='LOC105373540', Aliases='', description='uncharacterized LOC105373540', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107888299, end_position_on_the_genomic_accession=107895228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271449, CurrentID=0, Status='live', Symbol='RPL21P126', Aliases='RPL21_59_1605', description='ribosomal protein L21 pseudogene 126', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55635794, end_position_on_the_genomic_accession=55636346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101926980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926980, CurrentID=0, Status='live', Symbol='LOC101926980', Aliases='', description='uncharacterized LOC101926980', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223992709, end_position_on_the_genomic_accession=224004864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  373156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373156, CurrentID=0, Status='live', Symbol='GSTK1', Aliases='GST, GST 13-13, GST13, GST13-13-1, hGSTK1, GSTK1', description='glutathione S-transferase kappa 1', other_designations='GST class-kappa|glutathione S-transferase k1|glutathione S-transferase subunit 13 homolog', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143263429, end_position_on_the_genomic_accession=143269129, orientation='plus', exon_count=8, OMIM=602321, no_hdr0=''),
  101926982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926982, CurrentID=0, Status='live', Symbol='LOC101926982', Aliases='', description='protein GVQW1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9649308, end_position_on_the_genomic_accession=9665708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  373159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373159, CurrentID=0, Status='live', Symbol='STK19B', Aliases='RP2, STK19P, bPG116M5.10', description='serine/threonine kinase 19B, pseudogene', other_designations='serine/threonine kinase 19 pseudogene', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32013390, end_position_on_the_genomic_accession=32014184, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100118954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100118954, CurrentID=0, Status='live', Symbol='NRBF2P5', Aliases='', description='nuclear receptor binding factor 2 pseudogene 5', other_designations='', map_location='10p', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5775834, end_position_on_the_genomic_accession=5778774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376310, CurrentID=0, Status='live', Symbol='LOC105376310', Aliases='', description='uncharacterized LOC105376310', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134293930, end_position_on_the_genomic_accession=134295491, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378546, CurrentID=0, Status='live', Symbol='LOC105378546', Aliases='', description='uncharacterized LOC105378546', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125544088, end_position_on_the_genomic_accession=125555556, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100288584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288584, CurrentID=0, Status='live', Symbol='CCDC34P1', Aliases='', description='coiled-coil domain containing 34 pseudogene 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112068006, end_position_on_the_genomic_accession=112068528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378547, CurrentID=0, Status='live', Symbol='LOC105378547', Aliases='', description='uncharacterized LOC105378547', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125567331, end_position_on_the_genomic_accession=125574268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270679, CurrentID=0, Status='live', Symbol='LINC01432', Aliases='', description='long intergenic non-protein coding RNA 1432', other_designations='', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22054090, end_position_on_the_genomic_accession=22074654, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926985, CurrentID=0, Status='live', Symbol='LOC101926985', Aliases='', description='uncharacterized LOC101926985', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42569856, end_position_on_the_genomic_accession=42579544, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  171423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=171423, CurrentID=0, Status='live', Symbol='PDIA3P1', Aliases='ERp60, GRP58P, PDIA3P', description='protein disulfide isomerase family A member 3 pseudogene 1', other_designations='ER-60 protein processed pseudogene|glucose regulated protein, 58kDa pseudogene|protein disulfide isomerase family A, member 3 pseudogene 1|protein disulfide isomerase-associated 3 pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147177851, end_position_on_the_genomic_accession=147179949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101926986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926986, CurrentID=0, Status='live', Symbol='LOC101926986', Aliases='', description='glioma tumor suppressor candidate region gene 1 protein-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197632643, end_position_on_the_genomic_accession=197641357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378549, CurrentID=0, Status='live', Symbol='LOC105378549', Aliases='', description='uncharacterized LOC105378549', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126496847, end_position_on_the_genomic_accession=126503185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101926987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926987, CurrentID=0, Status='live', Symbol='DLGAP4-AS1', Aliases='', description='DLGAP4 antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36507702, end_position_on_the_genomic_accession=36573275, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101926989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926989, CurrentID=0, Status='live', Symbol='LOC101926989', Aliases='', description='uncharacterized LOC101926989', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36470728, end_position_on_the_genomic_accession=36472139, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480535, CurrentID=0, Status='live', Symbol='RN7SL820P', Aliases='', description='RNA, 7SL, cytoplasmic 820, pseudogene', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371947, CurrentID=0, Status='live', Symbol='LOC105371947', Aliases='', description='uncharacterized LOC105371947', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371897, CurrentID=0, Status='live', Symbol='LOC105371897', Aliases='', description='uncharacterized LOC105371897', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76376661, end_position_on_the_genomic_accession=76380066, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376312, CurrentID=0, Status='live', Symbol='LOC105376312', Aliases='', description='uncharacterized LOC105376312', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134648116, end_position_on_the_genomic_accession=134649288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378556, CurrentID=0, Status='live', Symbol='LOC105378556', Aliases='', description='uncharacterized LOC105378556', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129366283, end_position_on_the_genomic_accession=129368689, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101926994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926994, CurrentID=0, Status='live', Symbol='LOC101926994', Aliases='', description='uncharacterized LOC101926994', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401400, CurrentID=0, Status='live', Symbol='LOC401400', Aliases='', description='uncharacterized LOC401400', other_designations='', map_location='7q32.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288595, CurrentID=0, Status='live', Symbol='DYZ1L26', Aliases='', description='DYZ1 repeat sequence', other_designations='', map_location='Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=58911101, end_position_on_the_genomic_accession=58912517, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101926996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926996, CurrentID=0, Status='live', Symbol='RUNDC3A-AS1', Aliases='', description='RUNDC3A antisense RNA 1', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44299574, end_position_on_the_genomic_accession=44308431, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100381178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381178, CurrentID=0, Status='live', Symbol='TMPRSS11CP', Aliases='', description='transmembrane protease, serine 11C, pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67765914, end_position_on_the_genomic_accession=67774454, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101926997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926997, CurrentID=0, Status='live', Symbol='LOC101926997', Aliases='', description='uncharacterized LOC101926997', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2517250, end_position_on_the_genomic_accession=2519260, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616278, CurrentID=0, Status='live', Symbol='MIR4540', Aliases='', description='microRNA 4540', other_designations='hsa-mir-4540', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36864254, end_position_on_the_genomic_accession=36864308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616279, CurrentID=0, Status='live', Symbol='MIR3974', Aliases='', description='microRNA 3974', other_designations='hsa-mir-3974', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=17673299, end_position_on_the_genomic_accession=17673394, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616280, CurrentID=0, Status='live', Symbol='MIR4503', Aliases='', description='microRNA 4503', other_designations='hsa-mir-4503', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36952309, end_position_on_the_genomic_accession=36952391, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100381202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381202, CurrentID=0, Status='live', Symbol='NEK2P4', Aliases='', description='NEK2 pseudogene 4', other_designations='NIMA (never in mitosis gene a)-related kinase 2 pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131177418, end_position_on_the_genomic_accession=131179088, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100381204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381204, CurrentID=0, Status='live', Symbol='HCHGQ1', Aliases='', description='Hematocrit/hemoglobin quantitative trait locus 1', other_designations='', map_location='6q23', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609319, no_hdr0=''),
  100381205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381205, CurrentID=0, Status='live', Symbol='HCHGQ2', Aliases='', description='Hematocrit/hemoglobin quantitative trait locus 2', other_designations='', map_location='9q', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609320, no_hdr0=''),
  100381206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381206, CurrentID=0, Status='live', Symbol='HCHGQ3', Aliases='', description='Hematocrit/hemoglobin quantitative trait locus 3', other_designations='', map_location='22q12-q13', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613284, no_hdr0=''),
  100381207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381207, CurrentID=0, Status='live', Symbol='CIHL', Aliases='', description='hearing loss, cisplatin-induced, susceptibility to', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613290, no_hdr0=''),
  100381208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381208, CurrentID=0, Status='live', Symbol='RNY5P2', Aliases='', description='RNA, Ro-associated Y5 pseudogene 2', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197543376, end_position_on_the_genomic_accession=197543435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100381210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381210, CurrentID=0, Status='live', Symbol='CAMPD1', Aliases='', description='Camptodactyly 1', other_designations='', map_location='3q11.2-q13.12', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=114200, no_hdr0=''),
  100381211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381211, CurrentID=0, Status='live', Symbol='ARCODS', Aliases='', description='Ariculocondylar syndrome', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=602483, no_hdr0=''),
  100616282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616282, CurrentID=0, Status='live', Symbol='MIR4738', Aliases='', description='microRNA 4738', other_designations='hsa-mir-4738', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75784521, end_position_on_the_genomic_accession=75784607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100381215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381215, CurrentID=0, Status='live', Symbol='DYNLT3P2', Aliases='', description='dynein, light chain, Tctex-type 3 pseudogene 2', other_designations='dynein light chain Tctex-type 3 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128199901, end_position_on_the_genomic_accession=128201423, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616283, CurrentID=0, Status='live', Symbol='MIR4766', Aliases='mir-4766', description='microRNA 4766', other_designations='hsa-mir-4766', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40813883, end_position_on_the_genomic_accession=40813958, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616284, CurrentID=0, Status='live', Symbol='MIR4489', Aliases='mir-4489', description='microRNA 4489', other_designations='hsa-mir-4489', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65649192, end_position_on_the_genomic_accession=65649253, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378567, CurrentID=0, Status='live', Symbol='LOC105378567', Aliases='', description='uncharacterized LOC105378567', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132178445, end_position_on_the_genomic_accession=132184283, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  221927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=221927, CurrentID=0, Status='live', Symbol='BRAT1', Aliases='BAAT1, C7orf27, RMFSL', description='BRCA1-associated ATM activator 1', other_designations='BRCA1-associated protein required for ATM activation protein 1|HEAT repeat-containing protein C7orf27', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2537810, end_position_on_the_genomic_accession=2555758, orientation='minus', exon_count=17, OMIM=614506, no_hdr0=''),
  100616286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616286, CurrentID=0, Status='live', Symbol='MIR4676', Aliases='', description='microRNA 4676', other_designations='hsa-mir-4676', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72721029, end_position_on_the_genomic_accession=72721100, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616287, CurrentID=0, Status='live', Symbol='MIR4495', Aliases='', description='microRNA 4495', other_designations='hsa-mir-4495', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97939056, end_position_on_the_genomic_accession=97939121, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378570, CurrentID=0, Status='live', Symbol='LOC105378570', Aliases='', description='uncharacterized LOC105378570', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132455089, end_position_on_the_genomic_accession=132464024, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100616288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616288, CurrentID=0, Status='live', Symbol='MIR4665', Aliases='mir-4665', description='microRNA 4665', other_designations='hsa-mir-4665', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6007826, end_position_on_the_genomic_accession=6007904, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928344, CurrentID=0, Status='live', Symbol='LOC101928344', Aliases='', description='protein GVQW1-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=204025, end_position_on_the_genomic_accession=246533, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  100616289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616289, CurrentID=0, Status='live', Symbol='MIR4475', Aliases='', description='microRNA 4475', other_designations='hsa-mir-4475', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36823539, end_position_on_the_genomic_accession=36823599, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102216266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102216266, CurrentID=0, Status='live', Symbol='DEL3Q13.31', Aliases='C13DELq13.31', description='Chromosome 3q13.31 deletion syndrome', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615433, no_hdr0=''),
  102216268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102216268, CurrentID=0, Status='live', Symbol='CISTR', Aliases='CISTR-ACTACT, re52431, CISTR', description='chondrogenesis-associated transcript', other_designations='chondrogenic regulator lncRNA|cis- and trans-chromosomal chondrogenic regulator transcript', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53750447, end_position_on_the_genomic_accession=53757034, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100616290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616290, CurrentID=0, Status='live', Symbol='MIR4529', Aliases='', description='microRNA 4529', other_designations='hsa-mir-4529', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55479221, end_position_on_the_genomic_accession=55479298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378573, CurrentID=0, Status='live', Symbol='LOC105378573', Aliases='', description='uncharacterized LOC105378573', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133202193, end_position_on_the_genomic_accession=133218824, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100616291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616291, CurrentID=0, Status='live', Symbol='MIR4791', Aliases='mir-4791', description='microRNA 4791', other_designations='hsa-mir-4791', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19314848, end_position_on_the_genomic_accession=19314931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100381270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100381270, CurrentID=0, Status='live', Symbol='ZBED6', Aliases='MGR', description='zinc finger, BED-type containing 6', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203797523, end_position_on_the_genomic_accession=203800462, orientation='plus', exon_count=1, OMIM=613512, no_hdr0=''),
  100616292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616292, CurrentID=0, Status='live', Symbol='MIR4690', Aliases='', description='microRNA 4690', other_designations='hsa-mir-4690', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65636310, end_position_on_the_genomic_accession=65636369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378575, CurrentID=0, Status='live', Symbol='LOC105378575', Aliases='', description='uncharacterized LOC105378575', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133491452, end_position_on_the_genomic_accession=133514259, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652960, CurrentID=0, Status='live', Symbol='LOC100652960', Aliases='', description='SUZ RNA binding domain containing 1 pseudogene', other_designations='UPF0485 protein C1orf144 homolog', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86432247, end_position_on_the_genomic_accession=86432700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616293, CurrentID=0, Status='live', Symbol='MIR4510', Aliases='mir-4510', description='microRNA 4510', other_designations='hsa-mir-4510', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35926856, end_position_on_the_genomic_accession=35926923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100507067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507067, CurrentID=0, Status='live', Symbol='GOLGA8UP', Aliases='', description='golgin A8 family member U, pseudogene', other_designations='golgin A8 family, member U, pseudogene|golgin subfamily A member 8H-like', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=31087883, end_position_on_the_genomic_accession=31092183, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616294, CurrentID=0, Status='live', Symbol='MIR4740', Aliases='', description='microRNA 4740', other_designations='hsa-mir-4740', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81400716, end_position_on_the_genomic_accession=81400778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616295, CurrentID=0, Status='live', Symbol='MIR4764', Aliases='', description='microRNA 4764', other_designations='hsa-mir-4764', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33436582, end_position_on_the_genomic_accession=33436669, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378578, CurrentID=0, Status='live', Symbol='LOC105378578', Aliases='', description='histone-lysine N-methyltransferase SETMAR-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85148131, end_position_on_the_genomic_accession=85156883, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616296, CurrentID=0, Status='live', Symbol='MIR4678', Aliases='', description='microRNA 4678', other_designations='hsa-mir-4678', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87503881, end_position_on_the_genomic_accession=87503954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616297, CurrentID=0, Status='live', Symbol='MIR3977', Aliases='', description='microRNA 3977', other_designations='hsa-mir-3977', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82840155, end_position_on_the_genomic_accession=82840224, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389908, CurrentID=0, Status='live', Symbol='LOC389908', Aliases='', description='uncharacterized LOC389908', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616298, CurrentID=0, Status='live', Symbol='MIR4419B', Aliases='', description='microRNA 4419b', other_designations='hsa-mir-4419b', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128244506, end_position_on_the_genomic_accession=128244573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616299, CurrentID=0, Status='live', Symbol='MIR4450', Aliases='mir-4450', description='microRNA 4450', other_designations='hsa-mir-4450', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76573568, end_position_on_the_genomic_accession=76573632, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480541, CurrentID=0, Status='live', Symbol='RNU6ATAC20P', Aliases='', description='RNA, U6atac small nuclear 20, pseudogene', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616300, CurrentID=0, Status='live', Symbol='MIR4710', Aliases='mir-4710', description='microRNA 4710', other_designations='hsa-mir-4710', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104677694, end_position_on_the_genomic_accession=104677749, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422587, CurrentID=0, Status='live', Symbol='LOC100422587', Aliases='', description='platelet activating factor acetylhydrolase 1b regulatory subunit 1 pseudogene', other_designations='platelet-activating factor acetylhydrolase 1b, regulatory subunit 1 (45kDa) pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63797083, end_position_on_the_genomic_accession=63798131, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616301, CurrentID=0, Status='live', Symbol='MIR4674', Aliases='', description='microRNA 4674', other_designations='hsa-mir-4674', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136546173, end_position_on_the_genomic_accession=136546259, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616302, CurrentID=0, Status='live', Symbol='MIR548X2', Aliases='', description='microRNA 548x-2', other_designations='hsa-mir-548x-2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65966330, end_position_on_the_genomic_accession=65966429, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102216342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102216342, CurrentID=0, Status='live', Symbol='LINC01108', Aliases='LncRNA-ES1', description='long intergenic non-protein coding RNA 1108', other_designations='lncrna embryonic stem cell 1', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14280127, end_position_on_the_genomic_accession=14285454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106479504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479504, CurrentID=0, Status='live', Symbol='RN7SL778P', Aliases='', description='RNA, 7SL, cytoplasmic 778, pseudogene', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616303, CurrentID=0, Status='live', Symbol='MIR3688-2', Aliases='mir-3688-2', description='microRNA 3688-2', other_designations='hsa-mir-3688-2', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=159128805, end_position_on_the_genomic_accession=159128891, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954204, CurrentID=0, Status='live', Symbol='LINC00644', Aliases='', description='long intergenic non-protein coding RNA 644', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62134149, end_position_on_the_genomic_accession=62139973, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616304, CurrentID=0, Status='live', Symbol='MIR4499', Aliases='', description='microRNA 4499', other_designations='hsa-mir-4499', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20433778, end_position_on_the_genomic_accession=20433846, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928837, CurrentID=0, Status='live', Symbol='LOC101928837', Aliases='', description='uncharacterized LOC101928837', other_designations='', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76782581, end_position_on_the_genomic_accession=76783062, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616305, CurrentID=0, Status='live', Symbol='MIR548AE1', Aliases='', description='microRNA 548ae-1', other_designations='hsa-mir-548ae-1', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184378975, end_position_on_the_genomic_accession=184379044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927026, CurrentID=0, Status='live', Symbol='LOC101927026', Aliases='', description='uncharacterized LOC101927026', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9184664, end_position_on_the_genomic_accession=9657268, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100616307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616307, CurrentID=0, Status='live', Symbol='MIR4757', Aliases='', description='microRNA 4757', other_designations='hsa-mir-4757', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19348429, end_position_on_the_genomic_accession=19348505, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616308, CurrentID=0, Status='live', Symbol='MIR4666A', Aliases='MIR4666, hsa-mir-4666a', description='microRNA 4666a', other_designations='hsa-mir-4666|microRNA 4666', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228462074, end_position_on_the_genomic_accession=228462152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616309, CurrentID=0, Status='live', Symbol='MIR4472-2', Aliases='', description='microRNA 4472-2', other_designations='hsa-mir-4472-2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116428252, end_position_on_the_genomic_accession=116428318, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271451, CurrentID=0, Status='live', Symbol='RPL21P129', Aliases='RPL21_60_1619', description='ribosomal protein L21 pseudogene 129', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7750644, end_position_on_the_genomic_accession=7751190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84086, CurrentID=0, Status='live', Symbol='IGLCOR22-2', Aliases='IGLC/OR22-2', description='immunoglobulin lambda constant/OR22-2 (pseudogene)', other_designations='immunoglobulin lambda constant/OR22-2 pseudogene', map_location='22q12.2-q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371955, CurrentID=0, Status='live', Symbol='LOC105371955', Aliases='', description='uncharacterized LOC105371955', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2054823, end_position_on_the_genomic_accession=2061617, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  84087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84087, CurrentID=0, Status='live', Symbol='IGLVIVOR22-2', Aliases='', description='immunoglobulin lambda variable (IV)/OR22-2 (pseudogene)', other_designations='IGLV(IV)/OR22-2|immunoglobulin lambda variable (IV)/OR22-2 pseudogene', map_location='22q12.2-q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374415, CurrentID=0, Status='live', Symbol='LOC105374415', Aliases='', description='uncharacterized LOC105374415', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38718391, end_position_on_the_genomic_accession=38729124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84088, CurrentID=0, Status='live', Symbol='IGLVIVOR22-1', Aliases='', description='immunoglobulin lambda variable (IV)/OR22-1 (pseudogene)', other_designations='IGLV(IV)/OR22-1|immunoglobulin lambda variable (IV)/OR22-1 pseudogene', map_location='22q11.2-q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616313, CurrentID=0, Status='live', Symbol='MIR4749', Aliases='mir-4749', description='microRNA 4749', other_designations='hsa-mir-4749', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49854591, end_position_on_the_genomic_accession=49854651, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954264, CurrentID=0, Status='live', Symbol='RNVU1-4', Aliases='RNU1-102, RNU1-50, RNVU1-5, vU1.4, vU1.5', description='RNA, variant U1 small nuclear 4', other_designations='RNA, variant U1 small nuclear 5', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120942600, end_position_on_the_genomic_accession=120942763, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954266, CurrentID=0, Status='live', Symbol='RNVU1-14', Aliases='RNU1-37, vU1.14', description='RNA, variant U1 small nuclear 14', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145281116, end_position_on_the_genomic_accession=145281279, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954267, CurrentID=0, Status='live', Symbol='RNVU1-15', Aliases='RNU1-121, RNU1-66, RNVU1-16, vU1.15, vU1.16', description='RNA, variant U1 small nuclear 15', other_designations='RNA, variant U1 small nuclear 16', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144412576, end_position_on_the_genomic_accession=144412740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954268, CurrentID=0, Status='live', Symbol='RNVU1-20', Aliases='RNU1-110, vU1.20', description='RNA, variant U1 small nuclear 20', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101954269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954269, CurrentID=0, Status='live', Symbol='RNVU1-17', Aliases='RNU1-127, vU1.17', description='RNA, variant U1 small nuclear 17', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143699456, end_position_on_the_genomic_accession=143699619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101954271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954271, CurrentID=0, Status='live', Symbol='RNU6-9', Aliases='U6-9', description='RNA, U6 small nuclear 9', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=893484, end_position_on_the_genomic_accession=893590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954272, CurrentID=0, Status='live', Symbol='RNVU1-3', Aliases='RNU1-113, RNU1-151, RNVU1-12, vU1.12, vU1.3', description='RNA, variant U1 small nuclear 3', other_designations='RNA, variant U1 small nuclear 12', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148402715, end_position_on_the_genomic_accession=148402875, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954273, CurrentID=0, Status='live', Symbol='RNVU1-1', Aliases='RNU1-10P, RNU1-53, RNU1P100, U1.4, U1P14, vU1.1, vU1.10, RNVU1-1', description='RNA, variant U1 small nuclear 1', other_designations='RNA, variant U1 small nuclear 10', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148362370, end_position_on_the_genomic_accession=148362533, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616315, CurrentID=0, Status='live', Symbol='MIR4781', Aliases='mir-4781', description='microRNA 4781', other_designations='hsa-mir-4781', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54054079, end_position_on_the_genomic_accession=54054154, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101954276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954276, CurrentID=0, Status='live', Symbol='RNVU1-6', Aliases='RNU1-99, vU1.6', description='RNA, variant U1 small nuclear 6', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146052081, end_position_on_the_genomic_accession=146052244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101954277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954277, CurrentID=0, Status='live', Symbol='RNVU1-19', Aliases='RNU1-126, RNU1-147, RNVU1-13, vU1.13, vU1.19', description='RNA, variant U1 small nuclear 19', other_designations='RNA, variant U1 small nuclear 13', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120850819, end_position_on_the_genomic_accession=120850985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101954278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101954278, CurrentID=0, Status='live', Symbol='RNU6-8', Aliases='U6-8', description='RNA, U6 small nuclear 8', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32203163, end_position_on_the_genomic_accession=32203269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288637, CurrentID=0, Status='live', Symbol='LOC100288637', Aliases='', description='OTU deubiquitinase 7A pseudogene', other_designations='OTU domain containing 7A pseudogene', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30646115, end_position_on_the_genomic_accession=30773009, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100419844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419844, CurrentID=0, Status='live', Symbol='LOC100419844', Aliases='', description='zinc finger protein 610 pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52440094, end_position_on_the_genomic_accession=52442021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927038, CurrentID=0, Status='live', Symbol='LOC101927038', Aliases='', description='uncharacterized LOC101927038', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41764150, end_position_on_the_genomic_accession=41765571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928350, CurrentID=0, Status='live', Symbol='LOC101928350', Aliases='', description='uncharacterized LOC101928350', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11857183, end_position_on_the_genomic_accession=11881269, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927039, CurrentID=0, Status='live', Symbol='LOC101927039', Aliases='', description='uncharacterized LOC101927039', other_designations='KB-1043D8.6', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96235628, end_position_on_the_genomic_accession=96238031, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84096, CurrentID=0, Status='live', Symbol='IGLCOR22-1', Aliases='IGLC/OR22-1', description='immunoglobulin lambda constant/OR22-1 (pseudogene)', other_designations='immunoglobulin lambda constant/OR22-1 pseudogene', map_location='22q12.2-q12.3', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  373509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373509, CurrentID=0, Status='live', Symbol='USP50', Aliases='', description='ubiquitin specific peptidase 50', other_designations='inactive ubiquitin-specific peptidase 50|ubiquitin specific protease 50', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50493681, end_position_on_the_genomic_accession=50547439, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  84097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84097, CurrentID=0, Status='live', Symbol='IGLVIV-66-1', Aliases='', description='immunoglobulin lambda variable (IV)-66-1 (pseudogene)', other_designations='IGLV(IV)-66-1|immunoglobulin lambda variable (IV)-66-1 pseudogene', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22058479, end_position_on_the_genomic_accession=22058805, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378604, CurrentID=0, Status='live', Symbol='LOC105378604', Aliases='', description='uncharacterized LOC105378604', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3306636, end_position_on_the_genomic_accession=3323585, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927042, CurrentID=0, Status='live', Symbol='LOC101927042', Aliases='', description='uncharacterized LOC101927042', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38949647, end_position_on_the_genomic_accession=38998503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84099, CurrentID=0, Status='live', Symbol='ID2B', Aliases='', description='inhibitor of DNA binding 2B, dominant negative helix-loop-helix protein (pseudogene)', other_designations='striated muscle contraction regulatory protein', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62123423, end_position_on_the_genomic_accession=62124718, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928351, CurrentID=0, Status='live', Symbol='LOC101928351', Aliases='', description='uncharacterized LOC101928351', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159799101, end_position_on_the_genomic_accession=159824599, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616324, CurrentID=0, Status='live', Symbol='MIR1245B', Aliases='', description='microRNA 1245b', other_designations='hsa-mir-1245b', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188978093, end_position_on_the_genomic_accession=188978161, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378607, CurrentID=0, Status='live', Symbol='LOC105378607', Aliases='', description='uncharacterized LOC105378607', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3602893, end_position_on_the_genomic_accession=3607681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616325, CurrentID=0, Status='live', Symbol='MIR4460', Aliases='', description='microRNA 4460', other_designations='hsa-mir-4460', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129397062, end_position_on_the_genomic_accession=129397147, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616326, CurrentID=0, Status='live', Symbol='MIR4636', Aliases='mir-4636', description='microRNA 4636', other_designations='hsa-mir-4636', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9053816, end_position_on_the_genomic_accession=9053895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616327, CurrentID=0, Status='live', Symbol='MIR4484', Aliases='mir-4484', description='microRNA 4484', other_designations='hsa-mir-4484', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125819740, end_position_on_the_genomic_accession=125819822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106479505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479505, CurrentID=0, Status='live', Symbol='RN7SL783P', Aliases='', description='RNA, 7SL, cytoplasmic 783, pseudogene', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927048, CurrentID=0, Status='live', Symbol='LOC101927048', Aliases='', description='uncharacterized LOC101927048', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49815493, end_position_on_the_genomic_accession=49824766, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378611, CurrentID=0, Status='live', Symbol='LOC105378611', Aliases='', description='uncharacterized LOC105378611', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4582293, end_position_on_the_genomic_accession=4587722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616329, CurrentID=0, Status='live', Symbol='MIR4700', Aliases='mir-4700', description='microRNA 4700', other_designations='hsa-mir-4700', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120723193, end_position_on_the_genomic_accession=120723266, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616330, CurrentID=0, Status='live', Symbol='MIR4491', Aliases='mir-4491', description='microRNA 4491', other_designations='hsa-mir-4491', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111347757, end_position_on_the_genomic_accession=111347824, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378613, CurrentID=0, Status='live', Symbol='LOC105378613', Aliases='', description='uncharacterized LOC105378613', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19438076, end_position_on_the_genomic_accession=19447290, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371959, CurrentID=0, Status='live', Symbol='LOC105371959', Aliases='', description='uncharacterized LOC105371959', other_designations='CTD-2015H3.1', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1779902, end_position_on_the_genomic_accession=1782253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422593, CurrentID=0, Status='live', Symbol='LOC100422593', Aliases='', description='DCN1, defective in cullin neddylation 1, domain containing 3 (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178909710, end_position_on_the_genomic_accession=178910605, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616331, CurrentID=0, Status='live', Symbol='MIR4650-2', Aliases='', description='microRNA 4650-2', other_designations='hsa-mir-4650-2', map_location='7q11.22', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72697903, end_position_on_the_genomic_accession=72697978, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616332, CurrentID=0, Status='live', Symbol='MIR4716', Aliases='mir-4716', description='microRNA 4716', other_designations='hsa-mir-4716', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49169070, end_position_on_the_genomic_accession=49169153, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616333, CurrentID=0, Status='live', Symbol='MIR3689C', Aliases='', description='microRNA 3689c', other_designations='hsa-mir-3689c', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134849298, end_position_on_the_genomic_accession=134849369, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100288654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288654, CurrentID=0, Status='live', Symbol='NDUFB4P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B4 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 4, 15kDa pseudogene 1', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67513877, end_position_on_the_genomic_accession=67514377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378617, CurrentID=0, Status='live', Symbol='LOC105378617', Aliases='', description='uncharacterized LOC105378617', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30160482, end_position_on_the_genomic_accession=30199957, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616335, CurrentID=0, Status='live', Symbol='MIR219B', Aliases='MIR2964A, hsa-mir-219b, mir-219b', description='microRNA 219b', other_designations='hsa-mir-2964a|microRNA 2964a', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128392621, end_position_on_the_genomic_accession=128392708, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371960, CurrentID=0, Status='live', Symbol='LOC105371960', Aliases='', description='uncharacterized LOC105371960', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1513962, end_position_on_the_genomic_accession=1522174, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927056, CurrentID=0, Status='live', Symbol='LOC101927056', Aliases='', description='uncharacterized LOC101927056', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125827237, end_position_on_the_genomic_accession=125886077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927057, CurrentID=0, Status='live', Symbol='LOC101927057', Aliases='', description='uncharacterized LOC101927057', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31272011, end_position_on_the_genomic_accession=31275502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378620, CurrentID=0, Status='live', Symbol='LOC105378620', Aliases='', description='uncharacterized LOC105378620', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30749249, end_position_on_the_genomic_accession=30752871, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421691, CurrentID=0, Status='live', Symbol='LOC100421691', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41994079, end_position_on_the_genomic_accession=41994701, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419730, CurrentID=0, Status='live', Symbol='ZNF736P11Y', Aliases='', description='zinc finger protein 736 pseudogene 11, Y-linked', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22867141, end_position_on_the_genomic_accession=22867754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100288659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288659, CurrentID=0, Status='live', Symbol='ZNF317P1', Aliases='', description='zinc finger protein 317 pseudogene 1', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93646066, end_position_on_the_genomic_accession=93647094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372945, CurrentID=0, Status='live', Symbol='LOC105372945', Aliases='', description='uncharacterized LOC105372945', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221708840, end_position_on_the_genomic_accession=221714412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927060, CurrentID=0, Status='live', Symbol='LOC101927060', Aliases='', description='uncharacterized LOC101927060', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47056712, end_position_on_the_genomic_accession=47100282, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371961, CurrentID=0, Status='live', Symbol='LOC105371961', Aliases='', description='uncharacterized LOC105371961', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2350579, end_position_on_the_genomic_accession=2371117, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100288661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288661, CurrentID=0, Status='live', Symbol='SUB1P4', Aliases='', description='SUB1 homolog, transcriptional regulator pseudogene 4', other_designations='SUB1 homolog pseudogene 4|SUB1 pseudogene 4', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23537060, end_position_on_the_genomic_accession=23538132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927062, CurrentID=0, Status='live', Symbol='LOC101927062', Aliases='', description='uncharacterized LOC101927062', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26775495, end_position_on_the_genomic_accession=26822120, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645297, CurrentID=0, Status='live', Symbol='LOC645297', Aliases='', description='MAS related GPR family member X2 pseudogene', other_designations='MAS-related GPR, member X2 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18153277, end_position_on_the_genomic_accession=18154254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288663, CurrentID=0, Status='live', Symbol='LOC100288663', Aliases='', description='actin related protein 2/3 complex subunit 1A pseudogene', other_designations='actin related protein 2/3 complex, subunit 1A, 41kDa pseudogene', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208360709, end_position_on_the_genomic_accession=208361369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128297, CurrentID=0, Status='live', Symbol='LOC100128297', Aliases='', description='serine/arginine-rich splicing factor 10 pseudogene', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77956177, end_position_on_the_genomic_accession=77958884, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419737, CurrentID=0, Status='live', Symbol='LOC100419737', Aliases='', description='CTAGE family, member 5 pseudogene', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25031018, end_position_on_the_genomic_accession=25031920, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371962, CurrentID=0, Status='live', Symbol='LOC105371962', Aliases='', description='uncharacterized LOC105371962', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2506629, end_position_on_the_genomic_accession=2512439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616346, CurrentID=0, Status='live', Symbol='MIR4649', Aliases='mir-4649', description='microRNA 4649', other_designations='hsa-mir-4649', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44110849, end_position_on_the_genomic_accession=44110912, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927067, CurrentID=0, Status='live', Symbol='LINC01607', Aliases='', description='long intergenic non-protein coding RNA 1607', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79768142, end_position_on_the_genomic_accession=79802842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616348, CurrentID=0, Status='live', Symbol='MIR4659A', Aliases='mir-4659a', description='microRNA 4659a', other_designations='hsa-mir-4659a', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6745164, end_position_on_the_genomic_accession=6745244, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616349, CurrentID=0, Status='live', Symbol='MIR4451', Aliases='mir-4451', description='microRNA 4451', other_designations='hsa-mir-4451', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=85722468, end_position_on_the_genomic_accession=85722533, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616350, CurrentID=0, Status='live', Symbol='MIR4660', Aliases='', description='microRNA 4660', other_designations='hsa-mir-4660', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9048445, end_position_on_the_genomic_accession=9048518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371963, CurrentID=0, Status='live', Symbol='LOC105371963', Aliases='', description='uncharacterized LOC105371963', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2967000, end_position_on_the_genomic_accession=2980463, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616352, CurrentID=0, Status='live', Symbol='MIR4642', Aliases='', description='microRNA 4642', other_designations='hsa-mir-4642', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44435641, end_position_on_the_genomic_accession=44435722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616353, CurrentID=0, Status='live', Symbol='MIR4532', Aliases='mir-4532', description='microRNA 4532', other_designations='hsa-mir-4532', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57895394, end_position_on_the_genomic_accession=57895444, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100135756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100135756, CurrentID=0, Status='live', Symbol='ATG3P1', Aliases='AGP3, ATG3P, dJ54G6.1', description='autophagy related 3 pseudogene 1', other_designations='ATG3 autophagy related 3 homolog pseudogene 1', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=39328855, end_position_on_the_genomic_accession=39330000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390250, CurrentID=0, Status='live', Symbol='LOC390250', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide E, 25kDa pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110049528, end_position_on_the_genomic_accession=110050157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927074, CurrentID=0, Status='live', Symbol='LINC00682', Aliases='', description='long intergenic non-protein coding RNA 682', other_designations='', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41879521, end_position_on_the_genomic_accession=41882611, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927075, CurrentID=0, Status='live', Symbol='FGF10-AS1', Aliases='', description='FGF10 antisense RNA 1', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=44388732, end_position_on_the_genomic_accession=44413989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371964, CurrentID=0, Status='live', Symbol='LOC105371964', Aliases='', description='uncharacterized LOC105371964', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3347703, end_position_on_the_genomic_accession=3350608, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422598, CurrentID=0, Status='live', Symbol='SDHCP4', Aliases='', description='succinate dehydrogenase complex subunit C pseudogene 4', other_designations='succinate dehydrogenase complex, subunit C pseudogene 4|succinate dehydrogenase complex, subunit C, integral membrane protein, 15kDa pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17435024, end_position_on_the_genomic_accession=17436209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616356, CurrentID=0, Status='live', Symbol='MIR4774', Aliases='', description='microRNA 4774', other_designations='hsa-mir-4774', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168582943, end_position_on_the_genomic_accession=168583018, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616357, CurrentID=0, Status='live', Symbol='MIR3591', Aliases='', description='microRNA 3591', other_designations='hsa-mir-3591', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58451080, end_position_on_the_genomic_accession=58451152, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616358, CurrentID=0, Status='live', Symbol='MIR4800', Aliases='mir-4800', description='microRNA 4800', other_designations='hsa-mir-4800', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2250077, end_position_on_the_genomic_accession=2250156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928358, CurrentID=0, Status='live', Symbol='LOC101928358', Aliases='', description='uncharacterized LOC101928358', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108736540, end_position_on_the_genomic_accession=108738903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616359, CurrentID=0, Status='live', Symbol='MIR4730', Aliases='', description='microRNA 4730', other_designations='hsa-mir-4730', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80419418, end_position_on_the_genomic_accession=80419493, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378642, CurrentID=0, Status='live', Symbol='LOC105378642', Aliases='', description='uncharacterized LOC105378642', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34710022, end_position_on_the_genomic_accession=34724238, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100271453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271453, CurrentID=0, Status='live', Symbol='RPL21P130', Aliases='RPL21_63_1633', description='ribosomal protein L21 pseudogene 130', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17589570, end_position_on_the_genomic_accession=17590109, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927080, CurrentID=0, Status='live', Symbol='LOC101927080', Aliases='', description='uncharacterized LOC101927080', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47297159, end_position_on_the_genomic_accession=47362922, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100287614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287614, CurrentID=0, Status='live', Symbol='LOC100287614', Aliases='', description='ribosomal protein S10 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4621544, end_position_on_the_genomic_accession=4622357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927081, CurrentID=0, Status='live', Symbol='MIR4307HG', Aliases='', description='MIR4307 host gene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26873133, end_position_on_the_genomic_accession=26914743, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616362, CurrentID=0, Status='live', Symbol='MIR4533', Aliases='', description='microRNA 4533', other_designations='hsa-mir-4533', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60478111, end_position_on_the_genomic_accession=60478181, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378645, CurrentID=0, Status='live', Symbol='LOC105378645', Aliases='', description='uncharacterized LOC105378645', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35569815, end_position_on_the_genomic_accession=35578330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927790, CurrentID=0, Status='live', Symbol='LOC101927790', Aliases='', description='uncharacterized LOC101927790', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=159935817, end_position_on_the_genomic_accession=159937492, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928359, CurrentID=0, Status='live', Symbol='LOC101928359', Aliases='', description='uncharacterized LOC101928359', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=122421982, end_position_on_the_genomic_accession=122478081, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927084, CurrentID=0, Status='live', Symbol='LINC01359', Aliases='', description='long intergenic non-protein coding RNA 1359', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64979577, end_position_on_the_genomic_accession=65002476, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100616365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616365, CurrentID=0, Status='live', Symbol='MIR4425', Aliases='', description='microRNA 4425', other_designations='hsa-mir-4425', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25023503, end_position_on_the_genomic_accession=25023586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378648, CurrentID=0, Status='live', Symbol='LOC105378648', Aliases='', description='uncharacterized LOC105378648', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36686701, end_position_on_the_genomic_accession=36694967, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371966, CurrentID=0, Status='live', Symbol='LOC105371966', Aliases='', description='uncharacterized LOC105371966', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3877729, end_position_on_the_genomic_accession=3886864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616366, CurrentID=0, Status='live', Symbol='MIR4743', Aliases='', description='microRNA 4743', other_designations='hsa-mir-4743', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48670600, end_position_on_the_genomic_accession=48670668, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378649, CurrentID=0, Status='live', Symbol='LOC105378649', Aliases='', description='uncharacterized LOC105378649', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37531456, end_position_on_the_genomic_accession=37540213, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927087, CurrentID=0, Status='live', Symbol='LOC101927087', Aliases='', description='uncharacterized LOC101927087', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124499942, end_position_on_the_genomic_accession=124558434, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  346288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346288, CurrentID=0, Status='live', Symbol='SEPT14', Aliases='', description='septin 14', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55793544, end_position_on_the_genomic_accession=55862789, orientation='minus', exon_count=11, OMIM=612140, no_hdr0=''),
  101927089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927089, CurrentID=0, Status='live', Symbol='LOC101927089', Aliases='', description='uncharacterized LOC101927089', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=50620907, end_position_on_the_genomic_accession=50632732, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616370, CurrentID=0, Status='live', Symbol='MIR4771-1', Aliases='', description='microRNA 4771-1', other_designations='hsa-mir-4771-1', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87194786, end_position_on_the_genomic_accession=87194859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378653, CurrentID=0, Status='live', Symbol='LOC105378653', Aliases='', description='uncharacterized LOC105378653', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38047140, end_position_on_the_genomic_accession=38119530, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371967, CurrentID=0, Status='live', Symbol='LOC105371967', Aliases='', description='uncharacterized LOC105371967', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3652733, end_position_on_the_genomic_accession=3656282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419763, CurrentID=0, Status='live', Symbol='LOC100419763', Aliases='', description='zinc finger with KRAB and SCAN domains 8 pseudogene', other_designations='zinc finger protein 192 pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=145047724, end_position_on_the_genomic_accession=145048534, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616372, CurrentID=0, Status='live', Symbol='MIR4659B', Aliases='', description='microRNA 4659b', other_designations='hsa-mir-4659b', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6745168, end_position_on_the_genomic_accession=6745240, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927093, CurrentID=0, Status='live', Symbol='LOC101927093', Aliases='', description='uncharacterized LOC101927093', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30573471, end_position_on_the_genomic_accession=30576658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928361, CurrentID=0, Status='live', Symbol='LOC101928361', Aliases='', description='uncharacterized LOC101928361', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143602750, end_position_on_the_genomic_accession=143608759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927094, CurrentID=0, Status='live', Symbol='LOC101927094', Aliases='', description='uncharacterized LOC101927094', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3445792, end_position_on_the_genomic_accession=3448489, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378657, CurrentID=0, Status='live', Symbol='LOC105378657', Aliases='', description='uncharacterized LOC105378657', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38300262, end_position_on_the_genomic_accession=38503454, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371968, CurrentID=0, Status='live', Symbol='LOC105371968', Aliases='', description='uncharacterized LOC105371968', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=4493153, end_position_on_the_genomic_accession=4496393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616376, CurrentID=0, Status='live', Symbol='MIR4492', Aliases='mir-4492', description='microRNA 4492', other_designations='hsa-mir-4492', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118910708, end_position_on_the_genomic_accession=118910787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616377, CurrentID=0, Status='live', Symbol='MIR4803', Aliases='', description='microRNA 4803', other_designations='hsa-mir-4803', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72169467, end_position_on_the_genomic_accession=72169540, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616378, CurrentID=0, Status='live', Symbol='MIR4784', Aliases='', description='microRNA 4784', other_designations='hsa-mir-4784', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131491160, end_position_on_the_genomic_accession=131491236, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378661, CurrentID=0, Status='live', Symbol='LOC105378661', Aliases='', description='uncharacterized LOC105378661', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38923073, end_position_on_the_genomic_accession=38926507, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  373856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373856, CurrentID=0, Status='live', Symbol='USP41', Aliases='', description='ubiquitin specific peptidase 41', other_designations='deubiquitinating enzyme 41|ubiquitin specific protease 41|ubiquitin specific proteinase 41|ubiquitin thioesterase 41|ubiquitin thiolesterase 41|ubiquitin-specific-processing protease 41', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20363284, end_position_on_the_genomic_accession=20390758, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  373861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373861, CurrentID=0, Status='live', Symbol='HILS1', Aliases='H1.9', description='histone linker H1 domain, spermatid-specific 1, pseudogene', other_designations='spermatid-specific linker histone H1-like protein', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50171428, end_position_on_the_genomic_accession=50172476, orientation='minus', exon_count=2, OMIM=608101, no_hdr0=''),
  373863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=373863, CurrentID=0, Status='live', Symbol='DND1', Aliases='RBMS4', description='DND microRNA-mediated repression inhibitor 1', other_designations='RNA-binding motif, single-stranded-interacting protein 4|dead end homolog 1', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140670796, end_position_on_the_genomic_accession=140673586, orientation='minus', exon_count=4, OMIM=609385, no_hdr0=''),
  101927100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927100, CurrentID=0, Status='live', Symbol='LOC101927100', Aliases='', description='uncharacterized LOC101927100', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115602155, end_position_on_the_genomic_accession=115620419, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130180, CurrentID=0, Status='live', Symbol='LOC100130180', Aliases='', description='nuclear factor, erythroid 2 like 3 pseudogene', other_designations='nuclear factor (erythroid-derived 2)-like 3 pseudogene|nuclear factor, erythroid 2-like 3 pseudogene', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46789457, end_position_on_the_genomic_accession=46792286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371969, CurrentID=0, Status='live', Symbol='LOC105371969', Aliases='', description='uncharacterized LOC105371969', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5081182, end_position_on_the_genomic_accession=5098756, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927101, CurrentID=0, Status='live', Symbol='RPS3AP54', Aliases='', description='ribosomal protein S3a pseudogene 54', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6662366, end_position_on_the_genomic_accession=6663858, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378664, CurrentID=0, Status='live', Symbol='LOC105378664', Aliases='', description='uncharacterized LOC105378664', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39651176, end_position_on_the_genomic_accession=39655326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927102, CurrentID=0, Status='live', Symbol='ITFG1-AS1', Aliases='', description='ITFG1 antisense RNA 1', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47144068, end_position_on_the_genomic_accession=47162747, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616383, CurrentID=0, Status='live', Symbol='MIR4675', Aliases='', description='microRNA 4675', other_designations='hsa-mir-4675', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20551970, end_position_on_the_genomic_accession=20552046, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371970, CurrentID=0, Status='live', Symbol='LOC105371970', Aliases='', description='uncharacterized LOC105371970', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5353832, end_position_on_the_genomic_accession=5374753, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100616386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616386, CurrentID=0, Status='live', Symbol='MIR4654', Aliases='mir-4654', description='microRNA 4654', other_designations='hsa-mir-4654', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162157107, end_position_on_the_genomic_accession=162157182, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378669, CurrentID=0, Status='live', Symbol='LOC105378669', Aliases='', description='uncharacterized LOC105378669', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40014567, end_position_on_the_genomic_accession=40040153, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100616387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616387, CurrentID=0, Status='live', Symbol='MIR3064', Aliases='', description='microRNA 3064', other_designations='hsa-mir-3064', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64500774, end_position_on_the_genomic_accession=64500839, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616388, CurrentID=0, Status='live', Symbol='MIR4723', Aliases='mir-4723', description='microRNA 4723', other_designations='hsa-mir-4723', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28360654, end_position_on_the_genomic_accession=28360734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927109, CurrentID=0, Status='live', Symbol='LOC101927109', Aliases='', description='uncharacterized LOC101927109', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126064159, end_position_on_the_genomic_accession=126072277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616390, CurrentID=0, Status='live', Symbol='MIR4427', Aliases='', description='microRNA 4427', other_designations='hsa-mir-4427', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233624152, end_position_on_the_genomic_accession=233624219, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371971, CurrentID=0, Status='live', Symbol='LOC105371971', Aliases='', description='uncharacterized LOC105371971', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6336322, end_position_on_the_genomic_accession=6367247, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616391, CurrentID=0, Status='live', Symbol='MIR4684', Aliases='', description='microRNA 4684', other_designations='hsa-mir-4684', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22719517, end_position_on_the_genomic_accession=22719598, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927112, CurrentID=0, Status='live', Symbol='LINC01581', Aliases='', description='long intergenic non-protein coding RNA 1581', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93905403, end_position_on_the_genomic_accession=94107938, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378675, CurrentID=0, Status='live', Symbol='LOC105378675', Aliases='', description='uncharacterized LOC105378675', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40658046, end_position_on_the_genomic_accession=40666746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100775104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100775104, CurrentID=0, Status='live', Symbol='KLHL7-AS1', Aliases='PLATAK', description='KLHL7 antisense RNA 1 (head to head)', other_designations='KLHL7 antisense RNA 1 (non-protein coding)|placental transcript antisense to KLHL7', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23101228, end_position_on_the_genomic_accession=23105703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100775105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100775105, CurrentID=0, Status='live', Symbol='ERVK-18', Aliases='HERV-K18, c1_C, env', description='endogenous retrovirus group K member 18', other_designations='Envelope polyprotein|HERV-K_1q23.3 provirus ancestral Env polyprotein|HERV-K_1q23.3 provirus ancestral Pol protein|HERV-K_1q23.3 provirus ancestral Pro protein|IDDMK1,2 22 envelope protein|IDDMK1,2 22 superantigen|IDDMK1,2-22|endogenous retrovirus K-18|endogenous retrovirus group K, member 18', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100775107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100775107, CurrentID=0, Status='live', Symbol='RBM5-AS1', Aliases='LUST', description='RBM5 antisense RNA 1', other_designations='LUCA-15-specific transcript', map_location='3p21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50099603, end_position_on_the_genomic_accession=50100988, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378678, CurrentID=0, Status='live', Symbol='LOC105378678', Aliases='', description='uncharacterized LOC105378678', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41433051, end_position_on_the_genomic_accession=41464911, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616396, CurrentID=0, Status='live', Symbol='MIR4712', Aliases='', description='microRNA 4712', other_designations='hsa-mir-4712', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50360329, end_position_on_the_genomic_accession=50360410, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378679, CurrentID=0, Status='live', Symbol='LOC105378679', Aliases='', description='uncharacterized LOC105378679', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41905344, end_position_on_the_genomic_accession=41907112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616397, CurrentID=0, Status='live', Symbol='MIR4440', Aliases='mir-4440', description='microRNA 4440', other_designations='hsa-mir-4440', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239068817, end_position_on_the_genomic_accession=239068914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616398, CurrentID=0, Status='live', Symbol='MIR4681', Aliases='', description='microRNA 4681', other_designations='hsa-mir-4681', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119377972, end_position_on_the_genomic_accession=119378043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927119, CurrentID=0, Status='live', Symbol='LOC101927119', Aliases='', description='uncharacterized LOC101927119', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38081789, end_position_on_the_genomic_accession=38088741, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378682, CurrentID=0, Status='live', Symbol='LOC105378682', Aliases='', description='uncharacterized LOC105378682', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42841895, end_position_on_the_genomic_accession=42843443, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616400, CurrentID=0, Status='live', Symbol='MIR3198-2', Aliases='mir-3198-2', description='microRNA 3198-2', other_designations='hsa-mir-3198-2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54231397, end_position_on_the_genomic_accession=54231476, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927121, CurrentID=0, Status='live', Symbol='LOC101927121', Aliases='', description='uncharacterized LOC101927121', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74288078, end_position_on_the_genomic_accession=74408803, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378684, CurrentID=0, Status='live', Symbol='LOC105378684', Aliases='', description='uncharacterized LOC105378684', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42979698, end_position_on_the_genomic_accession=42985699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106479508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106479508, CurrentID=0, Status='live', Symbol='RN7SL789P', Aliases='', description='RNA, 7SL, cytoplasmic 789, pseudogene', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616403, CurrentID=0, Status='live', Symbol='MIR4691', Aliases='mir-4691', description='microRNA 4691', other_designations='hsa-mir-4691', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68033897, end_position_on_the_genomic_accession=68033981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927124, CurrentID=0, Status='live', Symbol='LOC101927124', Aliases='', description='uncharacterized LOC101927124', other_designations='CTD-2213F21.2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31420758, end_position_on_the_genomic_accession=31452883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84181, CurrentID=0, Status='live', Symbol='CHD6', Aliases='CHD-6, CHD5, RIGB', description='chromodomain helicase DNA binding protein 6', other_designations='ATP-dependent helicase CHD6|helicase C-terminal domain- and SNF2 N-terminal domain-containing protein|radiation-induced gene B protein', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41402101, end_position_on_the_genomic_accession=41618494, orientation='minus', exon_count=41, OMIM=616114, no_hdr0=''),
  105371974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371974, CurrentID=0, Status='live', Symbol='LOC105371974', Aliases='', description='uncharacterized LOC105371974', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7272319, end_position_on_the_genomic_accession=7279040, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616406, CurrentID=0, Status='live', Symbol='MIR4521', Aliases='mir-4521', description='microRNA 4521', other_designations='hsa-mir-4521', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8186945, end_position_on_the_genomic_accession=8187004, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  285412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285412, CurrentID=0, Status='live', Symbol='TOX4P1', Aliases='', description='TOX high mobility group box family member 4 pseudogene 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112455857, end_position_on_the_genomic_accession=112457986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  404024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=404024, CurrentID=0, Status='live', Symbol='TRIM26BP', Aliases='TRIM26P, TRIM26P1', description='tripartite motif containing 26B, pseudogene', other_designations='tripartite motif containing 26 pseudogene 1|tripartite motif-containing 26 pseudogene 1', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30238301, end_position_on_the_genomic_accession=30242279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927129, CurrentID=0, Status='live', Symbol='LINC01580', Aliases='', description='long intergenic non-protein coding RNA 1580', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93900701, end_position_on_the_genomic_accession=93984150, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100616410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616410, CurrentID=0, Status='live', Symbol='MIR4692', Aliases='', description='microRNA 4692', other_designations='hsa-mir-4692', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72783530, end_position_on_the_genomic_accession=72783592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371975, CurrentID=0, Status='live', Symbol='LOC105371975', Aliases='', description='uncharacterized LOC105371975', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7327449, end_position_on_the_genomic_accession=7359778, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616411, CurrentID=0, Status='live', Symbol='ERVK-16', Aliases='M3.8, c10_A', description='endogenous retrovirus group K member 16', other_designations='HERV-K_10p14 provirus Rec protein|endogenous retrovirus group K, member 16', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378694, CurrentID=0, Status='live', Symbol='LOC105378694', Aliases='', description='uncharacterized LOC105378694', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45867800, end_position_on_the_genomic_accession=45907053, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100616412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616412, CurrentID=0, Status='live', Symbol='ERVK3-5', Aliases='HML6-c12', description='endogenous retrovirus group K3 member 5', other_designations='endogenous retrovirus group K3, member 5', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100616413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616413, CurrentID=0, Status='live', Symbol='MIR4462', Aliases='', description='microRNA 4462', other_designations='hsa-mir-4462', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37555365, end_position_on_the_genomic_accession=37555422, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105755953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105755953, CurrentID=0, Status='live', Symbol='LOC105755953', Aliases='', description='uncharacterized LOC105755953', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105755954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105755954, CurrentID=0, Status='live', Symbol='LACAT8', Aliases='', description='lung adenocarcinoma-associated transcript 8', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927134, CurrentID=0, Status='live', Symbol='LINC01470', Aliases='', description='long intergenic non-protein coding RNA 1470', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152618965, end_position_on_the_genomic_accession=152972349, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616415, CurrentID=0, Status='live', Symbol='MIR4535', Aliases='', description='microRNA 4535', other_designations='hsa-mir-4535', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48780295, end_position_on_the_genomic_accession=48780353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378698, CurrentID=0, Status='live', Symbol='LOC105378698', Aliases='', description='uncharacterized LOC105378698', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46685083, end_position_on_the_genomic_accession=46686665, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371976, CurrentID=0, Status='live', Symbol='LOC105371976', Aliases='', description='uncharacterized LOC105371976', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7502050, end_position_on_the_genomic_accession=7509270, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927136, CurrentID=0, Status='live', Symbol='LOC101927136', Aliases='', description='uncharacterized LOC101927136', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53561289, end_position_on_the_genomic_accession=53573275, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100616417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616417, CurrentID=0, Status='live', Symbol='MIR4786', Aliases='mir-4786', description='microRNA 4786', other_designations='hsa-mir-4786', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239943015, end_position_on_the_genomic_accession=239943094, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378700, CurrentID=0, Status='live', Symbol='LOC105378700', Aliases='', description='uncharacterized LOC105378700', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47018311, end_position_on_the_genomic_accession=47021786, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927138, CurrentID=0, Status='live', Symbol='LOC101927138', Aliases='', description='uncharacterized LOC101927138', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42968744, end_position_on_the_genomic_accession=42971492, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390029, CurrentID=0, Status='live', Symbol='LOC390029', Aliases='', description='actin, beta pseudogene', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1802595, end_position_on_the_genomic_accession=1804194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927139, CurrentID=0, Status='live', Symbol='LOC101927139', Aliases='', description='uncharacterized LOC101927139', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66042500, end_position_on_the_genomic_accession=66050718, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378702, CurrentID=0, Status='live', Symbol='LOC105378702', Aliases='', description='uncharacterized LOC105378702', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47815880, end_position_on_the_genomic_accession=47819491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616420, CurrentID=0, Status='live', Symbol='MIR4744', Aliases='', description='microRNA 4744', other_designations='hsa-mir-4744', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49049687, end_position_on_the_genomic_accession=49049768, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378703, CurrentID=0, Status='live', Symbol='LOC105378703', Aliases='', description='uncharacterized LOC105378703', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48175647, end_position_on_the_genomic_accession=48181628, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371977, CurrentID=0, Status='live', Symbol='LOC105371977', Aliases='', description='uncharacterized LOC105371977', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8625008, end_position_on_the_genomic_accession=8627627, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616421, CurrentID=0, Status='live', Symbol='MIR4689', Aliases='mir-4689', description='microRNA 4689', other_designations='hsa-mir-4689', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5862672, end_position_on_the_genomic_accession=5862741, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927142, CurrentID=0, Status='live', Symbol='LOC101927142', Aliases='', description='uncharacterized LOC101927142', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100972648, end_position_on_the_genomic_accession=100977161, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378705, CurrentID=0, Status='live', Symbol='LOC105378705', Aliases='', description='uncharacterized LOC105378705', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48073565, end_position_on_the_genomic_accession=48074361, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927143, CurrentID=0, Status='live', Symbol='LOC101927143', Aliases='', description='uncharacterized LOC101927143', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224208747, end_position_on_the_genomic_accession=224213279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378706, CurrentID=0, Status='live', Symbol='LOC105378706', Aliases='', description='uncharacterized LOC105378706', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=49025595, end_position_on_the_genomic_accession=49140831, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84200, CurrentID=0, Status='live', Symbol='FLJ23519', Aliases='', description='uncharacterized protein FLJ23519', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378707, CurrentID=0, Status='live', Symbol='LOC105378707', Aliases='', description='uncharacterized LOC105378707', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48956662, end_position_on_the_genomic_accession=48962775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927145, CurrentID=0, Status='live', Symbol='LOC101927145', Aliases='', description='uncharacterized LOC101927145', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62133750, end_position_on_the_genomic_accession=62221384, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378708, CurrentID=0, Status='live', Symbol='LOC105378708', Aliases='', description='uncharacterized LOC105378708', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50024021, end_position_on_the_genomic_accession=50042404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378709, CurrentID=0, Status='live', Symbol='LOC105378709', Aliases='', description='uncharacterized LOC105378709', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50061018, end_position_on_the_genomic_accession=50063659, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616427, CurrentID=0, Status='live', Symbol='MIR5096', Aliases='', description='microRNA 5096', other_designations='hsa-mir-5096', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78820752, end_position_on_the_genomic_accession=78820821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378710, CurrentID=0, Status='live', Symbol='LOC105378710', Aliases='', description='uncharacterized LOC105378710', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50257640, end_position_on_the_genomic_accession=50258202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645314, CurrentID=0, Status='live', Symbol='LOC645314', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C17 pseudogene', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 17 pseudogene', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=193276582, end_position_on_the_genomic_accession=193277661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378711, CurrentID=0, Status='live', Symbol='LOC105378711', Aliases='', description='uncharacterized LOC105378711', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50206084, end_position_on_the_genomic_accession=50258259, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927149, CurrentID=0, Status='live', Symbol='LINC01471', Aliases='', description='long intergenic non-protein coding RNA 1471', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127497601, end_position_on_the_genomic_accession=127537780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378712, CurrentID=0, Status='live', Symbol='LOC105378712', Aliases='', description='uncharacterized LOC105378712', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50265199, end_position_on_the_genomic_accession=50321162, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100616430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616430, CurrentID=0, Status='live', Symbol='MIR4644', Aliases='', description='microRNA 4644', other_designations='hsa-mir-4644', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170330761, end_position_on_the_genomic_accession=170330844, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378713, CurrentID=0, Status='live', Symbol='LOC105378713', Aliases='', description='uncharacterized LOC105378713', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50298644, end_position_on_the_genomic_accession=50307609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927151, CurrentID=0, Status='live', Symbol='LOC101927151', Aliases='', description='uncharacterized LOC101927151', other_designations='', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27793467, end_position_on_the_genomic_accession=27806782, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928217, CurrentID=0, Status='live', Symbol='LOC101928217', Aliases='', description='uncharacterized LOC101928217', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4032391, end_position_on_the_genomic_accession=4074289, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101927152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927152, CurrentID=0, Status='live', Symbol='LINC01212', Aliases='', description='long intergenic non-protein coding RNA 1212', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69999744, end_position_on_the_genomic_accession=70015301, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378715, CurrentID=0, Status='live', Symbol='LOC105378715', Aliases='', description='uncharacterized LOC105378715', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50447416, end_position_on_the_genomic_accession=50480014, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927153, CurrentID=0, Status='live', Symbol='LOC101927153', Aliases='', description='uncharacterized LOC101927153', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93856560, end_position_on_the_genomic_accession=93878356, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378716, CurrentID=0, Status='live', Symbol='LOC105378716', Aliases='', description='uncharacterized LOC105378716', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51218412, end_position_on_the_genomic_accession=51236237, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100616434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616434, CurrentID=0, Status='live', Symbol='MIR4755', Aliases='mir-4755', description='microRNA 4755', other_designations='hsa-mir-4755', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34049119, end_position_on_the_genomic_accession=34049190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927155, CurrentID=0, Status='live', Symbol='LMO7-AS1', Aliases='UCHL3-AS1', description='LMO7 antisense RNA 1', other_designations='UCHL3 antisense RNA 1', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75604700, end_position_on_the_genomic_accession=75635994, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378718, CurrentID=0, Status='live', Symbol='LOC105378718', Aliases='', description='uncharacterized LOC105378718', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51669532, end_position_on_the_genomic_accession=51674021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371980, CurrentID=0, Status='live', Symbol='LOC105371980', Aliases='', description='uncharacterized LOC105371980', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9404664, end_position_on_the_genomic_accession=9407494, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927156, CurrentID=0, Status='live', Symbol='LOC101927156', Aliases='', description='uncharacterized LOC101927156', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181123837, end_position_on_the_genomic_accession=181399553, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378719, CurrentID=0, Status='live', Symbol='LOC105378719', Aliases='', description='uncharacterized LOC105378719', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51610692, end_position_on_the_genomic_accession=51623937, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100288757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288757, CurrentID=0, Status='live', Symbol='SUMO2P4', Aliases='', description='SUMO2 pseudogene 4', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 4', map_location='5q13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69068925, end_position_on_the_genomic_accession=69069200, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929963, CurrentID=0, Status='live', Symbol='LOC101929963', Aliases='', description='uncharacterized LOC101929963', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176753400, end_position_on_the_genomic_accession=176763093, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927158, CurrentID=0, Status='live', Symbol='RPSAP69', Aliases='', description='ribosomal protein SA pseudogene 69', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196982909, end_position_on_the_genomic_accession=196987573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616439, CurrentID=0, Status='live', Symbol='MIR4658', Aliases='', description='microRNA 4658', other_designations='hsa-mir-4658', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100156605, end_position_on_the_genomic_accession=100156669, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616440, CurrentID=0, Status='live', Symbol='MIR548AG2', Aliases='', description='microRNA 548ag-2', other_designations='hsa-mir-548ag-2', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60564562, end_position_on_the_genomic_accession=60564625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371981, CurrentID=0, Status='live', Symbol='LOC105371981', Aliases='', description='uncharacterized LOC105371981', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9708713, end_position_on_the_genomic_accession=9716356, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616441, CurrentID=0, Status='live', Symbol='MIR4474', Aliases='', description='microRNA 4474', other_designations='hsa-mir-4474', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=20502265, end_position_on_the_genomic_accession=20502342, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  84218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84218, CurrentID=0, Status='live', Symbol='TBC1D3F', Aliases='TBC1D3', description='TBC1 domain family member 3F', other_designations='TBC1 domain family member 3|TBC1 domain family, member 3F|prostate cancer gene 17 protein|protein TRE17-alpha', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36428612, end_position_on_the_genomic_accession=36439523, orientation='minus', exon_count=14, OMIM=610809, no_hdr0=''),
  105378725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378725, CurrentID=0, Status='live', Symbol='LOC105378725', Aliases='', description='uncharacterized LOC105378725', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53311983, end_position_on_the_genomic_accession=53321441, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100616443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616443, CurrentID=0, Status='live', Symbol='ERVK-15', Aliases='P1.10, c7_C', description='endogenous retrovirus group K member 15', other_designations='endogenous retrovirus group K, member 15', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378726, CurrentID=0, Status='live', Symbol='LOC105378726', Aliases='', description='uncharacterized LOC105378726', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53289570, end_position_on_the_genomic_accession=53311174, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100616444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616444, CurrentID=0, Status='live', Symbol='ERVK-20', Aliases='c11_B, env', description='endogenous retrovirus group K member 20', other_designations='endogenous retrovirus group K, member 20|envelope glycoprotein', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100270855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270855, CurrentID=0, Status='live', Symbol='RPL9P10', Aliases='RPL9_2_176', description='ribosomal protein L9 pseudogene 10', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234284951, end_position_on_the_genomic_accession=234285656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378727, CurrentID=0, Status='live', Symbol='LOC105378727', Aliases='', description='uncharacterized LOC105378727', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53286496, end_position_on_the_genomic_accession=53289558, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616445, CurrentID=0, Status='live', Symbol='ERVK-22', Aliases='c19_A', description='endogenous retrovirus group K member 22', other_designations='endogenous retrovirus group K, member 22|env protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112846, CurrentID=0, Status='live', Symbol='RBM8B', Aliases='', description='RNA binding motif protein 8B pseudogene', other_designations='RNA binding motif protein 8 pseudogene|ribonucleoprotein RBM8 pseudogene', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60397699, end_position_on_the_genomic_accession=60400514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371982, CurrentID=0, Status='live', Symbol='LOC105371982', Aliases='', description='uncharacterized LOC105371982', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9895227, end_position_on_the_genomic_accession=9897899, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84222, CurrentID=0, Status='live', Symbol='TMEM191A', Aliases='TMEM191AP', description='transmembrane protein 191A (pseudogene)', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20701114, end_position_on_the_genomic_accession=20704603, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100616447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616447, CurrentID=0, Status='live', Symbol='MIR4780', Aliases='', description='microRNA 4780', other_designations='hsa-mir-4780', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88082519, end_position_on_the_genomic_accession=88082599, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378730, CurrentID=0, Status='live', Symbol='LOC105378730', Aliases='', description='uncharacterized LOC105378730', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53337193, end_position_on_the_genomic_accession=53339291, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616448, CurrentID=0, Status='live', Symbol='MIR4792', Aliases='miR-4792', description='microRNA 4792', other_designations='hsa-mir-4792', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24521362, end_position_on_the_genomic_accession=24521435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100529067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529067, CurrentID=0, Status='live', Symbol='SERF2-C15ORF63', Aliases='', description='SERF2-C15orf63 readthrough', other_designations='', map_location='15q', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43791976, end_position_on_the_genomic_accession=43802589, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100616449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616449, CurrentID=0, Status='live', Symbol='MIR4725', Aliases='mir-4725', description='microRNA 4725', other_designations='hsa-mir-4725', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31575269, end_position_on_the_genomic_accession=31575358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616450, CurrentID=0, Status='live', Symbol='MIR548AG1', Aliases='', description='microRNA 548ag-1', other_designations='hsa-mir-548ag-1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=60922619, end_position_on_the_genomic_accession=60922684, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374286, CurrentID=0, Status='live', Symbol='CDRT1', Aliases='C170RF1, C17ORF1, C17ORF1A, FBXW10B, FBXW10P1, HREP, SM25H2', description='CMT1A duplicated region transcript 1', other_designations='Charcot-Marie-Tooth duplicated region transcript 1|F-box and WD repeat domain containing 10 pseudogene 1|F-box and WD-40 domain protein 10 pseudogene 1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15565482, end_position_on_the_genomic_accession=15619704, orientation='minus', exon_count=14, OMIM=604596, no_hdr0=''),
  105378733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378733, CurrentID=0, Status='live', Symbol='LOC105378733', Aliases='', description='uncharacterized LOC105378733', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53439781, end_position_on_the_genomic_accession=53457782, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371983, CurrentID=0, Status='live', Symbol='LOC105371983', Aliases='', description='uncharacterized LOC105371983', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10027198, end_position_on_the_genomic_accession=10028945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  374291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374291, CurrentID=0, Status='live', Symbol='NDUFS7', Aliases='CI-20, CI-20KD, MY017, PSST', description='NADH:ubiquinone oxidoreductase core subunit S7', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 7, 20kDa (NADH-coenzyme Q reductase)|NADH-coenzyme Q reductase|NADH-ubiquinone oxidoreductase 20 kDa subunit|NADH:ubiquinone oxidoreductase PSST subunit|PSST subunit|complex I 20kDa subunit|complex I, mitochondrial respiratory chain, 20-KD subunit|complex I-20kD', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1383884, end_position_on_the_genomic_accession=1395589, orientation='plus', exon_count=9, OMIM=601825, no_hdr0=''),
  101927172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927172, CurrentID=0, Status='live', Symbol='LOC101927172', Aliases='', description='uncharacterized LOC101927172', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101491257, end_position_on_the_genomic_accession=101492683, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378735, CurrentID=0, Status='live', Symbol='LOC105378735', Aliases='', description='uncharacterized LOC105378735', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54514233, end_position_on_the_genomic_accession=54517129, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372073, CurrentID=0, Status='live', Symbol='LOC105372073', Aliases='', description='uncharacterized LOC105372073', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37734473, end_position_on_the_genomic_accession=37774707, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616453, CurrentID=0, Status='live', Symbol='MIR4687', Aliases='', description='microRNA 4687', other_designations='hsa-mir-4687', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3856062, end_position_on_the_genomic_accession=3856141, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378736, CurrentID=0, Status='live', Symbol='LOC105378736', Aliases='', description='uncharacterized LOC105378736', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54975157, end_position_on_the_genomic_accession=54980460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  374308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374308, CurrentID=0, Status='live', Symbol='PTCHD3', Aliases='PTR', description='patched domain containing 3', other_designations='patched-related protein', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27398188, end_position_on_the_genomic_accession=27414368, orientation='minus', exon_count=4, OMIM=611791, no_hdr0=''),
  100616454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616454, CurrentID=0, Status='live', Symbol='MIR4497', Aliases='mir-4497', description='microRNA 4497', other_designations='hsa-mir-4497', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109833348, end_position_on_the_genomic_accession=109833436, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616455, CurrentID=0, Status='live', Symbol='MIR548H5', Aliases='', description='microRNA 548h-5', other_designations='hsa-mir-548h-5', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131792172, end_position_on_the_genomic_accession=131792231, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378738, CurrentID=0, Status='live', Symbol='LOC105378738', Aliases='', description='uncharacterized LOC105378738', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55945035, end_position_on_the_genomic_accession=55950119, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616456, CurrentID=0, Status='live', Symbol='MIR4476', Aliases='mir-4476', description='microRNA 4476', other_designations='hsa-mir-4476', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36893462, end_position_on_the_genomic_accession=36893531, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378739, CurrentID=0, Status='live', Symbol='LOC105378739', Aliases='', description='uncharacterized LOC105378739', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55915609, end_position_on_the_genomic_accession=55944973, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616457, CurrentID=0, Status='live', Symbol='MIR4693', Aliases='', description='microRNA 4693', other_designations='hsa-mir-4693', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103849906, end_position_on_the_genomic_accession=103849980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927178, CurrentID=0, Status='live', Symbol='LOC101927178', Aliases='', description='uncharacterized LOC101927178', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35081081, end_position_on_the_genomic_accession=35113130, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927179, CurrentID=0, Status='live', Symbol='LOC101927179', Aliases='', description='uncharacterized LOC101927179', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47831345, end_position_on_the_genomic_accession=47896848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  767557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767557, CurrentID=0, Status='live', Symbol='SYS1-DBNDD2', Aliases='C20orf169-DBNDD2', description='SYS1-DBNDD2 readthrough (NMD candidate)', other_designations='SYS1-DBNDD2 readthrough (non-protein coding)', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45363169, end_position_on_the_genomic_accession=45410610, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  767558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767558, CurrentID=0, Status='live', Symbol='LUZP6', Aliases='MPD6, MTPNUT', description='leucine zipper protein 6', other_designations="myeloproliferative disease-associated 6 kDa antigen|myeloproliferative disease-associated SEREX antigen|myeloproliferative disease-associated antigen, 6-kD|myotrophin 3'UTR transcript", map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135926755, end_position_on_the_genomic_accession=135977456, orientation='minus', exon_count=4, OMIM=611050, no_hdr0=''),
  767561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767561, CurrentID=0, Status='live', Symbol='SNORD113-1', Aliases='14q(I-1)', description='small nucleolar RNA, C/D box 113-1', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100924821, end_position_on_the_genomic_accession=100924890, orientation='plus', exon_count=1, OMIM=613650, no_hdr0=''),
  767562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767562, CurrentID=0, Status='live', Symbol='SNORD113-2', Aliases='14q(I-2)', description='small nucleolar RNA, C/D box 113-2', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100927342, end_position_on_the_genomic_accession=100927412, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767563, CurrentID=0, Status='live', Symbol='SNORD113-3', Aliases='14q(I-3)', description='small nucleolar RNA, C/D box 113-3', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100929919, end_position_on_the_genomic_accession=100929989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767564, CurrentID=0, Status='live', Symbol='SNORD113-4', Aliases='14q(I-4)', description='small nucleolar RNA, C/D box 113-4', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100936491, end_position_on_the_genomic_accession=100936564, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767565, CurrentID=0, Status='live', Symbol='SNORD113-5', Aliases='14q(I-5)', description='small nucleolar RNA, C/D box 113-5', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100938187, end_position_on_the_genomic_accession=100938263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767566, CurrentID=0, Status='live', Symbol='SNORD113-6', Aliases='14q(I-6)', description='small nucleolar RNA, C/D box 113-6', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100939556, end_position_on_the_genomic_accession=100939629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767567, CurrentID=0, Status='live', Symbol='SNORD113-7', Aliases='14q(I-7)', description='small nucleolar RNA, C/D box 113-7', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100941126, end_position_on_the_genomic_accession=100941201, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767568, CurrentID=0, Status='live', Symbol='SNORD113-8', Aliases='14q(I-8)', description='small nucleolar RNA, C/D box 113-8', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100943451, end_position_on_the_genomic_accession=100943523, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767569, CurrentID=0, Status='live', Symbol='SNORD113-9', Aliases='14q(I-9)', description='small nucleolar RNA, C/D box 113-9', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100945649, end_position_on_the_genomic_accession=100945719, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374354, CurrentID=0, Status='live', Symbol='NHLRC2', Aliases='', description='NHL repeat containing 2', other_designations='1200003G01Rik|novel NHL repeat domain containing protein', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113854632, end_position_on_the_genomic_accession=113912506, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  374355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374355, CurrentID=0, Status='live', Symbol='CCDC172', Aliases='C10orf96', description='coiled-coil domain containing 172', other_designations='UPF0628 protein C10orf96', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116324104, end_position_on_the_genomic_accession=116380029, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  101927182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927182, CurrentID=0, Status='live', Symbol='LOC101927182', Aliases='', description='uncharacterized LOC101927182', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=42090234, end_position_on_the_genomic_accession=42091012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  767575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767575, CurrentID=0, Status='live', Symbol='MRX71', Aliases='', description='mental retardation, X-linked 71', other_designations='', map_location='Xq24-q27', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  767577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767577, CurrentID=0, Status='live', Symbol='SNORD114-1', Aliases='14q(II-1)', description='small nucleolar RNA, C/D box 114-1', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100949833, end_position_on_the_genomic_accession=100949903, orientation='plus', exon_count=1, OMIM=613651, no_hdr0=''),
  767578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767578, CurrentID=0, Status='live', Symbol='SNORD114-2', Aliases='14q(II-2)', description='small nucleolar RNA, C/D box 114-2', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100951856, end_position_on_the_genomic_accession=100951932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767579, CurrentID=0, Status='live', Symbol='SNORD114-3', Aliases='14q(II-3)', description='small nucleolar RNA, C/D box 114-3', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100953349, end_position_on_the_genomic_accession=100953422, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767580, CurrentID=0, Status='live', Symbol='SNORD114-4', Aliases='14q(II-4)', description='small nucleolar RNA, C/D box 114-4', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100954344, end_position_on_the_genomic_accession=100954447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767581, CurrentID=0, Status='live', Symbol='SNORD114-5', Aliases='14q(II-5)', description='small nucleolar RNA, C/D box 114-5', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100955370, end_position_on_the_genomic_accession=100955438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767582, CurrentID=0, Status='live', Symbol='SNORD114-6', Aliases='14q(II-6)', description='small nucleolar RNA, C/D box 114-6', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100957166, end_position_on_the_genomic_accession=100957236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767583, CurrentID=0, Status='live', Symbol='SNORD114-7', Aliases='14q(II-7)', description='small nucleolar RNA, C/D box 114-7', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100963054, end_position_on_the_genomic_accession=100963129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767584, CurrentID=0, Status='live', Symbol='SNORD114-8', Aliases='14q(II-8)', description='small nucleolar RNA, C/D box 114-8', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100964781, end_position_on_the_genomic_accession=100964851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616464, CurrentID=0, Status='live', Symbol='MIR4778', Aliases='', description='microRNA 4778', other_designations='hsa-mir-4778', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66358249, end_position_on_the_genomic_accession=66358328, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378747, CurrentID=0, Status='live', Symbol='LOC105378747', Aliases='', description='uncharacterized LOC105378747', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=57172262, end_position_on_the_genomic_accession=57186219, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  767588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767588, CurrentID=0, Status='live', Symbol='SNORD114-10', Aliases='14q(II-10)', description='small nucleolar RNA, C/D box 114-10', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100967052, end_position_on_the_genomic_accession=100967122, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767589, CurrentID=0, Status='live', Symbol='SNORD114-11', Aliases='14q(II-11)', description='small nucleolar RNA, C/D box 114-11', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100968111, end_position_on_the_genomic_accession=100968184, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767590, CurrentID=0, Status='live', Symbol='SNORD114-12', Aliases='14q(II-12)', description='small nucleolar RNA, C/D box 114-12', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100968948, end_position_on_the_genomic_accession=100969021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767591, CurrentID=0, Status='live', Symbol='SNORD114-13', Aliases='14q(II-13)', description='small nucleolar RNA, C/D box 114-13', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100969879, end_position_on_the_genomic_accession=100969951, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767592, CurrentID=0, Status='live', Symbol='SNORD114-14', Aliases='14q(II-14)', description='small nucleolar RNA, C/D box 114-14', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100972103, end_position_on_the_genomic_accession=100972176, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767593, CurrentID=0, Status='live', Symbol='SNORD114-15', Aliases='14q(II-15)', description='small nucleolar RNA, C/D box 114-15', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100972670, end_position_on_the_genomic_accession=100972740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374378, CurrentID=0, Status='live', Symbol='GALNT18', Aliases='GALNACT18, GALNT15, GALNTL4, GalNAc-T15, GalNAc-T18', description='polypeptide N-acetylgalactosaminyltransferase 18', other_designations='GalNAc-transferase 18|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 4|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 15|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 18|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 4|galNAc-T-like protein 4|polypeptide GalNAc transferase 18|polypeptide GalNAc transferase-like protein 4|polypeptide N-acetylgalactosaminyltransferase-like protein 4|pp-GaNTase-like protein 4|protein-UDP acetylgalactosaminyltransferase-like protein 4|putative polypeptide N-acetylgalactosaminyltransferase-like protein 4', map_location='11p15.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11270874, end_position_on_the_genomic_accession=11622045, orientation='minus', exon_count=16, OMIM=615136, no_hdr0=''),
  767595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767595, CurrentID=0, Status='live', Symbol='SNORD114-17', Aliases='14q(II-17)', description='small nucleolar RNA, C/D box 114-17', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100974806, end_position_on_the_genomic_accession=100974879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767596, CurrentID=0, Status='live', Symbol='SNORD114-18', Aliases='14q(II-18)', description='small nucleolar RNA, C/D box 114-18', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100975825, end_position_on_the_genomic_accession=100975895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767597, CurrentID=0, Status='live', Symbol='SNORD114-19', Aliases='14q(II-19)', description='small nucleolar RNA, C/D box 114-19', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100976477, end_position_on_the_genomic_accession=100976550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767598, CurrentID=0, Status='live', Symbol='SNORD114-20', Aliases='14q(II-20)', description='small nucleolar RNA, C/D box 114-20', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100981004, end_position_on_the_genomic_accession=100981074, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374383, CurrentID=0, Status='live', Symbol='NCR3LG1', Aliases='B7-H6, B7H6, DKFZp686O24166', description='natural killer cell cytotoxicity receptor 3 ligand 1', other_designations='B7 homolog 6|putative Ig-like domain-containing protein DKFZp686O24166/DKFZp686I21167', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17351732, end_position_on_the_genomic_accession=17381661, orientation='plus', exon_count=9, OMIM=613714, no_hdr0=''),
  767600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767600, CurrentID=0, Status='live', Symbol='SNORD114-22', Aliases='14q(II-22)', description='small nucleolar RNA, C/D box 114-22', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100982926, end_position_on_the_genomic_accession=100982996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=442240, CurrentID=0, Status='live', Symbol='ZNF259P1', Aliases='354J5, ZNF259P', description='zinc finger protein 259 pseudogene 1', other_designations='zinc finger protein 259, pseudogene', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108785324, end_position_on_the_genomic_accession=108787059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616467, CurrentID=0, Status='live', Symbol='MIR4767', Aliases='mir-4767', description='microRNA 4767', other_designations='hsa-mir-4767', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=7147860, end_position_on_the_genomic_accession=7147937, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767604, CurrentID=0, Status='live', Symbol='SNORD114-24', Aliases='14q(II-24)', description='small nucleolar RNA, C/D box 114-24', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100984777, end_position_on_the_genomic_accession=100984847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767605, CurrentID=0, Status='live', Symbol='SNORD114-25', Aliases='14q(II-25)', description='small nucleolar RNA, C/D box 114-25', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100986057, end_position_on_the_genomic_accession=100986127, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767606, CurrentID=0, Status='live', Symbol='SNORD114-26', Aliases='14q(II-26)', description='small nucleolar RNA, C/D box 114-26', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100987046, end_position_on_the_genomic_accession=100987116, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  767608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767608, CurrentID=0, Status='live', Symbol='SNORD114-27', Aliases='14q(II-27)', description='small nucleolar RNA, C/D box 114-27', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100988161, end_position_on_the_genomic_accession=100988229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374393, CurrentID=0, Status='live', Symbol='FAM111B', Aliases='CANP, POIKTMP', description='family with sequence similarity 111 member B', other_designations='cancer-associated nucleoprotein|family with sequence similarity 111, member B', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59107185, end_position_on_the_genomic_accession=59127415, orientation='plus', exon_count=4, OMIM=615584, no_hdr0=''),
  767610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767610, CurrentID=0, Status='live', Symbol='SNORD114-29', Aliases='14q(II-29)', description='small nucleolar RNA, C/D box 114-29', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100990091, end_position_on_the_genomic_accession=100990159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374395, CurrentID=0, Status='live', Symbol='TMEM179B', Aliases='', description='transmembrane protein 179B', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62787399, end_position_on_the_genomic_accession=62790400, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  767612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767612, CurrentID=0, Status='live', Symbol='SNORD114-31', Aliases='14q(II-31)', description='small nucleolar RNA, C/D box 114-31', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100993236, end_position_on_the_genomic_accession=100993309, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  346389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346389, CurrentID=0, Status='live', Symbol='MACC1', Aliases='7A5, SH3BP4L', description='metastasis associated in colon cancer 1', other_designations='SH3 domain-containing protein 7a5|putative binding protein 7a5', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20134655, end_position_on_the_genomic_accession=20217390, orientation='minus', exon_count=7, OMIM=612646, no_hdr0=''),
  374403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374403, CurrentID=0, Status='live', Symbol='TBC1D10C', Aliases='CARABIN, EPI64C', description='TBC1 domain family member 10C', other_designations='TBC1 domain family, member 10C', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67403686, end_position_on_the_genomic_accession=67410090, orientation='plus', exon_count=11, OMIM=610831, no_hdr0=''),
  100616470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616470, CurrentID=0, Status='live', Symbol='MIR4488', Aliases='mir-4488', description='microRNA 4488', other_designations='hsa-mir-4488', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61508596, end_position_on_the_genomic_accession=61508657, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374407, CurrentID=0, Status='live', Symbol='DNAJB13', Aliases='RSPH16A, TSARG5, TSARG6', description='DnaJ heat shock protein family (Hsp40) member B13', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 13|DnaJ (Hsp40) related, subfamily B, member 13|DnaJ-like protein|radial spoke 16 homolog A|testis and spermatogenesis cell-related protein 6|testis spermatocyte apoptosis-related gene 6 protein|testis spermatogenesis apoptosis-related gene 3 protein|testis spermatogenesis apoptosis-related gene 6 protein|testis spermatogenesis apoptosis-related protein 6', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73950319, end_position_on_the_genomic_accession=73970287, orientation='plus', exon_count=16, OMIM=610263, no_hdr0=''),
  105371987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371987, CurrentID=0, Status='live', Symbol='LOC105371987', Aliases='', description='uncharacterized LOC105371987', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10375063, end_position_on_the_genomic_accession=10378308, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616471, CurrentID=0, Status='live', Symbol='MIR4798', Aliases='', description='microRNA 4798', other_designations='hsa-mir-4798', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7310450, end_position_on_the_genomic_accession=7310524, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378754, CurrentID=0, Status='live', Symbol='LOC105378754', Aliases='', description='uncharacterized LOC105378754', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59184277, end_position_on_the_genomic_accession=59201435, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616472, CurrentID=0, Status='live', Symbol='MIR4776-2', Aliases='mir-4776-2', description='microRNA 4776-2', other_designations='hsa-mir-4776-2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212926257, end_position_on_the_genomic_accession=212926336, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378755, CurrentID=0, Status='live', Symbol='LOC105378755', Aliases='', description='uncharacterized LOC105378755', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59491606, end_position_on_the_genomic_accession=59501173, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616473, CurrentID=0, Status='live', Symbol='MIR4432', Aliases='', description='microRNA 4432', other_designations='hsa-mir-4432', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60387362, end_position_on_the_genomic_accession=60387445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378756, CurrentID=0, Status='live', Symbol='LOC105378756', Aliases='', description='uncharacterized LOC105378756', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59289281, end_position_on_the_genomic_accession=59296460, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100616474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616474, CurrentID=0, Status='live', Symbol='MIR4715', Aliases='', description='microRNA 4715', other_designations='hsa-mir-4715', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25848747, end_position_on_the_genomic_accession=25848825, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616475, CurrentID=0, Status='live', Symbol='MIR548AD', Aliases='', description='microRNA 548ad', other_designations='hsa-mir-548ad', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=35471405, end_position_on_the_genomic_accession=35471486, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371988, CurrentID=0, Status='live', Symbol='LOC105371988', Aliases='', description='uncharacterized LOC105371988', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10404804, end_position_on_the_genomic_accession=10425250, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419868, CurrentID=0, Status='live', Symbol='ZNF33BP1', Aliases='', description='zinc finger protein 33B pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37793938, end_position_on_the_genomic_accession=37795175, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  374443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374443, CurrentID=0, Status='live', Symbol='LOC374443', Aliases='', description='C-type lectin domain family 2, member D pseudogene', other_designations='CLR pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9617284, end_position_on_the_genomic_accession=9658415, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100616477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616477, CurrentID=0, Status='live', Symbol='MIR4442', Aliases='mir-4442', description='microRNA 4442', other_designations='hsa-mir-4442', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25664873, end_position_on_the_genomic_accession=25664939, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616478, CurrentID=0, Status='live', Symbol='MIR4494', Aliases='', description='microRNA 4494', other_designations='hsa-mir-4494', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47364186, end_position_on_the_genomic_accession=47364269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  374454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374454, CurrentID=0, Status='live', Symbol='KRT77', Aliases='K1B, KRT1B', description='keratin 77, type II', other_designations='CK-1B|K77|cytokeratin-1B|keratin 1B|keratin-77|type-II keratin Kb39', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52689612, end_position_on_the_genomic_accession=52703534, orientation='minus', exon_count=10, OMIM=611158, no_hdr0=''),
  105378761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378761, CurrentID=0, Status='live', Symbol='LOC105378761', Aliases='', description='uncharacterized LOC105378761', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60183012, end_position_on_the_genomic_accession=60276791, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616479, CurrentID=0, Status='live', Symbol='MIR4635', Aliases='', description='microRNA 4635', other_designations='hsa-mir-4635', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1062896, end_position_on_the_genomic_accession=1062974, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  374462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374462, CurrentID=0, Status='live', Symbol='PTPRQ', Aliases='DFNB84, DFNB84A, PTPGMC1, R-PTP-Q', description='protein tyrosine phosphatase, receptor type, Q', other_designations='phosphotidylinositol phosphatase PTPRQ|protein-tyrosine phosphatase, receptor-type, expressed by glomerular mesangial cells', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80444274, end_position_on_the_genomic_accession=80680189, orientation='plus', exon_count=47, OMIM=603317, no_hdr0=''),
  100616480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616480, CurrentID=0, Status='live', Symbol='MIR4479', Aliases='', description='microRNA 4479', other_designations='hsa-mir-4479', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136886733, end_position_on_the_genomic_accession=136886803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390091, CurrentID=0, Status='live', Symbol='OR10AB1P', Aliases='', description='olfactory receptor family 10 subfamily AB member 1 pseudogene', other_designations='olfactory receptor OR11-91 pseudogene|olfactory receptor, family 10, subfamily AB, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7728401, end_position_on_the_genomic_accession=7729338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  374470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374470, CurrentID=0, Status='live', Symbol='C12orf42', Aliases='', description='chromosome 12 open reading frame 42', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103079645, end_position_on_the_genomic_accession=103525254, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  100419873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419873, CurrentID=0, Status='live', Symbol='ZNF519P1', Aliases='', description='zinc finger protein 519 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80298178, end_position_on_the_genomic_accession=80299202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100616482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616482, CurrentID=0, Status='live', Symbol='MIR4685', Aliases='', description='microRNA 4685', other_designations='hsa-mir-4685', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98431292, end_position_on_the_genomic_accession=98431360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378765, CurrentID=0, Status='live', Symbol='LOC105378765', Aliases='', description='uncharacterized LOC105378765', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61211942, end_position_on_the_genomic_accession=61230708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616483, CurrentID=0, Status='live', Symbol='MIR4751', Aliases='', description='microRNA 4751', other_designations='hsa-mir-4751', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49933064, end_position_on_the_genomic_accession=49933137, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928383, CurrentID=0, Status='live', Symbol='LOC101928383', Aliases='', description='uncharacterized LOC101928383', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173384655, end_position_on_the_genomic_accession=173479459, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100616484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616484, CurrentID=0, Status='live', Symbol='MIR4470', Aliases='mir-4470', description='microRNA 4470', other_designations='hsa-mir-4470', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61714788, end_position_on_the_genomic_accession=61714859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  374491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374491, CurrentID=0, Status='live', Symbol='TPTE2P6', Aliases='', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 6', other_designations='TPTE and PTEN homologous inositol lipid phosphatase pseudogene', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24580208, end_position_on_the_genomic_accession=24597676, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100616485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616485, CurrentID=0, Status='live', Symbol='MIR4447', Aliases='', description='microRNA 4447', other_designations='hsa-mir-4447', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116850277, end_position_on_the_genomic_accession=116850367, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271458, CurrentID=0, Status='live', Symbol='RPL23AP29', Aliases='RPL23A_11_287', description='ribosomal protein L23a pseudogene 29', other_designations='', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153370626, end_position_on_the_genomic_accession=153371064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  374500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374500, CurrentID=0, Status='live', Symbol='THSD1P1', Aliases='THSD1P', description='thrombospondin type 1 domain containing 1 pseudogene 1', other_designations='THSD1 pseudogene|thrombospondin, type I, domain containing 1 pseudogene 1', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52167719, end_position_on_the_genomic_accession=52168321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616486, CurrentID=0, Status='live', Symbol='MIR4698', Aliases='', description='microRNA 4698', other_designations='hsa-mir-4698', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47187812, end_position_on_the_genomic_accession=47187891, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616487, CurrentID=0, Status='live', Symbol='MIR4517', Aliases='mir-4517', description='microRNA 4517', other_designations='hsa-mir-4517', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28958583, end_position_on_the_genomic_accession=28958661, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390098, CurrentID=0, Status='live', Symbol='LOC390098', Aliases='', description='MAS related GPR family member X1 pseudogene', other_designations='MAS-related GPR, member X1 pseudogene', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18886924, end_position_on_the_genomic_accession=18888136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616488, CurrentID=0, Status='live', Symbol='MIR548AK', Aliases='', description='microRNA 548ak', other_designations='hsa-mir-548ak', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12130760, end_position_on_the_genomic_accession=12130816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376355, CurrentID=0, Status='live', Symbol='LOC105376355', Aliases='', description='uncharacterized LOC105376355', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3270280, end_position_on_the_genomic_accession=3277073, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928384, CurrentID=0, Status='live', Symbol='LOC101928384', Aliases='', description='uncharacterized LOC101928384', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134219085, end_position_on_the_genomic_accession=134220108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100616489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616489, CurrentID=0, Status='live', Symbol='MIR4417', Aliases='mir-4417', description='microRNA 4417', other_designations='hsa-mir-4417', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5564071, end_position_on_the_genomic_accession=5564143, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616490, CurrentID=0, Status='live', Symbol='MIR4706', Aliases='mir-4706', description='microRNA 4706', other_designations='hsa-mir-4706', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65044688, end_position_on_the_genomic_accession=65044769, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390101, CurrentID=0, Status='live', Symbol='LOC390101', Aliases='', description='MAS-related GPR, member X1 pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18971733, end_position_on_the_genomic_accession=18972922, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100616491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616491, CurrentID=0, Status='live', Symbol='MIR3978', Aliases='', description='microRNA 3978', other_designations='hsa-mir-3978', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=110082118, end_position_on_the_genomic_accession=110082218, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378774, CurrentID=0, Status='live', Symbol='LOC105378774', Aliases='', description='uncharacterized LOC105378774', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64369688, end_position_on_the_genomic_accession=64383571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100616492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616492, CurrentID=0, Status='live', Symbol='MIR378F', Aliases='mir-378f', description='microRNA 378f', other_designations='hsa-mir-378f', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23929070, end_position_on_the_genomic_accession=23929147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100529111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529111, CurrentID=0, Status='live', Symbol='LOC100529111', Aliases='', description='N-acetylneuraminate pyruvate lyase (dihydrodipicolinate synthase) pseudogene', other_designations='', map_location='2p22', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38769266, end_position_on_the_genomic_accession=38770417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112398, CurrentID=0, Status='live', Symbol='EGLN2', Aliases='EIT6, HIF-PH1, HIFPH1, HPH-1, HPH-3, PHD1', description='egl-9 family hypoxia-inducible factor 2', other_designations='HIF-prolyl hydroxylase 1|estrogen-induced tag 6|hypoxia-inducible factor prolyl hydroxylase 1|prolyl hydroxylase domain-containing protein 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40799143, end_position_on_the_genomic_accession=40808441, orientation='plus', exon_count=7, OMIM=606424, no_hdr0=''),
  112399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112399, CurrentID=0, Status='live', Symbol='EGLN3', Aliases='HIFP4H3, HIFPH3, PHD3', description='egl-9 family hypoxia-inducible factor 3', other_designations='HIF prolyl hydroxylase 3|HIF-PH3|HIF-prolyl hydroxylase 3|HPH-1|HPH-3|egl nine-like protein 3 isoform|hypoxia-inducible factor prolyl hydroxylase 3|prolyl hydroxylase domain-containing protein 3', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=33924215, end_position_on_the_genomic_accession=33951083, orientation='minus', exon_count=5, OMIM=606426, no_hdr0=''),
  112401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112401, CurrentID=0, Status='live', Symbol='BIRC8', Aliases='ILP-2, ILP2, hILP2', description='baculoviral IAP repeat containing 8', other_designations='IAP-like protein 2|baculoviral IAP repeat-containing 8|inhibitor of apoptosis-like protein 2|testis-specific inhibitor of apoptosis', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53289601, end_position_on_the_genomic_accession=53291622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928385, CurrentID=0, Status='live', Symbol='LOC101928385', Aliases='', description='uncharacterized LOC101928385', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32052906, end_position_on_the_genomic_accession=32053335, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100616494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616494, CurrentID=0, Status='live', Symbol='MIR1269B', Aliases='', description='microRNA 1269b', other_designations='hsa-mir-1269b', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12917268, end_position_on_the_genomic_accession=12917342, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100616495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616495, CurrentID=0, Status='live', Symbol='MIR2392', Aliases='', description='microRNA 2392', other_designations='hsa-mir-2392', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100814491, end_position_on_the_genomic_accession=100814574, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100616496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616496, CurrentID=0, Status='live', Symbol='ERVW-4', Aliases='C187-23, pol', description='endogenous retrovirus group W member 4', other_designations='endogenous retrovirus group W, member 4|putative reverse transcriptase', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378779, CurrentID=0, Status='live', Symbol='LOC105378779', Aliases='', description='uncharacterized LOC105378779', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67398498, end_position_on_the_genomic_accession=67399826, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100533947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533947, CurrentID=0, Status='live', Symbol='LOC100533947', Aliases='', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421597, CurrentID=0, Status='live', Symbol='LOC100421597', Aliases='', description='peptidylprolyl isomerase like 4 pseudogene', other_designations='peptidylprolyl isomerase (cyclophilin)-like 4 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7980314, end_position_on_the_genomic_accession=7982229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84273, CurrentID=0, Status='live', Symbol='NOA1', Aliases='C4orf14, MTG3, hAtNOS1, hNOA1, mAtNOS1', description='nitric oxide associated 1', other_designations='mitochondrial GTPase 3 homolog|nitric oxide associated-1|nitric oxide synthase, mitochondrial (putative)|putative ortholog of Arabidopsis mitochondrial nitric oxide synthase', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56963350, end_position_on_the_genomic_accession=56977660, orientation='minus', exon_count=7, OMIM=614919, no_hdr0=''),
  374569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374569, CurrentID=0, Status='live', Symbol='ASPG', Aliases='C14orf76, GPA/WT', description='asparaginase', other_designations='60-kDa-lysophospholipase|asparaginase homolog|wild-type L-asparaginase', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104085670, end_position_on_the_genomic_accession=104112761, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100616498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616498, CurrentID=0, Status='live', Symbol='MIR378E', Aliases='mir-378e', description='microRNA 378e', other_designations='hsa-mir-378e', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170028488, end_position_on_the_genomic_accession=170028566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376357, CurrentID=0, Status='live', Symbol='LOC105376357', Aliases='', description='uncharacterized LOC105376357', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3397447, end_position_on_the_genomic_accession=3398434, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928386, CurrentID=0, Status='live', Symbol='LOC101928386', Aliases='', description='uncharacterized LOC101928386', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143937067, end_position_on_the_genomic_accession=143964156, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927219, CurrentID=0, Status='live', Symbol='HNF4A-AS1', Aliases='uc002xlx', description='HNF4A antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44377743, end_position_on_the_genomic_accession=44395706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100529118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529118, CurrentID=0, Status='live', Symbol='DELXP22.11', Aliases='AUTSX4, CXDELp22.11', description='Autism, X-linked, susceptibility to, 4', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300830, no_hdr0=''),
  100616500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616500, CurrentID=0, Status='live', Symbol='MIR4683', Aliases='', description='microRNA 4683', other_designations='hsa-mir-4683', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35641172, end_position_on_the_genomic_accession=35641252, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378783, CurrentID=0, Status='live', Symbol='LOC105378783', Aliases='', description='uncharacterized LOC105378783', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68892377, end_position_on_the_genomic_accession=68904648, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927221, CurrentID=0, Status='live', Symbol='UBR5-AS1', Aliases='', description='UBR5 antisense RNA 1', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102239451, end_position_on_the_genomic_accession=102252174, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  767811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767811, CurrentID=0, Status='live', Symbol='H2BFXP', Aliases='', description='H2B histone family member X, pseudogene', other_designations='H2B histone family, member X, pseudogene', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103975924, end_position_on_the_genomic_accession=103978357, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927222, CurrentID=0, Status='live', Symbol='GAS2L1P1', Aliases='', description='growth arrest specific 2 like 1 pseudogene 1', other_designations='growth arrest-specific 2 like 1 pseudogene 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38486658, end_position_on_the_genomic_accession=38488210, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378785, CurrentID=0, Status='live', Symbol='LOC105378785', Aliases='', description='uncharacterized LOC105378785', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69055864, end_position_on_the_genomic_accession=69192476, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928387, CurrentID=0, Status='live', Symbol='LOC101928387', Aliases='', description='uncharacterized LOC101928387', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19590781, end_position_on_the_genomic_accession=19673089, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100507277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507277, CurrentID=0, Status='live', Symbol='LOC100507277', Aliases='', description='uncharacterized LOC100507277', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101442079, end_position_on_the_genomic_accession=101444315, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  112464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112464, CurrentID=0, Status='live', Symbol='PRKCDBP', Aliases='CAVIN3, HSRBC, SRBC, cavin-3', description='protein kinase C, delta binding protein', other_designations='sdr-related gene product that binds to c-kinase|serum deprivation response factor-related gene product that binds to C-kinase', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6318946, end_position_on_the_genomic_accession=6320510, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927224, CurrentID=0, Status='live', Symbol='LINC01049', Aliases='', description='long intergenic non-protein coding RNA 1049', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90493288, end_position_on_the_genomic_accession=90535342, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927225, CurrentID=0, Status='live', Symbol='LOC101927225', Aliases='', description='uncharacterized LOC101927225', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72682360, end_position_on_the_genomic_accession=72693241, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480582, CurrentID=0, Status='live', Symbol='RNU6-454P', Aliases='', description='RNA, U6 small nuclear 454, pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378788, CurrentID=0, Status='live', Symbol='LOC105378788', Aliases='', description='uncharacterized LOC105378788', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69431251, end_position_on_the_genomic_accession=69435400, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  374618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374618, CurrentID=0, Status='live', Symbol='TEX9', Aliases='', description='testis expressed 9', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56244085, end_position_on_the_genomic_accession=56459738, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  112476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112476, CurrentID=0, Status='live', Symbol='PRRT2', Aliases='BFIC2, BFIS2, DSPB3, DYT10, EKD1, FICCA, ICCA, IFITMD1, PKC', description='proline-rich transmembrane protein 2', other_designations='dispanin subfamily B member 3|dystonia 10|infantile convulsions and paroxysmal choreoathetosis|interferon induced transmembrane protein domain containing 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29811394, end_position_on_the_genomic_accession=29815881, orientation='plus', exon_count=4, OMIM=614386, no_hdr0=''),
  101927226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927226, CurrentID=0, Status='live', Symbol='LOC101927226', Aliases='', description='uncharacterized LOC101927226', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59616289, end_position_on_the_genomic_accession=59635536, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  112479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112479, CurrentID=0, Status='live', Symbol='ERI2', Aliases='EXOD1, ZGRF5', description='ERI1 exoribonuclease family member 2', other_designations='enhanced RNAi three prime mRNA exonuclease homolog 2|exonuclease domain containing 1|exonuclease domain-containing protein 1|exoribonuclease 2|zinc finger, GRF-type containing 5', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20780193, end_position_on_the_genomic_accession=20806473, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  112483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112483, CurrentID=0, Status='live', Symbol='SAT2', Aliases='SSAT2', description='spermidine/spermine N1-acetyltransferase family member 2', other_designations='diamine N-acetyltransferase 2|polyamine N-acetyltransferase 2|spermidine/spermine N(1)-acetyltransferase 2|spermidine/spermine N1-acetyltransferase 2|thialysine N-epsilon-acetyltransferase', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7626234, end_position_on_the_genomic_accession=7627876, orientation='minus', exon_count=6, OMIM=611463, no_hdr0=''),
  767846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767846, CurrentID=0, Status='live', Symbol='PFN1P2', Aliases='C1orf152, COAS3', description='profilin 1 pseudogene 2', other_designations='chromosome 1 amplified sequence 3', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120432204, end_position_on_the_genomic_accession=120434109, orientation='minus', exon_count=1, OMIM=608609, no_hdr0=''),
  112487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112487, CurrentID=0, Status='live', Symbol='DTD2', Aliases='C14orf126', description='D-tyrosyl-tRNA deacylase 2 (putative)', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31446036, end_position_on_the_genomic_accession=31457474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927228, CurrentID=0, Status='live', Symbol='LOC101927228', Aliases='', description='uncharacterized LOC101927228', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49466106, end_position_on_the_genomic_accession=49469875, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  767850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767850, CurrentID=0, Status='live', Symbol='PFN1P9', Aliases='', description='profilin 1 pseudogene 9', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119853205, end_position_on_the_genomic_accession=119853988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  767851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767851, CurrentID=0, Status='live', Symbol='PFN1P4', Aliases='PFN1P5', description='profilin 1 pseudogene 4', other_designations='profilin 1 pseudogene 5', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148129360, end_position_on_the_genomic_accession=148130138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  767853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=767853, CurrentID=0, Status='live', Symbol='PFN1P10', Aliases='', description='profilin 1 pseudogene 10', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21459424, end_position_on_the_genomic_accession=21460202, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  112495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112495, CurrentID=0, Status='live', Symbol='GTF3C6', Aliases='C6orf51, TFIIIC35, bA397G5.3', description='general transcription factor IIIC subunit 6', other_designations='TFIIIC 35 kDa subunit|general transcription factor IIIC, polypeptide 6, alpha 35kDa|transcription factor IIIC 35 kDa subunit|transcription factor IIIC 35kDa|transcription factor IIIC subunit 6', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110958560, end_position_on_the_genomic_accession=110967888, orientation='plus', exon_count=6, OMIM=611784, no_hdr0=''),
  101927230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927230, CurrentID=0, Status='live', Symbol='LOC101927230', Aliases='', description='uncharacterized LOC101927230', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50214694, end_position_on_the_genomic_accession=50215420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480583, CurrentID=0, Status='live', Symbol='RNU6-480P', Aliases='', description='RNA, U6 small nuclear 480, pseudogene', other_designations='', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100271459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271459, CurrentID=0, Status='live', Symbol='RPL23AP30', Aliases='RPL23A_12_308', description='ribosomal protein L23a pseudogene 30', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201066691, end_position_on_the_genomic_accession=201067191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  374650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374650, CurrentID=0, Status='live', Symbol='GOLGA6L5P', Aliases='GOLGA6L5, GOLGA6L8', description='golgin A6 family-like 5, pseudogene', other_designations='Golgin subfamily A member 6-like protein 8|golgi autoantigen, golgin subfamily a, 6-like 5 (pseudogene)|golgi autoantigen, golgin subfamily a, 6-like 8|golgi autoantigen, golgin subfamily a-like pseudogene|golgin A6 family-like 5 (pseudogene)', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84506168, end_position_on_the_genomic_accession=84516847, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100288831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288831, CurrentID=0, Status='live', Symbol='SRGAP3-AS3', Aliases='', description='SRGAP3 antisense RNA 3', other_designations='SRGAP3 antisense RNA 3 (non-protein coding)', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9216895, end_position_on_the_genomic_accession=9219508, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  374654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374654, CurrentID=0, Status='live', Symbol='KIF7', Aliases='ACLS, HLS2, JBTS12, UNQ340', description='kinesin family member 7', other_designations='EQYK340', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89608913, end_position_on_the_genomic_accession=89663074, orientation='minus', exon_count=24, OMIM=611254, no_hdr0=''),
  374655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374655, CurrentID=0, Status='live', Symbol='ZNF710', Aliases='', description='zinc finger protein 710', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89999046, end_position_on_the_genomic_accession=90082200, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  374659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374659, CurrentID=0, Status='live', Symbol='HDDC3', Aliases='(ppGpp)ase, MESH1', description='HD domain containing 3', other_designations="HD domain-containing protein 3|metazoan SpoT homolog 1|penta-phosphate guanosine-3'-pyrophosphohydrolase", map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90930918, end_position_on_the_genomic_accession=90932569, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927233, CurrentID=0, Status='live', Symbol='SEMA6A-AS1', Aliases='CTB-118N6.3', description='SEMA6A antisense RNA 1', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116447498, end_position_on_the_genomic_accession=116471673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376360, CurrentID=0, Status='live', Symbol='LOC105376360', Aliases='', description='uncharacterized LOC105376360', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3748505, end_position_on_the_genomic_accession=3750764, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  374666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374666, CurrentID=0, Status='live', Symbol='WASH3P', Aliases='FAM39DP', description='WAS protein family homolog 3 pseudogene', other_designations='family with sequence similarity 39, member D pseudogene', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101960813, end_position_on_the_genomic_accession=101976605, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105375934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375934, CurrentID=0, Status='live', Symbol='LOC105375934', Aliases='', description='uncharacterized LOC105375934', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=84701239, end_position_on_the_genomic_accession=84742581, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927234, CurrentID=0, Status='live', Symbol='LOC101927234', Aliases='', description='uncharacterized LOC101927234', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72514009, end_position_on_the_genomic_accession=72527998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480584, CurrentID=0, Status='live', Symbol='RNU6-496P', Aliases='', description='RNA, U6 small nuclear 496, pseudogene', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  374677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374677, CurrentID=0, Status='live', Symbol='WASH4P', Aliases='CXYorf1P, FAM39CP', description='WAS protein family homolog 4 pseudogene', other_designations='Putative WAS protein family homolog 4|family with sequence similarity 39, member C pseudogene|pseudogene of CXYorf1', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14079, end_position_on_the_genomic_accession=18091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927237, CurrentID=0, Status='live', Symbol='LOC101927237', Aliases='', description='uncharacterized LOC101927237', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67417305, end_position_on_the_genomic_accession=67422000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927238, CurrentID=0, Status='live', Symbol='LOC101927238', Aliases='', description='uncharacterized LOC101927238', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80200056, end_position_on_the_genomic_accession=80211350, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927239, CurrentID=0, Status='live', Symbol='LOC101927239', Aliases='', description='uncharacterized LOC101927239', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34169376, end_position_on_the_genomic_accession=34183632, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480585, CurrentID=0, Status='live', Symbol='RNU6-508P', Aliases='', description='RNA, U6 small nuclear 508, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378803, CurrentID=0, Status='live', Symbol='LOC105378803', Aliases='', description='leucine-rich repeat-containing protein 53-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74479955, end_position_on_the_genomic_accession=74537447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100419913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419913, CurrentID=0, Status='live', Symbol='LOC100419913', Aliases='', description='zinc finger protein 678 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50471769, end_position_on_the_genomic_accession=50472438, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378804, CurrentID=0, Status='live', Symbol='LOC105378804', Aliases='', description='leucine-rich repeat-containing protein 53', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74469617, end_position_on_the_genomic_accession=74476533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927242, CurrentID=0, Status='live', Symbol='LINC01430', Aliases='', description='long intergenic non-protein coding RNA 1430', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44448777, end_position_on_the_genomic_accession=44450557, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  112574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112574, CurrentID=0, Status='live', Symbol='SNX18', Aliases='SH3PX2, SH3PXD3B, SNAG1', description='sorting nexin 18', other_designations='SH3 and PX domain-containing protein 3B|sorting nexin associated golgi protein 1|sorting nexin-associated Golgi protein 1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54517759, end_position_on_the_genomic_accession=54546586, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390133, CurrentID=0, Status='live', Symbol='OR4A2P', Aliases='', description='olfactory receptor family 4 subfamily A member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 2 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54667299, end_position_on_the_genomic_accession=54668211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927243, CurrentID=0, Status='live', Symbol='LOC101927243', Aliases='', description='uncharacterized LOC101927243', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77415198, end_position_on_the_genomic_accession=77425444, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378806, CurrentID=0, Status='live', Symbol='LOC105378806', Aliases='', description='uncharacterized LOC105378806', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75964917, end_position_on_the_genomic_accession=76002941, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928391, CurrentID=0, Status='live', Symbol='LOC101928391', Aliases='', description='uncharacterized LOC101928391', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27031757, end_position_on_the_genomic_accession=27063997, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927244, CurrentID=0, Status='live', Symbol='LOC101927244', Aliases='', description='uncharacterized LOC101927244', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70706453, end_position_on_the_genomic_accession=70786468, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927245, CurrentID=0, Status='live', Symbol='LOC101927245', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102654318, end_position_on_the_genomic_accession=102730395, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  106480586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480586, CurrentID=0, Status='live', Symbol='RNU6-514P', Aliases='', description='RNA, U6 small nuclear 514, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390136, CurrentID=0, Status='live', Symbol='OR4A11P', Aliases='', description='olfactory receptor family 4 subfamily A member 11 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 11 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55318561, end_position_on_the_genomic_accession=55319478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  374739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374739, CurrentID=0, Status='live', Symbol='TEPP', Aliases='', description='testis, prostate and placenta expressed', other_designations='testis/prostate/placenta-expressed protein', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57976435, end_position_on_the_genomic_accession=57988354, orientation='plus', exon_count=8, OMIM=610264, no_hdr0=''),
  112597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112597, CurrentID=0, Status='live', Symbol='LINC00152', Aliases='C2orf59, NCRNA00152', description='long intergenic non-protein coding RNA 152', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87455455, end_position_on_the_genomic_accession=87521518, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390137, CurrentID=0, Status='live', Symbol='OR4A9P', Aliases='', description='olfactory receptor family 4 subfamily A member 9 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 9 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55388530, end_position_on_the_genomic_accession=55389473, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927247, CurrentID=0, Status='live', Symbol='LOC101927247', Aliases='', description='uncharacterized LOC101927247', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226186810, end_position_on_the_genomic_accession=226213192, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390138, CurrentID=0, Status='live', Symbol='OR4A10P', Aliases='OR4A25P', description='olfactory receptor family 4 subfamily A member 10 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 10 pseudogene|olfactory receptor, family 4, subfamily A, member 25 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55431958, end_position_on_the_genomic_accession=55432368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112609, CurrentID=0, Status='live', Symbol='MRAP2', Aliases='BMIQ18, C6orf117, bA51G5.2', description='melanocortin 2 receptor accessory protein 2', other_designations='MC2R accessory protein 2', map_location='6q14.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84024786, end_position_on_the_genomic_accession=84146282, orientation='plus', exon_count=12, OMIM=615410, no_hdr0=''),
  112611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112611, CurrentID=0, Status='live', Symbol='RWDD2A', Aliases='RWDD2, dJ747H23.2', description='RWD domain containing 2A', other_designations='RWD domain containing 2', map_location='6q14.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83193313, end_position_on_the_genomic_accession=83198932, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927249, CurrentID=0, Status='live', Symbol='PDZRN3-AS1', Aliases='', description='PDZRN3 antisense RNA 1', other_designations='PDZRN3 antisense RNA 1 (non-protein coding)', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73623568, end_position_on_the_genomic_accession=73627899, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  112616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112616, CurrentID=0, Status='live', Symbol='CMTM7', Aliases='CKLFSF7', description='CKLF-like MARVEL transmembrane domain containing 7', other_designations='chemokine-like factor super family 7|chemokine-like factor super family member 7 variant 2|chemokine-like factor superfamily 7|chemokine-like factor superfamily member 7', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32391671, end_position_on_the_genomic_accession=32454841, orientation='plus', exon_count=6, OMIM=607890, no_hdr0=''),
  105378812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378812, CurrentID=0, Status='live', Symbol='LOC105378812', Aliases='', description='uncharacterized LOC105378812', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=79278362, end_position_on_the_genomic_accession=79286728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100616530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616530, CurrentID=0, Status='live', Symbol='C8orf37-AS1', Aliases='', description='C8orf37 antisense RNA 1', other_designations='KB-1047C11.2|tospeak', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95268836, end_position_on_the_genomic_accession=95810143, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378813, CurrentID=0, Status='live', Symbol='LOC105378813', Aliases='', description='uncharacterized LOC105378813', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81158404, end_position_on_the_genomic_accession=81177845, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  374768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374768, CurrentID=0, Status='live', Symbol='SPEM1', Aliases='C17orf83', description='spermatid maturation 1', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7420360, end_position_on_the_genomic_accession=7421632, orientation='plus', exon_count=3, OMIM=615116, no_hdr0=''),
  105378814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378814, CurrentID=0, Status='live', Symbol='LOC105378814', Aliases='', description='uncharacterized LOC105378814', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=82621530, end_position_on_the_genomic_accession=82656278, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927252, CurrentID=0, Status='live', Symbol='LOC101927252', Aliases='', description='uncharacterized LOC101927252', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189338594, end_position_on_the_genomic_accession=189441118, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378815, CurrentID=0, Status='live', Symbol='LOC105378815', Aliases='', description='uncharacterized LOC105378815', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=82814123, end_position_on_the_genomic_accession=82853644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103106901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103106901, CurrentID=0, Status='live', Symbol='IGHV3-43D', Aliases='', description='immunoglobulin heavy variable 3-43D', other_designations='', map_location='14q32.33 not on reference assembly', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378816, CurrentID=0, Status='live', Symbol='LOC105378816', Aliases='', description='uncharacterized LOC105378816', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83287471, end_position_on_the_genomic_accession=83315236, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  374786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374786, CurrentID=0, Status='live', Symbol='EFCAB5', Aliases='', description='EF-hand calcium binding domain 5', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29929856, end_position_on_the_genomic_accession=30108452, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  392193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392193, CurrentID=0, Status='live', Symbol='RPL19P13', Aliases='RPL19_8_875', description='ribosomal protein L19 pseudogene 13', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11255990, end_position_on_the_genomic_accession=11256678, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103106902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103106902, CurrentID=0, Status='live', Symbol='ETL6', Aliases='', description='Epilepsy, familial temporal lobe, 6', other_designations='', map_location='3q25-q26', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615697, no_hdr0=''),
  103106903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103106903, CurrentID=0, Status='live', Symbol='LINC01276', Aliases='', description='long intergenic non-protein coding RNA 1276', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41502444, end_position_on_the_genomic_accession=41519852, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480588, CurrentID=0, Status='live', Symbol='RNU6-542P', Aliases='', description='RNA, U6 small nuclear 542, pseudogene', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927256, CurrentID=0, Status='live', Symbol='LOC101927256', Aliases='', description='uncharacterized LOC101927256', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2944554, end_position_on_the_genomic_accession=2947091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  1623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=1623, CurrentID=0, Status='live', Symbol='DBIP2', Aliases='DBIL1', description='diazepam binding inhibitor (GABA receptor modulator, acyl-CoA binding protein) pseudogene 2', other_designations='diazepam binding inhibitor-like 1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80603162, end_position_on_the_genomic_accession=80603631, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927257, CurrentID=0, Status='live', Symbol='LOC101927257', Aliases='', description='uncharacterized LOC101927257', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40521800, end_position_on_the_genomic_accession=40526707, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378820, CurrentID=0, Status='live', Symbol='LOC105378820', Aliases='', description='uncharacterized LOC105378820', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84630543, end_position_on_the_genomic_accession=84637724, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927258, CurrentID=0, Status='live', Symbol='LINC01474', Aliases='', description='long intergenic non-protein coding RNA 1474', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72871728, end_position_on_the_genomic_accession=72874109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928394, CurrentID=0, Status='live', Symbol='LOC101928394', Aliases='', description='uncharacterized LOC101928394', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71059580, end_position_on_the_genomic_accession=71162147, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  374819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374819, CurrentID=0, Status='live', Symbol='LRRC37A3', Aliases='LRRC37, LRRC37A', description='leucine-rich repeat containing 37 member A3', other_designations='leucine rich repeat containing 37, member A3', map_location='17q24.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64854147, end_position_on_the_genomic_accession=64919480, orientation='minus', exon_count=14, OMIM=616557, no_hdr0=''),
  101927260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927260, CurrentID=0, Status='live', Symbol='LOC101927260', Aliases='', description='uncharacterized LOC101927260', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111221749, end_position_on_the_genomic_accession=111248237, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378823, CurrentID=0, Status='live', Symbol='LOC105378823', Aliases='', description='uncharacterized LOC105378823', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85717695, end_position_on_the_genomic_accession=85728232, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378824, CurrentID=0, Status='live', Symbol='LOC105378824', Aliases='', description='uncharacterized LOC105378824', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86265375, end_position_on_the_genomic_accession=86320225, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927262, CurrentID=0, Status='live', Symbol='LOC101927262', Aliases='', description='melanoma-associated antigen C1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=286352, end_position_on_the_genomic_accession=294727, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100419853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419853, CurrentID=0, Status='live', Symbol='LOC100419853', Aliases='', description='transmembrane protein 69 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180967848, end_position_on_the_genomic_accession=180968567, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927263, CurrentID=0, Status='live', Symbol='LOC101927263', Aliases='', description='uncharacterized LOC101927263', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96354350, end_position_on_the_genomic_accession=96395016, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378826, CurrentID=0, Status='live', Symbol='LOC105378826', Aliases='', description='uncharacterized LOC105378826', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86520363, end_position_on_the_genomic_accession=86523557, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928395, CurrentID=0, Status='live', Symbol='LOC101928395', Aliases='', description='uncharacterized LOC101928395', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112703, CurrentID=0, Status='live', Symbol='FAM71E1', Aliases='', description='family with sequence similarity 71 member E1', other_designations='family with sequence similarity 71, member E1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50466643, end_position_on_the_genomic_accession=50476746, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927264, CurrentID=0, Status='live', Symbol='LOC101927264', Aliases='', description='UDP-glucuronosyltransferase 2B10-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68914732, end_position_on_the_genomic_accession=68924749, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378827, CurrentID=0, Status='live', Symbol='LOC105378827', Aliases='', description='uncharacterized LOC105378827', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86500097, end_position_on_the_genomic_accession=86503813, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927265, CurrentID=0, Status='live', Symbol='BNIP3P7', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18487322, end_position_on_the_genomic_accession=18492460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480590, CurrentID=0, Status='live', Symbol='RNU6-560P', Aliases='', description='RNA, U6 small nuclear 560, pseudogene', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112714, CurrentID=0, Status='live', Symbol='TUBA3E', Aliases='', description='tubulin alpha 3e', other_designations='alpha-tubulin 3E|tubulin, alpha 3e', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130191745, end_position_on_the_genomic_accession=130198461, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  374860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374860, CurrentID=0, Status='live', Symbol='ANKRD30B', Aliases='NY-BR-1.1', description='ankyrin repeat domain 30B', other_designations='breast cancer antigen NY-BR-1.1|serologically defined breast cancer antigen NY-BR-1.1', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14748160, end_position_on_the_genomic_accession=14855465, orientation='plus', exon_count=46, OMIM=616565, no_hdr0=''),
  105378829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378829, CurrentID=0, Status='live', Symbol='LOC105378829', Aliases='', description='uncharacterized LOC105378829', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86618449, end_position_on_the_genomic_accession=86623132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  374864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374864, CurrentID=0, Status='live', Symbol='CCDC178', Aliases='C18orf34', description='coiled-coil domain containing 178', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32937402, end_position_on_the_genomic_accession=33441040, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  374868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374868, CurrentID=0, Status='live', Symbol='ATP9B', Aliases='ATPASEP, ATPIIB, HUSSY-20, NEO1L, hMMR1', description='ATPase, class II, type 9B', other_designations='ATPase type IV, phospholipid transporting (P-type)|macrophage MHC receptor 1', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79047594, end_position_on_the_genomic_accession=79378283, orientation='plus', exon_count=41, OMIM=614446, no_hdr0=''),
  105378830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378830, CurrentID=0, Status='live', Symbol='LOC105378830', Aliases='', description='uncharacterized LOC105378830', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86571249, end_position_on_the_genomic_accession=86698263, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  768086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768086, CurrentID=0, Status='live', Symbol='TREHP1', Aliases='', description='trehalase pseudogene 1', other_designations='trehalase (brush-border membrane glycoprotein) pseudogene', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118687976, end_position_on_the_genomic_accession=118690600, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  374872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374872, CurrentID=0, Status='live', Symbol='C19orf35', Aliases='', description='chromosome 19 open reading frame 35', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2274632, end_position_on_the_genomic_accession=2282182, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927268, CurrentID=0, Status='live', Symbol='LOC101927268', Aliases='', description='uncharacterized LOC101927268', other_designations='XXbac-BPGBPG55C20.1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57855891, end_position_on_the_genomic_accession=57867101, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  374875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374875, CurrentID=0, Status='live', Symbol='HSD11B1L', Aliases='11-DH3, 11-beta-HSD3, HSD3, SCDR10, SCDR10B, SDR26C2', description='hydroxysteroid (11-beta) dehydrogenase 1-like', other_designations='11-beta-hydroxysteroid dehydrogenase type 3|short chain dehydrogenase/reductase 10|short chain dehydrogenase/reductase family 26C member 2|short chain dehydrogenase/reductase family 26C, member 2', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5680765, end_position_on_the_genomic_accession=5688523, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  374877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374877, CurrentID=0, Status='live', Symbol='C19orf45', Aliases='', description='chromosome 19 open reading frame 45', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7497559, end_position_on_the_genomic_accession=7515233, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  374879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374879, CurrentID=0, Status='live', Symbol='ZNF699', Aliases='hang', description='zinc finger protein 699', other_designations='hangover homolog|hangover, Drosophila, homolog of', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9293925, end_position_on_the_genomic_accession=9309766, orientation='minus', exon_count=6, OMIM=609571, no_hdr0=''),
  768096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768096, CurrentID=0, Status='live', Symbol='HAR1A', Aliases='HAR1F, LINC00064, NCRNA00064', description='highly accelerated region 1A (non-protein coding)', other_designations='highly accelerated region 1A (non-protein-coding RNA)|human accelerated region 1, forward|long intergenic non-protein coding RNA 64', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63101292, end_position_on_the_genomic_accession=63104385, orientation='plus', exon_count=2, OMIM=610556, no_hdr0=''),
  768097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768097, CurrentID=0, Status='live', Symbol='HAR1B', Aliases='HAR1R, LINC00065, NCRNA00065', description='highly accelerated region 1B (non-protein coding)', other_designations='highly accelerated region 1B (non-protein-coding RNA)|human accelerated region 1, reverse|long intergenic non-protein coding RNA 65', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63095493, end_position_on_the_genomic_accession=63102319, orientation='minus', exon_count=2, OMIM=610557, no_hdr0=''),
  374882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374882, CurrentID=0, Status='live', Symbol='TMEM205', Aliases='UNQ501', description='transmembrane protein 205', other_designations='MBC3205', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11342776, end_position_on_the_genomic_accession=11346414, orientation='minus', exon_count=5, OMIM=613771, no_hdr0=''),
  346470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346470, CurrentID=0, Status='live', Symbol='MRPL42P4', Aliases='', description='mitochondrial ribosomal protein L42 pseudogene 4', other_designations='', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47026122, end_position_on_the_genomic_accession=47026472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  374887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374887, CurrentID=0, Status='live', Symbol='YJEFN3', Aliases='', description='YjeF N-terminal domain containing 3', other_designations='apolipoprotein A1 binding protein|hYjeF_N3|hYjeF_N3-19p13.11|yjeF_N3', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19528861, end_position_on_the_genomic_accession=19537584, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  112744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112744, CurrentID=0, Status='live', Symbol='IL17F', Aliases='CANDF6, IL-17F, ML-1, ML1', description='interleukin 17F', other_designations='cytokine ML-1', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52236681, end_position_on_the_genomic_accession=52245689, orientation='minus', exon_count=4, OMIM=606496, no_hdr0=''),
  374890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374890, CurrentID=0, Status='live', Symbol='LOC374890', Aliases='', description='uncharacterized LOC374890', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22615557, end_position_on_the_genomic_accession=22623971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422637, CurrentID=0, Status='live', Symbol='LOC100422637', Aliases='', description='MX dynamin-like GTPase 1 pseudogene', other_designations='myxovirus (influenza virus) resistance 1, interferon-inducible protein p78 (mouse) pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10284951, end_position_on_the_genomic_accession=10285708, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378834, CurrentID=0, Status='live', Symbol='LOC105378834', Aliases='', description='uncharacterized LOC105378834', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87446491, end_position_on_the_genomic_accession=87451804, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  112752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112752, CurrentID=0, Status='live', Symbol='IFT43', Aliases='C14orf179, CED3', description='intraflagellar transport 43', other_designations='IFT complex A subunit|intraflagellar transport 43 homolog', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75985753, end_position_on_the_genomic_accession=76084073, orientation='plus', exon_count=11, OMIM=614068, no_hdr0=''),
  374897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374897, CurrentID=0, Status='live', Symbol='SBSN', Aliases='UNQ698', description='suprabasin', other_designations='HLAR698', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35523367, end_position_on_the_genomic_accession=35528351, orientation='minus', exon_count=6, OMIM=609969, no_hdr0=''),
  112755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112755, CurrentID=0, Status='live', Symbol='STX1B', Aliases='GEFSP91, STX1B2, STX1B', description='syntaxin 1B', other_designations='syntaxin-1B1|syntaxin-1B2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30989256, end_position_on_the_genomic_accession=31010638, orientation='minus', exon_count=10, OMIM=601485, no_hdr0=''),
  374900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374900, CurrentID=0, Status='live', Symbol='ZNF568', Aliases='', description='zinc finger protein 568', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36916329, end_position_on_the_genomic_accession=36997932, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101927273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927273, CurrentID=0, Status='live', Symbol='LINC01416', Aliases='', description='long intergenic non-protein coding RNA 1416', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55881688, end_position_on_the_genomic_accession=55920199, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928397, CurrentID=0, Status='live', Symbol='LOC101928397', Aliases='', description='uncharacterized LOC101928397', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143363038, end_position_on_the_genomic_accession=143364230, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  374907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374907, CurrentID=0, Status='live', Symbol='B3GNT8', Aliases='B3GALT7, BGALT15, beta3Gn-T8', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 8', other_designations='BGnT-8|UDP-Gal:betaGal beta 1,3-galactosyltransferase polypeptide 7|beta galactosyltransferase BGALT15|beta-1,3-Gn-T8|beta-1,3-N-acetylglucosaminyltransferase 8|beta1,3-N-acetylglucosaminyltransferase 8', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41425359, end_position_on_the_genomic_accession=41428856, orientation='minus', exon_count=2, OMIM=615357, no_hdr0=''),
  101927274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927274, CurrentID=0, Status='live', Symbol='LOC101927274', Aliases='', description='uncharacterized LOC101927274', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51312609, end_position_on_the_genomic_accession=51335162, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  112770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112770, CurrentID=0, Status='live', Symbol='GLMP', Aliases='C1orf85, NCU-G1', description='glycosylated lysosomal membrane protein', other_designations='kidney lysosomal membrane protein|kidney predominant protein NCU-G1|lysosomal protein NCU-G1', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156292687, end_position_on_the_genomic_accession=156295689, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927275, CurrentID=0, Status='live', Symbol='NDUFV2-AS1', Aliases='', description='NDUFV2 antisense RNA 1', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9121265, end_position_on_the_genomic_accession=9136720, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378838, CurrentID=0, Status='live', Symbol='LOC105378838', Aliases='', description='uncharacterized LOC105378838', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88615287, end_position_on_the_genomic_accession=88648319, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  374918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374918, CurrentID=0, Status='live', Symbol='IGFL1', Aliases='APRG644, UNQ644', description='IGF like family member 1', other_designations='IGF-like family member 1', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46229752, end_position_on_the_genomic_accession=46231243, orientation='plus', exon_count=4, OMIM=610544, no_hdr0=''),
  374920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374920, CurrentID=0, Status='live', Symbol='C19orf68', Aliases='', description='chromosome 19 open reading frame 68', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48170662, end_position_on_the_genomic_accession=48197621, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100421169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421169, CurrentID=0, Status='live', Symbol='LOC100421169', Aliases='', description='thyroid hormone receptor interactor 11 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25071521, end_position_on_the_genomic_accession=25072024, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378839, CurrentID=0, Status='live', Symbol='LOC105378839', Aliases='', description='uncharacterized LOC105378839', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88462888, end_position_on_the_genomic_accession=88465805, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927277, CurrentID=0, Status='live', Symbol='LOC101927277', Aliases='', description='uncharacterized LOC101927277', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31588040, end_position_on_the_genomic_accession=31595720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  374928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374928, CurrentID=0, Status='live', Symbol='ZNF773', Aliases='ZNF419B', description='zinc finger protein 773', other_designations='zinc finger protein 419B', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57499854, end_position_on_the_genomic_accession=57518437, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378840, CurrentID=0, Status='live', Symbol='LOC105378840', Aliases='', description='uncharacterized LOC105378840', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88945082, end_position_on_the_genomic_accession=88947495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927278, CurrentID=0, Status='live', Symbol='LOC101927278', Aliases='', description='uncharacterized LOC101927278', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98446647, end_position_on_the_genomic_accession=98453716, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378841, CurrentID=0, Status='live', Symbol='LOC105378841', Aliases='', description='uncharacterized LOC105378841', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89034791, end_position_on_the_genomic_accession=89102691, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101928398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928398, CurrentID=0, Status='live', Symbol='LOC101928398', Aliases='', description='uncharacterized LOC101928398', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38857539, end_position_on_the_genomic_accession=38858862, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927279, CurrentID=0, Status='live', Symbol='EP300-AS1', Aliases='', description='EP300 antisense RNA 1', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41185214, end_position_on_the_genomic_accession=41197501, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927280, CurrentID=0, Status='live', Symbol='HRAT56', Aliases='', description='heart tissue-associated transcript 56', other_designations='CTD-2281M20.1', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118350966, end_position_on_the_genomic_accession=118562117, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  112802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112802, CurrentID=0, Status='live', Symbol='KRT71', Aliases='HYPT13, K6IRS1, KRT6IRS, KRT6IRS1', description='keratin 71, type II', other_designations='CK-71|K71|cytokeratin-71|hK6irs|hK6irs1|keratin 6 irs|keratin-71|type II inner root sheath-specific keratin-K6irs1|type-II keratin Kb34', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52543909, end_position_on_the_genomic_accession=52553147, orientation='minus', exon_count=9, OMIM=608245, no_hdr0=''),
  101927281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927281, CurrentID=0, Status='live', Symbol='LOC101927281', Aliases='', description='uncharacterized LOC101927281', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73259286, end_position_on_the_genomic_accession=73371260, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  374955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374955, CurrentID=0, Status='live', Symbol='SPATA21', Aliases='spergen-2, spergen2', description='spermatogenesis associated 21', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16387117, end_position_on_the_genomic_accession=16437842, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  112812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112812, CurrentID=0, Status='live', Symbol='FDX1L', Aliases='FDX2', description='ferredoxin 1-like', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10310211, end_position_on_the_genomic_accession=10316015, orientation='minus', exon_count=5, OMIM=614585, no_hdr0=''),
  112817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112817, CurrentID=0, Status='live', Symbol='HOGA1', Aliases='C10orf65, DHDPS2, DHDPSL, HP3, NPL2', description='4-hydroxy-2-oxoglutarate aldolase 1', other_designations='DHDPS-like protein|N-acetylneuraminate pyruvate lyase 2 (putative)|dihydrodipicolinate synthase-like, mitochondrial|dihydrodipicolinate synthetase homolog 2|probable 2-keto-4-hydroxyglutarate aldolase|probable 4-hydroxy-2-oxoglutarate aldolase, mitochondrial|probable KHG-aldolase|protein 569272', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97584345, end_position_on_the_genomic_accession=97612802, orientation='plus', exon_count=7, OMIM=613597, no_hdr0=''),
  101927283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927283, CurrentID=0, Status='live', Symbol='LOC101927283', Aliases='', description='uncharacterized LOC101927283', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111264447, end_position_on_the_genomic_accession=111279486, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378846, CurrentID=0, Status='live', Symbol='LOC105378846', Aliases='', description='uncharacterized LOC105378846', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89900426, end_position_on_the_genomic_accession=89906621, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  374969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374969, CurrentID=0, Status='live', Symbol='SVBP', Aliases='CCDC23', description='small vasohibin binding protein', other_designations='coiled coil domain-containing protein 23|coiled-coil domain containing 23|coiled-coil domain-containing protein 23', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42807052, end_position_on_the_genomic_accession=42817388, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646696, CurrentID=0, Status='live', Symbol='LOC646696', Aliases='', description='MLX, MAX dimerization protein pseudogene', other_designations='MAX-like protein X pseudogene', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223704627, end_position_on_the_genomic_accession=223705337, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  374973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374973, CurrentID=0, Status='live', Symbol='TEX38', Aliases='ATPAF1-AS1, C1orf223, THEG4', description='testis expressed 38', other_designations='ATPAF1 antisense RNA 1|ATPAF1 antisense gene protein 1|testis highly expressed protein 4', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46669346, end_position_on_the_genomic_accession=46673594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100419957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419957, CurrentID=0, Status='live', Symbol='LOC100419957', Aliases='', description='ankyrin repeat domain 57 pseudogene', other_designations='sosondowah ankyrin repeat domain family member C pseudogene', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9082601, end_position_on_the_genomic_accession=9083909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  374977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374977, CurrentID=0, Status='live', Symbol='MROH7', Aliases='C1orf175, HEATR8', description='maestro heat-like repeat family member 7', other_designations='HEAT repeat-containing protein 8', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54641740, end_position_on_the_genomic_accession=54710267, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  101927286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927286, CurrentID=0, Status='live', Symbol='LOC101927286', Aliases='', description='uncharacterized LOC101927286', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97370371, end_position_on_the_genomic_accession=97521811, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  112840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112840, CurrentID=0, Status='live', Symbol='WDR89', Aliases='C14orf150, MSTP050', description='WD repeat domain 89', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63597039, end_position_on_the_genomic_accession=63641923, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  374986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374986, CurrentID=0, Status='live', Symbol='FAM73A', Aliases='', description='family with sequence similarity 73 member A', other_designations='family with sequence similarity 73, member A', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77779622, end_position_on_the_genomic_accession=77879540, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  374987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=374987, CurrentID=0, Status='live', Symbol='NEXN-AS1', Aliases='C1orf118', description='NEXN antisense RNA 1', other_designations='NEXN antisense RNA 1 (non-protein coding)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77881348, end_position_on_the_genomic_accession=77889539, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  768206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768206, CurrentID=0, Status='live', Symbol='PRCD', Aliases='RP36', description='progressive rod-cone degeneration', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76527586, end_position_on_the_genomic_accession=76553580, orientation='plus', exon_count=8, OMIM=610598, no_hdr0=''),
  112849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112849, CurrentID=0, Status='live', Symbol='L3HYPDH', Aliases='C14orf149', description='L-3-hydroxyproline dehydratase (trans-)', other_designations='probable proline racemase|trans-3-hydroxy-l-proline dehydratase|trans-3-hydroxyl-L-proline dehydratase|trans-L-3-hydroxyproline dehydratase', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59465511, end_position_on_the_genomic_accession=59484438, orientation='minus', exon_count=8, OMIM=614811, no_hdr0=''),
  768211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768211, CurrentID=0, Status='live', Symbol='RELL1', Aliases='', description='RELT-like 1', other_designations='receptor expressed in lymphoid tissues like 1|tmp_locus_29', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37588059, end_position_on_the_genomic_accession=37686377, orientation='minus', exon_count=10, OMIM=611212, no_hdr0=''),
  768212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768212, CurrentID=0, Status='live', Symbol='MIR758', Aliases='MIRN758, hsa-mir-758, mir-758', description='microRNA 758', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101026020, end_position_on_the_genomic_accession=101026107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  768213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768213, CurrentID=0, Status='live', Symbol='MIR671', Aliases='MIRN671, hsa-mir-671, mir-671', description='microRNA 671', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151238421, end_position_on_the_genomic_accession=151238538, orientation='plus', exon_count=1, OMIM=615245, no_hdr0=''),
  768214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768214, CurrentID=0, Status='live', Symbol='MIR668', Aliases='MIRN668, hsa-mir-668', description='microRNA 668', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101055258, end_position_on_the_genomic_accession=101055323, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  768215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768215, CurrentID=0, Status='live', Symbol='MIR767', Aliases='MIRN767, hsa-mir-767, mir-767', description='microRNA 767', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152393421, end_position_on_the_genomic_accession=152393529, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  768216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768216, CurrentID=0, Status='live', Symbol='MIR454', Aliases='MIRN454, hsa-mir-454, mir-454', description='microRNA 454', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59137758, end_position_on_the_genomic_accession=59137872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  768217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768217, CurrentID=0, Status='live', Symbol='MIR769', Aliases='MIRN769, hsa-mir-769, mir-769', description='microRNA 769', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46018932, end_position_on_the_genomic_accession=46019049, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  112858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112858, CurrentID=0, Status='live', Symbol='TP53RK', Aliases='BUD32, C20orf64, Nori-2, Nori-2p, PRPK, dJ101A2', description='TP53 regulating kinase', other_designations='EKC/KEOPS complex subunit TP53RK|atypical serine/threonine protein kinase TP53RK|p53-related protein kinase', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46684365, end_position_on_the_genomic_accession=46689637, orientation='minus', exon_count=2, OMIM=608679, no_hdr0=''),
  768219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768219, CurrentID=0, Status='live', Symbol='MIR802', Aliases='MIRN802, hsa-mir-802', description='microRNA 802', other_designations='', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35720715, end_position_on_the_genomic_accession=35720808, orientation='plus', exon_count=1, OMIM=616090, no_hdr0=''),
  768220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768220, CurrentID=0, Status='live', Symbol='MIR765', Aliases='MIRN765, hsa-mir-765, mir-765', description='microRNA 765', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156936131, end_position_on_the_genomic_accession=156936244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  768222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768222, CurrentID=0, Status='live', Symbol='MIR770', Aliases='MIRN770, hsa-mir-770', description='microRNA 770', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100852390, end_position_on_the_genomic_accession=100852487, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378853, CurrentID=0, Status='live', Symbol='LOC105378853', Aliases='', description='uncharacterized LOC105378853', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90782984, end_position_on_the_genomic_accession=90789487, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  112869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112869, CurrentID=0, Status='live', Symbol='SGF29', Aliases='CCDC101, STAF36', description='SAGA complex associated factor 29', other_designations='coiled-coil domain containing 101|coiled-coil domain-containing protein 101', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28553928, end_position_on_the_genomic_accession=28591790, orientation='plus', exon_count=10, OMIM=613374, no_hdr0=''),
  100288892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288892, CurrentID=0, Status='live', Symbol='COX6CP4', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 4', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36571930, end_position_on_the_genomic_accession=36572309, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378855, CurrentID=0, Status='live', Symbol='LOC105378855', Aliases='', description='uncharacterized LOC105378855', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90844356, end_position_on_the_genomic_accession=90851026, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  768239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768239, CurrentID=0, Status='live', Symbol='PSAPL1', Aliases='', description='prosaposin-like 1 (gene/pseudogene)', other_designations='prosaposin-like protein 1', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7430294, end_position_on_the_genomic_accession=7434973, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376372, CurrentID=0, Status='live', Symbol='LOC105376372', Aliases='', description='uncharacterized LOC105376372', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4384422, end_position_on_the_genomic_accession=4406350, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  375027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375027, CurrentID=0, Status='live', Symbol='S100A7P2', Aliases='S100A7L4, s100a7e', description='S100 calcium binding protein A7 pseudogene 2', other_designations='S100 calcium binding protein A7-like 4 (pseudogene)|S100 calcium binding protein A7A pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153444909, end_position_on_the_genomic_accession=153446663, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112885, CurrentID=0, Status='live', Symbol='PHF21B', Aliases='BHC80L, PHF4', description='PHD finger protein 21B', other_designations='PHD finger protein 4', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44868432, end_position_on_the_genomic_accession=45009928, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  375033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375033, CurrentID=0, Status='live', Symbol='PEAR1', Aliases='JEDI, MEGF12', description='platelet endothelial aggregation receptor 1', other_designations='hPEAR1|multiple EGF-like domains protein 12|multiple EGF-like-domains 12|multiple epidermal growth factor-like domains protein 12', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156893692, end_position_on_the_genomic_accession=156916434, orientation='plus', exon_count=25, OMIM=610278, no_hdr0=''),
  375035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375035, CurrentID=0, Status='live', Symbol='SFT2D2', Aliases='UNQ512, dJ747L4.C1.2', description='SFT2 domain containing 2', other_designations='SFT2 domain-containing protein 2', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168226017, end_position_on_the_genomic_accession=168242850, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378858, CurrentID=0, Status='live', Symbol='LOC105378858', Aliases='', description='uncharacterized LOC105378858', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94144809, end_position_on_the_genomic_accession=94161728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372008, CurrentID=0, Status='live', Symbol='LOC105372008', Aliases='', description='uncharacterized LOC105372008', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14758929, end_position_on_the_genomic_accession=14768083, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100419969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419969, CurrentID=0, Status='live', Symbol='LOC100419969', Aliases='', description='myosin light chain kinase pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125680725, end_position_on_the_genomic_accession=125682130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  375056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375056, CurrentID=0, Status='live', Symbol='MIA3', Aliases='ARNT, D320, TANGO, TANGO1, UNQ6077', description='melanoma inhibitory activity family member 3', other_designations='C219-reactive peptide|melanoma inhibitory activity family, member 3|transport and Golgi organization protein 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222618102, end_position_on_the_genomic_accession=222668012, orientation='plus', exon_count=31, OMIM=613455, no_hdr0=''),
  375057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375057, CurrentID=0, Status='live', Symbol='C1orf95', Aliases='', description='chromosome 1 open reading frame 95', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226548248, end_position_on_the_genomic_accession=226609214, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  375061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375061, CurrentID=0, Status='live', Symbol='FAM89A', Aliases='C1orf153', description='family with sequence similarity 89 member A', other_designations='family with sequence similarity 89, member A', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231018958, end_position_on_the_genomic_accession=231040249, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378863, CurrentID=0, Status='live', Symbol='LOC105378863', Aliases='', description='uncharacterized LOC105378863', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95028780, end_position_on_the_genomic_accession=95031531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128122, CurrentID=0, Status='live', Symbol='LOC100128122', Aliases='', description='Cbl proto-oncogene-like 1, E3 ubiquitin protein ligase pseudogene', other_designations='Cbl proto-oncogene, E3 ubiquitin protein ligase-like 1 pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121025279, end_position_on_the_genomic_accession=121026107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372680, CurrentID=0, Status='live', Symbol='LOC105372680', Aliases='', description='uncharacterized LOC105372680', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56142472, end_position_on_the_genomic_accession=56169754, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  112936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112936, CurrentID=0, Status='live', Symbol='VPS26B', Aliases='Pep8b', description='VPS26 retromer complex component B', other_designations='vesicle protein sorting 26B', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134224605, end_position_on_the_genomic_accession=134247792, orientation='plus', exon_count=7, OMIM=610027, no_hdr0=''),
  112937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112937, CurrentID=0, Status='live', Symbol='GLB1L3', Aliases='', description='galactosidase beta 1 like 3', other_designations='galactosidase beta 1-like 3|galactosidase, beta 1 like 3|galactosidase, beta 1-like 3', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134276322, end_position_on_the_genomic_accession=134321361, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  112939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112939, CurrentID=0, Status='live', Symbol='NACC1', Aliases='BEND8, BTBD14B, BTBD30, NAC-1, NAC1', description='nucleus accumbens associated 1, BEN and BTB (POZ) domain containing', other_designations='BEN domain containing 8|BTB/POZ domain-containing protein 14B|transcriptional repressor NAC1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13116847, end_position_on_the_genomic_accession=13141147, orientation='plus', exon_count=8, OMIM=610672, no_hdr0=''),
  112942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112942, CurrentID=0, Status='live', Symbol='CFAP36', Aliases='CCDC104', description='cilia and flagella associated protein 36', other_designations='coiled-coil domain containing 104|coiled-coil domain-containing protein 104', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55519595, end_position_on_the_genomic_accession=55545080, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105378867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378867, CurrentID=0, Status='live', Symbol='LOC105378867', Aliases='', description='uncharacterized LOC105378867', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97340752, end_position_on_the_genomic_accession=97389919, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  139061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139061, CurrentID=0, Status='live', Symbol='LOC139061', Aliases='', description='plasmolipin pseudogene', other_designations='plasma membrane proteolipid (plasmolipin) pseudogene', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45381526, end_position_on_the_genomic_accession=45381994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  112950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112950, CurrentID=0, Status='live', Symbol='MED8', Aliases='ARC32', description='mediator complex subunit 8', other_designations='activator-recruited cofactor 32 kDa component|mediator of RNA polymerase II transcription subunit MED8', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43383908, end_position_on_the_genomic_accession=43389812, orientation='minus', exon_count=7, OMIM=607956, no_hdr0=''),
  101927305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927305, CurrentID=0, Status='live', Symbol='LOC101927305', Aliases='', description='uncharacterized LOC101927305', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131764838, end_position_on_the_genomic_accession=131791482, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100271462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271462, CurrentID=0, Status='live', Symbol='RPL23AP40', Aliases='RPL23A_17_408', description='ribosomal protein L23a pseudogene 40', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138796851, end_position_on_the_genomic_accession=138797239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378869, CurrentID=0, Status='live', Symbol='LOC105378869', Aliases='', description='uncharacterized LOC105378869', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98143771, end_position_on_the_genomic_accession=98144838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927307, CurrentID=0, Status='live', Symbol='LOC101927307', Aliases='', description='uncharacterized LOC101927307', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227971380, end_position_on_the_genomic_accession=227972442, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421693, CurrentID=0, Status='live', Symbol='LOC100421693', Aliases='', description='NSA2 ribosome biogenesis homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96407284, end_position_on_the_genomic_accession=96407799, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  768328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=768328, CurrentID=0, Status='live', Symbol='EIF4EBP1P2', Aliases='', description='eukaryotic translation initiation factor 4E binding protein 1 pseudogene 2', other_designations='', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927308, CurrentID=0, Status='live', Symbol='LOC101927308', Aliases='', description='uncharacterized LOC101927308', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95590448, end_position_on_the_genomic_accession=95603044, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  112970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=112970, CurrentID=0, Status='live', Symbol='KTI12', Aliases='SBBI81, TOT4', description='KTI12 chromatin associated homolog', other_designations='KTI12 homolog, chromatin associated', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52032105, end_position_on_the_genomic_accession=52033800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100529207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529207, CurrentID=0, Status='live', Symbol='RAD51L3-RFFL', Aliases='', description='RAD51L3-RFFL readthrough', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35009112, end_position_on_the_genomic_accession=35121522, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927309, CurrentID=0, Status='live', Symbol='LOC101927309', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61249555, end_position_on_the_genomic_accession=61250890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378872, CurrentID=0, Status='live', Symbol='LOC105378872', Aliases='', description='uncharacterized LOC105378872', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101639560, end_position_on_the_genomic_accession=101785719, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101927310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927310, CurrentID=0, Status='live', Symbol='LOC101927310', Aliases='', description='uncharacterized LOC101927310', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97553284, end_position_on_the_genomic_accession=97560698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100529209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529209, CurrentID=0, Status='live', Symbol='RNASEK-C17orf49', Aliases='', description='RNASEK-C17orf49 readthrough', other_designations='RNASEK-C17orf49 readthrough (non-protein coding)', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7012417, end_position_on_the_genomic_accession=7017524, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  375133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375133, CurrentID=0, Status='live', Symbol='PI4KAP2', Aliases='', description='phosphatidylinositol 4-kinase, catalytic, alpha pseudogene 2', other_designations='Putative phosphatidylinositol 4-kinase alpha-like protein P2|phosphatidylinositol 4-kinase, catalytic, alpha polypeptide pseudogene 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21472998, end_position_on_the_genomic_accession=21517491, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  390203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390203, CurrentID=0, Status='live', Symbol='SRD5A3P1', Aliases='', description='steroid 5 alpha-reductase 3 pseudogene 1', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59898194, end_position_on_the_genomic_accession=59899134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  113000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113000, CurrentID=0, Status='live', Symbol='RPUSD1', Aliases='C16orf40, RLUCL', description='RNA pseudouridylate synthase domain containing 1', other_designations='ribosomal large subunit pseudouridine synthase C like|ribosomal large subunit pseudouridine synthase C-like protein', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=784974, end_position_on_the_genomic_accession=788387, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928405, CurrentID=0, Status='live', Symbol='LOC101928405', Aliases='', description='uncharacterized LOC101928405', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=165150006, end_position_on_the_genomic_accession=165158164, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100288914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288914, CurrentID=0, Status='live', Symbol='LOC100288914', Aliases='', description='mannosidase, beta A, lysosomal-like pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102734271, end_position_on_the_genomic_accession=102735483, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128061, CurrentID=0, Status='live', Symbol='C1orf131', Aliases='', description='chromosome 1 open reading frame 131', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231223763, end_position_on_the_genomic_accession=231241187, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378878, CurrentID=0, Status='live', Symbol='LOC105378878', Aliases='', description='uncharacterized LOC105378878', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103700592, end_position_on_the_genomic_accession=103701430, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422646, CurrentID=0, Status='live', Symbol='LOC100422646', Aliases='', description='1-aminocyclopropane-1-carboxylate synthase (inactive)-like pseudogene', other_designations='1-aminocyclopropane-1-carboxylate synthase homolog (Arabidopsis)(non-functional)-like pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7340744, end_position_on_the_genomic_accession=7342555, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378879, CurrentID=0, Status='live', Symbol='LOC105378879', Aliases='', description='uncharacterized LOC105378879', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=104132205, end_position_on_the_genomic_accession=104150706, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371725, CurrentID=0, Status='live', Symbol='LOC105371725', Aliases='', description='uncharacterized LOC105371725', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31425315, end_position_on_the_genomic_accession=31428318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  113026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113026, CurrentID=0, Status='live', Symbol='PLCD3', Aliases='PLC-delta-3', description='phospholipase C delta 3', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase delta-3|PLC delta3|phosphoinositide phospholipase C-delta-3|phospholipase C, delta 3', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45111640, end_position_on_the_genomic_accession=45132533, orientation='minus', exon_count=16, OMIM=608795, no_hdr0=''),
  102723526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723526, CurrentID=0, Status='live', Symbol='LOC102723526', Aliases='', description='uncharacterized LOC102723526', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17444057, end_position_on_the_genomic_accession=17445928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928406, CurrentID=0, Status='live', Symbol='LOC101928406', Aliases='', description='uncharacterized LOC101928406', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15994944, end_position_on_the_genomic_accession=16000491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  392206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392206, CurrentID=0, Status='live', Symbol='RPL10AP11', Aliases='RPL10A_4_880', description='ribosomal protein L10a pseudogene 11', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18383243, end_position_on_the_genomic_accession=18383948, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100419991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100419991, CurrentID=0, Status='live', Symbol='LOC100419991', Aliases='', description='phosphoserine phosphatase pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63057444, end_position_on_the_genomic_accession=63060084, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927320, CurrentID=0, Status='live', Symbol='ERICH3-AS1', Aliases='', description='ERICH3 antisense RNA 1', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74577430, end_position_on_the_genomic_accession=74626098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378883, CurrentID=0, Status='live', Symbol='LOC105378883', Aliases='', description='uncharacterized LOC105378883', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105830359, end_position_on_the_genomic_accession=105859474, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  375189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375189, CurrentID=0, Status='live', Symbol='PFN4', Aliases='', description='profilin family member 4', other_designations='profilin IV|profilin family, member 4', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24114809, end_position_on_the_genomic_accession=24123477, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  375190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375190, CurrentID=0, Status='live', Symbol='FAM228B', Aliases='', description='family with sequence similarity 228 member B', other_designations='UPF0638 protein B|family with sequence similarity 228, member B', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24076874, end_position_on_the_genomic_accession=24169638, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105378884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378884, CurrentID=0, Status='live', Symbol='LOC105378884', Aliases='', description='uncharacterized LOC105378884', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105927624, end_position_on_the_genomic_accession=106028371, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  375196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375196, CurrentID=0, Status='live', Symbol='LOC375196', Aliases='', description='uncharacterized LOC375196', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38959287, end_position_on_the_genomic_accession=38960344, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390213, CurrentID=0, Status='live', Symbol='DOC2GP', Aliases='', description='double C2-like domains, gamma, pseudogene', other_designations='double C2, gamma pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67613524, end_position_on_the_genomic_accession=67615664, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378886, CurrentID=0, Status='live', Symbol='LOC105378886', Aliases='', description='uncharacterized LOC105378886', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106041859, end_position_on_the_genomic_accession=106047677, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  346524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346524, CurrentID=0, Status='live', Symbol='OR2Q1P', Aliases='OR7-2', description='olfactory receptor family 2 subfamily Q member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily Q, member 1 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143980805, end_position_on_the_genomic_accession=143981932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390215, CurrentID=0, Status='live', Symbol='OR7E145P', Aliases='', description='olfactory receptor family 7 subfamily E member 145 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 145 pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67722459, end_position_on_the_genomic_accession=67723484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927325, CurrentID=0, Status='live', Symbol='LOC101927325', Aliases='', description='uncharacterized LOC101927325', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6505905, end_position_on_the_genomic_accession=6517035, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378888, CurrentID=0, Status='live', Symbol='LOC105378888', Aliases='', description='uncharacterized LOC105378888', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106831267, end_position_on_the_genomic_accession=106843817, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507379, CurrentID=0, Status='live', Symbol='LOC100507379', Aliases='', description='RNA binding motif protein 4 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33436268, end_position_on_the_genomic_accession=33436700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372014, CurrentID=0, Status='live', Symbol='LOC105372014', Aliases='', description='uncharacterized LOC105372014', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5629967, end_position_on_the_genomic_accession=5630924, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372681, CurrentID=0, Status='live', Symbol='LOC105372681', Aliases='', description='uncharacterized LOC105372681', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56497538, end_position_on_the_genomic_accession=56499022, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390218, CurrentID=0, Status='live', Symbol='IFITM9P', Aliases='', description='interferon induced transmembrane protein 9 pseudogene', other_designations='interferon induced transmembrane protein pseudogene', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69303351, end_position_on_the_genomic_accession=69303806, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927328, CurrentID=0, Status='live', Symbol='LOC101927328', Aliases='', description='uncharacterized LOC101927328', other_designations='CTD-2544H17.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122316955, end_position_on_the_genomic_accession=122323362, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  113091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113091, CurrentID=0, Status='live', Symbol='PTH2', Aliases='TIP39', description='parathyroid hormone 2', other_designations='tuberoinfundibular 39 residue protein', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49422414, end_position_on_the_genomic_accession=49426263, orientation='minus', exon_count=3, OMIM=608386, no_hdr0=''),
  101928408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928408, CurrentID=0, Status='live', Symbol='LOC101928408', Aliases='', description='uncharacterized LOC101928408', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97986551, end_position_on_the_genomic_accession=97988501, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378892, CurrentID=0, Status='live', Symbol='LOC105378892', Aliases='', description='uncharacterized LOC105378892', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109078320, end_position_on_the_genomic_accession=109090726, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390221, CurrentID=0, Status='live', Symbol='FOLR1P1', Aliases='FOLR1P, KB-1', description='folate receptor 1 (adult) pseudogene 1', other_designations='folate binding protein pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72158833, end_position_on_the_genomic_accession=72159456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  375248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375248, CurrentID=0, Status='live', Symbol='ANKRD36', Aliases='UNQ2430', description='ankyrin repeat domain 36', other_designations='ankyrin repeat domain-containing protein 36|ankyrin-related', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97113186, end_position_on_the_genomic_accession=97264521, orientation='plus', exon_count=81, OMIM='', no_hdr0=''),
  105378894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378894, CurrentID=0, Status='live', Symbol='LOC105378894', Aliases='', description='uncharacterized LOC105378894', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109720030, end_position_on_the_genomic_accession=109728731, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103456505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456505, CurrentID=0, Status='live', Symbol='TRG-CCC1-1', Aliases='TRNAG37', description='transfer RNA-Gly (CCC) 1-1', other_designations='tRNA-Gly-CCC-1-1|transfer RNA glycine 37 (anticodon CCC)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  113115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113115, CurrentID=0, Status='live', Symbol='MTFR2', Aliases='DUFD1, FAM54A', description='mitochondrial fission regulator 2', other_designations='DUF729 domain containing 1|DUF729 domain-containing protein 1|family with sequence similarity 54, member A|protein FAM54A', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136231029, end_position_on_the_genomic_accession=136253615, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  375260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375260, CurrentID=0, Status='live', Symbol='WASH2P', Aliases='FAM39B', description='WAS protein family homolog 2 pseudogene', other_designations='family with sequence similarity 39, member B', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113583653, end_position_on_the_genomic_accession=113599037, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105378896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378896, CurrentID=0, Status='live', Symbol='LOC105378896', Aliases='', description='uncharacterized LOC105378896', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110135495, end_position_on_the_genomic_accession=110138778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928409, CurrentID=0, Status='live', Symbol='LOC101928409', Aliases='', description='uncharacterized LOC101928409', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173507466, end_position_on_the_genomic_accession=173510957, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378897, CurrentID=0, Status='live', Symbol='LOC105378897', Aliases='', description='uncharacterized LOC105378897', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110165859, end_position_on_the_genomic_accession=110168717, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  113130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113130, CurrentID=0, Status='live', Symbol='CDCA5', Aliases='SORORIN', description='cell division cycle associated 5', other_designations='cell division cycle-associated protein 5|p35', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65061009, end_position_on_the_genomic_accession=65084150, orientation='minus', exon_count=16, OMIM=609374, no_hdr0=''),
  105378898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378898, CurrentID=0, Status='live', Symbol='LOC105378898', Aliases='', description='uncharacterized LOC105378898', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110472971, end_position_on_the_genomic_accession=110474980, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100310782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310782, CurrentID=0, Status='live', Symbol='LOC100310782', Aliases='', description='mitochondrial translational initiation factor 3 pseudogene', other_designations='', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=403697, end_position_on_the_genomic_accession=404549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  375287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375287, CurrentID=0, Status='live', Symbol='RBM43', Aliases='C2orf38', description='RNA binding motif protein 43', other_designations='RNA-binding motif protein 43', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=151248214, end_position_on_the_genomic_accession=151261901, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378900, CurrentID=0, Status='live', Symbol='LOC105378900', Aliases='', description='uncharacterized LOC105378900', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110522234, end_position_on_the_genomic_accession=110525306, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  113146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113146, CurrentID=0, Status='live', Symbol='AHNAK2', Aliases='C14orf78', description='AHNAK nucleoprotein 2', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104937253, end_position_on_the_genomic_accession=104978357, orientation='minus', exon_count=8, OMIM=608570, no_hdr0=''),
  103456511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103456511, CurrentID=0, Status='live', Symbol='TRL-CAA1-1', Aliases='TRNAL49', description='transfer RNA-Leu (CAA) 1-1', other_designations='transfer RNA leucine 49 (anticodon CAA)', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  375295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375295, CurrentID=0, Status='live', Symbol='LINC01116', Aliases='', description='long intergenic non-protein coding RNA 1116', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176629581, end_position_on_the_genomic_accession=176637574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  375298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375298, CurrentID=0, Status='live', Symbol='CERKL', Aliases='RP26', description='ceramide kinase like', other_designations='ceramide kinase-like', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=181536672, end_position_on_the_genomic_accession=181657107, orientation='minus', exon_count=14, OMIM=608381, no_hdr0=''),
  113157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113157, CurrentID=0, Status='live', Symbol='RPLP0P2', Aliases='RPLP0L2, RPLP0_3_1146', description='ribosomal protein, large, P0 pseudogene 2', other_designations='ribosomal protein P0 pseudogene 2', map_location='11cen-q12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61615036, end_position_on_the_genomic_accession=61639449, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  375307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375307, CurrentID=0, Status='live', Symbol='CATIP', Aliases='C2orf62', description='ciliogenesis associated TTC17 interacting protein', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218356832, end_position_on_the_genomic_accession=218368110, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100529240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529240, CurrentID=0, Status='live', Symbol='ZNF816-ZNF321P', Aliases='ZNF816-ZNF321', description='ZNF816-ZNF321P readthrough', other_designations='ZNF816-ZNF321 read-through transcript|ZNF816-ZNF321 readthrough', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52927135, end_position_on_the_genomic_accession=52962911, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  375316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375316, CurrentID=0, Status='live', Symbol='RBM44', Aliases='', description='RNA binding motif protein 44', other_designations='RNA-binding motif protein 44', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237798050, end_position_on_the_genomic_accession=237842808, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  113174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113174, CurrentID=0, Status='live', Symbol='SAAL1', Aliases='SPACIA1', description='serum amyloid A like 1', other_designations='serum amyloid A-like 1|synoviocyte proliferation-associated in collagen-induced arthritis 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18080343, end_position_on_the_genomic_accession=18106091, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  390233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390233, CurrentID=0, Status='live', Symbol='TRIM64FP', Aliases='', description='tripartite motif containing 64F, pseudogene', other_designations='', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95514827, end_position_on_the_genomic_accession=95520112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  113177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113177, CurrentID=0, Status='live', Symbol='IZUMO4', Aliases='C19orf36, IMAGE:4215339', description='IZUMO family member 4', other_designations='putative SAMP14 binding protein variant 1|putative SAMP14 binding protein variant 2|sperm 22 kDa protein c113', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2096598, end_position_on_the_genomic_accession=2099587, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  113178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113178, CurrentID=0, Status='live', Symbol='SCAMP4', Aliases='SCAMP-4', description='secretory carrier membrane protein 4', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1905374, end_position_on_the_genomic_accession=1926013, orientation='plus', exon_count=7, OMIM=613764, no_hdr0=''),
  113179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113179, CurrentID=0, Status='live', Symbol='ADAT3', Aliases='FWP005, MRT36, MST121, MSTP121, S863-5, TAD3', description='adenosine deaminase, tRNA-specific 3', other_designations='adenosine deaminase, tRNA-specific 3, TAD3 homolog|tRNA-specific adenosine deaminase 3 homolog|tRNA-specific adenosine deaminase-like protein 3|tRNA-specific adenosine-34 deaminase subunit ADAT3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1905372, end_position_on_the_genomic_accession=1913447, orientation='plus', exon_count=2, OMIM=615302, no_hdr0=''),
  105378906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378906, CurrentID=0, Status='live', Symbol='LOC105378906', Aliases='', description='uncharacterized LOC105378906', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112005933, end_position_on_the_genomic_accession=112008270, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465247, CurrentID=0, Status='live', Symbol='MIR548AY', Aliases='hsa-mir-548ay', description='microRNA 548ay', other_designations='microRNA mir-548ay', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32506283, end_position_on_the_genomic_accession=32506389, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100507397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507397, CurrentID=0, Status='live', Symbol='LOC100507397', Aliases='', description='uncharacterized LOC100507397', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=175546083, end_position_on_the_genomic_accession=175559262, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105378907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378907, CurrentID=0, Status='live', Symbol='LOC105378907', Aliases='', description='uncharacterized LOC105378907', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112077068, end_position_on_the_genomic_accession=112080078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  113189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113189, CurrentID=0, Status='live', Symbol='CHST14', Aliases='ATCS, D4ST1, EDSMC1, HNK1ST', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 14', other_designations='dermatan 4 sulfotransferase 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40470961, end_position_on_the_genomic_accession=40473158, orientation='plus', exon_count=1, OMIM=608429, no_hdr0=''),
  375337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375337, CurrentID=0, Status='live', Symbol='TOPAZ1', Aliases='C3orf77', description='testis and ovary specific PAZ domain containing 1', other_designations='testis and ovary-specific PAZ domain gene 1|topaz 1', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44241886, end_position_on_the_genomic_accession=44336610, orientation='plus', exon_count=21, OMIM=614412, no_hdr0=''),
  375341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375341, CurrentID=0, Status='live', Symbol='C3orf62', Aliases='', description='chromosome 3 open reading frame 62', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49268597, end_position_on_the_genomic_accession=49277075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  113201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113201, CurrentID=0, Status='live', Symbol='CASC4', Aliases='H63', description='cancer susceptibility candidate 4', other_designations='cancer susceptibility candidate gene 4 protein|gene associated with HER-2/neu overexpression', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44288616, end_position_on_the_genomic_accession=44415761, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  375346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375346, CurrentID=0, Status='live', Symbol='TMEM110', Aliases='STIMATE', description='transmembrane protein 110', other_designations='STIM-activating enhancer encoded by TMEM110', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52836756, end_position_on_the_genomic_accession=52897581, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927347, CurrentID=0, Status='live', Symbol='GRM7-AS3', Aliases='', description='GRM7 antisense RNA 3', other_designations='GRM7 antisense RNA 3 (non-protein coding)', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6632358, end_position_on_the_genomic_accession=6805449, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  375350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375350, CurrentID=0, Status='live', Symbol='COPS8P2', Aliases='', description='COP9 signalosome subunit 8 pseudogene 2', other_designations='COP9 constitutive photomorphogenic homolog subunit 8 pseudogene', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68144992, end_position_on_the_genomic_accession=68145773, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420020, CurrentID=0, Status='live', Symbol='LOC100420020', Aliases='', description='transmembrane protein 55B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65983682, end_position_on_the_genomic_accession=65984365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378911, CurrentID=0, Status='live', Symbol='LOC105378911', Aliases='', description='uncharacterized LOC105378911', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112857300, end_position_on_the_genomic_accession=112871382, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100420021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420021, CurrentID=0, Status='live', Symbol='LOC100420021', Aliases='', description='ankyrin repeat domain 33B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89863848, end_position_on_the_genomic_accession=89865597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113220, CurrentID=0, Status='live', Symbol='KIF12', Aliases='', description='kinesin family member 12', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114091623, end_position_on_the_genomic_accession=114100084, orientation='minus', exon_count=21, OMIM=611278, no_hdr0=''),
  100420022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420022, CurrentID=0, Status='live', Symbol='LOC100420022', Aliases='', description='ankyrin repeat domain 33B pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90051840, end_position_on_the_genomic_accession=90053735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480607, CurrentID=0, Status='live', Symbol='RNU6-761P', Aliases='', description='RNA, U6 small nuclear 761, pseudogene', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  113230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113230, CurrentID=0, Status='live', Symbol='LOC113230', Aliases='', description='uncharacterized protein LOC113230', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14073009, end_position_on_the_genomic_accession=14075062, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113235, CurrentID=0, Status='live', Symbol='SLC46A1', Aliases='G21, HCP1, PCFT', description='solute carrier family 46 (folate transporter), member 1', other_designations='heme carrier protein 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28394642, end_position_on_the_genomic_accession=28406212, orientation='minus', exon_count=5, OMIM=611672, no_hdr0=''),
  375387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375387, CurrentID=0, Status='live', Symbol='NRROS', Aliases='ELLP3030, GARPL1, LRRC33, UNQ3030', description='negative regulator of reactive oxygen species', other_designations='leucine-rich repeat-containing protein 33', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196639696, end_position_on_the_genomic_accession=196662004, orientation='plus', exon_count=3, OMIM=615322, no_hdr0=''),
  113246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113246, CurrentID=0, Status='live', Symbol='C12orf57', Aliases='C10, GRCC10', description='chromosome 12 open reading frame 57', other_designations='likely ortholog of mouse gene rich cluster, C10', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6943438, end_position_on_the_genomic_accession=6946003, orientation='plus', exon_count=4, OMIM=615140, no_hdr0=''),
  259173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259173, CurrentID=0, Status='live', Symbol='ALS2CL', Aliases='RN49018', description='ALS2 C-terminal like', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46668995, end_position_on_the_genomic_accession=46693704, orientation='minus', exon_count=29, OMIM=612402, no_hdr0=''),
  113251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113251, CurrentID=0, Status='live', Symbol='LARP4', Aliases='PP13296', description='La ribonucleoprotein domain family member 4', other_designations='La ribonucleoprotein domain family, member 4|c-Mpl binding protein', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50400809, end_position_on_the_genomic_accession=50480005, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  106480608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480608, CurrentID=0, Status='live', Symbol='RNU6-776P', Aliases='', description='RNA, U6 small nuclear 776, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  113263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113263, CurrentID=0, Status='live', Symbol='GLCCI1', Aliases='FAM117C, GCTR, GIG18, TSSN1', description='glucocorticoid induced 1', other_designations='glucocorticoid induced transcript 1', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7968743, end_position_on_the_genomic_accession=8089080, orientation='plus', exon_count=8, OMIM=614283, no_hdr0=''),
  149950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=149950, CurrentID=0, Status='live', Symbol='LOC149950', Aliases='', description='uncharacterized LOC149950', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32587479, end_position_on_the_genomic_accession=32608893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378921, CurrentID=0, Status='live', Symbol='LOC105378921', Aliases='', description='uncharacterized LOC105378921', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116407202, end_position_on_the_genomic_accession=116410261, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400943, CurrentID=0, Status='live', Symbol='LOC400943', Aliases='', description='uncharacterized LOC400943', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=10143627, end_position_on_the_genomic_accession=10143914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927359, CurrentID=0, Status='live', Symbol='LOC101927359', Aliases='', description='uncharacterized LOC101927359', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=133093860, end_position_on_the_genomic_accession=133149116, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  113277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113277, CurrentID=0, Status='live', Symbol='TMEM106A', Aliases='', description='transmembrane protein 106A', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43211827, end_position_on_the_genomic_accession=43220041, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  113278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113278, CurrentID=0, Status='live', Symbol='SLC52A3', Aliases='BVVLS, BVVLS1, C20orf54, RFT2, RFVT3, bA371L19.1, hRFT2', description='solute carrier family 52 (riboflavin transporter), member 3', other_designations='riboflavin transporter 2|solute carrier family 52 member 3', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=760080, end_position_on_the_genomic_accession=775985, orientation='minus', exon_count=6, OMIM=613350, no_hdr0=''),
  106480609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480609, CurrentID=0, Status='live', Symbol='RNU6-788P', Aliases='', description='RNA, U6 small nuclear 788, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  346562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346562, CurrentID=0, Status='live', Symbol='GNAT3', Aliases='GDCA', description='guanine nucleotide binding protein, alpha transducing 3', other_designations='gustatory G protein|gustducin alpha-3 chain|gustducin, alpha polypeptide', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80457292, end_position_on_the_genomic_accession=80527915, orientation='minus', exon_count=11, OMIM=139395, no_hdr0=''),
  105378925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378925, CurrentID=0, Status='live', Symbol='LOC105378925', Aliases='', description='uncharacterized LOC105378925', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116544615, end_position_on_the_genomic_accession=116550578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927363, CurrentID=0, Status='live', Symbol='LOC101927363', Aliases='', description='uncharacterized LOC101927363', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=32351038, end_position_on_the_genomic_accession=32353220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  375444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375444, CurrentID=0, Status='live', Symbol='C5orf34', Aliases='', description='chromosome 5 open reading frame 34', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43486701, end_position_on_the_genomic_accession=43515171, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105378926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378926, CurrentID=0, Status='live', Symbol='LOC105378926', Aliases='', description='uncharacterized LOC105378926', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116706439, end_position_on_the_genomic_accession=116710580, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102465251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465251, CurrentID=0, Status='live', Symbol='MIR6504', Aliases='hsa-mir-6504', description='microRNA 6504', other_designations='microRNA mir-6504', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81611348, end_position_on_the_genomic_accession=81611408, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  375449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375449, CurrentID=0, Status='live', Symbol='MAST4', Aliases='', description='microtubule associated serine/threonine kinase family member 4', other_designations='', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66596348, end_position_on_the_genomic_accession=67169598, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  390255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390255, CurrentID=0, Status='live', Symbol='LOC390255', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39B pseudogene', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119821222, end_position_on_the_genomic_accession=119822296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480610, CurrentID=0, Status='live', Symbol='RNU6-795P', Aliases='', description='RNA, U6 small nuclear 795, pseudogene', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288966, CurrentID=0, Status='live', Symbol='LOC100288966', Aliases='', description='POTE ankyrin domain family member D-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369191, CurrentID=0, Status='live', Symbol='LOC105369191', Aliases='', description='uncharacterized LOC105369191', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18486360, end_position_on_the_genomic_accession=18493078, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378931, CurrentID=0, Status='live', Symbol='LOC105378931', Aliases='', description='uncharacterized LOC105378931', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118136180, end_position_on_the_genomic_accession=118148481, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376387, CurrentID=0, Status='live', Symbol='LOC105376387', Aliases='', description='endogenous retrovirus group K member 7 Env polyprotein-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6837771, end_position_on_the_genomic_accession=7119802, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101927369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927369, CurrentID=0, Status='live', Symbol='LOC101927369', Aliases='', description='uncharacterized LOC101927369', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36116162, end_position_on_the_genomic_accession=36149062, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378932, CurrentID=0, Status='live', Symbol='LOC105378932', Aliases='', description='uncharacterized LOC105378932', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118555568, end_position_on_the_genomic_accession=118564265, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  375484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375484, CurrentID=0, Status='live', Symbol='SIMC1', Aliases='C5orf25, OOMA1, PLEIAD', description='SUMO-interacting motifs containing 1', other_designations='oocyte maturation associated 1|platform element for inhibition of autolytic degradation', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176238367, end_position_on_the_genomic_accession=176345989, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100420042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420042, CurrentID=0, Status='live', Symbol='CARSP1', Aliases='', description='cysteinyl-tRNA synthetase pseudogene 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68749501, end_position_on_the_genomic_accession=68750229, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480611, CurrentID=0, Status='live', Symbol='RNU6-815P', Aliases='', description='RNA, U6 small nuclear 815, pseudogene', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100288971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288971, CurrentID=0, Status='live', Symbol='LOC100288971', Aliases='', description='acireductone dioxygenase 1 pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102687953, end_position_on_the_genomic_accession=102688932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378935, CurrentID=0, Status='live', Symbol='LOC105378935', Aliases='', description='uncharacterized LOC105378935', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119483338, end_position_on_the_genomic_accession=119490108, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  728752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728752, CurrentID=0, Status='live', Symbol='LOC728752', Aliases='', description='uncharacterized LOC728752', other_designations='CTD-2630F21.1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36489627, end_position_on_the_genomic_accession=36491040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378936, CurrentID=0, Status='live', Symbol='LOC105378936', Aliases='', description='uncharacterized LOC105378936', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119727787, end_position_on_the_genomic_accession=119732090, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  400946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400946, CurrentID=0, Status='live', Symbol='LINC00954', Aliases='', description='long intergenic non-protein coding RNA 954', other_designations='', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19868854, end_position_on_the_genomic_accession=19885047, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100420046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420046, CurrentID=0, Status='live', Symbol='PRELID1P3', Aliases='', description='PRELI domain containing 1 pseudogene 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63427297, end_position_on_the_genomic_accession=63427943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378937, CurrentID=0, Status='live', Symbol='LOC105378937', Aliases='', description='uncharacterized LOC105378937', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119681348, end_position_on_the_genomic_accession=119711815, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  375513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375513, CurrentID=0, Status='live', Symbol='GUSBP4', Aliases='C6orf216, GUSBL2, SMA3-L, SMAC3L', description='glucuronidase, beta pseudogene 4', other_designations='glucuronidase, beta-like 2|spinal muscular atrophy candidate gene 3-like', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57919881, end_position_on_the_genomic_accession=57961446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480612, CurrentID=0, Status='live', Symbol='RNU6-824P', Aliases='', description='RNA, U6 small nuclear 824, pseudogene', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  375519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375519, CurrentID=0, Status='live', Symbol='GJB7', Aliases='CX25, bA136M9.1, connexin25', description='gap junction protein beta 7', other_designations='connexin 25|connexin-25|gap junction protein, beta 7, 25kDa', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87282978, end_position_on_the_genomic_accession=87329278, orientation='minus', exon_count=4, OMIM=611921, no_hdr0=''),
  101927376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927376, CurrentID=0, Status='live', Symbol='LOC101927376', Aliases='', description='uncharacterized LOC101927376', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10589366, end_position_on_the_genomic_accession=10606403, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422658, CurrentID=0, Status='live', Symbol='RAC1P9', Aliases='', description='ras-related C3 botulinum toxin substrate 1 pseudogene 9', other_designations='ras-related C3 botulinum toxin substrate 1 (rho family, small GTP binding protein Rac1) pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53779749, end_position_on_the_genomic_accession=53780263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390267, CurrentID=0, Status='live', Symbol='OR10N1P', Aliases='', description='olfactory receptor family 10 subfamily N member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily N, member 1 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124105052, end_position_on_the_genomic_accession=124105974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927377, CurrentID=0, Status='live', Symbol='LOC101927377', Aliases='', description='uncharacterized LOC101927377', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47352561, end_position_on_the_genomic_accession=47354633, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927378, CurrentID=0, Status='live', Symbol='LOC101927378', Aliases='', description='uncharacterized LOC101927378', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=84532476, end_position_on_the_genomic_accession=84584318, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390269, CurrentID=0, Status='live', Symbol='OR8F1P', Aliases='', description='olfactory receptor family 8 subfamily F member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily F, member 1 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124207184, end_position_on_the_genomic_accession=124208146, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100288979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100288979, CurrentID=0, Status='live', Symbol='DCAF13P2', Aliases='', description='DDB1 and CUL4 associated factor 13 pseudogene 2', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17202238, end_position_on_the_genomic_accession=17203145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373556, CurrentID=0, Status='live', Symbol='LOC105373556', Aliases='', description='uncharacterized LOC105373556', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111006201, end_position_on_the_genomic_accession=111050212, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927380, CurrentID=0, Status='live', Symbol='C9orf41-AS1', Aliases='', description='C9orf41 antisense RNA 1', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74952965, end_position_on_the_genomic_accession=74997382, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  113402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113402, CurrentID=0, Status='live', Symbol='SFT2D1', Aliases='C6orf83, pRGR1', description='SFT2 domain containing 1', other_designations='SFT2 domain-containing protein 1', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166319728, end_position_on_the_genomic_accession=166342606, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378943, CurrentID=0, Status='live', Symbol='LOC105378943', Aliases='', description='uncharacterized LOC105378943', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120718089, end_position_on_the_genomic_accession=120723566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927381, CurrentID=0, Status='live', Symbol='LINC01163', Aliases='TCONS_00017722', description='long intergenic non-protein coding RNA 1163', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128285950, end_position_on_the_genomic_accession=128317726, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378945, CurrentID=0, Status='live', Symbol='LOC105378945', Aliases='', description='uncharacterized LOC105378945', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120873852, end_position_on_the_genomic_accession=120911049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113419, CurrentID=0, Status='live', Symbol='TEX261', Aliases='TEG-261', description='testis expressed 261', other_designations='testis expressed sequence 261', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70985938, end_position_on_the_genomic_accession=70994873, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378946, CurrentID=0, Status='live', Symbol='LOC105378946', Aliases='', description='uncharacterized LOC105378946', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121492344, end_position_on_the_genomic_accession=121499740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376390, CurrentID=0, Status='live', Symbol='LOC105376390', Aliases='', description='uncharacterized LOC105376390', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7533207, end_position_on_the_genomic_accession=7536418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  375567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375567, CurrentID=0, Status='live', Symbol='VWC2', Aliases='PSST739, UNQ739', description='von Willebrand factor C domain containing 2', other_designations='brain-specific chordin-like protein|von Willebrand factor C domain-containing protein 2', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=49773661, end_position_on_the_genomic_accession=49921950, orientation='plus', exon_count=5, OMIM=611108, no_hdr0=''),
  105378947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378947, CurrentID=0, Status='live', Symbol='LOC105378947', Aliases='', description='proline-rich extensin-like protein EPR1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=586287, end_position_on_the_genomic_accession=611297, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927385, CurrentID=0, Status='live', Symbol='LOC101927385', Aliases='', description='uncharacterized LOC101927385', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97235411, end_position_on_the_genomic_accession=97298103, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  103752584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752584, CurrentID=0, Status='live', Symbol='CCND2-AS1', Aliases='', description='CCND2 antisense RNA 1', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4248767, end_position_on_the_genomic_accession=4276184, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  113444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113444, CurrentID=0, Status='live', Symbol='SMIM12', Aliases='C1orf212', description='small integral membrane protein 12', other_designations='UPF0767 protein C1orf212|alternative protein C1orf212', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34850362, end_position_on_the_genomic_accession=34859816, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  375593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375593, CurrentID=0, Status='live', Symbol='TRIM73', Aliases='TRIM50B', description='tripartite motif containing 73', other_designations='tripartite motif-containing protein 50B', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75395629, end_position_on_the_genomic_accession=75405613, orientation='plus', exon_count=5, OMIM=612549, no_hdr0=''),
  113451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113451, CurrentID=0, Status='live', Symbol='AZIN2', Aliases='ADC, AZI2, AZIB1, ODC-p, ODC1L, ODCp', description='antizyme inhibitor 2', other_designations='ODC antizyme inhibitor-2|ODC-like protein|ODC-paralogue|arginine decarboxylase|ornithine decarboxylase like|ornithine decarboxylase paralog|ornithine decarboxylase-like protein|ornithine decarboxylase-paralog', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33081110, end_position_on_the_genomic_accession=33178892, orientation='plus', exon_count=20, OMIM=608353, no_hdr0=''),
  113452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113452, CurrentID=0, Status='live', Symbol='TMEM54', Aliases='BCLP, CAC-1, CAC1', description='transmembrane protein 54', other_designations='beta-casein-like protein|protein CAC-1', map_location='1p35-p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32894595, end_position_on_the_genomic_accession=32901435, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  400949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400949, CurrentID=0, Status='live', Symbol='FKSG49', Aliases='FKSG52, FKSG55, FKSG56, FKSG57, FKSG60, FKSG61, FKSG64, FKSG65, FKSG66, FKSG67, FKSG68, FKSG70', description='FKSG49', other_designations='hepatocellular carcinoma-associated antigen HCA25b', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=75503347, end_position_on_the_genomic_accession=75504448, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927389, CurrentID=0, Status='live', Symbol='LOC101927389', Aliases='', description='uncharacterized LOC101927389', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55612978, end_position_on_the_genomic_accession=55650247, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  113457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113457, CurrentID=0, Status='live', Symbol='TUBA3D', Aliases='H2-ALPHA', description='tubulin alpha 3d', other_designations='alpha-tubulin isotype H2-alpha|tubulin alpha-2 chain|tubulin, alpha 3d', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131476007, end_position_on_the_genomic_accession=131482934, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100310835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100310835, CurrentID=0, Status='live', Symbol='LOC100310835', Aliases='', description='thyroid hormone receptor interactor 13 pseudogene', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53298167, end_position_on_the_genomic_accession=53301213, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927390, CurrentID=0, Status='live', Symbol='GTF2IP6', Aliases='', description='general transcription factor IIi pseudogene 6', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27024270, end_position_on_the_genomic_accession=27047984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  375607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375607, CurrentID=0, Status='live', Symbol='NAT16', Aliases='C7orf52', description='N-acetyltransferase 16 (GCN5-related, putative)', other_designations='putative N-acetyltransferase 16|putative N-acetyltransferase C7orf52', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101170493, end_position_on_the_genomic_accession=101195908, orientation='minus', exon_count=6, OMIM=615783, no_hdr0=''),
  375611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375611, CurrentID=0, Status='live', Symbol='SLC26A5', Aliases='DFNB61, PRES', description='solute carrier family 26 (anion exchanger), member 5', other_designations='prestin (motor protein)', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103352730, end_position_on_the_genomic_accession=103446177, orientation='minus', exon_count=22, OMIM=604943, no_hdr0=''),
  375612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375612, CurrentID=0, Status='live', Symbol='LHFPL3', Aliases='LHFPL4', description='lipoma HMGIC fusion partner-like 3', other_designations='LHFP-like protein 3|lipoma HMGIC fusion partner-like 4', map_location='7q22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104328656, end_position_on_the_genomic_accession=104908558, orientation='plus', exon_count=4, OMIM=609719, no_hdr0=''),
  390282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390282, CurrentID=0, Status='live', Symbol='LOC390282', Aliases='', description='eukaryotic translation initiation factor 3 subunit F pseudogene', other_designations='eukaryotic translation initiation factor 3, subunit F pseudogene', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5032391, end_position_on_the_genomic_accession=5032941, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  375616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375616, CurrentID=0, Status='live', Symbol='KCP', Aliases='CRIM21, NET67, KCP', description='kielin/chordin-like protein', other_designations='CRIM-2|KCP-1|cysteine rich BMP regulator 2 (chordin-like)|cysteine-rich BMP regulator 2|cysteine-rich motor neuron 2 protein|kielin|kielin/chordin-like protein 1', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128868934, end_position_on_the_genomic_accession=128910719, orientation='minus', exon_count=48, OMIM=609344, no_hdr0=''),
  101927392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927392, CurrentID=0, Status='live', Symbol='LOC101927392', Aliases='', description='uncharacterized LOC101927392', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105142860, end_position_on_the_genomic_accession=105188773, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378955, CurrentID=0, Status='live', Symbol='LOC105378955', Aliases='', description='tripartite motif-containing protein LOC642612', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90071054, end_position_on_the_genomic_accession=90073512, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927393, CurrentID=0, Status='live', Symbol='LOC101927393', Aliases='', description='uncharacterized LOC101927393', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42871414, end_position_on_the_genomic_accession=42873388, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378956, CurrentID=0, Status='live', Symbol='LOC105378956', Aliases='', description='uncharacterized LOC105378956', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120078113, end_position_on_the_genomic_accession=120102354, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928421, CurrentID=0, Status='live', Symbol='LOC101928421', Aliases='', description='uncharacterized LOC101928421', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35609754, end_position_on_the_genomic_accession=35629466, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927394, CurrentID=0, Status='live', Symbol='LOC101927394', Aliases='', description='uncharacterized LOC101927394', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=7952805, end_position_on_the_genomic_accession=8016307, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  222005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=222005, CurrentID=0, Status='live', Symbol='SLC25A5P3', Aliases='ANTP1', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 3', other_designations='adenine nucleotide translocator pseudogene 1 (clone 29)', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54419371, end_position_on_the_genomic_accession=54420600, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378959, CurrentID=0, Status='live', Symbol='LOC105378959', Aliases='', description='uncharacterized LOC105378959', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51937691, end_position_on_the_genomic_accession=51964435, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378960, CurrentID=0, Status='live', Symbol='LOC105378960', Aliases='', description='uncharacterized LOC105378960', other_designations='CTD-2203A3.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52004240, end_position_on_the_genomic_accession=52009849, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113510, CurrentID=0, Status='live', Symbol='HELQ', Aliases='HEL308', description='helicase, POLQ-like', other_designations='DNA helicase HEL308|POLQ-like helicase|mus308-like helicase', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83407343, end_position_on_the_genomic_accession=83455883, orientation='minus', exon_count=20, OMIM=606769, no_hdr0=''),
  105378961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378961, CurrentID=0, Status='live', Symbol='LOC105378961', Aliases='', description='uncharacterized LOC105378961', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52107478, end_position_on_the_genomic_accession=52137582, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928422, CurrentID=0, Status='live', Symbol='LOC101928422', Aliases='', description='protein FAM115C-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143620924, end_position_on_the_genomic_accession=143643937, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927399, CurrentID=0, Status='live', Symbol='PPIAP32', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 32', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130131716, end_position_on_the_genomic_accession=130132241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420072, CurrentID=0, Status='live', Symbol='EI24P4', Aliases='', description='etoposide induced 2.4 pseudogene 4', other_designations='etoposide induced 2.4 mRNA pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145005062, end_position_on_the_genomic_accession=145006145, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480617, CurrentID=0, Status='live', Symbol='RNU6-877P', Aliases='', description='RNA, U6 small nuclear 877, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927401, CurrentID=0, Status='live', Symbol='LOC101927401', Aliases='', description='cadherin-related family member 5-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228360910, end_position_on_the_genomic_accession=228368332, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927402, CurrentID=0, Status='live', Symbol='LOC101927402', Aliases='', description='uncharacterized proline-rich protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64522189, end_position_on_the_genomic_accession=64524611, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  113540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113540, CurrentID=0, Status='live', Symbol='CMTM1', Aliases='CKLFH, CKLFH1, CKLFSF1', description='CKLF-like MARVEL transmembrane domain containing 1', other_designations='chemokine-like factor super family 1|chemokine-like factor superfamily 1|chemokine-like factor superfamily member 1|chemokine-like factor-like protein CKLFH1', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66566391, end_position_on_the_genomic_accession=66579135, orientation='plus', exon_count=4, OMIM=607884, no_hdr0=''),
  105378966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378966, CurrentID=0, Status='live', Symbol='LOC105378966', Aliases='', description='uncharacterized LOC105378966', other_designations='CTD-2081C10.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53776836, end_position_on_the_genomic_accession=53819684, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  375686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375686, CurrentID=0, Status='live', Symbol='SPATC1', Aliases='SPATA15', description='spermatogenesis and centriole associated 1', other_designations='spermatogenesis and centriole-associated protein 1|spermatogenesis-associated protein 15|spermatogenic cell-specific Cdc20-binding protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144010993, end_position_on_the_genomic_accession=144047114, orientation='plus', exon_count=11, OMIM=610874, no_hdr0=''),
  101927404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927404, CurrentID=0, Status='live', Symbol='LOC101927404', Aliases='', description='uncharacterized LOC101927404', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64871872, end_position_on_the_genomic_accession=64993037, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  375690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375690, CurrentID=0, Status='live', Symbol='WASH5P', Aliases='', description='WAS protein family homolog 5 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=60951, end_position_on_the_genomic_accession=70966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373557, CurrentID=0, Status='live', Symbol='LOC105373557', Aliases='', description='uncharacterized LOC105373557', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111484328, end_position_on_the_genomic_accession=111510994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927405, CurrentID=0, Status='live', Symbol='LOC101927405', Aliases='', description='uncharacterized LOC101927405', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106491103, end_position_on_the_genomic_accession=106502188, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927406, CurrentID=0, Status='live', Symbol='LOC101927406', Aliases='', description='uncharacterized LOC101927406', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=194344269, end_position_on_the_genomic_accession=194419631, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  390297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390297, CurrentID=0, Status='live', Symbol='PSMC1P8', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 8', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=17383117, end_position_on_the_genomic_accession=17384658, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  375704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375704, CurrentID=0, Status='live', Symbol='ENHO', Aliases='C9orf165, UNQ470', description='energy homeostasis associated', other_designations='GAAI470|energy homeostasis-associated protein', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34521042, end_position_on_the_genomic_accession=34523039, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929973, CurrentID=0, Status='live', Symbol='LOC101929973', Aliases='', description='uncharacterized LOC101929973', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378970, CurrentID=0, Status='live', Symbol='LOC105378970', Aliases='', description='uncharacterized LOC105378970', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54957870, end_position_on_the_genomic_accession=54964041, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420080, CurrentID=0, Status='live', Symbol='HSPA8P7', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 7', other_designations='heat shock 70kDa protein 8 pseudogene 7', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=114774590, end_position_on_the_genomic_accession=114776853, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928424, CurrentID=0, Status='live', Symbol='LOC101928424', Aliases='', description='uncharacterized LOC101928424', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102452919, end_position_on_the_genomic_accession=102462038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  375719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375719, CurrentID=0, Status='live', Symbol='AQP7P1', Aliases='bA251O17.3', description='aquaporin 7 pseudogene 1', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63315243, end_position_on_the_genomic_accession=63334520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927410, CurrentID=0, Status='live', Symbol='LOC101927410', Aliases='', description='uncharacterized LOC101927410', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10661933, end_position_on_the_genomic_accession=10666887, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480619, CurrentID=0, Status='live', Symbol='RNU6-902P', Aliases='', description='RNA, U6 small nuclear 902, pseudogene', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927411, CurrentID=0, Status='live', Symbol='LOC101927411', Aliases='', description='uncharacterized LOC101927411', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32102088, end_position_on_the_genomic_accession=32105916, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378974, CurrentID=0, Status='live', Symbol='LOC105378974', Aliases='', description='uncharacterized LOC105378974', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55599922, end_position_on_the_genomic_accession=55605059, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103752585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103752585, CurrentID=0, Status='live', Symbol='RORB-AS1', Aliases='', description='RORB antisense RNA 1', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74473435, end_position_on_the_genomic_accession=74498553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  375743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375743, CurrentID=0, Status='live', Symbol='PTAR1', Aliases='', description='protein prenyltransferase alpha subunit repeat containing 1', other_designations='4930428J16Rik', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69709522, end_position_on_the_genomic_accession=69760044, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378976, CurrentID=0, Status='live', Symbol='LOC105378976', Aliases='', description='uncharacterized LOC105378976', other_designations='CTD-2227I18.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56059038, end_position_on_the_genomic_accession=56111111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928425, CurrentID=0, Status='live', Symbol='LOC101928425', Aliases='', description='protein shisa-5 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74060541, end_position_on_the_genomic_accession=74066364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  375748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375748, CurrentID=0, Status='live', Symbol='ERCC6L2', Aliases='BMFS2, C9orf102, RAD26L, SR278', description='excision repair cross-complementation group 6 like 2', other_designations='DNA repair and recombination protein RAD26-like|excision repair cross-complementation group 6-like 2|excision repair cross-complementing rodent repair deficiency, complementation group 6-like 2|putative repair and recombination helicase RAD26L|stretch responsive protein 278', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95875618, end_position_on_the_genomic_accession=96120476, orientation='plus', exon_count=28, OMIM=615667, no_hdr0=''),
  105373297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373297, CurrentID=0, Status='live', Symbol='LOC105373297', Aliases='', description='endogenous retrovirus group FC1 Env polyprotein', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=97841746, end_position_on_the_genomic_accession=97849585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113612, CurrentID=0, Status='live', Symbol='CYP2U1', Aliases='P450TEC, SPG49, SPG56', description='cytochrome P450 family 2 subfamily U member 1', other_designations='cytochrome P450, family 2, subfamily U, polypeptide 1|spastic paraplegia 49', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107931549, end_position_on_the_genomic_accession=107953461, orientation='plus', exon_count=5, OMIM=610670, no_hdr0=''),
  375757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375757, CurrentID=0, Status='live', Symbol='SWI5', Aliases='C9orf119, SAE3', description='SWI5 homologous recombination repair protein', other_designations='HBV DNAPTP1-transactivated protein A|SWI5 recombination repair homolog|protein SAE3 homolog', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128275384, end_position_on_the_genomic_accession=128288989, orientation='plus', exon_count=7, OMIM=616528, no_hdr0=''),
  375759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375759, CurrentID=0, Status='live', Symbol='C9orf50', Aliases='', description='chromosome 9 open reading frame 50', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129612225, end_position_on_the_genomic_accession=129622275, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100422666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422666, CurrentID=0, Status='live', Symbol='YWHABP2', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta pseudogene 2', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta polypeptide pseudogene', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18490239, end_position_on_the_genomic_accession=18490954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113622, CurrentID=0, Status='live', Symbol='ADPRHL1', Aliases='ARH2', description='ADP-ribosylhydrolase like 1', other_designations='ADP-ribosyl-hydrolase|ADP-ribosylhydrolase 2', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113421940, end_position_on_the_genomic_accession=113453524, orientation='minus', exon_count=8, OMIM=610620, no_hdr0=''),
  105369193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369193, CurrentID=0, Status='live', Symbol='LOC105369193', Aliases='', description='uncharacterized LOC105369193', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2320642, end_position_on_the_genomic_accession=2337998, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378980, CurrentID=0, Status='live', Symbol='LOC105378980', Aliases='', description='uncharacterized LOC105378980', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56896284, end_position_on_the_genomic_accession=56909861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  375775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375775, CurrentID=0, Status='live', Symbol='PNPLA7', Aliases='C9orf111, NTE-R1, NTEL1', description='patatin like phospholipase domain containing 7', other_designations='patatin-like phospholipase domain containing 7', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137459472, end_position_on_the_genomic_accession=137550895, orientation='minus', exon_count=39, OMIM=612122, no_hdr0=''),
  392226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392226, CurrentID=0, Status='live', Symbol='LOC392226', Aliases='', description='serine/threonine-protein kinase PLK1-like', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62143113, end_position_on_the_genomic_accession=62144593, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480621, CurrentID=0, Status='live', Symbol='RNU6-924P', Aliases='', description='RNA, U6 small nuclear 924, pseudogene', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  375790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375790, CurrentID=0, Status='live', Symbol='AGRN', Aliases='CMS8, CMSPPD', description='agrin', other_designations='agrin proteoglycan', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1020101, end_position_on_the_genomic_accession=1056119, orientation='plus', exon_count=40, OMIM=103320, no_hdr0=''),
  375791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375791, CurrentID=0, Status='live', Symbol='CYSRT1', Aliases='C9orf169', description='cysteine-rich tail protein 1', other_designations='UPF0574 protein C9orf169', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137224635, end_position_on_the_genomic_accession=137226311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378984, CurrentID=0, Status='live', Symbol='LOC105378984', Aliases='', description='uncharacterized LOC105378984', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58390281, end_position_on_the_genomic_accession=58400829, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  113655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113655, CurrentID=0, Status='live', Symbol='MFSD3', Aliases='', description='major facilitator superfamily domain containing 3', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144507267, end_position_on_the_genomic_accession=144511228, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390314, CurrentID=0, Status='live', Symbol='LOC390314', Aliases='', description='chromosome 20 open reading frame 27 pseudogene', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48800574, end_position_on_the_genomic_accession=48801088, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113675, CurrentID=0, Status='live', Symbol='SDSL', Aliases='SDH 2, SDS-RS1, TDH, cSDH', description='serine dehydratase-like', other_designations='L-serine deaminase|L-serine dehydratase/L-threonine deaminase|L-threonine dehydratase|serine dehydratase 2|serine dehydratase related sequence 1', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113422268, end_position_on_the_genomic_accession=113438277, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100289026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289026, CurrentID=0, Status='live', Symbol='LOC100289026', Aliases='', description='uncharacterized LOC100289026', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378990, CurrentID=0, Status='live', Symbol='LOC105378990', Aliases='', description='uncharacterized LOC105378990', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=59034281, end_position_on_the_genomic_accession=59065051, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100289028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289028, CurrentID=0, Status='live', Symbol='BNIP3P4', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 4', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64810866, end_position_on_the_genomic_accession=64811421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113691, CurrentID=0, Status='live', Symbol='TUBA3FP', Aliases='', description='tubulin alpha 3f pseudogene', other_designations='tubulin, alpha 3f, pseudogene|tubulin, alpha pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21008203, end_position_on_the_genomic_accession=21014287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378992, CurrentID=0, Status='live', Symbol='LOC105378992', Aliases='', description='uncharacterized LOC105378992', other_designations='CTC-436P18.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61201309, end_position_on_the_genomic_accession=61305062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  259249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259249, CurrentID=0, Status='live', Symbol='MRGPRX1', Aliases='GPCR, MGRG2, MRGX1, SNSR4', description='MAS related GPR family member X1', other_designations='G protein-coupled receptor MRGX1|G protein-coupled receptor SNSR3|MAS-related GPR, member X1|Mas-related G protein-coupled receptor G2|sensory neuron-specific G-protein coupled receptor 3/4', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18933813, end_position_on_the_genomic_accession=18935002, orientation='minus', exon_count=1, OMIM=607227, no_hdr0=''),
  375850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375850, CurrentID=0, Status='live', Symbol='RBMY2BP', Aliases='RBMY2B', description='RNA binding motif protein, Y-linked, family 2, member B pseudogene', other_designations='RNA binding motif protein, Y chromosome, family 2, member B', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22647156, end_position_on_the_genomic_accession=22655218, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390322, CurrentID=0, Status='live', Symbol='OR6C7P', Aliases='', description='olfactory receptor family 6 subfamily C member 7 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 7 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55342838, end_position_on_the_genomic_accession=55343776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378995, CurrentID=0, Status='live', Symbol='LOC105378995', Aliases='', description='uncharacterized LOC105378995', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61775230, end_position_on_the_genomic_accession=61784726, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378996, CurrentID=0, Status='live', Symbol='LOC105378996', Aliases='', description='uncharacterized LOC105378996', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61744227, end_position_on_the_genomic_accession=61759165, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  401703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401703, CurrentID=0, Status='live', Symbol='LOC401703', Aliases='', description='U2 small nuclear RNA auxiliary factor 1-like 4 pseudogene', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72280151, end_position_on_the_genomic_accession=72281107, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  113730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113730, CurrentID=0, Status='live', Symbol='KLHDC7B', Aliases='', description='kelch domain containing 7B', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50548033, end_position_on_the_genomic_accession=50551023, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106480624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480624, CurrentID=0, Status='live', Symbol='RNU6-955P', Aliases='', description='RNA, U6 small nuclear 955, pseudogene', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420108, CurrentID=0, Status='live', Symbol='LOC100420108', Aliases='', description='RNA binding motif protein 7 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=80162451, end_position_on_the_genomic_accession=80163212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378999, CurrentID=0, Status='live', Symbol='LOC105378999', Aliases='', description='uncharacterized LOC105378999', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=63883501, end_position_on_the_genomic_accession=63887417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927437, CurrentID=0, Status='live', Symbol='LOC101927437', Aliases='', description='uncharacterized LOC101927437', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99740870, end_position_on_the_genomic_accession=99957165, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  113746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113746, CurrentID=0, Status='live', Symbol='ODF3', Aliases='CT135, SHIPPO1, TISP50', description='outer dense fiber of sperm tails 3', other_designations='cancer/testis antigen 135|sperm tail protein SHIPPO1|transcript induced in spermiogenesis protein 50', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=196641, end_position_on_the_genomic_accession=200258, orientation='plus', exon_count=7, OMIM=608356, no_hdr0=''),
  105379001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379001, CurrentID=0, Status='live', Symbol='LOC105379001', Aliases='', description='uncharacterized LOC105379001', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64831774, end_position_on_the_genomic_accession=64884195, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  113763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113763, CurrentID=0, Status='live', Symbol='ZBED6CL', Aliases='C7orf29', description='ZBED6 C-terminal like', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150329849, end_position_on_the_genomic_accession=150332721, orientation='plus', exon_count=1, OMIM=615252, no_hdr0=''),
  195827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=195827, CurrentID=0, Status='live', Symbol='AAED1', Aliases='C9orf21', description='AhpC/TSA antioxidant enzyme domain containing 1', other_designations='UPF0308 protein C9orf21|ahpC/TSA antioxidant enzyme domain-containing protein 1', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96641251, end_position_on_the_genomic_accession=96655323, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  390332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390332, CurrentID=0, Status='live', Symbol='LOC390332', Aliases='', description='retinol dehydrogenase 16 (all-trans) pseudogene', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56981211, end_position_on_the_genomic_accession=56985960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420116, CurrentID=0, Status='live', Symbol='LOC100420116', Aliases='', description='transmembrane protein 132B pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127317709, end_position_on_the_genomic_accession=127319986, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113791, CurrentID=0, Status='live', Symbol='PIK3IP1', Aliases='HGFL, hHGFL(S)', description='phosphoinositide-3-kinase interacting protein 1', other_designations='kringle domain-containing protein HGFL', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31281593, end_position_on_the_genomic_accession=31292534, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480626, CurrentID=0, Status='live', Symbol='RNU6-977P', Aliases='', description='RNA, U6 small nuclear 977, pseudogene', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  375940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=375940, CurrentID=0, Status='live', Symbol='DMBT1P1', Aliases='', description='deleted in malignant brain tumors 1 pseudogene 1', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122756694, end_position_on_the_genomic_accession=122797645, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  101927446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927446, CurrentID=0, Status='live', Symbol='LOC101927446', Aliases='', description='uncharacterized LOC101927446', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90335223, end_position_on_the_genomic_accession=90344094, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  259265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259265, CurrentID=0, Status='live', Symbol='bA255A11.4', Aliases='', description='melanoma antigen pseudogene', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33572430, end_position_on_the_genomic_accession=33574635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  113802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113802, CurrentID=0, Status='live', Symbol='HENMT1', Aliases='C1orf59, HEN1', description='HEN1 methyltransferase homolog 1 (Arabidopsis)', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108648288, end_position_on_the_genomic_accession=108661526, orientation='minus', exon_count=10, OMIM=612178, no_hdr0=''),
  101927447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927447, CurrentID=0, Status='live', Symbol='LOC101927447', Aliases='', description='uncharacterized LOC101927447', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43275942, end_position_on_the_genomic_accession=43283830, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379010, CurrentID=0, Status='live', Symbol='LOC105379010', Aliases='', description='uncharacterized LOC105379010', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68648617, end_position_on_the_genomic_accession=68662370, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  392232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392232, CurrentID=0, Status='live', Symbol='LOC392232', Aliases='', description='transient receptor potential cation channel, subfamily A, member 1 pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72205299, end_position_on_the_genomic_accession=72251634, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  105379012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379012, CurrentID=0, Status='live', Symbol='LOC105379012', Aliases='', description='uncharacterized LOC105379012', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68483650, end_position_on_the_genomic_accession=68512736, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480627, CurrentID=0, Status='live', Symbol='RNU6-990P', Aliases='', description='RNA, U6 small nuclear 990, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420123, CurrentID=0, Status='live', Symbol='LOC100420123', Aliases='', description='MAS-related GPR, member X2 pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18836205, end_position_on_the_genomic_accession=18837084, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  113828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113828, CurrentID=0, Status='live', Symbol='FAM83F', Aliases='', description='family with sequence similarity 83 member F', other_designations='family with sequence similarity 83, member F', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39994949, end_position_on_the_genomic_accession=40030039, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  113829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113829, CurrentID=0, Status='live', Symbol='SLC35A4', Aliases='', description='solute carrier family 35 member A4', other_designations='solute carrier family 35 (UDP-galactose transporter), member A4|solute carrier family 35, member A4|tumor rejection antigen', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140564565, end_position_on_the_genomic_accession=140569104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113835, CurrentID=0, Status='live', Symbol='ZNF257', Aliases='BMZF-4, BMZF4', description='zinc finger protein 257', other_designations='bone marrow zinc finger 4', map_location='19q13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22048132, end_position_on_the_genomic_accession=22091109, orientation='plus', exon_count=7, OMIM=606957, no_hdr0=''),
  105379016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379016, CurrentID=0, Status='live', Symbol='LOC105379016', Aliases='', description='uncharacterized LOC105379016', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71259355, end_position_on_the_genomic_accession=71289856, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928433, CurrentID=0, Status='live', Symbol='LOC101928433', Aliases='', description='arabinogalactan protein 1-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=16760058, end_position_on_the_genomic_accession=16764994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390345, CurrentID=0, Status='live', Symbol='RPL10P13', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75688725, end_position_on_the_genomic_accession=75689087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480628, CurrentID=0, Status='live', Symbol='RNU6-1001P', Aliases='', description='RNA, U6 small nuclear 1001, pseudogene', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927457, CurrentID=0, Status='live', Symbol='LINC01522', Aliases='', description='long intergenic non-protein coding RNA 1522', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47980093, end_position_on_the_genomic_accession=47990117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420130, CurrentID=0, Status='live', Symbol='LOC100420130', Aliases='', description='RB1 inducible coiled-coil 1 pseudogene', other_designations='RB1-inducible coiled-coil 1 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98204677, end_position_on_the_genomic_accession=98205193, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376405, CurrentID=0, Status='live', Symbol='LOC105376405', Aliases='', description='uncharacterized LOC105376405', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10922309, end_position_on_the_genomic_accession=10929660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927459, CurrentID=0, Status='live', Symbol='LINC01608', Aliases='', description='long intergenic non-protein coding RNA 1608', other_designations='TCONS_00015136', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=110937690, end_position_on_the_genomic_accession=111027433, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  113877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113877, CurrentID=0, Status='live', Symbol='DFNB32', Aliases='', description='deafness, autosomal recessive 32', other_designations='', map_location='1p22.1-p13.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608653, no_hdr0=''),
  113878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=113878, CurrentID=0, Status='live', Symbol='DTX2', Aliases='RNF58', description='deltex 2, E3 ubiquitin ligase', other_designations='deltex homolog 2|deltex2|hDTX2|protein deltex-2|ring finger protein 58|zinc ion binding protein', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76461655, end_position_on_the_genomic_accession=76505998, orientation='plus', exon_count=15, OMIM=613141, no_hdr0=''),
  101927460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927460, CurrentID=0, Status='live', Symbol='LOC101927460', Aliases='', description='uncharacterized LOC101927460', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125493261, end_position_on_the_genomic_accession=125602227, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480629, CurrentID=0, Status='live', Symbol='RNU6-1007P', Aliases='', description='RNA, U6 small nuclear 1007, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420133, CurrentID=0, Status='live', Symbol='HSPA8P2', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 2', other_designations='heat shock 70kDa protein 8 pseudogene 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89546363, end_position_on_the_genomic_accession=89547196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379024, CurrentID=0, Status='live', Symbol='LOC105379024', Aliases='', description='uncharacterized LOC105379024', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69640366, end_position_on_the_genomic_accession=69647466, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376178, CurrentID=0, Status='live', Symbol='LOC105376178', Aliases='', description='uncharacterized LOC105376178', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100471135, end_position_on_the_genomic_accession=100472968, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379025, CurrentID=0, Status='live', Symbol='LOC105379025', Aliases='', description='uncharacterized LOC105379025', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71372548, end_position_on_the_genomic_accession=71375763, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390354, CurrentID=0, Status='live', Symbol='RPL18AP3', Aliases='RPL18A_6_1273, bcm182', description='ribosomal protein L18a pseudogene 3', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104265269, end_position_on_the_genomic_accession=104265892, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927464, CurrentID=0, Status='live', Symbol='LOC101927464', Aliases='', description='uncharacterized LOC101927464', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126095992, end_position_on_the_genomic_accession=126103932, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379027, CurrentID=0, Status='live', Symbol='LOC105379027', Aliases='', description='uncharacterized LOC105379027', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71750035, end_position_on_the_genomic_accession=71886003, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480630, CurrentID=0, Status='live', Symbol='RNU6-1016P', Aliases='', description='RNA, U6 small nuclear 1016, pseudogene', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372042, CurrentID=0, Status='live', Symbol='LOC105372042', Aliases='', description='uncharacterized LOC105372042', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27786527, end_position_on_the_genomic_accession=27805267, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379029, CurrentID=0, Status='live', Symbol='LOC105379029', Aliases='', description='uncharacterized LOC105379029', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73114872, end_position_on_the_genomic_accession=73120456, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129186, CurrentID=0, Status='live', Symbol='LOC100129186', Aliases='', description='uncharacterized LOC100129186', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=42950264, end_position_on_the_genomic_accession=42955114, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927467, CurrentID=0, Status='live', Symbol='LOC101927467', Aliases='', description='uncharacterized LOC101927467', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11062294, end_position_on_the_genomic_accession=11099148, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379030, CurrentID=0, Status='live', Symbol='LOC105379030', Aliases='', description='uncharacterized LOC105379030', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73132007, end_position_on_the_genomic_accession=73150691, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927468, CurrentID=0, Status='live', Symbol='LOC101927468', Aliases='', description='uncharacterized LOC101927468', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148208093, end_position_on_the_genomic_accession=148247071, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379031, CurrentID=0, Status='live', Symbol='LOC105379031', Aliases='', description='uncharacterized LOC105379031', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73246494, end_position_on_the_genomic_accession=73274998, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927469, CurrentID=0, Status='live', Symbol='LOC101927469', Aliases='', description='metacaspase-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14895762, end_position_on_the_genomic_accession=14913685, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106480631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480631, CurrentID=0, Status='live', Symbol='RNU6-1026P', Aliases='', description='RNA, U6 small nuclear 1026, pseudogene', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  346671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346671, CurrentID=0, Status='live', Symbol='LOC346671', Aliases='', description='ring finger protein 14 pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128218, CurrentID=0, Status='live', Symbol='TMEM125', Aliases='', description='transmembrane protein 125', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43269951, end_position_on_the_genomic_accession=43274002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370941, CurrentID=0, Status='live', Symbol='LOC105370941', Aliases='', description='uncharacterized LOC105370941', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39782415, end_position_on_the_genomic_accession=39785617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927472, CurrentID=0, Status='live', Symbol='LOC101927472', Aliases='', description='uncharacterized LOC101927472', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104323364, end_position_on_the_genomic_accession=104326745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379670, CurrentID=0, Status='live', Symbol='LOC105379670', Aliases='', description='uncharacterized LOC105379670', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390363, CurrentID=0, Status='live', Symbol='CLIC1P1', Aliases='', description='chloride intracellular channel 1 pseudogene 1', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120914392, end_position_on_the_genomic_accession=120915323, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  346673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=346673, CurrentID=0, Status='live', Symbol='STRA8', Aliases='', description='stimulated by retinoic acid 8', other_designations='stimulated by retinoic acid 8 homolog|stimulated by retinoic acid gene 8 homolog', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135231979, end_position_on_the_genomic_accession=135258661, orientation='plus', exon_count=11, OMIM=609987, no_hdr0=''),
  105379036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379036, CurrentID=0, Status='live', Symbol='LOC105379036', Aliases='', description='uncharacterized LOC105379036', other_designations='CTD-2292M14.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74034266, end_position_on_the_genomic_accession=74103216, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927474, CurrentID=0, Status='live', Symbol='LOC101927474', Aliases='', description='uncharacterized LOC101927474', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43570088, end_position_on_the_genomic_accession=43571760, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379037, CurrentID=0, Status='live', Symbol='LOC105379037', Aliases='', description='uncharacterized LOC105379037', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74258785, end_position_on_the_genomic_accession=74321276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480632, CurrentID=0, Status='live', Symbol='RNU6-1034P', Aliases='', description='RNA, U6 small nuclear 1034, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927476, CurrentID=0, Status='live', Symbol='LOC101927476', Aliases='', description='uncharacterized LOC101927476', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40262917, end_position_on_the_genomic_accession=40287721, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379039, CurrentID=0, Status='live', Symbol='LOC105379039', Aliases='', description='uncharacterized LOC105379039', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74883141, end_position_on_the_genomic_accession=74898113, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379040, CurrentID=0, Status='live', Symbol='LOC105379040', Aliases='', description='uncharacterized LOC105379040', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74948166, end_position_on_the_genomic_accession=74975758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  376132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376132, CurrentID=0, Status='live', Symbol='LRRC10', Aliases='HRLRRPA, LRRC10', description='leucine rich repeat containing 10', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69608564, end_position_on_the_genomic_accession=69611162, orientation='minus', exon_count=1, OMIM=610846, no_hdr0=''),
  101928438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928438, CurrentID=0, Status='live', Symbol='LOC101928438', Aliases='', description='uncharacterized LOC101928438', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=99585786, end_position_on_the_genomic_accession=99819889, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927479, CurrentID=0, Status='live', Symbol='LOC101927479', Aliases='', description='uncharacterized LOC101927479', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16155001, end_position_on_the_genomic_accession=16168216, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927480, CurrentID=0, Status='live', Symbol='LOC101927480', Aliases='', description='uncharacterized LOC101927480', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54847247, end_position_on_the_genomic_accession=54848291, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480633, CurrentID=0, Status='live', Symbol='RNU6-1044P', Aliases='', description='RNA, U6 small nuclear 1044, pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929976, CurrentID=0, Status='live', Symbol='LOC101929976', Aliases='', description='uncharacterized LOC101929976', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182671945, end_position_on_the_genomic_accession=182714608, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379045, CurrentID=0, Status='live', Symbol='LOC105379045', Aliases='', description='uncharacterized LOC105379045', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77772649, end_position_on_the_genomic_accession=77776194, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927483, CurrentID=0, Status='live', Symbol='LOC101927483', Aliases='', description='uncharacterized LOC101927483', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48194123, end_position_on_the_genomic_accession=48228207, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101928439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928439, CurrentID=0, Status='live', Symbol='LOC101928439', Aliases='', description='uncharacterized LOC101928439', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173516675, end_position_on_the_genomic_accession=173521135, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114026, CurrentID=0, Status='live', Symbol='ZIM3', Aliases='ZNF657', description='zinc finger, imprinted 3', other_designations='zinc finger protein 657', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57134096, end_position_on_the_genomic_accession=57145202, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480634, CurrentID=0, Status='live', Symbol='RNU6-1062P', Aliases='', description='RNA, U6 small nuclear 1062, pseudogene', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114034, CurrentID=0, Status='live', Symbol='TOE1', Aliases='', description='target of EGR1, member 1 (nuclear)', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45339670, end_position_on_the_genomic_accession=45343978, orientation='plus', exon_count=8, OMIM=613931, no_hdr0=''),
  114036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114036, CurrentID=0, Status='live', Symbol='LINC00310', Aliases='C21orf82, NCRNA00310', description='long intergenic non-protein coding RNA 310', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34180677, end_position_on_the_genomic_accession=34189919, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927486, CurrentID=0, Status='live', Symbol='LOC101927486', Aliases='', description='uncharacterized LOC101927486', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49568124, end_position_on_the_genomic_accession=49571934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  114038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114038, CurrentID=0, Status='live', Symbol='LINC00313', Aliases='C21orf84, CH507-42P11.5, NCRNA00313', description='long intergenic non-protein coding RNA 313', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43462094, end_position_on_the_genomic_accession=43478223, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114041, CurrentID=0, Status='live', Symbol='B3GALT5-AS1', Aliases='C21orf88', description='B3GALT5 antisense RNA 1', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39597147, end_position_on_the_genomic_accession=39612822, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114042, CurrentID=0, Status='live', Symbol='LINC00334', Aliases='C21orf89, NCRNA00334', description='long intergenic non-protein coding RNA 334', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=46678462, end_position_on_the_genomic_accession=46678563, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  114043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114043, CurrentID=0, Status='live', Symbol='TSPEAR-AS2', Aliases='C21orf90', description='TSPEAR antisense RNA 2', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44517215, end_position_on_the_genomic_accession=44518977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  114044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114044, CurrentID=0, Status='live', Symbol='MCM3AP-AS1', Aliases='C21orf85, MCM3AP-AS, MCM3APAS, MCM3APASB, NCRNA00031', description='MCM3AP antisense RNA 1', other_designations='MCM3 minichromosome maintenance deficient 3 associated protein antisense|MCM3AP antisense RNA 1 (non-protein coding)|minichromosome maintenance complex component 3 associated protein antisense', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46229231, end_position_on_the_genomic_accession=46251701, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379050, CurrentID=0, Status='live', Symbol='LOC105379050', Aliases='', description='uncharacterized LOC105379050', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82730531, end_position_on_the_genomic_accession=82735046, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114049, CurrentID=0, Status='live', Symbol='WBSCR22', Aliases='HASJ4442, HUSSY-3, MERM1, PP3381, WBMT', description='Williams Beuren syndrome chromosome region 22', other_designations='Williams-Beuren candidate region putative methyltransferase|Williams-Beuren syndrome chromosomal region 22 protein|bud site selection protein 23 homolog|metastasis-related methyltransferase 1|ribosome biogenesis methyltransferase WBSCR22', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73683566, end_position_on_the_genomic_accession=73698221, orientation='plus', exon_count=13, OMIM=615733, no_hdr0=''),
  105379051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379051, CurrentID=0, Status='live', Symbol='LOC105379051', Aliases='', description='uncharacterized LOC105379051', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82913367, end_position_on_the_genomic_accession=82924327, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  104054148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104054148, CurrentID=0, Status='live', Symbol='LOC104054148', Aliases='', description='uncharacterized LOC104054148', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101984432, end_position_on_the_genomic_accession=101994850, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  84545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84545, CurrentID=0, Status='live', Symbol='MRPL43', Aliases='L43mt, MRP-L43, bMRP36a', description='mitochondrial ribosomal protein L43', other_designations='mitochondrial ribosomal protein bMRP36a', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100977822, end_position_on_the_genomic_accession=100987515, orientation='minus', exon_count=8, OMIM=611848, no_hdr0=''),
  105379052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379052, CurrentID=0, Status='live', Symbol='LOC105379052', Aliases='', description='uncharacterized LOC105379052', other_designations='CTD-2227C6.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83011441, end_position_on_the_genomic_accession=83023190, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927490, CurrentID=0, Status='live', Symbol='LOC101927490', Aliases='', description='uncharacterized LOC101927490', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139618139, end_position_on_the_genomic_accession=139647339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480635, CurrentID=0, Status='live', Symbol='RNU6-1074P', Aliases='', description='RNA, U6 small nuclear 1074, pseudogene', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289091, CurrentID=0, Status='live', Symbol='LOC100289091', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit G2 pseudogene', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit G pseudogene', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58755629, end_position_on_the_genomic_accession=58756247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100289092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289092, CurrentID=0, Status='live', Symbol='ATP2A1-AS1', Aliases='', description='ATP2A1 antisense RNA 1', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28878957, end_position_on_the_genomic_accession=28879921, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379055, CurrentID=0, Status='live', Symbol='LOC105379055', Aliases='', description='uncharacterized LOC105379055', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83710383, end_position_on_the_genomic_accession=83714595, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927493, CurrentID=0, Status='live', Symbol='LOC101927493', Aliases='', description='uncharacterized LOC101927493', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1824237, end_position_on_the_genomic_accession=1834256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379056, CurrentID=0, Status='live', Symbol='LOC105379056', Aliases='', description='uncharacterized LOC105379056', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84492164, end_position_on_the_genomic_accession=84496472, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100289094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289094, CurrentID=0, Status='live', Symbol='LOC100289094', Aliases='', description='uncharacterized LOC100289094', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=241627222, end_position_on_the_genomic_accession=241629111, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  114086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114086, CurrentID=0, Status='live', Symbol='DBA2', Aliases='', description='Diamond-Blackfan anemia 2', other_designations='', map_location='8p23.3-p22', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606129, no_hdr0=''),
  114088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114088, CurrentID=0, Status='live', Symbol='TRIM9', Aliases='RNF91, SPRING', description='tripartite motif containing 9', other_designations='RING finger protein 91|homolog of rat RING finger Spring|tripartite motif-containing 9|tripartite motif-containing protein 9', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50975261, end_position_on_the_genomic_accession=51095704, orientation='minus', exon_count=13, OMIM=606555, no_hdr0=''),
  106480636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480636, CurrentID=0, Status='live', Symbol='RNU6-1081P', Aliases='', description='RNA, U6 small nuclear 1081, pseudogene', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927496, CurrentID=0, Status='live', Symbol='LOC101927496', Aliases='', description='uncharacterized LOC101927496', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21240700, end_position_on_the_genomic_accession=21242698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927497, CurrentID=0, Status='live', Symbol='LOC101927497', Aliases='', description='uncharacterized LOC101927497', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92836483, end_position_on_the_genomic_accession=92917187, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100420170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420170, CurrentID=0, Status='live', Symbol='LOC100420170', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108508574, end_position_on_the_genomic_accession=108509242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379061, CurrentID=0, Status='live', Symbol='LOC105379061', Aliases='', description='uncharacterized LOC105379061', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85629706, end_position_on_the_genomic_accession=85647805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114112, CurrentID=0, Status='live', Symbol='TXNRD3', Aliases='TGR, TR2, TRXR3', description='thioredoxin reductase 3', other_designations='thioredoxin and glutathione reductase|thioredoxin glutathione reductase|thioredoxin reductase 2|thioredoxin reductase TR2', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126607052, end_position_on_the_genomic_accession=126655124, orientation='minus', exon_count=16, OMIM=606235, no_hdr0=''),
  101927499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927499, CurrentID=0, Status='live', Symbol='LOC101927499', Aliases='', description='uncharacterized LOC101927499', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44312942, end_position_on_the_genomic_accession=44314999, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104054213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104054213, CurrentID=0, Status='live', Symbol='LINC01562', Aliases='', description='long intergenic non-protein coding RNA 1562', other_designations='RP11-296A18.3', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  376267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376267, CurrentID=0, Status='live', Symbol='RAB15', Aliases='', description='RAB15, member RAS oncogene family', other_designations='RAB15, member RAS onocogene family|Ras-related protein Rab-15 isoform AN1|Ras-related protein Rab-15 isoform AN2|Ras-related protein Rab-15 isoform AN3', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64945814, end_position_on_the_genomic_accession=64972290, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101927501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927501, CurrentID=0, Status='live', Symbol='LOC101927501', Aliases='', description='uncharacterized LOC101927501', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43176994, end_position_on_the_genomic_accession=43226598, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372049, CurrentID=0, Status='live', Symbol='LOC105372049', Aliases='', description='uncharacterized LOC105372049', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31228114, end_position_on_the_genomic_accession=31259438, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114131, CurrentID=0, Status='live', Symbol='UCN3', Aliases='SCP, SPC, UCNIII', description='urocortin 3', other_designations='prepro-urocortin 3|stresscopin|ucn III|urocortin III', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5365013, end_position_on_the_genomic_accession=5374206, orientation='plus', exon_count=2, OMIM=605901, no_hdr0=''),
  114132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114132, CurrentID=0, Status='live', Symbol='SIGLEC11', Aliases='', description='sialic acid binding Ig-like lectin 11', other_designations='siglec-11', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49948993, end_position_on_the_genomic_accession=49961172, orientation='minus', exon_count=11, OMIM=607157, no_hdr0=''),
  100289102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289102, CurrentID=0, Status='live', Symbol='EIF4A1P8', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 8', other_designations='', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92136861, end_position_on_the_genomic_accession=92138515, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114134, CurrentID=0, Status='live', Symbol='SLC2A13', Aliases='HMIT', description='solute carrier family 2 (facilitated glucose transporter), member 13', other_designations='H(+)-myo-inositol symporter|h(+)-myo-inositol cotransporter|proton (H+) myo-inositol symporter|solute carrier family 2 member 13', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39755021, end_position_on_the_genomic_accession=40106074, orientation='minus', exon_count=13, OMIM=611036, no_hdr0=''),
  101927503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927503, CurrentID=0, Status='live', Symbol='LOC101927503', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1039200, end_position_on_the_genomic_accession=1055749, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105379066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379066, CurrentID=0, Status='live', Symbol='LOC105379066', Aliases='', description='uncharacterized LOC105379066', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87412982, end_position_on_the_genomic_accession=87483172, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376414, CurrentID=0, Status='live', Symbol='LOC105376414', Aliases='', description='uncharacterized LOC105376414', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11765938, end_position_on_the_genomic_accession=11783605, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927504, CurrentID=0, Status='live', Symbol='LOC101927504', Aliases='', description='uncharacterized LOC101927504', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229870019, end_position_on_the_genomic_accession=229894066, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105373561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373561, CurrentID=0, Status='live', Symbol='LOC105373561', Aliases='', description='uncharacterized LOC105373561', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112286642, end_position_on_the_genomic_accession=112288965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927505, CurrentID=0, Status='live', Symbol='LOC101927505', Aliases='', description='uncharacterized LOC101927505', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101605513, end_position_on_the_genomic_accession=101614339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480638, CurrentID=0, Status='live', Symbol='RNU6-1110P', Aliases='', description='RNA, U6 small nuclear 1110, pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929977, CurrentID=0, Status='live', Symbol='LOC101929977', Aliases='', description='uncharacterized LOC101929977', other_designations='CTD-2256P15.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10342765, end_position_on_the_genomic_accession=10354085, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374434, CurrentID=0, Status='live', Symbol='LOC105374434', Aliases='', description='uncharacterized LOC105374434', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43296862, end_position_on_the_genomic_accession=43330696, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100507562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100507562, CurrentID=0, Status='live', Symbol='LOC100507562', Aliases='', description='uncharacterized LOC100507562', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21396455, end_position_on_the_genomic_accession=21602884, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927509, CurrentID=0, Status='live', Symbol='LOC101927509', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241267602, end_position_on_the_genomic_accession=241277002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114181, CurrentID=0, Status='live', Symbol='FAM8A5P', Aliases='', description='family with sequence similarity 8 member A5, pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114182, CurrentID=0, Status='live', Symbol='FAM8A6P', Aliases='', description='family with sequence similarity 8 member A6, pseudogene', other_designations='family with sequence similarity 8, member A1 pseudogene', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134603142, end_position_on_the_genomic_accession=134604239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480639, CurrentID=0, Status='live', Symbol='RNU6-1115P', Aliases='', description='RNA, U6 small nuclear 1115, pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114184, CurrentID=0, Status='live', Symbol='SIGLEC18P', Aliases='SIGLECP1', description='sialic acid binding Ig-like lectin 18, pseudogene', other_designations='SIGLEC18P, sialic acid binding Ig-like lectin 18, pseudogene|sialic acid binding Ig-like lectin 12 pseudogene|sialic acid binding Ig-like lectin, pseudogene 1', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51119830, end_position_on_the_genomic_accession=51120336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114185, CurrentID=0, Status='live', Symbol='SIGLEC26P', Aliases='SIGLECP10', description='sialic acid binding Ig-like lectin 26, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 10', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51468649, end_position_on_the_genomic_accession=51469530, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114186, CurrentID=0, Status='live', Symbol='SIGLEC27P', Aliases='SIGLECP11', description='sialic acid binding Ig-like lectin 27, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 11', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51508403, end_position_on_the_genomic_accession=51508828, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114187, CurrentID=0, Status='live', Symbol='SIGLEC28P', Aliases='SIGLECP12', description='sialic acid binding Ig-like lectin 28, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 12', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51545326, end_position_on_the_genomic_accession=51546087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114188, CurrentID=0, Status='live', Symbol='SIGLEC29P', Aliases='SIGLECP13', description='sialic acid binding Ig-like lectin 29, pseudogene', other_designations='CD33 molecule pseudogene|sialic acid binding Ig-like lectin, pseudogene 13', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51564438, end_position_on_the_genomic_accession=51564916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114189, CurrentID=0, Status='live', Symbol='SIGLEC31P', Aliases='SIGLECP15', description='sialic acid binding Ig-like lectin 31, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 15', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56589954, end_position_on_the_genomic_accession=56590491, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114190, CurrentID=0, Status='live', Symbol='SIGLEC30P', Aliases='SIGLECP14', description='sialic acid binding Ig-like lectin 30, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 14', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170115523, end_position_on_the_genomic_accession=170116397, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114192, CurrentID=0, Status='live', Symbol='SIGLEC20P', Aliases='SIGLECP1, SIGLECP4', description='sialic acid binding Ig-like lectin 20, pseudogene', other_designations='sialic acid binding Ig-like lectin 8 pseudogene|sialic acid binding Ig-like lectin, pseudogene 4', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51184814, end_position_on_the_genomic_accession=51185239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114194, CurrentID=0, Status='live', Symbol='SIGLEC21P', Aliases='SIGLECP5', description='sialic acid binding Ig-like lectin 21, pseudogene', other_designations='sialic acid binding Ig-like lectin 12 pseudogene|sialic acid binding Ig-like lectin, pseudogene 5', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51198856, end_position_on_the_genomic_accession=51199321, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114195, CurrentID=0, Status='live', Symbol='SIGLEC22P', Aliases='SIGLECP6', description='sialic acid binding Ig-like lectin 22, pseudogene', other_designations='CD33 molecule pseudogene|sialic acid binding Ig-like lectin, pseudogene 6', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51211110, end_position_on_the_genomic_accession=51211589, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114196, CurrentID=0, Status='live', Symbol='SIGLEC24P', Aliases='SIGLECP8', description='sialic acid binding Ig-like lectin 24, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 8', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51280835, end_position_on_the_genomic_accession=51281188, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114198, CurrentID=0, Status='live', Symbol='SIGLEC25P', Aliases='SIGLECP9', description='sialic acid binding Ig-like lectin 25, pseudogene', other_designations='sialic acid binding Ig-like lectin, pseudogene 9', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51461329, end_position_on_the_genomic_accession=51461803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927513, CurrentID=0, Status='live', Symbol='LOC101927513', Aliases='', description='uncharacterized LOC101927513', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119062942, end_position_on_the_genomic_accession=119068782, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928445, CurrentID=0, Status='live', Symbol='LOC101928445', Aliases='', description='uncharacterized LOC101928445', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179595904, end_position_on_the_genomic_accession=179603741, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927514, CurrentID=0, Status='live', Symbol='LOC101927514', Aliases='', description='uncharacterized LOC101927514', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126385573, end_position_on_the_genomic_accession=126395097, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  652549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652549, CurrentID=0, Status='live', Symbol='LOC652549', Aliases='', description='SNW domain containing 1 pseudogene', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78749072, end_position_on_the_genomic_accession=78750856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480640, CurrentID=0, Status='live', Symbol='RNU6-1127P', Aliases='', description='RNA, U6 small nuclear 1127, pseudogene', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927516, CurrentID=0, Status='live', Symbol='LOC101927516', Aliases='', description='uncharacterized LOC101927516', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140127329, end_position_on_the_genomic_accession=140138891, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379079, CurrentID=0, Status='live', Symbol='LOC105379079', Aliases='', description='uncharacterized LOC105379079', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91794228, end_position_on_the_genomic_accession=91797824, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101059906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059906, CurrentID=0, Status='live', Symbol='LOC101059906', Aliases='', description='collagen alpha-2(IV) chain-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=528484, end_position_on_the_genomic_accession=532263, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369197, CurrentID=0, Status='live', Symbol='LOC105369197', Aliases='', description='uncharacterized LOC105369197', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51803358, end_position_on_the_genomic_accession=51847484, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  114224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114224, CurrentID=0, Status='live', Symbol='PRO2852', Aliases='', description='uncharacterized protein PRO2852', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34186108, end_position_on_the_genomic_accession=34188127, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100289118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289118, CurrentID=0, Status='live', Symbol='LOC100289118', Aliases='', description='mitochondrial carrier 2 pseudogene', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=137535575, end_position_on_the_genomic_accession=137536144, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645388, CurrentID=0, Status='live', Symbol='LOC645388', Aliases='', description='adaptor-related protein complex 1, mu 2 subunit pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65469086, end_position_on_the_genomic_accession=65470381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376417, CurrentID=0, Status='live', Symbol='LOC105376417', Aliases='', description='uncharacterized LOC105376417', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12399734, end_position_on_the_genomic_accession=12403717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927519, CurrentID=0, Status='live', Symbol='LOC101927519', Aliases='', description='tubulin beta-7 chain-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29613288, end_position_on_the_genomic_accession=29614329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289120, CurrentID=0, Status='live', Symbol='LOC100289120', Aliases='', description='uncharacterized LOC100289120', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=135055800, end_position_on_the_genomic_accession=135057190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379083, CurrentID=0, Status='live', Symbol='LOC105379083', Aliases='', description='uncharacterized LOC105379083', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93135840, end_position_on_the_genomic_accession=93162489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480641, CurrentID=0, Status='live', Symbol='RNU6-1138P', Aliases='', description='RNA, U6 small nuclear 1138, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422687, CurrentID=0, Status='live', Symbol='LOC100422687', Aliases='', description='pleckstrin homology domain containing A1 pseudogene', other_designations='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 1 pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14620040, end_position_on_the_genomic_accession=14621243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379085, CurrentID=0, Status='live', Symbol='LOC105379085', Aliases='', description='uncharacterized LOC105379085', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95138564, end_position_on_the_genomic_accession=95259834, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101928447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928447, CurrentID=0, Status='live', Symbol='LOC101928447', Aliases='', description='uncharacterized LOC101928447', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76105974, end_position_on_the_genomic_accession=76132499, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379087, CurrentID=0, Status='live', Symbol='LOC105379087', Aliases='', description='uncharacterized LOC105379087', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94111695, end_position_on_the_genomic_accession=94251844, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  376412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376412, CurrentID=0, Status='live', Symbol='RNF126P1', Aliases='', description='ring finger protein 126 pseudogene 1', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57045478, end_position_on_the_genomic_accession=57046795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  376413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376413, CurrentID=0, Status='live', Symbol='LOC376413', Aliases='', description='PAD-like', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379088, CurrentID=0, Status='live', Symbol='LOC105379088', Aliases='', description='uncharacterized LOC105379088', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93621369, end_position_on_the_genomic_accession=93675492, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480642, CurrentID=0, Status='live', Symbol='RNU6-1148P', Aliases='', description='RNA, U6 small nuclear 1148, pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379090, CurrentID=0, Status='live', Symbol='LOC105379090', Aliases='', description='uncharacterized LOC105379090', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95667345, end_position_on_the_genomic_accession=95684773, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379091, CurrentID=0, Status='live', Symbol='LOC105379091', Aliases='', description='uncharacterized LOC105379091', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95861894, end_position_on_the_genomic_accession=95876809, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114293, CurrentID=0, Status='live', Symbol='ANBC', Aliases='', description='Aneurysmal bone cysts', other_designations='', map_location='16q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606179, no_hdr0=''),
  114294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114294, CurrentID=0, Status='live', Symbol='LACTB', Aliases='G24, MRPL56', description='lactamase beta', other_designations='lactamase, beta|mitochondrial 39S ribosomal protein L56', map_location='15q22.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63121800, end_position_on_the_genomic_accession=63142065, orientation='plus', exon_count=7, OMIM=608440, no_hdr0=''),
  105373562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373562, CurrentID=0, Status='live', Symbol='LOC105373562', Aliases='', description='uncharacterized LOC105373562', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112541656, end_position_on_the_genomic_accession=112542293, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114299, CurrentID=0, Status='live', Symbol='PALM2', Aliases='AKAP2', description='paralemmin 2', other_designations='A kinase (PRKA) anchor protein 2', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=109640788, end_position_on_the_genomic_accession=109951476, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101929105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929105, CurrentID=0, Status='live', Symbol='LOC101929105', Aliases='', description='uncharacterized LOC101929105', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16880918, end_position_on_the_genomic_accession=16887776, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379093, CurrentID=0, Status='live', Symbol='LOC105379093', Aliases='', description='uncharacterized LOC105379093', other_designations='CTD-2260A17.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96803952, end_position_on_the_genomic_accession=96806105, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480643, CurrentID=0, Status='live', Symbol='RNU6-1153P', Aliases='', description='RNA, U6 small nuclear 1153, pseudogene', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289131, CurrentID=0, Status='live', Symbol='COX6CP6', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 6', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70751067, end_position_on_the_genomic_accession=70751472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927532, CurrentID=0, Status='live', Symbol='LOC101927532', Aliases='', description='uncharacterized LOC101927532', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230002717, end_position_on_the_genomic_accession=230007162, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379095, CurrentID=0, Status='live', Symbol='LOC105379095', Aliases='', description='uncharacterized LOC105379095', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96142656, end_position_on_the_genomic_accession=96165500, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927533, CurrentID=0, Status='live', Symbol='LOC101927533', Aliases='', description='uncharacterized LOC101927533', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65897584, end_position_on_the_genomic_accession=65899512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379096, CurrentID=0, Status='live', Symbol='LOC105379096', Aliases='', description='uncharacterized LOC105379096', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95985889, end_position_on_the_genomic_accession=96072467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376420, CurrentID=0, Status='live', Symbol='LOC105376420', Aliases='', description='uncharacterized LOC105376420', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13413776, end_position_on_the_genomic_accession=13423758, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928449, CurrentID=0, Status='live', Symbol='LOC101928449', Aliases='', description='uncharacterized LOC101928449', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81953719, end_position_on_the_genomic_accession=81993133, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  114327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114327, CurrentID=0, Status='live', Symbol='EFHC1', Aliases='EJM1, dJ304B14.2', description='EF-hand domain (C-terminal) containing 1', other_designations='myoclonin-1', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52420196, end_position_on_the_genomic_accession=52495785, orientation='plus', exon_count=12, OMIM=608815, no_hdr0=''),
  106480644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480644, CurrentID=0, Status='live', Symbol='RNU6-1164P', Aliases='', description='RNA, U6 small nuclear 1164, pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114335, CurrentID=0, Status='live', Symbol='CGB1', Aliases='', description='chorionic gonadotropin, beta polypeptide 1', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49035569, end_position_on_the_genomic_accession=49036934, orientation='minus', exon_count=3, OMIM=608823, no_hdr0=''),
  114336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114336, CurrentID=0, Status='live', Symbol='CGB2', Aliases='', description='chorionic gonadotropin, beta polypeptide 2', other_designations='product of CGB2', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49031873, end_position_on_the_genomic_accession=49033238, orientation='plus', exon_count=3, OMIM=608824, no_hdr0=''),
  105372056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372056, CurrentID=0, Status='live', Symbol='LOC105372056', Aliases='', description='uncharacterized LOC105372056', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32680612, end_position_on_the_genomic_accession=32738451, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422690, CurrentID=0, Status='live', Symbol='SLC25A6P4', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene 4', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80077707, end_position_on_the_genomic_accession=80078249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100034248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034248, CurrentID=0, Status='live', Symbol='LOC100034248', Aliases='', description='uncharacterized LOC100034248', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=7011546, end_position_on_the_genomic_accession=7012062, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379101, CurrentID=0, Status='live', Symbol='LOC105379101', Aliases='', description='uncharacterized LOC105379101', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100246891, end_position_on_the_genomic_accession=100248331, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  376497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376497, CurrentID=0, Status='live', Symbol='SLC27A1', Aliases='ACSVL5, FATP, FATP-1, FATP1', description='solute carrier family 27 (fatty acid transporter), member 1', other_designations='fatty acid transport protein 1|solute carrier family 27 member 1', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17468774, end_position_on_the_genomic_accession=17506168, orientation='plus', exon_count=15, OMIM=600691, no_hdr0=''),
  101927539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927539, CurrentID=0, Status='live', Symbol='LOC101927539', Aliases='', description='uncharacterized LOC101927539', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57600864, end_position_on_the_genomic_accession=57608412, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100270858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270858, CurrentID=0, Status='live', Symbol='RPS2P15', Aliases='RPS2_3_202', description='ribosomal protein S2 pseudogene 15', other_designations='', map_location='2p23', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26101317, end_position_on_the_genomic_accession=26102136, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379103, CurrentID=0, Status='live', Symbol='LOC105379103', Aliases='', description='uncharacterized LOC105379103', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102669568, end_position_on_the_genomic_accession=102671059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480645, CurrentID=0, Status='live', Symbol='RNU6-1177P', Aliases='', description='RNA, U6 small nuclear 1177, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289141, CurrentID=0, Status='live', Symbol='LOC100289141', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17620534, end_position_on_the_genomic_accession=17620939, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379104, CurrentID=0, Status='live', Symbol='LOC105379104', Aliases='', description='uncharacterized LOC105379104', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102664252, end_position_on_the_genomic_accession=102726614, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369198, CurrentID=0, Status='live', Symbol='LOC105369198', Aliases='', description='uncharacterized LOC105369198', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30357290, end_position_on_the_genomic_accession=30358296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379105, CurrentID=0, Status='live', Symbol='LOC105379105', Aliases='', description='uncharacterized LOC105379105', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103045692, end_position_on_the_genomic_accession=103071029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101929243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929243, CurrentID=0, Status='live', Symbol='LOC101929243', Aliases='', description='uncharacterized LOC101929243', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34696899, end_position_on_the_genomic_accession=34697471, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420215, CurrentID=0, Status='live', Symbol='LOC100420215', Aliases='', description='RNA binding motif protein 45 pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=131712316, end_position_on_the_genomic_accession=131712980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379106, CurrentID=0, Status='live', Symbol='LOC105379106', Aliases='', description='uncharacterized LOC105379106', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=102974443, end_position_on_the_genomic_accession=102983479, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376422, CurrentID=0, Status='live', Symbol='LOC105376422', Aliases='', description='uncharacterized LOC105376422', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13548236, end_position_on_the_genomic_accession=13557170, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928451, CurrentID=0, Status='live', Symbol='LOC101928451', Aliases='', description='uncharacterized LOC101928451', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128531355, end_position_on_the_genomic_accession=128640002, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105379107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379107, CurrentID=0, Status='live', Symbol='LOC105379107', Aliases='', description='uncharacterized LOC105379107', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103670789, end_position_on_the_genomic_accession=103819228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379108, CurrentID=0, Status='live', Symbol='LOC105379108', Aliases='', description='uncharacterized LOC105379108', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103913404, end_position_on_the_genomic_accession=103915424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480646, CurrentID=0, Status='live', Symbol='RNU6-1189P', Aliases='', description='RNA, U6 small nuclear 1189, pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379109, CurrentID=0, Status='live', Symbol='LOC105379109', Aliases='', description='uncharacterized LOC105379109', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=104540194, end_position_on_the_genomic_accession=104773817, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105379110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379110, CurrentID=0, Status='live', Symbol='LOC105379110', Aliases='', description='uncharacterized LOC105379110', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=105243271, end_position_on_the_genomic_accession=105393540, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379111, CurrentID=0, Status='live', Symbol='LOC105379111', Aliases='', description='uncharacterized LOC105379111', other_designations='CTD-2374C24.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=104972974, end_position_on_the_genomic_accession=104979775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376423, CurrentID=0, Status='live', Symbol='LOC105376423', Aliases='', description='uncharacterized LOC105376423', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14182890, end_position_on_the_genomic_accession=14185567, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379112, CurrentID=0, Status='live', Symbol='LOC105379112', Aliases='', description='uncharacterized LOC105379112', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=106247370, end_position_on_the_genomic_accession=106250451, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100289150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289150, CurrentID=0, Status='live', Symbol='FAM197Y5P', Aliases='FAM197Y5', description='family with sequence similarity 197 Y-linked member 5, pseudogene', other_designations='family with sequence similarity 197, Y-linked, member 5, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9458743, end_position_on_the_genomic_accession=9464345, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379113, CurrentID=0, Status='live', Symbol='LOC105379113', Aliases='', description='uncharacterized LOC105379113', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=107126696, end_position_on_the_genomic_accession=107153491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480647, CurrentID=0, Status='live', Symbol='RNU6-1195P', Aliases='', description='RNA, U6 small nuclear 1195, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289151, CurrentID=0, Status='live', Symbol='DCLRE1CP1', Aliases='', description='DNA cross-link repair 1C pseudogene 1', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=15057357, end_position_on_the_genomic_accession=15077424, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379114, CurrentID=0, Status='live', Symbol='LOC105379114', Aliases='', description='uncharacterized LOC105379114', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108467402, end_position_on_the_genomic_accession=108486103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372690, CurrentID=0, Status='live', Symbol='LOC105372690', Aliases='', description='uncharacterized LOC105372690', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57580579, end_position_on_the_genomic_accession=57582585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379115, CurrentID=0, Status='live', Symbol='LOC105379115', Aliases='', description='uncharacterized LOC105379115', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109163663, end_position_on_the_genomic_accession=109186235, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376424, CurrentID=0, Status='live', Symbol='LOC105376424', Aliases='', description='uncharacterized LOC105376424', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14082947, end_position_on_the_genomic_accession=14089294, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283140, CurrentID=0, Status='live', Symbol='LOC283140', Aliases='', description='uncharacterized LOC283140', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112270749, end_position_on_the_genomic_accession=112362534, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379117, CurrentID=0, Status='live', Symbol='LOC105379117', Aliases='', description='uncharacterized LOC105379117', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109469055, end_position_on_the_genomic_accession=109573827, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106480648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480648, CurrentID=0, Status='live', Symbol='RNU6-1207P', Aliases='', description='RNA, U6 small nuclear 1207, pseudogene', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927556, CurrentID=0, Status='live', Symbol='LOC101927556', Aliases='', description='uncharacterized LOC101927556', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58129480, end_position_on_the_genomic_accession=58163295, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101059914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059914, CurrentID=0, Status='live', Symbol='LOC101059914', Aliases='', description='protein FAM218A-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164929491, end_position_on_the_genomic_accession=164930626, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927558, CurrentID=0, Status='live', Symbol='LOC101927558', Aliases='', description='uncharacterized LOC101927558', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15702337, end_position_on_the_genomic_accession=15727956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376425, CurrentID=0, Status='live', Symbol='LOC105376425', Aliases='', description='uncharacterized LOC105376425', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13990376, end_position_on_the_genomic_accession=14008040, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114475, CurrentID=0, Status='live', Symbol='AD7', Aliases='', description='Alzheimer disease 7', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606187, no_hdr0=''),
  114477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114477, CurrentID=0, Status='live', Symbol='ATOD6', Aliases='', description='Dermatitis, atopic, 6', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605845, no_hdr0=''),
  101927560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927560, CurrentID=0, Status='live', Symbol='LOC101927560', Aliases='', description='uncharacterized LOC101927560', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83801516, end_position_on_the_genomic_accession=83860546, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480649, CurrentID=0, Status='live', Symbol='RNU6-1219P', Aliases='', description='RNA, U6 small nuclear 1219, pseudogene', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927561, CurrentID=0, Status='live', Symbol='LOC101927561', Aliases='', description='uncharacterized LOC101927561', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120714117, end_position_on_the_genomic_accession=120716114, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379124, CurrentID=0, Status='live', Symbol='LOC105379124', Aliases='', description='uncharacterized LOC105379124', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=113294860, end_position_on_the_genomic_accession=113336545, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927562, CurrentID=0, Status='live', Symbol='LOC101927562', Aliases='', description='putative uncharacterized protein C11orf81', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1554607, end_position_on_the_genomic_accession=1556457, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645397, CurrentID=0, Status='live', Symbol='LOC645397', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1-like pseudogene', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18870987, end_position_on_the_genomic_accession=18873462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379128, CurrentID=0, Status='live', Symbol='LOC105379128', Aliases='', description='uncharacterized LOC105379128', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115172694, end_position_on_the_genomic_accession=115175456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480650, CurrentID=0, Status='live', Symbol='RNU6-1225P', Aliases='', description='RNA, U6 small nuclear 1225, pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927566, CurrentID=0, Status='live', Symbol='LOC101927566', Aliases='', description='uncharacterized LOC101927566', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=258275, end_position_on_the_genomic_accession=274363, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100422696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422696, CurrentID=0, Status='live', Symbol='LOC100422696', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 5 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102798990, end_position_on_the_genomic_accession=102799661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927568, CurrentID=0, Status='live', Symbol='LOC101927568', Aliases='', description='uncharacterized LOC101927568', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134595264, end_position_on_the_genomic_accession=134673081, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376427, CurrentID=0, Status='live', Symbol='LOC105376427', Aliases='', description='uncharacterized LOC105376427', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13694808, end_position_on_the_genomic_accession=13699088, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480651, CurrentID=0, Status='live', Symbol='RNU6-1237P', Aliases='', description='RNA, U6 small nuclear 1237, pseudogene', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927571, CurrentID=0, Status='live', Symbol='LOC101927571', Aliases='', description='uncharacterized LOC101927571', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22723491, end_position_on_the_genomic_accession=22786837, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114548, CurrentID=0, Status='live', Symbol='NLRP3', Aliases='AGTAVPRL, AII, AVP, C1orf7, CIAS1, CLR1.1, FCAS, FCAS1, FCU, MWS, NALP3, PYPAF1', description='NLR family, pyrin domain containing 3', other_designations='NACHT domain-, leucine-rich repeat-, and PYD-containing protein 3|NACHT, LRR and PYD containing protein 3|PYRIN-containing APAF1-like protein 1|caterpiller protein 1.1|cold autoinflammatory syndrome 1 protein|cold-induced autoinflammatory syndrome 1 protein|cryopyrin|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 3', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247415904, end_position_on_the_genomic_accession=247449108, orientation='plus', exon_count=13, OMIM=606416, no_hdr0=''),
  376693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376693, CurrentID=0, Status='live', Symbol='RPS10P7', Aliases='RPS10_2_147', description='ribosomal protein S10 pseudogene 7', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201519904, end_position_on_the_genomic_accession=201520602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927572, CurrentID=0, Status='live', Symbol='LOC101927572', Aliases='', description='uncharacterized LOC101927572', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36014499, end_position_on_the_genomic_accession=36045997, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927573, CurrentID=0, Status='live', Symbol='LOC101927573', Aliases='', description='uncharacterized LOC101927573', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=107689569, end_position_on_the_genomic_accession=107713406, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  652608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652608, CurrentID=0, Status='live', Symbol='LOC652608', Aliases='', description='60S ribosomal protein L6-like', other_designations='', map_location='X;Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1172583, end_position_on_the_genomic_accession=1175157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927574, CurrentID=0, Status='live', Symbol='LINC01186', Aliases='', description='long intergenic non-protein coding RNA 1186', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46325924, end_position_on_the_genomic_accession=46327674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114568, CurrentID=0, Status='live', Symbol='NMTC1', Aliases='', description='Nonmedullary thyroid carcinoma 1', other_designations='', map_location='2q21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606240, no_hdr0=''),
  114569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114569, CurrentID=0, Status='live', Symbol='MAL2', Aliases='', description='mal, T-cell differentiation protein 2 (gene/pseudogene)', other_designations='MAL proteolipid protein 2|MAL2 proteolipid protein', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119208371, end_position_on_the_genomic_accession=119245673, orientation='plus', exon_count=4, OMIM=609684, no_hdr0=''),
  114571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114571, CurrentID=0, Status='live', Symbol='SLC22A9', Aliases='HOAT4, OAT4, OAT7, UST3H, ust3', description='solute carrier family 22 (organic anion transporter), member 9', other_designations='organic anion transporter 4|organic anion transporter 7|solute carrier family 22 (organic anion/cation transporter), member 9', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63369789, end_position_on_the_genomic_accession=63410259, orientation='plus', exon_count=10, OMIM=607579, no_hdr0=''),
  105379138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379138, CurrentID=0, Status='live', Symbol='LOC105379138', Aliases='', description='uncharacterized LOC105379138', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117271934, end_position_on_the_genomic_accession=117280286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480652, CurrentID=0, Status='live', Symbol='RNU6-1249P', Aliases='', description='RNA, U6 small nuclear 1249, pseudogene', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927577, CurrentID=0, Status='live', Symbol='LOC101927577', Aliases='', description='uncharacterized LOC101927577', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66697030, end_position_on_the_genomic_accession=66703227, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420250, CurrentID=0, Status='live', Symbol='LOC100420250', Aliases='', description='MAGE family member A4 pseudogene', other_designations='melanoma antigen family A, 4 pseudogene|melanoma antigen family A4 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152682946, end_position_on_the_genomic_accession=152683843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106481272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481272, CurrentID=0, Status='live', Symbol='RNU6-295P', Aliases='', description='RNA, U6 small nuclear 295, pseudogene', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420251, CurrentID=0, Status='live', Symbol='LOC100420251', Aliases='', description='mitochondrial fission factor pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15191746, end_position_on_the_genomic_accession=15193127, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114599, CurrentID=0, Status='live', Symbol='SNORD15B', Aliases='RNU15B, U15B', description='small nucleolar RNA, C/D box 15B', other_designations='RNA, U15b small nucleolar|U15B small nucleolar RNA|U15B snRNA|U15B snoRNA', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75404421, end_position_on_the_genomic_accession=75404566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927580, CurrentID=0, Status='live', Symbol='LOC101927580', Aliases='', description='uncharacterized LOC101927580', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59855353, end_position_on_the_genomic_accession=60053973, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480653, CurrentID=0, Status='live', Symbol='RNU6-1263P', Aliases='', description='RNA, U6 small nuclear 1263, pseudogene', other_designations='', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114609, CurrentID=0, Status='live', Symbol='TIRAP', Aliases='BACTS1, Mal, MyD88-2, wyatt', description='toll-interleukin 1 receptor (TIR) domain containing adaptor protein', other_designations='MyD88 adapter-like protein|Toll-like receptor adaptor protein|adapter protein wyatt|adaptor protein Wyatt', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126281893, end_position_on_the_genomic_accession=126294933, orientation='plus', exon_count=7, OMIM=606252, no_hdr0=''),
  114610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114610, CurrentID=0, Status='live', Symbol='SAX1', Aliases='SPAX1', description='spastic ataxia 1 (autosomal dominant)', other_designations='', map_location='12p13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=108600, no_hdr0=''),
  114614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114614, CurrentID=0, Status='live', Symbol='MIR155HG', Aliases='BIC, MIRHG2, NCRNA00172', description='MIR155 host gene', other_designations='B-cell receptor inducible|BIC transcript|microRNA host gene 2 (non-protein coding)', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25562145, end_position_on_the_genomic_accession=25575168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927583, CurrentID=0, Status='live', Symbol='LOC101927583', Aliases='', description='uncharacterized LOC101927583', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57618405, end_position_on_the_genomic_accession=57621903, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928459, CurrentID=0, Status='live', Symbol='LOC101928459', Aliases='', description='uncharacterized LOC101928459', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166883679, end_position_on_the_genomic_accession=166888855, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114625, CurrentID=0, Status='live', Symbol='ERMAP', Aliases='BTN5, PRO2801, RD, SC', description='erythroblast membrane-associated protein (Scianna blood group)', other_designations='Radin blood group (Rd)|Scianna blood group (Sc)|radin blood group antigen|scianna blood group antigen', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42817105, end_position_on_the_genomic_accession=42844989, orientation='plus', exon_count=13, OMIM=609017, no_hdr0=''),
  101927585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927585, CurrentID=0, Status='live', Symbol='LOC101927585', Aliases='', description='uncharacterized LOC101927585', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16685985, end_position_on_the_genomic_accession=16704540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480654, CurrentID=0, Status='live', Symbol='RNU6-1268P', Aliases='', description='RNA, U6 small nuclear 1268, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420258, CurrentID=0, Status='live', Symbol='PLA2G12AP1', Aliases='', description='phospholipase A2 group XIIA pseudogene 1', other_designations='phospholipase A2, group XIIA pseudogene 1', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52368676, end_position_on_the_genomic_accession=52369242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  245711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245711, CurrentID=0, Status='live', Symbol='SPDYA', Aliases='RINGO3, RINGOA, SPDY1, SPY1', description='speedy/RINGO cell cycle regulator family member A', other_designations='RINGO A|hSpy/Ringo A|rapid inducer of G2/M progression in oocytes A|speedy homolog A|speedy-1', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28810834, end_position_on_the_genomic_accession=28850610, orientation='plus', exon_count=8, OMIM=614029, no_hdr0=''),
  100420259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420259, CurrentID=0, Status='live', Symbol='GDI2P2', Aliases='', description='GDP dissociation inhibitor 2 pseudogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72274543, end_position_on_the_genomic_accession=72275157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289188, CurrentID=0, Status='live', Symbol='FAM197Y4P', Aliases='FAM197Y4', description='family with sequence similarity 197 Y-linked member 4, pseudogene', other_designations='family with sequence similarity 197, Y-linked, member 4|family with sequence similarity 197, Y-linked, member 4, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9484323, end_position_on_the_genomic_accession=9485810, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376431, CurrentID=0, Status='live', Symbol='LOC105376431', Aliases='', description='uncharacterized LOC105376431', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15104511, end_position_on_the_genomic_accession=15140192, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  114655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114655, CurrentID=0, Status='live', Symbol='SNORD73B', Aliases='RNU73B, U73B', description='small nucleolar RNA, C/D box U73B (pseudogene)', other_designations='RNA, U73B small nucleolar pseudogene|U73B small nucleolar RNA|U73B snoRNA|small nucleolar RNA, C/D box U73B pseudogene', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151101957, end_position_on_the_genomic_accession=151102231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114657, CurrentID=0, Status='live', Symbol='COX7BP1', Aliases='bK714B7.1', description='cytochrome c oxidase subunit VIIb pseudogene 1', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=35194402, end_position_on_the_genomic_accession=35195032, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114659, CurrentID=0, Status='live', Symbol='LRRC37B', Aliases='LRRC37', description='leucine rich repeat containing 37B', other_designations='C66 SLIT-like testicular protein', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32007372, end_position_on_the_genomic_accession=32053500, orientation='plus', exon_count=16, OMIM=616558, no_hdr0=''),
  101927591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927591, CurrentID=0, Status='live', Symbol='DPP10-AS3', Aliases='', description='DPP10 antisense RNA 3', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=114828432, end_position_on_the_genomic_accession=114833548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390482, CurrentID=0, Status='live', Symbol='AKR1B1P5', Aliases='AKR1B1P', description='aldo-keto reductase family 1, member B1 pseudogene 5', other_designations='aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59135122, end_position_on_the_genomic_accession=59136451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927592, CurrentID=0, Status='live', Symbol='LOC101927592', Aliases='', description='uncharacterized LOC101927592', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126915227, end_position_on_the_genomic_accession=127060397, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  376818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376818, CurrentID=0, Status='live', Symbol='E2F6P2', Aliases='', description='E2F transcription factor 6 pseudogene 2', other_designations='E2F6 pseudogene 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21140441, end_position_on_the_genomic_accession=21142648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379155, CurrentID=0, Status='live', Symbol='LOC105379155', Aliases='', description='uncharacterized LOC105379155', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124403548, end_position_on_the_genomic_accession=124409975, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377929, CurrentID=0, Status='live', Symbol='LOC105377929', Aliases='', description='uncharacterized LOC105377929', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108030250, end_position_on_the_genomic_accession=108040387, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100289193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289193, CurrentID=0, Status='live', Symbol='LOC100289193', Aliases='', description='ubiquitin-conjugating enzyme E2L 3 pseudogene', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62292016, end_position_on_the_genomic_accession=62292334, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927595, CurrentID=0, Status='live', Symbol='LOC101927595', Aliases='', description='uncharacterized LOC101927595', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=303405, end_position_on_the_genomic_accession=306956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480656, CurrentID=0, Status='live', Symbol='RNU6-1292P', Aliases='', description='RNA, U6 small nuclear 1292, pseudogene', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  376844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376844, CurrentID=0, Status='live', Symbol='SDC4P', Aliases='', description='syndecan 4 pseudogene', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30481290, end_position_on_the_genomic_accession=30481756, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420269, CurrentID=0, Status='live', Symbol='FBP2P1', Aliases='', description='fructose-1,6-bisphosphatase 2 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87826498, end_position_on_the_genomic_accession=87837232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379675, CurrentID=0, Status='live', Symbol='LOC105379675', Aliases='', description='uncharacterized LOC105379675', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379160, CurrentID=0, Status='live', Symbol='LOC105379160', Aliases='', description='uncharacterized LOC105379160', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125804200, end_position_on_the_genomic_accession=125828076, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927598, CurrentID=0, Status='live', Symbol='LOC101927598', Aliases='', description='uncharacterized LOC101927598', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51763239, end_position_on_the_genomic_accession=51825019, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105379162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379162, CurrentID=0, Status='live', Symbol='LOC105379162', Aliases='', description='uncharacterized LOC105379162', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126664371, end_position_on_the_genomic_accession=126674976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927600, CurrentID=0, Status='live', Symbol='LOC101927600', Aliases='', description='uncharacterized LOC101927600', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112216882, end_position_on_the_genomic_accession=112227697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480657, CurrentID=0, Status='live', Symbol='RNU6-1297P', Aliases='', description='RNA, U6 small nuclear 1297, pseudogene', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  84657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84657, CurrentID=0, Status='live', Symbol='LINC00852', Aliases='C3orf42, GHRL-AS2, GHRLOS2, NAG73', description='long intergenic non-protein coding RNA 852', other_designations='GHRL antisense RNA 2 (non-protein coding)|NPC-related protein NAG73|ghrelin opposite strand RNA 2 (non-protein coding)|ghrelin opposite strand/antisense RNA 2 (tail to tail)', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10284419, end_position_on_the_genomic_accession=10285746, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  245802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245802, CurrentID=0, Status='live', Symbol='MS4A6E', Aliases='', description='membrane-spanning 4-domains, subfamily A, member 6E', other_designations='', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60334882, end_position_on_the_genomic_accession=60340968, orientation='plus', exon_count=4, OMIM=608402, no_hdr0=''),
  101927602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927602, CurrentID=0, Status='live', Symbol='LOC101927602', Aliases='', description='uncharacterized LOC101927602', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67365442, end_position_on_the_genomic_accession=67414309, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  245806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245806, CurrentID=0, Status='live', Symbol='VGLL2', Aliases='VGL2, VITO1', description='vestigial like family member 2', other_designations='Vestigial and Tondu related protein 1|vestigial like 2|vestigial-like family member 2', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117265558, end_position_on_the_genomic_accession=117273565, orientation='plus', exon_count=5, OMIM=609979, no_hdr0=''),
  101927445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927445, CurrentID=0, Status='live', Symbol='LOC101927445', Aliases='', description='uncharacterized LOC101927445', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101839949, end_position_on_the_genomic_accession=101910450, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101927603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927603, CurrentID=0, Status='live', Symbol='LOC101927603', Aliases='', description='uncharacterized LOC101927603', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85534067, end_position_on_the_genomic_accession=85538371, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245812, CurrentID=0, Status='live', Symbol='CNPY4', Aliases='PRAT4B', description='canopy FGF signaling regulator 4', other_designations='PRotein Associated with Tlr4|canopy 4 homolog', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100119642, end_position_on_the_genomic_accession=100125508, orientation='plus', exon_count=6, OMIM=610047, no_hdr0=''),
  106481273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481273, CurrentID=0, Status='live', Symbol='RNU6-298P', Aliases='', description='RNA, U6 small nuclear 298, pseudogene', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928463, CurrentID=0, Status='live', Symbol='LOC101928463', Aliases='', description='uncharacterized LOC101928463', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17092943, end_position_on_the_genomic_accession=17107584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927604, CurrentID=0, Status='live', Symbol='LOC101927604', Aliases='', description='uncharacterized LOC101927604', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230874846, end_position_on_the_genomic_accession=230879015, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  245821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245821, CurrentID=0, Status='live', Symbol='DFNB34', Aliases='', description='deafness, autosomal recessive 34', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420277, CurrentID=0, Status='live', Symbol='PRELID3BP9', Aliases='', description='PRELI domain containing 3B pseudogene 9', other_designations='slowmo homolog 2 (Drosophila) pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148418654, end_position_on_the_genomic_accession=148419256, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480658, CurrentID=0, Status='live', Symbol='RNU6-1310P', Aliases='', description='RNA, U6 small nuclear 1310, pseudogene', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100048907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048907, CurrentID=0, Status='live', Symbol='ZDHHC20P1', Aliases='', description='zinc finger, DHHC-type containing 20 pseudogene 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29708141, end_position_on_the_genomic_accession=29708547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114757, CurrentID=0, Status='live', Symbol='CYGB', Aliases='HGB, STAP', description='cytoglobin', other_designations='histoglobin|stellate cell activation-associated protein', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76527348, end_position_on_the_genomic_accession=76557597, orientation='minus', exon_count=7, OMIM=608759, no_hdr0=''),
  114758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114758, CurrentID=0, Status='live', Symbol='CSPG4P1Y', Aliases='CSPG4LY, CSPG4LYP1, CSPG4P1, CSPG4PY1', description='chondroitin sulfate proteoglycan 4 pseudogene 1, Y-linked', other_designations='chondroitin sulfate proteoglycan 4 pseudogene, Y-linked 1|chondroitin sulfate proteoglycan 4-like, Y-linked pseudogene 1', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25482908, end_position_on_the_genomic_accession=25486705, orientation='plus', exon_count=2, OMIM=400034, no_hdr0=''),
  114760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114760, CurrentID=0, Status='live', Symbol='TTTY3', Aliases='LINC00121, NCRNA00121A, TTY3, TTTY3', description='testis-specific transcript, Y-linked 3 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 121', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25728490, end_position_on_the_genomic_accession=25733388, orientation='plus', exon_count=3, OMIM=400036, no_hdr0=''),
  114761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114761, CurrentID=0, Status='live', Symbol='TTTY4', Aliases='LINC00123, NCRNA00123A, TTY4, TTTY4', description='testis-specific transcript, Y-linked 4 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 123', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22936455, end_position_on_the_genomic_accession=22973284, orientation='plus', exon_count=4, OMIM=400037, no_hdr0=''),
  104355137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355137, CurrentID=0, Status='live', Symbol='LINC01072', Aliases='', description='long intergenic non-protein coding RNA 1072', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20102701, end_position_on_the_genomic_accession=20103230, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  285329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285329, CurrentID=0, Status='live', Symbol='UBQLN4P1', Aliases='C3orf12, UBQLN4P', description='ubiquilin 4 pseudogene 1', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148984184, end_position_on_the_genomic_accession=148987750, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114769, CurrentID=0, Status='live', Symbol='CARD16', Aliases='COP, COP1, PSEUDO-ICE', description='caspase recruitment domain family member 16', other_designations='CARD only domain-containing protein 1|CARD only protein|CARD-only protein 1|caspase recruitment domain family, member 16|caspase recruitment domain-only protein 1|caspase-1 dominant-negative inhibitor pseudo-ICE|caspase-1 inhibitor COP|pseudo interleukin-1 beta converting enzyme|pseudo interleukin-1beta converting enzyme|pseudo-IL1B-converting enzyme', map_location='11q22', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105040306, end_position_on_the_genomic_accession=105045416, orientation='minus', exon_count=4, OMIM=615680, no_hdr0=''),
  114770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114770, CurrentID=0, Status='live', Symbol='PGLYRP2', Aliases='HMFT0141, PGLYRPL, PGRP-L, PGRPL, TAGL-like, tagL, tagL-alpha, tagl-beta', description='peptidoglycan recognition protein 2', other_designations='peptidoglycan recognition protein L|peptidoglycan recognition protein long', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15468645, end_position_on_the_genomic_accession=15479504, orientation='minus', exon_count=5, OMIM=608199, no_hdr0=''),
  114771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114771, CurrentID=0, Status='live', Symbol='PGLYRP3', Aliases='PGLYRPIalpha, PGRP-Ialpha, PGRPIA', description='peptidoglycan recognition protein 3', other_designations='peptidoglycan recognition protein I alpha|peptidoglycan recognition protein intermediate alpha', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153297589, end_position_on_the_genomic_accession=153312975, orientation='minus', exon_count=11, OMIM=608197, no_hdr0=''),
  105376435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376435, CurrentID=0, Status='live', Symbol='LOC105376435', Aliases='', description='uncharacterized LOC105376435', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16817712, end_position_on_the_genomic_accession=16821245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927609, CurrentID=0, Status='live', Symbol='LOC101927609', Aliases='', description='uncharacterized LOC101927609', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17279944, end_position_on_the_genomic_accession=17298538, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  392264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392264, CurrentID=0, Status='live', Symbol='LOC392264', Aliases='', description='zinc finger protein 532 pseudogene', other_designations='', map_location='8q24.12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119480278, end_position_on_the_genomic_accession=119481471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646709, CurrentID=0, Status='live', Symbol='LOC646709', Aliases='', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 1 pseudogene', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56342339, end_position_on_the_genomic_accession=56343724, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114780, CurrentID=0, Status='live', Symbol='PKD1L2', Aliases='PC1L2', description='polycystic kidney disease 1-like 2 (gene/pseudogene)', other_designations='PC1-like 2 protein|polycystin-1L2', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81100879, end_position_on_the_genomic_accession=81220370, orientation='minus', exon_count=44, OMIM=607894, no_hdr0=''),
  114781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114781, CurrentID=0, Status='live', Symbol='BTBD9', Aliases='dJ322I12.1', description='BTB (POZ) domain containing 9', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38168451, end_position_on_the_genomic_accession=38640547, orientation='minus', exon_count=17, OMIM=611237, no_hdr0=''),
  114783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114783, CurrentID=0, Status='live', Symbol='LMTK3', Aliases='LMR3, PPP1R101, TYKLM3', description='lemur tyrosine kinase 3', other_designations='protein phosphatase 1, regulatory subunit 101', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48485271, end_position_on_the_genomic_accession=48513905, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  114784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114784, CurrentID=0, Status='live', Symbol='CSMD2', Aliases='dJ1007G16.1, dJ1007G16.2, dJ947L8.1', description='CUB and Sushi multiple domains 2', other_designations='CUB and Sushi (SCR repeat) domain|CUB and sushi multiple domains protein 2', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33513999, end_position_on_the_genomic_accession=34165842, orientation='minus', exon_count=73, OMIM=608398, no_hdr0=''),
  114785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114785, CurrentID=0, Status='live', Symbol='MBD6', Aliases='', description='methyl-CpG binding domain protein 6', other_designations='methyl-CpG-binding protein MBD6', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57520710, end_position_on_the_genomic_accession=57530148, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  114786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114786, CurrentID=0, Status='live', Symbol='XKR4', Aliases='XRG4', description='X-linked Kx blood group related 4', other_designations='X Kell blood group precursor-related family, member 4|XK, Kell blood group complex subunit-related family, member 4', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55102457, end_position_on_the_genomic_accession=55526155, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  114787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114787, CurrentID=0, Status='live', Symbol='GPRIN1', Aliases='GRIN1', description='G protein regulated inducer of neurite outgrowth 1', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176595802, end_position_on_the_genomic_accession=176617338, orientation='minus', exon_count=3, OMIM=611239, no_hdr0=''),
  114788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114788, CurrentID=0, Status='live', Symbol='CSMD3', Aliases='', description='CUB and Sushi multiple domains 3', other_designations='CUB and sushi multiple domains protein 3', map_location='8q23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=112222928, end_position_on_the_genomic_accession=113437467, orientation='minus', exon_count=81, OMIM=608399, no_hdr0=''),
  114789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114789, CurrentID=0, Status='live', Symbol='SLC25A25', Aliases='MCSC, PCSCL, SCAMC-2', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 25', other_designations='mitochondrial ATP-Mg/Pi carrier protein 3|mitochondrial Ca(2+)-dependent solute carrier protein 3|short calcium-binding mitochondrial carrier 2|small calcium-binding mitochondrial carrier 2|solute carrier family 25, member 25', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128068200, end_position_on_the_genomic_accession=128109258, orientation='plus', exon_count=14, OMIM=608745, no_hdr0=''),
  114790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114790, CurrentID=0, Status='live', Symbol='STK11IP', Aliases='LIP1, LKB1IP1, STK11IP', description='serine/threonine kinase 11 interacting protein', other_designations='LKB1-interacting protein 1|STK11 interacting protein', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219597871, end_position_on_the_genomic_accession=219616451, orientation='plus', exon_count=29, OMIM=607172, no_hdr0=''),
  114791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114791, CurrentID=0, Status='live', Symbol='TUBGCP5', Aliases='GCP5', description='tubulin gamma complex associated protein 5', other_designations='GCP-5|gamma-tubulin complex component GCP5|tubulin, gamma complex associated protein 5', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22991677, end_position_on_the_genomic_accession=23039673, orientation='minus', exon_count=24, OMIM=608147, no_hdr0=''),
  114792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114792, CurrentID=0, Status='live', Symbol='KLHL32', Aliases='BKLHD5, KIAA1900, UG0030H05, dJ21F7.1', description='kelch like family member 32', other_designations='BTB and kelch domain containing 5|kelch-like family member 32', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96898013, end_position_on_the_genomic_accession=97142614, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  114793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114793, CurrentID=0, Status='live', Symbol='FMNL2', Aliases='FHOD2', description='formin like 2', other_designations='formin homology 2 domain containing 2|formin homology 2 domain-containing protein 2|formin-like 2', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152335170, end_position_on_the_genomic_accession=152649834, orientation='plus', exon_count=31, OMIM=616285, no_hdr0=''),
  114794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114794, CurrentID=0, Status='live', Symbol='ELFN2', Aliases='LRRC62, PPP1R29', description='extracellular leucine-rich repeat and fibronectin type III domain containing 2', other_designations='dJ63G5.3 (putative Leucine rich protein)|leucine-rich repeat and fibronectin type-III domain-containing protein 6|leucine-rich repeat-containing protein 62', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37340644, end_position_on_the_genomic_accession=37427470, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  114795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114795, CurrentID=0, Status='live', Symbol='TMEM132B', Aliases='', description='transmembrane protein 132B', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125186774, end_position_on_the_genomic_accession=125662377, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  376940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=376940, CurrentID=0, Status='live', Symbol='ZC3H6', Aliases='ZC3HDC6', description='zinc finger CCCH-type containing 6', other_designations='zinc finger CCCH type domain containing 6|zinc finger CCCH-type domain containing 6', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112275559, end_position_on_the_genomic_accession=112340063, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  114798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114798, CurrentID=0, Status='live', Symbol='SLITRK1', Aliases='LRRC12, TTM', description='SLIT and NTRK like family member 1', other_designations='SLIT and NTRK-like family, member 1|leucine-rich repeat-containing protein 12|slit and trk like gene 1', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=83877205, end_position_on_the_genomic_accession=83882393, orientation='minus', exon_count=2, OMIM=609678, no_hdr0=''),
  114799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114799, CurrentID=0, Status='live', Symbol='ESCO1', Aliases='A930014I12Rik, CTF, ECO1, EFO1, ESO1', description='establishment of sister chromatid cohesion N-acetyltransferase 1', other_designations='CTF7 homolog 1|ECO1 homolog 1|EFO1p|ESO1 homolog 1|N-acetyltransferase ESCO1 variant 2|establishment factor-like protein 1|establishment of cohesion 1 homolog 1|hEFO1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21529281, end_position_on_the_genomic_accession=21600736, orientation='minus', exon_count=13, OMIM=609674, no_hdr0=''),
  114800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114800, CurrentID=0, Status='live', Symbol='CCDC85A', Aliases='', description='coiled-coil domain containing 85A', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56184123, end_position_on_the_genomic_accession=56386174, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114801, CurrentID=0, Status='live', Symbol='TMEM200A', Aliases='KIAA1913, TTMA, TTMC', description='transmembrane protein 200A', other_designations='two transmembrane C', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130366281, end_position_on_the_genomic_accession=130443424, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114803, CurrentID=0, Status='live', Symbol='MYSM1', Aliases='2A-DUB, 2ADUB', description='Myb-like, SWIRM and MPN domains 1', other_designations='myb-like, SWIRM and MPN domain-containing protein 1', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58654739, end_position_on_the_genomic_accession=58700092, orientation='minus', exon_count=23, OMIM=612176, no_hdr0=''),
  100384884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100384884, CurrentID=0, Status='live', Symbol='IGHV1OR15-5', Aliases='IGHV1OR155', description='immunoglobulin heavy variable 1/OR15-5 (non-functional)', other_designations='IGHV1/OR15-5|immunoglobulin heavy variable 1/OR15-5 (pseudogene)|immunoglobulin heavy variable 1/OR15-5 pseudogene', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114805, CurrentID=0, Status='live', Symbol='GALNT13', Aliases='GalNAc-T13', description='polypeptide N-acetylgalactosaminyltransferase 13', other_designations='GalNAc transferase 13|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 13|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 13 (GalNAc-T13)|polypeptide GalNAc transferase 13|pp-GaNTase 13|protein-UDP acetylgalactosaminyltransferase 13', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153871898, end_position_on_the_genomic_accession=154453977, orientation='plus', exon_count=15, OMIM=608369, no_hdr0=''),
  101927615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927615, CurrentID=0, Status='live', Symbol='LOC101927615', Aliases='', description='uncharacterized LOC101927615', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9045567, end_position_on_the_genomic_accession=9064076, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114814, CurrentID=0, Status='live', Symbol='GNRHR2', Aliases='GnRH-II-R', description='gonadotropin releasing hormone receptor 2, pseudogene', other_designations='GnRH receptor II 5TM|Type II GnRH receptor|gonadotropin-releasing hormone (type 2) receptor 2, pseudogene', map_location='1q12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145919013, end_position_on_the_genomic_accession=145925341, orientation='plus', exon_count=4, OMIM=612875, no_hdr0=''),
  114815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114815, CurrentID=0, Status='live', Symbol='SORCS1', Aliases='hSorCS', description='sortilin-related VPS10 domain containing receptor 1', other_designations='VPS10 domain receptor protein SORCS 1', map_location='10q23-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106573663, end_position_on_the_genomic_accession=107181126, orientation='minus', exon_count=34, OMIM=606283, no_hdr0=''),
  114817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114817, CurrentID=0, Status='live', Symbol='CSPG4P5', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 5', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84398315, end_position_on_the_genomic_accession=84411700, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  114818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114818, CurrentID=0, Status='live', Symbol='KLHL29', Aliases='KBTBD9', description='kelch like family member 29', other_designations='kelch repeat and BTB (POZ) domain containing 9|kelch repeat and BTB domain-containing protein 9|kelch-like 29|kelch-like family member 29', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23385427, end_position_on_the_genomic_accession=23708613, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  114819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114819, CurrentID=0, Status='live', Symbol='CROCCP3', Aliases='CROCCL2, dJ798A10.2', description='ciliary rootlet coiled-coil, rootletin pseudogene 3', other_designations='ciliary rootlet coiled-coil, rootletin-like 2', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16467436, end_position_on_the_genomic_accession=16492701, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  114821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114821, CurrentID=0, Status='live', Symbol='ZBED9', Aliases='Buster4, SCAND3, ZFP38-L, ZNF305P2, ZNF452, dJ1186N24.3', description='zinc finger, BED-type containing 9', other_designations='SCAN domain containing 3|dJ1186N24.3 (novel zinc finger protein)|transposon-derived Buster4 transposase-like protein|zinc finger BED domain-containing protein 9|zinc finger protein 305 pseudogene 2|zinc finger protein 452', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28571338, end_position_on_the_genomic_accession=28616203, orientation='minus', exon_count=8, OMIM=615254, no_hdr0=''),
  114822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114822, CurrentID=0, Status='live', Symbol='RHPN1', Aliases='ODF5, RHOPHILIN, RHPN', description='rhophilin, Rho GTPase binding protein 1', other_designations='GTP-Rho-binding protein 1|outer dense fiber of sperm tails 5|rhophilin 1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143368842, end_position_on_the_genomic_accession=143384220, orientation='plus', exon_count=18, OMIM=601031, no_hdr0=''),
  114823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114823, CurrentID=0, Status='live', Symbol='LENG8', Aliases='pp13842', description='leukocyte receptor cluster (LRC) member 8', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54448719, end_position_on_the_genomic_accession=54462046, orientation='plus', exon_count=17, OMIM=616575, no_hdr0=''),
  114824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114824, CurrentID=0, Status='live', Symbol='PNMA5', Aliases='', description='paraneoplastic Ma antigen family member 5', other_designations='tumor antigen BJ-HCC-25', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152988824, end_position_on_the_genomic_accession=152994127, orientation='minus', exon_count=5, OMIM=300916, no_hdr0=''),
  114825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114825, CurrentID=0, Status='live', Symbol='PWWP2A', Aliases='MST101', description='PWWP domain containing 2A', other_designations='MSTP101', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160043562, end_position_on_the_genomic_accession=160120808, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  114826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114826, CurrentID=0, Status='live', Symbol='SMYD4', Aliases='ZMYND21', description='SET and MYND domain containing 4', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1779535, end_position_on_the_genomic_accession=1829881, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  114827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114827, CurrentID=0, Status='live', Symbol='FHAD1', Aliases='', description='forkhead-associated (FHA) phosphopeptide binding domain 1', other_designations='FHA domain-containing protein 1', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15236602, end_position_on_the_genomic_accession=15400283, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  101927619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927619, CurrentID=0, Status='live', Symbol='LOC101927619', Aliases='', description='uncharacterized LOC101927619', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=198299363, end_position_on_the_genomic_accession=198375097, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  245908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245908, CurrentID=0, Status='live', Symbol='DEFB105A', Aliases='BD-5, DEFB-5, DEFB105', description='defensin beta 105A', other_designations='beta defensin 5|defensin, beta 105A|defensin, beta 5', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7820779, end_position_on_the_genomic_accession=7823891, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  245909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245909, CurrentID=0, Status='live', Symbol='DEFB106A', Aliases='BD-6, DEFB-6, DEFB106', description='defensin beta 106A', other_designations='beta-defensin 6|defensin, beta 106A|defensin, beta 6', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7825172, end_position_on_the_genomic_accession=7829053, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  245910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245910, CurrentID=0, Status='live', Symbol='DEFB107A', Aliases='BD-7, DEFB-7, DEFB107', description='defensin beta 107A', other_designations='beta-defensin 7|defensin, beta 107A|defensin, beta 7', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7811720, end_position_on_the_genomic_accession=7815716, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245911, CurrentID=0, Status='live', Symbol='DEFB108B', Aliases='DEFB-8, hBD-8', description='defensin beta 108B', other_designations='beta-defensin 8|defensin, beta 108B|defensin, beta 8', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71833200, end_position_on_the_genomic_accession=71837562, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  245912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245912, CurrentID=0, Status='live', Symbol='DEFB109P1', Aliases='DEFB-9, DEFB109, DEFB109A', description='defensin beta 109 pseudogene 1', other_designations='defensin, beta 109, pseudogene 1|defensin, beta 9 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12393284, end_position_on_the_genomic_accession=12400365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245913, CurrentID=0, Status='live', Symbol='DEFB110', Aliases='DEFB-10, DEFB-11, DEFB111', description='defensin beta 110', other_designations='beta-defensin 10|beta-defensin 11|beta-defensin 111|defensin, beta 10|defensin, beta 110 locus', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50009138, end_position_on_the_genomic_accession=50021981, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  245915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245915, CurrentID=0, Status='live', Symbol='DEFB112', Aliases='DEFB-12', description='defensin beta 112', other_designations='beta-defensin 12|defensin, beta 112|defensin, beta 12', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=50043575, end_position_on_the_genomic_accession=50048651, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480661, CurrentID=0, Status='live', Symbol='RNU6-1336P', Aliases='', description='RNA, U6 small nuclear 1336, pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100048922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100048922, CurrentID=0, Status='live', Symbol='RPL35AP4', Aliases='BPG294E21.7', description='ribosomal protein L35a pseudogene 4', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33389356, end_position_on_the_genomic_accession=33389479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927621, CurrentID=0, Status='live', Symbol='LINC01534', Aliases='', description='long intergenic non-protein coding RNA 1534', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36685439, end_position_on_the_genomic_accession=36687449, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379184, CurrentID=0, Status='live', Symbol='LOC105379184', Aliases='', description='uncharacterized LOC105379184', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134111614, end_position_on_the_genomic_accession=134114629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245927, CurrentID=0, Status='live', Symbol='DEFB113', Aliases='DEFB-13', description='defensin beta 113', other_designations='beta-defensin 13|defensin, beta 113|defensin, beta 13', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49968677, end_position_on_the_genomic_accession=49969625, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245928, CurrentID=0, Status='live', Symbol='DEFB114', Aliases='DEFB-14, DEFB14', description='defensin beta 114', other_designations='beta-defensin 14|defensin, beta 114|defensin, beta 14', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49960292, end_position_on_the_genomic_accession=49964105, orientation='minus', exon_count=2, OMIM=615243, no_hdr0=''),
  245929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245929, CurrentID=0, Status='live', Symbol='DEFB115', Aliases='DEFB-15', description='defensin beta 115', other_designations='beta-defensin 15|defensin, beta 115|defensin, beta 15', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31257664, end_position_on_the_genomic_accession=31259632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  245930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245930, CurrentID=0, Status='live', Symbol='DEFB116', Aliases='DEFB-16', description='defensin beta 116', other_designations='beta-defensin 16|defensin, beta 116|defensin, beta 16', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31303212, end_position_on_the_genomic_accession=31308585, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245931, CurrentID=0, Status='live', Symbol='DEFB117', Aliases='DEFB-17', description='defensin beta 117 (pseudogene)', other_designations='defensin, beta 117 (pseudogene)|defensin, beta 17', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31360775, end_position_on_the_genomic_accession=31360903, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  245932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245932, CurrentID=0, Status='live', Symbol='DEFB119', Aliases='DEFB-19, DEFB-20, DEFB120, DEFB20, ESC42-RELA, ESC42-RELB', description='defensin beta 119', other_designations='defensin, beta 119|defensin, beta 120|defensin, beta 19|defensin, beta 20', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31377163, end_position_on_the_genomic_accession=31390649, orientation='minus', exon_count=5, OMIM=615997, no_hdr0=''),
  105379186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379186, CurrentID=0, Status='live', Symbol='LOC105379186', Aliases='', description='uncharacterized LOC105379186', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134412006, end_position_on_the_genomic_accession=134439793, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  245934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245934, CurrentID=0, Status='live', Symbol='DEFB121', Aliases='DEFB21, ESC42RELC', description='defensin beta 121', other_designations='beta-defensin 21|defensin, beta 121|defensin, beta 21', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31404845, end_position_on_the_genomic_accession=31419068, orientation='minus', exon_count=8, OMIM=616075, no_hdr0=''),
  245935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245935, CurrentID=0, Status='live', Symbol='DEFB122', Aliases='DEFB-22P, DEFB22, DEFB122', description='defensin beta 122 (pseudogene)', other_designations='defensin, beta 122 (pseudogene)|defensin, beta 22', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31421439, end_position_on_the_genomic_accession=31429180, orientation='minus', exon_count=3, OMIM=616077, no_hdr0=''),
  245936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245936, CurrentID=0, Status='live', Symbol='DEFB123', Aliases='DEFB-23, DEFB23, ESC42-RELD', description='defensin beta 123', other_designations='beta defensin 23|beta-defensin 23|defensin, beta 123|specific to testis and epididymis', map_location='20q11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31440519, end_position_on_the_genomic_accession=31450257, orientation='plus', exon_count=2, OMIM=616076, no_hdr0=''),
  245937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245937, CurrentID=0, Status='live', Symbol='DEFB124', Aliases='DEFB-24', description='defensin beta 124', other_designations='beta-defensin 24|defensin, beta 124|defensin, beta 24', map_location='20q11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31465506, end_position_on_the_genomic_accession=31473013, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245938, CurrentID=0, Status='live', Symbol='DEFB125', Aliases='DEFB-25', description='defensin beta 125', other_designations='beta defensin 25|beta-defensin 25|defensin, beta 125', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=87672, end_position_on_the_genomic_accession=96573, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  245939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245939, CurrentID=0, Status='live', Symbol='DEFB128', Aliases='DEFB-28, DEFB28, hBD-28', description='defensin beta 128', other_designations='beta-defensin 28|defensin, beta 128|defensin, beta 28', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=187853, end_position_on_the_genomic_accession=189681, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245940, CurrentID=0, Status='live', Symbol='DEFB130', Aliases='DEFB-30L, DEFB30, DEFB130', description='defensin beta 130', other_designations='beta-defensin 30|defensin, beta 130|defensin, beta 30', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12310962, end_position_on_the_genomic_accession=12318316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379188, CurrentID=0, Status='live', Symbol='LOC105379188', Aliases='', description='uncharacterized LOC105379188', other_designations='CTC-276P9.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135122067, end_position_on_the_genomic_accession=135131322, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480662, CurrentID=0, Status='live', Symbol='RNU1-30P', Aliases='', description='RNA, U1 small nuclear 30, pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114876, CurrentID=0, Status='live', Symbol='OSBPL1A', Aliases='ORP-1, ORP1, OSBPL1B', description='oxysterol binding protein like 1A', other_designations='OSBP-related protein 1|oxysterol binding protein-like 1A|oxysterol binding protein-like 1B|oxysterol-binding protein-related protein 1 variant 1|oxysterol-binding protein-related protein 1 variant 2', map_location='18q11.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24162045, end_position_on_the_genomic_accession=24397882, orientation='minus', exon_count=31, OMIM=606730, no_hdr0=''),
  114879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114879, CurrentID=0, Status='live', Symbol='OSBPL5', Aliases='OBPH1, ORP5', description='oxysterol binding protein like 5', other_designations='ORP-5|OSBP-related protein 5|oxysterol binding protein-like 5|oxysterol-binding protein homolog 1|oxysterol-binding protein homologue 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3087115, end_position_on_the_genomic_accession=3165352, orientation='minus', exon_count=23, OMIM=606733, no_hdr0=''),
  114880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114880, CurrentID=0, Status='live', Symbol='OSBPL6', Aliases='ORP6', description='oxysterol binding protein like 6', other_designations='ORP-6|OSBP-related protein 6|oxysterol binding protein-like 6', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178194481, end_position_on_the_genomic_accession=178399433, orientation='plus', exon_count=27, OMIM=606734, no_hdr0=''),
  114881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114881, CurrentID=0, Status='live', Symbol='OSBPL7', Aliases='ORP7', description='oxysterol binding protein like 7', other_designations='ORP-7|OSBP-related protein 7|oxysterol binding protein-like 7', map_location='17q21', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47807367, end_position_on_the_genomic_accession=47821826, orientation='minus', exon_count=24, OMIM=606735, no_hdr0=''),
  114882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114882, CurrentID=0, Status='live', Symbol='OSBPL8', Aliases='MST120, MSTP120, ORP8, OSBP10', description='oxysterol binding protein like 8', other_designations='ORP-8|OSBP-related protein 8|oxysterol binding protein-like 8', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76351797, end_position_on_the_genomic_accession=76559809, orientation='minus', exon_count=30, OMIM=606736, no_hdr0=''),
  114883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114883, CurrentID=0, Status='live', Symbol='OSBPL9', Aliases='ORP-9, ORP9', description='oxysterol binding protein like 9', other_designations='OSBP-related protein 9|oxysterol binding protein-like 9', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51616874, end_position_on_the_genomic_accession=51789219, orientation='plus', exon_count=28, OMIM=606737, no_hdr0=''),
  114884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114884, CurrentID=0, Status='live', Symbol='OSBPL10', Aliases='ORP10, OSBP9', description='oxysterol binding protein like 10', other_designations='ORP-10|OSBP-related protein 10|oxysterol binding protein-like 10', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31660825, end_position_on_the_genomic_accession=31981850, orientation='minus', exon_count=16, OMIM=606738, no_hdr0=''),
  114885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114885, CurrentID=0, Status='live', Symbol='OSBPL11', Aliases='ORP-11, ORP11, OSBP12, TCCCIA00292', description='oxysterol binding protein like 11', other_designations='OSBP-related protein 11|oxysterol binding protein-like 11', map_location='3q21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125528858, end_position_on_the_genomic_accession=125595537, orientation='minus', exon_count=13, OMIM=606739, no_hdr0=''),
  101927628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927628, CurrentID=0, Status='live', Symbol='LOC101927628', Aliases='', description='uncharacterized LOC101927628', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101758394, end_position_on_the_genomic_accession=101762342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  114890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114890, CurrentID=0, Status='live', Symbol='MRSS', Aliases='', description='Mental retardation, X-linked, with short stature', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300360, no_hdr0=''),
  106481274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481274, CurrentID=0, Status='live', Symbol='RNU6-301P', Aliases='', description='RNA, U6 small nuclear 301, pseudogene', other_designations='', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927629, CurrentID=0, Status='live', Symbol='LOC101927629', Aliases='', description='uncharacterized LOC101927629', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63730220, end_position_on_the_genomic_accession=63755868, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392268, CurrentID=0, Status='live', Symbol='LOC392268', Aliases='', description='neuropilin and tolloid like 2 pseudogene', other_designations='neuropilin (NRP) and tolloid (TLL)-like 2 pseudogene', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124141900, end_position_on_the_genomic_accession=124143509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114897, CurrentID=0, Status='live', Symbol='C1QTNF1', Aliases='CTRP1, GIP, ZSIG37', description='C1q and tumor necrosis factor related protein 1', other_designations='G protein coupled receptor interacting protein|g protein-coupled receptor-interacting protein', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79022934, end_position_on_the_genomic_accession=79049788, orientation='plus', exon_count=8, OMIM=610365, no_hdr0=''),
  114898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114898, CurrentID=0, Status='live', Symbol='C1QTNF2', Aliases='CTRP2, zacrp2', description='C1q and tumor necrosis factor related protein 2', other_designations='complement-c1q tumor necrosis factor-related protein 2', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160347754, end_position_on_the_genomic_accession=160370652, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  114899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114899, CurrentID=0, Status='live', Symbol='C1QTNF3', Aliases='C1ATNF3, CORCS, CORS, CORS-26, CORS26, CTRP3', description='C1q and tumor necrosis factor related protein 3', other_designations='cartonectin|collagenous repeat-containing sequence 26 kDa protein|collagenous repeat-containing sequence of 26-kDa|secretory protein CORS26', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34017858, end_position_on_the_genomic_accession=34043266, orientation='minus', exon_count=6, OMIM=612045, no_hdr0=''),
  114900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114900, CurrentID=0, Status='live', Symbol='C1QTNF4', Aliases='CTRP4, ZACRP4', description='C1q and tumor necrosis factor related protein 4', other_designations='complement-c1q tumor necrosis factor-related protein 4', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47589664, end_position_on_the_genomic_accession=47594409, orientation='minus', exon_count=3, OMIM=614911, no_hdr0=''),
  245973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=245973, CurrentID=0, Status='live', Symbol='ATP6V1C2', Aliases='ATP6C2, VMA5', description='ATPase, H+ transporting, lysosomal 42kDa, V1 subunit C2', other_designations='V-ATPase C2 subunit|V-ATPase subunit C 2|vacuolar proton pump subunit C 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10720974, end_position_on_the_genomic_accession=10785110, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  114902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114902, CurrentID=0, Status='live', Symbol='C1QTNF5', Aliases='CTRP5', description='C1q and tumor necrosis factor related protein 5', other_designations='C1q TNF-alpha-related protein 5|myonectin', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119338934, end_position_on_the_genomic_accession=119346673, orientation='minus', exon_count=15, OMIM=608752, no_hdr0=''),
  377047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377047, CurrentID=0, Status='live', Symbol='PRSS45', Aliases='TESSP5', description='protease, serine 45', other_designations='protease, serine, 45|putative testis serine protease 5|testis serine protease 5', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46742091, end_position_on_the_genomic_accession=46744755, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114904, CurrentID=0, Status='live', Symbol='C1QTNF6', Aliases='CTFP6, CTRP6, ZACRP6', description='C1q and tumor necrosis factor related protein 6', other_designations='complement-c1q tumor necrosis factor-related protein 6', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37180166, end_position_on_the_genomic_accession=37198301, orientation='minus', exon_count=5, OMIM=614910, no_hdr0=''),
  114905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114905, CurrentID=0, Status='live', Symbol='C1QTNF7', Aliases='CTRP7, ZACRP7', description='C1q and tumor necrosis factor related protein 7', other_designations='complement-c1q tumor necrosis factor-related protein 7', map_location='4p15.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15339818, end_position_on_the_genomic_accession=15446169, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114907, CurrentID=0, Status='live', Symbol='FBXO32', Aliases='Fbx32, MAFbx', description='F-box protein 32', other_designations='atrogin 1|atrogin-1|muscle atrophy F-box protein', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123497887, end_position_on_the_genomic_accession=123541253, orientation='minus', exon_count=9, OMIM=606604, no_hdr0=''),
  114908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114908, CurrentID=0, Status='live', Symbol='TMEM123', Aliases='KCT3, PORIMIN, PORMIN', description='transmembrane protein 123', other_designations='KCT-3|keratinocytes associated transmembrane protein 3|keratinocytes-associated transmembrane protein 3|pro oncosis receptor inducing membrane injury|pro-oncosis receptor inducing membrane injury|serine/threonine-rich receptor', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102396325, end_position_on_the_genomic_accession=102453044, orientation='minus', exon_count=5, OMIM=606356, no_hdr0=''),
  101927632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927632, CurrentID=0, Status='live', Symbol='LOC101927632', Aliases='', description='uncharacterized LOC101927632', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100435706, end_position_on_the_genomic_accession=100440064, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  114915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114915, CurrentID=0, Status='live', Symbol='EPB41L4A-AS1', Aliases='C5orf26, NCRNA00219, TIGA1', description='EPB41L4A antisense RNA 1', other_designations='EPB41L4A antisense RNA 1 (non-protein coding)|transcript induced by growth arrest 1', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112160526, end_position_on_the_genomic_accession=112162501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  114926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114926, CurrentID=0, Status='live', Symbol='SMIM19', Aliases='C8orf40', description='small integral membrane protein 19', other_designations='UPF0697 protein C8orf40', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42541155, end_position_on_the_genomic_accession=42552997, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114928, CurrentID=0, Status='live', Symbol='GPRASP2', Aliases='GASP2', description='G protein-coupled receptor associated sorting protein 2', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102712176, end_position_on_the_genomic_accession=102717733, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101927635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927635, CurrentID=0, Status='live', Symbol='LOC101927635', Aliases='', description='uncharacterized LOC101927635', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48506527, end_position_on_the_genomic_accession=48508839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114932, CurrentID=0, Status='live', Symbol='MRFAP1L1', Aliases='PP784', description='Morf4 family associated protein 1-like 1', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6707701, end_position_on_the_genomic_accession=6709879, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379198, CurrentID=0, Status='live', Symbol='LOC105379198', Aliases='', description='uncharacterized LOC105379198', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127218521, end_position_on_the_genomic_accession=127229743, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480664, CurrentID=0, Status='live', Symbol='RNU1-63P', Aliases='', description='RNA, U1 small nuclear 63, pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927636, CurrentID=0, Status='live', Symbol='LOC101927636', Aliases='', description='uncharacterized LOC101927636', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143912331, end_position_on_the_genomic_accession=143982454, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372076, CurrentID=0, Status='live', Symbol='LOC105372076', Aliases='', description='uncharacterized LOC105372076', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=38655254, end_position_on_the_genomic_accession=38733691, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927637, CurrentID=0, Status='live', Symbol='LOC101927637', Aliases='', description='uncharacterized LOC101927637', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127631269, end_position_on_the_genomic_accession=127636467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927638, CurrentID=0, Status='live', Symbol='LOC101927638', Aliases='', description='solute carrier family 22 member 24-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63264666, end_position_on_the_genomic_accession=63267617, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  400999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=400999, CurrentID=0, Status='live', Symbol='FLJ42351', Aliases='', description='uncharacterized LOC400999', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112641830, end_position_on_the_genomic_accession=112644180, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379202, CurrentID=0, Status='live', Symbol='LOC105379202', Aliases='', description='uncharacterized LOC105379202', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20012241, end_position_on_the_genomic_accession=20014711, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114960, CurrentID=0, Status='live', Symbol='TSGA13', Aliases='', description='testis specific 13', other_designations='testis specific, 13|testis-specific A13', map_location='7q32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130668646, end_position_on_the_genomic_accession=130687432, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927640, CurrentID=0, Status='live', Symbol='LOC101927640', Aliases='', description='uncharacterized LOC101927640', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112236596, end_position_on_the_genomic_accession=112306683, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390531, CurrentID=0, Status='live', Symbol='IGHV1OR15-9', Aliases='IGHV1OR159, VSIG7', description='immunoglobulin heavy variable 1/OR15-9 (non-functional)', other_designations='IGHV1/OR15-9|Putative V-set and immunoglobulin domain-containing protein 7|V-set and immunoglobulin domain containing 7', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=19964656, end_position_on_the_genomic_accession=19965101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480665, CurrentID=0, Status='live', Symbol='RNU1-78P', Aliases='', description='RNA, U1 small nuclear 78, pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927641, CurrentID=0, Status='live', Symbol='LOC101927641', Aliases='', description='uncharacterized LOC101927641', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199608068, end_position_on_the_genomic_accession=199659132, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422711, CurrentID=0, Status='live', Symbol='PRKRIRP2', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86012436, end_position_on_the_genomic_accession=86015419, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  114971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114971, CurrentID=0, Status='live', Symbol='PTPMT1', Aliases='DUSP23, MOSP, PLIP, PNAS-129', description='protein tyrosine phosphatase, mitochondrial 1', other_designations='NB4 apoptosis/differentiation related protein|PTEN-like phosphatase|phosphoinositide lipid phosphatase|protein-tyrosine phosphatase mitochondrial 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47565430, end_position_on_the_genomic_accession=47573461, orientation='plus', exon_count=4, OMIM=609538, no_hdr0=''),
  101927642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927642, CurrentID=0, Status='live', Symbol='LINC01444', Aliases='', description='long intergenic non-protein coding RNA 1444', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14969607, end_position_on_the_genomic_accession=14970468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  114984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114984, CurrentID=0, Status='live', Symbol='FLYWCH2', Aliases='', description='FLYWCH family member 2', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2882804, end_position_on_the_genomic_accession=2899382, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114987, CurrentID=0, Status='live', Symbol='WDR31', Aliases='', description='WD repeat domain 31', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113315265, end_position_on_the_genomic_accession=113340340, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  114990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114990, CurrentID=0, Status='live', Symbol='VASN', Aliases='SLITL2', description='vasorin', other_designations='protein slit-like 2|slit-like 2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4371848, end_position_on_the_genomic_accession=4383528, orientation='plus', exon_count=2, OMIM=608843, no_hdr0=''),
  114991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=114991, CurrentID=0, Status='live', Symbol='ZNF618', Aliases='FP13169, NEDD10', description='zinc finger protein 618', other_designations='neural precursor cell expressed, developmentally down-regulated 10', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113876282, end_position_on_the_genomic_accession=114056595, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  106480666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480666, CurrentID=0, Status='live', Symbol='RNU1-86P', Aliases='', description='RNA, U1 small nuclear 86, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115004, CurrentID=0, Status='live', Symbol='MB21D1', Aliases='C6orf150, cGAS, h-cGAS', description='Mab-21 domain containing 1', other_designations='cGAMP synthase|mab-21 domain-containing protein 1|protein MB21D1', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73416258, end_position_on_the_genomic_accession=73452330, orientation='minus', exon_count=6, OMIM=613973, no_hdr0=''),
  105379210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379210, CurrentID=0, Status='live', Symbol='LOC105379210', Aliases='', description='Ig heavy chain V-III region VH26-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=19986789, end_position_on_the_genomic_accession=19988203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  104355145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355145, CurrentID=0, Status='live', Symbol='NCOA7-AS1', Aliases='', description='NCOA7 antisense RNA 1', other_designations='', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125797861, end_position_on_the_genomic_accession=125818858, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375743, CurrentID=0, Status='live', Symbol='LOC105375743', Aliases='', description='uncharacterized LOC105375743', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124830160, end_position_on_the_genomic_accession=124857485, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105379211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379211, CurrentID=0, Status='live', Symbol='LOC105379211', Aliases='', description='uncharacterized LOC105379211', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20339369, end_position_on_the_genomic_accession=20351652, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100420321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420321, CurrentID=0, Status='live', Symbol='LOC100420321', Aliases='', description='melanoma antigen family A11 pseudogene', other_designations='melanoma antigen family A, 11 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149649177, end_position_on_the_genomic_accession=149650182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100909384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100909384, CurrentID=0, Status='live', Symbol='DUP16P11.2', Aliases='AUTS14B, C16DUPp11.2', description='Chromosome 16p11.2 duplication syndrome', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614671, no_hdr0=''),
  100909385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100909385, CurrentID=0, Status='live', Symbol='STUT4', Aliases='', description='Stuttering, familial persistent, 4', other_designations='', map_location='16q12.1-q23.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614668, no_hdr0=''),
  100909386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100909386, CurrentID=0, Status='live', Symbol='STUT3', Aliases='', description='Stuttering, familial persistent, 3', other_designations='', map_location='3q13.2-q13.33', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614655, no_hdr0=''),
  115019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115019, CurrentID=0, Status='live', Symbol='SLC26A9', Aliases='', description='solute carrier family 26 (anion exchanger), member 9', other_designations='anion transporter/exchanger protein 9|anion transporter/exchanger-9|solute carrier family 26, member 9', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205913048, end_position_on_the_genomic_accession=205943472, orientation='minus', exon_count=21, OMIM=608481, no_hdr0=''),
  101927650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927650, CurrentID=0, Status='live', Symbol='LOC101927650', Aliases='', description='uncharacterized LOC101927650', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=65141756, end_position_on_the_genomic_accession=65176573, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  115024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115024, CurrentID=0, Status='live', Symbol='NT5C3B', Aliases='NT5C3L', description="5'-nucleotidase, cytosolic IIIB", other_designations="5'-nucleotidase, cytosolic III-like|7-methylguanosine nucleotidase|N(7)-methylguanylate 5'-phosphatase|cN-III-like protein|cytosolic 5'-nucleotidase 3B|cytosolic 5'-nucleotidase III-like protein", map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41825082, end_position_on_the_genomic_accession=41836271, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101927651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927651, CurrentID=0, Status='live', Symbol='LOC101927651', Aliases='', description='uncharacterized LOC101927651', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76622781, end_position_on_the_genomic_accession=76639011, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246100, CurrentID=0, Status='live', Symbol='CTAG1A', Aliases='CT6.1, ESO1, LAGE-2, LAGE2A, NY-ESO-1', description='cancer/testis antigen 1A', other_designations='New York Esophageal Squamous Cell Carcinoma 1|autoimmunogenic cancer/testis antigen NY-ESO-1|cancer/testis antigen 1-A|cancer/testis antigen 6.1|l antigen family member 2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154585154, end_position_on_the_genomic_accession=154586816, orientation='plus', exon_count=3, OMIM=300657, no_hdr0=''),
  100422713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422713, CurrentID=0, Status='live', Symbol='LOC100422713', Aliases='', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 1 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31894094, end_position_on_the_genomic_accession=31895098, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927652, CurrentID=0, Status='live', Symbol='LOC101927652', Aliases='', description='uncharacterized LOC101927652', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26708162, end_position_on_the_genomic_accession=26742116, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  126037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126037, CurrentID=0, Status='live', Symbol='EEF1DP1', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 1', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14070325, end_position_on_the_genomic_accession=14071304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927653, CurrentID=0, Status='live', Symbol='LOC101927653', Aliases='', description='uncharacterized LOC101927653', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58565959, end_position_on_the_genomic_accession=58781716, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  246112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246112, CurrentID=0, Status='live', Symbol='TAF13P1', Aliases='TAF13P', description='TATA-box binding protein associated factor 13 pseudogene 1', other_designations='TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor pseudogene 1|TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 18kDa pseudogene|TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor, pseudogene|TBP-associated factor 13 pseudogene', map_location='6q14.3-q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87425640, end_position_on_the_genomic_accession=87426056, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928473, CurrentID=0, Status='live', Symbol='LOC101928473', Aliases='', description='uncharacterized LOC101928473', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70172151, end_position_on_the_genomic_accession=70175898, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  401002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401002, CurrentID=0, Status='live', Symbol='LOC401002', Aliases='', description='single stranded DNA binding protein 3 pseudogene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131352086, end_position_on_the_genomic_accession=131353239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246119, CurrentID=0, Status='live', Symbol='TTTY10', Aliases='NCRNA00133, TTY10', description='testis-specific transcript, Y-linked 10 (non-protein coding)', other_designations='testis transcript Y 10|transcript Y 10', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20465668, end_position_on_the_genomic_accession=20519228, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246122, CurrentID=0, Status='live', Symbol='TTTY7', Aliases='CLONE795723, LINC00129, NCRNA00129A, TTTY7B, TTY7, TTTY7', description='testis-specific transcript, Y-linked 7 (non-protein coding)', other_designations='Putative transcript Y 7 protein|long intergenic non-protein coding RNA 129|transcript Y 7', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6449468, end_position_on_the_genomic_accession=6457906, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246123, CurrentID=0, Status='live', Symbol='ZNF736P3Y', Aliases='ZNF381P, ZNF736P3, ZNF736PY3', description='zinc finger protein 736 pseudogene 3, Y-linked', other_designations='zinc finger protein 381, Y-linked (pseudogene)|zinc finger protein 736 pseudogene, Y-linked 3', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23050035, end_position_on_the_genomic_accession=23051168, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246126, CurrentID=0, Status='live', Symbol='TXLNGY', Aliases='CYorf15A, CYorf15B, TXLNG2P', description='taxilin gamma pseudogene, Y-linked', other_designations='lipopolysaccaride-specific response 5-like protein|taxilin gamma 2, pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=19567358, end_position_on_the_genomic_accession=19607170, orientation='plus', exon_count=11, OMIM=400031, no_hdr0=''),
  246132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246132, CurrentID=0, Status='live', Symbol='HCHOLA4', Aliases='', description='hypercholesterolemia, autosomal dominant 4', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246135, CurrentID=0, Status='live', Symbol='TAF9BP1', Aliases='', description='TATA-box binding protein associated factor 9b pseudogene 1', other_designations='TAF9B RNA polymerase II, TATA box binding protein (TBP)-associated factor, 31kDa pseudogene 1|TBP-associated factor 9-like pseudogene', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25754893, end_position_on_the_genomic_accession=25757543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152225, CurrentID=0, Status='live', Symbol='LOC152225', Aliases='', description='uncharacterized LOC152225', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101940859, end_position_on_the_genomic_accession=101997926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374440, CurrentID=0, Status='live', Symbol='LOC105374440', Aliases='', description='uncharacterized LOC105374440', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44801712, end_position_on_the_genomic_accession=44805359, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420331, CurrentID=0, Status='live', Symbol='LOC100420331', Aliases='', description='melanoma antigen family A10 pseudogene', other_designations='melanoma antigen family A, 10 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51162893, end_position_on_the_genomic_accession=51163487, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480669, CurrentID=0, Status='live', Symbol='RNU1-140P', Aliases='', description='RNA, U1 small nuclear 140, pseudogene', other_designations='', map_location='5q21.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927661, CurrentID=0, Status='live', Symbol='LOC101927661', Aliases='', description='uncharacterized LOC101927661', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67213408, end_position_on_the_genomic_accession=67244574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379224, CurrentID=0, Status='live', Symbol='LOC105379224', Aliases='', description='uncharacterized LOC105379224', other_designations='CTD-3023L14.2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8560948, end_position_on_the_genomic_accession=8574010, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  115098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115098, CurrentID=0, Status='live', Symbol='CCDC124', Aliases='', description='coiled-coil domain containing 124', other_designations='', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17933015, end_position_on_the_genomic_accession=17943985, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101927663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927663, CurrentID=0, Status='live', Symbol='LOC101927663', Aliases='', description='uncharacterized LOC101927663', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16633331, end_position_on_the_genomic_accession=16644733, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246175, CurrentID=0, Status='live', Symbol='CNOT6L', Aliases='CCR4b', description='CCR4-NOT transcription complex subunit 6-like', other_designations='CCR4-NOT transcription complex, subunit 6-like|carbon catabolite repressor protein 4 homolog B', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77713387, end_position_on_the_genomic_accession=77820238, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  246176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246176, CurrentID=0, Status='live', Symbol='GAS2L2', Aliases='GAR17', description='growth arrest specific 2 like 2', other_designations='GAS2-related protein on chromosome 17|growth arrest-specific 2 like 2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35744511, end_position_on_the_genomic_accession=35753745, orientation='minus', exon_count=6, OMIM=611398, no_hdr0=''),
  101927664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927664, CurrentID=0, Status='live', Symbol='LOC101927664', Aliases='', description='uncharacterized LOC101927664', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112376288, end_position_on_the_genomic_accession=112396404, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  115106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115106, CurrentID=0, Status='live', Symbol='HAUS1', Aliases='CCDC5, HEI-C, HEIC, HsT1461', description='HAUS augmin like complex subunit 1', other_designations='HAUS augmin-like complex, subunit 1|coiled-coil domain containing 5 (spindle associated)|coiled-coil domain-containing protein 5|enhancer of invasion-cluster', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46104332, end_position_on_the_genomic_accession=46128333, orientation='plus', exon_count=9, OMIM=608775, no_hdr0=''),
  246181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246181, CurrentID=0, Status='live', Symbol='AKR7L', Aliases='AFAR3, AFB1-AR3, AKR7A4', description='aldo-keto reductase family 7-like (gene/pseudogene)', other_designations='AFB1 aldehyde reductase 3|aflatoxin B1 aldehyde reductase member 4|aldo-keto reductase family 7 pseudogene|aldoketoreductase 7-like', map_location='1p35-p36.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19265982, end_position_on_the_genomic_accession=19274074, orientation='minus', exon_count=7, OMIM=608478, no_hdr0=''),
  246182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246182, CurrentID=0, Status='live', Symbol='AKR7A2P1', Aliases='AFARP1', description='aldo-keto reductase family 7, member A2 pseudogene 1', other_designations='AKR7 family pseudogene|aldo-keto reductase family 7, member A3 (aflatoxin aldehyde reductase) pseudogene', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112923350, end_position_on_the_genomic_accession=112924673, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  115111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115111, CurrentID=0, Status='live', Symbol='SLC26A7', Aliases='SUT2', description='solute carrier family 26 (anion exchanger), member 7', other_designations='solute carrier family 26, member 7|sulfate anion transporter', map_location='8q23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91209494, end_position_on_the_genomic_accession=91398155, orientation='plus', exon_count=22, OMIM=608479, no_hdr0=''),
  246184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246184, CurrentID=0, Status='live', Symbol='CDC26', Aliases='ANAPC12, APC12, C9orf17', description='cell division cycle 26', other_designations='CDC26 subunit of anaphase promoting complex|anaphase-promoting complex subunit 12|cell division cycle 26 homolog|cell division cycle protein 26 homolog', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113267008, end_position_on_the_genomic_accession=113275589, orientation='minus', exon_count=4, OMIM=614533, no_hdr0=''),
  100147751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100147751, CurrentID=0, Status='live', Symbol='RNU5B-3P', Aliases='', description='RNA, U5B small nuclear 3, pseudogene', other_designations='RNA, U5B small nuclear 3 pseudogene|RNA, U5B small nuclear pseudogene 3', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45034518, end_position_on_the_genomic_accession=45034589, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480670, CurrentID=0, Status='live', Symbol='RNU2-28P', Aliases='', description='RNA, U2 small nuclear 28, pseudogene', other_designations='', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927666, CurrentID=0, Status='live', Symbol='LOC101927666', Aliases='', description='uncharacterized LOC101927666', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58076891, end_position_on_the_genomic_accession=58083204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929732, CurrentID=0, Status='live', Symbol='LOC101929732', Aliases='LA16c-390H2.4', description='uncharacterized LOC101929732', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3542234, end_position_on_the_genomic_accession=3559589, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101059936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059936, CurrentID=0, Status='live', Symbol='LOC101059936', Aliases='', description='uncharacterized LOC101059936', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115123, CurrentID=0, Status='live', Symbol='MARCH3', Aliases='MARCH-III, RNF173', description='membrane associated ring-CH-type finger 3', other_designations='RING finger protein 173|membrane associated ring finger 3|membrane-associated RING-CH protein III|membrane-associated ring finger (C3HC4) 3, E3 ubiquitin protein ligase', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126867714, end_position_on_the_genomic_accession=127032402, orientation='minus', exon_count=11, OMIM=613333, no_hdr0=''),
  105379230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379230, CurrentID=0, Status='live', Symbol='LOC105379230', Aliases='', description='uncharacterized LOC105379230', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9365346, end_position_on_the_genomic_accession=9375166, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104355149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355149, CurrentID=0, Status='live', Symbol='LINC01039', Aliases='', description='long intergenic non-protein coding RNA 1039', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99577112, end_position_on_the_genomic_accession=99585995, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376447, CurrentID=0, Status='live', Symbol='LOC105376447', Aliases='', description='uncharacterized LOC105376447', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22247420, end_position_on_the_genomic_accession=22258542, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100289269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289269, CurrentID=0, Status='live', Symbol='LOC100289269', Aliases='', description='uncharacterized LOC100289269', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=24180452, end_position_on_the_genomic_accession=24231065, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105379232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379232, CurrentID=0, Status='live', Symbol='LOC105379232', Aliases='', description='uncharacterized LOC105379232', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9669449, end_position_on_the_genomic_accession=9672654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246210, CurrentID=0, Status='live', Symbol='IKBKGP1', Aliases='IKBKGP, deltaNEMO', description='inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase gamma pseudogene 1', other_designations='Delta NEMO', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154639391, end_position_on_the_genomic_accession=154648275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378805, CurrentID=0, Status='live', Symbol='LOC105378805', Aliases='', description='uncharacterized LOC105378805', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74465837, end_position_on_the_genomic_accession=74469124, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246213, CurrentID=0, Status='live', Symbol='SLC17A8', Aliases='DFNA25, VGLUT3', description='solute carrier family 17 (vesicular glutamate transporter), member 8', other_designations='solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 8|solute carrier family 17 member 8', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100357079, end_position_on_the_genomic_accession=100422059, orientation='plus', exon_count=12, OMIM=607557, no_hdr0=''),
  105379233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379233, CurrentID=0, Status='live', Symbol='LOC105379233', Aliases='', description='uncharacterized LOC105379233', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10055544, end_position_on_the_genomic_accession=10056299, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480671, CurrentID=0, Status='live', Symbol='RNU2-40P', Aliases='', description='RNA, U2 small nuclear 40, pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  649783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649783, CurrentID=0, Status='live', Symbol='LOC649783', Aliases='', description='HLA class II histocompatibility antigen-like', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927671, CurrentID=0, Status='live', Symbol='MIR202HG', Aliases='', description='MIR202 host gene', other_designations='MIR202 host gene (non-protein coding)', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133246479, end_position_on_the_genomic_accession=133247885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372083, CurrentID=0, Status='live', Symbol='LOC105372083', Aliases='', description='uncharacterized LOC105372083', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41449765, end_position_on_the_genomic_accession=41506866, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130497, CurrentID=0, Status='live', Symbol='LOC100130497', Aliases='', description='dnaJ homolog subfamily B member 6-like', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67278052, end_position_on_the_genomic_accession=67278779, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  104355150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355150, CurrentID=0, Status='live', Symbol='LINC01054', Aliases='', description='long intergenic non-protein coding RNA 1054', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114329757, end_position_on_the_genomic_accession=114333948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246230, CurrentID=0, Status='live', Symbol='CMAL', Aliases='', description='Capillary malformations, hereditary', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=163000, no_hdr0=''),
  101928477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928477, CurrentID=0, Status='live', Symbol='LOC101928477', Aliases='', description='uncharacterized LOC101928477', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102627111, end_position_on_the_genomic_accession=102683913, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  246243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246243, CurrentID=0, Status='live', Symbol='RNASEH1', Aliases='H1RNA, RNH1', description='ribonuclease H1', other_designations='RNase H1|ribonuclease H type II', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3531813, end_position_on_the_genomic_accession=3558371, orientation='minus', exon_count=13, OMIM=604123, no_hdr0=''),
  106480672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480672, CurrentID=0, Status='live', Symbol='RNU2-47P', Aliases='', description='RNA, U2 small nuclear 47, pseudogene', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372084, CurrentID=0, Status='live', Symbol='LOC105372084', Aliases='', description='uncharacterized LOC105372084', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41406799, end_position_on_the_genomic_accession=41421138, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372695, CurrentID=0, Status='live', Symbol='LOC105372695', Aliases='', description='uncharacterized LOC105372695', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58773392, end_position_on_the_genomic_accession=58774811, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246259, CurrentID=0, Status='live', Symbol='BMND3', Aliases='', description='Bone mineral density variability 3', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606928, no_hdr0=''),
  246260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246260, CurrentID=0, Status='live', Symbol='BDA1B', Aliases='', description='Brachydactyly, type A1, locus B', other_designations='', map_location='5p13.3-p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607004, no_hdr0=''),
  101927678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927678, CurrentID=0, Status='live', Symbol='LINC01429', Aliases='', description='long intergenic non-protein coding RNA 1429', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51831797, end_position_on_the_genomic_accession=51862913, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  170543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=170543, CurrentID=0, Status='live', Symbol='RNF11P2', Aliases='RNF11B, bA280O9.1', description='ring finger protein 11 pseudogene 2', other_designations='ring finger protein 11B, pseudogene', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=14547185, end_position_on_the_genomic_accession=14547623, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420351, CurrentID=0, Status='live', Symbol='LOC100420351', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148872255, end_position_on_the_genomic_accession=148872797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115196, CurrentID=0, Status='live', Symbol='ZNF554', Aliases='', description='zinc finger protein 554', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2819874, end_position_on_the_genomic_accession=2836735, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246269, CurrentID=0, Status='live', Symbol='LACE1', Aliases='AFG1', description='lactation elevated 1', other_designations='ATPase family gene 1 homolog|CG8520 gene product|protein AFG1 homolog', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108294894, end_position_on_the_genomic_accession=108525548, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  115201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115201, CurrentID=0, Status='live', Symbol='ATG4A', Aliases='APG4A, AUTL2', description='autophagy related 4A, cysteine peptidase', other_designations='APG4 autophagy 4 homolog A|ATG4 autophagy related 4 homolog A|AUT-like 2 cysteine endopeptidase|AUT-like 2, cysteine endopeptidase|autophagin 2|autophagin-2|autophagy-related cysteine endopeptidase 2|autophagy-related protein 4 homolog A|hAPG4A', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=108088916, end_position_on_the_genomic_accession=108154671, orientation='plus', exon_count=15, OMIM=300663, no_hdr0=''),
  100271477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271477, CurrentID=0, Status='live', Symbol='RPL31P53', Aliases='RPL31_19_1315', description='ribosomal protein L31 pseudogene 53', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57391165, end_position_on_the_genomic_accession=57391518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115207, CurrentID=0, Status='live', Symbol='KCTD12', Aliases='C13orf2, PFET1, PFETIN', description='potassium channel tetramerization domain containing 12', other_designations='potassium channel tetramerisation domain containing 12|predominantly fetal expressed T1 domain', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76880169, end_position_on_the_genomic_accession=76886405, orientation='minus', exon_count=1, OMIM=610521, no_hdr0=''),
  115209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115209, CurrentID=0, Status='live', Symbol='OMA1', Aliases='2010001O09Rik, DAB1, MPRP-1, YKR087C, ZMPOMA1, peptidase', description='OMA1 zinc metallopeptidase', other_designations='OMA1 homolog, zinc metallopeptidase|OMA1 zinc metallopeptidase homolog|metalloprotease related protein 1|metalloprotease-related protein 1|overlapping activity with M-AAA protease|overlapping with the m-AAA protease 1 homolog|zinc metallopeptidase OMA1', map_location='1p32.2-p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58480719, end_position_on_the_genomic_accession=58546799, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  246283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246283, CurrentID=0, Status='live', Symbol='TRBV22OR9-2', Aliases='TCRBV22S2, TCRBV29S2O', description='T cell receptor beta variable 22/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable 22/OR9-2 pseudogene|TRBV22/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33633853, end_position_on_the_genomic_accession=33634302, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246285, CurrentID=0, Status='live', Symbol='TRBV26OR9-2', Aliases='TCRBV26S2, TCRBV28S2, TCRBV28S2O', description='T cell receptor beta variable 26/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable 26/OR9-2 pseudogene|TRBV26/OR9-2', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33695580, end_position_on_the_genomic_accession=33696061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246286, CurrentID=0, Status='live', Symbol='TRBVAOR9-2', Aliases='TCRBV33S2, TRBVA/OR9-2', description='T cell receptor beta variable A/OR9-2 (pseudogene)', other_designations='T cell receptor beta variable A/OR9-2 pseudogene', map_location='9p21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33682270, end_position_on_the_genomic_accession=33682723, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100529855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529855, CurrentID=0, Status='live', Symbol='ZNF625-ZNF20', Aliases='', description='ZNF625-ZNF20 readthrough (NMD candidate)', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12131350, end_position_on_the_genomic_accession=12156731, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  246288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246288, CurrentID=0, Status='live', Symbol='DFNA46', Aliases='', description='deafness, autosomal dominant 46', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246289, CurrentID=0, Status='live', Symbol='DFNA47', Aliases='DFNB83', description='deafness, autosomal dominant 47', other_designations='', map_location='9p22-p21', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608652, no_hdr0=''),
  101927683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927683, CurrentID=0, Status='live', Symbol='LOC101927683', Aliases='', description='uncharacterized LOC101927683', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232727254, end_position_on_the_genomic_accession=232742715, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379248, CurrentID=0, Status='live', Symbol='LOC105379248', Aliases='', description='uncharacterized LOC105379248', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=360087, end_position_on_the_genomic_accession=366034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480674, CurrentID=0, Status='live', Symbol='ENPP7P5', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 5', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  377381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377381, CurrentID=0, Status='live', Symbol='RPL24P4', Aliases='RPL24_1_692, dJ125M3.1', description='ribosomal protein L24 pseudogene 4', other_designations='60S ribosomal protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42956334, end_position_on_the_genomic_accession=42956766, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372086, CurrentID=0, Status='live', Symbol='LOC105372086', Aliases='', description='uncharacterized LOC105372086', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41994945, end_position_on_the_genomic_accession=42004349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246312, CurrentID=0, Status='live', Symbol='C21orf91-OT1', Aliases='D21S2089E, NCRNA00285', description='C21orf91 overlapping transcript 1', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17777404, end_position_on_the_genomic_accession=17792509, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927687, CurrentID=0, Status='live', Symbol='LOC101927687', Aliases='', description='uncharacterized LOC101927687', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199867454, end_position_on_the_genomic_accession=199911171, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  390578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390578, CurrentID=0, Status='live', Symbol='RPLP0P10', Aliases='RPLP0_5_1426', description='ribosomal protein, large, P0 pseudogene 10', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40312284, end_position_on_the_genomic_accession=40313362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246318, CurrentID=0, Status='live', Symbol='FLJ13773', Aliases='', description='uncharacterized LOC246318', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246319, CurrentID=0, Status='live', Symbol='VAMAS6', Aliases='VTLG', description='Vitiligo', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=193200, no_hdr0=''),
  246321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246321, CurrentID=0, Status='live', Symbol='HSCR6', Aliases='HSCRS2', description='Hirschsprung disease, short-segment, 2', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606874, no_hdr0=''),
  246322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246322, CurrentID=0, Status='live', Symbol='HSCR7', Aliases='HSCRS3', description='Hirschsprung disease, short-segment, 3', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606875, no_hdr0=''),
  101927689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927689, CurrentID=0, Status='live', Symbol='LOC101927689', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12246587, end_position_on_the_genomic_accession=12247061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246329, CurrentID=0, Status='live', Symbol='STAC3', Aliases='NAM', description='SH3 and cysteine rich domain 3', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57243453, end_position_on_the_genomic_accession=57251193, orientation='minus', exon_count=12, OMIM=615521, no_hdr0=''),
  246330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246330, CurrentID=0, Status='live', Symbol='PELI3', Aliases='', description='pellino E3 ubiquitin protein ligase family member 3', other_designations='pellino homolog 3|protein pellino homolog 3', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66466327, end_position_on_the_genomic_accession=66477337, orientation='plus', exon_count=10, OMIM=609827, no_hdr0=''),
  105379253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379253, CurrentID=0, Status='live', Symbol='LOC105379253', Aliases='', description='uncharacterized LOC105379253', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65941571, end_position_on_the_genomic_accession=65960384, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  115265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115265, CurrentID=0, Status='live', Symbol='DDIT4L', Aliases='REDD2, Rtp801L', description='DNA damage inducible transcript 4 like', other_designations='DNA-damage-inducible transcript 4-like|HIF-1 responsive protein RTP801-like|REDD-2|homolog of mouse SMHS1|protein regulated in development and DNA damage response 2|regulated in development and DNA damage response 2', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100185110, end_position_on_the_genomic_accession=100190782, orientation='minus', exon_count=4, OMIM=607730, no_hdr0=''),
  101927691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927691, CurrentID=0, Status='live', Symbol='LOC101927691', Aliases='', description='uncharacterized LOC101927691', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=868498, end_position_on_the_genomic_accession=876613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  115273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115273, CurrentID=0, Status='live', Symbol='RAB42', Aliases='', description='RAB42, member RAS oncogene family', other_designations='RAB42, member RAS homolog family', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28592200, end_position_on_the_genomic_accession=28594586, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379255, CurrentID=0, Status='live', Symbol='LOC105379255', Aliases='', description='uncharacterized LOC105379255', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63549217, end_position_on_the_genomic_accession=63560997, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377933, CurrentID=0, Status='live', Symbol='LOC105377933', Aliases='', description='uncharacterized LOC105377933', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109265440, end_position_on_the_genomic_accession=109266428, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927693, CurrentID=0, Status='live', Symbol='LOC101927693', Aliases='', description='uncharacterized LOC101927693', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132002040, end_position_on_the_genomic_accession=132007022, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928481, CurrentID=0, Status='live', Symbol='LOC101928481', Aliases='', description='uncharacterized LOC101928481', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27590227, end_position_on_the_genomic_accession=27716779, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  115286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115286, CurrentID=0, Status='live', Symbol='SLC25A26', Aliases='SAMC', description='solute carrier family 25 (S-adenosylmethionine carrier), member 26', other_designations='mitochondrial S-adenosylmethionine transporter|solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 26', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66133610, end_position_on_the_genomic_accession=66380021, orientation='plus', exon_count=15, OMIM=611037, no_hdr0=''),
  115290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115290, CurrentID=0, Status='live', Symbol='FBXO17', Aliases='FBG4, FBX26, FBXO26, Fbx17', description='F-box protein 17', other_designations='F-box only protein 26|F-box protein FBG4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38941401, end_position_on_the_genomic_accession=38975813, orientation='minus', exon_count=7, OMIM=609094, no_hdr0=''),
  100289295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289295, CurrentID=0, Status='live', Symbol='NUP58P1', Aliases='NUPL1P1', description='nucleoporin 58kDa pseudogene 1', other_designations='nucleoporin like 1 pseudogene 1', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=125681508, end_position_on_the_genomic_accession=125682443, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379258, CurrentID=0, Status='live', Symbol='LOC105379258', Aliases='', description='uncharacterized LOC105379258', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64556010, end_position_on_the_genomic_accession=64573031, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  115294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115294, CurrentID=0, Status='live', Symbol='PCMTD1', Aliases='', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 1', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51817575, end_position_on_the_genomic_accession=51899186, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106480676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480676, CurrentID=0, Status='live', Symbol='KRT8P51', Aliases='', description='keratin 8 pseudogene 51', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927696, CurrentID=0, Status='live', Symbol='LOC101927696', Aliases='', description='uncharacterized LOC101927696', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64248080, end_position_on_the_genomic_accession=64252644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372088, CurrentID=0, Status='live', Symbol='LOC105372088', Aliases='', description='uncharacterized LOC105372088', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=43810688, end_position_on_the_genomic_accession=43933363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420370, CurrentID=0, Status='live', Symbol='TRIM60P7Y', Aliases='TRIM60P7, TRIM60PY7', description='tripartite motif containing 60 pseudogene 7, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 7, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 7', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25009703, end_position_on_the_genomic_accession=25010769, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376453, CurrentID=0, Status='live', Symbol='LOC105376453', Aliases='', description='uncharacterized LOC105376453', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23342233, end_position_on_the_genomic_accession=23359066, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480677, CurrentID=0, Status='live', Symbol='SQSTM1P1', Aliases='', description='sequestosome 1 pseudogene 1', other_designations='', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927701, CurrentID=0, Status='live', Symbol='LOC101927701', Aliases='', description='uncharacterized LOC101927701', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67796054, end_position_on_the_genomic_accession=67825562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372089, CurrentID=0, Status='live', Symbol='LOC105372089', Aliases='', description='uncharacterized LOC105372089', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44658455, end_position_on_the_genomic_accession=44664659, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  115330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115330, CurrentID=0, Status='live', Symbol='GPR146', Aliases='PGR8', description='G protein-coupled receptor 146', other_designations='G-protein coupled receptor PGR8', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1044573, end_position_on_the_genomic_accession=1059269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372696, CurrentID=0, Status='live', Symbol='LOC105372696', Aliases='', description='uncharacterized LOC105372696', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59470083, end_position_on_the_genomic_accession=59471988, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84759, CurrentID=0, Status='live', Symbol='PCGF1', Aliases='2010002K04Rik, NSPC1, RNF3A-2, RNF68', description='polycomb group ring finger 1', other_designations='RING finger protein 68|nervous system Polycomb-1', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74505043, end_position_on_the_genomic_accession=74507694, orientation='minus', exon_count=10, OMIM=610231, no_hdr0=''),
  106481277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481277, CurrentID=0, Status='live', Symbol='RNU6-310P', Aliases='', description='RNA, U6 small nuclear 310, pseudogene', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376454, CurrentID=0, Status='live', Symbol='LOC105376454', Aliases='', description='uncharacterized LOC105376454', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23543920, end_position_on_the_genomic_accession=23586268, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105379267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379267, CurrentID=0, Status='live', Symbol='LOC105379267', Aliases='', description='uncharacterized LOC105379267', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9292384, end_position_on_the_genomic_accession=9304613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  115350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115350, CurrentID=0, Status='live', Symbol='FCRL1', Aliases='CD307a, FCRH1, IFGP1, IRTA5', description='Fc receptor-like 1', other_designations='IFGP family protein 1|fc receptor homolog 1|fcR-like protein 1|hIFGP1|immune receptor translocation-associated protein 5|immunoglobulin superfamily Fc receptor, gp42', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157792249, end_position_on_the_genomic_accession=157820150, orientation='minus', exon_count=12, OMIM=606508, no_hdr0=''),
  100420377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420377, CurrentID=0, Status='live', Symbol='TRIM60P11Y', Aliases='TRIM60P11, TRIM60PY11', description='tripartite motif containing 60 pseudogene 11, Y-linked', other_designations='tripartite motif-containing 60 pseudogene 11, Y-linked|tripartite motif-containing 60 pseudogene, Y-linked 11', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24691782, end_position_on_the_genomic_accession=24692929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  115352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115352, CurrentID=0, Status='live', Symbol='FCRL3', Aliases='CD307c, FCRH3, IFGP3, IRTA3, SPAP2', description='Fc receptor-like 3', other_designations='Fc receptor homolog 3|IFGP family protein 3|SH2 domain-containing phosphatase anchor protein 2|fcR-like protein 3|hIFGP3|immune receptor translocation-associated protein 3|immunoglobulin superfamily receptor translocation associated protein 3', map_location='1q21-q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157676481, end_position_on_the_genomic_accession=157700985, orientation='minus', exon_count=16, OMIM=606510, no_hdr0=''),
  115353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115353, CurrentID=0, Status='live', Symbol='LRRC42', Aliases='dJ167A19.4', description='leucine rich repeat containing 42', other_designations='', map_location='1p33-p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53946326, end_position_on_the_genomic_accession=53968168, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100271478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271478, CurrentID=0, Status='live', Symbol='RPL31P56', Aliases='RPL31_20_1484', description='ribosomal protein L31 pseudogene 56', other_designations='', map_location='16q11.2-q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55352721, end_position_on_the_genomic_accession=55353022, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379269, CurrentID=0, Status='live', Symbol='LOC105379269', Aliases='', description='uncharacterized LOC105379269', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9375131, end_position_on_the_genomic_accession=9376612, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  115361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115361, CurrentID=0, Status='live', Symbol='GBP4', Aliases='Mpa2', description='guanylate binding protein 4', other_designations='GBP-4|GTP-binding protein 4|guanine nucleotide-binding protein 4', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89181148, end_position_on_the_genomic_accession=89198950, orientation='minus', exon_count=11, OMIM=612466, no_hdr0=''),
  115362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115362, CurrentID=0, Status='live', Symbol='GBP5', Aliases='GBP-5', description='guanylate binding protein 5', other_designations='GBP-TA antigen|GTP-binding protein 5|guanine nucleotide-binding protein 5', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89258950, end_position_on_the_genomic_accession=89272861, orientation='minus', exon_count=12, OMIM=611467, no_hdr0=''),
  105379270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379270, CurrentID=0, Status='live', Symbol='LOC105379270', Aliases='', description='uncharacterized LOC105379270', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9394048, end_position_on_the_genomic_accession=9396694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100289308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289308, CurrentID=0, Status='live', Symbol='LOC100289308', Aliases='', description='uncharacterized LOC100289308', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=23456337, end_position_on_the_genomic_accession=23461640, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379271, CurrentID=0, Status='live', Symbol='LOC105379271', Aliases='', description='uncharacterized LOC105379271', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927709, CurrentID=0, Status='live', Symbol='LOC101927709', Aliases='', description='uncharacterized LOC101927709', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118599783, end_position_on_the_genomic_accession=118602966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390600, CurrentID=0, Status='live', Symbol='LOC390600', Aliases='', description='cytokine induced apoptosis inhibitor 1 pseudogene', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68833728, end_position_on_the_genomic_accession=68835360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379273, CurrentID=0, Status='live', Symbol='LOC105379273', Aliases='', description='uncharacterized LOC105379273', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11161845, end_position_on_the_genomic_accession=11214984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480679, CurrentID=0, Status='live', Symbol='SUMO2P16', Aliases='', description='SUMO2 pseudogene 16', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927711, CurrentID=0, Status='live', Symbol='LOC101927711', Aliases='', description='uncharacterized LOC101927711', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232815545, end_position_on_the_genomic_accession=232907461, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101059945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059945, CurrentID=0, Status='live', Symbol='LOC101059945', Aliases='', description='uncharacterized LOC101059945', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115399, CurrentID=0, Status='live', Symbol='LRRC56', Aliases='', description='leucine rich repeat containing 56', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=519000, end_position_on_the_genomic_accession=554916, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  283172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283172, CurrentID=0, Status='live', Symbol='LOC283172', Aliases='', description='calponin 2 pseudogene', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133864550, end_position_on_the_genomic_accession=133865689, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  84771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84771, CurrentID=0, Status='live', Symbol='DDX11L2', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11 like 2', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 like 2|DEAD/H box polypeptide 11 like 2', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113599028, end_position_on_the_genomic_accession=113603717, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480680, CurrentID=0, Status='live', Symbol='SNRPGP18', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 18', other_designations='', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115416, CurrentID=0, Status='live', Symbol='MALSU1', Aliases='C7orf30, mtRsfA', description='mitochondrial assembly of ribosomal large subunit 1', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23299321, end_position_on_the_genomic_accession=23309561, orientation='plus', exon_count=4, OMIM=614624, no_hdr0=''),
  115426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115426, CurrentID=0, Status='live', Symbol='UHRF2', Aliases='NIRF, RNF107, URF2', description='ubiquitin-like with PHD and ring finger domains 2, E3 ubiquitin protein ligase', other_designations='Np95-like ring finger protein|RING finger protein 107|np95/ICBP90-like RING finger protein|nuclear protein 97|nuclear zinc finger protein NP97|ubiquitin-like PHD and RING finger domain-containing protein 2|ubiquitin-like, containing PHD and RING finger domains, 2|ubiquitin-like-containing PHD and RING finger domains protein 2', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6413148, end_position_on_the_genomic_accession=6507056, orientation='plus', exon_count=20, OMIM=615211, no_hdr0=''),
  105379282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379282, CurrentID=0, Status='live', Symbol='LOC105379282', Aliases='', description='glucose-dependent insulinotropic receptor-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226971537, end_position_on_the_genomic_accession=226977442, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100420392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420392, CurrentID=0, Status='live', Symbol='BTBD7P2', Aliases='', description='BTB (POZ) domain containing 7 pseudogene 2', other_designations='', map_location='10q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111322778, end_position_on_the_genomic_accession=111323402, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379283, CurrentID=0, Status='live', Symbol='LOC105379283', Aliases='', description='uncharacterized LOC105379283', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235126919, end_position_on_the_genomic_accession=235128187, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480681, CurrentID=0, Status='live', Symbol='DYNLL1P5', Aliases='', description='dynein, light chain, LC8-type 1 pseudogene 5', other_designations='', map_location='3q25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927721, CurrentID=0, Status='live', Symbol='LOC101927721', Aliases='', description='uncharacterized LOC101927721', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101299613, end_position_on_the_genomic_accession=101301344, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379284, CurrentID=0, Status='live', Symbol='LOC105379284', Aliases='', description='uncharacterized LOC105379284', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245671231, end_position_on_the_genomic_accession=245678589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927723, CurrentID=0, Status='live', Symbol='LOC101927723', Aliases='', description='uncharacterized LOC101927723', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68361214, end_position_on_the_genomic_accession=68370505, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390614, CurrentID=0, Status='live', Symbol='LOC390614', Aliases='', description='relaxin/insulin-like family peptide receptor 3 pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78831041, end_position_on_the_genomic_accession=78832030, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376458, CurrentID=0, Status='live', Symbol='LOC105376458', Aliases='', description='uncharacterized LOC105376458', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26192001, end_position_on_the_genomic_accession=26194622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928487, CurrentID=0, Status='live', Symbol='LOC101928487', Aliases='', description='uncharacterized LOC101928487', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218216103, end_position_on_the_genomic_accession=218217117, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379288, CurrentID=0, Status='live', Symbol='LOC105379288', Aliases='', description='uncharacterized LOC105379288', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12417985, end_position_on_the_genomic_accession=12428636, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480682, CurrentID=0, Status='live', Symbol='GTF3AP2', Aliases='', description='general transcription factor IIIA pseudogene 2', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379289, CurrentID=0, Status='live', Symbol='LOC105379289', Aliases='', description='uncharacterized LOC105379289', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12865466, end_position_on_the_genomic_accession=12890736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372697, CurrentID=0, Status='live', Symbol='LOC105372697', Aliases='', description='uncharacterized LOC105372697', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60332374, end_position_on_the_genomic_accession=60335793, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390618, CurrentID=0, Status='live', Symbol='LOC390618', Aliases='', description='ribosomal protein L7 like 1 pseudogene', other_designations='ribosomal protein L7-like 1 pseudogene', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79668345, end_position_on_the_genomic_accession=79669073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  377630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377630, CurrentID=0, Status='live', Symbol='USP17L2', Aliases='DUB-3, DUB3, USP17', description='ubiquitin specific peptidase 17-like family member 2', other_designations='deubiquitinating enzyme 17-like protein 2|deubiquitinating enzyme 3|deubiquitinating protein 3|ubiquitin carboxyl-terminal hydrolase 17-like protein 2|ubiquitin specific peptidase 17-like 2|ubiquitin thioesterase 17-like protein 2|ubiquitin thiolesterase 17-like protein 2|ubiquitin-specific-processing protease 17-like protein 2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12137168, end_position_on_the_genomic_accession=12138760, orientation='minus', exon_count=1, OMIM=610186, no_hdr0=''),
  353007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353007, CurrentID=0, Status='live', Symbol='3.8-1.2', Aliases='', description='HLA complex group 26 (non-protein coding) pseudogene', other_designations='3.8-1.2 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29995956, end_position_on_the_genomic_accession=29996411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106481278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481278, CurrentID=0, Status='live', Symbol='RNU6-314P', Aliases='', description='RNA, U6 small nuclear 314, pseudogene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369196, CurrentID=0, Status='live', Symbol='LOC105369196', Aliases='', description='uncharacterized LOC105369196', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74485674, end_position_on_the_genomic_accession=74491505, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100158257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100158257, CurrentID=0, Status='live', Symbol='NF1P7', Aliases='NF1L7', description='neurofibromin 1 pseudogene 7', other_designations='neurofibromin 1-like 7', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18898416, end_position_on_the_genomic_accession=18909674, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379292, CurrentID=0, Status='live', Symbol='LOC105379292', Aliases='', description='uncharacterized LOC105379292', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13833035, end_position_on_the_genomic_accession=13849100, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100158258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100158258, CurrentID=0, Status='live', Symbol='NEK2P3', Aliases='', description='NEK2 pseudogene 3', other_designations='NIMA (never in mitosis gene a)-related kinase 2 pseudogene', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18698004, end_position_on_the_genomic_accession=18700120, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373570, CurrentID=0, Status='live', Symbol='LOC105373570', Aliases='', description='uncharacterized LOC105373570', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113803963, end_position_on_the_genomic_accession=113818948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480683, CurrentID=0, Status='live', Symbol='AARSP1', Aliases='', description='alanyl-tRNA synthetase pseudogene 1', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115509, CurrentID=0, Status='live', Symbol='ZNF689', Aliases='TIPUH1', description='zinc finger protein 689', other_designations='transcription-involved protein upregulated in HCC 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30602558, end_position_on_the_genomic_accession=30610775, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100289333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289333, CurrentID=0, Status='live', Symbol='LOC100289333', Aliases='', description='uncharacterized LOC100289333', other_designations='CTD-2666L21.1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12195015, end_position_on_the_genomic_accession=12207574, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645431, CurrentID=0, Status='live', Symbol='FUT8-AS1', Aliases='', description='FUT8 antisense RNA 1', other_designations='FUT8 antisense RNA 1 (non-protein coding)', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65410592, end_position_on_the_genomic_accession=65412617, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100420406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420406, CurrentID=0, Status='live', Symbol='IRF5P1', Aliases='', description='interferon regulatory factor 5 pseudogene 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92972596, end_position_on_the_genomic_accession=92974741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379297, CurrentID=0, Status='live', Symbol='LOC105379297', Aliases='', description='uncharacterized LOC105379297', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16783215, end_position_on_the_genomic_accession=16916028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927735, CurrentID=0, Status='live', Symbol='LOC101927735', Aliases='', description='uncharacterized LOC101927735', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129208601, end_position_on_the_genomic_accession=129212662, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  377677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377677, CurrentID=0, Status='live', Symbol='CA13', Aliases='CAXIII', description='carbonic anhydrase XIII', other_designations='CA-XIII|carbonate dehydratase XIII', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85245487, end_position_on_the_genomic_accession=85284073, orientation='plus', exon_count=7, OMIM=611436, no_hdr0=''),
  106480684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480684, CurrentID=0, Status='live', Symbol='MTCO1P4', Aliases='', description='MT-CO1 pseudogene 4', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927736, CurrentID=0, Status='live', Symbol='LOC101927736', Aliases='', description='uncharacterized LOC101927736', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=233200428, end_position_on_the_genomic_accession=233206506, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379299, CurrentID=0, Status='live', Symbol='LOC105379299', Aliases='', description='uncharacterized LOC105379299', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17899858, end_position_on_the_genomic_accession=17908011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927737, CurrentID=0, Status='live', Symbol='LOC101927737', Aliases='', description='uncharacterized LOC101927737', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41958346, end_position_on_the_genomic_accession=41961535, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  353009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=353009, CurrentID=0, Status='live', Symbol='3.8-1.4', Aliases='', description='HLA complex group 26 (non-protein coding) pseudogene', other_designations='3.8-1.4 pseudogene', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29865915, end_position_on_the_genomic_accession=29867087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115548, CurrentID=0, Status='live', Symbol='FCHO2', Aliases='', description='FCH domain only 2', other_designations='FCH domain only protein 2', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72955981, end_position_on_the_genomic_accession=73090522, orientation='plus', exon_count=26, OMIM=613438, no_hdr0=''),
  115557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115557, CurrentID=0, Status='live', Symbol='ARHGEF25', Aliases='GEFT, p63RhoGEF', description='Rho guanine nucleotide exchange factor 25', other_designations='RAC/CDC42 exchange factor|Rho guanine nucleotide exchange factor (GEF) 25|RhoA/RAC/CDC42 exchange factor|guanine nucleotide exchange factor GEFT|rac/Cdc42/Rho exchange factor GEFT|rhoA/Rac/Cdc42 guanine nucleotide exchange factor GEFT', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57610180, end_position_on_the_genomic_accession=57617245, orientation='plus', exon_count=18, OMIM=610215, no_hdr0=''),
  115560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115560, CurrentID=0, Status='live', Symbol='ZNF501', Aliases='ZNF, ZNF52', description='zinc finger protein 501', other_designations='zinc finger protein 52', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44729606, end_position_on_the_genomic_accession=44737083, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480685, CurrentID=0, Status='live', Symbol='LLPHP1', Aliases='', description='LLP homolog, long-term synaptic facilitation (Aplysia) pseudogene 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420413, CurrentID=0, Status='live', Symbol='LOC100420413', Aliases='', description='G protein-coupled receptor 141 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37796193, end_position_on_the_genomic_accession=37796876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379304, CurrentID=0, Status='live', Symbol='LOC105379304', Aliases='', description='uncharacterized LOC105379304', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18854428, end_position_on_the_genomic_accession=18855280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  115572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115572, CurrentID=0, Status='live', Symbol='FAM46B', Aliases='', description='family with sequence similarity 46 member B', other_designations='family with sequence similarity 46, member B', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27005020, end_position_on_the_genomic_accession=27012842, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369206, CurrentID=0, Status='live', Symbol='LOC105369206', Aliases='', description='uncharacterized LOC105369206', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66464653, end_position_on_the_genomic_accession=66466706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379306, CurrentID=0, Status='live', Symbol='LOC105379306', Aliases='', description='uncharacterized LOC105379306', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19300523, end_position_on_the_genomic_accession=19311856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  115584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115584, CurrentID=0, Status='live', Symbol='SLC5A11', Aliases='KST1, RKST1, SGLT6, SMIT2', description='solute carrier family 5 (sodium/inositol cotransporter), member 11', other_designations='Na(+)/myo-inositol cotransporter 2|homolog of rabbit KST1|putative sodium-coupled cotransporter RKST1|sodium-dependent glucose cotransporter|sodium/glucose cotransporter KST1|sodium/myo-inositol transporter 2|solute carrier family 5 (sodium/glucose cotransporter), member 11', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24845863, end_position_on_the_genomic_accession=24911628, orientation='plus', exon_count=19, OMIM=610238, no_hdr0=''),
  105378808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378808, CurrentID=0, Status='live', Symbol='LOC105378808', Aliases='', description='uncharacterized LOC105378808', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77219567, end_position_on_the_genomic_accession=77231670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422491, CurrentID=0, Status='live', Symbol='LOC100422491', Aliases='', description='amyloid beta (A4) precursor protein-binding, family A, member 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34409648, end_position_on_the_genomic_accession=34410609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480686, CurrentID=0, Status='live', Symbol='RNY3P16', Aliases='', description='RNA, Ro-associated Y3 pseudogene 16', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379309, CurrentID=0, Status='live', Symbol='LOC105379309', Aliases='', description='uncharacterized LOC105379309', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19886234, end_position_on_the_genomic_accession=19922836, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379310, CurrentID=0, Status='live', Symbol='LOC105379310', Aliases='', description='uncharacterized LOC105379310', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20106548, end_position_on_the_genomic_accession=20109437, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927748, CurrentID=0, Status='live', Symbol='LOC101927748', Aliases='', description='uncharacterized LOC101927748', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61213965, end_position_on_the_genomic_accession=61220196, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645434, CurrentID=0, Status='live', Symbol='LINC01625', Aliases='', description='long intergenic non-protein coding RNA 1625', other_designations='hCG1786283', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139468995, end_position_on_the_genomic_accession=139474596, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  153527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=153527, CurrentID=0, Status='live', Symbol='ZMAT2', Aliases='Ptg-12, Snu23', description='zinc finger, matrin-type 2', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140698743, end_position_on_the_genomic_accession=140706663, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927750, CurrentID=0, Status='live', Symbol='ATP6V1B1-AS1', Aliases='', description='ATP6V1B1 antisense RNA 1', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70942013, end_position_on_the_genomic_accession=70948611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379313, CurrentID=0, Status='live', Symbol='LOC105379313', Aliases='', description='uncharacterized LOC105379313', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20225365, end_position_on_the_genomic_accession=20226480, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480687, CurrentID=0, Status='live', Symbol='C8orf59P2', Aliases='', description='chromosome 8 open reading frame 59 pseudogene 2', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927751, CurrentID=0, Status='live', Symbol='LOC101927751', Aliases='', description='E3 ubiquitin-protein ligase RING1-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100917250, end_position_on_the_genomic_accession=100918889, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372099, CurrentID=0, Status='live', Symbol='LOC105372099', Aliases='', description='uncharacterized LOC105372099', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246704, CurrentID=0, Status='live', Symbol='LINC00315', Aliases='C21orf93, NCRNA00315', description='long intergenic non-protein coding RNA 315', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.8', start_position_on_the_genomic_accession=46723031, end_position_on_the_genomic_accession=46724422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246705, CurrentID=0, Status='live', Symbol='LINC00314', Aliases='C21orf94, NCRNA00314', description='long intergenic non-protein coding RNA 314', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28013363, end_position_on_the_genomic_accession=28023209, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379315, CurrentID=0, Status='live', Symbol='LOC105379315', Aliases='', description='uncharacterized LOC105379315', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20915985, end_position_on_the_genomic_accession=20943031, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379316, CurrentID=0, Status='live', Symbol='LOC105379316', Aliases='', description='uncharacterized LOC105379316', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20952644, end_position_on_the_genomic_accession=20955312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246714, CurrentID=0, Status='live', Symbol='CKS1BP2', Aliases='bA192N10.2', description='CDC28 protein kinase regulatory subunit 1B pseudogene 2', other_designations='cell division control protein CKS1 pseudogene', map_location='10q11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29697935, end_position_on_the_genomic_accession=29698653, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246715, CurrentID=0, Status='live', Symbol='CKS1BP3', Aliases='', description='CDC28 protein kinase regulatory subunit 1B pseudogene 3', other_designations='cell division control protein CKS1 pseudogene', map_location='5q12.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62511761, end_position_on_the_genomic_accession=62512482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246716, CurrentID=0, Status='live', Symbol='YWHAQP2', Aliases='', description='YWHAQ pseudogene 2', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43644842, end_position_on_the_genomic_accession=43645876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  246717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246717, CurrentID=0, Status='live', Symbol='TCEB2P1', Aliases='', description='transcription elongation factor B (SIII), polypeptide 2 (18kDa, elongin B) pseudogene 1', other_designations='transcription elongation factor B (SIII), polypeptide 2 (18kD, elongin B) pseudogene', map_location='13q12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25365803, end_position_on_the_genomic_accession=25366199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246719, CurrentID=0, Status='live', Symbol='UBE2SP1', Aliases='', description='ubiquitin conjugating enzyme E2S pseudogene 1', other_designations='ubiquitin carrier protein E2-EPF pseudogene|ubiquitin-conjugating enzyme E2S pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15704079, end_position_on_the_genomic_accession=15705093, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246720, CurrentID=0, Status='live', Symbol='DYNLL1P1', Aliases='DNCL1P1', description='dynein, light chain, LC8-type 1 pseudogene 1', other_designations='dynein, cytoplasmic, light polypeptide 1 pseudogene 1|dynein, cytoplasmic, light polypeptide pseudogene', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81246010, end_position_on_the_genomic_accession=81246671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246721, CurrentID=0, Status='live', Symbol='POLR2J2', Aliases='HRPB11B, RPB11b1', description='polymerase (RNA) II (DNA directed) polypeptide J2', other_designations='DNA directed RNA polymerase II polypeptide J-related|DNA-directed RNA polymerase II subunit J2|RNA polymerase II subunit B11-b1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102636748, end_position_on_the_genomic_accession=102671729, orientation='minus', exon_count=9, OMIM=609881, no_hdr0=''),
  115650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115650, CurrentID=0, Status='live', Symbol='TNFRSF13C', Aliases='BAFF-R, BAFFR, BROMIX, CD268, CVID4, prolixin', description='tumor necrosis factor receptor superfamily member 13C', other_designations='B cell-activating factor receptor|B-cell-activating factor receptor|BAFF receptor|BLyS receptor 3|tumor necrosis factor receptor superfamily, member 13C', map_location='22q13.1-q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41925032, end_position_on_the_genomic_accession=41926817, orientation='minus', exon_count=3, OMIM=606269, no_hdr0=''),
  246723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246723, CurrentID=0, Status='live', Symbol='ATP5HP4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit D pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d, pseudogene 4', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68642443, end_position_on_the_genomic_accession=68643038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246724, CurrentID=0, Status='live', Symbol='LOC246724', Aliases='', description='DNA directed RNA polymerase II polypeptide J-related gene', other_designations='', map_location='7p15', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  115653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115653, CurrentID=0, Status='live', Symbol='KIR3DL3', Aliases='CD158Z, KIR2DS2, KIR3DL7, KIR44, KIRC1', description='killer cell immunoglobulin-like receptor, three domains, long cytoplasmic tail, 3', other_designations='CD158 antigen-like family member Z|killer cell Ig-like receptor KIR3DL7|killer cell immunoglobulin-like receptor two domains short cytoplasmic tail 2|killer cell inhibitory receptor 1', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54724447, end_position_on_the_genomic_accession=54736536, orientation='plus', exon_count=8, OMIM=610095, no_hdr0=''),
  100271480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271480, CurrentID=0, Status='live', Symbol='RPL31P59', Aliases='RPL31_23_1580', description='ribosomal protein L31 pseudogene 59', other_designations='', map_location='18p11.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3377926, end_position_on_the_genomic_accession=3378247, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927756, CurrentID=0, Status='live', Symbol='LOC101927756', Aliases='', description='uncharacterized LOC101927756', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61031227, end_position_on_the_genomic_accession=61049674, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246731, CurrentID=0, Status='live', Symbol='PPIHP1', Aliases='', description='peptidylprolyl isomerase H (cyclophilin H) pseudogene 1', other_designations='', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112029662, end_position_on_the_genomic_accession=112030394, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059954, CurrentID=0, Status='live', Symbol='LOC101059954', Aliases='', description='uncharacterized LOC101059954', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40523196, end_position_on_the_genomic_accession=40527002, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246734, CurrentID=0, Status='live', Symbol='NPCDR1', Aliases='NPCDRG, NPCR', description='nasopharyngeal carcinoma, down-regulated 1', other_designations='nasopharyngeal carcinoma-related protein', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=59957207, end_position_on_the_genomic_accession=59957535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246737, CurrentID=0, Status='live', Symbol='YWHAZP4', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta pseudogene 4', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', map_location='6q23.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127355742, end_position_on_the_genomic_accession=127358855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374444, CurrentID=0, Status='live', Symbol='LOC105374444', Aliases='', description='uncharacterized LOC105374444', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47627314, end_position_on_the_genomic_accession=47636543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420430, CurrentID=0, Status='live', Symbol='LOC100420430', Aliases='', description='glioblastoma amplified sequence pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146833501, end_position_on_the_genomic_accession=146834038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379321, CurrentID=0, Status='live', Symbol='LOC105379321', Aliases='', description='uncharacterized LOC105379321', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22695128, end_position_on_the_genomic_accession=22696326, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246744, CurrentID=0, Status='live', Symbol='STH', Aliases='MAPTIT', description='saitohin', other_designations='microtubule-associated protein tau (MAPT) intronic transcript', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45999250, end_position_on_the_genomic_accession=45999694, orientation='plus', exon_count=1, OMIM=607067, no_hdr0=''),
  101927759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927759, CurrentID=0, Status='live', Symbol='LOC101927759', Aliases='', description='uncharacterized LOC101927759', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3025694, end_position_on_the_genomic_accession=3027434, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  115677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115677, CurrentID=0, Status='live', Symbol='NOSTRIN', Aliases='DaIP2', description='nitric oxide synthase trafficking', other_designations='BM247 homolog|ENOS traffic inducer|eNOS-trafficking inducer|endothelial nitric oxide synthase traffic inducer|nitric oxide synthase traffic inducer|nitric oxide synthase trafficker|ortholog of mouse disabled 2 interacting proein 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168786539, end_position_on_the_genomic_accession=168865339, orientation='plus', exon_count=25, OMIM=607496, no_hdr0=''),
  246750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246750, CurrentID=0, Status='live', Symbol='MYAS1', Aliases='', description='Myasthenia gravis with thymus hyperplasia', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607085, no_hdr0=''),
  246751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246751, CurrentID=0, Status='live', Symbol='SMAR', Aliases='', description='Spinal muscular atrophy, chronic distal, autosomal recessive', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607088, no_hdr0=''),
  101927760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927760, CurrentID=0, Status='live', Symbol='LOC101927760', Aliases='', description='uncharacterized LOC101927760', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118241501, end_position_on_the_genomic_accession=118247881, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246754, CurrentID=0, Status='live', Symbol='MTVR2', Aliases='', description='mouse mammary tumor virus receptor homolog 2', other_designations='family with sequence similarity 89, member B pseudogene', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56884102, end_position_on_the_genomic_accession=56884913, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379323, CurrentID=0, Status='live', Symbol='LOC105379323', Aliases='', description='uncharacterized LOC105379323', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22898420, end_position_on_the_genomic_accession=22902058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480689, CurrentID=0, Status='live', Symbol='SAP18P1', Aliases='', description='SAP18 pseudogene 1', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927761, CurrentID=0, Status='live', Symbol='TH2LCRR', Aliases='TH2-LCR', description='T helper type 2 locus control region associated RNA', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132638158, end_position_on_the_genomic_accession=132642171, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379324, CurrentID=0, Status='live', Symbol='LOC105379324', Aliases='', description='uncharacterized LOC105379324', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22842164, end_position_on_the_genomic_accession=22846765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646797, CurrentID=0, Status='live', Symbol='LOC646797', Aliases='', description='glutamyl aminopeptidase (aminopeptidase A) pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50169808, end_position_on_the_genomic_accession=50190433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420434, CurrentID=0, Status='live', Symbol='STIP1P1', Aliases='', description='stress induced phosphoprotein 1 pseudogene 1', other_designations='stress-induced phosphoprotein 1 pseudogene 1|stress-induced-phosphoprotein 1 pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142915569, end_position_on_the_genomic_accession=142917431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374871, CurrentID=0, Status='live', Symbol='LOC105374871', Aliases='', description='uncharacterized LOC105374871', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=329769, end_position_on_the_genomic_accession=335021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  377841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377841, CurrentID=0, Status='live', Symbol='ENTPD8', Aliases='E-NTPDase, GLSR2492, NTPDase-8, UNQ2492', description='ectonucleoside triphosphate diphosphohydrolase 8', other_designations='E-NTPDase 8|liver ecto-ATP-diphosphohydrolase', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137434364, end_position_on_the_genomic_accession=137441662, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  115701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115701, CurrentID=0, Status='live', Symbol='ALPK2', Aliases='HAK', description='alpha kinase 2', other_designations='alpha-kinase 2|heart alpha-kinase|heart alpha-protein kinase', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58481247, end_position_on_the_genomic_accession=58629188, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  115703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115703, CurrentID=0, Status='live', Symbol='ARHGAP33', Aliases='NOMA-GAP, SNX26, TCGAP', description='Rho GTPase activating protein 33', other_designations='neurite outgrowth multiadaptor RhoGAP protein|rho-type GTPase-activating protein 33|sorting nexin 26|sorting nexin-26|tc10/CDC42 GTPase-activating protein', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35775515, end_position_on_the_genomic_accession=35788822, orientation='plus', exon_count=22, OMIM=614902, no_hdr0=''),
  115704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115704, CurrentID=0, Status='live', Symbol='EVI5L', Aliases='', description='ecotropic viral integration site 5-like', other_designations='ecotropic viral integration site 5-like protein', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7830273, end_position_on_the_genomic_accession=7864976, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  246777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246777, CurrentID=0, Status='live', Symbol='SPESP1', Aliases='ESP, SP-ESP', description='sperm equatorial segment protein 1', other_designations='equatorial segment protein|glycosylated 38 kDa sperm protein C-7/8', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68930500, end_position_on_the_genomic_accession=68946811, orientation='plus', exon_count=2, OMIM=609399, no_hdr0=''),
  246778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246778, CurrentID=0, Status='live', Symbol='IL27', Aliases='IL-27, IL-27AA, IL27p28, IL30, p28, IL27', description='interleukin 27', other_designations='IL-27 p28 subunit|IL-27 subunit alpha|IL-27-A|IL27-A|interleukin 30|interleukin-30', map_location='16p11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28499069, end_position_on_the_genomic_accession=28526399, orientation='minus', exon_count=6, OMIM=608273, no_hdr0=''),
  115708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115708, CurrentID=0, Status='live', Symbol='TRMT61A', Aliases='C14orf172, GCD14, Gcd14p, TRM61, hTRM61', description='tRNA methyltransferase 61A', other_designations='tRNA (adenine-N(1)-)-methyltransferase catalytic subunit TRM61|tRNA (adenine-N(1)-)-methyltransferase catalytic subunit TRMT61A|tRNA methyltransferase 61 homolog A|tRNA(m1A58)-methyltransferase subunit TRM61|tRNA(m1A58)-methyltransferase subunit TRMT61A|tRNA(m1A58)MTase subunit TRM61|tRNA(m1A58)MTase subunit TRMT61A', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103529172, end_position_on_the_genomic_accession=103537073, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246784, CurrentID=0, Status='live', Symbol='LOC246784', Aliases='', description='homolog of C. elegans smu-1 pseudogene', other_designations='smu-1 suppressor of mec-8 and unc-52 homolog (C. elegans) pseudogene', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157058910, end_position_on_the_genomic_accession=157060804, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  246785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=246785, CurrentID=0, Status='live', Symbol='OLA1P1', Aliases='', description='Obg-like ATPase 1 pseudogene 1', other_designations='PTD004 pseudogene', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42107350, end_position_on_the_genomic_accession=42109109, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480690, CurrentID=0, Status='live', Symbol='RPL23AP90', Aliases='', description='ribosomal protein L23a pseudogene 90', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369207, CurrentID=0, Status='live', Symbol='LOC105369207', Aliases='', description='uncharacterized LOC105369207', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57193537, end_position_on_the_genomic_accession=57201049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  115727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115727, CurrentID=0, Status='live', Symbol='RASGRP4', Aliases='', description='RAS guanyl releasing protein 4', other_designations='guanyl nucleotide releasing protein 4', map_location='19q13.1', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38409058, end_position_on_the_genomic_accession=38426305, orientation='minus', exon_count=17, OMIM=607320, no_hdr0=''),
  101927768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927768, CurrentID=0, Status='live', Symbol='LOC101927768', Aliases='', description='uncharacterized LOC101927768', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113969701, end_position_on_the_genomic_accession=114340729, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105379332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379332, CurrentID=0, Status='live', Symbol='LOC105379332', Aliases='', description='uncharacterized LOC105379332', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25685826, end_position_on_the_genomic_accession=25688759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378809, CurrentID=0, Status='live', Symbol='LOC105378809', Aliases='', description='uncharacterized LOC105378809', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77432955, end_position_on_the_genomic_accession=77437946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927770, CurrentID=0, Status='live', Symbol='LOC101927770', Aliases='', description='uncharacterized LOC101927770', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53552770, end_position_on_the_genomic_accession=53575308, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106480691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480691, CurrentID=0, Status='live', Symbol='DNAJC19P4', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 4', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 4', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130314, CurrentID=0, Status='live', Symbol='LOC100130314', Aliases='', description='uncharacterized LOC100130314', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=47933131, end_position_on_the_genomic_accession=47933496, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  115749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115749, CurrentID=0, Status='live', Symbol='C12orf56', Aliases='', description='chromosome 12 open reading frame 56', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64264762, end_position_on_the_genomic_accession=64390584, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  115752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115752, CurrentID=0, Status='live', Symbol='DIS3L', Aliases='DIS3L1', description="DIS3 like exosome 3'-5' exoribonuclease", other_designations='DIS3 mitotic control homolog-like', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66293256, end_position_on_the_genomic_accession=66333898, orientation='plus', exon_count=20, OMIM=614183, no_hdr0=''),
  100289373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289373, CurrentID=0, Status='live', Symbol='UBAC2-AS1', Aliases='', description='UBAC2 antisense RNA 1', other_designations='UBAC2 antisense RNA 1 (non-protein coding)', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99196374, end_position_on_the_genomic_accession=99200757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  115761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115761, CurrentID=0, Status='live', Symbol='ARL11', Aliases='ARLTS1', description='ADP ribosylation factor like GTPase 11', other_designations='ADP-ribosylation factor like GTPase 11|ADP-ribosylation factor-like 11|ADP-ribosylation factor-like tumor suppressor protein 1', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49628299, end_position_on_the_genomic_accession=49633872, orientation='plus', exon_count=3, OMIM=609351, no_hdr0=''),
  101927774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927774, CurrentID=0, Status='live', Symbol='LINC00824', Aliases='LINC01263', description='long intergenic non-protein coding RNA 824', other_designations='long intergenic non-protein coding RNA 1263', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128405269, end_position_on_the_genomic_accession=128564679, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  106480692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480692, CurrentID=0, Status='live', Symbol='RN7SKP186', Aliases='', description='RNA, 7SK small nuclear pseudogene 186', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379339, CurrentID=0, Status='live', Symbol='LOC105379339', Aliases='', description='uncharacterized LOC105379339', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26963663, end_position_on_the_genomic_accession=26974730, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101059958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059958, CurrentID=0, Status='live', Symbol='LOC101059958', Aliases='', description='collagen alpha-1(II) chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=69432331, end_position_on_the_genomic_accession=69437606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927777, CurrentID=0, Status='live', Symbol='SATB1-AS1', Aliases='', description='SATB1 antisense RNA 1', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18445237, end_position_on_the_genomic_accession=18530114, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372699, CurrentID=0, Status='live', Symbol='LOC105372699', Aliases='', description='uncharacterized LOC105372699', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=60748290, end_position_on_the_genomic_accession=60777746, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  115795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115795, CurrentID=0, Status='live', Symbol='FANCD2OS', Aliases='C3orf24', description='FANCD2 opposite strand', other_designations='Fanconi anemia group D2 protein opposite strand transcript protein', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10081320, end_position_on_the_genomic_accession=10108339, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100289380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289380, CurrentID=0, Status='live', Symbol='PTP4A1P5', Aliases='', description='protein tyrosine phosphatase type IVA, member 1 pseudogene 5', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27427721, end_position_on_the_genomic_accession=27428231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379343, CurrentID=0, Status='live', Symbol='LOC105379343', Aliases='', description='uncharacterized LOC105379343', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28177505, end_position_on_the_genomic_accession=28182479, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130316, CurrentID=0, Status='live', Symbol='LOC100130316', Aliases='', description='putative RNA-binding protein 15', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127170436, end_position_on_the_genomic_accession=127179847, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115811, CurrentID=0, Status='live', Symbol='IQCD', Aliases='4933433C09Rik, CFAP84, DRC10', description='IQ motif containing D', other_designations='dynein regulatory complex subunit 10', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113195441, end_position_on_the_genomic_accession=113221094, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  115817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115817, CurrentID=0, Status='live', Symbol='DHRS1', Aliases='SDR19C1', description='dehydrogenase/reductase (SDR family) member 1', other_designations='short chain dehydrogenase/reductase family 19C member 1|short chain dehydrogenase/reductase family 19C, member 1', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24290598, end_position_on_the_genomic_accession=24299833, orientation='minus', exon_count=10, OMIM=610410, no_hdr0=''),
  645441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645441, CurrentID=0, Status='live', Symbol='RPL17P6', Aliases='RPL17_1_64', description='ribosomal protein L17 pseudogene 6', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=77129074, end_position_on_the_genomic_accession=77129703, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  115825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115825, CurrentID=0, Status='live', Symbol='WDFY2', Aliases='PROF, WDF2, ZFYVE22', description='WD repeat and FYVE domain containing 2', other_designations='WD40 and FYVE domain containing 2|WD40- and FYVE domain-containing protein 2|propeller-FYVE protein|zinc finger FYVE domain-containing protein 22', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51584194, end_position_on_the_genomic_accession=51766799, orientation='plus', exon_count=16, OMIM=610418, no_hdr0=''),
  115827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115827, CurrentID=0, Status='live', Symbol='RAB3C', Aliases='', description='RAB3C, member RAS oncogene family', other_designations='', map_location='5q13', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58582127, end_position_on_the_genomic_accession=58859394, orientation='plus', exon_count=7, OMIM=612829, no_hdr0=''),
  101927785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927785, CurrentID=0, Status='live', Symbol='LOC101927785', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89471097, end_position_on_the_genomic_accession=89472879, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480694, CurrentID=0, Status='live', Symbol='RN7SL369P', Aliases='', description='RNA, 7SL, cytoplasmic 369, pseudogene', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420458, CurrentID=0, Status='live', Symbol='TATDN2P3', Aliases='', description='TatD DNase domain containing 2 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23244907, end_position_on_the_genomic_accession=23247181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  84843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84843, CurrentID=0, Status='live', Symbol='LOC84843', Aliases='', description='uncharacterized LOC84843', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.10', start_position_on_the_genomic_accession=31725010, end_position_on_the_genomic_accession=31726174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927788, CurrentID=0, Status='live', Symbol='LOC101927788', Aliases='', description='uncharacterized LOC101927788', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30403805, end_position_on_the_genomic_accession=30407268, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  377997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=377997, CurrentID=0, Status='live', Symbol='TSPY12P', Aliases='TSPYP2', description='testis specific protein, Y-linked 12, pseudogene', other_designations='testis specific protein, Y-linked 1 pseudogene|testis specific protein, Y-linked pseudogene 2', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7686801, end_position_on_the_genomic_accession=7689593, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378001, CurrentID=0, Status='live', Symbol='TSPY13P', Aliases='TSPYP3', description='testis specific protein, Y-linked 13, pseudogene', other_designations='testis specific protein, Y-linked 1 pseudogene|testis specific protein, Y-linked pseudogene 3', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9905457, end_position_on_the_genomic_accession=9908185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115861, CurrentID=0, Status='live', Symbol='NXNL1', Aliases='RDCVF, TXNL6', description='nucleoredoxin-like 1', other_designations='rod-derived cone viability factor|thioredoxin-like 6|thioredoxin-like protein 6', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17455425, end_position_on_the_genomic_accession=17460916, orientation='minus', exon_count=2, OMIM=608791, no_hdr0=''),
  378008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378008, CurrentID=0, Status='live', Symbol='CDY5P', Aliases='', description='chromodomain protein, Y-linked 5 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17721197, end_position_on_the_genomic_accession=17723195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378009, CurrentID=0, Status='live', Symbol='OFD1P1Y', Aliases='OFD1P1, OFD1PY1, OFDYP1', description='OFD1 pseudogene 1, Y-linked', other_designations='OFD1 pseudogene, Y-linked 1|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17809199, end_position_on_the_genomic_accession=17822834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378010, CurrentID=0, Status='live', Symbol='OFD1P2Y', Aliases='OFD1P2, OFD1PY2, OFDYP2', description='OFD1 pseudogene 2, Y-linked', other_designations='OFD1 pseudogene, Y-linked 2|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18080616, end_position_on_the_genomic_accession=18096811, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378011, CurrentID=0, Status='live', Symbol='CDY9P', Aliases='', description='chromodomain protein, Y-linked 9 pseudogene', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18182817, end_position_on_the_genomic_accession=18184815, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378013, CurrentID=0, Status='live', Symbol='TSPY14P', Aliases='TSPYP4', description='testis specific protein, Y-linked 14, pseudogene', other_designations='testis specific protein, Y-linked pseudogene 4', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21468129, end_position_on_the_genomic_accession=21470939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378014, CurrentID=0, Status='live', Symbol='CDY11P', Aliases='', description='chromodomain protein, Y-linked 11 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21696145, end_position_on_the_genomic_accession=21699173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369208, CurrentID=0, Status='live', Symbol='LOC105369208', Aliases='', description='uncharacterized LOC105369208', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68627865, end_position_on_the_genomic_accession=68631657, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645443, CurrentID=0, Status='live', Symbol='ATP5A1P3', Aliases='ATP5AP3', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1, cardiac muscle pseudogene|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, pseudogene 3', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72004903, end_position_on_the_genomic_accession=72006752, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130072, CurrentID=0, Status='live', Symbol='LOC100130072', Aliases='', description='heat-responsive protein 12 pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10478527, end_position_on_the_genomic_accession=10478932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378810, CurrentID=0, Status='live', Symbol='LOC105378810', Aliases='', description='uncharacterized LOC105378810', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=79270404, end_position_on_the_genomic_accession=79405420, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101927795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927795, CurrentID=0, Status='live', Symbol='LOC101927795', Aliases='', description='uncharacterized LOC101927795', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200780495, end_position_on_the_genomic_accession=200812170, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  84852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84852, CurrentID=0, Status='live', Symbol='ATP1A1-AS1', Aliases='ATP1A1OS, C1orf203', description='ATP1A1 antisense RNA 1', other_designations='ATP1A1 opposite strand', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116392865, end_position_on_the_genomic_accession=116418622, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379360, CurrentID=0, Status='live', Symbol='LOC105379360', Aliases='', description='uncharacterized LOC105379360', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31497312, end_position_on_the_genomic_accession=31542644, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  115908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115908, CurrentID=0, Status='live', Symbol='CTHRC1', Aliases='', description='collagen triple helix repeat containing 1', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103371509, end_position_on_the_genomic_accession=103383005, orientation='plus', exon_count=5, OMIM=610635, no_hdr0=''),
  101928502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928502, CurrentID=0, Status='live', Symbol='LOC101928502', Aliases='', description='uncharacterized LOC101928502', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85307448, end_position_on_the_genomic_accession=85312163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=7897, CurrentID=0, Status='live', Symbol='ZNF4', Aliases='', description='zinc finger protein-4', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129276, CurrentID=0, Status='live', Symbol='LOC100129276', Aliases='', description='RNA exonuclease 1 homolog', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55342316, end_position_on_the_genomic_accession=55344958, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927802, CurrentID=0, Status='live', Symbol='LINC00368', Aliases='', description='long intergenic non-protein coding RNA 368', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111095838, end_position_on_the_genomic_accession=111105902, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  115939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115939, CurrentID=0, Status='live', Symbol='TSR3', Aliases='C16orf42', description='TSR3, 20S rRNA accumulation, homolog (S. cerevisiae)', other_designations='probable ribosome biogenesis protein C16orf42', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1349240, end_position_on_the_genomic_accession=1351872, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100289404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289404, CurrentID=0, Status='live', Symbol='MTCO2P22', Aliases='', description='MT-CO2 pseudogene 22', other_designations='mitochondrially encoded cytochrome c oxidase II pseudogene', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100052857, end_position_on_the_genomic_accession=100053539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  115948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115948, CurrentID=0, Status='live', Symbol='CCDC151', Aliases='CILD30', description='coiled-coil domain containing 151', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11420605, end_position_on_the_genomic_accession=11435782, orientation='minus', exon_count=14, OMIM=615956, no_hdr0=''),
  115950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115950, CurrentID=0, Status='live', Symbol='ZNF653', Aliases='E430039K05Rik, ZIP67', description='zinc finger protein 653', other_designations='67 kDa zinc finger protein|zinc finger protein Zip67', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11483427, end_position_on_the_genomic_accession=11505923, orientation='minus', exon_count=9, OMIM=611371, no_hdr0=''),
  101927805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927805, CurrentID=0, Status='live', Symbol='LOC101927805', Aliases='', description='uncharacterized LOC101927805', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18585386, end_position_on_the_genomic_accession=18607187, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373573, CurrentID=0, Status='live', Symbol='LOC105373573', Aliases='', description='uncharacterized LOC105373573', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115537713, end_position_on_the_genomic_accession=115539107, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130321, CurrentID=0, Status='live', Symbol='LOC100130321', Aliases='', description='DNA fragmentation factor, 45kDa, alpha polypeptide pseudogene', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1730206, end_position_on_the_genomic_accession=1731452, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379369, CurrentID=0, Status='live', Symbol='LOC105379369', Aliases='', description='uncharacterized LOC105379369', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=35080575, end_position_on_the_genomic_accession=35093840, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  378108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378108, CurrentID=0, Status='live', Symbol='TRIM74', Aliases='TRIM50C', description='tripartite motif containing 74', other_designations='tripartite motif-containing 50C|tripartite motif-containing 74|tripartite motif-containing protein 50C', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72954849, end_position_on_the_genomic_accession=72969681, orientation='minus', exon_count=9, OMIM=612550, no_hdr0=''),
  101927808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927808, CurrentID=0, Status='live', Symbol='CLSTN2-AS1', Aliases='', description='CLSTN2 antisense RNA 1', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140505618, end_position_on_the_genomic_accession=140508789, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928504, CurrentID=0, Status='live', Symbol='LOC101928504', Aliases='', description='uncharacterized LOC101928504', other_designations='CTD-2506P8.6', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81220875, end_position_on_the_genomic_accession=81222462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376319, CurrentID=0, Status='live', Symbol='LOC105376319', Aliases='', description='uncharacterized LOC105376319', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135793056, end_position_on_the_genomic_accession=135795400, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927811, CurrentID=0, Status='live', Symbol='LOC101927811', Aliases='', description='uncharacterized LOC101927811', other_designations='CTA-293F17.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20328301, end_position_on_the_genomic_accession=20331762, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  115992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=115992, CurrentID=0, Status='live', Symbol='RNF166', Aliases='', description='ring finger protein 166', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88696495, end_position_on_the_genomic_accession=88706421, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927813, CurrentID=0, Status='live', Symbol='HMMR-AS1', Aliases='', description='HMMR antisense RNA 1', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163482604, end_position_on_the_genomic_accession=163494058, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376476, CurrentID=0, Status='live', Symbol='LOC105376476', Aliases='', description='uncharacterized LOC105376476', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29821314, end_position_on_the_genomic_accession=29829111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927814, CurrentID=0, Status='live', Symbol='LOC101927814', Aliases='', description='uncharacterized LOC101927814', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21589698, end_position_on_the_genomic_accession=21599385, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379377, CurrentID=0, Status='live', Symbol='LOC105379377', Aliases='', description='uncharacterized LOC105379377', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37066114, end_position_on_the_genomic_accession=37076857, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927815, CurrentID=0, Status='live', Symbol='LOC101927815', Aliases='', description='uncharacterized LOC101927815', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2530092, end_position_on_the_genomic_accession=2728451, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927817, CurrentID=0, Status='live', Symbol='LOC101927817', Aliases='', description='uncharacterized LOC101927817', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89430918, end_position_on_the_genomic_accession=89454494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  116028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116028, CurrentID=0, Status='live', Symbol='RMI2', Aliases='BLAP18, C16orf75', description='RecQ mediated genome instability 2', other_designations='BLM-associated protein of 18 kDa|RMI2, RecQ mediated genome instability 2, homolog', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11345438, end_position_on_the_genomic_accession=11351763, orientation='plus', exon_count=2, OMIM=612426, no_hdr0=''),
  101927818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927818, CurrentID=0, Status='live', Symbol='LOC101927818', Aliases='', description='uncharacterized LOC101927818', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116033676, end_position_on_the_genomic_accession=116055894, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100009613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100009613, CurrentID=0, Status='live', Symbol='ANO1-AS2', Aliases='C11orf78, NCRNA00224', description='ANO1 antisense RNA 2 (head to head)', other_designations='ANO1 antisense RNA 2 (non-protein coding)', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=70072434, end_position_on_the_genomic_accession=70075348, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928506, CurrentID=0, Status='live', Symbol='LOC101928506', Aliases='', description='uncharacterized LOC101928506', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105974326, end_position_on_the_genomic_accession=105979660, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  116039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116039, CurrentID=0, Status='live', Symbol='OSR2', Aliases='', description='odd-skipped related transciption factor 2', other_designations='odd-skipped related 2', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98938104, end_position_on_the_genomic_accession=98952110, orientation='plus', exon_count=7, OMIM=611297, no_hdr0=''),
  105378811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378811, CurrentID=0, Status='live', Symbol='LOC105378811', Aliases='', description='uncharacterized LOC105378811', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=79309584, end_position_on_the_genomic_accession=79323851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377141, CurrentID=0, Status='live', Symbol='LOC105377141', Aliases='', description='uncharacterized LOC105377141', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66567245, end_position_on_the_genomic_accession=66640280, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  401433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401433, CurrentID=0, Status='live', Symbol='LOC401433', Aliases='', description='uncharacterized LOC401433', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=151074131, end_position_on_the_genomic_accession=151077504, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927822, CurrentID=0, Status='live', Symbol='LOC101927822', Aliases='', description='uncharacterized LOC101927822', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133886496, end_position_on_the_genomic_accession=133902407, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379684, CurrentID=0, Status='live', Symbol='LOC105379684', Aliases='', description='uncharacterized LOC105379684', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390714, CurrentID=0, Status='live', Symbol='LOC390714', Aliases='', description='immunoglobulin heavy chain variable region', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33844758, end_position_on_the_genomic_accession=33858164, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116064, CurrentID=0, Status='live', Symbol='LRRC58', Aliases='', description='leucine rich repeat containing 58', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120324729, end_position_on_the_genomic_accession=120349361, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928507, CurrentID=0, Status='live', Symbol='LOC101928507', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116068, CurrentID=0, Status='live', Symbol='LYSMD3', Aliases='', description='LysM, putative peptidoglycan-binding, domain containing 3', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90515626, end_position_on_the_genomic_accession=90529584, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  116071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116071, CurrentID=0, Status='live', Symbol='BATF2', Aliases='SARI', description='basic leucine zipper transcription factor, ATF-like 2', other_designations='B-ATF-2|suppressor of AP-1 regulated by IFN', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64987945, end_position_on_the_genomic_accession=64997653, orientation='minus', exon_count=5, OMIM=614983, no_hdr0=''),
  105372114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372114, CurrentID=0, Status='live', Symbol='LOC105372114', Aliases='', description='uncharacterized LOC105372114', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50354141, end_position_on_the_genomic_accession=50374866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927827, CurrentID=0, Status='live', Symbol='LOC101927827', Aliases='', description='uncharacterized LOC101927827', other_designations='', map_location='9p11.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42579805, end_position_on_the_genomic_accession=42586534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  116085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116085, CurrentID=0, Status='live', Symbol='SLC22A12', Aliases='OAT4L, RST, URAT1', description='solute carrier family 22 (organic anion/urate transporter), member 12', other_designations='organic anion transporter 4-like protein|renal-specific transporter|solute carrier family 22 (organic anion/cation transporter), member 12|urate anion exchanger 1|urate transporter 1', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64590810, end_position_on_the_genomic_accession=64602353, orientation='plus', exon_count=10, OMIM=607096, no_hdr0=''),
  347028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347028, CurrentID=0, Status='live', Symbol='AFG3L2P1', Aliases='AFG3L2P', description='AFG3-like AAA ATPase 2 pseudogene 1', other_designations='AFG3 ATPase family gene 3-like 2 pseudogene 1|AFG3 ATPase family member 3-like 2 pseudogene 1', map_location='8p11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43247203, end_position_on_the_genomic_accession=43271960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116092, CurrentID=0, Status='live', Symbol='DNTTIP1', Aliases='C20orf167, Tdif1, dJ447F3.4', description='deoxynucleotidyltransferase, terminal, interacting protein 1', other_designations='TdT binding protein|tdT-interacting factor 1|terminal deoxynucleotidyltransferase-interacting factor 1', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45791937, end_position_on_the_genomic_accession=45811427, orientation='plus', exon_count=13, OMIM=611388, no_hdr0=''),
  116093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116093, CurrentID=0, Status='live', Symbol='DIRC1', Aliases='', description='disrupted in renal carcinoma 1', other_designations='disrupted in renal cancer protein', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188733738, end_position_on_the_genomic_accession=188839420, orientation='minus', exon_count=3, OMIM=606423, no_hdr0=''),
  101927829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927829, CurrentID=0, Status='live', Symbol='LOC101927829', Aliases='', description='uncharacterized LOC101927829', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=20342468, end_position_on_the_genomic_accession=20350928, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  267007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267007, CurrentID=0, Status='live', Symbol='PGBD3P3', Aliases='', description='piggyBac transposable element derived 3 pseudogene 3', other_designations='', map_location='12q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126510119, end_position_on_the_genomic_accession=126511276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929117, CurrentID=0, Status='live', Symbol='LOC101929117', Aliases='', description='uncharacterized LOC101929117', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26577376, end_position_on_the_genomic_accession=26579402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130326, CurrentID=0, Status='live', Symbol='LOC100130326', Aliases='', description='transmembrane protein 183A pseudogene', other_designations='', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=84691669, end_position_on_the_genomic_accession=84692591, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372115, CurrentID=0, Status='live', Symbol='LOC105372115', Aliases='', description='uncharacterized LOC105372115', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116113, CurrentID=0, Status='live', Symbol='FOXP4', Aliases='hFKHLA', description='forkhead box P4', other_designations='fork head-related protein like A|fork head-related protein-like A|winged-helix repressor FOXP4', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41546426, end_position_on_the_genomic_accession=41602384, orientation='plus', exon_count=19, OMIM=608924, no_hdr0=''),
  116115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116115, CurrentID=0, Status='live', Symbol='ZNF526', Aliases='', description='zinc finger protein 526', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42220340, end_position_on_the_genomic_accession=42228201, orientation='plus', exon_count=3, OMIM=614387, no_hdr0=''),
  101927833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927833, CurrentID=0, Status='live', Symbol='IFNL4P1', Aliases='', description='interferon, lambda 4 pseudogene 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39264382, end_position_on_the_genomic_accession=39265887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379396, CurrentID=0, Status='live', Symbol='LOC105379396', Aliases='', description='uncharacterized LOC105379396', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42695922, end_position_on_the_genomic_accession=42717097, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  116123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116123, CurrentID=0, Status='live', Symbol='FMO9P', Aliases='', description='flavin containing monooxygenase 9 pseudogene', other_designations='RP11-45J16.2', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166603916, end_position_on_the_genomic_accession=166625238, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101927834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927834, CurrentID=0, Status='live', Symbol='LOC101927834', Aliases='', description='uncharacterized LOC101927834', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127628425, end_position_on_the_genomic_accession=127630771, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927835, CurrentID=0, Status='live', Symbol='LOC101927835', Aliases='', description='uncharacterized LOC101927835', other_designations='CTD-2313P7.1', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=164448422, end_position_on_the_genomic_accession=164467402, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  116135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116135, CurrentID=0, Status='live', Symbol='LRRC3B', Aliases='LRP15', description='leucine rich repeat containing 3B', other_designations='leucine-rich repeat protein LRP15', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26622809, end_position_on_the_genomic_accession=26710776, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  116138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116138, CurrentID=0, Status='live', Symbol='KLHDC3', Aliases='PEAS, dJ20C7.3', description='kelch domain containing 3', other_designations='testis intracellular mediator protein', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43014103, end_position_on_the_genomic_accession=43021298, orientation='plus', exon_count=11, OMIM=611248, no_hdr0=''),
  116143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116143, CurrentID=0, Status='live', Symbol='WDR92', Aliases='', description='WD repeat domain 92', other_designations='WD repeat-containing protein Monad|monad', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68130149, end_position_on_the_genomic_accession=68157560, orientation='minus', exon_count=8, OMIM=610729, no_hdr0=''),
  116149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116149, CurrentID=0, Status='live', Symbol='RPL12P4', Aliases='RPL12_23_1714, dJ800C24.1', description='ribosomal protein L12 pseudogene 4', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55074556, end_position_on_the_genomic_accession=55075178, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116150, CurrentID=0, Status='live', Symbol='NUS1', Aliases='C6orf68, MGC:7199, NgBR, TANGO14', description='NUS1 dehydrodolichyl diphosphate synthase subunit', other_designations='Nogo-B receptor|di-trans,poly-cis-decaprenylcistransferase|nuclear undecaprenyl pyrophosphate synthase 1 homolog|transport and golgi organization 14 homolog', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117675454, end_position_on_the_genomic_accession=117710727, orientation='plus', exon_count=5, OMIM=610463, no_hdr0=''),
  116151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116151, CurrentID=0, Status='live', Symbol='FAM210B', Aliases='5A3, C20orf108, dJ1167H4.1', description='family with sequence similarity 210 member B', other_designations='family with sequence similarity 210, member B|transmembrane protein C20orf108', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56358927, end_position_on_the_genomic_accession=56368663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116154, CurrentID=0, Status='live', Symbol='PHACTR3', Aliases='C20orf101, H17739, PPP1R123, SCAPIN1, SCAPININ', description='phosphatase and actin regulator 3', other_designations='protein phosphatase 1, regulatory subunit 123|scaffold-associated PP1-inhibiting protein', map_location='20q13.32-q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59577509, end_position_on_the_genomic_accession=59847711, orientation='plus', exon_count=18, OMIM=608725, no_hdr0=''),
  101927839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927839, CurrentID=0, Status='live', Symbol='LOC101927839', Aliases='', description='uncharacterized LOC101927839', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2232678, end_position_on_the_genomic_accession=2233610, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379402, CurrentID=0, Status='live', Symbol='LOC105379402', Aliases='', description='uncharacterized LOC105379402', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116159, CurrentID=0, Status='live', Symbol='CYYR1', Aliases='C21orf95', description='cysteine/tyrosine-rich 1', other_designations='cysteine and tyrosine-rich 1|cysteine and tyrosine-rich protein 1 isoform 1,2,2bis,3,4|cysteine and tyrosine-rich protein 1 isoform 1,2,3,4b|cysteine and tyrosine-rich protein 1 isoform 1,2,3b|cysteine and tyrosine-rich protein 1 isoform 1,2,4|proline-rich domain-containing protein', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26466209, end_position_on_the_genomic_accession=26573284, orientation='minus', exon_count=6, OMIM=616020, no_hdr0=''),
  101927840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927840, CurrentID=0, Status='live', Symbol='LOC101927840', Aliases='', description='uncharacterized LOC101927840', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203932317, end_position_on_the_genomic_accession=203940810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379403, CurrentID=0, Status='live', Symbol='LOC105379403', Aliases='', description='uncharacterized LOC105379403', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116166, CurrentID=0, Status='live', Symbol='SORDL', Aliases='', description='sorbitol dehydrogenase-like', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927841, CurrentID=0, Status='live', Symbol='LOC101927841', Aliases='', description='uncharacterized LOC101927841', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379404, CurrentID=0, Status='live', Symbol='LOC105379404', Aliases='', description='uncharacterized LOC105379404', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101059971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059971, CurrentID=0, Status='live', Symbol='LOC101059971', Aliases='', description='uncharacterized LOC101059971', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=19898195, end_position_on_the_genomic_accession=19899863, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116173, CurrentID=0, Status='live', Symbol='CMTM5', Aliases='CKLFSF5', description='CKLF-like MARVEL transmembrane domain containing 5', other_designations='chemokine-like factor super family 5|chemokine-like factor superfamily member 5', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23376808, end_position_on_the_genomic_accession=23379772, orientation='plus', exon_count=6, OMIM=607888, no_hdr0=''),
  116179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116179, CurrentID=0, Status='live', Symbol='TGM7', Aliases='TGMZ', description='transglutaminase 7', other_designations='TG(Z)|TGZ|TGase Z|TGase-7|transglutaminase Z|transglutaminase-7', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43276281, end_position_on_the_genomic_accession=43302255, orientation='minus', exon_count=14, OMIM=606776, no_hdr0=''),
  101927844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927844, CurrentID=0, Status='live', Symbol='LOC101927844', Aliases='', description='uncharacterized LOC101927844', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87212669, end_position_on_the_genomic_accession=87251331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379407, CurrentID=0, Status='live', Symbol='LOC105379407', Aliases='', description='uncharacterized LOC105379407', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69475604, end_position_on_the_genomic_accession=69479894, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927845, CurrentID=0, Status='live', Symbol='LOC101927845', Aliases='', description='uncharacterized LOC101927845', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134849935, end_position_on_the_genomic_accession=134881899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130329, CurrentID=0, Status='live', Symbol='LOC100130329', Aliases='', description='solute carrier family 41 (magnesium transporter), member 2 pseudogene', other_designations='solute carrier family 41, member 2 pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11326139, end_position_on_the_genomic_accession=11330461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927846, CurrentID=0, Status='live', Symbol='LOC101927846', Aliases='', description='uncharacterized LOC101927846', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289448, CurrentID=0, Status='live', Symbol='REXO1L12P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 12, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 12, pseudogene|REX1, RNA exonuclease 1 homolog-like 1 pseudogene', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85714323, end_position_on_the_genomic_accession=85715984, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116211, CurrentID=0, Status='live', Symbol='TM4SF19', Aliases='OCTM4', description='transmembrane 4 L six family member 19', other_designations='osteoclast maturation-associated gene 4 protein|tetraspan membrane protein OCTM4', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196323546, end_position_on_the_genomic_accession=196338420, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101927849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927849, CurrentID=0, Status='live', Symbol='LOC101927849', Aliases='', description='uncharacterized LOC101927849', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=149154295, end_position_on_the_genomic_accession=149278123, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101928512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928512, CurrentID=0, Status='live', Symbol='LOC101928512', Aliases='', description='uncharacterized LOC101928512', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167455054, end_position_on_the_genomic_accession=167501062, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100131820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131820, CurrentID=0, Status='live', Symbol='LOC100131820', Aliases='', description='uncharacterized LOC100131820', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=9693671, end_position_on_the_genomic_accession=9694254, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  116224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116224, CurrentID=0, Status='live', Symbol='FAM122A', Aliases='C9orf42', description='family with sequence similarity 122A', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68780048, end_position_on_the_genomic_accession=68783693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  116225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116225, CurrentID=0, Status='live', Symbol='ZMYND19', Aliases='MIZIP', description='zinc finger, MYND-type containing 19', other_designations='MCH-R1-interacting zinc finger protein|melanin-concentrating hormone receptor 1 interacting zinc-finger protein|melanin-concentrating hormone receptor 1-interacting zinc finger protein|zinc finger, MYND domain containing 19', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137582076, end_position_on_the_genomic_accession=137590485, orientation='minus', exon_count=7, OMIM=611424, no_hdr0=''),
  101927851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927851, CurrentID=0, Status='live', Symbol='LOC101927851', Aliases='', description='uncharacterized LOC101927851', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234957342, end_position_on_the_genomic_accession=234963999, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  116228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116228, CurrentID=0, Status='live', Symbol='COX20', Aliases='FAM36A', description='COX20 cytochrome c oxidase assembly factor', other_designations='COX20 Cox2 chaperone homolog|family with sequence similarity 36, member A', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244835337, end_position_on_the_genomic_accession=244845057, orientation='plus', exon_count=4, OMIM=614698, no_hdr0=''),
  105372119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372119, CurrentID=0, Status='live', Symbol='LOC105372119', Aliases='', description='uncharacterized LOC105372119', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100287453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287453, CurrentID=0, Status='live', Symbol='LOC100287453', Aliases='', description='neurotrophin 4 pseudogene', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49042072, end_position_on_the_genomic_accession=49042928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116236, CurrentID=0, Status='live', Symbol='ABHD15', Aliases='', description='abhydrolase domain containing 15', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29560661, end_position_on_the_genomic_accession=29567024, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116238, CurrentID=0, Status='live', Symbol='TLCD1', Aliases='', description='TLC domain containing 1', other_designations='TLC domain-containing protein 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28724348, end_position_on_the_genomic_accession=28727935, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101927853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927853, CurrentID=0, Status='live', Symbol='LOC101927853', Aliases='', description='uncharacterized LOC101927853', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65796550, end_position_on_the_genomic_accession=65833497, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101927854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927854, CurrentID=0, Status='live', Symbol='LOC101927854', Aliases='', description='uncharacterized LOC101927854', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24151324, end_position_on_the_genomic_accession=24189956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928513, CurrentID=0, Status='live', Symbol='CATIP-AS1', Aliases='', description='CATIP antisense RNA 1', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218366665, end_position_on_the_genomic_accession=218367902, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100289455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289455, CurrentID=0, Status='live', Symbol='LOC100289455', Aliases='', description='uncharacterized LOC100289455', other_designations='', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22436387, end_position_on_the_genomic_accession=22438453, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379418, CurrentID=0, Status='live', Symbol='LOC105379418', Aliases='', description='uncharacterized LOC105379418', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10742027, end_position_on_the_genomic_accession=10753081, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116254, CurrentID=0, Status='live', Symbol='GINM1', Aliases='C6orf72, dJ12G14.2', description='glycoprotein integral membrane 1', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149566392, end_position_on_the_genomic_accession=149590931, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  116255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116255, CurrentID=0, Status='live', Symbol='MOGAT1', Aliases='DGAT2L, DGAT2L1, MGAT1', description='monoacylglycerol O-acyltransferase 1', other_designations='acyl-CoA:monoacylglycerol acyltransferase 1|diacylglycerol O-acyltransferase 2 like 1|diacylglycerol O-acyltransferase candidate 2|diacylglycerol acyltransferase 2-like protein 1|hDC2', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222671738, end_position_on_the_genomic_accession=222709930, orientation='plus', exon_count=6, OMIM=610268, no_hdr0=''),
  106480708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480708, CurrentID=0, Status='live', Symbol='RNU6-32P', Aliases='RNU6-32', description='RNA, U6 small nuclear 32, pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927856, CurrentID=0, Status='live', Symbol='LOC101927856', Aliases='', description='uncharacterized LOC101927856', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64703755, end_position_on_the_genomic_accession=64704260, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927857, CurrentID=0, Status='live', Symbol='LOC101927857', Aliases='', description='uncharacterized LOC101927857', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39655611, end_position_on_the_genomic_accession=39669012, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379420, CurrentID=0, Status='live', Symbol='LOC105379420', Aliases='', description='uncharacterized LOC105379420', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927858, CurrentID=0, Status='live', Symbol='LOC101927858', Aliases='', description='uncharacterized LOC101927858', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156655048, end_position_on_the_genomic_accession=156658121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927859, CurrentID=0, Status='live', Symbol='LOC101927859', Aliases='', description='serine/arginine repetitive matrix protein 2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20337984, end_position_on_the_genomic_accession=20361250, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  64975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64975, CurrentID=0, Status='live', Symbol='MRPL41', Aliases='BMRP, MRP-L27, MRPL27, PIG3, RPML27', description='mitochondrial ribosomal protein L41', other_designations='39S ribosomal protein L27 homolog|L41mt|MRP-L27 homolog|MRP-L41|bcl-2-interacting mitochondrial ribosomal protein L41|cell proliferation-inducing gene 3 protein|proliferation-inducing gene 3', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137551857, end_position_on_the_genomic_accession=137552555, orientation='plus', exon_count=2, OMIM=611846, no_hdr0=''),
  378426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378426, CurrentID=0, Status='live', Symbol='AIS2', Aliases='VAMAS3', description='autoimmune susceptibility 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608391, no_hdr0=''),
  378427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378427, CurrentID=0, Status='live', Symbol='AIS3', Aliases='VAMAS4', description='autoimmune susceptibility 3 (vitiligo specific)', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608392, no_hdr0=''),
  116285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116285, CurrentID=0, Status='live', Symbol='ACSM1', Aliases='BUCS1, MACS1', description='acyl-CoA synthetase medium-chain family member 1', other_designations='Butyrate CoA ligase|butyrate--CoA ligase 1|butyryl Coenzyme A synthetase 1|butyryl-coenzyme A synthetase 1|lipoate-activating enzyme|medium-chain acyl-CoA synthetase|middle-chain acyl-CoA synthetase 1', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20623233, end_position_on_the_genomic_accession=20697744, orientation='minus', exon_count=18, OMIM=614357, no_hdr0=''),
  101927862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927862, CurrentID=0, Status='live', Symbol='CTB-178M22.2', Aliases='', description='uncharacterized LOC101927862', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168229583, end_position_on_the_genomic_accession=168232372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379425, CurrentID=0, Status='live', Symbol='LOC105379425', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927863, CurrentID=0, Status='live', Symbol='LOC101927863', Aliases='', description='uncharacterized LOC101927863', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89491517, end_position_on_the_genomic_accession=89506568, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927864, CurrentID=0, Status='live', Symbol='LOC101927864', Aliases='', description='uncharacterized LOC101927864', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2366589, end_position_on_the_genomic_accession=2371204, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379427, CurrentID=0, Status='live', Symbol='LOC105379427', Aliases='', description='zinc finger protein 717-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11473858, end_position_on_the_genomic_accession=11484612, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379428, CurrentID=0, Status='live', Symbol='LOC105379428', Aliases='', description='uncharacterized LOC105379428', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=12798044, end_position_on_the_genomic_accession=12815761, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101927866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927866, CurrentID=0, Status='live', Symbol='LOC101927866', Aliases='', description='uncharacterized LOC101927866', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146921963, end_position_on_the_genomic_accession=146925218, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369211, CurrentID=0, Status='live', Symbol='LOC105369211', Aliases='', description='uncharacterized LOC105369211', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11559333, end_position_on_the_genomic_accession=11564088, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116328, CurrentID=0, Status='live', Symbol='C8orf34', Aliases='VEST-1, VEST1', description='chromosome 8 open reading frame 34', other_designations='protein VEST-1|vestibule 1|vestibule-1 protein', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68330722, end_position_on_the_genomic_accession=68819023, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  101927868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927868, CurrentID=0, Status='live', Symbol='GRK5-IT1', Aliases='', description='GRK5 intronic transcript 1', other_designations='GRK5 intronic transcript 1 (non-protein coding)', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119208069, end_position_on_the_genomic_accession=119212704, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379431, CurrentID=0, Status='live', Symbol='LOC105379431', Aliases='', description='uncharacterized LOC105379431', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=271648, end_position_on_the_genomic_accession=284291, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376487, CurrentID=0, Status='live', Symbol='LOC105376487', Aliases='', description='uncharacterized LOC105376487', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33249082, end_position_on_the_genomic_accession=33251099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928516, CurrentID=0, Status='live', Symbol='LOC101928516', Aliases='', description='uncharacterized LOC101928516', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=74069451, end_position_on_the_genomic_accession=74690727, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  116337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116337, CurrentID=0, Status='live', Symbol='PANX3', Aliases='PX3', description='pannexin 3', other_designations='gap junction protein pannexin 3', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124611557, end_position_on_the_genomic_accession=124620356, orientation='plus', exon_count=4, OMIM=608422, no_hdr0=''),
  378483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378483, CurrentID=0, Status='live', Symbol='MRX80', Aliases='', description='mental retardation, X-linked 80', other_designations='', map_location='Xq22-q24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  378484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378484, CurrentID=0, Status='live', Symbol='MRX82', Aliases='', description='mental retardation, X-linked 82', other_designations='', map_location='Xq24-q25', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300518, no_hdr0=''),
  101927870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927870, CurrentID=0, Status='live', Symbol='LOC101927870', Aliases='', description='uncharacterized LOC101927870', other_designations='CTB-107G13.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103445207, end_position_on_the_genomic_accession=103514007, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  116349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116349, CurrentID=0, Status='live', Symbol='EXOC3-AS1', Aliases='C5orf55', description='EXOC3 antisense RNA 1', other_designations='uncharacterized protein C5orf55', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=441835, end_position_on_the_genomic_accession=443160, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927872, CurrentID=0, Status='live', Symbol='LOC101927872', Aliases='', description='uncharacterized LOC101927872', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=135375091, end_position_on_the_genomic_accession=135457967, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101927873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927873, CurrentID=0, Status='live', Symbol='LINC01508', Aliases='', description='long intergenic non-protein coding RNA 1508', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90300896, end_position_on_the_genomic_accession=90433489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  116362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116362, CurrentID=0, Status='live', Symbol='RBP7', Aliases='CRBP4, CRBPIV', description='retinol binding protein 7', other_designations='CRABP-IV|CRABP4|cellular retinoic acid-binding protein 4|cellular retinoic acid-binding protein IV|putative cellular retinol-binding protein CRBP IV|retinoid binding protein 7|retinol binding protein 7, cellular', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9997197, end_position_on_the_genomic_accession=10016020, orientation='plus', exon_count=4, OMIM=608604, no_hdr0=''),
  101927874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927874, CurrentID=0, Status='live', Symbol='LOC101927874', Aliases='RARbeta1', description='uncharacterized LOC101927874', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24829344, end_position_on_the_genomic_accession=25120621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928517, CurrentID=0, Status='live', Symbol='LOC101928517', Aliases='', description='uncharacterized LOC101928517', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51152923, end_position_on_the_genomic_accession=51181966, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  116369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116369, CurrentID=0, Status='live', Symbol='SLC26A8', Aliases='SPGF3, TAT1', description='solute carrier family 26 (anion exchanger), member 8', other_designations='anion transporter/exchanger-8', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35943514, end_position_on_the_genomic_accession=36024636, orientation='minus', exon_count=22, OMIM=608480, no_hdr0=''),
  84931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=84931, CurrentID=0, Status='live', Symbol='LINC01101', Aliases='', description='long intergenic non-protein coding RNA 1101', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120464335, end_position_on_the_genomic_accession=120466349, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  116372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116372, CurrentID=0, Status='live', Symbol='LYPD1', Aliases='LYPDC1, PHTS', description='LY6/PLAUR domain containing 1', other_designations='putative HeLa tumor suppressor', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132644764, end_position_on_the_genomic_accession=132671497, orientation='minus', exon_count=5, OMIM=610450, no_hdr0=''),
  105379438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379438, CurrentID=0, Status='live', Symbol='LOC105379438', Aliases='', description='uncharacterized LOC105379438', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61894980, end_position_on_the_genomic_accession=61911166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927876, CurrentID=0, Status='live', Symbol='LOC101927876', Aliases='', description='uncharacterized LOC101927876', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18065657, end_position_on_the_genomic_accession=18074412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116379, CurrentID=0, Status='live', Symbol='IL22RA2', Aliases='CRF2-10, CRF2-S1, CRF2X, IL-22BP, IL-22R-alpha-2, IL-22RA2, ZCYTOR16', description='interleukin 22 receptor subunit alpha 2', other_designations='cytokine receptor class-II member 10|cytokine receptor family type 2, soluble 1|interleukin 22 receptor, alpha 2|interleukin 22-binding protein', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137143820, end_position_on_the_genomic_accession=137173648, orientation='minus', exon_count=7, OMIM=606648, no_hdr0=''),
  101927877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927877, CurrentID=0, Status='live', Symbol='LOC101927877', Aliases='', description='uncharacterized LOC101927877', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62808467, end_position_on_the_genomic_accession=62836371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927878, CurrentID=0, Status='live', Symbol='LOC101927878', Aliases='', description='uncharacterized LOC101927878', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66089211, end_position_on_the_genomic_accession=66096903, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927879, CurrentID=0, Status='live', Symbol='LOC101927879', Aliases='', description='uncharacterized LOC101927879', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39313889, end_position_on_the_genomic_accession=39328008, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100124332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124332, CurrentID=0, Status='live', Symbol='YBX2P1', Aliases='', description='Y-box binding protein 2 pseudogene 1', other_designations='Y box binding protein 2 pseudogene 1', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123991668, end_position_on_the_genomic_accession=123992671, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  64979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64979, CurrentID=0, Status='live', Symbol='MRPL36', Aliases='BRIP1, L36mt, MRP-L36, PRPL36, RPMJ', description='mitochondrial ribosomal protein L36', other_designations='BRCA1-interacting protein 1|putative BRCA1-interacting protein', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1798385, end_position_on_the_genomic_accession=1801423, orientation='minus', exon_count=4, OMIM=611842, no_hdr0=''),
  100420552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420552, CurrentID=0, Status='live', Symbol='RBBP8P1', Aliases='', description='retinoblastoma binding protein 8 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118632363, end_position_on_the_genomic_accession=118634230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373576, CurrentID=0, Status='live', Symbol='LOC105373576', Aliases='', description='uncharacterized LOC105373576', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=116294577, end_position_on_the_genomic_accession=116388224, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480713, CurrentID=0, Status='live', Symbol='USP9YP15', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 15', other_designations='', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927881, CurrentID=0, Status='live', Symbol='LOC101927881', Aliases='', description='uncharacterized LOC101927881', other_designations='', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128864600, end_position_on_the_genomic_accession=128868727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116412, CurrentID=0, Status='live', Symbol='ZNF837', Aliases='', description='zinc finger protein 837', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58367623, end_position_on_the_genomic_accession=58381022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927882, CurrentID=0, Status='live', Symbol='LOC101927882', Aliases='', description='uncharacterized LOC101927882', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7338842, end_position_on_the_genomic_accession=7341916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927884, CurrentID=0, Status='live', Symbol='LOC101927884', Aliases='', description='uncharacterized LOC101927884', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75524068, end_position_on_the_genomic_accession=75542706, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927885, CurrentID=0, Status='live', Symbol='CTB-78F1.2', Aliases='', description='uncharacterized LOC101927885', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=167573496, end_position_on_the_genomic_accession=167729424, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  116437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116437, CurrentID=0, Status='live', Symbol='LINC01257', Aliases='', description='long intergenic non-protein coding RNA 1257', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131165011, end_position_on_the_genomic_accession=131212931, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  116441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116441, CurrentID=0, Status='live', Symbol='TM4SF18', Aliases='L6D', description='transmembrane 4 L six family member 18', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149318498, end_position_on_the_genomic_accession=149333761, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  116442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116442, CurrentID=0, Status='live', Symbol='RAB39B', Aliases='MRX72, WSMN', description='RAB39B, member RAS oncogene family', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155258234, end_position_on_the_genomic_accession=155264589, orientation='minus', exon_count=2, OMIM=300774, no_hdr0=''),
  116443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116443, CurrentID=0, Status='live', Symbol='GRIN3A', Aliases='GluN3A, NMDAR-L, NR3A', description='glutamate receptor, ionotropic, N-methyl-D-aspartate 3A', other_designations='N-methyl-D-aspartate receptor subtype 3A|NMDAR3A|glutamate [NMDA] receptor subunit 3A', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101569352, end_position_on_the_genomic_accession=101738580, orientation='minus', exon_count=11, OMIM=606650, no_hdr0=''),
  116444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116444, CurrentID=0, Status='live', Symbol='GRIN3B', Aliases='GluN3B, NR3B', description='glutamate receptor, ionotropic, N-methyl-D-aspartate 3B', other_designations='N-methyl-D-aspartate receptor subtype 3B|NMDA receptor subunit 3B|NMDA type glutamate receptor subunit NR3B|NMDAR3B|glutamate [NMDA] receptor subunit 3B', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1000438, end_position_on_the_genomic_accession=1010223, orientation='plus', exon_count=9, OMIM=606651, no_hdr0=''),
  116447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116447, CurrentID=0, Status='live', Symbol='TOP1MT', Aliases='', description='topoisomerase (DNA) I, mitochondrial', other_designations='mitochondrial topoisomerase IB', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143309324, end_position_on_the_genomic_accession=143359977, orientation='minus', exon_count=16, OMIM=606387, no_hdr0=''),
  116448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116448, CurrentID=0, Status='live', Symbol='OLIG1', Aliases='BHLHB6, BHLHE21', description='oligodendrocyte transcription factor 1', other_designations='basic domain, helix-loop-helix protein, class B, 6|class B basic helix-loop-helix protein 6|class E basic helix-loop-helix protein 21|oligo1|oligodendrocyte lineage transcription factor 1|oligodendrocyte-specific bHLH transcription factor 1', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33070144, end_position_on_the_genomic_accession=33072422, orientation='plus', exon_count=1, OMIM=606385, no_hdr0=''),
  116449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116449, CurrentID=0, Status='live', Symbol='CLNK', Aliases='MIST', description='cytokine-dependent hematopoietic cell linker', other_designations='mast cell immunoreceptor signal transducer', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10489900, end_position_on_the_genomic_accession=10734853, orientation='minus', exon_count=23, OMIM=611434, no_hdr0=''),
  101927889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927889, CurrentID=0, Status='live', Symbol='LINC01153', Aliases='', description='long intergenic non-protein coding RNA 1153', other_designations='', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121178289, end_position_on_the_genomic_accession=121185966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100124394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124394, CurrentID=0, Status='live', Symbol='KLF3P2', Aliases='', description='Kruppel-like factor 3 (basic) pseudogene 2', other_designations='', map_location='3q', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156976878, end_position_on_the_genomic_accession=156978120, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116461, CurrentID=0, Status='live', Symbol='TSEN15', Aliases='C1orf19, sen15', description='tRNA splicing endonuclease subunit 15', other_designations='TSEN15 tRNA splicing endonuclease subunit|tRNA splicing endonuclease 15 homolog|tRNA-intron endonuclease Sen15', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184051651, end_position_on_the_genomic_accession=184095846, orientation='plus', exon_count=8, OMIM=608756, no_hdr0=''),
  105379453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379453, CurrentID=0, Status='live', Symbol='LOC105379453', Aliases='', description='uncharacterized LOC105379453', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67670264, end_position_on_the_genomic_accession=67686251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100124401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124401, CurrentID=0, Status='live', Symbol='LOC100124401', Aliases='', description='cytokine receptor-like factor 3 pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84031153, end_position_on_the_genomic_accession=84033500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124402, CurrentID=0, Status='live', Symbol='LOC100124402', Aliases='', description='ubiquitin-fold modifier conjugating enzyme 1 pseudogene', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=78486516, end_position_on_the_genomic_accession=78487400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927891, CurrentID=0, Status='live', Symbol='PKN2-AS1', Aliases='', description='PKN2 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88537513, end_position_on_the_genomic_accession=88685204, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369212, CurrentID=0, Status='live', Symbol='LOC105369212', Aliases='', description='uncharacterized LOC105369212', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96081411, end_position_on_the_genomic_accession=96152989, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100124412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124412, CurrentID=0, Status='live', Symbol='FAM138E', Aliases='F379', description='family with sequence similarity 138 member E', other_designations='family with sequence similarity 138, member E', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101954885, end_position_on_the_genomic_accession=101956355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100124414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124414, CurrentID=0, Status='live', Symbol='GINGF4', Aliases='GGF4, HGF4', description='gingival fibromatosis, hereditary, 4', other_designations='', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611010, no_hdr0=''),
  100124415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124415, CurrentID=0, Status='live', Symbol='EKD2', Aliases='', description='Episodic kinesigenic dyskinesia 2', other_designations='', map_location='16q13-q22.1', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611031, no_hdr0=''),
  105376492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376492, CurrentID=0, Status='live', Symbol='LOC105376492', Aliases='', description='uncharacterized LOC105376492', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34993657, end_position_on_the_genomic_accession=35007596, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929027, CurrentID=0, Status='live', Symbol='LOC101929027', Aliases='', description='uncharacterized LOC101929027', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31363481, end_position_on_the_genomic_accession=31370055, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100289495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289495, CurrentID=0, Status='live', Symbol='LOC100289495', Aliases='', description='uncharacterized LOC100289495', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166342631, end_position_on_the_genomic_accession=166351469, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  116496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116496, CurrentID=0, Status='live', Symbol='FAM129A', Aliases='C1orf24, GIG39, NIBAN', description='family with sequence similarity 129 member A', other_designations='cell growth-inhibiting gene 39 protein|family with sequence similarity 129, member A', map_location='1q25', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184791025, end_position_on_the_genomic_accession=184975138, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  101927896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927896, CurrentID=0, Status='live', Symbol='LOC101927896', Aliases='', description='uncharacterized LOC101927896', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234793204, end_position_on_the_genomic_accession=234888684, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  347097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347097, CurrentID=0, Status='live', Symbol='LOC347097', Aliases='', description='dermatan sulfate epimerase pseudogene', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68608899, end_position_on_the_genomic_accession=68612381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379460, CurrentID=0, Status='live', Symbol='LOC105379460', Aliases='', description='uncharacterized LOC105379460', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18490568, end_position_on_the_genomic_accession=18500319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927898, CurrentID=0, Status='live', Symbol='LOC101927898', Aliases='', description='uncharacterized LOC101927898', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63682336, end_position_on_the_genomic_accession=63690982, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  116511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116511, CurrentID=0, Status='live', Symbol='MAS1L', Aliases='MAS-L, MRG, dJ994E9.2', description='MAS1 proto-oncogene like, G protein-coupled receptor', other_designations='MAS-R|MAS1 oncogene-like|MAS1-like|mas-related G protein-coupled MRG', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29486766, end_position_on_the_genomic_accession=29487902, orientation='minus', exon_count=1, OMIM=607235, no_hdr0=''),
  116512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116512, CurrentID=0, Status='live', Symbol='MRGPRD', Aliases='MRGD, TGR7', description='MAS related GPR family member D', other_designations='G-protein coupled receptor TGR7|MAS-related GPR, member D|beta-alanine receptor|mas-related G protein-coupled MRGD', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68980021, end_position_on_the_genomic_accession=68980986, orientation='minus', exon_count=1, OMIM=607231, no_hdr0=''),
  128646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128646, CurrentID=0, Status='live', Symbol='SIRPD', Aliases='PTPNS1L2, dJ576H24.4', description='signal-regulatory protein delta', other_designations='SIRP-delta|protein tyrosine phosphatase non-receptor type substrate 1-like 2|protein tyrosine phosphatase, non-receptor type substrate 1-like 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1534223, end_position_on_the_genomic_accession=1559006, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  392322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392322, CurrentID=0, Status='live', Symbol='ATP5A1P8', Aliases='ATP5A1P9', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 8', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 9', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39654099, end_position_on_the_genomic_accession=39655527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116519, CurrentID=0, Status='live', Symbol='APOA5', Aliases='APOAV, RAP3', description='apolipoprotein A-V', other_designations='apo-AV|apolipoprotein A-V precursor variant 3|apolipoprotein A5|regeneration-associated protein 3', map_location='11q23', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116789370, end_position_on_the_genomic_accession=116792420, orientation='minus', exon_count=4, OMIM=606368, no_hdr0=''),
  100289500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289500, CurrentID=0, Status='live', Symbol='LOC100289500', Aliases='', description='zinc finger, AN1-type domain 1 pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92732018, end_position_on_the_genomic_accession=92732618, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390791, CurrentID=0, Status='live', Symbol='LOC390791', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40367941, end_position_on_the_genomic_accession=40368645, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927901, CurrentID=0, Status='live', Symbol='LOC101927901', Aliases='', description='uncharacterized LOC101927901', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67709047, end_position_on_the_genomic_accession=67729475, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379464, CurrentID=0, Status='live', Symbol='LOC105379464', Aliases='', description='uncharacterized LOC105379464', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28748971, end_position_on_the_genomic_accession=28774040, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927902, CurrentID=0, Status='live', Symbol='LOC101927902', Aliases='', description='uncharacterized LOC101927902', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104940933, end_position_on_the_genomic_accession=105002028, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  116534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116534, CurrentID=0, Status='live', Symbol='MRGPRE', Aliases='GPR167, MGRF, MRGE', description='MAS related GPR family member E', other_designations='G protein-coupled receptor 167|G-protein coupled receptor 167|MAS-related GPR, member E|Mas-related G protein-coupled receptor F|mas-related G protein-coupled MRGE', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3227811, end_position_on_the_genomic_accession=3232386, orientation='minus', exon_count=2, OMIM=607232, no_hdr0=''),
  116535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116535, CurrentID=0, Status='live', Symbol='MRGPRF', Aliases='GPR140, GPR168, MRGF, RTA', description='MAS related GPR family member F', other_designations='G protein-coupled receptor 140|G protein-coupled receptor 168|G protein-coupled receptor MrgF|G-protein coupled receptor 140|G-protein coupled receptor 168|MAS-related GPR, member F|mas-related G protein-coupled MRGF|mas-related gene F protein|seven transmembrane helix receptor', map_location='11q13.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69004394, end_position_on_the_genomic_accession=69013382, orientation='minus', exon_count=3, OMIM=607233, no_hdr0=''),
  105372428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372428, CurrentID=0, Status='live', Symbol='LOC105372428', Aliases='', description='uncharacterized LOC105372428', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47887210, end_position_on_the_genomic_accession=47887564, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116540, CurrentID=0, Status='live', Symbol='MRPL53', Aliases='L53MT', description='mitochondrial ribosomal protein L53', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74471958, end_position_on_the_genomic_accession=74472815, orientation='minus', exon_count=3, OMIM=611857, no_hdr0=''),
  116541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116541, CurrentID=0, Status='live', Symbol='MRPL54', Aliases='L54mt, MRP-L54', description='mitochondrial ribosomal protein L54', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3762667, end_position_on_the_genomic_accession=3767565, orientation='plus', exon_count=3, OMIM=611858, no_hdr0=''),
  101927904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927904, CurrentID=0, Status='live', Symbol='LOC101927904', Aliases='', description='uncharacterized LOC101927904', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3911511, end_position_on_the_genomic_accession=3936479, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928523, CurrentID=0, Status='live', Symbol='LOC101928523', Aliases='', description='uncharacterized LOC101928523', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103999709, end_position_on_the_genomic_accession=104000482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101927905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927905, CurrentID=0, Status='live', Symbol='LOC101927905', Aliases='', description='uncharacterized LOC101927905', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8235415, end_position_on_the_genomic_accession=8242564, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100271486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271486, CurrentID=0, Status='live', Symbol='RPL31P60', Aliases='RPL31_33_1645', description='ribosomal protein L31 pseudogene 60', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32963508, end_position_on_the_genomic_accession=32963865, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927906, CurrentID=0, Status='live', Symbol='LOC101927906', Aliases='', description='uncharacterized LOC101927906', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235641994, end_position_on_the_genomic_accession=235643820, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379469, CurrentID=0, Status='live', Symbol='LOC105379469', Aliases='', description='uncharacterized LOC105379469', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33370463, end_position_on_the_genomic_accession=33373967, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  378706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378706, CurrentID=0, Status='live', Symbol='RN7SL2', Aliases='7L1C, 7L30.1, 7SL1c, RNSRP2', description='RNA, 7SL, cytoplasmic 2', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49862551, end_position_on_the_genomic_accession=49862849, orientation='minus', exon_count=1, OMIM=612179, no_hdr0=''),
  378707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378707, CurrentID=0, Status='live', Symbol='RN7SL3', Aliases='7L30.2, RN7SL631P, RNSRP3', description='RNA, 7SL, cytoplasmic 3', other_designations='RNA, 7SL, cytoplasmic 631, pseudogene', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49853616, end_position_on_the_genomic_accession=49853914, orientation='minus', exon_count=0, OMIM=612180, no_hdr0=''),
  378708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378708, CurrentID=0, Status='live', Symbol='APITD1', Aliases='CENP-S, CENPS, FAAP16, MHF1', description='apoptosis-inducing, TAF9-like domain 1', other_designations='FANCM-interacting histone fold protein 1|Fanconi anemia-associated polypeptide of 16 kDa|apoptosis-inducing TAF9-like domain-containing protein 1', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10430102, end_position_on_the_genomic_accession=10442815, orientation='plus', exon_count=6, OMIM=609130, no_hdr0=''),
  101927908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927908, CurrentID=0, Status='live', Symbol='LOC101927908', Aliases='', description='uncharacterized LOC101927908', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=167164934, end_position_on_the_genomic_accession=167168401, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379472, CurrentID=0, Status='live', Symbol='LOC105379472', Aliases='', description='serine/arginine repetitive matrix protein 3-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30394113, end_position_on_the_genomic_accession=30395749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100124516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124516, CurrentID=0, Status='live', Symbol='SNORD58C', Aliases='U58', description='small nucleolar RNA, C/D box 58C', other_designations='SNORD58C snoRNA|U58 small nucleolar RNA|U58 snoRNA', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49489235, end_position_on_the_genomic_accession=49489324, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927910, CurrentID=0, Status='live', Symbol='LOC101927910', Aliases='', description='uncharacterized LOC101927910', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89912040, end_position_on_the_genomic_accession=89916412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101927911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927911, CurrentID=0, Status='live', Symbol='LOC101927911', Aliases='', description='uncharacterized LOC101927911', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2962247, end_position_on_the_genomic_accession=2965895, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379475, CurrentID=0, Status='live', Symbol='LOC105379475', Aliases='', description='uncharacterized LOC105379475', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34381587, end_position_on_the_genomic_accession=34386237, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100124533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124533, CurrentID=0, Status='live', Symbol='SCARNA27', Aliases='', description='small Cajal body-specific RNA 27', other_designations='SCARNA27 snoRNA', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8086408, end_position_on_the_genomic_accession=8086533, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124534, CurrentID=0, Status='live', Symbol='SNORA84', Aliases='', description='small nucleolar RNA, H/ACA box 84', other_designations='SNORA84 snoRNA', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92292461, end_position_on_the_genomic_accession=92292593, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124535, CurrentID=0, Status='live', Symbol='SNORA36C', Aliases='', description='small nucleolar RNA, H/ACA box 36C', other_designations='SNORA36C snoRNA|small nucleolar RNA, H/ACA box 36C (retrotransposed)', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69520044, end_position_on_the_genomic_accession=69520171, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124536, CurrentID=0, Status='live', Symbol='SNORA38B', Aliases='', description='small nucleolar RNA, H/ACA box 38B', other_designations='SNORA38B snoRNA|small nucleolar RNA, H/ACA box 38B (retrotransposed)', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67740670, end_position_on_the_genomic_accession=67740801, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124537, CurrentID=0, Status='live', Symbol='SNORA70B', Aliases='', description='small nucleolar RNA, H/ACA box 70B', other_designations='SNORA70B snoRNA|small nucleolar RNA, H/ACA box 70B (retrotransposed)', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61417244, end_position_on_the_genomic_accession=61417377, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124538, CurrentID=0, Status='live', Symbol='SNORA70C', Aliases='U70C', description='small nucleolar RNA, H/ACA box 70C', other_designations='SNORA70C snoRNA|U70 type 3 snoRT|small nucleolar RNA, H/ACA box 70C (retrotransposed)', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117181066, end_position_on_the_genomic_accession=117181201, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124539, CurrentID=0, Status='live', Symbol='SNORA11B', Aliases='', description='small nucleolar RNA, H/ACA box 11B', other_designations='SNORA11B snoRNA|small nucleolar RNA, H/ACA box 11B (retrotransposed)', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91126426, end_position_on_the_genomic_accession=91126552, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124540, CurrentID=0, Status='live', Symbol='SNORA11C', Aliases='', description='small nucleolar RNA, H/ACA box 11C', other_designations='SNORA11C snoRNA|small nucleolar RNA, H/ACA box 11C (retrotransposed)', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47388650, end_position_on_the_genomic_accession=47388776, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124541, CurrentID=0, Status='live', Symbol='SNORA11D', Aliases='', description='small nucleolar RNA, H/ACA box 11D', other_designations='SNORA11D snoRNA', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52190622, end_position_on_the_genomic_accession=52190748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101927915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927915, CurrentID=0, Status='live', Symbol='LOC101927915', Aliases='', description='uncharacterized LOC101927915', other_designations='', map_location='8q24.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=137406101, end_position_on_the_genomic_accession=137413588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390806, CurrentID=0, Status='live', Symbol='LOC390806', Aliases='', description='family with sequence similarity 136, member A pseudogene', other_designations='', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63610772, end_position_on_the_genomic_accession=63611186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101059986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059986, CurrentID=0, Status='live', Symbol='LOC101059986', Aliases='', description='cancer-related nucleoside-triphosphatase pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139482849, end_position_on_the_genomic_accession=139493141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927917, CurrentID=0, Status='live', Symbol='TUB-AS1', Aliases='', description='TUB antisense RNA 1', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8060038, end_position_on_the_genomic_accession=8069373, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  116629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116629, CurrentID=0, Status='live', Symbol='HTX3', Aliases='', description='Heterotaxy, visceral, 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606325, no_hdr0=''),
  101927919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927919, CurrentID=0, Status='live', Symbol='LOC101927919', Aliases='', description='uncharacterized LOC101927919', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117672813, end_position_on_the_genomic_accession=117675284, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101927920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927920, CurrentID=0, Status='live', Symbol='LOC101927920', Aliases='', description='probable ribosome biogenesis protein RLP24 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66760532, end_position_on_the_genomic_accession=66761076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379483, CurrentID=0, Status='live', Symbol='LOC105379483', Aliases='', description='uncharacterized LOC105379483', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30610098, end_position_on_the_genomic_accession=30613167, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379484, CurrentID=0, Status='live', Symbol='LOC105379484', Aliases='', description='uncharacterized LOC105379484', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5011976, end_position_on_the_genomic_accession=5012684, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101927922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927922, CurrentID=0, Status='live', Symbol='LINC01479', Aliases='HRAT39', description='long intergenic non-protein coding RNA 1479', other_designations='heart tissue-associated transcript 39', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67929235, end_position_on_the_genomic_accession=67970017, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379485, CurrentID=0, Status='live', Symbol='LOC105379485', Aliases='', description='uncharacterized LOC105379485', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5055740, end_position_on_the_genomic_accession=5061872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927923, CurrentID=0, Status='live', Symbol='LOC101927923', Aliases='', description='small integral membrane protein 12-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105353599, end_position_on_the_genomic_accession=105354656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378805, CurrentID=0, Status='live', Symbol='LINC-PINT', Aliases='LincRNA-Pint, MKLN1-AS1, PINT', description='long intergenic non-protein coding RNA, p53 induced transcript', other_designations='MKLN1 antisense RNA 1 (head to head)|p53 induced noncoding transcript', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130877562, end_position_on_the_genomic_accession=131109916, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  378807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378807, CurrentID=0, Status='live', Symbol='CATSPER4', Aliases='', description='cation channel, sperm associated 4', other_designations='ion channel CatSper4', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26190587, end_position_on_the_genomic_accession=26202542, orientation='plus', exon_count=11, OMIM=609121, no_hdr0=''),
  378810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378810, CurrentID=0, Status='live', Symbol='SNX19P1', Aliases='SNX19P', description='sorting nexin 19 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13525103, end_position_on_the_genomic_accession=13530011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379487, CurrentID=0, Status='live', Symbol='LOC105379487', Aliases='', description='uncharacterized LOC105379487', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5499383, end_position_on_the_genomic_accession=5502552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  652960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652960, CurrentID=0, Status='live', Symbol='BTF3P7', Aliases='', description='basic transcription factor 3 pseudogene 7', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6794656, end_position_on_the_genomic_accession=6795501, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378818, CurrentID=0, Status='live', Symbol='RPL37P3', Aliases='RL37P, RPL37_13_1720', description='ribosomal protein L37 pseudogene 3', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17894148, end_position_on_the_genomic_accession=17894507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378819, CurrentID=0, Status='live', Symbol='RPL37P4', Aliases='RL37P2, RPL37_10_1721', description='ribosomal protein L37 pseudogene 4', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19593707, end_position_on_the_genomic_accession=19594271, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130345, CurrentID=0, Status='live', Symbol='LOC100130345', Aliases='', description='cadherin-related family member 3-like', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63749816, end_position_on_the_genomic_accession=63785022, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379489, CurrentID=0, Status='live', Symbol='LOC105379489', Aliases='', description='uncharacterized LOC105379489', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5544930, end_position_on_the_genomic_accession=5546787, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  378825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378825, CurrentID=0, Status='live', Symbol='LINC00162', Aliases='C21orf113, NCRNA00162, NLC1-C, NLC1C, PRED74', description='long intergenic non-protein coding RNA 162', other_designations='Putative narcolepsy candidate region 1 gene C protein|Putative uncharacterized protein C21orf113|narcolepsy candidate region 1 gene C|narcolepsy candidate region gene 1C', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44999208, end_position_on_the_genomic_accession=45004727, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101059988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059988, CurrentID=0, Status='live', Symbol='BNIP3P41', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 41', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49549, end_position_on_the_genomic_accession=51521, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347127, CurrentID=0, Status='live', Symbol='SPATA31D5P', Aliases='FAM75D5, FAM75D5P', description='SPATA31 subfamily D, member 5, pseudogene', other_designations='FAM75-like protein FLJ46321 pseudogene|family with sequence similarity 75, member D5, pseudogene', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81913437, end_position_on_the_genomic_accession=81919927, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  378828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378828, CurrentID=0, Status='live', Symbol='LINC00317', Aliases='C21orf117, NCRNA00317, PRED89', description='long intergenic non-protein coding RNA 317', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=21723293, end_position_on_the_genomic_accession=21737319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379490, CurrentID=0, Status='live', Symbol='LOC105379490', Aliases='', description='uncharacterized LOC105379490', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6001516, end_position_on_the_genomic_accession=6009381, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  378832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378832, CurrentID=0, Status='live', Symbol='COL18A1-AS1', Aliases='C21orf123, NCRNA00175, PRED80', description='COL18A1 antisense RNA 1', other_designations='COL18A1 antisense RNA 1 (non-protein coding)', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45419716, end_position_on_the_genomic_accession=45425070, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379491, CurrentID=0, Status='live', Symbol='LOC105379491', Aliases='', description='uncharacterized LOC105379491', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6274186, end_position_on_the_genomic_accession=6276043, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652964, CurrentID=0, Status='live', Symbol='BTF3P3', Aliases='', description='basic transcription factor 3, pseudogene 3', other_designations='basic transcription factor 3 pseudogene', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22492500, end_position_on_the_genomic_accession=22493375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379493, CurrentID=0, Status='live', Symbol='CH507-145C22.1', Aliases='', description='uncharacterized LOC105379493', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6318421, end_position_on_the_genomic_accession=6359726, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  9506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=9506, CurrentID=0, Status='live', Symbol='PAGE4', Aliases='CT16.7, GAGE-9, GAGEC1, JM-27, JM27, PAGE-1, PAGE-4', description='PAGE family member 4', other_designations='P antigen family, member 4 (prostate associated)|g antigen family C member 1|prostate-associated gene protein 4', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49829265, end_position_on_the_genomic_accession=49834037, orientation='plus', exon_count=6, OMIM=300287, no_hdr0=''),
  101927931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927931, CurrentID=0, Status='live', Symbol='LINC01480', Aliases='', description='long intergenic non-protein coding RNA 1480', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41535498, end_position_on_the_genomic_accession=41536904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379494, CurrentID=0, Status='live', Symbol='LOC105379494', Aliases='', description='uncharacterized LOC105379494', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6374075, end_position_on_the_genomic_accession=6376537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116710, CurrentID=0, Status='live', Symbol='FRA18C', Aliases='', description='fragile site, aphidicolin type, common, fra(18)(q22.2)', other_designations='fragile site, folic acid type, rare, fra(18)(q22.1)', map_location='18q22.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927932, CurrentID=0, Status='live', Symbol='LOC101927932', Aliases='', description='uncharacterized LOC101927932', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58863756, end_position_on_the_genomic_accession=58888810, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379496, CurrentID=0, Status='live', Symbol='LOC105379496', Aliases='', description='uncharacterized LOC105379496', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6675125, end_position_on_the_genomic_accession=6677882, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376500, CurrentID=0, Status='live', Symbol='LOC105376500', Aliases='', description='uncharacterized LOC105376500', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38157046, end_position_on_the_genomic_accession=38169066, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928529, CurrentID=0, Status='live', Symbol='LOC101928529', Aliases='', description='uncharacterized LOC101928529', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44477741, end_position_on_the_genomic_accession=44483900, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379497, CurrentID=0, Status='live', Symbol='LOC105379497', Aliases='', description='uncharacterized LOC105379497', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6713083, end_position_on_the_genomic_accession=6714940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  116729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116729, CurrentID=0, Status='live', Symbol='PPP1R27', Aliases='DYSFIP1', description='protein phosphatase 1 regulatory subunit 27', other_designations='dysferlin interacting protein 1|dysferlin-interacting protein 1 (toonin)|protein phosphatase 1, regulatory subunit 27|toonin', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81833492, end_position_on_the_genomic_accession=81835050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  116730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116730, CurrentID=0, Status='live', Symbol='RPL37P2', Aliases='PSANK1, RPL37_9_1150', description='ribosomal protein L37 pseudogene 2', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67682693, end_position_on_the_genomic_accession=67683058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420607, CurrentID=0, Status='live', Symbol='RBBP8P2', Aliases='', description='retinoblastoma binding protein 8 pseudogene 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27077218, end_position_on_the_genomic_accession=27077954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376179, CurrentID=0, Status='live', Symbol='LOC105376179', Aliases='', description='uncharacterized LOC105376179', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100805620, end_position_on_the_genomic_accession=100852621, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  378881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378881, CurrentID=0, Status='live', Symbol='MIR381HG', Aliases='C14orf89, NCRNA00225', description='MIR381 host gene', other_designations='MIR381 host gene (non-protein coding)', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101045157, end_position_on_the_genomic_accession=101051795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  378882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378882, CurrentID=0, Status='live', Symbol='MRX84', Aliases='', description='mental retardation, X-linked 84', other_designations='', map_location='Xp11.3-q22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300505, no_hdr0=''),
  378884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378884, CurrentID=0, Status='live', Symbol='NHLRC1', Aliases='EPM2A, EPM2B, MALIN, bA204B7.2', description='NHL repeat containing E3 ubiquitin protein ligase 1', other_designations='NHL repeat containing 1|NHL repeat-containing protein 1', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18120487, end_position_on_the_genomic_accession=18122620, orientation='minus', exon_count=1, OMIM=608072, no_hdr0=''),
  378888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378888, CurrentID=0, Status='live', Symbol='HSN1B', Aliases='', description='Hereditary sensory neuropathy, type IB', other_designations='', map_location='3p24-p22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608088, no_hdr0=''),
  105379500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379500, CurrentID=0, Status='live', Symbol='LOC105379500', Aliases='', description='uncharacterized LOC105379500', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7002195, end_position_on_the_genomic_accession=7004952, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370174, CurrentID=0, Status='live', Symbol='LOC105370174', Aliases='', description='uncharacterized LOC105370174', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41517208, end_position_on_the_genomic_accession=41564091, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928530, CurrentID=0, Status='live', Symbol='LOC101928530', Aliases='', description='uncharacterized LOC101928530', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132887842, end_position_on_the_genomic_accession=132888571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379502, CurrentID=0, Status='live', Symbol='LOC105379502', Aliases='', description='uncharacterized LOC105379502', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7383973, end_position_on_the_genomic_accession=7386730, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132330, CurrentID=0, Status='live', Symbol='LOC100132330', Aliases='', description='mal, T-cell differentiation protein-like pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88818274, end_position_on_the_genomic_accession=88818438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100124696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124696, CurrentID=0, Status='live', Symbol='TEC', Aliases='', description='transient erythroblastopenia of childhood', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=227050, no_hdr0=''),
  652975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652975, CurrentID=0, Status='live', Symbol='RPL21P6', Aliases='RPL21_47_1368', description='ribosomal protein L21 pseudogene 6', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55413540, end_position_on_the_genomic_accession=55414102, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100124700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100124700, CurrentID=0, Status='live', Symbol='HOTAIR', Aliases='HOXAS, HOXC-AS4, HOXC11-AS1, NCRNA00072', description='HOX transcript antisense RNA', other_designations='HOX transcript antisense RNA (non-protein coding)|HOXC cluster antisense RNA 4 (non-protein coding)|hox transcript antisense intergenic RNA', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53962308, end_position_on_the_genomic_accession=53974956, orientation='minus', exon_count=7, OMIM=611400, no_hdr0=''),
  100420613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420613, CurrentID=0, Status='live', Symbol='LRRC34P2', Aliases='', description='leucine rich repeat containing 34 pseudogene 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55313032, end_position_on_the_genomic_accession=55313924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379505, CurrentID=0, Status='live', Symbol='LOC105379505', Aliases='', description='uncharacterized LOC105379505', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7854003, end_position_on_the_genomic_accession=7856885, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100289543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289543, CurrentID=0, Status='live', Symbol='LOC100289543', Aliases='', description='uncharacterized LOC100289543', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32454914, end_position_on_the_genomic_accession=32455437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378925, CurrentID=0, Status='live', Symbol='RNF148', Aliases='', description='ring finger protein 148', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122701666, end_position_on_the_genomic_accession=122702967, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928531, CurrentID=0, Status='live', Symbol='LOC101928531', Aliases='', description='uncharacterized LOC101928531', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15948508, end_position_on_the_genomic_accession=15954649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100420617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420617, CurrentID=0, Status='live', Symbol='LOC100420617', Aliases='', description='FERM and PDZ domain containing 2 pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46870858, end_position_on_the_genomic_accession=46873829, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390836, CurrentID=0, Status='live', Symbol='OR4K8P', Aliases='OR18-1, OR4K9P', description='olfactory receptor family 4 subfamily K member 8 pseudogene', other_designations='olfactory receptor OR18-1 pseudogene|olfactory receptor, family 4, subfamily K, member 8 pseudogene|olfactory receptor, family 4, subfamily K, member 9 pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14613210, end_position_on_the_genomic_accession=14614181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378938, CurrentID=0, Status='live', Symbol='MALAT1', Aliases='HCN, LINC00047, NCRNA00047, NEAT2, PRO2853, mascRNA', description='metastasis associated lung adenocarcinoma transcript 1 (non-protein coding)', other_designations='MALAT1-associated small cytoplasmic RNA|hepcarcin|long intergenic non-protein coding RNA 47|nuclear enriched abundant transcript 2|nuclear paraspeckle assembly transcript 2 (non-protein coding)', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65497762, end_position_on_the_genomic_accession=65506469, orientation='plus', exon_count=1, OMIM=607924, no_hdr0=''),
  100420618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420618, CurrentID=0, Status='live', Symbol='LOC100420618', Aliases='', description='phospholipase D family member 5 pseudogene', other_designations='phospholipase D family, member 5 pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38310046, end_position_on_the_genomic_accession=38333163, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927947, CurrentID=0, Status='live', Symbol='LOC101927947', Aliases='', description='uncharacterized LOC101927947', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154005132, end_position_on_the_genomic_accession=154298819, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  378948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378948, CurrentID=0, Status='live', Symbol='RBMY1B', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member B', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21511338, end_position_on_the_genomic_accession=21525786, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  378949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378949, CurrentID=0, Status='live', Symbol='RBMY1D', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member D', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21868388, end_position_on_the_genomic_accession=21900745, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  378950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378950, CurrentID=0, Status='live', Symbol='RBMY1E', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member E', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21902181, end_position_on_the_genomic_accession=21918044, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  378951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378951, CurrentID=0, Status='live', Symbol='RBMY1J', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member J', other_designations='y chromosome RNA recognition motif 2', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22403476, end_position_on_the_genomic_accession=22424890, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  378952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378952, CurrentID=0, Status='live', Symbol='RBMY2GP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member G pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6328189, end_position_on_the_genomic_accession=6343584, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116809, CurrentID=0, Status='live', Symbol='RPL38P1', Aliases='RPL38_1_417', description='ribosomal protein L38 pseudogene 1', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148695955, end_position_on_the_genomic_accession=148696248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  378955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378955, CurrentID=0, Status='live', Symbol='RBMY2JP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member J pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8127124, end_position_on_the_genomic_accession=8144630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=378956, CurrentID=0, Status='live', Symbol='RBMY2NP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member N pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9831418, end_position_on_the_genomic_accession=9835967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928532, CurrentID=0, Status='live', Symbol='LOC101928532', Aliases='', description='uncharacterized LOC101928532', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8745396, end_position_on_the_genomic_accession=8747703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379513, CurrentID=0, Status='live', Symbol='LOC105379513', Aliases='', description='uncharacterized LOC105379513', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9677284, end_position_on_the_genomic_accession=9678341, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480727, CurrentID=0, Status='live', Symbol='MTND6P11', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 6 pseudogene 11', other_designations='MT-ND6 pseudogene 11', map_location='2q22.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116828, CurrentID=0, Status='live', Symbol='N4BP2L2-IT2', Aliases='CG030', description='N4BPL2 intronic transcript 2', other_designations='N4BPL2 intronic transcript 2 (non-protein coding)|uncharacterized CG030', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32504506, end_position_on_the_genomic_accession=32509395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  116832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116832, CurrentID=0, Status='live', Symbol='RPL39L', Aliases='RPL39L1', description='ribosomal protein L39 like', other_designations='60S ribosomal protein L39-2|L39-2|ribosomal protein L39-like 1|ribosomal protein L39-like protein', map_location='3q27', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187120953, end_position_on_the_genomic_accession=187139475, orientation='minus', exon_count=3, OMIM=607547, no_hdr0=''),
  116833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116833, CurrentID=0, Status='live', Symbol='ANIB1', Aliases='AIB, AIB1', description='aneurysm, intracranial berry 1', other_designations='', map_location='7q11.2', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=105800, no_hdr0=''),
  116835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116835, CurrentID=0, Status='live', Symbol='HSPA12B', Aliases='C20orf60', description='heat shock protein family A (Hsp70) member 12B', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3732670, end_position_on_the_genomic_accession=3753111, orientation='plus', exon_count=15, OMIM=610702, no_hdr0=''),
  101927953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927953, CurrentID=0, Status='live', Symbol='C5orf66-AS1', Aliases='CTC-276P9.1, Epist', description='C5orf66 antisense RNA 1', other_designations='esophagus epithelial intergenic specific transcript', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=135038831, end_position_on_the_genomic_accession=135040047, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116840, CurrentID=0, Status='live', Symbol='CNTROB', Aliases='LIP8, PP1221', description='centrobin, centrosomal BRCA2 interacting protein', other_designations='LYST-interacting protein 8|LYST-interacting protein LIP8|centrosomal BRCA2-interacting protein', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7932124, end_position_on_the_genomic_accession=7949919, orientation='plus', exon_count=19, OMIM=611425, no_hdr0=''),
  116841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116841, CurrentID=0, Status='live', Symbol='SNAP47', Aliases='C1orf142, ESFI5812, HEL-S-290, HEL170, SNAP-47, SVAP1', description='synaptosome associated protein 47kDa', other_designations='epididymis luminal protein 170|epididymis secretory protein Li 290|synaptosomal-associated 47 kDa protein|synaptosomal-associated protein, 47kDa', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227734996, end_position_on_the_genomic_accession=227781231, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  116842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116842, CurrentID=0, Status='live', Symbol='LEAP2', Aliases='LEAP-2', description='liver expressed antimicrobial peptide 2', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132873666, end_position_on_the_genomic_accession=132874890, orientation='plus', exon_count=3, OMIM=611373, no_hdr0=''),
  116843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116843, CurrentID=0, Status='live', Symbol='SLC18B1', Aliases='C6orf192, dJ55C23.6', description='solute carrier family 18, subfamily B, member 1', other_designations='MFS-type transporter C6orf192|solute carrier family 18 member B1', map_location='6q22.3-q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132769368, end_position_on_the_genomic_accession=132804931, orientation='minus', exon_count=16, OMIM=613361, no_hdr0=''),
  116844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116844, CurrentID=0, Status='live', Symbol='LRG1', Aliases='HMFT1766, LRG', description='leucine-rich alpha-2-glycoprotein 1', other_designations='1300008B03Rik|2310031E04Rik|leucine rich alpha 2 glycoprotein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4537215, end_position_on_the_genomic_accession=4540024, orientation='minus', exon_count=2, OMIM=611289, no_hdr0=''),
  105379518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379518, CurrentID=0, Status='live', Symbol='LOC105379518', Aliases='', description='uncharacterized LOC105379518', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18233149, end_position_on_the_genomic_accession=18235228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  379003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379003, CurrentID=0, Status='live', Symbol='DFNB40', Aliases='', description='deafness, autosomal recessive 40', other_designations='', map_location='22q11.21-q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608264, no_hdr0=''),
  379005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379005, CurrentID=0, Status='live', Symbol='RBMY2OP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member O pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9951547, end_position_on_the_genomic_accession=9959843, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379006, CurrentID=0, Status='live', Symbol='RBMY2QP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member Q pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10022216, end_position_on_the_genomic_accession=10035290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379007, CurrentID=0, Status='live', Symbol='RBMY2SP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member S pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21311161, end_position_on_the_genomic_accession=21319435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379008, CurrentID=0, Status='live', Symbol='RBMY2TP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member T pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21430324, end_position_on_the_genomic_accession=21437347, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379009, CurrentID=0, Status='live', Symbol='RBMY2UP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member U pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22198322, end_position_on_the_genomic_accession=22201877, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379010, CurrentID=0, Status='live', Symbol='RBMY2VP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member V pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22381566, end_position_on_the_genomic_accession=22388096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379011, CurrentID=0, Status='live', Symbol='RBMY2WP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member W pseudogene', other_designations='RNA binding motif protein, Y-linked, family 2, pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22762324, end_position_on_the_genomic_accession=22769890, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379012, CurrentID=0, Status='live', Symbol='RBMY2XP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member X pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24396077, end_position_on_the_genomic_accession=24403633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379013, CurrentID=0, Status='live', Symbol='RNF138P1', Aliases='', description='ring finger protein 138, E3 ubiquitin protein ligase pseudogene 1', other_designations='ring finger protein 138 pseudogene 1', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55528842, end_position_on_the_genomic_accession=55534542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379521, CurrentID=0, Status='live', Symbol='LOC105379521', Aliases='', description='uncharacterized LOC105379521', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  379024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379024, CurrentID=0, Status='live', Symbol='RBMY2YP', Aliases='', description='RNA binding motif protein, Y-linked, family 2, member Y pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25266480, end_position_on_the_genomic_accession=25274045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379025, CurrentID=0, Status='live', Symbol='PSMA3-AS1', Aliases='', description='PSMA3 antisense RNA 1', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58265365, end_position_on_the_genomic_accession=58298137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  379027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379027, CurrentID=0, Status='live', Symbol='TSPY5P', Aliases='TSPYP5', description='testis specific protein, Y-linked 5, pseudogene', other_designations='testis specific protein, Y-linked 1 pseudogene|testis specific protein, Y-linked 5 (pseudogene)|testis specific protein, Y-linked pseudogene 5', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10066393, end_position_on_the_genomic_accession=10069196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379028, CurrentID=0, Status='live', Symbol='XKRYP1', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 1', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 1|XK, Kell blood group complex subunit-related, Y-linked pseudogene 1', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18507812, end_position_on_the_genomic_accession=18509608, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379029, CurrentID=0, Status='live', Symbol='XKRYP2', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 2', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 2|XK, Kell blood group complex subunit-related, Y-linked pseudogene 2', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18810798, end_position_on_the_genomic_accession=18812594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379030, CurrentID=0, Status='live', Symbol='XKRYP3', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 3', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 3|XK, Kell blood group complex subunit-related family, member 3 pseudogene|XK, Kell blood group complex subunit-related, Y-linked pseudogene 3', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23715687, end_position_on_the_genomic_accession=23717485, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379031, CurrentID=0, Status='live', Symbol='XKRYP4', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 4', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 4|XK, Kell blood group complex subunit-related, Y-linked pseudogene 4', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23916696, end_position_on_the_genomic_accession=23918494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  379032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379032, CurrentID=0, Status='live', Symbol='XKRYP5', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 5', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 5|XK, Kell blood group complex subunit-related, Y-linked pseudogene 5', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25751617, end_position_on_the_genomic_accession=25753415, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379033, CurrentID=0, Status='live', Symbol='XKRYP6', Aliases='', description='X-linked Kx blood group related, Y-linked pseudogene 6', other_designations='X Kell blood group precursor-related, Y-linked pseudogene 6|XK, Kell blood group complex subunit-related family, member 3 pseudogene|XK, Kell blood group complex subunit-related, Y-linked pseudogene 6', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25952590, end_position_on_the_genomic_accession=25954387, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  379034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=379034, CurrentID=0, Status='live', Symbol='LINC01181', Aliases='', description='long intergenic non-protein coding RNA 1181', other_designations='KB-1639H6.4', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103121032, end_position_on_the_genomic_accession=103140459, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379525, CurrentID=0, Status='live', Symbol='LOC105379525', Aliases='', description='uncharacterized LOC105379525', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  152286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152286, CurrentID=0, Status='live', Symbol='LOC152286', Aliases='', description='uncharacterized LOC152286', other_designations='', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927963, CurrentID=0, Status='live', Symbol='LOC101927963', Aliases='', description='uncharacterized LOC101927963', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141125940, end_position_on_the_genomic_accession=141130272, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379526, CurrentID=0, Status='live', Symbol='LOC105379526', Aliases='', description='uncharacterized LOC105379526', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928535, CurrentID=0, Status='live', Symbol='LOC101928535', Aliases='', description='uncharacterized LOC101928535', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106250019, end_position_on_the_genomic_accession=106264905, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100128333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128333, CurrentID=0, Status='live', Symbol='LOC100128333', Aliases='', description='uncharacterized LOC100128333', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=70090417, end_position_on_the_genomic_accession=70091286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379528, CurrentID=0, Status='live', Symbol='LOC105379528', Aliases='', description='Ig heavy chain V region 5A-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130353, CurrentID=0, Status='live', Symbol='LOC100130353', Aliases='', description='major facilitator superfamily domain containing 14B pseudogene', other_designations='hippocampus abundant transcript-like 1 pseudogene', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119402750, end_position_on_the_genomic_accession=119405388, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379529, CurrentID=0, Status='live', Symbol='LOC105379529', Aliases='', description='uncharacterized LOC105379529', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379530, CurrentID=0, Status='live', Symbol='LOC105379530', Aliases='', description='uncharacterized LOC105379530', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289568, CurrentID=0, Status='live', Symbol='LOC100289568', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A1, complex locus pseudogene', other_designations='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68972990, end_position_on_the_genomic_accession=68975995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116931, CurrentID=0, Status='live', Symbol='MED12L', Aliases='NOPAR, TNRC11L, TRALP, TRALPUSH', description='mediator complex subunit 12 like', other_designations='mediator of RNA polymerase II transcription, subunit 12 homolog-like|no opposite paired repeat protein|thyroid hormone receptor-associated-like protein|trinucleotide repeat-containing gene 11 protein-like', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151085665, end_position_on_the_genomic_accession=151436678, orientation='plus', exon_count=52, OMIM=611318, no_hdr0=''),
  116933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116933, CurrentID=0, Status='live', Symbol='CLRN1-AS1', Aliases='CLRN1OS, UCRP', description='CLRN1 antisense RNA 1', other_designations='CLRN1 antisense RNA 1 (non-protein coding)|Usher critical region protein pseudogene|clarin 1 opposite strand', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150972678, end_position_on_the_genomic_accession=151079830, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  85025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85025, CurrentID=0, Status='live', Symbol='TMEM60', Aliases='C7orf35, DC32', description='transmembrane protein 60', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77793728, end_position_on_the_genomic_accession=77798430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116936, CurrentID=0, Status='live', Symbol='RNU86', Aliases='U82, U86', description='RNA, U86 small nucleolar', other_designations='U82 small nucleolar RNA|U82 snoRNA|U86 snoRNA', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39316842, end_position_on_the_genomic_accession=39316896, orientation='minus', exon_count=1, OMIM=611069, no_hdr0=''),
  116937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116937, CurrentID=0, Status='live', Symbol='SNORD83A', Aliases='RNU83A, U83A', description='small nucleolar RNA, C/D box 83A', other_designations='RNA, U83A small nucleolar|U83a small nucleolar RNA|U83a snoRNA', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39315213, end_position_on_the_genomic_accession=39315307, orientation='minus', exon_count=1, OMIM=611070, no_hdr0=''),
  116938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116938, CurrentID=0, Status='live', Symbol='SNORD83B', Aliases='RNU83B, U83B', description='small nucleolar RNA, C/D box 83B', other_designations='RNA, U83B small nucleolar|U83b small nucleolar RNA|U83b snoRNA', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39313819, end_position_on_the_genomic_accession=39313911, orientation='minus', exon_count=1, OMIM=611071, no_hdr0=''),
  101927971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927971, CurrentID=0, Status='live', Symbol='LOC101927971', Aliases='', description='uncharacterized LOC101927971', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68395302, end_position_on_the_genomic_accession=68404259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379534, CurrentID=0, Status='live', Symbol='LOC105379534', Aliases='', description='proline-rich receptor-like protein kinase PERK2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101927972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927972, CurrentID=0, Status='live', Symbol='LOC101927972', Aliases='', description='uncharacterized LOC101927972', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5452468, end_position_on_the_genomic_accession=5458075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379535, CurrentID=0, Status='live', Symbol='LOC105379535', Aliases='', description='uncharacterized LOC105379535', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372254, CurrentID=0, Status='live', Symbol='LOC105372254', Aliases='', description='uncharacterized LOC105372254', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5692219, end_position_on_the_genomic_accession=5693257, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376508, CurrentID=0, Status='live', Symbol='LOC105376508', Aliases='', description='uncharacterized LOC105376508', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=697607, end_position_on_the_genomic_accession=706658, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  116966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116966, CurrentID=0, Status='live', Symbol='WDR17', Aliases='', description='WD repeat domain 17', other_designations='', map_location='4q34', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176065834, end_position_on_the_genomic_accession=176182828, orientation='plus', exon_count=33, OMIM=609005, no_hdr0=''),
  116969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116969, CurrentID=0, Status='live', Symbol='ART5', Aliases='ARTC5', description='ADP-ribosyltransferase 5', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 5|NAD(P)(+)--arginine ADP-ribosyltransferase 5|mono(ADP-ribosyl)transferase 5', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3638506, end_position_on_the_genomic_accession=3642316, orientation='minus', exon_count=6, OMIM=610625, no_hdr0=''),
  100132773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132773, CurrentID=0, Status='live', Symbol='LOC100132773', Aliases='', description='serine/threonine-protein phosphatase 4 regulatory subunit 2-like', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1614768, end_position_on_the_genomic_accession=1616469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105761008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105761008, CurrentID=0, Status='live', Symbol='TUC339', Aliases='', description='transcript uc.339', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420648, CurrentID=0, Status='live', Symbol='LOC100420648', Aliases='', description='Der1-like domain family, member 2 pseudogene', other_designations='derlin 2 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138047253, end_position_on_the_genomic_accession=138048097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372144, CurrentID=0, Status='live', Symbol='LOC105372144', Aliases='', description='uncharacterized LOC105372144', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58612583, end_position_on_the_genomic_accession=58615835, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101059998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101059998, CurrentID=0, Status='live', Symbol='LOC101059998', Aliases='', description='uncharacterized LOC101059998', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120844341, end_position_on_the_genomic_accession=120851350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116983, CurrentID=0, Status='live', Symbol='ACAP3', Aliases='CENTB5', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 3', other_designations='centaurin, beta 5|centaurin-beta-5|cnt-b5', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1292384, end_position_on_the_genomic_accession=1307925, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  116984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116984, CurrentID=0, Status='live', Symbol='ARAP2', Aliases='CENTD1, PARX', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 2', other_designations='Arf and Rho GAP adapter protein 2|centaurin, delta 1|centaurin-delta-1|cnt-d1', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36065998, end_position_on_the_genomic_accession=36244357, orientation='minus', exon_count=33, OMIM=606645, no_hdr0=''),
  116985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116985, CurrentID=0, Status='live', Symbol='ARAP1', Aliases='CENTD2', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 1', other_designations='ARF-GAP, RHO-GAP, ankyrin repeat, and pleckstrin homology domains-containing protein 1|centaurin, delta 2|centaurin-delta-2|cnt-d2', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72685069, end_position_on_the_genomic_accession=72752389, orientation='minus', exon_count=36, OMIM=606646, no_hdr0=''),
  116986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116986, CurrentID=0, Status='live', Symbol='AGAP2', Aliases='CENTG1, GGAP2, PIKE', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 2', other_designations='Arf GAP with GTP-binding protein-like, ANK repeat and PH domains 2|GTP-binding and GTPase activating protein 2|centaurin, gamma 1|cnt-g1|phosphatidylinositol 3-kinase enhancer|phosphatidylinositol-3-kinase enhancer|phosphoinositide 3-kinase enhancer', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57723761, end_position_on_the_genomic_accession=57742201, orientation='minus', exon_count=20, OMIM=605476, no_hdr0=''),
  116987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116987, CurrentID=0, Status='live', Symbol='AGAP1', Aliases='AGAP-1, CENTG2, GGAP1, cnt-g2', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 1', other_designations='Arf GAP with GTP-binding protein-like, ANK repeat and PH domains 1|GTP-binding and GTPase-activating protein 1|centaurin, gamma 2', map_location='2q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235494089, end_position_on_the_genomic_accession=236131800, orientation='plus', exon_count=22, OMIM=608651, no_hdr0=''),
  116988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=116988, CurrentID=0, Status='live', Symbol='AGAP3', Aliases='AGAP-3, CENTG3, CRAG, MRIP-1, cnt-g3', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 3', other_designations='CRAM-associated GTPase|CRMP (collapsin response mediator protein) associated|MR1-interacting protein|centaurin, gamma 3|centaurin-gamma-3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151085831, end_position_on_the_genomic_accession=151144436, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  105372707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372707, CurrentID=0, Status='live', Symbol='LOC105372707', Aliases='', description='uncharacterized LOC105372707', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62305432, end_position_on_the_genomic_accession=62308014, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379541, CurrentID=0, Status='live', Symbol='LOC105379541', Aliases='', description='Ig heavy chain V-III region VH26-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  267013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=267013, CurrentID=0, Status='live', Symbol='HLA-DPA3', Aliases='dJ1033B10.14', description='major histocompatibility complex, class II, DP alpha 3 (pseudogene)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33131197, end_position_on_the_genomic_accession=33131343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641293, CurrentID=0, Status='live', Symbol='RPL21P19', Aliases='RPL21_7_151', description='ribosomal protein L21 pseudogene 19', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204346757, end_position_on_the_genomic_accession=204347298, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420653, CurrentID=0, Status='live', Symbol='LOC100420653', Aliases='', description='guanosine monophosphate reductase pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103258570, end_position_on_the_genomic_accession=103259318, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379544, CurrentID=0, Status='live', Symbol='LOC105379544', Aliases='', description='uncharacterized LOC105379544', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641298, CurrentID=0, Status='live', Symbol='SMG1P1', Aliases='', description='SMG1 pseudogene 1', other_designations='PI-3-kinase-related kinase SMG-1|smg-1 homolog, phosphatidylinositol 3-kinase-related kinase pseudogene 1', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22437008, end_position_on_the_genomic_accession=22492220, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  105379545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379545, CurrentID=0, Status='live', Symbol='LOC105379545', Aliases='', description='uncharacterized LOC105379545', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372688, CurrentID=0, Status='live', Symbol='LOC105372688', Aliases='', description='uncharacterized LOC105372688', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57460474, end_position_on_the_genomic_accession=57463525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379546, CurrentID=0, Status='live', Symbol='LOC105379546', Aliases='', description='Ig heavy chain V-III region VH26-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641311, CurrentID=0, Status='live', Symbol='RPL31P11', Aliases='RPL31_1_108', description='ribosomal protein L31 pseudogene 11', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161683695, end_position_on_the_genomic_accession=161685252, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  641312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641312, CurrentID=0, Status='live', Symbol='SAR1AP2', Aliases='SAR1P2, SARA1P', description='SAR1 homolog A (S. cerevisiae) pseudogene 2', other_designations='SAR1 gene homolog pseudogene 2|SAR1 homolog pseudogene 2', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126272728, end_position_on_the_genomic_accession=126273322, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289584, CurrentID=0, Status='live', Symbol='NPM1P40', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 40', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203255734, end_position_on_the_genomic_accession=203256486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641314, CurrentID=0, Status='live', Symbol='LINC00594', Aliases='C15orf36', description='long intergenic non-protein coding RNA 594', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289585, CurrentID=0, Status='live', Symbol='LOC100289585', Aliases='', description='proliferation-inducing protein 25', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=73280258, end_position_on_the_genomic_accession=73280455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390876, CurrentID=0, Status='live', Symbol='LOC390876', Aliases='', description='ribosomal protein L35 pseudogene', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3130670, end_position_on_the_genomic_accession=3131437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379549, CurrentID=0, Status='live', Symbol='LOC105379549', Aliases='', description='uncharacterized LOC105379549', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129456, CurrentID=0, Status='live', Symbol='LOC100129456', Aliases='', description='heat shock protein 90kDa alpha family class B member 1 pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173575110, end_position_on_the_genomic_accession=173576341, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641339, CurrentID=0, Status='live', Symbol='ZNF674', Aliases='MRX92, ZNF673B', description='zinc finger protein 674', other_designations='zinc finger family member 674', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46497725, end_position_on_the_genomic_accession=46545459, orientation='minus', exon_count=9, OMIM=300573, no_hdr0=''),
  645482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645482, CurrentID=0, Status='live', Symbol='LOC645482', Aliases='', description='ATPase, H+ transporting, lysosomal 56/58kDa, V1 subunit B2 pseudogene', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121238676, end_position_on_the_genomic_accession=121240078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927989, CurrentID=0, Status='live', Symbol='LOC101927989', Aliases='', description='uncharacterized LOC101927989', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76491970, end_position_on_the_genomic_accession=76493918, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379553, CurrentID=0, Status='live', Symbol='LOC105379553', Aliases='', description='uncharacterized LOC105379553', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379554, CurrentID=0, Status='live', Symbol='LOC105379554', Aliases='', description='maFF-interacting protein', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372147, CurrentID=0, Status='live', Symbol='LOC105372147', Aliases='', description='uncharacterized LOC105372147', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58812696, end_position_on_the_genomic_accession=58834525, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379555, CurrentID=0, Status='live', Symbol='LOC105379555', Aliases='', description='uncharacterized LOC105379555', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641364, CurrentID=0, Status='live', Symbol='SLC7A11-AS1', Aliases='', description='SLC7A11 antisense RNA 1', other_designations='SLC7A11 antisense RNA 1 (non-protein coding)', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138089014, end_position_on_the_genomic_accession=138178177, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  641365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641365, CurrentID=0, Status='live', Symbol='LINC00616', Aliases='', description='long intergenic non-protein coding RNA 616', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138027422, end_position_on_the_genomic_accession=138130685, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  641367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641367, CurrentID=0, Status='live', Symbol='LOC641367', Aliases='', description='cyclin Y like 1 pseudogene', other_designations='cyclin Y-like 1 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21750745, end_position_on_the_genomic_accession=21753438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  641371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641371, CurrentID=0, Status='live', Symbol='ACOT1', Aliases='ACH2, CTE-1, LACH2', description='acyl-CoA thioesterase 1', other_designations='CTE-I|CTE-Ib|inducible cytosolic acyl-coenzyme A thioester hydrolase|long chain acyl-CoA hydrolase|long chain acyl-CoA thioester hydrolase', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73493234, end_position_on_the_genomic_accession=73543799, orientation='plus', exon_count=7, OMIM=614313, no_hdr0=''),
  117084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117084, CurrentID=0, Status='live', Symbol='CANDF1', Aliases='CMCT', description='candidiasis, familial 1', other_designations='candidiasis, familial chronic mucocutaneous, autosomal dominant, with thyroid disease', map_location='2p22.3-p21', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=114580, no_hdr0=''),
  641373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641373, CurrentID=0, Status='live', Symbol='NRON', Aliases='NCRNA00194', description='non-protein coding RNA, repressor of NFAT', other_designations='noncoding repressor of NFAT', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126407775, end_position_on_the_genomic_accession=126410504, orientation='minus', exon_count=1, OMIM=609618, no_hdr0=''),
  105379557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379557, CurrentID=0, Status='live', Symbol='LOC105379557', Aliases='', description='histone-lysine N-methyltransferase 2C-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641378, CurrentID=0, Status='live', Symbol='LOC641378', Aliases='', description='ADP ribosylation factor like GTPase 5A pseudogene', other_designations='ADP-ribosylation factor like GTPase 5A pseudogene|ADP-ribosylation factor-like 5A pseudogene', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76029488, end_position_on_the_genomic_accession=76031929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641379, CurrentID=0, Status='live', Symbol='ARL5AP4', Aliases='', description='ADP ribosylation factor like GTPase 5A pseudogene 4', other_designations='ADP-ribosylation factor like GTPase 5A pseudogene 4|ADP-ribosylation factor-like 5A pseudogene 4', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24685855, end_position_on_the_genomic_accession=24693184, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641380, CurrentID=0, Status='live', Symbol='ARL5AP2', Aliases='', description='ADP ribosylation factor like GTPase 5A pseudogene 2', other_designations='ADP-ribosylation factor like GTPase 5A pseudogene 2|ADP-ribosylation factor-like 5A pseudogene 2', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99011155, end_position_on_the_genomic_accession=99014222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379558, CurrentID=0, Status='live', Symbol='LOC105379558', Aliases='', description='uncharacterized LOC105379558', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641382, CurrentID=0, Status='live', Symbol='ARL5AP1', Aliases='', description='ADP ribosylation factor like GTPase 5A pseudogene 1', other_designations='ADP-ribosylation factor like GTPase 5A pseudogene 1|ADP-ribosylation factor-like 5A pseudogene 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119183482, end_position_on_the_genomic_accession=119187042, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641384, CurrentID=0, Status='live', Symbol='TMEM75', Aliases='', description='transmembrane protein 75', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127946559, end_position_on_the_genomic_accession=127948723, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101927997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927997, CurrentID=0, Status='live', Symbol='LOC101927997', Aliases='', description='proline-rich protein 23D1-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7525147, end_position_on_the_genomic_accession=7528605, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420670, CurrentID=0, Status='live', Symbol='NPAP1P8', Aliases='', description='nuclear pore associated protein 1 pseudogene 8', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82430118, end_position_on_the_genomic_accession=82431506, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379561, CurrentID=0, Status='live', Symbol='LOC105379561', Aliases='', description='uncharacterized LOC105379561', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420673, CurrentID=0, Status='live', Symbol='LOC100420673', Aliases='', description='RAB5A, member RAS oncogene family pseudogene', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4635345, end_position_on_the_genomic_accession=4636162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128748, CurrentID=0, Status='live', Symbol='RPL7AP13', Aliases='RPL7AL3, dJ348M17.1', description='ribosomal protein L7a pseudogene 13', other_designations='ribosomal protein L7a-like 3', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16684057, end_position_on_the_genomic_accession=16684405, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379566, CurrentID=0, Status='live', Symbol='LOC105379566', Aliases='', description='uncharacterized LOC105379566', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117143, CurrentID=0, Status='live', Symbol='TADA1', Aliases='ADA1, HFI1, STAF42L, hADA1, TADA1', description='transcriptional adaptor 1', other_designations='SPT3-associated factor 42 (STAF42)|transcriptional adapter 1-like protein|transcriptional adaptor 1 (HFI1 homolog, yeast)', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166856510, end_position_on_the_genomic_accession=166876417, orientation='minus', exon_count=8, OMIM=612763, no_hdr0=''),
  117144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117144, CurrentID=0, Status='live', Symbol='CATSPER1', Aliases='CATSPER, SPGF7', description='cation channel, sperm associated 1', other_designations='hCatSper|sperm ion channel|sperm-associated cation channel 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66016752, end_position_on_the_genomic_accession=66026519, orientation='minus', exon_count=12, OMIM=606389, no_hdr0=''),
  117145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117145, CurrentID=0, Status='live', Symbol='THEM4', Aliases='CTMP', description='thioesterase superfamily member 4', other_designations='C-terminal modulator protein|acyl-CoA thioesterase THEM4|carboxyl-terminal modulator protein', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151870866, end_position_on_the_genomic_accession=151909885, orientation='minus', exon_count=7, OMIM=606388, no_hdr0=''),
  101928543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928543, CurrentID=0, Status='live', Symbol='LOC101928543', Aliases='', description='uncharacterized LOC101928543', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1763693, end_position_on_the_genomic_accession=1781351, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105379567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379567, CurrentID=0, Status='live', Symbol='LOC105379567', Aliases='', description='uncharacterized LOC105379567', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117151, CurrentID=0, Status='live', Symbol='SLC31A1P1', Aliases='CTR1P, CTR1psi, SLC31A1P', description='solute carrier family 31 (copper transporter), member 1 pseudogene 1', other_designations='CRT1 pseudogene|solute carrier family 31 (copper transporters), member 1 pseudogene 1', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172599386, end_position_on_the_genomic_accession=172604183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  117153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117153, CurrentID=0, Status='live', Symbol='MIA2', Aliases='', description='melanoma inhibitory activity 2', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39233921, end_position_on_the_genomic_accession=39253371, orientation='plus', exon_count=6, OMIM=608001, no_hdr0=''),
  117154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117154, CurrentID=0, Status='live', Symbol='DACH2', Aliases='', description='dachshund family transcription factor 2', other_designations='', map_location='Xq21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86148451, end_position_on_the_genomic_accession=86832602, orientation='plus', exon_count=14, OMIM=300608, no_hdr0=''),
  117155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117155, CurrentID=0, Status='live', Symbol='CATSPER2', Aliases='', description='cation channel, sperm associated 2', other_designations='sperm ion channel', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43630562, end_position_on_the_genomic_accession=43648845, orientation='minus', exon_count=14, OMIM=607249, no_hdr0=''),
  117156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117156, CurrentID=0, Status='live', Symbol='SCGB3A2', Aliases='LU103, PNSP1, UGRP1, pnSP-1', description='secretoglobin, family 3A, member 2', other_designations='pneumo secretory protein 1|uteroglobin-related protein 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147878711, end_position_on_the_genomic_accession=147882193, orientation='plus', exon_count=3, OMIM=606531, no_hdr0=''),
  117157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117157, CurrentID=0, Status='live', Symbol='SH2D1B', Aliases='EAT2', description='SH2 domain containing 1B', other_designations='EAT-2|EWS/FLI1-activated transcript 2|SH2 domain-containing molecule EAT2', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162395266, end_position_on_the_genomic_accession=162412138, orientation='minus', exon_count=4, OMIM=608510, no_hdr0=''),
  117159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117159, CurrentID=0, Status='live', Symbol='DCD', Aliases='AIDD-1, DSEP, HCAP, PIF, DCD', description='dermcidin', other_designations='diffusible survival/evasion peptide|preproteolysin|proteolysis inducing factor|survival promoting peptide', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54644591, end_position_on_the_genomic_accession=54648493, orientation='minus', exon_count=6, OMIM=606634, no_hdr0=''),
  641451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641451, CurrentID=0, Status='live', Symbol='SNORA19', Aliases='ACA19', description='small nucleolar RNA, H/ACA box 19', other_designations='ACA19 small nucleolar RNA', map_location='10q26', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119060011, end_position_on_the_genomic_accession=119060138, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379570, CurrentID=0, Status='live', Symbol='LOC105379570', Aliases='', description='DLA class II histocompatibility antigen, DR-1 beta chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117166, CurrentID=0, Status='live', Symbol='WFIKKN1', Aliases='C16orf12, RJD2, WFDC20A, WFIKKN', description='WAP, follistatin/kazal, immunoglobulin, kunitz and netrin domain containing 1', other_designations='GASP-2|WAP four-disulfide core domain 20A|WAP, FS, Ig, KU, and NTR-containing protein|WAP, follistatin, immunoglobulin, kunitz and NTR domain-containing protein|growth and differentiation factor-associated serum protein 2|hGASP-2', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=631012, end_position_on_the_genomic_accession=634116, orientation='plus', exon_count=2, OMIM=608021, no_hdr0=''),
  641455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641455, CurrentID=0, Status='live', Symbol='POTEM', Aliases='ACT, P704P, POTE14beta', description='POTE ankyrin domain family member M', other_designations='ANKRD26-like family C member ENSP00000349402|POTE ankyrin domain family, member M|prostate-specific P704P', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18967434, end_position_on_the_genomic_accession=19003752, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  641456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641456, CurrentID=0, Status='live', Symbol='CRIP1P2', Aliases='', description='cysteine-rich protein 1 (intestinal) pseudogene 2', other_designations='', map_location='3p24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25686835, end_position_on_the_genomic_accession=25687261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641457, CurrentID=0, Status='live', Symbol='CRIP1P4', Aliases='', description='cysteine-rich protein 1 (intestinal) pseudogene 4', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24712231, end_position_on_the_genomic_accession=24712658, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  117177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117177, CurrentID=0, Status='live', Symbol='RAB3IP', Aliases='RABIN3, RABIN8', description='RAB3A interacting protein', other_designations='SSX2 interacting protein|SSX2-interacting protein|rab3A-interacting protein|rabin-3', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69738400, end_position_on_the_genomic_accession=69823204, orientation='plus', exon_count=16, OMIM=608686, no_hdr0=''),
  117178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117178, CurrentID=0, Status='live', Symbol='SSX2IP', Aliases='ADIP', description='synovial sarcoma, X breakpoint 2 interacting protein', other_designations='SSX2-interacting protein|afadin DIL domain-interacting protein', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84643707, end_position_on_the_genomic_accession=84690757, orientation='minus', exon_count=18, OMIM=608690, no_hdr0=''),
  641467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641467, CurrentID=0, Status='live', Symbol='TSC22D1-AS1', Aliases='', description='TSC22D1 antisense RNA 1', other_designations='TSC22D1 antisense RNA 1 (non-protein coding)', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44575896, end_position_on_the_genomic_accession=44580432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379573, CurrentID=0, Status='live', Symbol='LOC105379573', Aliases='', description='uncharacterized LOC105379573', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117187, CurrentID=0, Status='live', Symbol='ATOD3', Aliases='', description='Dermatitis, atopic, 3', other_designations='', map_location='20p', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605804, no_hdr0=''),
  117188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117188, CurrentID=0, Status='live', Symbol='ATOD5', Aliases='ATPD5', description='Dermatitis, atopic, 5', other_designations='', map_location='13q12-q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605844, no_hdr0=''),
  117189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117189, CurrentID=0, Status='live', Symbol='GRDX', Aliases='GD3', description='Graves disease, susceptibility to, X-linked', other_designations='', map_location='Xp11', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300351, no_hdr0=''),
  117190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117190, CurrentID=0, Status='live', Symbol='HPE6', Aliases='', description='holoprosencephaly 6', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=605934, no_hdr0=''),
  117191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117191, CurrentID=0, Status='live', Symbol='HYT1', Aliases='', description='Hypertension, essential, susceptibility to, 1', other_designations='', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603918, no_hdr0=''),
  117192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117192, CurrentID=0, Status='live', Symbol='MNRI', Aliases='', description='meningioma, radiation-induced', other_designations='', map_location='1p11', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606190, no_hdr0=''),
  117193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117193, CurrentID=0, Status='live', Symbol='THMA', Aliases='THM', description='Tibial hemimelia', other_designations='', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=275220, no_hdr0=''),
  117194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117194, CurrentID=0, Status='live', Symbol='MRGPRX2', Aliases='MGRG3, MRGX2', description='MAS related GPR family member X2', other_designations='G protein-coupled receptor MRGX2|MAS-related GPR, member X2|Mas-related G protein-coupled receptor G3', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19054455, end_position_on_the_genomic_accession=19071978, orientation='minus', exon_count=5, OMIM=607228, no_hdr0=''),
  117195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117195, CurrentID=0, Status='live', Symbol='MRGPRX3', Aliases='GPCR, MRGX3, SNSR1, SNSR2', description='MAS related GPR family member X3', other_designations='G protein-coupled receptor MRGX3|G protein-coupled receptor SNSR1|G protein-coupled receptor SNSR2|MAS-related GPR, member X3|mas-related GPCR member X3|sensory neuron-specific G-protein coupled receptor 1/2', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18103652, end_position_on_the_genomic_accession=18138491, orientation='plus', exon_count=4, OMIM=607229, no_hdr0=''),
  117196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117196, CurrentID=0, Status='live', Symbol='MRGPRX4', Aliases='GPCR, MRGX4, SNSR6', description='MAS related GPR family member X4', other_designations='G protein-coupled receptor MRGX4|G protein-coupled receptor SNSR5|G protein-coupled receptor SNSR6|MAS-related GPR, member X4|sensory neuron-specific G-protein coupled receptor 5/6', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18172837, end_position_on_the_genomic_accession=18174280, orientation='plus', exon_count=1, OMIM=607230, no_hdr0=''),
  101928013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928013, CurrentID=0, Status='live', Symbol='LOC101928013', Aliases='', description='uncharacterized LOC101928013', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93752987, end_position_on_the_genomic_accession=93847154, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128760, CurrentID=0, Status='live', Symbol='RPL21P3', Aliases='bA189K21.4', description='ribosomal protein L21 pseudogene 3', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18485225, end_position_on_the_genomic_accession=18486067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928545, CurrentID=0, Status='live', Symbol='SEPT7-AS1', Aliases='', description='SEPT7 antisense RNA 1 (head to head)', other_designations='', map_location='7p14.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35754964, end_position_on_the_genomic_accession=35800625, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372152, CurrentID=0, Status='live', Symbol='LOC105372152', Aliases='', description='uncharacterized LOC105372152', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60027149, end_position_on_the_genomic_accession=60044469, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  641510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641510, CurrentID=0, Status='live', Symbol='LOC641510', Aliases='', description='uncharacterized LOC641510', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641515, CurrentID=0, Status='live', Symbol='MGAT4EP', Aliases='', description='MGAT4 family member E, pseudogene', other_designations='MGAT4 family, member E, pseudogene|alpha-1,3-mannosyl-glycoprotein 4-beta-N-acetylglucosaminyltransferase-like protein pseudogene', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202820257, end_position_on_the_genomic_accession=202827225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  641516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641516, CurrentID=0, Status='live', Symbol='KC6', Aliases='', description='keratoconus gene 6', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41480271, end_position_on_the_genomic_accession=41520597, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  641517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641517, CurrentID=0, Status='live', Symbol='DEFB109P1B', Aliases='DEF109P1B, DEFB109', description='defensin beta 109 pseudogene 1B', other_designations='defensin, beta 109, pseudogene 1B|defensin, beta 9 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7312846, end_position_on_the_genomic_accession=7319951, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  641518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641518, CurrentID=0, Status='live', Symbol='LEF1-AS1', Aliases='LEF1NAT', description='LEF1 antisense RNA 1', other_designations='LEF1 antisense RNA 1 (non-protein coding)|LEF1 natural antisense transcript', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108167525, end_position_on_the_genomic_accession=108176430, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  117237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117237, CurrentID=0, Status='live', Symbol='SXI2', Aliases='', description='X inactivation, familial skewed, 2', other_designations='', map_location='Xq25-q26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300179, no_hdr0=''),
  117245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117245, CurrentID=0, Status='live', Symbol='HRASLS5', Aliases='HRLP5, HRSL5, RLP1, iNAT', description='HRAS like suppressor family member 5', other_designations='H-rev107-like protein 5|HRAS-like suppressor 5|HRAS-like suppressor family, member 5|calcium-independent phosphatidylethanolamine N-acyltransferase|lecithin-retinol acyltransferase (LRAT)-like protein-1', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63461404, end_position_on_the_genomic_accession=63491208, orientation='minus', exon_count=7, OMIM=611474, no_hdr0=''),
  117246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117246, CurrentID=0, Status='live', Symbol='FTSJ3', Aliases='EPCS3, SPB1', description='FtsJ homolog 3 (E. coli)', other_designations="2'-O-ribose RNA methyltransferase SPB1 homolog|SPB1 RNA methyltransferase homolog|protein ftsJ homolog 3|putative rRNA methyltransferase 3|rRNA (uridine-2'-O-)-methyltransferase 3", map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63819433, end_position_on_the_genomic_accession=63827671, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  117247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117247, CurrentID=0, Status='live', Symbol='SLC16A10', Aliases='MCT10, PRO0813, TAT1', description='solute carrier family 16 (aromatic amino acid transporter), member 10', other_designations='MCT 10|T-type amino acid transporter 1|aromatic amino acid transporter 1|solute carrier family 16 (monocarboxylic acid transporters), member 10|solute carrier family 16 member 10|solute carrier family 16, member 10 (aromatic amino acid transporter)', map_location='6q21-q22', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111087503, end_position_on_the_genomic_accession=111223405, orientation='plus', exon_count=10, OMIM=607550, no_hdr0=''),
  117248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117248, CurrentID=0, Status='live', Symbol='GALNT15', Aliases='GALNACT15, GALNT13, GALNT7, GALNTL2, PIH5', description='polypeptide N-acetylgalactosaminyltransferase 15', other_designations='UDP-GalNAc transferase T15|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 2|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 15|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 7|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 2|galNAc-T-like protein 2|polypeptide GalNAc transferase 15|polypeptide GalNAc transferase-like protein 2|polypeptide N-acetylgalactosaminyltransferase 13|polypeptide N-acetylgalactosaminyltransferase-like protein 2|pp-GaNTase-like protein 2|pp-GalNAc-T15|pregnancy-induced hypertension syndrome-related protein 5|protein-UDP acetylgalactosaminyltransferase-like protein 2', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16174579, end_position_on_the_genomic_accession=16246333, orientation='plus', exon_count=12, OMIM=615131, no_hdr0=''),
  105379585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379585, CurrentID=0, Status='live', Symbol='LOC105379585', Aliases='', description='uncharacterized LOC105379585', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100289623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289623, CurrentID=0, Status='live', Symbol='LOC100289623', Aliases='', description='ubiquitin-conjugating enzyme E2 variant 1 pseudogene', other_designations='', map_location='2q22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142543407, end_position_on_the_genomic_accession=142543972, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379586, CurrentID=0, Status='live', Symbol='LOC105379586', Aliases='', description='uncharacterized LOC105379586', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928547, CurrentID=0, Status='live', Symbol='LOC101928547', Aliases='', description='putative uncharacterized protein FLJ92257', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56825020, end_position_on_the_genomic_accession=56825698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641556, CurrentID=0, Status='live', Symbol='RPL36AP6', Aliases='RPL36A_12_1002, bA380I20.1', description='ribosomal protein L36a pseudogene 6', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107834457, end_position_on_the_genomic_accession=107834856, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379588, CurrentID=0, Status='live', Symbol='LOC105379588', Aliases='', description='uncharacterized LOC105379588', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379589, CurrentID=0, Status='live', Symbol='LOC105379589', Aliases='', description='uncharacterized protein FLJ40521-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372154, CurrentID=0, Status='live', Symbol='LOC105372154', Aliases='', description='uncharacterized LOC105372154', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60173198, end_position_on_the_genomic_accession=60177855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117283, CurrentID=0, Status='live', Symbol='IP6K3', Aliases='IHPK3, INSP6K3', description='inositol hexakisphosphate kinase 3', other_designations='ATP:1D-myo-inositol-hexakisphosphate phosphotransferase|InsP6 kinase 3|inositol hexaphosphate kinase 3', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33721637, end_position_on_the_genomic_accession=33746985, orientation='minus', exon_count=10, OMIM=606993, no_hdr0=''),
  117285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117285, CurrentID=0, Status='live', Symbol='DEFB118', Aliases='C20orf63, DEFB-18, ESC42, ESP13.6', description='defensin beta 118', other_designations='defensin, beta 118|defensin, beta 18|epididymal secretory protein 13.6|epididymus specific clone 42', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31368618, end_position_on_the_genomic_accession=31373923, orientation='plus', exon_count=2, OMIM=607650, no_hdr0=''),
  117286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117286, CurrentID=0, Status='live', Symbol='CIB3', Aliases='KIP3', description='calcium and integrin binding family member 3', other_designations='DNA-dependent protein kinase catalytic subunit-interacting protein 3', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16161368, end_position_on_the_genomic_accession=16173525, orientation='minus', exon_count=7, OMIM=610645, no_hdr0=''),
  117289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117289, CurrentID=0, Status='live', Symbol='TAGAP', Aliases='ARHGAP47, FKSG15, IDDM211, TAGAP', description='T-cell activation RhoGTPase activating protein', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159034468, end_position_on_the_genomic_accession=159045152, orientation='minus', exon_count=10, OMIM=609667, no_hdr0=''),
  641589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641589, CurrentID=0, Status='live', Symbol='LOC641589', Aliases='', description='neuroblastoma breakpoint family, member 19-like', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929125, CurrentID=0, Status='live', Symbol='LOC101929125', Aliases='', description='uncharacterized LOC101929125', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3808494, end_position_on_the_genomic_accession=3812434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928031, CurrentID=0, Status='live', Symbol='LOC101928031', Aliases='', description='uncharacterized LOC101928031', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=83522814, end_position_on_the_genomic_accession=83523482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645853, CurrentID=0, Status='live', Symbol='LOC645853', Aliases='', description='apolipoprotein A-I binding protein pseudogene', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178165669, end_position_on_the_genomic_accession=178166609, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379595, CurrentID=0, Status='live', Symbol='LOC105379595', Aliases='', description='olfactory receptor 11G2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131562, CurrentID=0, Status='live', Symbol='LOC100131562', Aliases='', description='chibby homolog 1 (Drosophila) pseudogene', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173478020, end_position_on_the_genomic_accession=173478400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929998, CurrentID=0, Status='live', Symbol='LOC101929998', Aliases='', description='uncharacterized LOC101929998', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154024916, end_position_on_the_genomic_accession=154066002, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105379596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379596, CurrentID=0, Status='live', Symbol='LOC105379596', Aliases='', description='uncharacterized LOC105379596', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641613, CurrentID=0, Status='live', Symbol='RPL21P5', Aliases='', description='ribosomal protein L21 pseudogene 5', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31111553, end_position_on_the_genomic_accession=31112283, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641614, CurrentID=0, Status='live', Symbol='RPL23AP8', Aliases='RPL23AP9, RPL23A_34_1351', description='ribosomal protein L23a pseudogene 8', other_designations='ribosomal protein L23a pseudogene 9', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34694155, end_position_on_the_genomic_accession=34694693, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641616, CurrentID=0, Status='live', Symbol='RPL9P3', Aliases='RPL9_8_1354', description='ribosomal protein L9 pseudogene 3', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35262433, end_position_on_the_genomic_accession=35263132, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100420707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420707, CurrentID=0, Status='live', Symbol='LOC100420707', Aliases='', description='DDB1 and CUL4 associated factor 6 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29233689, end_position_on_the_genomic_accession=29235137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653071, CurrentID=0, Status='live', Symbol='LOC653071', Aliases='', description='ankyrin repeat domain 57 pseudogene', other_designations='sosondowah ankyrin repeat domain family member C pseudogene', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928037, CurrentID=0, Status='live', Symbol='LOC101928037', Aliases='', description='uncharacterized LOC101928037', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141374183, end_position_on_the_genomic_accession=141382220, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379600, CurrentID=0, Status='live', Symbol='LOC105379600', Aliases='', description='uncharacterized LOC105379600', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641638, CurrentID=0, Status='live', Symbol='SNHG6', Aliases='HBII-276HG, NCRNA00058, U87HG', description='small nucleolar RNA host gene 6', other_designations='HBII-276 host|U87 host|small nucleolar RNA host gene (non-protein coding) 6|small nucleolar RNA host gene 6 (non-protein coding)', map_location='8q13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66921930, end_position_on_the_genomic_accession=66925542, orientation='minus', exon_count=4, OMIM=612215, no_hdr0=''),
  105379601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379601, CurrentID=0, Status='live', Symbol='LOC105379601', Aliases='', description='uncharacterized LOC105379601', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100125288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125288, CurrentID=0, Status='live', Symbol='ZGLP1', Aliases='GATAD3, GLP-1, GLP1', description='zinc finger, GATA-like protein 1', other_designations='GATA like protein 1|GATA like protein-1|GATA zinc finger domain containing 3|GATA-like protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10304803, end_position_on_the_genomic_accession=10309557, orientation='minus', exon_count=4, OMIM=611639, no_hdr0=''),
  100270862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270862, CurrentID=0, Status='live', Symbol='RPS3P4', Aliases='RPS3_1_652', description='ribosomal protein S3 pseudogene 4', other_designations='', map_location='6p25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7338761, end_position_on_the_genomic_accession=7339598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641648, CurrentID=0, Status='live', Symbol='SNORD87', Aliases='HBII-276, U87', description='small nucleolar RNA, C/D box 87', other_designations='C/D box snoRNA HBII-276|U87 snoRNA', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66922474, end_position_on_the_genomic_accession=66922549, orientation='minus', exon_count=1, OMIM=612216, no_hdr0=''),
  641649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641649, CurrentID=0, Status='live', Symbol='TMEM91', Aliases='DSPC3, IFITMD6', description='transmembrane protein 91', other_designations='dispanin subfamily C member 3|interferon induced transmembrane protein domain containing 6', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41363966, end_position_on_the_genomic_accession=41384083, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379603, CurrentID=0, Status='live', Symbol='LOC105379603', Aliases='', description='Ig lambda chain V-I region BL2-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641654, CurrentID=0, Status='live', Symbol='HEPN1', Aliases='', description='hepatocellular carcinoma, down-regulated 1', other_designations='HEPACAM opposite strand 1|cancer susceptibility gene HEPN1', map_location='11q24', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124919250, end_position_on_the_genomic_accession=124920677, orientation='plus', exon_count=1, OMIM=611641, no_hdr0=''),
  105372157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372157, CurrentID=0, Status='live', Symbol='LOC105372157', Aliases='', description='uncharacterized LOC105372157', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61652005, end_position_on_the_genomic_accession=61662445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928042, CurrentID=0, Status='live', Symbol='LOC101928042', Aliases='', description='uncharacterized LOC101928042', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32573958, end_position_on_the_genomic_accession=32614746, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  390933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390933, CurrentID=0, Status='live', Symbol='LOC390933', Aliases='', description='protein associated with topoisomerase II homolog 1 (yeast) pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40023377, end_position_on_the_genomic_accession=40025005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928043, CurrentID=0, Status='live', Symbol='LOC101928043', Aliases='', description='uncharacterized LOC101928043', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22023994, end_position_on_the_genomic_accession=22026048, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379606, CurrentID=0, Status='live', Symbol='LOC105379606', Aliases='', description='ice-structuring glycoprotein-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379607, CurrentID=0, Status='live', Symbol='LOC105379607', Aliases='', description='uncharacterized LOC105379607', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379608, CurrentID=0, Status='live', Symbol='LOC105379608', Aliases='', description='uncharacterized LOC105379608', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379609, CurrentID=0, Status='live', Symbol='LOC105379609', Aliases='', description='uncharacterized LOC105379609', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401442, CurrentID=0, Status='live', Symbol='LOC401442', Aliases='', description='uncharacterized LOC401442', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=738554, end_position_on_the_genomic_accession=740374, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372158, CurrentID=0, Status='live', Symbol='LOC105372158', Aliases='', description='uncharacterized LOC105372158', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61897669, end_position_on_the_genomic_accession=62018119, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130500, CurrentID=0, Status='live', Symbol='LOC100130500', Aliases='', description='S-phase kinase-associated protein 2, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96620552, end_position_on_the_genomic_accession=96621721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379610, CurrentID=0, Status='live', Symbol='LOC105379610', Aliases='', description='uncharacterized LOC105379610', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641695, CurrentID=0, Status='live', Symbol='LOC641695', Aliases='', description='reticulon 4 interacting protein 1 pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76137425, end_position_on_the_genomic_accession=76138612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641700, CurrentID=0, Status='live', Symbol='ECSCR', Aliases='ARIA, ECSM2', description='endothelial cell surface expressed chemotaxis and apoptosis regulator', other_designations='apoptosis regulator through modulating IAP expression|endothelial cell-specific molecule 2', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139454602, end_position_on_the_genomic_accession=139462743, orientation='minus', exon_count=11, OMIM=615736, no_hdr0=''),
  641702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641702, CurrentID=0, Status='live', Symbol='FAM138F', Aliases='F379', description='family with sequence similarity 138 member F', other_designations='F379 retina-specific protein|family with sequence similarity 138, member F', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=76220, end_position_on_the_genomic_accession=77690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379612, CurrentID=0, Status='live', Symbol='LOC105379612', Aliases='', description='uncharacterized LOC105379612', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641708, CurrentID=0, Status='live', Symbol='LOC641708', Aliases='', description='hydroxysteroid (17-beta) dehydrogenase 12 pseudogene', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157365988, end_position_on_the_genomic_accession=157366923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  106480747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480747, CurrentID=0, Status='live', Symbol='TUBB4BP5', Aliases='', description='tubulin beta 4B class IVb pseudogene 5', other_designations='tubulin, beta 4B class IVb pseudogene 5', map_location='4q27', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130370, CurrentID=0, Status='live', Symbol='LOC100130370', Aliases='', description='uncharacterized LOC100130370', other_designations='uncharacterized protein DKFZp761P0212', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81375498, end_position_on_the_genomic_accession=81386797, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101928052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928052, CurrentID=0, Status='live', Symbol='LOC101928052', Aliases='', description='uncharacterized LOC101928052', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=162022733, end_position_on_the_genomic_accession=162047550, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101928553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928553, CurrentID=0, Status='live', Symbol='LOC101928553', Aliases='', description='uncharacterized LOC101928553', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148876022, end_position_on_the_genomic_accession=148927123, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  641737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641737, CurrentID=0, Status='live', Symbol='LOC641737', Aliases='', description='uncharacterized LOC641737', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100125387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125387, CurrentID=0, Status='live', Symbol='HNRNPRP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein R pseudogene 1', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21623902, end_position_on_the_genomic_accession=21626376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100125392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125392, CurrentID=0, Status='live', Symbol='LOC100125392', Aliases='', description='TIMELESS interacting protein pseudogene', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53455216, end_position_on_the_genomic_accession=53457148, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100125393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125393, CurrentID=0, Status='live', Symbol='LOC100125393', Aliases='', description='TIMELESS interacting protein pseudogene', other_designations='', map_location='16q', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58091116, end_position_on_the_genomic_accession=58092990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641746, CurrentID=0, Status='live', Symbol='LOC641746', Aliases='', description='glycine cleavage system protein H (aminomethyl carrier) pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64582610, end_position_on_the_genomic_accession=64583751, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379620, CurrentID=0, Status='live', Symbol='LOC105379620', Aliases='', description='uncharacterized LOC105379620', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379621, CurrentID=0, Status='live', Symbol='LOC105379621', Aliases='', description='uncharacterized LOC105379621', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100125409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125409, CurrentID=0, Status='live', Symbol='LOC100125409', Aliases='', description='chromosome 4 open reading frame 27 pseudogene', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70113763, end_position_on_the_genomic_accession=70114308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376525, CurrentID=0, Status='live', Symbol='LOC105376525', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase family member 7-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3437496, end_position_on_the_genomic_accession=3447645, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100873792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873792, CurrentID=0, Status='live', Symbol='CBX3P4', Aliases='', description='chromobox 3 pseudogene 4', other_designations='chromobox homolog 3 pseudogene 4', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2787292, end_position_on_the_genomic_accession=2787582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641776, CurrentID=0, Status='live', Symbol='SPDYE14P', Aliases='', description='speedy/RINGO cell cycle regulator family member E14, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75310120, end_position_on_the_genomic_accession=75317417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100289662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100289662, CurrentID=0, Status='live', Symbol='COX6B1P5', Aliases='', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 5', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2234251, end_position_on_the_genomic_accession=2241909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  641790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641790, CurrentID=0, Status='live', Symbol='RPL31P39', Aliases='RPL31_11_842', description='ribosomal protein L31 pseudogene 39', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125300463, end_position_on_the_genomic_accession=125300904, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379628, CurrentID=0, Status='live', Symbol='LOC105379628', Aliases='', description='uncharacterized LOC105379628', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641808, CurrentID=0, Status='live', Symbol='FAM185BP', Aliases='FAM185B', description='family with sequence similarity 185 member B, pseudogene', other_designations='family with sequence similarity 185, member A pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77083886, end_position_on_the_genomic_accession=77122372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928067, CurrentID=0, Status='live', Symbol='LOC101928067', Aliases='', description='uncharacterized LOC101928067', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131983346, end_position_on_the_genomic_accession=131994971, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117531, CurrentID=0, Status='live', Symbol='TMC1', Aliases='DFNA36, DFNB11, DFNB7', description='transmembrane channel like 1', other_designations='transmembrane channel-like 1|transmembrane cochlear-expressed protein 1|transmembrane, cochlear expressed, 1', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72521801, end_position_on_the_genomic_accession=72836351, orientation='plus', exon_count=29, OMIM=606706, no_hdr0=''),
  117532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117532, CurrentID=0, Status='live', Symbol='TMC2', Aliases='C20orf145', description='transmembrane channel like 2', other_designations='transmembrane channel-like 2|transmembrane cochlear-expressed gene 2|transmembrane cochlear-expressed protein 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2536607, end_position_on_the_genomic_accession=2641869, orientation='plus', exon_count=21, OMIM=606707, no_hdr0=''),
  105379633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379633, CurrentID=0, Status='live', Symbol='LOC105379633', Aliases='', description='uncharacterized LOC105379633', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641842, CurrentID=0, Status='live', Symbol='SLC23A4P', Aliases='SLC23A4', description='solute carrier family 23 (nucleobase transporters), member 4, pseudogene', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135277217, end_position_on_the_genomic_accession=135313561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379635, CurrentID=0, Status='live', Symbol='LOC105379635', Aliases='', description='uncharacterized LOC105379635', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128820, CurrentID=0, Status='live', Symbol='CST9LP1', Aliases='CTES7C', description='cystatin 9-like pseudogene 1', other_designations='cystatin pseudogene', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23547764, end_position_on_the_genomic_accession=23550652, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379637, CurrentID=0, Status='live', Symbol='LOC105379637', Aliases='', description='collagen alpha-2(IV) chain-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928075, CurrentID=0, Status='live', Symbol='LOC101928075', Aliases='', description='uncharacterized LOC101928075', other_designations='CTD-2540L5.6', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70608798, end_position_on_the_genomic_accession=70641298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379638, CurrentID=0, Status='live', Symbol='LOC105379638', Aliases='', description='uncharacterized LOC105379638', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117579, CurrentID=0, Status='live', Symbol='RLN3', Aliases='H3, RXN3, ZINS4, insl7', description='relaxin 3', other_designations='insulin-like peptide 7|insulin-like peptide INSL7|prorelaxin H3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14028205, end_position_on_the_genomic_accession=14030971, orientation='plus', exon_count=2, OMIM=606855, no_hdr0=''),
  117581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117581, CurrentID=0, Status='live', Symbol='TWIST2', Aliases='AMS, BBRSAY, DERMO1, FFDD3, SETLSS, bHLHa39', description='twist family bHLH transcription factor 2', other_designations='class A basic helix-loop-helix protein 39|dermis-expressed protein 1|twist basic helix-loop-helix transcription factor 2|twist homolog 2|twist-related bHLH protein Dermo1', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238847934, end_position_on_the_genomic_accession=238910548, orientation='plus', exon_count=4, OMIM=607556, no_hdr0=''),
  117583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117583, CurrentID=0, Status='live', Symbol='PARD3B', Aliases='ALS2CR19, PAR3B, PAR3L, PAR3beta', description='par-3 family cell polarity regulator beta', other_designations='PAR3-L protein|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 19|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 19 protein|par-3 partitioning defective 3 homolog B|partitioning defective 3-like protein|partitioning-defective 3-beta', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204545793, end_position_on_the_genomic_accession=205620162, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  117584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117584, CurrentID=0, Status='live', Symbol='RFFL', Aliases='CARP-2, CARP2, FRING, RIFIFYLIN, RNF189, RNF34L', description='ring finger and FYVE-like domain containing E3 ubiquitin protein ligase', other_designations='FYVE-RING finger protein SAKURA|RING finger and FYVE-like domain-containing protein 1|RING finger protein 189|caspase regulator CARP2|caspases-8 and -10-associated RING finger protein 2|ring finger and FYVE-like domain containing 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35009112, end_position_on_the_genomic_accession=35089329, orientation='minus', exon_count=8, OMIM=609735, no_hdr0=''),
  105372339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372339, CurrentID=0, Status='live', Symbol='LOC105372339', Aliases='', description='uncharacterized LOC105372339', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23565585, end_position_on_the_genomic_accession=23569321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379641, CurrentID=0, Status='live', Symbol='LOC105379641', Aliases='', description='uncharacterized LOC105379641', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379642, CurrentID=0, Status='live', Symbol='LOC105379642', Aliases='', description='uncharacterized LOC105379642', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  284023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284023, CurrentID=0, Status='live', Symbol='LOC284023', Aliases='', description='uncharacterized LOC284023', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7913322, end_position_on_the_genomic_accession=7915947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379643, CurrentID=0, Status='live', Symbol='LOC105379643', Aliases='', description='uncharacterized LOC105379643', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928081, CurrentID=0, Status='live', Symbol='LOC101928081', Aliases='', description='uncharacterized LOC101928081', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163105505, end_position_on_the_genomic_accession=163119873, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117608, CurrentID=0, Status='live', Symbol='ZNF354B', Aliases='KID2', description='zinc finger protein 354B', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178859953, end_position_on_the_genomic_accession=178884434, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100420754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420754, CurrentID=0, Status='live', Symbol='SOCS5P1', Aliases='', description='suppressor of cytokine signaling 5 pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=86216785, end_position_on_the_genomic_accession=86217727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100125556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125556, CurrentID=0, Status='live', Symbol='FAM86JP', Aliases='', description='family with sequence similarity 86 member J, pseudogene', other_designations='family with sequence similarity 86, member J, pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125916601, end_position_on_the_genomic_accession=125930024, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  641922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641922, CurrentID=0, Status='live', Symbol='UNC93B2', Aliases='UNC93B2P', description='unc-93 homolog B2 pseudogene (C. elegans)', other_designations='Putative protein unc-93 homolog B1-like protein', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6855043, end_position_on_the_genomic_accession=6859830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  641924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641924, CurrentID=0, Status='live', Symbol='FAM86LP', Aliases='', description='family with sequence similarity 86 member L, pseudogene', other_designations='family with sequence similarity 86, member L, pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6930980, end_position_on_the_genomic_accession=6939613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100420758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420758, CurrentID=0, Status='live', Symbol='LOC100420758', Aliases='', description='poly(A) binding protein, nuclear 1 pseudogene', other_designations='PABII pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71419229, end_position_on_the_genomic_accession=71420983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379649, CurrentID=0, Status='live', Symbol='LOC105379649', Aliases='', description='uncharacterized LOC105379649', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641928, CurrentID=0, Status='live', Symbol='FLJ16734', Aliases='', description='uncharacterized LOC641928', other_designations='', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=153092312, end_position_on_the_genomic_accession=153092674, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101060020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060020, CurrentID=0, Status='live', Symbol='OLA1P3', Aliases='', description='Obg-like ATPase 1 pseudogene 3', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55870027, end_position_on_the_genomic_accession=55872602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379650, CurrentID=0, Status='live', Symbol='LOC105379650', Aliases='', description='killer cell immunoglobulin-like receptor 2DS1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379651, CurrentID=0, Status='live', Symbol='LOC105379651', Aliases='', description='uncharacterized LOC105379651', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379652, CurrentID=0, Status='live', Symbol='LOC105379652', Aliases='', description='DLA class II histocompatibility antigen, DR-1 beta chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379653, CurrentID=0, Status='live', Symbol='LOC105379653', Aliases='', description='uncharacterized LOC105379653', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420763, CurrentID=0, Status='live', Symbol='CKAP2P1', Aliases='', description='cytoskeleton associated protein 2 pseudogene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21110306, end_position_on_the_genomic_accession=21113432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928092, CurrentID=0, Status='live', Symbol='LOC101928092', Aliases='', description='uncharacterized LOC101928092', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=2566027, end_position_on_the_genomic_accession=2574621, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379655, CurrentID=0, Status='live', Symbol='LOC105379655', Aliases='', description='uncharacterized LOC105379655', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=641977, CurrentID=0, Status='live', Symbol='SEPT7P2', Aliases='SEPT13, SEPT7B', description='septin 7 pseudogene 2', other_designations='septin 13|septin 7B', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45723787, end_position_on_the_genomic_accession=45769018, orientation='minus', exon_count=13, OMIM=611563, no_hdr0=''),
  105372168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372168, CurrentID=0, Status='live', Symbol='LOC105372168', Aliases='', description='uncharacterized LOC105372168', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=65131920, end_position_on_the_genomic_accession=65144128, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  724065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=724065, CurrentID=0, Status='live', Symbol='LOC724065', Aliases='', description='chromosome 19 open reading frame 53 pseudogene', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111104933, end_position_on_the_genomic_accession=111105536, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642006, CurrentID=0, Status='live', Symbol='GUSBP10', Aliases='', description='glucuronidase, beta pseudogene 10', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57165670, end_position_on_the_genomic_accession=57180156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  642007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642007, CurrentID=0, Status='live', Symbol='GUSBP12', Aliases='', description='glucuronidase, beta pseudogene 12', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57200824, end_position_on_the_genomic_accession=57203669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100418606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418606, CurrentID=0, Status='live', Symbol='LOC100418606', Aliases='', description='MHC class I polypeptide-related sequence B pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128848, CurrentID=0, Status='live', Symbol='DKKL1P1', Aliases='DKKL2, dJ854E16.1', description='dickkopf-like 1 pseudogene 1', other_designations='dickkopf-like 1 (soggy) pseudogene|dickkopf-like 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31274053, end_position_on_the_genomic_accession=31274959, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390993, CurrentID=0, Status='live', Symbol='RPL7AP18', Aliases='RPL7A_1_5', description='ribosomal protein L7a pseudogene 18', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8057181, end_position_on_the_genomic_accession=8058487, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928103, CurrentID=0, Status='live', Symbol='LOC101928103', Aliases='', description='uncharacterized LOC101928103', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214810229, end_position_on_the_genomic_accession=214963605, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100420776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420776, CurrentID=0, Status='live', Symbol='CNTNAP3P6', Aliases='', description='contactin associated protein-like 3 pseudogene 6', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39778244, end_position_on_the_genomic_accession=39782549, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390996, CurrentID=0, Status='live', Symbol='CFL1P6', Aliases='CFLP6', description='cofilin 1 (non-muscle) pseudogene 6', other_designations='cofilin pseudogene 6', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10985830, end_position_on_the_genomic_accession=10991797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379669, CurrentID=0, Status='live', Symbol='LOC105379669', Aliases='', description='uncharacterized LOC105379669', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390998, CurrentID=0, Status='live', Symbol='LOC390998', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12220727, end_position_on_the_genomic_accession=12221101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101927473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927473, CurrentID=0, Status='live', Symbol='LOC101927473', Aliases='', description='uncharacterized LOC101927473', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91367552, end_position_on_the_genomic_accession=91380866, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379671, CurrentID=0, Status='live', Symbol='LOC105379671', Aliases='', description='uncharacterized LOC105379671', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928109, CurrentID=0, Status='live', Symbol='LOC101928109', Aliases='', description='uncharacterized LOC101928109', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134161923, end_position_on_the_genomic_accession=134172059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379673, CurrentID=0, Status='live', Symbol='LOC105379673', Aliases='', description='uncharacterized LOC105379673', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379674, CurrentID=0, Status='live', Symbol='LOC105379674', Aliases='', description='uncharacterized LOC105379674', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372171, CurrentID=0, Status='live', Symbol='LOC105372171', Aliases='', description='uncharacterized LOC105372171', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128859, CurrentID=0, Status='live', Symbol='BPIFB6', Aliases='BPIL3, LPLUNC6', description='BPI fold containing family B member 6', other_designations='BPI fold containing family B, member 6|bactericidal/permeability-increasing protein-like 3', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33030977, end_position_on_the_genomic_accession=33044047, orientation='plus', exon_count=16, OMIM=614110, no_hdr0=''),
  105379676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379676, CurrentID=0, Status='live', Symbol='LOC105379676', Aliases='', description='uncharacterized LOC105379676', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653149, CurrentID=0, Status='live', Symbol='NBPF6', Aliases='', description='neuroblastoma breakpoint family member 6', other_designations='neuroblastoma breakpoint family, member 6', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108421489, end_position_on_the_genomic_accession=108471903, orientation='plus', exon_count=22, OMIM=613996, no_hdr0=''),
  105379678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379678, CurrentID=0, Status='live', Symbol='LOC105379678', Aliases='', description='uncharacterized LOC105379678', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372172, CurrentID=0, Status='live', Symbol='LOC105372172', Aliases='', description='uncharacterized LOC105372172', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67767174, end_position_on_the_genomic_accession=67782176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060026, CurrentID=0, Status='live', Symbol='LOC101060026', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42896307, end_position_on_the_genomic_accession=42955499, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101927892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927892, CurrentID=0, Status='live', Symbol='LOC101927892', Aliases='', description='uncharacterized LOC101927892', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=137196244, end_position_on_the_genomic_accession=137201141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379680, CurrentID=0, Status='live', Symbol='LOC105379680', Aliases='', description='uncharacterized LOC105379680', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369221, CurrentID=0, Status='live', Symbol='LOC105369221', Aliases='', description='uncharacterized LOC105369221', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119723202, end_position_on_the_genomic_accession=119728800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379682, CurrentID=0, Status='live', Symbol='LOC105379682', Aliases='', description='uncharacterized LOC105379682', other_designations='CTC-229L21.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642131, CurrentID=0, Status='live', Symbol='LOC642131', Aliases='', description='putative V-set and immunoglobulin domain-containing-like protein IGHV4OR15-8', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22177502, end_position_on_the_genomic_accession=22185491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649873, CurrentID=0, Status='live', Symbol='LOC649873', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15506795, end_position_on_the_genomic_accession=15507538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391012, CurrentID=0, Status='live', Symbol='DYNLL1P3', Aliases='', description='dynein, light chain, LC8-type 1 pseudogene 3', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18513025, end_position_on_the_genomic_accession=18513395, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  117852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117852, CurrentID=0, Status='live', Symbol='TRIM78P', Aliases='TRIMP1', description='tripartite motif containing 78, pseudogene', other_designations='tripartite motif-containing pseudogene 1', map_location='11p15', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=5664412, end_position_on_the_genomic_accession=5687622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=117854, CurrentID=0, Status='live', Symbol='TRIM6', Aliases='RNF89', description='tripartite motif containing 6', other_designations='RING finger protein 89|tripartite motif-containing 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5596101, end_position_on_the_genomic_accession=5612958, orientation='plus', exon_count=9, OMIM=607564, no_hdr0=''),
  128869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128869, CurrentID=0, Status='live', Symbol='PIGU', Aliases='CDC91L1, GAB1', description='phosphatidylinositol glycan anchor biosynthesis class U', other_designations='CDC91 (cell division cycle 91, S. cerevisiae, homolog)-like 1|GPI transamidase component PIG-U|GPI transamidase subunit|cell division cycle 91-like 1 protein|cell division cycle protein 91-like 1|phosphatidylinositol glycan anchor biosynthesis, class U|protein CDC91-like 1', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34560542, end_position_on_the_genomic_accession=34677285, orientation='minus', exon_count=13, OMIM=608528, no_hdr0=''),
  653160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653160, CurrentID=0, Status='live', Symbol='LOC653160', Aliases='', description='uncharacterized LOC653160', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34975699, end_position_on_the_genomic_accession=34978706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379690, CurrentID=0, Status='live', Symbol='LOC105379690', Aliases='', description='uncharacterized LOC105379690', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391019, CurrentID=0, Status='live', Symbol='RPL29P6', Aliases='RPL29_1_20', description='ribosomal protein L29 pseudogene 6', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23244729, end_position_on_the_genomic_accession=23245345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379692, CurrentID=0, Status='live', Symbol='LOC105379692', Aliases='', description='uncharacterized LOC105379692', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401097, CurrentID=0, Status='live', Symbol='C3orf80', Aliases='', description='chromosome 3 open reading frame 80', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160223784, end_position_on_the_genomic_accession=160290143, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379694, CurrentID=0, Status='live', Symbol='LOC105379694', Aliases='', description='uncharacterized LOC105379694', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379695, CurrentID=0, Status='live', Symbol='LOC105379695', Aliases='', description='uncharacterized LOC105379695', other_designations='XXbac-BPG252P9.10', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642204, CurrentID=0, Status='live', Symbol='LOC642204', Aliases='', description='ankyrin repeat domain 20 family member A1 pseudogene', other_designations='ankyrin repeat domain 20 family, member A1 pseudogene|ankyrin repeat domain-containing protein 26-like', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  347333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347333, CurrentID=0, Status='live', Symbol='KRT8P14', Aliases='', description='keratin 8 pseudogene 14', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45632071, end_position_on_the_genomic_accession=45633810, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372779, CurrentID=0, Status='live', Symbol='LOC105372779', Aliases='', description='uncharacterized LOC105372779', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32306464, end_position_on_the_genomic_accession=32309154, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379698, CurrentID=0, Status='live', Symbol='LOC105379698', Aliases='', description='uncharacterized LOC105379698', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642236, CurrentID=0, Status='live', Symbol='FRG1JP', Aliases='', description='FSHD region gene 1 family member J, pseudogene', other_designations='FSHD region gene 1 pseudogene', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63832049, end_position_on_the_genomic_accession=63858641, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101928139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928139, CurrentID=0, Status='live', Symbol='LOC101928139', Aliases='', description='uncharacterized LOC101928139', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376541, CurrentID=0, Status='live', Symbol='LOC105376541', Aliases='', description='uncharacterized LOC105376541', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9004350, end_position_on_the_genomic_accession=9065242, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379702, CurrentID=0, Status='live', Symbol='LOC105379702', Aliases='', description='ice-structuring glycoprotein-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  150051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150051, CurrentID=0, Status='live', Symbol='LOC150051', Aliases='', description='homeobox protein cut-like 1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31559241, end_position_on_the_genomic_accession=31560487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642249, CurrentID=0, Status='live', Symbol='LOC642249', Aliases='', description='uncharacterized LOC642249', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379703, CurrentID=0, Status='live', Symbol='LOC105379703', Aliases='', description='uncharacterized LOC105379703', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642259, CurrentID=0, Status='live', Symbol='GAPDHP70', Aliases='GAPDHL15, GAPDL15', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 70', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 15, pseudogene', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88407977, end_position_on_the_genomic_accession=88409290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372177, CurrentID=0, Status='live', Symbol='LOC105372177', Aliases='', description='uncharacterized LOC105372177', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=69816896, end_position_on_the_genomic_accession=69835053, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  642262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642262, CurrentID=0, Status='live', Symbol='LOC642262', Aliases='', description='nuclear receptor coactivator 4 pseudogene', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139493841, end_position_on_the_genomic_accession=139497974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379705, CurrentID=0, Status='live', Symbol='LOC105379705', Aliases='', description='olfactory receptor 4F6-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100125918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100125918, CurrentID=0, Status='live', Symbol='LOC100125918', Aliases='', description='uncharacterized LOC100125918', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119703966, end_position_on_the_genomic_accession=119715901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642273, CurrentID=0, Status='live', Symbol='FAM110C', Aliases='', description='family with sequence similarity 110 member C', other_designations='family with sequence similarity 110, member C', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38814, end_position_on_the_genomic_accession=47115, orientation='minus', exon_count=4, OMIM=611395, no_hdr0=''),
  101928145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928145, CurrentID=0, Status='live', Symbol='LOC101928145', Aliases='', description='PNMA-like protein 2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46533669, end_position_on_the_genomic_accession=46534749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642280, CurrentID=0, Status='live', Symbol='ZNF876P', Aliases='', description='zinc finger protein 876, pseudogene', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=212600, end_position_on_the_genomic_accession=255985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379709, CurrentID=0, Status='live', Symbol='LOC105379709', Aliases='', description='uncharacterized LOC105379709', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642288, CurrentID=0, Status='live', Symbol='CSPG4P9', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 9', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642290, CurrentID=0, Status='live', Symbol='SLC25A1P5', Aliases='', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1 pseudogene 5', other_designations='', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27805427, end_position_on_the_genomic_accession=27806208, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379711, CurrentID=0, Status='live', Symbol='LOC105379711', Aliases='', description='uncharacterized LOC105379711', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642305, CurrentID=0, Status='live', Symbol='MAPRE1P2', Aliases='', description='MAPRE1 pseudogene 2', other_designations='', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33009267, end_position_on_the_genomic_accession=33011818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642306, CurrentID=0, Status='live', Symbol='RPL19P15', Aliases='RPL19_10_1014', description='ribosomal protein L19 pseudogene 15', other_designations='', map_location='9q34.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130651771, end_position_on_the_genomic_accession=130652464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642311, CurrentID=0, Status='live', Symbol='BMS1P15', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 15', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 15|BMS1 ribosome biogenesis factor pseudogene 15|Putative BMS1-like protein ENSP00000382360|Putative BMS1-like protein ENSP00000383048|Putative BMS1-like protein ENSP00000383088', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20153776, end_position_on_the_genomic_accession=20160826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642316, CurrentID=0, Status='live', Symbol='LOC642316', Aliases='', description='uncharacterized LOC642316', other_designations='', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=203313, end_position_on_the_genomic_accession=206392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642317, CurrentID=0, Status='live', Symbol='GAPDHP15', Aliases='GAPDH-Lp, GAPDP15, b55C20.4', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 15', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57967596, end_position_on_the_genomic_accession=57968738, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642319, CurrentID=0, Status='live', Symbol='TAGLN2P2', Aliases='', description='transgelin 2 pseudogene 2', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30249505, end_position_on_the_genomic_accession=30250119, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642320, CurrentID=0, Status='live', Symbol='RPL36AP31', Aliases='RPL36A_13_923', description='ribosomal protein L36a pseudogene 31', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86016758, end_position_on_the_genomic_accession=86017185, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642326, CurrentID=0, Status='live', Symbol='CCNYL3', Aliases='', description='cyclin Y-like 3', other_designations='Putative cyclin-Y-like protein 3', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=34257859, end_position_on_the_genomic_accession=34292628, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  391044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391044, CurrentID=0, Status='live', Symbol='PHBP3', Aliases='', description='prohibitin pseudogene 3', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58999676, end_position_on_the_genomic_accession=59000491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391045, CurrentID=0, Status='live', Symbol='SLC2A3P2', Aliases='', description='solute carrier family 2 (facilitated glucose transporter), member 3 pseudogene 2', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64982782, end_position_on_the_genomic_accession=64986108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642337, CurrentID=0, Status='live', Symbol='FTLP17', Aliases='', description='ferritin, light polypeptide pseudogene 17', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=104153329, end_position_on_the_genomic_accession=104154037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379718, CurrentID=0, Status='live', Symbol='LOC105379718', Aliases='', description='uncharacterized LOC105379718', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642345, CurrentID=0, Status='live', Symbol='MIR4500HG', Aliases='', description='MIR4500 host gene', other_designations='MIR4500 host gene (non-protein coding)|hCG1818123', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87443987, end_position_on_the_genomic_accession=87670963, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372180, CurrentID=0, Status='live', Symbol='LOC105372180', Aliases='', description='uncharacterized LOC105372180', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70205814, end_position_on_the_genomic_accession=70209778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379720, CurrentID=0, Status='live', Symbol='LOC105379720', Aliases='', description='uncharacterized LOC105379720', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642355, CurrentID=0, Status='live', Symbol='LOC642355', Aliases='', description='nucleoside diphosphate kinase, mitochondrial-like', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140435313, end_position_on_the_genomic_accession=140446806, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642357, CurrentID=0, Status='live', Symbol='EEF1GP5', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 5', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115702791, end_position_on_the_genomic_accession=115704195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126007, CurrentID=0, Status='live', Symbol='L1TD1P1', Aliases='', description='LINE-1 type transposase domain containing 1 pseudogene 1', other_designations='', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102016572, end_position_on_the_genomic_accession=102018788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126008, CurrentID=0, Status='live', Symbol='NUDT19P1', Aliases='', description='nudix hydrolase 19 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 19 pseudogene 1', map_location='Xq', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119220008, end_position_on_the_genomic_accession=119223245, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642361, CurrentID=0, Status='live', Symbol='LOC642361', Aliases='', description='uncharacterized LOC642361', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79825902, end_position_on_the_genomic_accession=79827602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379722, CurrentID=0, Status='live', Symbol='LOC105379722', Aliases='', description='uncharacterized LOC105379722', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642366, CurrentID=0, Status='live', Symbol='LOC642366', Aliases='', description='uncharacterized LOC642366', other_designations='CTD-2314G24.2', map_location='5q11.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51372736, end_position_on_the_genomic_accession=51383332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379723, CurrentID=0, Status='live', Symbol='LOC105379723', Aliases='', description='probable C-mannosyltransferase DPY19L1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642381, CurrentID=0, Status='live', Symbol='LOC642381', Aliases='', description='speckle type POZ protein like pseudogene', other_designations='speckle-type POZ protein-like pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68995339, end_position_on_the_genomic_accession=68997711, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101060035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060035, CurrentID=0, Status='live', Symbol='LOC101060035', Aliases='', description='dnaJ homolog subfamily C member 24 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133720495, end_position_on_the_genomic_accession=133720934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379725, CurrentID=0, Status='live', Symbol='LOC105379725', Aliases='', description='uncharacterized LOC105379725', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928163, CurrentID=0, Status='live', Symbol='LOC101928163', Aliases='', description='uncharacterized LOC101928163', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23521930, end_position_on_the_genomic_accession=23528419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642389, CurrentID=0, Status='live', Symbol='RBM17P2', Aliases='', description='RNA binding motif protein 17 pseudogene 2', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67211437, end_position_on_the_genomic_accession=67213123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642394, CurrentID=0, Status='live', Symbol='ADARB2-AS1', Aliases='C10orf109, NCRNA00168, bA466B20.1', description='ADARB2 antisense RNA 1', other_designations='ADARB2 antisense RNA 1 (non-protein coding)|hCG1773937', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1526630, end_position_on_the_genomic_accession=1556984, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642395, CurrentID=0, Status='live', Symbol='PSPC1P1', Aliases='', description='paraspeckle component 1 pseudogene 1', other_designations='paraspeckle protein 1 pseudogene', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19201637, end_position_on_the_genomic_accession=19227432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392375, CurrentID=0, Status='live', Symbol='OR13D2P', Aliases='', description='olfactory receptor family 13 subfamily D member 2 pseudogene', other_designations='olfactory receptor, family 13, subfamily D, member 2 pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104590018, end_position_on_the_genomic_accession=104590951, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928165, CurrentID=0, Status='live', Symbol='CASC17', Aliases='LINC00600', description='cancer susceptibility candidate 17 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 600', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=71097774, end_position_on_the_genomic_accession=71202179, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  642402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642402, CurrentID=0, Status='live', Symbol='GOLGA6L17P', Aliases='GOLGA6L21P', description='golgin A6 family-like 17, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 6 pseudogene|golgin A6 family, member A pseudogene|golgin A6 family-like 21, pseudogene', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82519590, end_position_on_the_genomic_accession=82526484, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  150759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150759, CurrentID=0, Status='live', Symbol='LINC00342', Aliases='NCRNA00342', description='long intergenic non-protein coding RNA 342', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95807052, end_position_on_the_genomic_accession=95826981, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642413, CurrentID=0, Status='live', Symbol='CTSLP6', Aliases='CTSL1P6, CTSLL6, bA479O17.1', description='cathepsin L pseudogene 6', other_designations='cathepsin L-like 6|cathepsin L1 pseudogene 6', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79870051, end_position_on_the_genomic_accession=79873915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642414, CurrentID=0, Status='live', Symbol='LOC642414', Aliases='', description='tripartite motif containing 64B pseudogene', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89820082, end_position_on_the_genomic_accession=89820493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391059, CurrentID=0, Status='live', Symbol='FRRS1', Aliases='SDFR2, SDR2', description='ferric-chelate reductase 1', other_designations='SDR-2|stromal cell derived factor receptor 2 homolog|stromal cell-derived receptor 2', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99708632, end_position_on_the_genomic_accession=99766630, orientation='minus', exon_count=20, OMIM=611578, no_hdr0=''),
  642422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642422, CurrentID=0, Status='live', Symbol='LOC642422', Aliases='', description='uncharacterized LOC642422', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1339880, end_position_on_the_genomic_accession=1341755, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928169, CurrentID=0, Status='live', Symbol='LOC101928169', Aliases='', description='uncharacterized LOC101928169', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95351251, end_position_on_the_genomic_accession=95352653, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642424, CurrentID=0, Status='live', Symbol='IGKV1OR10-1', Aliases='', description='immunoglobulin kappa variable 1/OR10-1 (pseudogene)', other_designations='IGKV1/OR10-1|immunoglobulin heavy variable 1/OR10-1 (pseudogene)|immunoglobulin heavy variable 1/OR21-1 (pseudogene)', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642425, CurrentID=0, Status='live', Symbol='TRIM53BP', Aliases='TRIM53B, TRIM53L, TRIM53L1, TRIM80P', description='tripartite motif containing 53B, pseudogene', other_designations='tripartite motif containing 80, pseudogene|tripartite motif-containing 49 pseudogene|tripartite motif-containing 53B', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89841993, end_position_on_the_genomic_accession=89851264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642426, CurrentID=0, Status='live', Symbol='LOC642426', Aliases='', description='uncharacterized LOC642426', other_designations='', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18630536, end_position_on_the_genomic_accession=18633634, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100420843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420843, CurrentID=0, Status='live', Symbol='KNOP1P4', Aliases='', description='lysine-rich nucleolar protein 1 pseudogene 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72986341, end_position_on_the_genomic_accession=72987947, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130394, CurrentID=0, Status='live', Symbol='LOC100130394', Aliases='', description='translation machinery associated 7 homolog (S. cerevisiae) pseudogene', other_designations='CCDC72 pseudogene|translational machinery associated 7 homolog pseudogene', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172656359, end_position_on_the_genomic_accession=172656734, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642441, CurrentID=0, Status='live', Symbol='LOC642441', Aliases='', description='uncharacterized LOC642441', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642443, CurrentID=0, Status='live', Symbol='ADH5P4', Aliases='ADH5B, RCADH5, bA707M13.2', description='alcohol dehydrogenase 5 (class III), chi polypeptide, pseudogene 4', other_designations='alcohol dehydrogenase 5B', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=65835477, end_position_on_the_genomic_accession=65838068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642446, CurrentID=0, Status='live', Symbol='TRIM64B', Aliases='', description='tripartite motif containing 64B', other_designations='tripartite motif-containing 64B', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89870127, end_position_on_the_genomic_accession=89878437, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  642448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642448, CurrentID=0, Status='live', Symbol='KRT18P23', Aliases='', description='keratin 18 pseudogene 23', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44566964, end_position_on_the_genomic_accession=44568305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642451, CurrentID=0, Status='live', Symbol='RPL7P46', Aliases='RPL7_26_1472', description='ribosomal protein L7 pseudogene 46', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11026961, end_position_on_the_genomic_accession=11028561, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642452, CurrentID=0, Status='live', Symbol='LOC642452', Aliases='', description='uncharacterized LOC642452', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=88996354, end_position_on_the_genomic_accession=88996941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928175, CurrentID=0, Status='live', Symbol='LOC101928175', Aliases='', description='uncharacterized LOC101928175', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=68327112, end_position_on_the_genomic_accession=68338808, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  642460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642460, CurrentID=0, Status='live', Symbol='ANKRD30BP1', Aliases='ANKRD30BP, C21orf110, PRED4', description='ankyrin repeat domain 30B pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13383842, end_position_on_the_genomic_accession=13427775, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642461, CurrentID=0, Status='live', Symbol='IARS2P1', Aliases='', description='isoleucyl-tRNA synthetase 2, mitochondrial pseudogene 1', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87598481, end_position_on_the_genomic_accession=87601619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347376, CurrentID=0, Status='live', Symbol='H3F3AP5', Aliases='p55', description='H3 histone, family 3A, pseudogene 5', other_designations='H3 histone, family 3A pseudogene', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50905355, end_position_on_the_genomic_accession=50906368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379739, CurrentID=0, Status='live', Symbol='LOC105379739', Aliases='', description='speedy protein E2-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642474, CurrentID=0, Status='live', Symbol='LOC642474', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69181660, end_position_on_the_genomic_accession=69182377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642475, CurrentID=0, Status='live', Symbol='MROH6', Aliases='C8orf73', description='maestro heat-like repeat family member 6', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143566189, end_position_on_the_genomic_accession=143572758, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  642477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642477, CurrentID=0, Status='live', Symbol='LINC01296', Aliases='', description='long intergenic non-protein coding RNA 1296', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19062377, end_position_on_the_genomic_accession=19107497, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105379741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379741, CurrentID=0, Status='live', Symbol='LOC105379741', Aliases='', description='uncharacterized LOC105379741', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100420851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100420851, CurrentID=0, Status='live', Symbol='LOC100420851', Aliases='', description='nitric oxide synthase 2, inducible pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36161607, end_position_on_the_genomic_accession=36227070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642484, CurrentID=0, Status='live', Symbol='LOC642484', Aliases='', description='uncharacterized LOC642484', other_designations='CTD-2008L17.2', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56003290, end_position_on_the_genomic_accession=56027819, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379742, CurrentID=0, Status='live', Symbol='LOC105379742', Aliases='', description='uncharacterized LOC105379742', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642487, CurrentID=0, Status='live', Symbol='LOC642487', Aliases='', description='H1 histone family, member O, oocyte-specific pseudogene', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69535638, end_position_on_the_genomic_accession=69536539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642489, CurrentID=0, Status='live', Symbol='FKBP1C', Aliases='bA184C23.2', description='FK506 binding protein 1C', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63211447, end_position_on_the_genomic_accession=63213022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642490, CurrentID=0, Status='live', Symbol='LOC642490', Aliases='', description='S-phase kinase-associated protein 1 pseudogene', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66879213, end_position_on_the_genomic_accession=66879924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642493, CurrentID=0, Status='live', Symbol='RPL29P25', Aliases='RPL29_15_1277', description='ribosomal protein L29 pseudogene 25', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110841557, end_position_on_the_genomic_accession=110842004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347381, CurrentID=0, Status='live', Symbol='LOC347381', Aliases='', description='hydroxyacyl-CoA dehydrogenase/3-ketoacyl-CoA thiolase/enoyl-CoA hydratase (trifunctional protein), beta subunit pseudogene', other_designations='hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), beta subunit pseudogene', map_location='Xp22.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=5305180, end_position_on_the_genomic_accession=5307179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642496, CurrentID=0, Status='live', Symbol='LOC642496', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A3 pseudogene', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69346610, end_position_on_the_genomic_accession=69358176, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414201, CurrentID=0, Status='live', Symbol='ZNF32-AS3', Aliases='C10orf44, ZNF32-OT1, bA402L1.3', description='ZNF32 antisense RNA 3', other_designations='ZNF32 antisense RNA 3 (non-protein coding)|ZNF32 overlapping transcript 1 (non-protein coding)', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43628817, end_position_on_the_genomic_accession=43674699, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  642502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642502, CurrentID=0, Status='live', Symbol='LOC642502', Aliases='', description='succinate dehydrogenase complex, subunit C, integral membrane protein, 15kDa pseudogene', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1857280, end_position_on_the_genomic_accession=1858461, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379745, CurrentID=0, Status='live', Symbol='LOC105379745', Aliases='', description='uncharacterized LOC105379745', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  285352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285352, CurrentID=0, Status='live', Symbol='KIF9-AS1', Aliases='', description='KIF9 antisense RNA 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47164370, end_position_on_the_genomic_accession=47244116, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  642513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642513, CurrentID=0, Status='live', Symbol='KCTD9P6', Aliases='', description='potassium channel tetramerization domain containing 9 pseudogene 6', other_designations='potassium channel tetramerisation domain containing 9 pseudogene', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31220015, end_position_on_the_genomic_accession=31220970, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379747, CurrentID=0, Status='live', Symbol='LOC105379747', Aliases='', description='uncharacterized LOC105379747', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642517, CurrentID=0, Status='live', Symbol='AGAP9', Aliases='AGAP-9, CTGLF6, bA301J7.2', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 9', other_designations='centaurin, gamma-like family, member 6|centaurin-gamma-like family member 6', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47501854, end_position_on_the_genomic_accession=47523638, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  391076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391076, CurrentID=0, Status='live', Symbol='HSD3BP1', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta, pseudogene 1', other_designations='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 1 pseudogene', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119467134, end_position_on_the_genomic_accession=119473999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379749, CurrentID=0, Status='live', Symbol='LOC105379749', Aliases='', description='uncharacterized LOC105379749', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642531, CurrentID=0, Status='live', Symbol='RPL36P11', Aliases='RPL36_1_620', description='ribosomal protein L36 pseudogene 11', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140609085, end_position_on_the_genomic_accession=140609403, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642533, CurrentID=0, Status='live', Symbol='LOC642533', Aliases='', description='uncharacterized LOC642533', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642538, CurrentID=0, Status='live', Symbol='C1DP2', Aliases='bA369J21.3', description='C1D nuclear receptor corepressor pseudogene 2', other_designations='C1D nuclear receptor co-repressor pseudogene 2', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80031610, end_position_on_the_genomic_accession=80032472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369637, CurrentID=0, Status='live', Symbol='LOC105369637', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101928189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928189, CurrentID=0, Status='live', Symbol='LOC101928189', Aliases='', description='uncharacterized LOC101928189', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25302735, end_position_on_the_genomic_accession=25338221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  642546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642546, CurrentID=0, Status='live', Symbol='HK2P1', Aliases='HK2P', description='hexokinase 2 pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80569434, end_position_on_the_genomic_accession=80575038, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642550, CurrentID=0, Status='live', Symbol='LOC642550', Aliases='', description='v-ral simian leukemia viral oncogene homolog A (ras related) pseudogene', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78793524, end_position_on_the_genomic_accession=78794114, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379753, CurrentID=0, Status='live', Symbol='LOC105379753', Aliases='', description='uncharacterized LOC105379753', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642554, CurrentID=0, Status='live', Symbol='LOC642554', Aliases='', description='leucine rich repeat containing 59 pseudogene', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63441223, end_position_on_the_genomic_accession=63443913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379754, CurrentID=0, Status='live', Symbol='LOC105379754', Aliases='', description='uncharacterized LOC105379754', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642559, CurrentID=0, Status='live', Symbol='POU5F1P3', Aliases='OCT4-pg3, OTF3L, POU5F1L, POU5FLC12', description='POU class 5 homeobox 1 pseudogene 3', other_designations='POU domain class 5, transcription factor 1 pseudogene 3|POU domain transcription factor OCT4-pg3', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8133769, end_position_on_the_genomic_accession=8134852, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101928193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928193, CurrentID=0, Status='live', Symbol='LOC101928193', Aliases='', description='uncharacterized LOC101928193', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133189786, end_position_on_the_genomic_accession=133192966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642569, CurrentID=0, Status='live', Symbol='TRIM53AP', Aliases='TRIM53, TRIM53P, TRIM80P', description='tripartite motif containing 53A, pseudogene', other_designations='tripartite motif containing 53, pseudogene|tripartite motif-containing 49 pseudogene|tripartite motif-containing 53', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89993541, end_position_on_the_genomic_accession=90002508, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  642574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642574, CurrentID=0, Status='live', Symbol='LOC642574', Aliases='', description='IQ motif and ankyrin repeat domain-containing protein LOC642574', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143766664, end_position_on_the_genomic_accession=143790701, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  642579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642579, CurrentID=0, Status='live', Symbol='LOC642579', Aliases='', description='tripartite motif containing 64B pseudogene', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90024098, end_position_on_the_genomic_accession=90024509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642580, CurrentID=0, Status='live', Symbol='HSPA8P14', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 14', other_designations='heat shock 70kDa protein 8 pseudogene 14', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111381783, end_position_on_the_genomic_accession=111385680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642585, CurrentID=0, Status='live', Symbol='VDAC1P1', Aliases='VDAC1P', description='voltage-dependent anion channel 1 pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80929424, end_position_on_the_genomic_accession=80931207, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642587, CurrentID=0, Status='live', Symbol='MIR205HG', Aliases='LINC00510', description='MIR205 host gene', other_designations='MIR205 host gene (non-protein coding)|NPC-A-5|long intergenic non-protein coding RNA 510', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209428823, end_position_on_the_genomic_accession=209432552, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372188, CurrentID=0, Status='live', Symbol='LOC105372188', Aliases='', description='uncharacterized LOC105372188', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71197958, end_position_on_the_genomic_accession=71203688, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  642592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642592, CurrentID=0, Status='live', Symbol='EIF5AP4', Aliases='EIF5AL3', description='eukaryotic translation initiation factor 5A pseudogene 4', other_designations='eukaryotic translation initiation factor 5A-like 3', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80246821, end_position_on_the_genomic_accession=80247797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379760, CurrentID=0, Status='live', Symbol='LOC105379760', Aliases='', description='olfactory receptor 7E24-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642597, CurrentID=0, Status='live', Symbol='C18orf42', Aliases='', description='chromosome 18 open reading frame 42', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5143673, end_position_on_the_genomic_accession=5198203, orientation='minus', exon_count=3, OMIM=616427, no_hdr0=''),
  105379762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379762, CurrentID=0, Status='live', Symbol='LOC105379762', Aliases='', description='uncharacterized LOC105379762', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369139, CurrentID=0, Status='live', Symbol='LOC105369139', Aliases='', description='killer cell immunoglobulin-like receptor 2DS3', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642612, CurrentID=0, Status='live', Symbol='TRIM49C', Aliases='TRIM49L2', description='tripartite motif containing 49C', other_designations='Tripartite motif-containing protein LOC642612|tripartite motif containing 49-like 2|tripartite motif-containing protein 49-like protein 2', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90030149, end_position_on_the_genomic_accession=90042029, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105369141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369141, CurrentID=0, Status='live', Symbol='LOC105369141', Aliases='', description='uncharacterized LOC105369141', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161002285, end_position_on_the_genomic_accession=161021003, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  642614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642614, CurrentID=0, Status='live', Symbol='C2orf69P2', Aliases='', description='chromosome 2 open reading frame 69 pseudogene 2', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35492323, end_position_on_the_genomic_accession=35494217, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369143, CurrentID=0, Status='live', Symbol='LOC105369143', Aliases='', description='uncharacterized LOC105369143', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176673224, end_position_on_the_genomic_accession=176680305, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369144, CurrentID=0, Status='live', Symbol='LOC105369144', Aliases='', description='uncharacterized LOC105369144', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34064449, end_position_on_the_genomic_accession=34115521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379764, CurrentID=0, Status='live', Symbol='LOC105379764', Aliases='', description='uncharacterized LOC105379764', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369146, CurrentID=0, Status='live', Symbol='LOC105369146', Aliases='', description='cadherin-related family member 4-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81472689, end_position_on_the_genomic_accession=81570356, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369147, CurrentID=0, Status='live', Symbol='LOC105369147', Aliases='', description='uncharacterized LOC105369147', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103224446, end_position_on_the_genomic_accession=103298433, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  642620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642620, CurrentID=0, Status='live', Symbol='LOC642620', Aliases='', description='uncharacterized LOC642620', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=157346819, end_position_on_the_genomic_accession=157347702, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369149, CurrentID=0, Status='live', Symbol='LOC105369149', Aliases='', description='uncharacterized LOC105369149', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9780376, end_position_on_the_genomic_accession=9792120, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100126270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126270, CurrentID=0, Status='live', Symbol='FMR1-AS1', Aliases='ASFMR1, FMR1-AS, FMR1AS, FMR4', description='FMR1 antisense RNA 1', other_designations='FMR1 antisense RNA 1 (non-protein coding)|antisense fragile X mental retardation protein variant A', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147909431, end_position_on_the_genomic_accession=147922158, orientation='minus', exon_count=4, OMIM=300805, no_hdr0=''),
  642623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642623, CurrentID=0, Status='live', Symbol='UBTFL1', Aliases='C11orf27, HMGPI', description='upstream binding transcription factor, RNA polymerase I-like 1', other_designations='putative upstream-binding factor 1-like protein 1', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90085950, end_position_on_the_genomic_accession=90087131, orientation='plus', exon_count=1, OMIM=613696, no_hdr0=''),
  105369152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369152, CurrentID=0, Status='live', Symbol='LOC105369152', Aliases='', description='uncharacterized LOC105369152', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44859708, end_position_on_the_genomic_accession=44882614, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369153, CurrentID=0, Status='live', Symbol='LOC105369153', Aliases='', description='endogenous retrovirus group K member 25 Env polyprotein-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45232945, end_position_on_the_genomic_accession=45275437, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369154, CurrentID=0, Status='live', Symbol='LOC105369154', Aliases='', description='polycystin-1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15136205, end_position_on_the_genomic_accession=15154766, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  642627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642627, CurrentID=0, Status='live', Symbol='MYL6P4', Aliases='', description='myosin light chain 6 pseudogene 4', other_designations='myosin, light chain 6 pseudogene 4|myosin, light chain 6, alkali, smooth muscle and non-muscle pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141811910, end_position_on_the_genomic_accession=141812257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369156, CurrentID=0, Status='live', Symbol='LOC105369156', Aliases='', description='uncharacterized LOC105369156', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24516883, end_position_on_the_genomic_accession=24541831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379766, CurrentID=0, Status='live', Symbol='LOC105379766', Aliases='', description='small nuclear ribonucleoprotein-associated protein N-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369158, CurrentID=0, Status='live', Symbol='LOC105369158', Aliases='', description='uncharacterized LOC105369158', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58289905, end_position_on_the_genomic_accession=58305718, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642631, CurrentID=0, Status='live', Symbol='TSPY15P', Aliases='', description='testis specific protein, Y-linked 15, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9548062, end_position_on_the_genomic_accession=9550873, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369160, CurrentID=0, Status='live', Symbol='LOC105369160', Aliases='', description='uncharacterized LOC105369160', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41469188, end_position_on_the_genomic_accession=41488965, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642633, CurrentID=0, Status='live', Symbol='FAM230B', Aliases='', description='family with sequence similarity 230 member B (non-protein coding)', other_designations='KB-1183D5.13|family with sequence similarity 230, member B (non-protein coding)', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21166903, end_position_on_the_genomic_accession=21192156, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105369162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369162, CurrentID=0, Status='live', Symbol='LOC105369162', Aliases='', description='uncharacterized LOC105369162', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48577177, end_position_on_the_genomic_accession=48595846, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376554, CurrentID=0, Status='live', Symbol='LOC105376554', Aliases='', description='uncharacterized LOC105376554', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12030875, end_position_on_the_genomic_accession=12061791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642636, CurrentID=0, Status='live', Symbol='RAD21L1', Aliases='RAD21L, dJ545L17.2', description='RAD21 cohesin complex component like 1', other_designations='RAD21-like 1', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1226043, end_position_on_the_genomic_accession=1261279, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105369165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369165, CurrentID=0, Status='live', Symbol='LOC105369165', Aliases='', description='uncharacterized LOC105369165', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52722715, end_position_on_the_genomic_accession=52961460, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369166, CurrentID=0, Status='live', Symbol='LOC105369166', Aliases='', description='uncharacterized LOC105369166', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65476316, end_position_on_the_genomic_accession=65628425, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105500239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105500239, CurrentID=0, Status='live', Symbol='XLOC_007697', Aliases='', description='uncharacterized LOC105500239', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105500240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105500240, CurrentID=0, Status='live', Symbol='XLOC_009911', Aliases='', description='uncharacterized LOC105500240', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642641, CurrentID=0, Status='live', Symbol='RPSAP54', Aliases='RPSA_26_1300', description='ribosomal protein SA pseudogene 54', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20961234, end_position_on_the_genomic_accession=20962268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369170, CurrentID=0, Status='live', Symbol='LOC105369170', Aliases='', description='uncharacterized LOC105369170', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142898658, end_position_on_the_genomic_accession=142926820, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  642643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642643, CurrentID=0, Status='live', Symbol='LOC642643', Aliases='', description='uncharacterized LOC642643', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18733735, end_position_on_the_genomic_accession=18750417, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105369172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369172, CurrentID=0, Status='live', Symbol='LOC105369172', Aliases='', description='uncharacterized LOC105369172', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165906085, end_position_on_the_genomic_accession=165987133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369173, CurrentID=0, Status='live', Symbol='LOC105369173', Aliases='', description='uncharacterized LOC105369173', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91216593, end_position_on_the_genomic_accession=91222539, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369174, CurrentID=0, Status='live', Symbol='LOC105369174', Aliases='', description='uncharacterized LOC105369174', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1061207, end_position_on_the_genomic_accession=1066390, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369175, CurrentID=0, Status='live', Symbol='LOC105369175', Aliases='', description='cilia- and flagella-associated protein 54-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53641740, end_position_on_the_genomic_accession=53739323, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126296, CurrentID=0, Status='live', Symbol='MIR298', Aliases='MIRN298, hsa-mir-298', description='microRNA 298', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58818226, end_position_on_the_genomic_accession=58818313, orientation='minus', exon_count=1, OMIM=614914, no_hdr0=''),
  100126297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126297, CurrentID=0, Status='live', Symbol='MIR300', Aliases='MIRN300, hsa-mir-300', description='microRNA 300', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101041363, end_position_on_the_genomic_accession=101041445, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372190, CurrentID=0, Status='live', Symbol='LOC105372190', Aliases='', description='uncharacterized LOC105372190', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=73681325, end_position_on_the_genomic_accession=73691316, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126299, CurrentID=0, Status='live', Symbol='VTRNA2-1', Aliases='CBL-3, CBL3, MIR886, MIRN886, VTRNA2, hsa-mir-886, hvg-5, nc886, svtRNA2-1a', description='vault RNA 2-1', other_designations='cord blood lymphocyte-derived noncoding RNA 3|microRNA 886|noncoding RNA 886', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136080498, end_position_on_the_genomic_accession=136080597, orientation='minus', exon_count=1, OMIM=614938, no_hdr0=''),
  642652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642652, CurrentID=0, Status='live', Symbol='RARRES2P8', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 8', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35543286, end_position_on_the_genomic_accession=35543758, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126301, CurrentID=0, Status='live', Symbol='MIR509-2', Aliases='MIRN509-2', description='microRNA 509-2', other_designations='hsa-mir-509-2', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147258760, end_position_on_the_genomic_accession=147258850, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126302, CurrentID=0, Status='live', Symbol='MIR450B', Aliases='MIRN450B, mir-450b', description='microRNA 450b', other_designations='hsa-mir-450b', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134540185, end_position_on_the_genomic_accession=134540262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126303, CurrentID=0, Status='live', Symbol='MIR890', Aliases='MIRN890, hsa-mir-890', description='microRNA 890', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145994275, end_position_on_the_genomic_accession=145994351, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126304, CurrentID=0, Status='live', Symbol='MIR891B', Aliases='MIRN891B', description='microRNA 891b', other_designations='hsa-mir-891b', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146001053, end_position_on_the_genomic_accession=146001131, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369185, CurrentID=0, Status='live', Symbol='LOC105369185', Aliases='', description='uncharacterized LOC105369185', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18425399, end_position_on_the_genomic_accession=18433513, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126306, CurrentID=0, Status='live', Symbol='MIR888', Aliases='MIRN888, hsa-mir-888', description='microRNA 888', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145994784, end_position_on_the_genomic_accession=145994860, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  642659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642659, CurrentID=0, Status='live', Symbol='HNRNPA1P48', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 48', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51645756, end_position_on_the_genomic_accession=51647116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126308, CurrentID=0, Status='live', Symbol='MIR541', Aliases='MIRN541, hsa-mir-541, mir-541', description='microRNA 541', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101064495, end_position_on_the_genomic_accession=101064578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126309, CurrentID=0, Status='live', Symbol='MIR875', Aliases='MIRN875, hsa-mir-875', description='microRNA 875', other_designations='', map_location='8q22.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99536786, end_position_on_the_genomic_accession=99536861, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126310, CurrentID=0, Status='live', Symbol='MIR876', Aliases='MIRN876, hsa-mir-876, mir-876', description='microRNA 876', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28863626, end_position_on_the_genomic_accession=28863706, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126311, CurrentID=0, Status='live', Symbol='MIR147B', Aliases='MIRN147B, mir-147b', description='microRNA 147b', other_designations='hsa-mir-147b', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45433050, end_position_on_the_genomic_accession=45433129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369192, CurrentID=0, Status='live', Symbol='LOC105369192', Aliases='', description='uncharacterized LOC105369192', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88284714, end_position_on_the_genomic_accession=88341825, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100126313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126313, CurrentID=0, Status='live', Symbol='MIR744', Aliases='MIRN744, hsa-mir-744, mir-744', description='microRNA 744', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12081899, end_position_on_the_genomic_accession=12081996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126314, CurrentID=0, Status='live', Symbol='MIR877', Aliases='MIRN877, hsa-mir-877, miRNA877, mir-877', description='microRNA 877', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30584332, end_position_on_the_genomic_accession=30584417, orientation='plus', exon_count=1, OMIM=611619, no_hdr0=''),
  100126315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126315, CurrentID=0, Status='live', Symbol='MIR665', Aliases='MIRN665, hsa-mir-665, mir-665', description='microRNA 665', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100875033, end_position_on_the_genomic_accession=100875104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126316, CurrentID=0, Status='live', Symbol='MIR873', Aliases='MIRN873, hsa-mir-873, mir-873', description='microRNA 873', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=28888879, end_position_on_the_genomic_accession=28888955, orientation='minus', exon_count=1, OMIM=616137, no_hdr0=''),
  100126317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126317, CurrentID=0, Status='live', Symbol='MIR374B', Aliases='MIRN374B, mir-374b', description='microRNA 374b', other_designations='hsa-mir-374b', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74218547, end_position_on_the_genomic_accession=74218618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126318, CurrentID=0, Status='live', Symbol='MIR301B', Aliases='MIRN301B, mir-301b', description='microRNA 301b', other_designations='hsa-mir-301b', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21652981, end_position_on_the_genomic_accession=21653058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126319, CurrentID=0, Status='live', Symbol='MIR216B', Aliases='MIRN216B, mir-216b', description='microRNA 216b', other_designations='hsa-mir-216b', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=56000714, end_position_on_the_genomic_accession=56000795, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126320, CurrentID=0, Status='live', Symbol='MIR920', Aliases='MIRN920, hsa-mir-920', description='microRNA 920', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24212421, end_position_on_the_genomic_accession=24212495, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126321, CurrentID=0, Status='live', Symbol='MIR922', Aliases='MIRN922, hsa-mir-922', description='microRNA 922', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197674496, end_position_on_the_genomic_accession=197674576, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369202, CurrentID=0, Status='live', Symbol='LOC105369202', Aliases='', description='uncharacterized LOC105369202', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15855340, end_position_on_the_genomic_accession=15863863, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126323, CurrentID=0, Status='live', Symbol='MIR924', Aliases='MIRN924, hsa-mir-924', description='microRNA 924', other_designations='', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39622123, end_position_on_the_genomic_accession=39622175, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126324, CurrentID=0, Status='live', Symbol='MIR934', Aliases='MIRN934, hsa-mir-934, mir-934', description='microRNA 934', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136550878, end_position_on_the_genomic_accession=136550960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126325, CurrentID=0, Status='live', Symbol='MIR935', Aliases='MIRN935, hsa-mir-935, mir-935', description='microRNA 935', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53982307, end_position_on_the_genomic_accession=53982397, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126326, CurrentID=0, Status='live', Symbol='MIR936', Aliases='MIRN936, hsa-mir-936', description='microRNA 936', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104048089, end_position_on_the_genomic_accession=104048186, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126327, CurrentID=0, Status='live', Symbol='MIR938', Aliases='MIRN938, hsa-mir-938', description='microRNA 938', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29602264, end_position_on_the_genomic_accession=29602346, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126328, CurrentID=0, Status='live', Symbol='MIR940', Aliases='MIRN940, hsa-mir-940, mir-940', description='microRNA 940', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2271747, end_position_on_the_genomic_accession=2271840, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126329, CurrentID=0, Status='live', Symbol='MIR941-1', Aliases='MIRN941-1, mir-941-1', description='microRNA 941-1', other_designations='hsa-mir-941-1', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63919449, end_position_on_the_genomic_accession=63919520, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126330, CurrentID=0, Status='live', Symbol='MIR941-4', Aliases='MIRN941-4', description='microRNA 941-4', other_designations='hsa-mir-941-4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63919756, end_position_on_the_genomic_accession=63919827, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126331, CurrentID=0, Status='live', Symbol='MIR942', Aliases='MIRN942, hsa-mir-942, mir-942', description='microRNA 942', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117094643, end_position_on_the_genomic_accession=117094728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126332, CurrentID=0, Status='live', Symbol='MIR943', Aliases='MIRN943, hsa-mir-943', description='microRNA 943', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1986384, end_position_on_the_genomic_accession=1986477, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126333, CurrentID=0, Status='live', Symbol='MIR708', Aliases='MIRN708, hsa-mir-708', description='microRNA 708', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=79402022, end_position_on_the_genomic_accession=79402109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126334, CurrentID=0, Status='live', Symbol='MIR885', Aliases='MIRN885, hsa-mir-885, mir-885', description='microRNA 885', other_designations='', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10394489, end_position_on_the_genomic_accession=10394562, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126335, CurrentID=0, Status='live', Symbol='MIR543', Aliases='MIRN543, hsa-mir-543, mir-543', description='microRNA 543', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101031987, end_position_on_the_genomic_accession=101032064, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126336, CurrentID=0, Status='live', Symbol='MIR208B', Aliases='MIRN208B, mir-208b', description='microRNA 208b', other_designations='hsa-mir-208b', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23417987, end_position_on_the_genomic_accession=23418063, orientation='minus', exon_count=1, OMIM=613613, no_hdr0=''),
  100126337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126337, CurrentID=0, Status='live', Symbol='MIR509-3', Aliases='MIRN509-3, mir-509-3', description='microRNA 509-3', other_designations='hsa-mir-509-3', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=147259652, end_position_on_the_genomic_accession=147259726, orientation='minus', exon_count=1, OMIM=300876, no_hdr0=''),
  100126338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126338, CurrentID=0, Status='live', Symbol='MIR937', Aliases='MIRN937, hsa-mir-937, mir-937', description='microRNA 937', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143812957, end_position_on_the_genomic_accession=143813042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126339, CurrentID=0, Status='live', Symbol='MIR941-2', Aliases='MIRN941-2', description='microRNA 941-2', other_designations='hsa-mir-941-2', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63919505, end_position_on_the_genomic_accession=63919576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126340, CurrentID=0, Status='live', Symbol='MIR944', Aliases='MIRN944, hsa-mir-944, mir-944', description='microRNA 944', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189829922, end_position_on_the_genomic_accession=189830009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126341, CurrentID=0, Status='live', Symbol='MIR891A', Aliases='MIRN891A', description='microRNA 891a', other_designations='hsa-mir-891a', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146027794, end_position_on_the_genomic_accession=146027872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126342, CurrentID=0, Status='live', Symbol='MIR892A', Aliases='MIRN892A', description='microRNA 892a', other_designations='hsa-mir-892a', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145996669, end_position_on_the_genomic_accession=145996743, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126343, CurrentID=0, Status='live', Symbol='MIR874', Aliases='MIRN874, hsa-mir-874, mir-874', description='microRNA 874', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137647572, end_position_on_the_genomic_accession=137647649, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  642696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642696, CurrentID=0, Status='live', Symbol='LOC642696', Aliases='', description='uncharacterized LOC642696', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12087563, end_position_on_the_genomic_accession=12089946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100126345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126345, CurrentID=0, Status='live', Symbol='MIR889', Aliases='MIRN889, hsa-mir-889, mir-889', description='microRNA 889', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101047901, end_position_on_the_genomic_accession=101047979, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126346, CurrentID=0, Status='live', Symbol='MIR190B', Aliases='MIRN190B, mir-190b', description='microRNA 190b', other_designations='hsa-mir-190b', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154193665, end_position_on_the_genomic_accession=154193743, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126347, CurrentID=0, Status='live', Symbol='MIR887', Aliases='MIRN887, hsa-mir-887', description='microRNA 887', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15935182, end_position_on_the_genomic_accession=15935260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126348, CurrentID=0, Status='live', Symbol='MIR760', Aliases='MIRN760, hsa-mir-760, mir-760', description='microRNA 760', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93846832, end_position_on_the_genomic_accession=93846911, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126349, CurrentID=0, Status='live', Symbol='MIR921', Aliases='MIRN921, hsa-mir-921', description='microRNA 921', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166154743, end_position_on_the_genomic_accession=166154798, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126350, CurrentID=0, Status='live', Symbol='MIR933', Aliases='MIRN933, hsa-mir-933, mir-933', description='microRNA 933', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175167633, end_position_on_the_genomic_accession=175167709, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126351, CurrentID=0, Status='live', Symbol='MIR939', Aliases='MIRN939, hsa-mir-939, mir-939', description='microRNA 939', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144394149, end_position_on_the_genomic_accession=144394230, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126352, CurrentID=0, Status='live', Symbol='MIR941-3', Aliases='MIRN941-3', description='microRNA 941-3', other_designations='hsa-mir-941-3', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63919561, end_position_on_the_genomic_accession=63919632, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369233, CurrentID=0, Status='live', Symbol='LOC105369233', Aliases='', description='macrophage mannose receptor 1-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126354, CurrentID=0, Status='live', Symbol='MIR297', Aliases='MIRN297, hsa-mir-297', description='microRNA 297', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110860582, end_position_on_the_genomic_accession=110860647, orientation='minus', exon_count=1, OMIM=615520, no_hdr0=''),
  100126355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126355, CurrentID=0, Status='live', Symbol='MIR365A', Aliases='MIR365-1, MIRN365-1, hsa-mir-365a, mir-365a', description='microRNA 365a', other_designations='hsa-mir-365-1|microRNA 365-1', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14309285, end_position_on_the_genomic_accession=14309371, orientation='plus', exon_count=1, OMIM=614735, no_hdr0=''),
  100126356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126356, CurrentID=0, Status='live', Symbol='MIR365B', Aliases='MIR365-2, MIRN365-2, hsa-mir-365b, mir-365b', description='microRNA 365b', other_designations='hsa-mir-365-2|microRNA 365-2', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31575411, end_position_on_the_genomic_accession=31575521, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  118421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118421, CurrentID=0, Status='live', Symbol='LINC00161', Aliases='C21orf100, NCRNA00161', description='long intergenic non-protein coding RNA 161', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28539318, end_position_on_the_genomic_accession=28540355, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369238, CurrentID=0, Status='live', Symbol='LOC105369238', Aliases='', description='synaptic vesicle 2-related protein', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369239, CurrentID=0, Status='live', Symbol='LOC105369239', Aliases='', description='multidrug resistance-associated protein 6', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118424, CurrentID=0, Status='live', Symbol='UBE2J2', Aliases='NCUBE-2, NCUBE2, PRO2121', description='ubiquitin conjugating enzyme E2, J2', other_designations='E2 ubiquitin-conjugating enzyme J2|non-canonical ubiquitin-conjugating enzyme 2|ubiquitin conjugating enzyme 6|ubiquitin-conjugating enzyme E2, J2 (UBC6 homolog, yeast)|yeast UBC6 homolog', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1253912, end_position_on_the_genomic_accession=1273854, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  118425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118425, CurrentID=0, Status='live', Symbol='PCAT4', Aliases='GDEP, PCA4, PCAN1', description='prostate cancer associated transcript 4 (non-protein coding)', other_designations='gene differentially expressed in prostate|prostate cancer gene 1', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79827471, end_position_on_the_genomic_accession=79863247, orientation='plus', exon_count=2, OMIM=609717, no_hdr0=''),
  118426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118426, CurrentID=0, Status='live', Symbol='BORCS5', Aliases='LOH12CR1, LOH1CR12', description='BLOC-1 related complex subunit 5', other_designations='loss of heterozygosity, 12, chromosomal region 1|myrlysin', map_location='12p12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12357079, end_position_on_the_genomic_accession=12469410, orientation='plus', exon_count=7, OMIM=616598, no_hdr0=''),
  118427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118427, CurrentID=0, Status='live', Symbol='OLFM3', Aliases='NOE3, NOELIN3, OPTIMEDIN', description='olfactomedin 3', other_designations='olfactomedin related ER localized protein 3', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101802567, end_position_on_the_genomic_accession=101997234, orientation='minus', exon_count=10, OMIM=607567, no_hdr0=''),
  105369244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369244, CurrentID=0, Status='live', Symbol='LOC105369244', Aliases='', description='putative uncharacterized protein LOC641298', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118429, CurrentID=0, Status='live', Symbol='ANTXR2', Aliases='CMG-2, CMG2, HFS, ISH, JHF', description='anthrax toxin receptor 2', other_designations='capillary morphogenesis gene 2 protein|capillary morphogenesis protein 2', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79901617, end_position_on_the_genomic_accession=80073472, orientation='minus', exon_count=18, OMIM=608041, no_hdr0=''),
  118430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118430, CurrentID=0, Status='live', Symbol='MUCL1', Aliases='SBEM', description='mucin-like 1', other_designations='small breast epithelial mucin', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54830357, end_position_on_the_genomic_accession=54858393, orientation='plus', exon_count=5, OMIM=610857, no_hdr0=''),
  101929279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929279, CurrentID=0, Status='live', Symbol='LOC101929279', Aliases='', description='uncharacterized LOC101929279', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30211852, end_position_on_the_genomic_accession=30232690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  118432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118432, CurrentID=0, Status='live', Symbol='RPL29P2', Aliases='RPL29_10_1510', description='ribosomal protein L29 pseudogene 2', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7754320, end_position_on_the_genomic_accession=7754976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  118433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118433, CurrentID=0, Status='live', Symbol='RPL23AP7', Aliases='RPL23AL1, RPL23A_6_267, bA395L14.9', description='ribosomal protein L23a pseudogene 7', other_designations='ribosomal protein L23a-like 1', map_location='2q14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113611239, end_position_on_the_genomic_accession=113627138, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369250, CurrentID=0, Status='live', Symbol='LOC105369250', Aliases='', description='uncharacterized LOC105369250', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369251, CurrentID=0, Status='live', Symbol='LOC105369251', Aliases='', description='ankyrin repeat domain-containing protein 33B-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369252, CurrentID=0, Status='live', Symbol='LOC105369252', Aliases='', description='hepatocyte growth factor-like protein', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379782, CurrentID=0, Status='live', Symbol='LOC105379782', Aliases='', description='uncharacterized LOC105379782', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369254, CurrentID=0, Status='live', Symbol='LOC105369254', Aliases='', description='espin-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369255, CurrentID=0, Status='live', Symbol='LOC105369255', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369256, CurrentID=0, Status='live', Symbol='LOC105369256', Aliases='', description='uncharacterized LOC105369256', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369257, CurrentID=0, Status='live', Symbol='LOC105369257', Aliases='', description='SAGA-associated factor 29 homolog', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118442, CurrentID=0, Status='live', Symbol='GPR62', Aliases='GPCR8, KPG_005', description='G protein-coupled receptor 62', other_designations='G-protein coupled receptor GPCR8|G-protein coupled receptor KPG_005|hGPCR8', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51955314, end_position_on_the_genomic_accession=51957504, orientation='plus', exon_count=1, OMIM=606917, no_hdr0=''),
  105369259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369259, CurrentID=0, Status='live', Symbol='LOC105369259', Aliases='', description='putative L-type amino acid transporter 1-like protein MLAS', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369260, CurrentID=0, Status='live', Symbol='LOC105369260', Aliases='', description='double homeobox protein 4-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369261, CurrentID=0, Status='live', Symbol='LOC105369261', Aliases='', description='leukosialin', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369262, CurrentID=0, Status='live', Symbol='LOC105369262', Aliases='', description='splicing factor 45-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369263, CurrentID=0, Status='live', Symbol='LOC105369263', Aliases='', description='uncharacterized LOC105369263', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369264, CurrentID=0, Status='live', Symbol='LOC105369264', Aliases='', description='tyrosine-protein phosphatase non-receptor type 20', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642737, CurrentID=0, Status='live', Symbol='SETP22', Aliases='', description='SET pseudogene 22', other_designations='SET translocation pseudogene', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97054683, end_position_on_the_genomic_accession=97057182, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  642738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642738, CurrentID=0, Status='live', Symbol='RPL21P69', Aliases='RPL21_27_774', description='ribosomal protein L21 pseudogene 69', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159526014, end_position_on_the_genomic_accession=159526567, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369267, CurrentID=0, Status='live', Symbol='LOC105369267', Aliases='', description='splicing factor 45-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642740, CurrentID=0, Status='live', Symbol='RPL9P11', Aliases='RPL9_1_9', description='ribosomal protein L9 pseudogene 11', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9576405, end_position_on_the_genomic_accession=9577081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642741, CurrentID=0, Status='live', Symbol='RPL3P7', Aliases='RPL3_3_732', description='ribosomal protein L3 pseudogene 7', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108004302, end_position_on_the_genomic_accession=108005594, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369270, CurrentID=0, Status='live', Symbol='LOC105369270', Aliases='', description='uncharacterized LOC105369270', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369271, CurrentID=0, Status='live', Symbol='LOC105369271', Aliases='', description='nuclear pore complex-interacting protein family member B7-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369272, CurrentID=0, Status='live', Symbol='LOC105369272', Aliases='', description='uncharacterized LOC105369272', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66348077, end_position_on_the_genomic_accession=66353487, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369273, CurrentID=0, Status='live', Symbol='LOC105369273', Aliases='', description='protein SSX4-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642746, CurrentID=0, Status='live', Symbol='EIF4A1P9', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 9', other_designations='', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2586187, end_position_on_the_genomic_accession=2587880, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369275, CurrentID=0, Status='live', Symbol='LOC105369275', Aliases='', description='COPII coat assembly protein SEC16-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118460, CurrentID=0, Status='live', Symbol='EXOSC6', Aliases='EAP4, MTR3, Mtr3p, hMtr3p, p11', description='exosome component 6', other_designations='Mtr3 (mRNA transport regulator 3)-homolog|exosome complex exonuclease MTR3|hMtr3|homolog of yeast mRNA transport regulator 3|mRNA transport regulator 3 homolog', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70250231, end_position_on_the_genomic_accession=70251930, orientation='minus', exon_count=1, OMIM=606490, no_hdr0=''),
  118461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118461, CurrentID=0, Status='live', Symbol='C10orf71', Aliases='', description='chromosome 10 open reading frame 71', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49297056, end_position_on_the_genomic_accession=49327492, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369278, CurrentID=0, Status='live', Symbol='LOC105369278', Aliases='', description='putative uncharacterized protein LOC641298', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369279, CurrentID=0, Status='live', Symbol='LOC105369279', Aliases='', description='golgin subfamily A member 6-like protein 2 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22767696, end_position_on_the_genomic_accession=22768242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369280, CurrentID=0, Status='live', Symbol='LOC105369280', Aliases='', description='cytochrome P450 2D4-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369281, CurrentID=0, Status='live', Symbol='LOC105369281', Aliases='', description='uncharacterized LOC105369281', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369282, CurrentID=0, Status='live', Symbol='LOC105369282', Aliases='', description='putative tripartite motif-containing protein 64B', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391115, CurrentID=0, Status='live', Symbol='OR6K5P', Aliases='', description='olfactory receptor family 6 subfamily K member 5 pseudogene', other_designations='olfactory receptor, family 6, subfamily K, member 5 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158742568, end_position_on_the_genomic_accession=158743415, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369284, CurrentID=0, Status='live', Symbol='LOC105369284', Aliases='', description='D(1B) dopamine receptor-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642757, CurrentID=0, Status='live', Symbol='TBC1D22A-AS1', Aliases='', description='TBC1D22A antisense RNA 1', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46914092, end_position_on_the_genomic_accession=46916042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369286, CurrentID=0, Status='live', Symbol='LOC105369286', Aliases='', description='uncharacterized LOC105369286', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31727484, end_position_on_the_genomic_accession=31767953, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  118471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118471, CurrentID=0, Status='live', Symbol='PRAP1', Aliases='PRO1195, UPA', description='proline-rich acidic protein 1', other_designations='epididymis tissue protein Li 178', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133347340, end_position_on_the_genomic_accession=133352683, orientation='plus', exon_count=5, OMIM=609776, no_hdr0=''),
  118472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118472, CurrentID=0, Status='live', Symbol='ZNF511', Aliases='', description='zinc finger protein 511', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133308889, end_position_on_the_genomic_accession=133313162, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369289, CurrentID=0, Status='live', Symbol='LOC105369289', Aliases='', description='uncharacterized LOC105369289', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26676135, end_position_on_the_genomic_accession=26681023, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642762, CurrentID=0, Status='live', Symbol='ABCD1P5', Aliases='', description='ATP binding cassette subfamily D member 1 pseudogene 5', other_designations='ATP-binding cassette, sub-family D (ALD), member 1 pseudogene|ATP-binding cassette, sub-family D (ALD), member 1, pseudogene 5|ATP-binding cassette, sub-family D-like pseudogene|hCG2042779', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91840046, end_position_on_the_genomic_accession=91843813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369291, CurrentID=0, Status='live', Symbol='LOC105369291', Aliases='', description='uncharacterized LOC105369291', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15502172, end_position_on_the_genomic_accession=15510091, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369292, CurrentID=0, Status='live', Symbol='LOC105369292', Aliases='', description='uncharacterized LOC105369292', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15263421, end_position_on_the_genomic_accession=15283708, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369293, CurrentID=0, Status='live', Symbol='LOC105369293', Aliases='', description='uncharacterized LOC105369293', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18413748, end_position_on_the_genomic_accession=18437554, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369294, CurrentID=0, Status='live', Symbol='LOC105369294', Aliases='', description='uncharacterized LOC105369294', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=40415614, end_position_on_the_genomic_accession=40422226, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379789, CurrentID=0, Status='live', Symbol='LOC105379789', Aliases='', description='uncharacterized LOC105379789', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369296, CurrentID=0, Status='live', Symbol='LOC105369296', Aliases='', description='phylloquinone omega-hydroxylase CYP4F2-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15979805, end_position_on_the_genomic_accession=15999015, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369297, CurrentID=0, Status='live', Symbol='LOC105369297', Aliases='', description='uncharacterized LOC105369297', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25649436, end_position_on_the_genomic_accession=25671222, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369298, CurrentID=0, Status='live', Symbol='LOC105369298', Aliases='', description='uncharacterized LOC105369298', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=23482779, end_position_on_the_genomic_accession=23484534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369299, CurrentID=0, Status='live', Symbol='LOC105369299', Aliases='', description='uncharacterized LOC105369299', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29296165, end_position_on_the_genomic_accession=29298696, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369300, CurrentID=0, Status='live', Symbol='LOC105369300', Aliases='', description='uncharacterized LOC105369300', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14740139, end_position_on_the_genomic_accession=14744489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369301, CurrentID=0, Status='live', Symbol='LOC105369301', Aliases='', description='uncharacterized LOC105369301', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36430294, end_position_on_the_genomic_accession=36482182, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369302, CurrentID=0, Status='live', Symbol='LOC105369302', Aliases='', description='uncharacterized LOC105369302', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15571198, end_position_on_the_genomic_accession=15627241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  118487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118487, CurrentID=0, Status='live', Symbol='CHCHD1', Aliases='C10orf34, C2360, MRP-S37', description='coiled-coil-helix-coiled-coil-helix domain containing 1', other_designations='28S ribosomal protein S37, mitochondrial|nuclear protein C2360', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73782050, end_position_on_the_genomic_accession=73783648, orientation='plus', exon_count=3, OMIM=608842, no_hdr0=''),
  642776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642776, CurrentID=0, Status='live', Symbol='DANT2', Aliases='', description='DXZ4 associated non-coding transcript 2, distal', other_designations='DXZ4 associated noncoding transcript 2', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115917271, end_position_on_the_genomic_accession=115969111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369305, CurrentID=0, Status='live', Symbol='LOC105369305', Aliases='', description='uncharacterized LOC105369305', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36811386, end_position_on_the_genomic_accession=36813696, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642778, CurrentID=0, Status='live', Symbol='NPIPA3', Aliases='', description='nuclear pore complex interacting protein family member A3', other_designations='nuclear pore complex interacting protein family, member A3', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14706473, end_position_on_the_genomic_accession=14726338, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  118491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118491, CurrentID=0, Status='live', Symbol='CFAP70', Aliases='TTC18', description='cilia and flagella associated protein 70', other_designations='TPR repeat protein 18|tetratricopeptide repeat domain 18|tetratricopeptide repeat protein 18|tetratricopeptide repeat-containing protein (LOC118491)', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73253758, end_position_on_the_genomic_accession=73361064, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  105369308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369308, CurrentID=0, Status='live', Symbol='LOC105369308', Aliases='', description='uncharacterized LOC105369308', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36632681, end_position_on_the_genomic_accession=36638498, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369309, CurrentID=0, Status='live', Symbol='LOC105369309', Aliases='', description='uncharacterized LOC105369309', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57189069, end_position_on_the_genomic_accession=57231589, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369310, CurrentID=0, Status='live', Symbol='LOC105369310', Aliases='', description='uncharacterized LOC105369310', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56902743, end_position_on_the_genomic_accession=56949853, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369311, CurrentID=0, Status='live', Symbol='LOC105369311', Aliases='', description='uncharacterized LOC105369311', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57322395, end_position_on_the_genomic_accession=57325956, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369312, CurrentID=0, Status='live', Symbol='LOC105369312', Aliases='', description='uncharacterized LOC105369312', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57392990, end_position_on_the_genomic_accession=57404155, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379792, CurrentID=0, Status='live', Symbol='LOC105379792', Aliases='', description='uncharacterized LOC105379792', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372444, CurrentID=0, Status='live', Symbol='LOC105372444', Aliases='', description='uncharacterized LOC105372444', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51271237, end_position_on_the_genomic_accession=51295425, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369315, CurrentID=0, Status='live', Symbol='LOC105369315', Aliases='', description='uncharacterized LOC105369315', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59095072, end_position_on_the_genomic_accession=59102572, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369316, CurrentID=0, Status='live', Symbol='LOC105369316', Aliases='', description='uncharacterized LOC105369316', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59260051, end_position_on_the_genomic_accession=59284695, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=789, CurrentID=0, Status='live', Symbol='SLC25A20P1', Aliases='CACTP, SLC25A20P', description='solute carrier family 25 (carnitine/acylcarnitine translocase), member 20 pseudogene 1', other_designations='', map_location='6p12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52246371, end_position_on_the_genomic_accession=52248104, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369318, CurrentID=0, Status='live', Symbol='LOC105369318', Aliases='', description='uncharacterized LOC105369318', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59674311, end_position_on_the_genomic_accession=59676041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  391121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391121, CurrentID=0, Status='live', Symbol='OR10J4', Aliases='OR10J4P, OST717', description='olfactory receptor family 10 subfamily J member 4 (gene/pseudogene)', other_designations='olfactory receptor 10J4|olfactory receptor, family 10, subfamily J, member 4 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159432205, end_position_on_the_genomic_accession=159433136, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369320, CurrentID=0, Status='live', Symbol='LOC105369320', Aliases='', description='uncharacterized LOC105369320', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60503824, end_position_on_the_genomic_accession=60507018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369321, CurrentID=0, Status='live', Symbol='LOC105369321', Aliases='', description='uncharacterized LOC105369321', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60607593, end_position_on_the_genomic_accession=60699116, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369322, CurrentID=0, Status='live', Symbol='LOC105369322', Aliases='', description='uncharacterized LOC105369322', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60792333, end_position_on_the_genomic_accession=60836139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369323, CurrentID=0, Status='live', Symbol='LOC105369323', Aliases='', description='uncharacterized LOC105369323', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60913166, end_position_on_the_genomic_accession=60913975, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642796, CurrentID=0, Status='live', Symbol='TSPY18P', Aliases='', description='testis specific protein, Y-linked 18, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9869703, end_position_on_the_genomic_accession=9870436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642797, CurrentID=0, Status='live', Symbol='DNAH10OS', Aliases='', description='dynein, axonemal, heavy chain 10 opposite strand', other_designations='dynein, axonemal, heavy chain 10 opposite strand (non-protein coding)', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=122984592, end_position_on_the_genomic_accession=122985484, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126446, CurrentID=0, Status='live', Symbol='DEPDC1P2', Aliases='', description='DEP domain containing 1 pseudogene 2', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105758048, end_position_on_the_genomic_accession=105761705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642799, CurrentID=0, Status='live', Symbol='NPIPA2', Aliases='NPIP', description='nuclear pore complex interacting protein family member A2', other_designations='nuclear pore complex interacting protein family, member A2', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14745595, end_position_on_the_genomic_accession=14765448, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105369328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369328, CurrentID=0, Status='live', Symbol='LOC105369328', Aliases='', description='uncharacterized LOC105369328', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369329, CurrentID=0, Status='live', Symbol='LOC105369329', Aliases='', description='uncharacterized LOC105369329', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61496311, end_position_on_the_genomic_accession=61506649, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  642802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642802, CurrentID=0, Status='live', Symbol='CCDC137P', Aliases='', description='coiled-coil domain containing 137 pseudogene', other_designations='', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72086890, end_position_on_the_genomic_accession=72087660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379795, CurrentID=0, Status='live', Symbol='LOC105379795', Aliases='', description='uncharacterized LOC105379795', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369332, CurrentID=0, Status='live', Symbol='LOC105369332', Aliases='', description='uncharacterized LOC105369332', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62832166, end_position_on_the_genomic_accession=62834056, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369333, CurrentID=0, Status='live', Symbol='LOC105369333', Aliases='', description='uncharacterized LOC105369333', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62919642, end_position_on_the_genomic_accession=62934768, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369334, CurrentID=0, Status='live', Symbol='LOC105369334', Aliases='', description='uncharacterized LOC105369334', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63245695, end_position_on_the_genomic_accession=63251074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369335, CurrentID=0, Status='live', Symbol='LOC105369335', Aliases='', description='uncharacterized LOC105369335', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63575053, end_position_on_the_genomic_accession=63590086, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369336, CurrentID=0, Status='live', Symbol='LOC105369336', Aliases='', description='uncharacterized LOC105369336', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63844467, end_position_on_the_genomic_accession=63863192, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  391124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391124, CurrentID=0, Status='live', Symbol='RPL27AP2', Aliases='RPL27A_1_106', description='ribosomal protein L27a pseudogene 2', other_designations='hCG2041321', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160018009, end_position_on_the_genomic_accession=160020784, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369338, CurrentID=0, Status='live', Symbol='LOC105369338', Aliases='', description='uncharacterized LOC105369338', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64028669, end_position_on_the_genomic_accession=64033449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369339, CurrentID=0, Status='live', Symbol='LOC105369339', Aliases='', description='uncharacterized LOC105369339', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64166232, end_position_on_the_genomic_accession=64170665, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642812, CurrentID=0, Status='live', Symbol='API5P1', Aliases='API5L1', description='apoptosis inhibitor 5 pseudogene 1', other_designations='API5 pseudogene 1|API5-like 1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116104366, end_position_on_the_genomic_accession=116108497, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369341, CurrentID=0, Status='live', Symbol='LOC105369341', Aliases='', description='uncharacterized LOC105369341', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64410118, end_position_on_the_genomic_accession=64420238, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369342, CurrentID=0, Status='live', Symbol='LOC105369342', Aliases='', description='uncharacterized LOC105369342', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64502460, end_position_on_the_genomic_accession=64505386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369343, CurrentID=0, Status='live', Symbol='MIR194-2HG', Aliases='', description='MIR194-2 host gene', other_designations='pri-microRNA-194-2, -192', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64888229, end_position_on_the_genomic_accession=64893452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369344, CurrentID=0, Status='live', Symbol='LOC105369344', Aliases='', description='uncharacterized LOC105369344', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65305624, end_position_on_the_genomic_accession=65314720, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369345, CurrentID=0, Status='live', Symbol='LOC105369345', Aliases='', description='uncharacterized LOC105369345', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65435347, end_position_on_the_genomic_accession=65445919, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642818, CurrentID=0, Status='live', Symbol='ST13P21', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 21', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108502379, end_position_on_the_genomic_accession=108503452, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928235, CurrentID=0, Status='live', Symbol='CTB-41I6.1', Aliases='', description='uncharacterized LOC101928235', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8965780, end_position_on_the_genomic_accession=8968423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369348, CurrentID=0, Status='live', Symbol='LOC105369348', Aliases='', description='uncharacterized LOC105369348', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65856249, end_position_on_the_genomic_accession=65856893, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379798, CurrentID=0, Status='live', Symbol='LOC105379798', Aliases='', description='uncharacterized LOC105379798', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369350, CurrentID=0, Status='live', Symbol='LOC105369350', Aliases='', description='uncharacterized LOC105369350', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65937400, end_position_on_the_genomic_accession=65945524, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369351, CurrentID=0, Status='live', Symbol='LOC105369351', Aliases='', description='uncharacterized LOC105369351', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66267741, end_position_on_the_genomic_accession=66269694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369352, CurrentID=0, Status='live', Symbol='LOC105369352', Aliases='', description='uncharacterized LOC105369352', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66261926, end_position_on_the_genomic_accession=66265666, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369353, CurrentID=0, Status='live', Symbol='LOC105369353', Aliases='', description='uncharacterized LOC105369353', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66259283, end_position_on_the_genomic_accession=66261745, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642826, CurrentID=0, Status='live', Symbol='BMS1P6', Aliases='BMS1LP2, BMS1LP6, BMS1P2, bA302K17.2', description='BMS1, ribosome biogenesis factor pseudogene 6', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 2|BMS1 pseudogene 6|BMS1 ribosome biogenesis factor pseudogene 6|BMS1-like, ribosome assembly protein pseudogene|BMS1L pseudogene 2|BMS1L pseudogene 6|ribosome biogenesis protein BMS1 homolog', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47535574, end_position_on_the_genomic_accession=47553523, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100126475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126475, CurrentID=0, Status='live', Symbol='TRNAL-UAA', Aliases='', description='transfer RNA leucine (anticodon UAA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642828, CurrentID=0, Status='live', Symbol='RPL6P10', Aliases='RPL6_9_488', description='ribosomal protein L6 pseudogene 10', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65573440, end_position_on_the_genomic_accession=65574371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126477, CurrentID=0, Status='live', Symbol='TRNAL-AAG', Aliases='', description='transfer RNA leucine (anticodon AAG)', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369358, CurrentID=0, Status='live', Symbol='LOC105369358', Aliases='', description='uncharacterized LOC105369358', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67758563, end_position_on_the_genomic_accession=67762743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126479, CurrentID=0, Status='live', Symbol='TRNAL-UAG', Aliases='', description='transfer RNA leucine (anticodon UAG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126480, CurrentID=0, Status='live', Symbol='TRNAL-CAG', Aliases='', description='transfer RNA leucine (anticodon CAG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126481, CurrentID=0, Status='live', Symbol='TRNAI-AAU', Aliases='', description='transfer RNA isoleucine (anticodon AAU)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26780622, end_position_on_the_genomic_accession=26780695, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126482, CurrentID=0, Status='live', Symbol='TRNAI-GAU', Aliases='', description='transfer RNA isoleucine (anticodon GAU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126483, CurrentID=0, Status='live', Symbol='TRNAI-UAU', Aliases='', description='transfer RNA isoleucine (anticodon UAU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126484, CurrentID=0, Status='live', Symbol='TRNAM-CAU', Aliases='', description='transfer RNA methionine (anticodon CAU)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26766234, end_position_on_the_genomic_accession=26766306, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126485, CurrentID=0, Status='live', Symbol='TRNAV-AAC', Aliases='', description='transfer RNA valine (anticodon AAC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369366, CurrentID=0, Status='live', Symbol='LOC105369366', Aliases='', description='uncharacterized LOC105369366', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69166791, end_position_on_the_genomic_accession=69177976, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100126487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126487, CurrentID=0, Status='live', Symbol='TRNAV-UAC', Aliases='', description='transfer RNA valine (anticodon UAC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642840, CurrentID=0, Status='live', Symbol='MEP1AP1', Aliases='MEP1AL1', description='meprin A subunit alpha pseudogene 1', other_designations='meprin A, alpha (PABA peptide hydrolase) pseudogene|meprin A, alpha pseudogene 1|meprin A, alpha-like 1 (pseudogene)|meprin A, alpha-like 1 (psueodgene)', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40187143, end_position_on_the_genomic_accession=40193921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126489, CurrentID=0, Status='live', Symbol='TRNAS-AGA', Aliases='', description='transfer RNA serine (anticodon AGA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369370, CurrentID=0, Status='live', Symbol='LOC105369370', Aliases='', description='uncharacterized LOC105369370', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=69298918, end_position_on_the_genomic_accession=69372512, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  642843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642843, CurrentID=0, Status='live', Symbol='CPSF4L', Aliases='', description='cleavage and polyadenylation specific factor 4-like', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73242343, end_position_on_the_genomic_accession=73262532, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100126492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126492, CurrentID=0, Status='live', Symbol='TRNAS-CGA', Aliases='', description='transfer RNA serine (anticodon CGA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126493, CurrentID=0, Status='live', Symbol='TRNAP-AGG', Aliases='', description='transfer RNA proline (anticodon AGG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642846, CurrentID=0, Status='live', Symbol='LOC642846', Aliases='', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11-like', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9283657, end_position_on_the_genomic_accession=9314088, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100126495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126495, CurrentID=0, Status='live', Symbol='TRNAP-UGG', Aliases='', description='transfer RNA proline (anticodon UGG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126496, CurrentID=0, Status='live', Symbol='TRNAP-CGG', Aliases='', description='transfer RNA proline (anticodon CGG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126497, CurrentID=0, Status='live', Symbol='TRNAT-AGU', Aliases='', description='transfer RNA threonine (anticodon AGU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369378, CurrentID=0, Status='live', Symbol='LOC105369378', Aliases='', description='uncharacterized LOC105369378', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72585698, end_position_on_the_genomic_accession=72588206, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100126499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126499, CurrentID=0, Status='live', Symbol='TRNAT-UGU', Aliases='', description='transfer RNA threonine (anticodon UGU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126500, CurrentID=0, Status='live', Symbol='TRNAT-CGU', Aliases='', description='transfer RNA threonine (anticodon CGU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126501, CurrentID=0, Status='live', Symbol='TRNAA-AGC', Aliases='', description='transfer RNA alanine (anticodon AGC)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57815974, end_position_on_the_genomic_accession=57816046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369382, CurrentID=0, Status='live', Symbol='LOC105369382', Aliases='', description='uncharacterized LOC105369382', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73253426, end_position_on_the_genomic_accession=73259158, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126503, CurrentID=0, Status='live', Symbol='TRNAA-UGC', Aliases='', description='transfer RNA alanine (anticodon UGC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126504, CurrentID=0, Status='live', Symbol='TRNAA-CGC', Aliases='', description='transfer RNA alanine (anticodon CGC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126505, CurrentID=0, Status='live', Symbol='TRNAY-AUA', Aliases='', description='transfer RNA tyrosine (anticodon AUA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126506, CurrentID=0, Status='live', Symbol='TRNAY-GUA', Aliases='', description='transfer RNA tyrosine (anticodon GUA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369387, CurrentID=0, Status='live', Symbol='LOC105369387', Aliases='', description='uncharacterized LOC105369387', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74698305, end_position_on_the_genomic_accession=74700678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100126508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126508, CurrentID=0, Status='live', Symbol='TRNAH-GUG', Aliases='', description='transfer RNA histidine(anticodon GUG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126509, CurrentID=0, Status='live', Symbol='TRNAQ-UUG', Aliases='', description='transfer RNA glutamine (anticodon UUG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126510, CurrentID=0, Status='live', Symbol='TRNAQ-CUG', Aliases='', description='transfer RNA glutamine (anticodon CUG)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145459658, end_position_on_the_genomic_accession=145459729, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379805, CurrentID=0, Status='live', Symbol='LOC105379805', Aliases='', description='uncharacterized LOC105379805', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126512, CurrentID=0, Status='live', Symbol='TRNAN-GUU', Aliases='', description='transfer RNA asparagine (anticodon GUU)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126513, CurrentID=0, Status='live', Symbol='TRNAK-UUU', Aliases='', description='transfer RNA lysine (anticodon UUU)', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126514, CurrentID=0, Status='live', Symbol='TRNAK-CUU', Aliases='', description='transfer RNA lysine (anticodon CUU)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54957869, end_position_on_the_genomic_accession=54957941, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369395, CurrentID=0, Status='live', Symbol='LOC105369395', Aliases='', description='uncharacterized LOC105369395', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76284397, end_position_on_the_genomic_accession=76321105, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100126516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126516, CurrentID=0, Status='live', Symbol='TRNAD-GUC', Aliases='', description='transfer RNA aspartic acid (anticodon GUC)', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642869, CurrentID=0, Status='live', Symbol='SETP4', Aliases='', description='SET pseudogene 4', other_designations='SET translocation (myeloid leukemia-associated) pseudogene|hCG1644608', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84754998, end_position_on_the_genomic_accession=84756477, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126518, CurrentID=0, Status='live', Symbol='TRNAE-CUC', Aliases='', description='transfer RNA glutamic acid (anticodon CUC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369399, CurrentID=0, Status='live', Symbol='LOC105369399', Aliases='', description='uncharacterized LOC105369399', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77047866, end_position_on_the_genomic_accession=77048740, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100126520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126520, CurrentID=0, Status='live', Symbol='TRNAC-GCA', Aliases='', description='transfer RNA cysteine (anticodon GCA)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93516277, end_position_on_the_genomic_accession=93516349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928244, CurrentID=0, Status='live', Symbol='TESC-AS1', Aliases='', description='TESC antisense RNA 1 (head to head)', other_designations='', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117099467, end_position_on_the_genomic_accession=117141489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126522, CurrentID=0, Status='live', Symbol='TRNAR-ACG', Aliases='', description='transfer RNA arginine (anticodon ACG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126523, CurrentID=0, Status='live', Symbol='TRNAR-CCG', Aliases='', description='transfer RNA arginine (anticodon CCG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126524, CurrentID=0, Status='live', Symbol='TRNAR-UCG', Aliases='', description='transfer RNA arginine (anticodon UCG)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369405, CurrentID=0, Status='live', Symbol='LOC105369405', Aliases='', description='uncharacterized LOC105369405', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=79595877, end_position_on_the_genomic_accession=79600351, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126526, CurrentID=0, Status='live', Symbol='TRNAR-UCU', Aliases='', description='transfer RNA arginine (anticodon UCU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126527, CurrentID=0, Status='live', Symbol='TRNAR-CCU', Aliases='', description='transfer RNA arginine (anticodon CCU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369408, CurrentID=0, Status='live', Symbol='LOC105369408', Aliases='', description='uncharacterized LOC105369408', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=80203226, end_position_on_the_genomic_accession=80225953, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100126529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126529, CurrentID=0, Status='live', Symbol='TRNAS-GCU', Aliases='', description='transfer RNA serine (anticodon GCU)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369410, CurrentID=0, Status='live', Symbol='LOC105369410', Aliases='', description='uncharacterized LOC105369410', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82472180, end_position_on_the_genomic_accession=82480087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126531, CurrentID=0, Status='live', Symbol='TRNAG-GCC', Aliases='', description='transfer RNA glycine (anticodon GCC)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126532, CurrentID=0, Status='live', Symbol='TRNAG-UCC', Aliases='', description='transfer RNA glycine (anticodon UCC)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126533, CurrentID=0, Status='live', Symbol='TRNAG-CCC', Aliases='', description='transfer RNA glycine (anticodon CCC)', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121016845, end_position_on_the_genomic_accession=121016915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126534, CurrentID=0, Status='live', Symbol='TRNASTOP-UCA', Aliases='', description='transfer RNA opal suppressor (anticodon UCA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379809, CurrentID=0, Status='live', Symbol='LOC105379809', Aliases='', description='uncharacterized LOC105379809', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100126536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126536, CurrentID=0, Status='live', Symbol='TRNASTOP-UUA', Aliases='', description='transfer RNA ochre suppressor(anticodon UUA)', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642889, CurrentID=0, Status='live', Symbol='FLJ16423', Aliases='', description='uncharacterized LOC642889', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=148441357, end_position_on_the_genomic_accession=148441761, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369418, CurrentID=0, Status='live', Symbol='LOC105369418', Aliases='', description='uncharacterized LOC105369418', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83891799, end_position_on_the_genomic_accession=83919607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642891, CurrentID=0, Status='live', Symbol='CHL1-AS2', Aliases='', description='CHL1 antisense RNA 2', other_designations='CHL1 antisense RNA 2 (non-protein coding)', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=237441, end_position_on_the_genomic_accession=239090, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  642892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642892, CurrentID=0, Status='live', Symbol='RPS29P3', Aliases='RPS29_7_451, bcm2758', description='ribosomal protein S29 pseudogene 3', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196536495, end_position_on_the_genomic_accession=196536792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379810, CurrentID=0, Status='live', Symbol='LOC105379810', Aliases='', description='uncharacterized LOC105379810', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369422, CurrentID=0, Status='live', Symbol='LOC105369422', Aliases='', description='uncharacterized LOC105369422', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87014123, end_position_on_the_genomic_accession=87037773, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369423, CurrentID=0, Status='live', Symbol='LOC105369423', Aliases='', description='uncharacterized LOC105369423', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87718354, end_position_on_the_genomic_accession=87719422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369424, CurrentID=0, Status='live', Symbol='LOC105369424', Aliases='', description='uncharacterized LOC105369424', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=88733201, end_position_on_the_genomic_accession=88735631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369425, CurrentID=0, Status='live', Symbol='LOC105369425', Aliases='', description='tripartite motif-containing protein LOC642612', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369426, CurrentID=0, Status='live', Symbol='LOC105369426', Aliases='', description='uncharacterized LOC105369426', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=90928903, end_position_on_the_genomic_accession=90930929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  118611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118611, CurrentID=0, Status='live', Symbol='C10orf90', Aliases='FATS, bA422P15.2', description='chromosome 10 open reading frame 90', other_designations='fragile-site associated tumor suppressor homolog', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126424997, end_position_on_the_genomic_accession=126671038, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  105369428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369428, CurrentID=0, Status='live', Symbol='LOC105369428', Aliases='', description='uncharacterized LOC105369428', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=91794339, end_position_on_the_genomic_accession=91872454, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105369429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369429, CurrentID=0, Status='live', Symbol='LOC105369429', Aliases='', description='uncharacterized LOC105369429', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92146409, end_position_on_the_genomic_accession=92153903, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369430, CurrentID=0, Status='live', Symbol='LOC105369430', Aliases='', description='uncharacterized LOC105369430', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92647278, end_position_on_the_genomic_accession=92654733, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369431, CurrentID=0, Status='live', Symbol='LOC105369431', Aliases='', description='uncharacterized LOC105369431', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92914138, end_position_on_the_genomic_accession=92915629, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369432, CurrentID=0, Status='live', Symbol='LOC105369432', Aliases='', description='uncharacterized LOC105369432', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93248395, end_position_on_the_genomic_accession=93252254, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379812, CurrentID=0, Status='live', Symbol='LOC105379812', Aliases='', description='uncharacterized LOC105379812', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369434, CurrentID=0, Status='live', Symbol='LOC105369434', Aliases='', description='uncharacterized LOC105369434', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642907, CurrentID=0, Status='live', Symbol='NIFKP8', Aliases='MKI67IPP8', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 8', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 8', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85181155, end_position_on_the_genomic_accession=85182125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369436, CurrentID=0, Status='live', Symbol='LOC105369436', Aliases='', description='uncharacterized LOC105369436', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94634738, end_position_on_the_genomic_accession=94657461, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  642909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642909, CurrentID=0, Status='live', Symbol='CTBP2P4', Aliases='', description='C-terminal binding protein 2 pseudogene 4', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98576341, end_position_on_the_genomic_accession=98577257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369438, CurrentID=0, Status='live', Symbol='LOC105369438', Aliases='', description='uncharacterized LOC105369438', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94675544, end_position_on_the_genomic_accession=94740400, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379813, CurrentID=0, Status='live', Symbol='LOC105379813', Aliases='', description='uncharacterized LOC105379813', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369440, CurrentID=0, Status='live', Symbol='LOC105369440', Aliases='', description='uncharacterized LOC105369440', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95513844, end_position_on_the_genomic_accession=95518014, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369441, CurrentID=0, Status='live', Symbol='LOC105369441', Aliases='', description='uncharacterized LOC105369441', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95611682, end_position_on_the_genomic_accession=95615158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369442, CurrentID=0, Status='live', Symbol='LOC105369442', Aliases='', description='uncharacterized LOC105369442', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96467978, end_position_on_the_genomic_accession=96499167, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369443, CurrentID=0, Status='live', Symbol='LOC105369443', Aliases='', description='uncharacterized LOC105369443', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=96507440, end_position_on_the_genomic_accession=96568658, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369444, CurrentID=0, Status='live', Symbol='LOC105369444', Aliases='', description='oocyte-secreted protein 2-like', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=60023803, end_position_on_the_genomic_accession=60030834, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369445, CurrentID=0, Status='live', Symbol='LOC105369445', Aliases='', description='uncharacterized LOC105369445', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66721448, end_position_on_the_genomic_accession=66727977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369446, CurrentID=0, Status='live', Symbol='LOC105369446', Aliases='', description='mitochondrial import receptor subunit TOM20 homolog pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77300933, end_position_on_the_genomic_accession=77314080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369447, CurrentID=0, Status='live', Symbol='LOC105369447', Aliases='', description='uncharacterized LOC105369447', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93197639, end_position_on_the_genomic_accession=93203569, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369448, CurrentID=0, Status='live', Symbol='LOC105369448', Aliases='', description='uncharacterized LOC105369448', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97011542, end_position_on_the_genomic_accession=97017278, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372199, CurrentID=0, Status='live', Symbol='LOC105372199', Aliases='', description='uncharacterized LOC105372199', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369450, CurrentID=0, Status='live', Symbol='LOC105369450', Aliases='', description='uncharacterized LOC105369450', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97270983, end_position_on_the_genomic_accession=97273429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369451, CurrentID=0, Status='live', Symbol='LOC105369451', Aliases='', description='uncharacterized LOC105369451', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97391216, end_position_on_the_genomic_accession=97399723, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126572, CurrentID=0, Status='live', Symbol='GJE1', Aliases='CX23', description='gap junction protein epsilon 1', other_designations='connexin-23|gap junction epsilon-1 protein|gap junction protein, epsilon 1, 23kDa', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=142133159, end_position_on_the_genomic_accession=142134922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369453, CurrentID=0, Status='live', Symbol='LOC105369453', Aliases='', description='uncharacterized LOC105369453', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97798443, end_position_on_the_genomic_accession=97802391, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369454, CurrentID=0, Status='live', Symbol='LOC105369454', Aliases='', description='uncharacterized LOC105369454', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=97878569, end_position_on_the_genomic_accession=97957056, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369455, CurrentID=0, Status='live', Symbol='LOC105369455', Aliases='', description='uncharacterized LOC105369455', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=98453812, end_position_on_the_genomic_accession=98495667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369456, CurrentID=0, Status='live', Symbol='LOC105369456', Aliases='', description='uncharacterized LOC105369456', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100054979, end_position_on_the_genomic_accession=100106102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642929, CurrentID=0, Status='live', Symbol='LOC642929', Aliases='', description='general transcription factor II, i pseudogene', other_designations='general transcription factor IIi pseudogene', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66159959, end_position_on_the_genomic_accession=66164906, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369458, CurrentID=0, Status='live', Symbol='LOC105369458', Aliases='', description='uncharacterized LOC105369458', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101320633, end_position_on_the_genomic_accession=101342762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369459, CurrentID=0, Status='live', Symbol='LOC105369459', Aliases='', description='uncharacterized LOC105369459', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101444099, end_position_on_the_genomic_accession=101452699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369460, CurrentID=0, Status='live', Symbol='LOC105369460', Aliases='', description='uncharacterized LOC105369460', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102267961, end_position_on_the_genomic_accession=102281239, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101928254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928254, CurrentID=0, Status='live', Symbol='LOC101928254', Aliases='', description='uncharacterized LOC101928254', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125229579, end_position_on_the_genomic_accession=125264291, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  642934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642934, CurrentID=0, Status='live', Symbol='HOST2', Aliases='lncRNA-HOST2', description='human ovarian cancer-specific transcript 2', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84167228, end_position_on_the_genomic_accession=84172188, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  642935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642935, CurrentID=0, Status='live', Symbol='ADAMTS7P4', Aliases='', description='ADAMTS7 pseudogene 4', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 7 pseudogene', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85255366, end_position_on_the_genomic_accession=85270854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369464, CurrentID=0, Status='live', Symbol='LOC105369464', Aliases='', description='uncharacterized LOC105369464', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=103864987, end_position_on_the_genomic_accession=103904816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369465, CurrentID=0, Status='live', Symbol='LOC105369465', Aliases='', description='uncharacterized LOC105369465', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104072187, end_position_on_the_genomic_accession=104092974, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  642938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642938, CurrentID=0, Status='live', Symbol='FAM196A', Aliases='C10orf141', description='family with sequence similarity 196 member A', other_designations='family with sequence similarity 196, member A', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127135426, end_position_on_the_genomic_accession=127196216, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369467, CurrentID=0, Status='live', Symbol='LOC105369467', Aliases='', description='uncharacterized LOC105369467', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105080453, end_position_on_the_genomic_accession=105082393, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369468, CurrentID=0, Status='live', Symbol='LOC105369468', Aliases='', description='uncharacterized LOC105369468', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105370792, end_position_on_the_genomic_accession=105541338, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369469, CurrentID=0, Status='live', Symbol='LOC105369469', Aliases='', description='uncharacterized LOC105369469', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=105162822, end_position_on_the_genomic_accession=105184806, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  257093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257093, CurrentID=0, Status='live', Symbol='PSME2P1', Aliases='PSME2P', description='proteasome activator subunit 2 pseudogene 1', other_designations='proteasome (prosome, macropain) activator subunit 2 pseudogene', map_location='5q21', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98213373, end_position_on_the_genomic_accession=98214165, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642943, CurrentID=0, Status='live', Symbol='LOC642943', Aliases='', description='SPATA31 subfamily A pseudogene', other_designations='SPATA31 subfamily A, member 4 pseudogene|family with sequence similarity 75, member A1 pseudogene', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92882567, end_position_on_the_genomic_accession=92889555, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369472, CurrentID=0, Status='live', Symbol='LOC105369472', Aliases='', description='uncharacterized LOC105369472', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=106085697, end_position_on_the_genomic_accession=106094509, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100126593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126593, CurrentID=0, Status='live', Symbol='MAFD4', Aliases='BPAD', description='Major affective disorder 4', other_designations='', map_location='16p12', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611247, no_hdr0=''),
  642946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642946, CurrentID=0, Status='live', Symbol='FLVCR1-AS1', Aliases='LQK1, NCRNA00292', description='FLVCR1 antisense RNA 1 (head to head)', other_designations='FLVCR1 antisense RNA 1 (non-protein coding)', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212856604, end_position_on_the_genomic_accession=212858138, orientation='minus', exon_count=2, OMIM=610864, no_hdr0=''),
  100126595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126595, CurrentID=0, Status='live', Symbol='MGS', Aliases='', description='Mungan syndrome', other_designations='', map_location='8q23-q24', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611376, no_hdr0=''),
  100126596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126596, CurrentID=0, Status='live', Symbol='CUX2P1', Aliases='', description='cut-like homeobox 2 pseudogene 1', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10257211, end_position_on_the_genomic_accession=10257376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126597, CurrentID=0, Status='live', Symbol='CUX2P2', Aliases='', description='cut-like homeobox 2 pseudogene 2', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369478, CurrentID=0, Status='live', Symbol='LOC105369478', Aliases='', description='uncharacterized LOC105369478', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118663, CurrentID=0, Status='live', Symbol='BTBD16', Aliases='C10orf87', description='BTB (POZ) domain containing 16', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122270828, end_position_on_the_genomic_accession=122338161, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  642952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642952, CurrentID=0, Status='live', Symbol='SNX18P2', Aliases='', description='sorting nexin 18 pseudogene 2', other_designations='sorting nexin associated golgi protein 1 pseudogene', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92893668, end_position_on_the_genomic_accession=92894505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369481, CurrentID=0, Status='live', Symbol='LOC105369481', Aliases='', description='uncharacterized LOC105369481', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109741667, end_position_on_the_genomic_accession=109746620, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642954, CurrentID=0, Status='live', Symbol='RBBP4P1', Aliases='RBBP4P', description='retinoblastoma binding protein 4 pseudogene 1', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14796181, end_position_on_the_genomic_accession=14798485, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369483, CurrentID=0, Status='live', Symbol='LOC105369483', Aliases='', description='uncharacterized LOC105369483', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109859559, end_position_on_the_genomic_accession=109860587, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  642956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642956, CurrentID=0, Status='live', Symbol='FABP5P9', Aliases='FABP5L9', description='fatty acid binding protein 5 pseudogene 9', other_designations='fatty acid binding protein 5-like 9', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85415167, end_position_on_the_genomic_accession=85415852, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369485, CurrentID=0, Status='live', Symbol='LOC105369485', Aliases='', description='uncharacterized LOC105369485', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110086060, end_position_on_the_genomic_accession=110088107, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  118670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118670, CurrentID=0, Status='live', Symbol='FAM24A', Aliases='', description='family with sequence similarity 24 member A', other_designations='family with sequence similarity 24, member A', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122910701, end_position_on_the_genomic_accession=122913111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  642959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642959, CurrentID=0, Status='live', Symbol='ALOX15P2', Aliases='', description='arachidonate 15-lipoxygenase pseudogene 2', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92913159, end_position_on_the_genomic_accession=92921557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  118672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118672, CurrentID=0, Status='live', Symbol='PSTK', Aliases='C10orf89', description='phosphoseryl-tRNA kinase', other_designations='O-phosphoseryl-tRNA(Sec) kinase', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122954235, end_position_on_the_genomic_accession=122997519, orientation='plus', exon_count=9, OMIM=611310, no_hdr0=''),
  105369489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369489, CurrentID=0, Status='live', Symbol='LOC105369489', Aliases='', description='uncharacterized LOC105369489', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111000830, end_position_on_the_genomic_accession=111021362, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369490, CurrentID=0, Status='live', Symbol='LOC105369490', Aliases='', description='uncharacterized LOC105369490', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111089882, end_position_on_the_genomic_accession=111097319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369491, CurrentID=0, Status='live', Symbol='LOC105369491', Aliases='', description='uncharacterized LOC105369491', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111320987, end_position_on_the_genomic_accession=111323074, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369492, CurrentID=0, Status='live', Symbol='LOC105369492', Aliases='', description='uncharacterized LOC105369492', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111466099, end_position_on_the_genomic_accession=111473436, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369493, CurrentID=0, Status='live', Symbol='LOC105369493', Aliases='', description='uncharacterized LOC105369493', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111670915, end_position_on_the_genomic_accession=111673314, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369494, CurrentID=0, Status='live', Symbol='LOC105369494', Aliases='', description='uncharacterized LOC105369494', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111915563, end_position_on_the_genomic_accession=111918618, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369495, CurrentID=0, Status='live', Symbol='LOC105369495', Aliases='', description='uncharacterized LOC105369495', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112097458, end_position_on_the_genomic_accession=112119132, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  642968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642968, CurrentID=0, Status='live', Symbol='FAM163B', Aliases='C9orf166', description='family with sequence similarity 163 member B', other_designations='family with sequence similarity 163, member B', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133577076, end_position_on_the_genomic_accession=133609920, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  642969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642969, CurrentID=0, Status='live', Symbol='LOC642969', Aliases='', description='phosphoglycerate mutase 1 (brain) pseudogene', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93639903, end_position_on_the_genomic_accession=93641619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369498, CurrentID=0, Status='live', Symbol='LOC105369498', Aliases='', description='uncharacterized LOC105369498', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112822542, end_position_on_the_genomic_accession=112835383, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379823, CurrentID=0, Status='live', Symbol='LOC105379823', Aliases='', description='uncharacterized LOC105379823', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369500, CurrentID=0, Status='live', Symbol='LOC105369500', Aliases='', description='uncharacterized LOC105369500', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113475669, end_position_on_the_genomic_accession=113479567, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375337, CurrentID=0, Status='live', Symbol='LOC105375337', Aliases='', description='uncharacterized LOC105375337', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67342852, end_position_on_the_genomic_accession=67356275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369502, CurrentID=0, Status='live', Symbol='LOC105369502', Aliases='', description='uncharacterized LOC105369502', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114411414, end_position_on_the_genomic_accession=114436319, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  642975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642975, CurrentID=0, Status='live', Symbol='LOC642975', Aliases='', description='transmembrane protein 230 pseudogene', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56690278, end_position_on_the_genomic_accession=56690719, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  642976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642976, CurrentID=0, Status='live', Symbol='GRIK1-AS1', Aliases='C21orf9, GRIK1-AS, GRIK1AS, NCRNA00110', description='GRIK1 antisense RNA 1', other_designations='GRIK1 antisense RNA (non-protein coding)|GRIK1 antisense RNA 1 (non-protein coding)', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29748175, end_position_on_the_genomic_accession=29764006, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369505, CurrentID=0, Status='live', Symbol='LOC105369505', Aliases='', description='uncharacterized LOC105369505', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114829405, end_position_on_the_genomic_accession=114831663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369506, CurrentID=0, Status='live', Symbol='LOC105369506', Aliases='', description='uncharacterized LOC105369506', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114779671, end_position_on_the_genomic_accession=114872773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369507, CurrentID=0, Status='live', Symbol='LOC105369507', Aliases='', description='uncharacterized LOC105369507', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115396723, end_position_on_the_genomic_accession=115397660, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369508, CurrentID=0, Status='live', Symbol='LOC105369508', Aliases='', description='uncharacterized LOC105369508', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115363210, end_position_on_the_genomic_accession=115378717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369509, CurrentID=0, Status='live', Symbol='LOC105369509', Aliases='', description='uncharacterized LOC105369509', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115333541, end_position_on_the_genomic_accession=115340262, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369510, CurrentID=0, Status='live', Symbol='LOC105369510', Aliases='', description='uncharacterized LOC105369510', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115331884, end_position_on_the_genomic_accession=115333244, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379825, CurrentID=0, Status='live', Symbol='LOC105379825', Aliases='', description='uncharacterized LOC105379825', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369512, CurrentID=0, Status='live', Symbol='LOC105369512', Aliases='', description='uncharacterized LOC105369512', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115900743, end_position_on_the_genomic_accession=115901736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369513, CurrentID=0, Status='live', Symbol='LOC105369513', Aliases='', description='uncharacterized LOC105369513', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116033385, end_position_on_the_genomic_accession=116081056, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369514, CurrentID=0, Status='live', Symbol='LOC105369514', Aliases='', description='uncharacterized LOC105369514', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116773236, end_position_on_the_genomic_accession=116773989, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  642987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642987, CurrentID=0, Status='live', Symbol='TMEM232', Aliases='', description='transmembrane protein 232', other_designations='', map_location='5q22.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110387273, end_position_on_the_genomic_accession=110738921, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  105369516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369516, CurrentID=0, Status='live', Symbol='LOC105369516', Aliases='', description='uncharacterized LOC105369516', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117496578, end_position_on_the_genomic_accession=117503835, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369517, CurrentID=0, Status='live', Symbol='LOC105369517', Aliases='', description='uncharacterized LOC105369517', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118092347, end_position_on_the_genomic_accession=118098978, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369518, CurrentID=0, Status='live', Symbol='LOC105369518', Aliases='', description='uncharacterized LOC105369518', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118565068, end_position_on_the_genomic_accession=118568236, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369519, CurrentID=0, Status='live', Symbol='LOC105369519', Aliases='', description='uncharacterized LOC105369519', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118692005, end_position_on_the_genomic_accession=118709304, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369520, CurrentID=0, Status='live', Symbol='LOC105369520', Aliases='', description='uncharacterized LOC105369520', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119642371, end_position_on_the_genomic_accession=119659530, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369521, CurrentID=0, Status='live', Symbol='LOC105369521', Aliases='', description='uncharacterized LOC105369521', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119581945, end_position_on_the_genomic_accession=119587369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369522, CurrentID=0, Status='live', Symbol='LOC105369522', Aliases='', description='uncharacterized LOC105369522', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119576371, end_position_on_the_genomic_accession=119579909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379827, CurrentID=0, Status='live', Symbol='LOC105379827', Aliases='', description='uncharacterized LOC105379827', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  642996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=642996, CurrentID=0, Status='live', Symbol='NFE2L3P1', Aliases='', description='nuclear factor, erythroid 2 like 3 pseudogene 1', other_designations='nuclear factor (erythroid-derived 2)-like 3 pseudogene 1|nuclear factor, erythroid 2-like 3 pseudogene 1|nuclear factor, erythroid-derived 2, like 3 pseudogene', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59967808, end_position_on_the_genomic_accession=59972107, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369525, CurrentID=0, Status='live', Symbol='LOC105369525', Aliases='', description='uncharacterized LOC105369525', other_designations='CTD-2523D13.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119709852, end_position_on_the_genomic_accession=119720455, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369526, CurrentID=0, Status='live', Symbol='LOC105369526', Aliases='', description='uncharacterized LOC105369526', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119905355, end_position_on_the_genomic_accession=119951015, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105369527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369527, CurrentID=0, Status='live', Symbol='LOC105369527', Aliases='', description='uncharacterized LOC105369527', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120026795, end_position_on_the_genomic_accession=120028995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369528, CurrentID=0, Status='live', Symbol='LOC105369528', Aliases='', description='uncharacterized LOC105369528', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119987958, end_position_on_the_genomic_accession=119994851, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643001, CurrentID=0, Status='live', Symbol='DUX4L17', Aliases='DUXY3', description='double homeobox 4 like 17', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11314921, end_position_on_the_genomic_accession=11316345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369530, CurrentID=0, Status='live', Symbol='LOC105369530', Aliases='', description='uncharacterized LOC105369530', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120188764, end_position_on_the_genomic_accession=120189452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643003, CurrentID=0, Status='live', Symbol='RPS26P28', Aliases='RPS26_11_544', description='ribosomal protein S26 pseudogene 28', other_designations='hCG1989844', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16902227, end_position_on_the_genomic_accession=16902666, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369532, CurrentID=0, Status='live', Symbol='LOC105369532', Aliases='', description='uncharacterized LOC105369532', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=120811252, end_position_on_the_genomic_accession=120819799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369533, CurrentID=0, Status='live', Symbol='LOC105369533', Aliases='', description='uncharacterized LOC105369533', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121237338, end_position_on_the_genomic_accession=121245159, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369534, CurrentID=0, Status='live', Symbol='LOC105369534', Aliases='', description='uncharacterized LOC105369534', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=121245199, end_position_on_the_genomic_accession=121247035, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643007, CurrentID=0, Status='live', Symbol='RPL36AP8', Aliases='RPL36A_20_1424', description='ribosomal protein L36a pseudogene 8', other_designations='hCG1787519', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34943051, end_position_on_the_genomic_accession=34943447, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643008, CurrentID=0, Status='live', Symbol='SMIM5', Aliases='C17orf109, PP12104', description='small integral membrane protein 5', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75633434, end_position_on_the_genomic_accession=75641406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369537, CurrentID=0, Status='live', Symbol='LOC105369537', Aliases='', description='uncharacterized LOC105369537', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122066451, end_position_on_the_genomic_accession=122070156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369538, CurrentID=0, Status='live', Symbol='LOC105369538', Aliases='', description='uncharacterized LOC105369538', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122748337, end_position_on_the_genomic_accession=122757325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369539, CurrentID=0, Status='live', Symbol='LOC105369539', Aliases='', description='uncharacterized LOC105369539', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123206432, end_position_on_the_genomic_accession=123207336, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369540, CurrentID=0, Status='live', Symbol='LOC105369540', Aliases='', description='uncharacterized LOC105369540', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123204312, end_position_on_the_genomic_accession=123208396, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379830, CurrentID=0, Status='live', Symbol='LOC105379830', Aliases='', description='uncharacterized LOC105379830', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369542, CurrentID=0, Status='live', Symbol='LOC105369542', Aliases='', description='uncharacterized LOC105369542', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123425085, end_position_on_the_genomic_accession=123431036, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369543, CurrentID=0, Status='live', Symbol='LOC105369543', Aliases='', description='uncharacterized LOC105369543', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123653977, end_position_on_the_genomic_accession=123672257, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369544, CurrentID=0, Status='live', Symbol='LOC105369544', Aliases='', description='uncharacterized LOC105369544', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123797387, end_position_on_the_genomic_accession=123817027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369545, CurrentID=0, Status='live', Symbol='LOC105369545', Aliases='', description='uncharacterized LOC105369545', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124082358, end_position_on_the_genomic_accession=124086660, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369546, CurrentID=0, Status='live', Symbol='LOC105369546', Aliases='', description='uncharacterized LOC105369546', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369547, CurrentID=0, Status='live', Symbol='LOC105369547', Aliases='', description='uncharacterized LOC105369547', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124950383, end_position_on_the_genomic_accession=124954061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369548, CurrentID=0, Status='live', Symbol='LOC105369548', Aliases='', description='uncharacterized LOC105369548', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125405273, end_position_on_the_genomic_accession=125410189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369549, CurrentID=0, Status='live', Symbol='LOC105369549', Aliases='', description='uncharacterized LOC105369549', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125258628, end_position_on_the_genomic_accession=125267323, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369550, CurrentID=0, Status='live', Symbol='STT3A-AS1', Aliases='', description='STT3A antisense RNA 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125570198, end_position_on_the_genomic_accession=125592635, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369551, CurrentID=0, Status='live', Symbol='LOC105369551', Aliases='', description='uncharacterized LOC105369551', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125696119, end_position_on_the_genomic_accession=125699621, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369552, CurrentID=0, Status='live', Symbol='LOC105369552', Aliases='', description='uncharacterized LOC105369552', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125827568, end_position_on_the_genomic_accession=125828337, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379832, CurrentID=0, Status='live', Symbol='LOC105379832', Aliases='', description='uncharacterized LOC105379832', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118738, CurrentID=0, Status='live', Symbol='ZNF488', Aliases='', description='zinc finger protein 488', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47365496, end_position_on_the_genomic_accession=47384333, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105369555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369555, CurrentID=0, Status='live', Symbol='LOC105369555', Aliases='', description='uncharacterized LOC105369555', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=125939136, end_position_on_the_genomic_accession=125943674, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369556, CurrentID=0, Status='live', Symbol='LOC105369556', Aliases='', description='uncharacterized LOC105369556', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126100501, end_position_on_the_genomic_accession=126106217, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369557, CurrentID=0, Status='live', Symbol='HP11014', Aliases='', description='uncharacterized LOC105369557', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126294298, end_position_on_the_genomic_accession=126297845, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369558, CurrentID=0, Status='live', Symbol='LOC105369558', Aliases='', description='uncharacterized LOC105369558', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126992210, end_position_on_the_genomic_accession=126995458, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643031, CurrentID=0, Status='live', Symbol='MTND5P10', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 5 pseudogene 10', other_designations='MT-ND5 pseudogene 10|mitochondrially encoded NADH dehydrogenase 5 pseudogene', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100046974, end_position_on_the_genomic_accession=100048785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369560, CurrentID=0, Status='live', Symbol='LOC105369560', Aliases='', description='uncharacterized LOC105369560', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126869368, end_position_on_the_genomic_accession=126873252, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369561, CurrentID=0, Status='live', Symbol='LOC105369561', Aliases='', description='uncharacterized LOC105369561', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126728170, end_position_on_the_genomic_accession=126731990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643034, CurrentID=0, Status='live', Symbol='DUX4L19', Aliases='DUXY4', description='double homeobox 4 like 19', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11332331, end_position_on_the_genomic_accession=11333764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369563, CurrentID=0, Status='live', Symbol='LOC105369563', Aliases='', description='uncharacterized LOC105369563', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128167198, end_position_on_the_genomic_accession=128184046, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643036, CurrentID=0, Status='live', Symbol='SLED1', Aliases='', description='proteoglycan 3 pseudogene', other_designations='RTFV9368', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184798296, end_position_on_the_genomic_accession=184799046, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643037, CurrentID=0, Status='live', Symbol='C11orf97', Aliases='LINC01171', description='chromosome 11 open reading frame 97', other_designations='long intergenic non-protein coding RNA 1171', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94512529, end_position_on_the_genomic_accession=94532123, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369566, CurrentID=0, Status='live', Symbol='LOC105369566', Aliases='', description='uncharacterized LOC105369566', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128413020, end_position_on_the_genomic_accession=128459813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106865375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106865375, CurrentID=0, Status='live', Symbol='RNA5SP334', Aliases='RN5S334', description='RNA, 5S ribosomal pseudogene 334', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369568, CurrentID=0, Status='live', Symbol='LOC105369568', Aliases='', description='uncharacterized LOC105369568', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128610399, end_position_on_the_genomic_accession=128621889, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369569, CurrentID=0, Status='live', Symbol='LOC105369569', Aliases='', description='uncharacterized LOC105369569', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128714936, end_position_on_the_genomic_accession=128721378, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369570, CurrentID=0, Status='live', Symbol='LOC105369570', Aliases='', description='uncharacterized LOC105369570', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128868844, end_position_on_the_genomic_accession=128887587, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369571, CurrentID=0, Status='live', Symbol='LOC105369571', Aliases='', description='uncharacterized LOC105369571', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129299742, end_position_on_the_genomic_accession=129302906, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369572, CurrentID=0, Status='live', Symbol='LOC105369572', Aliases='', description='uncharacterized LOC105369572', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129307296, end_position_on_the_genomic_accession=129331215, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126693, CurrentID=0, Status='live', Symbol='LINC00322', Aliases='C21orf136, NCRNA00322', description='long intergenic non-protein coding RNA 322', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43322209, end_position_on_the_genomic_accession=43332039, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369574, CurrentID=0, Status='live', Symbol='LOC105369574', Aliases='', description='uncharacterized LOC105369574', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130547221, end_position_on_the_genomic_accession=130558466, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101928273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928273, CurrentID=0, Status='live', Symbol='LOC101928273', Aliases='', description='uncharacterized LOC101928273', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=137878754, end_position_on_the_genomic_accession=137927932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369576, CurrentID=0, Status='live', Symbol='LOC105369576', Aliases='', description='uncharacterized LOC105369576', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130772355, end_position_on_the_genomic_accession=130785488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100689073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100689073, CurrentID=0, Status='live', Symbol='ARHGAP22-IT1', Aliases='', description='ARHGAP22 intronic transcript 1', other_designations='ARHGAP22 intronic transcript 1 (non-protein coding)', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643050, CurrentID=0, Status='live', Symbol='ACTR3BP1', Aliases='', description='ACTR3B pseudogene 1', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11395873, end_position_on_the_genomic_accession=11397516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369579, CurrentID=0, Status='live', Symbol='LOC105369579', Aliases='', description='uncharacterized LOC105369579', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=131074923, end_position_on_the_genomic_accession=131083468, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369580, CurrentID=0, Status='live', Symbol='LOC105369580', Aliases='', description='uncharacterized LOC105369580', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=132959935, end_position_on_the_genomic_accession=132983331, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369581, CurrentID=0, Status='live', Symbol='LOC105369581', Aliases='', description='uncharacterized LOC105369581', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=132736139, end_position_on_the_genomic_accession=132762196, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369582, CurrentID=0, Status='live', Symbol='LOC105369582', Aliases='', description='uncharacterized LOC105369582', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134035454, end_position_on_the_genomic_accession=134036981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379837, CurrentID=0, Status='live', Symbol='LOC105379837', Aliases='', description='uncharacterized LOC105379837', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369584, CurrentID=0, Status='live', Symbol='LOC105369584', Aliases='', description='uncharacterized LOC105369584', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134712438, end_position_on_the_genomic_accession=134715922, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369585, CurrentID=0, Status='live', Symbol='LOC105369585', Aliases='', description='uncharacterized LOC105369585', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134805768, end_position_on_the_genomic_accession=134807514, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369586, CurrentID=0, Status='live', Symbol='LOC105369586', Aliases='', description='uncharacterized LOC105369586', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134950485, end_position_on_the_genomic_accession=134951568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369587, CurrentID=0, Status='live', Symbol='LOC105369587', Aliases='', description='uncharacterized LOC105369587', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=135005163, end_position_on_the_genomic_accession=135010928, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369588, CurrentID=0, Status='live', Symbol='LOC105369588', Aliases='', description='uncharacterized LOC105369588', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=134996907, end_position_on_the_genomic_accession=135007912, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369589, CurrentID=0, Status='live', Symbol='LOC105369589', Aliases='', description='uncharacterized LOC105369589', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=135072411, end_position_on_the_genomic_accession=135076140, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369590, CurrentID=0, Status='live', Symbol='LOC105369590', Aliases='', description='uncharacterized LOC105369590', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119317224, end_position_on_the_genomic_accession=119321858, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369591, CurrentID=0, Status='live', Symbol='LOC105369591', Aliases='', description='uncharacterized LOC105369591', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=126111558, end_position_on_the_genomic_accession=126140687, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105369592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369592, CurrentID=0, Status='live', Symbol='LOC105369592', Aliases='', description='uncharacterized LOC105369592', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=163635, end_position_on_the_genomic_accession=183016, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369593, CurrentID=0, Status='live', Symbol='LOC105369593', Aliases='', description='uncharacterized LOC105369593', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106523, end_position_on_the_genomic_accession=111926, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369594, CurrentID=0, Status='live', Symbol='LOC105369594', Aliases='', description='uncharacterized LOC105369594', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=389977, end_position_on_the_genomic_accession=401548, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369595, CurrentID=0, Status='live', Symbol='LOC105369595', Aliases='', description='uncharacterized LOC105369595', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=585698, end_position_on_the_genomic_accession=591130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369596, CurrentID=0, Status='live', Symbol='LOC105369596', Aliases='', description='uncharacterized LOC105369596', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=711075, end_position_on_the_genomic_accession=728428, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369597, CurrentID=0, Status='live', Symbol='LOC105369597', Aliases='', description='uncharacterized LOC105369597', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=681742, end_position_on_the_genomic_accession=716584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369598, CurrentID=0, Status='live', Symbol='LOC105369598', Aliases='', description='uncharacterized LOC105369598', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=730918, end_position_on_the_genomic_accession=739279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369599, CurrentID=0, Status='live', Symbol='LOC105369599', Aliases='', description='uncharacterized LOC105369599', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1174065, end_position_on_the_genomic_accession=1174986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369600, CurrentID=0, Status='live', Symbol='LOC105369600', Aliases='', description='uncharacterized LOC105369600', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379840, CurrentID=0, Status='live', Symbol='LOC105379840', Aliases='', description='bcl-2-binding component 3-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369602, CurrentID=0, Status='live', Symbol='LOC105369602', Aliases='', description='uncharacterized LOC105369602', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1800619, end_position_on_the_genomic_accession=1807491, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369603, CurrentID=0, Status='live', Symbol='LOC105369603', Aliases='', description='uncharacterized LOC105369603', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2436763, end_position_on_the_genomic_accession=2451199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  118788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118788, CurrentID=0, Status='live', Symbol='PIK3AP1', Aliases='BCAP', description='phosphoinositide-3-kinase adaptor protein 1', other_designations='B cell adaptor protein|B-cell adapter for phosphoinositide 3-kinase|B-cell phosphoinositide 3-kinase adapter protein 1', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96593312, end_position_on_the_genomic_accession=96720522, orientation='minus', exon_count=19, OMIM=607942, no_hdr0=''),
  105369605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369605, CurrentID=0, Status='live', Symbol='LOC105369605', Aliases='', description='uncharacterized LOC105369605', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3230004, end_position_on_the_genomic_accession=3234890, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369606, CurrentID=0, Status='live', Symbol='LOC105369606', Aliases='', description='uncharacterized LOC105369606', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3315741, end_position_on_the_genomic_accession=3323452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  391169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391169, CurrentID=0, Status='live', Symbol='CDKN2AIPNLP1', Aliases='', description='CDKN2A interacting protein N-terminal like pseudogene 1', other_designations='hCG2040210', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226492438, end_position_on_the_genomic_accession=226493494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369608, CurrentID=0, Status='live', Symbol='LOC105369608', Aliases='', description='uncharacterized LOC105369608', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3907930, end_position_on_the_genomic_accession=3958606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369609, CurrentID=0, Status='live', Symbol='LOC105369609', Aliases='', description='uncharacterized LOC105369609', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3868373, end_position_on_the_genomic_accession=3916586, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369610, CurrentID=0, Status='live', Symbol='LOC105369610', Aliases='', description='uncharacterized LOC105369610', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3964336, end_position_on_the_genomic_accession=3965714, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369611, CurrentID=0, Status='live', Symbol='LOC105369611', Aliases='', description='uncharacterized LOC105369611', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4013027, end_position_on_the_genomic_accession=4019019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369612, CurrentID=0, Status='live', Symbol='LOC105369612', Aliases='', description='uncharacterized LOC105369612', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4173056, end_position_on_the_genomic_accession=4203932, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  129026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129026, CurrentID=0, Status='live', Symbol='LOC129026', Aliases='', description='gamma-glutamyltransferase-like activity 1 pseudogene', other_designations='', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22651599, end_position_on_the_genomic_accession=22654080, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369614, CurrentID=0, Status='live', Symbol='LOC105369614', Aliases='', description='uncharacterized LOC105369614', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4813560, end_position_on_the_genomic_accession=4829582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401127, CurrentID=0, Status='live', Symbol='LOC401127', Aliases='', description='WD repeat domain 5 pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39480255, end_position_on_the_genomic_accession=39481905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369616, CurrentID=0, Status='live', Symbol='LOC105369616', Aliases='', description='uncharacterized LOC105369616', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5198224, end_position_on_the_genomic_accession=5201964, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369617, CurrentID=0, Status='live', Symbol='LOC105369617', Aliases='', description='uncharacterized LOC105369617', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5122364, end_position_on_the_genomic_accession=5379083, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105369618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369618, CurrentID=0, Status='live', Symbol='LOC105369618', Aliases='', description='uncharacterized LOC105369618', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5392429, end_position_on_the_genomic_accession=5404187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379843, CurrentID=0, Status='live', Symbol='LOC105379843', Aliases='', description='uncharacterized LOC105379843', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369620, CurrentID=0, Status='live', Symbol='LOC105369620', Aliases='', description='uncharacterized LOC105369620', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5427318, end_position_on_the_genomic_accession=5431533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369621, CurrentID=0, Status='live', Symbol='LOC105369621', Aliases='', description='uncharacterized LOC105369621', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=5881614, end_position_on_the_genomic_accession=5912929, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369622, CurrentID=0, Status='live', Symbol='LOC105369622', Aliases='', description='uncharacterized LOC105369622', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6171414, end_position_on_the_genomic_accession=6174071, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369623, CurrentID=0, Status='live', Symbol='LOC105369623', Aliases='', description='uncharacterized LOC105369623', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6169855, end_position_on_the_genomic_accession=6180941, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369624, CurrentID=0, Status='live', Symbol='LOC105369624', Aliases='', description='uncharacterized LOC105369624', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6295280, end_position_on_the_genomic_accession=6297489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379844, CurrentID=0, Status='live', Symbol='LOC105379844', Aliases='', description='uncharacterized LOC105379844', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369626, CurrentID=0, Status='live', Symbol='LOC105369626', Aliases='', description='uncharacterized LOC105369626', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6365329, end_position_on_the_genomic_accession=6374352, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100913179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913179, CurrentID=0, Status='live', Symbol='LOC100913179', Aliases='', description='SRA stem-loop interacting RNA binding protein pseudogene', other_designations='', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25686158, end_position_on_the_genomic_accession=25686467, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  118812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118812, CurrentID=0, Status='live', Symbol='MORN4', Aliases='C10orf83, bA548K23.4, rtp', description='MORN repeat containing 4', other_designations='44050 protein|protein 44050|retinophilin|retinophilin homolog', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97614553, end_position_on_the_genomic_accession=97634156, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  118813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118813, CurrentID=0, Status='live', Symbol='ZFYVE27', Aliases='PROTRUDIN, SPG33', description='zinc finger, FYVE domain containing 27', other_designations='zinc finger FYVE domain containing 27', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97737121, end_position_on_the_genomic_accession=97760907, orientation='plus', exon_count=15, OMIM=610243, no_hdr0=''),
  100913182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913182, CurrentID=0, Status='live', Symbol='MARCKSL1P2', Aliases='', description='MARCKS-like 1 pseudogene 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138401839, end_position_on_the_genomic_accession=138403646, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100913183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913183, CurrentID=0, Status='live', Symbol='MARCKSL1P1', Aliases='', description='MARCKS-like 1 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103175373, end_position_on_the_genomic_accession=103177078, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369632, CurrentID=0, Status='live', Symbol='LOC105369632', Aliases='', description='uncharacterized LOC105369632', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6875765, end_position_on_the_genomic_accession=6890191, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369633, CurrentID=0, Status='live', Symbol='LOC105369633', Aliases='', description='uncharacterized LOC105369633', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6923179, end_position_on_the_genomic_accession=6928092, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100913186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913186, CurrentID=0, Status='live', Symbol='BDLN1', Aliases='', description='body length, human homolog of mouse body length 1', other_designations='', map_location='Xq24-q28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300054, no_hdr0=''),
  100913187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100913187, CurrentID=0, Status='live', Symbol='APOBEC3A_B', Aliases='', description='APOBEC3A and APOBEC3B deletion hybrid', other_designations='', map_location='22q13 alternate reference locus', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369636, CurrentID=0, Status='live', Symbol='LOC105369636', Aliases='', description='uncharacterized LOC105369636', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6983016, end_position_on_the_genomic_accession=7006303, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379846, CurrentID=0, Status='live', Symbol='LOC105379846', Aliases='', description='uncharacterized LOC105379846', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369638, CurrentID=0, Status='live', Symbol='LOC105369638', Aliases='', description='uncharacterized LOC105369638', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3302927, end_position_on_the_genomic_accession=3311463, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369639, CurrentID=0, Status='live', Symbol='LOC105369639', Aliases='', description='uncharacterized LOC105369639', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7166713, end_position_on_the_genomic_accession=7188831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369640, CurrentID=0, Status='live', Symbol='LOC105369640', Aliases='', description='uncharacterized LOC105369640', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7324582, end_position_on_the_genomic_accession=7326253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928284, CurrentID=0, Status='live', Symbol='LOC101928284', Aliases='', description='uncharacterized LOC101928284', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42733594, end_position_on_the_genomic_accession=42742422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369642, CurrentID=0, Status='live', Symbol='LOC105369642', Aliases='', description='uncharacterized LOC105369642', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8026328, end_position_on_the_genomic_accession=8027461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643115, CurrentID=0, Status='live', Symbol='KRT223P', Aliases='', description='keratin 223 pseudogene', other_designations='keratin 24 pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40717540, end_position_on_the_genomic_accession=40721932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369644, CurrentID=0, Status='live', Symbol='LOC105369644', Aliases='', description='uncharacterized LOC105369644', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8565567, end_position_on_the_genomic_accession=8567922, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369645, CurrentID=0, Status='live', Symbol='LOC105369645', Aliases='', description='uncharacterized LOC105369645', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8564567, end_position_on_the_genomic_accession=8566019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369646, CurrentID=0, Status='live', Symbol='LOC105369646', Aliases='', description='uncharacterized LOC105369646', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369647, CurrentID=0, Status='live', Symbol='LOC105369647', Aliases='', description='uncharacterized LOC105369647', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9403060, end_position_on_the_genomic_accession=9408752, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646736, CurrentID=0, Status='live', Symbol='LOC646736', Aliases='', description='uncharacterized LOC646736', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=226142794, end_position_on_the_genomic_accession=226180062, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369649, CurrentID=0, Status='live', Symbol='LOC105369649', Aliases='', description='uncharacterized LOC105369649', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9334243, end_position_on_the_genomic_accession=9346756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369650, CurrentID=0, Status='live', Symbol='LOC105369650', Aliases='', description='uncharacterized LOC105369650', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9323031, end_position_on_the_genomic_accession=9324593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369651, CurrentID=0, Status='live', Symbol='LOC105369651', Aliases='', description='uncharacterized LOC105369651', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369652, CurrentID=0, Status='live', Symbol='LOC105369652', Aliases='', description='uncharacterized LOC105369652', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9703626, end_position_on_the_genomic_accession=9711454, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379793, CurrentID=0, Status='live', Symbol='LOC105379793', Aliases='', description='uncharacterized LOC105379793', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643126, CurrentID=0, Status='live', Symbol='TRIM51JP', Aliases='', description='tripartite motif-containing 51J, pseudogene', other_designations='SPRY domain containing 5 pseudogene', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95574891, end_position_on_the_genomic_accession=95580939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130417, CurrentID=0, Status='live', Symbol='LOC100130417', Aliases='', description='uncharacterized LOC100130417', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=916818, end_position_on_the_genomic_accession=919692, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369656, CurrentID=0, Status='live', Symbol='LOC105369656', Aliases='', description='uncharacterized LOC105369656', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10280338, end_position_on_the_genomic_accession=10291285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369657, CurrentID=0, Status='live', Symbol='LOC105369657', Aliases='', description='uncharacterized LOC105369657', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10498721, end_position_on_the_genomic_accession=10537099, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369658, CurrentID=0, Status='live', Symbol='LOC105369658', Aliases='', description='uncharacterized LOC105369658', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10537152, end_position_on_the_genomic_accession=10551823, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  85343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85343, CurrentID=0, Status='live', Symbol='KRTAP2-5P', Aliases='KAP2A, KRTAP2P1', description='keratin associated protein 2-5, pseudogene', other_designations='keratin associated protein 2 pseudogene 1|keratin-associated protein 2A pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41071946, end_position_on_the_genomic_accession=41072346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126780, CurrentID=0, Status='live', Symbol='SNAR-G1', Aliases='', description='small ILF3/NF90-associated RNA G1', other_designations='small NF90-associated RNA G1|small nuclear ILF3/NF90-associated RNA G1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49037020, end_position_on_the_genomic_accession=49037147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126781, CurrentID=0, Status='live', Symbol='SNAR-F', Aliases='', description='small ILF3/NF90-associated RNA F', other_designations='small NF90-associated RNA F|small nuclear ILF3/NF90-associated RNA F', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50604963, end_position_on_the_genomic_accession=50605085, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369662, CurrentID=0, Status='live', Symbol='LOC105369662', Aliases='', description='uncharacterized LOC105369662', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12481716, end_position_on_the_genomic_accession=12484653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369663, CurrentID=0, Status='live', Symbol='LOC105369663', Aliases='', description='uncharacterized LOC105369663', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12721812, end_position_on_the_genomic_accession=12724319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643136, CurrentID=0, Status='live', Symbol='ZC3H11B', Aliases='ZC3HDC11B', description='zinc finger CCCH-type containing 11B pseudogene', other_designations='zinc finger CCCH-type domain containing 11B', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219607927, end_position_on_the_genomic_accession=219613145, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369665, CurrentID=0, Status='live', Symbol='LOC105369665', Aliases='', description='uncharacterized LOC105369665', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13102761, end_position_on_the_genomic_accession=13103445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369666, CurrentID=0, Status='live', Symbol='LOC105369666', Aliases='', description='uncharacterized LOC105369666', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13101356, end_position_on_the_genomic_accession=13102436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379851, CurrentID=0, Status='live', Symbol='LOC105379851', Aliases='', description='uncharacterized LOC105379851', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369668, CurrentID=0, Status='live', Symbol='LOC105369668', Aliases='', description='uncharacterized LOC105369668', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=13582055, end_position_on_the_genomic_accession=13620069, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369669, CurrentID=0, Status='live', Symbol='LOC105369669', Aliases='', description='uncharacterized LOC105369669', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14733577, end_position_on_the_genomic_accession=14770662, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369670, CurrentID=0, Status='live', Symbol='LOC105369670', Aliases='', description='uncharacterized LOC105369670', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14962604, end_position_on_the_genomic_accession=14989661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126791, CurrentID=0, Status='live', Symbol='EGOT', Aliases='EGO, NCRNA00190', description='eosinophil granule ontogeny transcript (non-protein coding)', other_designations='', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4749192, end_position_on_the_genomic_accession=4751590, orientation='minus', exon_count=2, OMIM=611662, no_hdr0=''),
  118856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118856, CurrentID=0, Status='live', Symbol='MMP21', Aliases='MMP-21', description='matrix metallopeptidase 21', other_designations='matrix metalloproteinase 21', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125764939, end_position_on_the_genomic_accession=125785224, orientation='minus', exon_count=8, OMIM=608416, no_hdr0=''),
  100126793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126793, CurrentID=0, Status='live', Symbol='GHRLOS', Aliases='GHRL-AS1, GHRLAS, NCRNA00068', description='ghrelin opposite strand/antisense RNA', other_designations='GHRL antisense RNA 1 (non-protein coding)', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10280952, end_position_on_the_genomic_accession=10293449, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100126794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126794, CurrentID=0, Status='live', Symbol='GAEC1', Aliases='', description='gene amplified in esophageal cancer 1', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612130, no_hdr0=''),
  105369675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369675, CurrentID=0, Status='live', Symbol='LOC105369675', Aliases='', description='uncharacterized LOC105369675', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16060829, end_position_on_the_genomic_accession=16236798, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369676, CurrentID=0, Status='live', Symbol='LOC105369676', Aliases='', description='uncharacterized LOC105369676', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16796594, end_position_on_the_genomic_accession=16854342, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369677, CurrentID=0, Status='live', Symbol='LOC105369677', Aliases='', description='uncharacterized LOC105369677', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16787936, end_position_on_the_genomic_accession=16988633, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100126798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126798, CurrentID=0, Status='live', Symbol='SNAR-A1', Aliases='SNAR-A53113498', description='small ILF3/NF90-associated RNA A1', other_designations='small NF90-associated RNA A53113498|small nuclear ILF3/NF90-associated RNA A1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47918429, end_position_on_the_genomic_accession=47918550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126799, CurrentID=0, Status='live', Symbol='SNAR-A2', Aliases='SNAR-A53129250', description='small ILF3/NF90-associated RNA A2', other_designations='small NF90-associated RNA A53129250|small nuclear ILF3/NF90-associated RNA A2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47934181, end_position_on_the_genomic_accession=47934302, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100126800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126800, CurrentID=0, Status='live', Symbol='SNAR-A12', Aliases='SNAR-A53102747', description='small ILF3/NF90-associated RNA A12', other_designations='small NF90-associated RNA AA53102747|small nuclear ILF3/NF90-associated RNA A12', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47907678, end_position_on_the_genomic_accession=47907799, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369681, CurrentID=0, Status='live', Symbol='LOC105369681', Aliases='', description='uncharacterized LOC105369681', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19276604, end_position_on_the_genomic_accession=19306609, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369682, CurrentID=0, Status='live', Symbol='LOC105369682', Aliases='', description='uncharacterized LOC105369682', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19843747, end_position_on_the_genomic_accession=19850510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643155, CurrentID=0, Status='live', Symbol='SMIM15', Aliases='C5orf43', description='small integral membrane protein 15', other_designations='UPF0542 protein C5orf43', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61157709, end_position_on_the_genomic_accession=61162475, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369684, CurrentID=0, Status='live', Symbol='LOC105369684', Aliases='', description='uncharacterized LOC105369684', other_designations='CTC-465D4.1', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20121029, end_position_on_the_genomic_accession=20136259, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  643157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643157, CurrentID=0, Status='live', Symbol='MEP1AP4', Aliases='', description='meprin A subunit alpha pseudogene 4', other_designations='meprin A, alpha (PABA peptide hydrolase) pseudogene|meprin A, alpha pseudogene 4', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42816451, end_position_on_the_genomic_accession=42823212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369686, CurrentID=0, Status='live', Symbol='LOC105369686', Aliases='', description='uncharacterized LOC105369686', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=20192444, end_position_on_the_genomic_accession=20211926, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643159, CurrentID=0, Status='live', Symbol='ATF4P3', Aliases='ATF4C', description='activating transcription factor 4 pseudogene 3', other_designations='activating transcription factor 4C', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76225665, end_position_on_the_genomic_accession=76227086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643160, CurrentID=0, Status='live', Symbol='CYMP', Aliases='', description='chymosin pseudogene', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110480766, end_position_on_the_genomic_accession=110491269, orientation='plus', exon_count=9, OMIM=118943, no_hdr0=''),
  643161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643161, CurrentID=0, Status='live', Symbol='FAM25A', Aliases='bA96C23.5', description='family with sequence similarity 25 member A', other_designations='family with sequence similarity 25, member A|protein FAM25', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87020289, end_position_on_the_genomic_accession=87024730, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643162, CurrentID=0, Status='live', Symbol='DKFZP779L1853', Aliases='', description='uncharacterized LOC643162', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=48592146, end_position_on_the_genomic_accession=48595814, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379855, CurrentID=0, Status='live', Symbol='LOC105379855', Aliases='', description='uncharacterized LOC105379855', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369692, CurrentID=0, Status='live', Symbol='LOC105369692', Aliases='', description='uncharacterized LOC105369692', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22578297, end_position_on_the_genomic_accession=22599815, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369693, CurrentID=0, Status='live', Symbol='LOC105369693', Aliases='', description='uncharacterized LOC105369693', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22859178, end_position_on_the_genomic_accession=22972463, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369694, CurrentID=0, Status='live', Symbol='LOC105369694', Aliases='', description='uncharacterized LOC105369694', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=23046113, end_position_on_the_genomic_accession=23126723, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369695, CurrentID=0, Status='live', Symbol='LOC105369695', Aliases='', description='uncharacterized LOC105369695', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24049663, end_position_on_the_genomic_accession=24081218, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643168, CurrentID=0, Status='live', Symbol='LOC643168', Aliases='', description='Rho GTPase activating protein 5 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54933381, end_position_on_the_genomic_accession=54939640, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  118881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118881, CurrentID=0, Status='live', Symbol='COMTD1', Aliases='', description='catechol-O-methyltransferase domain containing 1', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75233971, end_position_on_the_genomic_accession=75236057, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369698, CurrentID=0, Status='live', Symbol='LOC105369698', Aliases='', description='uncharacterized LOC105369698', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643171, CurrentID=0, Status='live', Symbol='GGT4P', Aliases='', description='gamma-glutamyltransferase 4 pseudogene', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18260350, end_position_on_the_genomic_accession=18271424, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369700, CurrentID=0, Status='live', Symbol='LOC105369700', Aliases='', description='uncharacterized LOC105369700', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369701, CurrentID=0, Status='live', Symbol='LOC105369701', Aliases='', description='uncharacterized LOC105369701', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25342605, end_position_on_the_genomic_accession=25357942, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369702, CurrentID=0, Status='live', Symbol='LOC105369702', Aliases='', description='uncharacterized LOC105369702', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26008337, end_position_on_the_genomic_accession=26011570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369703, CurrentID=0, Status='live', Symbol='LOC105369703', Aliases='', description='uncharacterized LOC105369703', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25944856, end_position_on_the_genomic_accession=25947557, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369704, CurrentID=0, Status='live', Symbol='LOC105369704', Aliases='', description='uncharacterized LOC105369704', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26273329, end_position_on_the_genomic_accession=26275988, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100126825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126825, CurrentID=0, Status='live', Symbol='GJA6P', Aliases='', description='gap junction protein alpha 6 pseudogene', other_designations='gap junction protein, alpha 1, 43kDa pseudogene|gap junction protein, alpha 6, pseudogene', map_location='Xp22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18534857, end_position_on_the_genomic_accession=18535789, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369706, CurrentID=0, Status='live', Symbol='LOC105369706', Aliases='', description='uncharacterized LOC105369706', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=26783108, end_position_on_the_genomic_accession=26819012, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369707, CurrentID=0, Status='live', Symbol='LOC105369707', Aliases='', description='uncharacterized LOC105369707', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27315357, end_position_on_the_genomic_accession=27332797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643180, CurrentID=0, Status='live', Symbol='CCT6P3', Aliases='CCT6-3P', description='chaperonin containing TCP1, subunit 6 (zeta) pseudogene 3', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65038354, end_position_on_the_genomic_accession=65074713, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  643181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643181, CurrentID=0, Status='live', Symbol='FADS2P1', Aliases='', description='fatty acid desaturase 2 pseudogene 1', other_designations='Fatty acid desaturase 2-like protein', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56890787, end_position_on_the_genomic_accession=56904509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643182, CurrentID=0, Status='live', Symbol='UBTFL8', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 8 (pseudogene)', other_designations='upstream binding transcription factor, RNA polymerase I pseudogene', map_location='3p26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5079578, end_position_on_the_genomic_accession=5084508, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369711, CurrentID=0, Status='live', Symbol='LOC105369711', Aliases='', description='uncharacterized LOC105369711', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28715346, end_position_on_the_genomic_accession=28795050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369712, CurrentID=0, Status='live', Symbol='LOC105369712', Aliases='', description='uncharacterized LOC105369712', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29052747, end_position_on_the_genomic_accession=29056133, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369713, CurrentID=0, Status='live', Symbol='LOC105369713', Aliases='', description='uncharacterized LOC105369713', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29560065, end_position_on_the_genomic_accession=29562626, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369714, CurrentID=0, Status='live', Symbol='LOC105369714', Aliases='', description='uncharacterized LOC105369714', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29518624, end_position_on_the_genomic_accession=29532657, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379859, CurrentID=0, Status='live', Symbol='LOC105379859', Aliases='', description='uncharacterized LOC105379859', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369716, CurrentID=0, Status='live', Symbol='LOC105369716', Aliases='', description='uncharacterized LOC105369716', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29873420, end_position_on_the_genomic_accession=29890563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369717, CurrentID=0, Status='live', Symbol='LOC105369717', Aliases='', description='uncharacterized LOC105369717', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30230785, end_position_on_the_genomic_accession=30296979, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369718, CurrentID=0, Status='live', Symbol='LOC105369718', Aliases='', description='uncharacterized LOC105369718', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30315009, end_position_on_the_genomic_accession=30320798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375399, CurrentID=0, Status='live', Symbol='LOC105375399', Aliases='', description='uncharacterized LOC105375399', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92627109, end_position_on_the_genomic_accession=92650952, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369720, CurrentID=0, Status='live', Symbol='LOC105369720', Aliases='', description='uncharacterized LOC105369720', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=30615475, end_position_on_the_genomic_accession=30624459, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643193, CurrentID=0, Status='live', Symbol='CALM2P4', Aliases='', description='calmodulin 2 pseudogene 4', other_designations='', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98640809, end_position_on_the_genomic_accession=98641307, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369722, CurrentID=0, Status='live', Symbol='LOC105369722', Aliases='', description='uncharacterized LOC105369722', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31723508, end_position_on_the_genomic_accession=31728860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369723, CurrentID=0, Status='live', Symbol='LOC105369723', Aliases='', description='uncharacterized LOC105369723', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31876966, end_position_on_the_genomic_accession=31887267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369724, CurrentID=0, Status='live', Symbol='LOC105369724', Aliases='', description='uncharacterized LOC105369724', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31748804, end_position_on_the_genomic_accession=31757072, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928298, CurrentID=0, Status='live', Symbol='LOC101928298', Aliases='', description='uncharacterized LOC101928298', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10058722, end_position_on_the_genomic_accession=10063502, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643198, CurrentID=0, Status='live', Symbol='RBPJP6', Aliases='', description='RBPJ pseudogene 6', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42497053, end_position_on_the_genomic_accession=42499323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653333, CurrentID=0, Status='live', Symbol='FAM86B2', Aliases='', description='family with sequence similarity 86 member B2', other_designations='family with sequence similarity 86, member B2|protein FAM86B2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12425614, end_position_on_the_genomic_accession=12436417, orientation='minus', exon_count=10, OMIM=616123, no_hdr0=''),
  105369728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369728, CurrentID=0, Status='live', Symbol='LOC105369728', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9647014, end_position_on_the_genomic_accession=9653535, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369729, CurrentID=0, Status='live', Symbol='ZNF970P', Aliases='', description='zinc finger protein 970, pseudogene', other_designations='zinc finger protein 850 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=37541740, end_position_on_the_genomic_accession=37577517, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369730, CurrentID=0, Status='live', Symbol='LOC105369730', Aliases='', description='uncharacterized LOC105369730', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39087493, end_position_on_the_genomic_accession=39095745, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369731, CurrentID=0, Status='live', Symbol='LOC105369731', Aliases='', description='uncharacterized LOC105369731', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39538794, end_position_on_the_genomic_accession=39550561, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369732, CurrentID=0, Status='live', Symbol='LOC105369732', Aliases='', description='uncharacterized LOC105369732', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40072155, end_position_on_the_genomic_accession=40085255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643205, CurrentID=0, Status='live', Symbol='RPL36P8', Aliases='RPL36_4_490', description='ribosomal protein L36 pseudogene 8', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76036750, end_position_on_the_genomic_accession=76037064, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369734, CurrentID=0, Status='live', Symbol='LOC105369734', Aliases='', description='uncharacterized LOC105369734', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40156032, end_position_on_the_genomic_accession=40168375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369735, CurrentID=0, Status='live', Symbol='LOC105369735', Aliases='', description='uncharacterized LOC105369735', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40207541, end_position_on_the_genomic_accession=40210471, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369736, CurrentID=0, Status='live', Symbol='LOC105369736', Aliases='', description='uncharacterized LOC105369736', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40354500, end_position_on_the_genomic_accession=40443896, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105369737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369737, CurrentID=0, Status='live', Symbol='LOC105369737', Aliases='', description='uncharacterized LOC105369737', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42023611, end_position_on_the_genomic_accession=42026585, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643210, CurrentID=0, Status='live', Symbol='EHMT1-IT1', Aliases='', description='EHMT1 intronic transcript 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137763022, end_position_on_the_genomic_accession=137764772, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379863, CurrentID=0, Status='live', Symbol='LOC105379863', Aliases='', description='uncharacterized LOC105379863', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  118924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118924, CurrentID=0, Status='live', Symbol='FRA10AC1', Aliases='C10orf4, F26C11.1-like, FRA10A', description='fragile site, folic acid type, rare, fra(10)(q23.3) or fra(10)(q24.2) candidate 1', other_designations='fragile site 10q23.3|rare folic acid-type fragile site, FRA(10)(q23.3), candidate gene 1', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93667883, end_position_on_the_genomic_accession=93702572, orientation='minus', exon_count=17, OMIM=608866, no_hdr0=''),
  105369741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369741, CurrentID=0, Status='live', Symbol='LOC105369741', Aliases='', description='uncharacterized LOC105369741', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=44322125, end_position_on_the_genomic_accession=44323258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369742, CurrentID=0, Status='live', Symbol='LOC105369742', Aliases='', description='uncharacterized LOC105369742', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45035134, end_position_on_the_genomic_accession=45045641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369743, CurrentID=0, Status='live', Symbol='LOC105369743', Aliases='', description='uncharacterized LOC105369743', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45378120, end_position_on_the_genomic_accession=45568843, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105369744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369744, CurrentID=0, Status='live', Symbol='LOC105369744', Aliases='', description='uncharacterized LOC105369744', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45595875, end_position_on_the_genomic_accession=45610610, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379864, CurrentID=0, Status='live', Symbol='LOC105379864', Aliases='', description='putative uncharacterized protein C11orf81', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369746, CurrentID=0, Status='live', Symbol='LOC105369746', Aliases='', description='uncharacterized LOC105369746', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47237671, end_position_on_the_genomic_accession=47250459, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369747, CurrentID=0, Status='live', Symbol='LOC105369747', Aliases='', description='uncharacterized LOC105369747', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47305917, end_position_on_the_genomic_accession=47307399, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  118932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118932, CurrentID=0, Status='live', Symbol='ANKRD22', Aliases='', description='ankyrin repeat domain 22', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88819899, end_position_on_the_genomic_accession=88851975, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369749, CurrentID=0, Status='live', Symbol='LOC105369749', Aliases='', description='uncharacterized LOC105369749', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=47846815, end_position_on_the_genomic_accession=47852657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369750, CurrentID=0, Status='live', Symbol='LOC105369750', Aliases='', description='uncharacterized LOC105369750', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48014103, end_position_on_the_genomic_accession=48031876, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379865, CurrentID=0, Status='live', Symbol='LOC105379865', Aliases='', description='uncharacterized LOC105379865', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643224, CurrentID=0, Status='live', Symbol='TUBBP5', Aliases='', description='tubulin beta pseudogene 5', other_designations='HSA9q34 beta-tubulin 4Q (TUBB4Q) pseudogene|tubulin, beta 8 pseudogene|tubulin, beta polypeptide pseudogene 5|tubulin, beta pseudogene 5', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138150113, end_position_on_the_genomic_accession=138177433, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369753, CurrentID=0, Status='live', Symbol='LOC105369753', Aliases='', description='uncharacterized LOC105369753', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48279016, end_position_on_the_genomic_accession=48307443, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643226, CurrentID=0, Status='live', Symbol='GRXCR2', Aliases='DFNB101', description='glutaredoxin, cysteine rich 2', other_designations='GRXCR1-like protein|glutaredoxin domain-containing cysteine-rich protein 1-like protein', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145858387, end_position_on_the_genomic_accession=145937176, orientation='minus', exon_count=6, OMIM=615762, no_hdr0=''),
  101928303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928303, CurrentID=0, Status='live', Symbol='LOC101928303', Aliases='', description='uncharacterized LOC101928303', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26225320, end_position_on_the_genomic_accession=26229840, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643228, CurrentID=0, Status='live', Symbol='GAPDHP30', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 30', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96068519, end_position_on_the_genomic_accession=96069731, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379866, CurrentID=0, Status='live', Symbol='LOC105379866', Aliases='', description='uncharacterized LOC105379866', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369758, CurrentID=0, Status='live', Symbol='LOC105369758', Aliases='', description='uncharacterized LOC105369758', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48998367, end_position_on_the_genomic_accession=49019809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369759, CurrentID=0, Status='live', Symbol='LOC105369759', Aliases='', description='uncharacterized LOC105369759', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49091398, end_position_on_the_genomic_accession=49096969, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643232, CurrentID=0, Status='live', Symbol='NRBF2P3', Aliases='', description='nuclear receptor binding factor 2 pseudogene 3', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110847583, end_position_on_the_genomic_accession=110849072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  118945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118945, CurrentID=0, Status='live', Symbol='CTSLP1', Aliases='CTSL1P1, CTSLL, CTSLL1', description='cathepsin L pseudogene 1', other_designations='cathepsin L-like 1 (pseudogene)|cathepsin L1 pseudogene 1', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87386361, end_position_on_the_genomic_accession=87390215, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369762, CurrentID=0, Status='live', Symbol='LOC105369762', Aliases='', description='uncharacterized LOC105369762', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49707721, end_position_on_the_genomic_accession=49731390, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379867, CurrentID=0, Status='live', Symbol='LOC105379867', Aliases='', description='uncharacterized LOC105379867', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643236, CurrentID=0, Status='live', Symbol='TMEM72', Aliases='C10orf127, KSP37', description='transmembrane protein 72', other_designations='kidney-specific secretory protein of 37 kDa', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44911308, end_position_on_the_genomic_accession=44937010, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369765, CurrentID=0, Status='live', Symbol='LOC105369765', Aliases='', description='uncharacterized LOC105369765', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50744893, end_position_on_the_genomic_accession=50763560, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369766, CurrentID=0, Status='live', Symbol='LOC105369766', Aliases='', description='uncharacterized LOC105369766', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369767, CurrentID=0, Status='live', Symbol='LOC105369767', Aliases='', description='uncharacterized LOC105369767', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51322606, end_position_on_the_genomic_accession=51324008, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369768, CurrentID=0, Status='live', Symbol='LOC105369768', Aliases='', description='uncharacterized LOC105369768', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51466778, end_position_on_the_genomic_accession=51467364, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379868, CurrentID=0, Status='live', Symbol='LOC105379868', Aliases='', description='uncharacterized LOC105379868', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369770, CurrentID=0, Status='live', Symbol='LOC105369770', Aliases='', description='uncharacterized LOC105369770', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52065120, end_position_on_the_genomic_accession=52069966, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369771, CurrentID=0, Status='live', Symbol='LOC105369771', Aliases='', description='uncharacterized LOC105369771', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52105711, end_position_on_the_genomic_accession=52118841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643244, CurrentID=0, Status='live', Symbol='RDXP1', Aliases='', description='radixin pseudogene 1', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4210290, end_position_on_the_genomic_accession=4212301, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369773, CurrentID=0, Status='live', Symbol='LOC105369773', Aliases='', description='uncharacterized LOC105369773', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53106824, end_position_on_the_genomic_accession=53115377, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643246, CurrentID=0, Status='live', Symbol='MAP1LC3B2', Aliases='ATG8G', description='microtubule associated protein 1 light chain 3 beta 2', other_designations='Microtubule-associated proteins 1A/1B light chain 3B-like protein|microtubule-associated protein 1 light chain 3 beta 2', map_location='12q24.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116559208, end_position_on_the_genomic_accession=116576620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369775, CurrentID=0, Status='live', Symbol='LOC105369775', Aliases='', description='uncharacterized LOC105369775', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53953556, end_position_on_the_genomic_accession=53956427, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369776, CurrentID=0, Status='live', Symbol='LOC105369776', Aliases='', description='uncharacterized LOC105369776', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54142602, end_position_on_the_genomic_accession=54147422, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369777, CurrentID=0, Status='live', Symbol='LOC105369777', Aliases='', description='uncharacterized LOC105369777', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54189097, end_position_on_the_genomic_accession=54201845, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369778, CurrentID=0, Status='live', Symbol='LOC105369778', Aliases='', description='uncharacterized protein MGC39606-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54916229, end_position_on_the_genomic_accession=54946248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928307, CurrentID=0, Status='live', Symbol='LOC101928307', Aliases='', description='uncharacterized LOC101928307', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68632663, end_position_on_the_genomic_accession=68634324, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369780, CurrentID=0, Status='live', Symbol='LOC105369780', Aliases='', description='uncharacterized LOC105369780', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56019745, end_position_on_the_genomic_accession=56041712, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  643253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643253, CurrentID=0, Status='live', Symbol='CCT6P1', Aliases='CCT6-5P, CCT6AP1', description='chaperonin containing TCP1, subunit 6 (zeta) pseudogene 1', other_designations='chaperonin containing TCP1, subunit 6A (zeta 1) pseudogene 1|chaperonin containing TCP1, subunit 6A pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65751105, end_position_on_the_genomic_accession=65763675, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105369782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369782, CurrentID=0, Status='live', Symbol='LOC105369782', Aliases='', description='uncharacterized LOC105369782', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56103205, end_position_on_the_genomic_accession=56104474, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643255, CurrentID=0, Status='live', Symbol='RARRES2P1', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 1', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75658515, end_position_on_the_genomic_accession=75659138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369784, CurrentID=0, Status='live', Symbol='LOC105369784', Aliases='', description='uncharacterized LOC105369784', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58081120, end_position_on_the_genomic_accession=58099618, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369785, CurrentID=0, Status='live', Symbol='LOC105369785', Aliases='', description='uncharacterized LOC105369785', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58087892, end_position_on_the_genomic_accession=58093386, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369786, CurrentID=0, Status='live', Symbol='LOC105369786', Aliases='', description='uncharacterized LOC105369786', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58176307, end_position_on_the_genomic_accession=58178151, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379871, CurrentID=0, Status='live', Symbol='LOC105379871', Aliases='', description='uncharacterized LOC105379871', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369788, CurrentID=0, Status='live', Symbol='LOC105369788', Aliases='', description='uncharacterized LOC105369788', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58444557, end_position_on_the_genomic_accession=58463985, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369789, CurrentID=0, Status='live', Symbol='LOC105369789', Aliases='', description='uncharacterized LOC105369789', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58550944, end_position_on_the_genomic_accession=58556861, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369790, CurrentID=0, Status='live', Symbol='LOC105369790', Aliases='', description='uncharacterized LOC105369790', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58588303, end_position_on_the_genomic_accession=58590854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643263, CurrentID=0, Status='live', Symbol='SSU72P6', Aliases='', description='SSU72 pseudogene 6', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4278697, end_position_on_the_genomic_accession=4279292, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643264, CurrentID=0, Status='live', Symbol='RPL21P106', Aliases='RPL21_48_1266', description='ribosomal protein L21 pseudogene 106', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92168366, end_position_on_the_genomic_accession=92168917, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369793, CurrentID=0, Status='live', Symbol='LOC105369793', Aliases='', description='uncharacterized LOC105369793', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61669037, end_position_on_the_genomic_accession=61682235, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369794, CurrentID=0, Status='live', Symbol='LOC105369794', Aliases='', description='uncharacterized LOC105369794', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=61682284, end_position_on_the_genomic_accession=61699884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369795, CurrentID=0, Status='live', Symbol='LOC105369795', Aliases='', description='uncharacterized LOC105369795', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62840361, end_position_on_the_genomic_accession=62896169, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  118980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118980, CurrentID=0, Status='live', Symbol='SFXN2', Aliases='', description='sideroflexin 2', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102714508, end_position_on_the_genomic_accession=102739189, orientation='plus', exon_count=17, OMIM=615570, no_hdr0=''),
  105369797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369797, CurrentID=0, Status='live', Symbol='LOC105369797', Aliases='', description='uncharacterized LOC105369797', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63545475, end_position_on_the_genomic_accession=63549368, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369798, CurrentID=0, Status='live', Symbol='LOC105369798', Aliases='', description='uncharacterized LOC105369798', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64099420, end_position_on_the_genomic_accession=64130267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369799, CurrentID=0, Status='live', Symbol='LOC105369799', Aliases='', description='uncharacterized LOC105369799', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64091935, end_position_on_the_genomic_accession=64098297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369800, CurrentID=0, Status='live', Symbol='LOC105369800', Aliases='', description='uncharacterized LOC105369800', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64091328, end_position_on_the_genomic_accession=64091918, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369801, CurrentID=0, Status='live', Symbol='LOC105369801', Aliases='', description='uncharacterized LOC105369801', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63994882, end_position_on_the_genomic_accession=64020588, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105369802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369802, CurrentID=0, Status='live', Symbol='LOC105369802', Aliases='', description='uncharacterized LOC105369802', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=64469211, end_position_on_the_genomic_accession=64470190, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  118987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=118987, CurrentID=0, Status='live', Symbol='PDZD8', Aliases='PDZK8', description='PDZ domain containing 8', other_designations='sarcoma antigen NY-SAR-84/NY-SAR-104', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117280489, end_position_on_the_genomic_accession=117375467, orientation='minus', exon_count=8, OMIM=614235, no_hdr0=''),
  106480799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480799, CurrentID=0, Status='live', Symbol='FCF1P4', Aliases='', description='FCF1 pseudogene 4', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369805, CurrentID=0, Status='live', Symbol='LOC105369805', Aliases='', description='uncharacterized LOC105369805', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369806, CurrentID=0, Status='live', Symbol='LOC105369806', Aliases='', description='uncharacterized LOC105369806', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65622310, end_position_on_the_genomic_accession=65629787, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369807, CurrentID=0, Status='live', Symbol='LOC105369807', Aliases='', description='uncharacterized LOC105369807', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65582207, end_position_on_the_genomic_accession=65622223, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369808, CurrentID=0, Status='live', Symbol='LOC105369808', Aliases='', description='uncharacterized LOC105369808', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65556993, end_position_on_the_genomic_accession=65560084, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372211, CurrentID=0, Status='live', Symbol='LOC105372211', Aliases='', description='uncharacterized LOC105372211', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76279575, end_position_on_the_genomic_accession=76318803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369810, CurrentID=0, Status='live', Symbol='LOC105369810', Aliases='', description='uncharacterized LOC105369810', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66027693, end_position_on_the_genomic_accession=66032392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369811, CurrentID=0, Status='live', Symbol='LOC105369811', Aliases='', description='uncharacterized LOC105369811', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66362397, end_position_on_the_genomic_accession=66371837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643284, CurrentID=0, Status='live', Symbol='RPS29P18', Aliases='RPS29_14_993', description='ribosomal protein S29 pseudogene 18', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87525263, end_position_on_the_genomic_accession=87525564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369813, CurrentID=0, Status='live', Symbol='LOC105369813', Aliases='', description='uncharacterized LOC105369813', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67380819, end_position_on_the_genomic_accession=67388542, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369814, CurrentID=0, Status='live', Symbol='LOC105369814', Aliases='', description='uncharacterized LOC105369814', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67545386, end_position_on_the_genomic_accession=67547872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369815, CurrentID=0, Status='live', Symbol='LOC105369815', Aliases='', description='uncharacterized LOC105369815', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67570954, end_position_on_the_genomic_accession=67589947, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369816, CurrentID=0, Status='live', Symbol='LOC105369816', Aliases='', description='uncharacterized LOC105369816', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67678217, end_position_on_the_genomic_accession=67683038, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369817, CurrentID=0, Status='live', Symbol='LOC105369817', Aliases='', description='uncharacterized LOC105369817', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67689427, end_position_on_the_genomic_accession=67692216, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369818, CurrentID=0, Status='live', Symbol='LOC105369818', Aliases='', description='uncharacterized LOC105369818', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68198465, end_position_on_the_genomic_accession=68313102, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369819, CurrentID=0, Status='live', Symbol='LOC105369819', Aliases='', description='uncharacterized LOC105369819', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68333666, end_position_on_the_genomic_accession=68337556, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369820, CurrentID=0, Status='live', Symbol='LOC105369820', Aliases='', description='uncharacterized LOC105369820', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68781811, end_position_on_the_genomic_accession=68793968, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369821, CurrentID=0, Status='live', Symbol='LOC105369821', Aliases='', description='uncharacterized LOC105369821', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69212336, end_position_on_the_genomic_accession=69222526, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369822, CurrentID=0, Status='live', Symbol='LOC105369822', Aliases='', description='uncharacterized LOC105369822', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69431551, end_position_on_the_genomic_accession=69461008, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369823, CurrentID=0, Status='live', Symbol='LOC105369823', Aliases='', description='uncharacterized LOC105369823', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=69624433, end_position_on_the_genomic_accession=69639013, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369824, CurrentID=0, Status='live', Symbol='LOC105369824', Aliases='', description='uncharacterized LOC105369824', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70064394, end_position_on_the_genomic_accession=70085135, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369825, CurrentID=0, Status='live', Symbol='LOC105369825', Aliases='', description='uncharacterized LOC105369825', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70254968, end_position_on_the_genomic_accession=70314854, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369826, CurrentID=0, Status='live', Symbol='LOC105369826', Aliases='', description='uncharacterized LOC105369826', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70180333, end_position_on_the_genomic_accession=70201014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369827, CurrentID=0, Status='live', Symbol='LOC105369827', Aliases='', description='uncharacterized LOC105369827', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70468096, end_position_on_the_genomic_accession=70513227, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643300, CurrentID=0, Status='live', Symbol='HSPD1P1', Aliases='HSPD1-2P', description='heat shock protein family D (Hsp60) member 1 pseudogene 1', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 1', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21882517, end_position_on_the_genomic_accession=21884766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369829, CurrentID=0, Status='live', Symbol='LOC105369829', Aliases='', description='uncharacterized LOC105369829', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=70719340, end_position_on_the_genomic_accession=70773765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369830, CurrentID=0, Status='live', Symbol='LOC105369830', Aliases='', description='uncharacterized LOC105369830', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71109841, end_position_on_the_genomic_accession=71118109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369831, CurrentID=0, Status='live', Symbol='LOC105369831', Aliases='', description='uncharacterized LOC105369831', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71217023, end_position_on_the_genomic_accession=71220821, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  119016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119016, CurrentID=0, Status='live', Symbol='AGAP4', Aliases='AGAP-4, AGAP-8, AGAP8, CTGLF1, CTGLF5, MRIP2', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 4', other_designations='ARF GTPase-activating protein|ArfGAP with GTPase domain, ankyrin repeat and PH domain 8|arf-GAP with GTPase, ANK repeat and PH domain-containing protein 8|centaurin, gamma-like family, member 1|centaurin, gamma-like family, member 5', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45825592, end_position_on_the_genomic_accession=45853877, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105369833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369833, CurrentID=0, Status='live', Symbol='LOC105369833', Aliases='', description='uncharacterized LOC105369833', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71527293, end_position_on_the_genomic_accession=71575195, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369834, CurrentID=0, Status='live', Symbol='LOC105369834', Aliases='', description='uncharacterized LOC105369834', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71430074, end_position_on_the_genomic_accession=71450832, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369835, CurrentID=0, Status='live', Symbol='LOC105369835', Aliases='', description='uncharacterized LOC105369835', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71659221, end_position_on_the_genomic_accession=71663006, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643308, CurrentID=0, Status='live', Symbol='RPL7P59', Aliases='', description='ribosomal protein L7 pseudogene 59', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145040351, end_position_on_the_genomic_accession=145040963, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643309, CurrentID=0, Status='live', Symbol='ACTBP1', Aliases='', description='actin, beta pseudogene 1', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46287816, end_position_on_the_genomic_accession=46289206, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643310, CurrentID=0, Status='live', Symbol='ST13P18', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 18', other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=93287736, end_position_on_the_genomic_accession=93288759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643311, CurrentID=0, Status='live', Symbol='CT47B1', Aliases='CT47.13, CT47A13', description='cancer/testis antigen family 47, member B1', other_designations='cancer/testis CT47 family, member 13|cancer/testis antigen 47.13|cancer/testis antigen family 147, member B1', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120872597, end_position_on_the_genomic_accession=120875925, orientation='minus', exon_count=3, OMIM=300790, no_hdr0=''),
  105369840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369840, CurrentID=0, Status='live', Symbol='LOC105369840', Aliases='', description='uncharacterized LOC105369840', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=73758675, end_position_on_the_genomic_accession=73846186, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369841, CurrentID=0, Status='live', Symbol='LOC105369841', Aliases='', description='uncharacterized LOC105369841', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74039417, end_position_on_the_genomic_accession=74048789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643314, CurrentID=0, Status='live', Symbol='KIAA0754', Aliases='', description='KIAA0754', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39409504, end_position_on_the_genomic_accession=39416482, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369843, CurrentID=0, Status='live', Symbol='LOC105369843', Aliases='', description='uncharacterized LOC105369843', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75015939, end_position_on_the_genomic_accession=75025839, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369844, CurrentID=0, Status='live', Symbol='LOC105369844', Aliases='', description='uncharacterized LOC105369844', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75663725, end_position_on_the_genomic_accession=75694971, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369845, CurrentID=0, Status='live', Symbol='LOC105369845', Aliases='', description='uncharacterized LOC105369845', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75524330, end_position_on_the_genomic_accession=75527868, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369846, CurrentID=0, Status='live', Symbol='LOC105369846', Aliases='', description='uncharacterized LOC105369846', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75772272, end_position_on_the_genomic_accession=75780953, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369847, CurrentID=0, Status='live', Symbol='LOC105369847', Aliases='', description='uncharacterized LOC105369847', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76002408, end_position_on_the_genomic_accession=76009591, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  119032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119032, CurrentID=0, Status='live', Symbol='BORCS7', Aliases='C10orf32', description='BLOC-1 related complex subunit 7', other_designations='diaskedin', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102854210, end_position_on_the_genomic_accession=102864961, orientation='plus', exon_count=6, OMIM=616600, no_hdr0=''),
  105369849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369849, CurrentID=0, Status='live', Symbol='LOC105369849', Aliases='', description='uncharacterized LOC105369849', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76293873, end_position_on_the_genomic_accession=76313462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369850, CurrentID=0, Status='live', Symbol='LOC105369850', Aliases='', description='uncharacterized LOC105369850', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76559842, end_position_on_the_genomic_accession=76615406, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369851, CurrentID=0, Status='live', Symbol='LOC105369851', Aliases='', description='uncharacterized LOC105369851', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76636119, end_position_on_the_genomic_accession=76664977, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369852, CurrentID=0, Status='live', Symbol='LOC105369852', Aliases='', description='uncharacterized LOC105369852', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77010832, end_position_on_the_genomic_accession=77013672, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379882, CurrentID=0, Status='live', Symbol='LOC105379882', Aliases='', description='uncharacterized LOC105379882', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369854, CurrentID=0, Status='live', Symbol='LOC105369854', Aliases='', description='uncharacterized LOC105369854', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77219603, end_position_on_the_genomic_accession=77229665, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643327, CurrentID=0, Status='live', Symbol='LOC643327', Aliases='', description='uncharacterized LOC643327', other_designations='', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3458703, end_position_on_the_genomic_accession=3469212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369856, CurrentID=0, Status='live', Symbol='LOC105369856', Aliases='', description='uncharacterized LOC105369856', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77383054, end_position_on_the_genomic_accession=77391277, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643329, CurrentID=0, Status='live', Symbol='KRT18P57', Aliases='', description='keratin 18 pseudogene 57', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111648228, end_position_on_the_genomic_accession=111649597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369858, CurrentID=0, Status='live', Symbol='LOC105369858', Aliases='', description='uncharacterized LOC105369858', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=77330250, end_position_on_the_genomic_accession=77380921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379883, CurrentID=0, Status='live', Symbol='LOC105379883', Aliases='', description='uncharacterized LOC105379883', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643332, CurrentID=0, Status='live', Symbol='ECRP', Aliases='', description='ribonuclease, RNase A family, 2 (liver, eosinophil-derived neurotoxin) pseudogene', other_designations='eosinophil cationic-related protein', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20919341, end_position_on_the_genomic_accession=20920176, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369861, CurrentID=0, Status='live', Symbol='LOC105369861', Aliases='', description='uncharacterized LOC105369861', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78423192, end_position_on_the_genomic_accession=78440788, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369862, CurrentID=0, Status='live', Symbol='LOC105369862', Aliases='', description='uncharacterized LOC105369862', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79090602, end_position_on_the_genomic_accession=79167321, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369863, CurrentID=0, Status='live', Symbol='LOC105369863', Aliases='', description='uncharacterized LOC105369863', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=78905071, end_position_on_the_genomic_accession=79036913, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  643336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643336, CurrentID=0, Status='live', Symbol='CDC42P2', Aliases='', description='cell division cycle 42 pseudogene 2', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55637420, end_position_on_the_genomic_accession=55638915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369865, CurrentID=0, Status='live', Symbol='LOC105369865', Aliases='', description='uncharacterized LOC105369865', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=79935281, end_position_on_the_genomic_accession=79941389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643338, CurrentID=0, Status='live', Symbol='C15orf62', Aliases='', description='chromosome 15 open reading frame 62', other_designations='uncharacterized protein FLJ75552, mitochondrial', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40769961, end_position_on_the_genomic_accession=40772449, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643339, CurrentID=0, Status='live', Symbol='LOC643339', Aliases='', description='uncharacterized LOC643339', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93003758, end_position_on_the_genomic_accession=93377736, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369868, CurrentID=0, Status='live', Symbol='LOC105369868', Aliases='', description='macrophage mannose receptor 1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80400695, end_position_on_the_genomic_accession=80423307, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101928322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928322, CurrentID=0, Status='live', Symbol='LOC101928322', Aliases='', description='uncharacterized LOC101928322', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10419538, end_position_on_the_genomic_accession=10462349, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369870, CurrentID=0, Status='live', Symbol='LOC105369870', Aliases='', description='uncharacterized LOC105369870', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80873588, end_position_on_the_genomic_accession=80886155, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643343, CurrentID=0, Status='live', Symbol='RPL5P4', Aliases='RPL5_1_31', description='ribosomal protein L5 pseudogene 4', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35350648, end_position_on_the_genomic_accession=35351666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369872, CurrentID=0, Status='live', Symbol='LOC105369872', Aliases='', description='uncharacterized LOC105369872', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=81343893, end_position_on_the_genomic_accession=81353236, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369873, CurrentID=0, Status='live', Symbol='LOC105369873', Aliases='', description='uncharacterized LOC105369873', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=82138345, end_position_on_the_genomic_accession=82308315, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369874, CurrentID=0, Status='live', Symbol='LOC105369874', Aliases='', description='uncharacterized LOC105369874', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=83458772, end_position_on_the_genomic_accession=83472608, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369875, CurrentID=0, Status='live', Symbol='LOC105369875', Aliases='', description='uncharacterized LOC105369875', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=84492413, end_position_on_the_genomic_accession=84508846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643348, CurrentID=0, Status='live', Symbol='LOC643348', Aliases='', description='ankyrin repeat domain 26 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55658996, end_position_on_the_genomic_accession=55678477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369877, CurrentID=0, Status='live', Symbol='LOC105369877', Aliases='', description='uncharacterized LOC105369877', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=86264558, end_position_on_the_genomic_accession=86270602, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369878, CurrentID=0, Status='live', Symbol='LOC105369878', Aliases='', description='uncharacterized LOC105369878', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87041916, end_position_on_the_genomic_accession=87055246, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369879, CurrentID=0, Status='live', Symbol='LOC105369879', Aliases='', description='uncharacterized LOC105369879', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87330739, end_position_on_the_genomic_accession=87332321, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369880, CurrentID=0, Status='live', Symbol='LOC105369880', Aliases='', description='uncharacterized LOC105369880', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87429152, end_position_on_the_genomic_accession=87440513, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643353, CurrentID=0, Status='live', Symbol='SAPCD2P3', Aliases='', description='suppressor APC domain containing 2 pseudogene 3', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66556152, end_position_on_the_genomic_accession=66557774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369882, CurrentID=0, Status='live', Symbol='LOC105369882', Aliases='', description='uncharacterized LOC105369882', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87815908, end_position_on_the_genomic_accession=87821541, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643355, CurrentID=0, Status='live', Symbol='LOC643355', Aliases='', description='uncharacterized LOC643355', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111989995, end_position_on_the_genomic_accession=111998842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369884, CurrentID=0, Status='live', Symbol='LOC105369884', Aliases='', description='uncharacterized LOC105369884', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88381174, end_position_on_the_genomic_accession=88419332, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  401136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401136, CurrentID=0, Status='live', Symbol='TMPRSS11BNL', Aliases='', description='TMPRSS11B N-terminal like, pseudogene', other_designations='TMPRSS11B N terminal-like|TMPRSS11B N-terminus-like protein|transmembrane protease serine 11B-like protein|transmembrane protease, serine 11B pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68184129, end_position_on_the_genomic_accession=68218080, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  643358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643358, CurrentID=0, Status='live', Symbol='RPS27AP16', Aliases='RPS27A_8_1488', description='ribosomal protein S27a pseudogene 16', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61055415, end_position_on_the_genomic_accession=61055931, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369887, CurrentID=0, Status='live', Symbol='LOC105369887', Aliases='', description='uncharacterized LOC105369887', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=88781950, end_position_on_the_genomic_accession=88784285, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369888, CurrentID=0, Status='live', Symbol='LOC105369888', Aliases='', description='uncharacterized LOC105369888', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89639581, end_position_on_the_genomic_accession=89647020, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369889, CurrentID=0, Status='live', Symbol='LOC105369889', Aliases='', description='uncharacterized LOC105369889', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89352603, end_position_on_the_genomic_accession=89455111, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369890, CurrentID=0, Status='live', Symbol='LOC105369890', Aliases='', description='uncharacterized LOC105369890', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89919422, end_position_on_the_genomic_accession=90112278, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105369891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369891, CurrentID=0, Status='live', Symbol='LOC105369891', Aliases='', description='uncharacterized LOC105369891', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89947600, end_position_on_the_genomic_accession=89949726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369892, CurrentID=0, Status='live', Symbol='LOC105369892', Aliases='', description='uncharacterized LOC105369892', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90253825, end_position_on_the_genomic_accession=90263894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643365, CurrentID=0, Status='live', Symbol='LINC00452', Aliases='LINC00453', description='long intergenic non-protein coding RNA 452', other_designations='long intergenic non-protein coding RNA 453', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113883717, end_position_on_the_genomic_accession=113924812, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105369894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369894, CurrentID=0, Status='live', Symbol='LOC105369894', Aliases='', description='uncharacterized LOC105369894', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=90576280, end_position_on_the_genomic_accession=90631103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643367, CurrentID=0, Status='live', Symbol='LOC643367', Aliases='', description='POM121 membrane glycoprotein (rat) pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70213756, end_position_on_the_genomic_accession=70218079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369896, CurrentID=0, Status='live', Symbol='LOC105369896', Aliases='', description='uncharacterized LOC105369896', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91365634, end_position_on_the_genomic_accession=91399373, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369897, CurrentID=0, Status='live', Symbol='LOC105369897', Aliases='', description='uncharacterized LOC105369897', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91326960, end_position_on_the_genomic_accession=91368601, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369898, CurrentID=0, Status='live', Symbol='LOC105369898', Aliases='', description='uncharacterized LOC105369898', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91228960, end_position_on_the_genomic_accession=91365436, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928327, CurrentID=0, Status='live', Symbol='LOC101928327', Aliases='', description='uncharacterized LOC101928327', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216870772, end_position_on_the_genomic_accession=216871639, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369900, CurrentID=0, Status='live', Symbol='LOC105369900', Aliases='', description='uncharacterized LOC105369900', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91875303, end_position_on_the_genomic_accession=91880655, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369901, CurrentID=0, Status='live', Symbol='LOC105369901', Aliases='', description='uncharacterized LOC105369901', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91918366, end_position_on_the_genomic_accession=91961098, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369902, CurrentID=0, Status='live', Symbol='LOC105369902', Aliases='', description='uncharacterized LOC105369902', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92349345, end_position_on_the_genomic_accession=92388403, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369903, CurrentID=0, Status='live', Symbol='LOC105369903', Aliases='', description='uncharacterized LOC105369903', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92272642, end_position_on_the_genomic_accession=92273897, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643376, CurrentID=0, Status='live', Symbol='BTBD18', Aliases='', description='BTB (POZ) domain containing 18', other_designations='hCG1730474', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57743514, end_position_on_the_genomic_accession=57753256, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369905, CurrentID=0, Status='live', Symbol='LOC105369905', Aliases='', description='uncharacterized LOC105369905', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92557249, end_position_on_the_genomic_accession=92629891, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369906, CurrentID=0, Status='live', Symbol='LOC105369906', Aliases='', description='uncharacterized LOC105369906', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=92989824, end_position_on_the_genomic_accession=93020571, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369907, CurrentID=0, Status='live', Symbol='LOC105369907', Aliases='', description='uncharacterized LOC105369907', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93090231, end_position_on_the_genomic_accession=93096154, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369908, CurrentID=0, Status='live', Symbol='LOC105369908', Aliases='', description='uncharacterized LOC105369908', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93096200, end_position_on_the_genomic_accession=93109302, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369909, CurrentID=0, Status='live', Symbol='LOC105369909', Aliases='', description='uncharacterized LOC105369909', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643382, CurrentID=0, Status='live', Symbol='TMEM253', Aliases='C14orf176, C14orf95, NCRNA00220', description='transmembrane protein 253', other_designations='transmembrane protein C14orf176', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21098937, end_position_on_the_genomic_accession=21103724, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369911, CurrentID=0, Status='live', Symbol='LOC105369911', Aliases='', description='uncharacterized LOC105369911', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=93894965, end_position_on_the_genomic_accession=93943606, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643384, CurrentID=0, Status='live', Symbol='UNC93B4', Aliases='', description='unc-93 homolog B4 pseudogene (C. elegans)', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4145582, end_position_on_the_genomic_accession=4147795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369913, CurrentID=0, Status='live', Symbol='LOC105369913', Aliases='', description='uncharacterized LOC105369913', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94042820, end_position_on_the_genomic_accession=94048613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369914, CurrentID=0, Status='live', Symbol='LOC105369914', Aliases='', description='uncharacterized LOC105369914', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94648497, end_position_on_the_genomic_accession=94651091, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643387, CurrentID=0, Status='live', Symbol='LOC643387', Aliases='', description='TAR DNA binding protein pseudogene', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238231686, end_position_on_the_genomic_accession=238234345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105369916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369916, CurrentID=0, Status='live', Symbol='LOC105369916', Aliases='', description='uncharacterized LOC105369916', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95337570, end_position_on_the_genomic_accession=95342887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369917, CurrentID=0, Status='live', Symbol='LOC105369917', Aliases='', description='uncharacterized LOC105369917', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95358388, end_position_on_the_genomic_accession=95380762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369918, CurrentID=0, Status='live', Symbol='LOC105369918', Aliases='', description='uncharacterized LOC105369918', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95738160, end_position_on_the_genomic_accession=95748904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369919, CurrentID=0, Status='live', Symbol='LOC105369919', Aliases='', description='uncharacterized LOC105369919', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95717106, end_position_on_the_genomic_accession=95725015, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369920, CurrentID=0, Status='live', Symbol='LOC105369920', Aliases='', description='uncharacterized LOC105369920', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95803097, end_position_on_the_genomic_accession=95823314, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369921, CurrentID=0, Status='live', Symbol='LOC105369921', Aliases='', description='uncharacterized LOC105369921', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95791782, end_position_on_the_genomic_accession=95838668, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643394, CurrentID=0, Status='live', Symbol='SPINK9', Aliases='LEKTI2', description='serine peptidase inhibitor, Kazal type 9', other_designations='lymphoepithelial Kazal-type-related inhibitor 2', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148335559, end_position_on_the_genomic_accession=148339849, orientation='plus', exon_count=4, OMIM=613511, no_hdr0=''),
  105369923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369923, CurrentID=0, Status='live', Symbol='LOC105369923', Aliases='', description='uncharacterized LOC105369923', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96096307, end_position_on_the_genomic_accession=96108624, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369924, CurrentID=0, Status='live', Symbol='LOC105369924', Aliases='', description='uncharacterized LOC105369924', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96160921, end_position_on_the_genomic_accession=96171999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369925, CurrentID=0, Status='live', Symbol='LOC105369925', Aliases='', description='uncharacterized LOC105369925', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369926, CurrentID=0, Status='live', Symbol='LOC105369926', Aliases='', description='uncharacterized LOC105369926', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97019569, end_position_on_the_genomic_accession=97025288, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369927, CurrentID=0, Status='live', Symbol='LOC105369927', Aliases='', description='uncharacterized LOC105369927', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97193747, end_position_on_the_genomic_accession=97228433, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369928, CurrentID=0, Status='live', Symbol='LOC105369928', Aliases='', description='uncharacterized LOC105369928', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97185082, end_position_on_the_genomic_accession=97212723, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643401, CurrentID=0, Status='live', Symbol='LINC01021', Aliases='', description='long intergenic non-protein coding RNA 1021', other_designations='', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=27472292, end_position_on_the_genomic_accession=27496401, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100127050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127050, CurrentID=0, Status='live', Symbol='VN2R21P', Aliases='', description='vomeronasal 2 receptor 21 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11953189, end_position_on_the_genomic_accession=11953474, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369931, CurrentID=0, Status='live', Symbol='LOC105369931', Aliases='', description='uncharacterized LOC105369931', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369932, CurrentID=0, Status='live', Symbol='LOC105369932', Aliases='', description='uncharacterized LOC105369932', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98294473, end_position_on_the_genomic_accession=98304683, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369933, CurrentID=0, Status='live', Symbol='LOC105369933', Aliases='', description='uncharacterized LOC105369933', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98337758, end_position_on_the_genomic_accession=98350599, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643406, CurrentID=0, Status='live', Symbol='LOC643406', Aliases='', description='uncharacterized LOC643406', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5471196, end_position_on_the_genomic_accession=5477134, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928333, CurrentID=0, Status='live', Symbol='LOC101928333', Aliases='', description='uncharacterized LOC101928333', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127215127, end_position_on_the_genomic_accession=127229921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369936, CurrentID=0, Status='live', Symbol='LOC105369936', Aliases='', description='uncharacterized LOC105369936', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101659838, end_position_on_the_genomic_accession=101665062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369937, CurrentID=0, Status='live', Symbol='LOC105369937', Aliases='', description='uncharacterized LOC105369937', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369938, CurrentID=0, Status='live', Symbol='LOC105369938', Aliases='', description='uncharacterized LOC105369938', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101600928, end_position_on_the_genomic_accession=101618839, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369939, CurrentID=0, Status='live', Symbol='LOC105369939', Aliases='', description='uncharacterized LOC105369939', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101694565, end_position_on_the_genomic_accession=101695841, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643412, CurrentID=0, Status='live', Symbol='BTF3P16', Aliases='', description='basic transcription factor 3 pseudogene 16', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101456985, end_position_on_the_genomic_accession=101457846, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928334, CurrentID=0, Status='live', Symbol='LOC101928334', Aliases='', description='uncharacterized LOC101928334', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=101150515, end_position_on_the_genomic_accession=101176555, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  643414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643414, CurrentID=0, Status='live', Symbol='LIPK', Aliases='LIPL2, bA186O14.2', description='lipase, family member K', other_designations='lipase K|lipase-like abhydrolase domain-containing protein 2|lipase-like, ab-hydrolase domain containing 2', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88706234, end_position_on_the_genomic_accession=88753092, orientation='plus', exon_count=16, OMIM=613922, no_hdr0=''),
  105369943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369943, CurrentID=0, Status='live', Symbol='LOC105369943', Aliases='', description='uncharacterized LOC105369943', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=102326911, end_position_on_the_genomic_accession=102350918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369944, CurrentID=0, Status='live', Symbol='LOC105369944', Aliases='', description='uncharacterized LOC105369944', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369945, CurrentID=0, Status='live', Symbol='LOC105369945', Aliases='', description='uncharacterized LOC105369945', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103496045, end_position_on_the_genomic_accession=103497695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643418, CurrentID=0, Status='live', Symbol='LIPN', Aliases='ARCI8, LI4, LIPL4, bA186O14.3', description='lipase, family member N', other_designations='lipase-like abhydrolase domain-containing protein 4', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88757226, end_position_on_the_genomic_accession=88778891, orientation='plus', exon_count=12, OMIM=613924, no_hdr0=''),
  101928335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928335, CurrentID=0, Status='live', Symbol='LOC101928335', Aliases='', description='uncharacterized LOC101928335', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107894597, end_position_on_the_genomic_accession=107935980, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105369948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369948, CurrentID=0, Status='live', Symbol='LOC105369948', Aliases='', description='uncharacterized LOC105369948', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104169972, end_position_on_the_genomic_accession=104177652, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643421, CurrentID=0, Status='live', Symbol='TPT1P13', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 13', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=121863469, end_position_on_the_genomic_accession=121864282, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369950, CurrentID=0, Status='live', Symbol='LOC105369950', Aliases='', description='uncharacterized LOC105369950', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104864911, end_position_on_the_genomic_accession=104870566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369951, CurrentID=0, Status='live', Symbol='LOC105369951', Aliases='', description='uncharacterized LOC105369951', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104994812, end_position_on_the_genomic_accession=105007672, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369952, CurrentID=0, Status='live', Symbol='LOC105369952', Aliases='', description='uncharacterized LOC105369952', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104984206, end_position_on_the_genomic_accession=104986387, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421008, CurrentID=0, Status='live', Symbol='SLC9B1P3', Aliases='', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1 pseudogene 3', other_designations='Na+/H+ exchanger domain containing 1 pseudogene|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1 pseudogene 3', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38640782, end_position_on_the_genomic_accession=38689076, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369954, CurrentID=0, Status='live', Symbol='KCCAT198', Aliases='', description='renal clear cell carcinoma-associated transcript 198', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105303552, end_position_on_the_genomic_accession=105326985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643427, CurrentID=0, Status='live', Symbol='RBMS2P1', Aliases='RBMS2P', description='RNA binding motif, single stranded interacting protein 2 pseudogene 1', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=94423468, end_position_on_the_genomic_accession=94425050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369956, CurrentID=0, Status='live', Symbol='LOC105369956', Aliases='', description='uncharacterized LOC105369956', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105426181, end_position_on_the_genomic_accession=105444135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369957, CurrentID=0, Status='live', Symbol='LOC105369957', Aliases='', description='uncharacterized LOC105369957', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105372127, end_position_on_the_genomic_accession=105412881, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105369958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369958, CurrentID=0, Status='live', Symbol='LOC105369958', Aliases='', description='uncharacterized LOC105369958', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105509775, end_position_on_the_genomic_accession=105550037, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369959, CurrentID=0, Status='live', Symbol='LOC105369959', Aliases='', description='uncharacterized LOC105369959', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105705196, end_position_on_the_genomic_accession=105708027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643432, CurrentID=0, Status='live', Symbol='TSG1', Aliases='', description='tumor suppressor TSG1', other_designations='', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93707083, end_position_on_the_genomic_accession=93776481, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105369961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369961, CurrentID=0, Status='live', Symbol='LOC105369961', Aliases='', description='uncharacterized LOC105369961', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106862622, end_position_on_the_genomic_accession=106885505, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105369962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369962, CurrentID=0, Status='live', Symbol='LOC105369962', Aliases='', description='uncharacterized LOC105369962', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107441958, end_position_on_the_genomic_accession=107449844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369963, CurrentID=0, Status='live', Symbol='LOC105369963', Aliases='', description='uncharacterized LOC105369963', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107795423, end_position_on_the_genomic_accession=107799183, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369964, CurrentID=0, Status='live', Symbol='LOC105369964', Aliases='', description='uncharacterized LOC105369964', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107910420, end_position_on_the_genomic_accession=107916265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105369965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369965, CurrentID=0, Status='live', Symbol='LOC105369965', Aliases='', description='uncharacterized LOC105369965', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108178077, end_position_on_the_genomic_accession=108234367, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  643438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643438, CurrentID=0, Status='live', Symbol='LOC643438', Aliases='', description='huntingtin interacting protein K pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148637180, end_position_on_the_genomic_accession=148640523, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369967, CurrentID=0, Status='live', Symbol='LOC105369967', Aliases='', description='uncharacterized LOC105369967', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108479644, end_position_on_the_genomic_accession=108516908, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105369968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369968, CurrentID=0, Status='live', Symbol='LOC105369968', Aliases='', description='uncharacterized LOC105369968', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108628467, end_position_on_the_genomic_accession=108645249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643441, CurrentID=0, Status='live', Symbol='LOC643441', Aliases='', description='uncharacterized LOC643441', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113196781, end_position_on_the_genomic_accession=113206253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369970, CurrentID=0, Status='live', Symbol='LOC105369970', Aliases='', description='uncharacterized LOC105369970', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=45990062, end_position_on_the_genomic_accession=46004827, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369971, CurrentID=0, Status='live', Symbol='LOC105369971', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51847781, end_position_on_the_genomic_accession=51853515, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105369972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369972, CurrentID=0, Status='live', Symbol='LOC105369972', Aliases='', description='endogenous retrovirus group K member 8 Pol protein-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58323535, end_position_on_the_genomic_accession=58332034, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643445, CurrentID=0, Status='live', Symbol='TRIM43CP', Aliases='', description='tripartite motif containing 43C, pseudogene', other_designations='tripartite motif containing 43 pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97025826, end_position_on_the_genomic_accession=97031646, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643446, CurrentID=0, Status='live', Symbol='RNPS1P1', Aliases='', description='RNA binding protein S1, serine-rich domain pseudogene 1', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11368821, end_position_on_the_genomic_accession=11373749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643447, CurrentID=0, Status='live', Symbol='KRT8P49', Aliases='', description='keratin 8 pseudogene 49', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4722491, end_position_on_the_genomic_accession=4724241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105369976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369976, CurrentID=0, Status='live', Symbol='LOC105369976', Aliases='', description='uncharacterized LOC105369976', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110082724, end_position_on_the_genomic_accession=110102376, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369977, CurrentID=0, Status='live', Symbol='LOC105369977', Aliases='', description='uncharacterized LOC105369977', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110470291, end_position_on_the_genomic_accession=110489282, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369978, CurrentID=0, Status='live', Symbol='LOC105369978', Aliases='', description='uncharacterized LOC105369978', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105369979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369979, CurrentID=0, Status='live', Symbol='LOC105369979', Aliases='', description='uncharacterized LOC105369979', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110662672, end_position_on_the_genomic_accession=110670451, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105369980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369980, CurrentID=0, Status='live', Symbol='LOC105369980', Aliases='', description='uncharacterized LOC105369980', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110956296, end_position_on_the_genomic_accession=110957761, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369981, CurrentID=0, Status='live', Symbol='LOC105369981', Aliases='', description='uncharacterized LOC105369981', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111012243, end_position_on_the_genomic_accession=111026007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643454, CurrentID=0, Status='live', Symbol='LOC643454', Aliases='', description='adaptor related protein complex 3, sigma 1 subunit pseudogene', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214482219, end_position_on_the_genomic_accession=214483489, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  347541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347541, CurrentID=0, Status='live', Symbol='MAGEB5', Aliases='CT3.3, MAGE-B5', description='MAGE family member B5', other_designations='MAGE family testis and tumor-specific protein|MAGE-B5 antigen|cancer/testis antigen 3.3|cancer/testis antigen family 3, member 3|melanoma antigen family B, 5|melanoma antigen family B5', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26216081, end_position_on_the_genomic_accession=26218270, orientation='plus', exon_count=3, OMIM=300466, no_hdr0=''),
  105369984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369984, CurrentID=0, Status='live', Symbol='LOC105369984', Aliases='', description='uncharacterized LOC105369984', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111396830, end_position_on_the_genomic_accession=111403313, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130428, CurrentID=0, Status='live', Symbol='LOC100130428', Aliases='', description='IGYY565', other_designations='', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=124659136, end_position_on_the_genomic_accession=124659965, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105369986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369986, CurrentID=0, Status='live', Symbol='LOC105369986', Aliases='', description='uncharacterized LOC105369986', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=112557761, end_position_on_the_genomic_accession=112558427, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105369987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369987, CurrentID=0, Status='live', Symbol='IMMP1LP2', Aliases='', description='inner mitochondrial membrane peptidase subunit 1 pseudogene 2', other_designations='IMP1 inner mitochondrial membrane peptidase-like (S. cerevisiae) pseudogene 2', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113015607, end_position_on_the_genomic_accession=113016114, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105369988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369988, CurrentID=0, Status='live', Symbol='LOC105369988', Aliases='', description='uncharacterized LOC105369988', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113136507, end_position_on_the_genomic_accession=113142361, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372217, CurrentID=0, Status='live', Symbol='LOC105372217', Aliases='', description='uncharacterized LOC105372217', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76980478, end_position_on_the_genomic_accession=76984157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369990, CurrentID=0, Status='live', Symbol='LOC105369990', Aliases='', description='uncharacterized LOC105369990', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113639043, end_position_on_the_genomic_accession=113644515, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105369991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369991, CurrentID=0, Status='live', Symbol='LOC105369991', Aliases='', description='uncharacterized LOC105369991', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113653888, end_position_on_the_genomic_accession=113656000, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105369992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369992, CurrentID=0, Status='live', Symbol='LOC105369992', Aliases='', description='uncharacterized LOC105369992', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113838259, end_position_on_the_genomic_accession=113840054, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369993, CurrentID=0, Status='live', Symbol='LOC105369993', Aliases='', description='uncharacterized LOC105369993', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114011259, end_position_on_the_genomic_accession=114031052, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105369994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369994, CurrentID=0, Status='live', Symbol='LOC105369994', Aliases='', description='uncharacterized LOC105369994', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114073181, end_position_on_the_genomic_accession=114077037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105369995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369995, CurrentID=0, Status='live', Symbol='LOC105369995', Aliases='', description='uncharacterized LOC105369995', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114077133, end_position_on_the_genomic_accession=114195426, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  119180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119180, CurrentID=0, Status='live', Symbol='LYZL2', Aliases='', description='lysozyme like 2', other_designations='lysozyme 2|lysozyme-2|lysozyme-like 2', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30606232, end_position_on_the_genomic_accession=30629762, orientation='minus', exon_count=8, OMIM=612748, no_hdr0=''),
  105369997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105369997, CurrentID=0, Status='live', Symbol='LOC105369997', Aliases='', description='uncharacterized LOC105369997', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114405713, end_position_on_the_genomic_accession=114407019, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643470, CurrentID=0, Status='live', Symbol='SURF6P1', Aliases='', description='surfeit 6 pseudogene 1', other_designations='', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17184470, end_position_on_the_genomic_accession=17194769, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643471, CurrentID=0, Status='live', Symbol='KRT18P12', Aliases='', description='keratin 18 pseudogene 12', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214532145, end_position_on_the_genomic_accession=214533524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370000, CurrentID=0, Status='live', Symbol='LOC105370000', Aliases='', description='uncharacterized LOC105370000', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114713841, end_position_on_the_genomic_accession=114734111, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  645553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645553, CurrentID=0, Status='live', Symbol='LOC645553', Aliases='', description='uncharacterized LOC645553', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46189029, end_position_on_the_genomic_accession=46230374, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370002, CurrentID=0, Status='live', Symbol='LOC105370002', Aliases='', description='uncharacterized LOC105370002', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115597042, end_position_on_the_genomic_accession=115641122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370003, CurrentID=0, Status='live', Symbol='LOC105370003', Aliases='', description='uncharacterized LOC105370003', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115373836, end_position_on_the_genomic_accession=115763260, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370004, CurrentID=0, Status='live', Symbol='LOC105370004', Aliases='', description='uncharacterized LOC105370004', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115808380, end_position_on_the_genomic_accession=115877463, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370005, CurrentID=0, Status='live', Symbol='LOC105370005', Aliases='', description='uncharacterized LOC105370005', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116340202, end_position_on_the_genomic_accession=116350727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370006, CurrentID=0, Status='live', Symbol='LOC105370006', Aliases='', description='uncharacterized LOC105370006', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116359383, end_position_on_the_genomic_accession=116360570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643479, CurrentID=0, Status='live', Symbol='FAM25E', Aliases='', description='family with sequence similarity 25 member E', other_designations='family with sequence similarity 25, member E', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=46310876, end_position_on_the_genomic_accession=46312766, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370008, CurrentID=0, Status='live', Symbol='LOC105370008', Aliases='', description='uncharacterized LOC105370008', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116504967, end_position_on_the_genomic_accession=116508216, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370009, CurrentID=0, Status='live', Symbol='LOC105370009', Aliases='', description='uncharacterized LOC105370009', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=116537010, end_position_on_the_genomic_accession=116554381, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370010, CurrentID=0, Status='live', Symbol='LOC105370010', Aliases='', description='uncharacterized LOC105370010', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117105477, end_position_on_the_genomic_accession=117121031, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370011, CurrentID=0, Status='live', Symbol='LOC105370011', Aliases='', description='uncharacterized LOC105370011', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117832668, end_position_on_the_genomic_accession=117844760, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370012, CurrentID=0, Status='live', Symbol='LOC105370012', Aliases='', description='uncharacterized LOC105370012', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=117513553, end_position_on_the_genomic_accession=117517190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370013, CurrentID=0, Status='live', Symbol='LOC105370013', Aliases='', description='uncharacterized LOC105370013', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118644692, end_position_on_the_genomic_accession=118650401, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643486, CurrentID=0, Status='live', Symbol='BRDTP1', Aliases='', description='bromodomain, testis-specific pseudogene 1', other_designations='hCG1811337', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96337086, end_position_on_the_genomic_accession=96337902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370015, CurrentID=0, Status='live', Symbol='LOC105370015', Aliases='', description='uncharacterized LOC105370015', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118773056, end_position_on_the_genomic_accession=118774733, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370016, CurrentID=0, Status='live', Symbol='LOC105370016', Aliases='', description='uncharacterized LOC105370016', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118782956, end_position_on_the_genomic_accession=118833536, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370017, CurrentID=0, Status='live', Symbol='LOC105370017', Aliases='', description='uncharacterized LOC105370017', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118781040, end_position_on_the_genomic_accession=118834639, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370018, CurrentID=0, Status='live', Symbol='LOC105370018', Aliases='', description='uncharacterized LOC105370018', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118837489, end_position_on_the_genomic_accession=118867985, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370019, CurrentID=0, Status='live', Symbol='LOC105370019', Aliases='', description='uncharacterized LOC105370019', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118871455, end_position_on_the_genomic_accession=118920284, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370020, CurrentID=0, Status='live', Symbol='LOC105370020', Aliases='', description='uncharacterized LOC105370020', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118945684, end_position_on_the_genomic_accession=118956189, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370021, CurrentID=0, Status='live', Symbol='LOC105370021', Aliases='', description='uncharacterized LOC105370021', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119069344, end_position_on_the_genomic_accession=119124985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370022, CurrentID=0, Status='live', Symbol='LOC105370022', Aliases='', description='uncharacterized LOC105370022', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119030928, end_position_on_the_genomic_accession=119052216, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370023, CurrentID=0, Status='live', Symbol='LOC105370023', Aliases='', description='uncharacterized LOC105370023', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118990414, end_position_on_the_genomic_accession=119001750, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643496, CurrentID=0, Status='live', Symbol='LOC643496', Aliases='', description='hCG1660379', other_designations='', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=168149687, end_position_on_the_genomic_accession=168414843, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370025, CurrentID=0, Status='live', Symbol='LOC105370025', Aliases='', description='uncharacterized LOC105370025', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119224700, end_position_on_the_genomic_accession=119250464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370026, CurrentID=0, Status='live', Symbol='LOC105370026', Aliases='', description='uncharacterized LOC105370026', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119210938, end_position_on_the_genomic_accession=119221452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  780854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780854, CurrentID=0, Status='live', Symbol='SNORD3D', Aliases='U3-4', description='small nucleolar RNA, C/D box 3D', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19112420, end_position_on_the_genomic_accession=19112636, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370028, CurrentID=0, Status='live', Symbol='LOC105370028', Aliases='', description='uncharacterized LOC105370028', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120700449, end_position_on_the_genomic_accession=120704370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370029, CurrentID=0, Status='live', Symbol='LOC105370029', Aliases='', description='uncharacterized LOC105370029', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120676075, end_position_on_the_genomic_accession=120687026, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370030, CurrentID=0, Status='live', Symbol='LOC105370030', Aliases='', description='uncharacterized LOC105370030', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121109589, end_position_on_the_genomic_accession=121120157, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643503, CurrentID=0, Status='live', Symbol='TARDBPP1', Aliases='TARDBPL, bA318P23.1', description='TAR DNA binding protein pseudogene 1', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6199598, end_position_on_the_genomic_accession=6202266, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370032, CurrentID=0, Status='live', Symbol='LOC105370032', Aliases='', description='uncharacterized LOC105370032', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121162335, end_position_on_the_genomic_accession=121209934, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370033, CurrentID=0, Status='live', Symbol='LOC105370033', Aliases='', description='uncharacterized LOC105370033', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370034, CurrentID=0, Status='live', Symbol='LOC105370034', Aliases='', description='uncharacterized LOC105370034', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121663524, end_position_on_the_genomic_accession=121690417, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  643507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643507, CurrentID=0, Status='live', Symbol='RPL7P57', Aliases='RPL7_22_1777', description='ribosomal protein L7 pseudogene 57', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47840963, end_position_on_the_genomic_accession=47841710, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370036, CurrentID=0, Status='live', Symbol='LOC105370036', Aliases='', description='uncharacterized LOC105370036', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370037, CurrentID=0, Status='live', Symbol='LOC105370037', Aliases='', description='uncharacterized LOC105370037', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122656949, end_position_on_the_genomic_accession=122661957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370038, CurrentID=0, Status='live', Symbol='LOC105370038', Aliases='', description='uncharacterized LOC105370038', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122836424, end_position_on_the_genomic_accession=122843921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370039, CurrentID=0, Status='live', Symbol='LOC105370039', Aliases='', description='uncharacterized LOC105370039', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122864140, end_position_on_the_genomic_accession=122865607, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370040, CurrentID=0, Status='live', Symbol='LOC105370040', Aliases='', description='uncharacterized LOC105370040', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123252030, end_position_on_the_genomic_accession=123260353, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370041, CurrentID=0, Status='live', Symbol='LOC105370041', Aliases='', description='uncharacterized LOC105370041', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123613313, end_position_on_the_genomic_accession=123617750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370042, CurrentID=0, Status='live', Symbol='LOC105370042', Aliases='', description='uncharacterized LOC105370042', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123712693, end_position_on_the_genomic_accession=123724108, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370043, CurrentID=0, Status='live', Symbol='LOC105370043', Aliases='', description='uncharacterized LOC105370043', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123919432, end_position_on_the_genomic_accession=123921707, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370044, CurrentID=0, Status='live', Symbol='LOC105370044', Aliases='', description='uncharacterized LOC105370044', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123764131, end_position_on_the_genomic_accession=123790927, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370045, CurrentID=0, Status='live', Symbol='LOC105370045', Aliases='', description='filamin-interacting protein FAM101A', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124295091, end_position_on_the_genomic_accession=124301978, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370046, CurrentID=0, Status='live', Symbol='LOC105370046', Aliases='', description='uncharacterized LOC105370046', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124205986, end_position_on_the_genomic_accession=124208380, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370047, CurrentID=0, Status='live', Symbol='LOC105370047', Aliases='', description='uncharacterized LOC105370047', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124512509, end_position_on_the_genomic_accession=124516861, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370048, CurrentID=0, Status='live', Symbol='LOC105370048', Aliases='', description='uncharacterized LOC105370048', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124395392, end_position_on_the_genomic_accession=124398099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372219, CurrentID=0, Status='live', Symbol='LOC105372219', Aliases='', description='uncharacterized LOC105372219', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78518575, end_position_on_the_genomic_accession=78522246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370050, CurrentID=0, Status='live', Symbol='LOC105370050', Aliases='', description='uncharacterized LOC105370050', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124865519, end_position_on_the_genomic_accession=124877917, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370051, CurrentID=0, Status='live', Symbol='LOC105370051', Aliases='', description='uncharacterized LOC105370051', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124882087, end_position_on_the_genomic_accession=124900419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370052, CurrentID=0, Status='live', Symbol='LOC105370052', Aliases='', description='uncharacterized LOC105370052', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125079841, end_position_on_the_genomic_accession=125083301, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127173, CurrentID=0, Status='live', Symbol='HLA-Y', Aliases='HLA-DEL, HLA-GOR', description='major histocompatibility complex, class I, Y (pseudogene)', other_designations='HLA-BEL/COQ/DEL', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370054, CurrentID=0, Status='live', Symbol='LOC105370054', Aliases='', description='uncharacterized LOC105370054', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125272659, end_position_on_the_genomic_accession=125281295, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85409, CurrentID=0, Status='live', Symbol='NKD2', Aliases='Naked2', description='naked cuticle homolog 2 (Drosophila)', other_designations='Dvl-binding protein NKD2', map_location='5p15.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1008820, end_position_on_the_genomic_accession=1038812, orientation='plus', exon_count=11, OMIM=607852, no_hdr0=''),
  105370056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370056, CurrentID=0, Status='live', Symbol='LOC105370056', Aliases='', description='uncharacterized LOC105370056', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126015767, end_position_on_the_genomic_accession=126034009, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643529, CurrentID=0, Status='live', Symbol='LINC00865', Aliases='', description='long intergenic non-protein coding RNA 865', other_designations='hCG2024094', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89829493, end_position_on_the_genomic_accession=89840861, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370058, CurrentID=0, Status='live', Symbol='LOC105370058', Aliases='', description='uncharacterized LOC105370058', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=125994250, end_position_on_the_genomic_accession=126002019, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643531, CurrentID=0, Status='live', Symbol='RPL29P26', Aliases='RPL29_14_1268', description='ribosomal protein L29 pseudogene 26', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95467262, end_position_on_the_genomic_accession=95467876, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370060, CurrentID=0, Status='live', Symbol='LOC105370060', Aliases='', description='uncharacterized LOC105370060', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126688366, end_position_on_the_genomic_accession=126720796, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370061, CurrentID=0, Status='live', Symbol='LOC105370061', Aliases='', description='uncharacterized LOC105370061', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126874801, end_position_on_the_genomic_accession=126888692, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370062, CurrentID=0, Status='live', Symbol='LOC105370062', Aliases='', description='uncharacterized LOC105370062', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127150529, end_position_on_the_genomic_accession=127163343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643535, CurrentID=0, Status='live', Symbol='KCTD9P2', Aliases='KCTD9L', description='potassium channel tetramerization domain containing 9 pseudogene 2', other_designations='Putative BTB/POZ domain-containing protein KCTD9-like protein|potassium channel tetramerisation domain containing 9 pseudogene 2', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=105405513, end_position_on_the_genomic_accession=105408882, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643536, CurrentID=0, Status='live', Symbol='VDAC1P10', Aliases='', description='voltage-dependent anion channel 1 pseudogene 10', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=215376498, end_position_on_the_genomic_accession=215378238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370065, CurrentID=0, Status='live', Symbol='LOC105370065', Aliases='', description='uncharacterized LOC105370065', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127274265, end_position_on_the_genomic_accession=127284300, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370066, CurrentID=0, Status='live', Symbol='LOC105370066', Aliases='', description='uncharacterized LOC105370066', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127459954, end_position_on_the_genomic_accession=127470411, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370067, CurrentID=0, Status='live', Symbol='LOC105370067', Aliases='', description='uncharacterized LOC105370067', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127486315, end_position_on_the_genomic_accession=127507532, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370068, CurrentID=0, Status='live', Symbol='LOC105370068', Aliases='', description='uncharacterized LOC105370068', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127649227, end_position_on_the_genomic_accession=127651340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370069, CurrentID=0, Status='live', Symbol='LOC105370069', Aliases='', description='uncharacterized LOC105370069', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=127719111, end_position_on_the_genomic_accession=127721624, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370070, CurrentID=0, Status='live', Symbol='CRAT8', Aliases='', description='cervical cancer-associated transcript 8', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128022365, end_position_on_the_genomic_accession=128027172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370071, CurrentID=0, Status='live', Symbol='LOC105370071', Aliases='', description='uncharacterized LOC105370071', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128072257, end_position_on_the_genomic_accession=128074306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370072, CurrentID=0, Status='live', Symbol='LOC105370072', Aliases='', description='uncharacterized LOC105370072', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128825680, end_position_on_the_genomic_accession=128829350, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370073, CurrentID=0, Status='live', Symbol='LOC105370073', Aliases='', description='uncharacterized LOC105370073', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129021727, end_position_on_the_genomic_accession=129028884, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370074, CurrentID=0, Status='live', Symbol='LOC105370074', Aliases='', description='uncharacterized LOC105370074', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129460604, end_position_on_the_genomic_accession=129468602, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370075, CurrentID=0, Status='live', Symbol='LOC105370075', Aliases='', description='uncharacterized LOC105370075', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370076, CurrentID=0, Status='live', Symbol='LOC105370076', Aliases='', description='uncharacterized LOC105370076', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129957583, end_position_on_the_genomic_accession=129975730, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  643549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643549, CurrentID=0, Status='live', Symbol='LOC643549', Aliases='', description='uncharacterized LOC643549', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=44562739, end_position_on_the_genomic_accession=44563631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643550, CurrentID=0, Status='live', Symbol='RPL12P33', Aliases='RPL12_13_1292', description='ribosomal protein L12 pseudogene 33', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120916650, end_position_on_the_genomic_accession=120917281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370079, CurrentID=0, Status='live', Symbol='LOC105370079', Aliases='', description='uncharacterized LOC105370079', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=130635787, end_position_on_the_genomic_accession=130638233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370080, CurrentID=0, Status='live', Symbol='LOC105370080', Aliases='', description='uncharacterized LOC105370080', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370081, CurrentID=0, Status='live', Symbol='LOC105370081', Aliases='', description='uncharacterized LOC105370081', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131028284, end_position_on_the_genomic_accession=131031474, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643554, CurrentID=0, Status='live', Symbol='LOC643554', Aliases='', description='ornithine aminotransferase pseudogene', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48152993, end_position_on_the_genomic_accession=48155971, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370083, CurrentID=0, Status='live', Symbol='LOC105370083', Aliases='', description='uncharacterized LOC105370083', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131298928, end_position_on_the_genomic_accession=131300689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370084, CurrentID=0, Status='live', Symbol='LOC105370084', Aliases='', description='uncharacterized LOC105370084', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131583582, end_position_on_the_genomic_accession=131591794, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370085, CurrentID=0, Status='live', Symbol='LOC105370085', Aliases='', description='uncharacterized LOC105370085', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131641715, end_position_on_the_genomic_accession=131648121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127206, CurrentID=0, Status='live', Symbol='KIAA1024L', Aliases='', description='KIAA1024-like', other_designations='', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129748191, end_position_on_the_genomic_accession=129765063, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370087, CurrentID=0, Status='live', Symbol='LOC105370087', Aliases='', description='uncharacterized LOC105370087', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131677708, end_position_on_the_genomic_accession=131681053, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370088, CurrentID=0, Status='live', Symbol='LOC105370088', Aliases='', description='uncharacterized LOC105370088', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131817183, end_position_on_the_genomic_accession=131823034, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370089, CurrentID=0, Status='live', Symbol='LOC105370089', Aliases='', description='uncharacterized LOC105370089', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=131808939, end_position_on_the_genomic_accession=131817759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370090, CurrentID=0, Status='live', Symbol='LOC105370090', Aliases='', description='uncharacterized LOC105370090', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132159626, end_position_on_the_genomic_accession=132162970, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643563, CurrentID=0, Status='live', Symbol='EEF1A1P15', Aliases='EEF1AL13', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 15', other_designations='eukaryotic translation elongation factor 1 alpha-like 13', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=98389510, end_position_on_the_genomic_accession=98390939, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370092, CurrentID=0, Status='live', Symbol='LOC105370092', Aliases='', description='uncharacterized LOC105370092', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132427420, end_position_on_the_genomic_accession=132437437, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370093, CurrentID=0, Status='live', Symbol='LOC105370093', Aliases='', description='uncharacterized LOC105370093', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132608003, end_position_on_the_genomic_accession=132610421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370094, CurrentID=0, Status='live', Symbol='LOC105370094', Aliases='', description='uncharacterized LOC105370094', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132595537, end_position_on_the_genomic_accession=132601788, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370095, CurrentID=0, Status='live', Symbol='LOC105370095', Aliases='', description='uncharacterized LOC105370095', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132762160, end_position_on_the_genomic_accession=132768211, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370096, CurrentID=0, Status='live', Symbol='LOC105370096', Aliases='', description='uncharacterized LOC105370096', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133065540, end_position_on_the_genomic_accession=133078646, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370097, CurrentID=0, Status='live', Symbol='LOC105370097', Aliases='', description='uncharacterized LOC105370097', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133106695, end_position_on_the_genomic_accession=133110158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370098, CurrentID=0, Status='live', Symbol='LOC105370098', Aliases='', description='uncharacterized LOC105370098', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18942475, end_position_on_the_genomic_accession=18944377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370099, CurrentID=0, Status='live', Symbol='LOC105370099', Aliases='', description='uncharacterized LOC105370099', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370100, CurrentID=0, Status='live', Symbol='LOC105370100', Aliases='', description='uncharacterized LOC105370100', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19561714, end_position_on_the_genomic_accession=19562532, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370101, CurrentID=0, Status='live', Symbol='LOC105370101', Aliases='', description='uncharacterized LOC105370101', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20105670, end_position_on_the_genomic_accession=20119929, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370102, CurrentID=0, Status='live', Symbol='LOC105370102', Aliases='', description='uncharacterized LOC105370102', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20295706, end_position_on_the_genomic_accession=20326221, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370103, CurrentID=0, Status='live', Symbol='LOC105370103', Aliases='', description='uncharacterized LOC105370103', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20550981, end_position_on_the_genomic_accession=20566765, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  643576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643576, CurrentID=0, Status='live', Symbol='LOC643576', Aliases='', description='phosphoglycerate mutase 1 (brain) pseudogene', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104030741, end_position_on_the_genomic_accession=104032464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370105, CurrentID=0, Status='live', Symbol='LOC105370105', Aliases='', description='uncharacterized LOC105370105', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21280284, end_position_on_the_genomic_accession=21293503, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370106, CurrentID=0, Status='live', Symbol='LOC105370106', Aliases='', description='uncharacterized LOC105370106', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370107, CurrentID=0, Status='live', Symbol='LOC105370107', Aliases='', description='uncharacterized LOC105370107', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370108, CurrentID=0, Status='live', Symbol='LOC105370108', Aliases='', description='uncharacterized LOC105370108', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22061697, end_position_on_the_genomic_accession=22155392, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372221, CurrentID=0, Status='live', Symbol='LOC105372221', Aliases='', description='uncharacterized LOC105372221', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78639294, end_position_on_the_genomic_accession=78641237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422855, CurrentID=0, Status='live', Symbol='MIR4301', Aliases='', description='microRNA 4301', other_designations='hsa-mir-4301', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113450023, end_position_on_the_genomic_accession=113450088, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370111, CurrentID=0, Status='live', Symbol='LOC105370111', Aliases='', description='uncharacterized LOC105370111', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22975808, end_position_on_the_genomic_accession=22978398, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370112, CurrentID=0, Status='live', Symbol='LOC105370112', Aliases='', description='uncharacterized LOC105370112', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23108590, end_position_on_the_genomic_accession=23129899, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370113, CurrentID=0, Status='live', Symbol='LOC105370113', Aliases='', description='uncharacterized LOC105370113', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23715165, end_position_on_the_genomic_accession=23738199, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643586, CurrentID=0, Status='live', Symbol='PKMP1', Aliases='', description='pyruvate kinase, muscle pseudogene 1', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114535899, end_position_on_the_genomic_accession=114537459, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370115, CurrentID=0, Status='live', Symbol='LOC105370115', Aliases='', description='uncharacterized LOC105370115', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23954464, end_position_on_the_genomic_accession=23969256, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370116, CurrentID=0, Status='live', Symbol='LOC105370116', Aliases='', description='uncharacterized LOC105370116', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370117, CurrentID=0, Status='live', Symbol='LOC105370117', Aliases='', description='uncharacterized LOC105370117', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24502341, end_position_on_the_genomic_accession=24529087, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370118, CurrentID=0, Status='live', Symbol='LOC105370118', Aliases='', description='uncharacterized LOC105370118', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24723116, end_position_on_the_genomic_accession=24738648, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370119, CurrentID=0, Status='live', Symbol='LOC105370119', Aliases='', description='uncharacterized LOC105370119', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25040200, end_position_on_the_genomic_accession=25046959, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370120, CurrentID=0, Status='live', Symbol='LOC105370120', Aliases='', description='uncharacterized LOC105370120', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25095620, end_position_on_the_genomic_accession=25119432, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370121, CurrentID=0, Status='live', Symbol='LOC105370121', Aliases='', description='uncharacterized LOC105370121', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25227264, end_position_on_the_genomic_accession=25235663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370122, CurrentID=0, Status='live', Symbol='LOC105370122', Aliases='', description='uncharacterized LOC105370122', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25975584, end_position_on_the_genomic_accession=25977551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  653399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653399, CurrentID=0, Status='live', Symbol='GSTTP2', Aliases='', description='glutathione S-transferase theta pseudogene 2', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643596, CurrentID=0, Status='live', Symbol='RNF224', Aliases='', description='ring finger protein 224', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137227566, end_position_on_the_genomic_accession=137229640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370125, CurrentID=0, Status='live', Symbol='LOC105370125', Aliases='', description='uncharacterized LOC105370125', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27177692, end_position_on_the_genomic_accession=27220649, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105370126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370126, CurrentID=0, Status='live', Symbol='LOC105370126', Aliases='', description='uncharacterized LOC105370126', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27368510, end_position_on_the_genomic_accession=27375005, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370127, CurrentID=0, Status='live', Symbol='LOC105370127', Aliases='', description='uncharacterized LOC105370127', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27445438, end_position_on_the_genomic_accession=27449185, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370128, CurrentID=0, Status='live', Symbol='LOC105370128', Aliases='', description='uncharacterized LOC105370128', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27667713, end_position_on_the_genomic_accession=27680599, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370129, CurrentID=0, Status='live', Symbol='LOC105370129', Aliases='', description='uncharacterized LOC105370129', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27738704, end_position_on_the_genomic_accession=27740133, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370130, CurrentID=0, Status='live', Symbol='LOC105370130', Aliases='', description='uncharacterized LOC105370130', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27901834, end_position_on_the_genomic_accession=27903597, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370131, CurrentID=0, Status='live', Symbol='LOC105370131', Aliases='', description='uncharacterized LOC105370131', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27988599, end_position_on_the_genomic_accession=27989785, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370132, CurrentID=0, Status='live', Symbol='LOC105370132', Aliases='', description='uncharacterized LOC105370132', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27980155, end_position_on_the_genomic_accession=27988526, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370133, CurrentID=0, Status='live', Symbol='LOC105370133', Aliases='', description='uncharacterized LOC105370133', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28270873, end_position_on_the_genomic_accession=28295819, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370134, CurrentID=0, Status='live', Symbol='LOC105370134', Aliases='', description='uncharacterized LOC105370134', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28312105, end_position_on_the_genomic_accession=28329602, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100137049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100137049, CurrentID=0, Status='live', Symbol='PLA2G4B', Aliases='HsT16992, cPLA2-beta', description='phospholipase A2 group IVB', other_designations='phospholipase A2, group IVB (cytosolic)', map_location='15q11.2-q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41838813, end_position_on_the_genomic_accession=41848148, orientation='plus', exon_count=20, OMIM=606088, no_hdr0=''),
  100130433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130433, CurrentID=0, Status='live', Symbol='LOC100130433', Aliases='', description='uncharacterized LOC100130433', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=87990841, end_position_on_the_genomic_accession=87991566, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643609, CurrentID=0, Status='live', Symbol='NR1H5P', Aliases='Fxrb, NR1H5', description='nuclear receptor subfamily 1 group H member 5, pseudogene', other_designations='nuclear receptor subfamily 1, group H, member 4 pseudogene|nuclear receptor subfamily 1, group H, member 5, pseudogene', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114834456, end_position_on_the_genomic_accession=114851544, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370138, CurrentID=0, Status='live', Symbol='LOC105370138', Aliases='', description='uncharacterized LOC105370138', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28721552, end_position_on_the_genomic_accession=28727631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370139, CurrentID=0, Status='live', Symbol='LOC105370139', Aliases='', description='uncharacterized LOC105370139', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=29695213, end_position_on_the_genomic_accession=29697815, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370140, CurrentID=0, Status='live', Symbol='LOC105370140', Aliases='', description='uncharacterized LOC105370140', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30062477, end_position_on_the_genomic_accession=30072158, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370141, CurrentID=0, Status='live', Symbol='LINC00385', Aliases='', description='long intergenic non-protein coding RNA 385', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30152602, end_position_on_the_genomic_accession=30169342, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370142, CurrentID=0, Status='live', Symbol='LOC105370142', Aliases='', description='uncharacterized LOC105370142', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30392701, end_position_on_the_genomic_accession=30399426, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643615, CurrentID=0, Status='live', Symbol='UBTFL6', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 6 (pseudogene)', other_designations='putative upstream-binding factor 1-like protein 6|upstream binding transcription factor, RNA polymerase I-like 1, pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97636585, end_position_on_the_genomic_accession=97637765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643616, CurrentID=0, Status='live', Symbol='MOP-1', Aliases='', description='MOP-1', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=82608700, end_position_on_the_genomic_accession=82609498, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643617, CurrentID=0, Status='live', Symbol='RPSAP8', Aliases='LAMR1P8, RPSA_26_1797', description='ribosomal protein SA pseudogene 8', other_designations='laminin receptor 1 pseudogene 8', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100155813, end_position_on_the_genomic_accession=100156850, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370146, CurrentID=0, Status='live', Symbol='LOC105370146', Aliases='', description='uncharacterized LOC105370146', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31020958, end_position_on_the_genomic_accession=31023742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370147, CurrentID=0, Status='live', Symbol='LOC105370147', Aliases='', description='uncharacterized LOC105370147', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31043221, end_position_on_the_genomic_accession=31046693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370148, CurrentID=0, Status='live', Symbol='LOC105370148', Aliases='', description='uncharacterized LOC105370148', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31162625, end_position_on_the_genomic_accession=31171761, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370149, CurrentID=0, Status='live', Symbol='LOC105370149', Aliases='', description='uncharacterized LOC105370149', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31364332, end_position_on_the_genomic_accession=31376313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370150, CurrentID=0, Status='live', Symbol='LOC105370150', Aliases='', description='uncharacterized LOC105370150', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31419989, end_position_on_the_genomic_accession=31470616, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  643623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643623, CurrentID=0, Status='live', Symbol='LOC643623', Aliases='', description='uncharacterized LOC643623', other_designations='', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125674353, end_position_on_the_genomic_accession=125720218, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370152, CurrentID=0, Status='live', Symbol='LOC105370152', Aliases='', description='uncharacterized LOC105370152', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31796375, end_position_on_the_genomic_accession=31848859, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370153, CurrentID=0, Status='live', Symbol='LOC105370153', Aliases='', description='uncharacterized LOC105370153', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33335443, end_position_on_the_genomic_accession=33350068, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370873, CurrentID=0, Status='live', Symbol='LOC105370873', Aliases='', description='uncharacterized LOC105370873', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68483202, end_position_on_the_genomic_accession=68488241, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643627, CurrentID=0, Status='live', Symbol='RPS18P8', Aliases='RPS18_5_650', description='ribosomal protein S18 pseudogene 8', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4979029, end_position_on_the_genomic_accession=4979532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370156, CurrentID=0, Status='live', Symbol='LOC105370156', Aliases='', description='uncharacterized LOC105370156', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33983989, end_position_on_the_genomic_accession=34000657, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370157, CurrentID=0, Status='live', Symbol='LOC105370157', Aliases='', description='uncharacterized LOC105370157', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=33860281, end_position_on_the_genomic_accession=33865163, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370158, CurrentID=0, Status='live', Symbol='LOC105370158', Aliases='', description='uncharacterized LOC105370158', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34347943, end_position_on_the_genomic_accession=34616292, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370159, CurrentID=0, Status='live', Symbol='LOC105370159', Aliases='', description='uncharacterized LOC105370159', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=34671006, end_position_on_the_genomic_accession=34689431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370160, CurrentID=0, Status='live', Symbol='LOC105370160', Aliases='', description='uncharacterized LOC105370160', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35288201, end_position_on_the_genomic_accession=35292783, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370161, CurrentID=0, Status='live', Symbol='LOC105370161', Aliases='', description='uncharacterized LOC105370161', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35691354, end_position_on_the_genomic_accession=35694680, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370162, CurrentID=0, Status='live', Symbol='LOC105370162', Aliases='', description='uncharacterized LOC105370162', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36074540, end_position_on_the_genomic_accession=36083561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370163, CurrentID=0, Status='live', Symbol='LOC105370163', Aliases='', description='uncharacterized LOC105370163', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=35856836, end_position_on_the_genomic_accession=35904240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370164, CurrentID=0, Status='live', Symbol='LOC105370164', Aliases='', description='uncharacterized LOC105370164', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36134018, end_position_on_the_genomic_accession=36146074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643637, CurrentID=0, Status='live', Symbol='GLYATL1P4', Aliases='', description='glycine-N-acyltransferase like 1 pseudogene 4', other_designations='glycine-N-acyltransferase-like 1 pseudogene 4', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59041036, end_position_on_the_genomic_accession=59052019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370166, CurrentID=0, Status='live', Symbol='LOC105370166', Aliases='', description='uncharacterized LOC105370166', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38184832, end_position_on_the_genomic_accession=38206253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370167, CurrentID=0, Status='live', Symbol='LOC105370167', Aliases='', description='uncharacterized LOC105370167', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38567066, end_position_on_the_genomic_accession=38567643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370168, CurrentID=0, Status='live', Symbol='LOC105370168', Aliases='', description='uncharacterized LOC105370168', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39208771, end_position_on_the_genomic_accession=39227906, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643641, CurrentID=0, Status='live', Symbol='ZNF862', Aliases='', description='zinc finger protein 862', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149838420, end_position_on_the_genomic_accession=149867479, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105370170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370170, CurrentID=0, Status='live', Symbol='LOC105370170', Aliases='', description='uncharacterized LOC105370170', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39504030, end_position_on_the_genomic_accession=39517130, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370171, CurrentID=0, Status='live', Symbol='LOC105370171', Aliases='', description='uncharacterized LOC105370171', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40063569, end_position_on_the_genomic_accession=40064394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370172, CurrentID=0, Status='live', Symbol='LOC105370172', Aliases='', description='uncharacterized LOC105370172', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40081592, end_position_on_the_genomic_accession=40480889, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105370173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370173, CurrentID=0, Status='live', Symbol='LOC105370173', Aliases='', description='uncharacterized LOC105370173', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41298776, end_position_on_the_genomic_accession=41303914, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643646, CurrentID=0, Status='live', Symbol='HSD17B1P1', Aliases='EDH17B1, EDHB17, HSD17, HSD17BP1, SDR28C1P1', description='hydroxysteroid (17-beta) dehydrogenase 1 pseudogene 1', other_designations='hydroxysteroid (17-beta) dehydrogenase pseudogene 1', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42545779, end_position_on_the_genomic_accession=42548975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370175, CurrentID=0, Status='live', Symbol='LOC105370175', Aliases='', description='uncharacterized LOC105370175', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41546886, end_position_on_the_genomic_accession=41552202, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643648, CurrentID=0, Status='live', Symbol='LINC01189', Aliases='', description='long intergenic non-protein coding RNA 1189', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62452490, end_position_on_the_genomic_accession=62522018, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370177, CurrentID=0, Status='live', Symbol='LOC105370177', Aliases='', description='uncharacterized LOC105370177', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42435205, end_position_on_the_genomic_accession=42485956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643650, CurrentID=0, Status='live', Symbol='LINC00842', Aliases='', description='long intergenic non-protein coding RNA 842', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46398362, end_position_on_the_genomic_accession=46453306, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370179, CurrentID=0, Status='live', Symbol='LOC105370179', Aliases='', description='uncharacterized LOC105370179', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43002936, end_position_on_the_genomic_accession=43007338, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370180, CurrentID=0, Status='live', Symbol='LOC105370180', Aliases='', description='uncharacterized LOC105370180', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43072861, end_position_on_the_genomic_accession=43077977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643653, CurrentID=0, Status='live', Symbol='RPL35P8', Aliases='RPL35_6_1759', description='ribosomal protein L35 pseudogene 8', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49213822, end_position_on_the_genomic_accession=49214275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370182, CurrentID=0, Status='live', Symbol='LOC105370182', Aliases='', description='uncharacterized LOC105370182', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=43984015, end_position_on_the_genomic_accession=44014959, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370183, CurrentID=0, Status='live', Symbol='LINC00390', Aliases='', description='long intergenic non-protein coding RNA 390', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44105446, end_position_on_the_genomic_accession=44147930, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370184, CurrentID=0, Status='live', Symbol='LOC105370184', Aliases='', description='uncharacterized LOC105370184', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44425071, end_position_on_the_genomic_accession=44428413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  119369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119369, CurrentID=0, Status='live', Symbol='NUDT9P1', Aliases='C10orf98, bA56M3.1', description='nudix hydrolase 9 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 9 pseudogene 1', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91151995, end_position_on_the_genomic_accession=91153083, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370186, CurrentID=0, Status='live', Symbol='LOC105370186', Aliases='', description='uncharacterized LOC105370186', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44651708, end_position_on_the_genomic_accession=44668409, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421047, CurrentID=0, Status='live', Symbol='RBMX2P3', Aliases='', description='RNA binding motif protein, X-linked 2 pseudogene 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119084801, end_position_on_the_genomic_accession=119086014, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370188, CurrentID=0, Status='live', Symbol='LOC105370188', Aliases='', description='uncharacterized LOC105370188', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45046178, end_position_on_the_genomic_accession=45048014, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391266, CurrentID=0, Status='live', Symbol='CNN2P7', Aliases='', description='calponin 2 pseudogene 7', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13826651, end_position_on_the_genomic_accession=13827824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370190, CurrentID=0, Status='live', Symbol='LOC105370190', Aliases='', description='uncharacterized LOC105370190', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45224489, end_position_on_the_genomic_accession=45243393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370191, CurrentID=0, Status='live', Symbol='LOC105370191', Aliases='', description='uncharacterized LOC105370191', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45930404, end_position_on_the_genomic_accession=45940838, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643664, CurrentID=0, Status='live', Symbol='SLC35G6', Aliases='AMAC1L3, TMEM21B', description='solute carrier family 35 member G6', other_designations='acyl-malonyl condensing enzyme 1-like 3|acyl-malonyl-condensing enzyme 1-like protein 3|protein AMAC1L3|solute carrier family 35, member G6|transmembrane protein 21B', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7481402, end_position_on_the_genomic_accession=7483064, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370193, CurrentID=0, Status='live', Symbol='LOC105370193', Aliases='', description='uncharacterized LOC105370193', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46278463, end_position_on_the_genomic_accession=46279841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370194, CurrentID=0, Status='live', Symbol='LOC105370194', Aliases='', description='uncharacterized LOC105370194', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46276632, end_position_on_the_genomic_accession=46321996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480812, CurrentID=0, Status='live', Symbol='PIGQP1', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class Q pseudogene 1', other_designations='phosphatidylinositol glycan anchor biosynthesis, class Q pseudogene 1', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643668, CurrentID=0, Status='live', Symbol='PSMC1P9', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 9', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=18692751, end_position_on_the_genomic_accession=18694295, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643669, CurrentID=0, Status='live', Symbol='CCER2', Aliases='', description='coiled-coil glutamate rich protein 2', other_designations='coiled-coil glutamate-rich protein 2', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38908980, end_position_on_the_genomic_accession=38914370, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370198, CurrentID=0, Status='live', Symbol='LOC105370198', Aliases='', description='uncharacterized LOC105370198', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48108123, end_position_on_the_genomic_accession=48166377, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370199, CurrentID=0, Status='live', Symbol='LOC105370199', Aliases='', description='uncharacterized LOC105370199', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370200, CurrentID=0, Status='live', Symbol='LOC105370200', Aliases='', description='uncharacterized LOC105370200', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48288915, end_position_on_the_genomic_accession=48296177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  119385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119385, CurrentID=0, Status='live', Symbol='AGAP11', Aliases='', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 11', other_designations='AGAP-11|ankyrin repeat and GTPase domain Arf GTPase activating protein 11', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86970741, end_position_on_the_genomic_accession=87010203, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105370202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370202, CurrentID=0, Status='live', Symbol='LOC105370202', Aliases='', description='uncharacterized LOC105370202', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48581179, end_position_on_the_genomic_accession=48599508, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643675, CurrentID=0, Status='live', Symbol='LOC643675', Aliases='', description='hCG2045751', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104907383, end_position_on_the_genomic_accession=104986701, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370204, CurrentID=0, Status='live', Symbol='LOC105370204', Aliases='', description='uncharacterized LOC105370204', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49616232, end_position_on_the_genomic_accession=49620615, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643677, CurrentID=0, Status='live', Symbol='CCDC168', Aliases='C13orf40', description='coiled-coil domain containing 168', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102729367, end_position_on_the_genomic_accession=102759072, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104355291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355291, CurrentID=0, Status='live', Symbol='LINC01392', Aliases='', description='long intergenic non-protein coding RNA 1392', other_designations='', map_location='7q31.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115123599, end_position_on_the_genomic_accession=115231355, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  119391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119391, CurrentID=0, Status='live', Symbol='GSTO2', Aliases='GSTO 2-2, bA127L20.1', description='glutathione S-transferase omega 2', other_designations='GSTO-2|MMA(V) reductase|bA127L20.1 (novel glutathione-S-transferase)|glutathione S-transferase omega 2-2|glutathione-S-transferase-like protein|glutathione-dependent dehydroascorbate reductase|monomethylarsonic acid reductase', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104268416, end_position_on_the_genomic_accession=104304948, orientation='plus', exon_count=10, OMIM=612314, no_hdr0=''),
  119392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119392, CurrentID=0, Status='live', Symbol='SFR1', Aliases='C10orf78, MEI5, MEIR5, bA373N18.1', description='SWI5-dependent homologous recombination repair protein 1', other_designations='MEI5 recombination repair protein homolog|SWI5-dependent recombination repair 1|SWI5-dependent recombination repair protein 1|meiosis protein 5 homolog', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=104122058, end_position_on_the_genomic_accession=104126385, orientation='plus', exon_count=5, OMIM=616527, no_hdr0=''),
  105370209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370209, CurrentID=0, Status='live', Symbol='LOC105370209', Aliases='', description='uncharacterized LOC105370209', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52759779, end_position_on_the_genomic_accession=52768918, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370210, CurrentID=0, Status='live', Symbol='LOC105370210', Aliases='', description='uncharacterized LOC105370210', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53555960, end_position_on_the_genomic_accession=53564303, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  119395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119395, CurrentID=0, Status='live', Symbol='CALHM3', Aliases='FAM26A, bA225H22.7', description='calcium homeostasis modulator 3', other_designations='family with sequence similarity 26, member A', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103472804, end_position_on_the_genomic_accession=103479240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370212, CurrentID=0, Status='live', Symbol='LOC105370212', Aliases='', description='uncharacterized LOC105370212', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54135880, end_position_on_the_genomic_accession=54145515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370213, CurrentID=0, Status='live', Symbol='LOC105370213', Aliases='', description='uncharacterized LOC105370213', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=54800172, end_position_on_the_genomic_accession=54841989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370214, CurrentID=0, Status='live', Symbol='LOC105370214', Aliases='', description='uncharacterized LOC105370214', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=56258461, end_position_on_the_genomic_accession=56735485, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370215, CurrentID=0, Status='live', Symbol='LOC105370215', Aliases='', description='uncharacterized LOC105370215', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57137091, end_position_on_the_genomic_accession=57137744, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370216, CurrentID=0, Status='live', Symbol='LOC105370216', Aliases='', description='uncharacterized LOC105370216', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57213047, end_position_on_the_genomic_accession=57216730, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643689, CurrentID=0, Status='live', Symbol='HNRNPA3P3', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 3', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140032304, end_position_on_the_genomic_accession=140033870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370218, CurrentID=0, Status='live', Symbol='LOC105370218', Aliases='', description='uncharacterized LOC105370218', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58627794, end_position_on_the_genomic_accession=58643520, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370219, CurrentID=0, Status='live', Symbol='LOC105370219', Aliases='', description='uncharacterized LOC105370219', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58653121, end_position_on_the_genomic_accession=58684724, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370220, CurrentID=0, Status='live', Symbol='LOC105370220', Aliases='', description='uncharacterized LOC105370220', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58715266, end_position_on_the_genomic_accession=58764288, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370221, CurrentID=0, Status='live', Symbol='LOC105370221', Aliases='', description='uncharacterized LOC105370221', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59086003, end_position_on_the_genomic_accession=59100873, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370222, CurrentID=0, Status='live', Symbol='LOC105370222', Aliases='', description='uncharacterized LOC105370222', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59280589, end_position_on_the_genomic_accession=59290797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370223, CurrentID=0, Status='live', Symbol='LOC105370223', Aliases='', description='uncharacterized LOC105370223', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59265905, end_position_on_the_genomic_accession=59280406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370224, CurrentID=0, Status='live', Symbol='LOC105370224', Aliases='', description='uncharacterized LOC105370224', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59119043, end_position_on_the_genomic_accession=59292524, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106480813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480813, CurrentID=0, Status='live', Symbol='RNU7-93P', Aliases='', description='RNA, U7 small nuclear 93 pseudogene', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370226, CurrentID=0, Status='live', Symbol='LOC105370226', Aliases='', description='uncharacterized LOC105370226', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=59659633, end_position_on_the_genomic_accession=59685141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643699, CurrentID=0, Status='live', Symbol='GOLGA8N', Aliases='GOLGA8R', description='golgin A8 family member N', other_designations='GOLGA8R', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32587328, end_position_on_the_genomic_accession=32607310, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105370228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370228, CurrentID=0, Status='live', Symbol='LOC105370228', Aliases='', description='uncharacterized LOC105370228', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60328478, end_position_on_the_genomic_accession=60381383, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370229, CurrentID=0, Status='live', Symbol='LOC105370229', Aliases='', description='uncharacterized LOC105370229', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60427788, end_position_on_the_genomic_accession=60432411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370230, CurrentID=0, Status='live', Symbol='LOC105370230', Aliases='', description='uncharacterized LOC105370230', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60616203, end_position_on_the_genomic_accession=60618893, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370231, CurrentID=0, Status='live', Symbol='LOC105370231', Aliases='', description='uncharacterized LOC105370231', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61563603, end_position_on_the_genomic_accession=61573889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370232, CurrentID=0, Status='live', Symbol='LOC105370232', Aliases='', description='uncharacterized LOC105370232', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62539677, end_position_on_the_genomic_accession=62574881, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370233, CurrentID=0, Status='live', Symbol='LOC105370233', Aliases='', description='uncharacterized LOC105370233', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62839134, end_position_on_the_genomic_accession=62897413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370234, CurrentID=0, Status='live', Symbol='LOC105370234', Aliases='', description='uncharacterized LOC105370234', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=62897473, end_position_on_the_genomic_accession=62911545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643707, CurrentID=0, Status='live', Symbol='GOLGA6L4', Aliases='', description='golgin A6 family-like 4', other_designations='golgi autoantigen, golgin subfamily a, 6-like 4', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84235773, end_position_on_the_genomic_accession=84245368, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105370236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370236, CurrentID=0, Status='live', Symbol='LOC105370236', Aliases='', description='uncharacterized LOC105370236', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63593893, end_position_on_the_genomic_accession=63673604, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370237, CurrentID=0, Status='live', Symbol='LOC105370237', Aliases='', description='uncharacterized LOC105370237', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63585463, end_position_on_the_genomic_accession=63586255, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370238, CurrentID=0, Status='live', Symbol='LOC105370238', Aliases='', description='uncharacterized LOC105370238', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=64436661, end_position_on_the_genomic_accession=64450363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643711, CurrentID=0, Status='live', Symbol='LOC643711', Aliases='', description='platelet activating factor acetylhydrolase 1b catalytic subunit 2 pseudogene', other_designations='platelet-activating factor acetylhydrolase 1b, catalytic subunit 2 (30kDa) pseudogene', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=97713412, end_position_on_the_genomic_accession=97756517, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105370240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370240, CurrentID=0, Status='live', Symbol='LOC105370240', Aliases='', description='uncharacterized LOC105370240', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=64738015, end_position_on_the_genomic_accession=64797734, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  347584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347584, CurrentID=0, Status='live', Symbol='TSPY11P', Aliases='TSPYP1', description='testis specific protein, Y-linked 11, pseudogene', other_designations='testis specific protein, Y-linked 1 pseudogene|testis specific protein, Y-linked pseudogene 1', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6266547, end_position_on_the_genomic_accession=6269275, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  643714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643714, CurrentID=0, Status='live', Symbol='CASC16', Aliases='LINC00918', description='cancer susceptibility candidate 16 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 918', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52552087, end_position_on_the_genomic_accession=52606975, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370243, CurrentID=0, Status='live', Symbol='LOC105370243', Aliases='', description='uncharacterized LOC105370243', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65902817, end_position_on_the_genomic_accession=65914949, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370244, CurrentID=0, Status='live', Symbol='LOC105370244', Aliases='', description='uncharacterized LOC105370244', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65929614, end_position_on_the_genomic_accession=65937161, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370245, CurrentID=0, Status='live', Symbol='LOC105370245', Aliases='', description='uncharacterized LOC105370245', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65995865, end_position_on_the_genomic_accession=66090248, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370246, CurrentID=0, Status='live', Symbol='LOC105370246', Aliases='', description='uncharacterized LOC105370246', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67305411, end_position_on_the_genomic_accession=67326132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643719, CurrentID=0, Status='live', Symbol='SCGB1B2P', Aliases='SCGB4A1P', description='secretoglobin, family 1B, member 2, pseudogene', other_designations='CTD-2378H7.2', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34576733, end_position_on_the_genomic_accession=34577691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370248, CurrentID=0, Status='live', Symbol='LOC105370248', Aliases='', description='uncharacterized LOC105370248', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67577620, end_position_on_the_genomic_accession=67596213, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643721, CurrentID=0, Status='live', Symbol='FLJ45508', Aliases='', description='uncharacterized LOC643721', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=105378613, end_position_on_the_genomic_accession=105380579, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370250, CurrentID=0, Status='live', Symbol='LOC105370250', Aliases='', description='uncharacterized LOC105370250', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67649902, end_position_on_the_genomic_accession=67660310, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643723, CurrentID=0, Status='live', Symbol='LYPLAL1-AS1', Aliases='', description='LYPLAL1 antisense RNA 1 (head to head)', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219080973, end_position_on_the_genomic_accession=219173788, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370252, CurrentID=0, Status='live', Symbol='LOC105370252', Aliases='', description='uncharacterized LOC105370252', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67769402, end_position_on_the_genomic_accession=67791191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  119437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119437, CurrentID=0, Status='live', Symbol='CTAGE7P', Aliases='CTAGE7, CTAGEP, bA500G10.2, rcCTAGE5', description='CTAGE family member 7, pseudogene', other_designations='CTAGE family pseudogene|CTAGE family, member 5 pseudogene|CTAGE family, member 7, pseudogene|meningioma expressed antigen 6 (coiled-coil proline-rich) pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130106009, end_position_on_the_genomic_accession=130108838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370254, CurrentID=0, Status='live', Symbol='LOC105370254', Aliases='', description='uncharacterized LOC105370254', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=68871070, end_position_on_the_genomic_accession=68874765, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106480814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480814, CurrentID=0, Status='live', Symbol='RNU7-99P', Aliases='', description='RNA, U7 small nuclear 99 pseudogene', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370256, CurrentID=0, Status='live', Symbol='LOC105370256', Aliases='', description='uncharacterized LOC105370256', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73592401, end_position_on_the_genomic_accession=73634420, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105370257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370257, CurrentID=0, Status='live', Symbol='LOC105370257', Aliases='', description='uncharacterized LOC105370257', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74158414, end_position_on_the_genomic_accession=74161956, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370258, CurrentID=0, Status='live', Symbol='LOC105370258', Aliases='', description='uncharacterized LOC105370258', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73879780, end_position_on_the_genomic_accession=73886129, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  347587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347587, CurrentID=0, Status='live', Symbol='MRPL57P10', Aliases='MRP63P10', description='mitochondrial ribosomal protein L57 pseudogene 10', other_designations='mitochondrial ribosomal protein 63 pseudogene 10', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10156183, end_position_on_the_genomic_accession=10156444, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370260, CurrentID=0, Status='live', Symbol='LOC105370260', Aliases='', description='uncharacterized LOC105370260', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=74649389, end_position_on_the_genomic_accession=74732799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643733, CurrentID=0, Status='live', Symbol='LOC643733', Aliases='', description='caspase 4, apoptosis-related cysteine peptidase pseudogene', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104901549, end_position_on_the_genomic_accession=104918175, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370262, CurrentID=0, Status='live', Symbol='LOC105370262', Aliases='', description='uncharacterized LOC105370262', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76229514, end_position_on_the_genomic_accession=76270904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106144537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106144537, CurrentID=0, Status='live', Symbol='LINC01173', Aliases='', description='long intergenic non-protein coding RNA 1173', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370264, CurrentID=0, Status='live', Symbol='LOC105370264', Aliases='', description='uncharacterized LOC105370264', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76552095, end_position_on_the_genomic_accession=76554739, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928388, CurrentID=0, Status='live', Symbol='PLA2G4E-AS1', Aliases='', description='PLA2G4E antisense RNA 1', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41972763, end_position_on_the_genomic_accession=41999094, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370266, CurrentID=0, Status='live', Symbol='LOC105370266', Aliases='', description='uncharacterized LOC105370266', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76664125, end_position_on_the_genomic_accession=76692312, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  391279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391279, CurrentID=0, Status='live', Symbol='OR7E23P', Aliases='OR21-1, OR21-3, OR7E92P', description='olfactory receptor family 7 subfamily E member 23 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 23 pseudogene|olfactory receptor, family 7, subfamily E, member 92 pseudogene', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32621062, end_position_on_the_genomic_accession=32622044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370268, CurrentID=0, Status='live', Symbol='LOC105370268', Aliases='', description='uncharacterized LOC105370268', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76833797, end_position_on_the_genomic_accession=76844819, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370269, CurrentID=0, Status='live', Symbol='LOC105370269', Aliases='', description='uncharacterized LOC105370269', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76932157, end_position_on_the_genomic_accession=76942960, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370270, CurrentID=0, Status='live', Symbol='LOC105370270', Aliases='', description='uncharacterized LOC105370270', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77280149, end_position_on_the_genomic_accession=77296165, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370271, CurrentID=0, Status='live', Symbol='LOC105370271', Aliases='', description='uncharacterized LOC105370271', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77462508, end_position_on_the_genomic_accession=77465855, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643744, CurrentID=0, Status='live', Symbol='HMGN2P4', Aliases='HMGN2L4', description='high mobility group nucleosomal binding domain 2 pseudogene 4', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 4|high-mobility group nucleosomal binding domain 2-like 4', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76241201, end_position_on_the_genomic_accession=76242405, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370273, CurrentID=0, Status='live', Symbol='LOC105370273', Aliases='', description='uncharacterized LOC105370273', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78963652, end_position_on_the_genomic_accession=78983499, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370274, CurrentID=0, Status='live', Symbol='LOC105370274', Aliases='', description='uncharacterized LOC105370274', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=79422603, end_position_on_the_genomic_accession=79429381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370275, CurrentID=0, Status='live', Symbol='LOC105370275', Aliases='', description='uncharacterized LOC105370275', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80121587, end_position_on_the_genomic_accession=80166394, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370276, CurrentID=0, Status='live', Symbol='LOC105370276', Aliases='', description='uncharacterized LOC105370276', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80356932, end_position_on_the_genomic_accession=80363131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643749, CurrentID=0, Status='live', Symbol='TRAF3IP2-AS1', Aliases='C6UAS, C6orf3, NCRNA00248, TRAF3IP2-AS2', description='TRAF3IP2 antisense RNA 1', other_designations='TRAF3IP2 antisense RNA 1 (non-protein coding)|TRAF3IP2 antisense RNA 2 (non-protein coding)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111483472, end_position_on_the_genomic_accession=111602295, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370278, CurrentID=0, Status='live', Symbol='LOC105370278', Aliases='', description='uncharacterized LOC105370278', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80695851, end_position_on_the_genomic_accession=80706675, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643751, CurrentID=0, Status='live', Symbol='CDC42P6', Aliases='', description='cell division cycle 42 pseudogene 6', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22726422, end_position_on_the_genomic_accession=22728023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643752, CurrentID=0, Status='live', Symbol='RAP1BL', Aliases='', description='RAP1B, member of RAS oncogene family pseudogene', other_designations='Rap1B-retro|Ras-related protein Rap-1b-like protein|hCG1757335', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76170085, end_position_on_the_genomic_accession=76174354, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370281, CurrentID=0, Status='live', Symbol='LOC105370281', Aliases='', description='uncharacterized LOC105370281', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81298609, end_position_on_the_genomic_accession=81302645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370282, CurrentID=0, Status='live', Symbol='LOC105370282', Aliases='', description='uncharacterized LOC105370282', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=81316401, end_position_on_the_genomic_accession=81339808, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  119467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119467, CurrentID=0, Status='live', Symbol='CLRN3', Aliases='TMEM12, USH3AL1', description='clarin 3', other_designations='transmembrane protein 12|usher syndrome type-3A-like protein 1', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127877841, end_position_on_the_genomic_accession=127892947, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377088, CurrentID=0, Status='live', Symbol='LOC105377088', Aliases='', description='uncharacterized LOC105377088', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52031121, end_position_on_the_genomic_accession=52035987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391282, CurrentID=0, Status='live', Symbol='RPL23AP12', Aliases='RPL23A_41_1732', description='ribosomal protein L23a pseudogene 12', other_designations='', map_location='21q22.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39127555, end_position_on_the_genomic_accession=39128076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370286, CurrentID=0, Status='live', Symbol='LOC105370286', Aliases='', description='uncharacterized LOC105370286', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=83802388, end_position_on_the_genomic_accession=83815806, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370287, CurrentID=0, Status='live', Symbol='LOC105370287', Aliases='', description='uncharacterized LOC105370287', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84064534, end_position_on_the_genomic_accession=84085137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370288, CurrentID=0, Status='live', Symbol='LOC105370288', Aliases='', description='putative uncharacterized protein C1orf99', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=83906573, end_position_on_the_genomic_accession=84068784, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370289, CurrentID=0, Status='live', Symbol='LOC105370289', Aliases='', description='uncharacterized LOC105370289', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84411852, end_position_on_the_genomic_accession=84571017, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372227, CurrentID=0, Status='live', Symbol='LOC105372227', Aliases='', description='uncharacterized LOC105372227', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79253577, end_position_on_the_genomic_accession=79255641, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643763, CurrentID=0, Status='live', Symbol='UG0898H09', Aliases='', description='uncharacterized LOC643763', other_designations='', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=62977861, end_position_on_the_genomic_accession=62984901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370292, CurrentID=0, Status='live', Symbol='LOC105370292', Aliases='', description='uncharacterized LOC105370292', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370293, CurrentID=0, Status='live', Symbol='LOC105370293', Aliases='', description='ribosome-binding protein 1-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643766, CurrentID=0, Status='live', Symbol='PSMC1P6', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 6', other_designations='', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87630078, end_position_on_the_genomic_accession=87631665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370295, CurrentID=0, Status='live', Symbol='LOC105370295', Aliases='', description='uncharacterized LOC105370295', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24538602, end_position_on_the_genomic_accession=24541913, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370296, CurrentID=0, Status='live', Symbol='LOC105370296', Aliases='', description='protein SPT2 homolog', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39601324, end_position_on_the_genomic_accession=39640412, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370297, CurrentID=0, Status='live', Symbol='LOC105370297', Aliases='', description='serine/arginine-rich splicing factor SR45-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44370118, end_position_on_the_genomic_accession=44374008, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643770, CurrentID=0, Status='live', Symbol='LOC643770', Aliases='', description='uncharacterized LOC643770', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98485544, end_position_on_the_genomic_accession=98503855, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370299, CurrentID=0, Status='live', Symbol='LOC105370299', Aliases='', description='uncharacterized LOC105370299', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=86402910, end_position_on_the_genomic_accession=86414564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370300, CurrentID=0, Status='live', Symbol='LOC105370300', Aliases='', description='uncharacterized LOC105370300', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=86792444, end_position_on_the_genomic_accession=86883286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643773, CurrentID=0, Status='live', Symbol='PNPLA10P', Aliases='', description='patatin like phospholipase domain containing 10 pseudogene', other_designations='patatin-like phospholipase domain containing 10 pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=127390561, end_position_on_the_genomic_accession=127392660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370302, CurrentID=0, Status='live', Symbol='LOC105370302', Aliases='', description='uncharacterized LOC105370302', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87334029, end_position_on_the_genomic_accession=87446308, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370303, CurrentID=0, Status='live', Symbol='LOC105370303', Aliases='', description='uncharacterized LOC105370303', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87679955, end_position_on_the_genomic_accession=87692607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370304, CurrentID=0, Status='live', Symbol='LOC105370304', Aliases='', description='uncharacterized LOC105370304', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=87853960, end_position_on_the_genomic_accession=87873712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370305, CurrentID=0, Status='live', Symbol='LOC105370305', Aliases='', description='uncharacterized LOC105370305', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88119798, end_position_on_the_genomic_accession=88138459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370306, CurrentID=0, Status='live', Symbol='LOC105370306', Aliases='', description='uncharacterized LOC105370306', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=88212367, end_position_on_the_genomic_accession=88236076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643779, CurrentID=0, Status='live', Symbol='RPLP0P5', Aliases='RPLP0_1_167', description='ribosomal protein, large, P0 pseudogene 5', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220316595, end_position_on_the_genomic_accession=220317683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370308, CurrentID=0, Status='live', Symbol='LINC01040', Aliases='TCONS_00021682', description='long intergenic non-protein coding RNA 1040', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89477608, end_position_on_the_genomic_accession=89480111, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370309, CurrentID=0, Status='live', Symbol='LOC105370309', Aliases='', description='uncharacterized LOC105370309', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89551237, end_position_on_the_genomic_accession=89559573, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370310, CurrentID=0, Status='live', Symbol='LOC105370310', Aliases='', description='uncharacterized LOC105370310', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90291247, end_position_on_the_genomic_accession=90300470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643783, CurrentID=0, Status='live', Symbol='USP46-AS1', Aliases='', description='USP46 antisense RNA 1', other_designations='CTD-2314I6.1|USP46 antisense RNA 1 (non-protein coding)', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52659406, end_position_on_the_genomic_accession=52661498, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370312, CurrentID=0, Status='live', Symbol='LOC105370312', Aliases='', description='uncharacterized LOC105370312', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91111764, end_position_on_the_genomic_accession=91145449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370313, CurrentID=0, Status='live', Symbol='LOC105370313', Aliases='', description='uncharacterized LOC105370313', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91241641, end_position_on_the_genomic_accession=91251661, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370314, CurrentID=0, Status='live', Symbol='LOC105370314', Aliases='', description='uncharacterized LOC105370314', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91251693, end_position_on_the_genomic_accession=91281078, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370315, CurrentID=0, Status='live', Symbol='LOC105370315', Aliases='', description='uncharacterized LOC105370315', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92610757, end_position_on_the_genomic_accession=92677690, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370316, CurrentID=0, Status='live', Symbol='LOC105370316', Aliases='', description='uncharacterized LOC105370316', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=92329272, end_position_on_the_genomic_accession=92331227, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370317, CurrentID=0, Status='live', Symbol='LOC105370317', Aliases='', description='uncharacterized LOC105370317', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93032987, end_position_on_the_genomic_accession=93034267, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643790, CurrentID=0, Status='live', Symbol='HMGN1P38', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 38', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 38', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92711694, end_position_on_the_genomic_accession=92712907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370319, CurrentID=0, Status='live', Symbol='LOC105370319', Aliases='', description='uncharacterized LOC105370319', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94488526, end_position_on_the_genomic_accession=94490069, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  119504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119504, CurrentID=0, Status='live', Symbol='ANAPC16', Aliases='APC16, C10orf104, CENP-27, MSAG, bA570G20.3', description='anaphase promoting complex subunit 16', other_designations='centromere protein 27|cyclosome subunit 16|metabolic syndrome-associated protein', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72216000, end_position_on_the_genomic_accession=72235860, orientation='plus', exon_count=5, OMIM=613427, no_hdr0=''),
  105370321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370321, CurrentID=0, Status='live', Symbol='LOC105370321', Aliases='', description='uncharacterized LOC105370321', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95403972, end_position_on_the_genomic_accession=95405018, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370322, CurrentID=0, Status='live', Symbol='LOC105370322', Aliases='', description='uncharacterized LOC105370322', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=95797083, end_position_on_the_genomic_accession=95801060, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370323, CurrentID=0, Status='live', Symbol='LOC105370323', Aliases='', description='uncharacterized LOC105370323', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96670615, end_position_on_the_genomic_accession=96678049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370324, CurrentID=0, Status='live', Symbol='LOC105370324', Aliases='', description='uncharacterized LOC105370324', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97539220, end_position_on_the_genomic_accession=97711044, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  643797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643797, CurrentID=0, Status='live', Symbol='AGVR6190', Aliases='', description='uncharacterized LOC643797', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92714200, end_position_on_the_genomic_accession=92715755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370326, CurrentID=0, Status='live', Symbol='LOC105370326', Aliases='', description='uncharacterized LOC105370326', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98395608, end_position_on_the_genomic_accession=98408327, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370327, CurrentID=0, Status='live', Symbol='LOC105370327', Aliases='', description='uncharacterized LOC105370327', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98328165, end_position_on_the_genomic_accession=98333238, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370328, CurrentID=0, Status='live', Symbol='LOC105370328', Aliases='', description='uncharacterized LOC105370328', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98066312, end_position_on_the_genomic_accession=98143991, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370329, CurrentID=0, Status='live', Symbol='LOC105370329', Aliases='', description='uncharacterized LOC105370329', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370330, CurrentID=0, Status='live', Symbol='LOC105370330', Aliases='', description='uncharacterized LOC105370330', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98755320, end_position_on_the_genomic_accession=98758154, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  85455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=85455, CurrentID=0, Status='live', Symbol='DISP2', Aliases='DISPB, HsT16908', description='dispatched homolog 2 (Drosophila)', other_designations='dispatched B', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40354359, end_position_on_the_genomic_accession=40371055, orientation='plus', exon_count=9, OMIM=607503, no_hdr0=''),
  105370332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370332, CurrentID=0, Status='live', Symbol='LOC105370332', Aliases='', description='uncharacterized LOC105370332', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99594230, end_position_on_the_genomic_accession=99606541, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370333, CurrentID=0, Status='live', Symbol='LOC105370333', Aliases='', description='uncharacterized LOC105370333', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100086228, end_position_on_the_genomic_accession=100088973, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370334, CurrentID=0, Status='live', Symbol='LOC105370334', Aliases='', description='uncharacterized LOC105370334', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100494670, end_position_on_the_genomic_accession=100506387, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370335, CurrentID=0, Status='live', Symbol='LOC105370335', Aliases='', description='uncharacterized LOC105370335', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100313328, end_position_on_the_genomic_accession=100364942, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105370336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370336, CurrentID=0, Status='live', Symbol='LOC105370336', Aliases='', description='uncharacterized LOC105370336', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102692351, end_position_on_the_genomic_accession=102693111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  150992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=150992, CurrentID=0, Status='live', Symbol='LINC00309', Aliases='NCRNA00309', description='long intergenic non-protein coding RNA 309', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64185078, end_position_on_the_genomic_accession=64205485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370338, CurrentID=0, Status='live', Symbol='LOC105370338', Aliases='', description='uncharacterized LOC105370338', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=103290969, end_position_on_the_genomic_accession=103306424, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370339, CurrentID=0, Status='live', Symbol='LOC105370339', Aliases='', description='uncharacterized LOC105370339', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=103392237, end_position_on_the_genomic_accession=103396516, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643812, CurrentID=0, Status='live', Symbol='KRTAP27-1', Aliases='', description='keratin associated protein 27-1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30337013, end_position_on_the_genomic_accession=30337694, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370341, CurrentID=0, Status='live', Symbol='LOC105370341', Aliases='', description='uncharacterized LOC105370341', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=104366414, end_position_on_the_genomic_accession=104368550, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370342, CurrentID=0, Status='live', Symbol='LOC105370342', Aliases='', description='uncharacterized LOC105370342', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=104851739, end_position_on_the_genomic_accession=104869164, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928401, CurrentID=0, Status='live', Symbol='LOC101928401', Aliases='', description='uncharacterized LOC101928401', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56534230, end_position_on_the_genomic_accession=56538130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370344, CurrentID=0, Status='live', Symbol='LOC105370344', Aliases='', description='uncharacterized LOC105370344', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105410675, end_position_on_the_genomic_accession=105414224, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370345, CurrentID=0, Status='live', Symbol='LOC105370345', Aliases='', description='uncharacterized LOC105370345', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105571962, end_position_on_the_genomic_accession=105706757, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105370346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370346, CurrentID=0, Status='live', Symbol='LOC105370346', Aliases='', description='uncharacterized LOC105370346', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106471747, end_position_on_the_genomic_accession=106473920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370347, CurrentID=0, Status='live', Symbol='LOC105370347', Aliases='', description='uncharacterized LOC105370347', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106479421, end_position_on_the_genomic_accession=106484457, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370348, CurrentID=0, Status='live', Symbol='LOC105370348', Aliases='', description='uncharacterized LOC105370348', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106568251, end_position_on_the_genomic_accession=106613835, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370349, CurrentID=0, Status='live', Symbol='LOC105370349', Aliases='', description='uncharacterized LOC105370349', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106773633, end_position_on_the_genomic_accession=106787337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370350, CurrentID=0, Status='live', Symbol='LOC105370350', Aliases='', description='uncharacterized LOC105370350', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107124862, end_position_on_the_genomic_accession=107132924, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370351, CurrentID=0, Status='live', Symbol='LOC105370351', Aliases='', description='uncharacterized LOC105370351', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107120884, end_position_on_the_genomic_accession=107135953, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370352, CurrentID=0, Status='live', Symbol='LOC105370352', Aliases='', description='uncharacterized LOC105370352', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107836230, end_position_on_the_genomic_accession=107838396, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370353, CurrentID=0, Status='live', Symbol='LOC105370353', Aliases='', description='uncharacterized LOC105370353', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=107465984, end_position_on_the_genomic_accession=107468710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370354, CurrentID=0, Status='live', Symbol='LOC105370354', Aliases='', description='uncharacterized LOC105370354', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=108242990, end_position_on_the_genomic_accession=108267801, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643827, CurrentID=0, Status='live', Symbol='CNTNAP3P2', Aliases='', description='contactin associated protein-like 3 pseudogene 2', other_designations='CASPR3 pseudogene', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67059460, end_position_on_the_genomic_accession=67294592, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105370356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370356, CurrentID=0, Status='live', Symbol='LOC105370356', Aliases='', description='uncharacterized LOC105370356', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109148533, end_position_on_the_genomic_accession=109150642, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370357, CurrentID=0, Status='live', Symbol='LINC00370', Aliases='', description='long intergenic non-protein coding RNA 370', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109282027, end_position_on_the_genomic_accession=109287470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370358, CurrentID=0, Status='live', Symbol='LOC105370358', Aliases='', description='uncharacterized LOC105370358', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109591740, end_position_on_the_genomic_accession=109606748, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  643831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643831, CurrentID=0, Status='live', Symbol='PDCL3P2', Aliases='', description='phosducin-like 3 pseudogene 2', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101871595, end_position_on_the_genomic_accession=101872579, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370360, CurrentID=0, Status='live', Symbol='LOC105370360', Aliases='', description='uncharacterized LOC105370360', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109796975, end_position_on_the_genomic_accession=109811425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928404, CurrentID=0, Status='live', Symbol='LOC101928404', Aliases='', description='uncharacterized LOC101928404', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=163161675, end_position_on_the_genomic_accession=163213023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643834, CurrentID=0, Status='live', Symbol='PGA3', Aliases='', description='pepsinogen 3, group I (pepsinogen A)', other_designations='pepsinogen A3|pepsinogen-3', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61203495, end_position_on_the_genomic_accession=61212878, orientation='plus', exon_count=9, OMIM=169710, no_hdr0=''),
  105370363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370363, CurrentID=0, Status='live', Symbol='LOC105370363', Aliases='', description='uncharacterized LOC105370363', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110822449, end_position_on_the_genomic_accession=110839127, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643836, CurrentID=0, Status='live', Symbol='ZFP62', Aliases='ZET, ZNF755', description='ZFP62 zinc finger protein', other_designations='zfp-62', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180845338, end_position_on_the_genomic_accession=180861286, orientation='minus', exon_count=5, OMIM=610281, no_hdr0=''),
  643837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643837, CurrentID=0, Status='live', Symbol='LINC01128', Aliases='', description='long intergenic non-protein coding RNA 1128', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=827591, end_position_on_the_genomic_accession=859446, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370366, CurrentID=0, Status='live', Symbol='LOC105370366', Aliases='', description='uncharacterized LOC105370366', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111663399, end_position_on_the_genomic_accession=111669523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370367, CurrentID=0, Status='live', Symbol='LOC105370367', Aliases='', description='collagen alpha-2(VIII) chain-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96086390, end_position_on_the_genomic_accession=96090864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370368, CurrentID=0, Status='live', Symbol='LOC105370368', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100673259, end_position_on_the_genomic_accession=100676985, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370369, CurrentID=0, Status='live', Symbol='LOC105370369', Aliases='', description='uncharacterized LOC105370369', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111623249, end_position_on_the_genomic_accession=111642014, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370370, CurrentID=0, Status='live', Symbol='LOC105370370', Aliases='', description='uncharacterized LOC105370370', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112180489, end_position_on_the_genomic_accession=112195688, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105370371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370371, CurrentID=0, Status='live', Symbol='LOC105370371', Aliases='', description='uncharacterized LOC105370371', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112183274, end_position_on_the_genomic_accession=112184109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370372, CurrentID=0, Status='live', Symbol='LOC105370372', Aliases='', description='uncharacterized LOC105370372', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112378738, end_position_on_the_genomic_accession=112476050, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370373, CurrentID=0, Status='live', Symbol='LOC105370373', Aliases='', description='uncharacterized LOC105370373', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112361613, end_position_on_the_genomic_accession=112376268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370374, CurrentID=0, Status='live', Symbol='LOC105370374', Aliases='', description='uncharacterized LOC105370374', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112736753, end_position_on_the_genomic_accession=112738367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643847, CurrentID=0, Status='live', Symbol='PGA4', Aliases='', description='pepsinogen 4, group I (pepsinogen A)', other_designations='pepsinogen A4|pepsinogen-4', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61222331, end_position_on_the_genomic_accession=61231707, orientation='plus', exon_count=9, OMIM=169720, no_hdr0=''),
  105370376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370376, CurrentID=0, Status='live', Symbol='LOC105370376', Aliases='', description='uncharacterized LOC105370376', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113165095, end_position_on_the_genomic_accession=113171597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370377, CurrentID=0, Status='live', Symbol='LOC105370377', Aliases='', description='uncharacterized LOC105370377', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113267504, end_position_on_the_genomic_accession=113269306, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370378, CurrentID=0, Status='live', Symbol='LOC105370378', Aliases='', description='splicing factor 3B subunit 4-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643851, CurrentID=0, Status='live', Symbol='RPL31P32', Aliases='RPL31_14_719', description='ribosomal protein L31 pseudogene 32', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85257288, end_position_on_the_genomic_accession=85257741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370380, CurrentID=0, Status='live', Symbol='LOC105370380', Aliases='', description='uncharacterized LOC105370380', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113817492, end_position_on_the_genomic_accession=113820631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643853, CurrentID=0, Status='live', Symbol='TMPPE', Aliases='', description='transmembrane protein with metallophosphoesterase domain', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33090416, end_position_on_the_genomic_accession=33096801, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643854, CurrentID=0, Status='live', Symbol='CTAGE9', Aliases='', description='CTAGE family member 9', other_designations='CTAGE family, member 9|cutaneous T-cell lymphoma-associated antigen 9|protein cTAGE-9', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=131708441, end_position_on_the_genomic_accession=131711017, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370383, CurrentID=0, Status='live', Symbol='LOC105370383', Aliases='', description='uncharacterized LOC105370383', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370384, CurrentID=0, Status='live', Symbol='LOC105370384', Aliases='', description='uncharacterized LOC105370384', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114244976, end_position_on_the_genomic_accession=114281572, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105370385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370385, CurrentID=0, Status='live', Symbol='LOC105370385', Aliases='', description='uncharacterized LOC105370385', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643858, CurrentID=0, Status='live', Symbol='KRT18P30', Aliases='', description='keratin 18 pseudogene 30', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85277425, end_position_on_the_genomic_accession=85285660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  104355297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355297, CurrentID=0, Status='live', Symbol='ABCA9-AS1', Aliases='', description='ABCA9 antisense RNA 1', other_designations='ABCA9 antisense RNA 1 (non-protein coding)', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69004714, end_position_on_the_genomic_accession=69018323, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370388, CurrentID=0, Status='live', Symbol='LOC105370388', Aliases='', description='uncharacterized LOC105370388', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370389, CurrentID=0, Status='live', Symbol='LOC105370389', Aliases='', description='uncharacterized LOC105370389', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19389307, end_position_on_the_genomic_accession=19392749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643862, CurrentID=0, Status='live', Symbol='SPDYE10P', Aliases='', description='speedy/RINGO cell cycle regulator family member E10, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73104020, end_position_on_the_genomic_accession=73111316, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643863, CurrentID=0, Status='live', Symbol='RPL17P34', Aliases='RPL17_10_1081', description='ribosomal protein L17 pseudogene 34', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93283969, end_position_on_the_genomic_accession=93284593, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370392, CurrentID=0, Status='live', Symbol='LOC105370392', Aliases='', description='uncharacterized LOC105370392', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643865, CurrentID=0, Status='live', Symbol='KRT126P', Aliases='KRT', description='keratin 126 pseudogene', other_designations='keratin 76 pseudogene', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52714138, end_position_on_the_genomic_accession=52721803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643866, CurrentID=0, Status='live', Symbol='CBLN3', Aliases='PRO1486', description='cerebellin 3 precursor', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24426532, end_position_on_the_genomic_accession=24429525, orientation='minus', exon_count=3, OMIM=612978, no_hdr0=''),
  105370395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370395, CurrentID=0, Status='live', Symbol='LOC105370395', Aliases='', description='uncharacterized LOC105370395', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20518748, end_position_on_the_genomic_accession=20521984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370396, CurrentID=0, Status='live', Symbol='LOC105370396', Aliases='', description='uncharacterized LOC105370396', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20652860, end_position_on_the_genomic_accession=20664117, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370397, CurrentID=0, Status='live', Symbol='LOC105370397', Aliases='', description='uncharacterized LOC105370397', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20791139, end_position_on_the_genomic_accession=20810050, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  643870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643870, CurrentID=0, Status='live', Symbol='TPT1P6', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 6', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85427251, end_position_on_the_genomic_accession=85428054, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370399, CurrentID=0, Status='live', Symbol='LOC105370399', Aliases='', description='uncharacterized LOC105370399', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22556953, end_position_on_the_genomic_accession=22559793, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643872, CurrentID=0, Status='live', Symbol='HMGN2P6', Aliases='HMGN2L6', description='high mobility group nucleosomal binding domain 2 pseudogene 6', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 6|high-mobility group nucleosomal binding domain 2-like 6', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25266253, end_position_on_the_genomic_accession=25267446, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643873, CurrentID=0, Status='live', Symbol='EIF4BP7', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 7', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111617654, end_position_on_the_genomic_accession=111621516, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370402, CurrentID=0, Status='live', Symbol='LOC105370402', Aliases='', description='uncharacterized LOC105370402', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22379819, end_position_on_the_genomic_accession=22432760, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  119587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119587, CurrentID=0, Status='live', Symbol='CPXM2', Aliases='CPX2, UNQ676', description='carboxypeptidase X (M14 family), member 2', other_designations='4632435C11Rik|carboxypeptidase Hlo|carboxypeptidase-like protein X2|cytosolic carboxypeptidase', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123722291, end_position_on_the_genomic_accession=123891984, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105370404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370404, CurrentID=0, Status='live', Symbol='LOC105370404', Aliases='', description='uncharacterized LOC105370404', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22702611, end_position_on_the_genomic_accession=22766506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370405, CurrentID=0, Status='live', Symbol='LOC105370405', Aliases='', description='uncharacterized LOC105370405', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22695939, end_position_on_the_genomic_accession=22702320, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370406, CurrentID=0, Status='live', Symbol='LOC105370406', Aliases='', description='uncharacterized LOC105370406', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23195228, end_position_on_the_genomic_accession=23214034, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370407, CurrentID=0, Status='live', Symbol='LOC105370407', Aliases='', description='uncharacterized LOC105370407', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23660501, end_position_on_the_genomic_accession=23665222, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643880, CurrentID=0, Status='live', Symbol='FKBP4P8', Aliases='', description='FK506 binding protein 4 pseudogene 8', other_designations='FK506 binding protein 4, 59kDa pseudogene 8', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66744465, end_position_on_the_genomic_accession=66746073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370409, CurrentID=0, Status='live', Symbol='LOC105370409', Aliases='', description='uncharacterized LOC105370409', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23699821, end_position_on_the_genomic_accession=23729741, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105370410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370410, CurrentID=0, Status='live', Symbol='LOC105370410', Aliases='', description='uncharacterized LOC105370410', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23699476, end_position_on_the_genomic_accession=23704396, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370411, CurrentID=0, Status='live', Symbol='LOC105370411', Aliases='', description='uncharacterized LOC105370411', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23927914, end_position_on_the_genomic_accession=23929673, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643884, CurrentID=0, Status='live', Symbol='SOCS5P5', Aliases='', description='suppressor of cytokine signaling 5 pseudogene 5', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113222166, end_position_on_the_genomic_accession=113224069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370413, CurrentID=0, Status='live', Symbol='LOC105370413', Aliases='', description='uncharacterized LOC105370413', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24638651, end_position_on_the_genomic_accession=24645234, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370414, CurrentID=0, Status='live', Symbol='LOC105370414', Aliases='', description='uncharacterized LOC105370414', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25092854, end_position_on_the_genomic_accession=25156313, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105370415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370415, CurrentID=0, Status='live', Symbol='LOC105370415', Aliases='', description='uncharacterized LOC105370415', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25246570, end_position_on_the_genomic_accession=25262931, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370416, CurrentID=0, Status='live', Symbol='LOC105370416', Aliases='', description='uncharacterized LOC105370416', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25827781, end_position_on_the_genomic_accession=25830174, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370417, CurrentID=0, Status='live', Symbol='LOC105370417', Aliases='', description='uncharacterized LOC105370417', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=25918408, end_position_on_the_genomic_accession=25920501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370418, CurrentID=0, Status='live', Symbol='LOC105370418', Aliases='', description='uncharacterized LOC105370418', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26382230, end_position_on_the_genomic_accession=26397318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370419, CurrentID=0, Status='live', Symbol='LOC105370419', Aliases='', description='uncharacterized LOC105370419', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=26952546, end_position_on_the_genomic_accession=26970939, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370420, CurrentID=0, Status='live', Symbol='LOC105370420', Aliases='', description='uncharacterized LOC105370420', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27077236, end_position_on_the_genomic_accession=27105768, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370421, CurrentID=0, Status='live', Symbol='LOC105370421', Aliases='', description='uncharacterized LOC105370421', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=27213312, end_position_on_the_genomic_accession=27220776, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370422, CurrentID=0, Status='live', Symbol='LOC105370422', Aliases='', description='uncharacterized LOC105370422', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28590808, end_position_on_the_genomic_accession=28613626, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370423, CurrentID=0, Status='live', Symbol='LOC105370423', Aliases='', description='uncharacterized LOC105370423', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28792018, end_position_on_the_genomic_accession=28830166, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  643896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643896, CurrentID=0, Status='live', Symbol='LOC643896', Aliases='', description='cAMP-regulated phosphoprotein 19kDa pseudogene', other_designations='cAMP-regulated phosphoprotein, 19kDa pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100509265, end_position_on_the_genomic_accession=100510358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370425, CurrentID=0, Status='live', Symbol='LOC105370425', Aliases='', description='uncharacterized LOC105370425', other_designations='CTD-2384A14.2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28975733, end_position_on_the_genomic_accession=29064973, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  643898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643898, CurrentID=0, Status='live', Symbol='ARL2BPP2', Aliases='ARL2BPL, bcm1040', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 2', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 2|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 2|ADP-ribosylation factor-like 2 binding protein pseudogene 2', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52763382, end_position_on_the_genomic_accession=52763893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370427, CurrentID=0, Status='live', Symbol='LOC105370427', Aliases='', description='uncharacterized LOC105370427', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=29378316, end_position_on_the_genomic_accession=29391835, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370428, CurrentID=0, Status='live', Symbol='LOC105370428', Aliases='', description='uncharacterized LOC105370428', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=29427312, end_position_on_the_genomic_accession=29436412, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370429, CurrentID=0, Status='live', Symbol='LOC105370429', Aliases='', description='uncharacterized LOC105370429', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=29928111, end_position_on_the_genomic_accession=29936050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370430, CurrentID=0, Status='live', Symbol='LOC105370430', Aliases='', description='uncharacterized LOC105370430', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30379187, end_position_on_the_genomic_accession=30383630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370431, CurrentID=0, Status='live', Symbol='LOC105370431', Aliases='', description='uncharacterized LOC105370431', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30411883, end_position_on_the_genomic_accession=30553405, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  643904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643904, CurrentID=0, Status='live', Symbol='RNF222', Aliases='', description='ring finger protein 222', other_designations='RING finger protein LOC643904', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8390704, end_position_on_the_genomic_accession=8397833, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  643905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643905, CurrentID=0, Status='live', Symbol='PRR21', Aliases='', description='proline rich 21', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240041813, end_position_on_the_genomic_accession=240042982, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  643906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643906, CurrentID=0, Status='live', Symbol='RPL7P27', Aliases='RPL7_11_720', description='ribosomal protein L7 pseudogene 27', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86086349, end_position_on_the_genomic_accession=86087175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370435, CurrentID=0, Status='live', Symbol='LOC105370435', Aliases='', description='uncharacterized LOC105370435', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31210782, end_position_on_the_genomic_accession=31266718, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370436, CurrentID=0, Status='live', Symbol='LOC105370436', Aliases='', description='uncharacterized LOC105370436', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31271956, end_position_on_the_genomic_accession=31279128, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643909, CurrentID=0, Status='live', Symbol='SPDYE9P', Aliases='', description='speedy/RINGO cell cycle regulator family member E9, pseudogene', other_designations='Williams Beuren syndrome chromosome region 19 pseudogene|speedy homolog E1 pseudogene|speedy homolog E9, pseudogene|speedy/RINGO cell cycle regulator family member E1 pseudogene', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73075984, end_position_on_the_genomic_accession=73083282, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370438, CurrentID=0, Status='live', Symbol='LOC105370438', Aliases='', description='uncharacterized LOC105370438', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31489956, end_position_on_the_genomic_accession=31558407, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  643911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643911, CurrentID=0, Status='live', Symbol='CRNDE', Aliases='CRNDEP, LINC00180, NCRNA00180, PNAS-108, lincIRX5', description='colorectal neoplasia differentially expressed (non-protein coding)', other_designations='long intergenic non-protein coding RNA 180', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54918863, end_position_on_the_genomic_accession=54929189, orientation='minus', exon_count=6, OMIM=615624, no_hdr0=''),
  105370440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370440, CurrentID=0, Status='live', Symbol='LOC105370440', Aliases='', description='uncharacterized LOC105370440', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32043836, end_position_on_the_genomic_accession=32057375, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370441, CurrentID=0, Status='live', Symbol='LOC105370441', Aliases='', description='uncharacterized LOC105370441', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=32656263, end_position_on_the_genomic_accession=32660566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370442, CurrentID=0, Status='live', Symbol='LOC105370442', Aliases='', description='uncharacterized LOC105370442', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=33511483, end_position_on_the_genomic_accession=33547810, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  643915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643915, CurrentID=0, Status='live', Symbol='MORF4L2P1', Aliases='MORF4LP4, MRG4', description='mortality factor 4 like 2 pseudogene 1', other_designations='mortality factor 4 like pseudogene 4', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54086685, end_position_on_the_genomic_accession=54088451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370444, CurrentID=0, Status='live', Symbol='LOC105370444', Aliases='', description='uncharacterized LOC105370444', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34095841, end_position_on_the_genomic_accession=34098491, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101928418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928418, CurrentID=0, Status='live', Symbol='LOC101928418', Aliases='', description='uncharacterized LOC101928418', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12671862, end_position_on_the_genomic_accession=12706135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370446, CurrentID=0, Status='live', Symbol='LOC105370446', Aliases='', description='uncharacterized LOC105370446', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=33959082, end_position_on_the_genomic_accession=33964147, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370447, CurrentID=0, Status='live', Symbol='LOC105370447', Aliases='', description='uncharacterized LOC105370447', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34539749, end_position_on_the_genomic_accession=34541706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370448, CurrentID=0, Status='live', Symbol='LOC105370448', Aliases='', description='uncharacterized LOC105370448', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34556669, end_position_on_the_genomic_accession=34559003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370449, CurrentID=0, Status='live', Symbol='LOC105370449', Aliases='', description='uncharacterized LOC105370449', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34551458, end_position_on_the_genomic_accession=34557133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370450, CurrentID=0, Status='live', Symbol='LOC105370450', Aliases='', description='uncharacterized LOC105370450', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35475930, end_position_on_the_genomic_accession=35482810, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370451, CurrentID=0, Status='live', Symbol='LOC105370451', Aliases='', description='uncharacterized LOC105370451', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36084014, end_position_on_the_genomic_accession=36090032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370452, CurrentID=0, Status='live', Symbol='LOC105370452', Aliases='', description='uncharacterized LOC105370452', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36047720, end_position_on_the_genomic_accession=36067486, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370453, CurrentID=0, Status='live', Symbol='LOC105370453', Aliases='', description='uncharacterized LOC105370453', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36523119, end_position_on_the_genomic_accession=36596897, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370454, CurrentID=0, Status='live', Symbol='LOC105370454', Aliases='', description='uncharacterized LOC105370454', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36491302, end_position_on_the_genomic_accession=36498241, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370455, CurrentID=0, Status='live', Symbol='LOC105370455', Aliases='', description='uncharacterized LOC105370455', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36649195, end_position_on_the_genomic_accession=36661690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370456, CurrentID=0, Status='live', Symbol='LOC105370456', Aliases='', description='uncharacterized LOC105370456', other_designations='CTD-2058B24.2|CTD-2058B24.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=38002646, end_position_on_the_genomic_accession=38038590, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928420, CurrentID=0, Status='live', Symbol='LINC01490', Aliases='', description='long intergenic non-protein coding RNA 1490', other_designations='', map_location='12q21.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80763154, end_position_on_the_genomic_accession=80770717, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370458, CurrentID=0, Status='live', Symbol='LOC105370458', Aliases='', description='uncharacterized LOC105370458', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39103140, end_position_on_the_genomic_accession=39103924, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370459, CurrentID=0, Status='live', Symbol='LOC105370459', Aliases='', description='uncharacterized LOC105370459', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39109817, end_position_on_the_genomic_accession=39114815, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  643932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643932, CurrentID=0, Status='live', Symbol='RPS3AP20', Aliases='RPS3A_6_581', description='ribosomal protein S3a pseudogene 20', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79284088, end_position_on_the_genomic_accession=79284946, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370461, CurrentID=0, Status='live', Symbol='LOC105370461', Aliases='', description='uncharacterized LOC105370461', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39432611, end_position_on_the_genomic_accession=39583099, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370462, CurrentID=0, Status='live', Symbol='LOC105370462', Aliases='', description='uncharacterized LOC105370462', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40318353, end_position_on_the_genomic_accession=40390545, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370463, CurrentID=0, Status='live', Symbol='LOC105370463', Aliases='', description='uncharacterized LOC105370463', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40230339, end_position_on_the_genomic_accession=40348046, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370464, CurrentID=0, Status='live', Symbol='LOC105370464', Aliases='', description='uncharacterized LOC105370464', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40217926, end_position_on_the_genomic_accession=40249746, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370465, CurrentID=0, Status='live', Symbol='LOC105370465', Aliases='', description='uncharacterized LOC105370465', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40399268, end_position_on_the_genomic_accession=40447295, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370466, CurrentID=0, Status='live', Symbol='LOC105370466', Aliases='', description='uncharacterized LOC105370466', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40573118, end_position_on_the_genomic_accession=40629697, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370467, CurrentID=0, Status='live', Symbol='LOC105370467', Aliases='', description='uncharacterized LOC105370467', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40767818, end_position_on_the_genomic_accession=40902216, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370468, CurrentID=0, Status='live', Symbol='LOC105370468', Aliases='', description='uncharacterized LOC105370468', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=41062891, end_position_on_the_genomic_accession=41093821, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643941, CurrentID=0, Status='live', Symbol='PHKG1P1', Aliases='', description='phosphorylase kinase, gamma 1 pseudogene 1', other_designations='phosphorylase kinase, gamma 1 (muscle) pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63232963, end_position_on_the_genomic_accession=63234246, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370470, CurrentID=0, Status='live', Symbol='LOC105370470', Aliases='', description='uncharacterized LOC105370470', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43308681, end_position_on_the_genomic_accession=43335865, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370471, CurrentID=0, Status='live', Symbol='LOC105370471', Aliases='', description='uncharacterized LOC105370471', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43990665, end_position_on_the_genomic_accession=44049473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370472, CurrentID=0, Status='live', Symbol='LOC105370472', Aliases='', description='uncharacterized LOC105370472', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44587160, end_position_on_the_genomic_accession=44590415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370473, CurrentID=0, Status='live', Symbol='LOC105370473', Aliases='', description='uncharacterized LOC105370473', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44507362, end_position_on_the_genomic_accession=44568002, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370474, CurrentID=0, Status='live', Symbol='LOC105370474', Aliases='', description='uncharacterized LOC105370474', other_designations='CTD-3028N15.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45082427, end_position_on_the_genomic_accession=45083991, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928423, CurrentID=0, Status='live', Symbol='LOC101928423', Aliases='', description='uncharacterized LOC101928423', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128205761, end_position_on_the_genomic_accession=128224431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370476, CurrentID=0, Status='live', Symbol='LOC105370476', Aliases='', description='uncharacterized LOC105370476', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45568001, end_position_on_the_genomic_accession=45569061, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643949, CurrentID=0, Status='live', Symbol='RPLP2P3', Aliases='RPLP2_2_1170', description='ribosomal protein, large, P2 pseudogene 3', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=107908347, end_position_on_the_genomic_accession=107908802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370478, CurrentID=0, Status='live', Symbol='LOC105370478', Aliases='', description='uncharacterized LOC105370478', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45744035, end_position_on_the_genomic_accession=45793230, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370479, CurrentID=0, Status='live', Symbol='LOC105370479', Aliases='', description='uncharacterized LOC105370479', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45706258, end_position_on_the_genomic_accession=45716018, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370480, CurrentID=0, Status='live', Symbol='LOC105370480', Aliases='', description='uncharacterized LOC105370480', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46578548, end_position_on_the_genomic_accession=46583457, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370481, CurrentID=0, Status='live', Symbol='LOC105370481', Aliases='', description='uncharacterized LOC105370481', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=46596162, end_position_on_the_genomic_accession=46660347, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  643954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643954, CurrentID=0, Status='live', Symbol='RPSAP43', Aliases='RPSA_12_742', description='ribosomal protein SA pseudogene 43', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114084098, end_position_on_the_genomic_accession=114084793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643955, CurrentID=0, Status='live', Symbol='ZNF733P', Aliases='ZNF733', description='zinc finger protein 733, pseudogene', other_designations='Putative zinc finger protein ENSP00000353728|zinc finger protein 479 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63291287, end_position_on_the_genomic_accession=63304056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370484, CurrentID=0, Status='live', Symbol='LOC105370484', Aliases='', description='uncharacterized LOC105370484', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49544143, end_position_on_the_genomic_accession=49550135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370485, CurrentID=0, Status='live', Symbol='LOC105370485', Aliases='', description='uncharacterized LOC105370485', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49987875, end_position_on_the_genomic_accession=49990483, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370486, CurrentID=0, Status='live', Symbol='LOC105370486', Aliases='', description='uncharacterized LOC105370486', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50174567, end_position_on_the_genomic_accession=50230822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370487, CurrentID=0, Status='live', Symbol='LOC105370487', Aliases='', description='uncharacterized LOC105370487', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50396329, end_position_on_the_genomic_accession=50406548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643960, CurrentID=0, Status='live', Symbol='RPS2P18', Aliases='RPS2_7_293', description='ribosomal protein S2 pseudogene 18', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173297610, end_position_on_the_genomic_accession=173298554, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370489, CurrentID=0, Status='live', Symbol='LOC105370489', Aliases='', description='uncharacterized LOC105370489', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50821862, end_position_on_the_genomic_accession=50823513, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370490, CurrentID=0, Status='live', Symbol='LOC105370490', Aliases='', description='uncharacterized LOC105370490', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50744324, end_position_on_the_genomic_accession=50746052, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370491, CurrentID=0, Status='live', Symbol='LOC105370491', Aliases='', description='uncharacterized LOC105370491', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50838283, end_position_on_the_genomic_accession=50860250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370492, CurrentID=0, Status='live', Symbol='LOC105370492', Aliases='', description='uncharacterized LOC105370492', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=50944563, end_position_on_the_genomic_accession=50946699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  643965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643965, CurrentID=0, Status='live', Symbol='TMEM88B', Aliases='', description='transmembrane protein 88B', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1426128, end_position_on_the_genomic_accession=1427787, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  119678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119678, CurrentID=0, Status='live', Symbol='OR52E2', Aliases='', description='olfactory receptor family 52 subfamily E member 2', other_designations='olfactory receptor, family 10, subfamily AC, member 1 pseudogene|olfactory receptor, family 52, subfamily E, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5058650, end_position_on_the_genomic_accession=5059627, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  119679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119679, CurrentID=0, Status='live', Symbol='OR52J3', Aliases='OR11-32', description='olfactory receptor family 52 subfamily J member 3', other_designations='olfactory receptor OR11-32|olfactory receptor, family 52, subfamily J, member 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5046526, end_position_on_the_genomic_accession=5047461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370496, CurrentID=0, Status='live', Symbol='LOC105370496', Aliases='', description='uncharacterized LOC105370496', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51391700, end_position_on_the_genomic_accession=51393703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370497, CurrentID=0, Status='live', Symbol='LOC105370497', Aliases='', description='uncharacterized LOC105370497', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51982520, end_position_on_the_genomic_accession=51987547, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  119682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119682, CurrentID=0, Status='live', Symbol='OR51L1', Aliases='OR11-31', description='olfactory receptor family 51 subfamily L member 1', other_designations='olfactory receptor OR11-31|olfactory receptor, family 51, subfamily L, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4998983, end_position_on_the_genomic_accession=4999930, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643971, CurrentID=0, Status='live', Symbol='RPL7P29', Aliases='RPL7_13_721', description='ribosomal protein L7 pseudogene 29', other_designations='', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=86970434, end_position_on_the_genomic_accession=86971258, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372234, CurrentID=0, Status='live', Symbol='LOC105372234', Aliases='', description='uncharacterized LOC105372234', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=633860, end_position_on_the_genomic_accession=634335, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370501, CurrentID=0, Status='live', Symbol='LOC105370501', Aliases='', description='uncharacterized LOC105370501', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  643974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643974, CurrentID=0, Status='live', Symbol='RPL6P13', Aliases='RPL6_10_498', description='ribosomal protein L6 pseudogene 13', other_designations='', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86870165, end_position_on_the_genomic_accession=86871433, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  119687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119687, CurrentID=0, Status='live', Symbol='OR51A7', Aliases='OR11-27', description='olfactory receptor family 51 subfamily A member 7', other_designations='olfactory receptor OR11-27|olfactory receptor, family 51, subfamily A, member 7', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4907370, end_position_on_the_genomic_accession=4908308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  643976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643976, CurrentID=0, Status='live', Symbol='FOXN3P1', Aliases='', description='forkhead box N3 pseudogene 1', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102546739, end_position_on_the_genomic_accession=102548214, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643977, CurrentID=0, Status='live', Symbol='FLJ32255', Aliases='', description='uncharacterized LOC643977', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42985398, end_position_on_the_genomic_accession=42993333, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370506, CurrentID=0, Status='live', Symbol='LOC105370506', Aliases='', description='uncharacterized LOC105370506', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54113339, end_position_on_the_genomic_accession=54114688, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370507, CurrentID=0, Status='live', Symbol='LOC105370507', Aliases='', description='uncharacterized LOC105370507', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=54043865, end_position_on_the_genomic_accession=54188479, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  119692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119692, CurrentID=0, Status='live', Symbol='OR51S1', Aliases='OR11-24', description='olfactory receptor family 51 subfamily S member 1', other_designations='olfactory receptor OR11-24|olfactory receptor, family 51, subfamily S, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4848203, end_position_on_the_genomic_accession=4859600, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643981, CurrentID=0, Status='live', Symbol='RPS3AP36', Aliases='RPS3A_16_1083', description='ribosomal protein S3a pseudogene 36', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95594561, end_position_on_the_genomic_accession=95595425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  119694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119694, CurrentID=0, Status='live', Symbol='OR51F2', Aliases='OR11-23', description='olfactory receptor family 51 subfamily F member 2', other_designations='olfactory receptor OR11-23|olfactory receptor, family 51, subfamily F, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4821386, end_position_on_the_genomic_accession=4822414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  119695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119695, CurrentID=0, Status='live', Symbol='OR52R1', Aliases='OR11-22', description='olfactory receptor family 52 subfamily R member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-22|olfactory receptor, family 52, subfamily R, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4803433, end_position_on_the_genomic_accession=4804380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  119696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119696, CurrentID=0, Status='live', Symbol='OR51N1P', Aliases='', description='olfactory receptor family 51 subfamily N member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily N, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4786754, end_position_on_the_genomic_accession=4787702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370513, CurrentID=0, Status='live', Symbol='LOC105370513', Aliases='', description='uncharacterized LOC105370513', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56095849, end_position_on_the_genomic_accession=56109398, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370514, CurrentID=0, Status='live', Symbol='LOC105370514', Aliases='', description='uncharacterized LOC105370514', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56389314, end_position_on_the_genomic_accession=56428250, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  643987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643987, CurrentID=0, Status='live', Symbol='HSFY4P', Aliases='', description='heat shock transcription factor, Y-linked 4, pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=20808776, end_position_on_the_genomic_accession=20843588, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  643988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643988, CurrentID=0, Status='live', Symbol='C1orf233', Aliases='', description='chromosome 1 open reading frame 233', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1598008, end_position_on_the_genomic_accession=1600096, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370517, CurrentID=0, Status='live', Symbol='LOC105370517', Aliases='', description='uncharacterized LOC105370517', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57491529, end_position_on_the_genomic_accession=57509906, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370518, CurrentID=0, Status='live', Symbol='LOC105370518', Aliases='', description='uncharacterized LOC105370518', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57578744, end_position_on_the_genomic_accession=57581514, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370519, CurrentID=0, Status='live', Symbol='LOC105370519', Aliases='', description='uncharacterized LOC105370519', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57569947, end_position_on_the_genomic_accession=57578282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  643992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643992, CurrentID=0, Status='live', Symbol='RPS2P26', Aliases='RPS2_8_601', description='ribosomal protein S2 pseudogene 26', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115264331, end_position_on_the_genomic_accession=115265037, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370521, CurrentID=0, Status='live', Symbol='LOC105370521', Aliases='', description='uncharacterized LOC105370521', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58399134, end_position_on_the_genomic_accession=58408330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370522, CurrentID=0, Status='live', Symbol='LOC105370522', Aliases='', description='uncharacterized LOC105370522', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58369613, end_position_on_the_genomic_accession=58395848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370523, CurrentID=0, Status='live', Symbol='LOC105370523', Aliases='', description='uncharacterized LOC105370523', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58634189, end_position_on_the_genomic_accession=58645715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  643996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=643996, CurrentID=0, Status='live', Symbol='VDAC2P2', Aliases='', description='voltage-dependent anion channel 2 pseudogene 2', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9036324, end_position_on_the_genomic_accession=9037523, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370525, CurrentID=0, Status='live', Symbol='LOC105370525', Aliases='', description='uncharacterized LOC105370525', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61534280, end_position_on_the_genomic_accession=61546566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  119710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119710, CurrentID=0, Status='live', Symbol='C11orf74', Aliases='HEPIS', description='chromosome 11 open reading frame 74', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36594229, end_position_on_the_genomic_accession=36659291, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105370527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370527, CurrentID=0, Status='live', Symbol='LOC105370527', Aliases='', description='uncharacterized LOC105370527', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62114601, end_position_on_the_genomic_accession=62117234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644000, CurrentID=0, Status='live', Symbol='ARAFP2', Aliases='ARAF2, ARAF2P, ARAFPS, PKS1', description='ARAF pseudogene 2', other_designations='v-raf murine sarcoma 3611 viral oncogene homolog 2, pseudogene|v-raf murine sarcoma 3611 viral oncogene homolog pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63404378, end_position_on_the_genomic_accession=63405398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370529, CurrentID=0, Status='live', Symbol='LOC105370529', Aliases='', description='uncharacterized LOC105370529', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62357919, end_position_on_the_genomic_accession=62505769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370530, CurrentID=0, Status='live', Symbol='LOC105370530', Aliases='', description='uncharacterized LOC105370530', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63178247, end_position_on_the_genomic_accession=63179993, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370531, CurrentID=0, Status='live', Symbol='LOC105370531', Aliases='', description='uncharacterized LOC105370531', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63123046, end_position_on_the_genomic_accession=63219108, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370532, CurrentID=0, Status='live', Symbol='LOC105370532', Aliases='', description='uncharacterized LOC105370532', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63642071, end_position_on_the_genomic_accession=63651508, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370533, CurrentID=0, Status='live', Symbol='LOC105370533', Aliases='', description='uncharacterized LOC105370533', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64552800, end_position_on_the_genomic_accession=64658022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370534, CurrentID=0, Status='live', Symbol='LOC105370534', Aliases='', description='uncharacterized LOC105370534', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64880649, end_position_on_the_genomic_accession=64902128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370535, CurrentID=0, Status='live', Symbol='LOC105370535', Aliases='', description='uncharacterized LOC105370535', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64940477, end_position_on_the_genomic_accession=64942310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370536, CurrentID=0, Status='live', Symbol='LOC105370536', Aliases='', description='uncharacterized LOC105370536', other_designations='CTD-2509G16.5', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65242018, end_position_on_the_genomic_accession=65318761, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370537, CurrentID=0, Status='live', Symbol='LOC105370537', Aliases='', description='uncharacterized LOC105370537', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65374866, end_position_on_the_genomic_accession=65378384, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  119722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119722, CurrentID=0, Status='live', Symbol='KRT18P14', Aliases='KRT18L4', description='keratin 18 pseudogene 14', other_designations='keratin 18-like 4', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35860185, end_position_on_the_genomic_accession=35861585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370539, CurrentID=0, Status='live', Symbol='LOC105370539', Aliases='', description='uncharacterized LOC105370539', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66127003, end_position_on_the_genomic_accession=66128773, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370540, CurrentID=0, Status='live', Symbol='LOC105370540', Aliases='', description='uncharacterized LOC105370540', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=66111228, end_position_on_the_genomic_accession=66127003, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  644013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644013, CurrentID=0, Status='live', Symbol='RPS6P3', Aliases='RPS6_1_251', description='ribosomal protein S6 pseudogene 3', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101509521, end_position_on_the_genomic_accession=101510343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370542, CurrentID=0, Status='live', Symbol='LOC105370542', Aliases='', description='uncharacterized LOC105370542', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67461846, end_position_on_the_genomic_accession=67468687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370543, CurrentID=0, Status='live', Symbol='LOC105370543', Aliases='', description='uncharacterized LOC105370543', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67600810, end_position_on_the_genomic_accession=67611030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644016, CurrentID=0, Status='live', Symbol='HSPD1P10', Aliases='HSPD1-4P', description='heat shock protein family D (Hsp60) member 1 pseudogene 10', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 10', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87298709, end_position_on_the_genomic_accession=87300641, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644017, CurrentID=0, Status='live', Symbol='ZNF734P', Aliases='ZNF734', description='zinc finger protein 734, pseudogene', other_designations='zinc finger protein 479 pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63444185, end_position_on_the_genomic_accession=63456893, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370546, CurrentID=0, Status='live', Symbol='LOC105370546', Aliases='', description='uncharacterized LOC105370546', other_designations='CTD-2566J3.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68124901, end_position_on_the_genomic_accession=68167516, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  644019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644019, CurrentID=0, Status='live', Symbol='CBWD6', Aliases='', description='COBW domain containing 6', other_designations='cobalamin synthase W domain-containing protein 6|cobalamin synthetase W domain-containing protein 6', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41131309, end_position_on_the_genomic_accession=41189364, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  644020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644020, CurrentID=0, Status='live', Symbol='RPL32P23', Aliases='RPL32_6_1020', description='ribosomal protein L32 pseudogene 23', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6071511, end_position_on_the_genomic_accession=6071921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644021, CurrentID=0, Status='live', Symbol='CTSLP3', Aliases='CTSL1P3, CTSLL3', description='cathepsin L pseudogene 3', other_designations='cathepsin L-like 3|cathepsin L1 pseudogene 3', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46758089, end_position_on_the_genomic_accession=46759065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644022, CurrentID=0, Status='live', Symbol='RPL22P19', Aliases='RPL22_8_1294', description='ribosomal protein L22 pseudogene 19', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=124935480, end_position_on_the_genomic_accession=124935865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370551, CurrentID=0, Status='live', Symbol='LOC105370551', Aliases='', description='uncharacterized LOC105370551', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=69479602, end_position_on_the_genomic_accession=69484436, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370552, CurrentID=0, Status='live', Symbol='LOC105370552', Aliases='', description='uncharacterized LOC105370552', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70244094, end_position_on_the_genomic_accession=70257866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370553, CurrentID=0, Status='live', Symbol='LOC105370553', Aliases='', description='uncharacterized LOC105370553', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70223259, end_position_on_the_genomic_accession=70252008, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370554, CurrentID=0, Status='live', Symbol='LOC105370554', Aliases='', description='transcription factor NF-E4-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70344946, end_position_on_the_genomic_accession=70347628, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480824, CurrentID=0, Status='live', Symbol='RNU7-151P', Aliases='', description='RNA, U7 small nuclear 151 pseudogene', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130447, CurrentID=0, Status='live', Symbol='RPL12P16', Aliases='RPL12_4_315', description='ribosomal protein L12 pseudogene 16', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203190692, end_position_on_the_genomic_accession=203191319, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644029, CurrentID=0, Status='live', Symbol='RPL7AP62', Aliases='RPL7A_28_1429', description='ribosomal protein L7a pseudogene 62', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=47730083, end_position_on_the_genomic_accession=47730960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644030, CurrentID=0, Status='live', Symbol='KRT18P58', Aliases='', description='keratin 18 pseudogene 58', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5993820, end_position_on_the_genomic_accession=5995209, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370559, CurrentID=0, Status='live', Symbol='LOC105370559', Aliases='', description='uncharacterized LOC105370559', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72382395, end_position_on_the_genomic_accession=72420819, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644032, CurrentID=0, Status='live', Symbol='PHKG1P2', Aliases='', description='phosphorylase kinase, gamma 1 pseudogene 2', other_designations='phosphorylase kinase, gamma 1 (muscle) pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63509303, end_position_on_the_genomic_accession=63510959, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370561, CurrentID=0, Status='live', Symbol='LOC105370561', Aliases='', description='uncharacterized LOC105370561', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73462579, end_position_on_the_genomic_accession=73464213, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370562, CurrentID=0, Status='live', Symbol='LOC105370562', Aliases='', description='uncharacterized LOC105370562', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73737266, end_position_on_the_genomic_accession=73747724, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644035, CurrentID=0, Status='live', Symbol='NONOP1', Aliases='', description='non-POU domain containing, octamer-binding pseudogene 1', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69755419, end_position_on_the_genomic_accession=69758070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370564, CurrentID=0, Status='live', Symbol='LOC105370564', Aliases='', description='uncharacterized LOC105370564', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74208080, end_position_on_the_genomic_accession=74217771, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  119749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119749, CurrentID=0, Status='live', Symbol='OR4C46', Aliases='', description='olfactory receptor family 4 subfamily C member 46', other_designations='olfactory receptor, family 4, subfamily C, member 46', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54603069, end_position_on_the_genomic_accession=54603998, orientation='minus', exon_count=1, OMIM=614273, no_hdr0=''),
  119750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119750, CurrentID=0, Status='live', Symbol='OR4C2P', Aliases='OR4C8P', description='olfactory receptor family 4 subfamily C member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily C, member 2 pseudogene|olfactory receptor, family 4, subfamily C, member 8 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48420043, end_position_on_the_genomic_accession=48421079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370567, CurrentID=0, Status='live', Symbol='LOC105370567', Aliases='', description='uncharacterized LOC105370567', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74654527, end_position_on_the_genomic_accession=74657674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370568, CurrentID=0, Status='live', Symbol='LOC105370568', Aliases='', description='uncharacterized LOC105370568', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74858006, end_position_on_the_genomic_accession=74859435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644041, CurrentID=0, Status='live', Symbol='C18orf63', Aliases='DKFZP781G0119', description='chromosome 18 open reading frame 63', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74315875, end_position_on_the_genomic_accession=74359187, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  644042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644042, CurrentID=0, Status='live', Symbol='SLC25A5P6', Aliases='ANTP6', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 6', other_designations='adenine nucleotide translocator pseudogene 6 (clone 3)', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186328455, end_position_on_the_genomic_accession=186329680, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370571, CurrentID=0, Status='live', Symbol='LOC105370571', Aliases='', description='uncharacterized LOC105370571', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75258137, end_position_on_the_genomic_accession=75269717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370572, CurrentID=0, Status='live', Symbol='LOC105370572', Aliases='', description='uncharacterized LOC105370572', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75515339, end_position_on_the_genomic_accession=75517779, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370573, CurrentID=0, Status='live', Symbol='LOC105370573', Aliases='', description='uncharacterized LOC105370573', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76079436, end_position_on_the_genomic_accession=76082288, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370574, CurrentID=0, Status='live', Symbol='LOC105370574', Aliases='', description='uncharacterized LOC105370574', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76475401, end_position_on_the_genomic_accession=76563075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370575, CurrentID=0, Status='live', Symbol='LOC105370575', Aliases='', description='uncharacterized LOC105370575', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76226769, end_position_on_the_genomic_accession=76232694, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370576, CurrentID=0, Status='live', Symbol='LOC105370576', Aliases='', description='uncharacterized LOC105370576', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76590680, end_position_on_the_genomic_accession=76596082, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370577, CurrentID=0, Status='live', Symbol='LOC105370577', Aliases='', description='uncharacterized LOC105370577', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76706060, end_position_on_the_genomic_accession=76726801, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370578, CurrentID=0, Status='live', Symbol='LINC01629', Aliases='linc-KIAA1737-2', description='long intergenic non-protein coding RNA 1629', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76959579, end_position_on_the_genomic_accession=76969939, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370579, CurrentID=0, Status='live', Symbol='LOC105370579', Aliases='', description='uncharacterized LOC105370579', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77032368, end_position_on_the_genomic_accession=77034940, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  119764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119764, CurrentID=0, Status='live', Symbol='OR4X2', Aliases='OR11-105', description='olfactory receptor family 4 subfamily X member 2 (gene/pseudogene)', other_designations='olfactory receptor OR11-105|olfactory receptor, family 4, subfamily X, member 2', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48245104, end_position_on_the_genomic_accession=48246015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  119765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119765, CurrentID=0, Status='live', Symbol='OR4B1', Aliases='OR11-106, OST208', description='olfactory receptor family 4 subfamily B member 1', other_designations='olfactory receptor OR11-106|olfactory receptor, family 4, subfamily B, member 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48216810, end_position_on_the_genomic_accession=48217739, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644054, CurrentID=0, Status='live', Symbol='FAM25C', Aliases='bA164N7.4', description='family with sequence similarity 25 member C', other_designations='family with sequence similarity 25, member C|protein FAM25', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47995322, end_position_on_the_genomic_accession=47999791, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370583, CurrentID=0, Status='live', Symbol='LOC105370583', Aliases='', description='uncharacterized LOC105370583', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79593695, end_position_on_the_genomic_accession=79633227, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370584, CurrentID=0, Status='live', Symbol='LOC105370584', Aliases='', description='uncharacterized LOC105370584', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79248878, end_position_on_the_genomic_accession=79255231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370585, CurrentID=0, Status='live', Symbol='LOC105370585', Aliases='', description='uncharacterized LOC105370585', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=79247064, end_position_on_the_genomic_accession=79248675, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644058, CurrentID=0, Status='live', Symbol='LOC644058', Aliases='', description='hCG2045048', other_designations='', map_location='6p24.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644059, CurrentID=0, Status='live', Symbol='PYY3', Aliases='', description='peptide YY, 3 (pseudogene)', other_designations='peptide YY pseudogene', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50156066, end_position_on_the_genomic_accession=50156596, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  119772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119772, CurrentID=0, Status='live', Symbol='OR52M1', Aliases='OR11-11P, OR52M3P, OR52M1', description='olfactory receptor family 52 subfamily M member 1', other_designations='olfactory receptor OR11-11|olfactory receptor, family 52, subfamily M, member 1 pseudogene|olfactory receptor, family 52, subfamily M, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4545191, end_position_on_the_genomic_accession=4546144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370589, CurrentID=0, Status='live', Symbol='LOC105370589', Aliases='', description='uncharacterized LOC105370589', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78703591, end_position_on_the_genomic_accession=78709969, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  119774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=119774, CurrentID=0, Status='live', Symbol='OR52K2', Aliases='OR11-7', description='olfactory receptor family 52 subfamily K member 2', other_designations='olfactory receptor OR11-7|olfactory receptor, family 52, subfamily K, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4449295, end_position_on_the_genomic_accession=4450361, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644063, CurrentID=0, Status='live', Symbol='HNRNPKP4', Aliases='', description='heterogeneous nuclear ribonucleoprotein K pseudogene 4', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96349554, end_position_on_the_genomic_accession=96350943, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370592, CurrentID=0, Status='live', Symbol='LOC105370592', Aliases='', description='uncharacterized LOC105370592', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80176785, end_position_on_the_genomic_accession=80184480, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370593, CurrentID=0, Status='live', Symbol='LOC105370593', Aliases='', description='uncharacterized LOC105370593', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80167414, end_position_on_the_genomic_accession=80188858, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370594, CurrentID=0, Status='live', Symbol='LOC105370594', Aliases='', description='uncharacterized LOC105370594', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81168156, end_position_on_the_genomic_accession=81171831, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370595, CurrentID=0, Status='live', Symbol='LOC105370595', Aliases='', description='uncharacterized LOC105370595', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81153631, end_position_on_the_genomic_accession=81154080, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644068, CurrentID=0, Status='live', Symbol='RPS14P3', Aliases='RPS14_1_16', description='ribosomal protein S14 pseudogene 3', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19607807, end_position_on_the_genomic_accession=19608644, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370597, CurrentID=0, Status='live', Symbol='LOC105370597', Aliases='', description='uncharacterized LOC105370597', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=80941387, end_position_on_the_genomic_accession=80941981, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644070, CurrentID=0, Status='live', Symbol='GSG1L2', Aliases='', description='GSG1-like 2', other_designations='GSG1-like protein 2|germ cell-specific gene 1-like protein 2-like|putative germ cell-specific gene 1-like protein 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9801473, end_position_on_the_genomic_accession=9809311, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370599, CurrentID=0, Status='live', Symbol='LOC105370599', Aliases='', description='uncharacterized LOC105370599', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=82420191, end_position_on_the_genomic_accession=82428634, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370600, CurrentID=0, Status='live', Symbol='LOC105370600', Aliases='', description='uncharacterized LOC105370600', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=83012206, end_position_on_the_genomic_accession=83039601, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370601, CurrentID=0, Status='live', Symbol='LOC105370601', Aliases='', description='uncharacterized LOC105370601', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=83286826, end_position_on_the_genomic_accession=83307472, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370602, CurrentID=0, Status='live', Symbol='LOC105370602', Aliases='', description='uncharacterized LOC105370602', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=83963372, end_position_on_the_genomic_accession=83974872, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370603, CurrentID=0, Status='live', Symbol='LOC105370603', Aliases='', description='uncharacterized LOC105370603', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=84216548, end_position_on_the_genomic_accession=84299016, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644076, CurrentID=0, Status='live', Symbol='GLYCAM1', Aliases='', description='glycosylation dependent cell adhesion molecule 1 (pseudogene)', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54608187, end_position_on_the_genomic_accession=54610462, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370605, CurrentID=0, Status='live', Symbol='LOC105370605', Aliases='', description='uncharacterized LOC105370605', other_designations='CTD-2128A3.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85387177, end_position_on_the_genomic_accession=85388765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370606, CurrentID=0, Status='live', Symbol='LOC105370606', Aliases='', description='uncharacterized LOC105370606', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=85979533, end_position_on_the_genomic_accession=86007621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644079, CurrentID=0, Status='live', Symbol='BCRP1', Aliases='BCR-1', description='breakpoint cluster region pseudogene 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24257400, end_position_on_the_genomic_accession=24263765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370608, CurrentID=0, Status='live', Symbol='LOC105370608', Aliases='', description='uncharacterized LOC105370608', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377146, CurrentID=0, Status='live', Symbol='LOC105377146', Aliases='', description='uncharacterized LOC105377146', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68500647, end_position_on_the_genomic_accession=68505875, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644082, CurrentID=0, Status='live', Symbol='RPL18AP8', Aliases='RPL18A_3_387', description='ribosomal protein L18a pseudogene 8', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96619269, end_position_on_the_genomic_accession=96619874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644083, CurrentID=0, Status='live', Symbol='LOC644083', Aliases='', description='uncharacterized LOC644083', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19399985, end_position_on_the_genomic_accession=19400561, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370612, CurrentID=0, Status='live', Symbol='LOC105370612', Aliases='', description='uncharacterized LOC105370612', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88160637, end_position_on_the_genomic_accession=88180198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928446, CurrentID=0, Status='live', Symbol='LOC101928446', Aliases='', description='uncharacterized LOC101928446', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=82773319, end_position_on_the_genomic_accession=82829638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370614, CurrentID=0, Status='live', Symbol='LOC105370614', Aliases='', description='uncharacterized LOC105370614', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88500878, end_position_on_the_genomic_accession=88515502, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370615, CurrentID=0, Status='live', Symbol='LOC105370615', Aliases='', description='uncharacterized LOC105370615', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89040490, end_position_on_the_genomic_accession=89075424, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370616, CurrentID=0, Status='live', Symbol='LOC105370616', Aliases='', description='uncharacterized LOC105370616', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89628921, end_position_on_the_genomic_accession=89642079, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370617, CurrentID=0, Status='live', Symbol='LOC105370617', Aliases='', description='uncharacterized LOC105370617', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89619317, end_position_on_the_genomic_accession=89643614, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370618, CurrentID=0, Status='live', Symbol='LOC105370618', Aliases='', description='uncharacterized LOC105370618', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90364433, end_position_on_the_genomic_accession=90368921, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370619, CurrentID=0, Status='live', Symbol='LOC105370619', Aliases='', description='uncharacterized LOC105370619', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90515806, end_position_on_the_genomic_accession=90517226, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370620, CurrentID=0, Status='live', Symbol='LOC105370620', Aliases='', description='uncharacterized LOC105370620', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90524568, end_position_on_the_genomic_accession=90537115, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  644093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644093, CurrentID=0, Status='live', Symbol='HCG2040054', Aliases='', description='uncharacterized LOC644093', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47527537, end_position_on_the_genomic_accession=47529164, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644094, CurrentID=0, Status='live', Symbol='PSMC1P12', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 12', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118614112, end_position_on_the_genomic_accession=118615668, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370623, CurrentID=0, Status='live', Symbol='LOC105370623', Aliases='', description='uncharacterized LOC105370623', other_designations='CTD-3035D6.2', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=90819440, end_position_on_the_genomic_accession=90828175, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  644096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644096, CurrentID=0, Status='live', Symbol='SDHAF1', Aliases='LYRM8', description='succinate dehydrogenase complex assembly factor 1', other_designations='LYR motif containing 8|LYR motif-containing protein 8|SDH assembly factor 1', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35995188, end_position_on_the_genomic_accession=35996318, orientation='plus', exon_count=1, OMIM=612848, no_hdr0=''),
  101928448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928448, CurrentID=0, Status='live', Symbol='C5orf67', Aliases='', description='chromosome 5 open reading frame 67', other_designations='Uncharacterized protein C5orf67', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56511394, end_position_on_the_genomic_accession=56606232, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370626, CurrentID=0, Status='live', Symbol='LOC105370626', Aliases='', description='uncharacterized LOC105370626', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91450012, end_position_on_the_genomic_accession=91458195, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370627, CurrentID=0, Status='live', Symbol='LOC105370627', Aliases='', description='uncharacterized LOC105370627', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92304414, end_position_on_the_genomic_accession=92319650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644100, CurrentID=0, Status='live', Symbol='ARL14EPL', Aliases='', description='ADP ribosylation factor like GTPase 14 effector protein like', other_designations='ADP-ribosylation factor like GTPase 14 effector protein like|ADP-ribosylation factor-like 14 effector protein-like', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116050393, end_position_on_the_genomic_accession=116059426, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644101, CurrentID=0, Status='live', Symbol='CBX3P9', Aliases='', description='chromobox 3 pseudogene 9', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 9', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116453017, end_position_on_the_genomic_accession=116453565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370630, CurrentID=0, Status='live', Symbol='LOC105370630', Aliases='', description='uncharacterized LOC105370630', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93295445, end_position_on_the_genomic_accession=93296187, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644103, CurrentID=0, Status='live', Symbol='RPL29P18', Aliases='RPL29_8_938', description='ribosomal protein L29 pseudogene 18', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105011067, end_position_on_the_genomic_accession=105011677, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370632, CurrentID=0, Status='live', Symbol='LOC105370632', Aliases='', description='uncharacterized LOC105370632', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93939860, end_position_on_the_genomic_accession=93947342, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370633, CurrentID=0, Status='live', Symbol='LOC105370633', Aliases='', description='uncharacterized LOC105370633', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93994704, end_position_on_the_genomic_accession=93997180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370634, CurrentID=0, Status='live', Symbol='LOC105370634', Aliases='', description='uncharacterized LOC105370634', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94144633, end_position_on_the_genomic_accession=94157298, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370635, CurrentID=0, Status='live', Symbol='LOC105370635', Aliases='', description='uncharacterized LOC105370635', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94460117, end_position_on_the_genomic_accession=94462642, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370636, CurrentID=0, Status='live', Symbol='LOC105370636', Aliases='', description='uncharacterized LOC105370636', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94558172, end_position_on_the_genomic_accession=94561340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644109, CurrentID=0, Status='live', Symbol='RPSAP50', Aliases='RPSA_24_1172', description='ribosomal protein SA pseudogene 50', other_designations='', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109982398, end_position_on_the_genomic_accession=109983011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370638, CurrentID=0, Status='live', Symbol='LOC105370638', Aliases='', description='uncharacterized LOC105370638', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95361346, end_position_on_the_genomic_accession=95362280, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370639, CurrentID=0, Status='live', Symbol='LOC105370639', Aliases='', description='uncharacterized LOC105370639', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95389764, end_position_on_the_genomic_accession=95393846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370640, CurrentID=0, Status='live', Symbol='LOC105370640', Aliases='', description='uncharacterized LOC105370640', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95396170, end_position_on_the_genomic_accession=95405078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370641, CurrentID=0, Status='live', Symbol='LOC105370641', Aliases='', description='uncharacterized protein C14orf113', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95463143, end_position_on_the_genomic_accession=95465924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370642, CurrentID=0, Status='live', Symbol='LOC105370642', Aliases='', description='uncharacterized LOC105370642', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370643, CurrentID=0, Status='live', Symbol='LOC105370643', Aliases='', description='uncharacterized LOC105370643', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95587666, end_position_on_the_genomic_accession=95590990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370644, CurrentID=0, Status='live', Symbol='LOC105370644', Aliases='', description='uncharacterized LOC105370644', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=95771576, end_position_on_the_genomic_accession=95774954, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370645, CurrentID=0, Status='live', Symbol='LOC105370645', Aliases='', description='uncharacterized LOC105370645', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96663792, end_position_on_the_genomic_accession=96681736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130450, CurrentID=0, Status='live', Symbol='RPL3P5', Aliases='RPL3_1_338', description='ribosomal protein L3 pseudogene 5', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236407790, end_position_on_the_genomic_accession=236409103, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644119, CurrentID=0, Status='live', Symbol='RPS14P9', Aliases='RPS14_6_722', description='ribosomal protein S14 pseudogene 9', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88265170, end_position_on_the_genomic_accession=88265585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370648, CurrentID=0, Status='live', Symbol='LOC105370648', Aliases='', description='uncharacterized LOC105370648', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97154857, end_position_on_the_genomic_accession=97158822, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370649, CurrentID=0, Status='live', Symbol='LOC105370649', Aliases='', description='uncharacterized LOC105370649', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97660793, end_position_on_the_genomic_accession=97662502, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370650, CurrentID=0, Status='live', Symbol='LOC105370650', Aliases='', description='uncharacterized LOC105370650', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97662737, end_position_on_the_genomic_accession=97688570, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370651, CurrentID=0, Status='live', Symbol='LOC105370651', Aliases='', description='uncharacterized LOC105370651', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97706182, end_position_on_the_genomic_accession=97797717, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370652, CurrentID=0, Status='live', Symbol='LOC105370652', Aliases='', description='uncharacterized LOC105370652', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98136010, end_position_on_the_genomic_accession=98165347, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370653, CurrentID=0, Status='live', Symbol='LOC105370653', Aliases='', description='uncharacterized LOC105370653', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98093306, end_position_on_the_genomic_accession=98116667, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370654, CurrentID=0, Status='live', Symbol='LOC105370654', Aliases='', description='uncharacterized LOC105370654', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98087981, end_position_on_the_genomic_accession=98093197, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101928453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928453, CurrentID=0, Status='live', Symbol='LOC101928453', Aliases='', description='uncharacterized LOC101928453', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14074284, end_position_on_the_genomic_accession=14087881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644128, CurrentID=0, Status='live', Symbol='RPL23AP53', Aliases='RPL23A_20_869', description='ribosomal protein L23a pseudogene 53', other_designations='', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=208344, end_position_on_the_genomic_accession=232318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644129, CurrentID=0, Status='live', Symbol='HMGB1P15', Aliases='', description='high mobility group box 1 pseudogene 15', other_designations='high-mobility group box 1 pseudogene 15', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50931114, end_position_on_the_genomic_accession=50931548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370658, CurrentID=0, Status='live', Symbol='LOC105370658', Aliases='', description='uncharacterized LOC105370658', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=98820553, end_position_on_the_genomic_accession=98823816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644131, CurrentID=0, Status='live', Symbol='CCT8P1', Aliases='CCT8-1P', description='chaperonin containing TCP1, subunit 8 (theta) pseudogene 1', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147203124, end_position_on_the_genomic_accession=147204984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370660, CurrentID=0, Status='live', Symbol='LOC105370660', Aliases='', description='uncharacterized LOC105370660', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99285174, end_position_on_the_genomic_accession=99287474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370661, CurrentID=0, Status='live', Symbol='LOC105370661', Aliases='', description='uncharacterized LOC105370661', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=99536097, end_position_on_the_genomic_accession=99539001, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370662, CurrentID=0, Status='live', Symbol='LOC105370662', Aliases='', description='uncharacterized LOC105370662', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100166926, end_position_on_the_genomic_accession=100172241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644135, CurrentID=0, Status='live', Symbol='LOC644135', Aliases='', description='uncharacterized LOC644135', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136012648, end_position_on_the_genomic_accession=136117169, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105370664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370664, CurrentID=0, Status='live', Symbol='LOC105370664', Aliases='', description='uncharacterized LOC105370664', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100407326, end_position_on_the_genomic_accession=100408399, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370665, CurrentID=0, Status='live', Symbol='LOC105370665', Aliases='', description='uncharacterized LOC105370665', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100573317, end_position_on_the_genomic_accession=100578891, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644138, CurrentID=0, Status='live', Symbol='GOT2P3', Aliases='GOT2L3, GOT2L3P', description='glutamic-oxaloacetic transaminase 2 pseudogene 3', other_designations='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene|glutamic-oxaloacetic transaminase 2-like 3 pseudogene|glutamic-oxaloacetic transaminase 2-like pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9641670, end_position_on_the_genomic_accession=9643211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644139, CurrentID=0, Status='live', Symbol='PIRT', Aliases='', description='phosphoinositide-interacting regulator of transient receptor potential channels', other_designations='phosphoinositide-interacting regulator of TRPV1', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10822475, end_position_on_the_genomic_accession=10838101, orientation='minus', exon_count=2, OMIM=612068, no_hdr0=''),
  105370668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370668, CurrentID=0, Status='live', Symbol='LOC105370668', Aliases='', description='uncharacterized LOC105370668', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100683782, end_position_on_the_genomic_accession=100692321, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370669, CurrentID=0, Status='live', Symbol='LOC105370669', Aliases='', description='uncharacterized LOC105370669', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100709067, end_position_on_the_genomic_accession=100709751, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370670, CurrentID=0, Status='live', Symbol='LOC105370670', Aliases='', description='uncharacterized LOC105370670', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101072830, end_position_on_the_genomic_accession=101077652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370671, CurrentID=0, Status='live', Symbol='LOC105370671', Aliases='', description='uncharacterized LOC105370671', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101357038, end_position_on_the_genomic_accession=101360242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370672, CurrentID=0, Status='live', Symbol='LOC105370672', Aliases='', description='uncharacterized LOC105370672', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101464153, end_position_on_the_genomic_accession=101477652, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  644145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644145, CurrentID=0, Status='live', Symbol='LOC644145', Aliases='', description='exocyst complex component 1 pseudogene', other_designations='exocyst -like pseudogene|hCG2040572', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55820071, end_position_on_the_genomic_accession=55837264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644146, CurrentID=0, Status='live', Symbol='RPS14P8', Aliases='RPS14_5_603', description='ribosomal protein S14 pseudogene 8', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116562521, end_position_on_the_genomic_accession=116563064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370675, CurrentID=0, Status='live', Symbol='LOC105370675', Aliases='', description='uncharacterized LOC105370675', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101566437, end_position_on_the_genomic_accession=101567849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370676, CurrentID=0, Status='live', Symbol='LOC105370676', Aliases='', description='uncharacterized LOC105370676', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101628349, end_position_on_the_genomic_accession=101731306, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370677, CurrentID=0, Status='live', Symbol='LOC105370677', Aliases='', description='uncharacterized LOC105370677', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102203883, end_position_on_the_genomic_accession=102205172, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644150, CurrentID=0, Status='live', Symbol='WIPF3', Aliases='CR16', description='WAS/WASL interacting protein family member 3', other_designations='WAS/WASL interacting protein family, member 3|corticosteroids and regional expression protein 16 homolog', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29806554, end_position_on_the_genomic_accession=29917066, orientation='plus', exon_count=14, OMIM=612432, no_hdr0=''),
  105370679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370679, CurrentID=0, Status='live', Symbol='LOC105370679', Aliases='', description='uncharacterized LOC105370679', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102510112, end_position_on_the_genomic_accession=102514323, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370680, CurrentID=0, Status='live', Symbol='LOC105370680', Aliases='', description='uncharacterized LOC105370680', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102517416, end_position_on_the_genomic_accession=102531291, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644153, CurrentID=0, Status='live', Symbol='DNM1P31', Aliases='DNM1DN4-2, DNM1DN4.2, DNM1DN4@', description='dynamin 1 pseudogene 31', other_designations='DNM1 pseudogene 31|DNM1DN4.2 duplicon', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32387418, end_position_on_the_genomic_accession=32388036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644154, CurrentID=0, Status='live', Symbol='RPL10AP8', Aliases='RPL10A_3_573', description='ribosomal protein L10a pseudogene 8', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72373219, end_position_on_the_genomic_accession=72382393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370683, CurrentID=0, Status='live', Symbol='LINC00677', Aliases='', description='long intergenic non-protein coding RNA 677', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103120849, end_position_on_the_genomic_accession=103123405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370684, CurrentID=0, Status='live', Symbol='LOC105370684', Aliases='', description='uncharacterized LOC105370684', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103139112, end_position_on_the_genomic_accession=103140121, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644157, CurrentID=0, Status='live', Symbol='DND1P1', Aliases='', description='DND microRNA-mediated repression inhibitor 1 pseudogene 1', other_designations='dead end homolog 1 pseudogene 1', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45585864, end_position_on_the_genomic_accession=45587424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370686, CurrentID=0, Status='live', Symbol='LOC105370686', Aliases='', description='uncharacterized LOC105370686', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103294074, end_position_on_the_genomic_accession=103317179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644159, CurrentID=0, Status='live', Symbol='SLC25A1P3', Aliases='', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1 pseudogene 3', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63931150, end_position_on_the_genomic_accession=63932233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644160, CurrentID=0, Status='live', Symbol='ACTG1P19', Aliases='', description='actin gamma 1 pseudogene 19', other_designations='actin, gamma 1 pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100731200, end_position_on_the_genomic_accession=100732362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370689, CurrentID=0, Status='live', Symbol='LOC105370689', Aliases='', description='uncharacterized LOC105370689', other_designations='CTD-2134A5.3', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103872106, end_position_on_the_genomic_accession=103880378, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370690, CurrentID=0, Status='live', Symbol='LOC105370690', Aliases='', description='uncharacterized LOC105370690', other_designations='CTD-2134A5.4', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103869379, end_position_on_the_genomic_accession=103879407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370691, CurrentID=0, Status='live', Symbol='LOC105370691', Aliases='', description='uncharacterized LOC105370691', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104079941, end_position_on_the_genomic_accession=104084033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370692, CurrentID=0, Status='live', Symbol='LOC105370692', Aliases='', description='uncharacterized LOC105370692', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644165, CurrentID=0, Status='live', Symbol='BCRP3', Aliases='BCR3, BCRL3, BCRL6', description='breakpoint cluster region pseudogene 3', other_designations='breakpoint cluster region-like 3|breakpoint cluster region-like 6', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24632915, end_position_on_the_genomic_accession=24653360, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  644166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644166, CurrentID=0, Status='live', Symbol='RPS26P20', Aliases='RPS26_7_290', description='ribosomal protein S26 pseudogene 20', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171374852, end_position_on_the_genomic_accession=171375299, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370695, CurrentID=0, Status='live', Symbol='LOC105370695', Aliases='', description='uncharacterized LOC105370695', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104459797, end_position_on_the_genomic_accession=104460799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644168, CurrentID=0, Status='live', Symbol='DRGX', Aliases='DRG11, PRRXL1', description='dorsal root ganglia homeobox', other_designations='paired related homeobox-like 1|paired-like homeodomain trancription factor DRG11', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49364066, end_position_on_the_genomic_accession=49396016, orientation='minus', exon_count=9, OMIM=606701, no_hdr0=''),
  644169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644169, CurrentID=0, Status='live', Symbol='LOC644169', Aliases='', description='voltage-dependent anion channel 1 pseudogene', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6488178, end_position_on_the_genomic_accession=6489377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370698, CurrentID=0, Status='live', Symbol='LOC105370698', Aliases='', description='uncharacterized LOC105370698', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105549485, end_position_on_the_genomic_accession=105559756, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370699, CurrentID=0, Status='live', Symbol='LOC105370699', Aliases='', description='uncharacterized LOC105370699', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106374914, end_position_on_the_genomic_accession=106376542, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644172, CurrentID=0, Status='live', Symbol='LOC644172', Aliases='', description='mitogen-activated protein kinase 8 interacting protein 1 pseudogene', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45600101, end_position_on_the_genomic_accession=45601862, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370701, CurrentID=0, Status='live', Symbol='LOC105370701', Aliases='', description='uncharacterized LOC105370701', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=106193827, end_position_on_the_genomic_accession=106196813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370702, CurrentID=0, Status='live', Symbol='LOC105370702', Aliases='', description='uncharacterized LOC105370702', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105812420, end_position_on_the_genomic_accession=105831600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370703, CurrentID=0, Status='live', Symbol='LOC105370703', Aliases='', description='uncharacterized LOC105370703', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370704, CurrentID=0, Status='live', Symbol='LOC105370704', Aliases='', description='putative zinc finger protein 355P', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370705, CurrentID=0, Status='live', Symbol='LOC105370705', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23056023, end_position_on_the_genomic_accession=23058476, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370706, CurrentID=0, Status='live', Symbol='LOC105370706', Aliases='', description='uncharacterized protein C14orf56', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=71216320, end_position_on_the_genomic_accession=71228231, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370707, CurrentID=0, Status='live', Symbol='LOC105370707', Aliases='', description='spidroin-1-like', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92508774, end_position_on_the_genomic_accession=92514640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370708, CurrentID=0, Status='live', Symbol='LOC105370708', Aliases='', description='uncharacterized LOC105370708', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104134783, end_position_on_the_genomic_accession=104138629, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370709, CurrentID=0, Status='live', Symbol='LOC105370709', Aliases='', description='uncharacterized LOC105370709', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372241, CurrentID=0, Status='live', Symbol='LOC105372241', Aliases='', description='uncharacterized LOC105372241', other_designations='CTC-265F19.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2621726, end_position_on_the_genomic_accession=2632747, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370711, CurrentID=0, Status='live', Symbol='LOC105370711', Aliases='', description='uncharacterized LOC105370711', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370712, CurrentID=0, Status='live', Symbol='LOC105370712', Aliases='', description='uncharacterized LOC105370712', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370713, CurrentID=0, Status='live', Symbol='LOC105370713', Aliases='', description='uncharacterized LOC105370713', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644186, CurrentID=0, Status='live', Symbol='SYCE3', Aliases='C22orf41, THEG2', description='synaptonemal complex central element protein 3', other_designations='testis highly expressed gene 2 protein|testis highly expressed protein 2', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50551112, end_position_on_the_genomic_accession=50562899, orientation='minus', exon_count=3, OMIM=615775, no_hdr0=''),
  105370715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370715, CurrentID=0, Status='live', Symbol='LOC105370715', Aliases='', description='uncharacterized LOC105370715', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21263162, end_position_on_the_genomic_accession=21270762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370716, CurrentID=0, Status='live', Symbol='LOC105370716', Aliases='', description='uncharacterized LOC105370716', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644189, CurrentID=0, Status='live', Symbol='LOC644189', Aliases='', description='acyl-CoA thioesterase 4 pseudogene', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36421174, end_position_on_the_genomic_accession=36422897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370718, CurrentID=0, Status='live', Symbol='LOC105370718', Aliases='', description='uncharacterized LOC105370718', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644191, CurrentID=0, Status='live', Symbol='RPS26P8', Aliases='RPS26_20_1551', description='ribosomal protein S26 pseudogene 8', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45608540, end_position_on_the_genomic_accession=45609003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644192, CurrentID=0, Status='live', Symbol='NR2F2-AS1', Aliases='', description='NR2F2 antisense RNA 1', other_designations='NR2F2 antisense RNA 1 (non-protein coding)', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96127360, end_position_on_the_genomic_accession=96327361, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100421136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421136, CurrentID=0, Status='live', Symbol='PGBD4P3', Aliases='', description='piggyBac transposable element derived 4 pseudogene 3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148307711, end_position_on_the_genomic_accession=148308250, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370722, CurrentID=0, Status='live', Symbol='LOC105370722', Aliases='', description='olfactory receptor 4S2-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21614052, end_position_on_the_genomic_accession=21614989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370723, CurrentID=0, Status='live', Symbol='LOC105370723', Aliases='', description='uncharacterized LOC105370723', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370724, CurrentID=0, Status='live', Symbol='LOC105370724', Aliases='', description='uncharacterized LOC105370724', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370725, CurrentID=0, Status='live', Symbol='LOC105370725', Aliases='', description='uncharacterized LOC105370725', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401164, CurrentID=0, Status='live', Symbol='LINC01060', Aliases='', description='long intergenic non-protein coding RNA 1060', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188455578, end_position_on_the_genomic_accession=188601908, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370727, CurrentID=0, Status='live', Symbol='LOC105370727', Aliases='', description='uncharacterized LOC105370727', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23381808, end_position_on_the_genomic_accession=23386759, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370728, CurrentID=0, Status='live', Symbol='LOC105370728', Aliases='', description='uncharacterized LOC105370728', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23403862, end_position_on_the_genomic_accession=23407335, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370729, CurrentID=0, Status='live', Symbol='LOC105370729', Aliases='', description='uncharacterized LOC105370729', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23475534, end_position_on_the_genomic_accession=23483289, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644202, CurrentID=0, Status='live', Symbol='RPL12P27', Aliases='RPL12_9_1087', description='ribosomal protein L12 pseudogene 27', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97309530, end_position_on_the_genomic_accession=97310046, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370731, CurrentID=0, Status='live', Symbol='LOC105370731', Aliases='', description='uncharacterized LOC105370731', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23974655, end_position_on_the_genomic_accession=24087770, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370732, CurrentID=0, Status='live', Symbol='LOC105370732', Aliases='', description='uncharacterized LOC105370732', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24142465, end_position_on_the_genomic_accession=24193786, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370733, CurrentID=0, Status='live', Symbol='LOC105370733', Aliases='', description='uncharacterized LOC105370733', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24101609, end_position_on_the_genomic_accession=24150495, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370734, CurrentID=0, Status='live', Symbol='LOC105370734', Aliases='', description='uncharacterized LOC105370734', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24098027, end_position_on_the_genomic_accession=24101511, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370735, CurrentID=0, Status='live', Symbol='LOC105370735', Aliases='', description='uncharacterized LOC105370735', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24412263, end_position_on_the_genomic_accession=24425453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370736, CurrentID=0, Status='live', Symbol='LOC105370736', Aliases='', description='uncharacterized LOC105370736', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370737, CurrentID=0, Status='live', Symbol='LOC105370737', Aliases='', description='uncharacterized LOC105370737', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25500934, end_position_on_the_genomic_accession=25578792, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370738, CurrentID=0, Status='live', Symbol='LOC105370738', Aliases='', description='uncharacterized LOC105370738', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25865284, end_position_on_the_genomic_accession=25877228, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370739, CurrentID=0, Status='live', Symbol='LOC105370739', Aliases='', description='uncharacterized LOC105370739', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26089631, end_position_on_the_genomic_accession=26142998, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370740, CurrentID=0, Status='live', Symbol='LOC105370740', Aliases='', description='uncharacterized LOC105370740', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26403750, end_position_on_the_genomic_accession=26477905, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  644213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644213, CurrentID=0, Status='live', Symbol='GAPDHP32', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 32', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119533548, end_position_on_the_genomic_accession=119534821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644214, CurrentID=0, Status='live', Symbol='PHBP11', Aliases='', description='prohibitin pseudogene 11', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223856582, end_position_on_the_genomic_accession=223857387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644215, CurrentID=0, Status='live', Symbol='LOC644215', Aliases='', description='uncharacterized LOC644215', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97401237, end_position_on_the_genomic_accession=97426151, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370744, CurrentID=0, Status='live', Symbol='LOC105370744', Aliases='', description='uncharacterized LOC105370744', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=29760037, end_position_on_the_genomic_accession=29762429, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370745, CurrentID=0, Status='live', Symbol='LOC105370745', Aliases='', description='uncharacterized LOC105370745', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30004972, end_position_on_the_genomic_accession=30033557, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644218, CurrentID=0, Status='live', Symbol='TRPC6P', Aliases='TRPC6L', description='transient receptor potential cation channel, subfamily C, member 6 pseudogene', other_designations='', map_location='Xq28;Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57171866, end_position_on_the_genomic_accession=57172767, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370747, CurrentID=0, Status='live', Symbol='LOC105370747', Aliases='', description='uncharacterized LOC105370747', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30180036, end_position_on_the_genomic_accession=30182290, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644220, CurrentID=0, Status='live', Symbol='ACTBP11', Aliases='', description='actin, beta pseudogene 11', other_designations='actin, gamma 1 pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223863145, end_position_on_the_genomic_accession=223864935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370749, CurrentID=0, Status='live', Symbol='LOC105370749', Aliases='', description='uncharacterized LOC105370749', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370750, CurrentID=0, Status='live', Symbol='LOC105370750', Aliases='', description='uncharacterized LOC105370750', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30534178, end_position_on_the_genomic_accession=30538421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644223, CurrentID=0, Status='live', Symbol='RPL9P4', Aliases='RPL9_9_1348', description='ribosomal protein L9 pseudogene 4', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=31462383, end_position_on_the_genomic_accession=31463422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370752, CurrentID=0, Status='live', Symbol='LOC105370752', Aliases='', description='uncharacterized LOC105370752', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31026206, end_position_on_the_genomic_accession=31037481, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370753, CurrentID=0, Status='live', Symbol='LOC105370753', Aliases='', description='uncharacterized LOC105370753', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31507847, end_position_on_the_genomic_accession=31526340, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370754, CurrentID=0, Status='live', Symbol='LOC105370754', Aliases='', description='uncharacterized LOC105370754', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32157791, end_position_on_the_genomic_accession=32164852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370755, CurrentID=0, Status='live', Symbol='LOC105370755', Aliases='', description='uncharacterized LOC105370755', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370756, CurrentID=0, Status='live', Symbol='LOC105370756', Aliases='', description='uncharacterized LOC105370756', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32686862, end_position_on_the_genomic_accession=32691697, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370757, CurrentID=0, Status='live', Symbol='LOC105370757', Aliases='', description='uncharacterized LOC105370757', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32655730, end_position_on_the_genomic_accession=32673137, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370759, CurrentID=0, Status='live', Symbol='LOC105370759', Aliases='', description='uncharacterized LOC105370759', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33194897, end_position_on_the_genomic_accession=33211952, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370760, CurrentID=0, Status='live', Symbol='LOC105370760', Aliases='', description='uncharacterized LOC105370760', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644233, CurrentID=0, Status='live', Symbol='AURKBPS1', Aliases='AurBps1', description='aurora kinase B pseudogene 1', other_designations='', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=108130962, end_position_on_the_genomic_accession=108132194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370762, CurrentID=0, Status='live', Symbol='LOC105370762', Aliases='', description='uncharacterized LOC105370762', other_designations='CTD-2125J1.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34651684, end_position_on_the_genomic_accession=34671056, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100127883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127883, CurrentID=0, Status='live', Symbol='IBTKP1', Aliases='', description='inhibitor of Bruton tyrosine kinase pseudogene 1', other_designations='IBTK pseudogene 1|inhibitor of Bruton agammaglobulinemia tyrosine kinase pseudogene 1', map_location='18q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=44039477, end_position_on_the_genomic_accession=44040339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644236, CurrentID=0, Status='live', Symbol='RPL7AP43', Aliases='RPL7A_17_815', description='ribosomal protein L7a pseudogene 43', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77115349, end_position_on_the_genomic_accession=77116208, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644237, CurrentID=0, Status='live', Symbol='GAPDHP33', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 33', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119595965, end_position_on_the_genomic_accession=119597250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370766, CurrentID=0, Status='live', Symbol='LOC105370766', Aliases='', description='uncharacterized LOC105370766', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35919920, end_position_on_the_genomic_accession=35940910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370767, CurrentID=0, Status='live', Symbol='LOC105370767', Aliases='', description='uncharacterized LOC105370767', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36310255, end_position_on_the_genomic_accession=36361476, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100127888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127888, CurrentID=0, Status='live', Symbol='SLCO4A1-AS1', Aliases='', description='SLCO4A1 antisense RNA 1', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62663027, end_position_on_the_genomic_accession=62666621, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100127889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127889, CurrentID=0, Status='live', Symbol='C10orf131', Aliases='bA690P14.3', description='chromosome 10 open reading frame 131', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95907965, end_position_on_the_genomic_accession=95938658, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  644242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644242, CurrentID=0, Status='live', Symbol='LINC00622', Aliases='', description='long intergenic non-protein coding RNA 622', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119597702, end_position_on_the_genomic_accession=119599291, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370771, CurrentID=0, Status='live', Symbol='LOC105370771', Aliases='', description='uncharacterized LOC105370771', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=36668898, end_position_on_the_genomic_accession=36674358, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370772, CurrentID=0, Status='live', Symbol='LOC105370772', Aliases='', description='uncharacterized LOC105370772', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=37364193, end_position_on_the_genomic_accession=37426756, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100127893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127893, CurrentID=0, Status='live', Symbol='RPL7P24', Aliases='RPL7_8_586', description='ribosomal protein L7 pseudogene 24', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80500312, end_position_on_the_genomic_accession=80501468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644246, CurrentID=0, Status='live', Symbol='KANSL1-AS1', Aliases='', description='KANSL1 antisense RNA 1', other_designations='KANSL1 antisense RNA 1 (non-protein coding)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46193573, end_position_on_the_genomic_accession=46196723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370775, CurrentID=0, Status='live', Symbol='LOC105370775', Aliases='', description='uncharacterized LOC105370775', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38671775, end_position_on_the_genomic_accession=38689146, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644248, CurrentID=0, Status='live', Symbol='SNCA-AS1', Aliases='', description='SNCA antisense RNA 1', other_designations='', map_location='4q22.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89836401, end_position_on_the_genomic_accession=89841991, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370777, CurrentID=0, Status='live', Symbol='LOC105370777', Aliases='', description='uncharacterized LOC105370777', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=38864806, end_position_on_the_genomic_accession=38881739, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370778, CurrentID=0, Status='live', Symbol='LOC105370778', Aliases='', description='uncharacterized LOC105370778', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39300168, end_position_on_the_genomic_accession=39310815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370779, CurrentID=0, Status='live', Symbol='LOC105370779', Aliases='', description='uncharacterized LOC105370779', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39213776, end_position_on_the_genomic_accession=39231074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370780, CurrentID=0, Status='live', Symbol='LOC105370780', Aliases='', description='uncharacterized LOC105370780', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39167740, end_position_on_the_genomic_accession=39180457, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  644253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644253, CurrentID=0, Status='live', Symbol='MAPK8IPP', Aliases='PRKM8IPP', description='mitogen-activated protein kinase 8 interacting protein, pseudogene', other_designations='PRKM8 interacting protein pseudogene', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46243606, end_position_on_the_genomic_accession=46245044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127902, CurrentID=0, Status='live', Symbol='BMI1P1', Aliases='', description='BMI1 proto-oncogene, polycomb ring finger pseudogene 1', other_designations='BMI1 polycomb ring finger oncogene pseudogene', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=67791910, end_position_on_the_genomic_accession=67793049, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260293, CurrentID=0, Status='live', Symbol='CYP4X1', Aliases='CYPIVX1', description='cytochrome P450 family 4 subfamily X member 1', other_designations='cytochrome P450, family 4, subfamily X, polypeptide 1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46961364, end_position_on_the_genomic_accession=47050751, orientation='plus', exon_count=14, OMIM=614999, no_hdr0=''),
  644256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644256, CurrentID=0, Status='live', Symbol='DND1P2', Aliases='', description='DND microRNA-mediated repression inhibitor 1 pseudogene 2', other_designations='dead end homolog 1 pseudogene 2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46259053, end_position_on_the_genomic_accession=46260613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100532735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100532735, CurrentID=0, Status='live', Symbol='INO80B-WBP1', Aliases='', description='INO80B-WBP1 readthrough (NMD candidate)', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74455023, end_position_on_the_genomic_accession=74460891, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100127906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127906, CurrentID=0, Status='live', Symbol='PQLC1P2', Aliases='', description='PQ loop repeat containing 1 pseudogene 2', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82691736, end_position_on_the_genomic_accession=82692506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127907, CurrentID=0, Status='live', Symbol='HNRNPCP7', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 7', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64500825, end_position_on_the_genomic_accession=64501729, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370788, CurrentID=0, Status='live', Symbol='LOC105370788', Aliases='', description='uncharacterized LOC105370788', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653510, CurrentID=0, Status='live', Symbol='CBWD4P', Aliases='', description='COBW domain containing 4 pseudogene', other_designations='COBW domain containing 1 pseudogene', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65287480, end_position_on_the_genomic_accession=65321282, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370790, CurrentID=0, Status='live', Symbol='LOC105370790', Aliases='', description='uncharacterized LOC105370790', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41479207, end_position_on_the_genomic_accession=41486369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127911, CurrentID=0, Status='live', Symbol='ST13P12', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 12', other_designations='', map_location='5q14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=82968890, end_position_on_the_genomic_accession=82977147, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370792, CurrentID=0, Status='live', Symbol='LOC105370792', Aliases='', description='uncharacterized LOC105370792', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41892762, end_position_on_the_genomic_accession=41895946, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370793, CurrentID=0, Status='live', Symbol='LOC105370793', Aliases='', description='uncharacterized LOC105370793', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42042827, end_position_on_the_genomic_accession=42044590, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370794, CurrentID=0, Status='live', Symbol='LOC105370794', Aliases='', description='uncharacterized LOC105370794', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42384560, end_position_on_the_genomic_accession=42386160, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370795, CurrentID=0, Status='live', Symbol='LOC105370795', Aliases='', description='uncharacterized LOC105370795', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42208380, end_position_on_the_genomic_accession=42231192, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370796, CurrentID=0, Status='live', Symbol='LOC105370796', Aliases='', description='uncharacterized LOC105370796', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44243911, end_position_on_the_genomic_accession=44286336, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370797, CurrentID=0, Status='live', Symbol='LOC105370797', Aliases='', description='uncharacterized LOC105370797', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44423037, end_position_on_the_genomic_accession=44426317, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370798, CurrentID=0, Status='live', Symbol='LOC105370798', Aliases='', description='uncharacterized LOC105370798', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44854559, end_position_on_the_genomic_accession=44856774, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370799, CurrentID=0, Status='live', Symbol='LOC105370799', Aliases='', description='uncharacterized LOC105370799', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45174447, end_position_on_the_genomic_accession=45188930, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370800, CurrentID=0, Status='live', Symbol='LOC105370800', Aliases='', description='uncharacterized LOC105370800', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45449695, end_position_on_the_genomic_accession=45454415, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370801, CurrentID=0, Status='live', Symbol='LOC105370801', Aliases='', description='uncharacterized LOC105370801', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45664235, end_position_on_the_genomic_accession=45677733, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127922, CurrentID=0, Status='live', Symbol='SUMO2P6', Aliases='', description='SUMO2 pseudogene 6', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 6', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174561775, end_position_on_the_genomic_accession=174562062, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370803, CurrentID=0, Status='live', Symbol='LOC105370803', Aliases='', description='uncharacterized LOC105370803', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=46607794, end_position_on_the_genomic_accession=46638995, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100127924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127924, CurrentID=0, Status='live', Symbol='RPS15AP14', Aliases='RPS15A_4_297', description='ribosomal protein S15a pseudogene 14', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=175184185, end_position_on_the_genomic_accession=175184655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644277, CurrentID=0, Status='live', Symbol='LOC644277', Aliases='', description='uncharacterized LOC644277', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111445886, end_position_on_the_genomic_accession=111456605, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105370806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370806, CurrentID=0, Status='live', Symbol='LOC105370806', Aliases='', description='uncharacterized LOC105370806', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48227184, end_position_on_the_genomic_accession=48230436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370807, CurrentID=0, Status='live', Symbol='LOC105370807', Aliases='', description='uncharacterized LOC105370807', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48188111, end_position_on_the_genomic_accession=48191796, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370808, CurrentID=0, Status='live', Symbol='LOC105370808', Aliases='', description='uncharacterized LOC105370808', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48526255, end_position_on_the_genomic_accession=48532376, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370809, CurrentID=0, Status='live', Symbol='LOC105370809', Aliases='', description='uncharacterized LOC105370809', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48645951, end_position_on_the_genomic_accession=48652016, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370810, CurrentID=0, Status='live', Symbol='LOC105370810', Aliases='', description='uncharacterized LOC105370810', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=48938242, end_position_on_the_genomic_accession=48947379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370811, CurrentID=0, Status='live', Symbol='LOC105370811', Aliases='', description='uncharacterized LOC105370811', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49490017, end_position_on_the_genomic_accession=49557758, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370812, CurrentID=0, Status='live', Symbol='LOC105370812', Aliases='', description='uncharacterized LOC105370812', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49812605, end_position_on_the_genomic_accession=49854022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644285, CurrentID=0, Status='live', Symbol='LOC644285', Aliases='', description='uncharacterized LOC644285', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87318415, end_position_on_the_genomic_accession=87323628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370814, CurrentID=0, Status='live', Symbol='LOC105370814', Aliases='', description='uncharacterized LOC105370814', other_designations='CTD-2378E12.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370815, CurrentID=0, Status='live', Symbol='LOC105370815', Aliases='', description='serine/arginine repetitive matrix protein 3-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51094593, end_position_on_the_genomic_accession=51096596, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370816, CurrentID=0, Status='live', Symbol='LOC105370816', Aliases='', description='uncharacterized LOC105370816', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51624309, end_position_on_the_genomic_accession=51655706, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370817, CurrentID=0, Status='live', Symbol='LOC105370817', Aliases='', description='uncharacterized LOC105370817', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51693346, end_position_on_the_genomic_accession=51715875, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105370818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370818, CurrentID=0, Status='live', Symbol='LOC105370818', Aliases='', description='uncharacterized LOC105370818', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51915877, end_position_on_the_genomic_accession=51947295, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370819, CurrentID=0, Status='live', Symbol='LOC105370819', Aliases='', description='uncharacterized LOC105370819', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52292024, end_position_on_the_genomic_accession=52304500, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100127940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127940, CurrentID=0, Status='live', Symbol='LOC100127940', Aliases='', description='uncharacterized LOC100127940', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=323479, end_position_on_the_genomic_accession=324640, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370821, CurrentID=0, Status='live', Symbol='LOC105370821', Aliases='', description='uncharacterized LOC105370821', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52679430, end_position_on_the_genomic_accession=52690175, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105370822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370822, CurrentID=0, Status='live', Symbol='LOC105370822', Aliases='', description='uncharacterized LOC105370822', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52804516, end_position_on_the_genomic_accession=53194054, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100421153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421153, CurrentID=0, Status='live', Symbol='DLEC1P1', Aliases='', description='deleted in lung and esophageal cancer 1 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38325237, end_position_on_the_genomic_accession=38329070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370824, CurrentID=0, Status='live', Symbol='LOC105370824', Aliases='', description='uncharacterized LOC105370824', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52790462, end_position_on_the_genomic_accession=52795850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644297, CurrentID=0, Status='live', Symbol='FAM215B', Aliases='ARL17A-IT1', description='family with sequence similarity 215 member B (non-protein coding)', other_designations='ARL17A intronic transcript 1 (non-protein coding)|family with sequence similarity 215, member B (non-protein coding)', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130456, CurrentID=0, Status='live', Symbol='LOC100130456', Aliases='', description='uncharacterized LOC100130456', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=1055992, end_position_on_the_genomic_accession=1056381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370827, CurrentID=0, Status='live', Symbol='LOC105370827', Aliases='', description='uncharacterized LOC105370827', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53902599, end_position_on_the_genomic_accession=53910971, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370828, CurrentID=0, Status='live', Symbol='LOC105370828', Aliases='', description='uncharacterized LOC105370828', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=53837628, end_position_on_the_genomic_accession=53916058, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100127949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127949, CurrentID=0, Status='live', Symbol='RPL15P16', Aliases='RPL15_9_1154', description='ribosomal protein L15 pseudogene 16', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72940498, end_position_on_the_genomic_accession=72941112, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127950, CurrentID=0, Status='live', Symbol='HECW1-IT1', Aliases='', description='HECW1 intronic transcript 1', other_designations='HECW1 intronic transcript 1 (non-protein coding)', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43117964, end_position_on_the_genomic_accession=43138343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100127951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127951, CurrentID=0, Status='live', Symbol='LOC100127951', Aliases='', description='ISPF6484', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=71760968, end_position_on_the_genomic_accession=71761234, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100127952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127952, CurrentID=0, Status='live', Symbol='ATF4P4', Aliases='ATF4B', description='activating transcription factor 4 pseudogene 4', other_designations='activating transcription factor 4 (tax-responsive enhancer element B67) pseudogene|activating transcription factor 4B', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113790038, end_position_on_the_genomic_accession=113791453, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370833, CurrentID=0, Status='live', Symbol='LOC105370833', Aliases='', description='uncharacterized LOC105370833', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56884855, end_position_on_the_genomic_accession=56888219, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370834, CurrentID=0, Status='live', Symbol='LOC105370834', Aliases='', description='uncharacterized LOC105370834', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57732421, end_position_on_the_genomic_accession=57775324, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370835, CurrentID=0, Status='live', Symbol='LOC105370835', Aliases='', description='uncharacterized LOC105370835', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58264735, end_position_on_the_genomic_accession=58278968, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370836, CurrentID=0, Status='live', Symbol='LOC105370836', Aliases='', description='uncharacterized LOC105370836', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=58212813, end_position_on_the_genomic_accession=58244327, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644309, CurrentID=0, Status='live', Symbol='RPS20P19', Aliases='RPS20_7_863', description='ribosomal protein S20 pseudogene 19', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152590546, end_position_on_the_genomic_accession=152591064, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100127958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127958, CurrentID=0, Status='live', Symbol='AGGF1P9', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 9', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 9', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35347321, end_position_on_the_genomic_accession=35348297, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370839, CurrentID=0, Status='live', Symbol='LOC105370839', Aliases='', description='uncharacterized LOC105370839', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60166939, end_position_on_the_genomic_accession=60255832, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100127960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127960, CurrentID=0, Status='live', Symbol='BCAS2P1', Aliases='', description='breast carcinoma amplified sequence 2 pseudogene 1', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=83132038, end_position_on_the_genomic_accession=83134358, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421156, CurrentID=0, Status='live', Symbol='DPPA4P1', Aliases='', description='developmental pluripotency associated 4 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66326249, end_position_on_the_genomic_accession=66326783, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644314, CurrentID=0, Status='live', Symbol='MT1IP', Aliases='MT1, MT1I, MTE', description='metallothionein 1I, pseudogene', other_designations='metallothionein 1I (pseudogene)|metallothionein E', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56676116, end_position_on_the_genomic_accession=56677877, orientation='plus', exon_count=4, OMIM=156355, no_hdr0=''),
  100127963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127963, CurrentID=0, Status='live', Symbol='RPS3AP48', Aliases='RPS3A_22_1469', description='ribosomal protein S3a pseudogene 48', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5270150, end_position_on_the_genomic_accession=5270998, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644316, CurrentID=0, Status='live', Symbol='FLJ43315', Aliases='', description='asparagine synthetase pseudogene', other_designations='asparagine synthetase (glutamine-hydrolyzing) pseudogene', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64016950, end_position_on_the_genomic_accession=64061612, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105370845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370845, CurrentID=0, Status='live', Symbol='LOC105370845', Aliases='', description='uncharacterized LOC105370845', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60655230, end_position_on_the_genomic_accession=60665460, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370846, CurrentID=0, Status='live', Symbol='LOC105370846', Aliases='', description='uncharacterized LOC105370846', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60632230, end_position_on_the_genomic_accession=60648209, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100127967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127967, CurrentID=0, Status='live', Symbol='KRT73-AS1', Aliases='', description='KRT73 antisense RNA 1', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52609851, end_position_on_the_genomic_accession=52615305, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370848, CurrentID=0, Status='live', Symbol='LOC105370848', Aliases='', description='uncharacterized LOC105370848', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=61834700, end_position_on_the_genomic_accession=61880589, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100127969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127969, CurrentID=0, Status='live', Symbol='VN1R96P', Aliases='', description='vomeronasal 1 receptor 96 pseudogene', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40121323, end_position_on_the_genomic_accession=40122230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370850, CurrentID=0, Status='live', Symbol='LOC105370850', Aliases='', description='uncharacterized LOC105370850', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62271343, end_position_on_the_genomic_accession=62302390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370851, CurrentID=0, Status='live', Symbol='LOC105370851', Aliases='', description='uncharacterized LOC105370851', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62267158, end_position_on_the_genomic_accession=62268411, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100127972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127972, CurrentID=0, Status='live', Symbol='LOC100127972', Aliases='', description='uncharacterized LOC100127972', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=40349750, end_position_on_the_genomic_accession=40350818, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644325, CurrentID=0, Status='live', Symbol='ADAM20P3', Aliases='', description='ADAM metallopeptidase domain 20 pseudogene 3', other_designations='disintegrin and metalloprotease domain protein pseudogene', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187746134, end_position_on_the_genomic_accession=187748205, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100127974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127974, CurrentID=0, Status='live', Symbol='LOC100127974', Aliases='', description='uncharacterized LOC100127974', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6569434, end_position_on_the_genomic_accession=6582659, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370855, CurrentID=0, Status='live', Symbol='LOC105370855', Aliases='', description='uncharacterized LOC105370855', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62657907, end_position_on_the_genomic_accession=62688118, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100127976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127976, CurrentID=0, Status='live', Symbol='VTI1BP3', Aliases='', description='vesicle transport through interaction with t-SNAREs 1B pseudogene 3', other_designations='vesicle transport through interaction with t-SNAREs homolog 1B pseudogene 3', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53131350, end_position_on_the_genomic_accession=53131987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370857, CurrentID=0, Status='live', Symbol='LOC105370857', Aliases='', description='uncharacterized LOC105370857', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644330, CurrentID=0, Status='live', Symbol='TPM3P5', Aliases='', description='tropomyosin 3 pseudogene 5', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41506068, end_position_on_the_genomic_accession=41508191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100127979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127979, CurrentID=0, Status='live', Symbol='FUNDC2P4', Aliases='', description='FUN14 domain containing 2 pseudogene 4', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39155536, end_position_on_the_genomic_accession=39156096, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370860, CurrentID=0, Status='live', Symbol='LOC105370860', Aliases='', description='uncharacterized LOC105370860', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64063165, end_position_on_the_genomic_accession=64072025, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644333, CurrentID=0, Status='live', Symbol='ATP5F1P3', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 3', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152748972, end_position_on_the_genomic_accession=152749735, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370862, CurrentID=0, Status='live', Symbol='LOC105370862', Aliases='', description='uncharacterized LOC105370862', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65083036, end_position_on_the_genomic_accession=65089961, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100127983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127983, CurrentID=0, Status='live', Symbol='C8orf88', Aliases='', description='chromosome 8 open reading frame 88', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90958478, end_position_on_the_genomic_accession=90985257, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105370864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370864, CurrentID=0, Status='live', Symbol='LOC105370864', Aliases='', description='uncharacterized LOC105370864', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65287009, end_position_on_the_genomic_accession=65293493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370865, CurrentID=0, Status='live', Symbol='LOC105370865', Aliases='', description='uncharacterized LOC105370865', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644338, CurrentID=0, Status='live', Symbol='RPL37AP8', Aliases='RPL37A_3_1174', description='ribosomal protein L37a pseudogene 8', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111889143, end_position_on_the_genomic_accession=111889827, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370867, CurrentID=0, Status='live', Symbol='LOC105370867', Aliases='', description='uncharacterized LOC105370867', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66019391, end_position_on_the_genomic_accession=66020589, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370868, CurrentID=0, Status='live', Symbol='LOC105370868', Aliases='', description='uncharacterized LOC105370868', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66278502, end_position_on_the_genomic_accession=66293552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127989, CurrentID=0, Status='live', Symbol='MED15P4', Aliases='', description='mediator complex subunit 15 pseudogene 4', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131534344, end_position_on_the_genomic_accession=131536989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370870, CurrentID=0, Status='live', Symbol='LOC105370870', Aliases='', description='uncharacterized LOC105370870', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67830198, end_position_on_the_genomic_accession=67832597, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101928489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928489, CurrentID=0, Status='live', Symbol='LOC101928489', Aliases='', description='uncharacterized LOC101928489', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73693903, end_position_on_the_genomic_accession=73696131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370872, CurrentID=0, Status='live', Symbol='LOC105370872', Aliases='', description='uncharacterized LOC105370872', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=68271011, end_position_on_the_genomic_accession=68277926, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127993, CurrentID=0, Status='live', Symbol='RPL15P5', Aliases='RPL15_1_277', description='ribosomal protein L15 pseudogene 5', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138278538, end_position_on_the_genomic_accession=138279451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370874, CurrentID=0, Status='live', Symbol='LOC105370874', Aliases='', description='uncharacterized LOC105370874', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69071395, end_position_on_the_genomic_accession=69073189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100127995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127995, CurrentID=0, Status='live', Symbol='RPL34P7', Aliases='RPL34_1_168', description='ribosomal protein L34 pseudogene 7', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226188868, end_position_on_the_genomic_accession=226189221, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370876, CurrentID=0, Status='live', Symbol='LOC105370876', Aliases='', description='uncharacterized LOC105370876', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69459530, end_position_on_the_genomic_accession=69462776, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105370877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370877, CurrentID=0, Status='live', Symbol='LOC105370877', Aliases='', description='uncharacterized LOC105370877', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70079656, end_position_on_the_genomic_accession=70084024, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370878, CurrentID=0, Status='live', Symbol='LOC105370878', Aliases='', description='uncharacterized LOC105370878', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70296244, end_position_on_the_genomic_accession=70297922, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100127999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100127999, CurrentID=0, Status='live', Symbol='RPL18AP11', Aliases='RPL18A_5_961', description='ribosomal protein L18a pseudogene 11', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=8713276, end_position_on_the_genomic_accession=8713894, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370880, CurrentID=0, Status='live', Symbol='LOC105370880', Aliases='', description='uncharacterized LOC105370880', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70503907, end_position_on_the_genomic_accession=70505995, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  120065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120065, CurrentID=0, Status='live', Symbol='OR5P2', Aliases='JCG3, JCG4', description='olfactory receptor family 5 subfamily P member 2', other_designations='olfactory receptor OR11-93|olfactory receptor, family 5, subfamily P, member 2|olfactory receptor-like protein JCG3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7795974, end_position_on_the_genomic_accession=7796942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  120066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120066, CurrentID=0, Status='live', Symbol='OR5P3', Aliases='JCG1', description='olfactory receptor family 5 subfamily P member 3', other_designations='olfactory receptor OR11-94|olfactory receptor, family 5, subfamily P, member 3|olfactory receptor-like protein JCG1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7825037, end_position_on_the_genomic_accession=7825972, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370883, CurrentID=0, Status='live', Symbol='LOC105370883', Aliases='', description='uncharacterized LOC105370883', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71290573, end_position_on_the_genomic_accession=71294862, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128004, CurrentID=0, Status='live', Symbol='RPL32P21', Aliases='RPL32_13_977', description='ribosomal protein L32 pseudogene 21', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37046835, end_position_on_the_genomic_accession=37047242, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644357, CurrentID=0, Status='live', Symbol='EEF1DP6', Aliases='', description='eukaryotic translation elongation factor 1 delta pseudogene 6', other_designations='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein) pseudogene', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4175463, end_position_on_the_genomic_accession=4175899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128006, CurrentID=0, Status='live', Symbol='LOC100128006', Aliases='', description='uncharacterized LOC100128006', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12760134, end_position_on_the_genomic_accession=12790284, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  120071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120071, CurrentID=0, Status='live', Symbol='GYLTL1B', Aliases='LARGE2, PP5656', description='glycosyltransferase-like 1B', other_designations='like-glycosyltransferase 2|ortholog of mouse glycosyltransferase-like 1B', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45920974, end_position_on_the_genomic_accession=45929096, orientation='plus', exon_count=17, OMIM=609709, no_hdr0=''),
  105370888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370888, CurrentID=0, Status='live', Symbol='LOC105370888', Aliases='', description='uncharacterized LOC105370888', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72407765, end_position_on_the_genomic_accession=72421603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370889, CurrentID=0, Status='live', Symbol='LOC105370889', Aliases='', description='uncharacterized LOC105370889', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72464587, end_position_on_the_genomic_accession=72471996, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128010, CurrentID=0, Status='live', Symbol='SUMO2P17', Aliases='', description='SUMO2 pseudogene 17', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene|small ubiquitin-like modifier 2 pseudogene', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48874908, end_position_on_the_genomic_accession=48875190, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128011, CurrentID=0, Status='live', Symbol='FAR1P1', Aliases='', description='fatty acyl-CoA reductase 1 pseudogene 1', other_designations='fatty acyl CoA reductase 1 pseudogene 1', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90170863, end_position_on_the_genomic_accession=90171907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128012, CurrentID=0, Status='live', Symbol='ISCA1P4', Aliases='', description='iron-sulfur cluster assembly 1 pseudogene 4', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 4', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40948487, end_position_on_the_genomic_accession=40949310, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370893, CurrentID=0, Status='live', Symbol='LOC105370893', Aliases='', description='uncharacterized LOC105370893', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74232124, end_position_on_the_genomic_accession=74236224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105370894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370894, CurrentID=0, Status='live', Symbol='LOC105370894', Aliases='', description='uncharacterized LOC105370894', other_designations='CTD-2311M21.5', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74381530, end_position_on_the_genomic_accession=74393827, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128015, CurrentID=0, Status='live', Symbol='HMGN1P26', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 26', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 26', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62361226, end_position_on_the_genomic_accession=62362502, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128016, CurrentID=0, Status='live', Symbol='WHSC1L2P', Aliases='HsT36289, NSD4P, WHSC1L1P', description='Wolf-Hirschhorn syndrome candidate 1-like 2, pseudogene', other_designations='Wolf-Hirschhorn syndrome candidate 1-like 1 pseudogene', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43718314, end_position_on_the_genomic_accession=43720436, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128017, CurrentID=0, Status='live', Symbol='TULP3P1', Aliases='', description='tubby like protein 3 pseudogene 1', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=96946595, end_position_on_the_genomic_accession=96948927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370898, CurrentID=0, Status='live', Symbol='LOC105370898', Aliases='', description='uncharacterized LOC105370898', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=74907217, end_position_on_the_genomic_accession=74911178, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370899, CurrentID=0, Status='live', Symbol='LOC105370899', Aliases='', description='uncharacterized LOC105370899', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128020, CurrentID=0, Status='live', Symbol='ACTR3P3', Aliases='', description='ACTR3 pseudogene 3', other_designations='ARP3 actin-related protein 3 homolog B pseudogene|ARP3 actin-related protein 3 homolog pseudogene 3', map_location='3q12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101163094, end_position_on_the_genomic_accession=101164039, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370901, CurrentID=0, Status='live', Symbol='LOC105370901', Aliases='', description='uncharacterized LOC105370901', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75451494, end_position_on_the_genomic_accession=75454466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370902, CurrentID=0, Status='live', Symbol='LOC105370902', Aliases='', description='uncharacterized LOC105370902', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75470763, end_position_on_the_genomic_accession=75503039, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128023, CurrentID=0, Status='live', Symbol='DPPA2P3', Aliases='', description='developmental pluripotency associated 2 pseudogene 3', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193993093, end_position_on_the_genomic_accession=193994238, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370904, CurrentID=0, Status='live', Symbol='LOC105370904', Aliases='', description='uncharacterized LOC105370904', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76972904, end_position_on_the_genomic_accession=76983228, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128025, CurrentID=0, Status='live', Symbol='WWTR1-AS1', Aliases='', description='WWTR1 antisense RNA 1', other_designations='WWTR1 antisense RNA 1 (non-protein coding)', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149658210, end_position_on_the_genomic_accession=149661364, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644378, CurrentID=0, Status='live', Symbol='GCNT6', Aliases='bA421M1.3', description='glucosaminyl (N-acetyl) transferase 6', other_designations='', map_location='6p24.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=10633993, end_position_on_the_genomic_accession=10647501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370907, CurrentID=0, Status='live', Symbol='LOC105370907', Aliases='', description='uncharacterized LOC105370907', other_designations='CTD-2071N1.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77787152, end_position_on_the_genomic_accession=77789400, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370908, CurrentID=0, Status='live', Symbol='LOC105370908', Aliases='', description='uncharacterized LOC105370908', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77957443, end_position_on_the_genomic_accession=77964006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128029, CurrentID=0, Status='live', Symbol='KNOP1P3', Aliases='', description='lysine-rich nucleolar protein 1 pseudogene 3', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51510142, end_position_on_the_genomic_accession=51511480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128030, CurrentID=0, Status='live', Symbol='RPL10P11', Aliases='RPL10_4_817', description='ribosomal protein L10 pseudogene 11', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80096521, end_position_on_the_genomic_accession=80097262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370911, CurrentID=0, Status='live', Symbol='LOC105370911', Aliases='', description='uncharacterized LOC105370911', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78252626, end_position_on_the_genomic_accession=78264144, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644384, CurrentID=0, Status='live', Symbol='RPL23AP71', Aliases='RPL23A_33_1349', description='ribosomal protein L23a pseudogene 71', other_designations='', map_location='14q13.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=34485978, end_position_on_the_genomic_accession=34486503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128033, CurrentID=0, Status='live', Symbol='PIGPP3', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class P pseudogene 3', other_designations='phosphatidylinositol glycan anchor biosynthesis, class P pseudogene 3', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33965186, end_position_on_the_genomic_accession=33965573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370914, CurrentID=0, Status='live', Symbol='LOC105370914', Aliases='', description='uncharacterized LOC105370914', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78755163, end_position_on_the_genomic_accession=78757709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644387, CurrentID=0, Status='live', Symbol='LOC644387', Aliases='', description='myelin protein zero-like 1 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64734615, end_position_on_the_genomic_accession=64736933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644388, CurrentID=0, Status='live', Symbol='RPS2P47', Aliases='RPS2_25_1553', description='ribosomal protein S2 pseudogene 47', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47159154, end_position_on_the_genomic_accession=47159998, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370917, CurrentID=0, Status='live', Symbol='LOC105370917', Aliases='', description='uncharacterized LOC105370917', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78998773, end_position_on_the_genomic_accession=79011931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644390, CurrentID=0, Status='live', Symbol='HNRNPLP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein L pseudogene 2', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34488669, end_position_on_the_genomic_accession=34490501, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  120103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120103, CurrentID=0, Status='live', Symbol='SLC36A4', Aliases='PAT4', description='solute carrier family 36 (proton/amino acid symporter), member 4', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93144171, end_position_on_the_genomic_accession=93197975, orientation='minus', exon_count=13, OMIM=613760, no_hdr0=''),
  105370920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370920, CurrentID=0, Status='live', Symbol='LOC105370920', Aliases='', description='uncharacterized LOC105370920', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81403931, end_position_on_the_genomic_accession=81409170, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128041, CurrentID=0, Status='live', Symbol='KRT19P4', Aliases='', description='keratin 19 pseudogene 4', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68260317, end_position_on_the_genomic_accession=68261533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128042, CurrentID=0, Status='live', Symbol='BRD7P6', Aliases='', description='bromodomain containing 7 pseudogene 6', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70342095, end_position_on_the_genomic_accession=70355918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128043, CurrentID=0, Status='live', Symbol='FAF2P1', Aliases='', description='Fas associated factor family member 2 pseudogene 1', other_designations='', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91765902, end_position_on_the_genomic_accession=91767208, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128044, CurrentID=0, Status='live', Symbol='HNRNPA1P11', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 11', other_designations='', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72618185, end_position_on_the_genomic_accession=72619420, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644397, CurrentID=0, Status='live', Symbol='LRRC37A17P', Aliases='', description='leucine-rich repeat containing 37 member A17, pseudogene', other_designations='leucine rich repeat containing 37, member A17, pseudogene|leucine rich repeat containing 37, member A2 pseudogene|leucine rich repeat containing 37A pseudogene', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47017190, end_position_on_the_genomic_accession=47054412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128046, CurrentID=0, Status='live', Symbol='PP13439', Aliases='', description='uncharacterized LOC100128046', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.10', start_position_on_the_genomic_accession=172992834, end_position_on_the_genomic_accession=172993292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105370927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370927, CurrentID=0, Status='live', Symbol='LOC105370927', Aliases='', description='uncharacterized LOC105370927', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370928, CurrentID=0, Status='live', Symbol='LOC105370928', Aliases='', description='uncharacterized LOC105370928', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82892802, end_position_on_the_genomic_accession=82944060, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370929, CurrentID=0, Status='live', Symbol='LOC105370929', Aliases='', description='homeobox protein Hox-D11-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83207433, end_position_on_the_genomic_accession=83227217, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  120114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120114, CurrentID=0, Status='live', Symbol='FAT3', Aliases='CDHF15, CDHR10, hFat3', description='FAT atypical cadherin 3', other_designations='FAT tumor suppressor homolog 3|cadherin family member 15|cadherin-related family member 10', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92314283, end_position_on_the_genomic_accession=92896470, orientation='plus', exon_count=29, OMIM=612483, no_hdr0=''),
  644403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644403, CurrentID=0, Status='live', Symbol='LARP1BP3', Aliases='', description='La ribonucleoprotein domain family member 1B pseudogene 3', other_designations='La ribonucleoprotein domain family, member 1B pseudogene 3', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134203783, end_position_on_the_genomic_accession=134205907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370932, CurrentID=0, Status='live', Symbol='LOC105370932', Aliases='', description='uncharacterized LOC105370932', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83283480, end_position_on_the_genomic_accession=83318631, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370933, CurrentID=0, Status='live', Symbol='LOC105370933', Aliases='', description='uncharacterized LOC105370933', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=83345642, end_position_on_the_genomic_accession=83350586, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128054, CurrentID=0, Status='live', Symbol='PLCE1-AS1', Aliases='', description='PLCE1 antisense RNA 1', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94279290, end_position_on_the_genomic_accession=94287070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128055, CurrentID=0, Status='live', Symbol='SMARCA5-AS1', Aliases='', description='SMARCA5 antisense RNA 1', other_designations='SMARCA5 antisense RNA 1 (non-protein coding)', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143513472, end_position_on_the_genomic_accession=143514837, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370936, CurrentID=0, Status='live', Symbol='LOC105370936', Aliases='', description='uncharacterized LOC105370936', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84329664, end_position_on_the_genomic_accession=84335458, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128057, CurrentID=0, Status='live', Symbol='NIPA2P3', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2 pseudogene 3', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19620752, end_position_on_the_genomic_accession=19631774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370938, CurrentID=0, Status='live', Symbol='LOC105370938', Aliases='', description='uncharacterized LOC105370938', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370939, CurrentID=0, Status='live', Symbol='LOC105370939', Aliases='', description='proline-rich protein HaeIII subfamily 1-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31391503, end_position_on_the_genomic_accession=31404730, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128060, CurrentID=0, Status='live', Symbol='RPS7P10', Aliases='', description='ribosomal protein S7 pseudogene 10', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21628398, end_position_on_the_genomic_accession=21629059, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128061, CurrentID=0, Status='live', Symbol='HNRNPDLP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein D-like pseudogene 2', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158424725, end_position_on_the_genomic_accession=158425303, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644414, CurrentID=0, Status='live', Symbol='DEFB131', Aliases='DEFB-31', description='defensin beta 131', other_designations='beta-defensin 31|defensin, beta 131|defensin, beta 31', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9444534, end_position_on_the_genomic_accession=9450514, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370943, CurrentID=0, Status='live', Symbol='LOC105370943', Aliases='', description='uncharacterized LOC105370943', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40904922, end_position_on_the_genomic_accession=40908532, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105370944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370944, CurrentID=0, Status='live', Symbol='LOC105370944', Aliases='', description='inositol hexakisphosphate and diphosphoinositol-pentakisphosphate kinase 1-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105370945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370945, CurrentID=0, Status='live', Symbol='LOC105370945', Aliases='', description='probable lysosomal cobalamin transporter', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78129038, end_position_on_the_genomic_accession=78143173, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370946, CurrentID=0, Status='live', Symbol='LOC105370946', Aliases='', description='uncharacterized LOC105370946', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84452694, end_position_on_the_genomic_accession=84458486, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370947, CurrentID=0, Status='live', Symbol='LOC105370947', Aliases='', description='uncharacterized LOC105370947', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84603901, end_position_on_the_genomic_accession=84616345, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370948, CurrentID=0, Status='live', Symbol='LOC105370948', Aliases='', description='uncharacterized LOC105370948', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85219730, end_position_on_the_genomic_accession=85225519, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105370949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370949, CurrentID=0, Status='live', Symbol='LOC105370949', Aliases='', description='uncharacterized LOC105370949', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85248948, end_position_on_the_genomic_accession=85251088, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105370950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370950, CurrentID=0, Status='live', Symbol='LOC105370950', Aliases='', description='uncharacterized LOC105370950', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85620185, end_position_on_the_genomic_accession=85632904, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128071, CurrentID=0, Status='live', Symbol='FAM229A', Aliases='', description='family with sequence similarity 229 member A', other_designations='family with sequence similarity 229, member A', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32361270, end_position_on_the_genomic_accession=32362243, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128072, CurrentID=0, Status='live', Symbol='RPL36AP27', Aliases='RPL36A_9_808', description='ribosomal protein L36a pseudogene 27', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44510972, end_position_on_the_genomic_accession=44511353, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644425, CurrentID=0, Status='live', Symbol='RARRES2P2', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 2', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133637285, end_position_on_the_genomic_accession=133637913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370954, CurrentID=0, Status='live', Symbol='LOC105370954', Aliases='', description='uncharacterized LOC105370954', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86932815, end_position_on_the_genomic_accession=86946290, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370955, CurrentID=0, Status='live', Symbol='LOC105370955', Aliases='', description='uncharacterized LOC105370955', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=87121594, end_position_on_the_genomic_accession=87178557, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128076, CurrentID=0, Status='live', Symbol='LOC100128076', Aliases='', description='protein tyrosine phosphatase pseudogene', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92132834, end_position_on_the_genomic_accession=92138629, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  644429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644429, CurrentID=0, Status='live', Symbol='HMGB3P15', Aliases='', description='high mobility group box 3 pseudogene 15', other_designations='high-mobility group box 3 pseudogene 15', map_location='4q22.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=94195919, end_position_on_the_genomic_accession=94196544, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370958, CurrentID=0, Status='live', Symbol='LOC105370958', Aliases='', description='uncharacterized LOC105370958', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88291460, end_position_on_the_genomic_accession=88322277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370959, CurrentID=0, Status='live', Symbol='LOC105370959', Aliases='', description='uncharacterized LOC105370959', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88344942, end_position_on_the_genomic_accession=88364656, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370960, CurrentID=0, Status='live', Symbol='LOC105370960', Aliases='', description='uncharacterized LOC105370960', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88796465, end_position_on_the_genomic_accession=88799055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128081, CurrentID=0, Status='live', Symbol='JAZF1-AS1', Aliases='', description='JAZF1 antisense RNA 1', other_designations='JAZF1 antisense RNA 1 (non-protein coding)', map_location='7p15.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28180457, end_position_on_the_genomic_accession=28241377, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  120146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120146, CurrentID=0, Status='live', Symbol='TRIM64', Aliases='C11orf28A, TRIM64', description='tripartite motif containing 64', other_designations='', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89966087, end_position_on_the_genomic_accession=89974072, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100128083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128083, CurrentID=0, Status='live', Symbol='RPS15AP40', Aliases='RPS15A_26_1770', description='ribosomal protein S15a pseudogene 40', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36374755, end_position_on_the_genomic_accession=36375213, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370964, CurrentID=0, Status='live', Symbol='LOC105370964', Aliases='', description='uncharacterized LOC105370964', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89454012, end_position_on_the_genomic_accession=89469224, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105370965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370965, CurrentID=0, Status='live', Symbol='LOC105370965', Aliases='', description='uncharacterized LOC105370965', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89821449, end_position_on_the_genomic_accession=89825989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105370966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370966, CurrentID=0, Status='live', Symbol='LOC105370966', Aliases='', description='uncharacterized LOC105370966', other_designations='CTD-2315E11.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90102684, end_position_on_the_genomic_accession=90137945, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105370967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370967, CurrentID=0, Status='live', Symbol='LOC105370967', Aliases='', description='uncharacterized LOC105370967', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90630599, end_position_on_the_genomic_accession=90631447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128088, CurrentID=0, Status='live', Symbol='LOC100128088', Aliases='', description='matrix metallopeptidase 1-like', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=102621798, end_position_on_the_genomic_accession=102625674, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105370969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370969, CurrentID=0, Status='live', Symbol='LOC105370969', Aliases='', description='uncharacterized LOC105370969', other_designations='CTD-3094K11.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90839594, end_position_on_the_genomic_accession=90853629, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644442, CurrentID=0, Status='live', Symbol='HMGB3P11', Aliases='', description='high mobility group box 3 pseudogene 11', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 11', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104771191, end_position_on_the_genomic_accession=104772047, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370971, CurrentID=0, Status='live', Symbol='LOC105370971', Aliases='', description='putative protein TPRXL', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91270022, end_position_on_the_genomic_accession=91272820, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644444, CurrentID=0, Status='live', Symbol='TMEM30C', Aliases='CDC50C', description='transmembrane protein 30C', other_designations='', map_location='3q12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=100185824, end_position_on_the_genomic_accession=100194186, orientation='plus', exon_count=3, OMIM=611030, no_hdr0=''),
  105373601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373601, CurrentID=0, Status='live', Symbol='LOC105373601', Aliases='', description='uncharacterized LOC105373601', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126443977, end_position_on_the_genomic_accession=126461948, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128094, CurrentID=0, Status='live', Symbol='LPCAT2BP', Aliases='', description='lysophosphatidylcholine acyltransferase 2b, pseudogene', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92066460, end_position_on_the_genomic_accession=92067773, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370975, CurrentID=0, Status='live', Symbol='LOC105370975', Aliases='', description='uncharacterized LOC105370975', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92645518, end_position_on_the_genomic_accession=92646744, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130461, CurrentID=0, Status='live', Symbol='CBX1P4', Aliases='', description='chromobox 1 pseudogene 4', other_designations='chromobox homolog 1 (HP1 beta homolog Drosophila ) pseudogene|chromobox homolog 1 pseudogene 4', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17282351, end_position_on_the_genomic_accession=17284552, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644449, CurrentID=0, Status='live', Symbol='HNRNPFP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein F pseudogene 1', other_designations='heterogeneous nuclear ribonucleoprotein H1 (H) pseudogene', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42040545, end_position_on_the_genomic_accession=42042845, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128098, CurrentID=0, Status='live', Symbol='ST8SIA6-AS1', Aliases='', description='ST8SIA6 antisense RNA 1', other_designations='', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17386936, end_position_on_the_genomic_accession=17408286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644451, CurrentID=0, Status='live', Symbol='SHQ1P1', Aliases='', description='SHQ1, H/ACA ribonucleoprotein assembly factor pseudogene 1', other_designations='SHQ1 homolog pseudogene', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50629327, end_position_on_the_genomic_accession=50631322, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370980, CurrentID=0, Status='live', Symbol='LOC105370980', Aliases='', description='putative uncharacterized protein UNQ9370/PRO34162', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93206194, end_position_on_the_genomic_accession=93206739, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105370981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370981, CurrentID=0, Status='live', Symbol='LOC105370981', Aliases='', description='uncharacterized LOC105370981', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=93212134, end_position_on_the_genomic_accession=93219778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128102, CurrentID=0, Status='live', Symbol='PHBP16', Aliases='', description='prohibitin pseudogene 16', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94055023, end_position_on_the_genomic_accession=94055983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128103, CurrentID=0, Status='live', Symbol='AGGF1P5', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 5', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 5', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35420224, end_position_on_the_genomic_accession=35421207, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370984, CurrentID=0, Status='live', Symbol='LOC105370984', Aliases='', description='uncharacterized LOC105370984', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94196779, end_position_on_the_genomic_accession=94206890, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421180, CurrentID=0, Status='live', Symbol='PGBD4P6', Aliases='', description='piggyBac transposable element derived 4 pseudogene 6', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143343970, end_position_on_the_genomic_accession=143344597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370986, CurrentID=0, Status='live', Symbol='LOC105370986', Aliases='', description='uncharacterized LOC105370986', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94677567, end_position_on_the_genomic_accession=94679693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644459, CurrentID=0, Status='live', Symbol='BEND3P1', Aliases='', description='BEN domain containing 3 pseudogene 1', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50658103, end_position_on_the_genomic_accession=50663946, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370988, CurrentID=0, Status='live', Symbol='LOC105370988', Aliases='', description='uncharacterized LOC105370988', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94764750, end_position_on_the_genomic_accession=94791174, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128109, CurrentID=0, Status='live', Symbol='MEF2BNBP1', Aliases='', description='MEF2B neighbor pseudogene 1', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19107851, end_position_on_the_genomic_accession=19108183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128110, CurrentID=0, Status='live', Symbol='CHCHD2P3', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 3', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34126925, end_position_on_the_genomic_accession=34127716, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105370991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370991, CurrentID=0, Status='live', Symbol='LOC105370991', Aliases='', description='uncharacterized LOC105370991', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95018801, end_position_on_the_genomic_accession=95172194, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  644464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644464, CurrentID=0, Status='live', Symbol='RPSAP61', Aliases='RPSA_31_1776', description='ribosomal protein SA pseudogene 61', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44741011, end_position_on_the_genomic_accession=44742034, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  260328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260328, CurrentID=0, Status='live', Symbol='ALRH', Aliases='', description='Allergic rhinitis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607154, no_hdr0=''),
  105370994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370994, CurrentID=0, Status='live', Symbol='LOC105370994', Aliases='', description='uncharacterized LOC105370994', other_designations='CTD-2536I1.1', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95209093, end_position_on_the_genomic_accession=95214013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128115, CurrentID=0, Status='live', Symbol='RPL26P35', Aliases='RPL26_15_1614', description='ribosomal protein L26 pseudogene 35', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76834392, end_position_on_the_genomic_accession=76834960, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105370996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105370996, CurrentID=0, Status='live', Symbol='LOC105370996', Aliases='', description='uncharacterized LOC105370996', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=95610009, end_position_on_the_genomic_accession=95825249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128117, CurrentID=0, Status='live', Symbol='RPL17P15', Aliases='RPL17_5_269', description='ribosomal protein L17 pseudogene 15', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119698589, end_position_on_the_genomic_accession=119700190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128118, CurrentID=0, Status='live', Symbol='CCNHP1', Aliases='', description='cyclin H pseudogene 1', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181905624, end_position_on_the_genomic_accession=181906482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128119, CurrentID=0, Status='live', Symbol='SLC25A51P3', Aliases='MCART5P', description='solute carrier family 25, member 51 pseudogene 3', other_designations='mitochondrial carrier triple repeat 1 pseudogene|mitochondrial carrier triple repeat 5 pseudogene', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80594750, end_position_on_the_genomic_accession=80595868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371000, CurrentID=0, Status='live', Symbol='LOC105371000', Aliases='', description='uncharacterized LOC105371000', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96444337, end_position_on_the_genomic_accession=96447875, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371001, CurrentID=0, Status='live', Symbol='LOC105371001', Aliases='', description='uncharacterized LOC105371001', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96537558, end_position_on_the_genomic_accession=96551697, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371002, CurrentID=0, Status='live', Symbol='LOC105371002', Aliases='', description='uncharacterized LOC105371002', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96699278, end_position_on_the_genomic_accession=96704866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371003, CurrentID=0, Status='live', Symbol='LOC105371003', Aliases='', description='uncharacterized LOC105371003', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97065639, end_position_on_the_genomic_accession=97070507, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128124, CurrentID=0, Status='live', Symbol='HGC6.3', Aliases='', description='uncharacterized LOC100128124', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167975924, end_position_on_the_genomic_accession=167976939, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371005, CurrentID=0, Status='live', Symbol='LOC105371005', Aliases='', description='uncharacterized LOC105371005', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97106064, end_position_on_the_genomic_accession=97119146, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128126, CurrentID=0, Status='live', Symbol='STAU2-AS1', Aliases='', description='STAU2 antisense RNA 1', other_designations='STAU2 antisense RNA 1 (non-protein coding)', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73420074, end_position_on_the_genomic_accession=73441518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128127, CurrentID=0, Status='live', Symbol='ADGRA1-AS1', Aliases='GPR123-AS1', description='adhesion G protein-coupled receptor A1 antisense RNA 1', other_designations='GPR123 antisense RNA 1', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133085244, end_position_on_the_genomic_accession=133086697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371008, CurrentID=0, Status='live', Symbol='LOC105371008', Aliases='', description='uncharacterized LOC105371008', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98000657, end_position_on_the_genomic_accession=98022201, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371009, CurrentID=0, Status='live', Symbol='LOC105371009', Aliases='', description='uncharacterized LOC105371009', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=97994815, end_position_on_the_genomic_accession=98018694, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644482, CurrentID=0, Status='live', Symbol='NDUFA3P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A3 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3 pseudogene 3|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3, 9kDa pseudogene', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225964169, end_position_on_the_genomic_accession=225964494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371011, CurrentID=0, Status='live', Symbol='LOC105371011', Aliases='', description='uncharacterized LOC105371011', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128132, CurrentID=0, Status='live', Symbol='TMCC1P1', Aliases='TMCC1P', description='transmembrane and coiled-coil domain family 1 pseudogene 1', other_designations='transmembrane and coiled-coil domains 1 pseudogene', map_location='8q23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105194708, end_position_on_the_genomic_accession=105197464, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371013, CurrentID=0, Status='live', Symbol='LOC105371013', Aliases='', description='uncharacterized LOC105371013', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99005698, end_position_on_the_genomic_accession=99010169, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371014, CurrentID=0, Status='live', Symbol='LOC105371014', Aliases='', description='uncharacterized LOC105371014', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=98831622, end_position_on_the_genomic_accession=98840393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  347713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347713, CurrentID=0, Status='live', Symbol='DFNA42', Aliases='DFNA52', description='deafness, autosomal dominant 42', other_designations='deafness, autosomal dominant 52', map_location='5q31.1-q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128136, CurrentID=0, Status='live', Symbol='NDUFA5P4', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 4', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 4|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35874267, end_position_on_the_genomic_accession=35875208, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260332, CurrentID=0, Status='live', Symbol='ECEL1P3', Aliases='', description='endothelin converting enzyme-like 1, pseudogene 3', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232351125, end_position_on_the_genomic_accession=232351362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371018, CurrentID=0, Status='live', Symbol='LOC105371018', Aliases='', description='uncharacterized LOC105371018', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99394357, end_position_on_the_genomic_accession=99401465, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128139, CurrentID=0, Status='live', Symbol='LINC01233', Aliases='XLOC_013014', description='long intergenic non-protein coding RNA 1233', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22532626, end_position_on_the_genomic_accession=22533494, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128140, CurrentID=0, Status='live', Symbol='RPS4XP17', Aliases='RPS4P17, RPS4X_7_1505', description='ribosomal protein S4X pseudogene 17', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=656732, end_position_on_the_genomic_accession=657596, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371021, CurrentID=0, Status='live', Symbol='LOC105371021', Aliases='', description='uncharacterized LOC105371021', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100349606, end_position_on_the_genomic_accession=100352140, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371022, CurrentID=0, Status='live', Symbol='LOC105371022', Aliases='', description='uncharacterized LOC105371022', other_designations='CTD-3076O17.2', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99976481, end_position_on_the_genomic_accession=99981726, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644495, CurrentID=0, Status='live', Symbol='HSP90AB7P', Aliases='', description='heat shock protein 90kDa alpha family class B member 7, pseudogene', other_designations='heat shock protein 90kDa alpha (cytosolic), class B member 1 pseudogene|heat shock protein 90kDa alpha (cytosolic), class B member 7, pseudogene', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=9721712, end_position_on_the_genomic_accession=9724505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644496, CurrentID=0, Status='live', Symbol='CTSLP4', Aliases='CTSL1P4, CTSLL4, bA564C4.3', description='cathepsin L pseudogene 4', other_designations='cathepsin L-like 4|cathepsin L1 pseudogene 4', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50704658, end_position_on_the_genomic_accession=50708915, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128145, CurrentID=0, Status='live', Symbol='CLCP2', Aliases='LGALS10P1', description='Charcot-Leyden crystal protein pseudogene 2', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=114306590, end_position_on_the_genomic_accession=114307196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644498, CurrentID=0, Status='live', Symbol='HMGN2P7', Aliases='HMGN2L7', description='high mobility group nucleosomal binding domain 2 pseudogene 7', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 7|high-mobility group nucleosomal binding domain 2-like 7', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15374253, end_position_on_the_genomic_accession=15375760, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644499, CurrentID=0, Status='live', Symbol='RPL39P21', Aliases='RPL39_9_576', description='ribosomal protein L39 pseudogene 21', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=74931608, end_position_on_the_genomic_accession=74932010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371028, CurrentID=0, Status='live', Symbol='LOC105371028', Aliases='', description='uncharacterized LOC105371028', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101494805, end_position_on_the_genomic_accession=101501769, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  260334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260334, CurrentID=0, Status='live', Symbol='TUBB8P12', Aliases='', description='tubulin beta 8 pseudogene 12', other_designations='HSA18p11 beta-tubulin 4Q pseudogene', map_location='18p11', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47390, end_position_on_the_genomic_accession=49557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644502, CurrentID=0, Status='live', Symbol='HMGB3P18', Aliases='', description='high mobility group box 3 pseudogene 18', other_designations='high-mobility group box 3 pseudogene 18', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121858137, end_position_on_the_genomic_accession=121858784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371031, CurrentID=0, Status='live', Symbol='LOC105371031', Aliases='', description='vasodilator-stimulated phosphoprotein-like', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89399901, end_position_on_the_genomic_accession=89437853, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371032, CurrentID=0, Status='live', Symbol='LOC105371032', Aliases='', description='polyadenylate-binding protein 4 pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94028501, end_position_on_the_genomic_accession=94034318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421188, CurrentID=0, Status='live', Symbol='GAPDHP57', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 57', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75455991, end_position_on_the_genomic_accession=75456595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371034, CurrentID=0, Status='live', Symbol='LOC105371034', Aliases='', description='uncharacterized LOC105371034', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371035, CurrentID=0, Status='live', Symbol='LOC105371035', Aliases='', description='uncharacterized LOC105371035', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371036, CurrentID=0, Status='live', Symbol='LOC105371036', Aliases='', description='uncharacterized LOC105371036', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=383436, end_position_on_the_genomic_accession=386986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644509, CurrentID=0, Status='live', Symbol='RBM22P6', Aliases='RBM22P10', description='RNA binding motif protein 22 pseudogene 6', other_designations='RNA binding motif protein 22 pseudogene 10', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52485045, end_position_on_the_genomic_accession=52485984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371038, CurrentID=0, Status='live', Symbol='LOC105371038', Aliases='', description='uncharacterized LOC105371038', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=648482, end_position_on_the_genomic_accession=649313, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644511, CurrentID=0, Status='live', Symbol='RPL13AP6', Aliases='RPL13A_7_1096, bA348N5.5', description='ribosomal protein L13a pseudogene 6', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110936603, end_position_on_the_genomic_accession=110937255, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  120224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120224, CurrentID=0, Status='live', Symbol='TMEM45B', Aliases='', description='transmembrane protein 45B', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=129815781, end_position_on_the_genomic_accession=129860003, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105371041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371041, CurrentID=0, Status='live', Symbol='LOC105371041', Aliases='', description='uncharacterized LOC105371041', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=849813, end_position_on_the_genomic_accession=853977, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128162, CurrentID=0, Status='live', Symbol='RPS6P7', Aliases='RPS6_3_716', description='ribosomal protein S6 pseudogene 7', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77491595, end_position_on_the_genomic_accession=77496743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  120227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120227, CurrentID=0, Status='live', Symbol='CYP2R1', Aliases='', description='cytochrome P450 family 2 subfamily R member 1', other_designations='cytochrome P450 2R1|cytochrome P450, family 2, R1|cytochrome P450, family 2, subfamily R, polypeptide 1', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14877436, end_position_on_the_genomic_accession=14894034, orientation='minus', exon_count=9, OMIM=608713, no_hdr0=''),
  100128164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128164, CurrentID=0, Status='live', Symbol='LOC100128164', Aliases='', description='four and a half LIM domains 1 pseudogene', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169943984, end_position_on_the_genomic_accession=169966734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644517, CurrentID=0, Status='live', Symbol='FAM86MP', Aliases='', description='family with sequence similarity 86 member M, pseudogene', other_designations='family with sequence similarity 86, member M, pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9693097, end_position_on_the_genomic_accession=9703594, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371046, CurrentID=0, Status='live', Symbol='LOC105371046', Aliases='LA16c-395F10.2', description='uncharacterized LOC105371046', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1604316, end_position_on_the_genomic_accession=1610328, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  260337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260337, CurrentID=0, Status='live', Symbol='UHRF2P1', Aliases='', description='ubiquitin-like with PHD and ring finger domains 2 pseudogene 1', other_designations='zinc finger protein Np97 pseudogene', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74105234, end_position_on_the_genomic_accession=74108821, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128168, CurrentID=0, Status='live', Symbol='RPS26P39', Aliases='RPS26_21_1107', description='ribosomal protein S26 pseudogene 39', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123171458, end_position_on_the_genomic_accession=123171898, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371049, CurrentID=0, Status='live', Symbol='LOC105371049', Aliases='', description='uncharacterized LOC105371049', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2091436, end_position_on_the_genomic_accession=2094884, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644522, CurrentID=0, Status='live', Symbol='RPL31P44', Aliases='RPL31_14_1046', description='ribosomal protein L31 pseudogene 44', other_designations='', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52389065, end_position_on_the_genomic_accession=52389525, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371051, CurrentID=0, Status='live', Symbol='LOC105371051', Aliases='', description='uncharacterized LOC105371051', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2308630, end_position_on_the_genomic_accession=2311570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644524, CurrentID=0, Status='live', Symbol='NKX2-4', Aliases='NKX2.4, NKX2D', description='NK2 homeobox 4', other_designations='NK2 transcription factor homolog D|NK2 transcription factor related, locus 4', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21395367, end_position_on_the_genomic_accession=21397409, orientation='minus', exon_count=2, OMIM=607808, no_hdr0=''),
  120237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120237, CurrentID=0, Status='live', Symbol='DBX1', Aliases='', description='developing brain homeobox 1', other_designations='developing brain homeobox protein 1', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20156214, end_position_on_the_genomic_accession=20160324, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371054, CurrentID=0, Status='live', Symbol='LOC105371054', Aliases='LA16c-325D7.1', description='uncharacterized LOC105371054', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2857854, end_position_on_the_genomic_accession=2861778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371055, CurrentID=0, Status='live', Symbol='LOC105371055', Aliases='LA16c-321D4.2', description='uncharacterized LOC105371055', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2939885, end_position_on_the_genomic_accession=2956156, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100128176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128176, CurrentID=0, Status='live', Symbol='UST-AS1', Aliases='', description='UST antisense RNA 1', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148954259, end_position_on_the_genomic_accession=148964684, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  544328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544328, CurrentID=0, Status='live', Symbol='OA11', Aliases='', description='Osteoarthritis QTL 11', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371058, CurrentID=0, Status='live', Symbol='LOC105371058', Aliases='', description='uncharacterized LOC105371058', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3077272, end_position_on_the_genomic_accession=3087169, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128179, CurrentID=0, Status='live', Symbol='RAP1BP2', Aliases='', description='RAP1B, member of RAS oncogene family pseudogene 2', other_designations='', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104062894, end_position_on_the_genomic_accession=104065087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371060, CurrentID=0, Status='live', Symbol='LOC105371060', Aliases='', description='uncharacterized LOC105371060', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3565040, end_position_on_the_genomic_accession=3579753, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128181, CurrentID=0, Status='live', Symbol='FMO10P', Aliases='', description='flavin containing monooxygenase 10, pseudogene', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166649150, end_position_on_the_genomic_accession=166699360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644534, CurrentID=0, Status='live', Symbol='HMGN1P11', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 11', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 11', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62510357, end_position_on_the_genomic_accession=62511569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128183, CurrentID=0, Status='live', Symbol='PABPC1P7', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 7', other_designations='poly(A) binding protein, cytoplasmic, pseudogene', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102896733, end_position_on_the_genomic_accession=102898679, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371064, CurrentID=0, Status='live', Symbol='LOC105371064', Aliases='', description='uncharacterized LOC105371064', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4895352, end_position_on_the_genomic_accession=4926651, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371065, CurrentID=0, Status='live', Symbol='LOC105371065', Aliases='', description='uncharacterized LOC105371065', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5133265, end_position_on_the_genomic_accession=5143508, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644538, CurrentID=0, Status='live', Symbol='SMIM10', Aliases='CXorf69', description='small integral membrane protein 10', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134990938, end_position_on_the_genomic_accession=134992473, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371067, CurrentID=0, Status='live', Symbol='LOC105371067', Aliases='', description='uncharacterized LOC105371067', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5198457, end_position_on_the_genomic_accession=5235849, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644540, CurrentID=0, Status='live', Symbol='TCEB1P3', Aliases='dJ254P11.1', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 3', other_designations='transcription elongation factor B (SIII), polypeptide 1 (15kDa, elongin C) pseudogene 3', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10174154, end_position_on_the_genomic_accession=10174777, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421194, CurrentID=0, Status='live', Symbol='PDK1P1', Aliases='', description='pyruvate dehydrogenase kinase, isozyme 1 pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=29317833, end_position_on_the_genomic_accession=29318952, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128190, CurrentID=0, Status='live', Symbol='SLC9B1P1', Aliases='NHEDC1P1', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1 pseudogene 1', other_designations='Na+/H+ exchanger domain containing 1 pseudogene 1|Putative protein SLC9B1-like 1|solute carrier family 9, subfamily B (cation proton antiporter 2), member 1 pseudogene 1', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11340390, end_position_on_the_genomic_accession=11385695, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653557, CurrentID=0, Status='live', Symbol='BMS1P8', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 8', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 8|BMS1 ribosome biogenesis factor pseudogene 8', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33683167, end_position_on_the_genomic_accession=33702132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371072, CurrentID=0, Status='live', Symbol='LOC105371072', Aliases='', description='uncharacterized LOC105371072', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8379167, end_position_on_the_genomic_accession=8404151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371073, CurrentID=0, Status='live', Symbol='LOC105371073', Aliases='', description='uncharacterized LOC105371073', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8664176, end_position_on_the_genomic_accession=8674373, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371074, CurrentID=0, Status='live', Symbol='LOC105371074', Aliases='', description='uncharacterized LOC105371074', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8961505, end_position_on_the_genomic_accession=8966990, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371075, CurrentID=0, Status='live', Symbol='LOC105371075', Aliases='', description='uncharacterized LOC105371075', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9294818, end_position_on_the_genomic_accession=9297974, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644548, CurrentID=0, Status='live', Symbol='EXOC5P1', Aliases='', description='exocyst complex component 5 pseudogene 1', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62816826, end_position_on_the_genomic_accession=62818791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371077, CurrentID=0, Status='live', Symbol='LOC105371077', Aliases='', description='uncharacterized LOC105371077', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9791751, end_position_on_the_genomic_accession=9809101, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371078, CurrentID=0, Status='live', Symbol='LOC105371078', Aliases='', description='uncharacterized LOC105371078', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10215430, end_position_on_the_genomic_accession=10252643, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371079, CurrentID=0, Status='live', Symbol='LOC105371079', Aliases='', description='uncharacterized LOC105371079', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10691170, end_position_on_the_genomic_accession=10692971, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371080, CurrentID=0, Status='live', Symbol='LOC105371080', Aliases='', description='uncharacterized LOC105371080', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10842199, end_position_on_the_genomic_accession=10889476, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371081, CurrentID=0, Status='live', Symbol='LOC105371081', Aliases='', description='uncharacterized LOC105371081', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11129371, end_position_on_the_genomic_accession=11131519, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644554, CurrentID=0, Status='live', Symbol='LOC644554', Aliases='', description='uncharacterized LOC644554', other_designations='CTD-2554C21.2', map_location='19q13.13', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37817411, end_position_on_the_genomic_accession=37826638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128203, CurrentID=0, Status='live', Symbol='ENSAP3', Aliases='', description='endosulfine alpha pseudogene 3', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214697597, end_position_on_the_genomic_accession=214697925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371084, CurrentID=0, Status='live', Symbol='LOC105371084', Aliases='', description='uncharacterized LOC105371084', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11851568, end_position_on_the_genomic_accession=11896551, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371085, CurrentID=0, Status='live', Symbol='LOC105371085', Aliases='', description='uncharacterized LOC105371085', other_designations='CTD-3037G24.4', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12550538, end_position_on_the_genomic_accession=12568499, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371086, CurrentID=0, Status='live', Symbol='LOC105371086', Aliases='', description='uncharacterized LOC105371086', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12546524, end_position_on_the_genomic_accession=12549278, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371087, CurrentID=0, Status='live', Symbol='LOC105371087', Aliases='', description='uncharacterized LOC105371087', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12372671, end_position_on_the_genomic_accession=12373904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371088, CurrentID=0, Status='live', Symbol='LOC105371088', Aliases='', description='uncharacterized LOC105371088', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12651918, end_position_on_the_genomic_accession=12653370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371089, CurrentID=0, Status='live', Symbol='LOC105371089', Aliases='', description='uncharacterized LOC105371089', other_designations='CTD-2583P5.3', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12759162, end_position_on_the_genomic_accession=12760745, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371090, CurrentID=0, Status='live', Symbol='LOC105371090', Aliases='', description='uncharacterized LOC105371090', other_designations='CTD-2583P5.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12745840, end_position_on_the_genomic_accession=12759157, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371091, CurrentID=0, Status='live', Symbol='LOC105371091', Aliases='', description='uncharacterized LOC105371091', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=13185656, end_position_on_the_genomic_accession=13203387, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371092, CurrentID=0, Status='live', Symbol='LOC105371092', Aliases='', description='uncharacterized LOC105371092', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=12937515, end_position_on_the_genomic_accession=12957598, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100421198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421198, CurrentID=0, Status='live', Symbol='LAMP1P1', Aliases='LAMP1L1', description='lysosomal-associated membrane protein 1 pseudogene 1', other_designations='hes 24/07/2003|lysosomal-associated membrane protein 1-like 1', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1864581, end_position_on_the_genomic_accession=1865687, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644566, CurrentID=0, Status='live', Symbol='TPM4P1', Aliases='', description='tropomyosin 4 pseudogene 1', other_designations='tropomyosin 2 (beta) pseudogene', map_location='10q24.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100114421, end_position_on_the_genomic_accession=100114990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371095, CurrentID=0, Status='live', Symbol='LOC105371095', Aliases='', description='collagen alpha-3(IX) chain-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14695769, end_position_on_the_genomic_accession=14703638, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371096, CurrentID=0, Status='live', Symbol='LOC105371096', Aliases='', description='uncharacterized LOC105371096', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15181166, end_position_on_the_genomic_accession=15195343, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128217, CurrentID=0, Status='live', Symbol='GTF3AP5', Aliases='', description='general transcription factor IIIA pseudogene 5', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14985298, end_position_on_the_genomic_accession=14986125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128218, CurrentID=0, Status='live', Symbol='PSMC6P3', Aliases='', description='proteasome 26S subunit, ATPase, 6 pseudogene 3', other_designations='proteasome (prosome, macropain) 26S subunit, ATPase, 6 pseudogene', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33970854, end_position_on_the_genomic_accession=33972928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128219, CurrentID=0, Status='live', Symbol='LOC100128219', Aliases='', description='uncharacterized LOC100128219', other_designations='', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.8', start_position_on_the_genomic_accession=7933418, end_position_on_the_genomic_accession=7933789, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128220, CurrentID=0, Status='live', Symbol='EDDM3DP', Aliases='', description='epididymal protein 3D, pseudogene', other_designations='epididymal protein 3D (pseudogene)', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20734559, end_position_on_the_genomic_accession=20735360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644573, CurrentID=0, Status='live', Symbol='SPCS2P2', Aliases='', description='signal peptidase complex subunit 2 homolog (S. cerevisiae) pseudogene 2', other_designations='', map_location='10q24.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100143072, end_position_on_the_genomic_accession=100143996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371102, CurrentID=0, Status='live', Symbol='LOC105371102', Aliases='', description='uncharacterized LOC105371102', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15504720, end_position_on_the_genomic_accession=15521028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371103, CurrentID=0, Status='live', Symbol='LOC105371103', Aliases='', description='uncharacterized LOC105371103', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15359069, end_position_on_the_genomic_accession=15380921, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644576, CurrentID=0, Status='live', Symbol='SNX18P9', Aliases='', description='sorting nexin 18 pseudogene 9', other_designations='sorting nexin associated golgi protein 1 pseudogene', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64439032, end_position_on_the_genomic_accession=64440151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371105, CurrentID=0, Status='live', Symbol='LOC105371105', Aliases='', description='uncharacterized LOC105371105', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16455617, end_position_on_the_genomic_accession=16473684, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128226, CurrentID=0, Status='live', Symbol='LOC100128226', Aliases='', description='hypothetic protein', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=112457364, end_position_on_the_genomic_accession=112465243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128227, CurrentID=0, Status='live', Symbol='HMGB1P20', Aliases='', description='high mobility group box 1 pseudogene 20', other_designations='high-mobility group box 1 pseudogene 20', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53234991, end_position_on_the_genomic_accession=53236326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371108, CurrentID=0, Status='live', Symbol='LOC105371108', Aliases='', description='uncharacterized LOC105371108', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18092973, end_position_on_the_genomic_accession=18106954, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371109, CurrentID=0, Status='live', Symbol='LOC105371109', Aliases='', description='uncharacterized LOC105371109', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371110, CurrentID=0, Status='live', Symbol='LOC105371110', Aliases='', description='uncharacterized LOC105371110', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  151121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151121, CurrentID=0, Status='live', Symbol='LOC151121', Aliases='', description='uncharacterized LOC151121', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129242173, end_position_on_the_genomic_accession=129273891, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371112, CurrentID=0, Status='live', Symbol='LOC105371112', Aliases='', description='uncharacterized LOC105371112', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371113, CurrentID=0, Status='live', Symbol='LOC105371113', Aliases='', description='uncharacterized LOC105371113', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19334177, end_position_on_the_genomic_accession=19337799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371114, CurrentID=0, Status='live', Symbol='LOC105371114', Aliases='', description='uncharacterized LOC105371114', other_designations='CTA-363E6.2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19344393, end_position_on_the_genomic_accession=19393463, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371115, CurrentID=0, Status='live', Symbol='LOC105371115', Aliases='', description='uncharacterized LOC105371115', other_designations='CTD-2380F24.1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19761161, end_position_on_the_genomic_accession=19766822, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128236, CurrentID=0, Status='live', Symbol='GAFA3', Aliases='', description='FGF-2 activity-associated protein 3', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=37868283, end_position_on_the_genomic_accession=37868438, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128237, CurrentID=0, Status='live', Symbol='YRDCP2', Aliases='', description='yrdC N(6)-threonylcarbamoyltransferase domain containing pseudogene 2', other_designations='yrdC domain containing pseudogene 2', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96954836, end_position_on_the_genomic_accession=96955458, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371118, CurrentID=0, Status='live', Symbol='LOC105371118', Aliases='', description='uncharacterized LOC105371118', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20208604, end_position_on_the_genomic_accession=20209281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644591, CurrentID=0, Status='live', Symbol='PPIAL4G', Aliases='COAS-2', description='peptidylprolyl isomerase A (cyclophilin A)-like 4G', other_designations='PPIase A-like 4A|PPIase A-like 4G|chromosome 1-amplified sequence 2|peptidylprolyl cis-trans isomerase A-like 4|peptidylprolyl cis-trans isomerase A-like 4A|peptidylprolyl cis-trans isomerase A-like 4G', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148482548, end_position_on_the_genomic_accession=148483285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128240, CurrentID=0, Status='live', Symbol='CD83P1', Aliases='CD83P, HB15p, b34I8.3', description='CD83 molecule pseudogene 1', other_designations='CD83 antigen (activated B lymphocytes, immunoglobulin superfamily) pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27560826, end_position_on_the_genomic_accession=27561771, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371121, CurrentID=0, Status='live', Symbol='LOC105371121', Aliases='', description='uncharacterized LOC105371121', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20830579, end_position_on_the_genomic_accession=20839811, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371122, CurrentID=0, Status='live', Symbol='LOC105371122', Aliases='', description='uncharacterized LOC105371122', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371123, CurrentID=0, Status='live', Symbol='LOC105371123', Aliases='', description='uncharacterized LOC105371123', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21227419, end_position_on_the_genomic_accession=21244079, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644596, CurrentID=0, Status='live', Symbol='SMIM10L2B', Aliases='LINC00087, NCRNA00087', description='small integral membrane protein 10 like 2B', other_designations='Small integral membrane protein 10-like protein 2B|long intergenic non-protein coding RNA 87', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135094985, end_position_on_the_genomic_accession=135098703, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371125, CurrentID=0, Status='live', Symbol='LOC105371125', Aliases='', description='uncharacterized LOC105371125', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21290470, end_position_on_the_genomic_accession=21294346, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371126, CurrentID=0, Status='live', Symbol='LOC105371126', Aliases='', description='uncharacterized LOC105371126', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21761324, end_position_on_the_genomic_accession=21793090, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371127, CurrentID=0, Status='live', Symbol='LOC105371127', Aliases='', description='uncharacterized LOC105371127', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644600, CurrentID=0, Status='live', Symbol='GEMIN8P2', Aliases='', description='gem nuclear organelle associated protein 8 pseudogene 2', other_designations='gem (nuclear organelle) associated protein 8 pseudogene 2', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58776540, end_position_on_the_genomic_accession=58777560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644601, CurrentID=0, Status='live', Symbol='BIN2P2', Aliases='', description='bridging integrator 2 pseudogene 2', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75910563, end_position_on_the_genomic_accession=75912068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128250, CurrentID=0, Status='live', Symbol='RPL31P55', Aliases='RPL31_31_1463', description='ribosomal protein L31 pseudogene 55', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96448126, end_position_on_the_genomic_accession=96448530, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371131, CurrentID=0, Status='live', Symbol='LOC105371131', Aliases='', description='uncharacterized LOC105371131', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22532772, end_position_on_the_genomic_accession=22538879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644604, CurrentID=0, Status='live', Symbol='EEF1A1P12', Aliases='EEF1AL10', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 12', other_designations='eukaryotic translation elongation factor 1 alpha-like 10', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106697038, end_position_on_the_genomic_accession=106698676, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128253, CurrentID=0, Status='live', Symbol='LOC100128253', Aliases='', description='uncharacterized LOC100128253', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=3405217, end_position_on_the_genomic_accession=3409390, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  120318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120318, CurrentID=0, Status='live', Symbol='RPS6P17', Aliases='RPS6_7_1168', description='ribosomal protein S6 pseudogene 17', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102295019, end_position_on_the_genomic_accession=102295839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371135, CurrentID=0, Status='live', Symbol='LOC105371135', Aliases='', description='uncharacterized LOC105371135', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371136, CurrentID=0, Status='live', Symbol='LOC105371136', Aliases='', description='uncharacterized LOC105371136', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  120321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120321, CurrentID=0, Status='live', Symbol='CSNK1A1P2', Aliases='', description='casein kinase 1, alpha 1 pseudogene 2', other_designations='', map_location='11q22.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=102806949, end_position_on_the_genomic_accession=102808326, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128258, CurrentID=0, Status='live', Symbol='RPS18P3', Aliases='RPS18_2_171', description='ribosomal protein S18 pseudogene 3', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227430482, end_position_on_the_genomic_accession=227431019, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371139, CurrentID=0, Status='live', Symbol='LOC105371139', Aliases='', description='uncharacterized LOC105371139', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128260, CurrentID=0, Status='live', Symbol='WASIR1', Aliases='NCRNA00286B', description='WASH and IL9R antisense RNA 1', other_designations='WASH and IL9R antisense RNA 1 (non-protein coding)|non-protein coding RNA 286B', map_location='Xq28; Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=59347608, end_position_on_the_genomic_accession=59349582, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371141, CurrentID=0, Status='live', Symbol='LOC105371141', Aliases='', description='uncharacterized LOC105371141', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24100771, end_position_on_the_genomic_accession=24116727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128262, CurrentID=0, Status='live', Symbol='TNK2-AS1', Aliases='', description='TNK2 antisense RNA 1', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=195634947, end_position_on_the_genomic_accession=195638128, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371143, CurrentID=0, Status='live', Symbol='LOC105371143', Aliases='', description='uncharacterized LOC105371143', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=24465527, end_position_on_the_genomic_accession=24479097, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128264, CurrentID=0, Status='live', Symbol='HTR5A-AS1', Aliases='HTR5AOS', description='HTR5A antisense RNA 1', other_designations='HTR5A opposite strand', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155067067, end_position_on_the_genomic_accession=155071557, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128265, CurrentID=0, Status='live', Symbol='LOC100128265', Aliases='', description='uncharacterized LOC100128265', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.9', start_position_on_the_genomic_accession=145507941, end_position_on_the_genomic_accession=145508828, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128266, CurrentID=0, Status='live', Symbol='RANP6', Aliases='', description='RAN, member RAS oncogene family pseudogene 6', other_designations='', map_location='4q34.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173633531, end_position_on_the_genomic_accession=173634415, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644619, CurrentID=0, Status='live', Symbol='INTS4P2', Aliases='INTS4L2', description='integrator complex subunit 4 pseudogene 2', other_designations='integrator complex subunit 4-like 2', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65647854, end_position_on_the_genomic_accession=65718688, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105371148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371148, CurrentID=0, Status='live', Symbol='LOC105371148', Aliases='', description='uncharacterized LOC105371148', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=25736010, end_position_on_the_genomic_accession=25743842, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371149, CurrentID=0, Status='live', Symbol='LOC105371149', Aliases='', description='uncharacterized LOC105371149', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26269114, end_position_on_the_genomic_accession=26270862, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371150, CurrentID=0, Status='live', Symbol='LOC105371150', Aliases='', description='uncharacterized LOC105371150', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26280443, end_position_on_the_genomic_accession=26310707, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644623, CurrentID=0, Status='live', Symbol='TPTE2P2', Aliases='', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 2', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=52219334, end_position_on_the_genomic_accession=52334284, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  105371152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371152, CurrentID=0, Status='live', Symbol='LOC105371152', Aliases='', description='uncharacterized LOC105371152', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=26584772, end_position_on_the_genomic_accession=26595452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644625, CurrentID=0, Status='live', Symbol='FUNDC2P3', Aliases='', description='FUN14 domain containing 2 pseudogene 3', other_designations='', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146171290, end_position_on_the_genomic_accession=146172072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128274, CurrentID=0, Status='live', Symbol='LOC100128274', Aliases='', description='putative p150', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=32433797, end_position_on_the_genomic_accession=32436074, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  644627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644627, CurrentID=0, Status='live', Symbol='RPL21P107', Aliases='RPL21_46_1336', description='ribosomal protein L21 pseudogene 107', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110762644, end_position_on_the_genomic_accession=110763176, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371156, CurrentID=0, Status='live', Symbol='LOC105371156', Aliases='', description='uncharacterized LOC105371156', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128277, CurrentID=0, Status='live', Symbol='RPL35AP20', Aliases='RPL35A_9_959', description='ribosomal protein L35a pseudogene 20', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6675215, end_position_on_the_genomic_accession=6675662, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371158, CurrentID=0, Status='live', Symbol='LOC105371158', Aliases='', description='uncharacterized LOC105371158', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28097474, end_position_on_the_genomic_accession=28111036, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928537, CurrentID=0, Status='live', Symbol='LINC01494', Aliases='', description='long intergenic non-protein coding RNA 1494', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218900814, end_position_on_the_genomic_accession=218930637, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  644632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644632, CurrentID=0, Status='live', Symbol='BMS1P11', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 11', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 11|BMS1 ribosome biogenesis factor pseudogene 11|BMS1-like protein ENSP00000382360 pseudogene', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64705734, end_position_on_the_genomic_accession=64709306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371161, CurrentID=0, Status='live', Symbol='LOC105371161', Aliases='', description='uncharacterized LOC105371161', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29133344, end_position_on_the_genomic_accession=29141824, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644634, CurrentID=0, Status='live', Symbol='FAM231D', Aliases='', description='family with sequence similarity 231 member D', other_designations='family with sequence similarity 231, member D', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149701425, end_position_on_the_genomic_accession=149705717, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371163, CurrentID=0, Status='live', Symbol='LOC105371163', Aliases='', description='uncharacterized LOC105371163', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28837639, end_position_on_the_genomic_accession=28842699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371164, CurrentID=0, Status='live', Symbol='LOC105371164', Aliases='', description='uncharacterized LOC105371164', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28597305, end_position_on_the_genomic_accession=28603054, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644637, CurrentID=0, Status='live', Symbol='NF1P6', Aliases='NF1L4, NF1L6', description='neurofibromin 1 pseudogene 6', other_designations='neurofibromin 1-like 6|putative neurofibromin 1-like protein 4/6', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15622601, end_position_on_the_genomic_accession=15632050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371166, CurrentID=0, Status='live', Symbol='LOC105371166', Aliases='', description='uncharacterized LOC105371166', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29808649, end_position_on_the_genomic_accession=29809533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371167, CurrentID=0, Status='live', Symbol='LOC105371167', Aliases='', description='uncharacterized LOC105371167', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29787513, end_position_on_the_genomic_accession=29788123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128288, CurrentID=0, Status='live', Symbol='LOC100128288', Aliases='', description='uncharacterized LOC100128288', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8358413, end_position_on_the_genomic_accession=8360541, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371169, CurrentID=0, Status='live', Symbol='LOC105371169', Aliases='', description='uncharacterized LOC105371169', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30370114, end_position_on_the_genomic_accession=30373010, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371170, CurrentID=0, Status='live', Symbol='LOC105371170', Aliases='', description='uncharacterized LOC105371170', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371171, CurrentID=0, Status='live', Symbol='LOC105371171', Aliases='', description='uncharacterized LOC105371171', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31487196, end_position_on_the_genomic_accession=31488492, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128292, CurrentID=0, Status='live', Symbol='DLG5-AS1', Aliases='', description='DLG5 antisense RNA 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77926812, end_position_on_the_genomic_accession=77929825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371173, CurrentID=0, Status='live', Symbol='LOC105371173', Aliases='', description='uncharacterized LOC105371173', other_designations='CTD-2014E2.2|CTD-2014E2.6', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31542751, end_position_on_the_genomic_accession=31553607, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371174, CurrentID=0, Status='live', Symbol='LOC105371174', Aliases='', description='uncharacterized LOC105371174', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32044693, end_position_on_the_genomic_accession=32050933, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371175, CurrentID=0, Status='live', Symbol='LOC105371175', Aliases='', description='uncharacterized LOC105371175', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143327631, end_position_on_the_genomic_accession=143345549, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371176, CurrentID=0, Status='live', Symbol='LOC105371176', Aliases='', description='Ig heavy chain V-III region VH26-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32995002, end_position_on_the_genomic_accession=32996770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644649, CurrentID=0, Status='live', Symbol='APOOP5', Aliases='', description='apolipoprotein O pseudogene 5', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59754132, end_position_on_the_genomic_accession=59755191, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128298, CurrentID=0, Status='live', Symbol='OCR1', Aliases='', description='ovarian cancer-related protein 1', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=200766000, end_position_on_the_genomic_accession=200766230, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644651, CurrentID=0, Status='live', Symbol='HMGB1P25', Aliases='', description='high mobility group box 1 pseudogene 25', other_designations='high-mobility group box 1 pseudogene 25', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9018263, end_position_on_the_genomic_accession=9018899, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  120364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120364, CurrentID=0, Status='live', Symbol='HNRNPA1P60', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 60', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=110787700, end_position_on_the_genomic_accession=110789069, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371181, CurrentID=0, Status='live', Symbol='LOC105371181', Aliases='', description='uncharacterized LOC105371181', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128302, CurrentID=0, Status='live', Symbol='SDHCP2', Aliases='', description='succinate dehydrogenase complex subunit C pseudogene 2', other_designations='succinate dehydrogenase complex, subunit C pseudogene 2|succinate dehydrogenase complex, subunit C, integral membrane protein, 15kDa pseudogene', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=92118645, end_position_on_the_genomic_accession=92119899, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371183, CurrentID=0, Status='live', Symbol='LOC105371183', Aliases='', description='uncharacterized LOC105371183', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33666640, end_position_on_the_genomic_accession=33676044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644656, CurrentID=0, Status='live', Symbol='LOC644656', Aliases='', description='uncharacterized LOC644656', other_designations='CTD-2371O3.3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9459556, end_position_on_the_genomic_accession=9460698, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128305, CurrentID=0, Status='live', Symbol='RPS13P1', Aliases='RPS13_1_111', description='ribosomal protein S13 pseudogene 1', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165671256, end_position_on_the_genomic_accession=165671731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371186, CurrentID=0, Status='live', Symbol='LOC105371186', Aliases='', description='nuclear pore complex-interacting protein family member A5-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644659, CurrentID=0, Status='live', Symbol='HMGB1P22', Aliases='', description='high mobility group box 1 pseudogene 22', other_designations='high-mobility group box 1 pseudogene 22', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125265263, end_position_on_the_genomic_accession=125265802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644660, CurrentID=0, Status='live', Symbol='RAD21-AS1', Aliases='C8orf81, NCRNA00255', description='RAD21 antisense RNA 1', other_designations='RAD21 antisense RNA 1 (non-protein coding)', map_location='8q24.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116874424, end_position_on_the_genomic_accession=116876868, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371189, CurrentID=0, Status='live', Symbol='LOC105371189', Aliases='', description='nuclear pore complex-interacting protein family member B9-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128310, CurrentID=0, Status='live', Symbol='LOC100128310', Aliases='', description='uncharacterized LOC100128310', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=60292562, end_position_on_the_genomic_accession=60294804, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372257, CurrentID=0, Status='live', Symbol='LOC105372257', Aliases='', description='uncharacterized LOC105372257', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6953211, end_position_on_the_genomic_accession=6954912, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  120376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120376, CurrentID=0, Status='live', Symbol='COLCA2', Aliases='C11orf93, CASC13', description='colorectal cancer associated 2', other_designations='cancer susceptibility candidate 13|cancer susceptibility candidate protein 13|colorectal cancer associated two', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111298546, end_position_on_the_genomic_accession=111308735, orientation='plus', exon_count=7, OMIM=615694, no_hdr0=''),
  105371193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371193, CurrentID=0, Status='live', Symbol='LOC105371193', Aliases='', description='putative HERC2-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371194, CurrentID=0, Status='live', Symbol='LOC105371194', Aliases='', description='putative HERC2-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  120379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120379, CurrentID=0, Status='live', Symbol='PIH1D2', Aliases='', description='PIH1 domain containing 2', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112052409, end_position_on_the_genomic_accession=112074171, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371196, CurrentID=0, Status='live', Symbol='LOC105371196', Aliases='', description='uncharacterized LOC105371196', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128317, CurrentID=0, Status='live', Symbol='LOC100128317', Aliases='', description='uncharacterized LOC100128317', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81576386, end_position_on_the_genomic_accession=81691406, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100128318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128318, CurrentID=0, Status='live', Symbol='RPL26P17', Aliases='RPL26_8_492', description='ribosomal protein L26 pseudogene 17', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76886045, end_position_on_the_genomic_accession=76891447, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371199, CurrentID=0, Status='live', Symbol='LOC105371199', Aliases='', description='uncharacterized LOC105371199', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35319031, end_position_on_the_genomic_accession=35327711, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644672, CurrentID=0, Status='live', Symbol='CLDN25', Aliases='', description='claudin 25', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113779796, end_position_on_the_genomic_accession=113780485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371201, CurrentID=0, Status='live', Symbol='LOC105371201', Aliases='', description='uncharacterized LOC105371201', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35398493, end_position_on_the_genomic_accession=35401780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371202, CurrentID=0, Status='live', Symbol='LOC105371202', Aliases='', description='uncharacterized LOC105371202', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35466230, end_position_on_the_genomic_accession=35476658, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371203, CurrentID=0, Status='live', Symbol='LOC105371203', Aliases='', description='uncharacterized LOC105371203', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35450771, end_position_on_the_genomic_accession=35457947, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128324, CurrentID=0, Status='live', Symbol='EXOGP1', Aliases='', description="endo/exonuclease (5'-3'), endonuclease G-like pseudogene 1", other_designations='', map_location='18q11.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21180843, end_position_on_the_genomic_accession=21183515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371205, CurrentID=0, Status='live', Symbol='LOC105371205', Aliases='', description='uncharacterized LOC105371205', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35555701, end_position_on_the_genomic_accession=35565022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371206, CurrentID=0, Status='live', Symbol='LOC105371206', Aliases='', description='uncharacterized LOC105371206', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144153169, end_position_on_the_genomic_accession=144171251, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128327, CurrentID=0, Status='live', Symbol='TRAPPC3L', Aliases='BET3L, bA259P20.2', description='trafficking protein particle complex 3-like', other_designations='BET3-like protein|TRAPPC3-like protein', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116496483, end_position_on_the_genomic_accession=116545610, orientation='minus', exon_count=6, OMIM=614137, no_hdr0=''),
  105371208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371208, CurrentID=0, Status='live', Symbol='LOC105371208', Aliases='', description='uncharacterized LOC105371208', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143976738, end_position_on_the_genomic_accession=143982381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128329, CurrentID=0, Status='live', Symbol='PRO2958', Aliases='', description='uncharacterized LOC100128329', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=15555549, end_position_on_the_genomic_accession=15555725, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371210, CurrentID=0, Status='live', Symbol='LOC105371210', Aliases='', description='uncharacterized LOC105371210', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143825905, end_position_on_the_genomic_accession=143831979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371211, CurrentID=0, Status='live', Symbol='LOC105371211', Aliases='', description='uncharacterized LOC105371211', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144472375, end_position_on_the_genomic_accession=144485779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644684, CurrentID=0, Status='live', Symbol='BMS1P12', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 12', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 12|BMS1 ribosome biogenesis factor pseudogene 12', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65189995, end_position_on_the_genomic_accession=65193571, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371213, CurrentID=0, Status='live', Symbol='LOC105371213', Aliases='', description='uncharacterized LOC105371213', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144512466, end_position_on_the_genomic_accession=144532363, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371214, CurrentID=0, Status='live', Symbol='LOC105371214', Aliases='', description='uncharacterized LOC105371214', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653581, CurrentID=0, Status='live', Symbol='LOC653581', Aliases='', description='uncharacterized LOC653581', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=63131644, end_position_on_the_genomic_accession=63132942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  120400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120400, CurrentID=0, Status='live', Symbol='NXPE1', Aliases='FAM55A', description='neurexophilin and PC-esterase domain family member 1', other_designations='family with sequence similarity 55, member A|neurexophilin and PC-esterase domain family, member 1|protein FAM55A', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114517732, end_position_on_the_genomic_accession=114559895, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100128337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128337, CurrentID=0, Status='live', Symbol='IPO7P2', Aliases='', description='importin 7 pseudogene 2', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24121024, end_position_on_the_genomic_accession=24123814, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128338, CurrentID=0, Status='live', Symbol='FAM83H-AS1', Aliases='onco-lncRNA-3', description='FAM83H antisense RNA 1 (head to head)', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143734140, end_position_on_the_genomic_accession=143746337, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=347747, CurrentID=0, Status='live', Symbol='HPCQTL19', Aliases='', description='Prostate cancer aggressiveness quantitative trait locus on chromosome 19', other_designations='', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607592, no_hdr0=''),
  105371220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371220, CurrentID=0, Status='live', Symbol='LOC105371220', Aliases='', description='uncharacterized LOC105371220', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149047909, end_position_on_the_genomic_accession=149051273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128341, CurrentID=0, Status='live', Symbol='RBMX2P1', Aliases='', description='RNA binding motif protein, X-linked 2 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14829149, end_position_on_the_genomic_accession=14829989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  120406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120406, CurrentID=0, Status='live', Symbol='NXPE2', Aliases='FAM55B', description='neurexophilin and PC-esterase domain family member 2', other_designations='family with sequence similarity 55, member B|neurexophilin and PC-esterase domain family, member 2|protein FAM55B', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114648490, end_position_on_the_genomic_accession=114707117, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371223, CurrentID=0, Status='live', Symbol='LOC105371223', Aliases='', description='uncharacterized LOC105371223', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148680046, end_position_on_the_genomic_accession=148686764, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371224, CurrentID=0, Status='live', Symbol='LOC105371224', Aliases='', description='uncharacterized LOC105371224', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371225, CurrentID=0, Status='live', Symbol='LOC105371225', Aliases='', description='uncharacterized LOC105371225', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148290891, end_position_on_the_genomic_accession=148296853, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371226, CurrentID=0, Status='live', Symbol='LOC105371226', Aliases='', description='uncharacterized LOC105371226', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148271896, end_position_on_the_genomic_accession=148281363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128347, CurrentID=0, Status='live', Symbol='RPL15P15', Aliases='RPL15_4_1179', description='ribosomal protein L15 pseudogene 15', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115950190, end_position_on_the_genomic_accession=115950801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371228, CurrentID=0, Status='live', Symbol='LOC105371228', Aliases='', description='uncharacterized LOC105371228', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148009531, end_position_on_the_genomic_accession=148027696, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371229, CurrentID=0, Status='live', Symbol='LOC105371229', Aliases='', description='uncharacterized LOC105371229', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147908673, end_position_on_the_genomic_accession=147928673, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371230, CurrentID=0, Status='live', Symbol='LOC105371230', Aliases='', description='uncharacterized LOC105371230', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147700808, end_position_on_the_genomic_accession=147738850, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128351, CurrentID=0, Status='live', Symbol='RPL21P52', Aliases='RPL21_16_474', description='ribosomal protein L21 pseudogene 52', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47490946, end_position_on_the_genomic_accession=47491503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371232, CurrentID=0, Status='live', Symbol='LOC105371232', Aliases='', description='uncharacterized LOC105371232', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146954396, end_position_on_the_genomic_accession=146956851, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371233, CurrentID=0, Status='live', Symbol='LOC105371233', Aliases='', description='uncharacterized LOC105371233', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146448455, end_position_on_the_genomic_accession=146452348, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371234, CurrentID=0, Status='live', Symbol='LOC105371234', Aliases='', description='uncharacterized LOC105371234', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146430421, end_position_on_the_genomic_accession=146433108, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371235, CurrentID=0, Status='live', Symbol='LOC105371235', Aliases='', description='uncharacterized LOC105371235', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146410595, end_position_on_the_genomic_accession=146443993, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128356, CurrentID=0, Status='live', Symbol='LOC100128356', Aliases='', description='protein transactivated by hepatitis B virus E antigen', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=4482199, end_position_on_the_genomic_accession=4482522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371237, CurrentID=0, Status='live', Symbol='LOC105371237', Aliases='', description='uncharacterized LOC105371237', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=47953039, end_position_on_the_genomic_accession=47965551, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128358, CurrentID=0, Status='live', Symbol='RPS6P24', Aliases='RPS6_12_1345', description='ribosomal protein S6 pseudogene 24', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=30055161, end_position_on_the_genomic_accession=30055988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371239, CurrentID=0, Status='live', Symbol='LOC105371239', Aliases='', description='probable ribosome biogenesis protein RLP24 pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48600120, end_position_on_the_genomic_accession=48601978, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371240, CurrentID=0, Status='live', Symbol='LOC105371240', Aliases='', description='uncharacterized LOC105371240', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=48623438, end_position_on_the_genomic_accession=48748924, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  120425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120425, CurrentID=0, Status='live', Symbol='JAML', Aliases='AMICA, AMICA1, CREA7-1, CREA7-4, Gm638', description='junction adhesion molecule like', other_designations='adhesion molecule AMICA|adhesion molecule interacting with CXADR antigen 1|adhesion molecule, interacts with CXADR antigen 1|dendritic-cell specific protein CREA7-1|dendritic-cell specific protein CREA7-4', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118193727, end_position_on_the_genomic_accession=118225094, orientation='minus', exon_count=14, OMIM=609770, no_hdr0=''),
  644714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644714, CurrentID=0, Status='live', Symbol='LIMD1-AS1', Aliases='', description='LIMD1 antisense RNA 1', other_designations='LIMD1 antisense RNA 1 (non-protein coding)', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45678165, end_position_on_the_genomic_accession=45688882, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371243, CurrentID=0, Status='live', Symbol='LOC105371243', Aliases='', description='uncharacterized LOC105371243', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49312511, end_position_on_the_genomic_accession=49316515, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644716, CurrentID=0, Status='live', Symbol='BMS1P13', Aliases='BMS1P14', description='BMS1, ribosome biogenesis factor pseudogene 13', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 13|BMS1 pseudogene 14|BMS1 ribosome biogenesis factor pseudogene 13', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65828120, end_position_on_the_genomic_accession=65831696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644717, CurrentID=0, Status='live', Symbol='SAGE2P', Aliases='', description='sarcoma antigen 2, pseudogene', other_designations='putative SAGE1-like protein', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135638042, end_position_on_the_genomic_accession=135688495, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105371246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371246, CurrentID=0, Status='live', Symbol='LOC105371246', Aliases='', description='uncharacterized LOC105371246', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371247, CurrentID=0, Status='live', Symbol='LOC105371247', Aliases='', description='uncharacterized LOC105371247', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128368, CurrentID=0, Status='live', Symbol='VN1R21P', Aliases='', description='vomeronasal 1 receptor 21 pseudogene', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14089053, end_position_on_the_genomic_accession=14089858, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371249, CurrentID=0, Status='live', Symbol='LOC105371249', Aliases='', description='uncharacterized LOC105371249', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146282612, end_position_on_the_genomic_accession=146285212, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371250, CurrentID=0, Status='live', Symbol='LOC105371250', Aliases='', description='uncharacterized LOC105371250', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50505997, end_position_on_the_genomic_accession=50522561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371251, CurrentID=0, Status='live', Symbol='LOC105371251', Aliases='', description='uncharacterized LOC105371251', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50783859, end_position_on_the_genomic_accession=50805332, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128372, CurrentID=0, Status='live', Symbol='HMGN2P28', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 28', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 28', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2341357, end_position_on_the_genomic_accession=2342500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128373, CurrentID=0, Status='live', Symbol='LOC100128373', Aliases='', description='AKT interacting protein pseudogene', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=101217387, end_position_on_the_genomic_accession=101218231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128374, CurrentID=0, Status='live', Symbol='LOC100128374', Aliases='', description='uncharacterized LOC100128374', other_designations='', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644727, CurrentID=0, Status='live', Symbol='LOC644727', Aliases='', description='uncharacterized LOC644727', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=53788684, end_position_on_the_genomic_accession=53790926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371256, CurrentID=0, Status='live', Symbol='LOC105371256', Aliases='', description='uncharacterized LOC105371256', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51362002, end_position_on_the_genomic_accession=51377976, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128377, CurrentID=0, Status='live', Symbol='RPS29P7', Aliases='RPS29_2_60', description='ribosomal protein S29 pseudogene 7', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65154418, end_position_on_the_genomic_accession=65154738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128378, CurrentID=0, Status='live', Symbol='LINC00696', Aliases='C3orf74', description='long intergenic non-protein coding RNA 696', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52062094, end_position_on_the_genomic_accession=52065112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  644731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644731, CurrentID=0, Status='live', Symbol='IGKV1OR-3', Aliases='IGKV1/OR-3, IGKV1OR3, IGKVPZ3', description='immunoglobulin kappa variable 1/OR-3 (pseudogene)', other_designations='immunoglobulin kappa variable 1-5 (IGKV1-5) pseudogene|immunoglobulin kappa variable 1/OR-3 pseudogene|immunoglobulin kappa variable region pseudogene Z3', map_location='9q12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=70394426, end_position_on_the_genomic_accession=70396639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371260, CurrentID=0, Status='live', Symbol='LOC105371260', Aliases='', description='uncharacterized LOC105371260', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145927557, end_position_on_the_genomic_accession=145949052, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371261, CurrentID=0, Status='live', Symbol='LOC105371261', Aliases='', description='uncharacterized LOC105371261', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52239680, end_position_on_the_genomic_accession=52269440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371262, CurrentID=0, Status='live', Symbol='LOC105371262', Aliases='', description='uncharacterized LOC105371262', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52416790, end_position_on_the_genomic_accession=52419866, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371263, CurrentID=0, Status='live', Symbol='LOC105371263', Aliases='', description='uncharacterized LOC105371263', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52622464, end_position_on_the_genomic_accession=52630327, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128384, CurrentID=0, Status='live', Symbol='VN1R65P', Aliases='', description='vomeronasal 1 receptor 65 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31554053, end_position_on_the_genomic_accession=31554927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128385, CurrentID=0, Status='live', Symbol='FAM225B', Aliases='C9orf110, LINC00256B, NCRNA00256B', description='family with sequence similarity 225 member B (non-protein coding)', other_designations='family with sequence similarity 225, member B (non-protein coding)|long intergenic non-protein coding RNA 256B|non-protein coding RNA 256B', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113104723, end_position_on_the_genomic_accession=113111677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128386, CurrentID=0, Status='live', Symbol='LOC100128386', Aliases='', description='uncharacterized LOC100128386', other_designations='CTD-2383M3.1', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=100684161, end_position_on_the_genomic_accession=100687955, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644739, CurrentID=0, Status='live', Symbol='WASF4P', Aliases='WASF4', description='WAS protein family member 4, pseudogene', other_designations='WAS protein family member 2 pseudogene|WAS protein family, member 4, pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47803147, end_position_on_the_genomic_accession=47807533, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371268, CurrentID=0, Status='live', Symbol='LOC105371268', Aliases='', description='uncharacterized LOC105371268', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53510284, end_position_on_the_genomic_accession=53519535, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371269, CurrentID=0, Status='live', Symbol='LOC105371269', Aliases='', description='uncharacterized LOC105371269', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=53544359, end_position_on_the_genomic_accession=53560428, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100128390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128390, CurrentID=0, Status='live', Symbol='CALM2P1', Aliases='', description='calmodulin 2 pseudogene 1', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=70241247, end_position_on_the_genomic_accession=70242378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644743, CurrentID=0, Status='live', Symbol='KRT8P47', Aliases='', description='keratin 8 pseudogene 47', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44103063, end_position_on_the_genomic_accession=44104705, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371272, CurrentID=0, Status='live', Symbol='LOC105371272', Aliases='', description='uncharacterized LOC105371272', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54370408, end_position_on_the_genomic_accession=54374716, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644745, CurrentID=0, Status='live', Symbol='HSPD1P4', Aliases='HSPD1-5P', description='heat shock protein family D (Hsp60) member 1 pseudogene 4', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 4', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56510941, end_position_on_the_genomic_accession=56513190, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371274, CurrentID=0, Status='live', Symbol='LOC105371274', Aliases='', description='uncharacterized LOC105371274', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54634442, end_position_on_the_genomic_accession=54652152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371275, CurrentID=0, Status='live', Symbol='LOC105371275', Aliases='', description='uncharacterized LOC105371275', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54853553, end_position_on_the_genomic_accession=54893566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371276, CurrentID=0, Status='live', Symbol='LOC105371276', Aliases='', description='uncharacterized LOC105371276', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=54847216, end_position_on_the_genomic_accession=54854994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128397, CurrentID=0, Status='live', Symbol='NIFKP5', Aliases='MKI67IPP5', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 5', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 5', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90284323, end_position_on_the_genomic_accession=90285624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128398, CurrentID=0, Status='live', Symbol='LOC100128398', Aliases='', description='uncharacterized LOC100128398', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58002894, end_position_on_the_genomic_accession=58007206, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371279, CurrentID=0, Status='live', Symbol='LOC105371279', Aliases='', description='uncharacterized LOC105371279', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55366723, end_position_on_the_genomic_accession=55369570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128400, CurrentID=0, Status='live', Symbol='LOC100128400', Aliases='', description='uncharacterized LOC100128400', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.9', start_position_on_the_genomic_accession=37866565, end_position_on_the_genomic_accession=37866978, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371281, CurrentID=0, Status='live', Symbol='LOC105371281', Aliases='', description='uncharacterized LOC105371281', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  644754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644754, CurrentID=0, Status='live', Symbol='BOLA3P3', Aliases='', description='bolA family member 3 pseudogene 3', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126663150, end_position_on_the_genomic_accession=126663688, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371283, CurrentID=0, Status='live', Symbol='LOC105371283', Aliases='', description='uncharacterized LOC105371283', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371284, CurrentID=0, Status='live', Symbol='LOC105371284', Aliases='', description='uncharacterized LOC105371284', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56609415, end_position_on_the_genomic_accession=56611920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371285, CurrentID=0, Status='live', Symbol='LOC105371285', Aliases='', description='uncharacterized LOC105371285', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56684104, end_position_on_the_genomic_accession=56687406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128406, CurrentID=0, Status='live', Symbol='AZFP', Aliases='', description='AML-associated zinc finger protein', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=133089230, end_position_on_the_genomic_accession=133090337, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  644759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644759, CurrentID=0, Status='live', Symbol='TMPRSS11GP', Aliases='', description='transmembrane protease, serine 11G, pseudogene', other_designations='serine protease Desc4 pseudogene', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67991812, end_position_on_the_genomic_accession=67997439, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128408, CurrentID=0, Status='live', Symbol='RPL22P10', Aliases='RPL22_3_253', description='ribosomal protein L22 pseudogene 10', other_designations='', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106247371, end_position_on_the_genomic_accession=106247743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780896, CurrentID=0, Status='live', Symbol='ACTD', Aliases='', description='Acetabular dysplasia', other_designations='', map_location='13q22', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=142700, no_hdr0=''),
  100128410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128410, CurrentID=0, Status='live', Symbol='NPM1P34', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 34', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124280435, end_position_on_the_genomic_accession=124281622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928559, CurrentID=0, Status='live', Symbol='LINC01467', Aliases='', description='long intergenic non-protein coding RNA 1467', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81605347, end_position_on_the_genomic_accession=81623061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128412, CurrentID=0, Status='live', Symbol='VTA1P2', Aliases='', description='vesicle (multivesicular body) trafficking 1 pseudogene 2', other_designations='Vps20-associated 1 homolog pseudogene', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87575636, end_position_on_the_genomic_accession=87576532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371293, CurrentID=0, Status='live', Symbol='LOC105371293', Aliases='', description='uncharacterized LOC105371293', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58238573, end_position_on_the_genomic_accession=58249851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371294, CurrentID=0, Status='live', Symbol='LOC105371294', Aliases='', description='uncharacterized LOC105371294', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58295069, end_position_on_the_genomic_accession=58300850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644767, CurrentID=0, Status='live', Symbol='CH17-351M24.1', Aliases='', description='uncharacterized LOC644767', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79598624, end_position_on_the_genomic_accession=79601286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128416, CurrentID=0, Status='live', Symbol='RAB28P5', Aliases='', description='RAB28, member RAS oncogene family pseudogene 5', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136847197, end_position_on_the_genomic_accession=136850125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371297, CurrentID=0, Status='live', Symbol='LOC105371297', Aliases='', description='uncharacterized LOC105371297', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=58684669, end_position_on_the_genomic_accession=58707611, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371298, CurrentID=0, Status='live', Symbol='LOC105371298', Aliases='', description='uncharacterized LOC105371298', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59404810, end_position_on_the_genomic_accession=59463769, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371299, CurrentID=0, Status='live', Symbol='LOC105371299', Aliases='', description='uncharacterized LOC105371299', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=59944170, end_position_on_the_genomic_accession=59971667, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100128420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128420, CurrentID=0, Status='live', Symbol='LINC00892', Aliases='', description='long intergenic non-protein coding RNA 892', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136639543, end_position_on_the_genomic_accession=136642429, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644773, CurrentID=0, Status='live', Symbol='ACTR3BP6', Aliases='', description='ACTR3B pseudogene 6', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 6', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16486693, end_position_on_the_genomic_accession=16488336, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371302, CurrentID=0, Status='live', Symbol='LOC105371302', Aliases='', description='uncharacterized LOC105371302', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61365504, end_position_on_the_genomic_accession=61447716, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128423, CurrentID=0, Status='live', Symbol='RPS24P14', Aliases='RPS24_4_1115', description='ribosomal protein S24 pseudogene 14', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3559512, end_position_on_the_genomic_accession=3560024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371304, CurrentID=0, Status='live', Symbol='LOC105371304', Aliases='', description='uncharacterized LOC105371304', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=61457940, end_position_on_the_genomic_accession=61470116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644777, CurrentID=0, Status='live', Symbol='MEP1AP2', Aliases='MEP1AL2', description='meprin A subunit alpha pseudogene 2', other_designations='meprin A, alpha (PABA peptide hydrolase) pseudogene|meprin A, alpha pseudogene 2|meprin A, alpha-like 2 (pseudogene)', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371306, CurrentID=0, Status='live', Symbol='LOC105371306', Aliases='', description='uncharacterized LOC105371306', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=62679268, end_position_on_the_genomic_accession=62682241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371307, CurrentID=0, Status='live', Symbol='LOC105371307', Aliases='', description='uncharacterized LOC105371307', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63058088, end_position_on_the_genomic_accession=63090302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371308, CurrentID=0, Status='live', Symbol='LOC105371308', Aliases='', description='uncharacterized LOC105371308', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63286914, end_position_on_the_genomic_accession=63618092, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371309, CurrentID=0, Status='live', Symbol='LOC105371309', Aliases='', description='uncharacterized LOC105371309', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63847045, end_position_on_the_genomic_accession=63848553, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371310, CurrentID=0, Status='live', Symbol='LOC105371310', Aliases='', description='uncharacterized LOC105371310', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=64344305, end_position_on_the_genomic_accession=64535040, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100128431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128431, CurrentID=0, Status='live', Symbol='GEMIN8P1', Aliases='FAM51A2P, GEMIN8P', description='gem nuclear organelle associated protein 8 pseudogene 1', other_designations='family with sequence similarity 51, member A2 pseudogene|gem (nuclear organelle) associated protein 8 pseudogene 1', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=69803115, end_position_on_the_genomic_accession=69804142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371312, CurrentID=0, Status='live', Symbol='LOC105371312', Aliases='', description='uncharacterized LOC105371312', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145412897, end_position_on_the_genomic_accession=145422886, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371313, CurrentID=0, Status='live', Symbol='LOC105371313', Aliases='', description='uncharacterized LOC105371313', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=64904799, end_position_on_the_genomic_accession=64916386, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371314, CurrentID=0, Status='live', Symbol='LOC105371314', Aliases='', description='uncharacterized LOC105371314', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=65123755, end_position_on_the_genomic_accession=65138818, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371315, CurrentID=0, Status='live', Symbol='LOC105371315', Aliases='', description='uncharacterized LOC105371315', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66092839, end_position_on_the_genomic_accession=66110408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371316, CurrentID=0, Status='live', Symbol='LOC105371316', Aliases='', description='uncharacterized LOC105371316', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66134971, end_position_on_the_genomic_accession=66146599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371317, CurrentID=0, Status='live', Symbol='LOC105371317', Aliases='', description='uncharacterized LOC105371317', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66238828, end_position_on_the_genomic_accession=66267047, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644790, CurrentID=0, Status='live', Symbol='RPS15AP11', Aliases='RPS15A_3_41', description='ribosomal protein S15a pseudogene 11', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44780285, end_position_on_the_genomic_accession=44780745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371319, CurrentID=0, Status='live', Symbol='LOC105371319', Aliases='', description='uncharacterized LOC105371319', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67142021, end_position_on_the_genomic_accession=67143329, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371320, CurrentID=0, Status='live', Symbol='LOC105371320', Aliases='', description='uncharacterized LOC105371320', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67265114, end_position_on_the_genomic_accession=67279235, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371321, CurrentID=0, Status='live', Symbol='LOC105371321', Aliases='', description='uncharacterized LOC105371321', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68512109, end_position_on_the_genomic_accession=68514920, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652730, CurrentID=0, Status='live', Symbol='LINC00659', Aliases='', description='long intergenic non-protein coding RNA 659', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62774121, end_position_on_the_genomic_accession=62776856, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371323, CurrentID=0, Status='live', Symbol='LOC105371323', Aliases='', description='uncharacterized LOC105371323', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69324574, end_position_on_the_genomic_accession=69325534, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371324, CurrentID=0, Status='live', Symbol='LOC105371324', Aliases='', description='uncharacterized LOC105371324', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145411771, end_position_on_the_genomic_accession=145412825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371325, CurrentID=0, Status='live', Symbol='LOC105371325', Aliases='', description='uncharacterized LOC105371325', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=69522997, end_position_on_the_genomic_accession=69524557, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102465491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102465491, CurrentID=0, Status='live', Symbol='MIR6819', Aliases='hsa-mir-6819', description='microRNA 6819', other_designations='microRNA mir-6819', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36286847, end_position_on_the_genomic_accession=36286907, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371327, CurrentID=0, Status='live', Symbol='LOC105371327', Aliases='', description='uncharacterized LOC105371327', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100652736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652736, CurrentID=0, Status='live', Symbol='LINC01311', Aliases='', description='long intergenic non-protein coding RNA 1311', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19171706, end_position_on_the_genomic_accession=19172832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128449, CurrentID=0, Status='live', Symbol='IL9RP1', Aliases='', description='interleukin 9 receptor pseudogene 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138137629, end_position_on_the_genomic_accession=138141610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128450, CurrentID=0, Status='live', Symbol='RPS6P8', Aliases='RPS6_3_713', description='ribosomal protein S6 pseudogene 8', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73391038, end_position_on_the_genomic_accession=73391767, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652739, CurrentID=0, Status='live', Symbol='RAET1E-AS1', Aliases='RAET1G-AS1', description='RAET1E antisense RNA 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149863498, end_position_on_the_genomic_accession=149919508, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100652740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652740, CurrentID=0, Status='live', Symbol='PYCARD-AS1', Aliases='C16orf98, PYCARDOS', description='PYCARD antisense RNA 1', other_designations='PYCARD opposite strand', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31201885, end_position_on_the_genomic_accession=31203452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371333, CurrentID=0, Status='live', Symbol='LOC105371333', Aliases='', description='uncharacterized LOC105371333', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71410315, end_position_on_the_genomic_accession=71425946, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371334, CurrentID=0, Status='live', Symbol='LOC105371334', Aliases='', description='uncharacterized LOC105371334', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71583769, end_position_on_the_genomic_accession=71621373, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371335, CurrentID=0, Status='live', Symbol='LOC105371335', Aliases='', description='uncharacterized LOC105371335', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=71513811, end_position_on_the_genomic_accession=71517852, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128456, CurrentID=0, Status='live', Symbol='CCNB1IP1P3', Aliases='', description='cyclin B1 interacting protein 1 pseudogene 3', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=94776550, end_position_on_the_genomic_accession=94777801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644809, CurrentID=0, Status='live', Symbol='C15orf56', Aliases='', description='chromosome 15 open reading frame 56', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40250664, end_position_on_the_genomic_accession=40252909, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371338, CurrentID=0, Status='live', Symbol='LOC105371338', Aliases='', description='uncharacterized LOC105371338', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=72806104, end_position_on_the_genomic_accession=72812010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128459, CurrentID=0, Status='live', Symbol='CCNJP1', Aliases='CCNJP', description='cyclin J pseudogene 1', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56228631, end_position_on_the_genomic_accession=56229782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652748, CurrentID=0, Status='live', Symbol='TIMM23B', Aliases='TIMM23, bA592B15.7', description='translocase of inner mitochondrial membrane 23 homolog B (yeast)', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49942033, end_position_on_the_genomic_accession=49988839, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100652749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652749, CurrentID=0, Status='live', Symbol='SPATA8-AS1', Aliases='', description='SPATA8 antisense RNA 1 (head to head)', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96772005, end_position_on_the_genomic_accession=96783337, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  120526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120526, CurrentID=0, Status='live', Symbol='DNAJC24', Aliases='DPH4, JJJ3, ZCSL3', description='DnaJ heat shock protein family (Hsp40) member C24', other_designations='1700030A21Rik|CSL-type zinc finger-containing protein 3|DPH4 homolog (JJJ3, S. cerevisiae)|DPH4, JJJ3 homolog|DnaJ (Hsp40) homolog, subfamily C, member 24|diphthamide biosynthesis protein 4|zinc finger, CSL domain containing 3|zinc finger, CSL-type containing 3', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31369815, end_position_on_the_genomic_accession=31432835, orientation='plus', exon_count=6, OMIM=611072, no_hdr0=''),
  644815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644815, CurrentID=0, Status='live', Symbol='FAM83G', Aliases='PAWS1', description='family with sequence similarity 83 member G', other_designations='family with sequence similarity 83, member G|protein associated with SMAD1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18971068, end_position_on_the_genomic_accession=19004747, orientation='minus', exon_count=6, OMIM=615886, no_hdr0=''),
  120528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120528, CurrentID=0, Status='live', Symbol='CYCSP25', Aliases='HC10, HCP25', description='cytochrome c, somatic pseudogene 25', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=31280668, end_position_on_the_genomic_accession=31280987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128465, CurrentID=0, Status='live', Symbol='IFIT6P', Aliases='', description='interferon induced protein with tetratricopeptide repeats 6, pseudogene', other_designations='interferon-induced protein with tetratricopeptide repeats 6, pseudogene', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89362315, end_position_on_the_genomic_accession=89363468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371346, CurrentID=0, Status='live', Symbol='CPHXL', Aliases='', description='cytoplasmic polyadenylated homeobox like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75714024, end_position_on_the_genomic_accession=75732740, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371347, CurrentID=0, Status='live', Symbol='LOC105371347', Aliases='', description='uncharacterized LOC105371347', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75835347, end_position_on_the_genomic_accession=75860845, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644820, CurrentID=0, Status='live', Symbol='EEF1B2P3', Aliases='EEF1B4', description='eukaryotic translation elongation factor 1 beta 2 pseudogene 3', other_designations='eukaryotic translation elongation factor 1 beta 4 (pseudogene)|eukaryotic translation elongation factor 1 beta 4 pseudogene|hCG1983058', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24788347, end_position_on_the_genomic_accession=24789092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128469, CurrentID=0, Status='live', Symbol='SNRPCP3', Aliases='', description='small nuclear ribonucleoprotein polypeptide C pseudogene 3', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150719753, end_position_on_the_genomic_accession=150720841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  120534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120534, CurrentID=0, Status='live', Symbol='ARL14EP', Aliases='ARF7EP, C11orf46, dJ299F11.1', description='ADP ribosylation factor like GTPase 14 effector protein', other_designations='ADP-ribosylation factor like GTPase 14 effector protein|ADP-ribosylation factor-like 14 effector protein|ARF7 effector protein', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30323099, end_position_on_the_genomic_accession=30338458, orientation='plus', exon_count=5, OMIM=612295, no_hdr0=''),
  100652759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652759, CurrentID=0, Status='live', Symbol='PRICKLE2-AS1', Aliases='', description='PRICKLE2 antisense RNA 1', other_designations='PRICKLE2 antisense RNA 1 (non-protein coding)', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64067964, end_position_on_the_genomic_accession=64103131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371352, CurrentID=0, Status='live', Symbol='LOC105371352', Aliases='', description='uncharacterized LOC105371352', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79202624, end_position_on_the_genomic_accession=79211600, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371353, CurrentID=0, Status='live', Symbol='LOC105371353', Aliases='', description='uncharacterized LOC105371353', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78793581, end_position_on_the_genomic_accession=78796778, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371354, CurrentID=0, Status='live', Symbol='LOC105371354', Aliases='', description='uncharacterized LOC105371354', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376627, CurrentID=0, Status='live', Symbol='LOC105376627', Aliases='', description='uncharacterized LOC105376627', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35086800, end_position_on_the_genomic_accession=35097559, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371356, CurrentID=0, Status='live', Symbol='LOC105371356', Aliases='', description='uncharacterized LOC105371356', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=79675954, end_position_on_the_genomic_accession=79747103, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371357, CurrentID=0, Status='live', Symbol='LOC105371357', Aliases='', description='uncharacterized LOC105371357', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80051344, end_position_on_the_genomic_accession=80168480, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371358, CurrentID=0, Status='live', Symbol='LOC105371358', Aliases='', description='uncharacterized LOC105371358', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80080692, end_position_on_the_genomic_accession=80109575, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371359, CurrentID=0, Status='live', Symbol='LOC105371359', Aliases='', description='uncharacterized LOC105371359', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80257846, end_position_on_the_genomic_accession=80265764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371360, CurrentID=0, Status='live', Symbol='LOC105371360', Aliases='', description='uncharacterized LOC105371360', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80551855, end_position_on_the_genomic_accession=80555068, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371361, CurrentID=0, Status='live', Symbol='LOC105371361', Aliases='', description='uncharacterized LOC105371361', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81071690, end_position_on_the_genomic_accession=81076598, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644834, CurrentID=0, Status='live', Symbol='RPL36AP36', Aliases='RPL36A_13_1022', description='ribosomal protein L36a pseudogene 36', other_designations='', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13159421, end_position_on_the_genomic_accession=13159817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128483, CurrentID=0, Status='live', Symbol='TPT1P2', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 2', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208214696, end_position_on_the_genomic_accession=208222625, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652772, CurrentID=0, Status='live', Symbol='NNT-AS1', Aliases='', description='NNT antisense RNA 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43573185, end_position_on_the_genomic_accession=43603230, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391462, CurrentID=0, Status='live', Symbol='RPSAP24', Aliases='RPSA_6_295', description='ribosomal protein SA pseudogene 24', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174042933, end_position_on_the_genomic_accession=174043954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644838, CurrentID=0, Status='live', Symbol='LOC644838', Aliases='', description='uncharacterized LOC644838', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67123357, end_position_on_the_genomic_accession=67215319, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100652775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652775, CurrentID=0, Status='live', Symbol='HP11026', Aliases='', description='uncharacterized protein HP11026', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=32288081, end_position_on_the_genomic_accession=32289305, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371368, CurrentID=0, Status='live', Symbol='LOC105371368', Aliases='', description='uncharacterized LOC105371368', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83382865, end_position_on_the_genomic_accession=83399617, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100652777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652777, CurrentID=0, Status='live', Symbol='LOC100652777', Aliases='', description='group 10 secretory phospholipase A2-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14709070, end_position_on_the_genomic_accession=14733781, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371370, CurrentID=0, Status='live', Symbol='LOC105371370', Aliases='', description='uncharacterized LOC105371370', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83816441, end_position_on_the_genomic_accession=83831774, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371371, CurrentID=0, Status='live', Symbol='LOC105371371', Aliases='', description='SH3 domain-binding protein 1-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=83931236, end_position_on_the_genomic_accession=83951460, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  644844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644844, CurrentID=0, Status='live', Symbol='PHGR1', Aliases='', description='proline/histidine/glycine-rich 1', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40351033, end_position_on_the_genomic_accession=40356434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100652781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652781, CurrentID=0, Status='live', Symbol='SNX29P1', Aliases='RUNDC2B, RUNDC2L', description='sorting nexin 29 pseudogene 1', other_designations='Putative RUN domain-containing protein 2B|RUN domain containing 2B|RUNDC2-like protein|Sorting nexin 29 protein pseudogene 1', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21349363, end_position_on_the_genomic_accession=21385933, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371374, CurrentID=0, Status='live', Symbol='LOC105371374', Aliases='', description='uncharacterized LOC105371374', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84460654, end_position_on_the_genomic_accession=84468407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371375, CurrentID=0, Status='live', Symbol='LOC105371375', Aliases='', description='uncharacterized LOC105371375', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84504845, end_position_on_the_genomic_accession=84507679, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128496, CurrentID=0, Status='live', Symbol='C20orf78', Aliases='dJ1068E13.1', description='chromosome 20 open reading frame 78', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18809351, end_position_on_the_genomic_accession=18830203, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128497, CurrentID=0, Status='live', Symbol='LOC100128497', Aliases='', description='mitotic spindle assembly checkpoint protein MAD2A-like', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=76878272, end_position_on_the_genomic_accession=76928085, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644850, CurrentID=0, Status='live', Symbol='PDCL3P5', Aliases='', description='phosducin-like 3 pseudogene 5', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150827625, end_position_on_the_genomic_accession=150828620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371379, CurrentID=0, Status='live', Symbol='LOC105371379', Aliases='', description='uncharacterized LOC105371379', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84928220, end_position_on_the_genomic_accession=84943056, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371380, CurrentID=0, Status='live', Symbol='LOC105371380', Aliases='', description='uncharacterized LOC105371380', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85079345, end_position_on_the_genomic_accession=85084100, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371381, CurrentID=0, Status='live', Symbol='LOC105371381', Aliases='', description='uncharacterized LOC105371381', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85119962, end_position_on_the_genomic_accession=85124164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371382, CurrentID=0, Status='live', Symbol='LOC105371382', Aliases='', description='uncharacterized LOC105371382', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85146384, end_position_on_the_genomic_accession=85168989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100652791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652791, CurrentID=0, Status='live', Symbol='GS1-24F4.2', Aliases='', description='uncharacterized LOC100652791', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6835554, end_position_on_the_genomic_accession=6842454, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371384, CurrentID=0, Status='live', Symbol='LOC105371384', Aliases='', description='uncharacterized LOC105371384', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85575901, end_position_on_the_genomic_accession=85583486, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128505, CurrentID=0, Status='live', Symbol='ASTN2-AS1', Aliases='', description='ASTN2 antisense RNA 1', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116504283, end_position_on_the_genomic_accession=116562293, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105371386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371386, CurrentID=0, Status='live', Symbol='LOC105371386', Aliases='', description='uncharacterized LOC105371386', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85449116, end_position_on_the_genomic_accession=85465819, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371387, CurrentID=0, Status='live', Symbol='LOC105371387', Aliases='', description='uncharacterized LOC105371387', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=85182204, end_position_on_the_genomic_accession=85185673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100652796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652796, CurrentID=0, Status='live', Symbol='BOLA2P1', Aliases='', description='bolA family member 2 pseudogene 1', other_designations='bolA family member 2B pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=832143, end_position_on_the_genomic_accession=832512, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644861, CurrentID=0, Status='live', Symbol='NBPF13P', Aliases='', description='neuroblastoma breakpoint family member 13, pseudogene', other_designations='neuroblastoma breakpoint family, member 13 (pseudogene)|neuroblastoma breakpoint family, member 13, pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147099482, end_position_on_the_genomic_accession=147114346, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  644862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644862, CurrentID=0, Status='live', Symbol='RPS28P3', Aliases='RPS28_2_172', description='ribosomal protein S28 pseudogene 3', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228134782, end_position_on_the_genomic_accession=228136793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128511, CurrentID=0, Status='live', Symbol='NEBL-AS1', Aliases='', description='NEBL antisense RNA 1', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21173990, end_position_on_the_genomic_accession=21174923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128512, CurrentID=0, Status='live', Symbol='FXYD6P2', Aliases='', description='FXYD domain containing ion transport regulator 6 pseudogene 2', other_designations='FXYD domain containing ion transport regulator 6, pseudogene 2|FXYD domain-containing ion transport regulator 6, pseudogene 2', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38091764, end_position_on_the_genomic_accession=38092051, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652801, CurrentID=0, Status='live', Symbol='NXT1P1', Aliases='', description='NTF2-like export factor 1 pseudogene 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39169632, end_position_on_the_genomic_accession=39169950, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128514, CurrentID=0, Status='live', Symbol='ARHGAP42P5', Aliases='', description='Rho GTPase activating protein 42 pseudogene 5', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18622686, end_position_on_the_genomic_accession=18626786, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371395, CurrentID=0, Status='live', Symbol='LOC105371395', Aliases='', description='uncharacterized LOC105371395', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128516, CurrentID=0, Status='live', Symbol='ATP5F1P5', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 5', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 5', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122960327, end_position_on_the_genomic_accession=122961085, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371397, CurrentID=0, Status='live', Symbol='LOC105371397', Aliases='', description='uncharacterized LOC105371397', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87700832, end_position_on_the_genomic_accession=87703017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371398, CurrentID=0, Status='live', Symbol='LOC105371398', Aliases='', description='uncharacterized LOC105371398', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87678451, end_position_on_the_genomic_accession=87679463, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128519, CurrentID=0, Status='live', Symbol='RPS24P11', Aliases='RPS24_6_457', description='ribosomal protein S24 pseudogene 11', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9553035, end_position_on_the_genomic_accession=9554127, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371400, CurrentID=0, Status='live', Symbol='LOC105371400', Aliases='', description='uncharacterized LOC105371400', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=87881549, end_position_on_the_genomic_accession=87886124, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644873, CurrentID=0, Status='live', Symbol='LINC01184', Aliases='', description='long intergenic non-protein coding RNA 1184', other_designations='CTC-228N24.3', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128021551, end_position_on_the_genomic_accession=128083074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  120586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120586, CurrentID=0, Status='live', Symbol='OR8I2', Aliases='OR11-170', description='olfactory receptor family 8 subfamily I member 2', other_designations='olfactory receptor OR11-170|olfactory receptor, family 8, subfamily I, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56093308, end_position_on_the_genomic_accession=56094240, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371403, CurrentID=0, Status='live', Symbol='LOC105371403', Aliases='', description='uncharacterized LOC105371403', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149383616, end_position_on_the_genomic_accession=149390590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371404, CurrentID=0, Status='live', Symbol='LOC105371404', Aliases='', description='uncharacterized LOC105371404', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128525, CurrentID=0, Status='live', Symbol='UQCRFS1P1', Aliases='UQCRFSL1', description='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide 1 pseudogene 1', other_designations='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide-like 1', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=39875296, end_position_on_the_genomic_accession=39876308, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371406, CurrentID=0, Status='live', Symbol='LOC105371406', Aliases='', description='uncharacterized LOC105371406', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149734940, end_position_on_the_genomic_accession=149740512, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644881, CurrentID=0, Status='live', Symbol='RARRES2P4', Aliases='', description='retinoic acid receptor responder (tazarotene induced) 2 pseudogene 4', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190037796, end_position_on_the_genomic_accession=190038423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371410, CurrentID=0, Status='live', Symbol='LOC105371410', Aliases='', description='uncharacterized LOC105371410', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88870754, end_position_on_the_genomic_accession=88872216, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128531, CurrentID=0, Status='live', Symbol='LOC100128531', Aliases='', description='uncharacterized LOC100128531', other_designations='CTA-221G9.11', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25102417, end_position_on_the_genomic_accession=25112692, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371412, CurrentID=0, Status='live', Symbol='LOC105371412', Aliases='', description='uncharacterized LOC105371412', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89071039, end_position_on_the_genomic_accession=89073455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371413, CurrentID=0, Status='live', Symbol='LOC105371413', Aliases='', description='uncharacterized LOC105371413', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89254591, end_position_on_the_genomic_accession=89257193, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371414, CurrentID=0, Status='live', Symbol='LOC105371414', Aliases='', description='keratin-associated protein 5-2-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89296085, end_position_on_the_genomic_accession=89300962, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128535, CurrentID=0, Status='live', Symbol='RPS18P14', Aliases='RPS18_10_1747', description='ribosomal protein S18 pseudogene 14', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31559442, end_position_on_the_genomic_accession=31559977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100652824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652824, CurrentID=0, Status='live', Symbol='KIAA2012', Aliases='', description='KIAA2012', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202073253, end_position_on_the_genomic_accession=202205195, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  100128537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128537, CurrentID=0, Status='live', Symbol='C1orf132', Aliases='', description='chromosome 1 open reading frame 132', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207761719, end_position_on_the_genomic_accession=207869072, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  644890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644890, CurrentID=0, Status='live', Symbol='MEIG1', Aliases='SPATA39, bA2K17.3', description='meiosis/spermiogenesis associated 1', other_designations='meiosis expressed gene 1 homolog|spermatogenesis associated 39', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14954228, end_position_on_the_genomic_accession=14972857, orientation='plus', exon_count=4, OMIM=614174, no_hdr0=''),
  105371419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371419, CurrentID=0, Status='live', Symbol='LOC105371419', Aliases='', description='uncharacterized LOC105371419', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89846860, end_position_on_the_genomic_accession=89854276, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371420, CurrentID=0, Status='live', Symbol='LOC105371420', Aliases='', description='uncharacterized LOC105371420', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149615928, end_position_on_the_genomic_accession=149625411, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128541, CurrentID=0, Status='live', Symbol='ATP6V1G1P2', Aliases='ATP6V1GP2', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G1 pseudogene 2', other_designations='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G pseudogene 2', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47194003, end_position_on_the_genomic_accession=47194349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128542, CurrentID=0, Status='live', Symbol='BET1P1', Aliases='', description='Bet1 golgi vesicular membrane trafficking protein pseudogene 1', other_designations='blocked early in transport 1 homolog pseudogene', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150749596, end_position_on_the_genomic_accession=150751538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100652831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652831, CurrentID=0, Status='live', Symbol='NANOGNBP3', Aliases='', description='NANOGNB pseudogene 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101337312, end_position_on_the_genomic_accession=101338207, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644896, CurrentID=0, Status='live', Symbol='RPL7P7', Aliases='RPL7_1_7', description='ribosomal protein L7 pseudogene 7', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8786129, end_position_on_the_genomic_accession=8786951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371425, CurrentID=0, Status='live', Symbol='LOC105371425', Aliases='', description='uncharacterized LOC105371425', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=338869, end_position_on_the_genomic_accession=341409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371426, CurrentID=0, Status='live', Symbol='LOC105371426', Aliases='', description='uncharacterized LOC105371426', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=321423, end_position_on_the_genomic_accession=322457, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100652835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652835, CurrentID=0, Status='live', Symbol='RBSG2', Aliases='', description='retinoblastoma-specific gene 2', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=139084423, end_position_on_the_genomic_accession=139085476, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371428, CurrentID=0, Status='live', Symbol='LOC105371428', Aliases='', description='uncharacterized LOC105371428', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=252818, end_position_on_the_genomic_accession=268394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371429, CurrentID=0, Status='live', Symbol='LOC105371429', Aliases='', description='uncharacterized LOC105371429', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371430, CurrentID=0, Status='live', Symbol='LOC105371430', Aliases='', description='mucin-17-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=401242, end_position_on_the_genomic_accession=417165, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644903, CurrentID=0, Status='live', Symbol='GCC2-AS1', Aliases='', description='GCC2 antisense RNA 1', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108505162, end_position_on_the_genomic_accession=108534219, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371432, CurrentID=0, Status='live', Symbol='LOC105371432', Aliases='', description='uncharacterized LOC105371432', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=354716, end_position_on_the_genomic_accession=357581, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128553, CurrentID=0, Status='live', Symbol='CTAGE4', Aliases='cTAGE-4', description='CTAGE family member 4', other_designations='CTAGE family, member 4|cutaneous T-cell lymphoma-associated antigen 4|cutaneous-T-cell-lymphoma-specific tumor antigen 4|protein cTAGE-4', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144183455, end_position_on_the_genomic_accession=144186080, orientation='plus', exon_count=1, OMIM=608910, no_hdr0=''),
  100128554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128554, CurrentID=0, Status='live', Symbol='LOC100128554', Aliases='', description='uncharacterized LOC100128554', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126442481, end_position_on_the_genomic_accession=126472785, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  644907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644907, CurrentID=0, Status='live', Symbol='RPL32P18', Aliases='RPL32_8_807', description='ribosomal protein L32 pseudogene 18', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44467843, end_position_on_the_genomic_accession=44468981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371436, CurrentID=0, Status='live', Symbol='LOC105371436', Aliases='', description='uncharacterized LOC105371436', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150521058, end_position_on_the_genomic_accession=150535829, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371437, CurrentID=0, Status='live', Symbol='LOC105371437', Aliases='', description='uncharacterized LOC105371437', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150550283, end_position_on_the_genomic_accession=150551350, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100652846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652846, CurrentID=0, Status='live', Symbol='CACNA1C-AS1', Aliases='', description='CACNA1C antisense RNA 1', other_designations='CACNA1C antisense RNA 1 (non-protein coding)', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2676001, end_position_on_the_genomic_accession=2691157, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  644911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644911, CurrentID=0, Status='live', Symbol='RPL10P5', Aliases='RPL10_1_257', description='ribosomal protein L10 pseudogene 5', other_designations='', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108712174, end_position_on_the_genomic_accession=108712923, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128560, CurrentID=0, Status='live', Symbol='MYHAS', Aliases='', description='myosin heavy chain gene cluster antisense RNA', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10383132, end_position_on_the_genomic_accession=10625540, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105371441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371441, CurrentID=0, Status='live', Symbol='LOC105371441', Aliases='', description='uncharacterized LOC105371441', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151987235, end_position_on_the_genomic_accession=152010601, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  644914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644914, CurrentID=0, Status='live', Symbol='H3F3AP6', Aliases='p21', description='H3 histone, family 3A, pseudogene 6', other_designations='H3 histone pseudogene|H3 histone, family 3A pseudogene', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139697626, end_position_on_the_genomic_accession=139698635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391475, CurrentID=0, Status='live', Symbol='DYTN', Aliases='', description='dystrotelin', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206651621, end_position_on_the_genomic_accession=206718396, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105371444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371444, CurrentID=0, Status='live', Symbol='LOC105371444', Aliases='', description='uncharacterized LOC105371444', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152861509, end_position_on_the_genomic_accession=152868176, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652853, CurrentID=0, Status='live', Symbol='LINC01023', Aliases='HP07349', description='long intergenic non-protein coding RNA 1023', other_designations='CTC-329H14.4', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108727825, end_position_on_the_genomic_accession=108728261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371446, CurrentID=0, Status='live', Symbol='LOC105371446', Aliases='', description='uncharacterized LOC105371446', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153011132, end_position_on_the_genomic_accession=153024583, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644919, CurrentID=0, Status='live', Symbol='LOC644919', Aliases='', description='uncharacterized LOC644919', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=40954711, end_position_on_the_genomic_accession=41141048, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100652856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652856, CurrentID=0, Status='live', Symbol='LINC00408', Aliases='', description='long intergenic non-protein coding RNA 408', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18905439, end_position_on_the_genomic_accession=18926741, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100128569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128569, CurrentID=0, Status='live', Symbol='C19orf71', Aliases='', description='chromosome 19 open reading frame 71', other_designations='Uncharacterized protein ENSP00000327950', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3539157, end_position_on_the_genomic_accession=3544030, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371450, CurrentID=0, Status='live', Symbol='LOC105371450', Aliases='', description='uncharacterized LOC105371450', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155195289, end_position_on_the_genomic_accession=155201949, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371451, CurrentID=0, Status='live', Symbol='LOC105371451', Aliases='', description='uncharacterized LOC105371451', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155313708, end_position_on_the_genomic_accession=155317353, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644924, CurrentID=0, Status='live', Symbol='LOC644924', Aliases='', description='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55961208, end_position_on_the_genomic_accession=55962479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371453, CurrentID=0, Status='live', Symbol='LOC105371453', Aliases='', description='uncharacterized LOC105371453', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156645725, end_position_on_the_genomic_accession=156661441, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128574, CurrentID=0, Status='live', Symbol='ARMCX3-AS1', Aliases='', description='ARMCX3 antisense RNA 1', other_designations='ARMCX3 antisense RNA 1 (non-protein coding)', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=100877747, end_position_on_the_genomic_accession=100879148, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371455, CurrentID=0, Status='live', Symbol='LOC105371455', Aliases='', description='uncharacterized LOC105371455', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157275009, end_position_on_the_genomic_accession=157283617, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644928, CurrentID=0, Status='live', Symbol='RPS26P15', Aliases='RPS26_1_56', description='ribosomal protein S26 pseudogene 15', other_designations='', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58056050, end_position_on_the_genomic_accession=58056504, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652865, CurrentID=0, Status='live', Symbol='LINC00539', Aliases='LINC00422', description='long intergenic non-protein coding RNA 539', other_designations='long intergenic non-protein coding RNA 422', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21303512, end_position_on_the_genomic_accession=21344859, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371458, CurrentID=0, Status='live', Symbol='LOC105371458', Aliases='', description='uncharacterized LOC105371458', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157948476, end_position_on_the_genomic_accession=157950921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128579, CurrentID=0, Status='live', Symbol='CDKN2AIPNLP2', Aliases='', description='CDKN2A interacting protein N-terminal like pseudogene 2', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26826788, end_position_on_the_genomic_accession=26827532, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371460, CurrentID=0, Status='live', Symbol='LOC105371460', Aliases='', description='uncharacterized LOC105371460', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158140725, end_position_on_the_genomic_accession=158170029, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371461, CurrentID=0, Status='live', Symbol='LOC105371461', Aliases='', description='uncharacterized LOC105371461', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159056127, end_position_on_the_genomic_accession=159057750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  644934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644934, CurrentID=0, Status='live', Symbol='RPS26P50', Aliases='RPS26_18_1443', description='ribosomal protein S26 pseudogene 50', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64592955, end_position_on_the_genomic_accession=64593396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644935, CurrentID=0, Status='live', Symbol='ELF2P4', Aliases='', description='E74-like factor 2 pseudogene 4', other_designations='E74-like factor 2 (ets domain transcription factor) pseudogene|ELF2P4 pseudogene', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185545722, end_position_on_the_genomic_accession=185547818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  644936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644936, CurrentID=0, Status='live', Symbol='LOC644936', Aliases='', description='actin, beta pseudogene', other_designations='cytoplasmic beta-actin pseudogene', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80299085, end_position_on_the_genomic_accession=80300478, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  644937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644937, CurrentID=0, Status='live', Symbol='RPL10P8', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171366307, end_position_on_the_genomic_accession=171366675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371466, CurrentID=0, Status='live', Symbol='LOC105371466', Aliases='', description='uncharacterized LOC105371466', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160375506, end_position_on_the_genomic_accession=160408444, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371467, CurrentID=0, Status='live', Symbol='LOC105371467', Aliases='', description='uncharacterized LOC105371467', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160432865, end_position_on_the_genomic_accession=160438177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128588, CurrentID=0, Status='live', Symbol='LOC100128588', Aliases='', description='PIN2/TERF1 interacting, telomerase inhibitor 1 pseudogene', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112325509, end_position_on_the_genomic_accession=112326865, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128589, CurrentID=0, Status='live', Symbol='GAPDHP39', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 39', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138777760, end_position_on_the_genomic_accession=138779011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128590, CurrentID=0, Status='live', Symbol='SLC8A1-AS1', Aliases='', description='SLC8A1 antisense RNA 1', other_designations='SLC8A1 antisense RNA 1 (non-protein coding)', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39917634, end_position_on_the_genomic_accession=40255209, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  644943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644943, CurrentID=0, Status='live', Symbol='RASSF10', Aliases='', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 10', other_designations='', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13009423, end_position_on_the_genomic_accession=13012106, orientation='plus', exon_count=1, OMIM=614713, no_hdr0=''),
  105371472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371472, CurrentID=0, Status='live', Symbol='LOC105371472', Aliases='', description='uncharacterized LOC105371472', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161279465, end_position_on_the_genomic_accession=161281773, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  644945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644945, CurrentID=0, Status='live', Symbol='KRT16P3', Aliases='KERSMCR', description='keratin 16 pseudogene 3', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20501513, end_position_on_the_genomic_accession=20504498, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100128594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128594, CurrentID=0, Status='live', Symbol='LL0XNC01-116E7.2', Aliases='', description='uncharacterized LOC100128594', other_designations='', map_location='Xq22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103916984, end_position_on_the_genomic_accession=103920025, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100652883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652883, CurrentID=0, Status='live', Symbol='ALOX15P1', Aliases='ALOX15P', description='arachidonate 15-lipoxygenase pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6657719, end_position_on_the_genomic_accession=6661382, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100128596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128596, CurrentID=0, Status='live', Symbol='NDUFB9P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B9 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 9, 22kDa pseudogene', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131753648, end_position_on_the_genomic_accession=131754295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100652885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652885, CurrentID=0, Status='live', Symbol='SIRPB3P', Aliases='', description='signal-regulatory protein beta 3, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1668590, end_position_on_the_genomic_accession=1671240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371478, CurrentID=0, Status='live', Symbol='LOC105371478', Aliases='', description='breakpoint cluster region protein-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1225891, end_position_on_the_genomic_accession=1229724, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371479, CurrentID=0, Status='live', Symbol='LOC105371479', Aliases='', description='uncharacterized LOC105371479', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1180411, end_position_on_the_genomic_accession=1190368, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371480, CurrentID=0, Status='live', Symbol='LOC105371480', Aliases='', description='uncharacterized LOC105371480', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1148235, end_position_on_the_genomic_accession=1153884, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371481, CurrentID=0, Status='live', Symbol='LOC105371481', Aliases='', description='uncharacterized LOC105371481', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1013172, end_position_on_the_genomic_accession=1014081, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371482, CurrentID=0, Status='live', Symbol='LOC105371482', Aliases='', description='uncharacterized LOC105371482', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1007319, end_position_on_the_genomic_accession=1009390, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128603, CurrentID=0, Status='live', Symbol='C1GALT1P2', Aliases='', description='ore 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase 1 pseudogene 2', other_designations='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase 1 pseudogene|core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase, 1 pseudogene', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56192368, end_position_on_the_genomic_accession=56193835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371484, CurrentID=0, Status='live', Symbol='LOC105371484', Aliases='', description='uncharacterized LOC105371484', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1620360, end_position_on_the_genomic_accession=1626014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128605, CurrentID=0, Status='live', Symbol='RPS15AP13', Aliases='RPS15A_3_241', description='ribosomal protein S15a pseudogene 13', other_designations='', map_location='2p13.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72288732, end_position_on_the_genomic_accession=72289115, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371486, CurrentID=0, Status='live', Symbol='LOC105371486', Aliases='', description='uncharacterized LOC105371486', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128607, CurrentID=0, Status='live', Symbol='LOC100128607', Aliases='', description='uncharacterized LOC100128607', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=64530734, end_position_on_the_genomic_accession=64531174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371488, CurrentID=0, Status='live', Symbol='LOC105371488', Aliases='', description='uncharacterized LOC105371488', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2381800, end_position_on_the_genomic_accession=2384325, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644961, CurrentID=0, Status='live', Symbol='ACTG1P20', Aliases='', description='actin gamma 1 pseudogene 20', other_designations='actin, beta pseudogene|actin, gamma 1 pseudogene', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27323874, end_position_on_the_genomic_accession=27326525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  644962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644962, CurrentID=0, Status='live', Symbol='TNRC18P1', Aliases='', description='trinucleotide repeat containing 18 pseudogene 1', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140641191, end_position_on_the_genomic_accession=140643080, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100652899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652899, CurrentID=0, Status='live', Symbol='ANP32BP3', Aliases='', description='acidic nuclear phosphoprotein 32 family member B pseudogene 3', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member B pseudogene 3', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=81117139, end_position_on_the_genomic_accession=81118340, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371492, CurrentID=0, Status='live', Symbol='LOC105371492', Aliases='', description='uncharacterized LOC105371492', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3652550, end_position_on_the_genomic_accession=3657998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371493, CurrentID=0, Status='live', Symbol='LOC105371493', Aliases='', description='uncharacterized LOC105371493', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3643109, end_position_on_the_genomic_accession=3651755, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371494, CurrentID=0, Status='live', Symbol='LOC105371494', Aliases='', description='uncharacterized LOC105371494', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3697158, end_position_on_the_genomic_accession=3719024, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371495, CurrentID=0, Status='live', Symbol='LOC105371495', Aliases='', description='uncharacterized LOC105371495', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3846664, end_position_on_the_genomic_accession=3853318, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371496, CurrentID=0, Status='live', Symbol='LOC105371496', Aliases='', description='uncharacterized LOC105371496', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3977060, end_position_on_the_genomic_accession=3983089, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128617, CurrentID=0, Status='live', Symbol='UBA52P7', Aliases='RPL40_2_1281', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 7', other_designations='ribosomal protein L40 pseudogene', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=115592604, end_position_on_the_genomic_accession=115592983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128618, CurrentID=0, Status='live', Symbol='SLC25A3P2', Aliases='', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 3 pseudogene 2', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111992938, end_position_on_the_genomic_accession=111994240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371499, CurrentID=0, Status='live', Symbol='LOC105371499', Aliases='', description='uncharacterized LOC105371499', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4645322, end_position_on_the_genomic_accession=4665775, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  644972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644972, CurrentID=0, Status='live', Symbol='RPS3AP26', Aliases='RPS3A_11_824', description='ribosomal protein S3a pseudogene 26', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98385750, end_position_on_the_genomic_accession=98386607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128621, CurrentID=0, Status='live', Symbol='CYB5AP2', Aliases='CYB5P2', description='cytochrome b5 type A (microsomal) pseudogene 2', other_designations='cytochrome b-5 pseudogene 2', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169790057, end_position_on_the_genomic_accession=169790375, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644974, CurrentID=0, Status='live', Symbol='ALG1L2', Aliases='', description='ALG1, chitobiosyldiphosphodolichol beta-mannosyltransferase-like 2', other_designations='asparagine-linked glycosylation 1-like 2', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130081831, end_position_on_the_genomic_accession=130098467, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  644975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644975, CurrentID=0, Status='live', Symbol='FLJ30064', Aliases='', description='uncharacterized LOC644975', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=97937314, end_position_on_the_genomic_accession=97937712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371504, CurrentID=0, Status='live', Symbol='LOC105371504', Aliases='', description='uncharacterized LOC105371504', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128625, CurrentID=0, Status='live', Symbol='SRSF1P1', Aliases='', description='serine/arginine-rich splicing factor 1 pseudogene 1', other_designations='splicing factor, arginine/serine-rich 1 pseudogene', map_location='13q21.33', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69408130, end_position_on_the_genomic_accession=69408735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128626, CurrentID=0, Status='live', Symbol='PPP1R2P10', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 10', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 10', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63810349, end_position_on_the_genomic_accession=63811930, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128627, CurrentID=0, Status='live', Symbol='CDC42P3', Aliases='', description='cell division cycle 42 pseudogene 3', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142995828, end_position_on_the_genomic_accession=142997663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128628, CurrentID=0, Status='live', Symbol='RGS17P1', Aliases='', description='regulator of G-protein signaling 17 pseudogene 1', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40992736, end_position_on_the_genomic_accession=40993331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371509, CurrentID=0, Status='live', Symbol='LOC105371509', Aliases='', description='uncharacterized LOC105371509', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6342519, end_position_on_the_genomic_accession=6375012, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100652918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652918, CurrentID=0, Status='live', Symbol='GAPDHP75', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 75', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=189132551, end_position_on_the_genomic_accession=189133289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371511, CurrentID=0, Status='live', Symbol='LOC105371511', Aliases='', description='uncharacterized LOC105371511', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7116485, end_position_on_the_genomic_accession=7118010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371512, CurrentID=0, Status='live', Symbol='LOC105371512', Aliases='', description='uncharacterized LOC105371512', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7131264, end_position_on_the_genomic_accession=7140890, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371513, CurrentID=0, Status='live', Symbol='LOC105371513', Aliases='', description='uncharacterized LOC105371513', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7307186, end_position_on_the_genomic_accession=7307867, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371514, CurrentID=0, Status='live', Symbol='LOC105371514', Aliases='', description='uncharacterized LOC105371514', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371515, CurrentID=0, Status='live', Symbol='LOC105371515', Aliases='', description='uncharacterized LOC105371515', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162990522, end_position_on_the_genomic_accession=163006645, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  644988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644988, CurrentID=0, Status='live', Symbol='ABHD17AP6', Aliases='FAM108A7P', description='abhydrolase domain containing 17A pseudogene 6', other_designations='family with sequence similarity 108, member A7, pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20840687, end_position_on_the_genomic_accession=20844479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128637, CurrentID=0, Status='live', Symbol='EEF1A1P26', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 26', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14190806, end_position_on_the_genomic_accession=14192125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  644990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644990, CurrentID=0, Status='live', Symbol='THRB-AS1', Aliases='', description='THRB antisense RNA 1', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24494087, end_position_on_the_genomic_accession=24500011, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128639, CurrentID=0, Status='live', Symbol='HMGB1P48', Aliases='', description='high mobility group box 1 pseudogene 48', other_designations='high-mobility group box 1 pseudogene 48', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45530822, end_position_on_the_genomic_accession=45532623, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128640, CurrentID=0, Status='live', Symbol='ACVR2B-AS1', Aliases='', description='ACVR2B antisense RNA 1', other_designations='ACVR2B antisense RNA 1 (non-protein coding)', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38451027, end_position_on_the_genomic_accession=38454820, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100652929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652929, CurrentID=0, Status='live', Symbol='HP09025', Aliases='', description='uncharacterized LOC100652929', other_designations='CTD-2116F7.1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79707266, end_position_on_the_genomic_accession=79712317, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  644994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644994, CurrentID=0, Status='live', Symbol='SMIM23', Aliases='C5orf50', description='small integral membrane protein 23', other_designations='CTB-78H18.1', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171772642, end_position_on_the_genomic_accession=171791136, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100652931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652931, CurrentID=0, Status='live', Symbol='LOC100652931', Aliases='', description='RNA binding motif protein, Y-linked, family 1, member A1 pseudogene', other_designations='RNA binding motif protein, Y-linked, family 1, member B pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22316678, end_position_on_the_genomic_accession=22320384, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  644996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=644996, CurrentID=0, Status='live', Symbol='RPL32P22', Aliases='RPL32_14_1010', description='ribosomal protein L32 pseudogene 22', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112362615, end_position_on_the_genomic_accession=112363116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100652933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652933, CurrentID=0, Status='live', Symbol='IGANRP', Aliases='', description='uncharacterized LOC100652933', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=104883499, end_position_on_the_genomic_accession=104886825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128646, CurrentID=0, Status='live', Symbol='RPL10AP7', Aliases='RPL10A_2_460', description='ribosomal protein L10a pseudogene 7', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15730930, end_position_on_the_genomic_accession=15731661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371527, CurrentID=0, Status='live', Symbol='LOC105371527', Aliases='', description='uncharacterized LOC105371527', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164428763, end_position_on_the_genomic_accession=164431745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645000, CurrentID=0, Status='live', Symbol='RPS3P6', Aliases='RPS3_2_1244', description='ribosomal protein S3 pseudogene 6', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62021488, end_position_on_the_genomic_accession=62022175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645001, CurrentID=0, Status='live', Symbol='HNRNPA1P52', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 52', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41829609, end_position_on_the_genomic_accession=41831134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371530, CurrentID=0, Status='live', Symbol='LOC105371530', Aliases='', description='uncharacterized LOC105371530', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8765869, end_position_on_the_genomic_accession=8768280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652939, CurrentID=0, Status='live', Symbol='MTND2P28', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 28', other_designations='MT-ND2 pseudogene 28', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=629640, end_position_on_the_genomic_accession=630686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371532, CurrentID=0, Status='live', Symbol='LOC105371532', Aliases='', description='uncharacterized LOC105371532', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9238715, end_position_on_the_genomic_accession=9239707, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371533, CurrentID=0, Status='live', Symbol='LOC105371533', Aliases='', description='uncharacterized LOC105371533', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9235991, end_position_on_the_genomic_accession=9238984, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371534, CurrentID=0, Status='live', Symbol='LOC105371534', Aliases='', description='uncharacterized LOC105371534', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10711136, end_position_on_the_genomic_accession=10713168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371535, CurrentID=0, Status='live', Symbol='LOC105371535', Aliases='', description='uncharacterized LOC105371535', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371536, CurrentID=0, Status='live', Symbol='LOC105371536', Aliases='', description='uncharacterized LOC105371536', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10905381, end_position_on_the_genomic_accession=10922819, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100128657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128657, CurrentID=0, Status='live', Symbol='CHCHD4P2', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 4 pseudogene 2', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108097435, end_position_on_the_genomic_accession=108097851, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645010, CurrentID=0, Status='live', Symbol='LOC645010', Aliases='', description='uncharacterized LOC645010', other_designations='', map_location='1p36.22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=9887668, end_position_on_the_genomic_accession=9888329, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391491, CurrentID=0, Status='live', Symbol='GAMTP1', Aliases='GAMTP', description='guanidinoacetate N-methyltransferase pseudogene 1', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231810074, end_position_on_the_genomic_accession=231810400, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371540, CurrentID=0, Status='live', Symbol='LOC105371540', Aliases='', description='uncharacterized LOC105371540', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=12590761, end_position_on_the_genomic_accession=12605957, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371541, CurrentID=0, Status='live', Symbol='LOC105371541', Aliases='', description='uncharacterized LOC105371541', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=13298869, end_position_on_the_genomic_accession=13306791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128662, CurrentID=0, Status='live', Symbol='RPL27P3', Aliases='', description='ribosomal protein L27 pseudogene', other_designations='', map_location='1p36.23', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8798438, end_position_on_the_genomic_accession=8799557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645015, CurrentID=0, Status='live', Symbol='ALG1L8P', Aliases='', description='asparagine-linked glycosylation 1-like 8, pseudogene', other_designations='asparagine-linked glycosylation 1 homolog pseudogene|asparagine-linked glycosylation 1, beta-1,4-mannosyltransferase homolog pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67785270, end_position_on_the_genomic_accession=67793345, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128664, CurrentID=0, Status='live', Symbol='RPL22P20', Aliases='RPL22_10_1337', description='ribosomal protein L22 pseudogene 20', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18692032, end_position_on_the_genomic_accession=18692408, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128665, CurrentID=0, Status='live', Symbol='FYTTD1P1', Aliases='', description='forty-two-three domain containing 1 pseudogene 1', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101373129, end_position_on_the_genomic_accession=101376511, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645018, CurrentID=0, Status='live', Symbol='RPS2P20', Aliases='RPS2_7_524', description='ribosomal protein S2 pseudogene 20', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143349229, end_position_on_the_genomic_accession=143350140, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371547, CurrentID=0, Status='live', Symbol='LOC105371547', Aliases='', description='uncharacterized LOC105371547', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15175097, end_position_on_the_genomic_accession=15182015, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371548, CurrentID=0, Status='live', Symbol='LOC105371548', Aliases='', description='uncharacterized LOC105371548', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15413733, end_position_on_the_genomic_accession=15419298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371549, CurrentID=0, Status='live', Symbol='LOC105371549', Aliases='', description='uncharacterized LOC105371549', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15289351, end_position_on_the_genomic_accession=15311621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371550, CurrentID=0, Status='live', Symbol='LOC105371550', Aliases='', description='uncharacterized LOC105371550', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15277289, end_position_on_the_genomic_accession=15290204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371551, CurrentID=0, Status='live', Symbol='LOC105371551', Aliases='', description='uncharacterized LOC105371551', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16832687, end_position_on_the_genomic_accession=16835115, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645024, CurrentID=0, Status='live', Symbol='GAPDHP59', Aliases='GAPDHL3, GAPDL3', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 59', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 3, pseudogene', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187415531, end_position_on_the_genomic_accession=187416360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371553, CurrentID=0, Status='live', Symbol='LOC105371553', Aliases='', description='uncharacterized LOC105371553', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16816359, end_position_on_the_genomic_accession=16819007, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371554, CurrentID=0, Status='live', Symbol='LOC105371554', Aliases='', description='uncharacterized LOC105371554', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16535898, end_position_on_the_genomic_accession=16548583, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100128675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128675, CurrentID=0, Status='live', Symbol='HPN-AS1', Aliases='', description='HPN antisense RNA 1', other_designations='CTD-2527I21.11', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35059059, end_position_on_the_genomic_accession=35106304, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371556, CurrentID=0, Status='live', Symbol='LOC105371556', Aliases='', description='uncharacterized LOC105371556', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16003463, end_position_on_the_genomic_accession=16008953, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371557, CurrentID=0, Status='live', Symbol='LOC105371557', Aliases='', description='uncharacterized LOC105371557', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15831277, end_position_on_the_genomic_accession=15847957, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645030, CurrentID=0, Status='live', Symbol='NFIA-AS1', Aliases='', description='NFIA antisense RNA 1', other_designations='RP5-833A20.1', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61248945, end_position_on_the_genomic_accession=61253518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371559, CurrentID=0, Status='live', Symbol='LOC105371559', Aliases='', description='uncharacterized LOC105371559', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15744633, end_position_on_the_genomic_accession=15764581, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371560, CurrentID=0, Status='live', Symbol='LOC105371560', Aliases='', description='uncharacterized LOC105371560', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128681, CurrentID=0, Status='live', Symbol='CCNG2P1', Aliases='', description='cyclin G2 pseudogene 1', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75257735, end_position_on_the_genomic_accession=75258610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128682, CurrentID=0, Status='live', Symbol='LINC01531', Aliases='', description='long intergenic non-protein coding RNA 1531', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35405607, end_position_on_the_genomic_accession=35416840, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371563, CurrentID=0, Status='live', Symbol='LOC105371563', Aliases='', description='uncharacterized LOC105371563', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17034455, end_position_on_the_genomic_accession=17042140, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371564, CurrentID=0, Status='live', Symbol='SMCR2', Aliases='TCONS_00025215', description='Smith-Magenis syndrome chromosome region, candidate 2 (non-protein coding)', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17674119, end_position_on_the_genomic_accession=17681202, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  645037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645037, CurrentID=0, Status='live', Symbol='GAGE2B', Aliases='CT4.2, GAGE-2, GAGE-2B', description='G antigen 2B', other_designations='cancer/testis antigen 4.2|g antigen 2A/2B', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49331615, end_position_on_the_genomic_accession=49338933, orientation='plus', exon_count=5, OMIM=300726, no_hdr0=''),
  100652974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652974, CurrentID=0, Status='live', Symbol='SOD1P2', Aliases='', description='superoxide dismutase 1, soluble pseudogene 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51107580, end_position_on_the_genomic_accession=51108020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371567, CurrentID=0, Status='live', Symbol='LOC105371567', Aliases='', description='uncharacterized LOC105371567', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128688, CurrentID=0, Status='live', Symbol='GPR50-AS1', Aliases='', description='GPR50 antisense RNA 1', other_designations='GPR50 antisense RNA 1 (non-protein coding)', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=150343715, end_position_on_the_genomic_accession=150346308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371569, CurrentID=0, Status='live', Symbol='LOC105371569', Aliases='', description='uncharacterized LOC105371569', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18442884, end_position_on_the_genomic_accession=18448741, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128690, CurrentID=0, Status='live', Symbol='ELL2P4', Aliases='', description='elongation factor, RNA polymerase II, 2 pseudogene 4', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146555156, end_position_on_the_genomic_accession=146555766, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371571, CurrentID=0, Status='live', Symbol='LOC105371571', Aliases='', description='uncharacterized LOC105371571', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371572, CurrentID=0, Status='live', Symbol='LOC105371572', Aliases='', description='uncharacterized LOC105371572', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19048699, end_position_on_the_genomic_accession=19055936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371573, CurrentID=0, Status='live', Symbol='LOC105371573', Aliases='', description='uncharacterized LOC105371573', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19362720, end_position_on_the_genomic_accession=19369872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371574, CurrentID=0, Status='live', Symbol='LOC105371574', Aliases='', description='uncharacterized LOC105371574', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19417366, end_position_on_the_genomic_accession=19460967, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  260425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260425, CurrentID=0, Status='live', Symbol='MAGI3', Aliases='MAGI-3, dJ730K3.2', description='membrane associated guanylate kinase, WW and PDZ domain containing 3', other_designations='membrane-associated guanylate kinase inverted 3|membrane-associated guanylate kinase-related 3', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=113390474, end_position_on_the_genomic_accession=113685923, orientation='plus', exon_count=23, OMIM=615943, no_hdr0=''),
  100128696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128696, CurrentID=0, Status='live', Symbol='PABPC1P6', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 6', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91878138, end_position_on_the_genomic_accession=91880203, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371577, CurrentID=0, Status='live', Symbol='LOC105371577', Aliases='', description='uncharacterized LOC105371577', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19589660, end_position_on_the_genomic_accession=19590690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371578, CurrentID=0, Status='live', Symbol='LOC105371578', Aliases='', description='uncharacterized LOC105371578', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19560445, end_position_on_the_genomic_accession=19571471, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645051, CurrentID=0, Status='live', Symbol='GAGE13', Aliases='GAGE-12A, GAGE-13, GAGE12A', description='G antigen 13', other_designations='G antigen 12A', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=49188094, end_position_on_the_genomic_accession=49195366, orientation='plus', exon_count=5, OMIM=300734, no_hdr0=''),
  100652988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652988, CurrentID=0, Status='live', Symbol='LINC00702', Aliases='', description='long intergenic non-protein coding RNA 702', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4206876, end_position_on_the_genomic_accession=4243912, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128701, CurrentID=0, Status='live', Symbol='HNRNPA1P41', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 41', other_designations='heterogeneous nuclear ribonucleoprotein A1-like 2 pseudogene', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4944346, end_position_on_the_genomic_accession=4945962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645054, CurrentID=0, Status='live', Symbol='RPL6P4', Aliases='RPL6_3_194', description='ribosomal protein L6 pseudogene 4', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11102123, end_position_on_the_genomic_accession=11103026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371583, CurrentID=0, Status='live', Symbol='LOC105371583', Aliases='', description='uncharacterized LOC105371583', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20649345, end_position_on_the_genomic_accession=20665589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371584, CurrentID=0, Status='live', Symbol='LOC105371584', Aliases='', description='uncharacterized LOC105371584', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20668060, end_position_on_the_genomic_accession=20684476, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100128705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128705, CurrentID=0, Status='live', Symbol='AK4P4', Aliases='', description='adenylate kinase 4 pseudogene 4', other_designations='adenylate kinase 3-like 1 pseudogene', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5855786, end_position_on_the_genomic_accession=5856426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371586, CurrentID=0, Status='live', Symbol='LOC105371586', Aliases='', description='uncharacterized LOC105371586', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166080842, end_position_on_the_genomic_accession=166091335, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128707, CurrentID=0, Status='live', Symbol='NIFKP4', Aliases='MKI67IPP4', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 4', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 4', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75251691, end_position_on_the_genomic_accession=75252979, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371588, CurrentID=0, Status='live', Symbol='LOC105371588', Aliases='', description='uncharacterized LOC105371588', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21566595, end_position_on_the_genomic_accession=21574365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100652997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652997, CurrentID=0, Status='live', Symbol='LINC00706', Aliases='', description='long intergenic non-protein coding RNA 706', other_designations='', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6774303, end_position_on_the_genomic_accession=6779180, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128710, CurrentID=0, Status='live', Symbol='RPL9P30', Aliases='RPL9_11_1536', description='ribosomal protein L9 pseudogene 30', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29855735, end_position_on_the_genomic_accession=29856420, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  120775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120775, CurrentID=0, Status='live', Symbol='OR2D3', Aliases='OR11-89', description='olfactory receptor family 2 subfamily D member 3', other_designations='olfactory receptor OR11-89|olfactory receptor, family 2, subfamily D, member 3', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6921002, end_position_on_the_genomic_accession=6921994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  120776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120776, CurrentID=0, Status='live', Symbol='OR2D2', Aliases='OR11-610, OR2D1, hg27', description='olfactory receptor family 2 subfamily D member 2', other_designations='HB2|olfactory receptor 11-610|olfactory receptor 2D1|olfactory receptor OR11-88|olfactory receptor, family 2, subfamily D, member 1|olfactory receptor, family 2, subfamily D, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6891574, end_position_on_the_genomic_accession=6892500, orientation='minus', exon_count=1, OMIM=608494, no_hdr0=''),
  645065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645065, CurrentID=0, Status='live', Symbol='RPEP2', Aliases='', description='ribulose-5-phosphate-3-epimerase pseudogene 2', other_designations='', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26015669, end_position_on_the_genomic_accession=26016379, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128714, CurrentID=0, Status='live', Symbol='LOC100128714', Aliases='', description='uncharacterized LOC100128714', other_designations='', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=25902360, end_position_on_the_genomic_accession=26053120, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371595, CurrentID=0, Status='live', Symbol='LOC105371595', Aliases='', description='uncharacterized LOC105371595', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22234556, end_position_on_the_genomic_accession=22266387, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105371596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371596, CurrentID=0, Status='live', Symbol='LOC105371596', Aliases='', description='uncharacterized LOC105371596', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166927194, end_position_on_the_genomic_accession=166935289, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371597, CurrentID=0, Status='live', Symbol='LOC105371597', Aliases='', description='uncharacterized LOC105371597', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22266465, end_position_on_the_genomic_accession=22297344, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371598, CurrentID=0, Status='live', Symbol='LOC105371598', Aliases='', description='uncharacterized LOC105371598', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167088521, end_position_on_the_genomic_accession=167091370, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  653645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653645, CurrentID=0, Status='live', Symbol='TBC1D3P1-DHX40P1', Aliases='DHX40P, DHX40P1, TBC1D3P1', description='TBC1D3P1-DHX40P1 readthrough transcribed pseudogene', other_designations='TBC1 domain family, member 3 pseudogene 1|TBC1D3-DHX40 pseudogene|TBC1D3P1-DHX40P1 readthrough (non-protein coding)', map_location='17q23', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59976024, end_position_on_the_genomic_accession=60018975, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105371600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371600, CurrentID=0, Status='live', Symbol='LOC105371600', Aliases='', description='uncharacterized LOC105371600', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167051133, end_position_on_the_genomic_accession=167067149, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  645073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645073, CurrentID=0, Status='live', Symbol='GAGE12G', Aliases='AL4, CT4.7, GAGE-12G', description='G antigen 12G', other_designations='cancer/testis antigen 4.7|g antigen 12F/G/I', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49570400, end_position_on_the_genomic_accession=49577757, orientation='plus', exon_count=5, OMIM=300731, no_hdr0=''),
  105371602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371602, CurrentID=0, Status='live', Symbol='LOC105371602', Aliases='', description='uncharacterized LOC105371602', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167639712, end_position_on_the_genomic_accession=167643235, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  120787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120787, CurrentID=0, Status='live', Symbol='OR52W1', Aliases='OR11-71P, OR52W1', description='olfactory receptor family 52 subfamily W member 1', other_designations='olfactory receptor OR11-71|olfactory receptor, family 52, subfamily W, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6199224, end_position_on_the_genomic_accession=6200186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371604, CurrentID=0, Status='live', Symbol='LOC105371604', Aliases='', description='uncharacterized LOC105371604', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168638020, end_position_on_the_genomic_accession=168666637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128725, CurrentID=0, Status='live', Symbol='APOOP4', Aliases='', description='apolipoprotein O pseudogene 4', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68304168, end_position_on_the_genomic_accession=68304966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645078, CurrentID=0, Status='live', Symbol='ZCCHC23', Aliases='', description='zinc finger, CCHC domain containing 23', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109303035, end_position_on_the_genomic_accession=109320823, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371607, CurrentID=0, Status='live', Symbol='LOC105371607', Aliases='', description='uncharacterized LOC105371607', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=169854971, end_position_on_the_genomic_accession=169862649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371608, CurrentID=0, Status='live', Symbol='LOC105371608', Aliases='', description='uncharacterized LOC105371608', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170041476, end_position_on_the_genomic_accession=170046229, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  120793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120793, CurrentID=0, Status='live', Symbol='OR56A4', Aliases='OR11-49', description='olfactory receptor family 56 subfamily A member 4', other_designations='olfactory receptor OR11-49|olfactory receptor, family 56, subfamily A, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6002051, end_position_on_the_genomic_accession=6003148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371610, CurrentID=0, Status='live', Symbol='LOC105371610', Aliases='', description='uncharacterized LOC105371610', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170655406, end_position_on_the_genomic_accession=170660691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128731, CurrentID=0, Status='live', Symbol='OST4', Aliases='', description='oligosaccharyltransferase complex subunit 4 (non-catalytic)', other_designations='oligosaccharyltransferase 4 homolog', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27070472, end_position_on_the_genomic_accession=27071699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  120796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120796, CurrentID=0, Status='live', Symbol='OR56A1', Aliases='OR11-75', description='olfactory receptor family 56 subfamily A member 1', other_designations='olfactory receptor OR11-75|olfactory receptor, family 56, subfamily A, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6026671, end_position_on_the_genomic_accession=6027741, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128733, CurrentID=0, Status='live', Symbol='FCF1P3', Aliases='', description='FCF1 pseudogene 3', other_designations='FCF1 small subunit processome component homolog pseudogene 3', map_location='3q13.12', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106848671, end_position_on_the_genomic_accession=106849260, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653022, CurrentID=0, Status='live', Symbol='PP12719', Aliases='', description='uncharacterized LOC100653022', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=58211978, end_position_on_the_genomic_accession=58213155, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105371615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371615, CurrentID=0, Status='live', Symbol='LOC105371615', Aliases='', description='uncharacterized LOC105371615', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171849288, end_position_on_the_genomic_accession=171860154, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371616, CurrentID=0, Status='live', Symbol='LOC105371616', Aliases='', description='uncharacterized LOC105371616', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172515145, end_position_on_the_genomic_accession=172532473, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  391504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391504, CurrentID=0, Status='live', Symbol='RPL23AP38', Aliases='RPL23A_11_340', description='ribosomal protein L23a pseudogene 38', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1595537, end_position_on_the_genomic_accession=1596257, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371618, CurrentID=0, Status='live', Symbol='LOC105371618', Aliases='', description='uncharacterized LOC105371618', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=172831141, end_position_on_the_genomic_accession=172836075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371619, CurrentID=0, Status='live', Symbol='LOC105371619', Aliases='', description='uncharacterized LOC105371619', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173637717, end_position_on_the_genomic_accession=173681710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371620, CurrentID=0, Status='live', Symbol='LOC105371620', Aliases='', description='uncharacterized LOC105371620', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174022472, end_position_on_the_genomic_accession=174022985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371621, CurrentID=0, Status='live', Symbol='LOC105371621', Aliases='', description='uncharacterized LOC105371621', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174056888, end_position_on_the_genomic_accession=174109993, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645094, CurrentID=0, Status='live', Symbol='MYL12BP1', Aliases='', description='myosin light chain 12B pseudogene 1', other_designations='myosin, light chain 12B pseudogene 1|myosin, light chain 12B, regulatory pseudogene', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42571824, end_position_on_the_genomic_accession=42572622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371623, CurrentID=0, Status='live', Symbol='LOC105371623', Aliases='', description='uncharacterized LOC105371623', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175330254, end_position_on_the_genomic_accession=175335704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371624, CurrentID=0, Status='live', Symbol='LOC105371624', Aliases='', description='uncharacterized LOC105371624', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175287553, end_position_on_the_genomic_accession=175321717, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128745, CurrentID=0, Status='live', Symbol='RPLP1P8', Aliases='RPLP1_2_758', description='ribosomal protein, large, P1 pseudogene 8', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136364000, end_position_on_the_genomic_accession=136364498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371626, CurrentID=0, Status='live', Symbol='LOC105371626', Aliases='', description='uncharacterized LOC105371626', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177556499, end_position_on_the_genomic_accession=177576949, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371627, CurrentID=0, Status='live', Symbol='LOC105371627', Aliases='', description='uncharacterized LOC105371627', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=177819478, end_position_on_the_genomic_accession=177825572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645100, CurrentID=0, Status='live', Symbol='LOC645100', Aliases='', description='hCG1820764', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=61682516, end_position_on_the_genomic_accession=61682851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653650, CurrentID=0, Status='live', Symbol='PDPK2P', Aliases='', description='3-phosphoinositide dependent protein kinase 2, pseudogene', other_designations='3-phosphoinositide dependent protein kinase-1 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2616120, end_position_on_the_genomic_accession=2643295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128750, CurrentID=0, Status='live', Symbol='RBPMS-AS1', Aliases='', description='RBPMS antisense RNA 1', other_designations='CTD-3107M8.4', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30382119, end_position_on_the_genomic_accession=30385401, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128751, CurrentID=0, Status='live', Symbol='LOC100128751', Aliases='', description='INM04', other_designations='', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=167637861, end_position_on_the_genomic_accession=167640111, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645104, CurrentID=0, Status='live', Symbol='CLRN2', Aliases='', description='clarin 2', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17515165, end_position_on_the_genomic_accession=17527104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371633, CurrentID=0, Status='live', Symbol='LOC105371633', Aliases='', description='uncharacterized LOC105371633', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178796956, end_position_on_the_genomic_accession=178809783, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371634, CurrentID=0, Status='live', Symbol='LOC105371634', Aliases='', description='uncharacterized LOC105371634', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179814301, end_position_on_the_genomic_accession=179819418, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  103504730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504730, CurrentID=0, Status='live', Symbol='MIR9500', Aliases='hsa-mir-9500', description='microRNA 9500', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218823090, end_position_on_the_genomic_accession=218823154, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100128756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128756, CurrentID=0, Status='live', Symbol='RPS7P9', Aliases='RPS7_5_1072', description='ribosomal protein S7 pseudogene 9', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80716468, end_position_on_the_genomic_accession=80717134, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371637, CurrentID=0, Status='live', Symbol='LOC105371637', Aliases='', description='uncharacterized LOC105371637', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180415818, end_position_on_the_genomic_accession=180424253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371638, CurrentID=0, Status='live', Symbol='LOC105371638', Aliases='', description='uncharacterized LOC105371638', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180906903, end_position_on_the_genomic_accession=180909621, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371639, CurrentID=0, Status='live', Symbol='LOC105371639', Aliases='', description='uncharacterized LOC105371639', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=181447664, end_position_on_the_genomic_accession=181482521, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  120824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120824, CurrentID=0, Status='live', Symbol='TRIM51GP', Aliases='', description='tripartite motif-containing 51G, pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48975326, end_position_on_the_genomic_accession=48983826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391508, CurrentID=0, Status='live', Symbol='PGAM1P4', Aliases='', description='phosphoglycerate mutase 1 pseudogene 4', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='3p25.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9348443, end_position_on_the_genomic_accession=9349189, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371642, CurrentID=0, Status='live', Symbol='LOC105371642', Aliases='', description='uncharacterized LOC105371642', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182109969, end_position_on_the_genomic_accession=182117815, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371643, CurrentID=0, Status='live', Symbol='LOC105371643', Aliases='', description='uncharacterized LOC105371643', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182393830, end_position_on_the_genomic_accession=182405940, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128764, CurrentID=0, Status='live', Symbol='PRKRIRP6', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 6', other_designations='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of pseudogene', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26234206, end_position_on_the_genomic_accession=26237329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653053, CurrentID=0, Status='live', Symbol='TAS2R68P', Aliases='PS7, T2R68P', description='taste 2 receptor member 68 pseudogene', other_designations='Hosa(Japanese)-T2R68*|bitter taste receptor Ps7 pseudogene|taste receptor, type 2, member 68 pseudogene', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128766, CurrentID=0, Status='live', Symbol='RPL8P2', Aliases='', description='ribosomal protein L8 pseudogene 2', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30263413, end_position_on_the_genomic_accession=30270273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371647, CurrentID=0, Status='live', Symbol='LOC105371647', Aliases='', description='uncharacterized LOC105371647', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183621938, end_position_on_the_genomic_accession=183626939, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645120, CurrentID=0, Status='live', Symbol='AIFM1P1', Aliases='', description='apoptosis inducing factor, mitochondria associated 1 pseudogene 1', other_designations='apoptosis-inducing factor, mitochondrion-associated, 1 pseudogene 1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18735675, end_position_on_the_genomic_accession=18738829, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645121, CurrentID=0, Status='live', Symbol='CCNI2', Aliases='', description='cyclin I family member 2', other_designations='cyclin I family, member 2', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132747411, end_position_on_the_genomic_accession=132756255, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371650, CurrentID=0, Status='live', Symbol='LOC105371650', Aliases='', description='uncharacterized LOC105371650', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=184943244, end_position_on_the_genomic_accession=184946879, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128771, CurrentID=0, Status='live', Symbol='EIF4BP3', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 3', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96145974, end_position_on_the_genomic_accession=96147965, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371652, CurrentID=0, Status='live', Symbol='LOC105371652', Aliases='', description='uncharacterized LOC105371652', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185354272, end_position_on_the_genomic_accession=185370571, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371653, CurrentID=0, Status='live', Symbol='GS1-204I12.4', Aliases='', description='uncharacterized LOC105371653', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=185646566, end_position_on_the_genomic_accession=185657168, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645126, CurrentID=0, Status='live', Symbol='PFN1P6', Aliases='PFN1P7', description='profilin 1 pseudogene 6', other_designations='profilin 1 pseudogene 7', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144442469, end_position_on_the_genomic_accession=144443247, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371655, CurrentID=0, Status='live', Symbol='LOC105371655', Aliases='', description='uncharacterized LOC105371655', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187254011, end_position_on_the_genomic_accession=187255104, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371656, CurrentID=0, Status='live', Symbol='LOC105371656', Aliases='', description='uncharacterized LOC105371656', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187485456, end_position_on_the_genomic_accession=187547684, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645129, CurrentID=0, Status='live', Symbol='ZBTB45P1', Aliases='', description='zinc finger and BTB domain containing 45 pseudogene 1', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109986346, end_position_on_the_genomic_accession=109988363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371658, CurrentID=0, Status='live', Symbol='LOC105371658', Aliases='', description='uncharacterized LOC105371658', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=190264959, end_position_on_the_genomic_accession=190362234, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105371659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371659, CurrentID=0, Status='live', Symbol='LOC105371659', Aliases='', description='uncharacterized LOC105371659', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=190228930, end_position_on_the_genomic_accession=190264964, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371660, CurrentID=0, Status='live', Symbol='LOC105371660', Aliases='', description='uncharacterized LOC105371660', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=191221159, end_position_on_the_genomic_accession=191228512, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128781, CurrentID=0, Status='live', Symbol='FAM35CP', Aliases='FAM35A2, FAM35AP', description='family with sequence similarity 35 member C, pseudogene', other_designations='family with sequence similarity 35, member A pseudogene|family with sequence similarity 35, member C, pseudogene', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=87952526, end_position_on_the_genomic_accession=87955403, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128782, CurrentID=0, Status='live', Symbol='LINC00476', Aliases='C9orf130, NAG11, NAG12', description='long intergenic non-protein coding RNA 476', other_designations='nasopharyngeal carcinoma-associated gene 12', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95806088, end_position_on_the_genomic_accession=95875977, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371663, CurrentID=0, Status='live', Symbol='LOC105371663', Aliases='', description='uncharacterized LOC105371663', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192343561, end_position_on_the_genomic_accession=192381698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645136, CurrentID=0, Status='live', Symbol='MYCLP2', Aliases='', description='MYCL pseudogene 2', other_designations='v-myc myelocytomatosis viral oncogene homolog 1, lung carcinoma derived pseudogene', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57931924, end_position_on_the_genomic_accession=57934837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128785, CurrentID=0, Status='live', Symbol='GS85', Aliases='', description='uncharacterized LOC100128785', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=1146811, end_position_on_the_genomic_accession=1160695, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105371666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371666, CurrentID=0, Status='live', Symbol='LOC105371666', Aliases='', description='uncharacterized LOC105371666', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192863899, end_position_on_the_genomic_accession=192873200, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128787, CurrentID=0, Status='live', Symbol='LINC01349', Aliases='', description='long intergenic non-protein coding RNA 1349', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100627049, end_position_on_the_genomic_accession=100647004, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100128788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128788, CurrentID=0, Status='live', Symbol='SRRM2-AS1', Aliases='', description='SRRM2 antisense RNA 1', other_designations='CTD-2270P14.3|SRRM2 antisense RNA 1 (non-protein coding)', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2737076, end_position_on_the_genomic_accession=2752600, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105371669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371669, CurrentID=0, Status='live', Symbol='LOC105371669', Aliases='', description='uncharacterized LOC105371669', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=192951394, end_position_on_the_genomic_accession=192955456, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645142, CurrentID=0, Status='live', Symbol='PPIAL4D', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A)-like 4D', other_designations='PPIase A-like 4D', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145241415, end_position_on_the_genomic_accession=145242199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372273, CurrentID=0, Status='live', Symbol='LOC105372273', Aliases='', description='uncharacterized LOC105372273', other_designations='CTC-510F12.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11203628, end_position_on_the_genomic_accession=11216176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371672, CurrentID=0, Status='live', Symbol='LOC105371672', Aliases='', description='uncharacterized LOC105371672', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=195753458, end_position_on_the_genomic_accession=195763506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645145, CurrentID=0, Status='live', Symbol='ZBED1P1', Aliases='ZBED1P', description='zinc finger, BED-type containing 1 pseudogene 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110277269, end_position_on_the_genomic_accession=110293660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371674, CurrentID=0, Status='live', Symbol='LOC105371674', Aliases='', description='uncharacterized LOC105371674', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=196181894, end_position_on_the_genomic_accession=196186140, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421295, CurrentID=0, Status='live', Symbol='PRPS1P2', Aliases='PRPS1L2', description='phosphoribosyl pyrophosphate synthetase 1 pseudogene 2', other_designations='phosphoribosyl pyrophosphate synthetase 1-like 2', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125150654, end_position_on_the_genomic_accession=125151588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371676, CurrentID=0, Status='live', Symbol='LOC105371676', Aliases='', description='uncharacterized LOC105371676', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197273013, end_position_on_the_genomic_accession=197288046, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371677, CurrentID=0, Status='live', Symbol='LOC105371677', Aliases='', description='uncharacterized LOC105371677', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198452364, end_position_on_the_genomic_accession=198520229, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371678, CurrentID=0, Status='live', Symbol='LOC105371678', Aliases='', description='uncharacterized LOC105371678', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198642949, end_position_on_the_genomic_accession=198657564, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  120863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120863, CurrentID=0, Status='live', Symbol='DEPDC4', Aliases='DEP.4', description='DEP domain containing 4', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100203669, end_position_on_the_genomic_accession=100267172, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105371680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371680, CurrentID=0, Status='live', Symbol='LOC105371680', Aliases='', description='uncharacterized LOC105371680', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199448986, end_position_on_the_genomic_accession=199460401, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421296, CurrentID=0, Status='live', Symbol='SRSF11P1', Aliases='', description='serine/arginine-rich splicing factor 11 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19306857, end_position_on_the_genomic_accession=19308036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371682, CurrentID=0, Status='live', Symbol='LOC105371682', Aliases='', description='uncharacterized LOC105371682', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199947647, end_position_on_the_genomic_accession=199959459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371683, CurrentID=0, Status='live', Symbol='LOC105371683', Aliases='', description='uncharacterized LOC105371683', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202123600, end_position_on_the_genomic_accession=202125637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128804, CurrentID=0, Status='live', Symbol='AGGF1P8', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 8', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 8', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35258630, end_position_on_the_genomic_accession=35259626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645157, CurrentID=0, Status='live', Symbol='RPL21P61', Aliases='RPL21_24_663', description='ribosomal protein L21 pseudogene 61', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19143653, end_position_on_the_genomic_accession=19144195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645158, CurrentID=0, Status='live', Symbol='CBX3P2', Aliases='', description='chromobox 3 pseudogene 2', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 2', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2652170, end_position_on_the_genomic_accession=2655395, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128807, CurrentID=0, Status='live', Symbol='RPL35AP37', Aliases='RPL35A_16_1766', description='ribosomal protein L35a pseudogene 37', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15117437, end_position_on_the_genomic_accession=15117768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645160, CurrentID=0, Status='live', Symbol='RPL5P30', Aliases='RPL5_16_1181', description='ribosomal protein L5 pseudogene 30', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118560617, end_position_on_the_genomic_accession=118561638, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645161, CurrentID=0, Status='live', Symbol='RPL12P8', Aliases='RPL12_10_1052, bA57G10.3', description='ribosomal protein L12 pseudogene 8', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67874332, end_position_on_the_genomic_accession=67874891, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371690, CurrentID=0, Status='live', Symbol='LOC105371690', Aliases='', description='uncharacterized LOC105371690', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204170176, end_position_on_the_genomic_accession=204174118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128811, CurrentID=0, Status='live', Symbol='GPR158-AS1', Aliases='', description='GPR158 antisense RNA 1', other_designations='GPR158 antisense RNA 1 (non-protein coding)', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25158072, end_position_on_the_genomic_accession=25176276, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371692, CurrentID=0, Status='live', Symbol='LOC105371692', Aliases='', description='uncharacterized LOC105371692', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204594403, end_position_on_the_genomic_accession=204629711, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105371693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371693, CurrentID=0, Status='live', Symbol='LOC105371693', Aliases='', description='uncharacterized LOC105371693', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204708808, end_position_on_the_genomic_accession=204715114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645166, CurrentID=0, Status='live', Symbol='LOC645166', Aliases='', description='lymphocyte-specific protein 1 pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144716079, end_position_on_the_genomic_accession=144740806, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371695, CurrentID=0, Status='live', Symbol='LOC105371695', Aliases='', description='uncharacterized LOC105371695', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204716384, end_position_on_the_genomic_accession=204726206, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128816, CurrentID=0, Status='live', Symbol='LOC100128816', Aliases='', description='ACAH3104', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=3719475, end_position_on_the_genomic_accession=3721169, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128817, CurrentID=0, Status='live', Symbol='AK6P2', Aliases='', description='adenylate kinase 6 pseudogene 2', other_designations='adenylate kinase isoenzyme 6 pseudogene', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38078179, end_position_on_the_genomic_accession=38088813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128818, CurrentID=0, Status='live', Symbol='LPEQ6126', Aliases='', description='uncharacterized LOC100128818', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47115069, end_position_on_the_genomic_accession=47118799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371699, CurrentID=0, Status='live', Symbol='LOC105371699', Aliases='', description='uncharacterized LOC105371699', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205558262, end_position_on_the_genomic_accession=205569217, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105371700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371700, CurrentID=0, Status='live', Symbol='LOC105371700', Aliases='', description='uncharacterized LOC105371700', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205539955, end_position_on_the_genomic_accession=205543560, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645173, CurrentID=0, Status='live', Symbol='RPS2P48', Aliases='RPS2_26_1559', description='ribosomal protein S2 pseudogene 48', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=53756942, end_position_on_the_genomic_accession=53757901, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645174, CurrentID=0, Status='live', Symbol='RPL21P46', Aliases='RPL21_14_463', description='ribosomal protein L21 pseudogene 46', other_designations='', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=19813733, end_position_on_the_genomic_accession=19814269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645175, CurrentID=0, Status='live', Symbol='MEMO1P2', Aliases='', description='mediator of cell motility 1 pseudogene 2', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134897857, end_position_on_the_genomic_accession=134899298, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371704, CurrentID=0, Status='live', Symbol='LOC105371704', Aliases='', description='uncharacterized LOC105371704', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27230189, end_position_on_the_genomic_accession=27231798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371705, CurrentID=0, Status='live', Symbol='LOC105371705', Aliases='', description='uncharacterized LOC105371705', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27398138, end_position_on_the_genomic_accession=27405406, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371706, CurrentID=0, Status='live', Symbol='LOC105371706', Aliases='', description='uncharacterized LOC105371706', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27412707, end_position_on_the_genomic_accession=27413945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128827, CurrentID=0, Status='live', Symbol='FRG2FP', Aliases='', description='FSHD region gene 2 family member F, pseudogene', other_designations='FSHD region gene 2 family, member F, pseudogene|FSHD region gene 2 pseudogene|TCONS_l2_00020053', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198110109, end_position_on_the_genomic_accession=198112057, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  120892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120892, CurrentID=0, Status='live', Symbol='LRRK2', Aliases='AURA17, DARDARIN, PARK8, RIPK7, ROCO2', description='leucine-rich repeat kinase 2', other_designations='augmented in rheumatoid arthritis 17', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=40225011, end_position_on_the_genomic_accession=40369285, orientation='plus', exon_count=54, OMIM=609007, no_hdr0=''),
  105371709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371709, CurrentID=0, Status='live', Symbol='LOC105371709', Aliases='', description='uncharacterized LOC105371709', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28034151, end_position_on_the_genomic_accession=28041882, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128830, CurrentID=0, Status='live', Symbol='LINC01165', Aliases='', description='long intergenic non-protein coding RNA 1165', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132520827, end_position_on_the_genomic_accession=132522569, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  544437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544437, CurrentID=0, Status='live', Symbol='COPD26', Aliases='', description='Chronic obstructive pulmonary disease QTL 26', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371712, CurrentID=0, Status='live', Symbol='LOC105371712', Aliases='', description='uncharacterized LOC105371712', other_designations='CTD-2008P7.8', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28246454, end_position_on_the_genomic_accession=28250834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371713, CurrentID=0, Status='live', Symbol='LOC105371713', Aliases='', description='uncharacterized LOC105371713', other_designations='CTD-2350C19.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28405363, end_position_on_the_genomic_accession=28406785, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128834, CurrentID=0, Status='live', Symbol='CLK3P2', Aliases='', description='CDC like kinase 3 pseudogene 2', other_designations='CDC-like kinase 3 pseudogene 2', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247935814, end_position_on_the_genomic_accession=247937128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371715, CurrentID=0, Status='live', Symbol='LOC105371715', Aliases='', description='uncharacterized LOC105371715', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29009802, end_position_on_the_genomic_accession=29010582, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128836, CurrentID=0, Status='live', Symbol='HNRNPA1P36', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 36', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81807700, end_position_on_the_genomic_accession=81808930, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371717, CurrentID=0, Status='live', Symbol='LOC105371717', Aliases='', description='uncharacterized LOC105371717', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29140478, end_position_on_the_genomic_accession=29163947, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371718, CurrentID=0, Status='live', Symbol='LOC105371718', Aliases='', description='uncharacterized LOC105371718', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29204857, end_position_on_the_genomic_accession=29212706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645191, CurrentID=0, Status='live', Symbol='LINGO3', Aliases='LERN2, LRRN6B', description='leucine rich repeat and Ig domain containing 3', other_designations='hCG2040376|leucine rich repeat neuronal 6B|leucine-rich repeat neuronal protein 2|leucine-rich repeat neuronal protein 6B', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2289775, end_position_on_the_genomic_accession=2308157, orientation='minus', exon_count=2, OMIM=609792, no_hdr0=''),
  105371720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371720, CurrentID=0, Status='live', Symbol='LOC105371720', Aliases='', description='uncharacterized LOC105371720', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30236906, end_position_on_the_genomic_accession=30252233, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371721, CurrentID=0, Status='live', Symbol='LOC105371721', Aliases='', description='uncharacterized LOC105371721', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30310833, end_position_on_the_genomic_accession=30314370, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371722, CurrentID=0, Status='live', Symbol='LOC105371722', Aliases='', description='uncharacterized LOC105371722', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31090896, end_position_on_the_genomic_accession=31093414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371723, CurrentID=0, Status='live', Symbol='LOC105371723', Aliases='', description='uncharacterized LOC105371723', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30646543, end_position_on_the_genomic_accession=30663479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371724, CurrentID=0, Status='live', Symbol='LOC105371724', Aliases='', description='uncharacterized LOC105371724', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=31379325, end_position_on_the_genomic_accession=31382116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100653133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653133, CurrentID=0, Status='live', Symbol='LOC100653133', Aliases='', description='golgin subfamily A member 6-like protein 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32489200, end_position_on_the_genomic_accession=32491804, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100271515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271515, CurrentID=0, Status='live', Symbol='RPL37P19', Aliases='RPL37_10_1232', description='ribosomal protein L37 pseudogene 19', other_designations='', map_location='12q13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48054968, end_position_on_the_genomic_accession=48055253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371727, CurrentID=0, Status='live', Symbol='LOC105371727', Aliases='', description='uncharacterized LOC105371727', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371728, CurrentID=0, Status='live', Symbol='LOC105371728', Aliases='', description='uncharacterized LOC105371728', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32083148, end_position_on_the_genomic_accession=32097125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645201, CurrentID=0, Status='live', Symbol='HMGB1P26', Aliases='', description='high mobility group box 1 pseudogene 26', other_designations='high-mobility group box 1 pseudogene 26', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229705211, end_position_on_the_genomic_accession=229705863, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645202, CurrentID=0, Status='live', Symbol='LOC645202', Aliases='', description='golgin A6 family-like', other_designations='Golgin subfamily A member 6-like protein 2-like', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28349390, end_position_on_the_genomic_accession=28358086, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  645203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645203, CurrentID=0, Status='live', Symbol='TMEM14DP', Aliases='TMEM14D, bA524O24.3', description='transmembrane protein 14D, pseudogene', other_designations='transmembrane protein 14B pseudogene|transmembrane protein 14D (pseudogene)', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68544019, end_position_on_the_genomic_accession=68544940, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645204, CurrentID=0, Status='live', Symbol='HNRNPMP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein M pseudogene 1', other_designations='', map_location='15q15.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44447575, end_position_on_the_genomic_accession=44449349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128853, CurrentID=0, Status='live', Symbol='ZNF725P', Aliases='ZNF725', description='zinc finger protein 725, pseudogene', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23489866, end_position_on_the_genomic_accession=23505982, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645206, CurrentID=0, Status='live', Symbol='LINC00693', Aliases='', description='long intergenic non-protein coding RNA 693', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28575278, end_position_on_the_genomic_accession=28758337, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  645207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645207, CurrentID=0, Status='live', Symbol='EIF4E2P2', Aliases='', description='eukaryotic translation initiation factor 4E family member 2 pseudogene 2', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=115279095, end_position_on_the_genomic_accession=115280049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371736, CurrentID=0, Status='live', Symbol='LOC105371736', Aliases='', description='uncharacterized LOC105371736', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128857, CurrentID=0, Status='live', Symbol='B3GALNT1P1', Aliases='', description='beta-1,3-N-acetylgalactosaminyltransferase 1 pseudogene 1', other_designations='beta-1,3-N-acetylgalactosaminyltransferase 1 (globoside blood group) pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=69597885, end_position_on_the_genomic_accession=69599151, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101927131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927131, CurrentID=0, Status='live', Symbol='LOC101927131', Aliases='', description='uncharacterized LOC101927131', other_designations='', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=11465260, end_position_on_the_genomic_accession=11473175, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372783, CurrentID=0, Status='live', Symbol='LOC105372783', Aliases='', description='uncharacterized LOC105372783', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32953677, end_position_on_the_genomic_accession=32955915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645212, CurrentID=0, Status='live', Symbol='EIF3J-AS1', Aliases='', description='EIF3J antisense RNA 1 (head to head)', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44534505, end_position_on_the_genomic_accession=44536923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371741, CurrentID=0, Status='live', Symbol='LOC105371741', Aliases='', description='uncharacterized LOC105371741', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34852218, end_position_on_the_genomic_accession=34854004, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371742, CurrentID=0, Status='live', Symbol='LOC105371742', Aliases='', description='uncharacterized LOC105371742', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34760588, end_position_on_the_genomic_accession=34923396, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100128863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128863, CurrentID=0, Status='live', Symbol='RPS15AP29', Aliases='RPS15A_17_1092', description='ribosomal protein S15a pseudogene 29', other_designations='', map_location='10q24.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103125017, end_position_on_the_genomic_accession=103125416, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371744, CurrentID=0, Status='live', Symbol='LOC105371744', Aliases='', description='uncharacterized LOC105371744', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35868920, end_position_on_the_genomic_accession=35884725, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371745, CurrentID=0, Status='live', Symbol='LOC105371745', Aliases='', description='uncharacterized LOC105371745', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35884737, end_position_on_the_genomic_accession=35885860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371746, CurrentID=0, Status='live', Symbol='LOC105371746', Aliases='', description='uncharacterized LOC105371746', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36037595, end_position_on_the_genomic_accession=36062770, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128867, CurrentID=0, Status='live', Symbol='RBM22P3', Aliases='', description='RNA binding motif protein 22 pseudogene 3', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56490607, end_position_on_the_genomic_accession=56491637, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645220, CurrentID=0, Status='live', Symbol='ADIPOR1P1', Aliases='', description='adiponectin receptor 1 pseudogene 1', other_designations='adiponectin receptor 2 pseudogene', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22164576, end_position_on_the_genomic_accession=22165424, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371749, CurrentID=0, Status='live', Symbol='LOC105371749', Aliases='', description='uncharacterized LOC105371749', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36475947, end_position_on_the_genomic_accession=36486662, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371750, CurrentID=0, Status='live', Symbol='LOC105371750', Aliases='', description='uncharacterized LOC105371750', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36722942, end_position_on_the_genomic_accession=36750975, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  120935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120935, CurrentID=0, Status='live', Symbol='CCDC38', Aliases='', description='coiled-coil domain containing 38', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95867048, end_position_on_the_genomic_accession=95943254, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  645224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645224, CurrentID=0, Status='live', Symbol='RPL29P17', Aliases='RPL29_8_665', description='ribosomal protein L29 pseudogene 17', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20042342, end_position_on_the_genomic_accession=20042985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371753, CurrentID=0, Status='live', Symbol='LOC105371753', Aliases='', description='uncharacterized LOC105371753', other_designations='CTC-268N12.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37043576, end_position_on_the_genomic_accession=37052775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371754, CurrentID=0, Status='live', Symbol='LOC105371754', Aliases='', description='uncharacterized LOC105371754', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37710671, end_position_on_the_genomic_accession=37724862, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  120939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=120939, CurrentID=0, Status='live', Symbol='TMEM52B', Aliases='C12orf59', description='transmembrane protein 52B', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10170599, end_position_on_the_genomic_accession=10191804, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105371756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371756, CurrentID=0, Status='live', Symbol='LOC105371756', Aliases='', description='uncharacterized LOC105371756', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37649099, end_position_on_the_genomic_accession=37651481, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371757, CurrentID=0, Status='live', Symbol='LOC105371757', Aliases='', description='uncharacterized LOC105371757', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37797915, end_position_on_the_genomic_accession=37815697, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371758, CurrentID=0, Status='live', Symbol='LOC105371758', Aliases='', description='uncharacterized LOC105371758', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37963948, end_position_on_the_genomic_accession=37966515, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371759, CurrentID=0, Status='live', Symbol='LOC105371759', Aliases='', description='uncharacterized LOC105371759', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38167448, end_position_on_the_genomic_accession=38170014, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128880, CurrentID=0, Status='live', Symbol='RAD1P1', Aliases='', description='RAD1 pseudogene 1', other_designations='RAD1 homolog pseudogene 1', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119620875, end_position_on_the_genomic_accession=119621987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128881, CurrentID=0, Status='live', Symbol='VPS9D1-AS1', Aliases='', description='VPS9D1 antisense RNA 1', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89711856, end_position_on_the_genomic_accession=89718165, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371762, CurrentID=0, Status='live', Symbol='LOC105371762', Aliases='', description='uncharacterized LOC105371762', other_designations='CTB-58E17.9', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38719037, end_position_on_the_genomic_accession=38720365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371763, CurrentID=0, Status='live', Symbol='LOC105371763', Aliases='', description='uncharacterized LOC105371763', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38700976, end_position_on_the_genomic_accession=38703537, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371764, CurrentID=0, Status='live', Symbol='LOC105371764', Aliases='', description='uncharacterized LOC105371764', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38682087, end_position_on_the_genomic_accession=38705524, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100093630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100093630, CurrentID=0, Status='live', Symbol='SNHG8', Aliases='LINC00060, NCRNA00060', description='small nucleolar RNA host gene 8', other_designations='long intergenic non-protein coding RNA 60|small nucleolar RNA host gene (non-protein coding) 8|small nucleolar RNA host gene 8 (non-protein coding)', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118278762, end_position_on_the_genomic_accession=118279823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371766, CurrentID=0, Status='live', Symbol='LOC105371766', Aliases='', description='uncharacterized LOC105371766', other_designations='CTD-2206N4.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39026708, end_position_on_the_genomic_accession=39027126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371767, CurrentID=0, Status='live', Symbol='LOC105371767', Aliases='', description='uncharacterized LOC105371767', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39008089, end_position_on_the_genomic_accession=39027637, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100128888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128888, CurrentID=0, Status='live', Symbol='RPSAP36', Aliases='RPSA_14_525', description='ribosomal protein SA pseudogene 36', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143424974, end_position_on_the_genomic_accession=143425984, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128889, CurrentID=0, Status='live', Symbol='NDUFAF4P4', Aliases='', description='NADH:ubiquinone oxidoreductase complex assembly factor 4 pseudogene 4', other_designations='NADH dehydrogenase (ubiquinone) complex I, assembly factor 4 pseudogene 4', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174849673, end_position_on_the_genomic_accession=174850168, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128890, CurrentID=0, Status='live', Symbol='FAM66B', Aliases='FAM66E', description='family with sequence similarity 66 member B', other_designations='family with sequence similarity 66, member B|family with sequence similarity 66, member E', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7301611, end_position_on_the_genomic_accession=7355354, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100421311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421311, CurrentID=0, Status='live', Symbol='BRCC3P1', Aliases='', description='BRCA1/BRCA2-containing complex subunit 3 pseudogene 1', other_designations='BRCA1/BRCA2-containing complex, subunit 3 pseudogene 1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176308055, end_position_on_the_genomic_accession=176309210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371772, CurrentID=0, Status='live', Symbol='LOC105371772', Aliases='', description='uncharacterized LOC105371772', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40315193, end_position_on_the_genomic_accession=40318151, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128893, CurrentID=0, Status='live', Symbol='GATA6-AS1', Aliases='locus5689', description='GATA6 antisense RNA 1 (head to head)', other_designations='locus 5689', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22166898, end_position_on_the_genomic_accession=22168968, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128894, CurrentID=0, Status='live', Symbol='NAP1L1P2', Aliases='', description='nucleosome assembly protein 1-like 1 pseudogene 2', other_designations='', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2929370, end_position_on_the_genomic_accession=2930363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128895, CurrentID=0, Status='live', Symbol='LOC100128895', Aliases='', description='uncharacterized LOC100128895', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.11', start_position_on_the_genomic_accession=118478052, end_position_on_the_genomic_accession=118478405, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371776, CurrentID=0, Status='live', Symbol='LOC105371776', Aliases='', description='uncharacterized LOC105371776', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40801462, end_position_on_the_genomic_accession=40814128, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  645249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645249, CurrentID=0, Status='live', Symbol='MNX1-AS1', Aliases='', description='MNX1 antisense RNA 1 (head to head)', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157010857, end_position_on_the_genomic_accession=157016424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371778, CurrentID=0, Status='live', Symbol='LOC105371778', Aliases='', description='uncharacterized LOC105371778', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100128899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128899, CurrentID=0, Status='live', Symbol='RPL7L1P9', Aliases='', description='ribosomal protein L7 like 1 pseudogene 9', other_designations='ribosomal protein L7-like 1 pseudogene 9', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217760538, end_position_on_the_genomic_accession=217761296, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371780, CurrentID=0, Status='live', Symbol='LOC105371780', Aliases='', description='uncharacterized LOC105371780', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42072769, end_position_on_the_genomic_accession=42090097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371781, CurrentID=0, Status='live', Symbol='LOC105371781', Aliases='', description='uncharacterized LOC105371781', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42153414, end_position_on_the_genomic_accession=42157639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100128902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128902, CurrentID=0, Status='live', Symbol='PRDX3P3', Aliases='', description='peroxiredoxin 3 pseudogene 3', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38617422, end_position_on_the_genomic_accession=38617997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128903, CurrentID=0, Status='live', Symbol='NPM1P4', Aliases='NG4-5, NPMP4', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 4', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene 4', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27696073, end_position_on_the_genomic_accession=27697302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645256, CurrentID=0, Status='live', Symbol='GAPDHP66', Aliases='GAPDHL11, GAPDL11', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 66', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 11, pseudogene', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3977457, end_position_on_the_genomic_accession=3978686, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645257, CurrentID=0, Status='live', Symbol='RPL7AP37', Aliases='RPL7A_15_759', description='ribosomal protein L7a pseudogene 37', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136900320, end_position_on_the_genomic_accession=136901180, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371786, CurrentID=0, Status='live', Symbol='LOC105371786', Aliases='', description='uncharacterized LOC105371786', other_designations='CTC-501O10.1', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43669589, end_position_on_the_genomic_accession=43706240, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371787, CurrentID=0, Status='live', Symbol='LOC105371787', Aliases='', description='uncharacterized LOC105371787', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43914432, end_position_on_the_genomic_accession=43917378, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128908, CurrentID=0, Status='live', Symbol='LOC100128908', Aliases='', description='leishmanolysin homolog', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23100213, end_position_on_the_genomic_accession=23105804, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105371789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371789, CurrentID=0, Status='live', Symbol='LOC105371789', Aliases='', description='uncharacterized LOC105371789', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44110810, end_position_on_the_genomic_accession=44120524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371790, CurrentID=0, Status='live', Symbol='LOC105371790', Aliases='', description='uncharacterized LOC105371790', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44274088, end_position_on_the_genomic_accession=44275681, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  391533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391533, CurrentID=0, Status='live', Symbol='FLT1P1', Aliases='', description='fms-related tyrosine kinase 1 pseudogene 1', other_designations='Fms-like tyrosine kinase 1 pseudogene|fms-related tyrosine kinase 1 (vascular endothelial growth factor/vascular permeability factor receptor) pseudogene', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46142357, end_position_on_the_genomic_accession=46143790, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128912, CurrentID=0, Status='live', Symbol='RPS21P4', Aliases='RPS21_2_461', description='ribosomal protein S21 pseudogene 4', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16256237, end_position_on_the_genomic_accession=16256601, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371793, CurrentID=0, Status='live', Symbol='LOC105371793', Aliases='', description='uncharacterized LOC105371793', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44919427, end_position_on_the_genomic_accession=44924797, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645266, CurrentID=0, Status='live', Symbol='LOC645266', Aliases='', description='pre-mRNA processing factor 4B pseudogene', other_designations='PRP4 pre-mRNA processing factor 4 homolog B pseudogene', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114845594, end_position_on_the_genomic_accession=114848588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371795, CurrentID=0, Status='live', Symbol='LOC105371795', Aliases='', description='uncharacterized LOC105371795', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45168800, end_position_on_the_genomic_accession=45172643, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371796, CurrentID=0, Status='live', Symbol='LOC105371796', Aliases='', description='uncharacterized LOC105371796', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45363255, end_position_on_the_genomic_accession=45366588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645269, CurrentID=0, Status='live', Symbol='TMEM256P1', Aliases='', description='transmembrane protein 256 pseudogene 1', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69523223, end_position_on_the_genomic_accession=69523659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371798, CurrentID=0, Status='live', Symbol='LOC105371798', Aliases='', description='uncharacterized LOC105371798', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46909742, end_position_on_the_genomic_accession=46912793, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371799, CurrentID=0, Status='live', Symbol='LOC105371799', Aliases='', description='uncharacterized LOC105371799', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46228571, end_position_on_the_genomic_accession=46241441, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645272, CurrentID=0, Status='live', Symbol='LOC645272', Aliases='', description='adenosylmethionine decarboxylase 1 pseudogene', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157088108, end_position_on_the_genomic_accession=157088377, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371801, CurrentID=0, Status='live', Symbol='LOC105371801', Aliases='', description='uncharacterized LOC105371801', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45826972, end_position_on_the_genomic_accession=45829271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371802, CurrentID=0, Status='live', Symbol='LOC105371802', Aliases='', description='uncharacterized LOC105371802', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45657496, end_position_on_the_genomic_accession=45670915, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128923, CurrentID=0, Status='live', Symbol='RPS6P16', Aliases='RPS6_8_1176', description='ribosomal protein S6 pseudogene 16', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112235364, end_position_on_the_genomic_accession=112236043, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371804, CurrentID=0, Status='live', Symbol='LOC105371804', Aliases='', description='leucine-rich repeat-containing protein 37A3 pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371805, CurrentID=0, Status='live', Symbol='LOC105371805', Aliases='', description='uncharacterized LOC105371805', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45452410, end_position_on_the_genomic_accession=45457112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371806, CurrentID=0, Status='live', Symbol='LOC105371806', Aliases='', description='uncharacterized LOC105371806', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48011317, end_position_on_the_genomic_accession=48012240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371807, CurrentID=0, Status='live', Symbol='THRA1/BTR', Aliases='', description='uncharacterized LOC105371807', other_designations='thyroid hormone receptor alpha', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48294375, end_position_on_the_genomic_accession=48309040, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645280, CurrentID=0, Status='live', Symbol='SMPD4P1', Aliases='', description='sphingomyelin phosphodiesterase 4 pseudogene 1', other_designations='sphingomyelin phosphodiesterase 4, neutral membrane (neutral sphingomyelinase-3) pseudogene 1', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20604987, end_position_on_the_genomic_accession=20627319, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371809, CurrentID=0, Status='live', Symbol='LOC105371809', Aliases='', description='uncharacterized LOC105371809', other_designations='CTD-2377D24.6|CTD-2377D24.8', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48642453, end_position_on_the_genomic_accession=48707346, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100128930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128930, CurrentID=0, Status='live', Symbol='SDCBPP1', Aliases='', description='syndecan binding protein (syntenin) pseudogene 1', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6293100, end_position_on_the_genomic_accession=6294684, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371811, CurrentID=0, Status='live', Symbol='LOC105371811', Aliases='', description='uncharacterized LOC105371811', other_designations='CTD-2377D24.4', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48732984, end_position_on_the_genomic_accession=48743465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371812, CurrentID=0, Status='live', Symbol='LOC105371812', Aliases='', description='uncharacterized LOC105371812', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48752905, end_position_on_the_genomic_accession=48753609, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371813, CurrentID=0, Status='live', Symbol='LOC105371813', Aliases='', description='uncharacterized LOC105371813', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48812720, end_position_on_the_genomic_accession=48816958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128934, CurrentID=0, Status='live', Symbol='FXYD6P1', Aliases='', description='FXYD domain containing ion transport regulator 6 pseudogene 1', other_designations='FXYD domain containing ion transport regulator 6, pseudogene 1|FXYD domain-containing ion transport regulator 6, pseudogene 1', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42555440, end_position_on_the_genomic_accession=42555727, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371815, CurrentID=0, Status='live', Symbol='LOC105371815', Aliases='', description='uncharacterized LOC105371815', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48995041, end_position_on_the_genomic_accession=48997093, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128936, CurrentID=0, Status='live', Symbol='RPL10AP6', Aliases='RPL10A_1_376', description='ribosomal protein L10a pseudogene 6', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=61742424, end_position_on_the_genomic_accession=61743137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371817, CurrentID=0, Status='live', Symbol='LOC105371817', Aliases='', description='uncharacterized LOC105371817', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49895687, end_position_on_the_genomic_accession=49933565, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371818, CurrentID=0, Status='live', Symbol='LOC105371818', Aliases='', description='uncharacterized LOC105371818', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50157873, end_position_on_the_genomic_accession=50166103, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371819, CurrentID=0, Status='live', Symbol='LOC105371819', Aliases='', description='uncharacterized LOC105371819', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50238849, end_position_on_the_genomic_accession=50256593, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645292, CurrentID=0, Status='live', Symbol='HMGB1P6', Aliases='HMG1L6, HMGB1L6', description='high mobility group box 1 pseudogene 6', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 6 pseudogene|high-mobility group box 1 pseudogene 6|high-mobility group box 1-like 6', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71163472, end_position_on_the_genomic_accession=71165512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372278, CurrentID=0, Status='live', Symbol='LOC105372278', Aliases='', description='uncharacterized LOC105372278', other_designations='CTD-2192J16.11', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12552417, end_position_on_the_genomic_accession=12553644, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  121006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121006, CurrentID=0, Status='live', Symbol='FAM186A', Aliases='', description='family with sequence similarity 186 member A', other_designations='family with sequence similarity 186, member A', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50326852, end_position_on_the_genomic_accession=50396635, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105371823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371823, CurrentID=0, Status='live', Symbol='LOC105371823', Aliases='', description='uncharacterized LOC105371823', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50503412, end_position_on_the_genomic_accession=50508197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645296, CurrentID=0, Status='live', Symbol='RPL17P39', Aliases='RPL17_14_1450, hCG2004593', description='ribosomal protein L17 pseudogene 39', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=71341123, end_position_on_the_genomic_accession=71341747, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371825, CurrentID=0, Status='live', Symbol='LOC105371825', Aliases='', description='uncharacterized LOC105371825', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50886747, end_position_on_the_genomic_accession=50894057, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100128946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128946, CurrentID=0, Status='live', Symbol='LINC01310', Aliases='WI2-81516E3.1', description='long intergenic non-protein coding RNA 1310', other_designations='', map_location='22q13.32', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48866770, end_position_on_the_genomic_accession=48898386, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  645299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645299, CurrentID=0, Status='live', Symbol='RPS10P24', Aliases='RPS10_9_1493', description='ribosomal protein S10 pseudogene 24', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70079083, end_position_on_the_genomic_accession=70079661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371828, CurrentID=0, Status='live', Symbol='LOC105371828', Aliases='', description='uncharacterized LOC105371828', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51435568, end_position_on_the_genomic_accession=51445802, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371829, CurrentID=0, Status='live', Symbol='LOC105371829', Aliases='', description='uncharacterized LOC105371829', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=52231570, end_position_on_the_genomic_accession=52236498, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  121014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121014, CurrentID=0, Status='live', Symbol='TUBB8P5', Aliases='', description='tubulin beta 8 class VIII pseudogene 5', other_designations='tubulin, beta 2B class IIb pseudogene|tubulin, beta 8 class VIII pseudogene 5', map_location='12p11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=38201567, end_position_on_the_genomic_accession=38203930, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101928649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928649, CurrentID=0, Status='live', Symbol='CTC-338M12.4', Aliases='', description='uncharacterized LOC101928649', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181246541, end_position_on_the_genomic_accession=181257586, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371832, CurrentID=0, Status='live', Symbol='LOC105371832', Aliases='', description='uncharacterized LOC105371832', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54388522, end_position_on_the_genomic_accession=54670484, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371833, CurrentID=0, Status='live', Symbol='LOC105371833', Aliases='', description='uncharacterized LOC105371833', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=55510885, end_position_on_the_genomic_accession=55561032, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105371834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371834, CurrentID=0, Status='live', Symbol='LOC105371834', Aliases='', description='uncharacterized LOC105371834', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56530760, end_position_on_the_genomic_accession=56538939, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371835, CurrentID=0, Status='live', Symbol='LOC105371835', Aliases='', description='uncharacterized LOC105371835', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56835338, end_position_on_the_genomic_accession=56843682, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371836, CurrentID=0, Status='live', Symbol='LOC105371836', Aliases='', description='uncharacterized LOC105371836', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=56961146, end_position_on_the_genomic_accession=56975951, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371837, CurrentID=0, Status='live', Symbol='LOC105371837', Aliases='', description='uncharacterized LOC105371837', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371838, CurrentID=0, Status='live', Symbol='LOC105371838', Aliases='', description='uncharacterized LOC105371838', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57442448, end_position_on_the_genomic_accession=57447131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645311, CurrentID=0, Status='live', Symbol='EEF1GP7', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 7', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52573115, end_position_on_the_genomic_accession=52573818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645312, CurrentID=0, Status='live', Symbol='GLTPP1', Aliases='', description='glycolipid transfer protein pseudogene 1', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18188904, end_position_on_the_genomic_accession=18191238, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645313, CurrentID=0, Status='live', Symbol='SCDP1', Aliases='SCDP', description='stearoyl-CoA desaturase (delta-9-desaturase) pseudogene 1', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20784250, end_position_on_the_genomic_accession=20789614, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  121026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121026, CurrentID=0, Status='live', Symbol='CYCSP30', Aliases='HCP30', description='cytochrome c, somatic pseudogene 30', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87746547, end_position_on_the_genomic_accession=87750314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128963, CurrentID=0, Status='live', Symbol='RPL3P9', Aliases='RPL3_6_918', description='ribosomal protein L3 pseudogene 9', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79571095, end_position_on_the_genomic_accession=79572350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371844, CurrentID=0, Status='live', Symbol='LOC105371844', Aliases='', description='putative speedy protein E6', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59181320, end_position_on_the_genomic_accession=59183020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645317, CurrentID=0, Status='live', Symbol='CHCHD2P6', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 6', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15604528, end_position_on_the_genomic_accession=15605318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645318, CurrentID=0, Status='live', Symbol='PHB2P1', Aliases='', description='prohibitin 2 pseudogene 1', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109527994, end_position_on_the_genomic_accession=109529211, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371847, CurrentID=0, Status='live', Symbol='LOC105371847', Aliases='', description='uncharacterized LOC105371847', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59529867, end_position_on_the_genomic_accession=59550534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371848, CurrentID=0, Status='live', Symbol='LOC105371848', Aliases='', description='uncharacterized LOC105371848', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371849, CurrentID=0, Status='live', Symbol='LOC105371849', Aliases='', description='uncharacterized LOC105371849', other_designations='CTD-2319I12.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60079309, end_position_on_the_genomic_accession=60088695, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371850, CurrentID=0, Status='live', Symbol='LOC105371850', Aliases='', description='uncharacterized LOC105371850', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60402537, end_position_on_the_genomic_accession=60417404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645323, CurrentID=0, Status='live', Symbol='LINC00461', Aliases='EyeLinc1, VISC, Visc-1a, Visc-1b, Visc-2', description='long intergenic non-protein coding RNA 461', other_designations='visual cortex expressed', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88540779, end_position_on_the_genomic_accession=88684803, orientation='minus', exon_count=7, OMIM=616611, no_hdr0=''),
  105371852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371852, CurrentID=0, Status='live', Symbol='LOC105371852', Aliases='', description='uncharacterized LOC105371852', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62075464, end_position_on_the_genomic_accession=62077280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371853, CurrentID=0, Status='live', Symbol='LOC105371853', Aliases='', description='uncharacterized LOC105371853', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62275695, end_position_on_the_genomic_accession=62276590, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645326, CurrentID=0, Status='live', Symbol='RPSAP11', Aliases='LAMR, LAMR1P11, LAMRL1, LAMRP1, RPSA_9_357', description='ribosomal protein SA pseudogene 11', other_designations='laminin receptor 1 pseudogene 11|laminin receptor-like 1', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32190671, end_position_on_the_genomic_accession=32191697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371855, CurrentID=0, Status='live', Symbol='LOC105371855', Aliases='', description='uncharacterized LOC105371855', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62705948, end_position_on_the_genomic_accession=62758153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645328, CurrentID=0, Status='live', Symbol='MRPL40P1', Aliases='', description='mitochondrial ribosomal protein L40 pseudogene 1', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67351503, end_position_on_the_genomic_accession=67352041, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100128977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128977, CurrentID=0, Status='live', Symbol='MAPT-AS1', Aliases='', description='MAPT antisense RNA 1', other_designations='MAPT antisense RNA 1 (non-protein coding)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45843356, end_position_on_the_genomic_accession=45895513, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371858, CurrentID=0, Status='live', Symbol='LOC105371858', Aliases='', description='uncharacterized LOC105371858', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63960599, end_position_on_the_genomic_accession=63966573, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128979, CurrentID=0, Status='live', Symbol='LOC100128979', Aliases='', description='uncharacterized LOC100128979', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=63338020, end_position_on_the_genomic_accession=63341586, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  645332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645332, CurrentID=0, Status='live', Symbol='FAM86C2P', Aliases='', description='family with sequence similarity 86 member C2, pseudogene', other_designations='Putative protein FAM86A-like 1|family with sequence similarity 86, member C2, pseudogene', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67791767, end_position_on_the_genomic_accession=67805336, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105371861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371861, CurrentID=0, Status='live', Symbol='LOC105371861', Aliases='', description='uncharacterized LOC105371861', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64775659, end_position_on_the_genomic_accession=64777828, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100128982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128982, CurrentID=0, Status='live', Symbol='KATNBL1P4', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 4', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51930808, end_position_on_the_genomic_accession=51932036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129401, CurrentID=0, Status='live', Symbol='NUP35', Aliases='MP-44, MP44, NP44, NUP53', description='nucleoporin 35kDa', other_designations='mitotic phosphoprotein 44|nuclear pore complex protein Nup53', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183117490, end_position_on_the_genomic_accession=183161684, orientation='plus', exon_count=13, OMIM=608140, no_hdr0=''),
  100128984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128984, CurrentID=0, Status='live', Symbol='RPL35AP7', Aliases='RPL35A_1_109', description='ribosomal protein L35a pseudogene 7', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164921276, end_position_on_the_genomic_accession=164921692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128985, CurrentID=0, Status='live', Symbol='NPM1P38', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 38', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100050201, end_position_on_the_genomic_accession=100051361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371866, CurrentID=0, Status='live', Symbol='LOC105371866', Aliases='', description='uncharacterized LOC105371866', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=65471995, end_position_on_the_genomic_accession=65479900, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645339, CurrentID=0, Status='live', Symbol='SNRPD2P2', Aliases='', description='small nuclear ribonucleoprotein D2 pseudogene 2', other_designations='small nuclear ribonucleoprotein D2 polypeptide pseudogene', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231475762, end_position_on_the_genomic_accession=231476525, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100128988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128988, CurrentID=0, Status='live', Symbol='LOC100128988', Aliases='', description='uncharacterized LOC100128988', other_designations='', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40980037, end_position_on_the_genomic_accession=41024521, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  121053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121053, CurrentID=0, Status='live', Symbol='C12orf45', Aliases='', description='chromosome 12 open reading frame 45', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=104986320, end_position_on_the_genomic_accession=104994727, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100128990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128990, CurrentID=0, Status='live', Symbol='MED6P1', Aliases='', description='mediator complex subunit 6 pseudogene 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88048151, end_position_on_the_genomic_accession=88048656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100128991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128991, CurrentID=0, Status='live', Symbol='RSL24D1P9', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 9', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51612559, end_position_on_the_genomic_accession=51629500, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371872, CurrentID=0, Status='live', Symbol='LOC105371872', Aliases='', description='uncharacterized LOC105371872', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68460560, end_position_on_the_genomic_accession=68468505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645345, CurrentID=0, Status='live', Symbol='CHCHD2P9', Aliases='C9orf49, CHCHD9', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 9', other_designations='coiled-coil-helix-coiled-coil-helix domain containing 9', map_location='9q21.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79391267, end_position_on_the_genomic_accession=79392040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371874, CurrentID=0, Status='live', Symbol='LOC105371874', Aliases='', description='uncharacterized LOC105371874', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68822689, end_position_on_the_genomic_accession=68875259, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105371875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371875, CurrentID=0, Status='live', Symbol='LOC105371875', Aliases='', description='uncharacterized LOC105371875', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68800227, end_position_on_the_genomic_accession=68824420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645348, CurrentID=0, Status='live', Symbol='TUBA3GP', Aliases='', description='tubulin alpha 3g pseudogene', other_designations='tubulin, alpha 3g, pseudogene|tubulin, alpha pseudogene', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21065467, end_position_on_the_genomic_accession=21070173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371877, CurrentID=0, Status='live', Symbol='LOC105371877', Aliases='', description='uncharacterized LOC105371877', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68715487, end_position_on_the_genomic_accession=68733524, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100128998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100128998, CurrentID=0, Status='live', Symbol='C20orf181', Aliases='dJ824A14.1', description='chromosome 20 open reading frame 181', other_designations='tissue plasminogen activator', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63832609, end_position_on_the_genomic_accession=63848312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371879, CurrentID=0, Status='live', Symbol='LOC105371879', Aliases='', description='uncharacterized LOC105371879', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69358749, end_position_on_the_genomic_accession=69362942, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371880, CurrentID=0, Status='live', Symbol='LOC105371880', Aliases='', description='uncharacterized LOC105371880', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69476147, end_position_on_the_genomic_accession=69485461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371881, CurrentID=0, Status='live', Symbol='LOC105371881', Aliases='', description='uncharacterized LOC105371881', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69996080, end_position_on_the_genomic_accession=70075137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645354, CurrentID=0, Status='live', Symbol='PRAMEF36P', Aliases='', description='PRAME family member 36, pseudogene', other_designations='PRAME family member 1 pseudogene', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13127069, end_position_on_the_genomic_accession=13132274, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645355, CurrentID=0, Status='live', Symbol='MIR3976HG', Aliases='LINC01469', description='MIR3976 host gene', other_designations='MIR3976 host gene (non-protein coding)|long intergenic non-protein coding RNA 1469', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5748819, end_position_on_the_genomic_accession=5795901, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129004, CurrentID=0, Status='live', Symbol='RPL3P10', Aliases='RPL3_5_896', description='ribosomal protein L3 pseudogene 10', other_designations='', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39157482, end_position_on_the_genomic_accession=39158719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129005, CurrentID=0, Status='live', Symbol='APOOP2', Aliases='', description='apolipoprotein O pseudogene 2', other_designations='', map_location='3p11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=87594806, end_position_on_the_genomic_accession=87595384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371886, CurrentID=0, Status='live', Symbol='LOC105371886', Aliases='', description='uncharacterized LOC105371886', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=72340084, end_position_on_the_genomic_accession=72342572, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645359, CurrentID=0, Status='live', Symbol='PRAMEF26', Aliases='', description='PRAME family member 26', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13148751, end_position_on_the_genomic_accession=13153111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645360, CurrentID=0, Status='live', Symbol='HMGB3P5', Aliases='', description='high mobility group box 3 pseudogene 5', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 5', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110450605, end_position_on_the_genomic_accession=110451404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129009, CurrentID=0, Status='live', Symbol='DCAF13P1', Aliases='', description='DDB1 and CUL4 associated factor 13 pseudogene 1', other_designations='WD repeats and SOF1 domain containing pseudogene', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106125331, end_position_on_the_genomic_accession=106127895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371890, CurrentID=0, Status='live', Symbol='LOC105371890', Aliases='', description='uncharacterized LOC105371890', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74032300, end_position_on_the_genomic_accession=74043145, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129011, CurrentID=0, Status='live', Symbol='LSM6P1', Aliases='HsT818, LSM6P', description='LSM6 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 1', other_designations='LSM6 U6 small nuclear RNA and mRNA degradation associated pseudogene 1|LSM6 homolog, U6 small nuclear RNA associated pseudogene 1|LSM6 pseudogene 1', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61618336, end_position_on_the_genomic_accession=61618576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371892, CurrentID=0, Status='live', Symbol='LOC105371892', Aliases='', description='uncharacterized LOC105371892', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74534635, end_position_on_the_genomic_accession=74537886, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129013, CurrentID=0, Status='live', Symbol='ACTR3P2', Aliases='', description='ACTR3 pseudogene 2', other_designations='ARP3 actin-related protein 3 homolog pseudogene 2', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68771322, end_position_on_the_genomic_accession=68772578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371894, CurrentID=0, Status='live', Symbol='LOC105371894', Aliases='', description='uncharacterized LOC105371894', other_designations='CTD-2006K23.2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74672315, end_position_on_the_genomic_accession=74678040, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645367, CurrentID=0, Status='live', Symbol='GGT8P', Aliases='GGT1P1', description='gamma-glutamyltransferase 8 pseudogene', other_designations='gamma-glutamyltransferase 1 pseudogene 1', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91775342, end_position_on_the_genomic_accession=91782127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645368, CurrentID=0, Status='live', Symbol='EIF3KP3', Aliases='', description='eukaryotic translation initiation factor 3 subunit K pseudogene 3', other_designations='eukaryotic translation initiation factor 3, subunit K pseudogene 3', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116043374, end_position_on_the_genomic_accession=116044124, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645369, CurrentID=0, Status='live', Symbol='TMEM200C', Aliases='TTMA', description='transmembrane protein 200C', other_designations='transmembrane protein TTMA|two transmembrane domain family member A|two transmembrane domain-containing family member A', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5889412, end_position_on_the_genomic_accession=5897101, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371898, CurrentID=0, Status='live', Symbol='LOC105371898', Aliases='', description='prothymosin alpha-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76651387, end_position_on_the_genomic_accession=76652330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371899, CurrentID=0, Status='live', Symbol='LOC105371899', Aliases='', description='uncharacterized LOC105371899', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76953181, end_position_on_the_genomic_accession=76969199, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105371900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371900, CurrentID=0, Status='live', Symbol='LOC105371900', Aliases='', description='uncharacterized LOC105371900', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77107080, end_position_on_the_genomic_accession=77120082, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645373, CurrentID=0, Status='live', Symbol='RPL21P76', Aliases='RPL21_31_868', description='ribosomal protein L21 pseudogene 76', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158719672, end_position_on_the_genomic_accession=158720203, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371902, CurrentID=0, Status='live', Symbol='LOC105371902', Aliases='', description='uncharacterized LOC105371902', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77468563, end_position_on_the_genomic_accession=77470869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129023, CurrentID=0, Status='live', Symbol='ARF4P4', Aliases='', description='ADP ribosylation factor 4 pseudogene 4', other_designations='ADP-ribosylation factor 4 pseudogene 4', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=80881563, end_position_on_the_genomic_accession=80888885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371904, CurrentID=0, Status='live', Symbol='LOC105371904', Aliases='', description='uncharacterized LOC105371904', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77385575, end_position_on_the_genomic_accession=77388672, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371905, CurrentID=0, Status='live', Symbol='LOC105371905', Aliases='', description='uncharacterized LOC105371905', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77373730, end_position_on_the_genomic_accession=77377341, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371906, CurrentID=0, Status='live', Symbol='LACAT1', Aliases='', description='lung adenocarcinoma-associated transcript 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77253764, end_position_on_the_genomic_accession=77270801, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100129027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129027, CurrentID=0, Status='live', Symbol='LOC100129027', Aliases='', description='uncharacterized LOC100129027', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45827841, end_position_on_the_genomic_accession=45836419, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129028, CurrentID=0, Status='live', Symbol='RPLP1P4', Aliases='RPLP1_3_298', description='ribosomal protein, large, P1 pseudogene 4', other_designations='60S acidic ribosomal protein P1 pseudogene', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176200871, end_position_on_the_genomic_accession=176201381, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645381, CurrentID=0, Status='live', Symbol='TLE1P1', Aliases='', description='transducin like enhancer of split 1 pseudogene 1', other_designations='transducin-like enhancer of split 1 pseudogene 1|transducin-like enhancer of split 4 (E(sp1) homolog, Drosophila) pseudogene', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65407877, end_position_on_the_genomic_accession=65408674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645382, CurrentID=0, Status='live', Symbol='PRAMEF33P', Aliases='', description='PRAME family member 33, pseudogene', other_designations='PRAME family member 10-like', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13303516, end_position_on_the_genomic_accession=13309027, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129031, CurrentID=0, Status='live', Symbol='UBL5P3', Aliases='', description='ubiquitin like 5 pseudogene 3', other_designations='ubiquitin-like 5 pseudogene 3', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63203121, end_position_on_the_genomic_accession=63203545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129032, CurrentID=0, Status='live', Symbol='GLYATL1P1', Aliases='', description='glycine-N-acyltransferase like 1 pseudogene 1', other_designations='glycine-N-acyltransferase-like 1 pseudogene 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58978831, end_position_on_the_genomic_accession=58993688, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129033, CurrentID=0, Status='live', Symbol='QIQN5815', Aliases='', description='uncharacterized LOC100129033', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=5044980, end_position_on_the_genomic_accession=5045900, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129034, CurrentID=0, Status='live', Symbol='LOC100129034', Aliases='', description='uncharacterized LOC100129034', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=124353473, end_position_on_the_genomic_accession=124359186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645387, CurrentID=0, Status='live', Symbol='RPL6P27', Aliases='RPL6_10_1582', description='ribosomal protein L6 pseudogene 27', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6462092, end_position_on_the_genomic_accession=6463029, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371916, CurrentID=0, Status='live', Symbol='LOC105371916', Aliases='', description='uncharacterized LOC105371916', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79183267, end_position_on_the_genomic_accession=79186476, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371917, CurrentID=0, Status='live', Symbol='LOC105371917', Aliases='', description='uncharacterized LOC105371917', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79779674, end_position_on_the_genomic_accession=79781685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371918, CurrentID=0, Status='live', Symbol='LOC105371918', Aliases='', description='uncharacterized LOC105371918', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79799299, end_position_on_the_genomic_accession=79802529, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371919, CurrentID=0, Status='live', Symbol='LOC105371919', Aliases='', description='uncharacterized LOC105371919', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79823498, end_position_on_the_genomic_accession=79827704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129040, CurrentID=0, Status='live', Symbol='CCDC11P1', Aliases='', description='coiled-coil domain containing 11 pseudogene 1', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=39571480, end_position_on_the_genomic_accession=39572672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371921, CurrentID=0, Status='live', Symbol='LOC105371921', Aliases='', description='uncharacterized LOC105371921', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80885804, end_position_on_the_genomic_accession=80891713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371922, CurrentID=0, Status='live', Symbol='LOC105371922', Aliases='', description='uncharacterized LOC105371922', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80559049, end_position_on_the_genomic_accession=80587121, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371923, CurrentID=0, Status='live', Symbol='LOC105371923', Aliases='', description='uncharacterized LOC105371923', other_designations='CTD-2561B21.11', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81017969, end_position_on_the_genomic_accession=81028678, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105371924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371924, CurrentID=0, Status='live', Symbol='LOC105371924', Aliases='', description='uncharacterized LOC105371924', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81134510, end_position_on_the_genomic_accession=81137035, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129045, CurrentID=0, Status='live', Symbol='EIF3KP2', Aliases='', description='eukaryotic translation initiation factor 3 subunit K pseudogene 2', other_designations='eukaryotic translation initiation factor 3, subunit K pseudogene 2', map_location='3p24.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24555796, end_position_on_the_genomic_accession=24556463, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645398, CurrentID=0, Status='live', Symbol='NIFKP2', Aliases='MKI67IPP2', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 2', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 2', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174923046, end_position_on_the_genomic_accession=174923999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371927, CurrentID=0, Status='live', Symbol='LOC105371927', Aliases='', description='uncharacterized LOC105371927', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81480523, end_position_on_the_genomic_accession=81481284, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371928, CurrentID=0, Status='live', Symbol='LOC105371928', Aliases='', description='uncharacterized LOC105371928', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81491839, end_position_on_the_genomic_accession=81510246, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129049, CurrentID=0, Status='live', Symbol='PIGFP3', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis class F pseudogene 3', other_designations='phosphatidylinositol glycan anchor biosynthesis, class F pseudogene|phosphatidylinositol glycan anchor biosynthesis, class F, pseudogene 3', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30017163, end_position_on_the_genomic_accession=30017881, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645402, CurrentID=0, Status='live', Symbol='USP17L4', Aliases='', description='ubiquitin specific peptidase 17-like family member 4', other_designations='ubiquitin specific peptidase 17-like 4', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7337115, end_position_on_the_genomic_accession=7338707, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371931, CurrentID=0, Status='live', Symbol='LOC105371931', Aliases='', description='uncharacterized LOC105371931', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29368857, end_position_on_the_genomic_accession=29392458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371932, CurrentID=0, Status='live', Symbol='LOC105371932', Aliases='', description='uncharacterized LOC105371932', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35241324, end_position_on_the_genomic_accession=35244151, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100421338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421338, CurrentID=0, Status='live', Symbol='CNOT7P1', Aliases='', description='CCR4-NOT transcription complex subunit 7 pseudogene 1', other_designations='CCR4-NOT transcription complex, subunit 7 pseudogene 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69937163, end_position_on_the_genomic_accession=69938017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371934, CurrentID=0, Status='live', Symbol='LOC105371934', Aliases='', description='uncharacterized LOC105371934', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40318289, end_position_on_the_genomic_accession=40323603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  103504732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504732, CurrentID=0, Status='live', Symbol='MIR941-5', Aliases='', description='microRNA 941-5', other_designations='hsa-mir-941-5', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63919868, end_position_on_the_genomic_accession=63919939, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105371936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371936, CurrentID=0, Status='live', Symbol='LOC105371936', Aliases='', description='uncharacterized LOC105371936', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62344040, end_position_on_the_genomic_accession=62353998, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371937, CurrentID=0, Status='live', Symbol='LOC105371937', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62463848, end_position_on_the_genomic_accession=62479796, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371938, CurrentID=0, Status='live', Symbol='LOC105371938', Aliases='', description='uncharacterized LOC105371938', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81942018, end_position_on_the_genomic_accession=81955417, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101928667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928667, CurrentID=0, Status='live', Symbol='LOC101928667', Aliases='', description='uncharacterized LOC101928667', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36007008, end_position_on_the_genomic_accession=36024200, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  645412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645412, CurrentID=0, Status='live', Symbol='RPS16P9', Aliases='RPS16_5_1659', description='ribosomal protein S16 pseudogene 9', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45332629, end_position_on_the_genomic_accession=45333188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653702, CurrentID=0, Status='live', Symbol='RPL7P9', Aliases='RPL7_3_75', description='ribosomal protein L7 pseudogene 9', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96678808, end_position_on_the_genomic_accession=96679640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645414, CurrentID=0, Status='live', Symbol='PRAMEF19', Aliases='PRAMEF18', description='PRAME family member 19', other_designations='PRAME family member 18|PRAME family member-like', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13368431, end_position_on_the_genomic_accession=13371947, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371943, CurrentID=0, Status='live', Symbol='LOC105371943', Aliases='', description='uncharacterized LOC105371943', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82728916, end_position_on_the_genomic_accession=82732815, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371944, CurrentID=0, Status='live', Symbol='LOC105371944', Aliases='', description='uncharacterized LOC105371944', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82941490, end_position_on_the_genomic_accession=82942075, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  121129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121129, CurrentID=0, Status='live', Symbol='OR2AP1', Aliases='OR2AP1P', description='olfactory receptor family 2 subfamily AP member 1', other_designations='olfactory receptor OR12-9|olfactory receptor, family 2, subfamily AP, member 1 pseudogene|seven transmembrane helix receptor', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55574415, end_position_on_the_genomic_accession=55575344, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  121130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121130, CurrentID=0, Status='live', Symbol='OR10P1', Aliases='OR10P1P, OR10P2P, OR10P3P, OR12-7, OST701', description='olfactory receptor family 10 subfamily P member 1', other_designations='olfactory receptor 10P2|olfactory receptor 10P3|olfactory receptor OR12-7|olfactory receptor, family 10, subfamily P, member 1 pseudogene|olfactory receptor, family 10, subfamily P, member 2 pseudogene|olfactory receptor, family 10, subfamily P, member 3 pseudogene|seven transmembrane helix receptor', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55636892, end_position_on_the_genomic_accession=55637833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129067, CurrentID=0, Status='live', Symbol='NPM1P35', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 35', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62330842, end_position_on_the_genomic_accession=62332016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371948, CurrentID=0, Status='live', Symbol='LOC105371948', Aliases='', description='uncharacterized LOC105371948', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83193578, end_position_on_the_genomic_accession=83199883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371949, CurrentID=0, Status='live', Symbol='LOC105371949', Aliases='', description='uncharacterized LOC105371949', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645422, CurrentID=0, Status='live', Symbol='RPL10P12', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68626810, end_position_on_the_genomic_accession=68627187, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645423, CurrentID=0, Status='live', Symbol='SLC25A51P2', Aliases='MCART4P', description='solute carrier family 25, member 51 pseudogene 2', other_designations='mitochondrial carrier triple repeat 1 pseudogene|mitochondrial carrier triple repeat 4 pseudogene', map_location='18p11.23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=7134120, end_position_on_the_genomic_accession=7135881, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105371952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371952, CurrentID=0, Status='live', Symbol='LOC105371952', Aliases='', description='uncharacterized LOC105371952', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=623689, end_position_on_the_genomic_accession=632709, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645425, CurrentID=0, Status='live', Symbol='PRAMEF20', Aliases='PRAMEF21', description='PRAME family member 20', other_designations='OTTHUMT00000008157|PRAME family member 20/21|PRAME family member 21', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13410450, end_position_on_the_genomic_accession=13421354, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645426, CurrentID=0, Status='live', Symbol='TMEM191C', Aliases='', description='transmembrane protein 191C', other_designations='KB-1323B2.6', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21467140, end_position_on_the_genomic_accession=21469940, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100129075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129075, CurrentID=0, Status='live', Symbol='KTN1-AS1', Aliases='C14orf33, MYCLo-3', description='KTN1 antisense RNA 1', other_designations='KTN1 antisense RNA 1 (non-protein coding)', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55576157, end_position_on_the_genomic_accession=55580092, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371956, CurrentID=0, Status='live', Symbol='LOC105371956', Aliases='', description='uncharacterized LOC105371956', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=2034172, end_position_on_the_genomic_accession=2126369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100653365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653365, CurrentID=0, Status='live', Symbol='BED', Aliases='', description='Bornholm eye disease', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300843, no_hdr0=''),
  100653366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653366, CurrentID=0, Status='live', Symbol='PSC', Aliases='', description='Cholangitis, primary sclerosing', other_designations='', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613806, no_hdr0=''),
  100653367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653367, CurrentID=0, Status='live', Symbol='OFC13', Aliases='', description='Orofacial cleft 13', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613857, no_hdr0=''),
  645432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645432, CurrentID=0, Status='live', Symbol='ARRDC5', Aliases='', description='arrestin domain containing 5', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4890437, end_position_on_the_genomic_accession=4916820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100653369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653369, CurrentID=0, Status='live', Symbol='APMR3', Aliases='', description='Alopecia-mental retardation syndrome 3', other_designations='', map_location='18q11.2-q12.2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613930, no_hdr0=''),
  100653370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653370, CurrentID=0, Status='live', Symbol='MYP19', Aliases='', description='Myopia 19, autosomal dominant', other_designations='', map_location='5p15.1-p13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613969, no_hdr0=''),
  100653371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653371, CurrentID=0, Status='live', Symbol='CPVT3', Aliases='', description='Ventricular tachycardia, catecholaminergic polymorphic, 3', other_designations='', map_location='7p22-p14', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614021, no_hdr0=''),
  100653372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653372, CurrentID=0, Status='live', Symbol='SRXX3', Aliases='', description='46XX sex reversal 3', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300833, no_hdr0=''),
  100653373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653373, CurrentID=0, Status='live', Symbol='KONDS', Aliases='', description='Kondoh syndrome', other_designations='', map_location='1p36.32-p35.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606242, no_hdr0=''),
  100653374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653374, CurrentID=0, Status='live', Symbol='DEL17P13.1', Aliases='C17DELp13.1', description='Chromosome 17p13.1 deletion syndrome', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613776, no_hdr0=''),
  100653375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653375, CurrentID=0, Status='live', Symbol='ADIPQTL5', Aliases='', description='Adiponectin, serum level of, QTL5', other_designations='', map_location='16q23.3', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613836, no_hdr0=''),
  645440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645440, CurrentID=0, Status='live', Symbol='ATP5F1P6', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 6', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 6', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139614441, end_position_on_the_genomic_accession=139615204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100653377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653377, CurrentID=0, Status='live', Symbol='SPGF2', Aliases='ASG', description='Spermatogenic failure', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=108420, no_hdr0=''),
  100653378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653378, CurrentID=0, Status='live', Symbol='SRXX2', Aliases='DEL17q24.3, DUP17q24.3', description='46XX sex reversal 2', other_designations='', map_location='17q24.3-q25.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=278850, no_hdr0=''),
  100653379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653379, CurrentID=0, Status='live', Symbol='MYMY4', Aliases='CXDELq38', description='Moyamoya disease 4', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300845, no_hdr0=''),
  100653380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653380, CurrentID=0, Status='live', Symbol='DEL7Q11.23', Aliases='C7DELq11.23', description='Chromosome 7q11.23 deletion syndrome, distal, 1.2Mb', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613729, no_hdr0=''),
  100653381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653381, CurrentID=0, Status='live', Symbol='GEFSP8', Aliases='', description='Generalized epilepsy with febrile seizures plus, type 8', other_designations='', map_location='6q16.3-q22.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613828, no_hdr0=''),
  100653382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653382, CurrentID=0, Status='live', Symbol='DEL13Q14', Aliases='C13DELq14', description='Chromosome 13q14 deletion syndrome', other_designations='', map_location='13q14', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613884, no_hdr0=''),
  100653383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653383, CurrentID=0, Status='live', Symbol='PSMNSW', Aliases='', description='Parasomnia, sleepwalking type', other_designations='', map_location='20q12-q13.12', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613938, no_hdr0=''),
  100653384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653384, CurrentID=0, Status='live', Symbol='IGAN2', Aliases='', description='IgA nephropathy, susceptibility to, 2', other_designations='', map_location='2q36', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613944, no_hdr0=''),
  100653385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653385, CurrentID=0, Status='live', Symbol='DEL3PTERP25', Aliases='C3DELpterp25', description='3p- syndrome', other_designations='', map_location='3pter-p25', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613792, no_hdr0=''),
  105371978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371978, CurrentID=0, Status='live', Symbol='LOC105371978', Aliases='', description='uncharacterized LOC105371978', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8950484, end_position_on_the_genomic_accession=8952858, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371979, CurrentID=0, Status='live', Symbol='LOC105371979', Aliases='', description='uncharacterized LOC105371979', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9008807, end_position_on_the_genomic_accession=9017644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645452, CurrentID=0, Status='live', Symbol='RPL21P68', Aliases='RPL21_25_668', description='ribosomal protein L21 pseudogene 68', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25261199, end_position_on_the_genomic_accession=25261758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645453, CurrentID=0, Status='live', Symbol='RLIMP3', Aliases='', description='ring finger protein, LIM domain interacting pseudogene 3', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49794244, end_position_on_the_genomic_accession=49796097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653390, CurrentID=0, Status='live', Symbol='DFNB81', Aliases='', description='deafness, autosomal recessive 81', other_designations='', map_location='19p13.3-p13.11', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614129, no_hdr0=''),
  645455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645455, CurrentID=0, Status='live', Symbol='CEP170P1', Aliases='CEP170L, FAM68B, KIAA0470L', description='centrosomal protein 170kDa pseudogene 1', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118516340, end_position_on_the_genomic_accession=118554204, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105371984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371984, CurrentID=0, Status='live', Symbol='LOC105371984', Aliases='', description='uncharacterized LOC105371984', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10085727, end_position_on_the_genomic_accession=10093894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100653393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653393, CurrentID=0, Status='live', Symbol='CACD3', Aliases='', description='Choroidal dystrophy, central areolar 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613144, no_hdr0=''),
  105371986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371986, CurrentID=0, Status='live', Symbol='LOC105371986', Aliases='', description='uncharacterized LOC105371986', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10327067, end_position_on_the_genomic_accession=10373392, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129107, CurrentID=0, Status='live', Symbol='TCF3P1', Aliases='', description='transcription factor 3 pseudogene 1', other_designations='transcription factor 3 (E2A immunoglobulin enhancer binding factors E12/E47) pseudogene', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5110861, end_position_on_the_genomic_accession=5113421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653396, CurrentID=0, Status='live', Symbol='YPEL5P3', Aliases='', description='YPEL5 pseudogene 3', other_designations='yippee-like 5 pseudogene 3', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96059340, end_position_on_the_genomic_accession=96060455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371989, CurrentID=0, Status='live', Symbol='LOC105371989', Aliases='', description='uncharacterized LOC105371989', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10380435, end_position_on_the_genomic_accession=10401145, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371990, CurrentID=0, Status='live', Symbol='LOC105371990', Aliases='', description='uncharacterized LOC105371990', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10598743, end_position_on_the_genomic_accession=10601177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105371991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371991, CurrentID=0, Status='live', Symbol='LOC105371991', Aliases='', description='uncharacterized LOC105371991', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371992, CurrentID=0, Status='live', Symbol='LOC105371992', Aliases='', description='uncharacterized LOC105371992', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11093727, end_position_on_the_genomic_accession=11094682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371993, CurrentID=0, Status='live', Symbol='LOC105371993', Aliases='', description='uncharacterized LOC105371993', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11316436, end_position_on_the_genomic_accession=11319015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105371994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371994, CurrentID=0, Status='live', Symbol='LOC105371994', Aliases='', description='uncharacterized LOC105371994', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11454494, end_position_on_the_genomic_accession=11466385, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129115, CurrentID=0, Status='live', Symbol='LOC100129115', Aliases='', description='uncharacterized LOC100129115', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=40560916, end_position_on_the_genomic_accession=40561486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  645468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645468, CurrentID=0, Status='live', Symbol='LOC645468', Aliases='', description='insulin like growth factor 2 mRNA binding protein 3 pseudogene', other_designations='insulin-like growth factor 2 mRNA binding protein 3 pseudogene', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166699361, end_position_on_the_genomic_accession=166703092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105371997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371997, CurrentID=0, Status='live', Symbol='LOC105371997', Aliases='', description='uncharacterized LOC105371997', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12288216, end_position_on_the_genomic_accession=12291591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105371998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371998, CurrentID=0, Status='live', Symbol='LOC105371998', Aliases='', description='uncharacterized LOC105371998', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12437612, end_position_on_the_genomic_accession=12449298, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105371999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371999, CurrentID=0, Status='live', Symbol='LOC105371999', Aliases='', description='uncharacterized LOC105371999', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13526177, end_position_on_the_genomic_accession=13528376, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372000, CurrentID=0, Status='live', Symbol='LOC105372000', Aliases='', description='uncharacterized LOC105372000', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372001, CurrentID=0, Status='live', Symbol='LOC105372001', Aliases='', description='uncharacterized LOC105372001', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645474, CurrentID=0, Status='live', Symbol='S100A11P3', Aliases='', description='S100 calcium binding protein A11 pseudogene 3', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71506061, end_position_on_the_genomic_accession=71522087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129123, CurrentID=0, Status='live', Symbol='ACE3P', Aliases='ACE3', description='angiotensin I converting enzyme (peptidyl-dipeptidase A) 3, pseudogene', other_designations='angiotensin I converting enzyme (peptidyl-dipeptidase A) 1 pseudogene|angiotensin converting enzyme-3 pseudogene', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63507034, end_position_on_the_genomic_accession=63519827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129124, CurrentID=0, Status='live', Symbol='TVP23BP1', Aliases='FAM18B3P', description='TVP23B pseudogene 1', other_designations='family with sequence similarity 18, member B pseudogene|family with sequence similarity 18, member B3, pseudogene|trans-golgi network vesicle protein 23 homolog B pseudogene 1', map_location='15q12', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=26964915, end_position_on_the_genomic_accession=26965511, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372005, CurrentID=0, Status='live', Symbol='LOC105372005', Aliases='', description='uncharacterized LOC105372005', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14724763, end_position_on_the_genomic_accession=14725929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129126, CurrentID=0, Status='live', Symbol='LOC100129126', Aliases='', description='uncharacterized LOC100129126', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=63411914, end_position_on_the_genomic_accession=63441761, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372007, CurrentID=0, Status='live', Symbol='LOC105372007', Aliases='', description='uncharacterized LOC105372007', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14729251, end_position_on_the_genomic_accession=14747353, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129128, CurrentID=0, Status='live', Symbol='KHDC1L', Aliases='', description='KH homology domain containing 1-like', other_designations='RP11-257K9.7', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73223544, end_position_on_the_genomic_accession=73225452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645481, CurrentID=0, Status='live', Symbol='LOC645481', Aliases='', description='zinc finger protein 248 pseudogene', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26111738, end_position_on_the_genomic_accession=26115952, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372010, CurrentID=0, Status='live', Symbol='LOC105372010', Aliases='', description='uncharacterized LOC105372010', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14910358, end_position_on_the_genomic_accession=14917464, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372011, CurrentID=0, Status='live', Symbol='LOC105372011', Aliases='', description='uncharacterized LOC105372011', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14877793, end_position_on_the_genomic_accession=14888646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372012, CurrentID=0, Status='live', Symbol='LOC105372012', Aliases='', description='uncharacterized LOC105372012', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372013, CurrentID=0, Status='live', Symbol='LOC105372013', Aliases='', description='basic salivary proline-rich protein 2-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=4454758, end_position_on_the_genomic_accession=4485704, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  645486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645486, CurrentID=0, Status='live', Symbol='AHCYP7', Aliases='', description='adenosylhomocysteinase pseudogene 7', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50417289, end_position_on_the_genomic_accession=50418883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100653423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653423, CurrentID=0, Status='live', Symbol='BANF1P4', Aliases='', description='barrier to autointegration factor 1 pseudogene 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=174756470, end_position_on_the_genomic_accession=174757212, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372016, CurrentID=0, Status='live', Symbol='LOC105372016', Aliases='', description='uncharacterized LOC105372016', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21896936, end_position_on_the_genomic_accession=21916648, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372017, CurrentID=0, Status='live', Symbol='LOC105372017', Aliases='', description='uncharacterized LOC105372017', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22215925, end_position_on_the_genomic_accession=22236380, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129138, CurrentID=0, Status='live', Symbol='LOC100129138', Aliases='', description='THAP domain containing, apoptosis associated protein 3 pseudogene', other_designations='', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=104073023, end_position_on_the_genomic_accession=104077086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372019, CurrentID=0, Status='live', Symbol='LOC105372019', Aliases='', description='uncharacterized LOC105372019', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22347846, end_position_on_the_genomic_accession=22349645, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372020, CurrentID=0, Status='live', Symbol='LOC105372020', Aliases='', description='uncharacterized LOC105372020', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22458889, end_position_on_the_genomic_accession=22465011, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129141, CurrentID=0, Status='live', Symbol='RPL21P37', Aliases='RPL21_11_229', description='ribosomal protein L21 pseudogene 37', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62532542, end_position_on_the_genomic_accession=62533099, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372022, CurrentID=0, Status='live', Symbol='LOC105372022', Aliases='', description='uncharacterized LOC105372022', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23133586, end_position_on_the_genomic_accession=23135761, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129143, CurrentID=0, Status='live', Symbol='SMUG1P1', Aliases='', description='single-strand-selective monofunctional uracil-DNA glycosylase 1 pseudogene 1', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49650350, end_position_on_the_genomic_accession=49651162, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372024, CurrentID=0, Status='live', Symbol='LOC105372024', Aliases='', description='uncharacterized LOC105372024', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24135301, end_position_on_the_genomic_accession=24138868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372025, CurrentID=0, Status='live', Symbol='LOC105372025', Aliases='', description='uncharacterized LOC105372025', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24326553, end_position_on_the_genomic_accession=24401591, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372026, CurrentID=0, Status='live', Symbol='LOC105372026', Aliases='', description='uncharacterized LOC105372026', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24499999, end_position_on_the_genomic_accession=24504142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372027, CurrentID=0, Status='live', Symbol='LOC105372027', Aliases='', description='uncharacterized LOC105372027', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24667979, end_position_on_the_genomic_accession=24671701, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372028, CurrentID=0, Status='live', Symbol='LOC105372028', Aliases='', description='uncharacterized LOC105372028', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24725781, end_position_on_the_genomic_accession=24766728, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372029, CurrentID=0, Status='live', Symbol='LOC105372029', Aliases='', description='uncharacterized LOC105372029', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24793154, end_position_on_the_genomic_accession=24800110, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  121214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121214, CurrentID=0, Status='live', Symbol='SDR9C7', Aliases='RDHS, SDR-O, SDRO', description='short chain dehydrogenase/reductase family 9C, member 7', other_designations='RDH-S|orphan short-chain dehydrogenase / reductase|orphan short-chain dehydrogenase/reductase|retinol dehydrogenase similar protein', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56923154, end_position_on_the_genomic_accession=56934405, orientation='minus', exon_count=4, OMIM=609769, no_hdr0=''),
  105372285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372285, CurrentID=0, Status='live', Symbol='LOC105372285', Aliases='', description='uncharacterized LOC105372285', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14006900, end_position_on_the_genomic_accession=14008223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372032, CurrentID=0, Status='live', Symbol='LOC105372032', Aliases='', description='uncharacterized LOC105372032', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=25124069, end_position_on_the_genomic_accession=25127069, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372033, CurrentID=0, Status='live', Symbol='LOC105372033', Aliases='', description='uncharacterized LOC105372033', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=25957984, end_position_on_the_genomic_accession=25960503, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645506, CurrentID=0, Status='live', Symbol='GYG1P3', Aliases='', description='glycogenin 1 pseudogene 3', other_designations='glycogenin 1 pseudogene 1', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55222279, end_position_on_the_genomic_accession=55224076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372035, CurrentID=0, Status='live', Symbol='LOC105372035', Aliases='', description='uncharacterized LOC105372035', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=26822342, end_position_on_the_genomic_accession=26825158, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129156, CurrentID=0, Status='live', Symbol='RBBP6P1', Aliases='', description='retinoblastoma binding protein 6 pseudogene 1', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77738222, end_position_on_the_genomic_accession=77739049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372037, CurrentID=0, Status='live', Symbol='LOC105372037', Aliases='', description='uncharacterized LOC105372037', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27224717, end_position_on_the_genomic_accession=27247181, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372038, CurrentID=0, Status='live', Symbol='LOC105372038', Aliases='', description='uncharacterized LOC105372038', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27336942, end_position_on_the_genomic_accession=27342690, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372039, CurrentID=0, Status='live', Symbol='LOC105372039', Aliases='', description='uncharacterized LOC105372039', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27342780, end_position_on_the_genomic_accession=27344256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129160, CurrentID=0, Status='live', Symbol='NMNAT1P2', Aliases='', description='NMNAT1 pseudogene 2', other_designations='nicotinamide nucleotide adenylyltransferase 1 pseudogene', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=164343005, end_position_on_the_genomic_accession=164343850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372041, CurrentID=0, Status='live', Symbol='LOC105372041', Aliases='', description='uncharacterized LOC105372041', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=27375351, end_position_on_the_genomic_accession=27401903, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129162, CurrentID=0, Status='live', Symbol='RSL24D1P2', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 2', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62552980, end_position_on_the_genomic_accession=62561423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  121227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121227, CurrentID=0, Status='live', Symbol='LRIG3', Aliases='LIG3', description='leucine-rich repeats and immunoglobulin-like domains 3', other_designations='LIG-3', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=58872149, end_position_on_the_genomic_accession=58920538, orientation='minus', exon_count=21, OMIM=608870, no_hdr0=''),
  105372044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372044, CurrentID=0, Status='live', Symbol='LOC105372044', Aliases='', description='uncharacterized LOC105372044', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=29180742, end_position_on_the_genomic_accession=29236720, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372045, CurrentID=0, Status='live', Symbol='LOC105372045', Aliases='', description='uncharacterized LOC105372045', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=29629844, end_position_on_the_genomic_accession=29651532, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376650, CurrentID=0, Status='live', Symbol='LOC105376650', Aliases='', description='uncharacterized LOC105376650', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45008273, end_position_on_the_genomic_accession=45043897, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129167, CurrentID=0, Status='live', Symbol='RFESDP1', Aliases='', description='Rieske (Fe-S) domain containing pseudogene 1', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22848750, end_position_on_the_genomic_accession=22851440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645520, CurrentID=0, Status='live', Symbol='FAM138A', Aliases='F379, FAM138F', description='family with sequence similarity 138 member A', other_designations='F379 retina-specific protein|chromosome 1 F379 retina specific protein|family with sequence similarity 138, member A', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34611, end_position_on_the_genomic_accession=36081, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391576, CurrentID=0, Status='live', Symbol='OR7E129P', Aliases='', description='olfactory receptor family 7 subfamily E member 129 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 129 pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130021555, end_position_on_the_genomic_accession=130022579, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372050, CurrentID=0, Status='live', Symbol='LOC105372050', Aliases='', description='uncharacterized LOC105372050', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31510569, end_position_on_the_genomic_accession=31513674, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372051, CurrentID=0, Status='live', Symbol='LOC105372051', Aliases='', description='uncharacterized LOC105372051', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31556981, end_position_on_the_genomic_accession=31586451, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645524, CurrentID=0, Status='live', Symbol='FLJ36840', Aliases='', description='uncharacterized LOC645524', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100421358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421358, CurrentID=0, Status='live', Symbol='LOC100421358', Aliases='', description='CREB regulated transcription coactivator 2 pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62333657, end_position_on_the_genomic_accession=62335346, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129174, CurrentID=0, Status='live', Symbol='POLR2KP2', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide K, 7.0kDa pseudogene 2', other_designations='novel protein similar to polymerase (RNA) II (DNA directed) polypeptide K, 7.0kDa POLR2K', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48133956, end_position_on_the_genomic_accession=48134553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129175, CurrentID=0, Status='live', Symbol='LOC100129175', Aliases='', description='uncharacterized LOC100129175', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219002215, end_position_on_the_genomic_accession=219015722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129176, CurrentID=0, Status='live', Symbol='RPL13P6', Aliases='RPL13AP4, RPL13_5_1409', description='ribosomal protein L13 pseudogene 6', other_designations='ribosomal protein L13a pseudogene 4', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103041469, end_position_on_the_genomic_accession=103042094, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480874, CurrentID=0, Status='live', Symbol='RN7SKP132', Aliases='', description='RNA, 7SK small nuclear pseudogene 132', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372058, CurrentID=0, Status='live', Symbol='LOC105372058', Aliases='', description='uncharacterized LOC105372058', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32833433, end_position_on_the_genomic_accession=32925895, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100129179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129179, CurrentID=0, Status='live', Symbol='RPL9P5', Aliases='RPL9_9_1374', description='ribosomal protein L9 pseudogene 5', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58564058, end_position_on_the_genomic_accession=58564738, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129180, CurrentID=0, Status='live', Symbol='RPL36AP11', Aliases='RPL36A_2_73', description='ribosomal protein L36a pseudogene 11', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93190690, end_position_on_the_genomic_accession=93191086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372061, CurrentID=0, Status='live', Symbol='LOC105372061', Aliases='', description='uncharacterized LOC105372061', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34891105, end_position_on_the_genomic_accession=34907066, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372062, CurrentID=0, Status='live', Symbol='LOC105372062', Aliases='', description='uncharacterized LOC105372062', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=34735771, end_position_on_the_genomic_accession=34760073, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372063, CurrentID=0, Status='live', Symbol='LOC105372063', Aliases='', description='uncharacterized LOC105372063', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35397058, end_position_on_the_genomic_accession=35409074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372064, CurrentID=0, Status='live', Symbol='LOC105372064', Aliases='', description='uncharacterized LOC105372064', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35690076, end_position_on_the_genomic_accession=35696982, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372065, CurrentID=0, Status='live', Symbol='LOC105372065', Aliases='', description='uncharacterized LOC105372065', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35666625, end_position_on_the_genomic_accession=35677584, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645538, CurrentID=0, Status='live', Symbol='GOT2P1', Aliases='GOT2L1', description='glutamic-oxaloacetic transaminase 2 pseudogene 1', other_designations='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) (GOT2) pseudogene|glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene|glutamic-oxaloacetic transaminase 2-like 1 pseudogene', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55367465, end_position_on_the_genomic_accession=55368774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372067, CurrentID=0, Status='live', Symbol='LOC105372067', Aliases='', description='uncharacterized LOC105372067', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35984624, end_position_on_the_genomic_accession=35989502, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372068, CurrentID=0, Status='live', Symbol='LOC105372068', Aliases='', description='uncharacterized LOC105372068', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37274460, end_position_on_the_genomic_accession=37275728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372069, CurrentID=0, Status='live', Symbol='LOC105372069', Aliases='', description='uncharacterized LOC105372069', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36845224, end_position_on_the_genomic_accession=36923835, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372070, CurrentID=0, Status='live', Symbol='LOC105372070', Aliases='', description='uncharacterized LOC105372070', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36693431, end_position_on_the_genomic_accession=36714289, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372071, CurrentID=0, Status='live', Symbol='LOC105372071', Aliases='', description='uncharacterized LOC105372071', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=36660268, end_position_on_the_genomic_accession=36663555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121256, CurrentID=0, Status='live', Symbol='TMEM132D', Aliases='MOLT, PPP1R153', description='transmembrane protein 132D', other_designations='mature OL transmembrane protein|mature oligodendrocytes transmembrane protein|protein phosphatase 1, regulatory subunit 153', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=129071726, end_position_on_the_genomic_accession=129903730, orientation='minus', exon_count=9, OMIM=611257, no_hdr0=''),
  100129193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129193, CurrentID=0, Status='live', Symbol='MUPP', Aliases='MUP', description='major urinary protein, pseudogene', other_designations='major urinary protein pseudogene', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112959134, end_position_on_the_genomic_accession=112962261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372074, CurrentID=0, Status='live', Symbol='LOC105372074', Aliases='', description='uncharacterized LOC105372074', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=38466643, end_position_on_the_genomic_accession=38490284, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129195, CurrentID=0, Status='live', Symbol='ZSCAN16-AS1', Aliases='', description='ZSCAN16 antisense RNA 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28121795, end_position_on_the_genomic_accession=28137316, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121260, CurrentID=0, Status='live', Symbol='SLC15A4', Aliases='FP12591, PHT1, PTR4', description='solute carrier family 15 (oligopeptide transporter), member 4', other_designations='hPHT1|peptide transporter 4|peptide-histidine transporter 4|peptide/histidine transporter 1|solute carrier family 15, member 4', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=128793194, end_position_on_the_genomic_accession=128823996, orientation='minus', exon_count=9, OMIM=615806, no_hdr0=''),
  105372077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372077, CurrentID=0, Status='live', Symbol='LOC105372077', Aliases='', description='uncharacterized LOC105372077', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=38784910, end_position_on_the_genomic_accession=38807562, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372078, CurrentID=0, Status='live', Symbol='LOC105372078', Aliases='', description='uncharacterized LOC105372078', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39275636, end_position_on_the_genomic_accession=39340995, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129199, CurrentID=0, Status='live', Symbol='RPS3AP40', Aliases='RPS3A_19_1151', description='ribosomal protein S3a pseudogene 40', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67925607, end_position_on_the_genomic_accession=67926455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372080, CurrentID=0, Status='live', Symbol='LOC105372080', Aliases='', description='uncharacterized LOC105372080', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39909543, end_position_on_the_genomic_accession=39924690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100653489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653489, CurrentID=0, Status='live', Symbol='MTND4P14', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 4 pseudogene 14', other_designations='MT-ND4 pseudogene 14', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5107937, end_position_on_the_genomic_accession=5109290, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372082, CurrentID=0, Status='live', Symbol='LOC105372082', Aliases='', description='uncharacterized LOC105372082', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=40246273, end_position_on_the_genomic_accession=40249626, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421363, CurrentID=0, Status='live', Symbol='ANKRD20A17P', Aliases='', description='ankyrin repeat domain 20 family member A17, pseudogene', other_designations='ankyrin repeat domain 20 family, member A17, pseudogene|ankyrin repeat domain 20 family, member A2 pseudogene', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49502148, end_position_on_the_genomic_accession=49506025, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  121268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121268, CurrentID=0, Status='live', Symbol='RHEBL1', Aliases='RHEBL1c', description='Ras homolog enriched in brain like 1', other_designations='Ras homolog enriched in brain like 1 c|ras homolog enriched in brain like-1 c|ras homolog enriched in brain-like protein 1|rheb-like protein 1|rheb2', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49064676, end_position_on_the_genomic_accession=49070025, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372085, CurrentID=0, Status='live', Symbol='LOC105372085', Aliases='', description='uncharacterized LOC105372085', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=41873593, end_position_on_the_genomic_accession=41908263, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645558, CurrentID=0, Status='live', Symbol='FAM90A22P', Aliases='FAM90A22', description='family with sequence similarity 90 member A22, pseudogene', other_designations='family with sequence similarity 90, member A22, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7571433, end_position_on_the_genomic_accession=7576669, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372087, CurrentID=0, Status='live', Symbol='LOC105372087', Aliases='', description='uncharacterized LOC105372087', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=43884165, end_position_on_the_genomic_accession=43889199, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645560, CurrentID=0, Status='live', Symbol='ATP5C1P1', Aliases='ATP5C2, ATP5CL2', description='ATP synthase, H+ transporting, mitochondrial F1 complex, gamma polypeptide 1 pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, gamma polypeptide 2|ATP synthase, H+ transporting, mitochondrial F1 complex, gamma polypeptide-like 2', map_location='14q22.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=53990931, end_position_on_the_genomic_accession=53992023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  121273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121273, CurrentID=0, Status='live', Symbol='C12orf54', Aliases='HSD-29, HSD-30', description='chromosome 12 open reading frame 54', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48455581, end_position_on_the_genomic_accession=48496514, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  121274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121274, CurrentID=0, Status='live', Symbol='ZNF641', Aliases='', description='zinc finger protein 641', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48335491, end_position_on_the_genomic_accession=48351414, orientation='minus', exon_count=9, OMIM=613906, no_hdr0=''),
  121275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121275, CurrentID=0, Status='live', Symbol='OR10AD1', Aliases='OR10AD1P, OR12-1', description='olfactory receptor family 10 subfamily AD member 1', other_designations='olfactory receptor OR12-1|olfactory receptor, family 10, subfamily AD, member 1 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48201772, end_position_on_the_genomic_accession=48205682, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372092, CurrentID=0, Status='live', Symbol='LOC105372092', Aliases='', description='uncharacterized LOC105372092', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=45713194, end_position_on_the_genomic_accession=45723234, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129213, CurrentID=0, Status='live', Symbol='SLC39A12-AS1', Aliases='', description='SLC39A12 antisense RNA 1', other_designations='', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18001786, end_position_on_the_genomic_accession=18010562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  121278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121278, CurrentID=0, Status='live', Symbol='TPH2', Aliases='ADHD7, NTPH', description='tryptophan hydroxylase 2', other_designations='neuronal tryptophan hydroxylase|tryptophan 5-monooxygenase 2', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=71938846, end_position_on_the_genomic_accession=72035297, orientation='plus', exon_count=14, OMIM=607478, no_hdr0=''),
  105372095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372095, CurrentID=0, Status='live', Symbol='LOC105372095', Aliases='', description='uncharacterized LOC105372095', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46388107, end_position_on_the_genomic_accession=46389512, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129216, CurrentID=0, Status='live', Symbol='LOC100129216', Aliases='', description='beta-defensin 131-like', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71878453, end_position_on_the_genomic_accession=71884561, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372097, CurrentID=0, Status='live', Symbol='LOC105372097', Aliases='', description='uncharacterized LOC105372097', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46757443, end_position_on_the_genomic_accession=46769135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372098, CurrentID=0, Status='live', Symbol='LOC105372098', Aliases='', description='uncharacterized LOC105372098', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47077367, end_position_on_the_genomic_accession=47100870, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  645571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645571, CurrentID=0, Status='live', Symbol='PRADC1P1', Aliases='', description='protease-associated domain containing 1 pseudogene 1', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36976233, end_position_on_the_genomic_accession=36977040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645572, CurrentID=0, Status='live', Symbol='FAM90A23P', Aliases='FAM90A23', description='family with sequence similarity 90 member A23, pseudogene', other_designations='family with sequence similarity 90, member A23, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7579082, end_position_on_the_genomic_accession=7584313, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645573, CurrentID=0, Status='live', Symbol='RPS4XP19', Aliases='RPS4P19, RPS4X_9_1583', description='ribosomal protein S4X pseudogene 19', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9019998, end_position_on_the_genomic_accession=9020826, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372102, CurrentID=0, Status='live', Symbol='LOC105372102', Aliases='', description='uncharacterized LOC105372102', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48362582, end_position_on_the_genomic_accession=48396796, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129223, CurrentID=0, Status='live', Symbol='THCAT155', Aliases='', description='thyroid cancer-associated transcript 155', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3483712, end_position_on_the_genomic_accession=3492819, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372104, CurrentID=0, Status='live', Symbol='LOC105372104', Aliases='', description='uncharacterized LOC105372104', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48162404, end_position_on_the_genomic_accession=48185405, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372105, CurrentID=0, Status='live', Symbol='LOC105372105', Aliases='', description='uncharacterized LOC105372105', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48817796, end_position_on_the_genomic_accession=48841796, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372106, CurrentID=0, Status='live', Symbol='LOC105372106', Aliases='', description='uncharacterized LOC105372106', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48794110, end_position_on_the_genomic_accession=48798095, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100653515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100653515, CurrentID=0, Status='live', Symbol='CEP295NL', Aliases='DDC8, KIAA1731NL', description='CEP295 N-terminal like', other_designations='KIAA1731 N-terminal like protein|differential display clone 8', map_location='17q25', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78890579, end_position_on_the_genomic_accession=78903217, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372108, CurrentID=0, Status='live', Symbol='LOC105372108', Aliases='', description='uncharacterized LOC105372108', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=48962086, end_position_on_the_genomic_accession=48964745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129229, CurrentID=0, Status='live', Symbol='HNRNPCP10', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 10', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=144840777, end_position_on_the_genomic_accession=144841660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372110, CurrentID=0, Status='live', Symbol='LOC105372110', Aliases='', description='uncharacterized LOC105372110', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49016027, end_position_on_the_genomic_accession=49023814, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129231, CurrentID=0, Status='live', Symbol='HSPA8P20', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 20', other_designations='heat shock 70kDa protein 1-like pseudogene|heat shock 70kDa protein 8 pseudogene 20', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124204748, end_position_on_the_genomic_accession=124205807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372112, CurrentID=0, Status='live', Symbol='LOC105372112', Aliases='', description='uncharacterized LOC105372112', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49612709, end_position_on_the_genomic_accession=49738560, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372113, CurrentID=0, Status='live', Symbol='LOC105372113', Aliases='', description='uncharacterized LOC105372113', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50142300, end_position_on_the_genomic_accession=50144992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129234, CurrentID=0, Status='live', Symbol='RPL27P7', Aliases='RPL27_3_270', description='ribosomal protein L27 pseudogene 7', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119993168, end_position_on_the_genomic_accession=119993639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645587, CurrentID=0, Status='live', Symbol='LINC00614', Aliases='C10orf52', description='long intergenic non-protein coding RNA 614', other_designations='Putative uncharacterized protein C10orf52', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=27295734, end_position_on_the_genomic_accession=27300833, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372116, CurrentID=0, Status='live', Symbol='LOC105372116', Aliases='', description='uncharacterized LOC105372116', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50748169, end_position_on_the_genomic_accession=50752661, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129237, CurrentID=0, Status='live', Symbol='NPM1P21', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 21', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56100911, end_position_on_the_genomic_accession=56102361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372118, CurrentID=0, Status='live', Symbol='LOC105372118', Aliases='', description='uncharacterized LOC105372118', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51192906, end_position_on_the_genomic_accession=51196548, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129239, CurrentID=0, Status='live', Symbol='CXorf51A', Aliases='CXorf51', description='chromosome X open reading frame 51A', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146814104, end_position_on_the_genomic_accession=146814731, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101926930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101926930, CurrentID=0, Status='live', Symbol='LINC00378', Aliases='', description='long intergenic non-protein coding RNA 378', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60672955, end_position_on_the_genomic_accession=60695806, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372121, CurrentID=0, Status='live', Symbol='LOC105372121', Aliases='', description='uncharacterized LOC105372121', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=52048255, end_position_on_the_genomic_accession=52217260, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372122, CurrentID=0, Status='live', Symbol='LOC105372122', Aliases='', description='uncharacterized LOC105372122', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=53884526, end_position_on_the_genomic_accession=53887309, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129243, CurrentID=0, Status='live', Symbol='RPS15AP12', Aliases='RPS15A_5_166', description='ribosomal protein S15a pseudogene 12', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220143918, end_position_on_the_genomic_accession=220144376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372124, CurrentID=0, Status='live', Symbol='LOC105372124', Aliases='', description='uncharacterized LOC105372124', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54987756, end_position_on_the_genomic_accession=54993394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372125, CurrentID=0, Status='live', Symbol='LOC105372125', Aliases='', description='uncharacterized LOC105372125', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55079050, end_position_on_the_genomic_accession=55085363, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645598, CurrentID=0, Status='live', Symbol='WIZP1', Aliases='', description='widely interspaced zinc finger motifs pseudogene 1', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23403807, end_position_on_the_genomic_accession=23406102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372127, CurrentID=0, Status='live', Symbol='TCF4-AS1', Aliases='', description='TCF4 antisense RNA 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55480434, end_position_on_the_genomic_accession=55482944, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129248, CurrentID=0, Status='live', Symbol='PDSS1P1', Aliases='', description='prenyl (decaprenyl) diphosphate synthase, subunit 1 pseudogene 1', other_designations='polyprenyl pyrophosphate synthetase pseudogene', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=5084690, end_position_on_the_genomic_accession=5086266, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372129, CurrentID=0, Status='live', Symbol='LOC105372129', Aliases='', description='uncharacterized LOC105372129', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=55294262, end_position_on_the_genomic_accession=55297148, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129250, CurrentID=0, Status='live', Symbol='TOPORS-AS1', Aliases='C9orf133', description='TOPORS antisense RNA 1', other_designations='TOPORS antisense RNA 1 (non-protein coding)', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32551144, end_position_on_the_genomic_accession=32566860, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129251, CurrentID=0, Status='live', Symbol='FTLP4', Aliases='', description='ferritin, light polypeptide pseudogene 4', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30934587, end_position_on_the_genomic_accession=30936005, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372132, CurrentID=0, Status='live', Symbol='LOC105372132', Aliases='', description='uncharacterized LOC105372132', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56137599, end_position_on_the_genomic_accession=56150102, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372133, CurrentID=0, Status='live', Symbol='LOC105372133', Aliases='', description='uncharacterized LOC105372133', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56059828, end_position_on_the_genomic_accession=56063078, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372134, CurrentID=0, Status='live', Symbol='LOC105372134', Aliases='', description='uncharacterized LOC105372134', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56259892, end_position_on_the_genomic_accession=56306892, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372135, CurrentID=0, Status='live', Symbol='LOC105372135', Aliases='', description='uncharacterized LOC105372135', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=56573518, end_position_on_the_genomic_accession=56575165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129256, CurrentID=0, Status='live', Symbol='NMNAT1P5', Aliases='', description='NMNAT1 pseudogene 5', other_designations='nicotinamide nucleotide adenylyltransferase 1 pseudogene', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59964942, end_position_on_the_genomic_accession=59966079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372137, CurrentID=0, Status='live', Symbol='LOC105372137', Aliases='', description='uncharacterized LOC105372137', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57158836, end_position_on_the_genomic_accession=57169318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372138, CurrentID=0, Status='live', Symbol='LOC105372138', Aliases='', description='uncharacterized LOC105372138', other_designations='CTD-2526M8.2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57169403, end_position_on_the_genomic_accession=57334332, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129259, CurrentID=0, Status='live', Symbol='BRD7P5', Aliases='', description='bromodomain containing 7 pseudogene 5', other_designations='', map_location='13q32.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=94757310, end_position_on_the_genomic_accession=94759417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129260, CurrentID=0, Status='live', Symbol='RPL7L1P1', Aliases='', description='ribosomal protein L7 like 1 pseudogene 1', other_designations='ribosomal protein L7-like 1 pseudogene 1', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89833301, end_position_on_the_genomic_accession=89834052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372141, CurrentID=0, Status='live', Symbol='LOC105372141', Aliases='', description='uncharacterized LOC105372141', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57731839, end_position_on_the_genomic_accession=57735388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372142, CurrentID=0, Status='live', Symbol='LOC105372142', Aliases='', description='uncharacterized LOC105372142', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57686553, end_position_on_the_genomic_accession=57731619, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372143, CurrentID=0, Status='live', Symbol='LOC105372143', Aliases='', description='uncharacterized LOC105372143', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58188247, end_position_on_the_genomic_accession=58195710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  121328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121328, CurrentID=0, Status='live', Symbol='GGTA2P', Aliases='GGTA1P, GLYT2L, GLYT3, HGT2', description='glycoprotein, alpha-galactosyltransferase 2, pseudogene', other_designations='alpha-1,3-galactosyltransferase pseudogene|glycoprotein, alpha-galactosyltransferase 1 pseudogene', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=67265839, end_position_on_the_genomic_accession=67266962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372145, CurrentID=0, Status='live', Symbol='LOC105372145', Aliases='', description='uncharacterized LOC105372145', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58557022, end_position_on_the_genomic_accession=58566678, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372146, CurrentID=0, Status='live', Symbol='LOC105372146', Aliases='', description='uncharacterized LOC105372146', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58671923, end_position_on_the_genomic_accession=58780234, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645619, CurrentID=0, Status='live', Symbol='AK4P3', Aliases='AK3L2, AK3L2P', description='adenylate kinase 4 pseudogene 3', other_designations='adenylate kinase 3-like 2 (pseudogene)', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31614443, end_position_on_the_genomic_accession=31616574, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372148, CurrentID=0, Status='live', Symbol='LOC105372148', Aliases='', description='uncharacterized LOC105372148', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59685826, end_position_on_the_genomic_accession=59690928, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129269, CurrentID=0, Status='live', Symbol='LINC01160', Aliases='', description='long intergenic non-protein coding RNA 1160', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111599007, end_position_on_the_genomic_accession=111608318, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372150, CurrentID=0, Status='live', Symbol='LOC105372150', Aliases='', description='uncharacterized LOC105372150', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59516653, end_position_on_the_genomic_accession=59521996, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129271, CurrentID=0, Status='live', Symbol='C1orf68', Aliases='LEP7, XP32', description='chromosome 1 open reading frame 68', other_designations='skin-specific protein (xp32)', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152719522, end_position_on_the_genomic_accession=152720429, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129272, CurrentID=0, Status='live', Symbol='TOMM20P4', Aliases='', description='TOMM20 pseudogene 4', other_designations='translocase of outer mitochondrial membrane 20 homolog pseudogene 4', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73223011, end_position_on_the_genomic_accession=73223758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372153, CurrentID=0, Status='live', Symbol='LOC105372153', Aliases='', description='uncharacterized LOC105372153', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60167697, end_position_on_the_genomic_accession=60173163, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129274, CurrentID=0, Status='live', Symbol='LOC100129274', Aliases='', description='NGG1 interacting factor 3 like 1 pseudogene', other_designations='NIF3 NGG1 interacting factor 3-like 1 (S. cerevisiae) pseudogene', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=187714199, end_position_on_the_genomic_accession=187715322, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372155, CurrentID=0, Status='live', Symbol='LOC105372155', Aliases='', description='uncharacterized LOC105372155', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60294440, end_position_on_the_genomic_accession=60301145, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  121340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121340, CurrentID=0, Status='live', Symbol='SP7', Aliases='OI11, OI12, OSX, osterix', description='Sp7 transcription factor', other_designations='zinc finger protein osterix', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53326575, end_position_on_the_genomic_accession=53344793, orientation='minus', exon_count=5, OMIM=606633, no_hdr0=''),
  100129277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129277, CurrentID=0, Status='live', Symbol='TPMTP3', Aliases='', description='thiopurine S-methyltransferase pseudogene 3', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51979073, end_position_on_the_genomic_accession=51980058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129278, CurrentID=0, Status='live', Symbol='TDRD15', Aliases='', description='tudor domain containing 15', other_designations='tudor domain-containing protein 6-like', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21106652, end_position_on_the_genomic_accession=21147393, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129279, CurrentID=0, Status='live', Symbol='RPS27P14', Aliases='RPS27_7_572', description='ribosomal protein S27 pseudogene 14', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69469852, end_position_on_the_genomic_accession=69470187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372736, CurrentID=0, Status='live', Symbol='LOC105372736', Aliases='', description='uncharacterized LOC105372736', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372161, CurrentID=0, Status='live', Symbol='LOC105372161', Aliases='', description='uncharacterized LOC105372161', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63088532, end_position_on_the_genomic_accession=63097162, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372162, CurrentID=0, Status='live', Symbol='LOC105372162', Aliases='', description='uncharacterized LOC105372162', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63226872, end_position_on_the_genomic_accession=63228857, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129283, CurrentID=0, Status='live', Symbol='LDHBP3', Aliases='', description='lactate dehydrogenase B pseudogene 3', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=92840837, end_position_on_the_genomic_accession=92841624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129284, CurrentID=0, Status='live', Symbol='RPS15AP39', Aliases='RPS15A_14_1774', description='ribosomal protein S15a pseudogene 39', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41748183, end_position_on_the_genomic_accession=41748665, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372165, CurrentID=0, Status='live', Symbol='LOC105372165', Aliases='', description='uncharacterized LOC105372165', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=63966381, end_position_on_the_genomic_accession=63969891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645638, CurrentID=0, Status='live', Symbol='WFDC21P', Aliases='LNCDC, linc-DC, lnc-DC', description='WAP four-disulfide core domain 21, pseudogene', other_designations='WDNM1-like pseudogene|WDNM1-like, pseudogene|long non-coding RNA, dendritic cell differentiation', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60083566, end_position_on_the_genomic_accession=60088467, orientation='minus', exon_count=3, OMIM=615772, no_hdr0=''),
  105372167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372167, CurrentID=0, Status='live', Symbol='LOC105372167', Aliases='', description='uncharacterized LOC105372167', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=64630290, end_position_on_the_genomic_accession=64631350, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129288, CurrentID=0, Status='live', Symbol='MORF4L1P5', Aliases='MORF4LP5', description='mortality factor 4 like 1 pseudogene 5', other_designations='mortality factor 4 like pseudogene 5', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65327991, end_position_on_the_genomic_accession=65331194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645641, CurrentID=0, Status='live', Symbol='MESTP3', Aliases='', description='mesoderm specific transcript pseudogene 3', other_designations='mesoderm specific transcript homolog pseudogene 3', map_location='4p15.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28823247, end_position_on_the_genomic_accession=28824183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372170, CurrentID=0, Status='live', Symbol='LOC105372170', Aliases='', description='uncharacterized LOC105372170', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=66110545, end_position_on_the_genomic_accession=66111996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121355, CurrentID=0, Status='live', Symbol='GTSF1', Aliases='FAM112B', description='gametocyte specific factor 1', other_designations='family with sequence similarity 112, member B', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=54455952, end_position_on_the_genomic_accession=54473602, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  645644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645644, CurrentID=0, Status='live', Symbol='FLJ42627', Aliases='', description='uncharacterized LOC645644', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2638982, end_position_on_the_genomic_accession=2646129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372173, CurrentID=0, Status='live', Symbol='LOC105372173', Aliases='', description='uncharacterized LOC105372173', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=67672261, end_position_on_the_genomic_accession=67767068, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645646, CurrentID=0, Status='live', Symbol='FAM32CP', Aliases='FAM32C', description='family with sequence similarity 32 member C, pseudogene', other_designations='family with sequence similarity 32 member C (pseudogene)', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74743201, end_position_on_the_genomic_accession=74744802, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372175, CurrentID=0, Status='live', Symbol='LOC105372175', Aliases='', description='uncharacterized LOC105372175', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68258078, end_position_on_the_genomic_accession=68264803, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  121360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121360, CurrentID=0, Status='live', Symbol='OR9K1P', Aliases='OR12-4', description='olfactory receptor family 9 subfamily K member 1 pseudogene', other_designations='olfactory receptor OR12-4 pseudogene|olfactory receptor, family 9, subfamily K, member 1 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55115534, end_position_on_the_genomic_accession=55116472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129297, CurrentID=0, Status='live', Symbol='LOC100129297', Aliases='', description='uncharacterized LOC100129297', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129298, CurrentID=0, Status='live', Symbol='TPMTP4', Aliases='', description='thiopurine S-methyltransferase pseudogene 4', other_designations='', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86128826, end_position_on_the_genomic_accession=86129630, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645651, CurrentID=0, Status='live', Symbol='FAM90A14P', Aliases='FAM90A14', description='family with sequence similarity 90 member A14, pseudogene', other_designations='family with sequence similarity 90, member A14, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7713783, end_position_on_the_genomic_accession=7719015, orientation='plus', exon_count=0, OMIM=613050, no_hdr0=''),
  121364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121364, CurrentID=0, Status='live', Symbol='OR10A7', Aliases='OR12-6', description='olfactory receptor family 10 subfamily A member 7', other_designations='olfactory receptor OR12-6|olfactory receptor, family 10, subfamily A, member 7', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55221025, end_position_on_the_genomic_accession=55221975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372181, CurrentID=0, Status='live', Symbol='LOC105372181', Aliases='', description='uncharacterized LOC105372181', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70411987, end_position_on_the_genomic_accession=70430178, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645654, CurrentID=0, Status='live', Symbol='RPL31P48', Aliases='RPL31_25_1257', description='ribosomal protein L31 pseudogene 48', other_designations='', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=73648640, end_position_on_the_genomic_accession=73649086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101060144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060144, CurrentID=0, Status='live', Symbol='MED28P8', Aliases='', description='mediator complex subunit 28 pseudogene 8', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81557117, end_position_on_the_genomic_accession=81558520, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645656, CurrentID=0, Status='live', Symbol='RPSAP56', Aliases='RPSA_23_1496', description='ribosomal protein SA pseudogene 56', other_designations='', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73940567, end_position_on_the_genomic_accession=73941602, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372185, CurrentID=0, Status='live', Symbol='LOC105372185', Aliases='', description='uncharacterized LOC105372185', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70773435, end_position_on_the_genomic_accession=70786753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129306, CurrentID=0, Status='live', Symbol='RPS20P32', Aliases='RPS20_15_1306', description='ribosomal protein S20 pseudogene 32', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27017424, end_position_on_the_genomic_accession=27017935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129307, CurrentID=0, Status='live', Symbol='LOC100129307', Aliases='', description='putative UPF0607 protein ENSP00000383144', other_designations='UPF0607 protein ENSP00000383144-like', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77661374, end_position_on_the_genomic_accession=77662803, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129308, CurrentID=0, Status='live', Symbol='DNAJA1P1', Aliases='', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 1', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 1', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=58750914, end_position_on_the_genomic_accession=58752355, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372189, CurrentID=0, Status='live', Symbol='LOC105372189', Aliases='', description='uncharacterized LOC105372189', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=72147669, end_position_on_the_genomic_accession=72184725, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  25864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25864, CurrentID=0, Status='live', Symbol='ABHD14A', Aliases='DORZ1', description='abhydrolase domain containing 14A', other_designations='abhydrolase domain-containing protein 14A', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51975026, end_position_on_the_genomic_accession=51981200, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372191, CurrentID=0, Status='live', Symbol='LOC105372191', Aliases='', description='uncharacterized LOC105372191', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=73756384, end_position_on_the_genomic_accession=73784702, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129312, CurrentID=0, Status='live', Symbol='LOC100129312', Aliases='', description='uncharacterized LOC100129312', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=174562734, end_position_on_the_genomic_accession=174563102, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  645626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645626, CurrentID=0, Status='live', Symbol='LOC645626', Aliases='', description='coiled-coil domain containing 29-like', other_designations='ankyrin repeat domain 20 family, member A4 pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18811143, end_position_on_the_genomic_accession=18837340, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372194, CurrentID=0, Status='live', Symbol='LOC105372194', Aliases='', description='uncharacterized LOC105372194', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75070197, end_position_on_the_genomic_accession=75073950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129315, CurrentID=0, Status='live', Symbol='VN1R66P', Aliases='', description='vomeronasal 1 receptor 66 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31648318, end_position_on_the_genomic_accession=31649231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645668, CurrentID=0, Status='live', Symbol='TUFMP1', Aliases='TUFML', description='Tu translation elongation factor, mitochondrial pseudogene 1', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27082492, end_position_on_the_genomic_accession=27084123, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372197, CurrentID=0, Status='live', Symbol='LOC105372197', Aliases='', description='uncharacterized LOC105372197', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74731558, end_position_on_the_genomic_accession=74741271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372198, CurrentID=0, Status='live', Symbol='LOC105372198', Aliases='', description='uncharacterized LOC105372198', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75112183, end_position_on_the_genomic_accession=75125551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129319, CurrentID=0, Status='live', Symbol='SLC25A6P5', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene 5', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105849710, end_position_on_the_genomic_accession=105850593, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129320, CurrentID=0, Status='live', Symbol='HMGB3P10', Aliases='', description='high mobility group box 3 pseudogene 10', other_designations='high-mobility group box 3 pseudogene 10', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99698215, end_position_on_the_genomic_accession=99699197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372201, CurrentID=0, Status='live', Symbol='LOC105372201', Aliases='', description='uncharacterized LOC105372201', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75472804, end_position_on_the_genomic_accession=75478187, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372202, CurrentID=0, Status='live', Symbol='LOC105372202', Aliases='', description='uncharacterized LOC105372202', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75506283, end_position_on_the_genomic_accession=75509972, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372203, CurrentID=0, Status='live', Symbol='LOC105372203', Aliases='', description='uncharacterized LOC105372203', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76005038, end_position_on_the_genomic_accession=76015412, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645676, CurrentID=0, Status='live', Symbol='ASH1L-AS1', Aliases='', description='ASH1L antisense RNA 1', other_designations='ASH1L antisense RNA 1 (non-protein coding)', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155561981, end_position_on_the_genomic_accession=155563944, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129325, CurrentID=0, Status='live', Symbol='HNRNPA1P64', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 64', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=80451148, end_position_on_the_genomic_accession=80451997, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372206, CurrentID=0, Status='live', Symbol='LOC105372206', Aliases='', description='uncharacterized LOC105372206', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76197309, end_position_on_the_genomic_accession=76207220, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121391, CurrentID=0, Status='live', Symbol='KRT74', Aliases='ADWH, HTSS2, HYPT3, K6IRS4, KRT5C, KRT6IRS4', description='keratin 74, type II', other_designations='CK-74|K5C|K74|cytokeratin-74|keratin 5c|keratin 6 irs4|keratin-5c|keratin-74|type II inner root sheath-specific keratin-K6irs4|type-II keratin Kb37', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=52565722, end_position_on_the_genomic_accession=52574092, orientation='minus', exon_count=10, OMIM=608248, no_hdr0=''),
  105372208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372208, CurrentID=0, Status='live', Symbol='LOC105372208', Aliases='', description='uncharacterized LOC105372208', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76218199, end_position_on_the_genomic_accession=76221719, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129329, CurrentID=0, Status='live', Symbol='SNRPEP3', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 3', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20238817, end_position_on_the_genomic_accession=20239304, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645682, CurrentID=0, Status='live', Symbol='POU5F1P4', Aliases='Oct4-pg4, POU5FLC1', description='POU class 5 homeobox 1 pseudogene 4', other_designations='POU domain class 5, transcription factor 1 pseudogene 4|POU domain transcription factor OCT4-pg4', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155433180, end_position_on_the_genomic_accession=155434262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645683, CurrentID=0, Status='live', Symbol='RPL13AP3', Aliases='RPL13A_11_1370', description='ribosomal protein L13a pseudogene 3', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55766245, end_position_on_the_genomic_accession=55767717, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129332, CurrentID=0, Status='live', Symbol='RPL29P20', Aliases='RPL29_10_1009', description='ribosomal protein L29 pseudogene 20', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112173414, end_position_on_the_genomic_accession=112174013, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372213, CurrentID=0, Status='live', Symbol='LOC105372213', Aliases='', description='uncharacterized LOC105372213', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76722364, end_position_on_the_genomic_accession=76727967, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645686, CurrentID=0, Status='live', Symbol='GTF2IP3', Aliases='', description='general transcription factor IIi pseudogene 3', other_designations='general transcription factor II, i pseudogene|general transcription factor IIi, pseudogene 3', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19026001, end_position_on_the_genomic_accession=19032626, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645687, CurrentID=0, Status='live', Symbol='LINC00520', Aliases='C14orf34', description='long intergenic non-protein coding RNA 520', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55781135, end_position_on_the_genomic_accession=55796674, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645688, CurrentID=0, Status='live', Symbol='RPL12P38', Aliases='RPL12_14_1564', description='ribosomal protein L12 pseudogene 38', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60433678, end_position_on_the_genomic_accession=60434300, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129337, CurrentID=0, Status='live', Symbol='EIF4BP2', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 2', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21028983, end_position_on_the_genomic_accession=21030793, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372218, CurrentID=0, Status='live', Symbol='LOC105372218', Aliases='', description='uncharacterized LOC105372218', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=77974558, end_position_on_the_genomic_accession=77984809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129339, CurrentID=0, Status='live', Symbol='TPSP1', Aliases='MP-2', description='tryptase pseudogene 1', other_designations='tryptase delta 1 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1273724, end_position_on_the_genomic_accession=1276395, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129340, CurrentID=0, Status='live', Symbol='LOC100129340', Aliases='', description='mitofusin-1-like', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88187665, end_position_on_the_genomic_accession=88194459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645693, CurrentID=0, Status='live', Symbol='EEF1A1P22', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 22', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52937139, end_position_on_the_genomic_accession=52938518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645694, CurrentID=0, Status='live', Symbol='HTR5BP', Aliases='5-HT5B, GPR134, HTR5B', description='5-hydroxytryptamine (serotonin) receptor 5B, pseudogene', other_designations='5-hydroxytryptamine (serotonin) receptor 5 pseudogene|5-hydroxytryptamine (serotonin) receptor 5A pseudogene|G protein-coupled receptor 134', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117859427, end_position_on_the_genomic_accession=117903680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372223, CurrentID=0, Status='live', Symbol='LOC105372223', Aliases='', description='uncharacterized LOC105372223', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78816095, end_position_on_the_genomic_accession=78818911, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372224, CurrentID=0, Status='live', Symbol='LOC105372224', Aliases='', description='uncharacterized LOC105372224', other_designations='CTD-2382H12.1|CTD-2382H12.2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78908612, end_position_on_the_genomic_accession=78927373, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129345, CurrentID=0, Status='live', Symbol='LOC100129345', Aliases='', description='uncharacterized LOC100129345', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97632647, end_position_on_the_genomic_accession=97686658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372226, CurrentID=0, Status='live', Symbol='LOC105372226', Aliases='', description='uncharacterized LOC105372226', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79272306, end_position_on_the_genomic_accession=79300810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129347, CurrentID=0, Status='live', Symbol='LINC00484', Aliases='C9orf73', description='long intergenic non-protein coding RNA 484', other_designations='', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91118605, end_position_on_the_genomic_accession=91182719, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  645700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645700, CurrentID=0, Status='live', Symbol='ZNF890P', Aliases='', description='zinc finger protein 890, pseudogene', other_designations='zinc finger protein 33A pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5121310, end_position_on_the_genomic_accession=5144546, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372229, CurrentID=0, Status='live', Symbol='LOC105372229', Aliases='', description='uncharacterized LOC105372229', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79582886, end_position_on_the_genomic_accession=79586633, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372230, CurrentID=0, Status='live', Symbol='LOC105372230', Aliases='', description='uncharacterized LOC105372230', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80179224, end_position_on_the_genomic_accession=80185369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129351, CurrentID=0, Status='live', Symbol='AGGF1P4', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 4', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 4', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35389203, end_position_on_the_genomic_accession=35390044, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372232, CurrentID=0, Status='live', Symbol='LOC105372232', Aliases='', description='uncharacterized LOC105372232', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=490310, end_position_on_the_genomic_accession=507459, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372233, CurrentID=0, Status='live', Symbol='LOC105372233', Aliases='', description='uncharacterized LOC105372233', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=567433, end_position_on_the_genomic_accession=572148, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129354, CurrentID=0, Status='live', Symbol='NRADDP', Aliases='', description='neurotrophin receptor associated death domain, pseudogene', other_designations='NRADD pseudogene', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47011542, end_position_on_the_genomic_accession=47013467, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129355, CurrentID=0, Status='live', Symbol='BOD1P2', Aliases='', description='biorientation of chromosomes in cell division 1 pseudogene 2', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=5131200, end_position_on_the_genomic_accession=5132226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372236, CurrentID=0, Status='live', Symbol='LOC105372236', Aliases='', description='uncharacterized LOC105372236', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1262102, end_position_on_the_genomic_accession=1264602, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129357, CurrentID=0, Status='live', Symbol='RPL9P29', Aliases='RPL9_17_1576', description='ribosomal protein L9 pseudogene 29', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78845847, end_position_on_the_genomic_accession=78846437, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130503, CurrentID=0, Status='live', Symbol='LOC100130503', Aliases='', description='uncharacterized LOC100130503', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=35677837, end_position_on_the_genomic_accession=35683573, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372239, CurrentID=0, Status='live', Symbol='LOC105372239', Aliases='', description='uncharacterized LOC105372239', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2214471, end_position_on_the_genomic_accession=2221969, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372240, CurrentID=0, Status='live', Symbol='LOC105372240', Aliases='', description='uncharacterized LOC105372240', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2268493, end_position_on_the_genomic_accession=2269421, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129361, CurrentID=0, Status='live', Symbol='SMIM10L1', Aliases='', description='small integral membrane protein 10 like 1', other_designations='Small integral membrane protein 10-like protein 1|chromosome X open reading frame 69-like', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11171181, end_position_on_the_genomic_accession=11176020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372242, CurrentID=0, Status='live', Symbol='LOC105372242', Aliases='', description='uncharacterized LOC105372242', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3065622, end_position_on_the_genomic_accession=3075468, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  645715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645715, CurrentID=0, Status='live', Symbol='EEF1A1P24', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 24', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39358545, end_position_on_the_genomic_accession=39359919, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372244, CurrentID=0, Status='live', Symbol='LOC105372244', Aliases='', description='uncharacterized LOC105372244', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3296765, end_position_on_the_genomic_accession=3306949, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101928718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928718, CurrentID=0, Status='live', Symbol='LOC101928718', Aliases='', description='uncharacterized LOC101928718', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111745317, end_position_on_the_genomic_accession=111755509, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372246, CurrentID=0, Status='live', Symbol='LOC105372246', Aliases='', description='uncharacterized LOC105372246', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3481894, end_position_on_the_genomic_accession=3490729, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372247, CurrentID=0, Status='live', Symbol='LOC105372247', Aliases='', description='uncharacterized LOC105372247', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3668168, end_position_on_the_genomic_accession=3674402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129368, CurrentID=0, Status='live', Symbol='HNRNPH1P2', Aliases='', description='heterogeneous nuclear ribonucleoprotein H1 pseudogene 2', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143677152, end_position_on_the_genomic_accession=143678361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372249, CurrentID=0, Status='live', Symbol='LOC105372249', Aliases='', description='uncharacterized LOC105372249', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4580696, end_position_on_the_genomic_accession=4586023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129370, CurrentID=0, Status='live', Symbol='NDUFB9P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B9 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 9, 22kDa pseudogene', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=110809373, end_position_on_the_genomic_accession=110809894, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372251, CurrentID=0, Status='live', Symbol='LOC105372251', Aliases='', description='uncharacterized LOC105372251', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4776494, end_position_on_the_genomic_accession=4782243, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372252, CurrentID=0, Status='live', Symbol='LOC105372252', Aliases='', description='uncharacterized LOC105372252', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5336447, end_position_on_the_genomic_accession=5338448, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372253, CurrentID=0, Status='live', Symbol='LOC105372253', Aliases='', description='uncharacterized LOC105372253', other_designations='LLNLR-304G9.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5293413, end_position_on_the_genomic_accession=5294681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129374, CurrentID=0, Status='live', Symbol='RPL18P3', Aliases='RPL18_2_608', description='ribosomal protein L18 pseudogene 3', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121616215, end_position_on_the_genomic_accession=121616857, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372255, CurrentID=0, Status='live', Symbol='LOC105372255', Aliases='', description='uncharacterized LOC105372255', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6126373, end_position_on_the_genomic_accession=6199572, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372256, CurrentID=0, Status='live', Symbol='LOC105372256', Aliases='', description='uncharacterized LOC105372256', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6901992, end_position_on_the_genomic_accession=6938790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121441, CurrentID=0, Status='live', Symbol='NEDD1', Aliases='GCP-WD, TUBGCP7', description='neural precursor cell expressed, developmentally down-regulated 1', other_designations='NEDD-1|neural precursor cell expressed developmentally down-regulated protein 1', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96907223, end_position_on_the_genomic_accession=96953696, orientation='plus', exon_count=17, OMIM=600372, no_hdr0=''),
  105372258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372258, CurrentID=0, Status='live', Symbol='LOC105372258', Aliases='', description='uncharacterized LOC105372258', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7006642, end_position_on_the_genomic_accession=7012524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129379, CurrentID=0, Status='live', Symbol='RPS29P17', Aliases='RPS29_13_976', description='ribosomal protein S29 pseudogene 17', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35592747, end_position_on_the_genomic_accession=35593044, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372260, CurrentID=0, Status='live', Symbol='LOC105372260', Aliases='', description='protein capicua homolog', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372261, CurrentID=0, Status='live', Symbol='LOC105372261', Aliases='', description='uncharacterized LOC105372261', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7514353, end_position_on_the_genomic_accession=7524954, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129382, CurrentID=0, Status='live', Symbol='PRKRIRP4', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 4', other_designations='', map_location='11p14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23478373, end_position_on_the_genomic_accession=23480730, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372263, CurrentID=0, Status='live', Symbol='LOC105372263', Aliases='', description='uncharacterized LOC105372263', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7770622, end_position_on_the_genomic_accession=7785167, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372264, CurrentID=0, Status='live', Symbol='LOC105372264', Aliases='', description='uncharacterized LOC105372264', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7902649, end_position_on_the_genomic_accession=7903716, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129385, CurrentID=0, Status='live', Symbol='C9orf92', Aliases='Em:AL513424.1', description='chromosome 9 open reading frame 92', other_designations='', map_location='9p22.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=16203935, end_position_on_the_genomic_accession=16276313, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372266, CurrentID=0, Status='live', Symbol='LOC105372266', Aliases='', description='uncharacterized LOC105372266', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=8008390, end_position_on_the_genomic_accession=8015581, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129387, CurrentID=0, Status='live', Symbol='GABPB1-AS1', Aliases='', description='GABPB1 antisense RNA 1', other_designations='GABPB1 antisense RNA 1 (non-protein coding)', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50354174, end_position_on_the_genomic_accession=50358312, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645740, CurrentID=0, Status='live', Symbol='NOS2P1', Aliases='NOS2C', description='nitric oxide synthase 2 pseudogene 1', other_designations='inducible nitric oxide synthase pseudogene|nitric oxide synthase 2, inducible pseudogene|nitric oxide synthase 2C (pseudogene)', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27651441, end_position_on_the_genomic_accession=27664653, orientation='minus', exon_count=0, OMIM=600720, no_hdr0=''),
  105372269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372269, CurrentID=0, Status='live', Symbol='LOC105372269', Aliases='', description='HERV-K_22q11.23 provirus ancestral Gag polyprotein-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9435607, end_position_on_the_genomic_accession=9462529, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372270, CurrentID=0, Status='live', Symbol='LOC105372270', Aliases='', description='uncharacterized LOC105372270', other_designations='CTD-3116E22.8', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9709646, end_position_on_the_genomic_accession=9722410, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372271, CurrentID=0, Status='live', Symbol='LOC105372271', Aliases='', description='complement C3-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10037577, end_position_on_the_genomic_accession=10041022, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  121456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121456, CurrentID=0, Status='live', Symbol='SLC9A7P1', Aliases='', description='solute carrier family 9, subfamily A (NHE7, cation proton antiporter 7), member 7 pseudogene 1', other_designations='solute carrier family 9 (sodium/hydrogen exchanger), member 7 pseudogene 1', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98453840, end_position_on_the_genomic_accession=98457145, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  121457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121457, CurrentID=0, Status='live', Symbol='IKBIP', Aliases='IKIP', description='IKBKB interacting protein', other_designations='I kappa B kinase interacting protein|IKBKB-interacting protein|IKK interacting protein|IKK-interacting protein|i kappa-B kinase-interacting protein', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=98613404, end_position_on_the_genomic_accession=98645051, orientation='minus', exon_count=4, OMIM=609861, no_hdr0=''),
  100129394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129394, CurrentID=0, Status='live', Symbol='RPL12P42', Aliases='RPL12_15_1630', description='ribosomal protein L12 pseudogene 42', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13157910, end_position_on_the_genomic_accession=13158874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372275, CurrentID=0, Status='live', Symbol='LOC105372275', Aliases='', description='uncharacterized LOC105372275', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12116359, end_position_on_the_genomic_accession=12147765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129396, CurrentID=0, Status='live', Symbol='FAM106CP', Aliases='FAM106C', description='family with sequence similarity 106 member C, pseudogene', other_designations='family with sequence similarity 106, member C, pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16788743, end_position_on_the_genomic_accession=16790501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129397, CurrentID=0, Status='live', Symbol='LOC100129397', Aliases='', description='uncharacterized LOC100129397', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=46557037, end_position_on_the_genomic_accession=46557453, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129398, CurrentID=0, Status='live', Symbol='LPGAT1P1', Aliases='FAM34A2P, FAM34BP', description='lysophosphatidylglycerol acyltransferase 1 pseudogene 1', other_designations='family with sequence similarity 34, member A2 pseudogene|family with sequence similarity 34, member B pseudogene', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30983459, end_position_on_the_genomic_accession=30984321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372279, CurrentID=0, Status='live', Symbol='LOC105372279', Aliases='', description='uncharacterized LOC105372279', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12633083, end_position_on_the_genomic_accession=12640696, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372280, CurrentID=0, Status='live', Symbol='LOC105372280', Aliases='', description='uncharacterized LOC105372280', other_designations='guanine nucleotide binding protein (G protein), gamma 12-like', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12687444, end_position_on_the_genomic_accession=12688820, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129401, CurrentID=0, Status='live', Symbol='LYPLA1P1', Aliases='', description='lysophospholipase I pseudogene 1', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123230064, end_position_on_the_genomic_accession=123230754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129402, CurrentID=0, Status='live', Symbol='HAUS4P1', Aliases='', description='HAUS augmin like complex subunit 4 pseudogene 1', other_designations='HAUS augmin-like complex, subunit 4 pseudogene 1', map_location='1q24.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=170368918, end_position_on_the_genomic_accession=170370397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129403, CurrentID=0, Status='live', Symbol='EEF1GP4', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 4', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161324837, end_position_on_the_genomic_accession=161326238, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372284, CurrentID=0, Status='live', Symbol='LOC105372284', Aliases='', description='uncharacterized LOC105372284', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=13684209, end_position_on_the_genomic_accession=13724394, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129405, CurrentID=0, Status='live', Symbol='MSTO2P', Aliases='MSTO2', description='misato family member 2, pseudogene', other_designations='misato homolog 2 pseudogene', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155745768, end_position_on_the_genomic_accession=155750688, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100129406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129406, CurrentID=0, Status='live', Symbol='LOC100129406', Aliases='', description='uncharacterized LOC100129406', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=112740447, end_position_on_the_genomic_accession=112740848, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129407, CurrentID=0, Status='live', Symbol='LINC00684', Aliases='DMRTC1-AS1', description='long intergenic non-protein coding RNA 684', other_designations='DMRTC1 antisense RNA 1 (non-protein coding)', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72938163, end_position_on_the_genomic_accession=72938958, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372288, CurrentID=0, Status='live', Symbol='LOC105372288', Aliases='', description='uncharacterized LOC105372288', other_designations='CTC-548K16.2', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14305458, end_position_on_the_genomic_accession=14325431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372289, CurrentID=0, Status='live', Symbol='LOC105372289', Aliases='', description='uncharacterized LOC105372289', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14542588, end_position_on_the_genomic_accession=14562887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372290, CurrentID=0, Status='live', Symbol='LOC105372290', Aliases='', description='uncharacterized LOC105372290', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15827045, end_position_on_the_genomic_accession=15835604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645763, CurrentID=0, Status='live', Symbol='LOC645763', Aliases='', description='intersectin 1 (SH3 domain protein) pseudogene', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10664112, end_position_on_the_genomic_accession=10665818, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372292, CurrentID=0, Status='live', Symbol='LOC105372292', Aliases='', description='uncharacterized LOC105372292', other_designations='CTD-2231E14.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16065692, end_position_on_the_genomic_accession=16066302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372293, CurrentID=0, Status='live', Symbol='LOC105372293', Aliases='', description='uncharacterized LOC105372293', other_designations='CTD-2562J15.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16186280, end_position_on_the_genomic_accession=16190513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372294, CurrentID=0, Status='live', Symbol='LOC105372294', Aliases='', description='uncharacterized LOC105372294', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16382989, end_position_on_the_genomic_accession=16385122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372295, CurrentID=0, Status='live', Symbol='LOC105372295', Aliases='', description='uncharacterized LOC105372295', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16572507, end_position_on_the_genomic_accession=16578448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372296, CurrentID=0, Status='live', Symbol='LOC105372296', Aliases='', description='uncharacterized LOC105372296', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16628412, end_position_on_the_genomic_accession=16629722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372297, CurrentID=0, Status='live', Symbol='LOC105372297', Aliases='', description='uncharacterized LOC105372297', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129418, CurrentID=0, Status='live', Symbol='RPL31P16', Aliases='RPL31_3_207', description='ribosomal protein L31 pseudogene 16', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37194379, end_position_on_the_genomic_accession=37194759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645771, CurrentID=0, Status='live', Symbol='PARP4P2', Aliases='', description='poly(ADP-ribose) polymerase family member 4 pseudogene 2', other_designations='poly (ADP-ribose) polymerase family, member 4 pseudogene 2|poly(ADP-ribose) polymerase family, member 4 pseudogene 2', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=19349131, end_position_on_the_genomic_accession=19408148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372300, CurrentID=0, Status='live', Symbol='LOC105372300', Aliases='', description='uncharacterized LOC105372300', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18317573, end_position_on_the_genomic_accession=18321997, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129421, CurrentID=0, Status='live', Symbol='RPL18P2', Aliases='RPL18_1_137', description='ribosomal protein L18 pseudogene 2', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182328458, end_position_on_the_genomic_accession=182329058, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372302, CurrentID=0, Status='live', Symbol='LOC105372302', Aliases='', description='uncharacterized LOC105372302', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129423, CurrentID=0, Status='live', Symbol='EI24P1', Aliases='', description='etoposide induced 2.4 pseudogene 1', other_designations='etoposide induced 2.4 mRNA pseudogene', map_location='3q26.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176068036, end_position_on_the_genomic_accession=176068900, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129424, CurrentID=0, Status='live', Symbol='RPL19P12', Aliases='RPL19_4_829', description='ribosomal protein L19 pseudogene 12', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103141270, end_position_on_the_genomic_accession=103142403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372305, CurrentID=0, Status='live', Symbol='LOC105372305', Aliases='', description='uncharacterized LOC105372305', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19700475, end_position_on_the_genomic_accession=19708190, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372306, CurrentID=0, Status='live', Symbol='LOC105372306', Aliases='', description='uncharacterized LOC105372306', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19785839, end_position_on_the_genomic_accession=19791747, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129427, CurrentID=0, Status='live', Symbol='DDC-AS1', Aliases='', description='DDC antisense RNA 1', other_designations='', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50531759, end_position_on_the_genomic_accession=50543463, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129428, CurrentID=0, Status='live', Symbol='RANP5', Aliases='', description='RAN, member RAS oncogene family pseudogene 5', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109046783, end_position_on_the_genomic_accession=109047661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372309, CurrentID=0, Status='live', Symbol='LOC105372309', Aliases='', description='uncharacterized LOC105372309', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20042288, end_position_on_the_genomic_accession=20053287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372310, CurrentID=0, Status='live', Symbol='LOC105372310', Aliases='', description='uncharacterized LOC105372310', other_designations='CTC-260E6.6', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20125491, end_position_on_the_genomic_accession=20238810, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372311, CurrentID=0, Status='live', Symbol='LOC105372311', Aliases='', description='uncharacterized LOC105372311', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20214523, end_position_on_the_genomic_accession=20221781, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645784, CurrentID=0, Status='live', Symbol='ANKRD36BP2', Aliases='', description='ankyrin repeat domain 36B pseudogene 2', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88765902, end_position_on_the_genomic_accession=88806596, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105372313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372313, CurrentID=0, Status='live', Symbol='LOC105372313', Aliases='', description='uncharacterized LOC105372313', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20238992, end_position_on_the_genomic_accession=20271893, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121498, CurrentID=0, Status='live', Symbol='LDHAL6CP', Aliases='LDH6C', description='lactate dehydrogenase A-like 6C, pseudogene', other_designations='lactate dehydrogenase A-like 6B pseudogene', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63003428, end_position_on_the_genomic_accession=63004988, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372315, CurrentID=0, Status='live', Symbol='LOC105372315', Aliases='', description='endogenous retrovirus group S71 member 1 Env polyprotein', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20358181, end_position_on_the_genomic_accession=20371817, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372316, CurrentID=0, Status='live', Symbol='LOC105372316', Aliases='', description='uncharacterized LOC105372316', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20473003, end_position_on_the_genomic_accession=20570882, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372317, CurrentID=0, Status='live', Symbol='LOC105372317', Aliases='', description='uncharacterized LOC105372317', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20422251, end_position_on_the_genomic_accession=20454892, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129438, CurrentID=0, Status='live', Symbol='ZFRP1', Aliases='', description='zinc finger RNA binding protein pseudogene 1', other_designations='', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=62884072, end_position_on_the_genomic_accession=62884908, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372319, CurrentID=0, Status='live', Symbol='LOC105372319', Aliases='', description='putative zinc finger protein 56', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  121504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121504, CurrentID=0, Status='live', Symbol='HIST4H4', Aliases='H4/p', description='histone cluster 4, H4', other_designations='histone 4, H4', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14770720, end_position_on_the_genomic_accession=14771131, orientation='minus', exon_count=1, OMIM=615069, no_hdr0=''),
  105372321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372321, CurrentID=0, Status='live', Symbol='LOC105372321', Aliases='', description='uncharacterized LOC105372321', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21444104, end_position_on_the_genomic_accession=21464220, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  121506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121506, CurrentID=0, Status='live', Symbol='ERP27', Aliases='C12orf46, PDIA8', description='endoplasmic reticulum protein 27', other_designations='ER protein 27|endoplasmic reticulum protein 27 kDa|protein disulfide isomerase family A, member 8', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=14914027, end_position_on_the_genomic_accession=14938549, orientation='minus', exon_count=8, OMIM=610642, no_hdr0=''),
  105372323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372323, CurrentID=0, Status='live', Symbol='LOC105372323', Aliases='', description='uncharacterized LOC105372323', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21594741, end_position_on_the_genomic_accession=21629790, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372324, CurrentID=0, Status='live', Symbol='LOC105372324', Aliases='', description='uncharacterized LOC105372324', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21858670, end_position_on_the_genomic_accession=21874496, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372325, CurrentID=0, Status='live', Symbol='LOC105372325', Aliases='', description='uncharacterized LOC105372325', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21922336, end_position_on_the_genomic_accession=21937618, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  645798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645798, CurrentID=0, Status='live', Symbol='RPL3P3', Aliases='RPL3_2_1372', description='ribosomal protein L3 pseudogene 3', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56978348, end_position_on_the_genomic_accession=56979676, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129447, CurrentID=0, Status='live', Symbol='LOC100129447', Aliases='', description='uncharacterized LOC100129447', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=112012653, end_position_on_the_genomic_accession=112013102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  121512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121512, CurrentID=0, Status='live', Symbol='FGD4', Aliases='CMT4H, FRABP, ZFYVE6', description='FYVE, RhoGEF and PH domain containing 4', other_designations='FGD1 family, member 4|FGD1-related F-actin-binding protein|actin-filament binding protein frabin|zinc finger FYVE domain-containing protein 6', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=32399541, end_position_on_the_genomic_accession=32646050, orientation='plus', exon_count=25, OMIM=611104, no_hdr0=''),
  105372329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372329, CurrentID=0, Status='live', Symbol='LOC105372329', Aliases='', description='uncharacterized LOC105372329', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22348772, end_position_on_the_genomic_accession=22352462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372330, CurrentID=0, Status='live', Symbol='LOC105372330', Aliases='', description='uncharacterized LOC105372330', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22328483, end_position_on_the_genomic_accession=22331600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372331, CurrentID=0, Status='live', Symbol='LOC105372331', Aliases='', description='uncharacterized LOC105372331', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22595382, end_position_on_the_genomic_accession=22596049, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100313816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313816, CurrentID=0, Status='live', Symbol='PIRC62', Aliases='', description='piwi-interacting RNA cluster 62', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645805, CurrentID=0, Status='live', Symbol='DAZAP2P1', Aliases='', description='DAZ associated protein 2 pseudogene 1', other_designations='deleted in azoospermia associated protein 2 pseudogene', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202199463, end_position_on_the_genomic_accession=202202021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129454, CurrentID=0, Status='live', Symbol='SNRPEP1', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 1', other_designations='snRNP E protein pseudogene', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160326659, end_position_on_the_genomic_accession=160327125, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645807, CurrentID=0, Status='live', Symbol='RPL5P10', Aliases='RPL5_4_365', description='ribosomal protein L5 pseudogene 10', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40586486, end_position_on_the_genomic_accession=40587484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645808, CurrentID=0, Status='live', Symbol='HSPD1P2', Aliases='HSP60P2, HSPD-9P, HSPDP2', description='heat shock protein family D (Hsp60) member 1 pseudogene 2', other_designations='chaperonin (HSP60) non-functional pseudogene 2|heat shock 60kD protein 1 (chaperonin) pseudogene 2|heat shock 60kDa protein 1 (chaperonin) pseudogene 2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7890511, end_position_on_the_genomic_accession=7892757, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645809, CurrentID=0, Status='live', Symbol='RPL23AP56', Aliases='RPL23A_21_952', description='ribosomal protein L23a pseudogene 56', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134978827, end_position_on_the_genomic_accession=134979282, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372338, CurrentID=0, Status='live', Symbol='LOC105372338', Aliases='', description='uncharacterized LOC105372338', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23549825, end_position_on_the_genomic_accession=23557899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645811, CurrentID=0, Status='live', Symbol='CCDC154', Aliases='C16orf29', description='coiled-coil domain containing 154', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1433425, end_position_on_the_genomic_accession=1444844, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100129460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129460, CurrentID=0, Status='live', Symbol='DPY19L1P1', Aliases='', description='DPY19L1 pseudogene 1', other_designations='dpy-19-like 1 pseudogene 1|dpy-19-like 2 pseudogene', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32580939, end_position_on_the_genomic_accession=32719168, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100129461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129461, CurrentID=0, Status='live', Symbol='LYRM4-AS1', Aliases='', description='LYRM4 antisense RNA 1', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5003816, end_position_on_the_genomic_accession=5240496, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372342, CurrentID=0, Status='live', Symbol='LOC105372342', Aliases='', description='uncharacterized LOC105372342', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9405651, end_position_on_the_genomic_accession=9408256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372343, CurrentID=0, Status='live', Symbol='LOC105372343', Aliases='', description='uncharacterized LOC105372343', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17389069, end_position_on_the_genomic_accession=17394158, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129464, CurrentID=0, Status='live', Symbol='LINC01546', Aliases='CXorf28', description='long intergenic non-protein coding RNA 1546', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3271820, end_position_on_the_genomic_accession=3284653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372345, CurrentID=0, Status='live', Symbol='LOC105372345', Aliases='', description='uncharacterized LOC105372345', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27681259, end_position_on_the_genomic_accession=27685262, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372346, CurrentID=0, Status='live', Symbol='LOC105372346', Aliases='', description='uncharacterized LOC105372346', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27983489, end_position_on_the_genomic_accession=28000401, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372347, CurrentID=0, Status='live', Symbol='LOC105372347', Aliases='', description='uncharacterized LOC105372347', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27886281, end_position_on_the_genomic_accession=27909543, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100129468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129468, CurrentID=0, Status='live', Symbol='BOD1P1', Aliases='', description='biorientation of chromosomes in cell division 1 pseudogene 1', other_designations='', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3415124, end_position_on_the_genomic_accession=3416263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372349, CurrentID=0, Status='live', Symbol='LOC105372349', Aliases='', description='uncharacterized LOC105372349', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28444160, end_position_on_the_genomic_accession=28448976, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372350, CurrentID=0, Status='live', Symbol='LOC105372350', Aliases='', description='uncharacterized LOC105372350', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28869400, end_position_on_the_genomic_accession=28879160, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372351, CurrentID=0, Status='live', Symbol='LOC105372351', Aliases='', description='uncharacterized LOC105372351', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29083096, end_position_on_the_genomic_accession=29093051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  121536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121536, CurrentID=0, Status='live', Symbol='AEBP2', Aliases='', description='AE binding protein 2', other_designations='AE(adipocyte enhancer)-binding protein 2|AE-binding protein 2|adipocyte enhancer-binding protein 2', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19439492, end_position_on_the_genomic_accession=19522239, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105372353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372353, CurrentID=0, Status='live', Symbol='LOC105372353', Aliases='', description='uncharacterized LOC105372353', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=29630264, end_position_on_the_genomic_accession=29665243, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372354, CurrentID=0, Status='live', Symbol='LOC105372354', Aliases='', description='uncharacterized LOC105372354', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=30716100, end_position_on_the_genomic_accession=30719674, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372355, CurrentID=0, Status='live', Symbol='LOC105372355', Aliases='', description='uncharacterized LOC105372355', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31023359, end_position_on_the_genomic_accession=31030274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129476, CurrentID=0, Status='live', Symbol='LOC100129476', Aliases='', description='uncharacterized LOC100129476', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7049847, end_position_on_the_genomic_accession=7051869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372357, CurrentID=0, Status='live', Symbol='LOC105372357', Aliases='', description='uncharacterized LOC105372357', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31117800, end_position_on_the_genomic_accession=31125767, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372358, CurrentID=0, Status='live', Symbol='LOC105372358', Aliases='', description='uncharacterized LOC105372358', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31143416, end_position_on_the_genomic_accession=31147663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372359, CurrentID=0, Status='live', Symbol='LOC105372359', Aliases='', description='uncharacterized LOC105372359', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31437567, end_position_on_the_genomic_accession=31438645, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129480, CurrentID=0, Status='live', Symbol='MKRN2OS', Aliases='C3orf83, MKRN2-AS1', description='MKRN2 opposite strand', other_designations='MKRN2 antisense RNA 1 (non-protein coding)|MKRN2 antisense gene protein 1', map_location='3p25.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12539717, end_position_on_the_genomic_accession=12561056, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372361, CurrentID=0, Status='live', Symbol='LOC105372361', Aliases='', description='uncharacterized LOC105372361', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31713013, end_position_on_the_genomic_accession=31718834, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129482, CurrentID=0, Status='live', Symbol='ZNF37BP', Aliases='KOX21, ZNF37B', description='zinc finger protein 37B, pseudogene', other_designations='zinc finger protein 37b (KOX 21)', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42513510, end_position_on_the_genomic_accession=42552870, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105372363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372363, CurrentID=0, Status='live', Symbol='LOC105372363', Aliases='', description='uncharacterized LOC105372363', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32071512, end_position_on_the_genomic_accession=32072099, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645836, CurrentID=0, Status='live', Symbol='USP17L3', Aliases='USP17B, USP17F', description='ubiquitin specific peptidase 17-like family member 3', other_designations='deubiquitinating enzyme 17-like protein 3|ubiquitin specific peptidase 17-like 3|ubiquitin thioesterase 17-like protein 3|ubiquitin-specific-processing protease 17-like protein 3', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7976393, end_position_on_the_genomic_accession=7977985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  121549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121549, CurrentID=0, Status='live', Symbol='ASCL4', Aliases='HASH4, bHLHa44', description='achaete-scute family bHLH transcription factor 4', other_designations='ASH-4|achaete-scute complex homolog 4|achaete-scute complex-like 4|achaete-scute-like protein 4|class A basic helix-loop-helix protein 44|class II bHLH protein ASCL4', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107774385, end_position_on_the_genomic_accession=107776644, orientation='plus', exon_count=1, OMIM=609155, no_hdr0=''),
  105372366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372366, CurrentID=0, Status='live', Symbol='LOC105372366', Aliases='', description='uncharacterized LOC105372366', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32347485, end_position_on_the_genomic_accession=32349700, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  121551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121551, CurrentID=0, Status='live', Symbol='BTBD11', Aliases='ABTB2B', description='BTB (POZ) domain containing 11', other_designations='ANK-repeat BTB domain containing protein|BTB/POZ domain-containing protein 11', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=107317168, end_position_on_the_genomic_accession=107659642, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  645840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645840, CurrentID=0, Status='live', Symbol='TXNRD3NB', Aliases='TR2IT1, TXNRD3IT1, TXNRD3NT1', description='thioredoxin reductase 3 neighbor', other_designations='TXNRD3 neighbor gene protein|thioredoxin reductase 2 intronic transcript 1|thioredoxin reductase 3 intronic transcript 1|thioredoxin reductase 3 neighbor gene protein|thioredoxin reductase 3 new transcript 1', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126571779, end_position_on_the_genomic_accession=126608555, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645841, CurrentID=0, Status='live', Symbol='TMEM248P1', Aliases='', description='transmembrane protein 248 pseudogene 1', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=125723273, end_position_on_the_genomic_accession=125724217, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  25870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=25870, CurrentID=0, Status='live', Symbol='SUMF2', Aliases='pFGE', description='sulfatase modifying factor 2', other_designations='C-alpha-formyglycine-generating enzyme 2|C-alpha-formylglycine-generating enzyme 2|paralog of the formylglycine-generating enzyme', map_location='7q11.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56064007, end_position_on_the_genomic_accession=56080672, orientation='plus', exon_count=10, OMIM=607940, no_hdr0=''),
  645843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645843, CurrentID=0, Status='live', Symbol='TMEM14EP', Aliases='TMEM14E', description='transmembrane protein 14E, pseudogene', other_designations='transmembrane protein 14E', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152339698, end_position_on_the_genomic_accession=152340990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372372, CurrentID=0, Status='live', Symbol='LOC105372372', Aliases='', description='uncharacterized LOC105372372', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34348049, end_position_on_the_genomic_accession=34359413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372373, CurrentID=0, Status='live', Symbol='LOC105372373', Aliases='', description='zinc finger protein 772', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645846, CurrentID=0, Status='live', Symbol='RPL7AP58', Aliases='RPL7A_20_1131', description='ribosomal protein L7a pseudogene 58', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=29725738, end_position_on_the_genomic_accession=29726621, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372375, CurrentID=0, Status='live', Symbol='LOC105372375', Aliases='', description='uncharacterized LOC105372375', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34772337, end_position_on_the_genomic_accession=34774689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376661, CurrentID=0, Status='live', Symbol='LOC105376661', Aliases='', description='uncharacterized LOC105376661', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46126290, end_position_on_the_genomic_accession=46135646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372377, CurrentID=0, Status='live', Symbol='LOC105372377', Aliases='', description='uncharacterized LOC105372377', other_designations='CTC-523E23.14|CTC-523E23.15', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34898965, end_position_on_the_genomic_accession=34905231, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372378, CurrentID=0, Status='live', Symbol='LOC105372378', Aliases='', description='uncharacterized LOC105372378', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35042534, end_position_on_the_genomic_accession=35055066, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372379, CurrentID=0, Status='live', Symbol='LOC105372379', Aliases='', description='uncharacterized LOC105372379', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35205058, end_position_on_the_genomic_accession=35216008, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372380, CurrentID=0, Status='live', Symbol='LOC105372380', Aliases='', description='uncharacterized LOC105372380', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35259071, end_position_on_the_genomic_accession=35261404, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372381, CurrentID=0, Status='live', Symbol='LOC105372381', Aliases='', description='uncharacterized LOC105372381', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35417940, end_position_on_the_genomic_accession=35419382, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129503, CurrentID=0, Status='live', Symbol='PP13', Aliases='', description='vegetative cell wall protein gp1', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81003341, end_position_on_the_genomic_accession=81005126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372384, CurrentID=0, Status='live', Symbol='LOC105372384', Aliases='', description='uncharacterized LOC105372384', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35959291, end_position_on_the_genomic_accession=35964064, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  645857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645857, CurrentID=0, Status='live', Symbol='RPS10P17', Aliases='RPS10_5_1013', description='ribosomal protein S10 pseudogene 17', other_designations='', map_location='9q33.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=125263536, end_position_on_the_genomic_accession=125264101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372386, CurrentID=0, Status='live', Symbol='LOC105372386', Aliases='', description='acrosin-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129507, CurrentID=0, Status='live', Symbol='MAN1A2P1', Aliases='', description='mannosidase, alpha, class 1A, member 2 pseudogene 1', other_designations='', map_location='19q12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=28790610, end_position_on_the_genomic_accession=28793070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372388, CurrentID=0, Status='live', Symbol='LOC105372388', Aliases='', description='uncharacterized LOC105372388', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36528760, end_position_on_the_genomic_accession=36535689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372389, CurrentID=0, Status='live', Symbol='LOC105372389', Aliases='', description='formin-like protein 10', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645862, CurrentID=0, Status='live', Symbol='CHCHD4P1', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 4 pseudogene 1', other_designations='', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92442077, end_position_on_the_genomic_accession=92442494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372391, CurrentID=0, Status='live', Symbol='LOC105372391', Aliases='', description='uncharacterized LOC105372391', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36667117, end_position_on_the_genomic_accession=36686820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645864, CurrentID=0, Status='live', Symbol='MAGEB17', Aliases='', description='MAGE family member B17', other_designations='melanoma antigen family B, 17 (pseudogene)|melanoma antigen family B17', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=16167481, end_position_on_the_genomic_accession=16172097, orientation='plus', exon_count=3, OMIM=300763, no_hdr0=''),
  105372393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372393, CurrentID=0, Status='live', Symbol='LOC105372393', Aliases='', description='uncharacterized LOC105372393', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37558649, end_position_on_the_genomic_accession=37561008, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372394, CurrentID=0, Status='live', Symbol='LOC105372394', Aliases='', description='uncharacterized LOC105372394', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37807812, end_position_on_the_genomic_accession=37817611, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129515, CurrentID=0, Status='live', Symbol='LINC00633', Aliases='', description='long intergenic non-protein coding RNA 633', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135118955, end_position_on_the_genomic_accession=135120477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129516, CurrentID=0, Status='live', Symbol='LOC100129516', Aliases='', description='uncharacterized LOC100129516', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197506187, end_position_on_the_genomic_accession=197506747, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129517, CurrentID=0, Status='live', Symbol='RPL26P23', Aliases='RPL26_13_865', description='ribosomal protein L26 pseudogene 23', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156633314, end_position_on_the_genomic_accession=156633820, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129518, CurrentID=0, Status='live', Symbol='LOC100129518', Aliases='', description='uncharacterized LOC100129518', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159760258, end_position_on_the_genomic_accession=159762332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372399, CurrentID=0, Status='live', Symbol='LOC105372399', Aliases='', description='uncharacterized LOC105372399', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39547468, end_position_on_the_genomic_accession=39555859, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129520, CurrentID=0, Status='live', Symbol='TEX13C', Aliases='', description='TEX13 family member C', other_designations='testis expressed sequence 13-like', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=125319811, end_position_on_the_genomic_accession=125325214, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372401, CurrentID=0, Status='live', Symbol='LOC105372401', Aliases='', description='uncharacterized LOC105372401', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41307206, end_position_on_the_genomic_accession=41310135, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372402, CurrentID=0, Status='live', Symbol='LOC105372402', Aliases='', description='uncharacterized LOC105372402', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41261289, end_position_on_the_genomic_accession=41264282, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372403, CurrentID=0, Status='live', Symbol='LOC105372403', Aliases='', description='uncharacterized LOC105372403', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41331701, end_position_on_the_genomic_accession=41332121, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372404, CurrentID=0, Status='live', Symbol='LOC105372404', Aliases='', description='uncharacterized LOC105372404', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41600525, end_position_on_the_genomic_accession=41605984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372405, CurrentID=0, Status='live', Symbol='LOC105372405', Aliases='', description='uncharacterized LOC105372405', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41641321, end_position_on_the_genomic_accession=41663250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376662, CurrentID=0, Status='live', Symbol='LOC105376662', Aliases='', description='uncharacterized LOC105376662', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46213653, end_position_on_the_genomic_accession=46220948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645879, CurrentID=0, Status='live', Symbol='FAM90A12P', Aliases='FAM90A12', description='family with sequence similarity 90 member A12, pseudogene', other_designations='family with sequence similarity 90, member A12, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8027074, end_position_on_the_genomic_accession=8032305, orientation='minus', exon_count=0, OMIM=613048, no_hdr0=''),
  100129528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129528, CurrentID=0, Status='live', Symbol='MUC8', Aliases='', description='mucin 8', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132471100, end_position_on_the_genomic_accession=132474630, orientation='minus', exon_count=2, OMIM=601932, no_hdr0=''),
  105372409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372409, CurrentID=0, Status='live', Symbol='LOC105372409', Aliases='', description='uncharacterized LOC105372409', other_designations='CTC-490G23.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43364068, end_position_on_the_genomic_accession=43369094, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372410, CurrentID=0, Status='live', Symbol='LOC105372410', Aliases='', description='uncharacterized LOC105372410', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43636862, end_position_on_the_genomic_accession=43638028, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372411, CurrentID=0, Status='live', Symbol='LOC105372411', Aliases='', description='uncharacterized LOC105372411', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43619579, end_position_on_the_genomic_accession=43634407, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  645884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645884, CurrentID=0, Status='live', Symbol='RPS7P5', Aliases='RPS7_2_183', description='ribosomal protein S7 pseudogene 5', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240007524, end_position_on_the_genomic_accession=240013260, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372413, CurrentID=0, Status='live', Symbol='LOC105372413', Aliases='', description='uncharacterized LOC105372413', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129534, CurrentID=0, Status='live', Symbol='LOC100129534', Aliases='', description='small nuclear ribonucleoprotein polypeptide N pseudogene', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2350414, end_position_on_the_genomic_accession=2352661, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  121599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121599, CurrentID=0, Status='live', Symbol='SPIC', Aliases='SPI-C', description='Spi-C transcription factor (Spi-1/PU.1 related)', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101475323, end_position_on_the_genomic_accession=101487566, orientation='plus', exon_count=6, OMIM=612568, no_hdr0=''),
  105372416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372416, CurrentID=0, Status='live', Symbol='LOC105372416', Aliases='', description='uncharacterized LOC105372416', other_designations='CTB-171A8.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44714604, end_position_on_the_genomic_accession=44739637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121601, CurrentID=0, Status='live', Symbol='ANO4', Aliases='TMEM16D', description='anoctamin 4', other_designations='transmembrane protein 16D', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100717463, end_position_on_the_genomic_accession=101128641, orientation='plus', exon_count=33, OMIM=610111, no_hdr0=''),
  105372418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372418, CurrentID=0, Status='live', Symbol='LOC105372418', Aliases='', description='uncharacterized LOC105372418', other_designations='CTB-129P6.11', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44950049, end_position_on_the_genomic_accession=44950970, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372419, CurrentID=0, Status='live', Symbol='LOC105372419', Aliases='', description='uncharacterized LOC105372419', other_designations='CTB-179K24.3', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45085329, end_position_on_the_genomic_accession=45090275, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129540, CurrentID=0, Status='live', Symbol='LOC100129540', Aliases='', description='uncharacterized LOC100129540', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=79271231, end_position_on_the_genomic_accession=79278268, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372421, CurrentID=0, Status='live', Symbol='LOC105372421', Aliases='', description='uncharacterized LOC105372421', other_designations='CTC-344H19.4', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46023348, end_position_on_the_genomic_accession=46028489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129542, CurrentID=0, Status='live', Symbol='RPS29P21', Aliases='RPS29_16_1562', description='ribosomal protein S29 pseudogene 21', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=59912163, end_position_on_the_genomic_accession=59912454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129543, CurrentID=0, Status='live', Symbol='ZNF730', Aliases='', description='zinc finger protein 730', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23116975, end_position_on_the_genomic_accession=23147212, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372424, CurrentID=0, Status='live', Symbol='LOC105372424', Aliases='', description='uncharacterized LOC105372424', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46160875, end_position_on_the_genomic_accession=46181210, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372425, CurrentID=0, Status='live', Symbol='LOC105372425', Aliases='', description='uncharacterized LOC105372425', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372426, CurrentID=0, Status='live', Symbol='LOC105372426', Aliases='', description='uncharacterized LOC105372426', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46390454, end_position_on_the_genomic_accession=46405338, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645899, CurrentID=0, Status='live', Symbol='RPS28P4', Aliases='RPS28_1_332', description='ribosomal protein S28 pseudogene 4', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231256066, end_position_on_the_genomic_accession=231256457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645900, CurrentID=0, Status='live', Symbol='LAPTM4BP1', Aliases='', description='lysosomal protein transmembrane 4 beta pseudogene 1', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153379814, end_position_on_the_genomic_accession=153380964, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372429, CurrentID=0, Status='live', Symbol='LOC105372429', Aliases='', description='uncharacterized LOC105372429', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48262828, end_position_on_the_genomic_accession=48271720, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372430, CurrentID=0, Status='live', Symbol='LOC105372430', Aliases='', description='uncharacterized LOC105372430', other_designations='CTC-273B12.5', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48465471, end_position_on_the_genomic_accession=48469261, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102095631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102095631, CurrentID=0, Status='live', Symbol='VUR7', Aliases='', description='Vesicoureteral reflux 7', other_designations='', map_location='12p11-q13', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615390, no_hdr0=''),
  100129552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129552, CurrentID=0, Status='live', Symbol='RPS26P38', Aliases='RPS26_15_1076', description='ribosomal protein S26 pseudogene 38', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87642579, end_position_on_the_genomic_accession=87643026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129553, CurrentID=0, Status='live', Symbol='RPL13AP16', Aliases='RPL13A_7_832', description='ribosomal protein L13a pseudogene 16', other_designations='', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105819492, end_position_on_the_genomic_accession=105820127, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372434, CurrentID=0, Status='live', Symbol='LOC105372434', Aliases='', description='uncharacterized LOC105372434', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49748412, end_position_on_the_genomic_accession=49757894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129555, CurrentID=0, Status='live', Symbol='HMGN1P7', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 7', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 7', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=93987946, end_position_on_the_genomic_accession=93989062, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129556, CurrentID=0, Status='live', Symbol='RPS10P27', Aliases='RPS10_12_1588', description='ribosomal protein S10 pseudogene 27', other_designations='', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23672204, end_position_on_the_genomic_accession=23672785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129557, CurrentID=0, Status='live', Symbol='HNRNPA3P8', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 8', other_designations='', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80216261, end_position_on_the_genomic_accession=80217264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372438, CurrentID=0, Status='live', Symbol='LOC105372438', Aliases='', description='uncharacterized LOC105372438', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50511418, end_position_on_the_genomic_accession=50534776, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372439, CurrentID=0, Status='live', Symbol='LOC105372439', Aliases='', description='uncharacterized LOC105372439', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50788407, end_position_on_the_genomic_accession=50793288, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372440, CurrentID=0, Status='live', Symbol='LOC105372440', Aliases='', description='uncharacterized LOC105372440', other_designations='CTD-2568A17.1', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50775645, end_position_on_the_genomic_accession=50786198, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372441, CurrentID=0, Status='live', Symbol='LOC105372441', Aliases='KLK3e', description='uncharacterized LOC105372441', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50830466, end_position_on_the_genomic_accession=50852956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372442, CurrentID=0, Status='live', Symbol='LOC105372442', Aliases='', description='uncharacterized LOC105372442', other_designations='CTB-147C22.9', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50963556, end_position_on_the_genomic_accession=50974726, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101928751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928751, CurrentID=0, Status='live', Symbol='LOC101928751', Aliases='', description='uncharacterized LOC101928751', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=175877343, end_position_on_the_genomic_accession=175880464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645916, CurrentID=0, Status='live', Symbol='RPS11P2', Aliases='RPS11_1_184', description='ribosomal protein S11 pseudogene 2', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240636534, end_position_on_the_genomic_accession=240637073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653786, CurrentID=0, Status='live', Symbol='LOC653786', Aliases='', description='otoancorin pseudogene', other_designations='', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22545698, end_position_on_the_genomic_accession=22576865, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  645918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645918, CurrentID=0, Status='live', Symbol='RPS13P7', Aliases='RPS13_3_1434', description='ribosomal protein S13 pseudogene 7', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57555905, end_position_on_the_genomic_accession=57556408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372447, CurrentID=0, Status='live', Symbol='LOC105372447', Aliases='', description='uncharacterized LOC105372447', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51612439, end_position_on_the_genomic_accession=51615841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372448, CurrentID=0, Status='live', Symbol='LOC105372448', Aliases='', description='uncharacterized LOC105372448', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51779699, end_position_on_the_genomic_accession=51783610, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  544560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544560, CurrentID=0, Status='live', Symbol='AASTH30', Aliases='', description='Allergic/atopic asthma related QTL 30', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645922, CurrentID=0, Status='live', Symbol='S100A7L2', Aliases='S100a7b', description='S100 calcium binding protein A7 like 2', other_designations='S100 calcium binding protein A7-like 2|S100 calcium-binding protein A7-like 2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153436995, end_position_on_the_genomic_accession=153440027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372451, CurrentID=0, Status='live', Symbol='LOC105372451', Aliases='', description='uncharacterized LOC105372451', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52751656, end_position_on_the_genomic_accession=52763185, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129572, CurrentID=0, Status='live', Symbol='LOC100129572', Aliases='', description='uncharacterized LOC100129572', other_designations='', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=148872792, end_position_on_the_genomic_accession=148873403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372453, CurrentID=0, Status='live', Symbol='LOC105372453', Aliases='', description='uncharacterized LOC105372453', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52981524, end_position_on_the_genomic_accession=52985003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372454, CurrentID=0, Status='live', Symbol='LOC105372454', Aliases='', description='uncharacterized LOC105372454', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53132894, end_position_on_the_genomic_accession=53136470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129575, CurrentID=0, Status='live', Symbol='MR1P1', Aliases='MR1B', description='major histocompatibility complex, class I-related pseudogene', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180964516, end_position_on_the_genomic_accession=180969362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372456, CurrentID=0, Status='live', Symbol='LOC105372456', Aliases='', description='uncharacterized LOC105372456', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53746910, end_position_on_the_genomic_accession=53752139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372457, CurrentID=0, Status='live', Symbol='LOC105372457', Aliases='', description='uncharacterized LOC105372457', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53853560, end_position_on_the_genomic_accession=53869721, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  121642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121642, CurrentID=0, Status='live', Symbol='ALKBH2', Aliases='ABH2', description='alkB homolog 2, alpha-ketoglutarate-dependent dioxygenase', other_designations='2OG-Fe(II) oxy DC1|alkB, alkylation repair homolog 2|alkylated DNA repair protein alkB homolog 2|alpha-ketoglutarate-dependent dioxygenase alkB homolog 2|oxy DC1', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109088188, end_position_on_the_genomic_accession=109093631, orientation='minus', exon_count=4, OMIM=610602, no_hdr0=''),
  121643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121643, CurrentID=0, Status='live', Symbol='FOXN4', Aliases='', description='forkhead box N4', other_designations='forkhead/winged helix transcription factor FOXN4', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109277978, end_position_on_the_genomic_accession=109310490, orientation='minus', exon_count=14, OMIM=609429, no_hdr0=''),
  105372738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372738, CurrentID=0, Status='live', Symbol='LOC105372738', Aliases='', description='uncharacterized LOC105372738', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14676699, end_position_on_the_genomic_accession=14678599, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372461, CurrentID=0, Status='live', Symbol='LOC105372461', Aliases='', description='uncharacterized LOC105372461', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54561978, end_position_on_the_genomic_accession=54571720, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372462, CurrentID=0, Status='live', Symbol='LOC105372462', Aliases='', description='uncharacterized LOC105372462', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55409433, end_position_on_the_genomic_accession=55428232, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129583, CurrentID=0, Status='live', Symbol='FAM47E', Aliases='', description='family with sequence similarity 47 member E', other_designations='family with sequence similarity 47, member E', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76214040, end_position_on_the_genomic_accession=76283783, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100129584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129584, CurrentID=0, Status='live', Symbol='RPL17P44', Aliases='RPL17_28_1610', description='ribosomal protein L17 pseudogene 44', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=62415644, end_position_on_the_genomic_accession=62416257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372465, CurrentID=0, Status='live', Symbol='LOC105372465', Aliases='', description='uncharacterized LOC105372465', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55753671, end_position_on_the_genomic_accession=55755370, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372466, CurrentID=0, Status='live', Symbol='LOC105372466', Aliases='', description='uncharacterized LOC105372466', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56148100, end_position_on_the_genomic_accession=56149645, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129587, CurrentID=0, Status='live', Symbol='VN1R34P', Aliases='', description='vomeronasal 1 receptor 34 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63934464, end_position_on_the_genomic_accession=63935331, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645940, CurrentID=0, Status='live', Symbol='HNRNPA1P32', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 32', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30517854, end_position_on_the_genomic_accession=30518861, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372469, CurrentID=0, Status='live', Symbol='LOC105372469', Aliases='', description='uncharacterized LOC105372469', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56311791, end_position_on_the_genomic_accession=56314597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372470, CurrentID=0, Status='live', Symbol='LOC105372470', Aliases='', description='uncharacterized LOC105372470', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56280599, end_position_on_the_genomic_accession=56283805, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129591, CurrentID=0, Status='live', Symbol='RPL29P24', Aliases='RPL29_16_1290', description='ribosomal protein L29 pseudogene 24', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120522539, end_position_on_the_genomic_accession=120523451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372472, CurrentID=0, Status='live', Symbol='LOC105372472', Aliases='', description='uncharacterized LOC105372472', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56689282, end_position_on_the_genomic_accession=56759721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421428, CurrentID=0, Status='live', Symbol='RNGTTP1', Aliases='', description="RNA guanylyltransferase and 5'-phosphatase pseudogene 1", other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25643545, end_position_on_the_genomic_accession=25644219, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129594, CurrentID=0, Status='live', Symbol='FAM8A3P', Aliases='', description='family with sequence similarity 8 member A3, pseudogene', other_designations='family with sequence similarity 8, member A1 pseudogene', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149318393, end_position_on_the_genomic_accession=149324851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129595, CurrentID=0, Status='live', Symbol='PRUNEP1', Aliases='', description='prune exopolyphosphatase pseudogene 1', other_designations='prune homolog pseudogene 1|prune pseudogene', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26115812, end_position_on_the_genomic_accession=26118987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372476, CurrentID=0, Status='live', Symbol='LOC105372476', Aliases='', description='uncharacterized LOC105372476', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57264033, end_position_on_the_genomic_accession=57280483, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129597, CurrentID=0, Status='live', Symbol='LINC00462', Aliases='', description='long intergenic non-protein coding RNA 462', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=48576974, end_position_on_the_genomic_accession=48580901, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645950, CurrentID=0, Status='live', Symbol='HNRNPA1P1', Aliases='HNRPA1P1, b34I8.2, hnRNPA1-Lp', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27523447, end_position_on_the_genomic_accession=27524537, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129599, CurrentID=0, Status='live', Symbol='RPS29P14', Aliases='RPS29_3_802', description='ribosomal protein S29 pseudogene 14', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=33036212, end_position_on_the_genomic_accession=33036382, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372480, CurrentID=0, Status='live', Symbol='LOC105372480', Aliases='', description='uncharacterized LOC105372480', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58314780, end_position_on_the_genomic_accession=58327277, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  121665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121665, CurrentID=0, Status='live', Symbol='SPPL3', Aliases='IMP2, MDHV1887, PRO4332, PSH1, PSL4', description='signal peptide peptidase like 3', other_designations='SPP-like 3|intramembrane protease 2|presenilin homologous protein 1|presenilin-like protein 4', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=120762510, end_position_on_the_genomic_accession=120904352, orientation='minus', exon_count=13, OMIM=608240, no_hdr0=''),
  645954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645954, CurrentID=0, Status='live', Symbol='SVILP1', Aliases='', description='supervillin pseudogene 1', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30692274, end_position_on_the_genomic_accession=30717013, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100129603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129603, CurrentID=0, Status='live', Symbol='LOC100129603', Aliases='', description='uncharacterized LOC100129603', other_designations='', map_location='7p22.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3140917, end_position_on_the_genomic_accession=3158573, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129604, CurrentID=0, Status='live', Symbol='PABPC1P12', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 12', other_designations='', map_location='10p11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=38635995, end_position_on_the_genomic_accession=38638082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645957, CurrentID=0, Status='live', Symbol='RPS3P7', Aliases='RPS3_3_1501', description='ribosomal protein S3 pseudogene 7', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78825252, end_position_on_the_genomic_accession=78826085, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645958, CurrentID=0, Status='live', Symbol='RPS18P9', Aliases='RPS18_6_767', description='ribosomal protein S18 pseudogene 9', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149593616, end_position_on_the_genomic_accession=149594602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372487, CurrentID=0, Status='live', Symbol='LOC105372487', Aliases='', description='transformer-2 protein homolog alpha-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=31349235, end_position_on_the_genomic_accession=31383997, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372488, CurrentID=0, Status='live', Symbol='LOC105372488', Aliases='', description='uncharacterized LOC105372488', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34675954, end_position_on_the_genomic_accession=34678661, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645961, CurrentID=0, Status='live', Symbol='SPATA31C2', Aliases='FAM75C2', description='SPATA31 subfamily C, member 2', other_designations='family with sequence similarity 75, member C2', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88129305, end_position_on_the_genomic_accession=88134985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  645962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645962, CurrentID=0, Status='live', Symbol='RPL23AP20', Aliases='RPL23A_6_188', description='ribosomal protein L23a pseudogene 20', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241916064, end_position_on_the_genomic_accession=241916587, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421431, CurrentID=0, Status='live', Symbol='HNRNPA3P11', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 11', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=57247088, end_position_on_the_genomic_accession=57247976, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372492, CurrentID=0, Status='live', Symbol='LOC105372492', Aliases='', description='uncharacterized LOC105372492', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=913548, end_position_on_the_genomic_accession=938162, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372493, CurrentID=0, Status='live', Symbol='LOC105372493', Aliases='', description='uncharacterized LOC105372493', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1027720, end_position_on_the_genomic_accession=1118497, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372494, CurrentID=0, Status='live', Symbol='LOC105372494', Aliases='', description='uncharacterized LOC105372494', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1016847, end_position_on_the_genomic_accession=1025695, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  645967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645967, CurrentID=0, Status='live', Symbol='LOC645967', Aliases='', description='uncharacterized LOC645967', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149717607, end_position_on_the_genomic_accession=149718896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  645968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645968, CurrentID=0, Status='live', Symbol='RPS3AP25', Aliases='RPS3A_9_820', description='ribosomal protein S3a pseudogene 25', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94694981, end_position_on_the_genomic_accession=94695839, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421432, CurrentID=0, Status='live', Symbol='HNRNPA1P44', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 44', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65815862, end_position_on_the_genomic_accession=65817558, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129618, CurrentID=0, Status='live', Symbol='THAP5P2', Aliases='', description='THAP domain containing 5 pseudogene 2', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16275059, end_position_on_the_genomic_accession=16276512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372499, CurrentID=0, Status='live', Symbol='LOC105372499', Aliases='', description='uncharacterized LOC105372499', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=1512515, end_position_on_the_genomic_accession=1540023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129620, CurrentID=0, Status='live', Symbol='LOC100129620', Aliases='', description='uncharacterized LOC100129620', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99004276, end_position_on_the_genomic_accession=99148852, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  645973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645973, CurrentID=0, Status='live', Symbol='ATP5F1P2', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 2', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94738654, end_position_on_the_genomic_accession=94739461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645974, CurrentID=0, Status='live', Symbol='PABPC1L2B', Aliases='RBM32B', description='poly(A) binding protein, cytoplasmic 1-like 2B', other_designations='RNA binding motif protein 32B|RNA-binding motif protein 32|RNA-binding protein 32', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73003513, end_position_on_the_genomic_accession=73005712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421433, CurrentID=0, Status='live', Symbol='HNRNPA1P71', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 71', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=33012130, end_position_on_the_genomic_accession=33013065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372504, CurrentID=0, Status='live', Symbol='LOC105372504', Aliases='', description='uncharacterized LOC105372504', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2534528, end_position_on_the_genomic_accession=2543543, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372505, CurrentID=0, Status='live', Symbol='LOC105372505', Aliases='', description='uncharacterized LOC105372505', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2613359, end_position_on_the_genomic_accession=2644140, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372506, CurrentID=0, Status='live', Symbol='LOC105372506', Aliases='', description='uncharacterized LOC105372506', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2703712, end_position_on_the_genomic_accession=2709578, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129627, CurrentID=0, Status='live', Symbol='PPP1R1AP2', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 1A pseudogene 2', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 1A pseudogene 2', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35912020, end_position_on_the_genomic_accession=35912522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645980, CurrentID=0, Status='live', Symbol='CFL1P4', Aliases='CFLL4, CFLP4', description='cofilin 1 (non-muscle) pseudogene 4', other_designations='cofilin pseudogene 4|cofilin-like 4', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241993206, end_position_on_the_genomic_accession=241993883, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  544570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544570, CurrentID=0, Status='live', Symbol='AASTH25', Aliases='', description='Allergic/atopic asthma related QTL 25', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372510, CurrentID=0, Status='live', Symbol='LOC105372510', Aliases='', description='uncharacterized LOC105372510', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4564341, end_position_on_the_genomic_accession=4572084, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129631, CurrentID=0, Status='live', Symbol='IGHV2OR16-5', Aliases='IGHV2OR165', description='immunoglobulin heavy variable 2/OR16-5 (non-functional)', other_designations='IGHV2/OR16-5', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=32859034, end_position_on_the_genomic_accession=32859575, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422935, CurrentID=0, Status='live', Symbol='MIR3118-4', Aliases='', description='microRNA 3118-4', other_designations='hsa-mir-3118-4', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21843750, end_position_on_the_genomic_accession=21843824, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372513, CurrentID=0, Status='live', Symbol='LOC105372513', Aliases='', description='uncharacterized LOC105372513', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5480786, end_position_on_the_genomic_accession=5489424, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129634, CurrentID=0, Status='live', Symbol='RPL9P19', Aliases='RPL9_4_783', description='ribosomal protein L9 pseudogene 19', other_designations='', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8938771, end_position_on_the_genomic_accession=8939454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372515, CurrentID=0, Status='live', Symbol='LOC105372515', Aliases='', description='uncharacterized LOC105372515', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5831328, end_position_on_the_genomic_accession=5833217, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129636, CurrentID=0, Status='live', Symbol='LOC100129636', Aliases='', description='uncharacterized LOC100129636', other_designations='XXbac-BPG258E24.10', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29036021, end_position_on_the_genomic_accession=29076740, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372517, CurrentID=0, Status='live', Symbol='LOC105372517', Aliases='', description='uncharacterized LOC105372517', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=6765590, end_position_on_the_genomic_accession=6767484, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372518, CurrentID=0, Status='live', Symbol='LOC105372518', Aliases='', description='uncharacterized LOC105372518', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7698353, end_position_on_the_genomic_accession=7724996, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372519, CurrentID=0, Status='live', Symbol='LOC105372519', Aliases='', description='uncharacterized LOC105372519', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=7848222, end_position_on_the_genomic_accession=7879444, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129640, CurrentID=0, Status='live', Symbol='DDX55P1', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 55 pseudogene 1', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11512710, end_position_on_the_genomic_accession=11514138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  645993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645993, CurrentID=0, Status='live', Symbol='RPL31P57', Aliases='RPL31_21_1570', description='ribosomal protein L31 pseudogene 57', other_designations='hCG1793136', map_location='17q23.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64308276, end_position_on_the_genomic_accession=64308716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372522, CurrentID=0, Status='live', Symbol='LOC105372522', Aliases='', description='uncharacterized LOC105372522', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=8962199, end_position_on_the_genomic_accession=8968281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129643, CurrentID=0, Status='live', Symbol='POM121L11P', Aliases='', description='POM121 transmembrane nucleoporin like 11, pseudogene', other_designations='POM121 membrane glycoprotein-like 11, pseudogene|POM121 transmembrane nucleoporin-like 11, pseudogene', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23309850, end_position_on_the_genomic_accession=23321987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645996, CurrentID=0, Status='live', Symbol='NAP1L6', Aliases='', description='nucleosome assembly protein 1-like 6', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73126037, end_position_on_the_genomic_accession=73128080, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129645, CurrentID=0, Status='live', Symbol='RPL18P9', Aliases='RPL18_3_1196', description='ribosomal protein L18 pseudogene 9', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=4109496, end_position_on_the_genomic_accession=4119955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  645998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645998, CurrentID=0, Status='live', Symbol='RPL10AP5', Aliases='RPL10A_2_189', description='ribosomal protein L10a pseudogene 5', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242365160, end_position_on_the_genomic_accession=242365848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129647, CurrentID=0, Status='live', Symbol='RPL5P35', Aliases='RPL5_16_1760', description='ribosomal protein L5 pseudogene 35', other_designations='', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49757559, end_position_on_the_genomic_accession=49759053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646000, CurrentID=0, Status='live', Symbol='SLC35G4', Aliases='AMAC1L1P, SLC35G4', description='solute carrier family 35 member G4', other_designations='acyl-malonyl-condensing enzyme 1-like protein 1|protein AMAC1L1|putative solute carrier family 35 member G4 pseudogene|solute carrier family 35, member G4|solute carrier family 35, member G4, pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11609558, end_position_on_the_genomic_accession=11610612, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129649, CurrentID=0, Status='live', Symbol='PGDP2', Aliases='', description='phosphogluconate dehydrogenase pseudogene 2', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75599999, end_position_on_the_genomic_accession=75601719, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129650, CurrentID=0, Status='live', Symbol='RPS26P55', Aliases='RPS26_30_1651', description='ribosomal protein S26 pseudogene 55', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34533354, end_position_on_the_genomic_accession=34534086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372531, CurrentID=0, Status='live', Symbol='LOC105372531', Aliases='', description='uncharacterized LOC105372531', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11952879, end_position_on_the_genomic_accession=11958943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372532, CurrentID=0, Status='live', Symbol='LOC105372532', Aliases='', description='uncharacterized LOC105372532', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12651076, end_position_on_the_genomic_accession=12673203, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372533, CurrentID=0, Status='live', Symbol='LOC105372533', Aliases='', description='uncharacterized LOC105372533', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12649845, end_position_on_the_genomic_accession=12692468, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129654, CurrentID=0, Status='live', Symbol='TCF24', Aliases='TCF-24, bHLHa25', description='transcription factor 24', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66946501, end_position_on_the_genomic_accession=66962590, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372535, CurrentID=0, Status='live', Symbol='LOC105372535', Aliases='', description='uncharacterized LOC105372535', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=12935948, end_position_on_the_genomic_accession=12940821, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372536, CurrentID=0, Status='live', Symbol='LOC105372536', Aliases='', description='uncharacterized LOC105372536', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13104760, end_position_on_the_genomic_accession=13125077, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129657, CurrentID=0, Status='live', Symbol='RPL17P18', Aliases='RPL17_5_454', description='ribosomal protein L17 pseudogene 18', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197850365, end_position_on_the_genomic_accession=197850990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372538, CurrentID=0, Status='live', Symbol='LOC105372538', Aliases='', description='uncharacterized LOC105372538', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13261420, end_position_on_the_genomic_accession=13325070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100421439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421439, CurrentID=0, Status='live', Symbol='TPT1P3', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141709010, end_position_on_the_genomic_accession=141709727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372540, CurrentID=0, Status='live', Symbol='LOC105372540', Aliases='', description='uncharacterized LOC105372540', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=13787359, end_position_on_the_genomic_accession=13791782, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646013, CurrentID=0, Status='live', Symbol='RPS17P3', Aliases='RPS17B, RPS17L2, RPS17_7_1537', description='ribosomal protein S17 pseudogene 3', other_designations='ribosomal protein S17-like 2|ribosomal protein S17b-like 2 (pseudogene)|ribosomal protein pseudogene S17b', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30830877, end_position_on_the_genomic_accession=30831360, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129662, CurrentID=0, Status='live', Symbol='FGF13-AS1', Aliases='', description='FGF13 antisense RNA 1', other_designations='FGF13 antisense RNA 1 (non-protein coding)', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138712107, end_position_on_the_genomic_accession=138716605, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  121727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121727, CurrentID=0, Status='live', Symbol='PEX12P1', Aliases='', description='peroxisomal biogenesis factor 12 pseudogene 1', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89991543, end_position_on_the_genomic_accession=89992895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129664, CurrentID=0, Status='live', Symbol='PRELID3BP1', Aliases='SLMO2P1', description='PRELI domain containing 3B pseudogene 1', other_designations='SLMO2 pseudogene 1|slowmo homolog 2 pseudogene 1', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220467859, end_position_on_the_genomic_accession=220468462, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372545, CurrentID=0, Status='live', Symbol='LOC105372545', Aliases='', description='uncharacterized LOC105372545', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17216771, end_position_on_the_genomic_accession=17218180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372546, CurrentID=0, Status='live', Symbol='LOC105372546', Aliases='', description='uncharacterized LOC105372546', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17264082, end_position_on_the_genomic_accession=17327491, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646019, CurrentID=0, Status='live', Symbol='CBY3', Aliases='', description='chibby homolog 3 (Drosophila)', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179678558, end_position_on_the_genomic_accession=179680974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372548, CurrentID=0, Status='live', Symbol='LOC105372548', Aliases='', description='uncharacterized LOC105372548', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=17872636, end_position_on_the_genomic_accession=17883258, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129669, CurrentID=0, Status='live', Symbol='IZUMO3', Aliases='C9orf134, bA20A20.1', description='IZUMO family member 3', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=24543214, end_position_on_the_genomic_accession=24545930, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100129670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129670, CurrentID=0, Status='live', Symbol='RPL18P8', Aliases='RPL18_5_1138', description='ribosomal protein L18 pseudogene 8', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=39161440, end_position_on_the_genomic_accession=39161973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646023, CurrentID=0, Status='live', Symbol='ADORA2A-AS1', Aliases='C22orf45', description='ADORA2A antisense RNA 1', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24429210, end_position_on_the_genomic_accession=24494815, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  646024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646024, CurrentID=0, Status='live', Symbol='RAET1K', Aliases='', description='retinoic acid early transcript 1K pseudogene', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149998019, end_position_on_the_genomic_accession=150005144, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372553, CurrentID=0, Status='live', Symbol='LOC105372553', Aliases='', description='uncharacterized LOC105372553', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20022507, end_position_on_the_genomic_accession=20025611, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372554, CurrentID=0, Status='live', Symbol='LOC105372554', Aliases='', description='uncharacterized LOC105372554', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20232046, end_position_on_the_genomic_accession=20232873, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372555, CurrentID=0, Status='live', Symbol='LOC105372555', Aliases='', description='uncharacterized LOC105372555', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20729260, end_position_on_the_genomic_accession=20749637, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372556, CurrentID=0, Status='live', Symbol='LOC105372556', Aliases='', description='uncharacterized LOC105372556', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21087616, end_position_on_the_genomic_accession=21106341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  646029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646029, CurrentID=0, Status='live', Symbol='LOC646029', Aliases='', description='uncharacterized LOC646029', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=77571089, end_position_on_the_genomic_accession=77575662, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401225, CurrentID=0, Status='live', Symbol='FLJ31183', Aliases='', description='uncharacterized LOC401225', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372559, CurrentID=0, Status='live', Symbol='LOC105372559', Aliases='', description='uncharacterized LOC105372559', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372560, CurrentID=0, Status='live', Symbol='LOC105372560', Aliases='', description='uncharacterized LOC105372560', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22217054, end_position_on_the_genomic_accession=22223265, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372561, CurrentID=0, Status='live', Symbol='LOC105372561', Aliases='', description='uncharacterized LOC105372561', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22260495, end_position_on_the_genomic_accession=22275074, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372562, CurrentID=0, Status='live', Symbol='LOC105372562', Aliases='', description='uncharacterized LOC105372562', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22449441, end_position_on_the_genomic_accession=22467245, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372563, CurrentID=0, Status='live', Symbol='LOC105372563', Aliases='', description='uncharacterized LOC105372563', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22468027, end_position_on_the_genomic_accession=22471553, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372564, CurrentID=0, Status='live', Symbol='LOC105372564', Aliases='', description='uncharacterized LOC105372564', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22510143, end_position_on_the_genomic_accession=22524166, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129685, CurrentID=0, Status='live', Symbol='RPL21P125', Aliases='RPL21_55_1506', description='ribosomal protein L21 pseudogene 125', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3665177, end_position_on_the_genomic_accession=3665740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372566, CurrentID=0, Status='live', Symbol='LOC105372566', Aliases='', description='uncharacterized LOC105372566', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22667515, end_position_on_the_genomic_accession=22686694, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372567, CurrentID=0, Status='live', Symbol='LOC105372567', Aliases='', description='uncharacterized LOC105372567', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22772813, end_position_on_the_genomic_accession=22774034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130514, CurrentID=0, Status='live', Symbol='PGAM1P1', Aliases='', description='phosphoglycerate mutase 1 pseudogene 1', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58161198, end_position_on_the_genomic_accession=58161751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372569, CurrentID=0, Status='live', Symbol='LOC105372569', Aliases='', description='uncharacterized LOC105372569', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23030889, end_position_on_the_genomic_accession=23039219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129690, CurrentID=0, Status='live', Symbol='RAD1P2', Aliases='', description='RAD1 pseudogene 2', other_designations='RAD1 homolog pseudogene 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159080940, end_position_on_the_genomic_accession=159081693, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372571, CurrentID=0, Status='live', Symbol='LOC105372571', Aliases='', description='uncharacterized LOC105372571', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23321000, end_position_on_the_genomic_accession=23335863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129692, CurrentID=0, Status='live', Symbol='EIF4BP9', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 9', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66074344, end_position_on_the_genomic_accession=66076363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372573, CurrentID=0, Status='live', Symbol='LOC105372573', Aliases='', description='uncharacterized LOC105372573', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23667149, end_position_on_the_genomic_accession=23674842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129694, CurrentID=0, Status='live', Symbol='R3HDM2P2', Aliases='', description='R3H domain containing 2 pseudogene 2', other_designations='', map_location='6q16.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104017630, end_position_on_the_genomic_accession=104019781, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372575, CurrentID=0, Status='live', Symbol='LOC105372575', Aliases='', description='uncharacterized LOC105372575', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23810659, end_position_on_the_genomic_accession=23821189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129696, CurrentID=0, Status='live', Symbol='TFP1', Aliases='TFP', description='transferrin pseudogene 1', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133691722, end_position_on_the_genomic_accession=133709674, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646049, CurrentID=0, Status='live', Symbol='HMGN2P20', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 20', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 20', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24330307, end_position_on_the_genomic_accession=24332356, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372578, CurrentID=0, Status='live', Symbol='LOC105372578', Aliases='', description='uncharacterized LOC105372578', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=24929217, end_position_on_the_genomic_accession=24932985, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372579, CurrentID=0, Status='live', Symbol='LOC105372579', Aliases='', description='uncharacterized LOC105372579', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25229228, end_position_on_the_genomic_accession=25237008, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646052, CurrentID=0, Status='live', Symbol='SLC20A1P3', Aliases='', description='solute carrier family 20 (phosphate transporter), member 1 pseudogene 3', other_designations='', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=19966630, end_position_on_the_genomic_accession=19968365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372581, CurrentID=0, Status='live', Symbol='LOC105372581', Aliases='', description='uncharacterized LOC105372581', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25394253, end_position_on_the_genomic_accession=25408643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372582, CurrentID=0, Status='live', Symbol='LOC105372582', Aliases='', description='uncharacterized LOC105372582', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25751191, end_position_on_the_genomic_accession=25768644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  252839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252839, CurrentID=0, Status='live', Symbol='TMEM9', Aliases='TMEM9A', description='transmembrane protein 9', other_designations='dermal papilla-derived protein 4', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201134771, end_position_on_the_genomic_accession=201171582, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  252840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252840, CurrentID=0, Status='live', Symbol='SRP68P2', Aliases='SRPSMCR', description='signal recognition particle 68kDa pseudogene 2', other_designations='signal recognition particle 68kD pseudogene', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18527559, end_position_on_the_genomic_accession=18528462, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  252841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252841, CurrentID=0, Status='live', Symbol='SRP68P1', Aliases='SRPSMCR', description='signal recognition particle 68kDa pseudogene 1', other_designations='signal recognition particle 68kD pseudogene|signal recognition particle, Smith Magenis syndrome chromosome region', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16787134, end_position_on_the_genomic_accession=16788239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  252842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252842, CurrentID=0, Status='live', Symbol='AAT1', Aliases='FAA1', description='Aortic aneurysm, familial thoracic 1', other_designations='', map_location='11q23.3-q24', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607086, no_hdr0=''),
  252843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252843, CurrentID=0, Status='live', Symbol='AAT2', Aliases='FAA2', description='Aortic aneurysm, familial thoracic 2', other_designations='', map_location='5q13-q14', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607087, no_hdr0=''),
  100129708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129708, CurrentID=0, Status='live', Symbol='IPO7P1', Aliases='', description='importin 7 pseudogene 1', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51921864, end_position_on_the_genomic_accession=51922384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372589, CurrentID=0, Status='live', Symbol='LOC105372589', Aliases='', description='uncharacterized LOC105372589', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31677434, end_position_on_the_genomic_accession=31716844, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372590, CurrentID=0, Status='live', Symbol='LOC105372590', Aliases='', description='uncharacterized LOC105372590', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32362609, end_position_on_the_genomic_accession=32393526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121775, CurrentID=0, Status='live', Symbol='CHDM', Aliases='', description='Chordoma', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=215400, no_hdr0=''),
  105372592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372592, CurrentID=0, Status='live', Symbol='LOC105372592', Aliases='', description='uncharacterized LOC105372592', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32536593, end_position_on_the_genomic_accession=32541950, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646065, CurrentID=0, Status='live', Symbol='PMM2P2', Aliases='', description='phosphomannomutase 2 pseudogene 2', other_designations='', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12192203, end_position_on_the_genomic_accession=12192934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129714, CurrentID=0, Status='live', Symbol='RCC2P8', Aliases='', description='regulator of chromosome condensation 2 pseudogene 8', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108788748, end_position_on_the_genomic_accession=108789784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372595, CurrentID=0, Status='live', Symbol='LOC105372595', Aliases='', description='uncharacterized LOC105372595', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33590178, end_position_on_the_genomic_accession=33596981, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129716, CurrentID=0, Status='live', Symbol='ARRDC3-AS1', Aliases='', description='ARRDC3 antisense RNA 1', other_designations='ARRDC3 antisense RNA 1 (non-protein coding)', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91380347, end_position_on_the_genomic_accession=91420715, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372597, CurrentID=0, Status='live', Symbol='LOC105372597', Aliases='', description='uncharacterized LOC105372597', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34002857, end_position_on_the_genomic_accession=34003619, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129718, CurrentID=0, Status='live', Symbol='RPS13P3', Aliases='RPS13_3_211', description='ribosomal protein S13 pseudogene 3', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42469790, end_position_on_the_genomic_accession=42470304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372599, CurrentID=0, Status='live', Symbol='LOC105372599', Aliases='', description='uncharacterized LOC105372599', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34703647, end_position_on_the_genomic_accession=34704182, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372600, CurrentID=0, Status='live', Symbol='LOC105372600', Aliases='', description='uncharacterized LOC105372600', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372601, CurrentID=0, Status='live', Symbol='LOC105372601', Aliases='', description='uncharacterized LOC105372601', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34991407, end_position_on_the_genomic_accession=34993743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646074, CurrentID=0, Status='live', Symbol='POM121L10P', Aliases='', description='POM121 transmembrane nucleoporin like 10, pseudogene', other_designations='POM121 membrane glycoprotein-like 1 pseudogene|POM121 membrane glycoprotein-like 10, pseudogene|POM121 transmembrane nucleoporin-like 10, pseudogene|POM121-like 1', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24645166, end_position_on_the_genomic_accession=24659147, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100129723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129723, CurrentID=0, Status='live', Symbol='HMGB1P12', Aliases='HMGB1L12', description='high mobility group box 1 pseudogene 12', other_designations='high-mobility group box 1 pseudogene 12|high-mobility group box 1-like 12|high-mobility group box 2 pseudogene', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111434427, end_position_on_the_genomic_accession=111435015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372604, CurrentID=0, Status='live', Symbol='LOC105372604', Aliases='', description='uncharacterized LOC105372604', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36147281, end_position_on_the_genomic_accession=36149901, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646077, CurrentID=0, Status='live', Symbol='RPL29P23', Aliases='RPL29_11_1134', description='ribosomal protein L29 pseudogene 23', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33189915, end_position_on_the_genomic_accession=33190554, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129726, CurrentID=0, Status='live', Symbol='LINC01126', Aliases='', description='long intergenic non-protein coding RNA 1126', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43227211, end_position_on_the_genomic_accession=43228855, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129727, CurrentID=0, Status='live', Symbol='TMED10P2', Aliases='', description='transmembrane emp24-like trafficking protein 10 (yeast) pseudogene 2', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128538017, end_position_on_the_genomic_accession=128538631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646080, CurrentID=0, Status='live', Symbol='RAB11FIP1P1', Aliases='', description='RAB11 family interacting protein 1 (class I) pseudogene 1', other_designations='Rab coupling protein pseudogene', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74202786, end_position_on_the_genomic_accession=74204656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  121793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121793, CurrentID=0, Status='live', Symbol='TEX29', Aliases='C13orf16, bA474D23.1', description='testis expressed 29', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=111316185, end_position_on_the_genomic_accession=111344247, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372739, CurrentID=0, Status='live', Symbol='LOC105372739', Aliases='', description='uncharacterized LOC105372739', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=15186637, end_position_on_the_genomic_accession=15189854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646083, CurrentID=0, Status='live', Symbol='PA2G4P6', Aliases='PA2G4L3', description='proliferation-associated 2G4 pseudogene 6', other_designations='proliferation-associated 2G4, 38kDa pseudogene', map_location='9q22.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89449437, end_position_on_the_genomic_accession=89451764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129732, CurrentID=0, Status='live', Symbol='RPS6P15', Aliases='RPS6_7_1097', description='ribosomal protein S6 pseudogene 15', other_designations='', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=111498176, end_position_on_the_genomic_accession=111498941, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646085, CurrentID=0, Status='live', Symbol='PSMC1P7', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 7', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161328705, end_position_on_the_genomic_accession=161330262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646086, CurrentID=0, Status='live', Symbol='WDR45BP1', Aliases='WDR45LP1', description='WD repeat domain 45B pseudogene 1', other_designations='WDR45-like pseudogene 1', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32110190, end_position_on_the_genomic_accession=32112695, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646087, CurrentID=0, Status='live', Symbol='IMP3P1', Aliases='', description='IMP3 pseudogene 1', other_designations='IMP3, U3 small nucleolar ribonucleoprotein, homolog pseudogene', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=10829758, end_position_on_the_genomic_accession=10830308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129736, CurrentID=0, Status='live', Symbol='CDV3P1', Aliases='', description='CDV3 pseudogene 1', other_designations='CDV3 homolog pseudogene 1', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96776774, end_position_on_the_genomic_accession=96777887, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372617, CurrentID=0, Status='live', Symbol='LOC105372617', Aliases='', description='uncharacterized LOC105372617', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40298752, end_position_on_the_genomic_accession=40301661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646090, CurrentID=0, Status='live', Symbol='RHPN2P1', Aliases='', description='rhophilin, Rho GTPase binding protein 2 pseudogene 1', other_designations='Putative rhophilin-2-like protein', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20228617, end_position_on_the_genomic_accession=20277094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129739, CurrentID=0, Status='live', Symbol='DYNLL1P4', Aliases='', description='dynein, light chain, LC8-type 1 pseudogene 4', other_designations='', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113789542, end_position_on_the_genomic_accession=113789807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372620, CurrentID=0, Status='live', Symbol='LOC105372620', Aliases='', description='uncharacterized LOC105372620', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=40677232, end_position_on_the_genomic_accession=40678675, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646093, CurrentID=0, Status='live', Symbol='RPS8P7', Aliases='RPS8_3_645', description='ribosomal protein S8 pseudogene 7', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180226333, end_position_on_the_genomic_accession=180227000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129742, CurrentID=0, Status='live', Symbol='RPS3AP13', Aliases='', description='ribosomal protein S3A pseudogene', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159689671, end_position_on_the_genomic_accession=159690344, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129743, CurrentID=0, Status='live', Symbol='RPL7P14', Aliases='RPL7_5_313', description='ribosomal protein L7 pseudogene 14', other_designations='', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203040105, end_position_on_the_genomic_accession=203040719, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646096, CurrentID=0, Status='live', Symbol='CHEK2P2', Aliases='', description='checkpoint kinase 2 pseudogene 2', other_designations='CHK2 checkpoint homolog pseudogene', map_location='15q11.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20282744, end_position_on_the_genomic_accession=20291558, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372625, CurrentID=0, Status='live', Symbol='LOC105372625', Aliases='', description='uncharacterized LOC105372625', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43820925, end_position_on_the_genomic_accession=43824807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372626, CurrentID=0, Status='live', Symbol='LOC105372626', Aliases='', description='uncharacterized LOC105372626', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43894696, end_position_on_the_genomic_accession=43895292, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372627, CurrentID=0, Status='live', Symbol='LOC105372627', Aliases='', description='uncharacterized LOC105372627', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  252884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252884, CurrentID=0, Status='live', Symbol='ZNF396', Aliases='ZSCAN14', description='zinc finger protein 396', other_designations='zinc finger and SCAN domain-containing protein 14', map_location='18q12', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35366697, end_position_on_the_genomic_accession=35377337, orientation='minus', exon_count=8, OMIM=609600, no_hdr0=''),
  105372629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372629, CurrentID=0, Status='live', Symbol='LOC105372629', Aliases='', description='uncharacterized LOC105372629', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44347767, end_position_on_the_genomic_accession=44355780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372630, CurrentID=0, Status='live', Symbol='LOC105372630', Aliases='', description='uncharacterized LOC105372630', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45230841, end_position_on_the_genomic_accession=45290352, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105372631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372631, CurrentID=0, Status='live', Symbol='LOC105372631', Aliases='', description='uncharacterized LOC105372631', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45427318, end_position_on_the_genomic_accession=45446730, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646104, CurrentID=0, Status='live', Symbol='RPL32P16', Aliases='RPL32_7_768', description='ribosomal protein L32 pseudogene 16', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151096159, end_position_on_the_genomic_accession=151099933, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372633, CurrentID=0, Status='live', Symbol='LOC105372633', Aliases='', description='uncharacterized LOC105372633', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46406768, end_position_on_the_genomic_accession=46444971, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372634, CurrentID=0, Status='live', Symbol='LOC105372634', Aliases='', description='uncharacterized LOC105372634', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46387700, end_position_on_the_genomic_accession=46393520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421455, CurrentID=0, Status='live', Symbol='EIF4EP4', Aliases='', description='eukaryotic translation initiation factor 4E pseudogene 4', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=81463036, end_position_on_the_genomic_accession=81463613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646108, CurrentID=0, Status='live', Symbol='RPS7P14', Aliases='RPS7_5_1787', description='ribosomal protein S7 pseudogene 14', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74409483, end_position_on_the_genomic_accession=74410144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372637, CurrentID=0, Status='live', Symbol='LOC105372637', Aliases='', description='uncharacterized LOC105372637', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47950810, end_position_on_the_genomic_accession=47961287, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372638, CurrentID=0, Status='live', Symbol='LOC105372638', Aliases='', description='uncharacterized LOC105372638', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48054523, end_position_on_the_genomic_accession=48057838, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372639, CurrentID=0, Status='live', Symbol='LOC105372639', Aliases='', description='uncharacterized LOC105372639', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48024795, end_position_on_the_genomic_accession=48067686, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372640, CurrentID=0, Status='live', Symbol='LOC105372640', Aliases='', description='uncharacterized LOC105372640', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48119691, end_position_on_the_genomic_accession=48126814, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646113, CurrentID=0, Status='live', Symbol='LINC00643', Aliases='', description='long intergenic non-protein coding RNA 643', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62117357, end_position_on_the_genomic_accession=62134186, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646114, CurrentID=0, Status='live', Symbol='AKAP8P1', Aliases='', description='A-kinase anchoring protein 8 pseudogene 1', other_designations='A kinase (PRKA) anchor protein 8 pseudogene 1', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11011749, end_position_on_the_genomic_accession=11013098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372643, CurrentID=0, Status='live', Symbol='LOC105372643', Aliases='', description='uncharacterized LOC105372643', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48394162, end_position_on_the_genomic_accession=48399443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372644, CurrentID=0, Status='live', Symbol='LOC105372644', Aliases='', description='uncharacterized LOC105372644', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48409272, end_position_on_the_genomic_accession=48410716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372645, CurrentID=0, Status='live', Symbol='LOC105372645', Aliases='', description='uncharacterized LOC105372645', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48471320, end_position_on_the_genomic_accession=48477559, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376670, CurrentID=0, Status='live', Symbol='LOC105376670', Aliases='', description='60S ribosomal protein L13a pseudogene', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14435699, end_position_on_the_genomic_accession=14441702, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646119, CurrentID=0, Status='live', Symbol='RPS8P4', Aliases='RPS8_2_1103', description='ribosomal protein S8 pseudogene 4', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119638433, end_position_on_the_genomic_accession=119639134, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372648, CurrentID=0, Status='live', Symbol='LOC105372648', Aliases='', description='uncharacterized LOC105372648', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=48835602, end_position_on_the_genomic_accession=48888173, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129769, CurrentID=0, Status='live', Symbol='AKR1B1P6', Aliases='', description='aldo-keto reductase family 1, member B1 pseudogene 6', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=8545649, end_position_on_the_genomic_accession=8546759, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372650, CurrentID=0, Status='live', Symbol='LOC105372650', Aliases='', description='uncharacterized LOC105372650', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49517321, end_position_on_the_genomic_accession=49520374, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372651, CurrentID=0, Status='live', Symbol='LOC105372651', Aliases='', description='uncharacterized LOC105372651', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49594585, end_position_on_the_genomic_accession=49602528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129772, CurrentID=0, Status='live', Symbol='TMPOP1', Aliases='Em:AB014080.3, TMPOL1', description='thymopoietin pseudogene 1', other_designations='thymopoietin-like 1', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30466225, end_position_on_the_genomic_accession=30468869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372653, CurrentID=0, Status='live', Symbol='LOC105372653', Aliases='', description='uncharacterized LOC105372653', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49915822, end_position_on_the_genomic_accession=49928854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  121838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121838, CurrentID=0, Status='live', Symbol='LINC00284', Aliases='NCRNA00284', description='long intergenic non-protein coding RNA 284', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44022335, end_position_on_the_genomic_accession=44030463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129775, CurrentID=0, Status='live', Symbol='LOC100129775', Aliases='', description='uncharacterized LOC100129775', other_designations='', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=74990231, end_position_on_the_genomic_accession=74990629, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372656, CurrentID=0, Status='live', Symbol='LOC105372656', Aliases='', description='uncharacterized LOC105372656', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50252723, end_position_on_the_genomic_accession=50264246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129777, CurrentID=0, Status='live', Symbol='BNIP3P6', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 6', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18418715, end_position_on_the_genomic_accession=18419246, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372658, CurrentID=0, Status='live', Symbol='LOC105372658', Aliases='', description='uncharacterized LOC105372658', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50378609, end_position_on_the_genomic_accession=50397021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372659, CurrentID=0, Status='live', Symbol='LOC105372659', Aliases='', description='uncharacterized LOC105372659', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50667520, end_position_on_the_genomic_accession=50674627, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372660, CurrentID=0, Status='live', Symbol='LOC105372660', Aliases='', description='uncharacterized LOC105372660', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50871177, end_position_on_the_genomic_accession=50874367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129781, CurrentID=0, Status='live', Symbol='MIR193BHG', Aliases='', description='MIR193B host gene', other_designations='', map_location='16p13.12', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14321463, end_position_on_the_genomic_accession=14326519, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129782, CurrentID=0, Status='live', Symbol='ATP5A1P4', Aliases='ATP5AP4', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1, cardiac muscle pseudogene|ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit, pseudogene 4', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=62604113, end_position_on_the_genomic_accession=62605722, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372663, CurrentID=0, Status='live', Symbol='LOC105372663', Aliases='', description='uncharacterized LOC105372663', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=51366364, end_position_on_the_genomic_accession=51396988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129784, CurrentID=0, Status='live', Symbol='RPL17P26', Aliases='RPL17_14_707', description='ribosomal protein L17 pseudogene 26', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56871149, end_position_on_the_genomic_accession=56871780, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372665, CurrentID=0, Status='live', Symbol='LOC105372665', Aliases='', description='uncharacterized LOC105372665', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52609797, end_position_on_the_genomic_accession=52619168, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372666, CurrentID=0, Status='live', Symbol='LOC105372666', Aliases='', description='uncharacterized LOC105372666', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52540823, end_position_on_the_genomic_accession=52693843, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105372667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372667, CurrentID=0, Status='live', Symbol='LOC105372667', Aliases='', description='uncharacterized LOC105372667', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52469241, end_position_on_the_genomic_accession=52471666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372668, CurrentID=0, Status='live', Symbol='LOC105372668', Aliases='', description='uncharacterized LOC105372668', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52840946, end_position_on_the_genomic_accession=52895423, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372669, CurrentID=0, Status='live', Symbol='LOC105372669', Aliases='', description='uncharacterized LOC105372669', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53434174, end_position_on_the_genomic_accession=53461505, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372670, CurrentID=0, Status='live', Symbol='LOC105372670', Aliases='', description='uncharacterized LOC105372670', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53399262, end_position_on_the_genomic_accession=53401377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372671, CurrentID=0, Status='live', Symbol='LOC105372671', Aliases='', description='uncharacterized LOC105372671', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53514111, end_position_on_the_genomic_accession=53525280, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129792, CurrentID=0, Status='live', Symbol='CCDC152', Aliases='CH5400', description='coiled-coil domain containing 152', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42756818, end_position_on_the_genomic_accession=42802437, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100129793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129793, CurrentID=0, Status='live', Symbol='TLK1P1', Aliases='TLK1P, TLK1ps', description='tousled like kinase 1 pseudogene 1', other_designations='tousled-like kinase 1 pseudogene 1', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122685940, end_position_on_the_genomic_accession=122688431, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129794, CurrentID=0, Status='live', Symbol='SLC25A21-AS1', Aliases='', description='SLC25A21 antisense RNA 1', other_designations='', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=37172026, end_position_on_the_genomic_accession=37173794, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100129795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129795, CurrentID=0, Status='live', Symbol='NPM1P50', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 50', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75079171, end_position_on_the_genomic_accession=75080228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129796, CurrentID=0, Status='live', Symbol='RPL13P13', Aliases='RPL13_5_1672', description='ribosomal protein L13 pseudogene 13', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56217168, end_position_on_the_genomic_accession=56217952, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372677, CurrentID=0, Status='live', Symbol='LOC105372677', Aliases='', description='uncharacterized LOC105372677', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55803900, end_position_on_the_genomic_accession=55829685, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372678, CurrentID=0, Status='live', Symbol='LOC105372678', Aliases='', description='uncharacterized LOC105372678', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=55839011, end_position_on_the_genomic_accession=55858371, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129799, CurrentID=0, Status='live', Symbol='RPL13AP21', Aliases='RPL13A_10_1231', description='ribosomal protein L13a pseudogene 21', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46004026, end_position_on_the_genomic_accession=46004685, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129800, CurrentID=0, Status='live', Symbol='ZNF807', Aliases='ZNF181P1P, ZNF807', description='zinc finger protein 807', other_designations='zinc finger protein 807, pseudogene', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34541692, end_position_on_the_genomic_accession=34545404, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129801, CurrentID=0, Status='live', Symbol='RPL12P37', Aliases='RPL12_18_1579', description='ribosomal protein L12 pseudogene 37', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80976399, end_position_on_the_genomic_accession=80976990, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372682, CurrentID=0, Status='live', Symbol='LOC105372682', Aliases='', description='uncharacterized LOC105372682', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56558911, end_position_on_the_genomic_accession=56596608, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105372683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372683, CurrentID=0, Status='live', Symbol='LOC105372683', Aliases='', description='uncharacterized LOC105372683', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56975134, end_position_on_the_genomic_accession=56985778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372684, CurrentID=0, Status='live', Symbol='LOC105372684', Aliases='', description='uncharacterized LOC105372684', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57048749, end_position_on_the_genomic_accession=57051527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372685, CurrentID=0, Status='live', Symbol='LOC105372685', Aliases='', description='uncharacterized LOC105372685', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57117189, end_position_on_the_genomic_accession=57135984, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372686, CurrentID=0, Status='live', Symbol='LOC105372686', Aliases='', description='uncharacterized LOC105372686', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57158789, end_position_on_the_genomic_accession=57162383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480895, CurrentID=0, Status='live', Symbol='RN7SKP200', Aliases='', description='RNA, 7SK small nuclear pseudogene 200', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129808, CurrentID=0, Status='live', Symbol='RPL24P6', Aliases='RPL24_3_450', description='ribosomal protein L24 pseudogene 6', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195554570, end_position_on_the_genomic_accession=195555127, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646161, CurrentID=0, Status='live', Symbol='RPL26P21', Aliases='RPL26_11_785', description='ribosomal protein L26 pseudogene 21', other_designations='', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=13871784, end_position_on_the_genomic_accession=13872556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103504737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504737, CurrentID=0, Status='live', Symbol='MIR8485', Aliases='hsa-mir-8485', description='microRNA 8485', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=50696172, end_position_on_the_genomic_accession=50696262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422941, CurrentID=0, Status='live', Symbol='MIR4272', Aliases='', description='microRNA 4272', other_designations='hsa-mir-4272', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67225464, end_position_on_the_genomic_accession=67225527, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  252949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252949, CurrentID=0, Status='live', Symbol='TTTY17A', Aliases='NCRNA00140, TTTY17', description='testis-specific transcript, Y-linked 17A (non-protein coding)', other_designations='', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22851584, end_position_on_the_genomic_accession=22852715, orientation='plus', exon_count=2, OMIM=400040, no_hdr0=''),
  252950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252950, CurrentID=0, Status='live', Symbol='TTTY18', Aliases='NCRNA00143', description='testis-specific transcript, Y-linked 18 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8683370, end_position_on_the_genomic_accession=8683878, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  252951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252951, CurrentID=0, Status='live', Symbol='TTTY20', Aliases='NCRNA00145', description='testis-specific transcript, Y-linked 20 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9329880, end_position_on_the_genomic_accession=9334832, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  252952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252952, CurrentID=0, Status='live', Symbol='TTTY19', Aliases='NCRNA00144', description='testis-specific transcript, Y-linked 19 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8704472, end_position_on_the_genomic_accession=8705283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  252953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252953, CurrentID=0, Status='live', Symbol='TTTY21', Aliases='NCRNA00146', description='testis-specific transcript, Y-linked 21 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9717653, end_position_on_the_genomic_accession=9721296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  252954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252954, CurrentID=0, Status='live', Symbol='TTTY22', Aliases='NCRNA00147', description='testis-specific transcript, Y-linked 22 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9801153, end_position_on_the_genomic_accession=9813245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  252955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252955, CurrentID=0, Status='live', Symbol='TTTY23', Aliases='NCRNA00148B, TTTY23', description='testis-specific transcript, Y-linked 23 (non-protein coding)', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9910798, end_position_on_the_genomic_accession=9911962, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  252956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252956, CurrentID=0, Status='live', Symbol='TPM3P1', Aliases='TPM3L2', description='tropomyosin 3 pseudogene 1', other_designations='tropomyosin 3-like 2 pseudogene', map_location='7q31.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116970897, end_position_on_the_genomic_accession=116972985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372701, CurrentID=0, Status='live', Symbol='LOC105372701', Aliases='', description='uncharacterized LOC105372701', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=61179148, end_position_on_the_genomic_accession=61185077, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646174, CurrentID=0, Status='live', Symbol='C16orf90', Aliases='', description='chromosome 16 open reading frame 90', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3493484, end_position_on_the_genomic_accession=3495721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646175, CurrentID=0, Status='live', Symbol='RPL36AP48', Aliases='RPL36A_23_1571', description='ribosomal protein L36a pseudogene 48', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67224700, end_position_on_the_genomic_accession=67225086, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129824, CurrentID=0, Status='live', Symbol='PIP5K1P2', Aliases='', description='phosphatidylinositol-4-phosphate 5-kinase, type I, pseudogene 2', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=159230996, end_position_on_the_genomic_accession=159235329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646177, CurrentID=0, Status='live', Symbol='SPATA31E2P', Aliases='FAM75E2P', description='SPATA31 subfamily E, member 2, pseudogene', other_designations='family with sequence similarity 75, member E2, pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20593367, end_position_on_the_genomic_accession=20597348, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372706, CurrentID=0, Status='live', Symbol='LOC105372706', Aliases='', description='uncharacterized LOC105372706', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62101143, end_position_on_the_genomic_accession=62124048, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100129827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129827, CurrentID=0, Status='live', Symbol='MRVI1-AS1', Aliases='', description='MRVI1 antisense RNA 1', other_designations='MRVI1 antisense RNA 1 (non-protein coding)', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10541236, end_position_on_the_genomic_accession=10599932, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372708, CurrentID=0, Status='live', Symbol='LOC105372708', Aliases='', description='uncharacterized LOC105372708', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62373817, end_position_on_the_genomic_accession=62378978, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372709, CurrentID=0, Status='live', Symbol='LOC105372709', Aliases='', description='uncharacterized LOC105372709', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62402231, end_position_on_the_genomic_accession=62406314, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372710, CurrentID=0, Status='live', Symbol='LOC105372710', Aliases='', description='uncharacterized LOC105372710', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62427792, end_position_on_the_genomic_accession=62449546, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100129831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129831, CurrentID=0, Status='live', Symbol='EPWW6493', Aliases='', description='uncharacterized LOC100129831', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=2294613, end_position_on_the_genomic_accession=2295571, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372712, CurrentID=0, Status='live', Symbol='WI2-87327B8.2', Aliases='', description='uncharacterized LOC105372712', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62601377, end_position_on_the_genomic_accession=62603809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  252969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252969, CurrentID=0, Status='live', Symbol='NEIL2', Aliases='NEH2, NEI2', description='nei-like DNA glycosylase 2', other_designations='DNA glycosylase/AP lyase Neil2|DNA-(apurinic or apyrimidinic site) lyase Neil2|endonuclease VIII-like 2|nei endonuclease VIII-like 2|nei homolog 2|nei like 2|nei-like protein 2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11769663, end_position_on_the_genomic_accession=11787345, orientation='plus', exon_count=6, OMIM=608933, no_hdr0=''),
  105372714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372714, CurrentID=0, Status='live', Symbol='LOC105372714', Aliases='', description='uncharacterized LOC105372714', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33140337, end_position_on_the_genomic_accession=33163020, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372715, CurrentID=0, Status='live', Symbol='LOC105372715', Aliases='', description='uncharacterized LOC105372715', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49608935, end_position_on_the_genomic_accession=49633617, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100129836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129836, CurrentID=0, Status='live', Symbol='COL4A2-AS2', Aliases='', description='COL4A2 antisense RNA 2', other_designations='COL4A2 antisense RNA 2 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110453466, end_position_on_the_genomic_accession=110463641, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  646189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646189, CurrentID=0, Status='live', Symbol='RPL18AP13', Aliases='RPL18A_8_1622', description='ribosomal protein L18a pseudogene 13', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11521931, end_position_on_the_genomic_accession=11522550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372718, CurrentID=0, Status='live', Symbol='LOC105372718', Aliases='', description='uncharacterized LOC105372718', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63025482, end_position_on_the_genomic_accession=63028012, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372719, CurrentID=0, Status='live', Symbol='LOC105372719', Aliases='', description='uncharacterized LOC105372719', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63116222, end_position_on_the_genomic_accession=63122754, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646192, CurrentID=0, Status='live', Symbol='NOP56P1', Aliases='NOL5B, NOL5BP, NOP56-L, dJ207F6.1', description='NOP56 ribonucleoprotein pseudogene 1', other_designations='NOP56 ribonucleoprotein homolog pseudogene 1|nucleolar protein 5B pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28783633, end_position_on_the_genomic_accession=28784004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646193, CurrentID=0, Status='live', Symbol='TRAPPC13P1', Aliases='', description='trafficking protein particle complex 13 pseudogene 1', other_designations='', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=76655505, end_position_on_the_genomic_accession=76656735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129842, CurrentID=0, Status='live', Symbol='ZNF737', Aliases='ZNF102', description='zinc finger protein 737', other_designations='zinc finger protein 102 (Y3)', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20531224, end_position_on_the_genomic_accession=20565847, orientation='minus', exon_count=9, OMIM=603984, no_hdr0=''),
  646195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646195, CurrentID=0, Status='live', Symbol='RPS28P7', Aliases='RPS28_3_1160', description='ribosomal protein S28 pseudogene 7', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82689530, end_position_on_the_genomic_accession=82689904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646196, CurrentID=0, Status='live', Symbol='SMARCE1P6', Aliases='', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1 pseudogene 6', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26149092, end_position_on_the_genomic_accession=26151272, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129845, CurrentID=0, Status='live', Symbol='PCOLCE-AS1', Aliases='', description='PCOLCE antisense RNA 1', other_designations='PCOLCE antisense RNA 1 (non-protein coding)', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100589401, end_position_on_the_genomic_accession=100604038, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105372726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372726, CurrentID=0, Status='live', Symbol='LOC105372726', Aliases='', description='uncharacterized LOC105372726', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63653152, end_position_on_the_genomic_accession=63655003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  252983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252983, CurrentID=0, Status='live', Symbol='STXBP4', Aliases='Synip', description='syntaxin binding protein 4', other_designations='STX4-interacting protein|syntaxin 4 interacting protein|syntaxin 4-interacting protein', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54968735, end_position_on_the_genomic_accession=55164285, orientation='plus', exon_count=19, OMIM=610415, no_hdr0=''),
  105372728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372728, CurrentID=0, Status='live', Symbol='LOC105372728', Aliases='', description='protein capicua homolog pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646201, CurrentID=0, Status='live', Symbol='BASP1P1', Aliases='', description='brain abundant, membrane attached signal protein 1 pseudogene 1', other_designations='', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=22897030, end_position_on_the_genomic_accession=22898181, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  646202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646202, CurrentID=0, Status='live', Symbol='TLK2P1', Aliases='TLK2ps1', description='tousled like kinase 2 pseudogene 1', other_designations='tousled-like kinase 2 pseudogene 1', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34036704, end_position_on_the_genomic_accession=34040141, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372731, CurrentID=0, Status='live', Symbol='LOC105372731', Aliases='', description='CMT1A duplicated region transcript 15 protein-like protein', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9026619, end_position_on_the_genomic_accession=9043096, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  121916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121916, CurrentID=0, Status='live', Symbol='CYCSP35', Aliases='HC8, HCP35', description='cytochrome c, somatic pseudogene 35', other_designations='', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98629684, end_position_on_the_genomic_accession=98629996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129853, CurrentID=0, Status='live', Symbol='DNAJC19P1', Aliases='', description='DnaJ heat shock protein family (Hsp40) member C19 pseudogene 1', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19 pseudogene 1', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67833763, end_position_on_the_genomic_accession=67834084, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372734, CurrentID=0, Status='live', Symbol='LOC105372734', Aliases='', description='uncharacterized LOC105372734', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9972439, end_position_on_the_genomic_accession=9986333, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372735, CurrentID=0, Status='live', Symbol='LOC105372735', Aliases='', description='uncharacterized LOC105372735', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10500932, end_position_on_the_genomic_accession=10514209, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129856, CurrentID=0, Status='live', Symbol='RPL37AP5', Aliases='RPL37A_2_859', description='ribosomal protein L37a pseudogene 5', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139227478, end_position_on_the_genomic_accession=139227859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372737, CurrentID=0, Status='live', Symbol='CH507-210P18.4', Aliases='', description='uncharacterized LOC105372737', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100129858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129858, CurrentID=0, Status='live', Symbol='SCOC-AS1', Aliases='', description='SCOC antisense RNA 1', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140283726, end_position_on_the_genomic_accession=140373392, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  252995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=252995, CurrentID=0, Status='live', Symbol='FNDC5', Aliases='FRCP2, irisin', description='fibronectin type III domain containing 5', other_designations='fibronectin type III repeat-containing protein 2', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32862268, end_position_on_the_genomic_accession=32872492, orientation='minus', exon_count=7, OMIM=611906, no_hdr0=''),
  100129860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129860, CurrentID=0, Status='live', Symbol='LOC100129860', Aliases='', description='uncharacterized LOC100129860', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52199840, end_position_on_the_genomic_accession=52201430, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129861, CurrentID=0, Status='live', Symbol='RSL24D1P7', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 7', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5071289, end_position_on_the_genomic_accession=5072924, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372742, CurrentID=0, Status='live', Symbol='LOC105372742', Aliases='', description='uncharacterized LOC105372742', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=17928945, end_position_on_the_genomic_accession=17931326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100421473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421473, CurrentID=0, Status='live', Symbol='TVP23BP2', Aliases='FAM18B4P', description='TVP23B pseudogene 2', other_designations='family with sequence similarity 18, member B pseudogene|family with sequence similarity 18, member B4, pseudogene|trans-golgi network vesicle protein 23 homolog B pseudogene 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74628328, end_position_on_the_genomic_accession=74628942, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372744, CurrentID=0, Status='live', Symbol='LOC105372744', Aliases='', description='uncharacterized LOC105372744', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19649775, end_position_on_the_genomic_accession=19657603, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372745, CurrentID=0, Status='live', Symbol='LOC105372745', Aliases='', description='uncharacterized LOC105372745', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=19899946, end_position_on_the_genomic_accession=20022665, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100129866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129866, CurrentID=0, Status='live', Symbol='EI24P2', Aliases='', description='etoposide induced 2.4 pseudogene 2', other_designations='etoposide induced 2.4 mRNA pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158454198, end_position_on_the_genomic_accession=158455270, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129867, CurrentID=0, Status='live', Symbol='RPL30P14', Aliases='RPL30_8_1609', description='ribosomal protein L30 pseudogene 14', other_designations='', map_location='18q21.33', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=61404808, end_position_on_the_genomic_accession=61405275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372748, CurrentID=0, Status='live', Symbol='LOC105372748', Aliases='', description='uncharacterized LOC105372748', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23358430, end_position_on_the_genomic_accession=23371752, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372749, CurrentID=0, Status='live', Symbol='LOC105372749', Aliases='', description='uncharacterized LOC105372749', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23704243, end_position_on_the_genomic_accession=23709301, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372750, CurrentID=0, Status='live', Symbol='LOC105372750', Aliases='', description='uncharacterized LOC105372750', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23812622, end_position_on_the_genomic_accession=23876401, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100129871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129871, CurrentID=0, Status='live', Symbol='GRPEL2P3', Aliases='', description='GrpE-like 2, mitochondrial (E. coli) pseudogene 3', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94782888, end_position_on_the_genomic_accession=94783526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372752, CurrentID=0, Status='live', Symbol='LOC105372752', Aliases='', description='uncharacterized LOC105372752', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25450336, end_position_on_the_genomic_accession=25466499, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372753, CurrentID=0, Status='live', Symbol='LOC105372753', Aliases='', description='uncharacterized LOC105372753', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25496794, end_position_on_the_genomic_accession=25518895, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372754, CurrentID=0, Status='live', Symbol='LOC105372754', Aliases='', description='uncharacterized LOC105372754', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25572131, end_position_on_the_genomic_accession=25579769, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646227, CurrentID=0, Status='live', Symbol='LOC646227', Aliases='', description='cAMP-regulated phosphoprotein 19kDa pseudogene', other_designations='cAMP-regulated phosphoprotein 19 pseudogene|cAMP-regulated phosphoprotein, 19kDa pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180982174, end_position_on_the_genomic_accession=180987378, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253012, CurrentID=0, Status='live', Symbol='HEPACAM2', Aliases='MIKI', description='HEPACAM family member 2', other_designations='mitotic kinetics regulator', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93188586, end_position_on_the_genomic_accession=93232297, orientation='minus', exon_count=13, OMIM=614133, no_hdr0=''),
  253013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253013, CurrentID=0, Status='live', Symbol='RPL31P8', Aliases='RPL31_8_567', description='ribosomal protein L31 pseudogene 8', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60429863, end_position_on_the_genomic_accession=60430302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372758, CurrentID=0, Status='live', Symbol='LOC105372758', Aliases='', description='uncharacterized LOC105372758', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26432006, end_position_on_the_genomic_accession=26437483, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646231, CurrentID=0, Status='live', Symbol='AGPAT5P1', Aliases='', description='1-acylglycerol-3-phosphate O-acyltransferase 5 pseudogene 1', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=3293808, end_position_on_the_genomic_accession=3294809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129880, CurrentID=0, Status='live', Symbol='NENFP2', Aliases='', description='neudesin neurotrophic factor pseudogene 2', other_designations='', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101963992, end_position_on_the_genomic_accession=101964471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253017, CurrentID=0, Status='live', Symbol='TECRL', Aliases='GPSN2L, SRD5A2L2, TERL', description='trans-2,3-enoyl-CoA reductase-like', other_designations='glycoprotein, synaptic 2-like|steroid 5 alpha-reductase 2-like 2|steroid 5-alpha-reductase 2-like 2 protein', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64276298, end_position_on_the_genomic_accession=64409494, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  253018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253018, CurrentID=0, Status='live', Symbol='HCG27', Aliases='bCX101P6.9, bPG299F13.9, bQB115I13.2', description='HLA complex group 27 (non-protein coding)', other_designations='bPG299F13.9 bCX101P6.9 bQB115I13.2', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31197760, end_position_on_the_genomic_accession=31203968, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372763, CurrentID=0, Status='live', Symbol='LOC105372763', Aliases='', description='uncharacterized LOC105372763', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=27638581, end_position_on_the_genomic_accession=27674774, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100129884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129884, CurrentID=0, Status='live', Symbol='LOC100129884', Aliases='', description='uncharacterized LOC100129884', other_designations='', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=72221587, end_position_on_the_genomic_accession=72222406, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372765, CurrentID=0, Status='live', Symbol='LOC105372765', Aliases='', description='uncharacterized LOC105372765', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28922497, end_position_on_the_genomic_accession=28923668, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372766, CurrentID=0, Status='live', Symbol='LOC105372766', Aliases='', description='uncharacterized LOC105372766', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28993096, end_position_on_the_genomic_accession=28994162, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481321, CurrentID=0, Status='live', Symbol='RNU6-450P', Aliases='', description='RNA, U6 small nuclear 450, pseudogene', other_designations='', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  121952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=121952, CurrentID=0, Status='live', Symbol='METTL21EP', Aliases='METTL21CP1', description='methyltransferase like 21E, pseudogene', other_designations='methyltransferase like 21C pseudogene 1', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102880099, end_position_on_the_genomic_accession=102896033, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646241, CurrentID=0, Status='live', Symbol='LOC646241', Aliases='', description='uncharacterized LOC646241', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17807277, end_position_on_the_genomic_accession=17930489, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105372770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372770, CurrentID=0, Status='live', Symbol='LOC105372770', Aliases='', description='uncharacterized LOC105372770', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29492138, end_position_on_the_genomic_accession=29495926, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646243, CurrentID=0, Status='live', Symbol='CXADRP2', Aliases='', description='coxsackie virus and adenovirus receptor pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21371493, end_position_on_the_genomic_accession=21373968, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372772, CurrentID=0, Status='live', Symbol='LOC105372772', Aliases='', description='uncharacterized LOC105372772', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30393250, end_position_on_the_genomic_accession=30475708, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100421478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421478, CurrentID=0, Status='live', Symbol='FAM71BP1', Aliases='', description='family with sequence similarity 71 member B pseudogene 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25937911, end_position_on_the_genomic_accession=25938437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129894, CurrentID=0, Status='live', Symbol='LOC100129894', Aliases='', description='uncharacterized LOC100129894', other_designations='', map_location='10p15.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=1324679, end_position_on_the_genomic_accession=1325065, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372775, CurrentID=0, Status='live', Symbol='LOC105372775', Aliases='', description='uncharacterized LOC105372775', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31014384, end_position_on_the_genomic_accession=31019190, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372776, CurrentID=0, Status='live', Symbol='LOC105372776', Aliases='', description='uncharacterized LOC105372776', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30805580, end_position_on_the_genomic_accession=31004816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646249, CurrentID=0, Status='live', Symbol='PCED1CP', Aliases='', description='PC-esterase domain containing 1C, pseudogene', other_designations='family with sequence similarity 113, member A pseudogene|hCG1645016', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212832161, end_position_on_the_genomic_accession=212832925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372778, CurrentID=0, Status='live', Symbol='LOC105372778', Aliases='', description='uncharacterized LOC105372778', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32156296, end_position_on_the_genomic_accession=32247342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  544615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=544615, CurrentID=0, Status='live', Symbol='BW56', Aliases='', description='Body weight QTL 56', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372780, CurrentID=0, Status='live', Symbol='LOC105372780', Aliases='', description='uncharacterized LOC105372780', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32543589, end_position_on_the_genomic_accession=32547223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372781, CurrentID=0, Status='live', Symbol='LOC105372781', Aliases='', description='uncharacterized LOC105372781', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32598794, end_position_on_the_genomic_accession=32601315, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129902, CurrentID=0, Status='live', Symbol='RPS29P11', Aliases='RPS29_8_466', description='ribosomal protein S29 pseudogene 11', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25678816, end_position_on_the_genomic_accession=25679109, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253039, CurrentID=0, Status='live', Symbol='PSMD5-AS1', Aliases='', description='PSMD5 antisense RNA 1 (head to head)', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120843042, end_position_on_the_genomic_accession=120854373, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372784, CurrentID=0, Status='live', Symbol='LOC105372784', Aliases='', description='uncharacterized LOC105372784', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32926856, end_position_on_the_genomic_accession=33019816, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129905, CurrentID=0, Status='live', Symbol='RPS27P19', Aliases='RPS27_7_1180', description='ribosomal protein S27 pseudogene 19', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117035740, end_position_on_the_genomic_accession=117036085, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372786, CurrentID=0, Status='live', Symbol='LOC105372786', Aliases='', description='uncharacterized LOC105372786', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33077484, end_position_on_the_genomic_accession=33078260, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372787, CurrentID=0, Status='live', Symbol='LOC105372787', Aliases='', description='uncharacterized LOC105372787', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=33172958, end_position_on_the_genomic_accession=33195517, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  253044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253044, CurrentID=0, Status='live', Symbol='LINGO1-AS1', Aliases='', description='LINGO1 antisense RNA 1', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=77641764, end_position_on_the_genomic_accession=77649240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372789, CurrentID=0, Status='live', Symbol='LOC105372789', Aliases='', description='uncharacterized LOC105372789', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34016903, end_position_on_the_genomic_accession=34019035, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646262, CurrentID=0, Status='live', Symbol='LACTBL1', Aliases='', description='lactamase beta like 1', other_designations='lactamase, beta-like 1', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22953043, end_position_on_the_genomic_accession=22972316, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105372791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372791, CurrentID=0, Status='live', Symbol='LOC105372791', Aliases='', description='uncharacterized LOC105372791', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34353037, end_position_on_the_genomic_accession=34375396, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372959, CurrentID=0, Status='live', Symbol='LOC105372959', Aliases='', description='uncharacterized LOC105372959', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372793, CurrentID=0, Status='live', Symbol='LOC105372793', Aliases='', description='uncharacterized LOC105372793', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=34404723, end_position_on_the_genomic_accession=34418061, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372794, CurrentID=0, Status='live', Symbol='LOC105372794', Aliases='', description='uncharacterized LOC105372794', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=36107893, end_position_on_the_genomic_accession=36111219, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372795, CurrentID=0, Status='live', Symbol='LOC105372795', Aliases='', description='uncharacterized LOC105372795', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37063040, end_position_on_the_genomic_accession=37064216, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480667, CurrentID=0, Status='live', Symbol='RNU1-101P', Aliases='', description='RNA, U1 small nuclear 101, pseudogene', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646269, CurrentID=0, Status='live', Symbol='RPS4XP8', Aliases='RPS4P8, RPS4X_5_772', description='ribosomal protein S4X pseudogene 8', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=154576256, end_position_on_the_genomic_accession=154577141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646270, CurrentID=0, Status='live', Symbol='ELL2P1', Aliases='', description='elongation factor, RNA polymerase II, 2 pseudogene 1', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158174353, end_position_on_the_genomic_accession=158178075, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129919, CurrentID=0, Status='live', Symbol='HMGB1P45', Aliases='', description='high mobility group box 1 pseudogene 45', other_designations='high-mobility group box 1 pseudogene 45|high-mobility group box 2 pseudogene', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50398862, end_position_on_the_genomic_accession=50399370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129920, CurrentID=0, Status='live', Symbol='HNRNPA1P40', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 40', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74354443, end_position_on_the_genomic_accession=74355268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646273, CurrentID=0, Status='live', Symbol='HSPD1P15', Aliases='HSPD1-15P', description='heat shock protein family D (Hsp60) member 1 pseudogene 15', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 15', map_location='5p14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=19233366, end_position_on_the_genomic_accession=19234539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372802, CurrentID=0, Status='live', Symbol='LOC105372802', Aliases='', description='uncharacterized LOC105372802', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38503440, end_position_on_the_genomic_accession=38543341, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372803, CurrentID=0, Status='live', Symbol='LOC105372803', Aliases='', description='uncharacterized LOC105372803', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=39011235, end_position_on_the_genomic_accession=39015106, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129924, CurrentID=0, Status='live', Symbol='LOC100129924', Aliases='', description='uncharacterized LOC100129924', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42787990, end_position_on_the_genomic_accession=42798230, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  646278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646278, CurrentID=0, Status='live', Symbol='PDCD6IPP2', Aliases='', description='PDCD6IP pseudogene 2', other_designations='programmed cell death 6 interacting protein pseudogene', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28789834, end_position_on_the_genomic_accession=28856574, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105372807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372807, CurrentID=0, Status='live', Symbol='LOC105372807', Aliases='', description='uncharacterized LOC105372807', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41357305, end_position_on_the_genomic_accession=41368087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129928, CurrentID=0, Status='live', Symbol='RPL31P5', Aliases='RPL31_29_1379', description='ribosomal protein L31 pseudogene 5', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63567912, end_position_on_the_genomic_accession=63568280, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421484, CurrentID=0, Status='live', Symbol='MARK2P11', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 11', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157984644, end_position_on_the_genomic_accession=157989984, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646282, CurrentID=0, Status='live', Symbol='AZGP1P1', Aliases='', description='alpha-2-glycoprotein 1, zinc-binding pseudogene 1', other_designations='alpha-2-glycoprotein 1, zinc pseudogene 1', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99980762, end_position_on_the_genomic_accession=99984237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372811, CurrentID=0, Status='live', Symbol='LOC105372811', Aliases='', description='uncharacterized LOC105372811', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=26395033, end_position_on_the_genomic_accession=26413109, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372812, CurrentID=0, Status='live', Symbol='LOC105372812', Aliases='', description='uncharacterized LOC105372812', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41637881, end_position_on_the_genomic_accession=41641787, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129933, CurrentID=0, Status='live', Symbol='GLYATL1P2', Aliases='', description='glycine-N-acyltransferase like 1 pseudogene 2', other_designations='glycine-N-acyltransferase-like 1 pseudogene 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58884947, end_position_on_the_genomic_accession=58893460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129934, CurrentID=0, Status='live', Symbol='RPL34P14', Aliases='RPL34_9_693', description='ribosomal protein L34 pseudogene 14', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43327852, end_position_on_the_genomic_accession=43328476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421485, CurrentID=0, Status='live', Symbol='MARK2P5', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15384231, end_position_on_the_genomic_accession=15384860, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372816, CurrentID=0, Status='live', Symbol='LOC105372816', Aliases='', description='uncharacterized LOC105372816', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=42663282, end_position_on_the_genomic_accession=42666981, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100129937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129937, CurrentID=0, Status='live', Symbol='KNOP1P1', Aliases='', description='lysine-rich nucleolar protein 1 pseudogene 1', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=24583186, end_position_on_the_genomic_accession=24585113, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129938, CurrentID=0, Status='live', Symbol='LOC100129938', Aliases='', description='uncharacterized LOC100129938', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=55752349, end_position_on_the_genomic_accession=55753132, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129939, CurrentID=0, Status='live', Symbol='RPS10P31', Aliases='RPS10_16_1758', description='ribosomal protein S10 pseudogene 31', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45275527, end_position_on_the_genomic_accession=45276135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129940, CurrentID=0, Status='live', Symbol='LOC100129940', Aliases='', description='uncharacterized LOC100129940', other_designations='', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65851227, end_position_on_the_genomic_accession=65883544, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100129941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129941, CurrentID=0, Status='live', Symbol='RSL24D1P5', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 5', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63055275, end_position_on_the_genomic_accession=63055765, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646294, CurrentID=0, Status='live', Symbol='RPS2P29', Aliases='RPS2_9_695', description='ribosomal protein S2 pseudogene 29', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43538785, end_position_on_the_genomic_accession=43539726, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372823, CurrentID=0, Status='live', Symbol='LOC105372823', Aliases='', description='uncharacterized LOC105372823', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6145979, end_position_on_the_genomic_accession=6149671, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372824, CurrentID=0, Status='live', Symbol='LOC105372824', Aliases='', description='uncharacterized protein C21orf124', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43737000, end_position_on_the_genomic_accession=43738097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372825, CurrentID=0, Status='live', Symbol='LOC105372825', Aliases='', description='uncharacterized LOC105372825', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43778630, end_position_on_the_genomic_accession=43779271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372826, CurrentID=0, Status='live', Symbol='LOC105372826', Aliases='', description='uncharacterized LOC105372826', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43945403, end_position_on_the_genomic_accession=43946514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122011, CurrentID=0, Status='live', Symbol='CSNK1A1L', Aliases='CK1', description='casein kinase 1, alpha 1-like', other_designations='CKI-alpha-like|casein kinase I alpha S-like', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=37103260, end_position_on_the_genomic_accession=37105664, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  646300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646300, CurrentID=0, Status='live', Symbol='COL6A4P2', Aliases='COL6A4', description='collagen, type VI, alpha 4 pseudogene 2', other_designations='alpha collagen containing double von Willebrand factor A domains|collagen VI alpha 4 pseudogene 2', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130212820, end_position_on_the_genomic_accession=130273806, orientation='plus', exon_count=23, OMIM=616612, no_hdr0=''),
  105372829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372829, CurrentID=0, Status='live', Symbol='LOC105372829', Aliases='', description='uncharacterized LOC105372829', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43993528, end_position_on_the_genomic_accession=43997486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372830, CurrentID=0, Status='live', Symbol='LOC105372830', Aliases='', description='uncharacterized LOC105372830', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372831, CurrentID=0, Status='live', Symbol='CH507-9B2.8', Aliases='', description='uncharacterized LOC105372831', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5067516, end_position_on_the_genomic_accession=5089824, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100129952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129952, CurrentID=0, Status='live', Symbol='RPL39P10', Aliases='RPL39_4_182', description='ribosomal protein L39 pseudogene 10', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239052571, end_position_on_the_genomic_accession=239052952, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372833, CurrentID=0, Status='live', Symbol='LOC105372833', Aliases='', description='uncharacterized LOC105372833', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44250813, end_position_on_the_genomic_accession=44251521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372834, CurrentID=0, Status='live', Symbol='LOC105372834', Aliases='', description='uncharacterized LOC105372834', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44473512, end_position_on_the_genomic_accession=44477220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  139363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139363, CurrentID=0, Status='live', Symbol='MAGEA13P', Aliases='', description='MAGE family member A13, pseudogene', other_designations='melanoma antigen family A, 13 pseudogene|melanoma antigen family A13, pseudogene', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138482689, end_position_on_the_genomic_accession=138484285, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129956, CurrentID=0, Status='live', Symbol='LOC100129956', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153051342, end_position_on_the_genomic_accession=153052505, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646309, CurrentID=0, Status='live', Symbol='NAMPTP1', Aliases='NAMPTL, PBEF2, bA92J19.4', description='nicotinamide phosphoribosyltransferase pseudogene 1', other_designations='nicotinamide phosphoribosyltransferase-like|novel protein similar to Pre-B cell enhancing factor (PBEF)|pre-B-cell colony enhancing factor 2', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36521736, end_position_on_the_genomic_accession=36524236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129958, CurrentID=0, Status='live', Symbol='KRT8P44', Aliases='', description='keratin 8 pseudogene 44', other_designations='', map_location='6q26', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=162568303, end_position_on_the_genomic_accession=162569966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129959, CurrentID=0, Status='live', Symbol='AK3P4', Aliases='', description='adenylate kinase 3 pseudogene 4', other_designations='', map_location='5q22.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115402340, end_position_on_the_genomic_accession=115402964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103504738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103504738, CurrentID=0, Status='live', Symbol='MIR6859-4', Aliases='hsa-mir-6859-4', description='microRNA 6859-4', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=17052, end_position_on_the_genomic_accession=17119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100129961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129961, CurrentID=0, Status='live', Symbol='CCNT2-AS1', Aliases='', description='CCNT2 antisense RNA 1', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134866633, end_position_on_the_genomic_accession=134918606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105372842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372842, CurrentID=0, Status='live', Symbol='LOC105372842', Aliases='', description='uncharacterized LOC105372842', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46036807, end_position_on_the_genomic_accession=46039533, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372843, CurrentID=0, Status='live', Symbol='LOC105372843', Aliases='', description='uncharacterized LOC105372843', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=43191604, end_position_on_the_genomic_accession=43195876, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646316, CurrentID=0, Status='live', Symbol='TERF1P3', Aliases='', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene 3', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=136353510, end_position_on_the_genomic_accession=136356337, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372845, CurrentID=0, Status='live', Symbol='LOC105372845', Aliases='', description='uncharacterized LOC105372845', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15651869, end_position_on_the_genomic_accession=15689274, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105372846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372846, CurrentID=0, Status='live', Symbol='LOC105372846', Aliases='', description='uncharacterized LOC105372846', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=15517078, end_position_on_the_genomic_accession=15543919, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372847, CurrentID=0, Status='live', Symbol='LOC105372847', Aliases='', description='uncharacterized LOC105372847', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372848, CurrentID=0, Status='live', Symbol='LOC105372848', Aliases='', description='uncharacterized LOC105372848', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17085953, end_position_on_the_genomic_accession=17094027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100129969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129969, CurrentID=0, Status='live', Symbol='FAM205C', Aliases='FAM205CP', description='family with sequence similarity 205 member C', other_designations='family with sequence similarity 205, member C|family with sequence similarity 205, member C, pseudogene|putative family with sequence similarity 205, member C protein|transmembrane protein C9orf144B pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34889108, end_position_on_the_genomic_accession=34895740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372850, CurrentID=0, Status='live', Symbol='LOC105372850', Aliases='', description='uncharacterized LOC105372850', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17563728, end_position_on_the_genomic_accession=17570154, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372851, CurrentID=0, Status='live', Symbol='LOC105372851', Aliases='', description='uncharacterized LOC105372851', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17878120, end_position_on_the_genomic_accession=17881192, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646324, CurrentID=0, Status='live', Symbol='LINC00607', Aliases='', description='long intergenic non-protein coding RNA 607', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215611563, end_position_on_the_genomic_accession=215843536, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100129973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129973, CurrentID=0, Status='live', Symbol='LOC100129973', Aliases='', description='uncharacterized LOC100129973', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52180026, end_position_on_the_genomic_accession=52205879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122038, CurrentID=0, Status='live', Symbol='TOMM22P3', Aliases='', description='TOMM22 pseudogene 3', other_designations='hCG1643176|translocase of outer mitochondrial membrane 22 homolog pseudogene 3', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32953857, end_position_on_the_genomic_accession=32954464, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100129975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129975, CurrentID=0, Status='live', Symbol='RPL10P15', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9820150, end_position_on_the_genomic_accession=9820518, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100129976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129976, CurrentID=0, Status='live', Symbol='BRI3BPP1', Aliases='', description='BRI3 binding protein pseudogene 1', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21800480, end_position_on_the_genomic_accession=21801187, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372857, CurrentID=0, Status='live', Symbol='LOC105372857', Aliases='', description='uncharacterized LOC105372857', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206057447, end_position_on_the_genomic_accession=206072340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122042, CurrentID=0, Status='live', Symbol='RXFP2', Aliases='GPR106, GREAT, INSL3R, LGR8, LGR8.1, RXFPR2', description='relaxin/insulin-like family peptide receptor 2', other_designations='G protein coupled receptor affecting testicular descent|G-protein coupled receptor 106|G-protein coupled receptor affecting testicular descent|leucine-rich repeat-containing G protein-coupled receptor 8|leucine-rich repeat-containing G-protein coupled receptor 8|relaxin family peptide receptor 2', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=31739542, end_position_on_the_genomic_accession=31802872, orientation='plus', exon_count=18, OMIM=606655, no_hdr0=''),
  105372859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372859, CurrentID=0, Status='live', Symbol='LOC105372859', Aliases='', description='uncharacterized LOC105372859', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19291914, end_position_on_the_genomic_accession=19351129, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372860, CurrentID=0, Status='live', Symbol='LOC105372860', Aliases='', description='uncharacterized LOC105372860', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19448238, end_position_on_the_genomic_accession=19451688, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372861, CurrentID=0, Status='live', Symbol='LOC105372861', Aliases='', description='uncharacterized LOC105372861', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19727703, end_position_on_the_genomic_accession=19738707, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122046, CurrentID=0, Status='live', Symbol='TEX26', Aliases='C13orf26', description='testis expressed 26', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30932684, end_position_on_the_genomic_accession=30975502, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100129983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129983, CurrentID=0, Status='live', Symbol='UBE2CP3', Aliases='', description='ubiquitin conjugating enzyme E2C pseudogene 3', other_designations='ubiquitin-conjugating enzyme E2C pseudogene 3', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57072549, end_position_on_the_genomic_accession=57073311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372864, CurrentID=0, Status='live', Symbol='LOC105372864', Aliases='', description='uncharacterized LOC105372864', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20244793, end_position_on_the_genomic_accession=20253854, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372865, CurrentID=0, Status='live', Symbol='LOC105372865', Aliases='', description='uncharacterized LOC105372865', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372866, CurrentID=0, Status='live', Symbol='LOC105372866', Aliases='', description='uncharacterized LOC105372866', other_designations='XXbac-B33L19.4', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20343229, end_position_on_the_genomic_accession=20354333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372867, CurrentID=0, Status='live', Symbol='LOC105372867', Aliases='', description='putative uncharacterized protein LOC388882', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18750562, end_position_on_the_genomic_accession=18757906, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105372868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372868, CurrentID=0, Status='live', Symbol='LOC105372868', Aliases='', description='uncharacterized LOC105372868', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372869, CurrentID=0, Status='live', Symbol='LOC105372869', Aliases='', description='uncharacterized LOC105372869', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206126028, end_position_on_the_genomic_accession=206134117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372870, CurrentID=0, Status='live', Symbol='LOC105372870', Aliases='', description='uncharacterized LOC105372870', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18548798, end_position_on_the_genomic_accession=18574563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100129991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129991, CurrentID=0, Status='live', Symbol='RPL38P2', Aliases='RPL38_1_244', description='ribosomal protein L38 pseudogene 2', other_designations='', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77792751, end_position_on_the_genomic_accession=77793095, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253128, CurrentID=0, Status='live', Symbol='LINC00612', Aliases='C12orf33', description='long intergenic non-protein coding RNA 612', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9055589, end_position_on_the_genomic_accession=9065070, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372873, CurrentID=0, Status='live', Symbol='LOC105372873', Aliases='', description='uncharacterized LOC105372873', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206147199, end_position_on_the_genomic_accession=206165682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100129994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129994, CurrentID=0, Status='live', Symbol='GRPEL2P2', Aliases='', description='GrpE-like 2, mitochondrial (E. coli) pseudogene 2', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64697875, end_position_on_the_genomic_accession=64698548, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421495, CurrentID=0, Status='live', Symbol='ACTR6P1', Aliases='', description='ACTR6 pseudogene 1', other_designations='ARP6 actin-related protein 6 homolog pseudogene 1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106836501, end_position_on_the_genomic_accession=106837571, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122060, CurrentID=0, Status='live', Symbol='SLAIN1', Aliases='C13orf32', description='SLAIN motif family member 1', other_designations='SLAIN motif family, member 1', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=77697854, end_position_on_the_genomic_accession=77764242, orientation='plus', exon_count=10, OMIM=610491, no_hdr0=''),
  105372877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372877, CurrentID=0, Status='live', Symbol='LOC105372877', Aliases='', description='uncharacterized LOC105372877', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206748331, end_position_on_the_genomic_accession=206751594, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100129998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129998, CurrentID=0, Status='live', Symbol='RPL5P27', Aliases='RPL5_14_1100', description='ribosomal protein L5 pseudogene 27', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116749864, end_position_on_the_genomic_accession=116751332, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372879, CurrentID=0, Status='live', Symbol='LOC105372879', Aliases='', description='uncharacterized LOC105372879', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=206885309, end_position_on_the_genomic_accession=206896495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372880, CurrentID=0, Status='live', Symbol='LOC105372880', Aliases='', description='uncharacterized LOC105372880', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=207416345, end_position_on_the_genomic_accession=207425700, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421496, CurrentID=0, Status='live', Symbol='MARK2P4', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176012028, end_position_on_the_genomic_accession=176014524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130002, CurrentID=0, Status='live', Symbol='HNRNPA1P23', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 23', other_designations='', map_location='3q21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122317609, end_position_on_the_genomic_accession=122318603, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130003, CurrentID=0, Status='live', Symbol='SNRPGP2', Aliases='HsT2742', description='small nuclear ribonucleoprotein polypeptide G pseudogene 2', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54406754, end_position_on_the_genomic_accession=54407195, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372884, CurrentID=0, Status='live', Symbol='LOC105372884', Aliases='', description='uncharacterized LOC105372884', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208207260, end_position_on_the_genomic_accession=208243294, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372885, CurrentID=0, Status='live', Symbol='LOC105372885', Aliases='', description='uncharacterized LOC105372885', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208122608, end_position_on_the_genomic_accession=208128688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130006, CurrentID=0, Status='live', Symbol='EIF4A1P11', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 11', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179201624, end_position_on_the_genomic_accession=179201926, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253143, CurrentID=0, Status='live', Symbol='PRR14L', Aliases='C22orf30', description='proline rich 14-like', other_designations='proline rich 14-like protein', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31681347, end_position_on_the_genomic_accession=31750173, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  646360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646360, CurrentID=0, Status='live', Symbol='TMEM161BP1', Aliases='', description='transmembrane protein 161B pseudogene 1', other_designations='hCG25653', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37337414, end_position_on_the_genomic_accession=37339771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130009, CurrentID=0, Status='live', Symbol='HMGA1P8', Aliases='HMGA1-p', description='high mobility group AT-hook 1 pseudogene 8 (functional)', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74416811, end_position_on_the_genomic_accession=74418691, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646362, CurrentID=0, Status='live', Symbol='FTH1P13', Aliases='FTHL13, FTHP3', description='ferritin, heavy polypeptide 1 pseudogene 13', other_designations='ferritin, heavy polypeptide pseudogene 3|ferritin, heavy polypeptide-like 13', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65900774, end_position_on_the_genomic_accession=65901450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372891, CurrentID=0, Status='live', Symbol='LOC105372891', Aliases='', description='uncharacterized LOC105372891', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208626741, end_position_on_the_genomic_accession=208636018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372892, CurrentID=0, Status='live', Symbol='LOC105372892', Aliases='', description='uncharacterized LOC105372892', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=208671329, end_position_on_the_genomic_accession=208728604, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100421498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421498, CurrentID=0, Status='live', Symbol='MARK2P8', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 8', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128748156, end_position_on_the_genomic_accession=128749482, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646366, CurrentID=0, Status='live', Symbol='MAS1LP1', Aliases='MAS1LP', description='MAS1L pseudogene 1', other_designations='MAS1 oncogene-like pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29475025, end_position_on_the_genomic_accession=29476071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130015, CurrentID=0, Status='live', Symbol='URAHP', Aliases='URAH', description='urate (hydroxyiso-) hydrolase, pseudogene', other_designations='5-hydroxyisourate hydrolase pseudogene|HIU hydrolase', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90039761, end_position_on_the_genomic_accession=90047783, orientation='minus', exon_count=6, OMIM=615805, no_hdr0=''),
  253152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253152, CurrentID=0, Status='live', Symbol='EPHX4', Aliases='ABHD7, EH4, EPHXRP', description='epoxide hydrolase 4', other_designations='abhydrolase domain containing 7|abhydrolase domain-containing protein 7|epoxide hydrolase-related protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92029976, end_position_on_the_genomic_accession=92063536, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100130017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130017, CurrentID=0, Status='live', Symbol='FTLP10', Aliases='', description='ferritin, light polypeptide pseudogene 10', other_designations='', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68182292, end_position_on_the_genomic_accession=68212470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646370, CurrentID=0, Status='live', Symbol='IGHV1OR15-3', Aliases='IGHV1OR153', description='immunoglobulin heavy variable 1/OR15-3 (pseudogene)', other_designations='IGHV1/OR15-3|immunoglobulin heavy variable 1/OR15-3 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21735637, end_position_on_the_genomic_accession=21736005, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100421499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421499, CurrentID=0, Status='live', Symbol='MARK2P17', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 17', other_designations='MAP/microtubule affinity-regulating kinase 4 pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128842753, end_position_on_the_genomic_accession=128844368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372900, CurrentID=0, Status='live', Symbol='LOC105372900', Aliases='', description='uncharacterized LOC105372900', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209680924, end_position_on_the_genomic_accession=209690613, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372901, CurrentID=0, Status='live', Symbol='LOC105372901', Aliases='', description='uncharacterized LOC105372901', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210708591, end_position_on_the_genomic_accession=210753925, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372902, CurrentID=0, Status='live', Symbol='LOC105372902', Aliases='', description='uncharacterized LOC105372902', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211257807, end_position_on_the_genomic_accession=211259191, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372903, CurrentID=0, Status='live', Symbol='LOC105372903', Aliases='', description='uncharacterized LOC105372903', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211604320, end_position_on_the_genomic_accession=211609562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372904, CurrentID=0, Status='live', Symbol='LOC105372904', Aliases='', description='uncharacterized LOC105372904', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211713176, end_position_on_the_genomic_accession=211715994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646377, CurrentID=0, Status='live', Symbol='PYHIN5P', Aliases='', description='pyrin and HIN domain family member 5, pseudogene', other_designations='pyrin and HIN domain family, member 1 pseudogene|pyrin and HIN domain family, member 5, pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158878746, end_position_on_the_genomic_accession=158880896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646378, CurrentID=0, Status='live', Symbol='DDX18P5', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 5', other_designations='', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130295001, end_position_on_the_genomic_accession=130297384, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646379, CurrentID=0, Status='live', Symbol='IGHV1OR15-4', Aliases='IGHV1OR154', description='immunoglobulin heavy variable 1/OR15-4 (pseudogene)', other_designations='IGHV1/OR15-4|immunoglobulin heavy variable 1/OR15-4 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=22482589, end_position_on_the_genomic_accession=22490023, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372908, CurrentID=0, Status='live', Symbol='LOC105372908', Aliases='', description='uncharacterized LOC105372908', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212506402, end_position_on_the_genomic_accession=212514782, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130029, CurrentID=0, Status='live', Symbol='HMGA1P6', Aliases='HMGA1L6', description='high mobility group AT-hook 1 pseudogene 6', other_designations='high mobility group AT-hook 1-like 6|high mobility group protein gene A1-like 6', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23134168, end_position_on_the_genomic_accession=23135915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646382, CurrentID=0, Status='live', Symbol='DPPA5P1', Aliases='', description='developmental pluripotency associated 5 pseudogene 1', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52425745, end_position_on_the_genomic_accession=52426250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421501, CurrentID=0, Status='live', Symbol='MARK2P1', Aliases='', description='MAP/microtubule affinity-regulating kinase 2 pseudogene 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9854905, end_position_on_the_genomic_accession=9855532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372912, CurrentID=0, Status='live', Symbol='LOC105372912', Aliases='', description='uncharacterized LOC105372912', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=213731423, end_position_on_the_genomic_accession=213863908, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372913, CurrentID=0, Status='live', Symbol='LOC105372913', Aliases='', description='uncharacterized LOC105372913', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214430947, end_position_on_the_genomic_accession=214438403, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130034, CurrentID=0, Status='live', Symbol='RPL36AP12', Aliases='RPL36A_2_79', description='ribosomal protein L36a pseudogene 12', other_designations='', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100651885, end_position_on_the_genomic_accession=100652317, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372915, CurrentID=0, Status='live', Symbol='LOC105372915', Aliases='', description='uncharacterized LOC105372915', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214900601, end_position_on_the_genomic_accession=214905479, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372916, CurrentID=0, Status='live', Symbol='LOC105372916', Aliases='', description='uncharacterized LOC105372916', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214823526, end_position_on_the_genomic_accession=214833280, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481322, CurrentID=0, Status='live', Symbol='RNU6-452P', Aliases='', description='RNA, U6 small nuclear 452, pseudogene', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130038, CurrentID=0, Status='live', Symbol='OSBPL9P2', Aliases='', description='oxysterol binding protein like 9 pseudogene 2', other_designations='oxysterol binding protein-like 9 pseudogene 2', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=91113876, end_position_on_the_genomic_accession=91116547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253175, CurrentID=0, Status='live', Symbol='CDY1B', Aliases='CDY', description='chromodomain protein, Y-linked, 1B', other_designations='chromodomain protein, Y chromosome, 1, centromeric|testis-specific chromodomain protein on Y, centromeric', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24035386, end_position_on_the_genomic_accession=24048016, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646392, CurrentID=0, Status='live', Symbol='NIP7P2', Aliases='', description='NIP7, nucleolar pre-rRNA processing protein pseudogene 2', other_designations='nuclear import 7 homolog pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132401253, end_position_on_the_genomic_accession=132401792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372921, CurrentID=0, Status='live', Symbol='LOC105372921', Aliases='', description='uncharacterized LOC105372921', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130042, CurrentID=0, Status='live', Symbol='MTHFD2P6', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 6', other_designations='', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41967140, end_position_on_the_genomic_accession=41968078, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100421503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421503, CurrentID=0, Status='live', Symbol='MARK3P2', Aliases='', description='MAP/microtubule affinity-regulating kinase 3 pseudogene 2', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12857314, end_position_on_the_genomic_accession=12858686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646396, CurrentID=0, Status='live', Symbol='REREP3', Aliases='', description='arginine-glutamic acid dipeptide (RE) repeats pseudogene 3', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22258614, end_position_on_the_genomic_accession=22282880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372925, CurrentID=0, Status='live', Symbol='LOC105372925', Aliases='', description='uncharacterized LOC105372925', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218912757, end_position_on_the_genomic_accession=218918358, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372926, CurrentID=0, Status='live', Symbol='LOC105372926', Aliases='', description='uncharacterized LOC105372926', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219795448, end_position_on_the_genomic_accession=219822603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105372927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372927, CurrentID=0, Status='live', Symbol='LOC105372927', Aliases='', description='uncharacterized LOC105372927', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220289208, end_position_on_the_genomic_accession=220299259, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372928, CurrentID=0, Status='live', Symbol='LOC105372928', Aliases='', description='uncharacterized LOC105372928', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220487738, end_position_on_the_genomic_accession=220494506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372929, CurrentID=0, Status='live', Symbol='LOC105372929', Aliases='', description='uncharacterized LOC105372929', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220483165, end_position_on_the_genomic_accession=220487584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105372930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372930, CurrentID=0, Status='live', Symbol='LOC105372930', Aliases='', description='uncharacterized LOC105372930', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220974373, end_position_on_the_genomic_accession=220982379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372931, CurrentID=0, Status='live', Symbol='LOC105372931', Aliases='', description='uncharacterized LOC105372931', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221298513, end_position_on_the_genomic_accession=221302539, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372932, CurrentID=0, Status='live', Symbol='LOC105372932', Aliases='', description='uncharacterized LOC105372932', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=221304188, end_position_on_the_genomic_accession=221379923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646405, CurrentID=0, Status='live', Symbol='TPTE2P1', Aliases='', description='transmembrane phosphoinositide 3-phosphatase and tensin homolog 2 pseudogene 1', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=24928755, end_position_on_the_genomic_accession=24968469, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  253190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253190, CurrentID=0, Status='live', Symbol='SERHL2', Aliases='dJ222E13.1', description='serine hydrolase-like 2', other_designations='', map_location='22q13', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42518735, end_position_on_the_genomic_accession=42574382, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105372935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372935, CurrentID=0, Status='live', Symbol='LOC105372935', Aliases='', description='uncharacterized LOC105372935', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21192790, end_position_on_the_genomic_accession=21204019, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372936, CurrentID=0, Status='live', Symbol='LOC105372936', Aliases='', description='uncharacterized LOC105372936', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646409, CurrentID=0, Status='live', Symbol='DGAT2L7P', Aliases='DGAT2L7', description='diacylglycerol O-acyltransferase 2-like 7, pseudogene', other_designations='diacylglycerol O-acyltransferase 2 like protein 7|diacylglycerol O-acyltransferase 2-like 7 pseudogene|putative diacylglycerol O-acyltransferase 2-like protein 7|putative diacylglycerol O-acyltransferase 2-like protein DGAT2L7P', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101201813, end_position_on_the_genomic_accession=101203275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372938, CurrentID=0, Status='live', Symbol='LOC105372938', Aliases='', description='uncharacterized LOC105372938', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20522070, end_position_on_the_genomic_accession=20524365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130059, CurrentID=0, Status='live', Symbol='VN1R70P', Aliases='', description='vomeronasal 1 receptor 70 pseudogene', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35836852, end_position_on_the_genomic_accession=35837791, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130060, CurrentID=0, Status='live', Symbol='GNPATP', Aliases='', description='glyceronephosphate O-acyltransferase pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=60654071, end_position_on_the_genomic_accession=60655974, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372941, CurrentID=0, Status='live', Symbol='LOC105372941', Aliases='', description='uncharacterized LOC105372941', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105372942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372942, CurrentID=0, Status='live', Symbol='LOC105372942', Aliases='', description='uncharacterized LOC105372942', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18349938, end_position_on_the_genomic_accession=18361189, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  285378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285378, CurrentID=0, Status='live', Symbol='LOC285378', Aliases='', description='uncharacterized LOC285378', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130064, CurrentID=0, Status='live', Symbol='EIF4BP4', Aliases='', description='eukaryotic translation initiation factor 4B pseudogene 4', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42340591, end_position_on_the_genomic_accession=42342466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139604, CurrentID=0, Status='live', Symbol='MAGEB16', Aliases='', description='MAGE family member B16', other_designations='MAGE-B16 antigen|melanoma antigen family B, 16 (pseudogene)|melanoma antigen family B16', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35795541, end_position_on_the_genomic_accession=35803762, orientation='plus', exon_count=4, OMIM=300762, no_hdr0=''),
  100130066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130066, CurrentID=0, Status='live', Symbol='RPL9P22', Aliases='RPL9_6_1163', description='ribosomal protein L9 pseudogene 22', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=84639965, end_position_on_the_genomic_accession=84640659, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130067, CurrentID=0, Status='live', Symbol='HPRT1P1', Aliases='', description='hypoxanthine phosphoribosyltransferase 1 pseudogene 1', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15864937, end_position_on_the_genomic_accession=15865569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372948, CurrentID=0, Status='live', Symbol='LOC105372948', Aliases='', description='uncharacterized LOC105372948', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22820981, end_position_on_the_genomic_accession=23016133, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372949, CurrentID=0, Status='live', Symbol='LL22NC03-102D1.18', Aliases='', description='uncharacterized LOC105372949', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22692756, end_position_on_the_genomic_accession=22711594, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130070, CurrentID=0, Status='live', Symbol='RPS27P13', Aliases='RPS27_6_465', description='ribosomal protein S27 pseudogene 13', other_designations='', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22594161, end_position_on_the_genomic_accession=22594507, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130071, CurrentID=0, Status='live', Symbol='LOC100130071', Aliases='', description='GSQS6193', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=6326515, end_position_on_the_genomic_accession=6327292, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646424, CurrentID=0, Status='live', Symbol='SPINK8', Aliases='', description='serine peptidase inhibitor, Kazal type 8 (putative)', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48306846, end_position_on_the_genomic_accession=48328341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130073, CurrentID=0, Status='live', Symbol='IQSEC3P2', Aliases='', description='IQ motif and Sec7 domain 3 pseudogene 2', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11500471, end_position_on_the_genomic_accession=11501022, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130074, CurrentID=0, Status='live', Symbol='KRT17P8', Aliases='', description='keratin 17 pseudogene 8', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9127730, end_position_on_the_genomic_accession=9128509, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130075, CurrentID=0, Status='live', Symbol='LOC100130075', Aliases='', description='SUZ RNA binding domain containing 1 pseudogene', other_designations='UPF0485 protein C1orf144 homolog', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68804437, end_position_on_the_genomic_accession=68805494, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105372956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372956, CurrentID=0, Status='live', Symbol='LOC105372956', Aliases='', description='uncharacterized LOC105372956', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222479518, end_position_on_the_genomic_accession=222499064, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372957, CurrentID=0, Status='live', Symbol='LOC105372957', Aliases='', description='uncharacterized LOC105372957', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23566970, end_position_on_the_genomic_accession=23573098, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130078, CurrentID=0, Status='live', Symbol='LOC100130078', Aliases='', description='uncharacterized LOC100130078', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=76667337, end_position_on_the_genomic_accession=76667714, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130079, CurrentID=0, Status='live', Symbol='RFPL4AP6', Aliases='', description='ret finger protein-like 4A pseudogene 6', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29433341, end_position_on_the_genomic_accession=29434956, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372960, CurrentID=0, Status='live', Symbol='LOC105372960', Aliases='', description='uncharacterized LOC105372960', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24394214, end_position_on_the_genomic_accession=24408485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646433, CurrentID=0, Status='live', Symbol='RPL32P19', Aliases='RPL32_10_878', description='ribosomal protein L32 pseudogene 19', other_designations='hCG1794003', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=15943068, end_position_on_the_genomic_accession=15943532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372962, CurrentID=0, Status='live', Symbol='LOC105372962', Aliases='', description='uncharacterized LOC105372962', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24700937, end_position_on_the_genomic_accession=24707758, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130083, CurrentID=0, Status='live', Symbol='HBCBP', Aliases='', description='HBcAg-binding protein', other_designations='', map_location='12q13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=56996784, end_position_on_the_genomic_accession=56998440, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105372964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372964, CurrentID=0, Status='live', Symbol='LOC105372964', Aliases='', description='uncharacterized LOC105372964', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25636780, end_position_on_the_genomic_accession=25642226, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372965, CurrentID=0, Status='live', Symbol='LOC105372965', Aliases='', description='uncharacterized LOC105372965', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25481195, end_position_on_the_genomic_accession=25512847, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130086, CurrentID=0, Status='live', Symbol='HSFX2', Aliases='', description='heat shock transcription factor family, X-linked 2', other_designations='heat shock transcription factor family, X linked 2', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149592512, end_position_on_the_genomic_accession=149595310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105372967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372967, CurrentID=0, Status='live', Symbol='LOC105372967', Aliases='', description='uncharacterized LOC105372967', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25419104, end_position_on_the_genomic_accession=25423374, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646440, CurrentID=0, Status='live', Symbol='CCT3P1', Aliases='CCT3-1P', description='chaperonin containing TCP1, subunit 3 (gamma) pseudogene 1', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16275677, end_position_on_the_genomic_accession=16276277, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372969, CurrentID=0, Status='live', Symbol='LOC105372969', Aliases='', description='uncharacterized LOC105372969', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=25280703, end_position_on_the_genomic_accession=25283065, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130090, CurrentID=0, Status='live', Symbol='SFR1P1', Aliases='', description='SFR1 pseudogene 1', other_designations='MEI5 meiotic recombination protein homolog pseudogene|SWI5-dependent recombination repair 1 pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120202059, end_position_on_the_genomic_accession=120202778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372971, CurrentID=0, Status='live', Symbol='LOC105372971', Aliases='', description='uncharacterized LOC105372971', other_designations='CTA-796E4.4', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26161910, end_position_on_the_genomic_accession=26169304, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372972, CurrentID=0, Status='live', Symbol='LOC105372972', Aliases='', description='uncharacterized LOC105372972', other_designations='CTA-796E4.3', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26152151, end_position_on_the_genomic_accession=26164287, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130093, CurrentID=0, Status='live', Symbol='LOC100130093', Aliases='', description='uncharacterized LOC100130093', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227728539, end_position_on_the_genomic_accession=227734377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646446, CurrentID=0, Status='live', Symbol='RPL27P2', Aliases='RPL27_1_105', description='ribosomal protein L27 pseudogene 2', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159759144, end_position_on_the_genomic_accession=159759620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372975, CurrentID=0, Status='live', Symbol='LOC105372975', Aliases='', description='uncharacterized LOC105372975', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26445284, end_position_on_the_genomic_accession=26446409, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372976, CurrentID=0, Status='live', Symbol='LOC105372976', Aliases='', description='uncharacterized LOC105372976', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26795118, end_position_on_the_genomic_accession=26803012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130097, CurrentID=0, Status='live', Symbol='KIF28P', Aliases='KLP-6', description='kinesin family member 28, pseudogene', other_designations='kinesin-like protein family member 6-like', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=246770907, end_position_on_the_genomic_accession=246839587, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  646450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646450, CurrentID=0, Status='live', Symbol='ARIH2OS', Aliases='C3orf71', description='ariadne homolog 2 opposite strand', other_designations='ariadne-2 homolog opposite strand protein', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48917788, end_position_on_the_genomic_accession=48919385, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130099, CurrentID=0, Status='live', Symbol='MIPEPP2', Aliases='', description='mitochondrial intermediate peptidase pseudogene 2', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238777125, end_position_on_the_genomic_accession=238778729, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372980, CurrentID=0, Status='live', Symbol='LOC105372980', Aliases='', description='uncharacterized LOC105372980', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27436318, end_position_on_the_genomic_accession=27441934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372981, CurrentID=0, Status='live', Symbol='LOC105372981', Aliases='', description='uncharacterized LOC105372981', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=27504921, end_position_on_the_genomic_accession=27561485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105372982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372982, CurrentID=0, Status='live', Symbol='LOC105372982', Aliases='', description='uncharacterized LOC105372982', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=28116741, end_position_on_the_genomic_accession=28161375, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372983, CurrentID=0, Status='live', Symbol='LOC105372983', Aliases='', description='uncharacterized LOC105372983', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29032779, end_position_on_the_genomic_accession=29036568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105372984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372984, CurrentID=0, Status='live', Symbol='LOC105372984', Aliases='', description='uncharacterized LOC105372984', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222751251, end_position_on_the_genomic_accession=222773179, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  646457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646457, CurrentID=0, Status='live', Symbol='C19orf67', Aliases='', description='chromosome 19 open reading frame 67', other_designations='UPF0575 protein C19orf67-like', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14081601, end_position_on_the_genomic_accession=14085795, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130106, CurrentID=0, Status='live', Symbol='RPS23P7', Aliases='', description='ribosomal protein S23 pseudogene 7', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=62397366, end_position_on_the_genomic_accession=62398167, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372987, CurrentID=0, Status='live', Symbol='LOC105372987', Aliases='', description='uncharacterized LOC105372987', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29392826, end_position_on_the_genomic_accession=29402914, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130108, CurrentID=0, Status='live', Symbol='ZNF849P', Aliases='', description='zinc finger protein 849, pseudogene', other_designations='zinc finger protein 98 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22684793, end_position_on_the_genomic_accession=22686327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130109, CurrentID=0, Status='live', Symbol='SNRPEP4', Aliases='', description='small nuclear ribonucleoprotein polypeptide E pseudogene 4', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5576486, end_position_on_the_genomic_accession=5576951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130110, CurrentID=0, Status='live', Symbol='ZSWIM5P1', Aliases='', description='zinc finger, SWIM-type containing 5 pseudogene 1', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15768720, end_position_on_the_genomic_accession=15771425, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646463, CurrentID=0, Status='live', Symbol='UBE2HP1', Aliases='UBE2HP', description='ubiquitin conjugating enzyme E2H pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2H (UBC8 homolog, yeast) pseudogene 1|ubiquitin-conjugating enzyme E2H pseudogene 1', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81254885, end_position_on_the_genomic_accession=81255770, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372992, CurrentID=0, Status='live', Symbol='LOC105372992', Aliases='', description='uncharacterized LOC105372992', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30929948, end_position_on_the_genomic_accession=30930633, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372993, CurrentID=0, Status='live', Symbol='LOC105372993', Aliases='', description='uncharacterized LOC105372993', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30889643, end_position_on_the_genomic_accession=30922223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060171, CurrentID=0, Status='live', Symbol='ARMC4P1', Aliases='', description='armadillo repeat containing 4 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27280762, end_position_on_the_genomic_accession=27329145, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646467, CurrentID=0, Status='live', Symbol='CNN2P1', Aliases='', description='calponin 2 pseudogene 1', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30046076, end_position_on_the_genomic_accession=30047242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372996, CurrentID=0, Status='live', Symbol='LOC105372996', Aliases='', description='uncharacterized LOC105372996', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=31150838, end_position_on_the_genomic_accession=31159716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130117, CurrentID=0, Status='live', Symbol='DDX6P2', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box helicase 6 pseudogene 2', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 6 pseudogene 2', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=86275493, end_position_on_the_genomic_accession=86277320, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105372998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372998, CurrentID=0, Status='live', Symbol='LOC105372998', Aliases='', description='uncharacterized LOC105372998', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  122183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122183, CurrentID=0, Status='live', Symbol='PRR20A', Aliases='PRR20', description='proline rich 20A', other_designations='proline rich 20|proline-rich protein 20', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57140918, end_position_on_the_genomic_accession=57143939, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373000, CurrentID=0, Status='live', Symbol='LOC105373000', Aliases='', description='uncharacterized LOC105373000', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32068612, end_position_on_the_genomic_accession=32121748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373001, CurrentID=0, Status='live', Symbol='LOC105373001', Aliases='', description='uncharacterized LOC105373001', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32167432, end_position_on_the_genomic_accession=32190857, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373002, CurrentID=0, Status='live', Symbol='LOC105373002', Aliases='', description='uncharacterized LOC105373002', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32641199, end_position_on_the_genomic_accession=32661215, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373003, CurrentID=0, Status='live', Symbol='LOC105373003', Aliases='', description='uncharacterized LOC105373003', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32587214, end_position_on_the_genomic_accession=32589050, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  253260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253260, CurrentID=0, Status='live', Symbol='RICTOR', Aliases='AVO3, PIA, hAVO3', description='RPTOR independent companion of MTOR, complex 2', other_designations='AVO3 homolog|TORC2-specific protein AVO3|pianissimo', map_location='5p13.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38937920, end_position_on_the_genomic_accession=39074419, orientation='minus', exon_count=40, OMIM=609022, no_hdr0=''),
  105373005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373005, CurrentID=0, Status='live', Symbol='LOC105373005', Aliases='', description='uncharacterized LOC105373005', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33066631, end_position_on_the_genomic_accession=33221758, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373006, CurrentID=0, Status='live', Symbol='LOC105373006', Aliases='', description='uncharacterized LOC105373006', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=33108529, end_position_on_the_genomic_accession=33110638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646479, CurrentID=0, Status='live', Symbol='ADAM24P', Aliases='ADAM24', description='ADAM metallopeptidase domain 24, pseudogene', other_designations='ADAM metallopeptidase domain 29 pseudogene|a disintegrin and metalloproteinase domain 24', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17469518, end_position_on_the_genomic_accession=17471672, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646480, CurrentID=0, Status='live', Symbol='FABP9', Aliases='PERF, PERF15, T-FABP, TLBP', description='fatty acid binding protein 9, testis', other_designations='lipid-binding protein|testis lipid binding protein|testis lipid-binding protein|testis-type fatty acid-binding protein', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81458383, end_position_on_the_genomic_accession=81461523, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373009, CurrentID=0, Status='live', Symbol='LL22NC03-13G6.2', Aliases='', description='uncharacterized LOC105373009', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34192027, end_position_on_the_genomic_accession=34210540, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373010, CurrentID=0, Status='live', Symbol='LL22NC03-86D4.1', Aliases='', description='uncharacterized LOC105373010', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34017445, end_position_on_the_genomic_accession=34218732, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  646483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646483, CurrentID=0, Status='live', Symbol='RPL6P19', Aliases='RPL6_14_853', description='ribosomal protein L6 pseudogene 19', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137721968, end_position_on_the_genomic_accession=137722886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646484, CurrentID=0, Status='live', Symbol='LOC646484', Aliases='', description='uncharacterized LOC646484', other_designations='', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=141418880, end_position_on_the_genomic_accession=141422484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373013, CurrentID=0, Status='live', Symbol='LOC105373013', Aliases='', description='uncharacterized LOC105373013', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=34628599, end_position_on_the_genomic_accession=34665829, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646486, CurrentID=0, Status='live', Symbol='FABP12', Aliases='', description='fatty acid binding protein 12', other_designations='Probable fatty acid-binding protein ENSP00000353650|fatty acid binding protein ORF|intracellular fatty acid-binding protein FABP12', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81524981, end_position_on_the_genomic_accession=81590596, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  122199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122199, CurrentID=0, Status='live', Symbol='LGMNP1', Aliases='LGMN2P', description='legumain pseudogene 1', other_designations='legumain 2 pseudogene', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=64957556, end_position_on_the_genomic_accession=64959518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253272, CurrentID=0, Status='live', Symbol='PCGF7P', Aliases='RNF134P1', description='polycomb group ring finger 7 pseudogene', other_designations='polycomb group ring finger 6 pseudogene|ring finger protein 134 pseudogene 1', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22137924, end_position_on_the_genomic_accession=22140098, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646489, CurrentID=0, Status='live', Symbol='RPL21P70', Aliases='RPL21_4_205', description='ribosomal protein L21 pseudogene 70', other_designations='', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31291182, end_position_on_the_genomic_accession=31291634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373018, CurrentID=0, Status='live', Symbol='LOC105373018', Aliases='', description='uncharacterized LOC105373018', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36461086, end_position_on_the_genomic_accession=36462804, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130139, CurrentID=0, Status='live', Symbol='UBE2D3P2', Aliases='', description='ubiquitin conjugating enzyme E2D 3 pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast) pseudogene 2|ubiquitin-conjugating enzyme E2D 3 pseudogene 2', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37134351, end_position_on_the_genomic_accession=37135164, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373020, CurrentID=0, Status='live', Symbol='LOC105373020', Aliases='', description='uncharacterized LOC105373020', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36496169, end_position_on_the_genomic_accession=36497984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373021, CurrentID=0, Status='live', Symbol='LOC105373021', Aliases='', description='uncharacterized LOC105373021', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36703918, end_position_on_the_genomic_accession=36721203, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373022, CurrentID=0, Status='live', Symbol='LOC105373022', Aliases='', description='uncharacterized LOC105373022', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36888030, end_position_on_the_genomic_accession=36889304, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373023, CurrentID=0, Status='live', Symbol='LOC105373023', Aliases='', description='uncharacterized LOC105373023', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36918472, end_position_on_the_genomic_accession=36925979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378895, CurrentID=0, Status='live', Symbol='LOC105378895', Aliases='', description='uncharacterized LOC105378895', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110031960, end_position_on_the_genomic_accession=110035511, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373025, CurrentID=0, Status='live', Symbol='LOC105373025', Aliases='', description='uncharacterized LOC105373025', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223372811, end_position_on_the_genomic_accession=223393209, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646498, CurrentID=0, Status='live', Symbol='C3orf84', Aliases='', description='chromosome 3 open reading frame 84', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49177539, end_position_on_the_genomic_accession=49192181, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373027, CurrentID=0, Status='live', Symbol='LOC105373027', Aliases='', description='uncharacterized LOC105373027', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38078641, end_position_on_the_genomic_accession=38080841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130148, CurrentID=0, Status='live', Symbol='MAPT-IT1', Aliases='', description='MAPT intronic transcript 1', other_designations='MAPT intronic transcript 1 (non-protein coding)', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45895783, end_position_on_the_genomic_accession=45898798, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373029, CurrentID=0, Status='live', Symbol='LOC105373029', Aliases='', description='uncharacterized LOC105373029', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38434316, end_position_on_the_genomic_accession=38436696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373030, CurrentID=0, Status='live', Symbol='LOC105373030', Aliases='', description='uncharacterized LOC105373030', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38410357, end_position_on_the_genomic_accession=38423930, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373031, CurrentID=0, Status='live', Symbol='LOC105373031', Aliases='', description='uncharacterized LOC105373031', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38570500, end_position_on_the_genomic_accession=38571202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373032, CurrentID=0, Status='live', Symbol='LOC105373032', Aliases='', description='uncharacterized LOC105373032', other_designations='CTA-150C2.13', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38886838, end_position_on_the_genomic_accession=38925336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646505, CurrentID=0, Status='live', Symbol='CLK3P1', Aliases='', description='CDC like kinase 3 pseudogene 1', other_designations='CDC-like kinase 3 pseudogene 1', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19895562, end_position_on_the_genomic_accession=19895982, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646506, CurrentID=0, Status='live', Symbol='LOC646506', Aliases='', description='lysosomal-associated membrane protein 1 pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=33410971, end_position_on_the_genomic_accession=33412170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130155, CurrentID=0, Status='live', Symbol='MIR124-2HG', Aliases='LINC00966', description='MIR124-2 host gene', other_designations='MIR124-2 host gene (non-protein coding)|long intergenic non-protein coding RNA 966', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64373218, end_position_on_the_genomic_accession=64387504, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646508, CurrentID=0, Status='live', Symbol='FAM90A27P', Aliases='', description='family with sequence similarity 90 member A27, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53282290, end_position_on_the_genomic_accession=53284349, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376683, CurrentID=0, Status='live', Symbol='LOC105376683', Aliases='', description='uncharacterized LOC105376683', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=105224974, end_position_on_the_genomic_accession=105233332, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130158, CurrentID=0, Status='live', Symbol='KNOP1P5', Aliases='', description='lysine-rich nucleolar protein 1 pseudogene 5', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126072573, end_position_on_the_genomic_accession=126074247, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373039, CurrentID=0, Status='live', Symbol='LOC105373039', Aliases='', description='uncharacterized LOC105373039', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40660738, end_position_on_the_genomic_accession=40684920, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373040, CurrentID=0, Status='live', Symbol='LOC105373040', Aliases='', description='uncharacterized LOC105373040', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40753074, end_position_on_the_genomic_accession=40758681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646513, CurrentID=0, Status='live', Symbol='LOC646513', Aliases='', description='VLGN1945', other_designations='', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30435550, end_position_on_the_genomic_accession=30436238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373042, CurrentID=0, Status='live', Symbol='LOC105373042', Aliases='', description='uncharacterized LOC105373042', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41370653, end_position_on_the_genomic_accession=41381842, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373043, CurrentID=0, Status='live', Symbol='LOC105373043', Aliases='', description='uncharacterized LOC105373043', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41447858, end_position_on_the_genomic_accession=41448898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373044, CurrentID=0, Status='live', Symbol='LOC105373044', Aliases='', description='uncharacterized LOC105373044', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41551709, end_position_on_the_genomic_accession=41560937, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646517, CurrentID=0, Status='live', Symbol='LOC646517', Aliases='', description='Wilms tumor 1 associated protein pseudogene', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51410084, end_position_on_the_genomic_accession=51410537, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373046, CurrentID=0, Status='live', Symbol='LOC105373046', Aliases='', description='uncharacterized LOC105373046', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223738470, end_position_on_the_genomic_accession=223740411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130167, CurrentID=0, Status='live', Symbol='NUS1P4', Aliases='', description='NUS1 dehydrodolichyl diphosphate synthase subunit pseudogene 4', other_designations='nuclear undecaprenyl pyrophosphate synthase 1 homolog (S. cerevisiae) pseudogene 4', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=98610301, end_position_on_the_genomic_accession=98610891, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646520, CurrentID=0, Status='live', Symbol='SUCLA2P1', Aliases='Em:AB014080.4, SUCLA2P', description='succinate-CoA ligase, ADP-forming, beta subunit pseudogene 1', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30468873, end_position_on_the_genomic_accession=30470922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373049, CurrentID=0, Status='live', Symbol='LOC105373049', Aliases='', description='uncharacterized LOC105373049', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42898515, end_position_on_the_genomic_accession=42907387, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646522, CurrentID=0, Status='live', Symbol='LOC646522', Aliases='', description='uncharacterized LOC646522', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=133783671, end_position_on_the_genomic_accession=133810376, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373051, CurrentID=0, Status='live', Symbol='LOC105373051', Aliases='', description='uncharacterized LOC105373051', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43212599, end_position_on_the_genomic_accession=43213684, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130172, CurrentID=0, Status='live', Symbol='LOC100130172', Aliases='', description='uncharacterized LOC100130172', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138032793, end_position_on_the_genomic_accession=138037066, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373053, CurrentID=0, Status='live', Symbol='LOC105373053', Aliases='', description='uncharacterized LOC105373053', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=43373994, end_position_on_the_genomic_accession=43391396, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646526, CurrentID=0, Status='live', Symbol='HMGB1P18', Aliases='', description='high mobility group box 1 pseudogene 18', other_designations='high-mobility group box 1 pseudogene 18', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=80283352, end_position_on_the_genomic_accession=80283969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646527, CurrentID=0, Status='live', Symbol='RPS3AP44', Aliases='RPS3A_18_1305', description='ribosomal protein S3a pseudogene 44', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26576858, end_position_on_the_genomic_accession=26577721, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373056, CurrentID=0, Status='live', Symbol='LOC105373056', Aliases='', description='uncharacterized LOC105373056', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224095982, end_position_on_the_genomic_accession=224105904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130177, CurrentID=0, Status='live', Symbol='LOC100130177', Aliases='', description='uracil DNA glycosylase pseudogene', other_designations='uracil-DNA glycosylase pseudogene', map_location='5q33.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158188980, end_position_on_the_genomic_accession=158189911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253314, CurrentID=0, Status='live', Symbol='EIF4E1B', Aliases='', description='eukaryotic translation initiation factor 4E family member 1B', other_designations='', map_location='5q35.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176630640, end_position_on_the_genomic_accession=176646641, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  646531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646531, CurrentID=0, Status='live', Symbol='YBX1P2', Aliases='', description='Y-box binding protein 1 pseudogene 2', other_designations='TCAG_1818537|Y box binding protein 1 pseudogene 2', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105581855, end_position_on_the_genomic_accession=105583407, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646532, CurrentID=0, Status='live', Symbol='RPL19P5', Aliases='RPL19_2_325', description='ribosomal protein L19 pseudogene 5', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218354042, end_position_on_the_genomic_accession=218354573, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373061, CurrentID=0, Status='live', Symbol='LOC105373061', Aliases='', description='uncharacterized LOC105373061', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224058866, end_position_on_the_genomic_accession=224073200, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373062, CurrentID=0, Status='live', Symbol='LOC105373062', Aliases='', description='uncharacterized LOC105373062', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44859463, end_position_on_the_genomic_accession=44864239, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130183, CurrentID=0, Status='live', Symbol='RPS24P19', Aliases='RPS24_9_1812', description='ribosomal protein S24 pseudogene 19', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=133670932, end_position_on_the_genomic_accession=133671460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373064, CurrentID=0, Status='live', Symbol='LOC105373064', Aliases='', description='uncharacterized LOC105373064', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373065, CurrentID=0, Status='live', Symbol='LOC105373065', Aliases='', description='uncharacterized LOC105373065', other_designations='CTA-268H5.14', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45262272, end_position_on_the_genomic_accession=45269122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646538, CurrentID=0, Status='live', Symbol='RPL7P10', Aliases='RPL7_2_66', description='ribosomal protein L7 pseudogene 10', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81098195, end_position_on_the_genomic_accession=81099026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130187, CurrentID=0, Status='live', Symbol='TUBB3P2', Aliases='', description='tubulin beta 3 class III pseudogene 2', other_designations='tubulin, beta 3 class III pseudogene 2', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134734942, end_position_on_the_genomic_accession=134735080, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373068, CurrentID=0, Status='live', Symbol='LOC105373068', Aliases='', description='uncharacterized LOC105373068', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45647367, end_position_on_the_genomic_accession=45656679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373069, CurrentID=0, Status='live', Symbol='LOC105373069', Aliases='', description='uncharacterized LOC105373069', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45647284, end_position_on_the_genomic_accession=45657227, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130190, CurrentID=0, Status='live', Symbol='SRRM1P3', Aliases='', description='serine/arginine repetitive matrix 1 pseudogene 3', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136983274, end_position_on_the_genomic_accession=136985751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373071, CurrentID=0, Status='live', Symbol='LOC105373071', Aliases='', description='uncharacterized LOC105373071', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=45882056, end_position_on_the_genomic_accession=45894406, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130192, CurrentID=0, Status='live', Symbol='RPS12P31', Aliases='RPS12_17_1644', description='ribosomal protein S12 pseudogene 31', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32636078, end_position_on_the_genomic_accession=32636475, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130193, CurrentID=0, Status='live', Symbol='LOC100130193', Aliases='', description='uncharacterized LOC100130193', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=19600475, end_position_on_the_genomic_accession=19600873, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  122258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122258, CurrentID=0, Status='live', Symbol='SPACA7', Aliases='C13orf28', description='sperm acrosome associated 7', other_designations='sperm acrosome-associated protein 7', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112375881, end_position_on_the_genomic_accession=112434695, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105373075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373075, CurrentID=0, Status='live', Symbol='LOC105373075', Aliases='', description='uncharacterized LOC105373075', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=46548748, end_position_on_the_genomic_accession=46553360, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130196, CurrentID=0, Status='live', Symbol='HNRNPCP9', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 9', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67642713, end_position_on_the_genomic_accession=67661015, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130197, CurrentID=0, Status='live', Symbol='EFCAB14-AS1', Aliases='KIAA0494-AS1', description='EFCAB14 antisense RNA 1', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46674036, end_position_on_the_genomic_accession=46692098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373078, CurrentID=0, Status='live', Symbol='LOC105373078', Aliases='', description='uncharacterized LOC105373078', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47141247, end_position_on_the_genomic_accession=47144648, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373079, CurrentID=0, Status='live', Symbol='LOC105373079', Aliases='', description='uncharacterized LOC105373079', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=47368815, end_position_on_the_genomic_accession=47373707, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373080, CurrentID=0, Status='live', Symbol='LOC105373080', Aliases='', description='uncharacterized LOC105373080', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48340030, end_position_on_the_genomic_accession=48341748, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373081, CurrentID=0, Status='live', Symbol='LOC105373081', Aliases='', description='uncharacterized LOC105373081', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48351340, end_position_on_the_genomic_accession=48360755, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373082, CurrentID=0, Status='live', Symbol='LOC105373082', Aliases='', description='uncharacterized LOC105373082', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=48466321, end_position_on_the_genomic_accession=48468510, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646555, CurrentID=0, Status='live', Symbol='LOC646555', Aliases='', description='AT-rich interaction domain 3B pseudogene', other_designations='AT rich interactive domain 3B (BRIGHT-like) pseudogene|AT rich interactive domain 3B pseudogene|pseudogene similar to dead ringer-like 2', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=81501585, end_position_on_the_genomic_accession=81505741, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130204, CurrentID=0, Status='live', Symbol='NIPA2P2', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2 pseudogene 2', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75083621, end_position_on_the_genomic_accession=75085261, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646557, CurrentID=0, Status='live', Symbol='OSTCP3', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 3', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53616804, end_position_on_the_genomic_accession=53617252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373086, CurrentID=0, Status='live', Symbol='LOC105373086', Aliases='', description='uncharacterized LOC105373086', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49095656, end_position_on_the_genomic_accession=49098633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130207, CurrentID=0, Status='live', Symbol='LOC100130207', Aliases='', description='uncharacterized LOC100130207', other_designations='', map_location='3p26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3684804, end_position_on_the_genomic_accession=3827421, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373088, CurrentID=0, Status='live', Symbol='LOC105373088', Aliases='', description='uncharacterized LOC105373088', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49534101, end_position_on_the_genomic_accession=49535499, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373089, CurrentID=0, Status='live', Symbol='LOC105373089', Aliases='', description='uncharacterized LOC105373089', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49538002, end_position_on_the_genomic_accession=49541612, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373090, CurrentID=0, Status='live', Symbol='LOC105373090', Aliases='', description='uncharacterized LOC105373090', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49781805, end_position_on_the_genomic_accession=49783435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130211, CurrentID=0, Status='live', Symbol='EEF1A1P29', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 29', other_designations='', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86159926, end_position_on_the_genomic_accession=86161368, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373092, CurrentID=0, Status='live', Symbol='LOC105373092', Aliases='', description='uncharacterized LOC105373092', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=49891815, end_position_on_the_genomic_accession=49897676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373093, CurrentID=0, Status='live', Symbol='LOC105373093', Aliases='', description='uncharacterized LOC105373093', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646566, CurrentID=0, Status='live', Symbol='RPLP1P12', Aliases='RPLP1_6_1670', description='ribosomal protein, large, P1 pseudogene 12', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53775470, end_position_on_the_genomic_accession=53775926, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646567, CurrentID=0, Status='live', Symbol='OSTCP2', Aliases='', description='oligosaccharyltransferase complex subunit pseudogene 2', other_designations='oligosaccharyltransferase complex subunit OSTC-like', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26985362, end_position_on_the_genomic_accession=26986349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373096, CurrentID=0, Status='live', Symbol='LOC105373096', Aliases='', description='uncharacterized LOC105373096', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50318028, end_position_on_the_genomic_accession=50321416, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373097, CurrentID=0, Status='live', Symbol='LOC105373097', Aliases='', description='uncharacterized LOC105373097', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  777642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777642, CurrentID=0, Status='live', Symbol='RP34', Aliases='', description='retinitis pigmentosa 34 (X-linked recessive)', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300605, no_hdr0=''),
  777643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777643, CurrentID=0, Status='live', Symbol='GLC1M', Aliases='', description='glaucoma 1, open angle, M (juvenile-onset)', other_designations='', map_location='5q', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610535, no_hdr0=''),
  646572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646572, CurrentID=0, Status='live', Symbol='CHCHD3P3', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 3 pseudogene 3', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27200836, end_position_on_the_genomic_accession=27201473, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  777645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777645, CurrentID=0, Status='live', Symbol='GLC1N', Aliases='', description='glaucoma 1, open angle, N (juvenile-onset)', other_designations='', map_location='15q22-q24', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=611274, no_hdr0=''),
  105373102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373102, CurrentID=0, Status='live', Symbol='LOC105373102', Aliases='', description='uncharacterized LOC105373102', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=1392349, end_position_on_the_genomic_accession=1396881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646575, CurrentID=0, Status='live', Symbol='CYP4F10P', Aliases='', description='cytochrome P450 family 4 subfamily F member 10, pseudogene', other_designations='cytochrome P450, family 4, subfamily F, polypeptide 10, pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15664106, end_position_on_the_genomic_accession=15667240, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646576, CurrentID=0, Status='live', Symbol='HHIP-AS1', Aliases='', description='HHIP antisense RNA 1', other_designations='HHIP antisense RNA 1 (non-protein coding)', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144642916, end_position_on_the_genomic_accession=144646019, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646577, CurrentID=0, Status='live', Symbol='HMGB1P16', Aliases='', description='high mobility group box 1 pseudogene 16', other_designations='high-mobility group box 1 pseudogene 16', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=36211114, end_position_on_the_genomic_accession=36211732, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  777650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777650, CurrentID=0, Status='live', Symbol='LOC777650', Aliases='', description='UBX domain protein 2B pseudogene', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157193674, end_position_on_the_genomic_accession=157198523, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373107, CurrentID=0, Status='live', Symbol='LOC105373107', Aliases='', description='uncharacterized LOC105373107', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224434719, end_position_on_the_genomic_accession=224437949, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480909, CurrentID=0, Status='live', Symbol='RN7SKP245', Aliases='', description='RNA, 7SK small nuclear pseudogene 245', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646581, CurrentID=0, Status='live', Symbol='IFNWP2', Aliases='', description='interferon, omega 1 pseudogene 2', other_designations='interferon omega pseudogene|interferon, omega 2 (pseudogene)', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21420234, end_position_on_the_genomic_accession=21420813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373110, CurrentID=0, Status='live', Symbol='LOC105373110', Aliases='', description='uncharacterized LOC105373110', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=225447823, end_position_on_the_genomic_accession=225462489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130231, CurrentID=0, Status='live', Symbol='LINC00861', Aliases='', description='long intergenic non-protein coding RNA 861', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125922523, end_position_on_the_genomic_accession=125951197, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373112, CurrentID=0, Status='live', Symbol='LOC105373112', Aliases='', description='uncharacterized LOC105373112', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226067714, end_position_on_the_genomic_accession=226070276, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130233, CurrentID=0, Status='live', Symbol='NPM1P8', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 8', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=32205766, end_position_on_the_genomic_accession=32206935, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130234, CurrentID=0, Status='live', Symbol='EEF1A1P27', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 27', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20582079, end_position_on_the_genomic_accession=20583210, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130235, CurrentID=0, Status='live', Symbol='KATNBL1P2', Aliases='', description='katanin p80 subunit B-like 1 pseudogene 2', other_designations='', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94650552, end_position_on_the_genomic_accession=94651434, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646588, CurrentID=0, Status='live', Symbol='LOC646588', Aliases='', description='uncharacterized LOC646588', other_designations='', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25655209, end_position_on_the_genomic_accession=25670674, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373117, CurrentID=0, Status='live', Symbol='LOC105373117', Aliases='', description='uncharacterized LOC105373117', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226691111, end_position_on_the_genomic_accession=226699896, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130238, CurrentID=0, Status='live', Symbol='LOC100130238', Aliases='', description='uncharacterized LOC100130238', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=132275391, end_position_on_the_genomic_accession=132280900, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130239, CurrentID=0, Status='live', Symbol='UBA52P6', Aliases='RPL40_3_969', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 6', other_designations='ribosomal protein L40 pseudogene', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22011950, end_position_on_the_genomic_accession=22012538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646592, CurrentID=0, Status='live', Symbol='ZNF652P1', Aliases='ZNF652P', description='zinc finger protein 652 pseudogene 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50786255, end_position_on_the_genomic_accession=50788696, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373121, CurrentID=0, Status='live', Symbol='LOC105373121', Aliases='', description='uncharacterized LOC105373121', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227060258, end_position_on_the_genomic_accession=227089636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373122, CurrentID=0, Status='live', Symbol='LOC105373122', Aliases='', description='uncharacterized LOC105373122', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227510329, end_position_on_the_genomic_accession=227520380, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646595, CurrentID=0, Status='live', Symbol='UBE2CP2', Aliases='', description='ubiquitin conjugating enzyme E2C pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2C pseudogene 2', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=22900486, end_position_on_the_genomic_accession=22900995, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373124, CurrentID=0, Status='live', Symbol='LOC105373124', Aliases='', description='uncharacterized LOC105373124', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228508047, end_position_on_the_genomic_accession=228510103, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646597, CurrentID=0, Status='live', Symbol='CHMP1AP1', Aliases='', description='charged multivesicular body protein 1A pseudogene 1', other_designations='chromatin modifying protein 1A pseudogene 1', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27686809, end_position_on_the_genomic_accession=27687440, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373126, CurrentID=0, Status='live', Symbol='LOC105373126', Aliases='', description='uncharacterized LOC105373126', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=9738685, end_position_on_the_genomic_accession=9741053, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130247, CurrentID=0, Status='live', Symbol='COX6CP3', Aliases='', description='cytochrome c oxidase subunit VIc pseudogene 3', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1176650, end_position_on_the_genomic_accession=1177072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646600, CurrentID=0, Status='live', Symbol='IGF2BP2-AS1', Aliases='C3orf65', description='IGF2BP2 antisense RNA 1', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185713252, end_position_on_the_genomic_accession=185729787, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100130249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130249, CurrentID=0, Status='live', Symbol='PP2672', Aliases='', description='uncharacterized LOC100130249', other_designations='', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234757619, end_position_on_the_genomic_accession=234760374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373130, CurrentID=0, Status='live', Symbol='LOC105373130', Aliases='', description='uncharacterized LOC105373130', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=11367770, end_position_on_the_genomic_accession=11370523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646603, CurrentID=0, Status='live', Symbol='C4orf51', Aliases='', description='chromosome 4 open reading frame 51', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145680104, end_position_on_the_genomic_accession=145800392, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105373132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373132, CurrentID=0, Status='live', Symbol='LOC105373132', Aliases='', description='uncharacterized LOC105373132', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228647614, end_position_on_the_genomic_accession=228650007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130253, CurrentID=0, Status='live', Symbol='RCBTB2P1', Aliases='', description='regulator of chromosome condensation (RCC1) and BTB (POZ) domain containing protein 2 pseudogene 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88784171, end_position_on_the_genomic_accession=88787426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373134, CurrentID=0, Status='live', Symbol='LOC105373134', Aliases='', description='uncharacterized LOC105373134', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13155710, end_position_on_the_genomic_accession=13164864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373135, CurrentID=0, Status='live', Symbol='LOC105373135', Aliases='', description='uncharacterized LOC105373135', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13243494, end_position_on_the_genomic_accession=13251172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373136, CurrentID=0, Status='live', Symbol='LOC105373136', Aliases='', description='uncharacterized LOC105373136', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13320800, end_position_on_the_genomic_accession=13334817, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  646609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646609, CurrentID=0, Status='live', Symbol='CLIC4P1', Aliases='', description='chloride intracellular channel 4 pseudogene 1', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22747692, end_position_on_the_genomic_accession=22748231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373138, CurrentID=0, Status='live', Symbol='LOC105373138', Aliases='', description='uncharacterized LOC105373138', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13985730, end_position_on_the_genomic_accession=13996343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130259, CurrentID=0, Status='live', Symbol='SALL4P2', Aliases='', description='spalt-like transcription factor 4 pseudogene 2', other_designations='sal-like 4 pseudogene 2', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129871300, end_position_on_the_genomic_accession=129877570, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130260, CurrentID=0, Status='live', Symbol='EEF1GP2', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 2', other_designations='', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147922182, end_position_on_the_genomic_accession=147923422, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373141, CurrentID=0, Status='live', Symbol='LOC105373141', Aliases='', description='uncharacterized LOC105373141', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=15951822, end_position_on_the_genomic_accession=16091874, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646614, CurrentID=0, Status='live', Symbol='BUB3P1', Aliases='', description='BUB3 mitotic checkpoint protein pseudogene 1', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108993718, end_position_on_the_genomic_accession=108995060, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373143, CurrentID=0, Status='live', Symbol='LOC105373143', Aliases='', description='uncharacterized LOC105373143', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228948712, end_position_on_the_genomic_accession=228965297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373144, CurrentID=0, Status='live', Symbol='LINC01456', Aliases='', description='long intergenic non-protein coding RNA 1456', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=17970167, end_position_on_the_genomic_accession=17982594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130265, CurrentID=0, Status='live', Symbol='ARL2BPP4', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 4', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 4|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 4|ADP-ribosylation factor-like 2 binding protein pseudogene 4', map_location='5q23.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130529401, end_position_on_the_genomic_accession=130530093, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373146, CurrentID=0, Status='live', Symbol='LOC105373146', Aliases='', description='uncharacterized LOC105373146', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21285871, end_position_on_the_genomic_accession=21374119, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646619, CurrentID=0, Status='live', Symbol='ATP5G1P3', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 3', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 3', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138889255, end_position_on_the_genomic_accession=138889957, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646620, CurrentID=0, Status='live', Symbol='RPL3P8', Aliases='RPL3_4_835', description='ribosomal protein L3 pseudogene 8', other_designations='', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=109998379, end_position_on_the_genomic_accession=109999646, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373149, CurrentID=0, Status='live', Symbol='LOC105373149', Aliases='', description='uncharacterized LOC105373149', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26391930, end_position_on_the_genomic_accession=26410110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373150, CurrentID=0, Status='live', Symbol='LOC105373150', Aliases='', description='uncharacterized LOC105373150', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27153617, end_position_on_the_genomic_accession=27399005, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  122335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122335, CurrentID=0, Status='live', Symbol='PPIAP24', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 24', other_designations='peptidylprolyl isomerase A, pseudogene', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106863606, end_position_on_the_genomic_accession=106864328, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646624, CurrentID=0, Status='live', Symbol='RPL27P4', Aliases='RPL27_1_28', description='ribosomal protein L27 pseudogene 4', other_designations='', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28982244, end_position_on_the_genomic_accession=28982644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646625, CurrentID=0, Status='live', Symbol='URAD', Aliases='PRHOXNB', description='ureidoimidazoline (2-oxo-4-hydroxy-4-carboxy-5-) decarboxylase', other_designations='OHCU decarboxylase|parahox cluster neighbor|parahox neighbor', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27978106, end_position_on_the_genomic_accession=27988637, orientation='minus', exon_count=2, OMIM=615804, no_hdr0=''),
  100130274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130274, CurrentID=0, Status='live', Symbol='CCDC166', Aliases='', description='coiled-coil domain containing 166', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143706694, end_position_on_the_genomic_accession=143708109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646627, CurrentID=0, Status='live', Symbol='LYPD8', Aliases='', description='LY6/PLAUR domain containing 8', other_designations='phospholipase inhibitor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248721230, end_position_on_the_genomic_accession=248755792, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105373156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373156, CurrentID=0, Status='live', Symbol='LOC105373156', Aliases='', description='uncharacterized LOC105373156', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=6226361, end_position_on_the_genomic_accession=6230313, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130277, CurrentID=0, Status='live', Symbol='CDC27P2', Aliases='', description='cell division cycle 27 pseudogene 2', other_designations='cell division cycle 27 homolog pseudogene 2|cell division cycle protein 27 homolog', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10190370, end_position_on_the_genomic_accession=10192415, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646630, CurrentID=0, Status='live', Symbol='CHCHD2P8', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 8', other_designations='hCG1790806', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28107595, end_position_on_the_genomic_accession=28107876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130279, CurrentID=0, Status='live', Symbol='HMGB1P44', Aliases='', description='high mobility group box 1 pseudogene 44', other_designations='high-mobility group box 1 pseudogene 44|high-mobility group box 2 pseudogene', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77963940, end_position_on_the_genomic_accession=77964530, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373160, CurrentID=0, Status='live', Symbol='LOC105373160', Aliases='', description='uncharacterized LOC105373160', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229452751, end_position_on_the_genomic_accession=229456753, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373161, CurrentID=0, Status='live', Symbol='LOC105373161', Aliases='', description='uncharacterized LOC105373161', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229778614, end_position_on_the_genomic_accession=229807848, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373162, CurrentID=0, Status='live', Symbol='LOC105373162', Aliases='', description='uncharacterized LOC105373162', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229814067, end_position_on_the_genomic_accession=229866962, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105373163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373163, CurrentID=0, Status='live', Symbol='LOC105373163', Aliases='', description='uncharacterized LOC105373163', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229896720, end_position_on_the_genomic_accession=229900680, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373164, CurrentID=0, Status='live', Symbol='LOC105373164', Aliases='', description='uncharacterized LOC105373164', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230496123, end_position_on_the_genomic_accession=230499024, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130285, CurrentID=0, Status='live', Symbol='LOC100130285', Aliases='', description='uncharacterized LOC100130285', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=640069, end_position_on_the_genomic_accession=641735, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130286, CurrentID=0, Status='live', Symbol='RPL34P11', Aliases='RPL34_6_354', description='ribosomal protein L34 pseudogene 11', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28447725, end_position_on_the_genomic_accession=28448077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373167, CurrentID=0, Status='live', Symbol='LOC105373167', Aliases='', description='uncharacterized LOC105373167', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=230889557, end_position_on_the_genomic_accession=230899570, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646640, CurrentID=0, Status='live', Symbol='TTC39DP', Aliases='', description='tetratricopeptide repeat domain 39D, pseudogene', other_designations='tetratricopeptide repeat domain 39B pseudogene', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38763585, end_position_on_the_genomic_accession=38765200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130289, CurrentID=0, Status='live', Symbol='SNRPGP10', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 10', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205351221, end_position_on_the_genomic_accession=205351607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373170, CurrentID=0, Status='live', Symbol='LOC105373170', Aliases='', description='uncharacterized LOC105373170', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=231811488, end_position_on_the_genomic_accession=231848787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646643, CurrentID=0, Status='live', Symbol='SBK2', Aliases='SGK069', description='SH3 domain binding kinase family member 2', other_designations='SH3 domain binding kinase family, member 2|SH3-binding domain kinase family, member 2|Sugen kinase 69', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55529336, end_position_on_the_genomic_accession=55537174, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373172, CurrentID=0, Status='live', Symbol='LOC105373172', Aliases='', description='uncharacterized LOC105373172', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232043918, end_position_on_the_genomic_accession=232057638, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373173, CurrentID=0, Status='live', Symbol='LOC105373173', Aliases='', description='uncharacterized LOC105373173', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38327851, end_position_on_the_genomic_accession=38344669, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253430, CurrentID=0, Status='live', Symbol='IPMK', Aliases='', description='inositol polyphosphate multikinase', other_designations='inositol 1,3,4,6-tetrakisphosphate 5-kinase', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58191517, end_position_on_the_genomic_accession=58267937, orientation='minus', exon_count=6, OMIM=609851, no_hdr0=''),
  105373175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373175, CurrentID=0, Status='live', Symbol='LOC105373175', Aliases='', description='uncharacterized LOC105373175', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39258101, end_position_on_the_genomic_accession=39299786, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373176, CurrentID=0, Status='live', Symbol='LOC105373176', Aliases='', description='uncharacterized LOC105373176', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39383653, end_position_on_the_genomic_accession=39386773, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373177, CurrentID=0, Status='live', Symbol='LOC105373177', Aliases='', description='uncharacterized LOC105373177', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39774119, end_position_on_the_genomic_accession=39808433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130298, CurrentID=0, Status='live', Symbol='LOC100130298', Aliases='', description='hCG1816373-like', other_designations='', map_location='8q12.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60966121, end_position_on_the_genomic_accession=60967748, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373179, CurrentID=0, Status='live', Symbol='LOC105373179', Aliases='', description='uncharacterized LOC105373179', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=39726792, end_position_on_the_genomic_accession=39849655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130300, CurrentID=0, Status='live', Symbol='RPS27AP13', Aliases='RPS27A_4_904', description='ribosomal protein S27a pseudogene 13', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53887008, end_position_on_the_genomic_accession=53887479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130301, CurrentID=0, Status='live', Symbol='C8orf89', Aliases='', description='chromosome 8 open reading frame 89', other_designations='', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73241364, end_position_on_the_genomic_accession=73285712, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100130302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130302, CurrentID=0, Status='live', Symbol='SUPT20HL1', Aliases='FAM48B1', description='SPT20 homolog, SAGA complex component-like 1', other_designations='family with sequence similarity 48, member B1|protein FAM48B1|suppressor of Ty 20 homolog (S. cerevisiae)-like 1', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=24362761, end_position_on_the_genomic_accession=24365424, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373183, CurrentID=0, Status='live', Symbol='LOC105373183', Aliases='', description='uncharacterized LOC105373183', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40309531, end_position_on_the_genomic_accession=40312544, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373184, CurrentID=0, Status='live', Symbol='LOC105373184', Aliases='', description='uncharacterized LOC105373184', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41276689, end_position_on_the_genomic_accession=41287911, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373185, CurrentID=0, Status='live', Symbol='LOC105373185', Aliases='', description='uncharacterized LOC105373185', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=41275713, end_position_on_the_genomic_accession=41276778, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646658, CurrentID=0, Status='live', Symbol='SYNDIG1L', Aliases='CAPUCIN, DSPC1, IFITMD4, TMEM90A', description='synapse differentiation inducing 1-like', other_designations='caudate- and putamen-enriched sequence|caudate-and putamen-enriched sequence|dispanin subfamily C member 1|interferon induced transmembrane protein domain containing 4|transmembrane protein 90A', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74405893, end_position_on_the_genomic_accession=74437631, orientation='minus', exon_count=5, OMIM=609999, no_hdr0=''),
  100130307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130307, CurrentID=0, Status='live', Symbol='SPDYE19P', Aliases='', description='speedy/RINGO cell cycle regulator family member E19, pseudogene', other_designations='speedy homolog E2 pseudogene|speedy/RINGO cell cycle regulator family member E2 pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5036485, end_position_on_the_genomic_accession=5037885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646660, CurrentID=0, Status='live', Symbol='RPS20P26', Aliases='RPS20_11_1143', description='ribosomal protein S20 pseudogene 26', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47191146, end_position_on_the_genomic_accession=47191656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373189, CurrentID=0, Status='live', Symbol='LOC105373189', Aliases='', description='uncharacterized LOC105373189', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=43855661, end_position_on_the_genomic_accession=43873724, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130310, CurrentID=0, Status='live', Symbol='MSANTD2P1', Aliases='', description='Myb/SANT-like DNA-binding domain containing 2 pseudogene 1', other_designations='', map_location='21q21.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=23101187, end_position_on_the_genomic_accession=23103071, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130311, CurrentID=0, Status='live', Symbol='C17orf107', Aliases='', description='chromosome 17 open reading frame 107', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4899462, end_position_on_the_genomic_accession=4905329, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373192, CurrentID=0, Status='live', Symbol='LOC105373192', Aliases='', description='uncharacterized LOC105373192', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45766092, end_position_on_the_genomic_accession=45771315, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130313, CurrentID=0, Status='live', Symbol='RPS6P23', Aliases='', description='ribosomal protein S6 pseudogene 23', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99173554, end_position_on_the_genomic_accession=99174491, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373194, CurrentID=0, Status='live', Symbol='LOC105373194', Aliases='', description='uncharacterized LOC105373194', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47204140, end_position_on_the_genomic_accession=47218588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373195, CurrentID=0, Status='live', Symbol='LOC105373195', Aliases='', description='uncharacterized LOC105373195', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49139605, end_position_on_the_genomic_accession=49156253, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373196, CurrentID=0, Status='live', Symbol='LOC105373196', Aliases='', description='uncharacterized LOC105373196', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232361502, end_position_on_the_genomic_accession=232370410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373197, CurrentID=0, Status='live', Symbol='LOC105373197', Aliases='', description='uncharacterized LOC105373197', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49307856, end_position_on_the_genomic_accession=49315804, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646670, CurrentID=0, Status='live', Symbol='LOC646670', Aliases='', description='uncharacterized LOC646670', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72637065, end_position_on_the_genomic_accession=72649548, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100130319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130319, CurrentID=0, Status='live', Symbol='TAF9BP2', Aliases='', description='TATA-box binding protein associated factor 9b pseudogene 2', other_designations='TAF9B RNA polymerase II, TATA box binding protein (TBP)-associated factor, 31kDa pseudogene 2', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114809373, end_position_on_the_genomic_accession=114811297, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646672, CurrentID=0, Status='live', Symbol='RPL15P20', Aliases='RPL15_7_1476', description='ribosomal protein L15 pseudogene 20', other_designations='', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15931697, end_position_on_the_genomic_accession=15932329, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373201, CurrentID=0, Status='live', Symbol='LOC105373201', Aliases='', description='uncharacterized LOC105373201', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=232910501, end_position_on_the_genomic_accession=232918421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130322, CurrentID=0, Status='live', Symbol='GAFA2', Aliases='', description='FGF-2 activity-associated protein 2', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373203, CurrentID=0, Status='live', Symbol='LOC105373203', Aliases='', description='uncharacterized LOC105373203', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51159780, end_position_on_the_genomic_accession=51171366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373204, CurrentID=0, Status='live', Symbol='LOC105373204', Aliases='', description='uncharacterized LOC105373204', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51171383, end_position_on_the_genomic_accession=51396466, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  253461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253461, CurrentID=0, Status='live', Symbol='ZBTB38', Aliases='CIBZ, PPP1R171, ZNF921', description='zinc finger and BTB domain containing 38', other_designations='protein phosphatase 1, regulatory subunit 171', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141324213, end_position_on_the_genomic_accession=141449792, orientation='plus', exon_count=14, OMIM=612218, no_hdr0=''),
  105373206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373206, CurrentID=0, Status='live', Symbol='LOC105373206', Aliases='', description='uncharacterized LOC105373206', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234273128, end_position_on_the_genomic_accession=234275220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130327, CurrentID=0, Status='live', Symbol='NDUFA5P5', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A5 pseudogene 5', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5 pseudogene 5|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5, 13kDa pseudogene', map_location='3p11.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88601055, end_position_on_the_genomic_accession=88602368, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373208, CurrentID=0, Status='live', Symbol='LOC105373208', Aliases='', description='uncharacterized LOC105373208', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234678678, end_position_on_the_genomic_accession=234688831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373209, CurrentID=0, Status='live', Symbol='LOC105373209', Aliases='', description='uncharacterized LOC105373209', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234676028, end_position_on_the_genomic_accession=234683176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373210, CurrentID=0, Status='live', Symbol='LOC105373210', Aliases='', description='uncharacterized LOC105373210', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234745770, end_position_on_the_genomic_accession=234748723, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373211, CurrentID=0, Status='live', Symbol='LOC105373211', Aliases='', description='uncharacterized LOC105373211', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234953870, end_position_on_the_genomic_accession=234957397, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373212, CurrentID=0, Status='live', Symbol='LOC105373212', Aliases='', description='uncharacterized LOC105373212', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=235906134, end_position_on_the_genomic_accession=235972738, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646685, CurrentID=0, Status='live', Symbol='RPL23AP41', Aliases='RPL23A_18_409', description='ribosomal protein L23a pseudogene 41', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140902194, end_position_on_the_genomic_accession=140902651, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373214, CurrentID=0, Status='live', Symbol='LOC105373214', Aliases='', description='uncharacterized LOC105373214', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=54528083, end_position_on_the_genomic_accession=54530055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130335, CurrentID=0, Status='live', Symbol='NAP1L1P3', Aliases='', description='nucleosome assembly protein 1-like 1 pseudogene 3', other_designations='', map_location='3q13.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120805866, end_position_on_the_genomic_accession=120806742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646688, CurrentID=0, Status='live', Symbol='RPL4P2', Aliases='RPL4_1_115', description='ribosomal protein L4 pseudogene 2', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166747323, end_position_on_the_genomic_accession=166748749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373217, CurrentID=0, Status='live', Symbol='LOC105373217', Aliases='', description='uncharacterized LOC105373217', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  122402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122402, CurrentID=0, Status='live', Symbol='TDRD9', Aliases='C14orf75, HIG-1, NET54', description='tudor domain containing 9', other_designations='hypoxia-inducible HIG-1|tudor domain-containing protein 9', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103928438, end_position_on_the_genomic_accession=104052667, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  105373219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373219, CurrentID=0, Status='live', Symbol='LOC105373219', Aliases='', description='uncharacterized LOC105373219', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238116367, end_position_on_the_genomic_accession=238125255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373220, CurrentID=0, Status='live', Symbol='LOC105373220', Aliases='', description='uncharacterized LOC105373220', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238323088, end_position_on_the_genomic_accession=238445519, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  646693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646693, CurrentID=0, Status='live', Symbol='CNN2P10', Aliases='', description='calponin 2 pseudogene 10', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166796245, end_position_on_the_genomic_accession=166797338, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130342, CurrentID=0, Status='live', Symbol='SCGB2B3P', Aliases='C2B, SCGB4A3P', description='secretoglobin, family 2B, member 3, pseudogene', other_designations='secretoglobin-like pseudogene', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34645873, end_position_on_the_genomic_accession=34646052, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  151473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151473, CurrentID=0, Status='live', Symbol='SLC16A14', Aliases='MCT14', description='solute carrier family 16 member 14', other_designations='MCT 14|monocarboxylic acid transporter 14|solute carrier family 16 (monocarboxylic acid transporters), member 14|solute carrier family 16, member 14 (monocarboxylic acid transporter 14)', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230034974, end_position_on_the_genomic_accession=230068999, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105373224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373224, CurrentID=0, Status='live', Symbol='LOC105373224', Aliases='', description='uncharacterized LOC105373224', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=239909479, end_position_on_the_genomic_accession=239952987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122409, CurrentID=0, Status='live', Symbol='RPL21P12', Aliases='', description='ribosomal protein L21 pseudogene 12', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103148557, end_position_on_the_genomic_accession=103149438, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646698, CurrentID=0, Status='live', Symbol='ZSCAN5D', Aliases='ZNF495DP, ZSCAN5D', description='zinc finger and SCAN domain containing 5D', other_designations='Putative zinc finger and SCAN domain-containing protein 5D|zinc finger and SCAN domain containing 5D, pseudogene', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56243925, end_position_on_the_genomic_accession=56247467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130347, CurrentID=0, Status='live', Symbol='RPS6P5', Aliases='RPS6_2_487', description='ribosomal protein S6 pseudogene 5', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64767106, end_position_on_the_genomic_accession=64767938, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130348, CurrentID=0, Status='live', Symbol='TMEM262', Aliases='', description='transmembrane protein 262', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65088212, end_position_on_the_genomic_accession=65089990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646701, CurrentID=0, Status='live', Symbol='DPPA5P4', Aliases='', description='developmental pluripotency associated 5 pseudogene 4', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=75292008, end_position_on_the_genomic_accession=75292349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646702, CurrentID=0, Status='live', Symbol='HLA-DPA2', Aliases='HLA-DP2A', description='major histocompatibility complex, class II, DP alpha 2 (pseudogene)', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33091482, end_position_on_the_genomic_accession=33093314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373231, CurrentID=0, Status='live', Symbol='LOC105373231', Aliases='', description='uncharacterized LOC105373231', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241817793, end_position_on_the_genomic_accession=241818765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122416, CurrentID=0, Status='live', Symbol='ANKRD9', Aliases='', description='ankyrin repeat domain 9', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102503649, end_position_on_the_genomic_accession=102509791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373233, CurrentID=0, Status='live', Symbol='LOC105373233', Aliases='', description='uncharacterized LOC105373233', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241585025, end_position_on_the_genomic_accession=241588836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  777778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=777778, CurrentID=0, Status='live', Symbol='OPA6', Aliases='ROA1', description='optic atrophy 6 (autosomal recessive)', other_designations='recessive optic atrophy 1', map_location='8q21.13-q22.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=258500, no_hdr0=''),
  105373235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373235, CurrentID=0, Status='live', Symbol='LOC105373235', Aliases='', description='uncharacterized LOC105373235', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242266626, end_position_on_the_genomic_accession=242269632, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373236, CurrentID=0, Status='live', Symbol='LOC105373236', Aliases='', description='uncharacterized LOC105373236', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242178295, end_position_on_the_genomic_accession=242210406, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100130357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130357, CurrentID=0, Status='live', Symbol='LOC100130357', Aliases='', description='uncharacterized LOC100130357', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13279295, end_position_on_the_genomic_accession=13297297, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646710, CurrentID=0, Status='live', Symbol='RPL19P4', Aliases='RPL19_1_276', description='ribosomal protein L19 pseudogene 4', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130293789, end_position_on_the_genomic_accession=130294479, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130359, CurrentID=0, Status='live', Symbol='MED28P4', Aliases='', description='mediator complex subunit 28 pseudogene 4', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45951602, end_position_on_the_genomic_accession=45952342, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373240, CurrentID=0, Status='live', Symbol='LOC105373240', Aliases='', description='uncharacterized LOC105373240', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66757149, end_position_on_the_genomic_accession=66780371, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130361, CurrentID=0, Status='live', Symbol='CXorf49', Aliases='CXorf49B', description='chromosome X open reading frame 49', other_designations='putative uncharacterized protein CXorf49', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71714374, end_position_on_the_genomic_accession=71718285, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373242, CurrentID=0, Status='live', Symbol='LOC105373242', Aliases='', description='uncharacterized LOC105373242', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=69031633, end_position_on_the_genomic_accession=69083878, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373243, CurrentID=0, Status='live', Symbol='LOC105373243', Aliases='', description='uncharacterized LOC105373243', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70223429, end_position_on_the_genomic_accession=70281252, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373244, CurrentID=0, Status='live', Symbol='LOC105373244', Aliases='', description='uncharacterized LOC105373244', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70427361, end_position_on_the_genomic_accession=70435494, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373245, CurrentID=0, Status='live', Symbol='LOC105373245', Aliases='', description='uncharacterized LOC105373245', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70928227, end_position_on_the_genomic_accession=70947559, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373246, CurrentID=0, Status='live', Symbol='LOC105373246', Aliases='', description='uncharacterized LOC105373246', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71143188, end_position_on_the_genomic_accession=71151226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646719, CurrentID=0, Status='live', Symbol='NIPBL-AS1', Aliases='NIPBL-AS', description='NIPBL antisense RNA 1 (head to head)', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36871361, end_position_on_the_genomic_accession=36876694, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130368, CurrentID=0, Status='live', Symbol='SNRPGP4', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 4', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69318690, end_position_on_the_genomic_accession=69319127, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646721, CurrentID=0, Status='live', Symbol='FAM60DP', Aliases='', description='family with sequence similarity 60 member D, pseudogene', other_designations='family with sequence similarity 60, member A pseudogene', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23634867, end_position_on_the_genomic_accession=23635507, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373250, CurrentID=0, Status='live', Symbol='LOC105373250', Aliases='', description='uncharacterized LOC105373250', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72306058, end_position_on_the_genomic_accession=72311741, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646723, CurrentID=0, Status='live', Symbol='KRT18P31', Aliases='', description='keratin 18 pseudogene 31', other_designations='', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=36885204, end_position_on_the_genomic_accession=36886556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373252, CurrentID=0, Status='live', Symbol='LOC105373252', Aliases='', description='uncharacterized LOC105373252', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74413027, end_position_on_the_genomic_accession=74421393, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130373, CurrentID=0, Status='live', Symbol='LOC100130373', Aliases='', description='uncharacterized LOC100130373', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36150413, end_position_on_the_genomic_accession=36155900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130374, CurrentID=0, Status='live', Symbol='RPL13P14', Aliases='RPL13_6_1708', description='ribosomal protein L13 pseudogene 14', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46863545, end_position_on_the_genomic_accession=46864098, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373255, CurrentID=0, Status='live', Symbol='LOC105373255', Aliases='', description='uncharacterized LOC105373255', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243494789, end_position_on_the_genomic_accession=243497823, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  253512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253512, CurrentID=0, Status='live', Symbol='SLC25A30', Aliases='KMCP1', description='solute carrier family 25 member 30', other_designations='solute carrier family 25, member 30', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45393316, end_position_on_the_genomic_accession=45418455, orientation='minus', exon_count=14, OMIM=610793, no_hdr0=''),
  105373257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373257, CurrentID=0, Status='live', Symbol='LOC105373257', Aliases='', description='uncharacterized LOC105373257', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=243163274, end_position_on_the_genomic_accession=243169614, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646730, CurrentID=0, Status='live', Symbol='LOC646730', Aliases='', description='uncharacterized LOC646730', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141660344, end_position_on_the_genomic_accession=141730357, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373259, CurrentID=0, Status='live', Symbol='LOC105373259', Aliases='', description='uncharacterized LOC105373259', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244066410, end_position_on_the_genomic_accession=244068711, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373260, CurrentID=0, Status='live', Symbol='LOC105373260', Aliases='', description='uncharacterized LOC105373260', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244117990, end_position_on_the_genomic_accession=244126166, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373261, CurrentID=0, Status='live', Symbol='LOC105373261', Aliases='', description='uncharacterized LOC105373261', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244108416, end_position_on_the_genomic_accession=244118489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373262, CurrentID=0, Status='live', Symbol='LOC105373262', Aliases='', description='uncharacterized LOC105373262', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=244230514, end_position_on_the_genomic_accession=244325151, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130383, CurrentID=0, Status='live', Symbol='HIGD1AP12', Aliases='', description='HIG1 hypoxia inducible domain family member 1A pseudogene 12', other_designations='HIG1 hypoxia inducible domain family, member 1A pseudogene 12', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111380270, end_position_on_the_genomic_accession=111380569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130384, CurrentID=0, Status='live', Symbol='CENPUP1', Aliases='', description='centromere protein U pseudogene 1', other_designations='MLF1 interacting protein pseudogene', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13756040, end_position_on_the_genomic_accession=13757564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646737, CurrentID=0, Status='live', Symbol='RPS14P6', Aliases='RPS14_3_529', description='ribosomal protein S14 pseudogene 6', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152613824, end_position_on_the_genomic_accession=152614370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130386, CurrentID=0, Status='live', Symbol='LINC00552', Aliases='', description='long intergenic non-protein coding RNA 552', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113748511, end_position_on_the_genomic_accession=113751089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373267, CurrentID=0, Status='live', Symbol='LOC105373267', Aliases='', description='uncharacterized LOC105373267', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245367478, end_position_on_the_genomic_accession=245371488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130388, CurrentID=0, Status='live', Symbol='EEF1E1P1', Aliases='', description='eukaryotic translation elongation factor 1 epsilon 1 pseudogene 1', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=111887886, end_position_on_the_genomic_accession=111889485, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373269, CurrentID=0, Status='live', Symbol='LOC105373269', Aliases='', description='uncharacterized LOC105373269', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245895209, end_position_on_the_genomic_accession=245931560, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373270, CurrentID=0, Status='live', Symbol='LOC105373270', Aliases='', description='uncharacterized LOC105373270', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=245812590, end_position_on_the_genomic_accession=245813854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130391, CurrentID=0, Status='live', Symbol='YWHAQP5', Aliases='', description='YWHAQ pseudogene 5', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98694115, end_position_on_the_genomic_accession=98695493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130392, CurrentID=0, Status='live', Symbol='RPL9P20', Aliases='RPL9_7_879', description='ribosomal protein L9 pseudogene 20', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17150203, end_position_on_the_genomic_accession=17150909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373273, CurrentID=0, Status='live', Symbol='LOC105373273', Aliases='', description='uncharacterized LOC105373273', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247273429, end_position_on_the_genomic_accession=247288212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373274, CurrentID=0, Status='live', Symbol='LOC105373274', Aliases='', description='uncharacterized LOC105373274', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247700079, end_position_on_the_genomic_accession=247749120, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373275, CurrentID=0, Status='live', Symbol='LOC105373275', Aliases='', description='uncharacterized LOC105373275', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248046934, end_position_on_the_genomic_accession=248095544, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130396, CurrentID=0, Status='live', Symbol='ATP5LP4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit G pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit g, pseudogene 4|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit g, pseudogene 4', map_location='4q31.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148703526, end_position_on_the_genomic_accession=148703832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373277, CurrentID=0, Status='live', Symbol='LOC105373277', Aliases='', description='uncharacterized LOC105373277', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373278, CurrentID=0, Status='live', Symbol='LOC105373278', Aliases='', description='uncharacterized LOC105373278', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373279, CurrentID=0, Status='live', Symbol='LOC105373279', Aliases='', description='uncharacterized LOC105373279', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248468553, end_position_on_the_genomic_accession=248482053, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130400, CurrentID=0, Status='live', Symbol='RPL21P92', Aliases='RPL21_38_1051', description='ribosomal protein L21 pseudogene 92', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=67793729, end_position_on_the_genomic_accession=67794279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646753, CurrentID=0, Status='live', Symbol='RPS26P2', Aliases='RPS26_19_971, bA98O11.1', description='ribosomal protein S26 pseudogene 2', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30831809, end_position_on_the_genomic_accession=30832249, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646754, CurrentID=0, Status='live', Symbol='TRIM64C', Aliases='', description='tripartite motif containing 64C', other_designations='tripartite motif-containing 64C', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49053714, end_position_on_the_genomic_accession=49059112, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646755, CurrentID=0, Status='live', Symbol='MFAP1P1', Aliases='', description='microfibrillar associated protein 1 pseudogene 1', other_designations='microfibrillar-associated protein 1 pseudogene 1', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30541643, end_position_on_the_genomic_accession=30542897, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377108, CurrentID=0, Status='live', Symbol='LOC105377108', Aliases='', description='uncharacterized LOC105377108', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58606360, end_position_on_the_genomic_accession=58613713, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130405, CurrentID=0, Status='live', Symbol='HMGB3P26', Aliases='', description='high mobility group box 3 pseudogene 26', other_designations='high-mobility group box 3 pseudogene 26', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103364215, end_position_on_the_genomic_accession=103364844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646758, CurrentID=0, Status='live', Symbol='PTPN2P2', Aliases='', description='protein tyrosine phosphatase, non-receptor type 2 pseudogene 2', other_designations='protein tyrosine phosphatase, non-receptor type 2 (pseudogene 2)', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=30553666, end_position_on_the_genomic_accession=30555010, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646759, CurrentID=0, Status='live', Symbol='UBTFL7', Aliases='', description='upstream binding transcription factor, RNA polymerase I-like 7 (pseudogene)', other_designations='hCG1990491|upstream binding transcription factor, RNA polymerase I pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49081916, end_position_on_the_genomic_accession=49083097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373288, CurrentID=0, Status='live', Symbol='LOC105373288', Aliases='', description='uncharacterized LOC105373288', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=82769185, end_position_on_the_genomic_accession=82799411, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130409, CurrentID=0, Status='live', Symbol='BTF3P15', Aliases='', description='basic transcription factor 3 pseudogene 15', other_designations='', map_location='10q25.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=109534655, end_position_on_the_genomic_accession=109535129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646762, CurrentID=0, Status='live', Symbol='LOC646762', Aliases='', description='uncharacterized LOC646762', other_designations='', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29645922, end_position_on_the_genomic_accession=29685138, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373291, CurrentID=0, Status='live', Symbol='LOC105373291', Aliases='', description='uncharacterized LOC105373291', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373292, CurrentID=0, Status='live', Symbol='LOC105373292', Aliases='', description='uncharacterized LOC105373292', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=90234591, end_position_on_the_genomic_accession=90265420, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373293, CurrentID=0, Status='live', Symbol='LOC105373293', Aliases='', description='uncharacterized LOC105373293', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91772227, end_position_on_the_genomic_accession=91776815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646766, CurrentID=0, Status='live', Symbol='RPL35P2', Aliases='RPL35_1_683', description='ribosomal protein L35 pseudogene 2', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34263270, end_position_on_the_genomic_accession=34263723, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373295, CurrentID=0, Status='live', Symbol='LOC105373295', Aliases='', description='uncharacterized LOC105373295', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96393024, end_position_on_the_genomic_accession=96402393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373296, CurrentID=0, Status='live', Symbol='LOC105373296', Aliases='', description='uncharacterized LOC105373296', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=96499408, end_position_on_the_genomic_accession=96503136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122481, CurrentID=0, Status='live', Symbol='AK7', Aliases='AK 7', description='adenylate kinase 7', other_designations='ATP-AMP transphosphorylase 7|putative adenylate kinase 7', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96392111, end_position_on_the_genomic_accession=96489427, orientation='plus', exon_count=19, OMIM=615364, no_hdr0=''),
  100130418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130418, CurrentID=0, Status='live', Symbol='CECR7', Aliases='SAHL1', description='cat eye syndrome chromosome region, candidate 7 (non-protein coding)', other_designations='', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17036570, end_position_on_the_genomic_accession=17058792, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373299, CurrentID=0, Status='live', Symbol='LOC105373299', Aliases='', description='uncharacterized LOC105373299', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102141728, end_position_on_the_genomic_accession=102143946, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373300, CurrentID=0, Status='live', Symbol='LOC105373300', Aliases='', description='uncharacterized LOC105373300', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103538191, end_position_on_the_genomic_accession=103585507, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101928894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928894, CurrentID=0, Status='live', Symbol='LOC101928894', Aliases='', description='uncharacterized LOC101928894', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=46171955, end_position_on_the_genomic_accession=46177106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253558, CurrentID=0, Status='live', Symbol='LCLAT1', Aliases='1AGPAT8, AGPAT8, ALCAT1, HSRG1849, LYCAT, UNQ1849', description='lysocardiolipin acyltransferase 1', other_designations='1-AGP acyltransferase 8|1-AGPAT 8|1-acylglycerol-3-phosphate O-acyltransferase 8|acyl-CoA:lysocardiolipin acyltransferase 1', map_location='2p23.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30447226, end_position_on_the_genomic_accession=30644225, orientation='plus', exon_count=10, OMIM=614241, no_hdr0=''),
  253559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253559, CurrentID=0, Status='live', Symbol='CADM2', Aliases='IGSF4D, NECL3, Necl-3, SynCAM 2, synCAM2', description='cell adhesion molecule 2', other_designations='immunoglobulin superfamily member 4D|immunoglobulin superfamily, member 4D|nectin-like 3|nectin-like protein 3|synaptic cell adhesion molecule 2', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=84958982, end_position_on_the_genomic_accession=86074429, orientation='plus', exon_count=13, OMIM=609938, no_hdr0=''),
  105373304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373304, CurrentID=0, Status='live', Symbol='LOC105373304', Aliases='', description='uncharacterized LOC105373304', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=105949716, end_position_on_the_genomic_accession=106007917, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373305, CurrentID=0, Status='live', Symbol='LOC105373305', Aliases='', description='uncharacterized LOC105373305', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=106008172, end_position_on_the_genomic_accession=106009400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646778, CurrentID=0, Status='live', Symbol='LOC646778', Aliases='', description='uncharacterized LOC646778', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.9', start_position_on_the_genomic_accession=29265049, end_position_on_the_genomic_accession=29266701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373307, CurrentID=0, Status='live', Symbol='LOC105373307', Aliases='', description='uncharacterized LOC105373307', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107144177, end_position_on_the_genomic_accession=107145943, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646780, CurrentID=0, Status='live', Symbol='PHKA1P1', Aliases='', description='phosphorylase kinase, alpha 1 pseudogene 1', other_designations='phosphorylase kinase, alpha 1 (muscle) pseudogene', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90892956, end_position_on_the_genomic_accession=90893927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373309, CurrentID=0, Status='live', Symbol='LOC105373309', Aliases='', description='uncharacterized LOC105373309', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107935766, end_position_on_the_genomic_accession=107938625, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373310, CurrentID=0, Status='live', Symbol='LOC105373310', Aliases='', description='uncharacterized LOC105373310', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373311, CurrentID=0, Status='live', Symbol='LOC105373311', Aliases='', description='uncharacterized LOC105373311', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109730965, end_position_on_the_genomic_accession=109734332, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  646784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646784, CurrentID=0, Status='live', Symbol='RPL5P6', Aliases='RPL5_2_71', description='ribosomal protein L5 pseudogene 6', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91023875, end_position_on_the_genomic_accession=91024867, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646785, CurrentID=0, Status='live', Symbol='RPS10P13', Aliases='RPS10_8_685', description='ribosomal protein S10 pseudogene 13', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34744140, end_position_on_the_genomic_accession=34744718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373314, CurrentID=0, Status='live', Symbol='LOC105373314', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like protein 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=115454266, end_position_on_the_genomic_accession=115465479, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646787, CurrentID=0, Status='live', Symbol='RPL31P37', Aliases='RPL31_18_839', description='ribosomal protein L31 pseudogene 37', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122234508, end_position_on_the_genomic_accession=122234948, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373316, CurrentID=0, Status='live', Symbol='LOC105373316', Aliases='', description='collagen alpha-4(IV) chain-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  253573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253573, CurrentID=0, Status='live', Symbol='LOC253573', Aliases='', description='uncharacterized LOC253573', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186454981, end_position_on_the_genomic_accession=186493661, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373318, CurrentID=0, Status='live', Symbol='LOC105373318', Aliases='', description='uncharacterized LOC105373318', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248931413, end_position_on_the_genomic_accession=248937105, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646791, CurrentID=0, Status='live', Symbol='ANP32BP1', Aliases='', description='acidic nuclear phosphoprotein 32 family member B pseudogene 1', other_designations='acidic (leucine-rich) nuclear phosphoprotein 32 family, member B pseudogene 1', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75321582, end_position_on_the_genomic_accession=75322706, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373320, CurrentID=0, Status='live', Symbol='LOC105373320', Aliases='', description='uncharacterized LOC105373320', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116726936, end_position_on_the_genomic_accession=116751276, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373321, CurrentID=0, Status='live', Symbol='LOC105373321', Aliases='', description='uncharacterized LOC105373321', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118556489, end_position_on_the_genomic_accession=118557504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373322, CurrentID=0, Status='live', Symbol='LOC105373322', Aliases='', description='uncharacterized LOC105373322', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118698546, end_position_on_the_genomic_accession=118703478, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373323, CurrentID=0, Status='live', Symbol='LOC105373323', Aliases='', description='uncharacterized LOC105373323', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=119990253, end_position_on_the_genomic_accession=119993869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373324, CurrentID=0, Status='live', Symbol='LOC105373324', Aliases='', description='uncharacterized LOC105373324', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162724, end_position_on_the_genomic_accession=190544, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  122509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122509, CurrentID=0, Status='live', Symbol='IFI27L1', Aliases='FAM14B, ISG12C', description='interferon, alpha-inducible protein 27-like 1', other_designations='ISG12(c) protein|family with sequence similarity 14, member B|interferon-stimulated gene 12c protein', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94081293, end_position_on_the_genomic_accession=94102714, orientation='plus', exon_count=7, OMIM=611320, no_hdr0=''),
  253582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253582, CurrentID=0, Status='live', Symbol='TMEM244', Aliases='C6orf191, bA174C7.4', description='transmembrane protein 244', other_designations='putative transmembrane protein C6orf191', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129831244, end_position_on_the_genomic_accession=129861503, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  646799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646799, CurrentID=0, Status='live', Symbol='ZAR1L', Aliases='Z3CXXC7, ZAR2', description='zygote arrest 1-like', other_designations='zinc finger, 3CxxC-type 7', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=32303719, end_position_on_the_genomic_accession=32312582, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130448, CurrentID=0, Status='live', Symbol='MRS2P1', Aliases='MRS2LP1', description='MRS2 pseudogene 1', other_designations='MRS2 magnesium homeostasis factor homolog pseudogene 1|MRS2-like, magnesium homeostasis factor pseudogene 1', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124920430, end_position_on_the_genomic_accession=124921974, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130449, CurrentID=0, Status='live', Symbol='PP14571', Aliases='', description='uncharacterized LOC100130449', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240449419, end_position_on_the_genomic_accession=240456700, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  646802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646802, CurrentID=0, Status='live', Symbol='CYP4F62P', Aliases='CYP4F43P', description='cytochrome P450 family 4 subfamily F member 62, pseudogene', other_designations='CYP4F-se11[6:7:8]|Uncharacterized protein C2orf14-like 1|cytochrome P450, family 4, subfamily F, polypeptide 43, pseudogene|cytochrome P450, family 4, subfamily F, polypeptide 62, pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130435662, end_position_on_the_genomic_accession=130441681, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373331, CurrentID=0, Status='live', Symbol='LOC105373331', Aliases='', description='uncharacterized LOC105373331', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124380907, end_position_on_the_genomic_accession=124389830, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373332, CurrentID=0, Status='live', Symbol='LOC105373332', Aliases='', description='uncharacterized LOC105373332', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=127492645, end_position_on_the_genomic_accession=127499469, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373333, CurrentID=0, Status='live', Symbol='LOC105373333', Aliases='', description='uncharacterized LOC105373333', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=127939302, end_position_on_the_genomic_accession=127945639, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373334, CurrentID=0, Status='live', Symbol='LOC105373334', Aliases='', description='uncharacterized LOC105373334', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129675888, end_position_on_the_genomic_accession=129678412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373335, CurrentID=0, Status='live', Symbol='LOC105373335', Aliases='', description='uncharacterized LOC105373335', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=129882292, end_position_on_the_genomic_accession=129934390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  646808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646808, CurrentID=0, Status='live', Symbol='LAGE3P1', Aliases='', description='L antigen family member 3 pseudogene 1', other_designations='L antigen family, member 3 pseudogene 1', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33019705, end_position_on_the_genomic_accession=33020132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  91370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=91370, CurrentID=0, Status='live', Symbol='MGC20647', Aliases='', description='uncharacterized protein MGC20647', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100130458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130458, CurrentID=0, Status='live', Symbol='LOC100130458', Aliases='', description='uncharacterized LOC100130458', other_designations='', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37036130, end_position_on_the_genomic_accession=37038869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130459, CurrentID=0, Status='live', Symbol='LOC100130459', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4029056, end_position_on_the_genomic_accession=4107936, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130460, CurrentID=0, Status='live', Symbol='CAND1.11', Aliases='', description='uncharacterized LOC100130460', other_designations='', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10308313, end_position_on_the_genomic_accession=10430673, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  122525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122525, CurrentID=0, Status='live', Symbol='C14orf28', Aliases='DRIP-1, DRIP1, c14_5270', description='chromosome 14 open reading frame 28', other_designations='dopamine receptor interacting protein 1|dopamine receptor-interacting protein 1', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=44897304, end_position_on_the_genomic_accession=44907257, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130462, CurrentID=0, Status='live', Symbol='HNF4GP1', Aliases='', description='hepatocyte nuclear factor 4, gamma pseudogene 1', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=55996489, end_position_on_the_genomic_accession=56000450, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373343, CurrentID=0, Status='live', Symbol='LOC105373343', Aliases='', description='uncharacterized LOC105373343', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140216026, end_position_on_the_genomic_accession=140216767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373344, CurrentID=0, Status='live', Symbol='LOC105373344', Aliases='', description='uncharacterized LOC105373344', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140517763, end_position_on_the_genomic_accession=140520395, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646817, CurrentID=0, Status='live', Symbol='SETSIP', Aliases='SETP18', description='SET-like protein', other_designations='SET nuclear oncogene pseudogene|SET pseudogene 18|SET pseudogene protein 18|SET similar protein', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92074436, end_position_on_the_genomic_accession=92075472, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130466, CurrentID=0, Status='live', Symbol='NUCKS1P1', Aliases='', description='nuclear casein kinase and cyclin-dependent kinase substrate 1 pseudogene 1', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227410173, end_position_on_the_genomic_accession=227423782, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646819, CurrentID=0, Status='live', Symbol='RPS15AP19', Aliases='RPS15A_5_686', description='ribosomal protein S15a pseudogene 19', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35555846, end_position_on_the_genomic_accession=35556304, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646820, CurrentID=0, Status='live', Symbol='RPL19P19', Aliases='RPL19_8_1674', description='ribosomal protein L19 pseudogene 19', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57953087, end_position_on_the_genomic_accession=57953795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  282566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282566, CurrentID=0, Status='live', Symbol='LINC00515', Aliases='C21orf71, PRED21', description='long intergenic non-protein coding RNA 515', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=25582775, end_position_on_the_genomic_accession=25583224, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373350, CurrentID=0, Status='live', Symbol='LOC105373350', Aliases='', description='uncharacterized LOC105373350', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=316255, end_position_on_the_genomic_accession=336807, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373351, CurrentID=0, Status='live', Symbol='LOC105373351', Aliases='', description='uncharacterized LOC105373351', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=367535, end_position_on_the_genomic_accession=387147, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373352, CurrentID=0, Status='live', Symbol='LOC105373352', Aliases='', description='uncharacterized LOC105373352', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=557813, end_position_on_the_genomic_accession=561982, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373353, CurrentID=0, Status='live', Symbol='LOC105373353', Aliases='', description='uncharacterized LOC105373353', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=635637, end_position_on_the_genomic_accession=646098, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373354, CurrentID=0, Status='live', Symbol='LOC105373354', Aliases='', description='uncharacterized LOC105373354', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=677387, end_position_on_the_genomic_accession=686799, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373355, CurrentID=0, Status='live', Symbol='LOC105373355', Aliases='', description='uncharacterized LOC105373355', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646828, CurrentID=0, Status='live', Symbol='LOC646828', Aliases='', description='carbonic anhydrase VA, mitochondrial pseudogene', other_designations='mitochondrial carbonic anhydrase VA pseudogene', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21532885, end_position_on_the_genomic_accession=21550376, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373357, CurrentID=0, Status='live', Symbol='LOC105373357', Aliases='', description='uncharacterized LOC105373357', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373358, CurrentID=0, Status='live', Symbol='LOC105373358', Aliases='', description='uncharacterized LOC105373358', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=697224, end_position_on_the_genomic_accession=715523, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373359, CurrentID=0, Status='live', Symbol='LOC105373359', Aliases='', description='uncharacterized LOC105373359', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1309202, end_position_on_the_genomic_accession=1314449, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130480, CurrentID=0, Status='live', Symbol='LINC01387', Aliases='C18orf64', description='long intergenic non-protein coding RNA 1387', other_designations='putative uncharacterized protein C18orf64|putative uncharacterized protein LOC100130480', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=6511416, end_position_on_the_genomic_accession=6590653, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130481, CurrentID=0, Status='live', Symbol='CCT6P4', Aliases='CCT6-4P', description='chaperonin containing TCP1, subunit 6 (zeta) pseudogene 4', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190714708, end_position_on_the_genomic_accession=190717377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373362, CurrentID=0, Status='live', Symbol='LOC105373362', Aliases='', description='uncharacterized LOC105373362', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1414509, end_position_on_the_genomic_accession=1417633, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373363, CurrentID=0, Status='live', Symbol='LOC105373363', Aliases='', description='uncharacterized LOC105373363', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1543550, end_position_on_the_genomic_accession=1556582, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130484, CurrentID=0, Status='live', Symbol='UBFD1P1', Aliases='', description='ubiquitin family domain containing 1 pseudogene 1', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98433964, end_position_on_the_genomic_accession=98434844, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646837, CurrentID=0, Status='live', Symbol='EEF1GP1', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 1', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125033433, end_position_on_the_genomic_accession=125035389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373366, CurrentID=0, Status='live', Symbol='IDS2', Aliases='', description='uncharacterized LOC105373366', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149511554, end_position_on_the_genomic_accession=149526523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646839, CurrentID=0, Status='live', Symbol='TPM3P8', Aliases='', description='tropomyosin 3 pseudogene 8', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230571895, end_position_on_the_genomic_accession=230573956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122552, CurrentID=0, Status='live', Symbol='PPIAP4', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 4', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39108195, end_position_on_the_genomic_accession=39108928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  122553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122553, CurrentID=0, Status='live', Symbol='TRAPPC6B', Aliases='TPC6', description='trafficking protein particle complex 6B', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39147811, end_position_on_the_genomic_accession=39170430, orientation='minus', exon_count=7, OMIM=610397, no_hdr0=''),
  105373370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373370, CurrentID=0, Status='live', Symbol='LOC105373370', Aliases='', description='uncharacterized LOC105373370', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152554383, end_position_on_the_genomic_accession=152698714, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373371, CurrentID=0, Status='live', Symbol='LOC105373371', Aliases='', description='uncharacterized LOC105373371', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1801806, end_position_on_the_genomic_accession=1803582, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373372, CurrentID=0, Status='live', Symbol='LOC105373372', Aliases='', description='uncharacterized LOC105373372', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152930263, end_position_on_the_genomic_accession=152931310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373373, CurrentID=0, Status='live', Symbol='LOC105373373', Aliases='', description='uncharacterized LOC105373373', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153225649, end_position_on_the_genomic_accession=153230357, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373374, CurrentID=0, Status='live', Symbol='LOC105373374', Aliases='', description='uncharacterized LOC105373374', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153084383, end_position_on_the_genomic_accession=153092263, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130495, CurrentID=0, Status='live', Symbol='RPS12P26', Aliases='RPS12_12_1461', description='ribosomal protein S12 pseudogene 26', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90206482, end_position_on_the_genomic_accession=90206995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373376, CurrentID=0, Status='live', Symbol='LOC105373376', Aliases='', description='uncharacterized LOC105373376', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  646849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646849, CurrentID=0, Status='live', Symbol='NPM1P28', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 28', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=147590992, end_position_on_the_genomic_accession=147591835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373378, CurrentID=0, Status='live', Symbol='LOC105373378', Aliases='', description='uncharacterized LOC105373378', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153355529, end_position_on_the_genomic_accession=153366474, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  646851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646851, CurrentID=0, Status='live', Symbol='FAM227A', Aliases='', description='family with sequence similarity 227 member A', other_designations='Protein FAM227A|family with sequence similarity 227, member A', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=38578120, end_position_on_the_genomic_accession=38656629, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105373380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373380, CurrentID=0, Status='live', Symbol='LOC105373380', Aliases='', description='uncharacterized LOC105373380', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=1798741, end_position_on_the_genomic_accession=1811738, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373381, CurrentID=0, Status='live', Symbol='LOC105373381', Aliases='', description='extracellular matrix protein 2-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153484796, end_position_on_the_genomic_accession=153487088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373382, CurrentID=0, Status='live', Symbol='LOC105373382', Aliases='', description='SH3 domain-containing protein C23A1.17-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153477156, end_position_on_the_genomic_accession=153483800, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  253639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253639, CurrentID=0, Status='live', Symbol='ZNF620', Aliases='', description='zinc finger protein 620', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40504989, end_position_on_the_genomic_accession=40518221, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373384, CurrentID=0, Status='live', Symbol='LOC105373384', Aliases='', description='MORF4 family-associated protein 1-like', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153606095, end_position_on_the_genomic_accession=153608033, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373385, CurrentID=0, Status='live', Symbol='LOC105373385', Aliases='', description='uncharacterized LOC105373385', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153841578, end_position_on_the_genomic_accession=153859194, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373386, CurrentID=0, Status='live', Symbol='LOC105373386', Aliases='', description='formin-like protein 5', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154333979, end_position_on_the_genomic_accession=154342742, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130507, CurrentID=0, Status='live', Symbol='LOC100130507', Aliases='', description='uncharacterized LOC100130507', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=134205136, end_position_on_the_genomic_accession=134208984, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373388, CurrentID=0, Status='live', Symbol='LOC105373388', Aliases='', description='uncharacterized LOC105373388', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373389, CurrentID=0, Status='live', Symbol='LOC105373389', Aliases='', description='uncharacterized LOC105373389', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=2484408, end_position_on_the_genomic_accession=2495203, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646862, CurrentID=0, Status='live', Symbol='RNF225', Aliases='', description='ring finger protein 225', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58395657, end_position_on_the_genomic_accession=58398954, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373391, CurrentID=0, Status='live', Symbol='LOC105373391', Aliases='', description='uncharacterized LOC105373391', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=3128061, end_position_on_the_genomic_accession=3129950, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646864, CurrentID=0, Status='live', Symbol='ZNF723P', Aliases='ZNF723', description='zinc finger protein 723, pseudogene', other_designations='zinc finger protein 92 pseudogene', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22832239, end_position_on_the_genomic_accession=22858484, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  285381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=285381, CurrentID=0, Status='live', Symbol='DPH3', Aliases='DELGIP, DELGIP1, DESR1A, KTI11, ZCSL2, DPH3', description='diphthamide biosynthesis 3', other_designations='CSL-type zinc finger-containing protein 2|DPH3 homolog (KTI11, S. cerevisiae)|DPH3, KTI11 homolog|DPH3A, KTI11 homolog A|Del-GEF interacting protein|deafness locus putative guanine nucleotide exchange factor interacting protein 1|delGEF-interacting protein 1|diphtheria toxin and Pseudomonas exotoxin A sensitivity required gene 1|zinc finger, CSL domain containing 2|zinc finger, CSL-type containing 2', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16257061, end_position_on_the_genomic_accession=16264989, orientation='minus', exon_count=3, OMIM=608959, no_hdr0=''),
  253650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253650, CurrentID=0, Status='live', Symbol='ANKRD18A', Aliases='', description='ankyrin repeat domain 18A', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38542965, end_position_on_the_genomic_accession=38620655, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  105373395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373395, CurrentID=0, Status='live', Symbol='LOC105373395', Aliases='', description='uncharacterized LOC105373395', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=4020896, end_position_on_the_genomic_accession=4025307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373396, CurrentID=0, Status='live', Symbol='LOC105373396', Aliases='', description='uncharacterized LOC105373396', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=4323242, end_position_on_the_genomic_accession=4330011, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373397, CurrentID=0, Status='live', Symbol='LOC105373397', Aliases='', description='uncharacterized LOC105373397', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=4096389, end_position_on_the_genomic_accession=4140262, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373398, CurrentID=0, Status='live', Symbol='LOC105373398', Aliases='', description='uncharacterized LOC105373398', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=5273718, end_position_on_the_genomic_accession=5295080, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130519, CurrentID=0, Status='live', Symbol='TMEM221', Aliases='', description='transmembrane protein 221', other_designations='Putative transmembrane protein ENSP00000342162', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17435509, end_position_on_the_genomic_accession=17448660, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130520, CurrentID=0, Status='live', Symbol='LOC100130520', Aliases='CD300H, CD300Hs', description='CD300c molecule-like', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74557246, end_position_on_the_genomic_accession=74567534, orientation='minus', exon_count=5, OMIM=616560, no_hdr0=''),
  646873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646873, CurrentID=0, Status='live', Symbol='PRELID3BP10', Aliases='', description='PRELI domain containing 3B pseudogene 10', other_designations='slowmo homolog 2 (Drosophila) pseudogene', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127295620, end_position_on_the_genomic_accession=127296200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130522, CurrentID=0, Status='live', Symbol='PARD6G-AS1', Aliases='', description='PARD6G antisense RNA 1', other_designations='PARD6G antisense RNA 1 (non-protein coding)', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=80147924, end_position_on_the_genomic_accession=80178432, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130523, CurrentID=0, Status='live', Symbol='RPL5P8', Aliases='RPL5_3_323', description='ribosomal protein L5 pseudogene 8', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=214280589, end_position_on_the_genomic_accession=214281524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130524, CurrentID=0, Status='live', Symbol='RPL6P23', Aliases='', description='ribosomal protein L6 pseudogene', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94201883, end_position_on_the_genomic_accession=94203072, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122589, CurrentID=0, Status='live', Symbol='RPLP0P3', Aliases='RPLP0_4_1355', description='ribosomal protein, large, P0 pseudogene 3', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35388055, end_position_on_the_genomic_accession=35389146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373406, CurrentID=0, Status='live', Symbol='LOC105373406', Aliases='', description='uncharacterized LOC105373406', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7382669, end_position_on_the_genomic_accession=7387037, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646879, CurrentID=0, Status='live', Symbol='ARL2BPP3', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 3', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 3|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 3|ADP-ribosylation factor-like 2 binding protein pseudogene 3', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36794484, end_position_on_the_genomic_accession=36795050, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130528, CurrentID=0, Status='live', Symbol='SUCLG2P4', Aliases='', description='succinate-CoA ligase, GDP-forming, beta subunit pseudogene 4', other_designations='succinyl-CoA ligase [GDP-forming] subunit beta, mitochondrial pseudogene', map_location='5p13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29001743, end_position_on_the_genomic_accession=29004712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253665, CurrentID=0, Status='live', Symbol='TCEB2P2', Aliases='', description='transcription elongation factor B (SIII), polypeptide 2 (18kDa, elongin B) pseudogene 2', other_designations='ELOB pseudogene|RNA polymerase II transcription factor SIII subunit B pseudogene|SIII p18 pseudogene|elongin 18 kDa subunit pseudogene|elongin B pseudogene|rig-1 protein - mouse|transcription elongation factor B (SIII), polypeptide 2 (18kD, elongin B) pseudogene|transcription elongation factor B polypeptide 2 pseudogene', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130032868, end_position_on_the_genomic_accession=130033277, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373410, CurrentID=0, Status='live', Symbol='LINC00299', Aliases='', description='long intergenic non-protein coding RNA 299', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8357374, end_position_on_the_genomic_accession=8362275, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373411, CurrentID=0, Status='live', Symbol='LOC105373411', Aliases='', description='uncharacterized LOC105373411', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8598896, end_position_on_the_genomic_accession=8622855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646884, CurrentID=0, Status='live', Symbol='RPL29P28', Aliases='RPL29_17_1307', description='ribosomal protein L29 pseudogene 28', other_designations='', map_location='13q13.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36926438, end_position_on_the_genomic_accession=36926831, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373413, CurrentID=0, Status='live', Symbol='LOC105373413', Aliases='', description='uncharacterized LOC105373413', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9004077, end_position_on_the_genomic_accession=9046176, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373414, CurrentID=0, Status='live', Symbol='LOC105373414', Aliases='', description='uncharacterized LOC105373414', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9103858, end_position_on_the_genomic_accession=9109859, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130535, CurrentID=0, Status='live', Symbol='VIM2P', Aliases='VIM2, VIMP1, lncRNA-CIR', description='vimentin 2, pseudogene (functional)', other_designations='RP11-162L10.1|long non-coding RNA, cartilage injury-related|vimentin pseudogene 1 (functional)', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126602355, end_position_on_the_genomic_accession=126603651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373416, CurrentID=0, Status='live', Symbol='LOC105373416', Aliases='', description='uncharacterized LOC105373416', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9110423, end_position_on_the_genomic_accession=9117032, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130537, CurrentID=0, Status='live', Symbol='RPL31P26', Aliases='RPL31_7_527', description='ribosomal protein L31 pseudogene 26', other_designations='', map_location='4q31.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147425930, end_position_on_the_genomic_accession=147426363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373418, CurrentID=0, Status='live', Symbol='LOC105373418', Aliases='', description='uncharacterized LOC105373418', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9638737, end_position_on_the_genomic_accession=9641402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130539, CurrentID=0, Status='live', Symbol='C10orf142', Aliases='', description='chromosome 10 open reading frame 142', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44292750, end_position_on_the_genomic_accession=44294649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646892, CurrentID=0, Status='live', Symbol='SH2D7', Aliases='', description='SH2 domain containing 7', other_designations='Putative SH2 domain-containing protein LOC646892|hCG38941', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78077805, end_position_on_the_genomic_accession=78104051, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105373421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373421, CurrentID=0, Status='live', Symbol='LOC105373421', Aliases='', description='uncharacterized LOC105373421', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9805310, end_position_on_the_genomic_accession=9807483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373422, CurrentID=0, Status='live', Symbol='LOC105373422', Aliases='', description='uncharacterized LOC105373422', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=9939097, end_position_on_the_genomic_accession=9950829, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373423, CurrentID=0, Status='live', Symbol='LOC105373423', Aliases='', description='uncharacterized LOC105373423', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10030593, end_position_on_the_genomic_accession=10040346, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373424, CurrentID=0, Status='live', Symbol='LOC105373424', Aliases='', description='uncharacterized LOC105373424', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10198596, end_position_on_the_genomic_accession=10202060, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373425, CurrentID=0, Status='live', Symbol='LOC105373425', Aliases='', description='uncharacterized LOC105373425', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=10406377, end_position_on_the_genomic_accession=10412513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  646898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646898, CurrentID=0, Status='live', Symbol='RPL12P40', Aliases='RPL12_19_1596', description='ribosomal protein L12 pseudogene 40', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37672695, end_position_on_the_genomic_accession=37673173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130547, CurrentID=0, Status='live', Symbol='LOC100130547', Aliases='', description='uncharacterized LOC100130547', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=140118567, end_position_on_the_genomic_accession=140122265, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130548, CurrentID=0, Status='live', Symbol='LOC100130548', Aliases='', description='uncharacterized LOC100130548', other_designations='', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134054290, end_position_on_the_genomic_accession=134059139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130549, CurrentID=0, Status='live', Symbol='ISCA1P6', Aliases='', description='iron-sulfur cluster assembly 1 pseudogene 6', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 6', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128517289, end_position_on_the_genomic_accession=128519258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373430, CurrentID=0, Status='live', Symbol='LOC105373430', Aliases='', description='uncharacterized LOC105373430', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11832219, end_position_on_the_genomic_accession=11866277, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  646903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646903, CurrentID=0, Status='live', Symbol='LOC646903', Aliases='', description='uncharacterized LOC646903', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149971279, end_position_on_the_genomic_accession=149973242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  122616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122616, CurrentID=0, Status='live', Symbol='C14orf79', Aliases='', description='chromosome 14 open reading frame 79', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104986279, end_position_on_the_genomic_accession=105000307, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373433, CurrentID=0, Status='live', Symbol='LOC105373433', Aliases='', description='uncharacterized LOC105373433', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=12104599, end_position_on_the_genomic_accession=12108303, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122618, CurrentID=0, Status='live', Symbol='PLD4', Aliases='C14orf175', description='phospholipase D family member 4', other_designations='PLD 4|choline phosphatase 4|phosphatidylcholine-hydrolyzing phospholipase D4|phospholipase D family, member 4', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104924821, end_position_on_the_genomic_accession=104937788, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100130555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130555, CurrentID=0, Status='live', Symbol='RPL23AP68', Aliases='RPL23A_26_1264', description='ribosomal protein L23a pseudogene 68', other_designations='', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87170036, end_position_on_the_genomic_accession=87180833, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373436, CurrentID=0, Status='live', Symbol='LOC105373436', Aliases='', description='uncharacterized LOC105373436', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13277937, end_position_on_the_genomic_accession=13331663, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130557, CurrentID=0, Status='live', Symbol='NFYC-AS1', Aliases='0808y08y', description='NFYC antisense RNA 1', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40689080, end_position_on_the_genomic_accession=40692261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  122622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122622, CurrentID=0, Status='live', Symbol='ADSSL1', Aliases='', description='adenylosuccinate synthase like 1', other_designations='AMPSase 1|IMP--aspartate ligase 1|M-type adenylosuccinate synthetase|adSS 1|adenylosuccinate synthetase, basic isozyme|adenylosuccinate synthetase, muscle isozyme', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104724197, end_position_on_the_genomic_accession=104747325, orientation='plus', exon_count=16, OMIM=612498, no_hdr0=''),
  100421589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421589, CurrentID=0, Status='live', Symbol='LRRC37A7P', Aliases='', description='leucine-rich repeat containing 37 member A7, pseudogene', other_designations='leucine rich repeat containing 37, member A7, pseudogene|leucine rich repeat containing 37A pseudogene', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31724227, end_position_on_the_genomic_accession=31725686, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646912, CurrentID=0, Status='live', Symbol='RPL7P52', Aliases='RPL7_21_1751', description='ribosomal protein L7 pseudogene 52', other_designations='', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=40106253, end_position_on_the_genomic_accession=40107086, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130561, CurrentID=0, Status='live', Symbol='HMGB1P10', Aliases='HMG1L10, HMGB1L10, bK445C9.3', description='high mobility group box 1 pseudogene 10', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 10|high-mobility group box 1 pseudogene 10|high-mobility group box 1-like 10', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26560428, end_position_on_the_genomic_accession=26561616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130562, CurrentID=0, Status='live', Symbol='RPS2P5', Aliases='RPS2_20_1284, bcm384', description='ribosomal protein S2 pseudogene 5', other_designations='', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118246061, end_position_on_the_genomic_accession=118247011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646915, CurrentID=0, Status='live', Symbol='ZNF806', Aliases='', description='zinc finger protein 806', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132307144, end_position_on_the_genomic_accession=132318747, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130564, CurrentID=0, Status='live', Symbol='SEPT7P7', Aliases='', description='septin 7 pseudogene 7', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98601749, end_position_on_the_genomic_accession=98608236, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373445, CurrentID=0, Status='live', Symbol='LOC105373445', Aliases='', description='uncharacterized LOC105373445', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=16529379, end_position_on_the_genomic_accession=16547574, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130566, CurrentID=0, Status='live', Symbol='EIF1P4', Aliases='', description='eukaryotic translation initiation factor 1 pseudogene 4', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2822657, end_position_on_the_genomic_accession=2822994, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122631, CurrentID=0, Status='live', Symbol='CEND1P1', Aliases='', description='cell cycle exit and neuronal differentiation 1 pseudogene 1', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104450806, end_position_on_the_genomic_accession=104451834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373448, CurrentID=0, Status='live', Symbol='LOC105373448', Aliases='', description='uncharacterized LOC105373448', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17433586, end_position_on_the_genomic_accession=17441992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373449, CurrentID=0, Status='live', Symbol='LOC105373449', Aliases='', description='uncharacterized LOC105373449', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17762476, end_position_on_the_genomic_accession=17801314, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373450, CurrentID=0, Status='live', Symbol='ZPAXP', Aliases='', description='zona pellucida glycoprotein AX, pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17847700, end_position_on_the_genomic_accession=17849474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373451, CurrentID=0, Status='live', Symbol='LOC105373451', Aliases='', description='uncharacterized LOC105373451', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18280367, end_position_on_the_genomic_accession=18311575, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130572, CurrentID=0, Status='live', Symbol='RAD23BP2', Aliases='', description='RAD23 homolog B, nucleotide excision repair protein pseudogene 2', other_designations='RAD23 homolog B (S. cerevisiae) pseudogene 2', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83663419, end_position_on_the_genomic_accession=83664625, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373453, CurrentID=0, Status='live', Symbol='LOC105373453', Aliases='', description='uncharacterized LOC105373453', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18492236, end_position_on_the_genomic_accession=18502923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373454, CurrentID=0, Status='live', Symbol='LOC105373454', Aliases='', description='uncharacterized LOC105373454', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18442420, end_position_on_the_genomic_accession=18514629, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373455, CurrentID=0, Status='live', Symbol='LOC105373455', Aliases='', description='uncharacterized LOC105373455', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19040672, end_position_on_the_genomic_accession=19066012, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373456, CurrentID=0, Status='live', Symbol='LOC105373456', Aliases='', description='uncharacterized LOC105373456', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=18563942, end_position_on_the_genomic_accession=19089337, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  646929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646929, CurrentID=0, Status='live', Symbol='ANKRD26P2', Aliases='', description='ankyrin repeat domain 26 pseudogene 2', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=38909228, end_position_on_the_genomic_accession=38936204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253714, CurrentID=0, Status='live', Symbol='MMS22L', Aliases='C6orf167, dJ39B17.2', description='MMS22 like, DNA repair protein', other_designations='MMS22-like, DNA repair protein', map_location='6q16.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=97142161, end_position_on_the_genomic_accession=97283437, orientation='minus', exon_count=32, OMIM=615614, no_hdr0=''),
  326275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326275, CurrentID=0, Status='live', Symbol='RPL12P7', Aliases='RPL12_16_1383', description='ribosomal protein L12 pseudogene 7', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=68693071, end_position_on_the_genomic_accession=68693626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373460, CurrentID=0, Status='live', Symbol='LOC105373460', Aliases='', description='uncharacterized LOC105373460', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19475576, end_position_on_the_genomic_accession=19520010, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130581, CurrentID=0, Status='live', Symbol='LINC00910', Aliases='', description='long intergenic non-protein coding RNA 910', other_designations='CTD-3014M21.4', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=43369845, end_position_on_the_genomic_accession=43388898, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373462, CurrentID=0, Status='live', Symbol='LOC105373462', Aliases='', description='uncharacterized LOC105373462', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19901891, end_position_on_the_genomic_accession=19903112, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373463, CurrentID=0, Status='live', Symbol='LOC105373463', Aliases='', description='uncharacterized LOC105373463', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=20052138, end_position_on_the_genomic_accession=20126029, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100130584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130584, CurrentID=0, Status='live', Symbol='CLIC4P2', Aliases='', description='chloride intracellular channel 4 pseudogene 2', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6562554, end_position_on_the_genomic_accession=6563889, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326276, CurrentID=0, Status='live', Symbol='SF3B4P1', Aliases='SF3B4P', description='splicing factor 3b, subunit 4 pseudogene 1', other_designations='splicing factor 3b, subunit 4, 49kD pseudogene|splicing factor 3b, subunit 4, 49kDa pseudogene|splicing factor 3b, subunit 4, pseudogene', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=67199041, end_position_on_the_genomic_accession=67200560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  646938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646938, CurrentID=0, Status='live', Symbol='LOC646938', Aliases='', description='TBC1 domain family, member 2B pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78752037, end_position_on_the_genomic_accession=78753392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  122651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122651, CurrentID=0, Status='live', Symbol='RNASE11', Aliases='C14orf6, HEL-S-84p, RAJ1', description='ribonuclease, RNase A family, 11 (non-active)', other_designations='RNase 11|epididymis secretory sperm binding protein Li 84p|ribonuclease 11|ribonuclease A J1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20582892, end_position_on_the_genomic_accession=20590258, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  253724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253724, CurrentID=0, Status='live', Symbol='TTC41P', Aliases='GNN, GNNP', description='tetratricopeptide repeat domain 41, pseudogene', other_designations='Grp94 neighboring nucleotidase pseudogene', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=103843749, end_position_on_the_genomic_accession=103930211, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  253725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253725, CurrentID=0, Status='live', Symbol='FAM21C', Aliases='FAM21A, VPEF', description='family with sequence similarity 21 member C', other_designations='family with sequence similarity 21, member C|vaccinia virus penetration factor', map_location='10q11.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45727200, end_position_on_the_genomic_accession=45792964, orientation='plus', exon_count=32, OMIM=613631, no_hdr0=''),
  646942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646942, CurrentID=0, Status='live', Symbol='RPS10P14', Aliases='RPS10_9_805', description='ribosomal protein S10 pseudogene 14', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=37375487, end_position_on_the_genomic_accession=37376073, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373471, CurrentID=0, Status='live', Symbol='LOC105373471', Aliases='', description='uncharacterized LOC105373471', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162420082, end_position_on_the_genomic_accession=162447346, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373472, CurrentID=0, Status='live', Symbol='LOC105373472', Aliases='', description='uncharacterized LOC105373472', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178538473, end_position_on_the_genomic_accession=178542606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373473, CurrentID=0, Status='live', Symbol='LOC105373473', Aliases='', description='uncharacterized LOC105373473', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=183869125, end_position_on_the_genomic_accession=183872705, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373474, CurrentID=0, Status='live', Symbol='LOC105373474', Aliases='', description='uncharacterized LOC105373474', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210293123, end_position_on_the_genomic_accession=210321710, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373475, CurrentID=0, Status='live', Symbol='LOC105373475', Aliases='', description='uncharacterized LOC105373475', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220111577, end_position_on_the_genomic_accession=220112602, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373476, CurrentID=0, Status='live', Symbol='LOC105373476', Aliases='', description='uncharacterized LOC105373476', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=222500737, end_position_on_the_genomic_accession=222501556, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646949, CurrentID=0, Status='live', Symbol='RPL23P6', Aliases='RPL23_1_689', description='ribosomal protein L23 pseudogene 6', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39958351, end_position_on_the_genomic_accession=39958859, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373478, CurrentID=0, Status='live', Symbol='LOC105373478', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159842321, end_position_on_the_genomic_accession=159849260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646951, CurrentID=0, Status='live', Symbol='FAM188B2', Aliases='C3orf76, CLRN1', description='family with sequence similarity 188 member B2', other_designations='Putative UPF0526 protein B|family with sequence similarity 188, member B2', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.11', start_position_on_the_genomic_accession=150588832, end_position_on_the_genomic_accession=150621044, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  122664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122664, CurrentID=0, Status='live', Symbol='TPPP2', Aliases='C14orf8, CT152, P18, p25beta', description='tubulin polymerization-promoting protein family member 2', other_designations='TPPP/p18|protein p25-beta', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21030186, end_position_on_the_genomic_accession=21036352, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  122665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122665, CurrentID=0, Status='live', Symbol='RNASE8', Aliases='', description='ribonuclease, RNase A family, 8', other_designations='RNase 8, placental', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21057822, end_position_on_the_genomic_accession=21058455, orientation='plus', exon_count=1, OMIM=612485, no_hdr0=''),
  646954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646954, CurrentID=0, Status='live', Symbol='YWHAQP4', Aliases='', description='YWHAQ pseudogene 4', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163932532, end_position_on_the_genomic_accession=163933243, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326279, CurrentID=0, Status='live', Symbol='RPL15P2', Aliases='RPL15_6_1404', description='ribosomal protein L15 pseudogene 2', other_designations='', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94799419, end_position_on_the_genomic_accession=94801519, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646956, CurrentID=0, Status='live', Symbol='HNRNPCP3', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 3', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79236332, end_position_on_the_genomic_accession=79237205, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373485, CurrentID=0, Status='live', Symbol='LOC105373485', Aliases='', description='uncharacterized LOC105373485', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=13519497, end_position_on_the_genomic_accession=13533126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130606, CurrentID=0, Status='live', Symbol='NPM1P51', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 51', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41666810, end_position_on_the_genomic_accession=41667855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373487, CurrentID=0, Status='live', Symbol='LOC105373487', Aliases='', description='uncharacterized LOC105373487', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=94947784, end_position_on_the_genomic_accession=94959950, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646960, CurrentID=0, Status='live', Symbol='PRSS56', Aliases='MCOP6', description='protease, serine 56', other_designations='protease, serine, 56|putative serine protease 56', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232520463, end_position_on_the_genomic_accession=232525716, orientation='plus', exon_count=13, OMIM=613858, no_hdr0=''),
  100130609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130609, CurrentID=0, Status='live', Symbol='PP12708', Aliases='', description='uncharacterized LOC100130609', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=139073926, end_position_on_the_genomic_accession=139077724, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646962, CurrentID=0, Status='live', Symbol='HRCT1', Aliases='LGLL338, PRO537, UNQ338', description='histidine rich carboxyl terminus 1', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35906192, end_position_on_the_genomic_accession=35907141, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373491, CurrentID=0, Status='live', Symbol='LOC105373491', Aliases='', description='uncharacterized LOC105373491', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95711719, end_position_on_the_genomic_accession=95718523, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373492, CurrentID=0, Status='live', Symbol='LOC105373492', Aliases='', description='uncharacterized LOC105373492', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95786366, end_position_on_the_genomic_accession=95789577, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130613, CurrentID=0, Status='live', Symbol='PRR32', Aliases='CXorf64', description='proline rich 32', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=126819764, end_position_on_the_genomic_accession=126821786, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  646966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646966, CurrentID=0, Status='live', Symbol='RPL26P16', Aliases='RPL26_3_532', description='ribosomal protein L26 pseudogene 16', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164920577, end_position_on_the_genomic_accession=164921137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373495, CurrentID=0, Status='live', Symbol='LOC105373495', Aliases='', description='uncharacterized LOC105373495', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96307263, end_position_on_the_genomic_accession=96321893, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373496, CurrentID=0, Status='live', Symbol='LOC105373496', Aliases='', description='uncharacterized LOC105373496', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96527940, end_position_on_the_genomic_accession=96532304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373497, CurrentID=0, Status='live', Symbol='LOC105373497', Aliases='', description='uncharacterized LOC105373497', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97037448, end_position_on_the_genomic_accession=97059182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130618, CurrentID=0, Status='live', Symbol='LOC100130618', Aliases='', description='tetratricopeptide repeat domain 14 pseudogene', other_designations='', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122159279, end_position_on_the_genomic_accession=122162781, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373499, CurrentID=0, Status='live', Symbol='LOC105373499', Aliases='', description='uncharacterized LOC105373499', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97607120, end_position_on_the_genomic_accession=97612902, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373500, CurrentID=0, Status='live', Symbol='LOC105373500', Aliases='', description='uncharacterized LOC105373500', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97997643, end_position_on_the_genomic_accession=98001796, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130621, CurrentID=0, Status='live', Symbol='RANP4', Aliases='', description='RAN, member RAS oncogene family pseudogene 4', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136806941, end_position_on_the_genomic_accession=136807944, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373502, CurrentID=0, Status='live', Symbol='LOC105373502', Aliases='', description='uncharacterized LOC105373502', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98331063, end_position_on_the_genomic_accession=98333595, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373503, CurrentID=0, Status='live', Symbol='LOC105373503', Aliases='', description='uncharacterized LOC105373503', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=98698432, end_position_on_the_genomic_accession=98708207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130624, CurrentID=0, Status='live', Symbol='RPL15P22', Aliases='RPL15_10_1742', description='ribosomal protein L15 pseudogene 22', other_designations='', map_location='22q12.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=26886845, end_position_on_the_genomic_accession=26887581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373505, CurrentID=0, Status='live', Symbol='LOC105373505', Aliases='', description='uncharacterized LOC105373505', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99679174, end_position_on_the_genomic_accession=99686597, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130626, CurrentID=0, Status='live', Symbol='TCEB1P18', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 18', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68375322, end_position_on_the_genomic_accession=68375662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373507, CurrentID=0, Status='live', Symbol='LOC105373507', Aliases='', description='uncharacterized LOC105373507', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100603729, end_position_on_the_genomic_accession=100609353, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373508, CurrentID=0, Status='live', Symbol='LOC105373508', Aliases='', description='uncharacterized LOC105373508', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100739947, end_position_on_the_genomic_accession=100747314, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  646981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646981, CurrentID=0, Status='live', Symbol='PTPN2P1', Aliases='PTPN2P, PTPTP1', description='protein tyrosine phosphatase, non-receptor type 2 pseudogene 1', other_designations='hCG2025063|protein tyrosine phosphatase, non-receptor type 2 (pseudogene 1)|protein tyrosine phosphatase, non-receptor type-like pseudogene', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=178746344, end_position_on_the_genomic_accession=178747830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  646982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646982, CurrentID=0, Status='live', Symbol='LINC00598', Aliases='TTL', description='long intergenic non-protein coding RNA 598', other_designations='twelve-thirteen translocation leukemia transcript (non-protein coding)', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=40347132, end_position_on_the_genomic_accession=40481006, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100421601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421601, CurrentID=0, Status='live', Symbol='PPATP2', Aliases='', description='phosphoribosyl pyrophosphate amidotransferase pseudogene 2', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=78699384, end_position_on_the_genomic_accession=78699887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373512, CurrentID=0, Status='live', Symbol='LOC105373512', Aliases='', description='uncharacterized LOC105373512', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101308781, end_position_on_the_genomic_accession=101312792, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253769, CurrentID=0, Status='live', Symbol='WDR27', Aliases='', description='WD repeat domain 27', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169426407, end_position_on_the_genomic_accession=169702063, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  105373514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373514, CurrentID=0, Status='live', Symbol='LOC105373514', Aliases='', description='uncharacterized LOC105373514', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=101551588, end_position_on_the_genomic_accession=101567902, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373515, CurrentID=0, Status='live', Symbol='LOC105373515', Aliases='', description='uncharacterized LOC105373515', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102063806, end_position_on_the_genomic_accession=102065330, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130636, CurrentID=0, Status='live', Symbol='SPATA31D2P', Aliases='FAM75D2P', description='SPATA31 subfamily D, member 2, pseudogene', other_designations='FAM75-like protein FLJ46321 pseudogene|family with sequence similarity 75, member D2, pseudogene', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81966563, end_position_on_the_genomic_accession=81971291, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130637, CurrentID=0, Status='live', Symbol='MORF4L1P7', Aliases='', description='mortality factor 4 like 1 pseudogene 7', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176367706, end_position_on_the_genomic_accession=176368635, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374182, CurrentID=0, Status='live', Symbol='LOC105374182', Aliases='', description='uncharacterized LOC105374182', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158924039, end_position_on_the_genomic_accession=158926862, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373519, CurrentID=0, Status='live', Symbol='LOC105373519', Aliases='', description='uncharacterized LOC105373519', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102967408, end_position_on_the_genomic_accession=102984520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  122704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122704, CurrentID=0, Status='live', Symbol='MRPL52', Aliases='', description='mitochondrial ribosomal protein L52', other_designations='L52mt|MRP-L52', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=22829851, end_position_on_the_genomic_accession=22835037, orientation='plus', exon_count=6, OMIM=611856, no_hdr0=''),
  646993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646993, CurrentID=0, Status='live', Symbol='HMGB3P24', Aliases='', description='high mobility group box 3 pseudogene 24', other_designations='high-mobility group box 3 pseudogene 24', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36302682, end_position_on_the_genomic_accession=36305020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122706, CurrentID=0, Status='live', Symbol='PSMB11', Aliases='BETA5T', description='proteasome (prosome, macropain) subunit, beta type, 11', other_designations='proteasome subunit beta 5T|proteasome subunit beta-5t', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23042167, end_position_on_the_genomic_accession=23044060, orientation='plus', exon_count=1, OMIM=611137, no_hdr0=''),
  105373523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373523, CurrentID=0, Status='live', Symbol='LOC105373523', Aliases='', description='uncharacterized LOC105373523', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104115006, end_position_on_the_genomic_accession=104158335, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  646996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646996, CurrentID=0, Status='live', Symbol='RAB42P1', Aliases='RAB42, RAB42P', description='RAB42, member RAS oncogene family, pseudogene 1', other_designations='RAB42, member RAS homolog family, pseudogene', map_location='14q32.11', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89891784, end_position_on_the_genomic_accession=89892910, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373525, CurrentID=0, Status='live', Symbol='LOC105373525', Aliases='', description='uncharacterized LOC105373525', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=104583152, end_position_on_the_genomic_accession=104585357, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  253782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253782, CurrentID=0, Status='live', Symbol='CERS6', Aliases='CERS5, LASS6', description='ceramide synthase 6', other_designations='LAG1 homolog, ceramide synthase 6|longevity assurance homolog 6', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168456249, end_position_on_the_genomic_accession=168775134, orientation='plus', exon_count=12, OMIM=615336, no_hdr0=''),
  646999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=646999, CurrentID=0, Status='live', Symbol='LOC646999', Aliases='', description='akirin 1 pseudogene', other_designations='', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39609628, end_position_on_the_genomic_accession=39612109, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647000, CurrentID=0, Status='live', Symbol='TUBBP2', Aliases='', description='tubulin beta pseudogene 2', other_designations='tubulin, beta polypeptide pseudogene 2|tubulin, beta pseudogene 2', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41383682, end_position_on_the_genomic_accession=41385076, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647001, CurrentID=0, Status='live', Symbol='GAPDHP35', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 35', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28239483, end_position_on_the_genomic_accession=28240727, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373530, CurrentID=0, Status='live', Symbol='LOC105373530', Aliases='', description='uncharacterized LOC105373530', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105928193, end_position_on_the_genomic_accession=105937732, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373531, CurrentID=0, Status='live', Symbol='LOC105373531', Aliases='', description='uncharacterized LOC105373531', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=105943180, end_position_on_the_genomic_accession=105971483, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373532, CurrentID=0, Status='live', Symbol='LOC105373532', Aliases='', description='uncharacterized LOC105373532', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106083052, end_position_on_the_genomic_accession=106093568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373533, CurrentID=0, Status='live', Symbol='LOC105373533', Aliases='', description='uncharacterized LOC105373533', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106376452, end_position_on_the_genomic_accession=106378247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130654, CurrentID=0, Status='live', Symbol='LOC100130654', Aliases='', description='uncharacterized LOC100130654', other_designations='', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=37060738, end_position_on_the_genomic_accession=37061285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373535, CurrentID=0, Status='live', Symbol='LOC105373535', Aliases='', description='uncharacterized LOC105373535', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107028963, end_position_on_the_genomic_accession=107032908, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373536, CurrentID=0, Status='live', Symbol='LOC105373536', Aliases='', description='uncharacterized LOC105373536', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107260700, end_position_on_the_genomic_accession=107365873, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  647009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647009, CurrentID=0, Status='live', Symbol='RWDD4P2', Aliases='FAM28CP', description='RWD domain containing 4 pseudogene 2', other_designations='family with sequence similarity 28, member C pseudogene', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39852697, end_position_on_the_genomic_accession=39855260, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373538, CurrentID=0, Status='live', Symbol='LOC105373538', Aliases='', description='uncharacterized LOC105373538', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107382472, end_position_on_the_genomic_accession=107542816, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105373539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373539, CurrentID=0, Status='live', Symbol='LOC105373539', Aliases='', description='uncharacterized LOC105373539', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107638478, end_position_on_the_genomic_accession=107651650, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130660, CurrentID=0, Status='live', Symbol='HNRNPA1P45', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 45', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35400691, end_position_on_the_genomic_accession=35401666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647013, CurrentID=0, Status='live', Symbol='KCTD9P3', Aliases='', description='potassium channel tetramerization domain containing 9 pseudogene 3', other_designations='potassium channel tetramerisation domain containing 9 pseudogene', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37477149, end_position_on_the_genomic_accession=37478314, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373542, CurrentID=0, Status='live', Symbol='LOC105373542', Aliases='', description='uncharacterized LOC105373542', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=108882111, end_position_on_the_genomic_accession=108885485, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  647015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647015, CurrentID=0, Status='live', Symbol='CDC37P2', Aliases='', description='cell division cycle 37 pseudogene 2', other_designations='cell division cycle 37 homolog pseudogene 2', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28413774, end_position_on_the_genomic_accession=28415018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647016, CurrentID=0, Status='live', Symbol='EIF2S2P7', Aliases='', description='eukaryotic translation initiation factor 2 subunit 2 beta pseudogene 7', other_designations='eukaryotic translation initiation factor 2, subunit 2 beta pseudogene 7', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57048352, end_position_on_the_genomic_accession=57049336, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373545, CurrentID=0, Status='live', Symbol='LOC105373545', Aliases='', description='uncharacterized LOC105373545', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109694486, end_position_on_the_genomic_accession=109701282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373546, CurrentID=0, Status='live', Symbol='LOC105373546', Aliases='', description='uncharacterized LOC105373546', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109733089, end_position_on_the_genomic_accession=109764662, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373547, CurrentID=0, Status='live', Symbol='LOC105373547', Aliases='', description='uncharacterized LOC105373547', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109918860, end_position_on_the_genomic_accession=109922848, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  647020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647020, CurrentID=0, Status='live', Symbol='LOC647020', Aliases='', description='uncharacterized LOC647020', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.9', start_position_on_the_genomic_accession=81372889, end_position_on_the_genomic_accession=81410492, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  253805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253805, CurrentID=0, Status='live', Symbol='LOC253805', Aliases='', description='uncharacterized LOC253805', other_designations='', map_location='11q25', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373550, CurrentID=0, Status='live', Symbol='LOC105373550', Aliases='', description='uncharacterized LOC105373550', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110341770, end_position_on_the_genomic_accession=110351400, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130671, CurrentID=0, Status='live', Symbol='IFNWP9', Aliases='', description='interferon, omega 1 pseudogene 9', other_designations='interferon, omega 9 (pseudogene)', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21190146, end_position_on_the_genomic_accession=21191244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647024, CurrentID=0, Status='live', Symbol='C6orf132', Aliases='bA7K24.2', description='chromosome 6 open reading frame 132', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42101119, end_position_on_the_genomic_accession=42142977, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373553, CurrentID=0, Status='live', Symbol='LOC105373553', Aliases='', description='uncharacterized LOC105373553', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110611118, end_position_on_the_genomic_accession=110612096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373554, CurrentID=0, Status='live', Symbol='LOC105373554', Aliases='', description='uncharacterized LOC105373554', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110678188, end_position_on_the_genomic_accession=110692786, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326291, CurrentID=0, Status='live', Symbol='RPL18P1', Aliases='RPL18_7_1361', description='ribosomal protein L18 pseudogene 1', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=48489897, end_position_on_the_genomic_accession=48490467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  122740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122740, CurrentID=0, Status='live', Symbol='OR4K14', Aliases='OR14-18, OR14-22', description='olfactory receptor family 4 subfamily K member 14', other_designations='olfactory receptor OR14-18 pseudogene|olfactory receptor OR14-22|olfactory receptor, family 4, subfamily K, member 14', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20014261, end_position_on_the_genomic_accession=20015193, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130677, CurrentID=0, Status='live', Symbol='NFYAP1', Aliases='NFYAP', description='nuclear transcription factor Y subunit alpha pseudogene 1', other_designations='nuclear transcription factor Y, alpha pseudogene 1', map_location='13q21.31', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63975780, end_position_on_the_genomic_accession=63976423, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122742, CurrentID=0, Status='live', Symbol='OR4L1', Aliases='OR14-28, OR4L2P', description='olfactory receptor family 4 subfamily L member 1', other_designations='olfactory receptor 4L2|olfactory receptor OR14-28|olfactory receptor, family 4, subfamily L, member 1|olfactory receptor, family 4, subfamily L, member 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20060045, end_position_on_the_genomic_accession=20060983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928937, CurrentID=0, Status='live', Symbol='LOC101928937', Aliases='', description='uncharacterized LOC101928937', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=99093359, end_position_on_the_genomic_accession=99105011, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373560, CurrentID=0, Status='live', Symbol='LOC105373560', Aliases='', description='uncharacterized LOC105373560', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112061080, end_position_on_the_genomic_accession=112064122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  647033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647033, CurrentID=0, Status='live', Symbol='PA2G4P4', Aliases='PA2G4L1', description='proliferation-associated 2G4 pseudogene 4', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156809271, end_position_on_the_genomic_accession=156812021, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  647034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647034, CurrentID=0, Status='live', Symbol='RPS14P10', Aliases='RPS14_7_847', description='ribosomal protein S14 pseudogene 10', other_designations='', map_location='7q32.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=131665513, end_position_on_the_genomic_accession=131665908, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373563, CurrentID=0, Status='live', Symbol='LOC105373563', Aliases='', description='uncharacterized LOC105373563', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112889625, end_position_on_the_genomic_accession=112908546, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  122748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122748, CurrentID=0, Status='live', Symbol='OR11H6', Aliases='', description='olfactory receptor family 11 subfamily H member 6', other_designations='olfactory receptor OR14-35|olfactory receptor, family 11, subfamily H, member 6', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20223710, end_position_on_the_genomic_accession=20224702, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373565, CurrentID=0, Status='live', Symbol='LOC105373565', Aliases='', description='uncharacterized LOC105373565', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113213365, end_position_on_the_genomic_accession=113214540, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373566, CurrentID=0, Status='live', Symbol='LOC105373566', Aliases='', description='uncharacterized LOC105373566', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113353530, end_position_on_the_genomic_accession=113360000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326293, CurrentID=0, Status='live', Symbol='RPL23AP10', Aliases='RPL23A_28_1405', description='ribosomal protein L23a pseudogene 10', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=96456529, end_position_on_the_genomic_accession=96457053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373568, CurrentID=0, Status='live', Symbol='LOC105373568', Aliases='', description='uncharacterized LOC105373568', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113457224, end_position_on_the_genomic_accession=113502749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373569, CurrentID=0, Status='live', Symbol='LOC105373569', Aliases='', description='uncharacterized LOC105373569', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113643188, end_position_on_the_genomic_accession=113645824, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  647042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647042, CurrentID=0, Status='live', Symbol='GOLGA6L10', Aliases='GOLGA6L18', description='golgin A6 family-like 10', other_designations='golgi autoantigen, golgin subfamily a, 6-like 10|golgin A6 family-like 18', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82339993, end_position_on_the_genomic_accession=82349475, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  253827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253827, CurrentID=0, Status='live', Symbol='MSRB3', Aliases='DFNB74', description='methionine sulfoxide reductase B3', other_designations='methionine-R-sulfoxide reductase B3, mitochondrial', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=65278643, end_position_on_the_genomic_accession=65466907, orientation='plus', exon_count=9, OMIM=613719, no_hdr0=''),
  647044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647044, CurrentID=0, Status='live', Symbol='FAM220BP', Aliases='C9orf51', description='family with sequence similarity 220 member B, pseudogene', other_designations='family with sequence similarity 220, member B, pseudogene', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38526662, end_position_on_the_genomic_accession=38528814, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  326294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326294, CurrentID=0, Status='live', Symbol='RPL23AP11', Aliases='RPL23A_29_1408', description='ribosomal protein L23a pseudogene 11', other_designations='', map_location='14q32.31', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=102662193, end_position_on_the_genomic_accession=102662728, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130694, CurrentID=0, Status='live', Symbol='LOC100130694', Aliases='', description='uncharacterized LOC100130694', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.9', start_position_on_the_genomic_accession=10147249, end_position_on_the_genomic_accession=10148466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647047, CurrentID=0, Status='live', Symbol='TCP1P1', Aliases='CCT1-3P, TCP1L1', description='t-complex 1 pseudogene 1', other_designations='t-complex 1-like 1', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42794772, end_position_on_the_genomic_accession=42796859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253832, CurrentID=0, Status='live', Symbol='ZDHHC20', Aliases='4933421L13Rik, DHHC-20', description='zinc finger, DHHC-type containing 20', other_designations='zinc finger DHHC domain-containing protein 20', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=21372571, end_position_on_the_genomic_accession=21459370, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105373577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373577, CurrentID=0, Status='live', Symbol='LOC105373577', Aliases='', description='uncharacterized LOC105373577', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=117701114, end_position_on_the_genomic_accession=117704474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130698, CurrentID=0, Status='live', Symbol='LOC100130698', Aliases='', description='uncharacterized LOC100130698', other_designations='', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80247751, end_position_on_the_genomic_accession=80255408, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130699, CurrentID=0, Status='live', Symbol='RSL24D1P10', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 10', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121671373, end_position_on_the_genomic_accession=121671872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130700, CurrentID=0, Status='live', Symbol='TP53TG3HP', Aliases='', description='TP53 target 3 family member H, pseudogene', other_designations='TP53 target 3B pseudogene', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35504376, end_position_on_the_genomic_accession=35505417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373581, CurrentID=0, Status='live', Symbol='LOC105373581', Aliases='', description='uncharacterized LOC105373581', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119265588, end_position_on_the_genomic_accession=119271421, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373582, CurrentID=0, Status='live', Symbol='LOC105373582', Aliases='', description='uncharacterized LOC105373582', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120167960, end_position_on_the_genomic_accession=120173745, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373583, CurrentID=0, Status='live', Symbol='LOC105373583', Aliases='', description='uncharacterized LOC105373583', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120235702, end_position_on_the_genomic_accession=120246141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373584, CurrentID=0, Status='live', Symbol='LOC105373584', Aliases='', description='uncharacterized LOC105373584', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120317386, end_position_on_the_genomic_accession=120319159, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122769, CurrentID=0, Status='live', Symbol='LRR1', Aliases='4-1BBLRR, LRR-1, PPIL5', description='leucine rich repeat protein 1', other_designations='4-1BB-mediated signaling molecule|4-1BB-mediated-signaling molecule|LRR-repeat protein 1|cyclophilin-like 5|peptidylprolyl isomerase (cyclophilin)-like 5|peptidylprolyl isomerase-like 5', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49598697, end_position_on_the_genomic_accession=49614672, orientation='plus', exon_count=7, OMIM=609193, no_hdr0=''),
  105373586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373586, CurrentID=0, Status='live', Symbol='LOC105373586', Aliases='', description='uncharacterized LOC105373586', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120596297, end_position_on_the_genomic_accession=120604040, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373587, CurrentID=0, Status='live', Symbol='LOC105373587', Aliases='', description='uncharacterized LOC105373587', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121049463, end_position_on_the_genomic_accession=121057784, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  647060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647060, CurrentID=0, Status='live', Symbol='SPATA31A1', Aliases='C9orf36, C9orf36A, FAM75A1, FAM75A2, SPATA31A2', description='SPATA31 subfamily A, member 1', other_designations='SPATA31 subfamily A, member 2|family with sequence similarity 75, member A1|family with sequence similarity 75, member A2|protein FAM75A1|protein FAM75A2|spermatogenesis-associated protein 31A2', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39355667, end_position_on_the_genomic_accession=39361962, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  122773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122773, CurrentID=0, Status='live', Symbol='KLHDC1', Aliases='MST025', description='kelch domain containing 1', other_designations='', map_location='14q21.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49693105, end_position_on_the_genomic_accession=49753152, orientation='plus', exon_count=14, OMIM=611281, no_hdr0=''),
  105373590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373590, CurrentID=0, Status='live', Symbol='LOC105373590', Aliases='', description='uncharacterized LOC105373590', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121775393, end_position_on_the_genomic_accession=121790364, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373591, CurrentID=0, Status='live', Symbol='LOC105373591', Aliases='', description='uncharacterized LOC105373591', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=122233397, end_position_on_the_genomic_accession=122243224, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373592, CurrentID=0, Status='live', Symbol='LOC105373592', Aliases='', description='uncharacterized LOC105373592', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121902311, end_position_on_the_genomic_accession=122432879, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100130713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130713, CurrentID=0, Status='live', Symbol='LOC100130713', Aliases='', description='uncharacterized LOC100130713', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=45882690, end_position_on_the_genomic_accession=45883894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373594, CurrentID=0, Status='live', Symbol='LOC105373594', Aliases='', description='uncharacterized LOC105373594', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=122800093, end_position_on_the_genomic_accession=122808893, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373595, CurrentID=0, Status='live', Symbol='LOC105373595', Aliases='', description='uncharacterized LOC105373595', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=122874891, end_position_on_the_genomic_accession=122886858, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373596, CurrentID=0, Status='live', Symbol='LOC105373596', Aliases='', description='uncharacterized LOC105373596', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=122987903, end_position_on_the_genomic_accession=122999839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130717, CurrentID=0, Status='live', Symbol='CECR5-AS1', Aliases='CECR4, NCRNA00017', description='CECR5 antisense RNA 1', other_designations='CECR5 antisense RNA 1 (non-protein coding)|cat eye syndrome chromosome region, candidate 4 (non-protein coding)', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=17159389, end_position_on_the_genomic_accession=17165445, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130718, CurrentID=0, Status='live', Symbol='RPL34P26', Aliases='RPL34_12_1312', description='ribosomal protein L34 pseudogene 26', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=50361410, end_position_on_the_genomic_accession=50361763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130719, CurrentID=0, Status='live', Symbol='UQCRBP2', Aliases='', description='ubiquinol-cytochrome c reductase binding protein pseudogene 2', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=162541332, end_position_on_the_genomic_accession=162541659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373600, CurrentID=0, Status='live', Symbol='LOC105373600', Aliases='', description='uncharacterized LOC105373600', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126109779, end_position_on_the_genomic_accession=126118017, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130721, CurrentID=0, Status='live', Symbol='RSL24D1P3', Aliases='', description='ribosomal L24 domain containing 1 pseudogene 3', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65335174, end_position_on_the_genomic_accession=65336005, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  122786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122786, CurrentID=0, Status='live', Symbol='FRMD6', Aliases='C14orf31, EX1, Willin, c14_5320', description='FERM domain containing 6', other_designations='4.1 ezrin radixin moesin (FERM)-containing protein|expanded homolog', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=51489100, end_position_on_the_genomic_accession=51730727, orientation='plus', exon_count=18, OMIM=614555, no_hdr0=''),
  105373603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373603, CurrentID=0, Status='live', Symbol='LOC105373603', Aliases='', description='uncharacterized LOC105373603', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=126893083, end_position_on_the_genomic_accession=126898796, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647076, CurrentID=0, Status='live', Symbol='BRD7P2', Aliases='', description='bromodomain containing 7 pseudogene 2', other_designations='', map_location='3q25.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=160100691, end_position_on_the_genomic_accession=160102912, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647077, CurrentID=0, Status='live', Symbol='LOC647077', Aliases='', description='ribosomal protein L14 pseudogene', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61710597, end_position_on_the_genomic_accession=61710978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373606, CurrentID=0, Status='live', Symbol='LOC105373606', Aliases='', description='endogenous retrovirus group K member 25 Pol protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127337789, end_position_on_the_genomic_accession=127375404, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373607, CurrentID=0, Status='live', Symbol='LOC105373607', Aliases='', description='uncharacterized LOC105373607', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127467025, end_position_on_the_genomic_accession=127489810, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373608, CurrentID=0, Status='live', Symbol='LOC105373608', Aliases='', description='uncharacterized LOC105373608', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=127426234, end_position_on_the_genomic_accession=127470996, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  647081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647081, CurrentID=0, Status='live', Symbol='SDHDP2', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 2', other_designations='succinate dehydrogenase complex, subunit D pseudogene|succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 2', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135444440, end_position_on_the_genomic_accession=135445740, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373610, CurrentID=0, Status='live', Symbol='LOC105373610', Aliases='', description='uncharacterized LOC105373610', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128375519, end_position_on_the_genomic_accession=128379070, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373611, CurrentID=0, Status='live', Symbol='LOC105373611', Aliases='', description='uncharacterized LOC105373611', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=128404939, end_position_on_the_genomic_accession=128644230, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253868, CurrentID=0, Status='live', Symbol='C20orf166-AS1', Aliases='C20orf200, NCRNA00335', description='C20orf166 antisense RNA 1', other_designations='', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62543070, end_position_on_the_genomic_accession=62551561, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130733, CurrentID=0, Status='live', Symbol='LRRC70', Aliases='SLRN', description='leucine rich repeat containing 70', other_designations='synleurin', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62578735, end_position_on_the_genomic_accession=62581448, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373614, CurrentID=0, Status='live', Symbol='LOC105373614', Aliases='', description='uncharacterized LOC105373614', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129546812, end_position_on_the_genomic_accession=129591308, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  647087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647087, CurrentID=0, Status='live', Symbol='C7orf73', Aliases='PL-5283', description='chromosome 7 open reading frame 73', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135662473, end_position_on_the_genomic_accession=135676412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373616, CurrentID=0, Status='live', Symbol='LOC105373616', Aliases='', description='uncharacterized LOC105373616', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129877899, end_position_on_the_genomic_accession=129883679, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373617, CurrentID=0, Status='live', Symbol='LOC105373617', Aliases='', description='uncharacterized LOC105373617', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130107684, end_position_on_the_genomic_accession=130110357, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373618, CurrentID=0, Status='live', Symbol='LOC105373618', Aliases='', description='uncharacterized LOC105373618', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130391653, end_position_on_the_genomic_accession=130393729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373619, CurrentID=0, Status='live', Symbol='LOC105373619', Aliases='', description='collagen alpha-1(XI) chain-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131382330, end_position_on_the_genomic_accession=131394225, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373620, CurrentID=0, Status='live', Symbol='LOC105373620', Aliases='', description='uncharacterized LOC105373620', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131228565, end_position_on_the_genomic_accession=131245710, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373621, CurrentID=0, Status='live', Symbol='LOC105373621', Aliases='', description='uncharacterized LOC105373621', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132335051, end_position_on_the_genomic_accession=132338397, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130742, CurrentID=0, Status='live', Symbol='LRRC69', Aliases='', description='leucine rich repeat containing 69', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91102619, end_position_on_the_genomic_accession=91219236, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373623, CurrentID=0, Status='live', Symbol='LOC105373623', Aliases='', description='uncharacterized LOC105373623', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132656417, end_position_on_the_genomic_accession=132666990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373624, CurrentID=0, Status='live', Symbol='LOC105373624', Aliases='', description='uncharacterized LOC105373624', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132943687, end_position_on_the_genomic_accession=132952620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122809, CurrentID=0, Status='live', Symbol='SOCS4', Aliases='SOCS7', description='suppressor of cytokine signaling 4', other_designations='SH2 domain containing SOCS box protein|SOCS-4|SOCS-7|suppressor of cytokine signaling 7', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55027126, end_position_on_the_genomic_accession=55049489, orientation='plus', exon_count=3, OMIM=616337, no_hdr0=''),
  105373626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373626, CurrentID=0, Status='live', Symbol='LOC105373626', Aliases='', description='uncharacterized LOC105373626', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=133927335, end_position_on_the_genomic_accession=133952961, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  647099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647099, CurrentID=0, Status='live', Symbol='RPL23AP42', Aliases='RPL23A_13_431', description='ribosomal protein L23a pseudogene 42', other_designations='hCG16001', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161429072, end_position_on_the_genomic_accession=161429620, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373628, CurrentID=0, Status='live', Symbol='LOC105373628', Aliases='', description='uncharacterized LOC105373628', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134018705, end_position_on_the_genomic_accession=134031272, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373629, CurrentID=0, Status='live', Symbol='LOC105373629', Aliases='', description='uncharacterized LOC105373629', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134718275, end_position_on_the_genomic_accession=134722526, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130750, CurrentID=0, Status='live', Symbol='RPS8P9', Aliases='RPS8_4_974', description='ribosomal protein S8 pseudogene 9', other_designations='', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34223931, end_position_on_the_genomic_accession=34224651, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373631, CurrentID=0, Status='live', Symbol='LOC105373631', Aliases='', description='uncharacterized LOC105373631', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135225838, end_position_on_the_genomic_accession=135238968, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373632, CurrentID=0, Status='live', Symbol='LOC105373632', Aliases='', description='uncharacterized LOC105373632', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136088441, end_position_on_the_genomic_accession=136114702, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373633, CurrentID=0, Status='live', Symbol='LOC105373633', Aliases='', description='uncharacterized LOC105373633', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136544831, end_position_on_the_genomic_accession=136575840, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373634, CurrentID=0, Status='live', Symbol='LOC105373634', Aliases='', description='uncharacterized LOC105373634', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=137493466, end_position_on_the_genomic_accession=137507560, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  647107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647107, CurrentID=0, Status='live', Symbol='LINC01192', Aliases='CT64', description='long intergenic non-protein coding RNA 1192', other_designations='cancer/testis antigen 64 (non-protein coding)', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=163177243, end_position_on_the_genomic_accession=163303301, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130756, CurrentID=0, Status='live', Symbol='UQCRHP1', Aliases='Em:AF129756.18, UQCRHP5', description='ubiquinol-cytochrome c reductase hinge protein pseudogene 1', other_designations='ubiquinol-cytochrome c reductase hinge protein pseudogene 5', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31611083, end_position_on_the_genomic_accession=31611356, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373637, CurrentID=0, Status='live', Symbol='LOC105373637', Aliases='', description='uncharacterized LOC105373637', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138492204, end_position_on_the_genomic_accession=138500791, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130758, CurrentID=0, Status='live', Symbol='METTL15P1', Aliases='METT5D2', description='methyltransferase like 15 pseudogene 1', other_designations='methyltransferase 5 domain containing 2', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156711403, end_position_on_the_genomic_accession=156715029, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373639, CurrentID=0, Status='live', Symbol='LOC105373639', Aliases='', description='uncharacterized LOC105373639', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138910055, end_position_on_the_genomic_accession=138917218, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647112, CurrentID=0, Status='live', Symbol='RPL27P6', Aliases='RPL27_2_233', description='ribosomal protein L27 pseudogene 6', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63684271, end_position_on_the_genomic_accession=63684746, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130761, CurrentID=0, Status='live', Symbol='LOC100130761', Aliases='', description='uncharacterized LOC100130761', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=161998819, end_position_on_the_genomic_accession=161999181, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373642, CurrentID=0, Status='live', Symbol='LOC105373642', Aliases='', description='uncharacterized LOC105373642', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139234686, end_position_on_the_genomic_accession=139359561, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  647115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647115, CurrentID=0, Status='live', Symbol='FLJ36848', Aliases='', description='uncharacterized LOC647115', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=64068283, end_position_on_the_genomic_accession=64069230, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373644, CurrentID=0, Status='live', Symbol='LOC105373644', Aliases='', description='uncharacterized LOC105373644', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139384818, end_position_on_the_genomic_accession=139388334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373645, CurrentID=0, Status='live', Symbol='LOC105373645', Aliases='', description='uncharacterized LOC105373645', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=139411297, end_position_on_the_genomic_accession=139474824, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  122830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122830, CurrentID=0, Status='live', Symbol='NAA30', Aliases='C14orf35, MAK3, Mak3p, NAT12, NAT12P', description='N(alpha)-acetyltransferase 30, NatC catalytic subunit', other_designations='N-acetyltransferase 12 (GCN5-related, putative)|N-acetyltransferase MAK3 homolog|N-alpha-acetyltransferase 30, NatC catalytic subunit|natC catalytic subunit|putative N-acetyltransferase', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=57390553, end_position_on_the_genomic_accession=57412748, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373647, CurrentID=0, Status='live', Symbol='LOC105373647', Aliases='', description='uncharacterized LOC105373647', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140079959, end_position_on_the_genomic_accession=140117665, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373648, CurrentID=0, Status='live', Symbol='LOC105373648', Aliases='', description='uncharacterized LOC105373648', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=141846829, end_position_on_the_genomic_accession=141855907, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  647121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647121, CurrentID=0, Status='live', Symbol='EMBP1', Aliases='', description='embigin pseudogene 1', other_designations='embigin homolog pseudogene', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121519112, end_position_on_the_genomic_accession=121571888, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373650, CurrentID=0, Status='live', Symbol='LOC105373650', Aliases='', description='uncharacterized LOC105373650', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=140712158, end_position_on_the_genomic_accession=140738920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130771, CurrentID=0, Status='live', Symbol='EFCAB10', Aliases='', description='EF-hand calcium binding domain 10', other_designations='EF-hand domain-containing protein', map_location='7q22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105565133, end_position_on_the_genomic_accession=105581529, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373652, CurrentID=0, Status='live', Symbol='LOC105373652', Aliases='', description='uncharacterized LOC105373652', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=142862572, end_position_on_the_genomic_accession=142871212, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  122837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122837, CurrentID=0, Status='live', Symbol='UBA52P3', Aliases='RPL40_5_1373', description='ubiquitin A-52 residue ribosomal protein fusion product 1 pseudogene 3', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58141238, end_position_on_the_genomic_accession=58141758, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373654, CurrentID=0, Status='live', Symbol='LOC105373654', Aliases='', description='uncharacterized LOC105373654', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=143628243, end_position_on_the_genomic_accession=143719630, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130775, CurrentID=0, Status='live', Symbol='RPS27P23', Aliases='RPS27_13_1271', description='ribosomal protein S27 pseudogene 23', other_designations='ribosomal protein S27-like pseudogene', map_location='12q23.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101424409, end_position_on_the_genomic_accession=101424751, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130776, CurrentID=0, Status='live', Symbol='AGAP2-AS1', Aliases='PUNISHER', description='AGAP2 antisense RNA 1', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57726240, end_position_on_the_genomic_accession=57728356, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373657, CurrentID=0, Status='live', Symbol='LOC105373657', Aliases='', description='uncharacterized LOC105373657', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144485908, end_position_on_the_genomic_accession=144504067, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  122842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122842, CurrentID=0, Status='live', Symbol='PPIAP5', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 5', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=59181571, end_position_on_the_genomic_accession=59182144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373659, CurrentID=0, Status='live', Symbol='LOC105373659', Aliases='', description='uncharacterized LOC105373659', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=144450972, end_position_on_the_genomic_accession=144485659, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  647132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647132, CurrentID=0, Status='live', Symbol='LOC647132', Aliases='', description='DEAH (Asp-Glu-Ala-His) box polypeptide 29 pseudogene', other_designations='', map_location='1q31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=189989536, end_position_on_the_genomic_accession=189993203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373661, CurrentID=0, Status='live', Symbol='LOC105373661', Aliases='', description='uncharacterized LOC105373661', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145188108, end_position_on_the_genomic_accession=145263207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130782, CurrentID=0, Status='live', Symbol='RPS15AP15', Aliases='RPS15A_6_237', description='ribosomal protein S15a pseudogene 15', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65511768, end_position_on_the_genomic_accession=65512146, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647135, CurrentID=0, Status='live', Symbol='SRGAP2B', Aliases='SRGAP2L, SRGAP2P2', description='SLIT-ROBO Rho GTPase activating protein 2B', other_designations='SLIT-ROBO Rho GTPase activating protein 2 pseudogene 2|SLIT-ROBO Rho GTPase activating protein 2B (pseudogene)', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=144887191, end_position_on_the_genomic_accession=145095328, orientation='minus', exon_count=12, OMIM=614703, no_hdr0=''),
  105373664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373664, CurrentID=0, Status='live', Symbol='LOC105373664', Aliases='', description='uncharacterized LOC105373664', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=145600921, end_position_on_the_genomic_accession=145631847, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100130785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130785, CurrentID=0, Status='live', Symbol='SALL4P5', Aliases='', description='spalt-like transcription factor 4 pseudogene 5', other_designations='sal-like 4 pseudogene 5', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=22989823, end_position_on_the_genomic_accession=22991579, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130786, CurrentID=0, Status='live', Symbol='RPS3AP22', Aliases='RPS3A_9_594', description='ribosomal protein S3a pseudogene 22', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88381909, end_position_on_the_genomic_accession=88382771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373667, CurrentID=0, Status='live', Symbol='LOC105373667', Aliases='', description='uncharacterized LOC105373667', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146203715, end_position_on_the_genomic_accession=146413225, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373668, CurrentID=0, Status='live', Symbol='LOC105373668', Aliases='', description='uncharacterized LOC105373668', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146856681, end_position_on_the_genomic_accession=146878427, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373669, CurrentID=0, Status='live', Symbol='LOC105373669', Aliases='', description='uncharacterized LOC105373669', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=146837770, end_position_on_the_genomic_accession=146856686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373670, CurrentID=0, Status='live', Symbol='LOC105373670', Aliases='', description='uncharacterized LOC105373670', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147088083, end_position_on_the_genomic_accession=147390670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373671, CurrentID=0, Status='live', Symbol='LOC105373671', Aliases='', description='uncharacterized LOC105373671', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147858938, end_position_on_the_genomic_accession=147903047, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100130792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130792, CurrentID=0, Status='live', Symbol='RPL29P27', Aliases='RPL29_13_1221', description='ribosomal protein L29 pseudogene 27', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=28564547, end_position_on_the_genomic_accession=28565171, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647145, CurrentID=0, Status='live', Symbol='LOC647145', Aliases='', description='serine/arginine-rich splicing factor 8 pseudogene', other_designations='serine/arginine-rich splicing factor 2 pseudogene', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47052459, end_position_on_the_genomic_accession=47055020, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130794, CurrentID=0, Status='live', Symbol='NDUFS5P7', Aliases='', description='NADH:ubiquinone oxidoreductase subunit S5 pseudogene 7', other_designations='NADH dehydrogenase (ubiquinone) Fe-S protein 5, 15kDa (NADH-coenzyme Q reductase) pseudogene 7', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29021201, end_position_on_the_genomic_accession=29021694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  326311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326311, CurrentID=0, Status='live', Symbol='RPL9P6', Aliases='RPL9_10_1398', description='ribosomal protein L9 pseudogene 6', other_designations='', map_location='14q31.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=81776920, end_position_on_the_genomic_accession=81777633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373676, CurrentID=0, Status='live', Symbol='LOC105373676', Aliases='', description='uncharacterized LOC105373676', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148951085, end_position_on_the_genomic_accession=148955218, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253933, CurrentID=0, Status='live', Symbol='OR7E39P', Aliases='OR7E138P, hg611', description='olfactory receptor family 7 subfamily E member 39 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 138 pseudogene|olfactory receptor, family 7, subfamily E, member 39 pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6834823, end_position_on_the_genomic_accession=6835794, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130798, CurrentID=0, Status='live', Symbol='LOC100130798', Aliases='', description='uncharacterized LOC100130798', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178968225, end_position_on_the_genomic_accession=178990116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  253935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253935, CurrentID=0, Status='live', Symbol='ANGPTL5', Aliases='', description='angiopoietin like 5', other_designations='angiopoietin-like 5|angiopoietin-like protein 5', map_location='11q22.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101890674, end_position_on_the_genomic_accession=101916715, orientation='minus', exon_count=9, OMIM=607666, no_hdr0=''),
  105373680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373680, CurrentID=0, Status='live', Symbol='LOC105373680', Aliases='', description='uncharacterized LOC105373680', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149745651, end_position_on_the_genomic_accession=149749703, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326312, CurrentID=0, Status='live', Symbol='RPS12P1', Aliases='RPS12_13_1388', description='ribosomal protein S12 pseudogene 1', other_designations='', map_location='14q24.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=72963747, end_position_on_the_genomic_accession=72964053, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130802, CurrentID=0, Status='live', Symbol='RPL39P13', Aliases='RPL39_3_72', description='ribosomal protein L39 pseudogene 13', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91600112, end_position_on_the_genomic_accession=91600504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647155, CurrentID=0, Status='live', Symbol='RPS15AP20', Aliases='RPS15A_10_701', description='ribosomal protein S15a pseudogene 20', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51537130, end_position_on_the_genomic_accession=51537575, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647156, CurrentID=0, Status='live', Symbol='ACTR3BP3', Aliases='', description='ACTR3B pseudogene 3', other_designations='ARP3 actin-related protein 3 homolog B pseudogene 3', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32378627, end_position_on_the_genomic_accession=32380268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373685, CurrentID=0, Status='live', Symbol='LOC105373685', Aliases='', description='uncharacterized LOC105373685', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=150426604, end_position_on_the_genomic_accession=150428905, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647158, CurrentID=0, Status='live', Symbol='RPSAP33', Aliases='RPSA_10_433', description='ribosomal protein SA pseudogene 33', other_designations='', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168666484, end_position_on_the_genomic_accession=168667371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253943, CurrentID=0, Status='live', Symbol='YTHDF3', Aliases='', description='YTH N(6)-methyladenosine RNA binding protein 3', other_designations='YTH domain family protein 3|YTH domain family, member 3', map_location='8q12.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63168553, end_position_on_the_genomic_accession=63212788, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373688, CurrentID=0, Status='live', Symbol='LOC105373688', Aliases='', description='uncharacterized LOC105373688', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152178860, end_position_on_the_genomic_accession=152180718, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373689, CurrentID=0, Status='live', Symbol='LOC105373689', Aliases='', description='uncharacterized LOC105373689', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152175871, end_position_on_the_genomic_accession=152190687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  647162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647162, CurrentID=0, Status='live', Symbol='KRT18P33', Aliases='', description='keratin 18 pseudogene 33', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65666469, end_position_on_the_genomic_accession=65667794, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373691, CurrentID=0, Status='live', Symbol='LOC105373691', Aliases='', description='uncharacterized LOC105373691', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=152958936, end_position_on_the_genomic_accession=153028286, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  122876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122876, CurrentID=0, Status='live', Symbol='GPHB5', Aliases='B5, GPB5, ZLUT1', description='glycoprotein hormone beta 5', other_designations='glycoprotein beta 5|glycoprotein hormone beta subunit|thyrostimulin subunit beta', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63312835, end_position_on_the_genomic_accession=63318879, orientation='minus', exon_count=3, OMIM=609652, no_hdr0=''),
  105373693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373693, CurrentID=0, Status='live', Symbol='LOC105373693', Aliases='', description='uncharacterized LOC105373693', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154486463, end_position_on_the_genomic_accession=154509447, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647166, CurrentID=0, Status='live', Symbol='LINC00371', Aliases='LINC00372', description='long intergenic non-protein coding RNA 371', other_designations='long intergenic non-protein coding RNA 372', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51082848, end_position_on_the_genomic_accession=51172388, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  647167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647167, CurrentID=0, Status='live', Symbol='EEF1A1P14', Aliases='EEF1AL12', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 14', other_designations='eukaryotic translation elongation factor 1 alpha 1 (EEF1A1) pseudogene|eukaryotic translation elongation factor 1 alpha-like 12 pseudogene', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=194188904, end_position_on_the_genomic_accession=194190542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373696, CurrentID=0, Status='live', Symbol='LOC105373696', Aliases='', description='uncharacterized LOC105373696', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=154950224, end_position_on_the_genomic_accession=155054274, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  647169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647169, CurrentID=0, Status='live', Symbol='GSTA12P', Aliases='GSTAP5', description='glutathione S-transferase alpha 12, pseudogene', other_designations='glutathione S-transferase alpha 3 pseudogene', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52765280, end_position_on_the_genomic_accession=52778355, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373698, CurrentID=0, Status='live', Symbol='LOC105373698', Aliases='', description='uncharacterized LOC105373698', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155263627, end_position_on_the_genomic_accession=155281981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373699, CurrentID=0, Status='live', Symbol='LOC105373699', Aliases='', description='uncharacterized LOC105373699', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=155363585, end_position_on_the_genomic_accession=155417986, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130820, CurrentID=0, Status='live', Symbol='CCNG1P1', Aliases='', description='cyclin G1 pseudogene 1', other_designations='', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132698783, end_position_on_the_genomic_accession=132699666, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130821, CurrentID=0, Status='live', Symbol='OSBPL9P1', Aliases='', description='oxysterol binding protein like 9 pseudogene 1', other_designations='oxysterol binding protein-like 9 pseudogene 1', map_location='3p12.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80163337, end_position_on_the_genomic_accession=80164968, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647174, CurrentID=0, Status='live', Symbol='SERPINE3', Aliases='', description='serpin peptidase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 3', other_designations='nexin-related serine protease inhibitor', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=51338519, end_position_on_the_genomic_accession=51365454, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  253959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253959, CurrentID=0, Status='live', Symbol='RALGAPA1', Aliases='GARNL1, GRIPE, RalGAPalpha1, TULIP1, p240', description='Ral GTPase activating protein, alpha subunit 1 (catalytic)', other_designations='GAP-related interacting protein to E12|GTPase activating RANGAP domain-like 1|GTPase activating Rap/RanGAP domain-like 1|tuberin-like protein 1', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35538352, end_position_on_the_genomic_accession=35809304, orientation='minus', exon_count=46, OMIM=608884, no_hdr0=''),
  122888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122888, CurrentID=0, Status='live', Symbol='PTBP1P', Aliases='HNRNPIP', description='polypyrimidine tract binding protein 1 pseudogene', other_designations='', map_location='14q23.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=65279220, end_position_on_the_genomic_accession=65281655, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647177, CurrentID=0, Status='live', Symbol='GSTA10P', Aliases='GSTAP2', description='glutathione S-transferase alpha 10, pseudogene', other_designations='glutathione S-transferase alpha 3 pseudogene', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52873014, end_position_on_the_genomic_accession=52890798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253962, CurrentID=0, Status='live', Symbol='CACNA1G-AS1', Aliases='', description='CACNA1G antisense RNA 1', other_designations='CACNA1G antisense RNA 1 (non-protein coding)', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50556207, end_position_on_the_genomic_accession=50562108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130827, CurrentID=0, Status='live', Symbol='SBK3', Aliases='SGK110', description='SH3 domain binding kinase family member 3', other_designations='SH3 domain binding kinase family, member 3|SH3-binding domain kinase family member 3|Sugen kinase 110', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55540559, end_position_on_the_genomic_accession=55545543, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373708, CurrentID=0, Status='live', Symbol='LOC105373708', Aliases='', description='uncharacterized LOC105373708', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156666600, end_position_on_the_genomic_accession=156694613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421634, CurrentID=0, Status='live', Symbol='CUBNP3', Aliases='', description='cubilin (intrinsic factor-cobalamin receptor) pseudogene 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45127314, end_position_on_the_genomic_accession=45148915, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373710, CurrentID=0, Status='live', Symbol='LOC105373710', Aliases='', description='uncharacterized LOC105373710', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=156968414, end_position_on_the_genomic_accession=157056592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373711, CurrentID=0, Status='live', Symbol='LOC105373711', Aliases='', description='uncharacterized LOC105373711', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157392922, end_position_on_the_genomic_accession=157404877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373712, CurrentID=0, Status='live', Symbol='LOC105373712', Aliases='', description='uncharacterized LOC105373712', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157464941, end_position_on_the_genomic_accession=157491080, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373713, CurrentID=0, Status='live', Symbol='LOC105373713', Aliases='', description='uncharacterized LOC105373713', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157725716, end_position_on_the_genomic_accession=157728415, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  253970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253970, CurrentID=0, Status='live', Symbol='SFTA3', Aliases='NANCI, SFTPH, SP-H', description='surfactant associated 3', other_designations='Nkx2.1-associated noncoding intergenic RNA|putative protein SFTA3|surfactant-associated protein H', map_location='14q13.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=36473288, end_position_on_the_genomic_accession=36513785, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  647187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647187, CurrentID=0, Status='live', Symbol='IGHV3OR16-6', Aliases='IGHV3OR166', description='immunoglobulin heavy variable 3/OR16-6 (pseudogene)', other_designations='IGHV3/OR16-6|immunoglobulin heavy variable 3/OR16-6 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32916796, end_position_on_the_genomic_accession=32978333, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105373716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373716, CurrentID=0, Status='live', Symbol='LOC105373716', Aliases='', description='uncharacterized LOC105373716', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159234309, end_position_on_the_genomic_accession=159241064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373717, CurrentID=0, Status='live', Symbol='LOC105373717', Aliases='', description='uncharacterized LOC105373717', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=160035060, end_position_on_the_genomic_accession=160141970, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  647190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647190, CurrentID=0, Status='live', Symbol='RPS16P5', Aliases='RPS16_3_702', description='ribosomal protein S16 pseudogene 5', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53334460, end_position_on_the_genomic_accession=53337617, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373719, CurrentID=0, Status='live', Symbol='LOC105373719', Aliases='', description='uncharacterized LOC105373719', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161067776, end_position_on_the_genomic_accession=161160272, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373720, CurrentID=0, Status='live', Symbol='LOC105373720', Aliases='', description='uncharacterized LOC105373720', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161469583, end_position_on_the_genomic_accession=161489464, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373721, CurrentID=0, Status='live', Symbol='LOC105373721', Aliases='', description='uncharacterized LOC105373721', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161902104, end_position_on_the_genomic_accession=161913163, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373722, CurrentID=0, Status='live', Symbol='LOC105373722', Aliases='', description='uncharacterized LOC105373722', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161883345, end_position_on_the_genomic_accession=161902093, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647195, CurrentID=0, Status='live', Symbol='PRR13P1', Aliases='', description='proline rich 13 pseudogene 1', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198197586, end_position_on_the_genomic_accession=198198241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  253980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253980, CurrentID=0, Status='live', Symbol='KCTD13', Aliases='BACURD1, FKSG86, PDIP1, POLDIP1, hBACURD1', description='potassium channel tetramerization domain containing 13', other_designations='BTB/POZ domain-containing protein KCTD13|TNFAIP1-like protein|polymerase delta-interacting protein 1|potassium channel tetramerisation domain containing 13', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29906335, end_position_on_the_genomic_accession=29926232, orientation='minus', exon_count=8, OMIM=608947, no_hdr0=''),
  105373725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373725, CurrentID=0, Status='live', Symbol='LOC105373725', Aliases='', description='uncharacterized LOC105373725', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162839635, end_position_on_the_genomic_accession=162854772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  253982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253982, CurrentID=0, Status='live', Symbol='ASPHD1', Aliases='', description='aspartate beta-hydroxylase domain containing 1', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29900826, end_position_on_the_genomic_accession=29920112, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373727, CurrentID=0, Status='live', Symbol='LOC105373727', Aliases='', description='uncharacterized LOC105373727', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=163259328, end_position_on_the_genomic_accession=163329379, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100130848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130848, CurrentID=0, Status='live', Symbol='TSPY20P', Aliases='', description='testis specific protein, Y-linked 20, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=10035018, end_position_on_the_genomic_accession=10038389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  647201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647201, CurrentID=0, Status='live', Symbol='RPL23AP16', Aliases='RPL23A_4_145', description='ribosomal protein L23a pseudogene 16', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=199371876, end_position_on_the_genomic_accession=199372286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  253986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=253986, CurrentID=0, Status='live', Symbol='RPL10AP2', Aliases='RPL10A_5_900', description='ribosomal protein L10a pseudogene 2', other_designations='', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47157087, end_position_on_the_genomic_accession=47157802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373731, CurrentID=0, Status='live', Symbol='LOC105373731', Aliases='', description='uncharacterized LOC105373731', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=166485072, end_position_on_the_genomic_accession=166547759, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373732, CurrentID=0, Status='live', Symbol='LOC105373732', Aliases='', description='uncharacterized LOC105373732', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168211339, end_position_on_the_genomic_accession=168213423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373733, CurrentID=0, Status='live', Symbol='LOC105373733', Aliases='', description='uncharacterized LOC105373733', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168341189, end_position_on_the_genomic_accession=168344168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373734, CurrentID=0, Status='live', Symbol='LOC105373734', Aliases='', description='uncharacterized LOC105373734', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=168344314, end_position_on_the_genomic_accession=168421771, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100130855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130855, CurrentID=0, Status='live', Symbol='USP3-AS1', Aliases='', description='USP3 antisense RNA 1', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=63587710, end_position_on_the_genomic_accession=63600827, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130856, CurrentID=0, Status='live', Symbol='LOC100130856', Aliases='', description='uncharacterized LOC100130856', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.9', start_position_on_the_genomic_accession=6377008, end_position_on_the_genomic_accession=6379184, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130857, CurrentID=0, Status='live', Symbol='LOC100130857', Aliases='', description='uncharacterized LOC100130857', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=30055543, end_position_on_the_genomic_accession=30055992, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105373738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373738, CurrentID=0, Status='live', Symbol='LOC105373738', Aliases='', description='uncharacterized LOC105373738', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171771784, end_position_on_the_genomic_accession=171793618, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373739, CurrentID=0, Status='live', Symbol='LOC105373739', Aliases='', description='uncharacterized LOC105373739', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171903126, end_position_on_the_genomic_accession=171916801, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373740, CurrentID=0, Status='live', Symbol='LOC105373740', Aliases='', description='uncharacterized LOC105373740', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171916949, end_position_on_the_genomic_accession=171922316, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373741, CurrentID=0, Status='live', Symbol='LOC105373741', Aliases='', description='uncharacterized LOC105373741', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172137373, end_position_on_the_genomic_accession=172153485, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373742, CurrentID=0, Status='live', Symbol='LOC105373742', Aliases='', description='uncharacterized LOC105373742', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172324055, end_position_on_the_genomic_accession=172330554, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  647215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647215, CurrentID=0, Status='live', Symbol='MROH3P', Aliases='C1orf81', description='maestro heat-like repeat family member 3, pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200916635, end_position_on_the_genomic_accession=200966676, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  105373744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373744, CurrentID=0, Status='live', Symbol='LOC105373744', Aliases='', description='uncharacterized LOC105373744', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173435475, end_position_on_the_genomic_accession=173446779, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373745, CurrentID=0, Status='live', Symbol='LOC105373745', Aliases='', description='uncharacterized LOC105373745', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=173813386, end_position_on_the_genomic_accession=173816126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130866, CurrentID=0, Status='live', Symbol='IFNWP15', Aliases='', description='interferon, omega 1 pseudogene 15', other_designations='interferon, omega 15 (pseudogene)', map_location='9p22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21178510, end_position_on_the_genomic_accession=21179598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647219, CurrentID=0, Status='live', Symbol='ASCL5', Aliases='bHLHa47', description='achaete-scute family bHLH transcription factor 5', other_designations='ASH-5|achaete-scute complex homolog 5|class A basic helix-loop-helix protein 47|hASH5', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201113953, end_position_on_the_genomic_accession=201115372, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373748, CurrentID=0, Status='live', Symbol='LOC105373748', Aliases='', description='uncharacterized LOC105373748', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174487305, end_position_on_the_genomic_accession=174489311, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373749, CurrentID=0, Status='live', Symbol='LOC105373749', Aliases='', description='uncharacterized LOC105373749', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174649248, end_position_on_the_genomic_accession=174657611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373750, CurrentID=0, Status='live', Symbol='LOC105373750', Aliases='', description='uncharacterized LOC105373750', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174772904, end_position_on_the_genomic_accession=174782066, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130871, CurrentID=0, Status='live', Symbol='ARMC8P1', Aliases='', description='armadillo repeat containing 8 pseudogene 1', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38452785, end_position_on_the_genomic_accession=38454370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130872, CurrentID=0, Status='live', Symbol='LOC100130872', Aliases='', description='uncharacterized LOC100130872', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1195783, end_position_on_the_genomic_accession=1208962, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130873, CurrentID=0, Status='live', Symbol='CARM1P1', Aliases='CARM1L', description='coactivator associated arginine methyltransferase 1 pseudogene 1', other_designations='coactivator-associated arginine methyltransferase 1 pseudogene 1', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2943562, end_position_on_the_genomic_accession=3053404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373754, CurrentID=0, Status='live', Symbol='LOC105373754', Aliases='', description='uncharacterized LOC105373754', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176611437, end_position_on_the_genomic_accession=176616503, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373755, CurrentID=0, Status='live', Symbol='LOC105373755', Aliases='', description='uncharacterized LOC105373755', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176495530, end_position_on_the_genomic_accession=176589896, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373756, CurrentID=0, Status='live', Symbol='LOC105373756', Aliases='', description='uncharacterized LOC105373756', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176849010, end_position_on_the_genomic_accession=176862400, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  254013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254013, CurrentID=0, Status='live', Symbol='METTL20', Aliases='C12orf72, ETFB-KMT', description='methyltransferase like 20', other_designations='ETFB lysine methyltransferase|methyltransferase-like protein 20', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31647160, end_position_on_the_genomic_accession=31669086, orientation='plus', exon_count=6, OMIM=615256, no_hdr0=''),
  122942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122942, CurrentID=0, Status='live', Symbol='FXNP1', Aliases='FRDAP', description='frataxin pseudogene 1', other_designations='Friedreich ataxia pseudogene', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=78059500, end_position_on_the_genomic_accession=78060384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130879, CurrentID=0, Status='live', Symbol='ZNRF2P3', Aliases='', description='zinc and ring finger 2 pseudogene 3', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79981477, end_position_on_the_genomic_accession=79982597, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130880, CurrentID=0, Status='live', Symbol='LOC100130880', Aliases='', description='uncharacterized LOC100130880', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137953348, end_position_on_the_genomic_accession=137957966, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  122945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122945, CurrentID=0, Status='live', Symbol='NOXRED1', Aliases='C14orf148', description='NADP-dependent oxidoreductase domain containing 1', other_designations='pyrroline-5-carboxylate reductase-like protein C14orf148', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77394021, end_position_on_the_genomic_accession=77426011, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373762, CurrentID=0, Status='live', Symbol='LOC105373762', Aliases='', description='uncharacterized LOC105373762', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177471907, end_position_on_the_genomic_accession=177476741, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373763, CurrentID=0, Status='live', Symbol='LOC105373763', Aliases='', description='uncharacterized LOC105373763', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177836950, end_position_on_the_genomic_accession=177840285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647236, CurrentID=0, Status='live', Symbol='RPL39P15', Aliases='RPL39_5_239', description='ribosomal protein L39 pseudogene 15', other_designations='60S ribosomal protein L39 pseudogene', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70253678, end_position_on_the_genomic_accession=70254058, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373765, CurrentID=0, Status='live', Symbol='LOC105373765', Aliases='', description='uncharacterized LOC105373765', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178924631, end_position_on_the_genomic_accession=178927625, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122950, CurrentID=0, Status='live', Symbol='CYCSP1', Aliases='HC5, HCP36', description='cytochrome c, somatic pseudogene 1', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=76633714, end_position_on_the_genomic_accession=76634028, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130887, CurrentID=0, Status='live', Symbol='ATE1-AS1', Aliases='', description='ATE1 antisense RNA 1', other_designations='ATE1 antisense RNA 1 (head to head)|ENST00000437593', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121928312, end_position_on_the_genomic_accession=121951965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373768, CurrentID=0, Status='live', Symbol='LOC105373768', Aliases='', description='uncharacterized LOC105373768', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180222897, end_position_on_the_genomic_accession=180224210, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130889, CurrentID=0, Status='live', Symbol='PSORS1C3', Aliases='NCRNA00196', description='psoriasis susceptibility 1 candidate 3 (non-protein coding)', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31173735, end_position_on_the_genomic_accession=31177899, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130890, CurrentID=0, Status='live', Symbol='TSTD3', Aliases='', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 3', other_designations='rhodanese domain-containing protein 3', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99520827, end_position_on_the_genomic_accession=99579288, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373771, CurrentID=0, Status='live', Symbol='LOC105373771', Aliases='', description='uncharacterized LOC105373771', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=180719967, end_position_on_the_genomic_accession=180724583, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130892, CurrentID=0, Status='live', Symbol='RPL7P32', Aliases='RPL7P, RPL7_15_834', description='ribosomal protein L7 pseudogene 32', other_designations='WUGSC:H_RG054D04.1', map_location='7q31.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108510213, end_position_on_the_genomic_accession=108511042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373773, CurrentID=0, Status='live', Symbol='LOC105373773', Aliases='', description='uncharacterized LOC105373773', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=182643258, end_position_on_the_genomic_accession=182644844, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130894, CurrentID=0, Status='live', Symbol='VAC14-AS1', Aliases='', description='VAC14 antisense RNA 1', other_designations='', map_location='16q22.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=70755060, end_position_on_the_genomic_accession=70773251, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105373775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373775, CurrentID=0, Status='live', Symbol='LOC105373775', Aliases='', description='uncharacterized LOC105373775', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184049140, end_position_on_the_genomic_accession=184051402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373776, CurrentID=0, Status='live', Symbol='LOC105373776', Aliases='', description='uncharacterized LOC105373776', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184278487, end_position_on_the_genomic_accession=184395388, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  122961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122961, CurrentID=0, Status='live', Symbol='ISCA2', Aliases='HBLD1, ISA2, MMDS4, c14_5557', description='iron-sulfur cluster assembly 2', other_designations='HESB-like domain-containing protein 1', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=74493720, end_position_on_the_genomic_accession=74495568, orientation='plus', exon_count=4, OMIM=615317, no_hdr0=''),
  100130898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130898, CurrentID=0, Status='live', Symbol='FLYWCH1P1', Aliases='', description='FLYWCH-type zinc finger 1 pseudogene 1', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180827942, end_position_on_the_genomic_accession=180829918, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373779, CurrentID=0, Status='live', Symbol='LOC105373779', Aliases='', description='uncharacterized LOC105373779', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184461929, end_position_on_the_genomic_accession=184467187, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373780, CurrentID=0, Status='live', Symbol='LOC105373780', Aliases='', description='proline-rich protein 18-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=184593627, end_position_on_the_genomic_accession=184598985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373781, CurrentID=0, Status='live', Symbol='LOC105373781', Aliases='', description='uncharacterized LOC105373781', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185166735, end_position_on_the_genomic_accession=185220392, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373782, CurrentID=0, Status='live', Symbol='LOC105373782', Aliases='', description='uncharacterized LOC105373782', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185011651, end_position_on_the_genomic_accession=185167113, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373783, CurrentID=0, Status='live', Symbol='LOC105373783', Aliases='', description='uncharacterized LOC105373783', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=185483053, end_position_on_the_genomic_accession=185509615, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373784, CurrentID=0, Status='live', Symbol='LOC105373784', Aliases='', description='uncharacterized LOC105373784', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186132083, end_position_on_the_genomic_accession=186161648, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373785, CurrentID=0, Status='live', Symbol='LOC105373785', Aliases='', description='uncharacterized LOC105373785', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=186421058, end_position_on_the_genomic_accession=186456754, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  122970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=122970, CurrentID=0, Status='live', Symbol='ACOT4', Aliases='PTE-Ib, PTE1B, PTE2B', description='acyl-CoA thioesterase 4', other_designations='PTE-2b|peroxisomal acyl coenzyme A thioester hydrolase Ib|peroxisomal acyl-CoA thioesterase 2B|peroxisomal long-chain acyl-CoA thioesterase Ib', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=73591706, end_position_on_the_genomic_accession=73595766, orientation='plus', exon_count=3, OMIM=614314, no_hdr0=''),
  105373787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373787, CurrentID=0, Status='live', Symbol='LOC105373787', Aliases='', description='uncharacterized LOC105373787', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187144438, end_position_on_the_genomic_accession=187149713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373788, CurrentID=0, Status='live', Symbol='LOC105373788', Aliases='', description='uncharacterized LOC105373788', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187803653, end_position_on_the_genomic_accession=187806590, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373789, CurrentID=0, Status='live', Symbol='LOC105373789', Aliases='', description='uncharacterized LOC105373789', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188596510, end_position_on_the_genomic_accession=188600686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373790, CurrentID=0, Status='live', Symbol='LOC105373790', Aliases='', description='uncharacterized LOC105373790', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188666708, end_position_on_the_genomic_accession=188751299, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373791, CurrentID=0, Status='live', Symbol='LOC105373791', Aliases='', description='uncharacterized LOC105373791', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188966458, end_position_on_the_genomic_accession=188984799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  254048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254048, CurrentID=0, Status='live', Symbol='UBN2', Aliases='', description='ubinuclein 2', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139229565, end_position_on_the_genomic_accession=139329421, orientation='plus', exon_count=19, OMIM=613841, no_hdr0=''),
  647265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647265, CurrentID=0, Status='live', Symbol='CYP2AB1P', Aliases='', description='cytochrome P450 family 2 subfamily AB member 1, pseudogene', other_designations='cytochrome P450, family 2, subfamily AB, polypeptide 1, pseudogene', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183895165, end_position_on_the_genomic_accession=183908547, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  254050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254050, CurrentID=0, Status='live', Symbol='LRRC43', Aliases='', description='leucine rich repeat containing 43', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122167719, end_position_on_the_genomic_accession=122203471, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105373795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373795, CurrentID=0, Status='live', Symbol='LOC105373795', Aliases='', description='uncharacterized LOC105373795', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189884558, end_position_on_the_genomic_accession=189921080, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100130916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130916, CurrentID=0, Status='live', Symbol='LOC100130916', Aliases='', description='HSAL5836', other_designations='', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=98880080, end_position_on_the_genomic_accession=98880355, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373797, CurrentID=0, Status='live', Symbol='LOC105373797', Aliases='', description='uncharacterized LOC105373797', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189823418, end_position_on_the_genomic_accession=189824528, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373798, CurrentID=0, Status='live', Symbol='LOC105373798', Aliases='', description='uncharacterized LOC105373798', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190281195, end_position_on_the_genomic_accession=190282710, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373799, CurrentID=0, Status='live', Symbol='LOC105373799', Aliases='', description='uncharacterized LOC105373799', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190219976, end_position_on_the_genomic_accession=190244880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130920, CurrentID=0, Status='live', Symbol='LOC100130920', Aliases='', description='uncharacterized LOC100130920', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=6608299, end_position_on_the_genomic_accession=6609178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373801, CurrentID=0, Status='live', Symbol='LOC105373801', Aliases='', description='uncharacterized LOC105373801', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190625330, end_position_on_the_genomic_accession=190649638, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100130922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130922, CurrentID=0, Status='live', Symbol='RPL22P22', Aliases='RPL22_10_1790', description='ribosomal protein L22 pseudogene 22', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=82506434, end_position_on_the_genomic_accession=82506812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130923, CurrentID=0, Status='live', Symbol='PSMA2P1', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 2 pseudogene 1', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87329356, end_position_on_the_genomic_accession=87330050, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647276, CurrentID=0, Status='live', Symbol='RPL4P4', Aliases='RPL4_2_444', description='ribosomal protein L4 pseudogene 4', other_designations='', map_location='3q27.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185417408, end_position_on_the_genomic_accession=185418843, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  154449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=154449, CurrentID=0, Status='live', Symbol='LOC154449', Aliases='', description='uncharacterized LOC154449', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170254334, end_position_on_the_genomic_accession=170262569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373806, CurrentID=0, Status='live', Symbol='LOC105373806', Aliases='', description='uncharacterized LOC105373806', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191020574, end_position_on_the_genomic_accession=191030969, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373807, CurrentID=0, Status='live', Symbol='LOC105373807', Aliases='', description='uncharacterized LOC105373807', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191615026, end_position_on_the_genomic_accession=191637497, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373808, CurrentID=0, Status='live', Symbol='LOC105373808', Aliases='', description='uncharacterized LOC105373808', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191661083, end_position_on_the_genomic_accession=191677906, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  254065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254065, CurrentID=0, Status='live', Symbol='BRWD3', Aliases='BRODL, MRX93', description='bromodomain and WD repeat domain containing 3', other_designations='bromo domain-containing protein disrupted in leukemia|novel WD repeat domain protein', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=80669488, end_position_on_the_genomic_accession=80809734, orientation='minus', exon_count=42, OMIM=300553, no_hdr0=''),
  105373810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373810, CurrentID=0, Status='live', Symbol='LOC105373810', Aliases='', description='uncharacterized LOC105373810', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191714481, end_position_on_the_genomic_accession=191720598, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373811, CurrentID=0, Status='live', Symbol='LOC105373811', Aliases='', description='uncharacterized LOC105373811', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191736165, end_position_on_the_genomic_accession=191745498, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130932, CurrentID=0, Status='live', Symbol='SNRPGP15', Aliases='', description='small nuclear ribonucleoprotein polypeptide G pseudogene 15', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14489300, end_position_on_the_genomic_accession=14489874, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130933, CurrentID=0, Status='live', Symbol='SMIM6', Aliases='C17orf110', description='small integral membrane protein 6', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75644822, end_position_on_the_genomic_accession=75647977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647286, CurrentID=0, Status='live', Symbol='RD3L', Aliases='TDRD9-AS1, TDRD9AS1', description='retinal degeneration 3-like', other_designations='TDRD9 antisense 1|TDRD9 antisense RNA 1 (non-protein coding)|retinal degeneration protein 3-like', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103940426, end_position_on_the_genomic_accession=103942308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130935, CurrentID=0, Status='live', Symbol='CSAG4', Aliases='', description='CSAG family member 4 (pseudogene)', other_designations='CSAG family, member 2 pseudogene|CSAG family, member 4 (pseudogene)|chondrosarcoma-associated gene 2/3 protein pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152733827, end_position_on_the_genomic_accession=152740985, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  647288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647288, CurrentID=0, Status='live', Symbol='CTAGE11P', Aliases='', description='CTAGE family member 11, pseudogene', other_designations='CTAGE family, member 11, pseudogene|CTAGE family, member 5 pseudogene', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75237753, end_position_on_the_genomic_accession=75240381, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373817, CurrentID=0, Status='live', Symbol='LOC105373817', Aliases='', description='uncharacterized LOC105373817', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195365827, end_position_on_the_genomic_accession=195371228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373818, CurrentID=0, Status='live', Symbol='LOC105373818', Aliases='', description='uncharacterized LOC105373818', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195569631, end_position_on_the_genomic_accession=195574469, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373819, CurrentID=0, Status='live', Symbol='LOC105373819', Aliases='', description='uncharacterized LOC105373819', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195795937, end_position_on_the_genomic_accession=195801312, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373820, CurrentID=0, Status='live', Symbol='LOC105373820', Aliases='', description='uncharacterized LOC105373820', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196155265, end_position_on_the_genomic_accession=196156431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373821, CurrentID=0, Status='live', Symbol='LOC105373821', Aliases='', description='uncharacterized LOC105373821', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196457255, end_position_on_the_genomic_accession=196475692, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373822, CurrentID=0, Status='live', Symbol='LOC105373822', Aliases='', description='uncharacterized LOC105373822', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196428253, end_position_on_the_genomic_accession=196437190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373823, CurrentID=0, Status='live', Symbol='LOC105373823', Aliases='', description='uncharacterized LOC105373823', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=196638320, end_position_on_the_genomic_accession=196639539, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373824, CurrentID=0, Status='live', Symbol='LOC105373824', Aliases='', description='uncharacterized LOC105373824', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197311045, end_position_on_the_genomic_accession=197318938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373825, CurrentID=0, Status='live', Symbol='LOC105373825', Aliases='', description='uncharacterized LOC105373825', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197247108, end_position_on_the_genomic_accession=197249055, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  647298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647298, CurrentID=0, Status='live', Symbol='HSPD1P8', Aliases='HSPD1-12P', description='heat shock protein family D (Hsp60) member 1 pseudogene 8', other_designations='heat shock 60kDa protein 1 (chaperonin) pseudogene 8', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78849102, end_position_on_the_genomic_accession=78851406, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373827, CurrentID=0, Status='live', Symbol='LOC105373827', Aliases='', description='uncharacterized LOC105373827', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197818597, end_position_on_the_genomic_accession=197820913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373828, CurrentID=0, Status='live', Symbol='LOC105373828', Aliases='', description='uncharacterized LOC105373828', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197798213, end_position_on_the_genomic_accession=197810235, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373829, CurrentID=0, Status='live', Symbol='LOC105373829', Aliases='', description='uncharacterized LOC105373829', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=198160281, end_position_on_the_genomic_accession=198165781, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647302, CurrentID=0, Status='live', Symbol='LSM3P3', Aliases='', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated pseudogene 3', other_designations='LSM3 U6 small nuclear RNA and mRNA degradation associated pseudogene 3|LSM3 homolog, U6 small nuclear RNA associated pseudogene 3|LSM3 pseudogene 3', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85101840, end_position_on_the_genomic_accession=85102699, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373831, CurrentID=0, Status='live', Symbol='LOC105373831', Aliases='', description='uncharacterized LOC105373831', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=198493308, end_position_on_the_genomic_accession=198772330, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  123016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123016, CurrentID=0, Status='live', Symbol='TTC8', Aliases='BBS8, RP51', description='tetratricopeptide repeat domain 8', other_designations='Bardet-Biedl syndrome type 8|TPR repeat protein 8', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=88824153, end_position_on_the_genomic_accession=88877996, orientation='plus', exon_count=18, OMIM=608132, no_hdr0=''),
  105373833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373833, CurrentID=0, Status='live', Symbol='LOC105373833', Aliases='', description='uncharacterized LOC105373833', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200086797, end_position_on_the_genomic_accession=200090376, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130954, CurrentID=0, Status='live', Symbol='LINC01502', Aliases='', description='long intergenic non-protein coding RNA 1502', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135574925, end_position_on_the_genomic_accession=135587112, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  647307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647307, CurrentID=0, Status='live', Symbol='PEBP1P2', Aliases='', description='phosphatidylethanolamine binding protein 1 pseudogene 2', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85341281, end_position_on_the_genomic_accession=85341838, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373836, CurrentID=0, Status='live', Symbol='LOC105373836', Aliases='', description='uncharacterized LOC105373836', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201105763, end_position_on_the_genomic_accession=201115417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647309, CurrentID=0, Status='live', Symbol='GMNC', Aliases='GEMC1', description='geminin coiled-coil domain containing', other_designations='', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190843268, end_position_on_the_genomic_accession=190862743, orientation='minus', exon_count=7, OMIM=614448, no_hdr0=''),
  100130958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130958, CurrentID=0, Status='live', Symbol='SYCE1L', Aliases='MRP2', description='synaptonemal complex central element protein 1-like', other_designations='meiosis-related protein', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=77199452, end_position_on_the_genomic_accession=77213079, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373839, CurrentID=0, Status='live', Symbol='LOC105373839', Aliases='', description='uncharacterized LOC105373839', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201754648, end_position_on_the_genomic_accession=201757293, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373840, CurrentID=0, Status='live', Symbol='LOC105373840', Aliases='', description='uncharacterized LOC105373840', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201781052, end_position_on_the_genomic_accession=201782110, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373841, CurrentID=0, Status='live', Symbol='LOC105373841', Aliases='', description='uncharacterized LOC105373841', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202055390, end_position_on_the_genomic_accession=202057408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130962, CurrentID=0, Status='live', Symbol='ATP5G1P7', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 7', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 7', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69432864, end_position_on_the_genomic_accession=69433413, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  254099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254099, CurrentID=0, Status='live', Symbol='LINC01342', Aliases='', description='long intergenic non-protein coding RNA 1342', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1137017, end_position_on_the_genomic_accession=1144056, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254100, CurrentID=0, Status='live', Symbol='SSSCA1-AS1', Aliases='', description='SSSCA1 antisense RNA 1 (head to head)', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65569220, end_position_on_the_genomic_accession=65570413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  647317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647317, CurrentID=0, Status='live', Symbol='GYG1P2', Aliases='', description='glycogenin 1 pseudogene 2', other_designations='glycogenin 1 pseudogene 3', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=82823039, end_position_on_the_genomic_accession=82824777, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254102, CurrentID=0, Status='live', Symbol='EHBP1L1', Aliases='', description='EH domain binding protein 1 like 1', other_designations='EH domain binding protein 1-like 1|tangerin', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65576038, end_position_on_the_genomic_accession=65592650, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  100130967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130967, CurrentID=0, Status='live', Symbol='C6orf99', Aliases='yR211F11.1', description='chromosome 6 open reading frame 99', other_designations='', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158869855, end_position_on_the_genomic_accession=158918005, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100130968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130968, CurrentID=0, Status='live', Symbol='RPL7L1P4', Aliases='', description='ribosomal protein L7 like 1 pseudogene 4', other_designations='ribosomal protein L7-like 1 pseudogene 4', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117778826, end_position_on_the_genomic_accession=117779590, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373849, CurrentID=0, Status='live', Symbol='LOC105373849', Aliases='', description='uncharacterized LOC105373849', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206084253, end_position_on_the_genomic_accession=206086860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130970, CurrentID=0, Status='live', Symbol='CTBP2P2', Aliases='', description='C-terminal binding protein 2 pseudogene 2', other_designations='', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=94646363, end_position_on_the_genomic_accession=94647996, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647323, CurrentID=0, Status='live', Symbol='LOC647323', Aliases='', description='uncharacterized LOC647323', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193957372, end_position_on_the_genomic_accession=194003659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  123036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123036, CurrentID=0, Status='live', Symbol='TC2N', Aliases='C14orf47, C2CD1, MTAC2D1, Tac2-N', description='tandem C2 domains, nuclear', other_designations='C2 calcium-dependent domain containing 1|membrane targeting (tandem) C2 domain containing 1|tandem C2 protein in nucleus', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=91779751, end_position_on_the_genomic_accession=91867536, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105373853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373853, CurrentID=0, Status='live', Symbol='LOC105373853', Aliases='', description='uncharacterized LOC105373853', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207238186, end_position_on_the_genomic_accession=207248433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373854, CurrentID=0, Status='live', Symbol='LOC105373854', Aliases='', description='uncharacterized LOC105373854', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208064717, end_position_on_the_genomic_accession=208077198, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373855, CurrentID=0, Status='live', Symbol='LOC105373855', Aliases='', description='uncharacterized LOC105373855', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208400918, end_position_on_the_genomic_accession=208406646, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373856, CurrentID=0, Status='live', Symbol='LOC105373856', Aliases='', description='uncharacterized LOC105373856', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208778472, end_position_on_the_genomic_accession=208793803, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  123041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123041, CurrentID=0, Status='live', Symbol='SLC24A4', Aliases='AI2A5, NCKX4, SHEP6, SLC24A2', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 4', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 4', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=92322581, end_position_on_the_genomic_accession=92501481, orientation='plus', exon_count=26, OMIM=609840, no_hdr0=''),
  105373858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373858, CurrentID=0, Status='live', Symbol='LOC105373858', Aliases='', description='uncharacterized LOC105373858', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209236049, end_position_on_the_genomic_accession=209280638, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373859, CurrentID=0, Status='live', Symbol='LOC105373859', Aliases='', description='uncharacterized LOC105373859', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=210105420, end_position_on_the_genomic_accession=210107520, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100130980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130980, CurrentID=0, Status='live', Symbol='RPS2P35', Aliases='RPS2_12_1000', description='ribosomal protein S2 pseudogene 35', other_designations='', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100116719, end_position_on_the_genomic_accession=100117674, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373861, CurrentID=0, Status='live', Symbol='LOC105373861', Aliases='', description='uncharacterized LOC105373861', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212733155, end_position_on_the_genomic_accession=212750599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373862, CurrentID=0, Status='live', Symbol='LOC105373862', Aliases='', description='uncharacterized LOC105373862', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=212870612, end_position_on_the_genomic_accession=212903415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100130983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130983, CurrentID=0, Status='live', Symbol='RPL26P28', Aliases='RPL26_8_1018', description='ribosomal protein L26 pseudogene 28', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5304380, end_position_on_the_genomic_accession=5304901, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373864, CurrentID=0, Status='live', Symbol='LOC105373864', Aliases='', description='uncharacterized LOC105373864', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=213167787, end_position_on_the_genomic_accession=213174270, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060200, CurrentID=0, Status='live', Symbol='ZNF891', Aliases='', description='zinc finger protein 891', other_designations='hCG1646157', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133119969, end_position_on_the_genomic_accession=133130473, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254122, CurrentID=0, Status='live', Symbol='SNX32', Aliases='SNX6B', description='sorting nexin 32', other_designations='sorting nexin 6B|sorting nexin-6B', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65833939, end_position_on_the_genomic_accession=65855049, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  326343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=326343, CurrentID=0, Status='live', Symbol='MT1DP', Aliases='MTM', description='metallothionein 1D, pseudogene (functional)', other_designations='metallothionein 1A pseudogene|metallothionein 1D (pseudogene)|metallothionein M', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56643687, end_position_on_the_genomic_accession=56644941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100130988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130988, CurrentID=0, Status='live', Symbol='C7orf72', Aliases='', description='chromosome 7 open reading frame 72', other_designations='', map_location='7p12.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50095861, end_position_on_the_genomic_accession=50159256, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105373869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373869, CurrentID=0, Status='live', Symbol='LINC01614', Aliases='LCAL4', description='long intergenic non-protein coding RNA 1614', other_designations='lung cancer associated lncRNA 4', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215717969, end_position_on_the_genomic_accession=215720946, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373870, CurrentID=0, Status='live', Symbol='LOC105373870', Aliases='', description='uncharacterized LOC105373870', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=215897172, end_position_on_the_genomic_accession=215912688, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373871, CurrentID=0, Status='live', Symbol='LOC105373871', Aliases='', description='uncharacterized LOC105373871', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216558248, end_position_on_the_genomic_accession=216576856, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  254128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254128, CurrentID=0, Status='live', Symbol='NIFK-AS1', Aliases='', description='NIFK antisense RNA 1', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121649654, end_position_on_the_genomic_accession=121728560, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373873, CurrentID=0, Status='live', Symbol='LOC105373873', Aliases='', description='uncharacterized LOC105373873', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217060568, end_position_on_the_genomic_accession=217066968, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130994, CurrentID=0, Status='live', Symbol='UQCRFS1P3', Aliases='', description='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide 1 pseudogene 3', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19637843, end_position_on_the_genomic_accession=19639197, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130995, CurrentID=0, Status='live', Symbol='ZNF969P', Aliases='', description='zinc finger protein 969, pseudogene', other_designations='zinc finger protein 845 pseudogene', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110416260, end_position_on_the_genomic_accession=110418504, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130996, CurrentID=0, Status='live', Symbol='YPEL5P1', Aliases='', description='YPEL5 pseudogene 1', other_designations='yippee-like 5 pseudogene 1', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182182038, end_position_on_the_genomic_accession=182183133, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373877, CurrentID=0, Status='live', Symbol='DIRC3-AS1', Aliases='', description='DIRC3 antisense RNA 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217259390, end_position_on_the_genomic_accession=217319557, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105373878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373878, CurrentID=0, Status='live', Symbol='LOC105373878', Aliases='', description='uncharacterized LOC105373878', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217978763, end_position_on_the_genomic_accession=217993768, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373879, CurrentID=0, Status='live', Symbol='LOC105373879', Aliases='', description='uncharacterized LOC105373879', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=217957314, end_position_on_the_genomic_accession=217970141, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373880, CurrentID=0, Status='live', Symbol='LOC105373880', Aliases='', description='uncharacterized LOC105373880', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218316584, end_position_on_the_genomic_accession=218318603, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373881, CurrentID=0, Status='live', Symbol='LOC105373881', Aliases='', description='uncharacterized LOC105373881', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218306389, end_position_on_the_genomic_accession=218322038, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105373882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373882, CurrentID=0, Status='live', Symbol='LOC105373882', Aliases='', description='uncharacterized LOC105373882', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218894010, end_position_on_the_genomic_accession=218897041, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373883, CurrentID=0, Status='live', Symbol='LOC105373883', Aliases='', description='uncharacterized LOC105373883', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219362506, end_position_on_the_genomic_accession=219368390, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373884, CurrentID=0, Status='live', Symbol='LOC105373884', Aliases='', description='uncharacterized LOC105373884', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219388354, end_position_on_the_genomic_accession=219403626, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373885, CurrentID=0, Status='live', Symbol='LOC105373885', Aliases='', description='uncharacterized LOC105373885', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219481732, end_position_on_the_genomic_accession=219517047, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373886, CurrentID=0, Status='live', Symbol='LOC105373886', Aliases='', description='uncharacterized LOC105373886', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219730126, end_position_on_the_genomic_accession=219737928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373887, CurrentID=0, Status='live', Symbol='LOC105373887', Aliases='', description='uncharacterized LOC105373887', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219738826, end_position_on_the_genomic_accession=219745176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373888, CurrentID=0, Status='live', Symbol='LOC105373888', Aliases='', description='uncharacterized LOC105373888', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219791732, end_position_on_the_genomic_accession=219794499, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373889, CurrentID=0, Status='live', Symbol='LOC105373889', Aliases='', description='uncharacterized LOC105373889', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219903502, end_position_on_the_genomic_accession=219904931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373890, CurrentID=0, Status='live', Symbol='LOC105373890', Aliases='', description='uncharacterized LOC105373890', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219904937, end_position_on_the_genomic_accession=219940772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373891, CurrentID=0, Status='live', Symbol='LOC105373891', Aliases='', description='uncharacterized LOC105373891', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219960660, end_position_on_the_genomic_accession=220035183, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100131012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131012, CurrentID=0, Status='live', Symbol='YBX1P4', Aliases='', description='Y-box binding protein 1 pseudogene 4', other_designations='Y box binding protein 1 pseudogene 4', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152128837, end_position_on_the_genomic_accession=152129656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373893, CurrentID=0, Status='live', Symbol='LOC105373893', Aliases='', description='uncharacterized LOC105373893', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=220105803, end_position_on_the_genomic_accession=220451110, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373894, CurrentID=0, Status='live', Symbol='LOC105373894', Aliases='', description='uncharacterized LOC105373894', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=220510650, end_position_on_the_genomic_accession=220515136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373895, CurrentID=0, Status='live', Symbol='LOC105373895', Aliases='', description='uncharacterized LOC105373895', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=220568462, end_position_on_the_genomic_accession=220634124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373896, CurrentID=0, Status='live', Symbol='LOC105373896', Aliases='', description='uncharacterized LOC105373896', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=220740285, end_position_on_the_genomic_accession=220798343, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100131017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131017, CurrentID=0, Status='live', Symbol='ZNF316', Aliases='ENST00000305834, MZF-3', description='zinc finger protein 316', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6637322, end_position_on_the_genomic_accession=6658279, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373898, CurrentID=0, Status='live', Symbol='LOC105373898', Aliases='', description='uncharacterized LOC105373898', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221655795, end_position_on_the_genomic_accession=221670487, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373899, CurrentID=0, Status='live', Symbol='LOC105373899', Aliases='', description='uncharacterized LOC105373899', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=221637529, end_position_on_the_genomic_accession=221738686, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100131020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131020, CurrentID=0, Status='live', Symbol='RPL9P23', Aliases='RPL9_5_1139', description='ribosomal protein L9 pseudogene 23', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=41536125, end_position_on_the_genomic_accession=41538986, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373901, CurrentID=0, Status='live', Symbol='LOC105373901', Aliases='', description='uncharacterized LOC105373901', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222221400, end_position_on_the_genomic_accession=222259470, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254158, CurrentID=0, Status='live', Symbol='CXorf58', Aliases='', description='chromosome X open reading frame 58', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23907932, end_position_on_the_genomic_accession=23939507, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105373903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373903, CurrentID=0, Status='live', Symbol='LOC105373903', Aliases='', description='uncharacterized LOC105373903', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222760825, end_position_on_the_genomic_accession=222777844, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373904, CurrentID=0, Status='live', Symbol='LOC105373904', Aliases='', description='uncharacterized LOC105373904', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222917339, end_position_on_the_genomic_accession=222921272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373905, CurrentID=0, Status='live', Symbol='LOC105373905', Aliases='', description='uncharacterized LOC105373905', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223102614, end_position_on_the_genomic_accession=223112389, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373906, CurrentID=0, Status='live', Symbol='LOC105373906', Aliases='', description='uncharacterized LOC105373906', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223203056, end_position_on_the_genomic_accession=223212133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131027, CurrentID=0, Status='live', Symbol='SLC25A5P4', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 4', other_designations='', map_location='13q21.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=57314665, end_position_on_the_genomic_accession=57316417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373908, CurrentID=0, Status='live', Symbol='LOC105373908', Aliases='', description='uncharacterized LOC105373908', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224199294, end_position_on_the_genomic_accession=224204044, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373909, CurrentID=0, Status='live', Symbol='LOC105373909', Aliases='', description='uncharacterized LOC105373909', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224603811, end_position_on_the_genomic_accession=224664957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373910, CurrentID=0, Status='live', Symbol='LOC105373910', Aliases='', description='uncharacterized LOC105373910', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=224678661, end_position_on_the_genomic_accession=224717757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131031, CurrentID=0, Status='live', Symbol='ANKRD18EP', Aliases='', description='ankyrin repeat domain 18E, pseudogene', other_designations='ankyrin repeat domain 18B pseudogene', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39110324, end_position_on_the_genomic_accession=39112919, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  123096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123096, CurrentID=0, Status='live', Symbol='SLC25A29', Aliases='C14orf69, CACL, ORNT3', description='solute carrier family 25 (mitochondrial carnitine/acylcarnitine carrier), member 29', other_designations='carnitine-acylcarnitine translocase like|carnitine/acylcarnitine translocase-like|mitochondrial carnitine/acylcarnitine carrier protein CACL|mitochondrial ornithine transporter 3|solute carrier family 25, member 29', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100279177, end_position_on_the_genomic_accession=100306547, orientation='minus', exon_count=10, OMIM=615064, no_hdr0=''),
  100131033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131033, CurrentID=0, Status='live', Symbol='PPIGP1', Aliases='', description='peptidylprolyl isomerase G (cyclophilin G) pseudogene 1', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=155491145, end_position_on_the_genomic_accession=155493812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254170, CurrentID=0, Status='live', Symbol='FBXO33', Aliases='BMND12, Fbx33, c14_5247', description='F-box protein 33', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39396373, end_position_on_the_genomic_accession=39432500, orientation='minus', exon_count=4, OMIM=609103, no_hdr0=''),
  123099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123099, CurrentID=0, Status='live', Symbol='DEGS2', Aliases='C14orf66, DES2, FADS8', description='delta(4)-desaturase, sphingolipid 2', other_designations='degenerative spermatocyte homolog 2, lipid desaturase|dihydroceramide desaturase 2|sphingolipid 4-desaturase|sphingolipid C4-hydroxylase/delta 4-desaturase|sphingolipid C4-monooxygenase|sphingolipid delta 4 desaturase/C-4 hydroxylase|sphingolipid delta(4)-desaturase 2|sphingolipid delta(4)-desaturase/C4-hydroxylase DES2', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100146416, end_position_on_the_genomic_accession=100166882, orientation='minus', exon_count=5, OMIM=610862, no_hdr0=''),
  100131036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131036, CurrentID=0, Status='live', Symbol='HMGB3P28', Aliases='', description='high mobility group box 3 pseudogene 28', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 28', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12450414, end_position_on_the_genomic_accession=12451072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254173, CurrentID=0, Status='live', Symbol='TTLL10', Aliases='TTLL5', description='tubulin tyrosine ligase like 10', other_designations='tubulin tyrosine ligase-like family member 10|tubulin tyrosine ligase-like family, member 10|tubulin tyrosine ligase-like family, member 5|tubulin--tyrosine ligase-like protein 10', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1173898, end_position_on_the_genomic_accession=1197935, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105373918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373918, CurrentID=0, Status='live', Symbol='LOC105373918', Aliases='', description='uncharacterized LOC105373918', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227962602, end_position_on_the_genomic_accession=228042985, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  123103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123103, CurrentID=0, Status='live', Symbol='KLHL33', Aliases='', description='kelch like family member 33', other_designations='kelch-like 33|kelch-like family member 33', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20427651, end_position_on_the_genomic_accession=20436310, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373920, CurrentID=0, Status='live', Symbol='LOC105373920', Aliases='', description='uncharacterized LOC105373920', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=228689337, end_position_on_the_genomic_accession=228716684, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373921, CurrentID=0, Status='live', Symbol='LOC105373921', Aliases='', description='uncharacterized LOC105373921', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=228580073, end_position_on_the_genomic_accession=228595870, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373922, CurrentID=0, Status='live', Symbol='LOC105373922', Aliases='', description='uncharacterized LOC105373922', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229249195, end_position_on_the_genomic_accession=229258703, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373923, CurrentID=0, Status='live', Symbol='LOC105373923', Aliases='', description='uncharacterized LOC105373923', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229610191, end_position_on_the_genomic_accession=229612542, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373924, CurrentID=0, Status='live', Symbol='LOC105373924', Aliases='', description='uncharacterized LOC105373924', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229956260, end_position_on_the_genomic_accession=229958999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373925, CurrentID=0, Status='live', Symbol='LOC105373925', Aliases='', description='uncharacterized LOC105373925', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230385337, end_position_on_the_genomic_accession=230401049, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100131046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131046, CurrentID=0, Status='live', Symbol='SALL4P3', Aliases='', description='spalt-like transcription factor 4 pseudogene 3', other_designations='sal-like 4 pseudogene 3', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20155582, end_position_on_the_genomic_accession=20158626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131047, CurrentID=0, Status='live', Symbol='LOC100131047', Aliases='', description='uncharacterized LOC100131047', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38906495, end_position_on_the_genomic_accession=38953107, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105373928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373928, CurrentID=0, Status='live', Symbol='LOC105373928', Aliases='', description='uncharacterized LOC105373928', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231611782, end_position_on_the_genomic_accession=231613585, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373929, CurrentID=0, Status='live', Symbol='LOC105373929', Aliases='', description='uncharacterized LOC105373929', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232595382, end_position_on_the_genomic_accession=232610764, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373930, CurrentID=0, Status='live', Symbol='LOC105373930', Aliases='', description='uncharacterized LOC105373930', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233357822, end_position_on_the_genomic_accession=233360222, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254187, CurrentID=0, Status='live', Symbol='TSGA10IP', Aliases='FAM161C', description='testis specific 10 interacting protein', other_designations='testis specific, 10 interacting protein|tsga10-interacting protein', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65945445, end_position_on_the_genomic_accession=65960392, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373932, CurrentID=0, Status='live', Symbol='LOC105373932', Aliases='', description='uncharacterized LOC105373932', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234200537, end_position_on_the_genomic_accession=234215305, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=64210, CurrentID=0, Status='live', Symbol='MMS19', Aliases='MET18L, hMMS19, MMS19', description='MMS19 homolog, cytosolic iron-sulfur assembly component', other_designations='MET18 homolog|MMS19 cytosolic iron-sulfur assembly component|MMS19 nucleotide excision repair homolog|MMS19-like (MET18 homolog, S. cerevisiae)|MMS19-like protein|homolog of yeast MMS19', map_location='10q24-q25', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97458324, end_position_on_the_genomic_accession=97498794, orientation='minus', exon_count=32, OMIM=614777, no_hdr0=''),
  105373934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373934, CurrentID=0, Status='live', Symbol='LOC105373934', Aliases='', description='uncharacterized LOC105373934', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234446072, end_position_on_the_genomic_accession=234454606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131055, CurrentID=0, Status='live', Symbol='RPS4XP15', Aliases='RPS4P15, RPS4X_7_1265', description='ribosomal protein S4X pseudogene 15', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='12q21.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=87944886, end_position_on_the_genomic_accession=87950818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373936, CurrentID=0, Status='live', Symbol='LOC105373936', Aliases='', description='uncharacterized LOC105373936', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234596554, end_position_on_the_genomic_accession=234632705, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373937, CurrentID=0, Status='live', Symbol='LOC105373937', Aliases='', description='uncharacterized LOC105373937', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234900827, end_position_on_the_genomic_accession=234915106, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373938, CurrentID=0, Status='live', Symbol='LOC105373938', Aliases='', description='uncharacterized LOC105373938', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234932601, end_position_on_the_genomic_accession=234951586, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373939, CurrentID=0, Status='live', Symbol='LOC105373939', Aliases='', description='uncharacterized LOC105373939', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=234976940, end_position_on_the_genomic_accession=234978778, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131060, CurrentID=0, Status='live', Symbol='LINC01135', Aliases='', description='long intergenic non-protein coding RNA 1135', other_designations='', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=58785151, end_position_on_the_genomic_accession=58899712, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373941, CurrentID=0, Status='live', Symbol='LOC105373941', Aliases='', description='uncharacterized LOC105373941', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235379050, end_position_on_the_genomic_accession=235381061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373942, CurrentID=0, Status='live', Symbol='LOC105373942', Aliases='', description='uncharacterized LOC105373942', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=235746385, end_position_on_the_genomic_accession=235790474, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  647415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647415, CurrentID=0, Status='live', Symbol='ALG1L5P', Aliases='', description='asparagine-linked glycosylation 1-like 5, pseudogene', other_designations='asparagine-linked glycosylation 1 homolog pseudogene|asparagine-linked glycosylation 1-like pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6923812, end_position_on_the_genomic_accession=6931831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373944, CurrentID=0, Status='live', Symbol='LOC105373944', Aliases='', description='uncharacterized LOC105373944', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236161366, end_position_on_the_genomic_accession=236164208, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131065, CurrentID=0, Status='live', Symbol='VN1R38P', Aliases='', description='vomeronasal 1 receptor 38 pseudogene', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64179988, end_position_on_the_genomic_accession=64180916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373946, CurrentID=0, Status='live', Symbol='LOC105373946', Aliases='', description='uncharacterized LOC105373946', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236603142, end_position_on_the_genomic_accession=236604117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131067, CurrentID=0, Status='live', Symbol='CKMT2-AS1', Aliases='', description='CKMT2 antisense RNA 1', other_designations='CTC-281B15.1', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=81237565, end_position_on_the_genomic_accession=81301569, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131068, CurrentID=0, Status='live', Symbol='ERVI-1', Aliases='', description='endogenous retrovirus group I member 1', other_designations='endogenous retrovirus group I, member 1', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373949, CurrentID=0, Status='live', Symbol='LOC105373949', Aliases='', description='uncharacterized LOC105373949', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236774139, end_position_on_the_genomic_accession=236780767, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106029310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106029310, CurrentID=0, Status='live', Symbol='DANT1', Aliases='', description='DXZ4 associated non-coding transcript 1, proximal', other_designations='DXZ4 associated noncoding transcript 1', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373951, CurrentID=0, Status='live', Symbol='LOC105373951', Aliases='', description='uncharacterized LOC105373951', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236953636, end_position_on_the_genomic_accession=236980061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373952, CurrentID=0, Status='live', Symbol='LOC105373952', Aliases='', description='uncharacterized LOC105373952', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237145833, end_position_on_the_genomic_accession=237152822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373953, CurrentID=0, Status='live', Symbol='LOC105373953', Aliases='', description='uncharacterized LOC105373953', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237213738, end_position_on_the_genomic_accession=237257644, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373954, CurrentID=0, Status='live', Symbol='LOC105373954', Aliases='', description='uncharacterized LOC105373954', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237290280, end_position_on_the_genomic_accession=237297921, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105373955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373955, CurrentID=0, Status='live', Symbol='LOC105373955', Aliases='', description='uncharacterized LOC105373955', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237421420, end_position_on_the_genomic_accession=237434822, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373956, CurrentID=0, Status='live', Symbol='LOC105373956', Aliases='', description='uncharacterized LOC105373956', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237536039, end_position_on_the_genomic_accession=237548656, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131077, CurrentID=0, Status='live', Symbol='OK/SW-CL.36', Aliases='', description='OK/SW-CL.36', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=84701688, end_position_on_the_genomic_accession=84701888, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373958, CurrentID=0, Status='live', Symbol='LOC105373958', Aliases='', description='uncharacterized LOC105373958', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237592956, end_position_on_the_genomic_accession=237595972, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373959, CurrentID=0, Status='live', Symbol='LOC105373959', Aliases='', description='uncharacterized LOC105373959', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=237676758, end_position_on_the_genomic_accession=237691882, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131080, CurrentID=0, Status='live', Symbol='LOC100131080', Aliases='', description='ribosomal protein S14 pseudogene', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=8508052, end_position_on_the_genomic_accession=8508644, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373961, CurrentID=0, Status='live', Symbol='LOC105373961', Aliases='', description='uncharacterized LOC105373961', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238298476, end_position_on_the_genomic_accession=238300271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373962, CurrentID=0, Status='live', Symbol='LOC105373962', Aliases='', description='uncharacterized LOC105373962', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=238720324, end_position_on_the_genomic_accession=238744084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131083, CurrentID=0, Status='live', Symbol='RPS8P5', Aliases='RPS8_2_374', description='ribosomal protein S8 pseudogene 5', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57530293, end_position_on_the_genomic_accession=57530973, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131084, CurrentID=0, Status='live', Symbol='DNM1P17', Aliases='DNM1DN17, DNM1DN17@', description='dynamin 1 pseudogene 17', other_designations='DNM1 duplicon 17 cluster|DNM1 pseudogene 17', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30357154, end_position_on_the_genomic_accession=30358984, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131085, CurrentID=0, Status='live', Symbol='RPL7P12', Aliases='RPL7_3_208', description='ribosomal protein L7 pseudogene 12', other_designations='', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38231548, end_position_on_the_genomic_accession=38232383, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373966, CurrentID=0, Status='live', Symbol='LOC105373966', Aliases='', description='uncharacterized LOC105373966', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=239656421, end_position_on_the_genomic_accession=239662405, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131087, CurrentID=0, Status='live', Symbol='RPLP1P11', Aliases='RPLP1_5_1519', description='ribosomal protein, large, P1 pseudogene 11', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16044492, end_position_on_the_genomic_accession=16044996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105373968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373968, CurrentID=0, Status='live', Symbol='LOC105373968', Aliases='', description='uncharacterized LOC105373968', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240252144, end_position_on_the_genomic_accession=240256455, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254225, CurrentID=0, Status='live', Symbol='RNF169', Aliases='', description='ring finger protein 169', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74748855, end_position_on_the_genomic_accession=74842413, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105373970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373970, CurrentID=0, Status='live', Symbol='LOC105373970', Aliases='', description='uncharacterized LOC105373970', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241007452, end_position_on_the_genomic_accession=241033724, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105373971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373971, CurrentID=0, Status='live', Symbol='LOC105373971', Aliases='', description='uncharacterized LOC105373971', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241167275, end_position_on_the_genomic_accession=241188369, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  254228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254228, CurrentID=0, Status='live', Symbol='FAM26E', Aliases='C6orf188, dJ493F7.3', description='family with sequence similarity 26 member E', other_designations='family with sequence similarity 26, member E', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=116511645, end_position_on_the_genomic_accession=116518546, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373973, CurrentID=0, Status='live', Symbol='LOC105373973', Aliases='', description='uncharacterized LOC105373973', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241360878, end_position_on_the_genomic_accession=241372559, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131094, CurrentID=0, Status='live', Symbol='DPP9-AS1', Aliases='', description='DPP9 antisense RNA 1', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4679282, end_position_on_the_genomic_accession=4685948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373975, CurrentID=0, Status='live', Symbol='LOC105373975', Aliases='', description='uncharacterized LOC105373975', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241574995, end_position_on_the_genomic_accession=241577204, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131096, CurrentID=0, Status='live', Symbol='TNRC6C-AS1', Aliases='', description='TNRC6C antisense RNA 1', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78107401, end_position_on_the_genomic_accession=78111799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373977, CurrentID=0, Status='live', Symbol='LOC105373977', Aliases='', description='uncharacterized LOC105373977', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241852952, end_position_on_the_genomic_accession=241856276, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105373978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373978, CurrentID=0, Status='live', Symbol='LOC105373978', Aliases='', description='uncharacterized LOC105373978', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241879041, end_position_on_the_genomic_accession=241881306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480938, CurrentID=0, Status='live', Symbol='RN7SL60P', Aliases='', description='RNA, 7SL, cytoplasmic 60, pseudogene', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105373980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373980, CurrentID=0, Status='live', Symbol='LOC105373980', Aliases='', description='uncharacterized LOC105373980', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242037224, end_position_on_the_genomic_accession=242042233, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373981, CurrentID=0, Status='live', Symbol='LOC105373981', Aliases='', description='uncharacterized LOC105373981', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=242143861, end_position_on_the_genomic_accession=242155858, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105373982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373982, CurrentID=0, Status='live', Symbol='LOC105373982', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=96869182, end_position_on_the_genomic_accession=96870390, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131103, CurrentID=0, Status='live', Symbol='TATDN2P1', Aliases='', description='TatD DNase domain containing 2 pseudogene 1', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44284066, end_position_on_the_genomic_accession=44286639, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254240, CurrentID=0, Status='live', Symbol='BPIFC', Aliases='BPIL2', description='BPI fold containing family C', other_designations='BPI-like 2|bactericidal/permeability-increasing protein-like 2', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32413845, end_position_on_the_genomic_accession=32464469, orientation='minus', exon_count=20, OMIM=614109, no_hdr0=''),
  123169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123169, CurrentID=0, Status='live', Symbol='LEO1', Aliases='RDL', description='LEO1 homolog, Paf1/RNA polymerase II complex component', other_designations='Leo1 Paf1/RNA polymerase II complex component|Leo1, Paf1/RNA polymerase II complex component, homolog|replicative senescence down-regulated leo1-like protein', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51938025, end_position_on_the_genomic_accession=51971801, orientation='minus', exon_count=13, OMIM=610507, no_hdr0=''),
  105373986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373986, CurrentID=0, Status='live', Symbol='LOC105373986', Aliases='', description='uncharacterized LOC105373986', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=110041483, end_position_on_the_genomic_accession=110055097, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100131107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131107, CurrentID=0, Status='live', Symbol='LOC100131107', Aliases='', description='putative UPF0607 protein ENSP00000383783', other_designations='UPF0607 protein ENSP00000383783-like', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152079413, end_position_on_the_genomic_accession=152080924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105373988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373988, CurrentID=0, Status='live', Symbol='LOC105373988', Aliases='', description='uncharacterized LOC105373988', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=118077311, end_position_on_the_genomic_accession=118088892, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105373989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373989, CurrentID=0, Status='live', Symbol='LOC105373989', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=120328666, end_position_on_the_genomic_accession=120346060, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373990, CurrentID=0, Status='live', Symbol='LOC105373990', Aliases='', description='trichohyalin-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=121610402, end_position_on_the_genomic_accession=121649344, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105373991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373991, CurrentID=0, Status='live', Symbol='LOC105373991', Aliases='', description='uncharacterized LOC105373991', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=236553699, end_position_on_the_genomic_accession=236569257, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105373992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373992, CurrentID=0, Status='live', Symbol='LOC105373992', Aliases='', description='uncharacterized LOC105373992', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94648076, end_position_on_the_genomic_accession=94657263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373993, CurrentID=0, Status='live', Symbol='LOC105373993', Aliases='', description='uncharacterized LOC105373993', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=95096796, end_position_on_the_genomic_accession=95152937, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105373994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373994, CurrentID=0, Status='live', Symbol='LOC105373994', Aliases='', description='uncharacterized LOC105373994', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=97834488, end_position_on_the_genomic_accession=97874588, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  254251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254251, CurrentID=0, Status='live', Symbol='LCORL', Aliases='MLR1', description='ligand dependent nuclear receptor corepressor like', other_designations='LCOR-like protein|MBLK1-related protein|ligand dependent nuclear receptor corepressor-like|transcription factor MLR1', map_location='4p15.31', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17841199, end_position_on_the_genomic_accession=18021860, orientation='minus', exon_count=11, OMIM=611799, no_hdr0=''),
  105373996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373996, CurrentID=0, Status='live', Symbol='LOC105373996', Aliases='', description='uncharacterized LOC105373996', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98147979, end_position_on_the_genomic_accession=98167850, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131117, CurrentID=0, Status='live', Symbol='ZBTB20-AS1', Aliases='', description='ZBTB20 antisense RNA 1', other_designations='ZBTB20 antisense RNA 1 (non-protein coding)', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114351811, end_position_on_the_genomic_accession=114388978, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100131118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131118, CurrentID=0, Status='live', Symbol='VN1R67P', Aliases='', description='vomeronasal 1 receptor 67 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31801606, end_position_on_the_genomic_accession=31802481, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105373999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373999, CurrentID=0, Status='live', Symbol='LOC105373999', Aliases='', description='uncharacterized LOC105373999', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98262772, end_position_on_the_genomic_accession=98314674, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374000, CurrentID=0, Status='live', Symbol='LOC105374000', Aliases='', description='uncharacterized LOC105374000', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98367878, end_position_on_the_genomic_accession=98368256, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374001, CurrentID=0, Status='live', Symbol='LOC105374001', Aliases='', description='uncharacterized LOC105374001', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98523074, end_position_on_the_genomic_accession=98526120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374002, CurrentID=0, Status='live', Symbol='LOC105374002', Aliases='', description='uncharacterized LOC105374002', other_designations='CTD-2021J15.1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98902009, end_position_on_the_genomic_accession=98905343, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374003, CurrentID=0, Status='live', Symbol='LOC105374003', Aliases='', description='uncharacterized LOC105374003', other_designations='CTD-2021J15.2', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=98981297, end_position_on_the_genomic_accession=98998402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647476, CurrentID=0, Status='live', Symbol='FRG2EP', Aliases='', description='FSHD region gene 2 family member E, pseudogene', other_designations='FSHD region gene 2 family, member C pseudogene|FSHD region gene 2 family, member E, pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29414075, end_position_on_the_genomic_accession=29416961, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374005, CurrentID=0, Status='live', Symbol='LOC105374005', Aliases='', description='uncharacterized LOC105374005', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99501758, end_position_on_the_genomic_accession=99545753, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374006, CurrentID=0, Status='live', Symbol='LOC105374006', Aliases='', description='uncharacterized LOC105374006', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99506229, end_position_on_the_genomic_accession=99514783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254263, CurrentID=0, Status='live', Symbol='CNIH2', Aliases='CNIH-2, Cnil', description='cornichon family AMPA receptor auxiliary protein 2', other_designations='cornichon homolog 2', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66278201, end_position_on_the_genomic_accession=66284214, orientation='plus', exon_count=6, OMIM=611288, no_hdr0=''),
  105374008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374008, CurrentID=0, Status='live', Symbol='LOC105374008', Aliases='', description='uncharacterized LOC105374008', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=99598064, end_position_on_the_genomic_accession=99637261, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  647481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647481, CurrentID=0, Status='live', Symbol='LOC647481', Aliases='', description='family with sequence similarity 91, member A1 pseudogene', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148434856, end_position_on_the_genomic_accession=148436924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131130, CurrentID=0, Status='live', Symbol='HNRNPA1P13', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 13', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136428832, end_position_on_the_genomic_accession=136430054, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131131, CurrentID=0, Status='live', Symbol='LOC100131131', Aliases='', description='AHPA9419', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=102737460, end_position_on_the_genomic_accession=102739005, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  254268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254268, CurrentID=0, Status='live', Symbol='AKNAD1', Aliases='C1orf62', description='AKNA domain containing 1', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108815898, end_position_on_the_genomic_accession=108858242, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105374013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374013, CurrentID=0, Status='live', Symbol='LOC105374013', Aliases='', description='endogenous retrovirus group K member 5 Gag polyprotein', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101687198, end_position_on_the_genomic_accession=101709759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374014, CurrentID=0, Status='live', Symbol='LOC105374014', Aliases='', description='uncharacterized LOC105374014', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101868674, end_position_on_the_genomic_accession=101892579, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131135, CurrentID=0, Status='live', Symbol='MTHFD2P4', Aliases='', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2, methenyltetrahydrofolate cyclohydrolase pseudogene 4', other_designations='', map_location='4q32.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=162323107, end_position_on_the_genomic_accession=162323904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254272, CurrentID=0, Status='live', Symbol='TBC1D28', Aliases='', description='TBC1 domain family member 28', other_designations='TBC1 domain family, member 28', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18634434, end_position_on_the_genomic_accession=18644433, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100131137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131137, CurrentID=0, Status='live', Symbol='BSPH1', Aliases='BSP1, ELSPBP2', description='binder of sperm protein homolog 1', other_designations='binder of sperm 1|bovine seminal plasma protein homolog 1|bovine seminal plasma protein-like 1|epididymal sperm binding protein 2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47968046, end_position_on_the_genomic_accession=47992170, orientation='minus', exon_count=7, OMIM=612213, no_hdr0=''),
  100131138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131138, CurrentID=0, Status='live', Symbol='LINC01405', Aliases='', description='long intergenic non-protein coding RNA 1405', other_designations='', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=110936602, end_position_on_the_genomic_accession=110937448, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374019, CurrentID=0, Status='live', Symbol='LOC105374019', Aliases='', description='uncharacterized LOC105374019', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104334239, end_position_on_the_genomic_accession=104336218, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374020, CurrentID=0, Status='live', Symbol='LOC105374020', Aliases='', description='uncharacterized LOC105374020', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104359464, end_position_on_the_genomic_accession=104456585, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131141, CurrentID=0, Status='live', Symbol='VN1R95P', Aliases='', description='vomeronasal 1 receptor 95 pseudogene', other_designations='', map_location='19q11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27803541, end_position_on_the_genomic_accession=27804392, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374022, CurrentID=0, Status='live', Symbol='LOC105374022', Aliases='', description='uncharacterized LOC105374022', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104773116, end_position_on_the_genomic_accession=104806604, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  123207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123207, CurrentID=0, Status='live', Symbol='C15orf40', Aliases='', description='chromosome 15 open reading frame 40', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82988308, end_position_on_the_genomic_accession=83012298, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105374024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374024, CurrentID=0, Status='live', Symbol='LOC105374024', Aliases='', description='uncharacterized LOC105374024', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=105872182, end_position_on_the_genomic_accession=105880885, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374025, CurrentID=0, Status='live', Symbol='LOC105374025', Aliases='', description='uncharacterized LOC105374025', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106035475, end_position_on_the_genomic_accession=106060321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374026, CurrentID=0, Status='live', Symbol='LOC105374026', Aliases='', description='uncharacterized LOC105374026', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106060422, end_position_on_the_genomic_accession=106070156, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374027, CurrentID=0, Status='live', Symbol='LOC105374027', Aliases='', description='uncharacterized LOC105374027', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106176883, end_position_on_the_genomic_accession=106202585, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374028, CurrentID=0, Status='live', Symbol='LOC105374028', Aliases='', description='uncharacterized LOC105374028', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106417755, end_position_on_the_genomic_accession=106428341, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374029, CurrentID=0, Status='live', Symbol='LOC105374029', Aliases='', description='uncharacterized LOC105374029', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=106684901, end_position_on_the_genomic_accession=106750844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100131150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131150, CurrentID=0, Status='live', Symbol='LOC100131150', Aliases='', description='uncharacterized LOC100131150', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=66764061, end_position_on_the_genomic_accession=66764783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131151, CurrentID=0, Status='live', Symbol='RPLP0P8', Aliases='RPLP0_3_381', description='ribosomal protein, large, P0 pseudogene 8', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=68633495, end_position_on_the_genomic_accession=68634399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374032, CurrentID=0, Status='live', Symbol='LOC105374032', Aliases='', description='uncharacterized LOC105374032', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108222610, end_position_on_the_genomic_accession=108293706, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374033, CurrentID=0, Status='live', Symbol='LOC105374033', Aliases='', description='uncharacterized LOC105374033', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108713549, end_position_on_the_genomic_accession=108727630, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374034, CurrentID=0, Status='live', Symbol='LOC105374034', Aliases='', description='uncharacterized LOC105374034', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=108742490, end_position_on_the_genomic_accession=108748179, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  647507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647507, CurrentID=0, Status='live', Symbol='CNN2P4', Aliases='CNN2P5', description='calponin 2 pseudogene 4', other_designations='calponin 2 pseudogene 5', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42971169, end_position_on_the_genomic_accession=42973327, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374036, CurrentID=0, Status='live', Symbol='LOC105374036', Aliases='', description='uncharacterized LOC105374036', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=110447235, end_position_on_the_genomic_accession=110471580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374037, CurrentID=0, Status='live', Symbol='LOC105374037', Aliases='', description='uncharacterized LOC105374037', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=110638487, end_position_on_the_genomic_accession=110667287, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374038, CurrentID=0, Status='live', Symbol='LOC105374038', Aliases='', description='uncharacterized LOC105374038', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111270922, end_position_on_the_genomic_accession=111275718, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254295, CurrentID=0, Status='live', Symbol='PHYHD1', Aliases='', description='phytanoyl-CoA dioxygenase domain containing 1', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128920895, end_position_on_the_genomic_accession=128942041, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100131160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131160, CurrentID=0, Status='live', Symbol='RPL39P20', Aliases='RPL39_10_584', description='ribosomal protein L39 pseudogene 20', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80315622, end_position_on_the_genomic_accession=80316018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131161, CurrentID=0, Status='live', Symbol='RPSAP39', Aliases='RPSA_12_495', description='ribosomal protein SA pseudogene 39', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80161053, end_position_on_the_genomic_accession=80162081, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374042, CurrentID=0, Status='live', Symbol='LOC105374042', Aliases='', description='uncharacterized LOC105374042', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112301477, end_position_on_the_genomic_accession=112329779, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254299, CurrentID=0, Status='live', Symbol='LINC00677', Aliases='C14orf74', description='long intergenic non-protein coding RNA 677', other_designations='', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  123228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123228, CurrentID=0, Status='live', Symbol='SENP8', Aliases='DEN1, NEDP1, PRSC2', description='SUMO/sentrin peptidase family member, NEDD8 specific', other_designations='NEDD8 specific-protease cysteine 2|NEDD8-specific protease 1|SUMO sentrin specific protease family member 8|SUMO/sentrin specific peptidase family member 8|deneddylase 1|deneddylase-1|protease, cysteine 2', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72114258, end_position_on_the_genomic_accession=72140970, orientation='plus', exon_count=5, OMIM=608659, no_hdr0=''),
  105374045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374045, CurrentID=0, Status='live', Symbol='LOC105374045', Aliases='', description='uncharacterized LOC105374045', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113263091, end_position_on_the_genomic_accession=113268292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374046, CurrentID=0, Status='live', Symbol='LOC105374046', Aliases='', description='uncharacterized LOC105374046', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113223637, end_position_on_the_genomic_accession=113241939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374047, CurrentID=0, Status='live', Symbol='LOC105374047', Aliases='', description='uncharacterized LOC105374047', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113729782, end_position_on_the_genomic_accession=113734132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374048, CurrentID=0, Status='live', Symbol='LOC105374048', Aliases='', description='uncharacterized LOC105374048', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114001691, end_position_on_the_genomic_accession=114016296, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374049, CurrentID=0, Status='live', Symbol='LOC105374049', Aliases='', description='uncharacterized LOC105374049', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114730204, end_position_on_the_genomic_accession=114745735, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131170, CurrentID=0, Status='live', Symbol='LOC100131170', Aliases='', description='uncharacterized LOC100131170', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.9', start_position_on_the_genomic_accession=76700600, end_position_on_the_genomic_accession=76702250, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929019, CurrentID=0, Status='live', Symbol='LOC101929019', Aliases='', description='uncharacterized LOC101929019', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=12223451, end_position_on_the_genomic_accession=12251286, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374052, CurrentID=0, Status='live', Symbol='LOC105374052', Aliases='', description='uncharacterized LOC105374052', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=116170818, end_position_on_the_genomic_accession=116180516, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131173, CurrentID=0, Status='live', Symbol='RPS29P12', Aliases='RPS29_9_646', description='ribosomal protein S29 pseudogene 12', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180945699, end_position_on_the_genomic_accession=180945993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374054, CurrentID=0, Status='live', Symbol='LOC105374054', Aliases='', description='uncharacterized LOC105374054', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117239217, end_position_on_the_genomic_accession=117259548, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374055, CurrentID=0, Status='live', Symbol='LOC105374055', Aliases='', description='uncharacterized LOC105374055', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117358144, end_position_on_the_genomic_accession=117371737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131176, CurrentID=0, Status='live', Symbol='WDR86-AS1', Aliases='', description='WDR86 antisense RNA 1', other_designations='WDR86 antisense RNA 1 (non-protein coding)', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151409161, end_position_on_the_genomic_accession=151413354, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374057, CurrentID=0, Status='live', Symbol='LOC105374057', Aliases='', description='uncharacterized LOC105374057', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=117678693, end_position_on_the_genomic_accession=117690978, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374058, CurrentID=0, Status='live', Symbol='LOC105374058', Aliases='', description='uncharacterized LOC105374058', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118488800, end_position_on_the_genomic_accession=118493856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374059, CurrentID=0, Status='live', Symbol='LOC105374059', Aliases='', description='uncharacterized LOC105374059', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=118771421, end_position_on_the_genomic_accession=118773172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647532, CurrentID=0, Status='live', Symbol='FARSBP1', Aliases='', description='phenylalanyl-tRNA synthetase beta subunit pseudogene 1', other_designations='phenylalanine-tRNA synthetase-like, beta subunit pseudogene|phenylalanyl-tRNA synthetase, beta subunit pseudogene 1', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80776044, end_position_on_the_genomic_accession=80778077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374061, CurrentID=0, Status='live', Symbol='LOC105374061', Aliases='', description='uncharacterized LOC105374061', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119399728, end_position_on_the_genomic_accession=119401814, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374062, CurrentID=0, Status='live', Symbol='LOC105374062', Aliases='', description='uncharacterized LOC105374062', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119599213, end_position_on_the_genomic_accession=119606055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131183, CurrentID=0, Status='live', Symbol='HMGN2P30', Aliases='', description='high mobility group nucleosomal binding domain 2 pseudogene 30', other_designations='high-mobility group nucleosomal binding domain 2 pseudogene 30', map_location='7p14.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42113251, end_position_on_the_genomic_accession=42114451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374064, CurrentID=0, Status='live', Symbol='LOC105374064', Aliases='', description='uncharacterized LOC105374064', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=119699868, end_position_on_the_genomic_accession=119702921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374065, CurrentID=0, Status='live', Symbol='LOC105374065', Aliases='', description='uncharacterized LOC105374065', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120145695, end_position_on_the_genomic_accession=120185355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374066, CurrentID=0, Status='live', Symbol='LOC105374066', Aliases='', description='uncharacterized LOC105374066', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=120285319, end_position_on_the_genomic_accession=120289196, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131187, CurrentID=0, Status='live', Symbol='TSTD1', Aliases='KAT', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 1', other_designations='putative thiosulfate sulfurtransferase KAT|thiosulfate sulfurtransferase KAT, putative', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161037631, end_position_on_the_genomic_accession=161038984, orientation='minus', exon_count=4, OMIM=616041, no_hdr0=''),
  100131188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131188, CurrentID=0, Status='live', Symbol='LIPT1P1', Aliases='', description='lipoyltransferase 1 pseudogene 1', other_designations='', map_location='13q33.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=102002130, end_position_on_the_genomic_accession=102003296, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421694, CurrentID=0, Status='live', Symbol='GLULP5', Aliases='', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 5', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114104934, end_position_on_the_genomic_accession=114105922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374070, CurrentID=0, Status='live', Symbol='LOC105374070', Aliases='', description='uncharacterized LOC105374070', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=121569624, end_position_on_the_genomic_accession=121601404, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374071, CurrentID=0, Status='live', Symbol='LOC105374071', Aliases='', description='uncharacterized LOC105374071', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=122514966, end_position_on_the_genomic_accession=122531898, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374072, CurrentID=0, Status='live', Symbol='LOC105374072', Aliases='', description='uncharacterized LOC105374072', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123016462, end_position_on_the_genomic_accession=123023168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131193, CurrentID=0, Status='live', Symbol='CCDC183-AS1', Aliases='KIAA1984-AS1', description='CCDC183 antisense RNA 1', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136803927, end_position_on_the_genomic_accession=136808848, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374074, CurrentID=0, Status='live', Symbol='LOC105374074', Aliases='', description='uncharacterized LOC105374074', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=123246444, end_position_on_the_genomic_accession=123277011, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131195, CurrentID=0, Status='live', Symbol='LOC100131195', Aliases='', description='uncharacterized LOC100131195', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=43486731, end_position_on_the_genomic_accession=43487111, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131196, CurrentID=0, Status='live', Symbol='RPS25P8', Aliases='RPS25_6_1012', description='ribosomal protein S25 pseudogene 8', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123029332, end_position_on_the_genomic_accession=123029817, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374077, CurrentID=0, Status='live', Symbol='LOC105374077', Aliases='', description='uncharacterized LOC105374077', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124168178, end_position_on_the_genomic_accession=124171712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374078, CurrentID=0, Status='live', Symbol='LOC105374078', Aliases='', description='uncharacterized LOC105374078', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=124812018, end_position_on_the_genomic_accession=124827700, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  123263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123263, CurrentID=0, Status='live', Symbol='MTFMT', Aliases='COXPD15, FMT1', description='mitochondrial methionyl-tRNA formyltransferase', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65001512, end_position_on_the_genomic_accession=65029639, orientation='minus', exon_count=10, OMIM=611766, no_hdr0=''),
  123264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123264, CurrentID=0, Status='live', Symbol='SLC51B', Aliases='OSTB, OSTBETA', description='solute carrier family 51 beta subunit', other_designations='OST-beta|organic solute transporter beta subunit|solute carrier family 51 subunit beta|solute carrier family 51, beta subunit', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65045370, end_position_on_the_genomic_accession=65053396, orientation='plus', exon_count=5, OMIM=612085, no_hdr0=''),
  105374081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374081, CurrentID=0, Status='live', Symbol='LOC105374081', Aliases='', description='uncharacterized LOC105374081', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125414839, end_position_on_the_genomic_accession=125425880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374082, CurrentID=0, Status='live', Symbol='LOC105374082', Aliases='', description='uncharacterized LOC105374082', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125904962, end_position_on_the_genomic_accession=125908478, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374083, CurrentID=0, Status='live', Symbol='LOC105374083', Aliases='', description='uncharacterized LOC105374083', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126157523, end_position_on_the_genomic_accession=126170467, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374084, CurrentID=0, Status='live', Symbol='LOC105374084', Aliases='', description='uncharacterized LOC105374084', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126094483, end_position_on_the_genomic_accession=126103543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131205, CurrentID=0, Status='live', Symbol='RPL21P28', Aliases='RPL21_8_160', description='ribosomal protein L21 pseudogene 28', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212051486, end_position_on_the_genomic_accession=212052051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131206, CurrentID=0, Status='live', Symbol='RPL29P16', Aliases='RPL29_9_696', description='ribosomal protein L29 pseudogene 16', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44089125, end_position_on_the_genomic_accession=44089763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131207, CurrentID=0, Status='live', Symbol='DNAJA1P2', Aliases='', description='DnaJ heat shock protein family (Hsp40) member A1 pseudogene 2', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1 pseudogene 2', map_location='2q23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=153761288, end_position_on_the_genomic_accession=153762796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131208, CurrentID=0, Status='live', Symbol='SNAP25-AS1', Aliases='', description='SNAP25 antisense RNA 1', other_designations='SNAP25 antisense RNA 1 (non-protein coding)', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10023812, end_position_on_the_genomic_accession=10219506, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374089, CurrentID=0, Status='live', Symbol='LOC105374089', Aliases='', description='uncharacterized LOC105374089', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126432796, end_position_on_the_genomic_accession=126435544, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374090, CurrentID=0, Status='live', Symbol='LOC105374090', Aliases='', description='uncharacterized LOC105374090', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126559013, end_position_on_the_genomic_accession=126604715, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131211, CurrentID=0, Status='live', Symbol='NEMP2', Aliases='TMEM194B', description='nuclear envelope integral membrane protein 2', other_designations='UPF0571 transmembrane protein|transmembrane protein 194B', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=190454092, end_position_on_the_genomic_accession=190543625, orientation='minus', exon_count=15, OMIM=616497, no_hdr0=''),
  105374092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374092, CurrentID=0, Status='live', Symbol='LOC105374092', Aliases='', description='uncharacterized LOC105374092', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127038469, end_position_on_the_genomic_accession=127043436, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131213, CurrentID=0, Status='live', Symbol='ZNF503-AS2', Aliases='C10orf41, NCRNA00245', description='ZNF503 antisense RNA 2', other_designations='ZNF503 antisense RNA 2 (non-protein coding)', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75401528, end_position_on_the_genomic_accession=75408982, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374094, CurrentID=0, Status='live', Symbol='LOC105374094', Aliases='', description='uncharacterized LOC105374094', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127384654, end_position_on_the_genomic_accession=127388831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374095, CurrentID=0, Status='live', Symbol='LOC105374095', Aliases='', description='uncharacterized LOC105374095', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127746937, end_position_on_the_genomic_accession=127749473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374096, CurrentID=0, Status='live', Symbol='LOC105374096', Aliases='', description='uncharacterized LOC105374096', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=127855800, end_position_on_the_genomic_accession=127859263, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  647569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647569, CurrentID=0, Status='live', Symbol='PFN1P12', Aliases='', description='profilin 1 pseudogene 12', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143619988, end_position_on_the_genomic_accession=143620767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374098, CurrentID=0, Status='live', Symbol='LOC105374098', Aliases='', description='uncharacterized LOC105374098', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128171999, end_position_on_the_genomic_accession=128194462, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  123283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123283, CurrentID=0, Status='live', Symbol='TARSL2', Aliases='', description='threonyl-tRNA synthetase-like 2', other_designations='probable threonyl-tRNA synthetase 2, cytoplasmic|thrRS|threonine--tRNA ligase|threonyl-tRNA synthetase|threonyl-tRNA synthetase-like protein 2', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101653752, end_position_on_the_genomic_accession=101724442, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  105374100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374100, CurrentID=0, Status='live', Symbol='LOC105374100', Aliases='', description='uncharacterized LOC105374100', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128999962, end_position_on_the_genomic_accession=129001545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374101, CurrentID=0, Status='live', Symbol='LOC105374101', Aliases='', description='uncharacterized LOC105374101', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129184028, end_position_on_the_genomic_accession=129196302, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374102, CurrentID=0, Status='live', Symbol='LOC105374102', Aliases='', description='uncharacterized LOC105374102', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129230742, end_position_on_the_genomic_accession=129236086, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254359, CurrentID=0, Status='live', Symbol='ZDHHC24', Aliases='', description='zinc finger, DHHC-type containing 24', other_designations='DHHC-24|zinc finger DHHC domain-containing protein 24|zinc finger, DHHC domain containing 24', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66521379, end_position_on_the_genomic_accession=66546200, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374104, CurrentID=0, Status='live', Symbol='LOC105374104', Aliases='', description='uncharacterized LOC105374104', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130111795, end_position_on_the_genomic_accession=130120569, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100421700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421700, CurrentID=0, Status='live', Symbol='BNIP3P29', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 29', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22065828, end_position_on_the_genomic_accession=22066398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374106, CurrentID=0, Status='live', Symbol='LOC105374106', Aliases='', description='uncharacterized LOC105374106', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130274909, end_position_on_the_genomic_accession=130276798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374107, CurrentID=0, Status='live', Symbol='LOC105374107', Aliases='', description='uncharacterized LOC105374107', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130813723, end_position_on_the_genomic_accession=130836545, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374108, CurrentID=0, Status='live', Symbol='LOC105374108', Aliases='', description='uncharacterized LOC105374108', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130844840, end_position_on_the_genomic_accession=130846883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131229, CurrentID=0, Status='live', Symbol='RPL6P17', Aliases='RPL6_4_660', description='ribosomal protein L6 pseudogene 17', other_designations='', map_location='6p23', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15102946, end_position_on_the_genomic_accession=15104211, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131230, CurrentID=0, Status='live', Symbol='RPL30P4', Aliases='RPL30_3_359', description='ribosomal protein L30 pseudogene 4', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32635165, end_position_on_the_genomic_accession=32635639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374111, CurrentID=0, Status='live', Symbol='LOC105374111', Aliases='', description='uncharacterized LOC105374111', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132047045, end_position_on_the_genomic_accession=132083115, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131232, CurrentID=0, Status='live', Symbol='LOC100131232', Aliases='', description='uncharacterized LOC100131232', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.8', start_position_on_the_genomic_accession=65559640, end_position_on_the_genomic_accession=65560235, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374113, CurrentID=0, Status='live', Symbol='LOC105374113', Aliases='', description='uncharacterized LOC105374113', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131802703, end_position_on_the_genomic_accession=131872160, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100131234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131234, CurrentID=0, Status='live', Symbol='MIR181A1HG', Aliases='', description='MIR181A1 host gene', other_designations='MIR181A1 host gene (non-protein coding)|familial acute myelogenous leukemia related factor', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198808003, end_position_on_the_genomic_accession=198937429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374115, CurrentID=0, Status='live', Symbol='LOC105374115', Aliases='', description='uncharacterized LOC105374115', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=132802394, end_position_on_the_genomic_accession=132804631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131236, CurrentID=0, Status='live', Symbol='PCBP2P3', Aliases='', description='poly(rC) binding protein 2 pseudogene 3', other_designations='', map_location='5q14.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91843687, end_position_on_the_genomic_accession=91844528, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647589, CurrentID=0, Status='live', Symbol='ANHX', Aliases='', description='anomalous homeobox', other_designations='hCG2007354', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=133218312, end_position_on_the_genomic_accession=133236220, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105374118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374118, CurrentID=0, Status='live', Symbol='LOC105374118', Aliases='', description='uncharacterized LOC105374118', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134325021, end_position_on_the_genomic_accession=134327299, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374119, CurrentID=0, Status='live', Symbol='LOC105374119', Aliases='', description='uncharacterized LOC105374119', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134312189, end_position_on_the_genomic_accession=134321167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374120, CurrentID=0, Status='live', Symbol='LOC105374120', Aliases='', description='uncharacterized LOC105374120', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135108500, end_position_on_the_genomic_accession=135125259, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374121, CurrentID=0, Status='live', Symbol='LOC105374121', Aliases='', description='uncharacterized LOC105374121', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=134851757, end_position_on_the_genomic_accession=134854520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374122, CurrentID=0, Status='live', Symbol='LOC105374122', Aliases='', description='uncharacterized LOC105374122', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135354814, end_position_on_the_genomic_accession=135516374, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374123, CurrentID=0, Status='live', Symbol='LOC105374123', Aliases='', description='uncharacterized LOC105374123', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=135782616, end_position_on_the_genomic_accession=135793932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131244, CurrentID=0, Status='live', Symbol='ANKRD63', Aliases='', description='ankyrin repeat domain 63', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40278369, end_position_on_the_genomic_accession=40282987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374125, CurrentID=0, Status='live', Symbol='LOC105374125', Aliases='', description='uncharacterized LOC105374125', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136331972, end_position_on_the_genomic_accession=136338003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131246, CurrentID=0, Status='live', Symbol='RPL9P27', Aliases='RPL9_12_1427', description='ribosomal protein L9 pseudogene 27', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40480651, end_position_on_the_genomic_accession=40481350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374127, CurrentID=0, Status='live', Symbol='LOC105374127', Aliases='', description='uncharacterized LOC105374127', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138017340, end_position_on_the_genomic_accession=138023661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374128, CurrentID=0, Status='live', Symbol='LOC105374128', Aliases='', description='uncharacterized LOC105374128', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138046372, end_position_on_the_genomic_accession=138058052, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374129, CurrentID=0, Status='live', Symbol='LOC105374129', Aliases='', description='uncharacterized LOC105374129', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=138410305, end_position_on_the_genomic_accession=138415087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131250, CurrentID=0, Status='live', Symbol='SPCS2P1', Aliases='SPCS2P', description='signal peptidase complex subunit 2 homolog (S. cerevisiae) pseudogene 1', other_designations='', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43131019, end_position_on_the_genomic_accession=43131687, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374131, CurrentID=0, Status='live', Symbol='LOC105374131', Aliases='', description='uncharacterized LOC105374131', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139422191, end_position_on_the_genomic_accession=139436609, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374132, CurrentID=0, Status='live', Symbol='LOC105374132', Aliases='', description='uncharacterized LOC105374132', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=140449385, end_position_on_the_genomic_accession=140474559, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374133, CurrentID=0, Status='live', Symbol='LOC105374133', Aliases='', description='uncharacterized LOC105374133', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141116700, end_position_on_the_genomic_accession=141131164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374134, CurrentID=0, Status='live', Symbol='LOC105374134', Aliases='', description='uncharacterized LOC105374134', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141851579, end_position_on_the_genomic_accession=141861140, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421705, CurrentID=0, Status='live', Symbol='BNIP3P20', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 20', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20351508, end_position_on_the_genomic_accession=20352077, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374136, CurrentID=0, Status='live', Symbol='LOC105374136', Aliases='', description='uncharacterized LOC105374136', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142596530, end_position_on_the_genomic_accession=142656898, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374137, CurrentID=0, Status='live', Symbol='LOC105374137', Aliases='', description='uncharacterized LOC105374137', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=143122502, end_position_on_the_genomic_accession=143132846, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254394, CurrentID=0, Status='live', Symbol='MCM9', Aliases='C6orf61, MCMDC1, ODG4, dJ329L24.1, dJ329L24.3', description='minichromosome maintenance 9 homologous recombination repair factor', other_designations='DNA replication licensing factor MCM9|mini-chromosome maintenance deficient domain-containing protein 1|minichromosome maintenance complex component 9', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118813449, end_position_on_the_genomic_accession=118935162, orientation='minus', exon_count=13, OMIM=610098, no_hdr0=''),
  105374139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374139, CurrentID=0, Status='live', Symbol='LOC105374139', Aliases='', description='uncharacterized LOC105374139', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=144375634, end_position_on_the_genomic_accession=144379498, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374140, CurrentID=0, Status='live', Symbol='LOC105374140', Aliases='', description='uncharacterized LOC105374140', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=144416106, end_position_on_the_genomic_accession=144475176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131261, CurrentID=0, Status='live', Symbol='PNRC2P1', Aliases='', description='proline rich nuclear receptor coactivator 2 pseudogene 1', other_designations='proline-rich nuclear receptor coactivator 2 pseudogene 1', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117776523, end_position_on_the_genomic_accession=117778909, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  254398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254398, CurrentID=0, Status='live', Symbol='ARHGAP42P4', Aliases='', description='Rho GTPase activating protein 42 pseudogene 4', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19688300, end_position_on_the_genomic_accession=19691196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374143, CurrentID=0, Status='live', Symbol='LOC105374143', Aliases='', description='uncharacterized LOC105374143', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145569268, end_position_on_the_genomic_accession=145573597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374144, CurrentID=0, Status='live', Symbol='LOC105374144', Aliases='', description='uncharacterized LOC105374144', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145625515, end_position_on_the_genomic_accession=145650558, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  645691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645691, CurrentID=0, Status='live', Symbol='HNRNPA1P35', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 35', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201063026, end_position_on_the_genomic_accession=201064248, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374146, CurrentID=0, Status='live', Symbol='LOC105374146', Aliases='', description='uncharacterized LOC105374146', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148193120, end_position_on_the_genomic_accession=148220595, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374147, CurrentID=0, Status='live', Symbol='LOC105374147', Aliases='', description='uncharacterized LOC105374147', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148384766, end_position_on_the_genomic_accession=148409741, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131268, CurrentID=0, Status='live', Symbol='TPT1P1', Aliases='TPT1P', description='tumor protein, translationally-controlled 1 pseudogene 1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=31840248, end_position_on_the_genomic_accession=31841055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374149, CurrentID=0, Status='live', Symbol='LOC105374149', Aliases='', description='uncharacterized LOC105374149', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148280993, end_position_on_the_genomic_accession=148401075, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374150, CurrentID=0, Status='live', Symbol='LOC105374150', Aliases='', description='uncharacterized LOC105374150', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=148440083, end_position_on_the_genomic_accession=148464985, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105374151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374151, CurrentID=0, Status='live', Symbol='LOC105374151', Aliases='', description='uncharacterized LOC105374151', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149392726, end_position_on_the_genomic_accession=149404785, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374152, CurrentID=0, Status='live', Symbol='LOC105374152', Aliases='', description='uncharacterized LOC105374152', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149952408, end_position_on_the_genomic_accession=149959859, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374153, CurrentID=0, Status='live', Symbol='LOC105374153', Aliases='', description='uncharacterized LOC105374153', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150039221, end_position_on_the_genomic_accession=150040741, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374154, CurrentID=0, Status='live', Symbol='LOC105374154', Aliases='', description='uncharacterized LOC105374154', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150275938, end_position_on_the_genomic_accession=150281089, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131275, CurrentID=0, Status='live', Symbol='NSAP11', Aliases='', description='nervous system abundant protein 11', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=18578415, end_position_on_the_genomic_accession=18591915, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374156, CurrentID=0, Status='live', Symbol='LOC105374156', Aliases='', description='protein FAM188B2-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150871019, end_position_on_the_genomic_accession=150878949, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374157, CurrentID=0, Status='live', Symbol='LOC105374157', Aliases='', description='uncharacterized LOC105374157', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151619929, end_position_on_the_genomic_accession=151625992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374158, CurrentID=0, Status='live', Symbol='LOC105374158', Aliases='', description='uncharacterized LOC105374158', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151628574, end_position_on_the_genomic_accession=151657988, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100421709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421709, CurrentID=0, Status='live', Symbol='BNIP3P38', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 38', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23236318, end_position_on_the_genomic_accession=23236851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374160, CurrentID=0, Status='live', Symbol='LOC105374160', Aliases='', description='uncharacterized LOC105374160', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=151796280, end_position_on_the_genomic_accession=151806085, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100131281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131281, CurrentID=0, Status='live', Symbol='GLUD1P9', Aliases='', description='glutamate dehydrogenase 1 pseudogene 9', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111230007, end_position_on_the_genomic_accession=111231330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  123346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123346, CurrentID=0, Status='live', Symbol='HIGD2B', Aliases='HIGD2BP', description='HIG1 hypoxia inducible domain family member 2B', other_designations='HIG1 domain family, member 2B pseudogene|HIG1 hypoxia inducible domain family, member 2B (pseudogene)', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72675782, end_position_on_the_genomic_accession=72686149, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374163, CurrentID=0, Status='live', Symbol='LOC105374163', Aliases='', description='uncharacterized LOC105374163', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=152495381, end_position_on_the_genomic_accession=152496854, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374164, CurrentID=0, Status='live', Symbol='LOC105374164', Aliases='', description='uncharacterized LOC105374164', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153203843, end_position_on_the_genomic_accession=153248505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374165, CurrentID=0, Status='live', Symbol='LOC105374165', Aliases='', description='uncharacterized LOC105374165', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=153383526, end_position_on_the_genomic_accession=153980105, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100131286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131286, CurrentID=0, Status='live', Symbol='RPS19P5', Aliases='RPS19_1_737', description='ribosomal protein S19 pseudogene 5', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110562175, end_position_on_the_genomic_accession=110569363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374167, CurrentID=0, Status='live', Symbol='LOC105374167', Aliases='', description='uncharacterized LOC105374167', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154251013, end_position_on_the_genomic_accession=154253108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374168, CurrentID=0, Status='live', Symbol='LOC105374168', Aliases='', description='uncharacterized LOC105374168', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154258274, end_position_on_the_genomic_accession=154274314, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374169, CurrentID=0, Status='live', Symbol='LOC105374169', Aliases='', description='uncharacterized LOC105374169', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154795869, end_position_on_the_genomic_accession=154797774, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374170, CurrentID=0, Status='live', Symbol='LOC105374170', Aliases='', description='uncharacterized LOC105374170', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=154863718, end_position_on_the_genomic_accession=154887414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  123355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123355, CurrentID=0, Status='live', Symbol='LRRC28', Aliases='', description='leucine rich repeat containing 28', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99251362, end_position_on_the_genomic_accession=99386309, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  254428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254428, CurrentID=0, Status='live', Symbol='SLC41A1', Aliases='MgtE', description='solute carrier family 41 (magnesium transporter), member 1', other_designations='solute carrier family 41, member 1', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205789093, end_position_on_the_genomic_accession=205813759, orientation='minus', exon_count=12, OMIM=610801, no_hdr0=''),
  105374173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374173, CurrentID=0, Status='live', Symbol='LOC105374173', Aliases='', description='uncharacterized LOC105374173', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155512246, end_position_on_the_genomic_accession=155519741, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374174, CurrentID=0, Status='live', Symbol='LOC105374174', Aliases='', description='uncharacterized LOC105374174', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155854527, end_position_on_the_genomic_accession=155862815, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374175, CurrentID=0, Status='live', Symbol='LOC105374175', Aliases='', description='uncharacterized LOC105374175', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156475448, end_position_on_the_genomic_accession=156482466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374176, CurrentID=0, Status='live', Symbol='LOC105374176', Aliases='', description='uncharacterized LOC105374176', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156448200, end_position_on_the_genomic_accession=156457009, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374177, CurrentID=0, Status='live', Symbol='LOC105374177', Aliases='', description='uncharacterized LOC105374177', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157082833, end_position_on_the_genomic_accession=157088523, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374178, CurrentID=0, Status='live', Symbol='LOC105374178', Aliases='', description='uncharacterized LOC105374178', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157242121, end_position_on_the_genomic_accession=157251022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374179, CurrentID=0, Status='live', Symbol='LOC105374179', Aliases='', description='uncharacterized LOC105374179', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157234309, end_position_on_the_genomic_accession=157242111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374180, CurrentID=0, Status='live', Symbol='LOC105374180', Aliases='', description='uncharacterized LOC105374180', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158702581, end_position_on_the_genomic_accession=158704381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374181, CurrentID=0, Status='live', Symbol='LOC105374181', Aliases='', description='uncharacterized LOC105374181', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=158866998, end_position_on_the_genomic_accession=158884636, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131302, CurrentID=0, Status='live', Symbol='PTCD2P1', Aliases='', description='pentatricopeptide repeat domain 2 pseudogene 1', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20524289, end_position_on_the_genomic_accession=20525415, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  254439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254439, CurrentID=0, Status='live', Symbol='C11orf86', Aliases='', description='chromosome 11 open reading frame 86', other_designations='', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66975283, end_position_on_the_genomic_accession=66977008, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374184, CurrentID=0, Status='live', Symbol='LOC105374184', Aliases='', description='uncharacterized LOC105374184', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159706903, end_position_on_the_genomic_accession=159763214, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105374185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374185, CurrentID=0, Status='live', Symbol='LOC105374185', Aliases='', description='uncharacterized LOC105374185', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161030593, end_position_on_the_genomic_accession=161042495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131306, CurrentID=0, Status='live', Symbol='BNIP3P21', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 21', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20382406, end_position_on_the_genomic_accession=20382978, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374187, CurrentID=0, Status='live', Symbol='LOC105374187', Aliases='', description='uncharacterized LOC105374187', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=161269246, end_position_on_the_genomic_accession=161312003, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374188, CurrentID=0, Status='live', Symbol='LOC105374188', Aliases='', description='uncharacterized LOC105374188', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=163304191, end_position_on_the_genomic_accession=163381162, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131309, CurrentID=0, Status='live', Symbol='PRKRIRP9', Aliases='', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor) pseudogene 9', other_designations='', map_location='4p12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=45323745, end_position_on_the_genomic_accession=45324751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647662, CurrentID=0, Status='live', Symbol='FLJ40039', Aliases='', description='uncharacterized LOC647662', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374191, CurrentID=0, Status='live', Symbol='LOC105374191', Aliases='', description='uncharacterized LOC105374191', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=164450709, end_position_on_the_genomic_accession=164620475, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374192, CurrentID=0, Status='live', Symbol='LOC105374192', Aliases='', description='uncharacterized LOC105374192', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=165072027, end_position_on_the_genomic_accession=165083824, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374193, CurrentID=0, Status='live', Symbol='LOC105374193', Aliases='', description='uncharacterized LOC105374193', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166900656, end_position_on_the_genomic_accession=166971623, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105374194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374194, CurrentID=0, Status='live', Symbol='LOC105374194', Aliases='inv3q26.1', description='uncharacterized LOC105374194', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166898683, end_position_on_the_genomic_accession=166931824, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101060211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060211, CurrentID=0, Status='live', Symbol='CT45A7', Aliases='', description='cancer/testis antigen family 45, member A7', other_designations='cancer/testis antigen 45A7|cancer/testis antigen family 45 member A5-like', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135829229, end_position_on_the_genomic_accession=135846600, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374196, CurrentID=0, Status='live', Symbol='LOC105374196', Aliases='', description='uncharacterized LOC105374196', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167098552, end_position_on_the_genomic_accession=167106760, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374197, CurrentID=0, Status='live', Symbol='LOC105374197', Aliases='', description='uncharacterized LOC105374197', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167469301, end_position_on_the_genomic_accession=167476367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374198, CurrentID=0, Status='live', Symbol='LOC105374198', Aliases='', description='uncharacterized LOC105374198', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167671754, end_position_on_the_genomic_accession=167683641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374199, CurrentID=0, Status='live', Symbol='LOC105374199', Aliases='', description='uncharacterized LOC105374199', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=168897433, end_position_on_the_genomic_accession=168901868, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131320, CurrentID=0, Status='live', Symbol='RAB6C-AS1', Aliases='', description='RAB6C antisense RNA 1', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129966592, end_position_on_the_genomic_accession=129980472, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100421716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421716, CurrentID=0, Status='live', Symbol='UBE2Q2P4Y', Aliases='UBE2Q2PY1', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 4, Y-linked', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 4, Y-linked|ubiquitin-conjugating enzyme E2Q family member 2 pseudogene, Y-linked 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24232826, end_position_on_the_genomic_accession=24238993, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374202, CurrentID=0, Status='live', Symbol='LOC105374202', Aliases='', description='uncharacterized LOC105374202', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169659979, end_position_on_the_genomic_accession=169661151, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374203, CurrentID=0, Status='live', Symbol='LOC105374203', Aliases='', description='uncharacterized LOC105374203', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169657482, end_position_on_the_genomic_accession=169663004, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374204, CurrentID=0, Status='live', Symbol='LOC105374204', Aliases='', description='uncharacterized LOC105374204', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169533077, end_position_on_the_genomic_accession=169536254, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374205, CurrentID=0, Status='live', Symbol='LOC105374205', Aliases='', description='uncharacterized LOC105374205', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169447775, end_position_on_the_genomic_accession=169477052, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374206, CurrentID=0, Status='live', Symbol='LOC105374206', Aliases='', description='uncharacterized LOC105374206', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169281004, end_position_on_the_genomic_accession=169282760, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131327, CurrentID=0, Status='live', Symbol='UQCRC2P1', Aliases='', description='ubiquinol-cytochrome c reductase core protein II pseudogene 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46310393, end_position_on_the_genomic_accession=46311976, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131328, CurrentID=0, Status='live', Symbol='FAM210CP', Aliases='', description='family with sequence similarity 210 member C, pseudogene', other_designations='family with sequence similarity 210, member C, pseudogene', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27349715, end_position_on_the_genomic_accession=27350290, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374209, CurrentID=0, Status='live', Symbol='LOC105374209', Aliases='', description='uncharacterized LOC105374209', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170102947, end_position_on_the_genomic_accession=170121923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374210, CurrentID=0, Status='live', Symbol='LOC105374210', Aliases='', description='uncharacterized LOC105374210', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170410537, end_position_on_the_genomic_accession=170418777, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374211, CurrentID=0, Status='live', Symbol='LOC105374211', Aliases='', description='uncharacterized LOC105374211', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170768815, end_position_on_the_genomic_accession=170837620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374212, CurrentID=0, Status='live', Symbol='LOC105374212', Aliases='', description='uncharacterized LOC105374212', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170710268, end_position_on_the_genomic_accession=170734074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131333, CurrentID=0, Status='live', Symbol='RPL18P6', Aliases='RPL18_2_942', description='ribosomal protein L18 pseudogene 6', other_designations='', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109899168, end_position_on_the_genomic_accession=109903153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374214, CurrentID=0, Status='live', Symbol='LOC105374214', Aliases='', description='uncharacterized LOC105374214', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=170434921, end_position_on_the_genomic_accession=170460697, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374215, CurrentID=0, Status='live', Symbol='LOC105374215', Aliases='', description='uncharacterized LOC105374215', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171328262, end_position_on_the_genomic_accession=171338406, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374216, CurrentID=0, Status='live', Symbol='LOC105374216', Aliases='', description='uncharacterized LOC105374216', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171101640, end_position_on_the_genomic_accession=171161262, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374217, CurrentID=0, Status='live', Symbol='LOC105374217', Aliases='', description='uncharacterized LOC105374217', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171942486, end_position_on_the_genomic_accession=171959585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131338, CurrentID=0, Status='live', Symbol='RPL15P13', Aliases='RPL15_8_1099', description='ribosomal protein L15 pseudogene 13', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114883298, end_position_on_the_genomic_accession=114901344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131339, CurrentID=0, Status='live', Symbol='RPL32P4', Aliases='RPL32L, RPL32_11_924', description='ribosomal protein L32 pseudogene 4', other_designations='', map_location='8q21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86043052, end_position_on_the_genomic_accession=86043543, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374220, CurrentID=0, Status='live', Symbol='LOC105374220', Aliases='', description='uncharacterized LOC105374220', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173111052, end_position_on_the_genomic_accession=173200932, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374221, CurrentID=0, Status='live', Symbol='LOC105374221', Aliases='', description='uncharacterized LOC105374221', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172960524, end_position_on_the_genomic_accession=172964079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374222, CurrentID=0, Status='live', Symbol='LOC105374222', Aliases='', description='uncharacterized LOC105374222', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172945665, end_position_on_the_genomic_accession=172947418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374223, CurrentID=0, Status='live', Symbol='LOC105374223', Aliases='', description='uncharacterized LOC105374223', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173206683, end_position_on_the_genomic_accession=173213669, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  647696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647696, CurrentID=0, Status='live', Symbol='CDC27P1', Aliases='', description='cell division cycle 27 pseudogene 1', other_designations='cell division cycle 27 homolog pseudogene 1', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132261069, end_position_on_the_genomic_accession=132263124, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374225, CurrentID=0, Status='live', Symbol='LOC105374225', Aliases='', description='uncharacterized LOC105374225', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=173731376, end_position_on_the_genomic_accession=173735975, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374226, CurrentID=0, Status='live', Symbol='LOC105374226', Aliases='', description='uncharacterized LOC105374226', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=174417102, end_position_on_the_genomic_accession=174422373, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131347, CurrentID=0, Status='live', Symbol='LOC100131347', Aliases='', description='RAD52 motif containing 1 pseudogene', other_designations='RAD52 motif 1 pseudogene', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39057019, end_position_on_the_genomic_accession=39081451, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374228, CurrentID=0, Status='live', Symbol='LOC105374228', Aliases='', description='uncharacterized LOC105374228', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176362721, end_position_on_the_genomic_accession=176395980, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374229, CurrentID=0, Status='live', Symbol='LOC105374229', Aliases='', description='uncharacterized LOC105374229', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176693719, end_position_on_the_genomic_accession=176718590, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374230, CurrentID=0, Status='live', Symbol='LOC105374230', Aliases='', description='uncharacterized LOC105374230', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=176544632, end_position_on_the_genomic_accession=176565281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374231, CurrentID=0, Status='live', Symbol='LOC105374231', Aliases='', description='uncharacterized LOC105374231', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177327208, end_position_on_the_genomic_accession=177337453, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374232, CurrentID=0, Status='live', Symbol='LOC105374232', Aliases='', description='uncharacterized LOC105374232', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177569684, end_position_on_the_genomic_accession=177594352, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374233, CurrentID=0, Status='live', Symbol='LOC105374233', Aliases='', description='uncharacterized LOC105374233', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=177954105, end_position_on_the_genomic_accession=177960138, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374234, CurrentID=0, Status='live', Symbol='LOC105374234', Aliases='', description='uncharacterized LOC105374234', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178015827, end_position_on_the_genomic_accession=178065846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374235, CurrentID=0, Status='live', Symbol='LOC105374235', Aliases='', description='uncharacterized LOC105374235', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178333128, end_position_on_the_genomic_accession=178385392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374236, CurrentID=0, Status='live', Symbol='LOC105374236', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178493315, end_position_on_the_genomic_accession=178493812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374237, CurrentID=0, Status='live', Symbol='LOC105374237', Aliases='', description='uncharacterized LOC105374237', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=178908495, end_position_on_the_genomic_accession=178921933, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131358, CurrentID=0, Status='live', Symbol='RPL3P12', Aliases='RPL3_3_1807', description='ribosomal protein L3 pseudogene 12', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=122538232, end_position_on_the_genomic_accession=122539516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374239, CurrentID=0, Status='live', Symbol='LOC105374239', Aliases='', description='uncharacterized LOC105374239', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=179969413, end_position_on_the_genomic_accession=179979894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374240, CurrentID=0, Status='live', Symbol='LOC105374240', Aliases='', description='uncharacterized LOC105374240', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180867987, end_position_on_the_genomic_accession=180871402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374241, CurrentID=0, Status='live', Symbol='LOC105374241', Aliases='', description='uncharacterized LOC105374241', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=181331109, end_position_on_the_genomic_accession=181338280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374242, CurrentID=0, Status='live', Symbol='LOC105374242', Aliases='', description='uncharacterized LOC105374242', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182025654, end_position_on_the_genomic_accession=182027708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374243, CurrentID=0, Status='live', Symbol='LOC105374243', Aliases='', description='uncharacterized LOC105374243', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182149525, end_position_on_the_genomic_accession=182182761, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105374244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374244, CurrentID=0, Status='live', Symbol='LOC105374244', Aliases='', description='uncharacterized LOC105374244', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182498214, end_position_on_the_genomic_accession=182536766, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374245, CurrentID=0, Status='live', Symbol='LOC105374245', Aliases='', description='uncharacterized LOC105374245', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182783331, end_position_on_the_genomic_accession=182793353, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131366, CurrentID=0, Status='live', Symbol='LINC00605', Aliases='', description='long intergenic non-protein coding RNA 605', other_designations='', map_location='14q32.32', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103187221, end_position_on_the_genomic_accession=103189028, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374247, CurrentID=0, Status='live', Symbol='LOC105374247', Aliases='', description='uncharacterized LOC105374247', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183108214, end_position_on_the_genomic_accession=183114410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374248, CurrentID=0, Status='live', Symbol='LOC105374248', Aliases='', description='uncharacterized LOC105374248', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183630029, end_position_on_the_genomic_accession=183632957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421724, CurrentID=0, Status='live', Symbol='BNIP3P5', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 5', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42314083, end_position_on_the_genomic_accession=42314585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131370, CurrentID=0, Status='live', Symbol='DDX18P6', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 6', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91053009, end_position_on_the_genomic_accession=91057025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374251, CurrentID=0, Status='live', Symbol='LOC105374251', Aliases='', description='uncharacterized LOC105374251', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184508529, end_position_on_the_genomic_accession=184552863, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374252, CurrentID=0, Status='live', Symbol='LOC105374252', Aliases='', description='uncharacterized LOC105374252', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184490936, end_position_on_the_genomic_accession=184504476, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374253, CurrentID=0, Status='live', Symbol='LOC105374253', Aliases='', description='uncharacterized LOC105374253', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184773273, end_position_on_the_genomic_accession=184779295, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131374, CurrentID=0, Status='live', Symbol='PABPC1P5', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 5', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11336403, end_position_on_the_genomic_accession=11337109, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374255, CurrentID=0, Status='live', Symbol='LOC105374255', Aliases='', description='uncharacterized LOC105374255', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=185452481, end_position_on_the_genomic_accession=185461586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374256, CurrentID=0, Status='live', Symbol='LOC105374256', Aliases='', description='uncharacterized LOC105374256', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186095120, end_position_on_the_genomic_accession=186097674, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374257, CurrentID=0, Status='live', Symbol='LOC105374257', Aliases='', description='uncharacterized LOC105374257', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186121542, end_position_on_the_genomic_accession=186127275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131378, CurrentID=0, Status='live', Symbol='C11orf91', Aliases='', description='chromosome 11 open reading frame 91', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33698108, end_position_on_the_genomic_accession=33700740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131379, CurrentID=0, Status='live', Symbol='RPL36AP1', Aliases='RPL36A_19_1371', description='ribosomal protein L36a pseudogene 1', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=56683377, end_position_on_the_genomic_accession=56683777, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374260, CurrentID=0, Status='live', Symbol='LOC105374260', Aliases='', description='uncharacterized LOC105374260', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187224956, end_position_on_the_genomic_accession=187226382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929054, CurrentID=0, Status='live', Symbol='LOC101929054', Aliases='', description='uncharacterized LOC101929054', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52239258, end_position_on_the_genomic_accession=52241097, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374262, CurrentID=0, Status='live', Symbol='LOC105374262', Aliases='', description='uncharacterized LOC105374262', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187522840, end_position_on_the_genomic_accession=187535944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374263, CurrentID=0, Status='live', Symbol='LOC105374263', Aliases='', description='uncharacterized LOC105374263', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187792595, end_position_on_the_genomic_accession=187795687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374264, CurrentID=0, Status='live', Symbol='LOC105374264', Aliases='', description='uncharacterized LOC105374264', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187835692, end_position_on_the_genomic_accession=187890242, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374265, CurrentID=0, Status='live', Symbol='LOC105374265', Aliases='', description='uncharacterized LOC105374265', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187931469, end_position_on_the_genomic_accession=187933225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374266, CurrentID=0, Status='live', Symbol='LOC105374266', Aliases='', description='uncharacterized LOC105374266', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187966856, end_position_on_the_genomic_accession=187976421, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131387, CurrentID=0, Status='live', Symbol='RPL39P32', Aliases='RPL39_13_1539', description='ribosomal protein L39 pseudogene 32', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35566460, end_position_on_the_genomic_accession=35566866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131388, CurrentID=0, Status='live', Symbol='NDUFA3P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A3 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3 pseudogene 1|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3, 9kDa pseudogene', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44297096, end_position_on_the_genomic_accession=44297421, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374269, CurrentID=0, Status='live', Symbol='LOC105374269', Aliases='', description='uncharacterized LOC105374269', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189261243, end_position_on_the_genomic_accession=189270190, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131390, CurrentID=0, Status='live', Symbol='SP9', Aliases='ZNF990', description='Sp9 transcription factor', other_designations='Sp9 transcription factor homolog|zinc finger protein 990', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=174334960, end_position_on_the_genomic_accession=174346005, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374271, CurrentID=0, Status='live', Symbol='LOC105374271', Aliases='', description='uncharacterized LOC105374271', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=189645063, end_position_on_the_genomic_accession=189655993, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254528, CurrentID=0, Status='live', Symbol='MEIOB', Aliases='C16orf73, gs129', description='meiosis specific with OB domains', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1833983, end_position_on_the_genomic_accession=1872178, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100421728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421728, CurrentID=0, Status='live', Symbol='NEK4P3', Aliases='', description='NIMA-related kinase 4 pseudogene 3', other_designations='NIMA (never in mitosis gene a)-related kinase 4 pseudogene 3', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65054055, end_position_on_the_genomic_accession=65055006, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374274, CurrentID=0, Status='live', Symbol='LOC105374274', Aliases='', description='uncharacterized LOC105374274', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190788498, end_position_on_the_genomic_accession=190793679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  254531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254531, CurrentID=0, Status='live', Symbol='LPCAT4', Aliases='AGPAT7, AYTL3, LPAAT-eta, LPEAT2', description='lysophosphatidylcholine acyltransferase 4', other_designations='1-AGP acyltransferase 7|1-AGPAT 7|1-acylglycerol-3-phosphate O-acyltransferase 7 (lysophosphatidic acid acyltransferase, eta)|1-acylglycerophosphocholine O-acyltransferase|1-acylglycerophosphoserine O-acyltransferase|1-alkenylglycerophosphoethanolamine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|PLSC domain containing protein|acyl-CoA:lysophosphatidylethanolamine acyltransferase 2|acyltransferase like 3|acyltransferase-like 3|lysophosphatidylethanolamine acyltransferase 2|plasmalogen synthase', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34358888, end_position_on_the_genomic_accession=34367196, orientation='minus', exon_count=16, OMIM=612039, no_hdr0=''),
  100131396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131396, CurrentID=0, Status='live', Symbol='RPS6P14', Aliases='RPS6_6_1039', description='ribosomal protein S6 pseudogene 14', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48292711, end_position_on_the_genomic_accession=48293417, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374277, CurrentID=0, Status='live', Symbol='LOC105374277', Aliases='', description='uncharacterized LOC105374277', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192005107, end_position_on_the_genomic_accession=192037435, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374278, CurrentID=0, Status='live', Symbol='LOC105374278', Aliases='', description='uncharacterized LOC105374278', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=192566796, end_position_on_the_genomic_accession=192574213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480948, CurrentID=0, Status='live', Symbol='RN7SL100P', Aliases='', description='RNA, 7SL, cytoplasmic 100, pseudogene', other_designations='', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374280, CurrentID=0, Status='live', Symbol='LOC105374280', Aliases='', description='uncharacterized LOC105374280', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193237989, end_position_on_the_genomic_accession=193240998, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374281, CurrentID=0, Status='live', Symbol='LOC105374281', Aliases='', description='uncharacterized LOC105374281', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193589886, end_position_on_the_genomic_accession=193593109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374282, CurrentID=0, Status='live', Symbol='LOC105374282', Aliases='', description='uncharacterized LOC105374282', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193770959, end_position_on_the_genomic_accession=193772375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374283, CurrentID=0, Status='live', Symbol='LOC105374283', Aliases='', description='uncharacterized LOC105374283', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193787249, end_position_on_the_genomic_accession=193792084, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374284, CurrentID=0, Status='live', Symbol='LOC105374284', Aliases='', description='uncharacterized LOC105374284', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193832779, end_position_on_the_genomic_accession=193838008, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374285, CurrentID=0, Status='live', Symbol='LOC105374285', Aliases='', description='uncharacterized LOC105374285', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193840599, end_position_on_the_genomic_accession=193844009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374286, CurrentID=0, Status='live', Symbol='LOC105374286', Aliases='', description='uncharacterized LOC105374286', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193874783, end_position_on_the_genomic_accession=193885446, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374287, CurrentID=0, Status='live', Symbol='LOC105374287', Aliases='', description='uncharacterized LOC105374287', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=193873307, end_position_on_the_genomic_accession=193940154, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374288, CurrentID=0, Status='live', Symbol='LOC105374288', Aliases='', description='uncharacterized LOC105374288', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194110005, end_position_on_the_genomic_accession=194111829, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374289, CurrentID=0, Status='live', Symbol='LOC105374289', Aliases='', description='uncharacterized LOC105374289', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194276597, end_position_on_the_genomic_accession=194278113, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374290, CurrentID=0, Status='live', Symbol='LOC105374290', Aliases='', description='uncharacterized LOC105374290', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194499007, end_position_on_the_genomic_accession=194527436, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100421731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421731, CurrentID=0, Status='live', Symbol='VPS26AP1', Aliases='', description='VPS26A pseudogene 1', other_designations='vacuolar protein sorting 26 homolog A pseudogene 1', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113919222, end_position_on_the_genomic_accession=113920197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374292, CurrentID=0, Status='live', Symbol='LOC105374292', Aliases='', description='uncharacterized LOC105374292', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194769561, end_position_on_the_genomic_accession=194825688, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374293, CurrentID=0, Status='live', Symbol='LOC105374293', Aliases='', description='uncharacterized LOC105374293', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194762173, end_position_on_the_genomic_accession=194773151, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374294, CurrentID=0, Status='live', Symbol='LOC105374294', Aliases='', description='uncharacterized LOC105374294', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194981642, end_position_on_the_genomic_accession=194998957, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131415, CurrentID=0, Status='live', Symbol='OR7E163P', Aliases='', description='olfactory receptor family 7 subfamily E member 163 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 163, pseudogene', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  254552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254552, CurrentID=0, Status='live', Symbol='NUDT8', Aliases='', description='nudix hydrolase 8', other_designations='nucleoside diphosphate-linked moiety X motif 8, mitochondrial|nudix (nucleoside diphosphate linked moiety X)-type motif 8|nudix motif 8', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67627938, end_position_on_the_genomic_accession=67629937, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374297, CurrentID=0, Status='live', Symbol='LOC105374297', Aliases='', description='uncharacterized LOC105374297', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195647421, end_position_on_the_genomic_accession=195653263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131418, CurrentID=0, Status='live', Symbol='RPS27P22', Aliases='RPS27_12_1218', description='ribosomal protein S27 pseudogene 22', other_designations='ribosomal protein S27-like pseudogene', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=22722229, end_position_on_the_genomic_accession=22722469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374299, CurrentID=0, Status='live', Symbol='LOC105374299', Aliases='', description='proline-rich protein 36-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195815607, end_position_on_the_genomic_accession=195818776, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374300, CurrentID=0, Status='live', Symbol='LOC105374300', Aliases='', description='uncharacterized LOC105374300', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196018635, end_position_on_the_genomic_accession=196035453, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374301, CurrentID=0, Status='live', Symbol='LOC105374301', Aliases='', description='mucin-1-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195995072, end_position_on_the_genomic_accession=196005093, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374302, CurrentID=0, Status='live', Symbol='LOC105374302', Aliases='', description='uncharacterized LOC105374302', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196052120, end_position_on_the_genomic_accession=196055071, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  254559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254559, CurrentID=0, Status='live', Symbol='MIR9-3HG', Aliases='LINC00925', description='MIR9-3 host gene', other_designations='CTD-2335A18.1|long intergenic non-protein coding RNA 925', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89378042, end_position_on_the_genomic_accession=89398487, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374304, CurrentID=0, Status='live', Symbol='LOC105374304', Aliases='', description='uncharacterized LOC105374304', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196212954, end_position_on_the_genomic_accession=196220393, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100131425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131425, CurrentID=0, Status='live', Symbol='LOC100131425', Aliases='', description='LSM5 homolog, U6 small nuclear RNA associated (S. cerevisiae) pseudogene', other_designations='', map_location='7p21.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=16888621, end_position_on_the_genomic_accession=16888888, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374306, CurrentID=0, Status='live', Symbol='LOC105374306', Aliases='', description='uncharacterized LOC105374306', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196496214, end_position_on_the_genomic_accession=196500706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374307, CurrentID=0, Status='live', Symbol='LOC105374307', Aliases='', description='uncharacterized LOC105374307', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196505845, end_position_on_the_genomic_accession=196506892, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374308, CurrentID=0, Status='live', Symbol='LOC105374308', Aliases='', description='uncharacterized LOC105374308', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197419865, end_position_on_the_genomic_accession=197458406, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106480949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480949, CurrentID=0, Status='live', Symbol='RN7SL104P', Aliases='', description='RNA, 7SL, cytoplasmic 104, pseudogene', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374310, CurrentID=0, Status='live', Symbol='LOC105374310', Aliases='', description='uncharacterized LOC105374310', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=197830807, end_position_on_the_genomic_accession=197852553, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374311, CurrentID=0, Status='live', Symbol='LOC105374311', Aliases='', description='uncharacterized LOC105374311', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=198028011, end_position_on_the_genomic_accession=198035876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374312, CurrentID=0, Status='live', Symbol='LOC105374312', Aliases='', description='coiled-coil domain-containing protein 86-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125756192, end_position_on_the_genomic_accession=125769389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374313, CurrentID=0, Status='live', Symbol='LOC105374313', Aliases='', description='uncharacterized LOC105374313', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=150090269, end_position_on_the_genomic_accession=150151046, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100131434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131434, CurrentID=0, Status='live', Symbol='LINC00893', Aliases='', description='long intergenic non-protein coding RNA 893', other_designations='gene W', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=149527591, end_position_on_the_genomic_accession=149539752, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374315, CurrentID=0, Status='live', Symbol='LOC105374315', Aliases='', description='uncharacterized LOC105374315', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184591546, end_position_on_the_genomic_accession=184632004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374316, CurrentID=0, Status='live', Symbol='LOC105374316', Aliases='', description='uncharacterized LOC105374316', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21070940, end_position_on_the_genomic_accession=21100195, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374317, CurrentID=0, Status='live', Symbol='LOC105374317', Aliases='', description='uncharacterized LOC105374317', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21265966, end_position_on_the_genomic_accession=21285701, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374318, CurrentID=0, Status='live', Symbol='LOC105374318', Aliases='', description='uncharacterized LOC105374318', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=21317682, end_position_on_the_genomic_accession=21321008, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131439, CurrentID=0, Status='live', Symbol='CD300LD', Aliases='CD300D, CLM-4, CMRF35-A4, CMRF35A4', description='CD300 molecule like family member d', other_designations='CD300 antigen-like family member D|CD300 molecule-like family member d|immune receptor CD300d', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74579972, end_position_on_the_genomic_accession=74592231, orientation='minus', exon_count=4, OMIM=616301, no_hdr0=''),
  100131440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131440, CurrentID=0, Status='live', Symbol='CDC42P4', Aliases='', description='cell division cycle 42 pseudogene 4', other_designations='cell division cycle 42 (GTP binding protein, 25kDa) pseudogene', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109555108, end_position_on_the_genomic_accession=109556904, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374321, CurrentID=0, Status='live', Symbol='LOC105374321', Aliases='', description='uncharacterized LOC105374321', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22192843, end_position_on_the_genomic_accession=22228925, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131442, CurrentID=0, Status='live', Symbol='RCC2P5', Aliases='', description='regulator of chromosome condensation 2 pseudogene 5', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96753181, end_position_on_the_genomic_accession=96754684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374323, CurrentID=0, Status='live', Symbol='LOC105374323', Aliases='', description='uncharacterized LOC105374323', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23357527, end_position_on_the_genomic_accession=23381923, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374324, CurrentID=0, Status='live', Symbol='LOC105374324', Aliases='', description='uncharacterized LOC105374324', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23811283, end_position_on_the_genomic_accession=23828566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374325, CurrentID=0, Status='live', Symbol='LOC105374325', Aliases='', description='uncharacterized LOC105374325', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23501925, end_position_on_the_genomic_accession=23524344, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374326, CurrentID=0, Status='live', Symbol='LOC105374326', Aliases='', description='uncharacterized LOC105374326', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23441550, end_position_on_the_genomic_accession=23467457, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374327, CurrentID=0, Status='live', Symbol='LOC105374327', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=23398355, end_position_on_the_genomic_accession=23399918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374328, CurrentID=0, Status='live', Symbol='LOC105374328', Aliases='', description='uncharacterized LOC105374328', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24166000, end_position_on_the_genomic_accession=24174833, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374329, CurrentID=0, Status='live', Symbol='LOC105374329', Aliases='', description='uncharacterized LOC105374329', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24404028, end_position_on_the_genomic_accession=24412395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374330, CurrentID=0, Status='live', Symbol='LOC105374330', Aliases='', description='uncharacterized LOC105374330', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24766164, end_position_on_the_genomic_accession=24767963, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131451, CurrentID=0, Status='live', Symbol='TLR12P', Aliases='', description='toll-like receptor 12, pseudogene', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33466249, end_position_on_the_genomic_accession=33468975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374332, CurrentID=0, Status='live', Symbol='LOC105374332', Aliases='', description='uncharacterized LOC105374332', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=25724083, end_position_on_the_genomic_accession=25725529, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374333, CurrentID=0, Status='live', Symbol='LOC105374333', Aliases='', description='uncharacterized LOC105374333', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26046329, end_position_on_the_genomic_accession=26093870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131454, CurrentID=0, Status='live', Symbol='DBIL5P', Aliases='ELP2P', description='diazepam binding inhibitor-like 5, pseudogene', other_designations='endozepine-like peptide 2 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=752333, end_position_on_the_genomic_accession=755336, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374335, CurrentID=0, Status='live', Symbol='LOC105374335', Aliases='', description='uncharacterized LOC105374335', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135236, end_position_on_the_genomic_accession=138691, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374336, CurrentID=0, Status='live', Symbol='LOC105374336', Aliases='', description='uncharacterized LOC105374336', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=307335, end_position_on_the_genomic_accession=310075, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131457, CurrentID=0, Status='live', Symbol='SLC25A1P1', Aliases='', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1 pseudogene 1', other_designations='', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=85934667, end_position_on_the_genomic_accession=85935850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374338, CurrentID=0, Status='live', Symbol='LOC105374338', Aliases='', description='uncharacterized LOC105374338', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=577002, end_position_on_the_genomic_accession=584529, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106480950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480950, CurrentID=0, Status='live', Symbol='RN7SL108P', Aliases='', description='RNA, 7SL, cytoplasmic 108, pseudogene', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374340, CurrentID=0, Status='live', Symbol='LOC105374340', Aliases='', description='uncharacterized LOC105374340', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=841704, end_position_on_the_genomic_accession=846821, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374341, CurrentID=0, Status='live', Symbol='LOC105374341', Aliases='', description='uncharacterized LOC105374341', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=872266, end_position_on_the_genomic_accession=880752, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374342, CurrentID=0, Status='live', Symbol='LOC105374342', Aliases='', description='uncharacterized LOC105374342', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1005950, end_position_on_the_genomic_accession=1007971, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374343, CurrentID=0, Status='live', Symbol='LOC105374343', Aliases='', description='uncharacterized LOC105374343', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1043643, end_position_on_the_genomic_accession=1051502, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374344, CurrentID=0, Status='live', Symbol='LOC105374344', Aliases='', description='uncharacterized LOC105374344', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1113639, end_position_on_the_genomic_accession=1132977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374345, CurrentID=0, Status='live', Symbol='LOC105374345', Aliases='', description='uncharacterized LOC105374345', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1256420, end_position_on_the_genomic_accession=1270109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374346, CurrentID=0, Status='live', Symbol='LOC105374346', Aliases='', description='uncharacterized LOC105374346', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1287686, end_position_on_the_genomic_accession=1288290, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131467, CurrentID=0, Status='live', Symbol='RPS6P9', Aliases='RPS6_4_936', description='ribosomal protein S6 pseudogene 9', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103768240, end_position_on_the_genomic_accession=103769063, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374348, CurrentID=0, Status='live', Symbol='LOC105374348', Aliases='', description='uncharacterized LOC105374348', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1500920, end_position_on_the_genomic_accession=1503642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131469, CurrentID=0, Status='live', Symbol='RPSAP62', Aliases='RPSA_32_1778', description='ribosomal protein SA pseudogene 62', other_designations='', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=53323027, end_position_on_the_genomic_accession=53323718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374350, CurrentID=0, Status='live', Symbol='LOC105374350', Aliases='', description='uncharacterized LOC105374350', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1623704, end_position_on_the_genomic_accession=1626874, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374351, CurrentID=0, Status='live', Symbol='LOC105374351', Aliases='', description='uncharacterized LOC105374351', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2049287, end_position_on_the_genomic_accession=2054402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374352, CurrentID=0, Status='live', Symbol='LOC105374352', Aliases='', description='uncharacterized LOC105374352', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2342581, end_position_on_the_genomic_accession=2344660, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374353, CurrentID=0, Status='live', Symbol='LOC105374353', Aliases='', description='uncharacterized LOC105374353', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=2463014, end_position_on_the_genomic_accession=2466198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374354, CurrentID=0, Status='live', Symbol='LOC105374354', Aliases='', description='uncharacterized LOC105374354', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3450104, end_position_on_the_genomic_accession=3453280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374355, CurrentID=0, Status='live', Symbol='LOC105374355', Aliases='', description='uncharacterized LOC105374355', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3474938, end_position_on_the_genomic_accession=3476390, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374356, CurrentID=0, Status='live', Symbol='LOC105374356', Aliases='', description='uncharacterized LOC105374356', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3544860, end_position_on_the_genomic_accession=3545928, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374357, CurrentID=0, Status='live', Symbol='LOC105374357', Aliases='', description='uncharacterized LOC105374357', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3645103, end_position_on_the_genomic_accession=3653896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131478, CurrentID=0, Status='live', Symbol='NPM1P33', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 33', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=203772434, end_position_on_the_genomic_accession=203773596, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377151, CurrentID=0, Status='live', Symbol='LOC105377151', Aliases='', description='uncharacterized LOC105377151', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70167809, end_position_on_the_genomic_accession=70172568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374360, CurrentID=0, Status='live', Symbol='LOC105374360', Aliases='', description='uncharacterized LOC105374360', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=4370372, end_position_on_the_genomic_accession=4381912, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131481, CurrentID=0, Status='live', Symbol='RPL21P38', Aliases='RPL21_7_292', description='ribosomal protein L21 pseudogene 38', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=171587050, end_position_on_the_genomic_accession=171587606, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374362, CurrentID=0, Status='live', Symbol='LOC105374362', Aliases='', description='uncharacterized LOC105374362', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5189718, end_position_on_the_genomic_accession=5203914, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374363, CurrentID=0, Status='live', Symbol='LOC105374363', Aliases='', description='uncharacterized LOC105374363', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27356739, end_position_on_the_genomic_accession=27367501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374364, CurrentID=0, Status='live', Symbol='LOC105374364', Aliases='', description='uncharacterized LOC105374364', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6203552, end_position_on_the_genomic_accession=6221936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374365, CurrentID=0, Status='live', Symbol='LOC105374365', Aliases='', description='uncharacterized LOC105374365', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6568729, end_position_on_the_genomic_accession=6574227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374366, CurrentID=0, Status='live', Symbol='LOC105374366', Aliases='', description='uncharacterized LOC105374366', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6670728, end_position_on_the_genomic_accession=6673912, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374367, CurrentID=0, Status='live', Symbol='LOC105374367', Aliases='', description='uncharacterized LOC105374367', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6763516, end_position_on_the_genomic_accession=6767620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374368, CurrentID=0, Status='live', Symbol='LOC105374368', Aliases='', description='uncharacterized LOC105374368', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7064885, end_position_on_the_genomic_accession=7066616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106480951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480951, CurrentID=0, Status='live', Symbol='RN7SL113P', Aliases='', description='RNA, 7SL, cytoplasmic 113, pseudogene', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374370, CurrentID=0, Status='live', Symbol='LOC105374370', Aliases='', description='uncharacterized LOC105374370', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7070576, end_position_on_the_genomic_accession=7099217, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374371, CurrentID=0, Status='live', Symbol='LOC105374371', Aliases='', description='uncharacterized LOC105374371', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7257182, end_position_on_the_genomic_accession=7263224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374372, CurrentID=0, Status='live', Symbol='LOC105374372', Aliases='', description='uncharacterized LOC105374372', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7910373, end_position_on_the_genomic_accession=7919635, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131493, CurrentID=0, Status='live', Symbol='ZIK1P1', Aliases='', description='zinc finger protein interacting with K protein 1 pseudogene 1', other_designations='zinc finger protein interacting with K protein 1 homolog pseudogene 1', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124172932, end_position_on_the_genomic_accession=124174498, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374374, CurrentID=0, Status='live', Symbol='LOC105374374', Aliases='', description='uncharacterized LOC105374374', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8322201, end_position_on_the_genomic_accession=8323965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374375, CurrentID=0, Status='live', Symbol='LOC105374375', Aliases='', description='uncharacterized LOC105374375', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=8476938, end_position_on_the_genomic_accession=8493515, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131496, CurrentID=0, Status='live', Symbol='LOC100131496', Aliases='', description='uncharacterized LOC100131496', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=47318502, end_position_on_the_genomic_accession=47320754, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374377, CurrentID=0, Status='live', Symbol='LOC105374377', Aliases='', description='BCL2/adenovirus E1B 19 kDa protein-interacting protein 3 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=48204, end_position_on_the_genomic_accession=50279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374378, CurrentID=0, Status='live', Symbol='LOC105374378', Aliases='', description='uncharacterized LOC105374378', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27716068, end_position_on_the_genomic_accession=27761921, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374379, CurrentID=0, Status='live', Symbol='LOC105374379', Aliases='', description='collagen alpha-1(X) chain-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=1032699, end_position_on_the_genomic_accession=1039509, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374380, CurrentID=0, Status='live', Symbol='LOC105374380', Aliases='', description='golgin subfamily A member 6-like protein 22', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=6007972, end_position_on_the_genomic_accession=6019180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374381, CurrentID=0, Status='live', Symbol='LOC105374381', Aliases='', description='uncharacterized LOC105374381', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28348958, end_position_on_the_genomic_accession=28350574, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374382, CurrentID=0, Status='live', Symbol='LOC105374382', Aliases='', description='uncharacterized LOC105374382', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28358011, end_position_on_the_genomic_accession=28365746, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374383, CurrentID=0, Status='live', Symbol='LOC105374383', Aliases='', description='uncharacterized LOC105374383', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28377355, end_position_on_the_genomic_accession=28387129, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374384, CurrentID=0, Status='live', Symbol='LOC105374384', Aliases='', description='uncharacterized LOC105374384', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=28678455, end_position_on_the_genomic_accession=28683308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374385, CurrentID=0, Status='live', Symbol='LOC105374385', Aliases='', description='uncharacterized LOC105374385', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29063827, end_position_on_the_genomic_accession=29088122, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374386, CurrentID=0, Status='live', Symbol='LOC105374386', Aliases='', description='uncharacterized LOC105374386', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29088776, end_position_on_the_genomic_accession=29097607, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  647859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647859, CurrentID=0, Status='live', Symbol='LOC647859', Aliases='', description='occludin pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71074203, end_position_on_the_genomic_accession=71093070, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374388, CurrentID=0, Status='live', Symbol='LOC105374388', Aliases='', description='uncharacterized LOC105374388', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374389, CurrentID=0, Status='live', Symbol='LOC105374389', Aliases='', description='uncharacterized LOC105374389', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29474736, end_position_on_the_genomic_accession=29475449, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131510, CurrentID=0, Status='live', Symbol='LOC100131510', Aliases='', description='uncharacterized LOC100131510', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=25592818, end_position_on_the_genomic_accession=25593527, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374391, CurrentID=0, Status='live', Symbol='LOC105374391', Aliases='', description='uncharacterized LOC105374391', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33467432, end_position_on_the_genomic_accession=33520434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374392, CurrentID=0, Status='live', Symbol='LOC105374392', Aliases='', description='uncharacterized LOC105374392', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=33888950, end_position_on_the_genomic_accession=33905281, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374393, CurrentID=0, Status='live', Symbol='LOC105374393', Aliases='', description='uncharacterized LOC105374393', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=34093954, end_position_on_the_genomic_accession=34106062, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374394, CurrentID=0, Status='live', Symbol='LOC105374394', Aliases='', description='uncharacterized LOC105374394', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=34120687, end_position_on_the_genomic_accession=34269742, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100131515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131515, CurrentID=0, Status='live', Symbol='MST152', Aliases='', description='MSTP152', other_designations='', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=151368389, end_position_on_the_genomic_accession=151370114, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374396, CurrentID=0, Status='live', Symbol='LOC105374396', Aliases='', description='uncharacterized LOC105374396', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=35859398, end_position_on_the_genomic_accession=35887783, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374397, CurrentID=0, Status='live', Symbol='LOC105374397', Aliases='', description='uncharacterized LOC105374397', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=35948341, end_position_on_the_genomic_accession=35972234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374398, CurrentID=0, Status='live', Symbol='LOC105374398', Aliases='', description='uncharacterized LOC105374398', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=35981205, end_position_on_the_genomic_accession=35985266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480952, CurrentID=0, Status='live', Symbol='RN7SL118P', Aliases='', description='RNA, 7SL, cytoplasmic 118, pseudogene', other_designations='', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374400, CurrentID=0, Status='live', Symbol='LOC105374400', Aliases='', description='uncharacterized LOC105374400', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36496196, end_position_on_the_genomic_accession=36582717, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374401, CurrentID=0, Status='live', Symbol='LOC105374401', Aliases='', description='uncharacterized LOC105374401', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374402, CurrentID=0, Status='live', Symbol='LOC105374402', Aliases='', description='uncharacterized LOC105374402', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37112005, end_position_on_the_genomic_accession=37129587, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374403, CurrentID=0, Status='live', Symbol='LOC105374403', Aliases='', description='uncharacterized LOC105374403', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37454053, end_position_on_the_genomic_accession=37472389, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374404, CurrentID=0, Status='live', Symbol='LOC105374404', Aliases='', description='uncharacterized LOC105374404', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=37705409, end_position_on_the_genomic_accession=37715086, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374405, CurrentID=0, Status='live', Symbol='LOC105374405', Aliases='', description='uncharacterized LOC105374405', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30164209, end_position_on_the_genomic_accession=30207730, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100131526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131526, CurrentID=0, Status='live', Symbol='RPL26P6', Aliases='RPL26_9_1063', description='ribosomal protein L26 pseudogene 6', other_designations='', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73422216, end_position_on_the_genomic_accession=73422759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  123591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123591, CurrentID=0, Status='live', Symbol='TMEM266', Aliases='C15orf27', description='transmembrane protein 266', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76059946, end_position_on_the_genomic_accession=76204963, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  254664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254664, CurrentID=0, Status='live', Symbol='MAPK6PS2', Aliases='', description='mitogen-activated protein kinase 6 pseudogene 2', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=22435036, end_position_on_the_genomic_accession=22439067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374409, CurrentID=0, Status='live', Symbol='LOC105374409', Aliases='', description='uncharacterized LOC105374409', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38279999, end_position_on_the_genomic_accession=38288140, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374410, CurrentID=0, Status='live', Symbol='LOC105374410', Aliases='', description='uncharacterized LOC105374410', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38330705, end_position_on_the_genomic_accession=38385785, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374411, CurrentID=0, Status='live', Symbol='LOC105374411', Aliases='', description='uncharacterized LOC105374411', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38565759, end_position_on_the_genomic_accession=38573332, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131532, CurrentID=0, Status='live', Symbol='LOC100131532', Aliases='', description='uncharacterized LOC100131532', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169962623, end_position_on_the_genomic_accession=169967678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374413, CurrentID=0, Status='live', Symbol='LOC105374413', Aliases='', description='uncharacterized LOC105374413', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38744122, end_position_on_the_genomic_accession=38748737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374414, CurrentID=0, Status='live', Symbol='LOC105374414', Aliases='', description='uncharacterized LOC105374414', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30177191, end_position_on_the_genomic_accession=30191224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131535, CurrentID=0, Status='live', Symbol='RPL5P28', Aliases='RPL5_11_1110', description='ribosomal protein L5 pseudogene 28', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133143507, end_position_on_the_genomic_accession=133145350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374416, CurrentID=0, Status='live', Symbol='LOC105374416', Aliases='', description='uncharacterized LOC105374416', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=38952696, end_position_on_the_genomic_accession=38965784, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374417, CurrentID=0, Status='live', Symbol='LOC105374417', Aliases='', description='uncharacterized LOC105374417', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39113029, end_position_on_the_genomic_accession=39121003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374418, CurrentID=0, Status='live', Symbol='LOC105374418', Aliases='', description='uncharacterized LOC105374418', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=39067937, end_position_on_the_genomic_accession=39101763, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131539, CurrentID=0, Status='live', Symbol='ZNF705E', Aliases='', description='zinc finger protein 705E', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71814045, end_position_on_the_genomic_accession=71821547, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131540, CurrentID=0, Status='live', Symbol='LOC100131540', Aliases='', description='uncharacterized LOC100131540', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=91619549, end_position_on_the_genomic_accession=91623358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131541, CurrentID=0, Status='live', Symbol='LOC100131541', Aliases='', description='uncharacterized LOC100131541', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.9', start_position_on_the_genomic_accession=95987997, end_position_on_the_genomic_accession=95989129, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  123606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123606, CurrentID=0, Status='live', Symbol='NIPA1', Aliases='FSP3, SPG6', description='non imprinted in Prader-Willi/Angelman syndrome 1', other_designations='non-imprinted in Prader-Willi/Angelman syndrome 1|non-imprinted in Prader-Willi/Angelman syndrome region protein 1|spastic paraplegia 6 protein', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22786225, end_position_on_the_genomic_accession=22829789, orientation='plus', exon_count=6, OMIM=608145, no_hdr0=''),
  105374423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374423, CurrentID=0, Status='live', Symbol='LOC105374423', Aliases='', description='uncharacterized LOC105374423', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41540668, end_position_on_the_genomic_accession=41555414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374424, CurrentID=0, Status='live', Symbol='LOC105374424', Aliases='', description='uncharacterized LOC105374424', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41753412, end_position_on_the_genomic_accession=41765314, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374425, CurrentID=0, Status='live', Symbol='LOC105374425', Aliases='', description='uncharacterized LOC105374425', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41794205, end_position_on_the_genomic_accession=41799955, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131546, CurrentID=0, Status='live', Symbol='ECEL1P1', Aliases='', description='endothelin converting enzyme-like 1, pseudogene 1', other_designations='endothelin-converting enzyme-like 1-like', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232416028, end_position_on_the_genomic_accession=232419816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374427, CurrentID=0, Status='live', Symbol='LOC105374427', Aliases='', description='uncharacterized LOC105374427', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131548, CurrentID=0, Status='live', Symbol='RPS6P4', Aliases='RPS6_2_432', description='ribosomal protein S6 pseudogene 4', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=163102039, end_position_on_the_genomic_accession=163116154, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106480953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480953, CurrentID=0, Status='live', Symbol='RN7SL120P', Aliases='', description='RNA, 7SL, cytoplasmic 120, pseudogene', other_designations='', map_location='3p26.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131550, CurrentID=0, Status='live', Symbol='CNN2P12', Aliases='', description='calponin 2 pseudogene 12', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18726656, end_position_on_the_genomic_accession=18727441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131551, CurrentID=0, Status='live', Symbol='LINC00887', Aliases='', description='long intergenic non-protein coding RNA 887', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=194301199, end_position_on_the_genomic_accession=194312804, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374432, CurrentID=0, Status='live', Symbol='LOC105374432', Aliases='', description='uncharacterized LOC105374432', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43223188, end_position_on_the_genomic_accession=43282951, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374433, CurrentID=0, Status='live', Symbol='LOC105374433', Aliases='', description='uncharacterized LOC105374433', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43316068, end_position_on_the_genomic_accession=43341359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422999, CurrentID=0, Status='live', Symbol='MIR4278', Aliases='', description='microRNA 4278', other_designations='hsa-mir-4278', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6827853, end_position_on_the_genomic_accession=6827921, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374435, CurrentID=0, Status='live', Symbol='LOC105374435', Aliases='', description='uncharacterized LOC105374435', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30823452, end_position_on_the_genomic_accession=30835366, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374436, CurrentID=0, Status='live', Symbol='LOC105374436', Aliases='', description='uncharacterized LOC105374436', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43457569, end_position_on_the_genomic_accession=43492566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374437, CurrentID=0, Status='live', Symbol='LOC105374437', Aliases='', description='uncharacterized LOC105374437', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43760602, end_position_on_the_genomic_accession=43766333, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374438, CurrentID=0, Status='live', Symbol='LOC105374438', Aliases='', description='uncharacterized LOC105374438', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=43979999, end_position_on_the_genomic_accession=44016794, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374439, CurrentID=0, Status='live', Symbol='LOC105374439', Aliases='', description='uncharacterized LOC105374439', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44726585, end_position_on_the_genomic_accession=44776409, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  123624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123624, CurrentID=0, Status='live', Symbol='AGBL1', Aliases='CCP4, FECD8', description='ATP/GTP binding protein-like 1', other_designations='', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=86079787, end_position_on_the_genomic_accession=87031178, orientation='plus', exon_count=27, OMIM=615496, no_hdr0=''),
  100131561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131561, CurrentID=0, Status='live', Symbol='FKSG29', Aliases='', description='FKSG29', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99351420, end_position_on_the_genomic_accession=99352027, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374442, CurrentID=0, Status='live', Symbol='LOC105374442', Aliases='', description='uncharacterized LOC105374442', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47092847, end_position_on_the_genomic_accession=47096145, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374443, CurrentID=0, Status='live', Symbol='LOC105374443', Aliases='', description='uncharacterized LOC105374443', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=47638210, end_position_on_the_genomic_accession=47645063, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131564, CurrentID=0, Status='live', Symbol='CCDC18-AS1', Aliases='', description='CCDC18 antisense RNA 1', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=93310109, end_position_on_the_genomic_accession=93345811, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100131565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131565, CurrentID=0, Status='live', Symbol='EIF4EP5', Aliases='', description='eukaryotic translation initiation factor 4E pseudogene 5', other_designations='', map_location='8p22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14309034, end_position_on_the_genomic_accession=14309871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374446, CurrentID=0, Status='live', Symbol='LOC105374446', Aliases='', description='uncharacterized LOC105374446', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=30861069, end_position_on_the_genomic_accession=30879094, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374447, CurrentID=0, Status='live', Symbol='LOC105374447', Aliases='', description='uncharacterized LOC105374447', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49239246, end_position_on_the_genomic_accession=49244366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131568, CurrentID=0, Status='live', Symbol='RBM43P1', Aliases='', description='RNA binding motif protein 43 pseudogene 1', other_designations='', map_location='3p14.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69470425, end_position_on_the_genomic_accession=69472531, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374449, CurrentID=0, Status='live', Symbol='LOC105374449', Aliases='', description='uncharacterized LOC105374449', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31773853, end_position_on_the_genomic_accession=31803056, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374450, CurrentID=0, Status='live', Symbol='LOC105374450', Aliases='', description='uncharacterized LOC105374450', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31855215, end_position_on_the_genomic_accession=31865811, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374451, CurrentID=0, Status='live', Symbol='LOC105374451', Aliases='', description='uncharacterized LOC105374451', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=31956453, end_position_on_the_genomic_accession=31968268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131572, CurrentID=0, Status='live', Symbol='RPS27P9', Aliases='RPS27_1_33', description='ribosomal protein S27 pseudogene 9', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37556215, end_position_on_the_genomic_accession=37556553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374453, CurrentID=0, Status='live', Symbol='LOC105374453', Aliases='', description='uncharacterized LOC105374453', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=32885988, end_position_on_the_genomic_accession=32903087, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374454, CurrentID=0, Status='live', Symbol='LOC105374454', Aliases='', description='uncharacterized LOC105374454', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33555045, end_position_on_the_genomic_accession=33557450, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374455, CurrentID=0, Status='live', Symbol='LOC105374455', Aliases='', description='uncharacterized LOC105374455', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33600613, end_position_on_the_genomic_accession=33656498, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374456, CurrentID=0, Status='live', Symbol='LOC105374456', Aliases='', description='uncharacterized LOC105374456', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34047951, end_position_on_the_genomic_accession=34067084, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131577, CurrentID=0, Status='live', Symbol='BMS1P19', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 19', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 19|BMS1 ribosome biogenesis factor pseudogene 19', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=109667132, end_position_on_the_genomic_accession=109668349, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374458, CurrentID=0, Status='live', Symbol='LOC105374458', Aliases='', description='uncharacterized LOC105374458', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34702783, end_position_on_the_genomic_accession=34736518, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101929087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929087, CurrentID=0, Status='live', Symbol='LOC101929087', Aliases='', description='small ubiquitin-related modifier 2-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153144205, end_position_on_the_genomic_accession=153144588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374460, CurrentID=0, Status='live', Symbol='LOC105374460', Aliases='', description='uncharacterized LOC105374460', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=34943766, end_position_on_the_genomic_accession=34990735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131581, CurrentID=0, Status='live', Symbol='LOC100131581', Aliases='', description='uncharacterized LOC100131581', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=12480186, end_position_on_the_genomic_accession=12481558, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374462, CurrentID=0, Status='live', Symbol='LOC105374462', Aliases='', description='uncharacterized LOC105374462', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=36365957, end_position_on_the_genomic_accession=36378814, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374463, CurrentID=0, Status='live', Symbol='LOC105374463', Aliases='', description='uncharacterized LOC105374463', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37211091, end_position_on_the_genomic_accession=37216193, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374464, CurrentID=0, Status='live', Symbol='LOC105374464', Aliases='', description='uncharacterized LOC105374464', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37455124, end_position_on_the_genomic_accession=37613081, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374465, CurrentID=0, Status='live', Symbol='LOC105374465', Aliases='', description='uncharacterized LOC105374465', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37719544, end_position_on_the_genomic_accession=37737598, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374466, CurrentID=0, Status='live', Symbol='LOC105374466', Aliases='', description='uncharacterized LOC105374466', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38112991, end_position_on_the_genomic_accession=38121988, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374467, CurrentID=0, Status='live', Symbol='LOC105374467', Aliases='', description='uncharacterized LOC105374467', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38269093, end_position_on_the_genomic_accession=38272203, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374468, CurrentID=0, Status='live', Symbol='LOC105374468', Aliases='', description='uncharacterized LOC105374468', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38408942, end_position_on_the_genomic_accession=38411923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374469, CurrentID=0, Status='live', Symbol='LOC105374469', Aliases='', description='uncharacterized LOC105374469', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38694895, end_position_on_the_genomic_accession=38733550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374470, CurrentID=0, Status='live', Symbol='LOC105374470', Aliases='', description='uncharacterized LOC105374470', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38836370, end_position_on_the_genomic_accession=38850121, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374471, CurrentID=0, Status='live', Symbol='LOC105374471', Aliases='', description='uncharacterized LOC105374471', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38935829, end_position_on_the_genomic_accession=38953511, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374472, CurrentID=0, Status='live', Symbol='LOC105374472', Aliases='', description='uncharacterized LOC105374472', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40592103, end_position_on_the_genomic_accession=40602796, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131593, CurrentID=0, Status='live', Symbol='LOC100131593', Aliases='', description='SAP domain containing ribonucleoprotein pseudogene', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101528417, end_position_on_the_genomic_accession=101529619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  647946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647946, CurrentID=0, Status='live', Symbol='MIR924HG', Aliases='LINC00669', description='MIR924 host gene', other_designations='long intergenic non-protein coding RNA 669', map_location='18q12.2-q12.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39206924, end_position_on_the_genomic_accession=39751995, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374475, CurrentID=0, Status='live', Symbol='LOC105374475', Aliases='', description='uncharacterized LOC105374475', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40633256, end_position_on_the_genomic_accession=40639622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374476, CurrentID=0, Status='live', Symbol='LOC105374476', Aliases='', description='uncharacterized LOC105374476', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9987982, end_position_on_the_genomic_accession=10016540, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374477, CurrentID=0, Status='live', Symbol='LOC105374477', Aliases='', description='uncharacterized LOC105374477', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10062741, end_position_on_the_genomic_accession=10074756, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374478, CurrentID=0, Status='live', Symbol='LOC105374478', Aliases='', description='uncharacterized LOC105374478', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10142809, end_position_on_the_genomic_accession=10158233, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374479, CurrentID=0, Status='live', Symbol='LOC105374479', Aliases='', description='uncharacterized LOC105374479', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10785781, end_position_on_the_genomic_accession=10789778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374480, CurrentID=0, Status='live', Symbol='LOC105374480', Aliases='', description='uncharacterized LOC105374480', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10621048, end_position_on_the_genomic_accession=10630285, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374481, CurrentID=0, Status='live', Symbol='LOC105374481', Aliases='', description='uncharacterized LOC105374481', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10569868, end_position_on_the_genomic_accession=10577385, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374482, CurrentID=0, Status='live', Symbol='LOC105374482', Aliases='', description='uncharacterized LOC105374482', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10455041, end_position_on_the_genomic_accession=10531030, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374483, CurrentID=0, Status='live', Symbol='LOC105374483', Aliases='', description='uncharacterized LOC105374483', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11776305, end_position_on_the_genomic_accession=11789866, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374484, CurrentID=0, Status='live', Symbol='LOC105374484', Aliases='', description='uncharacterized LOC105374484', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11740947, end_position_on_the_genomic_accession=11769852, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374485, CurrentID=0, Status='live', Symbol='LOC105374485', Aliases='', description='uncharacterized LOC105374485', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11730217, end_position_on_the_genomic_accession=11734988, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374486, CurrentID=0, Status='live', Symbol='LOC105374486', Aliases='', description='uncharacterized LOC105374486', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11429160, end_position_on_the_genomic_accession=11438574, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374487 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374487, CurrentID=0, Status='live', Symbol='LOC105374487', Aliases='', description='uncharacterized LOC105374487', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=11309554, end_position_on_the_genomic_accession=11312798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131608, CurrentID=0, Status='live', Symbol='PRR23D1', Aliases='', description='proline rich 23 domain containing 1', other_designations='proline-rich protein 23C-like', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7536780, end_position_on_the_genomic_accession=7542450, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131609, CurrentID=0, Status='live', Symbol='HNRNPA1P2', Aliases='HNRPA1P2, dJ372B18.1', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 2', other_designations='', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=32325198, end_position_on_the_genomic_accession=32326521, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374490, CurrentID=0, Status='live', Symbol='LOC105374490', Aliases='', description='uncharacterized LOC105374490', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=12540087, end_position_on_the_genomic_accession=12571354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131611, CurrentID=0, Status='live', Symbol='PGAM4P2', Aliases='', description='phosphoglycerate mutase family member 4 pseudogene 2', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=115764114, end_position_on_the_genomic_accession=115764761, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374492, CurrentID=0, Status='live', Symbol='LOC105374492', Aliases='', description='uncharacterized LOC105374492', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=12469901, end_position_on_the_genomic_accession=12609447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374493, CurrentID=0, Status='live', Symbol='LOC105374493', Aliases='', description='uncharacterized LOC105374493', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=12876608, end_position_on_the_genomic_accession=12973023, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131614, CurrentID=0, Status='live', Symbol='RPS4XP22', Aliases='RPS4P22, RPS4X_8_1621', description='ribosomal protein S4X pseudogene 22', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9683239, end_position_on_the_genomic_accession=9684096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374495, CurrentID=0, Status='live', Symbol='LOC105374495', Aliases='', description='uncharacterized LOC105374495', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13484449, end_position_on_the_genomic_accession=13489889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131616, CurrentID=0, Status='live', Symbol='CASP3P1', Aliases='', description='caspase 3 pseudogene 1', other_designations='caspase 3, apoptosis-related cysteine peptidase pseudogene 1', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70660654, end_position_on_the_genomic_accession=70661261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374497, CurrentID=0, Status='live', Symbol='LOC105374497', Aliases='', description='uncharacterized LOC105374497', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=40686110, end_position_on_the_genomic_accession=40970267, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374498, CurrentID=0, Status='live', Symbol='LOC105374498', Aliases='', description='uncharacterized LOC105374498', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14954421, end_position_on_the_genomic_accession=14966710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374499, CurrentID=0, Status='live', Symbol='LOC105374499', Aliases='', description='uncharacterized LOC105374499', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15374777, end_position_on_the_genomic_accession=15379776, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374500, CurrentID=0, Status='live', Symbol='LOC105374500', Aliases='', description='uncharacterized LOC105374500', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15358561, end_position_on_the_genomic_accession=15362520, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374501, CurrentID=0, Status='live', Symbol='LOC105374501', Aliases='', description='uncharacterized LOC105374501', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=15124843, end_position_on_the_genomic_accession=15131675, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374502, CurrentID=0, Status='live', Symbol='LOC105374502', Aliases='', description='uncharacterized LOC105374502', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16124284, end_position_on_the_genomic_accession=16126143, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131623, CurrentID=0, Status='live', Symbol='NUTF2P6', Aliases='', description='nuclear transport factor 2 pseudogene 6', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35282515, end_position_on_the_genomic_accession=35282985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  123688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123688, CurrentID=0, Status='live', Symbol='HYKK', Aliases='AGPHD1', description='hydroxylysine kinase', other_designations='5-hydroxy-L-lysine kinase|5-hydroxylysine kinase|aminoglycoside phosphotransferase domain containing 1|aminoglycoside phosphotransferase domain-containing protein 1', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78507564, end_position_on_the_genomic_accession=78537373, orientation='plus', exon_count=7, OMIM=614681, no_hdr0=''),
  105374505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374505, CurrentID=0, Status='live', Symbol='LOC105374505', Aliases='', description='uncharacterized LOC105374505', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=16400500, end_position_on_the_genomic_accession=16598545, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100131626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131626, CurrentID=0, Status='live', Symbol='LOC100131626', Aliases='', description='uncharacterized LOC100131626', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118381220, end_position_on_the_genomic_accession=118418815, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  647979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647979, CurrentID=0, Status='live', Symbol='NORAD', Aliases='LINC00657', description='non-coding RNA activated by DNA damage', other_designations='long intergenic non-protein coding RNA 657', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36045618, end_position_on_the_genomic_accession=36050960, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374508, CurrentID=0, Status='live', Symbol='LOC105374508', Aliases='', description='uncharacterized LOC105374508', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17612181, end_position_on_the_genomic_accession=17614632, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131629, CurrentID=0, Status='live', Symbol='ARL2BPP7', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 7', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 7|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 7|ADP-ribosylation factor-like 2 binding protein pseudogene 7', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102053863, end_position_on_the_genomic_accession=102055826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374510, CurrentID=0, Status='live', Symbol='LOC105374510', Aliases='', description='uncharacterized LOC105374510', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=18521168, end_position_on_the_genomic_accession=18832932, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374511, CurrentID=0, Status='live', Symbol='LOC105374511', Aliases='', description='uncharacterized LOC105374511', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=19856282, end_position_on_the_genomic_accession=19937686, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374512, CurrentID=0, Status='live', Symbol='LOC105374512', Aliases='', description='uncharacterized LOC105374512', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=19998929, end_position_on_the_genomic_accession=20003979, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374513, CurrentID=0, Status='live', Symbol='LOC105374513', Aliases='', description='uncharacterized LOC105374513', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20006917, end_position_on_the_genomic_accession=20026275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374514, CurrentID=0, Status='live', Symbol='LOC105374514', Aliases='', description='uncharacterized LOC105374514', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=20206013, end_position_on_the_genomic_accession=20212958, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131635, CurrentID=0, Status='live', Symbol='LOC100131635', Aliases='', description='hCG1645011-like', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=187702366, end_position_on_the_genomic_accession=187732415, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374516, CurrentID=0, Status='live', Symbol='LOC105374516', Aliases='', description='uncharacterized LOC105374516', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=21582076, end_position_on_the_genomic_accession=21613722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254773, CurrentID=0, Status='live', Symbol='LYG2', Aliases='LYGH', description='lysozyme G-like 2', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99198503, end_position_on_the_genomic_accession=99256732, orientation='minus', exon_count=8, OMIM=616547, no_hdr0=''),
  100131638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131638, CurrentID=0, Status='live', Symbol='CCT4P2', Aliases='CCT4-1P', description='chaperonin containing TCP1, subunit 4 (delta) pseudogene 2', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65270912, end_position_on_the_genomic_accession=65273021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374519, CurrentID=0, Status='live', Symbol='LOC105374519', Aliases='', description='uncharacterized LOC105374519', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22350921, end_position_on_the_genomic_accession=22361152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130579, CurrentID=0, Status='live', Symbol='PHBP20', Aliases='', description='prohibitin pseudogene 20', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72450658, end_position_on_the_genomic_accession=72451472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131641, CurrentID=0, Status='live', Symbol='RBM22P12', Aliases='', description='RNA binding motif protein 22 pseudogene 12', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31680771, end_position_on_the_genomic_accession=31682012, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  254778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254778, CurrentID=0, Status='live', Symbol='C8orf46', Aliases='', description='chromosome 8 open reading frame 46', other_designations='', map_location='8q13.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66493256, end_position_on_the_genomic_accession=66518524, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374523, CurrentID=0, Status='live', Symbol='LOC105374523', Aliases='', description='uncharacterized LOC105374523', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23302544, end_position_on_the_genomic_accession=23400408, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374524, CurrentID=0, Status='live', Symbol='LOC105374524', Aliases='', description='uncharacterized LOC105374524', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=22997576, end_position_on_the_genomic_accession=23332387, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  105374525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374525, CurrentID=0, Status='live', Symbol='LOC105374525', Aliases='', description='uncharacterized LOC105374525', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23578638, end_position_on_the_genomic_accession=23581482, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374526, CurrentID=0, Status='live', Symbol='LOC105374526', Aliases='', description='uncharacterized LOC105374526', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=23782554, end_position_on_the_genomic_accession=23787773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  254783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254783, CurrentID=0, Status='live', Symbol='OR6C74', Aliases='', description='olfactory receptor family 6 subfamily C member 74', other_designations='olfactory receptor, family 6, subfamily C, member 74', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55239762, end_position_on_the_genomic_accession=55248226, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  648000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648000, CurrentID=0, Status='live', Symbol='RPL7P23', Aliases='RPL7_7_580', description='ribosomal protein L7 pseudogene 23', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77582354, end_position_on_the_genomic_accession=77583200, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374529, CurrentID=0, Status='live', Symbol='LOC105374529', Aliases='', description='uncharacterized LOC105374529', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24436648, end_position_on_the_genomic_accession=24470650, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254786, CurrentID=0, Status='live', Symbol='OR6C3', Aliases='OST709', description='olfactory receptor family 6 subfamily C member 3', other_designations='HSA8|olfactory receptor, family 6, subfamily C, member 3', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55331701, end_position_on_the_genomic_accession=55332636, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100421771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421771, CurrentID=0, Status='live', Symbol='UBE2Q2P8', Aliases='', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 8', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 8', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84248418, end_position_on_the_genomic_accession=84256032, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374532, CurrentID=0, Status='live', Symbol='LOC105374532', Aliases='', description='uncharacterized LOC105374532', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24788709, end_position_on_the_genomic_accession=24790995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374533, CurrentID=0, Status='live', Symbol='LOC105374533', Aliases='', description='uncharacterized LOC105374533', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=24800485, end_position_on_the_genomic_accession=24802726, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374534, CurrentID=0, Status='live', Symbol='LOC105374534', Aliases='', description='uncharacterized LOC105374534', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=25107601, end_position_on_the_genomic_accession=25112765, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131655, CurrentID=0, Status='live', Symbol='LOC100131655', Aliases='', description='uncharacterized LOC100131655', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=76794732, end_position_on_the_genomic_accession=76822295, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  123720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123720, CurrentID=0, Status='live', Symbol='WHAMM', Aliases='WHAMM1, WHDC1', description='WAS protein homolog associated with actin, golgi membranes and microtubules', other_designations='WAS protein homology region 2 domain-containing protein 1|WH2 domain-containing protein 1', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82808169, end_position_on_the_genomic_accession=82834861, orientation='plus', exon_count=13, OMIM=612393, no_hdr0=''),
  105374537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374537, CurrentID=0, Status='live', Symbol='LOC105374537', Aliases='', description='uncharacterized LOC105374537', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42413741, end_position_on_the_genomic_accession=42425087, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  123722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123722, CurrentID=0, Status='live', Symbol='FSD2', Aliases='SPRYD1', description='fibronectin type III and SPRY domain containing 2', other_designations='SPRY domain containing 1|SPRY domain-containing protein 1', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82755365, end_position_on_the_genomic_accession=82806211, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105374539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374539, CurrentID=0, Status='live', Symbol='LOC105374539', Aliases='', description='uncharacterized LOC105374539', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26020866, end_position_on_the_genomic_accession=26024646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374540, CurrentID=0, Status='live', Symbol='LOC105374540', Aliases='', description='uncharacterized LOC105374540', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26070761, end_position_on_the_genomic_accession=26104251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131661, CurrentID=0, Status='live', Symbol='FTLP5', Aliases='', description='ferritin, light polypeptide pseudogene 5', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3870794, end_position_on_the_genomic_accession=3872207, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374542, CurrentID=0, Status='live', Symbol='LOC105374542', Aliases='', description='uncharacterized LOC105374542', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26288142, end_position_on_the_genomic_accession=26299235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374543, CurrentID=0, Status='live', Symbol='LOC105374543', Aliases='', description='uncharacterized LOC105374543', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42335559, end_position_on_the_genomic_accession=42346725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374544, CurrentID=0, Status='live', Symbol='LOC105374544', Aliases='', description='uncharacterized LOC105374544', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26268225, end_position_on_the_genomic_accession=26275467, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374545, CurrentID=0, Status='live', Symbol='LOC105374545', Aliases='', description='uncharacterized LOC105374545', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=26181106, end_position_on_the_genomic_accession=26197981, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  139420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139420, CurrentID=0, Status='live', Symbol='PPP4R3CP', Aliases='FLFL3P, SMEK3P, smk1', description='protein phosphatase 4 regulatory subunit 3C, pseudogene', other_designations='SMEK homolog 3, suppressor of mek1 (Dictyostelium) pseudogene|protein phosphatase 4, regulatory subunit 3C, pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27460211, end_position_on_the_genomic_accession=27463341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374547, CurrentID=0, Status='live', Symbol='LOC105374547', Aliases='', description='uncharacterized LOC105374547', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27645973, end_position_on_the_genomic_accession=27656496, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374548, CurrentID=0, Status='live', Symbol='LOC105374548', Aliases='', description='uncharacterized LOC105374548', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27669972, end_position_on_the_genomic_accession=27719249, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106480957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480957, CurrentID=0, Status='live', Symbol='RN7SL131P', Aliases='', description='RNA, 7SL, cytoplasmic 131, pseudogene', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374550, CurrentID=0, Status='live', Symbol='LOC105374550', Aliases='', description='uncharacterized LOC105374550', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27819188, end_position_on_the_genomic_accession=27830173, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131671, CurrentID=0, Status='live', Symbol='SEC13P1', Aliases='', description='SEC13 homolog, nuclear pore and COPII coat complex component pseudogene 1', other_designations='SEC13 homolog (S. cerevisiae) pseudogene 1', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33033814, end_position_on_the_genomic_accession=33035007, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131672, CurrentID=0, Status='live', Symbol='RPL21P87', Aliases='RPL21_35_1007', description='ribosomal protein L21 pseudogene 87', other_designations='', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110024841, end_position_on_the_genomic_accession=110025398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374553, CurrentID=0, Status='live', Symbol='LOC105374553', Aliases='', description='uncharacterized LOC105374553', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28024176, end_position_on_the_genomic_accession=28108950, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374554, CurrentID=0, Status='live', Symbol='LOC105374554', Aliases='', description='uncharacterized LOC105374554', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28117558, end_position_on_the_genomic_accession=28288204, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374555, CurrentID=0, Status='live', Symbol='LOC105374555', Aliases='', description='uncharacterized LOC105374555', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42582316, end_position_on_the_genomic_accession=42623401, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374556, CurrentID=0, Status='live', Symbol='LOC105374556', Aliases='', description='uncharacterized LOC105374556', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28393311, end_position_on_the_genomic_accession=28402926, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374557, CurrentID=0, Status='live', Symbol='LOC105374557', Aliases='', description='uncharacterized LOC105374557', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28465155, end_position_on_the_genomic_accession=28479381, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131678, CurrentID=0, Status='live', Symbol='CCNB3P1', Aliases='', description='cyclin B3 pseudogene 1', other_designations='CCNB3 pseudogene 1', map_location='5p14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=26739753, end_position_on_the_genomic_accession=26741183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374559, CurrentID=0, Status='live', Symbol='LOC105374559', Aliases='', description='uncharacterized LOC105374559', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=28624554, end_position_on_the_genomic_accession=28630095, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131680, CurrentID=0, Status='live', Symbol='LOC100131680', Aliases='', description='52 kDa repressor of the inhibitor of the protein kinase-like', other_designations='', map_location='6q14.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=76660390, end_position_on_the_genomic_accession=76663399, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  123745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123745, CurrentID=0, Status='live', Symbol='PLA2G4E', Aliases='', description='phospholipase A2 group IVE', other_designations='cPLA2-epsilon|phospholipase A2, group IVE', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41981582, end_position_on_the_genomic_accession=42050703, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  105374562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374562, CurrentID=0, Status='live', Symbol='LOC105374562', Aliases='', description='uncharacterized LOC105374562', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=29797903, end_position_on_the_genomic_accession=29801039, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374563, CurrentID=0, Status='live', Symbol='LOC105374563', Aliases='', description='uncharacterized LOC105374563', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=30747985, end_position_on_the_genomic_accession=30752119, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374564, CurrentID=0, Status='live', Symbol='LOC105374564', Aliases='', description='uncharacterized LOC105374564', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=31499797, end_position_on_the_genomic_accession=31506308, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374565, CurrentID=0, Status='live', Symbol='LOC105374565', Aliases='', description='uncharacterized LOC105374565', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=31506468, end_position_on_the_genomic_accession=31558821, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374566, CurrentID=0, Status='live', Symbol='LOC105374566', Aliases='', description='uncharacterized LOC105374566', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=31702060, end_position_on_the_genomic_accession=31729518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374567, CurrentID=0, Status='live', Symbol='LOC105374567', Aliases='', description='proline-rich protein 2-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=42808519, end_position_on_the_genomic_accession=42971692, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100131688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131688, CurrentID=0, Status='live', Symbol='VN1R68P', Aliases='', description='vomeronasal 1 receptor 68 pseudogene', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35118172, end_position_on_the_genomic_accession=35118992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131689, CurrentID=0, Status='live', Symbol='LOC100131689', Aliases='', description='pumilio domain-containing protein KIAA0020-like', other_designations='HBV X-transactivated gene 5 protein pseudogene', map_location='14q22.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52286768, end_position_on_the_genomic_accession=52289434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374570, CurrentID=0, Status='live', Symbol='LOC105374570', Aliases='', description='uncharacterized LOC105374570', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43127570, end_position_on_the_genomic_accession=43132357, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  254827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254827, CurrentID=0, Status='live', Symbol='NAALADL2', Aliases='', description='N-acetylated alpha-linked acidic dipeptidase-like 2', other_designations='N-acetylated alpha-linked acidic dipeptidase 2|NAALADase L2|glutamate carboxypeptidase II-type non-peptidase homologue', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=174440995, end_position_on_the_genomic_accession=175809510, orientation='plus', exon_count=22, OMIM=608806, no_hdr0=''),
  105374572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374572, CurrentID=0, Status='live', Symbol='LOC105374572', Aliases='', description='uncharacterized LOC105374572', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=44490730, end_position_on_the_genomic_accession=44493731, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131693, CurrentID=0, Status='live', Symbol='EIF4EP2', Aliases='', description='eukaryotic translation initiation factor 4E pseudogene 2', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49423287, end_position_on_the_genomic_accession=49424905, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374574, CurrentID=0, Status='live', Symbol='LOC105374574', Aliases='', description='uncharacterized LOC105374574', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45002564, end_position_on_the_genomic_accession=45004868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131695, CurrentID=0, Status='live', Symbol='RNF11P1', Aliases='', description='ring finger protein 11 pseudogene 1', other_designations='', map_location='2q32.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189928361, end_position_on_the_genomic_accession=189931042, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374576, CurrentID=0, Status='live', Symbol='LOC105374576', Aliases='', description='uncharacterized LOC105374576', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45207540, end_position_on_the_genomic_accession=45209655, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374577, CurrentID=0, Status='live', Symbol='LOC105374577', Aliases='', description='uncharacterized LOC105374577', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45323517, end_position_on_the_genomic_accession=45338630, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374578, CurrentID=0, Status='live', Symbol='LOC105374578', Aliases='', description='uncharacterized LOC105374578', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45963410, end_position_on_the_genomic_accession=45965365, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374579, CurrentID=0, Status='live', Symbol='LOC105374579', Aliases='', description='uncharacterized LOC105374579', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45786145, end_position_on_the_genomic_accession=45788599, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374580, CurrentID=0, Status='live', Symbol='LOC105374580', Aliases='', description='uncharacterized LOC105374580', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=45674701, end_position_on_the_genomic_accession=45675476, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374581, CurrentID=0, Status='live', Symbol='LOC105374581', Aliases='', description='uncharacterized LOC105374581', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46251663, end_position_on_the_genomic_accession=46256777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374582, CurrentID=0, Status='live', Symbol='LOC105374582', Aliases='', description='uncharacterized LOC105374582', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46392286, end_position_on_the_genomic_accession=46394232, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374583, CurrentID=0, Status='live', Symbol='LOC105374583', Aliases='', description='uncharacterized LOC105374583', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46291678, end_position_on_the_genomic_accession=46338138, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374584, CurrentID=0, Status='live', Symbol='LOC105374584', Aliases='', description='uncharacterized LOC105374584', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46636408, end_position_on_the_genomic_accession=46655968, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374585, CurrentID=0, Status='live', Symbol='LOC105374585', Aliases='', description='uncharacterized LOC105374585', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46675847, end_position_on_the_genomic_accession=46682929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374586, CurrentID=0, Status='live', Symbol='LOC105374586', Aliases='', description='uncharacterized LOC105374586', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46863534, end_position_on_the_genomic_accession=46875320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374587, CurrentID=0, Status='live', Symbol='LOC105374587', Aliases='', description='uncharacterized LOC105374587', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46777785, end_position_on_the_genomic_accession=46781473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100287704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287704, CurrentID=0, Status='live', Symbol='LOC100287704', Aliases='', description='uncharacterized LOC100287704', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63349070, end_position_on_the_genomic_accession=63351774, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379770, CurrentID=0, Status='live', Symbol='LOC105379770', Aliases='', description='uncharacterized LOC105379770', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374590, CurrentID=0, Status='live', Symbol='LOC105374590', Aliases='', description='uncharacterized LOC105374590', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47914338, end_position_on_the_genomic_accession=47915783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  123775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123775, CurrentID=0, Status='live', Symbol='C16orf46', Aliases='', description='chromosome 16 open reading frame 46', other_designations='', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=81053497, end_position_on_the_genomic_accession=81077267, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374592, CurrentID=0, Status='live', Symbol='LOC105374592', Aliases='', description='uncharacterized LOC105374592', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=47926499, end_position_on_the_genomic_accession=47937881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131713, CurrentID=0, Status='live', Symbol='RPL29P11', Aliases='RPL29_2_362', description='ribosomal protein L29 pseudogene 11', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37016398, end_position_on_the_genomic_accession=37017034, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374594, CurrentID=0, Status='live', Symbol='LOC105374594', Aliases='', description='uncharacterized LOC105374594', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48799305, end_position_on_the_genomic_accession=48815072, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374595, CurrentID=0, Status='live', Symbol='LOC105374595', Aliases='', description='uncharacterized LOC105374595', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=49307722, end_position_on_the_genomic_accession=49413132, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374596, CurrentID=0, Status='live', Symbol='LOC105374596', Aliases='', description='uncharacterized LOC105374596', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52370606, end_position_on_the_genomic_accession=52389837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374597, CurrentID=0, Status='live', Symbol='LOC105374597', Aliases='', description='uncharacterized LOC105374597', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=51976710, end_position_on_the_genomic_accession=51984284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374598, CurrentID=0, Status='live', Symbol='LOC105374598', Aliases='', description='uncharacterized LOC105374598', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52691303, end_position_on_the_genomic_accession=52698089, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374599, CurrentID=0, Status='live', Symbol='LOC105374599', Aliases='', description='uncharacterized LOC105374599', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52713264, end_position_on_the_genomic_accession=52717040, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374600, CurrentID=0, Status='live', Symbol='LOC105374600', Aliases='', description='uncharacterized LOC105374600', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53579791, end_position_on_the_genomic_accession=53603408, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374601, CurrentID=0, Status='live', Symbol='LOC105374601', Aliases='', description='uncharacterized LOC105374601', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=53604280, end_position_on_the_genomic_accession=53606080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374602, CurrentID=0, Status='live', Symbol='LOC105374602', Aliases='', description='uncharacterized LOC105374602', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24543, end_position_on_the_genomic_accession=51010, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  123787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123787, CurrentID=0, Status='live', Symbol='PRSS29P', Aliases='ISP2', description='protease, serine 29, pseudogene', other_designations='ISP2-like protein|implantation serine proteinase 2|implantation serine proteinase 2-like protein|protease, serine, 29 pseudogene|protease, serine, 29, pseudogene|putative serine protease 29', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1261003, end_position_on_the_genomic_accession=1264004, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374604, CurrentID=0, Status='live', Symbol='LOC105374604', Aliases='', description='uncharacterized LOC105374604', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=299836, end_position_on_the_genomic_accession=303312, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374605, CurrentID=0, Status='live', Symbol='LOC105374605', Aliases='', description='uncharacterized LOC105374605', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=292512, end_position_on_the_genomic_accession=297463, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131726, CurrentID=0, Status='live', Symbol='FAM83A-AS1', Aliases='HCCC11, HCCC11_v1, HCCC11_v2', description='FAM83A antisense RNA 1', other_designations='HCC-related HCC-C11_v3', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123201172, end_position_on_the_genomic_accession=123202743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  254863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254863, CurrentID=0, Status='live', Symbol='TMEM256', Aliases='C17orf61', description='transmembrane protein 256', other_designations='UPF0451 protein C17orf61', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7402974, end_position_on_the_genomic_accession=7404131, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374608, CurrentID=0, Status='live', Symbol='LOC105374608', Aliases='', description='uncharacterized LOC105374608', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=600501, end_position_on_the_genomic_accession=610822, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  282776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=282776, CurrentID=0, Status='live', Symbol='OR8V1P', Aliases='OR11-267', description='olfactory receptor family 8 subfamily V member 1 pseudogene', other_designations='olfactory receptor, family 8, subfamily V, member 1 pseudogene', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56188331, end_position_on_the_genomic_accession=56189251, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131730, CurrentID=0, Status='live', Symbol='RPL37AP7', Aliases='RPL37A_4_1128', description='ribosomal protein L37a pseudogene 7', other_designations='', map_location='11p14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27483819, end_position_on_the_genomic_accession=27484191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374611, CurrentID=0, Status='live', Symbol='LOC105374611', Aliases='', description='uncharacterized LOC105374611', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54146527, end_position_on_the_genomic_accession=54149618, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374612, CurrentID=0, Status='live', Symbol='LOC105374612', Aliases='', description='uncharacterized LOC105374612', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1110733, end_position_on_the_genomic_accession=1117428, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131733, CurrentID=0, Status='live', Symbol='USP30-AS1', Aliases='', description='USP30 antisense RNA 1', other_designations='USP30 antisense RNA 1 (non-protein coding)', map_location='12q24.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=109052041, end_position_on_the_genomic_accession=109053965, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374614, CurrentID=0, Status='live', Symbol='LOC105374614', Aliases='', description='uncharacterized LOC105374614', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1704253, end_position_on_the_genomic_accession=1725328, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374615, CurrentID=0, Status='live', Symbol='LOC105374615', Aliases='', description='uncharacterized LOC105374615', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54106303, end_position_on_the_genomic_accession=54108991, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374616, CurrentID=0, Status='live', Symbol='LOC105374616', Aliases='', description='uncharacterized LOC105374616', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1991981, end_position_on_the_genomic_accession=1993420, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374617, CurrentID=0, Status='live', Symbol='LOC105374617', Aliases='', description='uncharacterized LOC105374617', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2066842, end_position_on_the_genomic_accession=2071611, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374618, CurrentID=0, Status='live', Symbol='LOC105374618', Aliases='', description='uncharacterized LOC105374618', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=2093662, end_position_on_the_genomic_accession=2119386, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  123803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123803, CurrentID=0, Status='live', Symbol='NTAN1', Aliases='PNAA, PNAD', description='N-terminal asparagine amidase', other_designations='protein N-terminal Asn amidase|protein N-terminal asparagine amidase|protein NH2-terminal asparagine deamidase|protein NTN-amidase', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15037853, end_position_on_the_genomic_accession=15056079, orientation='minus', exon_count=11, OMIM=615367, no_hdr0=''),
  100131740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131740, CurrentID=0, Status='live', Symbol='CIR1P1', Aliases='', description='corepressor interacting with RBPJ, 1 pseudogene 1', other_designations='', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154559315, end_position_on_the_genomic_accession=154560518, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374621, CurrentID=0, Status='live', Symbol='LOC105374621', Aliases='', description='uncharacterized LOC105374621', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3035625, end_position_on_the_genomic_accession=3043011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374622, CurrentID=0, Status='live', Symbol='LOC105374622', Aliases='', description='uncharacterized LOC105374622', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3160158, end_position_on_the_genomic_accession=3161379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254879, CurrentID=0, Status='live', Symbol='OR2T6', Aliases='OR2T6P, OR2T9, OST703', description='olfactory receptor family 2 subfamily T member 6', other_designations='olfactory receptor 2T9|olfactory receptor, family 2, subfamily T, member 6 pseudogene|seven transmembrane helix receptor', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248387609, end_position_on_the_genomic_accession=248388535, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374624, CurrentID=0, Status='live', Symbol='LOC105374624', Aliases='', description='uncharacterized LOC105374624', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3452816, end_position_on_the_genomic_accession=3462314, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374625, CurrentID=0, Status='live', Symbol='LOC105374625', Aliases='', description='uncharacterized LOC105374625', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3778347, end_position_on_the_genomic_accession=3788064, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374626, CurrentID=0, Status='live', Symbol='LOC105374626', Aliases='', description='uncharacterized LOC105374626', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=3961158, end_position_on_the_genomic_accession=3964006, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  123811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123811, CurrentID=0, Status='live', Symbol='FOPNL', Aliases='C16orf63, FOR20, PHSECRG2', description='FGFR1OP N-terminal like', other_designations='FOP-related protein of 20 kDa|lisH domain-containing protein C16orf63|pluripotent embryonic stem cell-related protein', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15865719, end_position_on_the_genomic_accession=15888649, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104326052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326052, CurrentID=0, Status='live', Symbol='NRIR', Aliases='lncRNA-CMPK2', description='negative regulator of interferon response (non-protein coding)', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6828554, end_position_on_the_genomic_accession=6840464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104326053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326053, CurrentID=0, Status='live', Symbol='LINC00363', Aliases='TCONS_00021586', description='long intergenic non-protein coding RNA 363', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=93045341, end_position_on_the_genomic_accession=93057926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104326054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326054, CurrentID=0, Status='live', Symbol='LINC00399', Aliases='TCONS_00021596', description='long intergenic non-protein coding RNA 399', other_designations='', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=109400696, end_position_on_the_genomic_accession=109401641, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104326055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326055, CurrentID=0, Status='live', Symbol='APOA1-AS', Aliases='', description='APOA1 antisense RNA', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=116836117, end_position_on_the_genomic_accession=116855729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104326056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326056, CurrentID=0, Status='live', Symbol='ZNF511-PRAP1', Aliases='', description='ZNF511-PRAP1 readthrough', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104326057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326057, CurrentID=0, Status='live', Symbol='GACAT1', Aliases='LINC00876', description='gastric cancer associated transcript 1 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 876', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=107754112, end_position_on_the_genomic_accession=107822129, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104326058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326058, CurrentID=0, Status='live', Symbol='SMAD1-AS1', Aliases='', description='SMAD1 antisense RNA 1', other_designations='ENST00000513542', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145514578, end_position_on_the_genomic_accession=145517194, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131755, CurrentID=0, Status='live', Symbol='ARMCX4', Aliases='CXorf35, GASP4', description='armadillo repeat containing, X-linked 4', other_designations='armadillo repeat containing, X-linked 4 pseudogene', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101418263, end_position_on_the_genomic_accession=101535988, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100131756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131756, CurrentID=0, Status='live', Symbol='LOC100131756', Aliases='', description='uncharacterized LOC100131756', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=43705872, end_position_on_the_genomic_accession=43707974, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374637, CurrentID=0, Status='live', Symbol='LOC105374637', Aliases='', description='uncharacterized LOC105374637', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=5843992, end_position_on_the_genomic_accession=5850465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374638, CurrentID=0, Status='live', Symbol='LOC105374638', Aliases='', description='uncharacterized LOC105374638', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6339488, end_position_on_the_genomic_accession=6369751, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100131759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131759, CurrentID=0, Status='live', Symbol='RPL19P3', Aliases='RPL19_1_59', description='ribosomal protein L19 pseudogene 3', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63788641, end_position_on_the_genomic_accession=63789334, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131760, CurrentID=0, Status='live', Symbol='GXYLT1P5', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 5', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42974789, end_position_on_the_genomic_accession=42976241, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131761, CurrentID=0, Status='live', Symbol='RPS12P12', Aliases='RPS12_5_662', description='ribosomal protein S12 pseudogene 12', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17953569, end_position_on_the_genomic_accession=17953942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374642, CurrentID=0, Status='live', Symbol='LOC105374642', Aliases='', description='uncharacterized LOC105374642', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6870743, end_position_on_the_genomic_accession=6888120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131763, CurrentID=0, Status='live', Symbol='LOC100131763', Aliases='', description='uncharacterized LOC100131763', other_designations='', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=242457243, end_position_on_the_genomic_accession=242457629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100423006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423006, CurrentID=0, Status='live', Symbol='MIR3179-3', Aliases='', description='microRNA 3179-3', other_designations='hsa-mir-3179-3', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18411894, end_position_on_the_genomic_accession=18411977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374645, CurrentID=0, Status='live', Symbol='LOC105374645', Aliases='', description='uncharacterized LOC105374645', other_designations='CTD-2296D1.3|CTD-2296D1.4', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=7362936, end_position_on_the_genomic_accession=7373142, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100131766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131766, CurrentID=0, Status='live', Symbol='OK/SW-CL.58', Aliases='', description='OK/SW-CL.58', other_designations='', map_location='13q12.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=29064863, end_position_on_the_genomic_accession=29065033, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374647, CurrentID=0, Status='live', Symbol='LOC105374647', Aliases='', description='uncharacterized LOC105374647', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8525155, end_position_on_the_genomic_accession=8563160, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105374648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374648, CurrentID=0, Status='live', Symbol='LOC105374648', Aliases='', description='uncharacterized LOC105374648', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9854396, end_position_on_the_genomic_accession=9891237, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374649, CurrentID=0, Status='live', Symbol='LOC105374649', Aliases='', description='uncharacterized LOC105374649', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=9751414, end_position_on_the_genomic_accession=9776840, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374650, CurrentID=0, Status='live', Symbol='LOC105374650', Aliases='', description='uncharacterized LOC105374650', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10207378, end_position_on_the_genomic_accession=10223999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374651, CurrentID=0, Status='live', Symbol='LOC105374651', Aliases='', description='uncharacterized LOC105374651', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10316100, end_position_on_the_genomic_accession=10332129, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374652, CurrentID=0, Status='live', Symbol='LOC105374652', Aliases='', description='uncharacterized LOC105374652', other_designations='CTD-2154B17.4', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10767738, end_position_on_the_genomic_accession=10770295, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374653, CurrentID=0, Status='live', Symbol='LOC105374653', Aliases='', description='uncharacterized LOC105374653', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54674382, end_position_on_the_genomic_accession=54684908, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  254910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254910, CurrentID=0, Status='live', Symbol='LCE5A', Aliases='LEP18, SPRL5A', description='late cornified envelope 5A', other_designations='late envelope protein 18|small proline rich-like (epidermal differentiation complex) 5A|small proline-rich-like epidermal differentiation complex protein 5A', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152510844, end_position_on_the_genomic_accession=152512177, orientation='plus', exon_count=2, OMIM=612619, no_hdr0=''),
  100131775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131775, CurrentID=0, Status='live', Symbol='PSMA2P2', Aliases='', description='proteasome (prosome, macropain) subunit, alpha type, 2 pseudogene 2', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93940086, end_position_on_the_genomic_accession=93942877, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374656, CurrentID=0, Status='live', Symbol='LOC105374656', Aliases='', description='uncharacterized LOC105374656', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12354684, end_position_on_the_genomic_accession=12368765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131777, CurrentID=0, Status='live', Symbol='DPPA2P4', Aliases='', description='developmental pluripotency associated 2 pseudogene 4', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56686837, end_position_on_the_genomic_accession=56688298, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374658, CurrentID=0, Status='live', Symbol='LOC105374658', Aliases='', description='uncharacterized LOC105374658', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=12888001, end_position_on_the_genomic_accession=13032890, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374659, CurrentID=0, Status='live', Symbol='LOC105374659', Aliases='', description='uncharacterized LOC105374659', other_designations='CTC-305H11.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=13144042, end_position_on_the_genomic_accession=13175981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374660, CurrentID=0, Status='live', Symbol='LOC105374660', Aliases='', description='uncharacterized LOC105374660', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=13516343, end_position_on_the_genomic_accession=13578394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374661, CurrentID=0, Status='live', Symbol='LOC105374661', Aliases='', description='uncharacterized LOC105374661', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14883562, end_position_on_the_genomic_accession=14892660, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374662, CurrentID=0, Status='live', Symbol='LOC105374662', Aliases='', description='uncharacterized LOC105374662', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55080230, end_position_on_the_genomic_accession=55086883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374663, CurrentID=0, Status='live', Symbol='LOC105374663', Aliases='', description='uncharacterized LOC105374663', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=16441947, end_position_on_the_genomic_accession=16446524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131784, CurrentID=0, Status='live', Symbol='ARL2BPP8', Aliases='', description='ADP ribosylation factor like GTPase 2 binding protein pseudogene 8', other_designations='ADP-ribosylation factor like 2 binding protein pseudogene 8|ADP-ribosylation factor like GTPase 2 binding protein pseudogene 8|ADP-ribosylation factor-like 2 binding protein pseudogene 8', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=54752875, end_position_on_the_genomic_accession=54754677, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374665, CurrentID=0, Status='live', Symbol='LOC105374665', Aliases='', description='uncharacterized LOC105374665', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17491582, end_position_on_the_genomic_accession=17495538, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374666, CurrentID=0, Status='live', Symbol='LOC105374666', Aliases='', description='uncharacterized LOC105374666', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17743834, end_position_on_the_genomic_accession=17785733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131787, CurrentID=0, Status='live', Symbol='RPS27P7', Aliases='RPS27_3_126', description='ribosomal protein S27 pseudogene 7', other_designations='', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173741643, end_position_on_the_genomic_accession=173741987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131788, CurrentID=0, Status='live', Symbol='RPL26P29', Aliases='RPL26_14_1054', description='ribosomal protein L26 pseudogene 29', other_designations='', map_location='10q21.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68424560, end_position_on_the_genomic_accession=68424997, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131789, CurrentID=0, Status='live', Symbol='VN1R46P', Aliases='', description='vomeronasal 1 receptor 46 pseudogene', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43219928, end_position_on_the_genomic_accession=43220809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374670, CurrentID=0, Status='live', Symbol='LOC105374670', Aliases='', description='uncharacterized LOC105374670', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=19035492, end_position_on_the_genomic_accession=19038790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374671, CurrentID=0, Status='live', Symbol='LOC105374671', Aliases='', description='uncharacterized LOC105374671', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20290977, end_position_on_the_genomic_accession=20293079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131792, CurrentID=0, Status='live', Symbol='LOC100131792', Aliases='', description='uncharacterized LOC100131792', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=126187126, end_position_on_the_genomic_accession=126189758, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374673, CurrentID=0, Status='live', Symbol='LOC105374673', Aliases='', description='uncharacterized LOC105374673', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=20849717, end_position_on_the_genomic_accession=20875809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374674, CurrentID=0, Status='live', Symbol='LOC105374674', Aliases='', description='uncharacterized LOC105374674', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55142272, end_position_on_the_genomic_accession=55170440, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929123, CurrentID=0, Status='live', Symbol='LOC101929123', Aliases='', description='uncharacterized LOC101929123', other_designations='', map_location='4p15.32', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=17171757, end_position_on_the_genomic_accession=17186057, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100131796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131796, CurrentID=0, Status='live', Symbol='DACOR1', Aliases='LP2570, TCONS_00023265', description='DNMT1-associated colon cancer repressed lncRNA 1', other_designations='', map_location='15q22.33', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=65123616, end_position_on_the_genomic_accession=65123969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131797, CurrentID=0, Status='live', Symbol='RPL15P19', Aliases='', description='ribosomal protein L15 pseudogene', other_designations='', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=49177301, end_position_on_the_genomic_accession=49178795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374678, CurrentID=0, Status='live', Symbol='LOC105374678', Aliases='', description='uncharacterized LOC105374678', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21031933, end_position_on_the_genomic_accession=21116540, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105374679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374679, CurrentID=0, Status='live', Symbol='LOC105374679', Aliases='', description='uncharacterized LOC105374679', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55824158, end_position_on_the_genomic_accession=55874935, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131800, CurrentID=0, Status='live', Symbol='UBE2D3P4', Aliases='', description='ubiquitin conjugating enzyme E2D 3 pseudogene 4', other_designations='ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast) pseudogene 4|ubiquitin-conjugating enzyme E2D 3 pseudogene 4|ubiquitin-conjugating enzyme pseudogene', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84020414, end_position_on_the_genomic_accession=84020742, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131801, CurrentID=0, Status='live', Symbol='PET100', Aliases='C19orf79', description='PET100 homolog', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7629785, end_position_on_the_genomic_accession=7631624, orientation='plus', exon_count=5, OMIM=614770, no_hdr0=''),
  105374682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374682, CurrentID=0, Status='live', Symbol='LOC105374682', Aliases='', description='uncharacterized LOC105374682', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=22212449, end_position_on_the_genomic_accession=22216197, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374683, CurrentID=0, Status='live', Symbol='LOC105374683', Aliases='', description='uncharacterized LOC105374683', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21823363, end_position_on_the_genomic_accession=21826613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374684, CurrentID=0, Status='live', Symbol='LOC105374684', Aliases='', description='uncharacterized LOC105374684', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=21741574, end_position_on_the_genomic_accession=21802252, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131805, CurrentID=0, Status='live', Symbol='RPL6P18', Aliases='RPL6_13_667', description='ribosomal protein L6 pseudogene 18', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23102659, end_position_on_the_genomic_accession=23103577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374686, CurrentID=0, Status='live', Symbol='LOC105374686', Aliases='', description='uncharacterized LOC105374686', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23039054, end_position_on_the_genomic_accession=23093949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374687, CurrentID=0, Status='live', Symbol='LOC105374687', Aliases='', description='uncharacterized LOC105374687', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23159844, end_position_on_the_genomic_accession=23171595, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  123872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123872, CurrentID=0, Status='live', Symbol='DNAAF1', Aliases='CILD13, LRRC50, ODA7', description='dynein, axonemal, assembly factor 1', other_designations='leucine rich repeat containing 50|leucine-rich repeat-containing protein 50|outer row dynein assembly 7 homolog', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84145260, end_position_on_the_genomic_accession=84177918, orientation='plus', exon_count=15, OMIM=613190, no_hdr0=''),
  100271569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271569, CurrentID=0, Status='live', Symbol='RPS26P44', Aliases='RPS26_16_1195', description='ribosomal protein S26 pseudogene 44', other_designations='', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3468923, end_position_on_the_genomic_accession=3469359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131810, CurrentID=0, Status='live', Symbol='ADH5P3', Aliases='', description='alcohol dehydrogenase 5 (class III), chi polypeptide, pseudogene 3', other_designations='alcohol dehydrogenase 5 (class III), chi polypeptide pseudogene', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=240170143, end_position_on_the_genomic_accession=240171530, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374691, CurrentID=0, Status='live', Symbol='LOC105374691', Aliases='', description='uncharacterized LOC105374691', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=24840762, end_position_on_the_genomic_accession=24851023, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  123876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123876, CurrentID=0, Status='live', Symbol='ACSM2A', Aliases='A-923A4.1, ACSM2', description='acyl-CoA synthetase medium-chain family member 2A', other_designations='Homolog of rat kidney-specific (KS)|acyl-CoA synthetase medium-chain family member 2|butyrate--CoA ligase 2A|butyryl-coenzyme A synthetase 2A|middle-chain acyl-CoA synthetase 2A', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20451526, end_position_on_the_genomic_accession=20487669, orientation='plus', exon_count=15, OMIM=614358, no_hdr0=''),
  105374693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374693, CurrentID=0, Status='live', Symbol='LOC105374693', Aliases='', description='spidroin-2-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=25190596, end_position_on_the_genomic_accession=25520991, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  254950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254950, CurrentID=0, Status='live', Symbol='KRTAP15-1', Aliases='KAP15.1', description='keratin associated protein 15-1', other_designations='', map_location='21q22.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30440328, end_position_on_the_genomic_accession=30440780, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  123879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123879, CurrentID=0, Status='live', Symbol='DCUN1D3', Aliases='44M2.4, SCCRO3', description='DCN1, defective in cullin neddylation 1, domain containing 3', other_designations='DCUN1 domain-containing protein 3|defective in cullin neddylation protein 1-like protein 3|squamous cell carcinoma-related oncogene 3', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20858074, end_position_on_the_genomic_accession=20900239, orientation='minus', exon_count=3, OMIM=616167, no_hdr0=''),
  100131816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131816, CurrentID=0, Status='live', Symbol='UBE2DNL', Aliases='', description='ubiquitin conjugating enzyme E2D N-terminal like (pseudogene)', other_designations='ubiquitin-conjugating enzyme E2D N-terminal like (pseudogene)', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=84934151, end_position_on_the_genomic_accession=84934890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374697, CurrentID=0, Status='live', Symbol='LOC105374697', Aliases='', description='uncharacterized LOC105374697', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57720807, end_position_on_the_genomic_accession=57726143, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374698, CurrentID=0, Status='live', Symbol='LOC105374698', Aliases='', description='uncharacterized LOC105374698', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=28285979, end_position_on_the_genomic_accession=28287763, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929127, CurrentID=0, Status='live', Symbol='PGM5P3-AS1', Aliases='FAM233B', description='PGM5P3 antisense RNA 1', other_designations='family with sequence similarity 233, member B', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72690, end_position_on_the_genomic_accession=88826, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  254956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254956, CurrentID=0, Status='live', Symbol='MORN5', Aliases='C9orf113, C9orf18', description='MORN repeat containing 5', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122159858, end_position_on_the_genomic_accession=122200088, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105374701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374701, CurrentID=0, Status='live', Symbol='LOC105374701', Aliases='', description='uncharacterized LOC105374701', other_designations='CTD-2134P3.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=28548406, end_position_on_the_genomic_accession=28809029, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  254958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254958, CurrentID=0, Status='live', Symbol='REXO1L1P', Aliases='GOR, REXO1L1', description='REX1, RNA exonuclease 1 homolog-like 1, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 1, pseudogene|REX1, RNA exonuclease 1 homolog-like 1|antigen GOR homolog|exonuclease GOR', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85661470, end_position_on_the_genomic_accession=85663497, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100131823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131823, CurrentID=0, Status='live', Symbol='SUMO2P7', Aliases='', description='SUMO2 pseudogene 7', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 7', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50235845, end_position_on_the_genomic_accession=50236129, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374704, CurrentID=0, Status='live', Symbol='LOC105374704', Aliases='', description='uncharacterized LOC105374704', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=29880526, end_position_on_the_genomic_accession=29882103, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131825, CurrentID=0, Status='live', Symbol='CADM3-AS1', Aliases='', description='CADM3 antisense RNA 1', other_designations='CTA-134P22.2', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159195985, end_position_on_the_genomic_accession=159202422, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131826, CurrentID=0, Status='live', Symbol='LOC100131826', Aliases='', description='TSSP3028', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=77905998, end_position_on_the_genomic_accession=77907656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100131827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131827, CurrentID=0, Status='live', Symbol='ZNF717', Aliases='X17, ZNF838', description='zinc finger protein 717', other_designations='krueppel-like factor X17|kruppel-like zinc finger factor X17|zinc finger protein 838', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75716445, end_position_on_the_genomic_accession=75785583, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100131828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131828, CurrentID=0, Status='live', Symbol='CIR1P2', Aliases='', description='corepressor interacting with RBPJ, 1 pseudogene 2', other_designations='', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114713439, end_position_on_the_genomic_accession=114714951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374709, CurrentID=0, Status='live', Symbol='LOC105374709', Aliases='', description='uncharacterized LOC105374709', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31201505, end_position_on_the_genomic_accession=31213922, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374710, CurrentID=0, Status='live', Symbol='LOC105374710', Aliases='', description='uncharacterized LOC105374710', other_designations='CTD-2152M20.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32106149, end_position_on_the_genomic_accession=32122020, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374711, CurrentID=0, Status='live', Symbol='LOC105374711', Aliases='', description='uncharacterized LOC105374711', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32097671, end_position_on_the_genomic_accession=32098396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374712, CurrentID=0, Status='live', Symbol='LOC105374712', Aliases='', description='uncharacterized LOC105374712', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=31846412, end_position_on_the_genomic_accession=31855024, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374713, CurrentID=0, Status='live', Symbol='LOC105374713', Aliases='', description='uncharacterized LOC105374713', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32505640, end_position_on_the_genomic_accession=32571326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374714, CurrentID=0, Status='live', Symbol='LOC105374714', Aliases='', description='uncharacterized LOC105374714', other_designations='CTD-2004A9.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=32634208, end_position_on_the_genomic_accession=32652024, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374715, CurrentID=0, Status='live', Symbol='LOC105374715', Aliases='', description='uncharacterized LOC105374715', other_designations='CTD-2066L21.2|CTD-2066L21.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33010865, end_position_on_the_genomic_accession=33023321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374716, CurrentID=0, Status='live', Symbol='LOC105374716', Aliases='', description='uncharacterized LOC105374716', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=33229224, end_position_on_the_genomic_accession=33258419, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  254973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=254973, CurrentID=0, Status='live', Symbol='OR1L4', Aliases='OR1L5, OR9-29, OR9-E, OST046', description='olfactory receptor family 1 subfamily L member 4', other_designations='olfactory receptor 1L5|olfactory receptor 9-E|olfactory receptor OR9-29|olfactory receptor, family 1, subfamily L, member 4|olfactory receptor, family 1, subfamily L, member 5', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122723990, end_position_on_the_genomic_accession=122724925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374718, CurrentID=0, Status='live', Symbol='LOC105374718', Aliases='', description='uncharacterized LOC105374718', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34156876, end_position_on_the_genomic_accession=34169120, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374719, CurrentID=0, Status='live', Symbol='LOC105374719', Aliases='', description='uncharacterized LOC105374719', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34264917, end_position_on_the_genomic_accession=34269419, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  123904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123904, CurrentID=0, Status='live', Symbol='NRN1L', Aliases='UNQ2446', description='neuritin 1-like', other_designations='MRCC2446', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67884878, end_position_on_the_genomic_accession=67886368, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374721, CurrentID=0, Status='live', Symbol='LOC105374721', Aliases='', description='uncharacterized LOC105374721', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34585802, end_position_on_the_genomic_accession=34590317, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131842, CurrentID=0, Status='live', Symbol='HNRNPCP8', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 8', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='11q14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=87024584, end_position_on_the_genomic_accession=87025776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100421803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421803, CurrentID=0, Status='live', Symbol='ENPP7P9', Aliases='', description='ectonucleotide pyrophosphatase/phosphodiesterase 7 pseudogene 9', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3967033, end_position_on_the_genomic_accession=3967660, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374724, CurrentID=0, Status='live', Symbol='LOC105374724', Aliases='', description='uncharacterized LOC105374724', other_designations='CTD-2113L7.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=35676736, end_position_on_the_genomic_accession=35686311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376356, CurrentID=0, Status='live', Symbol='LOC105376356', Aliases='', description='uncharacterized LOC105376356', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3305132, end_position_on_the_genomic_accession=3313998, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131846, CurrentID=0, Status='live', Symbol='RPL23AP83', Aliases='RPL23A_38_1781', description='ribosomal protein L23a pseudogene 83', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56242937, end_position_on_the_genomic_accession=56243409, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374727, CurrentID=0, Status='live', Symbol='LOC105374727', Aliases='', description='uncharacterized LOC105374727', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37249051, end_position_on_the_genomic_accession=37250644, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374728, CurrentID=0, Status='live', Symbol='LOC105374728', Aliases='', description='uncharacterized LOC105374728', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37916469, end_position_on_the_genomic_accession=37920929, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374729, CurrentID=0, Status='live', Symbol='LOC105374729', Aliases='', description='uncharacterized LOC105374729', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37948305, end_position_on_the_genomic_accession=37951055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374730, CurrentID=0, Status='live', Symbol='LOC105374730', Aliases='', description='uncharacterized LOC105374730', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38074801, end_position_on_the_genomic_accession=38084348, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374731, CurrentID=0, Status='live', Symbol='LOC105374731', Aliases='', description='uncharacterized LOC105374731', other_designations='CTD-2116N24.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38024191, end_position_on_the_genomic_accession=38183932, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374732, CurrentID=0, Status='live', Symbol='LOC105374732', Aliases='', description='uncharacterized LOC105374732', other_designations='CTD-2263F21.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38460778, end_position_on_the_genomic_accession=38462817, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374733, CurrentID=0, Status='live', Symbol='LOC105374733', Aliases='', description='uncharacterized LOC105374733', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38812099, end_position_on_the_genomic_accession=38813637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374734, CurrentID=0, Status='live', Symbol='LOC105374734', Aliases='', description='uncharacterized LOC105374734', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=38604654, end_position_on_the_genomic_accession=38608420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100423009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423009, CurrentID=0, Status='live', Symbol='MIR4329', Aliases='', description='microRNA 4329', other_designations='hsa-mir-4329', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=112780718, end_position_on_the_genomic_accession=112780788, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  123920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123920, CurrentID=0, Status='live', Symbol='CMTM3', Aliases='BNAS2, CKLFSF3', description='CKLF-like MARVEL transmembrane domain containing 3', other_designations='chemokine-like factor superfamily member 3', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=66604032, end_position_on_the_genomic_accession=66613892, orientation='plus', exon_count=6, OMIM=607886, no_hdr0=''),
  105374737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374737, CurrentID=0, Status='live', Symbol='LOC105374737', Aliases='', description='uncharacterized LOC105374737', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40599880, end_position_on_the_genomic_accession=40631723, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374738, CurrentID=0, Status='live', Symbol='LOC105374738', Aliases='', description='uncharacterized LOC105374738', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=40678776, end_position_on_the_genomic_accession=40679489, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131859, CurrentID=0, Status='live', Symbol='LOC100131859', Aliases='', description='SAP domain containing ribonucleoprotein pseudogene', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99442338, end_position_on_the_genomic_accession=99443509, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131860, CurrentID=0, Status='live', Symbol='LOC100131860', Aliases='', description='uncharacterized LOC100131860', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=81444997, end_position_on_the_genomic_accession=81445900, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374741, CurrentID=0, Status='live', Symbol='LOC105374741', Aliases='', description='uncharacterized LOC105374741', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41281222, end_position_on_the_genomic_accession=41350070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374742, CurrentID=0, Status='live', Symbol='LOC105374742', Aliases='', description='uncharacterized LOC105374742', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=41733479, end_position_on_the_genomic_accession=41737454, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131863, CurrentID=0, Status='live', Symbol='RPS18P5', Aliases='RPS18, RPS18_1_389', description='ribosomal protein S18 pseudogene 5', other_designations='', map_location='3q12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=101576451, end_position_on_the_genomic_accession=101576990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131864, CurrentID=0, Status='live', Symbol='LOC100131864', Aliases='', description='uncharacterized LOC100131864', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=6160103, end_position_on_the_genomic_accession=6161095, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  648217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648217, CurrentID=0, Status='live', Symbol='RPL37P23', Aliases='RPL37_8_1665', description='ribosomal protein L37 pseudogene 23', other_designations='CTC-471J1.9', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52142626, end_position_on_the_genomic_accession=52143396, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131866, CurrentID=0, Status='live', Symbol='RPS2P9', Aliases='RPS2_5_144', description='ribosomal protein S2 pseudogene 9', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198949147, end_position_on_the_genomic_accession=198949849, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374747, CurrentID=0, Status='live', Symbol='LOC105374747', Aliases='', description='uncharacterized LOC105374747', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=42974417, end_position_on_the_genomic_accession=42975949, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374748, CurrentID=0, Status='live', Symbol='LOC105374748', Aliases='', description='uncharacterized LOC105374748', other_designations='CTD-2035E11.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43018857, end_position_on_the_genomic_accession=43024240, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374749, CurrentID=0, Status='live', Symbol='LOC105374749', Aliases='', description='uncharacterized LOC105374749', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43511264, end_position_on_the_genomic_accession=43529468, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374750, CurrentID=0, Status='live', Symbol='LOC105374750', Aliases='', description='uncharacterized LOC105374750', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43483916, end_position_on_the_genomic_accession=43509815, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374751, CurrentID=0, Status='live', Symbol='LOC105374751', Aliases='', description='uncharacterized LOC105374751', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=45226524, end_position_on_the_genomic_accession=45228894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374752, CurrentID=0, Status='live', Symbol='LOC105374752', Aliases='', description='atherin-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51194, end_position_on_the_genomic_accession=57883, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105374753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374753, CurrentID=0, Status='live', Symbol='LOC105374753', Aliases='', description='class E vacuolar protein-sorting machinery protein hse1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=34656306, end_position_on_the_genomic_accession=34658489, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131874, CurrentID=0, Status='live', Symbol='CFL1P3', Aliases='CFLL3, CFLP3', description='cofilin 1 (non-muscle) pseudogene 3', other_designations='cofilin pseudogene 3|cofilin-like 3', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63842791, end_position_on_the_genomic_accession=63843807, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374755, CurrentID=0, Status='live', Symbol='LOC105374755', Aliases='', description='uncharacterized LOC105374755', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60264918, end_position_on_the_genomic_accession=60269545, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374756, CurrentID=0, Status='live', Symbol='LOC105374756', Aliases='', description='uncharacterized LOC105374756', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60334399, end_position_on_the_genomic_accession=60353073, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374757, CurrentID=0, Status='live', Symbol='LOC105374757', Aliases='', description='uncharacterized LOC105374757', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60470081, end_position_on_the_genomic_accession=60472959, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374758, CurrentID=0, Status='live', Symbol='LOC105374758', Aliases='', description='uncharacterized LOC105374758', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61151433, end_position_on_the_genomic_accession=61162483, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131879, CurrentID=0, Status='live', Symbol='ZYG11AP1', Aliases='', description='zyg-11 family member A, cell cycle regulator pseudogene 1', other_designations='zyg-11 homolog A pseudogene 1', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=102611104, end_position_on_the_genomic_accession=102611969, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374760, CurrentID=0, Status='live', Symbol='LOC105374760', Aliases='', description='uncharacterized LOC105374760', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61793509, end_position_on_the_genomic_accession=61798714, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131881, CurrentID=0, Status='live', Symbol='EI24P3', Aliases='', description='etoposide induced 2.4 pseudogene 3', other_designations='etoposide induced 2.4 mRNA pseudogene', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44524744, end_position_on_the_genomic_accession=44525805, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374762, CurrentID=0, Status='live', Symbol='LOC105374762', Aliases='', description='uncharacterized LOC105374762', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=62535872, end_position_on_the_genomic_accession=62538475, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104326187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326187, CurrentID=0, Status='live', Symbol='EA8', Aliases='', description='Episodic ataxia, type 8', other_designations='', map_location='1p36.13-p34.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616055, no_hdr0=''),
  104326188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326188, CurrentID=0, Status='live', Symbol='LINC01202', Aliases='', description='long intergenic non-protein coding RNA 1202', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161910462, end_position_on_the_genomic_accession=162001196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104326189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326189, CurrentID=0, Status='live', Symbol='LMO7DN-IT1', Aliases='C13orf45-IT1, TCONS_00021569', description='LMO7DN intronic transcript 1', other_designations='', map_location='13q22.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=75876886, end_position_on_the_genomic_accession=75881127, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  255022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255022, CurrentID=0, Status='live', Symbol='CALHM1', Aliases='FAM26C', description='calcium homeostasis modulator 1', other_designations='family with sequence similarity 26, member C', map_location='10q24.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=103453387, end_position_on_the_genomic_accession=103458891, orientation='minus', exon_count=2, OMIM=612234, no_hdr0=''),
  104326191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326191, CurrentID=0, Status='live', Symbol='LINC01336', Aliases='', description='long intergenic non-protein coding RNA 1336', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75047719, end_position_on_the_genomic_accession=75052643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104326192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326192, CurrentID=0, Status='live', Symbol='PTPRJ-AS1', Aliases='', description='PTPRJ antisense RNA 1', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104326193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104326193, CurrentID=0, Status='live', Symbol='DLX2-AS1', Aliases='TCONS_00003049', description='DLX2 antisense RNA 1 (head to head)', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=172103006, end_position_on_the_genomic_accession=172109982, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374770, CurrentID=0, Status='live', Symbol='LOC105374770', Aliases='', description='uncharacterized LOC105374770', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64438586, end_position_on_the_genomic_accession=64445588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255027, CurrentID=0, Status='live', Symbol='MPV17L', Aliases='M-LPH, MLPH1, MLPH21, MPV17L', description='MPV17 mitochondrial membrane protein-like', other_designations='M-LP homolog|Mpv17-like protein type 1|Mpv17-like protein type 2', map_location='16p13.11', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=15395698, end_position_on_the_genomic_accession=15409686, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374772, CurrentID=0, Status='live', Symbol='LOC105374772', Aliases='', description='uncharacterized LOC105374772', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64581257, end_position_on_the_genomic_accession=64607175, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131893, CurrentID=0, Status='live', Symbol='RPL3P11', Aliases='RPL3_7_963', description='ribosomal protein L3 pseudogene 11', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14040657, end_position_on_the_genomic_accession=14041887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131894, CurrentID=0, Status='live', Symbol='LOC100131894', Aliases='', description='uncharacterized LOC100131894', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=131360918, end_position_on_the_genomic_accession=131361883, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  255031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255031, CurrentID=0, Status='live', Symbol='LINC00957', Aliases='', description='long intergenic non-protein coding RNA 957', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44039049, end_position_on_the_genomic_accession=44044296, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374776, CurrentID=0, Status='live', Symbol='LOC105374776', Aliases='', description='uncharacterized LOC105374776', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64788893, end_position_on_the_genomic_accession=64797081, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131897, CurrentID=0, Status='live', Symbol='FAM196B', Aliases='C5orf57', description='family with sequence similarity 196 member B', other_designations='family with sequence similarity 196, member B', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169861291, end_position_on_the_genomic_accession=169980740, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374778, CurrentID=0, Status='live', Symbol='LOC105374778', Aliases='', description='uncharacterized LOC105374778', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65044712, end_position_on_the_genomic_accession=65056197, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131899, CurrentID=0, Status='live', Symbol='E2F4P1', Aliases='', description='E2F transcription factor 4, p107/p130-binding pseudogene 1', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39553817, end_position_on_the_genomic_accession=39554940, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374780, CurrentID=0, Status='live', Symbol='LOC105374780', Aliases='', description='uncharacterized LOC105374780', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65442462, end_position_on_the_genomic_accession=65456524, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374781, CurrentID=0, Status='live', Symbol='LOC105374781', Aliases='', description='uncharacterized LOC105374781', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=65487255, end_position_on_the_genomic_accession=65519262, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131902, CurrentID=0, Status='live', Symbol='KRTAP25-1', Aliases='KAP25.1', description='keratin associated protein 25-1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30289145, end_position_on_the_genomic_accession=30289514, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929141, CurrentID=0, Status='live', Symbol='LOC101929141', Aliases='', description='coiled-coil domain-containing protein 144B-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18834662, end_position_on_the_genomic_accession=18865564, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374784, CurrentID=0, Status='live', Symbol='LOC105374784', Aliases='', description='uncharacterized LOC105374784', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66441476, end_position_on_the_genomic_accession=66448980, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131905, CurrentID=0, Status='live', Symbol='RPS27P21', Aliases='RPS27_9_1229', description='ribosomal protein S27 pseudogene 21', other_designations='', map_location='12q12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=42678817, end_position_on_the_genomic_accession=42679158, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  123970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=123970, CurrentID=0, Status='live', Symbol='C16orf78', Aliases='', description='chromosome 16 open reading frame 78', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=49373808, end_position_on_the_genomic_accession=49399431, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  255043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255043, CurrentID=0, Status='live', Symbol='TMEM86B', Aliases='', description='transmembrane protein 86B', other_designations='alkenylglycerophosphocholine hydrolase|alkenylglycerophosphoethanolamine hydrolase', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55226634, end_position_on_the_genomic_accession=55229264, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374788, CurrentID=0, Status='live', Symbol='LOC105374788', Aliases='', description='uncharacterized LOC105374788', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=67522673, end_position_on_the_genomic_accession=67543548, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374789, CurrentID=0, Status='live', Symbol='LOC105374789', Aliases='', description='uncharacterized LOC105374789', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68088914, end_position_on_the_genomic_accession=68111556, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  648262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648262, CurrentID=0, Status='live', Symbol='CICP26', Aliases='', description='capicua transcriptional repressor pseudogene 26', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227976373, end_position_on_the_genomic_accession=227978204, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374791, CurrentID=0, Status='live', Symbol='LOC105374791', Aliases='', description='uncharacterized LOC105374791', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68831188, end_position_on_the_genomic_accession=68837207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374792, CurrentID=0, Status='live', Symbol='LOC105374792', Aliases='', description='uncharacterized LOC105374792', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68875541, end_position_on_the_genomic_accession=68882029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131913, CurrentID=0, Status='live', Symbol='RPL5P20', Aliases='RPL5_9_664', description='ribosomal protein L5 pseudogene 20', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19348110, end_position_on_the_genomic_accession=19349119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374794, CurrentID=0, Status='live', Symbol='LOC105374794', Aliases='', description='uncharacterized LOC105374794', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70673291, end_position_on_the_genomic_accession=70680753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374795, CurrentID=0, Status='live', Symbol='LOC105374795', Aliases='', description='uncharacterized LOC105374795', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70973540, end_position_on_the_genomic_accession=70984116, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374796, CurrentID=0, Status='live', Symbol='LOC105374796', Aliases='', description='uncharacterized LOC105374796', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71000446, end_position_on_the_genomic_accession=71002934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374797, CurrentID=0, Status='live', Symbol='LOC105374797', Aliases='', description='uncharacterized LOC105374797', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=71258262, end_position_on_the_genomic_accession=71276445, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374798, CurrentID=0, Status='live', Symbol='LOC105374798', Aliases='', description='uncharacterized LOC105374798', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72010400, end_position_on_the_genomic_accession=72033799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374799, CurrentID=0, Status='live', Symbol='LOC105374799', Aliases='', description='uncharacterized LOC105374799', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=72172114, end_position_on_the_genomic_accession=72176272, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374800, CurrentID=0, Status='live', Symbol='LOC105374800', Aliases='', description='uncharacterized LOC105374800', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73053329, end_position_on_the_genomic_accession=73055511, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  255057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255057, CurrentID=0, Status='live', Symbol='CBARP', Aliases='BARP, C19orf26, DOS', description='calcium channel, voltage-dependent, beta subunit associated regulatory protein', other_designations='downstream of Stk11', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1229948, end_position_on_the_genomic_accession=1241763, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105374802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374802, CurrentID=0, Status='live', Symbol='LOC105374802', Aliases='', description='uncharacterized LOC105374802', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73305946, end_position_on_the_genomic_accession=73307875, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374803, CurrentID=0, Status='live', Symbol='LOC105374803', Aliases='', description='uncharacterized LOC105374803', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73307970, end_position_on_the_genomic_accession=73316078, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100131924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131924, CurrentID=0, Status='live', Symbol='BLOC1S2P1', Aliases='', description='biogenesis of lysosomal organelles complex-1, subunit 2 pseudogene 1', other_designations='', map_location='Xq11.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=64726555, end_position_on_the_genomic_accession=64727155, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  255061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255061, CurrentID=0, Status='live', Symbol='TAC4', Aliases='EK, HK-1, HK1, PPT-C', description='tachykinin 4 (hemokinin)', other_designations='endokinin|preprotachykinin-C', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49838309, end_position_on_the_genomic_accession=49848017, orientation='minus', exon_count=5, OMIM=607833, no_hdr0=''),
  105374806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374806, CurrentID=0, Status='live', Symbol='LOC105374806', Aliases='', description='uncharacterized LOC105374806', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73844799, end_position_on_the_genomic_accession=73858028, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374807, CurrentID=0, Status='live', Symbol='LOC105374807', Aliases='', description='uncharacterized LOC105374807', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73820775, end_position_on_the_genomic_accession=73830838, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374808, CurrentID=0, Status='live', Symbol='LOC105374808', Aliases='', description='uncharacterized LOC105374808', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74701696, end_position_on_the_genomic_accession=74708947, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374809, CurrentID=0, Status='live', Symbol='LOC105374809', Aliases='', description='uncharacterized LOC105374809', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74928262, end_position_on_the_genomic_accession=74932595, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374810, CurrentID=0, Status='live', Symbol='LOC105374810', Aliases='', description='uncharacterized LOC105374810', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75345230, end_position_on_the_genomic_accession=75348373, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  648283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648283, CurrentID=0, Status='live', Symbol='TERF1P4', Aliases='', description='telomeric repeat binding factor (NIMA-interacting) 1 pseudogene 4', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=83748825, end_position_on_the_genomic_accession=83750101, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131932, CurrentID=0, Status='live', Symbol='RCC2P3', Aliases='', description='regulator of chromosome condensation 2 pseudogene 3', other_designations='', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=138122203, end_position_on_the_genomic_accession=138124257, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374813, CurrentID=0, Status='live', Symbol='LOC105374813', Aliases='', description='uncharacterized LOC105374813', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=75710960, end_position_on_the_genomic_accession=75752096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131934, CurrentID=0, Status='live', Symbol='LETM1P2', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 1, pseudogene 2', other_designations='leucine zipper-EF-hand containing transmembrane protein 1 pseudogene', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58044440, end_position_on_the_genomic_accession=58048251, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131935, CurrentID=0, Status='live', Symbol='GAPDHP41', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 41', other_designations='', map_location='6p11.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=60719138, end_position_on_the_genomic_accession=60720271, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374816, CurrentID=0, Status='live', Symbol='LOC105374816', Aliases='', description='uncharacterized LOC105374816', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77987876, end_position_on_the_genomic_accession=77999030, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374817, CurrentID=0, Status='live', Symbol='LOC105374817', Aliases='', description='uncharacterized LOC105374817', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=77924542, end_position_on_the_genomic_accession=77955034, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100131938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131938, CurrentID=0, Status='live', Symbol='LOC100131938', Aliases='', description='uncharacterized LOC100131938', other_designations='', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.9', start_position_on_the_genomic_accession=162876894, end_position_on_the_genomic_accession=162877320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374819, CurrentID=0, Status='live', Symbol='LOC105374819', Aliases='', description='uncharacterized LOC105374819', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78558150, end_position_on_the_genomic_accession=78570361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100131940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131940, CurrentID=0, Status='live', Symbol='SNX5P1', Aliases='', description='sorting nexin 5 pseudogene 1', other_designations='', map_location='4q21.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76344548, end_position_on_the_genomic_accession=76345734, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374821, CurrentID=0, Status='live', Symbol='LOC105374821', Aliases='', description='uncharacterized LOC105374821', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=78786134, end_position_on_the_genomic_accession=78791332, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  648294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648294, CurrentID=0, Status='live', Symbol='RPL23AP57', Aliases='RPL23A_24_958', description='ribosomal protein L23a pseudogene 57', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6639141, end_position_on_the_genomic_accession=6639611, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100131943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131943, CurrentID=0, Status='live', Symbol='LOC100131943', Aliases='', description='uncharacterized LOC100131943', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.10', start_position_on_the_genomic_accession=19958912, end_position_on_the_genomic_accession=19959681, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105374824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374824, CurrentID=0, Status='live', Symbol='LOC105374824', Aliases='', description='uncharacterized LOC105374824', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=80343396, end_position_on_the_genomic_accession=80364019, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100423012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423012, CurrentID=0, Status='live', Symbol='MIR3177', Aliases='mir-3177', description='microRNA 3177', other_designations='hsa-mir-3177', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1734985, end_position_on_the_genomic_accession=1735066, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  255082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255082, CurrentID=0, Status='live', Symbol='CASC2', Aliases='C10orf5', description='cancer susceptibility candidate 2 (non-protein coding)', other_designations='', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118046821, end_position_on_the_genomic_accession=118210153, orientation='plus', exon_count=6, OMIM=608598, no_hdr0=''),
  100131947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131947, CurrentID=0, Status='live', Symbol='PRPF38AP2', Aliases='', description='PRP38 domain containing A pseudogene 2', other_designations='PRP38 pre-mRNA processing factor 38 domain containing A pseudogene', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17577285, end_position_on_the_genomic_accession=17578215, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374828, CurrentID=0, Status='live', Symbol='LOC105374828', Aliases='', description='uncharacterized LOC105374828', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81708832, end_position_on_the_genomic_accession=81709753, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374829, CurrentID=0, Status='live', Symbol='LOC105374829', Aliases='', description='uncharacterized LOC105374829', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=81868769, end_position_on_the_genomic_accession=81895284, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374830, CurrentID=0, Status='live', Symbol='LOC105374830', Aliases='', description='uncharacterized LOC105374830', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82408595, end_position_on_the_genomic_accession=82424499, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374831, CurrentID=0, Status='live', Symbol='LOC105374831', Aliases='', description='uncharacterized LOC105374831', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82825993, end_position_on_the_genomic_accession=82832960, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374832, CurrentID=0, Status='live', Symbol='LOC105374832', Aliases='', description='uncharacterized LOC105374832', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=82882460, end_position_on_the_genomic_accession=82937536, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131953, CurrentID=0, Status='live', Symbol='ACTG1P22', Aliases='', description='actin gamma 1 pseudogene 22', other_designations='actin, gamma 1 pseudogene', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57764954, end_position_on_the_genomic_accession=57766053, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374834, CurrentID=0, Status='live', Symbol='LOC105374834', Aliases='', description='uncharacterized LOC105374834', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=83518674, end_position_on_the_genomic_accession=83541904, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374835, CurrentID=0, Status='live', Symbol='LOC105374835', Aliases='', description='uncharacterized LOC105374835', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131956, CurrentID=0, Status='live', Symbol='RPS3P5', Aliases='RPS3_2_726', description='ribosomal protein S3 pseudogene 5', other_designations='', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99575682, end_position_on_the_genomic_accession=99576532, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374837, CurrentID=0, Status='live', Symbol='LOC105374837', Aliases='', description='uncharacterized LOC105374837', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84888975, end_position_on_the_genomic_accession=84902938, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131958, CurrentID=0, Status='live', Symbol='RPL15P12', Aliases='RPL15_6_949', description='ribosomal protein L15 pseudogene 12', other_designations='', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=129864478, end_position_on_the_genomic_accession=129865169, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374839, CurrentID=0, Status='live', Symbol='LOC105374839', Aliases='', description='uncharacterized LOC105374839', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85242035, end_position_on_the_genomic_accession=85253179, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374840, CurrentID=0, Status='live', Symbol='LOC105374840', Aliases='', description='uncharacterized LOC105374840', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85389024, end_position_on_the_genomic_accession=85390209, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131961, CurrentID=0, Status='live', Symbol='ST13P19', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 19', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210265637, end_position_on_the_genomic_accession=210267239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374842, CurrentID=0, Status='live', Symbol='LOC105374842', Aliases='', description='uncharacterized LOC105374842', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85681347, end_position_on_the_genomic_accession=85683560, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  648315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648315, CurrentID=0, Status='live', Symbol='VN2R9P', Aliases='', description='vomeronasal 2 receptor 9 pseudogene', other_designations='calcium-sensing receptor pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58040768, end_position_on_the_genomic_accession=58066959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100131964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131964, CurrentID=0, Status='live', Symbol='RPL7L1P3', Aliases='', description='ribosomal protein L7 like 1 pseudogene 3', other_designations='ribosomal protein L7-like 1 pseudogene 3', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75922750, end_position_on_the_genomic_accession=75923455, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255101, CurrentID=0, Status='live', Symbol='CFAP65', Aliases='CCDC108', description='cilia and flagella associated protein 65', other_designations='coiled-coil domain containing 108', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219002846, end_position_on_the_genomic_accession=219041551, orientation='minus', exon_count=36, OMIM=614270, no_hdr0=''),
  105374846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374846, CurrentID=0, Status='live', Symbol='LOC105374846', Aliases='', description='uncharacterized LOC105374846', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86795302, end_position_on_the_genomic_accession=86822306, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374847, CurrentID=0, Status='live', Symbol='LOC105374847', Aliases='', description='uncharacterized LOC105374847', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87249072, end_position_on_the_genomic_accession=87253878, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  255104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255104, CurrentID=0, Status='live', Symbol='TMCO4', Aliases='', description='transmembrane and coiled-coil domains 4', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19682213, end_position_on_the_genomic_accession=19800353, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105374849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374849, CurrentID=0, Status='live', Symbol='LOC105374849', Aliases='', description='uncharacterized LOC105374849', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374850, CurrentID=0, Status='live', Symbol='LOC105374850', Aliases='', description='uncharacterized LOC105374850', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100131971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131971, CurrentID=0, Status='live', Symbol='RPS26P54', Aliases='RPS26_21_1607', description='ribosomal protein S26 pseudogene 54', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59761532, end_position_on_the_genomic_accession=59761985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374852, CurrentID=0, Status='live', Symbol='LOC105374852', Aliases='', description='uncharacterized LOC105374852', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88016791, end_position_on_the_genomic_accession=88021354, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374853, CurrentID=0, Status='live', Symbol='LOC105374853', Aliases='', description='uncharacterized LOC105374853', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88714478, end_position_on_the_genomic_accession=88723820, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374854, CurrentID=0, Status='live', Symbol='LOC105374854', Aliases='', description='uncharacterized LOC105374854', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88813829, end_position_on_the_genomic_accession=88829980, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101929153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929153, CurrentID=0, Status='live', Symbol='LOC101929153', Aliases='', description='uncharacterized LOC101929153', other_designations='CTD-2161F6.3', map_location='5p15.32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=4773481, end_position_on_the_genomic_accession=4774865, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374856, CurrentID=0, Status='live', Symbol='LOC105374856', Aliases='', description='uncharacterized LOC105374856', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88853976, end_position_on_the_genomic_accession=88855610, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374857, CurrentID=0, Status='live', Symbol='LOC105374857', Aliases='', description='uncharacterized LOC105374857', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89100403, end_position_on_the_genomic_accession=89101843, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374858, CurrentID=0, Status='live', Symbol='LOC105374858', Aliases='', description='uncharacterized LOC105374858', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89034562, end_position_on_the_genomic_accession=89039896, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374859, CurrentID=0, Status='live', Symbol='LOC105374859', Aliases='', description='uncharacterized LOC105374859', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=88966814, end_position_on_the_genomic_accession=88989684, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131980, CurrentID=0, Status='live', Symbol='ZNF705G', Aliases='', description='zinc finger protein 705G', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7357976, end_position_on_the_genomic_accession=7362968, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  124045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124045, CurrentID=0, Status='live', Symbol='SPATA33', Aliases='C16orf55', description='spermatogenesis associated 33', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89657744, end_position_on_the_genomic_accession=89670458, orientation='plus', exon_count=4, OMIM=615409, no_hdr0=''),
  105374862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374862, CurrentID=0, Status='live', Symbol='LOC105374862', Aliases='', description='uncharacterized LOC105374862', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89843625, end_position_on_the_genomic_accession=89850844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255119, CurrentID=0, Status='live', Symbol='C4orf22', Aliases='', description='chromosome 4 open reading frame 22', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80335326, end_position_on_the_genomic_accession=80963756, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100131984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131984, CurrentID=0, Status='live', Symbol='ANKRD57P1', Aliases='', description='ankyrin repeat domain 57 pseudogene 1', other_designations='', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11167992, end_position_on_the_genomic_accession=11169160, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374865, CurrentID=0, Status='live', Symbol='LOC105374865', Aliases='', description='Ig kappa chain V-I region HK102-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374866, CurrentID=0, Status='live', Symbol='LOC105374866', Aliases='', description='uncharacterized LOC105374866', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=90245395, end_position_on_the_genomic_accession=90250362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374867, CurrentID=0, Status='live', Symbol='LOC105374867', Aliases='', description='uncharacterized LOC105374867', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91561595, end_position_on_the_genomic_accession=91571328, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374868, CurrentID=0, Status='live', Symbol='LOC105374868', Aliases='', description='uncharacterized LOC105374868', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91769214, end_position_on_the_genomic_accession=91774850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374869, CurrentID=0, Status='live', Symbol='LOC105374869', Aliases='', description='uncharacterized LOC105374869', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=182309, end_position_on_the_genomic_accession=187745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374870, CurrentID=0, Status='live', Symbol='LOC105374870', Aliases='', description='uncharacterized LOC105374870', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=230930, end_position_on_the_genomic_accession=237089, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131991, CurrentID=0, Status='live', Symbol='RPL29P8', Aliases='RPL29_1_300', description='ribosomal protein L29 pseudogene 8', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176807034, end_position_on_the_genomic_accession=176809748, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  124056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124056, CurrentID=0, Status='live', Symbol='NOXO1', Aliases='P41NOX, P41NOXA, P41NOXB, P41NOXC, SH3PXD5, SNX28', description='NADPH oxidase organizer 1', other_designations='NADPH oxidase regulatory protein|Nox organizer 1|SH3 and PX domain-containing protein 5|nox-organizing protein 1|regulatory protein P41NOX', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1978917, end_position_on_the_genomic_accession=1981549, orientation='minus', exon_count=8, OMIM=611256, no_hdr0=''),
  105374873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374873, CurrentID=0, Status='live', Symbol='LOC105374873', Aliases='', description='uncharacterized LOC105374873', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=729074, end_position_on_the_genomic_accession=758947, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  255130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255130, CurrentID=0, Status='live', Symbol='IGFBP7-AS1', Aliases='', description='IGFBP7 antisense RNA 1', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57109762, end_position_on_the_genomic_accession=57205299, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100131995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131995, CurrentID=0, Status='live', Symbol='UQCRHP3', Aliases='', description='ubiquinol-cytochrome c reductase hinge protein pseudogene 3', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43732259, end_position_on_the_genomic_accession=43732700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374876, CurrentID=0, Status='live', Symbol='LOC105374876', Aliases='', description='uncharacterized LOC105374876', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=775545, end_position_on_the_genomic_accession=780628, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100131997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131997, CurrentID=0, Status='live', Symbol='FAM27E3', Aliases='', description='family with sequence similarity 27 member E3', other_designations='family with sequence similarity 27, member E3', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67717498, end_position_on_the_genomic_accession=67719179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100131998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131998, CurrentID=0, Status='live', Symbol='RRN3P3', Aliases='', description='RRN3 homolog, RNA polymerase I transcription factor pseudogene 3', other_designations='RNA polymerase I transcription factor homolog (S. cerevisiae) pseudogene 3|RRN3 RNA polymerase I transcription factor homolog pseudogene', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22419546, end_position_on_the_genomic_accession=22437715, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374879, CurrentID=0, Status='live', Symbol='LOC105374879', Aliases='', description='uncharacterized LOC105374879', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1284931, end_position_on_the_genomic_accession=1305966, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374880, CurrentID=0, Status='live', Symbol='LOC105374880', Aliases='', description='uncharacterized LOC105374880', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1383792, end_position_on_the_genomic_accession=1389470, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100271536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271536, CurrentID=0, Status='live', Symbol='RPL7AP56', Aliases='RPL7A_21_1137', description='ribosomal protein L7a pseudogene 56', other_designations='', map_location='11p12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=37725645, end_position_on_the_genomic_accession=37726490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374882, CurrentID=0, Status='live', Symbol='LOC105374882', Aliases='', description='uncharacterized LOC105374882', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1529123, end_position_on_the_genomic_accession=1544090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374883, CurrentID=0, Status='live', Symbol='LOC105374883', Aliases='', description='uncharacterized LOC105374883', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1489489, end_position_on_the_genomic_accession=1492524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132004, CurrentID=0, Status='live', Symbol='LOC100132004', Aliases='', description='putative ankyrin repeat domain-containing protein 30B-like', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40754862, end_position_on_the_genomic_accession=40768065, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100132005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132005, CurrentID=0, Status='live', Symbol='LOC100132005', Aliases='', description='uncharacterized LOC100132005', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.10', start_position_on_the_genomic_accession=38255058, end_position_on_the_genomic_accession=38255618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132006, CurrentID=0, Status='live', Symbol='LOC100132006', Aliases='', description='uncharacterized LOC100132006', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=8647639, end_position_on_the_genomic_accession=8651087, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  648359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648359, CurrentID=0, Status='live', Symbol='ABHD17AP3', Aliases='FAM108A4, FAM108A4P', description='abhydrolase domain containing 17A pseudogene 3', other_designations='family with sequence similarity 108, member A4, pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214605457, end_position_on_the_genomic_accession=214609203, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374888, CurrentID=0, Status='live', Symbol='LOC105374888', Aliases='', description='uncharacterized LOC105374888', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=2989723, end_position_on_the_genomic_accession=2999292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374889, CurrentID=0, Status='live', Symbol='LOC105374889', Aliases='', description='uncharacterized LOC105374889', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3157676, end_position_on_the_genomic_accession=3158702, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374890, CurrentID=0, Status='live', Symbol='LOC105374890', Aliases='', description='uncharacterized LOC105374890', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3410719, end_position_on_the_genomic_accession=3417979, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374891, CurrentID=0, Status='live', Symbol='LOC105374891', Aliases='', description='uncharacterized LOC105374891', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3309250, end_position_on_the_genomic_accession=3313646, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374892, CurrentID=0, Status='live', Symbol='LOC105374892', Aliases='', description='uncharacterized LOC105374892', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3662821, end_position_on_the_genomic_accession=3665982, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374893, CurrentID=0, Status='live', Symbol='LOC105374893', Aliases='', description='uncharacterized LOC105374893', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3893036, end_position_on_the_genomic_accession=3895552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132014, CurrentID=0, Status='live', Symbol='LOC100132014', Aliases='', description='uncharacterized LOC100132014', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.8', start_position_on_the_genomic_accession=115325748, end_position_on_the_genomic_accession=115326215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132015, CurrentID=0, Status='live', Symbol='TEX13D', Aliases='', description='TEX13 family member D', other_designations='testis expressed 13A pseudogene|testis expressed sequence 13-like pseudogene', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=124332918, end_position_on_the_genomic_accession=124333910, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374896, CurrentID=0, Status='live', Symbol='LOC105374896', Aliases='', description='uncharacterized LOC105374896', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4599288, end_position_on_the_genomic_accession=4602420, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374897, CurrentID=0, Status='live', Symbol='LOC105374897', Aliases='', description='uncharacterized LOC105374897', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=4902373, end_position_on_the_genomic_accession=4909245, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374898, CurrentID=0, Status='live', Symbol='LOC105374898', Aliases='', description='uncharacterized LOC105374898', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5993272, end_position_on_the_genomic_accession=5998801, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374899, CurrentID=0, Status='live', Symbol='LOC105374899', Aliases='', description='uncharacterized LOC105374899', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6325201, end_position_on_the_genomic_accession=6326274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374900, CurrentID=0, Status='live', Symbol='LOC105374900', Aliases='', description='uncharacterized LOC105374900', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6659371, end_position_on_the_genomic_accession=6681179, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374901, CurrentID=0, Status='live', Symbol='LOC105374901', Aliases='', description='uncharacterized LOC105374901', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6752066, end_position_on_the_genomic_accession=6759366, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374902, CurrentID=0, Status='live', Symbol='LOC105374902', Aliases='', description='uncharacterized LOC105374902', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6885676, end_position_on_the_genomic_accession=6898568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374903, CurrentID=0, Status='live', Symbol='LOC105374903', Aliases='', description='uncharacterized LOC105374903', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6872275, end_position_on_the_genomic_accession=6890466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374904, CurrentID=0, Status='live', Symbol='LOC105374904', Aliases='', description='uncharacterized LOC105374904', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=6994096, end_position_on_the_genomic_accession=6995825, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132025, CurrentID=0, Status='live', Symbol='LOC100132025', Aliases='', description='transmembrane domain-containing protein ENSP00000320207-like', other_designations='Putative transmembrane domain-containing protein ENSP00000320207', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.8', start_position_on_the_genomic_accession=20956228, end_position_on_the_genomic_accession=20957057, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105374906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374906, CurrentID=0, Status='live', Symbol='LOC105374906', Aliases='', description='uncharacterized LOC105374906', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7672678, end_position_on_the_genomic_accession=7699197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374907, CurrentID=0, Status='live', Symbol='LOC105374907', Aliases='', description='uncharacterized LOC105374907', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7727627, end_position_on_the_genomic_accession=7731870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374908, CurrentID=0, Status='live', Symbol='LOC105374908', Aliases='', description='uncharacterized LOC105374908', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7912481, end_position_on_the_genomic_accession=7916025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124093, CurrentID=0, Status='live', Symbol='CCDC78', Aliases='C16orf25, CNM4, JFP10, hsCCDC78', description='coiled-coil domain containing 78', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=722582, end_position_on_the_genomic_accession=726881, orientation='minus', exon_count=12, OMIM=614666, no_hdr0=''),
  104457406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104457406, CurrentID=0, Status='live', Symbol='MIR222HG', Aliases='Lnc-Ang362', description='MIR222 host gene', other_designations='MIR222 host gene (non-protein coding)', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  255167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255167, CurrentID=0, Status='live', Symbol='LINC01018', Aliases='SRHC', description='long intergenic non-protein coding RNA 1018', other_designations='CTD-2195M18.1', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=6582136, end_position_on_the_genomic_accession=6588500, orientation='plus', exon_count=5, OMIM=616385, no_hdr0=''),
  105374912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374912, CurrentID=0, Status='live', Symbol='LOC105374912', Aliases='', description='uncharacterized LOC105374912', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8307531, end_position_on_the_genomic_accession=8326285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374913, CurrentID=0, Status='live', Symbol='LOC105374913', Aliases='', description='uncharacterized LOC105374913', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8925897, end_position_on_the_genomic_accession=8939632, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374914, CurrentID=0, Status='live', Symbol='LOC105374914', Aliases='', description='uncharacterized LOC105374914', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8965319, end_position_on_the_genomic_accession=9051657, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105374915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374915, CurrentID=0, Status='live', Symbol='LOC105374915', Aliases='', description='uncharacterized LOC105374915', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=9188908, end_position_on_the_genomic_accession=9227648, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132036, CurrentID=0, Status='live', Symbol='NDUFB4P2', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B4 pseudogene 2', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 4, 15kDa pseudogene 2', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=59528792, end_position_on_the_genomic_accession=59529311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132037, CurrentID=0, Status='live', Symbol='RPS29P10', Aliases='RPS29_1_224', description='ribosomal protein S29 pseudogene 10', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61589493, end_position_on_the_genomic_accession=61589663, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374918, CurrentID=0, Status='live', Symbol='LOC105374918', Aliases='', description='uncharacterized LOC105374918', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10140701, end_position_on_the_genomic_accession=10144465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374919, CurrentID=0, Status='live', Symbol='LOC105374919', Aliases='', description='uncharacterized LOC105374919', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10041205, end_position_on_the_genomic_accession=10062391, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390139, CurrentID=0, Status='live', Symbol='OR4A17P', Aliases='OR4A22P', description='olfactory receptor family 4 subfamily A member 17 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 17 pseudogene|olfactory receptor, family 4, subfamily A, member 22 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55444436, end_position_on_the_genomic_accession=55445339, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255177, CurrentID=0, Status='live', Symbol='LOC255177', Aliases='', description='uncharacterized LOC255177', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62196161, end_position_on_the_genomic_accession=62203318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374922, CurrentID=0, Status='live', Symbol='LOC105374922', Aliases='', description='uncharacterized LOC105374922', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10570210, end_position_on_the_genomic_accession=10581238, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374923, CurrentID=0, Status='live', Symbol='LOC105374923', Aliases='', description='uncharacterized LOC105374923', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=10684212, end_position_on_the_genomic_accession=10686111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255180, CurrentID=0, Status='live', Symbol='GOLGA2P11', Aliases='', description='golgin A2 pseudogene 11', other_designations='golgin subfamily A member 2-like', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62230984, end_position_on_the_genomic_accession=62252380, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105374925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374925, CurrentID=0, Status='live', Symbol='LOC105374925', Aliases='', description='uncharacterized LOC105374925', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11291535, end_position_on_the_genomic_accession=11336978, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105374926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374926, CurrentID=0, Status='live', Symbol='LOC105374926', Aliases='', description='uncharacterized LOC105374926', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11219861, end_position_on_the_genomic_accession=11232127, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132047, CurrentID=0, Status='live', Symbol='RPL22P6', Aliases='RPL22_1_94', description='ribosomal protein L22 pseudogene 6', other_designations='', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121142049, end_position_on_the_genomic_accession=121142438, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374928, CurrentID=0, Status='live', Symbol='LOC105374928', Aliases='', description='uncharacterized LOC105374928', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11417401, end_position_on_the_genomic_accession=11535327, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100132049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132049, CurrentID=0, Status='live', Symbol='RPS19P4', Aliases='RPS19_2_563', description='ribosomal protein S19 pseudogene 4', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53413477, end_position_on_the_genomic_accession=53414260, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374930, CurrentID=0, Status='live', Symbol='LOC105374930', Aliases='', description='uncharacterized LOC105374930', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=11727246, end_position_on_the_genomic_accession=11731247, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  255187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255187, CurrentID=0, Status='live', Symbol='LOC255187', Aliases='', description='uncharacterized LOC255187', other_designations='', map_location='5q33.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149063317, end_position_on_the_genomic_accession=149109787, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374932, CurrentID=0, Status='live', Symbol='LOC105374932', Aliases='', description='uncharacterized LOC105374932', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13273187, end_position_on_the_genomic_accession=13276683, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255189, CurrentID=0, Status='live', Symbol='PLA2G4F', Aliases='PLA2G4FZ', description='phospholipase A2 group IVF', other_designations='cPLA2-zeta|phospholipase A2, group IVF', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=42141134, end_position_on_the_genomic_accession=42156641, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  105374934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374934, CurrentID=0, Status='live', Symbol='LOC105374934', Aliases='', description='uncharacterized LOC105374934', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12828524, end_position_on_the_genomic_accession=12834796, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374935, CurrentID=0, Status='live', Symbol='LOC105374935', Aliases='', description='uncharacterized LOC105374935', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12695221, end_position_on_the_genomic_accession=12747458, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374936, CurrentID=0, Status='live', Symbol='LOC105374936', Aliases='', description='uncharacterized LOC105374936', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=13351083, end_position_on_the_genomic_accession=13354188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255193, CurrentID=0, Status='live', Symbol='CSNK1G2-AS1', Aliases='C19orf34', description='CSNK1G2 antisense RNA 1', other_designations='CSNK1G2 antisense RNA 1 (non-protein coding)', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1952527, end_position_on_the_genomic_accession=1954549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132058, CurrentID=0, Status='live', Symbol='GPS2P2', Aliases='', description='G protein pathway suppressor 2 pseudogene 2', other_designations='', map_location='3q27.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=186772612, end_position_on_the_genomic_accession=186773579, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105374939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374939, CurrentID=0, Status='live', Symbol='LOC105374939', Aliases='', description='uncharacterized LOC105374939', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14098683, end_position_on_the_genomic_accession=14107284, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480970, CurrentID=0, Status='live', Symbol='RN7SL180P', Aliases='', description='RNA, 7SL, cytoplasmic 180, pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132061, CurrentID=0, Status='live', Symbol='RPL9P14', Aliases='RPL9_3_318', description='ribosomal protein L9 pseudogene 14', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208031291, end_position_on_the_genomic_accession=208031992, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132062, CurrentID=0, Status='live', Symbol='LOC100132062', Aliases='', description='uncharacterized LOC100132062', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181323506, end_position_on_the_genomic_accession=181328195, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374943, CurrentID=0, Status='live', Symbol='LOC105374943', Aliases='', description='uncharacterized LOC105374943', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14389891, end_position_on_the_genomic_accession=14394068, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374944, CurrentID=0, Status='live', Symbol='LOC105374944', Aliases='', description='uncharacterized LOC105374944', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14860416, end_position_on_the_genomic_accession=14865591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374945, CurrentID=0, Status='live', Symbol='LOC105374945', Aliases='', description='uncharacterized LOC105374945', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=14876139, end_position_on_the_genomic_accession=15009028, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132066, CurrentID=0, Status='live', Symbol='ALG1L3P', Aliases='', description='asparagine-linked glycosylation 1-like 3, pseudogene', other_designations='asparagine-linked glycosylation 1 homolog pseudogene|asparagine-linked glycosylation 1-like pseudogene|beta-1,4-mannosyltransferase pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9702844, end_position_on_the_genomic_accession=9710812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374947, CurrentID=0, Status='live', Symbol='LOC105374947', Aliases='', description='uncharacterized LOC105374947', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15533404, end_position_on_the_genomic_accession=15560573, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374948, CurrentID=0, Status='live', Symbol='LOC105374948', Aliases='', description='uncharacterized LOC105374948', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15829647, end_position_on_the_genomic_accession=15834463, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374949, CurrentID=0, Status='live', Symbol='LOC105374949', Aliases='', description='uncharacterized LOC105374949', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=15956416, end_position_on_the_genomic_accession=15980976, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132070, CurrentID=0, Status='live', Symbol='FAM231EP', Aliases='', description='family with sequence similarity 231 member E, pseudogene', other_designations='family with sequence similarity 231, member E, pseudogene', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16868363, end_position_on_the_genomic_accession=16868868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374951, CurrentID=0, Status='live', Symbol='LOC105374951', Aliases='', description='uncharacterized LOC105374951', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=17015795, end_position_on_the_genomic_accession=17034225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255208, CurrentID=0, Status='live', Symbol='TUBB8P9', Aliases='', description='tubulin beta 8 class VIII pseudogene 9', other_designations='HSA1q42.3 beta-tubulin 4Q pseudogene|tubulin, beta 8 class VIII pseudogene 9|tubulin, beta 8 pseudogene', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227506182, end_position_on_the_genomic_accession=227508469, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132073, CurrentID=0, Status='live', Symbol='CCNB2P1', Aliases='', description='cyclin B2 pseudogene 1', other_designations='CCNB2 pseudogene 1', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7909113, end_position_on_the_genomic_accession=7910455, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132074, CurrentID=0, Status='live', Symbol='FOXO6', Aliases='', description='forkhead box O6', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41361931, end_position_on_the_genomic_accession=41383591, orientation='plus', exon_count=2, OMIM=611457, no_hdr0=''),
  105374955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374955, CurrentID=0, Status='live', Symbol='LOC105374955', Aliases='', description='uncharacterized LOC105374955', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18496636, end_position_on_the_genomic_accession=18517420, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374956, CurrentID=0, Status='live', Symbol='LOC105374956', Aliases='', description='uncharacterized LOC105374956', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18522797, end_position_on_the_genomic_accession=18723964, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100132077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132077, CurrentID=0, Status='live', Symbol='LOC100132077', Aliases='', description='uncharacterized LOC100132077', other_designations='', map_location='9q22.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94332476, end_position_on_the_genomic_accession=94360948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132078, CurrentID=0, Status='live', Symbol='LOC100132078', Aliases='', description='uncharacterized LOC100132078', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111414242, end_position_on_the_genomic_accession=111418186, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132079, CurrentID=0, Status='live', Symbol='KARSP3', Aliases='', description='lysyl-tRNA synthetase pseudogene 3', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=68002423, end_position_on_the_genomic_accession=68004451, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374960, CurrentID=0, Status='live', Symbol='LOC105374960', Aliases='', description='uncharacterized LOC105374960', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19449711, end_position_on_the_genomic_accession=19459348, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374961, CurrentID=0, Status='live', Symbol='LOC105374961', Aliases='', description='uncharacterized LOC105374961', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19317177, end_position_on_the_genomic_accession=19322245, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374962, CurrentID=0, Status='live', Symbol='LOC105374962', Aliases='', description='uncharacterized LOC105374962', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19729427, end_position_on_the_genomic_accession=19750457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105374963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374963, CurrentID=0, Status='live', Symbol='LOC105374963', Aliases='', description='uncharacterized LOC105374963', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=19807730, end_position_on_the_genomic_accession=19814175, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  255220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255220, CurrentID=0, Status='live', Symbol='TXNDC8', Aliases='SPTRX-3, TRX6, bA427L11.2', description='thioredoxin domain containing 8 (spermatozoa)', other_designations='sperm-specific thioredoxin 3|spermatid-specific thioredoxin-3|spermatocyte/spermatid-specific thioredoxin-3|thioredoxin 6', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110276799, end_position_on_the_genomic_accession=110337887, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  103539957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103539957, CurrentID=0, Status='live', Symbol='ERVK-26', Aliases='', description='endogenous retrovirus group K, member 26', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103539958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103539958, CurrentID=0, Status='live', Symbol='TRG-CCC8-1', Aliases='', description='transfer RNA-Gly (CCC) 8-1', other_designations='tRNA-Gly-CCC-8-1', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374967, CurrentID=0, Status='live', Symbol='LOC105374967', Aliases='', description='uncharacterized LOC105374967', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20974075, end_position_on_the_genomic_accession=20999650, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124152, CurrentID=0, Status='live', Symbol='IQCK', Aliases='', description='IQ motif containing K', other_designations='', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=19716456, end_position_on_the_genomic_accession=19857585, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105375406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375406, CurrentID=0, Status='live', Symbol='LOC105375406', Aliases='', description='uncharacterized LOC105375406', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=94926167, end_position_on_the_genomic_accession=94928540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480971, CurrentID=0, Status='live', Symbol='RN7SL183P', Aliases='', description='RNA, 7SL, cytoplasmic 183, pseudogene', other_designations='', map_location='6q15', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105374971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374971, CurrentID=0, Status='live', Symbol='LOC105374971', Aliases='', description='uncharacterized LOC105374971', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22349226, end_position_on_the_genomic_accession=22590129, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105374972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374972, CurrentID=0, Status='live', Symbol='LOC105374972', Aliases='', description='uncharacterized LOC105374972', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22654160, end_position_on_the_genomic_accession=22717922, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374973, CurrentID=0, Status='live', Symbol='LOC105374973', Aliases='', description='uncharacterized LOC105374973', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=22644191, end_position_on_the_genomic_accession=22662205, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105374974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374974, CurrentID=0, Status='live', Symbol='LOC105374974', Aliases='', description='uncharacterized LOC105374974', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23016751, end_position_on_the_genomic_accession=23177072, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  255231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255231, CurrentID=0, Status='live', Symbol='MCOLN2', Aliases='TRP-ML2, TRPML2', description='mucolipin 2', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84925583, end_position_on_the_genomic_accession=84997122, orientation='minus', exon_count=18, OMIM=607399, no_hdr0=''),
  105374976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374976, CurrentID=0, Status='live', Symbol='LOC105374976', Aliases='', description='uncharacterized LOC105374976', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23337813, end_position_on_the_genomic_accession=23584381, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105374977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374977, CurrentID=0, Status='live', Symbol='LOC105374977', Aliases='', description='uncharacterized LOC105374977', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=23605949, end_position_on_the_genomic_accession=23626334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374978, CurrentID=0, Status='live', Symbol='LOC105374978', Aliases='', description='uncharacterized LOC105374978', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=24742302, end_position_on_the_genomic_accession=24751646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132099, CurrentID=0, Status='live', Symbol='LOC100132099', Aliases='', description='FRSS1829', other_designations='', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.9', start_position_on_the_genomic_accession=98699124, end_position_on_the_genomic_accession=98699345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132100, CurrentID=0, Status='live', Symbol='DNM1P32', Aliases='DNM1DN4-5, DNM1DN4.5, DNM1DN4@', description='dynamin 1 pseudogene 32', other_designations='DNM1 pseudogene 32|DNM1DN4.5 duplicon', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32434748, end_position_on_the_genomic_accession=32439743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132101, CurrentID=0, Status='live', Symbol='HERC2P7', Aliases='', description='hect domain and RLD 2 pseudogene 7', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22479153, end_position_on_the_genomic_accession=22482374, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105374982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374982, CurrentID=0, Status='live', Symbol='LOC105374982', Aliases='', description='uncharacterized LOC105374982', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25121716, end_position_on_the_genomic_accession=25135311, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255239, CurrentID=0, Status='live', Symbol='ANKK1', Aliases='PKK2', description='ankyrin repeat and kinase domain containing 1', other_designations='X-kinase|protein kinase PKK2|sgK288|sugen kinase 288', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113386038, end_position_on_the_genomic_accession=113400418, orientation='plus', exon_count=10, OMIM=608774, no_hdr0=''),
  100132104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132104, CurrentID=0, Status='live', Symbol='AK4P2', Aliases='', description='adenylate kinase 4 pseudogene 2', other_designations='adenylate kinase 3-like 1 pseudogene', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55551888, end_position_on_the_genomic_accession=55552554, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105374985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374985, CurrentID=0, Status='live', Symbol='LOC105374985', Aliases='', description='uncharacterized LOC105374985', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26280354, end_position_on_the_genomic_accession=26285027, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374986, CurrentID=0, Status='live', Symbol='LOC105374986', Aliases='', description='uncharacterized LOC105374986', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26257024, end_position_on_the_genomic_accession=26262753, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374987, CurrentID=0, Status='live', Symbol='LOC105374987', Aliases='', description='uncharacterized LOC105374987', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26331844, end_position_on_the_genomic_accession=26340273, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105374988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374988, CurrentID=0, Status='live', Symbol='LOC105374988', Aliases='', description='uncharacterized LOC105374988', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26569078, end_position_on_the_genomic_accession=26574698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374989, CurrentID=0, Status='live', Symbol='LOC105374989', Aliases='', description='uncharacterized LOC105374989', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26602050, end_position_on_the_genomic_accession=26606661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374990, CurrentID=0, Status='live', Symbol='LOC105374990', Aliases='', description='uncharacterized LOC105374990', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26706841, end_position_on_the_genomic_accession=26730118, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132111, CurrentID=0, Status='live', Symbol='LOC100132111', Aliases='', description='uncharacterized LOC100132111', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151838469, end_position_on_the_genomic_accession=151844165, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374992, CurrentID=0, Status='live', Symbol='LOC105374992', Aliases='', description='uncharacterized LOC105374992', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26779537, end_position_on_the_genomic_accession=26803593, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105374993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374993, CurrentID=0, Status='live', Symbol='LOC105374993', Aliases='', description='uncharacterized LOC105374993', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26771568, end_position_on_the_genomic_accession=26794817, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105374994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374994, CurrentID=0, Status='live', Symbol='LOC105374994', Aliases='', description='uncharacterized LOC105374994', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26813827, end_position_on_the_genomic_accession=26827989, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132115, CurrentID=0, Status='live', Symbol='RBM22P11', Aliases='RBM22P8', description='RNA binding motif protein 22 pseudogene 11', other_designations='RNA binding motif protein 22 pseudogene 8', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52526903, end_position_on_the_genomic_accession=52527842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132116, CurrentID=0, Status='live', Symbol='ACTA2-AS1', Aliases='ZXF1, uc001kfo.1', description='ACTA2 antisense RNA 1', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88932684, end_position_on_the_genomic_accession=88939974, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105374997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374997, CurrentID=0, Status='live', Symbol='LOC105374997', Aliases='', description='uncharacterized LOC105374997', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28421575, end_position_on_the_genomic_accession=28424819, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105374998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374998, CurrentID=0, Status='live', Symbol='LOC105374998', Aliases='', description='uncharacterized LOC105374998', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28399819, end_position_on_the_genomic_accession=28403595, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105374999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105374999, CurrentID=0, Status='live', Symbol='LOC105374999', Aliases='', description='uncharacterized LOC105374999', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28039343, end_position_on_the_genomic_accession=28062955, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480972, CurrentID=0, Status='live', Symbol='RN7SL186P', Aliases='', description='RNA, 7SL, cytoplasmic 186, pseudogene', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132121, CurrentID=0, Status='live', Symbol='ETF1P3', Aliases='SUP45L4', description='eukaryotic translation termination factor 1 pseudogene 3', other_designations='sup45 (yeast omnipotent suppressor 45) homolog-like 4', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=65794345, end_position_on_the_genomic_accession=65798029, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132122, CurrentID=0, Status='live', Symbol='RPL35AP5', Aliases='RPL35A_2_150', description='ribosomal protein L35a pseudogene 5', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203835540, end_position_on_the_genomic_accession=203835983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375003, CurrentID=0, Status='live', Symbol='LOC105375003', Aliases='', description='uncharacterized LOC105375003', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29108630, end_position_on_the_genomic_accession=29114417, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132124, CurrentID=0, Status='live', Symbol='TSPY17P', Aliases='', description='testis specific protein, Y-linked 17, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6520418, end_position_on_the_genomic_accession=6520929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375005, CurrentID=0, Status='live', Symbol='LOC105375005', Aliases='', description='uncharacterized LOC105375005', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29284072, end_position_on_the_genomic_accession=29334443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375006, CurrentID=0, Status='live', Symbol='LOC105375006', Aliases='', description='uncharacterized LOC105375006', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29205063, end_position_on_the_genomic_accession=29232329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132127, CurrentID=0, Status='live', Symbol='RPL39P14', Aliases='RPL39_6_311', description='ribosomal protein L39 pseudogene 14', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202231648, end_position_on_the_genomic_accession=202232040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375008, CurrentID=0, Status='live', Symbol='LOC105375008', Aliases='', description='uncharacterized LOC105375008', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29473743, end_position_on_the_genomic_accession=29490405, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375009, CurrentID=0, Status='live', Symbol='LOC105375009', Aliases='', description='uncharacterized LOC105375009', other_designations='XXbac-BPG13B8.10', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29497500, end_position_on_the_genomic_accession=29510558, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375010, CurrentID=0, Status='live', Symbol='LOC105375010', Aliases='', description='uncharacterized LOC105375010', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29848017, end_position_on_the_genomic_accession=29857386, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375011, CurrentID=0, Status='live', Symbol='LOC105375011', Aliases='', description='uncharacterized LOC105375011', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29876073, end_position_on_the_genomic_accession=29887324, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132132, CurrentID=0, Status='live', Symbol='DSTNP5', Aliases='', description='destrin (actin depolymerizing factor) pseudogene 5', other_designations='', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=204306842, end_position_on_the_genomic_accession=204307361, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375013, CurrentID=0, Status='live', Symbol='HCG20', Aliases='', description='HLA complex group 20 (non-protein coding)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30744491, end_position_on_the_genomic_accession=30791578, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375014, CurrentID=0, Status='live', Symbol='LOC105375014', Aliases='', description='uncharacterized LOC105375014', other_designations='XXbac-BPG27H4.8', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30839529, end_position_on_the_genomic_accession=30848150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124199, CurrentID=0, Status='live', Symbol='EEF1A1P38', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 38', other_designations='', map_location='16p12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=27133483, end_position_on_the_genomic_accession=27135179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375016, CurrentID=0, Status='live', Symbol='LOC105375016', Aliases='', description='uncharacterized LOC105375016', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31357276, end_position_on_the_genomic_accession=31358155, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375017, CurrentID=0, Status='live', Symbol='LOC105375017', Aliases='', description='uncharacterized LOC105375017', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31439083, end_position_on_the_genomic_accession=31440913, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375018, CurrentID=0, Status='live', Symbol='LOC105375018', Aliases='', description='uncharacterized LOC105375018', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31689711, end_position_on_the_genomic_accession=31691316, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  255275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255275, CurrentID=0, Status='live', Symbol='MYADML2', Aliases='', description='myeloid-associated differentiation marker-like 2', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81939645, end_position_on_the_genomic_accession=81947233, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375020, CurrentID=0, Status='live', Symbol='LOC105375020', Aliases='', description='uncharacterized LOC105375020', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31737432, end_position_on_the_genomic_accession=31739943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375021, CurrentID=0, Status='live', Symbol='LOC105375021', Aliases='', description='uncharacterized LOC105375021', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33131014, end_position_on_the_genomic_accession=33143300, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132142, CurrentID=0, Status='live', Symbol='RPS3AP4', Aliases='', description='ribosomal protein S3A pseudogene 4', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35468247, end_position_on_the_genomic_accession=35469400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375023, CurrentID=0, Status='live', Symbol='LOC105375023', Aliases='', description='uncharacterized LOC105375023', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33633876, end_position_on_the_genomic_accession=33635421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375024, CurrentID=0, Status='live', Symbol='LOC105375024', Aliases='', description='uncharacterized LOC105375024', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33789172, end_position_on_the_genomic_accession=33803004, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375025, CurrentID=0, Status='live', Symbol='LOC105375025', Aliases='', description='uncharacterized LOC105375025', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33820300, end_position_on_the_genomic_accession=33822132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132146, CurrentID=0, Status='live', Symbol='LOC100132146', Aliases='', description='uncharacterized LOC100132146', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46612435, end_position_on_the_genomic_accession=46626543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132147, CurrentID=0, Status='live', Symbol='LOC100132147', Aliases='', description='uncharacterized LOC100132147', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=16860691, end_position_on_the_genomic_accession=16862580, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375028, CurrentID=0, Status='live', Symbol='LOC105375028', Aliases='', description='uncharacterized LOC105375028', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34687099, end_position_on_the_genomic_accession=34688874, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375029, CurrentID=0, Status='live', Symbol='LOC105375029', Aliases='', description='uncharacterized LOC105375029', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=34713143, end_position_on_the_genomic_accession=34720142, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480973, CurrentID=0, Status='live', Symbol='RN7SL189P', Aliases='', description='RNA, 7SL, cytoplasmic 189, pseudogene', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375031, CurrentID=0, Status='live', Symbol='LOC105375031', Aliases='', description='uncharacterized LOC105375031', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35192547, end_position_on_the_genomic_accession=35213648, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375032, CurrentID=0, Status='live', Symbol='LOC105375032', Aliases='', description='uncharacterized LOC105375032', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35633534, end_position_on_the_genomic_accession=35653116, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132153, CurrentID=0, Status='live', Symbol='TUBB2BP1', Aliases='', description='tubulin beta 2B class IIb pseudogene 1', other_designations='tubulin, beta 2B class IIb pseudogene 1|tubulin, beta 2B pseudogene', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3177039, end_position_on_the_genomic_accession=3179715, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132154, CurrentID=0, Status='live', Symbol='LOC100132154', Aliases='', description='putative ankyrin repeat domain-containing protein 30B-like', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64878467, end_position_on_the_genomic_accession=64889049, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375035, CurrentID=0, Status='live', Symbol='LOC105375035', Aliases='', description='uncharacterized LOC105375035', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35968943, end_position_on_the_genomic_accession=35992503, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  124220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124220, CurrentID=0, Status='live', Symbol='ZG16B', Aliases='EECP, HRPE773, JCLN2, PAUF, PRO1567', description='zymogen granule protein 16B', other_designations='endocrine and exocrine protein|jacalin-like lectin domain containing 2|pancreatic adenocarcinoma upregulated factor', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2830172, end_position_on_the_genomic_accession=2832284, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  124221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124221, CurrentID=0, Status='live', Symbol='PRSS30P', Aliases='Disp, TMPRSS8, TMPRSS8P', description='protease, serine, 30 pseudogene', other_designations='protease, serine, 30 homolog, pseudogene|protease, serine, 30, pseudogene|transmembrane protease, serine 8 homolog, pseudogene|transmembrane protease, serine pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2839573, end_position_on_the_genomic_accession=2842751, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  124222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124222, CurrentID=0, Status='live', Symbol='PAQR4', Aliases='', description='progestin and adipoQ receptor family member IV', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2969245, end_position_on_the_genomic_accession=2973489, orientation='plus', exon_count=3, OMIM=614578, no_hdr0=''),
  100132159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132159, CurrentID=0, Status='live', Symbol='CALML3-AS1', Aliases='', description='CALML3 antisense RNA 1', other_designations='', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5514244, end_position_on_the_genomic_accession=5526246, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375040, CurrentID=0, Status='live', Symbol='LOC105375040', Aliases='', description='uncharacterized LOC105375040', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37334128, end_position_on_the_genomic_accession=37354004, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375041, CurrentID=0, Status='live', Symbol='LOC105375041', Aliases='', description='uncharacterized LOC105375041', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37547624, end_position_on_the_genomic_accession=37548601, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375042, CurrentID=0, Status='live', Symbol='LOC105375042', Aliases='', description='uncharacterized LOC105375042', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=37543553, end_position_on_the_genomic_accession=37546762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132163, CurrentID=0, Status='live', Symbol='PHKA2-AS1', Aliases='', description='PHKA2 antisense RNA 1', other_designations='PHKA2 antisense RNA 1 (non-protein coding)', map_location='Xp22.13', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=18890296, end_position_on_the_genomic_accession=18894975, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375044, CurrentID=0, Status='live', Symbol='LOC105375044', Aliases='', description='uncharacterized LOC105375044', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38640395, end_position_on_the_genomic_accession=38664180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375045, CurrentID=0, Status='live', Symbol='LOC105375045', Aliases='', description='uncharacterized LOC105375045', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=38242880, end_position_on_the_genomic_accession=38246616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132166, CurrentID=0, Status='live', Symbol='IL9RP4', Aliases='', description='interleukin 9 receptor pseudogene 4', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79721, end_position_on_the_genomic_accession=88561, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132167, CurrentID=0, Status='live', Symbol='LOC100132167', Aliases='', description='uncharacterized LOC100132167', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39984323, end_position_on_the_genomic_accession=39989420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375048, CurrentID=0, Status='live', Symbol='LOC105375048', Aliases='', description='uncharacterized LOC105375048', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=39807845, end_position_on_the_genomic_accession=39819734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132169, CurrentID=0, Status='live', Symbol='WASIR2', Aliases='NCRNA00286A', description='WASH and IL9R antisense RNA 2', other_designations='WASH and IL9R antisense RNA 2 (non-protein coding)|non-protein coding RNA 286A', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22916, end_position_on_the_genomic_accession=25114, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375050, CurrentID=0, Status='live', Symbol='LOC105375050', Aliases='', description='uncharacterized LOC105375050', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40271606, end_position_on_the_genomic_accession=40276310, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375051, CurrentID=0, Status='live', Symbol='LOC105375051', Aliases='', description='uncharacterized LOC105375051', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40524041, end_position_on_the_genomic_accession=40529389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255308, CurrentID=0, Status='live', Symbol='LOC255308', Aliases='', description='eukaryotic translation initiation factor 2 subunit gamma pseudogene', other_designations='eukaryotic translation initiation factor 2, subunit 3 gamma, 52kDa pseudogene', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10505883, end_position_on_the_genomic_accession=10508527, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132173, CurrentID=0, Status='live', Symbol='PPP1R26P3', Aliases='KIAA0649P3', description='protein phosphatase 1 regulatory subunit 26 pseudogene 3', other_designations='protein phosphatase 1, regulatory subunit 26 pseudogene 3', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18402096, end_position_on_the_genomic_accession=18404068, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375054, CurrentID=0, Status='live', Symbol='LOC105375054', Aliases='', description='uncharacterized LOC105375054', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=40858546, end_position_on_the_genomic_accession=40868551, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375055, CurrentID=0, Status='live', Symbol='LOC105375055', Aliases='', description='uncharacterized LOC105375055', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41221275, end_position_on_the_genomic_accession=41225181, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375056, CurrentID=0, Status='live', Symbol='LOC105375056', Aliases='', description='uncharacterized LOC105375056', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41140894, end_position_on_the_genomic_accession=41178551, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  255313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255313, CurrentID=0, Status='live', Symbol='CT47A11', Aliases='CT47.11', description='cancer/testis antigen family 47, member A11', other_designations='cancer/testis CT47 family, member 11', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120933840, end_position_on_the_genomic_accession=120937198, orientation='minus', exon_count=4, OMIM=300592, no_hdr0=''),
  105375058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375058, CurrentID=0, Status='live', Symbol='LOC105375058', Aliases='', description='uncharacterized LOC105375058', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41452889, end_position_on_the_genomic_accession=41487130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375059, CurrentID=0, Status='live', Symbol='LOC105375059', Aliases='', description='uncharacterized LOC105375059', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42013860, end_position_on_the_genomic_accession=42017799, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  648532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648532, CurrentID=0, Status='live', Symbol='NRBF2P1', Aliases='', description='nuclear receptor binding factor 2 pseudogene 1', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35841581, end_position_on_the_genomic_accession=35843425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  124245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124245, CurrentID=0, Status='live', Symbol='ZC3H18', Aliases='NHN1', description='zinc finger CCCH-type containing 18', other_designations='conserved nuclear protein NHN1|nuclear protein NHN1', map_location='16q24.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88570379, end_position_on_the_genomic_accession=88631964, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105375062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375062, CurrentID=0, Status='live', Symbol='LOC105375062', Aliases='', description='uncharacterized LOC105375062', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42746421, end_position_on_the_genomic_accession=42781996, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132183, CurrentID=0, Status='live', Symbol='FP6628', Aliases='', description='uncharacterized LOC100132183', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=94725262, end_position_on_the_genomic_accession=94728360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375064, CurrentID=0, Status='live', Symbol='LOC105375064', Aliases='', description='uncharacterized LOC105375064', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43281417, end_position_on_the_genomic_accession=43283628, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100423020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423020, CurrentID=0, Status='live', Symbol='MIR4258', Aliases='', description='microRNA 4258', other_designations='hsa-mir-4258', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154975693, end_position_on_the_genomic_accession=154975783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375066, CurrentID=0, Status='live', Symbol='LOC105375066', Aliases='', description='uncharacterized LOC105375066', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43722548, end_position_on_the_genomic_accession=43737945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375067, CurrentID=0, Status='live', Symbol='LOC105375067', Aliases='', description='uncharacterized LOC105375067', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43899879, end_position_on_the_genomic_accession=43902490, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  255324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255324, CurrentID=0, Status='live', Symbol='EPGN', Aliases='ALGV3072, EPG, PRO9904', description='epithelial mitogen', other_designations='epithelial mitogen homolog', map_location='4q13.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74308270, end_position_on_the_genomic_accession=74317298, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375069, CurrentID=0, Status='live', Symbol='LOC105375069', Aliases='', description='uncharacterized LOC105375069', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43836932, end_position_on_the_genomic_accession=43838732, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375070, CurrentID=0, Status='live', Symbol='LOC105375070', Aliases='', description='uncharacterized LOC105375070', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43797121, end_position_on_the_genomic_accession=43821702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375071, CurrentID=0, Status='live', Symbol='LOC105375071', Aliases='', description='uncharacterized LOC105375071', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44010454, end_position_on_the_genomic_accession=44048929, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375072, CurrentID=0, Status='live', Symbol='LOC105375072', Aliases='', description='uncharacterized LOC105375072', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44185989, end_position_on_the_genomic_accession=44189941, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375073, CurrentID=0, Status='live', Symbol='LOC105375073', Aliases='', description='uncharacterized LOC105375073', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=44306943, end_position_on_the_genomic_accession=44308325, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255330, CurrentID=0, Status='live', Symbol='NUP210P1', Aliases='C3orf46', description='nucleoporin 210kDa pseudogene 1', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=126662081, end_position_on_the_genomic_accession=126671939, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132195, CurrentID=0, Status='live', Symbol='TRIM60P13', Aliases='', description='tripartite motif containing 60 pseudogene 13', other_designations='tripartite motif-containing 60 pseudogene 13', map_location='13q31.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=89248860, end_position_on_the_genomic_accession=89250671, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375076, CurrentID=0, Status='live', Symbol='LOC105375076', Aliases='', description='uncharacterized LOC105375076', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45574453, end_position_on_the_genomic_accession=45578085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375077, CurrentID=0, Status='live', Symbol='LOC105375077', Aliases='', description='uncharacterized LOC105375077', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45555181, end_position_on_the_genomic_accession=45576785, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105375078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375078, CurrentID=0, Status='live', Symbol='LOC105375078', Aliases='', description='uncharacterized LOC105375078', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=45911967, end_position_on_the_genomic_accession=45930080, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132199, CurrentID=0, Status='live', Symbol='RPL36AP44', Aliases='RPL36A_21_1441', description='ribosomal protein L36a pseudogene 44', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=60390338, end_position_on_the_genomic_accession=60390731, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375080, CurrentID=0, Status='live', Symbol='LOC105375080', Aliases='', description='uncharacterized LOC105375080', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=46931865, end_position_on_the_genomic_accession=46937165, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132201, CurrentID=0, Status='live', Symbol='ZSWIM5P2', Aliases='', description='zinc finger, SWIM-type containing 5 pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20583758, end_position_on_the_genomic_accession=20586458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132202, CurrentID=0, Status='live', Symbol='LOC100132202', Aliases='', description='golgin subfamily A member 6-like protein 22', other_designations='', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28553488, end_position_on_the_genomic_accession=28562239, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375083, CurrentID=0, Status='live', Symbol='LOC105375083', Aliases='', description='uncharacterized LOC105375083', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49773367, end_position_on_the_genomic_accession=49782499, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132204, CurrentID=0, Status='live', Symbol='DNM1P29', Aliases='DNM1DN3-2, DNM1DN3.2, DNM1DN3@', description='dynamin 1 pseudogene 29', other_designations='DNM1 pseudogene 29|DNM1DN3.2 duplicon', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30393346, end_position_on_the_genomic_accession=30398394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375085, CurrentID=0, Status='live', Symbol='LOC105375085', Aliases='', description='uncharacterized LOC105375085', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51398321, end_position_on_the_genomic_accession=51408679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375086, CurrentID=0, Status='live', Symbol='LOC105375086', Aliases='', description='uncharacterized LOC105375086', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52068808, end_position_on_the_genomic_accession=52073134, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375087, CurrentID=0, Status='live', Symbol='LOC105375087', Aliases='', description='uncharacterized LOC105375087', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=51651839, end_position_on_the_genomic_accession=51685413, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375088, CurrentID=0, Status='live', Symbol='LOC105375088', Aliases='', description='uncharacterized LOC105375088', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52237114, end_position_on_the_genomic_accession=52238823, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375089, CurrentID=0, Status='live', Symbol='LOC105375089', Aliases='', description='uncharacterized LOC105375089', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52522989, end_position_on_the_genomic_accession=52531140, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124274, CurrentID=0, Status='live', Symbol='GPR139', Aliases='GPRg1, PGR3', description='G protein-coupled receptor 139', other_designations='G protein-coupled receptor PGR3|G-protein-coupled receptor PGR3|g(q)-coupled orphan receptor GPRg1', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20031455, end_position_on_the_genomic_accession=20073778, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375091, CurrentID=0, Status='live', Symbol='LOC105375091', Aliases='', description='uncharacterized LOC105375091', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52882076, end_position_on_the_genomic_accession=52905737, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375092, CurrentID=0, Status='live', Symbol='LOC105375092', Aliases='', description='uncharacterized LOC105375092', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53064693, end_position_on_the_genomic_accession=53065792, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  255349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255349, CurrentID=0, Status='live', Symbol='TMEM211', Aliases='bA9F11.1', description='transmembrane protein 211', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24934773, end_position_on_the_genomic_accession=24939347, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375094, CurrentID=0, Status='live', Symbol='LOC105375094', Aliases='', description='uncharacterized LOC105375094', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53193543, end_position_on_the_genomic_accession=53262771, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100423021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423021, CurrentID=0, Status='live', Symbol='MIR4298', Aliases='', description='microRNA 4298', other_designations='hsa-mir-4298', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1859464, end_position_on_the_genomic_accession=1859536, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375096, CurrentID=0, Status='live', Symbol='LOC105375096', Aliases='', description='uncharacterized LOC105375096', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=53848296, end_position_on_the_genomic_accession=53850799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132217, CurrentID=0, Status='live', Symbol='LOC100132217', Aliases='', description='uncharacterized LOC100132217', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375098, CurrentID=0, Status='live', Symbol='LOC105375098', Aliases='', description='uncharacterized LOC105375098', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54301168, end_position_on_the_genomic_accession=54302708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375099, CurrentID=0, Status='live', Symbol='LOC105375099', Aliases='', description='uncharacterized LOC105375099', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55862875, end_position_on_the_genomic_accession=55869546, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375100, CurrentID=0, Status='live', Symbol='LOC105375100', Aliases='', description='uncharacterized LOC105375100', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=56012808, end_position_on_the_genomic_accession=56019173, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375101, CurrentID=0, Status='live', Symbol='LOC105375101', Aliases='', description='uncharacterized LOC105375101', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57821241, end_position_on_the_genomic_accession=57846232, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375102, CurrentID=0, Status='live', Symbol='LOC105375102', Aliases='', description='uncharacterized LOC105375102', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57811149, end_position_on_the_genomic_accession=57822628, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375103, CurrentID=0, Status='live', Symbol='LOC105375103', Aliases='', description='uncharacterized LOC105375103', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57792090, end_position_on_the_genomic_accession=57798702, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375104, CurrentID=0, Status='live', Symbol='LOC105375104', Aliases='', description='uncharacterized LOC105375104', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=364143, end_position_on_the_genomic_accession=380671, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375105, CurrentID=0, Status='live', Symbol='LOC105375105', Aliases='', description='uncharacterized LOC105375105', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=20400748, end_position_on_the_genomic_accession=20402772, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375106, CurrentID=0, Status='live', Symbol='LOC105375106', Aliases='', description='uncharacterized LOC105375106', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25273478, end_position_on_the_genomic_accession=25279431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375107, CurrentID=0, Status='live', Symbol='LOC105375107', Aliases='', description='uncharacterized LOC105375107', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27312169, end_position_on_the_genomic_accession=27314612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375108, CurrentID=0, Status='live', Symbol='LOC105375108', Aliases='', description='uncharacterized LOC105375108', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30464493, end_position_on_the_genomic_accession=30467855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132229, CurrentID=0, Status='live', Symbol='LOC100132229', Aliases='', description='nuclear pore associated protein 1 pseudogene', other_designations='Putative UPF0607 protein ENSP00000382826', map_location='8p21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27010456, end_position_on_the_genomic_accession=27012444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132230, CurrentID=0, Status='live', Symbol='FAM201CP', Aliases='FAM201C', description='family with sequence similarity 201 member C, pseudogene', other_designations='family with sequence similarity 201, member C, pseudogene', map_location='Yq11.1', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=13251502, end_position_on_the_genomic_accession=13263811, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375111, CurrentID=0, Status='live', Symbol='LOC105375111', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57927950, end_position_on_the_genomic_accession=57946670, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375112, CurrentID=0, Status='live', Symbol='LOC105375112', Aliases='', description='putative protein FAM157B', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=31567, end_position_on_the_genomic_accession=37465, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375113, CurrentID=0, Status='live', Symbol='LOC105375113', Aliases='', description='uncharacterized LOC105375113', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=70809, end_position_on_the_genomic_accession=75742, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132234, CurrentID=0, Status='live', Symbol='LINC00543', Aliases='', description='long intergenic non-protein coding RNA 543', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.10', start_position_on_the_genomic_accession=28527669, end_position_on_the_genomic_accession=28529500, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375115, CurrentID=0, Status='live', Symbol='LOC105375115', Aliases='', description='uncharacterized LOC105375115', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=174920, end_position_on_the_genomic_accession=176114, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375116, CurrentID=0, Status='live', Symbol='LOC105375116', Aliases='', description='TRIO and F-actin-binding protein-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=176197, end_position_on_the_genomic_accession=180728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375117, CurrentID=0, Status='live', Symbol='LOC105375117', Aliases='', description='uncharacterized LOC105375117', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=181485, end_position_on_the_genomic_accession=183814, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255374, CurrentID=0, Status='live', Symbol='MBLAC1', Aliases='', description='metallo-beta-lactamase domain containing 1', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100126694, end_position_on_the_genomic_accession=100128724, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375119, CurrentID=0, Status='live', Symbol='LOC105375119', Aliases='', description='uncharacterized LOC105375119', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=612961, end_position_on_the_genomic_accession=614572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106480976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480976, CurrentID=0, Status='live', Symbol='RN7SL200P', Aliases='', description='RNA, 7SL, cytoplasmic 200, pseudogene', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375121, CurrentID=0, Status='live', Symbol='LOC105375121', Aliases='', description='uncharacterized LOC105375121', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1521724, end_position_on_the_genomic_accession=1525609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375122, CurrentID=0, Status='live', Symbol='LOC105375122', Aliases='', description='uncharacterized LOC105375122', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1592357, end_position_on_the_genomic_accession=1598609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375123, CurrentID=0, Status='live', Symbol='LOC105375123', Aliases='', description='uncharacterized LOC105375123', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1601733, end_position_on_the_genomic_accession=1603722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375124, CurrentID=0, Status='live', Symbol='LOC105375124', Aliases='', description='uncharacterized LOC105375124', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1720767, end_position_on_the_genomic_accession=1725035, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375125, CurrentID=0, Status='live', Symbol='LOC105375125', Aliases='', description='uncharacterized LOC105375125', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1718018, end_position_on_the_genomic_accession=1718978, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375126, CurrentID=0, Status='live', Symbol='LOC105375126', Aliases='', description='uncharacterized LOC105375126', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=2182527, end_position_on_the_genomic_accession=2205792, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132247, CurrentID=0, Status='live', Symbol='NPIPB5', Aliases='', description='nuclear pore complex interacting protein family, member B5', other_designations='nuclear pore complex interacting protein related', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=22513523, end_position_on_the_genomic_accession=22536535, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105375128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375128, CurrentID=0, Status='live', Symbol='LOC105375128', Aliases='', description='uncharacterized LOC105375128', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1817459, end_position_on_the_genomic_accession=1819236, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132249, CurrentID=0, Status='live', Symbol='LOC100132249', Aliases='', description='uncharacterized LOC100132249', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62834247, end_position_on_the_genomic_accession=62838368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375130, CurrentID=0, Status='live', Symbol='LOC105375130', Aliases='', description='uncharacterized LOC105375130', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3094030, end_position_on_the_genomic_accession=3101381, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132251, CurrentID=0, Status='live', Symbol='ABHD17AP9', Aliases='FAM108A10P', description='abhydrolase domain containing 17A pseudogene 9', other_designations='family with sequence similarity 108, member A10, pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33140868, end_position_on_the_genomic_accession=33144023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375132, CurrentID=0, Status='live', Symbol='LOC105375132', Aliases='', description='uncharacterized LOC105375132', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=3260333, end_position_on_the_genomic_accession=3302522, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375133, CurrentID=0, Status='live', Symbol='LOC105375133', Aliases='', description='uncharacterized LOC105375133', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=4533745, end_position_on_the_genomic_accession=4569716, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375134, CurrentID=0, Status='live', Symbol='LOC105375134', Aliases='', description='uncharacterized LOC105375134', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5033776, end_position_on_the_genomic_accession=5034750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375135, CurrentID=0, Status='live', Symbol='LOC105375135', Aliases='', description='uncharacterized LOC105375135', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5686025, end_position_on_the_genomic_accession=5697583, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375136, CurrentID=0, Status='live', Symbol='LOC105375136', Aliases='', description='uncharacterized LOC105375136', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6588876, end_position_on_the_genomic_accession=6589546, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132257, CurrentID=0, Status='live', Symbol='LOC100132257', Aliases='', description='Putative ankyrin repeat domain-containing protein ENSP00000383069', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=19108950, end_position_on_the_genomic_accession=19118312, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  255394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255394, CurrentID=0, Status='live', Symbol='TCP11L2', Aliases='', description='t-complex 11, testis-specific-like 2', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=106302784, end_position_on_the_genomic_accession=106347587, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105375139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375139, CurrentID=0, Status='live', Symbol='LOC105375139', Aliases='', description='uncharacterized LOC105375139', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7929358, end_position_on_the_genomic_accession=7968022, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375140, CurrentID=0, Status='live', Symbol='LOC105375140', Aliases='', description='uncharacterized LOC105375140', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7883075, end_position_on_the_genomic_accession=7884695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375141, CurrentID=0, Status='live', Symbol='LOC105375141', Aliases='', description='uncharacterized LOC105375141', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=7738534, end_position_on_the_genomic_accession=7742155, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375142, CurrentID=0, Status='live', Symbol='LOC105375142', Aliases='', description='uncharacterized LOC105375142', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8092701, end_position_on_the_genomic_accession=8095681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375143, CurrentID=0, Status='live', Symbol='LOC105375143', Aliases='', description='uncharacterized LOC105375143', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=8209800, end_position_on_the_genomic_accession=8216296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132264, CurrentID=0, Status='live', Symbol='RNFT1P2', Aliases='', description='ring finger protein, transmembrane 1 pseudogene 2', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=78170470, end_position_on_the_genomic_accession=78171243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132265, CurrentID=0, Status='live', Symbol='RPL32P15', Aliases='RPL32_5_690', description='ribosomal protein L32 pseudogene 15', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=41307958, end_position_on_the_genomic_accession=41310305, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375146, CurrentID=0, Status='live', Symbol='LOC105375146', Aliases='', description='uncharacterized LOC105375146', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9961075, end_position_on_the_genomic_accession=9986360, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  255403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255403, CurrentID=0, Status='live', Symbol='ZNF718', Aliases='', description='zinc finger protein 718', other_designations='', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124476, end_position_on_the_genomic_accession=202306, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100132268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132268, CurrentID=0, Status='live', Symbol='FKBP4P6', Aliases='FKBP4P4', description='FK506 binding protein 4 pseudogene 6', other_designations='FK506 binding protein 4, 59kDa pseudogene 4|FK506 binding protein 4, 59kDa pseudogene 6', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42384661, end_position_on_the_genomic_accession=42386230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375149, CurrentID=0, Status='live', Symbol='LOC105375149', Aliases='', description='uncharacterized LOC105375149', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10090343, end_position_on_the_genomic_accession=10161810, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375150, CurrentID=0, Status='live', Symbol='LOC105375150', Aliases='', description='uncharacterized LOC105375150', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=10206694, end_position_on_the_genomic_accession=10213281, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375151, CurrentID=0, Status='live', Symbol='LOC105375151', Aliases='', description='uncharacterized LOC105375151', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11567342, end_position_on_the_genomic_accession=11594174, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375152, CurrentID=0, Status='live', Symbol='LOC105375152', Aliases='', description='uncharacterized LOC105375152', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=11407063, end_position_on_the_genomic_accession=11417286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132273, CurrentID=0, Status='live', Symbol='NDUFA6-AS1', Aliases='', description='NDUFA6 antisense RNA 1 (head to head)', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42090933, end_position_on_the_genomic_accession=42125349, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375154, CurrentID=0, Status='live', Symbol='LOC105375154', Aliases='', description='uncharacterized LOC105375154', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12406089, end_position_on_the_genomic_accession=12407895, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100423023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423023, CurrentID=0, Status='live', Symbol='MIR3197', Aliases='', description='microRNA 3197', other_designations='hsa-mir-3197', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=41167557, end_position_on_the_genomic_accession=41167629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375156, CurrentID=0, Status='live', Symbol='LOC105375156', Aliases='', description='uncharacterized LOC105375156', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12490483, end_position_on_the_genomic_accession=12496343, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375157, CurrentID=0, Status='live', Symbol='LOC105375157', Aliases='', description='uncharacterized LOC105375157', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=12674036, end_position_on_the_genomic_accession=12675534, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  648630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648630, CurrentID=0, Status='live', Symbol='TSSK5P', Aliases='TSSK5P12, TSSKps1, TSSK5P', description='testis-specific serine kinase 5, pseudogene', other_designations='testis-specific serine kinase 5 pseudogene 1|testis-specific serine kinase 5 pseudogene 2', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144140864, end_position_on_the_genomic_accession=144143664, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375159, CurrentID=0, Status='live', Symbol='LOC105375159', Aliases='', description='uncharacterized LOC105375159', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=13092106, end_position_on_the_genomic_accession=13217613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375160, CurrentID=0, Status='live', Symbol='LOC105375160', Aliases='', description='uncharacterized LOC105375160', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=13717782, end_position_on_the_genomic_accession=13728612, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375161, CurrentID=0, Status='live', Symbol='LOC105375161', Aliases='', description='uncharacterized LOC105375161', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=13728653, end_position_on_the_genomic_accession=13743574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375162, CurrentID=0, Status='live', Symbol='LOC105375162', Aliases='', description='uncharacterized LOC105375162', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14051111, end_position_on_the_genomic_accession=14061781, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375163, CurrentID=0, Status='live', Symbol='LOC105375163', Aliases='', description='uncharacterized LOC105375163', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14799370, end_position_on_the_genomic_accession=14816694, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375164, CurrentID=0, Status='live', Symbol='LOC105375164', Aliases='', description='uncharacterized LOC105375164', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=14375774, end_position_on_the_genomic_accession=14382903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132285, CurrentID=0, Status='live', Symbol='KIR2DS2', Aliases='183ActI, CD158J, CD158b, NKAT-5, NKAT5, cl-49', description='killer cell immunoglobulin-like receptor, two domains, short cytoplasmic tail, 2', other_designations='CD158 antigen-like family member J|MHC class I NK cell receptor|NK receptor 183 ActI|killer-cell Ig-like receptor|killer-cell immunoglobulin-like receptor two domains short tail 2 protein|natural killer associated transcript 5|natural killer cell inhibitory receptor|p58 killer cell inhibitory receptor KIR-K7a|p58 natural killer cell receptor clone CL-49', map_location='19q13.4', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=604953, no_hdr0=''),
  100132286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132286, CurrentID=0, Status='live', Symbol='NDUFB8P3', Aliases='', description='NADH:ubiquinone oxidoreductase subunit B8 pseudogene 3', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, 19kDa pseudogene|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 8, pseudogene 3', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34674223, end_position_on_the_genomic_accession=34710264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132287, CurrentID=0, Status='live', Symbol='LOC100132287', Aliases='', description='uncharacterized LOC100132287', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=490756, end_position_on_the_genomic_accession=495445, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132288, CurrentID=0, Status='live', Symbol='TEKT4P2', Aliases='MAFIPL, TEKT4P', description='tektin 4 pseudogene 2', other_designations='Tektin-4 like protein LOC389833', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9068356, end_position_on_the_genomic_accession=9129761, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132289, CurrentID=0, Status='live', Symbol='ABCB10P4', Aliases='', description='ATP binding cassette subfamily B member 10 pseudogene 4', other_designations='ATP-binding cassette, sub-family B (MDR/TAP), member 10 pseudogene 4', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28506627, end_position_on_the_genomic_accession=28508793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  255426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255426, CurrentID=0, Status='live', Symbol='RASGEF1C', Aliases='', description='RasGEF domain family member 1C', other_designations='RasGEF domain family, member 1C', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180100795, end_position_on_the_genomic_accession=180209130, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100132291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132291, CurrentID=0, Status='live', Symbol='RPS27P29', Aliases='RPS27_17_1641', description='ribosomal protein S27 pseudogene 29', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23597307, end_position_on_the_genomic_accession=23597649, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375172, CurrentID=0, Status='live', Symbol='LOC105375172', Aliases='', description='uncharacterized LOC105375172', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17430587, end_position_on_the_genomic_accession=17463138, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375173, CurrentID=0, Status='live', Symbol='LOC105375173', Aliases='', description='uncharacterized LOC105375173', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=17900286, end_position_on_the_genomic_accession=17958952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375174, CurrentID=0, Status='live', Symbol='LOC105375174', Aliases='', description='uncharacterized LOC105375174', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19112481, end_position_on_the_genomic_accession=19115880, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124359, CurrentID=0, Status='live', Symbol='CDYL2', Aliases='PCCP1', description='chromodomain protein, Y-like 2', other_designations='CDY-like 2|prostate cancer candidate protein 1, PCCP1', map_location='16q23.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80597899, end_position_on_the_genomic_accession=80805043, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105375176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375176, CurrentID=0, Status='live', Symbol='LOC105375176', Aliases='', description='uncharacterized LOC105375176', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18885889, end_position_on_the_genomic_accession=18902613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375177, CurrentID=0, Status='live', Symbol='LOC105375177', Aliases='', description='uncharacterized LOC105375177', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18653149, end_position_on_the_genomic_accession=18659887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375178, CurrentID=0, Status='live', Symbol='LOC105375178', Aliases='', description='uncharacterized LOC105375178', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=18543292, end_position_on_the_genomic_accession=18545930, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375179, CurrentID=0, Status='live', Symbol='LOC105375179', Aliases='', description='uncharacterized LOC105375179', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19403696, end_position_on_the_genomic_accession=19433670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106480978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480978, CurrentID=0, Status='live', Symbol='RN7SL208P', Aliases='', description='RNA, 7SL, cytoplasmic 208, pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375181, CurrentID=0, Status='live', Symbol='LOC105375181', Aliases='', description='uncharacterized LOC105375181', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=20104918, end_position_on_the_genomic_accession=20119655, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375182, CurrentID=0, Status='live', Symbol='LOC105375182', Aliases='', description='uncharacterized LOC105375182', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=19944909, end_position_on_the_genomic_accession=19984764, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375183, CurrentID=0, Status='live', Symbol='LOC105375183', Aliases='', description='uncharacterized LOC105375183', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=21560086, end_position_on_the_genomic_accession=21576423, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132304, CurrentID=0, Status='live', Symbol='LOC100132304', Aliases='', description='long intergenic non-protein coding RNA 684', other_designations='LINC00684', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72781865, end_position_on_the_genomic_accession=72782660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375185, CurrentID=0, Status='live', Symbol='LOC105375185', Aliases='', description='uncharacterized LOC105375185', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23021721, end_position_on_the_genomic_accession=23030374, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132306, CurrentID=0, Status='live', Symbol='FAM91A4P', Aliases='', description='family with sequence similarity 91 member A4, pseudogene', other_designations='family with sequence similarity 91, member A4 pseudogene', map_location='1p11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120983370, end_position_on_the_genomic_accession=120988697, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375187, CurrentID=0, Status='live', Symbol='LOC105375187', Aliases='', description='uncharacterized LOC105375187', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23586694, end_position_on_the_genomic_accession=23597266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132308, CurrentID=0, Status='live', Symbol='SLC29A4P2', Aliases='', description='solute carrier family 29 (nucleoside transporters), member 4 pseudogene 2', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63556598, end_position_on_the_genomic_accession=63562588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375189, CurrentID=0, Status='live', Symbol='LOC105375189', Aliases='', description='uncharacterized LOC105375189', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=24227786, end_position_on_the_genomic_accession=24234603, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132310, CurrentID=0, Status='live', Symbol='FCF1P11', Aliases='', description='FCF1 pseudogene 11', other_designations='FCF1 small subunit processome component homolog pseudogene 11', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139198556, end_position_on_the_genomic_accession=139199142, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132311, CurrentID=0, Status='live', Symbol='RPL38P3', Aliases='RPL38_2_479', description='ribosomal protein L38 pseudogene 3', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54976115, end_position_on_the_genomic_accession=54976397, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375192, CurrentID=0, Status='live', Symbol='LOC105375192', Aliases='', description='tRNA-splicing endonuclease subunit Sen15 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25265922, end_position_on_the_genomic_accession=25266289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375193, CurrentID=0, Status='live', Symbol='LOC105375193', Aliases='', description='uncharacterized LOC105375193', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25321503, end_position_on_the_genomic_accession=25321920, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375194, CurrentID=0, Status='live', Symbol='LOC105375194', Aliases='', description='uncharacterized LOC105375194', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25332890, end_position_on_the_genomic_accession=25333714, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375195, CurrentID=0, Status='live', Symbol='LOC105375195', Aliases='', description='uncharacterized LOC105375195', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25329205, end_position_on_the_genomic_accession=25359926, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375196, CurrentID=0, Status='live', Symbol='LOC105375196', Aliases='', description='uncharacterized LOC105375196', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25486968, end_position_on_the_genomic_accession=25496732, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375197, CurrentID=0, Status='live', Symbol='LOC105375197', Aliases='', description='uncharacterized LOC105375197', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25762758, end_position_on_the_genomic_accession=25777196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375198, CurrentID=0, Status='live', Symbol='LOC105375198', Aliases='', description='uncharacterized LOC105375198', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26101638, end_position_on_the_genomic_accession=26123242, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100873838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100873838, CurrentID=0, Status='live', Symbol='RNU5A-8P', Aliases='', description='RNA, U5A small nuclear 8, pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=210374154, end_position_on_the_genomic_accession=210374267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375200, CurrentID=0, Status='live', Symbol='LOC105375200', Aliases='', description='uncharacterized LOC105375200', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=25941537, end_position_on_the_genomic_accession=25945831, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132321, CurrentID=0, Status='live', Symbol='NIPA2P1', Aliases='', description='non imprinted in Prader-Willi/Angelman syndrome 2 pseudogene 1', other_designations='', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91319999, end_position_on_the_genomic_accession=91321194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101180898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180898, CurrentID=0, Status='live', Symbol='LINC00237', Aliases='NCRNA00237', description='long intergenic non-protein coding RNA 237', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614992, no_hdr0=''),
  101180899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180899, CurrentID=0, Status='live', Symbol='UCH1LAS', Aliases='', description='antisense UCHL1', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614991, no_hdr0=''),
  101180900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180900, CurrentID=0, Status='live', Symbol='MSE', Aliases='', description='myelinating Schwann cell element', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614996, no_hdr0=''),
  101180901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180901, CurrentID=0, Status='live', Symbol='IPO11-LRRC70', Aliases='', description='IPO11-LRRC70 readthrough', other_designations='', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62578735, end_position_on_the_genomic_accession=62628589, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375206, CurrentID=0, Status='live', Symbol='LOC105375206', Aliases='', description='uncharacterized LOC105375206', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27379933, end_position_on_the_genomic_accession=27381328, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375207, CurrentID=0, Status='live', Symbol='LOC105375207', Aliases='', description='uncharacterized LOC105375207', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27387778, end_position_on_the_genomic_accession=27409118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375208, CurrentID=0, Status='live', Symbol='LOC105375208', Aliases='', description='uncharacterized LOC105375208', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=28037569, end_position_on_the_genomic_accession=28063180, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375209, CurrentID=0, Status='live', Symbol='LOC105375209', Aliases='', description='uncharacterized LOC105375209', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27943309, end_position_on_the_genomic_accession=27955715, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106480979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480979, CurrentID=0, Status='live', Symbol='RN7SL213P', Aliases='', description='RNA, 7SL, cytoplasmic 213, pseudogene', other_designations='', map_location='14q24.1', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101180907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180907, CurrentID=0, Status='live', Symbol='USH1K', Aliases='', description='Usher syndrome 1K (autosomal recessive)', other_designations='', map_location='10p11.21-q21.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614990, no_hdr0=''),
  100132332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132332, CurrentID=0, Status='live', Symbol='CNOT7P2', Aliases='', description='CCR4-NOT transcription complex subunit 7 pseudogene 2', other_designations='CCR4-NOT transcription complex, subunit 7 pseudogene 2', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115564601, end_position_on_the_genomic_accession=115565429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375213, CurrentID=0, Status='live', Symbol='LOC105375213', Aliases='', description='uncharacterized LOC105375213', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29639695, end_position_on_the_genomic_accession=29645161, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375214, CurrentID=0, Status='live', Symbol='LOC105375214', Aliases='', description='uncharacterized LOC105375214', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29510185, end_position_on_the_genomic_accession=29512557, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375215, CurrentID=0, Status='live', Symbol='LOC105375215', Aliases='', description='uncharacterized LOC105375215', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=29987806, end_position_on_the_genomic_accession=30026305, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375216, CurrentID=0, Status='live', Symbol='LOC105375216', Aliases='', description='uncharacterized LOC105375216', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30130579, end_position_on_the_genomic_accession=30134670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124401, CurrentID=0, Status='live', Symbol='ANKS3', Aliases='', description='ankyrin repeat and sterile alpha motif domain containing 3', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4696510, end_position_on_the_genomic_accession=4734377, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  124402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124402, CurrentID=0, Status='live', Symbol='UBALD1', Aliases='FAM100A, PP11303', description='UBA-like domain containing 1', other_designations='family with sequence similarity 100, member A|protein FAM100A', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4608883, end_position_on_the_genomic_accession=4614926, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  648691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648691, CurrentID=0, Status='live', Symbol='LL22NC03-63E9.3', Aliases='', description='uncharacterized LOC648691', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22559343, end_position_on_the_genomic_accession=22566599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124404, CurrentID=0, Status='live', Symbol='SEPT12', Aliases='SPGF10', description='septin 12', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4777609, end_position_on_the_genomic_accession=4788521, orientation='minus', exon_count=10, OMIM=611562, no_hdr0=''),
  100132341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132341, CurrentID=0, Status='live', Symbol='CLUHP3', Aliases='C16orf67, KIAA0664L3, KIAA0664P3, URLC3', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 3', other_designations='KIAA0664-like 3|up-regulated in lung cancer 3', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31700613, end_position_on_the_genomic_accession=31707437, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375222, CurrentID=0, Status='live', Symbol='LOC105375222', Aliases='', description='uncharacterized LOC105375222', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30983868, end_position_on_the_genomic_accession=31031031, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375223, CurrentID=0, Status='live', Symbol='LOC105375223', Aliases='', description='uncharacterized LOC105375223', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=32029290, end_position_on_the_genomic_accession=32030412, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255480, CurrentID=0, Status='live', Symbol='TBX5-AS1', Aliases='', description='TBX5 antisense RNA 1', other_designations='TBX5 antisense RNA 1 (non-protein coding)', map_location='12q24.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114408195, end_position_on_the_genomic_accession=114412832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132346, CurrentID=0, Status='live', Symbol='HSPE1P7', Aliases='', description='heat shock protein family E (Hsp10) member 1 pseudogene 7', other_designations='heat shock 10kDa protein 1 (chaperonin 10) pseudogene|heat shock 10kDa protein 1 pseudogene 7', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74612091, end_position_on_the_genomic_accession=74613249, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  124411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124411, CurrentID=0, Status='live', Symbol='ZNF720', Aliases='', description='zinc finger protein 720', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31713229, end_position_on_the_genomic_accession=31794641, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100132348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132348, CurrentID=0, Status='live', Symbol='LOC100132348', Aliases='', description='uncharacterized LOC100132348', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.10', start_position_on_the_genomic_accession=88917060, end_position_on_the_genomic_accession=89028910, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375229, CurrentID=0, Status='live', Symbol='LOC105375229', Aliases='', description='uncharacterized LOC105375229', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35190686, end_position_on_the_genomic_accession=35199219, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375230, CurrentID=0, Status='live', Symbol='LOC105375230', Aliases='', description='uncharacterized LOC105375230', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35384949, end_position_on_the_genomic_accession=35387287, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375231, CurrentID=0, Status='live', Symbol='LOC105375231', Aliases='', description='uncharacterized LOC105375231', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=35594220, end_position_on_the_genomic_accession=35609712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255488, CurrentID=0, Status='live', Symbol='RNF144B', Aliases='IBRDC2, PIR2, bA528A10.3, p53RFP', description='ring finger protein 144B', other_designations='IBR domain containing 2|IBR domain-containing protein 2|p53-inducible RING finger protein', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=18277515, end_position_on_the_genomic_accession=18468874, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375233, CurrentID=0, Status='live', Symbol='LOC105375233', Aliases='', description='uncharacterized LOC105375233', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=36001519, end_position_on_the_genomic_accession=36030574, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132354, CurrentID=0, Status='live', Symbol='LINC01512', Aliases='HI-LNC77', description='long intergenic non-protein coding RNA 1512', other_designations='Human Islet Long Non Coding RNA 77|TCONS_00011120', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=43891028, end_position_on_the_genomic_accession=43938207, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132355, CurrentID=0, Status='live', Symbol='SLC25A1P4', Aliases='', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1 pseudogene 4', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35173557, end_position_on_the_genomic_accession=35174091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132356, CurrentID=0, Status='live', Symbol='LOC100132356', Aliases='', description='uncharacterized LOC100132356', other_designations='', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43065187, end_position_on_the_genomic_accession=43066971, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375237, CurrentID=0, Status='live', Symbol='LOC105375237', Aliases='', description='uncharacterized LOC105375237', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39506588, end_position_on_the_genomic_accession=39522502, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132358, CurrentID=0, Status='live', Symbol='EEF1A1P34', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 34', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18782427, end_position_on_the_genomic_accession=18783690, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375239, CurrentID=0, Status='live', Symbol='LOC105375239', Aliases='', description='uncharacterized LOC105375239', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39327557, end_position_on_the_genomic_accession=39335202, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106480980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480980, CurrentID=0, Status='live', Symbol='RN7SL215P', Aliases='', description='RNA, 7SL, cytoplasmic 215, pseudogene', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375241, CurrentID=0, Status='live', Symbol='LOC105375241', Aliases='', description='uncharacterized LOC105375241', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41108868, end_position_on_the_genomic_accession=41149795, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375242, CurrentID=0, Status='live', Symbol='LOC105375242', Aliases='', description='uncharacterized LOC105375242', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40858425, end_position_on_the_genomic_accession=40902292, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132363, CurrentID=0, Status='live', Symbol='LOC100132363', Aliases='', description='uncharacterized LOC100132363', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=611061, end_position_on_the_genomic_accession=611723, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375244, CurrentID=0, Status='live', Symbol='LOC105375244', Aliases='', description='uncharacterized LOC105375244', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=40565173, end_position_on_the_genomic_accession=40595291, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100423026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423026, CurrentID=0, Status='live', Symbol='MIR4299', Aliases='', description='microRNA 4299', other_designations='hsa-mir-4299', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11656651, end_position_on_the_genomic_accession=11656722, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375246, CurrentID=0, Status='live', Symbol='LOC105375246', Aliases='', description='uncharacterized LOC105375246', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41332089, end_position_on_the_genomic_accession=41355671, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375247, CurrentID=0, Status='live', Symbol='LOC105375247', Aliases='', description='uncharacterized LOC105375247', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41656421, end_position_on_the_genomic_accession=41667312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375248, CurrentID=0, Status='live', Symbol='LOC105375248', Aliases='', description='uncharacterized LOC105375248', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=41933593, end_position_on_the_genomic_accession=41936933, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375249, CurrentID=0, Status='live', Symbol='LOC105375249', Aliases='', description='uncharacterized LOC105375249', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42326358, end_position_on_the_genomic_accession=42338382, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375250, CurrentID=0, Status='live', Symbol='LOC105375250', Aliases='', description='uncharacterized LOC105375250', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42496296, end_position_on_the_genomic_accession=42516554, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375251, CurrentID=0, Status='live', Symbol='LOC105375251', Aliases='', description='uncharacterized LOC105375251', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42591291, end_position_on_the_genomic_accession=42650382, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100132372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132372, CurrentID=0, Status='live', Symbol='RPS10P28', Aliases='RPS10_14_1658', description='ribosomal protein S10 pseudogene 28', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42669978, end_position_on_the_genomic_accession=42670861, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375253, CurrentID=0, Status='live', Symbol='LOC105375253', Aliases='', description='uncharacterized LOC105375253', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42953227, end_position_on_the_genomic_accession=42956129, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375254, CurrentID=0, Status='live', Symbol='LOC105375254', Aliases='', description='uncharacterized LOC105375254', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=42972332, end_position_on_the_genomic_accession=43044627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375255, CurrentID=0, Status='live', Symbol='LOC105375255', Aliases='', description='uncharacterized LOC105375255', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=43649716, end_position_on_the_genomic_accession=43650639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255512, CurrentID=0, Status='live', Symbol='TOLLIP-AS1', Aliases='', description='TOLLIP antisense RNA 1 (head to head)', other_designations='CTD-2245O6.1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1309708, end_position_on_the_genomic_accession=1310707, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  648729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648729, CurrentID=0, Status='live', Symbol='RPS15AP24', Aliases='RPS15A_13_889', description='ribosomal protein S15a pseudogene 24', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30117340, end_position_on_the_genomic_accession=30117809, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375258, CurrentID=0, Status='live', Symbol='LOC105375258', Aliases='', description='uncharacterized LOC105375258', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44147667, end_position_on_the_genomic_accession=44149773, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375259, CurrentID=0, Status='live', Symbol='LOC105375259', Aliases='', description='uncharacterized LOC105375259', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44795586, end_position_on_the_genomic_accession=44796159, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132380, CurrentID=0, Status='live', Symbol='AGGF1P1', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 1', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 1', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190040984, end_position_on_the_genomic_accession=190044286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375261, CurrentID=0, Status='live', Symbol='LOC105375261', Aliases='', description='uncharacterized LOC105375261', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=44848439, end_position_on_the_genomic_accession=44849565, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124446, CurrentID=0, Status='live', Symbol='TMEM219', Aliases='IGFBP-3R', description='transmembrane protein 219', other_designations='insulin-like growth factor binding protein-3 receptor', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29962030, end_position_on_the_genomic_accession=29973052, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375263, CurrentID=0, Status='live', Symbol='LOC105375263', Aliases='', description='uncharacterized LOC105375263', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45534589, end_position_on_the_genomic_accession=45539918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  255520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255520, CurrentID=0, Status='live', Symbol='ELMOD2', Aliases='9830169G11Rik', description='ELMO/CED-12 domain containing 2', other_designations='ELMO domain containing 2', map_location='4q31.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=140524158, end_position_on_the_genomic_accession=140553770, orientation='plus', exon_count=11, OMIM=610196, no_hdr0=''),
  386593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386593, CurrentID=0, Status='live', Symbol='CHKB-CPT1B', Aliases='CHKL-CPT1B', description='CHKB-CPT1B readthrough (NMD candidate)', other_designations='CHKB-CPT1B readthrough (non-protein coding)|choline kinase-like, carnitine palmitoyltransferase 1B (muscle) transcription unit', map_location='22q13.33', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=50568861, end_position_on_the_genomic_accession=50582999, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  100132386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132386, CurrentID=0, Status='live', Symbol='KRTAP4-9', Aliases='KAP4.9', description='keratin associated protein 4-9', other_designations='keratin-associated protein 4.9|ultrahigh sulfur keratin-associated protein 4.9', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41105389, end_position_on_the_genomic_accession=41106488, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375267, CurrentID=0, Status='live', Symbol='LOC105375267', Aliases='', description='uncharacterized LOC105375267', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=46969692, end_position_on_the_genomic_accession=46981879, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  648740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648740, CurrentID=0, Status='live', Symbol='ACTG1P4', Aliases='ACTGP4', description='actin gamma 1 pseudogene 4', other_designations='ACTB pseudogene|actin, beta pseudogene|actin, gamma 1 pseudogene 4|actin, gamma 2, smooth muscle, enteric pseudogene|actin, gamma pseudogene 4', map_location='1p21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103569404, end_position_on_the_genomic_accession=103571391, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  386597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386597, CurrentID=0, Status='live', Symbol='RNF144A-AS1', Aliases='', description='RNF144A antisense RNA 1', other_designations='RNF144A antisense RNA 1 (non-protein coding)', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6912276, end_position_on_the_genomic_accession=6918682, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  124454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124454, CurrentID=0, Status='live', Symbol='EARS2', Aliases='COXPD12, MSE1, gluRS', description='glutamyl-tRNA synthetase 2, mitochondrial', other_designations='glutamate tRNA ligase 2, mitochondrial|glutamate--tRNA ligase|probable glutamyl-tRNA synthetase, mitochondrial', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=23522013, end_position_on_the_genomic_accession=23557375, orientation='minus', exon_count=11, OMIM=612799, no_hdr0=''),
  105375271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375271, CurrentID=0, Status='live', Symbol='LOC105375271', Aliases='', description='uncharacterized LOC105375271', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48111246, end_position_on_the_genomic_accession=48114370, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=23524, CurrentID=0, Status='live', Symbol='SRRM2', Aliases='300-KD, CWF21, Cwc21, HSPC075, SRL300, SRm300', description='serine/arginine repetitive matrix 2', other_designations='300 kDa nuclear matrix antigen|RNA binding protein|SR-related nuclear matrix protein of 300 kDa|ser/Arg-related nuclear matrix protein of 300 kDa|serine/arginine-rich splicing factor-related nuclear matrix protein of 300 kDa|splicing coactivator subunit SRm300|tax-responsive enhancer element-binding protein 803|taxREB803', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2752329, end_position_on_the_genomic_accession=2771412, orientation='plus', exon_count=18, OMIM=606032, no_hdr0=''),
  105375273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375273, CurrentID=0, Status='live', Symbol='LOC105375273', Aliases='', description='uncharacterized LOC105375273', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=48708469, end_position_on_the_genomic_accession=48711484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375275, CurrentID=0, Status='live', Symbol='LOC105375275', Aliases='', description='uncharacterized LOC105375275', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=50382723, end_position_on_the_genomic_accession=50399390, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124460, CurrentID=0, Status='live', Symbol='SNX20', Aliases='SLIC1', description='sorting nexin 20', other_designations='selectin ligand interactor cytoplasmic 1|selectin ligand interactor cytoplasmic-1|selectin ligand-interactor cytoplasmic 1', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50666300, end_position_on_the_genomic_accession=50681353, orientation='minus', exon_count=5, OMIM=613281, no_hdr0=''),
  105375277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375277, CurrentID=0, Status='live', Symbol='LOC105375277', Aliases='', description='uncharacterized LOC105375277', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51544219, end_position_on_the_genomic_accession=51579388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375278, CurrentID=0, Status='live', Symbol='LOC105375278', Aliases='', description='uncharacterized LOC105375278', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=51707969, end_position_on_the_genomic_accession=51717668, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132399, CurrentID=0, Status='live', Symbol='GAGE12D', Aliases='GAGE-12B', description='G antigen 12D', other_designations='GAGE-12D|g antigen 12C/D/E', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49541722, end_position_on_the_genomic_accession=49549107, orientation='plus', exon_count=5, OMIM=300728, no_hdr0=''),
  101180976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101180976, CurrentID=0, Status='live', Symbol='IFNL4', Aliases='IFNAN', description='interferon, lambda 4 (gene/pseudogene)', other_designations='IFN-lambda-4', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39246314, end_position_on_the_genomic_accession=39248856, orientation='minus', exon_count=5, OMIM=615090, no_hdr0=''),
  101929224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929224, CurrentID=0, Status='live', Symbol='LOC101929224', Aliases='', description='uncharacterized LOC101929224', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=200669507, end_position_on_the_genomic_accession=200694250, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132402, CurrentID=0, Status='live', Symbol='PGGT1BP2', Aliases='', description='protein geranylgeranyltransferase type I, beta subunit pseudogene 2', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79892151, end_position_on_the_genomic_accession=79893175, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132403, CurrentID=0, Status='live', Symbol='FAM157B', Aliases='', description='family with sequence similarity 157 member B', other_designations='family with sequence similarity 157, member B', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138215351, end_position_on_the_genomic_accession=138252775, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105375284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375284, CurrentID=0, Status='live', Symbol='LOC105375284', Aliases='', description='uncharacterized LOC105375284', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=54993024, end_position_on_the_genomic_accession=54995619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375285, CurrentID=0, Status='live', Symbol='LOC105375285', Aliases='', description='uncharacterized LOC105375285', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55366329, end_position_on_the_genomic_accession=55368010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132406, CurrentID=0, Status='live', Symbol='NBPF10', Aliases='AB6, AG1, NBPF9', description='neuroblastoma breakpoint family member 10', other_designations='neuroblastoma breakpoint family, member 10', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146066318, end_position_on_the_genomic_accession=146144804, orientation='minus', exon_count=90, OMIM=614000, no_hdr0=''),
  105375287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375287, CurrentID=0, Status='live', Symbol='LOC105375287', Aliases='', description='uncharacterized LOC105375287', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55587731, end_position_on_the_genomic_accession=55596180, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  386616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386616, CurrentID=0, Status='live', Symbol='SCKL3', Aliases='', description='Seckel syndrome 3', other_designations='', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  386617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386617, CurrentID=0, Status='live', Symbol='KCTD8', Aliases='', description='potassium channel tetramerization domain containing 8', other_designations='potassium channel tetramerisation domain containing 8', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=44173903, end_position_on_the_genomic_accession=44448885, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  386618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386618, CurrentID=0, Status='live', Symbol='KCTD4', Aliases='bA321C24.3', description='potassium channel tetramerization domain containing 4', other_designations='potassium channel tetramerisation domain containing 4', map_location='13q14.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45192853, end_position_on_the_genomic_accession=45211102, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  401316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401316, CurrentID=0, Status='live', Symbol='EIF4HP1', Aliases='', description='eukaryotic translation initiation factor 4H pseudogene 1', other_designations='Williams-Beuren syndrome chromosome region 1 pseudogene', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27456372, end_position_on_the_genomic_accession=27458857, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375292, CurrentID=0, Status='live', Symbol='LOC105375292', Aliases='', description='uncharacterized LOC105375292', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56381765, end_position_on_the_genomic_accession=56388488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132413, CurrentID=0, Status='live', Symbol='GKN3P', Aliases='', description='gastrokine 3, pseudogene', other_designations='gastrokine-3', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68922505, end_position_on_the_genomic_accession=68926834, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375294, CurrentID=0, Status='live', Symbol='LOC105375294', Aliases='', description='uncharacterized LOC105375294', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56836225, end_position_on_the_genomic_accession=56849683, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132415, CurrentID=0, Status='live', Symbol='SOCS5P4', Aliases='', description='suppressor of cytokine signaling 5 pseudogene 4', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71043214, end_position_on_the_genomic_accession=71045470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375296, CurrentID=0, Status='live', Symbol='LOC105375296', Aliases='', description='uncharacterized LOC105375296', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57140610, end_position_on_the_genomic_accession=57148387, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132417, CurrentID=0, Status='live', Symbol='FCGR1C', Aliases='CD64c, FCRIC, IGFR1, IGFRC', description='Fc fragment of IgG, high affinity Ic, receptor (CD64), pseudogene', other_designations='Fc fragment of IgG, high affinity Ic, receptor for (CD64)|immunoglobulin gamma Fc receptor 1C', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143874743, end_position_on_the_genomic_accession=143883733, orientation='plus', exon_count=6, OMIM=601503, no_hdr0=''),
  100132418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132418, CurrentID=0, Status='live', Symbol='LOC100132418', Aliases='', description='Uncharacterized PRO1102-like protein LOC100132418', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.9', start_position_on_the_genomic_accession=51402263, end_position_on_the_genomic_accession=51404617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  386627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386627, CurrentID=0, Status='live', Symbol='SAP30L-AS1', Aliases='GALNT10-AS1', description='SAP30L antisense RNA 1 (head to head)', other_designations='CTB-157D17.1|GALNT10 antisense RNA 1', map_location='5q33.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=154389769, end_position_on_the_genomic_accession=154445822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132420, CurrentID=0, Status='live', Symbol='ANKRD36P1', Aliases='', description='ankyrin repeat domain 36 pseudogene 1', other_designations='', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26409520, end_position_on_the_genomic_accession=26439003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132421, CurrentID=0, Status='live', Symbol='DUX4L18', Aliases='DUXY2', description='double homeobox 4 like 18', other_designations='', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=11321558, end_position_on_the_genomic_accession=11322983, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375302, CurrentID=0, Status='live', Symbol='LOC105375302', Aliases='', description='protein FAM182A-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=57770209, end_position_on_the_genomic_accession=57840981, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375303, CurrentID=0, Status='live', Symbol='LOC105375303', Aliases='', description='uncharacterized LOC105375303', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1616880, end_position_on_the_genomic_accession=1642739, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375304, CurrentID=0, Status='live', Symbol='LOC105375304', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=26372438, end_position_on_the_genomic_accession=26376794, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375305, CurrentID=0, Status='live', Symbol='LOC105375305', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=39948746, end_position_on_the_genomic_accession=39950638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375306, CurrentID=0, Status='live', Symbol='LOC105375306', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56829776, end_position_on_the_genomic_accession=56832244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  124491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124491, CurrentID=0, Status='live', Symbol='TMEM170A', Aliases='TMEM170', description='transmembrane protein 170A', other_designations='transmembrane protein 170', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75443244, end_position_on_the_genomic_accession=75464737, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375308, CurrentID=0, Status='live', Symbol='LOC105375308', Aliases='', description='uncharacterized LOC105375308', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375309, CurrentID=0, Status='live', Symbol='LOC105375309', Aliases='', description='uncharacterized LOC105375309', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375310, CurrentID=0, Status='live', Symbol='LOC105375310', Aliases='', description='uncharacterized LOC105375310', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375311, CurrentID=0, Status='live', Symbol='LOC105375311', Aliases='', description='uncharacterized LOC105375311', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  124496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124496, CurrentID=0, Status='live', Symbol='ATP5F1P7', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit B1 pseudogene 7', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1 pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit B1 pseudogene 7', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=75709059, end_position_on_the_genomic_accession=75709815, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375313, CurrentID=0, Status='live', Symbol='LOC105375313', Aliases='', description='uncharacterized LOC105375313', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63699206, end_position_on_the_genomic_accession=63707105, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375314, CurrentID=0, Status='live', Symbol='LOC105375314', Aliases='', description='uncharacterized LOC105375314', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63714587, end_position_on_the_genomic_accession=63718963, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375315, CurrentID=0, Status='live', Symbol='LOC105375315', Aliases='', description='uncharacterized LOC105375315', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63727306, end_position_on_the_genomic_accession=63739638, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375316, CurrentID=0, Status='live', Symbol='LOC105375316', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63765655, end_position_on_the_genomic_accession=63767001, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375317, CurrentID=0, Status='live', Symbol='LOC105375317', Aliases='', description='uncharacterized LOC105375317', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64032106, end_position_on_the_genomic_accession=64039611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375318, CurrentID=0, Status='live', Symbol='LOC105375318', Aliases='', description='uncharacterized LOC105375318', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63925903, end_position_on_the_genomic_accession=63966074, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  648791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648791, CurrentID=0, Status='live', Symbol='PPP1R3G', Aliases='', description='protein phosphatase 1 regulatory subunit 3G', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 3G|protein phosphatase 1, regulatory subunit 3G|putative protein phosphatase 1 regulatory inhibitor subunit 3G', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5085486, end_position_on_the_genomic_accession=5087221, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375320, CurrentID=0, Status='live', Symbol='LOC105375320', Aliases='', description='uncharacterized LOC105375320', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64381838, end_position_on_the_genomic_accession=64401995, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132441, CurrentID=0, Status='live', Symbol='GXYLT1P4', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 4', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66026330, end_position_on_the_genomic_accession=66027784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375322, CurrentID=0, Status='live', Symbol='LOC105375322', Aliases='', description='uncharacterized LOC105375322', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64294505, end_position_on_the_genomic_accession=64308015, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105375323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375323, CurrentID=0, Status='live', Symbol='LOC105375323', Aliases='', description='uncharacterized LOC105375323', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64287420, end_position_on_the_genomic_accession=64292262, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375324, CurrentID=0, Status='live', Symbol='LOC105375324', Aliases='', description='uncharacterized LOC105375324', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64402039, end_position_on_the_genomic_accession=64412499, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  386653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386653, CurrentID=0, Status='live', Symbol='IL31', Aliases='IL-31', description='interleukin 31', other_designations='', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=122172030, end_position_on_the_genomic_accession=122182963, orientation='minus', exon_count=4, OMIM=609509, no_hdr0=''),
  386654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386654, CurrentID=0, Status='live', Symbol='OSTF1P1', Aliases='OSTF1P', description='osteoclast stimulating factor 1 pseudogene 1', other_designations='OSTF1 pseudogene', map_location='12q24.1-q24.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=114588051, end_position_on_the_genomic_accession=114588799, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375327, CurrentID=0, Status='live', Symbol='LOC105375327', Aliases='', description='uncharacterized LOC105375327', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64631046, end_position_on_the_genomic_accession=64634081, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124512, CurrentID=0, Status='live', Symbol='METTL23', Aliases='C17orf95, MRT44', description='methyltransferase like 23', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76725873, end_position_on_the_genomic_accession=76733881, orientation='plus', exon_count=9, OMIM=615262, no_hdr0=''),
  105375329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375329, CurrentID=0, Status='live', Symbol='LOC105375329', Aliases='', description='uncharacterized LOC105375329', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65026794, end_position_on_the_genomic_accession=65029426, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132450, CurrentID=0, Status='live', Symbol='IQSEC3P1', Aliases='', description='IQ motif and Sec7 domain 3 pseudogene 1', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=2752597, end_position_on_the_genomic_accession=2753132, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100271539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271539, CurrentID=0, Status='live', Symbol='RPL7AP61', Aliases='RPL7A_25_1330', description='ribosomal protein L7a pseudogene 61', other_designations='', map_location='13q32.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=97901652, end_position_on_the_genomic_accession=97902253, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375332, CurrentID=0, Status='live', Symbol='LOC105375332', Aliases='', description='uncharacterized LOC105375332', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65072097, end_position_on_the_genomic_accession=65073445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375333, CurrentID=0, Status='live', Symbol='LOC105375333', Aliases='', description='uncharacterized LOC105375333', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65531315, end_position_on_the_genomic_accession=65553682, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375334, CurrentID=0, Status='live', Symbol='LOC105375334', Aliases='', description='uncharacterized LOC105375334', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=65303555, end_position_on_the_genomic_accession=65338614, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100423029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423029, CurrentID=0, Status='live', Symbol='MIR3163', Aliases='', description='microRNA 3163', other_designations='hsa-mir-3163', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66934434, end_position_on_the_genomic_accession=66934506, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132456, CurrentID=0, Status='live', Symbol='GUSBP7', Aliases='', description='glucuronidase, beta pseudogene 7', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100066239, end_position_on_the_genomic_accession=100066417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  648809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648809, CurrentID=0, Status='live', Symbol='EFTUD1P1', Aliases='FAM42B, HsT19321', description='elongation factor Tu GTP binding domain containing 1 pseudogene 1', other_designations='elongation factor Tu GTP-binding domain-containing protein 1 pseudogene|family with sequence similarity 42, member B', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84080187, end_position_on_the_genomic_accession=84126601, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375338, CurrentID=0, Status='live', Symbol='LOC105375338', Aliases='', description='uncharacterized LOC105375338', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67229939, end_position_on_the_genomic_accession=67236948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375339, CurrentID=0, Status='live', Symbol='LOC105375339', Aliases='', description='uncharacterized LOC105375339', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=66958301, end_position_on_the_genomic_accession=66960925, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375340, CurrentID=0, Status='live', Symbol='LOC105375340', Aliases='', description='uncharacterized LOC105375340', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=67595651, end_position_on_the_genomic_accession=67604480, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375341, CurrentID=0, Status='live', Symbol='LOC105375341', Aliases='', description='uncharacterized LOC105375341', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=68155126, end_position_on_the_genomic_accession=68330391, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105375342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375342, CurrentID=0, Status='live', Symbol='LOC105375342', Aliases='', description='uncharacterized LOC105375342', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69040178, end_position_on_the_genomic_accession=69046781, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132463, CurrentID=0, Status='live', Symbol='CLDN24', Aliases='CLDN21', description='claudin 24', other_designations='claudin 21', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183321764, end_position_on_the_genomic_accession=183322426, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  386672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386672, CurrentID=0, Status='live', Symbol='KRTAP10-4', Aliases='KAP10.4, KAP18-4, KRTAP10.4, KRTAP18-4, KRTAP18.4', description='keratin associated protein 10-4', other_designations='high sulfur keratin-associated protein 10.4|keratin-associated protein 18.4', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44573729, end_position_on_the_genomic_accession=44575371, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375345, CurrentID=0, Status='live', Symbol='LOC105375345', Aliases='', description='uncharacterized LOC105375345', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=69283014, end_position_on_the_genomic_accession=69304166, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  386674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386674, CurrentID=0, Status='live', Symbol='KRTAP10-6', Aliases='KAP10.6, KAP18.6, KRTAP18-6, KRTAP18.6', description='keratin associated protein 10-6', other_designations='high sulfur keratin-associated protein 10.6|keratin associated protein 18-6', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44591268, end_position_on_the_genomic_accession=44592505, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  386675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386675, CurrentID=0, Status='live', Symbol='KRTAP10-7', Aliases='KAP10.7, KAP18.7, KRTAP18-7', description='keratin associated protein 10-7', other_designations='high sulfur keratin-associated protein 10.7|keratin associated protein 18-7', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44600597, end_position_on_the_genomic_accession=44602174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  386676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386676, CurrentID=0, Status='live', Symbol='KRTAP10-9', Aliases='KAP10.9, KAP18.9, KRTAP18-9', description='keratin associated protein 10-9', other_designations='high sulfur keratin-associated protein 10.9|keratin associated protein 18-9|keratin-associated protein 10.9|keratin-associated protein 18-9|keratin-associated protein 18.9', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44627123, end_position_on_the_genomic_accession=44628378, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  386677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386677, CurrentID=0, Status='live', Symbol='KRTAP10-1', Aliases='KAP10.1, KAP18-1, KAP18.1, KRTAP10.1, KRTAP18-1, KRTAP18.1', description='keratin associated protein 10-1', other_designations='high sulfur keratin-associated protein 10.1|keratin-associated protein 10.1|keratin-associated protein 18-1|keratin-associated protein 18.1', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44539185, end_position_on_the_genomic_accession=44540195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  386678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386678, CurrentID=0, Status='live', Symbol='KRTAP10-11', Aliases='KAP10.11, KAP18.11, KRTAP18-11, KRTAP18.11', description='keratin associated protein 10-11', other_designations='high sulfur keratin-associated protein 10.11|keratin associated protein 18-11|keratin-associated protein 10.11|keratin-associated protein 18-11|keratin-associated protein 18.11', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44646414, end_position_on_the_genomic_accession=44647650, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  124535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124535, CurrentID=0, Status='live', Symbol='HSF5', Aliases='HSF 5, HSTF 5', description='heat shock transcription factor family member 5', other_designations='', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58420167, end_position_on_the_genomic_accession=58488401, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  386680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386680, CurrentID=0, Status='live', Symbol='KRTAP10-5', Aliases='KAP10.5, KAP18-5, KAP18.5, KRTAP10.5, KRTAP18-5, KRTAP18.1, KRTAP18.5', description='keratin associated protein 10-5', other_designations='high sulfur keratin-associated protein 10.5|keratin-associated protein 18.5', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44579455, end_position_on_the_genomic_accession=44580604, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  386681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386681, CurrentID=0, Status='live', Symbol='KRTAP10-8', Aliases='KAP10.8, KRTAP18-8, KRTAP18.8', description='keratin associated protein 10-8', other_designations='high sulfur keratin-associated protein 10.8|keratin associated protein 18-8|keratin-associated protein 10.8|keratin-associated protein 18-8|keratin-associated protein 18.8', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44612079, end_position_on_the_genomic_accession=44612954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  124538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124538, CurrentID=0, Status='live', Symbol='OR4D2', Aliases='BC2009, OR17-24', description='olfactory receptor family 4 subfamily D member 2', other_designations='B-lymphocyte membrane protein BC2009|G-protein coupled membrane protein|olfactory receptor OR17-24|olfactory receptor, family 4, subfamily D, member 2', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58169656, end_position_on_the_genomic_accession=58170579, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  386683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386683, CurrentID=0, Status='live', Symbol='KRTAP12-3', Aliases='KRTAP12.3', description='keratin associated protein 12-3', other_designations='high sulfur keratin-associated protein 12.3|keratin-associated protein 12.3', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44657932, end_position_on_the_genomic_accession=44658341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  124540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124540, CurrentID=0, Status='live', Symbol='MSI2', Aliases='MSI2H', description='musashi RNA binding protein 2', other_designations='musashi RNA-binding protein 2|musashi homolog 2|musashi-2', map_location='17q22', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57256570, end_position_on_the_genomic_accession=57700477, orientation='plus', exon_count=19, OMIM=607897, no_hdr0=''),
  386685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386685, CurrentID=0, Status='live', Symbol='KRTAP10-12', Aliases='KAP10.12, KRTAP18-12, KRTAP18.12', description='keratin associated protein 10-12', other_designations='high sulfur keratin-associated protein 10.12|keratin-associated protein 18-12', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44697172, end_position_on_the_genomic_accession=44698044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375358, CurrentID=0, Status='live', Symbol='LOC105375358', Aliases='', description='uncharacterized LOC105375358', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=76902506, end_position_on_the_genomic_accession=76919127, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  386687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386687, CurrentID=0, Status='live', Symbol='OFD1P3Y', Aliases='OFD1P19, OFD1P19Y, OFD1P3, OFD1PY19, OFD1PY3, OFDYP3', description='OFD1 pseudogene 3, Y-linked', other_designations='OFD1 pseudogene 19, Y-linked|OFD1 pseudogene, Y-linked 19|OFD1 pseudogene, Y-linked 3|oral-facial-digital syndrome 1 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9034110, end_position_on_the_genomic_accession=9040188, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386688, CurrentID=0, Status='live', Symbol='OFD1P4Y', Aliases='OFD1P4, OFD1PY4, OFDYP4', description='OFD1 pseudogene 4, Y-linked', other_designations='OFD1 pseudogene, Y-linked 4|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18453050, end_position_on_the_genomic_accession=18469259, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386689, CurrentID=0, Status='live', Symbol='OFD1P5Y', Aliases='OFD1P5, OFD1PY5, OFDYP5', description='OFD1 pseudogene 5, Y-linked', other_designations='OFD1 pseudogene, Y-linked 5|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18581963, end_position_on_the_genomic_accession=18629179, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386690, CurrentID=0, Status='live', Symbol='OFD1P6Y', Aliases='OFD1P2, OFD1P6, OFD1PY6, OFDYP6', description='OFD1 pseudogene 6, Y-linked', other_designations='OFD1 pseudogene 2|OFD1 pseudogene, Y-linked 6|oral-facial-digital syndrome 1 pseudogene|oral-facial-digital syndrome 1, pseudogene 2', map_location='Yq11.22', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18673727, end_position_on_the_genomic_accession=18738440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386691, CurrentID=0, Status='live', Symbol='OFD1P7Y', Aliases='OFD1P7, OFD1PY7, OFDYP7', description='OFD1 pseudogene 7, Y-linked', other_designations='OFD1 pseudogene, Y-linked 7|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18848162, end_position_on_the_genomic_accession=18867381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386692, CurrentID=0, Status='live', Symbol='OFD1P8Y', Aliases='OFD1P8, OFD1PY8, OFDYP8', description='OFD1 pseudogene 8, Y-linked', other_designations='OFD1 pseudogene, Y-linked 8|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21972211, end_position_on_the_genomic_accession=22005684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386693, CurrentID=0, Status='live', Symbol='OFD1P9Y', Aliases='OFD1P9, OFD1PY9, OFDYP9', description='OFD1 pseudogene 9, Y-linked', other_designations='OFD1 pseudogene, Y-linked 9|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22580707, end_position_on_the_genomic_accession=22614183, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386694, CurrentID=0, Status='live', Symbol='OFD1P10Y', Aliases='OFD1P10, OFD1PY10, OFDYP10', description='OFD1 pseudogene 10, Y-linked', other_designations='OFD1 pseudogene, Y-linked 10|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23581595, end_position_on_the_genomic_accession=23614898, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386695, CurrentID=0, Status='live', Symbol='OFD1P11Y', Aliases='OFD1P11, OFD1PY11, OFDYP11', description='OFD1 pseudogene 11, Y-linked', other_designations='OFD1 pseudogene, Y-linked 11|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23770823, end_position_on_the_genomic_accession=23798250, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386696, CurrentID=0, Status='live', Symbol='OFD1P12Y', Aliases='OFD1P12, OFD1PY12, OFDYP12', description='OFD1 pseudogene 12, Y-linked', other_designations='OFD1 pseudogene, Y-linked 12|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23983300, end_position_on_the_genomic_accession=23995181, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386697, CurrentID=0, Status='live', Symbol='OFD1P13Y', Aliases='OFD1P13, OFD1PY13, OFDYP13', description='OFD1 pseudogene 13, Y-linked', other_designations='OFD1 pseudogene, Y-linked 13|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25674995, end_position_on_the_genomic_accession=25697426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  255626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255626, CurrentID=0, Status='live', Symbol='HIST1H2BA', Aliases='H2BFU, STBP, TSH2B, TSH2B.1, bA317E16.3', description='histone cluster 1, H2ba', other_designations='H2B histone family, member U, (testis-specific)|histone 1, H2ba|histone H2B, testis|testis-specific histone H2B', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25726909, end_position_on_the_genomic_accession=25727345, orientation='plus', exon_count=1, OMIM=609904, no_hdr0=''),
  386699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386699, CurrentID=0, Status='live', Symbol='OFD1P15Y', Aliases='OFD1P15, OFD1PY15, OFDYP15', description='OFD1 pseudogene 15, Y-linked', other_designations='OFD1 pseudogene, Y-linked 15|oral-facial-digital syndrome 1 pseudogene', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26069665, end_position_on_the_genomic_accession=26088595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375372, CurrentID=0, Status='live', Symbol='LOC105375372', Aliases='', description='uncharacterized LOC105375372', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=79948936, end_position_on_the_genomic_accession=79957033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375373, CurrentID=0, Status='live', Symbol='LOC105375373', Aliases='', description='uncharacterized LOC105375373', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=80330291, end_position_on_the_genomic_accession=80334950, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375374, CurrentID=0, Status='live', Symbol='LOC105375374', Aliases='', description='uncharacterized LOC105375374', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=82341205, end_position_on_the_genomic_accession=82354347, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255631, CurrentID=0, Status='live', Symbol='COL24A1', Aliases='', description='collagen, type XXIV, alpha 1', other_designations='', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85729233, end_position_on_the_genomic_accession=86156791, orientation='minus', exon_count=61, OMIM=610025, no_hdr0=''),
  100132496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132496, CurrentID=0, Status='live', Symbol='RPS24P9', Aliases='RPS24_5_400', description='ribosomal protein S24 pseudogene 9', other_designations='', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125787648, end_position_on_the_genomic_accession=125788250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375377, CurrentID=0, Status='live', Symbol='LOC105375377', Aliases='', description='uncharacterized LOC105375377', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83168155, end_position_on_the_genomic_accession=83175460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375378, CurrentID=0, Status='live', Symbol='LOC105375378', Aliases='', description='uncharacterized LOC105375378', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=83471781, end_position_on_the_genomic_accession=83482249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132499, CurrentID=0, Status='live', Symbol='RPS29P16', Aliases='RPS29_11_830', description='ribosomal protein S29 pseudogene 16', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103348571, end_position_on_the_genomic_accession=103348859, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375380, CurrentID=0, Status='live', Symbol='LINC00972', Aliases='', description='long intergenic non-protein coding RNA 972', other_designations='', map_location='7q21.11', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=85475227, end_position_on_the_genomic_accession=85489338, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124565, CurrentID=0, Status='live', Symbol='SLC38A10', Aliases='PP1744', description='solute carrier family 38 member 10', other_designations='solute carrier family 38, member 10', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81244811, end_position_on_the_genomic_accession=81295433, orientation='minus', exon_count=18, OMIM=616525, no_hdr0=''),
  105375382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375382, CurrentID=0, Status='live', Symbol='LOC105375382', Aliases='', description='uncharacterized LOC105375382', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=86773314, end_position_on_the_genomic_accession=86795187, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375383, CurrentID=0, Status='live', Symbol='LOC105375383', Aliases='', description='uncharacterized LOC105375383', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87058643, end_position_on_the_genomic_accession=87111964, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132504, CurrentID=0, Status='live', Symbol='TSPY19P', Aliases='', description='testis specific protein, Y-linked 19, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6304302, end_position_on_the_genomic_accession=6305349, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375385, CurrentID=0, Status='live', Symbol='LOC105375385', Aliases='', description='uncharacterized LOC105375385', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=87888066, end_position_on_the_genomic_accession=87926669, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132506, CurrentID=0, Status='live', Symbol='FAM197Y3', Aliases='', description='family with sequence similarity 197 Y-linked member 3', other_designations='family with sequence similarity 197, Y-linked, member 3', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=9340620, end_position_on_the_genomic_accession=9342813, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375387, CurrentID=0, Status='live', Symbol='LOC105375387', Aliases='', description='uncharacterized LOC105375387', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=89443974, end_position_on_the_genomic_accession=89497132, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375388, CurrentID=0, Status='live', Symbol='LOC105375388', Aliases='', description='uncharacterized LOC105375388', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90289096, end_position_on_the_genomic_accession=90294895, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132509, CurrentID=0, Status='live', Symbol='ZNF114P1', Aliases='', description='zinc finger protein 114 pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13838730, end_position_on_the_genomic_accession=13839795, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132510, CurrentID=0, Status='live', Symbol='GLRXP3', Aliases='GLRXL', description='glutaredoxin (thioltransferase) pseudogene 3', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161750456, end_position_on_the_genomic_accession=161751837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375391, CurrentID=0, Status='live', Symbol='LOC105375391', Aliases='', description='uncharacterized LOC105375391', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91287754, end_position_on_the_genomic_accession=91292339, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375392, CurrentID=0, Status='live', Symbol='LOC105375392', Aliases='', description='uncharacterized LOC105375392', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91311311, end_position_on_the_genomic_accession=91316055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255649, CurrentID=0, Status='live', Symbol='OOSP1', Aliases='', description='oocyte secreted protein 1, pseudogene', other_designations='oocyte-secreted protein 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59942879, end_position_on_the_genomic_accession=59995845, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105375394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375394, CurrentID=0, Status='live', Symbol='LOC105375394', Aliases='', description='uncharacterized LOC105375394', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=91496562, end_position_on_the_genomic_accession=91515322, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375395, CurrentID=0, Status='live', Symbol='LOC105375395', Aliases='', description='uncharacterized LOC105375395', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92319932, end_position_on_the_genomic_accession=92323807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  386724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386724, CurrentID=0, Status='live', Symbol='AMIGO3', Aliases='ALI3, AMIGO-3', description='adhesion molecule with Ig-like domain 3', other_designations='alivin-3|amphoterin-induced gene and ORF 3', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49716834, end_position_on_the_genomic_accession=49719805, orientation='minus', exon_count=1, OMIM=615691, no_hdr0=''),
  386725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386725, CurrentID=0, Status='live', Symbol='FAM8A7P', Aliases='', description='family with sequence similarity 8 member A7, pseudogene', other_designations='family with sequence similarity 8, member A7 pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8177826, end_position_on_the_genomic_accession=8179167, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386726, CurrentID=0, Status='live', Symbol='FAM8A4P', Aliases='FAM8A8P', description='family with sequence similarity 8 member A4, pseudogene', other_designations='family with sequence similarity 8, member A4 pseudogene|family with sequence similarity 8, member A8 pseudogene', map_location='Yq11.21', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12321481, end_position_on_the_genomic_accession=12323357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  124583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124583, CurrentID=0, Status='live', Symbol='CANT1', Aliases='DBQD, SCAN-1, SCAN1, SHAPY', description='calcium activated nucleotidase 1', other_designations='Ca2+-dependent endoplasmic reticulum nucleoside diphosphatase|apyrase homolog|micromelic dwarfism with vertebral and metaphyseal abnormalities and advanced carpotarsal ossification|putative MAPK-activating protein PM09|putative NF-kappa-B-activating protein 107|soluble Ca-activated nucleotidase, isozyme 1|soluble calcium-activated nucleotidase SCAN-1', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78991716, end_position_on_the_genomic_accession=79009817, orientation='minus', exon_count=8, OMIM=613165, no_hdr0=''),
  386728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386728, CurrentID=0, Status='live', Symbol='FAM8A10P', Aliases='', description='family with sequence similarity 8 member A10, pseudogene', other_designations='family with sequence similarity 8, member A10 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22726008, end_position_on_the_genomic_accession=22729254, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386729, CurrentID=0, Status='live', Symbol='CDY4P', Aliases='', description='chromodomain protein, Y-linked 4 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12801341, end_position_on_the_genomic_accession=12804258, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386730, CurrentID=0, Status='live', Symbol='CDY6P', Aliases='', description='chromodomain protein, Y-linked 6 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17881998, end_position_on_the_genomic_accession=17884134, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386731, CurrentID=0, Status='live', Symbol='CDY7P', Aliases='', description='chromodomain protein, Y-linked 7 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17951268, end_position_on_the_genomic_accession=17953564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386732, CurrentID=0, Status='live', Symbol='CDY8P', Aliases='', description='chromodomain protein, Y-linked 8 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18021873, end_position_on_the_genomic_accession=18024009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375405, CurrentID=0, Status='live', Symbol='LOC105375405', Aliases='', description='uncharacterized LOC105375405', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95035452, end_position_on_the_genomic_accession=95214946, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  124590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124590, CurrentID=0, Status='live', Symbol='USH1G', Aliases='ANKS4A, SANS', description='Usher syndrome 1G (autosomal recessive)', other_designations='scaffold protein containing ankyrin repeats and SAM domain', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74916083, end_position_on_the_genomic_accession=74923263, orientation='minus', exon_count=4, OMIM=607696, no_hdr0=''),
  386735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386735, CurrentID=0, Status='live', Symbol='CDY13P', Aliases='', description='chromodomain protein, Y-linked 13 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=22520588, end_position_on_the_genomic_accession=22521796, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386736, CurrentID=0, Status='live', Symbol='CDY14P', Aliases='', description='chromodomain protein, Y-linked 14 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23674278, end_position_on_the_genomic_accession=23675485, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386737, CurrentID=0, Status='live', Symbol='CDY15P', Aliases='', description='chromodomain protein, Y-linked 15 pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23854271, end_position_on_the_genomic_accession=23857512, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386738, CurrentID=0, Status='live', Symbol='CDY16P', Aliases='', description='chromodomain protein, Y-linked 16 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23861470, end_position_on_the_genomic_accession=23862800, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386739, CurrentID=0, Status='live', Symbol='CDY17P', Aliases='', description='chromodomain protein, Y-linked 17 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24049777, end_position_on_the_genomic_accession=24051913, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386740, CurrentID=0, Status='live', Symbol='CDY18P', Aliases='', description='chromodomain protein, Y-linked 18 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24103786, end_position_on_the_genomic_accession=24107233, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386741, CurrentID=0, Status='live', Symbol='CDY19P', Aliases='', description='chromodomain protein, Y-linked 19 pseudogene', other_designations='', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25562907, end_position_on_the_genomic_accession=25566354, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386742, CurrentID=0, Status='live', Symbol='CDY20P', Aliases='', description='chromodomain protein, Y-linked 20 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25618218, end_position_on_the_genomic_accession=25620354, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  124599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124599, CurrentID=0, Status='live', Symbol='CD300LB', Aliases='CD300b, CLM-7, CLM7, CMRF35-A2, IREM-3, IREM3, TREM-5, TREM5', description='CD300 molecule like family member b', other_designations='CD300 antigen like family member B|CD300 molecule-like family member b|immune receptor expressed on myeloid cells 3|leukocyte mono-Ig-like receptor 5|triggering receptor expressed on myeloid cells 5', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74521174, end_position_on_the_genomic_accession=74531474, orientation='minus', exon_count=4, OMIM=610705, no_hdr0=''),
  386744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386744, CurrentID=0, Status='live', Symbol='CDY22P', Aliases='', description='chromodomain protein, Y-linked 22 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25812592, end_position_on_the_genomic_accession=25815833, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  386745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386745, CurrentID=0, Status='live', Symbol='CDY23P', Aliases='', description='chromodomain protein, Y-linked 23 pseudogene', other_designations='chromodomain protein, Y-linked, 1 pseudogene', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=25994591, end_position_on_the_genomic_accession=25995798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  386746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386746, CurrentID=0, Status='live', Symbol='MRGPRG', Aliases='GPR169, MRGG', description='MAS related GPR family member G', other_designations='G protein-coupled receptor 169|G protein-coupled receptor MRGG|MAS-related GPR, member G', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3217944, end_position_on_the_genomic_accession=3218813, orientation='minus', exon_count=1, OMIM=607234, no_hdr0=''),
  105375419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375419, CurrentID=0, Status='live', Symbol='LOC105375419', Aliases='', description='uncharacterized LOC105375419', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98652639, end_position_on_the_genomic_accession=98657602, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375420, CurrentID=0, Status='live', Symbol='LOC105375420', Aliases='', description='uncharacterized LOC105375420', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99047889, end_position_on_the_genomic_accession=99081068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132541, CurrentID=0, Status='live', Symbol='GXYLT1P3', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 3', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40348499, end_position_on_the_genomic_accession=40353658, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375422, CurrentID=0, Status='live', Symbol='LOC105375422', Aliases='', description='uncharacterized LOC105375422', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99552152, end_position_on_the_genomic_accession=99554822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375423, CurrentID=0, Status='live', Symbol='LOC105375423', Aliases='', description='uncharacterized LOC105375423', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99999451, end_position_on_the_genomic_accession=100011750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375424, CurrentID=0, Status='live', Symbol='LOC105375424', Aliases='', description='uncharacterized LOC105375424', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100108484, end_position_on_the_genomic_accession=100109898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375425, CurrentID=0, Status='live', Symbol='LOC105375425', Aliases='', description='uncharacterized LOC105375425', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100130671, end_position_on_the_genomic_accession=100143647, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375426, CurrentID=0, Status='live', Symbol='LOC105375426', Aliases='', description='uncharacterized LOC105375426', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100264867, end_position_on_the_genomic_accession=100267483, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132547, CurrentID=0, Status='live', Symbol='RPL26P33', Aliases='RPL26_16_1236', description='ribosomal protein L26 pseudogene 33', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=50365599, end_position_on_the_genomic_accession=50366130, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100874219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874219, CurrentID=0, Status='live', Symbol='TMEM212-AS1', Aliases='', description='TMEM212 antisense RNA 1', other_designations='TMEM212 antisense RNA 1 (non-protein coding)', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171876352, end_position_on_the_genomic_accession=171900740, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  386757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386757, CurrentID=0, Status='live', Symbol='SLC6A10P', Aliases='CT-2, CT2, SLC6A10, SLC6A10pA', description='solute carrier family 6 (neurotransmitter transporter), member 10, pseudogene', other_designations='creatine transporter-2|solute carrier family 6 (neurotransmitter transporter, creatine), member 10, pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32877476, end_position_on_the_genomic_accession=32885142, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  386758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386758, CurrentID=0, Status='live', Symbol='ZNF582-AS1', Aliases='', description='ZNF582 antisense RNA 1 (head to head)', other_designations='ZNF582 antisense RNA 1 (non-protein coding)', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56393676, end_position_on_the_genomic_accession=56399175, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375431, CurrentID=0, Status='live', Symbol='LOC105375431', Aliases='', description='uncharacterized LOC105375431', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100942059, end_position_on_the_genomic_accession=100964228, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375432, CurrentID=0, Status='live', Symbol='LOC105375432', Aliases='', description='uncharacterized LOC105375432', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102240549, end_position_on_the_genomic_accession=102247538, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375433, CurrentID=0, Status='live', Symbol='LOC105375433', Aliases='', description='uncharacterized LOC105375433', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102407507, end_position_on_the_genomic_accession=102421280, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375434, CurrentID=0, Status='live', Symbol='LOC105375434', Aliases='', description='uncharacterized LOC105375434', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103149735, end_position_on_the_genomic_accession=103200285, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375435, CurrentID=0, Status='live', Symbol='LOC105375435', Aliases='', description='uncharacterized LOC105375435', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=103534590, end_position_on_the_genomic_accession=103539064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375436, CurrentID=0, Status='live', Symbol='LOC105375436', Aliases='', description='uncharacterized LOC105375436', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=104738551, end_position_on_the_genomic_accession=104753116, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375437, CurrentID=0, Status='live', Symbol='LOC105375437', Aliases='', description='uncharacterized LOC105375437', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=105807643, end_position_on_the_genomic_accession=105816676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375438, CurrentID=0, Status='live', Symbol='LOC105375438', Aliases='', description='uncharacterized LOC105375438', other_designations='CTB-111H14.1', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106461958, end_position_on_the_genomic_accession=106466551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375439, CurrentID=0, Status='live', Symbol='LOC105375439', Aliases='', description='uncharacterized LOC105375439', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106588137, end_position_on_the_genomic_accession=106598900, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375440, CurrentID=0, Status='live', Symbol='LOC105375440', Aliases='', description='uncharacterized LOC105375440', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=106624104, end_position_on_the_genomic_accession=106628849, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375441, CurrentID=0, Status='live', Symbol='LOC105375441', Aliases='', description='uncharacterized LOC105375441', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107066925, end_position_on_the_genomic_accession=107079609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  124626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124626, CurrentID=0, Status='live', Symbol='ZPBP2', Aliases='ZPBPL', description='zona pellucida binding protein 2', other_designations='ZPBP-like protein', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39868164, end_position_on_the_genomic_accession=39877896, orientation='plus', exon_count=8, OMIM=608499, no_hdr0=''),
  105375443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375443, CurrentID=0, Status='live', Symbol='LOC105375443', Aliases='', description='uncharacterized LOC105375443', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107631894, end_position_on_the_genomic_accession=107643352, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375444, CurrentID=0, Status='live', Symbol='LOC105375444', Aliases='', description='uncharacterized LOC105375444', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=107856479, end_position_on_the_genomic_accession=107870881, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132565, CurrentID=0, Status='live', Symbol='GOLGA8F', Aliases='', description='golgin A8 family member F', other_designations='golgi autoantigen, golgin subfamily a, 8F|golgin A8 family, member F pseudogene|golgin-like hypothetical protein LOC440321', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28378638, end_position_on_the_genomic_accession=28391266, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105375446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375446, CurrentID=0, Status='live', Symbol='LOC105375446', Aliases='', description='uncharacterized LOC105375446', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108084271, end_position_on_the_genomic_accession=108091374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375447, CurrentID=0, Status='live', Symbol='LOC105375447', Aliases='', description='uncharacterized LOC105375447', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108554543, end_position_on_the_genomic_accession=108561807, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375448, CurrentID=0, Status='live', Symbol='LOC105375448', Aliases='', description='uncharacterized LOC105375448', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=108577418, end_position_on_the_genomic_accession=108614273, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375449, CurrentID=0, Status='live', Symbol='LOC105375449', Aliases='', description='uncharacterized LOC105375449', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110531189, end_position_on_the_genomic_accession=110554153, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375450, CurrentID=0, Status='live', Symbol='LOC105375450', Aliases='', description='uncharacterized LOC105375450', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110436521, end_position_on_the_genomic_accession=110456520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375451, CurrentID=0, Status='live', Symbol='LOC105375451', Aliases='', description='uncharacterized LOC105375451', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110432748, end_position_on_the_genomic_accession=110534687, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375452, CurrentID=0, Status='live', Symbol='LOC105375452', Aliases='', description='uncharacterized LOC105375452', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=110555234, end_position_on_the_genomic_accession=110588106, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  124637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124637, CurrentID=0, Status='live', Symbol='CYB5D1', Aliases='', description='cytochrome b5 domain containing 1', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7857746, end_position_on_the_genomic_accession=7863486, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375454, CurrentID=0, Status='live', Symbol='LOC105375454', Aliases='', description='uncharacterized LOC105375454', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111840694, end_position_on_the_genomic_accession=111846992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375455, CurrentID=0, Status='live', Symbol='LOC105375455', Aliases='', description='uncharacterized LOC105375455', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=111756974, end_position_on_the_genomic_accession=111758372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375456, CurrentID=0, Status='live', Symbol='LOC105375456', Aliases='', description='uncharacterized LOC105375456', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112370880, end_position_on_the_genomic_accession=112391337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  124641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124641, CurrentID=0, Status='live', Symbol='OVCA2', Aliases='', description='ovarian tumor suppressor candidate 2', other_designations='candidate tumor suppressor in ovarian cancer 2|ovarian cancer gene-2 protein', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=2041983, end_position_on_the_genomic_accession=2043431, orientation='plus', exon_count=2, OMIM=607896, no_hdr0=''),
  105375458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375458, CurrentID=0, Status='live', Symbol='LOC105375458', Aliases='', description='uncharacterized LOC105375458', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112522274, end_position_on_the_genomic_accession=112617338, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375459, CurrentID=0, Status='live', Symbol='LOC105375459', Aliases='', description='uncharacterized LOC105375459', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=112752834, end_position_on_the_genomic_accession=112758678, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132580, CurrentID=0, Status='live', Symbol='RBMXP4', Aliases='', description='RNA binding motif protein, X-linked pseudogene 4', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109346133, end_position_on_the_genomic_accession=109347610, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375461, CurrentID=0, Status='live', Symbol='LOC105375461', Aliases='', description='uncharacterized LOC105375461', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115682791, end_position_on_the_genomic_accession=115684594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375462, CurrentID=0, Status='live', Symbol='LOC105375462', Aliases='', description='uncharacterized LOC105375462', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=115756735, end_position_on_the_genomic_accession=115758871, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375463, CurrentID=0, Status='live', Symbol='LOC105375463', Aliases='', description='uncharacterized LOC105375463', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116285341, end_position_on_the_genomic_accession=116349745, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105375464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375464, CurrentID=0, Status='live', Symbol='LOC105375464', Aliases='', description='uncharacterized LOC105375464', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116235868, end_position_on_the_genomic_accession=116243799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375465, CurrentID=0, Status='live', Symbol='LOC105375465', Aliases='', description='uncharacterized LOC105375465', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=116940339, end_position_on_the_genomic_accession=116947897, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375466, CurrentID=0, Status='live', Symbol='LOC105375466', Aliases='', description='uncharacterized LOC105375466', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117232821, end_position_on_the_genomic_accession=117241161, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375467, CurrentID=0, Status='live', Symbol='LOC105375467', Aliases='', description='uncharacterized LOC105375467', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117286797, end_position_on_the_genomic_accession=117322563, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375468, CurrentID=0, Status='live', Symbol='LOC105375468', Aliases='', description='uncharacterized LOC105375468', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=117447916, end_position_on_the_genomic_accession=117475729, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255725, CurrentID=0, Status='live', Symbol='OR52B2', Aliases='OR11-70', description='olfactory receptor family 52 subfamily B member 2', other_designations='olfactory receptor OR11-70|olfactory receptor, family 52, subfamily B, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6169355, end_position_on_the_genomic_accession=6170326, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375470, CurrentID=0, Status='live', Symbol='LOC105375470', Aliases='', description='uncharacterized LOC105375470', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118152314, end_position_on_the_genomic_accession=118184002, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132591, CurrentID=0, Status='live', Symbol='SERBP1P4', Aliases='', description='SERPINE1 mRNA binding protein 1 pseudogene 4', other_designations='', map_location='Xq21.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=91560438, end_position_on_the_genomic_accession=91561598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375472, CurrentID=0, Status='live', Symbol='LOC105375472', Aliases='', description='uncharacterized LOC105375472', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118474808, end_position_on_the_genomic_accession=118480594, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132593, CurrentID=0, Status='live', Symbol='LOC100132593', Aliases='', description='uncharacterized LOC100132593', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.12', start_position_on_the_genomic_accession=101604887, end_position_on_the_genomic_accession=101605799, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132594, CurrentID=0, Status='live', Symbol='PGAM1P5', Aliases='', description='phosphoglycerate mutase 1 pseudogene 5', other_designations='', map_location='12q22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=95649255, end_position_on_the_genomic_accession=95673999, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375475, CurrentID=0, Status='live', Symbol='LVCAT5', Aliases='', description='liver cancer-associated transcript 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=119764937, end_position_on_the_genomic_accession=119895309, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132596, CurrentID=0, Status='live', Symbol='XGY2', Aliases='XG, XGPY, XGPY2', description='Xg pseudogene, Y-linked 2', other_designations='Xg blood group pseudogene', map_location='Yp11.31', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=2752296, end_position_on_the_genomic_accession=2774996, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132597, CurrentID=0, Status='live', Symbol='EIF3EP2', Aliases='', description='eukaryotic translation initiation factor 3 subunit E pseudogene 2', other_designations='eukaryotic translation initiation factor 3, subunit E pseudogene 2', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=162088259, end_position_on_the_genomic_accession=162090276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132598, CurrentID=0, Status='live', Symbol='ZNF965P', Aliases='', description='zinc finger protein 965, pseudogene', other_designations='zinc finger protein 114 pseudogene', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18733719, end_position_on_the_genomic_accession=18734428, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375479, CurrentID=0, Status='live', Symbol='LOC105375479', Aliases='', description='uncharacterized LOC105375479', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121480303, end_position_on_the_genomic_accession=121490711, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375480, CurrentID=0, Status='live', Symbol='LOC105375480', Aliases='', description='uncharacterized LOC105375480', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121731682, end_position_on_the_genomic_accession=121756004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132601, CurrentID=0, Status='live', Symbol='GAPDHP68', Aliases='GAPDHL13, GAPDL13', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 68', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 13, pseudogene', map_location='7p21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=9614908, end_position_on_the_genomic_accession=9616186, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255738, CurrentID=0, Status='live', Symbol='PCSK9', Aliases='FH3, HCHOLA3, LDLCQ1, NARC-1, NARC1, PC9', description='proprotein convertase subtilisin/kexin type 9', other_designations='convertase subtilisin/kexin type 9 preproprotein|neural apoptosis regulated convertase 1|subtilisin/kexin-like protease PC9', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55039476, end_position_on_the_genomic_accession=55064853, orientation='plus', exon_count=14, OMIM=607786, no_hdr0=''),
  105375483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375483, CurrentID=0, Status='live', Symbol='LOC105375483', Aliases='', description='uncharacterized LOC105375483', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=123994631, end_position_on_the_genomic_accession=124027646, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375484, CurrentID=0, Status='live', Symbol='LOC105375484', Aliases='', description='uncharacterized LOC105375484', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124274235, end_position_on_the_genomic_accession=124288855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132605, CurrentID=0, Status='live', Symbol='LOC100132605', Aliases='', description='uncharacterized LOC100132605', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.9', start_position_on_the_genomic_accession=663645, end_position_on_the_genomic_accession=675900, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100132606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132606, CurrentID=0, Status='live', Symbol='AGGF1P7', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 7', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 7', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35547855, end_position_on_the_genomic_accession=35548835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255743, CurrentID=0, Status='live', Symbol='NPNT', Aliases='EGFL6L, POEM', description='nephronectin', other_designations='preosteoblast EGF-like repeat protein with MAM domain', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105894526, end_position_on_the_genomic_accession=105971671, orientation='plus', exon_count=17, OMIM=610306, no_hdr0=''),
  105375488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375488, CurrentID=0, Status='live', Symbol='LOC105375488', Aliases='', description='uncharacterized LOC105375488', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=126233570, end_position_on_the_genomic_accession=126253648, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375489, CurrentID=0, Status='live', Symbol='LOC105375489', Aliases='', description='uncharacterized LOC105375489', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127591582, end_position_on_the_genomic_accession=127592196, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375490, CurrentID=0, Status='live', Symbol='LOC105375490', Aliases='', description='uncharacterized LOC105375490', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=127485197, end_position_on_the_genomic_accession=127590059, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375491, CurrentID=0, Status='live', Symbol='LOC105375491', Aliases='', description='uncharacterized LOC105375491', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128049578, end_position_on_the_genomic_accession=128061568, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132612, CurrentID=0, Status='live', Symbol='LOC100132612', Aliases='', description='uncharacterized LOC100132612', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.7', start_position_on_the_genomic_accession=97286119, end_position_on_the_genomic_accession=97286634, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375493, CurrentID=0, Status='live', Symbol='LOC105375493', Aliases='', description='uncharacterized LOC105375493', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128121847, end_position_on_the_genomic_accession=128126321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375494, CurrentID=0, Status='live', Symbol='LOC105375494', Aliases='', description='uncharacterized LOC105375494', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128235683, end_position_on_the_genomic_accession=128236798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375495, CurrentID=0, Status='live', Symbol='LOC105375495', Aliases='', description='uncharacterized LOC105375495', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128434063, end_position_on_the_genomic_accession=128455594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375496, CurrentID=0, Status='live', Symbol='LOC105375496', Aliases='', description='uncharacterized LOC105375496', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128822551, end_position_on_the_genomic_accession=128828979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132617, CurrentID=0, Status='live', Symbol='SNX18P23', Aliases='', description='sorting nexin 18 pseudogene 23', other_designations='', map_location='4p11', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49232915, end_position_on_the_genomic_accession=49233959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132618, CurrentID=0, Status='live', Symbol='ZRANB2-AS1', Aliases='', description='ZRANB2 antisense RNA 1', other_designations='ZRANB2 antisense RNA 1 (non-protein coding)', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71046506, end_position_on_the_genomic_accession=71067184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132619, CurrentID=0, Status='live', Symbol='ATP5A1P10', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1 pseudogene 10', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66721435, end_position_on_the_genomic_accession=66722864, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375500, CurrentID=0, Status='live', Symbol='LOC105375500', Aliases='', description='uncharacterized LOC105375500', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129350981, end_position_on_the_genomic_accession=129352477, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124685, CurrentID=0, Status='live', Symbol='MYL6P5', Aliases='', description='myosin light chain 6 pseudogene 5', other_designations='myosin, light chain 6 pseudogene 5|myosin, light chain 6, alkali, smooth muscle and non-muscle pseudogene', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=71620938, end_position_on_the_genomic_accession=71621631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255758, CurrentID=0, Status='live', Symbol='TCTEX1D2', Aliases='', description='Tctex1 domain containing 2', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196291219, end_position_on_the_genomic_accession=196318294, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375503, CurrentID=0, Status='live', Symbol='LOC105375503', Aliases='', description='uncharacterized LOC105375503', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130318649, end_position_on_the_genomic_accession=130344572, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375504, CurrentID=0, Status='live', Symbol='LOC105375504', Aliases='', description='uncharacterized LOC105375504', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130365650, end_position_on_the_genomic_accession=130367127, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375505, CurrentID=0, Status='live', Symbol='LOC105375505', Aliases='', description='uncharacterized LOC105375505', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130441976, end_position_on_the_genomic_accession=130443226, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255762, CurrentID=0, Status='live', Symbol='PDZD9', Aliases='C16orf65', description='PDZ domain containing 9', other_designations='PDZ domain-containing protein C16orf65', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21957540, end_position_on_the_genomic_accession=22001127, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375507, CurrentID=0, Status='live', Symbol='LOC105375507', Aliases='', description='uncharacterized LOC105375507', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130595572, end_position_on_the_genomic_accession=130597009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  648980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648980, CurrentID=0, Status='live', Symbol='BCRP5', Aliases='BCRL5', description='breakpoint cluster region pseudogene 5', other_designations='breakpoint cluster region-like 5', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20695438, end_position_on_the_genomic_accession=20698959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375509, CurrentID=0, Status='live', Symbol='LOC105375509', Aliases='', description='uncharacterized LOC105375509', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=130791462, end_position_on_the_genomic_accession=130832757, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132630, CurrentID=0, Status='live', Symbol='CICP3', Aliases='', description='capicua transcriptional repressor pseudogene 3', other_designations='capicua homolog pseudogene 3|protein capicua homolog', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=722046, end_position_on_the_genomic_accession=725822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375511, CurrentID=0, Status='live', Symbol='LOC105375511', Aliases='', description='uncharacterized LOC105375511', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=132403448, end_position_on_the_genomic_accession=132410331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375512, CurrentID=0, Status='live', Symbol='LOC105375512', Aliases='', description='uncharacterized LOC105375512', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133099071, end_position_on_the_genomic_accession=133108591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375513, CurrentID=0, Status='live', Symbol='LOC105375513', Aliases='', description='uncharacterized LOC105375513', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134073215, end_position_on_the_genomic_accession=134102412, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375514, CurrentID=0, Status='live', Symbol='LOC105375514', Aliases='', description='uncharacterized LOC105375514', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134072427, end_position_on_the_genomic_accession=134098189, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  648987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=648987, CurrentID=0, Status='live', Symbol='LOC648987', Aliases='', description='uncharacterized LOC648987', other_designations='CTD-2201E18.3', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43014729, end_position_on_the_genomic_accession=43067439, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375516, CurrentID=0, Status='live', Symbol='LOC105375516', Aliases='', description='uncharacterized LOC105375516', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134119317, end_position_on_the_genomic_accession=134125756, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375517, CurrentID=0, Status='live', Symbol='LOC105375517', Aliases='', description='uncharacterized LOC105375517', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134368823, end_position_on_the_genomic_accession=134374270, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375518, CurrentID=0, Status='live', Symbol='LOC105375518', Aliases='', description='uncharacterized LOC105375518', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134366771, end_position_on_the_genomic_accession=134384225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375519, CurrentID=0, Status='live', Symbol='LOC105375519', Aliases='', description='uncharacterized LOC105375519', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134415576, end_position_on_the_genomic_accession=134432407, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375520, CurrentID=0, Status='live', Symbol='LOC105375520', Aliases='', description='uncharacterized LOC105375520', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=134549248, end_position_on_the_genomic_accession=134558334, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132641, CurrentID=0, Status='live', Symbol='BTF3P14', Aliases='', description='basic transcription factor 3 pseudogene 14', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6322415, end_position_on_the_genomic_accession=6323228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375522, CurrentID=0, Status='live', Symbol='LOC105375522', Aliases='', description='uncharacterized LOC105375522', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135758457, end_position_on_the_genomic_accession=135771179, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375523, CurrentID=0, Status='live', Symbol='LOC105375523', Aliases='', description='uncharacterized LOC105375523', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136025622, end_position_on_the_genomic_accession=136439517, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375524, CurrentID=0, Status='live', Symbol='LOC105375524', Aliases='', description='uncharacterized LOC105375524', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136743805, end_position_on_the_genomic_accession=136752301, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132645, CurrentID=0, Status='live', Symbol='RPL17P41', Aliases='RPL17_26_1572', description='ribosomal protein L17 pseudogene 41', other_designations='', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=67784708, end_position_on_the_genomic_accession=67785328, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375526, CurrentID=0, Status='live', Symbol='LOC105375526', Aliases='', description='uncharacterized LOC105375526', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137363712, end_position_on_the_genomic_accession=137365167, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255783, CurrentID=0, Status='live', Symbol='INAFM1', Aliases='PRR24', description='InaF-motif containing 1', other_designations='inaF-motif-containing protein 1|proline rich 24|proline-rich protein 24', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47274402, end_position_on_the_genomic_accession=47275723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375528, CurrentID=0, Status='live', Symbol='LOC105375528', Aliases='', description='uncharacterized LOC105375528', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=137303168, end_position_on_the_genomic_accession=137321266, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375529, CurrentID=0, Status='live', Symbol='LOC105375529', Aliases='', description='uncharacterized LOC105375529', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139424735, end_position_on_the_genomic_accession=139425500, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375530, CurrentID=0, Status='live', Symbol='LOC105375530', Aliases='', description='uncharacterized LOC105375530', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139584072, end_position_on_the_genomic_accession=139615848, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375531, CurrentID=0, Status='live', Symbol='LOC105375531', Aliases='', description='uncharacterized LOC105375531', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139970393, end_position_on_the_genomic_accession=139978524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375532, CurrentID=0, Status='live', Symbol='LOC105375532', Aliases='', description='uncharacterized LOC105375532', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=139939363, end_position_on_the_genomic_accession=139947147, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375533, CurrentID=0, Status='live', Symbol='LOC105375533', Aliases='', description='uncharacterized LOC105375533', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140073427, end_position_on_the_genomic_accession=140083502, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375534, CurrentID=0, Status='live', Symbol='LOC105375534', Aliases='', description='uncharacterized LOC105375534', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140294529, end_position_on_the_genomic_accession=140297436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375535, CurrentID=0, Status='live', Symbol='LOC105375535', Aliases='', description='uncharacterized LOC105375535', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140640967, end_position_on_the_genomic_accession=140652431, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132656, CurrentID=0, Status='live', Symbol='FKBP4P1', Aliases='', description='FK506 binding protein 4 pseudogene 1', other_designations='FK506 binding protein 4, 59kDa pseudogene 1', map_location='4q26', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118192727, end_position_on_the_genomic_accession=118194895, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375537, CurrentID=0, Status='live', Symbol='LOC105375537', Aliases='', description='uncharacterized LOC105375537', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141056880, end_position_on_the_genomic_accession=141070707, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375538, CurrentID=0, Status='live', Symbol='LOC105375538', Aliases='', description='uncharacterized LOC105375538', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141500095, end_position_on_the_genomic_accession=141551231, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929267, CurrentID=0, Status='live', Symbol='LOC101929267', Aliases='', description='uncharacterized LOC101929267', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122963987, end_position_on_the_genomic_accession=122977631, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375540, CurrentID=0, Status='live', Symbol='LOC105375540', Aliases='', description='uncharacterized LOC105375540', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142309750, end_position_on_the_genomic_accession=142312366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375541, CurrentID=0, Status='live', Symbol='LOC105375541', Aliases='', description='uncharacterized LOC105375541', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142813301, end_position_on_the_genomic_accession=142817433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255798, CurrentID=0, Status='live', Symbol='SMCO1', Aliases='C3orf43', description='single-pass membrane protein with coiled-coil domains 1', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=196506879, end_position_on_the_genomic_accession=196515366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375543, CurrentID=0, Status='live', Symbol='LOC105375543', Aliases='', description='uncharacterized LOC105375543', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=142937555, end_position_on_the_genomic_accession=142939092, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375544, CurrentID=0, Status='live', Symbol='LOC105375544', Aliases='', description='uncharacterized LOC105375544', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143029260, end_position_on_the_genomic_accession=143039487, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375545, CurrentID=0, Status='live', Symbol='LOC105375545', Aliases='', description='uncharacterized LOC105375545', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143255307, end_position_on_the_genomic_accession=143286202, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375546, CurrentID=0, Status='live', Symbol='LOC105375546', Aliases='', description='uncharacterized LOC105375546', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143237575, end_position_on_the_genomic_accession=143263243, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375547, CurrentID=0, Status='live', Symbol='LOC105375547', Aliases='', description='uncharacterized LOC105375547', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143495359, end_position_on_the_genomic_accession=143496986, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375548, CurrentID=0, Status='live', Symbol='LOC105375548', Aliases='', description='uncharacterized LOC105375548', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144058490, end_position_on_the_genomic_accession=144084827, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375549, CurrentID=0, Status='live', Symbol='LOC105375549', Aliases='', description='uncharacterized LOC105375549', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143885446, end_position_on_the_genomic_accession=143901949, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375550, CurrentID=0, Status='live', Symbol='LOC105375550', Aliases='', description='uncharacterized LOC105375550', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143836868, end_position_on_the_genomic_accession=143875981, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375551, CurrentID=0, Status='live', Symbol='LOC105375551', Aliases='', description='uncharacterized LOC105375551', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=145042675, end_position_on_the_genomic_accession=145165679, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100132672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132672, CurrentID=0, Status='live', Symbol='GXYLT1P6', Aliases='', description='glucoside xylosyltransferase 1 pseudogene 6', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64488535, end_position_on_the_genomic_accession=64489985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  255809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255809, CurrentID=0, Status='live', Symbol='C19orf38', Aliases='HIDE1', description='chromosome 19 open reading frame 38', other_designations='Highly expressed in immature dendritic cell transcript 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10836653, end_position_on_the_genomic_accession=10869791, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375554, CurrentID=0, Status='live', Symbol='LOC105375554', Aliases='', description='uncharacterized LOC105375554', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148381638, end_position_on_the_genomic_accession=148436899, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  124739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124739, CurrentID=0, Status='live', Symbol='USP43', Aliases='', description='ubiquitin specific peptidase 43', other_designations='deubiquitinating enzyme 43|ubiquitin specific protease 43|ubiquitin thioesterase 43|ubiquitin thiolesterase 43|ubiquitin-specific-processing protease 43', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=9643493, end_position_on_the_genomic_accession=9729687, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  255812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255812, CurrentID=0, Status='live', Symbol='SDHAP1', Aliases='SDHAL1, SDHALP1', description='succinate dehydrogenase complex subunit A, flavoprotein pseudogene 1', other_designations='succinate dehydrogenase complex, subunit A, flavoprotein pseudogene 1|succinate dehydrogenase complex, subunit A, flavoprotein-like 1', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195959921, end_position_on_the_genomic_accession=195990279, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100132677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132677, CurrentID=0, Status='live', Symbol='BSN-AS2', Aliases='', description='BSN antisense RNA 2 (head to head)', other_designations='BSN antisense RNA 2 (non-protein coding)', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49549306, end_position_on_the_genomic_accession=49554366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132678, CurrentID=0, Status='live', Symbol='RPL7AP45', Aliases='RPL7A_15_979', description='ribosomal protein L7a pseudogene 45', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41479933, end_position_on_the_genomic_accession=41480534, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375559, CurrentID=0, Status='live', Symbol='LOC105375559', Aliases='', description='uncharacterized LOC105375559', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149397231, end_position_on_the_genomic_accession=149401431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375560, CurrentID=0, Status='live', Symbol='LOC105375560', Aliases='', description='uncharacterized LOC105375560', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150332701, end_position_on_the_genomic_accession=150334318, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132681, CurrentID=0, Status='live', Symbol='NFU1P1', Aliases='', description='NFU1 iron-sulfur cluster scaffold pseudogene 1', other_designations='NFU1 iron-sulfur cluster scaffold homolog pseudogene 1', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39644041, end_position_on_the_genomic_accession=39644822, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375562, CurrentID=0, Status='live', Symbol='LOC105375562', Aliases='', description='uncharacterized LOC105375562', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149973528, end_position_on_the_genomic_accession=149984139, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132683, CurrentID=0, Status='live', Symbol='PDCL3P3', Aliases='', description='phosducin-like 3 pseudogene 3', other_designations='', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17877266, end_position_on_the_genomic_accession=17878196, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375564, CurrentID=0, Status='live', Symbol='LOC105375564', Aliases='', description='uncharacterized LOC105375564', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150379844, end_position_on_the_genomic_accession=150384006, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375565, CurrentID=0, Status='live', Symbol='LOC105375565', Aliases='', description='uncharacterized LOC105375565', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150648476, end_position_on_the_genomic_accession=150686725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375566, CurrentID=0, Status='live', Symbol='LOC105375566', Aliases='', description='uncharacterized LOC105375566', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150761587, end_position_on_the_genomic_accession=150782108, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  124751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124751, CurrentID=0, Status='live', Symbol='KRBA2', Aliases='', description='KRAB-A domain containing 2', other_designations='KRAB A domain containing 2', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8368637, end_position_on_the_genomic_accession=8376711, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375568, CurrentID=0, Status='live', Symbol='LOC105375568', Aliases='', description='uncharacterized LOC105375568', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151024995, end_position_on_the_genomic_accession=151028267, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375569, CurrentID=0, Status='live', Symbol='LOC105375569', Aliases='', description='uncharacterized LOC105375569', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151248747, end_position_on_the_genomic_accession=151257102, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375570, CurrentID=0, Status='live', Symbol='LOC105375570', Aliases='', description='uncharacterized LOC105375570', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151946951, end_position_on_the_genomic_accession=151954995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375571, CurrentID=0, Status='live', Symbol='LOC105375571', Aliases='', description='uncharacterized LOC105375571', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151955119, end_position_on_the_genomic_accession=151956345, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375572, CurrentID=0, Status='live', Symbol='LOC105375572', Aliases='', description='uncharacterized LOC105375572', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152100984, end_position_on_the_genomic_accession=152102433, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375573, CurrentID=0, Status='live', Symbol='LOC105375573', Aliases='', description='uncharacterized LOC105375573', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152071793, end_position_on_the_genomic_accession=152085656, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375574, CurrentID=0, Status='live', Symbol='LOC105375574', Aliases='', description='uncharacterized LOC105375574', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=152513883, end_position_on_the_genomic_accession=152523309, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100423037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100423037, CurrentID=0, Status='live', Symbol='MIR3176', Aliases='mir-3176', description='microRNA 3176', other_designations='hsa-mir-3176', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=543277, end_position_on_the_genomic_accession=543366, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375576, CurrentID=0, Status='live', Symbol='LOC105375576', Aliases='', description='uncharacterized LOC105375576', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153190795, end_position_on_the_genomic_accession=153225409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375577, CurrentID=0, Status='live', Symbol='LOC105375577', Aliases='', description='uncharacterized LOC105375577', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153389597, end_position_on_the_genomic_accession=153396975, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375578, CurrentID=0, Status='live', Symbol='LOC105375578', Aliases='', description='uncharacterized LOC105375578', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153423960, end_position_on_the_genomic_accession=153436251, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132699, CurrentID=0, Status='live', Symbol='NFU1P2', Aliases='', description='NFU1 iron-sulfur cluster scaffold pseudogene 2', other_designations='NFU1 iron-sulfur cluster scaffold homolog pseudogene 2', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98076814, end_position_on_the_genomic_accession=98077500, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375580, CurrentID=0, Status='live', Symbol='LOC105375580', Aliases='', description='uncharacterized LOC105375580', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154825116, end_position_on_the_genomic_accession=154865463, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375581, CurrentID=0, Status='live', Symbol='LOC105375581', Aliases='', description='uncharacterized LOC105375581', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154614759, end_position_on_the_genomic_accession=154624910, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375582, CurrentID=0, Status='live', Symbol='LOC105375582', Aliases='', description='uncharacterized LOC105375582', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154232355, end_position_on_the_genomic_accession=154248178, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375583, CurrentID=0, Status='live', Symbol='LOC105375583', Aliases='', description='uncharacterized LOC105375583', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=153754378, end_position_on_the_genomic_accession=153755904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375584, CurrentID=0, Status='live', Symbol='LOC105375584', Aliases='', description='uncharacterized LOC105375584', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155169336, end_position_on_the_genomic_accession=155173319, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375585, CurrentID=0, Status='live', Symbol='LOC105375585', Aliases='', description='uncharacterized LOC105375585', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155222627, end_position_on_the_genomic_accession=155227237, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375586, CurrentID=0, Status='live', Symbol='LOC105375586', Aliases='', description='uncharacterized LOC105375586', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155203700, end_position_on_the_genomic_accession=155214634, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132707, CurrentID=0, Status='live', Symbol='PAXIP1-AS2', Aliases='PAXIP1OS', description='PAXIP1 antisense RNA 2', other_designations='PAXIP1 opposite strand', map_location='7q36.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=154928517, end_position_on_the_genomic_accession=154949490, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100132708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132708, CurrentID=0, Status='live', Symbol='CYP4F30P', Aliases='C2orf14', description='cytochrome P450 family 4 subfamily F member 30, pseudogene', other_designations='4F-se9[6:7:8]|cytochrome P450, family 4, subfamily F, polypeptide 30, pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130680050, end_position_on_the_genomic_accession=130685863, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  124773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124773, CurrentID=0, Status='live', Symbol='C17orf64', Aliases='', description='chromosome 17 open reading frame 64', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60421085, end_position_on_the_genomic_accession=60431426, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375590, CurrentID=0, Status='live', Symbol='LOC105375590', Aliases='', description='uncharacterized LOC105375590', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155269530, end_position_on_the_genomic_accession=155278415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375591, CurrentID=0, Status='live', Symbol='LOC105375591', Aliases='', description='uncharacterized LOC105375591', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155288609, end_position_on_the_genomic_accession=155298252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132712, CurrentID=0, Status='live', Symbol='NAMPTP2', Aliases='', description='nicotinamide phosphoribosyltransferase pseudogene 2', other_designations='', map_location='5q31.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=145002238, end_position_on_the_genomic_accession=145005801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375593, CurrentID=0, Status='live', Symbol='LOC105375593', Aliases='', description='uncharacterized LOC105375593', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155364222, end_position_on_the_genomic_accession=155370289, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375594, CurrentID=0, Status='live', Symbol='LOC105375594', Aliases='', description='uncharacterized LOC105375594', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155414884, end_position_on_the_genomic_accession=155418811, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375596, CurrentID=0, Status='live', Symbol='LOC105375596', Aliases='', description='uncharacterized LOC105375596', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=155939891, end_position_on_the_genomic_accession=155944973, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375597, CurrentID=0, Status='live', Symbol='LOC105375597', Aliases='', description='uncharacterized LOC105375597', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156012680, end_position_on_the_genomic_accession=156033398, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375598, CurrentID=0, Status='live', Symbol='LOC105375598', Aliases='', description='uncharacterized LOC105375598', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156522098, end_position_on_the_genomic_accession=156525230, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124783, CurrentID=0, Status='live', Symbol='SPATA32', Aliases='AEP2, C17orf46, TEX34, VAD1.2', description='spermatogenesis associated 32', other_designations='CTD-2020K17.4|acrosome expressed 2|testis expressed 34|testis-expressed sequence 34 protein', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45247274, end_position_on_the_genomic_accession=45262112, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105375600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375600, CurrentID=0, Status='live', Symbol='LOC105375600', Aliases='', description='uncharacterized LOC105375600', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156223471, end_position_on_the_genomic_accession=156236167, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375601, CurrentID=0, Status='live', Symbol='LOC105375601', Aliases='', description='uncharacterized LOC105375601', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156194064, end_position_on_the_genomic_accession=156209890, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132722, CurrentID=0, Status='live', Symbol='RPL15P21', Aliases='RPL15_8_1514', description='ribosomal protein L15 pseudogene 21', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10860531, end_position_on_the_genomic_accession=10861462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375603, CurrentID=0, Status='live', Symbol='LOC105375603', Aliases='', description='uncharacterized LOC105375603', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156140074, end_position_on_the_genomic_accession=156145989, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132724, CurrentID=0, Status='live', Symbol='DCAF13P3', Aliases='', description='DDB1 and CUL4 associated factor 13 pseudogene 3', other_designations='WD repeats and SOF1 domain containing pseudogene', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=50944129, end_position_on_the_genomic_accession=50946573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375605, CurrentID=0, Status='live', Symbol='LOC105375605', Aliases='', description='uncharacterized LOC105375605', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156942430, end_position_on_the_genomic_accession=156945780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124790, CurrentID=0, Status='live', Symbol='HEXIM2', Aliases='L3', description='hexamethylene bis-acetamide inducible 2', other_designations='MAQ1 paralog|hexamethylene bis-acetamide-inducible protein 2|hexamethylene-bis-acetamide-inducible transcript 2', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45160026, end_position_on_the_genomic_accession=45170040, orientation='plus', exon_count=7, OMIM=615695, no_hdr0=''),
  105375607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375607, CurrentID=0, Status='live', Symbol='LOC105375607', Aliases='', description='uncharacterized LOC105375607', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157111166, end_position_on_the_genomic_accession=157124009, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375608, CurrentID=0, Status='live', Symbol='LOC105375608', Aliases='', description='uncharacterized LOC105375608', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157368183, end_position_on_the_genomic_accession=157373439, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375609, CurrentID=0, Status='live', Symbol='LOC105375609', Aliases='', description='uncharacterized LOC105375609', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157465524, end_position_on_the_genomic_accession=157466177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375610, CurrentID=0, Status='live', Symbol='LOC105375610', Aliases='', description='uncharacterized LOC105375610', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157501191, end_position_on_the_genomic_accession=157518614, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100132731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132731, CurrentID=0, Status='live', Symbol='LOC100132731', Aliases='', description='uncharacterized LOC100132731', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128910159, end_position_on_the_genomic_accession=128965645, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375612, CurrentID=0, Status='live', Symbol='LOC105375612', Aliases='', description='uncharacterized LOC105375612', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158433391, end_position_on_the_genomic_accession=158447848, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132733, CurrentID=0, Status='live', Symbol='ANKRD20A13P', Aliases='', description='ankyrin repeat domain 20 family member A13, pseudogene', other_designations='ankyrin repeat domain 20 family, member A13, pseudogene', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375614, CurrentID=0, Status='live', Symbol='LOC105375614', Aliases='', description='uncharacterized LOC105375614', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=157614023, end_position_on_the_genomic_accession=157618771, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132735, CurrentID=0, Status='live', Symbol='LOC100132735', Aliases='', description='uncharacterized LOC100132735', other_designations='', map_location='6q24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=139771073, end_position_on_the_genomic_accession=139860471, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375616, CurrentID=0, Status='live', Symbol='LINC01022', Aliases='', description='long intergenic non-protein coding RNA 1022', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=158590236, end_position_on_the_genomic_accession=158591120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124801, CurrentID=0, Status='live', Symbol='LSM12', Aliases='PNAS-135', description='LSM12 homolog', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44034635, end_position_on_the_genomic_accession=44067619, orientation='minus', exon_count=7, OMIM=611793, no_hdr0=''),
  105375618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375618, CurrentID=0, Status='live', Symbol='LOC105375618', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63753006, end_position_on_the_genomic_accession=63754024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375619, CurrentID=0, Status='live', Symbol='LOC105375619', Aliases='', description='helicase SRCAP-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63761930, end_position_on_the_genomic_accession=63763923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375620, CurrentID=0, Status='live', Symbol='LOC105375620', Aliases='', description='basic salivary proline-rich protein 2-like', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=129609399, end_position_on_the_genomic_accession=129612722, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  255877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255877, CurrentID=0, Status='live', Symbol='BCL6B', Aliases='BAZF, ZBTB28, ZNF62', description='B-cell CLL/lymphoma 6, member B', other_designations='B-cell CLL/lymphoma 6, member B (zinc finger protein)|bcl6-associated zinc finger protein|zinc finger protein 62', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7023050, end_position_on_the_genomic_accession=7029642, orientation='plus', exon_count=9, OMIM=608992, no_hdr0=''),
  100132742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132742, CurrentID=0, Status='live', Symbol='RPL17P7', Aliases='RPL17L, RPL17_3_82', description='ribosomal protein L17 pseudogene 7', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108992254, end_position_on_the_genomic_accession=108992871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375623, CurrentID=0, Status='live', Symbol='LOC105375623', Aliases='', description='uncharacterized LOC105375623', other_designations='CTD-3118D11.3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86180420, end_position_on_the_genomic_accession=86212477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  124808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124808, CurrentID=0, Status='live', Symbol='CCDC43', Aliases='', description='coiled-coil domain containing 43', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44677437, end_position_on_the_genomic_accession=44689797, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375625, CurrentID=0, Status='live', Symbol='LOC105375625', Aliases='', description='uncharacterized LOC105375625', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=86616775, end_position_on_the_genomic_accession=86627996, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375626, CurrentID=0, Status='live', Symbol='LOC105375626', Aliases='', description='uncharacterized LOC105375626', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87609834, end_position_on_the_genomic_accession=87668439, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375627, CurrentID=0, Status='live', Symbol='LOC105375627', Aliases='', description='uncharacterized LOC105375627', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87925732, end_position_on_the_genomic_accession=87926224, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375628, CurrentID=0, Status='live', Symbol='LOC105375628', Aliases='', description='uncharacterized LOC105375628', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=87960981, end_position_on_the_genomic_accession=87971037, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375629, CurrentID=0, Status='live', Symbol='LOC105375629', Aliases='', description='uncharacterized LOC105375629', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88552698, end_position_on_the_genomic_accession=88598872, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375630, CurrentID=0, Status='live', Symbol='LOC105375630', Aliases='', description='uncharacterized LOC105375630', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88825672, end_position_on_the_genomic_accession=88886096, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375631, CurrentID=0, Status='live', Symbol='LOC105375631', Aliases='', description='uncharacterized LOC105375631', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88807904, end_position_on_the_genomic_accession=88894051, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105375632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375632, CurrentID=0, Status='live', Symbol='LOC105375632', Aliases='', description='uncharacterized LOC105375632', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=88910805, end_position_on_the_genomic_accession=88952076, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  124817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124817, CurrentID=0, Status='live', Symbol='CNTD1', Aliases='CNTD', description='cyclin N-terminal domain containing 1', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42798667, end_position_on_the_genomic_accession=42811587, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375634, CurrentID=0, Status='live', Symbol='LOC105375634', Aliases='', description='uncharacterized LOC105375634', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90845954, end_position_on_the_genomic_accession=90953980, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375635, CurrentID=0, Status='live', Symbol='LOC105375635', Aliases='', description='uncharacterized LOC105375635', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=90806475, end_position_on_the_genomic_accession=90812939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375636, CurrentID=0, Status='live', Symbol='GS1-251I9.3', Aliases='', description='uncharacterized LOC105375636', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91015621, end_position_on_the_genomic_accession=91018383, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375637, CurrentID=0, Status='live', Symbol='LOC105375637', Aliases='', description='uncharacterized LOC105375637', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=91261444, end_position_on_the_genomic_accession=91263949, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375638, CurrentID=0, Status='live', Symbol='LOC105375638', Aliases='', description='uncharacterized LOC105375638', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92407964, end_position_on_the_genomic_accession=92413639, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132759, CurrentID=0, Status='live', Symbol='VN1R7P', Aliases='ORLP1, PHB4C5', description='vomeronasal 1 receptor 7 pseudogene', other_designations='olfactory receptor-like pseudogene 1|pheromone receptor pseudogene', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10357653, end_position_on_the_genomic_accession=10359380, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375640, CurrentID=0, Status='live', Symbol='LOC105375640', Aliases='', description='uncharacterized LOC105375640', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=92984566, end_position_on_the_genomic_accession=93017710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101929284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929284, CurrentID=0, Status='live', Symbol='LOC101929284', Aliases='', description='uncharacterized LOC101929284', other_designations='', map_location='5p15.31', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=8839732, end_position_on_the_genomic_accession=8881525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375642, CurrentID=0, Status='live', Symbol='LOC105375642', Aliases='', description='uncharacterized LOC105375642', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93327721, end_position_on_the_genomic_accession=93342882, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375643, CurrentID=0, Status='live', Symbol='LOC105375643', Aliases='', description='uncharacterized LOC105375643', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93226139, end_position_on_the_genomic_accession=93230444, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375644, CurrentID=0, Status='live', Symbol='LOC105375644', Aliases='', description='uncharacterized LOC105375644', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93156981, end_position_on_the_genomic_accession=93174333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375645, CurrentID=0, Status='live', Symbol='LOC105375645', Aliases='', description='uncharacterized LOC105375645', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93741419, end_position_on_the_genomic_accession=93744614, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132766, CurrentID=0, Status='live', Symbol='PABPC1P8', Aliases='', description='poly(A) binding protein, cytoplasmic 1 pseudogene 8', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42279729, end_position_on_the_genomic_accession=42281795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375647, CurrentID=0, Status='live', Symbol='LOC105375647', Aliases='', description='uncharacterized LOC105375647', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94144155, end_position_on_the_genomic_accession=94168090, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375648, CurrentID=0, Status='live', Symbol='LOC105375648', Aliases='', description='uncharacterized LOC105375648', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94225566, end_position_on_the_genomic_accession=94228074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375649, CurrentID=0, Status='live', Symbol='LOC105375649', Aliases='', description='uncharacterized LOC105375649', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=94253662, end_position_on_the_genomic_accession=94260061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375650, CurrentID=0, Status='live', Symbol='LOC105375650', Aliases='', description='uncharacterized LOC105375650', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95064174, end_position_on_the_genomic_accession=95116455, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375651, CurrentID=0, Status='live', Symbol='LOC105375651', Aliases='', description='uncharacterized LOC105375651', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=95729379, end_position_on_the_genomic_accession=95733947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132772, CurrentID=0, Status='live', Symbol='RAB28P3', Aliases='', description='RAB28, member RAS oncogene family pseudogene 3', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42493229, end_position_on_the_genomic_accession=42494409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375653, CurrentID=0, Status='live', Symbol='LOC105375653', Aliases='', description='uncharacterized LOC105375653', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96419770, end_position_on_the_genomic_accession=96459472, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132774, CurrentID=0, Status='live', Symbol='KDM4A-AS1', Aliases='', description='KDM4A antisense RNA 1', other_designations='KDM4A antisense RNA 1 (non-protein coding)', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43699724, end_position_on_the_genomic_accession=43707341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375655, CurrentID=0, Status='live', Symbol='LOC105375655', Aliases='', description='uncharacterized LOC105375655', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97148039, end_position_on_the_genomic_accession=97171806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375656, CurrentID=0, Status='live', Symbol='LOC105375656', Aliases='', description='uncharacterized LOC105375656', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97188224, end_position_on_the_genomic_accession=97193986, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375657, CurrentID=0, Status='live', Symbol='LOC105375657', Aliases='', description='uncharacterized LOC105375657', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=97512979, end_position_on_the_genomic_accession=97519147, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124842, CurrentID=0, Status='live', Symbol='TMEM132E', Aliases='DFNB99', description='transmembrane protein 132E', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=34580749, end_position_on_the_genomic_accession=34639319, orientation='plus', exon_count=11, OMIM=616178, no_hdr0=''),
  105375659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375659, CurrentID=0, Status='live', Symbol='LOC105375659', Aliases='', description='uncharacterized LOC105375659', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98254308, end_position_on_the_genomic_accession=98305689, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375660, CurrentID=0, Status='live', Symbol='LOC105375660', Aliases='', description='uncharacterized LOC105375660', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98176486, end_position_on_the_genomic_accession=98249528, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132781, CurrentID=0, Status='live', Symbol='LOC100132781', Aliases='', description='cyclin Y like 1 pseudogene', other_designations='cyclin Y-like 1 pseudogene', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96868900, end_position_on_the_genomic_accession=96872228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375662, CurrentID=0, Status='live', Symbol='LOC105375662', Aliases='', description='uncharacterized LOC105375662', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=98373037, end_position_on_the_genomic_accession=98397739, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  255919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255919, CurrentID=0, Status='live', Symbol='CNEP1R1', Aliases='C16orf69, NEP1-R1, TMEM188, TMP125', description='CTD nuclear envelope phosphatase 1 regulatory subunit 1', other_designations='nuclear envelope phosphatase 1-regulatory subunit 1|transmembrane protein 188', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=50025206, end_position_on_the_genomic_accession=50037088, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105375664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375664, CurrentID=0, Status='live', Symbol='LOC105375664', Aliases='', description='uncharacterized LOC105375664', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=99655362, end_position_on_the_genomic_accession=99689589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649137, CurrentID=0, Status='live', Symbol='ZSCAN5CP', Aliases='ZNF495C, ZSCAN5C', description='zinc finger and SCAN domain containing 5C, pseudogene', other_designations='Zinc finger and SCAN domain-containing protein 5-like protein 2', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56205785, end_position_on_the_genomic_accession=56209454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100313927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313927, CurrentID=0, Status='live', Symbol='PIRC29', Aliases='', description='piwi-interacting RNA cluster 29', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375667, CurrentID=0, Status='live', Symbol='LOC105375667', Aliases='', description='uncharacterized LOC105375667', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100106005, end_position_on_the_genomic_accession=100107768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375668, CurrentID=0, Status='live', Symbol='LOC105375668', Aliases='', description='uncharacterized LOC105375668', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100234182, end_position_on_the_genomic_accession=100241939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375669, CurrentID=0, Status='live', Symbol='LOC105375669', Aliases='', description='uncharacterized LOC105375669', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100217172, end_position_on_the_genomic_accession=100230686, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  255926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255926, CurrentID=0, Status='live', Symbol='ADAM5', Aliases='ADAM5P, TMDCII', description='ADAM metallopeptidase domain 5 (pseudogene)', other_designations='ADAM metallopeptidase domain 5, pseudogene|a disintegrin and metalloproteinase domain 5|tMDC II', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=39314663, end_position_on_the_genomic_accession=39417226, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105375671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375671, CurrentID=0, Status='live', Symbol='LOC105375671', Aliases='', description='uncharacterized LOC105375671', other_designations='KB-1083B1.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100618528, end_position_on_the_genomic_accession=100620020, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  255928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255928, CurrentID=0, Status='live', Symbol='SYT14', Aliases='SCAR11, sytXIV', description='synaptotagmin 14', other_designations='synaptotagmin XIV', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209938174, end_position_on_the_genomic_accession=210164291, orientation='plus', exon_count=14, OMIM=610949, no_hdr0=''),
  124857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124857, CurrentID=0, Status='live', Symbol='WFIKKN2', Aliases='GASP-1, WFDC20B, WFIKKNRP, hGASP-1', description='WAP, follistatin/kazal, immunoglobulin, kunitz and netrin domain containing 2', other_designations='WAP four-disulfide core domain 20B|WAP, FS, Ig, two KU and NTR module related protein|WAP, follistatin, immunoglobulin, kunitz and NTR domain-containing-related protein|WFIKKN-related protein|growth and differentiation factor-associated serum protein 1|multivalent protease inhibitor protein', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=50835244, end_position_on_the_genomic_accession=50842348, orientation='plus', exon_count=2, OMIM=610895, no_hdr0=''),
  105375674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375674, CurrentID=0, Status='live', Symbol='LOC105375674', Aliases='', description='uncharacterized LOC105375674', other_designations='KB-1460A1.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101121818, end_position_on_the_genomic_accession=101126631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132795, CurrentID=0, Status='live', Symbol='RPL12P32', Aliases='RPL12_12_1226', description='ribosomal protein L12 pseudogene 32', other_designations='', map_location='12p11.21', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=31948246, end_position_on_the_genomic_accession=31948871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132796, CurrentID=0, Status='live', Symbol='NIFKP6', Aliases='MKI67IPP6', description='nucleolar protein interacting with the FHA domain of MKI67 pseudogene 6', other_designations='MKI67 (FHA domain) interacting nucleolar phosphoprotein pseudogene 6', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51396238, end_position_on_the_genomic_accession=51403599, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375677, CurrentID=0, Status='live', Symbol='LOC105375677', Aliases='', description='uncharacterized LOC105375677', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101175166, end_position_on_the_genomic_accession=101183996, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375678, CurrentID=0, Status='live', Symbol='LOC105375678', Aliases='', description='uncharacterized LOC105375678', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101214031, end_position_on_the_genomic_accession=101321227, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105375679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375679, CurrentID=0, Status='live', Symbol='LOC105375679', Aliases='', description='uncharacterized LOC105375679', other_designations='KB-1410C5.2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101261550, end_position_on_the_genomic_accession=101262978, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132800, CurrentID=0, Status='live', Symbol='LOC100132800', Aliases='', description='uncharacterized LOC100132800', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.8', start_position_on_the_genomic_accession=56785886, end_position_on_the_genomic_accession=56786421, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375681, CurrentID=0, Status='live', Symbol='LOC105375681', Aliases='', description='uncharacterized LOC105375681', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102062648, end_position_on_the_genomic_accession=102077466, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375682, CurrentID=0, Status='live', Symbol='LOC105375682', Aliases='', description='uncharacterized LOC105375682', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101856685, end_position_on_the_genomic_accession=101861166, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132803, CurrentID=0, Status='live', Symbol='HNRNPH1P1', Aliases='', description='heterogeneous nuclear ribonucleoprotein H1 pseudogene 1', other_designations='heterogeneous nuclear ribonucleoprotein H1 (H) pseudogene', map_location='6q25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159712551, end_position_on_the_genomic_accession=159713985, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375684, CurrentID=0, Status='live', Symbol='LOC105375684', Aliases='', description='uncharacterized LOC105375684', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102541067, end_position_on_the_genomic_accession=102551575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375685, CurrentID=0, Status='live', Symbol='LOC105375685', Aliases='', description='uncharacterized LOC105375685', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102680344, end_position_on_the_genomic_accession=102685814, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375686, CurrentID=0, Status='live', Symbol='LOC105375686', Aliases='', description='uncharacterized LOC105375686', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=102669611, end_position_on_the_genomic_accession=102680094, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649159, CurrentID=0, Status='live', Symbol='LINC00273', Aliases='NCRNA00273, TOP', description='long intergenic non-protein coding RNA 273', other_designations='NCRNA00273-1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=34158585, end_position_on_the_genomic_accession=34160036, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  124872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124872, CurrentID=0, Status='live', Symbol='B4GALNT2', Aliases='B4GALT, GALGT2', description='beta-1,4-N-acetyl-galactosaminyl transferase 2', other_designations='UDP-GalNAc:Neu5Aca2-3Galb-R b1,4-N-acetylgalactosaminyltransferase|UDP-GalNAc:Neu5Acalpha2-3Galbeta-R beta1,4-N-acetylgalactosaminyltransferase|beta 1,4 N-acetylgalactosaminyltransferase/betal,4 GalNAcT/Sda-GalNAcT|beta-4-N-acetylgalactosaminyltransferase|sd(a) beta-1,4-GalNAc transferase|sda beta-1,4-GalNAc transferase', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=49132460, end_position_on_the_genomic_accession=49169989, orientation='plus', exon_count=13, OMIM=111730, no_hdr0=''),
  105375689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375689, CurrentID=0, Status='live', Symbol='LOC105375689', Aliases='', description='uncharacterized LOC105375689', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103958757, end_position_on_the_genomic_accession=104005876, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375690, CurrentID=0, Status='live', Symbol='LOC105375690', Aliases='', description='uncharacterized LOC105375690', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=103487644, end_position_on_the_genomic_accession=103499488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375691, CurrentID=0, Status='live', Symbol='LOC105375691', Aliases='', description='uncharacterized LOC105375691', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104465876, end_position_on_the_genomic_accession=104484896, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132812, CurrentID=0, Status='live', Symbol='LOC100132812', Aliases='', description='nascent polypeptide-associated complex alpha subunit 2 pseudogene', other_designations='', map_location='8q13.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67569377, end_position_on_the_genomic_accession=67570195, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132813, CurrentID=0, Status='live', Symbol='LOC100132813', Aliases='', description='uncharacterized LOC100132813', other_designations='', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=109644115, end_position_on_the_genomic_accession=109648032, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375694, CurrentID=0, Status='live', Symbol='LOC105375694', Aliases='', description='uncharacterized LOC105375694', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=104827383, end_position_on_the_genomic_accession=104895717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132815, CurrentID=0, Status='live', Symbol='IPO5P1', Aliases='', description='importin 5 pseudogene 1', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23255053, end_position_on_the_genomic_accession=23274251, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375696, CurrentID=0, Status='live', Symbol='LOC105375696', Aliases='', description='uncharacterized LOC105375696', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105543280, end_position_on_the_genomic_accession=105557528, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102723513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723513, CurrentID=0, Status='live', Symbol='ZNF30-AS1', Aliases='', description='ZNF30 antisense RNA 1', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34923015, end_position_on_the_genomic_accession=34926826, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375698, CurrentID=0, Status='live', Symbol='LOC105375698', Aliases='', description='uncharacterized LOC105375698', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105530432, end_position_on_the_genomic_accession=105534428, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375699, CurrentID=0, Status='live', Symbol='LOC105375699', Aliases='', description='uncharacterized LOC105375699', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=105823178, end_position_on_the_genomic_accession=105834043, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375700, CurrentID=0, Status='live', Symbol='LOC105375700', Aliases='', description='uncharacterized LOC105375700', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=106646437, end_position_on_the_genomic_accession=106657734, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375701, CurrentID=0, Status='live', Symbol='LOC105375701', Aliases='', description='uncharacterized LOC105375701', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107165865, end_position_on_the_genomic_accession=107166757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375702, CurrentID=0, Status='live', Symbol='LOC105375702', Aliases='', description='uncharacterized LOC105375702', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107252026, end_position_on_the_genomic_accession=107256843, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375703, CurrentID=0, Status='live', Symbol='LOC105375703', Aliases='', description='uncharacterized LOC105375703', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=107776392, end_position_on_the_genomic_accession=107778747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387032, CurrentID=0, Status='live', Symbol='ZKSCAN4', Aliases='P1P373C6, ZNF307, ZNF427, ZSCAN36', description='zinc finger with KRAB and SCAN domains 4', other_designations='p373c6.1 (novel C2H2 type zinc finger protein)|zinc finger protein 307|zinc finger protein 427', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28241705, end_position_on_the_genomic_accession=28259271, orientation='minus', exon_count=9, OMIM=611643, no_hdr0=''),
  105375705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375705, CurrentID=0, Status='live', Symbol='LOC105375705', Aliases='', description='uncharacterized LOC105375705', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111178854, end_position_on_the_genomic_accession=111179736, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375706, CurrentID=0, Status='live', Symbol='LOC105375706', Aliases='', description='uncharacterized LOC105375706', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=111376628, end_position_on_the_genomic_accession=111617595, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  649179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649179, CurrentID=0, Status='live', Symbol='PRAMENP', Aliases='PRAMEF24P, PRAMEL', description='PRAME N-terminal-like, pseudogene', other_designations='PRAME family member 24, pseudogene|preferentially expressed antigen in melanoma-like', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21991099, end_position_on_the_genomic_accession=22044033, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  387036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387036, CurrentID=0, Status='live', Symbol='GUSBP2', Aliases='GUSBL1, GUSBP4, SMA3-L, SMAC3L, SMAC3L2, b55C20.1, bA239L20.1, bA239L20.5, bGLU-Lp', description='glucuronidase, beta pseudogene 2', other_designations='glucuronidase, beta pseudogene 4|glucuronidase, beta-like 1|spinal muscular atrophy candidate gene 3-like 2', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26871487, end_position_on_the_genomic_accession=26956554, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375709, CurrentID=0, Status='live', Symbol='LOC105375709', Aliases='', description='uncharacterized LOC105375709', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=113432178, end_position_on_the_genomic_accession=113433272, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375710, CurrentID=0, Status='live', Symbol='LOC105375710', Aliases='', description='uncharacterized LOC105375710', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=114294502, end_position_on_the_genomic_accession=114311356, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  255967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=255967, CurrentID=0, Status='live', Symbol='PAN3', Aliases='', description='PAN3 poly(A) specific ribonuclease subunit', other_designations='PAB-dependent poly(A)-specific ribonuclease subunit 3|PAB1P-dependent poly(A)-nuclease|PABP-dependent poly(A) nuclease 3|PABP1-dependent poly A-specific ribonuclease subunit PAN3|PAN deadenylation complex subunit 3|PAN3 poly(A) specific ribonuclease subunit homolog', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=28136717, end_position_on_the_genomic_accession=28295338, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  100132832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132832, CurrentID=0, Status='live', Symbol='PMS2P9', Aliases='PMS2L17, PMS2LP1, PMSR5', description='PMS1 homolog 2, mismatch repair system component pseudogene 9', other_designations='PMS2 postmeiotic segregation increased 2 pseudogene|postmeiotic segregation increased 2 pseudogene 9|postmeiotic segregation increased 2-like 17|postmeiotic segregation increased 2-like pseudogene 1', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=77039480, end_position_on_the_genomic_accession=77053038, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375713, CurrentID=0, Status='live', Symbol='LOC105375713', Aliases='', description='uncharacterized LOC105375713', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=116648933, end_position_on_the_genomic_accession=116654642, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  649186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649186, CurrentID=0, Status='live', Symbol='VN2R17P', Aliases='', description='vomeronasal 2 receptor 17 pseudogene', other_designations='calcium-sensing receptor pseudogene|seven transmembrane helix receptor', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56317283, end_position_on_the_genomic_accession=56317885, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387043, CurrentID=0, Status='live', Symbol='RPL13AP', Aliases='RPL13A_5_675, dJ271M21.3', description='ribosomal protein L13a pseudogene', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=29582508, end_position_on_the_genomic_accession=29583025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387044, CurrentID=0, Status='live', Symbol='RPL13P', Aliases='BBC1p, RPL13_3_673, dJ111M5.1', description='ribosomal protein L13 pseudogene', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28861400, end_position_on_the_genomic_accession=28862016, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375717, CurrentID=0, Status='live', Symbol='LOC105375717', Aliases='', description='uncharacterized LOC105375717', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117273754, end_position_on_the_genomic_accession=117311087, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  387046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387046, CurrentID=0, Status='live', Symbol='RPL8P1', Aliases='RPL2p, RPL8P, dJ15D7.1', description='ribosomal protein L8 pseudogene 1', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27652591, end_position_on_the_genomic_accession=27652733, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375719, CurrentID=0, Status='live', Symbol='LOC105375719', Aliases='', description='uncharacterized LOC105375719', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117096230, end_position_on_the_genomic_accession=117107679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375720, CurrentID=0, Status='live', Symbol='LOC105375720', Aliases='', description='uncharacterized LOC105375720', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117613500, end_position_on_the_genomic_accession=117632750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132841, CurrentID=0, Status='live', Symbol='NMNAT1P4', Aliases='', description='NMNAT1 pseudogene 4', other_designations='nicotinamide nucleotide adenylyltransferase 1 pseudogene', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145456005, end_position_on_the_genomic_accession=145456838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375722, CurrentID=0, Status='live', Symbol='LOC105375722', Aliases='', description='uncharacterized LOC105375722', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=117890464, end_position_on_the_genomic_accession=117894233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375723, CurrentID=0, Status='live', Symbol='LOC105375723', Aliases='', description='uncharacterized LOC105375723', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118833338, end_position_on_the_genomic_accession=118861895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375724, CurrentID=0, Status='live', Symbol='LOC105375724', Aliases='', description='uncharacterized LOC105375724', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=118281186, end_position_on_the_genomic_accession=118424994, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105375725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375725, CurrentID=0, Status='live', Symbol='LOC105375725', Aliases='', description='uncharacterized LOC105375725', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=119152052, end_position_on_the_genomic_accession=119203654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100313929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100313929, CurrentID=0, Status='live', Symbol='PIRC49', Aliases='', description='piwi-interacting RNA cluster 49', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387055, CurrentID=0, Status='live', Symbol='LINC00533', Aliases='C6orf38, C6orf39, bA373N24.1', description='long intergenic non-protein coding RNA 533', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  124912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124912, CurrentID=0, Status='live', Symbol='SPACA3', Aliases='ALLP17, CT54, LYC3, LYZL3, SLLP1', description='sperm acrosome associated 3', other_designations='cancer/testis antigen 54|lysozyme-like acrosomal sperm-specific secretory protein ALLP-17|lysozyme-like protein 3|lysozyme-like sperm-specific secretory protein ALLP17|sperm lysozyme like protein 1|sperm lysozyme-like protein 1|sperm lyzozyme-like acrosomal protein 1|sperm protein reactive with ASA|sperm protein reactive with antisperm antibodies', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=32991864, end_position_on_the_genomic_accession=32997874, orientation='plus', exon_count=5, OMIM=612749, no_hdr0=''),
  649201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649201, CurrentID=0, Status='live', Symbol='LOC649201', Aliases='', description='paraneoplastic antigen like 6A-like', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.10', start_position_on_the_genomic_accession=152584314, end_position_on_the_genomic_accession=152587591, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375730, CurrentID=0, Status='live', Symbol='LOC105375730', Aliases='', description='uncharacterized LOC105375730', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=120058882, end_position_on_the_genomic_accession=120107546, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375731, CurrentID=0, Status='live', Symbol='LOC105375731', Aliases='', description='uncharacterized LOC105375731', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121158673, end_position_on_the_genomic_accession=121162325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375732, CurrentID=0, Status='live', Symbol='LOC105375732', Aliases='', description='uncharacterized LOC105375732', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=121668951, end_position_on_the_genomic_accession=121684593, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375733, CurrentID=0, Status='live', Symbol='LOC105375733', Aliases='', description='uncharacterized LOC105375733', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122338736, end_position_on_the_genomic_accession=122402417, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375734, CurrentID=0, Status='live', Symbol='LOC105375734', Aliases='', description='uncharacterized LOC105375734', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122413690, end_position_on_the_genomic_accession=122568733, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  387063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387063, CurrentID=0, Status='live', Symbol='HMGA1P5', Aliases='HMGA1L5', description='high mobility group AT-hook 1 pseudogene 5', other_designations='high mobility group AT-hook 1-like 5', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75275136, end_position_on_the_genomic_accession=75276536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375736, CurrentID=0, Status='live', Symbol='LOC105375736', Aliases='', description='uncharacterized LOC105375736', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123215267, end_position_on_the_genomic_accession=123217228, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387065, CurrentID=0, Status='live', Symbol='HMGA1P7', Aliases='HMGA1L7', description='high mobility group AT-hook 1 pseudogene 7', other_designations='high mobility group AT-hook 1-like 7', map_location='6q23.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=134113962, end_position_on_the_genomic_accession=134117680, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387066, CurrentID=0, Status='live', Symbol='SNHG5', Aliases='C6orf160, LINC00044, NCRNA00044, U50HG, bA33E24.2', description='small nucleolar RNA host gene 5', other_designations='U50 host|long intergenic non-protein coding RNA 44|small nucleolar RNA host gene (non-protein coding) 5|small nucleolar RNA host gene 5 (non-protein coding)', map_location='6q14.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=85677007, end_position_on_the_genomic_accession=85678733, orientation='minus', exon_count=6, OMIM=613263, no_hdr0=''),
  124923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124923, CurrentID=0, Status='live', Symbol='SGK494', Aliases='', description='uncharacterized serine/threonine-protein kinase SgK494', other_designations='sugen kinase 494', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28607964, end_position_on_the_genomic_accession=28614193, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105375740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375740, CurrentID=0, Status='live', Symbol='LOC105375740', Aliases='', description='uncharacterized LOC105375740', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124568827, end_position_on_the_genomic_accession=124583975, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  124925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124925, CurrentID=0, Status='live', Symbol='SEZ6', Aliases='BSRPC', description='seizure related 6 homolog (mouse)', other_designations='SEZ-6|hSEZ-6', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28954905, end_position_on_the_genomic_accession=29006440, orientation='minus', exon_count=18, OMIM=616666, no_hdr0=''),
  105375742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375742, CurrentID=0, Status='live', Symbol='LOC105375742', Aliases='', description='uncharacterized LOC105375742', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124935130, end_position_on_the_genomic_accession=124941930, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100132863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132863, CurrentID=0, Status='live', Symbol='HMGB1P37', Aliases='', description='high mobility group box 1 pseudogene 37', other_designations='SP100 nuclear antigen pseudogene|high-mobility group box 1 pseudogene 37', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121519852, end_position_on_the_genomic_accession=121520409, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375744, CurrentID=0, Status='live', Symbol='LOC105375744', Aliases='', description='uncharacterized LOC105375744', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=124996988, end_position_on_the_genomic_accession=124998198, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375745, CurrentID=0, Status='live', Symbol='LOC105375745', Aliases='', description='uncharacterized LOC105375745', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125466957, end_position_on_the_genomic_accession=125624733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  124930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124930, CurrentID=0, Status='live', Symbol='ANKRD13B', Aliases='', description='ankyrin repeat domain 13B', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=29593509, end_position_on_the_genomic_accession=29614761, orientation='plus', exon_count=16, OMIM=615124, no_hdr0=''),
  105375747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375747, CurrentID=0, Status='live', Symbol='LOC105375747', Aliases='', description='uncharacterized LOC105375747', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125749055, end_position_on_the_genomic_accession=125749943, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375748, CurrentID=0, Status='live', Symbol='LOC105375748', Aliases='', description='uncharacterized LOC105375748', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=125859825, end_position_on_the_genomic_accession=125873101, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375749, CurrentID=0, Status='live', Symbol='LOC105375749', Aliases='', description='uncharacterized LOC105375749', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126497158, end_position_on_the_genomic_accession=126501053, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  256006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256006, CurrentID=0, Status='live', Symbol='ANKRD31', Aliases='', description='ankyrin repeat domain 31', other_designations='ankyrin repeat domain-containing protein 31', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75068297, end_position_on_the_genomic_accession=75236887, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  124935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124935, CurrentID=0, Status='live', Symbol='SLC43A2', Aliases='LAT4', description='solute carrier family 43 (amino acid system L transporter), member 2', other_designations='L-type amino acid transporter 4|amino acid transporter|solute carrier family 43, member 2', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1569254, end_position_on_the_genomic_accession=1630018, orientation='minus', exon_count=18, OMIM=610791, no_hdr0=''),
  124936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124936, CurrentID=0, Status='live', Symbol='CYB5D2', Aliases='', description='cytochrome b5 domain containing 2', other_designations='cytochrome b5 domain-containing protein 2', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4143168, end_position_on_the_genomic_accession=4157701, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375753, CurrentID=0, Status='live', Symbol='LOC105375753', Aliases='', description='uncharacterized LOC105375753', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=126750468, end_position_on_the_genomic_accession=126846987, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  387082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387082, CurrentID=0, Status='live', Symbol='SUMO4', Aliases='IDDM5, SMT3H4, SUMO-4, dJ281H8.4', description='small ubiquitin-like modifier 4', other_designations='SMT3 suppressor of mif two 3 homolog 2|SMT3 suppressor of mif two 3 homolog 4|insulin-dependent diabetes mellitus 5|small ubiquitin-like modifier 4 protein', map_location='6q25', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149400359, end_position_on_the_genomic_accession=149401046, orientation='plus', exon_count=1, OMIM=608829, no_hdr0=''),
  100421975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421975, CurrentID=0, Status='live', Symbol='CBX3P7', Aliases='', description='chromobox 3 pseudogene 7', other_designations='chromobox homolog 3 (HP1 gamma homolog, Drosophila) pseudogene|chromobox homolog 3 pseudogene 7', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89293744, end_position_on_the_genomic_accession=89294246, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375756, CurrentID=0, Status='live', Symbol='LOC105375756', Aliases='', description='uncharacterized LOC105375756', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=128799956, end_position_on_the_genomic_accession=128816667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  654338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654338, CurrentID=0, Status='live', Symbol='LOC654338', Aliases='', description='brain cytoplasmic RNA 1 (non-protein coding) pseudogene', other_designations='brain cytoplasmic RNA 1, pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13809557, end_position_on_the_genomic_accession=13809759, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375758, CurrentID=0, Status='live', Symbol='LOC105375758', Aliases='', description='uncharacterized LOC105375758', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130526944, end_position_on_the_genomic_accession=130549775, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375759, CurrentID=0, Status='live', Symbol='LOC105375759', Aliases='', description='uncharacterized LOC105375759', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=131096378, end_position_on_the_genomic_accession=131144908, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  124944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124944, CurrentID=0, Status='live', Symbol='C17orf49', Aliases='BAP18, HEPIS', description='chromosome 17 open reading frame 49', other_designations='BPTF associated protein of 18 kDa|BPTF-associated protein of 18 kDa|MLL1/MLL complex subunit C17orf49|human embryo lung cellular protein interacting with SARS-CoV nsp-10', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7014737, end_position_on_the_genomic_accession=7017524, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105375761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375761, CurrentID=0, Status='live', Symbol='LOC105375761', Aliases='', description='uncharacterized LOC105375761', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=130884567, end_position_on_the_genomic_accession=130907176, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387090, CurrentID=0, Status='live', Symbol='RPL36AP5', Aliases='RPL36A_9_691, bA256G5.1', description='ribosomal protein L36a pseudogene 5', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42499678, end_position_on_the_genomic_accession=42500079, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375763, CurrentID=0, Status='live', Symbol='LOC105375763', Aliases='', description='uncharacterized LOC105375763', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=131463041, end_position_on_the_genomic_accession=131470316, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375764, CurrentID=0, Status='live', Symbol='LOC105375764', Aliases='', description='uncharacterized LOC105375764', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132381541, end_position_on_the_genomic_accession=132384431, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256021, CurrentID=0, Status='live', Symbol='LINC01619', Aliases='C12orf79', description='long intergenic non-protein coding RNA 1619', other_designations='', map_location='12q21.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=91984976, end_position_on_the_genomic_accession=92142671, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375766, CurrentID=0, Status='live', Symbol='LOC105375766', Aliases='', description='uncharacterized LOC105375766', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132687462, end_position_on_the_genomic_accession=132707668, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375767, CurrentID=0, Status='live', Symbol='LOC105375767', Aliases='', description='uncharacterized LOC105375767', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=132675639, end_position_on_the_genomic_accession=132676075, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375768, CurrentID=0, Status='live', Symbol='LOC105375768', Aliases='', description='uncharacterized LOC105375768', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133013771, end_position_on_the_genomic_accession=133017586, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  654340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654340, CurrentID=0, Status='live', Symbol='RNPC3P1', Aliases='', description='RNA binding region (RNP1, RRM) containing 3 pseudogene 1', other_designations='RNA-binding region (RNP1, RRM) containing 3 pseudogene 1', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70124855, end_position_on_the_genomic_accession=70126656, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375770, CurrentID=0, Status='live', Symbol='LOC105375770', Aliases='', description='uncharacterized LOC105375770', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133326551, end_position_on_the_genomic_accession=133354334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132891, CurrentID=0, Status='live', Symbol='MSC-AS1', Aliases='', description='MSC antisense RNA 1', other_designations='', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71843123, end_position_on_the_genomic_accession=72056312, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375772, CurrentID=0, Status='live', Symbol='LOC105375772', Aliases='', description='uncharacterized LOC105375772', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133359189, end_position_on_the_genomic_accession=133361524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387101, CurrentID=0, Status='live', Symbol='RPL29P4', Aliases='RPL29P3, RPL29_6_745, bA632C17A.1', description='ribosomal protein L29 pseudogene 4', other_designations='', map_location='6q22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117998928, end_position_on_the_genomic_accession=117999582, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375774, CurrentID=0, Status='live', Symbol='LOC105375774', Aliases='', description='uncharacterized LOC105375774', other_designations='CTD-2342N23.3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=134719008, end_position_on_the_genomic_accession=134723907, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387103, CurrentID=0, Status='live', Symbol='CENPW', Aliases='C6orf173, CENP-W, CUG2', description='centromere protein W', other_designations='cancer-up-regulated gene 2 protein|cancer-upregulated gene 2', map_location='6q22.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126339789, end_position_on_the_genomic_accession=126483320, orientation='plus', exon_count=5, OMIM=611264, no_hdr0=''),
  387104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387104, CurrentID=0, Status='live', Symbol='SOGA3', Aliases='C6orf174, dJ403A15.3', description='SOGA family member 3', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127438406, end_position_on_the_genomic_accession=127519355, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  124961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124961, CurrentID=0, Status='live', Symbol='ZFP3', Aliases='ZNF752', description='ZFP3 zinc finger protein', other_designations='zfp-3|zinc finger protein 752|zinc finger protein homologous to Zfp-3 in mouse|zinc finger protein-3', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5078459, end_position_on_the_genomic_accession=5096374, orientation='plus', exon_count=2, OMIM=194480, no_hdr0=''),
  105375778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375778, CurrentID=0, Status='live', Symbol='LOC105375778', Aliases='', description='uncharacterized LOC105375778', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140349812, end_position_on_the_genomic_accession=140353348, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387107, CurrentID=0, Status='live', Symbol='ATP5J2P2', Aliases='bA249L21.4', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene 2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2 pseudogene 2|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit f, isoform 2 pseudogene 2|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene 2', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108907477, end_position_on_the_genomic_accession=108907896, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375780, CurrentID=0, Status='live', Symbol='LOC105375780', Aliases='', description='uncharacterized LOC105375780', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140007462, end_position_on_the_genomic_accession=140009425, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375781, CurrentID=0, Status='live', Symbol='LOC105375781', Aliases='', description='uncharacterized LOC105375781', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=139911238, end_position_on_the_genomic_accession=139917182, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375782, CurrentID=0, Status='live', Symbol='LOC105375782', Aliases='', description='uncharacterized LOC105375782', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=140464874, end_position_on_the_genomic_accession=140468915, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387111, CurrentID=0, Status='live', Symbol='LINC00222', Aliases='C6orf181, NCRNA00222, dJ354J5.2', description='long intergenic non-protein coding RNA 222', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108751654, end_position_on_the_genomic_accession=108769942, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375784, CurrentID=0, Status='live', Symbol='LOC105375784', Aliases='', description='uncharacterized LOC105375784', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141058453, end_position_on_the_genomic_accession=141062493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375785, CurrentID=0, Status='live', Symbol='LOC105375785', Aliases='', description='uncharacterized LOC105375785', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141194868, end_position_on_the_genomic_accession=141202539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100132906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132906, CurrentID=0, Status='live', Symbol='RPL13AP10', Aliases='RPL13A_2_89', description='ribosomal protein L13a pseudogene 10', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114001433, end_position_on_the_genomic_accession=114003343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375787, CurrentID=0, Status='live', Symbol='LOC105375787', Aliases='', description='uncharacterized LOC105375787', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141254649, end_position_on_the_genomic_accession=141256817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375788, CurrentID=0, Status='live', Symbol='LOC105375788', Aliases='', description='uncharacterized LOC105375788', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141507302, end_position_on_the_genomic_accession=141509650, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387117, CurrentID=0, Status='live', Symbol='UBQLN1P1', Aliases='UBQLN1P', description='ubiquilin 1 pseudogene 1', other_designations='', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30358594, end_position_on_the_genomic_accession=30364111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  124974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124974, CurrentID=0, Status='live', Symbol='TXNP4', Aliases='', description='thioredoxin pseudogene 4', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4572176, end_position_on_the_genomic_accession=4572660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  124975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124975, CurrentID=0, Status='live', Symbol='GGT6', Aliases='', description='gamma-glutamyltransferase 6', other_designations='gamma-glutamyltransferase 6 homolog|gamma-glutamyltranspeptidase 6|glutathione hydrolase 6', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4556927, end_position_on_the_genomic_accession=4560596, orientation='minus', exon_count=4, OMIM=612341, no_hdr0=''),
  124976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124976, CurrentID=0, Status='live', Symbol='SPNS2', Aliases='', description='spinster homolog 2 (Drosophila)', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4498834, end_position_on_the_genomic_accession=4539933, orientation='plus', exon_count=13, OMIM=612584, no_hdr0=''),
  105375793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375793, CurrentID=0, Status='live', Symbol='LOC105375793', Aliases='', description='uncharacterized LOC105375793', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142920130, end_position_on_the_genomic_accession=142925092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387122, CurrentID=0, Status='live', Symbol='WASF5P', Aliases='', description='WAS protein family member 5, pseudogene', other_designations='WAS protein family, member 3 pseudogene|WAS protein family, member 5, pseudogene', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31287385, end_position_on_the_genomic_accession=31289164, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256051, CurrentID=0, Status='live', Symbol='ZNF549', Aliases='', description='zinc finger protein 549', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57527081, end_position_on_the_genomic_accession=57540886, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132916, CurrentID=0, Status='live', Symbol='FAM159B', Aliases='', description='family with sequence similarity 159 member B', other_designations='family with sequence similarity 159, member B', map_location='5q12.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64690308, end_position_on_the_genomic_accession=64718190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132917, CurrentID=0, Status='live', Symbol='RAP1BP3', Aliases='', description='RAP1B, member of RAS oncogene family pseudogene 3', other_designations='', map_location='6q22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117431591, end_position_on_the_genomic_accession=117432053, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132918, CurrentID=0, Status='live', Symbol='EIF4EP3', Aliases='', description='eukaryotic translation initiation factor 4E pseudogene 3', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=182460220, end_position_on_the_genomic_accession=182461055, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375799, CurrentID=0, Status='live', Symbol='LOC105375799', Aliases='', description='uncharacterized LOC105375799', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143663118, end_position_on_the_genomic_accession=143668478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100132920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132920, CurrentID=0, Status='live', Symbol='CICP4', Aliases='', description='capicua transcriptional repressor pseudogene 4', other_designations='capicua homolog pseudogene 4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=64291211, end_position_on_the_genomic_accession=64293040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387129, CurrentID=0, Status='live', Symbol='NPSR1', Aliases='ASRT2, GPR154, GPRA, NPSR, PGR14, VRR1', description='neuropeptide S receptor 1', other_designations='G-protein coupled receptor 154|G-protein coupled receptor PGR14|G-protein coupled receptor for asthma susceptibility|vasopressin receptor-related receptor 1', map_location='7p14.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=34658239, end_position_on_the_genomic_accession=34878332, orientation='plus', exon_count=11, OMIM=608595, no_hdr0=''),
  100132922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132922, CurrentID=0, Status='live', Symbol='RPL6P11', Aliases='', description='ribosomal protein L6 pseudogene', other_designations='', map_location='4q32.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157968392, end_position_on_the_genomic_accession=157969562, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132923, CurrentID=0, Status='live', Symbol='FAM66D', Aliases='', description='family with sequence similarity 66 member D', other_designations='family with sequence similarity 66, member D', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12115782, end_position_on_the_genomic_accession=12151189, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100132924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132924, CurrentID=0, Status='live', Symbol='RPL13AP19', Aliases='RPL13A_6_1040', description='ribosomal protein L13a pseudogene 19', other_designations='', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48745202, end_position_on_the_genomic_accession=48752108, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  124989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124989, CurrentID=0, Status='live', Symbol='EFCAB13', Aliases='C17orf57', description='EF-hand calcium binding domain 13', other_designations='EF-hand domain-containing protein C17orf57', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47323961, end_position_on_the_genomic_accession=47441312, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  105375806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375806, CurrentID=0, Status='live', Symbol='LOC105375806', Aliases='', description='uncharacterized LOC105375806', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144854871, end_position_on_the_genomic_accession=144857698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375807, CurrentID=0, Status='live', Symbol='LOC105375807', Aliases='', description='ephrin type-B receptor 6-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375808, CurrentID=0, Status='live', Symbol='LOC105375808', Aliases='', description='uncharacterized LOC105375808', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100132929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132929, CurrentID=0, Status='live', Symbol='FAM25BP', Aliases='FAM25A, FAM25B, FAM25C, FAM25G', description='family with sequence similarity 25 member B, pseudogene', other_designations='family with sequence similarity 25, member B, pseudogene|protein FAM25B', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46369087, end_position_on_the_genomic_accession=46373563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375810, CurrentID=0, Status='live', Symbol='LOC105375810', Aliases='', description='uncharacterized LOC105375810', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  124995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124995, CurrentID=0, Status='live', Symbol='MRPL10', Aliases='L10MT, MRP-L10, MRP-L8, MRPL8, RPML8', description='mitochondrial ribosomal protein L10', other_designations='39S ribosomal protein L8, mitochondrial|L8mt', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47823272, end_position_on_the_genomic_accession=47831541, orientation='minus', exon_count=6, OMIM=611825, no_hdr0=''),
  100132932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132932, CurrentID=0, Status='live', Symbol='FAM197Y9', Aliases='', description='family with sequence similarity 197 Y-linked member 9', other_designations='family with sequence similarity 197, Y-linked, member 9', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=6130045, end_position_on_the_genomic_accession=6132240, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  124997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=124997, CurrentID=0, Status='live', Symbol='WDR81', Aliases='CAMRQ2, PPP1R166', description='WD repeat domain 81', other_designations='protein phosphatase 1, regulatory subunit 166', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1716523, end_position_on_the_genomic_accession=1738599, orientation='plus', exon_count=12, OMIM=614218, no_hdr0=''),
  105375814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375814, CurrentID=0, Status='live', Symbol='LOC105375814', Aliases='', description='uncharacterized LOC105375814', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46928860, end_position_on_the_genomic_accession=46951840, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375815, CurrentID=0, Status='live', Symbol='LOC105375815', Aliases='', description='uncharacterized LOC105375815', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47003712, end_position_on_the_genomic_accession=47084213, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  649288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649288, CurrentID=0, Status='live', Symbol='AK4P6', Aliases='', description='adenylate kinase 4 pseudogene 6', other_designations='adenylate kinase 3-like 1 pseudogene', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151027090, end_position_on_the_genomic_accession=151027617, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375817, CurrentID=0, Status='live', Symbol='LOC105375817', Aliases='', description='POTE ankyrin domain family member A-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47100920, end_position_on_the_genomic_accession=47168769, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100132938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132938, CurrentID=0, Status='live', Symbol='AQP7P4', Aliases='', description='aquaporin 7 pseudogene 4', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63133597, end_position_on_the_genomic_accession=63143456, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375819, CurrentID=0, Status='live', Symbol='LOC105375819', Aliases='', description='uncharacterized LOC105375819', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=47989028, end_position_on_the_genomic_accession=47999102, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  256076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256076, CurrentID=0, Status='live', Symbol='COL6A5', Aliases='COL29A1, VWA4', description='collagen, type VI, alpha 5', other_designations='collagen, type XXIX, alpha 1|von Willebrand factor A domain-containing protein 4', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=130345516, end_position_on_the_genomic_accession=130484846, orientation='plus', exon_count=44, OMIM=611916, no_hdr0=''),
  105375821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375821, CurrentID=0, Status='live', Symbol='LOC105375821', Aliases='', description='uncharacterized LOC105375821', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48422998, end_position_on_the_genomic_accession=48474576, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  649294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649294, CurrentID=0, Status='live', Symbol='LOC649294', Aliases='', description='uncharacterized LOC649294', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375823, CurrentID=0, Status='live', Symbol='LOC105375823', Aliases='', description='uncharacterized LOC105375823', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48597087, end_position_on_the_genomic_accession=48621018, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375824, CurrentID=0, Status='live', Symbol='LOC105375824', Aliases='', description='uncharacterized LOC105375824', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48627871, end_position_on_the_genomic_accession=48635955, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375825, CurrentID=0, Status='live', Symbol='LOC105375825', Aliases='', description='uncharacterized LOC105375825', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48780612, end_position_on_the_genomic_accession=48828608, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375826, CurrentID=0, Status='live', Symbol='LOC105375826', Aliases='', description='uncharacterized LOC105375826', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=49168089, end_position_on_the_genomic_accession=49193297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  649299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649299, CurrentID=0, Status='live', Symbol='RPL36AP51', Aliases='RPL36A_25_1638', description='ribosomal protein L36a pseudogene 51', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=21233306, end_position_on_the_genomic_accession=21233698, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100132948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132948, CurrentID=0, Status='live', Symbol='FAM27C', Aliases='FAM27A, FAM27A1, FAM27A3', description='family with sequence similarity 27 member C', other_designations='family with sequence similarity 27, member A|family with sequence similarity 27, member C', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61854084, end_position_on_the_genomic_accession=61855340, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  256085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256085, CurrentID=0, Status='live', Symbol='WRBP1', Aliases='', description='tryptophan rich basic protein pseudogene 1', other_designations='hCG1790950', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112693438, end_position_on_the_genomic_accession=112693791, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375830, CurrentID=0, Status='live', Symbol='LOC105375830', Aliases='', description='uncharacterized LOC105375830', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=50926492, end_position_on_the_genomic_accession=50934774, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375831, CurrentID=0, Status='live', Symbol='LOC105375831', Aliases='', description='uncharacterized LOC105375831', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51107069, end_position_on_the_genomic_accession=51113064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375832, CurrentID=0, Status='live', Symbol='LOC105375832', Aliases='', description='uncharacterized LOC105375832', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51209594, end_position_on_the_genomic_accession=51251028, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  649305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649305, CurrentID=0, Status='live', Symbol='LOC649305', Aliases='', description='uncharacterized LOC649305', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375834, CurrentID=0, Status='live', Symbol='LOC105375834', Aliases='', description='uncharacterized LOC105375834', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51647314, end_position_on_the_genomic_accession=51649283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375835, CurrentID=0, Status='live', Symbol='LOC105375835', Aliases='', description='uncharacterized LOC105375835', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52714343, end_position_on_the_genomic_accession=52778045, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105375836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375836, CurrentID=0, Status='live', Symbol='LOC105375836', Aliases='', description='uncharacterized LOC105375836', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53177571, end_position_on_the_genomic_accession=53228012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375837, CurrentID=0, Status='live', Symbol='LOC105375837', Aliases='', description='uncharacterized LOC105375837', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53276676, end_position_on_the_genomic_accession=53303362, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375838, CurrentID=0, Status='live', Symbol='LOC105375838', Aliases='', description='uncharacterized LOC105375838', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=53397360, end_position_on_the_genomic_accession=53399838, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375839, CurrentID=0, Status='live', Symbol='LOC105375839', Aliases='', description='uncharacterized LOC105375839', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54042987, end_position_on_the_genomic_accession=54045833, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375840, CurrentID=0, Status='live', Symbol='LOC105375840', Aliases='', description='uncharacterized LOC105375840', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54373207, end_position_on_the_genomic_accession=54395505, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100132961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132961, CurrentID=0, Status='live', Symbol='RPS21P2', Aliases='RPS21_1_254', description='ribosomal protein S21 pseudogene 2', other_designations='', map_location='2q12.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106288744, end_position_on_the_genomic_accession=106289100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375842, CurrentID=0, Status='live', Symbol='LOC105375842', Aliases='', description='uncharacterized LOC105375842', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54769810, end_position_on_the_genomic_accession=54775298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100132963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132963, CurrentID=0, Status='live', Symbol='SMIM9', Aliases='CXorf68', description='small integral membrane protein 9', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154823348, end_position_on_the_genomic_accession=154834662, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100652999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100652999, CurrentID=0, Status='live', Symbol='LOC100652999', Aliases='', description='uncharacterized LOC100652999', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53513984, end_position_on_the_genomic_accession=53517608, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100132965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132965, CurrentID=0, Status='live', Symbol='GYG1P1', Aliases='', description='glycogenin 1 pseudogene 1', other_designations='glycogenin 1 pseudogene 2', map_location='8q21.11', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73904599, end_position_on_the_genomic_accession=73915106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375846, CurrentID=0, Status='live', Symbol='LOC105375846', Aliases='', description='uncharacterized LOC105375846', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55695369, end_position_on_the_genomic_accession=55729495, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132967, CurrentID=0, Status='live', Symbol='FAM223A', Aliases='CXorf52, LINC00204A, NCRNA00204, NCRNA00204A, SPCX', description='family with sequence similarity 223 member A (non-protein coding)', other_designations='family with sequence similarity 223, member A (non-protein coding)|long intergenic non-protein coding RNA 204A', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154571248, end_position_on_the_genomic_accession=154571960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375848, CurrentID=0, Status='live', Symbol='LOC105375848', Aliases='', description='uncharacterized LOC105375848', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56436676, end_position_on_the_genomic_accession=56438667, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375849, CurrentID=0, Status='live', Symbol='LOC105375849', Aliases='', description='uncharacterized LOC105375849', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56394928, end_position_on_the_genomic_accession=56434882, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105375850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375850, CurrentID=0, Status='live', Symbol='LOC105375850', Aliases='', description='uncharacterized LOC105375850', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56565250, end_position_on_the_genomic_accession=56566703, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375851, CurrentID=0, Status='live', Symbol='LOC105375851', Aliases='', description='uncharacterized LOC105375851', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56638922, end_position_on_the_genomic_accession=56656474, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  649324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649324, CurrentID=0, Status='live', Symbol='PRAMEF34P', Aliases='', description='PRAME family member 34, pseudogene', other_designations='PRAME family member 1 pseudogene', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13094515, end_position_on_the_genomic_accession=13099717, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132973, CurrentID=0, Status='live', Symbol='TCEB1P29', Aliases='', description='transcription elongation factor B (SIII), polypeptide 1 pseudogene 29', other_designations='transcription elongation factor B (SIII), polypeptide 1 (15kDa, elongin C) pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9182169, end_position_on_the_genomic_accession=9183130, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375854, CurrentID=0, Status='live', Symbol='LOC105375854', Aliases='', description='uncharacterized LOC105375854', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57142676, end_position_on_the_genomic_accession=57154821, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375855, CurrentID=0, Status='live', Symbol='LOC105375855', Aliases='', description='uncharacterized LOC105375855', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57462714, end_position_on_the_genomic_accession=57552718, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375856, CurrentID=0, Status='live', Symbol='LOC105375856', Aliases='', description='uncharacterized LOC105375856', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=57855480, end_position_on_the_genomic_accession=57958510, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375857, CurrentID=0, Status='live', Symbol='LOC105375857', Aliases='', description='uncharacterized LOC105375857', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=58519023, end_position_on_the_genomic_accession=58527370, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  649330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649330, CurrentID=0, Status='live', Symbol='HNRNPCL3', Aliases='HNRNPCL2', description='heterogeneous nuclear ribonucleoprotein C-like 3', other_designations='heterogeneous nuclear ribonucleoprotein C-like 2|hnRNP C-like-2', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13061232, end_position_on_the_genomic_accession=13062237, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100132979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132979, CurrentID=0, Status='live', Symbol='GOLGA8DP', Aliases='GOLGA8D', description='golgin A8 family member D, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 8D|golgin A8 family, member D (pseudogene)|golgin A8 family, member D, pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23157340, end_position_on_the_genomic_accession=23170783, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105375860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375860, CurrentID=0, Status='live', Symbol='LOC105375860', Aliases='', description='uncharacterized LOC105375860', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=59593361, end_position_on_the_genomic_accession=59603073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375861, CurrentID=0, Status='live', Symbol='LOC105375861', Aliases='', description='uncharacterized LOC105375861', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60058460, end_position_on_the_genomic_accession=60116407, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105375862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375862, CurrentID=0, Status='live', Symbol='LOC105375862', Aliases='', description='uncharacterized LOC105375862', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60046783, end_position_on_the_genomic_accession=60050059, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375863, CurrentID=0, Status='live', Symbol='LOC105375863', Aliases='', description='uncharacterized LOC105375863', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60135864, end_position_on_the_genomic_accession=60138043, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375864, CurrentID=0, Status='live', Symbol='LOC105375864', Aliases='', description='uncharacterized LOC105375864', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60284982, end_position_on_the_genomic_accession=60385396, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105375865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375865, CurrentID=0, Status='live', Symbol='LOC105375865', Aliases='', description='uncharacterized LOC105375865', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60493757, end_position_on_the_genomic_accession=60514501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  125050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125050, CurrentID=0, Status='live', Symbol='RN7SK', Aliases='7SK', description='RNA, 7SK small nuclear', other_designations='7SK small nuclear RNA|RNA, 7SK, nuclear', map_location='6p12.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=52995620, end_position_on_the_genomic_accession=52995951, orientation='plus', exon_count=1, OMIM=606515, no_hdr0=''),
  100132987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132987, CurrentID=0, Status='live', Symbol='LINC00856', Aliases='', description='long intergenic non-protein coding RNA 856', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78248625, end_position_on_the_genomic_accession=78252645, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100132988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132988, CurrentID=0, Status='live', Symbol='RPL37AP3', Aliases='RPL37A_1_734', description='ribosomal protein L37a pseudogene 3', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108817680, end_position_on_the_genomic_accession=108817926, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375869, CurrentID=0, Status='live', Symbol='LOC105375869', Aliases='', description='uncharacterized LOC105375869', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60676848, end_position_on_the_genomic_accession=60679027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  256126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256126, CurrentID=0, Status='live', Symbol='SYCE2', Aliases='CESC1', description='synaptonemal complex central element protein 2', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12899076, end_position_on_the_genomic_accession=12919343, orientation='minus', exon_count=6, OMIM=611487, no_hdr0=''),
  105375871 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375871, CurrentID=0, Status='live', Symbol='LOC105375871', Aliases='', description='uncharacterized LOC105375871', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=61810480, end_position_on_the_genomic_accession=61825224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100132992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132992, CurrentID=0, Status='live', Symbol='HMGN1P30', Aliases='', description='high mobility group nucleosome binding domain 1 pseudogene 30', other_designations='high-mobility group nucleosome binding domain 1 pseudogene 30', map_location='18q21.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=58017657, end_position_on_the_genomic_accession=58019061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100132993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132993, CurrentID=0, Status='live', Symbol='RPS21P3', Aliases='RPS21_2_494', description='ribosomal protein S21 pseudogene 3', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=78768449, end_position_on_the_genomic_accession=78768812, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  125058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125058, CurrentID=0, Status='live', Symbol='TBC1D16', Aliases='', description='TBC1 domain family member 16', other_designations='TBC1 domain family, member 16', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=79932343, end_position_on_the_genomic_accession=80035867, orientation='minus', exon_count=15, OMIM=616637, no_hdr0=''),
  100132995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132995, CurrentID=0, Status='live', Symbol='RPS8P3', Aliases='HsT2707, RPS8_7_1602', description='ribosomal protein S8 pseudogene 3', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=51844950, end_position_on_the_genomic_accession=51845634, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375876, CurrentID=0, Status='live', Symbol='LOC105375876', Aliases='', description='uncharacterized LOC105375876', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=63687057, end_position_on_the_genomic_accession=63691935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  125061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125061, CurrentID=0, Status='live', Symbol='AFMID', Aliases='FKF, KF, KFA', description='arylformamidase', other_designations='FKF|KFA|KFase|N-formylkynurenine formamidase|probable arylformamidase', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78185102, end_position_on_the_genomic_accession=78207701, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105375878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375878, CurrentID=0, Status='live', Symbol='LOC105375878', Aliases='', description='uncharacterized LOC105375878', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64803890, end_position_on_the_genomic_accession=64817542, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375879, CurrentID=0, Status='live', Symbol='LOC105375879', Aliases='', description='uncharacterized LOC105375879', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=64798938, end_position_on_the_genomic_accession=64817458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  649352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649352, CurrentID=0, Status='live', Symbol='LOC649352', Aliases='', description='ubiquitin carboxyl-terminal hydrolase 17-like protein 2-like', other_designations='ubiquitin carboxyl-terminal hydrolase 17-like pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12379267, end_position_on_the_genomic_accession=12379789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375881, CurrentID=0, Status='live', Symbol='LOC105375881', Aliases='', description='uncharacterized LOC105375881', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65215978, end_position_on_the_genomic_accession=65220407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375882, CurrentID=0, Status='live', Symbol='LOC105375882', Aliases='', description='uncharacterized LOC105375882', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65721990, end_position_on_the_genomic_accession=65729414, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375883, CurrentID=0, Status='live', Symbol='LOC105375883', Aliases='', description='uncharacterized LOC105375883', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=65979831, end_position_on_the_genomic_accession=66023482, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105375884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375884, CurrentID=0, Status='live', Symbol='LOC105375884', Aliases='', description='uncharacterized LOC105375884', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=66613542, end_position_on_the_genomic_accession=66614828, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133005, CurrentID=0, Status='live', Symbol='RASA4DP', Aliases='', description='RAS p21 protein activator 4CD, pseudogene', other_designations='RAS p21 protein activator 4 pseudogene', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102678706, end_position_on_the_genomic_accession=102690471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375886, CurrentID=0, Status='live', Symbol='LOC105375886', Aliases='', description='uncharacterized LOC105375886', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=67556043, end_position_on_the_genomic_accession=67614760, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375887, CurrentID=0, Status='live', Symbol='LOC105375887', Aliases='', description='uncharacterized LOC105375887', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=68202201, end_position_on_the_genomic_accession=68220914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  256144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256144, CurrentID=0, Status='live', Symbol='OR4C3', Aliases='OR11-98', description='olfactory receptor family 4 subfamily C member 3', other_designations='olfactory receptor OR11-98|olfactory receptor, family 4, subfamily C, member 3', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48324941, end_position_on_the_genomic_accession=48325930, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375889, CurrentID=0, Status='live', Symbol='LOC105375889', Aliases='', description='uncharacterized LOC105375889', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=69711544, end_position_on_the_genomic_accession=69719959, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375890, CurrentID=0, Status='live', Symbol='LOC105375890', Aliases='', description='uncharacterized LOC105375890', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70525104, end_position_on_the_genomic_accession=70535939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375891, CurrentID=0, Status='live', Symbol='LOC105375891', Aliases='', description='uncharacterized LOC105375891', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70789356, end_position_on_the_genomic_accession=70796948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256148, CurrentID=0, Status='live', Symbol='OR4S1', Aliases='OR11-100', description='olfactory receptor family 4 subfamily S member 1', other_designations='olfactory receptor OR11-100|olfactory receptor, family 4, subfamily S, member 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48306223, end_position_on_the_genomic_accession=48307152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375893, CurrentID=0, Status='live', Symbol='LOC105375893', Aliases='', description='uncharacterized LOC105375893', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71510264, end_position_on_the_genomic_accession=71531863, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375894, CurrentID=0, Status='live', Symbol='LOC105375894', Aliases='', description='uncharacterized LOC105375894', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=71155444, end_position_on_the_genomic_accession=71199220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375895, CurrentID=0, Status='live', Symbol='LOC105375895', Aliases='', description='uncharacterized LOC105375895', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72117742, end_position_on_the_genomic_accession=72121729, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375896, CurrentID=0, Status='live', Symbol='LOC105375896', Aliases='', description='uncharacterized LOC105375896', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72194210, end_position_on_the_genomic_accession=72227255, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375897, CurrentID=0, Status='live', Symbol='LOC105375897', Aliases='', description='uncharacterized LOC105375897', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72516233, end_position_on_the_genomic_accession=72533346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375898, CurrentID=0, Status='live', Symbol='LOC105375898', Aliases='', description='uncharacterized LOC105375898', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=72749371, end_position_on_the_genomic_accession=72753570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375899, CurrentID=0, Status='live', Symbol='LOC105375899', Aliases='', description='uncharacterized LOC105375899', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73367893, end_position_on_the_genomic_accession=73370588, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375900, CurrentID=0, Status='live', Symbol='LOC105375900', Aliases='', description='uncharacterized LOC105375900', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73285830, end_position_on_the_genomic_accession=73290671, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375901, CurrentID=0, Status='live', Symbol='LOC105375901', Aliases='', description='uncharacterized LOC105375901', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=73224411, end_position_on_the_genomic_accession=73232366, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256158, CurrentID=0, Status='live', Symbol='HMCN2', Aliases='', description='hemicentin 2', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130265759, end_position_on_the_genomic_accession=130434123, orientation='plus', exon_count=100, OMIM='', no_hdr0=''),
  105375903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375903, CurrentID=0, Status='live', Symbol='LOC105375903', Aliases='', description='uncharacterized LOC105375903', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=74199394, end_position_on_the_genomic_accession=74208281, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133024, CurrentID=0, Status='live', Symbol='AMD1P1', Aliases='', description='adenosylmethionine decarboxylase 1 pseudogene 1', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20350052, end_position_on_the_genomic_accession=20351100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375905, CurrentID=0, Status='live', Symbol='LOC105375905', Aliases='', description='uncharacterized LOC105375905', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=75122360, end_position_on_the_genomic_accession=75136864, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375906, CurrentID=0, Status='live', Symbol='LOC105375906', Aliases='', description='uncharacterized LOC105375906', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=76044505, end_position_on_the_genomic_accession=76071389, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375907, CurrentID=0, Status='live', Symbol='LOC105375907', Aliases='', description='uncharacterized LOC105375907', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77000325, end_position_on_the_genomic_accession=77007572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375908, CurrentID=0, Status='live', Symbol='LOC105375908', Aliases='', description='uncharacterized LOC105375908', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77033426, end_position_on_the_genomic_accession=77052999, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133029, CurrentID=0, Status='live', Symbol='LOC100133029', Aliases='', description='sirtuin 5 pseudogene', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68043329, end_position_on_the_genomic_accession=68044234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375910, CurrentID=0, Status='live', Symbol='LOC105375910', Aliases='', description='uncharacterized LOC105375910', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=77505801, end_position_on_the_genomic_accession=77515343, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375911, CurrentID=0, Status='live', Symbol='LOC105375911', Aliases='', description='uncharacterized LOC105375911', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78406793, end_position_on_the_genomic_accession=78665979, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375912, CurrentID=0, Status='live', Symbol='LOC105375912', Aliases='', description='uncharacterized LOC105375912', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78903196, end_position_on_the_genomic_accession=78948907, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375913, CurrentID=0, Status='live', Symbol='LOC105375913', Aliases='', description='uncharacterized LOC105375913', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78938583, end_position_on_the_genomic_accession=78947675, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375914, CurrentID=0, Status='live', Symbol='LOC105375914', Aliases='', description='uncharacterized LOC105375914', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=78805216, end_position_on_the_genomic_accession=78937680, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375915, CurrentID=0, Status='live', Symbol='LOC105375915', Aliases='', description='uncharacterized LOC105375915', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=79282079, end_position_on_the_genomic_accession=79288805, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133036, CurrentID=0, Status='live', Symbol='FAM95B1', Aliases='', description='family with sequence similarity 95 member B1', other_designations='family with sequence similarity 95, member B1', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40323571, end_position_on_the_genomic_accession=40329220, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100133037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133037, CurrentID=0, Status='live', Symbol='RPL7P4', Aliases='Em:AB023049.2, RPL7_10_678', description='ribosomal protein L7 pseudogene 4', other_designations='', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30696746, end_position_on_the_genomic_accession=30697536, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133038, CurrentID=0, Status='live', Symbol='RPL18AP7', Aliases='RPL18A_2_364', description='ribosomal protein L18a pseudogene 7', other_designations='', map_location='3p22.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38526746, end_position_on_the_genomic_accession=38527362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392221, CurrentID=0, Status='live', Symbol='CHCHD2P10', Aliases='', description='coiled-coil-helix-coiled-coil-helix domain containing 2 pseudogene 10', other_designations='', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=54191285, end_position_on_the_genomic_accession=54192054, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133040, CurrentID=0, Status='live', Symbol='PPP1R26P5', Aliases='KIAA0649P5', description='protein phosphatase 1 regulatory subunit 26 pseudogene 5', other_designations='protein phosphatase 1, regulatory subunit 26 pseudogene 5', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=21341596, end_position_on_the_genomic_accession=21343569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375921, CurrentID=0, Status='live', Symbol='LOC105375921', Aliases='', description='uncharacterized LOC105375921', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80483811, end_position_on_the_genomic_accession=80485545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375922, CurrentID=0, Status='live', Symbol='LOC105375922', Aliases='', description='uncharacterized LOC105375922', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80440094, end_position_on_the_genomic_accession=80451051, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  649395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649395, CurrentID=0, Status='live', Symbol='YWHAEP1', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon pseudogene 1', other_designations='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64433693, end_position_on_the_genomic_accession=64435520, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100133044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133044, CurrentID=0, Status='live', Symbol='BCRP8', Aliases='', description='breakpoint cluster region pseudogene 8', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23179923, end_position_on_the_genomic_accession=23180172, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375925, CurrentID=0, Status='live', Symbol='LOC105375925', Aliases='', description='uncharacterized LOC105375925', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81333530, end_position_on_the_genomic_accession=81347136, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387254, CurrentID=0, Status='live', Symbol='SLC7A5P2', Aliases='IMAA, MMAA', description='solute carrier family 7 (amino acid transporter light chain, L system), member 5 pseudogene 2', other_designations='LAT1-3TM protein 2|SLC7A5 pseudogene|hLAT1-3TM|solute carrier family 7 (cationic amino acid transporter, y+ system), member 5 pseudogene 2', map_location='16p12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=21517893, end_position_on_the_genomic_accession=21520444, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  125111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125111, CurrentID=0, Status='live', Symbol='GJD3', Aliases='CX31.9, Cx30.2, GJA11, GJC1', description='gap junction protein delta 3', other_designations='connexin 31.9|connexin-31.9|gap junction alpha-11 protein|gap junction chi-1 protein|gap junction protein, delta 3, 31.9kDa', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40360652, end_position_on_the_genomic_accession=40364693, orientation='minus', exon_count=1, OMIM=607425, no_hdr0=''),
  105375928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375928, CurrentID=0, Status='live', Symbol='LOC105375928', Aliases='', description='uncharacterized LOC105375928', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=81842031, end_position_on_the_genomic_accession=81849812, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  125113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125113, CurrentID=0, Status='live', Symbol='KRT222', Aliases='KA21P, KRT222', description='keratin 222, type II', other_designations='keratin 222 pseudogene|keratin-222 pseudogene', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40655620, end_position_on_the_genomic_accession=40665164, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375930, CurrentID=0, Status='live', Symbol='LOC105375930', Aliases='', description='uncharacterized LOC105375930', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=82160784, end_position_on_the_genomic_accession=82179726, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  125115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125115, CurrentID=0, Status='live', Symbol='KRT40', Aliases='CK-40, K40, KA36', description='keratin 40, type I', other_designations='cytokeratin-40|keratin-40|type I hair keratin KA36', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40977716, end_position_on_the_genomic_accession=40987139, orientation='minus', exon_count=9, OMIM=616679, no_hdr0=''),
  105375932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375932, CurrentID=0, Status='live', Symbol='LOC105375932', Aliases='', description='uncharacterized LOC105375932', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=83674853, end_position_on_the_genomic_accession=83693285, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133053, CurrentID=0, Status='live', Symbol='CXorf51B', Aliases='', description='chromosome X open reading frame 51B', other_designations='', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=146809784, end_position_on_the_genomic_accession=146810411, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133054, CurrentID=0, Status='live', Symbol='RPS9P3', Aliases='RPS9_1_598', description='ribosomal protein S9 pseudogene 3', other_designations='', map_location='5q21.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98954369, end_position_on_the_genomic_accession=98955009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387263, CurrentID=0, Status='live', Symbol='C6orf120', Aliases='', description='chromosome 6 open reading frame 120', other_designations='', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169702161, end_position_on_the_genomic_accession=169706306, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387264, CurrentID=0, Status='live', Symbol='KRTAP5-1', Aliases='KRN1L, KRTAP5.1', description='keratin associated protein 5-1', other_designations='keratin-associated protein 5.1|ultrahigh sulfur keratin-associated protein 5.1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1584342, end_position_on_the_genomic_accession=1585283, orientation='minus', exon_count=1, OMIM=148022, no_hdr0=''),
  105375937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375937, CurrentID=0, Status='live', Symbol='LOC105375937', Aliases='', description='uncharacterized LOC105375937', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85495699, end_position_on_the_genomic_accession=85534714, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  387266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387266, CurrentID=0, Status='live', Symbol='KRTAP5-3', Aliases='KRTAP5-9, KRTAP5.3', description='keratin associated protein 5-3', other_designations='keratin-associated protein 5-9|keratin-associated protein 5.3|keratin-associated protein 5.9|ultrahigh sulfur keratin-associated protein 5.3', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1607565, end_position_on_the_genomic_accession=1608463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387267, CurrentID=0, Status='live', Symbol='KRTAP5-4', Aliases='KRTAP5.4', description='keratin associated protein 5-4', other_designations='keratin-associated protein 5.4|ultrahigh sulfur keratin-associated protein 5.4', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375940, CurrentID=0, Status='live', Symbol='LOC105375940', Aliases='', description='uncharacterized LOC105375940', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105375941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375941, CurrentID=0, Status='live', Symbol='LOC105375941', Aliases='', description='uncharacterized LOC105375941', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133062, CurrentID=0, Status='live', Symbol='FAM217AP1', Aliases='', description='family with sequence similarity 217 member A pseudogene 1', other_designations='', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=500270, end_position_on_the_genomic_accession=501493, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392225, CurrentID=0, Status='live', Symbol='PDCL3P1', Aliases='', description='phosducin-like 3 pseudogene 1', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=60387201, end_position_on_the_genomic_accession=60388192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133064, CurrentID=0, Status='live', Symbol='PPP1R26P2', Aliases='KIAA0649P2', description='protein phosphatase 1 regulatory subunit 26 pseudogene 2', other_designations='protein phosphatase 1, regulatory subunit 26 pseudogene 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18649828, end_position_on_the_genomic_accession=18653812, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387273, CurrentID=0, Status='live', Symbol='KRTAP5-10', Aliases='KRTAP5.10', description='keratin associated protein 5-10', other_designations='keratin-associated protein 5.10|ultrahigh sulfur keratin-associated protein 5.10', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71565563, end_position_on_the_genomic_accession=71566620, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105375946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375946, CurrentID=0, Status='live', Symbol='LOC105375946', Aliases='', description='uncharacterized LOC105375946', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=252825, end_position_on_the_genomic_accession=275616, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  387275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387275, CurrentID=0, Status='live', Symbol='KRTAP5-13P', Aliases='KRTAP5.P1, KRTAP5P1', description='keratin associated protein 5-13, pseudogene', other_designations='keratin associated protein 5 pseudogene 1', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71568552, end_position_on_the_genomic_accession=71569001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387276, CurrentID=0, Status='live', Symbol='KRTAP5-14P', Aliases='KRTAP5.P2, KRTAP5P2', description='keratin associated protein 5-14, pseudogene', other_designations='keratin associated protein 5 pseudogene 2', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71579512, end_position_on_the_genomic_accession=71579981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375949, CurrentID=0, Status='live', Symbol='LOC105375949', Aliases='', description='uncharacterized LOC105375949', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=805236, end_position_on_the_genomic_accession=826756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375950, CurrentID=0, Status='live', Symbol='LOC105375950', Aliases='', description='uncharacterized LOC105375950', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=971948, end_position_on_the_genomic_accession=976558, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375951, CurrentID=0, Status='live', Symbol='LOC105375951', Aliases='', description='uncharacterized LOC105375951', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1948313, end_position_on_the_genomic_accession=1962500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375952, CurrentID=0, Status='live', Symbol='LOC105375952', Aliases='', description='uncharacterized LOC105375952', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=1927369, end_position_on_the_genomic_accession=1930259, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387281, CurrentID=0, Status='live', Symbol='LCRB', Aliases='', description='locus control region, beta', other_designations='LCR-beta', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5275850, end_position_on_the_genomic_accession=5291371, orientation='minus', exon_count=0, OMIM=152424, no_hdr0=''),
  100133074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133074, CurrentID=0, Status='live', Symbol='SLC25A5P9', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 9', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75752290, end_position_on_the_genomic_accession=75753539, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105375955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375955, CurrentID=0, Status='live', Symbol='LOC105375955', Aliases='', description='uncharacterized LOC105375955', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2289117, end_position_on_the_genomic_accession=2296883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375956, CurrentID=0, Status='live', Symbol='LOC105375956', Aliases='', description='uncharacterized LOC105375956', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2407793, end_position_on_the_genomic_accession=2431600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375957, CurrentID=0, Status='live', Symbol='LOC105375957', Aliases='', description='uncharacterized LOC105375957', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2652927, end_position_on_the_genomic_accession=2667976, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105375958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375958, CurrentID=0, Status='live', Symbol='LOC105375958', Aliases='', description='uncharacterized LOC105375958', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3198916, end_position_on_the_genomic_accession=3203621, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133079, CurrentID=0, Status='live', Symbol='SLC25A1P2', Aliases='', description='solute carrier family 25 (mitochondrial carrier; citrate transporter), member 1 pseudogene 2', other_designations='', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=63397454, end_position_on_the_genomic_accession=63399592, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  125144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125144, CurrentID=0, Status='live', Symbol='LRRC75A-AS1', Aliases='C17orf45, C17orf76-AS1, FAM211A-AS1, NCRNA00188, TSAP19', description='LRRC75A antisense RNA 1', other_designations='FAM211A antisense RNA 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16438987, end_position_on_the_genomic_accession=16470648, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375961, CurrentID=0, Status='live', Symbol='LOC105375961', Aliases='', description='uncharacterized LOC105375961', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3463947, end_position_on_the_genomic_accession=3467001, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375962, CurrentID=0, Status='live', Symbol='LOC105375962', Aliases='', description='uncharacterized LOC105375962', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3791783, end_position_on_the_genomic_accession=3802626, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375963, CurrentID=0, Status='live', Symbol='LOC105375963', Aliases='', description='uncharacterized LOC105375963', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3806395, end_position_on_the_genomic_accession=3809509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375964, CurrentID=0, Status='live', Symbol='LOC105375964', Aliases='', description='uncharacterized LOC105375964', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=3971801, end_position_on_the_genomic_accession=3994658, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375965, CurrentID=0, Status='live', Symbol='LOC105375965', Aliases='', description='uncharacterized LOC105375965', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4347052, end_position_on_the_genomic_accession=4431600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  125150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125150, CurrentID=0, Status='live', Symbol='ZSWIM7', Aliases='SWS1', description='zinc finger, SWIM-type containing 7', other_designations='SWIM domain-containing and Srs2-interacting protein 1 homolog|SWIM-type zinc finger domain-containing protein 7', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=15976560, end_position_on_the_genomic_accession=15999692, orientation='minus', exon_count=5, OMIM=614535, no_hdr0=''),
  256223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256223, CurrentID=0, Status='live', Symbol='CDRT15L2', Aliases='', description='CMT1A duplicated region transcript 15-like 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20579724, end_position_on_the_genomic_accession=20580911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105375968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375968, CurrentID=0, Status='live', Symbol='LOC105375968', Aliases='', description='uncharacterized LOC105375968', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6704273, end_position_on_the_genomic_accession=6707780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375969, CurrentID=0, Status='live', Symbol='LOC105375969', Aliases='', description='uncharacterized LOC105375969', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7176077, end_position_on_the_genomic_accession=7210817, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105375970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375970, CurrentID=0, Status='live', Symbol='LOC105375970', Aliases='', description='uncharacterized LOC105375970', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7304570, end_position_on_the_genomic_accession=7344574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  256227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256227, CurrentID=0, Status='live', Symbol='STEAP1B', Aliases='', description='STEAP family member 1B', other_designations='STEAP family protein MGC87042', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22419444, end_position_on_the_genomic_accession=22500282, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105375972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375972, CurrentID=0, Status='live', Symbol='LOC105375972', Aliases='', description='uncharacterized LOC105375972', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=9799392, end_position_on_the_genomic_accession=9803795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133093, CurrentID=0, Status='live', Symbol='FAM25G', Aliases='FAM25A, FAM25B, FAM25C, bA301J7.4', description='family with sequence similarity 25 member G', other_designations='family with sequence similarity 25, member G|protein FAM25', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47487236, end_position_on_the_genomic_accession=47491700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649446, CurrentID=0, Status='live', Symbol='DLGAP1-AS1', Aliases='HsT914', description='DLGAP1 antisense RNA 1', other_designations='DLGAP1 antisense RNA 1 (non-protein coding)', map_location='18p11.31', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=3594114, end_position_on_the_genomic_accession=3597379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105375975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375975, CurrentID=0, Status='live', Symbol='LOC105375975', Aliases='', description='histone-lysine N-methyltransferase SETMAR-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=11610222, end_position_on_the_genomic_accession=11618749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375976, CurrentID=0, Status='live', Symbol='LOC105375976', Aliases='', description='uncharacterized LOC105375976', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=12098790, end_position_on_the_genomic_accession=12159115, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105375977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375977, CurrentID=0, Status='live', Symbol='LOC105375977', Aliases='', description='uncharacterized LOC105375977', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13440907, end_position_on_the_genomic_accession=13487526, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375978, CurrentID=0, Status='live', Symbol='LOC105375978', Aliases='', description='uncharacterized LOC105375978', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14588797, end_position_on_the_genomic_accession=14602832, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375979, CurrentID=0, Status='live', Symbol='LOC105375979', Aliases='', description='uncharacterized LOC105375979', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=14779929, end_position_on_the_genomic_accession=14795530, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  256236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256236, CurrentID=0, Status='live', Symbol='NAPSB', Aliases='NAP1L, NAP2, NAPBP, NAPSB', description='napsin B aspartic peptidase, pseudogene', other_designations='napsin 2|napsin A aspartic peptidase pseudogene|napsin B aspartic peptidase pseudogene|napsin B pseudogene|napsin-A pseudogene|pronapsin B', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50333800, end_position_on_the_genomic_accession=50344748, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105375981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375981, CurrentID=0, Status='live', Symbol='LOC105375981', Aliases='', description='uncharacterized LOC105375981', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=15147001, end_position_on_the_genomic_accession=15165105, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105375982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375982, CurrentID=0, Status='live', Symbol='LOC105375982', Aliases='', description='uncharacterized LOC105375982', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=16705522, end_position_on_the_genomic_accession=16722471, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133103, CurrentID=0, Status='live', Symbol='RPL7L1P13', Aliases='', description='ribosomal protein L7 like 1 pseudogene 13', other_designations='ribosomal protein L7-like 1 pseudogene 13', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110399764, end_position_on_the_genomic_accession=110400516, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  125168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125168, CurrentID=0, Status='live', Symbol='RIBIN', Aliases='', description='rRNA promoter binding protein', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  780529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780529, CurrentID=0, Status='live', Symbol='LOC780529', Aliases='', description='uncharacterized LOC780529', other_designations='', map_location='13q31', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  125170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125170, CurrentID=0, Status='live', Symbol='MIEF2', Aliases='MID49, SMCR7', description='mitochondrial elongation factor 2', other_designations='Smith-Magenis syndrome chromosomal region candidate gene 7 protein|Smith-Magenis syndrome chromosome region, candidate 7|mitochondrial dynamic protein MID49|mitochondrial dynamic protein of 49 kDa|mitochondrial dynamics protein of 49 kDa', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18260534, end_position_on_the_genomic_accession=18265781, orientation='plus', exon_count=5, OMIM=615498, no_hdr0=''),
  387315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387315, CurrentID=0, Status='live', Symbol='MTND2P1', Aliases='', description='mitochondrially encoded NADH:ubiquinone oxidoreductase core subunit 2 pseudogene 1', other_designations='MT-ND2 pseudogene 1|MTND2 pseudogene 1|NADH dehydrogenase 2 pseudogene 1', map_location='21q22.12', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=35891226, end_position_on_the_genomic_accession=35892151, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387316, CurrentID=0, Status='live', Symbol='VN1R10P', Aliases='FKSG83, VNR6I1P, b24o18.2, hs6V1-1p', description='vomeronasal 1 receptor 10 pseudogene', other_designations='vomeronasal 1 receptor 1 pseudogene|vomeronasal olfactory receptor, (chromosome 6) subtype I, member 1 pseudogene', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27324761, end_position_on_the_genomic_accession=27325963, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387317, CurrentID=0, Status='live', Symbol='VN1R11P', Aliases='VNR6I2P, dJ86C11.7, hs6V1-2p', description='vomeronasal 1 receptor 11 pseudogene', other_designations='vomeronasal 1 receptor 4 pseudogene', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27083292, end_position_on_the_genomic_accession=27084216, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375990, CurrentID=0, Status='live', Symbol='LOC105375990', Aliases='', description='uncharacterized LOC105375990', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=22917085, end_position_on_the_genomic_accession=22934992, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387319, CurrentID=0, Status='live', Symbol='HIST1H2APS1', Aliases='H2AFSP, bA317E16.5', description='histone cluster 1, H2a, pseudogene 1', other_designations='histone 1, H2a, pseudogene 1', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=25732433, end_position_on_the_genomic_accession=25732861, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387320, CurrentID=0, Status='live', Symbol='VN1R14P', Aliases='VNR6I5P, bA457M11.4, hs6M1-5p', description='vomeronasal 1 receptor 14 pseudogene', other_designations='', map_location='6p22-p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26631085, end_position_on_the_genomic_accession=26631736, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387321, CurrentID=0, Status='live', Symbol='VN1R12P', Aliases='VNR6I3P, bA373D17.2, hs6V1-3p', description='vomeronasal 1 receptor 12 pseudogene', other_designations='', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27033346, end_position_on_the_genomic_accession=27034343, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387322, CurrentID=0, Status='live', Symbol='VN1R13P', Aliases='VNR6I4P, bA373D17.3, hs6V1-4p', description='vomeronasal 1 receptor 13 pseudogene', other_designations='vomeronasal 1 receptor 1 pseudogene', map_location='6p21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=27060370, end_position_on_the_genomic_accession=27061267, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387323, CurrentID=0, Status='live', Symbol='HIST1H2APS3', Aliases='H2AFUP, dJ34B20.8', description='histone cluster 1, H2a, pseudogene 3', other_designations='histone 1, H2a, pseudogene 3', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26233122, end_position_on_the_genomic_accession=26233255, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375996, CurrentID=0, Status='live', Symbol='LOC105375996', Aliases='', description='uncharacterized LOC105375996', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=25541925, end_position_on_the_genomic_accession=25570050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387325, CurrentID=0, Status='live', Symbol='HIST1H1PS1', Aliases='H1F6P, dJ34B20.16', description='histone cluster 1, H1, pseudogene 1', other_designations='histone 1, H1, pseudogene 1', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26195566, end_position_on_the_genomic_accession=26195771, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375998, CurrentID=0, Status='live', Symbol='LOC105375998', Aliases='', description='uncharacterized LOC105375998', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26328496, end_position_on_the_genomic_accession=26346259, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105375999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375999, CurrentID=0, Status='live', Symbol='LOC105375999', Aliases='', description='uncharacterized LOC105375999', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=26410979, end_position_on_the_genomic_accession=26426874, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  387328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387328, CurrentID=0, Status='live', Symbol='ZNF322P1', Aliases='ZNF322, ZNF322B', description='zinc finger protein 322 pseudogene 1', other_designations='putative zinc finger protein 322B|zinc finger protein 322B', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97195351, end_position_on_the_genomic_accession=97200145, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133121, CurrentID=0, Status='live', Symbol='FAM27B', Aliases='FAM27A2', description='family with sequence similarity 27 member B', other_designations='family with sequence similarity 27, member B', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67725483, end_position_on_the_genomic_accession=67726743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133122, CurrentID=0, Status='live', Symbol='RBPJP7', Aliases='', description='RBPJ pseudogene 7', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62601052, end_position_on_the_genomic_accession=62603327, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133123, CurrentID=0, Status='live', Symbol='LINC01203', Aliases='', description='long intergenic non-protein coding RNA 1203', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13335241, end_position_on_the_genomic_accession=13341825, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  387332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387332, CurrentID=0, Status='live', Symbol='TBPL2', Aliases='TBP2, TRF3', description='TATA-box binding protein like 2', other_designations='TATA box binding protein like 2|TATA box-binding protein-related factor 3|TBP-like protein 2|TBP-related factor 3', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55414210, end_position_on_the_genomic_accession=55440545, orientation='minus', exon_count=7, OMIM=608964, no_hdr0=''),
  105376005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376005, CurrentID=0, Status='live', Symbol='LOC105376005', Aliases='', description='uncharacterized LOC105376005', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=29593124, end_position_on_the_genomic_accession=29597311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376006, CurrentID=0, Status='live', Symbol='LOC105376006', Aliases='', description='uncharacterized LOC105376006', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=29717574, end_position_on_the_genomic_accession=29729544, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376007, CurrentID=0, Status='live', Symbol='LOC105376007', Aliases='', description='uncharacterized LOC105376007', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=29832533, end_position_on_the_genomic_accession=29882712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133128, CurrentID=0, Status='live', Symbol='LOC100133128', Aliases='', description='Beta-defensin 108B-like', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.11', start_position_on_the_genomic_accession=9400867, end_position_on_the_genomic_accession=9405291, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376009, CurrentID=0, Status='live', Symbol='LOC105376009', Aliases='', description='uncharacterized LOC105376009', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31119066, end_position_on_the_genomic_accession=31128552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387338, CurrentID=0, Status='live', Symbol='NSUN4', Aliases='SHTAP', description='NOP2/Sun RNA methyltransferase family member 4', other_designations='NOL1/NOP2/Sun domain family member 4|NOP2/Sun domain family, member 4|putative methyltransferase NSUN4|sperm head and tail associated protein', map_location='1p34', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46340177, end_position_on_the_genomic_accession=46365152, orientation='plus', exon_count=7, OMIM=615394, no_hdr0=''),
  105376011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376011, CurrentID=0, Status='live', Symbol='LOC105376011', Aliases='', description='uncharacterized LOC105376011', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31882443, end_position_on_the_genomic_accession=31917022, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105376012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376012, CurrentID=0, Status='live', Symbol='LOC105376012', Aliases='', description='uncharacterized LOC105376012', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31848686, end_position_on_the_genomic_accession=31978375, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376013, CurrentID=0, Status='live', Symbol='LOC105376013', Aliases='', description='uncharacterized LOC105376013', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=31836543, end_position_on_the_genomic_accession=31849205, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376014, CurrentID=0, Status='live', Symbol='LOC105376014', Aliases='', description='uncharacterized LOC105376014', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32448445, end_position_on_the_genomic_accession=32454942, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376015, CurrentID=0, Status='live', Symbol='LOC105376015', Aliases='', description='uncharacterized LOC105376015', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32603907, end_position_on_the_genomic_accession=32609001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376016, CurrentID=0, Status='live', Symbol='LOC105376016', Aliases='', description='uncharacterized LOC105376016', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32738971, end_position_on_the_genomic_accession=32767437, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  649489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649489, CurrentID=0, Status='live', Symbol='PPP1R2P5', Aliases='', description='protein phosphatase 1 regulatory inhibitor subunit 2 pseudogene 5', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 2 pseudogene 5', map_location='2q12.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106940683, end_position_on_the_genomic_accession=106941682, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376018, CurrentID=0, Status='live', Symbol='LOC105376018', Aliases='', description='uncharacterized LOC105376018', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33348793, end_position_on_the_genomic_accession=33351595, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283011, CurrentID=0, Status='live', Symbol='FLJ37201', Aliases='', description='tigger transposable element derived 2 pseudogene', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89691300, end_position_on_the_genomic_accession=89697928, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376020, CurrentID=0, Status='live', Symbol='LOC105376020', Aliases='', description='uncharacterized LOC105376020', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33401679, end_position_on_the_genomic_accession=33410280, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376021, CurrentID=0, Status='live', Symbol='LOC105376021', Aliases='', description='uncharacterized LOC105376021', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33490737, end_position_on_the_genomic_accession=33491594, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  125206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125206, CurrentID=0, Status='live', Symbol='SLC5A10', Aliases='SGLT-5, SGLT5', description='solute carrier family 5 (sodium/sugar cotransporter), member 10', other_designations='Na(+)/glucose cotransporter 5|solute carrier family 5 (sodium/glucose cotransporter), member 10', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18950676, end_position_on_the_genomic_accession=19020691, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105376023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376023, CurrentID=0, Status='live', Symbol='LOC105376023', Aliases='', description='uncharacterized LOC105376023', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34568990, end_position_on_the_genomic_accession=34571027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  125208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125208, CurrentID=0, Status='live', Symbol='RPS2P46', Aliases='RPS2_22_1525', description='ribosomal protein S2 pseudogene 46', other_designations='', map_location='17p11', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19445929, end_position_on_the_genomic_accession=19446875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256281, CurrentID=0, Status='live', Symbol='NUDT14', Aliases='UGPP, UGPPase', description='nudix hydrolase 14', other_designations='UDP-sugar diphosphatase|UDPG pyrophosphatase|nudix (nucleoside diphosphate linked moiety X)-type motif 14', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105172937, end_position_on_the_genomic_accession=105181323, orientation='minus', exon_count=6, OMIM=609219, no_hdr0=''),
  105376026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376026, CurrentID=0, Status='live', Symbol='LOC105376026', Aliases='', description='uncharacterized LOC105376026', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35491136, end_position_on_the_genomic_accession=35491732, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376027, CurrentID=0, Status='live', Symbol='LOC105376027', Aliases='', description='uncharacterized LOC105376027', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35899270, end_position_on_the_genomic_accession=35904820, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376028, CurrentID=0, Status='live', Symbol='LOC105376028', Aliases='', description='uncharacterized LOC105376028', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36206881, end_position_on_the_genomic_accession=36214010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387357, CurrentID=0, Status='live', Symbol='THEMIS', Aliases='C6orf190, C6orf207, GASP, SPOT, TSEPA, bA325O24.3, bA325O24.4', description='thymocyte selection associated', other_designations='GRB2-associated protein|signaling phosphoprotein specific for T cells|thymocyte expressed molecule involved in selection|thymocyte selection pathway associated|thymocyte-expressed molecule involved in selection', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127708194, end_position_on_the_genomic_accession=127918631, orientation='minus', exon_count=10, OMIM=613607, no_hdr0=''),
  105376030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376030, CurrentID=0, Status='live', Symbol='LOC105376030', Aliases='', description='uncharacterized LOC105376030', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36779758, end_position_on_the_genomic_accession=36825728, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  387359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387359, CurrentID=0, Status='live', Symbol='USP9YP4', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 4', other_designations='USP9Y pseudogene 4|ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9189639, end_position_on_the_genomic_accession=9204613, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376032, CurrentID=0, Status='live', Symbol='LOC105376032', Aliases='', description='uncharacterized LOC105376032', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37002780, end_position_on_the_genomic_accession=37007642, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387361, CurrentID=0, Status='live', Symbol='USP9YP34', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 34', other_designations='ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17984249, end_position_on_the_genomic_accession=17993387, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387362, CurrentID=0, Status='live', Symbol='USP9YP1', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 1', other_designations='USP9Y pseudogene 1|ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18540894, end_position_on_the_genomic_accession=18544232, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387363, CurrentID=0, Status='live', Symbol='USP9YP2', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 2', other_designations='USP9Y pseudogene 2|ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yq11.221-q11.222', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18776175, end_position_on_the_genomic_accession=18779513, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387364, CurrentID=0, Status='live', Symbol='USP9YP3', Aliases='', description='ubiquitin specific peptidase 9, Y-linked pseudogene 3', other_designations='USP9Y pseudogene 3|ubiquitin specific protease 9, Y-linked pseudogene', map_location='Yq11.223', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21661826, end_position_on_the_genomic_accession=21674111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376037, CurrentID=0, Status='live', Symbol='LOC105376037', Aliases='', description='uncharacterized LOC105376037', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37789046, end_position_on_the_genomic_accession=37798671, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376038, CurrentID=0, Status='live', Symbol='LOC105376038', Aliases='', description='uncharacterized LOC105376038', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37942289, end_position_on_the_genomic_accession=37948490, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376039, CurrentID=0, Status='live', Symbol='LOC105376039', Aliases='', description='uncharacterized LOC105376039', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38070282, end_position_on_the_genomic_accession=38074639, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376040, CurrentID=0, Status='live', Symbol='LOC105376040', Aliases='', description='uncharacterized LOC105376040', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38354893, end_position_on_the_genomic_accession=38376369, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  256297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256297, CurrentID=0, Status='live', Symbol='PTF1A', Aliases='PACA, PAGEN2, PTF1-p48, bHLHa29', description='pancreas specific transcription factor, 1a', other_designations='bHLH transcription factor p48|class A basic helix-loop-helix protein 29|class II bHLH protein PTF1A|exocrine pancreas-specific transcription factor p48|p48 DNA-binding subunit of transcription factor PTF1|pancreas-specific transcription factor 1a', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23192531, end_position_on_the_genomic_accession=23194252, orientation='plus', exon_count=2, OMIM=607194, no_hdr0=''),
  100133162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133162, CurrentID=0, Status='live', Symbol='RPL32P34', Aliases='RPL32_22_1617', description='ribosomal protein L32 pseudogene 34', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5277734, end_position_on_the_genomic_accession=5278137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133163, CurrentID=0, Status='live', Symbol='BCRP7', Aliases='', description='breakpoint cluster region pseudogene 7', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18852284, end_position_on_the_genomic_accession=18858640, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  125228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125228, CurrentID=0, Status='live', Symbol='FAM210A', Aliases='C18orf19, HsT2329', description='family with sequence similarity 210 member A', other_designations='family with sequence similarity 210, member A|uncharacterized protein C18orf19', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13663347, end_position_on_the_genomic_accession=13726592, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  392242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392242, CurrentID=0, Status='live', Symbol='REXO1L8P', Aliases='REXO1L8', description='REX1, RNA exonuclease 1 homolog-like 8, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 8, pseudogene|REX1, RNA exonuclease 1 homolog-like 8', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85642725, end_position_on_the_genomic_accession=85644522, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256302, CurrentID=0, Status='live', Symbol='NATD1', Aliases='C17orf103, Gtlf3b', description='N-acetyltransferase domain containing 1', other_designations='N-acetyltransferase domain-containing protein 1|gene trap locus F3b|protein GTLF3B|transcript expressed during hematopoiesis 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21238870, end_position_on_the_genomic_accession=21253266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376047, CurrentID=0, Status='live', Symbol='LOC105376047', Aliases='', description='uncharacterized LOC105376047', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40106732, end_position_on_the_genomic_accession=40110189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133168, CurrentID=0, Status='live', Symbol='CCR12P', Aliases='GPR183P1', description='chemokine (C-C motif) receptor 12, pseudogene', other_designations='G protein-coupled receptor 183 pseudogene 1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99407926, end_position_on_the_genomic_accession=99408968, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376049, CurrentID=0, Status='live', Symbol='LOC105376049', Aliases='', description='uncharacterized LOC105376049', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39951809, end_position_on_the_genomic_accession=39962655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133170, CurrentID=0, Status='live', Symbol='RPL11P1', Aliases='RPL11_1_236', description='ribosomal protein L11 pseudogene 1', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64960008, end_position_on_the_genomic_accession=64960616, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654387, CurrentID=0, Status='live', Symbol='RPL31P6', Aliases='HsT19508, RPL31_30_1462', description='ribosomal protein L31 pseudogene 6', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92719431, end_position_on_the_genomic_accession=92719712, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133172, CurrentID=0, Status='live', Symbol='FAM66A', Aliases='', description='family with sequence similarity 66 member A', other_designations='family with sequence similarity 66, member A', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12362019, end_position_on_the_genomic_accession=12411001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  256309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256309, CurrentID=0, Status='live', Symbol='CCDC110', Aliases='CT52, KM-HN-1, KMHN1', description='coiled-coil domain containing 110', other_designations='cancer/testis antigen 52|cancer/testis antigen KM-HN-1', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185445182, end_position_on_the_genomic_accession=185471759, orientation='minus', exon_count=12, OMIM=609488, no_hdr0=''),
  105376054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376054, CurrentID=0, Status='live', Symbol='LOC105376054', Aliases='', description='uncharacterized LOC105376054', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40355509, end_position_on_the_genomic_accession=40363850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133175, CurrentID=0, Status='live', Symbol='LRLE1', Aliases='', description='liver-related low express protein 1', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=8046158, end_position_on_the_genomic_accession=8046331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376056, CurrentID=0, Status='live', Symbol='LOC105376056', Aliases='', description='putative zinc finger protein LOC441666', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=40708764, end_position_on_the_genomic_accession=40711900, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376057, CurrentID=0, Status='live', Symbol='LOC105376057', Aliases='', description='protein FRG1B-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41212196, end_position_on_the_genomic_accession=41219978, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  125242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125242, CurrentID=0, Status='live', Symbol='KRT18P8', Aliases='', description='keratin 18 pseudogene 8', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=9678242, end_position_on_the_genomic_accession=9678598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376059, CurrentID=0, Status='live', Symbol='LOC105376059', Aliases='', description='uncharacterized LOC105376059', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41025450, end_position_on_the_genomic_accession=41036403, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133180, CurrentID=0, Status='live', Symbol='DDX3P1', Aliases='DDX3YP1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, pseudogene 1', other_designations='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, Y-linked pseudogene 1', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74121043, end_position_on_the_genomic_accession=74139454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376061, CurrentID=0, Status='live', Symbol='LOC105376061', Aliases='', description='uncharacterized LOC105376061', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42569300, end_position_on_the_genomic_accession=42614282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376062, CurrentID=0, Status='live', Symbol='LOC105376062', Aliases='', description='uncharacterized LOC105376062', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42506454, end_position_on_the_genomic_accession=42507610, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376063, CurrentID=0, Status='live', Symbol='LOC105376063', Aliases='', description='uncharacterized LOC105376063', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41759330, end_position_on_the_genomic_accession=41764976, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376064, CurrentID=0, Status='live', Symbol='LOC105376064', Aliases='', description='uncharacterized LOC105376064', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41683259, end_position_on_the_genomic_accession=41699071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376065, CurrentID=0, Status='live', Symbol='LOC105376065', Aliases='', description='uncharacterized LOC105376065', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41500064, end_position_on_the_genomic_accession=41565697, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376066, CurrentID=0, Status='live', Symbol='LOC105376066', Aliases='', description='uncharacterized LOC105376066', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41356260, end_position_on_the_genomic_accession=41438754, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105376067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376067, CurrentID=0, Status='live', Symbol='LOC105376067', Aliases='', description='uncharacterized LOC105376067', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=43092108, end_position_on_the_genomic_accession=43093868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376068, CurrentID=0, Status='live', Symbol='LOC105376068', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=38620801, end_position_on_the_genomic_accession=38622698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376069, CurrentID=0, Status='live', Symbol='LOC105376069', Aliases='', description='uncharacterized LOC105376069', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68466271, end_position_on_the_genomic_accession=68468985, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133190, CurrentID=0, Status='live', Symbol='ADIRF-AS1', Aliases='', description='ADIRF antisense RNA 1', other_designations='', map_location='10q23.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.10', start_position_on_the_genomic_accession=88725498, end_position_on_the_genomic_accession=88731068, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376071, CurrentID=0, Status='live', Symbol='LOC105376071', Aliases='', description='uncharacterized LOC105376071', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68626284, end_position_on_the_genomic_accession=68635090, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376072, CurrentID=0, Status='live', Symbol='LOC105376072', Aliases='', description='uncharacterized LOC105376072', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=68541047, end_position_on_the_genomic_accession=68644441, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  256329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256329, CurrentID=0, Status='live', Symbol='LMNTD2', Aliases='C11orf35', description='lamin tail domain containing 2', other_designations='', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=554850, end_position_on_the_genomic_accession=560780, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105376074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376074, CurrentID=0, Status='live', Symbol='LOC105376074', Aliases='', description='uncharacterized LOC105376074', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69392263, end_position_on_the_genomic_accession=69399279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376075, CurrentID=0, Status='live', Symbol='LOC105376075', Aliases='', description='uncharacterized LOC105376075', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69756580, end_position_on_the_genomic_accession=69759709, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376076, CurrentID=0, Status='live', Symbol='LOC105376076', Aliases='', description='uncharacterized LOC105376076', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69760081, end_position_on_the_genomic_accession=69799066, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376077, CurrentID=0, Status='live', Symbol='LOC105376077', Aliases='', description='uncharacterized LOC105376077', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=69817191, end_position_on_the_genomic_accession=69820640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376078, CurrentID=0, Status='live', Symbol='LOC105376078', Aliases='', description='uncharacterized LOC105376078', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70671066, end_position_on_the_genomic_accession=70676221, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376079, CurrentID=0, Status='live', Symbol='LOC105376079', Aliases='', description='uncharacterized LOC105376079', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70466030, end_position_on_the_genomic_accession=70470570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100526416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526416, CurrentID=0, Status='live', Symbol='AK4P5', Aliases='', description='adenylate kinase 4 pseudogene 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80077729, end_position_on_the_genomic_accession=80078362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376081, CurrentID=0, Status='live', Symbol='LOC105376081', Aliases='', description='uncharacterized LOC105376081', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73090355, end_position_on_the_genomic_accession=73113916, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376082, CurrentID=0, Status='live', Symbol='LOC105376082', Aliases='', description='uncharacterized LOC105376082', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=73240387, end_position_on_the_genomic_accession=73259135, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100133203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133203, CurrentID=0, Status='live', Symbol='RAB28P1', Aliases='', description='RAB28, member RAS oncogene family pseudogene 1', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39735288, end_position_on_the_genomic_accession=39736466, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133204, CurrentID=0, Status='live', Symbol='C9orf147', Aliases='bA32M23.3', description='chromosome 9 open reading frame 147', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112469663, end_position_on_the_genomic_accession=112490453, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100133205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133205, CurrentID=0, Status='live', Symbol='LINC00240', Aliases='C6orf41, NCRNA00240, bA373D17.1', description='long intergenic non-protein coding RNA 240', other_designations='', map_location='6p22.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=26956993, end_position_on_the_genomic_accession=27023974, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133206, CurrentID=0, Status='live', Symbol='RBM22P5', Aliases='', description='RNA binding motif protein 22 pseudogene 5', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76285868, end_position_on_the_genomic_accession=76288148, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376087, CurrentID=0, Status='live', Symbol='LOC105376087', Aliases='', description='uncharacterized LOC105376087', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74345606, end_position_on_the_genomic_accession=74371417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376088 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376088, CurrentID=0, Status='live', Symbol='LOC105376088', Aliases='', description='uncharacterized LOC105376088', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74630706, end_position_on_the_genomic_accession=74642433, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376089, CurrentID=0, Status='live', Symbol='LOC105376089', Aliases='', description='uncharacterized LOC105376089', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=74921137, end_position_on_the_genomic_accession=74927985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133210, CurrentID=0, Status='live', Symbol='COX6B1P7', Aliases='', description='cytochrome c oxidase subunit VIb polypeptide 1 (ubiquitous) pseudogene 7', other_designations='', map_location='1p31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=68282368, end_position_on_the_genomic_accession=68282620, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376091, CurrentID=0, Status='live', Symbol='LOC105376091', Aliases='', description='uncharacterized LOC105376091', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=75852853, end_position_on_the_genomic_accession=75879360, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133212, CurrentID=0, Status='live', Symbol='ABHD17AP7', Aliases='FAM108A8P', description='abhydrolase domain containing 17A pseudogene 7', other_designations='family with sequence similarity 108, member A8, pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32726611, end_position_on_the_genomic_accession=32729550, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376093, CurrentID=0, Status='live', Symbol='LOC105376093', Aliases='', description='uncharacterized LOC105376093', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76515628, end_position_on_the_genomic_accession=76565728, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376094, CurrentID=0, Status='live', Symbol='LOC105376094', Aliases='', description='uncharacterized LOC105376094', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76644891, end_position_on_the_genomic_accession=76678115, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376095, CurrentID=0, Status='live', Symbol='LOC105376095', Aliases='', description='uncharacterized LOC105376095', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76605855, end_position_on_the_genomic_accession=76644806, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376096, CurrentID=0, Status='live', Symbol='LOC105376096', Aliases='', description='uncharacterized LOC105376096', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=77034899, end_position_on_the_genomic_accession=77055130, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376097, CurrentID=0, Status='live', Symbol='LOC105376097', Aliases='', description='uncharacterized LOC105376097', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=78861067, end_position_on_the_genomic_accession=78881933, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376098, CurrentID=0, Status='live', Symbol='LOC105376098', Aliases='', description='uncharacterized LOC105376098', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79728664, end_position_on_the_genomic_accession=79733869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256355, CurrentID=0, Status='live', Symbol='RPS2P32', Aliases='RPS2_14_794', description='ribosomal protein S2 pseudogene 32', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23490388, end_position_on_the_genomic_accession=23491428, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  256356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256356, CurrentID=0, Status='live', Symbol='GK5', Aliases='', description='glycerol kinase 5 (putative)', other_designations='ATP:glycerol 3-phosphotransferase 5|GK 5|glycerokinase 5', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=142157527, end_position_on_the_genomic_accession=142225607, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  105376101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376101, CurrentID=0, Status='live', Symbol='LOC105376101', Aliases='', description='uncharacterized LOC105376101', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79862207, end_position_on_the_genomic_accession=79887883, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100133222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133222, CurrentID=0, Status='live', Symbol='RPL39P27', Aliases='RPL39_13_1220', description='ribosomal protein L39 pseudogene 27', other_designations='', map_location='12p12.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=25409113, end_position_on_the_genomic_accession=25409494, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376103, CurrentID=0, Status='live', Symbol='LOC105376103', Aliases='', description='uncharacterized LOC105376103', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80360125, end_position_on_the_genomic_accession=80451467, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376104, CurrentID=0, Status='live', Symbol='LOC105376104', Aliases='', description='uncharacterized LOC105376104', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80617999, end_position_on_the_genomic_accession=80642381, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376105, CurrentID=0, Status='live', Symbol='LOC105376105', Aliases='', description='uncharacterized LOC105376105', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=80870110, end_position_on_the_genomic_accession=80871206, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376106, CurrentID=0, Status='live', Symbol='LOC105376106', Aliases='', description='uncharacterized LOC105376106', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81587788, end_position_on_the_genomic_accession=81590158, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376107, CurrentID=0, Status='live', Symbol='LOC105376107', Aliases='', description='uncharacterized LOC105376107', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81977177, end_position_on_the_genomic_accession=82426373, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  256364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256364, CurrentID=0, Status='live', Symbol='EML3', Aliases='ELP95', description='echinoderm microtubule associated protein like 3', other_designations='EMAP-3', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62602218, end_position_on_the_genomic_accession=62612765, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  105376109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376109, CurrentID=0, Status='live', Symbol='LOC105376109', Aliases='', description='uncharacterized LOC105376109', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81877082, end_position_on_the_genomic_accession=81911614, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376110, CurrentID=0, Status='live', Symbol='LOC105376110', Aliases='', description='uncharacterized LOC105376110', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82531052, end_position_on_the_genomic_accession=82540408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376111, CurrentID=0, Status='live', Symbol='LOC105376111', Aliases='', description='uncharacterized LOC105376111', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82451611, end_position_on_the_genomic_accession=82455231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376112, CurrentID=0, Status='live', Symbol='LOC105376112', Aliases='', description='uncharacterized LOC105376112', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83065213, end_position_on_the_genomic_accession=83074143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  256369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256369, CurrentID=0, Status='live', Symbol='LINC00521', Aliases='C14orf48, c14_5713', description='long intergenic non-protein coding RNA 521', other_designations='', map_location='14q32.12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=93997270, end_position_on_the_genomic_accession=94011695, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100133234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133234, CurrentID=0, Status='live', Symbol='SBF1P1', Aliases='', description='SET binding factor 1 pseudogene 1', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=55449197, end_position_on_the_genomic_accession=55455321, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100133235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133235, CurrentID=0, Status='live', Symbol='RPL26P14', Aliases='RPL26_7_282', description='ribosomal protein L26 pseudogene 14', other_designations='', map_location='2q22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=147801847, end_position_on_the_genomic_accession=147810564, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376116, CurrentID=0, Status='live', Symbol='LOC105376116', Aliases='', description='uncharacterized LOC105376116', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84278313, end_position_on_the_genomic_accession=84290136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376117, CurrentID=0, Status='live', Symbol='LOC105376117', Aliases='', description='uncharacterized LOC105376117', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=84825479, end_position_on_the_genomic_accession=84832791, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376118, CurrentID=0, Status='live', Symbol='LOC105376118', Aliases='', description='uncharacterized LOC105376118', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85030441, end_position_on_the_genomic_accession=85040934, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376119, CurrentID=0, Status='live', Symbol='LOC105376119', Aliases='', description='uncharacterized LOC105376119', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85261554, end_position_on_the_genomic_accession=85289479, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376120, CurrentID=0, Status='live', Symbol='LOC105376120', Aliases='', description='uncharacterized LOC105376120', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85482160, end_position_on_the_genomic_accession=85485364, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376121, CurrentID=0, Status='live', Symbol='LOC105376121', Aliases='', description='uncharacterized LOC105376121', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85401359, end_position_on_the_genomic_accession=85479962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376122, CurrentID=0, Status='live', Symbol='LOC105376122', Aliases='', description='uncharacterized LOC105376122', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85526167, end_position_on_the_genomic_accession=85529187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376123, CurrentID=0, Status='live', Symbol='LOC105376123', Aliases='', description='uncharacterized LOC105376123', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86590919, end_position_on_the_genomic_accession=86595010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256380, CurrentID=0, Status='live', Symbol='SCML4', Aliases='dJ47M23.1', description='sex comb on midleg-like 4 (Drosophila)', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=107697297, end_position_on_the_genomic_accession=107848659, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105376125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376125, CurrentID=0, Status='live', Symbol='LOC105376125', Aliases='', description='uncharacterized LOC105376125', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86751037, end_position_on_the_genomic_accession=86758048, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376126, CurrentID=0, Status='live', Symbol='LOC105376126', Aliases='', description='uncharacterized LOC105376126', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87220388, end_position_on_the_genomic_accession=87280515, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376127, CurrentID=0, Status='live', Symbol='LOC105376127', Aliases='', description='uncharacterized LOC105376127', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87384846, end_position_on_the_genomic_accession=87386759, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376128 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376128, CurrentID=0, Status='live', Symbol='LOC105376128', Aliases='', description='uncharacterized LOC105376128', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87751753, end_position_on_the_genomic_accession=87759256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100327036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100327036, CurrentID=0, Status='live', Symbol='PPP2R2DP1', Aliases='', description='protein phosphatase 2 regulatory subunit B, delta pseudogene 1', other_designations='protein phosphatase 2, regulatory subunit B, delta pseudogene 1', map_location='3p', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38051278, end_position_on_the_genomic_accession=38052400, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376130, CurrentID=0, Status='live', Symbol='LOC105376130', Aliases='', description='uncharacterized LOC105376130', other_designations='XXyac-YM21GA2.3', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87847983, end_position_on_the_genomic_accession=87853395, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100133251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133251, CurrentID=0, Status='live', Symbol='PRR23D2', Aliases='', description='proline rich 23 domain containing 2', other_designations='proline-rich protein 23C-like', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7778591, end_position_on_the_genomic_accession=7784165, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376132, CurrentID=0, Status='live', Symbol='LOC105376132', Aliases='', description='uncharacterized LOC105376132', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87967134, end_position_on_the_genomic_accession=87968871, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376133, CurrentID=0, Status='live', Symbol='LOC105376133', Aliases='', description='uncharacterized LOC105376133', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88113637, end_position_on_the_genomic_accession=88179227, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376134, CurrentID=0, Status='live', Symbol='LOC105376134', Aliases='', description='uncharacterized LOC105376134', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=88527960, end_position_on_the_genomic_accession=88534100, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133255, CurrentID=0, Status='live', Symbol='RPS14P5', Aliases='RPS14_3_249', description='ribosomal protein S14 pseudogene 5', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87654855, end_position_on_the_genomic_accession=87655376, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376136, CurrentID=0, Status='live', Symbol='LOC105376136', Aliases='', description='uncharacterized LOC105376136', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89288578, end_position_on_the_genomic_accession=89311105, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376137, CurrentID=0, Status='live', Symbol='LOC105376137', Aliases='', description='uncharacterized LOC105376137', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89527557, end_position_on_the_genomic_accession=89556920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  256394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256394, CurrentID=0, Status='live', Symbol='SERPINA11', Aliases='', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 11', other_designations='antiproteinase-like 2|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 11', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94442464, end_position_on_the_genomic_accession=94452850, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100133259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133259, CurrentID=0, Status='live', Symbol='CBX1P5', Aliases='CBX9', description='chromobox 1 pseudogene 5', other_designations='chromobox homolog 1 (HP1 beta homolog Drosophila ) pseudogene|chromobox homolog 1 pseudogene 5|chromobox homolog 9', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=166810829, end_position_on_the_genomic_accession=166812743, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376140, CurrentID=0, Status='live', Symbol='LOC105376140', Aliases='', description='uncharacterized LOC105376140', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=89696901, end_position_on_the_genomic_accession=89702412, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133261, CurrentID=0, Status='live', Symbol='LOC100133261', Aliases='', description='cysteine and histidine rich domain containing 1 pseudogene', other_designations='cysteine and histidine-rich domain (CHORD) containing 1 pseudogene', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164725533, end_position_on_the_genomic_accession=164727661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133262, CurrentID=0, Status='live', Symbol='ATP6V0E1P3', Aliases='', description='ATPase, H+ transporting, lysosomal 9kDa, V0 subunit e1 pseudogene 3', other_designations='', map_location='2p22.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=33601983, end_position_on_the_genomic_accession=33602785, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376143, CurrentID=0, Status='live', Symbol='LOC105376143', Aliases='', description='uncharacterized LOC105376143', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90931901, end_position_on_the_genomic_accession=90934309, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376144, CurrentID=0, Status='live', Symbol='LOC105376144', Aliases='', description='uncharacterized LOC105376144', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90997707, end_position_on_the_genomic_accession=91001881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133265, CurrentID=0, Status='live', Symbol='BMS1P7', Aliases='BMS1LP7, bA164N7.1', description='BMS1, ribosome biogenesis factor pseudogene 7', other_designations='BMS1 pseudogene 7|BMS1 ribosome biogenesis factor pseudogene 7|BMS1L pseudogene 7', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48050178, end_position_on_the_genomic_accession=48060116, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376146, CurrentID=0, Status='live', Symbol='LOC105376146', Aliases='', description='uncharacterized LOC105376146', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91424704, end_position_on_the_genomic_accession=91427140, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133267, CurrentID=0, Status='live', Symbol='LOC100133267', Aliases='', description='defensin, beta 130-like', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12064389, end_position_on_the_genomic_accession=12071747, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100133268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133268, CurrentID=0, Status='live', Symbol='HAUS6P2', Aliases='', description='HAUS augmin like complex subunit 6 pseudogene 2', other_designations='HAUS augmin-like complex, subunit 6 pseudogene 2', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31341374, end_position_on_the_genomic_accession=31344854, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376149, CurrentID=0, Status='live', Symbol='LOC105376149', Aliases='', description='uncharacterized LOC105376149', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92621621, end_position_on_the_genomic_accession=92631713, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376150, CurrentID=0, Status='live', Symbol='LOC105376150', Aliases='', description='uncharacterized LOC105376150', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93133012, end_position_on_the_genomic_accession=93148244, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376151, CurrentID=0, Status='live', Symbol='LOC105376151', Aliases='', description='uncharacterized LOC105376151', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=93321231, end_position_on_the_genomic_accession=93327581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  125336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125336, CurrentID=0, Status='live', Symbol='LOXHD1', Aliases='DFNB77, LH2D1', description='lipoxygenase homology domains 1', other_designations='', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=46476961, end_position_on_the_genomic_accession=46657114, orientation='minus', exon_count=48, OMIM=613072, no_hdr0=''),
  100133273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133273, CurrentID=0, Status='live', Symbol='RPL21P80', Aliases='RPL21_32_892', description='ribosomal protein L21 pseudogene 80', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34874120, end_position_on_the_genomic_accession=34874680, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376154, CurrentID=0, Status='live', Symbol='LOC105376154', Aliases='', description='uncharacterized LOC105376154', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=94309961, end_position_on_the_genomic_accession=94323998, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376155, CurrentID=0, Status='live', Symbol='LOC105376155', Aliases='', description='uncharacterized LOC105376155', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95355308, end_position_on_the_genomic_accession=95367189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376156, CurrentID=0, Status='live', Symbol='LOC105376156', Aliases='', description='uncharacterized LOC105376156', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95387442, end_position_on_the_genomic_accession=95426832, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376157, CurrentID=0, Status='live', Symbol='LOC105376157', Aliases='', description='uncharacterized LOC105376157', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95523219, end_position_on_the_genomic_accession=95550471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  387486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387486, CurrentID=0, Status='live', Symbol='LINC00320', Aliases='C21orf131, NCRNA00320, PRED14', description='long intergenic non-protein coding RNA 320', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=20742590, end_position_on_the_genomic_accession=20803108, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105376159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376159, CurrentID=0, Status='live', Symbol='LINC00092', Aliases='', description='long intergenic non-protein coding RNA 92', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96024347, end_position_on_the_genomic_accession=96027963, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376160, CurrentID=0, Status='live', Symbol='LOC105376160', Aliases='', description='uncharacterized LOC105376160', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95852072, end_position_on_the_genomic_accession=95869580, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376161, CurrentID=0, Status='live', Symbol='LOC105376161', Aliases='', description='uncharacterized LOC105376161', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=95764922, end_position_on_the_genomic_accession=95776282, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376162, CurrentID=0, Status='live', Symbol='LOC105376162', Aliases='', description='uncharacterized LOC105376162', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=96245435, end_position_on_the_genomic_accession=96263454, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  387491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387491, CurrentID=0, Status='live', Symbol='GAPDHP16', Aliases='GAPDP16', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 16', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14774303, end_position_on_the_genomic_accession=14775480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133284, CurrentID=0, Status='live', Symbol='LOC100133284', Aliases='', description='striatin, calmodulin binding protein pseudogene', other_designations='', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=25017531, end_position_on_the_genomic_accession=25017774, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133285, CurrentID=0, Status='live', Symbol='PCDH8P1', Aliases='', description='protocadherin 8 pseudogene 1', other_designations='', map_location='13q14.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=53199981, end_position_on_the_genomic_accession=53202412, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387494, CurrentID=0, Status='live', Symbol='RPL18AP2', Aliases='', description='ribosomal protein L18a pseudogene 2', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=46420497, end_position_on_the_genomic_accession=46421043, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387495, CurrentID=0, Status='live', Symbol='RPS26P5', Aliases='RPS26_31_1718', description='ribosomal protein S26 pseudogene 5', other_designations='', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16121243, end_position_on_the_genomic_accession=16121665, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387496, CurrentID=0, Status='live', Symbol='RASL11A', Aliases='', description='RAS-like family 11 member A', other_designations='RAS-like, family 11, member A', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=27270327, end_position_on_the_genomic_accession=27273767, orientation='plus', exon_count=6, OMIM=612403, no_hdr0=''),
  105376169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376169, CurrentID=0, Status='live', Symbol='LOC105376169', Aliases='', description='uncharacterized LOC105376169', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97786666, end_position_on_the_genomic_accession=97793879, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376170, CurrentID=0, Status='live', Symbol='LOC105376170', Aliases='', description='uncharacterized LOC105376170', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97682437, end_position_on_the_genomic_accession=97683355, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376171, CurrentID=0, Status='live', Symbol='LOC105376171', Aliases='', description='uncharacterized LOC105376171', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97633922, end_position_on_the_genomic_accession=97636453, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376172, CurrentID=0, Status='live', Symbol='LOC105376172', Aliases='', description='uncharacterized LOC105376172', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97985127, end_position_on_the_genomic_accession=97993713, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376173, CurrentID=0, Status='live', Symbol='LOC105376173', Aliases='', description='uncharacterized LOC105376173', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=98697534, end_position_on_the_genomic_accession=98701700, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133294, CurrentID=0, Status='live', Symbol='RPL34P16', Aliases='RPL34_4_651', description='ribosomal protein L34 pseudogene 16', other_designations='', map_location='6p25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=5788310, end_position_on_the_genomic_accession=5788732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387503, CurrentID=0, Status='live', Symbol='HCG4P1', Aliases='HCGIV-01', description='HLA complex group 4 pseudogene 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387504, CurrentID=0, Status='live', Symbol='HCG4P2', Aliases='HCGIV-02', description='HLA complex group 4 pseudogene 2', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387505, CurrentID=0, Status='live', Symbol='HCG9P1', Aliases='HCGIX-1', description='HLA complex group 9 pseudogene 1', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100874220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100874220, CurrentID=0, Status='live', Symbol='PCNPP4', Aliases='', description='PEST containing nuclear protein pseudogene 4', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75537548, end_position_on_the_genomic_accession=75537803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256435, CurrentID=0, Status='live', Symbol='ST6GALNAC3', Aliases='PRO7177, SIAT7C, ST6GALNACIII, STY', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 3', other_designations='GalNAc alpha-2,6-sialyltransferase III|SIAT7-C|ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltran|ST6GALNAC III|alpha-N-acetylgalactosaminide alpha-2,6-sialyltransferase III|sialyltransferase 7 ((alpha-N-acetylneuraminyl-2,3-beta-galactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) C|sialyltransferase 7C', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76074704, end_position_on_the_genomic_accession=76634601, orientation='plus', exon_count=7, OMIM=610133, no_hdr0=''),
  105376180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376180, CurrentID=0, Status='live', Symbol='LOC105376180', Aliases='', description='uncharacterized LOC105376180', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100891625, end_position_on_the_genomic_accession=100898673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133301, CurrentID=0, Status='live', Symbol='FAM231B', Aliases='', description='family with sequence similarity 231 member B', other_designations='UPF0627 protein ENSP00000364708|family with sequence similarity 231, member B', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16538724, end_position_on_the_genomic_accession=16540035, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376182, CurrentID=0, Status='live', Symbol='LOC105376182', Aliases='', description='uncharacterized LOC105376182', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101226765, end_position_on_the_genomic_accession=101238760, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376183, CurrentID=0, Status='live', Symbol='LOC105376183', Aliases='', description='uncharacterized LOC105376183', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101213670, end_position_on_the_genomic_accession=101222768, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376184, CurrentID=0, Status='live', Symbol='LOC105376184', Aliases='', description='uncharacterized LOC105376184', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101443042, end_position_on_the_genomic_accession=101459046, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376185, CurrentID=0, Status='live', Symbol='LOC105376185', Aliases='', description='uncharacterized LOC105376185', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101527774, end_position_on_the_genomic_accession=101533727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376186, CurrentID=0, Status='live', Symbol='LOC105376186', Aliases='', description='uncharacterized LOC105376186', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=101595106, end_position_on_the_genomic_accession=101597925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  125371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125371, CurrentID=0, Status='live', Symbol='RPL22P1', Aliases='RPL22_5_435', description='ribosomal protein L22 pseudogene 1', other_designations='EBER-associated protein pseudogene 1|Epstein-Barr virus small RNA-associated protein pseudogene 1|Epstein-Barr-encoded RNA-associated protein pseudogene 1', map_location='3q26.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=169483618, end_position_on_the_genomic_accession=169484020, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133308, CurrentID=0, Status='live', Symbol='RSU1P2', Aliases='', description='Ras suppressor protein 1 pseudogene 2', other_designations='', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45099476, end_position_on_the_genomic_accession=45154596, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100133309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133309, CurrentID=0, Status='live', Symbol='EIF4A2P3', Aliases='', description='eukaryotic translation initiation factor 4A2 pseudogene 3', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58242043, end_position_on_the_genomic_accession=58242956, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133310, CurrentID=0, Status='live', Symbol='HMGB1P28', Aliases='', description='high mobility group box 1 pseudogene 28', other_designations='high-mobility group box 1 pseudogene 28', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=41841615, end_position_on_the_genomic_accession=41844052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133311, CurrentID=0, Status='live', Symbol='HOXA-AS3', Aliases='HOXA6as', description='HOXA cluster antisense RNA 3', other_designations='HOXA cluster antisense RNA 3 (non-protein coding)', map_location='7p15.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=27140364, end_position_on_the_genomic_accession=27155928, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376192, CurrentID=0, Status='live', Symbol='LOC105376192', Aliases='', description='uncharacterized LOC105376192', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=103916596, end_position_on_the_genomic_accession=103926196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387521, CurrentID=0, Status='live', Symbol='TMEM189', Aliases='KUA', description='transmembrane protein 189', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50123737, end_position_on_the_genomic_accession=50153798, orientation='minus', exon_count=7, OMIM=610994, no_hdr0=''),
  387522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387522, CurrentID=0, Status='live', Symbol='TMEM189-UBE2V1', Aliases='CROC-1B, CROC1B, KUA-UEV', description='TMEM189-UBE2V1 readthrough', other_designations='TMEM189-UBE2V1 readthrough transcript|transmembrane protein 189-ubiquitin-conjugating enzyme E2 variant 1 read-through', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=50081124, end_position_on_the_genomic_accession=50153798, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100133315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133315, CurrentID=0, Status='live', Symbol='LOC100133315', Aliases='', description='transient receptor potential cation channel, subfamily C, member 2-like', other_designations='', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71865509, end_position_on_the_genomic_accession=71928447, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376196, CurrentID=0, Status='live', Symbol='LOC105376196', Aliases='', description='uncharacterized LOC105376196', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104927177, end_position_on_the_genomic_accession=104939050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376197, CurrentID=0, Status='live', Symbol='LOC105376197', Aliases='', description='uncharacterized LOC105376197', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105091515, end_position_on_the_genomic_accession=105154960, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376198, CurrentID=0, Status='live', Symbol='LOC105376198', Aliases='', description='uncharacterized LOC105376198', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105181441, end_position_on_the_genomic_accession=105194244, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133319, CurrentID=0, Status='live', Symbol='PRO1804', Aliases='', description='uncharacterized LOC100133319', other_designations='', map_location='17q24.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=69232975, end_position_on_the_genomic_accession=69235012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376200, CurrentID=0, Status='live', Symbol='LOC105376200', Aliases='', description='uncharacterized LOC105376200', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=105645434, end_position_on_the_genomic_accession=105656218, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100133321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133321, CurrentID=0, Status='live', Symbol='KLF17P1', Aliases='', description='Kruppel-like factor 17 pseudogene 1', other_designations='', map_location='12q14.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=62234390, end_position_on_the_genomic_accession=62235515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376202, CurrentID=0, Status='live', Symbol='LOC105376202', Aliases='', description='uncharacterized LOC105376202', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106310441, end_position_on_the_genomic_accession=106323864, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376203, CurrentID=0, Status='live', Symbol='LOC105376203', Aliases='', description='uncharacterized LOC105376203', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106615080, end_position_on_the_genomic_accession=106615965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376204, CurrentID=0, Status='live', Symbol='LOC105376204', Aliases='', description='uncharacterized LOC105376204', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=106857663, end_position_on_the_genomic_accession=106861009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376205, CurrentID=0, Status='live', Symbol='LOC105376205', Aliases='', description='uncharacterized LOC105376205', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107566823, end_position_on_the_genomic_accession=107665327, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105376206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376206, CurrentID=0, Status='live', Symbol='LOC105376206', Aliases='', description='uncharacterized LOC105376206', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107505084, end_position_on_the_genomic_accession=107515215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387535, CurrentID=0, Status='live', Symbol='HCRP1', Aliases='', description='hepatocellular carcinoma-related HCRP1', other_designations='', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=42326762, end_position_on_the_genomic_accession=42326920, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376208, CurrentID=0, Status='live', Symbol='LOC105376208', Aliases='', description='uncharacterized LOC105376208', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107798224, end_position_on_the_genomic_accession=107812147, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376209, CurrentID=0, Status='live', Symbol='LOC105376209', Aliases='', description='uncharacterized LOC105376209', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107758534, end_position_on_the_genomic_accession=107799871, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100133330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133330, CurrentID=0, Status='live', Symbol='RBM17P1', Aliases='', description='RNA binding motif protein 17 pseudogene 1', other_designations='', map_location='9p13.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39154022, end_position_on_the_genomic_accession=39155708, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133331, CurrentID=0, Status='live', Symbol='LOC100133331', Aliases='', description='uncharacterized LOC100133331', other_designations='', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=725759, end_position_on_the_genomic_accession=730351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376212, CurrentID=0, Status='live', Symbol='LOC105376212', Aliases='', description='uncharacterized LOC105376212', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108214659, end_position_on_the_genomic_accession=108273699, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376213, CurrentID=0, Status='live', Symbol='LOC105376213', Aliases='', description='uncharacterized LOC105376213', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108078530, end_position_on_the_genomic_accession=108083760, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376214, CurrentID=0, Status='live', Symbol='LOC105376214', Aliases='', description='uncharacterized LOC105376214', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108048185, end_position_on_the_genomic_accession=108416996, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  256471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256471, CurrentID=0, Status='live', Symbol='MFSD8', Aliases='CCMD, CLN7', description='major facilitator superfamily domain containing 8', other_designations='ceroid-lipofuscinosis, neuronal 7, late infantile', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127917805, end_position_on_the_genomic_accession=127965984, orientation='minus', exon_count=13, OMIM=611124, no_hdr0=''),
  256472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256472, CurrentID=0, Status='live', Symbol='TMEM151A', Aliases='TMEM151', description='transmembrane protein 151A', other_designations='transmembrane protein 151', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66291902, end_position_on_the_genomic_accession=66296664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376217, CurrentID=0, Status='live', Symbol='LOC105376217', Aliases='', description='uncharacterized LOC105376217', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110174702, end_position_on_the_genomic_accession=110179240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376218, CurrentID=0, Status='live', Symbol='LOC105376218', Aliases='', description='uncharacterized LOC105376218', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=110617151, end_position_on_the_genomic_accession=110620609, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376219, CurrentID=0, Status='live', Symbol='LOC105376219', Aliases='', description='uncharacterized LOC105376219', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111121880, end_position_on_the_genomic_accession=111285148, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376220, CurrentID=0, Status='live', Symbol='LOC105376220', Aliases='', description='uncharacterized LOC105376220', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=112028538, end_position_on_the_genomic_accession=112031297, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376221, CurrentID=0, Status='live', Symbol='LOC105376221', Aliases='', description='uncharacterized LOC105376221', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113588969, end_position_on_the_genomic_accession=113590340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376222, CurrentID=0, Status='live', Symbol='LOC105376222', Aliases='', description='uncharacterized LOC105376222', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113614957, end_position_on_the_genomic_accession=113628746, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376223, CurrentID=0, Status='live', Symbol='LOC105376223', Aliases='', description='uncharacterized LOC105376223', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113647895, end_position_on_the_genomic_accession=113682855, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376224, CurrentID=0, Status='live', Symbol='LOC105376224', Aliases='', description='uncharacterized LOC105376224', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114304680, end_position_on_the_genomic_accession=114318337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387553, CurrentID=0, Status='live', Symbol='DNM1P18', Aliases='DNM1DN18', description='dynamin 1 pseudogene 18', other_designations='DNM1 duplicon 18|DNM1 pseudogene 18', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42787090, end_position_on_the_genomic_accession=42788829, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387554, CurrentID=0, Status='live', Symbol='DNM1P19', Aliases='DNM1DN19', description='dynamin 1 pseudogene 19', other_designations='DNM1 duplicon 19|DNM1 pseudogene 19', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=100153596, end_position_on_the_genomic_accession=100156359, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376227, CurrentID=0, Status='live', Symbol='LOC105376227', Aliases='', description='uncharacterized LOC105376227', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114388106, end_position_on_the_genomic_accession=114391520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376228, CurrentID=0, Status='live', Symbol='LOC105376228', Aliases='', description='uncharacterized LOC105376228', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114493399, end_position_on_the_genomic_accession=114497437, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376229, CurrentID=0, Status='live', Symbol='LOC105376229', Aliases='', description='uncharacterized LOC105376229', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114558686, end_position_on_the_genomic_accession=114566443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376230, CurrentID=0, Status='live', Symbol='LOC105376230', Aliases='', description='uncharacterized LOC105376230', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=114656304, end_position_on_the_genomic_accession=114658336, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387559, CurrentID=0, Status='live', Symbol='FGF7P1', Aliases='HsT283', description='fibroblast growth factor 7 pseudogene 1', other_designations='fibroblast growth factor 7 (keratinocyte growth factor) pseudogene', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14885896, end_position_on_the_genomic_accession=14890091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  780776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780776, CurrentID=0, Status='live', Symbol='TVP23A', Aliases='FAM18A, YDR084C', description='trans-golgi network vesicle protein 23 homolog A (S. cerevisiae)', other_designations='family with sequence similarity 18, member A|protein FAM18A', map_location='16p13.13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=10761364, end_position_on_the_genomic_accession=10818859, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105376233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376233, CurrentID=0, Status='live', Symbol='LOC105376233', Aliases='', description='uncharacterized LOC105376233', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115253243, end_position_on_the_genomic_accession=115262527, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  780778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780778, CurrentID=0, Status='live', Symbol='USP10P2', Aliases='', description='ubiquitin specific peptidase10 pseudogene 2', other_designations='ubiquitin specific peptidase 10 pseudogene', map_location='14q11.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=18634112, end_position_on_the_genomic_accession=18637315, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422055, CurrentID=0, Status='live', Symbol='VPS35P1', Aliases='', description='VPS35 retromer complex component pseudogene 1', other_designations='vacuolar protein sorting 35 homolog (S. cerevisiae) pseudogene 1', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35090502, end_position_on_the_genomic_accession=35104837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376236, CurrentID=0, Status='live', Symbol='LOC105376236', Aliases='', description='uncharacterized LOC105376236', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=115989961, end_position_on_the_genomic_accession=116013910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376237, CurrentID=0, Status='live', Symbol='LOC105376237', Aliases='', description='uncharacterized LOC105376237', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117196310, end_position_on_the_genomic_accession=117259296, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376238, CurrentID=0, Status='live', Symbol='LOC105376238', Aliases='', description='uncharacterized LOC105376238', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116845339, end_position_on_the_genomic_accession=116848386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376239, CurrentID=0, Status='live', Symbol='LOC105376239', Aliases='', description='uncharacterized LOC105376239', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116838470, end_position_on_the_genomic_accession=116843126, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376240, CurrentID=0, Status='live', Symbol='LOC105376240', Aliases='', description='uncharacterized LOC105376240', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=116597543, end_position_on_the_genomic_accession=116644642, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387569, CurrentID=0, Status='live', Symbol='ACF', Aliases='', description='Asymmetric crying facies (Cayler cardiofacial syndrome)', other_designations='', map_location='22q11', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=125520, no_hdr0=''),
  387570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387570, CurrentID=0, Status='live', Symbol='DUH1', Aliases='DUH', description='Dyschromatosis universalis hereditaria', other_designations='', map_location='6q24.2-q25.2', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=127500, no_hdr0=''),
  105376243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376243, CurrentID=0, Status='live', Symbol='LOC105376243', Aliases='', description='uncharacterized LOC105376243', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=117840282, end_position_on_the_genomic_accession=117876249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387572, CurrentID=0, Status='live', Symbol='PAND1', Aliases='PAND', description='Panic disorder 1', other_designations='', map_location='13q22-q32', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=167870, no_hdr0=''),
  387573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387573, CurrentID=0, Status='live', Symbol='CHDS3', Aliases='', description='Coronary heart disease, susceptibility to, 3', other_designations='', map_location='Xq23-q26', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300464, no_hdr0=''),
  387574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387574, CurrentID=0, Status='live', Symbol='CMTDIA', Aliases='CMTDI2', description='Charcot-Marie-Tooth disease, dominant intermediate 2', other_designations='', map_location='10q24.1-q25.1', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606483, no_hdr0=''),
  387575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387575, CurrentID=0, Status='live', Symbol='HYT3', Aliases='', description='Hypertension, essential, susceptibility to, 3', other_designations='', map_location='2p25-p24', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607329, no_hdr0=''),
  387576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387576, CurrentID=0, Status='live', Symbol='MGR5', Aliases='', description='Migraine with or without aura, susceptibility to, 5', other_designations='', map_location='19p13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607508, no_hdr0=''),
  387577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387577, CurrentID=0, Status='live', Symbol='AUTS3', Aliases='', description='Autism, susceptibility to, 3', other_designations='', map_location='13q14.2-q14.1', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608049, no_hdr0=''),
  387578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387578, CurrentID=0, Status='live', Symbol='AITD1', Aliases='', description='Autoimmune thyroid disease, susceptibility to, 1', other_designations='', map_location='6p11', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608173, no_hdr0=''),
  387579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387579, CurrentID=0, Status='live', Symbol='AITD2', Aliases='', description='Autoimmune thyroid disease, susceptibility to, 2', other_designations='', map_location='5q31-q33', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608174, no_hdr0=''),
  105376252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376252, CurrentID=0, Status='live', Symbol='LOC105376252', Aliases='', description='uncharacterized LOC105376252', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120380873, end_position_on_the_genomic_accession=120383349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387581, CurrentID=0, Status='live', Symbol='AITD4', Aliases='', description='Autoimmune thyroid disease, susceptibility to, 4', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608176, no_hdr0=''),
  387582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387582, CurrentID=0, Status='live', Symbol='KAZA1', Aliases='KAZA', description='Kala-azar (visceral leishmaniasis), susceptibility to', other_designations='', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608207, no_hdr0=''),
  387583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387583, CurrentID=0, Status='live', Symbol='SPG25', Aliases='', description='spastic paraplegia 25 (autosomal recessive, with disc herniation)', other_designations='', map_location='6q23-q24.1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608220, no_hdr0=''),
  387584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387584, CurrentID=0, Status='live', Symbol='CHDS2', Aliases='', description='Coronary heart disease, susceptibility to, 2', other_designations='', map_location='2q21.1-q22', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608316, no_hdr0=''),
  387585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387585, CurrentID=0, Status='live', Symbol='CHDS4', Aliases='', description='Coronary heart disease, susceptibility to, 4', other_designations='', map_location='14q32', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=608318, no_hdr0=''),
  105376258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376258, CurrentID=0, Status='live', Symbol='LOC105376258', Aliases='', description='uncharacterized LOC105376258', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121695351, end_position_on_the_genomic_accession=121698908, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376259, CurrentID=0, Status='live', Symbol='LOC105376259', Aliases='', description='uncharacterized LOC105376259', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121679198, end_position_on_the_genomic_accession=121685218, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376260, CurrentID=0, Status='live', Symbol='LOC105376260', Aliases='', description='uncharacterized LOC105376260', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122083077, end_position_on_the_genomic_accession=122092679, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376261, CurrentID=0, Status='live', Symbol='TTLL11-IT1', Aliases='', description='TTLL11 intronic transcript 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121884611, end_position_on_the_genomic_accession=121896783, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387590, CurrentID=0, Status='live', Symbol='TPTEP1', Aliases='psiTPTE22', description='transmembrane phosphatase with tensin homology pseudogene 1', other_designations='TPTE pseudogene|endogenous retroviral protease pseudogene', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16601911, end_position_on_the_genomic_accession=16648830, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376263, CurrentID=0, Status='live', Symbol='LOC105376263', Aliases='', description='uncharacterized LOC105376263', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122755756, end_position_on_the_genomic_accession=122785108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376264, CurrentID=0, Status='live', Symbol='LOC105376264', Aliases='', description='uncharacterized LOC105376264', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123079494, end_position_on_the_genomic_accession=123087380, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376265, CurrentID=0, Status='live', Symbol='LOC105376265', Aliases='', description='uncharacterized LOC105376265', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=123304825, end_position_on_the_genomic_accession=123322610, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  780810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780810, CurrentID=0, Status='live', Symbol='PAICSP3', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 3', other_designations='', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73326955, end_position_on_the_genomic_accession=73331060, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780811, CurrentID=0, Status='live', Symbol='PAICSP5', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 5', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143545041, end_position_on_the_genomic_accession=143546311, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  780812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780812, CurrentID=0, Status='live', Symbol='PAICSP6', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 6', other_designations='', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143783127, end_position_on_the_genomic_accession=143784279, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387597, CurrentID=0, Status='live', Symbol='ILDR2', Aliases='C1orf32, dJ782G3.1', description='immunoglobulin-like domain containing receptor 2', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=166913204, end_position_on_the_genomic_accession=166989093, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  780814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780814, CurrentID=0, Status='live', Symbol='PAICSP1', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 1', other_designations='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase carboxylase pseudogene 1', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=37878061, end_position_on_the_genomic_accession=37912435, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780815, CurrentID=0, Status='live', Symbol='PAICSP2', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 2', other_designations='', map_location='9q22', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=91611266, end_position_on_the_genomic_accession=91612531, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  780816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780816, CurrentID=0, Status='live', Symbol='PAICSP7', Aliases='', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoimidazole succinocarboxamide synthetase pseudogene 7', other_designations='', map_location='Xq21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=94636503, end_position_on_the_genomic_accession=94639599, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387601, CurrentID=0, Status='live', Symbol='SLC22A25', Aliases='HIMTP, UST6', description='solute carrier family 22 member 25', other_designations='MGI:2442751, MGI:2385316, MGI:3042283, MGI:3645714, MGI:3605624, MGI:2442750|organic anion transporter UST6|putative UST1-like organic anion transporter|solute carrier family 22, member 25', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63162266, end_position_on_the_genomic_accession=63243516, orientation='minus', exon_count=13, OMIM=610792, no_hdr0=''),
  105376274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376274, CurrentID=0, Status='live', Symbol='LOC105376274', Aliases='', description='uncharacterized LOC105376274', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126124919, end_position_on_the_genomic_accession=126149127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376275, CurrentID=0, Status='live', Symbol='LOC105376275', Aliases='', description='uncharacterized LOC105376275', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126226089, end_position_on_the_genomic_accession=126235851, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376276, CurrentID=0, Status='live', Symbol='LOC105376276', Aliases='', description='uncharacterized LOC105376276', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126516491, end_position_on_the_genomic_accession=126518861, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376277, CurrentID=0, Status='live', Symbol='LOC105376277', Aliases='', description='uncharacterized LOC105376277', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=126608094, end_position_on_the_genomic_accession=126613724, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376278, CurrentID=0, Status='live', Symbol='LOC105376278', Aliases='', description='uncharacterized LOC105376278', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127174789, end_position_on_the_genomic_accession=127188018, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376279, CurrentID=0, Status='live', Symbol='LOC105376279', Aliases='', description='uncharacterized LOC105376279', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127022556, end_position_on_the_genomic_accession=127023250, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  256536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256536, CurrentID=0, Status='live', Symbol='TCERG1L', Aliases='', description='transcription elongation regulator 1 like', other_designations='5730476P14Rik', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131092391, end_position_on_the_genomic_accession=131311721, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105376281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376281, CurrentID=0, Status='live', Symbol='LOC105376281', Aliases='', description='uncharacterized LOC105376281', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127536585, end_position_on_the_genomic_accession=127537336, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376282, CurrentID=0, Status='live', Symbol='LOC105376282', Aliases='', description='uncharacterized LOC105376282', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127579128, end_position_on_the_genomic_accession=127592233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376283, CurrentID=0, Status='live', Symbol='LOC105376283', Aliases='', description='uncharacterized LOC105376283', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127693176, end_position_on_the_genomic_accession=127695181, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387612, CurrentID=0, Status='live', Symbol='RAC1P1', Aliases='HsT2789', description='ras-related C3 botulinum toxin substrate 1 pseudogene 1', other_designations='ras-related C3 botulinum toxin substrate 1 (rho family, small GTP binding protein Rac1) pseudogene', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24685478, end_position_on_the_genomic_accession=24686365, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376285, CurrentID=0, Status='live', Symbol='LOC105376285', Aliases='', description='uncharacterized LOC105376285', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128196832, end_position_on_the_genomic_accession=128199962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376286, CurrentID=0, Status='live', Symbol='LOC105376286', Aliases='', description='uncharacterized LOC105376286', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128313825, end_position_on_the_genomic_accession=128315559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387615, CurrentID=0, Status='live', Symbol='MAP1LC3P', Aliases='HsT312', description='microtubule associated protein 1 light chain 3 pseudogene', other_designations='microtubule-associated protein 1 light chain 3 pseudogene', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54954413, end_position_on_the_genomic_accession=54955037, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376288, CurrentID=0, Status='live', Symbol='LOC105376288', Aliases='', description='uncharacterized LOC105376288', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128771158, end_position_on_the_genomic_accession=128781872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376289, CurrentID=0, Status='live', Symbol='LOC105376289', Aliases='', description='uncharacterized LOC105376289', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129203649, end_position_on_the_genomic_accession=129209885, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376290, CurrentID=0, Status='live', Symbol='LOC105376290', Aliases='', description='uncharacterized LOC105376290', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129257778, end_position_on_the_genomic_accession=129260922, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376291, CurrentID=0, Status='live', Symbol='LOC105376291', Aliases='', description='uncharacterized LOC105376291', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129438621, end_position_on_the_genomic_accession=129447098, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  125476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125476, CurrentID=0, Status='live', Symbol='INO80C', Aliases='C18orf37, IES6, hIes6', description='INO80 complex subunit C', other_designations='IES6 homolog', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=35468327, end_position_on_the_genomic_accession=35497991, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376293, CurrentID=0, Status='live', Symbol='LOC105376293', Aliases='', description='uncharacterized LOC105376293', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130260278, end_position_on_the_genomic_accession=130266936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376294, CurrentID=0, Status='live', Symbol='LOC105376294', Aliases='', description='uncharacterized LOC105376294', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130490658, end_position_on_the_genomic_accession=130502550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376295, CurrentID=0, Status='live', Symbol='LOC105376295', Aliases='', description='uncharacterized LOC105376295', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130532452, end_position_on_the_genomic_accession=130536194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376296, CurrentID=0, Status='live', Symbol='LOC105376296', Aliases='', description='uncharacterized LOC105376296', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130782359, end_position_on_the_genomic_accession=130817460, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376297, CurrentID=0, Status='live', Symbol='LOC105376297', Aliases='', description='uncharacterized LOC105376297', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=130940346, end_position_on_the_genomic_accession=130945524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376298, CurrentID=0, Status='live', Symbol='LOC105376298', Aliases='', description='uncharacterized LOC105376298', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131088125, end_position_on_the_genomic_accession=131093868, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376299, CurrentID=0, Status='live', Symbol='LOC105376299', Aliases='', description='uncharacterized LOC105376299', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131347306, end_position_on_the_genomic_accession=131358960, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387628, CurrentID=0, Status='live', Symbol='KGFLP1', Aliases='', description='fibroblast growth factor 7 pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62376256, end_position_on_the_genomic_accession=62435519, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376301, CurrentID=0, Status='live', Symbol='LOC105376301', Aliases='', description='uncharacterized LOC105376301', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=131516558, end_position_on_the_genomic_accession=131517803, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376302, CurrentID=0, Status='live', Symbol='LOC105376302', Aliases='', description='uncharacterized LOC105376302', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132103220, end_position_on_the_genomic_accession=132104582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392285, CurrentID=0, Status='live', Symbol='PPIAP33', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 33', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=7499155, end_position_on_the_genomic_accession=7598374, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  125488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125488, CurrentID=0, Status='live', Symbol='TTC39C', Aliases='C18orf17, HsT2697', description='tetratricopeptide repeat domain 39C', other_designations='TPR repeat protein 39C', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=23992773, end_position_on_the_genomic_accession=24135610, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105376305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376305, CurrentID=0, Status='live', Symbol='LOC105376305', Aliases='', description='uncharacterized LOC105376305', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133010567, end_position_on_the_genomic_accession=133023931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376306, CurrentID=0, Status='live', Symbol='LOC105376306', Aliases='', description='uncharacterized LOC105376306', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133028147, end_position_on_the_genomic_accession=133030492, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  151987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=151987, CurrentID=0, Status='live', Symbol='PPP4R2', Aliases='PP4R2', description='protein phosphatase 4 regulatory subunit 2', other_designations='protein phosphatase 4, regulatory subunit 2', map_location='3p13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72996968, end_position_on_the_genomic_accession=73065860, orientation='plus', exon_count=9, OMIM=613822, no_hdr0=''),
  780852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780852, CurrentID=0, Status='live', Symbol='SNORD3B-2', Aliases='U3-2B, U3b2', description='small nucleolar RNA, C/D box 3B-2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19063920, end_position_on_the_genomic_accession=19064136, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  780853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780853, CurrentID=0, Status='live', Symbol='SNORD3C', Aliases='U3-3', description='small nucleolar RNA, C/D box 3C', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19190029, end_position_on_the_genomic_accession=19190245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387638, CurrentID=0, Status='live', Symbol='C10orf113', Aliases='bA165O3.1', description='chromosome 10 open reading frame 113', other_designations='', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21125763, end_position_on_the_genomic_accession=21146559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376311, CurrentID=0, Status='live', Symbol='LOC105376311', Aliases='', description='uncharacterized LOC105376311', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134472769, end_position_on_the_genomic_accession=134483232, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  387640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387640, CurrentID=0, Status='live', Symbol='SKIDA1', Aliases='C10orf140, DLN-1', description='SKI/DACH domain containing 1', other_designations='protein DLN-1', map_location='10p12.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21510754, end_position_on_the_genomic_accession=21527188, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105376313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376313, CurrentID=0, Status='live', Symbol='LOC105376313', Aliases='', description='uncharacterized LOC105376313', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134944536, end_position_on_the_genomic_accession=134949027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376314, CurrentID=0, Status='live', Symbol='LOC105376314', Aliases='', description='uncharacterized LOC105376314', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134934956, end_position_on_the_genomic_accession=134943558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376315, CurrentID=0, Status='live', Symbol='LOC105376315', Aliases='', description='uncharacterized LOC105376315', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135268142, end_position_on_the_genomic_accession=135277644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387644, CurrentID=0, Status='live', Symbol='LINC00202-1', Aliases='C10orf51, LINC00202, NCRNA00202, bB27G4.1', description='long intergenic non-protein coding RNA 202-1', other_designations='long intergenic non-protein coding RNA 202', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26931206, end_position_on_the_genomic_accession=26942001, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376317, CurrentID=0, Status='live', Symbol='LOC105376317', Aliases='', description='uncharacterized LOC105376317', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135675200, end_position_on_the_genomic_accession=135685111, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  387646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387646, CurrentID=0, Status='live', Symbol='LRRC37A6P', Aliases='', description='leucine-rich repeat containing 37 member A6, pseudogene', other_designations='leucine rich repeat containing 37, member A family pseudogene|leucine rich repeat containing 37, member A6, pseudogene', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27245853, end_position_on_the_genomic_accession=27252306, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387647, CurrentID=0, Status='live', Symbol='PTCHD3P1', Aliases='', description='patched domain containing 3 pseudogene 1', other_designations='', map_location='10p11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29421476, end_position_on_the_genomic_accession=29422238, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100526656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526656, CurrentID=0, Status='live', Symbol='GNG10P1', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 10 pseudogene 1', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125958874, end_position_on_the_genomic_accession=125960469, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376321, CurrentID=0, Status='live', Symbol='LOC105376321', Aliases='', description='uncharacterized LOC105376321', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135702112, end_position_on_the_genomic_accession=135704361, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376322, CurrentID=0, Status='live', Symbol='LOC105376322', Aliases='', description='uncharacterized LOC105376322', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136013956, end_position_on_the_genomic_accession=136016099, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100312843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100312843, CurrentID=0, Status='live', Symbol='VN1R19P', Aliases='', description='vomeronasal 1 receptor 19 pseudogene', other_designations='', map_location='2q21.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=132108590, end_position_on_the_genomic_accession=132109005, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376324, CurrentID=0, Status='live', Symbol='LOC105376324', Aliases='', description='uncharacterized LOC105376324', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136523199, end_position_on_the_genomic_accession=136533069, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376325, CurrentID=0, Status='live', Symbol='LOC105376325', Aliases='', description='uncharacterized LOC105376325', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136640166, end_position_on_the_genomic_accession=136644479, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376326, CurrentID=0, Status='live', Symbol='LOC105376326', Aliases='', description='uncharacterized LOC105376326', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136881926, end_position_on_the_genomic_accession=136886256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376327, CurrentID=0, Status='live', Symbol='LOC105376327', Aliases='', description='uncharacterized LOC105376327', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137052358, end_position_on_the_genomic_accession=137053375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100526664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526664, CurrentID=0, Status='live', Symbol='LY75-CD302', Aliases='', description='LY75-CD302 readthrough', other_designations='DEC-205/DCL-1 fusion', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159768628, end_position_on_the_genomic_accession=159904756, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  105376329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376329, CurrentID=0, Status='live', Symbol='LOC105376329', Aliases='', description='uncharacterized LOC105376329', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137224109, end_position_on_the_genomic_accession=137225233, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  256586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256586, CurrentID=0, Status='live', Symbol='LYSMD2', Aliases='', description='LysM, putative peptidoglycan-binding, domain containing 2', other_designations='', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51723064, end_position_on_the_genomic_accession=51751453, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376331, CurrentID=0, Status='live', Symbol='LOC105376331', Aliases='', description='uncharacterized LOC105376331', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137943654, end_position_on_the_genomic_accession=137947022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376332, CurrentID=0, Status='live', Symbol='LOC105376332', Aliases='', description='uncharacterized LOC105376332', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138286330, end_position_on_the_genomic_accession=138317732, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376333, CurrentID=0, Status='live', Symbol='LOC105376333', Aliases='', description='formin-like protein 6', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=138276875, end_position_on_the_genomic_accession=138279535, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376334, CurrentID=0, Status='live', Symbol='LOC105376334', Aliases='', description='ankyrin repeat domain-containing protein 36B pseudogene', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=79822358, end_position_on_the_genomic_accession=79851565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376335, CurrentID=0, Status='live', Symbol='LOC105376335', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83920901, end_position_on_the_genomic_accession=83924902, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376336, CurrentID=0, Status='live', Symbol='LOC105376336', Aliases='', description='glutamate receptor ionotropic, NMDA 2D-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=107267050, end_position_on_the_genomic_accession=107283631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376337, CurrentID=0, Status='live', Symbol='LOC105376337', Aliases='', description='uncharacterized LOC105376337', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121194174, end_position_on_the_genomic_accession=121212539, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376338, CurrentID=0, Status='live', Symbol='LOC105376338', Aliases='', description='uncharacterized LOC105376338', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44696, end_position_on_the_genomic_accession=46683, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376339, CurrentID=0, Status='live', Symbol='LOC105376339', Aliases='', description='uncharacterized LOC105376339', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=665959, end_position_on_the_genomic_accession=666681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376340, CurrentID=0, Status='live', Symbol='LOC105376340', Aliases='', description='uncharacterized LOC105376340', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=353610, end_position_on_the_genomic_accession=362481, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133461, CurrentID=0, Status='live', Symbol='LOC100133461', Aliases='', description='uncharacterized LOC100133461', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=3673593, end_position_on_the_genomic_accession=3677855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376342, CurrentID=0, Status='live', Symbol='LOC105376342', Aliases='', description='uncharacterized LOC105376342', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1359590, end_position_on_the_genomic_accession=1361951, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376343, CurrentID=0, Status='live', Symbol='LOC105376343', Aliases='', description='uncharacterized LOC105376343', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1319332, end_position_on_the_genomic_accession=1328781, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376344, CurrentID=0, Status='live', Symbol='LOC105376344', Aliases='', description='uncharacterized LOC105376344', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1152310, end_position_on_the_genomic_accession=1159347, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376345, CurrentID=0, Status='live', Symbol='LOC105376345', Aliases='', description='uncharacterized LOC105376345', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1742274, end_position_on_the_genomic_accession=1772289, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376346, CurrentID=0, Status='live', Symbol='LOC105376346', Aliases='', description='uncharacterized LOC105376346', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=1971813, end_position_on_the_genomic_accession=1990380, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133467, CurrentID=0, Status='live', Symbol='LOC100133467', Aliases='', description='uncharacterized LOC100133467', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376348, CurrentID=0, Status='live', Symbol='LOC105376348', Aliases='', description='uncharacterized LOC105376348', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2169164, end_position_on_the_genomic_accession=2189505, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376349, CurrentID=0, Status='live', Symbol='LOC105376349', Aliases='', description='uncharacterized LOC105376349', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2318426, end_position_on_the_genomic_accession=2325108, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376350, CurrentID=0, Status='live', Symbol='LOC105376350', Aliases='', description='uncharacterized LOC105376350', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=2501908, end_position_on_the_genomic_accession=2618724, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  780895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780895, CurrentID=0, Status='live', Symbol='DUPC1', Aliases='', description='Dupuytren contracture 1', other_designations='', map_location='16q11.1-q22', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=126900, no_hdr0=''),
  387680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387680, CurrentID=0, Status='live', Symbol='FAM21A', Aliases='FAM21B, bA56A21.1, bA98I6.1', description='family with sequence similarity 21 member A', other_designations='WASH complex subunit FAM21B|family with sequence similarity 21, member A|family with sequence similarity 21, member B', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50067888, end_position_on_the_genomic_accession=50133509, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  780897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780897, CurrentID=0, Status='live', Symbol='HPP1', Aliases='FPH1, HPP, MUH', description='hyperpigmentation, progressive, 1', other_designations='familial progressive hyperpigmentation 1', map_location='19pter-p13.1', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=145250, no_hdr0=''),
  780898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780898, CurrentID=0, Status='live', Symbol='APMR1', Aliases='', description='Alopecia-mental retardation syndrome', other_designations='', map_location='3q26.3-q27.3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=203650, no_hdr0=''),
  780899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780899, CurrentID=0, Status='live', Symbol='DIH2', Aliases='', description='Hernia, congenital diaphragmatic 2', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=222400, no_hdr0=''),
  780900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780900, CurrentID=0, Status='live', Symbol='FGS5', Aliases='', description='FG syndrome 5', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=300581, no_hdr0=''),
  100526693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526693, CurrentID=0, Status='live', Symbol='ARPC4-TTLL3', Aliases='', description='ARPC4-TTLL3 readthrough', other_designations='ARPC4-TTLL3 read-through transcript', map_location='3p', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=9792548, end_position_on_the_genomic_accession=9836356, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100526694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526694, CurrentID=0, Status='live', Symbol='MSANTD3-TMEFF1', Aliases='C9orf3-TMEFF1, C9orf30-TMEFF1', description='MSANTD3-TMEFF1 readthrough', other_designations='', map_location='9q', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100441906, end_position_on_the_genomic_accession=100577636, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  780903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780903, CurrentID=0, Status='live', Symbol='HHT3', Aliases='ORW3', description='Osler-Rendu-Weber syndrome 3', other_designations='', map_location='5q31.3-q32', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601101, no_hdr0=''),
  780904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780904, CurrentID=0, Status='live', Symbol='PSS', Aliases='', description='Potocki-Shaffer syndrome', other_designations='', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=601224, no_hdr0=''),
  780905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780905, CurrentID=0, Status='live', Symbol='EA3', Aliases='', description='Episodic ataxia, type 3', other_designations='', map_location='1q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606554, no_hdr0=''),
  780906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780906, CurrentID=0, Status='live', Symbol='SLI3', Aliases='', description='Specific language impairment QTL, 3', other_designations='', map_location='13q21', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607134, no_hdr0=''),
  105376363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376363, CurrentID=0, Status='live', Symbol='LOC105376363', Aliases='', description='uncharacterized LOC105376363', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=3759190, end_position_on_the_genomic_accession=3769321, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  780908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780908, CurrentID=0, Status='live', Symbol='PEE2', Aliases='', description='Preeclampsia/eclampsia 2', other_designations='', map_location='2p25', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609402, no_hdr0=''),
  780909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780909, CurrentID=0, Status='live', Symbol='PEE3', Aliases='', description='Preeclampsia/eclampsia 3', other_designations='', map_location='9p13', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609403, no_hdr0=''),
  387694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387694, CurrentID=0, Status='live', Symbol='SH2D4B', Aliases='', description='SH2 domain containing 4B', other_designations='novel SH2 domain-containing protein', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80537902, end_position_on_the_genomic_accession=80646560, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  780911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780911, CurrentID=0, Status='live', Symbol='DEL10Q26', Aliases='C10q26DEL, TQDS', description='Chromosome 10q deletion syndrome', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609625, no_hdr0=''),
  780912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780912, CurrentID=0, Status='live', Symbol='AD10', Aliases='', description='Alzheimer disease-10', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609636, no_hdr0=''),
  780913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780913, CurrentID=0, Status='live', Symbol='EIG4', Aliases='', description='Epilepsy, idiopathic generalized, susceptibility to 4', other_designations='', map_location='10q25-q26', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609750, no_hdr0=''),
  780914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780914, CurrentID=0, Status='live', Symbol='ZD1', Aliases='', description='Zygodactyly 1', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609815, no_hdr0=''),
  780915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780915, CurrentID=0, Status='live', Symbol='CHNG3', Aliases='RTSH', description='Hypothyroidism, congenital, nongoitrous, 3', other_designations='', map_location='15q25.3-q26.1', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609893, no_hdr0=''),
  780916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780916, CurrentID=0, Status='live', Symbol='GINGF3', Aliases='GGF3, HGF3', description='gingival fibromatosis, hereditary, 3', other_designations='', map_location='2p23.3-p22.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609955, no_hdr0=''),
  780917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780917, CurrentID=0, Status='live', Symbol='NKCD', Aliases='', description='Natural killer cell deficiency, familial isolated', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609981, no_hdr0=''),
  105376374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376374, CurrentID=0, Status='live', Symbol='LOC105376374', Aliases='', description='uncharacterized LOC105376374', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4757070, end_position_on_the_genomic_accession=4764045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  780919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780919, CurrentID=0, Status='live', Symbol='HRPT3', Aliases='', description='Hyperparathyroidism 3', other_designations='', map_location='2p14-p13.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610071, no_hdr0=''),
  780920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780920, CurrentID=0, Status='live', Symbol='MPD3', Aliases='', description='Myopathy, distal 3', other_designations='', map_location='8p22-q11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610099, no_hdr0=''),
  780921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780921, CurrentID=0, Status='live', Symbol='KYPSC1', Aliases='', description='Kyphoscoliosis 1', other_designations='', map_location='5p13', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610170, no_hdr0=''),
  780922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780922, CurrentID=0, Status='live', Symbol='SPD3', Aliases='', description='Synpolydactyly 3', other_designations='', map_location='14q11.2-q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610234, no_hdr0=''),
  387707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387707, CurrentID=0, Status='live', Symbol='CC2D2B', Aliases='C10orf130, bA248J23.4', description='coiled-coil and C2 domain containing 2B', other_designations='', map_location='10q24.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95974029, end_position_on_the_genomic_accession=96035750, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105376380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376380, CurrentID=0, Status='live', Symbol='LOC105376380', Aliases='', description='uncharacterized LOC105376380', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5549599, end_position_on_the_genomic_accession=5552069, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  780925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=780925, CurrentID=0, Status='live', Symbol='CHMRQ', Aliases='', description='Cerebellar hypoplasia, mental retardation, and quadrupedal locomotion', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=610185, no_hdr0=''),
  105376382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376382, CurrentID=0, Status='live', Symbol='LOC105376382', Aliases='', description='uncharacterized LOC105376382', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5594984, end_position_on_the_genomic_accession=5596094, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376383, CurrentID=0, Status='live', Symbol='LOC105376383', Aliases='', description='uncharacterized LOC105376383', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=5607994, end_position_on_the_genomic_accession=5610439, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387712, CurrentID=0, Status='live', Symbol='ENO4', Aliases='C10orf134', description='enolase family member 4', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116849212, end_position_on_the_genomic_accession=116943974, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105376385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376385, CurrentID=0, Status='live', Symbol='LOC105376385', Aliases='', description='uncharacterized LOC105376385', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6745784, end_position_on_the_genomic_accession=6755414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376386, CurrentID=0, Status='live', Symbol='LOC105376386', Aliases='', description='uncharacterized LOC105376386', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7096285, end_position_on_the_genomic_accession=7097675, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387715, CurrentID=0, Status='live', Symbol='ARMS2', Aliases='ARMD8', description='age-related maculopathy susceptibility 2', other_designations='', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122454663, end_position_on_the_genomic_accession=122457352, orientation='plus', exon_count=2, OMIM=611313, no_hdr0=''),
  105376388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376388, CurrentID=0, Status='live', Symbol='LOC105376388', Aliases='', description='uncharacterized LOC105376388', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7465661, end_position_on_the_genomic_accession=7472896, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376389, CurrentID=0, Status='live', Symbol='LOC105376389', Aliases='', description='uncharacterized LOC105376389', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7485462, end_position_on_the_genomic_accession=7491755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256646, CurrentID=0, Status='live', Symbol='NUTM1', Aliases='C15orf55, FAM22H, NUT', description='NUT midline carcinoma, family member 1', other_designations='nuclear protein in testis', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34343251, end_position_on_the_genomic_accession=34357737, orientation='plus', exon_count=8, OMIM=608963, no_hdr0=''),
  105376391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376391, CurrentID=0, Status='live', Symbol='LOC105376391', Aliases='', description='uncharacterized LOC105376391', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=7541399, end_position_on_the_genomic_accession=7548704, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387720, CurrentID=0, Status='live', Symbol='LOC387720', Aliases='', description='collagen alpha-1(X) chain', other_designations='', map_location='10q26.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127734610, end_position_on_the_genomic_accession=127737275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376393, CurrentID=0, Status='live', Symbol='LOC105376393', Aliases='', description='uncharacterized LOC105376393', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8015794, end_position_on_the_genomic_accession=8018607, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376394, CurrentID=0, Status='live', Symbol='LOC105376394', Aliases='', description='uncharacterized LOC105376394', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8035581, end_position_on_the_genomic_accession=8042845, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387723, CurrentID=0, Status='live', Symbol='LINC00959', Aliases='', description='long intergenic non-protein coding RNA 959', other_designations='', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130063896, end_position_on_the_genomic_accession=130110817, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376396, CurrentID=0, Status='live', Symbol='LOC105376396', Aliases='', description='uncharacterized LOC105376396', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8382525, end_position_on_the_genomic_accession=8401504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376397, CurrentID=0, Status='live', Symbol='LOC105376397', Aliases='', description='uncharacterized LOC105376397', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8435740, end_position_on_the_genomic_accession=8454127, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376398, CurrentID=0, Status='live', Symbol='LOC105376398', Aliases='', description='uncharacterized LOC105376398', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8572747, end_position_on_the_genomic_accession=8594412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100526735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526735, CurrentID=0, Status='live', Symbol='CRYZP2', Aliases='', description='crystallin zeta pseudogene 2', other_designations='crystallin, zeta (quinone reductase) pseudogene 2', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=97914418, end_position_on_the_genomic_accession=97917424, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100526736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526736, CurrentID=0, Status='live', Symbol='LOC100526736', Aliases='', description='quinone oxidoreductase-like protein 2 pseudogene', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=85246137, end_position_on_the_genomic_accession=85248851, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100526737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526737, CurrentID=0, Status='live', Symbol='RBM14-RBM4', Aliases='COAZ', description='RBM14-RBM4 readthrough', other_designations='RBM14/RBM4 fusion|transcriptional coactivator CoAZ', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66616582, end_position_on_the_genomic_accession=66646473, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100526738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526738, CurrentID=0, Status='live', Symbol='SUMO2P5', Aliases='', description='SUMO2 pseudogene 5', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 5', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75225424, end_position_on_the_genomic_accession=75225668, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100526739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526739, CurrentID=0, Status='live', Symbol='APITD1-CORT', Aliases='', description='APITD1-CORT readthrough', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10430102, end_position_on_the_genomic_accession=10452003, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100526740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526740, CurrentID=0, Status='live', Symbol='ATP5J2-PTCD1', Aliases='', description='ATP5J2-PTCD1 readthrough', other_designations='ATP5J2-PTCD1 read-through transcript', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=99416739, end_position_on_the_genomic_accession=99466201, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  387733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387733, CurrentID=0, Status='live', Symbol='IFITM5', Aliases='BRIL, DSPA1, Hrmp1, OI5, fragilis4', description='interferon induced transmembrane protein 5', other_designations='bone-restricted ifitm-like protein|bone-restricted interferon-induced transmembrane protein-like protein|dispanin subfamily A member 1', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=298200, end_position_on_the_genomic_accession=299526, orientation='minus', exon_count=2, OMIM=614757, no_hdr0=''),
  100526742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526742, CurrentID=0, Status='live', Symbol='DEL16P12.1P11.2', Aliases='C16DELp12.1p11.2', description='Chromosome 16p12.2-p11.2 deletion syndrome', other_designations='', map_location='16q12.2-p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613604, no_hdr0=''),
  100526743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526743, CurrentID=0, Status='live', Symbol='DUP17Q23.1Q23.2', Aliases='C17DUPq23.1q23.2', description='Chromosome 17q23.1-q23.2 duplication syndrome', other_designations='', map_location='17q23.1-q23.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613618, no_hdr0=''),
  105376408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376408, CurrentID=0, Status='live', Symbol='LOC105376408', Aliases='', description='uncharacterized LOC105376408', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=10775621, end_position_on_the_genomic_accession=10783760, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376409, CurrentID=0, Status='live', Symbol='LOC105376409', Aliases='', description='uncharacterized LOC105376409', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11350575, end_position_on_the_genomic_accession=11355466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376410, CurrentID=0, Status='live', Symbol='LOC105376410', Aliases='', description='uncharacterized LOC105376410', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11497422, end_position_on_the_genomic_accession=11517354, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376411, CurrentID=0, Status='live', Symbol='LOC105376411', Aliases='', description='uncharacterized LOC105376411', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11463783, end_position_on_the_genomic_accession=11467204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376412, CurrentID=0, Status='live', Symbol='LOC105376412', Aliases='', description='uncharacterized LOC105376412', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11672957, end_position_on_the_genomic_accession=11679646, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376413, CurrentID=0, Status='live', Symbol='LOC105376413', Aliases='', description='uncharacterized LOC105376413', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11636399, end_position_on_the_genomic_accession=11682029, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  387742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387742, CurrentID=0, Status='live', Symbol='FAM99A', Aliases='', description='family with sequence similarity 99 member A (non-protein coding)', other_designations='family with sequence similarity 99, member A (non-protein coding)', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1665599, end_position_on_the_genomic_accession=1667856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376415, CurrentID=0, Status='live', Symbol='LOC105376415', Aliases='', description='uncharacterized LOC105376415', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=11784286, end_position_on_the_genomic_accession=11785332, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376416, CurrentID=0, Status='live', Symbol='LOC105376416', Aliases='', description='uncharacterized LOC105376416', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12043673, end_position_on_the_genomic_accession=12046077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387745, CurrentID=0, Status='live', Symbol='FAM86GP', Aliases='', description='family with sequence similarity 86 member G, pseudogene', other_designations='family with sequence similarity 86, member G, pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3408240, end_position_on_the_genomic_accession=3422577, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376418, CurrentID=0, Status='live', Symbol='LOC105376418', Aliases='', description='uncharacterized LOC105376418', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=12907636, end_position_on_the_genomic_accession=12965475, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387747, CurrentID=0, Status='live', Symbol='OR52B3P', Aliases='PJCG7', description='olfactory receptor family 52 subfamily B member 3 pseudogene', other_designations='olfactory receptor, family 52, subfamily B, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4378282, end_position_on_the_genomic_accession=4379224, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  387748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387748, CurrentID=0, Status='live', Symbol='OR56B1', Aliases='OR11-65P, OR56B1', description='olfactory receptor family 56 subfamily B member 1', other_designations='olfactory receptor OR11-65|olfactory receptor, family 56, subfamily B, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5736448, end_position_on_the_genomic_accession=5737538, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283078, CurrentID=0, Status='live', Symbol='MKX', Aliases='C10orf48, IFRX, IRXL1', description='mohawk homeobox', other_designations='Iroquois family related homeodomain protein|iroquois homeobox protein-like 1', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27672874, end_position_on_the_genomic_accession=27745849, orientation='minus', exon_count=9, OMIM=601332, no_hdr0=''),
  647654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=647654, CurrentID=0, Status='live', Symbol='CDC27P5', Aliases='', description='cell division cycle 27 pseudogene 5', other_designations='cell division cycle protein 27 homolog pseudogene', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30917666, end_position_on_the_genomic_accession=30919582, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387751, CurrentID=0, Status='live', Symbol='GVINP1', Aliases='GVIN1, GVIN1P, VLIG-1, VLIG1', description='GTPase, very large interferon inducible pseudogene 1', other_designations='GTPase, very large interferon inducible 1, pseudogene|very large inducible GTPase 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6713139, end_position_on_the_genomic_accession=6721879, orientation='minus', exon_count=1, OMIM=616121, no_hdr0=''),
  100526760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526760, CurrentID=0, Status='live', Symbol='ABHD14A-ACY1', Aliases='', description='ABHD14A-ACY1 readthrough', other_designations='ABHD14A-ACY1 readthrough (non-protein coding)', map_location='3p', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51975026, end_position_on_the_genomic_accession=51989202, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100526761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526761, CurrentID=0, Status='live', Symbol='CCDC169-SOHLH2', Aliases='C13orf38-SOHLH2', description='CCDC169-SOHLH2 readthrough', other_designations='', map_location='13q', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36168208, end_position_on_the_genomic_accession=36297855, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105376426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376426, CurrentID=0, Status='live', Symbol='LOC105376426', Aliases='', description='uncharacterized LOC105376426', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=13710018, end_position_on_the_genomic_accession=13730725, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  387755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387755, CurrentID=0, Status='live', Symbol='INSC', Aliases='', description='inscuteable homolog (Drosophila)', other_designations='inscuteable spindle orientation adaptor protein', map_location='11p15.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15111430, end_position_on_the_genomic_accession=15247208, orientation='plus', exon_count=21, OMIM=610668, no_hdr0=''),
  105376428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376428, CurrentID=0, Status='live', Symbol='LOC105376428', Aliases='', description='uncharacterized LOC105376428', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14655127, end_position_on_the_genomic_accession=14660616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376429, CurrentID=0, Status='live', Symbol='LOC105376429', Aliases='', description='uncharacterized LOC105376429', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14652718, end_position_on_the_genomic_accession=14655065, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387758, CurrentID=0, Status='live', Symbol='FIBIN', Aliases='', description='fin bud initiation factor homolog (zebrafish)', other_designations='', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26994081, end_position_on_the_genomic_accession=26997087, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100526767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526767, CurrentID=0, Status='live', Symbol='RNF103-CHMP3', Aliases='RNF103-VPS24', description='RNF103-CHMP3 readthrough', other_designations='RNF103-CHMP3 read-through|RNF103-VPS24 readthrough', map_location='2p', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=86503430, end_position_on_the_genomic_accession=86721122, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376432, CurrentID=0, Status='live', Symbol='LOC105376432', Aliases='', description='protein artemis-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15016256, end_position_on_the_genomic_accession=15073080, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  414152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414152, CurrentID=0, Status='live', Symbol='C10orf105', Aliases='', description='chromosome 10 open reading frame 105', other_designations='', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71711701, end_position_on_the_genomic_accession=71737824, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376434, CurrentID=0, Status='live', Symbol='LOC105376434', Aliases='', description='uncharacterized LOC105376434', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=15172897, end_position_on_the_genomic_accession=15211298, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  256691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256691, CurrentID=0, Status='live', Symbol='MAMDC2', Aliases='', description='MAM domain containing 2', other_designations='MAM domain containing 1|MAM domain-containing proteoglycan|mamcan', map_location='9q21.12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=70043519, end_position_on_the_genomic_accession=70226972, orientation='plus', exon_count=18, OMIM=612879, no_hdr0=''),
  100526772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526772, CurrentID=0, Status='live', Symbol='TMEM110-MUSTN1', Aliases='', description='TMEM110-MUSTN1 readthrough', other_designations='TMEM110-MUSTN1 read-through transcript', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52833115, end_position_on_the_genomic_accession=52897581, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100526773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526773, CurrentID=0, Status='live', Symbol='EPPIN-WFDC6', Aliases='SPINLW1-WFDC6', description='EPPIN-WFDC6 readthrough', other_designations='SPINLW1-WFDC6 read-through transcript|SPINLW1-WFDC6 readthrough', map_location='20q', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45536280, end_position_on_the_genomic_accession=45547426, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376438, CurrentID=0, Status='live', Symbol='LOC105376438', Aliases='', description='uncharacterized LOC105376438', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17695892, end_position_on_the_genomic_accession=17696768, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376439, CurrentID=0, Status='live', Symbol='LOC105376439', Aliases='', description='uncharacterized LOC105376439', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=18512582, end_position_on_the_genomic_accession=18539265, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100526776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526776, CurrentID=0, Status='live', Symbol='CSNK1A1P3', Aliases='', description='casein kinase 1, alpha 1 pseudogene 3', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98833382, end_position_on_the_genomic_accession=98833708, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376441, CurrentID=0, Status='live', Symbol='LOC105376441', Aliases='', description='uncharacterized LOC105376441', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=19156571, end_position_on_the_genomic_accession=19160232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376442, CurrentID=0, Status='live', Symbol='LOC105376442', Aliases='', description='uncharacterized LOC105376442', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=20597130, end_position_on_the_genomic_accession=20610771, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376443, CurrentID=0, Status='live', Symbol='LOC105376443', Aliases='', description='uncharacterized LOC105376443', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21054239, end_position_on_the_genomic_accession=21101141, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376444, CurrentID=0, Status='live', Symbol='LOC105376444', Aliases='', description='uncharacterized LOC105376444', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21368479, end_position_on_the_genomic_accession=21372943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376445, CurrentID=0, Status='live', Symbol='LOC105376445', Aliases='', description='uncharacterized LOC105376445', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=21414787, end_position_on_the_genomic_accession=21418654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376446, CurrentID=0, Status='live', Symbol='LOC105376446', Aliases='', description='uncharacterized LOC105376446', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22225994, end_position_on_the_genomic_accession=22246083, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387775, CurrentID=0, Status='live', Symbol='SLC22A10', Aliases='OAT5, hOAT5', description='solute carrier family 22 member 10', other_designations='organic anion transporter 5|solute carrier family 22 (organic anion/cation transporter), member 10|solute carrier family 22, member 10', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=63268720, end_position_on_the_genomic_accession=63311774, orientation='plus', exon_count=15, OMIM=607580, no_hdr0=''),
  105376448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376448, CurrentID=0, Status='live', Symbol='LOC105376448', Aliases='', description='uncharacterized LOC105376448', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22426822, end_position_on_the_genomic_accession=22428526, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376449, CurrentID=0, Status='live', Symbol='LOC105376449', Aliases='', description='uncharacterized LOC105376449', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22447138, end_position_on_the_genomic_accession=22462623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  387778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387778, CurrentID=0, Status='live', Symbol='SPDYC', Aliases='RINGOC, Ringo2', description='speedy/RINGO cell cycle regulator family member C', other_designations='RINGO C|hSpy/Ringo C|rapid inducer of G2/M progression in oocytes C|speedy C|speedy homolog C', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65169994, end_position_on_the_genomic_accession=65173368, orientation='plus', exon_count=7, OMIM=614030, no_hdr0=''),
  105376451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376451, CurrentID=0, Status='live', Symbol='LOC105376451', Aliases='', description='uncharacterized LOC105376451', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22854084, end_position_on_the_genomic_accession=22868246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376452, CurrentID=0, Status='live', Symbol='LOC105376452', Aliases='', description='uncharacterized LOC105376452', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=22964735, end_position_on_the_genomic_accession=22995934, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649925, CurrentID=0, Status='live', Symbol='SETP16', Aliases='', description='SET pseudogene 16', other_designations='SET nuclear oncogene pseudogene', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=118833797, end_position_on_the_genomic_accession=118834565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256710, CurrentID=0, Status='live', Symbol='GLIPR1L1', Aliases='ALKN2972, PRO7434', description='GLI pathogenesis-related 1 like 1', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=75334639, end_position_on_the_genomic_accession=75370390, orientation='plus', exon_count=7, OMIM=610395, no_hdr0=''),
  100526791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526791, CurrentID=0, Status='live', Symbol='DEL19P13.13', Aliases='C19DELp13.13, C19DUPp13.13, DUP19p13.13', description='Chromosome 19p13.13 deletion syndrome', other_designations='', map_location='19p13.13', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613638, no_hdr0=''),
  105376456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376456, CurrentID=0, Status='live', Symbol='LOC105376456', Aliases='', description='uncharacterized LOC105376456', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=24753408, end_position_on_the_genomic_accession=24778075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100526793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526793, CurrentID=0, Status='live', Symbol='CKS1BP5', Aliases='', description='CDC28 protein kinase regulatory subunit 1B pseudogene 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=144439143, end_position_on_the_genomic_accession=144439962, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  256714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256714, CurrentID=0, Status='live', Symbol='MAP7D2', Aliases='', description='MAP7 domain containing 2', other_designations='MAP7 domain containing 2 brain|MAP7 domain containing 2 protein varaint 1|MAP7 domain containing 2 protein varaint 2|MAP7 domain-containing protein 2 brain specific variant 1|MAP7 domain-containing protein 2 brain specific variant 2', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=20006713, end_position_on_the_genomic_accession=20116996, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  387787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387787, CurrentID=0, Status='live', Symbol='LIPT2', Aliases='', description='lipoyl(octanoyl) transferase 2 (putative)', other_designations='lipoate-protein ligase B|lipoyl/octanoyl transferase|octanoyl-[acyl-carrier-protein]-protein N-octanoyltransferase|putative octanoyltransferase, mitochondrial', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=74490727, end_position_on_the_genomic_accession=74493719, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376797, CurrentID=0, Status='live', Symbol='LOC105376797', Aliases='', description='uncharacterized LOC105376797', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4565752, end_position_on_the_genomic_accession=4572178, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376461, CurrentID=0, Status='live', Symbol='LOC105376461', Aliases='', description='uncharacterized LOC105376461', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26637822, end_position_on_the_genomic_accession=26642958, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376462, CurrentID=0, Status='live', Symbol='LOC105376462', Aliases='', description='uncharacterized LOC105376462', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=26663175, end_position_on_the_genomic_accession=26675736, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649935, CurrentID=0, Status='live', Symbol='LOC649935', Aliases='', description='UPF0607 protein ENSP00000381514-like', other_designations='Putative UPF0607 protein ENSP00000381514', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.10', start_position_on_the_genomic_accession=68608956, end_position_on_the_genomic_accession=68610190, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376464, CurrentID=0, Status='live', Symbol='LOC105376464', Aliases='', description='uncharacterized LOC105376464', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=27774954, end_position_on_the_genomic_accession=27796423, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376466, CurrentID=0, Status='live', Symbol='LOC105376466', Aliases='', description='uncharacterized LOC105376466', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28303162, end_position_on_the_genomic_accession=28306271, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378875, CurrentID=0, Status='live', Symbol='LOC105378875', Aliases='', description='uncharacterized LOC105378875', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=102676778, end_position_on_the_genomic_accession=102680950, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376468, CurrentID=0, Status='live', Symbol='LOC105376468', Aliases='', description='uncharacterized LOC105376468', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28433008, end_position_on_the_genomic_accession=28436221, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376469, CurrentID=0, Status='live', Symbol='LOC105376469', Aliases='', description='uncharacterized LOC105376469', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28485011, end_position_on_the_genomic_accession=28488022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376470, CurrentID=0, Status='live', Symbol='LOC105376470', Aliases='', description='uncharacterized LOC105376470', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28708104, end_position_on_the_genomic_accession=28712450, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376471, CurrentID=0, Status='live', Symbol='LOC105376471', Aliases='', description='uncharacterized LOC105376471', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29108194, end_position_on_the_genomic_accession=29133234, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376472, CurrentID=0, Status='live', Symbol='LOC105376472', Aliases='', description='uncharacterized LOC105376472', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29229433, end_position_on_the_genomic_accession=29235506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376473, CurrentID=0, Status='live', Symbol='LOC105376473', Aliases='', description='uncharacterized LOC105376473', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29373026, end_position_on_the_genomic_accession=29374919, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  649946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=649946, CurrentID=0, Status='live', Symbol='RPL23AP64', Aliases='RPL23A_25_1182', description='ribosomal protein L23a pseudogene 64', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=119002953, end_position_on_the_genomic_accession=119003593, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376475, CurrentID=0, Status='live', Symbol='LOC105376475', Aliases='', description='uncharacterized LOC105376475', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=29782397, end_position_on_the_genomic_accession=29784039, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387804, CurrentID=0, Status='live', Symbol='VSTM5', Aliases='C11orf90', description='V-set and transmembrane domain containing 5', other_designations='', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93820569, end_position_on_the_genomic_accession=93850502, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376477, CurrentID=0, Status='live', Symbol='LOC105376477', Aliases='', description='uncharacterized LOC105376477', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30197081, end_position_on_the_genomic_accession=30206208, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376478, CurrentID=0, Status='live', Symbol='LOC105376478', Aliases='', description='uncharacterized LOC105376478', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30529557, end_position_on_the_genomic_accession=30533734, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376479, CurrentID=0, Status='live', Symbol='LOC105376479', Aliases='', description='uncharacterized LOC105376479', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30587717, end_position_on_the_genomic_accession=30590045, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376480, CurrentID=0, Status='live', Symbol='LOC105376480', Aliases='', description='uncharacterized LOC105376480', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=30723219, end_position_on_the_genomic_accession=30723740, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376481, CurrentID=0, Status='live', Symbol='LOC105376481', Aliases='', description='uncharacterized LOC105376481', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31033023, end_position_on_the_genomic_accession=31113399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387810, CurrentID=0, Status='live', Symbol='LOC387810', Aliases='', description='uncharacterized LOC387810', other_designations='', map_location='11q23.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=112534220, end_position_on_the_genomic_accession=112555802, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376483, CurrentID=0, Status='live', Symbol='LOC105376483', Aliases='', description='uncharacterized LOC105376483', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31134265, end_position_on_the_genomic_accession=31139329, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100526820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526820, CurrentID=0, Status='live', Symbol='CAHM', Aliases='LINC00468', description='colon adenocarcinoma hypermethylated (non-protein coding)', other_designations='long intergenic non-protein coding RNA 468', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163413065, end_position_on_the_genomic_accession=163413950, orientation='minus', exon_count=1, OMIM=615930, no_hdr0=''),
  105376485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376485, CurrentID=0, Status='live', Symbol='LOC105376485', Aliases='', description='uncharacterized LOC105376485', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=31980788, end_position_on_the_genomic_accession=31992578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376486, CurrentID=0, Status='live', Symbol='LOC105376486', Aliases='', description='uncharacterized LOC105376486', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=32887259, end_position_on_the_genomic_accession=32895979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100526823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526823, CurrentID=0, Status='live', Symbol='TST2', Aliases='', description='Tuberculin skin test reactivity QTL', other_designations='', map_location='5p15', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613637, no_hdr0=''),
  105376488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376488, CurrentID=0, Status='live', Symbol='LOC105376488', Aliases='', description='uncharacterized LOC105376488', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33336755, end_position_on_the_genomic_accession=33384153, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100526825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526825, CurrentID=0, Status='live', Symbol='INMT-FAM188B', Aliases='', description='INMT-FAM188B readthrough (NMD candidate)', other_designations='INMT-FAM188B readthrough (non-protein coding)', map_location='7p', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=30752135, end_position_on_the_genomic_accession=30892387, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  105376490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376490, CurrentID=0, Status='live', Symbol='LOC105376490', Aliases='', description='uncharacterized LOC105376490', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33579036, end_position_on_the_genomic_accession=33594205, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376491, CurrentID=0, Status='live', Symbol='LOC105376491', Aliases='', description='uncharacterized LOC105376491', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=34973018, end_position_on_the_genomic_accession=34975578, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100133612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133612, CurrentID=0, Status='live', Symbol='LINC01134', Aliases='', description='long intergenic non-protein coding RNA 1134', other_designations='', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3900404, end_position_on_the_genomic_accession=3915447, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376493, CurrentID=0, Status='live', Symbol='LOC105376493', Aliases='', description='uncharacterized LOC105376493', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=35219803, end_position_on_the_genomic_accession=35244679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100526830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526830, CurrentID=0, Status='live', Symbol='SLX1A-SULT1A3', Aliases='', description='SLX1A-SULT1A3 readthrough (NMD candidate)', other_designations='SLX1A-SULT1A3 readthrough (non-protein coding)', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30194433, end_position_on_the_genomic_accession=30204329, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100526831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526831, CurrentID=0, Status='live', Symbol='SLX1B-SULT1A4', Aliases='', description='SLX1B-SULT1A4 readthrough (NMD candidate)', other_designations='', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29455091, end_position_on_the_genomic_accession=29464980, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100526832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526832, CurrentID=0, Status='live', Symbol='PHOSPHO2-KLHL23', Aliases='', description='PHOSPHO2-KLHL23 readthrough', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169694454, end_position_on_the_genomic_accession=169751886, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100526833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526833, CurrentID=0, Status='live', Symbol='SEPT5-GP1BB', Aliases='', description='SEPT5-GP1BB readthrough', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19717220, end_position_on_the_genomic_accession=19724774, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105376498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376498, CurrentID=0, Status='live', Symbol='LOC105376498', Aliases='', description='uncharacterized LOC105376498', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37677097, end_position_on_the_genomic_accession=37680351, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100526835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526835, CurrentID=0, Status='live', Symbol='FPGT-TNNI3K', Aliases='', description='FPGT-TNNI3K readthrough', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74198212, end_position_on_the_genomic_accession=74544432, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  100526836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526836, CurrentID=0, Status='live', Symbol='BLOC1S5-TXNDC5', Aliases='MUTED-TXNDC5', description='BLOC1S5-TXNDC5 readthrough (NMD candidate)', other_designations='BLOC1S5-TXNDC5 readthrough (non-protein coding)|MUTED-TXNDC5 readthrough (non-protein coding)', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=7881250, end_position_on_the_genomic_accession=8064414, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100526837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526837, CurrentID=0, Status='live', Symbol='EEF1E1-BLOC1S5', Aliases='EEF1E1-MUTED', description='EEF1E1-BLOC1S5 readthrough (NMD candidate)', other_designations='EEF1E1-MUTED readthrough', map_location='6p', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=8013567, end_position_on_the_genomic_accession=8102595, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376502, CurrentID=0, Status='live', Symbol='LOC105376502', Aliases='', description='microtubule cross-linking factor 1-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8046827, end_position_on_the_genomic_accession=8050862, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376503, CurrentID=0, Status='live', Symbol='LOC105376503', Aliases='', description='uncharacterized LOC105376503', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=283062, end_position_on_the_genomic_accession=288263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376504, CurrentID=0, Status='live', Symbol='LOC105376504', Aliases='', description='uncharacterized LOC105376504', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=325703, end_position_on_the_genomic_accession=327479, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376505, CurrentID=0, Status='live', Symbol='LOC105376505', Aliases='', description='uncharacterized LOC105376505', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=330065, end_position_on_the_genomic_accession=332192, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100526842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100526842, CurrentID=0, Status='live', Symbol='RPL17-C18orf32', Aliases='', description='RPL17-C18orf32 readthrough', other_designations='', map_location='18q', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=49481178, end_position_on_the_genomic_accession=49492565, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376507, CurrentID=0, Status='live', Symbol='LOC105376507', Aliases='', description='uncharacterized LOC105376507', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=513844, end_position_on_the_genomic_accession=515068, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387836, CurrentID=0, Status='live', Symbol='CLEC2A', Aliases='INPE5792, KACL, PILAR, UNQ5792', description='C-type lectin domain family 2 member A', other_designations='C-type lectin domain family 2, member A|keratinocyte-associated C-type lectin|proliferation-induced lymphocyte-associated receptor', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9880903, end_position_on_the_genomic_accession=9932407, orientation='minus', exon_count=8, OMIM=612087, no_hdr0=''),
  387837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387837, CurrentID=0, Status='live', Symbol='CLEC12B', Aliases='UNQ5782', description='C-type lectin domain family 12 member B', other_designations='C-type lectin domain family 12, member B|macrophage antigen h', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10009047, end_position_on_the_genomic_accession=10018800, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376510, CurrentID=0, Status='live', Symbol='LOC105376510', Aliases='', description='uncharacterized LOC105376510', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=881306, end_position_on_the_genomic_accession=881569, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376511, CurrentID=0, Status='live', Symbol='LOC105376511', Aliases='', description='uncharacterized LOC105376511', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1061345, end_position_on_the_genomic_accession=1063848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376512, CurrentID=0, Status='live', Symbol='LOC105376512', Aliases='', description='uncharacterized LOC105376512', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1298146, end_position_on_the_genomic_accession=1302574, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387841, CurrentID=0, Status='live', Symbol='RPL13AP20', Aliases='RPL13A_9_1211', description='ribosomal protein L13a pseudogene 20', other_designations='hCG27695', map_location='12p13.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=12875477, end_position_on_the_genomic_accession=12876136, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376514, CurrentID=0, Status='live', Symbol='LOC105376514', Aliases='', description='uncharacterized LOC105376514', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1454629, end_position_on_the_genomic_accession=1456337, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376515, CurrentID=0, Status='live', Symbol='LOC105376515', Aliases='', description='uncharacterized LOC105376515', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1451348, end_position_on_the_genomic_accession=1454380, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376516, CurrentID=0, Status='live', Symbol='LOC105376516', Aliases='', description='uncharacterized LOC105376516', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1424113, end_position_on_the_genomic_accession=1429630, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387845, CurrentID=0, Status='live', Symbol='EEF1A1P16', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 16', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=16990670, end_position_on_the_genomic_accession=16992082, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376518, CurrentID=0, Status='live', Symbol='LOC105376518', Aliases='', description='uncharacterized LOC105376518', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2310754, end_position_on_the_genomic_accession=2313647, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376519, CurrentID=0, Status='live', Symbol='LOC105376519', Aliases='', description='uncharacterized LOC105376519', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2353059, end_position_on_the_genomic_accession=2354434, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  125704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125704, CurrentID=0, Status='live', Symbol='FAM69C', Aliases='C18orf51', description='family with sequence similarity 69 member C', other_designations='family with sequence similarity 69, member C', map_location='18q22.3', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=74435728, end_position_on_the_genomic_accession=74464715, orientation='minus', exon_count=6, OMIM=614544, no_hdr0=''),
  387849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387849, CurrentID=0, Status='live', Symbol='REP15', Aliases='RAB15EP', description='RAB15 effector protein', other_designations='', map_location='12p11.22', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=27696495, end_position_on_the_genomic_accession=27697633, orientation='plus', exon_count=1, OMIM=610848, no_hdr0=''),
  105376522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376522, CurrentID=0, Status='live', Symbol='LOC105376522', Aliases='', description='uncharacterized LOC105376522', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2793768, end_position_on_the_genomic_accession=2799443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376523, CurrentID=0, Status='live', Symbol='LOC105376523', Aliases='', description='uncharacterized LOC105376523', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=2680734, end_position_on_the_genomic_accession=2697018, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376524, CurrentID=0, Status='live', Symbol='LOC105376524', Aliases='', description='uncharacterized LOC105376524', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3057504, end_position_on_the_genomic_accession=3063242, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  654466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654466, CurrentID=0, Status='live', Symbol='KGFLP2', Aliases='', description='keratinocyte growth factor-like protein 2', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39813784, end_position_on_the_genomic_accession=39874566, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376526, CurrentID=0, Status='live', Symbol='LOC105376526', Aliases='', description='uncharacterized LOC105376526', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3473571, end_position_on_the_genomic_accession=3478053, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376527, CurrentID=0, Status='live', Symbol='LOC105376527', Aliases='', description='uncharacterized LOC105376527', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3854323, end_position_on_the_genomic_accession=3855603, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387856, CurrentID=0, Status='live', Symbol='CCDC184', Aliases='C12orf68', description='coiled-coil domain containing 184', other_designations='', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48183583, end_position_on_the_genomic_accession=48185926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376529, CurrentID=0, Status='live', Symbol='LOC105376529', Aliases='', description='uncharacterized LOC105376529', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5509587, end_position_on_the_genomic_accession=5513464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376530, CurrentID=0, Status='live', Symbol='LOC105376530', Aliases='', description='uncharacterized LOC105376530', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6362782, end_position_on_the_genomic_accession=6371269, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376531, CurrentID=0, Status='live', Symbol='LOC105376531', Aliases='', description='uncharacterized LOC105376531', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6733720, end_position_on_the_genomic_accession=6746480, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376532, CurrentID=0, Status='live', Symbol='LOC105376532', Aliases='', description='uncharacterized LOC105376532', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=6690914, end_position_on_the_genomic_accession=6704471, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376533, CurrentID=0, Status='live', Symbol='LOC105376533', Aliases='', description='uncharacterized LOC105376533', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7701745, end_position_on_the_genomic_accession=7721067, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376534, CurrentID=0, Status='live', Symbol='LOC105376534', Aliases='', description='uncharacterized LOC105376534', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7676470, end_position_on_the_genomic_accession=7700287, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  414169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414169, CurrentID=0, Status='live', Symbol='BANF1P2', Aliases='BANF1L1, bA122K13.8', description='barrier to autointegration factor 1 pseudogene 2', other_designations='barrier to autointegration factor 1-like 1', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133344278, end_position_on_the_genomic_accession=133344999, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376536, CurrentID=0, Status='live', Symbol='LOC105376536', Aliases='', description='uncharacterized LOC105376536', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8229885, end_position_on_the_genomic_accession=8233020, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376537, CurrentID=0, Status='live', Symbol='LOC105376537', Aliases='', description='uncharacterized LOC105376537', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8317218, end_position_on_the_genomic_accession=8335622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376538, CurrentID=0, Status='live', Symbol='LOC105376538', Aliases='', description='uncharacterized LOC105376538', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8659429, end_position_on_the_genomic_accession=8665617, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387867, CurrentID=0, Status='live', Symbol='RPSAP12', Aliases='LAMR1P12, LAMRL2, LAMRP2, RPSA_18_1254', description='ribosomal protein SA pseudogene 12', other_designations='laminin receptor 1 pseudogene 12|laminin receptor-like 2', map_location='12q15', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=68552925, end_position_on_the_genomic_accession=68553958, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376540, CurrentID=0, Status='live', Symbol='LOC105376540', Aliases='', description='uncharacterized LOC105376540', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=8932959, end_position_on_the_genomic_accession=8936954, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387869, CurrentID=0, Status='live', Symbol='LOC387869', Aliases='', description='microtubule associated protein 1 light chain 3 beta pseudogene', other_designations='microtubule-associated protein 1 light chain 3 beta pseudogene', map_location='12q21.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=74663768, end_position_on_the_genomic_accession=74664141, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376542, CurrentID=0, Status='live', Symbol='LOC105376542', Aliases='', description='uncharacterized LOC105376542', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9116175, end_position_on_the_genomic_accession=9121858, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376543, CurrentID=0, Status='live', Symbol='LOC105376543', Aliases='', description='uncharacterized LOC105376543', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=9539854, end_position_on_the_genomic_accession=9556816, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376544, CurrentID=0, Status='live', Symbol='LOC105376544', Aliases='', description='uncharacterized LOC105376544', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10301745, end_position_on_the_genomic_accession=10304753, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376545, CurrentID=0, Status='live', Symbol='LOC105376545', Aliases='', description='uncharacterized LOC105376545', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10811284, end_position_on_the_genomic_accession=10825690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376546, CurrentID=0, Status='live', Symbol='LOC105376546', Aliases='', description='uncharacterized LOC105376546', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10843991, end_position_on_the_genomic_accession=10847945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376547, CurrentID=0, Status='live', Symbol='LOC105376547', Aliases='', description='uncharacterized LOC105376547', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10834837, end_position_on_the_genomic_accession=10838454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100133668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133668, CurrentID=0, Status='live', Symbol='RPS27P24', Aliases='RPS27_10_1247', description='ribosomal protein S27 pseudogene 24', other_designations='', map_location='12q14.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=63760303, end_position_on_the_genomic_accession=63760643, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100133669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133669, CurrentID=0, Status='live', Symbol='LOC100133669', Aliases='', description='uncharacterized LOC100133669', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=142982031, end_position_on_the_genomic_accession=143018390, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376550, CurrentID=0, Status='live', Symbol='LOC105376550', Aliases='', description='uncharacterized LOC105376550', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11073774, end_position_on_the_genomic_accession=11177321, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376551, CurrentID=0, Status='live', Symbol='LOC105376551', Aliases='', description='uncharacterized LOC105376551', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11475188, end_position_on_the_genomic_accession=11476452, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376552, CurrentID=0, Status='live', Symbol='LOC105376552', Aliases='', description='uncharacterized LOC105376552', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11293731, end_position_on_the_genomic_accession=11299188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376553, CurrentID=0, Status='live', Symbol='LOC105376553', Aliases='', description='uncharacterized LOC105376553', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=11635326, end_position_on_the_genomic_accession=11642781, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387882, CurrentID=0, Status='live', Symbol='C12orf75', Aliases='AGD3, OCC-1, OCC1', description='chromosome 12 open reading frame 75', other_designations='adipogenesis down-regulated 3|overexpressed in colon carcinoma-1|putative overexpressed in colon carcinoma-1 protein variant C', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=105330636, end_position_on_the_genomic_accession=105371518, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376555, CurrentID=0, Status='live', Symbol='LOC105376555', Aliases='', description='uncharacterized LOC105376555', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12213417, end_position_on_the_genomic_accession=12224656, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376556, CurrentID=0, Status='live', Symbol='LOC105376556', Aliases='', description='uncharacterized LOC105376556', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=12619302, end_position_on_the_genomic_accession=12669097, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  387885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387885, CurrentID=0, Status='live', Symbol='CFAP73', Aliases='CCDC42B, MIA2', description='cilia and flagella associated protein 73', other_designations='coiled-coil domain containing 42B', map_location='12q24.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=113149181, end_position_on_the_genomic_accession=113159276, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105376558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376558, CurrentID=0, Status='live', Symbol='LOC105376558', Aliases='', description='uncharacterized LOC105376558', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13111503, end_position_on_the_genomic_accession=13112981, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  256815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256815, CurrentID=0, Status='live', Symbol='C10orf67', Aliases='C10orf115, LINC01552', description='chromosome 10 open reading frame 67', other_designations='', map_location='10p12.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=23203816, end_position_on_the_genomic_accession=23239792, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  105376560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376560, CurrentID=0, Status='live', Symbol='LOC105376560', Aliases='', description='uncharacterized LOC105376560', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13271757, end_position_on_the_genomic_accession=13277572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376561, CurrentID=0, Status='live', Symbol='LOC105376561', Aliases='', description='uncharacterized LOC105376561', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13844758, end_position_on_the_genomic_accession=13872421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  387890 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387890, CurrentID=0, Status='live', Symbol='TMEM233', Aliases='DSPB2, IFITMD2', description='transmembrane protein 233', other_designations='dispanin subfamily B member 2|interferon induced transmembrane protein domain containing 2|interferon-induced transmembrane domain-containing protein D2', map_location='12q24.23', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=119592636, end_position_on_the_genomic_accession=119654442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376563, CurrentID=0, Status='live', Symbol='LOC105376563', Aliases='', description='uncharacterized LOC105376563', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13966451, end_position_on_the_genomic_accession=13969081, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376564, CurrentID=0, Status='live', Symbol='LOC105376564', Aliases='', description='uncharacterized LOC105376564', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=14451831, end_position_on_the_genomic_accession=14453257, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387893, CurrentID=0, Status='live', Symbol='KMT5A', Aliases='PR-Set7, SET07, SET8, SETD8', description='lysine (K)-specific methyltransferase 5A', other_designations='H4-K20-HMTase SETD8|H4-K20-specific histone methyltransferase|H4K20-specific histone methyltransferase splice variant Set8b|PR/SET domain containing protein 8|PR/SET domain-containing protein 07|PR/SET07|SET domain containing (lysine methyltransferase) 8|SET domain-containing protein 8|histone-lysine N-methyltransferase SETD8|lysine N-methyltransferase 5A', map_location='12q24.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123383778, end_position_on_the_genomic_accession=123409358, orientation='plus', exon_count=10, OMIM=607240, no_hdr0=''),
  105376566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376566, CurrentID=0, Status='live', Symbol='LOC105376566', Aliases='', description='uncharacterized LOC105376566', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15478934, end_position_on_the_genomic_accession=15483131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387895, CurrentID=0, Status='live', Symbol='LINC00944', Aliases='', description='long intergenic non-protein coding RNA 944', other_designations='', map_location='12q24.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=126730701, end_position_on_the_genomic_accession=126772262, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376568, CurrentID=0, Status='live', Symbol='LOC105376568', Aliases='', description='uncharacterized LOC105376568', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15571836, end_position_on_the_genomic_accession=15623691, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376569, CurrentID=0, Status='live', Symbol='LOC105376569', Aliases='', description='uncharacterized LOC105376569', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15562156, end_position_on_the_genomic_accession=15604112, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376570, CurrentID=0, Status='live', Symbol='LOC105376570', Aliases='', description='uncharacterized LOC105376570', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=15910572, end_position_on_the_genomic_accession=15930917, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376571, CurrentID=0, Status='live', Symbol='LOC105376571', Aliases='', description='uncharacterized LOC105376571', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16613023, end_position_on_the_genomic_accession=16655850, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376572, CurrentID=0, Status='live', Symbol='LOC105376572', Aliases='', description='uncharacterized LOC105376572', other_designations='CTD-3096P4.1', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16022699, end_position_on_the_genomic_accession=16047546, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376573, CurrentID=0, Status='live', Symbol='LOC105376573', Aliases='', description='uncharacterized LOC105376573', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=16768344, end_position_on_the_genomic_accession=16776362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376574, CurrentID=0, Status='live', Symbol='LOC105376574', Aliases='', description='uncharacterized LOC105376574', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17014750, end_position_on_the_genomic_accession=17020512, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376575, CurrentID=0, Status='live', Symbol='LOC105376575', Aliases='', description='uncharacterized LOC105376575', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17333650, end_position_on_the_genomic_accession=17337515, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376576, CurrentID=0, Status='live', Symbol='LOC105376576', Aliases='', description='uncharacterized LOC105376576', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=17350698, end_position_on_the_genomic_accession=17351866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376577, CurrentID=0, Status='live', Symbol='LOC105376577', Aliases='', description='uncharacterized LOC105376577', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18349238, end_position_on_the_genomic_accession=18350914, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376578, CurrentID=0, Status='live', Symbol='LOC105376578', Aliases='', description='uncharacterized LOC105376578', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18634870, end_position_on_the_genomic_accession=18635997, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376579, CurrentID=0, Status='live', Symbol='LOC105376579', Aliases='', description='uncharacterized LOC105376579', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18706325, end_position_on_the_genomic_accession=18708821, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376580, CurrentID=0, Status='live', Symbol='LOC105376580', Aliases='', description='uncharacterized LOC105376580', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19297174, end_position_on_the_genomic_accession=19310713, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376581, CurrentID=0, Status='live', Symbol='LOC105376581', Aliases='', description='uncharacterized LOC105376581', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20008911, end_position_on_the_genomic_accession=20024779, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376582, CurrentID=0, Status='live', Symbol='LOC105376582', Aliases='', description='uncharacterized LOC105376582', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19505947, end_position_on_the_genomic_accession=19510279, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  387911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387911, CurrentID=0, Status='live', Symbol='C1QTNF9B', Aliases='CTRP9B', description='C1q and tumor necrosis factor related protein 9B', other_designations='C1q/TNF-related protein 9B|collagen triple helix repeat-containing', map_location='13q12.12', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=23891099, end_position_on_the_genomic_accession=23898503, orientation='minus', exon_count=7, OMIM=614148, no_hdr0=''),
  105376584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376584, CurrentID=0, Status='live', Symbol='LOC105376584', Aliases='', description='uncharacterized LOC105376584', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20560068, end_position_on_the_genomic_accession=20572998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376585, CurrentID=0, Status='live', Symbol='LOC105376585', Aliases='', description='uncharacterized LOC105376585', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=20880672, end_position_on_the_genomic_accession=20926897, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387914, CurrentID=0, Status='live', Symbol='SHISA2', Aliases='C13orf13, PRO28631, TMEM46, WGAR9166, bA398O19.2, hShisa', description='shisa family member 2', other_designations='shisa homolog 2|transmembrane protein 46', map_location='13q12.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=26044597, end_position_on_the_genomic_accession=26059587, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  650059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650059, CurrentID=0, Status='live', Symbol='RPL21P82', Aliases='RPL21_34_997', description='ribosomal protein L21 pseudogene 82', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92006024, end_position_on_the_genomic_accession=92006577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376588, CurrentID=0, Status='live', Symbol='LOC105376588', Aliases='', description='uncharacterized LOC105376588', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22492131, end_position_on_the_genomic_accession=22510387, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376589, CurrentID=0, Status='live', Symbol='LOC105376589', Aliases='', description='uncharacterized LOC105376589', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22489532, end_position_on_the_genomic_accession=22491469, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376590, CurrentID=0, Status='live', Symbol='LOC105376590', Aliases='', description='uncharacterized LOC105376590', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22829476, end_position_on_the_genomic_accession=22885695, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376591, CurrentID=0, Status='live', Symbol='LOC105376591', Aliases='', description='uncharacterized LOC105376591', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23116579, end_position_on_the_genomic_accession=23163139, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376592, CurrentID=0, Status='live', Symbol='LOC105376592', Aliases='', description='uncharacterized LOC105376592', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=23163563, end_position_on_the_genomic_accession=23192385, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  387921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387921, CurrentID=0, Status='live', Symbol='NHLRC3', Aliases='', description='NHL repeat containing 3', other_designations='', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=39038311, end_position_on_the_genomic_accession=39050109, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376594, CurrentID=0, Status='live', Symbol='LOC105376594', Aliases='', description='uncharacterized LOC105376594', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=24259385, end_position_on_the_genomic_accession=24262168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387923, CurrentID=0, Status='live', Symbol='SERP2', Aliases='C13orf21, bA269C23.1', description='stress-associated endoplasmic reticulum protein family member 2', other_designations='ribosome-associated membrane protein RAMP4-2', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=44373842, end_position_on_the_genomic_accession=44397714, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  387924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387924, CurrentID=0, Status='live', Symbol='OGFOD1P1', Aliases='', description='2-oxoglutarate and iron-dependent oxygenase domain containing 1 pseudogene 1', other_designations='', map_location='13q14.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=49210090, end_position_on_the_genomic_accession=49212314, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376597, CurrentID=0, Status='live', Symbol='LOC105376597', Aliases='', description='uncharacterized LOC105376597', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=25853804, end_position_on_the_genomic_accession=25924375, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376598, CurrentID=0, Status='live', Symbol='LOC105376598', Aliases='', description='uncharacterized LOC105376598', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26328603, end_position_on_the_genomic_accession=26332111, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376599, CurrentID=0, Status='live', Symbol='LOC105376599', Aliases='', description='uncharacterized LOC105376599', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=26287334, end_position_on_the_genomic_accession=26294073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376600, CurrentID=0, Status='live', Symbol='LOC105376600', Aliases='', description='uncharacterized LOC105376600', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27219748, end_position_on_the_genomic_accession=27235670, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376601, CurrentID=0, Status='live', Symbol='LOC105376601', Aliases='', description='uncharacterized LOC105376601', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=27294005, end_position_on_the_genomic_accession=27296085, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376602, CurrentID=0, Status='live', Symbol='LOC105376602', Aliases='', description='uncharacterized LOC105376602', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28661235, end_position_on_the_genomic_accession=28679640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376603, CurrentID=0, Status='live', Symbol='LOC105376603', Aliases='', description='uncharacterized LOC105376603', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28902282, end_position_on_the_genomic_accession=29063683, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376604, CurrentID=0, Status='live', Symbol='LOC105376604', Aliases='', description='uncharacterized LOC105376604', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=28894431, end_position_on_the_genomic_accession=28941029, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  387933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387933, CurrentID=0, Status='live', Symbol='HNRNPA3P5', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 5', other_designations='', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=65787544, end_position_on_the_genomic_accession=65788763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387934, CurrentID=0, Status='live', Symbol='FABP5P1', Aliases='FABP5, FABP5L1', description='fatty acid binding protein 5 pseudogene 1', other_designations='fatty acid binding protein 5-like 1 (pseudogene)', map_location='13q22.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=73100363, end_position_on_the_genomic_accession=73101045, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101929445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929445, CurrentID=0, Status='live', Symbol='LOC101929445', Aliases='', description='uncharacterized LOC101929445', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42871521, end_position_on_the_genomic_accession=42874060, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376608, CurrentID=0, Status='live', Symbol='LOC105376608', Aliases='', description='uncharacterized LOC105376608', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30363241, end_position_on_the_genomic_accession=30374007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376609, CurrentID=0, Status='live', Symbol='LOC105376609', Aliases='', description='uncharacterized LOC105376609', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30584149, end_position_on_the_genomic_accession=30630436, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376610, CurrentID=0, Status='live', Symbol='LOC105376610', Aliases='', description='uncharacterized LOC105376610', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=30890119, end_position_on_the_genomic_accession=30904873, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387939, CurrentID=0, Status='live', Symbol='LOC387939', Aliases='', description='hCG2045417', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=84592560, end_position_on_the_genomic_accession=84611629, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376612, CurrentID=0, Status='live', Symbol='LOC105376612', Aliases='', description='uncharacterized LOC105376612', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32003042, end_position_on_the_genomic_accession=32008054, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376613, CurrentID=0, Status='live', Symbol='LOC105376613', Aliases='', description='uncharacterized LOC105376613', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32041964, end_position_on_the_genomic_accession=32047124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376614, CurrentID=0, Status='live', Symbol='LOC105376614', Aliases='', description='uncharacterized LOC105376614', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32702285, end_position_on_the_genomic_accession=32715828, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376615, CurrentID=0, Status='live', Symbol='LOC105376615', Aliases='', description='uncharacterized LOC105376615', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=32956200, end_position_on_the_genomic_accession=33015362, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376616, CurrentID=0, Status='live', Symbol='LOC105376616', Aliases='', description='uncharacterized LOC105376616', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33080606, end_position_on_the_genomic_accession=33084577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376617, CurrentID=0, Status='live', Symbol='LOC105376617', Aliases='', description='uncharacterized LOC105376617', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33669559, end_position_on_the_genomic_accession=33697150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376618, CurrentID=0, Status='live', Symbol='LOC105376618', Aliases='', description='uncharacterized LOC105376618', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33654288, end_position_on_the_genomic_accession=33656045, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929447, CurrentID=0, Status='live', Symbol='LOC101929447', Aliases='', description='uncharacterized LOC101929447', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=33073729, end_position_on_the_genomic_accession=33116806, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376620 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376620, CurrentID=0, Status='live', Symbol='LOC105376620', Aliases='', description='uncharacterized LOC105376620', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33814141, end_position_on_the_genomic_accession=33822384, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376621, CurrentID=0, Status='live', Symbol='LOC105376621', Aliases='', description='uncharacterized LOC105376621', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33893959, end_position_on_the_genomic_accession=33911209, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376622, CurrentID=0, Status='live', Symbol='LOC105376622', Aliases='', description='uncharacterized LOC105376622', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34432361, end_position_on_the_genomic_accession=34436378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376623, CurrentID=0, Status='live', Symbol='LOC105376623', Aliases='', description='uncharacterized LOC105376623', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34570888, end_position_on_the_genomic_accession=34578608, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256880, CurrentID=0, Status='live', Symbol='LOC256880', Aliases='', description='uncharacterized LOC256880', other_designations='', map_location='4q23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=99950494, end_position_on_the_genomic_accession=100037705, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376625, CurrentID=0, Status='live', Symbol='LOC105376625', Aliases='', description='uncharacterized LOC105376625', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=34824571, end_position_on_the_genomic_accession=34825834, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376626, CurrentID=0, Status='live', Symbol='LOC105376626', Aliases='', description='uncharacterized LOC105376626', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35050136, end_position_on_the_genomic_accession=35067254, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  130195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130195, CurrentID=0, Status='live', Symbol='HSPA8P6', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 6', other_designations='heat shock 70kDa protein 8 pseudogene 6', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=209072600, end_position_on_the_genomic_accession=209074151, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376628, CurrentID=0, Status='live', Symbol='LOC105376628', Aliases='', description='uncharacterized LOC105376628', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=35176747, end_position_on_the_genomic_accession=35186825, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376629, CurrentID=0, Status='live', Symbol='LOC105376629', Aliases='', description='uncharacterized LOC105376629', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36420742, end_position_on_the_genomic_accession=36428106, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376630, CurrentID=0, Status='live', Symbol='LOC105376630', Aliases='', description='uncharacterized LOC105376630', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36148180, end_position_on_the_genomic_accession=36151184, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376631, CurrentID=0, Status='live', Symbol='LOC105376631', Aliases='', description='uncharacterized LOC105376631', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=36709049, end_position_on_the_genomic_accession=36710990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376632, CurrentID=0, Status='live', Symbol='LOC105376632', Aliases='', description='uncharacterized LOC105376632', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=37471812, end_position_on_the_genomic_accession=37489050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376633, CurrentID=0, Status='live', Symbol='LOC105376633', Aliases='', description='uncharacterized LOC105376633', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=37938601, end_position_on_the_genomic_accession=37954663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376634, CurrentID=0, Status='live', Symbol='LOC105376634', Aliases='', description='uncharacterized LOC105376634', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38200942, end_position_on_the_genomic_accession=38337044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376635, CurrentID=0, Status='live', Symbol='LOC105376635', Aliases='', description='uncharacterized LOC105376635', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=38646466, end_position_on_the_genomic_accession=38686693, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  256892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256892, CurrentID=0, Status='live', Symbol='OR51F1', Aliases='OR11-21P, OR51F1', description='olfactory receptor family 51 subfamily F member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-21 pseudogene|olfactory receptor, family 51, subfamily F, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4768979, end_position_on_the_genomic_accession=4769917, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376637, CurrentID=0, Status='live', Symbol='LOC105376637', Aliases='', description='uncharacterized LOC105376637', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=39480665, end_position_on_the_genomic_accession=39779540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376638, CurrentID=0, Status='live', Symbol='LOC105376638', Aliases='', description='uncharacterized LOC105376638', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=40826421, end_position_on_the_genomic_accession=40829545, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376639, CurrentID=0, Status='live', Symbol='LOC105376639', Aliases='', description='uncharacterized LOC105376639', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=41853693, end_position_on_the_genomic_accession=41875785, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376640, CurrentID=0, Status='live', Symbol='LOC105376640', Aliases='', description='uncharacterized LOC105376640', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=41521167, end_position_on_the_genomic_accession=41714477, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376641, CurrentID=0, Status='live', Symbol='LOC105376641', Aliases='', description='uncharacterized LOC105376641', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=42067015, end_position_on_the_genomic_accession=42080205, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376642, CurrentID=0, Status='live', Symbol='LOC105376642', Aliases='', description='uncharacterized LOC105376642', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=42253828, end_position_on_the_genomic_accession=42261931, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376643, CurrentID=0, Status='live', Symbol='LOC105376643', Aliases='', description='uncharacterized LOC105376643', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43329208, end_position_on_the_genomic_accession=43335334, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376644, CurrentID=0, Status='live', Symbol='LOC105376644', Aliases='', description='uncharacterized LOC105376644', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43874136, end_position_on_the_genomic_accession=43880741, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376645, CurrentID=0, Status='live', Symbol='LOC105376645', Aliases='', description='uncharacterized LOC105376645', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44510112, end_position_on_the_genomic_accession=44517170, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376646, CurrentID=0, Status='live', Symbol='LOC105376646', Aliases='', description='uncharacterized LOC105376646', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44518827, end_position_on_the_genomic_accession=44526066, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376647, CurrentID=0, Status='live', Symbol='LOC105376647', Aliases='', description='uncharacterized LOC105376647', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44689380, end_position_on_the_genomic_accession=44701306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376648, CurrentID=0, Status='live', Symbol='LOC105376648', Aliases='', description='uncharacterized LOC105376648', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44901003, end_position_on_the_genomic_accession=44903885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387977, CurrentID=0, Status='live', Symbol='NOVA1-AS1', Aliases='C14orf22', description='NOVA1 antisense RNA 1 (head to head)', other_designations='NOVA1 antisense RNA 1 (non-protein coding)', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387978, CurrentID=0, Status='live', Symbol='LINC01551', Aliases='C14orf23, c14_5148', description='long intergenic non-protein coding RNA 1551', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28772704, end_position_on_the_genomic_accession=28794794, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376651, CurrentID=0, Status='live', Symbol='LOC105376651', Aliases='', description='uncharacterized LOC105376651', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=44983932, end_position_on_the_genomic_accession=45008175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376652, CurrentID=0, Status='live', Symbol='LOC105376652', Aliases='', description='uncharacterized LOC105376652', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45106485, end_position_on_the_genomic_accession=45145307, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376653, CurrentID=0, Status='live', Symbol='LOC105376653', Aliases='', description='uncharacterized LOC105376653', other_designations='CTD-2560E9.5', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45253914, end_position_on_the_genomic_accession=45254882, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376654, CurrentID=0, Status='live', Symbol='LOC105376654', Aliases='', description='uncharacterized LOC105376654', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45531002, end_position_on_the_genomic_accession=45533527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376655, CurrentID=0, Status='live', Symbol='LOC105376655', Aliases='', description='uncharacterized LOC105376655', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45569343, end_position_on_the_genomic_accession=45578134, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376656, CurrentID=0, Status='live', Symbol='LOC105376656', Aliases='', description='uncharacterized LOC105376656', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45578340, end_position_on_the_genomic_accession=45582398, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376657, CurrentID=0, Status='live', Symbol='LOC105376657', Aliases='', description='uncharacterized LOC105376657', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45739160, end_position_on_the_genomic_accession=45741128, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376658, CurrentID=0, Status='live', Symbol='LOC105376658', Aliases='', description='uncharacterized LOC105376658', other_designations='CTD-2210P24.3', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45750823, end_position_on_the_genomic_accession=45751602, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376659, CurrentID=0, Status='live', Symbol='LOC105376659', Aliases='', description='uncharacterized LOC105376659', other_designations='CTD-2210P24.6', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45813055, end_position_on_the_genomic_accession=45825632, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376660, CurrentID=0, Status='live', Symbol='LOC105376660', Aliases='', description='uncharacterized LOC105376660', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=45898184, end_position_on_the_genomic_accession=45899921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929454, CurrentID=0, Status='live', Symbol='LOC101929454', Aliases='', description='uncharacterized LOC101929454', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15191646, end_position_on_the_genomic_accession=15266560, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  387990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=387990, CurrentID=0, Status='live', Symbol='TOMM20L', Aliases='UNQ9438', description='translocase of outer mitochondrial membrane 20 homolog (yeast)-like', other_designations='translocase of outer mitochondrial membrane 20 homolog type I', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=58395926, end_position_on_the_genomic_accession=58417080, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376663, CurrentID=0, Status='live', Symbol='LOC105376663', Aliases='', description='uncharacterized LOC105376663', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47153440, end_position_on_the_genomic_accession=47155136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376664, CurrentID=0, Status='live', Symbol='LOC105376664', Aliases='', description='uncharacterized LOC105376664', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=47001698, end_position_on_the_genomic_accession=47011880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376665, CurrentID=0, Status='live', Symbol='LOC105376665', Aliases='', description='uncharacterized LOC105376665', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49554362, end_position_on_the_genomic_accession=49558290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376666, CurrentID=0, Status='live', Symbol='LOC105376666', Aliases='', description='putative uncharacterized protein FLJ44672', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=50070331, end_position_on_the_genomic_accession=50079477, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376667, CurrentID=0, Status='live', Symbol='LOC105376667', Aliases='', description='uncharacterized LOC105376667', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=574016, end_position_on_the_genomic_accession=577669, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376668, CurrentID=0, Status='live', Symbol='LOC105376668', Aliases='', description='putative uncharacterized protein DDB_G0291608', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=1621284, end_position_on_the_genomic_accession=1629036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376669, CurrentID=0, Status='live', Symbol='LOC105376669', Aliases='', description='ret finger protein-like 4A', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=13478521, end_position_on_the_genomic_accession=13479503, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100133790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133790, CurrentID=0, Status='live', Symbol='MUC3', Aliases='', description='intestinal mucin-like', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100133791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133791, CurrentID=0, Status='live', Symbol='ERV18-1', Aliases='', description='endogenous retrovirus group 18 member 1', other_designations='endogenous retrovirus group 18, member 1', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376672, CurrentID=0, Status='live', Symbol='LOC105376672', Aliases='', description='uncharacterized LOC105376672', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3944540, end_position_on_the_genomic_accession=3949012, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  152048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152048, CurrentID=0, Status='live', Symbol='FLJ31715', Aliases='', description='uncharacterized FLJ31715', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376674, CurrentID=0, Status='live', Symbol='LOC105376674', Aliases='', description='uncharacterized LOC105376674', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4432146, end_position_on_the_genomic_accession=4443683, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376675 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376675, CurrentID=0, Status='live', Symbol='LOC105376675', Aliases='', description='uncharacterized LOC105376675', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10566990, end_position_on_the_genomic_accession=10570094, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376676, CurrentID=0, Status='live', Symbol='LOC105376676', Aliases='', description='uncharacterized LOC105376676', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11095762, end_position_on_the_genomic_accession=11124254, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  256933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256933, CurrentID=0, Status='live', Symbol='NPB', Aliases='L7, PPL7, PPNPB', description='neuropeptide B', other_designations='prepro-NPB|preproneuropeptide B|preproprotein L7', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81902196, end_position_on_the_genomic_accession=81902905, orientation='plus', exon_count=2, OMIM=607996, no_hdr0=''),
  105376678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376678, CurrentID=0, Status='live', Symbol='LOC105376678', Aliases='', description='ovostatin-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=9448794, end_position_on_the_genomic_accession=9503089, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  388007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388007, CurrentID=0, Status='live', Symbol='SERPINA13P', Aliases='SERPINA13, UNQ6121', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 13, pseudogene', other_designations='protease inhibitor|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 13|serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 13 (pseudogene)', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94640725, end_position_on_the_genomic_accession=94646994, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376680, CurrentID=0, Status='live', Symbol='LOC105376680', Aliases='', description='uncharacterized LOC105376680', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=4973391, end_position_on_the_genomic_accession=4983958, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376681, CurrentID=0, Status='live', Symbol='LOC105376681', Aliases='', description='uncharacterized LOC105376681', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5088251, end_position_on_the_genomic_accession=5098548, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376682 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376682, CurrentID=0, Status='live', Symbol='LOC105376682', Aliases='', description='uncharacterized LOC105376682', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5131856, end_position_on_the_genomic_accession=5135373, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388011, CurrentID=0, Status='live', Symbol='LINC01550', Aliases='C14orf64', description='long intergenic non-protein coding RNA 1550', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=97925610, end_position_on_the_genomic_accession=97978124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376684, CurrentID=0, Status='live', Symbol='LOC105376684', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110710644, end_position_on_the_genomic_accession=110714084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  414194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414194, CurrentID=0, Status='live', Symbol='CCNYL2', Aliases='C10orf21P, bA178A10.2, CCNYL2', description='cyclin Y-like 2, pseudogene', other_designations='cyclin-Y-like protein 2', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42408174, end_position_on_the_genomic_accession=42472240, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105376686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376686, CurrentID=0, Status='live', Symbol='LOC105376686', Aliases='', description='uncharacterized LOC105376686', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5490842, end_position_on_the_genomic_accession=5494629, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388015, CurrentID=0, Status='live', Symbol='RTL1', Aliases='MART1, Mar1, PEG11', description='retrotransposon-like 1', other_designations='mammalian retrotransposon derived protein 1|paternally expressed gene 11 protein|retrotransposon-derived protein PEG11', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=100880655, end_position_on_the_genomic_accession=100884847, orientation='minus', exon_count=1, OMIM=611896, no_hdr0=''),
  105376688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376688, CurrentID=0, Status='live', Symbol='LOC105376688', Aliases='', description='uncharacterized LOC105376688', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6385583, end_position_on_the_genomic_accession=6387249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376689, CurrentID=0, Status='live', Symbol='LOC105376689', Aliases='', description='uncharacterized LOC105376689', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7701686, end_position_on_the_genomic_accession=7704034, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376690, CurrentID=0, Status='live', Symbol='LOC105376690', Aliases='', description='uncharacterized LOC105376690', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7700704, end_position_on_the_genomic_accession=7701651, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  125875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125875, CurrentID=0, Status='live', Symbol='CLDND2', Aliases='', description='claudin domain containing 2', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=51366957, end_position_on_the_genomic_accession=51369241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376692, CurrentID=0, Status='live', Symbol='LOC105376692', Aliases='', description='uncharacterized LOC105376692', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=7236685, end_position_on_the_genomic_accession=7240058, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388021, CurrentID=0, Status='live', Symbol='TMEM179', Aliases='C14orf90A, TMEM179', description='transmembrane protein 179', other_designations='transmembrane protein 179A', map_location='14q32.33', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104590864, end_position_on_the_genomic_accession=104604787, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376694 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376694, CurrentID=0, Status='live', Symbol='LOC105376694', Aliases='', description='uncharacterized LOC105376694', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8003181, end_position_on_the_genomic_accession=8007029, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376695, CurrentID=0, Status='live', Symbol='LOC105376695', Aliases='', description='uncharacterized LOC105376695', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8240090, end_position_on_the_genomic_accession=8248643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376696, CurrentID=0, Status='live', Symbol='LOC105376696', Aliases='', description='uncharacterized LOC105376696', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=8905612, end_position_on_the_genomic_accession=8926858, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376697, CurrentID=0, Status='live', Symbol='LOC105376697', Aliases='', description='uncharacterized LOC105376697', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9241963, end_position_on_the_genomic_accession=9242942, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376698, CurrentID=0, Status='live', Symbol='LOC105376698', Aliases='', description='uncharacterized LOC105376698', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23338745, end_position_on_the_genomic_accession=23341887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376699, CurrentID=0, Status='live', Symbol='LOC105376699', Aliases='', description='uncharacterized LOC105376699', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24258648, end_position_on_the_genomic_accession=24260606, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  650172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650172, CurrentID=0, Status='live', Symbol='CCT4P1', Aliases='CCT4-2P', description='chaperonin containing TCP1, subunit 4 (delta) pseudogene 1', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140997744, end_position_on_the_genomic_accession=140999788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  256957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256957, CurrentID=0, Status='live', Symbol='HEATR9', Aliases='C17orf66', description='HEAT repeat containing 9', other_designations='HEAT repeat-containing protein 9', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=35854946, end_position_on_the_genomic_accession=35868891, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105376702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376702, CurrentID=0, Status='live', Symbol='LOC105376702', Aliases='', description='uncharacterized LOC105376702', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=27271527, end_position_on_the_genomic_accession=27285057, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376703, CurrentID=0, Status='live', Symbol='LOC105376703', Aliases='', description='uncharacterized LOC105376703', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30486293, end_position_on_the_genomic_accession=30489311, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376704, CurrentID=0, Status='live', Symbol='LOC105376704', Aliases='', description='uncharacterized LOC105376704', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30427814, end_position_on_the_genomic_accession=30471009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376705, CurrentID=0, Status='live', Symbol='LOC105376705', Aliases='', description='uncharacterized LOC105376705', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30418624, end_position_on_the_genomic_accession=30423814, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376706, CurrentID=0, Status='live', Symbol='LOC105376706', Aliases='', description='uncharacterized LOC105376706', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9500752, end_position_on_the_genomic_accession=9503503, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376707, CurrentID=0, Status='live', Symbol='LOC105376707', Aliases='', description='uncharacterized LOC105376707', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=31344872, end_position_on_the_genomic_accession=31348801, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376708, CurrentID=0, Status='live', Symbol='LOC105376708', Aliases='', description='uncharacterized LOC105376708', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32358729, end_position_on_the_genomic_accession=32368362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  125893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125893, CurrentID=0, Status='live', Symbol='ZNF816', Aliases='ZNF816A', description='zinc finger protein 816', other_designations='zinc finger protein 816A', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52949379, end_position_on_the_genomic_accession=52962911, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376710, CurrentID=0, Status='live', Symbol='LOC105376710', Aliases='', description='uncharacterized LOC105376710', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32462966, end_position_on_the_genomic_accession=32473443, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376711, CurrentID=0, Status='live', Symbol='LOC105376711', Aliases='', description='uncharacterized LOC105376711', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32459271, end_position_on_the_genomic_accession=32465194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376712, CurrentID=0, Status='live', Symbol='LOC105376712', Aliases='', description='uncharacterized LOC105376712', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9628447, end_position_on_the_genomic_accession=9629428, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376713, CurrentID=0, Status='live', Symbol='LOC105376713', Aliases='', description='uncharacterized LOC105376713', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40998995, end_position_on_the_genomic_accession=41011952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376714, CurrentID=0, Status='live', Symbol='LOC105376714', Aliases='', description='uncharacterized LOC105376714', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45448410, end_position_on_the_genomic_accession=45469925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376715, CurrentID=0, Status='live', Symbol='LOC105376715', Aliases='', description='uncharacterized LOC105376715', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55221283, end_position_on_the_genomic_accession=55223274, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376716, CurrentID=0, Status='live', Symbol='LOC105376716', Aliases='', description='uncharacterized LOC105376716', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=55311735, end_position_on_the_genomic_accession=55314050, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376717, CurrentID=0, Status='live', Symbol='LOC105376717', Aliases='', description='uncharacterized LOC105376717', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9850784, end_position_on_the_genomic_accession=9855490, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376718, CurrentID=0, Status='live', Symbol='LOC105376718', Aliases='', description='uncharacterized LOC105376718', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66858573, end_position_on_the_genomic_accession=66864416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376719, CurrentID=0, Status='live', Symbol='LOC105376719', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=9996423, end_position_on_the_genomic_accession=9998089, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376720, CurrentID=0, Status='live', Symbol='LOC105376720', Aliases='', description='uncharacterized LOC105376720', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=72609002, end_position_on_the_genomic_accession=72624335, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376721, CurrentID=0, Status='live', Symbol='LOC105376721', Aliases='', description='uncharacterized LOC105376721', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84436868, end_position_on_the_genomic_accession=84438089, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376722, CurrentID=0, Status='live', Symbol='LOC105376722', Aliases='', description='uncharacterized LOC105376722', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84389017, end_position_on_the_genomic_accession=84392007, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  256979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256979, CurrentID=0, Status='live', Symbol='SUN3', Aliases='SUNC1', description='Sad1 and UNC84 domain containing 3', other_designations='Sad1 and UNC84 domain containing 1|hypothetical protein MGC33329|sad1/unc-84 domain-containing protein 1', map_location='7p12.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=47987148, end_position_on_the_genomic_accession=48029127, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105376724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376724, CurrentID=0, Status='live', Symbol='LOC105376724', Aliases='', description='uncharacterized LOC105376724', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84215102, end_position_on_the_genomic_accession=84220891, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376725 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376725, CurrentID=0, Status='live', Symbol='LOC105376725', Aliases='', description='uncharacterized LOC105376725', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10358001, end_position_on_the_genomic_accession=10366735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376726, CurrentID=0, Status='live', Symbol='LOC105376726', Aliases='', description='uncharacterized LOC105376726', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=85273166, end_position_on_the_genomic_accession=85274158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  283129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283129, CurrentID=0, Status='live', Symbol='C11orf85', Aliases='MAJIN', description='chromosome 11 open reading frame 85', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=64938230, end_position_on_the_genomic_accession=64972112, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105376728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376728, CurrentID=0, Status='live', Symbol='LOC105376728', Aliases='', description='uncharacterized LOC105376728', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96460729, end_position_on_the_genomic_accession=96463149, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376729, CurrentID=0, Status='live', Symbol='LOC105376729', Aliases='', description='uncharacterized LOC105376729', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376730, CurrentID=0, Status='live', Symbol='LOC105376730', Aliases='', description='uncharacterized LOC105376730', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101871867, end_position_on_the_genomic_accession=101873791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  256987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=256987, CurrentID=0, Status='live', Symbol='SERINC5', Aliases='C5orf12, TPO1', description='serine incorporator 5', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80111227, end_position_on_the_genomic_accession=80256082, orientation='minus', exon_count=16, OMIM=614551, no_hdr0=''),
  105376732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376732, CurrentID=0, Status='live', Symbol='LOC105376732', Aliases='', description='uncharacterized LOC105376732', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=89087561, end_position_on_the_genomic_accession=89094089, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376733, CurrentID=0, Status='live', Symbol='LOC105376733', Aliases='', description='uncharacterized LOC105376733', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=10809018, end_position_on_the_genomic_accession=10815658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376734, CurrentID=0, Status='live', Symbol='LOC105376734', Aliases='', description='uncharacterized LOC105376734', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101755720, end_position_on_the_genomic_accession=101758532, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  125919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125919, CurrentID=0, Status='live', Symbol='ZNF543', Aliases='', description='zinc finger protein 543', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57320497, end_position_on_the_genomic_accession=57330776, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376736, CurrentID=0, Status='live', Symbol='LOC105376736', Aliases='', description='uncharacterized LOC105376736', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11099721, end_position_on_the_genomic_accession=11102100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376737, CurrentID=0, Status='live', Symbol='LOC105376737', Aliases='', description='uncharacterized LOC105376737', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11311498, end_position_on_the_genomic_accession=11316839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376738, CurrentID=0, Status='live', Symbol='LOC105376738', Aliases='', description='uncharacterized LOC105376738', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376739, CurrentID=0, Status='live', Symbol='LOC105376739', Aliases='', description='uncharacterized LOC105376739', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11357968, end_position_on_the_genomic_accession=11364461, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376740, CurrentID=0, Status='live', Symbol='LOC105376740', Aliases='', description='uncharacterized LOC105376740', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11371744, end_position_on_the_genomic_accession=11377509, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392358, CurrentID=0, Status='live', Symbol='RPS6P13', Aliases='RPS6_6_992', description='ribosomal protein S6 pseudogene 13', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86374935, end_position_on_the_genomic_accession=86375747, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376742, CurrentID=0, Status='live', Symbol='LOC105376742', Aliases='', description='uncharacterized LOC105376742', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=11908121, end_position_on_the_genomic_accession=11914327, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376743, CurrentID=0, Status='live', Symbol='LOC105376743', Aliases='', description='uncharacterized LOC105376743', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12471980, end_position_on_the_genomic_accession=12497047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  257000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257000, CurrentID=0, Status='live', Symbol='TINCR', Aliases='LINC00036, NCRNA00036, PLAC2, onco-lncRNA-16', description='tissue differentiation-inducing non-protein coding RNA', other_designations='long intergenic non-protein coding RNA 36|placenta-specific 2 (non-protein coding)|terminal differentiation-induced ncRNA', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5558167, end_position_on_the_genomic_accession=5568034, orientation='minus', exon_count=3, OMIM=615241, no_hdr0=''),
  105376745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376745, CurrentID=0, Status='live', Symbol='LOC105376745', Aliases='', description='uncharacterized LOC105376745', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12822724, end_position_on_the_genomic_accession=12823223, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376746, CurrentID=0, Status='live', Symbol='LOC105376746', Aliases='', description='formin-like protein 5', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=80887, end_position_on_the_genomic_accession=83044, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  125931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125931, CurrentID=0, Status='live', Symbol='CEACAM20', Aliases='UNQ9366', description='carcinoembryonic antigen-related cell adhesion molecule 20', other_designations='GPAD9366', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44504833, end_position_on_the_genomic_accession=44529945, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  388076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388076, CurrentID=0, Status='live', Symbol='RPS8P10', Aliases='RPS8_6_1418', description='ribosomal protein S8 pseudogene 10', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22152218, end_position_on_the_genomic_accession=22152918, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388077, CurrentID=0, Status='live', Symbol='IGHV1OR15-1', Aliases='IGHV1/OR15-1, IGHV1OR151', description='immunoglobulin heavy variable 1/OR15-1 (non-functional)', other_designations='immunoglobulin heavy variable 1/OR15-1 (pseudogene)|immunoglobulin heavy variable 1/OR15-1 pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22160288, end_position_on_the_genomic_accession=22160934, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376750, CurrentID=0, Status='live', Symbol='LOC105376750', Aliases='', description='uncharacterized LOC105376750', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18073874, end_position_on_the_genomic_accession=18076287, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376751, CurrentID=0, Status='live', Symbol='LOC105376751', Aliases='', description='uncharacterized LOC105376751', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18238290, end_position_on_the_genomic_accession=18256304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376752, CurrentID=0, Status='live', Symbol='LOC105376752', Aliases='', description='NPIP-like protein 1', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18358072, end_position_on_the_genomic_accession=18372027, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  654504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654504, CurrentID=0, Status='live', Symbol='KIAA0125P1', Aliases='FAM30B, HsT16028', description='KIAA0125 pseudogene 1', other_designations='family with sequence similarity 30, member B', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  650226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650226, CurrentID=0, Status='live', Symbol='LOC650226', Aliases='', description='ankyrin repeat domain 26 pseudogene', other_designations='', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56423704, end_position_on_the_genomic_accession=56448375, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376755, CurrentID=0, Status='live', Symbol='LOC105376755', Aliases='', description='uncharacterized LOC105376755', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376756, CurrentID=0, Status='live', Symbol='LOC105376756', Aliases='', description='uncharacterized LOC105376756', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13687118, end_position_on_the_genomic_accession=13691876, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376757, CurrentID=0, Status='live', Symbol='LOC105376757', Aliases='', description='uncharacterized LOC105376757', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14922090, end_position_on_the_genomic_accession=14923846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376758, CurrentID=0, Status='live', Symbol='LOC105376758', Aliases='', description='uncharacterized LOC105376758', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14816222, end_position_on_the_genomic_accession=14826229, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376759, CurrentID=0, Status='live', Symbol='LOC105376759', Aliases='', description='uncharacterized LOC105376759', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14801038, end_position_on_the_genomic_accession=14814936, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376760, CurrentID=0, Status='live', Symbol='LOC105376760', Aliases='', description='uncharacterized LOC105376760', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376761, CurrentID=0, Status='live', Symbol='LOC105376761', Aliases='', description='uncharacterized LOC105376761', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376762, CurrentID=0, Status='live', Symbol='LOC105376762', Aliases='', description='uncharacterized LOC105376762', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14388171, end_position_on_the_genomic_accession=14419973, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  257019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257019, CurrentID=0, Status='live', Symbol='FRMD3', Aliases='4.1O, EPB41L4O, EPB41LO, P410', description='FERM domain containing 3', other_designations='band 4.1-like protein 4|band 4.1-like protein 4O|ovary type protein 4.1|protein 4.1O', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=83242990, end_position_on_the_genomic_accession=83538433, orientation='minus', exon_count=21, OMIM=607619, no_hdr0=''),
  650236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650236, CurrentID=0, Status='live', Symbol='PRAMEF32P', Aliases='', description='PRAME family member 32, pseudogene', other_designations='PRAME family member 1 pseudogene', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13272907, end_position_on_the_genomic_accession=13275627, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376765, CurrentID=0, Status='live', Symbol='LOC105376765', Aliases='', description='uncharacterized LOC105376765', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15289697, end_position_on_the_genomic_accession=15305562, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  125950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125950, CurrentID=0, Status='live', Symbol='RAVER1', Aliases='', description='ribonucleoprotein, PTB-binding 1', other_designations='protein raver-1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10316212, end_position_on_the_genomic_accession=10333638, orientation='minus', exon_count=15, OMIM=609950, no_hdr0=''),
  105378877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378877, CurrentID=0, Status='live', Symbol='LOC105378877', Aliases='', description='uncharacterized LOC105378877', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103669003, end_position_on_the_genomic_accession=103681271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376768, CurrentID=0, Status='live', Symbol='LOC105376768', Aliases='', description='uncharacterized LOC105376768', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414208, CurrentID=0, Status='live', Symbol='ZNF32-AS2', Aliases='ZNF32OS2, b402L1.7', description='ZNF32 antisense RNA 2', other_designations='ZNF32 antisense RNA 2 (non-protein coding)|zinc finger protein 32 opposite strand 2', map_location='10q11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43645942, end_position_on_the_genomic_accession=43648019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376770, CurrentID=0, Status='live', Symbol='LOC105376770', Aliases='', description='uncharacterized LOC105376770', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55653124, end_position_on_the_genomic_accession=55655652, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376771, CurrentID=0, Status='live', Symbol='LOC105376771', Aliases='', description='uncharacterized LOC105376771', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57917458, end_position_on_the_genomic_accession=57921242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376772, CurrentID=0, Status='live', Symbol='LOC105376772', Aliases='', description='uncharacterized LOC105376772', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74406535, end_position_on_the_genomic_accession=74407239, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376773, CurrentID=0, Status='live', Symbol='LOC105376773', Aliases='', description='uncharacterized LOC105376773', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74388537, end_position_on_the_genomic_accession=74389055, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  125958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125958, CurrentID=0, Status='live', Symbol='OR7D4', Aliases='OR19-7, OR19-B, OR19BP, hg105, OR7D4', description='olfactory receptor family 7 subfamily D member 4', other_designations='odorant receptor family 7 subfamily D member 4 RT|olfactory receptor OR19-7|olfactory receptor, family 7, subfamily D, member 4 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9213850, end_position_on_the_genomic_accession=9214871, orientation='minus', exon_count=1, OMIM=611538, no_hdr0=''),
  204962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=204962, CurrentID=0, Status='live', Symbol='SLC44A5', Aliases='CTL5', description='solute carrier family 44 member 5', other_designations='solute carrier family 44, member 5', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75202131, end_position_on_the_genomic_accession=75724322, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  105376776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376776, CurrentID=0, Status='live', Symbol='LOC105376776', Aliases='', description='uncharacterized LOC105376776', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=78418620, end_position_on_the_genomic_accession=78437737, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376777, CurrentID=0, Status='live', Symbol='LOC105376777', Aliases='', description='uncharacterized LOC105376777', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=84291444, end_position_on_the_genomic_accession=84294838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  125962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125962, CurrentID=0, Status='live', Symbol='OR7G1', Aliases='OR19-15, OR19-8P, OR7G1', description='olfactory receptor family 7 subfamily G member 1', other_designations='olfactory receptor 19-15|olfactory receptor OR19-8|olfactory receptor, family 7, subfamily G, member 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9114828, end_position_on_the_genomic_accession=9115763, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  125963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125963, CurrentID=0, Status='live', Symbol='OR1M1', Aliases='OR19-5, OR19-6', description='olfactory receptor family 1 subfamily M member 1', other_designations='olfactory receptor 19-6|olfactory receptor OR19-5|olfactory receptor, family 1, subfamily M, member 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9093245, end_position_on_the_genomic_accession=9094186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105376780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376780, CurrentID=0, Status='live', Symbol='LOC105376780', Aliases='', description='uncharacterized LOC105376780', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89232741, end_position_on_the_genomic_accession=89234453, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  125965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125965, CurrentID=0, Status='live', Symbol='COX6B2', Aliases='COXVIB2, CT59', description='cytochrome c oxidase subunit VIb polypeptide 2 (testis)', other_designations='COX VIb-2|cancer/testis antigen 59|cytochrome c oxidase subunit VIb, testes-specific', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55349702, end_position_on_the_genomic_accession=55354814, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376782, CurrentID=0, Status='live', Symbol='LOC105376782', Aliases='', description='uncharacterized LOC105376782', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90156049, end_position_on_the_genomic_accession=90157203, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376783, CurrentID=0, Status='live', Symbol='LOC105376783', Aliases='', description='uncharacterized LOC105376783', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90133834, end_position_on_the_genomic_accession=90137822, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388112, CurrentID=0, Status='live', Symbol='NANOGP8', Aliases='NANOG, NANOGP1, PN8', description='Nanog homeobox pseudogene 8', other_designations='NANOG pseudogene-1', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35083226, end_position_on_the_genomic_accession=35085308, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376785, CurrentID=0, Status='live', Symbol='LOC105376785', Aliases='', description='uncharacterized LOC105376785', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90102673, end_position_on_the_genomic_accession=90106120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376786, CurrentID=0, Status='live', Symbol='LOC105376786', Aliases='', description='uncharacterized LOC105376786', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90101179, end_position_on_the_genomic_accession=90101838, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388115, CurrentID=0, Status='live', Symbol='C15orf52', Aliases='', description='chromosome 15 open reading frame 52', other_designations='', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=40331452, end_position_on_the_genomic_accession=40340967, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  125972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125972, CurrentID=0, Status='live', Symbol='CALR3', Aliases='CMH19, CRT2, CT93', description='calreticulin 3', other_designations='calreticulin-2|calsperin|cancer/testis antigen 93', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16479057, end_position_on_the_genomic_accession=16496192, orientation='minus', exon_count=9, OMIM=611414, no_hdr0=''),
  105376789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376789, CurrentID=0, Status='live', Symbol='LOC105376789', Aliases='', description='uncharacterized LOC105376789', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81800413, end_position_on_the_genomic_accession=81807271, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376790, CurrentID=0, Status='live', Symbol='LOC105376790', Aliases='', description='uncharacterized LOC105376790', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82380823, end_position_on_the_genomic_accession=82383611, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376791, CurrentID=0, Status='live', Symbol='LOC105376791', Aliases='', description='uncharacterized LOC105376791', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=82701905, end_position_on_the_genomic_accession=82705329, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376792, CurrentID=0, Status='live', Symbol='LOC105376792', Aliases='', description='uncharacterized LOC105376792', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388121, CurrentID=0, Status='live', Symbol='TNFAIP8L3', Aliases='TIPE3', description='TNF alpha induced protein 8 like 3', other_designations='TNF alpha-induced protein 8-like protein 3|TNFAIP8-like protein 3|tumor necrosis factor, alpha induced protein 8 like 3|tumor necrosis factor, alpha-induced protein 8-like 3', map_location='15q21.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=51056596, end_position_on_the_genomic_accession=51105276, orientation='minus', exon_count=6, OMIM=616438, no_hdr0=''),
  388122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388122, CurrentID=0, Status='live', Symbol='RPSAP55', Aliases='RPSA_28_1433', description='ribosomal protein SA pseudogene 55', other_designations='', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=52885519, end_position_on_the_genomic_accession=52886542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376795, CurrentID=0, Status='live', Symbol='LOC105376795', Aliases='', description='uncharacterized LOC105376795', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=503027, end_position_on_the_genomic_accession=511347, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376796, CurrentID=0, Status='live', Symbol='LOC105376796', Aliases='', description='uncharacterized LOC105376796', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16516224, end_position_on_the_genomic_accession=16521801, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  125981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125981, CurrentID=0, Status='live', Symbol='ACER1', Aliases='ALKCDase1, ASAH3', description='alkaline ceramidase 1', other_designations='CTB-180A7.3|N-acylsphingosine amidohydrolase (alkaline ceramidase) 3|N-acylsphingosine amidohydrolase 3|acylsphingosine deacylase 3|alkCDase 1|alkaline CDase 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6306142, end_position_on_the_genomic_accession=6353444, orientation='minus', exon_count=7, OMIM=613491, no_hdr0=''),
  105376798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376798, CurrentID=0, Status='live', Symbol='LOC105376798', Aliases='', description='uncharacterized LOC105376798', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5094894, end_position_on_the_genomic_accession=5111791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376799, CurrentID=0, Status='live', Symbol='LOC105376799', Aliases='', description='uncharacterized LOC105376799', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8128569, end_position_on_the_genomic_accession=8139209, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376800, CurrentID=0, Status='live', Symbol='LOC105376800', Aliases='', description='uncharacterized LOC105376800', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8304001, end_position_on_the_genomic_accession=8305065, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376801, CurrentID=0, Status='live', Symbol='LOC105376801', Aliases='', description='uncharacterized LOC105376801', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8295474, end_position_on_the_genomic_accession=8303446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376802, CurrentID=0, Status='live', Symbol='LOC105376802', Aliases='', description='uncharacterized LOC105376802', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=19118607, end_position_on_the_genomic_accession=19125845, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376803, CurrentID=0, Status='live', Symbol='LOC105376803', Aliases='', description='uncharacterized LOC105376803', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  125988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125988, CurrentID=0, Status='live', Symbol='C19orf70', Aliases='MIC13, P117, QIL1', description='chromosome 19 open reading frame 70', other_designations='protein QIL1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5678419, end_position_on_the_genomic_accession=5680900, orientation='minus', exon_count=6, OMIM=616658, no_hdr0=''),
  105376805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376805, CurrentID=0, Status='live', Symbol='LOC105376805', Aliases='', description='uncharacterized LOC105376805', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16888538, end_position_on_the_genomic_accession=16889686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  257062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257062, CurrentID=0, Status='live', Symbol='CATSPERD', Aliases='TMEM146', description='catsper channel auxiliary subunit delta', other_designations='catSper-delta|catSperdelta|transmembrane protein 146', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5720401, end_position_on_the_genomic_accession=5778734, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  388135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388135, CurrentID=0, Status='live', Symbol='C15orf59', Aliases='', description='chromosome 15 open reading frame 59', other_designations='', map_location='15q24.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=73735431, end_position_on_the_genomic_accession=73753187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376808, CurrentID=0, Status='live', Symbol='LOC105376808', Aliases='', description='uncharacterized LOC105376808', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=17135704, end_position_on_the_genomic_accession=17138955, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376809 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376809, CurrentID=0, Status='live', Symbol='LOC105376809', Aliases='', description='uncharacterized LOC105376809', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18000422, end_position_on_the_genomic_accession=18010510, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376810, CurrentID=0, Status='live', Symbol='LOC105376810', Aliases='', description='uncharacterized LOC105376810', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376811, CurrentID=0, Status='live', Symbol='LOC105376811', Aliases='', description='laminin subunit beta-2-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  257068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257068, CurrentID=0, Status='live', Symbol='PLCXD2', Aliases='', description='phosphatidylinositol-specific phospholipase C, X domain containing 2', other_designations='', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=111674676, end_position_on_the_genomic_accession=111846447, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  125997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=125997, CurrentID=0, Status='live', Symbol='MBD3L2', Aliases='', description='methyl-CpG binding domain protein 3-like 2', other_designations='CTB-25J19.1|MBD3-like protein 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7049340, end_position_on_the_genomic_accession=7051735, orientation='plus', exon_count=2, OMIM=607964, no_hdr0=''),
  105376814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376814, CurrentID=0, Status='live', Symbol='LOC105376814', Aliases='', description='uncharacterized LOC105376814', other_designations='RP11-174G17', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18159986, end_position_on_the_genomic_accession=18176677, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376815, CurrentID=0, Status='live', Symbol='LOC105376815', Aliases='', description='uncharacterized LOC105376815', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18971113, end_position_on_the_genomic_accession=19050967, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376816, CurrentID=0, Status='live', Symbol='LOC105376816', Aliases='', description='uncharacterized LOC105376816', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19052480, end_position_on_the_genomic_accession=19055855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376817, CurrentID=0, Status='live', Symbol='LOC105376817', Aliases='', description='uncharacterized LOC105376817', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19512540, end_position_on_the_genomic_accession=19529924, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376818, CurrentID=0, Status='live', Symbol='LOC105376818', Aliases='', description='uncharacterized LOC105376818', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19509890, end_position_on_the_genomic_accession=19512483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  126003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126003, CurrentID=0, Status='live', Symbol='TRAPPC5', Aliases='TRS31', description='trafficking protein particle complex 5', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7680821, end_position_on_the_genomic_accession=7682862, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376820, CurrentID=0, Status='live', Symbol='LOC105376820', Aliases='', description='uncharacterized LOC105376820', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19408709, end_position_on_the_genomic_accession=19416641, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100133941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100133941, CurrentID=0, Status='live', Symbol='CD24', Aliases='CD24A', description='CD24 molecule', other_designations='CD24 antigen (small cell lung carcinoma cluster 4 antigen)', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106969831, end_position_on_the_genomic_accession=106976855, orientation='minus', exon_count=6, OMIM=600074, no_hdr0=''),
  126006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126006, CurrentID=0, Status='live', Symbol='PCP2', Aliases='GPSM4', description='Purkinje cell protein 2', other_designations='CTD-3214H19.6|Purkinje cell-specific protein 2|Purkinje cell-specific protein L7', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7631614, end_position_on_the_genomic_accession=7636967, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376823, CurrentID=0, Status='live', Symbol='LOC105376823', Aliases='', description='uncharacterized LOC105376823', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19860166, end_position_on_the_genomic_accession=19889388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388152, CurrentID=0, Status='live', Symbol='GOLGA2P7', Aliases='', description='golgin A2 pseudogene 7', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84198848, end_position_on_the_genomic_accession=84230168, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376825, CurrentID=0, Status='live', Symbol='LOC105376825', Aliases='', description='uncharacterized LOC105376825', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20198996, end_position_on_the_genomic_accession=20202827, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376826, CurrentID=0, Status='live', Symbol='LOC105376826', Aliases='', description='uncharacterized LOC105376826', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20210720, end_position_on_the_genomic_accession=20213372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376827, CurrentID=0, Status='live', Symbol='LOC105376827', Aliases='', description='uncharacterized LOC105376827', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20243101, end_position_on_the_genomic_accession=20244622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376828, CurrentID=0, Status='live', Symbol='LOC105376828', Aliases='', description='uncharacterized LOC105376828', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20271641, end_position_on_the_genomic_accession=20274919, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376829, CurrentID=0, Status='live', Symbol='LOC105376829', Aliases='', description='uncharacterized LOC105376829', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20919271, end_position_on_the_genomic_accession=20936232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  126014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126014, CurrentID=0, Status='live', Symbol='OSCAR', Aliases='PIGR3, PIgR-3', description='osteoclast associated, immunoglobulin-like receptor', other_designations='osteoclast associated receptor OSCAR-S1|osteoclast associated receptor OSCAR-S2|poly-Ig receptor 3|polymeric immunoglobulin receptor 3', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=54094668, end_position_on_the_genomic_accession=54102686, orientation='minus', exon_count=6, OMIM=606862, no_hdr0=''),
  100271549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271549, CurrentID=0, Status='live', Symbol='RPS12P20', Aliases='RPS12_10_1153', description='ribosomal protein S12 pseudogene 20', other_designations='', map_location='11q13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=72708186, end_position_on_the_genomic_accession=72708565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376832, CurrentID=0, Status='live', Symbol='LOC105376832', Aliases='', description='uncharacterized LOC105376832', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21412679, end_position_on_the_genomic_accession=21416543, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  126017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126017, CurrentID=0, Status='live', Symbol='ZNF813', Aliases='', description='zinc finger protein 813', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53467735, end_position_on_the_genomic_accession=53494292, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376834, CurrentID=0, Status='live', Symbol='LOC105376834', Aliases='', description='uncharacterized LOC105376834', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21585689, end_position_on_the_genomic_accession=21591187, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376835, CurrentID=0, Status='live', Symbol='LOC105376835', Aliases='', description='uncharacterized LOC105376835', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21732236, end_position_on_the_genomic_accession=21780475, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376836, CurrentID=0, Status='live', Symbol='LOC105376836', Aliases='', description='uncharacterized LOC105376836', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36657893, end_position_on_the_genomic_accession=36715155, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388165, CurrentID=0, Status='live', Symbol='UBE2Q2P1', Aliases='UBE2QP1', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 1', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 1|ubiquitin-conjugating enzyme E2Q family pseudogene 1', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=84527196, end_position_on_the_genomic_accession=84570795, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105376838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376838, CurrentID=0, Status='live', Symbol='LOC105376838', Aliases='', description='uncharacterized LOC105376838', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46169095, end_position_on_the_genomic_accession=46169962, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376839, CurrentID=0, Status='live', Symbol='LRRC37A', Aliases='', description='leucine rich repeat containing 37A', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45585786, end_position_on_the_genomic_accession=45597843, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376840, CurrentID=0, Status='live', Symbol='LOC105376840', Aliases='', description='uncharacterized LOC105376840', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=45448019, end_position_on_the_genomic_accession=45450611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376841, CurrentID=0, Status='live', Symbol='LOC105376841', Aliases='', description='uncharacterized LOC105376841', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22031028, end_position_on_the_genomic_accession=22052869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376842, CurrentID=0, Status='live', Symbol='LOC105376842', Aliases='', description='uncharacterized LOC105376842', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=58019031, end_position_on_the_genomic_accession=58026739, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376843, CurrentID=0, Status='live', Symbol='SNHG25', Aliases='', description='small nucleolar RNA host gene 25', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64144950, end_position_on_the_genomic_accession=64146810, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376844, CurrentID=0, Status='live', Symbol='LOC105376844', Aliases='', description='uncharacterized LOC105376844', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64835657, end_position_on_the_genomic_accession=64910624, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  257101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257101, CurrentID=0, Status='live', Symbol='ZNF683', Aliases='Hobit', description='zinc finger protein 683', other_designations='homolog of Blimp-1 in T cells|hypothetical protein MGC33414', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26361632, end_position_on_the_genomic_accession=26374533, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376846, CurrentID=0, Status='live', Symbol='LOC105376846', Aliases='', description='uncharacterized LOC105376846', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=71891824, end_position_on_the_genomic_accession=71907746, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376847, CurrentID=0, Status='live', Symbol='LOC105376847', Aliases='', description='uncharacterized LOC105376847', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=76700190, end_position_on_the_genomic_accession=76707304, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376848, CurrentID=0, Status='live', Symbol='LOC105376848', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=57129175, end_position_on_the_genomic_accession=57132329, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376849, CurrentID=0, Status='live', Symbol='LOC105376849', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60119031, end_position_on_the_genomic_accession=60120552, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  257106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257106, CurrentID=0, Status='live', Symbol='ARHGAP30', Aliases='', description='Rho GTPase activating protein 30', other_designations='rho-type GTPase-activating protein 30', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161046942, end_position_on_the_genomic_accession=161069971, orientation='minus', exon_count=13, OMIM=614264, no_hdr0=''),
  105376851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376851, CurrentID=0, Status='live', Symbol='LOC105376851', Aliases='', description='uncharacterized LOC105376851', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=11635709, end_position_on_the_genomic_accession=11641066, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376852, CurrentID=0, Status='live', Symbol='LOC105376852', Aliases='', description='uncharacterized LOC105376852', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=13942343, end_position_on_the_genomic_accession=13947553, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  152078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152078, CurrentID=0, Status='live', Symbol='PQLC2L', Aliases='C3orf55', description='PQ loop repeat containing 2-like', other_designations='PQ-loop repeat-containing protein 2-like', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157543344, end_position_on_the_genomic_accession=157677750, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  388182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388182, CurrentID=0, Status='live', Symbol='SPATA41', Aliases='HSD47', description='spermatogenesis associated 41 (non-protein coding)', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=100344457, end_position_on_the_genomic_accession=100350718, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376855, CurrentID=0, Status='live', Symbol='LOC105376855', Aliases='', description='uncharacterized LOC105376855', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=12431366, end_position_on_the_genomic_accession=12436206, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376856, CurrentID=0, Status='live', Symbol='LOC105376856', Aliases='', description='uncharacterized LOC105376856', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22416331, end_position_on_the_genomic_accession=22418938, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376857, CurrentID=0, Status='live', Symbol='LOC105376857', Aliases='', description='uncharacterized LOC105376857', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=22883330, end_position_on_the_genomic_accession=22889663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376858, CurrentID=0, Status='live', Symbol='LOC105376858', Aliases='', description='uncharacterized LOC105376858', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23424835, end_position_on_the_genomic_accession=23429733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376859, CurrentID=0, Status='live', Symbol='LOC105376859', Aliases='', description='uncharacterized LOC105376859', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23548517, end_position_on_the_genomic_accession=23557350, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376860, CurrentID=0, Status='live', Symbol='LOC105376860', Aliases='', description='uncharacterized LOC105376860', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23790624, end_position_on_the_genomic_accession=23793698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376861, CurrentID=0, Status='live', Symbol='LOC105376861', Aliases='', description='uncharacterized LOC105376861', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23825545, end_position_on_the_genomic_accession=23838629, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376862, CurrentID=0, Status='live', Symbol='LOC105376862', Aliases='', description='uncharacterized LOC105376862', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23949007, end_position_on_the_genomic_accession=23959141, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376863, CurrentID=0, Status='live', Symbol='LOC105376863', Aliases='', description='uncharacterized LOC105376863', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24010505, end_position_on_the_genomic_accession=24028914, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376864, CurrentID=0, Status='live', Symbol='LOC105376864', Aliases='', description='uncharacterized LOC105376864', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24098669, end_position_on_the_genomic_accession=24104537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  414224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414224, CurrentID=0, Status='live', Symbol='AGAP12P', Aliases='CTGLF12P', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 12, pseudogene', other_designations='centaurin, gamma-like family, member 12 pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48010132, end_position_on_the_genomic_accession=48031640, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376866, CurrentID=0, Status='live', Symbol='LOC105376866', Aliases='', description='uncharacterized LOC105376866', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=21704848, end_position_on_the_genomic_accession=21706061, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376867, CurrentID=0, Status='live', Symbol='LOC105376867', Aliases='', description='uncharacterized LOC105376867', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24112496, end_position_on_the_genomic_accession=24129813, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126052, CurrentID=0, Status='live', Symbol='TGIF1P1', Aliases='', description='TGFB-induced factor homeobox 1 pseudogene 1', other_designations='TG-interacting factor (TALE family homeobox) processed pseudogene|TGFB-induced factor (TALE family homeobox) pseudogene', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46136288, end_position_on_the_genomic_accession=46137597, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376869, CurrentID=0, Status='live', Symbol='LOC105376869', Aliases='', description='uncharacterized LOC105376869', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=50255761, end_position_on_the_genomic_accession=50261055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376870, CurrentID=0, Status='live', Symbol='LOC105376870', Aliases='', description='uncharacterized LOC105376870', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=57763791, end_position_on_the_genomic_accession=57794459, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388199, CurrentID=0, Status='live', Symbol='PRR25', Aliases='gs64', description='proline rich 25', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=805443, end_position_on_the_genomic_accession=813861, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376872, CurrentID=0, Status='live', Symbol='LOC105376872', Aliases='', description='uncharacterized LOC105376872', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=70470174, end_position_on_the_genomic_accession=70531730, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376873, CurrentID=0, Status='live', Symbol='LOC105376873', Aliases='', description='uncharacterized LOC105376873', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=71777919, end_position_on_the_genomic_accession=71799893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376874, CurrentID=0, Status='live', Symbol='LOC105376874', Aliases='', description='uncharacterized LOC105376874', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=75165615, end_position_on_the_genomic_accession=75168258, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376875, CurrentID=0, Status='live', Symbol='LOC105376875', Aliases='', description='uncharacterized LOC105376875', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=79680361, end_position_on_the_genomic_accession=79688319, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376876, CurrentID=0, Status='live', Symbol='LOC105376876', Aliases='', description='uncharacterized LOC105376876', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24857343, end_position_on_the_genomic_accession=24865338, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376877, CurrentID=0, Status='live', Symbol='LOC105376877', Aliases='', description='collagen alpha-2(IV) chain-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=78652970, end_position_on_the_genomic_accession=78658276, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376878, CurrentID=0, Status='live', Symbol='LOC105376878', Aliases='', description='uncharacterized LOC105376878', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24932512, end_position_on_the_genomic_accession=24964506, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376879, CurrentID=0, Status='live', Symbol='LOC105376879', Aliases='', description='uncharacterized LOC105376879', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25041172, end_position_on_the_genomic_accession=25047346, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376880, CurrentID=0, Status='live', Symbol='LOC105376880', Aliases='', description='uncharacterized LOC105376880', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25131181, end_position_on_the_genomic_accession=25132913, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376881, CurrentID=0, Status='live', Symbol='LOC105376881', Aliases='', description='uncharacterized LOC105376881', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25290741, end_position_on_the_genomic_accession=25338211, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376882, CurrentID=0, Status='live', Symbol='LOC105376882', Aliases='', description='uncharacterized LOC105376882', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25269462, end_position_on_the_genomic_accession=25274764, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376883, CurrentID=0, Status='live', Symbol='LOC105376883', Aliases='', description='uncharacterized LOC105376883', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25753603, end_position_on_the_genomic_accession=25756063, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126068, CurrentID=0, Status='live', Symbol='ZNF441', Aliases='', description='zinc finger protein 441', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11767000, end_position_on_the_genomic_accession=11784078, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  126069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126069, CurrentID=0, Status='live', Symbol='ZNF491', Aliases='', description='zinc finger protein 491', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11798477, end_position_on_the_genomic_accession=11809622, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  126070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126070, CurrentID=0, Status='live', Symbol='ZNF440', Aliases='', description='zinc finger protein 440', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11814284, end_position_on_the_genomic_accession=11835617, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376887, CurrentID=0, Status='live', Symbol='LOC105376887', Aliases='', description='uncharacterized LOC105376887', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26165313, end_position_on_the_genomic_accession=26169802, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  257144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257144, CurrentID=0, Status='live', Symbol='GCSAM', Aliases='GCAT2, GCET2, HGAL', description='germinal center-associated, signaling and motility', other_designations='germinal center B-cell-expressed transcript 2 protein|germinal center expressed transcript 2|germinal center-associated lymphoma protein|human germinal center-associated lymphoma', map_location='3q13.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=112120841, end_position_on_the_genomic_accession=112136705, orientation='minus', exon_count=6, OMIM=607792, no_hdr0=''),
  105376889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376889, CurrentID=0, Status='live', Symbol='LOC105376889', Aliases='', description='uncharacterized LOC105376889', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27207820, end_position_on_the_genomic_accession=27208690, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  126074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126074, CurrentID=0, Status='live', Symbol='SWSAP1', Aliases='C19orf39, SWS1AP1, ZSWIM7AP1', description='SWIM-type zinc finger 7 associated protein 1', other_designations='SWIM-type zinc finger 7-associated protein 1|SWS1-associated protein 1|ZSWIM7-associated protein 1|zinc finger, SWIM-type containing 7 associated protein 1', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11374707, end_position_on_the_genomic_accession=11376951, orientation='plus', exon_count=2, OMIM=614536, no_hdr0=''),
  126075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126075, CurrentID=0, Status='live', Symbol='CCDC159', Aliases='', description='coiled-coil domain containing 159', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11346505, end_position_on_the_genomic_accession=11354944, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105376892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376892, CurrentID=0, Status='live', Symbol='LOC105376892', Aliases='', description='uncharacterized LOC105376892', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27522506, end_position_on_the_genomic_accession=27530506, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376893 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376893, CurrentID=0, Status='live', Symbol='LOC105376893', Aliases='', description='uncharacterized LOC105376893', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27849816, end_position_on_the_genomic_accession=27850118, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376894, CurrentID=0, Status='live', Symbol='LOC105376894', Aliases='', description='uncharacterized LOC105376894', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27828451, end_position_on_the_genomic_accession=27831528, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100134015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134015, CurrentID=0, Status='live', Symbol='UBOX5-AS1', Aliases='', description='UBOX5 antisense RNA 1', other_designations='UBOX5 antisense RNA 1 (non-protein coding)', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=3106911, end_position_on_the_genomic_accession=3150867, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  650368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650368, CurrentID=0, Status='live', Symbol='TSSC2', Aliases='', description='tumor suppressing subtransferable candidate 2 pseudogene', other_designations='asparagine-linked glycosylation 1 homolog (yeast, beta-1,4-mannosyltransferase) (ALG1) pseudogene|asparagine-linked glycosylation 1-like pseudogene|beta-1,4-mannosyltransferase pseudogene|tumor-supressing STF cDNA 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=3380961, end_position_on_the_genomic_accession=3409148, orientation='plus', exon_count=11, OMIM=608999, no_hdr0=''),
  105376897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376897, CurrentID=0, Status='live', Symbol='LOC105376897', Aliases='', description='uncharacterized LOC105376897', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28927222, end_position_on_the_genomic_accession=28942792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376898, CurrentID=0, Status='live', Symbol='LOC105376898', Aliases='', description='uncharacterized LOC105376898', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28870482, end_position_on_the_genomic_accession=28886843, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376899, CurrentID=0, Status='live', Symbol='LOC105376899', Aliases='', description='uncharacterized LOC105376899', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32072219, end_position_on_the_genomic_accession=32074122, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388228, CurrentID=0, Status='live', Symbol='SBK1', Aliases='SBK', description='SH3 domain binding kinase 1', other_designations='SH3-binding domain kinase 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28258703, end_position_on_the_genomic_accession=28323849, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376901, CurrentID=0, Status='live', Symbol='LOC105376901', Aliases='', description='uncharacterized LOC105376901', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46144648, end_position_on_the_genomic_accession=46158277, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376902, CurrentID=0, Status='live', Symbol='LOC105376902', Aliases='', description='uncharacterized LOC105376902', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49056589, end_position_on_the_genomic_accession=49060965, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376903, CurrentID=0, Status='live', Symbol='LOC105376903', Aliases='', description='uncharacterized LOC105376903', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52835761, end_position_on_the_genomic_accession=52837792, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257160, CurrentID=0, Status='live', Symbol='RNF214', Aliases='', description='ring finger protein 214', other_designations='', map_location='11q23.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=117232625, end_position_on_the_genomic_accession=117285688, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105376905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376905, CurrentID=0, Status='live', Symbol='LOC105376905', Aliases='', description='uncharacterized LOC105376905', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55605753, end_position_on_the_genomic_accession=55610280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376906, CurrentID=0, Status='live', Symbol='LOC105376906', Aliases='', description='endogenous retrovirus group K member 19 Pol protein', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=27520104, end_position_on_the_genomic_accession=27642524, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376907, CurrentID=0, Status='live', Symbol='LOC105376907', Aliases='', description='uncharacterized LOC105376907', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36269531, end_position_on_the_genomic_accession=36278351, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376908, CurrentID=0, Status='live', Symbol='LOC105376908', Aliases='', description='uncharacterized LOC105376908', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41135638, end_position_on_the_genomic_accession=41193052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376909, CurrentID=0, Status='live', Symbol='LOC105376909', Aliases='', description='uncharacterized LOC105376909', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43901660, end_position_on_the_genomic_accession=43935219, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376910, CurrentID=0, Status='live', Symbol='LOC105376910', Aliases='', description='uncharacterized LOC105376910', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26045895, end_position_on_the_genomic_accession=26049057, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376911, CurrentID=0, Status='live', Symbol='LOC105376911', Aliases='', description='putative uncharacterized protein encoded by LINC00269', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28677907, end_position_on_the_genomic_accession=28681528, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376912, CurrentID=0, Status='live', Symbol='LOC105376912', Aliases='', description='uncharacterized LOC105376912', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=71141, end_position_on_the_genomic_accession=72759, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  257169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257169, CurrentID=0, Status='live', Symbol='C9orf43', Aliases='', description='chromosome 9 open reading frame 43', other_designations='', map_location='9q32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=113410644, end_position_on_the_genomic_accession=113429684, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105376914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376914, CurrentID=0, Status='live', Symbol='LOC105376914', Aliases='', description='uncharacterized LOC105376914', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9590469, end_position_on_the_genomic_accession=9596320, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376915 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376915, CurrentID=0, Status='live', Symbol='LOC105376915', Aliases='', description='uncharacterized LOC105376915', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19822944, end_position_on_the_genomic_accession=19867098, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376916, CurrentID=0, Status='live', Symbol='LOC105376916', Aliases='', description='uncharacterized LOC105376916', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20858206, end_position_on_the_genomic_accession=20863696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376917, CurrentID=0, Status='live', Symbol='LOC105376917', Aliases='', description='uncharacterized LOC105376917', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22455376, end_position_on_the_genomic_accession=22532491, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105376918 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376918, CurrentID=0, Status='live', Symbol='LOC105376918', Aliases='', description='uncharacterized LOC105376918', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23988718, end_position_on_the_genomic_accession=24001346, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376919, CurrentID=0, Status='live', Symbol='LOC105376919', Aliases='', description='nucleolar protein 58-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23466457, end_position_on_the_genomic_accession=23471128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100134040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134040, CurrentID=0, Status='live', Symbol='LOC100134040', Aliases='', description='uncharacterized LOC100134040', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150031334, end_position_on_the_genomic_accession=150074885, orientation='plus', exon_count=24, OMIM=615171, no_hdr0=''),
  257177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257177, CurrentID=0, Status='live', Symbol='CFAP126', Aliases='C1orf192, Flattop, Fltp', description='cilia and flagella associated protein 126', other_designations='UPF0740 protein C1orf192|cilia- and flagella-associated protein 126', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161364731, end_position_on_the_genomic_accession=161367883, orientation='minus', exon_count=5, OMIM=616119, no_hdr0=''),
  105376922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376922, CurrentID=0, Status='live', Symbol='LOC105376922', Aliases='', description='uncharacterized LOC105376922', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=483639, end_position_on_the_genomic_accession=493542, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376923, CurrentID=0, Status='live', Symbol='LOC105376923', Aliases='', description='uncharacterized LOC105376923', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1182207, end_position_on_the_genomic_accession=1194419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376924, CurrentID=0, Status='live', Symbol='LOC105376924', Aliases='', description='uncharacterized LOC105376924', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=1165712, end_position_on_the_genomic_accession=1177237, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376925, CurrentID=0, Status='live', Symbol='LOC105376925', Aliases='', description='uncharacterized LOC105376925', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3008944, end_position_on_the_genomic_accession=3012076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388254, CurrentID=0, Status='live', Symbol='HERC2P5', Aliases='D16F37S6', description='hect domain and RLD 2 pseudogene 5', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32741244, end_position_on_the_genomic_accession=32775468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388255, CurrentID=0, Status='live', Symbol='IGHV3OR16-8', Aliases='IGHV3/OR16-8, IGHV3OR168', description='immunoglobulin heavy variable 3/OR16-8 (non-functional)', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32997726, end_position_on_the_genomic_accession=33009646, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105376928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376928, CurrentID=0, Status='live', Symbol='LOC105376928', Aliases='', description='uncharacterized LOC105376928', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3249883, end_position_on_the_genomic_accession=3382650, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105376929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376929, CurrentID=0, Status='live', Symbol='LOC105376929', Aliases='', description='uncharacterized LOC105376929', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4269295, end_position_on_the_genomic_accession=4303242, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376930, CurrentID=0, Status='live', Symbol='LOC105376930', Aliases='', description='uncharacterized LOC105376930', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4041378, end_position_on_the_genomic_accession=4045953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376931, CurrentID=0, Status='live', Symbol='LOC105376931', Aliases='', description='uncharacterized LOC105376931', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=3925112, end_position_on_the_genomic_accession=3963144, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376932, CurrentID=0, Status='live', Symbol='LOC105376932', Aliases='', description='uncharacterized LOC105376932', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4705181, end_position_on_the_genomic_accession=4710356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376933, CurrentID=0, Status='live', Symbol='LOC105376933', Aliases='', description='uncharacterized LOC105376933', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4581920, end_position_on_the_genomic_accession=4584241, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376934, CurrentID=0, Status='live', Symbol='LOC105376934', Aliases='', description='uncharacterized LOC105376934', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4994674, end_position_on_the_genomic_accession=5027038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  126119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126119, CurrentID=0, Status='live', Symbol='JOSD2', Aliases='SBBI54', description='Josephin domain containing 2', other_designations='josephin domain-containing protein 2', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50505997, end_position_on_the_genomic_accession=50511355, orientation='minus', exon_count=5, OMIM=615324, no_hdr0=''),
  105376936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376936, CurrentID=0, Status='live', Symbol='LOC105376936', Aliases='', description='uncharacterized LOC105376936', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5437577, end_position_on_the_genomic_accession=5441605, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376937, CurrentID=0, Status='live', Symbol='LOC105376937', Aliases='', description='uncharacterized LOC105376937', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5673675, end_position_on_the_genomic_accession=5684180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  257194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257194, CurrentID=0, Status='live', Symbol='NEGR1', Aliases='DMML2433, IGLON4, KILON, Ntra', description='neuronal growth regulator 1', other_designations='IgLON family member 4|a kindred of IgLON|neurotractin', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71402942, end_position_on_the_genomic_accession=72282847, orientation='minus', exon_count=9, OMIM=613173, no_hdr0=''),
  126123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126123, CurrentID=0, Status='live', Symbol='IZUMO2', Aliases='C19orf41, PLAL6978, PRO21961, SCRL', description='IZUMO family member 2', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50152548, end_position_on_the_genomic_accession=50163281, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105376940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376940, CurrentID=0, Status='live', Symbol='LOC105376940', Aliases='', description='uncharacterized LOC105376940', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5740696, end_position_on_the_genomic_accession=5755710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376941, CurrentID=0, Status='live', Symbol='LOC105376941', Aliases='', description='uncharacterized LOC105376941', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=5815040, end_position_on_the_genomic_accession=5842609, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376942, CurrentID=0, Status='live', Symbol='LOC105376942', Aliases='', description='uncharacterized LOC105376942', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6067096, end_position_on_the_genomic_accession=6366159, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105376943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376943, CurrentID=0, Status='live', Symbol='LOC105376943', Aliases='', description='uncharacterized LOC105376943', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=6780425, end_position_on_the_genomic_accession=6783195, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388272, CurrentID=0, Status='live', Symbol='C16orf87', Aliases='', description='chromosome 16 open reading frame 87', other_designations='', map_location='16q11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=46802047, end_position_on_the_genomic_accession=46831162, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  126129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126129, CurrentID=0, Status='live', Symbol='CPT1C', Aliases='CATL1, CPT1-B, CPT1P, CPTI-B, CPTIC, SPG73', description='carnitine palmitoyltransferase 1C', other_designations='carnitine O-palmitoyltransferase I, brain isoform|carnitine palmitoyltransferase I related C', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49690878, end_position_on_the_genomic_accession=49713731, orientation='plus', exon_count=25, OMIM=608846, no_hdr0=''),
  257202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257202, CurrentID=0, Status='live', Symbol='GPX6', Aliases='GPX5p, GPXP3, GPx-6, GSHPx-6, dJ1186N24, dJ1186N24.1', description='glutathione peroxidase 6', other_designations='glutathione peroxidase 6 (olfactory)|glutathione peroxidase pseudogene 3', map_location='6p22.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28503296, end_position_on_the_genomic_accession=28515793, orientation='minus', exon_count=5, OMIM=607913, no_hdr0=''),
  257203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257203, CurrentID=0, Status='live', Symbol='DSCR9', Aliases='NCRNA00038', description='Down syndrome critical region 9 (non-protein coding)', other_designations='Down syndrome critical region gene 9 (non-protein coding)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37208503, end_position_on_the_genomic_accession=37221736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388276, CurrentID=0, Status='live', Symbol='C16orf97', Aliases='', description='chromosome 16 open reading frame 97', other_designations='', map_location='16q12.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52026352, end_position_on_the_genomic_accession=52078517, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  126133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126133, CurrentID=0, Status='live', Symbol='ALDH16A1', Aliases='', description='aldehyde dehydrogenase 16 family member A1', other_designations='aldehyde dehydrogenase 16 family, member A1', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49453216, end_position_on_the_genomic_accession=49471050, orientation='plus', exon_count=19, OMIM=613358, no_hdr0=''),
  105376950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376950, CurrentID=0, Status='live', Symbol='LOC105376950', Aliases='', description='uncharacterized LOC105376950', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=10913950, end_position_on_the_genomic_accession=10928067, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376951, CurrentID=0, Status='live', Symbol='LOC105376951', Aliases='', description='uncharacterized LOC105376951', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11141372, end_position_on_the_genomic_accession=11154389, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376952, CurrentID=0, Status='live', Symbol='LOC105376952', Aliases='', description='uncharacterized LOC105376952', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=11565753, end_position_on_the_genomic_accession=11576352, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376953, CurrentID=0, Status='live', Symbol='LOC105376953', Aliases='', description='uncharacterized LOC105376953', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12792304, end_position_on_the_genomic_accession=12797304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388282, CurrentID=0, Status='live', Symbol='LOC388282', Aliases='', description='uncharacterized LOC388282', other_designations='CTD-2600O9.1', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=57810066, end_position_on_the_genomic_accession=57816927, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376955, CurrentID=0, Status='live', Symbol='LOC105376955', Aliases='', description='nuclear pore membrane glycoprotein 210-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=12882360, end_position_on_the_genomic_accession=12885099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388284, CurrentID=0, Status='live', Symbol='C16orf86', Aliases='', description='chromosome 16 open reading frame 86', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67666806, end_position_on_the_genomic_accession=67668758, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376957, CurrentID=0, Status='live', Symbol='LOC105376957', Aliases='', description='uncharacterized LOC105376957', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=13895530, end_position_on_the_genomic_accession=13926911, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376958, CurrentID=0, Status='live', Symbol='LOC105376958', Aliases='', description='uncharacterized LOC105376958', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14224679, end_position_on_the_genomic_accession=14230753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376959, CurrentID=0, Status='live', Symbol='LOC105376959', Aliases='', description='uncharacterized LOC105376959', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14390294, end_position_on_the_genomic_accession=14399181, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105376960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376960, CurrentID=0, Status='live', Symbol='LOC105376960', Aliases='', description='uncharacterized LOC105376960', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14601052, end_position_on_the_genomic_accession=14609305, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388289, CurrentID=0, Status='live', Symbol='C16orf47', Aliases='', description='chromosome 16 open reading frame 47', other_designations='', map_location='16q22.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=73126638, end_position_on_the_genomic_accession=73144447, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  257218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257218, CurrentID=0, Status='live', Symbol='SHPRH', Aliases='bA545I5.2', description='SNF2 histone linker PHD RING helicase, E3 ubiquitin protein ligase', other_designations='2610103K11Rik|SNF2, histone-linker, PHD and RING finger domain-containing helicase', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145862721, end_position_on_the_genomic_accession=145964285, orientation='minus', exon_count=36, OMIM=608048, no_hdr0=''),
  126147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126147, CurrentID=0, Status='live', Symbol='NTN5', Aliases='', description='netrin 5', other_designations='netrin-1-like protein', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48661407, end_position_on_the_genomic_accession=48673081, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105376964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376964, CurrentID=0, Status='live', Symbol='LOC105376964', Aliases='', description='uncharacterized LOC105376964', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15499648, end_position_on_the_genomic_accession=15504180, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376965, CurrentID=0, Status='live', Symbol='LOC105376965', Aliases='', description='uncharacterized LOC105376965', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=15444318, end_position_on_the_genomic_accession=15450061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376966, CurrentID=0, Status='live', Symbol='LOC105376966', Aliases='', description='uncharacterized LOC105376966', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16128073, end_position_on_the_genomic_accession=16132757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100291105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100291105, CurrentID=0, Status='live', Symbol='LOC100291105', Aliases='', description='uncharacterized LOC100291105', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=57383726, end_position_on_the_genomic_accession=57392492, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376968, CurrentID=0, Status='live', Symbol='LOC105376968', Aliases='', description='uncharacterized LOC105376968', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16001380, end_position_on_the_genomic_accession=16049684, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376969, CurrentID=0, Status='live', Symbol='LOC105376969', Aliases='', description='uncharacterized LOC105376969', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16472791, end_position_on_the_genomic_accession=16480777, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376970, CurrentID=0, Status='live', Symbol='LOC105376970', Aliases='', description='uncharacterized LOC105376970', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16605577, end_position_on_the_genomic_accession=16631681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376971, CurrentID=0, Status='live', Symbol='LOC105376971', Aliases='', description='uncharacterized LOC105376971', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105376972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376972, CurrentID=0, Status='live', Symbol='LOC105376972', Aliases='', description='uncharacterized LOC105376972', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=16881791, end_position_on_the_genomic_accession=16884164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376973, CurrentID=0, Status='live', Symbol='LOC105376973', Aliases='', description='uncharacterized LOC105376973', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37499707, end_position_on_the_genomic_accession=37517257, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105376974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376974, CurrentID=0, Status='live', Symbol='LOC105376974', Aliases='', description='uncharacterized LOC105376974', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=17140982, end_position_on_the_genomic_accession=17157719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104266957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266957, CurrentID=0, Status='live', Symbol='KBTBD11-OT1', Aliases='', description='KBTBD11 overlapping transcript 1', other_designations='KBTBD11 overlapping transcript 1 (non-protein coding)', map_location='8p23.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1971397, end_position_on_the_genomic_accession=1976444, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376976, CurrentID=0, Status='live', Symbol='LOC105376976', Aliases='', description='uncharacterized LOC105376976', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18745977, end_position_on_the_genomic_accession=18843164, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376977, CurrentID=0, Status='live', Symbol='LOC105376977', Aliases='', description='uncharacterized LOC105376977', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46204443, end_position_on_the_genomic_accession=46212987, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105376978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376978, CurrentID=0, Status='live', Symbol='LOC105376978', Aliases='', description='uncharacterized LOC105376978', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=18949263, end_position_on_the_genomic_accession=18991062, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  414243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414243, CurrentID=0, Status='live', Symbol='LINC00595', Aliases='C10orf101', description='long intergenic non-protein coding RNA 595', other_designations='', map_location='10q22.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78267328, end_position_on_the_genomic_accession=78280213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  257236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257236, CurrentID=0, Status='live', Symbol='CCDC96', Aliases='', description='coiled-coil domain containing 96', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7040849, end_position_on_the_genomic_accession=7043001, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104266958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104266958, CurrentID=0, Status='live', Symbol='LINC01151', Aliases='', description='long intergenic non-protein coding RNA 1151', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122670385, end_position_on_the_genomic_accession=122694106, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105376982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376982, CurrentID=0, Status='live', Symbol='LOC105376982', Aliases='', description='uncharacterized LOC105376982', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19390760, end_position_on_the_genomic_accession=19484728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105376983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376983, CurrentID=0, Status='live', Symbol='LOC105376983', Aliases='', description='uncharacterized LOC105376983', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=19342617, end_position_on_the_genomic_accession=19348351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  257240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257240, CurrentID=0, Status='live', Symbol='KLHL34', Aliases='', description='kelch like family member 34', other_designations='RP11-450P7.3|kelch-like 34|kelch-like family member 34', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=21655349, end_position_on_the_genomic_accession=21658387, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105376985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376985, CurrentID=0, Status='live', Symbol='LOC105376985', Aliases='', description='uncharacterized LOC105376985', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52660165, end_position_on_the_genomic_accession=52667932, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376986, CurrentID=0, Status='live', Symbol='LOC105376986', Aliases='', description='uncharacterized LOC105376986', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52210623, end_position_on_the_genomic_accession=52357555, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376987, CurrentID=0, Status='live', Symbol='LOC105376987', Aliases='', description='uncharacterized LOC105376987', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21041230, end_position_on_the_genomic_accession=21150115, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105376988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376988, CurrentID=0, Status='live', Symbol='LOC105376988', Aliases='', description='uncharacterized LOC105376988', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21169369, end_position_on_the_genomic_accession=21221327, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105376989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376989, CurrentID=0, Status='live', Symbol='LOC105376989', Aliases='', description='uncharacterized LOC105376989', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56101774, end_position_on_the_genomic_accession=56103602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376990, CurrentID=0, Status='live', Symbol='LOC105376990', Aliases='', description='uncharacterized LOC105376990', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21645291, end_position_on_the_genomic_accession=21657732, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  414245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414245, CurrentID=0, Status='live', Symbol='DNAJC9-AS1', Aliases='C10orf103, bA537A6.3', description='DNAJC9 antisense RNA 1', other_designations='DNAJC9 antisense RNA 1 (non-protein coding)', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73251622, end_position_on_the_genomic_accession=73254343, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105376992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376992, CurrentID=0, Status='live', Symbol='LOC105376992', Aliases='', description='uncharacterized LOC105376992', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58003850, end_position_on_the_genomic_accession=58004580, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105376993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376993, CurrentID=0, Status='live', Symbol='LOC105376993', Aliases='', description='uncharacterized LOC105376993', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=23643669, end_position_on_the_genomic_accession=23654875, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376994 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376994, CurrentID=0, Status='live', Symbol='LOC105376994', Aliases='', description='uncharacterized LOC105376994', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24446191, end_position_on_the_genomic_accession=24473379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388323, CurrentID=0, Status='live', Symbol='GLTPD2', Aliases='', description='glycolipid transfer protein domain containing 2', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4788422, end_position_on_the_genomic_accession=4790592, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388324, CurrentID=0, Status='live', Symbol='INCA1', Aliases='HSD45', description='inhibitor of CDK, cyclin A1 interacting protein 1', other_designations='', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=4988130, end_position_on_the_genomic_accession=4997922, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  388325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388325, CurrentID=0, Status='live', Symbol='SCIMP', Aliases='C17orf87, UNQ5783', description='SLP adaptor and CSK interacting membrane protein', other_designations='DTFT5783|SLP65/SLP76, Csk-interacting membrane protein|transmembrane protein C17orf87', map_location='17p13.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=5208920, end_position_on_the_genomic_accession=5234860, orientation='minus', exon_count=6, OMIM=614406, no_hdr0=''),
  105376998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376998, CurrentID=0, Status='live', Symbol='LOC105376998', Aliases='', description='uncharacterized LOC105376998', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25226343, end_position_on_the_genomic_accession=25255436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388327, CurrentID=0, Status='live', Symbol='C17orf100', Aliases='', description='chromosome 17 open reading frame 100', other_designations='hCG1985469', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=6651739, end_position_on_the_genomic_accession=6653298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377000, CurrentID=0, Status='live', Symbol='LOC105377000', Aliases='', description='uncharacterized LOC105377000', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=25831302, end_position_on_the_genomic_accession=25850056, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377001, CurrentID=0, Status='live', Symbol='LOC105377001', Aliases='', description='uncharacterized LOC105377001', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26064874, end_position_on_the_genomic_accession=26080722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  650474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650474, CurrentID=0, Status='live', Symbol='TPSP2', Aliases='', description='tryptase pseudogene 2', other_designations='mastin|tryptase beta pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1286343, end_position_on_the_genomic_accession=1289073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377003, CurrentID=0, Status='live', Symbol='LOC105377003', Aliases='', description='uncharacterized LOC105377003', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=26619649, end_position_on_the_genomic_accession=26622655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377004, CurrentID=0, Status='live', Symbol='LOC105377004', Aliases='', description='uncharacterized LOC105377004', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27216575, end_position_on_the_genomic_accession=27220883, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388333, CurrentID=0, Status='live', Symbol='SPDYE4', Aliases='', description='speedy/RINGO cell cycle regulator family member E4', other_designations='putative Speedy protein E4|speedy homolog E4', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=8746868, end_position_on_the_genomic_accession=8759412, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377006, CurrentID=0, Status='live', Symbol='LOC105377006', Aliases='', description='uncharacterized LOC105377006', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=27712337, end_position_on_the_genomic_accession=27714006, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388335, CurrentID=0, Status='live', Symbol='TMEM220', Aliases='', description='transmembrane protein 220', other_designations='', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=10713322, end_position_on_the_genomic_accession=10730329, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  388336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388336, CurrentID=0, Status='live', Symbol='SHISA6', Aliases='', description='shisa family member 6', other_designations='shisa homolog 6', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=11241417, end_position_on_the_genomic_accession=11564063, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105377009 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377009, CurrentID=0, Status='live', Symbol='LOC105377009', Aliases='', description='uncharacterized LOC105377009', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=29037908, end_position_on_the_genomic_accession=29057585, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377010, CurrentID=0, Status='live', Symbol='LOC105377010', Aliases='', description='uncharacterized LOC105377010', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=28590476, end_position_on_the_genomic_accession=28616673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388339, CurrentID=0, Status='live', Symbol='RPS18P12', Aliases='RPS18_4_1516', description='ribosomal protein S18 pseudogene 12', other_designations='', map_location='17p12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=14705038, end_position_on_the_genomic_accession=14705576, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377012, CurrentID=0, Status='live', Symbol='LOC105377012', Aliases='', description='uncharacterized LOC105377012', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30180494, end_position_on_the_genomic_accession=30184848, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388341, CurrentID=0, Status='live', Symbol='LRRC75A', Aliases='C17orf76, FAM211A', description='leucine rich repeat containing 75A', other_designations='family with sequence similarity 211, member A|leucine-rich repeat-containing protein C17orf76|leucine-rich repeat-containing protein FAM211A', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16442005, end_position_on_the_genomic_accession=16492191, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377014, CurrentID=0, Status='live', Symbol='LOC105377014', Aliases='', description='uncharacterized LOC105377014', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30511721, end_position_on_the_genomic_accession=30523547, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377015, CurrentID=0, Status='live', Symbol='LOC105377015', Aliases='', description='uncharacterized LOC105377015', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30603886, end_position_on_the_genomic_accession=30606485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388344, CurrentID=0, Status='live', Symbol='RPL13P12', Aliases='RPL13_4_1521', description='ribosomal protein L13 pseudogene 12', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=17383355, end_position_on_the_genomic_accession=17384063, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377017, CurrentID=0, Status='live', Symbol='LOC105377017', Aliases='', description='uncharacterized LOC105377017', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=30783663, end_position_on_the_genomic_accession=30796774, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377018, CurrentID=0, Status='live', Symbol='LOC105377018', Aliases='', description='uncharacterized LOC105377018', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31201432, end_position_on_the_genomic_accession=31202631, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377019, CurrentID=0, Status='live', Symbol='LOC105377019', Aliases='', description='uncharacterized LOC105377019', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31462571, end_position_on_the_genomic_accession=31469799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126204, CurrentID=0, Status='live', Symbol='NLRP13', Aliases='CLR19.7, NALP13, NOD14, PAN13', description='NLR family, pyrin domain containing 13', other_designations='NACHT, LRR and PYD containing protein 13|NACHT, leucine rich repeat and PYD containing 13|nucleotide-binding oligomerization domain protein 14|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 13', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55891692, end_position_on_the_genomic_accession=55932336, orientation='minus', exon_count=12, OMIM=609660, no_hdr0=''),
  126205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126205, CurrentID=0, Status='live', Symbol='NLRP8', Aliases='CLR19.2, NALP8, NOD16, PAN4', description='NLR family, pyrin domain containing 8', other_designations='NACHT, LRR and PYD containing protein 8|NACHT, leucine rich repeat and PYD containing 8|PYRIN and NACHT-containing protein 4|nucleotide-binding oligomerization domain protein 16|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 8', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55947832, end_position_on_the_genomic_accession=55988629, orientation='plus', exon_count=10, OMIM=609659, no_hdr0=''),
  126206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126206, CurrentID=0, Status='live', Symbol='NLRP5', Aliases='CLR19.8, MATER, NALP5, PAN11, PYPAF8', description='NLR family, pyrin domain containing 5', other_designations='NACHT, leucine rich repeat and PYD containing 5|mater protein homolog|maternal antigen that embryos require|nucleotide-binding oligomerization domain, leucine rich repeat and pyrin domain containing 5', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55986783, end_position_on_the_genomic_accession=56061813, orientation='plus', exon_count=16, OMIM=609658, no_hdr0=''),
  105377023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377023, CurrentID=0, Status='live', Symbol='LOC105377023', Aliases='', description='uncharacterized LOC105377023', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=33783824, end_position_on_the_genomic_accession=33798491, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  126208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126208, CurrentID=0, Status='live', Symbol='ZNF787', Aliases='TIP20', description='zinc finger protein 787', other_designations='TTF-I-interacting peptide 20|transcription termination factor I interacting peptide 20', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56087363, end_position_on_the_genomic_accession=56121373, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377025, CurrentID=0, Status='live', Symbol='LOC105377025', Aliases='', description='uncharacterized LOC105377025', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34174974, end_position_on_the_genomic_accession=34181702, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377026, CurrentID=0, Status='live', Symbol='LOC105377026', Aliases='', description='uncharacterized LOC105377026', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=34540174, end_position_on_the_genomic_accession=34543756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377027, CurrentID=0, Status='live', Symbol='LOC105377027', Aliases='', description='uncharacterized LOC105377027', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35509998, end_position_on_the_genomic_accession=35529153, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377028, CurrentID=0, Status='live', Symbol='LOC105377028', Aliases='', description='uncharacterized LOC105377028', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=35814632, end_position_on_the_genomic_accession=35833875, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377029, CurrentID=0, Status='live', Symbol='LOC105377029', Aliases='', description='uncharacterized LOC105377029', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36768995, end_position_on_the_genomic_accession=36772585, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377030, CurrentID=0, Status='live', Symbol='LOC105377030', Aliases='', description='uncharacterized LOC105377030', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36775892, end_position_on_the_genomic_accession=36784280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377031, CurrentID=0, Status='live', Symbol='LOC105377031', Aliases='', description='uncharacterized LOC105377031', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=36819283, end_position_on_the_genomic_accession=36822892, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377032, CurrentID=0, Status='live', Symbol='LOC105377032', Aliases='', description='uncharacterized LOC105377032', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=37379121, end_position_on_the_genomic_accession=37383264, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377033, CurrentID=0, Status='live', Symbol='LOC105377033', Aliases='', description='uncharacterized LOC105377033', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38103490, end_position_on_the_genomic_accession=38122350, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377034, CurrentID=0, Status='live', Symbol='LOC105377034', Aliases='', description='uncharacterized LOC105377034', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38813624, end_position_on_the_genomic_accession=38816345, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388363, CurrentID=0, Status='live', Symbol='RPS7P1', Aliases='RPS7_4_1531', description='ribosomal protein S7 pseudogene 1', other_designations='', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=28467778, end_position_on_the_genomic_accession=28468454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388364, CurrentID=0, Status='live', Symbol='TMIGD1', Aliases='TMIGD, UNQ9372', description='transmembrane and immunoglobulin domain containing 1', other_designations='AWKS9372', map_location='17q11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=30316329, end_position_on_the_genomic_accession=30334076, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377037, CurrentID=0, Status='live', Symbol='LOC105377037', Aliases='', description='uncharacterized LOC105377037', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39374722, end_position_on_the_genomic_accession=39381013, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377038, CurrentID=0, Status='live', Symbol='LOC105377038', Aliases='', description='uncharacterized LOC105377038', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39456699, end_position_on_the_genomic_accession=39459303, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377039, CurrentID=0, Status='live', Symbol='LOC105377039', Aliases='', description='uncharacterized LOC105377039', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39778654, end_position_on_the_genomic_accession=39802904, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105377040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377040, CurrentID=0, Status='live', Symbol='LOC105377040', Aliases='', description='uncharacterized LOC105377040', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40062087, end_position_on_the_genomic_accession=40083774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377041, CurrentID=0, Status='live', Symbol='LOC105377041', Aliases='', description='uncharacterized LOC105377041', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40627615, end_position_on_the_genomic_accession=40645433, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377042, CurrentID=0, Status='live', Symbol='LOC105377042', Aliases='', description='uncharacterized LOC105377042', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40871132, end_position_on_the_genomic_accession=40875601, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377043, CurrentID=0, Status='live', Symbol='LOC105377043', Aliases='', description='uncharacterized LOC105377043', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40719862, end_position_on_the_genomic_accession=40911068, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  388372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388372, CurrentID=0, Status='live', Symbol='CCL4L1', Aliases='AT744.2, CCL4L, LAG-1, LAG1, SCYA4L, SCYA4L1', description='chemokine (C-C motif) ligand 4-like 1', other_designations='CC chemokine ligand 4L1|chemokine (C-C motif) ligand 4-like 1, telomeric|lymphocyte activation gene 1|macrophage inflammatory protein-1b2|small inducible cytokine A4-like', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=603782, no_hdr0=''),
  414254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414254, CurrentID=0, Status='live', Symbol='RPSAP6', Aliases='LAMR1P6, RPSA_23_1065', description='ribosomal protein SA pseudogene 6', other_designations='laminin receptor 1 pseudogene 6', map_location='10q22.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=74371453, end_position_on_the_genomic_accession=74372486, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377046, CurrentID=0, Status='live', Symbol='LOC105377046', Aliases='', description='uncharacterized LOC105377046', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=41261330, end_position_on_the_genomic_accession=41267816, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126231, CurrentID=0, Status='live', Symbol='ZNF573', Aliases='', description='zinc finger protein 573', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37735131, end_position_on_the_genomic_accession=37779636, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105377048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377048, CurrentID=0, Status='live', Symbol='LOC105377048', Aliases='', description='uncharacterized LOC105377048', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42247925, end_position_on_the_genomic_accession=42253494, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377049, CurrentID=0, Status='live', Symbol='LOC105377049', Aliases='', description='uncharacterized LOC105377049', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42669453, end_position_on_the_genomic_accession=42683801, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377050, CurrentID=0, Status='live', Symbol='LOC105377050', Aliases='', description='uncharacterized LOC105377050', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  126235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126235, CurrentID=0, Status='live', Symbol='RPS4XP21', Aliases='RPS4P21, RPS4X_10_1649', description='ribosomal protein S4X pseudogene 21', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34092476, end_position_on_the_genomic_accession=34093366, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377052, CurrentID=0, Status='live', Symbol='LOC105377052', Aliases='', description='uncharacterized LOC105377052', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43260357, end_position_on_the_genomic_accession=43262575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388381, CurrentID=0, Status='live', Symbol='C17orf98', Aliases='', description='chromosome 17 open reading frame 98', other_designations='', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=38825392, end_position_on_the_genomic_accession=38841500, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377054, CurrentID=0, Status='live', Symbol='LOC105377054', Aliases='', description='uncharacterized LOC105377054', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=43754341, end_position_on_the_genomic_accession=43757591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377055, CurrentID=0, Status='live', Symbol='LOC105377055', Aliases='', description='uncharacterized LOC105377055', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44326553, end_position_on_the_genomic_accession=44338015, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377056 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377056, CurrentID=0, Status='live', Symbol='LOC105377056', Aliases='', description='uncharacterized LOC105377056', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=44715798, end_position_on_the_genomic_accession=44729553, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  257313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257313, CurrentID=0, Status='live', Symbol='UTS2B', Aliases='U2B, URP, UTS2D', description='urotensin 2B', other_designations='U-IIB|UIIB|prepro-URP|urotensin 2 domain containing|urotensin II-related peptide|urotensin IIB|urotensin-2 domain-containing protein', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191267155, end_position_on_the_genomic_accession=191330536, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  126242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126242, CurrentID=0, Status='live', Symbol='AKR1B1P7', Aliases='', description='aldo-keto reductase family 1, member B1 pseudogene 7', other_designations='aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33360126, end_position_on_the_genomic_accession=33361465, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388387, CurrentID=0, Status='live', Symbol='LINC00671', Aliases='', description='long intergenic non-protein coding RNA 671', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=42874674, end_position_on_the_genomic_accession=42898734, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377060, CurrentID=0, Status='live', Symbol='LOC105377060', Aliases='', description='uncharacterized LOC105377060', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=22744670, end_position_on_the_genomic_accession=22746973, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388389, CurrentID=0, Status='live', Symbol='CCDC103', Aliases='CILD17, PR46b, SMH', description='coiled-coil domain containing 103', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44899712, end_position_on_the_genomic_accession=44903679, orientation='plus', exon_count=4, OMIM=614677, no_hdr0=''),
  105377062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377062, CurrentID=0, Status='live', Symbol='LOC105377062', Aliases='', description='uncharacterized LOC105377062', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45327833, end_position_on_the_genomic_accession=45346520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377063, CurrentID=0, Status='live', Symbol='LOC105377063', Aliases='', description='uncharacterized LOC105377063', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=45995943, end_position_on_the_genomic_accession=46001283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126248, CurrentID=0, Status='live', Symbol='WDR88', Aliases='PQWD', description='WD repeat domain 88', other_designations='PQQ repeat and WD repeat domain containing|PQQ repeat and WD repeat-containing protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33132029, end_position_on_the_genomic_accession=33175799, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105377065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377065, CurrentID=0, Status='live', Symbol='LOC105377065', Aliases='', description='uncharacterized LOC105377065', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46067693, end_position_on_the_genomic_accession=46084076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388394, CurrentID=0, Status='live', Symbol='RPRML', Aliases='', description='reprimo-like', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=46978156, end_position_on_the_genomic_accession=46979248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377067, CurrentID=0, Status='live', Symbol='LOC105377067', Aliases='', description='uncharacterized LOC105377067', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46130894, end_position_on_the_genomic_accession=46190381, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377068, CurrentID=0, Status='live', Symbol='LOC105377068', Aliases='', description='uncharacterized LOC105377068', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46414779, end_position_on_the_genomic_accession=46424092, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377069, CurrentID=0, Status='live', Symbol='LOC105377069', Aliases='', description='uncharacterized LOC105377069', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46471550, end_position_on_the_genomic_accession=46476219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377070, CurrentID=0, Status='live', Symbol='LOC105377070', Aliases='', description='uncharacterized LOC105377070', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46504977, end_position_on_the_genomic_accession=46514933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377071, CurrentID=0, Status='live', Symbol='LOC105377071', Aliases='', description='uncharacterized LOC105377071', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46646420, end_position_on_the_genomic_accession=46665048, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377072, CurrentID=0, Status='live', Symbol='LOC105377072', Aliases='', description='uncharacterized LOC105377072', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46812083, end_position_on_the_genomic_accession=46813142, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388401, CurrentID=0, Status='live', Symbol='RPL7P48', Aliases='RPL7_19_1558', description='ribosomal protein L7 pseudogene 48', other_designations='', map_location='17q21.33', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51502422, end_position_on_the_genomic_accession=51503253, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377074, CurrentID=0, Status='live', Symbol='LOC105377074', Aliases='', description='uncharacterized LOC105377074', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=47532995, end_position_on_the_genomic_accession=47537671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126259, CurrentID=0, Status='live', Symbol='TMIGD2', Aliases='IGPR-1, IGPR1', description='transmembrane and immunoglobulin domain containing 2', other_designations='CD28 homolog|immunoglobulin and proline-rich receptor 1|immunoglobulin-containing and proline-rich receptor 1|transmembrane and immunoglobulin domain-containing protein 2 variant 2|transmembrane and immunoglobulin domain-containing protein 2 variant 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4292227, end_position_on_the_genomic_accession=4302582, orientation='minus', exon_count=5, OMIM=614715, no_hdr0=''),
  105377076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377076, CurrentID=0, Status='live', Symbol='LOC105377076', Aliases='', description='uncharacterized LOC105377076', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48275182, end_position_on_the_genomic_accession=48287631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377077, CurrentID=0, Status='live', Symbol='LOC105377077', Aliases='', description='uncharacterized LOC105377077', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48430149, end_position_on_the_genomic_accession=48432207, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377078, CurrentID=0, Status='live', Symbol='LOC105377078', Aliases='', description='uncharacterized LOC105377078', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48729740, end_position_on_the_genomic_accession=48740328, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388407, CurrentID=0, Status='live', Symbol='C17orf82', Aliases='', description='chromosome 17 open reading frame 82', other_designations='', map_location='17q23.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=61411751, end_position_on_the_genomic_accession=61413280, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377080, CurrentID=0, Status='live', Symbol='LOC105377080', Aliases='', description='uncharacterized LOC105377080', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48979910, end_position_on_the_genomic_accession=48982934, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377081, CurrentID=0, Status='live', Symbol='LOC105377081', Aliases='', description='uncharacterized LOC105377081', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49553623, end_position_on_the_genomic_accession=49554072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377082, CurrentID=0, Status='live', Symbol='LOC105377082', Aliases='', description='uncharacterized LOC105377082', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50454087, end_position_on_the_genomic_accession=50455566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377083, CurrentID=0, Status='live', Symbol='LOC105377083', Aliases='', description='uncharacterized LOC105377083', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50534962, end_position_on_the_genomic_accession=50539700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377084, CurrentID=0, Status='live', Symbol='LOC105377084', Aliases='', description='uncharacterized LOC105377084', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=50598983, end_position_on_the_genomic_accession=50600922, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377085, CurrentID=0, Status='live', Symbol='LOC105377085', Aliases='', description='uncharacterized LOC105377085', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51507070, end_position_on_the_genomic_accession=51536393, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377086, CurrentID=0, Status='live', Symbol='LOC105377086', Aliases='', description='uncharacterized LOC105377086', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51627772, end_position_on_the_genomic_accession=51662415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377087, CurrentID=0, Status='live', Symbol='LOC105377087', Aliases='', description='uncharacterized LOC105377087', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51754638, end_position_on_the_genomic_accession=51807090, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126272, CurrentID=0, Status='live', Symbol='EID2B', Aliases='EID-2B, EID-3', description='EP300 interacting inhibitor of differentiation 2B', other_designations='EID-2-like inhibitor of differentiation-3', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39531533, end_position_on_the_genomic_accession=39532867, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  402374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=402374, CurrentID=0, Status='live', Symbol='OR13D3P', Aliases='', description='olfactory receptor family 13 subfamily D member 3 pseudogene', other_designations='olfactory receptor, family 13, subfamily D, member 3 pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104722336, end_position_on_the_genomic_accession=104723269, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377090, CurrentID=0, Status='live', Symbol='LOC105377090', Aliases='', description='uncharacterized LOC105377090', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388419, CurrentID=0, Status='live', Symbol='BTBD17', Aliases='BTBD17A, LGALS3BPL, TANGO10A', description='BTB (POZ) domain containing 17', other_designations='BTB/POZ and BACK domain-containing protein LOC388419|transport and golgi organization 10 homolog A', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=74356416, end_position_on_the_genomic_accession=74363652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377092, CurrentID=0, Status='live', Symbol='LOC105377092', Aliases='', description='uncharacterized LOC105377092', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52606292, end_position_on_the_genomic_accession=52619538, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377093, CurrentID=0, Status='live', Symbol='LOC105377093', Aliases='', description='uncharacterized LOC105377093', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52796756, end_position_on_the_genomic_accession=52797810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377094, CurrentID=0, Status='live', Symbol='LOC105377094', Aliases='', description='uncharacterized LOC105377094', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53350498, end_position_on_the_genomic_accession=53371273, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377095, CurrentID=0, Status='live', Symbol='LOC105377095', Aliases='', description='uncharacterized LOC105377095', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53986291, end_position_on_the_genomic_accession=54035004, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105377096 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377096, CurrentID=0, Status='live', Symbol='LOC105377096', Aliases='', description='uncharacterized LOC105377096', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55186729, end_position_on_the_genomic_accession=55189602, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377097, CurrentID=0, Status='live', Symbol='LOC105377097', Aliases='', description='uncharacterized LOC105377097', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=55228283, end_position_on_the_genomic_accession=55259344, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  126282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126282, CurrentID=0, Status='live', Symbol='TNFAIP8L1', Aliases='TIPE1', description='TNF alpha induced protein 8 like 1', other_designations='TNF alpha-induced protein 8-like protein 1|TNFAIP8-like protein 1|oxidative stress-regulated gene-beta|oxy-beta|tumor necrosis factor, alpha induced protein 8 like 1|tumor necrosis factor, alpha-induced protein 8-like 1|tumor necrosis factor, alpha-induced protein 8-like protein 1', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=4639514, end_position_on_the_genomic_accession=4655568, orientation='plus', exon_count=4, OMIM=615869, no_hdr0=''),
  105377099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377099, CurrentID=0, Status='live', Symbol='LOC105377099', Aliases='', description='uncharacterized LOC105377099', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=56435970, end_position_on_the_genomic_accession=56441377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388428, CurrentID=0, Status='live', Symbol='AATK-AS1', Aliases='', description='AATK antisense RNA 1', other_designations='AATK antisense RNA 1 (non-protein coding)', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81165507, end_position_on_the_genomic_accession=81183164, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377101, CurrentID=0, Status='live', Symbol='LOC105377101', Aliases='', description='uncharacterized LOC105377101', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57149882, end_position_on_the_genomic_accession=57153150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257358, CurrentID=0, Status='live', Symbol='LINC01366', Aliases='', description='long intergenic non-protein coding RNA 1366', other_designations='CTB-114C7.3', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170331393, end_position_on_the_genomic_accession=170335100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377103, CurrentID=0, Status='live', Symbol='LOC105377103', Aliases='', description='uncharacterized LOC105377103', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57904285, end_position_on_the_genomic_accession=57920752, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377104, CurrentID=0, Status='live', Symbol='LOC105377104', Aliases='', description='uncharacterized LOC105377104', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57959568, end_position_on_the_genomic_accession=57986620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377105, CurrentID=0, Status='live', Symbol='LOC105377105', Aliases='', description='uncharacterized LOC105377105', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58162547, end_position_on_the_genomic_accession=58170635, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377106, CurrentID=0, Status='live', Symbol='LOC105377106', Aliases='', description='uncharacterized LOC105377106', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58351627, end_position_on_the_genomic_accession=58354408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377107, CurrentID=0, Status='live', Symbol='LOC105377107', Aliases='', description='uncharacterized LOC105377107', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58617068, end_position_on_the_genomic_accession=58618624, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  257364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257364, CurrentID=0, Status='live', Symbol='SNX33', Aliases='SH3PX3, SH3PXD3C, SNX30', description='sorting nexin 33', other_designations='SH3 and PX domain containing 3|SH3 and PX domain-containing protein 3', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75649007, end_position_on_the_genomic_accession=75658627, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100134229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134229, CurrentID=0, Status='live', Symbol='JHDM1D-AS1', Aliases='', description='JHDM1D antisense RNA 1 (head to head)', other_designations='', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140177261, end_position_on_the_genomic_accession=140179640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377110, CurrentID=0, Status='live', Symbol='LOC105377110', Aliases='', description='uncharacterized LOC105377110', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59050176, end_position_on_the_genomic_accession=59536998, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  126295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126295, CurrentID=0, Status='live', Symbol='ZNF57', Aliases='ZNF424', description='zinc finger protein 57', other_designations='zinc finger protein 424', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2900898, end_position_on_the_genomic_accession=2918646, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377112, CurrentID=0, Status='live', Symbol='LOC105377112', Aliases='', description='uncharacterized LOC105377112', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59984903, end_position_on_the_genomic_accession=59987531, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377113, CurrentID=0, Status='live', Symbol='LOC105377113', Aliases='', description='uncharacterized LOC105377113', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=59850926, end_position_on_the_genomic_accession=59879503, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126298, CurrentID=0, Status='live', Symbol='IRGQ', Aliases='FKSG271, IRGQ', description='immunity-related GTPase family, Q', other_designations='immunity-related GTPase family, Q1', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43584367, end_position_on_the_genomic_accession=43596229, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  126299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126299, CurrentID=0, Status='live', Symbol='ZNF428', Aliases='C19orf37, Zfp428', description='zinc finger protein 428', other_designations='enzyme-like protein PIT13', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43607224, end_position_on_the_genomic_accession=43619862, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377116, CurrentID=0, Status='live', Symbol='LOC105377116', Aliases='', description='uncharacterized LOC105377116', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62211394, end_position_on_the_genomic_accession=62218799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377117, CurrentID=0, Status='live', Symbol='LOC105377117', Aliases='', description='uncharacterized LOC105377117', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62144898, end_position_on_the_genomic_accession=62150052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377118, CurrentID=0, Status='live', Symbol='LOC105377118', Aliases='', description='uncharacterized LOC105377118', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62086623, end_position_on_the_genomic_accession=62130271, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377119, CurrentID=0, Status='live', Symbol='LOC105377119', Aliases='', description='uncharacterized LOC105377119', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63703708, end_position_on_the_genomic_accession=63711737, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377120, CurrentID=0, Status='live', Symbol='LOC105377120', Aliases='', description='uncharacterized LOC105377120', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=63969986, end_position_on_the_genomic_accession=63973999, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377121, CurrentID=0, Status='live', Symbol='LOC105377121', Aliases='', description='uncharacterized LOC105377121', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64024105, end_position_on_the_genomic_accession=64026360, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126306, CurrentID=0, Status='live', Symbol='JSRP1', Aliases='JP-45, JP45', description='junctional sarcoplasmic reticulum protein 1', other_designations='2310032K21Rik|homolog of mouse skeletal muscle sarcoplasmic reticulum protein JP-45|junctional-face membrane protein of 45 kDa homolog|skeletal muscle sarcoplasmic reticulum protein JP-45', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2252251, end_position_on_the_genomic_accession=2256423, orientation='minus', exon_count=7, OMIM=608743, no_hdr0=''),
  105377123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377123, CurrentID=0, Status='live', Symbol='LOC105377123', Aliases='', description='uncharacterized LOC105377123', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=64453513, end_position_on_the_genomic_accession=64494217, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  126308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126308, CurrentID=0, Status='live', Symbol='MOB3A', Aliases='MOB1C, MOBKL2A, moblak', description='MOB kinase activator 3A', other_designations='MOB LAK|MOB-LAK|MOB1, Mps One Binder kinase activator-like 2A|mob1 homolog 2A|mps one binder kinase activator-like 2A', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=2071032, end_position_on_the_genomic_accession=2096502, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377125, CurrentID=0, Status='live', Symbol='LOC105377125', Aliases='', description='uncharacterized LOC105377125', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65107817, end_position_on_the_genomic_accession=65115415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377126, CurrentID=0, Status='live', Symbol='LOC105377126', Aliases='', description='uncharacterized LOC105377126', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65113796, end_position_on_the_genomic_accession=65131047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377127, CurrentID=0, Status='live', Symbol='LOC105377127', Aliases='', description='uncharacterized LOC105377127', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65161632, end_position_on_the_genomic_accession=65174828, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388456, CurrentID=0, Status='live', Symbol='LOC388456', Aliases='', description='uncharacterized LOC388456', other_designations='', map_location='18p11.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=926083, end_position_on_the_genomic_accession=927993, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377129, CurrentID=0, Status='live', Symbol='LOC105377129', Aliases='', description='uncharacterized LOC105377129', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65974389, end_position_on_the_genomic_accession=65980286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377130, CurrentID=0, Status='live', Symbol='LOC105377130', Aliases='', description='uncharacterized LOC105377130', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=65717325, end_position_on_the_genomic_accession=65720231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377131, CurrentID=0, Status='live', Symbol='LOC105377131', Aliases='', description='putative uncharacterized protein YIL100C-A', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388460, CurrentID=0, Status='live', Symbol='RPL6P22', Aliases='RPL6_7_890', description='ribosomal protein L6 pseudogene 22', other_designations='', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33859348, end_position_on_the_genomic_accession=33860248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377133, CurrentID=0, Status='live', Symbol='LOC105377133', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=48424610, end_position_on_the_genomic_accession=48429742, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377134, CurrentID=0, Status='live', Symbol='LOC105377134', Aliases='', description='uncharacterized LOC105377134', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14027416, end_position_on_the_genomic_accession=14060261, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377135, CurrentID=0, Status='live', Symbol='LOC105377135', Aliases='', description='transcription initiation factor TFIID subunit 4-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52999045, end_position_on_the_genomic_accession=53048123, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377136, CurrentID=0, Status='live', Symbol='LOC105377136', Aliases='', description='uncharacterized LOC105377136', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=32880277, end_position_on_the_genomic_accession=32893764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  126321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126321, CurrentID=0, Status='live', Symbol='MFSD12', Aliases='C19orf28, PP3501', description='major facilitator superfamily domain containing 12', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3538273, end_position_on_the_genomic_accession=3557588, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105377138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377138, CurrentID=0, Status='live', Symbol='LOC105377138', Aliases='', description='uncharacterized LOC105377138', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44158736, end_position_on_the_genomic_accession=44159922, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100134259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134259, CurrentID=0, Status='live', Symbol='LINC01119', Aliases='', description='long intergenic non-protein coding RNA 1119', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46827864, end_position_on_the_genomic_accession=46859006, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388468, CurrentID=0, Status='live', Symbol='POTEC', Aliases='A26B2, CT104.6, POTE-18, POTE18', description='POTE ankyrin domain family member C', other_designations='ANKRD26-like family B member 2|ANKRD26-like family B, member 2|POTE ankyrin domain family, member C|cancer/testis antigen family 104, member 6|prostate, ovary, testis-expressed protein on chromosome 18', map_location='18p11.21', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14509778, end_position_on_the_genomic_accession=14543600, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  257397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257397, CurrentID=0, Status='live', Symbol='TAB3', Aliases='MAP3K7IP3, NAP1', description='TGF-beta activated kinase 1/MAP3K7 binding protein 3', other_designations='NF-kappa-B-activating protein 1|NFkB activating protein 1|TAB-3|TAK1-binding protein 3|TGF-beta-activated kinase 1-binding protein 3|mitogen-activated protein kinase kinase kinase 7 interacting protein 3', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30827442, end_position_on_the_genomic_accession=30889394, orientation='minus', exon_count=15, OMIM=300480, no_hdr0=''),
  126326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126326, CurrentID=0, Status='live', Symbol='GIPC3', Aliases='C19orf64, DFNB15, DFNB72, DFNB95', description='GIPC PDZ domain containing family member 3', other_designations='GIPC PDZ domain containing family, member 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=3585541, end_position_on_the_genomic_accession=3593541, orientation='plus', exon_count=6, OMIM=608792, no_hdr0=''),
  105377143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377143, CurrentID=0, Status='live', Symbol='LOC105377143', Aliases='', description='uncharacterized LOC105377143', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=66730378, end_position_on_the_genomic_accession=66738314, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  126328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126328, CurrentID=0, Status='live', Symbol='NDUFA11', Aliases='B14.7, CI-B14.7', description='NADH:ubiquinone oxidoreductase subunit A11', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 11, 14.7kDa|NADH-ubiquinone oxidoreductase subunit B14.7|complex I B14.7 subunit', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=5891276, end_position_on_the_genomic_accession=5904013, orientation='minus', exon_count=6, OMIM=612638, no_hdr0=''),
  105377145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377145, CurrentID=0, Status='live', Symbol='LOC105377145', Aliases='', description='uncharacterized LOC105377145', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67240855, end_position_on_the_genomic_accession=67246465, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388474, CurrentID=0, Status='live', Symbol='RPL7AP66', Aliases='RPL7A_30_1597', description='ribosomal protein L7a pseudogene 66', other_designations='', map_location='18q12.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=39334817, end_position_on_the_genomic_accession=39335692, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377147, CurrentID=0, Status='live', Symbol='LOC105377147', Aliases='', description='uncharacterized LOC105377147', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67858576, end_position_on_the_genomic_accession=67888150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377148, CurrentID=0, Status='live', Symbol='LOC105377148', Aliases='', description='uncharacterized LOC105377148', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=67617023, end_position_on_the_genomic_accession=67620497, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377149, CurrentID=0, Status='live', Symbol='LOC105377149', Aliases='', description='uncharacterized LOC105377149', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=69794312, end_position_on_the_genomic_accession=69799693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388478, CurrentID=0, Status='live', Symbol='SDCCAG3P1', Aliases='SDCCAG3L', description='serologically defined colon cancer antigen 3 pseudogene 1', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=60009985, end_position_on_the_genomic_accession=60011073, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  257407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257407, CurrentID=0, Status='live', Symbol='C2orf72', Aliases='', description='chromosome 2 open reading frame 72', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231037566, end_position_on_the_genomic_accession=231049714, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377152, CurrentID=0, Status='live', Symbol='LOC105377152', Aliases='', description='uncharacterized LOC105377152', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70575130, end_position_on_the_genomic_accession=70598459, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377153, CurrentID=0, Status='live', Symbol='LOC105377153', Aliases='', description='uncharacterized LOC105377153', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=70617886, end_position_on_the_genomic_accession=70622434, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377154, CurrentID=0, Status='live', Symbol='LOC105377154', Aliases='', description='uncharacterized LOC105377154', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71566702, end_position_on_the_genomic_accession=71571112, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377155, CurrentID=0, Status='live', Symbol='LOC105377155', Aliases='', description='uncharacterized LOC105377155', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71627685, end_position_on_the_genomic_accession=71628667, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377156, CurrentID=0, Status='live', Symbol='LOC105377156', Aliases='', description='uncharacterized LOC105377156', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71785331, end_position_on_the_genomic_accession=71789374, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377157, CurrentID=0, Status='live', Symbol='LOC105377157', Aliases='', description='uncharacterized LOC105377157', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71993699, end_position_on_the_genomic_accession=71996134, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377158, CurrentID=0, Status='live', Symbol='LOC105377158', Aliases='', description='uncharacterized LOC105377158', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72234120, end_position_on_the_genomic_accession=72307259, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  257415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257415, CurrentID=0, Status='live', Symbol='FAM133B', Aliases='', description='family with sequence similarity 133 member B', other_designations='family with sequence similarity 133, member B', map_location='7q21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=92560758, end_position_on_the_genomic_accession=92590394, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105377160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377160, CurrentID=0, Status='live', Symbol='LOC105377160', Aliases='', description='uncharacterized LOC105377160', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72458720, end_position_on_the_genomic_accession=72463575, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377161, CurrentID=0, Status='live', Symbol='LOC105377161', Aliases='', description='uncharacterized LOC105377161', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72582130, end_position_on_the_genomic_accession=72647837, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377162, CurrentID=0, Status='live', Symbol='LOC105377162', Aliases='', description='uncharacterized LOC105377162', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72663301, end_position_on_the_genomic_accession=72664253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377163, CurrentID=0, Status='live', Symbol='LOC105377163', Aliases='', description='uncharacterized LOC105377163', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72912587, end_position_on_the_genomic_accession=72929887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377164, CurrentID=0, Status='live', Symbol='LOC105377164', Aliases='', description='uncharacterized LOC105377164', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=73693884, end_position_on_the_genomic_accession=73707023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929538, CurrentID=0, Status='live', Symbol='LOC101929538', Aliases='', description='uncharacterized LOC101929538', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=128681240, end_position_on_the_genomic_accession=128687011, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377166, CurrentID=0, Status='live', Symbol='LOC105377166', Aliases='', description='uncharacterized LOC105377166', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74614973, end_position_on_the_genomic_accession=74622500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377167, CurrentID=0, Status='live', Symbol='LOC105377167', Aliases='', description='uncharacterized LOC105377167', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74825210, end_position_on_the_genomic_accession=74884243, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377168, CurrentID=0, Status='live', Symbol='LOC105377168', Aliases='', description='uncharacterized LOC105377168', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=74900471, end_position_on_the_genomic_accession=74907586, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  126353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126353, CurrentID=0, Status='live', Symbol='MISP', Aliases='C19orf21', description='mitotic spindle positioning', other_designations='caprice|mitotic interactor and substrate of Plk1|mitotic spindle positioning protein', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=748421, end_position_on_the_genomic_accession=764318, orientation='plus', exon_count=7, OMIM=615289, no_hdr0=''),
  105377170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377170, CurrentID=0, Status='live', Symbol='LOC105377170', Aliases='', description='uncharacterized LOC105377170', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75393739, end_position_on_the_genomic_accession=75396371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  388499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388499, CurrentID=0, Status='live', Symbol='LYPLA2P2', Aliases='', description='lysophospholipase II pseudogene 2', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7878619, end_position_on_the_genomic_accession=7880232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377172, CurrentID=0, Status='live', Symbol='LOC105377172', Aliases='', description='uncharacterized LOC105377172', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=78038717, end_position_on_the_genomic_accession=78048884, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377173, CurrentID=0, Status='live', Symbol='LOC105377173', Aliases='', description='uncharacterized LOC105377173', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=78052738, end_position_on_the_genomic_accession=78065497, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377174, CurrentID=0, Status='live', Symbol='LOC105377174', Aliases='', description='uncharacterized LOC105377174', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80823158, end_position_on_the_genomic_accession=80825427, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388503, CurrentID=0, Status='live', Symbol='C3P1', Aliases='CPLP', description='complement component 3 precursor pseudogene', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10041356, end_position_on_the_genomic_accession=10074138, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  105377176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377176, CurrentID=0, Status='live', Symbol='LOC105377176', Aliases='', description='uncharacterized LOC105377176', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80679966, end_position_on_the_genomic_accession=80722517, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377177, CurrentID=0, Status='live', Symbol='LOC105377177', Aliases='', description='uncharacterized LOC105377177', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80519774, end_position_on_the_genomic_accession=80769829, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377178, CurrentID=0, Status='live', Symbol='LOC105377178', Aliases='', description='uncharacterized LOC105377178', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=81840547, end_position_on_the_genomic_accession=81872955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388507, CurrentID=0, Status='live', Symbol='ZNF788', Aliases='', description='zinc finger family member 788', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=12092263, end_position_on_the_genomic_accession=12114678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126364, CurrentID=0, Status='live', Symbol='LRRC25', Aliases='MAPA', description='leucine rich repeat containing 25', other_designations='monocyte and plasmacytoid activated molecule|monocyte and plasmacytoid-activated protein', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=18391137, end_position_on_the_genomic_accession=18397625, orientation='minus', exon_count=2, OMIM=607518, no_hdr0=''),
  105377181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377181, CurrentID=0, Status='live', Symbol='LOC105377181', Aliases='', description='uncharacterized LOC105377181', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83391660, end_position_on_the_genomic_accession=83404337, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377182, CurrentID=0, Status='live', Symbol='LOC105377182', Aliases='', description='uncharacterized LOC105377182', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18758535, end_position_on_the_genomic_accession=18769783, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  650655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650655, CurrentID=0, Status='live', Symbol='ABCA17P', Aliases='ABCA17', description='ATP binding cassette subfamily A member 17, pseudogene', other_designations='ATP-binding cassette, sub-family A (ABC1), member 17, pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2340922, end_position_on_the_genomic_accession=2426699, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  388512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388512, CurrentID=0, Status='live', Symbol='CLEC17A', Aliases='', description='C-type lectin domain family 17 member A', other_designations='prolectin', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14582914, end_position_on_the_genomic_accession=14611144, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105377185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377185, CurrentID=0, Status='live', Symbol='LOC105377185', Aliases='', description='uncharacterized LOC105377185', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19933599, end_position_on_the_genomic_accession=19937972, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126370, CurrentID=0, Status='live', Symbol='OR1I1', Aliases='OR19-20P, OR1I1Q, OR1I1', description='olfactory receptor family 1 subfamily I member 1', other_designations='olfactory receptor 19-20|olfactory receptor, family 1, subfamily I, member 1', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15087066, end_position_on_the_genomic_accession=15088133, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377187, CurrentID=0, Status='live', Symbol='LOC105377187', Aliases='', description='uncharacterized LOC105377187', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83851576, end_position_on_the_genomic_accession=83857499, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377188, CurrentID=0, Status='live', Symbol='LOC105377188', Aliases='', description='uncharacterized LOC105377188', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=83953628, end_position_on_the_genomic_accession=84052828, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105377189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377189, CurrentID=0, Status='live', Symbol='LOC105377189', Aliases='', description='putative uncharacterized protein ENSP00000383309', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20360709, end_position_on_the_genomic_accession=20362095, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126374, CurrentID=0, Status='live', Symbol='WTIP', Aliases='', description='Wilms tumor 1 interacting protein', other_designations='WT1-interacting protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34481901, end_position_on_the_genomic_accession=34509133, orientation='plus', exon_count=11, OMIM=614790, no_hdr0=''),
  126375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126375, CurrentID=0, Status='live', Symbol='ZNF792', Aliases='', description='zinc finger protein 792', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=34956354, end_position_on_the_genomic_accession=34964282, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377192, CurrentID=0, Status='live', Symbol='LOC105377192', Aliases='', description='uncharacterized LOC105377192', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=84882014, end_position_on_the_genomic_accession=84892953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377193, CurrentID=0, Status='live', Symbol='LOC105377193', Aliases='', description='uncharacterized LOC105377193', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=84910298, end_position_on_the_genomic_accession=84918834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377194, CurrentID=0, Status='live', Symbol='LOC105377194', Aliases='', description='uncharacterized LOC105377194', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=85627285, end_position_on_the_genomic_accession=85636277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388523, CurrentID=0, Status='live', Symbol='ZNF728', Aliases='', description='zinc finger protein 728', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22974883, end_position_on_the_genomic_accession=23003208, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  388524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388524, CurrentID=0, Status='live', Symbol='RPSAP58', Aliases='RPSA_30_1642', description='ribosomal protein SA pseudogene 58', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=23763014, end_position_on_the_genomic_accession=23828135, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  650669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650669, CurrentID=0, Status='live', Symbol='GAS6-AS1', Aliases='', description='GAS6 antisense RNA 1', other_designations='GAS6 antisense RNA 1 (non-protein coding)', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113815610, end_position_on_the_genomic_accession=113842841, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126382, CurrentID=0, Status='live', Symbol='NR2C2AP', Aliases='TRA16', description='nuclear receptor 2C2-associated protein', other_designations='TR4 orphan receptor associated protein TRA16|TR4 orphan receptor-associated 16 kDa protein|repressor for TR4 transactivation', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=19201409, end_position_on_the_genomic_accession=19203435, orientation='minus', exon_count=6, OMIM=608719, no_hdr0=''),
  105377199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377199, CurrentID=0, Status='live', Symbol='LOC105377199', Aliases='', description='uncharacterized LOC105377199', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36388167, end_position_on_the_genomic_accession=36411036, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377200, CurrentID=0, Status='live', Symbol='LOC105377200', Aliases='', description='zinc finger protein Rlf-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88059221, end_position_on_the_genomic_accession=88129791, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377201, CurrentID=0, Status='live', Symbol='LOC105377201', Aliases='', description='uncharacterized LOC105377201', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=36872546, end_position_on_the_genomic_accession=36872992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377202, CurrentID=0, Status='live', Symbol='LOC105377202', Aliases='', description='uncharacterized LOC105377202', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=88275217, end_position_on_the_genomic_accession=88327779, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388531, CurrentID=0, Status='live', Symbol='RGS9BP', Aliases='PERRS, R9AP, RGS9', description='regulator of G-protein signaling 9 binding protein', other_designations='RGS9 anchor protein|RGS9-anchoring protein|regulator of G protein signaling 9 binding protein', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32675407, end_position_on_the_genomic_accession=32678300, orientation='plus', exon_count=1, OMIM=607814, no_hdr0=''),
  388532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388532, CurrentID=0, Status='live', Symbol='RPL21P131', Aliases='RPL21_64_1647', description='ribosomal protein L21 pseudogene 131', other_designations='', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=33589167, end_position_on_the_genomic_accession=33589728, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388533, CurrentID=0, Status='live', Symbol='KRTDAP', Aliases='KDAP, UNQ467', description='keratinocyte differentiation-associated protein', other_designations='KIPV467', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35487324, end_position_on_the_genomic_accession=35490531, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377206, CurrentID=0, Status='live', Symbol='LOC105377206', Aliases='', description='5E5 antigen-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=71580821, end_position_on_the_genomic_accession=71584038, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  152137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152137, CurrentID=0, Status='live', Symbol='CCDC50', Aliases='C3orf6, DFNA44, YMER', description='coiled-coil domain containing 50', other_designations='protein Ymer', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=191329082, end_position_on_the_genomic_accession=191398670, orientation='plus', exon_count=13, OMIM=611051, no_hdr0=''),
  388536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388536, CurrentID=0, Status='live', Symbol='ZNF790', Aliases='', description='zinc finger protein 790', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=36818322, end_position_on_the_genomic_accession=36850787, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  126393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126393, CurrentID=0, Status='live', Symbol='HSPB6', Aliases='HEL55, Hsp20, PPP1R91', description='heat shock protein family B (small) member 6', other_designations='epididymis luminal protein 55|heat shock 20 kDa-like protein p20|heat shock protein family B (small) member B6|heat shock protein, alpha-crystallin-related, B6|protein phosphatase 1, regulatory subunit 91', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35754566, end_position_on_the_genomic_accession=35757029, orientation='minus', exon_count=3, OMIM=610695, no_hdr0=''),
  105377210 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377210, CurrentID=0, Status='live', Symbol='LOC105377210', Aliases='', description='uncharacterized LOC105377210', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=56560878, end_position_on_the_genomic_accession=56573249, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377211, CurrentID=0, Status='live', Symbol='LOC105377211', Aliases='', description='uncharacterized LOC105377211', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63136067, end_position_on_the_genomic_accession=63155860, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  257468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257468, CurrentID=0, Status='live', Symbol='MEIS3P2', Aliases='', description='Meis homeobox 3 pseudogene 2', other_designations='Meis1 homolog 3 pseudogene 2|putative homeobox protein Meis3-like 2', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20588849, end_position_on_the_genomic_accession=20590815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377213, CurrentID=0, Status='live', Symbol='LOC105377213', Aliases='', description='uncharacterized LOC105377213', other_designations='XX-FW80269A6.1', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=151515546, end_position_on_the_genomic_accession=151516270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377214, CurrentID=0, Status='live', Symbol='LOC105377214', Aliases='', description='uncharacterized LOC105377214', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120057290, end_position_on_the_genomic_accession=120061414, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377215, CurrentID=0, Status='live', Symbol='LOC105377215', Aliases='', description='uncharacterized LOC105377215', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=12662323, end_position_on_the_genomic_accession=12663793, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377216, CurrentID=0, Status='live', Symbol='LOC105377216', Aliases='', description='uncharacterized LOC105377216', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=15455989, end_position_on_the_genomic_accession=15460850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377217, CurrentID=0, Status='live', Symbol='LOC105377217', Aliases='', description='uncharacterized LOC105377217', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=17553294, end_position_on_the_genomic_accession=17558834, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  126402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126402, CurrentID=0, Status='live', Symbol='CCDC105', Aliases='', description='coiled-coil domain containing 105', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15010727, end_position_on_the_genomic_accession=15023271, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377219, CurrentID=0, Status='live', Symbol='LOC105377219', Aliases='', description='uncharacterized LOC105377219', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18216751, end_position_on_the_genomic_accession=18217177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377220, CurrentID=0, Status='live', Symbol='LOC105377220', Aliases='', description='uncharacterized LOC105377220', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18347177, end_position_on_the_genomic_accession=18352716, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377221, CurrentID=0, Status='live', Symbol='LOC105377221', Aliases='', description='uncharacterized LOC105377221', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18484189, end_position_on_the_genomic_accession=18517866, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  388550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388550, CurrentID=0, Status='live', Symbol='CEACAM22P', Aliases='', description='carcinoembryonic antigen-related cell adhesion molecule 22, pseudogene', other_designations='carcinoembryonic antigen-related cell adhesion molecule 2, pseudogene', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44537058, end_position_on_the_genomic_accession=44556917, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  388551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388551, CurrentID=0, Status='live', Symbol='CEACAM16', Aliases='CEAL2, DFNA4B', description='carcinoembryonic antigen-related cell adhesion molecule 16', other_designations='carcinoembryonic antigen like-2 protein', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44697504, end_position_on_the_genomic_accession=44710714, orientation='plus', exon_count=8, OMIM=614591, no_hdr0=''),
  388552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388552, CurrentID=0, Status='live', Symbol='BLOC1S3', Aliases='BLOS3, HPS8, RP', description='biogenesis of lysosomal organelles complex-1, subunit 3', other_designations='BLOC-1 subunit 3|biogenesis of lysosome-related organelles complex-1, subunit 3|reduced pigmentation, mouse, homolog of', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45178745, end_position_on_the_genomic_accession=45217078, orientation='plus', exon_count=8, OMIM=609762, no_hdr0=''),
  388553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388553, CurrentID=0, Status='live', Symbol='BHMG1', Aliases='HMGDC', description='basic helix-loop-helix and HMG box domain containing 1', other_designations='HMG domain containing', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=45723892, end_position_on_the_genomic_accession=45764539, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  126410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126410, CurrentID=0, Status='live', Symbol='CYP4F22', Aliases='ARCI5, INLNE, LI3', description='cytochrome P450 family 4 subfamily F member 22', other_designations='cytochrome P450, family 2, subfamily E, polypeptide 2 homolog|cytochrome P450, family 4, subfamily F, polypeptide 22', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15508484, end_position_on_the_genomic_accession=15552317, orientation='plus', exon_count=16, OMIM=611495, no_hdr0=''),
  388555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388555, CurrentID=0, Status='live', Symbol='IGFL3', Aliases='UNQ483', description='IGF like family member 3', other_designations='IGF-like family member 3|RPRC483', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46119863, end_position_on_the_genomic_accession=46124705, orientation='minus', exon_count=4, OMIM=610546, no_hdr0=''),
  388556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388556, CurrentID=0, Status='live', Symbol='RPS9P4', Aliases='RPS9_1_1662', description='ribosomal protein S9 pseudogene 4', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49719426, end_position_on_the_genomic_accession=49720137, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377229, CurrentID=0, Status='live', Symbol='LOC105377229', Aliases='', description='uncharacterized LOC105377229', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=7831622, end_position_on_the_genomic_accession=7858758, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388558, CurrentID=0, Status='live', Symbol='ZNF808', Aliases='', description='zinc finger protein 808', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52527652, end_position_on_the_genomic_accession=52570178, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  388559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388559, CurrentID=0, Status='live', Symbol='ZNF888', Aliases='', description='zinc finger protein 888', other_designations='CTD-2331H12.6', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=52905893, end_position_on_the_genomic_accession=52923452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  388560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388560, CurrentID=0, Status='live', Symbol='ERVFRD-2', Aliases='', description='endogenous retrovirus group FRD member 2', other_designations='endogenous retrovirus group FRD, member 2', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  388561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388561, CurrentID=0, Status='live', Symbol='ZNF761', Aliases='ZNF468', description='zinc finger protein 761', other_designations='zinc finger protein 468', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53431974, end_position_on_the_genomic_accession=53458261, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377234, CurrentID=0, Status='live', Symbol='LOC105377234', Aliases='', description='uncharacterized LOC105377234', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8693336, end_position_on_the_genomic_accession=8695298, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377235, CurrentID=0, Status='live', Symbol='LOC105377235', Aliases='', description='uncharacterized LOC105377235', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=8721851, end_position_on_the_genomic_accession=8727521, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  388564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388564, CurrentID=0, Status='live', Symbol='TMEM238', Aliases='', description='transmembrane protein 238', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=55379244, end_position_on_the_genomic_accession=55384259, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377237, CurrentID=0, Status='live', Symbol='LOC105377237', Aliases='', description='uncharacterized LOC105377237', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=21837427, end_position_on_the_genomic_accession=21839144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388566, CurrentID=0, Status='live', Symbol='ZNF470', Aliases='CZF-1', description='zinc finger protein 470', other_designations='chondrogenesis zinc finger protein 1', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=56567081, end_position_on_the_genomic_accession=56588903, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  388567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388567, CurrentID=0, Status='live', Symbol='ZNF749', Aliases='', description='zinc finger protein 749', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57428568, end_position_on_the_genomic_accession=57447485, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377240, CurrentID=0, Status='live', Symbol='LOC105377240', Aliases='', description='uncharacterized LOC105377240', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=23435271, end_position_on_the_genomic_accession=23474400, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388569, CurrentID=0, Status='live', Symbol='ZNF324B', Aliases='', description='zinc finger protein 324B', other_designations='', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58418560, end_position_on_the_genomic_accession=58457832, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377242, CurrentID=0, Status='live', Symbol='LOC105377242', Aliases='', description='uncharacterized LOC105377242', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24926665, end_position_on_the_genomic_accession=24929994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929551, CurrentID=0, Status='live', Symbol='LOC101929551', Aliases='', description='uncharacterized LOC101929551', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=7886767, end_position_on_the_genomic_accession=7899655, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377244 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377244, CurrentID=0, Status='live', Symbol='LOC105377244', Aliases='', description='uncharacterized LOC105377244', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=26404926, end_position_on_the_genomic_accession=26409003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377245, CurrentID=0, Status='live', Symbol='LOC105377245', Aliases='', description='uncharacterized LOC105377245', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=58939321, end_position_on_the_genomic_accession=58988152, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388574, CurrentID=0, Status='live', Symbol='RPL23AP87', Aliases='', description='ribosomal protein L23a pseudogene 87', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=83226897, end_position_on_the_genomic_accession=83240804, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377247, CurrentID=0, Status='live', Symbol='LOC105377247', Aliases='', description='uncharacterized LOC105377247', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=59040587, end_position_on_the_genomic_accession=59046957, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126432, CurrentID=0, Status='live', Symbol='RINL', Aliases='', description='Ras and Rab interactor like', other_designations='Ras and Rab interactor-like', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38867830, end_position_on_the_genomic_accession=38878279, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  126433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126433, CurrentID=0, Status='live', Symbol='FBXO27', Aliases='FBG5, Fbx27', description='F-box protein 27', other_designations='F-box protein FBG5|F-box/G-domain protein 5', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39024021, end_position_on_the_genomic_accession=39032596, orientation='minus', exon_count=6, OMIM=609099, no_hdr0=''),
  105377250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377250, CurrentID=0, Status='live', Symbol='LOC105377250', Aliases='', description='uncharacterized LOC105377250', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=61808443, end_position_on_the_genomic_accession=61818311, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388579, CurrentID=0, Status='live', Symbol='TUBB8P11', Aliases='', description='tubulin beta 8 class VIII pseudogene 11', other_designations='TUBB4Q pseudogene|tubulin, beta 8 class VIII pseudogene 11', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=873291, end_position_on_the_genomic_accession=874514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377252, CurrentID=0, Status='live', Symbol='LOC105377252', Aliases='', description='uncharacterized LOC105377252', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62522724, end_position_on_the_genomic_accession=62563929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388581, CurrentID=0, Status='live', Symbol='FAM132A', Aliases='C1QDC2, C1QTNF12, CTRP12', description='family with sequence similarity 132 member A', other_designations='C1q domain containing 2|C1q/TNF-related protein 12|adipose-derived insulin-sensitizing factor|complement C1q tumor necrosis factor-related protein 12|family with sequence similarity 132, member A|protein FAM132A', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1242446, end_position_on_the_genomic_accession=1247218, orientation='minus', exon_count=9, OMIM=616593, no_hdr0=''),
  105377254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377254, CurrentID=0, Status='live', Symbol='LOC105377254', Aliases='', description='uncharacterized LOC105377254', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=63761544, end_position_on_the_genomic_accession=63794955, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377255, CurrentID=0, Status='live', Symbol='LOC105377255', Aliases='', description='uncharacterized LOC105377255', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64870687, end_position_on_the_genomic_accession=64873426, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377256, CurrentID=0, Status='live', Symbol='LOC105377256', Aliases='', description='uncharacterized LOC105377256', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65153731, end_position_on_the_genomic_accession=65241794, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  388585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388585, CurrentID=0, Status='live', Symbol='HES5', Aliases='bHLHb38', description='hes family bHLH transcription factor 5', other_designations='class B basic helix-loop-helix protein 38|hairy and enhancer of split 5', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2528745, end_position_on_the_genomic_accession=2530245, orientation='minus', exon_count=3, OMIM=607348, no_hdr0=''),
  105377258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377258, CurrentID=0, Status='live', Symbol='LOC105377258', Aliases='', description='uncharacterized LOC105377258', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65780638, end_position_on_the_genomic_accession=65784672, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377259, CurrentID=0, Status='live', Symbol='LOC105377259', Aliases='', description='uncharacterized LOC105377259', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=65868476, end_position_on_the_genomic_accession=65886184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388588, CurrentID=0, Status='live', Symbol='SMIM1', Aliases='Vel', description='small integral membrane protein 1 (Vel blood group)', other_designations='vel blood group antigen', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3772761, end_position_on_the_genomic_accession=3775982, orientation='plus', exon_count=4, OMIM=615242, no_hdr0=''),
  105377261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377261, CurrentID=0, Status='live', Symbol='LOC105377261', Aliases='', description='uncharacterized LOC105377261', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=66003268, end_position_on_the_genomic_accession=66046620, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377262, CurrentID=0, Status='live', Symbol='LOC105377262', Aliases='', description='uncharacterized LOC105377262', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=66862864, end_position_on_the_genomic_accession=67077385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388591, CurrentID=0, Status='live', Symbol='RNF207', Aliases='C1orf188', description='ring finger protein 207', other_designations='', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6205775, end_position_on_the_genomic_accession=6221299, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105377264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377264, CurrentID=0, Status='live', Symbol='LOC105377264', Aliases='', description='uncharacterized LOC105377264', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=67748546, end_position_on_the_genomic_accession=67753709, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377265, CurrentID=0, Status='live', Symbol='LOC105377265', Aliases='', description='uncharacterized LOC105377265', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69047083, end_position_on_the_genomic_accession=69056931, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377266, CurrentID=0, Status='live', Symbol='LOC105377266', Aliases='', description='uncharacterized LOC105377266', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69220530, end_position_on_the_genomic_accession=69225008, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388595, CurrentID=0, Status='live', Symbol='TMEM82', Aliases='', description='transmembrane protein 82', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=15742492, end_position_on_the_genomic_accession=15747797, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377268, CurrentID=0, Status='live', Symbol='LOC105377268', Aliases='', description='uncharacterized LOC105377268', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69822501, end_position_on_the_genomic_accession=69838531, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377269, CurrentID=0, Status='live', Symbol='LOC105377269', Aliases='', description='uncharacterized LOC105377269', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=69986834, end_position_on_the_genomic_accession=69993435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377270, CurrentID=0, Status='live', Symbol='LOC105377270', Aliases='', description='uncharacterized LOC105377270', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70171510, end_position_on_the_genomic_accession=70185437, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377271, CurrentID=0, Status='live', Symbol='LOC105377271', Aliases='', description='uncharacterized LOC105377271', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=71622509, end_position_on_the_genomic_accession=71653233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377272, CurrentID=0, Status='live', Symbol='LOC105377272', Aliases='', description='uncharacterized LOC105377272', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73710339, end_position_on_the_genomic_accession=73734745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377273, CurrentID=0, Status='live', Symbol='LOC105377273', Aliases='', description='uncharacterized LOC105377273', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=72953842, end_position_on_the_genomic_accession=72998555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377274, CurrentID=0, Status='live', Symbol='LOC105377274', Aliases='', description='uncharacterized LOC105377274', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=73998773, end_position_on_the_genomic_accession=74007941, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392447, CurrentID=0, Status='live', Symbol='RPL32P36', Aliases='RPL32_24_1772', description='ribosomal protein L32 pseudogene 36', other_designations='', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=40897216, end_position_on_the_genomic_accession=40908390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377276, CurrentID=0, Status='live', Symbol='LOC105377276', Aliases='', description='uncharacterized LOC105377276', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74278045, end_position_on_the_genomic_accession=74330316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388605, CurrentID=0, Status='live', Symbol='HS6ST1P1', Aliases='HS6ST1P', description='heparan sulfate 6-O-sulfotransferase 1 pseudogene 1', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=21428268, end_position_on_the_genomic_accession=21432234, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388606, CurrentID=0, Status='live', Symbol='SDHDP7', Aliases='', description='succinate dehydrogenase complex subunit D, integral membrane protein pseudogene 7', other_designations='succinate dehydrogenase complex, subunit D, integral membrane protein pseudogene 7', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=25398695, end_position_on_the_genomic_accession=25400004, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377279, CurrentID=0, Status='live', Symbol='LOC105377279', Aliases='', description='uncharacterized LOC105377279', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74925704, end_position_on_the_genomic_accession=74932997, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377280, CurrentID=0, Status='live', Symbol='LOC105377280', Aliases='', description='uncharacterized LOC105377280', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74883352, end_position_on_the_genomic_accession=74896108, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377281, CurrentID=0, Status='live', Symbol='LOC105377281', Aliases='', description='uncharacterized LOC105377281', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=74881532, end_position_on_the_genomic_accession=74882935, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388610, CurrentID=0, Status='live', Symbol='TRNP1', Aliases='C1orf225, TNRP', description='TMF1-regulated nuclear protein 1', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26993667, end_position_on_the_genomic_accession=27000886, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388611, CurrentID=0, Status='live', Symbol='CD164L2', Aliases='EAPG6122, UNQ6122', description='CD164 sialomucin-like 2', other_designations='', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=27379100, end_position_on_the_genomic_accession=27383333, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377284, CurrentID=0, Status='live', Symbol='LOC105377284', Aliases='', description='uncharacterized LOC105377284', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=75822951, end_position_on_the_genomic_accession=75838700, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101929558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929558, CurrentID=0, Status='live', Symbol='LOC101929558', Aliases='', description='uncharacterized LOC101929558', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105743893, end_position_on_the_genomic_accession=105896968, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377286, CurrentID=0, Status='live', Symbol='LOC105377286', Aliases='', description='uncharacterized LOC105377286', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76239728, end_position_on_the_genomic_accession=76251673, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377287, CurrentID=0, Status='live', Symbol='LOC105377287', Aliases='', description='uncharacterized LOC105377287', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76418764, end_position_on_the_genomic_accession=76428702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377288, CurrentID=0, Status='live', Symbol='LOC105377288', Aliases='', description='uncharacterized LOC105377288', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76353288, end_position_on_the_genomic_accession=76354264, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377289, CurrentID=0, Status='live', Symbol='LOC105377289', Aliases='', description='uncharacterized LOC105377289', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76742078, end_position_on_the_genomic_accession=76802524, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377290, CurrentID=0, Status='live', Symbol='LOC105377290', Aliases='', description='uncharacterized LOC105377290', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76624149, end_position_on_the_genomic_accession=76646863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377291, CurrentID=0, Status='live', Symbol='LOC105377291', Aliases='', description='uncharacterized LOC105377291', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77030976, end_position_on_the_genomic_accession=77041202, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377292, CurrentID=0, Status='live', Symbol='LOC105377292', Aliases='', description='uncharacterized LOC105377292', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76996623, end_position_on_the_genomic_accession=77011727, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  388621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388621, CurrentID=0, Status='live', Symbol='RPL21P20', Aliases='RPL21_3_36', description='ribosomal protein L21 pseudogene 20', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40333042, end_position_on_the_genomic_accession=40333598, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377294, CurrentID=0, Status='live', Symbol='LOC105377294', Aliases='', description='uncharacterized LOC105377294', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=76906678, end_position_on_the_genomic_accession=76949691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377295, CurrentID=0, Status='live', Symbol='LOC105377295', Aliases='', description='uncharacterized LOC105377295', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77392275, end_position_on_the_genomic_accession=77409281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377296, CurrentID=0, Status='live', Symbol='LOC105377296', Aliases='', description='uncharacterized LOC105377296', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77580909, end_position_on_the_genomic_accession=77589175, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377297 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377297, CurrentID=0, Status='live', Symbol='LOC105377297', Aliases='', description='uncharacterized LOC105377297', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77630637, end_position_on_the_genomic_accession=77636939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377298, CurrentID=0, Status='live', Symbol='LOC105377298', Aliases='', description='uncharacterized LOC105377298', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79288759, end_position_on_the_genomic_accession=79298703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126307, CurrentID=0, Status='live', Symbol='MIR892B', Aliases='MIRN892B', description='microRNA 892b', other_designations='hsa-mir-892b', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=145997198, end_position_on_the_genomic_accession=145997274, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377300, CurrentID=0, Status='live', Symbol='LOC105377300', Aliases='', description='uncharacterized LOC105377300', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79377825, end_position_on_the_genomic_accession=79399209, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377301, CurrentID=0, Status='live', Symbol='LOC105377301', Aliases='', description='uncharacterized LOC105377301', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=79663718, end_position_on_the_genomic_accession=79696454, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388630, CurrentID=0, Status='live', Symbol='TRABD2B', Aliases='TIKI2', description='TraB domain containing 2B', other_designations='TRAB domain-containing protein 2B|UPF0632 protein A|heart, kidney and adipose-enriched transmembrane protein homolog', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47760528, end_position_on_the_genomic_accession=47997371, orientation='minus', exon_count=12, OMIM=614913, no_hdr0=''),
  105377303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377303, CurrentID=0, Status='live', Symbol='LOC105377303', Aliases='', description='uncharacterized LOC105377303', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80181264, end_position_on_the_genomic_accession=80191465, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377304, CurrentID=0, Status='live', Symbol='LOC105377304', Aliases='', description='uncharacterized LOC105377304', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=80219529, end_position_on_the_genomic_accession=80255440, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388633, CurrentID=0, Status='live', Symbol='LDLRAD1', Aliases='', description='low density lipoprotein receptor class A domain containing 1', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54007298, end_position_on_the_genomic_accession=54018186, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377306 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377306, CurrentID=0, Status='live', Symbol='LOC105377306', Aliases='', description='uncharacterized LOC105377306', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=81499239, end_position_on_the_genomic_accession=81502129, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377307, CurrentID=0, Status='live', Symbol='LOC105377307', Aliases='', description='uncharacterized LOC105377307', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82285013, end_position_on_the_genomic_accession=82297508, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377308, CurrentID=0, Status='live', Symbol='LOC105377308', Aliases='', description='uncharacterized LOC105377308', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82242959, end_position_on_the_genomic_accession=82285250, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377309, CurrentID=0, Status='live', Symbol='LOC105377309', Aliases='', description='uncharacterized LOC105377309', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82430923, end_position_on_the_genomic_accession=82441166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377310, CurrentID=0, Status='live', Symbol='LOC105377310', Aliases='', description='syncytin-1-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82469579, end_position_on_the_genomic_accession=82480384, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377311, CurrentID=0, Status='live', Symbol='LOC105377311', Aliases='', description='uncharacterized LOC105377311', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83115241, end_position_on_the_genomic_accession=83138388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377312, CurrentID=0, Status='live', Symbol='LOC105377312', Aliases='', description='uncharacterized LOC105377312', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83109941, end_position_on_the_genomic_accession=83120977, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377313, CurrentID=0, Status='live', Symbol='LOC105377313', Aliases='', description='uncharacterized LOC105377313', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83297869, end_position_on_the_genomic_accession=83316932, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377314, CurrentID=0, Status='live', Symbol='LOC105377314', Aliases='', description='uncharacterized LOC105377314', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83360571, end_position_on_the_genomic_accession=83374227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377315 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377315, CurrentID=0, Status='live', Symbol='LOC105377315', Aliases='', description='uncharacterized LOC105377315', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=83796446, end_position_on_the_genomic_accession=83803132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  650788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650788, CurrentID=0, Status='live', Symbol='RPS28P9', Aliases='RPS28_5_1524', description='ribosomal protein S28 pseudogene 9', other_designations='', map_location='17p11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18572721, end_position_on_the_genomic_accession=18573117, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377317, CurrentID=0, Status='live', Symbol='LOC105377317', Aliases='', description='uncharacterized LOC105377317', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84443077, end_position_on_the_genomic_accession=84456355, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388646, CurrentID=0, Status='live', Symbol='GBP7', Aliases='GBP4L', description='guanylate binding protein 7', other_designations='GBP-7|GTP-binding protein 7|guanine nucleotide-binding protein 7', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89131751, end_position_on_the_genomic_accession=89176040, orientation='minus', exon_count=11, OMIM=612468, no_hdr0=''),
  105377319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377319, CurrentID=0, Status='live', Symbol='LOC105377319', Aliases='', description='uncharacterized LOC105377319', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=85646373, end_position_on_the_genomic_accession=85660799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377320, CurrentID=0, Status='live', Symbol='LOC105377320', Aliases='', description='uncharacterized LOC105377320', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87065373, end_position_on_the_genomic_accession=87073658, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388649, CurrentID=0, Status='live', Symbol='C1orf146', Aliases='', description='chromosome 1 open reading frame 146', other_designations='', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92217912, end_position_on_the_genomic_accession=92245812, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  388650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388650, CurrentID=0, Status='live', Symbol='FAM69A', Aliases='', description='family with sequence similarity 69 member A', other_designations='family with sequence similarity 69, member A', map_location='1p22.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92832729, end_position_on_the_genomic_accession=92961522, orientation='minus', exon_count=7, OMIM=614542, no_hdr0=''),
  105377323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377323, CurrentID=0, Status='live', Symbol='LOC105377323', Aliases='', description='uncharacterized LOC105377323', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=87657202, end_position_on_the_genomic_accession=87673022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100134444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134444, CurrentID=0, Status='live', Symbol='KCNJ18', Aliases='KIR2.6, TTPP2', description='potassium channel, inwardly rectifying subfamily J, member 18', other_designations='inward rectifier K(+) channel Kir2.6|thyrotoxic periodic paralysis susceptibility ion channel', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=21692523, end_position_on_the_genomic_accession=21705700, orientation='plus', exon_count=4, OMIM=613236, no_hdr0=''),
  100134445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134445, CurrentID=0, Status='live', Symbol='LOC100134445', Aliases='', description='uncharacterized LOC100134445', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377326, CurrentID=0, Status='live', Symbol='LOC105377326', Aliases='', description='uncharacterized LOC105377326', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=88361575, end_position_on_the_genomic_accession=88362153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  414301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414301, CurrentID=0, Status='live', Symbol='DDI1', Aliases='', description='DNA damage inducible 1 homolog 1', other_designations='DDI1, DNA-damage inducible 1, homolog 1|DNA-damage inducible 1 homolog 1|DNA-damage inducible protein 1', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=104036580, end_position_on_the_genomic_accession=104039194, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377328, CurrentID=0, Status='live', Symbol='LOC105377328', Aliases='', description='uncharacterized LOC105377328', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89456541, end_position_on_the_genomic_accession=89485559, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377329, CurrentID=0, Status='live', Symbol='LOC105377329', Aliases='', description='uncharacterized LOC105377329', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89490972, end_position_on_the_genomic_accession=89720426, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105377330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377330, CurrentID=0, Status='live', Symbol='LOC105377330', Aliases='', description='uncharacterized LOC105377330', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=89739184, end_position_on_the_genomic_accession=89741540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377331, CurrentID=0, Status='live', Symbol='LOC105377331', Aliases='', description='uncharacterized LOC105377331', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=90039614, end_position_on_the_genomic_accession=90074847, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377332, CurrentID=0, Status='live', Symbol='LOC105377332', Aliases='', description='uncharacterized LOC105377332', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=91268910, end_position_on_the_genomic_accession=91303522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377333, CurrentID=0, Status='live', Symbol='LOC105377333', Aliases='', description='uncharacterized LOC105377333', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=90244057, end_position_on_the_genomic_accession=90265847, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388662, CurrentID=0, Status='live', Symbol='SLC6A17', Aliases='MRT48, NTT4', description='solute carrier family 6 (neutral amino acid transporter), member 17', other_designations='orphan sodium- and chloride-dependent neurotransmitter transporter NTT4|sodium-dependent neurotransmitter transporter NTT4|solute carrier family 6 (neurotransmitter transporter), member 17|solute carrier family 6, member 17', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110150510, end_position_on_the_genomic_accession=110202202, orientation='plus', exon_count=13, OMIM=610299, no_hdr0=''),
  105377335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377335, CurrentID=0, Status='live', Symbol='LOC105377335', Aliases='', description='uncharacterized LOC105377335', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95549142, end_position_on_the_genomic_accession=95552322, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  650808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650808, CurrentID=0, Status='live', Symbol='RPL4P1', Aliases='RPL4_2_1340', description='ribosomal protein L4 pseudogene 1', other_designations='', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21750212, end_position_on_the_genomic_accession=21751630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377337, CurrentID=0, Status='live', Symbol='LOC105377337', Aliases='', description='uncharacterized LOC105377337', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=95555609, end_position_on_the_genomic_accession=95586299, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377338, CurrentID=0, Status='live', Symbol='LOC105377338', Aliases='', description='uncharacterized LOC105377338', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=96170456, end_position_on_the_genomic_accession=96631238, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  388667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388667, CurrentID=0, Status='live', Symbol='C1orf137', Aliases='', description='chromosome 1 open reading frame 137', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116694112, end_position_on_the_genomic_accession=116706603, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377340, CurrentID=0, Status='live', Symbol='LOC105377340', Aliases='', description='uncharacterized LOC105377340', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98183546, end_position_on_the_genomic_accession=98203799, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377341, CurrentID=0, Status='live', Symbol='LOC105377341', Aliases='', description='uncharacterized LOC105377341', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98658912, end_position_on_the_genomic_accession=98664550, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126526, CurrentID=0, Status='live', Symbol='C19orf47', Aliases='', description='chromosome 19 open reading frame 47', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40295271, end_position_on_the_genomic_accession=40348527, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105377343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377343, CurrentID=0, Status='live', Symbol='LOC105377343', Aliases='', description='uncharacterized LOC105377343', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=98783794, end_position_on_the_genomic_accession=98801296, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377344, CurrentID=0, Status='live', Symbol='LOC105377344', Aliases='', description='uncharacterized LOC105377344', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100483390, end_position_on_the_genomic_accession=100487025, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377345, CurrentID=0, Status='live', Symbol='LOC105377345', Aliases='', description='uncharacterized LOC105377345', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=100531936, end_position_on_the_genomic_accession=100546470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377346, CurrentID=0, Status='live', Symbol='LOC105377346', Aliases='', description='uncharacterized LOC105377346', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=101640946, end_position_on_the_genomic_accession=101650217, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377347, CurrentID=0, Status='live', Symbol='LOC105377347', Aliases='', description='uncharacterized LOC105377347', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102597670, end_position_on_the_genomic_accession=102616271, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377348, CurrentID=0, Status='live', Symbol='LOC105377348', Aliases='', description='uncharacterized LOC105377348', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102828096, end_position_on_the_genomic_accession=102844099, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388677, CurrentID=0, Status='live', Symbol='NOTCH2NL', Aliases='N2N', description='notch 2 N-terminal like', other_designations='Notch homolog 2 N-terminal like protein', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146151908, end_position_on_the_genomic_accession=146229032, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377350, CurrentID=0, Status='live', Symbol='LOC105377350', Aliases='', description='uncharacterized LOC105377350', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104283396, end_position_on_the_genomic_accession=104394349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377351, CurrentID=0, Status='live', Symbol='LOC105377351', Aliases='', description='uncharacterized LOC105377351', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104892899, end_position_on_the_genomic_accession=104895394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126536, CurrentID=0, Status='live', Symbol='LINC00661', Aliases='', description='long intergenic non-protein coding RNA 661', other_designations='', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=16015634, end_position_on_the_genomic_accession=16027462, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377353, CurrentID=0, Status='live', Symbol='LOC105377353', Aliases='', description='uncharacterized LOC105377353', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=105700029, end_position_on_the_genomic_accession=105706265, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377354, CurrentID=0, Status='live', Symbol='LOC105377354', Aliases='', description='uncharacterized LOC105377354', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106279498, end_position_on_the_genomic_accession=106296212, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377355, CurrentID=0, Status='live', Symbol='LOC105377355', Aliases='', description='uncharacterized LOC105377355', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106433495, end_position_on_the_genomic_accession=106457332, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377356, CurrentID=0, Status='live', Symbol='LOC105377356', Aliases='', description='uncharacterized LOC105377356', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=106525843, end_position_on_the_genomic_accession=106843997, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  126541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126541, CurrentID=0, Status='live', Symbol='OR10H4', Aliases='OR19-28', description='olfactory receptor family 10 subfamily H member 4', other_designations='olfactory receptor OR19-28|olfactory receptor, family 10, subfamily H, member 4', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15949008, end_position_on_the_genomic_accession=15949958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377358, CurrentID=0, Status='live', Symbol='LOC105377358', Aliases='', description='uncharacterized LOC105377358', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=107546467, end_position_on_the_genomic_accession=107590397, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377359, CurrentID=0, Status='live', Symbol='LOC105377359', Aliases='', description='uncharacterized LOC105377359', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108668142, end_position_on_the_genomic_accession=108679401, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377360, CurrentID=0, Status='live', Symbol='LOC105377360', Aliases='', description='uncharacterized LOC105377360', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109185760, end_position_on_the_genomic_accession=109206004, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377361, CurrentID=0, Status='live', Symbol='LOC105377361', Aliases='', description='uncharacterized LOC105377361', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109613094, end_position_on_the_genomic_accession=109619117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377362, CurrentID=0, Status='live', Symbol='LOC105377362', Aliases='', description='uncharacterized LOC105377362', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110365290, end_position_on_the_genomic_accession=110373433, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100422243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422243, CurrentID=0, Status='live', Symbol='NPM1P27', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 27', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93682519, end_position_on_the_genomic_accession=93683816, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388692, CurrentID=0, Status='live', Symbol='LOC388692', Aliases='', description='uncharacterized LOC388692', other_designations='', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=143792495, end_position_on_the_genomic_accession=143797108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126549, CurrentID=0, Status='live', Symbol='ANKLE1', Aliases='ANKRD41, LEM3, LEMD6', description='ankyrin repeat and LEM domain containing 1', other_designations='LEM domain containing 6|LEM-domain containing 3|ankyrin repeat domain-containing protein 41', map_location='19p13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=17281645, end_position_on_the_genomic_accession=17287646, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377366, CurrentID=0, Status='live', Symbol='LOC105377366', Aliases='', description='uncharacterized LOC105377366', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111642252, end_position_on_the_genomic_accession=111648785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388695, CurrentID=0, Status='live', Symbol='LYSMD1', Aliases='SB145', description='LysM, putative peptidoglycan-binding, domain containing 1', other_designations='RP11-68I18.5', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151156810, end_position_on_the_genomic_accession=151165948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377368, CurrentID=0, Status='live', Symbol='LOC105377368', Aliases='', description='histone-lysine N-methyltransferase SETMAR-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=111826331, end_position_on_the_genomic_accession=111844523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388697, CurrentID=0, Status='live', Symbol='HRNR', Aliases='FLG3, S100A16, S100a18', description='hornerin', other_designations='filaggrin family member 3|intermediate filament-associated protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152212076, end_position_on_the_genomic_accession=152224196, orientation='minus', exon_count=3, OMIM=616293, no_hdr0=''),
  388698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388698, CurrentID=0, Status='live', Symbol='FLG2', Aliases='IFPS', description='filaggrin family member 2', other_designations='ifapsoriasin|intermediate filament-associated and psoriasis susceptibility protein|intermediate filament-associated and psoriasis-susceptibility protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152348735, end_position_on_the_genomic_accession=152360010, orientation='minus', exon_count=4, OMIM=616284, no_hdr0=''),
  388699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388699, CurrentID=0, Status='live', Symbol='LINC00302', Aliases='C1orf46, NCRNA00302, xp33', description='long intergenic non-protein coding RNA 302', other_designations='skin-specific protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152655429, end_position_on_the_genomic_accession=152656805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377372, CurrentID=0, Status='live', Symbol='LOC105377372', Aliases='', description='uncharacterized LOC105377372', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=112523664, end_position_on_the_genomic_accession=112546850, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388701, CurrentID=0, Status='live', Symbol='C1orf189', Aliases='', description='chromosome 1 open reading frame 189', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154199085, end_position_on_the_genomic_accession=154206365, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377374, CurrentID=0, Status='live', Symbol='LOC105377374', Aliases='', description='uncharacterized LOC105377374', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113402482, end_position_on_the_genomic_accession=113434480, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377375, CurrentID=0, Status='live', Symbol='LOC105377375', Aliases='', description='uncharacterized LOC105377375', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=113871746, end_position_on_the_genomic_accession=113876592, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377376, CurrentID=0, Status='live', Symbol='LOC105377376', Aliases='', description='uncharacterized LOC105377376', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114023390, end_position_on_the_genomic_accession=114036551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377377, CurrentID=0, Status='live', Symbol='LOC105377377', Aliases='', description='uncharacterized LOC105377377', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114269418, end_position_on_the_genomic_accession=114276941, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377378, CurrentID=0, Status='live', Symbol='LOC105377378', Aliases='', description='uncharacterized LOC105377378', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114334783, end_position_on_the_genomic_accession=114364951, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  388707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388707, CurrentID=0, Status='live', Symbol='RPSAP18', Aliases='RPSA_4_107', description='ribosomal protein SA pseudogene 18', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=160266269, end_position_on_the_genomic_accession=160267303, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377380, CurrentID=0, Status='live', Symbol='LOC105377380', Aliases='', description='uncharacterized LOC105377380', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114549781, end_position_on_the_genomic_accession=114555668, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377381, CurrentID=0, Status='live', Symbol='LOC105377381', Aliases='', description='uncharacterized LOC105377381', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114683067, end_position_on_the_genomic_accession=114689484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377382, CurrentID=0, Status='live', Symbol='LOC105377382', Aliases='', description='uncharacterized LOC105377382', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=114724153, end_position_on_the_genomic_accession=114731737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126567, CurrentID=0, Status='live', Symbol='C2CD4C', Aliases='FAM148C, KIAA1957, NLF3', description='C2 calcium-dependent domain containing 4C', other_designations='family with sequence similarity 148, member C|nuclear localized factor 3|nuclear-localized factor 3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=405435, end_position_on_the_genomic_accession=409170, orientation='minus', exon_count=3, OMIM=610336, no_hdr0=''),
  105377384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377384, CurrentID=0, Status='live', Symbol='LOC105377384', Aliases='', description='uncharacterized LOC105377384', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=116344096, end_position_on_the_genomic_accession=116355209, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  257641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=257641, CurrentID=0, Status='live', Symbol='NPCA1', Aliases='NPC1', description='Nasopharyngeal carcinoma 1', other_designations='', map_location='4p15.1-q12', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607107, no_hdr0=''),
  388714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388714, CurrentID=0, Status='live', Symbol='FMO6P', Aliases='FMO6', description='flavin containing monooxygenase 6 pseudogene', other_designations='novel flavin containing monooxygenase', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171137740, end_position_on_the_genomic_accession=171161563, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100134507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134507, CurrentID=0, Status='live', Symbol='LOC100134507', Aliases='', description='metaxin 2 pseudogene', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=125344972, end_position_on_the_genomic_accession=125345754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377388, CurrentID=0, Status='live', Symbol='LOC105377388', Aliases='', description='uncharacterized LOC105377388', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117081069, end_position_on_the_genomic_accession=117105395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377389, CurrentID=0, Status='live', Symbol='LOC105377389', Aliases='', description='uncharacterized LOC105377389', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117315158, end_position_on_the_genomic_accession=117360795, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377390, CurrentID=0, Status='live', Symbol='LOC105377390', Aliases='', description='uncharacterized LOC105377390', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=117361333, end_position_on_the_genomic_accession=117389140, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  388719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388719, CurrentID=0, Status='live', Symbol='LINC00272', Aliases='C1orf120, NCRNA00272', description='long intergenic non-protein coding RNA 272', other_designations='RP1-223H12.3', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182407621, end_position_on_the_genomic_accession=182414813, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377392, CurrentID=0, Status='live', Symbol='LOC105377392', Aliases='', description='uncharacterized LOC105377392', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118251501, end_position_on_the_genomic_accession=118265951, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377393, CurrentID=0, Status='live', Symbol='LOC105377393', Aliases='', description='uncharacterized LOC105377393', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=118879225, end_position_on_the_genomic_accession=118884685, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388722, CurrentID=0, Status='live', Symbol='C1orf53', Aliases='', description='chromosome 1 open reading frame 53', other_designations='', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197902552, end_position_on_the_genomic_accession=197907367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377395, CurrentID=0, Status='live', Symbol='LOC105377395', Aliases='', description='uncharacterized LOC105377395', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119117846, end_position_on_the_genomic_accession=119128366, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377396, CurrentID=0, Status='live', Symbol='LOC105377396', Aliases='', description='uncharacterized LOC105377396', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=119628579, end_position_on_the_genomic_accession=119636722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126581, CurrentID=0, Status='live', Symbol='MRPL57P1', Aliases='MRP63P1', description='mitochondrial ribosomal protein L57 pseudogene 1', other_designations='mitochondrial ribosomal protein 63 pseudogene 1', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114279002, end_position_on_the_genomic_accession=114279286, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377398, CurrentID=0, Status='live', Symbol='LOC105377398', Aliases='', description='uncharacterized LOC105377398', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120563459, end_position_on_the_genomic_accession=120571233, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377399, CurrentID=0, Status='live', Symbol='LOC105377399', Aliases='', description='uncharacterized LOC105377399', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=120922946, end_position_on_the_genomic_accession=120927628, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377400, CurrentID=0, Status='live', Symbol='LOC105377400', Aliases='', description='uncharacterized LOC105377400', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121072780, end_position_on_the_genomic_accession=121081450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377401, CurrentID=0, Status='live', Symbol='LOC105377401', Aliases='', description='uncharacterized LOC105377401', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=121551527, end_position_on_the_genomic_accession=121582702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388730, CurrentID=0, Status='live', Symbol='TMEM81', Aliases='HC3107, KVLA2788, UNQ2788', description='transmembrane protein 81', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205083129, end_position_on_the_genomic_accession=205084460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377403, CurrentID=0, Status='live', Symbol='LOC105377403', Aliases='', description='uncharacterized LOC105377403', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=122077250, end_position_on_the_genomic_accession=122078310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377404, CurrentID=0, Status='live', Symbol='LOC105377404', Aliases='', description='uncharacterized LOC105377404', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123325659, end_position_on_the_genomic_accession=123326515, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377405, CurrentID=0, Status='live', Symbol='LOC105377405', Aliases='', description='uncharacterized LOC105377405', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=123533823, end_position_on_the_genomic_accession=123546326, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377406, CurrentID=0, Status='live', Symbol='LOC105377406', Aliases='', description='uncharacterized LOC105377406', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124185261, end_position_on_the_genomic_accession=124310723, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377407, CurrentID=0, Status='live', Symbol='LOC105377407', Aliases='', description='uncharacterized LOC105377407', other_designations='CTD-2325B11.1', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124033509, end_position_on_the_genomic_accession=124252040, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377408, CurrentID=0, Status='live', Symbol='LOC105377408', Aliases='', description='uncharacterized LOC105377408', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124342723, end_position_on_the_genomic_accession=124420328, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377409, CurrentID=0, Status='live', Symbol='LOC105377409', Aliases='', description='uncharacterized LOC105377409', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126221667, end_position_on_the_genomic_accession=126243094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377410, CurrentID=0, Status='live', Symbol='LOC105377410', Aliases='', description='uncharacterized LOC105377410', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126280041, end_position_on_the_genomic_accession=126284395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377411, CurrentID=0, Status='live', Symbol='LOC105377411', Aliases='', description='uncharacterized LOC105377411', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=126462239, end_position_on_the_genomic_accession=126464306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377412, CurrentID=0, Status='live', Symbol='LOC105377412', Aliases='', description='uncharacterized LOC105377412', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127565063, end_position_on_the_genomic_accession=127618205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377413, CurrentID=0, Status='live', Symbol='LOC105377413', Aliases='', description='uncharacterized LOC105377413', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127618888, end_position_on_the_genomic_accession=127632844, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377414, CurrentID=0, Status='live', Symbol='LOC105377414', Aliases='', description='uncharacterized LOC105377414', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127763373, end_position_on_the_genomic_accession=127776918, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388743, CurrentID=0, Status='live', Symbol='CAPN8', Aliases='nCL-2', description='calpain 8', other_designations='new calpain 2|stomach-specific M-type calpain', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=223609153, end_position_on_the_genomic_accession=223665734, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  105377416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377416, CurrentID=0, Status='live', Symbol='LOC105377416', Aliases='', description='uncharacterized LOC105377416', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129311029, end_position_on_the_genomic_accession=129314000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377417, CurrentID=0, Status='live', Symbol='LOC105377417', Aliases='', description='uncharacterized LOC105377417', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129363105, end_position_on_the_genomic_accession=129383409, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377418, CurrentID=0, Status='live', Symbol='LOC105377418', Aliases='', description='uncharacterized LOC105377418', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130023078, end_position_on_the_genomic_accession=130030760, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377419, CurrentID=0, Status='live', Symbol='LOC105377419', Aliases='', description='uncharacterized LOC105377419', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130124131, end_position_on_the_genomic_accession=130138189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377420, CurrentID=0, Status='live', Symbol='LOC105377420', Aliases='', description='uncharacterized LOC105377420', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130376254, end_position_on_the_genomic_accession=130388079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377421, CurrentID=0, Status='live', Symbol='LOC105377421', Aliases='', description='uncharacterized LOC105377421', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=130896432, end_position_on_the_genomic_accession=130901863, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377422, CurrentID=0, Status='live', Symbol='LOC105377422', Aliases='', description='uncharacterized LOC105377422', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131104704, end_position_on_the_genomic_accession=131136251, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377423, CurrentID=0, Status='live', Symbol='LOC105377423', Aliases='', description='uncharacterized LOC105377423', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131341809, end_position_on_the_genomic_accession=131343825, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377424, CurrentID=0, Status='live', Symbol='LOC105377424', Aliases='', description='uncharacterized LOC105377424', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131522923, end_position_on_the_genomic_accession=131528144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388753, CurrentID=0, Status='live', Symbol='COA6', Aliases='C1orf31', description='cytochrome c oxidase assembly factor 6', other_designations='', map_location='1q42.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=234373437, end_position_on_the_genomic_accession=234384049, orientation='plus', exon_count=4, OMIM=614772, no_hdr0=''),
  105377426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377426, CurrentID=0, Status='live', Symbol='LOC105377426', Aliases='', description='uncharacterized LOC105377426', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=131900279, end_position_on_the_genomic_accession=131901281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377427, CurrentID=0, Status='live', Symbol='LOC105377427', Aliases='', description='uncharacterized LOC105377427', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132002554, end_position_on_the_genomic_accession=132006399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377428, CurrentID=0, Status='live', Symbol='LOC105377428', Aliases='', description='uncharacterized LOC105377428', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132027265, end_position_on_the_genomic_accession=132033697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  650901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=650901, CurrentID=0, Status='live', Symbol='RPS2P12', Aliases='RPS2_2_35', description='ribosomal protein S2 pseudogene 12', other_designations='', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39962524, end_position_on_the_genomic_accession=39963441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377430, CurrentID=0, Status='live', Symbol='LOC105377430', Aliases='', description='uncharacterized LOC105377430', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=132743337, end_position_on_the_genomic_accession=132757295, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388759, CurrentID=0, Status='live', Symbol='C1orf229', Aliases='', description='chromosome 1 open reading frame 229', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247110160, end_position_on_the_genomic_accession=247112417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377432, CurrentID=0, Status='live', Symbol='LOC105377432', Aliases='', description='uncharacterized LOC105377432', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=133059045, end_position_on_the_genomic_accession=133060028, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388761, CurrentID=0, Status='live', Symbol='OR14A2', Aliases='OR5AX1, OR5AX1P', description='olfactory receptor family 14 subfamily A member 2', other_designations='olfactory receptor, family 14, subfamily A, member 2|olfactory receptor, family 5, subfamily AX, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247722899, end_position_on_the_genomic_accession=247724243, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388762, CurrentID=0, Status='live', Symbol='OR2M1P', Aliases='JCG10, OR2M1, OST037', description='olfactory receptor family 2 subfamily M member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily M, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248122136, end_position_on_the_genomic_accession=248122780, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377435, CurrentID=0, Status='live', Symbol='LOC105377435', Aliases='', description='uncharacterized LOC105377435', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134512088, end_position_on_the_genomic_accession=134518019, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377436, CurrentID=0, Status='live', Symbol='LOC105377436', Aliases='', description='uncharacterized LOC105377436', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134544623, end_position_on_the_genomic_accession=134640651, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377437, CurrentID=0, Status='live', Symbol='LOC105377437', Aliases='', description='uncharacterized LOC105377437', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134934262, end_position_on_the_genomic_accession=134947019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377438, CurrentID=0, Status='live', Symbol='LOC105377438', Aliases='', description='uncharacterized LOC105377438', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=134932702, end_position_on_the_genomic_accession=134940305, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377439, CurrentID=0, Status='live', Symbol='LOC105377439', Aliases='', description='uncharacterized LOC105377439', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135112763, end_position_on_the_genomic_accession=135123645, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377440, CurrentID=0, Status='live', Symbol='LOC105377440', Aliases='', description='uncharacterized LOC105377440', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=136690577, end_position_on_the_genomic_accession=136750919, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422256, CurrentID=0, Status='live', Symbol='NPM1P23', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 23', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=24096617, end_position_on_the_genomic_accession=24097360, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  126626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126626, CurrentID=0, Status='live', Symbol='GABPB2', Aliases='', description='GA binding protein transcription factor, beta subunit 2', other_designations='GABP subunit beta-2|GABPB-2', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151070604, end_position_on_the_genomic_accession=151129119, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105377443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377443, CurrentID=0, Status='live', Symbol='LOC105377443', Aliases='', description='uncharacterized LOC105377443', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137545731, end_position_on_the_genomic_accession=137603430, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377444, CurrentID=0, Status='live', Symbol='LOC105377444', Aliases='', description='uncharacterized LOC105377444', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137707205, end_position_on_the_genomic_accession=137734957, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377445, CurrentID=0, Status='live', Symbol='LOC105377445', Aliases='', description='uncharacterized LOC105377445', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137751401, end_position_on_the_genomic_accession=137757644, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377446, CurrentID=0, Status='live', Symbol='LOC105377446', Aliases='', description='uncharacterized LOC105377446', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138386245, end_position_on_the_genomic_accession=138414349, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377447, CurrentID=0, Status='live', Symbol='LOC105377447', Aliases='', description='uncharacterized LOC105377447', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137943279, end_position_on_the_genomic_accession=138051701, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105377448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377448, CurrentID=0, Status='live', Symbol='LOC105377448', Aliases='', description='uncharacterized LOC105377448', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138995912, end_position_on_the_genomic_accession=139012646, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377449, CurrentID=0, Status='live', Symbol='LOC105377449', Aliases='', description='uncharacterized LOC105377449', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139405367, end_position_on_the_genomic_accession=139412933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377450, CurrentID=0, Status='live', Symbol='LOC105377450', Aliases='', description='uncharacterized LOC105377450', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139393591, end_position_on_the_genomic_accession=139420276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377451, CurrentID=0, Status='live', Symbol='LOC105377451', Aliases='', description='uncharacterized LOC105377451', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139568548, end_position_on_the_genomic_accession=139578883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377452, CurrentID=0, Status='live', Symbol='LOC105377452', Aliases='', description='uncharacterized LOC105377452', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139929515, end_position_on_the_genomic_accession=139931574, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  126637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126637, CurrentID=0, Status='live', Symbol='TCHHL1', Aliases='S100A17, THHL1', description='trichohyalin like 1', other_designations='S100 calcium binding protein A17|S100 calcium-binding protein A17|basalin|intermediate filament-associated protein|trichohyalin-like 1', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152084144, end_position_on_the_genomic_accession=152089064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  126638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126638, CurrentID=0, Status='live', Symbol='RPTN', Aliases='', description='repetin', other_designations='intermediate filament-associated protein', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=152153595, end_position_on_the_genomic_accession=152159228, orientation='minus', exon_count=3, OMIM=613259, no_hdr0=''),
  105377455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377455, CurrentID=0, Status='live', Symbol='LOC105377455', Aliases='', description='uncharacterized LOC105377455', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=141428813, end_position_on_the_genomic_accession=141442575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377456, CurrentID=0, Status='live', Symbol='LOC105377456', Aliases='', description='uncharacterized LOC105377456', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=141568718, end_position_on_the_genomic_accession=141577157, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377457, CurrentID=0, Status='live', Symbol='LOC105377457', Aliases='', description='uncharacterized LOC105377457', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=142768637, end_position_on_the_genomic_accession=142792128, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377458, CurrentID=0, Status='live', Symbol='LOC105377458', Aliases='', description='uncharacterized LOC105377458', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143817794, end_position_on_the_genomic_accession=143829471, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377459, CurrentID=0, Status='live', Symbol='LOC105377459', Aliases='', description='uncharacterized LOC105377459', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=143843839, end_position_on_the_genomic_accession=143886988, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377460, CurrentID=0, Status='live', Symbol='LOC105377460', Aliases='', description='uncharacterized LOC105377460', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144080937, end_position_on_the_genomic_accession=144090185, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  388789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388789, CurrentID=0, Status='live', Symbol='LINC00493', Aliases='', description='long intergenic non-protein coding RNA 493', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18567429, end_position_on_the_genomic_accession=18569560, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377462, CurrentID=0, Status='live', Symbol='LOC105377462', Aliases='', description='uncharacterized LOC105377462', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144216135, end_position_on_the_genomic_accession=144562161, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377463, CurrentID=0, Status='live', Symbol='LOC105377463', Aliases='', description='uncharacterized LOC105377463', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145177543, end_position_on_the_genomic_accession=145178747, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377464, CurrentID=0, Status='live', Symbol='LOC105377464', Aliases='', description='uncharacterized LOC105377464', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145197366, end_position_on_the_genomic_accession=145232661, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377465, CurrentID=0, Status='live', Symbol='LOC105377465', Aliases='', description='uncharacterized LOC105377465', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145196745, end_position_on_the_genomic_accession=145203554, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377466, CurrentID=0, Status='live', Symbol='LOC105377466', Aliases='', description='uncharacterized LOC105377466', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=145335245, end_position_on_the_genomic_accession=145343528, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  388795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388795, CurrentID=0, Status='live', Symbol='EFCAB8', Aliases='', description='EF-hand calcium binding domain 8', other_designations='EF-hand domain-containing protein ENSP00000383366', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32860995, end_position_on_the_genomic_accession=32961845, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  388796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388796, CurrentID=0, Status='live', Symbol='SNHG17', Aliases='', description='small nucleolar RNA host gene 17', other_designations='small nucleolar RNA host gene 17 (non-protein coding)', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38420592, end_position_on_the_genomic_accession=38435375, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377469, CurrentID=0, Status='live', Symbol='LOC105377469', Aliases='', description='uncharacterized LOC105377469', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146105677, end_position_on_the_genomic_accession=146109625, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377470, CurrentID=0, Status='live', Symbol='LOC105377470', Aliases='', description='uncharacterized LOC105377470', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146122258, end_position_on_the_genomic_accession=146131967, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388799, CurrentID=0, Status='live', Symbol='FAM209B', Aliases='C20orf107, dJ1153D9.4', description='family with sequence similarity 209 member B', other_designations='family with sequence similarity 209, member B|uncharacterized protein C20orf107', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56526195, end_position_on_the_genomic_accession=56536521, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377472, CurrentID=0, Status='live', Symbol='LOC105377472', Aliases='', description='uncharacterized LOC105377472', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146294871, end_position_on_the_genomic_accession=146299071, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377473, CurrentID=0, Status='live', Symbol='LOC105377473', Aliases='', description='uncharacterized LOC105377473', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146934008, end_position_on_the_genomic_accession=146983894, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105377474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377474, CurrentID=0, Status='live', Symbol='LOC105377474', Aliases='', description='uncharacterized LOC105377474', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=146833237, end_position_on_the_genomic_accession=146833627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377475, CurrentID=0, Status='live', Symbol='LOC105377475', Aliases='', description='uncharacterized LOC105377475', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147002239, end_position_on_the_genomic_accession=147020964, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377476, CurrentID=0, Status='live', Symbol='LOC105377476', Aliases='', description='uncharacterized LOC105377476', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147071147, end_position_on_the_genomic_accession=147073226, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  126661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126661, CurrentID=0, Status='live', Symbol='CCDC163P', Aliases='C1orf231', description='coiled-coil domain containing 163, pseudogene', other_designations='', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45493926, end_position_on_the_genomic_accession=45500079, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377478, CurrentID=0, Status='live', Symbol='LOC105377478', Aliases='', description='uncharacterized LOC105377478', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=147557287, end_position_on_the_genomic_accession=147567438, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377479, CurrentID=0, Status='live', Symbol='LOC105377479', Aliases='', description='uncharacterized LOC105377479', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=149516880, end_position_on_the_genomic_accession=149521264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377480, CurrentID=0, Status='live', Symbol='LOC105377480', Aliases='', description='uncharacterized LOC105377480', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148940941, end_position_on_the_genomic_accession=148942425, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377481, CurrentID=0, Status='live', Symbol='LOC105377481', Aliases='', description='uncharacterized LOC105377481', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148783893, end_position_on_the_genomic_accession=148835218, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105377482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377482, CurrentID=0, Status='live', Symbol='LOC105377482', Aliases='', description='uncharacterized LOC105377482', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=148719365, end_position_on_the_genomic_accession=148720357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929591, CurrentID=0, Status='live', Symbol='KIZ-AS1', Aliases='', description='KIZ antisense RNA 1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21194404, end_position_on_the_genomic_accession=21218289, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  126668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126668, CurrentID=0, Status='live', Symbol='TDRD10', Aliases='', description='tudor domain containing 10', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154502219, end_position_on_the_genomic_accession=154548147, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  126669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126669, CurrentID=0, Status='live', Symbol='SHE', Aliases='', description='Src homology 2 domain containing E', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154469637, end_position_on_the_genomic_accession=154502050, orientation='minus', exon_count=7, OMIM=610482, no_hdr0=''),
  105377486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377486, CurrentID=0, Status='live', Symbol='LOC105377486', Aliases='', description='uncharacterized LOC105377486', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151401077, end_position_on_the_genomic_accession=151409002, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  388815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388815, CurrentID=0, Status='live', Symbol='MIR99AHG', Aliases='C21orf34, C21orf35, FLJ38295, LINC00478, MONC', description='mir-99a-let-7c cluster host gene', other_designations='long intergenic non-protein coding RNA 478|megakaryocytic oncogenic non-coding RNA|mir-99a-let-7c cluster host gene (non-protein coding)', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=16070522, end_position_on_the_genomic_accession=16609775, orientation='plus', exon_count=10, OMIM=615964, no_hdr0=''),
  105377488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377488, CurrentID=0, Status='live', Symbol='LOC105377488', Aliases='', description='uncharacterized LOC105377488', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=151912883, end_position_on_the_genomic_accession=151940051, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100422264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422264, CurrentID=0, Status='live', Symbol='CYP51A1P3', Aliases='CYP51P3', description='cytochrome P450 family 51 subfamily A member 1 pseudogene 3', other_designations='cytochrome P450, family 51, pseudogene 3|cytochrome P450, family 51, subfamily A, polypeptide 1 pseudogene 3', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148479353, end_position_on_the_genomic_accession=148480760, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388818, CurrentID=0, Status='live', Symbol='KRTAP26-1', Aliases='', description='keratin associated protein 26-1', other_designations='', map_location='21q22.11', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=30319131, end_position_on_the_genomic_accession=30320289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377491, CurrentID=0, Status='live', Symbol='LOC105377491', Aliases='', description='uncharacterized LOC105377491', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152316469, end_position_on_the_genomic_accession=152320442, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377492, CurrentID=0, Status='live', Symbol='LOC105377492', Aliases='', description='uncharacterized LOC105377492', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152268345, end_position_on_the_genomic_accession=152310880, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377493, CurrentID=0, Status='live', Symbol='LOC105377493', Aliases='', description='uncharacterized LOC105377493', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152552250, end_position_on_the_genomic_accession=152553758, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377494, CurrentID=0, Status='live', Symbol='LOC105377494', Aliases='', description='uncharacterized LOC105377494', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152666363, end_position_on_the_genomic_accession=152670139, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  283257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283257, CurrentID=0, Status='live', Symbol='TRIM51FP', Aliases='RNF18B, TRIM49B', description='tripartite motif-containing 51F, pseudogene', other_designations='Putative tripartite motif-containing protein 49B|SPRY domain containing 5 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49832966, end_position_on_the_genomic_accession=49839330, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377496, CurrentID=0, Status='live', Symbol='LOC105377496', Aliases='', description='uncharacterized LOC105377496', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153226159, end_position_on_the_genomic_accession=153249213, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377497, CurrentID=0, Status='live', Symbol='LOC105377497', Aliases='', description='uncharacterized LOC105377497', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153640174, end_position_on_the_genomic_accession=153641506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377498, CurrentID=0, Status='live', Symbol='LOC105377498', Aliases='', description='uncharacterized LOC105377498', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153527527, end_position_on_the_genomic_accession=153542511, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377499, CurrentID=0, Status='live', Symbol='LOC105377499', Aliases='', description='uncharacterized LOC105377499', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=153712738, end_position_on_the_genomic_accession=153750679, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377500, CurrentID=0, Status='live', Symbol='LOC105377500', Aliases='', description='uncharacterized LOC105377500', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=154766265, end_position_on_the_genomic_accession=154781873, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377501, CurrentID=0, Status='live', Symbol='LOC105377501', Aliases='', description='uncharacterized LOC105377501', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155076202, end_position_on_the_genomic_accession=155082503, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388830 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388830, CurrentID=0, Status='live', Symbol='LINC00316', Aliases='C21orf111, NCRNA00316, PRED59', description='long intergenic non-protein coding RNA 316', other_designations='', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=45338590, end_position_on_the_genomic_accession=45341990, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377503, CurrentID=0, Status='live', Symbol='LOC105377503', Aliases='', description='uncharacterized LOC105377503', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155428514, end_position_on_the_genomic_accession=155448605, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377504, CurrentID=0, Status='live', Symbol='LOC105377504', Aliases='', description='uncharacterized LOC105377504', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155521550, end_position_on_the_genomic_accession=155522709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377505, CurrentID=0, Status='live', Symbol='LOC105377505', Aliases='', description='uncharacterized LOC105377505', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155495882, end_position_on_the_genomic_accession=155517196, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377506, CurrentID=0, Status='live', Symbol='LOC105377506', Aliases='', description='uncharacterized LOC105377506', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155713593, end_position_on_the_genomic_accession=155758233, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377507, CurrentID=0, Status='live', Symbol='LOC105377507', Aliases='', description='uncharacterized LOC105377507', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=155854761, end_position_on_the_genomic_accession=155858898, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377508, CurrentID=0, Status='live', Symbol='LOC105377508', Aliases='', description='uncharacterized LOC105377508', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=156305864, end_position_on_the_genomic_accession=156330534, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377509, CurrentID=0, Status='live', Symbol='LOC105377509', Aliases='', description='uncharacterized LOC105377509', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=157803425, end_position_on_the_genomic_accession=158059684, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377510, CurrentID=0, Status='live', Symbol='LOC105377510', Aliases='', description='uncharacterized LOC105377510', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=158356444, end_position_on_the_genomic_accession=158394903, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  126695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126695, CurrentID=0, Status='live', Symbol='KDF1', Aliases='C1orf172', description='keratinocyte differentiation factor 1', other_designations='RP11-344H11.3', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=26949556, end_position_on_the_genomic_accession=26960484, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377512, CurrentID=0, Status='live', Symbol='LOC105377512', Aliases='', description='uncharacterized LOC105377512', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=160518864, end_position_on_the_genomic_accession=160553346, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100422268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422268, CurrentID=0, Status='live', Symbol='BRIX1P1', Aliases='', description='BRX1, biogenesis of ribosomes pseudogene 1', other_designations='BRX1, biogenesis of ribosomes, homolog pseudogene', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=51435594, end_position_on_the_genomic_accession=51436519, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377514, CurrentID=0, Status='live', Symbol='LOC105377514', Aliases='', description='uncharacterized LOC105377514', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=161066632, end_position_on_the_genomic_accession=161124893, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377515, CurrentID=0, Status='live', Symbol='LOC105377515', Aliases='', description='uncharacterized LOC105377515', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=161234008, end_position_on_the_genomic_accession=161252202, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377516, CurrentID=0, Status='live', Symbol='LOC105377516', Aliases='', description='uncharacterized LOC105377516', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=162705767, end_position_on_the_genomic_accession=162736057, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377517, CurrentID=0, Status='live', Symbol='LOC105377517', Aliases='', description='uncharacterized LOC105377517', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163172276, end_position_on_the_genomic_accession=163179118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377518, CurrentID=0, Status='live', Symbol='LOC105377518', Aliases='', description='uncharacterized LOC105377518', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=163521544, end_position_on_the_genomic_accession=163528661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377519, CurrentID=0, Status='live', Symbol='LOC105377519', Aliases='', description='uncharacterized LOC105377519', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165348411, end_position_on_the_genomic_accession=165361542, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377520, CurrentID=0, Status='live', Symbol='LOC105377520', Aliases='', description='uncharacterized LOC105377520', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165505222, end_position_on_the_genomic_accession=165507464, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388849, CurrentID=0, Status='live', Symbol='CCDC188', Aliases='', description='coiled-coil domain containing 188', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20148417, end_position_on_the_genomic_accession=20151259, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377522, CurrentID=0, Status='live', Symbol='LOC105377522', Aliases='', description='uncharacterized LOC105377522', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=167308199, end_position_on_the_genomic_accession=167424409, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377523 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377523, CurrentID=0, Status='live', Symbol='LOC105377523', Aliases='', description='uncharacterized LOC105377523', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168037651, end_position_on_the_genomic_accession=168042709, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377524, CurrentID=0, Status='live', Symbol='LOC105377524', Aliases='', description='uncharacterized LOC105377524', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168169239, end_position_on_the_genomic_accession=168199047, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377525, CurrentID=0, Status='live', Symbol='LOC105377525', Aliases='', description='uncharacterized LOC105377525', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168778905, end_position_on_the_genomic_accession=168792280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377526, CurrentID=0, Status='live', Symbol='LOC105377526', Aliases='', description='uncharacterized LOC105377526', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=168650152, end_position_on_the_genomic_accession=168677325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377527, CurrentID=0, Status='live', Symbol='LOC105377527', Aliases='', description='uncharacterized LOC105377527', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169059361, end_position_on_the_genomic_accession=169072194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377528 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377528, CurrentID=0, Status='live', Symbol='LOC105377528', Aliases='', description='uncharacterized LOC105377528', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169201549, end_position_on_the_genomic_accession=169229929, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377529, CurrentID=0, Status='live', Symbol='LOC105377529', Aliases='', description='uncharacterized LOC105377529', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169273995, end_position_on_the_genomic_accession=169331093, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377530, CurrentID=0, Status='live', Symbol='LOC105377530', Aliases='', description='uncharacterized LOC105377530', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=169912547, end_position_on_the_genomic_accession=169941396, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377531, CurrentID=0, Status='live', Symbol='LOC105377531', Aliases='', description='uncharacterized LOC105377531', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170055175, end_position_on_the_genomic_accession=170057423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377532, CurrentID=0, Status='live', Symbol='LOC105377532', Aliases='', description='DNA dC->dU-editing enzyme APOBEC-3G-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170092143, end_position_on_the_genomic_accession=170107095, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377533, CurrentID=0, Status='live', Symbol='LOC105377533', Aliases='', description='uncharacterized LOC105377533', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=170289846, end_position_on_the_genomic_accession=170316537, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377534, CurrentID=0, Status='live', Symbol='LOC105377534', Aliases='', description='uncharacterized LOC105377534', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171276412, end_position_on_the_genomic_accession=171288406, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377535, CurrentID=0, Status='live', Symbol='LOC105377535', Aliases='', description='uncharacterized LOC105377535', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171512354, end_position_on_the_genomic_accession=171537237, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377536, CurrentID=0, Status='live', Symbol='LOC105377536', Aliases='', description='uncharacterized LOC105377536', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=171635985, end_position_on_the_genomic_accession=171656810, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377537, CurrentID=0, Status='live', Symbol='LOC105377537', Aliases='', description='uncharacterized LOC105377537', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=172002981, end_position_on_the_genomic_accession=172031055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377538, CurrentID=0, Status='live', Symbol='LOC105377538', Aliases='', description='uncharacterized LOC105377538', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173322052, end_position_on_the_genomic_accession=173329694, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377539, CurrentID=0, Status='live', Symbol='LOC105377539', Aliases='', description='uncharacterized LOC105377539', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173314749, end_position_on_the_genomic_accession=173318041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377540, CurrentID=0, Status='live', Symbol='LOC105377540', Aliases='', description='uncharacterized LOC105377540', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173363780, end_position_on_the_genomic_accession=173370689, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377541, CurrentID=0, Status='live', Symbol='LOC105377541', Aliases='', description='uncharacterized LOC105377541', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173334745, end_position_on_the_genomic_accession=173357566, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377542, CurrentID=0, Status='live', Symbol='LOC105377542', Aliases='', description='uncharacterized LOC105377542', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173527382, end_position_on_the_genomic_accession=173583168, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929601, CurrentID=0, Status='live', Symbol='LOC101929601', Aliases='', description='putative exonuclease GOR', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85735314, end_position_on_the_genomic_accession=85737341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377544, CurrentID=0, Status='live', Symbol='LOC105377544', Aliases='', description='uncharacterized LOC105377544', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=173953892, end_position_on_the_genomic_accession=173980185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377545, CurrentID=0, Status='live', Symbol='LOC105377545', Aliases='', description='uncharacterized LOC105377545', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174086763, end_position_on_the_genomic_accession=174191903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377546 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377546, CurrentID=0, Status='live', Symbol='LOC105377546', Aliases='', description='uncharacterized LOC105377546', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174005836, end_position_on_the_genomic_accession=174018083, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  126731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126731, CurrentID=0, Status='live', Symbol='CCSAP', Aliases='C1orf96, CSAP', description='centriole, cilia and spindle-associated protein', other_designations='centriole and spindle-associated protein', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=229321005, end_position_on_the_genomic_accession=229342941, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377548, CurrentID=0, Status='live', Symbol='LOC105377548', Aliases='', description='uncharacterized LOC105377548', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174536716, end_position_on_the_genomic_accession=174541049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377549, CurrentID=0, Status='live', Symbol='LOC105377549', Aliases='', description='uncharacterized LOC105377549', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174610106, end_position_on_the_genomic_accession=174612239, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377550, CurrentID=0, Status='live', Symbol='LOC105377550', Aliases='', description='uncharacterized LOC105377550', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174613625, end_position_on_the_genomic_accession=174614497, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377551, CurrentID=0, Status='live', Symbol='LOC105377551', Aliases='', description='uncharacterized LOC105377551', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=174630419, end_position_on_the_genomic_accession=174635436, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377552, CurrentID=0, Status='live', Symbol='LOC105377552', Aliases='', description='uncharacterized LOC105377552', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175013052, end_position_on_the_genomic_accession=175034309, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377553, CurrentID=0, Status='live', Symbol='LOC105377553', Aliases='', description='uncharacterized LOC105377553', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175948364, end_position_on_the_genomic_accession=175951614, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388882, CurrentID=0, Status='live', Symbol='LOC388882', Aliases='', description='uncharacterized LOC388882', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23462082, end_position_on_the_genomic_accession=23486980, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  152195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152195, CurrentID=0, Status='live', Symbol='NUDT16P1', Aliases='NUDT16P', description='nudix hydrolase 16 pseudogene 1', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 16 pseudogene 1|nudix-type motif 16 pseudogene', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131361845, end_position_on_the_genomic_accession=131365122, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377556, CurrentID=0, Status='live', Symbol='LOC105377556', Aliases='', description='uncharacterized LOC105377556', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=176875712, end_position_on_the_genomic_accession=176908072, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  388885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388885, CurrentID=0, Status='live', Symbol='RPS10P29', Aliases='', description='ribosomal protein S10 pseudogene 29', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24269283, end_position_on_the_genomic_accession=24270188, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388886, CurrentID=0, Status='live', Symbol='LRRC75B', Aliases='C22orf36, FAM211B', description='leucine rich repeat containing 75B', other_designations='family with sequence similarity 211, member B|leucine-rich repeat-containing protein C22orf36|leucine-rich repeat-containing protein FAM211B', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24585623, end_position_on_the_genomic_accession=24593104, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377559, CurrentID=0, Status='live', Symbol='LOC105377559', Aliases='', description='uncharacterized LOC105377559', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177686673, end_position_on_the_genomic_accession=177691160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377560, CurrentID=0, Status='live', Symbol='LOC105377560', Aliases='', description='uncharacterized LOC105377560', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=177756383, end_position_on_the_genomic_accession=177758371, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377561, CurrentID=0, Status='live', Symbol='LOC105377561', Aliases='', description='uncharacterized LOC105377561', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=178007857, end_position_on_the_genomic_accession=178038597, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377562, CurrentID=0, Status='live', Symbol='LOC105377562', Aliases='', description='uncharacterized LOC105377562', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=178136775, end_position_on_the_genomic_accession=178163582, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377563, CurrentID=0, Status='live', Symbol='LOC105377563', Aliases='', description='uncharacterized LOC105377563', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=178813040, end_position_on_the_genomic_accession=178817326, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377564, CurrentID=0, Status='live', Symbol='LOC105377564', Aliases='', description='uncharacterized LOC105377564', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=179964859, end_position_on_the_genomic_accession=180029565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377565, CurrentID=0, Status='live', Symbol='LOC105377565', Aliases='', description='uncharacterized LOC105377565', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=180460330, end_position_on_the_genomic_accession=180532829, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377566, CurrentID=0, Status='live', Symbol='LOC105377566', Aliases='', description='uncharacterized LOC105377566', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=180095500, end_position_on_the_genomic_accession=180117090, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377567, CurrentID=0, Status='live', Symbol='LOC105377567', Aliases='', description='uncharacterized LOC105377567', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=180058957, end_position_on_the_genomic_accession=180554846, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377568, CurrentID=0, Status='live', Symbol='LOC105377568', Aliases='', description='uncharacterized LOC105377568', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=180905496, end_position_on_the_genomic_accession=180929240, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377569, CurrentID=0, Status='live', Symbol='LOC105377569', Aliases='', description='uncharacterized LOC105377569', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181036447, end_position_on_the_genomic_accession=181060566, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377570 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377570, CurrentID=0, Status='live', Symbol='LOC105377570', Aliases='', description='uncharacterized LOC105377570', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182793924, end_position_on_the_genomic_accession=182799582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  126755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126755, CurrentID=0, Status='live', Symbol='LRRC38', Aliases='', description='leucine rich repeat containing 38', other_designations='BK channel auxiliary gamma subunit LRRC38|BK channel auxilliary gamma subunit LRRC38', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13474977, end_position_on_the_genomic_accession=13513781, orientation='minus', exon_count=4, OMIM=615212, no_hdr0=''),
  105377572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377572, CurrentID=0, Status='live', Symbol='LOC105377572', Aliases='', description='uncharacterized LOC105377572', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182204528, end_position_on_the_genomic_accession=182216815, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377573 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377573, CurrentID=0, Status='live', Symbol='LOC105377573', Aliases='', description='uncharacterized LOC105377573', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181820005, end_position_on_the_genomic_accession=181821630, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377574, CurrentID=0, Status='live', Symbol='LOC105377574', Aliases='', description='uncharacterized LOC105377574', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=181447619, end_position_on_the_genomic_accession=181867837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377575, CurrentID=0, Status='live', Symbol='LOC105377575', Aliases='', description='uncharacterized LOC105377575', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182826353, end_position_on_the_genomic_accession=182832727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377576, CurrentID=0, Status='live', Symbol='LOC105377576', Aliases='', description='uncharacterized LOC105377576', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182871595, end_position_on_the_genomic_accession=182881167, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377577, CurrentID=0, Status='live', Symbol='LOC105377577', Aliases='', description='uncharacterized LOC105377577', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=182937523, end_position_on_the_genomic_accession=182943279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388906, CurrentID=0, Status='live', Symbol='OGFRP1', Aliases='', description='opioid growth factor receptor pseudogene 1', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42269753, end_position_on_the_genomic_accession=42274862, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388907, CurrentID=0, Status='live', Symbol='RPL5P34', Aliases='RPL5_15_1754', description='ribosomal protein L5 pseudogene 34', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=42776345, end_position_on_the_genomic_accession=42777370, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377580, CurrentID=0, Status='live', Symbol='LOC105377580', Aliases='', description='uncharacterized LOC105377580', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183558584, end_position_on_the_genomic_accession=183567306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  651053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651053, CurrentID=0, Status='live', Symbol='DDX18P2', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18 pseudogene 2', other_designations='', map_location='15q22.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=59871343, end_position_on_the_genomic_accession=59873553, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  388910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388910, CurrentID=0, Status='live', Symbol='LINC00207', Aliases='NCRNA00207', description='long intergenic non-protein coding RNA 207', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=44569340, end_position_on_the_genomic_accession=44572449, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126767, CurrentID=0, Status='live', Symbol='AADACL3', Aliases='', description='arylacetamide deacetylase-like 3', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12716115, end_position_on_the_genomic_accession=12728759, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377584, CurrentID=0, Status='live', Symbol='LOC105377584', Aliases='', description='uncharacterized LOC105377584', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184265036, end_position_on_the_genomic_accession=184274264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377585, CurrentID=0, Status='live', Symbol='LOC105377585', Aliases='', description='uncharacterized LOC105377585', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184506460, end_position_on_the_genomic_accession=184538536, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377586, CurrentID=0, Status='live', Symbol='LVCAT8', Aliases='', description='liver cancer-associated transcript 8', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184618616, end_position_on_the_genomic_accession=184625697, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377587, CurrentID=0, Status='live', Symbol='LOC105377587', Aliases='', description='uncharacterized LOC105377587', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184826230, end_position_on_the_genomic_accession=184844206, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377588, CurrentID=0, Status='live', Symbol='LOC105377588', Aliases='', description='uncharacterized LOC105377588', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185081109, end_position_on_the_genomic_accession=185086206, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379774, CurrentID=0, Status='live', Symbol='LOC105379774', Aliases='', description='uncharacterized LOC105379774', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377590, CurrentID=0, Status='live', Symbol='LOC105377590', Aliases='', description='uncharacterized LOC105377590', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185471674, end_position_on_the_genomic_accession=185472263, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377591, CurrentID=0, Status='live', Symbol='LOC105377591', Aliases='', description='uncharacterized LOC105377591', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185370725, end_position_on_the_genomic_accession=185392840, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377592, CurrentID=0, Status='live', Symbol='LOC105377592', Aliases='', description='uncharacterized LOC105377592', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186015183, end_position_on_the_genomic_accession=186018644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100134713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134713, CurrentID=0, Status='live', Symbol='NDUFB2-AS1', Aliases='', description='NDUFB2 antisense RNA 1', other_designations='NDUFB2 antisense RNA 1 (non-protein coding)', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=140695336, end_position_on_the_genomic_accession=140697077, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377594, CurrentID=0, Status='live', Symbol='LOC105377594', Aliases='', description='protein FAM149A-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186104788, end_position_on_the_genomic_accession=186124062, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377595, CurrentID=0, Status='live', Symbol='RPSAP70', Aliases='', description='ribosomal protein SA pseudogene 70', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186158229, end_position_on_the_genomic_accession=186158843, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377596, CurrentID=0, Status='live', Symbol='LOC105377596', Aliases='', description='uncharacterized LOC105377596', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186501069, end_position_on_the_genomic_accession=186545752, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388925, CurrentID=0, Status='live', Symbol='PSMC1P10', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 10', other_designations='', map_location='2p24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=17385072, end_position_on_the_genomic_accession=17386612, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377598, CurrentID=0, Status='live', Symbol='LOC105377598', Aliases='', description='uncharacterized LOC105377598', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186921167, end_position_on_the_genomic_accession=186930893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392501, CurrentID=0, Status='live', Symbol='RPL26P36', Aliases='RPL26_16_1792', description='ribosomal protein L26 pseudogene 36', other_designations='', map_location='Xq21.32', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=92676476, end_position_on_the_genomic_accession=92677052, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377600, CurrentID=0, Status='live', Symbol='LOC105377600', Aliases='', description='uncharacterized LOC105377600', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187201980, end_position_on_the_genomic_accession=187214091, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377601, CurrentID=0, Status='live', Symbol='LOC105377601', Aliases='', description='uncharacterized LOC105377601', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187421482, end_position_on_the_genomic_accession=187437691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377602, CurrentID=0, Status='live', Symbol='LOC105377602', Aliases='', description='uncharacterized LOC105377602', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187370076, end_position_on_the_genomic_accession=187372098, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388931, CurrentID=0, Status='live', Symbol='MFSD2B', Aliases='', description='major facilitator superfamily domain containing 2B', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24010083, end_position_on_the_genomic_accession=24025874, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105377604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377604, CurrentID=0, Status='live', Symbol='LOC105377604', Aliases='', description='uncharacterized LOC105377604', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=187555620, end_position_on_the_genomic_accession=187646425, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  126789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126789, CurrentID=0, Status='live', Symbol='PUSL1', Aliases='', description='pseudouridylate synthase-like 1', other_designations='tRNA pseudouridylate synthase-like 1|tRNA-uridine isomerase-like 1', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1308571, end_position_on_the_genomic_accession=1311677, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377606, CurrentID=0, Status='live', Symbol='LOC105377606', Aliases='', description='uncharacterized LOC105377606', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=188159217, end_position_on_the_genomic_accession=188161142, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100527943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527943, CurrentID=0, Status='live', Symbol='TGIF2-C20orf24', Aliases='', description='TGIF2-C20orf24 readthrough', other_designations='', map_location='20q', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36574554, end_position_on_the_genomic_accession=36612557, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126792, CurrentID=0, Status='live', Symbol='B3GALT6', Aliases='EDSP2, SEMDJL1, beta3GalT6', description='Beta-1,3-galactosyltransferase 6', other_designations='GAG GalTII|UDP-Gal:betaGal beta 1,3-galactosyltransferase 6|UDP-Gal:betaGal beta 1,3-galactosyltransferase polypeptide 6|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 6|beta-1,3-GalTase 6|beta3Gal-T6|galactosyltransferase II|galactosylxylosylprotein 3-beta-galactosyltransferase', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1232249, end_position_on_the_genomic_accession=1235041, orientation='plus', exon_count=1, OMIM=615291, no_hdr0=''),
  100527945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527945, CurrentID=0, Status='live', Symbol='NDUFA12P1', Aliases='', description='NADH:ubiquinone oxidoreductase subunit A12 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 12 pseudogene 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100027769, end_position_on_the_genomic_accession=100028174, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100527946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527946, CurrentID=0, Status='live', Symbol='PDCD5P1', Aliases='', description='programmed cell death 5 pseudogene 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=19413152, end_position_on_the_genomic_accession=19413588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  388939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388939, CurrentID=0, Status='live', Symbol='C2orf71', Aliases='RP54', description='chromosome 2 open reading frame 71', other_designations='', map_location='2p23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=29061373, end_position_on_the_genomic_accession=29074838, orientation='minus', exon_count=4, OMIM=613425, no_hdr0=''),
  100527948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527948, CurrentID=0, Status='live', Symbol='HMGB1P27', Aliases='', description='high mobility group box 1 pseudogene 27', other_designations='high-mobility group box 1 pseudogene 27', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=191174233, end_position_on_the_genomic_accession=191174835, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100527949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527949, CurrentID=0, Status='live', Symbol='GIMAP1-GIMAP5', Aliases='', description='GIMAP1-GIMAP5 readthrough', other_designations='', map_location='7q', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150716557, end_position_on_the_genomic_accession=150743649, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100527950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527950, CurrentID=0, Status='live', Symbol='GJA9-MYCBP', Aliases='', description='GJA9-MYCBP readthrough', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38862490, end_position_on_the_genomic_accession=38881626, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377615, CurrentID=0, Status='live', Symbol='LOC105377615', Aliases='', description='uncharacterized LOC105377615', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189550782, end_position_on_the_genomic_accession=189551957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377616, CurrentID=0, Status='live', Symbol='LOC105377616', Aliases='', description='uncharacterized LOC105377616', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189714056, end_position_on_the_genomic_accession=189722576, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377617, CurrentID=0, Status='live', Symbol='LINC01596', Aliases='', description='long intergenic non-protein coding RNA 1596', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189881692, end_position_on_the_genomic_accession=189888086, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388946, CurrentID=0, Status='live', Symbol='TMEM247', Aliases='', description='transmembrane protein 247', other_designations='transmembrane protein ENSP00000343375', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46479497, end_position_on_the_genomic_accession=46484579, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377619, CurrentID=0, Status='live', Symbol='LOC105377619', Aliases='', description='uncharacterized LOC105377619', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=189817574, end_position_on_the_genomic_accession=189820145, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  388948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388948, CurrentID=0, Status='live', Symbol='LINC01118', Aliases='', description='long intergenic non-protein coding RNA 1118', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=46816668, end_position_on_the_genomic_accession=46822660, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377621, CurrentID=0, Status='live', Symbol='LOC105377621', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=102418707, end_position_on_the_genomic_accession=102501891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377622, CurrentID=0, Status='live', Symbol='LOC105377622', Aliases='', description='uncharacterized LOC105377622', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=139600624, end_position_on_the_genomic_accession=139608579, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  388951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388951, CurrentID=0, Status='live', Symbol='TSPYL6', Aliases='', description='TSPY-like 6', other_designations='TSPY-like protein 6', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54253178, end_position_on_the_genomic_accession=54256272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100527960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527960, CurrentID=0, Status='live', Symbol='MROH7-TTC4', Aliases='HEATR8-TTC4', description='MROH7-TTC4 readthrough (NMD candidate)', other_designations='HEATR8-TTC4 readthrough', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54641740, end_position_on_the_genomic_accession=54742657, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  105377625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377625, CurrentID=0, Status='live', Symbol='LOC105377625', Aliases='', description='uncharacterized LOC105377625', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=22198763, end_position_on_the_genomic_accession=22199227, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377626, CurrentID=0, Status='live', Symbol='LOC105377626', Aliases='', description='uncharacterized LOC105377626', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24832080, end_position_on_the_genomic_accession=24834476, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100527963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527963, CurrentID=0, Status='live', Symbol='PMF1-BGLAP', Aliases='PMF1', description='PMF1-BGLAP readthrough', other_designations='polyamine-modulated factor 1', map_location='1q', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156212988, end_position_on_the_genomic_accession=156243332, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100527964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527964, CurrentID=0, Status='live', Symbol='LHX4-AS1', Aliases='', description='LHX4 antisense RNA 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=180269663, end_position_on_the_genomic_accession=180274681, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377629, CurrentID=0, Status='live', Symbol='LOC105377629', Aliases='', description='uncharacterized LOC105377629', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57706583, end_position_on_the_genomic_accession=57711600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377630, CurrentID=0, Status='live', Symbol='LOC105377630', Aliases='', description='uncharacterized LOC105377630', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57694907, end_position_on_the_genomic_accession=57705174, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377631, CurrentID=0, Status='live', Symbol='LOC105377631', Aliases='', description='uncharacterized LOC105377631', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=57536146, end_position_on_the_genomic_accession=57551728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388960, CurrentID=0, Status='live', Symbol='C2orf78', Aliases='COG5373, hCG1989538', description='chromosome 2 open reading frame 78', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73784189, end_position_on_the_genomic_accession=73817148, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377633, CurrentID=0, Status='live', Symbol='LOC105377633', Aliases='', description='uncharacterized LOC105377633', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61471393, end_position_on_the_genomic_accession=61471845, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  388962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388962, CurrentID=0, Status='live', Symbol='BOLA3', Aliases='MMDS2', description='bolA family member 3', other_designations='bolA homolog 3', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74135401, end_position_on_the_genomic_accession=74147912, orientation='minus', exon_count=4, OMIM=613183, no_hdr0=''),
  388963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388963, CurrentID=0, Status='live', Symbol='C2orf81', Aliases='hCG40743', description='chromosome 2 open reading frame 81', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74414176, end_position_on_the_genomic_accession=74417717, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  126820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126820, CurrentID=0, Status='live', Symbol='WDR63', Aliases='DIC3, NYD-SP29', description='WD repeat domain 63', other_designations='testis development protein NYD-SP29 (NYD-SP29)', map_location='1p22.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85062298, end_position_on_the_genomic_accession=85133138, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  388965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388965, CurrentID=0, Status='live', Symbol='FUNDC2P2', Aliases='', description='FUN14 domain containing 2 pseudogene 2', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84290682, end_position_on_the_genomic_accession=84292200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377638, CurrentID=0, Status='live', Symbol='LOC105377638', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=87406253, end_position_on_the_genomic_accession=87411116, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126823, CurrentID=0, Status='live', Symbol='KLHDC9', Aliases='KARCA1', description='kelch domain containing 9', other_designations='kelch/ankyrin repeat-containing cyclin A1-interacting protein', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161098361, end_position_on_the_genomic_accession=161100348, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377640, CurrentID=0, Status='live', Symbol='LOC105377640', Aliases='', description='uncharacterized LOC105377640', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=4162469, end_position_on_the_genomic_accession=4252096, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  388969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=388969, CurrentID=0, Status='live', Symbol='C2orf68', Aliases='HCRCN81', description='chromosome 2 open reading frame 68', other_designations='', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85605253, end_position_on_the_genomic_accession=85612056, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100527978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100527978, CurrentID=0, Status='live', Symbol='TMEM56-RWDD3', Aliases='', description='TMEM56-RWDD3 readthrough', other_designations='', map_location='1p', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95117923, end_position_on_the_genomic_accession=95244955, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377643, CurrentID=0, Status='live', Symbol='LOC105377643', Aliases='', description='uncharacterized LOC105377643', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=38950448, end_position_on_the_genomic_accession=39052002, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105377644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377644, CurrentID=0, Status='live', Symbol='LOC105377644', Aliases='', description='uncharacterized LOC105377644', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=39342852, end_position_on_the_genomic_accession=39383223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377645, CurrentID=0, Status='live', Symbol='LOC105377645', Aliases='', description='uncharacterized LOC105377645', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40184378, end_position_on_the_genomic_accession=40200230, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377646, CurrentID=0, Status='live', Symbol='LOC105377646', Aliases='', description='uncharacterized LOC105377646', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52829286, end_position_on_the_genomic_accession=52830660, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377647, CurrentID=0, Status='live', Symbol='LOC105377647', Aliases='', description='uncharacterized LOC105377647', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62374234, end_position_on_the_genomic_accession=62375799, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377648, CurrentID=0, Status='live', Symbol='LOC105377648', Aliases='', description='uncharacterized LOC105377648', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=72561174, end_position_on_the_genomic_accession=72561819, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377649, CurrentID=0, Status='live', Symbol='LOC105377649', Aliases='', description='uncharacterized LOC105377649', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=80760909, end_position_on_the_genomic_accession=80761490, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377650, CurrentID=0, Status='live', Symbol='LOC105377650', Aliases='', description='uncharacterized LOC105377650', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=58171777, end_position_on_the_genomic_accession=58181840, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377651, CurrentID=0, Status='live', Symbol='LOC105377651', Aliases='', description='uncharacterized LOC105377651', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=32005858, end_position_on_the_genomic_accession=32022484, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377652, CurrentID=0, Status='live', Symbol='LOC105377652', Aliases='', description='uncharacterized LOC105377652', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52044803, end_position_on_the_genomic_accession=52046924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377653, CurrentID=0, Status='live', Symbol='LOC105377653', Aliases='', description='uncharacterized LOC105377653', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53038956, end_position_on_the_genomic_accession=53051191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377654, CurrentID=0, Status='live', Symbol='LOC105377654', Aliases='', description='uncharacterized LOC105377654', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53375303, end_position_on_the_genomic_accession=53377646, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377655, CurrentID=0, Status='live', Symbol='LOC105377655', Aliases='', description='uncharacterized LOC105377655', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=53908012, end_position_on_the_genomic_accession=53917583, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377656, CurrentID=0, Status='live', Symbol='LOC105377656', Aliases='', description='uncharacterized LOC105377656', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54332873, end_position_on_the_genomic_accession=54356104, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377657, CurrentID=0, Status='live', Symbol='LOC105377657', Aliases='', description='uncharacterized LOC105377657', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54756430, end_position_on_the_genomic_accession=54818455, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377658, CurrentID=0, Status='live', Symbol='LOC105377658', Aliases='', description='uncharacterized LOC105377658', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=54836077, end_position_on_the_genomic_accession=54845413, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377659, CurrentID=0, Status='live', Symbol='LOC105377659', Aliases='', description='uncharacterized LOC105377659', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55378392, end_position_on_the_genomic_accession=55386432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377660, CurrentID=0, Status='live', Symbol='LOC105377660', Aliases='', description='uncharacterized LOC105377660', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55656472, end_position_on_the_genomic_accession=55664655, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377661, CurrentID=0, Status='live', Symbol='LOC105377661', Aliases='', description='uncharacterized LOC105377661', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55821053, end_position_on_the_genomic_accession=55825012, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377662, CurrentID=0, Status='live', Symbol='LOC105377662', Aliases='', description='uncharacterized LOC105377662', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55951393, end_position_on_the_genomic_accession=55974326, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377663, CurrentID=0, Status='live', Symbol='LOC105377663', Aliases='', description='uncharacterized LOC105377663', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=55890393, end_position_on_the_genomic_accession=55948732, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377664, CurrentID=0, Status='live', Symbol='LOC105377664', Aliases='', description='uncharacterized LOC105377664', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56305856, end_position_on_the_genomic_accession=56308284, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377665, CurrentID=0, Status='live', Symbol='LOC105377665', Aliases='', description='uncharacterized LOC105377665', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56523346, end_position_on_the_genomic_accession=56530486, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377666, CurrentID=0, Status='live', Symbol='LOC105377666', Aliases='', description='uncharacterized LOC105377666', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56683190, end_position_on_the_genomic_accession=56696367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377667, CurrentID=0, Status='live', Symbol='LOC105377667', Aliases='', description='uncharacterized LOC105377667', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56697217, end_position_on_the_genomic_accession=56704466, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377668, CurrentID=0, Status='live', Symbol='LOC105377668', Aliases='', description='uncharacterized LOC105377668', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56794901, end_position_on_the_genomic_accession=56800897, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377669, CurrentID=0, Status='live', Symbol='LOC105377669', Aliases='', description='uncharacterized LOC105377669', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=57008430, end_position_on_the_genomic_accession=57010236, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100528006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528006, CurrentID=0, Status='live', Symbol='DPPA5P3', Aliases='', description='developmental pluripotency associated 5 pseudogene 3', other_designations='', map_location='10p15', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=6329134, end_position_on_the_genomic_accession=6329477, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100528007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528007, CurrentID=0, Status='live', Symbol='BORCS7-ASMT', Aliases='C10orf32-AS3MT, C10orf32-ASMT', description='BORCS7-ASMT readthrough (NMD candidate)', other_designations='', map_location='10q', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102854210, end_position_on_the_genomic_accession=102901899, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  651144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651144, CurrentID=0, Status='live', Symbol='NACAP2', Aliases='', description='nascent polypeptide associated complex alpha subunit pseudogene 2', other_designations='nascent polypeptide-associated complex alpha subunit pseudogene|nascent-polypeptide-associated complex alpha polypeptide pseudogene 2', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119996996, end_position_on_the_genomic_accession=119997748, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377673, CurrentID=0, Status='live', Symbol='LOC105377673', Aliases='', description='uncharacterized LOC105377673', other_designations='CTC-340I23.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156703578, end_position_on_the_genomic_accession=156742880, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377674, CurrentID=0, Status='live', Symbol='LOC105377674', Aliases='', description='uncharacterized LOC105377674', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=156290511, end_position_on_the_genomic_accession=156376289, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  126859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126859, CurrentID=0, Status='live', Symbol='AXDND1', Aliases='C1orf125', description='axonemal dynein light chain domain containing 1', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=179365718, end_position_on_the_genomic_accession=179554735, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  105377676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377676, CurrentID=0, Status='live', Symbol='LOC105377676', Aliases='', description='uncharacterized LOC105377676', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=157696216, end_position_on_the_genomic_accession=157711208, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377677, CurrentID=0, Status='live', Symbol='LOC105377677', Aliases='', description='uncharacterized LOC105377677', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158045316, end_position_on_the_genomic_accession=158051280, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377678, CurrentID=0, Status='live', Symbol='LOC105377678', Aliases='', description='uncharacterized LOC105377678', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158226021, end_position_on_the_genomic_accession=158251077, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377679, CurrentID=0, Status='live', Symbol='LOC105377679', Aliases='', description='uncharacterized LOC105377679', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=158387401, end_position_on_the_genomic_accession=158396126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100528016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528016, CurrentID=0, Status='live', Symbol='TMX2-CTNND1', Aliases='', description='TMX2-CTNND1 readthrough (NMD candidate)', other_designations='TMX2-CTNND1 readthrough (non-protein coding)', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57712523, end_position_on_the_genomic_accession=57819180, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100528017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528017, CurrentID=0, Status='live', Symbol='SAA2-SAA4', Aliases='', description='SAA2-SAA4 readthrough', other_designations='SAA2-SAA2 read-through transcript', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=18231355, end_position_on_the_genomic_accession=18248674, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100528018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528018, CurrentID=0, Status='live', Symbol='ARL2-SNX15', Aliases='', description='ARL2-SNX15 readthrough (NMD candidate)', other_designations='', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65014113, end_position_on_the_genomic_accession=65040572, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100528019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528019, CurrentID=0, Status='live', Symbol='HSPB2-C11orf52', Aliases='', description='HSPB2-C11orf52 readthrough (NMD candidate)', other_designations='HSPB2-C11orf52 readthrough (non-protein coding)', map_location='11q', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=111912736, end_position_on_the_genomic_accession=111926871, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  126868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126868, CurrentID=0, Status='live', Symbol='MAB21L3', Aliases='C1orf161', description='mab-21-like 3 (C. elegans)', other_designations='', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116075792, end_position_on_the_genomic_accession=116135240, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100528021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528021, CurrentID=0, Status='live', Symbol='ST20-MTHFS', Aliases='', description='ST20-MTHFS readthrough', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79843547, end_position_on_the_genomic_accession=79923106, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100528022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528022, CurrentID=0, Status='live', Symbol='BLOC1S1-RDH5', Aliases='', description='BLOC1S1-RDH5 readthrough', other_designations='', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55716034, end_position_on_the_genomic_accession=55724742, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  389015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389015, CurrentID=0, Status='live', Symbol='SLC9A4', Aliases='NHE4', description='solute carrier family 9, subfamily A (NHE4, cation proton antiporter 4), member 4', other_designations='NHE-4|Na(+)/H(+) exchanger 4|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 (sodium/hydrogen exchanger), member 4|solute carrier family 9 member 4', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102473282, end_position_on_the_genomic_accession=102533974, orientation='plus', exon_count=12, OMIM=600531, no_hdr0=''),
  100528024 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528024, CurrentID=0, Status='live', Symbol='DEL11P13', Aliases='C11DELp13, WAGR', description='Wilms tumor, aniridia, genitourinary anomalies and mental retardation syndrome', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=194072, no_hdr0=''),
  100528025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528025, CurrentID=0, Status='live', Symbol='SXGQTL1', Aliases='', description='Sex hormone-binding blobulin circulating level QTL 1', other_designations='', map_location='17p13', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613498, no_hdr0=''),
  100528026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528026, CurrentID=0, Status='live', Symbol='DEL4Q21', Aliases='C4DELq21', description='Chromosome 4q21 deletion syndrome', other_designations='', map_location='4q21', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613509, no_hdr0=''),
  100528027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528027, CurrentID=0, Status='live', Symbol='RJBS', Aliases='', description='Rajab syndrome', other_designations='', map_location='2q36.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613658, no_hdr0=''),
  105377692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377692, CurrentID=0, Status='live', Symbol='LOC105377692', Aliases='', description='uncharacterized LOC105377692', other_designations='CTC-529G1.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=160685310, end_position_on_the_genomic_accession=160692005, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377693, CurrentID=0, Status='live', Symbol='LOC105377693', Aliases='', description='uncharacterized LOC105377693', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=161059772, end_position_on_the_genomic_accession=161085376, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100528030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528030, CurrentID=0, Status='live', Symbol='POC1B-GALNT4', Aliases='', description='POC1B-GALNT4 readthrough', other_designations='POC1B-GALNT4 fusion', map_location='12q', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89519412, end_position_on_the_genomic_accession=89526262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389023, CurrentID=0, Status='live', Symbol='DPP10-AS1', Aliases='', description='DPP10 antisense RNA 1', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=115144048, end_position_on_the_genomic_accession=115161343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100528032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528032, CurrentID=0, Status='live', Symbol='KLRC4-KLRK1', Aliases='', description='KLRC4-KLRK1 readthrough', other_designations='NK cell receptor D|NKG2-D-activating NK receptor|killer cell lectin-like receptor subfamily K member 1', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10372353, end_position_on_the_genomic_accession=10410146, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105377697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377697, CurrentID=0, Status='live', Symbol='LOC105377697', Aliases='', description='uncharacterized LOC105377697', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=162534936, end_position_on_the_genomic_accession=162558535, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377698, CurrentID=0, Status='live', Symbol='LOC105377698', Aliases='', description='uncharacterized LOC105377698', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=162623348, end_position_on_the_genomic_accession=162624440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101929627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929627, CurrentID=0, Status='live', Symbol='LOC101929627', Aliases='', description='putative exonuclease GOR', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85744489, end_position_on_the_genomic_accession=85746516, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377700, CurrentID=0, Status='live', Symbol='LOC105377700', Aliases='', description='uncharacterized LOC105377700', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=163397143, end_position_on_the_genomic_accession=163437228, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377701, CurrentID=0, Status='live', Symbol='LOC105377701', Aliases='', description='uncharacterized LOC105377701', other_designations='CTB-181F24.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=165219920, end_position_on_the_genomic_accession=165243358, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105377702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377702, CurrentID=0, Status='live', Symbol='LOC105377702', Aliases='', description='uncharacterized LOC105377702', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=165140173, end_position_on_the_genomic_accession=165160126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377703, CurrentID=0, Status='live', Symbol='LOC105377703', Aliases='', description='uncharacterized LOC105377703', other_designations='CTC-340A15.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=164973249, end_position_on_the_genomic_accession=165171593, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377704, CurrentID=0, Status='live', Symbol='LOC105377704', Aliases='', description='uncharacterized LOC105377704', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=165843235, end_position_on_the_genomic_accession=165847804, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389033, CurrentID=0, Status='live', Symbol='LOC389033', Aliases='', description='placenta specific 9 pseudogene', other_designations='placenta-specific 9 pseudogene', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=129922862, end_position_on_the_genomic_accession=129934317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377706, CurrentID=0, Status='live', Symbol='LOC105377706', Aliases='', description='uncharacterized LOC105377706', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=166845621, end_position_on_the_genomic_accession=166895725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377707, CurrentID=0, Status='live', Symbol='LOC105377707', Aliases='', description='uncharacterized LOC105377707', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168226523, end_position_on_the_genomic_accession=168227960, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389036, CurrentID=0, Status='live', Symbol='ACT', Aliases='', description='actin-like protein (ACT) gene', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377709, CurrentID=0, Status='live', Symbol='LOC105377709', Aliases='', description='uncharacterized LOC105377709', other_designations='CTC-353G13.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=167937077, end_position_on_the_genomic_accession=167953560, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377710, CurrentID=0, Status='live', Symbol='LOC105377710', Aliases='', description='uncharacterized LOC105377710', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=167411155, end_position_on_the_genomic_accession=167418869, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377711, CurrentID=0, Status='live', Symbol='LOC105377711', Aliases='', description='uncharacterized LOC105377711', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168350144, end_position_on_the_genomic_accession=168352132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377712, CurrentID=0, Status='live', Symbol='LOC105377712', Aliases='', description='uncharacterized LOC105377712', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169079896, end_position_on_the_genomic_accession=169091143, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377713, CurrentID=0, Status='live', Symbol='LOC105377713', Aliases='', description='uncharacterized LOC105377713', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168886582, end_position_on_the_genomic_accession=168892251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377714, CurrentID=0, Status='live', Symbol='LOC105377714', Aliases='', description='uncharacterized LOC105377714', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=169458436, end_position_on_the_genomic_accession=169583608, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  389043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389043, CurrentID=0, Status='live', Symbol='LINC01120', Aliases='', description='long intergenic non-protein coding RNA 1120', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131402901, end_position_on_the_genomic_accession=131408228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377716 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377716, CurrentID=0, Status='live', Symbol='LOC105377716', Aliases='', description='uncharacterized LOC105377716', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170642233, end_position_on_the_genomic_accession=170681437, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377717, CurrentID=0, Status='live', Symbol='LOC105377717', Aliases='', description='uncharacterized LOC105377717', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171139432, end_position_on_the_genomic_accession=171143752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377718, CurrentID=0, Status='live', Symbol='LOC105377718', Aliases='', description='uncharacterized LOC105377718', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=170911170, end_position_on_the_genomic_accession=170936355, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377719, CurrentID=0, Status='live', Symbol='LOC105377719', Aliases='', description='uncharacterized LOC105377719', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171443609, end_position_on_the_genomic_accession=171451889, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377720, CurrentID=0, Status='live', Symbol='LOC105377720', Aliases='', description='uncharacterized LOC105377720', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171494922, end_position_on_the_genomic_accession=171502737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377721, CurrentID=0, Status='live', Symbol='LOC105377721', Aliases='', description='uncharacterized LOC105377721', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171469835, end_position_on_the_genomic_accession=171495642, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377722, CurrentID=0, Status='live', Symbol='LOC105377722', Aliases='', description='uncharacterized LOC105377722', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171566903, end_position_on_the_genomic_accession=171582890, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377723, CurrentID=0, Status='live', Symbol='LOC105377723', Aliases='', description='uncharacterized LOC105377723', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171738385, end_position_on_the_genomic_accession=171742789, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377724, CurrentID=0, Status='live', Symbol='LOC105377724', Aliases='', description='uncharacterized LOC105377724', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=171718777, end_position_on_the_genomic_accession=171742482, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  389053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389053, CurrentID=0, Status='live', Symbol='HNRNPKP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein K pseudogene 2', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=136199117, end_position_on_the_genomic_accession=136200503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100528062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528062, CurrentID=0, Status='live', Symbol='ARMCX5-GPRASP2', Aliases='', description='ARMCX5-GPRASP2 readthrough', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=102599348, end_position_on_the_genomic_accession=102717733, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377727, CurrentID=0, Status='live', Symbol='LOC105377727', Aliases='', description='uncharacterized LOC105377727', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172479229, end_position_on_the_genomic_accession=172535787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100528064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100528064, CurrentID=0, Status='live', Symbol='NEDD8-MDP1', Aliases='', description='NEDD8-MDP1 readthrough', other_designations='NEDD8-MDP1 read-through transcript', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=24213937, end_position_on_the_genomic_accession=24232370, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377729, CurrentID=0, Status='live', Symbol='LOC105377729', Aliases='', description='uncharacterized LOC105377729', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172605446, end_position_on_the_genomic_accession=172616842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389058, CurrentID=0, Status='live', Symbol='SP5', Aliases='', description='Sp5 transcription factor', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170714738, end_position_on_the_genomic_accession=170717988, orientation='plus', exon_count=2, OMIM=609391, no_hdr0=''),
  105377731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377731, CurrentID=0, Status='live', Symbol='LOC105377731', Aliases='', description='uncharacterized LOC105377731', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173269431, end_position_on_the_genomic_accession=173280929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377732, CurrentID=0, Status='live', Symbol='LOC105377732', Aliases='', description='uncharacterized LOC105377732', other_designations='CTB-164N12.1|CTB-32H22.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173385052, end_position_on_the_genomic_accession=173524433, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  126917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126917, CurrentID=0, Status='live', Symbol='IFFO2', Aliases='', description='intermediate filament family orphan 2', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18904280, end_position_on_the_genomic_accession=18956667, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377734, CurrentID=0, Status='live', Symbol='LOC105377734', Aliases='', description='uncharacterized LOC105377734', other_designations='CTB-33O18.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173561470, end_position_on_the_genomic_accession=173574123, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422305, CurrentID=0, Status='live', Symbol='PSMC1P13', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 13', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=69893856, end_position_on_the_genomic_accession=69895391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377736, CurrentID=0, Status='live', Symbol='LOC105377736', Aliases='', description='uncharacterized LOC105377736', other_designations='CTB-43E15.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173689217, end_position_on_the_genomic_accession=173708197, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377737, CurrentID=0, Status='live', Symbol='LOC105377737', Aliases='', description='uncharacterized LOC105377737', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174237915, end_position_on_the_genomic_accession=174243501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377738, CurrentID=0, Status='live', Symbol='LOC105377738', Aliases='', description='uncharacterized LOC105377738', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174530845, end_position_on_the_genomic_accession=174532457, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377739, CurrentID=0, Status='live', Symbol='LOC105377739', Aliases='', description='uncharacterized LOC105377739', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174522566, end_position_on_the_genomic_accession=174548113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377740, CurrentID=0, Status='live', Symbol='LOC105377740', Aliases='', description='uncharacterized LOC105377740', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174615581, end_position_on_the_genomic_accession=174621838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377741, CurrentID=0, Status='live', Symbol='LOC105377741', Aliases='', description='uncharacterized LOC105377741', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174721321, end_position_on_the_genomic_accession=174724123, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377742, CurrentID=0, Status='live', Symbol='LOC105377742', Aliases='', description='uncharacterized LOC105377742', other_designations='CTD-2532K18.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=174804970, end_position_on_the_genomic_accession=174826714, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377743, CurrentID=0, Status='live', Symbol='LOC105377743', Aliases='', description='uncharacterized LOC105377743', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=175691855, end_position_on_the_genomic_accession=175697196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389072, CurrentID=0, Status='live', Symbol='PLEKHM3', Aliases='DAPR, PLEKHM1L', description='pleckstrin homology domain containing M3', other_designations='PH domain-containing family M member 3|differentiation associated protein|pleckstrin homology domain containing, family M, member 3', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=207821288, end_position_on_the_genomic_accession=208040372, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  389073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389073, CurrentID=0, Status='live', Symbol='C2orf80', Aliases='GONDA1', description='chromosome 2 open reading frame 80', other_designations='gonad development associated 1', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208165347, end_position_on_the_genomic_accession=208190258, orientation='minus', exon_count=10, OMIM=615536, no_hdr0=''),
  105377746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377746, CurrentID=0, Status='live', Symbol='LOC105377746', Aliases='', description='uncharacterized LOC105377746', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=176031325, end_position_on_the_genomic_accession=176035274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389075, CurrentID=0, Status='live', Symbol='RESP18', Aliases='', description='regulated endocrine specific protein 18', other_designations='regulated endocrine-specific protein 18 homolog', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219327409, end_position_on_the_genomic_accession=219336785, orientation='minus', exon_count=9, OMIM=612721, no_hdr0=''),
  100134868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134868, CurrentID=0, Status='live', Symbol='LOC100134868', Aliases='', description='uncharacterized LOC100134868', other_designations='', map_location='20p11.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=26009799, end_position_on_the_genomic_accession=26021795, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100134869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134869, CurrentID=0, Status='live', Symbol='UBE2Q2P2', Aliases='UBE2Q2P3, UBE2QP2', description='ubiquitin conjugating enzyme E2Q family member 2 pseudogene 2', other_designations='ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 2|ubiquitin-conjugating enzyme E2Q family member 2 pseudogene 3|ubiquitin-conjugating enzyme E2Q family pseudogene 2', map_location='15q25', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82355136, end_position_on_the_genomic_accession=82415611, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377750, CurrentID=0, Status='live', Symbol='LOC105377750', Aliases='', description='uncharacterized LOC105377750', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177554628, end_position_on_the_genomic_accession=177575299, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377751, CurrentID=0, Status='live', Symbol='LOC105377751', Aliases='', description='uncharacterized LOC105377751', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177707966, end_position_on_the_genomic_accession=177722358, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105377752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377752, CurrentID=0, Status='live', Symbol='LOC105377752', Aliases='', description='uncharacterized LOC105377752', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177801111, end_position_on_the_genomic_accession=177807458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377753, CurrentID=0, Status='live', Symbol='LOC105377753', Aliases='', description='uncharacterized LOC105377753', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=177864399, end_position_on_the_genomic_accession=177871742, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=258010, CurrentID=0, Status='live', Symbol='SVIP', Aliases='', description='small VCP/p97-interacting protein', other_designations='', map_location='11p14.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=22822052, end_position_on_the_genomic_accession=22829836, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377755, CurrentID=0, Status='live', Symbol='LOC105377755', Aliases='', description='uncharacterized LOC105377755', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178232837, end_position_on_the_genomic_accession=178237077, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389084, CurrentID=0, Status='live', Symbol='C2orf82', Aliases='UNQ830', description='chromosome 2 open reading frame 82', other_designations='ASCL830', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=232869119, end_position_on_the_genomic_accession=232878695, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377757, CurrentID=0, Status='live', Symbol='LOC105377757', Aliases='', description='uncharacterized LOC105377757', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178365833, end_position_on_the_genomic_accession=178369499, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377758, CurrentID=0, Status='live', Symbol='LOC105377758', Aliases='', description='uncharacterized LOC105377758', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178302270, end_position_on_the_genomic_accession=178315869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377759, CurrentID=0, Status='live', Symbol='LOC105377759', Aliases='', description='uncharacterized LOC105377759', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179346588, end_position_on_the_genomic_accession=179349065, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377760, CurrentID=0, Status='live', Symbol='LOC105377760', Aliases='', description='uncharacterized LOC105377760', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179373431, end_position_on_the_genomic_accession=179375753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377761, CurrentID=0, Status='live', Symbol='LOC105377761', Aliases='', description='uncharacterized LOC105377761', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179456228, end_position_on_the_genomic_accession=179457903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389090, CurrentID=0, Status='live', Symbol='OR6B2', Aliases='OR2-1P, OR6B2', description='olfactory receptor family 6 subfamily B member 2', other_designations='olfactory receptor OR2-1|olfactory receptor, family 6, subfamily B, member 2 pseudogene', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240028965, end_position_on_the_genomic_accession=240033858, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377763, CurrentID=0, Status='live', Symbol='LOC105377763', Aliases='', description='uncharacterized LOC105377763', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=179657784, end_position_on_the_genomic_accession=179674052, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377764, CurrentID=0, Status='live', Symbol='LOC105377764', Aliases='', description='uncharacterized LOC105377764', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180349442, end_position_on_the_genomic_accession=180352353, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377765, CurrentID=0, Status='live', Symbol='LOC105377765', Aliases='', description='uncharacterized LOC105377765', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180684096, end_position_on_the_genomic_accession=180687502, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377766, CurrentID=0, Status='live', Symbol='LOC105377766', Aliases='', description='uncharacterized LOC105377766', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180822811, end_position_on_the_genomic_accession=180826497, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377767, CurrentID=0, Status='live', Symbol='LOC105377767', Aliases='', description='uncharacterized LOC105377767', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181045682, end_position_on_the_genomic_accession=181046131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377768, CurrentID=0, Status='live', Symbol='LOC105377768', Aliases='', description='uncharacterized LOC105377768', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=180951074, end_position_on_the_genomic_accession=181040099, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377769, CurrentID=0, Status='live', Symbol='LOC105377769', Aliases='', description='uncharacterized LOC105377769', other_designations='CTC-338M12.5', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181192023, end_position_on_the_genomic_accession=181194247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377770, CurrentID=0, Status='live', Symbol='LOC105377770', Aliases='', description='uncharacterized LOC105377770', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181454360, end_position_on_the_genomic_accession=181460306, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377771, CurrentID=0, Status='live', Symbol='LOC105377771', Aliases='', description='uncharacterized LOC105377771', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=181295347, end_position_on_the_genomic_accession=181318407, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377772, CurrentID=0, Status='live', Symbol='LOC105377772', Aliases='', description='chromosome alignment-maintaining phosphoprotein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178006273, end_position_on_the_genomic_accession=178007707, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389101, CurrentID=0, Status='live', Symbol='RPL23AP43', Aliases='RPL23A_16_360', description='ribosomal protein L23a pseudogene 43', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32781559, end_position_on_the_genomic_accession=32786180, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377774, CurrentID=0, Status='live', Symbol='LOC105377774', Aliases='', description='uncharacterized LOC105377774', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=651299, end_position_on_the_genomic_accession=662434, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377775, CurrentID=0, Status='live', Symbol='LOC105377775', Aliases='', description='uncharacterized LOC105377775', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1428297, end_position_on_the_genomic_accession=1437302, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377776, CurrentID=0, Status='live', Symbol='LOC105377776', Aliases='', description='uncharacterized LOC105377776', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1368648, end_position_on_the_genomic_accession=1373228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  126961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126961, CurrentID=0, Status='live', Symbol='HIST2H3C', Aliases='H3, H3.2, H3/M, H3F2, H3FM, H3FN', description='histone cluster 2, H3c', other_designations='H3 histone family, member M|H3 histone, family 2|histone 2, H3c|histone H3/m', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149840687, end_position_on_the_genomic_accession=149841193, orientation='minus', exon_count=1, OMIM=142780, no_hdr0=''),
  651250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651250, CurrentID=0, Status='live', Symbol='LRRC37A16P', Aliases='', description='leucine-rich repeat containing 37 member A16, pseudogene', other_designations='leucine rich repeat containing 37, member A16, pseudogene|leucine rich repeat containing 37, member A3 pseudogene', map_location='17q24.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=68126475, end_position_on_the_genomic_accession=68152471, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377779, CurrentID=0, Status='live', Symbol='LOC105377779', Aliases='', description='uncharacterized LOC105377779', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=1810338, end_position_on_the_genomic_accession=1814148, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377780, CurrentID=0, Status='live', Symbol='LOC105377780', Aliases='', description='uncharacterized LOC105377780', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2034491, end_position_on_the_genomic_accession=2039974, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377781, CurrentID=0, Status='live', Symbol='LOC105377781', Aliases='', description='uncharacterized LOC105377781', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2156673, end_position_on_the_genomic_accession=2196528, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377782 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377782, CurrentID=0, Status='live', Symbol='LOC105377782', Aliases='', description='uncharacterized LOC105377782', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2199670, end_position_on_the_genomic_accession=2206182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377783, CurrentID=0, Status='live', Symbol='LOC105377783', Aliases='', description='uncharacterized LOC105377783', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2251353, end_position_on_the_genomic_accession=2254139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389112, CurrentID=0, Status='live', Symbol='RPS27P4', Aliases='RPS27_5_366', description='ribosomal protein S27 pseudogene 4', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40761512, end_position_on_the_genomic_accession=40761856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  126969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=126969, CurrentID=0, Status='live', Symbol='SLC44A3', Aliases='CTL3', description='solute carrier family 44 member 3', other_designations='solute carrier family 44, member 3', map_location='1p21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94820342, end_position_on_the_genomic_accession=94895247, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  389114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389114, CurrentID=0, Status='live', Symbol='ZNF662', Aliases='', description='zinc finger protein 662', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42905910, end_position_on_the_genomic_accession=42919333, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377787, CurrentID=0, Status='live', Symbol='LOC105377787', Aliases='', description='uncharacterized LOC105377787', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2900319, end_position_on_the_genomic_accession=2923688, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377788, CurrentID=0, Status='live', Symbol='LOC105377788', Aliases='', description='uncharacterized LOC105377788', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=2834836, end_position_on_the_genomic_accession=2926781, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377789, CurrentID=0, Status='live', Symbol='LOC105377789', Aliases='', description='uncharacterized LOC105377789', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=4528043, end_position_on_the_genomic_accession=4537829, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389118, CurrentID=0, Status='live', Symbol='CDHR4', Aliases='CDH29, PRO34300', description='cadherin related family member 4', other_designations='Cadherin-like protein UNQ9392/PRO34300|VLLR9392|cadherin-like 29|cadherin-like protein 29', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49790732, end_position_on_the_genomic_accession=49803164, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  389119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389119, CurrentID=0, Status='live', Symbol='FAM212A', Aliases='C3orf54, INKA1', description='family with sequence similarity 212 member A', other_designations='family with sequence similarity 212, member A', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49803233, end_position_on_the_genomic_accession=49805033, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389120, CurrentID=0, Status='live', Symbol='ACTBP13', Aliases='', description='actin, beta pseudogene 13', other_designations='', map_location='3p21.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=49873345, end_position_on_the_genomic_accession=49877306, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377793, CurrentID=0, Status='live', Symbol='LOC105377793', Aliases='', description='uncharacterized LOC105377793', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5534066, end_position_on_the_genomic_accession=5545612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377794, CurrentID=0, Status='live', Symbol='LOC105377794', Aliases='', description='uncharacterized LOC105377794', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5595231, end_position_on_the_genomic_accession=5596726, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389123, CurrentID=0, Status='live', Symbol='IQCF2', Aliases='', description='IQ motif containing F2', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51861629, end_position_on_the_genomic_accession=51863424, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389124, CurrentID=0, Status='live', Symbol='IQCF5', Aliases='', description='IQ motif containing F5', other_designations='', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=51873721, end_position_on_the_genomic_accession=51875584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389125, CurrentID=0, Status='live', Symbol='MUSTN1', Aliases='MUSTANG', description='musculoskeletal, embryonic nuclear protein 1', other_designations='musculoskeletal temporally activated novel', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52833115, end_position_on_the_genomic_accession=52835219, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377798, CurrentID=0, Status='live', Symbol='LOC105377798', Aliases='', description='uncharacterized LOC105377798', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6859027, end_position_on_the_genomic_accession=6865673, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389127, CurrentID=0, Status='live', Symbol='RPS10P10', Aliases='RPS10_4_378', description='ribosomal protein S10 pseudogene 10', other_designations='', map_location='3p14.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=62926343, end_position_on_the_genomic_accession=62926928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377800, CurrentID=0, Status='live', Symbol='LOC105377800', Aliases='', description='uncharacterized LOC105377800', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7575171, end_position_on_the_genomic_accession=7584130, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377801, CurrentID=0, Status='live', Symbol='LOC105377801', Aliases='', description='uncharacterized LOC105377801', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7520340, end_position_on_the_genomic_accession=7520815, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377802, CurrentID=0, Status='live', Symbol='LOC105377802', Aliases='', description='uncharacterized LOC105377802', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7252291, end_position_on_the_genomic_accession=7256411, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377803, CurrentID=0, Status='live', Symbol='LOC105377803', Aliases='', description='uncharacterized LOC105377803', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7138531, end_position_on_the_genomic_accession=7186452, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377804, CurrentID=0, Status='live', Symbol='LOC105377804', Aliases='', description='uncharacterized LOC105377804', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113752096, end_position_on_the_genomic_accession=113754369, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377805, CurrentID=0, Status='live', Symbol='LOC105377805', Aliases='', description='basic salivary proline-rich protein 3-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113741875, end_position_on_the_genomic_accession=113752839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  503635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=503635, CurrentID=0, Status='live', Symbol='DUXAP6', Aliases='', description='double homeobox A pseudogene 6', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92281438, end_position_on_the_genomic_accession=92282474, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377807, CurrentID=0, Status='live', Symbol='LOC105377807', Aliases='', description='uncharacterized LOC105377807', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=113933165, end_position_on_the_genomic_accession=113938181, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389136, CurrentID=0, Status='live', Symbol='VGLL3', Aliases='VGL-3, VGL3', description='vestigial like family member 3', other_designations='colon carcinoma related protein|vestigial like 3|vestigial-like 3|vestigial-like family member 3', map_location='3p12.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=86937973, end_position_on_the_genomic_accession=86991287, orientation='minus', exon_count=5, OMIM=609980, no_hdr0=''),
  389137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389137, CurrentID=0, Status='live', Symbol='ARMC10P1', Aliases='', description='armadillo repeat containing 10 pseudogene 1', other_designations='', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=94505538, end_position_on_the_genomic_accession=94507650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377810 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377810, CurrentID=0, Status='live', Symbol='LOC105377810', Aliases='', description='uncharacterized LOC105377810', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377811, CurrentID=0, Status='live', Symbol='LOC105377811', Aliases='', description='uncharacterized LOC105377811', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105377812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377812, CurrentID=0, Status='live', Symbol='LOC105377812', Aliases='', description='uncharacterized LOC105377812', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389141, CurrentID=0, Status='live', Symbol='RPSAP29', Aliases='RPSA_6_392', description='ribosomal protein SA pseudogene 29', other_designations='', map_location='3q13.13', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=110682216, end_position_on_the_genomic_accession=110683237, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100134934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134934, CurrentID=0, Status='live', Symbol='TEN1', Aliases='C17orf106', description='TEN1 CST complex subunit', other_designations='TEN1 telomerase capping complex subunit homolog|protein telomeric pathways with STN1 homolog|telomere length regulation protein TEN1 homolog|telomeric pathways in association with Stn1, number 1', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75979217, end_position_on_the_genomic_accession=76000586, orientation='plus', exon_count=4, OMIM=613130, no_hdr0=''),
  105377815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377815, CurrentID=0, Status='live', Symbol='LOC105377815', Aliases='', description='uncharacterized LOC105377815', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=63846957, end_position_on_the_genomic_accession=63859179, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389144, CurrentID=0, Status='live', Symbol='OR7E130P', Aliases='OST702', description='olfactory receptor family 7 subfamily E member 130 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 130 pseudogene', map_location='3q21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=125703385, end_position_on_the_genomic_accession=125704347, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100134937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100134937, CurrentID=0, Status='live', Symbol='LOC100134937', Aliases='', description='uncharacterized LOC100134937', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  127002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127002, CurrentID=0, Status='live', Symbol='ATXN7L2', Aliases='', description='ataxin 7-like 2', other_designations='ataxin 7-like 1', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109483911, end_position_on_the_genomic_accession=109492804, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  127003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127003, CurrentID=0, Status='live', Symbol='C1orf194', Aliases='', description='chromosome 1 open reading frame 194', other_designations='', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=109105951, end_position_on_the_genomic_accession=109113870, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377820, CurrentID=0, Status='live', Symbol='LOC105377820', Aliases='', description='uncharacterized LOC105377820', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35255045, end_position_on_the_genomic_accession=35255540, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377821, CurrentID=0, Status='live', Symbol='LOC105377821', Aliases='', description='collagen alpha-2(I) chain-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22392722, end_position_on_the_genomic_accession=22399180, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377822, CurrentID=0, Status='live', Symbol='LOC105377822', Aliases='', description='uncharacterized LOC105377822', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=22388744, end_position_on_the_genomic_accession=22389971, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389151, CurrentID=0, Status='live', Symbol='PRR23B', Aliases='', description='proline rich 23B', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139019031, end_position_on_the_genomic_accession=139020926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  389152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389152, CurrentID=0, Status='live', Symbol='PRR23C', Aliases='', description='proline rich 23C', other_designations='proline-rich protein 23A', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=139042102, end_position_on_the_genomic_accession=139044892, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377825, CurrentID=0, Status='live', Symbol='LOC105377825', Aliases='', description='putative protein FAM157B', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=114736, end_position_on_the_genomic_accession=133559, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377826, CurrentID=0, Status='live', Symbol='LOC105377826', Aliases='', description='uncharacterized LOC105377826', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=81905, end_position_on_the_genomic_accession=86392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377827, CurrentID=0, Status='live', Symbol='LOC105377827', Aliases='', description='scavenger receptor class F member 2-like', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=63309, end_position_on_the_genomic_accession=64666, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389156, CurrentID=0, Status='live', Symbol='RPL21P39', Aliases='RPL21_14_415', description='ribosomal protein L21 pseudogene 39', other_designations='', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=145824243, end_position_on_the_genomic_accession=145824778, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377829, CurrentID=0, Status='live', Symbol='LOC105377829', Aliases='', description='uncharacterized LOC105377829', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389158, CurrentID=0, Status='live', Symbol='PLSCR5', Aliases='', description='phospholipid scramblase family member 5', other_designations='phospholipid scramblase family, member 5', map_location='3q24', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=146576608, end_position_on_the_genomic_accession=146606216, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377831, CurrentID=0, Status='live', Symbol='LOC105377831', Aliases='', description='uncharacterized LOC105377831', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389160, CurrentID=0, Status='live', Symbol='CPHL1P', Aliases='CPHL1', description='ceruloplasmin and hephaestin-like 1 pseudogene', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149241362, end_position_on_the_genomic_accession=149275607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389161, CurrentID=0, Status='live', Symbol='ANKUB1', Aliases='C3orf16', description='ankyrin repeat and ubiquitin domain containing 1', other_designations='ankyrin repeat and ubiquitin domain-containing 1', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149760648, end_position_on_the_genomic_accession=149792843, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  127018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127018, CurrentID=0, Status='live', Symbol='LYPLAL1', Aliases='Q96AV0', description='lysophospholipase-like 1', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=219173831, end_position_on_the_genomic_accession=219212865, orientation='plus', exon_count=8, OMIM=616548, no_hdr0=''),
  105377835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377835, CurrentID=0, Status='live', Symbol='LOC105377835', Aliases='', description='uncharacterized LOC105377835', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=60784951, end_position_on_the_genomic_accession=60789765, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377836, CurrentID=0, Status='live', Symbol='LOC105377836', Aliases='', description='uncharacterized LOC105377836', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63926492, end_position_on_the_genomic_accession=63930763, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377837, CurrentID=0, Status='live', Symbol='LOC105377837', Aliases='', description='uncharacterized LOC105377837', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=64979060, end_position_on_the_genomic_accession=64994288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377838, CurrentID=0, Status='live', Symbol='LOC105377838', Aliases='', description='uncharacterized LOC105377838', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=65989471, end_position_on_the_genomic_accession=65998637, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377839, CurrentID=0, Status='live', Symbol='LOC105377839', Aliases='', description='uncharacterized LOC105377839', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66054412, end_position_on_the_genomic_accession=66055177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389168, CurrentID=0, Status='live', Symbol='SETP14', Aliases='', description='SET pseudogene 14', other_designations='SET translocation pseudogene', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=155987415, end_position_on_the_genomic_accession=155989042, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377841, CurrentID=0, Status='live', Symbol='LOC105377841', Aliases='', description='uncharacterized LOC105377841', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=66134951, end_position_on_the_genomic_accession=66176338, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  389170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389170, CurrentID=0, Status='live', Symbol='LEKR1', Aliases='', description='leucine, glutamate and lysine rich 1', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=156826307, end_position_on_the_genomic_accession=157047659, orientation='plus', exon_count=22, OMIM=613536, no_hdr0=''),
  105377843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377843, CurrentID=0, Status='live', Symbol='LOC105377843', Aliases='', description='uncharacterized LOC105377843', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67282677, end_position_on_the_genomic_accession=67301037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377844, CurrentID=0, Status='live', Symbol='LOC105377844', Aliases='', description='uncharacterized LOC105377844', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67995806, end_position_on_the_genomic_accession=67998743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377845, CurrentID=0, Status='live', Symbol='LOC105377845', Aliases='', description='uncharacterized LOC105377845', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=67886974, end_position_on_the_genomic_accession=67932198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377846, CurrentID=0, Status='live', Symbol='LOC105377846', Aliases='', description='uncharacterized LOC105377846', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=68091854, end_position_on_the_genomic_accession=68113612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377847, CurrentID=0, Status='live', Symbol='LOC105377847', Aliases='', description='uncharacterized LOC105377847', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70384175, end_position_on_the_genomic_accession=70399208, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377848, CurrentID=0, Status='live', Symbol='LOC105377848', Aliases='', description='uncharacterized LOC105377848', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70348121, end_position_on_the_genomic_accession=70381447, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  389177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389177, CurrentID=0, Status='live', Symbol='TMEM212', Aliases='', description='transmembrane protein 212', other_designations='', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=171843349, end_position_on_the_genomic_accession=171867846, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377850, CurrentID=0, Status='live', Symbol='LOC105377850', Aliases='', description='uncharacterized LOC105377850', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=70882101, end_position_on_the_genomic_accession=70905324, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  389179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389179, CurrentID=0, Status='live', Symbol='EEF1A1P8', Aliases='EEF1AL6', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 8', other_designations='eukaryotic translation elongation factor 1 alpha-like 6 pseudogene', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184026369, end_position_on_the_genomic_accession=184027756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389180, CurrentID=0, Status='live', Symbol='5-HT3C2', Aliases='', description='5-hydroxytryptamine (serotonin) receptor 3, family member E pseudogene', other_designations='5-hydroxytryptamine (serotonin) receptor 3E, ionotropic pseudogene', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184076646, end_position_on_the_genomic_accession=184084331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377853, CurrentID=0, Status='live', Symbol='LOC105377853', Aliases='', description='uncharacterized LOC105377853', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=71469804, end_position_on_the_genomic_accession=71479480, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377854, CurrentID=0, Status='live', Symbol='LOC105377854', Aliases='', description='uncharacterized LOC105377854', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73003921, end_position_on_the_genomic_accession=73005994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377855 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377855, CurrentID=0, Status='live', Symbol='LOC105377855', Aliases='', description='uncharacterized LOC105377855', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72945826, end_position_on_the_genomic_accession=72994626, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377856, CurrentID=0, Status='live', Symbol='LOC105377856', Aliases='', description='uncharacterized LOC105377856', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=72614330, end_position_on_the_genomic_accession=72621869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377857, CurrentID=0, Status='live', Symbol='LOC105377857', Aliases='', description='uncharacterized LOC105377857', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73523618, end_position_on_the_genomic_accession=73571599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377858, CurrentID=0, Status='live', Symbol='LOC105377858', Aliases='', description='uncharacterized LOC105377858', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=74594253, end_position_on_the_genomic_accession=74734294, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377859, CurrentID=0, Status='live', Symbol='LOC105377859', Aliases='', description='uncharacterized LOC105377859', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=75357119, end_position_on_the_genomic_accession=75362418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377860, CurrentID=0, Status='live', Symbol='LOC105377860', Aliases='', description='uncharacterized LOC105377860', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=76522374, end_position_on_the_genomic_accession=76593800, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377861, CurrentID=0, Status='live', Symbol='LOC105377861', Aliases='', description='uncharacterized LOC105377861', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=76724217, end_position_on_the_genomic_accession=76726450, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377862, CurrentID=0, Status='live', Symbol='LOC105377862', Aliases='', description='uncharacterized LOC105377862', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=76774936, end_position_on_the_genomic_accession=77137973, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105377863 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377863, CurrentID=0, Status='live', Symbol='LOC105377863', Aliases='', description='uncharacterized LOC105377863', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77427884, end_position_on_the_genomic_accession=77433654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377864, CurrentID=0, Status='live', Symbol='LOC105377864', Aliases='', description='uncharacterized LOC105377864', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77403604, end_position_on_the_genomic_accession=77503988, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105377865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377865, CurrentID=0, Status='live', Symbol='LOC105377865', Aliases='', description='uncharacterized LOC105377865', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=77925902, end_position_on_the_genomic_accession=78300807, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377866, CurrentID=0, Status='live', Symbol='LOC105377866', Aliases='', description='uncharacterized LOC105377866', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79537645, end_position_on_the_genomic_accession=79631131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377867, CurrentID=0, Status='live', Symbol='LOC105377867', Aliases='', description='uncharacterized LOC105377867', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79444037, end_position_on_the_genomic_accession=79447068, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377868, CurrentID=0, Status='live', Symbol='LOC105377868', Aliases='', description='uncharacterized LOC105377868', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80259696, end_position_on_the_genomic_accession=80262907, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389197, CurrentID=0, Status='live', Symbol='C4orf50', Aliases='', description='chromosome 4 open reading frame 50', other_designations='', map_location='4p16.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=5957112, end_position_on_the_genomic_accession=6008060, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105377870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377870, CurrentID=0, Status='live', Symbol='LOC105377870', Aliases='', description='uncharacterized LOC105377870', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=80904608, end_position_on_the_genomic_accession=80924038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389199, CurrentID=0, Status='live', Symbol='LOC389199', Aliases='', description='uncharacterized LOC389199', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=7938921, end_position_on_the_genomic_accession=7940296, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377872, CurrentID=0, Status='live', Symbol='LOC105377872', Aliases='', description='uncharacterized LOC105377872', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81692567, end_position_on_the_genomic_accession=81740179, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105377873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377873, CurrentID=0, Status='live', Symbol='LOC105377873', Aliases='', description='uncharacterized LOC105377873', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=81760295, end_position_on_the_genomic_accession=81768415, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377874, CurrentID=0, Status='live', Symbol='LOC105377874', Aliases='', description='uncharacterized LOC105377874', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82271426, end_position_on_the_genomic_accession=82272641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127059, CurrentID=0, Status='live', Symbol='OR2M5', Aliases='OR2M5P', description='olfactory receptor family 2 subfamily M member 5', other_designations='olfactory receptor, family 2, subfamily M, member 5 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248145148, end_position_on_the_genomic_accession=248146086, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377876, CurrentID=0, Status='live', Symbol='LOC105377876', Aliases='', description='uncharacterized LOC105377876', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82598014, end_position_on_the_genomic_accession=82688414, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377877, CurrentID=0, Status='live', Symbol='LOC105377877', Aliases='', description='uncharacterized LOC105377877', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=82819565, end_position_on_the_genomic_accession=82836480, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389206, CurrentID=0, Status='live', Symbol='BEND4', Aliases='CCDC4', description='BEN domain containing 4', other_designations='coiled-coil domain containing 4|coiled-coil domain-containing protein 4', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42110853, end_position_on_the_genomic_accession=42152878, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  389207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389207, CurrentID=0, Status='live', Symbol='GRXCR1', Aliases='DFNB25, PPP1R88', description='glutaredoxin, cysteine rich 1', other_designations='protein phosphatase 1, regulatory subunit 88', map_location='4p13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=42893266, end_position_on_the_genomic_accession=43030658, orientation='plus', exon_count=5, OMIM=613283, no_hdr0=''),
  389208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389208, CurrentID=0, Status='live', Symbol='TMPRSS11F', Aliases='', description='transmembrane protease, serine 11F', other_designations='airway trypsin-like protease 4', map_location='4q13.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=68053198, end_position_on_the_genomic_accession=68129869, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105377881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377881, CurrentID=0, Status='live', Symbol='LOC105377881', Aliases='', description='uncharacterized LOC105377881', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=84961735, end_position_on_the_genomic_accession=84967809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127066, CurrentID=0, Status='live', Symbol='OR14C36', Aliases='OR5BF1', description='olfactory receptor family 14 subfamily C member 36', other_designations='olfactory receptor 5BF1|olfactory receptor OR1-59|olfactory receptor, family 14, subfamily C, member 36|olfactory receptor, family 5, subfamily BF, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248348775, end_position_on_the_genomic_accession=248349713, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377883, CurrentID=0, Status='live', Symbol='LOC105377883', Aliases='', description='uncharacterized LOC105377883', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=88217318, end_position_on_the_genomic_accession=88249176, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127068, CurrentID=0, Status='live', Symbol='OR2T34', Aliases='', description='olfactory receptor family 2 subfamily T member 34', other_designations='olfactory receptor OR1-63|olfactory receptor, family 2, subfamily T, member 34', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248573801, end_position_on_the_genomic_accession=248574757, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  127069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127069, CurrentID=0, Status='live', Symbol='OR2T10', Aliases='OR1-64', description='olfactory receptor family 2 subfamily T member 10', other_designations='olfactory receptor OR1-64|olfactory receptor, family 2, subfamily T, member 10', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248592830, end_position_on_the_genomic_accession=248593768, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100271556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271556, CurrentID=0, Status='live', Symbol='RPS12P32', Aliases='RPS12_16_1623', description='ribosomal protein S12 pseudogene 32', other_designations='', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=11523356, end_position_on_the_genomic_accession=11523853, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377887, CurrentID=0, Status='live', Symbol='LOC105377887', Aliases='', description='uncharacterized LOC105377887', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89621403, end_position_on_the_genomic_accession=89624577, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377888, CurrentID=0, Status='live', Symbol='LOC105377888', Aliases='', description='uncharacterized LOC105377888', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=89537818, end_position_on_the_genomic_accession=89566972, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  389217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389217, CurrentID=0, Status='live', Symbol='SETP20', Aliases='', description='SET pseudogene 20', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=109553246, end_position_on_the_genomic_accession=109554094, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  127074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127074, CurrentID=0, Status='live', Symbol='OR2T4', Aliases='OR1-60Q, OR2T4', description='olfactory receptor family 2 subfamily T member 4', other_designations='olfactory receptor OR1-60|olfactory receptor, family 2, subfamily T, member 4', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248361581, end_position_on_the_genomic_accession=248362627, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105377891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377891, CurrentID=0, Status='live', Symbol='LOC105377891', Aliases='', description='uncharacterized LOC105377891', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90331360, end_position_on_the_genomic_accession=90357265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377892, CurrentID=0, Status='live', Symbol='LOC105377892', Aliases='', description='uncharacterized LOC105377892', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=90587211, end_position_on_the_genomic_accession=90590736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127077, CurrentID=0, Status='live', Symbol='OR2T11', Aliases='OR2T11Q', description='olfactory receptor family 2 subfamily T member 11 (gene/pseudogene)', other_designations='olfactory receptor OR1-65|olfactory receptor, family 2, subfamily T, member 11', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248626178, end_position_on_the_genomic_accession=248627128, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377894, CurrentID=0, Status='live', Symbol='LOC105377894', Aliases='', description='uncharacterized LOC105377894', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=91719752, end_position_on_the_genomic_accession=91729795, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389223, CurrentID=0, Status='live', Symbol='EEF1A1P35', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 35', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=135045464, end_position_on_the_genomic_accession=135046847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377896, CurrentID=0, Status='live', Symbol='LOC105377896', Aliases='', description='uncharacterized LOC105377896', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92584904, end_position_on_the_genomic_accession=92590151, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377897 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377897, CurrentID=0, Status='live', Symbol='LOC105377897', Aliases='', description='uncharacterized LOC105377897', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=92979599, end_position_on_the_genomic_accession=92986870, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377898, CurrentID=0, Status='live', Symbol='LOC105377898', Aliases='', description='uncharacterized LOC105377898', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93030894, end_position_on_the_genomic_accession=93069761, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377899, CurrentID=0, Status='live', Symbol='LOC105377899', Aliases='', description='uncharacterized LOC105377899', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93418648, end_position_on_the_genomic_accession=93503360, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377900 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377900, CurrentID=0, Status='live', Symbol='LOC105377900', Aliases='', description='uncharacterized LOC105377900', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=93674395, end_position_on_the_genomic_accession=93702017, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377901, CurrentID=0, Status='live', Symbol='LOC105377901', Aliases='', description='uncharacterized LOC105377901', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=94546225, end_position_on_the_genomic_accession=94573259, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  127086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127086, CurrentID=0, Status='live', Symbol='XRCC6P3', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 6 pseudogene 3', other_designations='thyroid autoantigen 70kDa pseudogene', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=220313709, end_position_on_the_genomic_accession=220315730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377903, CurrentID=0, Status='live', Symbol='LOC105377903', Aliases='', description='uncharacterized LOC105377903', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=95487907, end_position_on_the_genomic_accession=95519565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377904, CurrentID=0, Status='live', Symbol='LOC105377904', Aliases='', description='uncharacterized LOC105377904', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96200343, end_position_on_the_genomic_accession=96399031, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105377905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377905, CurrentID=0, Status='live', Symbol='LOC105377905', Aliases='', description='uncharacterized LOC105377905', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96052968, end_position_on_the_genomic_accession=96056660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377906, CurrentID=0, Status='live', Symbol='LOC105377906', Aliases='', description='uncharacterized LOC105377906', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96526041, end_position_on_the_genomic_accession=96531033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377907, CurrentID=0, Status='live', Symbol='LOC105377907', Aliases='', description='uncharacterized LOC105377907', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96499957, end_position_on_the_genomic_accession=96517331, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377908, CurrentID=0, Status='live', Symbol='LOC105377908', Aliases='', description='uncharacterized LOC105377908', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=96717622, end_position_on_the_genomic_accession=96721365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377909, CurrentID=0, Status='live', Symbol='LOC105377909', Aliases='', description='uncharacterized LOC105377909', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=97456571, end_position_on_the_genomic_accession=97456979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377910, CurrentID=0, Status='live', Symbol='LOC105377910', Aliases='', description='uncharacterized LOC105377910', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=98517894, end_position_on_the_genomic_accession=98519842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377911, CurrentID=0, Status='live', Symbol='LOC105377911', Aliases='', description='uncharacterized LOC105377911', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=100149639, end_position_on_the_genomic_accession=100156113, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389240, CurrentID=0, Status='live', Symbol='NACA3P', Aliases='NACA3', description='NACA family member 3 pseudogene', other_designations='nascent-polypeptide-associated complex alpha polypeptide pseudogene', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164943243, end_position_on_the_genomic_accession=164944040, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377913, CurrentID=0, Status='live', Symbol='LOC105377913', Aliases='', description='uncharacterized LOC105377913', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=102160999, end_position_on_the_genomic_accession=102225366, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105377914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377914, CurrentID=0, Status='live', Symbol='LOC105377914', Aliases='', description='uncharacterized LOC105377914', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=103000964, end_position_on_the_genomic_accession=103162055, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  127099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127099, CurrentID=0, Status='live', Symbol='RPLP1P3', Aliases='RPLP1_1_124', description='ribosomal protein, large, P1 pseudogene 3', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171824602, end_position_on_the_genomic_accession=171825041, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377916, CurrentID=0, Status='live', Symbol='LOC105377916', Aliases='', description='uncharacterized LOC105377916', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=103535038, end_position_on_the_genomic_accession=103540791, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377917 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377917, CurrentID=0, Status='live', Symbol='LOC105377917', Aliases='', description='uncharacterized LOC105377917', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104275487, end_position_on_the_genomic_accession=104286853, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389246, CurrentID=0, Status='live', Symbol='VTI1BP2', Aliases='', description='vesicle transport through interaction with t-SNAREs 1B pseudogene 2', other_designations='VTI1B pseudogene|vesicle transport through interaction with t-SNAREs homolog 1B pseudogene 2', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183483350, end_position_on_the_genomic_accession=183484416, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  389247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389247, CurrentID=0, Status='live', Symbol='LOC389247', Aliases='', description='uncharacterized LOC389247', other_designations='', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183494737, end_position_on_the_genomic_accession=183504515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377920, CurrentID=0, Status='live', Symbol='LOC105377920', Aliases='', description='uncharacterized LOC105377920', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105275772, end_position_on_the_genomic_accession=105277981, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377921 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377921, CurrentID=0, Status='live', Symbol='LOC105377921', Aliases='', description='uncharacterized LOC105377921', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105453668, end_position_on_the_genomic_accession=105479087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377922, CurrentID=0, Status='live', Symbol='LOC105377922', Aliases='', description='uncharacterized LOC105377922', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105611547, end_position_on_the_genomic_accession=105632271, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377923 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377923, CurrentID=0, Status='live', Symbol='LOC105377923', Aliases='', description='uncharacterized LOC105377923', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=105637105, end_position_on_the_genomic_accession=105893062, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377924, CurrentID=0, Status='live', Symbol='LOC105377924', Aliases='', description='uncharacterized LOC105377924', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106451496, end_position_on_the_genomic_accession=106457267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377925 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377925, CurrentID=0, Status='live', Symbol='LOC105377925', Aliases='', description='uncharacterized LOC105377925', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106587888, end_position_on_the_genomic_accession=106588771, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377926 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377926, CurrentID=0, Status='live', Symbol='LOC105377926', Aliases='', description='uncharacterized LOC105377926', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106686982, end_position_on_the_genomic_accession=106699946, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377927, CurrentID=0, Status='live', Symbol='LOC105377927', Aliases='', description='uncharacterized LOC105377927', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106702868, end_position_on_the_genomic_accession=106716503, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377928, CurrentID=0, Status='live', Symbol='LOC105377928', Aliases='', description='uncharacterized LOC105377928', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=106881938, end_position_on_the_genomic_accession=106884664, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389257, CurrentID=0, Status='live', Symbol='LRRC14B', Aliases='', description='leucine rich repeat containing 14B', other_designations='', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=191511, end_position_on_the_genomic_accession=195353, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377930, CurrentID=0, Status='live', Symbol='LOC105377930', Aliases='', description='uncharacterized LOC105377930', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108120200, end_position_on_the_genomic_accession=108122489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377931 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377931, CurrentID=0, Status='live', Symbol='LOC105377931', Aliases='', description='uncharacterized LOC105377931', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108468773, end_position_on_the_genomic_accession=108492382, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377932, CurrentID=0, Status='live', Symbol='LOC105377932', Aliases='', description='uncharacterized LOC105377932', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=108359371, end_position_on_the_genomic_accession=108407461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  728743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=728743, CurrentID=0, Status='live', Symbol='LOC728743', Aliases='', description='zinc finger protein pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150405752, end_position_on_the_genomic_accession=150412470, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377934, CurrentID=0, Status='live', Symbol='LOC105377934', Aliases='', description='uncharacterized LOC105377934', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109183930, end_position_on_the_genomic_accession=109203797, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377935, CurrentID=0, Status='live', Symbol='LOC105377935', Aliases='', description='uncharacterized LOC105377935', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=109482768, end_position_on_the_genomic_accession=109495331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377936 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377936, CurrentID=0, Status='live', Symbol='LOC105377936', Aliases='', description='uncharacterized LOC105377936', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110415700, end_position_on_the_genomic_accession=110420130, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377937, CurrentID=0, Status='live', Symbol='LOC105377937', Aliases='', description='uncharacterized LOC105377937', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110241041, end_position_on_the_genomic_accession=110254214, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377938, CurrentID=0, Status='live', Symbol='LOC105377938', Aliases='', description='uncharacterized LOC105377938', other_designations='CTA-331P3.1', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110475352, end_position_on_the_genomic_accession=110479436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377939 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377939, CurrentID=0, Status='live', Symbol='LOC105377939', Aliases='', description='uncharacterized LOC105377939', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110549563, end_position_on_the_genomic_accession=110555691, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  127124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127124, CurrentID=0, Status='live', Symbol='ATP6V1G3', Aliases='ATP6G3, Vma10', description='ATPase, H+ transporting, lysosomal 13kDa, V1 subunit G3', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) subunit G3|V-ATPase 13 kDa subunit 3|V-ATPase G subunit 3|V-ATPase G3 subunit|V-ATPase subunit G 3|vacuolar ATP synthase subunit G 3|vacuolar proton pump G subunit 3|vacuolar proton pump subunit G 3|vacuolar proton pump, subunit G3', map_location='1q31.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=198523222, end_position_on_the_genomic_accession=198540945, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377941, CurrentID=0, Status='live', Symbol='LOC105377941', Aliases='', description='uncharacterized LOC105377941', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110575411, end_position_on_the_genomic_accession=110589420, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377942, CurrentID=0, Status='live', Symbol='LOC105377942', Aliases='', description='uncharacterized LOC105377942', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=110943460, end_position_on_the_genomic_accession=110958478, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100135063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100135063, CurrentID=0, Status='live', Symbol='GTF3AP6', Aliases='', description='general transcription factor IIIA pseudogene 6', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=118880187, end_position_on_the_genomic_accession=118881250, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100135064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100135064, CurrentID=0, Status='live', Symbol='GTF3AP1', Aliases='', description='general transcription factor IIIA pseudogene 1', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=6196040, end_position_on_the_genomic_accession=6196776, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283332, CurrentID=0, Status='live', Symbol='LOC283332', Aliases='', description='uncharacterized LOC283332', other_designations='', map_location='12q13.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=49908882, end_position_on_the_genomic_accession=49911863, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377946 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377946, CurrentID=0, Status='live', Symbol='LOC105377946', Aliases='', description='uncharacterized LOC105377946', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111900305, end_position_on_the_genomic_accession=111909477, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377947, CurrentID=0, Status='live', Symbol='LOC105377947', Aliases='', description='uncharacterized LOC105377947', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=111980871, end_position_on_the_genomic_accession=111986954, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372482, CurrentID=0, Status='live', Symbol='LOC105372482', Aliases='', description='uncharacterized LOC105372482', other_designations='CTD-3138B18.5', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=58258988, end_position_on_the_genomic_accession=58278791, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377949, CurrentID=0, Status='live', Symbol='LOC105377949', Aliases='', description='uncharacterized LOC105377949', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112689900, end_position_on_the_genomic_accession=112771666, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105377950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377950, CurrentID=0, Status='live', Symbol='LOC105377950', Aliases='', description='uncharacterized LOC105377950', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112880153, end_position_on_the_genomic_accession=112904294, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377951, CurrentID=0, Status='live', Symbol='LOC105377951', Aliases='', description='uncharacterized LOC105377951', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112983720, end_position_on_the_genomic_accession=112997698, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377952, CurrentID=0, Status='live', Symbol='LOC105377952', Aliases='', description='uncharacterized LOC105377952', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113094282, end_position_on_the_genomic_accession=113111755, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377953, CurrentID=0, Status='live', Symbol='LOC105377953', Aliases='', description='uncharacterized LOC105377953', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113174911, end_position_on_the_genomic_accession=113197540, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377954, CurrentID=0, Status='live', Symbol='LOC105377954', Aliases='', description='uncharacterized LOC105377954', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113172041, end_position_on_the_genomic_accession=113173547, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377955, CurrentID=0, Status='live', Symbol='LOC105377955', Aliases='', description='uncharacterized LOC105377955', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113168495, end_position_on_the_genomic_accession=113194877, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377956, CurrentID=0, Status='live', Symbol='LOC105377956', Aliases='', description='uncharacterized LOC105377956', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113122122, end_position_on_the_genomic_accession=113131546, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377957, CurrentID=0, Status='live', Symbol='LOC105377957', Aliases='', description='uncharacterized LOC105377957', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=113424362, end_position_on_the_genomic_accession=113441493, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  651430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651430, CurrentID=0, Status='live', Symbol='FLJ38576', Aliases='', description='uncharacterized LOC651430', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=186189032, end_position_on_the_genomic_accession=186191490, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105377959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377959, CurrentID=0, Status='live', Symbol='LOC105377959', Aliases='', description='uncharacterized LOC105377959', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=114477350, end_position_on_the_genomic_accession=114545331, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105377960 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377960, CurrentID=0, Status='live', Symbol='LOC105377960', Aliases='', description='uncharacterized LOC105377960', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=115046774, end_position_on_the_genomic_accession=115053480, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389289, CurrentID=0, Status='live', Symbol='ANXA2R', Aliases='AX2R, AXIIR, C5orf39', description='annexin A2 receptor', other_designations='annexin II receptor', map_location='5p12', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=43039080, end_position_on_the_genomic_accession=43040345, orientation='minus', exon_count=1, OMIM=611296, no_hdr0=''),
  105377962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377962, CurrentID=0, Status='live', Symbol='LOC105377962', Aliases='', description='uncharacterized LOC105377962', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=115633586, end_position_on_the_genomic_accession=115643898, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  283335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283335, CurrentID=0, Status='live', Symbol='LOC283335', Aliases='', description='uncharacterized LOC283335', other_designations='', map_location='12q13.13', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53043189, end_position_on_the_genomic_accession=53054438, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377964, CurrentID=0, Status='live', Symbol='LOC105377964', Aliases='', description='uncharacterized LOC105377964', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=117261489, end_position_on_the_genomic_accession=117264784, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389293, CurrentID=0, Status='live', Symbol='ISCA1P1', Aliases='ISCA1L', description='iron-sulfur cluster assembly 1 pseudogene 1', other_designations='iron-sulfur cluster assembly 1 homolog pseudogene 1', map_location='5q12.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=62775376, end_position_on_the_genomic_accession=62777344, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377966, CurrentID=0, Status='live', Symbol='LOC105377966', Aliases='', description='uncharacterized LOC105377966', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118290417, end_position_on_the_genomic_accession=118325365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377967, CurrentID=0, Status='live', Symbol='LOC105377967', Aliases='', description='uncharacterized LOC105377967', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118085903, end_position_on_the_genomic_accession=118095452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377968, CurrentID=0, Status='live', Symbol='LOC105377968', Aliases='', description='uncharacterized LOC105377968', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118335360, end_position_on_the_genomic_accession=118448439, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377969, CurrentID=0, Status='live', Symbol='LOC105377969', Aliases='', description='uncharacterized LOC105377969', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118664515, end_position_on_the_genomic_accession=118687766, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377970, CurrentID=0, Status='live', Symbol='LOC105377970', Aliases='', description='uncharacterized LOC105377970', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118496603, end_position_on_the_genomic_accession=118540443, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377971, CurrentID=0, Status='live', Symbol='LOC105377971', Aliases='', description='uncharacterized LOC105377971', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=118466478, end_position_on_the_genomic_accession=118489030, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377972, CurrentID=0, Status='live', Symbol='LOC105377972', Aliases='', description='uncharacterized LOC105377972', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119654068, end_position_on_the_genomic_accession=119703623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377973, CurrentID=0, Status='live', Symbol='LOC105377973', Aliases='', description='uncharacterized LOC105377973', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119473412, end_position_on_the_genomic_accession=119478658, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105377974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377974, CurrentID=0, Status='live', Symbol='LOC105377974', Aliases='', description='uncharacterized LOC105377974', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119209090, end_position_on_the_genomic_accession=119219345, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377975, CurrentID=0, Status='live', Symbol='LOC105377975', Aliases='', description='uncharacterized LOC105377975', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119843066, end_position_on_the_genomic_accession=119845084, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377976, CurrentID=0, Status='live', Symbol='LOC105377976', Aliases='', description='uncharacterized LOC105377976', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=119877930, end_position_on_the_genomic_accession=119880655, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377977, CurrentID=0, Status='live', Symbol='LOC105377977', Aliases='', description='uncharacterized LOC105377977', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=120970994, end_position_on_the_genomic_accession=121004252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377978, CurrentID=0, Status='live', Symbol='LOC105377978', Aliases='', description='uncharacterized LOC105377978', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121865959, end_position_on_the_genomic_accession=121884715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377979, CurrentID=0, Status='live', Symbol='LOC105377979', Aliases='', description='uncharacterized LOC105377979', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121767376, end_position_on_the_genomic_accession=122055650, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  651452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651452, CurrentID=0, Status='live', Symbol='POM121L3P', Aliases='POM121L3, dJ831C21.3', description='POM121 transmembrane nucleoporin like 3, pseudogene', other_designations='POM121 membrane glycoprotein (rat homolog)-like 3|POM121 membrane glycoprotein-like 3, pseudogene|POM121 transmembrane nucleoporin-like 3, pseudogene', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23989292, end_position_on_the_genomic_accession=23990447, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105377981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377981, CurrentID=0, Status='live', Symbol='LOC105377981', Aliases='', description='uncharacterized LOC105377981', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123610434, end_position_on_the_genomic_accession=123668971, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377982, CurrentID=0, Status='live', Symbol='LOC105377982', Aliases='', description='uncharacterized LOC105377982', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123519439, end_position_on_the_genomic_accession=123570916, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105377983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377983, CurrentID=0, Status='live', Symbol='LOC105377983', Aliases='', description='uncharacterized LOC105377983', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123506595, end_position_on_the_genomic_accession=123509746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377984 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377984, CurrentID=0, Status='live', Symbol='LOC105377984', Aliases='', description='uncharacterized LOC105377984', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=123309152, end_position_on_the_genomic_accession=123336000, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377985, CurrentID=0, Status='live', Symbol='LOC105377985', Aliases='', description='uncharacterized LOC105377985', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125267735, end_position_on_the_genomic_accession=125276098, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105377986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377986, CurrentID=0, Status='live', Symbol='LOC105377986', Aliases='', description='uncharacterized LOC105377986', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125645673, end_position_on_the_genomic_accession=125749208, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377987, CurrentID=0, Status='live', Symbol='LOC105377987', Aliases='', description='uncharacterized LOC105377987', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=125916320, end_position_on_the_genomic_accession=125919121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377988, CurrentID=0, Status='live', Symbol='LOC105377988', Aliases='', description='uncharacterized LOC105377988', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127119975, end_position_on_the_genomic_accession=127180309, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105377989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377989, CurrentID=0, Status='live', Symbol='LOC105377989', Aliases='', description='uncharacterized LOC105377989', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126998850, end_position_on_the_genomic_accession=127119967, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105377990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377990, CurrentID=0, Status='live', Symbol='LOC105377990', Aliases='', description='uncharacterized LOC105377990', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126877676, end_position_on_the_genomic_accession=126943596, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105377991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377991, CurrentID=0, Status='live', Symbol='LOC105377991', Aliases='', description='uncharacterized LOC105377991', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126874058, end_position_on_the_genomic_accession=126882804, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  389320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389320, CurrentID=0, Status='live', Symbol='TEX43', Aliases='C5orf48, Tseg7', description='testis expressed 43', other_designations='testis specific expressed gene 7', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126631722, end_position_on_the_genomic_accession=126636282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105377993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377993, CurrentID=0, Status='live', Symbol='LOC105377993', Aliases='', description='uncharacterized LOC105377993', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=126692611, end_position_on_the_genomic_accession=126696228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389322, CurrentID=0, Status='live', Symbol='HNRNPKP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein K pseudogene 1', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127511462, end_position_on_the_genomic_accession=127512840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105377995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377995, CurrentID=0, Status='live', Symbol='LOC105377995', Aliases='', description='uncharacterized LOC105377995', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127599426, end_position_on_the_genomic_accession=127608188, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105377996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377996, CurrentID=0, Status='live', Symbol='LOC105377996', Aliases='', description='uncharacterized LOC105377996', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127608338, end_position_on_the_genomic_accession=127632311, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105377997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377997, CurrentID=0, Status='live', Symbol='LOC105377997', Aliases='', description='uncharacterized LOC105377997', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127663887, end_position_on_the_genomic_accession=127684761, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105377998 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377998, CurrentID=0, Status='live', Symbol='LOC105377998', Aliases='', description='uncharacterized LOC105377998', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=128648538, end_position_on_the_genomic_accession=128702758, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105377999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105377999, CurrentID=0, Status='live', Symbol='LOC105377999', Aliases='', description='uncharacterized LOC105377999', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=129855628, end_position_on_the_genomic_accession=129947895, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378000, CurrentID=0, Status='live', Symbol='LOC105378000', Aliases='', description='uncharacterized LOC105378000', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130120122, end_position_on_the_genomic_accession=130121326, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378001 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378001, CurrentID=0, Status='live', Symbol='LOC105378001', Aliases='', description='uncharacterized LOC105378001', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130687613, end_position_on_the_genomic_accession=130697583, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378002, CurrentID=0, Status='live', Symbol='LOC105378002', Aliases='', description='uncharacterized LOC105378002', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130719230, end_position_on_the_genomic_accession=130723221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378003, CurrentID=0, Status='live', Symbol='LOC105378003', Aliases='', description='uncharacterized LOC105378003', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=130741373, end_position_on_the_genomic_accession=130770750, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389332, CurrentID=0, Status='live', Symbol='LOC389332', Aliases='', description='uncharacterized LOC389332', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136191468, end_position_on_the_genomic_accession=136193163, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  389333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389333, CurrentID=0, Status='live', Symbol='PROB1', Aliases='C5orf65', description='proline-rich basic protein 1', other_designations='weakly similar to basic proline-rich protein', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139391946, end_position_on_the_genomic_accession=139395196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378006, CurrentID=0, Status='live', Symbol='LOC105378006', Aliases='', description='uncharacterized LOC105378006', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132074859, end_position_on_the_genomic_accession=132077657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378007, CurrentID=0, Status='live', Symbol='LOC105378007', Aliases='', description='uncharacterized LOC105378007', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=132876632, end_position_on_the_genomic_accession=132920124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389336, CurrentID=0, Status='live', Symbol='C5orf46', Aliases='SSSP1', description='chromosome 5 open reading frame 46', other_designations='CTC-327F10.5|skin and saliva secreted protein 1', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147885908, end_position_on_the_genomic_accession=147906589, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  389337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389337, CurrentID=0, Status='live', Symbol='ARHGEF37', Aliases='', description='Rho guanine nucleotide exchange factor 37', other_designations='Rho guanine nucleotide exchange factor (GEF) 37', map_location='5q32', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149551632, end_position_on_the_genomic_accession=149634968, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  105378010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378010, CurrentID=0, Status='live', Symbol='LOC105378010', Aliases='', description='uncharacterized LOC105378010', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135106845, end_position_on_the_genomic_accession=135110423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378011, CurrentID=0, Status='live', Symbol='LOC105378011', Aliases='', description='uncharacterized LOC105378011', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135195914, end_position_on_the_genomic_accession=135236190, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378012, CurrentID=0, Status='live', Symbol='LOC105378012', Aliases='', description='uncharacterized LOC105378012', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135796399, end_position_on_the_genomic_accession=135805073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378013, CurrentID=0, Status='live', Symbol='LOC105378013', Aliases='', description='uncharacterized LOC105378013', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=135628897, end_position_on_the_genomic_accession=135634680, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389342, CurrentID=0, Status='live', Symbol='RPL10P9', Aliases='RPL10_1_637', description='ribosomal protein L10 pseudogene 9', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=168616329, end_position_on_the_genomic_accession=168617055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378015, CurrentID=0, Status='live', Symbol='LOC105378015', Aliases='', description='uncharacterized LOC105378015', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136550421, end_position_on_the_genomic_accession=136552602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378016 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378016, CurrentID=0, Status='live', Symbol='LOC105378016', Aliases='', description='uncharacterized LOC105378016', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=136756569, end_position_on_the_genomic_accession=136765538, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378017, CurrentID=0, Status='live', Symbol='LOC105378017', Aliases='', description='uncharacterized LOC105378017', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137305772, end_position_on_the_genomic_accession=137314115, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378018, CurrentID=0, Status='live', Symbol='LOC105378018', Aliases='', description='uncharacterized LOC105378018', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137664066, end_position_on_the_genomic_accession=137673295, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378019, CurrentID=0, Status='live', Symbol='LOC105378019', Aliases='', description='uncharacterized LOC105378019', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137824575, end_position_on_the_genomic_accession=137860852, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378020, CurrentID=0, Status='live', Symbol='LOC105378020', Aliases='', description='uncharacterized LOC105378020', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137945042, end_position_on_the_genomic_accession=137957661, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378021, CurrentID=0, Status='live', Symbol='LOC105378021', Aliases='', description='uncharacterized LOC105378021', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138015543, end_position_on_the_genomic_accession=138034865, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378022, CurrentID=0, Status='live', Symbol='LOC105378022', Aliases='', description='uncharacterized LOC105378022', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=138830570, end_position_on_the_genomic_accession=138833614, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  283345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283345, CurrentID=0, Status='live', Symbol='RPL13P5', Aliases='RPL13-2, RPL13L, RPL13_4_1199, RRPL13L', description='ribosomal protein L13 pseudogene 5', other_designations='RPL13-2 pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6883981, end_position_on_the_genomic_accession=6884614, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100418754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418754, CurrentID=0, Status='live', Symbol='CLUHP6', Aliases='', description='clustered mitochondria (cluA/CLU1) homolog pseudogene 6', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32196026, end_position_on_the_genomic_accession=32196966, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378025, CurrentID=0, Status='live', Symbol='LOC105378025', Aliases='', description='uncharacterized LOC105378025', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140140612, end_position_on_the_genomic_accession=140148171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378026, CurrentID=0, Status='live', Symbol='LOC105378026', Aliases='', description='uncharacterized LOC105378026', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140610366, end_position_on_the_genomic_accession=140663992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378027, CurrentID=0, Status='live', Symbol='LOC105378027', Aliases='', description='uncharacterized LOC105378027', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140539233, end_position_on_the_genomic_accession=140785751, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105378028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378028, CurrentID=0, Status='live', Symbol='LOC105378028', Aliases='', description='uncharacterized LOC105378028', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=140866466, end_position_on_the_genomic_accession=140880595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  152274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152274, CurrentID=0, Status='live', Symbol='LOC152274', Aliases='', description='uncharacterized LOC152274', other_designations='', map_location='3p25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=14802894, end_position_on_the_genomic_accession=14811414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378030, CurrentID=0, Status='live', Symbol='LOC105378030', Aliases='', description='uncharacterized LOC105378030', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=141752985, end_position_on_the_genomic_accession=141798153, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378031, CurrentID=0, Status='live', Symbol='LOC105378031', Aliases='', description='uncharacterized LOC105378031', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=141850131, end_position_on_the_genomic_accession=142030422, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379802, CurrentID=0, Status='live', Symbol='LOC105379802', Aliases='', description='uncharacterized LOC105379802', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378033, CurrentID=0, Status='live', Symbol='LOC105378033', Aliases='', description='uncharacterized LOC105378033', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143115213, end_position_on_the_genomic_accession=143165071, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389362, CurrentID=0, Status='live', Symbol='PSMG4', Aliases='C6orf86, PAC4, bA506K6.2', description='proteasome (prosome, macropain) assembly chaperone 4', other_designations='PAC-4|hPAC4', map_location='6p25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=3254449, end_position_on_the_genomic_accession=3303019, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378035, CurrentID=0, Status='live', Symbol='LOC105378035', Aliases='', description='uncharacterized LOC105378035', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=143513092, end_position_on_the_genomic_accession=143534580, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378036, CurrentID=0, Status='live', Symbol='LOC105378036', Aliases='', description='uncharacterized LOC105378036', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=144110171, end_position_on_the_genomic_accession=144125408, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378037, CurrentID=0, Status='live', Symbol='LOC105378037', Aliases='', description='uncharacterized LOC105378037', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145138119, end_position_on_the_genomic_accession=145154520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378038, CurrentID=0, Status='live', Symbol='LOC105378038', Aliases='', description='uncharacterized LOC105378038', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145744806, end_position_on_the_genomic_accession=145750461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378039, CurrentID=0, Status='live', Symbol='LOC105378039', Aliases='', description='uncharacterized LOC105378039', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=145753806, end_position_on_the_genomic_accession=145767491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378040, CurrentID=0, Status='live', Symbol='LOC105378040', Aliases='', description='uncharacterized LOC105378040', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146731004, end_position_on_the_genomic_accession=146771347, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378041, CurrentID=0, Status='live', Symbol='LOC105378041', Aliases='', description='uncharacterized LOC105378041', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146720751, end_position_on_the_genomic_accession=146730759, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378042, CurrentID=0, Status='live', Symbol='LOC105378042', Aliases='', description='uncharacterized LOC105378042', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=146857697, end_position_on_the_genomic_accession=146863335, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378043, CurrentID=0, Status='live', Symbol='LOC105378043', Aliases='', description='uncharacterized LOC105378043', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=147431729, end_position_on_the_genomic_accession=147440077, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378044, CurrentID=0, Status='live', Symbol='LOC105378044', Aliases='', description='uncharacterized LOC105378044', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=147740975, end_position_on_the_genomic_accession=147743297, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378045, CurrentID=0, Status='live', Symbol='LOC105378045', Aliases='', description='uncharacterized LOC105378045', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=148267749, end_position_on_the_genomic_accession=148272345, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378046, CurrentID=0, Status='live', Symbol='LOC105378046', Aliases='', description='uncharacterized LOC105378046', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149049948, end_position_on_the_genomic_accession=149053415, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378047, CurrentID=0, Status='live', Symbol='LOC105378047', Aliases='', description='uncharacterized LOC105378047', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149027229, end_position_on_the_genomic_accession=149032754, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389376, CurrentID=0, Status='live', Symbol='SFTA2', Aliases='GSGL541, SFTPG, SP-G, UNQ541', description='surfactant associated 2', other_designations='surfactant associated protein G|surfactant-associated protein G', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=30931350, end_position_on_the_genomic_accession=30932175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378049, CurrentID=0, Status='live', Symbol='LOC105378049', Aliases='', description='uncharacterized LOC105378049', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149243282, end_position_on_the_genomic_accession=149257567, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378050 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378050, CurrentID=0, Status='live', Symbol='LOC105378050', Aliases='', description='uncharacterized LOC105378050', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149685683, end_position_on_the_genomic_accession=149691408, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378051, CurrentID=0, Status='live', Symbol='LOC105378051', Aliases='', description='multimerin-2-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149884511, end_position_on_the_genomic_accession=149905461, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378052 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378052, CurrentID=0, Status='live', Symbol='LOC105378052', Aliases='', description='uncharacterized LOC105378052', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=149934490, end_position_on_the_genomic_accession=149936324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378053, CurrentID=0, Status='live', Symbol='LOC105378053', Aliases='', description='uncharacterized LOC105378053', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150079413, end_position_on_the_genomic_accession=150119617, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378054, CurrentID=0, Status='live', Symbol='LOC105378054', Aliases='', description='uncharacterized LOC105378054', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150269744, end_position_on_the_genomic_accession=150306111, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389383, CurrentID=0, Status='live', Symbol='CLPSL2', Aliases='AAAL3045, C6orf126, UNQ3045, dJ510O8.5', description='colipase like 2', other_designations='colipase-like 2|colipase-like protein C6orf126', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=35776594, end_position_on_the_genomic_accession=35779552, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389384, CurrentID=0, Status='live', Symbol='C6orf222', Aliases='', description='chromosome 6 open reading frame 222', other_designations='', map_location='6p21.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36315757, end_position_on_the_genomic_accession=36336977, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105378057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378057, CurrentID=0, Status='live', Symbol='LOC105378057', Aliases='', description='uncharacterized LOC105378057', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151088103, end_position_on_the_genomic_accession=151151137, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389386, CurrentID=0, Status='live', Symbol='LAP3P2', Aliases='', description='leucine aminopeptidase 3 pseudogene 2', other_designations='', map_location='6p21.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=36673750, end_position_on_the_genomic_accession=36675445, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378059, CurrentID=0, Status='live', Symbol='LOC105378059', Aliases='', description='uncharacterized LOC105378059', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152061131, end_position_on_the_genomic_accession=152088443, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378060, CurrentID=0, Status='live', Symbol='LOC105378060', Aliases='', description='uncharacterized LOC105378060', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=151747959, end_position_on_the_genomic_accession=151764293, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378061, CurrentID=0, Status='live', Symbol='LOC105378061', Aliases='', description='uncharacterized LOC105378061', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152657976, end_position_on_the_genomic_accession=152687915, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378062, CurrentID=0, Status='live', Symbol='LOC105378062', Aliases='', description='uncharacterized LOC105378062', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152863579, end_position_on_the_genomic_accession=152871493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127247, CurrentID=0, Status='live', Symbol='ASB17', Aliases='Asb-17', description='ankyrin repeat and SOCS box containing 17', other_designations='ankyrin repeat and SOCS box-containing 17', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=75918873, end_position_on_the_genomic_accession=75932433, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378064, CurrentID=0, Status='live', Symbol='LOC105378064', Aliases='', description='uncharacterized LOC105378064', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=152838426, end_position_on_the_genomic_accession=152845806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929688, CurrentID=0, Status='live', Symbol='LINC01251', Aliases='', description='long intergenic non-protein coding RNA 1251', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=33732975, end_position_on_the_genomic_accession=33738416, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378066, CurrentID=0, Status='live', Symbol='LOC105378066', Aliases='', description='uncharacterized LOC105378066', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153304632, end_position_on_the_genomic_accession=153427337, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  389395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389395, CurrentID=0, Status='live', Symbol='HNRNPA3P4', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 4', other_designations='', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=48148498, end_position_on_the_genomic_accession=48150045, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389396, CurrentID=0, Status='live', Symbol='GLYATL3', Aliases='C6orf140, bA28H17.2', description='glycine-N-acyltransferase like 3', other_designations='glycine-N-acyltransferase-like 3', map_location='6p12.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=49499958, end_position_on_the_genomic_accession=49528064, orientation='plus', exon_count=6, OMIM=614763, no_hdr0=''),
  127253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127253, CurrentID=0, Status='live', Symbol='TYW3', Aliases='C1orf171', description='tRNA-yW synthesizing protein 3 homolog', other_designations='tRNA(Phe) 7-((3-amino-3-carboxypropyl)-4-demethylwyosine(37)-N(4))-methyltransferase|tRNA-yW-synthesizing protein 3', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74733143, end_position_on_the_genomic_accession=74766677, orientation='plus', exon_count=6, OMIM=611245, no_hdr0=''),
  127254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127254, CurrentID=0, Status='live', Symbol='ERICH3', Aliases='C1orf173', description='glutamate rich 3', other_designations='RP11-653A5.1|glutamate-rich 3', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74568111, end_position_on_the_genomic_accession=74673738, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  127255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127255, CurrentID=0, Status='live', Symbol='LRRIQ3', Aliases='LRRC44', description='leucine-rich repeats and IQ motif containing 3', other_designations='leucine rich repeat containing 44|leucine-rich repeat-containing protein 44', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=74026015, end_position_on_the_genomic_accession=74198201, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  389400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389400, CurrentID=0, Status='live', Symbol='GFRAL', Aliases='C6orf144, GRAL, UNQ9356, bA360D14.1', description='GDNF family receptor alpha like', other_designations='IVFI9356', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55327469, end_position_on_the_genomic_accession=55402493, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105378073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378073, CurrentID=0, Status='live', Symbol='LOC105378073', Aliases='', description='uncharacterized LOC105378073', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156498848, end_position_on_the_genomic_accession=156501135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378074, CurrentID=0, Status='live', Symbol='LOC105378074', Aliases='', description='uncharacterized LOC105378074', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=156776114, end_position_on_the_genomic_accession=156778422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378075, CurrentID=0, Status='live', Symbol='LOC105378075', Aliases='', description='uncharacterized LOC105378075', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157240567, end_position_on_the_genomic_accession=157253516, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389404, CurrentID=0, Status='live', Symbol='RPL9P18', Aliases='RPL9_6_709', description='ribosomal protein L9 pseudogene 18', other_designations='', map_location='6q12', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=63615736, end_position_on_the_genomic_accession=63616393, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378077, CurrentID=0, Status='live', Symbol='LOC105378077', Aliases='', description='uncharacterized LOC105378077', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=157558231, end_position_on_the_genomic_accession=157561455, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127262, CurrentID=0, Status='live', Symbol='TPRG1L', Aliases='FAM79A, h-mover, mover', description='tumor protein p63 regulated 1-like', other_designations='RP11-46F15.3|family with sequence similarity 79, member A|mossy fiber terminal-associated vertebrate-specific presynaptic protein', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3624992, end_position_on_the_genomic_accession=3630131, orientation='plus', exon_count=5, OMIM=611460, no_hdr0=''),
  105378079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378079, CurrentID=0, Status='live', Symbol='LOC105378079', Aliases='', description='uncharacterized LOC105378079', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158071801, end_position_on_the_genomic_accession=158074119, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378080, CurrentID=0, Status='live', Symbol='LOC105378080', Aliases='', description='uncharacterized LOC105378080', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=158858080, end_position_on_the_genomic_accession=158864498, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378081, CurrentID=0, Status='live', Symbol='LOC105378081', Aliases='', description='uncharacterized LOC105378081', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159051766, end_position_on_the_genomic_accession=159121467, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378082, CurrentID=0, Status='live', Symbol='LOC105378082', Aliases='', description='uncharacterized LOC105378082', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159042252, end_position_on_the_genomic_accession=159046703, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378083, CurrentID=0, Status='live', Symbol='LOC105378083', Aliases='', description='uncharacterized LOC105378083', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159000329, end_position_on_the_genomic_accession=159042223, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378084, CurrentID=0, Status='live', Symbol='LOC105378084', Aliases='', description='uncharacterized LOC105378084', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159375331, end_position_on_the_genomic_accession=159383327, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378085 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378085, CurrentID=0, Status='live', Symbol='LOC105378085', Aliases='', description='uncharacterized LOC105378085', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159586953, end_position_on_the_genomic_accession=159595202, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378086, CurrentID=0, Status='live', Symbol='LOC105378086', Aliases='', description='uncharacterized LOC105378086', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159615418, end_position_on_the_genomic_accession=159618776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378087 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378087, CurrentID=0, Status='live', Symbol='LOC105378087', Aliases='', description='uncharacterized LOC105378087', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=159917041, end_position_on_the_genomic_accession=159919259, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389416, CurrentID=0, Status='live', Symbol='BDH2P1', Aliases='DHRS6P1, bA758C21.1', description='3-hydroxybutyrate dehydrogenase, type 2 pseudogene 1', other_designations='dehydrogenase/reductase (SDR family) member 6 pseudogene 1', map_location='6q16.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=99174671, end_position_on_the_genomic_accession=99177570, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378089, CurrentID=0, Status='live', Symbol='LOC105378089', Aliases='', description='uncharacterized LOC105378089', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160485910, end_position_on_the_genomic_accession=160506798, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378090 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378090, CurrentID=0, Status='live', Symbol='LOC105378090', Aliases='', description='uncharacterized LOC105378090', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160369745, end_position_on_the_genomic_accession=160391996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378091 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378091, CurrentID=0, Status='live', Symbol='LOC105378091', Aliases='', description='uncharacterized LOC105378091', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160776300, end_position_on_the_genomic_accession=160807169, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378092, CurrentID=0, Status='live', Symbol='LOC105378092', Aliases='', description='uncharacterized LOC105378092', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=160872888, end_position_on_the_genomic_accession=160895704, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389421, CurrentID=0, Status='live', Symbol='LIN28B', Aliases='CSDD2', description='lin-28 homolog B (C. elegans)', other_designations='Lin-28.2|lin-28B', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=104936999, end_position_on_the_genomic_accession=105083332, orientation='plus', exon_count=7, OMIM=611044, no_hdr0=''),
  389422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389422, CurrentID=0, Status='live', Symbol='C6orf183', Aliases='bA487F23.3', description='chromosome 6 open reading frame 183', other_designations='', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.11', start_position_on_the_genomic_accession=109505723, end_position_on_the_genomic_accession=109551438, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105378095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378095, CurrentID=0, Status='live', Symbol='LOC105378095', Aliases='', description='uncharacterized LOC105378095', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163209279, end_position_on_the_genomic_accession=163211640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389424, CurrentID=0, Status='live', Symbol='PA2G4P5', Aliases='PA2G4L2', description='proliferation-associated 2G4 pseudogene 5', other_designations='proliferation-associated 2G4, 38kDa pseudogene', map_location='6q21', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=112615690, end_position_on_the_genomic_accession=112618030, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  127281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127281, CurrentID=0, Status='live', Symbol='FAM213B', Aliases='C1orf93', description='family with sequence similarity 213 member B', other_designations='family with sequence similarity 213, member B|prostamide/PG F synthase|prostamide/PGF synthase', map_location='1p36.32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2586460, end_position_on_the_genomic_accession=2591469, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378098, CurrentID=0, Status='live', Symbol='LOC105378098', Aliases='', description='uncharacterized LOC105378098', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=161470977, end_position_on_the_genomic_accession=161484045, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378099, CurrentID=0, Status='live', Symbol='LOC105378099', Aliases='', description='uncharacterized LOC105378099', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163368306, end_position_on_the_genomic_accession=163371187, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389428, CurrentID=0, Status='live', Symbol='RPL5P18', Aliases='RPL5_11_750', description='ribosomal protein L5 pseudogene 18', other_designations='', map_location='6q22.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=127362486, end_position_on_the_genomic_accession=127363457, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378101 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378101, CurrentID=0, Status='live', Symbol='LOC105378101', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163411037, end_position_on_the_genomic_accession=163414279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378102, CurrentID=0, Status='live', Symbol='LOC105378102', Aliases='', description='uncharacterized LOC105378102', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163919543, end_position_on_the_genomic_accession=164058966, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378103 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378103, CurrentID=0, Status='live', Symbol='LOC105378103', Aliases='', description='uncharacterized LOC105378103', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163903537, end_position_on_the_genomic_accession=163919119, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389432, CurrentID=0, Status='live', Symbol='SAMD5', Aliases='dJ875H10.1', description='sterile alpha motif domain containing 5', other_designations='SAM domain containing 1|SAM domain-containing protein 5', map_location='6q24.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=147508648, end_position_on_the_genomic_accession=147737541, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378105 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378105, CurrentID=0, Status='live', Symbol='LOC105378105', Aliases='', description='uncharacterized LOC105378105', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=163711976, end_position_on_the_genomic_accession=163714377, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  389434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389434, CurrentID=0, Status='live', Symbol='IYD', Aliases='C6orf71, DEHAL1, TDH4, dJ422F24.1', description='iodotyrosine deiodinase', other_designations='IYD-1|iodotyrosine dehalogenase 1', map_location='6q25.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=150368861, end_position_on_the_genomic_accession=150404629, orientation='plus', exon_count=7, OMIM=612025, no_hdr0=''),
  389435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389435, CurrentID=0, Status='live', Symbol='RPL27AP6', Aliases='RPL27A_4_770', description='ribosomal protein L27a pseudogene 6', other_designations='', map_location='6q25.2', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=153282249, end_position_on_the_genomic_accession=153282754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378108, CurrentID=0, Status='live', Symbol='LOC105378108', Aliases='', description='uncharacterized LOC105378108', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164305347, end_position_on_the_genomic_accession=164337800, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378109, CurrentID=0, Status='live', Symbol='LOC105378109', Aliases='', description='uncharacterized LOC105378109', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164293408, end_position_on_the_genomic_accession=164300747, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  127294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127294, CurrentID=0, Status='live', Symbol='MYOM3', Aliases='', description='myomesin 3', other_designations='myomesin family, member 3', map_location='1p36.11', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=24056041, end_position_on_the_genomic_accession=24112175, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  105378111 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378111, CurrentID=0, Status='live', Symbol='MEAT6', Aliases='', description='melanoma-associated transcript 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164792575, end_position_on_the_genomic_accession=164827657, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378112, CurrentID=0, Status='live', Symbol='LOC105378112', Aliases='', description='uncharacterized LOC105378112', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=164927835, end_position_on_the_genomic_accession=164928940, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378113, CurrentID=0, Status='live', Symbol='LOC105378113', Aliases='', description='uncharacterized LOC105378113', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165309663, end_position_on_the_genomic_accession=165316728, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378114, CurrentID=0, Status='live', Symbol='LOC105378114', Aliases='', description='uncharacterized LOC105378114', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165773201, end_position_on_the_genomic_accession=165775945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378115, CurrentID=0, Status='live', Symbol='LOC105378115', Aliases='', description='uncharacterized LOC105378115', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165668730, end_position_on_the_genomic_accession=165676224, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378116, CurrentID=0, Status='live', Symbol='LOC105378116', Aliases='', description='uncharacterized LOC105378116', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=165408845, end_position_on_the_genomic_accession=165413654, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378117, CurrentID=0, Status='live', Symbol='LOC105378117', Aliases='', description='uncharacterized LOC105378117', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166025366, end_position_on_the_genomic_accession=166042646, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378118, CurrentID=0, Status='live', Symbol='LOC105378118', Aliases='', description='uncharacterized LOC105378118', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166275587, end_position_on_the_genomic_accession=166277085, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378119, CurrentID=0, Status='live', Symbol='LOC105378119', Aliases='', description='protein GVQW1-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166943096, end_position_on_the_genomic_accession=166946675, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378120, CurrentID=0, Status='live', Symbol='LOC105378120', Aliases='', description='uncharacterized LOC105378120', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166968040, end_position_on_the_genomic_accession=166974440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378121, CurrentID=0, Status='live', Symbol='LOC105378121', Aliases='', description='uncharacterized LOC105378121', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=166982651, end_position_on_the_genomic_accession=166988392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378122, CurrentID=0, Status='live', Symbol='LOC105378122', Aliases='', description='uncharacterized LOC105378122', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167112399, end_position_on_the_genomic_accession=167122594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378123, CurrentID=0, Status='live', Symbol='LOC105378123', Aliases='', description='uncharacterized LOC105378123', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167165525, end_position_on_the_genomic_accession=167170007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378124 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378124, CurrentID=0, Status='live', Symbol='LOC105378124', Aliases='', description='uncharacterized LOC105378124', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167146414, end_position_on_the_genomic_accession=167169128, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378125, CurrentID=0, Status='live', Symbol='LOC105378125', Aliases='', description='uncharacterized LOC105378125', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167385109, end_position_on_the_genomic_accession=167404984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378126, CurrentID=0, Status='live', Symbol='LOC105378126', Aliases='', description='uncharacterized LOC105378126', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167423235, end_position_on_the_genomic_accession=167436372, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378127, CurrentID=0, Status='live', Symbol='LOC105378127', Aliases='', description='uncharacterized LOC105378127', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167459262, end_position_on_the_genomic_accession=167463531, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  651600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651600, CurrentID=0, Status='live', Symbol='RPL36P16', Aliases='RPL36_7_1657', description='ribosomal protein L36 pseudogene 16', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41169961, end_position_on_the_genomic_accession=41170068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378129, CurrentID=0, Status='live', Symbol='LOC105378129', Aliases='', description='uncharacterized LOC105378129', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167544446, end_position_on_the_genomic_accession=167551392, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389458, CurrentID=0, Status='live', Symbol='RBAKDN', Aliases='', description='RBAK downstream neighbor (non-protein coding)', other_designations='', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5072060, end_position_on_the_genomic_accession=5073223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378131, CurrentID=0, Status='live', Symbol='LOC105378131', Aliases='', description='uncharacterized LOC105378131', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167574834, end_position_on_the_genomic_accession=167588832, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378132, CurrentID=0, Status='live', Symbol='LOC105378132', Aliases='', description='uncharacterized LOC105378132', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167686992, end_position_on_the_genomic_accession=167696379, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378133, CurrentID=0, Status='live', Symbol='LOC105378133', Aliases='', description='uncharacterized LOC105378133', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167677263, end_position_on_the_genomic_accession=167679526, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378134, CurrentID=0, Status='live', Symbol='LOC105378134', Aliases='', description='uncharacterized LOC105378134', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167815894, end_position_on_the_genomic_accession=167823247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378135, CurrentID=0, Status='live', Symbol='LOC105378135', Aliases='', description='uncharacterized LOC105378135', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168045105, end_position_on_the_genomic_accession=168046162, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378136, CurrentID=0, Status='live', Symbol='LOC105378136', Aliases='', description='uncharacterized LOC105378136', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168229765, end_position_on_the_genomic_accession=168231364, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378137, CurrentID=0, Status='live', Symbol='LOC105378137', Aliases='', description='uncharacterized LOC105378137', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168225280, end_position_on_the_genomic_accession=168227074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378138, CurrentID=0, Status='live', Symbol='LOC105378138', Aliases='', description='uncharacterized LOC105378138', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168298073, end_position_on_the_genomic_accession=168302153, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378139, CurrentID=0, Status='live', Symbol='LOC105378139', Aliases='', description='uncharacterized LOC105378139', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168333251, end_position_on_the_genomic_accession=168334523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378140, CurrentID=0, Status='live', Symbol='LOC105378140', Aliases='', description='uncharacterized LOC105378140', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168410452, end_position_on_the_genomic_accession=168413100, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378141, CurrentID=0, Status='live', Symbol='LOC105378141', Aliases='', description='uncharacterized LOC105378141', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168413354, end_position_on_the_genomic_accession=168421325, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378142, CurrentID=0, Status='live', Symbol='LOC105378142', Aliases='', description='uncharacterized LOC105378142', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168695000, end_position_on_the_genomic_accession=168697957, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378143, CurrentID=0, Status='live', Symbol='LOC105378143', Aliases='', description='uncharacterized LOC105378143', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168711148, end_position_on_the_genomic_accession=168716888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389472, CurrentID=0, Status='live', Symbol='RPS26P32', Aliases='RPS26, RPS26_13_792', description='ribosomal protein S26 pseudogene 32', other_designations='', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22773570, end_position_on_the_genomic_accession=22774017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389473, CurrentID=0, Status='live', Symbol='LOC389473', Aliases='', description='chromosome 5 open reading frame 13 pseudogene', other_designations='neuronal regeneration related protein homolog (rat) pseudogene', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23480583, end_position_on_the_genomic_accession=23482569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378146, CurrentID=0, Status='live', Symbol='LOC105378146', Aliases='', description='uncharacterized LOC105378146', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=168919702, end_position_on_the_genomic_accession=168964340, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378147, CurrentID=0, Status='live', Symbol='LOC105378147', Aliases='', description='uncharacterized LOC105378147', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169355736, end_position_on_the_genomic_accession=169396423, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378148, CurrentID=0, Status='live', Symbol='LOC105378148', Aliases='', description='uncharacterized LOC105378148', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169330544, end_position_on_the_genomic_accession=169340581, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378149, CurrentID=0, Status='live', Symbol='LOC105378149', Aliases='', description='uncharacterized LOC105378149', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=169899161, end_position_on_the_genomic_accession=169904735, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378150, CurrentID=0, Status='live', Symbol='LOC105378150', Aliases='', description='uncharacterized LOC105378150', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170023727, end_position_on_the_genomic_accession=170028797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378151, CurrentID=0, Status='live', Symbol='LOC105378151', Aliases='', description='uncharacterized LOC105378151', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170081100, end_position_on_the_genomic_accession=170126847, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378152, CurrentID=0, Status='live', Symbol='LOC105378152', Aliases='', description='uncharacterized LOC105378152', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170130197, end_position_on_the_genomic_accession=170142348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378153, CurrentID=0, Status='live', Symbol='LOC105378153', Aliases='', description='uncharacterized LOC105378153', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170153579, end_position_on_the_genomic_accession=170165670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378154, CurrentID=0, Status='live', Symbol='LOC105378154', Aliases='', description='uncharacterized LOC105378154', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170404831, end_position_on_the_genomic_accession=170412183, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378155, CurrentID=0, Status='live', Symbol='LOC105378155', Aliases='', description='uncharacterized LOC105378155', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170370495, end_position_on_the_genomic_accession=170394115, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  651628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651628, CurrentID=0, Status='live', Symbol='EEF1GP3', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 3', other_designations='', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=40596122, end_position_on_the_genomic_accession=40597519, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378157, CurrentID=0, Status='live', Symbol='LOC105378157', Aliases='', description='uncharacterized LOC105378157', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170469031, end_position_on_the_genomic_accession=170497372, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378158, CurrentID=0, Status='live', Symbol='LOC105378158', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=83708512, end_position_on_the_genomic_accession=83736441, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127343, CurrentID=0, Status='live', Symbol='DMBX1', Aliases='MBX, OTX3, PAXB', description='diencephalon/mesencephalon homeobox 1', other_designations='homeoprotein MBX|orthodenticle homolog 3|paired-like homeobox protein DMBX1', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46489837, end_position_on_the_genomic_accession=46514226, orientation='plus', exon_count=7, OMIM=607410, no_hdr0=''),
  105378160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378160, CurrentID=0, Status='live', Symbol='LOC105378160', Aliases='', description='centromere protein J-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167357356, end_position_on_the_genomic_accession=167371630, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378161, CurrentID=0, Status='live', Symbol='LOC105378161', Aliases='', description='uncharacterized LOC105378161', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=170461022, end_position_on_the_genomic_accession=170463114, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378162, CurrentID=0, Status='live', Symbol='LOC105378162', Aliases='', description='uncharacterized LOC105378162', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378163, CurrentID=0, Status='live', Symbol='LOC105378163', Aliases='', description='uncharacterized LOC105378163', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378164, CurrentID=0, Status='live', Symbol='LOC105378164', Aliases='', description='uncharacterized LOC105378164', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389493, CurrentID=0, Status='live', Symbol='NUPR2', Aliases='NUPR1L', description='nuclear protein 2, transcriptional regulator', other_designations='nuclear protein, transcriptional regulator, 1-like', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=56114681, end_position_on_the_genomic_accession=56116397, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378166, CurrentID=0, Status='live', Symbol='LOC105378166', Aliases='', description='uncharacterized LOC105378166', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378167, CurrentID=0, Status='live', Symbol='LOC105378167', Aliases='', description='vegetative cell wall protein gp1-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378168, CurrentID=0, Status='live', Symbol='LOC105378168', Aliases='', description='mucin-5AC-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378169, CurrentID=0, Status='live', Symbol='LOC105378169', Aliases='', description='uncharacterized LOC105378169', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378170, CurrentID=0, Status='live', Symbol='LOC105378170', Aliases='', description='uncharacterized LOC105378170', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378171, CurrentID=0, Status='live', Symbol='LOC105378171', Aliases='', description='uncharacterized LOC105378171', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19042918, end_position_on_the_genomic_accession=19051808, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  651644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651644, CurrentID=0, Status='live', Symbol='LOC651644', Aliases='', description='LIM and senescent cell antigen-like domains 2 pseudogene', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=36065073, end_position_on_the_genomic_accession=36065726, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378173 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378173, CurrentID=0, Status='live', Symbol='LOC105378173', Aliases='', description='uncharacterized LOC105378173', other_designations='CTD-2311B13.7|CTD-2314B22.1', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19365097, end_position_on_the_genomic_accession=19382340, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378174, CurrentID=0, Status='live', Symbol='LOC105378174', Aliases='', description='uncharacterized LOC105378174', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19348216, end_position_on_the_genomic_accession=19357099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378175, CurrentID=0, Status='live', Symbol='LOC105378175', Aliases='', description='uncharacterized LOC105378175', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21135383, end_position_on_the_genomic_accession=21135857, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378176 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378176, CurrentID=0, Status='live', Symbol='LOC105378176', Aliases='', description='proline-rich receptor-like protein kinase PERK2', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23298916, end_position_on_the_genomic_accession=23307276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378177, CurrentID=0, Status='live', Symbol='LOC105378177', Aliases='', description='uncharacterized LOC105378177', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=45855372, end_position_on_the_genomic_accession=45857042, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378178, CurrentID=0, Status='live', Symbol='LOC105378178', Aliases='', description='uncharacterized LOC105378178', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49055643, end_position_on_the_genomic_accession=49151927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378179, CurrentID=0, Status='live', Symbol='LOC105378179', Aliases='', description='uncharacterized LOC105378179', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=49860974, end_position_on_the_genomic_accession=49868143, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378180, CurrentID=0, Status='live', Symbol='LOC105378180', Aliases='', description='uncharacterized LOC105378180', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52069334, end_position_on_the_genomic_accession=52074144, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378181, CurrentID=0, Status='live', Symbol='LOC105378181', Aliases='', description='uncharacterized LOC105378181', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=52982577, end_position_on_the_genomic_accession=53024708, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378182 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378182, CurrentID=0, Status='live', Symbol='LOC105378182', Aliases='', description='uncharacterized LOC105378182', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=61113519, end_position_on_the_genomic_accession=61122877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378183, CurrentID=0, Status='live', Symbol='LOC105378183', Aliases='', description='uncharacterized LOC105378183', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=103317257, end_position_on_the_genomic_accession=103318539, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378184 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378184, CurrentID=0, Status='live', Symbol='LOC105378184', Aliases='', description='uncharacterized LOC105378184', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105754186, end_position_on_the_genomic_accession=105760351, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378185, CurrentID=0, Status='live', Symbol='LOC105378185', Aliases='', description='uncharacterized LOC105378185', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105680283, end_position_on_the_genomic_accession=105681852, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100271558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271558, CurrentID=0, Status='live', Symbol='RPS15AP23', Aliases='RPS15A_12_849', description='ribosomal protein S15a pseudogene 23', other_designations='', map_location='7q31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=133315009, end_position_on_the_genomic_accession=133315411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378187, CurrentID=0, Status='live', Symbol='LOC105378187', Aliases='', description='uncharacterized LOC105378187', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105655024, end_position_on_the_genomic_accession=105663100, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378188, CurrentID=0, Status='live', Symbol='LOC105378188', Aliases='', description='uncharacterized LOC105378188', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=105609312, end_position_on_the_genomic_accession=105619188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378189, CurrentID=0, Status='live', Symbol='LOC105378189', Aliases='', description='uncharacterized LOC105378189', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60575033, end_position_on_the_genomic_accession=60643945, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378190, CurrentID=0, Status='live', Symbol='LOC105378190', Aliases='', description='uncharacterized LOC105378190', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=104216110, end_position_on_the_genomic_accession=104218144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422381, CurrentID=0, Status='live', Symbol='PPP1R10P1', Aliases='PPP1R10P', description='protein phosphatase 1 regulatory subunit 10 pseudogene 1', other_designations='protein phosphatase 1, regulatory (inhibitor) subunit 10 pseudogene 1|protein phosphatase 1, regulatory subunit 10 pseudogene 1', map_location='11q24.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=130789229, end_position_on_the_genomic_accession=130790875, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378192, CurrentID=0, Status='live', Symbol='LOC105378192', Aliases='', description='uncharacterized LOC105378192', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112862586, end_position_on_the_genomic_accession=112865576, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378193, CurrentID=0, Status='live', Symbol='LOC105378193', Aliases='', description='serine/arginine repetitive matrix protein 5-like', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=112013841, end_position_on_the_genomic_accession=112036593, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378194, CurrentID=0, Status='live', Symbol='LOC105378194', Aliases='', description='uncharacterized LOC105378194', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378195, CurrentID=0, Status='live', Symbol='LOC105378195', Aliases='', description='uncharacterized LOC105378195', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=139667824, end_position_on_the_genomic_accession=139672147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389524, CurrentID=0, Status='live', Symbol='GTF2IRD2B', Aliases='', description='GTF2I repeat domain containing 2B', other_designations='GTF2I repeat domain-containing protein 2B|GTF2IRD2 beta|general transcription factor 2 I repeat domain-containing 2 beta|transcription factor GTF2IRD2-beta', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75092396, end_position_on_the_genomic_accession=75149817, orientation='plus', exon_count=17, OMIM=608900, no_hdr0=''),
  101929710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929710, CurrentID=0, Status='live', Symbol='LOC101929710', Aliases='', description='uncharacterized LOC101929710', other_designations='CTD-2337A12.1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95962001, end_position_on_the_genomic_accession=96631085, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378198 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378198, CurrentID=0, Status='live', Symbol='LOC105378198', Aliases='', description='uncharacterized LOC105378198', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=140711460, end_position_on_the_genomic_accession=140733719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378199, CurrentID=0, Status='live', Symbol='LOC105378199', Aliases='', description='uncharacterized LOC105378199', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141260224, end_position_on_the_genomic_accession=141263663, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378200, CurrentID=0, Status='live', Symbol='LOC105378200', Aliases='', description='uncharacterized LOC105378200', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141576241, end_position_on_the_genomic_accession=141583212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127385, CurrentID=0, Status='live', Symbol='OR10J5', Aliases='OR1-28', description='olfactory receptor family 10 subfamily J member 5', other_designations='olfactory receptor OR1-28|olfactory receptor, family 10, subfamily J, member 5', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159535078, end_position_on_the_genomic_accession=159536007, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378202, CurrentID=0, Status='live', Symbol='LOC105378202', Aliases='', description='uncharacterized LOC105378202', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141822050, end_position_on_the_genomic_accession=141828787, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378203, CurrentID=0, Status='live', Symbol='LOC105378203', Aliases='', description='uncharacterized LOC105378203', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=141951597, end_position_on_the_genomic_accession=141955721, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378204, CurrentID=0, Status='live', Symbol='LOC105378204', Aliases='', description='uncharacterized LOC105378204', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142088384, end_position_on_the_genomic_accession=142103124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378205, CurrentID=0, Status='live', Symbol='LOC105378205', Aliases='', description='uncharacterized LOC105378205', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=142810780, end_position_on_the_genomic_accession=142821323, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378206, CurrentID=0, Status='live', Symbol='LOC105378206', Aliases='', description='uncharacterized LOC105378206', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143258639, end_position_on_the_genomic_accession=143285987, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127391, CurrentID=0, Status='live', Symbol='TMCO2', Aliases='dJ39G22.2', description='transmembrane and coiled-coil domains 2', other_designations='dJ39G22.2 (novel protein)', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40247901, end_position_on_the_genomic_accession=40251693, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378208, CurrentID=0, Status='live', Symbol='LOC105378208', Aliases='', description='uncharacterized LOC105378208', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=143531378, end_position_on_the_genomic_accession=143560079, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  389537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389537, CurrentID=0, Status='live', Symbol='OR7E38P', Aliases='OR7E76, OST127', description='olfactory receptor family 7 subfamily E member 38 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 38 pseudogene|olfactory receptor, family 7, subfamily E, member 76', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97966092, end_position_on_the_genomic_accession=97967077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389538, CurrentID=0, Status='live', Symbol='MGC72080', Aliases='', description='MGC72080 pseudogene', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97966596, end_position_on_the_genomic_accession=97972326, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  392603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392603, CurrentID=0, Status='live', Symbol='ZNF736P2Y', Aliases='ZNF736P2, ZNF736PY2', description='zinc finger protein 736 pseudogene 2, Y-linked', other_designations='zinc finger protein 736 pseudogene, Y-linked 2', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=24679737, end_position_on_the_genomic_accession=24684855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  127396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127396, CurrentID=0, Status='live', Symbol='ZNF684', Aliases='', description='zinc finger protein 684', other_designations='hypothetical protein MGC27466', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40531155, end_position_on_the_genomic_accession=40548169, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  389541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389541, CurrentID=0, Status='live', Symbol='LAMTOR4', Aliases='C7orf59', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 4', other_designations='UPF0539 protein C7orf59|late endosomal/lysosomal adaptor and MAPK and MTOR activator 4', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=100148907, end_position_on_the_genomic_accession=100154210, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378214, CurrentID=0, Status='live', Symbol='LOC105378214', Aliases='', description='uncharacterized LOC105378214', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146917931, end_position_on_the_genomic_accession=146923541, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378215, CurrentID=0, Status='live', Symbol='CTB-99A3.1', Aliases='', description='uncharacterized LOC105378215', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=146563161, end_position_on_the_genomic_accession=146621829, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378216, CurrentID=0, Status='live', Symbol='LOC105378216', Aliases='', description='uncharacterized LOC105378216', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147432428, end_position_on_the_genomic_accession=147438455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378217, CurrentID=0, Status='live', Symbol='LOC105378217', Aliases='', description='uncharacterized LOC105378217', other_designations='CTC-255N20.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147174510, end_position_on_the_genomic_accession=147234960, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378218, CurrentID=0, Status='live', Symbol='LOC105378218', Aliases='', description='uncharacterized LOC105378218', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=147721490, end_position_on_the_genomic_accession=147725809, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378219, CurrentID=0, Status='live', Symbol='LOC105378219', Aliases='', description='uncharacterized LOC105378219', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148147523, end_position_on_the_genomic_accession=148152080, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378220, CurrentID=0, Status='live', Symbol='LOC105378220', Aliases='', description='macrophage receptor MARCO-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=148238565, end_position_on_the_genomic_accession=148244034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389549, CurrentID=0, Status='live', Symbol='FEZF1', Aliases='FEZ, HH22, ZNF312B', description='FEZ family zinc finger 1', other_designations='zinc finger protein 312B', map_location='7q31.32', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=122301309, end_position_on_the_genomic_accession=122311853, orientation='minus', exon_count=7, OMIM=613301, no_hdr0=''),
  127406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127406, CurrentID=0, Status='live', Symbol='RPSAP20', Aliases='RPSA_1_53', description='ribosomal protein SA pseudogene 20', other_designations='', map_location='1p32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56207472, end_position_on_the_genomic_accession=56208529, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378223, CurrentID=0, Status='live', Symbol='LOC105378223', Aliases='', description='uncharacterized LOC105378223', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149371301, end_position_on_the_genomic_accession=149375112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378224, CurrentID=0, Status='live', Symbol='LOC105378224', Aliases='', description='uncharacterized LOC105378224', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=149703277, end_position_on_the_genomic_accession=149721636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378225, CurrentID=0, Status='live', Symbol='LOC105378225', Aliases='', description='uncharacterized LOC105378225', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150334876, end_position_on_the_genomic_accession=150340649, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378226, CurrentID=0, Status='live', Symbol='LOC105378226', Aliases='', description='uncharacterized LOC105378226', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150353282, end_position_on_the_genomic_accession=150355670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378227, CurrentID=0, Status='live', Symbol='LOC105378227', Aliases='', description='uncharacterized LOC105378227', other_designations='CTC-345K18.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150666003, end_position_on_the_genomic_accession=150672340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378228, CurrentID=0, Status='live', Symbol='LOC105378228', Aliases='', description='uncharacterized LOC105378228', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=150985576, end_position_on_the_genomic_accession=151020483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378229, CurrentID=0, Status='live', Symbol='LOC105378229', Aliases='', description='uncharacterized LOC105378229', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151173630, end_position_on_the_genomic_accession=151175458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389558, CurrentID=0, Status='live', Symbol='FAM180A', Aliases='UNQ1940', description='family with sequence similarity 180 member A', other_designations='HWKM1940|family with sequence similarity 180, member A', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=135729598, end_position_on_the_genomic_accession=135748867, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378231, CurrentID=0, Status='live', Symbol='LOC105378231', Aliases='', description='uncharacterized LOC105378231', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151637915, end_position_on_the_genomic_accession=151655449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378232, CurrentID=0, Status='live', Symbol='LOC105378232', Aliases='', description='uncharacterized LOC105378232', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151626346, end_position_on_the_genomic_accession=151628832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378233 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378233, CurrentID=0, Status='live', Symbol='LOC105378233', Aliases='', description='uncharacterized LOC105378233', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151567254, end_position_on_the_genomic_accession=151583152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378234, CurrentID=0, Status='live', Symbol='LOC105378234', Aliases='', description='uncharacterized LOC105378234', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151352583, end_position_on_the_genomic_accession=151436929, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378235, CurrentID=0, Status='live', Symbol='LOC105378235', Aliases='', description='uncharacterized LOC105378235', other_designations='CTB-113P19.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=151752500, end_position_on_the_genomic_accession=151758528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378236, CurrentID=0, Status='live', Symbol='LOC105378236', Aliases='', description='uncharacterized LOC105378236', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152294645, end_position_on_the_genomic_accession=152329868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378237, CurrentID=0, Status='live', Symbol='LOC105378237', Aliases='', description='uncharacterized LOC105378237', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=152337668, end_position_on_the_genomic_accession=152378423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378238, CurrentID=0, Status='live', Symbol='LOC105378238', Aliases='', description='uncharacterized LOC105378238', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=153861919, end_position_on_the_genomic_accession=153977134, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101929717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929717, CurrentID=0, Status='live', Symbol='LOC101929717', Aliases='', description='uncharacterized LOC101929717', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=113142350, end_position_on_the_genomic_accession=113167731, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378240, CurrentID=0, Status='live', Symbol='LOC105378240', Aliases='', description='uncharacterized LOC105378240', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378241, CurrentID=0, Status='live', Symbol='LOC105378241', Aliases='', description='uncharacterized LOC105378241', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  651714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651714, CurrentID=0, Status='live', Symbol='MUC20P1', Aliases='', description='mucin 20, cell surface associated pseudogene 1', other_designations='', map_location='3q29', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=195618625, end_position_on_the_genomic_accession=195620237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378243, CurrentID=0, Status='live', Symbol='LOC105378243', Aliases='', description='uncharacterized LOC105378243', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  127428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127428, CurrentID=0, Status='live', Symbol='TCEANC2', Aliases='C1orf83', description='transcription elongation factor A (SII) N-terminal and central domain containing 2', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54053572, end_position_on_the_genomic_accession=54112519, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378245, CurrentID=0, Status='live', Symbol='LOC105378245', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378246, CurrentID=0, Status='live', Symbol='LOC105378246', Aliases='', description='uncharacterized LOC105378246', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=39746008, end_position_on_the_genomic_accession=39802611, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378247, CurrentID=0, Status='live', Symbol='LOC105378247', Aliases='', description='uncharacterized LOC105378247', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=41979460, end_position_on_the_genomic_accession=41989168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378248 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378248, CurrentID=0, Status='live', Symbol='LOC105378248', Aliases='', description='uncharacterized LOC105378248', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=46269111, end_position_on_the_genomic_accession=46275520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378249, CurrentID=0, Status='live', Symbol='LOC105378249', Aliases='', description='uncharacterized LOC105378249', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=51839120, end_position_on_the_genomic_accession=51846959, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378250 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378250, CurrentID=0, Status='live', Symbol='LOC105378250', Aliases='', description='uncharacterized LOC105378250', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=53754357, end_position_on_the_genomic_accession=53762272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127435, CurrentID=0, Status='live', Symbol='PODN', Aliases='PCAN, SLRR5A', description='podocan', other_designations='podocan proteoglycan', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53062052, end_position_on_the_genomic_accession=53085502, orientation='plus', exon_count=14, OMIM=608661, no_hdr0=''),
  105378252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378252, CurrentID=0, Status='live', Symbol='LOC105378252', Aliases='', description='uncharacterized LOC105378252', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80624016, end_position_on_the_genomic_accession=80707218, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378253, CurrentID=0, Status='live', Symbol='LOC105378253', Aliases='', description='uncharacterized LOC105378253', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80497241, end_position_on_the_genomic_accession=80514070, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378254, CurrentID=0, Status='live', Symbol='LOC105378254', Aliases='', description='tenascin-N-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=80424259, end_position_on_the_genomic_accession=80435269, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378255, CurrentID=0, Status='live', Symbol='LOC105378255', Aliases='', description='uncharacterized LOC105378255', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=89222379, end_position_on_the_genomic_accession=89230010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378256, CurrentID=0, Status='live', Symbol='LOC105378256', Aliases='', description='uncharacterized LOC105378256', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108983755, end_position_on_the_genomic_accession=109007750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378257, CurrentID=0, Status='live', Symbol='LOC105378257', Aliases='', description='uncharacterized LOC105378257', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=118381602, end_position_on_the_genomic_accession=118383725, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378258 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378258, CurrentID=0, Status='live', Symbol='LOC105378258', Aliases='', description='uncharacterized LOC105378258', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=121059290, end_position_on_the_genomic_accession=121074209, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378259, CurrentID=0, Status='live', Symbol='LOC105378259', Aliases='', description='uncharacterized LOC105378259', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=123384221, end_position_on_the_genomic_accession=123385447, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378260, CurrentID=0, Status='live', Symbol='LOC105378260', Aliases='', description='histone-lysine N-methyltransferase 2C', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10462907, end_position_on_the_genomic_accession=10500830, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105378261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378261, CurrentID=0, Status='live', Symbol='LOC105378261', Aliases='', description='uncharacterized LOC105378261', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378262, CurrentID=0, Status='live', Symbol='LOC105378262', Aliases='', description='uncharacterized LOC105378262', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378263, CurrentID=0, Status='live', Symbol='LOC105378263', Aliases='', description='uncharacterized LOC105378263', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378264, CurrentID=0, Status='live', Symbol='LOC105378264', Aliases='', description='uncharacterized LOC105378264', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378265, CurrentID=0, Status='live', Symbol='LOC105378265', Aliases='', description='uncharacterized LOC105378265', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378266, CurrentID=0, Status='live', Symbol='LOC105378266', Aliases='', description='forkhead box protein L2-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378267, CurrentID=0, Status='live', Symbol='LOC105378267', Aliases='', description='uncharacterized LOC105378267', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42192464, end_position_on_the_genomic_accession=42242291, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378268, CurrentID=0, Status='live', Symbol='LOC105378268', Aliases='', description='uncharacterized LOC105378268', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42638464, end_position_on_the_genomic_accession=42645542, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378269, CurrentID=0, Status='live', Symbol='LOC105378269', Aliases='', description='uncharacterized LOC105378269', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=42755142, end_position_on_the_genomic_accession=42769621, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378270 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378270, CurrentID=0, Status='live', Symbol='LOC105378270', Aliases='', description='uncharacterized LOC105378270', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43323017, end_position_on_the_genomic_accession=43327938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389599, CurrentID=0, Status='live', Symbol='STRADBP1', Aliases='ALS2CR2P1, STLK6ps1', description='STE20-related kinase adaptor beta pseudogene 1', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 2 pseudogene', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=150511974, end_position_on_the_genomic_accession=150514047, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378887 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378887, CurrentID=0, Status='live', Symbol='LOC105378887', Aliases='', description='uncharacterized LOC105378887', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106448066, end_position_on_the_genomic_accession=106485680, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378273, CurrentID=0, Status='live', Symbol='LOC105378273', Aliases='', description='uncharacterized LOC105378273', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43598612, end_position_on_the_genomic_accession=43604486, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  651746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651746, CurrentID=0, Status='live', Symbol='ANKRD33B', Aliases='', description='ankyrin repeat domain 33B', other_designations='Ankyrin repeat domain-containing protein LOC651746', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=10564323, end_position_on_the_genomic_accession=10657816, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283387, CurrentID=0, Status='live', Symbol='LOC283387', Aliases='', description='uncharacterized LOC283387', other_designations='', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=57869173, end_position_on_the_genomic_accession=57893819, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  401509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401509, CurrentID=0, Status='live', Symbol='ZNF658B', Aliases='', description='zinc finger protein 658B, pseudogene', other_designations='zinc finger protein 658 pseudogene', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39443815, end_position_on_the_genomic_accession=39464526, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378277, CurrentID=0, Status='live', Symbol='LOC105378277', Aliases='', description='uncharacterized LOC105378277', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=43901365, end_position_on_the_genomic_accession=43912343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378278, CurrentID=0, Status='live', Symbol='LOC105378278', Aliases='', description='uncharacterized LOC105378278', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44293483, end_position_on_the_genomic_accession=44349067, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378279, CurrentID=0, Status='live', Symbol='LOC105378279', Aliases='', description='uncharacterized LOC105378279', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44402733, end_position_on_the_genomic_accession=44459911, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378280, CurrentID=0, Status='live', Symbol='LOC105378280', Aliases='', description='uncharacterized LOC105378280', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44945060, end_position_on_the_genomic_accession=44954748, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378281, CurrentID=0, Status='live', Symbol='LOC105378281', Aliases='', description='uncharacterized LOC105378281', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=44969111, end_position_on_the_genomic_accession=44970174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389610, CurrentID=0, Status='live', Symbol='XKR5', Aliases='HARL2754, UNQ2754, XRG5A, XRG5BM', description='X-linked Kx blood group related 5', other_designations='X Kell blood group precursor-related family, member 5|XK, Kell blood group complex subunit-related family, member 5|XK-related protein 5a', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6808517, end_position_on_the_genomic_accession=6835644, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  389611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389611, CurrentID=0, Status='live', Symbol='FAM90A3P', Aliases='FAM90A3', description='family with sequence similarity 90 member A3, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7262866, end_position_on_the_genomic_accession=7268098, orientation='plus', exon_count=0, OMIM=613042, no_hdr0=''),
  105378284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378284, CurrentID=0, Status='live', Symbol='LOC105378284', Aliases='', description='uncharacterized LOC105378284', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45071705, end_position_on_the_genomic_accession=45074924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378285, CurrentID=0, Status='live', Symbol='LOC105378285', Aliases='', description='uncharacterized LOC105378285', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45046940, end_position_on_the_genomic_accession=45064542, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378286, CurrentID=0, Status='live', Symbol='LOC105378286', Aliases='', description='uncharacterized LOC105378286', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45464275, end_position_on_the_genomic_accession=45511565, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  554174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554174, CurrentID=0, Status='live', Symbol='LOC554174', Aliases='', description='uncharacterized LOC554174', other_designations='', map_location='22q13.31', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  389616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389616, CurrentID=0, Status='live', Symbol='OR7E125P', Aliases='PJCG6', description='olfactory receptor family 7 subfamily E member 125 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 125 pseudogene|seven transmembrane helix receptor', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7246619, end_position_on_the_genomic_accession=7247587, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378289, CurrentID=0, Status='live', Symbol='LOC105378289', Aliases='', description='uncharacterized LOC105378289', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46607751, end_position_on_the_genomic_accession=46612020, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389618, CurrentID=0, Status='live', Symbol='FAM90A6P', Aliases='', description='family with sequence similarity 90 member A6, pseudogene', other_designations='family with sequence similarity 90, member A6, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7548493, end_position_on_the_genomic_accession=7553723, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378291, CurrentID=0, Status='live', Symbol='LOC105378291', Aliases='', description='uncharacterized LOC105378291', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47203531, end_position_on_the_genomic_accession=47210958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378292, CurrentID=0, Status='live', Symbol='LOC105378292', Aliases='', description='uncharacterized LOC105378292', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47413387, end_position_on_the_genomic_accession=47414620, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378293, CurrentID=0, Status='live', Symbol='LOC105378293', Aliases='', description='uncharacterized LOC105378293', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48140547, end_position_on_the_genomic_accession=48145048, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378294, CurrentID=0, Status='live', Symbol='LOC105378294', Aliases='', description='uncharacterized LOC105378294', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48043394, end_position_on_the_genomic_accession=48049232, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378295, CurrentID=0, Status='live', Symbol='LOC105378295', Aliases='', description='uncharacterized LOC105378295', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47959322, end_position_on_the_genomic_accession=47968165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378296, CurrentID=0, Status='live', Symbol='LOC105378296', Aliases='', description='uncharacterized LOC105378296', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47729532, end_position_on_the_genomic_accession=47733997, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  127481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127481, CurrentID=0, Status='live', Symbol='S100A7P1', Aliases='S100A7L3, s100a7d', description='S100 calcium binding protein A7 pseudogene 1', other_designations='S100 calcium binding protein A7-like 3', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153427008, end_position_on_the_genomic_accession=153428401, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389626, CurrentID=0, Status='live', Symbol='OR7E161P', Aliases='', description='olfactory receptor family 7 subfamily E member 161 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 161 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11928568, end_position_on_the_genomic_accession=11929563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378299 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378299, CurrentID=0, Status='live', Symbol='LOC105378299', Aliases='', description='uncharacterized LOC105378299', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48831657, end_position_on_the_genomic_accession=48837052, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378300, CurrentID=0, Status='live', Symbol='LOC105378300', Aliases='', description='uncharacterized LOC105378300', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=49419269, end_position_on_the_genomic_accession=49472901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378301, CurrentID=0, Status='live', Symbol='LOC105378301', Aliases='', description='uncharacterized LOC105378301', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50322934, end_position_on_the_genomic_accession=50342166, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389630, CurrentID=0, Status='live', Symbol='FAM90A15P', Aliases='FAM90A15', description='family with sequence similarity 90 member A15, pseudogene', other_designations='family with sequence similarity 90, member A15, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7255243, end_position_on_the_genomic_accession=7260476, orientation='plus', exon_count=0, OMIM=613051, no_hdr0=''),
  389631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389631, CurrentID=0, Status='live', Symbol='ZNF705CP', Aliases='ZNF705C', description='zinc finger protein 705C, pseudogene', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12354362, end_position_on_the_genomic_accession=12361857, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378304 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378304, CurrentID=0, Status='live', Symbol='LOC105378304', Aliases='', description='uncharacterized LOC105378304', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=51732917, end_position_on_the_genomic_accession=51737355, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389633, CurrentID=0, Status='live', Symbol='FAM90A25P', Aliases='', description='family with sequence similarity 90 member A25, pseudogene', other_designations='family with sequence similarity 90, member A25, pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12414521, end_position_on_the_genomic_accession=12418032, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389634, CurrentID=0, Status='live', Symbol='LINC00937', Aliases='', description='long intergenic non-protein coding RNA 937', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8356963, end_position_on_the_genomic_accession=8390752, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378307, CurrentID=0, Status='live', Symbol='LOC105378307', Aliases='', description='uncharacterized LOC105378307', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52782141, end_position_on_the_genomic_accession=52810960, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378308, CurrentID=0, Status='live', Symbol='LOC105378308', Aliases='', description='uncharacterized LOC105378308', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52946431, end_position_on_the_genomic_accession=52965394, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378309, CurrentID=0, Status='live', Symbol='LOC105378309', Aliases='', description='uncharacterized LOC105378309', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=52972263, end_position_on_the_genomic_accession=53030299, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378310, CurrentID=0, Status='live', Symbol='LOC105378310', Aliases='', description='uncharacterized LOC105378310', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=53151803, end_position_on_the_genomic_accession=53163045, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127495, CurrentID=0, Status='live', Symbol='LRRC39', Aliases='', description='leucine rich repeat containing 39', other_designations='densin hlg', map_location='1p21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=100148448, end_position_on_the_genomic_accession=100178273, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105378312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378312, CurrentID=0, Status='live', Symbol='LOC105378312', Aliases='', description='uncharacterized LOC105378312', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=55938671, end_position_on_the_genomic_accession=56151992, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389641, CurrentID=0, Status='live', Symbol='LOC389641', Aliases='', description='uncharacterized LOC389641', other_designations='', map_location='8p21.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23225221, end_position_on_the_genomic_accession=23230926, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378314, CurrentID=0, Status='live', Symbol='LOC105378314', Aliases='', description='uncharacterized LOC105378314', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=57698901, end_position_on_the_genomic_accession=57847359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389643, CurrentID=0, Status='live', Symbol='NUGGC', Aliases='C8orf80, HMFN0672, SLIP-GC', description='nuclear GTPase, germinal center associated', other_designations='GTPase SLIP-GC|speckled-like pattern in the germinal center', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28021964, end_position_on_the_genomic_accession=28084063, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  389644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389644, CurrentID=0, Status='live', Symbol='RPL5P22', Aliases='RPL5_12_886', description='ribosomal protein L5 pseudogene 22', other_designations='', map_location='8p21.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28299608, end_position_on_the_genomic_accession=28300631, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378317, CurrentID=0, Status='live', Symbol='LOC105378317', Aliases='', description='uncharacterized LOC105378317', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=58608991, end_position_on_the_genomic_accession=58614513, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378318, CurrentID=0, Status='live', Symbol='LOC105378318', Aliases='', description='uncharacterized LOC105378318', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59578698, end_position_on_the_genomic_accession=59596363, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378319, CurrentID=0, Status='live', Symbol='LOC105378319', Aliases='', description='uncharacterized LOC105378319', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=59629370, end_position_on_the_genomic_accession=59649659, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378320, CurrentID=0, Status='live', Symbol='LOC105378320', Aliases='', description='uncharacterized LOC105378320', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60730384, end_position_on_the_genomic_accession=60745082, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389649, CurrentID=0, Status='live', Symbol='C8orf86', Aliases='', description='chromosome 8 open reading frame 86', other_designations='', map_location='8p11.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38504582, end_position_on_the_genomic_accession=38528665, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378322, CurrentID=0, Status='live', Symbol='LOC105378322', Aliases='', description='uncharacterized LOC105378322', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=60898763, end_position_on_the_genomic_accession=60903988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378323, CurrentID=0, Status='live', Symbol='LOC105378323', Aliases='', description='uncharacterized LOC105378323', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=61098982, end_position_on_the_genomic_accession=61103756, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389652, CurrentID=0, Status='live', Symbol='ASNSP1', Aliases='ASNSL1', description='asparagine synthetase pseudogene 1', other_designations='asparagine synthetase-like 1', map_location='8q11.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=46579206, end_position_on_the_genomic_accession=46617739, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378325, CurrentID=0, Status='live', Symbol='LOC105378325', Aliases='', description='uncharacterized LOC105378325', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62243506, end_position_on_the_genomic_accession=62304141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378326, CurrentID=0, Status='live', Symbol='LOC105378326', Aliases='', description='uncharacterized LOC105378326', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=62219117, end_position_on_the_genomic_accession=62234277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389655, CurrentID=0, Status='live', Symbol='RPL29P19', Aliases='RPL29_7_901', description='ribosomal protein L29 pseudogene 19', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=48384537, end_position_on_the_genomic_accession=48385192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378328, CurrentID=0, Status='live', Symbol='LOC105378328', Aliases='', description='uncharacterized LOC105378328', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63123929, end_position_on_the_genomic_accession=63125555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378329, CurrentID=0, Status='live', Symbol='LOC105378329', Aliases='', description='uncharacterized LOC105378329', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63591693, end_position_on_the_genomic_accession=63625310, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  389658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389658, CurrentID=0, Status='live', Symbol='FAM150A', Aliases='UNQ9433', description='family with sequence similarity 150 member A', other_designations='RPLK9433|family with sequence similarity 150, member A', map_location='8q11.23', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=52534037, end_position_on_the_genomic_accession=52565461, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378331, CurrentID=0, Status='live', Symbol='LOC105378331', Aliases='', description='uncharacterized LOC105378331', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63159504, end_position_on_the_genomic_accession=63167571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378332 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378332, CurrentID=0, Status='live', Symbol='LOC105378332', Aliases='', description='uncharacterized LOC105378332', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=63887402, end_position_on_the_genomic_accession=64091251, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105378333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378333, CurrentID=0, Status='live', Symbol='LOC105378333', Aliases='', description='uncharacterized LOC105378333', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64029392, end_position_on_the_genomic_accession=64037449, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389662, CurrentID=0, Status='live', Symbol='SEPT10P1', Aliases='DKFZP564M1416', description='septin 10 pseudogene 1', other_designations='', map_location='8q12.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=56476830, end_position_on_the_genomic_accession=56478090, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378335, CurrentID=0, Status='live', Symbol='LOC105378335', Aliases='', description='uncharacterized LOC105378335', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64589407, end_position_on_the_genomic_accession=64626971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378336, CurrentID=0, Status='live', Symbol='LOC105378336', Aliases='', description='uncharacterized LOC105378336', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=64866865, end_position_on_the_genomic_accession=64883782, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378337, CurrentID=0, Status='live', Symbol='LOC105378337', Aliases='', description='uncharacterized LOC105378337', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65479192, end_position_on_the_genomic_accession=65481670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378338, CurrentID=0, Status='live', Symbol='LOC105378338', Aliases='', description='uncharacterized LOC105378338', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65745581, end_position_on_the_genomic_accession=65811389, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378339, CurrentID=0, Status='live', Symbol='LOC105378339', Aliases='', description='uncharacterized LOC105378339', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=65837944, end_position_on_the_genomic_accession=65891451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389668, CurrentID=0, Status='live', Symbol='XKR9', Aliases='XRG9', description='X-linked Kx blood group related 9', other_designations='X Kell blood group precursor-related family, member 9|XK, Kell blood group complex subunit-related family, member 9', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70669363, end_position_on_the_genomic_accession=71065788, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105378341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378341, CurrentID=0, Status='live', Symbol='LOC105378341', Aliases='', description='uncharacterized LOC105378341', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=68199247, end_position_on_the_genomic_accession=68216861, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378342, CurrentID=0, Status='live', Symbol='LOC105378342', Aliases='', description='uncharacterized LOC105378342', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69046959, end_position_on_the_genomic_accession=69048980, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  651815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=651815, CurrentID=0, Status='live', Symbol='KLF3P1', Aliases='', description='Kruppel-like factor 3 (basic) pseudogene 1', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172253780, end_position_on_the_genomic_accession=172255740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389672, CurrentID=0, Status='live', Symbol='RPSAP47', Aliases='RPSA_19_921', description='ribosomal protein SA pseudogene 47', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=80558794, end_position_on_the_genomic_accession=80559827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378345, CurrentID=0, Status='live', Symbol='LOC105378345', Aliases='', description='uncharacterized LOC105378345', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69534935, end_position_on_the_genomic_accession=69539186, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  389674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389674, CurrentID=0, Status='live', Symbol='HNRNPA1P4', Aliases='HNRPA1P4', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 4', other_designations='', map_location='8q21.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=82290908, end_position_on_the_genomic_accession=82292391, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378347, CurrentID=0, Status='live', Symbol='LOC105378347', Aliases='', description='uncharacterized LOC105378347', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69808805, end_position_on_the_genomic_accession=69820887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389676, CurrentID=0, Status='live', Symbol='C8orf87', Aliases='', description='chromosome 8 open reading frame 87', other_designations='', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93134095, end_position_on_the_genomic_accession=93166850, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389677, CurrentID=0, Status='live', Symbol='RBM12B', Aliases='MGC:33837', description='RNA binding motif protein 12B', other_designations='RNA-binding motif protein 12B', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=93731500, end_position_on_the_genomic_accession=93741012, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  127534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127534, CurrentID=0, Status='live', Symbol='GJB4', Aliases='CX30.3, EKV', description='gap junction protein beta 4', other_designations='connexin 30.3|connexin-30.3|gap junction protein, beta 4, 30.3kDa', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34759741, end_position_on_the_genomic_accession=34763724, orientation='plus', exon_count=3, OMIM=605425, no_hdr0=''),
  105378351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378351, CurrentID=0, Status='live', Symbol='LOC105378351', Aliases='', description='uncharacterized LOC105378351', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71003507, end_position_on_the_genomic_accession=71005764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378352, CurrentID=0, Status='live', Symbol='LOC105378352', Aliases='', description='uncharacterized LOC105378352', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71245821, end_position_on_the_genomic_accession=71251690, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378353, CurrentID=0, Status='live', Symbol='LOC105378353', Aliases='', description='uncharacterized LOC105378353', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71338806, end_position_on_the_genomic_accession=71348526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378354, CurrentID=0, Status='live', Symbol='LOC105378354', Aliases='', description='uncharacterized LOC105378354', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71704182, end_position_on_the_genomic_accession=71709087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378355, CurrentID=0, Status='live', Symbol='LOC105378355', Aliases='', description='uncharacterized LOC105378355', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=71551050, end_position_on_the_genomic_accession=71554006, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  127540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127540, CurrentID=0, Status='live', Symbol='HMGB4', Aliases='dJ1007G16.5', description='high mobility group box 4', other_designations='HMG2 like', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33860475, end_position_on_the_genomic_accession=33864791, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378357, CurrentID=0, Status='live', Symbol='LOC105378357', Aliases='', description='uncharacterized LOC105378357', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=72942163, end_position_on_the_genomic_accession=72945702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378358, CurrentID=0, Status='live', Symbol='LOC105378358', Aliases='', description='uncharacterized LOC105378358', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73182493, end_position_on_the_genomic_accession=73234843, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378359, CurrentID=0, Status='live', Symbol='LOC105378359', Aliases='', description='uncharacterized LOC105378359', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=73139474, end_position_on_the_genomic_accession=73149555, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127544, CurrentID=0, Status='live', Symbol='RNF19B', Aliases='IBRDC3, NKLAM', description='ring finger protein 19B', other_designations='IBR domain-containing protein 3|natural killer lytic-associated molecule', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32932498, end_position_on_the_genomic_accession=32965266, orientation='minus', exon_count=10, OMIM=610872, no_hdr0=''),
  127545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127545, CurrentID=0, Status='live', Symbol='RPL18AP4', Aliases='RPL18A_2_30', description='ribosomal protein L18a pseudogene 4', other_designations='', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32979907, end_position_on_the_genomic_accession=32980525, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  389690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389690, CurrentID=0, Status='live', Symbol='MROH5', Aliases='', description='maestro heat-like repeat family member 5', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=141433829, end_position_on_the_genomic_accession=141507230, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  105378363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378363, CurrentID=0, Status='live', Symbol='LOC105378363', Aliases='', description='uncharacterized LOC105378363', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75385343, end_position_on_the_genomic_accession=75389935, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389692, CurrentID=0, Status='live', Symbol='MAFA', Aliases='RIPE3b1, hMafA', description='v-maf avian musculoaponeurotic fibrosarcoma oncogene homolog A', other_designations='pancreatic beta-cell-specific transcriptional activator|transcription factor RIPE3b1', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143428060, end_position_on_the_genomic_accession=143430432, orientation='minus', exon_count=1, OMIM=610303, no_hdr0=''),
  101929738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929738, CurrentID=0, Status='live', Symbol='LOC101929738', Aliases='', description='putative POM121-like protein 1', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18846449, end_position_on_the_genomic_accession=18848852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  127550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127550, CurrentID=0, Status='live', Symbol='A3GALT2', Aliases='A3GALT2P, IGBS3S', description='alpha 1,3-galactosyltransferase 2', other_designations='alpha 1,3-galactosyltransferase 2, pseudogene|iGb3 synthase|isoglobotriaosylceramide synthase', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33306766, end_position_on_the_genomic_accession=33321098, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378367, CurrentID=0, Status='live', Symbol='LOC105378367', Aliases='', description='uncharacterized LOC105378367', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=75712455, end_position_on_the_genomic_accession=75743848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378368, CurrentID=0, Status='live', Symbol='LOC105378368', Aliases='', description='uncharacterized LOC105378368', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77618871, end_position_on_the_genomic_accession=77628404, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378369, CurrentID=0, Status='live', Symbol='LOC105378369', Aliases='', description='uncharacterized LOC105378369', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77422984, end_position_on_the_genomic_accession=77432687, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378370, CurrentID=0, Status='live', Symbol='LOC105378370', Aliases='', description='uncharacterized LOC105378370', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77344593, end_position_on_the_genomic_accession=77347463, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422411, CurrentID=0, Status='live', Symbol='PSMD8P1', Aliases='', description='proteasome 26S subunit, non-ATPase, 8 pseudogene 1', other_designations='proteasome (prosome, macropain) 26S subunit, non-ATPase, 8 pseudogene|proteasome 26S non-ATPase subunit 8 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=154414272, end_position_on_the_genomic_accession=154415222, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378372, CurrentID=0, Status='live', Symbol='LOC105378372', Aliases='', description='uncharacterized LOC105378372', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=77702443, end_position_on_the_genomic_accession=77711109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378373, CurrentID=0, Status='live', Symbol='LOC105378373', Aliases='', description='uncharacterized LOC105378373', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78148441, end_position_on_the_genomic_accession=78149333, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378374, CurrentID=0, Status='live', Symbol='LOC105378374', Aliases='', description='uncharacterized LOC105378374', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78164902, end_position_on_the_genomic_accession=78178137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378375, CurrentID=0, Status='live', Symbol='LOC105378375', Aliases='', description='uncharacterized LOC105378375', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78179099, end_position_on_the_genomic_accession=78193046, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378376, CurrentID=0, Status='live', Symbol='LOC105378376', Aliases='', description='uncharacterized LOC105378376', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78213718, end_position_on_the_genomic_accession=78233436, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378377, CurrentID=0, Status='live', Symbol='LOC105378377', Aliases='', description='uncharacterized LOC105378377', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78354274, end_position_on_the_genomic_accession=78356101, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378378, CurrentID=0, Status='live', Symbol='LOC105378378', Aliases='', description='uncharacterized LOC105378378', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78463521, end_position_on_the_genomic_accession=78473947, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378379, CurrentID=0, Status='live', Symbol='LOC105378379', Aliases='', description='uncharacterized LOC105378379', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78645054, end_position_on_the_genomic_accession=78665940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378380, CurrentID=0, Status='live', Symbol='LOC105378380', Aliases='', description='uncharacterized LOC105378380', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79056000, end_position_on_the_genomic_accession=79056703, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378381, CurrentID=0, Status='live', Symbol='LOC105378381', Aliases='', description='uncharacterized LOC105378381', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=78834823, end_position_on_the_genomic_accession=78872932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378382, CurrentID=0, Status='live', Symbol='LOC105378382', Aliases='', description='uncharacterized LOC105378382', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79404303, end_position_on_the_genomic_accession=79420065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378383, CurrentID=0, Status='live', Symbol='LOC105378383', Aliases='', description='uncharacterized LOC105378383', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79447575, end_position_on_the_genomic_accession=79454750, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378384, CurrentID=0, Status='live', Symbol='LOC105378384', Aliases='', description='uncharacterized LOC105378384', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79798198, end_position_on_the_genomic_accession=79799154, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378385, CurrentID=0, Status='live', Symbol='LOC105378385', Aliases='', description='uncharacterized LOC105378385', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=79630836, end_position_on_the_genomic_accession=79632186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378386, CurrentID=0, Status='live', Symbol='LOC105378386', Aliases='', description='uncharacterized LOC105378386', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=80649552, end_position_on_the_genomic_accession=80653734, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389715, CurrentID=0, Status='live', Symbol='FAM205BP', Aliases='C9orf144, FAM205B', description='family with sequence similarity 205 member B, pseudogene', other_designations='family with sequence similarity 205, member B, pseudogene', map_location='9p13.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34830267, end_position_on_the_genomic_accession=34838586, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378388, CurrentID=0, Status='live', Symbol='LOC105378388', Aliases='', description='uncharacterized LOC105378388', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=81831162, end_position_on_the_genomic_accession=81837085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378389, CurrentID=0, Status='live', Symbol='LOC105378389', Aliases='', description='uncharacterized LOC105378389', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=81838784, end_position_on_the_genomic_accession=81850419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378390, CurrentID=0, Status='live', Symbol='LOC105378390', Aliases='', description='uncharacterized LOC105378390', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=81864104, end_position_on_the_genomic_accession=81869012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378391, CurrentID=0, Status='live', Symbol='LOC105378391', Aliases='', description='uncharacterized LOC105378391', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=82037414, end_position_on_the_genomic_accession=82043641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378392, CurrentID=0, Status='live', Symbol='LOC105378392', Aliases='', description='uncharacterized LOC105378392', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83188962, end_position_on_the_genomic_accession=83195025, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378393, CurrentID=0, Status='live', Symbol='LOC105378393', Aliases='', description='uncharacterized LOC105378393', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83546315, end_position_on_the_genomic_accession=83547979, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378394, CurrentID=0, Status='live', Symbol='LOC105378394', Aliases='', description='uncharacterized LOC105378394', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83521509, end_position_on_the_genomic_accession=83548449, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127579, CurrentID=0, Status='live', Symbol='DCST2', Aliases='', description='DC-STAMP domain containing 2', other_designations='', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=155017448, end_position_on_the_genomic_accession=155033781, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  105378396 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378396, CurrentID=0, Status='live', Symbol='LOC105378396', Aliases='', description='uncharacterized LOC105378396', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83584616, end_position_on_the_genomic_accession=83649658, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378397, CurrentID=0, Status='live', Symbol='LOC105378397', Aliases='', description='uncharacterized LOC105378397', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83672406, end_position_on_the_genomic_accession=83677118, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378398, CurrentID=0, Status='live', Symbol='LOC105378398', Aliases='', description='uncharacterized LOC105378398', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=83953305, end_position_on_the_genomic_accession=83970831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101060351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060351, CurrentID=0, Status='live', Symbol='TBC1D3K', Aliases='', description='TBC1 domain family member 3K', other_designations='TBC1 domain family member-like|TBC1 domain family, member 3K', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=37923805, end_position_on_the_genomic_accession=37935371, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105378400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378400, CurrentID=0, Status='live', Symbol='LOC105378400', Aliases='', description='uncharacterized LOC105378400', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84847273, end_position_on_the_genomic_accession=84874407, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378401 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378401, CurrentID=0, Status='live', Symbol='LOC105378401', Aliases='', description='uncharacterized LOC105378401', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=84947323, end_position_on_the_genomic_accession=84951362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389730, CurrentID=0, Status='live', Symbol='SPATA31A6', Aliases='FAM75A6', description='SPATA31 subfamily A, member 6', other_designations='family with sequence similarity 75, member A6|protein FAM75A6', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42183659, end_position_on_the_genomic_accession=42189887, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378403, CurrentID=0, Status='live', Symbol='LOC105378403', Aliases='', description='uncharacterized LOC105378403', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85377499, end_position_on_the_genomic_accession=85380116, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378404, CurrentID=0, Status='live', Symbol='LOC105378404', Aliases='', description='uncharacterized LOC105378404', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=85537119, end_position_on_the_genomic_accession=85580521, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105378405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378405, CurrentID=0, Status='live', Symbol='LOC105378405', Aliases='', description='uncharacterized LOC105378405', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86522065, end_position_on_the_genomic_accession=86525101, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378406, CurrentID=0, Status='live', Symbol='LOC105378406', Aliases='', description='uncharacterized LOC105378406', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86556328, end_position_on_the_genomic_accession=86566402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378407, CurrentID=0, Status='live', Symbol='LOC105378407', Aliases='', description='uncharacterized LOC105378407', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86546027, end_position_on_the_genomic_accession=86556154, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378408, CurrentID=0, Status='live', Symbol='LOC105378408', Aliases='', description='uncharacterized LOC105378408', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=86534096, end_position_on_the_genomic_accession=86558059, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  654780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654780, CurrentID=0, Status='live', Symbol='SFPQ', Aliases='', description='splicing factor proline/glutamine-rich', other_designations='', map_location='16q24.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.9', start_position_on_the_genomic_accession=84226164, end_position_on_the_genomic_accession=84230659, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378410, CurrentID=0, Status='live', Symbol='LOC105378410', Aliases='', description='uncharacterized LOC105378410', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87189779, end_position_on_the_genomic_accession=87200025, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378411, CurrentID=0, Status='live', Symbol='LOC105378411', Aliases='', description='uncharacterized LOC105378411', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=87658444, end_position_on_the_genomic_accession=87658945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389740, CurrentID=0, Status='live', Symbol='ADGRF5P1', Aliases='GPR116P1', description='adhesion G protein-coupled receptor F5 pseudogene 1', other_designations='G protein-coupled receptor 116 pseudogene 1', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62860847, end_position_on_the_genomic_accession=62862193, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  389741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389741, CurrentID=0, Status='live', Symbol='GLIDR', Aliases='LINC01172, TCONS_00015562', description='glioblastoma down-regulated RNA', other_designations='long intergenic non-protein coding RNA 1172', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39803496, end_position_on_the_genomic_accession=39810159, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378414, CurrentID=0, Status='live', Symbol='LOC105378414', Aliases='', description='uncharacterized LOC105378414', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88042318, end_position_on_the_genomic_accession=88060177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378415, CurrentID=0, Status='live', Symbol='LOC105378415', Aliases='', description='uncharacterized LOC105378415', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88066878, end_position_on_the_genomic_accession=88104389, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378416, CurrentID=0, Status='live', Symbol='LOC105378416', Aliases='', description='uncharacterized LOC105378416', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88120313, end_position_on_the_genomic_accession=88122102, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378417, CurrentID=0, Status='live', Symbol='LOC105378417', Aliases='', description='uncharacterized LOC105378417', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88995788, end_position_on_the_genomic_accession=89002993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127602, CurrentID=0, Status='live', Symbol='DNAH14', Aliases='C1orf67, Dnahc14, HL-18, HL18', description='dynein, axonemal, heavy chain 14', other_designations='axonemal beta dynein heavy chain 14|ciliary dynein heavy chain 14|dynein, axonemal, heavy polypeptide 14', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=224929595, end_position_on_the_genomic_accession=225399294, orientation='plus', exon_count=92, OMIM=603341, no_hdr0=''),
  389747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389747, CurrentID=0, Status='live', Symbol='MEP1AP3', Aliases='', description='meprin A subunit alpha pseudogene 3', other_designations='meprin A, alpha (PABA peptide hydrolase) pseudogene|meprin A, alpha pseudogene 3', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66182324, end_position_on_the_genomic_accession=66189100, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378420 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378420, CurrentID=0, Status='live', Symbol='LOC105378420', Aliases='', description='uncharacterized LOC105378420', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89646285, end_position_on_the_genomic_accession=89650780, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378421, CurrentID=0, Status='live', Symbol='LOC105378421', Aliases='', description='uncharacterized LOC105378421', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89645414, end_position_on_the_genomic_accession=89650718, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378422, CurrentID=0, Status='live', Symbol='LOC105378422', Aliases='', description='uncharacterized LOC105378422', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89531701, end_position_on_the_genomic_accession=89592727, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378423, CurrentID=0, Status='live', Symbol='LOC105378423', Aliases='', description='uncharacterized LOC105378423', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=89789148, end_position_on_the_genomic_accession=89804646, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127608, CurrentID=0, Status='live', Symbol='OR2AJ1', Aliases='OR2AJ1PQ, OR2AJ1', description='olfactory receptor family 2 subfamily AJ member 1', other_designations='olfactory receptor OR1-52|olfactory receptor, family 2, subfamily AJ, member 1 pseudogene', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247933877, end_position_on_the_genomic_accession=247934819, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378425, CurrentID=0, Status='live', Symbol='LOC105378425', Aliases='', description='uncharacterized LOC105378425', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90065757, end_position_on_the_genomic_accession=90079837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378426, CurrentID=0, Status='live', Symbol='LOC105378426', Aliases='', description='uncharacterized LOC105378426', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90173231, end_position_on_the_genomic_accession=90175714, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378427, CurrentID=0, Status='live', Symbol='XLOC_008559', Aliases='', description='uncharacterized LOC105378427', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90981963, end_position_on_the_genomic_accession=90990537, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389756, CurrentID=0, Status='live', Symbol='AQP7P2', Aliases='', description='aquaporin 7 pseudogene 2', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64621504, end_position_on_the_genomic_accession=64638864, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378429, CurrentID=0, Status='live', Symbol='LOC105378429', Aliases='', description='uncharacterized LOC105378429', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91103140, end_position_on_the_genomic_accession=91161299, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378430, CurrentID=0, Status='live', Symbol='LOC105378430', Aliases='', description='uncharacterized LOC105378430', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91033166, end_position_on_the_genomic_accession=91040929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378431, CurrentID=0, Status='live', Symbol='LOC105378431', Aliases='', description='uncharacterized LOC105378431', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=90997579, end_position_on_the_genomic_accession=91023424, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378432, CurrentID=0, Status='live', Symbol='LOC105378432', Aliases='', description='uncharacterized LOC105378432', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91575199, end_position_on_the_genomic_accession=91580815, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389761, CurrentID=0, Status='live', Symbol='SPATA31D4', Aliases='FAM75D4', description='SPATA31 subfamily D, member 4', other_designations='FAM75-like protein FLJ43859|family with sequence similarity 75, member D4|putative protein FAM75D4', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81928428, end_position_on_the_genomic_accession=81934998, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389762, CurrentID=0, Status='live', Symbol='SPATA31D3', Aliases='FAM75D3', description='SPATA31 subfamily D, member 3', other_designations='FAM75-like protein FLJ44082|Putative FAM75-like protein FLJ44082|family with sequence similarity 75, member D3|putative protein FAM75D3', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81943500, end_position_on_the_genomic_accession=81950094, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  389763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389763, CurrentID=0, Status='live', Symbol='SPATA31D1', Aliases='FAM75D1', description='SPATA31 subfamily D, member 1', other_designations='FAM75-like protein FLJ46321|family with sequence similarity 75, member D1|protein FAM75D1', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=81987667, end_position_on_the_genomic_accession=81995256, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378436, CurrentID=0, Status='live', Symbol='LOC105378436', Aliases='', description='uncharacterized LOC105378436', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93556792, end_position_on_the_genomic_accession=93561103, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389765, CurrentID=0, Status='live', Symbol='LOC389765', Aliases='', description='kinesin family member 27 pseudogene', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=85806002, end_position_on_the_genomic_accession=85842879, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  389766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389766, CurrentID=0, Status='live', Symbol='C9orf153', Aliases='bA507D14.1', description='chromosome 9 open reading frame 153', other_designations='', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=86220265, end_position_on_the_genomic_accession=86259657, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  127623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127623, CurrentID=0, Status='live', Symbol='OR2B11', Aliases='', description='olfactory receptor family 2 subfamily B member 11', other_designations='novel 7 transmembrane receptor (rhodopsin family) protein|olfactory receptor, family 2, subfamily B, member 11', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247451029, end_position_on_the_genomic_accession=247451982, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  645834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645834, CurrentID=0, Status='live', Symbol='KRT8P15', Aliases='', description='keratin 8 pseudogene 15', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202840086, end_position_on_the_genomic_accession=202841857, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378441, CurrentID=0, Status='live', Symbol='LOC105378441', Aliases='', description='uncharacterized LOC105378441', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=95970301, end_position_on_the_genomic_accession=95988250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378442, CurrentID=0, Status='live', Symbol='LOC105378442', Aliases='', description='uncharacterized LOC105378442', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96627897, end_position_on_the_genomic_accession=96640553, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378443, CurrentID=0, Status='live', Symbol='LOC105378443', Aliases='', description='uncharacterized LOC105378443', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=96587098, end_position_on_the_genomic_accession=96609704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378444, CurrentID=0, Status='live', Symbol='LOC105378444', Aliases='', description='uncharacterized LOC105378444', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97077631, end_position_on_the_genomic_accession=97102199, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378445, CurrentID=0, Status='live', Symbol='LOC105378445', Aliases='', description='uncharacterized LOC105378445', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97195885, end_position_on_the_genomic_accession=97205977, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378446, CurrentID=0, Status='live', Symbol='LOC105378446', Aliases='', description='uncharacterized LOC105378446', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97215161, end_position_on_the_genomic_accession=97219075, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378447, CurrentID=0, Status='live', Symbol='LOC105378447', Aliases='', description='uncharacterized LOC105378447', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97219139, end_position_on_the_genomic_accession=97231983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378448, CurrentID=0, Status='live', Symbol='LOC105378448', Aliases='', description='uncharacterized LOC105378448', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=97335446, end_position_on_the_genomic_accession=97343203, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378449, CurrentID=0, Status='live', Symbol='LOC105378449', Aliases='', description='uncharacterized LOC105378449', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=98361315, end_position_on_the_genomic_accession=98368518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378450, CurrentID=0, Status='live', Symbol='LOC105378450', Aliases='', description='uncharacterized LOC105378450', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=99624782, end_position_on_the_genomic_accession=99653622, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378451, CurrentID=0, Status='live', Symbol='LOC105378451', Aliases='', description='uncharacterized LOC105378451', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100093045, end_position_on_the_genomic_accession=100109690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378452, CurrentID=0, Status='live', Symbol='LOC105378452', Aliases='', description='uncharacterized LOC105378452', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=100463787, end_position_on_the_genomic_accession=100479878, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378453, CurrentID=0, Status='live', Symbol='LOC105378453', Aliases='', description='uncharacterized LOC105378453', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101284452, end_position_on_the_genomic_accession=101285679, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378454, CurrentID=0, Status='live', Symbol='LOC105378454', Aliases='', description='uncharacterized LOC105378454', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101311052, end_position_on_the_genomic_accession=101313035, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101929753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929753, CurrentID=0, Status='live', Symbol='LOC101929753', Aliases='', description='uncharacterized LOC101929753', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=170340517, end_position_on_the_genomic_accession=170350943, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378456, CurrentID=0, Status='live', Symbol='LOC105378456', Aliases='', description='uncharacterized LOC105378456', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101701994, end_position_on_the_genomic_accession=101704541, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378457, CurrentID=0, Status='live', Symbol='LOC105378457', Aliases='', description='uncharacterized LOC105378457', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101767136, end_position_on_the_genomic_accession=101771576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378458, CurrentID=0, Status='live', Symbol='LOC105378458', Aliases='', description='uncharacterized LOC105378458', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=101735891, end_position_on_the_genomic_accession=101767056, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  389787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389787, CurrentID=0, Status='live', Symbol='TPT1P9', Aliases='', description='tumor protein, translationally-controlled 1 pseudogene 9', other_designations='', map_location='9q33.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=118082693, end_position_on_the_genomic_accession=118083475, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378460 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378460, CurrentID=0, Status='live', Symbol='LOC105378460', Aliases='', description='uncharacterized LOC105378460', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102650029, end_position_on_the_genomic_accession=102655131, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378461, CurrentID=0, Status='live', Symbol='LOC105378461', Aliases='', description='uncharacterized LOC105378461', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102802754, end_position_on_the_genomic_accession=102806129, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378462 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378462, CurrentID=0, Status='live', Symbol='LOC105378462', Aliases='', description='uncharacterized LOC105378462', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=102866082, end_position_on_the_genomic_accession=102869412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389791, CurrentID=0, Status='live', Symbol='PTGES2-AS1', Aliases='', description='PTGES2 antisense RNA 1 (head to head)', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=128128529, end_position_on_the_genomic_accession=128130634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389792, CurrentID=0, Status='live', Symbol='IER5L', Aliases='bA247A12.2', description='immediate early response 5-like', other_designations='', map_location='9q34.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=129175552, end_position_on_the_genomic_accession=129178261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378465, CurrentID=0, Status='live', Symbol='LOC105378465', Aliases='', description='uncharacterized LOC105378465', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105155149, end_position_on_the_genomic_accession=105177761, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378466, CurrentID=0, Status='live', Symbol='LOC105378466', Aliases='', description='uncharacterized LOC105378466', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105328109, end_position_on_the_genomic_accession=105346524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378467, CurrentID=0, Status='live', Symbol='LOC105378467', Aliases='', description='uncharacterized LOC105378467', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=105495884, end_position_on_the_genomic_accession=105496919, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378468, CurrentID=0, Status='live', Symbol='LOC105378468', Aliases='', description='uncharacterized LOC105378468', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106011448, end_position_on_the_genomic_accession=106014022, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  654790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654790, CurrentID=0, Status='live', Symbol='PCP4L1', Aliases='IQM1', description='Purkinje cell protein 4 like 1', other_designations='PCP4-like protein 1', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161258727, end_position_on_the_genomic_accession=161285450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378470, CurrentID=0, Status='live', Symbol='LOC105378470', Aliases='', description='uncharacterized LOC105378470', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106183792, end_position_on_the_genomic_accession=106188591, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  389799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389799, CurrentID=0, Status='live', Symbol='CFAP77', Aliases='C9orf171', description='cilia and flagella associated protein 77', other_designations='', map_location='9q34.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=132410043, end_position_on_the_genomic_accession=132573319, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105378472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378472, CurrentID=0, Status='live', Symbol='LOC105378472', Aliases='', description='uncharacterized LOC105378472', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106567867, end_position_on_the_genomic_accession=106571232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378473, CurrentID=0, Status='live', Symbol='LOC105378473', Aliases='', description='uncharacterized LOC105378473', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=106649104, end_position_on_the_genomic_accession=106676640, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378474, CurrentID=0, Status='live', Symbol='LOC105378474', Aliases='', description='uncharacterized LOC105378474', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=107367827, end_position_on_the_genomic_accession=107404836, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378475 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378475, CurrentID=0, Status='live', Symbol='LOC105378475', Aliases='', description='uncharacterized LOC105378475', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=107363299, end_position_on_the_genomic_accession=107376386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378476, CurrentID=0, Status='live', Symbol='LOC105378476', Aliases='', description='uncharacterized LOC105378476', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=108252088, end_position_on_the_genomic_accession=108295350, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378477, CurrentID=0, Status='live', Symbol='LOC105378477', Aliases='', description='uncharacterized LOC105378477', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=108395654, end_position_on_the_genomic_accession=108466397, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378478, CurrentID=0, Status='live', Symbol='LOC105378478', Aliases='', description='uncharacterized LOC105378478', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=108710206, end_position_on_the_genomic_accession=108761624, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378479, CurrentID=0, Status='live', Symbol='LOC105378479', Aliases='', description='uncharacterized LOC105378479', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110147955, end_position_on_the_genomic_accession=110167921, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378480, CurrentID=0, Status='live', Symbol='LOC105378480', Aliases='', description='uncharacterized LOC105378480', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110220735, end_position_on_the_genomic_accession=110223385, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  127665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127665, CurrentID=0, Status='live', Symbol='ZNF648', Aliases='', description='zinc finger protein 648', other_designations='', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182053766, end_position_on_the_genomic_accession=182062505, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378482, CurrentID=0, Status='live', Symbol='LOC105378482', Aliases='', description='uncharacterized LOC105378482', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110428769, end_position_on_the_genomic_accession=110460444, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378483, CurrentID=0, Status='live', Symbol='LOC105378483', Aliases='', description='uncharacterized LOC105378483', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=110564737, end_position_on_the_genomic_accession=110567588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389812 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389812, CurrentID=0, Status='live', Symbol='LCN15', Aliases='PRO6093, UNQ2541', description='lipocalin 15', other_designations='MSFL2541', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136759634, end_position_on_the_genomic_accession=136770114, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  389813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389813, CurrentID=0, Status='live', Symbol='C9orf172', Aliases='', description='chromosome 9 open reading frame 172', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=136844415, end_position_on_the_genomic_accession=136847345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  127670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127670, CurrentID=0, Status='live', Symbol='TEDDM1', Aliases='EDDM9, Epdd1, HE9, HEL-S-45e, TMEM45C', description='transmembrane epididymal protein 1', other_designations='epididymal protein 9|epididymis secretory sperm binding protein Li 45e|human epididymis-specific protein 9|putative membrane protein HE9|transmembrane protein 45C', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182398117, end_position_on_the_genomic_accession=182400616, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  127671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127671, CurrentID=0, Status='live', Symbol='TEDDM2P', Aliases='TEDDM1P1', description='transmembrane epididymal protein 2, pseudogene', other_designations='transmembrane epididymal protein 1 pseudogene 1', map_location='1q25.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=182441697, end_position_on_the_genomic_accession=182442451, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  389816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389816, CurrentID=0, Status='live', Symbol='LRRC26', Aliases='CAPC, bA350O14.10', description='leucine rich repeat containing 26', other_designations='BK channel auxiliary gamma subunit LRRC26|BK channel auxilliary gamma subunit LRRC26|cytokeratin associated protein|cytokeratin-associated protein in cancer', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=137168758, end_position_on_the_genomic_accession=137170039, orientation='minus', exon_count=2, OMIM=613505, no_hdr0=''),
  105378489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378489, CurrentID=0, Status='live', Symbol='LOC105378489', Aliases='', description='uncharacterized LOC105378489', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113222943, end_position_on_the_genomic_accession=113226906, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378490, CurrentID=0, Status='live', Symbol='LOC105378490', Aliases='', description='uncharacterized LOC105378490', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113306050, end_position_on_the_genomic_accession=113320971, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378491 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378491, CurrentID=0, Status='live', Symbol='LOC105378491', Aliases='', description='uncharacterized LOC105378491', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=113522764, end_position_on_the_genomic_accession=113532132, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378492, CurrentID=0, Status='live', Symbol='LOC105378492', Aliases='', description='uncharacterized LOC105378492', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114008191, end_position_on_the_genomic_accession=114022695, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378493 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378493, CurrentID=0, Status='live', Symbol='LOC105378493', Aliases='', description='uncharacterized LOC105378493', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114027353, end_position_on_the_genomic_accession=114031286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378494 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378494, CurrentID=0, Status='live', Symbol='LOC105378494', Aliases='', description='uncharacterized LOC105378494', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114320825, end_position_on_the_genomic_accession=114326885, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378495, CurrentID=0, Status='live', Symbol='LOC105378495', Aliases='', description='uncharacterized LOC105378495', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114661396, end_position_on_the_genomic_accession=114670580, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378496, CurrentID=0, Status='live', Symbol='LOC105378496', Aliases='', description='uncharacterized LOC105378496', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=115072610, end_position_on_the_genomic_accession=115087019, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378497, CurrentID=0, Status='live', Symbol='LOC105378497', Aliases='', description='uncharacterized LOC105378497', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=115903252, end_position_on_the_genomic_accession=115919160, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378498, CurrentID=0, Status='live', Symbol='LOC105378498', Aliases='', description='uncharacterized LOC105378498', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116769783, end_position_on_the_genomic_accession=116773756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389827, CurrentID=0, Status='live', Symbol='TMEM8C', Aliases='MYOMAKER, TMEM226', description='transmembrane protein 8C', other_designations='transmembrane protein 226', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133514586, end_position_on_the_genomic_accession=133528606, orientation='minus', exon_count=6, OMIM=615345, no_hdr0=''),
  105378500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378500, CurrentID=0, Status='live', Symbol='LOC105378500', Aliases='', description='uncharacterized LOC105378500', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117240038, end_position_on_the_genomic_accession=117242763, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378501, CurrentID=0, Status='live', Symbol='LOC105378501', Aliases='', description='uncharacterized LOC105378501', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117376964, end_position_on_the_genomic_accession=117402556, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378502, CurrentID=0, Status='live', Symbol='LOC105378502', Aliases='', description='uncharacterized LOC105378502', other_designations='CTA-109P11.4', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117436231, end_position_on_the_genomic_accession=117455509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127687, CurrentID=0, Status='live', Symbol='C1orf122', Aliases='ALAESM', description='chromosome 1 open reading frame 122', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37807801, end_position_on_the_genomic_accession=37809454, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378504, CurrentID=0, Status='live', Symbol='LOC105378504', Aliases='', description='uncharacterized LOC105378504', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=117935021, end_position_on_the_genomic_accession=117940266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378505, CurrentID=0, Status='live', Symbol='LOC105378505', Aliases='', description='uncharacterized LOC105378505', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118260470, end_position_on_the_genomic_accession=118270163, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389834, CurrentID=0, Status='live', Symbol='LOC389834', Aliases='', description='ankyrin repeat domain 57 pseudogene', other_designations='sosondowah ankyrin repeat domain family member C pseudogene', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378507, CurrentID=0, Status='live', Symbol='LOC105378507', Aliases='', description='uncharacterized LOC105378507', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118498175, end_position_on_the_genomic_accession=118507761, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100616355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100616355, CurrentID=0, Status='live', Symbol='MIR4531', Aliases='', description='microRNA 4531', other_designations='hsa-mir-4531', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=44653686, end_position_on_the_genomic_accession=44653732, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378509, CurrentID=0, Status='live', Symbol='LOC105378509', Aliases='', description='uncharacterized LOC105378509', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118900896, end_position_on_the_genomic_accession=118902516, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378510, CurrentID=0, Status='live', Symbol='LOC105378510', Aliases='', description='uncharacterized LOC105378510', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119426744, end_position_on_the_genomic_accession=119429244, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378511 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378511, CurrentID=0, Status='live', Symbol='LOC105378511', Aliases='', description='uncharacterized LOC105378511', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119334630, end_position_on_the_genomic_accession=119336432, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389840 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389840, CurrentID=0, Status='live', Symbol='MAP3K15', Aliases='ASK3, bA723P2.3', description='mitogen-activated protein kinase kinase kinase 15', other_designations='MAPK/ERK kinase kinase 15|MEK kinase 15|MEKK 15|apoptosis signal-regulating kinase 3', map_location='Xp22.12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=19360058, end_position_on_the_genomic_accession=19515261, orientation='minus', exon_count=33, OMIM=300820, no_hdr0=''),
  105378513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378513, CurrentID=0, Status='live', Symbol='LOC105378513', Aliases='', description='uncharacterized LOC105378513', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119792973, end_position_on_the_genomic_accession=119818208, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389842, CurrentID=0, Status='live', Symbol='LOC389842', Aliases='', description='RAN binding protein 1 pseudogene', other_designations='', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=25645592, end_position_on_the_genomic_accession=25646439, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378515 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378515, CurrentID=0, Status='live', Symbol='LOC105378515', Aliases='', description='uncharacterized LOC105378515', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120181239, end_position_on_the_genomic_accession=120337537, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  127700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127700, CurrentID=0, Status='live', Symbol='OSCP1', Aliases='C1orf102, NOR1', description='organic solute carrier partner 1', other_designations='hOSCP1|organic solute carrier protein 1|organic solute transport protein 1|oxidored nitro domain containing protein|oxidored-nitro domain-containing protein 1', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36417906, end_position_on_the_genomic_accession=36450485, orientation='minus', exon_count=13, OMIM=608854, no_hdr0=''),
  105378517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378517, CurrentID=0, Status='live', Symbol='LOC105378517', Aliases='', description='uncharacterized LOC105378517', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120608592, end_position_on_the_genomic_accession=120814855, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378518, CurrentID=0, Status='live', Symbol='LOC105378518', Aliases='', description='uncharacterized LOC105378518', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120954167, end_position_on_the_genomic_accession=120954817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127703, CurrentID=0, Status='live', Symbol='C1orf216', Aliases='', description='chromosome 1 open reading frame 216', other_designations='', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35713875, end_position_on_the_genomic_accession=35719189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378520, CurrentID=0, Status='live', Symbol='LOC105378520', Aliases='', description='uncharacterized LOC105378520', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=120984966, end_position_on_the_genomic_accession=120996754, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389849, CurrentID=0, Status='live', Symbol='GAPDHP65', Aliases='GAPDHL10, GAPDL10', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 65', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 10, pseudogene', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46439512, end_position_on_the_genomic_accession=46440785, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378522, CurrentID=0, Status='live', Symbol='LOC105378522', Aliases='', description='uncharacterized LOC105378522', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=121129539, end_position_on_the_genomic_accession=121156053, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  127707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127707, CurrentID=0, Status='live', Symbol='KLHDC7A', Aliases='', description='kelch domain containing 7A', other_designations='', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=18480930, end_position_on_the_genomic_accession=18485999, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389852, CurrentID=0, Status='live', Symbol='SPACA5', Aliases='LYC5, LYZL5, PNPK6288, SLLP-X, SLLP2, UNQ6288', description='sperm acrosome associated 5', other_designations='lysozyme-like protein 5|sperm lysozyme-like protein 2|sperm-specific lysozyme-like protein X', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48004336, end_position_on_the_genomic_accession=48010162, orientation='plus', exon_count=6, OMIM=300593, no_hdr0=''),
  105378525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378525, CurrentID=0, Status='live', Symbol='LOC105378525', Aliases='', description='uncharacterized LOC105378525', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122449053, end_position_on_the_genomic_accession=122461358, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378526, CurrentID=0, Status='live', Symbol='LOC105378526', Aliases='', description='uncharacterized LOC105378526', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122489633, end_position_on_the_genomic_accession=122493965, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378527, CurrentID=0, Status='live', Symbol='LOC105378527', Aliases='', description='uncharacterized LOC105378527', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=122478644, end_position_on_the_genomic_accession=122484021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389856, CurrentID=0, Status='live', Symbol='USP27X', Aliases='USP22L, USP27', description='ubiquitin specific peptidase 27, X-linked', other_designations='X-linked ubiquitin carboxyl-terminal hydrolase 27|deubiquitinating enzyme 27|ubiquitin specific protease 27, X chromosome|ubiquitin thioesterase 27|ubiquitin-specific-processing protease 27', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49879867, end_position_on_the_genomic_accession=49882565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  389857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389857, CurrentID=0, Status='live', Symbol='CENPVP1', Aliases='CENPVL1, PRR6L1', description='centromere protein V pseudogene 1', other_designations='centromere protein V-like 1|proline rich 6-like 1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=51710829, end_position_on_the_genomic_accession=51712131, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378530, CurrentID=0, Status='live', Symbol='LOC105378530', Aliases='', description='uncharacterized LOC105378530', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123543131, end_position_on_the_genomic_accession=123561427, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378531, CurrentID=0, Status='live', Symbol='LOC105378531', Aliases='', description='uncharacterized LOC105378531', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123352686, end_position_on_the_genomic_accession=123558526, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  389860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389860, CurrentID=0, Status='live', Symbol='PAGE2B', Aliases='CT16.5, GAGEE3, PAGE-2B', description='PAGE family member 2B', other_designations='G antigen, family E, 3|P antigen family, member 2B|prostate-associated gene 2B protein', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55029540, end_position_on_the_genomic_accession=55078909, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378533, CurrentID=0, Status='live', Symbol='LOC105378533', Aliases='', description='uncharacterized LOC105378533', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123776758, end_position_on_the_genomic_accession=123777850, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378534 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378534, CurrentID=0, Status='live', Symbol='LOC105378534', Aliases='', description='uncharacterized LOC105378534', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123893412, end_position_on_the_genomic_accession=123894324, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378535, CurrentID=0, Status='live', Symbol='LOC105378535', Aliases='', description='uncharacterized LOC105378535', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=123912912, end_position_on_the_genomic_accession=123944112, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378536, CurrentID=0, Status='live', Symbol='LOC105378536', Aliases='', description='uncharacterized LOC105378536', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124133796, end_position_on_the_genomic_accession=124151258, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378537, CurrentID=0, Status='live', Symbol='LOC105378537', Aliases='', description='uncharacterized LOC105378537', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124546809, end_position_on_the_genomic_accession=124548753, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389866, CurrentID=0, Status='live', Symbol='SERBP1P1', Aliases='SERBP1P', description='SERPINE1 mRNA binding protein 1 pseudogene 1', other_designations='PAI-1 mRNA binding protein pseudogene', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=68783373, end_position_on_the_genomic_accession=68786430, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378539, CurrentID=0, Status='live', Symbol='LOC105378539', Aliases='', description='uncharacterized LOC105378539', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124616752, end_position_on_the_genomic_accession=124621928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378540, CurrentID=0, Status='live', Symbol='LOC105378540', Aliases='', description='uncharacterized LOC105378540', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124998179, end_position_on_the_genomic_accession=125001494, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378541, CurrentID=0, Status='live', Symbol='LOC105378541', Aliases='', description='uncharacterized LOC105378541', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125228567, end_position_on_the_genomic_accession=125234053, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378542, CurrentID=0, Status='live', Symbol='LOC105378542', Aliases='', description='uncharacterized LOC105378542', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125263101, end_position_on_the_genomic_accession=125275272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378543, CurrentID=0, Status='live', Symbol='LOC105378543', Aliases='', description='uncharacterized LOC105378543', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125453388, end_position_on_the_genomic_accession=125494138, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378544, CurrentID=0, Status='live', Symbol='LOC105378544', Aliases='', description='uncharacterized LOC105378544', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=125521865, end_position_on_the_genomic_accession=125531051, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389873, CurrentID=0, Status='live', Symbol='SEPHS1P4', Aliases='', description='selenophosphate synthetase 1 pseudogene 4', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=73769248, end_position_on_the_genomic_accession=73770420, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  389874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389874, CurrentID=0, Status='live', Symbol='ZCCHC13', Aliases='CNBP2, ZNF9L', description='zinc finger, CCHC domain containing 13', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74304190, end_position_on_the_genomic_accession=74305034, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  127731 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127731, CurrentID=0, Status='live', Symbol='VWA5B1', Aliases='', description='von Willebrand factor A domain containing 5B1', other_designations='', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20290907, end_position_on_the_genomic_accession=20360221, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  105378548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378548, CurrentID=0, Status='live', Symbol='LOC105378548', Aliases='', description='uncharacterized LOC105378548', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126607547, end_position_on_the_genomic_accession=126616525, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  127733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127733, CurrentID=0, Status='live', Symbol='UBXN10', Aliases='UBXD3', description='UBX domain protein 10', other_designations='UBX domain containing 3|UBX domain-containing protein 3', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20183302, end_position_on_the_genomic_accession=20196048, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378550, CurrentID=0, Status='live', Symbol='LOC105378550', Aliases='', description='uncharacterized LOC105378550', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=126903418, end_position_on_the_genomic_accession=126905140, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378551, CurrentID=0, Status='live', Symbol='LOC105378551', Aliases='', description='uncharacterized LOC105378551', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127363312, end_position_on_the_genomic_accession=127396674, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378552, CurrentID=0, Status='live', Symbol='LOC105378552', Aliases='', description='uncharacterized LOC105378552', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=127000317, end_position_on_the_genomic_accession=127026424, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378553, CurrentID=0, Status='live', Symbol='LOC105378553', Aliases='', description='uncharacterized LOC105378553', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128056197, end_position_on_the_genomic_accession=128063076, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378554 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378554, CurrentID=0, Status='live', Symbol='LOC105378554', Aliases='', description='uncharacterized LOC105378554', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128685404, end_position_on_the_genomic_accession=128691379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378555, CurrentID=0, Status='live', Symbol='LOC105378555', Aliases='', description='uncharacterized LOC105378555', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=128781954, end_position_on_the_genomic_accession=128813103, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  389884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389884, CurrentID=0, Status='live', Symbol='PA2G4P1', Aliases='PA2G4P', description='proliferation-associated 2G4 pseudogene 1', other_designations='proliferation-associated 2G4, 38kDa pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120868348, end_position_on_the_genomic_accession=120869962, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378557, CurrentID=0, Status='live', Symbol='LOC105378557', Aliases='', description='uncharacterized LOC105378557', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129370328, end_position_on_the_genomic_accession=129381122, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378558 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378558, CurrentID=0, Status='live', Symbol='LOC105378558', Aliases='', description='uncharacterized LOC105378558', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129859710, end_position_on_the_genomic_accession=129867134, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378559, CurrentID=0, Status='live', Symbol='LOC105378559', Aliases='', description='uncharacterized LOC105378559', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129768365, end_position_on_the_genomic_accession=129770983, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378560, CurrentID=0, Status='live', Symbol='LOC105378560', Aliases='', description='uncharacterized LOC105378560', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=129715442, end_position_on_the_genomic_accession=129717867, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378561, CurrentID=0, Status='live', Symbol='LOC105378561', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130127353, end_position_on_the_genomic_accession=130132705, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378562, CurrentID=0, Status='live', Symbol='LOC105378562', Aliases='', description='uncharacterized LOC105378562', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=130685608, end_position_on_the_genomic_accession=130697052, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378563, CurrentID=0, Status='live', Symbol='LOC105378563', Aliases='', description='uncharacterized LOC105378563', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131202553, end_position_on_the_genomic_accession=131208816, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378564 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378564, CurrentID=0, Status='live', Symbol='LOC105378564', Aliases='', description='uncharacterized LOC105378564', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131830526, end_position_on_the_genomic_accession=131833634, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378565, CurrentID=0, Status='live', Symbol='LOC105378565', Aliases='', description='uncharacterized LOC105378565', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131860662, end_position_on_the_genomic_accession=131875324, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378566, CurrentID=0, Status='live', Symbol='LOC105378566', Aliases='', description='uncharacterized LOC105378566', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131875451, end_position_on_the_genomic_accession=131879783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389895, CurrentID=0, Status='live', Symbol='LOC389895', Aliases='', description='chromosome 16 open reading frame 72-like', other_designations='', map_location='Xq27.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=140091667, end_position_on_the_genomic_accession=140092911, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378568, CurrentID=0, Status='live', Symbol='LOC105378568', Aliases='', description='uncharacterized LOC105378568', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132434606, end_position_on_the_genomic_accession=132442346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378569, CurrentID=0, Status='live', Symbol='LOC105378569', Aliases='', description='uncharacterized LOC105378569', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132460589, end_position_on_the_genomic_accession=132466073, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389898 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389898, CurrentID=0, Status='live', Symbol='UBE2NL', Aliases='Li174', description='ubiquitin conjugating enzyme E2N-like (gene/pseudogene)', other_designations='epididymis tissue protein Li 174|ubiquitin-conjugating enzyme E2N-like', map_location='Xq27.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=143884071, end_position_on_the_genomic_accession=143885255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378571, CurrentID=0, Status='live', Symbol='LOC105378571', Aliases='', description='uncharacterized LOC105378571', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132849213, end_position_on_the_genomic_accession=132857536, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378572, CurrentID=0, Status='live', Symbol='LOC105378572', Aliases='', description='uncharacterized LOC105378572', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=132943367, end_position_on_the_genomic_accession=132947851, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389901, CurrentID=0, Status='live', Symbol='XRCC6P2', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 6 pseudogene 2', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150230839, end_position_on_the_genomic_accession=150269789, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378574, CurrentID=0, Status='live', Symbol='LOC105378574', Aliases='', description='uncharacterized LOC105378574', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133194487, end_position_on_the_genomic_accession=133195631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  389903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389903, CurrentID=0, Status='live', Symbol='CSAG3', Aliases='CSAG3A, CT24.2', description='CSAG family, member 3', other_designations='CSAG family, member 3A|cancer/testis antigen 24.2|taxol resistance associated gene 3|taxol-resistant-associated gene 3 protein', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=152753921, end_position_on_the_genomic_accession=152760222, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378576, CurrentID=0, Status='live', Symbol='LOC105378576', Aliases='', description='uncharacterized LOC105378576', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133462356, end_position_on_the_genomic_accession=133464555, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378577, CurrentID=0, Status='live', Symbol='LOC105378577', Aliases='', description='basic proline-rich protein-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46283257, end_position_on_the_genomic_accession=46285118, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389906 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389906, CurrentID=0, Status='live', Symbol='LOC389906', Aliases='', description='zinc finger protein 839 pseudogene', other_designations='', map_location='Xp22.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=3817535, end_position_on_the_genomic_accession=3843894, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378579, CurrentID=0, Status='live', Symbol='LOC105378579', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=119198557, end_position_on_the_genomic_accession=119207578, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378580, CurrentID=0, Status='live', Symbol='LOC105378580', Aliases='', description='uncharacterized LOC105378580', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=803398, end_position_on_the_genomic_accession=805130, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378581, CurrentID=0, Status='live', Symbol='LOC105378581', Aliases='', description='uncharacterized LOC105378581', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=778798, end_position_on_the_genomic_accession=784251, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378582, CurrentID=0, Status='live', Symbol='LOC105378582', Aliases='', description='uncharacterized LOC105378582', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=688825, end_position_on_the_genomic_accession=732141, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378583, CurrentID=0, Status='live', Symbol='LOC105378583', Aliases='', description='uncharacterized LOC105378583', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=982073, end_position_on_the_genomic_accession=989979, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378584, CurrentID=0, Status='live', Symbol='LOC105378584', Aliases='', description='uncharacterized LOC105378584', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1093333, end_position_on_the_genomic_accession=1094038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378585 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378585, CurrentID=0, Status='live', Symbol='LOC105378585', Aliases='', description='uncharacterized LOC105378585', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1420245, end_position_on_the_genomic_accession=1423287, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378586, CurrentID=0, Status='live', Symbol='LOC105378586', Aliases='', description='uncharacterized LOC105378586', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1600288, end_position_on_the_genomic_accession=1607378, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378587, CurrentID=0, Status='live', Symbol='LOC105378587', Aliases='', description='uncharacterized LOC105378587', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1625285, end_position_on_the_genomic_accession=1626651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378588, CurrentID=0, Status='live', Symbol='LOC105378588', Aliases='', description='uncharacterized LOC105378588', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1745725, end_position_on_the_genomic_accession=1751773, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378589, CurrentID=0, Status='live', Symbol='LOC105378589', Aliases='', description='uncharacterized LOC105378589', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2013222, end_position_on_the_genomic_accession=2016584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378590, CurrentID=0, Status='live', Symbol='LOC105378590', Aliases='', description='uncharacterized LOC105378590', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2034986, end_position_on_the_genomic_accession=2047444, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378591, CurrentID=0, Status='live', Symbol='LOC105378591', Aliases='', description='uncharacterized LOC105378591', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2049201, end_position_on_the_genomic_accession=2050143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378592, CurrentID=0, Status='live', Symbol='LOC105378592', Aliases='', description='uncharacterized protein FLJ14100', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2208942, end_position_on_the_genomic_accession=2215211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378593, CurrentID=0, Status='live', Symbol='LOC105378593', Aliases='', description='uncharacterized LOC105378593', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2207013, end_position_on_the_genomic_accession=2208854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378594, CurrentID=0, Status='live', Symbol='LOC105378594', Aliases='', description='nuclear speckle splicing regulatory protein 1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2280383, end_position_on_the_genomic_accession=2284676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378595, CurrentID=0, Status='live', Symbol='LOC105378595', Aliases='', description='uncharacterized LOC105378595', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2314980, end_position_on_the_genomic_accession=2319268, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378596 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378596, CurrentID=0, Status='live', Symbol='LOC105378596', Aliases='', description='uncharacterized LOC105378596', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2567830, end_position_on_the_genomic_accession=2570106, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378597, CurrentID=0, Status='live', Symbol='LOC105378597', Aliases='', description='uncharacterized LOC105378597', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2633111, end_position_on_the_genomic_accession=2636620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  652070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652070, CurrentID=0, Status='live', Symbol='SCFV', Aliases='', description='single-chain Fv fragment', other_designations='immunglobulin heavy chain variable region', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378599, CurrentID=0, Status='live', Symbol='LOC105378599', Aliases='', description='uncharacterized LOC105378599', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2773495, end_position_on_the_genomic_accession=2774394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378600, CurrentID=0, Status='live', Symbol='LOC105378600', Aliases='', description='uncharacterized LOC105378600', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2766073, end_position_on_the_genomic_accession=2766810, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378601, CurrentID=0, Status='live', Symbol='LOC105378601', Aliases='', description='uncharacterized LOC105378601', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2762488, end_position_on_the_genomic_accession=2763382, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378602, CurrentID=0, Status='live', Symbol='LOC105378602', Aliases='', description='uncharacterized LOC105378602', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2701552, end_position_on_the_genomic_accession=2702276, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378603, CurrentID=0, Status='live', Symbol='LOC105378603', Aliases='', description='uncharacterized LOC105378603', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=2700560, end_position_on_the_genomic_accession=2702756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389932, CurrentID=0, Status='live', Symbol='AKR1C6P', Aliases='TAKR', description='aldo-keto reductase family 1, member C6, pseudogene', other_designations='aldo-keto reductase, truncated|truncated AKR|truncated aldo-keto reductase', map_location='10p15.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=4871667, end_position_on_the_genomic_accession=4916273, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105378605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378605, CurrentID=0, Status='live', Symbol='LOC105378605', Aliases='', description='uncharacterized LOC105378605', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3210736, end_position_on_the_genomic_accession=3213909, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378606, CurrentID=0, Status='live', Symbol='LOC105378606', Aliases='', description='uncharacterized LOC105378606', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3161279, end_position_on_the_genomic_accession=3167573, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  389935 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389935, CurrentID=0, Status='live', Symbol='PRPF38AP1', Aliases='', description='PRP38 domain containing A pseudogene 1', other_designations='PRP38 pre-mRNA processing factor 38 domain containing A pseudogene', map_location='10p14', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=8161260, end_position_on_the_genomic_accession=8162194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378608, CurrentID=0, Status='live', Symbol='LOC105378608', Aliases='', description='uncharacterized LOC105378608', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3546961, end_position_on_the_genomic_accession=3550526, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378609, CurrentID=0, Status='live', Symbol='LOC105378609', Aliases='', description='uncharacterized LOC105378609', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3674735, end_position_on_the_genomic_accession=3677526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378610, CurrentID=0, Status='live', Symbol='LOC105378610', Aliases='', description='uncharacterized LOC105378610', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=3658649, end_position_on_the_genomic_accession=3665313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127795, CurrentID=0, Status='live', Symbol='C1orf87', Aliases='CREF', description='chromosome 1 open reading frame 87', other_designations='carcinoma-related EF-hand protein', map_location='1p32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59990394, end_position_on_the_genomic_accession=60073770, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105378612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378612, CurrentID=0, Status='live', Symbol='LOC105378612', Aliases='', description='uncharacterized LOC105378612', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13670248, end_position_on_the_genomic_accession=13674725, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389941 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389941, CurrentID=0, Status='live', Symbol='C1QL3', Aliases='C1QTNF13, C1ql, CTRP13, K100', description='complement component 1, q subcomponent-like 3', other_designations='C1q and tumor necrosis factor-related protein 13|C1q/TNF-related protein 13', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=16513743, end_position_on_the_genomic_accession=16522005, orientation='minus', exon_count=2, OMIM=615227, no_hdr0=''),
  105378614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378614, CurrentID=0, Status='live', Symbol='LOC105378614', Aliases='', description='uncharacterized LOC105378614', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=19591802, end_position_on_the_genomic_accession=19596872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378615, CurrentID=0, Status='live', Symbol='LOC105378615', Aliases='', description='uncharacterized LOC105378615', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23987408, end_position_on_the_genomic_accession=23987788, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378616, CurrentID=0, Status='live', Symbol='LOC105378616', Aliases='', description='uncharacterized LOC105378616', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28643433, end_position_on_the_genomic_accession=28648731, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101929780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929780, CurrentID=0, Status='live', Symbol='NBPF25P', Aliases='WI2-925H4.1', description='neuroblastoma breakpoint family member 25, pseudogene', other_designations='neuroblastoma breakpoint family member 25 pseudogene|neuroblastoma breakpoint family, member 25, pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145572345, end_position_on_the_genomic_accession=145607858, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  105378618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378618, CurrentID=0, Status='live', Symbol='LOC105378618', Aliases='', description='uncharacterized LOC105378618', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30253430, end_position_on_the_genomic_accession=30261152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378619, CurrentID=0, Status='live', Symbol='LOC105378619', Aliases='', description='uncharacterized LOC105378619', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30410930, end_position_on_the_genomic_accession=30412521, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389948, CurrentID=0, Status='live', Symbol='MKNK2P1', Aliases='', description='MAP kinase interacting serine/threonine kinase 2 pseudogene 1', other_designations='', map_location='10p11.21', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=36944059, end_position_on_the_genomic_accession=36944672, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378621 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378621, CurrentID=0, Status='live', Symbol='LOC105378621', Aliases='', description='uncharacterized LOC105378621', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30775107, end_position_on_the_genomic_accession=30782380, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378622, CurrentID=0, Status='live', Symbol='LOC105378622', Aliases='', description='uncharacterized LOC105378622', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=30824139, end_position_on_the_genomic_accession=30834282, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378623, CurrentID=0, Status='live', Symbol='LOC105378623', Aliases='', description='uncharacterized LOC105378623', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31383332, end_position_on_the_genomic_accession=31389830, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378624, CurrentID=0, Status='live', Symbol='LOC105378624', Aliases='', description='uncharacterized LOC105378624', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31459330, end_position_on_the_genomic_accession=31471520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378625, CurrentID=0, Status='live', Symbol='LOC105378625', Aliases='', description='uncharacterized LOC105378625', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31448374, end_position_on_the_genomic_accession=31449586, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378626, CurrentID=0, Status='live', Symbol='LOC105378626', Aliases='', description='uncharacterized LOC105378626', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31568236, end_position_on_the_genomic_accession=31577922, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378627, CurrentID=0, Status='live', Symbol='LOC105378627', Aliases='', description='uncharacterized LOC105378627', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31549577, end_position_on_the_genomic_accession=31568204, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378628, CurrentID=0, Status='live', Symbol='LOC105378628', Aliases='', description='uncharacterized LOC105378628', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=31650457, end_position_on_the_genomic_accession=31651485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378629, CurrentID=0, Status='live', Symbol='LOC105378629', Aliases='', description='uncharacterized LOC105378629', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32343136, end_position_on_the_genomic_accession=32351795, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378630, CurrentID=0, Status='live', Symbol='LOC105378630', Aliases='', description='uncharacterized LOC105378630', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32394513, end_position_on_the_genomic_accession=32395663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378631, CurrentID=0, Status='live', Symbol='LOC105378631', Aliases='', description='uncharacterized LOC105378631', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32876027, end_position_on_the_genomic_accession=32882505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378632, CurrentID=0, Status='live', Symbol='LOC105378632', Aliases='', description='uncharacterized LOC105378632', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32901660, end_position_on_the_genomic_accession=32903203, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378633 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378633, CurrentID=0, Status='live', Symbol='LOC105378633', Aliases='', description='metallothionein-1E pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32951279, end_position_on_the_genomic_accession=32951510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378634, CurrentID=0, Status='live', Symbol='LOC105378634', Aliases='', description='uncharacterized LOC105378634', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32927453, end_position_on_the_genomic_accession=32934354, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378635, CurrentID=0, Status='live', Symbol='LOC105378635', Aliases='', description='uncharacterized LOC105378635', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33077884, end_position_on_the_genomic_accession=33081102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100271561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100271561, CurrentID=0, Status='live', Symbol='RPS20P29', Aliases='RPS20_14_1201', description='ribosomal protein S20 pseudogene 29', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=7860449, end_position_on_the_genomic_accession=7860764, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378637, CurrentID=0, Status='live', Symbol='LOC105378637', Aliases='', description='uncharacterized LOC105378637', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33790991, end_position_on_the_genomic_accession=33804803, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378638, CurrentID=0, Status='live', Symbol='LOC105378638', Aliases='', description='uncharacterized LOC105378638', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=33605002, end_position_on_the_genomic_accession=33610152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378639, CurrentID=0, Status='live', Symbol='LOC105378639', Aliases='', description='uncharacterized LOC105378639', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34279848, end_position_on_the_genomic_accession=34295942, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378640, CurrentID=0, Status='live', Symbol='LOC105378640', Aliases='', description='uncharacterized LOC105378640', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34364822, end_position_on_the_genomic_accession=34366065, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378641, CurrentID=0, Status='live', Symbol='LOC105378641', Aliases='', description='uncharacterized LOC105378641', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34457859, end_position_on_the_genomic_accession=34686927, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  389970 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389970, CurrentID=0, Status='live', Symbol='GAPDHP21', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 21', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='10q21.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=55667138, end_position_on_the_genomic_accession=55668420, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378643, CurrentID=0, Status='live', Symbol='LOC105378643', Aliases='', description='uncharacterized LOC105378643', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=34742408, end_position_on_the_genomic_accession=34799371, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378644, CurrentID=0, Status='live', Symbol='LOC105378644', Aliases='', description='uncharacterized LOC105378644', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35141439, end_position_on_the_genomic_accession=35146825, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127829, CurrentID=0, Status='live', Symbol='ARL8A', Aliases='ARL10B, GIE2', description='ADP ribosylation factor like GTPase 8A', other_designations='ADP-ribosylation factor like GTPase 8A|ADP-ribosylation factor-like 10B|ADP-ribosylation factor-like 8A|novel small G protein indispensable for equal chromosome segregation 2', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202133404, end_position_on_the_genomic_accession=202144743, orientation='minus', exon_count=7, OMIM=616597, no_hdr0=''),
  105378646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378646, CurrentID=0, Status='live', Symbol='LOC105378646', Aliases='', description='uncharacterized LOC105378646', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=35665308, end_position_on_the_genomic_accession=35685441, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378647, CurrentID=0, Status='live', Symbol='LOC105378647', Aliases='', description='uncharacterized LOC105378647', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=36009604, end_position_on_the_genomic_accession=36019780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  389976 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389976, CurrentID=0, Status='live', Symbol='ACTBP14', Aliases='', description='actin, beta pseudogene 14', other_designations='actin, gamma 1 pseudogene', map_location='10q22.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=69022778, end_position_on_the_genomic_accession=69023866, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  127833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127833, CurrentID=0, Status='live', Symbol='SYT2', Aliases='CMS7, MYSPC, SytII', description='synaptotagmin 2', other_designations='synaptotagmin II', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202587637, end_position_on_the_genomic_accession=202710444, orientation='minus', exon_count=10, OMIM=600104, no_hdr0=''),
  105378650 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378650, CurrentID=0, Status='live', Symbol='LOC105378650', Aliases='', description='uncharacterized LOC105378650', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37765729, end_position_on_the_genomic_accession=37767313, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378651, CurrentID=0, Status='live', Symbol='LOC105378651', Aliases='', description='uncharacterized LOC105378651', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37932061, end_position_on_the_genomic_accession=37945314, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378652, CurrentID=0, Status='live', Symbol='LOC105378652', Aliases='', description='uncharacterized LOC105378652', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37910157, end_position_on_the_genomic_accession=37918656, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422458, CurrentID=0, Status='live', Symbol='NUDT19P2', Aliases='', description='nudix hydrolase 19 pseudogene 2', other_designations='nudix (nucleoside diphosphate linked moiety X)-type motif 19 pseudogene 2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=103007071, end_position_on_the_genomic_accession=103008064, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378654, CurrentID=0, Status='live', Symbol='LOC105378654', Aliases='', description='uncharacterized LOC105378654', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38141654, end_position_on_the_genomic_accession=38213095, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378655, CurrentID=0, Status='live', Symbol='LOC105378655', Aliases='', description='uncharacterized LOC105378655', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38129480, end_position_on_the_genomic_accession=38141542, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378656, CurrentID=0, Status='live', Symbol='LOC105378656', Aliases='', description='uncharacterized LOC105378656', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38475196, end_position_on_the_genomic_accession=38476255, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  127841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127841, CurrentID=0, Status='live', Symbol='LINC00628', Aliases='', description='long intergenic non-protein coding RNA 628', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204368430, end_position_on_the_genomic_accession=204369719, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378658, CurrentID=0, Status='live', Symbol='LOC105378658', Aliases='', description='uncharacterized LOC105378658', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38518246, end_position_on_the_genomic_accession=38532800, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378659, CurrentID=0, Status='live', Symbol='LOC105378659', Aliases='', description='uncharacterized LOC105378659', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38553525, end_position_on_the_genomic_accession=38558188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378660, CurrentID=0, Status='live', Symbol='LOC105378660', Aliases='', description='uncharacterized LOC105378660', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38692952, end_position_on_the_genomic_accession=38728533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  127845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127845, CurrentID=0, Status='live', Symbol='GOLT1A', Aliases='CGI-141, GOT1, HMFN1187, YMR292W', description='golgi transport 1A', other_designations='hGOT1b|putative protein product of HMFN1187', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=204198160, end_position_on_the_genomic_accession=204214092, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  389990 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389990, CurrentID=0, Status='live', Symbol='RPA2P2', Aliases='', description='replication protein A2 pseudogene 2', other_designations='replication protein A2, 32kDa pseudogene', map_location='10q23.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=81136543, end_position_on_the_genomic_accession=81137323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378663, CurrentID=0, Status='live', Symbol='LOC105378663', Aliases='', description='uncharacterized LOC105378663', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=38859918, end_position_on_the_genomic_accession=38876226, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  389992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389992, CurrentID=0, Status='live', Symbol='RPL7P34', Aliases='RPL7_14_1078', description='ribosomal protein L7 pseudogene 34', other_designations='', map_location='10q23.31', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=88618582, end_position_on_the_genomic_accession=88618935, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378665, CurrentID=0, Status='live', Symbol='LOC105378665', Aliases='', description='uncharacterized LOC105378665', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39808159, end_position_on_the_genomic_accession=39817442, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378666, CurrentID=0, Status='live', Symbol='LOC105378666', Aliases='', description='uncharacterized LOC105378666', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39779984, end_position_on_the_genomic_accession=39781424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378667, CurrentID=0, Status='live', Symbol='LOC105378667', Aliases='', description='uncharacterized LOC105378667', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39824055, end_position_on_the_genomic_accession=39838083, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378668, CurrentID=0, Status='live', Symbol='LOC105378668', Aliases='', description='uncharacterized LOC105378668', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39897741, end_position_on_the_genomic_accession=39899098, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  389997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=389997, CurrentID=0, Status='live', Symbol='NIP7P1', Aliases='', description='NIP7, nucleolar pre-rRNA processing protein pseudogene 1', other_designations='nuclear import 7 homolog pseudogene', map_location='10q23.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=93105921, end_position_on_the_genomic_accession=93107289, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378670, CurrentID=0, Status='live', Symbol='LOC105378670', Aliases='', description='uncharacterized LOC105378670', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40023512, end_position_on_the_genomic_accession=40039121, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378671, CurrentID=0, Status='live', Symbol='LOC105378671', Aliases='', description='uncharacterized LOC105378671', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40147701, end_position_on_the_genomic_accession=40161312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378672, CurrentID=0, Status='live', Symbol='LOC105378672', Aliases='', description='uncharacterized LOC105378672', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40487269, end_position_on_the_genomic_accession=40508671, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378673 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378673, CurrentID=0, Status='live', Symbol='LOC105378673', Aliases='', description='uncharacterized LOC105378673', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40563485, end_position_on_the_genomic_accession=40568562, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378674, CurrentID=0, Status='live', Symbol='LOC105378674', Aliases='', description='uncharacterized LOC105378674', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40668540, end_position_on_the_genomic_accession=40673849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390003, CurrentID=0, Status='live', Symbol='GUCY2GP', Aliases='GUCY2G', description='guanylate cyclase 2G, pseudogene', other_designations='guanylate cyclase 2G homolog, pseudogene', map_location='10q25.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=112308178, end_position_on_the_genomic_accession=112356595, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105378676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378676, CurrentID=0, Status='live', Symbol='LOC105378676', Aliases='', description='uncharacterized LOC105378676', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=40847981, end_position_on_the_genomic_accession=40857494, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378677, CurrentID=0, Status='live', Symbol='LOC105378677', Aliases='', description='uncharacterized LOC105378677', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=41427201, end_position_on_the_genomic_accession=41429394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390006, CurrentID=0, Status='live', Symbol='PPIAP19', Aliases='bA108K1.2', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 19', other_designations='', map_location='10q25.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=114689943, end_position_on_the_genomic_accession=114690683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390007, CurrentID=0, Status='live', Symbol='TOMM22P5', Aliases='', description='TOMM22 pseudogene 5', other_designations='translocase of outer mitochondrial membrane 22 homolog pseudogene 5', map_location='10q26.11', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=118643759, end_position_on_the_genomic_accession=118644394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378680 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378680, CurrentID=0, Status='live', Symbol='LOC105378680', Aliases='', description='uncharacterized LOC105378680', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42311267, end_position_on_the_genomic_accession=42323675, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378681, CurrentID=0, Status='live', Symbol='LOC105378681', Aliases='', description='uncharacterized LOC105378681', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42680158, end_position_on_the_genomic_accession=42681381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390010 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390010, CurrentID=0, Status='live', Symbol='NKX1-2', Aliases='C10orf121, NKX-1.1, bB238F13.2', description='NK1 homeobox 2', other_designations='NK1 transcription factor related, locus 2|homeobox protein SAX-1', map_location='10q26.13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=124447429, end_position_on_the_genomic_accession=124449981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378683, CurrentID=0, Status='live', Symbol='LOC105378683', Aliases='', description='uncharacterized LOC105378683', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42775877, end_position_on_the_genomic_accession=42784890, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390012 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390012, CurrentID=0, Status='live', Symbol='OR6L1P', Aliases='', description='olfactory receptor family 6 subfamily L member 1 pseudogene', other_designations='olfactory receptor, family 6, subfamily L, member 1 pseudogene', map_location='10q26.3', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=133575056, end_position_on_the_genomic_accession=133576135, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378685, CurrentID=0, Status='live', Symbol='LOC105378685', Aliases='', description='uncharacterized LOC105378685', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43181692, end_position_on_the_genomic_accession=43250544, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105378686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378686, CurrentID=0, Status='live', Symbol='LOC105378686', Aliases='', description='uncharacterized LOC105378686', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43172776, end_position_on_the_genomic_accession=43175257, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378687, CurrentID=0, Status='live', Symbol='LOC105378687', Aliases='', description='uncharacterized LOC105378687', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43354258, end_position_on_the_genomic_accession=43358630, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378688, CurrentID=0, Status='live', Symbol='LOC105378688', Aliases='', description='uncharacterized LOC105378688', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=43937892, end_position_on_the_genomic_accession=43940543, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378689, CurrentID=0, Status='live', Symbol='LOC105378689', Aliases='', description='uncharacterized LOC105378689', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44172746, end_position_on_the_genomic_accession=44174548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378690 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378690, CurrentID=0, Status='live', Symbol='LOC105378690', Aliases='', description='uncharacterized LOC105378690', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44688341, end_position_on_the_genomic_accession=44724114, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378691, CurrentID=0, Status='live', Symbol='LOC105378691', Aliases='', description='uncharacterized LOC105378691', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45296584, end_position_on_the_genomic_accession=45303663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378692, CurrentID=0, Status='live', Symbol='LOC105378692', Aliases='', description='uncharacterized LOC105378692', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45503889, end_position_on_the_genomic_accession=45504329, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378693, CurrentID=0, Status='live', Symbol='LOC105378693', Aliases='', description='uncharacterized LOC105378693', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=45750896, end_position_on_the_genomic_accession=45760837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  106481095 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481095, CurrentID=0, Status='live', Symbol='RN7SL632P', Aliases='', description='RNA, 7SL, cytoplasmic 632, pseudogene', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378695 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378695, CurrentID=0, Status='live', Symbol='LOC105378695', Aliases='', description='uncharacterized LOC105378695', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46171283, end_position_on_the_genomic_accession=46175408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378696, CurrentID=0, Status='live', Symbol='LOC105378696', Aliases='', description='uncharacterized LOC105378696', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46533462, end_position_on_the_genomic_accession=46535497, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378697, CurrentID=0, Status='live', Symbol='LOC105378697', Aliases='', description='uncharacterized LOC105378697', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46621695, end_position_on_the_genomic_accession=46624879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  652170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652170, CurrentID=0, Status='live', Symbol='COX4I1P2', Aliases='', description='cytochrome c oxidase subunit IV isoform 1 pseudogene 2', other_designations='cytochrome c oxidase subunit IV pseudogene', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45671845, end_position_on_the_genomic_accession=45672391, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378699, CurrentID=0, Status='live', Symbol='LOC105378699', Aliases='', description='uncharacterized LOC105378699', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=46765057, end_position_on_the_genomic_accession=46767638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390028, CurrentID=0, Status='live', Symbol='OR4F2P', Aliases='HS191N21, OR4F2, S191N21, hs6M1-11', description='olfactory receptor family 4 subfamily F member 2 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 2 pseudogene', map_location='11p15.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86649, end_position_on_the_genomic_accession=87586, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378701, CurrentID=0, Status='live', Symbol='LOC105378701', Aliases='', description='uncharacterized LOC105378701', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=47172851, end_position_on_the_genomic_accession=47177065, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390030 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390030, CurrentID=0, Status='live', Symbol='OR55B1P', Aliases='OR55B2P, OR55C1P', description='olfactory receptor family 55 subfamily B member 1 pseudogene', other_designations='olfactory receptor, family 55, subfamily B, member 1 pseudogene|olfactory receptor, family 55, subfamily B, member 2 pseudogene|olfactory receptor, family 55, subfamily C, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4146315, end_position_on_the_genomic_accession=4147294, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390031 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390031, CurrentID=0, Status='live', Symbol='SSU72P2', Aliases='', description='SSU72 pseudogene 2', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4242056, end_position_on_the_genomic_accession=4242650, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378704, CurrentID=0, Status='live', Symbol='LOC105378704', Aliases='', description='uncharacterized LOC105378704', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48100475, end_position_on_the_genomic_accession=48161444, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390033, CurrentID=0, Status='live', Symbol='SSU72P7', Aliases='', description='SSU72 pseudogene 7', other_designations='SSU72 RNA polymerase II CTD phosphatase homolog pseudogene 7', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4338660, end_position_on_the_genomic_accession=4339244, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390034, CurrentID=0, Status='live', Symbol='OR51R1P', Aliases='', description='olfactory receptor family 51 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 51, subfamily R, member 1 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4418060, end_position_on_the_genomic_accession=4419008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390035 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390035, CurrentID=0, Status='live', Symbol='OR52K3P', Aliases='', description='olfactory receptor family 52 subfamily K member 3 pseudogene', other_designations='olfactory receptor, family 52, subfamily K, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4474810, end_position_on_the_genomic_accession=4475756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390036 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390036, CurrentID=0, Status='live', Symbol='OR52K1', Aliases='OR11-8', description='olfactory receptor family 52 subfamily K member 1', other_designations='olfactory receptor OR11-8|olfactory receptor, family 52, subfamily K, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4488879, end_position_on_the_genomic_accession=4489908, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390037, CurrentID=0, Status='live', Symbol='OR52I1', Aliases='OR11-13', description='olfactory receptor family 52 subfamily I member 1', other_designations='olfactory receptor OR11-13|olfactory receptor, family 52, subfamily I, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4594039, end_position_on_the_genomic_accession=4595013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390038 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390038, CurrentID=0, Status='live', Symbol='OR51D1', Aliases='OR11-14Q, OR51D1', description='olfactory receptor family 51 subfamily D member 1', other_designations='olfactory receptor OR11-14|olfactory receptor, family 51, subfamily D, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4639715, end_position_on_the_genomic_accession=4640838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390039, CurrentID=0, Status='live', Symbol='OR51A9P', Aliases='', description='olfactory receptor family 51 subfamily A member 9 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 9 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4660831, end_position_on_the_genomic_accession=4661729, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390040, CurrentID=0, Status='live', Symbol='OR51F5P', Aliases='', description='olfactory receptor family 51 subfamily F member 5 pseudogene', other_designations='olfactory receptor, family 51, subfamily F, member 5 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4709569, end_position_on_the_genomic_accession=4710510, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390041, CurrentID=0, Status='live', Symbol='OR51C4P', Aliases='', description='olfactory receptor family 51 subfamily C member 4 pseudogene', other_designations='olfactory receptor, family 51, subfamily C, member 4 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4718612, end_position_on_the_genomic_accession=4719559, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378714, CurrentID=0, Status='live', Symbol='LOC105378714', Aliases='', description='uncharacterized LOC105378714', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=50423505, end_position_on_the_genomic_accession=50425355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390043, CurrentID=0, Status='live', Symbol='OR51F3P', Aliases='', description='olfactory receptor family 51 subfamily F member 3 pseudogene', other_designations='olfactory receptor, family 51, subfamily F, member 3 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4736177, end_position_on_the_genomic_accession=4737077, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390044, CurrentID=0, Status='live', Symbol='OR51F4P', Aliases='', description='olfactory receptor family 51 subfamily F member 4 pseudogene', other_designations='olfactory receptor, family 51, subfamily F, member 4 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4752045, end_position_on_the_genomic_accession=4752943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378717 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378717, CurrentID=0, Status='live', Symbol='LOC105378717', Aliases='', description='uncharacterized LOC105378717', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51326753, end_position_on_the_genomic_accession=51328589, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390046, CurrentID=0, Status='live', Symbol='OR51A6P', Aliases='', description='olfactory receptor family 51 subfamily A member 6 pseudogene', other_designations='olfactory receptor, family 51, subfamily A, member 6 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=4889577, end_position_on_the_genomic_accession=4890595, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422469, CurrentID=0, Status='live', Symbol='B3GNT2P1', Aliases='', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 2 pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=99886625, end_position_on_the_genomic_accession=99888106, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378720, CurrentID=0, Status='live', Symbol='LOC105378720', Aliases='', description='uncharacterized LOC105378720', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51516613, end_position_on_the_genomic_accession=51600488, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105378721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378721, CurrentID=0, Status='live', Symbol='LOC105378721', Aliases='', description='uncharacterized LOC105378721', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51793990, end_position_on_the_genomic_accession=51798131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378722, CurrentID=0, Status='live', Symbol='LOC105378722', Aliases='', description='protein argonaute 1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52552934, end_position_on_the_genomic_accession=52565732, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105378723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378723, CurrentID=0, Status='live', Symbol='LOC105378723', Aliases='', description='uncharacterized LOC105378723', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52530967, end_position_on_the_genomic_accession=52552709, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378724, CurrentID=0, Status='live', Symbol='LOC105378724', Aliases='', description='uncharacterized LOC105378724', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53180461, end_position_on_the_genomic_accession=53191428, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390053, CurrentID=0, Status='live', Symbol='OR52A4P', Aliases='OR52A4', description='olfactory receptor family 52 subfamily A member 4 pseudogene', other_designations='olfactory receptor, family 52, subfamily A, member 4, pseudogene|putative olfactory receptor 52A4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5120664, end_position_on_the_genomic_accession=5121578, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390054, CurrentID=0, Status='live', Symbol='OR52A5', Aliases='OR11-33', description='olfactory receptor family 52 subfamily A member 5', other_designations="odorant receptor HOR3'beta5|olfactory receptor OR11-33|olfactory receptor, family 52, subfamily A, member 5", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5131692, end_position_on_the_genomic_accession=5132642, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100529063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529063, CurrentID=0, Status='live', Symbol='BCL2L2-PABPN1', Aliases='', description='BCL2L2-PABPN1 readthrough', other_designations='', map_location='14q', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=23306762, end_position_on_the_genomic_accession=23326185, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105378728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378728, CurrentID=0, Status='live', Symbol='LOC105378728', Aliases='', description='uncharacterized LOC105378728', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53255174, end_position_on_the_genomic_accession=53269366, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378729, CurrentID=0, Status='live', Symbol='LOC105378729', Aliases='', description='uncharacterized LOC105378729', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53348296, end_position_on_the_genomic_accession=53350340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390058, CurrentID=0, Status='live', Symbol='OR51B6', Aliases="HOR5'Beta6", description='olfactory receptor family 51 subfamily B member 6', other_designations="odorant receptor HOR5'beta6|olfactory receptor 51B5|olfactory receptor, family 51, subfamily B, member 6", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5351508, end_position_on_the_genomic_accession=5352446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390059, CurrentID=0, Status='live', Symbol='OR51M1', Aliases="HOR5'Beta7, OR11-40", description='olfactory receptor family 51 subfamily M member 1', other_designations="odorant receptor HOR5'beta7|olfactory receptor OR11-40|olfactory receptor, family 51, subfamily M, member 1", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5389377, end_position_on_the_genomic_accession=5390434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378732, CurrentID=0, Status='live', Symbol='LOC105378732', Aliases='', description='uncharacterized LOC105378732', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=53328230, end_position_on_the_genomic_accession=53337109, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390061, CurrentID=0, Status='live', Symbol='OR51Q1', Aliases='', description='olfactory receptor family 51 subfamily Q member 1 (gene/pseudogene)', other_designations='olfactory receptor, family 51, subfamily Q, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5422111, end_position_on_the_genomic_accession=5423206, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378734, CurrentID=0, Status='live', Symbol='LOC105378734', Aliases='', description='uncharacterized LOC105378734', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=54516412, end_position_on_the_genomic_accession=54528746, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390063, CurrentID=0, Status='live', Symbol='OR51I1', Aliases='', description='olfactory receptor family 51 subfamily I member 1', other_designations="HOR5'Beta11|odorant receptor HOR5'beta11|olfactory receptor OR11-39|olfactory receptor, family 51, subfamily I, member 1", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5440542, end_position_on_the_genomic_accession=5441553, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390064, CurrentID=0, Status='live', Symbol='OR51I2', Aliases='OR11-38', description='olfactory receptor family 51 subfamily I member 2', other_designations="HOR5'Beta12|odorant receptor HOR5'beta12|olfactory receptor OR11-38|olfactory receptor, family 51, subfamily I, member 2", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5453408, end_position_on_the_genomic_accession=5454477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378737, CurrentID=0, Status='live', Symbol='LOC105378737', Aliases='', description='uncharacterized LOC105378737', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55953137, end_position_on_the_genomic_accession=55975733, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390066 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390066, CurrentID=0, Status='live', Symbol='OR52D1', Aliases='OR11-43', description='olfactory receptor family 52 subfamily D member 1', other_designations="HOR 5'Beta14|odorant receptor HOR5'beta14|olfactory receptor OR11-43|olfactory receptor, family 52, subfamily D, member 1", map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5488685, end_position_on_the_genomic_accession=5489749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390067, CurrentID=0, Status='live', Symbol='OR52H1', Aliases='OR11-41, OR11-45', description='olfactory receptor family 52 subfamily H member 1', other_designations='olfactory receptor OR11-41 pseudogene|olfactory receptor OR11-45|olfactory receptor, family 52, subfamily H, member 1', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5544561, end_position_on_the_genomic_accession=5545523, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378740, CurrentID=0, Status='live', Symbol='LOC105378740', Aliases='', description='uncharacterized LOC105378740', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=55783681, end_position_on_the_genomic_accession=55854984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378741, CurrentID=0, Status='live', Symbol='LOC105378741', Aliases='', description='uncharacterized LOC105378741', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56045841, end_position_on_the_genomic_accession=56060859, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378742, CurrentID=0, Status='live', Symbol='LOC105378742', Aliases='', description='uncharacterized LOC105378742', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=56154655, end_position_on_the_genomic_accession=56194042, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101929801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929801, CurrentID=0, Status='live', Symbol='FAM91A2P', Aliases='', description='family with sequence similarity 91 member A2, pseudogene', other_designations='family with sequence similarity 91, member A1 pseudogene|family with sequence similarity 91, member A2 pseudogene', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149675140, end_position_on_the_genomic_accession=149680481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390072, CurrentID=0, Status='live', Symbol='OR52N4', Aliases='OR11-64', description='olfactory receptor family 52 subfamily N member 4 (gene/pseudogene)', other_designations='olfactory receptor OR11-64|olfactory receptor, family 52, subfamily N, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5722259, end_position_on_the_genomic_accession=5755958, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  390073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390073, CurrentID=0, Status='live', Symbol='OR56B2P', Aliases='OR11-63, OR56B2', description='olfactory receptor family 56 subfamily B member 2 pseudogene', other_designations='Olfactory receptor 56B2|olfactory receptor OR11-63|olfactory receptor, family 56, subfamily B, member 2 pseudogene|seven transmembrane helix receptor', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5765161, end_position_on_the_genomic_accession=5766045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378746, CurrentID=0, Status='live', Symbol='LOC105378746', Aliases='', description='uncharacterized LOC105378746', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=57953813, end_position_on_the_genomic_accession=57991232, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  390075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390075, CurrentID=0, Status='live', Symbol='OR52N5', Aliases='OR11-62Q, OR52N5', description='olfactory receptor family 52 subfamily N member 5', other_designations='olfactory receptor OR11-62|olfactory receptor, family 52, subfamily N, member 5', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5777634, end_position_on_the_genomic_accession=5778667, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378748, CurrentID=0, Status='live', Symbol='LOC105378748', Aliases='', description='uncharacterized LOC105378748', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=57017405, end_position_on_the_genomic_accession=57049117, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  390077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390077, CurrentID=0, Status='live', Symbol='OR52N2', Aliases='OR11-57', description='olfactory receptor family 52 subfamily N member 2', other_designations='olfactory receptor OR11-57|olfactory receptor, family 52, subfamily N, member 2', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5808905, end_position_on_the_genomic_accession=5821400, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390078, CurrentID=0, Status='live', Symbol='OR52E6', Aliases='OR11-58', description='olfactory receptor family 52 subfamily E member 6', other_designations='olfactory receptor OR11-58|olfactory receptor, family 52, subfamily E, member 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5840956, end_position_on_the_genomic_accession=5841897, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390079, CurrentID=0, Status='live', Symbol='OR52E8', Aliases='OR11-54', description='olfactory receptor family 52 subfamily E member 8', other_designations='olfactory receptor OR11-54|olfactory receptor, family 52, subfamily E, member 8 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5856749, end_position_on_the_genomic_accession=5857702, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378752, CurrentID=0, Status='live', Symbol='LOC105378752', Aliases='', description='uncharacterized LOC105378752', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59011497, end_position_on_the_genomic_accession=59029907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390081, CurrentID=0, Status='live', Symbol='OR52E4', Aliases='OR11-55', description='olfactory receptor family 52 subfamily E member 4', other_designations='olfactory receptor OR11-55|olfactory receptor, family 52, subfamily E, member 4', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5880630, end_position_on_the_genomic_accession=5885294, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390082, CurrentID=0, Status='live', Symbol='OR52E5', Aliases='OR11-56', description='olfactory receptor family 52 subfamily E member 5', other_designations='olfactory receptor OR11-56 pseudogene|olfactory receptor, family 52, subfamily E, member 5', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5900777, end_position_on_the_genomic_accession=5901925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390083 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390083, CurrentID=0, Status='live', Symbol='OR56A3', Aliases='OR56A2PP, OR56A6, OR56A3', description='olfactory receptor family 56 subfamily A member 3', other_designations='olfactory receptor 56A6|olfactory receptor OR11-51|olfactory receptor, family 56, subfamily A, member 2 pseudogene|olfactory receptor, family 56, subfamily A, member 3 pseudogene|olfactory receptor, family 56, subfamily A, member 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5947347, end_position_on_the_genomic_accession=5948294, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390084, CurrentID=0, Status='live', Symbol='OR56A5', Aliases='OR56A5P', description='olfactory receptor family 56 subfamily A member 5', other_designations='olfactory receptor, family 56, subfamily A, member 5 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=5967553, end_position_on_the_genomic_accession=5968494, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378757, CurrentID=0, Status='live', Symbol='LOC105378757', Aliases='', description='uncharacterized LOC105378757', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59766771, end_position_on_the_genomic_accession=59807539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378758, CurrentID=0, Status='live', Symbol='LOC105378758', Aliases='', description='uncharacterized LOC105378758', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=59807592, end_position_on_the_genomic_accession=59815740, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  127943 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=127943, CurrentID=0, Status='live', Symbol='FCRLB', Aliases='FCRL2, FCRLM2, FCRLY, FREB-2, FcRY', description='Fc receptor-like B', other_designations='fc receptor homolog expressed in B cells protein 2|fc receptor-like and mucin-like protein 2|fc receptor-like protein 2|fc receptor-related protein Y', map_location='1q23.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161721544, end_position_on_the_genomic_accession=161728143, orientation='plus', exon_count=9, OMIM=609251, no_hdr0=''),
  105378760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378760, CurrentID=0, Status='live', Symbol='LOC105378760', Aliases='', description='uncharacterized LOC105378760', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60242823, end_position_on_the_genomic_accession=60258619, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100529097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529097, CurrentID=0, Status='live', Symbol='RPL36A-HNRNPH2', Aliases='', description='RPL36A-HNRNPH2 readthrough', other_designations='', map_location='Xq22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101390890, end_position_on_the_genomic_accession=101414140, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378762 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378762, CurrentID=0, Status='live', Symbol='LOC105378762', Aliases='', description='uncharacterized LOC105378762', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60522927, end_position_on_the_genomic_accession=60529035, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100529099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529099, CurrentID=0, Status='live', Symbol='DUP2Q31.1', Aliases='C2DUPq31.1', description='Chromosome 2q31.1 duplication syndrome', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613681, no_hdr0=''),
  105378764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378764, CurrentID=0, Status='live', Symbol='LOC105378764', Aliases='', description='uncharacterized LOC105378764', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=60893266, end_position_on_the_genomic_accession=60925000, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390093, CurrentID=0, Status='live', Symbol='OR10A6', Aliases='OR11-96', description='olfactory receptor family 10 subfamily A member 6 (gene/pseudogene)', other_designations='olfactory receptor OR11-96|olfactory receptor, family 10, subfamily A, member 6', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=7927718, end_position_on_the_genomic_accession=7928662, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378766, CurrentID=0, Status='live', Symbol='LOC105378766', Aliases='', description='uncharacterized LOC105378766', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61551257, end_position_on_the_genomic_accession=61567847, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378767, CurrentID=0, Status='live', Symbol='LOC105378767', Aliases='', description='uncharacterized LOC105378767', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62356546, end_position_on_the_genomic_accession=62395506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378768, CurrentID=0, Status='live', Symbol='LOC105378768', Aliases='', description='uncharacterized LOC105378768', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62423976, end_position_on_the_genomic_accession=62432921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378769, CurrentID=0, Status='live', Symbol='LOC105378769', Aliases='', description='uncharacterized LOC105378769', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=62975974, end_position_on_the_genomic_accession=63022126, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105378770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378770, CurrentID=0, Status='live', Symbol='LOC105378770', Aliases='', description='uncharacterized LOC105378770', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63173148, end_position_on_the_genomic_accession=63198709, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378771, CurrentID=0, Status='live', Symbol='LOC105378771', Aliases='', description='uncharacterized LOC105378771', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=63669659, end_position_on_the_genomic_accession=63728051, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378772, CurrentID=0, Status='live', Symbol='LOC105378772', Aliases='', description='uncharacterized LOC105378772', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64033935, end_position_on_the_genomic_accession=64035179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378773, CurrentID=0, Status='live', Symbol='LOC105378773', Aliases='', description='uncharacterized LOC105378773', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=64191521, end_position_on_the_genomic_accession=64196045, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390102, CurrentID=0, Status='live', Symbol='PCNAP4', Aliases='', description='proliferating cell nuclear antigen pseudogene 4', other_designations='', map_location='11p15.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=19274529, end_position_on_the_genomic_accession=19275268, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378775, CurrentID=0, Status='live', Symbol='LOC105378775', Aliases='', description='cell death regulator Aven-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=65048336, end_position_on_the_genomic_accession=65069390, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378776, CurrentID=0, Status='live', Symbol='LOC105378776', Aliases='', description='uncharacterized LOC105378776', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66416559, end_position_on_the_genomic_accession=66439650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378777, CurrentID=0, Status='live', Symbol='LOC105378777', Aliases='', description='uncharacterized LOC105378777', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=66611192, end_position_on_the_genomic_accession=66619849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378778, CurrentID=0, Status='live', Symbol='LOC105378778', Aliases='', description='uncharacterized LOC105378778', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67044310, end_position_on_the_genomic_accession=67049008, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390107, CurrentID=0, Status='live', Symbol='RPS24P15', Aliases='RPS24_5_1135', description='ribosomal protein S24 pseudogene 15', other_designations='', map_location='11p13', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=33236970, end_position_on_the_genomic_accession=33237478, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378780, CurrentID=0, Status='live', Symbol='LOC105378780', Aliases='', description='uncharacterized LOC105378780', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67562057, end_position_on_the_genomic_accession=67590379, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378781, CurrentID=0, Status='live', Symbol='LOC105378781', Aliases='', description='uncharacterized LOC105378781', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=67753931, end_position_on_the_genomic_accession=67783501, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390110, CurrentID=0, Status='live', Symbol='ACCSL', Aliases='', description='1-aminocyclopropane-1-carboxylate synthase (inactive)-like', other_designations='1-aminocyclopropane-1-carboxylate synthase homolog (Arabidopsis)(non-functional)-like|1-aminocyclopropane-1-carboxylate synthase-like protein 2|ACC synthase-like protein 2', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=43921084, end_position_on_the_genomic_accession=44060307, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100529119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529119, CurrentID=0, Status='live', Symbol='SCLL', Aliases='', description='Chromosome 8p11 myeloproliferative syndrome', other_designations='', map_location='8p11', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613523, no_hdr0=''),
  105378784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378784, CurrentID=0, Status='live', Symbol='LOC105378784', Aliases='', description='uncharacterized LOC105378784', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69020353, end_position_on_the_genomic_accession=69054436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390113, CurrentID=0, Status='live', Symbol='OR4X1', Aliases='OR11-104', description='olfactory receptor family 4 subfamily X member 1 (gene/pseudogene)', other_designations='olfactory receptor OR11-104|olfactory receptor, family 4, subfamily X, member 1', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48263861, end_position_on_the_genomic_accession=48264778, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378786 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378786, CurrentID=0, Status='live', Symbol='LOC105378786', Aliases='', description='uncharacterized LOC105378786', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69215515, end_position_on_the_genomic_accession=69249487, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378787, CurrentID=0, Status='live', Symbol='LOC105378787', Aliases='', description='uncharacterized LOC105378787', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69197086, end_position_on_the_genomic_accession=69229641, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  390116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390116, CurrentID=0, Status='live', Symbol='OR4A40P', Aliases='', description='olfactory receptor family 4 subfamily A member 40 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 40 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48512321, end_position_on_the_genomic_accession=48513264, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390117, CurrentID=0, Status='live', Symbol='OR4A43P', Aliases='', description='olfactory receptor family 4 subfamily A member 43 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 43 pseudogene', map_location='11p11.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=48526065, end_position_on_the_genomic_accession=48526975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378790, CurrentID=0, Status='live', Symbol='LOC105378790', Aliases='', description='uncharacterized LOC105378790', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=69557832, end_position_on_the_genomic_accession=69564068, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378791 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378791, CurrentID=0, Status='live', Symbol='LOC105378791', Aliases='', description='uncharacterized LOC105378791', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70038662, end_position_on_the_genomic_accession=70043126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378792, CurrentID=0, Status='live', Symbol='LOC105378792', Aliases='', description='uncharacterized LOC105378792', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70020823, end_position_on_the_genomic_accession=70036600, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378793 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378793, CurrentID=0, Status='live', Symbol='LOC105378793', Aliases='', description='uncharacterized LOC105378793', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70449091, end_position_on_the_genomic_accession=70461596, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378794, CurrentID=0, Status='live', Symbol='LOC105378794', Aliases='', description='uncharacterized LOC105378794', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=70640360, end_position_on_the_genomic_accession=70648909, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378795 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378795, CurrentID=0, Status='live', Symbol='LOC105378795', Aliases='', description='uncharacterized LOC105378795', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=71083811, end_position_on_the_genomic_accession=71088902, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378796, CurrentID=0, Status='live', Symbol='LOC105378796', Aliases='', description='uncharacterized LOC105378796', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72409644, end_position_on_the_genomic_accession=72466359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378797, CurrentID=0, Status='live', Symbol='LOC105378797', Aliases='', description='uncharacterized LOC105378797', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72283096, end_position_on_the_genomic_accession=72670716, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105378798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378798, CurrentID=0, Status='live', Symbol='LOC105378798', Aliases='', description='uncharacterized LOC105378798', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72697398, end_position_on_the_genomic_accession=72766696, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105378799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378799, CurrentID=0, Status='live', Symbol='LOC105378799', Aliases='', description='uncharacterized LOC105378799', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=72848336, end_position_on_the_genomic_accession=72854302, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378800, CurrentID=0, Status='live', Symbol='LOC105378800', Aliases='', description='uncharacterized LOC105378800', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=73129096, end_position_on_the_genomic_accession=73342515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378801, CurrentID=0, Status='live', Symbol='LOC105378801', Aliases='', description='uncharacterized LOC105378801', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=73451088, end_position_on_the_genomic_accession=73472982, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378802, CurrentID=0, Status='live', Symbol='LOC105378802', Aliases='', description='uncharacterized LOC105378802', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=73583282, end_position_on_the_genomic_accession=73607277, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390131, CurrentID=0, Status='live', Symbol='OR4A6P', Aliases='', description='olfactory receptor family 4 subfamily A member 6 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 6 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54724848, end_position_on_the_genomic_accession=54725817, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  652276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652276, CurrentID=0, Status='live', Symbol='LOC652276', Aliases='', description='potassium channel tetramerization domain containing 5 pseudogene', other_designations='Uncharacterized protein PRO0461|potassium channel tetramerisation domain containing 5 pseudogene', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2603384, end_position_on_the_genomic_accession=2630494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100529141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529141, CurrentID=0, Status='live', Symbol='LGTN', Aliases='', description='ligatin', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=151625, no_hdr0=''),
  390134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390134, CurrentID=0, Status='live', Symbol='OR4A4P', Aliases='OR4A4', description='olfactory receptor family 4 subfamily A member 4 pseudogene', other_designations='olfactory receptor, family 4, subfamily A, member 4 pseudogene', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=54659667, end_position_on_the_genomic_accession=54660596, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378807, CurrentID=0, Status='live', Symbol='LOC105378807', Aliases='', description='uncharacterized LOC105378807', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=76587063, end_position_on_the_genomic_accession=76675311, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100529144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529144, CurrentID=0, Status='live', Symbol='CORO7-PAM16', Aliases='', description='CORO7-PAM16 readthrough', other_designations='', map_location='16p', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=4340251, end_position_on_the_genomic_accession=4416961, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  100529145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529145, CurrentID=0, Status='live', Symbol='TEN1-CDK3', Aliases='C17orf106-CDK3', description='TEN1-CDK3 readthrough (NMD candidate)', other_designations='TEN1-CDK3 readthrough (non-protein coding)', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=75979217, end_position_on_the_genomic_accession=76005999, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100529146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529146, CurrentID=0, Status='live', Symbol='DER22T11-22', Aliases='', description='Emanuel syndrome', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609029, no_hdr0=''),
  139562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139562, CurrentID=0, Status='live', Symbol='OTUD6A', Aliases='DUBA2, HSHIN6', description='OTU deubiquitinase 6A', other_designations='HIN-6 protease', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70062491, end_position_on_the_genomic_accession=70064179, orientation='plus', exon_count=1, OMIM=300714, no_hdr0=''),
  100529148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529148, CurrentID=0, Status='live', Symbol='MDNS', Aliases='', description='Mammary-digital-nail syndrome', other_designations='', map_location='22q12.3-q13.1', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613689, no_hdr0=''),
  100529149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529149, CurrentID=0, Status='live', Symbol='DER22T8-22', Aliases='', description='Supernumerary der(22)t(8-22) syndrome', other_designations='', map_location='22q11.2', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613700, no_hdr0=''),
  390142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390142, CurrentID=0, Status='live', Symbol='OR5D13', Aliases='', description='olfactory receptor family 5 subfamily D member 13 (gene/pseudogene)', other_designations='olfactory receptor OR11-142|olfactory receptor OR11-148|olfactory receptor, family 5, subfamily D, member 13', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55773438, end_position_on_the_genomic_accession=55774382, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100529151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529151, CurrentID=0, Status='live', Symbol='IBD11', Aliases='', description='Inflammatory bowel disease 11', other_designations='', map_location='7q22', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=191390, no_hdr0=''),
  390144 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390144, CurrentID=0, Status='live', Symbol='OR5D16', Aliases='OR11-154', description='olfactory receptor family 5 subfamily D member 16', other_designations='olfactory receptor OR11-154|olfactory receptor, family 5, subfamily D, member 16', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55838752, end_position_on_the_genomic_accession=55839738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378817, CurrentID=0, Status='live', Symbol='LOC105378817', Aliases='', description='uncharacterized LOC105378817', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=83065928, end_position_on_the_genomic_accession=83086308, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378818, CurrentID=0, Status='live', Symbol='LOC105378818', Aliases='', description='uncharacterized LOC105378818', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84076331, end_position_on_the_genomic_accession=84078014, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378819, CurrentID=0, Status='live', Symbol='LOC105378819', Aliases='', description='uncharacterized LOC105378819', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84578091, end_position_on_the_genomic_accession=84583620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390148, CurrentID=0, Status='live', Symbol='OR5W2', Aliases='OR5W2P, OR5W3P', description='olfactory receptor family 5 subfamily W member 2', other_designations='olfactory receptor 5W3|olfactory receptor OR11-155|olfactory receptor, family 5, subfamily W, member 2|olfactory receptor, family 5, subfamily W, member 3 pseudogene', map_location='11q11', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=55913650, end_position_on_the_genomic_accession=55914582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378821, CurrentID=0, Status='live', Symbol='LOC105378821', Aliases='', description='uncharacterized LOC105378821', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85225579, end_position_on_the_genomic_accession=85245412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378822, CurrentID=0, Status='live', Symbol='LOC105378822', Aliases='', description='uncharacterized LOC105378822', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=85624203, end_position_on_the_genomic_accession=85628292, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390151, CurrentID=0, Status='live', Symbol='OR8H2', Aliases='OR11-171', description='olfactory receptor family 8 subfamily H member 2', other_designations='olfactory receptor OR11-171|olfactory receptor, family 8, subfamily H, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56105043, end_position_on_the_genomic_accession=56105981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390152, CurrentID=0, Status='live', Symbol='OR8H3', Aliases='OR11-172', description='olfactory receptor family 8 subfamily H member 3', other_designations='olfactory receptor OR11-172|olfactory receptor, family 8, subfamily H, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56122373, end_position_on_the_genomic_accession=56123311, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378825, CurrentID=0, Status='live', Symbol='LOC105378825', Aliases='', description='uncharacterized LOC105378825', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86328446, end_position_on_the_genomic_accession=86330110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390154 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390154, CurrentID=0, Status='live', Symbol='OR5T3', Aliases='OR11-178Q, OR5T3', description='olfactory receptor family 5 subfamily T member 3', other_designations='olfactory receptor OR11-178|olfactory receptor, family 5, subfamily T, member 3', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56252200, end_position_on_the_genomic_accession=56253222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390155, CurrentID=0, Status='live', Symbol='OR5T1', Aliases='OR11-179P, OR5T1', description='olfactory receptor family 5 subfamily T member 1', other_designations='olfactory receptor OR11-179|olfactory receptor, family 5, subfamily T, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56275639, end_position_on_the_genomic_accession=56276619, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378828, CurrentID=0, Status='live', Symbol='LOC105378828', Aliases='', description='uncharacterized LOC105378828', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86703498, end_position_on_the_genomic_accession=86704754, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390157, CurrentID=0, Status='live', Symbol='OR8K1', Aliases='OR11-182, OR8N1P', description='olfactory receptor family 8 subfamily K member 1', other_designations='olfactory receptor OR11-182|olfactory receptor, family 8, subfamily K, member 1|olfactory receptor, family 8, subfamily N, member 1 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56346039, end_position_on_the_genomic_accession=56346998, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390158, CurrentID=0, Status='live', Symbol='RPL5P29', Aliases='RPL5_15_1144', description='ribosomal protein L5 pseudogene 29', other_designations='', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56357273, end_position_on_the_genomic_accession=56358286, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378831, CurrentID=0, Status='live', Symbol='LOC105378831', Aliases='', description='uncharacterized LOC105378831', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=86821597, end_position_on_the_genomic_accession=86825909, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378832 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378832, CurrentID=0, Status='live', Symbol='LOC105378832', Aliases='', description='uncharacterized LOC105378832', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87219918, end_position_on_the_genomic_accession=87261689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378833, CurrentID=0, Status='live', Symbol='LOC105378833', Aliases='', description='uncharacterized LOC105378833', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87394492, end_position_on_the_genomic_accession=87430418, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390162, CurrentID=0, Status='live', Symbol='OR5M9', Aliases='OR11-190', description='olfactory receptor family 5 subfamily M member 9', other_designations='olfactory receptor OR11-190|olfactory receptor, family 5, subfamily M, member 9', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56462469, end_position_on_the_genomic_accession=56463401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378835, CurrentID=0, Status='live', Symbol='LOC105378835', Aliases='', description='uncharacterized LOC105378835', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87526334, end_position_on_the_genomic_accession=87528656, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378836, CurrentID=0, Status='live', Symbol='LOC105378836', Aliases='', description='uncharacterized LOC105378836', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=87965171, end_position_on_the_genomic_accession=87987040, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378837 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378837, CurrentID=0, Status='live', Symbol='LOC105378837', Aliases='', description='uncharacterized LOC105378837', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=88175239, end_position_on_the_genomic_accession=88271652, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390166, CurrentID=0, Status='live', Symbol='OR5M5P', Aliases='', description='olfactory receptor family 5 subfamily M member 5 pseudogene', other_designations='olfactory receptor OR11-187 pseudogene|olfactory receptor OR11-192 pseudogene|olfactory receptor OR11-198 pseudogene|olfactory receptor, family 5, subfamily M, member 5 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56526567, end_position_on_the_genomic_accession=56527502, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390167, CurrentID=0, Status='live', Symbol='OR5M10', Aliases='OR11-207', description='olfactory receptor family 5 subfamily M member 10', other_designations='olfactory receptor OR11-207|olfactory receptor, family 5, subfamily M, member 10', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56576774, end_position_on_the_genomic_accession=56577721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390168, CurrentID=0, Status='live', Symbol='OR5M1', Aliases='OR11-208, OST050', description='olfactory receptor family 5 subfamily M member 1', other_designations='olfactory receptor OR11-208|olfactory receptor, family 5, subfamily M, member 1|seven transmembrane helix receptor', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56612555, end_position_on_the_genomic_accession=56613502, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  128025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128025, CurrentID=0, Status='live', Symbol='WDR64', Aliases='', description='WD repeat domain 64', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241652278, end_position_on_the_genomic_accession=241802745, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  105378842 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378842, CurrentID=0, Status='live', Symbol='LOC105378842', Aliases='', description='uncharacterized LOC105378842', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89128417, end_position_on_the_genomic_accession=89179775, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378843, CurrentID=0, Status='live', Symbol='LOC105378843', Aliases='', description='uncharacterized LOC105378843', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89217044, end_position_on_the_genomic_accession=89218034, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378844, CurrentID=0, Status='live', Symbol='LOC105378844', Aliases='', description='uncharacterized LOC105378844', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89471601, end_position_on_the_genomic_accession=89479972, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378845, CurrentID=0, Status='live', Symbol='LOC105378845', Aliases='', description='uncharacterized LOC105378845', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89544057, end_position_on_the_genomic_accession=89555164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390174, CurrentID=0, Status='live', Symbol='OR9G1', Aliases='OR9G5', description='olfactory receptor family 9 subfamily G member 1', other_designations='olfactory receptor 9G5|olfactory receptor OR11-114|olfactory receptor OR11-210|olfactory receptor, family 9, subfamily G, member 1|olfactory receptor, family 9, subfamily G, member 5', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56700388, end_position_on_the_genomic_accession=56701305, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378847 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378847, CurrentID=0, Status='live', Symbol='LOC105378847', Aliases='', description='uncharacterized LOC105378847', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89967191, end_position_on_the_genomic_accession=89986800, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378848, CurrentID=0, Status='live', Symbol='LOC105378848', Aliases='', description='uncharacterized LOC105378848', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90115181, end_position_on_the_genomic_accession=90118603, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378849, CurrentID=0, Status='live', Symbol='LOC105378849', Aliases='', description='uncharacterized LOC105378849', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90219803, end_position_on_the_genomic_accession=90285806, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105378850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378850, CurrentID=0, Status='live', Symbol='LOC105378850', Aliases='', description='uncharacterized LOC105378850', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90390187, end_position_on_the_genomic_accession=90401055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378851 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378851, CurrentID=0, Status='live', Symbol='LOC105378851', Aliases='', description='uncharacterized LOC105378851', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90475857, end_position_on_the_genomic_accession=90488088, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378852, CurrentID=0, Status='live', Symbol='LOC105378852', Aliases='', description='uncharacterized LOC105378852', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90504685, end_position_on_the_genomic_accession=90547930, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  390181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390181, CurrentID=0, Status='live', Symbol='OR5AK2', Aliases='', description='olfactory receptor family 5 subfamily AK member 2', other_designations='olfactory receptor, family 5, subfamily AK, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=56988914, end_position_on_the_genomic_accession=56989843, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378854, CurrentID=0, Status='live', Symbol='LOC105378854', Aliases='', description='uncharacterized LOC105378854', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=90851126, end_position_on_the_genomic_accession=90854899, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390183, CurrentID=0, Status='live', Symbol='RPS4XP13', Aliases='RPS4P13, RPS4X_3_1145', description='ribosomal protein S4X pseudogene 13', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=57576229, end_position_on_the_genomic_accession=57577105, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378856, CurrentID=0, Status='live', Symbol='LOC105378856', Aliases='', description='uncharacterized LOC105378856', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=91488866, end_position_on_the_genomic_accession=91499054, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378857, CurrentID=0, Status='live', Symbol='LOC105378857', Aliases='', description='uncharacterized LOC105378857', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=92214631, end_position_on_the_genomic_accession=92229004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390186 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390186, CurrentID=0, Status='live', Symbol='OR5B10P', Aliases='OR13-34, OR13-64, OR13-67, OR5B10, OR5B11, OR5B11P, OR5B18P, OR5B4P', description='olfactory receptor family 5 subfamily B member 10 pseudogene', other_designations='olfactory receptor, family 5, subfamily B, member 10 pseudogene|olfactory receptor, family 5, subfamily B, member 11 pseudogene|olfactory receptor, family 5, subfamily B, member 18 pseudogene|olfactory receptor, family 5, subfamily B, member 4 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58348985, end_position_on_the_genomic_accession=58349638, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378859 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378859, CurrentID=0, Status='live', Symbol='LOC105378859', Aliases='', description='uncharacterized LOC105378859', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94311301, end_position_on_the_genomic_accession=94330050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378860, CurrentID=0, Status='live', Symbol='LOC105378860', Aliases='', description='uncharacterized LOC105378860', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94683472, end_position_on_the_genomic_accession=94688112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378861, CurrentID=0, Status='live', Symbol='LOC105378861', Aliases='', description='uncharacterized LOC105378861', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=94599330, end_position_on_the_genomic_accession=94636984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390190, CurrentID=0, Status='live', Symbol='OR5B2', Aliases='OR11-240, OST073', description='olfactory receptor family 5 subfamily B member 2', other_designations='olfactory receptor OR11-240|olfactory receptor, family 5, subfamily B, member 2', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58422265, end_position_on_the_genomic_accession=58423313, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390191, CurrentID=0, Status='live', Symbol='OR5B12', Aliases='OR11-241P, OR5B16, OST743, OR5B12', description='olfactory receptor family 5 subfamily B member 12', other_designations='olfactory receptor 5B16|olfactory receptor OR11-241|olfactory receptor, family 5, subfamily B, member 12 pseudogene|olfactory receptor, family 5, subfamily B, member 16', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=58439120, end_position_on_the_genomic_accession=58440173, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378864, CurrentID=0, Status='live', Symbol='LOC105378864', Aliases='', description='uncharacterized LOC105378864', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=95315795, end_position_on_the_genomic_accession=95317984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378865, CurrentID=0, Status='live', Symbol='LOC105378865', Aliases='', description='uncharacterized LOC105378865', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96268861, end_position_on_the_genomic_accession=96359189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378866, CurrentID=0, Status='live', Symbol='LOC105378866', Aliases='', description='uncharacterized LOC105378866', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96482428, end_position_on_the_genomic_accession=96524341, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390195, CurrentID=0, Status='live', Symbol='OR5AN1', Aliases='OR11-244', description='olfactory receptor family 5 subfamily AN member 1', other_designations='muscone receptor|olfactory receptor OR11-244|olfactory receptor, family 5, subfamily AN, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59364459, end_position_on_the_genomic_accession=59365394, orientation='plus', exon_count=1, OMIM=615702, no_hdr0=''),
  105378868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378868, CurrentID=0, Status='live', Symbol='LOC105378868', Aliases='', description='uncharacterized LOC105378868', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=97108138, end_position_on_the_genomic_accession=97116436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390197, CurrentID=0, Status='live', Symbol='OR4D10', Aliases='OR11-251P, OST711, OR4D10', description='olfactory receptor family 4 subfamily D member 10', other_designations='olfactory receptor OR11-251|olfactory receptor, family 4, subfamily D, member 10 pseudogene|seven transmembrane helix receptor', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59477430, end_position_on_the_genomic_accession=59478365, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378870, CurrentID=0, Status='live', Symbol='LOC105378870', Aliases='', description='uncharacterized LOC105378870', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=98188471, end_position_on_the_genomic_accession=98209472, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390199, CurrentID=0, Status='live', Symbol='OR4D9', Aliases='OR11-253', description='olfactory receptor family 4 subfamily D member 9', other_designations='olfactory receptor OR11-253|olfactory receptor, family 4, subfamily D, member 9 pseudogene', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59514913, end_position_on_the_genomic_accession=59515857, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378891, CurrentID=0, Status='live', Symbol='LOC105378891', Aliases='', description='uncharacterized LOC105378891', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108833763, end_position_on_the_genomic_accession=108837003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390201, CurrentID=0, Status='live', Symbol='OR10V1', Aliases='OR11-256', description='olfactory receptor family 10 subfamily V member 1', other_designations='olfactory receptor OR11-256|olfactory receptor, family 10, subfamily V, member 1', map_location='11q12.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=59712916, end_position_on_the_genomic_accession=59713845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378874, CurrentID=0, Status='live', Symbol='LOC105378874', Aliases='', description='uncharacterized LOC105378874', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=102647854, end_position_on_the_genomic_accession=102660714, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100529211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529211, CurrentID=0, Status='live', Symbol='TMEM256-PLSCR3', Aliases='C17orf61-PLSCR3', description='TMEM256-PLSCR3 readthrough (NMD candidate)', other_designations='', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7389727, end_position_on_the_genomic_accession=7404131, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105378876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378876, CurrentID=0, Status='live', Symbol='LOC105378876', Aliases='', description='uncharacterized LOC105378876', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=103650532, end_position_on_the_genomic_accession=103651378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390205, CurrentID=0, Status='live', Symbol='LRRC10B', Aliases='', description='leucine rich repeat containing 10B', other_designations='LRRC10-like protein ENSP00000367315', map_location='11q12.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=61508800, end_position_on_the_genomic_accession=61511018, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390206, CurrentID=0, Status='live', Symbol='RPS2P37', Aliases='RPS2_14_1147', description='ribosomal protein S2 pseudogene 37', other_designations='', map_location='11q12.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=62011721, end_position_on_the_genomic_accession=62012648, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100529215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529215, CurrentID=0, Status='live', Symbol='ZNF559-ZNF177', Aliases='', description='ZNF559-ZNF177 readthrough', other_designations='', map_location='19p', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9324226, end_position_on_the_genomic_accession=9382617, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105378880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378880, CurrentID=0, Status='live', Symbol='LOC105378880', Aliases='', description='uncharacterized LOC105378880', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105001026, end_position_on_the_genomic_accession=105025102, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378881, CurrentID=0, Status='live', Symbol='LOC105378881', Aliases='', description='uncharacterized LOC105378881', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105473602, end_position_on_the_genomic_accession=105489957, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378882, CurrentID=0, Status='live', Symbol='LOC105378882', Aliases='', description='uncharacterized LOC105378882', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=105668856, end_position_on_the_genomic_accession=105673290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390211, CurrentID=0, Status='live', Symbol='KRT8P26', Aliases='', description='keratin 8 pseudogene 26', other_designations='', map_location='11q13.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=65726867, end_position_on_the_genomic_accession=65728015, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390212, CurrentID=0, Status='live', Symbol='GPR152', Aliases='PGR5', description='G protein-coupled receptor 152', other_designations='G protein-coupled receptor PGR5', map_location='11q13.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67451301, end_position_on_the_genomic_accession=67452729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100529221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529221, CurrentID=0, Status='live', Symbol='DEL6Q11Q14', Aliases='C6DELq11q14', description='Chromosome 6q11-q14 deletion syndrome', other_designations='', map_location='6q11-q14', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613544, no_hdr0=''),
  100529222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529222, CurrentID=0, Status='live', Symbol='STQTL23', Aliases='', description='Stature QTL 23', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613548, no_hdr0=''),
  100529223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529223, CurrentID=0, Status='live', Symbol='STQTL24', Aliases='', description='Stature QTL 24', other_designations='', map_location='2p16', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613549, no_hdr0=''),
  100529224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529224, CurrentID=0, Status='live', Symbol='FWS', Aliases='', description='Forsythe-Wakeling syndrome', other_designations='', map_location='1p33-p31.1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613606, no_hdr0=''),
  105378889 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378889, CurrentID=0, Status='live', Symbol='LOC105378889', Aliases='', description='uncharacterized LOC105378889', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=106818188, end_position_on_the_genomic_accession=106878302, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100529226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529226, CurrentID=0, Status='live', Symbol='DUP17Q21.31', Aliases='C17DUPq21.31', description='Chromosome 17q21.31 duplication syndrome', other_designations='', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613533, no_hdr0=''),
  100529227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529227, CurrentID=0, Status='live', Symbol='STQTL22', Aliases='', description='Stature QTL 22', other_designations='', map_location='16q24', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613547, no_hdr0=''),
  100529228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529228, CurrentID=0, Status='live', Symbol='TRIP4Q32.1Q32.2', Aliases='C4TRIPq32.1q32.2', description='Chromosome 4q32.1-q32.2 triplication syndrome', other_designations='', map_location='4q32.1-q32.2', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613603, no_hdr0=''),
  128077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128077, CurrentID=0, Status='live', Symbol='LIX1L', Aliases='', description='limb and CNS expressed 1 like', other_designations='Lix1 homolog (chicken) like|Lix1 homolog-like', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145933424, end_position_on_the_genomic_accession=145958019, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100529230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529230, CurrentID=0, Status='live', Symbol='LGMD1H', Aliases='', description='limb girdle muscular dystrophy 1H (autosomal dominant)', other_designations='', map_location='3p25.1-p23', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613530, no_hdr0=''),
  100529231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529231, CurrentID=0, Status='live', Symbol='MACST', Aliases='', description='Macrostomia', other_designations='', map_location='1p34-p32', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613545, no_hdr0=''),
  390224 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390224, CurrentID=0, Status='live', Symbol='OR2AT2P', Aliases='', description='olfactory receptor family 2 subfamily AT member 2 pseudogene', other_designations='olfactory receptor, family 2, subfamily AT, member 2 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75071120, end_position_on_the_genomic_accession=75072091, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390225, CurrentID=0, Status='live', Symbol='OR2AT1P', Aliases='', description='olfactory receptor family 2 subfamily AT member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily AT, member 1 pseudogene', map_location='11q13.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=75131103, end_position_on_the_genomic_accession=75131872, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390226, CurrentID=0, Status='live', Symbol='GUCY2EP', Aliases='GC-E, GCD, GUCY2E', description='guanylate cyclase 2E, pseudogene', other_designations='olfactory guanylate cyclase', map_location='11q13.5', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=76680166, end_position_on_the_genomic_accession=76721789, orientation='minus', exon_count=21, OMIM=601138, no_hdr0=''),
  105378899 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378899, CurrentID=0, Status='live', Symbol='LOC105378899', Aliases='', description='uncharacterized LOC105378899', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110472525, end_position_on_the_genomic_accession=110473646, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390228, CurrentID=0, Status='live', Symbol='RPL7AP54', Aliases='RPL7A_22_1161', description='ribosomal protein L7a pseudogene 54', other_designations='', map_location='11q14.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=82957393, end_position_on_the_genomic_accession=82958273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378901 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378901, CurrentID=0, Status='live', Symbol='LOC105378901', Aliases='', description='uncharacterized LOC105378901', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110749204, end_position_on_the_genomic_accession=110753073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378902 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378902, CurrentID=0, Status='live', Symbol='LOC105378902', Aliases='', description='uncharacterized LOC105378902', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=110877153, end_position_on_the_genomic_accession=110929157, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390231, CurrentID=0, Status='live', Symbol='TRIM77', Aliases='TRIM77P', description='tripartite motif containing 77', other_designations='TRIM77-isoform|putative tripartite motif-containing protein 64D|putative tripartite motif-containing protein 77|tripartite motif containing 77, pseudogene', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=89710299, end_position_on_the_genomic_accession=89717872, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100885787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100885787, CurrentID=0, Status='live', Symbol='DEL8Q12Q21', Aliases='C8DELq12q21', description='Bor-Duane hydrocephalus contiguous gene syndrome', other_designations='', map_location='8q12.2-q21.2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=600257, no_hdr0=''),
  100529241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529241, CurrentID=0, Status='live', Symbol='HSPE1-MOB4', Aliases='HSPE1-PHOCN', description='HSPE1-MOB4 readthrough', other_designations='', map_location='2q33', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197499997, end_position_on_the_genomic_accession=197553699, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100529242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529242, CurrentID=0, Status='live', Symbol='DEL1Q41Q42', Aliases='C1DELq41q42', description='Chromosome 1q41-q42 deletion syndrome', other_designations='', map_location='1q41-q42', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=612530, no_hdr0=''),
  100529243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529243, CurrentID=0, Status='live', Symbol='DEL2P12P11.2', Aliases='C2DELp12p11.2', description='Chromosome 2p12-p11.2 deletion syndrome', other_designations='', map_location='2p12-p11.2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613564, no_hdr0=''),
  105378908 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378908, CurrentID=0, Status='live', Symbol='LOC105378908', Aliases='', description='uncharacterized LOC105378908', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112269862, end_position_on_the_genomic_accession=112273735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378909 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378909, CurrentID=0, Status='live', Symbol='LOC105378909', Aliases='', description='uncharacterized LOC105378909', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112229673, end_position_on_the_genomic_accession=112360633, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378910, CurrentID=0, Status='live', Symbol='LOC105378910', Aliases='', description='uncharacterized LOC105378910', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112715357, end_position_on_the_genomic_accession=112717794, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390239, CurrentID=0, Status='live', Symbol='NDUFB11P1', Aliases='NDUFB11P', description='NADH:ubiquinone oxidoreductase subunit B11 pseudogene 1', other_designations='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 11, 17.3kDa pseudogene|NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 11, pseudogene 1', map_location='11q14.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=92335938, end_position_on_the_genomic_accession=92336513, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378912 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378912, CurrentID=0, Status='live', Symbol='HP08777', Aliases='', description='uncharacterized LOC105378912', other_designations='hypothetical protein HP08777', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112850344, end_position_on_the_genomic_accession=112877871, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378913 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378913, CurrentID=0, Status='live', Symbol='LOC105378913', Aliases='', description='uncharacterized LOC105378913', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114545318, end_position_on_the_genomic_accession=114548202, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378914, CurrentID=0, Status='live', Symbol='LOC105378914', Aliases='', description='uncharacterized LOC105378914', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=114581739, end_position_on_the_genomic_accession=114586880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390243, CurrentID=0, Status='live', Symbol='IZUMO1R', Aliases='FOLR4, Folbp3, JUNO', description='IZUMO1 receptor, JUNO', other_designations='FR-delta|IZUMO1 receptor protein JUNO|folate receptor 4 (delta) homolog|folate receptor 4, delta (putative)|folate receptor delta|probable folate receptor delta', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=94304561, end_position_on_the_genomic_accession=94307762, orientation='plus', exon_count=5, OMIM=615737, no_hdr0=''),
  105378916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378916, CurrentID=0, Status='live', Symbol='LOC105378916', Aliases='', description='uncharacterized LOC105378916', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=115099560, end_position_on_the_genomic_accession=115102681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390245, CurrentID=0, Status='live', Symbol='KDM4E', Aliases='JMJD2E, KDM4DL', description='lysine (K)-specific demethylase 4E', other_designations='KDM4D-like protein', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=95025258, end_position_on_the_genomic_accession=95027596, orientation='plus', exon_count=1, OMIM=616581, no_hdr0=''),
  128102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128102, CurrentID=0, Status='live', Symbol='HSD3BP4', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta, pseudogene 4', other_designations='3 beta-hydroxysteroid dehydrogenase homolog|3-beta-hydroxysteroid dehydrogenase, tissue-type heart|hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 1 pseudogene', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119563880, end_position_on_the_genomic_accession=119572576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105378919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378919, CurrentID=0, Status='live', Symbol='LOC105378919', Aliases='', description='uncharacterized LOC105378919', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116240964, end_position_on_the_genomic_accession=116242077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378920 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378920, CurrentID=0, Status='live', Symbol='LOC105378920', Aliases='', description='uncharacterized LOC105378920', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116289248, end_position_on_the_genomic_accession=116336339, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100529257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529257, CurrentID=0, Status='live', Symbol='SYNJ2BP-COX16', Aliases='', description='SYNJ2BP-COX16 readthrough', other_designations='', map_location='14q24', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=70325081, end_position_on_the_genomic_accession=70417090, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105378922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378922, CurrentID=0, Status='live', Symbol='LOC105378922', Aliases='', description='uncharacterized LOC105378922', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116493016, end_position_on_the_genomic_accession=116505088, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390251, CurrentID=0, Status='live', Symbol='LOC390251', Aliases='', description='SH3-domain GRB2-like 1 pseudogene', other_designations='', map_location='11q23.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=113789087, end_position_on_the_genomic_accession=113790097, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378924, CurrentID=0, Status='live', Symbol='LOC105378924', Aliases='', description='uncharacterized LOC105378924', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116591470, end_position_on_the_genomic_accession=116595796, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100529261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529261, CurrentID=0, Status='live', Symbol='CHURC1-FNTB', Aliases='', description='CHURC1-FNTB readthrough', other_designations='', map_location='14q23', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=64914361, end_position_on_the_genomic_accession=65062655, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100529262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529262, CurrentID=0, Status='live', Symbol='MIA-RAB4B', Aliases='', description='MIA-RAB4B readthrough (NMD candidate)', other_designations='', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40775537, end_position_on_the_genomic_accession=40796944, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105378927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378927, CurrentID=0, Status='live', Symbol='LOC105378927', Aliases='', description='uncharacterized LOC105378927', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116882384, end_position_on_the_genomic_accession=116885828, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100529264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100529264, CurrentID=0, Status='live', Symbol='RAB4B-EGLN2', Aliases='RERT-lncRNA', description='RAB4B-EGLN2 readthrough (NMD candidate)', other_designations='RAB4B-EGLN2 readthrough (non-protein coding)|RAB4B-EGLN2 readthrough long non-coding RNA', map_location='19q', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=40778219, end_position_on_the_genomic_accession=40808441, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105378929 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378929, CurrentID=0, Status='live', Symbol='LOC105378929', Aliases='', description='uncharacterized LOC105378929', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=117815461, end_position_on_the_genomic_accession=117821182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378930 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378930, CurrentID=0, Status='live', Symbol='LOC105378930', Aliases='', description='uncharacterized LOC105378930', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=118185256, end_position_on_the_genomic_accession=118206679, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390259, CurrentID=0, Status='live', Symbol='BSX', Aliases='BSX1', description='brain specific homeobox', other_designations='', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=122977649, end_position_on_the_genomic_accession=122981671, orientation='minus', exon_count=3, OMIM=611074, no_hdr0=''),
  390260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390260, CurrentID=0, Status='live', Symbol='OR6X1', Aliases='OR11-270', description='olfactory receptor family 6 subfamily X member 1', other_designations='olfactory receptor OR11-270|olfactory receptor, family 6, subfamily X, member 1', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123753580, end_position_on_the_genomic_accession=123754518, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390261, CurrentID=0, Status='live', Symbol='OR6M1', Aliases='OR11-271', description='olfactory receptor family 6 subfamily M member 1', other_designations='olfactory receptor OR11-271|olfactory receptor, family 6, subfamily M, member 1|olfactory receptor, family 6, subfamily M, member 2 pseudogene', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=123805408, end_position_on_the_genomic_accession=123806349, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105378934 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378934, CurrentID=0, Status='live', Symbol='LOC105378934', Aliases='', description='uncharacterized LOC105378934', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119327415, end_position_on_the_genomic_accession=119362420, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100422505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422505, CurrentID=0, Status='live', Symbol='ATP8A2P1', Aliases='', description='ATPase, aminophospholipid transporter, class I, type 8A, member 2 pseudogene 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=37248118, end_position_on_the_genomic_accession=37264637, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390264, CurrentID=0, Status='live', Symbol='OR10G4', Aliases='OR11-278', description='olfactory receptor family 10 subfamily G member 4', other_designations='olfactory receptor OR11-278|olfactory receptor, family 10, subfamily G, member 4', map_location='11q24.1', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124015575, end_position_on_the_genomic_accession=124016510, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390265, CurrentID=0, Status='live', Symbol='OR10G7', Aliases='OR11-283', description='olfactory receptor family 10 subfamily G member 7', other_designations='olfactory receptor OR11-283|olfactory receptor, family 10, subfamily G, member 7', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124038066, end_position_on_the_genomic_accession=124039001, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390266, CurrentID=0, Status='live', Symbol='OR10D4P', Aliases='OR10D4, OR10D6P', description='olfactory receptor family 10 subfamily D member 4 pseudogene', other_designations='Olfactory receptor 10D4|olfactory receptor, family 10, subfamily D, member 4 pseudogene|olfactory receptor, family 10, subfamily D, member 6 pseudogene|seven transmembrane helix receptor', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124093436, end_position_on_the_genomic_accession=124094732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652411, CurrentID=0, Status='live', Symbol='RPSAP44', Aliases='RPSA_16_706', description='ribosomal protein SA pseudogene 44', other_designations='', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=54624527, end_position_on_the_genomic_accession=54625523, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378940, CurrentID=0, Status='live', Symbol='LOC105378940', Aliases='', description='uncharacterized LOC105378940', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120120477, end_position_on_the_genomic_accession=120136719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259197, CurrentID=0, Status='live', Symbol='NCR3', Aliases='1C7, CD337, LY117, MALS, NKp30', description='natural cytotoxicity triggering receptor 3', other_designations='NK-p30|activating NK-A1 receptor|activating natural killer receptor p30|lymphocyte antigen 117|natural killer cell p30-related protein', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31588883, end_position_on_the_genomic_accession=31593019, orientation='minus', exon_count=5, OMIM=611550, no_hdr0=''),
  105378942 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378942, CurrentID=0, Status='live', Symbol='LOC105378942', Aliases='', description='uncharacterized LOC105378942', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120375275, end_position_on_the_genomic_accession=120377885, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390271, CurrentID=0, Status='live', Symbol='OR8B3', Aliases='OR11-311', description='olfactory receptor family 8 subfamily B member 3', other_designations='olfactory receptor OR11-311|olfactory receptor, family 8, subfamily B, member 3', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124396410, end_position_on_the_genomic_accession=124399522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378944 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378944, CurrentID=0, Status='live', Symbol='LOC105378944', Aliases='', description='uncharacterized LOC105378944', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120716939, end_position_on_the_genomic_accession=120723764, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390273 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390273, CurrentID=0, Status='live', Symbol='OR8A2P', Aliases='', description='olfactory receptor family 8 subfamily A member 2 pseudogene', other_designations='olfactory receptor, family 8, subfamily A, member 2 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124460565, end_position_on_the_genomic_accession=124461527, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390274, CurrentID=0, Status='live', Symbol='OR8B10P', Aliases='', description='olfactory receptor family 8 subfamily B member 10 pseudogene', other_designations='olfactory receptor, family 8, subfamily B, member 10 pseudogene', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124516310, end_position_on_the_genomic_accession=124517242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390275, CurrentID=0, Status='live', Symbol='OR8A1', Aliases='OR11-318, OST025', description='olfactory receptor family 8 subfamily A member 1', other_designations='olfactory receptor OR11-318|olfactory receptor, family 8, subfamily A, member 1', map_location='11q24.2', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=124570003, end_position_on_the_genomic_accession=124602563, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378948, CurrentID=0, Status='live', Symbol='LOC105378948', Aliases='', description='skin secretory protein xP2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1065635, end_position_on_the_genomic_accession=1069326, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105378949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378949, CurrentID=0, Status='live', Symbol='LOC105378949', Aliases='', description='basic salivary proline-rich protein 2-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=1885960, end_position_on_the_genomic_accession=1893010, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378950 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378950, CurrentID=0, Status='live', Symbol='LOC105378950', Aliases='', description='mitogen-activated protein kinase 7-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=39488834, end_position_on_the_genomic_accession=39491993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  652423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652423, CurrentID=0, Status='live', Symbol='NPM1P36', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 36', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='6p12.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=55939753, end_position_on_the_genomic_accession=55940720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378952 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378952, CurrentID=0, Status='live', Symbol='KLF18', Aliases='', description='Kruppel-like factor 18', other_designations='KLF pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44137796, end_position_on_the_genomic_accession=44141631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378953 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378953, CurrentID=0, Status='live', Symbol='LOC105378953', Aliases='', description='UPF0587 protein C1orf123 homolog pseudogene', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=99464390, end_position_on_the_genomic_accession=99464980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378954, CurrentID=0, Status='live', Symbol='LOC105378954', Aliases='', description='uncharacterized LOC105378954', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120339116, end_position_on_the_genomic_accession=120341679, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390283, CurrentID=0, Status='live', Symbol='RPL31P10', Aliases='RPL31_15_1198', description='ribosomal protein L31 pseudogene 10', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6409596, end_position_on_the_genomic_accession=6410040, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390284 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390284, CurrentID=0, Status='live', Symbol='SRP14P1', Aliases='', description='signal recognition particle 14kDa (homologous Alu RNA binding protein) pseudogene 1', other_designations='', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6438475, end_position_on_the_genomic_accession=6439137, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378957, CurrentID=0, Status='live', Symbol='LOC105378957', Aliases='', description='basic salivary proline-rich protein 2-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91498319, end_position_on_the_genomic_accession=91506909, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105378958 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378958, CurrentID=0, Status='live', Symbol='LOC105378958', Aliases='', description='adhesive plaque matrix protein-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91507532, end_position_on_the_genomic_accession=91513275, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  259215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259215, CurrentID=0, Status='live', Symbol='LY6G6F', Aliases='C6orf21, G6f, LY6G6D, NG32', description='lymphocyte antigen 6 complex, locus G6F', other_designations='lymphocyte antigen 6 complex, locus G6D', map_location='6p21.33', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=31706907, end_position_on_the_genomic_accession=31710595, orientation='plus', exon_count=6, OMIM=611404, no_hdr0=''),
  390288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390288, CurrentID=0, Status='live', Symbol='OR7E84P', Aliases='OR7E54P, OST185', description='olfactory receptor family 7 subfamily E member 84 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 54 pseudogene|olfactory receptor, family 7, subfamily E, member 84 pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9468826, end_position_on_the_genomic_accession=9470249, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  259217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259217, CurrentID=0, Status='live', Symbol='HSPA12A', Aliases='', description='heat shock protein family A (Hsp70) member 12A', other_designations='heat shock 70kD protein 12A|heat shock 70kDa protein 12A', map_location='10q26.12', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=116671192, end_position_on_the_genomic_accession=116850814, orientation='minus', exon_count=17, OMIM=610701, no_hdr0=''),
  105378962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378962, CurrentID=0, Status='live', Symbol='LOC105378962', Aliases='', description='uncharacterized LOC105378962', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52289023, end_position_on_the_genomic_accession=52339234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105378963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378963, CurrentID=0, Status='live', Symbol='LOC105378963', Aliases='', description='uncharacterized LOC105378963', other_designations='CTD-2288O8.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52674689, end_position_on_the_genomic_accession=52689973, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378964 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378964, CurrentID=0, Status='live', Symbol='LOC105378964', Aliases='', description='uncharacterized LOC105378964', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=52710153, end_position_on_the_genomic_accession=52751776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378965, CurrentID=0, Status='live', Symbol='LOC105378965', Aliases='', description='uncharacterized LOC105378965', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=53299659, end_position_on_the_genomic_accession=53326543, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  390294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390294, CurrentID=0, Status='live', Symbol='HNRNPABP1', Aliases='', description='heterogeneous nuclear ribonucleoprotein A/B pseudogene 1', other_designations='', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10111738, end_position_on_the_genomic_accession=10112471, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378967 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378967, CurrentID=0, Status='live', Symbol='LOC105378967', Aliases='', description='uncharacterized LOC105378967', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54617771, end_position_on_the_genomic_accession=54648022, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378968, CurrentID=0, Status='live', Symbol='LOC105378968', Aliases='', description='uncharacterized LOC105378968', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54633027, end_position_on_the_genomic_accession=54642207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128153, CurrentID=0, Status='live', Symbol='SPATA17', Aliases='IQCH, MSRG-11, MSRG11', description='spermatogenesis associated 17', other_designations='IQ motif containing H|spermatogenesis-related protein 11', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=217631324, end_position_on_the_genomic_accession=217868151, orientation='plus', exon_count=12, OMIM=611032, no_hdr0=''),
  390298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390298, CurrentID=0, Status='live', Symbol='TIMM17BP1', Aliases='', description='translocase of inner mitochondrial membrane 17 homolog B (yeast) pseudogene 1', other_designations='', map_location='12p12.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=17400391, end_position_on_the_genomic_accession=17401191, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378971 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378971, CurrentID=0, Status='live', Symbol='LOC105378971', Aliases='', description='uncharacterized LOC105378971', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55039614, end_position_on_the_genomic_accession=55042380, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378972, CurrentID=0, Status='live', Symbol='LOC105378972', Aliases='', description='uncharacterized LOC105378972', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55212255, end_position_on_the_genomic_accession=55214911, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390301, CurrentID=0, Status='live', Symbol='ST13P9', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 9', other_designations='', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=33454346, end_position_on_the_genomic_accession=33455455, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  259230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259230, CurrentID=0, Status='live', Symbol='SGMS1', Aliases='MOB, MOB1, SMS1, TMEM23, hmob33', description='sphingomyelin synthase 1', other_designations='medulla oblongata-derived protein|protein Mob|transmembrane protein 23', map_location='10q11.2', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50305585, end_position_on_the_genomic_accession=50623977, orientation='minus', exon_count=15, OMIM=611573, no_hdr0=''),
  105378975 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378975, CurrentID=0, Status='live', Symbol='LOC105378975', Aliases='', description='uncharacterized LOC105378975', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=55613394, end_position_on_the_genomic_accession=55620559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  259232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259232, CurrentID=0, Status='live', Symbol='NALCN', Aliases='CLIFAHDD, CanIon, IHPRF, INNFD, VGCNL1, bA430M15.1', description='sodium leak channel, non selective', other_designations='four repeat voltage-gated ion channel|sodium leak channel, non-selective|voltage gated channel like 1|voltage gated channel-like protein 1', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=101053776, end_position_on_the_genomic_accession=101417205, orientation='minus', exon_count=45, OMIM=611549, no_hdr0=''),
  105378977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378977, CurrentID=0, Status='live', Symbol='LOC105378977', Aliases='', description='uncharacterized LOC105378977', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56296696, end_position_on_the_genomic_accession=56350724, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259234, CurrentID=0, Status='live', Symbol='DSCR10', Aliases='', description='Down syndrome critical region 10 (non-protein coding)', other_designations='Down syndrome critical region gene 10 (non-protein coding)', map_location='21q22.13', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=38206156, end_position_on_the_genomic_accession=38208644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378979, CurrentID=0, Status='live', Symbol='LOC105378979', Aliases='', description='growth/differentiation factor 3-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=56770327, end_position_on_the_genomic_accession=56773143, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  259236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259236, CurrentID=0, Status='live', Symbol='TMIE', Aliases='DFNB6', description='transmembrane inner ear', other_designations='transmembrane inner ear protein', map_location='3p21', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=46693842, end_position_on_the_genomic_accession=46710923, orientation='plus', exon_count=5, OMIM=607237, no_hdr0=''),
  105378981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378981, CurrentID=0, Status='live', Symbol='LOC105378981', Aliases='', description='uncharacterized LOC105378981', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57251031, end_position_on_the_genomic_accession=57252444, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105378982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378982, CurrentID=0, Status='live', Symbol='LOC105378982', Aliases='', description='uncharacterized LOC105378982', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=57803865, end_position_on_the_genomic_accession=57810349, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259239, CurrentID=0, Status='live', Symbol='WFDC11', Aliases='WAP11', description='WAP four-disulfide core domain 11', other_designations='protease inhibitor WAP11', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45648563, end_position_on_the_genomic_accession=45670239, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  259240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259240, CurrentID=0, Status='live', Symbol='WFDC9', Aliases='WAP9, dJ688G8.2', description='WAP four-disulfide core domain 9', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45607939, end_position_on_the_genomic_accession=45631268, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  390313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390313, CurrentID=0, Status='live', Symbol='OR5BS1P', Aliases='OR5BS1', description='olfactory receptor family 5 subfamily BS member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily BS, member 1 pseudogene', map_location='12q13.11', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=48559882, end_position_on_the_genomic_accession=48560814, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105378986 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378986, CurrentID=0, Status='live', Symbol='LOC105378986', Aliases='', description='uncharacterized LOC105378986', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=58814667, end_position_on_the_genomic_accession=58817121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105378987 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378987, CurrentID=0, Status='live', Symbol='LOC105378987', Aliases='', description='uncharacterized LOC105378987', other_designations='CTD-2254N19.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60021288, end_position_on_the_genomic_accession=60032848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105378988 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378988, CurrentID=0, Status='live', Symbol='LOC105378988', Aliases='', description='uncharacterized LOC105378988', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=59768393, end_position_on_the_genomic_accession=59817023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105378989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378989, CurrentID=0, Status='live', Symbol='LOC105378989', Aliases='', description='uncharacterized LOC105378989', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=59618802, end_position_on_the_genomic_accession=59649084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390318, CurrentID=0, Status='live', Symbol='OR10U1P', Aliases='', description='olfactory receptor family 10 subfamily U member 1 pseudogene', other_designations='olfactory receptor, family 10, subfamily U, member 1 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55193860, end_position_on_the_genomic_accession=55194815, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378991, CurrentID=0, Status='live', Symbol='LOC105378991', Aliases='', description='uncharacterized LOC105378991', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=60914863, end_position_on_the_genomic_accession=60919573, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390320, CurrentID=0, Status='live', Symbol='OR6C5P', Aliases='', description='olfactory receptor family 6 subfamily C member 5 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 5 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55311835, end_position_on_the_genomic_accession=55312757, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390321, CurrentID=0, Status='live', Symbol='OR6C1', Aliases='OST267', description='olfactory receptor family 6 subfamily C member 1', other_designations='olfactory receptor, family 6, subfamily C, member 1', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55320600, end_position_on_the_genomic_accession=55321538, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  128178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128178, CurrentID=0, Status='live', Symbol='EDARADD', Aliases='ECTD11A, ECTD11B, ED3, EDA3', description='EDAR-associated death domain', other_designations='EDAR-associated death domain protein|crinkled homolog|ectodysplasia A receptor associated death domain|ectodysplasin A receptor associated adapter protein', map_location='1q42.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=236394380, end_position_on_the_genomic_accession=236484708, orientation='plus', exon_count=7, OMIM=606603, no_hdr0=''),
  390323 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390323, CurrentID=0, Status='live', Symbol='OR6C75', Aliases='', description='olfactory receptor family 6 subfamily C member 75', other_designations='olfactory receptor, family 6, subfamily C, member 75', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55365111, end_position_on_the_genomic_accession=55366049, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390324, CurrentID=0, Status='live', Symbol='OR6C71P', Aliases='', description='olfactory receptor family 6 subfamily C member 71 pseudogene', other_designations='olfactory receptor, family 6, subfamily C, member 71 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55376800, end_position_on_the_genomic_accession=55377735, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105378997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378997, CurrentID=0, Status='live', Symbol='LOC105378997', Aliases='', description='uncharacterized LOC105378997', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=61624358, end_position_on_the_genomic_accession=61631575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390326, CurrentID=0, Status='live', Symbol='OR6C76', Aliases='', description='olfactory receptor family 6 subfamily C member 76', other_designations='olfactory receptor, family 6, subfamily C, member 76', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55426254, end_position_on_the_genomic_accession=55427192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390327, CurrentID=0, Status='live', Symbol='OR6C70', Aliases='', description='olfactory receptor family 6 subfamily C member 70', other_designations='olfactory receptor, family 6, subfamily C, member 70', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55469200, end_position_on_the_genomic_accession=55659844, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105379000 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379000, CurrentID=0, Status='live', Symbol='LOC105379000', Aliases='', description='uncharacterized LOC105379000', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=64151280, end_position_on_the_genomic_accession=64165343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390329, CurrentID=0, Status='live', Symbol='OR6U2P', Aliases='OR6U1P', description='olfactory receptor family 6 subfamily U member 2 pseudogene', other_designations='olfactory receptor, family 6, subfamily U, member 1 pseudogene|olfactory receptor, family 6, subfamily U, member 2 pseudogene', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55611623, end_position_on_the_genomic_accession=55612554, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379002, CurrentID=0, Status='live', Symbol='LOC105379002', Aliases='', description='uncharacterized LOC105379002', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66504152, end_position_on_the_genomic_accession=66511835, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379003, CurrentID=0, Status='live', Symbol='LOC105379003', Aliases='', description='uncharacterized LOC105379003', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66346433, end_position_on_the_genomic_accession=66440488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379004, CurrentID=0, Status='live', Symbol='LOC105379004', Aliases='', description='uncharacterized LOC105379004', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=66280649, end_position_on_the_genomic_accession=66298081, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379005, CurrentID=0, Status='live', Symbol='LOC105379005', Aliases='', description='uncharacterized LOC105379005', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67692566, end_position_on_the_genomic_accession=67800939, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379006 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379006, CurrentID=0, Status='live', Symbol='LOC105379006', Aliases='', description='uncharacterized LOC105379006', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67632676, end_position_on_the_genomic_accession=67646787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379007 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379007, CurrentID=0, Status='live', Symbol='LOC105379007', Aliases='', description='uncharacterized LOC105379007', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=67910871, end_position_on_the_genomic_accession=67950678, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379008 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379008, CurrentID=0, Status='live', Symbol='LOC105379008', Aliases='', description='uncharacterized LOC105379008', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68239378, end_position_on_the_genomic_accession=68243345, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100421702 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100421702, CurrentID=0, Status='live', Symbol='BNIP3P40', Aliases='', description='BCL2/adenovirus E1B 19kDa interacting protein 3 pseudogene 40', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=24098427, end_position_on_the_genomic_accession=24098980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  259266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259266, CurrentID=0, Status='live', Symbol='ASPM', Aliases='ASP, Calmbp1, MCPH5', description='abnormal spindle microtubule assembly', other_designations='asp (abnormal spindle) homolog, microcephaly associated', map_location='1q31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=197084127, end_position_on_the_genomic_accession=197146694, orientation='minus', exon_count=28, OMIM=605481, no_hdr0=''),
  105379011 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379011, CurrentID=0, Status='live', Symbol='LOC105379011', Aliases='', description='uncharacterized LOC105379011', other_designations='CTC-537E7.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68508245, end_position_on_the_genomic_accession=68759670, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  390340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390340, CurrentID=0, Status='live', Symbol='PDCL3P7', Aliases='', description='phosducin-like 3 pseudogene 7', other_designations='phosducin-like protein 3-like', map_location='12q14.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=66275933, end_position_on_the_genomic_accession=66276649, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379013, CurrentID=0, Status='live', Symbol='LOC105379013', Aliases='', description='uncharacterized LOC105379013', other_designations='CTC-537E7.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68429586, end_position_on_the_genomic_accession=68444977, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379014 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379014, CurrentID=0, Status='live', Symbol='LOC105379014', Aliases='', description='uncharacterized LOC105379014', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68792666, end_position_on_the_genomic_accession=68832704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379015 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379015, CurrentID=0, Status='live', Symbol='LOC105379015', Aliases='', description='uncharacterized LOC105379015', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=68941950, end_position_on_the_genomic_accession=68951328, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100233156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100233156, CurrentID=0, Status='live', Symbol='LOC100233156', Aliases='', description='tektin 4 pseudogene', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  652489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652489, CurrentID=0, Status='live', Symbol='SUMO2P3', Aliases='', description='SUMO2 pseudogene 3', other_designations='SMT3 suppressor of mif two 3 homolog 2 pseudogene 3', map_location='7p11.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=55732024, end_position_on_the_genomic_accession=55733013, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379018 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379018, CurrentID=0, Status='live', Symbol='LOC105379018', Aliases='', description='uncharacterized LOC105379018', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70798133, end_position_on_the_genomic_accession=70804737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379019 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379019, CurrentID=0, Status='live', Symbol='LOC105379019', Aliases='', description='uncharacterized LOC105379019', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70738921, end_position_on_the_genomic_accession=70741927, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379020 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379020, CurrentID=0, Status='live', Symbol='LOC105379020', Aliases='', description='uncharacterized LOC105379020', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70585910, end_position_on_the_genomic_accession=70598722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379021 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379021, CurrentID=0, Status='live', Symbol='LOC105379021', Aliases='', description='uncharacterized LOC105379021', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70382241, end_position_on_the_genomic_accession=70401369, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379022 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379022, CurrentID=0, Status='live', Symbol='LOC105379022', Aliases='', description='uncharacterized LOC105379022', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70050964, end_position_on_the_genomic_accession=70054666, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379023 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379023, CurrentID=0, Status='live', Symbol='LOC105379023', Aliases='', description='uncharacterized LOC105379023', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69923084, end_position_on_the_genomic_accession=69924616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390352, CurrentID=0, Status='live', Symbol='EEF1A1P33', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 33', other_designations='', map_location='12q23.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=96901094, end_position_on_the_genomic_accession=96902698, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128209, CurrentID=0, Status='live', Symbol='KLF17', Aliases='ZNF393, Zfp393', description='Kruppel-like factor 17', other_designations='novel zinc-finger protein|zinc finger protein 393', map_location='1p34.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=44043927, end_position_on_the_genomic_accession=44135140, orientation='plus', exon_count=10, OMIM=609602, no_hdr0=''),
  259282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259282, CurrentID=0, Status='live', Symbol='BOD1L1', Aliases='BOD1L, FAM44A', description='biorientation of chromosomes in cell division 1-like 1', other_designations='family with sequence similarity 44, member A', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13568738, end_position_on_the_genomic_accession=13627705, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  259283 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259283, CurrentID=0, Status='live', Symbol='MDS2', Aliases='', description='myelodysplastic syndrome 2 translocation associated', other_designations='', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=23627334, end_position_on_the_genomic_accession=23640566, orientation='plus', exon_count=7, OMIM=607305, no_hdr0=''),
  105379028 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379028, CurrentID=0, Status='live', Symbol='LOC105379028', Aliases='', description='uncharacterized LOC105379028', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=72074584, end_position_on_the_genomic_accession=72108954, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  259285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259285, CurrentID=0, Status='live', Symbol='TAS2R39', Aliases='T2R39, T2R57', description='taste 2 receptor member 39', other_designations='taste receptor type 2 member 57|taste receptor, type 2, member 39', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143183419, end_position_on_the_genomic_accession=143184435, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259286, CurrentID=0, Status='live', Symbol='TAS2R40', Aliases='GPR60, T2R40, T2R58', description='taste 2 receptor member 40', other_designations='G protein-coupled receptor 60|G-protein coupled receptor 60|taste receptor type 2 member 58|taste receptor, type 2, member 40', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143222079, end_position_on_the_genomic_accession=143223050, orientation='plus', exon_count=1, OMIM=613964, no_hdr0=''),
  259287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259287, CurrentID=0, Status='live', Symbol='TAS2R41', Aliases='T2R41, T2R59', description='taste 2 receptor member 41', other_designations='taste receptor type 2 member 59|taste receptor, type 2, member 41', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143477873, end_position_on_the_genomic_accession=143478796, orientation='plus', exon_count=1, OMIM=613965, no_hdr0=''),
  105379032 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379032, CurrentID=0, Status='live', Symbol='LOC105379032', Aliases='', description='uncharacterized LOC105379032', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=73287647, end_position_on_the_genomic_accession=73290488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259289 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259289, CurrentID=0, Status='live', Symbol='TAS2R43', Aliases='T2R43, T2R52', description='taste 2 receptor member 43', other_designations='taste receptor type 2 member 52|taste receptor, type 2, member 43', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11091287, end_position_on_the_genomic_accession=11092313, orientation='minus', exon_count=1, OMIM=612668, no_hdr0=''),
  259290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259290, CurrentID=0, Status='live', Symbol='TAS2R31', Aliases='T2R31, T2R44, T2R53, TAS2R44', description='taste 2 receptor member 31', other_designations='taste receptor type 2 member 44|taste receptor type 2 member 53|taste receptor, type 2, member 31|taste receptor, type 2, member 44', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11030387, end_position_on_the_genomic_accession=11031407, orientation='minus', exon_count=1, OMIM=612669, no_hdr0=''),
  259291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259291, CurrentID=0, Status='live', Symbol='TAS2R45', Aliases='GPR59, T2R45, ZG24P', description='taste 2 receptor member 45', other_designations='G-protein coupled receptor 59|taste receptor, type 2, member 45', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=613967, no_hdr0=''),
  259292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259292, CurrentID=0, Status='live', Symbol='TAS2R46', Aliases='T2R46, T2R54', description='taste 2 receptor member 46', other_designations='taste receptor type 2 member 54|taste receptor, type 2, member 46', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11061365, end_position_on_the_genomic_accession=11062294, orientation='minus', exon_count=1, OMIM=612774, no_hdr0=''),
  259293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259293, CurrentID=0, Status='live', Symbol='TAS2R30', Aliases='T2R30, T2R47, TAS2R47', description='taste 2 receptor member 30', other_designations='taste receptor T2R47|taste receptor type 2 member 47|taste receptor, type 2, member 30|taste receptor, type 2, member 47|type 2 taste receptor member 30|type 2 taste receptor member 47', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11133285, end_position_on_the_genomic_accession=11134244, orientation='minus', exon_count=1, OMIM=613963, no_hdr0=''),
  259294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259294, CurrentID=0, Status='live', Symbol='TAS2R19', Aliases='MSTP058, T2R19, T2R23, T2R48, TAS2R23, TAS2R48', description='taste 2 receptor member 19', other_designations='taste receptor, type 2, member 19|taste receptor, type 2, member 23|taste receptor, type 2, member 48', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11021619, end_position_on_the_genomic_accession=11022620, orientation='minus', exon_count=1, OMIM=613961, no_hdr0=''),
  259295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259295, CurrentID=0, Status='live', Symbol='TAS2R20', Aliases='T2R20, T2R49, T2R56, TAS2R49', description='taste 2 receptor member 20', other_designations='taste receptor type 2 member 49|taste receptor type 2 member 56|taste receptor, type 2, member 20|taste receptor, type 2, member 49', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10995962, end_position_on_the_genomic_accession=10997875, orientation='minus', exon_count=1, OMIM=613962, no_hdr0=''),
  259296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259296, CurrentID=0, Status='live', Symbol='TAS2R50', Aliases='T2R50, T2R51, TAS2R51', description='taste 2 receptor member 50', other_designations='taste receptor type 2 member 51|taste receptor, type 2, member 50', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=10985913, end_position_on_the_genomic_accession=10986912, orientation='minus', exon_count=1, OMIM=609627, no_hdr0=''),
  105379041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379041, CurrentID=0, Status='live', Symbol='LOC105379041', Aliases='', description='uncharacterized LOC105379041', other_designations='CTC-235G5.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76081412, end_position_on_the_genomic_accession=76082986, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379042, CurrentID=0, Status='live', Symbol='LOC105379042', Aliases='', description='uncharacterized LOC105379042', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76294995, end_position_on_the_genomic_accession=76311454, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379043 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379043, CurrentID=0, Status='live', Symbol='LOC105379043', Aliases='', description='uncharacterized LOC105379043', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=76975740, end_position_on_the_genomic_accession=76981882, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379044 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379044, CurrentID=0, Status='live', Symbol='LOC105379044', Aliases='', description='uncharacterized LOC105379044', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=77127580, end_position_on_the_genomic_accession=77131128, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128229, CurrentID=0, Status='live', Symbol='TSACC', Aliases='C1orf182, SIP, SSTK-IP', description='TSSK6 activating co-chaperone', other_designations='SSTK-interacting protein (SSTK-IP)', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156337314, end_position_on_the_genomic_accession=156346995, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379046, CurrentID=0, Status='live', Symbol='LOC105379046', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79513678, end_position_on_the_genomic_accession=79515303, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379047, CurrentID=0, Status='live', Symbol='LOC105379047', Aliases='', description='uncharacterized LOC105379047', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79547573, end_position_on_the_genomic_accession=79579305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379048 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379048, CurrentID=0, Status='live', Symbol='LOC105379048', Aliases='', description='uncharacterized LOC105379048', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=79859861, end_position_on_the_genomic_accession=79975850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379049, CurrentID=0, Status='live', Symbol='LOC105379049', Aliases='', description='uncharacterized LOC105379049', other_designations='CTD-2015H6.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80483120, end_position_on_the_genomic_accession=80487966, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390378, CurrentID=0, Status='live', Symbol='PPIAP28', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 28', other_designations='', map_location='13q12.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=20179064, end_position_on_the_genomic_accession=20179462, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  259307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259307, CurrentID=0, Status='live', Symbol='IL4I1', Aliases='FIG1, LAAO, LAO', description='interleukin 4 induced 1', other_designations='Fig-1 protein|IL4-induced protein 1|interleukin four induced 1', map_location='19q13.3-q13.4', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49889654, end_position_on_the_genomic_accession=49929539, orientation='minus', exon_count=12, OMIM=609742, no_hdr0=''),
  259308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=259308, CurrentID=0, Status='live', Symbol='FAM205A', Aliases='C9orf144B', description='family with sequence similarity 205 member A', other_designations='family with sequence similarity 205, member A|transmembrane protein C9orf144B', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34723053, end_position_on_the_genomic_accession=34729538, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379053, CurrentID=0, Status='live', Symbol='LOC105379053', Aliases='', description='uncharacterized LOC105379053', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83388268, end_position_on_the_genomic_accession=83398262, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379054, CurrentID=0, Status='live', Symbol='LOC105379054', Aliases='', description='uncharacterized LOC105379054', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=83507040, end_position_on_the_genomic_accession=83562372, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  128239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128239, CurrentID=0, Status='live', Symbol='IQGAP3', Aliases='', description='IQ motif containing GTPase activating protein 3', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156525405, end_position_on_the_genomic_accession=156572604, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  128240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128240, CurrentID=0, Status='live', Symbol='APOA1BP', Aliases='AIBP, YJEFN1', description='apolipoprotein A-I binding protein', other_designations='AI-BP|NAD(P)HX epimerase|apoA-I binding protein|apolipoprotein A-I-binding protein|yjeF N-terminal domain-containing protein 1|yjeF_N1', map_location='1q21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156591766, end_position_on_the_genomic_accession=156594299, orientation='plus', exon_count=6, OMIM=608862, no_hdr0=''),
  105379057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379057, CurrentID=0, Status='live', Symbol='LOC105379057', Aliases='', description='uncharacterized LOC105379057', other_designations='CTD-2269F5.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84384790, end_position_on_the_genomic_accession=84538785, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105379058 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379058, CurrentID=0, Status='live', Symbol='LOC105379058', Aliases='', description='uncharacterized LOC105379058', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84161458, end_position_on_the_genomic_accession=84176489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379059, CurrentID=0, Status='live', Symbol='LOC105379059', Aliases='', description='uncharacterized LOC105379059', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84566751, end_position_on_the_genomic_accession=84590484, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379060 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379060, CurrentID=0, Status='live', Symbol='LOC105379060', Aliases='', description='uncharacterized LOC105379060', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=84590536, end_position_on_the_genomic_accession=84605813, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422526, CurrentID=0, Status='live', Symbol='LOC100422526', Aliases='', description='proteasome activator subunit 3 pseudogene', other_designations='proteasome (prosome, macropain) activator subunit 3 (PA28 gamma; Ki) pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=161890940, end_position_on_the_genomic_accession=161891677, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379062 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379062, CurrentID=0, Status='live', Symbol='LOC105379062', Aliases='', description='uncharacterized LOC105379062', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=85805313, end_position_on_the_genomic_accession=85856116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379063 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379063, CurrentID=0, Status='live', Symbol='LOC105379063', Aliases='', description='uncharacterized LOC105379063', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86010709, end_position_on_the_genomic_accession=86031103, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379064 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379064, CurrentID=0, Status='live', Symbol='LOC105379064', Aliases='', description='uncharacterized LOC105379064', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=86355018, end_position_on_the_genomic_accession=86429992, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379065 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379065, CurrentID=0, Status='live', Symbol='LOC105379065', Aliases='', description='uncharacterized LOC105379065', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87066565, end_position_on_the_genomic_accession=87075365, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390394, CurrentID=0, Status='live', Symbol='EIF4A1P5', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 5', other_designations='eukaryotic translation initiation factor 4A pseudogene', map_location='13q13.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=36938115, end_position_on_the_genomic_accession=36939546, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379067, CurrentID=0, Status='live', Symbol='LOC105379067', Aliases='', description='uncharacterized LOC105379067', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87372742, end_position_on_the_genomic_accession=87376386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379068 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379068, CurrentID=0, Status='live', Symbol='LOC105379068', Aliases='', description='uncharacterized LOC105379068', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87364306, end_position_on_the_genomic_accession=87366801, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379069, CurrentID=0, Status='live', Symbol='LOC105379069', Aliases='', description='uncharacterized LOC105379069', other_designations='CTD-2316B1.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=87680560, end_position_on_the_genomic_accession=87715578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379070, CurrentID=0, Status='live', Symbol='LOC105379070', Aliases='', description='uncharacterized LOC105379070', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88100277, end_position_on_the_genomic_accession=88111441, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379071, CurrentID=0, Status='live', Symbol='LOC105379071', Aliases='', description='uncharacterized LOC105379071', other_designations='CTC-467M3.3', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=88691761, end_position_on_the_genomic_accession=88693863, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379072, CurrentID=0, Status='live', Symbol='LOC105379072', Aliases='', description='uncharacterized LOC105379072', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89003917, end_position_on_the_genomic_accession=89468238, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105379073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379073, CurrentID=0, Status='live', Symbol='LOC105379073', Aliases='', description='uncharacterized LOC105379073', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89401004, end_position_on_the_genomic_accession=89430351, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  390402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390402, CurrentID=0, Status='live', Symbol='AKR1B1P4', Aliases='ALDRL4', description='aldo-keto reductase family 1, member B1 pseudogene 4', other_designations='aldehyde reductase (aldose reductase)-like 4|aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=45687183, end_position_on_the_genomic_accession=45688510, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390403 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390403, CurrentID=0, Status='live', Symbol='FAM206BP', Aliases='', description='family with sequence similarity 206 member B, pseudogene', other_designations='', map_location='13q14.13', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=46270077, end_position_on_the_genomic_accession=46270622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379076, CurrentID=0, Status='live', Symbol='LOC105379076', Aliases='', description='uncharacterized LOC105379076', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=89581218, end_position_on_the_genomic_accession=89794279, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379077, CurrentID=0, Status='live', Symbol='LOC105379077', Aliases='', description='uncharacterized LOC105379077', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=90712377, end_position_on_the_genomic_accession=90741177, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379078, CurrentID=0, Status='live', Symbol='LOC105379078', Aliases='', description='uncharacterized LOC105379078', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=91687476, end_position_on_the_genomic_accession=91721364, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  390407 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390407, CurrentID=0, Status='live', Symbol='EIF4A1P6', Aliases='', description='eukaryotic translation initiation factor 4A1 pseudogene 6', other_designations='', map_location='13q21.2', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=60636719, end_position_on_the_genomic_accession=60638111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379080, CurrentID=0, Status='live', Symbol='LOC105379080', Aliases='', description='uncharacterized LOC105379080', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=92329173, end_position_on_the_genomic_accession=92334443, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379081, CurrentID=0, Status='live', Symbol='LOC105379081', Aliases='', description='uncharacterized LOC105379081', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=92082507, end_position_on_the_genomic_accession=92416514, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379082, CurrentID=0, Status='live', Symbol='LOC105379082', Aliases='', description='uncharacterized LOC105379082', other_designations='CTC-458G6.2|CTC-529L17.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=92556655, end_position_on_the_genomic_accession=92939630, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  390411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390411, CurrentID=0, Status='live', Symbol='NPM1P22', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 22', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene|nucleophosmin pseudogene', map_location='13q21.32', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=67831564, end_position_on_the_genomic_accession=67832843, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379084 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379084, CurrentID=0, Status='live', Symbol='LOC105379084', Aliases='', description='uncharacterized LOC105379084', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93571682, end_position_on_the_genomic_accession=93574817, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390413, CurrentID=0, Status='live', Symbol='RPL7P44', Aliases='RPL7_17_1325', description='ribosomal protein L7 pseudogene 44', other_designations='', map_location='13q22.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=76981112, end_position_on_the_genomic_accession=76981937, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379086 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379086, CurrentID=0, Status='live', Symbol='LOC105379086', Aliases='', description='uncharacterized LOC105379086', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=94577869, end_position_on_the_genomic_accession=94580575, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390415, CurrentID=0, Status='live', Symbol='CCT5P2', Aliases='CCT5-P1', description='chaperonin containing TCP1, subunit 5 (epsilon) pseudogene 2', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=78908930, end_position_on_the_genomic_accession=78910547, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128272, CurrentID=0, Status='live', Symbol='ARHGEF19', Aliases='WGEF', description='Rho guanine nucleotide exchange factor 19', other_designations='Rho guanine nucleotide exchange factor (GEF) 19|ephexin-2|weakly similar to Rho GEF', map_location='1p36.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=16197856, end_position_on_the_genomic_accession=16212671, orientation='minus', exon_count=17, OMIM=612496, no_hdr0=''),
  105379089 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379089, CurrentID=0, Status='live', Symbol='LOC105379089', Aliases='', description='uncharacterized LOC105379089', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95366377, end_position_on_the_genomic_accession=95391172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390418, CurrentID=0, Status='live', Symbol='KRT18P27', Aliases='', description='keratin 18 pseudogene 27', other_designations='', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=90230331, end_position_on_the_genomic_accession=90231737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390419, CurrentID=0, Status='live', Symbol='PPIAP23', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 23', other_designations='peptidylprolyl isomerase pseudogene', map_location='13q31.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=91271883, end_position_on_the_genomic_accession=91272619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379092 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379092, CurrentID=0, Status='live', Symbol='LOC105379092', Aliases='', description='uncharacterized LOC105379092', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96808234, end_position_on_the_genomic_accession=96814158, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390421, CurrentID=0, Status='live', Symbol='GAPDHP22', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 22', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=99190579, end_position_on_the_genomic_accession=99190991, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379094 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379094, CurrentID=0, Status='live', Symbol='LOC105379094', Aliases='', description='uncharacterized LOC105379094', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=96748621, end_position_on_the_genomic_accession=96752513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390423, CurrentID=0, Status='live', Symbol='ARF4P3', Aliases='', description='ADP ribosylation factor 4 pseudogene 3', other_designations='ADP-ribosylation factor 4 pseudogene 3', map_location='13q32.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=100755392, end_position_on_the_genomic_accession=100757162, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390424, CurrentID=0, Status='live', Symbol='ATP5G1P5', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C1 (subunit 9) pseudogene 5', other_designations='ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C1 (subunit 9) pseudogene 5', map_location='13q33.3', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=106663392, end_position_on_the_genomic_accession=106663950, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379097 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379097, CurrentID=0, Status='live', Symbol='LOC105379097', Aliases='', description='uncharacterized LOC105379097', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=95964871, end_position_on_the_genomic_accession=95980527, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379098 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379098, CurrentID=0, Status='live', Symbol='LOC105379098', Aliases='', description='uncharacterized LOC105379098', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98066789, end_position_on_the_genomic_accession=98161239, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105379099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379099, CurrentID=0, Status='live', Symbol='LOC105379099', Aliases='', description='uncharacterized LOC105379099', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=98841082, end_position_on_the_genomic_accession=98841928, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379100, CurrentID=0, Status='live', Symbol='LOC105379100', Aliases='', description='uncharacterized LOC105379100', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=100397558, end_position_on_the_genomic_accession=100443717, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  390429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390429, CurrentID=0, Status='live', Symbol='OR4N2', Aliases='OR14-13, OR14-8', description='olfactory receptor family 4 subfamily N member 2', other_designations='olfactory receptor OR14-13|olfactory receptor OR14-8|olfactory receptor, family 4, subfamily N, member 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19797293, end_position_on_the_genomic_accession=19828973, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379102, CurrentID=0, Status='live', Symbol='LOC105379102', Aliases='', description='uncharacterized LOC105379102', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=101591655, end_position_on_the_genomic_accession=101815736, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390431, CurrentID=0, Status='live', Symbol='OR4K2', Aliases='OR14-15', description='olfactory receptor family 4 subfamily K member 2', other_designations='olfactory receptor OR14-15|olfactory receptor, family 4, subfamily K, member 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=19876119, end_position_on_the_genomic_accession=19877303, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390432, CurrentID=0, Status='live', Symbol='OR4Q2', Aliases='OR4Q2P', description='olfactory receptor family 4 subfamily Q member 2 (gene/pseudogene)', other_designations='Olfactory receptor 4Q2|olfactory receptor, family 4, subfamily Q, member 2 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20002162, end_position_on_the_genomic_accession=20003092, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390433, CurrentID=0, Status='live', Symbol='OR4K13', Aliases='OR14-27', description='olfactory receptor family 4 subfamily K member 13', other_designations='olfactory receptor OR14-27|olfactory receptor, family 4, subfamily K, member 13', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20033844, end_position_on_the_genomic_accession=20034758, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390434, CurrentID=0, Status='live', Symbol='OR4U1P', Aliases='OR14-23', description='olfactory receptor family 4 subfamily U member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily U, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20043916, end_position_on_the_genomic_accession=20044855, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390435, CurrentID=0, Status='live', Symbol='OR4T1P', Aliases='', description='olfactory receptor family 4 subfamily T member 1 pseudogene', other_designations='olfactory receptor, family 4, subfamily T, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20093356, end_position_on_the_genomic_accession=20094276, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390436, CurrentID=0, Status='live', Symbol='OR4K17', Aliases='OR14-29', description='olfactory receptor family 4 subfamily K member 17', other_designations='olfactory receptor OR14-29|olfactory receptor, family 4, subfamily K, member 17', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20117407, end_position_on_the_genomic_accession=20118438, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390437, CurrentID=0, Status='live', Symbol='OR4N5', Aliases='', description='olfactory receptor family 4 subfamily N member 5', other_designations='olfactory receptor OR14-33|olfactory receptor, family 4, subfamily N, member 5', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20143736, end_position_on_the_genomic_accession=20144662, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390438 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390438, CurrentID=0, Status='live', Symbol='OR11G1P', Aliases='', description='olfactory receptor family 11 subfamily G member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily G, member 1 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20179127, end_position_on_the_genomic_accession=20180135, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390439, CurrentID=0, Status='live', Symbol='OR11G2', Aliases='OR14-34', description='olfactory receptor family 11 subfamily G member 2', other_designations='olfactory receptor OR14-34|olfactory receptor, family 11, subfamily G, member 2', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20197336, end_position_on_the_genomic_accession=20198373, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390440, CurrentID=0, Status='live', Symbol='OR11H5P', Aliases='', description='olfactory receptor family 11 subfamily H member 5 pseudogene', other_designations='olfactory receptor, family 11, subfamily H, member 5 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20209298, end_position_on_the_genomic_accession=20210196, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390441, CurrentID=0, Status='live', Symbol='OR11H7', Aliases='OR11H7P', description='olfactory receptor family 11 subfamily H member 7 (gene/pseudogene)', other_designations='olfactory receptor 11H7|olfactory receptor, family 11, subfamily H, member 7 pseudogene', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20229403, end_position_on_the_genomic_accession=20230344, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390442, CurrentID=0, Status='live', Symbol='OR11H4', Aliases='OR14-36', description='olfactory receptor family 11 subfamily H member 4', other_designations='olfactory receptor OR14-36|olfactory receptor, family 11, subfamily H, member 4', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20242792, end_position_on_the_genomic_accession=20243766, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390443, CurrentID=0, Status='live', Symbol='RNASE9', Aliases='HEL128, RAK1, h461', description='ribonuclease, RNase A family, 9 (non-active)', other_designations='ribonuclear enzyme|ribonuclease A K1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=20556093, end_position_on_the_genomic_accession=20560931, orientation='minus', exon_count=5, OMIM=614014, no_hdr0=''),
  105379116 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379116, CurrentID=0, Status='live', Symbol='LOC105379116', Aliases='', description='uncharacterized LOC105379116', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109451511, end_position_on_the_genomic_accession=109466180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390445, CurrentID=0, Status='live', Symbol='OR5AU1', Aliases='OR14-38', description='olfactory receptor family 5 subfamily AU member 1', other_designations='olfactory receptor OR14-38|olfactory receptor, family 5, subfamily AU, member 1', map_location='14q11.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=21154937, end_position_on_the_genomic_accession=21156025, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379118 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379118, CurrentID=0, Status='live', Symbol='LOC105379118', Aliases='', description='uncharacterized LOC105379118', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109666064, end_position_on_the_genomic_accession=109677749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379119, CurrentID=0, Status='live', Symbol='LOC105379119', Aliases='', description='uncharacterized LOC105379119', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109907504, end_position_on_the_genomic_accession=109912259, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379120, CurrentID=0, Status='live', Symbol='LOC105379120', Aliases='', description='uncharacterized LOC105379120', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=110289248, end_position_on_the_genomic_accession=110323455, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379121, CurrentID=0, Status='live', Symbol='LOC105379121', Aliases='', description='uncharacterized LOC105379121', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111008933, end_position_on_the_genomic_accession=111014620, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379122, CurrentID=0, Status='live', Symbol='LOC105379122', Aliases='', description='uncharacterized LOC105379122', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=111826784, end_position_on_the_genomic_accession=111845362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379123, CurrentID=0, Status='live', Symbol='LOC105379123', Aliases='', description='uncharacterized LOC105379123', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112922443, end_position_on_the_genomic_accession=112934261, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  128308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128308, CurrentID=0, Status='live', Symbol='MRPL55', Aliases='AAVG5835, L55nt, MRP-L55, PRO19675', description='mitochondrial ribosomal protein L55', other_designations='L55mt', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228106679, end_position_on_the_genomic_accession=228110953, orientation='minus', exon_count=7, OMIM=611859, no_hdr0=''),
  105379125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379125, CurrentID=0, Status='live', Symbol='LOC105379125', Aliases='', description='uncharacterized LOC105379125', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112981129, end_position_on_the_genomic_accession=112984656, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379126, CurrentID=0, Status='live', Symbol='LOC105379126', Aliases='', description='uncharacterized LOC105379126', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112971853, end_position_on_the_genomic_accession=112976618, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379127 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379127, CurrentID=0, Status='live', Symbol='LOC105379127', Aliases='', description='uncharacterized LOC105379127', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=113776433, end_position_on_the_genomic_accession=113777286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128312, CurrentID=0, Status='live', Symbol='HIST3H2BB', Aliases='H2Bb', description='histone cluster 3, H2bb', other_designations='H2B type 12|histone 3, H2bb', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228458107, end_position_on_the_genomic_accession=228458558, orientation='plus', exon_count=1, OMIM=615046, no_hdr0=''),
  105379129 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379129, CurrentID=0, Status='live', Symbol='LOC105379129', Aliases='', description='uncharacterized LOC105379129', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115296928, end_position_on_the_genomic_accession=115338873, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379130, CurrentID=0, Status='live', Symbol='LOC105379130', Aliases='', description='uncharacterized LOC105379130', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115364202, end_position_on_the_genomic_accession=115370216, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379131, CurrentID=0, Status='live', Symbol='LOC105379131', Aliases='', description='uncharacterized LOC105379131', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=115813620, end_position_on_the_genomic_accession=115816111, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379132, CurrentID=0, Status='live', Symbol='LOC105379132', Aliases='', description='uncharacterized LOC105379132', other_designations='CTB-118N6.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116574916, end_position_on_the_genomic_accession=116632174, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105379133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379133, CurrentID=0, Status='live', Symbol='LOC105379133', Aliases='', description='uncharacterized LOC105379133', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116764471, end_position_on_the_genomic_accession=116810049, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379134 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379134, CurrentID=0, Status='live', Symbol='LOC105379134', Aliases='', description='uncharacterized LOC105379134', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116917240, end_position_on_the_genomic_accession=116919948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379135, CurrentID=0, Status='live', Symbol='LOC105379135', Aliases='', description='uncharacterized LOC105379135', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116899713, end_position_on_the_genomic_accession=116917909, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379136, CurrentID=0, Status='live', Symbol='LOC105379136', Aliases='', description='uncharacterized LOC105379136', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116899356, end_position_on_the_genomic_accession=116913594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379137, CurrentID=0, Status='live', Symbol='LOC105379137', Aliases='', description='uncharacterized LOC105379137', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=116921245, end_position_on_the_genomic_accession=116953427, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  128322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128322, CurrentID=0, Status='live', Symbol='NUTF2P4', Aliases='', description='nuclear transport factor 2 pseudogene 4', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=112748095, end_position_on_the_genomic_accession=112748709, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379139, CurrentID=0, Status='live', Symbol='LOC105379139', Aliases='', description='uncharacterized LOC105379139', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117356982, end_position_on_the_genomic_accession=117370388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379140, CurrentID=0, Status='live', Symbol='LOC105379140', Aliases='', description='uncharacterized LOC105379140', other_designations='CTD-3179P9.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=117986398, end_position_on_the_genomic_accession=118001124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379141, CurrentID=0, Status='live', Symbol='LOC105379141', Aliases='', description='uncharacterized LOC105379141', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118577113, end_position_on_the_genomic_accession=118581581, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379142 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379142, CurrentID=0, Status='live', Symbol='LOC105379142', Aliases='', description='uncharacterized LOC105379142', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118696909, end_position_on_the_genomic_accession=118737666, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379143, CurrentID=0, Status='live', Symbol='LOC105379143', Aliases='', description='uncharacterized LOC105379143', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=118996312, end_position_on_the_genomic_accession=119070865, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  390472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390472, CurrentID=0, Status='live', Symbol='KRT8P2', Aliases='', description='keratin 8 pseudogene 2', other_designations='', map_location='14q21.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=43540673, end_position_on_the_genomic_accession=43541515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379145, CurrentID=0, Status='live', Symbol='LOC105379145', Aliases='', description='uncharacterized LOC105379145', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=120433121, end_position_on_the_genomic_accession=120451857, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379146, CurrentID=0, Status='live', Symbol='LOC105379146', Aliases='', description='uncharacterized LOC105379146', other_designations='CTD-2334D19.1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=120781225, end_position_on_the_genomic_accession=120792869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379147, CurrentID=0, Status='live', Symbol='LOC105379147', Aliases='', description='uncharacterized LOC105379147', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=120868488, end_position_on_the_genomic_accession=120873008, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379148, CurrentID=0, Status='live', Symbol='LOC105379148', Aliases='', description='uncharacterized LOC105379148', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121865088, end_position_on_the_genomic_accession=121872394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379149, CurrentID=0, Status='live', Symbol='LOC105379149', Aliases='', description='uncharacterized LOC105379149', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121829423, end_position_on_the_genomic_accession=121878798, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379150 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379150, CurrentID=0, Status='live', Symbol='LOC105379150', Aliases='', description='uncharacterized LOC105379150', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121821696, end_position_on_the_genomic_accession=121847689, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379151, CurrentID=0, Status='live', Symbol='LOC105379151', Aliases='', description='uncharacterized LOC105379151', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=122545300, end_position_on_the_genomic_accession=122567141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652624, CurrentID=0, Status='live', Symbol='RPSAP46', Aliases='RPSA_14_827', description='ribosomal protein SA pseudogene 46', other_designations='', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=101203600, end_position_on_the_genomic_accession=101204571, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379153 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379153, CurrentID=0, Status='live', Symbol='LOC105379153', Aliases='', description='uncharacterized LOC105379153', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123075763, end_position_on_the_genomic_accession=123083927, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128338, CurrentID=0, Status='live', Symbol='DRAM2', Aliases='PRO180, TMEM77, WWFQ154', description='DNA damage regulated autophagy modulator 2', other_designations='DNA-damage regulated autophagy modulator 2|RP5-1180E21.1|damage regulated autophagy modulator 2|transmembrane protein 77', map_location='1p13.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111117332, end_position_on_the_genomic_accession=111140216, orientation='minus', exon_count=11, OMIM=613360, no_hdr0=''),
  390483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390483, CurrentID=0, Status='live', Symbol='SALL4P7', Aliases='', description='spalt-like transcription factor 4 pseudogene 7', other_designations='sal-like 4 pseudogene 7', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60493863, end_position_on_the_genomic_accession=60496873, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379156, CurrentID=0, Status='live', Symbol='LOC105379156', Aliases='', description='uncharacterized LOC105379156', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124796866, end_position_on_the_genomic_accession=124801928, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379157 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379157, CurrentID=0, Status='live', Symbol='LOC105379157', Aliases='', description='uncharacterized LOC105379157', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124869171, end_position_on_the_genomic_accession=124876110, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379158, CurrentID=0, Status='live', Symbol='LOC105379158', Aliases='', description='uncharacterized LOC105379158', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=124986602, end_position_on_the_genomic_accession=124996103, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105379159 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379159, CurrentID=0, Status='live', Symbol='LOC105379159', Aliases='', description='uncharacterized LOC105379159', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=125388081, end_position_on_the_genomic_accession=125391260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  128344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128344, CurrentID=0, Status='live', Symbol='PIFO', Aliases='C1orf88, pitchfork', description='primary cilia formation', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111324515, end_position_on_the_genomic_accession=111353017, orientation='plus', exon_count=8, OMIM=614234, no_hdr0=''),
  105379161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379161, CurrentID=0, Status='live', Symbol='LOC105379161', Aliases='', description='uncharacterized LOC105379161', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=126432175, end_position_on_the_genomic_accession=126491786, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  128346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128346, CurrentID=0, Status='live', Symbol='C1orf162', Aliases='', description='chromosome 1 open reading frame 162', other_designations='', map_location='1p13.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111467487, end_position_on_the_genomic_accession=111478512, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105379163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379163, CurrentID=0, Status='live', Symbol='LOC105379163', Aliases='', description='uncharacterized LOC105379163', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127032113, end_position_on_the_genomic_accession=127036662, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379164, CurrentID=0, Status='live', Symbol='LOC105379164', Aliases='', description='uncharacterized LOC105379164', other_designations='CTC-548H10.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127651743, end_position_on_the_genomic_accession=127663762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379165, CurrentID=0, Status='live', Symbol='LOC105379165', Aliases='', description='uncharacterized LOC105379165', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127940234, end_position_on_the_genomic_accession=127978690, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379166, CurrentID=0, Status='live', Symbol='LOC105379166', Aliases='', description='uncharacterized LOC105379166', other_designations='CTC-228N24.2', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127826777, end_position_on_the_genomic_accession=127858418, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379167 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379167, CurrentID=0, Status='live', Symbol='LOC105379167', Aliases='', description='uncharacterized LOC105379167', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128640362, end_position_on_the_genomic_accession=128659185, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379168, CurrentID=0, Status='live', Symbol='LOC105379168', Aliases='', description='uncharacterized LOC105379168', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=128537972, end_position_on_the_genomic_accession=128845694, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379169 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379169, CurrentID=0, Status='live', Symbol='LOC105379169', Aliases='', description='uncharacterized LOC105379169', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=129345429, end_position_on_the_genomic_accession=129352457, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379170, CurrentID=0, Status='live', Symbol='LOC105379170', Aliases='', description='uncharacterized LOC105379170', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130120168, end_position_on_the_genomic_accession=130124343, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379171, CurrentID=0, Status='live', Symbol='LOC105379171', Aliases='', description='uncharacterized LOC105379171', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=130337422, end_position_on_the_genomic_accession=130379839, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379172 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379172, CurrentID=0, Status='live', Symbol='LOC105379172', Aliases='', description='uncharacterized LOC105379172', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=131019893, end_position_on_the_genomic_accession=131068547, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  390501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390501, CurrentID=0, Status='live', Symbol='MPPE1P1', Aliases='MPPE1P', description='metallophosphoesterase 1 pseudogene 1', other_designations='', map_location='14q31.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=89110108, end_position_on_the_genomic_accession=89111223, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390502, CurrentID=0, Status='live', Symbol='SERPINA2', Aliases='ARGS, ATR, PILP, psiATR, SERPINA2', description='serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 2 (gene/pseudogene)', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 2|serpin A2|serpin peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 2, pseudogene', map_location='14q32.13', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=94364318, end_position_on_the_genomic_accession=94366702, orientation='minus', exon_count=2, OMIM=107410, no_hdr0=''),
  100422545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422545, CurrentID=0, Status='live', Symbol='HIRAP1', Aliases='', description='histone cell cycle regulator pseudogene 1', other_designations='HIR histone cell cycle regulation defective homolog A pseudogene 1', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=25691539, end_position_on_the_genomic_accession=25692328, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128360, CurrentID=0, Status='live', Symbol='OR10T2', Aliases='OR1-3', description='olfactory receptor family 10 subfamily T member 2', other_designations='olfactory receptor OR1-3|olfactory receptor, family 10, subfamily T, member 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158398522, end_position_on_the_genomic_accession=158399466, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379177 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379177, CurrentID=0, Status='live', Symbol='LOC105379177', Aliases='', description='uncharacterized LOC105379177', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132958513, end_position_on_the_genomic_accession=132962693, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379178 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379178, CurrentID=0, Status='live', Symbol='LOC105379178', Aliases='', description='uncharacterized LOC105379178', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133111068, end_position_on_the_genomic_accession=133124476, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379179, CurrentID=0, Status='live', Symbol='LOC105379179', Aliases='', description='uncharacterized LOC105379179', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133799395, end_position_on_the_genomic_accession=133802808, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379180, CurrentID=0, Status='live', Symbol='CTB-3M24.3', Aliases='', description='uncharacterized LOC105379180', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133387523, end_position_on_the_genomic_accession=133395362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379181, CurrentID=0, Status='live', Symbol='CTB-49A3.2', Aliases='', description='uncharacterized LOC105379181', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=133207887, end_position_on_the_genomic_accession=133225143, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  128366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128366, CurrentID=0, Status='live', Symbol='OR6P1', Aliases='', description='olfactory receptor family 6 subfamily P member 1', other_designations='olfactory receptor OR1-12|olfactory receptor, family 6, subfamily P, member 1|seven transmembrane helix receptor', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158562651, end_position_on_the_genomic_accession=158563604, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  128367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128367, CurrentID=0, Status='live', Symbol='OR10X1', Aliases='OR1-13, OR1-14P, OR10X1', description='olfactory receptor family 10 subfamily X member 1 (gene/pseudogene)', other_designations='olfactory receptor OR1-13 pseudogene|olfactory receptor OR1-14|olfactory receptor, family 10, subfamily X, member 1 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158578919, end_position_on_the_genomic_accession=158579899, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  128368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128368, CurrentID=0, Status='live', Symbol='OR10Z1', Aliases='OR1-15', description='olfactory receptor family 10 subfamily Z member 1', other_designations='olfactory receptor OR1-15|olfactory receptor, family 10, subfamily Z, member 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158606439, end_position_on_the_genomic_accession=158607380, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379185 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379185, CurrentID=0, Status='live', Symbol='LOC105379185', Aliases='', description='uncharacterized LOC105379185', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134074366, end_position_on_the_genomic_accession=134106400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128370, CurrentID=0, Status='live', Symbol='OR6K4P', Aliases='', description='olfactory receptor family 6 subfamily K member 4 pseudogene', other_designations='olfactory receptor, family 6, subfamily K, member 4 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158724113, end_position_on_the_genomic_accession=158725067, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128371, CurrentID=0, Status='live', Symbol='OR6K6', Aliases='OR1-21', description='olfactory receptor family 6 subfamily K member 6', other_designations='olfactory receptor OR1-21|olfactory receptor, family 6, subfamily K, member 6', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158754816, end_position_on_the_genomic_accession=158755847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  128372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128372, CurrentID=0, Status='live', Symbol='OR6N1', Aliases='OR1-22', description='olfactory receptor family 6 subfamily N member 1', other_designations='olfactory receptor OR1-22|olfactory receptor, family 6, subfamily N, member 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158765744, end_position_on_the_genomic_accession=158766682, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379189, CurrentID=0, Status='live', Symbol='LOC105379189', Aliases='', description='uncharacterized LOC105379189', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136021581, end_position_on_the_genomic_accession=136025893, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379190, CurrentID=0, Status='live', Symbol='LOC105379190', Aliases='', description='uncharacterized LOC105379190', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136144377, end_position_on_the_genomic_accession=136157399, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379191, CurrentID=0, Status='live', Symbol='CTB-1I21.1', Aliases='', description='uncharacterized LOC105379191', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136753934, end_position_on_the_genomic_accession=136755619, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379192, CurrentID=0, Status='live', Symbol='LOC105379192', Aliases='', description='uncharacterized LOC105379192', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=137128136, end_position_on_the_genomic_accession=137130969, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379193 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379193, CurrentID=0, Status='live', Symbol='LOC105379193', Aliases='', description='uncharacterized LOC105379193', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138543331, end_position_on_the_genomic_accession=138587034, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379194, CurrentID=0, Status='live', Symbol='LOC105379194', Aliases='', description='uncharacterized LOC105379194', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138744429, end_position_on_the_genomic_accession=138754097, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379195, CurrentID=0, Status='live', Symbol='LOC105379195', Aliases='', description='uncharacterized LOC105379195', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138960691, end_position_on_the_genomic_accession=138970365, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379196, CurrentID=0, Status='live', Symbol='LOC105379196', Aliases='', description='protein mago nashi homolog pseudogene', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=51427996, end_position_on_the_genomic_accession=51445124, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379197, CurrentID=0, Status='live', Symbol='LOC105379197', Aliases='', description='pleckstrin homology domain-containing family A member 7-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=103432459, end_position_on_the_genomic_accession=103503988, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  652670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652670, CurrentID=0, Status='live', Symbol='RPS3AP15', Aliases='RPS3A_2_382', description='ribosomal protein S3a pseudogene 15', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75542105, end_position_on_the_genomic_accession=75542953, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379199, CurrentID=0, Status='live', Symbol='LOC105379199', Aliases='', description='uncharacterized LOC105379199', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=127264453, end_position_on_the_genomic_accession=127291123, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379200 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379200, CurrentID=0, Status='live', Symbol='LOC105379200', Aliases='', description='skin secretory protein xP2-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=132495146, end_position_on_the_genomic_accession=132497253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379201 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379201, CurrentID=0, Status='live', Symbol='LOC105379201', Aliases='', description='putative uncharacterized protein ENSP00000344348', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=49551061, end_position_on_the_genomic_accession=49559237, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390530, CurrentID=0, Status='live', Symbol='IGHV1OR21-1', Aliases='', description='immunoglobulin heavy variable 1/OR21-1 (non-functional)', other_designations='IGHV1/OR21-1|immunoglobulin heavy variable 1/OR21-1 pseudogene', map_location='21p11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=10649390, end_position_on_the_genomic_accession=10649835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128387, CurrentID=0, Status='live', Symbol='TATDN3', Aliases='', description='TatD DNase domain containing 3', other_designations='', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=212791828, end_position_on_the_genomic_accession=212816825, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105379204 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379204, CurrentID=0, Status='live', Symbol='LOC105379204', Aliases='', description='uncharacterized LOC105379204', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=20144881, end_position_on_the_genomic_accession=20147794, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379205 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379205, CurrentID=0, Status='live', Symbol='LOC105379205', Aliases='', description='uncharacterized LOC105379205', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21179912, end_position_on_the_genomic_accession=21191753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379206 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379206, CurrentID=0, Status='live', Symbol='LOC105379206', Aliases='', description='uncharacterized LOC105379206', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21127480, end_position_on_the_genomic_accession=21134735, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390535, CurrentID=0, Status='live', Symbol='GOLGA8EP', Aliases='GOLGA8E', description='golgin A8 family member E, pseudogene', other_designations='golgi autoantigen, golgin subfamily a, 8E|golgin A8 family, member E, pseudogene', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22424673, end_position_on_the_genomic_accession=22438026, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  105379208 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379208, CurrentID=0, Status='live', Symbol='LOC105379208', Aliases='', description='uncharacterized LOC105379208', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21752748, end_position_on_the_genomic_accession=21759926, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379209, CurrentID=0, Status='live', Symbol='LOC105379209', Aliases='', description='uncharacterized LOC105379209', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=21574615, end_position_on_the_genomic_accession=21579838, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390538, CurrentID=0, Status='live', Symbol='OR4M2', Aliases='OR15-3', description='olfactory receptor family 4 subfamily M member 2', other_designations='olfactory receptor OR15-3|olfactory receptor, family 4, subfamily M, member 2', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22080527, end_position_on_the_genomic_accession=22081610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390539, CurrentID=0, Status='live', Symbol='OR4N3P', Aliases='OR15-2', description='olfactory receptor family 4 subfamily N member 3 pseudogene', other_designations='olfactory receptor, family 4, subfamily N, member 3 pseudogene|seven transmembrane helix receptor', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22125511, end_position_on_the_genomic_accession=22126447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379212 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379212, CurrentID=0, Status='live', Symbol='LOC105379212', Aliases='', description='uncharacterized LOC105379212', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51943, end_position_on_the_genomic_accession=53959, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379213 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379213, CurrentID=0, Status='live', Symbol='LOC105379213', Aliases='', description='uncharacterized LOC105379213', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=96600, end_position_on_the_genomic_accession=107727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379214, CurrentID=0, Status='live', Symbol='LOC105379214', Aliases='', description='uncharacterized LOC105379214', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176545, end_position_on_the_genomic_accession=180102, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379215 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379215, CurrentID=0, Status='live', Symbol='LOC105379215', Aliases='', description='uncharacterized LOC105379215', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7713187, end_position_on_the_genomic_accession=7721556, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379216 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379216, CurrentID=0, Status='live', Symbol='LOC105379216', Aliases='', description='uncharacterized LOC105379216', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7800198, end_position_on_the_genomic_accession=7800673, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379217, CurrentID=0, Status='live', Symbol='LOC105379217', Aliases='', description='uncharacterized LOC105379217', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7830346, end_position_on_the_genomic_accession=7834991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379218 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379218, CurrentID=0, Status='live', Symbol='LOC105379218', Aliases='', description='uncharacterized LOC105379218', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8137924, end_position_on_the_genomic_accession=8153282, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379219 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379219, CurrentID=0, Status='live', Symbol='LOC105379219', Aliases='', description='uncharacterized LOC105379219', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8101125, end_position_on_the_genomic_accession=8143741, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379220, CurrentID=0, Status='live', Symbol='LOC105379220', Aliases='', description='uncharacterized LOC105379220', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8070371, end_position_on_the_genomic_accession=8071735, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379221 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379221, CurrentID=0, Status='live', Symbol='LOC105379221', Aliases='', description='uncharacterized LOC105379221', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8220050, end_position_on_the_genomic_accession=8227930, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379222 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379222, CurrentID=0, Status='live', Symbol='LOC105379222', Aliases='', description='uncharacterized LOC105379222', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8389265, end_position_on_the_genomic_accession=8399297, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379223 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379223, CurrentID=0, Status='live', Symbol='LOC105379223', Aliases='', description='uncharacterized LOC105379223', other_designations='CTA-398F10.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8414205, end_position_on_the_genomic_accession=8424623, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  128408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128408, CurrentID=0, Status='live', Symbol='BHLHE23', Aliases='BETA4, BHLHB4, bA305P22.3', description='basic helix-loop-helix family member e23', other_designations='basic helix-loop-helix domain containing, class B, 4|basic helix-loop-helix family, member e23|class B basic helix-loop-helix protein 4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63005979, end_position_on_the_genomic_accession=63007035, orientation='minus', exon_count=1, OMIM=609331, no_hdr0=''),
  105379225 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379225, CurrentID=0, Status='live', Symbol='LOC105379225', Aliases='', description='uncharacterized LOC105379225', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8686065, end_position_on_the_genomic_accession=8686979, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379226, CurrentID=0, Status='live', Symbol='LOC105379226', Aliases='', description='uncharacterized LOC105379226', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8991622, end_position_on_the_genomic_accession=8995522, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379227 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379227, CurrentID=0, Status='live', Symbol='LOC105379227', Aliases='', description='uncharacterized LOC105379227', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=8988277, end_position_on_the_genomic_accession=9002790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379228 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379228, CurrentID=0, Status='live', Symbol='LOC105379228', Aliases='', description='uncharacterized LOC105379228', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9351266, end_position_on_the_genomic_accession=9359681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379229 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379229, CurrentID=0, Status='live', Symbol='LOC105379229', Aliases='', description='uncharacterized LOC105379229', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9383971, end_position_on_the_genomic_accession=9415960, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128414, CurrentID=0, Status='live', Symbol='NKAIN4', Aliases='C20orf58, FAM77A, bA261N11.2', description='Na+/K+ transporting ATPase interacting 4', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 4', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=63240784, end_position_on_the_genomic_accession=63254540, orientation='minus', exon_count=8, OMIM=612873, no_hdr0=''),
  105379231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379231, CurrentID=0, Status='live', Symbol='LOC105379231', Aliases='', description='uncharacterized LOC105379231', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9363288, end_position_on_the_genomic_accession=9413878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390560, CurrentID=0, Status='live', Symbol='DNM1DN4-1', Aliases='DNM1DN4.1', description='DNM1DN4.1 duplicon', other_designations='', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390561, CurrentID=0, Status='live', Symbol='HERC2P10', Aliases='', description='hect domain and RLD 2 pseudogene 10', other_designations='H_DJ1086D14.1|WUGSC:H_DJ1086D14.1', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30818036, end_position_on_the_genomic_accession=30827783, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379234, CurrentID=0, Status='live', Symbol='LOC105379234', Aliases='', description='uncharacterized LOC105379234', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10049130, end_position_on_the_genomic_accession=10053053, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379235 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379235, CurrentID=0, Status='live', Symbol='LOC105379235', Aliases='', description='uncharacterized LOC105379235', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=9985986, end_position_on_the_genomic_accession=10054993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379236, CurrentID=0, Status='live', Symbol='LOC105379236', Aliases='', description='uncharacterized LOC105379236', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10483207, end_position_on_the_genomic_accession=10487478, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379237 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379237, CurrentID=0, Status='live', Symbol='LOC105379237', Aliases='', description='uncharacterized LOC105379237', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10578724, end_position_on_the_genomic_accession=10580693, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379238, CurrentID=0, Status='live', Symbol='LOC105379238', Aliases='', description='uncharacterized LOC105379238', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10667199, end_position_on_the_genomic_accession=10672600, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379239, CurrentID=0, Status='live', Symbol='LOC105379239', Aliases='', description='uncharacterized LOC105379239', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=10856083, end_position_on_the_genomic_accession=10860279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379240, CurrentID=0, Status='live', Symbol='LOC105379240', Aliases='', description='uncharacterized LOC105379240', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11020565, end_position_on_the_genomic_accession=11033189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379241 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379241, CurrentID=0, Status='live', Symbol='LOC105379241', Aliases='', description='uncharacterized LOC105379241', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11553051, end_position_on_the_genomic_accession=11555802, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379242, CurrentID=0, Status='live', Symbol='LOC105379242', Aliases='', description='uncharacterized LOC105379242', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11632214, end_position_on_the_genomic_accession=11660901, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379243 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379243, CurrentID=0, Status='live', Symbol='LOC105379243', Aliases='', description='uncharacterized LOC105379243', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11787017, end_position_on_the_genomic_accession=11801355, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390572 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390572, CurrentID=0, Status='live', Symbol='PRELID1P4', Aliases='', description='PRELI domain containing 1 pseudogene 4', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=35121226, end_position_on_the_genomic_accession=35122248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379245 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379245, CurrentID=0, Status='live', Symbol='LOC105379245', Aliases='', description='uncharacterized LOC105379245', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12000761, end_position_on_the_genomic_accession=12007367, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379246 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379246, CurrentID=0, Status='live', Symbol='LOC105379246', Aliases='', description='uncharacterized LOC105379246', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12043489, end_position_on_the_genomic_accession=12046779, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379247, CurrentID=0, Status='live', Symbol='LOC105379247', Aliases='', description='uncharacterized LOC105379247', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11264160, end_position_on_the_genomic_accession=11284725, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  390576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390576, CurrentID=0, Status='live', Symbol='RPS15P8', Aliases='RPS15_4_1425', description='ribosomal protein S15 pseudogene 8', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=37490678, end_position_on_the_genomic_accession=37491139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379249, CurrentID=0, Status='live', Symbol='LOC105379249', Aliases='', description='uncharacterized LOC105379249', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61174252, end_position_on_the_genomic_accession=61175665, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128434, CurrentID=0, Status='live', Symbol='VSTM2L', Aliases='C20orf102, dJ1118M15.2', description='V-set and transmembrane domain containing 2 like', other_designations='', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=37903097, end_position_on_the_genomic_accession=37945350, orientation='plus', exon_count=4, OMIM=616537, no_hdr0=''),
  105379251 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379251, CurrentID=0, Status='live', Symbol='LOC105379251', Aliases='', description='uncharacterized LOC105379251', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65655751, end_position_on_the_genomic_accession=65664509, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379252 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379252, CurrentID=0, Status='live', Symbol='LOC105379252', Aliases='', description='coiled-coil domain-containing protein 29-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66047079, end_position_on_the_genomic_accession=66109222, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  390581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390581, CurrentID=0, Status='live', Symbol='ATP5HP1', Aliases='HsT18164', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit D pseudogene 1', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit d, pseudogene 1|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit d, pseudogene 1', map_location='15q15.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=43267831, end_position_on_the_genomic_accession=43268425, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379254, CurrentID=0, Status='live', Symbol='LOC105379254', Aliases='', description='uncharacterized LOC105379254', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65917795, end_position_on_the_genomic_accession=65927987, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128439, CurrentID=0, Status='live', Symbol='SNHG11', Aliases='C20orf198, LINC00101, NCRNA00101', description='small nucleolar RNA host gene 11', other_designations='long intergenic non-protein coding RNA 101|small nucleolar RNA host gene (non-protein coding) 11|small nucleolar RNA host gene 11 (non-protein coding)', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38446654, end_position_on_the_genomic_accession=38450921, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379256, CurrentID=0, Status='live', Symbol='LOC105379256', Aliases='', description='uncharacterized LOC105379256', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64711980, end_position_on_the_genomic_accession=64715140, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379257, CurrentID=0, Status='live', Symbol='LOC105379257', Aliases='', description='uncharacterized LOC105379257', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64603835, end_position_on_the_genomic_accession=64618296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390586 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390586, CurrentID=0, Status='live', Symbol='LOC390586', Aliases='', description='hydroxyacyl-CoA dehydrogenase pseudogene', other_designations='hydroxyacyl-Coenzyme A dehydrogenase pseudogene', map_location='15q21.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=56191584, end_position_on_the_genomic_accession=56193350, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379259, CurrentID=0, Status='live', Symbol='LOC105379259', Aliases='', description='uncharacterized LOC105379259', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64516017, end_position_on_the_genomic_accession=64527283, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379260 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379260, CurrentID=0, Status='live', Symbol='LOC105379260', Aliases='', description='putative zinc finger protein LOC441666', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64832222, end_position_on_the_genomic_accession=64835363, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379261 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379261, CurrentID=0, Status='live', Symbol='LOC105379261', Aliases='', description='golgin subfamily A member 6-like protein 22', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22768435, end_position_on_the_genomic_accession=22769121, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379262, CurrentID=0, Status='live', Symbol='LOC105379262', Aliases='', description='uncharacterized LOC105379262', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62288067, end_position_on_the_genomic_accession=62298373, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379263, CurrentID=0, Status='live', Symbol='LOC105379263', Aliases='', description='uncharacterized LOC105379263', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62254327, end_position_on_the_genomic_accession=62353265, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105379264 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379264, CurrentID=0, Status='live', Symbol='LOC105379264', Aliases='', description='uncharacterized LOC105379264', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9264844, end_position_on_the_genomic_accession=9269708, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379265, CurrentID=0, Status='live', Symbol='LOC105379265', Aliases='', description='uncharacterized LOC105379265', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9274397, end_position_on_the_genomic_accession=9282943, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390594, CurrentID=0, Status='live', Symbol='KBTBD13', Aliases='HCG1645727, NEM6', description='kelch repeat and BTB (POZ) domain containing 13', other_designations='nemaline myopathy type 6', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65076816, end_position_on_the_genomic_accession=65079938, orientation='plus', exon_count=1, OMIM=613727, no_hdr0=''),
  390595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390595, CurrentID=0, Status='live', Symbol='UBAP1L', Aliases='UBAP-1L', description='ubiquitin associated protein 1 like', other_designations='ubiquitin associated protein 1-like', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=65092760, end_position_on_the_genomic_accession=65106647, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379268, CurrentID=0, Status='live', Symbol='LOC105379268', Aliases='', description='uncharacterized LOC105379268', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9354851, end_position_on_the_genomic_accession=9355431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390597, CurrentID=0, Status='live', Symbol='LETM1P1', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 1, pseudogene 1', other_designations='leucine zipper-EF-hand containing transmembrane protein 1 pseudogene', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66808567, end_position_on_the_genomic_accession=66810669, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390598, CurrentID=0, Status='live', Symbol='SKOR1', Aliases='CORL1, FUSSEL15, LBXCOR1', description='SKI family transcriptional corepressor 1', other_designations='LBX1 corepressor 1|Lbxcor1 homolog|corepressor for LBX1|functional Smad-suppressing element on chromosome 15|functional smad suppressing element 15|fussel-15|ladybird homeobox corepressor 1|transcriptional corepressor CORL1', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=67819704, end_position_on_the_genomic_accession=67834561, orientation='plus', exon_count=16, OMIM=611273, no_hdr0=''),
  100422561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422561, CurrentID=0, Status='live', Symbol='LOC100422561', Aliases='', description='guanine nucleotide binding protein (G protein), beta polypeptide 2-like 1 pseudogene', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=108784098, end_position_on_the_genomic_accession=108785539, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379272, CurrentID=0, Status='live', Symbol='LOC105379272', Aliases='', description='putative ankyrin repeat domain-containing protein 20A12 pseudogene', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390601 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390601, CurrentID=0, Status='live', Symbol='KRT8P9', Aliases='', description='keratin 8 pseudogene 9', other_designations='', map_location='15q23', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=70858637, end_position_on_the_genomic_accession=70860362, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379274 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379274, CurrentID=0, Status='live', Symbol='LOC105379274', Aliases='', description='uncharacterized LOC105379274', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379275, CurrentID=0, Status='live', Symbol='LOC105379275', Aliases='', description='uncharacterized LOC105379275', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379276, CurrentID=0, Status='live', Symbol='LOC105379276', Aliases='', description='uncharacterized LOC105379276', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379277 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379277, CurrentID=0, Status='live', Symbol='LOC105379277', Aliases='', description='uncharacterized LOC105379277', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18171511, end_position_on_the_genomic_accession=18176697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379278 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379278, CurrentID=0, Status='live', Symbol='LOC105379278', Aliases='', description='uncharacterized LOC105379278', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18339183, end_position_on_the_genomic_accession=18348633, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379279 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379279, CurrentID=0, Status='live', Symbol='LOC105379279', Aliases='', description='uncharacterized LOC105379279', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379280 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379280, CurrentID=0, Status='live', Symbol='LOC105379280', Aliases='', description='uncharacterized LOC105379280', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379281, CurrentID=0, Status='live', Symbol='LOC105379281', Aliases='', description='uncharacterized LOC105379281', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390610 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390610, CurrentID=0, Status='live', Symbol='KRT8P23', Aliases='', description='keratin 8 pseudogene 23', other_designations='', map_location='15q24.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=76978829, end_position_on_the_genomic_accession=76980527, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128467, CurrentID=0, Status='live', Symbol='RPL12P11', Aliases='RPL12L2, RPL12_22_1703, dJ661I20.1', description='ribosomal protein L12 pseudogene 11', other_designations='ribosomal protein L12-like 2|ribosomal protein S12 pseudogene 11', map_location='20q12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=41547335, end_position_on_the_genomic_accession=41547966, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390612, CurrentID=0, Status='live', Symbol='RPL18P11', Aliases='RPL18_4_1455', description='ribosomal protein L18 pseudogene 11', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=78661008, end_position_on_the_genomic_accession=78661641, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379285, CurrentID=0, Status='live', Symbol='LOC105379285', Aliases='', description='uncharacterized LOC105379285', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248445607, end_position_on_the_genomic_accession=248548838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379286 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379286, CurrentID=0, Status='live', Symbol='LOC105379286', Aliases='', description='uncharacterized LOC105379286', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379287 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379287, CurrentID=0, Status='live', Symbol='LOC105379287', Aliases='', description='uncharacterized LOC105379287', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390616, CurrentID=0, Status='live', Symbol='ANKRD34C', Aliases='', description='ankyrin repeat domain 34C', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79282804, end_position_on_the_genomic_accession=79298239, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390617, CurrentID=0, Status='live', Symbol='LOC390617', Aliases='', description='transcription factor Dp-1 pseudogene', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=79556565, end_position_on_the_genomic_accession=79560673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379290 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379290, CurrentID=0, Status='live', Symbol='LOC105379290', Aliases='', description='uncharacterized LOC105379290', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12985151, end_position_on_the_genomic_accession=13038442, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105379291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379291, CurrentID=0, Status='live', Symbol='LOC105379291', Aliases='', description='uncharacterized LOC105379291', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13462698, end_position_on_the_genomic_accession=13479139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652764 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652764, CurrentID=0, Status='live', Symbol='HPX-2', Aliases='', description='homeobox HPX-2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379293, CurrentID=0, Status='live', Symbol='LOC105379293', Aliases='', description='uncharacterized LOC105379293', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13986484, end_position_on_the_genomic_accession=13995932, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379294, CurrentID=0, Status='live', Symbol='LOC105379294', Aliases='', description='uncharacterized LOC105379294', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14723150, end_position_on_the_genomic_accession=14758738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379295, CurrentID=0, Status='live', Symbol='LOC105379295', Aliases='', description='uncharacterized LOC105379295', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=14404739, end_position_on_the_genomic_accession=14420797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379296, CurrentID=0, Status='live', Symbol='LOC105379296', Aliases='', description='uncharacterized LOC105379296', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=16619532, end_position_on_the_genomic_accession=16623828, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  128481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128481, CurrentID=0, Status='live', Symbol='RPL27AP', Aliases='dJ1028D15.2', description='ribosomal protein L27a pseudogene', other_designations='', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=43652508, end_position_on_the_genomic_accession=43653340, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379298, CurrentID=0, Status='live', Symbol='LOC105379298', Aliases='', description='uncharacterized LOC105379298', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=17126123, end_position_on_the_genomic_accession=17150130, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390627, CurrentID=0, Status='live', Symbol='CSPG4P10', Aliases='', description='chondroitin sulfate proteoglycan 4 pseudogene 10', other_designations='', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82459471, end_position_on_the_genomic_accession=82470008, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379300 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379300, CurrentID=0, Status='live', Symbol='LOC105379300', Aliases='', description='uncharacterized LOC105379300', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19245599, end_position_on_the_genomic_accession=19277381, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379301, CurrentID=0, Status='live', Symbol='LOC105379301', Aliases='', description='uncharacterized LOC105379301', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19091724, end_position_on_the_genomic_accession=19145182, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  128486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128486, CurrentID=0, Status='live', Symbol='FITM2', Aliases='C20orf142, Fit2, dJ881L22.2', description='fat storage-inducing transmembrane protein 2', other_designations='fat inducing transcript 2|fat-inducing protein 2|fat-inducing transcript 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=44306557, end_position_on_the_genomic_accession=44311249, orientation='minus', exon_count=2, OMIM=612029, no_hdr0=''),
  105379303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379303, CurrentID=0, Status='live', Symbol='LOC105379303', Aliases='', description='uncharacterized LOC105379303', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18989279, end_position_on_the_genomic_accession=19003259, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  128488 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128488, CurrentID=0, Status='live', Symbol='WFDC12', Aliases='C20orf122, SWAM2, WAP2, dJ211D12.4', description='WAP four-disulfide core domain 12', other_designations='protease inhibitor WAP2|putative protease inhibitor WAP12|single WAP motif protein 2|whey acidic protein 2', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45123426, end_position_on_the_genomic_accession=45124465, orientation='minus', exon_count=3, OMIM=609872, no_hdr0=''),
  105379305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379305, CurrentID=0, Status='live', Symbol='LOC105379305', Aliases='', description='uncharacterized LOC105379305', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=18608999, end_position_on_the_genomic_accession=18612069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390634, CurrentID=0, Status='live', Symbol='KRT18P47', Aliases='', description='keratin 18 pseudogene 47', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88959761, end_position_on_the_genomic_accession=88961111, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379307, CurrentID=0, Status='live', Symbol='LOC105379307', Aliases='', description='uncharacterized LOC105379307', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19640644, end_position_on_the_genomic_accession=19651126, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379308, CurrentID=0, Status='live', Symbol='LOC105379308', Aliases='', description='uncharacterized LOC105379308', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=19549697, end_position_on_the_genomic_accession=19551552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  390637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390637, CurrentID=0, Status='live', Symbol='GDPGP1', Aliases='C15orf58, VTC2', description='GDP-D-glucose phosphorylase 1', other_designations='GDP-D-glucose phosphorylase C15orf58', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90234255, end_position_on_the_genomic_accession=90242080, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390638 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390638, CurrentID=0, Status='live', Symbol='LOC390638', Aliases='', description='golgin A2 pseudogene', other_designations='', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=91031947, end_position_on_the_genomic_accession=91035295, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379311, CurrentID=0, Status='live', Symbol='LOC105379311', Aliases='', description='uncharacterized LOC105379311', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20078739, end_position_on_the_genomic_accession=20125001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379312 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379312, CurrentID=0, Status='live', Symbol='LOC105379312', Aliases='', description='uncharacterized LOC105379312', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20125425, end_position_on_the_genomic_accession=20133093, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  128497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128497, CurrentID=0, Status='live', Symbol='SPATA25', Aliases='C20orf165, TSG23, dJ337O18.8', description='spermatogenesis associated 25', other_designations='testis-specific gene 23 protein', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45886486, end_position_on_the_genomic_accession=45888653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379314, CurrentID=0, Status='live', Symbol='LOC105379314', Aliases='', description='uncharacterized LOC105379314', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20343384, end_position_on_the_genomic_accession=20370627, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  390643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390643, CurrentID=0, Status='live', Symbol='PGAM1P12', Aliases='', description='phosphoglycerate mutase 1 pseudogene 12', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=96441569, end_position_on_the_genomic_accession=96442313, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128500, CurrentID=0, Status='live', Symbol='RPL13P2', Aliases='RPL13P3, RPL13_7_1707, dJ599F21.2', description='ribosomal protein L13 pseudogene 2', other_designations='ribosomal protein L13 pseudogene 3', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46099475, end_position_on_the_genomic_accession=46100193, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379317 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379317, CurrentID=0, Status='live', Symbol='LOC105379317', Aliases='', description='uncharacterized LOC105379317', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21604991, end_position_on_the_genomic_accession=21618185, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379318 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379318, CurrentID=0, Status='live', Symbol='LOC105379318', Aliases='', description='uncharacterized LOC105379318', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21850241, end_position_on_the_genomic_accession=21861318, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379319, CurrentID=0, Status='live', Symbol='LOC105379319', Aliases='', description='uncharacterized LOC105379319', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22188429, end_position_on_the_genomic_accession=22191478, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390648, CurrentID=0, Status='live', Symbol='OR4F6', Aliases='OR15-15, OR4F12', description='olfactory receptor family 4 subfamily F member 6', other_designations='olfactory receptor 4F12|olfactory receptor OR15-15|olfactory receptor, family 4, subfamily F, member 12|olfactory receptor, family 4, subfamily F, member 6', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101805720, end_position_on_the_genomic_accession=101806658, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  390649 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390649, CurrentID=0, Status='live', Symbol='OR4F15', Aliases='', description='olfactory receptor family 4 subfamily F member 15', other_designations='olfactory receptor OR15-14|olfactory receptor, family 4, subfamily F, member 15', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101818187, end_position_on_the_genomic_accession=101819125, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  128506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128506, CurrentID=0, Status='live', Symbol='OCSTAMP', Aliases='C20orf123, OC-STAMP, dJ257E24.3', description='osteoclast stimulatory transmembrane protein', other_designations='transmembrane protein C20orf123', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=46541031, end_position_on_the_genomic_accession=46550574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390651 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390651, CurrentID=0, Status='live', Symbol='OR4F13P', Aliases='GPCRLTM7', description='olfactory receptor family 4 subfamily F member 13 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 13 pseudogene|putative olfactory receptor GPCRLTM7', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101842119, end_position_on_the_genomic_accession=101850324, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  390652 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390652, CurrentID=0, Status='live', Symbol='OR4F28P', Aliases='', description='olfactory receptor family 4 subfamily F member 28 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 28 pseudogene', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=101875964, end_position_on_the_genomic_accession=101876901, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379325 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379325, CurrentID=0, Status='live', Symbol='LOC105379325', Aliases='', description='uncharacterized LOC105379325', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=22811405, end_position_on_the_genomic_accession=22815386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379326 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379326, CurrentID=0, Status='live', Symbol='LOC105379326', Aliases='', description='uncharacterized LOC105379326', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23263110, end_position_on_the_genomic_accession=23264728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379327 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379327, CurrentID=0, Status='live', Symbol='LOC105379327', Aliases='', description='uncharacterized LOC105379327', other_designations='CTC-756D1.3', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=23492298, end_position_on_the_genomic_accession=23493891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379328 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379328, CurrentID=0, Status='live', Symbol='LOC105379328', Aliases='', description='uncharacterized LOC105379328', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24090100, end_position_on_the_genomic_accession=24105096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379329 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379329, CurrentID=0, Status='live', Symbol='LOC105379329', Aliases='', description='uncharacterized LOC105379329', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24884088, end_position_on_the_genomic_accession=24924631, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379330, CurrentID=0, Status='live', Symbol='LOC105379330', Aliases='', description='uncharacterized LOC105379330', other_designations='CTD-2168K21.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=24991877, end_position_on_the_genomic_accession=25009973, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379331, CurrentID=0, Status='live', Symbol='LOC105379331', Aliases='', description='uncharacterized LOC105379331', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25350604, end_position_on_the_genomic_accession=25361501, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  390660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390660, CurrentID=0, Status='live', Symbol='ADAMTS7P1', Aliases='ADAMTS7P2', description='ADAMTS7 pseudogene 1', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 7 pseudogene|ADAMTS7 pseudogene 2', map_location='15q25.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82293283, end_position_on_the_genomic_accession=82334561, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105379333 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379333, CurrentID=0, Status='live', Symbol='LOC105379333', Aliases='', description='uncharacterized LOC105379333', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25670608, end_position_on_the_genomic_accession=25685084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379334, CurrentID=0, Status='live', Symbol='LOC105379334', Aliases='', description='uncharacterized LOC105379334', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=25876481, end_position_on_the_genomic_accession=25879101, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379335, CurrentID=0, Status='live', Symbol='LOC105379335', Aliases='', description='uncharacterized LOC105379335', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26121735, end_position_on_the_genomic_accession=26127590, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  390664 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390664, CurrentID=0, Status='live', Symbol='C1QTNF8', Aliases='CTRP8, UNQ5829', description='C1q and tumor necrosis factor related protein 8', other_designations='C1q/TNF-related protein 8', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1088226, end_position_on_the_genomic_accession=1096306, orientation='minus', exon_count=5, OMIM=614147, no_hdr0=''),
  105379337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379337, CurrentID=0, Status='live', Symbol='LOC105379337', Aliases='', description='uncharacterized LOC105379337', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26440491, end_position_on_the_genomic_accession=26442195, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379338, CurrentID=0, Status='live', Symbol='LOC105379338', Aliases='', description='uncharacterized LOC105379338', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=26854444, end_position_on_the_genomic_accession=26858782, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390667, CurrentID=0, Status='live', Symbol='PTX4', Aliases='C16orf38', description='pentraxin 4', other_designations='neuronal pentraxin-like protein C16orf38|pentraxin 4, long', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1485939, end_position_on_the_genomic_accession=1488467, orientation='minus', exon_count=3, OMIM=613442, no_hdr0=''),
  105379340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379340, CurrentID=0, Status='live', Symbol='LOC105379340', Aliases='', description='uncharacterized LOC105379340', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27171491, end_position_on_the_genomic_accession=27211333, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379341, CurrentID=0, Status='live', Symbol='LOC105379341', Aliases='', description='uncharacterized LOC105379341', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27590111, end_position_on_the_genomic_accession=27598127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379342, CurrentID=0, Status='live', Symbol='LOC105379342', Aliases='', description='uncharacterized LOC105379342', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27902371, end_position_on_the_genomic_accession=27910289, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390671, CurrentID=0, Status='live', Symbol='RPL18P12', Aliases='RPL18_8_1468', description='ribosomal protein L18 pseudogene 12', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=3496167, end_position_on_the_genomic_accession=3496803, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379344 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379344, CurrentID=0, Status='live', Symbol='LOC105379344', Aliases='', description='uncharacterized LOC105379344', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28600475, end_position_on_the_genomic_accession=28621699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379345 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379345, CurrentID=0, Status='live', Symbol='LOC105379345', Aliases='', description='uncharacterized LOC105379345', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28622749, end_position_on_the_genomic_accession=28644408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379346, CurrentID=0, Status='live', Symbol='LOC105379346', Aliases='', description='uncharacterized LOC105379346', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29015033, end_position_on_the_genomic_accession=29037444, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379347 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379347, CurrentID=0, Status='live', Symbol='LOC105379347', Aliases='', description='uncharacterized LOC105379347', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=28763639, end_position_on_the_genomic_accession=28769882, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379348, CurrentID=0, Status='live', Symbol='LOC105379348', Aliases='', description='uncharacterized LOC105379348', other_designations='CTD-2647L4.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29124130, end_position_on_the_genomic_accession=29140526, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379349, CurrentID=0, Status='live', Symbol='LOC105379349', Aliases='', description='uncharacterized LOC105379349', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29368444, end_position_on_the_genomic_accession=29373349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379350, CurrentID=0, Status='live', Symbol='LOC105379350', Aliases='', description='uncharacterized LOC105379350', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29533702, end_position_on_the_genomic_accession=29544551, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379351 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379351, CurrentID=0, Status='live', Symbol='LOC105379351', Aliases='', description='uncharacterized LOC105379351', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29564027, end_position_on_the_genomic_accession=29583895, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379352, CurrentID=0, Status='live', Symbol='LOC105379352', Aliases='', description='uncharacterized LOC105379352', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29595496, end_position_on_the_genomic_accession=29602274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379353 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379353, CurrentID=0, Status='live', Symbol='LOC105379353', Aliases='', description='uncharacterized LOC105379353', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29667871, end_position_on_the_genomic_accession=29671288, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379354, CurrentID=0, Status='live', Symbol='LOC105379354', Aliases='', description='uncharacterized LOC105379354', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=29862367, end_position_on_the_genomic_accession=29882387, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379355, CurrentID=0, Status='live', Symbol='LOC105379355', Aliases='', description='uncharacterized LOC105379355', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30740783, end_position_on_the_genomic_accession=30743145, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379356, CurrentID=0, Status='live', Symbol='LOC105379356', Aliases='', description='uncharacterized LOC105379356', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30732477, end_position_on_the_genomic_accession=30734324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379357 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379357, CurrentID=0, Status='live', Symbol='LOC105379357', Aliases='', description='uncharacterized LOC105379357', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30888775, end_position_on_the_genomic_accession=30901672, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379358, CurrentID=0, Status='live', Symbol='LOC105379358', Aliases='', description='uncharacterized LOC105379358', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31176320, end_position_on_the_genomic_accession=31179268, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379359 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379359, CurrentID=0, Status='live', Symbol='LOC105379359', Aliases='', description='uncharacterized LOC105379359', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=31143630, end_position_on_the_genomic_accession=31177479, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  390688 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390688, CurrentID=0, Status='live', Symbol='CDC37P1', Aliases='A-761H5.2', description='cell division cycle 37 pseudogene 1', other_designations='Putative CDC37-like protein LOC390688|cell division cycle 37 homolog pseudogene 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=28700176, end_position_on_the_genomic_accession=28701611, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379361, CurrentID=0, Status='live', Symbol='LOC105379361', Aliases='', description='uncharacterized LOC105379361', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=32428771, end_position_on_the_genomic_accession=32429877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379362 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379362, CurrentID=0, Status='live', Symbol='LOC105379362', Aliases='', description='uncharacterized LOC105379362', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=32927945, end_position_on_the_genomic_accession=33048479, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379363, CurrentID=0, Status='live', Symbol='LOC105379363', Aliases='', description='uncharacterized LOC105379363', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33604878, end_position_on_the_genomic_accession=33609611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379364, CurrentID=0, Status='live', Symbol='LOC105379364', Aliases='', description='uncharacterized LOC105379364', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=33755809, end_position_on_the_genomic_accession=34039059, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379365, CurrentID=0, Status='live', Symbol='LOC105379365', Aliases='', description='uncharacterized LOC105379365', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34174806, end_position_on_the_genomic_accession=34227631, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379366 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379366, CurrentID=0, Status='live', Symbol='LOC105379366', Aliases='', description='uncharacterized LOC105379366', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34135683, end_position_on_the_genomic_accession=34175233, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105379367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379367, CurrentID=0, Status='live', Symbol='LOC105379367', Aliases='', description='uncharacterized LOC105379367', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34858652, end_position_on_the_genomic_accession=34874675, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379368, CurrentID=0, Status='live', Symbol='LOC105379368', Aliases='', description='uncharacterized LOC105379368', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=34967483, end_position_on_the_genomic_accession=34974950, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  128553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128553, CurrentID=0, Status='live', Symbol='TSHZ2', Aliases='C20orf17, OVC10-2, TSH2, ZABC2, ZNF218', description='teashirt zinc finger homeobox 2', other_designations='cell growth-inhibiting protein 7|ovarian cancer-related protein 10-2|serologically defined colon cancer antigen 33 like|teashirt family zinc finger 2|zinc finger protein 218', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=52972309, end_position_on_the_genomic_accession=53495330, orientation='plus', exon_count=4, OMIM=614118, no_hdr0=''),
  105379370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379370, CurrentID=0, Status='live', Symbol='LOC105379370', Aliases='', description='uncharacterized LOC105379370', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36191524, end_position_on_the_genomic_accession=36192942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379371, CurrentID=0, Status='live', Symbol='LOC105379371', Aliases='', description='uncharacterized LOC105379371', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36195043, end_position_on_the_genomic_accession=36266992, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379372 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379372, CurrentID=0, Status='live', Symbol='LOC105379372', Aliases='', description='uncharacterized LOC105379372', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36323573, end_position_on_the_genomic_accession=36326636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379373 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379373, CurrentID=0, Status='live', Symbol='LOC105379373', Aliases='', description='uncharacterized LOC105379373', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36537393, end_position_on_the_genomic_accession=36758831, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379374, CurrentID=0, Status='live', Symbol='LOC105379374', Aliases='', description='uncharacterized LOC105379374', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36765194, end_position_on_the_genomic_accession=36779164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379375 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379375, CurrentID=0, Status='live', Symbol='LOC105379375', Aliases='', description='uncharacterized LOC105379375', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36883170, end_position_on_the_genomic_accession=36925165, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379376, CurrentID=0, Status='live', Symbol='LOC105379376', Aliases='', description='uncharacterized LOC105379376', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=36963764, end_position_on_the_genomic_accession=36982005, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390705 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390705, CurrentID=0, Status='live', Symbol='LOC390705', Aliases='', description="protein phosphatase 2 regulatory subunit B'', beta pseudogene", other_designations="protein phosphatase 2, regulatory subunit B'', beta pseudogene", map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32289547, end_position_on_the_genomic_accession=32289981, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379378, CurrentID=0, Status='live', Symbol='LOC105379378', Aliases='', description='uncharacterized LOC105379378', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37581515, end_position_on_the_genomic_accession=37582959, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379379 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379379, CurrentID=0, Status='live', Symbol='LOC105379379', Aliases='', description='uncharacterized LOC105379379', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37559915, end_position_on_the_genomic_accession=37565633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379380, CurrentID=0, Status='live', Symbol='LOC105379380', Aliases='', description='uncharacterized LOC105379380', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37602804, end_position_on_the_genomic_accession=37658398, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379381, CurrentID=0, Status='live', Symbol='LOC105379381', Aliases='', description='uncharacterized LOC105379381', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=37790370, end_position_on_the_genomic_accession=37794545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379382, CurrentID=0, Status='live', Symbol='LOC105379382', Aliases='', description='uncharacterized LOC105379382', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38156220, end_position_on_the_genomic_accession=38163494, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379383, CurrentID=0, Status='live', Symbol='LOC105379383', Aliases='', description='uncharacterized LOC105379383', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38552075, end_position_on_the_genomic_accession=38559255, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379384 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379384, CurrentID=0, Status='live', Symbol='LOC105379384', Aliases='', description='uncharacterized LOC105379384', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=38565066, end_position_on_the_genomic_accession=38578485, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379385, CurrentID=0, Status='live', Symbol='LOC105379385', Aliases='', description='uncharacterized LOC105379385', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40104110, end_position_on_the_genomic_accession=40173333, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105379386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379386, CurrentID=0, Status='live', Symbol='LOC105379386', Aliases='', description='uncharacterized LOC105379386', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40114693, end_position_on_the_genomic_accession=40127118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379387, CurrentID=0, Status='live', Symbol='LOC105379387', Aliases='', description='uncharacterized LOC105379387', other_designations='CTA-392C11.1', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40333173, end_position_on_the_genomic_accession=40353117, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379388, CurrentID=0, Status='live', Symbol='LOC105379388', Aliases='', description='uncharacterized LOC105379388', other_designations='CTA-392C11.2', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40369989, end_position_on_the_genomic_accession=40392796, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379389 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379389, CurrentID=0, Status='live', Symbol='LOC105379389', Aliases='', description='uncharacterized LOC105379389', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40904039, end_position_on_the_genomic_accession=40928831, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379390, CurrentID=0, Status='live', Symbol='LOC105379390', Aliases='', description='uncharacterized LOC105379390', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=40929835, end_position_on_the_genomic_accession=40933923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379391, CurrentID=0, Status='live', Symbol='LOC105379391', Aliases='', description='uncharacterized LOC105379391', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41435298, end_position_on_the_genomic_accession=41440419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379392, CurrentID=0, Status='live', Symbol='LOC105379392', Aliases='', description='uncharacterized LOC105379392', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41797795, end_position_on_the_genomic_accession=41801977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379393 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379393, CurrentID=0, Status='live', Symbol='LOC105379393', Aliases='', description='uncharacterized LOC105379393', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42051904, end_position_on_the_genomic_accession=42104989, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379394 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379394, CurrentID=0, Status='live', Symbol='LOC105379394', Aliases='', description='uncharacterized LOC105379394', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42016987, end_position_on_the_genomic_accession=42039641, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379395 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379395, CurrentID=0, Status='live', Symbol='LOC105379395', Aliases='', description='uncharacterized LOC105379395', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=42297848, end_position_on_the_genomic_accession=42301357, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128580, CurrentID=0, Status='live', Symbol='RPS4XP3', Aliases='RPS4L, RPS4P3, dJ843L14.2', description='ribosomal protein S4X pseudogene 3', other_designations='ribosomal protein S4 pseudogene 3', map_location='20q13.31', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=56601666, end_position_on_the_genomic_accession=56602059, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379397 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379397, CurrentID=0, Status='live', Symbol='LOC105379397', Aliases='', description='uncharacterized LOC105379397', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=43486726, end_position_on_the_genomic_accession=43510775, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379398 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379398, CurrentID=0, Status='live', Symbol='LOC105379398', Aliases='', description='uncharacterized LOC105379398', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=20405485, end_position_on_the_genomic_accession=20413289, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379399, CurrentID=0, Status='live', Symbol='LOC105379399', Aliases='', description='uncharacterized LOC105379399', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=21227720, end_position_on_the_genomic_accession=21297077, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379400, CurrentID=0, Status='live', Symbol='LOC105379400', Aliases='', description='uncharacterized LOC105379400', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  652873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652873, CurrentID=0, Status='live', Symbol='IGKV2OR2-7', Aliases='IGKV2/OR2-7, IGKV2OR27', description='immunoglobulin kappa variable 2/OR2-7 (pseudogene)', other_designations='immunoglobulin kappa variable 2/OR2-7 pseudogene', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97372095, end_position_on_the_genomic_accession=97372999, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390730, CurrentID=0, Status='live', Symbol='PHBP21', Aliases='', description='prohibitin pseudogene 21', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=52934812, end_position_on_the_genomic_accession=52935634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283575, CurrentID=0, Status='live', Symbol='LOC283575', Aliases='', description='uncharacterized LOC283575', other_designations='', map_location='14q24.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=77041049, end_position_on_the_genomic_accession=77069503, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  390732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390732, CurrentID=0, Status='live', Symbol='CES1P2', Aliases='', description='carboxylesterase 1 pseudogene 2', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=55724925, end_position_on_the_genomic_accession=55750211, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  105379405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379405, CurrentID=0, Status='live', Symbol='LOC105379405', Aliases='', description='uncharacterized LOC105379405', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=66483128, end_position_on_the_genomic_accession=66483820, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379406, CurrentID=0, Status='live', Symbol='LOC105379406', Aliases='', description='uncharacterized LOC105379406', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=67779601, end_position_on_the_genomic_accession=67784428, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  390735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390735, CurrentID=0, Status='live', Symbol='RPS15AP34', Aliases='RPS15A_22_1489', description='ribosomal protein S15a pseudogene 34', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=63350759, end_position_on_the_genomic_accession=63351148, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379408 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379408, CurrentID=0, Status='live', Symbol='LOC105379408', Aliases='', description='uncharacterized LOC105379408', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379409, CurrentID=0, Status='live', Symbol='LOC105379409', Aliases='', description='uncharacterized LOC105379409', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=71775845, end_position_on_the_genomic_accession=71786680, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379410 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379410, CurrentID=0, Status='live', Symbol='LOC105379410', Aliases='', description='uncharacterized LOC105379410', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=86580011, end_position_on_the_genomic_accession=86591195, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379411, CurrentID=0, Status='live', Symbol='LOC105379411', Aliases='', description='uncharacterized LOC105379411', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=73580220, end_position_on_the_genomic_accession=73648096, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379412, CurrentID=0, Status='live', Symbol='LOC105379412', Aliases='', description='uncharacterized LOC105379412', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379413 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379413, CurrentID=0, Status='live', Symbol='LOC105379413', Aliases='', description='uncharacterized LOC105379413', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1744535, end_position_on_the_genomic_accession=1776605, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379414, CurrentID=0, Status='live', Symbol='LOC105379414', Aliases='', description='uncharacterized LOC105379414', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1744771, end_position_on_the_genomic_accession=1745632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379415, CurrentID=0, Status='live', Symbol='LOC105379415', Aliases='', description='uncharacterized LOC105379415', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1732556, end_position_on_the_genomic_accession=1744502, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379416, CurrentID=0, Status='live', Symbol='LOC105379416', Aliases='', description='serine/arginine repetitive matrix protein 1-like', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=1392008, end_position_on_the_genomic_accession=1395560, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379417, CurrentID=0, Status='live', Symbol='LOC105379417', Aliases='', description='ankyrin repeat domain-containing protein 20A2-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128602, CurrentID=0, Status='live', Symbol='C20orf85', Aliases='LLC1, bA196N14.1', description='chromosome 20 open reading frame 85', other_designations='low in lung cancer 1', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=58150927, end_position_on_the_genomic_accession=58161130, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379419 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379419, CurrentID=0, Status='live', Symbol='LOC105379419', Aliases='', description='Ig kappa chain V-I region Walker-like', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10685662, end_position_on_the_genomic_accession=10685987, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390748, CurrentID=0, Status='live', Symbol='PABPN1L', Aliases='PABPNL1, ePABP2', description='poly(A) binding protein, nuclear 1-like (cytoplasmic)', other_designations='ePABP-2|embryonic poly(A) binding protein 2|embryonic poly(A)-binding protein 2|embryonic poly(A)-binding protein type II|poly(A)-binding protein nuclear-like 1', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=88862623, end_position_on_the_genomic_accession=88870156, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105379421 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379421, CurrentID=0, Status='live', Symbol='LOC105379421', Aliases='', description='collagen alpha-1(III) chain-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379422, CurrentID=0, Status='live', Symbol='LOC105379422', Aliases='', description='octapeptide-repeat protein T2-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379423, CurrentID=0, Status='live', Symbol='LOC105379423', Aliases='', description='putative uncharacterized protein DDB_G0271982', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379424, CurrentID=0, Status='live', Symbol='LOC105379424', Aliases='', description='uncharacterized LOC105379424', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128609 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128609, CurrentID=0, Status='live', Symbol='MRPS16P2', Aliases='MRPS16P, bA379F14.1', description='mitochondrial ribosomal protein S16 pseudogene 2', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59113075, end_position_on_the_genomic_accession=59113932, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379426, CurrentID=0, Status='live', Symbol='LOC105379426', Aliases='', description='uncharacterized LOC105379426', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128611, CurrentID=0, Status='live', Symbol='ZNF831', Aliases='C20orf174, dJ492J12.1', description='zinc finger protein 831', other_designations='', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59123428, end_position_on_the_genomic_accession=59259113, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  390756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390756, CurrentID=0, Status='live', Symbol='OR3A4P', Aliases='OLFRA05, OLFRA06, OR17-13, OR17-16, OR17-24, OR17-25, OR24, OR25, OR3A4, OR3A5P', description='olfactory receptor family 3 subfamily A member 4 pseudogene', other_designations='olfactory receptor OR17-13 pseudogene|olfactory receptor OR17-16 pseudogene|olfactory receptor, family 3, subfamily A, member 4 pseudogene|olfactory receptor, family 3, subfamily A, member 5 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3310245, end_position_on_the_genomic_accession=3311446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379429, CurrentID=0, Status='live', Symbol='LOC105379429', Aliases='', description='uncharacterized LOC105379429', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=11709501, end_position_on_the_genomic_accession=11716715, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379430 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379430, CurrentID=0, Status='live', Symbol='LOC105379430', Aliases='', description='uncharacterized LOC105379430', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=10846697, end_position_on_the_genomic_accession=10848407, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128615, CurrentID=0, Status='live', Symbol='PIEZO1P1', Aliases='C20orf83, FAM38CP, bA164D18.1', description='piezo-type mechanosensitive ion channel component 1 pseudogene 1', other_designations='family with sequence similarity 38, member C, pseudogene', map_location='20q13.32', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=59460619, end_position_on_the_genomic_accession=59460837, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652904, CurrentID=0, Status='live', Symbol='CKS1BP6', Aliases='', description='CDC28 protein kinase regulatory subunit 1B pseudogene 6', other_designations='', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30617368, end_position_on_the_genomic_accession=30618175, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379433, CurrentID=0, Status='live', Symbol='LOC105379433', Aliases='', description='immunoglobulin superfamily member 2-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=89661730, end_position_on_the_genomic_accession=89683251, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105376896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376896, CurrentID=0, Status='live', Symbol='LOC105376896', Aliases='', description='uncharacterized LOC105376896', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28239511, end_position_on_the_genomic_accession=28247324, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379435, CurrentID=0, Status='live', Symbol='LOC105379435', Aliases='', description='uncharacterized LOC105379435', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61296748, end_position_on_the_genomic_accession=61297691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379436, CurrentID=0, Status='live', Symbol='LOC105379436', Aliases='', description='uncharacterized LOC105379436', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61600871, end_position_on_the_genomic_accession=61609320, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379437, CurrentID=0, Status='live', Symbol='LOC105379437', Aliases='', description='uncharacterized LOC105379437', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=61549134, end_position_on_the_genomic_accession=61555128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  390766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390766, CurrentID=0, Status='live', Symbol='RNASEH1P2', Aliases='', description='ribonuclease H1 pseudogene 2', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=16683402, end_position_on_the_genomic_accession=16684230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379439, CurrentID=0, Status='live', Symbol='LOC105379439', Aliases='', description='uncharacterized LOC105379439', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62125621, end_position_on_the_genomic_accession=62129209, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379440, CurrentID=0, Status='live', Symbol='LOC105379440', Aliases='', description='uncharacterized LOC105379440', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62877498, end_position_on_the_genomic_accession=62880628, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379441, CurrentID=0, Status='live', Symbol='LOC105379441', Aliases='', description='uncharacterized LOC105379441', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63254475, end_position_on_the_genomic_accession=63270221, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379442, CurrentID=0, Status='live', Symbol='LOC105379442', Aliases='', description='uncharacterized LOC105379442', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63408510, end_position_on_the_genomic_accession=63422122, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379443 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379443, CurrentID=0, Status='live', Symbol='LOC105379443', Aliases='', description='monofunctional C1-tetrahydrofolate synthase, mitochondrial-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65143073, end_position_on_the_genomic_accession=65203646, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105379444 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379444, CurrentID=0, Status='live', Symbol='LOC105379444', Aliases='', description='uncharacterized LOC105379444', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65383315, end_position_on_the_genomic_accession=65384351, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379445 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379445, CurrentID=0, Status='live', Symbol='LOC105379445', Aliases='', description='uncharacterized LOC105379445', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65400319, end_position_on_the_genomic_accession=65403841, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105379446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379446, CurrentID=0, Status='live', Symbol='LOC105379446', Aliases='', description='uncharacterized LOC105379446', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65445307, end_position_on_the_genomic_accession=65454950, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379447, CurrentID=0, Status='live', Symbol='LOC105379447', Aliases='', description='uncharacterized LOC105379447', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65546931, end_position_on_the_genomic_accession=65551631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379448, CurrentID=0, Status='live', Symbol='LOC105379448', Aliases='', description='uncharacterized LOC105379448', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66593659, end_position_on_the_genomic_accession=66599035, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379449 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379449, CurrentID=0, Status='live', Symbol='LOC105379449', Aliases='', description='uncharacterized LOC105379449', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66695989, end_position_on_the_genomic_accession=66702329, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379450, CurrentID=0, Status='live', Symbol='LOC105379450', Aliases='', description='uncharacterized LOC105379450', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67006925, end_position_on_the_genomic_accession=67009079, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379451, CurrentID=0, Status='live', Symbol='LOC105379451', Aliases='', description='uncharacterized LOC105379451', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67459183, end_position_on_the_genomic_accession=67462788, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379452, CurrentID=0, Status='live', Symbol='LOC105379452', Aliases='', description='uncharacterized LOC105379452', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67492450, end_position_on_the_genomic_accession=67515381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  128637 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128637, CurrentID=0, Status='live', Symbol='TBC1D20', Aliases='C20orf140, WARBM4', description='TBC1 domain family member 20', other_designations='TBC1 domain family, member 20', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=435477, end_position_on_the_genomic_accession=462553, orientation='minus', exon_count=10, OMIM=611663, no_hdr0=''),
  105379454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379454, CurrentID=0, Status='live', Symbol='LOC105379454', Aliases='', description='uncharacterized LOC105379454', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=67651435, end_position_on_the_genomic_accession=67668172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379455 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379455, CurrentID=0, Status='live', Symbol='LOC105379455', Aliases='', description='uncharacterized LOC105379455', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379456, CurrentID=0, Status='live', Symbol='LOC105379456', Aliases='', description='uncharacterized LOC105379456', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379457 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379457, CurrentID=0, Status='live', Symbol='LOC105379457', Aliases='', description='uncharacterized LOC105379457', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379458, CurrentID=0, Status='live', Symbol='LOC105379458', Aliases='', description='uncharacterized LOC105379458', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379459, CurrentID=0, Status='live', Symbol='LOC105379459', Aliases='', description='uncharacterized LOC105379459', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390788, CurrentID=0, Status='live', Symbol='CCL3P1', Aliases='CCL3L2, G0S19-3, LD78gamma, SCYA3L2', description='chemokine (C-C motif) ligand 3 pseudogene 1', other_designations='LD78gamma pseudogene|chemokine (C-C motif) ligand 3-like 2|small inducible cytokine A3-like 2', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=609467, no_hdr0=''),
  105379461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379461, CurrentID=0, Status='live', Symbol='LOC105379461', Aliases='', description='uncharacterized LOC105379461', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20674418, end_position_on_the_genomic_accession=20718430, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390790, CurrentID=0, Status='live', Symbol='ARL5C', Aliases='ARL12', description='ADP ribosylation factor like GTPase 5C', other_designations='ADP-ribosylation factor like GTPase 5C|ADP-ribosylation factor-like 12|ADP-ribosylation factor-like 5C|ADP-ribosylation factor-like protein 12', map_location='17q12', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39156894, end_position_on_the_genomic_accession=39166161, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  105379463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379463, CurrentID=0, Status='live', Symbol='LOC105379463', Aliases='', description='uncharacterized LOC105379463', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390792 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390792, CurrentID=0, Status='live', Symbol='KRT39', Aliases='CK-39, K39, KA35', description='keratin 39, type I', other_designations='cytokeratin-39|keratin-39|type I hair keratin KA35', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=40958417, end_position_on_the_genomic_accession=40966892, orientation='minus', exon_count=7, OMIM=616678, no_hdr0=''),
  105379465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379465, CurrentID=0, Status='live', Symbol='LOC105379465', Aliases='', description='uncharacterized LOC105379465', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32734591, end_position_on_the_genomic_accession=32741112, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379466 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379466, CurrentID=0, Status='live', Symbol='LOC105379466', Aliases='', description='uncharacterized LOC105379466', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32235135, end_position_on_the_genomic_accession=32235834, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379467, CurrentID=0, Status='live', Symbol='LOC105379467', Aliases='', description='uncharacterized LOC105379467', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=32188321, end_position_on_the_genomic_accession=32194848, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379468 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379468, CurrentID=0, Status='live', Symbol='LOC105379468', Aliases='', description='uncharacterized LOC105379468', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33203844, end_position_on_the_genomic_accession=33207348, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  128653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128653, CurrentID=0, Status='live', Symbol='C20orf141', Aliases='dJ860F19.4', description='chromosome 20 open reading frame 141', other_designations='', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=2814987, end_position_on_the_genomic_accession=2815830, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379470, CurrentID=0, Status='live', Symbol='LOC105379470', Aliases='', description='uncharacterized LOC105379470', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33469234, end_position_on_the_genomic_accession=33472738, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105379471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379471, CurrentID=0, Status='live', Symbol='LOC105379471', Aliases='', description='putative uncharacterized protein encoded by LINC00612', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=20524287, end_position_on_the_genomic_accession=20531839, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390800, CurrentID=0, Status='live', Symbol='RPL7L1P5', Aliases='', description='ribosomal protein L7 like 1 pseudogene 5', other_designations='ribosomal protein L7-like 1 pseudogene 5', map_location='17q21.31', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=44365017, end_position_on_the_genomic_accession=44365751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379473 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379473, CurrentID=0, Status='live', Symbol='LOC105379473', Aliases='', description='uncharacterized LOC105379473', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33385983, end_position_on_the_genomic_accession=33387728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379474 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379474, CurrentID=0, Status='live', Symbol='LOC105379474', Aliases='', description='uncharacterized LOC105379474', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33484765, end_position_on_the_genomic_accession=33486510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422595, CurrentID=0, Status='live', Symbol='PCMTD1P3', Aliases='', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 1 pseudogene 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=23721310, end_position_on_the_genomic_accession=23721810, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379476 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379476, CurrentID=0, Status='live', Symbol='LOC105379476', Aliases='', description='uncharacterized LOC105379476', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29401775, end_position_on_the_genomic_accession=29408058, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379477, CurrentID=0, Status='live', Symbol='LOC105379477', Aliases='', description='uncharacterized LOC105379477', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29454923, end_position_on_the_genomic_accession=29476568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379478 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379478, CurrentID=0, Status='live', Symbol='LINC01598', Aliases='', description='long intergenic non-protein coding RNA 1598', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29548674, end_position_on_the_genomic_accession=29559760, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379479, CurrentID=0, Status='live', Symbol='DUX4L34', Aliases='', description='double homeobox 4 like 34, pseudogene', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29410284, end_position_on_the_genomic_accession=29411803, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379480, CurrentID=0, Status='live', Symbol='LOC105379480', Aliases='', description='putative uncharacterized protein SPANXA2-OT1', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=29748947, end_position_on_the_genomic_accession=29754673, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379481 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379481, CurrentID=0, Status='live', Symbol='LOC105379481', Aliases='', description='uncharacterized LOC105379481', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30514052, end_position_on_the_genomic_accession=30531183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379482 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379482, CurrentID=0, Status='live', Symbol='LOC105379482', Aliases='', description='uncharacterized LOC105379482', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=30585339, end_position_on_the_genomic_accession=30587476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390811, CurrentID=0, Status='live', Symbol='POLR3KP2', Aliases='', description='polymerase (RNA) III (DNA directed) polypeptide K, 12.3 kDa pseudogene 2', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73162770, end_position_on_the_genomic_accession=73163384, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128668 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128668, CurrentID=0, Status='live', Symbol='RPL7AP12', Aliases='RPL7AL2, dJ189G13.1', description='ribosomal protein L7a pseudogene 12', other_designations='ribosomal protein L7a like 2|ribosomal protein L7a-like 2', map_location='20p13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=4590935, end_position_on_the_genomic_accession=4591321, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379768, CurrentID=0, Status='live', Symbol='LOC105379768', Aliases='', description='uncharacterized LOC105379768', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379486, CurrentID=0, Status='live', Symbol='LOC105379486', Aliases='', description='uncharacterized LOC105379486', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=5093473, end_position_on_the_genomic_accession=5099713, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  652959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652959, CurrentID=0, Status='live', Symbol='BTF3P5', Aliases='', description='basic transcription factor 3 pseudogene 5', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55434896, end_position_on_the_genomic_accession=55435719, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390816, CurrentID=0, Status='live', Symbol='THA1P', Aliases='GLY1', description='threonine aldolase 1, pseudogene', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=78248590, end_position_on_the_genomic_accession=78256185, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652961, CurrentID=0, Status='live', Symbol='BTF3P8', Aliases='', description='basic transcription factor 3 pseudogene 8', other_designations='', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=63766759, end_position_on_the_genomic_accession=63767588, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128674 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128674, CurrentID=0, Status='live', Symbol='PROKR2', Aliases='GPR73L1, GPR73b, GPRg2, HH3, KAL3, PKR2, dJ680N4.3', description='prokineticin receptor 2', other_designations='G protein-coupled receptor 73-like 1|G-protein coupled receptor I5E|PK-R2', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5299874, end_position_on_the_genomic_accession=5317547, orientation='minus', exon_count=3, OMIM=607123, no_hdr0=''),
  652963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652963, CurrentID=0, Status='live', Symbol='BTF3P12', Aliases='BTF3L2, HUMBTFC', description='basic transcription factor 3 pseudogene 12', other_designations='basic transcription factor 3, like 2|basic transcription factor 3, pseudogene|basic transcription factor 3-like 2, pseudogene|lambda h23', map_location='8q13.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=70273263, end_position_on_the_genomic_accession=70274168, orientation='plus', exon_count=0, OMIM=603738, no_hdr0=''),
  128676 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128676, CurrentID=0, Status='live', Symbol='RPS18P1', Aliases='RPS18_9_1682, dJ1022P6.1', description='ribosomal protein S18 pseudogene 1', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5532878, end_position_on_the_genomic_accession=5533335, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  652965 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652965, CurrentID=0, Status='live', Symbol='SNORA48', Aliases='ACA48', description='small nucleolar RNA, H/ACA box 48', other_designations='ACA48 snoRNA', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7574713, end_position_on_the_genomic_accession=7574847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  652966 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652966, CurrentID=0, Status='live', Symbol='SNORD10', Aliases='mgU6-77', description='small nucleolar RNA, C/D box 10', other_designations='mgU6-77 snoRNA', map_location='17p13.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=7576811, end_position_on_the_genomic_accession=7576958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379495 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379495, CurrentID=0, Status='live', Symbol='LOC105379495', Aliases='', description='uncharacterized LOC105379495', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6574582, end_position_on_the_genomic_accession=6576681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652968 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652968, CurrentID=0, Status='live', Symbol='GATSL3', Aliases='', description='GATS protein-like 3', other_designations='', map_location='22q12', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=30285118, end_position_on_the_genomic_accession=30289627, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  652969 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652969, CurrentID=0, Status='live', Symbol='SLC6A21P', Aliases='SLC6A21', description='solute carrier family 6 member 21, pseudogene', other_designations='solute carrier family 6 (neurotransmitter transporter), member 21 (pseudogene)|solute carrier family 6 (neurotransmitter transporter), member 21, pseudogene|solute carrier family 6 (neutral amino acid transporter), member 15 pseudogene|solute carrier family 6, member 21 pseudogene|solute carrier family 6, member 21, pseudogene', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49252333, end_position_on_the_genomic_accession=49256340, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379498 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379498, CurrentID=0, Status='live', Symbol='LOC105379498', Aliases='', description='uncharacterized LOC105379498', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6721497, end_position_on_the_genomic_accession=6725231, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379499, CurrentID=0, Status='live', Symbol='LOC105379499', Aliases='CH507-145C22.1, CH507-154B10.1, CH507-154B10.2, CH507-24F1.1', description='uncharacterized LOC105379499', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=6986189, end_position_on_the_genomic_accession=6997775, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  652972 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652972, CurrentID=0, Status='live', Symbol='LRRC37A5P', Aliases='C9orf29', description='leucine-rich repeat containing 37 member A5, pseudogene', other_designations='leucine rich repeat containing 37, member A5, pseudogene', map_location='9q31.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=111602831, end_position_on_the_genomic_accession=111613553, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105379501 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379501, CurrentID=0, Status='live', Symbol='LOC105379501', Aliases='', description='uncharacterized LOC105379501', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7040152, end_position_on_the_genomic_accession=7042009, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  652974 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652974, CurrentID=0, Status='live', Symbol='RPL17P2', Aliases='RPL17_13_1376', description='ribosomal protein L17 pseudogene 2', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60212453, end_position_on_the_genomic_accession=60212954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390831, CurrentID=0, Status='live', Symbol='PMM2P1', Aliases='', description='phosphomannomutase 2 pseudogene 1', other_designations='', map_location='18p11.22', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=10636670, end_position_on_the_genomic_accession=10637401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379504 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379504, CurrentID=0, Status='live', Symbol='LOC105379504', Aliases='', description='uncharacterized LOC105379504', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=7777354, end_position_on_the_genomic_accession=7788049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100422600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422600, CurrentID=0, Status='live', Symbol='SDHCP1', Aliases='', description='succinate dehydrogenase complex subunit C pseudogene 1', other_designations='succinate dehydrogenase complex, subunit C pseudogene 1', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=68866491, end_position_on_the_genomic_accession=68868158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  652978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652978, CurrentID=0, Status='live', Symbol='RPL7AP2', Aliases='RPL7A_26_1357', description='ribosomal protein L7a pseudogene 2', other_designations='', map_location='14q21.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=39157122, end_position_on_the_genomic_accession=39157916, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379507, CurrentID=0, Status='live', Symbol='LOC105379507', Aliases='', description='uncharacterized LOC105379507', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8449461, end_position_on_the_genomic_accession=8450648, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652980, CurrentID=0, Status='live', Symbol='RPS28P1', Aliases='', description='ribosomal protein S28 pseudogene 1', other_designations='', map_location='14q23.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=63808435, end_position_on_the_genomic_accession=63809316, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379509, CurrentID=0, Status='live', Symbol='LOC105379509', Aliases='', description='fibril-forming collagen alpha chain-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8388342, end_position_on_the_genomic_accession=8404720, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379510, CurrentID=0, Status='live', Symbol='LOC105379510', Aliases='', description='collagen alpha-1(II) chain-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=8433534, end_position_on_the_genomic_accession=8448927, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652983 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652983, CurrentID=0, Status='live', Symbol='RPL35AP2', Aliases='RPL35A_10_973, bA571F15.2', description='ribosomal protein L35a pseudogene 2', other_designations='', map_location='9p', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34191493, end_position_on_the_genomic_accession=34191909, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379512 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379512, CurrentID=0, Status='live', Symbol='LOC105379512', Aliases='', description='uncharacterized LOC105379512', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9634286, end_position_on_the_genomic_accession=9648103, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  652985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652985, CurrentID=0, Status='live', Symbol='RPL7AP4', Aliases='RPL7A_27_1369', description='ribosomal protein L7a pseudogene 4', other_designations='', map_location='14q22.3', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=55543376, end_position_on_the_genomic_accession=55544231, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379514 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379514, CurrentID=0, Status='live', Symbol='LOC105379514', Aliases='', description='uncharacterized LOC105379514', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=9661337, end_position_on_the_genomic_accession=9681385, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  390843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390843, CurrentID=0, Status='live', Symbol='EIF4A3P1', Aliases='', description='eukaryotic translation initiation factor 4A3 pseudogene 1', other_designations='EIF4A3 pseudogene 1', map_location='18q11.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=24522074, end_position_on_the_genomic_accession=24522908, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390844 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390844, CurrentID=0, Status='live', Symbol='ARIH2P1', Aliases='ARIH2P', description='ariadne RBR E3 ubiquitin protein ligase 2 pseudogene 1', other_designations='ariadne homolog 2 pseudogene 1', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=28651530, end_position_on_the_genomic_accession=28653411, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379517, CurrentID=0, Status='live', Symbol='LOC105379517', Aliases='', description='uncharacterized LOC105379517', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390846 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390846, CurrentID=0, Status='live', Symbol='LOC390846', Aliases='', description='golgi SNAP receptor complex member 2 pseudogene', other_designations='hCG1642947', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=31726010, end_position_on_the_genomic_accession=31726396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652991 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652991, CurrentID=0, Status='live', Symbol='SKOR2', Aliases='CH18515, CORL2, FUSSEL18', description='SKI family transcriptional corepressor 2', other_designations='LBX1 corepressor 1-like protein|functional Smad-suppressing element on chromosome 18|functional smad suppressing element 18|fussel-18|ladybird homeobox corepressor 1-like protein', map_location='18q21.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=47206203, end_position_on_the_genomic_accession=47249630, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  390848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390848, CurrentID=0, Status='live', Symbol='RPS15AP35', Aliases='RPS15A_23_1592', description='ribosomal protein S15a pseudogene 35', other_designations='', map_location='18q12.1', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=32075846, end_position_on_the_genomic_accession=32076302, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  652993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652993, CurrentID=0, Status='live', Symbol='LOC652993', Aliases='', description='uncharacterized LOC652993', other_designations='', map_location='12q21', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379522, CurrentID=0, Status='live', Symbol='LOC105379522', Aliases='', description='zinc finger protein 717-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  652995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=652995, CurrentID=0, Status='live', Symbol='UCA1', Aliases='CUDR, LINC00178, NCRNA00178, UCAT1, onco-lncRNA-36', description='urothelial cancer associated 1 (non-protein coding)', other_designations='cancer up-regulated drug resistant|long intergenic non-protein coding RNA 178', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=15828947, end_position_on_the_genomic_accession=15836321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105379524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379524, CurrentID=0, Status='live', Symbol='LOC105379524', Aliases='', description='uncharacterized LOC105379524', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128709, CurrentID=0, Status='live', Symbol='RPL23AP6', Aliases='RPL23A_36_1683, dJ1068F16.1', description='ribosomal protein L23a pseudogene 6', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10349112, end_position_on_the_genomic_accession=10349639, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128710, CurrentID=0, Status='live', Symbol='SLX4IP', Aliases='C20orf94, bA204H22.1, bA254M13.1, dJ1099D15.3', description='SLX4 interacting protein', other_designations='SLX4-interacting protein|UPF0492 protein C20orf94', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=10434547, end_position_on_the_genomic_accession=10638546, orientation='plus', exon_count=13, OMIM=615958, no_hdr0=''),
  105379527 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379527, CurrentID=0, Status='live', Symbol='LOC105379527', Aliases='', description='uncharacterized LOC105379527', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390856, CurrentID=0, Status='live', Symbol='RPL29P32', Aliases='RPL29_21_1603', description='ribosomal protein L29 pseudogene 32', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=53605274, end_position_on_the_genomic_accession=53605633, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390857, CurrentID=0, Status='live', Symbol='RPSAP57', Aliases='RPSA_29_1604', description='ribosomal protein SA pseudogene 57', other_designations='', map_location='18q21.2', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=54744673, end_position_on_the_genomic_accession=54745656, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390858, CurrentID=0, Status='live', Symbol='OACYLP', Aliases='ACYL3P', description='O-acyltransferase like, pseudogene', other_designations='ACYL3 pseudogene|acyltransferase 3, pseudogene', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=59035679, end_position_on_the_genomic_accession=59053214, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379531, CurrentID=0, Status='live', Symbol='LOC105379531', Aliases='', description='uncharacterized LOC105379531', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379532, CurrentID=0, Status='live', Symbol='LOC105379532', Aliases='', description='uncharacterized LOC105379532', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379533, CurrentID=0, Status='live', Symbol='LOC105379533', Aliases='', description='uncharacterized LOC105379533', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128718, CurrentID=0, Status='live', Symbol='RPS11P1', Aliases='RPS11_3_1684, dJ1060K6.1', description='ribosomal protein S11 pseudogene 1', other_designations='', map_location='20p12.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=11345383, end_position_on_the_genomic_accession=11345855, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422605, CurrentID=0, Status='live', Symbol='ST13P16', Aliases='', description='suppression of tumorigenicity 13 (colon carcinoma) (Hsp70 interacting protein) pseudogene 16', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=87444398, end_position_on_the_genomic_accession=87445178, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379536, CurrentID=0, Status='live', Symbol='LOC105379536', Aliases='', description='uncharacterized LOC105379536', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379537 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379537, CurrentID=0, Status='live', Symbol='LOC105379537', Aliases='', description='uncharacterized LOC105379537', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379538, CurrentID=0, Status='live', Symbol='LOC105379538', Aliases='', description='putative HERC2-like protein 3', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379539, CurrentID=0, Status='live', Symbol='LOC105379539', Aliases='', description='tyrosine-protein phosphatase non-receptor type 23-like', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379540, CurrentID=0, Status='live', Symbol='LOC105379540', Aliases='', description='uncharacterized LOC105379540', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422606, CurrentID=0, Status='live', Symbol='SLC25A3P3', Aliases='', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 3 pseudogene', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=1159189, end_position_on_the_genomic_accession=1159772, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379542 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379542, CurrentID=0, Status='live', Symbol='LOC105379542', Aliases='', description='ribosome biogenesis protein BMS1 homolog', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379543, CurrentID=0, Status='live', Symbol='LOC105379543', Aliases='', description='uncharacterized LOC105379543', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390872, CurrentID=0, Status='live', Symbol='OR4F8P', Aliases='OR4F20P, OR4F9P', description='olfactory receptor family 4 subfamily F member 8 pseudogene', other_designations='olfactory receptor, family 4, subfamily F, member 20 pseudogene|olfactory receptor, family 4, subfamily F, member 8 pseudogene|olfactory receptor, family 4, subfamily F, member 9 pseudogene', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=156282, end_position_on_the_genomic_accession=157216, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653017 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653017, CurrentID=0, Status='live', Symbol='TBC1D3P3', Aliases='USP6P1', description='TBC1 domain family member 3 pseudogene 3', other_designations='TBC1 domain family, member 3 pseudogene 3|ubiquitin specific peptidase 6 pseudogene 1', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=20543480, end_position_on_the_genomic_accession=20550964, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390874, CurrentID=0, Status='live', Symbol='ONECUT3', Aliases='OC3', description='one cut homeobox 3', other_designations='OC-3|one cut domain, family member 3|transcription factor ONECUT-3', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=1753663, end_position_on_the_genomic_accession=1775445, orientation='plus', exon_count=2, OMIM=611294, no_hdr0=''),
  105379547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379547, CurrentID=0, Status='live', Symbol='LOC105379547', Aliases='', description='uncharacterized LOC105379547', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379548, CurrentID=0, Status='live', Symbol='LOC105379548', Aliases='', description='uncharacterized LOC105379548', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390877, CurrentID=0, Status='live', Symbol='LOC390877', Aliases='', description='adenylate kinase isoenzyme 1-like', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=6393250, end_position_on_the_genomic_accession=6412404, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105379550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379550, CurrentID=0, Status='live', Symbol='LOC105379550', Aliases='', description='putative uncharacterized protein LOC388882', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379551 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379551, CurrentID=0, Status='live', Symbol='LOC105379551', Aliases='', description='uncharacterized LOC105379551', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379552, CurrentID=0, Status='live', Symbol='LOC105379552', Aliases='', description='uncharacterized LOC105379552', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422608 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422608, CurrentID=0, Status='live', Symbol='SLC25A24P2', Aliases='', description='SLC25A24 pseudogene 2', other_designations='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 24 pseudogene', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=108383689, end_position_on_the_genomic_accession=108421239, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390882 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390882, CurrentID=0, Status='live', Symbol='OR7G2', Aliases='OR19-6, OST260', description='olfactory receptor family 7 subfamily G member 2', other_designations='OR19-13|olfactory receptor 19-13|olfactory receptor OR19-6|olfactory receptor, family 7, subfamily G, member 2', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9102269, end_position_on_the_genomic_accession=9103306, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390883 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390883, CurrentID=0, Status='live', Symbol='OR7G3', Aliases='OST085', description='olfactory receptor family 7 subfamily G member 3', other_designations='olfactory receptor OR19-9|olfactory receptor, family 7, subfamily G, member 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9126012, end_position_on_the_genomic_accession=9126950, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379556, CurrentID=0, Status='live', Symbol='LOC105379556', Aliases='', description='nascent polypeptide-associated complex subunit alpha, muscle-specific form-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390885, CurrentID=0, Status='live', Symbol='OR7E18P', Aliases='HSTPCR26P, OR19-14, OR7E61, OR7E98P, TPCR26', description='olfactory receptor family 7 subfamily E member 18 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 18 pseudogene|olfactory receptor, family 7, subfamily E, member 61|olfactory receptor, family 7, subfamily E, member 98 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9260977, end_position_on_the_genomic_accession=9261797, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390886, CurrentID=0, Status='live', Symbol='OR7H1P', Aliases='OR7H1', description='olfactory receptor family 7 subfamily H member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily H, member 1 pseudogene', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=9278433, end_position_on_the_genomic_accession=9279369, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379559 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379559, CurrentID=0, Status='live', Symbol='LOC105379559', Aliases='', description='Ig kappa chain V-I region Hau-like', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379560, CurrentID=0, Status='live', Symbol='LOC105379560', Aliases='', description='uncharacterized LOC105379560', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128745 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128745, CurrentID=0, Status='live', Symbol='RPLP0P1', Aliases='dJ777L9.2', description='ribosomal protein, large, P0 pseudogene 1', other_designations='', map_location='20p12.1', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=16586333, end_position_on_the_genomic_accession=16586788, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379562, CurrentID=0, Status='live', Symbol='LOC105379562', Aliases='', description='uncharacterized LOC105379562', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379563, CurrentID=0, Status='live', Symbol='LOC105379563', Aliases='', description='uncharacterized LOC105379563', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390892 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390892, CurrentID=0, Status='live', Symbol='OR7A10', Aliases='BC85395_3, OR19-18', description='olfactory receptor family 7 subfamily A member 10', other_designations='OST027|olfactory receptor OR19-18|olfactory receptor, family 7, subfamily A, member 10', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14840948, end_position_on_the_genomic_accession=14841877, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379565 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379565, CurrentID=0, Status='live', Symbol='LOC105379565', Aliases='', description='uncharacterized LOC105379565', other_designations='', map_location='', chromosome='Un', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390894 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390894, CurrentID=0, Status='live', Symbol='OR7A2P', Aliases='OLF4p, OR19-18, OR7A2, OR7A7, hg1003', description='olfactory receptor family 7 subfamily A member 2 pseudogene', other_designations='Olfactory receptor 7A2|olfactory receptor, family 7, subfamily A, member 2 pseudogene|olfactory receptor, family 7, subfamily A, member 7|seven transmembrane helix receptor', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14863586, end_position_on_the_genomic_accession=14865262, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  390895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390895, CurrentID=0, Status='live', Symbol='OR7A1P', Aliases='OLF4p, OR19-3, OR7A6P, hg513', description='olfactory receptor family 7 subfamily A member 1 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 1 pseudogene|olfactory receptor, family 7, subfamily A, member 6 pseudogene', map_location='19p13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=14891859, end_position_on_the_genomic_accession=14892804, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379568 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379568, CurrentID=0, Status='live', Symbol='LOC105379568', Aliases='', description='uncharacterized LOC105379568', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379569 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379569, CurrentID=0, Status='live', Symbol='LOC105379569', Aliases='', description='uncharacterized LOC105379569', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653042, CurrentID=0, Status='live', Symbol='REXO1L4P', Aliases='', description='REX1, RNA exonuclease 1 homolog-like 4, pseudogene', other_designations='REX1, RNA exonuclease 1 homolog (S. cerevisiae)-like 4, pseudogene|REX1, RNA exonuclease 1 homolog-like 4 (pseudogene)', map_location='8q21.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=85802518, end_position_on_the_genomic_accession=85804545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379571 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379571, CurrentID=0, Status='live', Symbol='LOC105379571', Aliases='', description='uncharacterized LOC105379571', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128756, CurrentID=0, Status='live', Symbol='RPL15P1', Aliases='RPL15_9_1687, dJ726N1.3', description='ribosomal protein L15 pseudogene 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=18090059, end_position_on_the_genomic_accession=18090751, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653045 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653045, CurrentID=0, Status='live', Symbol='PTPN20CP', Aliases='PTPN20C, bA164N7.3', description='protein tyrosine phosphatase, non-receptor type 20C, pseudogene', other_designations='', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48105157, end_position_on_the_genomic_accession=48140426, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379574, CurrentID=0, Status='live', Symbol='LOC105379574', Aliases='', description='uncharacterized LOC105379574', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379575 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379575, CurrentID=0, Status='live', Symbol='LOC105379575', Aliases='', description='collagen alpha-2(IV) chain-like', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390904 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390904, CurrentID=0, Status='live', Symbol='KRT18P40', Aliases='', description='keratin 18 pseudogene 40', other_designations='', map_location='19p12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=20961058, end_position_on_the_genomic_accession=20963425, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379577 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379577, CurrentID=0, Status='live', Symbol='LOC105379577', Aliases='', description='uncharacterized LOC105379577', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379578, CurrentID=0, Status='live', Symbol='LOC105379578', Aliases='', description='uncharacterized LOC105379578', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379580 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379580, CurrentID=0, Status='live', Symbol='LOC105379580', Aliases='', description='uncharacterized LOC105379580', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379581, CurrentID=0, Status='live', Symbol='LOC105379581', Aliases='', description='uncharacterized LOC105379581', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379582 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379582, CurrentID=0, Status='live', Symbol='LOC105379582', Aliases='', description='uncharacterized LOC105379582', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379583, CurrentID=0, Status='live', Symbol='LOC105379583', Aliases='', description='uncharacterized LOC105379583', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379584, CurrentID=0, Status='live', Symbol='LOC105379584', Aliases='', description='uncharacterized LOC105379584', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  401459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401459, CurrentID=0, Status='live', Symbol='FLJ46365', Aliases='', description='uncharacterized LOC401459', other_designations='', map_location='8q11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.9', start_position_on_the_genomic_accession=49665514, end_position_on_the_genomic_accession=49666906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  128770 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128770, CurrentID=0, Status='live', Symbol='EIF4E2P1', Aliases='EIF4EL4, dJ1049G11.3', description='eukaryotic translation initiation factor 4E family member 2 pseudogene 1', other_designations='eukaryotic translation initiation factor 4E member 2 pseudogene 1|eukaryotic translation initiation factor 4E-like 4|novel protein similar to eukaryotic translation initiation factor 4E-like 4 EIF4EL4', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20659688, end_position_on_the_genomic_accession=20659964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379587, CurrentID=0, Status='live', Symbol='LOC105379587', Aliases='', description='putative ubiquitin carboxyl-terminal hydrolase 17-like protein 23', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390916, CurrentID=0, Status='live', Symbol='NUDT19', Aliases='RP2', description='nudix hydrolase 19', other_designations='CTC-379B2.4|nucleoside diphosphate-linked moiety X motif 19, mitochondrial|nudix (nucleoside diphosphate linked moiety X)-type motif 19|nudix motif 19', map_location='19q13.11', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=32666296, end_position_on_the_genomic_accession=32713796, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  653061 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653061, CurrentID=0, Status='live', Symbol='GOLGA8S', Aliases='', description='golgin A8 family member S', other_designations='golgin A8 family, member B pseudogene|golgin A8 family, member S', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=23354748, end_position_on_the_genomic_accession=23368324, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  128774 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128774, CurrentID=0, Status='live', Symbol='MRPS11P1', Aliases='dJ746H2.1', description='mitochondrial ribosomal protein S11 pseudogene 1', other_designations='', map_location='20p11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=20853831, end_position_on_the_genomic_accession=20854645, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379591 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379591, CurrentID=0, Status='live', Symbol='LOC105379591', Aliases='', description='uncharacterized LOC105379591', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379592, CurrentID=0, Status='live', Symbol='LOC105379592', Aliases='', description='uncharacterized LOC105379592', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379593 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379593, CurrentID=0, Status='live', Symbol='LOC105379593', Aliases='', description='uncharacterized LOC105379593', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379594, CurrentID=0, Status='live', Symbol='LOC105379594', Aliases='', description='uncharacterized LOC105379594', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653067 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653067, CurrentID=0, Status='live', Symbol='XAGE1E', Aliases='CT12.1, CT12.1C, CT12.1D, CT12.1E, CTP9, GAGED2, XAGE-1, XAGE1, XAGE1C, XAGE1D', description='X antigen family member 1E', other_designations='G antigen family D member 2|X antigen family, member 1C|X antigen family, member 1D|X antigen family, member 1E|cancer/testis antigen 12.1|cancer/testis antigen family 12, member 1c|cancer/testis antigen family 12, member 1d|cancer/testis antigen family 12, member 1e|cancer/testis associated protein|protein XAGE-1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52512077, end_position_on_the_genomic_accession=52517221, orientation='minus', exon_count=4, OMIM='300289,300744,300745', no_hdr0=''),
  390924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390924, CurrentID=0, Status='live', Symbol='EEF1A1P7', Aliases='EEF1AL5', description='eukaryotic translation elongation factor 1 alpha 1 pseudogene 7', other_designations='eukaryotic translation elongation factor 1 alpha 2 pseudogene|eukaryotic translation elongation factor 1 alpha-like 5 pseudogene', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=35382172, end_position_on_the_genomic_accession=35384055, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379597, CurrentID=0, Status='live', Symbol='LOC105379597', Aliases='', description='uncharacterized LOC105379597', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379598, CurrentID=0, Status='live', Symbol='LOC105379598', Aliases='', description='uncharacterized LOC105379598', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390927, CurrentID=0, Status='live', Symbol='ZNF793', Aliases='', description='zinc finger protein 793', other_designations='', map_location='19q13.12', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=37506447, end_position_on_the_genomic_accession=37543337, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  390928 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390928, CurrentID=0, Status='live', Symbol='ACP7', Aliases='PAPL', description='acid phosphatase 7, tartrate resistant (putative)', other_designations='purple acid phosphatase long form', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=39084305, end_position_on_the_genomic_accession=39111489, orientation='plus', exon_count=14, OMIM=610490, no_hdr0=''),
  653073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653073, CurrentID=0, Status='live', Symbol='GOLGA8J', Aliases='', description='golgin A8 family member J', other_designations='Golgin subfamily A member 8-like protein 3|golgin A8 family, member J', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30082375, end_position_on_the_genomic_accession=30096701, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105379602 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379602, CurrentID=0, Status='live', Symbol='LOC105379602', Aliases='', description='uncharacterized LOC105379602', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653075, CurrentID=0, Status='live', Symbol='GOLGA8T', Aliases='', description='golgin A8 family member T', other_designations='golgin A8 family, member A pseudogene|golgin A8 family, member T', map_location='15q13.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=30135787, end_position_on_the_genomic_accession=30147192, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  105379604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379604, CurrentID=0, Status='live', Symbol='LOC105379604', Aliases='', description='set1/Ash2 histone methyltransferase complex subunit ASH2 pseudogene', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379605 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379605, CurrentID=0, Status='live', Symbol='LOC105379605', Aliases='', description='uncharacterized LOC105379605', other_designations='', map_location='', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128790, CurrentID=0, Status='live', Symbol='SLC25A6P1', Aliases='SLC25A5L, dJ1065O2.2', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6, pseudogene 1', other_designations='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene', map_location='20p11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=21742352, end_position_on_the_genomic_accession=21743076, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653079 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653079, CurrentID=0, Status='live', Symbol='RPL21P98', Aliases='RPL21_47_1259', description='ribosomal protein L21 pseudogene 98', other_designations='', map_location='12q21.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=76660365, end_position_on_the_genomic_accession=76660928, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653080, CurrentID=0, Status='live', Symbol='LOC653080', Aliases='', description='glucuronidase, beta pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70785397, end_position_on_the_genomic_accession=70798698, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390937 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390937, CurrentID=0, Status='live', Symbol='LOC390937', Aliases='', description='Ets2 repressor factor-like', other_designations='', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=41907535, end_position_on_the_genomic_accession=41928122, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  653082 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653082, CurrentID=0, Status='live', Symbol='ZDHHC11B', Aliases='DHHC-11B', description='zinc finger, DHHC-type containing 11B', other_designations='Probable palmitoyltransferase ZDHHC11B|Zinc finger DHHC domain-containing protein 11B', map_location='5p15.33', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=710355, end_position_on_the_genomic_accession=784749, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  105379611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379611, CurrentID=0, Status='live', Symbol='LOC105379611', Aliases='', description='uncharacterized LOC105379611', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390940, CurrentID=0, Status='live', Symbol='PINLYP', Aliases='', description='phospholipase A2 inhibitor and LY6/PLAUR domain containing', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43576800, end_position_on_the_genomic_accession=43582104, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105379613 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379613, CurrentID=0, Status='live', Symbol='LOC105379613', Aliases='', description='uncharacterized LOC105379613', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379614 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379614, CurrentID=0, Status='live', Symbol='LOC105379614', Aliases='', description='uncharacterized LOC105379614', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379615 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379615, CurrentID=0, Status='live', Symbol='LOC105379615', Aliases='', description='uncharacterized LOC105379615', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379616 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379616, CurrentID=0, Status='live', Symbol='LOC105379616', Aliases='', description='uncharacterized LOC105379616', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379617, CurrentID=0, Status='live', Symbol='LOC105379617', Aliases='', description='zinc finger protein 430-like', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379618 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379618, CurrentID=0, Status='live', Symbol='LOC105379618', Aliases='', description='uncharacterized LOC105379618', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379619, CurrentID=0, Status='live', Symbol='LOC105379619', Aliases='', description='uncharacterized LOC105379619', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390948 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390948, CurrentID=0, Status='live', Symbol='HNRNPMP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein M pseudogene 2', other_designations='', map_location='19q13.32', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=46809584, end_position_on_the_genomic_accession=46810179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  390949 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390949, CurrentID=0, Status='live', Symbol='RPL23AP80', Aliases='RPL23A_35_1661', description='ribosomal protein L23a pseudogene 80', other_designations='', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=47794466, end_position_on_the_genomic_accession=47794902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379622, CurrentID=0, Status='live', Symbol='LOC105379622', Aliases='', description='uncharacterized LOC105379622', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379623 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379623, CurrentID=0, Status='live', Symbol='LOC105379623', Aliases='', description='uncharacterized LOC105379623', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379624, CurrentID=0, Status='live', Symbol='LOC105379624', Aliases='', description='uncharacterized LOC105379624', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379625, CurrentID=0, Status='live', Symbol='LOC105379625', Aliases='', description='uncharacterized LOC105379625', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379626 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379626, CurrentID=0, Status='live', Symbol='LOC105379626', Aliases='', description='uncharacterized LOC105379626', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379627, CurrentID=0, Status='live', Symbol='LOC105379627', Aliases='', description='uncharacterized LOC105379627', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390956, CurrentID=0, Status='live', Symbol='LOC390956', Aliases='', description='peptidyl-prolyl cis-trans isomerase A pseudogene', other_designations='', map_location='19q13.41', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=50926868, end_position_on_the_genomic_accession=50935870, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379629, CurrentID=0, Status='live', Symbol='LOC105379629', Aliases='', description='uncharacterized LOC105379629', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379630, CurrentID=0, Status='live', Symbol='LOC105379630', Aliases='', description='mucin-1-like', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379631, CurrentID=0, Status='live', Symbol='LOC105379631', Aliases='', description='uncharacterized LOC105379631', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379632, CurrentID=0, Status='live', Symbol='LOC105379632', Aliases='', description='uncharacterized LOC105379632', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128817, CurrentID=0, Status='live', Symbol='CSTL1', Aliases='CTES1, RCET11, dJ322G13.4', description='cystatin-like 1', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23439639, end_position_on_the_genomic_accession=23461091, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105379634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379634, CurrentID=0, Status='live', Symbol='LOC105379634', Aliases='', description='uncharacterized LOC105379634', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390963, CurrentID=0, Status='live', Symbol='ZNF818P', Aliases='ZNF818', description='zinc finger protein 818, pseudogene', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53212948, end_position_on_the_genomic_accession=53216102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653108 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653108, CurrentID=0, Status='live', Symbol='CXADRP1', Aliases='CAR, CXADRP', description='coxsackie virus and adenovirus receptor pseudogene 1', other_designations='', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13675941, end_position_on_the_genomic_accession=13677035, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128821, CurrentID=0, Status='live', Symbol='CST9L', Aliases='CTES7B, bA218C14.1', description='cystatin 9-like', other_designations='testatin', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23564732, end_position_on_the_genomic_accession=23568749, orientation='minus', exon_count=3, OMIM=616536, no_hdr0=''),
  128822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128822, CurrentID=0, Status='live', Symbol='CST9', Aliases='CLM, CTES7A', description='cystatin 9 (testatin)', other_designations='cystatin-like molecule', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=23602410, end_position_on_the_genomic_accession=23605973, orientation='minus', exon_count=2, OMIM=616543, no_hdr0=''),
  105379639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379639, CurrentID=0, Status='live', Symbol='LOC105379639', Aliases='', description='uncharacterized LOC105379639', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379640 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379640, CurrentID=0, Status='live', Symbol='LOC105379640', Aliases='', description='uncharacterized LOC105379640', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653113 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653113, CurrentID=0, Status='live', Symbol='FAM86FP', Aliases='', description='family with sequence similarity 86 member F, pseudogene', other_designations='family with sequence similarity 86, member F, pseudogene', map_location='12p13.31', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=8231049, end_position_on_the_genomic_accession=8242946, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  128826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128826, CurrentID=0, Status='live', Symbol='MIR1-1HG', Aliases='C20orf166, MIR133A2HG, dJ353C17.1', description='MIR1-1 host gene', other_designations='MIR1-1 host gene protein|MIR133A2 host|uncharacterized protein C20orf166', map_location='20q13.33', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=62550453, end_position_on_the_genomic_accession=62570764, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  653115 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653115, CurrentID=0, Status='live', Symbol='HNRNPA3P6', Aliases='', description='heterogeneous nuclear ribonucleoprotein A3 pseudogene 6', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75214459, end_position_on_the_genomic_accession=75215648, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379644 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379644, CurrentID=0, Status='live', Symbol='LOC105379644', Aliases='', description='uncharacterized LOC105379644', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379645 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379645, CurrentID=0, Status='live', Symbol='LOC105379645', Aliases='', description='killer cell immunoglobulin-like receptor 2DL2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379646 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379646, CurrentID=0, Status='live', Symbol='LOC105379646', Aliases='', description='uncharacterized LOC105379646', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379647, CurrentID=0, Status='live', Symbol='LOC105379647', Aliases='', description='uncharacterized LOC105379647', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379648 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379648, CurrentID=0, Status='live', Symbol='LOC105379648', Aliases='', description='DLA class II histocompatibility antigen, DR-1 beta chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653121 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653121, CurrentID=0, Status='live', Symbol='ZBTB8A', Aliases='BOZF1, ZBTB8, ZNF916A', description='zinc finger and BTB domain containing 8A', other_designations='BTB/POZ and zinc-finger domain-containing factor|BTB/POZ and zinc-finger domains factor on chromosome 1|zinc finger and BTB domain containing 8', map_location='1p35.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=32539145, end_position_on_the_genomic_accession=32605950, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  653122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653122, CurrentID=0, Status='live', Symbol='LSM12P1', Aliases='LSM12P', description='LSM12 pseudogene 1', other_designations='LSM12 homolog pseudogene 1', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=35523579, end_position_on_the_genomic_accession=35525774, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653123, CurrentID=0, Status='live', Symbol='FAM74A6', Aliases='', description='family with sequence similarity 74 member A6', other_designations='family with sequence similarity 74, member A6', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=60929782, end_position_on_the_genomic_accession=60935727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  390980 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390980, CurrentID=0, Status='live', Symbol='ZNF805', Aliases='', description='zinc finger protein 805', other_designations='CTC-444N24.8', map_location='19q13.43', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=57240685, end_position_on_the_genomic_accession=57262738, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  653125 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653125, CurrentID=0, Status='live', Symbol='GOLGA8K', Aliases='', description='golgin A8 family member K', other_designations='Golgin subfamily A member 8K|golgin A8 family, member K|putative golgin subfamily A member 8I-like', map_location='15q13.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=32389585, end_position_on_the_genomic_accession=32403292, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105379654 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379654, CurrentID=0, Status='live', Symbol='LOC105379654', Aliases='', description='uncharacterized LOC105379654', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100422625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422625, CurrentID=0, Status='live', Symbol='GOT2P5', Aliases='GOT2Y', description='glutamic-oxaloacetic transaminase 2 pseudogene 5', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=6837699, end_position_on_the_genomic_accession=6838252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379656, CurrentID=0, Status='live', Symbol='LOC105379656', Aliases='', description='uncharacterized LOC105379656', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379657, CurrentID=0, Status='live', Symbol='LOC105379657', Aliases='', description='uncharacterized LOC105379657', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379658, CurrentID=0, Status='live', Symbol='LOC105379658', Aliases='', description='DLA class II histocompatibility antigen, DR-1 beta chain-like', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379659, CurrentID=0, Status='live', Symbol='LOC105379659', Aliases='', description='uncharacterized LOC105379659', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379660 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379660, CurrentID=0, Status='live', Symbol='LOC105379660', Aliases='', description='uncharacterized LOC105379660', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379661 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379661, CurrentID=0, Status='live', Symbol='LOC105379661', Aliases='', description='uncharacterized LOC105379661', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379662 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379662, CurrentID=0, Status='live', Symbol='LOC105379662', Aliases='', description='killer cell immunoglobulin-like receptor 2DS1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379663 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379663, CurrentID=0, Status='live', Symbol='LOC105379663', Aliases='', description='uncharacterized LOC105379663', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  390992 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390992, CurrentID=0, Status='live', Symbol='HES3', Aliases='bHLHb43', description='hes family bHLH transcription factor 3', other_designations='class B basic helix-loop-helix protein 43|hairy and enhancer of split 3', map_location='1p36.31', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=6244158, end_position_on_the_genomic_accession=6245578, orientation='plus', exon_count=4, OMIM=609971, no_hdr0=''),
  386679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=386679, CurrentID=0, Status='live', Symbol='KRTAP10-2', Aliases='KAP10.2, KAP18-2, KAP18.2, KRTAP10.2, KRTAP18-2, KRTAP18.2', description='keratin associated protein 10-2', other_designations='high sulfur keratin-associated protein 10.2|keratin-associated protein 18-2|keratin-associated protein 18.2', map_location='21q22.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=44550357, end_position_on_the_genomic_accession=44551505, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379666 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379666, CurrentID=0, Status='live', Symbol='LOC105379666', Aliases='', description='uncharacterized LOC105379666', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379667, CurrentID=0, Status='live', Symbol='LOC105379667', Aliases='', description='uncharacterized LOC105379667', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653140, CurrentID=0, Status='live', Symbol='FAM228A', Aliases='C2orf84', description='family with sequence similarity 228 member A', other_designations='UPF0638 protein C2orf84|family with sequence similarity 228, member A', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24175043, end_position_on_the_genomic_accession=24191698, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  128853 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128853, CurrentID=0, Status='live', Symbol='DUSP15', Aliases='C20orf57, VHY', description='dual specificity phosphatase 15', other_designations='VH1-related member Y|dual specificity phosphatase-like 15|vaccinia virus VH1-related dual-specific protein phosphatase Y', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=31861067, end_position_on_the_genomic_accession=31870676, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  128854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128854, CurrentID=0, Status='live', Symbol='TSPY26P', Aliases='TSPYL3, bA392M18.1', description='testis specific protein, Y-linked 26, pseudogene', other_designations='TSPY-like 3 (pseudogene)', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=32189146, end_position_on_the_genomic_accession=32190360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  390999 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=390999, CurrentID=0, Status='live', Symbol='PRAMEF12', Aliases='', description='PRAME family member 12', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=12763223, end_position_on_the_genomic_accession=12777906, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105379672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379672, CurrentID=0, Status='live', Symbol='LOC105379672', Aliases='', description='uncharacterized LOC105379672', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653145, CurrentID=0, Status='live', Symbol='ANXA8', Aliases='ANX8, CH17-360D5.2', description='annexin A8', other_designations='VAC-beta|annexin VIII|annexin-8|vascular anticoagulant-beta', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=47467993, end_position_on_the_genomic_accession=47993768, orientation='minus', exon_count=21, OMIM=602396, no_hdr0=''),
  391002 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391002, CurrentID=0, Status='live', Symbol='PRAMEF8', Aliases='PRAMEF24', description='PRAME family member 8', other_designations='putative PRAME family member 24', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13281304, end_position_on_the_genomic_accession=13285137, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  391003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391003, CurrentID=0, Status='live', Symbol='PRAMEF18', Aliases='', description='PRAME family member 18', other_designations='PRAME family member-like', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13222695, end_position_on_the_genomic_accession=13226154, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391004 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391004, CurrentID=0, Status='live', Symbol='PRAMEF17', Aliases='', description='PRAME family member 17', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13389632, end_position_on_the_genomic_accession=13392629, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  128861 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128861, CurrentID=0, Status='live', Symbol='BPIFA3', Aliases='C20orf71, SPLUNC3', description='BPI fold containing family A member 3', other_designations='BPI fold containing family A, member 3|short long palate, lung and nasal epithelium carcinoma associated 3', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33217310, end_position_on_the_genomic_accession=33227806, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  128862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128862, CurrentID=0, Status='live', Symbol='RPL12P3', Aliases='RPL12_17_1694, bA49G10.7', description='ribosomal protein L12 pseudogene 3', other_designations='', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33272262, end_position_on_the_genomic_accession=33272657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379679, CurrentID=0, Status='live', Symbol='LOC105379679', Aliases='', description='uncharacterized LOC105379679', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128864 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128864, CurrentID=0, Status='live', Symbol='C20orf144', Aliases='dJ63M2.6', description='chromosome 20 open reading frame 144', other_designations='bcl-2-like protein from testis|bclt', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33662286, end_position_on_the_genomic_accession=33663915, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105379681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379681, CurrentID=0, Status='live', Symbol='LOC105379681', Aliases='', description='uncharacterized LOC105379681', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128866 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128866, CurrentID=0, Status='live', Symbol='CHMP4B', Aliases='C20orf178, CHMP4A, CTPP3, CTRCT31, SNF7, SNF7-2, Shax1, VPS32B, Vps32-2, dJ553F4.4', description='charged multivesicular body protein 4B', other_designations='SNF7 homolog associated with Alix 1|Snf7 homologue associated with Alix 1|chromatin modifying protein 4B|chromatin-modifying protein 4b|hSnf7-2|hVps32-2|vacuolar protein sorting-associated protein 32-2|vacuolar protein-sorting-associated protein 32-2', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33811304, end_position_on_the_genomic_accession=33854367, orientation='plus', exon_count=5, OMIM=610897, no_hdr0=''),
  105379683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379683, CurrentID=0, Status='live', Symbol='LOC105379683', Aliases='', description='uncharacterized LOC105379683', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653156 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653156, CurrentID=0, Status='live', Symbol='RPL21P53', Aliases='RPL21_19_517', description='ribosomal protein L21 pseudogene 53', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=127812674, end_position_on_the_genomic_accession=127813240, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391013, CurrentID=0, Status='live', Symbol='PLA2G2C', Aliases='', description='phospholipase A2 group IIC', other_designations='phosphatidylcholine 2-acylhydrolase-like protein GIIC|phospholipase A2, group IIC (possible pseudogene)', map_location='1p36.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=20163076, end_position_on_the_genomic_accession=20163980, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105379686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379686, CurrentID=0, Status='live', Symbol='LOC105379686', Aliases='', description='uncharacterized LOC105379686', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379687, CurrentID=0, Status='live', Symbol='LOC105379687', Aliases='', description='uncharacterized LOC105379687', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128872, CurrentID=0, Status='live', Symbol='HMGB3P1', Aliases='HMG4L, HMGB3L1, dJ18C9.3', description='high mobility group box 3 pseudogene 1', other_designations='high-mobility group box 3 pseudogene 1|high-mobility group box 3-like 1|high-mobility group box pseudogene 1', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=34833575, end_position_on_the_genomic_accession=34834462, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379689, CurrentID=0, Status='live', Symbol='LOC105379689', Aliases='', description='F-box only protein 25-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653162, CurrentID=0, Status='live', Symbol='RPSAP9', Aliases='LAMR1P9, RPSA_20_986', description='ribosomal protein SA pseudogene 9', other_designations='laminin receptor 1 pseudogene 9', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76398599, end_position_on_the_genomic_accession=76400038, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653163 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653163, CurrentID=0, Status='live', Symbol='LCN1P2', Aliases='LCN1L2, bA244N20.11', description='lipocalin 1 pseudogene 2', other_designations='lipocalin 1-like 2', map_location='9q34.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=133317505, end_position_on_the_genomic_accession=133318569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  128876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128876, CurrentID=0, Status='live', Symbol='FAM83C', Aliases='C20orf128, dJ614O4.7', description='family with sequence similarity 83 member C', other_designations='family with sequence similarity 83, member C', map_location='20q11.22', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=35285731, end_position_on_the_genomic_accession=35292422, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379693, CurrentID=0, Status='live', Symbol='LOC105379693', Aliases='', description='uncharacterized LOC105379693', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653166 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653166, CurrentID=0, Status='live', Symbol='OR1D4', Aliases='OR17-30', description='olfactory receptor family 1 subfamily D member 4 (gene/pseudogene)', other_designations='olfactory receptor 17-30|olfactory receptor 1D4|olfactory receptor OR17-1|olfactory receptor pseudogene|olfactory receptor, family 1, subfamily D, member 4 pseudogene', map_location='17p13.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=3240676, end_position_on_the_genomic_accession=3241265, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  128879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128879, CurrentID=0, Status='live', Symbol='HMGB3P2', Aliases='HMG4L2, HMGB3L2, dJ550H1.1', description='high mobility group box 3 pseudogene 2', other_designations='high mobility group protein pseudogene|high-mobility group (nonhistone chromosomal) protein 4-like 2|high-mobility group box 3 pseudogene 2|high-mobility group box 3-like 2', map_location='20q11.23', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=36076000, end_position_on_the_genomic_accession=36076754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379696, CurrentID=0, Status='live', Symbol='LOC105379696', Aliases='', description='uncharacterized LOC105379696', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379697, CurrentID=0, Status='live', Symbol='LOC105379697', Aliases='', description='uncharacterized LOC105379697', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391026, CurrentID=0, Status='live', Symbol='ACTN4P2', Aliases='', description='actinin, alpha 4 pseudogene 2', other_designations='alpha-actinin-4-like', map_location='1p34.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=37767588, end_position_on_the_genomic_accession=37779538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379699 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379699, CurrentID=0, Status='live', Symbol='LOC105379699', Aliases='', description='uncharacterized LOC105379699', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379700, CurrentID=0, Status='live', Symbol='LOC105379700', Aliases='', description='uncharacterized LOC105379700', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379701, CurrentID=0, Status='live', Symbol='LOC105379701', Aliases='', description='uncharacterized LOC105379701', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653174, CurrentID=0, Status='live', Symbol='TSPY6P', Aliases='TSPYP6', description='testis specific protein, Y-linked 6, pseudogene', other_designations='testis specific protein, Y-linked 6 (pseudogene)|testis specific protein, Y-linked pseudogene 6', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9507550, end_position_on_the_genomic_accession=9510363, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653175 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653175, CurrentID=0, Status='live', Symbol='HAUS6P1', Aliases='', description='HAUS augmin like complex subunit 6 pseudogene 1', other_designations='HAUS augmin-like complex, subunit 6 pseudogene 1', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53187136, end_position_on_the_genomic_accession=53188941, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379704 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379704, CurrentID=0, Status='live', Symbol='LOC105379704', Aliases='', description='CASP-like protein 4A1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391033 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391033, CurrentID=0, Status='live', Symbol='MKRN8P', Aliases='MKRNP4, ZNF127L4', description='makorin ring finger protein 8, pseudogene', other_designations='makorin ring finger protein 1 pseudogene|makorin ring finger protein pseudogene 4|zinc finger protein 127-like 4', map_location='1p34.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=42889631, end_position_on_the_genomic_accession=42892440, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379706, CurrentID=0, Status='live', Symbol='LOC105379706', Aliases='', description='uncharacterized LOC105379706', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379707, CurrentID=0, Status='live', Symbol='LOC105379707', Aliases='', description='uncharacterized LOC105379707', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379708 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379708, CurrentID=0, Status='live', Symbol='LOC105379708', Aliases='', description='uncharacterized LOC105379708', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391037 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391037, CurrentID=0, Status='live', Symbol='SKINTL', Aliases='BTN12, SKINT1, SKINTP', description='Skint-like, pseudogene', other_designations='', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=48101715, end_position_on_the_genomic_accession=48182428, orientation='minus', exon_count=10, OMIM=616392, no_hdr0=''),
  105379710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379710, CurrentID=0, Status='live', Symbol='LOC105379710', Aliases='', description='uncharacterized LOC105379710', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391039 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391039, CurrentID=0, Status='live', Symbol='CFL1P2', Aliases='CFLL2, CFLP2', description='cofilin 1 (non-muscle) pseudogene 2', other_designations='cofilin pseudogene 2|cofilin-like 2', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=51157440, end_position_on_the_genomic_accession=51158540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391040 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391040, CurrentID=0, Status='live', Symbol='BTF3L4P3', Aliases='', description='basic transcription factor 3-like 4 pseudogene 3', other_designations='', map_location='6q23.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=137543502, end_position_on_the_genomic_accession=137544443, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379713, CurrentID=0, Status='live', Symbol='LOC105379713', Aliases='', description='uncharacterized LOC105379713', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379714, CurrentID=0, Status='live', Symbol='LOC105379714', Aliases='', description='uncharacterized LOC105379714', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379715 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379715, CurrentID=0, Status='live', Symbol='LOC105379715', Aliases='', description='uncharacterized LOC105379715', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653188, CurrentID=0, Status='live', Symbol='GUSBP3', Aliases='', description='glucuronidase, beta pseudogene 3', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=69639463, end_position_on_the_genomic_accession=69710527, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  653189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653189, CurrentID=0, Status='live', Symbol='BTF3L4P2', Aliases='BTF3L4P', description='basic transcription factor 3-like 4 pseudogene 2', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=159003650, end_position_on_the_genomic_accession=159004718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653190, CurrentID=0, Status='live', Symbol='ABCC6P1', Aliases='', description='ATP binding cassette subfamily C member 6 pseudogene 1 (functional)', other_designations='ATP-binding cassette, sub-family C, member 6 pseudogene 1 (functional)', map_location='16p12.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18571248, end_position_on_the_genomic_accession=18598289, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105379719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379719, CurrentID=0, Status='live', Symbol='LOC105379719', Aliases='', description='uncharacterized LOC105379719', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653192, CurrentID=0, Status='live', Symbol='TRIM43B', Aliases='', description='tripartite motif containing 43B', other_designations='Tripartite motif-containing protein 43-like protein 1|tripartite motif-containing 43B', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95476967, end_position_on_the_genomic_accession=95484731, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105379721 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379721, CurrentID=0, Status='live', Symbol='LOC105379721', Aliases='', description='uncharacterized LOC105379721', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653194, CurrentID=0, Status='live', Symbol='KHDC1P1', Aliases='', description='KH homology domain containing 1 pseudogene 1', other_designations='KH homology domain-containing protein 1-like', map_location='6q13', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=73208981, end_position_on_the_genomic_accession=73210972, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391051 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391051, CurrentID=0, Status='live', Symbol='UOX', Aliases='UOXP, URICASE', description='urate oxidase, pseudogene', other_designations='', map_location='1p22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=84364958, end_position_on_the_genomic_accession=84397893, orientation='minus', exon_count=9, OMIM=191540, no_hdr0=''),
  105379724 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379724, CurrentID=0, Status='live', Symbol='LOC105379724', Aliases='', description='uncharacterized LOC105379724', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391053, CurrentID=0, Status='live', Symbol='CAPNS1P1', Aliases='', description='calpain, small subunit 1 pseudogene 1', other_designations='', map_location='1p22.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=89394021, end_position_on_the_genomic_accession=89394660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379726, CurrentID=0, Status='live', Symbol='LOC105379726', Aliases='', description='uncharacterized LOC105379726', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653199 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653199, CurrentID=0, Status='live', Symbol='TCAF1P1', Aliases='FAM115B', description='TRPM8 channel-associated factor 1 pseudogene 1', other_designations='family with sequence similarity 115, member A pseudogene|family with sequence similarity 115, member B (pseudogene)', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143598038, end_position_on_the_genomic_accession=143609179, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379728, CurrentID=0, Status='live', Symbol='LOC105379728', Aliases='', description='uncharacterized LOC105379728', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379729, CurrentID=0, Status='live', Symbol='LOC105379729', Aliases='', description='uncharacterized LOC105379729', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379730, CurrentID=0, Status='live', Symbol='LOC105379730', Aliases='', description='uncharacterized LOC105379730', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653203 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653203, CurrentID=0, Status='live', Symbol='FAM230A', Aliases='DGCR15', description='family with sequence similarity 230 member A', other_designations='family with sequence similarity 230, member A', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18493207, end_position_on_the_genomic_accession=18500595, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105379732 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379732, CurrentID=0, Status='live', Symbol='LOC105379732', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379733, CurrentID=0, Status='live', Symbol='LOC105379733', Aliases='', description='putative uncharacterized protein FLJ46235', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379734, CurrentID=0, Status='live', Symbol='LOC105379734', Aliases='', description='uncharacterized LOC105379734', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379735, CurrentID=0, Status='live', Symbol='LOC105379735', Aliases='', description='integrator complex subunit 4-like protein 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379736 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379736, CurrentID=0, Status='live', Symbol='LOC105379736', Aliases='', description='uncharacterized LOC105379736', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379737, CurrentID=0, Status='live', Symbol='LOC105379737', Aliases='', description='uncharacterized LOC105379737', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379738, CurrentID=0, Status='live', Symbol='LOC105379738', Aliases='', description='uncharacterized LOC105379738', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  283631 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283631, CurrentID=0, Status='live', Symbol='BTF3P2', Aliases='', description='basic transcription factor 3, pseudogene 2', other_designations='', map_location='14q12', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=28672982, end_position_on_the_genomic_accession=28673841, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379740 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379740, CurrentID=0, Status='live', Symbol='LOC105379740', Aliases='', description='uncharacterized LOC105379740', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391069 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391069, CurrentID=0, Status='live', Symbol='GAPDHP64', Aliases='GAPDHL9, GAPDL9', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 64', other_designations='glyceraldehyde-3-phosphate dehydrogenase-like 9, pseudogene', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=116713635, end_position_on_the_genomic_accession=116714911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653214, CurrentID=0, Status='live', Symbol='PPIAP22', Aliases='PPIA3L, PPIAL3', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene 22', other_designations='peptidylprolyl isomerase A (cyclophilin A)-like 3|peptidylprolyl isomerase A pseudogene', map_location='21q21.1', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=18857737, end_position_on_the_genomic_accession=18858476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379743, CurrentID=0, Status='live', Symbol='LOC105379743', Aliases='', description='uncharacterized LOC105379743', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379744, CurrentID=0, Status='live', Symbol='LOC105379744', Aliases='', description='uncharacterized LOC105379744', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391073 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391073, CurrentID=0, Status='live', Symbol='GAPDHP74', Aliases='', description='glyceraldehyde-3-phosphate dehydrogenase pseudogene 74', other_designations='', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119434460, end_position_on_the_genomic_accession=119435089, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379746, CurrentID=0, Status='live', Symbol='LOC105379746', Aliases='', description='uncharacterized LOC105379746', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391075, CurrentID=0, Status='live', Symbol='GAPDHP23', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 23', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119461835, end_position_on_the_genomic_accession=119463092, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653220 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653220, CurrentID=0, Status='live', Symbol='XAGE1B', Aliases='CT12.1, CT12.1A, CT12.1B, CTP9, GAGED2, XAGE-1, XAGE1, XAGE1A', description='X antigen family member 1B', other_designations='G antigen family D member 2|X antigen family, member 1A|X antigen family, member 1B|cancer/testis antigen 12.1|cancer/testis antigen family 12, member 1a|cancer/testis antigen family 12, member 1b|cancer/testis associated protein|protein XAGE-1', map_location='Xp11.22', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=52495668, end_position_on_the_genomic_accession=52500812, orientation='plus', exon_count=4, OMIM='300742,300743', no_hdr0=''),
  391077 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391077, CurrentID=0, Status='live', Symbol='GAPDHP58', Aliases='GAPDHL1, GAPDL1', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 58', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene|glyceraldehyde-3-phosphate dehydrogenase-like 1, pseudogene', map_location='1p12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119495789, end_position_on_the_genomic_accession=119496916, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379750 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379750, CurrentID=0, Status='live', Symbol='LOC105379750', Aliases='', description='uncharacterized LOC105379750', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379751 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379751, CurrentID=0, Status='live', Symbol='LOC105379751', Aliases='', description='uncharacterized LOC105379751', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379752, CurrentID=0, Status='live', Symbol='LOC105379752', Aliases='', description='uncharacterized LOC105379752', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391081 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391081, CurrentID=0, Status='live', Symbol='HSD3BP5', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta, pseudogene 5', other_designations='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 1 pseudogene', map_location='1p13.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=119601253, end_position_on_the_genomic_accession=119609449, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653226 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653226, CurrentID=0, Status='live', Symbol='SRP9P1', Aliases='SRP9L1', description='signal recognition particle 9 pseudogene 1', other_designations='hCG1781062|signal recognition particle 9-like 1', map_location='10q23.32', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=91806043, end_position_on_the_genomic_accession=91807533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379755 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379755, CurrentID=0, Status='live', Symbol='LOC105379755', Aliases='', description='uncharacterized LOC105379755', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379756 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379756, CurrentID=0, Status='live', Symbol='LOC105379756', Aliases='', description='uncharacterized LOC105379756', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379757 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379757, CurrentID=0, Status='live', Symbol='LOC105379757', Aliases='', description='uncharacterized LOC105379757', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379758, CurrentID=0, Status='live', Symbol='LOC105379758', Aliases='', description='uncharacterized LOC105379758', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379759 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379759, CurrentID=0, Status='live', Symbol='LOC105379759', Aliases='', description='uncharacterized LOC105379759', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653232 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653232, CurrentID=0, Status='live', Symbol='RPL15P3', Aliases='RPL15_3_658, bA511S14.1', description='ribosomal protein L15 pseudogene 3', other_designations='', map_location='6p24.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=12514074, end_position_on_the_genomic_accession=12514775, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379761 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379761, CurrentID=0, Status='live', Symbol='LOC105379761', Aliases='', description='uncharacterized LOC105379761', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653234 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653234, CurrentID=0, Status='live', Symbol='AGAP10P', Aliases='AGAP10, CTGLF10P, CTGLF7, bA144G6.2, bA358L16.1', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 10 pseudogene', other_designations='centaurin, gamma-like family, member 10 pseudogene|centaurin, gamma-like family, member 7', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=45678692, end_position_on_the_genomic_accession=45700553, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379763 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379763, CurrentID=0, Status='live', Symbol='LOC105379763', Aliases='', description='uncharacterized LOC105379763', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653236, CurrentID=0, Status='live', Symbol='RPL23AP49', Aliases='RPL23A_12_383', description='ribosomal protein L23a pseudogene 49', other_designations='', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75624686, end_position_on_the_genomic_accession=75625153, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379765, CurrentID=0, Status='live', Symbol='LOC105379765', Aliases='', description='uncharacterized LOC105379765', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653238 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653238, CurrentID=0, Status='live', Symbol='GTF2H2B', Aliases='', description='general transcription factor IIH subunit 2B (pseudogene)', other_designations='general transcription factor IIH, polypeptide 2B (pseudogene)', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=70415370, end_position_on_the_genomic_accession=70450362, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105379767 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379767, CurrentID=0, Status='live', Symbol='LOC105379767', Aliases='', description='uncharacterized LOC105379767', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653240 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653240, CurrentID=0, Status='live', Symbol='KRTAP4-11', Aliases='KAP4.11, KAP4.14, KRTAP4-14, KRTAP4.14', description='keratin associated protein 4-11', other_designations='keratin associated protein 4-14|keratin-associated protein 4-14|keratin-associated protein 4.11|keratin-associated protein 4.14|ultrahigh sulfur keratin-associated protein 4.14', map_location='17q21.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=41117181, end_position_on_the_genomic_accession=41118354, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379769, CurrentID=0, Status='live', Symbol='LOC105379769', Aliases='', description='uncharacterized LOC105379769', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128954 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128954, CurrentID=0, Status='live', Symbol='GAB4', Aliases='', description='GRB2 associated binding protein family member 4', other_designations='GRB2-associated binder 2-like protein|GRB2-associated binder 4|GRB2-associated binding protein family, member 4|growth factor receptor bound protein 2-associated protein 4', map_location='22q11.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=16961936, end_position_on_the_genomic_accession=17008278, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  391099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391099, CurrentID=0, Status='live', Symbol='KRT8P28', Aliases='', description='keratin 8 pseudogene 28', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=151949395, end_position_on_the_genomic_accession=151950990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379772 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379772, CurrentID=0, Status='live', Symbol='LOC105379772', Aliases='', description='uncharacterized LOC105379772', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379773, CurrentID=0, Status='live', Symbol='LOC105379773', Aliases='', description='uncharacterized LOC105379773', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391102 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391102, CurrentID=0, Status='live', Symbol='RPLP0P4', Aliases='RPLP0_1_100', description='ribosomal protein, large, P0 pseudogene 4', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=153225007, end_position_on_the_genomic_accession=153226072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653247, CurrentID=0, Status='live', Symbol='PRB2', Aliases='IB-9, PRPPRB1, Ps, cP7', description='proline-rich protein BstNI subfamily 2', other_designations='Basic proline-rich peptide P-E|con1 glycoprotein|proline-rich protein BstNI, subfamily-2 (parotid size variant)|salivary proline-rich protein 2', map_location='12p13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=11391540, end_position_on_the_genomic_accession=11395564, orientation='minus', exon_count=4, OMIM=168810, no_hdr0=''),
  391104 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391104, CurrentID=0, Status='live', Symbol='VHLL', Aliases='VHLP, VLP', description='von Hippel-Lindau tumor suppressor-like', other_designations='VHL pseudogene|VHL-like protein', map_location='1q22', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=156298624, end_position_on_the_genomic_accession=156299637, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379777, CurrentID=0, Status='live', Symbol='LOC105379777', Aliases='', description='uncharacterized LOC105379777', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391106, CurrentID=0, Status='live', Symbol='VDAC1P9', Aliases='', description='voltage-dependent anion channel 1 pseudogene 9', other_designations='', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=157724200, end_position_on_the_genomic_accession=157725026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391107, CurrentID=0, Status='live', Symbol='OR10K2', Aliases='OR1-4', description='olfactory receptor family 10 subfamily K member 2', other_designations='olfactory receptor OR1-4|olfactory receptor, family 10, subfamily K, member 2', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158419928, end_position_on_the_genomic_accession=158420866, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379780 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379780, CurrentID=0, Status='live', Symbol='LOC105379780', Aliases='', description='uncharacterized LOC105379780', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391109 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391109, CurrentID=0, Status='live', Symbol='OR10K1', Aliases='OR1-6', description='olfactory receptor family 10 subfamily K member 1', other_designations='olfactory receptor OR1-6|olfactory receptor, family 10, subfamily K, member 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158465562, end_position_on_the_genomic_accession=158466503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391110 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391110, CurrentID=0, Status='live', Symbol='OR10R3P', Aliases='OR1-9', description='olfactory receptor family 10 subfamily R member 3 pseudogene', other_designations='olfactory receptor OR1-9 pseudogene|olfactory receptor, family 10, subfamily R, member 3 pseudogene', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158491220, end_position_on_the_genomic_accession=158492158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379783, CurrentID=0, Status='live', Symbol='LOC105379783', Aliases='', description='uncharacterized LOC105379783', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391112 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391112, CurrentID=0, Status='live', Symbol='OR6Y1', Aliases='OR1-11, OR6Y2', description='olfactory receptor family 6 subfamily Y member 1', other_designations='olfactory receptor 6Y2|olfactory receptor OR1-11|olfactory receptor, family 6, subfamily Y, member 1', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158547128, end_position_on_the_genomic_accession=158548105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379785 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379785, CurrentID=0, Status='live', Symbol='LOC105379785', Aliases='', description='extensin-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391114 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391114, CurrentID=0, Status='live', Symbol='OR6K3', Aliases='OR1-18', description='olfactory receptor family 6 subfamily K member 3', other_designations='olfactory receptor OR1-18|olfactory receptor, family 6, subfamily K, member 3|seven transmembrane helix receptor', map_location='1q23.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=158717168, end_position_on_the_genomic_accession=158718115, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  653259 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653259, CurrentID=0, Status='live', Symbol='AGAP14P', Aliases='CTGLF11P, bA145E20.1', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 14, pseudogene', other_designations='Putative centaurin-gamma-like family member 11P|centaurin, gamma-like family, member 11 pseudogene', map_location='10q11.22', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46337224, end_position_on_the_genomic_accession=46359023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379788 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379788, CurrentID=0, Status='live', Symbol='LOC105379788', Aliases='', description='uncharacterized LOC105379788', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391117 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391117, CurrentID=0, Status='live', Symbol='OR10J2P', Aliases='', description='olfactory receptor family 10 subfamily J member 2 pseudogene', other_designations='olfactory receptor, family 10, subfamily J, member 2 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159279043, end_position_on_the_genomic_accession=159280000, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379790, CurrentID=0, Status='live', Symbol='LOC105379790', Aliases='', description='uncharacterized LOC105379790', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391119 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391119, CurrentID=0, Status='live', Symbol='OR10J7P', Aliases='', description='olfactory receptor family 10 subfamily J member 7 pseudogene', other_designations='olfactory receptor, family 10, subfamily J, member 7 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159351051, end_position_on_the_genomic_accession=159352023, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391120, CurrentID=0, Status='live', Symbol='OR10J9P', Aliases='', description='olfactory receptor family 10 subfamily J member 9 pseudogene', other_designations='olfactory receptor, family 10, subfamily J, member 9 pseudogene', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159405422, end_position_on_the_genomic_accession=159406408, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128977 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128977, CurrentID=0, Status='live', Symbol='C22orf39', Aliases='', description='chromosome 22 open reading frame 39', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=19440886, end_position_on_the_genomic_accession=19448232, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379794, CurrentID=0, Status='live', Symbol='LOC105379794', Aliases='', description='uncharacterized LOC105379794', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391123 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391123, CurrentID=0, Status='live', Symbol='VSIG8', Aliases='', description='V-set and immunoglobulin domain containing 8', other_designations='', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=159854316, end_position_on_the_genomic_accession=159862657, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  653268 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653268, CurrentID=0, Status='live', Symbol='AGAP7P', Aliases='AGAP-7, AGAP7, CTGLF4, bA109G10.1', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 7, pseudogene', other_designations='arf-GAP with GTPase, ANK repeat and PH domain-containing protein 7 pseudogene|centaurin, gamma-like family, member 4|putative Arf-GAP with GTPase, ANK repeat and PH domain-containing protein 7', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=46109495, end_position_on_the_genomic_accession=46131678, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  653269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653269, CurrentID=0, Status='live', Symbol='POTEI', Aliases='POTE2beta', description='POTE ankyrin domain family member I', other_designations='POTE ankyrin domain family, member I', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130462245, end_position_on_the_genomic_accession=130509355, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  391126 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391126, CurrentID=0, Status='live', Symbol='RPL26P12', Aliases='RPL26_4_113', description='ribosomal protein L26 pseudogene 12', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=165819342, end_position_on_the_genomic_accession=165819831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379799 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379799, CurrentID=0, Status='live', Symbol='LOC105379799', Aliases='', description='uncharacterized LOC105379799', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379800 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379800, CurrentID=0, Status='live', Symbol='LOC105379800', Aliases='', description='uncharacterized LOC105379800', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379801, CurrentID=0, Status='live', Symbol='LOC105379801', Aliases='', description='uncharacterized LOC105379801', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391130 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391130, CurrentID=0, Status='live', Symbol='RPS17P6', Aliases='RPS17_1_116', description='ribosomal protein S17 pseudogene 6', other_designations='', map_location='1q24.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=167162380, end_position_on_the_genomic_accession=167162862, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653275, CurrentID=0, Status='live', Symbol='CFC1B', Aliases='', description='cripto, FRL-1, cryptic family 1B', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130521094, end_position_on_the_genomic_accession=130527992, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105379804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379804, CurrentID=0, Status='live', Symbol='LOC105379804', Aliases='', description='uncharacterized LOC105379804', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  128989 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128989, CurrentID=0, Status='live', Symbol='TANGO2', Aliases='C22orf25', description='transport and golgi organization 2 homolog', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20017000, end_position_on_the_genomic_accession=20065926, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105379806 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379806, CurrentID=0, Status='live', Symbol='LOC105379806', Aliases='', description='uncharacterized LOC105379806', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391135, CurrentID=0, Status='live', Symbol='RPL4P3', Aliases='RPL4_1_122', description='ribosomal protein L4 pseudogene 3', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171683066, end_position_on_the_genomic_accession=171684492, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379808, CurrentID=0, Status='live', Symbol='LOC105379808', Aliases='', description='uncharacterized LOC105379808', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391137 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391137, CurrentID=0, Status='live', Symbol='RPS15P3', Aliases='RPS15_2_123', description='ribosomal protein S15 pseudogene 3', other_designations='', map_location='1q24.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=171801973, end_position_on_the_genomic_accession=171813970, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653282, CurrentID=0, Status='live', Symbol='CT47A7', Aliases='CT47.7', description='cancer/testis antigen family 47, member A7', other_designations='cancer/testis CT47 family, member 7', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=120953282, end_position_on_the_genomic_accession=120956600, orientation='minus', exon_count=4, OMIM=300786, no_hdr0=''),
  391139 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391139, CurrentID=0, Status='live', Symbol='GOT2P2', Aliases='GOT2L2', description='glutamic-oxaloacetic transaminase 2 pseudogene 2', other_designations='glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2) pseudogene|glutamic-oxaloacetic transaminase 2, mitochondrial (aspartate aminotransferase 2)(GOT2) pseudogene|glutamic-oxaloacetic transaminase 2-like 2 pseudogene', map_location='1q25.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=173140504, end_position_on_the_genomic_accession=173142372, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391140, CurrentID=0, Status='live', Symbol='RPL13P7', Aliases='RPL13_1_131', description='ribosomal protein L13 pseudogene 7', other_designations='', map_location='1q25.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=176231184, end_position_on_the_genomic_accession=176231830, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  128997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=128997, CurrentID=0, Status='live', Symbol='SLC9A3P2', Aliases='', description='solute carrier family 9, subfamily A (NHE3, cation proton antiporter 3), member 3 pseudogene 2', other_designations='solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 pseudogene 2|solute carrier family 9 (sodium/hydrogen exchanger), member 3 pseudogene 2', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=20652834, end_position_on_the_genomic_accession=20654727, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379814, CurrentID=0, Status='live', Symbol='LOC105379814', Aliases='', description='uncharacterized LOC105379814', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379815, CurrentID=0, Status='live', Symbol='LOC105379815', Aliases='', description='uncharacterized LOC105379815', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653288, CurrentID=0, Status='live', Symbol='TSPY7P', Aliases='TSPYP7', description='testis specific protein, Y-linked 7, pseudogene', other_designations='testis specific protein, Y-linked 7 (pseudogene)|testis specific protein, Y-linked pseudogene 7', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=9378076, end_position_on_the_genomic_accession=9380871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379817, CurrentID=0, Status='live', Symbol='LOC105379817', Aliases='', description='uncharacterized LOC105379817', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379818 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379818, CurrentID=0, Status='live', Symbol='LOC105379818', Aliases='', description='uncharacterized LOC105379818', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379819, CurrentID=0, Status='live', Symbol='LOC105379819', Aliases='', description='uncharacterized LOC105379819', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379820, CurrentID=0, Status='live', Symbol='LOC105379820', Aliases='', description='uncharacterized LOC105379820', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379821 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379821, CurrentID=0, Status='live', Symbol='LOC105379821', Aliases='', description='uncharacterized LOC105379821', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379822 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379822, CurrentID=0, Status='live', Symbol='LOC105379822', Aliases='', description='uncharacterized LOC105379822', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391151, CurrentID=0, Status='live', Symbol='RPL10P4', Aliases='', description='ribosomal protein L10 pseudogene', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=201978651, end_position_on_the_genomic_accession=201979011, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379824 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379824, CurrentID=0, Status='live', Symbol='LOC105379824', Aliases='', description='uncharacterized LOC105379824', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  130728 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130728, CurrentID=0, Status='live', Symbol='RPL7P13', Aliases='RPL7_4_218', description='ribosomal protein L7 pseudogene 13', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=49878612, end_position_on_the_genomic_accession=49879434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379826 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379826, CurrentID=0, Status='live', Symbol='LOC105379826', Aliases='', description='uncharacterized LOC105379826', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391155 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391155, CurrentID=0, Status='live', Symbol='KRT8P29', Aliases='', description='keratin 8 pseudogene 29', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=203872574, end_position_on_the_genomic_accession=203874187, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379828, CurrentID=0, Status='live', Symbol='LOC105379828', Aliases='', description='uncharacterized LOC105379828', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379829 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379829, CurrentID=0, Status='live', Symbol='LOC105379829', Aliases='', description='histone-lysine N-methyltransferase SETMAR-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391158 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391158, CurrentID=0, Status='live', Symbol='TFDP1P', Aliases='', description='transcription factor Dp-1 pseudogene', other_designations='', map_location='1q32.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=209232113, end_position_on_the_genomic_accession=209234517, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379831, CurrentID=0, Status='live', Symbol='LINC00475', Aliases='', description='long intergenic non-protein coding RNA 475', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391160 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391160, CurrentID=0, Status='live', Symbol='ARPC3P2', Aliases='', description='actin related protein 2/3 complex subunit 3 pseudogene 2', other_designations='actin related protein 2/3 complex, subunit 3 pseudogene 2|actin related protein 2/3 complex, subunit 3, 21kDa pseudogene', map_location='1q32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=211442254, end_position_on_the_genomic_accession=211443003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379833, CurrentID=0, Status='live', Symbol='LOC105379833', Aliases='', description='uncharacterized LOC105379833', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391162, CurrentID=0, Status='live', Symbol='GAPDHP24', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 24', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=214870546, end_position_on_the_genomic_accession=214871922, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379835, CurrentID=0, Status='live', Symbol='LOC105379835', Aliases='', description='uncharacterized LOC105379835', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653308, CurrentID=0, Status='live', Symbol='ASAH2B', Aliases='ASAH2C, ASAH2L, bA449O16.3, bA98I6.3', description='N-acylsphingosine amidohydrolase (non-lysosomal ceramidase) 2B', other_designations='ASAH2-like protein|putative inactive N-acylsphingosine amidohydrolase 2B|putative inactive non-lysosomal ceramidase B', map_location='10q11.23', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=50739899, end_position_on_the_genomic_accession=50757188, orientation='plus', exon_count=6, OMIM=610987, no_hdr0=''),
  391165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391165, CurrentID=0, Status='live', Symbol='RPS26P17', Aliases='RPS26_2_165', description='ribosomal protein S26 pseudogene 17', other_designations='', map_location='1q41', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=218338546, end_position_on_the_genomic_accession=218338831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379838 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379838, CurrentID=0, Status='live', Symbol='LOC105379838', Aliases='', description='uncharacterized LOC105379838', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379839 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379839, CurrentID=0, Status='live', Symbol='LOC105379839', Aliases='', description='uncharacterized LOC105379839', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391168, CurrentID=0, Status='live', Symbol='RPS3AP7', Aliases='RPS3A_5_169', description='ribosomal protein S3a pseudogene 7', other_designations='', map_location='1q42.12', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=226438532, end_position_on_the_genomic_accession=226439394, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129025 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129025, CurrentID=0, Status='live', Symbol='ZNF280A', Aliases="3'OY11.1, SUHW1, ZNF280, ZNF636", description='zinc finger protein 280A', other_designations='suppressor of hairy wing homolog 1|zinc finger protein 280|zinc finger protein 636', map_location='22q11.22', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=22513736, end_position_on_the_genomic_accession=22520281, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  653314 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653314, CurrentID=0, Status='live', Symbol='RPL19P9', Aliases='RPL19_7_643', description='ribosomal protein L19 pseudogene 9', other_designations='', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178055572, end_position_on_the_genomic_accession=178056268, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391171, CurrentID=0, Status='live', Symbol='TUBB8P10', Aliases='', description='tubulin beta 8 class VIII pseudogene 10', other_designations='tubulin, beta 8 class VIII pseudogene 10', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=227492851, end_position_on_the_genomic_accession=227494826, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653316 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653316, CurrentID=0, Status='live', Symbol='FAM153C', Aliases='NY-REN-7-like', description='family with sequence similarity 153 member C', other_designations='family with sequence similarity 153, member C, pseudogene', map_location='5q35.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=178006966, end_position_on_the_genomic_accession=178049087, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  105379845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379845, CurrentID=0, Status='live', Symbol='LOC105379845', Aliases='', description='uncharacterized LOC105379845', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391174 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391174, CurrentID=0, Status='live', Symbol='FAM96AP2', Aliases='', description='family with sequence similarity 96 member A pseudogene 2', other_designations='', map_location='1q42.13', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=228114631, end_position_on_the_genomic_accession=228115484, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653319, CurrentID=0, Status='live', Symbol='KIAA0895L', Aliases='', description='KIAA0895-like', other_designations='', map_location='16q22.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=67175602, end_position_on_the_genomic_accession=67183980, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  105379848 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379848, CurrentID=0, Status='live', Symbol='LOC105379848', Aliases='', description='uncharacterized LOC105379848', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379849 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379849, CurrentID=0, Status='live', Symbol='LOC105379849', Aliases='', description='uncharacterized LOC105379849', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379850, CurrentID=0, Status='live', Symbol='LOC105379850', Aliases='', description='uncharacterized LOC105379850', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391179 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391179, CurrentID=0, Status='live', Symbol='KRT18P32', Aliases='', description='keratin 18 pseudogene 32', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=238491305, end_position_on_the_genomic_accession=238492701, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379852, CurrentID=0, Status='live', Symbol='LOC105379852', Aliases='', description='uncharacterized LOC105379852', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391181, CurrentID=0, Status='live', Symbol='RPL6P3', Aliases='RPL6_2_187', description='ribosomal protein L6 pseudogene 3', other_designations='', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241831909, end_position_on_the_genomic_accession=241832827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379854, CurrentID=0, Status='live', Symbol='LOC105379854', Aliases='', description='uncharacterized LOC105379854', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391183 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391183, CurrentID=0, Status='live', Symbol='RSL24D1P4', Aliases='RSL24D1P19', description='ribosomal L24 domain containing 1 pseudogene 4', other_designations='ribosomal L24 domain containing 1 pseudogene 19', map_location='1q43', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=242772568, end_position_on_the_genomic_accession=242775305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379856, CurrentID=0, Status='live', Symbol='LOC105379856', Aliases='', description='uncharacterized LOC105379856', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379857, CurrentID=0, Status='live', Symbol='LOC105379857', Aliases='', description='uncharacterized LOC105379857', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379858 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379858, CurrentID=0, Status='live', Symbol='LOC105379858', Aliases='', description='uncharacterized LOC105379858', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  152579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152579, CurrentID=0, Status='live', Symbol='SCFD2', Aliases='STXBP1L1', description='sec1 family domain containing 2', other_designations='syntaxin binding protein 1-like 1|syntaxin-binding protein 1-like 1', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=52872982, end_position_on_the_genomic_accession=53366111, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105379860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379860, CurrentID=0, Status='live', Symbol='LOC105379860', Aliases='', description='uncharacterized LOC105379860', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391189 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391189, CurrentID=0, Status='live', Symbol='OR11L1', Aliases='', description='olfactory receptor family 11 subfamily L member 1', other_designations='olfactory receptor OR1-50|olfactory receptor, family 11, subfamily L, member 1', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247840928, end_position_on_the_genomic_accession=247841896, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391190 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391190, CurrentID=0, Status='live', Symbol='OR2L8', Aliases='', description='olfactory receptor family 2 subfamily L member 8 (gene/pseudogene)', other_designations='olfactory receptor OR1-46|olfactory receptor, family 2, subfamily L, member 8', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247948858, end_position_on_the_genomic_accession=247949796, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391191 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391191, CurrentID=0, Status='live', Symbol='OR2AK2', Aliases='OR1-47, OR2AK1P', description='olfactory receptor family 2 subfamily AK member 2', other_designations='olfactory receptor 2AK1|olfactory receptor OR1-47|olfactory receptor, family 2, subfamily AK, member 1 pseudogene|olfactory receptor, family 2, subfamily AK, member 2', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=247965332, end_position_on_the_genomic_accession=247966339, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391192 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391192, CurrentID=0, Status='live', Symbol='OR2L3', Aliases='', description='olfactory receptor family 2 subfamily L member 3', other_designations='olfactory receptor, family 2, subfamily L, member 3', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248060682, end_position_on_the_genomic_accession=248061620, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  129049 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129049, CurrentID=0, Status='live', Symbol='SGSM1', Aliases='RUTBC2', description='small G protein signaling modulator 1', other_designations='RUN and TBC1 domain containing 2|RUN and TBC1 domain-containing protein 2|small G protein signaling modulator 1 protein', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=24806221, end_position_on_the_genomic_accession=24926846, orientation='plus', exon_count=26, OMIM=611417, no_hdr0=''),
  391194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391194, CurrentID=0, Status='live', Symbol='OR2M2', Aliases='OR2M2Q, OST423', description='olfactory receptor family 2 subfamily M member 2', other_designations='olfactory receptor, family 2, subfamily M, member 2', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248179986, end_position_on_the_genomic_accession=248181029, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391195 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391195, CurrentID=0, Status='live', Symbol='OR2T33', Aliases='OR1-56', description='olfactory receptor family 2 subfamily T member 33', other_designations='novel 7 transmembrane receptor (rhodopsin family) protein|olfactory receptor OR1-56|olfactory receptor, family 2, subfamily T, member 33', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248272852, end_position_on_the_genomic_accession=248273814, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391196, CurrentID=0, Status='live', Symbol='OR2M7', Aliases='OR1-58', description='olfactory receptor family 2 subfamily M member 7', other_designations='olfactory receptor OR1-58|olfactory receptor, family 2, subfamily M, member 7', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248323630, end_position_on_the_genomic_accession=248324568, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105379869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379869, CurrentID=0, Status='live', Symbol='LOC105379869', Aliases='', description='uncharacterized LOC105379869', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379870, CurrentID=0, Status='live', Symbol='LOC105379870', Aliases='', description='uncharacterized LOC105379870', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101060400 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060400, CurrentID=0, Status='live', Symbol='LOC101060400', Aliases='', description='uncharacterized LOC101060400', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48634804, end_position_on_the_genomic_accession=48647072, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105379872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379872, CurrentID=0, Status='live', Symbol='LOC105379872', Aliases='', description='uncharacterized LOC105379872', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379873 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379873, CurrentID=0, Status='live', Symbol='LOC105379873', Aliases='', description='uncharacterized LOC105379873', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379874 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379874, CurrentID=0, Status='live', Symbol='LOC105379874', Aliases='', description='uncharacterized LOC105379874', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379875, CurrentID=0, Status='live', Symbol='LOC105379875', Aliases='', description='uncharacterized LOC105379875', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379876 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379876, CurrentID=0, Status='live', Symbol='LOC105379876', Aliases='', description='uncharacterized LOC105379876', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379877, CurrentID=0, Status='live', Symbol='LOC105379877', Aliases='', description='uncharacterized LOC105379877', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379878, CurrentID=0, Status='live', Symbol='LOC105379878', Aliases='', description='uncharacterized LOC105379878', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379879, CurrentID=0, Status='live', Symbol='LOC105379879', Aliases='', description='uncharacterized LOC105379879', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379880, CurrentID=0, Status='live', Symbol='LOC105379880', Aliases='', description='uncharacterized LOC105379880', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379881, CurrentID=0, Status='live', Symbol='LOC105379881', Aliases='', description='uncharacterized LOC105379881', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653354 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653354, CurrentID=0, Status='live', Symbol='ARMCX7P', Aliases='ARMCX6-ps, GASP10ps', description='armadillo repeat containing, X-linked 7, pseudogene', other_designations='armadillo repeat containing, X-linked 6 pseudogene', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101597371, end_position_on_the_genomic_accession=101598068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391211 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391211, CurrentID=0, Status='live', Symbol='OR2G6', Aliases='', description='olfactory receptor family 2 subfamily G member 6', other_designations='olfactory receptor, family 2, subfamily G, member 6', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248521647, end_position_on_the_genomic_accession=248522597, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  105379884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379884, CurrentID=0, Status='live', Symbol='LOC105379884', Aliases='', description='uncharacterized LOC105379884', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379885 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379885, CurrentID=0, Status='live', Symbol='LOC105379885', Aliases='', description='uncharacterized LOC105379885', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105379886 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379886, CurrentID=0, Status='live', Symbol='LOC105379886', Aliases='', description='putative uncharacterized protein encoded by LINC00174', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653361 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653361, CurrentID=0, Status='live', Symbol='NCF1', Aliases='NCF1A, NOXO2, SH3PXD1A, p47phox', description='neutrophil cytosolic factor 1', other_designations='47 kDa autosomal chronic granulomatous disease protein|47 kDa neutrophil oxidase factor|NADPH oxidase organizer 2|NCF-1|NCF-47K|SH3 and PX domain-containing protein 1A|neutrophil NADPH oxidase factor 1|neutrophil cytosolic factor 1, (chronic granulomatous disease, autosomal 1)|nox organizer 2|nox-organizing protein 2|p47-phox', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=74773962, end_position_on_the_genomic_accession=74789376, orientation='plus', exon_count=13, OMIM=608512, no_hdr0=''),
  653363 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653363, CurrentID=0, Status='live', Symbol='TEX28P2', Aliases='CXorf2B, pTEX', description='testis expressed 28 pseudogene 2', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154159207, end_position_on_the_genomic_accession=154180155, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653364, CurrentID=0, Status='live', Symbol='HSPB1P2', Aliases='HSPBL3', description='heat shock protein family B (small) member 1 pseudogene 2', other_designations='heat shock 27kDa protein 1 pseudogene 2|heat shock 27kDa protein-like 3 pseudogene', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=49233800, end_position_on_the_genomic_accession=49234540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129080 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129080, CurrentID=0, Status='live', Symbol='EMID1', Aliases='EMI5, EMU1', description='EMI domain containing 1', other_designations='emilin and multimerin domain-containing protein 1', map_location='22q12.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=29205912, end_position_on_the_genomic_accession=29259597, orientation='plus', exon_count=23, OMIM=608926, no_hdr0=''),
  101054525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101054525, CurrentID=0, Status='live', Symbol='LOC101054525', Aliases='', description='progesterone receptor antisense RNA', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=101129077, end_position_on_the_genomic_accession=101159270, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  391230 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391230, CurrentID=0, Status='live', Symbol='LOC391230', Aliases='', description='KN motif and ankyrin repeat domain-containing protein 1-like', other_designations='', map_location='20p12.3', chromosome='20', genomic_nucleotide_accession_version='NC_000020.10', start_position_on_the_genomic_accession=5908717, end_position_on_the_genomic_accession=5915051, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  653380 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653380, CurrentID=0, Status='live', Symbol='LOC653380', Aliases='', description='TBC1 domain family member 3C-like protein ENSP00000341742', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653381, CurrentID=0, Status='live', Symbol='SORD2P', Aliases='SORD2', description='sorbitol dehydrogenase 2, pseudogene', other_designations='SORD pseudogene|sorbitol dehydrogenase pseudogene', map_location='15q21.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=44825499, end_position_on_the_genomic_accession=44847563, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391239 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391239, CurrentID=0, Status='live', Symbol='LOC391239', Aliases='', description='V-set domain containing T cell activation inhibitor 1 pseudogene', other_designations='', map_location='20p11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=25099105, end_position_on_the_genomic_accession=25099925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101929996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929996, CurrentID=0, Status='live', Symbol='LOC101929996', Aliases='', description='uncharacterized LOC101929996', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=183543604, end_position_on_the_genomic_accession=183557201, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  391242 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391242, CurrentID=0, Status='live', Symbol='PUDPP3', Aliases='HDHD1DP, HDHD1P3', description="pseudouridine 5'-phosphatase pseudogene 3", other_designations='HDHD family member 1D, pseudogene|haloacid dehalogenase-like hydrolase domain containing 1 pseudogene 3', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33047520, end_position_on_the_genomic_accession=33048194, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129099 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129099, CurrentID=0, Status='live', Symbol='CPSF1P1', Aliases='dJ90G24.5', description='cleavage and polyadenylation specific factor 1 pseudogene 1', other_designations='cleavage and polyadenylation specific factor 1, 160kDa pseudogene 1', map_location='22q12.3', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=32269381, end_position_on_the_genomic_accession=32273194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  106146146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106146146, CurrentID=0, Status='live', Symbol='LINC01079', Aliases='TCONS_00021533', description='long intergenic non-protein coding RNA 1079', other_designations='', map_location='13q12.2', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653390, CurrentID=0, Status='live', Symbol='RRN3P2', Aliases='', description='RRN3 homolog, RNA polymerase I transcription factor pseudogene 2', other_designations='RNA polymerase I transcription factor homolog (S. cerevisiae) pseudogene 2|RRN3 RNA polymerase I transcription factor homolog pseudogene|RRN3-like protein 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29074842, end_position_on_the_genomic_accession=29116718, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100422669 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422669, CurrentID=0, Status='live', Symbol='LOC100422669', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide pseudogene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=45052780, end_position_on_the_genomic_accession=45053490, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139599 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139599, CurrentID=0, Status='live', Symbol='MAGEE2', Aliases='HCA3', description='MAGE family member E2', other_designations='MAGE-E2 antigen|hepatocellular carcinoma-associated protein 3|hepatocellular carcinoma-associated protein HCA3|melanoma antigen family E, 2|melanoma antigen family E2', map_location='Xq13.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=75782988, end_position_on_the_genomic_accession=75785244, orientation='minus', exon_count=1, OMIM=300760, no_hdr0=''),
  391253 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391253, CurrentID=0, Status='live', Symbol='SPINT4', Aliases='C20orf137, SPINT3, dJ601O1.1', description='serine peptidase inhibitor, Kunitz type 4', other_designations='serine protease inhibitor, Kunitz type 1|serine protease inhibitor, Kunitz type 4', map_location='20q13.12', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=45722349, end_position_on_the_genomic_accession=45725696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100137047 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100137047, CurrentID=0, Status='live', Symbol='JMJD7', Aliases='', description='jumonji domain containing 7', other_designations='jumonji domain-containing protein 7', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=41828085, end_position_on_the_genomic_accession=41837587, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  391256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391256, CurrentID=0, Status='live', Symbol='KRT18P4', Aliases='dJ963K23.1', description='keratin 18 pseudogene 4', other_designations='', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49956740, end_position_on_the_genomic_accession=49958087, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391257 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391257, CurrentID=0, Status='live', Symbol='SUMO1P1', Aliases='PIC1L, UBL2, UBL6', description='SUMO1 pseudogene 1', other_designations='ubiquitin-like 2|ubiquitin-like 6', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=53874498, end_position_on_the_genomic_accession=53875709, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422671 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422671, CurrentID=0, Status='live', Symbol='LOC100422671', Aliases='', description='cell division cycle 73, Paf1/RNA polymerase II complex component, homolog (S. cerevisiae) pseudogene', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=79854684, end_position_on_the_genomic_accession=79855635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653404, CurrentID=0, Status='live', Symbol='FOXD4L6', Aliases='', description='forkhead box D4-like 6', other_designations='FOXD4-like 6', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=41126251, end_position_on_the_genomic_accession=41128975, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  653406 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653406, CurrentID=0, Status='live', Symbol='LOC653406', Aliases='', description='NLR family, apoptosis inhibitory protein pseudogene', other_designations='', map_location='5q13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=71101256, end_position_on_the_genomic_accession=71128756, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379771, CurrentID=0, Status='live', Symbol='LOC105379771', Aliases='', description='uncharacterized LOC105379771', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105371042 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371042, CurrentID=0, Status='live', Symbol='LOC105371042', Aliases='', description='uncharacterized LOC105371042', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=1126624, end_position_on_the_genomic_accession=1134385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106821730 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106821730, CurrentID=0, Status='live', Symbol='BUB1B-PAK6', Aliases='', description='BUB1B-PAK6 readthrough', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391267, CurrentID=0, Status='live', Symbol='ANKRD20A11P', Aliases='C21orf81', description='ankyrin repeat domain 20 family member A11, pseudogene', other_designations='ankyrin repeat domain 20 family, member A11, pseudogene|ankyrin repeat domain 20 family, member A3 pseudogene|ankyrin repeat domain 20 family, member A4 pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=13943769, end_position_on_the_genomic_accession=13980444, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  391269 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391269, CurrentID=0, Status='live', Symbol='ANKRD20A18P', Aliases='C21orf126', description='ankyrin repeat domain 20 family member A18, pseudogene', other_designations='ankyrin repeat domain 20 family, member A18, pseudogene|ankyrin repeat domain 20 family, member A3 pseudogene', map_location='21q11.2', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=14064219, end_position_on_the_genomic_accession=14069088, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  401471 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401471, CurrentID=0, Status='live', Symbol='LOC401471', Aliases='', description='uncharacterized LOC401471', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=101008330, end_position_on_the_genomic_accession=101009860, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  391276 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391276, CurrentID=0, Status='live', Symbol='THUMPD1P1', Aliases='', description='THUMP domain containing 1 pseudogene 1', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=28901756, end_position_on_the_genomic_accession=28903768, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  152594 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152594, CurrentID=0, Status='live', Symbol='LOC152594', Aliases='', description='NIN1/RPN12 binding protein 1 pseudogene', other_designations='NIN1/RPN12 binding protein 1 homolog (S. cerevisiae) pseudogene', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=138560647, end_position_on_the_genomic_accession=138562387, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653423 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653423, CurrentID=0, Status='live', Symbol='SPAG11A', Aliases='EDDM2A, HE2', description='sperm associated antigen 11A', other_designations='epididymal protein 2A|human epididymis-specific protein 2|sperm antigen HE2', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7847864, end_position_on_the_genomic_accession=7863797, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  129138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129138, CurrentID=0, Status='live', Symbol='ANKRD54', Aliases='LIAR', description='ankyrin repeat domain 54', other_designations='lyn-interacting ankyrin repeat protein', map_location='22q13.1', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=37830854, end_position_on_the_genomic_accession=37848621, orientation='minus', exon_count=10, OMIM=613383, no_hdr0=''),
  653427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653427, CurrentID=0, Status='live', Symbol='FOXD4L5', Aliases='bA15J10.2', description='forkhead box D4-like 5', other_designations='FOXD4-like 5', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=65282101, end_position_on_the_genomic_accession=65285209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100130760 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130760, CurrentID=0, Status='live', Symbol='LOC100130760', Aliases='', description='jumonji domain containing 4 pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129045808, end_position_on_the_genomic_accession=129047000, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653435, CurrentID=0, Status='live', Symbol='GUSBP6', Aliases='', description='glucuronidase, beta pseudogene 6', other_designations='glucuronidase, beta pseudogene', map_location='7q11.21', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=64100209, end_position_on_the_genomic_accession=64120538, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653436, CurrentID=0, Status='live', Symbol='ANKRD20A7P', Aliases='', description='ankyrin repeat domain 20 family member A7, pseudogene', other_designations='ankyrin repeat domain 20 family, member A2 pseudogene|ankyrin repeat domain 20 family, member A7, pseudogene', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42852563, end_position_on_the_genomic_accession=42895739, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653437, CurrentID=0, Status='live', Symbol='AQP12B', Aliases='INSSA3', description='aquaporin 12B', other_designations='AQP-12B|insulin synthesis associated 3', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240676416, end_position_on_the_genomic_accession=240685768, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  653440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653440, CurrentID=0, Status='live', Symbol='WASH6P', Aliases='CXYorf1, FAM39A, WASH', description='WAS protein family homolog 6 pseudogene', other_designations='actin nucleation promoting factor|family with sequence similarity 39, member A', map_location='Xq28;Yq12', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=57208178, end_position_on_the_genomic_accession=57212192, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653441 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653441, CurrentID=0, Status='live', Symbol='PHC1P1', Aliases='PHC1B', description='polyhomeotic homolog 1 (Drosophila) pseudogene 1', other_designations='polyhomeotic homolog 1B', map_location='12q13.2', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=55409686, end_position_on_the_genomic_accession=55414943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653447, CurrentID=0, Status='live', Symbol='HNRNPCP2', Aliases='', description='heterogeneous nuclear ribonucleoprotein C pseudogene 2', other_designations='heterogeneous nuclear ribonucleoprotein C (C1/C2) pseudogene', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=189923169, end_position_on_the_genomic_accession=189924890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100268168 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100268168, CurrentID=0, Status='live', Symbol='LOC100268168', Aliases='', description='uncharacterized LOC100268168', other_designations='CTC-308K20.1', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=172954782, end_position_on_the_genomic_accession=172959368, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  653453 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653453, CurrentID=0, Status='live', Symbol='LOC653453', Aliases='', description='fibroblast growth factor 7 pseudogene', other_designations='', map_location='9p11.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=42625430, end_position_on_the_genomic_accession=42629639, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653458, CurrentID=0, Status='live', Symbol='BMS1P9', Aliases='', description='BMS1, ribosome biogenesis factor pseudogene 9', other_designations='BMS1 homolog, ribosome assembly protein pseudogene|BMS1 pseudogene 9|BMS1 ribosome biogenesis factor pseudogene 9', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63055646, end_position_on_the_genomic_accession=63059217, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653464 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653464, CurrentID=0, Status='live', Symbol='SRGAP2C', Aliases='SRGAP2P1', description='SLIT-ROBO Rho GTPase activating protein 2C', other_designations='SLIT-ROBO Rho GTPase activating protein 2 pseudogene 1', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121184967, end_position_on_the_genomic_accession=121392824, orientation='plus', exon_count=13, OMIM=614704, no_hdr0=''),
  391322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391322, CurrentID=0, Status='live', Symbol='LOC391322', Aliases='', description='D-dopachrome tautomerase-like', other_designations='', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391334, CurrentID=0, Status='live', Symbol='LOC391334', Aliases='', description='actin, gamma 2, smooth muscle, enteric pseudogene', other_designations='', map_location='22q13.2', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=41074167, end_position_on_the_genomic_accession=41075230, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653479, CurrentID=0, Status='live', Symbol='MRPL45P2', Aliases='', description='mitochondrial ribosomal protein L45 pseudogene 2', other_designations='', map_location='17q21.32', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=47450227, end_position_on_the_genomic_accession=47492620, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  653483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653483, CurrentID=0, Status='live', Symbol='MLLT4-AS1', Aliases='C6orf124, HGC6.4, dJ431P23.3', description='MLLT4 antisense RNA 1 (head to head)', other_designations='MLLT4 antisense RNA 1 (non-protein coding)', map_location='6q27', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=167823890, end_position_on_the_genomic_accession=167826796, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  653486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653486, CurrentID=0, Status='live', Symbol='SCGB1C2', Aliases='', description='secretoglobin, family 1C, member 2', other_designations='secretoglobin, family 1C, member 1-like', map_location='17p', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=132364, end_position_on_the_genomic_accession=139067, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100422685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422685, CurrentID=0, Status='live', Symbol='LOC100422685', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 6 pseudogene', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=136788284, end_position_on_the_genomic_accession=136788792, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653489, CurrentID=0, Status='live', Symbol='RGPD3', Aliases='RGP3', description='RANBP2-like and GRIP domain containing 3', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=106391360, end_position_on_the_genomic_accession=106468345, orientation='minus', exon_count=28, OMIM=612706, no_hdr0=''),
  653492 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653492, CurrentID=0, Status='live', Symbol='PSG10P', Aliases='PSG10, PSG12', description='pregnancy specific beta-1-glycoprotein 10, pseudogene', other_designations='pregnancy specific beta-1-glycoprotein 1 pseudogene|pregnancy specific beta-1-glycoprotein 12|pregnancy specific beta-1-glycoprotein 120|pregnancy-specific beta-1 glycoprotein', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=42836997, end_position_on_the_genomic_accession=42855718, orientation='minus', exon_count=7, OMIM=176399, no_hdr0=''),
  391349 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391349, CurrentID=0, Status='live', Symbol='LOC391349', Aliases='', description='phosphatidylinositol-4,5-bisphosphate 3-kinase, catalytic subunit delta pseudogene', other_designations='', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6495880, end_position_on_the_genomic_accession=6496798, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653499, CurrentID=0, Status='live', Symbol='LGALS7B', Aliases='GAL7, Gal-7, HKL-14, LGALS7, PI7', description='lectin, galactoside-binding, soluble, 7B', other_designations='galectin 7B|galectin-7B|keratinocyte lectin 14|lectin galactoside-binding soluble 7|p53-induced gene 1 protein|p53-induced protein 1', map_location='19q13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=38789210, end_position_on_the_genomic_accession=38791754, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  391356 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391356, CurrentID=0, Status='live', Symbol='PTRHD1', Aliases='C2orf79', description='peptidyl-tRNA hydrolase domain containing 1', other_designations='peptidyl-tRNA hydrolase domain-containing protein 1', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=24790267, end_position_on_the_genomic_accession=24793382, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100034743 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100034743, CurrentID=0, Status='live', Symbol='PDZK1P1', Aliases='OTTHUMT00000038524, PDZK1P2', description='PDZ domain containing 1 pseudogene 1', other_designations='PDZ domain containing 1 pseudogene 2', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147993862, end_position_on_the_genomic_accession=148014956, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  391358 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391358, CurrentID=0, Status='live', Symbol='TRMT112P6', Aliases='', description='tRNA methyltransferase 11-2 homolog (S. cerevisiae) pseudogene 6', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26028033, end_position_on_the_genomic_accession=26028624, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653503, CurrentID=0, Status='live', Symbol='LOC653503', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 10 pseudogene', other_designations='GNG10 pseudogene', map_location='11p15.4', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=10271211, end_position_on_the_genomic_accession=10272288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653505, CurrentID=0, Status='live', Symbol='PPIAL4A', Aliases='COAS-2, COAS2, PPIAL4, PPIAL4B', description='peptidylprolyl isomerase A (cyclophilin A)-like 4A', other_designations='PPIase A-like 4B|PPIase A-like 4G|chromosome 1-amplified sequence 2|chromosome one-amplified sequence 2|cyclophilin homolog overexpressed in liver cancer|cyclophilin-LC (COAS2)|peptidyl-prolyl cis-trans isomerase A-like 4A/B/C|peptidylprolyl cis-trans isomerase A-like 4A|peptidylprolyl cis-trans isomerase A-like 4B|peptidylprolyl isomerase A (cyclophilin A)-like 4B', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=120889746, end_position_on_the_genomic_accession=120890530, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101928945 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928945, CurrentID=0, Status='live', Symbol='LOC101928945', Aliases='', description='uracil-DNA glycosylase pseudogene', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=51275209, end_position_on_the_genomic_accession=51279401, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391365, CurrentID=0, Status='live', Symbol='SULT6B1', Aliases='', description='sulfotransferase family 6B member 1', other_designations='ST6B1|sulfotransferase SULT6B1|sulfotransferase family, cytosolic, 6B, member 1|thyroxine sulfotransferase', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=37167820, end_position_on_the_genomic_accession=37188547, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  260294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260294, CurrentID=0, Status='live', Symbol='NSUN5P2', Aliases='NOL1R2, NSUN5C, WBSCR20B, WBSCR20C', description='NOP2/Sun RNA methyltransferase family member 5 pseudogene 2', other_designations='NOL1/NOP2/Sun domain family, member 5C|NOP2/Sun domain family, member 5 pseudogene 2|NOP2/Sun domain family, member 5C (pseudogene)|Williams Beuren syndrome chromosome region 20C', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=72948293, end_position_on_the_genomic_accession=72954763, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  391367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391367, CurrentID=0, Status='live', Symbol='GAPDHP25', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 25', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='2p22.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38285340, end_position_on_the_genomic_accession=38286595, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653513 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653513, CurrentID=0, Status='live', Symbol='LOC653513', Aliases='', description='phosphodiesterase 4D interacting protein-like', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149245804, end_position_on_the_genomic_accession=149322070, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  391370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391370, CurrentID=0, Status='live', Symbol='RPS12P4', Aliases='RPS12_1_210', description='ribosomal protein S12 pseudogene 4', other_designations='', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=41850181, end_position_on_the_genomic_accession=41850683, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414754 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414754, CurrentID=0, Status='live', Symbol='ACTG1P11', Aliases='', description='actin gamma 1 pseudogene 11', other_designations='actin, gamma 1 pseudogene 11|actin, gamma pseudogene', map_location='Yq11.221', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18147810, end_position_on_the_genomic_accession=18149718, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653519 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653519, CurrentID=0, Status='live', Symbol='GPR89A', Aliases='GPHR, GPR89, GPR89B, SH120, UNQ192', description='G protein-coupled receptor 89A', other_designations='protein GPR89|putative MAPK-activating protein PM01|putative NF-kappa-B-activating protein 90', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=145607960, end_position_on_the_genomic_accession=145670977, orientation='plus', exon_count=17, OMIM=612821, no_hdr0=''),
  391378 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391378, CurrentID=0, Status='live', Symbol='HNRNPA1P66', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 66', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=63751694, end_position_on_the_genomic_accession=63752619, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  283683 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283683, CurrentID=0, Status='live', Symbol='LOC283683', Aliases='', description='uncharacterized LOC283683', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=22757857, end_position_on_the_genomic_accession=22778737, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  391383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391383, CurrentID=0, Status='live', Symbol='LOC391383', Aliases='', description='WD repeat domain 4 pseudogene', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=68447891, end_position_on_the_genomic_accession=68449186, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391387, CurrentID=0, Status='live', Symbol='RPSAP28', Aliases='RPSA_4_242', description='ribosomal protein SA pseudogene 28', other_designations='', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=73369951, end_position_on_the_genomic_accession=73370963, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422693 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422693, CurrentID=0, Status='live', Symbol='LOC100422693', Aliases='', description='polypeptide N-acetylgalactosaminyltransferase 1 pseudogene', other_designations='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 1 (GalNAc-T1) pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60063004, end_position_on_the_genomic_accession=60064009, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130161, CurrentID=0, Status='live', Symbol='LOC100130161', Aliases='', description='developmental pluripotency associated 4 pseudogene', other_designations='', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=10166969, end_position_on_the_genomic_accession=10167819, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653541 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653541, CurrentID=0, Status='live', Symbol='LOC653541', Aliases='', description='double homeobox, 4-like', other_designations='', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.10', start_position_on_the_genomic_accession=191223374, end_position_on_the_genomic_accession=191224345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653543 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653543, CurrentID=0, Status='live', Symbol='DUX4L7', Aliases='', description='double homeobox 4 like 7', other_designations='double homeobox protein 4-like protein 7', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190071232, end_position_on_the_genomic_accession=190072516, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653544 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653544, CurrentID=0, Status='live', Symbol='DUX4L6', Aliases='', description='double homeobox 4 like 6', other_designations='double homeobox protein 4-like protein 6', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190074525, end_position_on_the_genomic_accession=190075809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653545, CurrentID=0, Status='live', Symbol='DUX4L5', Aliases='', description='double homeobox 4 like 5', other_designations='double homeobox protein 4-like protein 5', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190077818, end_position_on_the_genomic_accession=190079102, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  260330 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260330, CurrentID=0, Status='live', Symbol='ALPPP', Aliases='', description='alkaline phosphatase, pseudogene', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  260331 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260331, CurrentID=0, Status='live', Symbol='LOC260331', Aliases='ECEL1P1', description='endothelin-converting enzyme-like 1-like', other_designations='endothelin converting enzyme-like 1, pseudogene 1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653548, CurrentID=0, Status='live', Symbol='DUX4L3', Aliases='', description='double homeobox 4 like 3', other_designations='double homeobox protein 4-like protein 3', map_location='4q35.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=190087705, end_position_on_the_genomic_accession=190088989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391405 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391405, CurrentID=0, Status='live', Symbol='LOC391405', Aliases='', description='immunoglobulin kappa light chain variable region', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.11', start_position_on_the_genomic_accession=92222523, end_position_on_the_genomic_accession=92260784, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653550 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653550, CurrentID=0, Status='live', Symbol='TP53TG3C', Aliases='TP53TG3, TP53TG3E, TP53TG3F', description='TP53 target 3C', other_designations='TP53-inducible gene 3 protein', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=33194264, end_position_on_the_genomic_accession=33196858, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  260335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260335, CurrentID=0, Status='live', Symbol='ELF2P2', Aliases='', description='E74-like factor 2 pseudogene 2', other_designations='E74-like factor 2 (ets domain transcription factor) pseudogene|ELF2P2 pseudogene', map_location='6p25.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=1512341, end_position_on_the_genomic_accession=1515565, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260336 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260336, CurrentID=0, Status='live', Symbol='ELF2P3', Aliases='', description='E74-like factor 2 pseudogene 3', other_designations='E74-like factor 2 (ets domain transcription factor) pseudogene|ELF2P3 pseudogene', map_location='9q21', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87816474, end_position_on_the_genomic_accession=87819454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653553 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653553, CurrentID=0, Status='live', Symbol='HSPB1P1', Aliases='HSPBL2', description='heat shock protein family B (small) member 1 pseudogene 1', other_designations='heat shock 27kD protein-like 2|heat shock 27kDa protein 1 pseudogene 1|heat shock 27kDa protein-like 2 pseudogene', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72007759, end_position_on_the_genomic_accession=72008135, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260338 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260338, CurrentID=0, Status='live', Symbol='TUBB8P4', Aliases='', description='tubulin beta 8 class VIII pseudogene 4', other_designations='HSA12cenq11 beta-tubulin 4Q (TUBB4Q) pseudogene|tubulin, beta 2B class IIb pseudogene|tubulin, beta 2B pseudogene|tubulin, beta 8 class VIII pseudogene 4', map_location='12p11.1', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=34164773, end_position_on_the_genomic_accession=34166954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260339 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260339, CurrentID=0, Status='live', Symbol='LOC260339', Aliases='', description='transcription factor A, mitochondrial pseudogene', other_designations='processed pseudogene mtTFA 3', map_location='6q22.31', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=121478167, end_position_on_the_genomic_accession=121479288, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260340, CurrentID=0, Status='live', Symbol='TFAMP2', Aliases='MTTF1, TCF6L3', description='transcription factor A, mitochondrial pseudogene 2', other_designations='processed pseudogene mtTFA 2|transcription factor 6-like 3 (mitochondrial transcription factor 1-like)', map_location='11q22.3', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=109905922, end_position_on_the_genomic_accession=109907866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260341 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260341, CurrentID=0, Status='live', Symbol='TFAMP1', Aliases='MTTF1, TCF6L1', description='transcription factor A, mitochondrial pseudogene 1', other_designations='processed pseudogene mtTFA 1|transcription factor 6-like 1 (mitochondrial transcription factor 1-like)', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1614470, end_position_on_the_genomic_accession=1616692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  260342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260342, CurrentID=0, Status='live', Symbol='COPS8P1', Aliases='BA345E19.2', description='COP9 signalosome subunit 8 pseudogene 1', other_designations='COP9 constitutive photomorphogenic homolog subunit 8 pseudogene', map_location='Xq21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=86798601, end_position_on_the_genomic_accession=86799226, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653562, CurrentID=0, Status='live', Symbol='SLC6A10PB', Aliases='', description='solute carrier family 6 (neurotransmitter transporter), member 8 pseudogene', other_designations='sodium- and chloride-dependent creatine transporter 1-like|solute carrier family 6 (neurotransmitter transporter, creatine), member 8 pseudogene', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653566, CurrentID=0, Status='live', Symbol='SPCS2P4', Aliases='', description='signal peptidase complex subunit 2 homolog (S. cerevisiae) pseudogene 4', other_designations='signal peptidase complex subunit 2 homolog pseudogene', map_location='1p35.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=28095035, end_position_on_the_genomic_accession=28096458, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653567 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653567, CurrentID=0, Status='live', Symbol='TMEM236', Aliases='FAM23A, FAM23B, bA162I21.2, bA16O1.2', description='transmembrane protein 236', other_designations='family with sequence similarity 23, member A|family with sequence similarity 23, member B', map_location='10p12.33', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=17752210, end_position_on_the_genomic_accession=17800887, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  391428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391428, CurrentID=0, Status='live', Symbol='SEPHS1P7', Aliases='', description='selenophosphate synthetase 1 pseudogene 7', other_designations='', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=114260049, end_position_on_the_genomic_accession=114260840, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129285 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129285, CurrentID=0, Status='live', Symbol='PPP1R21', Aliases='CCDC128, KLRAQ1', description='protein phosphatase 1 regulatory subunit 21', other_designations='KLRAQ motif containing 1|KLRAQ motif-containing protein 1|coiled-coil domain containing 128|coiled-coil domain-containing protein 128|protein phosphatase 1, regulatory subunit 21', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=48440769, end_position_on_the_genomic_accession=48515392, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  283692 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283692, CurrentID=0, Status='live', Symbol='CPEB1-AS1', Aliases='', description='CPEB1 antisense RNA 1', other_designations='', map_location='15q25.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=82647770, end_position_on_the_genomic_accession=82692820, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  653579 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653579, CurrentID=0, Status='live', Symbol='CD177P1', Aliases='CD177P', description='CD177 molecule pseudogene 1', other_designations='CD177 pseudogene', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43371891, end_position_on_the_genomic_accession=43379125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391436, CurrentID=0, Status='live', Symbol='MED15P5', Aliases='', description='mediator complex subunit 15 pseudogene 5', other_designations='', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130251502, end_position_on_the_genomic_accession=130252737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129293 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129293, CurrentID=0, Status='live', Symbol='TRABD2A', Aliases='C2orf89, TIKI1', description='TraB domain containing 2A', other_designations='TRAB domain-containing protein 2A|UPF0632 protein C2orf89', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=84821650, end_position_on_the_genomic_accession=84881975, orientation='minus', exon_count=9, OMIM=614912, no_hdr0=''),
  653583 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653583, CurrentID=0, Status='live', Symbol='PHLDB3', Aliases='', description='pleckstrin homology-like domain, family B, member 3', other_designations='', map_location='19q13.31', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=43474341, end_position_on_the_genomic_accession=43504920, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  653588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653588, CurrentID=0, Status='live', Symbol='LOC653588', Aliases='', description='Sjogren syndrome antigen B (autoantigen La) pseudogene', other_designations='', map_location='Xq11.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=62779123, end_position_on_the_genomic_accession=62780626, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  414766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414766, CurrentID=0, Status='live', Symbol='HCG16', Aliases='HCG15, dJ25J6.7', description='HLA complex group 16 (non-protein coding)', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=28984949, end_position_on_the_genomic_accession=29008251, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  653590 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653590, CurrentID=0, Status='live', Symbol='GGTLC5P', Aliases='GGT', description='gamma-glutamyltransferase light chain 5 pseudogene', other_designations='', map_location='22q11.21', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=18340163, end_position_on_the_genomic_accession=18346065, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129303 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129303, CurrentID=0, Status='live', Symbol='TMEM150A', Aliases='TM6P1, TMEM150', description='transmembrane protein 150A', other_designations='fasting-inducible integral membrane protein TM6P1|transmembrane protein 150', map_location='2p11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=85598547, end_position_on_the_genomic_accession=85602699, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  391448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391448, CurrentID=0, Status='live', Symbol='LOC391448', Aliases='', description='mannosidase, endo-alpha-like pseudogene', other_designations='', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=135897955, end_position_on_the_genomic_accession=135899166, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653598, CurrentID=0, Status='live', Symbol='PPIAL4C', Aliases='COAS-2', description='peptidylprolyl isomerase A (cyclophilin A)-like 4C', other_designations='PPIase A-like 4A|PPIase A-like 4C|chromosome 1-amplified sequence 2|chromosome one-amplified sequence 2|cyclophilin homolog overexpressed in liver cancer', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149583823, end_position_on_the_genomic_accession=149584607, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391458 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391458, CurrentID=0, Status='live', Symbol='KRT18P46', Aliases='', description='keratin 18 pseudogene 46', other_designations='', map_location='2q24.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=161580070, end_position_on_the_genomic_accession=161581408, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653604 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653604, CurrentID=0, Status='live', Symbol='HIST2H3D', Aliases='', description='histone cluster 2, H3d', other_designations='histone 2, H3d', map_location='1q21.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=149813225, end_position_on_the_genomic_accession=149813681, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  653606 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653606, CurrentID=0, Status='live', Symbol='PRAMEF22', Aliases='PRAMEF3L', description='PRAME family member 22', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13005048, end_position_on_the_genomic_accession=13005624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100130766 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130766, CurrentID=0, Status='live', Symbol='LOC100130766', Aliases='', description='metastasis associated 1 family member 3 pseudogene', other_designations='metastasis associated 1 family, member 3 pseudogene', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157961374, end_position_on_the_genomic_accession=157962900, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391465, CurrentID=0, Status='live', Symbol='LOC391465', Aliases='', description='STIP1 homology and U-box containing protein 1, E3 ubiquitin protein ligase pseudogene', other_designations='', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=177177697, end_position_on_the_genomic_accession=177178458, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391467 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391467, CurrentID=0, Status='live', Symbol='IMPDH1P7', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 7', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=187001139, end_position_on_the_genomic_accession=187003449, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391470 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391470, CurrentID=0, Status='live', Symbol='AHCYP5', Aliases='', description='adenosylhomocysteinase pseudogene 5', other_designations='', map_location='2q32.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=195582793, end_position_on_the_genomic_accession=195584031, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391472 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391472, CurrentID=0, Status='live', Symbol='BICD1P1', Aliases='', description='bicaudal D homolog 1 (Drosophila) pseudogene 1', other_designations='', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200844305, end_position_on_the_genomic_accession=200845409, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  260402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260402, CurrentID=0, Status='live', Symbol='IS1', Aliases='AIS, AIS1', description='Adolescent idiopathic scoliosis', other_designations='', map_location='19p13.3', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=181800, no_hdr0=''),
  653619 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653619, CurrentID=0, Status='live', Symbol='PRAMEF15', Aliases='', description='PRAME family member 15', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=13315581, end_position_on_the_genomic_accession=13322595, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  260404 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260404, CurrentID=0, Status='live', Symbol='LOC260404', Aliases='', description='DKFZp434F1819 related gene', other_designations='', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101447996 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101447996, CurrentID=0, Status='live', Symbol='RNVU1-8', Aliases='RNU1-145, vU1.8', description='RNA, variant U1 small nuclear 8', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=147084616, end_position_on_the_genomic_accession=147084756, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391485, CurrentID=0, Status='live', Symbol='KRT8P30', Aliases='', description='keratin 8 pseudogene 30', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218953401, end_position_on_the_genomic_accession=218953983, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  260415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260415, CurrentID=0, Status='live', Symbol='SCASI', Aliases='SCA24, SCAR4', description='spinocerebellar ataxia with saccadic intrusions', other_designations='spinocerebellar ataxia 24 (with saccadic intrusions)', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607317, no_hdr0=''),
  391490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391490, CurrentID=0, Status='live', Symbol='RPL23AP26', Aliases='RPL23A_14_334', description='ribosomal protein L23a pseudogene 26', other_designations='', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231585864, end_position_on_the_genomic_accession=231590927, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653635 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653635, CurrentID=0, Status='live', Symbol='WASH7P', Aliases='FAM39F, WASH5P', description='WAS protein family homolog 7 pseudogene', other_designations='WAS protein family homolog 5 pseudogene|family with sequence similarity 39, member F', map_location='1p36.33', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=14362, end_position_on_the_genomic_accession=29370, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100422710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422710, CurrentID=0, Status='live', Symbol='LOC100422710', Aliases='', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 1 pseudogene', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=233545482, end_position_on_the_genomic_accession=233546942, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  260422 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260422, CurrentID=0, Status='live', Symbol='LOC260422', Aliases='', description='actin related protein 2/3 complex subunit 1A pseudogene', other_designations='actin related protein 2/3 complex, subunit 1A, 41kDa pseudogene', map_location='4q21.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=86912928, end_position_on_the_genomic_accession=86914917, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653639 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653639, CurrentID=0, Status='live', Symbol='LYPLA2P1', Aliases='APT, LYPLA2L, dJ570F3.6', description='lysophospholipase II pseudogene 1', other_designations='dJ570F3.6 (novel protein similar to lysophospholipase II (LYPLA2))|lysophospholipase 2 like', map_location='6p21.32', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=33364724, end_position_on_the_genomic_accession=33366362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391496 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391496, CurrentID=0, Status='live', Symbol='OR5S1P', Aliases='', description='olfactory receptor family 5 subfamily S member 1 pseudogene', other_designations='olfactory receptor, family 5, subfamily S, member 1 pseudogene', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=240109102, end_position_on_the_genomic_accession=240110049, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653641 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653641, CurrentID=0, Status='live', Symbol='GOLGA6C', Aliases='', description='golgin A6 family member C', other_designations='Golgin subfamily A member 6C|golgi autoantigen, golgin subfamily a, 6C|golgin A6 family, member C', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75258558, end_position_on_the_genomic_accession=75273455, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  653643 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653643, CurrentID=0, Status='live', Symbol='GOLGA6D', Aliases='', description='golgin A6 family member D', other_designations='golgi autoantigen, golgin subfamily a, 6D|golgin A6 family, member D', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75282841, end_position_on_the_genomic_accession=75295807, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  260429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260429, CurrentID=0, Status='live', Symbol='PRSS33', Aliases='EOS', description='protease, serine 33', other_designations='protease, serine, 33|serine protease EOS', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2783953, end_position_on_the_genomic_accession=2787607, orientation='minus', exon_count=7, OMIM=613797, no_hdr0=''),
  260431 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260431, CurrentID=0, Status='live', Symbol='COPD', Aliases='', description='Pulmonary disease, chronic obstructive, severe early-onset', other_designations='', map_location='2q', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606963, no_hdr0=''),
  260432 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260432, CurrentID=0, Status='live', Symbol='SLSN3', Aliases='', description='Senior-Loken syndrome 3', other_designations='', map_location='3q22', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=606995, no_hdr0=''),
  414776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414776, CurrentID=0, Status='live', Symbol='HCG19P', Aliases='', description='HLA complex group 19 pseudogene', other_designations='', map_location='6p21.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  260434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260434, CurrentID=0, Status='live', Symbol='PYDC1', Aliases='ASC2, POP1, PYC1, cPOP1', description='PYD (pyrin domain) containing 1', other_designations='PAAD-only protein 1|cellular POP1|pyrin-only protein 1', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31215962, end_position_on_the_genomic_accession=31217093, orientation='minus', exon_count=2, OMIM=615700, no_hdr0=''),
  260436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=260436, CurrentID=0, Status='live', Symbol='FDCSP', Aliases='C4orf7, FDC-SP', description='follicular dendritic cell secreted protein', other_designations='FDC secreted protein', map_location='4q13', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=70226071, end_position_on_the_genomic_accession=70235252, orientation='plus', exon_count=5, OMIM=607241, no_hdr0=''),
  653653 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653653, CurrentID=0, Status='live', Symbol='LOC653653', Aliases='', description='adaptor-related protein complex 1 sigma 2 subunit pseudogene', other_designations='adaptor-related protein complex 1, sigma 2 subunit pseudogene', map_location='17q23.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=60101760, end_position_on_the_genomic_accession=60102919, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101930041 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930041, CurrentID=0, Status='live', Symbol='LOC101930041', Aliases='', description='UDP-glucuronosyltransferase 2B10-like', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  653656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653656, CurrentID=0, Status='live', Symbol='MBD3L4', Aliases='', description='methyl-CpG binding domain protein 3-like 4', other_designations='MBD3-like 4|MBD3-like protein 4|Putative methyl-CpG-binding domain protein 3-like 4', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7037744, end_position_on_the_genomic_accession=7040826, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  653657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653657, CurrentID=0, Status='live', Symbol='MBD3L3', Aliases='', description='methyl-CpG binding domain protein 3-like 3', other_designations='MBD3-like 3|MBD3-like protein 3', map_location='19p13.2', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=7056205, end_position_on_the_genomic_accession=7069615, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  653658 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653658, CurrentID=0, Status='live', Symbol='RPS23P8', Aliases='RPS23_5_1783', description='ribosomal protein S23 pseudogene 8', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=70962962, end_position_on_the_genomic_accession=70963342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653659 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653659, CurrentID=0, Status='live', Symbol='TMEM183B', Aliases='C1orf37-DUP', description='transmembrane protein 183B', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149981661, end_position_on_the_genomic_accession=149983366, orientation='minus', exon_count=1, OMIM=611365, no_hdr0=''),
  391518 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391518, CurrentID=0, Status='live', Symbol='VENTXP7', Aliases='HPX42, VENTX1', description='VENT homeobox pseudogene 7', other_designations='VENT-like homeobox 1|VENT-like homeobox pseudogene 7', map_location='3p24.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=21405726, end_position_on_the_genomic_accession=21406708, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653665 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653665, CurrentID=0, Status='live', Symbol='RPL21P93', Aliases='RPL21_36_1027', description='ribosomal protein L21 pseudogene 93', other_designations='', map_location='10p12.1', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=28898950, end_position_on_the_genomic_accession=28899524, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391524 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391524, CurrentID=0, Status='live', Symbol='THRAP3P1', Aliases='THRAP3L', description='thyroid hormone receptor associated protein 3 pseudogene 1', other_designations='', map_location='3p24.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=31452643, end_position_on_the_genomic_accession=31456091, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391526 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391526, CurrentID=0, Status='live', Symbol='IGBP1P3', Aliases='', description='immunoglobulin (CD79A) binding protein 1 pseudogene 3', other_designations='', map_location='3p22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=32620700, end_position_on_the_genomic_accession=32621914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391530, CurrentID=0, Status='live', Symbol='SALL4P6', Aliases='', description='spalt-like transcription factor 4 pseudogene 6', other_designations='sal-like 4 pseudogene 6', map_location='3p22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=42321005, end_position_on_the_genomic_accession=42322692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391532 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391532, CurrentID=0, Status='live', Symbol='LOC391532', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75746727, end_position_on_the_genomic_accession=75747439, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653677 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653677, CurrentID=0, Status='live', Symbol='SEC1P', Aliases='SEC1', description='secretory blood group 1, pseudogene', other_designations='alpha(1,2) fucosyltransferase pseudogene', map_location='19q13.33', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=48638039, end_position_on_the_genomic_accession=48682245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  103283057 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103283057, CurrentID=0, Status='live', Symbol='LINC01362', Aliases='', description='long intergenic non-protein coding RNA 1362', other_designations='', map_location='1p31.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=82903183, end_position_on_the_genomic_accession=82997583, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  391538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391538, CurrentID=0, Status='live', Symbol='ALDOAP1', Aliases='', description='aldolase, fructose-bisphosphate A pseudogene 1', other_designations='aldolase A, fructose-bisphosphate pseudogene 1', map_location='3p21.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=52193045, end_position_on_the_genomic_accession=52194991, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391539, CurrentID=0, Status='live', Symbol='RPS25P4', Aliases='RPS25_2_371', description='ribosomal protein S25 pseudogene 4', other_designations='', map_location='3p21.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=53474657, end_position_on_the_genomic_accession=53475148, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653687 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653687, CurrentID=0, Status='live', Symbol='FAM226B', Aliases='CXorf50B, LINC00246B, NCRNA00246B', description='family with sequence similarity 226 member B (non-protein coding)', other_designations='family with sequence similarity 226, member B (non-protein coding)|hCG1731871|long intergenic non-protein coding RNA 246B|non-protein coding RNA 246B', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72777073, end_position_on_the_genomic_accession=72779095, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  653689 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653689, CurrentID=0, Status='live', Symbol='GSTT2B', Aliases='GSTT2P', description='glutathione S-transferase theta 2B (gene/pseudogene)', other_designations='GST class-theta-2|glutathione S-transferase theta 2|glutathione S-transferase theta-2', map_location='22q11.23', chromosome='22', genomic_nucleotide_accession_version='NC_000022.11', start_position_on_the_genomic_accession=23957414, end_position_on_the_genomic_accession=23961234, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  129402 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129402, CurrentID=0, Status='live', Symbol='LIN28AP1', Aliases='', description='LIN28A pseudogene 1', other_designations='lin-28 homolog A pseudogene 1', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=183260642, end_position_on_the_genomic_accession=183261263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653691 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653691, CurrentID=0, Status='live', Symbol='TCAF2P1', Aliases='FAM115D, FAM139B', description='TRPM8 channel-associated factor 2 pseudogene 1', other_designations='family with sequence similarity 115, member D (pseudogene)|family with sequence similarity 139, member B', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143809990, end_position_on_the_genomic_accession=143815389, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  391548 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391548, CurrentID=0, Status='live', Symbol='OR7E55P', Aliases='OR2DG, OR3.2, OR7E56P, OR7E57P, OST007, OST013', description='olfactory receptor family 7 subfamily E member 55 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 55 pseudogene|olfactory receptor, family 7, subfamily E, member 56 pseudogene|olfactory receptor, family 7, subfamily E, member 57 pseudogene', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75370412, end_position_on_the_genomic_accession=75371362, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391552, CurrentID=0, Status='live', Symbol='AGGF1P3', Aliases='', description='angiogenic factor with G-patch and FHA domains 1 pseudogene 3', other_designations='angiogenic factor with G patch and FHA domains 1 pseudogene 3', map_location='3p12.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=75653450, end_position_on_the_genomic_accession=75655581, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653698, CurrentID=0, Status='live', Symbol='LOC653698', Aliases='', description='solute carrier family 25 (pyrimidine nucleotide carrier), member 33 pseudogene', other_designations='', map_location='11p11.12', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=49433427, end_position_on_the_genomic_accession=49434662, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391556 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391556, CurrentID=0, Status='live', Symbol='LOC391556', Aliases='', description='retinoblastoma binding protein 4 pseudogene', other_designations='retinoblastoma binding protein 7 pseudogene', map_location='3q11.2', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=96566295, end_position_on_the_genomic_accession=96567576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101448070 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101448070, CurrentID=0, Status='live', Symbol='GLC1P', Aliases='', description='glaucoma 1, open angle, P', other_designations='', map_location='12q14', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615141, no_hdr0=''),
  101448071 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101448071, CurrentID=0, Status='live', Symbol='BWQTL4', Aliases='', description='Birth weight QTL4', other_designations='', map_location='6p22.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615192, no_hdr0=''),
  101448072 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101448072, CurrentID=0, Status='live', Symbol='RLS8', Aliases='', description='Restless legs syndrome, susceptibility to, 8', other_designations='', map_location='5q31', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=615197, no_hdr0=''),
  391562 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391562, CurrentID=0, Status='live', Symbol='LOC391562', Aliases='', description='ezrin pseudogene', other_designations='', map_location='3q13.11', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=104502701, end_position_on_the_genomic_accession=104504116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653707 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653707, CurrentID=0, Status='live', Symbol='PDCL2P1', Aliases='', description='phosducin-like 2 pseudogene 1', other_designations='', map_location='Xp22.11', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=23198318, end_position_on_the_genomic_accession=23198710, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391566 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391566, CurrentID=0, Status='live', Symbol='LOC391566', Aliases='', description='histone cluster 1, H2bj pseudogene', other_designations='', map_location='3q13.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=114103247, end_position_on_the_genomic_accession=114103628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653712, CurrentID=0, Status='live', Symbol='LOC653712', Aliases='', description='intraflagellar transport 122 homolog (Chlamydomonas) pseudogene', other_designations='', map_location='3q21.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=128861508, end_position_on_the_genomic_accession=128871541, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  653720 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653720, CurrentID=0, Status='live', Symbol='GOLGA8M', Aliases='', description='golgin A8 family member M', other_designations='Putative golgin subfamily A member 6-like protein 7|golgin A8 family, member M', map_location='15q13.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=28698695, end_position_on_the_genomic_accession=28718510, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  391578 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391578, CurrentID=0, Status='live', Symbol='LOC391578', Aliases='', description='MAF1 homolog (S. cerevisiae) pseudogene', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=133490806, end_position_on_the_genomic_accession=133491560, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  65263 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=65263, CurrentID=0, Status='live', Symbol='PYCRL', Aliases='PYCR3', description='pyrroline-5-carboxylate reductase-like', other_designations='P5C reductase 3|P5CR 3|pyrroline-5-carboxylate reductase-like protein', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=143603913, end_position_on_the_genomic_accession=143609614, orientation='minus', exon_count=6, OMIM=616408, no_hdr0=''),
  391581 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391581, CurrentID=0, Status='live', Symbol='RPL31P23', Aliases='RPL31_5_407', description='ribosomal protein L31 pseudogene 23', other_designations='', map_location='3q22.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=136205489, end_position_on_the_genomic_accession=136205943, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930053, CurrentID=0, Status='live', Symbol='LOC101930053', Aliases='', description='uncharacterized LOC101930053', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2423858, end_position_on_the_genomic_accession=2494223, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  391584 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391584, CurrentID=0, Status='live', Symbol='KRT18P35', Aliases='', description='keratin 18 pseudogene 35', other_designations='', map_location='3q23', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=141470548, end_position_on_the_genomic_accession=141471950, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391587 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391587, CurrentID=0, Status='live', Symbol='LOC391587', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='3q25.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=149984254, end_position_on_the_genomic_accession=149984989, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391589, CurrentID=0, Status='live', Symbol='KRT18P34', Aliases='', description='keratin 18 pseudogene 34', other_designations='', map_location='3q25.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=157162609, end_position_on_the_genomic_accession=157163934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129446 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129446, CurrentID=0, Status='live', Symbol='XIRP2', Aliases='CMYA3', description='xin actin binding repeat containing 2', other_designations='beta-xin|cardiomyopathy associated 3|cardiomyopathy-associated protein 3|myomaxin|xeplin|xin actin-binding repeat containing 2|xin repeat protein 2', map_location='2q24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=166888487, end_position_on_the_genomic_accession=167259753, orientation='plus', exon_count=13, OMIM=609778, no_hdr0=''),
  391592 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391592, CurrentID=0, Status='live', Symbol='BZW1P1', Aliases='', description='basic leucine zipper and W2 domains 1 pseudogene 1', other_designations='basic leucine zipper and W2 domains 2 pseudogene', map_location='3q26.31', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=172424195, end_position_on_the_genomic_accession=172427097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653737 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653737, CurrentID=0, Status='live', Symbol='RPL21P119', Aliases='RPL21_56_1471', description='ribosomal protein L21 pseudogene 119', other_designations='', map_location='16p13.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=9156342, end_position_on_the_genomic_accession=9156921, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129450 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129450, CurrentID=0, Status='live', Symbol='TYW5', Aliases='C2orf60, hTYW5', description='tRNA-yW synthesizing protein 5', other_designations='jmjC domain-containing protein C2orf60|tRNA yW-synthesizing enzyme 5|tRNA(Phe) (7-(3-amino-3-carboxypropyl)wyosine(37)-C(2))-hydroxylase', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=199928911, end_position_on_the_genomic_accession=199955736, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  391595 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391595, CurrentID=0, Status='live', Symbol='RPL32P10', Aliases='RPL32_3_439', description='ribosomal protein L32 pseudogene 10', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180887670, end_position_on_the_genomic_accession=180888075, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391598, CurrentID=0, Status='live', Symbol='RPSAP31', Aliases='RPSA_11_442', description='ribosomal protein SA pseudogene 31', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=183887522, end_position_on_the_genomic_accession=183888537, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391600 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391600, CurrentID=0, Status='live', Symbol='LOC391600', Aliases='', description='transmembrane epididymal protein 1 pseudogene', other_designations='', map_location='3q27.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=184499582, end_position_on_the_genomic_accession=184500273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391603 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391603, CurrentID=0, Status='live', Symbol='NMNAT1P3', Aliases='', description='NMNAT1 pseudogene 3', other_designations='nicotinamide nucleotide adenylyltransferase 1 pseudogene', map_location='3q28', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=190166764, end_position_on_the_genomic_accession=190167629, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653753 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653753, CurrentID=0, Status='live', Symbol='VN1R6P', Aliases='VNR19I2, ZVNH2, ZVNR2', description='vomeronasal 1 receptor 6 pseudogene', other_designations='vomeronasal olfactory receptor, (chromosome 19) subtype I, member 2', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53315137, end_position_on_the_genomic_accession=53316043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653758 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653758, CurrentID=0, Status='live', Symbol='TPM3P6', Aliases='', description='tropomyosin 3 pseudogene 6', other_designations='', map_location='19q13.42', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=53479282, end_position_on_the_genomic_accession=53480365, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930059 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930059, CurrentID=0, Status='live', Symbol='LOC101930059', Aliases='', description='uncharacterized LOC101930059', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106480539 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106480539, CurrentID=0, Status='live', Symbol='RNU4ATAC7P', Aliases='', description='RNA, U4atac small nuclear 7, pseudogene', other_designations='', map_location='20q13.2', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391622, CurrentID=0, Status='live', Symbol='USP17L6P', Aliases='DUB4, USP17L6, vDUB4', description='ubiquitin specific peptidase 17-like family member 6, pseudogene', other_designations='chorionic villi deubiquitinating enzyme|deubiquitinating enzyme DUB4|ubiquitin specific peptidase 17-like 6 (pseudogene)|ubiquitin specific peptidase 17-like 6, pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9367874, end_position_on_the_genomic_accession=9369070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  391627 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391627, CurrentID=0, Status='live', Symbol='USP17L9P', Aliases='RS447, USP17, USP17A, USP17H, USP17I, USP17J, USP17K, USP17L, USP17M', description='ubiquitin specific peptidase 17-like family member 9, pseudogene', other_designations='deubiquitinating enzyme 17|ubiquitin carboxyl-terminal hydrolase 17|ubiquitin specific peptidase 17|ubiquitin specific protease 17|ubiquitin thioesterase 17|ubiquitin thiolesterase 17|ubiquitin-specific-processing protease 17', map_location='4p15', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9358382, end_position_on_the_genomic_accession=9359974, orientation='plus', exon_count=1, OMIM=607011, no_hdr0=''),
  653773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653773, CurrentID=0, Status='live', Symbol='RPL31P17', Aliases='RPL31_4_330', description='ribosomal protein L31 pseudogene 17', other_designations='', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222726430, end_position_on_the_genomic_accession=222726875, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391630 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391630, CurrentID=0, Status='live', Symbol='OR7E83P', Aliases='OR7E134P', description='olfactory receptor family 7 subfamily E member 83 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 83 pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9512823, end_position_on_the_genomic_accession=9513987, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391632 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391632, CurrentID=0, Status='live', Symbol='OR7E35P', Aliases='OR7E120, OST018', description='olfactory receptor family 7 subfamily E member 35 pseudogene', other_designations='olfactory receptor, family 7, subfamily A, member 120|olfactory receptor, family 7, subfamily E, member 35 pseudogene', map_location='4p16.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=9754893, end_position_on_the_genomic_accession=9755806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391634 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391634, CurrentID=0, Status='live', Symbol='HSP90AB2P', Aliases='HSP90BB', description='heat shock protein 90kDa alpha family class B member 2, pseudogene', other_designations='heat shock protein 90Bb|heat shock protein 90kDa alpha (cytosolic), class B member 2, pseudogene|heat shock protein HSP 90-beta', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13336387, end_position_on_the_genomic_accession=13338931, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391636, CurrentID=0, Status='live', Symbol='LOC391636', Aliases='', description='chromosome 9 open reading frame 78 pseudogene', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=13977166, end_position_on_the_genomic_accession=13978025, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653781, CurrentID=0, Status='live', Symbol='POTEJ', Aliases='POTE2beta', description='POTE ankyrin domain family member J', other_designations='POTE ankyrin domain family, member J', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=130611413, end_position_on_the_genomic_accession=130658448, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  105375409 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375409, CurrentID=0, Status='live', Symbol='LOC105375409', Aliases='', description='uncharacterized LOC105375409', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=95660322, end_position_on_the_genomic_accession=95719917, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  653784 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653784, CurrentID=0, Status='live', Symbol='MZT2A', Aliases='FAM128A, MOZART2A', description='mitotic spindle organizing protein 2A', other_designations='family with sequence similarity 128, member A|mitotic-spindle organizing protein associated with a ring of gamma-tubulin 2A', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131469724, end_position_on_the_genomic_accession=131492719, orientation='minus', exon_count=7, OMIM=613449, no_hdr0=''),
  391642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391642, CurrentID=0, Status='live', Symbol='IGBP1P5', Aliases='', description='immunoglobulin (CD79A) binding protein 1 pseudogene 5', other_designations='immunoglobulin (CD79A) binding protein 1 pseudogene pseudogene', map_location='4p15.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=27585134, end_position_on_the_genomic_accession=27586305, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653789 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653789, CurrentID=0, Status='live', Symbol='RPL23AP25', Aliases='RPL23A_7_192', description='ribosomal protein L23a pseudogene 25', other_designations='', map_location='1q44', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248936704, end_position_on_the_genomic_accession=248937069, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391647 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391647, CurrentID=0, Status='live', Symbol='KRT18P25', Aliases='', description='keratin 18 pseudogene 25', other_designations='', map_location='4p14', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=40020192, end_position_on_the_genomic_accession=40021578, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653794 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653794, CurrentID=0, Status='live', Symbol='TRIM60P14', Aliases='', description='tripartite motif containing 60 pseudogene 14', other_designations='tripartite motif-containing 60 pseudogene 14', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164915566, end_position_on_the_genomic_accession=164916980, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391655 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391655, CurrentID=0, Status='live', Symbol='RPL17P20', Aliases='RPL17_10_481', description='ribosomal protein L17 pseudogene 20', other_designations='', map_location='4q12', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=56710021, end_position_on_the_genomic_accession=56710646, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391656, CurrentID=0, Status='live', Symbol='RPS15AP17', Aliases='RPS15A_8_485', description='ribosomal protein S15a pseudogene 17', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=62105620, end_position_on_the_genomic_accession=62106079, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391657 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391657, CurrentID=0, Status='live', Symbol='DPP3P1', Aliases='', description='dipeptidyl-peptidase 3 pseudogene 1', other_designations='', map_location='4q13.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=64430868, end_position_on_the_genomic_accession=64433434, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  729217 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=729217, CurrentID=0, Status='live', Symbol='LOC729217', Aliases='', description='emopamil binding protein-like pseudogene', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=64260326, end_position_on_the_genomic_accession=64261241, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  653805 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653805, CurrentID=0, Status='live', Symbol='RPS27P8', Aliases='RPS27_1_148', description='ribosomal protein S27 pseudogene 8', other_designations='', map_location='1q32.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=202471806, end_position_on_the_genomic_accession=202472150, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928955 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928955, CurrentID=0, Status='live', Symbol='LOC101928955', Aliases='', description='uncharacterized LOC101928955', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=62894681, end_position_on_the_genomic_accession=62899543, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  653808 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653808, CurrentID=0, Status='live', Symbol='ZG16', Aliases='JCLN, JCLN1A, ZG16', description='zymogen granule protein 16', other_designations='jacalin-like lectin domain containing|secretory lectin ZG16|zymogen granule protein 16 homolog', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=29760827, end_position_on_the_genomic_accession=29781648, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  129521 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129521, CurrentID=0, Status='live', Symbol='NMS', Aliases='', description='neuromedin S', other_designations='prepro-NMS', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100470482, end_position_on_the_genomic_accession=100483280, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  129522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129522, CurrentID=0, Status='live', Symbol='LOC129522', Aliases='', description='ralA binding protein 1 pseudogene', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=100424688, end_position_on_the_genomic_accession=100426661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391670 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391670, CurrentID=0, Status='live', Symbol='HNRNPA1P56', Aliases='', description='heterogeneous nuclear ribonucleoprotein A1 pseudogene 56', other_designations='', map_location='4q21.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=77987509, end_position_on_the_genomic_accession=77988874, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391672, CurrentID=0, Status='live', Symbol='IGBP1P4', Aliases='', description='immunoglobulin (CD79A) binding protein 1 pseudogene 4', other_designations='', map_location='4q21.22', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=82401419, end_position_on_the_genomic_accession=82402607, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129530 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129530, CurrentID=0, Status='live', Symbol='LYG1', Aliases='SALW1939', description='lysozyme G-like 1', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99284238, end_position_on_the_genomic_accession=99304742, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  129531 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129531, CurrentID=0, Status='live', Symbol='MITD1', Aliases='', description='microtubule interacting and trafficking domain containing 1', other_designations='MIT, microtubule interacting and transport, domain containing 1', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=99161427, end_position_on_the_genomic_accession=99181061, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  653820 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653820, CurrentID=0, Status='live', Symbol='FAM72B', Aliases='p17', description='family with sequence similarity 72 member B', other_designations='RP11-439A17.6|amyloid-beta peptide-induced protein p17|family with sequence similarity 72, member B', map_location='1p11.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=121167646, end_position_on_the_genomic_accession=121186896, orientation='minus', exon_count=7, OMIM=614711, no_hdr0=''),
  391679 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391679, CurrentID=0, Status='live', Symbol='RPL6P14', Aliases='RPL6_2_505', description='ribosomal protein L6 pseudogene 14', other_designations='', map_location='4q24', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=104886110, end_position_on_the_genomic_accession=104886955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391681 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391681, CurrentID=0, Status='live', Symbol='EXOC7P1', Aliases='', description='exocyst complex component 7 pseudogene 1', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=108417708, end_position_on_the_genomic_accession=108419763, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391686 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391686, CurrentID=0, Status='live', Symbol='LYPLA1P2', Aliases='', description='lysophospholipase I pseudogene 2', other_designations='', map_location='4q25', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=110945112, end_position_on_the_genomic_accession=110945829, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101448202 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101448202, CurrentID=0, Status='live', Symbol='LOC101448202', Aliases='', description='uncharacterized LOC101448202', other_designations='', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=134819415, end_position_on_the_genomic_accession=134872618, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105371747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371747, CurrentID=0, Status='live', Symbol='LOC105371747', Aliases='', description='uncharacterized LOC105371747', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=36225283, end_position_on_the_genomic_accession=36237902, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391696 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391696, CurrentID=0, Status='live', Symbol='TECRP2', Aliases='', description='trans-2,3-enoyl-CoA reductase pseudogene 2', other_designations='', map_location='4q28.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=124462758, end_position_on_the_genomic_accession=124463860, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391697 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391697, CurrentID=0, Status='live', Symbol='ZSWIM5P3', Aliases='', description='zinc finger, SWIM-type containing 5 pseudogene 3', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129135932, end_position_on_the_genomic_accession=129139740, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391698 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391698, CurrentID=0, Status='live', Symbol='EEF1GP8', Aliases='', description='eukaryotic translation elongation factor 1 gamma pseudogene 8', other_designations='', map_location='4q28.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=129903001, end_position_on_the_genomic_accession=129904244, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391701 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391701, CurrentID=0, Status='live', Symbol='RPS23P2', Aliases='RPS23_2_521', description='ribosomal protein S23 pseudogene 2', other_designations='', map_location='4q28.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=137448053, end_position_on_the_genomic_accession=137448550, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391703 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391703, CurrentID=0, Status='live', Symbol='KRT18P51', Aliases='', description='keratin 18 pseudogene 51', other_designations='', map_location='4q31.21', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=144572238, end_position_on_the_genomic_accession=144572992, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129560, CurrentID=0, Status='live', Symbol='AHCYP4', Aliases='', description='adenosylhomocysteinase pseudogene 4', other_designations='', map_location='2q22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=138901242, end_position_on_the_genomic_accession=138902832, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391706 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391706, CurrentID=0, Status='live', Symbol='RPS3AP18', Aliases='RPS3A_5_528', description='ribosomal protein S3a pseudogene 18', other_designations='', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152551234, end_position_on_the_genomic_accession=152552383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129563 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129563, CurrentID=0, Status='live', Symbol='DIS3L2', Aliases='FAM6A, PRLMNS, hDIS3L2', description="DIS3 like 3'-5' exoribonuclease 2", other_designations='DIS3 mitotic control homolog-like 2|family with sequence similarity 6, member A', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231961583, end_position_on_the_genomic_accession=232343968, orientation='plus', exon_count=21, OMIM=614184, no_hdr0=''),
  283738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283738, CurrentID=0, Status='live', Symbol='NTRK3-AS1', Aliases='', description='NTRK3 antisense RNA 1', other_designations='NTRK3 antisense RNA 1 (non-protein coding)', map_location='15q25.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=88252730, end_position_on_the_genomic_accession=88271066, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  391710 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391710, CurrentID=0, Status='live', Symbol='LOC391710', Aliases='', description='tripartite motif-containing 75 pseudogene', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164939333, end_position_on_the_genomic_accession=164939990, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391711 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391711, CurrentID=0, Status='live', Symbol='LOC391711', Aliases='', description='tripartite motif-containing protein 60-like', other_designations='', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164948094, end_position_on_the_genomic_accession=164949393, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391712 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391712, CurrentID=0, Status='live', Symbol='TRIM61', Aliases='RNF35', description='tripartite motif containing 61', other_designations='ring finger protein 35|tripartite motif-containing 61', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=164954446, end_position_on_the_genomic_accession=164977666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  653857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653857, CurrentID=0, Status='live', Symbol='ACTR3C', Aliases='ARP11', description='ARP3 actin-related protein 3 homolog C (yeast)', other_designations='actin-related Arp11|actin-related protein 11', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=149887178, end_position_on_the_genomic_accession=150323669, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  391714 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391714, CurrentID=0, Status='live', Symbol='TRIM75P', Aliases='TRIM75', description='tripartite motif containing 75, pseudogene', other_designations='tripartite motif-containing 75|tripartite motif-containing protein 75', map_location='4q32.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=165058968, end_position_on_the_genomic_accession=165060562, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152667 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152667, CurrentID=0, Status='live', Symbol='FAM192BP', Aliases='FAM192B, rcNIP30', description='family with sequence similarity 192 member B, pseudogene', other_designations='family with sequence similarity 192, member A pseudogene|family with sequence similarity 192, member B, pseudogene', map_location='4q31.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152927430, end_position_on_the_genomic_accession=152928984, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391718 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391718, CurrentID=0, Status='live', Symbol='TSEN2P1', Aliases='', description='tRNA splicing endonuclease subunit 2 pseudogene 1', other_designations='TSEN2 tRNA splicing endonuclease subunit pseudogene 1|tRNA splicing endonuclease 2 homolog pseudogene 1', map_location='4q34.2', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=175494570, end_position_on_the_genomic_accession=175496813, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391719 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391719, CurrentID=0, Status='live', Symbol='RPL19P8', Aliases='RPL19_5_537', description='ribosomal protein L19 pseudogene 8', other_designations='', map_location='4q34.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=178286671, end_position_on_the_genomic_accession=178287364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391722 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391722, CurrentID=0, Status='live', Symbol='MYL12BP2', Aliases='', description='myosin light chain 12B pseudogene 2', other_designations='myosin regulatory light chain 12B pseudogene|myosin, light chain 12B pseudogene 2', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=184299114, end_position_on_the_genomic_accession=184299770, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391723 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391723, CurrentID=0, Status='live', Symbol='HELT', Aliases='HCM1228, HESL, Mgn, bHLHb44', description='helt bHLH transcription factor', other_designations='HES/HEY-like transcription factor|Heslike|Hey-like transcription factor|Hey-like transcriptional repressor|Megane bHLH factor', map_location='4q35.1', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=185018841, end_position_on_the_genomic_accession=185020804, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101927461 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101927461, CurrentID=0, Status='live', Symbol='LOC101927461', Aliases='', description='uncharacterized LOC101927461', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=131312172, end_position_on_the_genomic_accession=131315690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  391738 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391738, CurrentID=0, Status='live', Symbol='RPL29P13', Aliases='RPL29_5_543', description='ribosomal protein L29 pseudogene 13', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=13637791, end_position_on_the_genomic_accession=13638387, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391739 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391739, CurrentID=0, Status='live', Symbol='CCT6P2', Aliases='CCT6-1P', description='chaperonin containing TCP1, subunit 6 (zeta) pseudogene 2', other_designations='', map_location='5p15.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=14639412, end_position_on_the_genomic_accession=14641021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391741 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391741, CurrentID=0, Status='live', Symbol='LOC391741', Aliases='', description='actin, beta-like 2 pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=15450709, end_position_on_the_genomic_accession=15451831, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391742, CurrentID=0, Status='live', Symbol='LOC391742', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='Putative TAF11-like protein ENSP00000332601|TBP-associated factor 11 pseudogene', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17498231, end_position_on_the_genomic_accession=17498827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391744 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391744, CurrentID=0, Status='live', Symbol='LOC391744', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17511832, end_position_on_the_genomic_accession=17512238, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391746 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391746, CurrentID=0, Status='live', Symbol='LOC391746', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17521701, end_position_on_the_genomic_accession=17522497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391747 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391747, CurrentID=0, Status='live', Symbol='LOC391747', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17528569, end_position_on_the_genomic_accession=17529365, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129607, CurrentID=0, Status='live', Symbol='CMPK2', Aliases='TMPK2, TYKi, UMP-CMPK2', description='cytidine monophosphate (UMP-CMP) kinase 2, mitochondrial', other_designations='cytidylate kinase 2|mitochondrial UMP-CMP kinase|nucleoside-diphosphate kinase|thymidine monophosphate kinase 2|thymidylate kinase family LPS-inducible member', map_location='2p25.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=6840417, end_position_on_the_genomic_accession=6866635, orientation='minus', exon_count=8, OMIM=611787, no_hdr0=''),
  101929207 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929207, CurrentID=0, Status='live', Symbol='LOC101929207', Aliases='', description='uncharacterized LOC101929207', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=5507875, end_position_on_the_genomic_accession=5510002, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391765 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391765, CurrentID=0, Status='live', Symbol='LOC391765', Aliases='', description='H3 histone, family 3B (H3.3B) pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17625568, end_position_on_the_genomic_accession=17625959, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391768 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391768, CurrentID=0, Status='live', Symbol='LOC391768', Aliases='', description='TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 28kDa pseudogene', other_designations='', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17649040, end_position_on_the_genomic_accession=17649635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391769, CurrentID=0, Status='live', Symbol='H3.Y', Aliases='', description='histone cluster 2, H3d pseudogene', other_designations='H3 histone, family 3B (H3.3B) pseudogene', map_location='5p15.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=17655126, end_position_on_the_genomic_accession=17655538, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391771 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391771, CurrentID=0, Status='live', Symbol='LOC391771', Aliases='', description='protein tyrosine phosphatase, non-receptor type 11 pseudogene', other_designations='', map_location='5p14.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=23303963, end_position_on_the_genomic_accession=23305342, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391777 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391777, CurrentID=0, Status='live', Symbol='RPS4XP6', Aliases='RPS4P6, RPS4X_2_555', description='ribosomal protein S4X pseudogene 6', other_designations='ribosomal protein S4, X-linked pseudogene', map_location='5p13.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=37085005, end_position_on_the_genomic_accession=37085918, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  653924 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=653924, CurrentID=0, Status='live', Symbol='LOC653924', Aliases='', description='glycerol-3-phosphate acyltransferase 2, mitochondrial pseudogene', other_designations='', map_location='2q11.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=97081436, end_position_on_the_genomic_accession=97083472, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129642 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129642, CurrentID=0, Status='live', Symbol='MBOAT2', Aliases='LPCAT4, OACT2', description='membrane bound O-acyltransferase domain containing 2', other_designations='1-acylglycerophosphate O-acyltransferase|1-acylglycerophosphoethanolamine O-acyltransferase|LPAAT|LPEAT|LPLAT 2|O-acyltransferase (membrane bound) domain containing 2|O-acyltransferase domain-containing protein 2|lyso-PA acyltransferase|lyso-PE acyltransferase|lysophosphatidic acid acyltransferase|lysophosphatidylethanolamine acyltransferase|membrane-bound O-acyltransferase domain-containing protein 2', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=8852690, end_position_on_the_genomic_accession=9003747, orientation='minus', exon_count=14, OMIM=611949, no_hdr0=''),
  103021164 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103021164, CurrentID=0, Status='live', Symbol='CASC21', Aliases='CARLo-2, LINC01244', description='cancer susceptibility candidate 21 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1244', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127244637, end_position_on_the_genomic_accession=127392631, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103021165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103021165, CurrentID=0, Status='live', Symbol='CASC19', Aliases='CARLo-6, LINC01245', description='cancer susceptibility candidate 19 (non-protein coding)', other_designations='long intergenic non-protein coding RNA 1245', map_location='8q24.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=127187785, end_position_on_the_genomic_accession=127197627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  391798 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391798, CurrentID=0, Status='live', Symbol='LOC391798', Aliases='', description='karyopherin alpha 7 (importin alpha 8) pseudogene', other_designations='', map_location='5q13.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=75831121, end_position_on_the_genomic_accession=75832793, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  129656 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129656, CurrentID=0, Status='live', Symbol='LOC129656', Aliases='', description='CREB regulated transcription coactivator 1 pseudogene', other_designations='', map_location='2p16.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=52570648, end_position_on_the_genomic_accession=52572524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391803 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391803, CurrentID=0, Status='live', Symbol='KRT18P45', Aliases='', description='keratin 18 pseudogene 45', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80288400, end_position_on_the_genomic_accession=80289841, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391804, CurrentID=0, Status='live', Symbol='RBMX2P5', Aliases='', description='RNA binding motif protein, X-linked 2 pseudogene 5', other_designations='', map_location='5q14.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=80330780, end_position_on_the_genomic_accession=80332592, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  554247 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554247, CurrentID=0, Status='live', Symbol='LOC554247', Aliases='', description='uncharacterized LOC554247', other_designations='', map_location='18q23', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  391811 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391811, CurrentID=0, Status='live', Symbol='POLD2P1', Aliases='', description='polymerase (DNA directed), delta 2, accessory subunit pseudogene 1', other_designations='polymerase (DNA directed), delta 2, regulatory subunit 50kDa pseudogene 1', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=93267338, end_position_on_the_genomic_accession=93268684, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391813, CurrentID=0, Status='live', Symbol='LOC391813', Aliases='', description='zinc finger, DHHC-type containing 3 pseudogene', other_designations='', map_location='5q15', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=97737894, end_position_on_the_genomic_accession=97738900, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418836, CurrentID=0, Status='live', Symbol='LOC100418836', Aliases='', description='metadherin pseudogene', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=22140396, end_position_on_the_genomic_accession=22141117, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391819 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391819, CurrentID=0, Status='live', Symbol='KRT18P42', Aliases='', description='keratin 18 pseudogene 42', other_designations='', map_location='5q21.3', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=109588289, end_position_on_the_genomic_accession=109589660, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100130778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130778, CurrentID=0, Status='live', Symbol='LOC100130778', Aliases='', description='uncharacterized LOC100130778', other_designations='', map_location='12q24.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.10', start_position_on_the_genomic_accession=130074252, end_position_on_the_genomic_accession=130075239, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  391825 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391825, CurrentID=0, Status='live', Symbol='RPL23AP44', Aliases='RPL23A_14_607', description='ribosomal protein L23a pseudogene 44', other_designations='', map_location='5q23.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=121575848, end_position_on_the_genomic_accession=121576057, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391827, CurrentID=0, Status='live', Symbol='KRT18P16', Aliases='', description='keratin 18 pseudogene 16', other_designations='', map_location='5q23.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=123636080, end_position_on_the_genomic_accession=123637467, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129684 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129684, CurrentID=0, Status='live', Symbol='CNTNAP5', Aliases='caspr5', description='contactin associated protein-like 5', other_designations='cell recognition molecule Caspr5', map_location='2q14.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=124025287, end_position_on_the_genomic_accession=124921201, orientation='plus', exon_count=24, OMIM=610519, no_hdr0=''),
  129685 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129685, CurrentID=0, Status='live', Symbol='TAF8', Aliases='II, TAF, TAFII-43, TAFII43, TBN', description='TATA-box binding protein associated factor 8', other_designations='TAF(II)43|TAF8 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 43kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, A, 45/50kDa|TATA box binding protein associated factor 8|TBP-associated factor 43 kDa|TBP-associated factor 8|TBP-associated factor TAFII43|TBP-associated factor, RNA polymerase II, 43 kD|hTAFII43|protein taube nuss|taube nuss homolog|transcription initiation factor TFIID 43 kDa subunit', map_location='6p21.1', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=42050513, end_position_on_the_genomic_accession=42087462, orientation='plus', exon_count=11, OMIM=609514, no_hdr0=''),
  100129291 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100129291, CurrentID=0, Status='live', Symbol='LOC100129291', Aliases='', description='uncharacterized protein CXorf49-like', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71662995, end_position_on_the_genomic_accession=71666785, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  391833 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391833, CurrentID=0, Status='live', Symbol='RPS10P11', Aliases='RPS10_6_615', description='ribosomal protein S10 pseudogene 11', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=134422825, end_position_on_the_genomic_accession=134423090, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391834 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391834, CurrentID=0, Status='live', Symbol='LOC391834', Aliases='', description='ankyrin repeat domain 49 pseudogene', other_designations='', map_location='5q31.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=136857628, end_position_on_the_genomic_accession=136858342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391836, CurrentID=0, Status='live', Symbol='RPL10AP10', Aliases='RPL10A_4_618', description='ribosomal protein L10a pseudogene 10', other_designations='', map_location='5q31.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=138610655, end_position_on_the_genomic_accession=138620964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  391845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391845, CurrentID=0, Status='live', Symbol='RPS15P6', Aliases='RPS15_3_634', description='ribosomal protein S15 pseudogene 6', other_designations='', map_location='5q34', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=164600971, end_position_on_the_genomic_accession=164601470, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  391850 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=391850, CurrentID=0, Status='live', Symbol='RPL7AP33', Aliases='RPL7A_12_639', description='ribosomal protein L7a pseudogene 33', other_designations='', map_location='5q35.1', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=173225777, end_position_on_the_genomic_accession=173226428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105379775 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379775, CurrentID=0, Status='live', Symbol='LOC105379775', Aliases='', description='uncharacterized LOC105379775', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554249 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=554249, CurrentID=0, Status='live', Symbol='LOC554249', Aliases='', description='uncharacterized LOC554249', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=39873857, end_position_on_the_genomic_accession=39875395, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  101928962 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928962, CurrentID=0, Status='live', Symbol='LOC101928962', Aliases='', description='uncharacterized LOC101928962', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=114528843, end_position_on_the_genomic_accession=114542751, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100131910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100131910, CurrentID=0, Status='live', Symbol='LOC100131910', Aliases='', description='uncharacterized LOC100131910', other_designations='', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.10', start_position_on_the_genomic_accession=141043797, end_position_on_the_genomic_accession=141046059, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654029 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654029, CurrentID=0, Status='live', Symbol='RPS10P15', Aliases='RPS10_3_844', description='ribosomal protein S10 pseudogene 15', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128570195, end_position_on_the_genomic_accession=128570725, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930107 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930107, CurrentID=0, Status='live', Symbol='LOC101930107', Aliases='', description='uncharacterized LOC101930107', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103021294 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103021294, CurrentID=0, Status='live', Symbol='ABALON', Aliases='INXS', description='apoptotic BCL2L1-antisense long non-coding RNA', other_designations='apoptotic BCL-X-associated long ncRNA|intronic BCL-XS-inducing lncRNA', map_location='20q11.21', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=616018, no_hdr0=''),
  103021295 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103021295, CurrentID=0, Status='live', Symbol='LOC103021295', Aliases='', description='uncharacterized LOC103021295', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205957925, end_position_on_the_genomic_accession=205958388, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103021296 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=103021296, CurrentID=0, Status='live', Symbol='LOC103021296', Aliases='', description='uncharacterized LOC103021296', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=205934913, end_position_on_the_genomic_accession=205978095, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  129787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129787, CurrentID=0, Status='live', Symbol='TMEM18', Aliases='', description='transmembrane protein 18', other_designations='', map_location='2p25.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=667973, end_position_on_the_genomic_accession=677439, orientation='minus', exon_count=5, OMIM=613220, no_hdr0=''),
  129790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129790, CurrentID=0, Status='live', Symbol='C7orf13', Aliases='MY040', description='chromosome 7 open reading frame 13', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=156638366, end_position_on_the_genomic_accession=156640654, orientation='minus', exon_count=1, OMIM=610242, no_hdr0=''),
  105375411 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375411, CurrentID=0, Status='live', Symbol='LOC105375411', Aliases='', description='uncharacterized LOC105375411', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96450178, end_position_on_the_genomic_accession=96509218, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  129804 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129804, CurrentID=0, Status='live', Symbol='FBLN7', Aliases='TM14', description='fibulin 7', other_designations='FIBL-7', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=112138385, end_position_on_the_genomic_accession=112188218, orientation='plus', exon_count=11, OMIM=611551, no_hdr0=''),
  129807 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129807, CurrentID=0, Status='live', Symbol='NEU4', Aliases='', description='neuraminidase 4 (sialidase)', other_designations='N-acetyl-alpha-neuraminidase 4|neuraminidase 4|sialidase 4', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241809065, end_position_on_the_genomic_accession=241817413, orientation='plus', exon_count=6, OMIM=608527, no_hdr0=''),
  100287905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100287905, CurrentID=0, Status='live', Symbol='TTTY31P', Aliases='', description='testis-specific transcript, Y-linked 31, pseudogene', other_designations='', map_location='Yp11.2', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.9', start_position_on_the_genomic_accession=8795089, end_position_on_the_genomic_accession=8796046, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  87171 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=87171, CurrentID=0, Status='live', Symbol='ARPC3P1', Aliases='ARPC3B, dJ470L14.3', description='actin related protein 2/3 complex subunit 3 pseudogene 1', other_designations='actin related protein 2/3 complex, subunit 3 pseudogene 1|actin related protein 2/3 complex, subunit 3, 21kDa pseudogene|actin related protein 2/3 complex, subunit 3B pseudogene|actin related protein 2/3 complex, subunit 3B, 21kDa', map_location='20q13.13', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=49134451, end_position_on_the_genomic_accession=49135273, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129831 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129831, CurrentID=0, Status='live', Symbol='RBM45', Aliases='DRB1, RB-1', description='RNA binding motif protein 45', other_designations='RNA-binding motif protein 45|developmentally regulated RNA-binding protein 1|developmentally-regulated RNA-binding protein 1|putative RNA binding protein RB-1', map_location='2q31.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=178112417, end_position_on_the_genomic_accession=178129656, orientation='plus', exon_count=10, OMIM=608888, no_hdr0=''),
  129841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129841, CurrentID=0, Status='live', Symbol='RPSAP25', Aliases='RPSA_7_299', description='ribosomal protein SA pseudogene 25', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=176242035, end_position_on_the_genomic_accession=176243070, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  129852 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129852, CurrentID=0, Status='live', Symbol='C2orf73', Aliases='', description='chromosome 2 open reading frame 73', other_designations='', map_location='2p16.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=54330013, end_position_on_the_genomic_accession=54383673, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101930122 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930122, CurrentID=0, Status='live', Symbol='LOC101930122', Aliases='', description='general transcription factor II-I-like', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105378903 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105378903, CurrentID=0, Status='live', Symbol='LOC105378903', Aliases='', description='uncharacterized LOC105378903', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=111323840, end_position_on_the_genomic_accession=111324572, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  129868 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129868, CurrentID=0, Status='live', Symbol='TRIM43', Aliases='TRIM43A', description='tripartite motif containing 43', other_designations='tripartite motif-containing 43', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95592018, end_position_on_the_genomic_accession=95599771, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  129870 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129870, CurrentID=0, Status='live', Symbol='UBTFL3', Aliases='UBTFL5', description='upstream binding transcription factor, RNA polymerase I-like 3 (pseudogene)', other_designations='HMG box-containing protein LOC129870|putative upstream-binding factor 1-like protein 3/5', map_location='2q11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=95625250, end_position_on_the_genomic_accession=95626404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379776 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379776, CurrentID=0, Status='live', Symbol='LOC105379776', Aliases='', description='uncharacterized LOC105379776', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  129880 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129880, CurrentID=0, Status='live', Symbol='BBS5', Aliases='', description='Bardet-Biedl syndrome 5', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169479496, end_position_on_the_genomic_accession=169506655, orientation='plus', exon_count=12, OMIM=603650, no_hdr0=''),
  129881 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=129881, CurrentID=0, Status='live', Symbol='CCDC173', Aliases='C2orf77', description='coiled-coil domain containing 173', other_designations='', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169645425, end_position_on_the_genomic_accession=169694421, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  654170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654170, CurrentID=0, Status='live', Symbol='RPS16P8', Aliases='RPS16_4_1529', description='ribosomal protein S16 pseudogene 8', other_designations='', map_location='17q11.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=27280429, end_position_on_the_genomic_accession=27280961, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392027 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392027, CurrentID=0, Status='live', Symbol='LOC392027', Aliases='', description='ribosome binding protein 1 homolog 180kDa (dog) pseudogene', other_designations='ribosome receptor pseudogene', map_location='7p12.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=53002420, end_position_on_the_genomic_accession=53003429, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  101930131 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930131, CurrentID=0, Status='live', Symbol='LOC101930131', Aliases='', description='uncharacterized LOC101930131', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=90138063, end_position_on_the_genomic_accession=90156007, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  401497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401497, CurrentID=0, Status='live', Symbol='LINC01242', Aliases='', description='long intergenic non-protein coding RNA 1242', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=30388935, end_position_on_the_genomic_accession=30408454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  654231 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654231, CurrentID=0, Status='live', Symbol='OCM', Aliases='OCM1, OM, ONCM', description='oncomodulin', other_designations='hCG18255|oncomodulin 1|parvalbumin beta', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=5859804, end_position_on_the_genomic_accession=5886363, orientation='plus', exon_count=5, OMIM=164795, no_hdr0=''),
  102388891 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102388891, CurrentID=0, Status='live', Symbol='LOC102388891', Aliases='', description='family with sequence similarity 195, member A pseudogene', other_designations='', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131795372, end_position_on_the_genomic_accession=131796170, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392100 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392100, CurrentID=0, Status='live', Symbol='PSMC1P3', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase, 1 pseudogene 3', other_designations='proteasome 26S subunit, ATPase, 1 pseudogene', map_location='7q33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=136713826, end_position_on_the_genomic_accession=136715411, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392106, CurrentID=0, Status='live', Symbol='OR9A3P', Aliases='OR9A6P', description='olfactory receptor family 9 subfamily A member 3 pseudogene', other_designations='olfactory receptor, family 9, subfamily A, member 3 pseudogene|olfactory receptor, family 9, subfamily A, member 6 pseudogene', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141862859, end_position_on_the_genomic_accession=141863818, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654254 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654254, CurrentID=0, Status='live', Symbol='ZNF732', Aliases='', description='zinc finger protein 732', other_designations='zinc finger protein LOC654254', map_location='4p16.3', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=270675, end_position_on_the_genomic_accession=305321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  139628 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139628, CurrentID=0, Status='live', Symbol='FOXR2', Aliases='FOXN6', description='forkhead box R2', other_designations='forkhead box protein N6', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55623400, end_position_on_the_genomic_accession=55626188, orientation='plus', exon_count=1, OMIM=300949, no_hdr0=''),
  106481165 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=106481165, CurrentID=0, Status='live', Symbol='RNU6ATAC19P', Aliases='', description='RNA, U6atac small nuclear 19, pseudogene', other_designations='', map_location='2q31.3', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  392132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392132, CurrentID=0, Status='live', Symbol='OR2R1P', Aliases='OR2R1, OST058', description='olfactory receptor family 2 subfamily R member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily B, member 2 pseudogene|olfactory receptor, family 2, subfamily R, member 1, pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143488462, end_position_on_the_genomic_accession=143489409, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392133 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392133, CurrentID=0, Status='live', Symbol='OR10AC1', Aliases='OR10AC1P', description='olfactory receptor family 10 subfamily AC member 1 (gene/pseudogene)', other_designations='Olfactory receptor OR7-5|Putative olfactory receptor 10AC1|olfactory receptor, family 10, subfamily AC, member 1 pseudogene|olfactory receptor, family 10, subfamily W, member 1 pseudogene|seven transmembrane helix receptor', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=143510967, end_position_on_the_genomic_accession=143511911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392138 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392138, CurrentID=0, Status='live', Symbol='OR2A25', Aliases='OR2A24PP, OR2A27, OR2A25', description='olfactory receptor family 2 subfamily A member 25', other_designations='olfactory receptor 2A27|olfactory receptor, family 2, subfamily A, member 24 pseudogene|olfactory receptor, family 2, subfamily A, member 25 pseudogene|olfactory receptor, family 2, subfamily A, member 27', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144074220, end_position_on_the_genomic_accession=144075152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392140, CurrentID=0, Status='live', Symbol='OR2A13P', Aliases='', description='olfactory receptor family 2 subfamily A member 13 pseudogene', other_designations='olfactory receptor, family 2, subfamily A, member 13 pseudogene', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144141935, end_position_on_the_genomic_accession=144143141, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392145 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392145, CurrentID=0, Status='live', Symbol='LOC392145', Aliases='', description='exosome component 6 pseudogene', other_designations='', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=148445810, end_position_on_the_genomic_accession=148446395, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  139629 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=139629, CurrentID=0, Status='live', Symbol='XAGE-4', Aliases='', description='XAGE-4 protein', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=55654692, end_position_on_the_genomic_accession=55657961, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  130013 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130013, CurrentID=0, Status='live', Symbol='ACMSD', Aliases='', description='aminocarboxymuconate semialdehyde decarboxylase', other_designations='picolinate carboxylase', map_location='2q21.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=134838348, end_position_on_the_genomic_accession=134902076, orientation='plus', exon_count=12, OMIM=608889, no_hdr0=''),
  101930149 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930149, CurrentID=0, Status='live', Symbol='LOC101930149', Aliases='', description='uncharacterized LOC101930149', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=13338309, end_position_on_the_genomic_accession=13357501, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  152742 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152742, CurrentID=0, Status='live', Symbol='LINC01085', Aliases='', description='long intergenic non-protein coding RNA 1085', other_designations='', map_location='4p15.33', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=14111968, end_position_on_the_genomic_accession=14140052, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  130026 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130026, CurrentID=0, Status='live', Symbol='ICA1L', Aliases='ALS2CR14, ALS2CR15', description='islet cell autoantigen 1 like', other_designations='Ica69-related protein|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 14|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 15|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 14 protein|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 15 protein|islet cell autoantigen 1,69kDa-like', map_location='2q33.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=202773150, end_position_on_the_genomic_accession=202871985, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100422823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422823, CurrentID=0, Status='live', Symbol='MIR4300', Aliases='', description='microRNA 4300', other_designations='hsa-mir-4300', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=81890741, end_position_on_the_genomic_accession=81890836, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654319 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654319, CurrentID=0, Status='live', Symbol='SNORA5A', Aliases='ACA5, ACA5A/B/C', description='small nucleolar RNA, H/ACA box 5A', other_designations='', map_location='7p13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=45104349, end_position_on_the_genomic_accession=45104482, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654320 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654320, CurrentID=0, Status='live', Symbol='SNORA8', Aliases='ACA8', description='small nucleolar RNA, H/ACA box 8', other_designations='ACA8 small nucleolar RNA|ACA8 snoRNA', map_location='11q21', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=93732361, end_position_on_the_genomic_accession=93732499, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654321 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654321, CurrentID=0, Status='live', Symbol='SNORA75', Aliases='U23', description='small nucleolar RNA, H/ACA box 75', other_designations='U23 small nucleolar RNA|U23 snoRNA', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=231455800, end_position_on_the_genomic_accession=231455936, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654322 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654322, CurrentID=0, Status='live', Symbol='SNORA13', Aliases='ACA13', description='small nucleolar RNA, H/ACA box 13', other_designations='ACA13 small nucleolar RNA within TIGA1 host', map_location='5q22.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=112161485, end_position_on_the_genomic_accession=112161617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392180 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392180, CurrentID=0, Status='live', Symbol='LOC392180', Aliases='', description='transcriptional adaptor 1 pseudogene', other_designations='', map_location='8p23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=5848697, end_position_on_the_genomic_accession=5849692, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392181 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392181, CurrentID=0, Status='live', Symbol='RPL23AP96', Aliases='', description='ribosomal protein L23a pseudogene 96', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=6898881, end_position_on_the_genomic_accession=6899411, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392187 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392187, CurrentID=0, Status='live', Symbol='LOC392187', Aliases='', description='ubiquitin carboxyl-terminal hydrolase 17-like protein 2-like', other_designations='', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7967344, end_position_on_the_genomic_accession=7968544, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392188 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392188, CurrentID=0, Status='live', Symbol='USP17L8', Aliases='USP17L10', description='ubiquitin specific peptidase 17-like family member 8', other_designations='ubiquitin specific peptidase 17-like 8', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=7971661, end_position_on_the_genomic_accession=7973253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654337 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654337, CurrentID=0, Status='live', Symbol='BCYRN1P3', Aliases='', description='brain cytoplasmic RNA 1, pseudogene 3', other_designations='', map_location='13q11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=18709709, end_position_on_the_genomic_accession=18709902, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392194 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392194, CurrentID=0, Status='live', Symbol='OR7E158P', Aliases='', description='olfactory receptor family 7 subfamily E member 158 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 158 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=11919896, end_position_on_the_genomic_accession=11920795, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392196 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392196, CurrentID=0, Status='live', Symbol='LOC392196', Aliases='', description='ubiquitin specific peptidase 17-like family member 2 pseudogene', other_designations='deubiquitinating enzyme 3 pseudogene|ubiquitin carboxyl-terminal hydrolase 17-like pseudogene|ubiquitin specific peptidase 17-like 2 pseudogene', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12127858, end_position_on_the_genomic_accession=12129297, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392197 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392197, CurrentID=0, Status='live', Symbol='USP17L7', Aliases='', description='ubiquitin specific peptidase 17-like family member 7', other_designations='ubiquitin specific peptidase 17-like 7', map_location='8p23.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=12132417, end_position_on_the_genomic_accession=12134009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654342 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654342, CurrentID=0, Status='live', Symbol='LOC654342', Aliases='', description='lymphocyte-specific protein 1 pseudogene', other_designations='', map_location='2p11.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=91636683, end_position_on_the_genomic_accession=91659949, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422828 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422828, CurrentID=0, Status='live', Symbol='MIR4287', Aliases='', description='microRNA 4287', other_designations='hsa-mir-4287', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=27886039, end_position_on_the_genomic_accession=27886116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654346 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654346, CurrentID=0, Status='live', Symbol='LGALS9C', Aliases='', description='lectin, galactoside-binding, soluble, 9C', other_designations='gal-9C|galectin 9 like|galectin-9-like protein B|lectin, galactoside-binding, soluble, 9 (galectin 9) pseudogene', map_location='17p11.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=18476785, end_position_on_the_genomic_accession=18494945, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  654348 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654348, CurrentID=0, Status='live', Symbol='PRAMEF16', Aliases='', description='PRAME family member 16', other_designations='', map_location='1p36.21', chromosome='1', genomic_nucleotide_accession_version='NC_000001.10', start_position_on_the_genomic_accession=13495254, end_position_on_the_genomic_accession=13498260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  654350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654350, CurrentID=0, Status='live', Symbol='RPL9P25', Aliases='RPL9_13_1445', description='ribosomal protein L9 pseudogene 25', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66379238, end_position_on_the_genomic_accession=66379937, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392209 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392209, CurrentID=0, Status='live', Symbol='HSPA8P11', Aliases='', description='heat shock protein family A (Hsp70) member 8 pseudogene 11', other_designations='heat shock 70kDa protein 8 pseudogene 11', map_location='8p12', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30237286, end_position_on_the_genomic_accession=30239389, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392214 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392214, CurrentID=0, Status='live', Symbol='KRT18P37', Aliases='', description='keratin 18 pseudogene 37', other_designations='', map_location='8p11.21', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=41511239, end_position_on_the_genomic_accession=41512590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  130074 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130074, CurrentID=0, Status='live', Symbol='FAM168B', Aliases='MANI', description='family with sequence similarity 168 member B', other_designations='family with sequence similarity 168, member B|p20|protein FAM168B', map_location='2q21.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=131047876, end_position_on_the_genomic_accession=131093431, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  130075 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130075, CurrentID=0, Status='live', Symbol='OR9A4', Aliases='', description='olfactory receptor family 9 subfamily A member 4', other_designations='olfactory receptor OR7-1|olfactory receptor, family 9, subfamily A, member 4', map_location='7q34', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=141873479, end_position_on_the_genomic_accession=141986420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  654364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654364, CurrentID=0, Status='live', Symbol='NME1-NME2', Aliases='NM23-LV, NMELV', description='NME1-NME2 readthrough', other_designations='NME1-NME2 readthrough transcript', map_location='17q21.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=51153536, end_position_on_the_genomic_accession=51171744, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  654365 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654365, CurrentID=0, Status='live', Symbol='RPS27P3', Aliases='RPS27_8_1194', description='ribosomal protein S27 pseudogene 3', other_designations='', map_location='12p13.32', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=3211606, end_position_on_the_genomic_accession=3211951, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654369 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654369, CurrentID=0, Status='live', Symbol='RPL37P5', Aliases='RPL37_12_1377', description='ribosomal protein L37 pseudogene 5', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60572010, end_position_on_the_genomic_accession=60572351, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654370 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654370, CurrentID=0, Status='live', Symbol='RPL7AP3', Aliases='', description='ribosomal protein L7a pseudogene 3', other_designations='', map_location='14q13.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=35235115, end_position_on_the_genomic_accession=35235653, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654371, CurrentID=0, Status='live', Symbol='RPS15AP4', Aliases='RPS15A_20_1378', description='ribosomal protein S15a pseudogene 4', other_designations='', map_location='14q23.1', chromosome='14', genomic_nucleotide_accession_version='NC_000014.9', start_position_on_the_genomic_accession=60602106, end_position_on_the_genomic_accession=60602551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101930161 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101930161, CurrentID=0, Status='live', Symbol='LOC101930161', Aliases='', description='RNA polymerase II transcription factor SIII subunit A3-like', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  654376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654376, CurrentID=0, Status='live', Symbol='RPL12P9', Aliases='RPL12_18_1725', description='ribosomal protein L12 pseudogene 9', other_designations='', map_location='21q21.3', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=29127649, end_position_on_the_genomic_accession=29128275, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654377, CurrentID=0, Status='live', Symbol='CHES1L1', Aliases='C18orf7', description='checkpoint suppressor 1-like 1', other_designations='', map_location='18q21.32', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  654381 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654381, CurrentID=0, Status='live', Symbol='RPL7P5', Aliases='RPL7_18_1464', description='ribosomal protein L7 pseudogene 5', other_designations='', map_location='15q26.3', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=99551314, end_position_on_the_genomic_accession=99552128, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654382, CurrentID=0, Status='live', Symbol='RPL19P2', Aliases='RPL19_3_373', description='ribosomal protein L19 pseudogene 2', other_designations='', map_location='3p14.3', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=57134103, end_position_on_the_genomic_accession=57134802, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654383 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654383, CurrentID=0, Status='live', Symbol='RPL13P4', Aliases='HsT16723, RPL13_6_1454', description='ribosomal protein L13 pseudogene 4', other_designations='', map_location='15q24.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=75388236, end_position_on_the_genomic_accession=75388876, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654385 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654385, CurrentID=0, Status='live', Symbol='RPL21P15', Aliases='HsT18206, RPL21_54_1444', description='ribosomal protein L21 pseudogene 15', other_designations='', map_location='15q22.31', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=64631077, end_position_on_the_genomic_accession=64631533, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654386 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654386, CurrentID=0, Status='live', Symbol='RPL26P5', Aliases='HsT41398', description='ribosomal protein L26 pseudogene 5', other_designations='', map_location='15q26.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=94273496, end_position_on_the_genomic_accession=94274002, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422835, CurrentID=0, Status='live', Symbol='MIR3183', Aliases='mir-3183', description='microRNA 3183', other_designations='hsa-mir-3183', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=1022476, end_position_on_the_genomic_accession=1022559, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  654388 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654388, CurrentID=0, Status='live', Symbol='RPL32P2', Aliases='HsT20607, RPL32_9_1423', description='ribosomal protein L32 pseudogene 2', other_designations='', map_location='15q14', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=34085858, end_position_on_the_genomic_accession=34086366, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654390, CurrentID=0, Status='live', Symbol='RPS27P2', Aliases='HsT16324, RPS27_11_1420', description='ribosomal protein S27 pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=24910714, end_position_on_the_genomic_accession=24911257, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654391, CurrentID=0, Status='live', Symbol='RPL31P7', Aliases='RPL31_22_1578', description='ribosomal protein L31 pseudogene 7', other_designations='', map_location='17q25.3', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=80602507, end_position_on_the_genomic_accession=80602955, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654392, CurrentID=0, Status='live', Symbol='RPL39P4', Aliases='RPL39_18_1545', description='ribosomal protein L39 pseudogene 4', other_designations='', map_location='17q21.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=39839175, end_position_on_the_genomic_accession=39839573, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422836 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422836, CurrentID=0, Status='live', Symbol='MIR3180-3', Aliases='', description='microRNA 3180-3', other_designations='hsa-mir-3180-3', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=18402178, end_position_on_the_genomic_accession=18402271, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  130106 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130106, CurrentID=0, Status='live', Symbol='CIB4', Aliases='KIP4', description='calcium and integrin binding family member 4', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=26573764, end_position_on_the_genomic_accession=26641411, orientation='minus', exon_count=10, OMIM=610646, no_hdr0=''),
  392255 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392255, CurrentID=0, Status='live', Symbol='GDF6', Aliases='BMP-13, BMP13, CDMP2, KFM, KFS, KFS1, KFSL, LCA17, MCOP4, MCOPCB6, SCDO4, SGM1', description='growth differentiation factor 6', other_designations='GDF-6|Klip-Feil malformation|Klippel-Feil malformation|Klippel-Feil syndrome|bone morphogenetic protein 13|growth/differentiation factor 16', map_location='8q22.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=96142330, end_position_on_the_genomic_accession=96160792, orientation='minus', exon_count=3, OMIM=601147, no_hdr0=''),
  392256 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392256, CurrentID=0, Status='live', Symbol='RPS26P6', Aliases='RPS26_10_933', description='ribosomal protein S26 pseudogene 6', other_designations='', map_location='8q22.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=100895740, end_position_on_the_genomic_accession=100896202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392262 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392262, CurrentID=0, Status='live', Symbol='RPSAP48', Aliases='RPSA_15_943', description='ribosomal protein SA pseudogene 48', other_designations='', map_location='8q23.2', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=110105076, end_position_on_the_genomic_accession=110105964, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130120 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130120, CurrentID=0, Status='live', Symbol='REG3G', Aliases='LPPM429, PAP IB, PAP-1B, PAP1B, PAPIB, REG III, REG-III, UNQ429', description='regenerating family member 3 gamma', other_designations='REG-3-gamma|pancreatitis-associated protein 1B|pancreatitis-associated protein IB|reg III-gamma|regenerating gene III|regenerating islet-derived 3 gamma|regenerating islet-derived protein III-gamma', map_location='2p12', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=79025686, end_position_on_the_genomic_accession=79028504, orientation='plus', exon_count=7, OMIM=609933, no_hdr0=''),
  392265 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392265, CurrentID=0, Status='live', Symbol='CDK5PS', Aliases='', description='cyclin-dependent kinase 5 pseudogene', other_designations='', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=122769229, end_position_on_the_genomic_accession=122770318, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392266 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392266, CurrentID=0, Status='live', Symbol='LOC392266', Aliases='', description='17beta-hydroxysteroid dehydrogenase pseudogene', other_designations='hydroxysteroid (17-beta) dehydrogenase 4 pseudogene', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123288332, end_position_on_the_genomic_accession=123291480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392267 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392267, CurrentID=0, Status='live', Symbol='IMPDH1P6', Aliases='', description='IMP (inosine monophosphate) dehydrogenase 1 pseudogene 6', other_designations='inosine monophosphate dehydrogenase 1 pseudogene', map_location='8q24.13', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123400522, end_position_on_the_genomic_accession=123402815, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654412 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654412, CurrentID=0, Status='live', Symbol='FAM138B', Aliases='F379, bA395L14.6', description='family with sequence similarity 138 member B', other_designations='F379 retina-specific protein|chromosome 2 F379 retina specific protein|family with sequence similarity 138, member B', map_location='2q14.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113577382, end_position_on_the_genomic_accession=113578852, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392271, CurrentID=0, Status='live', Symbol='RPL32P20', Aliases='RPL32_12_951', description='ribosomal protein L32 pseudogene 20', other_designations='', map_location='8q24.22', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=133325997, end_position_on_the_genomic_accession=133326404, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392275 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392275, CurrentID=0, Status='live', Symbol='SMPD5', Aliases='MA-nSMase', description='sphingomyelin phosphodiesterase 5', other_designations='CTD-3065J16.6|mitochondria-associated nSMase', map_location='8q24.3', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144047114, end_position_on_the_genomic_accession=144051294, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  130132 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130132, CurrentID=0, Status='live', Symbol='RFTN2', Aliases='C2orf11, Raftlin-2', description='raftlin family member 2', other_designations='raft-linking protein 2', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=197570800, end_position_on_the_genomic_accession=197676045, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100422841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422841, CurrentID=0, Status='live', Symbol='MIR3174', Aliases='', description='microRNA 3174', other_designations='hsa-mir-3174', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=90006755, end_position_on_the_genomic_accession=90006841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392281 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392281, CurrentID=0, Status='live', Symbol='GPS2P1', Aliases='', description='G protein pathway suppressor 2 pseudogene 1', other_designations='', map_location='9p24.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=2875387, end_position_on_the_genomic_accession=2876503, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392282 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392282, CurrentID=0, Status='live', Symbol='RPS5P6', Aliases='RPS5_2_957', description='ribosomal protein S5 pseudogene 6', other_designations='', map_location='9p24.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=4781417, end_position_on_the_genomic_accession=4782149, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654427 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654427, CurrentID=0, Status='live', Symbol='PCNPP1', Aliases='PCNPP', description='PEST containing nuclear protein pseudogene 1', other_designations='PEST proteolytic signal containing nuclear protein pseudogene', map_location='12q24.12', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=111666924, end_position_on_the_genomic_accession=111670379, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654429, CurrentID=0, Status='live', Symbol='LRTM2', Aliases='', description='leucine-rich repeats and transmembrane domains 2', other_designations='', map_location='12p13.33', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=1820267, end_position_on_the_genomic_accession=1836753, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  392288 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392288, CurrentID=0, Status='live', Symbol='MAP1LC3BP1', Aliases='', description='microtubule associated protein 1 light chain 3 beta pseudogene 1', other_designations='microtubule-associated protein 1 light chain 3 beta pseudogene 1', map_location='9p22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=19460815, end_position_on_the_genomic_accession=19463389, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654433, CurrentID=0, Status='live', Symbol='PAX8-AS1', Aliases='', description='PAX8 antisense RNA 1', other_designations='', map_location='2q13', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=113235527, end_position_on_the_genomic_accession=113267023, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  654434 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654434, CurrentID=0, Status='live', Symbol='SNHG20', Aliases='C17orf86, LINC00338, NCRNA00338, SCARNA16HG', description='small nucleolar RNA host gene 20', other_designations='long intergenic non-protein coding RNA 338|small nucleolar RNA host gene 20 (non-protein coding)', map_location='17q25.2', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=77088643, end_position_on_the_genomic_accession=77094986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392292 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392292, CurrentID=0, Status='live', Symbol='IFNWP5', Aliases='', description='interferon, omega 1 pseudogene 5', other_designations='interferon, omega 5 (pseudogene)', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21230821, end_position_on_the_genomic_accession=21231853, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105376947 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105376947, CurrentID=0, Status='live', Symbol='LOC105376947', Aliases='', description='uncharacterized LOC105376947', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=8091244, end_position_on_the_genomic_accession=8120195, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  392298 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392298, CurrentID=0, Status='live', Symbol='CTAGE12P', Aliases='', description='CTAGE family member 12, pseudogene', other_designations='CTAGE family, member 12, pseudogene', map_location='9p21.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=27607988, end_position_on_the_genomic_accession=27610967, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392301 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392301, CurrentID=0, Status='live', Symbol='SLC25A5P8', Aliases='', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 5 pseudogene 8', other_designations='', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32333195, end_position_on_the_genomic_accession=32334432, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392302 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392302, CurrentID=0, Status='live', Symbol='ASS1P12', Aliases='ASSP12', description='argininosuccinate synthetase 1 pseudogene 12', other_designations='argininosuccinate synthase 1 pseudogene|argininosuccinate synthetase pseudogene 12', map_location='9p21.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=32945894, end_position_on_the_genomic_accession=32947460, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392305 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392305, CurrentID=0, Status='live', Symbol='GLULP4', Aliases='GLULP', description='glutamate-ammonia ligase (glutamine synthetase) pseudogene 4', other_designations='glutamate-ammonia ligase (glutamine synthase) pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34917045, end_position_on_the_genomic_accession=34918408, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  130162 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130162, CurrentID=0, Status='live', Symbol='CLHC1', Aliases='C2orf63', description='clathrin heavy chain linker domain containing 1', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=55172547, end_position_on_the_genomic_accession=55232563, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  392307 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392307, CurrentID=0, Status='live', Symbol='FAM221B', Aliases='C9orf128', description='family with sequence similarity 221 member B', other_designations='family with sequence similarity 221, member B|uncharacterized protein C9orf128', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35817017, end_position_on_the_genomic_accession=35833805, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  392308 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392308, CurrentID=0, Status='live', Symbol='OR13E1P', Aliases='OR13E2, OST741', description='olfactory receptor family 13 subfamily E member 1 pseudogene', other_designations='olfactory receptor, family 13, subfamily E, member 1 pseudogene|olfactory receptor, family 13, subfamily E, member 2', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35859025, end_position_on_the_genomic_accession=35859980, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392309 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392309, CurrentID=0, Status='live', Symbol='OR13J1', Aliases='OR9-2', description='olfactory receptor family 13 subfamily J member 1', other_designations='olfactory receptor OR9-2|olfactory receptor, family 13, subfamily J, member 1', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35869463, end_position_on_the_genomic_accession=35870401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392310 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392310, CurrentID=0, Status='live', Symbol='PGAM1P2', Aliases='', description='phosphoglycerate mutase 1 pseudogene 2', other_designations='phosphoglycerate mutase 1 (brain) pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35943210, end_position_on_the_genomic_accession=35943934, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392311 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392311, CurrentID=0, Status='live', Symbol='OR13C6P', Aliases='OR37B', description='olfactory receptor family 13 subfamily C member 6 pseudogene', other_designations='Putative olfactory receptor ENSP00000348552|olfactory receptor, family 13, subfamily C, member 6 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=35991339, end_position_on_the_genomic_accession=35992285, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392313 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392313, CurrentID=0, Status='live', Symbol='OR2S1P', Aliases='OR37D, OST611', description='olfactory receptor family 2 subfamily S member 1 pseudogene', other_designations='olfactory receptor, family 2, subfamily S, member 1 pseudogene', map_location='9p13.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=36013185, end_position_on_the_genomic_accession=36014131, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105379778 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379778, CurrentID=0, Status='live', Symbol='LOC105379778', Aliases='', description='uncharacterized LOC105379778', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  654463 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654463, CurrentID=0, Status='live', Symbol='FER1L6', Aliases='C8ORFK23', description='fer-1-like family member 6', other_designations='fer-1-like 6', map_location='8q24.1', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=123851987, end_position_on_the_genomic_accession=124120061, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  100008588 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100008588, CurrentID=0, Status='live', Symbol='RNA18S5', Aliases='RN18S1', description='RNA, 18S ribosomal 5', other_designations='18S ribosomal RNA', map_location='22p12', chromosome='22', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101055624 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101055624, CurrentID=0, Status='live', Symbol='NYS7', Aliases='', description='Nystagmus 7, congenital', other_designations='', map_location='1q31-q32.2', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=614826, no_hdr0=''),
  101055625 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101055625, CurrentID=0, Status='live', Symbol='OVCH1-AS1', Aliases='', description='OVCH1 antisense RNA 1', other_designations='', map_location='12p', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=29389294, end_position_on_the_genomic_accession=29487324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392334 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392334, CurrentID=0, Status='live', Symbol='LOC392334', Aliases='', description='calponin 2 pseudogene', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=66029226, end_position_on_the_genomic_accession=66030366, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392335 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392335, CurrentID=0, Status='live', Symbol='MYO5BP2', Aliases='', description='myosin VB pseudogene 2', other_designations='', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=62857201, end_position_on_the_genomic_accession=62858483, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654480 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654480, CurrentID=0, Status='live', Symbol='ATP5G2P4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit C2 (subunit 9) pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C2 (subunit 9) pseudogene|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit C2 (subunit 9) pseudogene 4', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=42098603, end_position_on_the_genomic_accession=42099231, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654483, CurrentID=0, Status='live', Symbol='BOLA2B', Aliases='BOLA2', description='bolA family member 2B', other_designations='BolA-like protein 2 member B|bolA homolog 2B|bolA-like 2B|bolA-like protein 2B', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30192935, end_position_on_the_genomic_accession=30194306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  654484 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654484, CurrentID=0, Status='live', Symbol='ATP5J2P5', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene 5', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2 pseudogene 5|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene 5', map_location='12p13.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=6270025, end_position_on_the_genomic_accession=6270442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654485, CurrentID=0, Status='live', Symbol='ATP5J2P4', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene 4', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2 pseudogene 4|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene 4', map_location='17q', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64491518, end_position_on_the_genomic_accession=64491926, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  654486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654486, CurrentID=0, Status='live', Symbol='ATP5J2P6', Aliases='', description='ATP synthase, H+ transporting, mitochondrial Fo complex subunit F2 pseudogene 6', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2 pseudogene 6|ATP synthase, H+ transporting, mitochondrial Fo complex, subunit F2 pseudogene 6', map_location='15q22', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66414800, end_position_on_the_genomic_accession=66415225, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392343 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392343, CurrentID=0, Status='live', Symbol='CNN2P3', Aliases='', description='calponin 2 pseudogene 3', other_designations='', map_location='9q21.11', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=64485967, end_position_on_the_genomic_accession=64487107, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392350 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392350, CurrentID=0, Status='live', Symbol='BTF3P4', Aliases='', description='basic transcription factor 3 pseudogene 4', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=72107563, end_position_on_the_genomic_accession=72108422, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392352 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392352, CurrentID=0, Status='live', Symbol='LOC392352', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='9q21.13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=76549085, end_position_on_the_genomic_accession=76549813, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654499 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654499, CurrentID=0, Status='live', Symbol='KIAA0125P2', Aliases='FAM30C, HsT16041', description='KIAA0125 pseudogene 2', other_designations='', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  654500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654500, CurrentID=0, Status='live', Symbol='H2AFVP1', Aliases='H2AFVP, HsT19970', description='H2A histone family member V pseudogene 1', other_designations='H2A histone family, member V pseudogene|H2A histone family, member V, pseudogene 1', map_location='15q26.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92733671, end_position_on_the_genomic_accession=92734276, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422854 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422854, CurrentID=0, Status='live', Symbol='MIR3187', Aliases='mir-3187', description='microRNA 3187', other_designations='hsa-mir-3187', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=813584, end_position_on_the_genomic_accession=813653, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  654502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654502, CurrentID=0, Status='live', Symbol='IQCJ', Aliases='', description='IQ motif containing J', other_designations='calmodulin binding protein', map_location='3q25.32', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=159069252, end_position_on_the_genomic_accession=159266307, orientation='plus', exon_count=5, OMIM=611622, no_hdr0=''),
  654503 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654503, CurrentID=0, Status='live', Symbol='HMGB1P9', Aliases='HMG1L9, HMGB1L9', description='high mobility group box 1 pseudogene 9', other_designations='high-mobility group (nonhistone chromosomal) protein 1-like 9|high-mobility group box 1 pseudogene 9|high-mobility group box 1-like 9', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=218200276, end_position_on_the_genomic_accession=218201850, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392360 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392360, CurrentID=0, Status='live', Symbol='CTSL3P', Aliases='CTSL3, HCTSL-s', description='cathepsin L family member 3, pseudogene', other_designations='cathepsin L-like protein', map_location='9q21.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87772915, end_position_on_the_genomic_accession=87786884, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  654505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654505, CurrentID=0, Status='live', Symbol='H3F3AP1', Aliases='HsT18159', description='H3 histone, family 3A pseudogene 1', other_designations='H3 histone, family 3A, pseudogene 1', map_location='15q15.1', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=39950856, end_position_on_the_genomic_accession=39951834, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  654506 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654506, CurrentID=0, Status='live', Symbol='RPSAP7', Aliases='LAMR1P7, RPSA_16_1024', description='ribosomal protein SA pseudogene 7', other_designations='laminin receptor 1 pseudogene 7', map_location='10p13', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=14723099, end_position_on_the_genomic_accession=14724138, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  414919 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414919, CurrentID=0, Status='live', Symbol='C8orf82', Aliases='', description='chromosome 8 open reading frame 82', other_designations='', map_location='8q24', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=144526219, end_position_on_the_genomic_accession=144529074, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  392364 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392364, CurrentID=0, Status='live', Symbol='LOC392364', Aliases='', description='nuclear pore associated protein 1 pseudogene', other_designations='', map_location='9q22.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=87858043, end_position_on_the_genomic_accession=87859056, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392368 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392368, CurrentID=0, Status='live', Symbol='BEND3P2', Aliases='', description='BEN domain containing 3 pseudogene 2', other_designations='', map_location='9q22.31', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=92171231, end_position_on_the_genomic_accession=92173494, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422856, CurrentID=0, Status='live', Symbol='MIR3123', Aliases='mir-3123', description='microRNA 3123', other_designations='hsa-mir-3123', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=241132272, end_position_on_the_genomic_accession=241132346, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392371 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392371, CurrentID=0, Status='live', Symbol='KRT18P13', Aliases='KRT18L3', description='keratin 18 pseudogene 13', other_designations='keratin 18-like 3', map_location='9q22.33', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=97698872, end_position_on_the_genomic_accession=97700801, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392374 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392374, CurrentID=0, Status='live', Symbol='GAPDHP26', Aliases='', description='glyceraldehyde 3 phosphate dehydrogenase pseudogene 26', other_designations='glyceraldehyde-3-phosphate dehydrogenase pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=100975588, end_position_on_the_genomic_accession=100976737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422857 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422857, CurrentID=0, Status='live', Symbol='MIR4318', Aliases='', description='microRNA 4318', other_designations='hsa-mir-4318', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=37657135, end_position_on_the_genomic_accession=37657215, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392376 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392376, CurrentID=0, Status='live', Symbol='OR13C2', Aliases='OR37K', description='olfactory receptor family 13 subfamily C member 2', other_designations='olfactory receptor OR9-12|olfactory receptor, family 13, subfamily C, member 2', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104604671, end_position_on_the_genomic_accession=104605627, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392377 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392377, CurrentID=0, Status='live', Symbol='OR13C1P', Aliases='OR37M', description='olfactory receptor family 13 subfamily C member 1 pseudogene', other_designations='olfactory receptor, family 13, subfamily C, member 1 pseudogene', map_location='9q31.1', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=104656763, end_position_on_the_genomic_accession=104657716, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392382 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392382, CurrentID=0, Status='live', Symbol='RPL31P43', Aliases='RPL31_21_1004', description='ribosomal protein L31 pseudogene 43', other_designations='', map_location='9q31.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=108040185, end_position_on_the_genomic_accession=108040630, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  196469 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196469, CurrentID=0, Status='live', Symbol='LOC196469', Aliases='', description='DNA dC->dU-editing enzyme APOBEC-3G-like', other_designations='', map_location='12q23.3', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=108281106, end_position_on_the_genomic_accession=108282235, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928979, CurrentID=0, Status='live', Symbol='LOC101928979', Aliases='', description='uncharacterized LOC101928979', other_designations='', map_location='1q21.1', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=146052596, end_position_on_the_genomic_accession=146058799, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392387 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392387, CurrentID=0, Status='live', Symbol='AHCYP2', Aliases='', description='adenosylhomocysteinase pseudogene 2', other_designations='', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=120720573, end_position_on_the_genomic_accession=120722157, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375414 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375414, CurrentID=0, Status='live', Symbol='LOC105375414', Aliases='', description='uncharacterized LOC105375414', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96786823, end_position_on_the_genomic_accession=96804686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392390 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392390, CurrentID=0, Status='live', Symbol='OR1L6', Aliases='HG16, OR1L7, OR9-30', description='olfactory receptor family 1 subfamily L member 6', other_designations='olfactory receptor 1L7|olfactory receptor OR9-30|olfactory receptor, family 1, subfamily L, member 6|olfactory receptor, family 1, subfamily L, member 7', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122749848, end_position_on_the_genomic_accession=122750783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392391 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392391, CurrentID=0, Status='live', Symbol='OR5C1', Aliases='OR5C2P, OR9-31, OR9-F', description='olfactory receptor family 5 subfamily C member 1', other_designations='hRPK-465_F_21|olfactory receptor 5C2|olfactory receptor 9-F|olfactory receptor, family 5, subfamily C, member 1|olfactory receptor, family 5, subfamily C, member 2 pseudogene', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122788933, end_position_on_the_genomic_accession=122789895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392392 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392392, CurrentID=0, Status='live', Symbol='OR1K1', Aliases='hg99', description='olfactory receptor family 1 subfamily K member 1', other_designations='olfactory receptor, family 1, subfamily K, member 1', map_location='9q33.2', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=122800123, end_position_on_the_genomic_accession=122801073, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392399, CurrentID=0, Status='live', Symbol='LCN9', Aliases='9230102I19Rik, HEL129', description='lipocalin 9', other_designations='MUP-like lipocalin|epididymis luminal protein 129', map_location='9q34.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=135663322, end_position_on_the_genomic_accession=135666769, orientation='plus', exon_count=6, OMIM=612903, no_hdr0=''),
  100422862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422862, CurrentID=0, Status='live', Symbol='MIR548S', Aliases='', description='microRNA 548s', other_designations='hsa-mir-548s', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11767444, end_position_on_the_genomic_accession=11767525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  414927 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=414927, CurrentID=0, Status='live', Symbol='MGC34796', Aliases='', description='sepiapterin reductase (7,8-dihydrobiopterin:NADP+ oxidoreductase) pseudogene', other_designations='', map_location='1p32', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=61654242, end_position_on_the_genomic_accession=61656128, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  130271 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130271, CurrentID=0, Status='live', Symbol='PLEKHH2', Aliases='PLEKHH1L', description='pleckstrin homology, MyTH4 and FERM domain containing H2', other_designations='pleckstrin homology domain containing, family H (with MyTH4 domain) member 2', map_location='2p21', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=43637300, end_position_on_the_genomic_accession=43767987, orientation='plus', exon_count=31, OMIM=612723, no_hdr0=''),
  283856 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283856, CurrentID=0, Status='live', Symbol='LOC283856', Aliases='', description='uncharacterized LOC283856', other_designations='', map_location='16q12.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=56092987, end_position_on_the_genomic_accession=56191094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  392424 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392424, CurrentID=0, Status='live', Symbol='RPS5P8', Aliases='RPS5_5_1762', description='ribosomal protein S5 pseudogene 8', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=6422047, end_position_on_the_genomic_accession=6422773, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392425 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392425, CurrentID=0, Status='live', Symbol='RPS27AP17', Aliases='RPS27A_9_1763', description='ribosomal protein S27a pseudogene 17', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=6989329, end_position_on_the_genomic_accession=6990021, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392426 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392426, CurrentID=0, Status='live', Symbol='LOC392426', Aliases='', description='nucleolar and coiled-body phosphoprotein 1 pseudogene', other_designations='', map_location='Xp22.31', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=8980425, end_position_on_the_genomic_accession=8982486, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392428 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392428, CurrentID=0, Status='live', Symbol='NPM1P9', Aliases='', description='nucleophosmin 1 (nucleolar phosphoprotein B23, numatrin) pseudogene 9', other_designations='nucleophosmin (nucleolar phosphoprotein B23, numatrin) pseudogene', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=14834158, end_position_on_the_genomic_accession=14835417, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  102723549 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723549, CurrentID=0, Status='live', Symbol='LOC102723549', Aliases='', description='uncharacterized LOC102723549', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=30442719, end_position_on_the_genomic_accession=30446648, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  392433 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392433, CurrentID=0, Status='live', Symbol='MAGEB6P1', Aliases='MAGEB6B', description='MAGE family member B6 pseudogene 1', other_designations='melanoma antigen family B, 6 pseudogene 1|melanoma antigen family B, 6B pseudogene|melanoma antigen family B6 pseudogene 1', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=26160526, end_position_on_the_genomic_accession=26162413, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392435 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392435, CurrentID=0, Status='live', Symbol='LOC392435', Aliases='', description='MAGE family member B3 pseudogene', other_designations='melanoma antigen family B, 3 pseudogene|melanoma antigen family B3 pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27832524, end_position_on_the_genomic_accession=27833003, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392436 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392436, CurrentID=0, Status='live', Symbol='LOC392436', Aliases='', description='MAGE family member B4 pseudogene', other_designations='MAGE-B13 pseudogene|melanoma antigen family B, 4 pseudogene|melanoma antigen family B4 pseudogene', map_location='Xp21.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=27854192, end_position_on_the_genomic_accession=27855354, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392437 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392437, CurrentID=0, Status='live', Symbol='FTLP2', Aliases='FTLL2', description='ferritin, light polypeptide pseudogene 2', other_designations='ferritin, light polypeptide-like 2 pseudogene', map_location='Xp21.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=30630097, end_position_on_the_genomic_accession=30630975, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392439 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392439, CurrentID=0, Status='live', Symbol='LOC392439', Aliases='', description='serine/arginine-rich splicing factor 2 pseudogene', other_designations='', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=34387722, end_position_on_the_genomic_accession=34388437, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392440 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392440, CurrentID=0, Status='live', Symbol='LOC392440', Aliases='', description='MAGE family member B18 pseudogene', other_designations='melanoma antigen family B, 18 pseudogene|melanoma antigen family B18 pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=35058966, end_position_on_the_genomic_accession=35059973, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283860 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283860, CurrentID=0, Status='live', Symbol='LINC00304', Aliases='C16orf81, NCRNA00304', description='long intergenic non-protein coding RNA 304', other_designations='', map_location='16q24.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=89159220, end_position_on_the_genomic_accession=89163675, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392442, CurrentID=0, Status='live', Symbol='LOC392442', Aliases='', description='BCL2 associated athanogene 1 pseudogene', other_designations='BCL2-associated athanogene pseudogene', map_location='Xp11.4', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=38479252, end_position_on_the_genomic_accession=38480906, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422869 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422869, CurrentID=0, Status='live', Symbol='MIR3152', Aliases='', description='microRNA 3152', other_designations='hsa-mir-3152', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=18573306, end_position_on_the_genomic_accession=18573379, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392451 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392451, CurrentID=0, Status='live', Symbol='RPL19P20', Aliases='RPL19_12_1775', description='ribosomal protein L19 pseudogene 20', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=44649093, end_position_on_the_genomic_accession=44649798, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392452 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392452, CurrentID=0, Status='live', Symbol='LOC392452', Aliases='', description='mitochondrial fission factor pseudogene', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=45731088, end_position_on_the_genomic_accession=45731757, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102723324 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=102723324, CurrentID=0, Status='live', Symbol='LOC102723324', Aliases='', description='uncharacterized LOC102723324', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=121370444, end_position_on_the_genomic_accession=121390665, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  392454 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392454, CurrentID=0, Status='live', Symbol='PCNAP3', Aliases='', description='proliferating cell nuclear antigen pseudogene 3', other_designations='', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46058552, end_position_on_the_genomic_accession=46059657, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392456 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392456, CurrentID=0, Status='live', Symbol='LOC392456', Aliases='', description='catenin beta like 1 pseudogene', other_designations='catenin, beta like 1 pseudogene', map_location='Xp11.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=46418339, end_position_on_the_genomic_accession=46420097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392459 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392459, CurrentID=0, Status='live', Symbol='CXXC1P1', Aliases='CXorf25, NCRNA00236', description='CXXC finger protein 1 pseudogene 1', other_designations='CXXC finger 1 (PHD domain) pseudogene|Putative uncharacterized protein CXorf25', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=47707191, end_position_on_the_genomic_accession=47736628, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101060442 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101060442, CurrentID=0, Status='live', Symbol='LOC101060442', Aliases='', description='calcium-binding mitochondrial carrier protein SCaMC-1-like', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  645769 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645769, CurrentID=0, Status='live', Symbol='LOC645769', Aliases='', description='uncharacterized LOC645769', other_designations='', map_location='Xp22.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=13652497, end_position_on_the_genomic_accession=13653433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392465 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392465, CurrentID=0, Status='live', Symbol='GLOD5', Aliases='', description='glyoxalase domain containing 5', other_designations='', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=48761706, end_position_on_the_genomic_accession=48773648, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  196483 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=196483, CurrentID=0, Status='live', Symbol='EEF2KMT', Aliases='FAM86A, SB153, eEF2-KMT', description='eukaryotic elongation factor 2 lysine methyltransferase', other_designations='eEF2-lysine methyltransferase|family with sequence similarity 86, member A|protein FAM86A|putative protein N-methyltransferase FAM86A', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=5084300, end_position_on_the_genomic_accession=5097820, orientation='minus', exon_count=9, OMIM=615263, no_hdr0=''),
  105379779 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379779, CurrentID=0, Status='live', Symbol='LOC105379779', Aliases='', description='collagen alpha-1(I) chain-like', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  392477 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392477, CurrentID=0, Status='live', Symbol='LOC392477', Aliases='', description='malate dehydrogenase 1, NAD (soluble) pseudogene', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57766318, end_position_on_the_genomic_accession=57767522, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392479 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392479, CurrentID=0, Status='live', Symbol='KRT8P17', Aliases='', description='keratin 8 pseudogene 17', other_designations='', map_location='Xp11.21', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=57984605, end_position_on_the_genomic_accession=57986331, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283867 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283867, CurrentID=0, Status='live', Symbol='LINC00922', Aliases='', description='long intergenic non-protein coding RNA 922', other_designations='', map_location='16q21', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=65284499, end_position_on_the_genomic_accession=65576300, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  130340 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130340, CurrentID=0, Status='live', Symbol='AP1S3', Aliases='PSORS15', description='adaptor related protein complex 1 sigma 3 subunit', other_designations='adapter-related protein complex 1 subunit sigma-1C|adaptor protein complex AP-1 sigma-1C subunit|adaptor-related protein complex 1 sigma 3 subunit|adaptor-related protein complex 1 subunit sigma-1C|adaptor-related protein complex 1, sigma 3 subunit|clathrin assembly protein complex 1 sigma-1C small chain|golgi adaptor HA1/AP1 adaptin sigma-1C subunit|sigma 1C subunit of AP-1 clathrin', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=223755330, end_position_on_the_genomic_accession=223837602, orientation='minus', exon_count=6, OMIM=615781, no_hdr0=''),
  392485 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392485, CurrentID=0, Status='live', Symbol='LOC392485', Aliases='', description='ataxin 7-like 3 pseudogene', other_designations='', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66055590, end_position_on_the_genomic_accession=66056585, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392486 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392486, CurrentID=0, Status='live', Symbol='GPR165P', Aliases='PGR15', description='G protein-coupled receptor 165 pseudogene', other_designations='6530406P05Rik|G protein-coupled receptor 83 pseudogene|G protein-coupled receptor orphanA7', map_location='Xq12', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=66437967, end_position_on_the_genomic_accession=66441779, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392489 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392489, CurrentID=0, Status='live', Symbol='LOC392489', Aliases='', description='ras homolog family member G pseudogene', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=71352418, end_position_on_the_genomic_accession=71352964, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392490 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392490, CurrentID=0, Status='live', Symbol='FLJ44635', Aliases='', description='TPT1-like protein', other_designations='', map_location='Xq13.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72144184, end_position_on_the_genomic_accession=72161750, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105372796 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372796, CurrentID=0, Status='live', Symbol='LOC105372796', Aliases='', description='uncharacterized LOC105372796', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37341405, end_position_on_the_genomic_accession=37344292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100422877 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422877, CurrentID=0, Status='live', Symbol='MIR3202-2', Aliases='mir-3202-2', description='microRNA 3202-2', other_designations='hsa-mir-3202-2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=153981098, end_position_on_the_genomic_accession=153981176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392497, CurrentID=0, Status='live', Symbol='RPS6P26', Aliases='RPS6_11_1786', description='ribosomal protein S6 pseudogene 26', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=74376090, end_position_on_the_genomic_accession=74376926, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130355 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130355, CurrentID=0, Status='live', Symbol='C2orf76', Aliases='AIM29', description='chromosome 2 open reading frame 76', other_designations='', map_location='2q14.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=119300442, end_position_on_the_genomic_accession=119366826, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100422878 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422878, CurrentID=0, Status='live', Symbol='MIR3168', Aliases='', description='microRNA 3168', other_designations='hsa-mir-3168', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=41101019, end_position_on_the_genomic_accession=41101100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392505 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392505, CurrentID=0, Status='live', Symbol='RPL6P29', Aliases='RPL6_17_1796', description='ribosomal protein L6 pseudogene 29', other_designations='', map_location='Xq21.33', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=98251662, end_position_on_the_genomic_accession=98252576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422879 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422879, CurrentID=0, Status='live', Symbol='MIR3124', Aliases='mir-3124', description='microRNA 3124', other_designations='hsa-mir-3124', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=248826377, end_position_on_the_genomic_accession=248826443, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392508 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392508, CurrentID=0, Status='live', Symbol='LOC392508', Aliases='', description='RAD21 homolog (S. pombe) pseudogene', other_designations='', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100801177, end_position_on_the_genomic_accession=100803059, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392509 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392509, CurrentID=0, Status='live', Symbol='ARL13A', Aliases='ARL13, dJ341D10.2', description='ADP ribosylation factor like GTPase 13A', other_designations='ADP-ribosylation factor like GTPase 13A|ADP-ribosylation factor-like 13|ADP-ribosylation factor-like 13A', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=100969040, end_position_on_the_genomic_accession=100990831, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  392510 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392510, CurrentID=0, Status='live', Symbol='YWHAQP8', Aliases='', description='YWHAQ pseudogene 8', other_designations='tyrosine 3/tryptophan 5 -monooxygenase activation protein, theta polypeptide pseudogene', map_location='Xq22.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=101190459, end_position_on_the_genomic_accession=101191472, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130367 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130367, CurrentID=0, Status='live', Symbol='SGPP2', Aliases='SPP2', description='sphingosine-1-phosphate phosphatase 2', other_designations='SPPase2|hSPP2|sphingosine 1-phosphate phosphohydrolase 2|sphingosine-1-phosphatase 2|sphingosine-1-phosphate phosphotase 2', map_location='2q36.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=222424052, end_position_on_the_genomic_accession=222561842, orientation='plus', exon_count=6, OMIM=612827, no_hdr0=''),
  100130801 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130801, CurrentID=0, Status='live', Symbol='LOC100130801', Aliases='', description='lupus La protein pseudogene', other_designations='', map_location='9p23', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=13021632, end_position_on_the_genomic_accession=13049754, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392516 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392516, CurrentID=0, Status='live', Symbol='KRT18P49', Aliases='', description='keratin 18 pseudogene 49', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107428627, end_position_on_the_genomic_accession=107429520, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392517 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392517, CurrentID=0, Status='live', Symbol='NCBP2L', Aliases='dJ820B18.1', description='nuclear cap binding protein subunit 2-like', other_designations='', map_location='Xq22.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=107794261, end_position_on_the_genomic_accession=107794677, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392520 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392520, CurrentID=0, Status='live', Symbol='RPS5P7', Aliases='RPS5_3_1801', description='ribosomal protein S5 pseudogene 7', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=109853248, end_position_on_the_genomic_accession=109853968, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392522 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392522, CurrentID=0, Status='live', Symbol='RPL18AP15', Aliases='RPL18A_4_1802', description='ribosomal protein L18a pseudogene 15', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=111624163, end_position_on_the_genomic_accession=111624773, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392525, CurrentID=0, Status='live', Symbol='AKR1B1P8', Aliases='', description='aldo-keto reductase family 1, member B1 pseudogene 8', other_designations='aldo-keto reductase family 1, member B1 (aldose reductase) pseudogene', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116043451, end_position_on_the_genomic_accession=116044622, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392529 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392529, CurrentID=0, Status='live', Symbol='LOC392529', Aliases='', description='cordon-bleu WH2 repeat protein pseudogene', other_designations='cordon-bleu homolog pseudogene', map_location='Xq24', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=118938157, end_position_on_the_genomic_accession=118939148, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283875 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283875, CurrentID=0, Status='live', Symbol='LINC00514', Aliases='LA16c-380H5.1', description='long intergenic non-protein coding RNA 514', other_designations='', map_location='16p13.3', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=2989054, end_position_on_the_genomic_accession=2994509, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  392536 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392536, CurrentID=0, Status='live', Symbol='LOC392536', Aliases='', description='filamin binding LIM protein 1 pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=125326900, end_position_on_the_genomic_accession=125328021, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422884 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422884, CurrentID=0, Status='live', Symbol='MIR548U', Aliases='', description='microRNA 548u', other_designations='hsa-mir-548u', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000006.12', start_position_on_the_genomic_accession=57390132, end_position_on_the_genomic_accession=57390212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  392538 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392538, CurrentID=0, Status='live', Symbol='LOC392538', Aliases='', description='RuvB-like AAA ATPase 2 pseudogene', other_designations='', map_location='Xq25', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=128440929, end_position_on_the_genomic_accession=128441782, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  105375415 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375415, CurrentID=0, Status='live', Symbol='LOC105375415', Aliases='', description='uncharacterized LOC105375415', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=96945553, end_position_on_the_genomic_accession=96960417, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  392540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392540, CurrentID=0, Status='live', Symbol='OR11N1P', Aliases='', description='olfactory receptor family 11 subfamily N member 1 pseudogene', other_designations='olfactory receptor, family 11, subfamily N, member 1 pseudogene', map_location='Xq26.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=131324334, end_position_on_the_genomic_accession=131325313, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130399 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130399, CurrentID=0, Status='live', Symbol='ACVR1C', Aliases='ACVRLK7, ALK7', description='activin A receptor type IC', other_designations='ACTR-IC|ALK-7|activin A receptor, type IC|activin receptor type IC|activin receptor-like kinase 7', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=157526767, end_position_on_the_genomic_accession=157628887, orientation='minus', exon_count=10, OMIM=608981, no_hdr0=''),
  392545 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392545, CurrentID=0, Status='live', Symbol='HMGB3P31', Aliases='', description='high mobility group box 3 pseudogene 31', other_designations='high mobility group protein B3-like protein-like pseudogene|high-mobility group box 3 pseudogene 31', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=134961651, end_position_on_the_genomic_accession=134962404, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392547 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392547, CurrentID=0, Status='live', Symbol='CROCCP1', Aliases='CROCCP', description='ciliary rootlet coiled-coil, rootletin pseudogene 1', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=135432675, end_position_on_the_genomic_accession=135438912, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392552 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392552, CurrentID=0, Status='live', Symbol='KRT8P6', Aliases='', description='keratin 8 pseudogene 6', other_designations='', map_location='Xq26.3', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=138487470, end_position_on_the_genomic_accession=138488882, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392555 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392555, CurrentID=0, Status='live', Symbol='LOC392555', Aliases='', description='MAGE family member C2 pseudogene', other_designations='melanoma antigen family C, 2 pseudogene|melanoma antigen family C2 pseudogene', map_location='Xq27.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=142172670, end_position_on_the_genomic_accession=142174057, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392557, CurrentID=0, Status='live', Symbol='RPL19P21', Aliases='RPL19_13_1819', description='ribosomal protein L19 pseudogene 21', other_designations='', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=150993465, end_position_on_the_genomic_accession=150994170, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392561, CurrentID=0, Status='live', Symbol='OR3B1P', Aliases='', description='olfactory receptor family 3 subfamily B member 1 pseudogene', other_designations='olfactory receptor, family 3, subfamily B, member 1 pseudogene', map_location='Xq28', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=154658733, end_position_on_the_genomic_accession=154659661, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  101928985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101928985, CurrentID=0, Status='live', Symbol='LOC101928985', Aliases='', description='uncharacterized LOC101928985', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000011.10', start_position_on_the_genomic_accession=115732954, end_position_on_the_genomic_accession=115734740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  130429 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130429, CurrentID=0, Status='live', Symbol='RPSAP27', Aliases='RPSA_8_320', description='ribosomal protein SA pseudogene 27', other_designations='', map_location='2q34', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=208208059, end_position_on_the_genomic_accession=208209066, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392597 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392597, CurrentID=0, Status='live', Symbol='TRAPPC2P7', Aliases='SEDLP7', description='trafficking protein particle complex 2 pseudogene 7', other_designations='spondyloepiphyseal dysplasia, late, pseudogene 7', map_location='Yq11.23', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=18654056, end_position_on_the_genomic_accession=18656177, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422895 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422895, CurrentID=0, Status='live', Symbol='MIR4294', Aliases='', description='microRNA 4294', other_designations='hsa-mir-4294', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000010.11', start_position_on_the_genomic_accession=48985512, end_position_on_the_genomic_accession=48985587, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422896 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422896, CurrentID=0, Status='live', Symbol='MIR3140', Aliases='mir-3140', description='microRNA 3140', other_designations='hsa-mir-3140', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=152489327, end_position_on_the_genomic_accession=152489416, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392617, CurrentID=0, Status='live', Symbol='ELFN1', Aliases='PPP1R28', description='extracellular leucine-rich repeat and fibronectin type III domain containing 1', other_designations='extracellular leucine-rich repeat and fibronectin type III containing 1|extracellular leucine-rich repeat and fibronectin type-III domain-containing protein 1|protein phosphatase 1 regulatory subunit 28|protein phosphatase 1, regulatory subunit 28', map_location='7p22.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=1666056, end_position_on_the_genomic_accession=1747954, orientation='plus', exon_count=10, OMIM=614964, no_hdr0=''),
  105372272 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372272, CurrentID=0, Status='live', Symbol='LOC105372272', Aliases='', description='uncharacterized LOC105372272', other_designations='CTD-2369P2.5', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=10260103, end_position_on_the_genomic_accession=10283474, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  392636 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392636, CurrentID=0, Status='live', Symbol='AGMO', Aliases='TMEM195', description='alkylglycerol monooxygenase', other_designations='glyceryl-ether monooxygenase|transmembrane protein 195', map_location='7p21.2', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=15159124, end_position_on_the_genomic_accession=15562015, orientation='minus', exon_count=16, OMIM=613738, no_hdr0=''),
  654783 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654783, CurrentID=0, Status='live', Symbol='IFNNP1', Aliases='', description='interferon, nu 1, pseudogene', other_designations='', map_location='9p21.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=21058771, end_position_on_the_genomic_accession=21059323, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130497 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130497, CurrentID=0, Status='live', Symbol='OSR1', Aliases='ODD', description='odd-skipped related transciption factor 1', other_designations='odd-skipped homolog|odd-skipped related 1', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19351485, end_position_on_the_genomic_accession=19358611, orientation='minus', exon_count=4, OMIM=608891, no_hdr0=''),
  105372797 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372797, CurrentID=0, Status='live', Symbol='LOC105372797', Aliases='', description='uncharacterized LOC105372797', other_designations='', map_location='', chromosome='21', genomic_nucleotide_accession_version='NC_000021.9', start_position_on_the_genomic_accession=37478890, end_position_on_the_genomic_accession=37490168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  130500 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130500, CurrentID=0, Status='live', Symbol='CISD1P1', Aliases='CISD1B, ZCD1B', description='CDGSH iron sulfur domain 1 pseudogene 1', other_designations='CDGSH iron sulfur domain 1B pseudogene|zinc finger, CDGSH-type domain 1B', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19826212, end_position_on_the_genomic_accession=19826907, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  130502 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130502, CurrentID=0, Status='live', Symbol='TTC32', Aliases='', description='tetratricopeptide repeat domain 32', other_designations='TPR repeat protein 32', map_location='2p24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=19896753, end_position_on_the_genomic_accession=19901986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  130507 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130507, CurrentID=0, Status='live', Symbol='UBR3', Aliases='ZNF650', description='ubiquitin protein ligase E3 component n-recognin 3 (putative)', other_designations='N-recognin-3|ubiquitin-protein ligase E3-alpha-3|ubiquitin-protein ligase E3-alpha-III|zinc finger protein 650', map_location='2q31.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=169827470, end_position_on_the_genomic_accession=170084131, orientation='plus', exon_count=43, OMIM=613831, no_hdr0=''),
  284561 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=284561, CurrentID=0, Status='live', Symbol='LOC284561', Aliases='', description='uncharacterized LOC284561', other_designations='', map_location='1q21.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=148279042, end_position_on_the_genomic_accession=148291861, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422905 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422905, CurrentID=0, Status='live', Symbol='MIR4311', Aliases='', description='microRNA 4311', other_designations='hsa-mir-4311', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=66040233, end_position_on_the_genomic_accession=66040332, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  654813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654813, CurrentID=0, Status='live', Symbol='P2RY10P1', Aliases='GPR121P, P2Y10P1', description='purinergic receptor P2Y, G-protein coupled, 10 pseudogene 1', other_designations='', map_location='Xq21.1', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  654816 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654816, CurrentID=0, Status='live', Symbol='NCF1B', Aliases='NCF-1B, SH3PXD1B', description='neutrophil cytosolic factor 1B pseudogene', other_designations='putative SH3 and PX domain-containing protein 1B|putative neutrophil cytosol factor 1B', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=73220639, end_position_on_the_genomic_accession=73235945, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  654817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654817, CurrentID=0, Status='live', Symbol='NCF1C', Aliases='SH3PXD1C', description='neutrophil cytosolic factor 1C pseudogene', other_designations='', map_location='7q11.23', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=75156578, end_position_on_the_genomic_accession=75171998, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100422907 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422907, CurrentID=0, Status='live', Symbol='MIR3156-2', Aliases='', description='microRNA 3156-2', other_designations='hsa-mir-3156-2', map_location='', chromosome='18', genomic_nucleotide_accession_version='NC_000018.10', start_position_on_the_genomic_accession=14830166, end_position_on_the_genomic_accession=14830242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  130535 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130535, CurrentID=0, Status='live', Symbol='KCTD18', Aliases='6530404F10Rik', description='potassium channel tetramerization domain containing 18', other_designations='potassium channel tetramerisation domain containing 18', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=200488952, end_position_on_the_genomic_accession=200510094, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  130540 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130540, CurrentID=0, Status='live', Symbol='ALS2CR12', Aliases='', description='amyotrophic lateral sclerosis 2 chromosome region candidate 12', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 12|amyotrophic lateral sclerosis 2 chromosome region, candidate 12', map_location='2q33.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=201288271, end_position_on_the_genomic_accession=201357398, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  105375416 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375416, CurrentID=0, Status='live', Symbol='LOC105375416', Aliases='', description='uncharacterized LOC105375416', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97328530, end_position_on_the_genomic_accession=97372151, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  654835 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654835, CurrentID=0, Status='live', Symbol='FAM138C', Aliases='F379', description='family with sequence similarity 138 member C', other_designations='F379 retina-specific protein|family with sequence similarity 138, member C', map_location='9p24.3', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=34394, end_position_on_the_genomic_accession=35864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422910 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422910, CurrentID=0, Status='live', Symbol='MIR2861', Aliases='BMND15, MIRN2861', description='microRNA 2861', other_designations='hsa-mir-2861', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=127785918, end_position_on_the_genomic_accession=127786007, orientation='plus', exon_count=1, OMIM=613405, no_hdr0=''),
  654841 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=654841, CurrentID=0, Status='live', Symbol='LOC654841', Aliases='', description='uncharacterized LOC654841', other_designations='', map_location='2q36-q37', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=227221052, end_position_on_the_genomic_accession=227325164, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100422911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422911, CurrentID=0, Status='live', Symbol='MIR500B', Aliases='mir-500b', description='microRNA 500b', other_designations='hsa-mir-500b', map_location='Xp11.23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=50010672, end_position_on_the_genomic_accession=50010750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  130557 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130557, CurrentID=0, Status='live', Symbol='ZNF513', Aliases='HMFT0656, RP58', description='zinc finger protein 513', other_designations='', map_location='2p23.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=27377231, end_position_on_the_genomic_accession=27380866, orientation='minus', exon_count=5, OMIM=613598, no_hdr0=''),
  130560 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130560, CurrentID=0, Status='live', Symbol='SPATA3', Aliases='TSARG1', description='spermatogenesis associated 3', other_designations='testis and spermatogenesis cell apoptosis related gene 1|testis and spermatogenesis cell-related protein 1|testis spermatocyte apoptosis-related protein 1', map_location='2q37.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=230996124, end_position_on_the_genomic_accession=231025056, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100533622 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100533622, CurrentID=0, Status='live', Symbol='LOC100533622', Aliases='', description='sulfite oxidase pseudogene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000008.11', start_position_on_the_genomic_accession=83200952, end_position_on_the_genomic_accession=83202431, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392713 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392713, CurrentID=0, Status='live', Symbol='PMS2P12', Aliases='PMS2L12, PMSL12', description='PMS1 homolog 2, mismatch repair system component pseudogene 12', other_designations='postmeiotic segregation increased 2 pseudogene 12|postmeiotic segregation increased 2-like 12 pseudogene|postmeiotic segregation increased 2-like protein 12', map_location='7q22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=102337315, end_position_on_the_genomic_accession=102339116, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100418574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100418574, CurrentID=0, Status='live', Symbol='LOC100418574', Aliases='', description='ubiquitin specific peptidase 8 pseudogene', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  130574 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130574, CurrentID=0, Status='live', Symbol='LYPD6', Aliases='', description='LY6/PLAUR domain containing 6', other_designations='', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149329985, end_position_on_the_genomic_accession=149474148, orientation='plus', exon_count=6, OMIM=613359, no_hdr0=''),
  130576 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130576, CurrentID=0, Status='live', Symbol='LYPD6B', Aliases='CT116, LYPD7', description='LY6/PLAUR domain containing 6B', other_designations='cancer/testis antigen 116', map_location='2q23.2', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=149038467, end_position_on_the_genomic_accession=149215262, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  130589 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130589, CurrentID=0, Status='live', Symbol='GALM', Aliases='BLOCK25, GLAT, HEL-S-63p, IBD1', description='galactose mutarotase (aldose 1-epimerase)', other_designations='epididymis secretory sperm binding protein Li 63p|galactomutarotase|galactose enzyme activator', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=38665910, end_position_on_the_genomic_accession=38734767, orientation='plus', exon_count=9, OMIM=137030, no_hdr0=''),
  100130170 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130170, CurrentID=0, Status='live', Symbol='LOC100130170', Aliases='', description='chromosome 12 open reading frame 49 pseudogene', other_designations='', map_location='Xq26.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=130645601, end_position_on_the_genomic_accession=130646526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283911 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283911, CurrentID=0, Status='live', Symbol='LOC283911', Aliases='', description='uncharacterized LOC283911', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  392748 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392748, CurrentID=0, Status='live', Symbol='RPS27P17', Aliases='RPS27_5_819', description='ribosomal protein S27 pseudogene 17', other_designations='', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=93669769, end_position_on_the_genomic_accession=93670111, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  392752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392752, CurrentID=0, Status='live', Symbol='OR7E7P', Aliases='', description='olfactory receptor family 7 subfamily E member 7 pseudogene', other_designations='olfactory receptor, family 7, subfamily E, member 7 pseudogene', map_location='7q21.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97946983, end_position_on_the_genomic_accession=97947989, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130612 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130612, CurrentID=0, Status='live', Symbol='TMEM198', Aliases='TMEM198A', description='transmembrane protein 198', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219543663, end_position_on_the_genomic_accession=219550595, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105372598 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372598, CurrentID=0, Status='live', Symbol='LOC105372598', Aliases='', description='uncharacterized LOC105372598', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=33989480, end_position_on_the_genomic_accession=33992292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  130617 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130617, CurrentID=0, Status='live', Symbol='ZFAND2B', Aliases='AIRAPL', description='zinc finger, AN1-type domain 2B', other_designations='AIRAP-like protein|arsenite-inducible RNA-associated protein-like protein', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219206784, end_position_on_the_genomic_accession=219209651, orientation='plus', exon_count=10, OMIM=613474, no_hdr0=''),
  283914 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283914, CurrentID=0, Status='live', Symbol='LINC01566', Aliases='', description='long intergenic non-protein coding RNA 1566', other_designations='', map_location='16p11.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=35363416, end_position_on_the_genomic_accession=35391713, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105379781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105379781, CurrentID=0, Status='live', Symbol='LOC105379781', Aliases='', description='uncharacterized LOC105379781', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  392781 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392781, CurrentID=0, Status='live', Symbol='RPS2P31', Aliases='RPS2_15_841', description='ribosomal protein S2 pseudogene 31', other_designations='', map_location='7q31.33', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=124480502, end_position_on_the_genomic_accession=124481436, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  152845 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=152845, CurrentID=0, Status='live', Symbol='LOC152845', Aliases='', description='pleiomorphic adenoma gene-like 2 pseudogene', other_designations='', map_location='4q21.23', chromosome='4', genomic_nucleotide_accession_version='NC_000004.12', start_position_on_the_genomic_accession=84244003, end_position_on_the_genomic_accession=84245488, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  392787 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392787, CurrentID=0, Status='live', Symbol='LOC392787', Aliases='', description='filamin C pseudogene', other_designations='', map_location='7q32.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=128912660, end_position_on_the_genomic_accession=128923070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100379236 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100379236, CurrentID=0, Status='live', Symbol='LOC100379236', Aliases='', description='patatin-like phospholipase domain containing 4 pseudogene 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=14100967, end_position_on_the_genomic_accession=14101721, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  261726 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261726, CurrentID=0, Status='live', Symbol='TIPRL', Aliases='TIP, TIP41', description='TOR signaling pathway regulator', other_designations='TIP41, TOR signaling pathway regulator-like|putative MAPK-activating protein PM10|type 2A-interacting protein', map_location='1q23.2', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=168178845, end_position_on_the_genomic_accession=168202114, orientation='plus', exon_count=9, OMIM=611807, no_hdr0=''),
  261727 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261727, CurrentID=0, Status='live', Symbol='CRCL', Aliases='', description='Creatinine clearance QTL', other_designations='', map_location='3p', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM=607135, no_hdr0=''),
  261729 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261729, CurrentID=0, Status='live', Symbol='STEAP2', Aliases='IPCA1, PCANAP1, PUMPCn, STAMP1, STMP', description='STEAP2 metalloreductase', other_designations='STEAP family member 2, metalloreductase|SixTransMembrane Protein of Prostate 1|prostate cancer associated protein 1|prostate cancer-associated protein 1|protein up-regulated in metastatic prostate cancer|protein upregulated in metastatic prostate cancer|six transmembrane epithelial antigen of prostate 2|six transmembrane epithelial antigen of the prostate 2|six-transmembrane epithelial antigen of prostate 2', map_location='7q21.13', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=90211686, end_position_on_the_genomic_accession=90237678, orientation='plus', exon_count=8, OMIM=605094, no_hdr0=''),
  261734 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261734, CurrentID=0, Status='live', Symbol='NPHP4', Aliases='POC10, SLSN4', description='nephronophthisis 4', other_designations='POC10 centriolar protein homolog|nephroretinin', map_location='1p36', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=5862808, end_position_on_the_genomic_accession=5992505, orientation='minus', exon_count=36, OMIM=607215, no_hdr0=''),
  261735 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=261735, CurrentID=0, Status='live', Symbol='TVP23CP2', Aliases='CGI-148P', description='TVP23C pseudogene 2', other_designations='family with sequence similarity 18, member B1 pseudogene|family with sequence similarity 18, member B2 pseudogene|trans-golgi network vesicle protein 23 homolog C pseudogene 2', map_location='16p13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=14200332, end_position_on_the_genomic_accession=14202119, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283922 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283922, CurrentID=0, Status='live', Symbol='LOC283922', Aliases='', description='pyruvate dehydrogenase phosphatase regulatory subunit pseudogene', other_designations='', map_location='16q23.1', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=74332406, end_position_on_the_genomic_accession=74368255, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  130678 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130678, CurrentID=0, Status='live', Symbol='RPLP1P5', Aliases='RPLP1_1_196', description='ribosomal protein, large, P1 pseudogene 5', other_designations='', map_location='2p24.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=15869810, end_position_on_the_genomic_accession=15870280, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100126447 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126447, CurrentID=0, Status='live', Symbol='LOC100126447', Aliases='', description='uncharacterized LOC100126447', other_designations='', map_location='Xq23', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=116810931, end_position_on_the_genomic_accession=116817860, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105375417 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375417, CurrentID=0, Status='live', Symbol='LOC105375417', Aliases='', description='uncharacterized LOC105375417', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=97159617, end_position_on_the_genomic_accession=97177389, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  392843 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392843, CurrentID=0, Status='live', Symbol='IQCA1L', Aliases='IQCA1P1', description='IQ motif containing with AAA domain 1 like', other_designations='IQ and AAA domain-containing protein 1 pseudogene 1|IQ motif containing with AAA domain 1-like protein|putative IQ and AAA domain-containing protein 1 pseudogene 1|putative IQ and AAA domain-containing protein 1-like|putative IQ motif containing with AAA domain 1 pseudogene 1', map_location='7q36.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=151190873, end_position_on_the_genomic_accession=151205519, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  130700 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130700, CurrentID=0, Status='live', Symbol='PSMB3P2', Aliases='', description='proteasome (prosome, macropain) subunit, beta type, 3 pseudogene 2', other_designations='', map_location='2q35', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=216610286, end_position_on_the_genomic_accession=216611061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100126448 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100126448, CurrentID=0, Status='live', Symbol='LOC100126448', Aliases='', description='uncharacterized LOC100126448', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  392862 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392862, CurrentID=0, Status='live', Symbol='GRID2IP', Aliases='DELPHILIN', description='glutamate receptor, ionotropic, delta 2 (Grid2) interacting protein', other_designations='glutamate receptor, ionotropic, delta 2 (Grid2) interacting protein 1|glutamate receptor, ionotropic, delta 2-interacting protein 1', map_location='7p22.1', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=6496778, end_position_on_the_genomic_accession=6551436, orientation='minus', exon_count=23, OMIM=610639, no_hdr0=''),
  392872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392872, CurrentID=0, Status='live', Symbol='RPL12P10', Aliases='RPL12_8_793', description='ribosomal protein L12 pseudogene 10', other_designations='TCAG_1783340', map_location='7p15.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=22881821, end_position_on_the_genomic_accession=22882432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  283932 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283932, CurrentID=0, Status='live', Symbol='FBXL19-AS1', Aliases='NCRNA00095', description='FBXL19 antisense RNA 1 (head to head)', other_designations='FBXL19 antisense RNA 1 (non-protein coding)', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=30919319, end_position_on_the_genomic_accession=30923269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  130733 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130733, CurrentID=0, Status='live', Symbol='TMEM178A', Aliases='TMEM178', description='transmembrane protein 178A', other_designations='transmembrane protein 178', map_location='2p22.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=39665498, end_position_on_the_genomic_accession=39717963, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  283933 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283933, CurrentID=0, Status='live', Symbol='ZNF843', Aliases='', description='zinc finger protein 843', other_designations='', map_location='16p11.2', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=31435564, end_position_on_the_genomic_accession=31443027, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  645938 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=645938, CurrentID=0, Status='live', Symbol='LOC645938', Aliases='', description='capping protein (actin filament) muscle Z-line, alpha 1 pseudogene', other_designations='', map_location='Xq13.2', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=72727980, end_position_on_the_genomic_accession=72728821, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130749 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130749, CurrentID=0, Status='live', Symbol='CPO', Aliases='', description='carboxypeptidase O', other_designations='metallocarboxypeptidase C|metallocarboxypeptidase O', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206879383, end_position_on_the_genomic_accession=206971064, orientation='plus', exon_count=14, OMIM=609563, no_hdr0=''),
  401525 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401525, CurrentID=0, Status='live', Symbol='CDK2AP2P3', Aliases='', description='cyclin-dependent kinase 2 associated protein 2 pseudogene 3', other_designations='CDK2-associated protein 2 pseudogene', map_location='9q13', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=63774448, end_position_on_the_genomic_accession=63775347, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130752 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130752, CurrentID=0, Status='live', Symbol='MDH1B', Aliases='RP11-95H11', description='malate dehydrogenase 1B', other_designations='malate dehydrogenase 1B, NAD (soluble)', map_location='2q33.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206737763, end_position_on_the_genomic_accession=206765549, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105371053 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105371053, CurrentID=0, Status='live', Symbol='LOC105371053', Aliases='', description='uncharacterized LOC105371053', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  130773 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130773, CurrentID=0, Status='live', Symbol='RPL23AP37', Aliases='RPL23A_5_234', description='ribosomal protein L23a pseudogene 37', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=64347255, end_position_on_the_genomic_accession=64347745, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100130815 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130815, CurrentID=0, Status='live', Symbol='LOC100130815', Aliases='', description='uncharacterized LOC100130815', other_designations='', map_location='14q32.2', chromosome='14', genomic_nucleotide_accession_version='NC_000014.8', start_position_on_the_genomic_accession=96741383, end_position_on_the_genomic_accession=96743060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100132790 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100132790, CurrentID=0, Status='live', Symbol='LOC100132790', Aliases='', description='uncharacterized LOC100132790', other_designations='', map_location='9p12', chromosome='9', genomic_nucleotide_accession_version='NC_000009.11', start_position_on_the_genomic_accession=42796892, end_position_on_the_genomic_accession=42798049, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  130802 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130802, CurrentID=0, Status='live', Symbol='DNMT3AP1', Aliases='', description='DNA methyltransferase 3A pseudogene 1', other_designations='', map_location='2p14', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=66819881, end_position_on_the_genomic_accession=66822628, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  130813 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130813, CurrentID=0, Status='live', Symbol='C2orf50', Aliases='', description='chromosome 2 open reading frame 50', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11133053, end_position_on_the_genomic_accession=11154564, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  130814 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130814, CurrentID=0, Status='live', Symbol='PQLC3', Aliases='C2orf22', description='PQ loop repeat containing 3', other_designations='', map_location='2p25.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=11155372, end_position_on_the_genomic_accession=11178874, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105373672 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105373672, CurrentID=0, Status='live', Symbol='LOC105373672', Aliases='', description='uncharacterized LOC105373672', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=148111315, end_position_on_the_genomic_accession=148118379, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100422956 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422956, CurrentID=0, Status='live', Symbol='MIR3180-2', Aliases='mir-3180-2', description='microRNA 3180-2', other_designations='hsa-mir-3180-2', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000016.10', start_position_on_the_genomic_accession=16309879, end_position_on_the_genomic_accession=16309966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  130827 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130827, CurrentID=0, Status='live', Symbol='TMEM182', Aliases='', description='transmembrane protein 182', other_designations='', map_location='2q12.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=102736908, end_position_on_the_genomic_accession=102843838, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100422957 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422957, CurrentID=0, Status='live', Symbol='MIR3131', Aliases='mir-3131', description='microRNA 3131', other_designations='hsa-mir-3131', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219058688, end_position_on_the_genomic_accession=219058750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  392979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=392979, CurrentID=0, Status='live', Symbol='RPL18P4', Aliases='RPL18_1_838', description='ribosomal protein L18 pseudogene 4', other_designations='', map_location='7q31.31', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=121440812, end_position_on_the_genomic_accession=121441261, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422959 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422959, CurrentID=0, Status='live', Symbol='MIR4268', Aliases='', description='microRNA 4268', other_designations='hsa-mir-4268', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=219906502, end_position_on_the_genomic_accession=219906565, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105375418 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105375418, CurrentID=0, Status='live', Symbol='LOC105375418', Aliases='', description='uncharacterized LOC105375418', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=98668220, end_position_on_the_genomic_accession=98671131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100130817 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130817, CurrentID=0, Status='live', Symbol='LOC100130817', Aliases='', description='uncharacterized LOC100130817', other_designations='', map_location='16q13', chromosome='16', genomic_nucleotide_accession_version='NC_000016.8', start_position_on_the_genomic_accession=55734034, end_position_on_the_genomic_accession=55735390, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100422961 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422961, CurrentID=0, Status='live', Symbol='MIR4315-2', Aliases='', description='microRNA 4315-2', other_designations='hsa-mir-4315-2', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=64822030, end_position_on_the_genomic_accession=64822102, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  130865 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130865, CurrentID=0, Status='live', Symbol='RPL26P13', Aliases='RPL26_6_221', description='ribosomal protein L26 pseudogene 13', other_designations='', map_location='2p16.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=60711451, end_position_on_the_genomic_accession=60711934, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422963 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422963, CurrentID=0, Status='live', Symbol='MIR4290', Aliases='', description='microRNA 4290', other_designations='hsa-mir-4290', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=90023441, end_position_on_the_genomic_accession=90023535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  130872 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130872, CurrentID=0, Status='live', Symbol='AHSA2', Aliases='Hch1', description='AHA1, activator of heat shock 90kDa protein ATPase homolog 2 (yeast)', other_designations='', map_location='2p15', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=61177420, end_position_on_the_genomic_accession=61188573, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105372607 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372607, CurrentID=0, Status='live', Symbol='LOC105372607', Aliases='', description='uncharacterized LOC105372607', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  130888 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130888, CurrentID=0, Status='live', Symbol='FBXO36', Aliases='Fbx36', description='F-box protein 36', other_designations='', map_location='2q36.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=229922491, end_position_on_the_genomic_accession=230013109, orientation='plus', exon_count=5, OMIM=609105, no_hdr0=''),
  393046 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=393046, CurrentID=0, Status='live', Symbol='OR2A5', Aliases='OR2A11P, OR2A26, OR2A8, OR7-138, OR7-141', description='olfactory receptor family 2 subfamily A member 5', other_designations='olfactory receptor 2A26|olfactory receptor 2A8|olfactory receptor 7-138/7-141|olfactory receptor, family 2, subfamily A, member 11 pseudogene|olfactory receptor, family 2, subfamily A, member 26|olfactory receptor, family 2, subfamily A, member 5|olfactory receptor, family 2, subfamily A, member 8', map_location='7q35', chromosome='7', genomic_nucleotide_accession_version='NC_000007.14', start_position_on_the_genomic_accession=144050402, end_position_on_the_genomic_accession=144051337, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  130916 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130916, CurrentID=0, Status='live', Symbol='MTERF4', Aliases='MTERFD2', description='mitochondrial transcription termination factor 4', other_designations='MTERF domain containing 2|mTERF domain-containing protein 2', map_location='2q37.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=241074680, end_position_on_the_genomic_accession=241102359, orientation='minus', exon_count=9, OMIM=615393, no_hdr0=''),
  100422973 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422973, CurrentID=0, Status='live', Symbol='MIR3169', Aliases='', description='microRNA 3169', other_designations='hsa-mir-3169', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=61199798, end_position_on_the_genomic_accession=61199880, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  393076 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=393076, CurrentID=0, Status='live', Symbol='LOC393076', Aliases='', description='uncharacterized LOC393076', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=155897113, end_position_on_the_genomic_accession=156058961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  393078 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=393078, CurrentID=0, Status='live', Symbol='LOC393078', Aliases='', description='uncharacterized LOC393078', other_designations='', map_location='7q36.3', chromosome='7', genomic_nucleotide_accession_version='NC_000007.13', start_position_on_the_genomic_accession=157233688, end_position_on_the_genomic_accession=157263477, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  104355135 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355135, CurrentID=0, Status='live', Symbol='LINC00431', Aliases='', description='long intergenic non-protein coding RNA 431', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110966027, end_position_on_the_genomic_accession=110990602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  130940 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130940, CurrentID=0, Status='live', Symbol='CCDC148', Aliases='', description='coiled-coil domain containing 148', other_designations='', map_location='2q24.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=158171356, end_position_on_the_genomic_accession=158456824, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  104355136 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355136, CurrentID=0, Status='live', Symbol='LINC01033', Aliases='', description='long intergenic non-protein coding RNA 1033', other_designations='', map_location='5q11.2', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=54320944, end_position_on_the_genomic_accession=54415125, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  393093 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=393093, CurrentID=0, Status='live', Symbol='MYP4', Aliases='', description='myopia 4 (high grade, autosomal dominant)', other_designations='', map_location='7q36', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  130951 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=130951, CurrentID=0, Status='live', Symbol='M1AP', Aliases='C2orf65, D6Mm5e, SPATA37', description='meiosis 1 associated protein', other_designations='meiosis 1 arresting protein|spermatogenesis associated 37', map_location='2p13.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=74557883, end_position_on_the_genomic_accession=74648852, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100270709 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100270709, CurrentID=0, Status='live', Symbol='FTH1P27', Aliases='', description='ferritin, heavy polypeptide 1 pseudogene 27', other_designations='ferritin, heavy polypeptide-like 17 pseudogene', map_location='Xp21.1', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=37505267, end_position_on_the_genomic_accession=37506026, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422978 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422978, CurrentID=0, Status='live', Symbol='MIR3185', Aliases='', description='microRNA 3185', other_designations='hsa-mir-3185', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=48724408, end_position_on_the_genomic_accession=48724475, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101929003 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929003, CurrentID=0, Status='live', Symbol='LOC101929003', Aliases='', description='uncharacterized LOC101929003', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000005.10', start_position_on_the_genomic_accession=1792577, end_position_on_the_genomic_accession=1798622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100422979 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422979, CurrentID=0, Status='live', Symbol='MIR4324', Aliases='mir-4324', description='microRNA 4324', other_designations='hsa-mir-4324', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000019.10', start_position_on_the_genomic_accession=49308797, end_position_on_the_genomic_accession=49308868, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104355140 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355140, CurrentID=0, Status='live', Symbol='LINC00375', Aliases='', description='long intergenic non-protein coding RNA 375', other_designations='', map_location='13q31.1', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=85065087, end_position_on_the_genomic_accession=85148057, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104355141 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355141, CurrentID=0, Status='live', Symbol='LINC01143', Aliases='', description='long intergenic non-protein coding RNA 1143', other_designations='', map_location='2p13.3', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=70887871, end_position_on_the_genomic_accession=70889984, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  104355143 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355143, CurrentID=0, Status='live', Symbol='TXNDC12-AS1', Aliases='', description='TXNDC12 antisense RNA 1', other_designations='', map_location='1p32.3', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=52050918, end_position_on_the_genomic_accession=52052683, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  401533 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=401533, CurrentID=0, Status='live', Symbol='DDX10P2', Aliases='', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 10 pseudogene 2', other_designations='', map_location='9q21.32', chromosome='9', genomic_nucleotide_accession_version='NC_000009.12', start_position_on_the_genomic_accession=82085641, end_position_on_the_genomic_accession=82088868, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  100422985 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422985, CurrentID=0, Status='live', Symbol='MIR1184-2', Aliases='', description='microRNA 1184-2', other_designations='hsa-mir-1184-2', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_000023.11', start_position_on_the_genomic_accession=155383100, end_position_on_the_genomic_accession=155383198, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104355146 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355146, CurrentID=0, Status='live', Symbol='LINC00396', Aliases='', description='long intergenic non-protein coding RNA 396', other_designations='', map_location='13q34', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=110053285, end_position_on_the_genomic_accession=110054952, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105372611 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=105372611, CurrentID=0, Status='live', Symbol='LOC105372611', Aliases='', description='uncharacterized LOC105372611', other_designations='', map_location='', chromosome='20', genomic_nucleotide_accession_version='NC_000020.11', start_position_on_the_genomic_accession=38109439, end_position_on_the_genomic_accession=38112432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104355147 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355147, CurrentID=0, Status='live', Symbol='LINC00428', Aliases='', description='long intergenic non-protein coding RNA 428', other_designations='', map_location='13q14.11', chromosome='13', genomic_nucleotide_accession_version='NC_000013.11', start_position_on_the_genomic_accession=42842412, end_position_on_the_genomic_accession=42872744, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104355148 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355148, CurrentID=0, Status='live', Symbol='CYP4A22-AS1', Aliases='ncRNA-a3', description='CYP4A22 antisense RNA 1', other_designations='CYP4A22 antisense RNA 1 (non-protein coding)', map_location='1p33', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  101929005 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=101929005, CurrentID=0, Status='live', Symbol='LOC101929005', Aliases='', description='uncharacterized LOC101929005', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NC_000012.12', start_position_on_the_genomic_accession=101747248, end_position_on_the_genomic_accession=101748944, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  283981 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283981, CurrentID=0, Status='live', Symbol='LINC00685', Aliases='CXYorf10, NCRNA00107, PPP2R3B-AS1', description='long intergenic non-protein coding RNA 685', other_designations='OTTHUMT00000055574|PPP2R3B antisense RNA 1 (non-protein coding)', map_location='Xp22.33; Yp11.3', chromosome='Y', genomic_nucleotide_accession_version='NC_000024.10', start_position_on_the_genomic_accession=319145, end_position_on_the_genomic_accession=321332, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  283982 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=283982, CurrentID=0, Status='live', Symbol='LINC00469', Aliases='C17orf54', description='long intergenic non-protein coding RNA 469', other_designations='', map_location='17q25.1', chromosome='17', genomic_nucleotide_accession_version='NC_000017.11', start_position_on_the_genomic_accession=73749270, end_position_on_the_genomic_accession=73828537, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100130823 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100130823, CurrentID=0, Status='live', Symbol='RPS17P9', Aliases='RPS17_3_404', description='ribosomal protein S17 pseudogene 9', other_designations='', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=129985092, end_position_on_the_genomic_accession=129998936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131034 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131034, CurrentID=0, Status='live', Symbol='CPNE4', Aliases='COPN4, CPN4', description='copine IV', other_designations='copine 8', map_location='3q22.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=131533560, end_position_on_the_genomic_accession=132285547, orientation='minus', exon_count=25, OMIM=604208, no_hdr0=''),
  104355151 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355151, CurrentID=0, Status='live', Symbol='PWRN4', Aliases='LINC01084', description='Prader-Willi region non-protein coding RNA 4', other_designations='long intergenic non-protein coding RNA 1084', map_location='15q11.2', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104355152 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=104355152, CurrentID=0, Status='live', Symbol='LINC01090', Aliases='', description='long intergenic non-protein coding RNA 1090', other_designations='', map_location='2q32.1', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=188035596, end_position_on_the_genomic_accession=188287691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100422993 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422993, CurrentID=0, Status='live', Symbol='MIR3130-1', Aliases='MIR3130-3, mir-3130-1', description='microRNA 3130-1', other_designations='hsa-mir-3130-1|hsa-mir-3130-3|microRNA 3130-3', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000002.12', start_position_on_the_genomic_accession=206783234, end_position_on_the_genomic_accession=206783308, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  131054 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131054, CurrentID=0, Status='live', Symbol='LOC131054', Aliases='', description='ralA binding protein 1 pseudogene', other_designations='', map_location='3q26.33', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=180323110, end_position_on_the_genomic_accession=180325061, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  131055 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=131055, CurrentID=0, Status='live', Symbol='LOC131055', Aliases='', description='peptidylprolyl isomerase A (cyclophilin A) pseudogene', other_designations='', map_location='3q26.1', chromosome='3', genomic_nucleotide_accession_version='NC_000003.12', start_position_on_the_genomic_accession=167065789, end_position_on_the_genomic_accession=167066542, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100422995 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422995, CurrentID=0, Status='live', Symbol='MIR3175', Aliases='mir-3175', description='microRNA 3175', other_designations='hsa-mir-3175', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000015.10', start_position_on_the_genomic_accession=92904399, end_position_on_the_genomic_accession=92904475, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100422997 : NtData(tax_id=9606, Org_name='Homo sapiens', GeneID=100422997, CurrentID=0, Status='live', Symbol='MIR4257', Aliases='', description='microRNA 4257', other_designations='hsa-mir-4257', map_location='', chromosome='1', genomic_nucleotide_accession_version='NC_000001.11', start_position_on_the_genomic_accession=150551929, end_position_on_the_genomic_accession=150552014, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
}

